

-- MySQL dump 10.13  Distrib 5.5.47, for debian-linux-gnu (x86_64)
--
-- Host: localhost    Database: dla_tutorial
-- ------------------------------------------------------
-- Server version	5.5.47-0ubuntu0.12.04.1

CREATE DATABASE IF NOT EXISTS `dla_tutorial`;
USE `dla_tutorial`;

/*!40101 SET @OLD_CHARACTER_SET_CLIENT=@@CHARACTER_SET_CLIENT */;
/*!40101 SET @OLD_CHARACTER_SET_RESULTS=@@CHARACTER_SET_RESULTS */;
/*!40101 SET @OLD_COLLATION_CONNECTION=@@COLLATION_CONNECTION */;
/*!40101 SET NAMES utf8 */;
/*!40103 SET @OLD_TIME_ZONE=@@TIME_ZONE */;
/*!40103 SET TIME_ZONE='+00:00' */;
/*!40014 SET @OLD_UNIQUE_CHECKS=@@UNIQUE_CHECKS, UNIQUE_CHECKS=0 */;
/*!40014 SET @OLD_FOREIGN_KEY_CHECKS=@@FOREIGN_KEY_CHECKS, FOREIGN_KEY_CHECKS=0 */;
/*!40101 SET @OLD_SQL_MODE=@@SQL_MODE, SQL_MODE='NO_AUTO_VALUE_ON_ZERO' */;
/*!40111 SET @OLD_SQL_NOTES=@@SQL_NOTES, SQL_NOTES=0 */;

--
-- Table structure for table `blog_outcomes`
--

DROP TABLE IF EXISTS `blog_outcomes`;
/*!40101 SET @saved_cs_client     = @@character_set_client */;
/*!40101 SET character_set_client = utf8 */;
CREATE TABLE `blog_outcomes` (
  `user_id` int(11) NOT NULL,
  `gender` int(2) DEFAULT NULL,
  `age` int(3) unsigned DEFAULT NULL,
  `occu` varchar(32) DEFAULT NULL,
  `sign` varchar(16) DEFAULT NULL,
  `is_indunk` int(1) DEFAULT NULL,
  `is_student` int(1) DEFAULT NULL,
  `is_education` int(1) DEFAULT NULL,
  `is_technology` int(1) DEFAULT NULL,
  PRIMARY KEY (`user_id`)
) ENGINE=MyISAM DEFAULT CHARSET=latin1;
/*!40101 SET character_set_client = @saved_cs_client */;

--
-- Dumping data for table `blog_outcomes`
--

LOCK TABLES `blog_outcomes` WRITE;
/*!40000 ALTER TABLE `blog_outcomes` DISABLE KEYS */;
INSERT INTO `blog_outcomes` VALUES (3991108,1,17,'indUnk','Leo',1,NULL,NULL,NULL),(3417138,1,25,'Communications-Media','Taurus',0,0,NULL,NULL),(3673414,0,14,'Student','Scorpio',0,1,NULL,NULL),(3361075,1,16,'Student','Capricorn',0,1,NULL,NULL),(4115327,1,14,'indUnk','Libra',1,NULL,NULL,NULL),(3549541,1,16,'Student','Gemini',0,1,NULL,NULL),(3750679,1,46,'indUnk','Libra',1,NULL,NULL,NULL),(3600011,0,43,'Technology','Cancer',0,0,NULL,NULL),(3469665,1,23,'Military','Taurus',0,0,NULL,NULL),(4074230,0,35,'Technology','Pisces',0,0,NULL,NULL),(715237,0,26,'indUnk','Virgo',1,NULL,NULL,NULL),(3295265,0,23,'Consulting','Gemini',0,0,NULL,NULL),(791840,0,38,'Student','Aries',0,1,NULL,NULL),(3544826,1,37,'Fashion','Sagittarius',0,0,NULL,NULL),(3711805,0,23,'indUnk','Aries',1,NULL,NULL,NULL),(4216216,1,39,'indUnk','Leo',1,NULL,NULL,NULL),(3991209,1,14,'Student','Pisces',0,1,NULL,NULL),(1077954,0,26,'indUnk','Gemini',1,NULL,NULL,NULL),(3358293,1,26,'indUnk','Scorpio',1,NULL,NULL,NULL),(4190982,0,17,'Engineering','Sagittarius',0,0,NULL,NULL),(3671580,0,24,'Student','Leo',0,1,NULL,NULL),(3155970,1,24,'Science','Sagittarius',0,0,NULL,NULL),(1636406,0,23,'Student','Scorpio',0,1,NULL,NULL),(2935790,0,16,'Student','Taurus',0,1,NULL,NULL),(4263309,1,48,'Fashion','Pisces',0,0,NULL,NULL),(4220141,1,15,'indUnk','Leo',1,NULL,NULL,NULL),(2300555,1,24,'indUnk','Libra',1,NULL,NULL,NULL),(3498277,1,13,'indUnk','Pisces',1,NULL,NULL,NULL),(2678611,0,16,'Student','Capricorn',0,1,NULL,NULL),(1234212,0,27,'indUnk','Taurus',1,NULL,NULL,NULL),(3808913,1,25,'Arts','Capricorn',0,0,NULL,NULL),(3983686,1,27,'Non-Profit','Capricorn',0,0,NULL,NULL),(3838040,0,17,'Engineering','Cancer',0,0,NULL,NULL),(4142807,1,17,'Student','Capricorn',0,1,NULL,NULL),(4075308,1,15,'Student','Scorpio',0,1,NULL,NULL),(3421075,0,23,'Technology','Cancer',0,0,NULL,NULL),(4045197,1,17,'indUnk','Gemini',1,NULL,NULL,NULL),(3913126,0,26,'Communications-Media','Taurus',0,0,NULL,NULL),(3396638,1,47,'indUnk','Pisces',1,NULL,NULL,NULL),(1305545,0,34,'indUnk','Aquarius',1,NULL,NULL,NULL),(3560320,0,46,'Manufacturing','Libra',0,0,NULL,NULL),(3753830,1,13,'Student','Capricorn',0,1,NULL,NULL),(3659117,1,36,'Marketing','Cancer',0,0,NULL,NULL),(3558382,1,40,'indUnk','Aquarius',1,NULL,NULL,NULL),(3779062,1,24,'indUnk','Leo',1,NULL,NULL,NULL),(3734069,1,24,'indUnk','Scorpio',1,NULL,NULL,NULL),(3646978,1,35,'indUnk','Virgo',1,NULL,NULL,NULL),(3307733,1,26,'Manufacturing','Sagittarius',0,0,NULL,NULL),(3146636,1,48,'indUnk','Pisces',1,NULL,NULL,NULL),(3785626,1,15,'Student','Taurus',0,1,NULL,NULL),(4031020,1,27,'Marketing','Cancer',0,0,NULL,NULL),(4298922,1,27,'Marketing','Virgo',0,0,NULL,NULL),(1969237,0,43,'Sports-Recreation','Cancer',0,0,NULL,NULL),(4128156,0,47,'Consulting','Gemini',0,0,NULL,NULL),(3418455,0,26,'Student','Gemini',0,1,NULL,NULL),(4166693,1,24,'Education','Cancer',0,0,NULL,NULL),(3415194,1,45,'Government','Sagittarius',0,0,NULL,NULL),(1232458,1,38,'indUnk','Cancer',1,NULL,NULL,NULL),(3782048,0,16,'indUnk','Aquarius',1,NULL,NULL,NULL),(3694314,1,16,'indUnk','Cancer',1,NULL,NULL,NULL),(3823631,1,37,'Marketing','Cancer',0,0,NULL,NULL),(3374253,0,26,'Telecommunications','Libra',0,0,NULL,NULL),(3509206,1,17,'Student','Cancer',0,1,NULL,NULL),(3596192,0,40,'Technology','Scorpio',0,0,NULL,NULL),(3816397,1,42,'Technology','Virgo',0,0,NULL,NULL),(2680773,0,24,'Communications-Media','Gemini',0,0,NULL,NULL),(4040811,1,23,'indUnk','Cancer',1,NULL,NULL,NULL),(1401816,0,40,'indUnk','Pisces',1,NULL,NULL,NULL),(3155540,1,24,'Communications-Media','Sagittarius',0,0,NULL,NULL),(4248100,0,14,'Student','Aquarius',0,1,NULL,NULL),(3483901,1,23,'indUnk','Taurus',1,NULL,NULL,NULL),(3634853,0,44,'Technology','Virgo',0,0,NULL,NULL),(4184832,1,36,'indUnk','Cancer',1,NULL,NULL,NULL),(3038828,1,24,'indUnk','Aries',1,NULL,NULL,NULL),(1960271,0,35,'Internet','Aries',0,0,NULL,NULL),(2224636,1,17,'indUnk','Aries',1,NULL,NULL,NULL),(3306195,0,33,'indUnk','Leo',1,NULL,NULL,NULL),(3972268,0,17,'Science','Leo',0,0,NULL,NULL),(3679805,1,15,'Student','Virgo',0,1,NULL,NULL),(3663542,0,23,'Student','Cancer',0,1,NULL,NULL),(3471824,1,16,'Student','Scorpio',0,1,NULL,NULL),(3921381,1,27,'Technology','Cancer',0,0,NULL,NULL),(3599631,0,24,'Technology','Gemini',0,0,NULL,NULL),(4096204,1,17,'indUnk','Scorpio',1,NULL,NULL,NULL),(4056798,1,14,'Student','Virgo',0,1,NULL,NULL),(3539593,0,23,'Technology','Cancer',0,0,NULL,NULL),(3463885,0,42,'indUnk','Gemini',1,NULL,NULL,NULL),(4149935,0,15,'Student','Capricorn',0,1,NULL,NULL),(3856641,0,27,'Arts','Aries',0,0,NULL,NULL),(4308812,1,24,'Communications-Media','Aries',0,0,NULL,NULL),(4162840,0,17,'indUnk','Virgo',1,NULL,NULL,NULL),(4210644,1,14,'Student','Leo',0,1,NULL,NULL),(1785539,0,24,'Science','Scorpio',0,0,NULL,NULL),(3877108,1,27,'Student','Pisces',0,1,NULL,NULL),(4121917,1,15,'Student','Aries',0,1,NULL,NULL),(3749228,0,13,'indUnk','Taurus',1,NULL,NULL,NULL),(3130162,0,26,'Banking','Virgo',0,0,NULL,NULL),(4023489,1,25,'indUnk','Taurus',1,NULL,NULL,NULL),(3784161,1,17,'Student','Leo',0,1,NULL,NULL),(2880960,0,34,'Communications-Media','Pisces',0,0,NULL,NULL),(3759734,1,27,'Student','Leo',0,1,NULL,NULL),(3924062,0,17,'Non-Profit','Capricorn',0,0,NULL,NULL),(2311511,0,13,'indUnk','Taurus',1,NULL,NULL,NULL),(1808929,0,23,'Student','Leo',0,1,NULL,NULL),(4310425,1,27,'indUnk','Aries',1,NULL,NULL,NULL),(3986257,1,15,'Student','Gemini',0,1,NULL,NULL),(3992688,0,27,'Government','Virgo',0,0,NULL,NULL),(1807720,1,16,'Student','Leo',0,1,NULL,NULL),(3840157,1,14,'Student','Scorpio',0,1,NULL,NULL),(3125462,0,15,'Student','Taurus',0,1,NULL,NULL),(2607577,1,27,'Religion','Aquarius',0,0,NULL,NULL),(2967519,1,17,'Student','Gemini',0,1,NULL,NULL),(1298865,1,39,'indUnk','Aquarius',1,NULL,NULL,NULL),(4132315,0,15,'indUnk','Aries',1,NULL,NULL,NULL),(3889796,0,14,'indUnk','Pisces',1,NULL,NULL,NULL),(3532367,0,33,'Education','Aries',0,0,NULL,NULL),(1733718,1,24,'Student','Virgo',0,1,NULL,NULL),(4022192,1,17,'Student','Gemini',0,1,NULL,NULL),(3069909,1,35,'HumanResources','Virgo',0,0,NULL,NULL),(3595127,1,23,'Communications-Media','Cancer',0,0,NULL,NULL),(3861317,1,25,'Government','Aries',0,0,NULL,NULL),(3487773,1,24,'indUnk','Gemini',1,NULL,NULL,NULL),(3170057,1,26,'Communications-Media','Gemini',0,0,NULL,NULL),(3590306,1,23,'Accounting','Scorpio',0,0,NULL,NULL),(3339644,0,26,'Internet','Capricorn',0,0,NULL,NULL),(3533397,1,17,'Internet','Virgo',0,0,NULL,NULL),(3879020,0,24,'indUnk','Sagittarius',1,NULL,NULL,NULL),(4129852,0,17,'Student','Aries',0,1,NULL,NULL),(2831727,1,25,'indUnk','Aries',1,NULL,NULL,NULL),(4163223,0,23,'Technology','Pisces',0,0,NULL,NULL),(3463044,1,37,'indUnk','Pisces',1,NULL,NULL,NULL),(3476799,0,24,'indUnk','Leo',1,NULL,NULL,NULL),(3191071,1,14,'indUnk','Gemini',1,NULL,NULL,NULL),(1359631,0,26,'Internet','Cancer',0,0,NULL,NULL),(4259479,0,17,'RealEstate','Libra',0,0,NULL,NULL),(3905963,1,23,'Consulting','Taurus',0,0,NULL,NULL),(3911122,1,13,'indUnk','Taurus',1,NULL,NULL,NULL),(671748,0,27,'Communications-Media','Aries',0,0,NULL,NULL),(3869854,0,33,'Marketing','Sagittarius',0,0,NULL,NULL),(4077627,0,44,'Telecommunications','Scorpio',0,0,NULL,NULL),(3839683,1,23,'Student','Pisces',0,1,NULL,NULL),(3938699,0,17,'Student','Taurus',0,1,NULL,NULL),(3388444,0,24,'Communications-Media','Leo',0,0,NULL,NULL),(4248403,0,27,'indUnk','Cancer',1,NULL,NULL,NULL),(3893150,0,16,'Student','Virgo',0,1,NULL,NULL),(3206777,0,24,'Education','Libra',0,0,NULL,NULL),(3821963,1,15,'Student','Leo',0,1,NULL,NULL),(3735544,1,13,'Student','Libra',0,1,NULL,NULL),(1007188,0,48,'Religion','Libra',0,0,NULL,NULL),(3936941,0,26,'Consulting','Aquarius',0,0,NULL,NULL),(4067087,1,39,'Education','Leo',0,0,NULL,NULL),(3320022,0,26,'Telecommunications','Gemini',0,0,NULL,NULL),(3311843,0,41,'indUnk','Libra',1,NULL,NULL,NULL),(3466740,1,23,'indUnk','Virgo',1,NULL,NULL,NULL),(4044445,1,40,'indUnk','Virgo',1,NULL,NULL,NULL),(3677153,0,15,'indUnk','Leo',1,NULL,NULL,NULL),(3485630,1,15,'Arts','Aries',0,0,NULL,NULL),(3624333,1,37,'indUnk','Gemini',1,NULL,NULL,NULL),(1788355,0,24,'indUnk','Taurus',1,NULL,NULL,NULL),(3811749,1,26,'indUnk','Virgo',1,NULL,NULL,NULL),(3851968,1,14,'indUnk','Cancer',1,NULL,NULL,NULL),(3417999,0,33,'Non-Profit','Virgo',0,0,NULL,NULL),(3853532,1,35,'HumanResources','Aquarius',0,0,NULL,NULL),(3877932,0,23,'Student','Aquarius',0,1,NULL,NULL),(1729882,0,48,'Education','Taurus',0,0,NULL,NULL),(4015298,1,42,'indUnk','Taurus',1,NULL,NULL,NULL),(4178914,0,17,'indUnk','Aquarius',1,NULL,NULL,NULL),(4008193,1,17,'Student','Gemini',0,1,NULL,NULL),(3554687,0,15,'indUnk','Taurus',1,NULL,NULL,NULL),(1261679,1,16,'indUnk','Gemini',1,NULL,NULL,NULL),(4280767,1,16,'Student','Gemini',0,1,NULL,NULL),(4115407,1,14,'Student','Capricorn',0,1,NULL,NULL),(3186371,0,27,'Law','Taurus',0,0,NULL,NULL),(3767607,0,40,'HumanResources','Capricorn',0,0,NULL,NULL),(3607694,1,23,'indUnk','Leo',1,NULL,NULL,NULL),(2862007,1,23,'indUnk','Capricorn',1,NULL,NULL,NULL),(4213470,0,13,'Student','Libra',0,1,NULL,NULL),(4102045,1,17,'indUnk','Gemini',1,NULL,NULL,NULL),(4264876,1,15,'indUnk','Capricorn',1,NULL,NULL,NULL),(4075835,1,27,'Construction','Scorpio',0,0,NULL,NULL),(4052552,0,16,'indUnk','Gemini',1,NULL,NULL,NULL),(3970154,1,23,'Student','Pisces',0,1,NULL,NULL),(4173841,1,43,'indUnk','Gemini',1,NULL,NULL,NULL),(3291131,0,25,'indUnk','Taurus',1,NULL,NULL,NULL),(3413707,0,15,'Chemicals','Taurus',0,0,NULL,NULL),(2813430,0,33,'indUnk','Taurus',1,NULL,NULL,NULL),(3883177,0,35,'Technology','Scorpio',0,0,NULL,NULL),(1629718,0,27,'Technology','Libra',0,0,NULL,NULL),(3672170,0,36,'Internet','Scorpio',0,0,NULL,NULL),(734023,1,14,'Student','Cancer',0,1,NULL,NULL),(3860353,1,24,'indUnk','Gemini',1,NULL,NULL,NULL),(3935450,1,25,'Arts','Aries',0,0,NULL,NULL),(4184856,0,14,'Non-Profit','Capricorn',0,0,NULL,NULL),(3558845,1,26,'indUnk','Pisces',1,NULL,NULL,NULL),(3306314,0,24,'Student','Taurus',0,1,NULL,NULL),(3394208,0,17,'Student','Capricorn',0,1,NULL,NULL),(4216577,1,38,'indUnk','Cancer',1,NULL,NULL,NULL),(2568090,1,24,'indUnk','Scorpio',1,NULL,NULL,NULL),(3823807,1,14,'indUnk','Virgo',1,NULL,NULL,NULL),(4235996,0,15,'Student','Taurus',0,1,NULL,NULL),(3367343,1,17,'indUnk','Capricorn',1,NULL,NULL,NULL),(4169661,1,41,'HumanResources','Aquarius',0,0,NULL,NULL),(3482840,1,13,'Student','Leo',0,1,NULL,NULL),(3460713,1,17,'Arts','Leo',0,0,NULL,NULL),(3674747,0,16,'Student','Aries',0,1,NULL,NULL),(3699086,0,16,'indUnk','Pisces',1,NULL,NULL,NULL),(4152098,0,26,'indUnk','Scorpio',1,NULL,NULL,NULL),(3443396,1,16,'Student','Gemini',0,1,NULL,NULL),(3371255,0,15,'Student','Pisces',0,1,NULL,NULL),(1980443,0,16,'Student','Virgo',0,1,NULL,NULL),(3326628,1,23,'indUnk','Capricorn',1,NULL,NULL,NULL),(3215937,1,26,'Accounting','Cancer',0,0,NULL,NULL),(3838218,0,13,'Student','Virgo',0,1,NULL,NULL),(3287144,0,16,'Student','Sagittarius',0,1,NULL,NULL),(828778,0,23,'Sports-Recreation','Sagittarius',0,0,NULL,NULL),(2818788,1,16,'Student','Sagittarius',0,1,NULL,NULL),(3072650,0,14,'indUnk','Sagittarius',1,NULL,NULL,NULL),(3712913,0,15,'Student','Virgo',0,1,NULL,NULL),(3723692,0,25,'Technology','Virgo',0,0,NULL,NULL),(4129418,1,34,'indUnk','Virgo',1,NULL,NULL,NULL),(3802286,0,23,'indUnk','Capricorn',1,NULL,NULL,NULL),(3888166,1,13,'Military','Aries',0,0,NULL,NULL),(3549714,1,24,'Education','Virgo',0,0,NULL,NULL),(3806796,0,27,'Arts','Pisces',0,0,NULL,NULL),(3764220,1,14,'Student','Virgo',0,1,NULL,NULL),(3950869,1,17,'Student','Virgo',0,1,NULL,NULL),(3321866,0,15,'Student','Sagittarius',0,1,NULL,NULL),(3785845,1,15,'Arts','Aries',0,0,NULL,NULL),(3478510,0,16,'Student','Aries',0,1,NULL,NULL),(4320505,1,33,'indUnk','Sagittarius',1,NULL,NULL,NULL),(3068590,0,45,'Technology','Virgo',0,0,NULL,NULL),(3785275,1,26,'indUnk','Pisces',1,NULL,NULL,NULL),(3407290,0,37,'Engineering','Leo',0,0,NULL,NULL),(4188274,1,16,'Student','Gemini',0,1,NULL,NULL),(3505165,0,14,'Student','Cancer',0,1,NULL,NULL),(3759312,0,24,'Arts','Cancer',0,0,NULL,NULL),(4287218,1,26,'indUnk','Gemini',1,NULL,NULL,NULL),(2977760,1,14,'indUnk','Libra',1,NULL,NULL,NULL),(3866526,0,26,'indUnk','Aquarius',1,NULL,NULL,NULL),(4116584,0,16,'Museums-Libraries','Cancer',0,0,NULL,NULL),(3992573,1,26,'indUnk','Aquarius',1,NULL,NULL,NULL),(3446654,0,25,'Telecommunications','Cancer',0,0,NULL,NULL),(3659651,0,17,'Student','Aquarius',0,1,NULL,NULL),(3451166,0,17,'indUnk','Aquarius',1,NULL,NULL,NULL),(4296269,1,27,'Publishing','Capricorn',0,0,NULL,NULL),(3831350,0,26,'Manufacturing','Capricorn',0,0,NULL,NULL),(3377942,1,24,'indUnk','Libra',1,NULL,NULL,NULL),(3983707,1,13,'Student','Scorpio',0,1,NULL,NULL),(3432308,1,25,'Education','Cancer',0,0,NULL,NULL),(2539406,0,23,'Government','Cancer',0,0,NULL,NULL),(3416335,0,16,'indUnk','Capricorn',1,NULL,NULL,NULL),(3354517,1,16,'Student','Aries',0,1,NULL,NULL),(3819668,1,25,'Marketing','Pisces',0,0,NULL,NULL),(4188004,1,17,'Student','Taurus',0,1,NULL,NULL),(669116,1,26,'indUnk','Gemini',1,NULL,NULL,NULL),(2753052,1,14,'Student','Scorpio',0,1,NULL,NULL),(4128797,0,27,'Technology','Sagittarius',0,0,NULL,NULL),(4189257,0,25,'Religion','Pisces',0,0,NULL,NULL),(3384726,0,17,'indUnk','Aquarius',1,NULL,NULL,NULL),(882528,0,17,'indUnk','Gemini',1,NULL,NULL,NULL),(4325889,0,14,'Student','Scorpio',0,1,NULL,NULL),(3768397,1,23,'indUnk','Capricorn',1,NULL,NULL,NULL),(3000240,0,17,'Student','Taurus',0,1,NULL,NULL),(4238152,0,14,'Student','Capricorn',0,1,NULL,NULL),(3490316,1,24,'LawEnforcement-Security','Aries',0,0,NULL,NULL),(2428393,1,17,'indUnk','Aquarius',1,NULL,NULL,NULL),(3678942,1,15,'Student','Cancer',0,1,NULL,NULL),(3729846,0,17,'Student','Sagittarius',0,1,NULL,NULL),(4210703,0,26,'indUnk','Capricorn',1,NULL,NULL,NULL),(3739942,1,15,'indUnk','Libra',1,NULL,NULL,NULL),(1778579,0,23,'indUnk','Gemini',1,NULL,NULL,NULL),(3670354,0,17,'indUnk','Leo',1,NULL,NULL,NULL),(3693380,1,14,'Fashion','Capricorn',0,0,NULL,NULL),(3522772,1,27,'Government','Leo',0,0,NULL,NULL),(3572381,0,40,'indUnk','Virgo',1,NULL,NULL,NULL),(3293519,0,25,'Consulting','Leo',0,0,NULL,NULL),(3646377,0,25,'Sports-Recreation','Cancer',0,0,NULL,NULL),(4122943,1,14,'Student','Leo',0,1,NULL,NULL),(4220994,1,25,'indUnk','Virgo',1,NULL,NULL,NULL),(4075064,0,23,'indUnk','Sagittarius',1,NULL,NULL,NULL),(1902121,1,27,'Engineering','Taurus',0,0,NULL,NULL),(1987893,1,38,'indUnk','Capricorn',1,NULL,NULL,NULL),(3527113,0,26,'Banking','Gemini',0,0,NULL,NULL),(2876677,0,17,'Student','Aries',0,1,NULL,NULL),(4238006,0,16,'indUnk','Sagittarius',1,NULL,NULL,NULL),(3505043,1,26,'indUnk','Pisces',1,NULL,NULL,NULL),(3993797,1,16,'Student','Libra',0,1,NULL,NULL),(3848892,0,16,'indUnk','Libra',1,NULL,NULL,NULL),(3599769,1,23,'Student','Libra',0,1,NULL,NULL),(3304415,0,25,'indUnk','Virgo',1,NULL,NULL,NULL),(1111991,1,33,'indUnk','Taurus',1,NULL,NULL,NULL),(3348160,0,24,'Government','Leo',0,0,NULL,NULL),(3356970,0,41,'Science','Gemini',0,0,NULL,NULL),(4038842,0,26,'indUnk','Libra',1,NULL,NULL,NULL),(3779963,1,16,'Arts','Virgo',0,0,NULL,NULL),(2726395,1,41,'Technology','Leo',0,0,NULL,NULL),(4055499,1,24,'Student','Aries',0,1,NULL,NULL),(3454992,0,33,'Technology','Libra',0,0,NULL,NULL),(3962812,0,17,'indUnk','Scorpio',1,NULL,NULL,NULL),(4060980,0,16,'indUnk','Libra',1,NULL,NULL,NULL),(3829312,0,27,'indUnk','Gemini',1,NULL,NULL,NULL),(3900039,1,17,'Student','Leo',0,1,NULL,NULL),(3375539,0,17,'Student','Aries',0,1,NULL,NULL),(3392887,1,24,'Science','Libra',0,0,NULL,NULL),(3736053,0,24,'Arts','Aries',0,0,NULL,NULL),(4166300,0,17,'Student','Taurus',0,1,NULL,NULL),(3632734,0,26,'Internet','Capricorn',0,0,NULL,NULL),(3987879,0,34,'indUnk','Scorpio',1,NULL,NULL,NULL),(3114279,0,24,'Student','Taurus',0,1,NULL,NULL),(3592700,1,25,'indUnk','Capricorn',1,NULL,NULL,NULL),(4060812,1,17,'indUnk','Virgo',1,NULL,NULL,NULL),(3831024,0,23,'Student','Pisces',0,1,NULL,NULL),(3699072,1,23,'indUnk','Sagittarius',1,NULL,NULL,NULL),(1503551,0,27,'Education','Sagittarius',0,0,NULL,NULL),(2246318,0,25,'Telecommunications','Libra',0,0,NULL,NULL),(3969316,1,33,'Internet','Aries',0,0,NULL,NULL),(4201404,0,27,'indUnk','Cancer',1,NULL,NULL,NULL),(3682262,1,15,'indUnk','Gemini',1,NULL,NULL,NULL),(1240274,0,23,'Communications-Media','Gemini',0,0,NULL,NULL),(3403051,0,26,'Telecommunications','Leo',0,0,NULL,NULL),(4151959,1,14,'indUnk','Gemini',1,NULL,NULL,NULL),(3722833,0,15,'Student','Libra',0,1,NULL,NULL),(3561749,1,15,'Education','Capricorn',0,0,NULL,NULL),(4224299,1,17,'Fashion','Scorpio',0,0,NULL,NULL),(3594873,1,16,'Student','Scorpio',0,1,NULL,NULL),(3448540,1,17,'indUnk','Virgo',1,NULL,NULL,NULL),(3663701,1,27,'Communications-Media','Cancer',0,0,NULL,NULL),(3466592,0,25,'indUnk','Pisces',1,NULL,NULL,NULL),(4312209,1,16,'indUnk','Virgo',1,NULL,NULL,NULL),(3356583,0,15,'Student','Leo',0,1,NULL,NULL),(4252081,0,26,'Arts','Libra',0,0,NULL,NULL),(2166338,0,17,'Student','Libra',0,1,NULL,NULL),(3386981,0,27,'indUnk','Cancer',1,NULL,NULL,NULL),(3400445,1,24,'Advertising','Leo',0,0,NULL,NULL),(3522747,0,36,'indUnk','Taurus',1,NULL,NULL,NULL),(4158987,1,13,'Student','Libra',0,1,NULL,NULL),(3984120,0,26,'Advertising','Sagittarius',0,0,NULL,NULL),(3659024,0,34,'Technology','Aquarius',0,0,NULL,NULL),(3757297,1,27,'Student','Pisces',0,1,NULL,NULL),(3829063,0,23,'indUnk','Taurus',1,NULL,NULL,NULL),(3986767,0,40,'Internet','Sagittarius',0,0,NULL,NULL),(2308025,0,23,'Law','Libra',0,0,NULL,NULL),(3353979,0,15,'Student','Sagittarius',0,1,NULL,NULL),(4027311,0,39,'Technology','Sagittarius',0,0,NULL,NULL),(3475992,1,33,'indUnk','Libra',1,NULL,NULL,NULL),(3546623,1,42,'indUnk','Libra',1,NULL,NULL,NULL),(1488330,0,17,'Non-Profit','Pisces',0,0,NULL,NULL),(3601175,1,15,'indUnk','Gemini',1,NULL,NULL,NULL),(3802748,1,26,'indUnk','Scorpio',1,NULL,NULL,NULL),(3672882,0,16,'indUnk','Gemini',1,NULL,NULL,NULL),(4227656,1,23,'indUnk','Cancer',1,NULL,NULL,NULL),(4177430,0,16,'Student','Virgo',0,1,NULL,NULL),(3819051,0,14,'Student','Leo',0,1,NULL,NULL),(3460990,1,27,'Education','Aquarius',0,0,NULL,NULL),(4074186,0,25,'indUnk','Cancer',1,NULL,NULL,NULL),(4288065,1,40,'indUnk','Taurus',1,NULL,NULL,NULL),(4216034,0,37,'Technology','Taurus',0,0,NULL,NULL),(3873570,1,34,'Education','Capricorn',0,0,NULL,NULL),(216833,0,24,'Student','Scorpio',0,1,NULL,NULL),(3858126,1,26,'Education','Capricorn',0,0,NULL,NULL),(4151531,0,25,'indUnk','Virgo',1,NULL,NULL,NULL),(1758432,1,23,'Engineering','Taurus',0,0,NULL,NULL),(3667367,1,25,'Non-Profit','Leo',0,0,NULL,NULL),(517266,0,36,'Education','Virgo',0,0,NULL,NULL),(3984458,0,17,'Science','Sagittarius',0,0,NULL,NULL),(3605346,1,25,'Government','Taurus',0,0,NULL,NULL),(3343294,1,25,'indUnk','Cancer',1,NULL,NULL,NULL),(3360363,1,23,'Student','Capricorn',0,1,NULL,NULL),(2284476,0,17,'indUnk','Taurus',1,NULL,NULL,NULL),(1223561,0,23,'Engineering','Cancer',0,0,NULL,NULL),(832265,0,27,'Technology','Libra',0,0,NULL,NULL),(3997213,0,15,'Student','Aquarius',0,1,NULL,NULL),(3569294,1,43,'Education','Gemini',0,0,NULL,NULL),(2259900,0,26,'Marketing','Taurus',0,0,NULL,NULL),(2033616,1,17,'Student','Virgo',0,1,NULL,NULL),(28451,0,27,'Internet','Aquarius',0,0,NULL,NULL),(3878251,1,25,'Construction','Pisces',0,0,NULL,NULL),(3525639,1,24,'LawEnforcement-Security','Scorpio',0,0,NULL,NULL),(3735538,0,14,'Student','Scorpio',0,1,NULL,NULL),(3813860,1,24,'indUnk','Sagittarius',1,NULL,NULL,NULL),(3723398,0,25,'Technology','Leo',0,0,NULL,NULL),(4225684,0,17,'Student','Sagittarius',0,1,NULL,NULL),(4082385,0,26,'indUnk','Aquarius',1,NULL,NULL,NULL),(3197709,0,24,'indUnk','Taurus',1,NULL,NULL,NULL),(3022585,1,27,'Education','Aquarius',0,0,NULL,NULL),(4170329,1,23,'Education','Aries',0,0,NULL,NULL),(3596048,1,16,'indUnk','Leo',1,NULL,NULL,NULL),(3958372,1,33,'Communications-Media','Virgo',0,0,NULL,NULL),(4108638,0,25,'Student','Scorpio',0,1,NULL,NULL),(3644129,1,24,'Tourism','Virgo',0,0,NULL,NULL),(3921274,0,23,'Communications-Media','Cancer',0,0,NULL,NULL),(3999305,0,16,'Student','Aries',0,1,NULL,NULL),(3994464,1,36,'Consulting','Sagittarius',0,0,NULL,NULL),(3760558,0,36,'Manufacturing','Aquarius',0,0,NULL,NULL),(743739,0,24,'Student','Sagittarius',0,1,NULL,NULL),(3738552,0,23,'Telecommunications','Capricorn',0,0,NULL,NULL),(4271205,1,26,'indUnk','Pisces',1,NULL,NULL,NULL),(3690855,0,25,'Student','Leo',0,1,NULL,NULL),(3541469,0,13,'indUnk','Scorpio',1,NULL,NULL,NULL),(3295677,1,43,'Technology','Scorpio',0,0,NULL,NULL),(4108942,0,17,'indUnk','Sagittarius',1,NULL,NULL,NULL),(4325536,1,25,'indUnk','Leo',1,NULL,NULL,NULL),(3489768,1,13,'Student','Aries',0,1,NULL,NULL),(2882045,0,15,'Student','Sagittarius',0,1,NULL,NULL),(3850561,1,24,'indUnk','Scorpio',1,NULL,NULL,NULL),(3403602,0,15,'Education','Scorpio',0,0,NULL,NULL),(3354778,0,17,'Student','Pisces',0,1,NULL,NULL),(3989154,0,33,'Non-Profit','Aries',0,0,NULL,NULL),(3529951,1,16,'indUnk','Libra',1,NULL,NULL,NULL),(3409476,1,36,'Education','Aquarius',0,0,NULL,NULL),(3940997,1,23,'Student','Aries',0,1,NULL,NULL),(4052055,1,35,'indUnk','Leo',1,NULL,NULL,NULL),(4196591,1,15,'indUnk','Leo',1,NULL,NULL,NULL),(4075808,1,35,'indUnk','Virgo',1,NULL,NULL,NULL),(4274229,1,14,'Student','Scorpio',0,1,NULL,NULL),(4007735,0,17,'Student','Capricorn',0,1,NULL,NULL),(3771600,1,14,'Student','Virgo',0,1,NULL,NULL),(4202671,0,17,'Student','Gemini',0,1,NULL,NULL),(3966870,0,23,'Technology','Virgo',0,0,NULL,NULL),(2952444,0,13,'indUnk','Aries',1,NULL,NULL,NULL),(3737807,1,17,'indUnk','Virgo',1,NULL,NULL,NULL),(3577655,1,17,'indUnk','Leo',1,NULL,NULL,NULL),(3650974,0,23,'Student','Scorpio',0,1,NULL,NULL),(3496773,0,24,'indUnk','Taurus',1,NULL,NULL,NULL),(3485073,0,15,'Education','Virgo',0,0,NULL,NULL),(4068475,0,17,'indUnk','Gemini',1,NULL,NULL,NULL),(3441506,1,26,'indUnk','Pisces',1,NULL,NULL,NULL),(3680589,1,35,'Arts','Libra',0,0,NULL,NULL),(3886562,0,23,'Student','Taurus',0,1,NULL,NULL),(3753470,0,15,'Student','Pisces',0,1,NULL,NULL),(3618866,1,14,'indUnk','Virgo',1,NULL,NULL,NULL),(3380464,0,48,'Communications-Media','Libra',0,0,NULL,NULL),(3411285,1,26,'indUnk','Libra',1,NULL,NULL,NULL),(3668030,1,23,'Student','Cancer',0,1,NULL,NULL),(4236023,0,17,'Student','Taurus',0,1,NULL,NULL),(3501362,0,25,'indUnk','Aquarius',1,NULL,NULL,NULL),(4024747,1,24,'indUnk','Aquarius',1,NULL,NULL,NULL),(4109605,1,15,'Student','Leo',0,1,NULL,NULL),(1668565,0,40,'indUnk','Aries',1,NULL,NULL,NULL),(3400091,1,23,'Law','Aquarius',0,0,NULL,NULL),(3587934,1,25,'Education','Aries',0,0,NULL,NULL),(4027719,1,23,'Biotech','Aries',0,0,NULL,NULL),(3434961,0,24,'indUnk','Pisces',1,NULL,NULL,NULL),(3573678,1,24,'Technology','Sagittarius',0,0,NULL,NULL),(4277394,0,13,'Student','Scorpio',0,1,NULL,NULL),(3576839,0,27,'Military','Virgo',0,0,NULL,NULL),(3196705,0,24,'indUnk','Scorpio',1,NULL,NULL,NULL),(2747575,0,17,'Student','Taurus',0,1,NULL,NULL),(3887824,1,16,'Student','Pisces',0,1,NULL,NULL),(1143104,0,23,'Student','Gemini',0,1,NULL,NULL),(3364899,0,17,'Student','Sagittarius',0,1,NULL,NULL),(3509141,1,16,'Student','Cancer',0,1,NULL,NULL),(3551359,0,23,'Communications-Media','Cancer',0,0,NULL,NULL),(1270372,1,33,'Manufacturing','Aries',0,0,NULL,NULL),(1197086,0,16,'Student','Aries',0,1,NULL,NULL),(4242799,1,26,'Transportation','Virgo',0,0,NULL,NULL),(3828451,1,16,'Student','Leo',0,1,NULL,NULL),(3771668,0,25,'Automotive','Libra',0,0,NULL,NULL),(3572119,1,17,'Arts','Sagittarius',0,0,NULL,NULL),(3719233,0,37,'indUnk','Pisces',1,NULL,NULL,NULL),(3835889,0,17,'Student','Cancer',0,1,NULL,NULL),(3851365,1,45,'Education','Cancer',0,0,NULL,NULL),(3324571,1,36,'indUnk','Sagittarius',1,NULL,NULL,NULL),(3139191,0,16,'indUnk','Aquarius',1,NULL,NULL,NULL),(3961625,0,14,'Student','Sagittarius',0,1,NULL,NULL),(3536867,0,17,'Student','Aquarius',0,1,NULL,NULL),(3655970,0,17,'Religion','Sagittarius',0,0,NULL,NULL),(4055095,0,33,'Government','Pisces',0,0,NULL,NULL),(3354828,0,16,'Internet','Gemini',0,0,NULL,NULL),(3560141,1,26,'Student','Scorpio',0,1,NULL,NULL),(3769478,0,24,'indUnk','Virgo',1,NULL,NULL,NULL),(3767918,0,15,'Technology','Capricorn',0,0,NULL,NULL),(3320342,1,23,'indUnk','Pisces',1,NULL,NULL,NULL),(4219169,1,14,'Student','Capricorn',0,1,NULL,NULL),(3556477,1,25,'indUnk','Libra',1,NULL,NULL,NULL),(4078761,1,35,'Religion','Taurus',0,0,NULL,NULL),(3907553,1,24,'indUnk','Aries',1,NULL,NULL,NULL),(3669958,0,24,'indUnk','Scorpio',1,NULL,NULL,NULL),(4204123,1,16,'indUnk','Scorpio',1,NULL,NULL,NULL),(3443825,0,25,'indUnk','Aries',1,NULL,NULL,NULL),(4049195,0,17,'Student','Scorpio',0,1,NULL,NULL),(3977680,1,16,'indUnk','Scorpio',1,NULL,NULL,NULL),(3732960,0,17,'Student','Sagittarius',0,1,NULL,NULL),(2645566,1,16,'Student','Virgo',0,1,NULL,NULL),(3471722,1,14,'indUnk','Capricorn',1,NULL,NULL,NULL),(4174428,0,24,'Education','Capricorn',0,0,NULL,NULL),(4185342,1,23,'indUnk','Gemini',1,NULL,NULL,NULL),(4127857,1,17,'Student','Gemini',0,1,NULL,NULL),(3828283,0,14,'indUnk','Cancer',1,NULL,NULL,NULL),(4257430,1,16,'Student','Aquarius',0,1,NULL,NULL),(3343256,0,23,'indUnk','Virgo',1,NULL,NULL,NULL),(3879492,0,14,'Student','Cancer',0,1,NULL,NULL),(3363036,1,23,'indUnk','Aries',1,NULL,NULL,NULL),(4279253,1,42,'indUnk','Aries',1,NULL,NULL,NULL),(1223044,0,16,'indUnk','Sagittarius',1,NULL,NULL,NULL),(1374061,0,34,'indUnk','Gemini',1,NULL,NULL,NULL),(3051898,1,16,'Student','Taurus',0,1,NULL,NULL),(3426788,0,23,'Technology','Scorpio',0,0,NULL,NULL),(4008178,0,14,'Student','Virgo',0,1,NULL,NULL),(3792567,1,35,'Transportation','Aquarius',0,0,NULL,NULL),(4015188,0,35,'Arts','Cancer',0,0,NULL,NULL),(3957657,0,16,'Student','Virgo',0,1,NULL,NULL),(3601759,0,25,'indUnk','Pisces',1,NULL,NULL,NULL),(3066719,0,27,'Technology','Pisces',0,0,NULL,NULL),(3982753,0,24,'indUnk','Pisces',1,NULL,NULL,NULL),(3878306,0,14,'Student','Taurus',0,1,NULL,NULL),(3330628,1,13,'Student','Virgo',0,1,NULL,NULL),(3546432,1,47,'Consulting','Gemini',0,0,NULL,NULL),(3372184,0,26,'Communications-Media','Pisces',0,0,NULL,NULL),(3503201,0,27,'indUnk','Cancer',1,NULL,NULL,NULL),(3626700,1,15,'Student','Libra',0,1,NULL,NULL),(3821176,1,42,'Education','Aquarius',0,0,NULL,NULL),(4033738,0,13,'Student','Pisces',0,1,NULL,NULL),(3656035,1,16,'indUnk','Leo',1,NULL,NULL,NULL),(3413492,0,27,'Manufacturing','Scorpio',0,0,NULL,NULL),(3419991,1,24,'indUnk','Cancer',1,NULL,NULL,NULL),(3832385,0,24,'indUnk','Libra',1,NULL,NULL,NULL),(3515252,1,27,'Education','Virgo',0,0,NULL,NULL),(1675646,0,25,'indUnk','Leo',1,NULL,NULL,NULL),(4042606,1,17,'Student','Capricorn',0,1,NULL,NULL),(4008161,1,17,'indUnk','Taurus',1,NULL,NULL,NULL),(4138081,0,17,'Marketing','Sagittarius',0,0,NULL,NULL),(3519860,1,16,'indUnk','Virgo',1,NULL,NULL,NULL),(4100934,1,23,'Arts','Sagittarius',0,0,NULL,NULL),(3976199,1,17,'indUnk','Capricorn',1,NULL,NULL,NULL),(4115894,1,15,'indUnk','Virgo',1,NULL,NULL,NULL),(4240336,1,17,'Student','Scorpio',0,1,NULL,NULL),(3469176,0,17,'Engineering','Scorpio',0,0,NULL,NULL),(1550687,1,24,'Communications-Media','Sagittarius',0,0,NULL,NULL),(1937270,0,36,'indUnk','Virgo',1,NULL,NULL,NULL),(3247172,1,36,'BusinessServices','Aries',0,0,NULL,NULL),(4200309,0,27,'Communications-Media','Libra',0,0,NULL,NULL),(3661106,1,16,'Student','Leo',0,1,NULL,NULL),(3371580,0,15,'indUnk','Capricorn',1,NULL,NULL,NULL),(3376440,0,37,'Communications-Media','Aries',0,0,NULL,NULL),(3721842,0,23,'Student','Aquarius',0,1,NULL,NULL),(3350341,1,35,'Marketing','Cancer',0,0,NULL,NULL),(4053312,1,33,'Architecture','Taurus',0,0,NULL,NULL),(4141669,0,47,'Religion','Cancer',0,0,NULL,NULL),(4236769,1,16,'Education','Sagittarius',0,0,NULL,NULL),(3422339,1,15,'Sports-Recreation','Cancer',0,0,NULL,NULL),(3623073,1,17,'Student','Capricorn',0,1,NULL,NULL),(3206585,1,25,'HumanResources','Leo',0,0,NULL,NULL),(3885215,1,27,'indUnk','Cancer',1,NULL,NULL,NULL),(3504445,1,15,'Student','Gemini',0,1,NULL,NULL),(1694057,1,25,'Arts','Capricorn',0,0,NULL,NULL),(3457789,0,15,'Student','Sagittarius',0,1,NULL,NULL),(4257468,1,17,'Student','Pisces',0,1,NULL,NULL),(3523319,0,24,'Internet','Pisces',0,0,NULL,NULL),(3970082,1,26,'indUnk','Gemini',1,NULL,NULL,NULL),(3632466,0,25,'indUnk','Libra',1,NULL,NULL,NULL),(3929371,1,23,'Science','Aquarius',0,0,NULL,NULL),(3642942,0,37,'Consulting','Taurus',0,0,NULL,NULL),(4163734,0,14,'Student','Libra',0,1,NULL,NULL),(3389016,1,27,'Technology','Aquarius',0,0,NULL,NULL),(3768627,1,15,'indUnk','Aquarius',1,NULL,NULL,NULL),(4256858,0,25,'indUnk','Taurus',1,NULL,NULL,NULL),(2839449,1,26,'Biotech','Aries',0,0,NULL,NULL),(2972526,1,25,'Consulting','Virgo',0,0,NULL,NULL),(3558648,0,34,'Internet','Sagittarius',0,0,NULL,NULL),(3344485,0,25,'indUnk','Capricorn',1,NULL,NULL,NULL),(3148889,1,26,'Communications-Media','Capricorn',0,0,NULL,NULL),(3797536,0,16,'indUnk','Libra',1,NULL,NULL,NULL),(3426176,0,17,'Student','Aries',0,1,NULL,NULL),(3705153,0,17,'Science','Scorpio',0,0,NULL,NULL),(3802511,1,25,'indUnk','Libra',1,NULL,NULL,NULL),(4028811,1,24,'Communications-Media','Pisces',0,0,NULL,NULL),(4198810,0,27,'indUnk','Sagittarius',1,NULL,NULL,NULL),(3377701,0,17,'Student','Libra',0,1,NULL,NULL),(483266,0,33,'indUnk','Taurus',1,NULL,NULL,NULL),(3288591,0,37,'Internet','Cancer',0,0,NULL,NULL),(3614613,0,14,'Student','Aquarius',0,1,NULL,NULL),(3582331,1,23,'indUnk','Pisces',1,NULL,NULL,NULL),(3570813,1,24,'indUnk','Gemini',1,NULL,NULL,NULL),(3654666,0,24,'indUnk','Taurus',1,NULL,NULL,NULL),(3948533,1,41,'indUnk','Pisces',1,NULL,NULL,NULL),(3737568,0,16,'indUnk','Pisces',1,NULL,NULL,NULL),(4176051,1,45,'Consulting','Pisces',0,0,NULL,NULL),(3776616,1,17,'Student','Sagittarius',0,1,NULL,NULL),(3569188,1,14,'Student','Gemini',0,1,NULL,NULL),(3428326,0,23,'Manufacturing','Leo',0,0,NULL,NULL),(1156278,0,37,'Internet','Taurus',0,0,NULL,NULL),(3490604,1,14,'Student','Cancer',0,1,NULL,NULL),(3308717,0,25,'Technology','Sagittarius',0,0,NULL,NULL),(1841346,1,35,'indUnk','Scorpio',1,NULL,NULL,NULL),(918175,1,27,'indUnk','Aquarius',1,NULL,NULL,NULL),(3464484,1,24,'indUnk','Aries',1,NULL,NULL,NULL),(2543400,0,17,'Student','Capricorn',0,1,NULL,NULL),(174357,1,23,'indUnk','Aquarius',1,NULL,NULL,NULL),(3524887,0,25,'indUnk','Gemini',1,NULL,NULL,NULL),(4177364,0,36,'Transportation','Taurus',0,0,NULL,NULL),(3683867,1,24,'Student','Leo',0,1,NULL,NULL),(3340813,0,16,'Arts','Aquarius',0,0,NULL,NULL),(2413978,0,15,'Education','Capricorn',0,0,NULL,NULL),(3930072,1,17,'indUnk','Libra',1,NULL,NULL,NULL),(3613005,0,15,'indUnk','Gemini',1,NULL,NULL,NULL),(3601621,0,23,'indUnk','Gemini',1,NULL,NULL,NULL),(2957170,0,26,'indUnk','Aquarius',1,NULL,NULL,NULL),(1405024,1,24,'indUnk','Cancer',1,NULL,NULL,NULL),(1794926,1,23,'Communications-Media','Sagittarius',0,0,NULL,NULL),(2004217,1,17,'Education','Leo',0,0,NULL,NULL),(4238678,1,13,'Student','Pisces',0,1,NULL,NULL),(4022853,0,17,'Student','Leo',0,1,NULL,NULL),(2101657,0,36,'Technology','Aries',0,0,NULL,NULL),(3551081,1,17,'Student','Sagittarius',0,1,NULL,NULL),(3719046,1,15,'Student','Aquarius',0,1,NULL,NULL),(3389942,0,17,'Technology','Gemini',0,0,NULL,NULL),(3956613,0,36,'indUnk','Pisces',1,NULL,NULL,NULL),(4285920,0,24,'Education','Scorpio',0,0,NULL,NULL),(3479211,1,27,'Advertising','Aries',0,0,NULL,NULL),(3961392,1,14,'indUnk','Gemini',1,NULL,NULL,NULL),(3959121,1,15,'indUnk','Aquarius',1,NULL,NULL,NULL),(4154216,1,44,'indUnk','Gemini',1,NULL,NULL,NULL),(3818197,1,25,'Student','Cancer',0,1,NULL,NULL),(4164945,0,16,'Student','Libra',0,1,NULL,NULL),(3425069,0,23,'Technology','Libra',0,0,NULL,NULL),(3336745,1,14,'indUnk','Capricorn',1,NULL,NULL,NULL),(3851875,1,34,'BusinessServices','Sagittarius',0,0,NULL,NULL),(1826527,0,25,'Science','Cancer',0,0,NULL,NULL),(4059083,0,41,'Communications-Media','Aries',0,0,NULL,NULL),(3894428,1,17,'indUnk','Cancer',1,NULL,NULL,NULL),(3902340,0,35,'Communications-Media','Leo',0,0,NULL,NULL),(3544112,1,23,'Arts','Libra',0,0,NULL,NULL),(2863105,0,43,'Non-Profit','Cancer',0,0,NULL,NULL),(3590069,1,24,'indUnk','Leo',1,NULL,NULL,NULL),(3323485,0,14,'Student','Sagittarius',0,1,NULL,NULL),(4127797,0,25,'Student','Libra',0,1,NULL,NULL),(3441713,1,17,'indUnk','Leo',1,NULL,NULL,NULL),(3309541,0,37,'Banking','Taurus',0,0,NULL,NULL),(3767000,1,15,'Student','Libra',0,1,NULL,NULL),(3413512,0,16,'Student','Pisces',0,1,NULL,NULL),(3376420,1,15,'Student','Virgo',0,1,NULL,NULL),(3308014,1,35,'indUnk','Libra',1,NULL,NULL,NULL),(3878084,0,13,'Student','Cancer',0,1,NULL,NULL),(4072259,1,24,'indUnk','Capricorn',1,NULL,NULL,NULL),(3146933,1,17,'Student','Gemini',0,1,NULL,NULL),(3810203,1,13,'indUnk','Gemini',1,NULL,NULL,NULL),(317581,0,26,'Technology','Scorpio',0,0,NULL,NULL),(3677344,1,36,'Automotive','Cancer',0,0,NULL,NULL),(3423418,0,25,'Student','Aries',0,1,NULL,NULL),(3471740,1,27,'RealEstate','Capricorn',0,0,NULL,NULL),(518116,1,27,'indUnk','Pisces',1,NULL,NULL,NULL),(2899366,0,17,'Student','Aquarius',0,1,NULL,NULL),(3327596,1,15,'indUnk','Leo',1,NULL,NULL,NULL),(3447369,0,17,'indUnk','Aquarius',1,NULL,NULL,NULL),(3903607,1,25,'Law','Taurus',0,0,NULL,NULL),(4303113,1,23,'indUnk','Leo',1,NULL,NULL,NULL),(2787343,0,14,'Education','Leo',0,0,NULL,NULL),(3875375,0,14,'indUnk','Libra',1,NULL,NULL,NULL),(3923707,1,15,'Student','Leo',0,1,NULL,NULL),(3431135,1,15,'Student','Libra',0,1,NULL,NULL),(3638066,0,16,'Student','Scorpio',0,1,NULL,NULL),(3600967,1,33,'Arts','Scorpio',0,0,NULL,NULL),(3556371,0,13,'Student','Sagittarius',0,1,NULL,NULL),(1599685,0,23,'Student','Capricorn',0,1,NULL,NULL),(2190377,1,16,'Student','Gemini',0,1,NULL,NULL),(3448971,1,15,'indUnk','Cancer',1,NULL,NULL,NULL),(3208138,0,13,'indUnk','Pisces',1,NULL,NULL,NULL),(3655173,0,26,'Law','Virgo',0,0,NULL,NULL),(3480958,1,24,'Government','Gemini',0,0,NULL,NULL),(3519292,0,14,'Student','Scorpio',0,1,NULL,NULL),(3692874,0,15,'LawEnforcement-Security','Virgo',0,0,NULL,NULL),(3252533,0,16,'indUnk','Scorpio',1,NULL,NULL,NULL),(3992091,1,26,'indUnk','Sagittarius',1,NULL,NULL,NULL),(3905634,0,17,'indUnk','Cancer',1,NULL,NULL,NULL),(4218355,0,24,'indUnk','Aries',1,NULL,NULL,NULL),(2536117,1,16,'Student','Scorpio',0,1,NULL,NULL),(3631998,0,25,'Technology','Taurus',0,0,NULL,NULL),(3499599,1,15,'Student','Aquarius',0,1,NULL,NULL),(2108549,1,27,'Technology','Virgo',0,0,NULL,NULL),(4292654,1,33,'Banking','Libra',0,0,NULL,NULL),(4200075,0,14,'Student','Scorpio',0,1,NULL,NULL),(3097708,1,16,'indUnk','Capricorn',1,NULL,NULL,NULL),(3724910,0,17,'Student','Virgo',0,1,NULL,NULL),(3951300,0,16,'indUnk','Pisces',1,NULL,NULL,NULL),(3538351,1,17,'indUnk','Sagittarius',1,NULL,NULL,NULL),(3331655,0,13,'indUnk','Pisces',1,NULL,NULL,NULL),(3711841,0,16,'indUnk','Libra',1,NULL,NULL,NULL),(3332079,1,16,'indUnk','Cancer',1,NULL,NULL,NULL),(3438987,1,25,'indUnk','Pisces',1,NULL,NULL,NULL),(3682338,1,16,'Student','Aquarius',0,1,NULL,NULL),(3325039,0,26,'indUnk','Aquarius',1,NULL,NULL,NULL),(3403416,0,27,'Banking','Gemini',0,0,NULL,NULL),(3665752,0,26,'Arts','Sagittarius',0,0,NULL,NULL),(3393040,0,26,'Religion','Libra',0,0,NULL,NULL),(3845636,1,17,'Student','Cancer',0,1,NULL,NULL),(3558872,1,47,'Education','Cancer',0,0,NULL,NULL),(3380244,1,23,'Student','Gemini',0,1,NULL,NULL),(4105180,1,15,'Student','Capricorn',0,1,NULL,NULL),(3667102,1,46,'Law','Libra',0,0,NULL,NULL),(3590380,1,37,'Government','Taurus',0,0,NULL,NULL),(3917555,1,13,'Student','Aries',0,1,NULL,NULL),(3780716,1,17,'Student','Aries',0,1,NULL,NULL),(4117248,1,16,'indUnk','Pisces',1,NULL,NULL,NULL),(3827259,0,23,'indUnk','Cancer',1,NULL,NULL,NULL),(3466070,0,16,'Student','Pisces',0,1,NULL,NULL),(1916038,1,14,'indUnk','Cancer',1,NULL,NULL,NULL),(3585158,1,27,'Non-Profit','Capricorn',0,0,NULL,NULL),(3752323,0,24,'Science','Gemini',0,0,NULL,NULL),(2965205,0,33,'indUnk','Gemini',1,NULL,NULL,NULL),(3745004,1,13,'Student','Pisces',0,1,NULL,NULL),(3792732,1,26,'indUnk','Leo',1,NULL,NULL,NULL),(3481803,0,16,'Student','Capricorn',0,1,NULL,NULL),(3599758,0,24,'Non-Profit','Aries',0,0,NULL,NULL),(4144596,0,26,'indUnk','Aquarius',1,NULL,NULL,NULL),(3496909,1,16,'Education','Sagittarius',0,0,NULL,NULL),(3753339,1,23,'Education','Taurus',0,0,NULL,NULL),(3483306,0,23,'Manufacturing','Gemini',0,0,NULL,NULL),(3957509,0,38,'Technology','Aquarius',0,0,NULL,NULL),(3806681,0,24,'Telecommunications','Sagittarius',0,0,NULL,NULL),(3619647,0,15,'indUnk','Pisces',1,NULL,NULL,NULL),(3889401,0,23,'Student','Taurus',0,1,NULL,NULL),(4101011,1,23,'indUnk','Cancer',1,NULL,NULL,NULL),(4173964,0,16,'Student','Virgo',0,1,NULL,NULL),(4051770,0,24,'Law','Sagittarius',0,0,NULL,NULL),(2751885,1,24,'indUnk','Pisces',1,NULL,NULL,NULL),(3870139,0,16,'Student','Scorpio',0,1,NULL,NULL),(3435900,1,34,'indUnk','Leo',1,NULL,NULL,NULL),(3602354,1,17,'Publishing','Capricorn',0,0,NULL,NULL),(4224651,0,25,'Technology','Gemini',0,0,NULL,NULL),(4206637,1,24,'Communications-Media','Capricorn',0,0,NULL,NULL),(4312005,0,17,'indUnk','Scorpio',1,NULL,NULL,NULL),(4086028,0,17,'Student','Scorpio',0,1,NULL,NULL),(3368285,0,27,'Tourism','Leo',0,0,NULL,NULL),(3994022,1,24,'Consulting','Aries',0,0,NULL,NULL),(3363764,0,27,'Architecture','Cancer',0,0,NULL,NULL),(3065484,0,14,'Education','Pisces',0,0,NULL,NULL),(3860954,1,17,'indUnk','Scorpio',1,NULL,NULL,NULL),(4165509,1,15,'Student','Libra',0,1,NULL,NULL),(3894623,0,34,'Education','Libra',0,0,NULL,NULL),(3325171,0,24,'Technology','Virgo',0,0,NULL,NULL),(3327285,0,25,'Technology','Taurus',0,0,NULL,NULL),(1225881,0,17,'Internet','Scorpio',0,0,NULL,NULL),(477665,1,25,'indUnk','Leo',1,NULL,NULL,NULL),(4103386,0,13,'Student','Virgo',0,1,NULL,NULL),(4213897,1,40,'Non-Profit','Gemini',0,0,NULL,NULL),(3224168,1,15,'Student','Capricorn',0,1,NULL,NULL),(3614153,0,24,'indUnk','Sagittarius',1,NULL,NULL,NULL),(3660667,0,16,'indUnk','Gemini',1,NULL,NULL,NULL),(3849830,0,13,'Student','Cancer',0,1,NULL,NULL),(3235391,1,13,'indUnk','Taurus',1,NULL,NULL,NULL),(3497723,1,26,'Student','Taurus',0,1,NULL,NULL),(3639495,1,24,'Education','Scorpio',0,0,NULL,NULL),(3805349,0,14,'Student','Virgo',0,1,NULL,NULL),(3860294,0,14,'indUnk','Capricorn',1,NULL,NULL,NULL),(3955951,0,47,'indUnk','Sagittarius',1,NULL,NULL,NULL),(3358017,1,23,'indUnk','Taurus',1,NULL,NULL,NULL),(3495545,0,26,'indUnk','Gemini',1,NULL,NULL,NULL),(3473164,1,25,'indUnk','Pisces',1,NULL,NULL,NULL),(3868312,0,25,'Education','Gemini',0,0,NULL,NULL),(3937692,1,26,'indUnk','Virgo',1,NULL,NULL,NULL),(3700178,0,17,'Student','Libra',0,1,NULL,NULL),(4321513,0,16,'Student','Taurus',0,1,NULL,NULL),(3997969,1,15,'indUnk','Taurus',1,NULL,NULL,NULL),(3499629,1,14,'indUnk','Virgo',1,NULL,NULL,NULL),(4058677,0,25,'Advertising','Cancer',0,0,NULL,NULL),(3580741,0,25,'Internet','Libra',0,0,NULL,NULL),(4061665,0,25,'Advertising','Leo',0,0,NULL,NULL),(3418474,0,15,'Student','Virgo',0,1,NULL,NULL),(2824507,0,26,'Government','Libra',0,0,NULL,NULL),(664485,0,25,'indUnk','Taurus',1,NULL,NULL,NULL),(2061087,1,33,'indUnk','Cancer',1,NULL,NULL,NULL),(4049268,0,17,'Military','Virgo',0,0,NULL,NULL),(3990866,0,23,'Internet','Cancer',0,0,NULL,NULL),(3366974,0,17,'Student','Libra',0,1,NULL,NULL),(2711915,0,25,'Arts','Taurus',0,0,NULL,NULL),(3400678,0,15,'indUnk','Leo',1,NULL,NULL,NULL),(3990427,1,25,'Banking','Sagittarius',0,0,NULL,NULL),(2314011,1,23,'Student','Capricorn',0,1,NULL,NULL),(3994011,0,38,'Government','Leo',0,0,NULL,NULL),(4239436,0,17,'Technology','Capricorn',0,0,NULL,NULL),(3739687,1,24,'Education','Virgo',0,0,NULL,NULL),(4256898,1,15,'Student','Taurus',0,1,NULL,NULL),(3751728,1,16,'Student','Taurus',0,1,NULL,NULL),(2116991,0,33,'Internet','Virgo',0,0,NULL,NULL),(3820930,1,23,'indUnk','Libra',1,NULL,NULL,NULL),(4125847,1,27,'indUnk','Gemini',1,NULL,NULL,NULL),(3545726,1,16,'Student','Aquarius',0,1,NULL,NULL),(4150699,1,24,'indUnk','Aquarius',1,NULL,NULL,NULL),(4231133,1,13,'indUnk','Scorpio',1,NULL,NULL,NULL),(3451587,0,48,'indUnk','Capricorn',1,NULL,NULL,NULL),(3412947,0,26,'Internet','Gemini',0,0,NULL,NULL),(3330196,0,23,'Marketing','Virgo',0,0,NULL,NULL),(673235,0,37,'Internet','Capricorn',0,0,NULL,NULL),(1292625,1,14,'indUnk','Taurus',1,NULL,NULL,NULL),(3650032,0,25,'Education','Aries',0,0,NULL,NULL),(3907107,0,15,'Student','Libra',0,1,NULL,NULL),(3908022,1,16,'Student','Pisces',0,1,NULL,NULL),(3550869,0,25,'BusinessServices','Virgo',0,0,NULL,NULL),(3685554,0,14,'Student','Gemini',0,1,NULL,NULL),(3575606,1,24,'indUnk','Taurus',1,NULL,NULL,NULL),(3715376,1,17,'Student','Gemini',0,1,NULL,NULL),(3502527,1,24,'indUnk','Scorpio',1,NULL,NULL,NULL),(3809990,0,38,'Military','Virgo',0,0,NULL,NULL),(3998459,1,23,'Student','Cancer',0,1,NULL,NULL),(3723604,0,15,'Student','Scorpio',0,1,NULL,NULL),(4047559,0,24,'indUnk','Cancer',1,NULL,NULL,NULL),(4016792,1,45,'Construction','Capricorn',0,0,NULL,NULL),(3700421,1,24,'Science','Aries',0,0,NULL,NULL),(3558898,0,26,'indUnk','Aquarius',1,NULL,NULL,NULL),(4185692,1,15,'indUnk','Taurus',1,NULL,NULL,NULL),(3565927,0,25,'indUnk','Scorpio',1,NULL,NULL,NULL),(3500159,1,37,'Education','Pisces',0,0,NULL,NULL),(3885952,0,24,'Technology','Aries',0,0,NULL,NULL),(3309500,0,24,'Consulting','Taurus',0,0,NULL,NULL),(3517315,1,23,'indUnk','Scorpio',1,NULL,NULL,NULL),(3647581,0,17,'Student','Sagittarius',0,1,NULL,NULL),(4139489,0,16,'Student','Scorpio',0,1,NULL,NULL),(3495224,0,16,'Student','Aries',0,1,NULL,NULL),(3931951,1,16,'Student','Virgo',0,1,NULL,NULL),(4204173,1,26,'Technology','Aquarius',0,0,NULL,NULL),(3835812,1,47,'Non-Profit','Pisces',0,0,NULL,NULL),(4293254,1,16,'Student','Cancer',0,1,NULL,NULL),(4095362,1,16,'Education','Pisces',0,0,NULL,NULL),(4130944,1,37,'Education','Aries',0,0,NULL,NULL),(4256976,0,24,'indUnk','Capricorn',1,NULL,NULL,NULL),(808956,1,25,'Government','Aries',0,0,NULL,NULL),(3746444,0,17,'Student','Leo',0,1,NULL,NULL),(4242609,0,16,'Student','Aries',0,1,NULL,NULL),(3833031,0,24,'indUnk','Aquarius',1,NULL,NULL,NULL),(3644724,1,36,'indUnk','Leo',1,NULL,NULL,NULL),(3099605,0,16,'Religion','Aries',0,0,NULL,NULL),(4073027,1,33,'Transportation','Sagittarius',0,0,NULL,NULL),(3700686,0,27,'Technology','Aquarius',0,0,NULL,NULL),(4075287,0,15,'indUnk','Sagittarius',1,NULL,NULL,NULL),(3297517,1,27,'indUnk','Pisces',1,NULL,NULL,NULL),(4232070,0,16,'indUnk','Aries',1,NULL,NULL,NULL),(3753019,1,17,'Student','Libra',0,1,NULL,NULL),(3857776,1,13,'indUnk','Taurus',1,NULL,NULL,NULL),(2490703,1,17,'Student','Libra',0,1,NULL,NULL),(3952263,1,13,'indUnk','Cancer',1,NULL,NULL,NULL),(3452845,0,25,'Banking','Taurus',0,0,NULL,NULL),(1196970,0,16,'indUnk','Sagittarius',1,NULL,NULL,NULL),(4142553,1,25,'indUnk','Pisces',1,NULL,NULL,NULL),(3816360,1,23,'Non-Profit','Pisces',0,0,NULL,NULL),(4162455,0,40,'Internet','Cancer',0,0,NULL,NULL),(3572448,0,23,'Student','Libra',0,1,NULL,NULL),(3502783,1,25,'Non-Profit','Cancer',0,0,NULL,NULL),(3494693,1,23,'Student','Gemini',0,1,NULL,NULL),(2238828,1,23,'indUnk','Libra',1,NULL,NULL,NULL),(4178311,1,16,'indUnk','Cancer',1,NULL,NULL,NULL),(4284264,0,27,'Telecommunications','Gemini',0,0,NULL,NULL),(950122,1,15,'Student','Virgo',0,1,NULL,NULL),(4176168,0,15,'Student','Leo',0,1,NULL,NULL),(4221834,1,24,'indUnk','Leo',1,NULL,NULL,NULL),(4021133,1,17,'indUnk','Libra',1,NULL,NULL,NULL),(3722603,0,16,'Arts','Capricorn',0,0,NULL,NULL),(3319097,0,15,'Student','Sagittarius',0,1,NULL,NULL),(3661997,1,13,'indUnk','Taurus',1,NULL,NULL,NULL),(3152463,0,17,'Student','Sagittarius',0,1,NULL,NULL),(3471937,1,23,'BusinessServices','Libra',0,0,NULL,NULL),(3481255,0,26,'indUnk','Sagittarius',1,NULL,NULL,NULL),(3239841,0,17,'Arts','Virgo',0,0,NULL,NULL),(4044095,0,27,'indUnk','Cancer',1,NULL,NULL,NULL),(4261141,0,15,'Student','Sagittarius',0,1,NULL,NULL),(3382977,0,15,'Student','Taurus',0,1,NULL,NULL),(2693059,0,17,'Student','Aries',0,1,NULL,NULL),(4186455,0,27,'Arts','Aries',0,0,NULL,NULL),(2673530,0,33,'indUnk','Aries',1,NULL,NULL,NULL),(4199600,1,24,'indUnk','Pisces',1,NULL,NULL,NULL),(4030244,1,27,'indUnk','Pisces',1,NULL,NULL,NULL),(3850554,0,33,'Technology','Aquarius',0,0,NULL,NULL),(4231869,1,17,'indUnk','Scorpio',1,NULL,NULL,NULL),(3766692,1,17,'indUnk','Pisces',1,NULL,NULL,NULL),(446275,1,17,'indUnk','Virgo',1,NULL,NULL,NULL),(3927367,1,15,'Student','Virgo',0,1,NULL,NULL),(3309198,0,14,'Student','Cancer',0,1,NULL,NULL),(3602089,0,15,'Student','Libra',0,1,NULL,NULL),(3818324,1,17,'Fashion','Taurus',0,0,NULL,NULL),(3496136,1,38,'indUnk','Libra',1,NULL,NULL,NULL),(4271044,0,23,'Student','Gemini',0,1,NULL,NULL),(1835108,1,24,'indUnk','Taurus',1,NULL,NULL,NULL),(4066793,0,26,'Engineering','Aries',0,0,NULL,NULL),(2694370,0,27,'indUnk','Taurus',1,NULL,NULL,NULL),(2510334,0,17,'Student','Pisces',0,1,NULL,NULL),(979795,0,26,'Technology','Taurus',0,0,NULL,NULL),(2606936,1,17,'Student','Virgo',0,1,NULL,NULL),(3972921,0,35,'Technology','Aries',0,0,NULL,NULL),(4255254,0,16,'indUnk','Virgo',1,NULL,NULL,NULL),(3485704,0,16,'Student','Pisces',0,1,NULL,NULL),(3870574,0,17,'Student','Cancer',0,1,NULL,NULL),(4179297,1,27,'Government','Aries',0,0,NULL,NULL),(4052167,1,35,'Education','Libra',0,0,NULL,NULL),(4110791,0,17,'Student','Virgo',0,1,NULL,NULL),(3743431,1,27,'indUnk','Virgo',1,NULL,NULL,NULL),(4313487,0,26,'indUnk','Sagittarius',1,NULL,NULL,NULL),(1254400,0,34,'Religion','Leo',0,0,NULL,NULL),(3586100,1,24,'Education','Leo',0,0,NULL,NULL),(4271040,1,15,'Student','Aries',0,1,NULL,NULL),(4141026,1,34,'Arts','Leo',0,0,NULL,NULL),(485307,0,17,'indUnk','Cancer',1,NULL,NULL,NULL),(3417884,1,17,'indUnk','Leo',1,NULL,NULL,NULL),(2310335,0,15,'Student','Aries',0,1,NULL,NULL),(4052963,1,14,'indUnk','Capricorn',1,NULL,NULL,NULL),(3431501,0,15,'indUnk','Sagittarius',1,NULL,NULL,NULL),(3302611,1,23,'indUnk','Cancer',1,NULL,NULL,NULL),(3670880,0,14,'Student','Capricorn',0,1,NULL,NULL),(3863615,0,27,'Internet','Scorpio',0,0,NULL,NULL),(3893735,0,17,'Student','Cancer',0,1,NULL,NULL),(4184069,1,27,'Museums-Libraries','Leo',0,0,NULL,NULL),(4067717,1,17,'Student','Taurus',0,1,NULL,NULL),(4004415,1,27,'Law','Aries',0,0,NULL,NULL),(3482303,1,33,'indUnk','Pisces',1,NULL,NULL,NULL),(3972276,0,17,'Education','Aquarius',0,0,NULL,NULL),(2811560,1,15,'Student','Sagittarius',0,1,NULL,NULL),(3291871,1,25,'Arts','Libra',0,0,NULL,NULL),(4112653,1,15,'Student','Sagittarius',0,1,NULL,NULL),(450169,0,17,'indUnk','Sagittarius',1,NULL,NULL,NULL),(4148541,0,15,'Student','Scorpio',0,1,NULL,NULL),(3542735,0,26,'Advertising','Aries',0,0,NULL,NULL),(942828,1,34,'indUnk','Cancer',1,NULL,NULL,NULL),(3993587,0,17,'Agriculture','Aquarius',0,0,NULL,NULL),(825029,1,17,'Student','Aries',0,1,NULL,NULL),(3568011,0,16,'Student','Leo',0,1,NULL,NULL),(3881689,0,44,'indUnk','Taurus',1,NULL,NULL,NULL),(3901750,1,14,'indUnk','Sagittarius',1,NULL,NULL,NULL),(2278942,0,26,'Technology','Virgo',0,0,NULL,NULL),(3716795,0,16,'indUnk','Sagittarius',1,NULL,NULL,NULL),(3969341,1,14,'Student','Cancer',0,1,NULL,NULL),(3126039,1,17,'Student','Pisces',0,1,NULL,NULL),(3686696,1,23,'Student','Aquarius',0,1,NULL,NULL),(911744,0,23,'Telecommunications','Capricorn',0,0,NULL,NULL),(4168666,0,24,'Government','Cancer',0,0,NULL,NULL),(3801580,0,33,'indUnk','Cancer',1,NULL,NULL,NULL),(4229520,0,16,'indUnk','Scorpio',1,NULL,NULL,NULL),(4170570,1,25,'indUnk','Libra',1,NULL,NULL,NULL),(4007094,0,17,'indUnk','Gemini',1,NULL,NULL,NULL),(4156867,1,16,'Student','Sagittarius',0,1,NULL,NULL),(3122944,1,24,'indUnk','Sagittarius',1,NULL,NULL,NULL),(4168699,0,15,'Student','Scorpio',0,1,NULL,NULL),(3649624,0,16,'Student','Sagittarius',0,1,NULL,NULL),(3838427,0,15,'indUnk','Taurus',1,NULL,NULL,NULL),(3566811,1,27,'Student','Gemini',0,1,NULL,NULL),(3723475,1,15,'Student','Cancer',0,1,NULL,NULL),(2903378,1,15,'indUnk','Capricorn',1,NULL,NULL,NULL),(4042056,1,15,'Arts','Virgo',0,0,NULL,NULL),(2366391,1,25,'Arts','Libra',0,0,NULL,NULL),(3290021,1,34,'Education','Aries',0,0,NULL,NULL),(4213837,0,16,'Education','Virgo',0,0,NULL,NULL),(1011311,1,17,'indUnk','Scorpio',1,NULL,NULL,NULL),(2864451,0,26,'Communications-Media','Taurus',0,0,NULL,NULL),(3471727,1,24,'indUnk','Virgo',1,NULL,NULL,NULL),(2710453,0,25,'Architecture','Gemini',0,0,NULL,NULL),(3324842,0,25,'Arts','Leo',0,0,NULL,NULL),(3855724,1,34,'Education','Cancer',0,0,NULL,NULL),(4154751,1,16,'Arts','Virgo',0,0,NULL,NULL),(3662868,0,14,'Student','Gemini',0,1,NULL,NULL),(2952959,1,24,'indUnk','Capricorn',1,NULL,NULL,NULL),(3801041,0,25,'Government','Pisces',0,0,NULL,NULL),(4122908,0,16,'Arts','Scorpio',0,0,NULL,NULL),(4176133,1,14,'Student','Scorpio',0,1,NULL,NULL),(3738990,0,14,'Student','Virgo',0,1,NULL,NULL),(4108504,0,15,'indUnk','Aries',1,NULL,NULL,NULL),(3976084,0,17,'indUnk','Scorpio',1,NULL,NULL,NULL),(3382799,1,26,'Consulting','Virgo',0,0,NULL,NULL),(4133033,1,24,'indUnk','Virgo',1,NULL,NULL,NULL),(4227371,1,24,'indUnk','Libra',1,NULL,NULL,NULL),(3457392,0,23,'Internet','Taurus',0,0,NULL,NULL),(4144593,0,16,'Student','Scorpio',0,1,NULL,NULL),(4300342,0,33,'Manufacturing','Aries',0,0,NULL,NULL),(3489929,1,25,'Student','Cancer',0,1,NULL,NULL),(4311120,1,25,'indUnk','Cancer',1,NULL,NULL,NULL),(468786,0,45,'Religion','Aries',0,0,NULL,NULL),(4097182,1,17,'Student','Leo',0,1,NULL,NULL),(1188176,1,35,'Education','Leo',0,0,NULL,NULL),(3889090,0,24,'Education','Pisces',0,0,NULL,NULL),(4150492,1,24,'RealEstate','Pisces',0,0,NULL,NULL),(3840303,0,27,'indUnk','Capricorn',1,NULL,NULL,NULL),(4042287,1,17,'indUnk','Capricorn',1,NULL,NULL,NULL),(2869535,1,24,'Student','Scorpio',0,1,NULL,NULL),(4168492,1,36,'indUnk','Gemini',1,NULL,NULL,NULL),(4039381,0,24,'Consulting','Libra',0,0,NULL,NULL),(3847790,1,36,'indUnk','Cancer',1,NULL,NULL,NULL),(3178789,0,36,'Publishing','Cancer',0,0,NULL,NULL),(3493704,0,26,'Engineering','Cancer',0,0,NULL,NULL),(3732159,0,27,'indUnk','Sagittarius',1,NULL,NULL,NULL),(4016135,0,36,'Technology','Gemini',0,0,NULL,NULL),(4159699,1,14,'Student','Taurus',0,1,NULL,NULL),(4043736,1,23,'Arts','Sagittarius',0,0,NULL,NULL),(3483262,0,24,'Banking','Libra',0,0,NULL,NULL),(3616679,1,47,'indUnk','Cancer',1,NULL,NULL,NULL),(4100455,1,24,'Religion','Cancer',0,0,NULL,NULL),(3826242,0,42,'indUnk','Aries',1,NULL,NULL,NULL),(2959033,1,24,'indUnk','Cancer',1,NULL,NULL,NULL),(4149956,1,45,'Law','Sagittarius',0,0,NULL,NULL),(1039908,1,16,'indUnk','Gemini',1,NULL,NULL,NULL),(3992553,1,23,'indUnk','Aries',1,NULL,NULL,NULL),(2228840,0,33,'Technology','Aries',0,0,NULL,NULL),(3294176,0,37,'indUnk','Leo',1,NULL,NULL,NULL),(3559653,1,25,'Student','Gemini',0,1,NULL,NULL),(3335423,1,33,'HumanResources','Sagittarius',0,0,NULL,NULL),(3693954,1,14,'Internet','Leo',0,0,NULL,NULL),(3242831,0,16,'indUnk','Taurus',1,NULL,NULL,NULL),(4284876,1,14,'indUnk','Gemini',1,NULL,NULL,NULL),(3875285,1,43,'Education','Pisces',0,0,NULL,NULL),(2542729,0,26,'Technology','Scorpio',0,0,NULL,NULL),(3840766,1,48,'indUnk','Libra',1,NULL,NULL,NULL),(4043476,0,15,'Student','Cancer',0,1,NULL,NULL),(3348721,0,26,'Technology','Libra',0,0,NULL,NULL),(3294001,1,17,'indUnk','Capricorn',1,NULL,NULL,NULL);
/*!40000 ALTER TABLE `blog_outcomes` ENABLE KEYS */;
UNLOCK TABLES;

--
-- Table structure for table `msgs`
--

DROP TABLE IF EXISTS `msgs`;
/*!40101 SET @saved_cs_client     = @@character_set_client */;
/*!40101 SET character_set_client = utf8 */;
CREATE TABLE `msgs` (
  `message_id` int(11) NOT NULL AUTO_INCREMENT,
  `user_id` int(10) unsigned DEFAULT NULL,
  `date` varchar(64) DEFAULT NULL,
  `created_time` datetime DEFAULT NULL,
  `message` text,
  PRIMARY KEY (`message_id`),
  KEY `user_id` (`user_id`),
  KEY `datedex` (`created_time`)
) ENGINE=MyISAM AUTO_INCREMENT=31675 DEFAULT CHARSET=latin1;
/*!40101 SET character_set_client = @saved_cs_client */;

--
-- Dumping data for table `msgs`
--

LOCK TABLES `msgs` WRITE;
/*!40000 ALTER TABLE `msgs` DISABLE KEYS */;
INSERT INTO `msgs` VALUES (1,3991108,'31,July,2004','2004-07-31 00:00:00','   can you believe it?? my mom wouln\'t let me go out on my b\'day...i was really really mad at her. still am. but i got more presents from my friends this year. so thats great.  '),(2,3991108,'25,July,2004','2004-07-25 00:00:00','   miss su  used to be my living skills teacher when i was in form three. she passed away recently due to appendix eruption. she was 47. it was a real shock to all of us. most of us found out through an sms forwarded by the headmistress. it was a saturday. i couldn\'t believe it when my friend told me. it seemed surreal. a few days before i just saw her.&nbsp;it didn\'t sink in till monday. her funeral was held at the school chapel. all of us were given a chance to pay our last respects to her. i\'m not sure if this is the right word to use but her funeral was beautiful. many people including a few of the teachers, her sister and students talked about how she touched their lives. many cried their hearts out.&nbsp;  she has touched me as well in many ways. when she was teaching me she used to do our kh project for us even though she grumbled a little and scolded us. she sewd almost the whole project for us. i\'m thankful to her for that. i don\'t think i could have finished it by myself.  during the last few months of her life she was touched by the lord. she joined the rcia. she was very keen to be baptised. her enthusiasm in getting to know god better puts me and many others to shame.she even said she would share with us(cathecism students)&nbsp;her experiences in the rcia. i was really looking forward to it. but unfortunately she went back to the lord before she could share with us.    miss su  was a very talkative person. once she starts you cannot stop her. she\'ll go from one topic to another. she was an open book. she would talk about everything from latest issues to family problems. she opened her heart to everyone. she would go out of her way to help people. she was one person who lived her life to the fullest. its hard to believe i wont be seeing her around school anymore.  i believe she is in heaven now. i also believe she died because god loved her so much that he wanted her by his side. she is by the side of her creator whom she loved so much.  thank you  miss su  for all you have done for us. you will be in our hearts forever.  '),(3,3991108,'24,July,2004','2004-07-24 00:00:00','   i\'m lookin for info about cancer in bm...i have to write an essay about it..so much homework so little time..and my test starts on monday...phew...how am i ever gona finish my revision???  '),(4,3991108,'24,July,2004','2004-07-24 00:00:00','   what a time to start blogging when the spm is only a few months away...i\'m actually banned from using de internet except for using it to find info to complete assignments...  '),(5,3991108,'01,August,2004','2004-08-01 00:00:00','   i cannot believe how many presents i\'m getting for my b\'day this year...its over but the presents are still coming...yay...but i\'m feeling kinda guilty coz i didn\'i give them presents for ther b\'days...i\'m terrible at remembering b\'days...  '),(6,3417138,'29,May,2004','2004-05-29 00:00:00','    So I\'ve just spent the last hour browsing Amazon.com.  Damn you! You are too good! I have settled on 5 choices:  1) Melon Collie and the Infinite Sadness By Smashing Pumpkins Double disc set. Was stolen from me quite a few years back. I miss them!   2) Franz Ferdinand By Franz Ferdinand You may have seen their \"Take Me Out\" Video on MTV. What a rockin\' tune. Must have.  3) Schmack! By Steriogram These kids are my age and younger. I just need to have them in my collection. What promise they have! (P.S. They have a video made almost entirely of yarn.... ponder that)  4) Stellastarr* By Stellastarr* These guys are too cool. My Coco has definitely hit on my TRL chart. They sound a little like Billy Idol (Dancin\' with Myself) & the Smiths. I know that sounds odd - but that\'s what it sounds like!)  5) Good News... By Modest Mouse \"Float On\" easily makes it into my top 10 of the month.  I know some of you think I have odd taste in music. I have attached some of their websites so that you can check it out for yourself. These are all edgy, indie punk style bands. If you like the Backstreet Boys, you need not go in. Happy Listening!  www.steriogram.com www.modestmouse.com www.franzferdinand.co.uk www.stellastarr.com   '),(7,3417138,'26,May,2004','2004-05-26 00:00:00','    Well, Jesper was taken away today. Apparently the lady down the street had called the Animal Shelter before she found out I was keeping it, and someone had been there looking for this particular kitty. So the crazy lady came and started wigging out in my yard because she missed her kitty so bad. Jesper was hesitant to go, but if it was her kitty, then she needed to take him... Guess what his real name was... Charlie Tuna. That lady was a freak. I will miss you Jesper.. you will alway be remembered on my blog!!!  :(   (I\'m crying)   '),(8,3417138,'24,May,2004','2004-05-24 00:00:00','    urlLink    ....yawn... Good morning, Wilson.... Wilson? Wilson? WILLLLLLLLLSSSSSOOOOOOOONNNNNN!!!!&nbsp; urlLink   '),(9,3417138,'23,May,2004','2004-05-23 00:00:00','    urlLink    I\'m getting pretty tired, Wilson... maybe I will just take a nap...you won\'t mind, right Wilson?&nbsp; urlLink   '),(10,3417138,'23,May,2004','2004-05-23 00:00:00','    urlLink    Oh Wilson, I love you. I\'m going to hold onto you forever...&nbsp; urlLink   '),(11,3417138,'23,May,2004','2004-05-23 00:00:00','    urlLink    Kitty\'s claws are digging into Tim\'s legs... note the expression...&nbsp; urlLink   '),(12,3417138,'23,May,2004','2004-05-23 00:00:00','    urlLink    An up close and personal encounter between Pup and Kitty. A few moments after this photo, Pup\'s ass was kicked by what I can only assume to be Puss N Boot\'s cousin... &nbsp; urlLink   '),(13,3417138,'23,May,2004','2004-05-23 00:00:00','    urlLink    Where\'s Kitty?&nbsp; urlLink   '),(14,3417138,'22,May,2004','2004-05-22 00:00:00','    urlLink    Please respond to the following:                                    and be nice!                                                  i ____ nikki.  nikki is ____. if i were alone in a room with nikki, i would _______. i think nikki should _____. nikki needs ______. i want to ____________ nikki. someday nikki will ________. nikki reminds me of _______. without nikki _______. memories of nikki are ________. nikki can be __________. __________ is how i describe meeting nikki. worst thing about nikki is _________. best thing about nikki is _________. i am ________ with nikki. if i could say one thing to nikki it would be _________. nikki says _________ alot. &nbsp; urlLink   '),(15,3417138,'22,May,2004','2004-05-22 00:00:00','    I just got back from seeing Shrek 2. What a cute little film. The best part of it has got to be \"Puss NBoots\", voiced by Antonio Banderas... Here kitty, kitty. I just can\'t see many more of these though... I hear they are making a 3 and 4?? Too much Shrek!   '),(16,3417138,'22,May,2004','2004-05-22 00:00:00','    urlLink    Today is the best day ever. All of the 80\'s movies are on TV. Molly Ringwold, I never quite got what it was about her (maybe I\'m the wrong sex to get it). But more power to her... where is she now?And can someone please bring back Jake from Sixteen Candles. He\'s in hiding somewhere making furniture. He was soooo hot!&nbsp; urlLink   '),(17,3417138,'22,May,2004','2004-05-22 00:00:00','    urlLink    I have to give another Hello to my old Roomie and good friend Ashleigh J (now Ashleigh J Lake) who\'s now a NEW mommy! Congratulations girl! I have all the blackmail pictures from the old days! I will post our trip to Vegas, baby, Vegas on here soon!!&nbsp; urlLink   '),(18,3417138,'22,May,2004','2004-05-22 00:00:00','    urlLink    This is Tim in Key West, FL. This is right across the street from where we stayed on Marathon. We definitely were wastin\' away in Margaritaville. &nbsp; urlLink   '),(19,3417138,'22,May,2004','2004-05-22 00:00:00','    urlLink    I have to give a shout out to my girl Alaina. She is one of my best friends in the whole wide world!  She and I have caused sooo much drama in our life, and we loved every minute of it! &nbsp; urlLink   '),(20,3417138,'22,May,2004','2004-05-22 00:00:00','    urlLink    Me and the girls aboard the Carnival Cruise.. we apparently drank all of the alcohol. From left to right: Samatha, Wendy, and Kim. Oh yeah, that\'s me in the back!&nbsp; urlLink   '),(21,3417138,'22,May,2004','2004-05-22 00:00:00','    urlLink    As a patron of the All You Can Drink Rum Cruise, it is customary for the captain of the boat to stop driving all together and join in on the excitement...&nbsp; urlLink   '),(22,3417138,'22,May,2004','2004-05-22 00:00:00','    urlLink    Here I am on the \"All you can drink rum cruise\" outside of Nassau. Man, those were the days...&nbsp; urlLink   '),(23,3417138,'22,May,2004','2004-05-22 00:00:00','    urlLink    Once upon a time there was a young girl named Nikki who went on a cruise to the Bahamas. For 4 days and nights she did nothing but drink. She had no sleep, lots of food, and lots of booze. Here is a shot of us posing for a Carnival Cruise Brochure... I sware we will be on there one day!  From left to right: Money Bags (claims to live next to Martha Stewart), Samatha, Tony (favorite part of the trip), Nikki and Wendy. &nbsp; urlLink   '),(24,3417138,'22,May,2004','2004-05-22 00:00:00','    urlLink    When I moved into a house I actually owned, I decided it was time for me to learn how to garden and landscape. Now, I am no professional, but please look at this LOVELY gardenia! They\'re ALIVE!!&nbsp; urlLink   '),(25,3417138,'22,May,2004','2004-05-22 00:00:00','    urlLink    Hi. My name is Bailey. I want to apologize for all of my photos throughout this thing. I am my mother\'s only child (despite what she thinks) and therefore she snaps photos of me like it is the last time she\'ll ever see me. I mean, I know I\'m cute and all, but geez, mom, give it a rest. Can\'t you see I\'m trying to sleep here, and I open my eye- and there you are with a flipping camera! Argh!&nbsp; urlLink   '),(26,3417138,'22,May,2004','2004-05-22 00:00:00','    urlLink    For those of you who have never been here... this is the view (after walking 2 minutes) from my house on beautiful Lake Martin, Alabama. I encourage everyone to make it here one day!&nbsp; urlLink   '),(27,3417138,'22,May,2004','2004-05-22 00:00:00','    urlLink    This is my cousin Eric, my Uncle Howie and Eric\'s 2 sons, Avery and Chandler. What little heart breakers!!! I\'m a little lopsided in this photo, they snapped it when Chandler was standing on my foot. This was taken out front of East 74 Restaurant in Charlotte. They have the BEST GYROS ever!!!&nbsp; urlLink   '),(28,3417138,'22,May,2004','2004-05-22 00:00:00','    urlLink    The BLUE eye.. natural light.&nbsp; urlLink   '),(29,3417138,'22,May,2004','2004-05-22 00:00:00','    urlLink    There isn\'t anything cuter than little boys... my cousin&nbsp; urlLink   '),(30,3417138,'22,May,2004','2004-05-22 00:00:00','    urlLink    And the clearest eyeof them all!&nbsp; urlLink   '),(31,3417138,'22,May,2004','2004-05-22 00:00:00','    urlLink    This is my eye. I had just gotten my digital camera and was fooling around..freaky freaky&nbsp; urlLink   '),(32,3417138,'22,May,2004','2004-05-22 00:00:00','    urlLink    This is my pup in the morning doing the infamous\"death roll\". If you are laying down, she will lay on top of your neck and begin flipping around like an alligator does with their prey....&nbsp; urlLink   '),(33,3417138,'22,May,2004','2004-05-22 00:00:00','    urlLink    This picture kind of reminds me of the Great Gatsby, because they always talk about the billboard with the set of eyes watching over the city... and Jesper outside the window is just taunting my darling Bailey.. if only she\'d show the rest of her, not just her eyes!&nbsp; urlLink   '),(34,3417138,'22,May,2004','2004-05-22 00:00:00','    So I saw Troy today. I was looking so forward to it. Not for the dumbed down version of the Iliad, because it didn\'t really stick to the story. I was looking forward to it because I am like every other woman in America - Extremely Hot for Brad Pitt. Although, to me, the real hottie of this film had to be Hector, played by Eric Bana (Hoot off of Black Hawk Down).  The movie wasn\'t exceptional by any means. Actually for the first hour I thought I was watching a made for TV movie. If you compared it to Gladiator, which I thoroughly enjoyed, it would be a B movie in comparison... that is until you get to the scene before Hector and Achilles fight. And for a good 30 minutes there, its actually a pretty good flick. But other than that, I think those actors can act a lot better than they actually did.  It just seemed way to showy, and not enough depth.  The one person you actually want to die doesn\'t die (other than Agamemnon). Basically this story says, go- be a coward, and sleep with married women... you\'ll live longer   '),(35,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    Another one of my watercolor on canvas using a Qtip. This was inspired by a real skyline seen in Clanton, AL.&nbsp; urlLink   '),(36,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    Boats. Done with exterior house paint on wood. Also used the biggest P.O.S. brush I could possibly find. I like all the shapes.&nbsp; urlLink   '),(37,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    My very first painting ever. I had a thing for purple. And guys made of stone.Hmmm.&nbsp; urlLink   '),(38,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    My first mixed media. Watercolor and pen and ink. The signature is from Brandon Slay, Olympic Gold Medalist in Wrestling down in Sydney. He was from Dallas, and when I lived there, he signed it for me. &nbsp; urlLink   '),(39,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    One of my earlier attempts at painting. Watercolor, black only.  \"Pig and Bunny\"&nbsp; urlLink   '),(40,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    Another busy roadway... &nbsp; urlLink   '),(41,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    Why not Ristorante?!? I learned it in my book! I didn\'t travel overseas to go to a place named RESTAURANT!!!&nbsp; urlLink   '),(42,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    So, Peggy Guggenheim lived to be 94... however, according to the engraved stone to the right, none of her 14 children lived to be over 20 years of age, most in the single digits. An interesting observation.&nbsp; urlLink   '),(43,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    Row, row, row your boat.&nbsp; urlLink   '),(44,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    In the square, this building looms overhead while Tippi Heddren is attacked by several... just kidding. I fed the birds. They had reason to attack.&nbsp; urlLink   '),(45,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    One of the most natural shots in all of Venice. I took this when I was in a water taxi, so it gave me a really nice angle to get the gondolas with. &nbsp; urlLink   '),(46,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    One to many at ole\' Harry\'s. For the boys who are interested, on the other side of the guy behind me, there were 2 Playboy Playmates. Crazy.&nbsp; urlLink   '),(47,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    Many a thing can be said for those who travel to foreign countries, only to end up in a bar named after an American. Oh Harry, we salute you!&nbsp; urlLink   '),(48,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    A canal at sunset on the island of Murano or Burano, i forget which. One makes lace, the other glass. This is the one that makes lace.&nbsp; urlLink   '),(49,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    Hotel Excelsior on the island of Lido in Venice Italy. We were one week shy of the Venice Film Festival... why, God, why!?!?!?!&nbsp; urlLink   '),(50,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    I am Bailey, queen of the sniffers!&nbsp; urlLink   '),(51,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    My piggy bank of champions, feed me, says the piggy.&nbsp; urlLink   '),(52,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    Stop staring at me, swan!&nbsp; urlLink   '),(53,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    My name is Jesper. You killed my father. Prepare to die.&nbsp; urlLink   '),(54,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    Left to right: Jeannine, Randi, Wendy, Me, Shannon, Kim, Alaina&nbsp; urlLink   '),(55,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    My girl Randi & Me... LOVE YOU!!!&nbsp; urlLink   '),(56,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    Invitation, program, guest book & bouquets...&nbsp; urlLink   '),(57,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    These are all miscellaneous pictures of me, friends and family (old and new) on St. Patrick\'s Day... &nbsp; urlLink   '),(58,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    This is the sweetest picture I\'ve ever seen... my Timmy really loves me! &nbsp; urlLink   '),(59,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    Me and my nieces Brooke & Brittany.. aka Chee Chee and Bee Bee.&nbsp; urlLink   '),(60,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    Dad (JoeJoe) and me. Awwww&nbsp; urlLink   '),(61,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    This is me on my wedding day! March 17, 2001. &nbsp; urlLink   '),(62,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    As you may have noticed, I kind of have a thing for John Cusack. And yes, my husband knows. Oh John, I love you! &nbsp; urlLink   '),(63,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    My, how tasty a Jesper kitty would be right about now. So soft, so pretty, so easy for me ... to poop on!&nbsp; urlLink   '),(64,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    This is \"Mr. Bojangles\", my sister abandoned him, as she does with all of her animals after they grow up.&nbsp; urlLink   '),(65,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    The new and improved Meowing Kitty.&nbsp; urlLink   '),(66,3417138,'21,May,2004','2004-05-21 00:00:00','    urlLink    Bailey, the first born. My angel!&nbsp; urlLink   '),(67,3417138,'21,May,2004','2004-05-21 00:00:00','    It has become clear to me that I must have been Cinderella or Dr. Dolittle in a former life. I can not seem to have only 1 pet. I have recently become a foster parent to one Mr. Bojangles, a Beta fish with quite an attitude. Also, I am now an adoptive parent to a very demure yet spunky Jesper (pronounced Yes-Purr, like Jesper Parnevik the golfer)the stray kitty that has made a permanent residence.  And of course, our first born, Bailey Irish Queen -- who, by the way, doesn\'t quite know what to think of all of these other animals intruding on her territory. (She should just be thankful there are no mice)   '),(68,3417138,'21,June,2004','2004-06-21 00:00:00','    urlLink    This is me in the \"Rumble Seat\". That was a real bitch to get into, and even more of a bitch to get out of! FINE CAR!!&nbsp; urlLink   '),(69,3417138,'20,June,2004','2004-06-20 00:00:00','    urlLink    This is me welcoming my car in. This car is owned by Bill France, who\'s father founded NASCAR (Alaina, eat your heart out!) The car is an exact replica of the one that was the original racer in the very first NASCAR race!! The navigator and driver both work for NASCAR in Daytona, and they hooked us up with cool info!&nbsp; urlLink   '),(70,3417138,'20,June,2004','2004-06-20 00:00:00','    urlLink    this is the one I call \"Silver Bullet\". No ladies,not that kind! It was fabulous!&nbsp; urlLink   '),(71,3417138,'20,June,2004','2004-06-20 00:00:00','    urlLink    This FINE Bentley rolled in... my jaw dropped of course. Oh Ben Affleck, won\'t you buy ME one????&nbsp; urlLink   '),(72,3417138,'20,June,2004','2004-06-20 00:00:00','    urlLink    And so the Great Race begins... Well, we were a pit stop really. The annual great race features antique cars that drive from Point A to Point B across the USA in 2 weeks. You can check out more info by logging on to www.greatrace.com   I was one of the sponsors/greeters, so I will post a few pics of the AWESOME cars that cruised into OUR town!&nbsp; urlLink   '),(73,3417138,'20,June,2004','2004-06-20 00:00:00','    urlLink    My Morning Jacket rockin\' the house. They look like metal - but they rock!&nbsp; urlLink   '),(74,3417138,'02,June,2004','2004-06-02 00:00:00','    urlLink    Random shot of the City Stages crowd (behind us!)&nbsp; urlLink   '),(75,3417138,'02,June,2004','2004-06-02 00:00:00','    urlLink    Alaina & I at this year\'s City Stages 2004 in Birmingham, AL... Cool acts like Stellastarr*, My Morning Jacket, Chris Robinson, Lynard Skynard, KID ROCK, Kanye West, Switchfoot, Los Lonely Boys, etc... It rocked!&nbsp; urlLink   '),(76,3417138,'02,June,2004','2004-06-02 00:00:00','    urlLink    Chris Robinson (formerly of the Black Crowes -pre-Yoko Hudson breaking up the band). His new sound is very cool, very weed-induced. He still sounds awesome! &nbsp; urlLink   '),(77,3417138,'02,June,2004','2004-06-02 00:00:00','    urlLink    Our first look at the new Chris Robinson & (get this) the New Earth Mud. Hmmm. Not sure how he could get away with that, but after the show I realized how. Gotta love the rockin\' hippies!&nbsp; urlLink   '),(78,3417138,'02,June,2004','2004-06-02 00:00:00','    urlLink    Shawn Christensen (lead singer) of Stellastarr*. RUN, do not WALK, to go buy this KICKASS album!! What a great show!!&nbsp; urlLink   '),(79,3417138,'02,June,2004','2004-06-02 00:00:00','    urlLink    This my friends, is the rockin\' band Stellastarr* They played one HELL of a show! I must say though, people didn\'t rock out like the should have... I LOVE YOU LEAD SINGER!!&nbsp; urlLink   '),(80,3417138,'02,June,2004','2004-06-02 00:00:00','    urlLink    This is a picture of Children\'s Harbor on Lake Martin. It is on the way to the Amphitheatre. Home of Alex City Jazz Fest!&nbsp; urlLink   '),(81,3417138,'02,June,2004','2004-06-02 00:00:00','    urlLink    This is a picture of Children&nbsp; urlLink   '),(82,3417138,'02,June,2004','2004-06-02 00:00:00','    urlLink    This is a picture of the Lake Martin Amphitheatre before everyone (the crowd)got there. It overlooks Lake Martin (background). The tour buses belong to Dickey Betts of the Allman Brothers and Los Lonely Boys.&nbsp; urlLink   '),(83,3417138,'02,June,2004','2004-06-02 00:00:00','    urlLink    Everybody needs a bosom for a pillow, everybody needs a bosom!&nbsp; urlLink   '),(84,3417138,'01,June,2004','2004-06-01 00:00:00','    urlLink    I\'m getting verrrrrrrrry sleeeeeeeeepy&nbsp; urlLink   '),(85,3417138,'01,June,2004','2004-06-01 00:00:00','    urlLink    man i\'m cute!&nbsp; urlLink   '),(86,3417138,'01,June,2004','2004-06-01 00:00:00','    urlLink    bailey & newby during playtime. turns out i\'m probably allergic to kitty. i haven\'t stopped sneezing, watery eyes, itchy face since she came home. I\'ve always had outside kitties. this is my first indoor! if anyone would like a precious kitty, let me know!&nbsp; urlLink   '),(87,3417138,'01,June,2004','2004-06-01 00:00:00','    urlLink    I am still the favorite. Don\'t let these kitties fool you!&nbsp; urlLink   '),(88,3417138,'01,June,2004','2004-06-01 00:00:00','    urlLink    This is Newby and Bailey, trying to get along.&nbsp; urlLink   '),(89,3417138,'01,June,2004','2004-06-01 00:00:00','    urlLink    Hello! I\'m Newby - the newest member of the Reeves family. I was adopted today from the Animal Shelter. I have white paws and belly, and the rest of me is black, except for my chin and 1/2 of my top lip(as you can see!) My new sister Bailey keeps trying to \"nip\" me. If you ask me, she\'s a little crazy. But I just sit there and let her do it, because I\'m only 6 weeks old. Bailey asked me if I did drugs too (bc she takes herbs everyday) and I said I brought my own. Got a dime bag of cat nip in my new room. Man, my mom\'s so cool...&nbsp; urlLink   '),(90,3417138,'05,July,2004','2004-07-05 00:00:00','    urlLink    Alright I know I said I wasn\'t posting the picture of her in a dress, but look how BEAUTIFUL!! She needs to lose about 10 inches in height though, AMAZON woman!&nbsp; urlLink   '),(91,3417138,'05,July,2004','2004-07-05 00:00:00','    urlLink    This shot, is a) SO COOL. b) SO FREAKING COOL. c) ROCKS THE CASBAH.  I mean, how often do you see a light, with LITE around it.. very cool shot. I have no idea who took this photo, but I assume it was taken at the Garage. &nbsp; urlLink   '),(92,3417138,'05,July,2004','2004-07-05 00:00:00','    urlLink    My best friend Josh... SO CUTE!! Ladies, I\'m his personal PR Rep. Just ask him :) Whoa... we TOTALLY have the same eyeballs in this picture. Man, I was drunk.&nbsp; urlLink   '),(93,3417138,'05,July,2004','2004-07-05 00:00:00','    urlLink    In this picture, one of my best friends is TOTALLY checking out my boobs. Yes, they are nice, aren\'t they? &nbsp; urlLink   '),(94,3417138,'05,July,2004','2004-07-05 00:00:00','    urlLink    this is what the night looked like to me... blurry!&nbsp; urlLink   '),(95,3417138,'05,July,2004','2004-07-05 00:00:00','    urlLink    Alaina and I at the Garage.... I\'m not sure what glass of wine this is. I do know, however, that drinking this wine makes you stick out your tongue for the rest of the night&nbsp; urlLink   '),(96,3417138,'05,July,2004','2004-07-05 00:00:00','    urlLink    This is Shannon, her boyfriend Matt, and I at the Garage in downtown Mobile on Saturday night.  Fox did well for herself, I LOVE that guy! &nbsp; urlLink   '),(97,3417138,'05,July,2004','2004-07-05 00:00:00','    urlLink    This weekend was a hoot! Here is Sat. July 3. Randi went bridal gown shopping and took Alaina and I for moral support and bridesmaid dress shopping. I am afraid of putting gown pictures on here of her bc I\'m not sure which one she\'s going to use...so this head shot of miss pretty princess will have to do! &nbsp; urlLink   '),(98,3417138,'05,July,2004','2004-07-05 00:00:00','    urlLink    This weekend was a hoot... here is Saturday, July 3. Randi went bridal gown shopping and took Alaina and I with her to search for bridesmaids dresses... I am not going to post any dress pictures of Randi, bc that\'s bad luck! But here is the bride to be!&nbsp; urlLink   '),(99,3417138,'04,August,2004','2004-08-04 00:00:00','    urlLink    If you can imagine me about to pass out from sheer excitement over seeing this guy, you\'re pretty much right on the money. What a cutie! (Another shot by moi.)&nbsp; urlLink   '),(100,3417138,'01,August,2004','2004-08-01 00:00:00','    urlLink    Look at that face! Nikki, star photographer!&nbsp; urlLink   '),(101,3417138,'01,August,2004','2004-08-01 00:00:00','    urlLink    This is me actually trying to engage in conversation. I have no idea what I said. All the other pictures of Brad I took, except for this one, and the one of him snuggling me.&nbsp; urlLink   '),(102,3417138,'01,August,2004','2004-08-01 00:00:00','    urlLink    Oh my goodness. Brad is so wonderful!&nbsp; urlLink   '),(103,3417138,'01,August,2004','2004-08-01 00:00:00','    urlLink    I thought this shot was really cool... something about a man with a guitar!&nbsp; urlLink   '),(104,3417138,'01,August,2004','2004-08-01 00:00:00','    urlLink    This is another snapshot I took that was in the cabin.... man, he is great.&nbsp; urlLink   '),(105,3417138,'01,August,2004','2004-08-01 00:00:00','    urlLink    This is a picture I took of Brad Cotter when he was giving an interview.....  :)&nbsp; urlLink   '),(106,3417138,'01,August,2004','2004-08-01 00:00:00','    urlLink    Brad Cotter, how I love thee! Let me count the ways.... This is Brad Cotter and I (this past year\'s Nashville Star winner) when he was doing a closed press conference on Lake Martin. We were inside Hank Williams cabin. Oh Brad!&nbsp; urlLink   '),(107,3417138,'01,August,2004','2004-08-01 00:00:00','    I am not one for pushing my beliefs on anyone, nor do I intend to here. Those who know me, will know that I try to be as open-minded as possible. The following is an excerpt from a website that I have found, that pretty much sums up how I feel about what is going on right now in the Middle East. If you feel the same, please feel free to visit the website also, and take a look around. If you do not, well, I respect your opinion, and hope that you will do the same for me. I don\'t agree with everything on the site, but I do agree with the Pledge of Resistance. I do not want to live my life in fear. I am not against the Soldiers, I\'m against the War.   We believe that as people living in the United States it is our responsibility    to resist the injustices done by our government, in our names Not in our name will you wage endless war there can be no more deaths    no more transfusions of blood for oil Not in our name will you invade countries bomb civilians, kill more children    letting history take its course over the graves of the nameless Not in our name will you erode the very freedoms you have claimed to fight for Not by our hands will we supply weapons and funding for    the annihilation of families on foreign soil Not by our mouths will we let fear silence us Not by our hearts will we allow whole peoples or countries to be deemed evil Not by our will and Not in our name We pledge resistance We pledge alliance with those who have come under attack for voicing opposition to the war or for their religion or ethnicity We pledge to make common cause with the people of the world to bring about justice,freedom and peace Another world is possible and we pledge to make it real.    urlLink http://www.notinourname.net/index.html   '),(108,3673414,'20,June,2004','2004-06-20 00:00:00','   Just thinking about my homework remaining, and i was bloody scared. so much homework left and my RE draft due on tuesday. haven\'t even finished the experiments yet. shyt.  i must try and finish as much as possible of my hw today. hopefully kuang is not reading this. she says that we have not much homework even though she has given us piles. two essays, one project and one debate. wtf. this is bad.  i also have to do chem creative assignment. anywae, no lamenting now. let\'s talk about my overseas exchange trip to perth.   me and a guy called yue hao went to perth for the first two weeks of the june holidays. it was damn fun there. i wuld love to be in their school. it is way more slack and fun. our school\'s size is mincemeat compared to it. wth. ggs is a pro school, man. the guys i stayed with had a farm bigger than s\'pore. wth. it was extremely fun there.  anywae, bye for now guys, thanks for reading and putting up with my self-pity.  SUpersonic   '),(109,3673414,'18,June,2004','2004-06-18 00:00:00','   hello everyone,  today was a damn funny day. in the morning, i went to nus to do my science mentorship programme. got there at 8.20. had agreed to meet kin kit at 8.30. then kin kit turns up at 9.20 and i couldm\'t contact him cause his phone was off. Anywae, still finished the experiments early. They are turning out wonderfully well.  anywae, when i was coming back, i saw this stray cat (actually i saw this yesterday, but...) and it was given food by this stranger. it turned round and shat on the food. That was damn funny. i laughed all the way home. lolz.  Then, come home, homework time. philo essay. wth, i only get half an hour of recreation daily. Lolz, this sux. anywae, started my blog. no time to change anything so used a stupid format. anywae, hope can blog often.   Anywae, overall, i think today was a very fun day. tomoro must go to nus again, so shall end now. thx for reading. ttbb.  Supersonic   '),(110,3673414,'08,July,2004','2004-07-08 00:00:00','   hi guys.  got to do a ot of homework today. finished my smp, at last. had a bit of a problem with kin kit, who i feel is shirking responsibility. kk, don\'t be insulted if u read this. anywae, today, feel like blogging about classmates:  Ackerley - nice guy, a bit mugger, concerned about grades, chinese orchestra, member of shortie team (our class is lame right), cute smile, smart  Ben Chen - Tall, waterpolo, a bit hunched over, tanned, a bit quiet, extremely smart, a bit mugger as well  Ben Lim - Tall, very mugger, smart, tries to be modest but sucks at it, nice guy anywae can\'t remember his cca (oh well)  Bernard - Betawesh, medium, tanned, strong, poser, scout (in our class, that means gay), nice guy, likes to call me \'nigger\'. haha anywae. he\'s a very nice guy (betawesh.pitas.com), neer does his work on time but improving, slacker, nice smile  James - very tall, poser, scout as well, nice guy, also a bit racist in a nice way, a slacker mainly  Andrew - short, very pale, red cross, very thin, not very sporty, likes to go around disturbing oher people. sleeps the whole day in class blatantly in the sight of teachers, don\'t noe what he does at home. anywae, he\'s a nice guy, if not a bit slow and sleepy  Xian Bin - Medium height, badminton, strong, poser, nice guy but tries not to be one. mugger but tries to present himself as a slacker. racist, anti-mugger  Dominic - One of the nicest guys in the class. shortie team but almost out of it already, plump, very racist (in a nice way), mugger (3.92 for GPA), very smart actually, strng ensemble  Melvyn - shortie team leader, short, nice guy, a bit sadistic, a bit of an act cute guy. he is cute actually, mugger, maths rep, pe rep and secretary. very smart. band, plays the flute really well  Boon Leng - Mugger, very quiet, a bit of a warped sense of humour, racist (hopefully in a nice way, can\'t tell), quiet strong but doesn\'t look it. national cadet corps (i think)  Kenny - medium height, mugger (science genius), quite dark, racist (nice way), red cross  William - tall, rugger, christian (always preaching to us), loud and racist (nice way), don\'t think he can become a missionary. doesn\'t look like one. mugger (geog genius)  Melvin - cross country, nice medium height guy, medium build, sat with him for one term, tries to  be modest, sucks at it, racist (nice way)  Ivan - shortie team member, enormous mugger (top geog for level), maybe just very smart, string ensemble, very sporty, supports man u (yay!!!)  Me - lolz  Solomon - haha, very different from everyone else, tall, tries to be different in every sense from the norm, racist against people from china (bokai), reads a lot, has written his own book (quite nice actaully, about 300 A4 pages long), nice guy, though no one really likes him, roomed with him on malaysia trip  Ning Rui - Medium  height, plump, chinese orchestra, arts, mugger, tries to be racist, poser, nice guy  Xian Ying - mugger, talks cock all the time, tries to be modest and humble, but nothing he says conveys that point, mugger (english rep), good at languagy stuff. (i suck at english)  Yi Rui - last member of shortie team, tanned, softballer, very sporty, slacker, nice guy but tries to be racist, sticks with his circle of friends.  Yi Yuan - very racist, science genius, mugger, pretends not to be one, quite pale, but sporty and strong  Zhi Yong - Nice guy, choir, mugger, a bit teacher-suck-up. very serious and responsible.  Jia Hui - Tall, air rifle, mugger (pro at maths), nice smile, christian (missionary), nicely though, a bit racist sometime (nicely), prefect  Si Rui - nice guy, medium height, scout, tanned, not racist, very kind (gave me his thermometer yesterday when i forgot and refuses to accept the payment), nice smile  Bokai - colgate teeth (goes to dentist evey day, i think), from china (ostracised), mugger, very competitive, not very nice  Lawrence - nice guy, short, very skinny, ncc, racist (nicely), mugger (science genius), science club, prefect  Shun Cheng - very tall, red cross, skinyy but very fit (ran 2.4 km in 9 minutes), mugger, competitive, very nice in a different way. quiet.  Yip - actually Daniel, but 3/4 of the guys in the school call him yip. softball captain. pro guy, genius but slacker at the same time, strong, nice guy (not racist)  Yup, and that\'s for my classmates. i\'ll post about my teachers next. btw, this is written 3 days after the start of this post, and my SMP is finished, yeah. bb  Signing off, supersonic   '),(111,3673414,'05,July,2004','2004-07-05 00:00:00','   Oh man,  kicking a ball round in the house, kicked one a bit too hard and hit a light, whihc broke. oh shit, am so screwed. my parents will kill me, they just repaired that very light yesterday. oh man, god, i swear i won\'t kick a ball again. oh no!  i have tons of homework left to do. hindi is like 3 weeks overdue. my teacher will kill me. only good point is that i can use malaysia trip as excuse.   btw, i was on a malaysia trip for the school. it was damn nice, got to know a lot of guys better. it was extremely fun, but that does not change the fact that i am bloody screwed. oh man, my parents are back, bye  Mayank   '),(112,3361075,'20,May,2004','2004-05-20 00:00:00','    Good song good song...Shania Twain is the only country type person i like. lol I worked the register all by myself today at work...so fun. I love work...i don\'t see y ppl complain lol....but i\'m told it gets old fast. The cutest little girl was talkin to me when she went through the line saying \"Be extra careful not to break the eggs...cuz then the egg monster will come, and we dont want that.\" lolol MCAS was pretty annoying today in school. It wasn\'t hard...just annoying. I was supposed to hang out with Eric and Melissa tomorrow night but I switched hours with a girl from work so i gotta work 5-10. But hey, more money for me :) 2 more days till the Kiss Concert! I\'m almost more excited to take the limo there lolol. But it kinda sucks that i have to get up at 5 to be ready for 7. Thats the same time i fricken get up for school. But its all worth it to see Maroon 5...   '),(113,3361075,'19,May,2004','2004-05-19 00:00:00','    How is everyone on this lovely night of May 19th?? I\'m simply wonderful...a little cold though. Today was pretty dull. Nothing out of the ordinary happened...besides the \"dead people\" walking around school today...that kinda bummed and creeped me out at the same time. Erika came over after school and we played a little volleyball-type thing with my sister with my giant red bouncy ball. It\'s amazing how many times we let the ball roll into the street before we realized that if we moved higher into the lawn that that probably wouldn\'t happen...lol. I decided to take a shot at asking my mom if I could drive around the neighborhood with Erika. That was a big NO. haha. Later, Jenn came and picked us up and we went to Jay Gees and got some ice cream. I got vanilla soft serve with extra gummi bears on top lol...lets just say i ate all the gummi bears and none of the ice cream...cuz thats how i do yo. lol There is officially 3 days until the Kiss Concert :)It\'s gonna be so much fun. I\'m planning on meeting Maroon 5, The Calling, Jessica Simpson, Kimberly Locke, and possibly Avril Lavigne...cuz i\'m curious to see if she really is a bitch. lol I\'m pretty upset that Simple Plan isn\'t gonna be there this year...but I guess I can deal. I think thats all I have to say for today. I should get to bed now...Math MCAS in the morning...fricken awesome huh? NOT! Goodluck to everyone else who has to take them...hope u do better than I do! :) Peace out homie g dawgs yo.   So I FaLL  I dOn\'T wAnNa FeEl ThIs sMaLL  YoU kNoW i JuSt CaN\'t HaNdLe ThIs  HaNdLe ThIs At aLL  AnD i\'LL jUsT FaLL  i\'LL lEt My HeArTbEaT dRoP  I fAlTeR aS tHe MuSiC sToPs  AnD yOu WaTcH mE aS I sTaLL  AnD wOnDeR wHeN i FaLL...                     ~SoMeThInG cOrPoRaTe   '),(114,3361075,'18,May,2004','2004-05-18 00:00:00','    Hey guys. This is my very first entry...isn\'t that exciting? I\'m excited. Hopefully I\'ll be able to keep up with this one. I sucked at the other one. Well, I just walked to my uncle\'s house from the orthodontist. Guess who gets their braces off next month? :)if u said Stephanie..ur riiight! I swear to God a tear fell down my cheek and i just got THE biggest smile on my face when he ever told me. lolol Today we had the English MCAS at school which was incredibly easily. The only thing I\'m worried about is Math. I sucked and 1/2 at Math this year. Geometry sucks...bring on the Algebra! lol This week, they put me on the schedule for work on Monday, Thursday, Friday, and Saturday. And figures, the first Saturday they put me on, I can\'t work cuz I have the Kiss 108 Concert :). So now i\'m panicking to find someone to take my hours...which seems to be impossible. So I think I\'m just gonna go sometime today to tell the guy that I can\'t come in. And thats the end of my story...I\'m off to listen to some Maroon 5...gonna see them on Saturday :)   '),(115,4115327,'16,August,2004','2004-08-16 00:00:00','   Yesterday I drew this awesome pic of the character in my new manga.  I mean seriously, heâ€™s hot.  Then I (technically) went to sleep. Well, today so far has beenâ€¦annoying, I suppose.  My evil stepdad woke me up at 6:30 AM and told me to take a shower so I could help my mom with my sister and brother â€˜cuz he was going to work.  So I complained a bit and took that shower.  By then he was gone, so I went to my mom and was about to tell her straight out how I felt when she told me that all she had said was â€œI would like some helpâ€ to L. but didnâ€™t say to wake me up.  And by the time she knew it was too late (sounds kinda like one of those trials on TV if you ask me.  Lol).  So I didnâ€™t get all pissy at her too much and ate the rest of the Chex cereal and went back to sleep. Then two hours later I woke up.  I played zoo with my little sister for a little bit (gag) and then got on here.  Thatâ€™s all Iâ€™ve done so far.  So Iâ€™m gonna go.  See ya.   '),(116,4115327,'15,August,2004','2004-08-15 00:00:00','   Today I went to the grocery store with L, my stepdad.  Or rather, I got up at 10 AM and was dragged to the grocery store with a wall.  â€˜Cuz I kept asking him stuff and he ignored me.  Am I really that annoying?  Well anyway, after that we went home and I carried up practically all of the groceries.  Then I made like a recluse and stayed in my room the whole day watching anime I had taped on Adult Swim.  Thatâ€™s it.  Leave me alone.   '),(117,4115327,'14,August,2004','2004-08-14 00:00:00','   Since I can never use the computer, hereâ€™s what I did yesterday.  Well, all I remember doing is watching X-files.  And then I played .Hack//Infection.  Iâ€™m at level 32 now and Blackrose and Mistral are at level 31.  I have six Ryu books and 5 have as many items I can get in Infection.  I have a level 14 twin blade weapon and a level 18 heavy blade weapon but I canâ€™t remember what level Mistralâ€™s weapon is atâ€¦ ^_^â€       See, Iâ€™m at the end of the game, in the last dungeon (except for the extra with the super-rare twin blade weapon in it) but every time I get to the end of the dungeon I decide to use a sprite ocarina and high tail it out of there.  Iâ€™m such a wimp.  Iâ€™ve faced the last monster (canâ€™t remember name -_-â€œ) once and I lost.  So now Iâ€™m afraid that Iâ€™ll face it again and lose.  Basically itâ€™s because by the time I get to the end of the dungeon Iâ€™ve gained two levels and gotten a bunch of stuff.  And I donâ€™t want to lose it.  So I use a sprite ocarina and gate out so I can save.  *Sighâ€¦*   Well, after that trial of my bravery, I went to my room and wrote down an idea I have for a new manga and inked some Cyco.  Then I went to bed.  But I couldnâ€™t got to sleep (stupid frickinâ€™ insomnia!!!!  I think too much ^^â€) and then I remember that I hadnâ€™t put in my tape to tape .Hack//Legend of the Twilight (Bracelet?) or rewound it.  So I got up and put the tape in and then I went to go ask my parents if I could get some Tylenol (lol) and rewind my tape.  They said yes to the Tylenol but no to the tape rewinding, â€˜cuz apparently I was supposed to do it before I went to bed.  Who would have thought?  Anyways, so I went back to sleep (technically).  Then later when Iâ€™m actually starting to go to sleep, my mom comes in and tells me to rewind my tape (sneaky, sneaky!) so I get up to do it and it turns out that it was rewound.  ARGH.  I felt so stupid.  But then I went to sleep for real.  Go me.  Now for today.  Today I made a new poster from a Game Informer magazine that I had two of and moved some posters around so I could fit it on my door.  (In case youâ€™re wondering, I have 92 posters on my doors and walls.  Most of them are pictures I drew and then stuck on the wall.  I also have a whole bunch of lists and cutouts stuck on the walls.  I am so dandy.  Lol.)  I also listened to Gravitation CDs and talked to my mom about my future.  And I watched some anime I taped (Inu Yasha, Lupin III, and others).  Now Iâ€™m going to go start on that manga I wrote down the idea for.  Ciao.   '),(118,4115327,'13,August,2004','2004-08-13 00:00:00','   Since I canâ€™t be on the computer late at night, Iâ€™ll tell you what I did last night.       After we watched the X-filesâ€¦I canâ€™t really remember what we didâ€¦Oh yeah!  How could I forget?  After we watched the X-files, my mom had to go to the pharmacy.  I asked her if she could take me to Barnes and Nobles too, but she said no.  Then we got in the car ready to leave and she asked me if I had my wallet.  I said â€œNo, why would I, weâ€™re only going to the pharmacy.â€  Then I thought about it for a second and I thought, â€œMaybe sheâ€™s going to take me after allâ€ so I said â€œShould I?â€  And she got the smile on her face like â€œIâ€™m going to take you even though I know I shouldnâ€™tâ€ so I went upstairs and got my wallet (I only had 2 bucks anyway but we got some money from my account too) and we went to the pharmacy.         At the pharmacy, she went in and left me in the van with my little brother and sister.  I really wish she hadnâ€™t done that.  â€˜Cuz I was listening to my Gravitation CD and then my little brother starts crying, and I donâ€™t know whatâ€™s wrong so I go back there and try to rock him.  But that doesnâ€™t work, so I think about it for a minute and then I think that maybe the mirror (thereâ€™s a mirror above his car seat in the car) is scaring him or something.  So I turn that around and he shuts up.     A few minutes later my mom comes back.  She bought me a Cosmo Girl Magazine.  I donâ€™t like â€œpreppyâ€ magazines that much but at least it gives me something to do, if only to make fun of all the varieties of makeup these people have to have.  You know what really annoys me?  When you read about these embarrassing situations these stars have been in and make such a big deal out of but youâ€™ve been in the same situation and itâ€™s not really that bad.  Yeah.     So anyway, she takes me to Barnes and Nobles and I look for a manga (of course).  But they donâ€™t have Gravitation #6.  Now that really pisses me off, â€˜cuz I know that my mom wonâ€™t take me anywhere else.  ERGH.  But, you know me (well maybe you donâ€™t but I do), and I have to get a manga anyway, cuz I canâ€™t walk into a bookstore without buying something unless I have no money.  But sometimes even then, I manage to buy something.  *Cough* (Being the spoiled little brat that I am, I usually make someone else give me money.  And itâ€™s usually Mictoriaâ€™s sister.  But I swear Iâ€™ll buy her something at Nekocon.  I really do.  I think.)  Okay, now Iâ€™m getting off the subject.     Anyways, I have to get a manga, so I buy Snowdrop #2 â€˜cuz I read the first one and it was really good.  But I still wish they had Gravitation #6.  I was really wondering about that, because they only had #4 and #5.  And I bought #4 and #5 a couple of days ago, so now they only have #5.  Those bastards.  -_-+   Well, we go home and eat and stuff and I get ready for bed and blah blah blah.  At around 9 â€˜o clock I decide to draw some Cyco.  So I do.  And it turns out really nicely, because I was in a major drawing mood.  So I was happy.  Then I went to sleep.  Well, sorda; I had a little bit of insomnia because I was thinking too much.  And this morning my head really hurt.  It still does. Now for what I did this morning.  I took 2 Tylenols and I put my CD player in the bathroom and listened to Gravitation while I was in the shower.  And thatâ€™s about it.  Now Iâ€™m here, typing this.  Iâ€™ll be back later to tell you what else I do.  Iâ€™m going to go type the description for the Yu-Gi-Oh! cards Iâ€™m going to sell on Ebay.  Goodbye.   '),(119,4115327,'12,August,2004','2004-08-12 00:00:00','   Iâ€™m gonna talk on here while my momâ€™s on the phone.  Weâ€™re still watching the X-files.  Anyways, my little sister has this book.  You know the books for toddlers that have buttons you can press while your reading?  Well, she has one of those.  And it scares me.  Really.  It does.  When you turn it on it says â€œJust press a buttonâ€ or something like that in this really scary distorted voice.  And she was playing with it while we were watching and it scared the crap out of me.  I was like, â€œI donâ€™t want to press a button!!!!  Noooooooooo!!!!!â€  Lol.  I want to go to my grammaâ€™s soon.  (Yes, I know I spell â€œgrandmaâ€ weird.)  I hope I can go there until my mini Taekwondo session starts.  That way I can hang out with all my friends that live down there.  (BTW, my mom is talking in this creepy baby voice to my little brother about how he just took a crap.  Sorry, but I didnâ€™t really need to have that information.)  I have at least four really good friends that live near where my gramma lives.  Yeah.  My momâ€™s still on the phone.  And my little sis is peeking at me through the gate near our door.  Interesting.  Oh crap.  I donâ€™t want to start saying that again.  *Sigh* I sure talk a lot about stuff that doesnâ€™t matter.  I really need to break that habit.  Oh, speaking of breaking habits, the song â€œBreaking the Habitâ€ by Linkin Park has to have been played on the radio at least a million times a day.  I mean, I have it on CD.  I donâ€™t need to hear it every five seconds!!!  When are they coming out with a new CD anyway?  I just donâ€™t see how people still are crazy about them even though they havenâ€™t released a new CD in a really long time.  I thought people had to have a new CD come out every week to be able to stay popular!  I guess I was wrong.  *Evil laugh* Oh, weâ€™re gonna watch the X-files again.  Iâ€™ve seen and own the first through the eighth season and Iâ€™m watching them over again.  I really want the ninth season.  My mom is asking me about desserts she should make.  Rice pudding, bread pudding, or a cake, etc.  I want brownies, but she has no chocolate.  Darn it.  I canâ€™t even have a simple brownie.  Does fate hate me so much?  I dunno, but I think that if there is a god of boredom that Iâ€™m his favorite.  Well, I gtg watch the X-files.  Bai~   '),(120,4115327,'11,August,2004','2004-08-11 00:00:00','   Well, I didnâ€™t test yesterday.  But thatâ€™s okay.  Iâ€™m gonna test in the â€œmini-sessionâ€ from August 23rd to September 23rd.  I couldnâ€™t test last night because Master B. had already told another girl in my class that she couldnâ€™t test yet.  So weâ€™re gonna test on the same day.  I sparred a lot though.  I sparred against four different people:  Once against the strongest guy in my class, twice against the Master and once against the new guy.  It looks like I didnâ€™t know my own strength though!  I sparred really well and everyone said I was really good at it.  I can kick really high too!  Mwuahahaha.  (Heheh, I can brag, canâ€™t I?)  I was wondering though, if some dude attacked me on the street, would I kick his ass with my Taekwondo moves or just bite and scratch or something?  Who knows?  Anyhoo, Iâ€™ve come to call the feeling I get after Taekwondo class â€œpost-Taekwondo giddiness.â€  â€˜Cuz every time after I have a class I feel like I can do anything.  It rocks!  So all you drug-addicts out there, forget that crap!!  Take a Martial Arts class!!  (Well thereâ€™s my speech for mankind for the day  ^_^â€)  And, wouldnâ€™t you know it, my Gravitation DVD (I canâ€™t seem to get off that subject) didnâ€™t come today.  AGAIN.  Well actually, I donâ€™t know how something â€œcanâ€™t come againâ€ but whatever, Iâ€™m tired. I finished reading Gravi books 4 and 5 today.  But I donâ€™t know why Shuichi started crying at the end of book 5, â€˜cuz he got to sing with Ryuichi!!  Itâ€™s driving me insane! (Not that Iâ€™m not already, but itâ€™s yet to be proven medically. Is that a word?)  â€¦â€¦â€¦ Time for some promotional talk for my manga!  Hahahaha!  *Ahem*  Oops, gotta go eat.  Brb. Okay, Iâ€™m back, but just for a little bit, I think.  I think Iâ€™m going to watch a few episodes of the X-files with my mom.  You may think thatâ€™s weird, but itâ€™s really not.  I love the X-files.  ^_^ Well, while Iâ€™m on that subject, the guy who plays Fox Mulder (Who is he again?  Isnâ€™t he Davidâ€¦something?)â€¦Wellâ€¦I think heâ€™s kinda hot.  Actually I think itâ€™s the character that I think is hot, not Davidâ€¦whatever.  Hang on; I must attend to my baby brother.  Oh, never mind, hereâ€™s my mom.  But weâ€™re gonna watch X-files.  So my rant ends here.   '),(121,4115327,'11,August,2004','2004-08-11 00:00:00','   Iâ€™m sitting here at the computer with my Taekwondo uniform on, waiting for my mom to put my hair up.  Iâ€™m feeling very nervous.  Iâ€™m afraid that I wonâ€™t be able to execute a move, or Iâ€™ll mess up on my poomse, or something like that when I take my test for yellow belt in an hour.  Thatâ€™s if Master B. didnâ€™t forget (I donâ€™t know why he would, but thatâ€™s my imagination for you) that I was going to take my test today.  â€˜Cuz today is the last day of my second Taekwondo session.  And if I donâ€™t do it today, I donâ€™t know when Iâ€™ll work up the guts to do it again.  Iâ€™m mostly afraid that Master B. forgot about it, â€˜cuz I donâ€™t know if Iâ€™ll tell him that he forgot, â€˜cuz Iâ€™m afraid to take it.  Or something.  I dunno.  Anyway, Iâ€™m just rambling.   Iâ€™ve been working on another song in Japanese.  But since I donâ€™t really know that much Japanese, itâ€™s kinda hard.  I wrote out the basic idea in English and looked up the translation.  But thatâ€™s all.   Lately Iâ€™ve been on a poster-making frenzy.  I drew and colored a Gravitation poster of Shuichi, Hiro, and Suguru and taped it on my wall.  I wrote the lyrics to â€œSuper Driveâ€ by Bad Luck in Japanese and put that on a piece of orange construction paper and put it on my wall.  I put up all my favorite pics from Cyco on one wall and it covers half of it.  Oh and I taped my massive â€œLe Voyage de Chihiroâ€ (Spirited Away in the U.S.) poster to my ceiling over my bed.  When Iâ€™m about to go to sleep and the lights are turned off it looks like a giant square black hole about to suck me into another dimension or something.  Anyways, I also started a kind of Kare Kano poster with flower stickers.  Iâ€™m writing out what they mean (like white rose = innocence) and putting the word in Japanese beside them.   Well, like I posted earlier, I forced my mom to take me to Barnes and Nobles to buy a manga.  But I got TWO!  Go me!  I got (of course) Gravitation #4 and #5.  But I also got the enjoyment of torturing Victoria â€˜cuz she canâ€™t find #4!  Mwuahahaha!!  Oh yes, Iâ€™m on an evil roll.  Hmmâ€¦that made me think of sandwiches, which made me think of food, which made my stomach hurt.  Ouch.  My friend Sophie (as I mentionedâ€¦somewhere) is in town and I want to see her!!  But no, we canâ€™t go see her because Iâ€™m stuck at my stupid house!!  Sheâ€™s been here for almost 4 weeks and Iâ€™ve only seen her once.  ONCE!!!  How crazy is that???  Iâ€™ll tell you.  Itâ€™s insane!!  You wanna know whatâ€™s also insane?? (Well, at least horribly irritating)  My Gravitation vol. 2 DVD didnâ€™t come today!!!  ARGH!!!  And I had plenty of time to watch it!!  ARRRRGGGGGHHHHHH!!!!!!!!!  Iâ€™ve been taping anime on Adult Swim and Toonami and I have 8 tapes of it that I havenâ€™t watched yet.  I watched one yesterday, but that alone took me four hours!!  I donâ€™t think Iâ€™ll ever watch them all!!  Well, itâ€™s storming and Iâ€™m done rambling so Iâ€™m gonna go.  Farewell.   '),(122,4115327,'10,August,2004','2004-08-10 00:00:00','   Looks like my blog buddy has abandoned me.  Guess I have to post on my own now.  Damn you, Mictoria, DAMN YOU!!! (I like saying that.)            Oh, I think the hit counter on my website is schitzophrenic or something because it says that 1333742 ppl have visited my site.  That\'s crazy!!  *sigh* If only it were true...But the same number of ppl as before have signed my g/b!! Urgh! It\'s so frickin\' confusing!!            On a completely different subject, I got my SOL scores the other day and I got Pass/Advanced for everything but Math!! (stupid Math; I only passed that.) And on my English ones I got close to 600!  Woo-hoo! Go me!!           Oh, and in my Taekwondo class on Monday my master said that I\'d improved a whole lot in front of the whole class and everyone clapped.  I\'m testing for my yellow belt today.  Yay.          Something that really annoys me:  when ppl in the other apartments go in and out of their apartments nonstop.  And when the mailman\'s late delivering the mail (usually happens when I\'m expecting something -_-).          Okay, now for what I\'m doing today.  I\'m gonna check the mail to see if my Gravitation DVD (vol. 2) is here yet, I\'m gonna make a few posters and crap, I\'m going to go with my mom to her stupid appointment and listen to Gravitation music on the way, I\'m gonna force her to take me to Barnes and Nobles to get a manga, and I\'m gonna test for my yellow belt.  Woo.          That\'s all I have to say.  My fingers hurt.  Sayonara.   '),(123,4115327,'10,August,2004','2004-08-10 00:00:00','   Me, Mictoria again, telling you that I now have my own blog.       urlLink www.beautifullydisfigured.blogspot.com       go there if you wish to learn about me and my wonderful(pathetic) life(!)      -V*ctor*a   '),(124,4115327,'08,August,2004','2004-08-08 00:00:00','   I have to return to the real world today...also known as home.  I had a \"vacation\" at my grandma\'s for a week and a half.  But now I must go home.  Damn it all.  Goodbye.   '),(125,4115327,'07,August,2004','2004-08-07 00:00:00','   The other day my friend Victoria and I typed a letter to my friend Joe cuz he wasn\'t on AIM.  Here\'s what we wrote:     Charlotte: HIIIIIIIYYYYYYYYYAAAAAAAAA!!!!!!!!      Victoria: My dearest joe........      Charlotte: ????      Victoria: But you\'re into hiiiiiiim. Isn\'t that splendifferous?      Charlotte: I don\'t understand you....Oh well, you fat piece of lard. Not. Anyways, Nee, oikakate oikakate shiroi kaze Nee, ............ai ni nate hana hirotagai Nee....something..........      Victoria: And you say you don\'t understand me??? Yes, we are curently obsessed with Gravitation......just a little. MUAHAHAHAHAHAHA!!!!!!!!!!!!      Charlotte: Didja know that marijuana in japanese is mariwana??      Victoria: That\'s very ironic......and a little strange......but also twisted......... but anyway, Charlotte, do you know who else I want on my \"Want to Hug\" List?      Charlotte: IIIIIIIIII DOOOOOOOOOOOO!!!!!!!!!!!!!! SHUICHI-CHAN!!!!!!!!!!!!!! OoOoOoOo guy-kissing!!!      Victoria: Yes, he is sexi, don\'t you agree Moe? His outfits are erm............unique. Especially that pink bondage-y one on your poster-thing you\'re drawing...........      Charlotte: I wanna SIIIIIIIIIIIIIINNNNNNNNNNGGGGGGGGGG Itsu mo.....forget it....no, wanna sell your soul................Nee, oikakate oikakate.....-_-\"      Victoria: Please don\'t......I\'ll have to go all opera on your ass..............Memory, all alone in the moonlight, I can smile on the old days...........      Charlotte: tojikomeru!!!!!!!!!!!!!!!!!!!!! (shut up)      Victoria: Bruhaha! Yes, I know over 50 Andrew Lloyd Webber songs!!!! And I have an amazing range!!!!!!!!!!!!!!! MUAHAHAHAHAHAHA!      Charlotte: Yes, but you cannot do that high C.....wait....High-C?? Hahahahahahahahahaaa!!!!!!!! *cough*      V    ictoria: Actually i can, and that really wasn\'t very funny...........You\'re losing your touch...........      Charlotte: Am not!!!!!!! Seguchi-san has a gay hat. Do you have a gay hat? Hmmm...I think I am losing my touch...*cough*      Victoria: Don\'t joke on Seguchi!! His hat is sexi!!!!! lol      Charlotte: More like......something bad that rhymes with sexy...      Victoria: What? Rexi?? That\'s my character!!!!!      Charlotte: Don\'t get all emotional on me!! Random Gravitation track: It might have been just a kiss for you, but it wasn\'t for me!!!! (Me: Oh...) ....*cries* I\'m losing my mind, too!!!!!!!!!      Victoria: Charlotte, you really have some issues to work out.................      Charlotte: I do work out....just not issues...I\'m not good at that...(yes, goo at that!!!!!!!!) Oh! I just thought of another use for tape!!! You can...eat it!!!! Or you can eat Shuichi, like Ryuichi does!!! (Su!!)      Victoria: I like Ryuichi......he\'s so cuuuuute..... even though he\'s 30...........hey, Tatsuha thinks so and he\'s 16..and a priest...... WHY are priests in manga always gay? Is it a trend? Like germans and gangs?      Charlotte: Yes, yes it is...But also, gay people are hot.....lol...even though that\'s kinda wrong....and ironic, cuz they don\'t like girls!!! But Shindo-kun is sooo cuuttteeee!!!! Mwuahahahahaa!!!      Victoria: Yes, my perspective on guys..... All the hot ones are either gay or taken..........      Charlotte: lol, that is so true....Ah, wait, they\'re either, gay, taken, or not real....^____^\"      Victoria: Yes, you are so obsessed. But anyway, Moe have you read the manga....it\'s erm..............yes, that\'s it \"erm...........\" The drawing style is completely different and the plotline is slightly differently paced, but it\'s really, REALLY, REALLY, good!!!!!      Charlotte: Yes, \"really REALLY, REALLY\" good!!!!! (squared) Hey, chase me, chase me, white breeze!!! toki o koetaiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiii!!!!!!!!!!!!! !!!!!!!!!!!!!!!!!!!!!!!!!! lol And that was my long dream. If only that was the punch line!!! Hey, Hiro, ask me if something good happened???      Victoria: My name\'s not Hiro, and I\'d rather not.................. It\'s Victoria remember? Vic-TOR-ee-Ah...........      Charlotte: I know you retard I wasn\'t talking to you!!!! Wait...who was I talking to???? ARGH!!!!!      Victoria: Hello? Sarcasm much? Hey, you know, maybe you\'re schizo......... Great, You\'re schizo and I have OCD. We\'re all crazy...... What does Moe have? depression? PMS?      Charlotte: You know, I was watching this commercial the other day, and it listed the symptoms for schizophrenia, and....I have them alll!!!!!!! It\'s creepy!!!! Oh, and it\'s not PMS, it\'s a \"depression session!!\" (insert evil laugh here)      Victoria: Yeah, I was doing this research paper on OCD and I looked up a test to see if you have OCD, and my score was 17 out of 18: You should definitely seek help for Obsessive-Compulsive Disorder... lovely eh?      Charlotte: Oh, yes....But am I really schizophrenic???????????? The only thing that didn\'t apply to me was seeing things....and if you see things, you don\'t even know, right???????? HELP ME!!!!!!!!      Victoria: Great, I\'m friends with a schizo. No, not a fluffy dog you idiot. NO I DON\'T WANT ANY DAMN OPTIONS FROM HP!!!!!!!!!!!!!!!!!!!!!      Charlotte: Huh?? Aranji?? Oh. I see how it is...you hate me now that you know that I\'m schizo, right?????? DAMN YOU!!!!!!!!!!!!!!! (chukushou!!!)      Victoria: I like K from gravi. He\'s trigger-happy, and American. Btw, didn\'t you know that all of us Americans have long blonde hair and big blue eyes? at least, that\'s how it is for anime and manga.... You know most of us Americans are brunettes? Some of us have pink hair........      Charlotte: Nooooo!!!!!!!!! Shu-chan\'s crying!!!!!!!!! DAMN, you YUKI!!!!!!!!!!!!!!! DAMN YOU!!!!!!!!!!!!!!!!!!! I want to hug him now...*huggle*      Victoria: Yeah, well we\'re kinda going on and on about nothing now, so we\'d better cut this short.......      Charlotte: Wow, we typed a lot............................................................................. ....................................................*brain-dead*................ ............Uh.................I know! I\'ll make up an imaginary friend!!! I\'ll name him Kyuchi-chan!! And I\'ll hug him and shcueeze him, and call him...Kyuchi-chan!!!! Baka baka baka baka ni!!!!!!      Victoria: Yeah, we\'ll go now......shut up Charlotte.      Charlotte: Sayonara, Moe-kun. Come on Kyuchi, let\'s go die your hair pink. *walks off*      Victoria: Yes and if you\'re feeling bored call me!!!! xxx-xxxx!!!! And may you bake a thousand pies.........Shut UP brain!!!!!!!!!!!!! Guten Tag and Auf Weidersein.      /\\/\\/\\/\\/\\/\\/\\/\\/\\/\\/\\/\\/\\/\\/\\/\\/______________________________(and we\'re dead)   '),(126,4115327,'06,August,2004','2004-08-06 00:00:00','   This is   Mictoria   again.   Aren\'t you excited?  Don\'t deny it, I know you are.   Okay, so Maybe you aren\'t.  Well you should be.    Why? I don\'t know/care/give a shit.  Pick one.  Uno.  Eins.      Don\'t you love me?  Okay, don\'t answer that.     Well, you know all about Marlit, so here\'s some stuff about me:   I\'m unusually tall.  I have  pink streaks  in my hair.  I like using big words and massive amounts of sarcasm.  Deal with it.  I have two cats and a hamster that looks like he was genetically fused with a guinea pig and a mop.  I am obsessed with Eerie Queerie and Gravitation.  And guys who are unusually pale......with dark hair.....   My manga collection spans over 70 with about 20-30 different series.   And I don\'t like crunchy peanut butter.  Pet peeve.      I shall be shutting up now.  Now scamper off and get on with your puny insignificant lives.   Muahahahahaha......   Aren\'t you glad I posted Marlit?   '),(127,4115327,'05,August,2004','2004-08-05 00:00:00','   I started Reti\'s song.  Here it is...   I\'m surrounded by people but I feel so alone They treat me like a steeple so high; but I\'m really just low   Why can\'t they see that I\'m just like everyone else? I\'d rather be your knight in shining armor than your stuck-up prince   But what is the most strange of all for you I don\'t want to be normal   So can I be more to you than everyone else While still being normal? It doesn\'t seem possible   I\'m surrounded by people but I feel so alone They treat me like a steeple so high; but I\'m really just low   Gotta go, Bai~   '),(128,4115327,'05,August,2004','2004-08-05 00:00:00','   I\'ve been trying to write a song for Reti to sing in Cyco.  I\'ve got the gist of what I want him to sing about but I can\'t seem to put it into words.  Oh well...I\'m at my Aunt\'s house right now.  My gramma and me came to visit for a while...so I\'m using her computer.  I wish I could work on Cyco right now.  Earlier, I was talking to my gramma about southern accents.  It annoys me when people say puddin\' instead of pudding.  lol. Anyway, she usually drops the \"g\" on most of her \"-ing\" words.  I discovered that I drop the \"g\" on \"going\".  Crap. Well this is an interesting topic...well, we left the house about 3 hours ago (I think) and went to the library (I was looking for a thesaurus so I could work on Reti\'s song) and my gramma got 4 free books.  I got two, Men are from Mars, Women are from Venus and Don\'t sweat the small stuff.  (Self-help books) lol. Then we came here. I have nothing else to say, I suppose I\'ll go.  Sayonara~   '),(129,4115327,'05,August,2004','2004-08-05 00:00:00','   Most of you (more likely all of you) probably have no idea what Cyco is.  It\'s my webcomic about me and my friends\' lives.  It\'s a comedy, of course.  Anyway, to the point...my website is being stupid, so I\'ll just post the new pics here.  Here they are:            Okay, well that\'s it.  bai-bai~   '),(130,4115327,'04,August,2004','2004-08-04 00:00:00','   Well...I\'m at Victoria\'s house again...Who would have guessed??  Not me, that\'s for sure.  (Well, maybe...)  You know what?  I hate DVD burners...and DVDRs...and that song that\'s playing on the TV...and...the world...oh yes, the world...I watched Gravitation AGAIN last night....woo...Required Weekly Test of the Emergency Alert System that always frickin starts right in the middle of Angelic Layer...darn them...*BEEEEEEEEEEEEEEEEP*  Ah yes, beep.  Well, I stayed up til midnight last night...aren\'t I special?  Nothing much is going on in my life...I\'m just bored, as usual...*sigh*  Sayonara...   '),(131,4115327,'04,August,2004','2004-08-04 00:00:00','   This is Victoria/Mictoria/Homie-G/That tall girl over there (Depending on who you are.)  Charlotte asked me to put my song on here, so I guess I shall, even though I\'ll probably regret it later.  It\'s more like poetry to me, but, whatever.  Tell me what you think, eh? E-mail is MOONCHILD8304500@aol.com.  Untitled (for now)  I don\'t know what these lyrics are about I just have to say something I need to get free I need to get out  Help me Don\'t let me drown Save me I don\'t want to end up like this  I don\'t know what to say to you I just want you to notice I need to run away I need to be loved  Rescue me Don\'t let me down Fix me I don\'t like being alone  Every time I look around It\'s all the same All falls down To one single thing Do i really care? Do I really need you?  Help me  Don\'t let me drown Save me  I never want to feel like this  (end song)  Yeah, that\'s it.  Might be crap, but who gives a shit? This is Mictoria signing off....   '),(132,4115327,'03,August,2004','2004-08-03 00:00:00','   I uploaded the Kyuchi and Aruhi pics that I drew in paint!  Here they are:         They don\'t have any backgrounds, though.  Oh well...*sigh*  Bai~   '),(133,4115327,'03,August,2004','2004-08-03 00:00:00','   Well, I finished my song! (I think.) If you ask me, it needs a little something more, but I\'m afraid that if I add any more verses it will be too long. I was planning for it to be a rather fast song, though. So maybe even if I add something it won\'t matter. Anyways, here it is, if you care:    I feel so helpless I need a purpose I want to cure this boredom I feel  and it\'s hard to do this by myself I think I need you it\'s difficult to tell  I try to escape this world when I can but there\'s only so much I can take and I give up again  I don\'t know what to do to get it through to you I can\'t seem to come out and say how I think about you every day  I feel so helpless I need a purpose I want to cure this emptiness I feel  When I find my other will it help free This breathtaking flutter that causes anxiety?  Everywhere I go I feel so alone and I want to know if you are the one  Who can cure my existence  I feel so helpless I need a purpose I want to cure this longing I feel  It needs something!!! Grr...stupid brain!! On a completely different note, my best friend from England is coming over in an hour to spend the night. She\'s bringing her GBA, too, so we can battle our Pokemon on Sapphire! Lol, I feel so...\"out.\" Oh well...I was never \"in\" in the first place so it doesn\'t really matter. I drew a picture in Paint of the character in my manga (Aruhi from The Legend of the Era-mixer Machine, or E.M.M.) and will probably put it up on my website some time before I go back to my house (I\'m at my gramma\'s for the week). I also drew a pic in Paint of a character that I thought up at my friend Victoria\'s house a few days ago. He was supposed to be my imaginary friend cuz I was lonely, but I wanted to draw him, too. His name is Kyuchi and he has blue hair and pink eyes and a \"Victoria-proclaimed\" hour-glass figure. Lol. He\'s in a band called Oxy-moron, but I can\'t decide whether to make him the drummer or one of the guitarists. He\'s gonna be in Cyco, my manga, eventually (in other words, the next time I update, which will be...most likely never ^_^\"). *sigh...* So I\'ll probably put that up eventually, too. I\'ve always wanted to be in a rockish type band called Psykobabble (Victoria named it) and be a lead singer. So I\'ll put that in Cyco, too. Maybe Kyuchi and I can be band mates...lol The reason I\'m typing this is cuz I have nothing else to do and my friend will be here in 30 min. Or she could already be here by the time I post this in my blog. Who knows... I just finished reading I.N.V.U. #2 and I want Sey (?) to wind up with Siho. But that\'s just my opinion. Sorda. Well, I don\'t really have anything to say so that\'s all, I guess. Bai~   '),(134,4115327,'03,August,2004','2004-08-03 00:00:00','   Yes, I slept until 12:00 p.m. again...and I went to bed at 10:00 p.m.!  *sigh* Maybe it was that talk \'til 5 a.m. on tuesday about bishounen that took it out of me...Anyways, I borrowed a lot of manga from my friends (Inu Yasha 4-9, I.N.V.U. 2 and 3, Model #2, Fruits Basket #4, and Eerie Queerie #3) and read DNAngel #3 at their house.  I also bought two manga at Books-a-Million yesterday:  Samurai Deeper KYO #7 and The Candidate for Goddess #1, cuz it\'s by the same guy who did DNAngel.  Altogether, on my shelves at home, I have 105 graphic novels that belong to me.  That\'s around $1050 that I\'ve spent on manga!! *faints* Wow...and I have around 11 different anime DVDs (that\'s about $270 bucks I\'ve spent on anime)! Not to mention all the different posters and other merchandise I have...I probably could have bought a car! (Not that I could drive it, but whatever.)  Oh well.  That\'s enough calculating for me.  See you...whenever.   '),(135,4115327,'03,August,2004','2004-08-03 00:00:00','   Since the stupid post-thingy didn\'t want to put my first sentence, here it is: Well, as I\'m having an obsession for Gravitation, I looked for a bunch of pics from the anime. I found a whole lot of cute ones ^_^ especially ones of Ryuichi and Shuichi. Go to    urlLink   http://cloverpetals.com/addiction/gallery.html      and see if you can find pics that match these phrases, cuz those are the ones I got:   '),(136,4115327,'02,August,2004','2004-08-02 00:00:00','   Well, as I\'m having an obsession for Gravitation, I looked for a bunch of pics from the anime. I found a whole lot of cute ones ^_^ especially ones of Ryuichi and Shuichi. Go to  urlLink http://cloverpetals.com/addiction/gallery.html  and see if you can find pics that match these phrases, cuz those are the ones I got:    (\"Hey, Hiro, ask me if something good happened.\" \"Hey, Shuichi, did something good happen?\" \"It\'s a se-cr-et!!\") , (\"And that was the long dream I had.\" \"Hai!\" \"If only that was the punchline\"),  (\"Su...\" \"Su?\" \"Su..\" \"Su?\" \"Suki desu!\") , (Me:  Oh... ), and I found a lot more, too. ^_^        Oh, and last night at was at my friend\'s house and we were talking about the perfect bishounen, and here\'s what I said: ...let\'s see...mysterious and aloof, yet not boring and sometimes laughable, has straight hair but isn\'t too straight, (you know, no bowl hair cuts, lol) and is sometimes messy. And then, uh, body-type: skinny but not a stick with some muscles (particulary in the chest area *cough*) and has to have a really nice personality. And the clothes-style would be either baggy, or kinda like japanese high-school boy uniforms (*sigh*). And not a prep. Definitely goth/punk.       oh yeah, the hair color can\'t be boring, like blonde (ugh!!!) or a flat brown. like maybe...black with orange streaks or pink, etc. and can have piercings except for on their nose unless it\'s a ring. uh...and can have lots of ear rings but not only one in each ear. oh, they can have tatoos too but not like all over just on the arms or the back yeah... but I can make a few exceptions of course lol       Oh and his name should be uncommon...Felix, Isaac, Piers or Ivan (from Golden Sun, lol)       I was also writing a song the other day and here are the lyrics I have so far: (not much and not very good, either...-_-)       I feel so helpless I need a purpose I want to cure this boredom I feel       and it\'s hard to do this by myself I think I need you but it\'s difficult to tell       I try to escape this world when I can but there\'s only so much I can take and I remember everything again       I don\'t know what to do to get it through to you I can\'t seem to come out and say how I think about you every day       I feel so helpless I need a purpose I need to cure this anxiety I feel        I\'ve been reading a bunch of webcomics lately, too. They are:  urlLink Megatokyo ,  urlLink Modern Sorcerer Shishiro ,  urlLink The Way to Your Heart ,  urlLink Path of a Fallen Angel ,  urlLink Arcana , and a Fruits Basket doujinshi called  urlLink Rainy Day Man .       That\'s it for now. Talk laterz. ^_______^   '),(137,4115327,'02,August,2004','2004-08-02 00:00:00','   I\'m always bored.  I guess that\'s why I\'m doing this blog.  But it\'s mostly because I want to talk about anime.  ^_^\"  I got the first Gravitation DVD yesterday and my aunt\'s ordering me the second one.  I\'m a little (lot) obsessed...  I have my own site that has my webcomics and other drawings on there, but I can\'t really just talk about anything on there.   I\'m also making an attempt at writing a song but I can\'t think of anything that rhymes...oh well.  My site URL is  urlLink www.narutokun.co.nr  if you want to look.  I also got the OVA soundtrack for Gravi the other day and my aunt\'s ordering me the TV tracks for my B-Day (September 28th).  I can\'t really think of anything else to write right now...so Bai~   '),(138,3549541,'05,June,2004','2004-06-05 00:00:00','   Just testing to see how the layout looks..   '),(139,3549541,'24,June,2004','2004-06-24 00:00:00','   err quite happy todae though. bought the book that i wanted: å¼ å°å¨´ï¼š ç›¸é€¢ã€‚  å…¨éƒ½æ˜¯ç²¾é«“ä¹‹ä¸€ã€‚æ˜¯æ•£æ–‡ã€‚ bz reading my book. lol. sooo wun update much. yeahz.  sch was sian.  Social Studies Mathematics (if u dun count Joshua poking at u full time behind u) Geography ( free period ) Chinese ( teacher din come but still got hmwk )   haix. goin to go back sch le. dotz. i feel so damn sick of it le. how am i gonna survive the darn o levels lor. imagine this: mock prelim. [okay. still okay] prelim [haiyo. so sian. study again]  o level [godz. I DUN WAN STUDY LE] haiyo. siao. å®¶ wrote some chinese stuff actually. but wun post it here. å®¶ä¸‘ä¸å¯å¤–ä¼ ï¼ å“ˆå“ˆã€‚bb.   '),(140,3549541,'22,June,2004','2004-06-22 00:00:00','   finally have the time to blog le. but it\'ll be a short one lah. literally have no time nowadays. so much hmwk to do and revision to do. argh. totally in no mood now. one thing i muz say though. jiesin im sorry. happy? u started it anyway. nvm.  in bad mood lah. wun blog. bye.   '),(141,3549541,'17,June,2004','2004-06-17 00:00:00','   heyy ppl. wun be updating anytime soon after this. have some prob wif my com.. have to wait for very long to fix lor.. lolz. soo too bad. haiyo. i go game le larh. byex.  stupid msn ppl. stupid com. stupid c drive.   '),(142,3549541,'15,June,2004','2004-06-15 00:00:00','   heyy. need help.. i rewashed everything off my com the other day rite.. then after that i reinstalled MSN version 6.2. then when i logged in, the screen wif all my contacts came out but after dat it became stuck. as in cannot move. the mouse can still move but it\'s in the bz mode le as in the hourglass mode. how? i need to use MSN. its my life.. haixxxx. *cries.   '),(143,3549541,'11,June,2004','2004-06-11 00:00:00','   heyy ppl.nv blog on monday lor. haha com kena washed. everything.. weepz. anyway. monday was boring larh. hmm. shall blog abt sundae.. haha. had yuhua concert to go to. hmm met nich at 2.00 at serangoon mrt. as usual. i was late wahaha. sooo met him then to my usual hangout : Plaza Singapura!! yeah.. haha.. after that met zakiah den went to st andrew cathedral to pick saga seeds. lolz. this time a lot a lot lor. with the help wif zak and nich.. haha. nich is expert in picking lor. he can find the saga seeds very fast leh. haha shant tell u all for u larh. but i guess band members shuld noe. nono. there\'s nth between us. haha family ties ba.. plus ive been always very bad to her ma.. lolz. the other ppl ah.. haiyo.  haha then after that went to take lots and loads and loads of photos. haha. siao le. stupid \"model\" photo gort passed around yesterdae. lolx. nort i wan de hor.  den after that went to eat.. then went to suntec city.. lolz. heyy there gort machine can wash photos lehz.. haha feel so high tech lor. haha.but the dumb machine ate 50 cents lor. haix. din realise the sign saying \"no change given\".. dotz.  almost late for the concert lor. we were like stuck at suntec at 7. 15? haha.. den we ran all the way there. tiring sia. like juz ate liddat.. but very fun larh. we were basically breaking like all of the traffic rules lor.. haha. anyway. yuhua is good lor. damn gd. esp the alumni band. haha. maybe we can go and revive the alumni band lor.. lolz. we gort josh, yijun and jingyuan leh.. haha and many others.. marcus , benson.. lol.. maybe can fight?!? wahaha. no larh..   haiyo.. i expected their godzilla eats las vegas to be gd le larh.. but din expect it to be soooo gd lor. haix. their showmanship still like superb lor. haha. i luv miss chan sia.. wooHOO!~ their sooo called encore was like.. can be considered as one full song le lor.. haha saw diana again. dun noe still rmb me or nort lor.. but i saw my best frenr again.. argh wilson sorrie i din call u to go but u came anyway larh.. lolx.. yeahz. gd rite.  those who din go shuld consider killing themselves leh.. lol, joanne and edo went too. or rather edo wif his girl-frenr. lolzokay okay enuff. then after that went to eat and take more photos. wasnt realli happy towards the end. and vaness shuld noe what i mean.. ill be giving them a piece of my mind larh.. haha. dun wanna blog abt yesterdae larh. boring. sundae was fun and very..  okay abt todae. nort gonna go out. supposed to go to outing todae but the sec3s postponed it. haix... nvm nvm.. jiayoux. im numb already. juz hope that it will go well. gd luck ppl.  okay okay.. am quite happy now larh.. lolx. yeah yeah yeah yeahz. haha. i no love le larh. give all away le..lolz bye   '),(144,3549541,'11,June,2004','2004-06-11 00:00:00','   anyway. read this. gort this from my sec1 junior:  (i tink its sick though, she sec1 onli leh) A husband, tired of his wife asking how she looks, buys her a full  length mirror. This does little to help. As now she just stands in front  of it, looking at herself, asking him how she looks.   One day, fresh out of the shower, she is yet again in front of the  mirror. Now complaining that her breasts are too small!  Uncharacteristically, the husband said. \"If you want your breasts to  grow, then every day take a piece of toilet paper and rub it between  your breast for a few seconds.\"   Willing to try anything, the wife fetches a piece of toilet paper,  stands in front of the mirror, rubbing it between her breasts.   \"How long will this take?\" she asks.   \"They\'ll grow larger over a period of years,\" he replies.   The wife stops,\" Why do you think rubbing this piece of toilet paper  between my breast everyday will make my breasts grow over the years?\"  she asks.   The husband shrugs.\" WHY NOT, IT WORKED FOR YOUR ASS, DIDN\'T IT!\"    ------------------------------------------------------------------- okay nvm on wif todae. woah. todae is one day which im back  urlLink home  early (dat means before 5 reach  urlLink home , assuming i spend abt 8 hours outside everyday) in the morning went to amk mrt there and reached there abt say 6.50? damn tiring, seemed as though my head weighed a ton. lucky my dad fetch. lucky me. lolz had to brief michelle and yong chun larh for their bm and dm understudy.. but i was supposed to meet them at say 7.20? haha. ended up that i din bring my headphones out then did some paperwork for band till then. met terry and chuan jian and nice them came to talk to me lor. haha. they were supposed to meet their frenrs at seven but when they reached at say 7.10? nobody lor. haha. and they were supposed to hve breakfast. lolz. wait i thot michelle lost the outing minutes lor. turned out as she typed it out and lost the original copy. hai wo rewrite everything according to my  urlLink memory . dotz. after briefing them, went up to the plaform. as usual, i was the earliest. saw pauline on the way though. lolz. next came vincent then after a while, matthew appeared out of thinair(idinseehim) b4 peizhi,lihuan joined us. lolz. and the latest was the organiser lor:xiwen. haha. saw some realli funny sights. and disgusting ones. saw ck. lolz. nvm.. haha.i was like omg ck.. then eeks. haha. nvm. then went to somerset to hve my seminar.. orh it was boring. seriously. i culdnt keep myself awake. oh yah. did i mention that anthony wanted to kill me. coz i told him to meet in sch or sth to brief his QMs, but i din say im goin. so i din go and he... haix. he waited for me i guess. wahhaa. he came in pe attire larh and attracted quite a lot of laughter too larh. nvm larh. ya gort ya dear daryl to sit wif u ma.. lolz. percussion.zzz.  nvm. the lecture was boring lor xcept for the break. haha. i enjoyed the  urlLink food  lor. weilong, me and vincent were like eating second and third helpings. haha. but we din finish all the fd on the table larh. we\'re nort pigs. tokin abt pigs: rmbed someone who suaned me that i can nv be a pig. i thot she was being nice or wat before she said that i dun even deserve to be a pig. juz becoz i sleep a lot.. she oso whatx. but nvm. it was onli suaning... lol  after that decided to go off wif my class. dun say i always nv go wif u all larh.. haha. lucky jiesin nort there. I solemnly proclaim: me and jiesin will hate each other for ever. lolz. her  urlLink dog  gort a castration yesterdae for like wat, prevention against prostrate  urlLink cancer . haha. errr. sooo we went to long john  urlLink silver  and saw yr, juliana they all there. victor immediately backed off. followed by joshua who tried to act blur and walk away. for god\'s sake, pls dun act lor. nvm. nort up to me to say anything lest anyone say me againz. okay larh. but ate there anyway.  after that walked then took the lift up to the arcade there. lost to matthew in one game. duno what to call that larh. but matt, jy and qy fought for racing oso larh. after that went up to bowling. rmb ppl i dun bowl sooo slacked there wif some ppl larh. matt treated us to cookies. famous amos! woohoo. nice sia. he gave me the big big ones. =) after that, receieved some soo called gd  urlLink news  larh.. okay nort very gd but im happy anyway... then after they finished me and matthew went off earlier and headed  urlLink home . the others went to pool. oh yah. i boasted to some of them abt the  urlLink movies  larh.. lolz recent de 50 1st dates: watched scooby doo: watched troy: watched van helsing: watched harry potter: watched shrek 2: watched e day after tml: watched spiderman: gonna watch garfield: gonna watch mean gals: gonna watch  lol yeah. think xiwen gort quite irritated by me. lol oh yah on the way back wif matt, we took to marina bay then took back to amk lor. haha. crapped lots wif him or rather bitched abt some ppl larh. at least he agrees wif me. but i think im evil in some way or another. im nort even involved but im like sooo interested..okay nvm. u all dun understand. errrr. supposed to meet ketsu todae and yesterdae but hes like ultra bz. haha. anyway ketsu is yuan ru\'s dream lover. from cat high lor. been suaning her. u noe the nice testimonial she gave me in frenster. it is becoz she wanted his picture. dotz. nvm. meeting him becoz he wants a score larh.. hes a clarinetist too though play db most of the  urlLink time  ba. one more thing shuld i go for choir concert? haha shuhui\'s paying $2 out of $12 for me leh. lol both of us leaders gort connection sia.shuld i or shuld i nort.   okay. everybodys like basically eradicated from singapore all go sarawak le lor. haha. less noisy though. not a lot of  urlLink sms . culd feel the effect instantly. used to receive 50 over smses a day but now haha.. nort even 25. yeah~haiyo. sundae still gort yuhua concert., maybe meeting nickky earlier again. vanessa u goin rite? u havent pay  urlLink money  yet.. band fund help u pay leh.. err. anyway sorrie ppl onli had 36 tix coz e  urlLink money  wasnt paid up in  urlLink time . see what happens when u all dun pay up.. grrr. i hope i can go. for all who duno, there\'s a band outing next tuesday and im officially a facilitator.lol  hmmm. i think crapped a lot todae le leh. i wun ever take 70 bus again. waited one whole entire hour for the stupid bus yesterdae at suntec. took 133  urlLink home  instead. grrr. angry. but nvm.. haix tonite have to study le. mum\'s been complaining. haiyo.. grandmama calling me eat le.. i go le hor. think gort salmon fish. haha.nice~ byex tc ppl.    love was when i loved u. one true  urlLink time . i hold u.   '),(145,3549541,'10,June,2004','2004-06-10 00:00:00','   grrrrrrrrrrrr. i typed an ultra long entry and the stupid thing gort problem. dun wan blog le. i hate my com. stupid virus.   '),(146,3549541,'09,June,2004','2004-06-09 00:00:00','   hey ppl. juz came back from woodlands. haiyo...todae long day sia. in the morning went to the sch to meet the sec 3s.. haiyo.. took very bery long/ but nvm. enjoyed it. haix. but kinda kena scolded by mr tan and mrs lim larh..  guess i deserved it.  then after that. it was raining realli heavily ... and very. but the air con could nort be on lor. or else shuang~! haha. but i couldnt realli hear myself speak.. sooo i shut all the doors and the windows in the temp classroom. sorrie to the sec3s, i tormented ya all.. haha. but u all culd sit leh. haha  hmm.. after that. decided to go wifthe sec3s to woodlands. actually wanted to go plaza sing wif zak de.. but the others din wan sooo suit them ba.. saw meisen oso.. hot pink sia>~  sooo we took to woodlands.. had a wonderous meal at Burger king anyway.. lol. luffed lots and stuff there.. reali cracked us up dere man. anyway. suhtyng, michelle, zakiah and me were dere larh.. saw some pretty funny stuff.  picture a boy sitting alone at the corner. he wanted to move away but wanted to leave his stuff on the table sooo he started talking to HIMSELF: \"this one gort mosquito one.. dun touch this one ah\" dotz.  and there was this scene when suhtyng was eating haha. she was like eating eating.. then suddenly she looked in another direction and looked back and she said \"uh oh\" and stared at the piece of thing which fell off her burger. haha.. shes soo damn cute lor haha.. no wonder shes my daughter. haiyo.. btw: someone wanted to be my great granddaughter leh.haha.yingchern was like complaining how come nv put her in da family.. dotz.  okay after that decided to follow them around. needed to buy stuff too.. haha. but felt quite paiseh to follow them to go around searching for clothes lor. esp in the ladies beach wear section dere. i was like the onli guy there. and all the other shops oso lor.  but nvm after that zakiah went off. leaving me wif suhtyng and michelle. haha. crapped lots wif them. i was quite picky wif the stuff i bought larh. haha. tormented them totally. suytyng was complaining full time abt me lor. wahhaa.. but finally gort it.. now i need onli one more \"ingredient\" haha.  oso played wif suhtyng there at the metro there.. lol.. it was like the t shirts hung up dere were too high for EVEN me to reach larh. and even joshua i wuld say.. so i used the normal bamboo stick to bring it down lor.. haha.was very very funny lor.. luves suhtyng\'s new t-shirt.. had words like this: language was invented becoz of the need to complain, complain and complain.. haha. then the back had onli a few words : COMPLAIN. haha. and its in bright orange. haha vanessa ur fav colour. haha.   broke todae anyway.. ttreated them to a bit of burger king.. den bought a lot of stuff. and went to take neoprint. haha. treated 20 cents more.(i noe im petty)  orhh!!! gort one more thing.. in the music shop in the causeway pt there gort some very very nice small versions of the instruments. u noe the small saxophone was soo nice lor.. if huiming saw it, she wuld totally scream lor.. haha. but its like thiry six dollars. cant buy for anybody lor.. haha. sorrie peepx culdnt lend u all money. nvm. next time then get chee hao to buy for us.. wahhaa. if dun wan buy for us ah.. then we use someone** ahemx to threaten u. haha. looks like everybody noes huh? lolz.  i wan a new bk lor.. it costs thirteen sth.. damn nice chinese san wen book lor.. by zhang xiao man. hk writer. she makes hell lot of sense lor. haha. even mich and suythng agreed wif me todae lor.. haha. okay nvm. shall get my mum to get it tml. wahaha. grr im feeling bored. and a bit sad. maybe its becoz of some stuff. duno larh. juz bored ba.. wanna chat wif some ppl on the phone lor.. haha. juz chatted to jy for a while.. anybody noe how to remove firewall. he needs help.. hmm. tml gonna watch hp wif yijun they all.. im excited lor. haha. though theres a music exchange tml oso.. haix guess im nort goin to go ba.. though i already brought the clarinets back.. grr. so heavy. haha. the tubist oso noe how to bring their mouthpiece back onli lor. sooo gd. haix. but other than cck, and sji oso gort regent and bedok north. u all ah.. what schools oso duno.. for dose who duno.. they\'re playing \"holding the truth\" sorrie to jeremy cant be dere to help ya.. maybe next time come and find me lor. friday nort gonna go band oso. maybe onli morning go.. haix very little ppl.. kkies. nvm.. i think i crap a lot lor.. haha. haiyo.. i go le larh. enuff for the day. bye ppl ~ byex.  is loving someone or being loved more fortunate?lolz.   '),(147,3549541,'08,June,2004','2004-06-08 00:00:00','   grr. im pissed . very.no mood to blog le lar. my mum accused me of sth. he culd jolly well hve done it on his own and then, it comes out as i didnt help.. haha. yah. the smaller brother gets everything watx. yah.  Mum says: im addicted to msn whatx. i talk crap wat.  i onli noe how to go out whatx. i refuse to help my brother. she is disappoined wif me. i dun even care abt my brother. i make this whole family a unpeaceful place.  FINE. if so.  kill me then.  juz becoz he needs some help, i hve to help. i was bz helO~? i get scolded. its juz sending an attachment in hotmail. i pointed u in the direction, freak. so u hve to go crying to mum for help. becoz of wat. \"big brother didnt help u\"  and mum comes out to scold me. and brother sits innocently on the floor watching the fight. grr. im pissed. very very very. cant he take some venturing. i duno dreamweaver.  my brother needs to use. mum call me do. i duno kena scold. simple as dat. USELESS BRAT I AM HOR.  so basic stuff i do : eat (waste money) sleeeep (waste air-con) go out (take money) side wif frenrs (dun help family) make this family un-warm   so tell me shuld i go or nort? since frenrs and family dun understand me le. i tried to help. i teach u maths, dear brother. i offered u help. U??? wahaha. u dun even bother to cover the blankets for me when its cold. u dun even close the windows when it\'s raining, late late at nite. u dun even bother to wash the plates, clear the table, make breakfast for urself? or even do a project without mama\'s help. ya sec1 ma hor? dun need do larh. need help for everything... haha. i dun think i was dat pathetic in sec1. who will it be when ya all need the photos in the digitial cam for ya project. it\'ll be me. nvm.  and so in the end what am i? worthless breed.  haha. smiling? that\'ll hve to come after ive finished being pissed. sorrie if this doesnt sound like the normal wenjie. the normal me was gone a long time ago. eradicated since my sec3 life started. no quote for todae. bye.   '),(148,3549541,'07,June,2004','2004-06-07 00:00:00','   haha. todae is one rare day when im alone at home lor. haha. my grandparents went out. soooo i have the morning alone. haha. wanna blog earlier de but couldnt wake up. lol. haix. nth to blog in the morning leh haha. haix. been watching tv~ hmmm had a wonderful breakfast i made for myself lor.haha haha recipe: 1) Spread garlic spread on two pieces of bread. 2) Toast. Leave to cool for a while.This is essential because if we dun do this, the oil from the garlic spread will mix wif the peanut butter and form a ugly suspension. 3) Spread peanut butter on the two pieces of garlic bread. 4) Add a piece of cheese on each piece of bread. 5) Add another layer of peanut butter onto the cheese. 6) Put the two pieces of bread together and dig in!   haha. i think it sounds disgusting to ya all.. like cheese is sour and peanut butter is sweet. but u peepx go try larh. last time home econs chuanjian they all say nort nice. but when they tried it, they almost finished everything haha. contradicting. see dun believe me.. haha try it, its nice~  arghZ i wanna go orchard. i juz wanna get out of the hse lor butttttt my mum dun allow lor. haha. grrr. haiyo.  godz.im bored.bored.bored.bored.bored.  anyway. gonna watch hp on thur wif sec4 peeps. lol. haix. dats all for now ba.give a longer post later ba.. haha.  see ya.~    *the greatest barrier on earth isn\'t life and death. e greatest barrier is that im standing in front of u and u duno that i love u*   '),(149,3549541,'07,June,2004','2004-06-07 00:00:00','   so its finally out eh. after the confession from someone. haha. so okay i noe its nort meant to be after all. i guessed it all along  ANYWAY . nvm. i shant bother at all.  Looking on the brighter side, I went SEntosa wif the sec3s todae. haha. though i din realli enjoy it but since it was sunny and all, brighter side than. Brighter as in literally. okay nvm. lamez. Okay shall wanna clear up sth here. Cheryl has nv had me as a bf larh. its juz playing ya? haha. secrets actually smsed me hor??? haha. nvm.. i shall luv ya too~ yeah shes my GF too.. haha.  to secrets: she wun get jealous becoz she doesnt exist.ya. dun worry. haha. and its nort even her anymore. yeah.  my stupid com has a virus lor. haix. cant read my mail can im too lazy to on the other com. haix.idiotic com. icant sleepwell.if u actually called last nite sleep. nvm. basically was soo tired but kena bothered by some stuff. meaningless stuff.STUPID> i havent watched harry potter anyway.nobody realli waited for me larh.. haha. must be the put off from my normal ego. but wateva. heard it sux sooo i duno. shuld i watch or shuld i nort watch.? orhhh i rmbed. who wuld wan to watch wif pathetic me rite.. wahaha.nvm.i rather watch wif juz cept hes nort free.but nvm the sec4s gort me wif them soo guess i\'ll go wif them instead of the normal gang. haha. thanx man~ ni men hai shi the best!  orhh one thing. tml. shuld i go for SJI prac.. have ta get tix from mr yea. haix. early morning. i feel like sleeping plus i dun feel like goin there. its juz like another amkps. where everybody\'s so close but ya\'re nort inside. nobody bothers to intro. okay nvm. since i\'ve decided to be antisocial, it doesnt reali matter.  lee hom gort new album lor. nort sold in s\'pore. haix.another ppl flying to  urlLink japan  lately. help me get kkies. i prefer his jap songs to his chin songs.haha. tokin abt discs, sorrie to nich. i din noe it will spoil oso. haha.  share something wif ya all:(to a certain someone oso)  ??????,???????????????? ???????, ??????????????? ????????,??????????????? ??,???????,???????????????????? ??,?????????????????? ?????????????? ?????????????? duno what to say le larh. i tink its nice. haha. gort it off a bk. haix yah.. nth much to say le lor. decided nort to go SJI le larh.. haha. i duno.    *the greatest barrier on earth isn\'t life and death. e greatest barrier is that im standing in front of u and u duno that i love u*   '),(150,3549541,'06,June,2004','2004-06-06 00:00:00','   okay. juz went out wif siva, nich, yuzhen and huiming. the concert was okay larh other than edwin *wif his classmates nort huiming ah* making a lot of noise,huiming luffing all thru out, siva and nich talking, lack of attention from the crowd and me being sad, i\'ll say it was quite a success. in any way, cheryl u do rawk.. haha u still luv me rite. althou i\'m your 5th bf rite. lol. anyway. cheryl (u noe that cheryl tay from amb, yesh her, u\'re nort seeing things) says she gort her first bf in pri 4,haha. then still gort a longest relationship which lasted for 2 years leh. ai qing chang pao. then now juggling between edwin and me. so saaaad. haix.  quite irritated durin the concert actually.  due to a few   reasons  : 1) lack of money. cheryl i bought a $20 worth of flowers for u kkies betta appreciate it. 2) recieved an sms from someone. or rather was waiting for dear SOMEONE to reply me. appartantly dat person din bother larh. 3) lamez but the music actually moved me somehow. haha. to irritating mode (one piece but forgot what le) strange what music can do nowadays. 4) nobody to go out wif after the concert. okay this is realli bad i noe.. nort blaming nich and siva but haix. nvm..  and sorrie to Juz who tried to get me out. haha. but u enjoyed urself anyway rite. without me.  and heard that some ppl will go on somewhere and enjoy themselves and basically if i din ask, i\'ll be left out rite. without me. again  and juz the other day. i was left out for sth. and they said they enjoyed it, it rawked. okay okay soo i waited for them and after that. they went off. without me again. this is getting too qiao~ to be ignored.  ç¬¬ä¸€æ¬¡ï¼Œæˆ‘å‘Šè¯‰è‡ªå·±ï¼Œ å¯èƒ½ä½ ä»¬å¤ªå¿™äº†ï¼Œå¿˜è®°äº†æˆ‘ï¼Œæˆ‘ä¸æ€ªä½ ä»¬ã€‚ ç¬¬äºŒæ¬¡ï¼Œæˆ‘å‘Šè¯‰è‡ªå·±ï¼Œ å¯èƒ½å½“æ—¶çš„æƒ…å†µä¸å…è®¸ä½ ä»¬è¿™æ ·ä½œï¼Œæˆ‘ä¸æ€ªä½ ä»¬ã€‚ ç¬¬ä¸‰æ¬¡ï¼Œæˆ‘å‘Šè¯‰è‡ªå·±ï¼Œ å½“æœ‹å‹çš„ï¼Œä¸è¦å¤ªè®¡è¾ƒã€‚ ç¬¬å››æ¬¡ï¼Œæˆ‘å‘Šè¯‰è‡ªå·±ï¼Œ æˆ‘å¯ä»¥å¤§æ–¹åœ°ä½“ä¼šã€‚ ç¬¬äº”æ¬¡ï¼Œæˆ‘å‘Šè¯‰è‡ªå·±ï¼Œ æˆ‘å¯ä»¥å°½é‡åŽ»æ˜Žç™½ ã€‚ ç¬¬å…­æ¬¡ï¼Œæˆ‘å‘Šè¯‰è‡ªå·±ï¼Œ æ˜¯æˆ‘å¤ªæ•æ„Ÿäº†ã€‚ ç¬¬ä¸ƒæ¬¡ï¼Œå¯¹ä¸èµ·ï¼Œ æˆ‘å·²ç»æ‰¾ä¸åˆ°ä»»ä½•çš„å€Ÿå£äº†ã€‚ ä½ ä»¬çš„ç¡®åœ¨take me for granted.  haix.. sorrie for the sudden change to chinese. i luv chin anyway. juz feeling kinda sad ba. with or without me, the world will still go on.  å³ä½¿æˆ‘ä¸åœ¨äº†ï¼Œ å¤ªé˜³ä¼šä¾ç„¶åœ°ç…§è€€ç€å¤§åœ°ï¼Œ é£Žä¼šä¾æ—§åœ°å¹éåƒå±±ä¸‡é‡Œï¼Œ èŠ±å„¿ä¾ç„¶ä¼šå¼€å¾—ç¿çƒ‚æ— æ¯”ã€‚ å› ä¸ºæˆ‘çš„ç¦»åŽ»å¹¶æ²¡æœ‰ä»»ä½•ä»·å€¼ã€‚ å°±ç®—æˆ‘èµ°äº†ï¼Œ ä½ ä¹Ÿæ„Ÿå—ä¸åˆ°æˆ‘çš„ç¦»åŽ»ã€‚  feelin kinda poetic anyway. nvm. all crappz.(juz as the sitename) hmm. ya larh but that\'s realli what i feel. haha. think i type quite long le larh. shall go off now. byex ~     *the greatest barrier on earth isn\'t life and death. e greatest barrier is that im standing in front of u and u duno that i love u*   '),(151,3549541,'05,June,2004','2004-06-05 00:00:00','   okay. i woke up at 8.30 again. haix.okay cancel that. nvm. haha. b4 i forget. MY BDAE IS ACTUALLY COMING LEH> rmb kkies.  haiyo. supposed to be doin hmwk but lied to my mum that i need ta check mail from mr tan. which is quite true but he din reply me. hmm i supposed to go offline now, but since u noe \'supposed\' soo heckx.  hey. thanx to vaness hor.. haha for the testimonial. thanx for announcing to da ENTIRE WORLD abt the colour and lee hom thing. haha.hmmmmmmmmmmmm.im booorreed.  later shall go out by my boi boi to hougang mall 1st lor.. haha. this guy ah. always so paiseh one but can crap soo much. he has his father character woOhOO~ but i hve nv been to hougang mall b4 lor and the bus goin there runs on a stupid loop service and goes onli like one way. sooo i basically have to go the whole round b4 reaching the first stop when its actually soo near. daMn 21 stops u noe.  err. btw anybody gort any nice band songs pls pass to me hor. in the mood for new ones. other than persis and jap series larh lol.  hmmm. other than all these. kinda hope tmr can go out wif someone. haha. u all duno de larh keep guessing.  to yuzhen and nich: u two hve no mama yet larh. but haha. papa\'s working h@rd.~!  errr. share a joke here. haha. contributed by jinjun actually: a racist one though but since it\'s racist, naturally my entire class noes. yeah~  what do u say when u agree wif a indian? > > > > > > > > > > > > > > > > > > > give up/       ERRR. u hve a POINT there.. dose ppl who get it gd. haha. dose ppl who dun. too bad too sad.OKAY i saddist.  haiya.. think nth more to add le lor.. haha....... shall go off to chat on msn le.. bye & smilez :P     *the greatest barrier on earth isn\'t life and death. e greatest barrier is that im standing in front of u and u duno that i love u*   '),(152,3549541,'05,June,2004','2004-06-05 00:00:00','   orhh helo ppl.. wahah. i finally gort my own blog lor. after quite some  urlLink time . but i muz realli thank my frenr for dat. lazy me wuld have nv started on a blog lor. hey.. thanx hor. are u secrets or vicky.. haha.  sidenote: the author is nort a good writer hor. pardon him for his lack of vocab and language.. ya larh. as if anybody will read.. haha kkies.  hmmm.. todae totally slacked at home. i mean i\'m supposed to study larh. okay in the morning was actually quite hardworking. gd start man.. but that was all.. i woke up at 8.30 (for the 1st time this yr) yah.. studied a bit of chemistry (10 minutes) then went off to watch TV. haha. i actually watched peach gal on channel 8 lor.. at 12. ppl catch it. lee hom\'s song \' aidejiushini \' as main theme song lehx! wOOhoO~! okay lame  -____-\"  hmm..in the afternoon, haha, slacked thru again. haiyo. when im like supposed to do physics.. okay. i noe im realli lazy but juz cant help it.   err. recently heard a quote in a workshop larh.. juz wanna share it here wif myself (ibetnobodywillreadmyblog)haha..  here goes: \"Everybody dies but not everybody lives.\"  okay larh.. it\'s like realli true larh.. but somehow my class wasnt too inspired by it. okay i admit im innocent wahaha.. innocent pig! haha.. betta than meisen le. she ah. onli noe how to famei. (sorrie.inoe u dun use net that\'s why attack u wahha)  hmmm.realised a lot of ppl are sad recently.. so ppl CHEER UP!!! rmb what i say: smile big big..  steps to make urself happy: 1) pull urself away from the tissue box. 2) walk to the bathroom and to the mirror (any reflecting surface may do larh) 3) kkies. i noe it sounds stupid. but it works for me larh.. ya... now  JUX SMILE !! keep the smile. think of the happy stuff. look into the mirror. see the real u.  kkies. lame~ nvm larh. follow this.. i noe this doesnt realli work.. but promise me.. nv take ur life away okay? sorrieZ  NVM. crapz a lot le.. tml oso goin out .. again haha.. wif my son and gal. haha. oh ya did i mention my  urlLink family tree  again? okay wait: i hve one father. i hve  NO  mother. (she ran away from us lor) i hve 3 sisters. i hve one nephew. i hve three daughters. i hve one son. haha.  hmmm. maybe dats for all todae. need ta send  urlLink mail  to mr tan le. abt band camp. hope he approves~  byex and take carez smilez.      *the greatest barrier on earth isn\'t life and death. e greatest barrier is that im standing in front of u and u duno that i love u*   '),(153,3549541,'02,July,2004','2004-07-02 00:00:00','   Hi! This is Wen Jie. Yup.  No really! Believe me. This is Wen Jie typing. =D  If I am an imposter, I\'ll kill myself -- uh I mean Wen Jie will kill me -- for impersonating me -- or I should say, for impersonating Wen Jie. So you can see from the above sentense, I am really Wen Jie. Yes. No kidding. Who else will type so weirdly?  I know I know. You all don\'t believe me because I\'m typing weirdly now but that\'s because I\'m making an effort to type proper English with capitalisation and all. So ya. This is Wen Jie. Wen Jie is the only way you should address me by. For I will not respond to any other name you call. That is because,  I am Wen Jie.   '),(154,3750679,'26,June,2004','2004-06-26 00:00:00','    This is my blog to the worldâ€¦  â€¦that never blogged to me (apologies to Dickinson, check it out at www.bartleby.com).  What a fitting first posting to my blog!  Clever.  Blogs are the ultimate in one-way communication, brought to us by the infinite self-publishing capabilities of the web.  Children in Beijing and the editorial staff at Al Jazeera can find out with the click of a mouse what I have to say about anything.  Everything!  If it crosses my mind, it will cross theirs, yours, everybodyâ€™s.  The good folks at the Independent School Technology Letter have come to yours truly, bloggingman, to get everybody hip to blogs.  Can do.  In fact, we decided to blog our way through the topic, for real.  Is there any other way to communicate?  From me to you.  I am giving Sherry and Will privileges to post to the blog as we go along, so they can help me keep it straight and blog it into shape.  What, you ask, is a blog?  Where you been?  If I spelled it â€˜blog, would that give you a clue?  Right, itâ€™s short for web log.  â€œLogâ€: noun, a record of performance, events, or day-to-day activities (check it out at www.m-w.com), and you know what â€œwebâ€ is.  Iâ€™m keeping track of things in my log, which I publish on the web, hence this is my web log, or blog.    Of course, all blogs exist in the blogosphere.  â€œBlogosphere (alternate: blogsphere) is the collective term encompassing all weblogs. Weblogs are heavily interconnected; bloggers read other blogs, link to them, and reference them in their own writing. Because of this, the interconnected blogs have grown their own cultureâ€ (check it out at: en.wikipedia.org).    Blogs have become incredibly popular, even in schools, which is where I work (identity concealed).  In fact, a guy at NECC this year said â€œyou must blog.â€  If I mustâ€¦  So here I go.  Kinda catchy, â€œyou must blog it.â€  When a problem comes along  You must blog it  Before a thought sits out too long  You must blog it  When something\'s going wrong  You must blog it   now blog it  into shape  shape it up  get straight  go forward  move ahead  try to detect it  it\'s not too late  to blog it  blog it good   Thatâ€™s the ticket!  Blog it good!   '),(155,3750679,'25,June,2004','2004-06-25 00:00:00','    Today is June 25th and the first day for this blog for  urlLink ISTL   '),(156,3600011,'10,June,2004','2004-06-10 00:00:00','    Hi everyone!!  My name is Kevin and I live in a suburb of Kansas City called Shawnee, KS.  I\'m starting this blog off by telling about what life is like living with exotic birds.  So, grab a soda and a snack, sit back, and enjoy!!  Approximately 5 years ago, my wife purchased a Cockatiel for me.  I had been interested in getting a bird so she purchased one for me.  I was thrilled to say the least.  It was just a baby and we finished hand feeding it.  That was an experience in itself.  We named him Sammy.  After a couple of months we decided to purchase a second one.  We were really enjoying the 2 Cockatiels.    At the same time, my dad had purchased 2 Cockatiels.  After having them for a year or so, he decided that he didn\'t want them anymore.  So, quess what?  Right!!  We brought them into our home.  Talk about a lot of noise!!  But it was a noise that we enjoyed.  We named our second tiel Daisy, thinking that it was a girl.  Well, we guessed wrong!!  It is a boy.  LOL  The 2 tiels that we took from my dad are Cassie (yes, an actual female) and Simon.  Sammy, Daisy, and Simon are are normal grey tiels.  Cassie is what is called a Lutino tiel.  A very bueatiful yellowish color.   In the summer of 2002 my wife and daughter and I went to south Florida for a week.  We decided that while we were there we wanted to look for a Congo African Grey.  We had read a lot about the Greys and had decided that we wanted to purchase one.  We found a large aviary in south Miami.  After visiting with them, we found a Grey that we wanted to buy.  One problem.  We were having some major heat waves and no airline would take the bird because of the heat.  So, we left Florida with no bird! :*(  A few days after returning home, I came home from work one day to find a baby Congo African Grey.  My wife and daughter had gone out that day and purchased one from a local breeder that we knew.  Wow!  Talk about being excited!!!  We named her Khiya (pronounced kie-ya)  No meaning to that word.  Just something we came up with.  Grey\'s have the intelligence of a 5 yr old child.  They have a vocabulary of approximately 1500 words.  They are absolutely amazing creatures.  She talks up a storm and is right now going thru the terrible 2\'s just like a kid does.  But she\'s extemely funny and entertaining!  Never a dull moment.  Ok, as Paul Harvey would say, now for the rest of the story.    Since then, we have obtained, either through adoptions or purchases (mostly adoptions), a Blue/Gold Macaw (Mr. Bones), an Umbrella Cockatoo (Lady), a Jenday Conure (Sunny), a Greencheek Conure (Carrie), a Green Quaker (Hawkeye), a Catalina Macaw (Cleo, a hybrid between a Blue/Gold & a Scarlet), and our latest purchase, a 5 month old Timneh African Grey (Libby).  Talk about a crazy family!!!  But, we love each one very much and get hours of entertainment from all of them.  The only bird out of that group that we actually purchased from a breeder was the Timneh. The rest we have adopted.  Sometimes when people go out and purchase a bird because they\'re cute and make pretty noises, they don\'t understand that this is a life long committment.  Our larger birds, including the 2 Greys, can live to be 80+ yrs!!!  That is longer than most of us will be around after we\'ve purchased a bird like that.  In our case, our birds will go to our daughter and son-in-law (she\'s now married).    Our club (www.beaknwings.org) gets calls weekly to come and rescue birds from people who either can\'t take care of anymore or who choose not to take care of anymore.  Some birds come to us in an abused state.  The club recently aquired a Cockatoo that had had all of its bones broken (literally, every bone) and stuffed in a shoe box.  The shoe box was taken to a pet store in southern Missouri and left.  After $1500, the majority of the bones were fixed.  The wings, however, could not be fixed properly.  They were left intact but will never be right.  Now, a lot of you may say, why not just put the bird down rather than spend that kind of money to repair it.  Well, would you spend that kind of money to fix up your son / daughter if they were in an accident?  I\'ll leave that at that.  Our Timneh is a special needs bird of sorts.  When she was a baby, her parents bit off all of her toes on both feet!!  She barely has enough left to grasp on to anything.  But, despite this handicap, she has adapted very well.  I just found her actually perching on the same wood perch that my Congo uses.  I was totally amazed.  She has over come her handicap. We picked her up in March from a breeder friend of ours.  She was concerned that she would not get a good home because of the lack of toes and knew that we would provide for her just as we do for the other 12 birds (oh, and our 2 dogs and 3 cats).   Ok, so what is it like with that many birds in the house?  CHAOTIC!!!!!  It is never a dull moment.  Cages are all over the house.  Actually all but 5 of the birds are in our living room, or what use to be our living room!!!  Our day starts off at 5:45a.  The birds wake up at 6:30a.  Their diet consists of a special pellet called Zupreem.  It is formulated with vitamins and, in the case of the variety we use, fruits.  We dehydrate vegetables every week and mix in with the pellets.  They also get fresh, filtered water each day.  We use newspaper the line the bottom of the cages so that all we have to do is pull out the dirty paper, throw it away, and put in fresh paper.  That kind of sums things up (very briefly).  I will write more later.  I hope that you what I\'ve written so far.  I would love to get feedback and any questions that you might have.  You can email me at dejavukc@hotmail.com.  Have a wonderful weekend!!!  Kevin   '),(157,3469665,'28,May,2004','2004-05-28 00:00:00','   Ok, so yesterday was better. There is just something about getting out of work and home by noon. It always brings a insipid smile to my face. So anyway, I got home, stripped down and hopped in bed. I\'ve got one of those beds with a pillowtop mattress...I added a feather bed on top of that, covered it with 300 count Egyptian cotton sheets and then put four more down comforters on top of that. It makes for some happy heavenly sleeping. Anyway, I got up after a solid three hour nap, fought with my family on the phone and then went to my Brazilian ju-jitsu class. Talk about an ass whooping. Good grief. My upper body didn\'t stop quivering until after dinner. It was an odd sensation. Anyway, after my class I went to my favorite grocery store, ordered a yum yum sandwich and headed back to my apartment where I sat cross-legged on the floor and ate. My best friend was there too, we were too tired to do much more than gossip feebly. I was going to meet up with the guys to go mountain biking, but I hurt too much. I wanted to go, but then again, I\'m still a bit skittish about my lack of mountain biking skills.  Today I have duty...for those of you who are unaware, duty is a cruel and unusual punishment the military dreamed up to ruin perfectly good weekends for Armed Forces members. Someone has to be on the ship at all times and we each take turns on a rotating basis enjoying this special form of torture. So I\'m here, sitting, hoping I can sneak over to my room and take a nap, wishing I could go home or at least get outside for a bit. Maybe I\'ll eat lunch outside.   This is what I hate...you get this bullshit line about joining the Navy and living an adventure, but in reality you\'re going to spend months on end staring at the same people, watching the same four walls, hoping that maybe if you\'re lucky they\'ll let you at some port for a day or so. Even better, they\'re repeating the same ports over and over and over again so you\'re really not going anywhere new. By the time I get off work, (when we\'re in home port) I\'m so desperate to see light and have a normal life that I end up staying up almost all night, just so I don\'t miss a second of my freedom. Only 3 years and 2 months and 5 days to go until I\'m free again. *sigh* I guess I\'d better go find somewhere to take a nap.   '),(158,3469665,'27,May,2004','2004-05-27 00:00:00','   So I\'m sitting here at work, watching people slog back and forth and for the life of me I can\'t figure out how I wound up working at a place so mundane. I would give anything to be outside right now. It\'s just hot enough with just the right amount of wind. I wouldn\'t mind hitting the trails for a run, give my legs a stretch. Maybe I\'d grab my bike and work on jumping stuff. I\'m supposed to go mountain biking with a friend this Tuesday, but the gap between my experience and his is vastly embarrassing. He\'s sweet enough that he pretends not to notice (even when I randomly throw myself into the bushes...damn those clipless pedals!) At least he\'s a good sport!  The trail I\'ve been riding is beautiful. I\'ve been running/riding it for months now, watching it turn from a dull, dead ashy brown to a vibrant, dripping green. I feel like I\'m watching a nephew grow up. The down side of all this fecund, viral growth is that the bugs are out in full force. Though I\'ve eaten my fair share, I am apparently not slowing their prolific procreation habits at all! They bounce off my sunglasses, my shirt, dive into my hair and get stuck to my sweat/mud covered legs. *shrug* At least I\'d be outside.   '),(159,3469665,'08,June,2004','2004-06-08 00:00:00','   Two weekends ago, I haphazardly stuffed my pack, grabbed my best friend J and headed for the woods. I picked a section of the Appalachian Trail for our little three day hike and J reluctantly acquiesced. We drove for about five glorious hours, through some of the most adorable towns I\'ve ever seen. These were the kind of places that didn\'t even have a McDonald\'s and had names like Cuckoo and Gordonville. We finally made it to the mountains, laced up our boots, threw on our packs and hit the trail. Little did I know, this backpacking trip was going to change my life.  Right after we got on the trail, we met a guy named M. He was headed to Rock Spring Hut, only about four miles down the trail. I decided that we weren\'t going there because I didn\'t want to be deluged with other hikers. After a few hours of hiking I changed my mind. You see, I\'ve always wanted to hike the Appalachian Trail and I figured that it would be a good idea to hit up this hut so I could meet other thru-hikers and nose around through their packs-to get a good idea of what I would need to buy and all. Anyway, we turned down the trail to the hut and set up camp on one of the pre-prepared camping flats (which actually wasn\'t all that flat...) Before all this, let me tell you about the trail. I\'m in love with backpacking. I don\'t know what it is, but it has a lot to do with being outside, breathing truly clean air and waking up to birds singing. It\'s idealistic I know, but I wish I could live out there. Anyway, this hike was the same sort of thing. Huge rock formations, narrow trails, the legendary white blazes on tall beautiful trees. There were open meadows, clean streams and wildflowers everywhere...what more could I ask for? I sped ahead of J, only because I enjoyed the solitude and because I can haul ass when I want to. I always sat on a boulder and waited for her though...to her credit, she kept up well.  Back at the Rock Spring Hut...so we are setting up camp and talking away with TJ and WL, two guys who were setting up camp too and we all decided to eat together. Everyone pooled together their food and shared hot dogs, avocado dip, chocolate and all sorts of other yummies. One of the best things about backpacking a lot is that you can eat whatever you want and as much as you want-you\'re burning 3-6,000 calories per day anyway.  The other thru-hikers were hilarious, a product of long periods of loneliness combined with mile after mile after mile of solid physical exertion made them talkative, friendly and informative. I got lots of great advice and even better, received a trail name from TJ (if you didn\'t know, thru-hikers use pseudonyms while hiking to sign the various trail books along the way.) After dinner and a lot of beer (I shared M\'s vodka and gatorade,) we all went to our tents. Right before I drifted off, I heard, in a loud voice, \"YEAH, I WAS WONDERING IF I COULD GET A TAXI TO COME PICK ME UP TOMORROW...\" I started laughing. Being the smart-ass that I am, I couldn\'t resist shouting up the hill at M \"AND COULD YOU GET HIM TO BRING A PIZZA? EXTRA CHEESE?\" At this point, the other hikers started laughing as well. Who brings a cell phone and calls a taxi from the AT? Fortunately M had a great sense of humor and we all had a good chuckle over it.  The next morning as we were packing up, one of the thru-hikers, WL asked if he could join J and I on our little lap around WhiteOak Canyon. We agreed to drop him off further up the trail so he wouldn\'t be behind schedule. We started off, adding another hiker to our party, JB which had our group topping off at four hikers. We found the WhiteOak trail and headed down it, WL and myself darting ahead. After a while we came upon our first waterfall (WhiteOak has a concentration of waterfalls which is why I originally picked the trail-I probably should have equated the waterfalls with immense changes in elevation, but I wasn\'t thinking that clearly!) Anyway, we got in our bathing suits and hopped into the chilly/freezing water for a little refreshment. It was great...there is nothing in this world that beats a dip in a waterfall, followed by some sun-drying on a rock, followed by some lunch cooked over a camp stove, followed by some more hiking. I was in heaven...honestly I could have stayed there forever. Anyway, we repeated our little swimming trip two more times, each time, hurriedly stripping down and jumping (against our common sense,) into a pool of frigid water. At this point WL (S) and I had discovered that we had tons of things in common including a trip to DC planned for the next weekend. We agreed to swap phone numbers and hang out when we got there.  The hike was a tinsy bit harder than I thought and J was ready to beat me senseless when we finally got out of the canyon. She was even more angry when I shot down the suggestion that we stay in a hotel for our next night. I did concede a hot dinner at the trail restaurant, but I could tell she wasn\'t happy with me.  Anyway we hit the trail again, this time rushing to find a place to camp before dark. After nearly running down the trail, S found us a spot and we put up our tents by headlamp. We each grabbed our booze and sat on our sleeping pads, getting good and drunk to the sound of light rain falling on the canopy of leaves over us. Finally J went to be while JB, S and I took a drunken night hike to a nearby outlook. We sat there for a while, watching the stars and finishing off our booze and then we stumbled back to our tents. The next morning, J took off early for the car while JB, S and I packed everything up. As we were walking to the car, I invited S to come stay with me in Virginia Beach (he said earlier that he missed the beach a lot,) and I\'d drive him to DC on the weekend. After a short hesitation, S agreed and headed back to the car for the ride home. We dropped off JB at his truck, swapped e-mail addresses and said goodbye through the misty morning air and J,S and I drove home...TO BE CONTINUED...   '),(160,4074230,'31,July,2004','2004-07-31 00:00:00','   I for one know what it means to be a liberal. I come from an extremely conservative family. As a matter of fact, pretty much like Sean Hannity and Rush kind of conservative. I came out as a liberal when I was about thirteen, and my grandfather told me he would rather shoot me in the head right then, than see me become a \"communist\" (and he wasn\'t kidding). Kind of reminiscent of the McCarthy era, don\'t you think? I saw us go to war in Viet Nam and trample the constitution. I saw the civil rights of a whole race of people being challenged because that\'s the way it\'s always been. I heard conservatives say it was neccesary to kill students at Kent State University who were exercising their constitutional rights. Well when you are faced with that kind of conservatism, you either fall in line or start questioning what is right. I did the latter. I\'m a Dem because I believe firmly in the power of the people to govern themselves through the democratic process. I believe in the power of the human spirit, to become better than what we used to be, to grow, to be alive. If the founding fathers wanted the constitution to be a museum piece with no room for improvement, they wouldn\'t have made a way for us to amend it,... there would only be one party, the conservative Republicans. Things would be the same as they were over two hundred years ago. We\'d still be burning witches, stoning homosexuals, and most of the women in this country would be wearing a large \"A\" emblasoned on their breast in crimson. For me liberalism is a flame... the flame of freedom that burns in the hand of a woman called Liberty. A flame that calls all people equal. She does not stand in the harbour to protect us from the infidel...but to welcome all those who are oppressed. We must ever go forward,...not back.  '),(161,4074230,'30,July,2004','2004-07-30 00:00:00','   urlLink Air America Radio   Personally I think this radio station is \"Fair and Balanced\"  '),(162,4074230,'30,July,2004','2004-07-30 00:00:00','   Why am I so hip on slamming Republicans?  Well, the main reason is look what Bush has done to the United States.  He has tossed us into a needless war, alienated our allies (Lets hope we donâ€™t need them any time soon), allows American jobs to go over seas, created the largest budget deficit in a long time (where did the surplus go that President Clinton created) and the list goes on and on.   Now, if anyone can give me a good reason not to slam Republicans I will be more than happy to sit down with you and discuss it.  I will also give you a chance to voice your â€œfair and balanced opinionâ€.  '),(163,4074230,'30,July,2004','2004-07-30 00:00:00','   Fair and Balanced!   urlLink  '),(164,4074230,'30,July,2004','2004-07-30 00:00:00','   urlLink CNN.com - Key points of Kerry\'s speech - Jul 29, 2004    All of these Key points sound very important to me?  '),(165,4074230,'29,July,2004','2004-07-29 00:00:00','   What American Political Party would you rather belong to?  The Democratic party who are not limited in their thinking and who are free from bigotry.  Or, the Republican party who are opposed to change and who tend to be restrained?  Before deciding if Democrats are better or if Republicans are lets see what the definitions are.  Democrat- An advocate of  democracy . Democrat A member of the Democratic Party.  Republican- Of, relating to, or characteristic of a  republic . Favoring a republic as the best form of government. Republican Of, relating to, characteristic of, or belonging to the Republican Party of the United States.  Definitions from Dictionary.com  Now lets define the synonymous names for both Democrats and Republicans.  Liberal-  Not limited  to or by established, traditional, orthodox, or authoritarian attitudes, views, or dogmas;  free from bigotry .  Favoring proposals for reform ,  open to new ideas for progress, and tolerant of the ideas and behavior of others; broad-minded.  Of, relating to, or characteristic of liberalism. Liberal Of, designating, or characteristic of a political party founded on or associated with principles of social and political liberalism, especially in Great Britain, Canada, and the United States.  Conservative- Favoring traditional views and values;  tending to oppose change . Traditional or  restrained  in style: a conservative dark suit. Moderate; cautious: a conservative estimate. Of or relating to the political philosophy of conservatism. Belonging to a conservative party, group, or movement. Conservative Of or belonging to the Conservative Party in the United Kingdom or the Progressive Conservative Party in Canada. Conservative Of or adhering to Conservative Judaism. Tending to conserve; preservative: the conservative use of natural resources.  What American Political Party would you rather belong to?  The Democratic party who are not limited in their thinking and who are free from bigotry.  Or, the Republican party who are opposed to change and who tend to be restrained?  '),(166,4074230,'17,August,2004','2004-08-17 00:00:00','   The other day I was out diligently campaigning for the Kerry-Edwards ticket, like any other good liberal. The local party has a serious shortage of precinct captains, and with my inherent fear of what will happen if GWB has four more years to plunder the countryside, I volunteered and went out and put some blisters on my feet, attempting to spread some of the hope I garnered from the Democratic Convention.  Some of my neighbors were staunch Republicans; just as I had suspected. You can pretty much tell who those people are in my neighborhood by a cursory inspection of their homes. My neighborhood is one of those, in the first throes of historical reconstruction. Those whose homes have already undergone refurbishing are pretty much conservative.  The poor are pretty much firm Democrats, with some exceptions! The exceptions are the Religious Right. This is pretty much a mystery to me. If Jesus had been born thirty-three years ago, I\'m certain he\'d be branded as an extreme leftist by the conservative pundits today. I\'m also just as sure that the televangelists would crucify him (they\'d probably nail him tighter to the cross than the Roman soldiers did two-thousand years ago; just to make sure he never rose from the dead). In short the money-changers have returned to the temple.  More to the point however, I walked up to the home of one of one of my neighbors who I knew was a disabled veteran (and not very well off I might add). I was not prepared for the reception I received. As soon as he saw my Kerry shirt, I was bombasted with a long line of expletives, followed by: \"Are you for Kerry? I won\'t vote for Kerry, he ain\'t no Christian. Bush, now he\'s a Christian. No I won\'t be votin\' for Kerry. I\'ll be votin\' for Bush, he\'s a Christian!  I didn\'t even try to change his mind in the least. I\'ve run up against this mindset before, (in fact I was reminded of how Jesus himself felt when people were so willingly blinded by false belief). They believe what they\'re told by the likes of Limbaugh, Hannity, O\'Reilly, Robertson, Falwell, and on and on and on.....oh yeah, let\'s not forget the pastor of the local [insert brand name here] fundamental bible believin\' church.  Now before I start getting a bunch of flaming hatemail, let me explain that I have no axe to grind with Christians as such (some of my best friends are Christians...), what I don\'t get is how someone could call themselves Christian and still back the Bush-Cheney ticket. Neither one seems to me to be very much like Christ at all, yet the very meaning of the word Christian is \"little christ\".  Seeing as how the bible very explicily states; \"by their fruits ye shall know them\", I\'d like to invite you to compare Bush-Cheney\'s actions with Jesus\' teachings in the following article by Ernest Partridge, entitled:          \"What would Jesus Do?\"'),(167,4074230,'12,August,2004','2004-08-12 00:00:00','   Interesting Bush Facts Bush\'s tax cuts overwhelmingly favor the wealthy. The top 1% will save more than $30,000 per year while 31% save nothing at all. Source:  urlLink Citizens for Tax Justice  and  urlLink AFL/CIO  The nation faces a record $500 billion deficit. The biggest cause of the deficit is Bush\'s tax cuts. Source:  urlLink www.govexec.com  The Bush Administration is systematically turning back 30 years of environmental progress. Source:  urlLink Sierra Club   President Bush is under-funding education with the smallest percentage increase in funding in eight years. Source:  urlLink Committee on Education and the Workforce   2.9 million jobs have been lost under the Bush Administration. Source:  urlLink New York Times   Bush misled the public about the reasons for going to war with Iraq. No weapons of mass destruction have been found. Source:  urlLink The Guardian  Bush\'s Medicare bill helps corporations, not the elderly. 43.6 million U.S. citizens do not have health care. Source:  urlLink www.misleader.org   Halliburton, the company where Vice President Dick Cheney was CEO, was given multi-billion dollar contracts in Iraq without competitive bidding. Source:  urlLink New York Review of Books   Pentagon officials found that Halliburton overcharged the U.S. government by $61 million for gasoline in Iraq. Source:  urlLink Associated Press   Bush has tried to roll back more than 200 of America \'s environmental laws. The United States emits more greenhouse gases than any other country. Source:  urlLink Rolling Stone     Bush: The Warrior 2 â€“ Number of nations George W. Bush has attacked and taken over since coming into office. 130 â€“ Approximate number of countries (out of a total of 191 recognized by the United Nations) with a U.S. military presence. 10 million â€“ Estimated number of people worldwide who took to the streets in opposition to the invasion of Iraq , setting an all-time record for simultaneous protest. $100 billion â€“ Estimated cost of the war in Iraq to U.S. citizens by the end of the year. 104 â€“ Number of American combat deaths in Iraq between May 2003 (when Bush landed on an aircraft carrier and declared an end to the open conflict and the middle of October 2003. 907 â€“ Number of U.S. deaths in Iraq as of July 26th, 2004. 37 â€“ Percentage of Americans who approve of Bush\'s handling of the situation with Iraq. Source:  urlLink NY Times  95 - Percentage of Army Reserve Soldiers in Iraq and other Middle East bases experiencing significant pay problems. Source:  urlLink GAO report  What the bipartisan 9/11 commission said on the link between Iraq and Al Qaeda: \"We have no credible evidence that Iraq and Al Qaeda cooperated on attacks against the United States.\"   Bush: Money Manager $28 billion - Amount of proposed cuts to veterans\' benefits in the congressional Republicans\' budget resolution for fiscal year 2004. $127 billion - Amount of U.S. budget surplus in fiscal year 2001, the last year Clinton was president. $374 billion - Amount of U.S. budget deficit in fiscal year 2003. #1 - This year\'s deficit will be the biggest in U.S. history. $7.29 trillion - Current  urlLink national debt . $1.98 billion - Amount on average the national debt increases each day. $23,396 â€“ Amount of each citizen\'s share of the national debt as of October 21, 2003. #1 â€“ Record for most bankruptcies filed in a single year (1.57 million) set in 2002. #1 â€“ Set record in 2003, for most residential real-estate foreclosures in a one-quarter period. 1.6 â€“ Percentage increase in economic growth since Bush took office, the slowest rate of increase over an equivalent period for any administration in 50 years.   Bush: Has Lots of Friends $11.5 million â€“ Amount of  urlLink hard money  Bush raised through the controversial â€œPioneerâ€ program. Pioneers pledged to raise at least $100,000 by bundling together checks from friends and family. 212 â€“ Total number of Pioneers identified by the Bush 2000 campaign. 524 â€“ Total number of Pioneers later revealed through court documents. 61 â€“ Number of Pioneers subsequently named to government posts. 19 â€“ Number of Pioneers subsequently appointed as U.S. ambassadors. 2 â€“ Number of Pioneers subsequently appointed to the  urlLink Cabinet  . $5.3 million â€“ Amount Bush raised this past September 30 toward his re-election campaign, breaking the one-day record he himself set.   Bush: The Environment 58 million- Number of acres of public lands Bush has opened to road building, logging, and drilling. 200 - Number of public-health and environmental laws Bush has worked to downgrade or weaken since taking office. #1 - Rank of U.S. worldwide in terms of greenhouse gas emissions.   Bush: More Like the French Than He Would Care to Admit 28 - Number of vacation days Bush took in August this year, the second-longest vacation of any president in U.S. history (Record holder: Richard M. Nixon.) 13 - Number of vacation days the average American receives each year.   Bush: Loves to Travel 65 â€“ Approximate number of fund-raisers attended in 2002. 0 â€“ Number of trips taken to Afghanistan before waging war against that country. 0 â€“ Number of trips taken to Iraq before waging war against that country. 0 â€“ Number of funerals or memorials Bush has attended for soldiers killed in Iraq.   Bush: Tough On Crime #1 â€“ First president to execute a federal prisoner in the last 40 years. 8 â€“ Number of days after that first execution that a second federal prisoner was executed. #1 â€“ As governor of Texas , executed more prisoners (152) than any governor in modern U.S. history.   Bush: Employment Record 2.4 million â€“ Number of Americans who lost their jobs during the first two and half years of the Bush administration. 9 million â€“ Number of workers unemployed as of September 2003. #1 â€“ The administration is well on its way to being the first since Herbert Hoover\'s to preside over an overall loss of jobs during its complete term in office.   Bush: Helping the Rich #1 â€“ Has assembled the wealthiest  urlLink Cabinet  in U.S. history. $10.9 million â€“ Average wealth of the members of Bush\'s original 16-person Cabinet. 75 â€“ Percentage of Americans unaffected by Bush\'s 2003 cuts in the  urlLink capital-gains  and dividend taxes. $42,000 â€“ Average savings members of Bush\'s Cabinet are expected to receive this year as a result of cuts in capital-gains and dividends taxes. $42,228 â€“  urlLink Median  household income in the U.S. in 2001. $116,000 â€“ Amount Vice President Dick Cheney is expected to save each year in taxes. 9 â€“ Number of members of Bush\'s Defense Policy Board who also sit on the corporate board of, or advise, at least one defense contractor.   Bush: Helping the Poor 43.6 million â€“ Number of Americans without health insurance as of 2002. $300 million â€“ Amount cut in December 2002 from the federal program that provides subsidies to poor families so they can heat their homes during the winter.   Bush: Foreign Relationships 35 â€“ Number of countries to which U.S. has suspended military assistance after they failed to sign agreements giving Americans immunity from prosecution before the International Criminal Court. #1 â€“ First American president to ignore the Geneva Convention on warfare (by refusing to allow inspectors access to U.S.-held prisoners of war).   Bush: Public Opinion 90 â€“ Percentage of Americans who approved of the way Bush was handling his job as president on September 26th , 2001. 67 â€“ Percentage of Americans who approved of the way Bush was handling his job as president on September 26th , 2002. 54 â€“ Percentage of Americans who approved of the way Bush was handling his job as president on September 30th , 2003. 50 â€“ Percentage of Americans who approved of the way Bush was handling his job as president on October 15th , 2003.  '),(168,4074230,'06,August,2004','2004-08-06 00:00:00','   In case you\'re wondering, they\'re still here, disguised as moderates. Anna Quindlen in Newsweek explains...  urlLink http://www.msnbc.msn.com/id/5569783/site/newsweek/  '),(169,4074230,'05,August,2004','2004-08-05 00:00:00','   John Kerry tells the Unity conference of minority journalists in response to a question, that if he had been president during the al Qaeda attacks on New York, he would have handled the situation differently.  Here is what he told the unity conference. \"Had I been reading to children and had my top aide whispered in my ear, \'America is under attack,\' I would have told those kids very politely and nicely that the president of the United States had something that he needed to attend to -- and I would have attended to it.\"  According to the 9/11 commission report, the President took between 5 and 7 minutes before reacting to the news from White House Chief of Staff Andrew Card that a second hijacked jet had crashed into the world trade center.  Our country is under attack and the president continues to read a childrenâ€™s book?  '),(170,4074230,'03,August,2004','2004-08-03 00:00:00','   urlLink CNN.com - McCain condemns anti-Kerry ad - Aug 5, 2004   McCain a REPUBLICAN condemns anti-Kerry ad  '),(171,4074230,'03,August,2004','2004-08-03 00:00:00','   I don\'t know about you, but I was more than a little pissed off at the coverage of the Democratic Convention. \"MSNBC is a little more liberal than FAUX\", I thought, so I watched the coverage of Chris Matthews roll out conservative nay-sayers of every sort to trash every good thing anyone said. Now, don\'t get me wrong, I\'m not in love with the Kerry-Edwards ticket. I\'ll support them of course, but I\'m a liberal-liberal, not a conservative-liberal. I have to use terminology like this, because the ultra-conservative pundits working for the Repubs can\'t see any middle ground anywhere and are busy painting moderate to conservative Dems as some type of \"Old Europe\" communists. For more info on where all of the 2004 Dem Primary candidates actually stand, check out  urlLink http://www.politicalcompass.org/  . But I digress...while watching all of the circus entertainment after the Al Sharpton speech, my body went numb and started to sweat profusely, I couldn\'t breathe. At first I thought I was having some sort of massive panic attack. Just then, a powerful force that felt as old as humanity took control of my body. I felt my hand move over to the remote and begin to push buttons all on it\'s own. Instead of my head spinning around and spitting pea soup, a strange thing happened. My hand switched the channel to CSPAN. I was awash with the enormity of it all as the mysterious force returned control of my body to me, materialized in front of me in the form of Lady Liberty, spoke three simple words, \"Think for yourself!\", then vanished. Okay, okay. Maybe  some  of that didn\'t really happen  exactly  that way, but I had to get your attention somehow! The point of all of this is that I really don\'t like some bobble-headed talking doll telling me how I\'m going to react to someone\'s speech. Some  real , intelligent political commentary might have been nice, but only during the dance party, not during a speech that someone else thinks is not important enough for me to hear. Even though CSPAN is not really independent television I guess it\'s the closest thing we have for now, and they covered the convention pretty much in it\'s entirety with no commercial interruptions, and direct from the pundit free zone. If you\'re as concerned about the state of the media as I am, check out the Washington Monthly article written by Ted Turner at  urlLink http://www.washingtonmonthly.com/features/2004/0407.turner.html (yeah, yeah, I\'m not a fan of Ted either, but he does have some substantive things to say). Then run, don\'t crawl over to Media Matters for America at  urlLink http://mediamatters.org/  and join in the activist network. Together we CAN make a difference!  '),(172,4074230,'02,August,2004','2004-08-02 00:00:00','   Bush says in his July 31st radio address...\"Because of my policy of strengthening the economy while enforcing spending discipline in Washington, we remain on pace to reduce the deficit by half in the next five years,\" Now the deficit is up 445 billion in 2004 from 374 billion last year.  Bush inheridted a 500 billion budget surplus when he came into office.  GO FIGURE!  '),(173,4074230,'02,August,2004','2004-08-02 00:00:00','   This one gives me more hope that \"help is on the way\" than the Democratic Convention.  urlLink http://www.latimes.com/news/opinion/commentary/la-oe-micklethwait28jul28,1,6511477.story  '),(174,4074230,'02,August,2004','2004-08-02 00:00:00','   urlLink CNN.com - Study: Fear shapes voters\'ï¿½views - Jul 30, 2004  I read an article on CNN.com today, that says Bush is invoking fear and death in order to gain votes refering to the 9/11/2001 terror attacks. How low can you get! see the entire article at the above link.  '),(175,4074230,'02,August,2004','2004-08-02 00:00:00','   The week before last, we saw the Department of Homeland Security deciding whether or not to go to Congress to ask for the power to cancel elections if terrorists attack or threaten to, before or during elections. \"There, there now G-Dubya. A little worried about the outcome of the elections?\" If we\'d used the National Guard for it\'s intended purpose (guarding the nation) instead of spreading them across the global oil fields, we might not need this precaution. But then again, if we use the same tactics to win as we did during the last election, things are going to look mighty fishy... Activists for democracy all over the nation went into action writing their congressmen and women, airing their fears that we\'ll have an unelected president again... or still... or whatever. The fact remains that Dubya\'s party is still in power... and that would look fishy too. We  need  an angle. Car bombs on Wall Street? \" Never before have we collected this kind of complete intellegence from the terrorists\".   ORANGE ALERT! Plastic on your windows!More to come from the Dirctor of Homeland Security, Tom Ridge... HERE COMES THE FEAR!  '),(176,715237,'03,September,2003','2003-09-03 00:00:00','   RAMBLING INCOHERENCE by Jason Latour     Well, it\'s another late Sunday night—or Monday morning, rather—and I have decided to write, the response to which I understand causes great rejoicing by the Iraqi people. Honestly I mean not to get into all that war talk again, but what doesn\'t cause the Iraqi people to rejoice anymore? It\'s like its free chalupa time at Taco Bell everyday, all day, over there now. Well, at least according to CNN, it is. I think what the Iraqi people need right now is some good service-station-brand ripple to celebrate with. I mean honestly, ripple is always cause for cheer—especially strawberry ripple.   Mmmm ripple...  As long as I\'m RAMBLING INCOHERENT-ly™, I saw on Bill Mahr a while back that Texas is trying to increase the laws against sodomy in their state—making it punishable by death or something ridiculous like that. For those of you that don\'t know, \"sodomy\" is defined as:   \"Any of various forms of sexual intercourse held to be unnatural or abnormal, especially anal intercourse or bestiality.\"  The notion of outlawing sodomy, is in my mind,ridiculous because—honestly—who defines what is deviant and what\'s not? Some people might think letting the woman get on top or doing it \"doggy\" style is deviant. Or strapping on a Louisville Slugger and strangling your partner to the sound of Beethoven\'s Fifth symphony--- honestly, who among us hasn\'t wanted to donkey punch our lover now and then? It might be rude but it\'s certainly not illegal or immoral activity in my eyes. Just stay away from the kids and barnyard animals and you\'re okay in my book, I guess. But that said, I would probably have sex with a cow, if it had a nice ass and some big sexy breasts...  It brings up the age-old question of who defines what is immoral for another? Take for example myself. I\'m the ripe old age of 26, which in my new found home of Gainesville, Florida, is looked upon as being about the equvialent of Yoda or that old guy with the booger finger from THE GOLDEN CHILD. Being a 26-year-old man in a town populated by beautiful, sexually ripe 18–21-year-old women is hard on the constitution. It\'s unbelievable. Now, I used to say, \"I\'d never touch a high school girl\"—which basically the freshmen girls around here still are—but when it\'s in your face—and I mean literally in your face—man, let\'s just say the temptation is GREAT. So great that it makes me want to grab a bottle of that aforementioned strawberry ripple and get REALLY REALLY REALLY schnauzelled. Because under the influence, we all know that it\'s okay to just say that you didn\'t know. But then you\'re entering Kobe Bryant territory and no one wants that.   Ah, Kobe. The whole thing is mind-numbingly stupid. I thought he was a giant idiot when I first heard he married a 19-year-old who was fresh out of high school. When you live in a world where fame and popularity determine sex status so mightily that Demi Moore is jumping Aston Kutcher, you just don\'t marry a 19-year-old girl (no matter how hot she is)! Then to make matters worse, once you get tired of that \"old hag\" you\'ve got at home, you jump the first preteen beer-tub girl that slants you the sweet eye. And then he either forces her to have sex with him (when we all know that there are millions of hot young women out there who would do it for free) or he just gets way \"too\" freaky (based on the bruises and missing hairs and such) and she can\'t take it when he leaves. When you\'re in the the public eye you just don\'t do that stuff. My advice to Kobe in the future is: if you can\'t keep in holstered then follow the lead of our mutual idol, Michael Jordan. Keep it on the DL with older women who know the deal. A 30-year-old basketball groupie isn\'t running home crying to her parents—no matter what happens. She\'s two-way paging her girlfriends to brag about the new Benzo she\'s pushing with the money you fronted her to keep quiet. Now I know this sounds hypocritical coming from me based on what I said earlier, but you have to realize that Kobe and I ain\'t exactly playing on the same ball field. He\'s in the majors and I\'m slumming in the optimist league.   So, what did we learn today? That if if you\'re travelling to Texas and you have a pention for sex with underage barnyard life don\'t be a Kobe—keep it on the down low!  '),(177,715237,'03,September,2003','2003-09-03 00:00:00','   RAMBLING INCOHERENCE By Jason Latour      I have now officially degenerated to full blown slob. I was eating some Tostitos and salsa moments ago when I somehow unexplicably lost all control of my motor functions and dumped a healthy scoop of \"Medium All Natural Salsa\" onto my t-shirt. Well like any good American, I just tried to scoop it off my shirt with a new chip only to be foiled by a second \"salsa foul\" as I redumped salsa onto my shirt again. So now I sit shirtless on my couch eating chips and salsa—my hands pathetically stretching crumbs of chips in hopeless vain for that last glob of tomato and pepper paste goodness at the bottom of the jar.   I have come to the realization that no matter how old I get, there is a certain level of slobbery that will always cling to me. Soon physical exercise  and weight lifting will no longer be neccessary. Much like your dad.   It\'s just something that dad\'s have. You reach a certain point and its as if you don\'t need to work out anymore. My dad, for instance, never ever, ever works out. In a weight room I can lift circles around him. BUT when it comes to twisting a bolt off a washer or lifting a home entertainment center, he\'s a workhorse. That, my friends, is \"MAN STRENGTH.\" When you can twist off a beer cap with your forearm, you have evolved.   I think we men need a revolution in this country. We need more salsa-ridden slobs sitting half naked on their couches watching WILD ON. We need to say that it\'s okay if I have a little paunch and, no, I don’t need no damn QUEER EYES on this STRAIGHT GUY, thank you very much. I can twist off a beer cap with my forearm and drive a nail through a two-by-four with my face—can you?   Don’t get me wrong I’m all for the sexual revolution of these times. I like that we can all exist peacefully and not (at least in public) be at each other’s throats. But I say we make the normal man the new J-Lo. She made the big ass fashionable in a time when skinny waifs populated the landscape. Let\'s do the same for those guys out there who maybe have a little too much back hair or just don\'t quite smell right. Oh, Al Bundy, where are you when we need you?  '),(178,715237,'03,September,2003','2003-09-03 00:00:00','   RAMBLING INCOHERENCE By  Jason Latour      I am broke. Broke as a joke.   Remember those old episodes of DUCK TALES where Uncle Scrooge would swim in his giant vault of money? Well I recall there was an episode where these mole men cause an earthquake and all the sudden Uncle Scrooge had like one coin and it was rolling around on the floor towards this huge crack. This was a man (well a duck) who had once had it all and that was his last gold coin rolling towards certain oblivion and all he could do was lust forward ---reaching in vain as it fell to the center of the earth.   Dear god the humanity of it all.   It pretty much describes what happened to me in Las Vegas this past summer.   I\'m just not lucky. There is no way of getting around it. I mean I need Karma lessons or something because I play the odds the best way you can and I still lose. Now don\'t get me wrong---I\'m not hating on Vegas. I LOVE the place. But it kicks my ass.   I guess the appeal, whatREALLY makes Vegas so great, is that it\'s just about the only place in this country where almost ANYTHING and EVERYTHING goes. It\'s a place where anyone with a couple hundred bucks in their pocket can blow off steam doing things that they would get arrested for in Smallville, Kansas. Though as much fun as I have there on my meager budget--- I can only imagine how it would be if I were rich.   Though if you’re rich, I mean REALLY rich, I guess you don’t need Vegas. You can do pretty much anything you want anyway. Which is my THEORY on why rich people are usually so weird—it’s because they’re bored. When you can afford to finally get that swimming pool full of Coca-Cola you wanted as a kid it changes you in really strange ways. It’s like some sort of super marijuana high that just kills all ambition and reason.  But of course that wouldn’t be a problem for me because you see---I have imagination.   For example I would make people climb poles for my amusement---at threat of death or beatings. Now how is that legal or even fun you might ask? Well we’re speaking of rich people here so what’s legal isn’t really an issue. Besides, the Japanese have game shows that are close to the things I\'m proposing already. It might seem stupid or lacking in the fun department but unless you’ve ever SEEN a fat guy or a midget climb a pole you just don\'t understand. The key is you have to to sweeten the deal for the contestants—put a prize at the top, like a gift certificate or a steak or a loved one hanging on for dear life. Imagine the side bets.  See sounds funner than you thought.  Of course gambling with human life isn\'t the only perk of the super rich---you get to date super models. Though even then you always get tired and all uppity and self-righteous and start having these ungoldly high standards about the women you should date. You start actually expecting your super model girlfriend to not only have perfect breasts but now she’s got to a Yale grad. And it’s not that this girl isn’t out there but trust me she’s harder to find that WALDO when you’re tripping on acid. So my question to the uppity rich is why stop there? As long as this is fantasy perfection, why not find other perfect combinations of women. Say, a  different supermodel for every household task. I\'d have a cook-model and a vacuuming-model and a dusting model---a gutter cleaning model--one to clean my collection of model porn. Ah it would just be a misogynistic heaven wouldn\'t it? Isn\'t that a dream that we can all get behind? I know I can.   So I guess the point is that I’m glad I get to go to Vegas every year and loose money. Gambling is the everyman’s vice. It’s risky and fun and on the edge without going too far. Sure I could save my money and invest wisely so I could have a better shot at being rich someday, but if being rich means that I have to start getting jacked up on cocaine and Viagra while watching men fight to the death in Thailand to get my kicks----then I just don’t think its for me.  '),(179,715237,'16,June,2004','2004-06-16 00:00:00','   RAMBLING INCOHERENCE by Jason Latour      In a shocking turn of events, pop sensation Britney Spears has been forced to cancel her summer tour due to a knee injury sustained on the set of her latest video shoot with rapper Calvin Broadus aka Snoop Dogg.   \"Yeah, I guess she just tried to bust a move and slipped.....slipped on this DEE-IACK!\" said Mr. Broadus.    The move will force fantasy owners to scramble for a replacement as Spears was second in the league only to Beyonce\' Knowles in AGPS (ass giggle per second)and first in BIYF (boobs in your face). Owners should consider trading for another star with a comprable slut to innocence ratio such as Jessica Simpson, Jessica Alba or Nancy Reagan.  '),(180,715237,'15,June,2004','2004-06-15 00:00:00','   RAMBLING INCOHERENCE by Jason Latour      Tonight was truly a night to remember as Rasheed Wallace and the Detroit Pistons crushed the Los Angeles Lakers 100-87 to win the NBA Finals 4 games to 1.   It wasn\'t even close.   The Pistons stifling defense made Kobe and Shaq look very human, Gary Payton look like an idiot and Karl Malone look washed up.   And I for one loved every minute of it.   Yes ladies and gentlemen, I did indeed pick the Pistons over the Lakers, though as you might expect I didn\'t think it was going to be this easy.   Granted I enjoy rooting against them, it makes the game fun. In fact with the possible exception of the Dallas Cowboys, they might be my favorite team ever to pull against. But that said I can\'t say I\'m not glad to see them go.   Next year will bring a much different Lakers squad in my estimation. Leaving us with the question \"what happens to this current bunch of Laker superstars?\"   Phil Jackson will ride off into the sunset, to smoke peyote on his ranch and go on dreamquests with naked indian shaman until of course another job like say coaching Lebron James or the Dream Team opens up in five or six years. He\'ll return to the court in Dalhi Lama gear, head shaven, dressed a pink robe, only to discover that Cleveland is no Shangri La and that Drew Carey doesn\'t really live there anymore.   Kobe will go to prison, not for rape but in a stroke most egomanical-- for sodomizing himself . He will lead the prison league in scoring average for two seasons before being released back into civilization. He will then renounce basketball and his former life and start a religion known as \"Kobe-ism\" where the central belief is that \"one must shoot---one must always shoot\".   Shaq...well Shaq will continue to grow, soon eclipsing our sun. This will force our government to propel him into space, where he will roam the universe in search of barren planets to feed on and quench his \"Shaq-sized hunger\".   Karl Malone will return to Lousiana, move to the bayou and begin a successful career running \"shine\". He and his pet possum \"Cheeto\" will become the most feared and wanted shine smugglers since Burt Reynolds played the Bandit.   Gary Payton will replace Rush Limbaugh as talk radio\'s number one shock jock. His levels of stupidity and egomania will force FCC regulations prohibiting him from referring to himself in the third person---a move which will eventually cause his head to explode.   Luke Walton will forgo a lucrative NBA future and move back in with his parents where he will be submitted to the daily ridicule of playing his father, Bill, in one on one matches. After a decade of studying the technique of delivering severe razor sharp elbows to the face and fused ankle pivot moves he will return to the NBA and win an NBA title, fufilling his destiny.   And all will be right with the world.  '),(181,715237,'15,June,2004','2004-06-15 00:00:00','   RAMBLING INCOHERENCE Seed of the Future By Jason Latour    Disclaimer: The following article is from a message board post I made on my fantasy baseball league message boards. Jason Ollman, Trey Flynn, Jason Wolf etc. are good friends of mine. If you\'ve never seen Jon Maynard\'s Tron suit website you should  urlLink check it out here.  But I warn you it will make you want to claw your own eyes out. This post was so insanely stupid I just thought it would be fun to post here. Enjoy.    SEED OF THE FUTURE!     Dateline- Gainesville, FL   After a harrowing journey through time and space Space Semen First Class Jason “Jingles” Ollman Jr. is happy to return to the loving bosom of his parents---Jason Ollman and Joseph “Trey” Flynn of Gainesville, FL.   Appearing in a flash of light and heat, the “One true, sperm warrior Tronite” arrived late last evening, interrupting the first quarter of Detroit’s victory over the L.A. Lakers with his blindingly horrid visage.   This strange visitor from another era offered his would be parents a vague but telling explanation of his origins:   “I come from the year 2040, seeking refuge, an outcast from society due to my perfect combination of your horrible, horrible genetic flaws!”   “I now return to the bosom of love that once bore my putrid soul! Take in me and drink of my stench!” cried the geekily clad freak.   Bystanders were then horrified as the three exchanged a heartfelt embrace, which may or may not have included some ass fingering.   “It doesn’t really surprise me,” said Jason Latour. “No, no shock at all. Makes complete sense. He looks exactly like I’d picture him. A horrific freak.”   “If my boy came out looking like that…I’d have to take him out back and put him down,” commented Jason Wolf.   Then greeting his parents a final farewell, this cosmic visitor from the unknown waddled back into his time pod, preparing for a return to his home.   “I now return to my time, relieved to now know that my parents are indeed the reason for my infinitesimal success and lifelong virginity. I am now content to live out my remaining years as head chef at Steak and Shake 2040, where the burgers are always greasy and the chances of success are gloomy. Farewell parental units, may you soon achieve my procreation, either through the miracle of your primitive science or through the force of your combined ass loving.”  '),(182,715237,'15,June,2004','2004-06-15 00:00:00','   Rambling Incoherence By Jason Latour      A while back I got one of those head set thingies for my cell phone, the goal being to cut down on the cancerous radiation leaking into my brain. Fast forward to today, I have the ear piece thing in so that the deadly cancerous cell phone isn\'t next to my head, but I look down and low and behold my phone is sitting in my lap---right over my, well  \"my boys\".   Is that Darwinism at work or what?   What do you choose in that situation? Your brain or your \"boys\"? Do you wanna be a drooling freak with functional genitalia or still be able to form coherent sentences yet impotent?   It\'s quite the dilemma.   Though if you think about it cell phones are just the tip of the iceberg when it comes to the scary stuff that’s out there. These days you’ve got the West Nile and the Mad Cow and a list of STD’s longer than Yao Ming’s arm. The world is a scary, scary place and I think that mysterious cell phone radiation is probably not a viable concern--- at least I hope.  Still it makes one yearn for the good old days of my father’s generation where you could screw a Vietnamese hooker and, if it burned afterwards, that was a sign that you were still alive?  '),(183,715237,'24,June,2004','2004-06-24 00:00:00','   Dwight Howard, no. 1 pick and property of the Orlando Magic-NBA brain farts.     So the actual NBA draft was tonight. I watched the first half or so and then arouse from my coma. Comments below:  FIRST ROUND-   1 Orlando Dwight Howard PF 6-10 240 Southwest Atlanta Christian- Orlando I could kiss you. Don\'t worry MAYBE he\'ll pan out.   2 Charlotte Emeka Okafor PF 6-9 252 Connecticut -Ch\'ching!   3 Chicago Ben Gordon PG 6-2 195 Connecticut- So Okafor and Gordon go 2 and 3, more for UCONN to brag about--great. Diana Taurasi can eat my ass.   4 L.A. Clippers Shaun Livingston PG 6-7 175 Peoria Central HS - Welcome to the NBA Shaun, the loaded gun is in your foot locker.  5 Washington (Traded to Dallas) Devin Harris PG 6-3 185 Wisconsin- Hey Devin, some guys get models and movie actresses, you get interns, it\'s better than playing for the Jazz.   6 Atlanta Josh Childress SG 6-7 205 Stanford - Anyone else think this kid looks like he came straight off the set of The X-Files?  7 Phoenix (Traded to Chicago) Luol Deng SF 6-8 220 Duke- The BULLS actually made a smart move? Someone tell me what the HELL is going on here?   8 Toronto Rafael Araujo C 6-11 280 BYU- Umm....who? No seriously is this a joke?   9 Philadelphia Andre Iguodala SG 6-6 200 Arizona- If one more guy gets compared to Scottie Pippen I\'ll fucking scream.    10 Cleveland Luke Jackson SF 6-7 215 Oregon- Resisting...urge to...make...stupid ...Star Wars ...joke.   11 Golden State Andris Biedrins PF 7-0 240 BK Skonto Riga (Latvia)- Latveria? Isn\'t that Dr. Doom\'s country?    12 Seattle Robert Swift C 7-0 245 Bakersfield HS- According to draft analyst Stephen Scrhamm he\'s a  poor man\'s Kyle Visser.  13 Portland Sebastian Telfair PG 5-11 170 Lincoln HS- Pack a lunch Portland you have offically been demoted to the special class.     14 Utah Kris Humphries PF 6-8 240 Minnesota- Utah drafted a big white boy...oh the shock.   15 Boston Al Jefferson PF 6-9 250 Prentiss HS- Isn\'t this guy already in the NBA? Al Jefferson? Al Harrington? Jeff Allerson? I\'m lost.   16 Utah Kirk Snyder SG 6-6 219 Nevada- I thought the guy was white. What? I did.  17 Atlanta Josh Smith SF 6-8 220 Oak Hill Academy HS- um another J. Smith?    18 New Orleans J.R. Smith SG 6-6 210 St. Benedict\'s Prep (HS)- am I reading this right?    19 Miami Dorell Wright SF 6-7 190 South Kent Prep HS- Get the feeling this guy\'s gonna be \"sneaking\" into a few Southbeach clubs?   20 Denver (Traded to Orlando) Jameer Nelson PG 6-0 190 Saint Joseph\'s- Gotta feeling some GM\'s might be kicking themselves over missing this one one day.   21 Utah (Traded to Dallas) Pavel Podkolzine C 7-5 303 Metis Varese- That dude is freaking HUGE! I hear he doesn\'t play much on his team in where-ever-the-fuck. I guess they just want him over here so the rest of the country can eat.   22 New Jersey (Traded to Portland) Viktor Khryapa SF 6-9 210 CSKA Moscow- Didn\'t this guy kill Apollo Creed?    23 Portland Sergey Monya SF 6-8 220 CSKA Moscow- Christ whatever happened to the good old days of the Cold War?   24 Boston Delonte West SG 6-4 180 Saint Joseph\'s- Mmmm Delmonte.   25 Boston Tony Allen SG 6-4 205 Oklahoma State - With a name like that, shouldn\'t he be on Broadway?   26 Sacramento Kevin Martin SG 6-7 195 Western Carolina- Wow it REALLY is a thin draft when Catamounts are going in the first round.    27 L.A. Lakers Sasha Vujacic PG 6-7 193 Snaidero Udine (Italy)- I\'m sure he would\'ve been a perfect compliment to Shaq and Kobe. Just that third superstar they were looking for.    28 San Antonio Beno Udrih PG 6-3 203 Breil Milano (Italy)- Pretty soon San Antonio is going need weapons inspectors with all these foriegn guys on the team.   29 Indiana David Harrison C 7-0 260 Colorado- That was my H.S. Driver\'s Ed teachers name. No joke.   SECOND ROUND- Coming soon. Maybe.  '),(184,715237,'21,June,2004','2004-06-21 00:00:00','   RAMBLING INCOHERENCE by Jason Latour   Time for something a little different today. No straight up jokes. Just a good old \"whats on my mind\" type blog.     Uconn\'s Emeka Okafor   So tomorrow (well technically today by the time of this post)is the 2004 NBA draft. A date which for many of you I\'m sure is circled on your calendars as you hold your breath in anticipation, hoping that Pavel Podkolzine falls into your teams lap. For the expansion Charlotte Bobcats, which I guess is now my hometown team though I don\'t live there anymore, it is the inaugural draft. A day where possibly, but not likely, the entire future of the franchise could be forged. So in keeping with this theme, I was watching ESPN and their preview coverage of the  forthcoming draft and the big debate is whether Orlando, who holds the number one selection, should take former UCONN big man and  College national player of the year Emeka Okafor or high school \"phenom\" Dwight Howard. This is of utmost importance to the Bobcats because they hold the second pick and will likely take whichever of the two Orlando doesn\'t.     Dwight Howard    Okay cut to the chase, so I\'m watching the coverage on ESPN and they begin to interview Dwight Howard. Seems like a good kid all and all, I\'m not sure I don\'t know him. But something about him irks me. He begins to talk about his religion, which I don\'t care to hear personally. Religion to me is best left out of sports, its a personal experience that has nothing to do with how many points you scored or TDs you threw. It would be like if someone went to work everyday and told you about the bible verses he read last night in the morning meetings. But the thing that bothered me the most about this was that the kid was talking about how he has stopped listening to \"all music but gospel\" and stopped doing most anything not related to scripture due to his beliefs. Look fine, okay I respect that you want to be a good person and you\'re entitled to your beliefs but when you\'re 18 I just don\'t think its wise to cut yourself off to life experience (which I consider the art forms of music and film,etc. to be a big part of) like that before you\'ve had a chance to form your world view. A small thing to get upset about I know but the clincher was when asked about his acting ability he mentions an aptitude for impressions, then offers up a scene featuring Debo from the movie FRIDAY. Thats right FRIDAY*. Praise Jesus Smokey.     *I love the movie FRIDAY.     Maybe the kid will grow and become more open minded. I hate to paint him as close minded not knowing him or more about him but it just kind of bothered me. I guess most of were different at 18 than we are now. Hopefully if he comes here he\'ll be a great player and a great citizen and I know he won\'t have much trouble finding bible studies to attend here in the \"bible belt\".   Personally I can\'t decide between the two, I prefer Okafor on a personal level but given that we more than likely won\'t be in contention for a few years it might work out if Howard falls into our lap. With Okafor and the plan of \"fiscal responsibility\" (aka tight waddedness) that owner Bob Johnson has laid out, I can see a scenario developing where Okafor stays on our team just long enough to get good, we do terrible and he bolts for a big market pay day. I hope I\'m wrong.   Hey at least I\'m not and Orlando fan and in my buddy  urlLink Trey\'s shoes    In other news, my comic book reading \"geekness\" continues. Which I guess is balanced out by that lengthy sports rant above. I recently have been dropping a lot of the books I read, in some cases out of lack of funds to continue purchasing them regularly and in other cases just sheer boredom. I\'m down to basically the following regular titles  urlLink 100 BULLETS,  SLEEPER, DAREDEVIL and I think now a title that will become the surprise entry....wait for it....freaking AQUAMAN.   Yeah you heard me.   Freaking I talk to fish and swim under-freaking-water AQUAMAN.    urlLink        click for full image   But in all seriousness, I always thought it would take a really good story to get me to read a regular freaking AQUAMAN title and I think I\'ve found it. The current arc is moving the character, who quite frankly always seemed either useless or out of context compared to his contemporaries (not to mention rather silly), into a much more relatable and interesting context.    urlLink   click for full image    Writer Will Pfiefer has set up a story arc in which the city of San Diego as fallen into the ocean (either by tidal wave or an earthquake, I missed the first couple of issues). Well as it turns out there are survivors, who somehow are now endowed with the ability to live under water but no longer out of it. As a result freaking Aquaman decides to look out for the safety of these people who are now a society of \"fish out of water\", all the while searching for the answers to their mysterious metamorphosis. This to me is a nice stroke by Pfiefer in that it puts freaking Aquaman into a context where the stories can still retain a fantasy element but at the same time make the environment somewhat relatable to readers (certainly more so than Atlantis).   urlLink    click for preview    I imagine in the future the new SD will be a new sort of \"kingdom\" for freaking Aquaman, allowing him to be more of a cop/mayor than an out and out king. A nice story so far, at least enough to get my attention and if nothing else Patrick Gleason\'s art is pretty nifty, reminiscent of another fav Doug Mahnke. Hopefully it will continue to hold my interest and make an interesting character out of something that\'s been on the fringe of lame for quite sometime.      Freaking Aquaman.  '),(185,715237,'21,June,2004','2004-06-21 00:00:00','   RAMBLING INCOHERENCE  by Jason Latour   Disturbing, disturbing news out of Hollweird as Matrix co-creator Larry Wachowski has aparently opted undergo a sex change. No I\'m not making this up:    \'Matrix\' co-creator ready for sex change azcentral.com April 30, 2004  When you\'re touted as one-half of the creative genius behind the science fiction movie of the decade, morphing between the sexes is probably just another research trip. At least that is one way to describe Larry Wachowski latest move.   According to the Chicago Sun Times, Wachowski, who created the Matrix series along with his brother Andy, is working toward becoming \"Linda,\" surgically and otherwise. \"As always, it was impossible to get any comment from the press-shy Wachowskis, but several longtime friends of the Rogers Park native confirm Wachowski is planning to complete the process of becoming a woman,\" reports the Chicago Sun Times.  When Thea Bloom, Larry\'s former wife, accused Wachowski of being \"extremely dishonest with me in our personal life,\" blaming the separation on \"very intimate circumstances concerning which I do not elaborate at this time for the reasons of his personal privacy,\" a sex change wasn\'t the first thing to come to mind. But those close to Larry relate the he has been \"living and dressing as a woman for some time.\"   Talk about there is no spoon.  '),(186,715237,'07,July,2004','2004-07-07 00:00:00','   Rambling Incoherence  By Jason Latour   Well is seems in the end it was much ado about nothing as far as the Los Angeles Lakers courtship of Duke University head basketball coach Mike Krzyzewski was concerned. In the end Coach K decided to stay in Durham, opting not to take the $40 million dollar deal that the Lakers reportedly laid on the table .     Coach K with Sweaty Robin, the Ultimate Duke \'tard.      Being a UNC Tarheels fan I have to say that the news doesn\'t really affect me in the way you might think it would. Given Duke\'s dominance in the UNC rivalry the past few years (yeah I admitted it, at least in the win column) you would think I would want the guy to go. But no in a lot of ways I think that a strong Duke team benefits Carolina much more than a weak one. The rivalry is a great draw for recruiting and on a personal level I don\'t want to hear any excuses about Coach K being gone when the Tar heels return to form (as they are expected to these next couple of years).   No, what has me more upset is the way all this is being handled by the media. In particular the latest story of how a lowly Duke student was able to convince Coach K to stay at Duke with a letter  he wrote that bascially went---  \"Dear Coach K, please stay at Duke. I once dreamed for so long about being a Duke player but given my one inch vertical and lack of opposable thumbs I soon realized it was a dream never to be. However I knew that by going to Duke I had an opportunity to be something greater than the neuro surgeon or third world dictator that my parents hoped sending me there would accomplish, I knew that  you would now be  my coach and I would become your most glorious sixth man.  I could cheer you on with all my heart as a coated my flabby body with blue lead based paint and cheered away until my throat was sorer than Jenna Jameson at the end of a long shoot. (Insert continued verbal mastrubation here). The end.\"   Please.   Lets be honest here, that godawful, sappy letter had little to no real sway over this decision. This came down to one thing---Coack K using the Lakers to get what he wanted, which was to stay at Duke and to get the adminstration to take notice that he is as coveted as a coach as he thinks he is. I guarantee you Coach K will , smartly, use this as a bargaining chip to get a bigger pay raise at some point down the line. Don\'t get me wrong, I think he did the right thing in staying but the way everyone is playing it up like he\'s a saint for doing so is ridiculous and the fact that people actually think this goofball who wrote the letter saved Duke Basketball is fucking retarded.   There honestly was no real choice here. Stay at Duke where you are a certain hall of famer and guaranteed an untarnished legacy that will stand for generations? Or risk a year in L.A. where you\'ve got a team full of egos run amok with more bad drama than a daily soap? Add in the fact that there may have been the possiblity of him coaching a Lakers team sans Kobe and Shaq. Honestly people its not rocket science.     So please spare me the heart wrenching cotton candy B.S. about how some goofy looking kid in stain defender dockers saved Duke basketball I get enough propaganda from FOX NEWS.    PS If someone has a copy of the real letter please forward it my way.  '),(187,715237,'04,July,2004','2004-07-04 00:00:00','   Rambling Incoherence The Godfather     Marlon Brando 1924-2004  On a slightly serious note:    This July 4th weekend marked the passing of one of the worlds more famous, most enigmatic and perhaps most talented actors as Marlon Brando died of lung cancer at the age of 80. Like many of us, upon hearing of his death I naturally found myself reflecting upon what I knew of his career and of the man himself.     Brando as Jor-El in SUPERMAN the movie.      As a kid I had my first and perhaps most memorable taste of his work as he portrayed Jor-El in the movie “Superman”, certainly not his most famous role but one that connected with my obsession with comic books and in retrospect an unbelievably apt casting choice. It was much later that  I encountered his work as the titular role in “The Godfather” and as Kurtz in “Apocalypse Now” and was startled at his breadth as an actor. And though while I could recognize the skill and ability he clearly possessed, it was more likely the choice of these roles, the sheer outlandishness (especially in “Apocalypse”) and the inherent comedy in that outlandishness that endeared Brando to me the most. I soon learned that this outlandish behavior carried over into the real world, that people regarded Brando as a talented yet off kilter celebrity who had, much like Kurtz in “Apocalypse Now”, left the “real” world long ago. And while this seemed to be the prevailing public opinion, I personally knew little about the real reason’s why he was considered such an aloof and eccentric personality.  I had heard of his son’s murder trial, of his outspoken activism and of his fondness for food and women but I still had no real basis as to where it all came from. His self-imposed isolation seemed odd given his life of renown and fame. Was he simply as he seemed; the poster boy for celebrity gone AWOL? Or was there more to it? The answer intrigues me to this day.     Brando as Kurtz....crazy...like a fox.    Tonight I was fortunate enough to catch an airing of “A&E’s Biography” which helped to shed some light into his life. I was able to see the events of his life in some sort of context and what I took away has in many ways changed my opinion about him. To me Brando seemed like a man who was very much driven by passions and desires that he could not articulate in any conventional means. After viewing the episode I would have to, in my limited knowledge, conclude that the passing of his mother, his struggles love and with his children all seem to mold his odd behavior and by someway form the indulgences and eccentricities that were outlets for his energy. He seemed very much a man who was blessed and cursed with the gift of being special. I think in someway he was searching for answers in his work, something that I can truly relate to. This seems never more evident than in his portrayal of Kurtz in “Apocalypse Now” a role that seems to mirror his own life. A man who steps out of our society and retreats to a faraway land (in this case Vietnam) where he is king of his own private world populated by unspoiled natives. It is oddly reflective of Brando’s purchase of a Tahitian island and subsequent expatriation in the sixties and perhaps very telling of the man himself.   In the wake of his passing there as been the expected media circus, as questions abound about his last days. Rumors that upon his death he owed in excess of 20 million dollars to debtors. I find it interesting that people must now, having truly known so little about him, grasp at straws in their search to find the “real” Marlon Brando. If anything his death should influence us to look past his obvious flaws and instead  revisit his work and take in films we may have overlooked in the past and reflect on what he left behind.  '),(188,715237,'04,July,2004','2004-07-04 00:00:00','   Rambling Incoherence  By Jason Latour   From CNN.com: LOS ANGELES, California (AP) -- Screen legend Elizabeth Taylor is fighting a family\'s claims that a Vincent van Gogh painting she owns was taken from one of their ancestors by Nazis.      Taylor, whose father bought her the painting at a London auction in 1963, has filed a lawsuit seeking a pre-emptive court declaration that she is the rightful owner of the painting, which hangs in the living room of her Bel-Air estate.  In a statement issued last month, Taylor said she has \"a tremendous amount of sympathy\" for those who lost possessions during World War II and the Holocaust. She said she told her lawyers to thoroughly investigate the ownership history of the 115-year-old van Gogh.   urlLink READ FULL ARTICLE HERE   Just a little commentary on this one:   Sweet Christ let the painting go. Assuming that these claims are remotely valid it makes this thing one of the most ridiculously greedy things I\'ve ever seen. \"I\'m sorry your family endured horrific torture at the hands of a genocidal madman but the Van Gogh looks great above my couch.\"   -J La  '),(189,715237,'04,July,2004','2004-07-04 00:00:00','   Rambling Incoherence  by Jason Latour   Once again another July 4th has come and gone and I for one must interject that it was once again pretty much a disappointment. As holidays go I would rate the 4th behind New Years and Valentines day for holidays with the least actual follow through on the hype. It\'s a great idea and all but when it comes down to it Independence Day is the Star Wars Episode One of holidays, all flash and hype---little to no pay off.   To me all the build up of waiting around for hours just to watch fireworks has the same allure of a NASCAR race. \"Hot damn Honey! I cain\'t wait ta watch sumthin\' \'splode!\"   That\'s not to say that it\'s a totally worthless day, oh by no means. Any excuse to take the day off and drink yourself into a stupor is certainly much better than your average weekday, by default making it worthy of a holiday. Hell, if it were up to me we\'d all be smashed drinking pine sap beer on Arbor Day.It just seems like every 4th of July people have to work the next day, so by the time the party\'s starting it\'s ending. Then you\'ve got a totally good buzz wasted over nothing.  And of course there is the food so thats not bad. Yeah I have no problem eating BBQ.  Damn I had to mention BBQ ribs. Ruins my whole argument.   Blasted delicious ribs.   -J La  '),(190,715237,'06,August,2004','2004-08-06 00:00:00','   Two quick tidbits of humor today.   First off, You have to love these fanfiction movie trailers.   www.theforce.net/theater/nonsw/worldsfinest/   I kept laughing because it looked like it was directed by Buster Gonads. Straight up porn style actors,costumes and lighting.   I just find it funny that these guys spend all this time and energy working on stuff like this. Why not put it to use on something of your own? I mean I guess it\'s kind of cool in the fact that they love the stuff so much but it\'s also just a little sad.   I just goes to show you how damn awesome Christopher Reeve was to pull off that costume. These muscle bound clods that everyone tries to use always look horrible.   And from CNN comes this latest brain fart by our presidente\':  And hopefully, Bush won\'t use his new line, spoken at a White House bill signing yesterday. \"Our enemies are innovative and resourceful, and so are we,\" Bush said. \"They never stop thinking about new ways to harm our country and our people, and neither do we.\"   The only think I\'m gonna miss about Bush is marveling in his uncanny ability to put his foot in his mouth.   -J La  '),(191,715237,'05,August,2004','2004-08-05 00:00:00','   Rambling Incoherence  by Jason Latour   Anyone who knows me knows that Batman is one of my favorite characters/concepts in superhero comics. I\'m certainly not alone in this one, I mean he is one of the most popular characters around. I recently stumbled upon something I found pretty cool (which I\'m sure boosts me well into the UBER GEEK category and expunges the equity I\'d built up from the \"regular world\" with the thirty or so sports posts I\'ve made in this blog), it\'s a short essay that archives the number of people Batman has killed viciously since his creation by Bob Kane back in the thirties. The interesting thing is how the modern Batman we all know and love is pretty much considered to be a paradigm of justice, someone who cherishes life first and foremost and believes in second chances. I\'m not saying this modern version is completely flawed, seeing as that I believe in many of those things myself, but damn this old style bad ass Batman was cool! It just goes to show how quickly things change when a character suddenly becomes a \"property\".   Enjoy:   \"Batman\'s Body Count   By Ray Tate   I\'ve always felt the Punisher, not the original but the remake, was something of a wuss, and rereading the Dark Knight Archives supports my opinion. Batman in his early days made the Punisher look like a bleeding heart and something of an amateurish boob. The Punisher is the poster child for the NRA. Batman used guns to kill only once. He instead preferred to slay his foes through brute strength. Observe, and be amazed at his astounding knack for execution.  In Detective Comics #27, his debut, Batman breaks a thug\'s neck with a \"deadly headlock\" then tosses the body over his shoulder and onto a rooftop. Following this incident, Batman with a solid left cross whacks Styles, and in anticipation of the 1989 Batman movie, the murderer falls into an acid vat. Batman does not lift a finger to save him. Instead, he chills with the line: \"A fitting end to his kind.\"  In Detective Comics #28, a fool attempts to knife Batman who briskly hoists him over the roof and to the welcoming concrete below.  In Detective Comics #29, he snaps the neck of Jabah, Dr. Death\'s goon. In a later scene you can still see Jabah\'s cooling on the floor.  Doctor Death returns in Detective Comics #30 with a brand new goon who finds his neck broken beneath Batman\'s boot. Batman really liked to break necks in the old days.  For the next two issues, Batman battles against, even when judged today, one of his weirdest foes. The Monk was half-vampire/half-werewolf and perhaps the first creature of mixed heritage apart from Hercules to appear in comic books. I think he even beat out Aquaman and Namor. Using silver bullets and a .45 Smith and Weson automatic Batman dispatches the Monk and Dala his female vampire servant. They were already dead, and technically they don\'t add to Batman\'s impressive scorecard. I mention them only for archivists. Though, he does not even question whether or not they are alive or dead. You can easily make the argument that both Angel and Spike are alive, but this Batman was more pragmatic than philosophical and in many ways more pure in his quest to eradicate crime, avenge his parents and prevent others to suffer his loss.  In Detective Comics #33, Batman destroys the dirigible of a would-be Hitler named Krueger. I don\'t actually know how many men were on the airship nor is the number indicated, but the range must be at least between ten to twenty even when accounting for science fiction additions to the design of the craft. As an encore, Batman tosses a knock-out pellet from his utility belt into the cockpit of Krueger\'s escape-plane. Just one for the big brain. No more escape planes nor politically correct parachutes. No survivors. Batman got them all. Kruger, now unconscious, smashes his plane into the river, and \"the body of Krueger was recovered from the water.\"  In Detective Comics #34, Batman chokes Duc Docteru--there\'s that predilection for necks--then lets his car fly off the cliff. Crispy.  In Detective Comics #35, Batman propels a swordsman onto his partner\'s sword. Batman then throws a statue at the mastermind to knock him out the window. Naturally, you think perhaps a flagpole or a copse of trees broke his fall. Ha-Ha-ha-ha-ha-ha-ha! We see blood pouring out his cracked skull in the next panel.  Think Robin\'s arrival lightened the Darker Knight\'s demeanor? Think again. Robin actually is no angel. His slingshot has done more than simply knock out his foes. Batman and Robin during the Boy Wonder\'s introduction in fact calmly watch Boss Zucco toss one of his men off a girder at a mob bought construction site. Robin even takes a snapshot. One for the police and one to pin to his wall. I made that last bit up. Robin didn\'t pin the photograph to his wall. He probably kept it in his wallet.  In Detective #39, Batman pushes a giant jade idol and a heavy throne onto six, count them, six members of the Tong of the Green Dragon.   Not at all afraid of repetition, Batman uses a heavy marble statue to kill three of the Joker\'s men. The Joker incidentally miraculously escaped a car crash that the Batman tried to orchestrate. Hey, cut the guy some slack. Batman probably single-handedly killed every mobster in Gotham City. Perhaps, he greeted any newcomers at the train depot and snapped their neck.  In Detective Comics #47, Batman whacks a hood over the head and lets him plow his car into a phone pole. What I believe to be Batman\'s final act of killing occurs in a classic battle against Hugo Strange. Batman after slipping a noose around the neck--the neck again--over one of Strange\'s Monster Men whisks the poor creature off the ground and uses the machine gun mounted on the Bat-Plane to riddle him with bullets. It should be noted that this is the only death over which Batman has felt any sorrow. His thoughts aloud went something like this: \"It saddens me to take an innocent human life, but in this case, I have no choice.\" It should also be noted that his thoughts do not contradict his previous characterization. Simply put, criminals are not innocent, and I doubt the original Batman envisioned by Bob Kane and Bill Finger believed they were even marginally human. The Punisher? More like the Pussycat.\"  '),(192,715237,'05,August,2004','2004-08-05 00:00:00','   I thought I\'d put up something I found today. Its my answers to one of those stupid email surveys that your lame ass friends always send you. I have not idea why I kept it, but I think though it\'s a bit dated (about 2-3 years old) it does show some insight into the labyrinth I dare call a mind. I have provided an updated commentary in parenthesis for some of the answers. Enjoy:    WHAT BOOK ARE YOU READING NOW?  SICK PUPPY by Carl Hiassen (about to start THE ADVENTURES OF KAVALIER AND CLAY by Michael Chabon when I get the energy. What? It\'s like 1000 pages.)  WHAT\'S ON YOUR MOUSE PAD?  I\'m on a laptop stupid.  FAVORITE BOARD GAME? Candy Land   FAVORITE MAGAZINE?  Cat Fancy (ahh such a stupid and unfunny answer)   CURRENT FAVORITE SONG? Miss you So Badly by Jimmy Buffett (Never Let Me Down- Kanye West)  FAVORITE SMELLS?  Meat on a grill.  FAVORITE SOUNDS? A woman screaming my name. (make that plural: women)  WORST FEELING IN THE WORLD?  A woman yelling another mans name.  WHAT IS THE FIRST THING YOU THINK OF WHEN YOU WAKE UP IN THE MORNING? Damn I\'m late for work!Aw, fuck it.(Work is for suckas)   ROLLER COASTER:  SCARY OR EXCITING?  Scary (Tell anyone I said that and I\'ll kill you.)  HOW MANY RINGS BEFORE YOU ANSWER THE PHONE?  None, I can\'t hear it cause I\'m on the goddamn internet filling out stupid surveys like this. (Ah the days of dial-up internet)  FUTURE DAUGHTER\'S NAME?  Wishyouwereaboy (Ahhh... misogynism.)  FAVORITE FOODS?  Cereal (I\'m currently on cereal hiatus it was getting out of hand. So currently I\'d answer BBQ ribs.)  DO YOU PREFER CHOCOLATE OR VANILLA?  Chocolate  DO YOU LIKE TO DRIVE FAST? My driving record would indicate yes.  DO YOU LIKE STORMS?  Yes  WHAT TYPE WAS YOUR FIRST CAR?  Nissan 200  IF YOU COULD MEET ONE PERSON, DEAD OR ALIVE, WHO WOULD IT BE? My real daddy? (sorry Dad)  FAVORITE DRINK?  Guinness (that stuff\'s too expensive, lets just say beer in general)  DO YOU EAT THE STEMS OF BROCCOLI?  Broccoli sucks ass  IF YOU COULD HAVE ANY JOB YOU WANTED WHAT WOULD IT BE? Britney Spears personal inseminator. (Still haven\'t heard back on that one, maybe I should send a follow up letter?)  IF YOU COULD DYE YOUR HAIR ANY COLOR, WHAT WOULD IT BE? Hair? What hair? (Sad but true)  EVER BEEN IN LOVE? Uh... real men don\'t fall in love...yeah thats it.  IS THE GLASS HALF EMPTY OR HALF FULL?  It better be all the way full or I\'m gonna beat the bartenders ass.  FAVORITE MOVIE(S)?  \"The Empire Strikes Back\" (\"The Big Lebowski\", \"The Good,The Bad and The Ugly\")  DO YOU TYPE WITH YOUR FINGERS ON THE RIGHT KEYS? I use my feet. (You don\'t really want to know)  WHAT\'S UNDER YOUR BED? Jimmy Hoffa (Lacy Peterson---damn going to hell for that one)  YOUR FAVORITE NUMBER?  5 (21, Blackjack bitch!)  FAVORITE SPORT TO WATCH?  Carolina Tarheels Basketball and Panthers football  SAY SOMETHING NICE ABOUT THE PERSON WHO SENT THIS TO YOU:  John Overman is a kindhearted, open minded, generous, forgiving person.(Whats one more lie on my road to hell?)  PERSON YOU SENT THIS TO WHO IS LEAST LIKELY TO RESPOND? George Dubya Bush (Though I bet the CIA knows I\'m typing this :))   -J La  '),(193,715237,'04,August,2004','2004-08-04 00:00:00','   Rambling Incoherence By Jason Latour       Oh where have you gone MJ?     Once again the Summer Olympics are upon us and as usual the only sport I really care about is basketball. Yeah Gymnastics and High Diving and competitive muff diving are great and all but basketball is the only sport that really grabs my interest.   The reasons are pretty obvious. Basketball is my favorite sport (next to football). I also possess a vast knowledge (hey that’s debatable I agree) of the sport and of the players themselves. And third there are no judges, usually pale and gauntly thin psuedo-erudites feigning interest as they flip through an East Latvian porno mag and try not to ash on their designer silk Capri pants.   Watching the original DREAM TEAM (the first US team featuring pro players and comprised of such legends as Michael Jordan, Larry Bird, Magic Johnson and Charles Barkley) run through the other countries like a young Hugh Hefner at a sorority house, using and abusing them in every orifice and position, was one of the highlights of my young life.  It was truly amazing to watch, as other teams seemed more interested in taking photos with our team than actually trying to compete against them.   Lately though our guys (being the US, and the NBA players in particular) have pretty much fallen flat on their collectively rich asses. It\'s a disturbing and harsh reality that the rest of the world has truly leveled the playing field, as evidenced by our disastrous performance in the World Championships two years ago, our 17 point loss to Italy (friggin\' ITALY!) yesterday and our 3 point win over Germany at the buzzer today.   Let me say that again. It took a 3-point prayer at the buzzer by Allen Iverson to beat GERMANY! Germany, a team that hasn\'t even qualified for the Olympics yet. FREAKING GERMANY, with it\'s great international stars like former Tar Heel Ademola Okulaja and The Baron Helmut Von Zemo! (Yeah I know they have NBA Allstar Dirk Nowitzki)   Christ, we had an easier time with NAZI Germany.   Of course as always excuses abound. There is a laundry list of NBA superstars such as Shaq, Kobe, KG, etc. who turned down the team. We don\'t have the time to gel like other teams. We have to go to play in countries where they use the metric system.   Bill Simmons, a writer for ESPN and Jimmy Kimmel recently wrote an article about this subject that I think pretty much sums up how I feel about it all. You can read it here:   http://sports.espn.go.com/espn/page2/story?page=simmons/040804  Over the course of which he outlines the team he would have assembled for the Olympics. Though I find both his reasoning and selection process pretty solid, I still don’t know that I really agree with his team. Here is my “dream” team, selected with the though that the guys who turned down Team USA would still do so.   HEAD COACH –  Larry Brown- Best coach around period. Just give him something to work with and it’s a cakewalk. Let him pick his own assistants, I’m sure he knows better than I.   STARTERS G-Dwayne Wade-Selected on pure talent. A stud superstar in the making. Plus he’s able to penetrate and finish better and stronger than most of the other guys available.  G-Richard Hamilton – Selected with the same caveat as Simmons that he would have accepted had he been asked earlier. A winner and perfect team player who possesses the best mid range game around, he would destroy any zone defense and run wild under international rules  F- Michael Jeffery Jordan- Ahhh the curve ball. Yeah here is where I go out there a little bit. A last hurrah for MJ. The greatest player of all time would still be able to lace them up and go out and compete with the best of them, let alone international players. Sure he can’t play all the minutes you’d like but he brings an intimidating presence to the court along with an amazing inherent knowledge of what to do on the court under any situation. His mid range shooting and post game would terrorize opponents. The other teams would be so busy either drooling over meeting him or wanting to best him that it would open up countless opportunities for the other guys.   F-Tim Duncan- MVP candidate year in year out. Champion. ‘Nuff said.   C-Rasheed Wallace- Since Ben Wallace turned down the team; why not get his running mate? Right at 7 feet, he\'s a great shot blocker and shooter who has a Euro’s touch and a street baller’s toughness. Sure he has attitude, but after playing for Larry Brown and winning a championship, its been proven that he can sacrifice for the good of the team.   BENCH  F Antawn Jamison- Sort of a tweener, but a guy who hustles and possesses all the intangibles. He can score in bunches without having plays ran for him and likes to rebound.   G/F- Michael Redd- Like Simmons I take this guy for his shooting if nothing else. Everything else he brings, which is a lot, is just butter.   F-Robert Horry- One word- clutch. A proven winner who rises his game when it counts. Provides passing, shot blocking, outside shooting, leadership and experience. Doesn\'t need to play or touch the ball but when he does he\'s definitely a threat.  G/F- Ron Artest- A mad dog on defense who chews metal and spits nails. Above average offensive skills and shooting ability. Only question is attitude, something that surrounding him good teammates and coaching eases that worry.   G-Lebron James- Plays multiple positions. Ridiculously talented. Unselfish and willing to be coached. Plus it doesn’t hurt that he is Lebron friggin’ James.   C-Brad Miller- A banger. A winner. A dirt dog on the boards. He would bring toughness and heart to a team needing it.   G. Richard Jefferson- A world class athlete, a winner and a guy I just like.   G – Chauncey Billups.- A great team player, a winner and a point guard who can slash, shoot and defend. Makes the cut above Sam Cassell due to his youth.   ALTERNATES   Carmelo Anthony- Another young stud superstar on the rise. Unfortunately there are just too many people at his position.   Sam Cassell- Again, a proven winner who can run the offense and score in bunches. Wouldn’t need to play a lot of minutes, but could provide tons of experience and saavy.   Wally Szczerbiak- As much as I bash the guy, he is one of the only pure shooters available. So long as he’s not counted on to score he could ease the load on the big guys by opening up sagging defenses.   Now I realize there are some holes on this team. We’re sacrificing 3 point shooting for mid-range ability, but with guys like Duncan, Wallace and Jordan who can all produce down low and a shorter three point line, you’d get a lot more wide open looks and you need guys willing to step in and take the easier shot as opposed to throwing one up wildly. There is also the lack of a true point guard, but hell outside of college you generally run into that problem in American hoops.   Anyways that’s my rant. I haven’t given up hope on our current team. I think they can still win it, which speaks more to my faith in Larry Brown than anything else. I do think the egos aren’t so great on this team and with some time to work out the kinks they could possibly become a force.   I just pray I don’t have to witness the sickening site of our guys piling on after hanging on to a 1 point win over Botswana.  '),(194,715237,'03,August,2004','2004-08-03 00:00:00','   Rambling Incoherence  By Jason Latour      OFFICE SPACE was REALLY on to something.    In an effort to swing this blog a little more into a free for all of Rambling topics and not just the typical tirade of bad comedy that it normal it is, allow me to wax philosophical for a moment.  It is official.  I’m back in the “Real World”.   That’s right no more hanging out on college campuses, stalking women decades my junior. I have officially withdrawn from grad school and am in the process of transitioning my life back to that of a “normal” 26-year-old adult.   Well---as normal as you can be for a cartoonist/writer/aspiring filmmaker with no job.   Not that I’m really concerned. You see the past 2 years or so I’ve managed to get by on some strange combination of luck, inventiveness and raw desperation and over the course of that time I’ve realized one thing---  Work is for suckas.  That’s been my credo for a while. But before you go jumping to conclusions, let me explain it.   Work, at least in the traditional sense of a nine to five, mindless drudgery where one just shows up in order to live check to check, is for suckas.     Honestly, I recommend to anyone and everyone---unless you truly love what you do---quit your day job. Yes that job that is dragging your soul down into the pits of depression and anxiety and wasted potential. Junk that bitch with a quickness.   I want you to realize something; I am in no way condoning just sitting on your ass the rest of your life and doing nothing. In fact I personally very much value  and respect hard work and sweat and study. So no I\'m not suggesting you sit on your ass, collect unemployment and watch Dave Chapelle all day (though it is a GREAT show). I\'m suggesting that you find what it is that you have a passion for and seek to make THAT your occupation.   Yeah I know that this sounds all cheesy and inspirational and from me  (a guy who has I guess achieved very little in the traditional sense) it might not hold much water, but I truly believe it’s the only way to true happiness. Sure it’s a pretty idealist and straightforward way to look at it. You’re right it is odd to hear this speech from my typically cynical ass. And yes, for some it’s a much easier road to hoe, but I think the rewards, no matter how tough they are to achieve, usually outweigh the risks.   For a long time I did a comic strip called 4 Seats Left (www.4seatsleft.com when it was up and running). But in the past year or so I’ve sort of abandoned it. Moved my passions on to other pursuits, other interests—writing, etc. Over that time I’ve often wondered why I felt unable to connect with the characters, why I felt unable to write it anymore? Had I simply outgrown it?   You see the main character, Bernie, is basically worthless. Middle twenties, lives at home, no job, no ambition, just a life of ease. But see that’s is all just parody. Parody of the way I feel on those mornings when I get up and feel that I’m a long ways away from what I’m aiming for. And maybe on some level that’s why I stepped away from the strip for a while, because I felt that the character (who was lets face it based on my “dark” side) didn’t reflect me anymore. But now I realize that Bernie is what he always has been, and that is a parody, a coping device for looking at myself objectively at that lack of drive and that complacency, which is really just insecurity, that most artists of any kind feel. And a way for me get that out into the world, have a chuckle about it and then get myself going again.   And in some ways this blog is a lot like that too. All my work reflects some sort of aspect of my personality and I think I’m beginning, after a long time, to understand that.   So find that thing that you’re good at, that you have an aptitude and passion for, and DO IT. It could be becoming a racecar driver, an artist, a doctor, a lawyer, a world dictator, a porn fluffer, it doesn’t matter. Do it. Even if it just starts out as a hobby, because I promise over the course of it you will learn more about yourself than you ever thought.   So maybe it’s not “Work is for suckas” that is my credo.  Maybe it’s merely “A day job is for suckas”  '),(195,3295265,'11,May,2004','2004-05-11 00:00:00','   It started off as a fine day. Got up early, came to work. As usual, my PC would give me some slight problems, but after a while, it would be okay.  New colleague really giving me lots of headache. Taught him thousand and one things and all he can remember is only 0.01% of everything. This is a real saddening case. Wish that I can just fire him and save my hair from turning all white.  Half hour to lunch. Wonder what\'s there to eat? Let\'s hope no more LG1 food. Getting sick and tired of the food there. Making me shirt and etc stink like crazy.  Toothache seriously aching like crazy now. Must see doc. Most probably, whole tooth will have to come out. Damn.  Phones also giving problem. LCD problem. aiii. Why do these LCDs always give humongous problems? Really cannot stand it already. Everytime an LCD fails, it stays in the workshop for dunno how long.  KTMB also another one. Giving problems like nobody\'s business. Wanna give them busines, they dun wan. Real tired of them already. Called them dunno how long, they dun answer. :( Hope they reply my sms.  After work, had a great time. Jessy called. We just talked about everything and anything. She was at the pasar malam. Then I told her that I\'ll call her right after she gets back. I did.   We talked and talked and talked for 2 hours and 30 minutes. We are just normal friends. We talked bout unique names for her kids (Zanique, Danielle, Cyclice / Celynn) and she commented a few for me (I Lai, He Lai, She Lai, They Lai, We Lai, It Lai). Both of us were just plain talking. It\'s hard to find someone that you can click to in just a short while. Plus, we did more than just click. We were close, quite close friends already.  Her boyfriend\'s coming back tomorrow. She must be really happy. Can tell from the tone on her voice.   Well, that\'s all for today. See ya.  '),(196,3295265,'18,May,2004','2004-05-18 00:00:00','   It seems to be a very long day today. Feeling sleepy and stuff. :(   Knight Online\'s gonna start charging really soon. Don\'t know whether to pay or not to pay. Why do these online games have to charge? Can\'t they remain free throughout ur entire life? I would seriously love free P2P games.  Samsung phones are really getting the best of me. I seem to be bored of Samsung phones already. I do not know why. They are basically something like Nokia also. Same phone functions, different name only.  '),(197,791840,'11,September,2002','2002-09-11 00:00:00','    >  This is not my blog, it\'s from yahoo.com  >As many of you know, I recently relocated to Florida.  As part of >getting my life back in order, I found and went to a new primary care >physician.  While discussing my medical history in detail with this >doctor, I brought up my history of ADD.  I was flabbergasted to hear >the doctor respond by saying that he didn\'t believe that ADD was a >real disorder, but rather one that was created in the late 60\'s and >early 70\'s to account for those that just couldn\'t deal with hard >work and discipline. Further, he said that it took a doctor with real >guts to prescribe a methamphetamine-based stimulant to anyone, let >alone someone who couldn\'t focus on their responsibilities. > >Yes...I was ready to strangle this guy.  Instead, I spent the next 5 >minutes rattling of information, research, web addresses and relevant >anecdotes to try and convince this guy that there was more to this >story than his perceptions of laziness. > >I don\'t think I got through.   Anyone else run into this before? > >BIG > >hey big this is mickey, and i have had a similar sun in with a pediatrician with my daughter. i  found a new one and got her back on track. this was after extensive testing  through her psych and documentation of add, depression, bouts of anxiety and  ptsd too. so it does happen and i just looked until i found someone who  actually cared about my daughter and her problems. good luck in your search  and your move. hugs, mickeymouse > >________________________________________________________________________ >________________________________________________________________________ > > > >Your use of Yahoo! Groups is subject to http://docs.yahoo.com/info/terms/ > >     _________________________________________________________________ MSN Photos is the easiest way to share and print your photos:  http://photos.msn.com/support/worldwide.aspx    ________________________________________________________________________ ________________________________________________________________________  Message: 2    Date: Tue, 25 Jun 2002 23:24:44 -0000    From: \"ryan1978_48415\"   Subject: Re: Digest Number 33  Oh my goodness I remember on ICQ well anyways I happen to be Vegan  nowadays but this one girl who was Vegan and later went back to Lacto- Ovo Vegetarian sent me an IM back when she was Vegan and I was Lacto- Ovo Vegetarian I had I dunno how many conversations with her but last  time we chatted and I mentioned about me having Autism and ADHD she  said that she didn\'t believe there was any such thing as ADD and that  she believes the Capitalists made that up to make money or whatever  she did end up admitting that she didn\'t know much about Autism  though after she was offline and I sent her a long note.I decided to  remove her from my ICQ friendlist later on cause as far as im  concerned the ignorant might as well swallow their words up.Oh my  goodness I also remember this one Canuck that I have chatted with(I  have a little bit of Canadian heritage myself so I am entitled to say  Canuck)anyways when I had made mention of where I pointed out the  fact that I am sure that i\'d know what to do if I do pass Autism or  anything else to my future kids this Canuck called me Cruel or  something I tried getting her to open her eyes but she\'s obviously an  ignorant and prejudice type of person.I was actually in an Autism  Chatroom that same time she didn\'t realize I was there but I asked  people in the room including the Boss of the room what I should say  to the ignorant person and finally I just said to her I Hope God  Forgives you for what you\'ve done she then said thanks and then she  said\"Let\'s not bring God into this\"I immediately decided that I  definately don\'t wanna talk to her no more.Ya know what though I am  left to wonder what if she actually has kids in the future and one of  em is an Autistic kid is she gonna tell that kid that it would be  cruel of him/her to have kids?   --- In ADD_Couch@y..., \"Gina Goodman\"   wrote: >  >  >  > >From: ADD_Couch@y... > >Reply-To: ADD_Couch@y... > >To: ADD_Couch@y... > >Subject: [ADD_Couch] Digest Number 33 > >Date: 25 Jun 2002 09:52:42 -0000 > > > > > >To unsubscribe from this group, send an email to: > >ADD_Couch-unsubscribe@y... > > > > > >------------------------------------------------------------------- ----- > > > >There is 1 message in this issue. > > > >Topics in this digest: > > > >       1. A run-in with my primary care physician... > >            From: \"biganburly\"   > > > > >  >_____________________________________________________________________ ___ >  >_____________________________________________________________________ ___ > > > >Message: 1 > >    Date: Mon, 24 Jun 2002 15:08:11 -0000 > >    From: \"biganburly\"   > >Subject: A run-in with my primary care physician... > > > >As many of you know, I recently relocated to Florida.  As part of > >getting my life back in order, I found and went to a new primary  care > >physician.  While discussing my medical history in detail with this > >doctor, I brought up my history of ADD.  I was flabbergasted to  hear > >the doctor respond by saying that he didn\'t believe that ADD was a > >real disorder, but rather one that was created in the late 60\'s and > >early 70\'s to account for those that just couldn\'t deal with hard > >work and discipline. Further, he said that it took a doctor with  real > >guts to prescribe a methamphetamine-based stimulant to anyone, let > >alone someone who couldn\'t focus on their responsibilities. > > > >Yes...I was ready to strangle this guy.  Instead, I spent the next  5 > >minutes rattling of information, research, web addresses and  relevant > >anecdotes to try and convince this guy that there was more to this > >story than his perceptions of laziness. > > > >I don\'t think I got through.   Anyone else run into this before? > > > >BIG > > > >hey big this is mickey, > and i have had a similar sun in with a pediatrician with my  daughter. i  > found a new one and got her back on track. this was after extensive  testing  > through her psych and documentation of add, depression, bouts of  anxiety and  > ptsd too. so it does happen and i just looked until i found someone  who  > actually cared about my daughter and her problems. good luck in  your search  > and your move. > hugs, > mickeymouse > > >  >_____________________________________________________________________ ___ >  >_____________________________________________________________________ ___ > > > > > > > >Your use of Yahoo! Groups is subject to  http://docs.yahoo.com/info/terms/ > > > > >  >  >  >  > _________________________________________________________________ > MSN Photos is the easiest way to share and print your photos:  > http://photos.msn.com/support/worldwide.aspx    ________________________________________________________________________ ________________________________________________________________________  Message: 3    Date: Wed, 26 Jun 2002 00:27:00 -0400 (EDT)    From: Constance SChlievert   Subject: Re: A run-in with my primary care physician...    Hi Big, yes as a matter of fact i have run into this in dealing with my own family\'s care, I know this is  hard to deal with a physicain in this manner and all i can offer you is this. Do not waste your  time wth him, He is obveously to lazy and uneducated to deal with the fact that add/adhd is real.  Get yourself another physician. you may even want to research the physicians a bit, before you make  a choice. I actually recieved a letter from a physician  at the center concerning one of the kids   that i get in the mornings saying the same darn thing. Yes dealing with people especialy  physicians can make us very angry, I have been struggleing with  a few here for a while you just can not get through. i have learned they are just not worh the  argument.   biganburly   wrote: As many of you know, I recently relocated to Florida.   As part of  getting my life back in order, I found and went to a new primary care  physician.  While discussing my medical history in detail with this  doctor, I brought up my history of ADD.  I was flabbergasted to hear  the doctor respond by saying that he didn\'t believe that ADD was a  real disorder, but rather one that was created in the late 60\'s and  early 70\'s to account for those that just couldn\'t deal with hard  work and discipline. Further, he said that it took a doctor with real  guts to prescribe a methamphetamine-based stimulant to anyone, let  alone someone who couldn\'t focus on their responsibilities.  Yes...I was ready to strangle this guy.  Instead, I spent the next 5  minutes rattling of information, research, web addresses and relevant  anecdotes to try and convince this guy that there was more to this  story than his perceptions of laziness.  I don\'t think I got through.   Anyone else run into this before?  BIG   To unsubscribe from this group, send an email to: ADD_Couch-unsubscribe@yahoogroups.com    Your use of Yahoo! Groups is subject to the Yahoo! Terms of Service.     --------------------------------- Post your ad for free now! Yahoo! Canada Personals   [This message contained attachments]    ________________________________________________________________________   '),(198,791840,'10,September,2002','2002-09-10 00:00:00','    As well, please enjoy the great feeling of stating your own opinion without feeling that you are being watched...there is good security in this.   '),(199,791840,'10,September,2002','2002-09-10 00:00:00','    This is a Blog for the Green Family to keep in communication with each other. Please use the site for communication in the family.   9/10/2002 6:45:46 PM | Stephen Olson] Well today I was at Dr. Dwyer\'s and managed to do some real good work with feeling accomplished. Also Richard from the CCA took me to Granville Street and Broadway to try to get into the new course through the Vancouver School Board. This is to complete the grade 11 math that I have to complete and it\'s better to do this in a structured setting. All I really need is to get a transcript from the Counselling office and then mange to give this to the 41st and Fraser location the John Oliver school location.   Things that are not right for me:   1. I sometimes have difficulty with defensiveness.   2. I do have trouble with relationships that are close to me, intimacy.   3. I have trouble with working with people who are close to me intimatly.   4. I have trouble with personal awareness many times.   5. I feel that I have to ask questions to feel at ease with others so that they will accept me and so that I will stay in their boundries.   6. Taking risks is difficult for me sometimes.   7. Feeling comfortable with others has always been a challange, and sometimes doing the basic requirements of life is difficult for me.   8. Sometimes I have trouble with procrastination with things that I don\'t want to do.   9. Budgeting and future planning is a challange for me.   10. Organizing is fine if I have the tools to do so, and time planning usually is the most difficult because it takes the most time to do.   11. I have difficulty using the telephone correctly sometimes esp. when it comes to personal converstaions with family.   12. I find myself being very able to handle a restricted lifestyle, but when it comes to a more complexities I tend to run into problems.   13. I find knowing and using skills that I have already quite enjoyable but learning new skills seem to take time developing, especially person to person skills.   I have to say to Mom that, I need to claim my independence of family even though I am almost 40 years of age at the moment.  Mom, you have done this yourself, and have continued to do this over time. I on the other hand have a non-grown child inside of me that is not fully a developed child.   My child\'s development has been stiffled by the fact that I have been put in a protected environment, and the outside world has not had any way of getting to me for a long time. So obviously I could not learn about this world since I was apart from it.  My difficulty now is that I need to know more about this world and understand more fully what this world means to me personally.   Inside my confidence has not grown in the world, since I was not in the world but kept apart from the world. You, Mom, have claimed your independence, and are a fully developed adult, and I need the chance to become one myself. I am just now comming to the point where I am comming out of the house emotionally as an adolecent.   This is a very crutial point.   I can feel shure and say that I feel reall good that this time has come. I now have to learn about the world around me much more than  others who are already fully developed adults.   Yes, I have asked for help from you, Mom, and asked you to be my coach, but you have to understand something. A child does whine, and cry when they are not feeling safe and a teenager does go against the wishes of their parents. I never really rebeled, I never really stated my independence from you.   I feel that I need to claim this independence, and state my position and claim my individualness as every adult has done.  I understand that nobody really rebells outrightly most of the time but they do claim their space and their own area of personal development.  You already know that I am very independent, and have always been independent and have had a hard time being together with another person because of my need for freedom.  This need for freedom, is very important for me.  I need to feel free enough to be able to state myself, and be free thinking.  As well as this motion is very important to me.  I need to be able to get away from things physically, and exert energy to this end.    I also need to have home as a place of rest and relaxation, and a place to get away from the world and escape into my own reality.  I ask for your help, and for Ron\'s, but I need this in moderation.  The problem for me is figuring out what moderation is.  What I need help with is finding out the middle ground, and the moderate.  I tend to do the extream thing whatever it may be. I tend to go beyond what is normal, and I don\'t do enough of what is within the boundries of normal.  The idea here generally is that I need to be moderated, but in a way that is not so direct maybe.   '),(200,791840,'12,November,2003','2003-11-12 00:00:00','    urlLink Vancouver Community College - News & Events   '),(201,3544826,'08,june,2004','2004-06-08 00:00:00','   This was a big moving weekend.  Now we are living in a cavern of boxes.  The question is always \"where is my...\" or \"where is the...\"  I was still packing while the movers were carrying stuff down.  It was so much work! At one point I sneezed and saw stars.  I never saw them so clearly.  Usually there is just a vague swimming something that disappears quickly, but this time I saw little perfectly round pinpoints of colored light that swam in little arcs, leaving fading trails behind them.  My little girl (my real daughter, not my inner little girl) is desperate.  What with all the moving and hauling and searching for things and cleaning and stuff, there is no one to play with her, to pay her attention, to take her to the park.  Remember \"it takes a village\"?  Where is the damn village?  We had a babysitting come by for a couple hours, and it was beautiful.  '),(202,3544826,'04,june,2004','2004-06-04 00:00:00','   In one of Bush\'s rare   press conferences  , he was asked who in the world we would be handing the Iraqi government to on June 30th.  Bush, as much in the dark as anyone, replied, \"We will find that out soon. That\'s what [UN envoy] Mr. Brahimi is doing; he\'s figuring out the nature of the entity we\'ll be handing sovereignty over.\"  In other words, you\'ll find out when I do.  The difficulty for the White House is that no one there knows anything about Iraq.  This is why they first called for elections and afterward realized that a free election would probably make Iraq a Shiite state (similar to Iran).  Whoops!  At first, it seemed like  urlLink  a complicated caucus-based election   would be the best way to control the result, but Iraqis saw through it.  So the White House threw up their hands and appealed to Brahimi...  \"The problem was that the Americans, after first agreeing to [Brahimi\'s] idea of a technocratic government, had changed their minds. They accepted the complaints of their friends on the governing council that they could not all be shunted aside. The Americans were also afraid that genuine independents might call for a US troop pullout.\"  '),(203,3544826,'04,june,2004','2004-06-04 00:00:00','   \"A poor fellow with a craze, Sir,\" said Mr. Dick, \"a simpleton, a weak-minded person â€” present company, you know!\" striking himself again, \"may do what wonderful people may not do. I\'ll bring them together, boy. I\'ll try. They\'ll not blame me. They\'ll not object to me. They\'ll not mind what I do, if it\'s wrong. I\'m only Mr. Dick. And who minds Dick? Dickâ€™s nobody! Whoo!\"    urlLink David Copperfield  , Charles Dickens'),(204,3544826,'29,july,2004','2004-07-29 00:00:00','   I saw a few news celebrities:  Chris Matthews talking on his cell phone, Tony Blankly yakking with somebody.  So it was not very exciting.  But do you know who is all over town?  LaRouche\'s people.  I was sitting outside Faneuil Hall, where unfortunately, they\'ve staked out a prime piece of the marketplace.  Over to my left is Chris Matthews \"Hardball\" stage, and in front of me are a bunch of the LaRouchers trying to hand out their campaign literature.  I tried to give off a \"not interested\" vibe, but in the end one of them came to talk to me, and I figured what the hell.  I talked to her.  It was not a pleasant conversation.  Everything I said, she contradicted.  She changed things I said to bend them to her point of view.  She had an answer instantly ready for any political problem.  Not a practical solution, though: just drastic, crazy stuff, like \"put all the banks into receivership\" and \"start bankruptcy proceedings on the IMF.\"  In the end, she started getting aggressive, and so I left.  I\'ve heard the LaRouchers described as a  urlLink political cult , and that fit with the weird feeling I got from the girl.  '),(205,3544826,'26,july,2004','2004-07-26 00:00:00','   \"...surely every grandeur groans for an orgy of quiet recollection.\"   -- Taylor Brady'),(206,3711805,'22,June,2004','2004-06-22 00:00:00','   This is actually my second attempt at maintaining a blog. I really dunno what happened to the first. But what some eyewitnesses tell me is that it was first washed off by a tidal wave larger than the one in \"The day after tomorrow\".  Hmmm...there have been wierd excuses than this actually. In fact, I was bereft of technology when I was doing my summer internship and modern civilization lost on about two months of the most sublime and supine thoughts that ever struck me. maybe, they will be lucky this time.  '),(207,3711805,'31,July,2004','2004-07-31 00:00:00','   Hello. Is there anybody in there?  Just nod if you can hear me. Is there anyone home?   Come on, now. I hear youâ€™re feeling down. Well I can ease your pain, Get you on your feet again.  Relax. I need some information first. Just the basic facts: Can you show me where it hurts?   There is no pain, you are receding. A distant shipâ€™s smoke on the horizon. You are only coming through in waves. Your lips move but I canâ€™t hear what youâ€™re sayinâ€™. When I was a child I had a fever. My hands felt just like two balloons. Now I got that feeling once again. I canâ€™t explain, you would not understand. This is not how I am. I have become comfortably numb.    â€œComfortably numbâ€ by Pink Floyd has always intrigued me. I have not seen â€œThe wallâ€ but it does not require a rocket scientist to figure out that the song is about death. If you are crazy about this song like I am, read on.   To understand the song, you first need to understand the Wall. The entire Wall is about a troubled life. The songs have different perspectives to different people. Pink grows up without a father, a father lost to war. In a life with no father, only hearing stories, and such, Pink turns to music. He writes lyrics, words being his only outlet of expressions. Now, he only has a mother, and his music. His mother keeps him under his wing. She is afraid of loosing him, like her husband. She helps construct a mental \"wall\"  Throughout his life, he decides that he wonâ€™t let anyone take take advantage of him. He falls in love with a woman and when he looses her, he decided to commit suicide. This is the part where â€œcomfortably numbâ€ comes on. It is how the shots are affecting his body and he is feeling numb where \"you are only coming through in waves, you lips move, but I canâ€™t hear what you are saying\"  PS I wont be blogging for a week. Yeah ...you can go ahead and pop the champagne.  '),(208,3711805,'29,July,2004','2004-07-29 00:00:00','   WOW!  This week has sure been very busy for me. I did not have time for the customary 25 daily trips to my blog. You would notice that by the value of the counter.  I have some very good news on the placement front. My next door neighbor got a pre-placement offer from WeP peripherals. Vijay has got a great pay package from WeP (which I cannot divulge for security reasons - my own security that is). We had a huge party the same night where we were all up until 4 in the morning and stuck our tongues out at songs like â€œpapa kehte haiâ€ (It must have been a loser who wrote that song, nevertheless it is an awesome song), â€œNovember rainâ€ (when it was actually July rain).   It was also great to see India win in style after quite a bit of time. Nice to see, they have not dropped the winning habit like a hot potato.I go off to Hyderabad in a couple of days and will be there for a week. There is something about the city that I just adore. A funny incident occurred when some of us were in Hyderabad a couple of years ago. We had heard so much about Hyderabadi biryani and we really wanted to dig into it. The conversation with the waiter went something like this:   Me:  Boss, khane ko kuch milega? (Not the brightest of questions considering that we were sitting in a hotel)  Waiter:  Hou (Not to be confused with its English counterpart â€œhowâ€, the former usually signifies a dynamic yes disguised with indifference unless ofcourse the Hyderabadi is speaking English)   Me again:  Hyderabadi biryani milega?  (This had him floored. Even the waiterâ€™s keen sense of judgment failed him. Looking back at it today, I have a feeling the word â€œHyderabadâ€ confused the poor soul. Maybe, that was because we were in Hyderabad!)   Waiter again:  uhh â€¦. Hyderabadi ?? Biryani â€¦haan Saab!  So, do they call American chopsuey just â€œchopsueyâ€ in the US? Do they play just â€œcheckersâ€ in the land of the dragons? Or Worse, do they just watch â€œblues 2â€ in Hyderabad??  '),(209,3711805,'26,July,2004','2004-07-26 00:00:00','   I am currently reading \"snapshots from hell\" by Tim Robinson which brilliantly summarizes what B-schools are like. The book is wild and hilarious and it has me in splits whenever I compare my own predicament to his.   Flashes from my own hell (when I was in the first year at NITIE) come back to me. Now that I am in the second year, it does not make things any easier. When I find terms like \"term structure of interest rates\", \"nelson Siegel function form\" and other equally obnoxious terms being flung at poor hapless me, I cant help pitying myself. Even Omar Khatib of Fist of God would have been put to shame at a heinous activity like this - testing the endurance limits of an innocent mechanical engineer whose only fault was that he dreamt of making it big in Business!   People tend to think that having an engineering background makes us whizzes with numbers and the like. I would say this is true - normally. But then there are the people like me who are at the fag left end of the bloody normal curve. I remember every moment of confusion, horror and abject humiliation I faced in my introductory Statistics course. I was stretched to what I thought was the maximum possible physical and mental&nbsp;limit of human endurance. Now I realize how wrong I was!   Six Sigma starts from where the last course on statistics ended. It talks about things that only people like Einstein (and fourteen others who understood his bally theory) would have an idea of. Making a guy like me (with an IQ level of 45 at my brightest) take this course would amount to a crime equivalent to sending a solitary soldier into hostile enemy fire with just a Swiss-knife.   So what prevents me from jumping off my fourth floor hostel room? Apart from the fact that the hostel windows are too small for me to get through, it is the presence of subjects like Organizational behavior, Human resource management and Strategic management, where I can jargon-dump* my way through 3-page cases and still score an AB.   * This is a special kind of  maneuver  which is in vogue at all B-schools. It employs words that are  incomprehensible and verbose  in nature and these words appear in  complex sentence structures  that are created&nbsp;with an intention&nbsp;to discourage the reader. The key to this technique is  underlining jargon  and using a  near-illegible handwriting .  '),(210,3711805,'24,July,2004','2004-07-24 00:00:00','   This issue is more important than life and death itselfâ€¦. Guess! &nbsp;You are hereby certified an Indian by yours truly if u answered â€œcricketâ€. That is what we all live and die for anyway, right?   â€œTendulkar slams doubtersâ€, â€œSachin â€“ Mr Allrounderâ€  were some of the headlines that appeared in the newspapers after India pulverized Bangladesh in the Asia cup. This is becoming a habit for the maestro. Let me make this clear that I am one of the greatest fans of the little master and I do not mean even an ounce of disrespect to him (I love my life, of course). But what makes me irate is, Sachin is sometimes oh-so-overrated by all of us. All it takes is a decent innings by him against even a club-level team and he is splashed all over the media. Wake up and smell the grass guys! I admit Sachin is brilliance personified and watching him is sheer delight, but why go overboard?   Sachin does not play for India because he is an entertainer (We have a full-time Sehwag for that). He is in the team because he is a match winner. Take away the two matches at Sharjah where he manhandled Australia and then bam! We have to think hard for a third game which he won for India. Contrast this to Dravid, who comes one down and we are overwhelmed by the amount of matches he has steered India to victory. This is of course in addition to the innumerable instances where he helped India avoid the ignominy of follow-on or defeat. Very few people who know the game would deny that Dravid in full flow is as classic as it can get. This reminds me of a similar debate between Beckham and Zidane. One brings in glamour and spectators and the other inspires more players into the game.  PS You can comment and add any other game that you think Sachin won for India.  '),(211,3711805,'23,July,2004','2004-07-23 00:00:00','   I am going to keep this short and simple. No complicated humor&nbsp;or unintended puns . This is to say  Bon Voyage  to two of my very best buds. One is starting out on a trans-Atlantic cruiser (lucky bugger!) and the â€œmatrixâ€ has the other.  &nbsp;   urlLink Shivali  joins Infy from the 26th of July and this would be her first venture into the IT industry. She has been a great friend for more than three years now and this is wishing her all the very best in whatever she chooses to do. What she would do in life, this I do not know but, I am sure, she would shine wherever she goes.  &nbsp;  OjaS boards one of those obscenely huge merchant ships and heads off on a voyage which will last more than 8 months. He has been a friend&nbsp;since I was so high&nbsp;and we have had some of the best times in our lives lazing away on the  katta and BH . This is hoping he earns zillionz of non-taxable dollars and remembers me -&nbsp; a friend in need who is always in need.  '),(212,3711805,'22,July,2004','2004-07-22 00:00:00','   It was a normal day, albeit with a difference. You see, most of my days begin at 7:35 in the morning, which gives me enough time to have a leisurely breakfast, check my mail, update myself with some blogs, perform my daily ablutions (note: no newspapers to clutter my mind )and head off to class at 9:04 â€“ which is the latest allowed entry time. The only difference was, â€œtodayâ€ began when GAM woke me up, at 8:40!   I remember the day when we were new in NITIE, we had an alumnus lecturing us about the big bad corporate world. He threw an innocuous looking question at us â€“ â€œwhat is management?â€ After negotiating a volley of answers that were extremely obscene in terms of both practicality and management jargons, he gave his 24-carat answer â€“ â€œmanagement is decision makingâ€  No truer words were ever spoken! For instance, I had the following decisions to make today â€“ all in a space of less than 5 minutes.   DECISION 1:  Do I wake up at all?  PROCESS:  Uhhh....Which was this first class anyway?I am never at my brightest in the mornings.Was it PM ? O my God!!! Hell yes!    DECISION 2:   Do I take a bath?  PROCESS:   This was easy, because there was a high possibility of a quiz being held in this class and I did not want to have people shying away from me just because I stink (Donâ€™t we all do, anyway?). My principle is simple when it comes to quizzes â€“ the more the merrier.   DECISION 3:   Do I have my breakfast?  PROCESS:  Now, this was a very tricky decision to make, something that even Jack Welch would have been proud of. On my way down to the mess, my intuition (so vital for a manager) told me that I had about 8 minutes and 35 seconds left (Duh...I could have also looked at a watch) before this class began. And the â€œstairway to heavenâ€ would take up 5 minutes. So what did I do? Quickly munch on two slices of bread with the hope that this would somehow last me until 1:40.   DECISION 4:  Do I stop writing this blog?  NO PROCESS REQUIRED:  yes  P.S. I was thinking about a way to say something about Nishant Prasad in this blog. Mission Accomplished !  '),(213,3711805,'20,July,2004','2004-07-20 00:00:00','   I was talking to a friend on different ways to make a blog popular. Both of us came to the conclusion that readership for a blog is extremely important and unless the â€œcommentsâ€ column looked full and populated, the whole effort is just not worth it. An â€œuncommentedâ€ blog looks like MF Hussainâ€™s Gajagamini â€“ Lots of big stars, talent in it. It even has an esoteric story line to go with it, but unfortunately no distributors.   Then, there are also the  â€œone blog wondersâ€  which have words of invaluable wisdom on them but the author never demonstrates the same virtuosity with words and worldly wisdom again, in fact he decides that he has done enough penning for a lifetime. I wish I could write once-in-a-lifetime blogs, but I canâ€™t. The second alternative is to keep rambling on and on and on until you donâ€™t have any visitors left. I have clearly chosen the second alternative. So, how do the dickens you know you donâ€™t have any visitors? (No, the blog counter does not show this, it ticks because I visit it daily â€¦.no! make that hourly) A sure-as-litmus test is to ask one of your closest friends in hostel (If you are not in a hostel, do not despair. Try the person in the next cubicle tearing his hair out over a bug) whom you know to be an unflinching reader of your blog to â€œcasuallyâ€ look over your latest blog. If he/she gives an excuse like â€œBut I have to go and wash my clothes now and then blow air over them to dry them. Once this is over, I will wash them again!â€ or â€œAre you aware that the next release 12.11003b is just 8 months away!! I have to work!â€ you know you have had it! Better go back to the more conventional methods of torturing people. By the way, did somebody say â€œThe pen is mightier than the sword??â€ This blog aims to lend a whole new perspective to it.  P.S. This blog is specially dedicated to the one and only GAM of Room no 466, better known to the world (read Banjara hills) as Gaurav Paji. It is because of his tenacity and persistence with my pen-downs that my blog is what it is. Also known as Ice man and the badshah of impromptu fart, he epitomizes everything that is â€œstrategicâ€ in nature.   P.P.S. This blog is also dedicated to all the other people in this world as well.   P.P.P.S I never thought this blog would ever hit double figures and be a month old. But then who thought the Congress would come to power!  By the way, enjoy the budget. No, I am not going to publish my special analysis of Budget 2004.  '),(214,3711805,'18,July,2004','2004-07-18 00:00:00','   Harry : kya main tumhe phone kar sakta hun ? Sally : Haan , mera number directory me hai Harry : Kool !! aur tumhara naam ? Sally : wo bhi directory me hai !  Today, I have decided to start writing short stories. Why on earth the sudden urge to become another O Henry??? Yeah, I agree with you when you say that we have more Henrys than this world can afford, but then I am human and I crave for my moments of glory. Of course that was only the secondary motive behind this. I actually wanted to do my own bit for civilization.   The main and the only characters in this plot are Harry and Sally. They meet in college (like all others do) and before you can say â€œHarry met Sallyâ€ Harry develops this huge crush on Sally. The next â€œlogicalâ€ thing for Harry is to try and woo Sally. Now, Harry is one of those unoriginal fellows who think that being sweet-like-honey to Sally would do the trick. To be fair to Harry, it does work in a way. Sally can now see thru Harry like a transparent bottle. (There is also the galâ€™s â€œsixth sense/intuitionâ€ thingy that a friend of mine keeps talking about .I frankly dunno much about it)   So, the picture at this point reads something like this: Harry likes Sally  Sally thinks Harry is in the â€œfriendsâ€™ zoneâ€  Ouch! That does not make a very pretty reading for Harry. Then the inevitable happens, Sally utters the words every guy dreads ofâ€“ â€œLetâ€™s be friends!â€  So, Harry and Sally become friends, in fact the very best of friends.Contarary to his own expectations, Harry starts enjoying being just-friends with Sally. They start understanding each other very well. So allâ€™s well that ends well? Not quite. Harry starts having these moments of â€œmadnessâ€ where he crosses that threshold of friendship. He crosses the line where friendship ends and something else begins. This naturally leads to unrealistic expectations on the part of Harry and arguments ensue between Harry and Sally who otherwise look very cute as friends.   So the situation now reads: Harry starts oscillating between being just friends and being something more Sally thinks Harry is the undisputed king of the â€œfriendsâ€™ zoneâ€  It is still not a pretty picture for Harry. It pains him when things do not work out the way he wants them to. And also the fact that Sally means so much for him - probably more than anything else in his life troubles him to no end. But then there comes a time in every guyâ€™s life when he must grow up and cease being that 10 year old kid when he got everything he wanted. So, it is time for Harry to realise that some things are not meant to be, and that he should let go. Also , being one of Sally\'s best friends also means living upto it. But does Harry know all this? Would he change for the better?   NB The author has still not decided on Harryâ€™s fate yet. Keep watching this space for more developments between the original Harry and Sally.  Disclaimer: The author would like to make clear that any resemblance to any person dead or alive is purely intentional :P.  '),(215,3711805,'17,July,2004','2004-07-17 00:00:00','   The proven way to concoct a movie so hot that, it has even the Eskimos getting out of their warm clothes.  Disclaimer: The above recipe may not be palatable to people who tend to maintain a serious disposition towards the culinary field. Some might even feel offended by the effrontery of the recipe here. The author begs pardon for this and would like to state that the above concoction has been influenced heavily by personal tastes. He accepts no moral/ethical/political responsibility for any side effects arising out of public consumption of this recipe.  Ingredients: Camcorder (Qty 1) Villains (Bush, Powell, Rice and Rumsfeld) (Qty 4) Controversy (100 ml) A fat beer-guzzling reporter standing in front of the Saudi embassy with the camcorder (Strictly 1, 2 would mean a nuclear holocaust) Gullible audience (The more the better) Media and the internet, preferably manipulated (To taste)  Take the camcorder and set off to the White house on one boring afternoon. Talk to some congressmen and ask them if they would like to enroll their children in the army in Iraq. Mix the four villains with 100 ml of controversy and heat them over a flame until they turn brownish red. â€œStir the potâ€ well until you smell something burning. Remove the pot from the oven and lay the misture on a newspaper-covered plate in the red hot condition. Garnish with some great â€œsneak peeksâ€ in the form of movie trailers on all forms of media. Serve Hot (do I need to mention that?) on a platter to the gullible audience and hope that the audience likes the dish and â€œmunchesâ€ on the crunchy villains.  '),(216,3711805,'16,July,2004','2004-07-16 00:00:00','   It is 1:56 in the night or the morning for the more human people reading this. I am on a high. No, I did not drink. The ambience got to me. Dancing for two hours on the trot does take its toll I guess.   I am no prude when it comes to what I believe people should do or not do. But I do believe in a code of morality that we as humans ought to follow. Does our life have a purpose? Are we bound by some mysterious string which holds us from crossing â€œthatâ€ line? Or are we just an improved form of animals (with the added ability to think rationally) and we have just four objectives in life (sleeping, eating, excretion and mating)?  When I hear/see people puking to glory after drinking too much, girls dancing the way they should not (the word â€œshouldâ€ is again a very subjective term and it should not construed as chauvinism on my part, I just happen to believe that girls for what they are deserve more respect than guys), people carried away in a delirious state, girls going to guysâ€™ rooms, and many more things that I am thankfully not an audience of, it gets me thinking. Is succumbing to carnal pleasures right? If yes, are we different than animals? If succumbing to materialistic desires is wrong, is it completely wrong or are we at a liberty to indulge occasionally? I mean, there must be some line where recreation ends and animal materialism begins. Is the objective of life to enjoy it to the hilt letting no other obstacle come in the way? Or listening to that tiny little voice inside us that sometimes squeaks to pass its judgment on what we do?  So what then is the objective of our life? I am a part of a class of 86 management students who have fires in their bellies and great plans for the way they are going to live life. Almost all of these plans (including mine) revolve around money, power, position and married life. Should the sphere of human thought be expanded to include something else too? If the answer is yes, pray tell me what something else is â€¦â€¦.  The objective of life surely has to be more than: Born: 4th Sep 1981  Died: 24th Dec 2041  '),(217,3711805,'16,July,2004','2004-07-16 00:00:00','   This Blog contains material which might be objectional to certain audiences. The author recommends descretion on the part of the reader.He does not accept responsibility for any mental side-effects that may occur after reading it    I absolutely abhor quizzes !A Quiz is the worst and the bluntest of all methods to prove that your mind is sluggish, you can\'t think fast, you can\'t think right ...and so on. Just how many times do we guys think about women and bathing models and when a question about \"Leela Chitnis was the first what??\" is asked, it just does not come back !!!! Same goes with sex. They say 99% of a guy\'s thoughts are about sex or related activities and the rest is about beer. SO why the hell does our memory fail us with \"Viagra\"??? It should have been a reflex action - When someone says Pfizer , you think Viagra .....  But jokes apart , we had a ball. It is amazing how quizzing can give u the goosebumps and all that. Especially when you are able to crack something like this and you dont do this everyday:    Connect  '),(218,3711805,'15,July,2004','2004-07-15 00:00:00','   It all started in our strategic management class. In one of those discussions where just about everyone is trying to emulate Porter, we come on to a discussion on \"who is a stakeholder in a company??\"  I would be lying if i said that it did not start all right. Conventional answers like \"owners\",\"employees\" and \"shareholders\" started pouring in. This is when yours truly decided to lend his brilliant creative streak to the above discussion by saying that \"society at large\" is a stakeholder. It must have been the late night or the missed breakfast (coz i dont drink), but the bottom line here is I actually said that. To top it , I also substantiated my answer with an example.  Now, all hell broke loose. The definition was quickly amended to include all those directly or indirectly affected by the company\'s performance as stakeholders. This was actually all right until the discussion took a nasty turn when one bright fella said that Standard fireworks has a stake in all companies coz if they dont do well , nobody gets a Diwali bonus and their crackers wont sell.  Somebody even dragged the poor leapord that has been stalking in and around NITIE all this time into the picture. I dont really remember how the poor cat got into this , but all I can do is take pity on it and hope that it never ever decides to eavesdrop again...For its own good ofcourse !  Somebody mentioned something about multicollinearity ....Well , that was the last we heard of that guy as the Prof decided to take things into her hands......   \"Dil pe mat le ....Haath me le!\"  '),(219,3711805,'14,July,2004','2004-07-14 00:00:00','   The sky was blue without a cloud, the grass green and seven members on either side....idyllic setting for a game of Futbol! ( I would like to interrupt here to mention that I am not a Pele or a Maradona ....in fact I am not even a novice when it comes to Futbol ....what do u call somebody who has played barely 4 hrs of football prior to this ??)  The match started off in real earnest. And then I started to weave my magic. I ran like the wind , dribbled like Figo , and drove like rooney. Hmmmm...I donâ€™t know how much of all that really took place...But hey!! This is pure passion ... I just donâ€™t know where reality ended and where the dream began!  SO, I ran like the wind for 25 minutes ....making a few intelligent passes .... But my team members were not as intelligent as I was and they were never in the places I passed to . I even made this Beckham-like-cross But alas! there was no Owen to tuck it into the goal ......  I really donâ€™t understand why people have to be a specialist defender or a midfielder. I mean, look at me ! I am a natural at any place on the field. The opponents soon realized my dexterity with the ball and my speed and they had me marked well ....Maybe that was the reason my team mates were not giving me the ball ....So I ran like the wind for another 25 minutes. You might think that I was not playing an active role by just running around â€¦..But then how many people here have heard of mind games in football â€¦..??  The whistel blew ......Needless to say â€¦.we had lost 2-3 â€¦.  '),(220,3711805,'13,July,2004','2004-07-13 00:00:00','   This is a little lengthy but very inspirational....   God: Hello. Did you call me?   Me: Called you? No, who is this?   God: This is GOD. I heard your prayers. So I thought I will chat.   Me: I do pray. Just makes me feel good. I am actually busy now. I am in the midst of something.   God: What are you busy at? Ants are busy too.   Me: Don\'t know. But I can\'t find free time. Life has become hectic. It\'s rush hour all the time.   God: Sure. Activity gets you busy. But productivity gets you results. Activity consumes time. Productivity frees it.   Me: I understand. But I still can\'t figure out. By the way, I was not expecting YOU to buzz me on instant messaging chat.   God: Well I wanted to resolve your fight for time, by giving you some clarity. In this net era, I wanted to reach you through the medium you are comfortable with.   Me: Tell me, why has life become complicated now?   God: Stop analyzing life. Just live it. Analysis is what makes it complicated.   Me: Why are we then constantly unhappy?   God: Your today is the tomorrow that you worried about yesterday. You are worrying because you are analyzing. Worrying has become your habit. That\'s why you are not happy.   Me: But how can we not worry when there is so much uncertainty?   God: Uncertainty is inevitable, but worrying is optional.   Me: But then, there is so much pain due to uncertainty.   God: Pain is inevitable, but suffering is optional.   Me: If suffering is optional, why do good people always suffer?   God: Diamond cannot be polished without friction. Gold cannot be purified without fire. Good people go through trials, but don\'t suffer. With that experience their life becomes better not bitter.   Me: You mean to say such experience is useful?   God: Yes. In every term, Experience is a hard teacher. She gives the test first and the lessons afterwards.   Me: But still, why should we go through such tests? Why can\'t we be free from problems?   God: Problems are Purposeful Roadblocks Offering Beneficial Lessons to Enhance Mental Strength. Inner strength comes from struggle and endurance, not when you are free from problems.   Me: Frankly in the midst of so many problems, we don\'t know where we are heading.   God: If you look outside you will not know where you are heading. Look inside. Looking outside, you dream. Looking inside, you awaken. Eyes provide sight. Heart provides insight.   Me: Sometimes not succeeding fast seems to hurt more than moving in the right direction. What should I do?   God: Success is a measure as decided by others. Satisfaction is a measure as decided by you. Knowing the road ahead is more satisfying than knowing you rode ahead. You work with the compass. Let others work with the clock.   Me: In tough times, how do you stay motivated?   God: Always look at how far you have come rather than how far you have to go. Always count your blessing, not what you are missing.   Me: What surprises you about people?   God: When they suffer they ask, \"Why me?\" When they prosper, they never ask \"Why me\". Everyone wishes to have truth on their side, but few want to be on the side of the truth.   Me: Sometimes I ask, who am I, why am I here. I can\'t get the answer.   God: Seek not to find who you are, but to determine who you want to be. Stop looking for a purpose as to why you are here. Create it. Life is not a process of discovery but a process of creation.   Me: How can I get the best out of life?   God: Face your past without regret. Handle your present with confidence. Prepare for the future without fear.   Me: One last question. Sometimes I feel my prayers are not answered.   God: There are no unanswered prayers. At times the answer is NO.   Me: Thank you for this wonderful chat. I am so happy to start the day with a new sense of inspiration.   God: Well, Keep the faith and drop the fear. Don\'t believe your doubts and doubt your beliefs. Life is a mystery to solve not a problem to resolve. Trust me. Life is wonderful if you know how to live. Best wishes for a good day!  '),(221,3711805,'13,July,2004','2004-07-13 00:00:00','   SO what was I doing all this while ? I frankly dunno . Blame my blognactivity partly on nothing interesting happening in my life and partly on me losing my creative streak ....(huh ...what ??)  Something new ...Gave my final summer project presentation in the college at last. It was actually a good experience and it was good to hear some intelligent questions at last ;) I mean when you have given a 2 hour presentation at your company out of which you spend half an hour on explaining why the normal curve is bell shaped ....you do tend to appreciate questions which r even remotely relevant to your project !  Back to my hostel room , which effectively means cacophonic hell for my neighbours. I have \"graduated\" in my metal music tastes (thou not yet to satanic levels like \"cradle of filth\") ....Am bordering somewhere in the doom-speed metal line ....    many people ask why the hell do they call all that noise \"music\" .... Petrucci of DT would retort \" I play my guitar fast bcoz ....I can !!\"   PS ...My respect for Schmie has immensely increased after the British GP .....  '),(222,3711805,'18,August,2004','2004-08-18 00:00:00','   I was reading  urlLink Shubham\'s blog  on nostalgia and I could not agree more with him. As I look back on my \'golden years\' spent in school and college, one thing stands out above all else - the list of my crushes.Hmmm... I hear you saying \"OK Ganesh, just because we read your blog doesn\'t mean we don\'t have anything better to do\". Well, if that is the case, for your own good, I would advise you to leave. But, if you have nothing else to do and are depressed, this might make you feel better. You would know you are not the only love-struck guy around (pun intended).  So, are you still with me?   I was an early starter who acted on instincts. My first crush was in the fifth standard. I don\'t remember much of her except that she had hazel eyes and her name was Ashwini. Of course, she left school when we joined the sixth standard. This would, as you would notice, be a standard ending to most of my crushes. So Ashwini, if you are reading this and you are already not married, it is never too late. Err... I dont think I remember your last name.  The next lasted for about two years in high school. Even at my charming best, I was not the young and enthusiastic teenager who would strike up a lovely conversation with a girl. Stuttering, stammering and staring were some of my more preferred maneuvers.I quickly became the undisputed king of the friends\' zone, something that \"terrifies, petrifies and mortifies\" all guys. Mine was the shoulder they needed when their boyfriends ignored them, when they didn\'t score well in exams or when they had a tiff with their best friends. This crush wanted me to \"remain friends\" even after we left school. I did what the best and the most unrelenting of guys would do in trying situations like this - remain friends.  Junior college was a drag - one of those situations where you know you are in college but how well you do in life depends on those crucial two years. So I made a great attempt to be a studios and a sincere student. This had the only effect it could have. I had a 12-feet magnetic force field effect on girls around me - the kind of magnetism in which they don\'t enter the field.(Now that I look back at it, I realize, not studying in those two crucial years would not have mattered a great deal after all)  Enter engineering college and times changed drastically. Fate put me into one of those unfortunate cliques. Yes, I had chosen mechanical engineering and fate had had the last snickering laugh. Yet, I managed my longest crush during this time. I was one of those determined individuals, who was not ready to let fate screw up his life. I decided to do it with my own hands and I did a damn neat job of it!   The Road ahead  I have given up on crushes now. But, if you are female and think that I am cute (I purposefully don\'t have a picture here, so don\'t scroll up to see my profile) and have not dumped me in the recent past, I might reconsider my decision.  '),(223,3711805,'15,August,2004','2004-08-15 00:00:00','   One of my recent mails to my friends working in IT companies in Pune:  are you guys allright ? shan replied yesterday ... ani and sau ....r u  guys still online ? just reply with a blank email so tht i know u r still alive   Ganesh   One of the guys replied:     --- Ganesh Iyer   wrote:    > are you guys working allright ? > shan replied yesterday ... > ani and sau ....r u  guys still online ? > just reply with a blank email so tht i know u r still alive   Yeah, He actually sent me a blank mail. I had no option but to reply in an equally fitting manner. I promptly sent him a mail which contained a single word.   Mazurdya    There were days when I used to meet one of my \"IT\" friends online and then he/she would chat like there was no tomorrow. They would ask me how I was, they would ask me how my parents were, they would ask me how our mutual friends were, they would ask me whether I thought the Sensex was going up .Heck ! They would even ask me how Tiger was. (The funny thing is I don\'t even have a dog called Tiger)  Now, what i get when I meet someone online is:   Me:   u there?   Him :  yes  Me:   What are you doing?   Him:   Duhh...like work dude...   (I actually used to get answers that were different from this and ranged from playing minesweeper to playing NFS 2 on the network)  Me:   So u r busy ?   Him:  (no reply)  Me:   well, Cya later   Him:  (no reply)  The message from the other side of the line is pretty clear.  \"I don\'t have no time for jobless people like you. I actually work.\"   I used to think I could get into one of those companies where I can live the good life. But , it now appears that I would actually have to *work* for a living too. Gosh ! how much would that suck after being a student all my life !  '),(224,3711805,'13,August,2004','2004-08-13 00:00:00','   Why ?  1. I got up late for the first class and am too early for the second.  2. I had my breakfast and I am blogging the first thing in the morning. I am just hoping that the Prof does not decide to conduct a surprise quiz today.  3. I am doing all of this with my email client  '),(225,3711805,'12,August,2004','2004-08-12 00:00:00','   â€œHow are you doing, Maâ€™am?â€    â€œI am fine, thank you very muchâ€   The question was addressed to the nice lady sitting opposite to me at Pizza hut. It had a touch of reverence and genuine concern to it. The guy waiting on us then turned to me and asked â€œAnd what will you have, sir?â€. He had a look that was cold and arrogant. He spoke to me as if I was a fugitive from Somalia.  His demeanor would have caused even the most conservative English butler to clap his hands and legs furiously and dance with heartfelt approval (Raising their left eyebrow a sixth of an inch is the farthest they go, I am told).   â€œAnd would you like some Garlic bread to start off?â€   The question wasnâ€™t even addressed to me. However, I butted in with a meek â€œyesâ€. But the nice lady was not in a mood to have bread and my petition was overruled without a second thought. It just did not matter.   â€œAnd Pepsi, Maâ€™am?â€  â€œNoâ€   So, I had pizza without Pepsi,Garlic bread and mustard sauce. And I paid 300 bucks for all of the above. When we left, the nice lady sounded the bell kept at the door which signalled that the customer was happy with the service.   â€œThank you Maâ€™am!!â€  shouted the guy.  I felt like sounding the damn gong until it never made a sound again.  P.S. I had a very nice time that evening  P.P.S. Really  '),(226,3711805,'10,August,2004','2004-08-10 00:00:00','   1.Aim to provide  quality content . A blog on what you thought of the non-superhuman treatment of spiderman 2 where you get into the psych of Peter Parker is cool. Writing about \"mujhse shaadi karogi\" is well ..... Err.... Suicidal.  2. Blog regularly . I cannot emphasize enough on this. A live testimonial would be this blog itself. I generally use a 48 hour alarm clock and I start typing whatever is on my mind when it goes off. The timing can be varied according to personal preferences and the current level of joblessness.  3.Define the  type of readers  and the content you want for them. It is essential to understand here that not many zombies (if any) read your blogs, so as a general rule, writing about the nuances of death is not advisable.  4.This is a tricky one. Try using  keywords  that attract major search engines like google and AltaVista. Subtlety is the key here. Eg.  Sachin Tendulkar  of  India  plays  cricket . He likes  music  and  free Mp3 . His favorite  teen celebrities  are  Sharapova  and  Britney Spears .  Sex : male. He hates  Pamela Anderson .  5. Get involved in other blogs  in a big way. Always leave thoughtful comments like \"visit my blog at http://iyerganesh.blogspot.com\"  6.Mention your blog wherever you can - email signatures, comments on other blogs, websites, friends\' diaries, caps, shirts, trousers, bikes, cars, and the big hoardings on the buses etc. The key here is to be imaginative.   Disclaimer:  The above are time tested methods which would bring the best and the most relevant readers to your blogs. The very fact that you are here reflects the efficacy of these methods. Why did I call this a disclaimer? It kinda looks cool. That is the seventh tip.  Finish in a manner that leaves the readers numb with excitement.  '),(227,3711805,'08,August,2004','2004-08-08 00:00:00','   I finally made it to Hyderabad. Unfortunately, I don\'t have pics to share of the trip for the  nisarg premis  amongst you. Not that we did not do anything worth capturing on the lens. Listed below are some of the finer moments of the trip.   - Saw  \"with me marriage will you do?\" . It must have been the stars or maybe the biryani at Hyderabad house, but the fact remains that I made a voluntary decision to go watch \"mujhse shaadi karogi\". Yeah, the Gestapo would have been better. By the end of the movie, I had a splitting headache and IMHO  Sallu bhai  in the movie was in a much better situation than I was. Atleast he did not have to see himself with a fake balloon on top of his head hamming away to Bollywood glory and Akkie doing the cool-guy stuff. The saving grace was the special appearance of some Indian cricketers in the end and the discovery that the bhelpuri of a movie was after all not original. Just a poor copy of Anger management with the prerequisite oomphs and sentiness which is so vital for Hindi films. Did I mention that David Dhawan was the Misdirector?  - Went to Begumpet and Hitech areas where there are like a thousand software firms around.I Was especially impressed by the colossal structures at Hitech city. Me and my friend were disappointed when we noted the decreasing aesthetic \"quality\" of receptionists at software firms. After a lot of deliberation and analysis on this , we blamed the BPOs and the callcenters for attracting the real \"talent\" out there with great paychecks rendering the S/W firms desolate.  - Saw Spidey 2. Enough has been said and done about this already. But personally, I liked the \"human\" side of spiderman. The script lacked conviction but the Special FX were great as usual. It had me wondering - how many times would MJ allow Peter Parker to go off with a \"Go get \'em tiger!\" while they are in the middle of a kiss?? Once or twice during the movie is okay I mean ... :D    Public notice:  There is a certain Mr. Hrishikesh Diwan who was kind enough to dedicate his entire blog to me. He was also kind enough to send me the link. What was not so kind of me is that I managed to lose it.(Dedications dont happen everyday. Was a bit over-excited i guess). So Hrish, if you are reading this, just add a comment to this blog and I will add your link to the sidebar. I get like a thousand hits everyday (I reset the counter daily to prevent it from blowing off) and more the blogging junta the merrier it gets !  '),(228,4216216,'30,July,2004','2004-07-30 00:00:00','   empty beer cans in your walk-in closet  waking up naked in your walk-in closet with bbq tongs and an onion still in your hands  a small petrified black object on the grill which at one time was some form of meat and no memory of starting the grill  skid marks in the living room  a video camera, tv and vcr set up in the bathroom  waking up Tuesday with 23 messeges from work which were left on Monday  acute pancreatitis  waking up on the hood of your car in the Nevada desert and wondering why you parked so far from the surf  finding your little dog immobilized wearing a full body duct tape \"suit of armor\"  answering the door of your apartment to a security guard saying \"I just wanted to check on you...as you may or may not know, I helped you up here last night after getting a call about a possible dead body in the bushes.\" '),(229,4216216,'29,July,2004','2004-07-29 00:00:00','   Did anyone see that photo of John Kerry on the flight deck of the space shuttle in that little bunny costume? Where were the ears, the pink nose and whiskers? To me he looked like a cross between an Oompa Loompa and Woody Allen in the movie \"Everything You Always Wanted to Know about Sex but . . .\" where he was playing a sperm with the exact costume . . .  only in white (of course).         Even funnier is the fact that after it hit the press, his camp got all panicked and embarrassed. They are calling it a \"leak\" saying that he just happened to be taking a \"legitimate\" tour  of the flight deck (who wouldn\'t have time for  that  during a campaign?) and the republicans got ahold of the photo and distributed it with the intention of making him look like a boob. Maybe that\'s true, but it doesn\'t take anyone else to make him look like the boob he is. He does just fine on his own. Who does Kerry\'s camp think they\'re kidding? They arranged this photo-op and it backfired on them. I know this is trivial but so is he and I\'m having fun with it. '),(230,4216216,'27,July,2004','2004-07-27 00:00:00','   &nbsp;&nbsp;&nbsp;&nbsp; I\'ve always heard how bad things do, or at least seem to, come in threes. I must say that I\'ve noticed that too. I don\'t give it any credence but it sure seems to turn out that way. So what the hell is going on here now?   &nbsp;&nbsp;&nbsp;&nbsp; My boss is pissed at me, my parents are pissed at me, my 17 year old is on probation and under house-arrest and trying to get out of control again (and thus pissed at me), all three of us have been sick off and on for a week and a half, I have IRS trouble, I got two tickets in one stop last weekend, because of that I\'m gonna have major trouble paying them plus the light bill and my ISP bill, my friends dad died, another friend is having child/marital troubles that may not get worked out, there is a case of lice going around my youngest daughters group of friends, my next door neighbor lost his job and now I feel like I need to help them out with groceries, my truck just got out of the shop after an $860.00 brake job which took two days, I needed those two days to get caught up at work, the ex wife is starting rehab on Monday meaning that I won\'t be given back the child support checks that the state garnishes from my wages (the girls live with me, btw). There may be more but I\'m tired of thinking of it all.   &nbsp;&nbsp;&nbsp;&nbsp; I would apologize for being so negative but I don\'t see the need as I\'m the only one that reads this. Cheers!   &nbsp;&nbsp;&nbsp;&nbsp;    &nbsp;&nbsp;&nbsp;&nbsp;    &nbsp;&nbsp;&nbsp;&nbsp;    &nbsp;&nbsp;&nbsp;&nbsp; '),(231,4216216,'22,July,2004','2004-07-22 00:00:00','   &nbsp;&nbsp;&nbsp;&nbsp; I am so sick and tired of teenaged girls that I could just spit Napalm. There is no other creature on this planet that is as self-centered, selfish, bitchy or sneaky. They think of no one and nothing else but thier own stupid, shallow-minded selves and thier pathetically shallow, narrow-minded little lives.    &nbsp;&nbsp;&nbsp;&nbsp; I just saw \"The Hives\" on Letterman. I had never heard of them until a few weeks ago when I saw a promo for the show. I understand that the high, tinny sound, the costumes and the haircuts are part of the shtick. And I defy anyone to tell me that the singer is not in love with Mick Jaggar . . . but I thought they were a pretty kick ass band.    &nbsp;&nbsp;&nbsp;&nbsp; I got attacked by blackbirds the other day.&nbsp;In Texas we have this ugly,  stupid  bird here called a Grackle. It\'s commonly called a blackbird. They look like a genetic experiment between a crow and a Myna bird gone awry. The other day I was in the Med Center near downtown walking&nbsp; down a block that had Magnolia trees all along the curb and I heard them squawking. I had never heard them make that sound before but I knew that it meant \"bad\". Then I noticed that three of them were following me, hopping from tree to tree . . . still making that awful sound that I now&nbsp;knew was directed at me. When I got out of the tree line (mind you this is in the med center) they started dive-bombing me. I dropped everything and just started swinging aimlessly. They got a few pecks in on my head and all I did was make a few blind swipes at them. When it was over I was laughing my ass off and noticed that everyone at the intersection was too (at me).&nbsp;Hopefully with me.    &nbsp;&nbsp;&nbsp;&nbsp;&nbsp; '),(232,4216216,'21,July,2004','2004-07-21 00:00:00','   &nbsp;&nbsp;&nbsp;&nbsp; I feel like I\'m at the end of my rope. I\'m exhausted with no light at the end of the tunnel, I have all the usual financial woes and&nbsp;I\'m convinced the girls are plotting my death. I\'m on to \'em. They are using those  daughter  powers to&nbsp;shut down most, if not  all  of my vital organs.&nbsp;Meaning,&nbsp;&nbsp;they push me over the edge, I die a \"natural\"&nbsp; death and they go \"scott-free\" to live whatever life it is that I\'m not letting them live.  &nbsp;&nbsp;&nbsp;&nbsp; Send help. '),(233,4216216,'17,July,2004','2004-07-17 00:00:00','   &nbsp;&nbsp;&nbsp;&nbsp; I have two questions to ask of anyone who may see this.&nbsp;&nbsp;&nbsp;(1)&nbsp; In your opinion, &nbsp;did it or did it not appear that Mike Piazza sold out Roger Clemens in that all-star game here last weekend?&nbsp;&nbsp; (2)&nbsp;&nbsp;Is there a band&nbsp; called \"flipmode\"? (what looked like a tour bus today with&nbsp; \'flipmode\' logos &nbsp;all over it nearly drove over me before I could get out of the left lane of I-10 to let it by). '),(234,4216216,'13,July,2004','2004-07-13 00:00:00','   &nbsp;&nbsp;&nbsp;&nbsp; I\'m&nbsp; so damned glad to be back home. In my opinion, San Antonio is not a very fun place to be. My job has had me there &nbsp;once or twice a month (several days each) for nearly 10 years. I\'m tired of driving and I\'m tired of hotel/motel rooms. &nbsp;I shouldn\'t bitch though, it used to be Palestine,&nbsp;Round Rock, Austin, Wimberley, Universal city ,&nbsp;San Antonio , then Corpus Christi. The out of town accounts have been whittled&nbsp;down to just San Antonio, and just today I &nbsp;finished training the guy who will be taking care of that for me from now on. &nbsp;So, hopefully, &nbsp;no regular&nbsp;monthly roadtrips anymore. &nbsp;(In all fairness, I should mention what San Antonio  does  offer like no other place: the riverwalk and the&nbsp;carefree mood it puts you in, the undisputably best Tex-Mex in the universe and a dense population of some seriously stunning Mexican cuties). '),(235,4216216,'13,July,2004','2004-07-13 00:00:00','   The horn on an automobile is no longer used for its original purpose. It is now the audible middle finger. '),(236,4216216,'18,August,2004','2004-08-18 00:00:00','   It\'s Wednesday! That\'s too many days away from Friday! :P The last few weeks have been so hectic, I so need a vacation...It\'s become apparent to me, that my transfer out-of-town is not going to happen until all of the audits and major pain in the ass work here is done! I suppose it\'s only fair not to stick a new manager with all of it, but I really didn\'t want my son to start school here and then have to leave in the middle of it. *sigh* I\'ve spent two years at this complex and I\'m so burned out on this place...the people here are driving me over the edge, which seems like a very short drive at this point! But anyhow, I wrote up a nice bitch letter for my tenants and attached the rules and regulations to it...(in hopes that they will actually read it), and had to attach a few information forms as well....so now I get to pass them out and wait for the morons, who can\'t read simple directions, to ask me what I need....uhg! My old boss told me I can\'t write big words on my memos, she told me to lower myself to their level and type my memos, to my tenants, in language that they can understand, hahahahaha....I don\'t know that I can think in redneck hillbilly terms......but I\'ll give it a shot! So that is my big goal for the day....that and my son gets his eye glasses today...yippee....no more whining about not being able to see! I\'ve been working 3 apartment complexes for the last few weeks. It\'s actually kind of been nice to be so busy, but I\'m tired.           On another note........I was forced to have lunch with a gay guy yesterday....was rather amusing...he told me I make him want to go straight...I suppose that was a compliment, although I about died of embarrassment when he kept telling everyone in the restaurant. Was certainly a different experience....one that I won\'t be doing again as long as I can help it! :)          Well, I\'m off to work I\'m evicting 2 people, so I have to find 2 replacement tenants.....that\'s always fun! One of them will get to live next to me...so I have to find someone who can tolerate me and doesn\'t get on my nerves...hahaha! Like that will ever happen!          Linda '),(237,4216216,'17,August,2004','2004-08-17 00:00:00','   Thanks QT for keeping things rolling here. I\'ve been really busy and have not been able to get anything done here. I only have the energy for one more sentence. If only I could explain. '),(238,4216216,'15,August,2004','2004-08-15 00:00:00','   Houston football  KICKS ASS  !!! Back later. '),(239,4216216,'14,August,2004','2004-08-14 00:00:00','   Ahhh, it\'s Saturday at last, no work, no school, and the weather is absolutely beautiful!  But, it hasn\'t started off well.....I\'m hoping it gets better......I was awakened by a drunk calling me at 4:00 am....a friend, whom apparently has more in mind than just the friendship he and I have agreed to have.....why me?  So....I talk to the drunk, who is asking me all sorts of personal questions,  for a good half hour or more...then manage to get back to sleep...wake up at 7:30 or so (Ever notice how you wake up early and full of energy on your days off and can hardly muster up the energy to get out of bed on work days?), my son makes me my coffee as usual, and all is going well, then the ex husband calls and wham, the day is ruined...seems he and my son were having a disagreement, my son had told him I wasn\'t home....so I hear my kid defending me with his dad and I\'m like, wtf is that all about, I grab his phone and say why are you threatening me to my kid?  He\'s telling my kid if I take him back to court again, he\'s going to burn my ass......HAHAHAHAHAHA......you would have to know me to understand this, but I do absolutely nothing wrong at all!  I don\'t even go out, I stay home and raise my child, like a good mother should....and his father on the other hand is an alcoholic and pothead....and does both in front of my child...I should probably take him back to court just for the satisfaction of proving him wrong.....but anyhow....we argue for a bit and it\'s done.  Now, for the rest of the day.....I have to evict 2 single mothers, each with a child today....for not paying their rent...my job really sux sometimes, but ho well, you gotta pay your rent if you want to have an apartment.  Sure hope this day gets better, and I hope others are having a better one! ;) (I know this day is going to get better and that later I\'m going to have something wonderful to post!)  Enough whining! Linda '),(240,4216216,'12,August,2004','2004-08-12 00:00:00','   I\'ve never been the type to \"surf\" the internet. I have always had something specific to do online. However, when all is done, I sometimes find myself with nothing left to do (exciting life, eh?). So, I thought I would offer some links for those in the same situation. Some of these links I came across by accident and some I actually hunted down. The weird ones were found by accident, the practical ones were actually searched for. I will list the practical links first, then the weird ones.     urlLink http://www.cooltext.com/  For those interested in HTML and web site design, this is a very handy site for getting pre-fab graphics. They have logos, fonts and buttons. Maybe more, I just don\'t remember. The choices are somewhat limited, but it\'s extremely easy to use and very fast.    urlLink http://www.findsounds.com/  This is a .wav locater. I understand that \"dailywav\" has thousands of .wavs, but this site allows you to search for specific sounds unrelated to movies, TV or cartoons although that is available too.    urlLink http://www.howstuffworks.com/  This has to be the best website   ever created  ! It\'s the finest site I have ever seen. Anything your mind can come up with, this site will tell you all about how it works, and even in greater detail than you may need (or care to know ).  A definite  must-see  site.    urlLink http://www.netsurfing.com/surfcam/index  This is a very helpful site for surfers and fisherman. It provides weather, tide and surf information. It also has a very cool surfing video game you can play. But the best feature on this site, in my opinion, is the surfcam. It shows the surf (surfside beach, my fav) from a camera mounted on a house right on the beach. The image normally reloads every ten minutes. I like it because I can take a look at the surf in the morning right from my desk here and then decide whether or not to go out there.      AND NOW . . . SOME ODDITIES OF THE WEB'),(241,4216216,'12,August,2004','2004-08-12 00:00:00','   I guess it pays to browse other blogs at 4 or 5 in the morning. I happened to wake up around 4:00 AM and found this:   Some young kook in Florida (kook: meaning, look at the guys profile pic) has a couple of photos of a supposed \"ufo\" over downtown Houston during the all-star game. It\'s a fabrication. A lame attempt at a hoax. If there really was something in the sky over downtown Houston that night it was a blimp.   For fun\'s sake, check out the link.     urlLink http://blogg.somewhiteguy.com/2004/07/ufo-over-houston.html '),(242,4216216,'12,August,2004','2004-08-12 00:00:00','   Summer is over at the greg household. The kids started school today. They are taking it  way  much better than  I  am. My 17 year old is on probation and under house arrest, so she is thrilled to just get out of the house. My 10 year old is one of those gifted/talented students, so  she\'s  back in business now . . . Me? I\'m pissed that I have to get up and get them to school at an hour that I think even animals are not even yet stirring.    Speaking of G/T students . . . mine is typical. She approached me with what seemed like ream of paper saying that \"these needs to be filled out by you tonight\" at 10 fucking thirty-something at night! It\'s either a lack of common sense or that she\'s just too involved in her projects to remember. I think it\'s the latter . . . I think. Who knows.   Thank God tomorrow is Friday. Don\'t forget the Houston/Dallas game on Saturday. (Death to Dallas). '),(243,4216216,'11,August,2004','2004-08-11 00:00:00','   Wait a minute . . . \"Red\" is a homeland security level. Scratch that. The point is: QT is now an official contributor to this blog and has full control of the dashboard to post as she feels inclined to do so. Instead of code red, think of it as you would a hurricane warning. I got confirmation this afternoon and I couldn\'t be happier. She will be a definite improvement to this blog.   Thanks QT ! '),(244,4216216,'11,August,2004','2004-08-11 00:00:00','   If I am lucky,  there may be a new member added to this blog.  I am not very optimistic about her having the time, energy or desire to write here, but at least the invitation was sent. If she  does  decide to write in whenever she has the time, I  promise  that it will be worth reading.         I have known her as QT for several years.  She\'s smarter, faster, meaner, wittier and tougher than anyone I know, including myself. She is  not  someone you would want to engage in a debate. Her humor and attitude are right-on,  and  it turns out she always seem to be right.         All I can do is wait and see if she writes in. I hope she does. '),(245,4216216,'09,August,2004','2004-08-09 00:00:00','   That\'s right. Sunday sucks. Especially Sunday  night . You know how it is. Even though you are off, you can\'t really indulge in anything because you have to be at work in the morning,  on  the   day of total scrutinization (Monday). If your late, slow, not looking  your best, still wearing a coctail dress . . . it will be decided by all that you were on a drug and alcohol binge all weekend. People . . . pfft. '),(246,4216216,'08,August,2004','2004-08-08 00:00:00','   I have nothing to say or talk about. Actually, I do, but it\'s all so negative even I don\'t want to get into it. I guess all I\'m doing now is leaving a current post so this place doesn\'t look dead. '),(247,3991209,'22,July,2004','2004-07-22 00:00:00','    nowadays damn&nbsp;sianz lor...everyday gt nth to do except those stupid maths h/w...do until siao ar...summore is graph lor...idiotic lor..still muz copy qns...go &amp; die la......lyk wat sehx....today damn suay oso lor...kanna caught by LL for piercing too much ear holes...mani mehz...6 onli rite..summore i onli put ear stick...transparent white lor...cannot mehz...piercing more than 1 pair doesn\'t mean i \"pai kia\" rite...juz fashion onli rite....lyk tt oso cannot...haiXxX...say galz cannot wear short pinafore...she herself also lyk to wear mini-skirt rite....lyk wat sehx...reali gt nth to say lor....haix.... '),(248,3991209,'16,August,2004','2004-08-16 00:00:00','    TMD! Ur can together anot...lyk tt i so difficult ....fuck sia...always lyk tt de....Tmd! tt bloody bitch....slap mi for wat sia...my fault mehz...chee bye....u free u can go fetch him ma,....y myuz it be mi ..knnbccb! AnD lastly...tt MS PANG HENG HOON! siao ar.....give so mani worksheets...tmd....so difficult summore..u tink we robot....coomon test oso no nid give so mani wksht ma...tmd....so mani wksht how to do finish! wah lao! cant take it liao...sumtymes i wish i\'m nt in express but in normal tech sehX.....HaIxXx...... '),(249,3991209,'10,August,2004','2004-08-10 00:00:00','    HeY...2day... i hav a most happiest day.....hahaz go out wif mi sis de frendz....siao sia they all...but i envy them.....their class veri united... the guys are open minded...they go out wif galz....can make fun...say jokes....haiXx....the whole class damn funi ....always....\"hey! ur mother porn star\"..\"see larx...no wonder ur whole family porn star\" \"ur mother hor chia long\"\"ur mother kana fucked by bangladesh porn star\" which mi class de ppl will nv tot of it lor....mi class....haix...not united....no fun....everytin SUX! '),(250,3991209,'08,August,2004','2004-08-08 00:00:00','    so wat...it\'s onli for tiz mth nia...everymth lyk tt den gd lor...den no nid to wry so much abt it lor....haixX...currently wishin for a new hp...best if i can get N7610 LOLx...fat hope lor.....when am i gettin a new hp!haix....*sobx* '),(251,3991209,'07,August,2004','2004-08-07 00:00:00','    haiyo...2day the C.I.P project...damn de sian....den so paiseh....the volunteer mark.....i say he write lyk gal gal de words...den he heard it...den call mi to write....lolx...but i tink he dun realli heard tt cuz he say\"juz lyk sumione said...my handwriting sux...lyk boiboi de word\"lolx...i din not say sux and did nt say boiboi...ish galgal...hahaxx...but hor...its a fact lor...the words write so small hor...den write slow slow...den so ugly...(i noe my handwriting oso ugly la...but hor...mine sumtime nice nice de)sianz lor...c.i.p project still got 4 lessons to go...waste my time onli.....haiXx.... '),(252,3991209,'06,August,2004','2004-08-06 00:00:00','    sianz....bo liao nth to do....haiXx....when can i get a NEW hp....so sian man...wif tiz useless budget lousy hp....haix....when my mum strikin 4D!!!! ohh...n found out today tt some(some onli) guys tend to be more \"shaui\"when they SiMlE...but hor...muz be those type of natural de...lyk when hear jokes then smile or laugh tt type....hahax.... '),(253,3991209,'05,August,2004','2004-08-05 00:00:00','    National day comin...the following mon..hahz...so gd...so mani dae no nid study...but gt h/w lehz...den librarians...haiXx still muz go bak on sat to do wat \"CIP Project\"wat the hack ish tiz!SiAnZaTiOnXx...2pid Louis...take mi de foto den cannot send...SiAnZ....aniwae..today juz a borin dae lor...ohh...juz find out 2dae...the \"les\"lyk to bring alot of tissues to sch....isn\'t she MAD???'),(254,3991209,'04,August,2004','2004-08-04 00:00:00','    NOW! EvEr SeEn AnYone SOOOOO STUPID....go change a samsungX430 for a nokia 8250!WAT SIA!   DAMN de STUPID....CoLoUr PhOnE ChAnGe To no colour de.....STUPID ARX!GuEss  WhO Tt STUPID IDIOT CHEE BYE KIA ISH!!!!MY MUM(even feel shame to say she is my mum!)   WAh111realli....reali reali reali STUPID !!!!HOPELESS BRAINLESS HAIX!!!KAN BU XIA QU! '),(255,3991209,'03,August,2004','2004-08-03 00:00:00','    enuff liao lor...always say i lyk him...can see u still lyk him...wat for push to mi lehz...summore....i dnnno him lor....den keep sayin i lyk him...sayin sumbody handsome or looks nice dosen\'t mean u lyk tt person rite...wah...den lyk tt sia....HaiXxx...reali dunno wat to say...and those ppl hu believe...den i tink ..oni gt 1 word to describe u lorX....STUPID! '),(256,1077954,'17,January,2003','2003-01-17 00:00:00','   Those who give up liberty for the sake of security deserve neither liberty nor security.--- Ben Franklin  '),(257,1077954,'17,January,2003','2003-01-17 00:00:00','   Welcome to my Blog!!!  '),(258,1077954,'19,January,2003','2003-01-19 00:00:00','   Football Fan\'s!! I went out to dinner with my family tonight. My cousins were in town so me, my mom & dad, my grandmother, & my aunt & uncle went out for Chinese at a great buffet place, & came home went to NFL.com to find out the Tampa Bay BUCS are the NFC Champions. Congrats to the BUCS!!!! I then turned on the Titians Vs. Raiders game to see the Titians were holding there own only down buy 3 point in the 3rd quarter, but the Raiders were the stronger team and prevailed, and the Oakland Raiders are the AFC Champions. Congrats to the Raiders!!!!  Super bowl XXXVII will be a good one!!!!!  '),(259,1077954,'22,January,2003','2003-01-22 00:00:00','   I took the  urlLink Dirty Mind test  !!!  my score was 80    This probably isn\'t much of a shocker, but you\'ve got a one-track mind! Either your hormones have taken over your brain or you\'re feeling deprived in some way, because s-e-x is the first thing that pops into your mind on almost any occasion. This may or may not be a problem, depending on the point to which your fantasizing takes over your mental processes - thinking about getting it on is a vital part of any healthy sex drive, but it can bring trouble when it\'s too extreme. As long as you are still able to shut off that part of your brain when need be (while working, for example), there\'s no cause for concern.  '),(260,1077954,'25,January,2003','2003-01-25 00:00:00','   How ya\'ll doing tonight i was at our local county fair tonight Volunteering selling refreshment in the expo building and got to see john anderson, what a great show. just thought i would share my cool experiance.  '),(261,1077954,'25,January,2003','2003-01-25 00:00:00','   Funny thing happened today. i live in Florida, South Florida actually, and in the 19 years i have lived in this state the news has never reported snow. thats right SNOW, yesterday(1/24/2003) in Sebastian, Fla there were snow flurries. i wasn\'t\' at all surprised since the tempatures were in the low 30\'s to low 50\'s.    on a lighter note i got Linux on my laptop FINALLY!!!  '),(262,1077954,'27,January,2003','2003-01-27 00:00:00','   Football Fans!!!!  Super Bowl XXXVII was a great game.   Tampa Bay Bucs - 48 Oakland Raiders - 21  Dexter Jackson as Super Bowl MVP was a good choice.  '),(263,1077954,'01,February,2003','2003-02-01 00:00:00','   Febuary 1, 2003   Crew of the space shuttle Columbia....REST IN PEACE heros!!!!  '),(264,1077954,'03,February,2003','2003-02-03 00:00:00','   Football Fans!!!!  -  Well the 2002 NFL season is officially over, and was caped off with a blowout victory buy the AFC in the Pro Bowl.  Final Score: AFC - 45 NFC - 20  Miami Dolphin Ricky Williams was the MVP of the game which made me happy since i am a devoted Fins fan.  Jason Taylor, Zach Thomas, and Sam Madison also had a great game.   Augest isn\'t close enought.  the highlight of the game other then Ricky being MVP was Colts QB Payton Manning Calling his kicker an Idiot for comments he made about Payton. funny stuff.  '),(265,1077954,'04,February,2003','2003-02-04 00:00:00','   RANT!!!!   -  All this talk about not going back into space is ridiculous. It is a tragic thing that happened on Saturday, but the Astronauts that risked there lives to explore space wouldn’t want us to abandon the program because they lost there lives.  Exploring space is like exploring the depths of the ocean. Many have died exploring the ocean, but that doesn’t mean we still don’t have submarines.   If you ask me the lose of 14 astronauts, and 2 shuttles in a 22 year period isn’t that bad of a number, yes it’s sad 7 astronauts lost there life in 1986 when challenger exploded on lift off, and it’s just as sad that 7 astronauts lost there life in Columbia on reentry, but those are the risks and the brave men and women that have that opportunity knew that before lift off and during reentry.  my thought are with the families that lost loved ones on Febuary 1, 2003  '),(266,1077954,'04,February,2003','2003-02-04 00:00:00','   Thanks goes to  urlLink Chris Pirillo  for the  urlLink STS-107 Tribute  link.  '),(267,1077954,'09,February,2003','2003-02-09 00:00:00','   Went to starbucks tonight and got a Grande White Chocolate Caffe Mocha  '),(268,1077954,'10,February,2003','2003-02-10 00:00:00','   Right now the Enzo  is my dream car!!!  '),(269,1077954,'10,February,2003','2003-02-10 00:00:00','   I am Scooby-Doo!!!!!           urlLink What     cartoon dog are you?       Brought to you by the good folks at  urlLink sacwriters.com .  '),(270,1077954,'23,March,2003','2003-03-23 00:00:00','   urlLink Nascar  -  I can\'t help but notice that  urlLink Dale Jr. (#8)  is starting his season like  urlLink Tony Stewart  did last year, bad showing at Daytona and had good season, and  urlLink Tony Stewart  won the WC Champ.     urlLink  GO DALE!!!!  '),(271,1077954,'23,March,2003','2003-03-23 00:00:00','   As I watch the Liberation of Iraq, I can only hope the men & women of the US Armed Services stay safe and come home safely.  '),(272,1077954,'27,March,2003','2003-03-27 00:00:00','   Computers - Tonight I decided to install Redhat 8 on my Pentium 4, well it\'s   2:42 A.M & running great.   I just hope i can get the programs i want to use to install.  '),(273,1077954,'28,March,2003','2003-03-28 00:00:00','   Rant!!!   -   Ever have one of those set backs in your life, where no matter what happened the outlook looked grim, what i am talking about is being unemployed. I mean no matter how many resumes i send out for a job that i know i can do, the response is, \" sorry we\'re looking for someone w/ a little more experience \", or \" we decide to go with someone else \". I have 3+ years experience building & repairing computers, and setting up networks, and can tackle most problems and resolve them. I am not looking to become the admin of a server datacentre (at least not tomorrow). I just want the opportunity that i deserve, and to top it off   UNEMPLOYMENT SUCKS!!!!!  '),(274,1077954,'31,March,2003','2003-03-31 00:00:00','   urlLink      Saturday I went a bought the new Linkin Park CD  urlLink Meteora , the cd is awsome, and while i was at  urlLink Winamp.com  i found a pretty cool skin from  urlLink Linkin Park  for winamp3.  '),(275,1077954,'06,April,2003','2003-04-06 00:00:00','   urlLink Nascar  -   urlLink Dale Earnhardt Jr.  Wins the Aaron\'s 499 @ Talladega superspeedway, that win bumped  urlLink Dale Jr.  into second in the Winston Cup point standings.  '),(276,1077954,'06,April,2003','2003-04-06 00:00:00','   Computers -   Tonight I had to reinstall windows on my laptop, the reason I had to reinstall windows was because Partition Magic 8, wouldn\'t allow me to resize my current partition.  It\'s a 30GB HD and I wanted to dual boot windows with mandrake 9.1.   The way I partitioned the drive:  Total size 27.94GB (30GB) 9.75GB - Windows XP Pro 9.75GB - Storage 8.40GB - Mandrake 9.1  I finally decided to just wipe the 1 week old install from when i had to wipe the drive to recover some bad sectors.  I had two drives that had that issue in the last two weeks.  The first one was one of my WD 100GB drives, but i wasn\'t two surprised since the drive had the original format from when i bought it, and i had written & deleted files on it since the drive was bought.  The Second was the 30GB Laptop drive, and I glad to say both are working.  '),(277,1077954,'13,April,2003','2003-04-13 00:00:00','   Computers   -  Last week i was about one BOSD away from throwing my PC across the room & going into debt, and buying a G4.   Thank god for Red Hat 8 & the Winblows XP(eXtra Pain) repair utility which helped me get all my important files(MP3\'s) off my HD just incase i was going to reformat the HD, &  then the kicker after i did all that, Xplay decided to clear out the  itunes.db  file on my ipod, & i was forced to restore the ipod to factory settings and resync the songs.   I wish steve jobs would come to the realization that porting OS X to the Intel & AMD Platform would be a good thing.  '),(278,1077954,'18,April,2003','2003-04-18 00:00:00','   My Uncle is the best he got my cousin & me tickets to the last show on the  urlLink OZZFEST 2003  tour on Aug. 28 in West Palm Beach @ Coral Sky Amphitheatre.  I am so stoked, i get to see Ozzy, my favorite band Disturbed will be there along with KoRn, Marilyn Manson, & Chevelle, plus many other good bands.  I can\'t wait till the summer.  '),(279,1077954,'18,April,2003','2003-04-18 00:00:00','   Computers   -  I finailly went WiFi for my laptop, i decided to go with the    urlLink SMC2404WBR   11/22Mbps router w/ 3 port switch which i am using as a Wireless access point, and for my laptop i got the  urlLink SMC2435W  11/22Mbps PCMCIA WiFi Card.   It\'s great I was impressed with the fact I could watch a divx movie on my laptop from my desktop.  '),(280,1077954,'20,May,2003','2003-05-20 00:00:00','   It was my lucky day -   Today I was curising along up US-1 on my way to a client and I wasn\'t paying attention to the the Speed-o-meter, and a cop stepped out in front of my truck and motioned to me to pull over, so i did, and he then asked me if i knew the speed limt and i responded to him buy saying is it 50 or 45, he said it\'s 45 and i clocked you at 55 he said \"Slow it down\" and he let me on my way with a warning.    I lucked out. if it was the interstate or near my town, i would have gotton a nice fat ticket.  '),(281,1077954,'15,June,2003','2003-06-15 00:00:00','   Today is MY 25th Birthday.........  '),(282,1077954,'15,June,2003','2003-06-15 00:00:00','   Friday night I had one of those dreams that was so realistic when i woke up i wished it was real.  '),(283,1077954,'01,July,2003','2003-07-01 00:00:00','   It\'s Tuesday July 1st  '),(284,1077954,'22,July,2003','2003-07-22 00:00:00','   My mood today.......  '),(285,1077954,'06,August,2003','2003-08-06 00:00:00','   I hate florida.....there is to much damn rain, i was sitting here after a long day at work and all i wanted to do was watch  urlLink The Screensavers  and my satellite signal keeps going out.   atleast i still have my DSL & can get on the internet, and can look for apps for my new pocket pc. I got the  urlLink HP Ipaq 1945  w/ Pocket PC 2003.  '),(286,1077954,'31,August,2003','2003-08-31 00:00:00','   Thursday Augest 28, 2003 was OZZFest, and it was awsome. All the bands i wanted to see Rocked harder then i expected.  I know it was a good show my ears were ringing for 24hours after the show and still have a slight ringing in them.  '),(287,1077954,'21,September,2003','2003-09-21 00:00:00','   I set up a picture blog over @  urlLink TextAmerica.com .   My Picblog is  urlLink http://mp3mooch.textamerica.com/  '),(288,1077954,'22,September,2003','2003-09-22 00:00:00','   The Miami   Dolphins   Beat the Buffalo Bills to go 2-1 on the season. wOOt!!!!  '),(289,1077954,'24,September,2003','2003-09-24 00:00:00','     '),(290,1077954,'25,September,2003','2003-09-25 00:00:00','   Stupid Copy Protection!!!!!  ok i bought a new CD yesterday Nickelback\'s The Long Road, and of course i brought it home popped it in the CD drive and ripped it to mp3 to put on my ipod, then i decided before I went to sleep to burn a copy to keep in the car, and put the original back in the sleeve and went to bed. This morning i got in my truck placed the cd in the head unit and all the tracks were popping and hissing. Now here is the part that pissed me off all the mp3\'s I ripped sound great no popping or hissing. unless it\'s my burner I think the copy protection is a little backwards, I can rip files to my HD but I can\'t make a copy of the disc to preserve the life of the intellectual property I paid my hard earned money for.  The reason I think it is copy protection is I copied the new Fuel CD Natural Selection, and it sounds like the original.  Right now I am trying to copy the disc w/ xcdrost in Redhat 9.  '),(291,1077954,'01,October,2003','2003-10-01 00:00:00','   WiFi card for Linux on laptop....   I finally found a wifi card that works with my laptop, the  urlLink   Netgear MA401   , I had one where i work that we use to get laptops on the internet to do updates and stuff when they don\'t have a NIC card, a USB port(for a usb nic), an O/B ethernet port, or when we don\'t have an open cat5 cable to plug the laptop to. So I tried it, & what da you know it worked. So i went to compusa after work and spent more money....being a geek is so expensive.   I am just happy I can use linux on my lappy w/ my wireless network.  '),(292,1077954,'19,October,2003','2003-10-19 00:00:00','   My new iPod!!!!!  I have pondered over it long enough, and since they released itunes for windows, I decided to get the new docking ipod.   unopened    Just out of the box     In it\'s dock    In the dock next to the 2nd Gen    I will admit i wasn\'t a big fan of the 3g ipod till i went to the apple store to look around(guys windows shopping), and actually held it, and i then pondered over it and walked in to the apple store Saturday and bought one. Now i just need a case.  '),(293,1077954,'01,November,2003','2003-11-01 00:00:00','   Hell has froze over!!!!    The reason i say this is i did something i thought i would never do, and i bought a DELL. Granted it\'s a Laptop, but it\'s still a DELL.  Now i think i got an awsome deal on it, i paid $1000 for a:     Inspiron 5100 2.4Ghz P4 Intel 845G Mobo 256 PC2100 DDR ATI Readon 7500 16MB(on the small side but it\'s enought) 40GB HD CDRW 15\" LCD Windows XP Home(which quickly became Professional)  I have only had it for 3 days now, and it\'s seem to be solid, hasn\'t crashed yet The Main reason i bought it was the 1394, & the CDRW.  One thing i saw which i haven\'t been able to confirm is, i saw it had USB 2.0 which would be awsome since i just bought a new docking ipod and the USB 2.0/Firewire cable for it.  It\'s biggest test will be in 3 weeks when i travel to VA to vist my brother, i will be on airplanes for a while, so i will need something to do, and from the looks of the spec on this lappy it will beable to handle some of my fav games.  I will let everyone know how this new purchase turns out.  '),(294,1077954,'14,November,2003','2003-11-14 00:00:00','   urlLink Which Pirates of the Caribbean character are you?    Fitting if you ask me  '),(295,1077954,'02,June,2004','2004-06-02 00:00:00','   \"Wise men talk because they have something to say; fools, because they have to say something.\"     -- Plato  '),(296,1077954,'02,June,2004','2004-06-02 00:00:00','   Quote:   \"A lie gets halfway around the world before the truth has a chance to get its pants on.\"  -- Winston Churchill  '),(297,1077954,'02,June,2004','2004-06-02 00:00:00','   This is a test, to see if comments are on  '),(298,1077954,'08,June,2004','2004-06-08 00:00:00','   I went to talk to the new owner of the repair shop I used to work at today, and he offered me my job back, the reason I left originally is a long story, and not worth going into, but after being out of work for 3 months I am happy that I am going back to work.  '),(299,1077954,'17,July,2004','2004-07-17 00:00:00','   http://www.bannister.org/software/emu.htm - emulater http://www.emulationstation.com/ - roms  '),(300,3358293,'19,May,2004','2004-05-19 00:00:00','   A week ago I was laying in bed and heard a \"pop-pop\" my knee felt like a rubber band being stretched as far as it would go then being snapped. I tried to buy a ghetto brace from Krogers and deal with the pain so I wouldn\'t have to go to the doctors again. Last Friday the pain got to the place I couldn\'t take it. My knee was making a crunching/grinding noise not the normal cracking I hear from my arthritis. So Saturday I send countless hours at the ER. They do x-rays and say nothing is broke. Damn it took a bunch of x-rays to tell you that? I need to become a doctor! I stated over and over again I didn\'t twist it and I could move it. So the doctor plays with my knee for a while scratches his head and says I think you might have torn ligaments and cartilage. So he dx me with internal derangement of the knee, gave me some pain pills, a brace which makes me have to hold my knee straight at all times and some crutches. He gave me a number to an orthopedic doctor to call and said to call them within the next ten days. So first thing Monday morning I called the number the ER had given me and they say you can\'t make this appointment you family doctor has to call us first and refer you. I made the mistake of thinking this might be easy! So I had to see my family doctor. He plays around with my knee for a while and says he agrees with the ER and wants me to get an MRI then he will call an orthopedic doctor for me. In the mean time I am suppose to keep the God awful brace on and continue to use the crutches. So tomorrow I go for my MRI. In the meantime I have had this stupid brace on my leg which has a steel plate down the back so I can\'t bend my knee so now when I try to bend my leg I can\'t at all. So no matter what the original problem was I will end up having to do physical therapy to get my leg working. I know I won\'t get the MRI results until mid next week so I\'m really frustrated right now. There\'s not much I can do with this brace on. The pain sucks but I try to work through it. My fear is I will have to get surgery and I am not looking forward to the at all. I think \"the wolf\" has me pinned down for now but soon I will be back on my feet and kick his ass!  The Lupus scares me a little more everyday. I am 26 looking at knee surgery because of it. I am in a support group and I hear all the stories of the woman in the group. Most of them have had countless kindney problems and many have had kidney transplants. I don\'t like to spend to much time thinking about my Lupus but somedays I do get stuck in fear of what next. I don\'t fear death for myself but I don\'t want my kids to grow up without me. I don\'t want to be a burden to my family and I often feel that way already. I take pride in doing things on my own and not depending on anyone but every year it seems I become more dependent on those around me. Whether it\'s for someone to open a jar or for someone to carry things down the stairs because for that day my body isn\'t working. I hate the fact that some days I just can\'t do things normal people do. I push myself as far as I can but sometimes my body won\'t do what I tell it to do. I understand it could be worse but these are my pains I know no worse pain then what I am feeling. To not be able to go out a toss a ball with my son or not be able to brush my daughters hair because my hands hurt doesn\'t seem fair to my kids.  '),(301,3358293,'19,May,2004','2004-05-19 00:00:00','   So much time spent on things that don\'t matter, so much time spent on others. We don\'t take the time to care anymore. We don\'t take the time to love anymore. I lost you a long time ago but it took me this long to see it. All these years just a faded memory. Why are we still here? Is it because once in a while we get that old feeling back? For the kids? Maybe because it has become comfortable...  My heart is still in your hands but your mind has drifted. You don\'t look at me the same. You are too busy to notice. Maybe my health has turned you away from me. Maybe the years have made us change so that we can\'t find common ground anymore.   I\'m tired of being that painting on the wall you notice from time to time. I\'m tired of being in the same room with you but still alone. I\'m tired of feeling nothing. I miss the passion. I miss the friendship. I miss the butterflys I use to get when you kissed me. I miss the way you use to look at me.  You pour your heart out to others but have nothing to say to me. You stay up all hours of the night with your buddies but it\'s lights out at nine with me. You are closer to people you have never met then you are to me. You don\'t even know who I am anymore. I bet I could ask questions about your online buddies and you would know all the answers. Do you even know of my dreams? Do you know my fears? Do you know anything about me anymore? Do you even care?   So many times we said forever but now that seems so far away. I\'m not sure I see forever anymore. I can\'t live my life like this. I can\'t beg you to love me. I can\'t beg for you attention. I want you to love me. I want you to want me. I want you to notice me but not because I asked you to.  So where do I go from here? What is best for the kids? What is best for me? What is best for you?  Maybe you would rather not having me around? Maybe I add too much stress to your life? Maybe we just got lazy? Maybe we figured we would always be here for each other so we don\'t have to work at it? Maybe we have taken each other for granted? Maybe we just don\'t care anymore? Maybe we don\'t think we deserve any better? Maybe love isn\'t forever?  '),(302,3358293,'19,May,2004','2004-05-19 00:00:00','   I am mom I am wife I am a homemaker At times I feel very useless At times I feel nothing To many people think if you donâ€™t bring home a pay check you donâ€™t work To many people think if you donâ€™t have a 9 to 5 you are lazy Where is my life going? What will I become? I had many dreams I had many plans Now I dream of nothing Now I donâ€™t make any plans I am only a mom I am only a wife I am only a homemaker  '),(303,3358293,'09,June,2004','2004-06-09 00:00:00','   So my dad calls today to let me know my Grandma\'s sister died. She had bone cancer and I know she\'s better off. So knowing this makes it all better right? No not really. Growing up me my brother and sister were the poor white trash of the family. Everyone knew what an asshole my dad was so either people took pity on us or wanted nothing to do with us. There were few people who gave a shit. Opal (my grandma\'s sister) was one of those people who cared for us not out of pity but just because she had a good heart. She also took great care of my grandma. So she was old and full of cancer so I should buck up and be happy that she\'s gone I guess...Most of my pain is for my Grandma she is about the only one left from her family and the years are getting to her. She feels very alone and I can\'t handle seeing my grandma in pain and I know I have to stay strong for her this weekend while we bury her sister. My grandparents are the world to me and I know their time is coming. I\'ve lost many people but my grandparents were my hope when I was a kid they are the ones who taught me about love. Before Steve and the kids it was only my grandparents. Now my life is full of love but I know some of that love will be taken away soon.  I know I will have to see my dad for the funeral. People always say to leave the past behind you. I try everyday to do this. I get sick just looking at my dad or hearing his voice. I start to feel sick to my stomach and end up with a headache. I would love to find forgiveness in my heart for him but he doesn\'t want it. He won\'t own up for the times he knocked around my mom. He won\'t own up for all the hate he gave us kids. Most of the time people who don\'t understand baggage have never had any of their own. I love people who have no clue what we went through say I should let it go and move on with my life. I have moved on with my life but the drama from my childhood will always be there with me. It reminds me to never let myself turn into the asshole my dad is. I know the pain a parent who doesn\'t love their child can bring so I make sure my kids know everyday I love them. Every time I see or talk to my dad my past and present meets. There\'s no moving on or shoving it aside it is there and I have to deal with it for the moment then when I see him in anther six months I deal with it again.   My leg is still hurting me. I go to the orthopedic doctor next week. It looks like I\'m heading towards surgery. All my ligaments in my knee are torn apart. So I\'ll either get some plastic in my knee or parts off a dead person. Kind of weird to think about having a dead persons body parts. I just want it over all ready I am sick of the waiting game.  So today is a shitty day. The good thing is there\'s only so much shit that happens before good things start to happen so soon my ship should come in right?  '),(304,4190982,'22,August,2004','2004-08-22 00:00:00','   hmmm!!! a very boring sunday to start off!! online the whole saturday nite !! tired!! woke up 12.30 p.m [sunday]  ouch!!! went cook lunch than wentback to sleep.. huh!~~ so boring!!!! why my weekend so boring!! arggg... stomach ache!!!  '),(305,4190982,'21,August,2004','2004-08-21 00:00:00','   83, 84, 85, 86, 90 ...kg.. argg!~~ my gosh.. i am growing fatter!~~     y? y god made me so fat? sob sob.. i need to loose weight..     lolz.. btw.. one of my friend name RYAN.. bet with me said if i loose weight.. he treat me lunch..     ooOOoo.. how i wish i can loosee weight..      how to lose weight? i dont go gym, nor lazy to diet? how? how? how?     no determination... lazy boy.. how? god plss show me a sign...  '),(306,4190982,'19,August,2004','2004-08-19 00:00:00','   ring ring ring!~~> its 12.30 p.m. wow!! late...        woke up and found i lost my towel. how mystereous it is..  search everywhere--> toilet, living room, bed room can\'t even find a trace of my red towel.. lolz       towel towel where r u?? after that i went outside to check, oh my fucking god my towel its out side... how come? shit!~~~ scary... nearly fainted.       after the towel searching, i went clean up my room.. now its clean and i like it..        clean up the toilet... wash my clothes !!1 huh.. tired...        after all this activities.... i went watch tv. volleyball match (aust vs usa).. n so stupidly aus losses 2 matches..argg@!!! y lose?        lolz--->mika mika  '),(307,4190982,'19,August,2004','2004-08-19 00:00:00','   OooOOOoo.. weekend is coming when it hits friday. huh!! oo how much i hate weekend.. boring boring week end.. nothing to do.. stranded inside a room. no where to go. how sad. lolz well today, its friday... i actually attended my 9.30 applied maths class.. finally i understand what my teacher gordan was teaching.. all these hardwork.. lolz.. after classs.. went mc D n eat.. lolz.. i am eventually declare myself broke and it should be said \"SERIOUSLY BROKE\" muahahaha. concidering i am broke, i actually left 5 dollars and finally i end up using the $5 to play arcade.... lose lose lose... wasted my money.. hahaha after that went have a stroll along swanston street..and after that head back to home. slept till 6.30 until a stupid prank call woke me up [@#%$&amp;@#@^]....... yishhhh!!!! finnally go online... went play gunbound on9 game.. hulalala its my luckly day and i actually won 4 matches out of 5 wah!!!! -->mood become more miserable when its gonna be saturday.. i hate it.. sob sob!!! i hate weekends.. wat to do? anything to do? where to go? able to go on9? can have fun? ewww!~~  '),(308,4190982,'19,August,2004','2004-08-19 00:00:00','   oh yes!!! after days and nites wondering about the blog...hahaha.. (happy). finally i know how to control the studpid blog. muahahaha.   i took me the whole 3 hour class to figure out how to do it. huh!! frustrated. now i am back!!!      - challenges always confront us   - make sure we accept it   - and fight until the last breath      - from time you fail   - never ever give up so easily   - as u fail, you learn from mistakes       -akimoto-  '),(309,4190982,'11,August,2004','2004-08-11 00:00:00','   ooo... headache.... today i\'ve been trying to change all the blog skins n stuff but i still dont know how to operate it. uh!~~ making me crazy. y can\'t i fix it? stupid enough? yish!!!     i dont know why its so hard to modify the blog skin n stuff?     uh!~~ i really gave up doing this.... make me sick n also waste my time.     well, i think i have to do it tomorrow.          ARggg!!!  '),(310,4190982,'10,August,2004','2004-08-10 00:00:00','   arg!~~ decision need to be made by today. i am ask to go to my foundation winter party and now i still havent decided that should i go or not? will it be fun? exciting? hmmmm.....   ------------------------------------------------------------------------------------------------    ohh ohhh.. yeah. finally i\'ve made up my decision to go to the party. well after being advice by my friend. [not really wanna go but there\'s no harm right]hmm.. today a very simple day for me. actually miss my pure maths class coz i am too lazy to wake up n hate to see my add maths teacher that is RAUHA!!! [she makes me SICK] than i eventually attend my computer science class.. lolz.. nah.. i am a very bad student.. didnt even do my comp science work.. online the whole period. hahaha.. hmm. i felt bad for my god sista --> Beckie. she got a diarrhea. huh!~~ just hope she recovers as soon as possibleMikakaka!~~~-akimoto-  '),(311,4190982,'10,August,2004','2004-08-10 00:00:00','   Hmm. What should i say? lolz. Nothing much happen to me. As usual, i went to uni and attend class and sitted for the english exam. pretty easy i should say but hope i can get high marks!~> Than went back home and slept till now. lolz. Well i guess this is for my 1st blog. Hope i can write more next time :D  '),(312,3671580,'29,June,2004','2004-06-29 00:00:00','    urlLink Michael Moore.com : Mike\'s Message : Mike\'s Latest News   Ok,  apparently the link didn\'t go through.  This is the web page, so if you want to see the theater, just scroll down to Minnesota.   '),(313,3671580,'28,June,2004','2004-06-28 00:00:00','    urlLink    Inside the theater.   '),(314,3671580,'28,June,2004','2004-06-28 00:00:00','    urlLink    This is the theater where I went to see Farenheit 9-11   '),(315,3671580,'28,June,2004','2004-06-28 00:00:00','    This afternoon I drove into uptown Minneapolis to catch a late afternoon showing of the new Michael Moore movie.  I was very impressed by the film. Yes the film is a very slanted propoganda-documentary, but when has Moore ever denied that?    I believe Moore\'s strength lies in his ability to combine chilling video clips with poignant music.  He is also very adept at connecting events that otherwise appear to have no connection. Nowhere are these strengths more evident than in this movie.    My biggest problem with the film was its fragmentation.  He begins with a 20 min. look at how Gore lost the election, proceedes to a 45 minute examination of the Bush-Saudi connection, continues by examining the Afghanistan war for 30 min. (along with its connections to the defense industry), and then finishes with an attack against the administration\'s Iraq war.  Although all these parts are interesting, there is no thread to tie them together.  Moore begins by discussing Bin Laden, and ends up on the old standby of Corporate Greed, and the plight of the downtrodden.    I think at times Moore has a tendency to go over the top, but I do believe he is genuine in his crusade, and therefore I respect him.  On tonight\'s Daily Show they showed an interview he did on CBS where the anchor clearly intended to ambush him.  However, Moore quickly turned that ambush on its head and confronted the Woman about CBS\'s supposed lack of journalistic integrity.  It was rather funny.   '),(316,3671580,'27,June,2004','2004-06-27 00:00:00','    urlLink WorldTribune.com: Move over Tiger: N. Korea\'s Kim shot 38 under par his 1st time out   This was just too good to pass up.   '),(317,3671580,'27,June,2004','2004-06-27 00:00:00','    Last night the people across the hall moved out, throwing what looked like all of their worldly posessions in front of the dumpster.  Included in that pile were a leather love seat, a chair, and a matching ottoman, all invaluable items for a poor grad student who will need to furnish an apartment in another month.    So, with the help of my roommate I emulated my good buddy Fred and pillaged the dumpster pile.  I am very happy with my latest accusitions, and will continue to keep a hawkish eye on the junk heap in case anything else of interest is added.   '),(318,3671580,'25,June,2004','2004-06-25 00:00:00','    urlLink    My pillaged Leather \"Like-Seat\"&nbsp; urlLink   '),(319,3671580,'25,June,2004','2004-06-25 00:00:00','    urlLink    My new chair in a temporary holding space.  Behind it, on the other side of the wall sits the matching Ottoman.&nbsp; urlLink   '),(320,3671580,'25,June,2004','2004-06-25 00:00:00','    Yesterday Minnesotans officially entered the 20th century with the arrival of the new light rail train.  I met my friend Elma and her roommate to check out the festivities first hand.  From the crowds of people waiting to ride the train, you would think that the people here have never ridden a subway train before.    After riding through several stations, we went back towards the university to meet up with Alec.  Following lunch at a chinese restaurant (my fortune stated:  \"Don\'t expect romantic attachments to be strictly logical or rational!\" ;) ) we headed back downtown to the Metrodome for the Twins-Brewers game.    The game itself was not noteworthy (the Twins lost 7-2).  However we were scolded during the singing of the national anthem by a militant ueber nationalist Brewers fan for not demonstrating proper respect for the country.    All in all this was a very good Saturday.   '),(321,3671580,'25,June,2004','2004-06-25 00:00:00','    urlLink    The Twentieth century has arrived!!!  Hooray!!!&nbsp; urlLink   '),(322,3671580,'25,June,2004','2004-06-25 00:00:00','    urlLink    View from the Lake Street station.&nbsp; urlLink   '),(323,3671580,'25,June,2004','2004-06-25 00:00:00','    urlLink    You\'d think these people had never rode a train before.  This is the line just to get on the platform to wait for the train.&nbsp; urlLink   '),(324,3671580,'25,June,2004','2004-06-25 00:00:00','    urlLink    Waiting for the train.&nbsp; urlLink   '),(325,3671580,'25,June,2004','2004-06-25 00:00:00','    urlLink    Lunch at Peking Garden. I reccommend the Szechuan Chicken.&nbsp; urlLink   '),(326,3671580,'25,June,2004','2004-06-25 00:00:00','    urlLink    My obligatory Metrodome shot.&nbsp; urlLink   '),(327,3671580,'25,June,2004','2004-06-25 00:00:00','    So, I bet you\'re all wondering what a bored grad student does on a Friday night.  Well, tonight I decided to try something different.  I wanted to recreate a great moment in Blogging history.  So, I got my digital camera, went to Rainbow Foods (a division of Roundys for those not in the know), and attempted to recreate  urlLink this great moment from May 30, 2004.     On this date in Blogging history, San Antonio intellectual Micah Valine astutely documented the rising costs of milk.  Placing his camera behind the milk containers, Micah artistically conveyed this serious social, economic, and potentially political issue.    Recreating this moment, I needed to place a half gallon of fruit juice under the camera to prop it up.  Also, unlike the original, my recreation can actually be accepted as the view from the other side, in that milk stalkers, uhhh. .. stockers actually place new milk in the case from the other side of the wall, ensuring regular product rotation, and fresh milk for you the consumer.    This has been a Great Moment in Blogging History.   '),(328,3671580,'24,June,2004','2004-06-24 00:00:00','    urlLink    Take 2&nbsp; urlLink   '),(329,3671580,'24,June,2004','2004-06-24 00:00:00','    urlLink    Take 3&nbsp; urlLink   '),(330,3671580,'24,June,2004','2004-06-24 00:00:00','    urlLink    Take four of my recreation&nbsp; urlLink   '),(331,3671580,'24,June,2004','2004-06-24 00:00:00','    urlLink    As you can see, milk prices have risen here too.&nbsp; urlLink   '),(332,3671580,'24,June,2004','2004-06-24 00:00:00','    Tonight I went to see the Darkness play at the River Centre in St. Paul.  The concert was awesome!!!!  Wow!  Mad props to Freddie for pushing me to buy the ticket.  It is easily the third best concert I\'ve seen, after the White Stripes, and the Foo Fighters.  At one point, the lead singer did a tour de force guitar solo, and came within 4 feet of me.  What a show!!!!   '),(333,3671580,'24,June,2004','2004-06-24 00:00:00','    urlLink    The Darkness rocked my socks off.&nbsp; urlLink   '),(334,3671580,'24,June,2004','2004-06-24 00:00:00','    urlLink    Me, getting ready to rock out with the Darkness.&nbsp; urlLink   '),(335,3671580,'24,June,2004','2004-06-24 00:00:00','    urlLink The Club For Growth - Home Page     urlLink Advertisement   This is a link to a new commercial being run by the Club for Growth.  I have no clue what the message is they are trying to convey.  This has to be the worst propaganda piece I\'ve ever seen.  To quote Adam Sandler on SNL:  \"Who are the Ad-wizards who came up with this one?\"   '),(336,3671580,'24,June,2004','2004-06-24 00:00:00','    urlLink CNN.com - Gene mutation makes tot super strong - Jun 24, 2004    Settling the argument once and for all.  Uebermensch ist ein Deutscher!   '),(337,3671580,'23,June,2004','2004-06-23 00:00:00','    Today I went to downtown St. Paul in order to purchase a ticket for friday night\'s Darkness concert, sans $8.50 surcharge.  It really is a nice city.  Much more pedestrian than Minneapolis.  Below are some pictures from St. Paul, and my walk into Minneapolis on Saturday afternoon.   '),(338,3671580,'22,June,2004','2004-06-22 00:00:00','    urlLink    Snoopy!  Part of the St. Paul campaign (like the bisons in Buffalo)&nbsp; urlLink   '),(339,3671580,'22,June,2004','2004-06-22 00:00:00','    urlLink    the Excel Center&nbsp; urlLink   '),(340,3671580,'22,June,2004','2004-06-22 00:00:00','    urlLink    Downtown St. Paul&nbsp; urlLink   '),(341,3671580,'22,June,2004','2004-06-22 00:00:00','    urlLink    St. Paul Cathedral&nbsp; urlLink   '),(342,3671580,'22,June,2004','2004-06-22 00:00:00','    urlLink    What trip would be complete without fowl?&nbsp; urlLink   '),(343,3671580,'22,June,2004','2004-06-22 00:00:00','    urlLink    Downtown Minneapolis from the East Bank&nbsp; urlLink   '),(344,3671580,'22,June,2004','2004-06-22 00:00:00','    urlLink    You\'re gonna make it after all&nbsp; urlLink   '),(345,3671580,'22,June,2004','2004-06-22 00:00:00','    urlLink Clinton\'s Empty \'Life\' (washingtonpost.com)    Ok Jeff, I took your advice, and read the post.  Not bad :)   '),(346,3671580,'22,June,2004','2004-06-22 00:00:00','    urlLink The New York Times > Opinion > Op-Ed Columnist: Magboula\'s Brush With Genocide   For those interested in Sudan, here is Kristof\'s latest column.   '),(347,3671580,'22,June,2004','2004-06-22 00:00:00','    urlLink The New York Times > Opinion > Op-Ed Columnist: The Great Cash Cow   Here\'s Safire\'s latest on the UN Oil for Food Scandal.  It\'s too bad he\'s the only person probing this issue.   '),(348,3671580,'22,June,2004','2004-06-22 00:00:00','    urlLink CNN.com - List of 100 best movie songs - Jun 23, 2004    Well, apparently not the Ghostbusters.  The AFI released its list of the 100 greatest film songs today, but failed to list the Ray Parker Jr. classic.  How this song could not make the top 100 is beyond me.  If Eminem makes the list, Ghostbusters deserve to be on there too. Harrumph!   '),(349,3671580,'22,June,2004','2004-06-22 00:00:00','    Bill Clinton, of that fabulous acting duo of Bill and Hillary.  Before the hoopla surrounding Clinton\'s new memoir arose this week, I was actually starting to change my opinion of the former president.  I began to look back at his accomplishments, and acknowledge, that he was a fairly successful president.    Now however, I am once again reminded why I dislike him so much.  Clinton, like his wife, is an egomaniacal, mendacious, Megalomaniac.  For a man who could remember absolutely nothing for eight years, he suddenly has an awful lot to say.   Although I have not read the book (and will not waste my time doing so), I have read enough about it to know that it contains nothing revealing, and that if anything it is just another frantic attempt by Clinton to cement his legacy.  Clinton does not confront his decision to pardon Mark Rich, a billionaire charged in 1983 with evading $48 million in taxes as well as running illegal oil deals with Iran during the hostage crisis.  Rich fled to Switzerland to evade the charges, but essentially bought a pardon with the help of his wife.  The pardon was heavily criticized by both Republicans and Democrats.   Clinton also fails to discuss the 1996 funding scandal involving Johnny Chung, a Chinese businessman who used his business connections in China to funnel millions of dollars into Clinton\'s campaign fund for the 1996 election.  In exchange for the money Clinton altered US Arms sales restriction laws to allow US companies to sell China military technology.  Democratic senator Joeseph Lieberman later said: \"The fundraising scandal of 1996 was a very real tragedy, with very real consequences for our democracy.\"  Finally, in an attempt to keep up with Hillary, Clinton also compares himself to Nelson Mandela. (Hillary compared her forgiveness of Bill with Mandela\'s forgiveness of White South Africans).  If this does not smack of Megalomania, I do not know what does.  So, for those of you still believing this is the best administration of the century, I leave you with this:  his  urlLink rap sheet .   '),(350,3671580,'21,June,2004','2004-06-21 00:00:00','    Listening to the events of and discourse over the Iraq war, I cannot help but think the world is ruled by one of my favorite authors, Kafka.    The Americans invaded the country in an attempt to remove the weapons of mass destruction, and end human rights abuses, yet turned out to be worse than the  urlLink deposed .  We have also learned that France and Russia, two of the staunchest defenders of peace, actually fought the war out of the fear that the massive corruption of the oil for food program would come to  light.   The list goes on and on, reading exactly like something Kafka himself would write.   Everyone argues that this war is the fault of the US, and that the US should therefore pay the price alone.  However, I think if anything, this is just exacerbating the problem.  Yes, the US occupation has been a disaster, but conditions could very easily be better if the world was just be more open to helping.    I\'ll conceede, the US hasn\'t been willing to ceede authority over Iraq, but who ultimately does that hurt?  Iraqis!  As the international community tries to stick it to the US for being the unopposed superpower, the situation just gets worse and worse inside Iraq.  Neighboring middle eastern countries do not want to see Iraq succeed out of the fear that a democratic government there will create problems for their despotism.     There\'s nothing we can do now about the war.  The US is there, and to completely pull out now would only make things worse.  Although this may be an opportunity to challenge the American unipolar order, the only losers in this are the Iraqis.    Everyone criticizes the US for acting, asking why it was so important to go to war over this.  As the corruption within the UN proved though, the relief was not achieving its desired end, and the enforcement of the No-Fly Zones was just a waste of money, material and life.  So, as my communist friends ask me all the time, why was it so wrong to take a chance?  Instead of continuing to plod along in a stagnant international environment, content contain Saddam, as long as he only made his own people\'s lives miserable, the Bush administration wanted to remake Iraq in the shadow of Western Liberal Democracy.    Perhaps if we start to work together, we can end this Kafkaesque nightmare.   '),(351,3671580,'21,June,2004','2004-06-21 00:00:00','    I\'m doing more photoshop work right now, listening to my favorite English radio station on ITunes.  Yesterday, England beat Croatia 4-2, to advance in the EM 2004.  To celebrate, the English have combined Wuhoo by the 5678s (a la Kill Bill fame) with a bunch of Soccer play by play calls and cheers.  Its kind of catchy, and they play it every 30 min.  Now having said that, its time for the daily 1:45 playing of Travis. . .   '),(352,3671580,'21,June,2004','2004-06-21 00:00:00','    urlLink The Zelikow Report   In today\'s New York Times, William Safire comments on the recently released Zelikow Report, an \"interim report\" issued by the Commission\'s staff.  The report, picked up by news media under headlines that dismiss any Iraq-Al Qaeda connection, has clearly been misconstrued.  As Safire clearly demonstrates, the report combines the issues of Iraqi-Qaeda relations (which there were), and an Iraqi connection to 9/11 (which there was none).  It attempts to argue on the basis of Iraq\'s 9/11 involvement (or lack thereof) that no connection between the two entities existed whatsoever.    I think Safire\'s column is yet another reminder of why the 9/11 commission is nothing but a joke.  It was designed to be a bipartisan effort to determine the root causes and failures that led to the terrorist strikes.   However, it has been more concerned with assigning blame solely on one party or the other.  As current and former top officials were questioned by the commission, the only member to question both sides with equal vigor was Bob Kerry.  The Zelikow report is yet another example of the commission\'s failure to do its job correctly.  Supposedly published without the commissions knowledge, the report seeks to discredit Bush more than uncover the truth.    I do not think you can completely blame either Republicans or Democrats for 9/11.  Yes the Bush administration did not fight the terrorist threat with as much zeal as it should have, but at the same time, Clinton had his chance to get Bin Laden in the late 1990\'s and failed.  There are also many other root failures that can be traced back to earlier administrations.    I sympathize with those families that lost relatives in the attacks.  However, I do not think it prudent to brand the Bush and Guilianni administrations as a band of murderers.  Giulianni had New York City arguably more prepared for a terrorist attack than any other American City.  Unfortunately his anti-terror headquarters was destroyed in the bombings.  At the same time, the mayor himself admitted that he believed an attack would most likely come in the form of a chemical or biological bomb, not kamikazi Boeings.  Yet, in the face of so much chaos, Gulianni\'s stoic leadership arguably prevented the situation from being much, much worse.  As the mayor himself has said, if anyone is to be blamed for the attacks, how about the 19 individuals who actually carried out the attack?  If we truly want to get to the root causes of the 9/11 failures, the commission must do what it was intended to do.  As Safire argues, it must show how Clinton\'s failure to respond to the USS Cole bombings contributed to 9/11.  It must demonstrate how Bush\'s refusal to replace George Tenet as CIA director would be beneficial, and finally it must demonstrate \"how Congress\'s intelligence oversight failed abysmally.\"  Only then, when all parties are criticized equally will the commission regain the credibility it has so shamelessly lost.   '),(353,3671580,'19,June,2004','2004-06-19 00:00:00','    Kristof continues his saga on Sudan in today\'s editorial.  http://www.nytimes.com/2004/06/19/opinion/19KRIS.html   '),(354,3671580,'18,June,2004','2004-06-18 00:00:00','    href=\"http://www.nytimes.com/2004/06/18/opinion/18FRI1.html\">Time for Action on Sudan   Many people may not be aware, but a brutal massacre has been raging in Sudan.  Although it is part of a greater civil war that has been raging nearly non-stop for decades, the last several months have seen a dramatic rise in the rate of murder.  Instead of recounting everything here, I will leave it to you to search for info on the background.  The American media (surprise surprise) hasn\'t covered this too extensively, but I have found the Economist and the New York Times to provide decent reporting. Keep an eye out for Nicholas Kristof  http://www.nytimes.com/2004/06/16/opinion/16KRIS.html?n=Top%2fOpinion%2fEditorials%20and%20Op%2dEd%2fOp%2dEd%2fColumnists>Dare We Call it Genocide?  he has been writing regularly on the subject for the past couple months, and provides a vivid description of the situation.    Although a discussion on Sudan has been active for a while, I am moved to post on it because of today\'s NYT op-ed column.  As the column clearly argues, although there has been much discussion and condemnation in diplomatic circles, there has been absolutely NO ACTION.    China, Pakistan and Algeria stonewall any attempt made by the UN to act, doing more to actually protect the Sudanese government.    At the same time, the Bush administration is too busy examining whether or not the disaster in Sudan legally meets the definition of Genocide.  I am abhorred by the failure of the Western democracies to do anything on behalf of the Sudanese people.  After the Holocaust, we hear all to often \"Never Again!\" in reference to preventing another Genocide from breaking out.  Yet, the remainder of the twentieth century is marked by the West\'s failure to take any action to back up this statement.  Nothing was done in Cambodia during the 1970\'s, and in the mid 1990\'s, the UN just stood around, and monitored the massacres in Rwanda.    President Clinton often points to the NATO action he led to end the killings in Kosovo (a little delayed, but better late than never right?), but where was he when the UN needed someone to lead a force into Kosovo?  In the one month the UN spent debating what to do, almost 1 million people were hacked to death.    Rwanda was supposed to change everything.  This was supposed to be the real eye opener.  Now we know that Genocide doesn\'t take months and months to develop, so expedited reaction is an even greater necessity.    Once again though, many thousands of people are dying, and those that can make a difference are sitting on their hands, claiming that they cannot act until they are positive that Sudan meets all the legal requirements of Genocide.  Does it really make a difference?  People are people!  Is it considered less morally reprehensible if it is deemed that the Sudanese government did not really attempt to wipe out every Black African/Christian/Animist?    Bush does deserve to be praised for negotiating a peace in the civil war, but permitting the Genocide now in exchange for maintaining the accord is pointless.  The same can be said for travel restrictions and sanctions.  As the Times points out, what good will that do on a government that is already banned from half the world and has no assets to freeze?  I think the biggest problem I have with the Bush administration\'s response is the sudden concern over legality.  They were all to eager to invade Iraq, and remove Saddam, who was such an urgent threat to national security.  Now, I supported the invasion of Iraq.  I think there were too many broken resolutions to overlook.  However, I think a much stronger case could be made for invading Sudan.  The humanitarian crisis in Sudan is much worse than the situation under Saddam.  At least Iraq did have a central government that was half heartedly committed to distributing foreign aid (yes this was only after half of the aid was confiscated for Saddam himself).    At the same time Sudan can also be classified as a failed state, a state with a very weak central government.  Its much more likely that the Sudanese government would support Terrorism, or allow terrorists to operate off of Sudanese soil.  They did house Osama at one point, and I don\'t find it far fetched to believe they would do it again if the price was right.    Furthermore, as a result of the peace treaty, Bush is immensely popular in the South.  It would be a perfect opportunity for him to deploy the military in a situation where he would not only receive immense praise from the international community, but also be openly embraced by the native populations!  Alas, this appears to be too much to ask.  The world will once again passionately cry out \"Never Again\" while standing idly by and not lifting a finger.   '),(355,3671580,'17,June,2004','2004-06-17 00:00:00','    I\'m hesitant to voice my grievances with the American Media, but I just cannot restrain myself right now.    This afternoon, two very newsworthy events occurred that deserved an immense amount of media attention.  However, based on the medias coverage of these events, you would never know of the second one.    The first event I speak of is the tragic beheading of Paul Johnson.  The story has dominated the headlines, and rightfully so.  However, the second event I speak of has barely received any attention.    So, what is this event you ask?  Well, I\'ll give you a hint:  It will have a significant impact on the current Geopolitical environment.  Don\'t know?  Here\'s another:  The combined economic strength of this area is almost equivalent to that of the United States.  Still no clue?  Ok, one last hint:  The agreement reached today will help pave the way for this region to become a Superpower that will rival the USA.  If you still don\'t know, I would like to say I am disappointed in you, but I can\'t because this event has received virtually NO ATTENTION in the media!!!!!  Of course I\'m speaking of the new EU constitution that was finally agreed upon today.  http://www.cnn.com/2004/WORLD/europe/06/18/eu.summit/index.html  Six months after failing to reach an agreement, European diplomats finally agreed on a constitution today.  Of course it still needs to be approved by 13 of the 25 members to take effect, but this is still a significant event.  However, this is the first step in creating a closer union among the EU states that could ultimately rival the US as a second superpower.  The consequences of such a union would be immense for the US.  If taken to an extreme, the EU could help usher in a new bi-polar age where the US no longer has the opportunity to create hegemonic order unopposed.   Sadly, this received no coverage in the media today.  It is currently listed #3 on www.cnn.com under the WORLD section.  That\'s right, this was not even important enough to make the normal headline section.  Apparently McDonald\'s food testing, a rapper who made a video in jail, and Bill Clinton\'s admission that he slept on the couch after coming clean to Hillary about his escapades with Monica are all more important than the EU.  As if this is bad enough, Fox News does not even MENTION the story.  I guess if it is not wrapped in Red White and Blue singing the high praises of W., then it is an anti-American, Benedict Arnold, Liberal conspiracy that is out to deep six everything good and wholesome.    I understand that the primary goal of American news media outlets is to focus on those events which are of relevance to Americans.  I also realize that my complaints are nothing new, and that I am just reiterating grievances raised by numerous others.  However, I am nevertheless disappointed by the coverage given to the EU today.   '),(356,3671580,'17,June,2004','2004-06-17 00:00:00','    urlLink    I just thought this was a really funny sign on my way to school this morning.&nbsp; urlLink   '),(357,3671580,'17,June,2004','2004-06-17 00:00:00','    urlLink    The view across the street from the movie theater of downtown Minneapolis.&nbsp; urlLink   '),(358,3671580,'17,June,2004','2004-06-17 00:00:00','    urlLink    The Movie Theater, part of a redeveloped stretch along the river.&nbsp; urlLink   '),(359,3671580,'17,June,2004','2004-06-17 00:00:00','    On the way home from school, I decided to stop by the movie theater and see the new Ben Stiller movie:  Dodgeball.  I had pretty low expectations for the film, but was pleasantly surprised in that it actually was pretty funny.    I thought the high point of the film was Jason Bateman\'s performance as Pepper Brooks, the color commentator for the ESPN 8 telecast of the dodgeball tourney.  Bateman perfectly captures the role of the inane commentator that one all to often finds on professional sport broadcasts.  On a different note, I was disappointed by the audience\'s failure to pick up on a hilarious joke involving the German dodgeball team.  I won\'t spoil it for anyone, but I thought it was very subtle, and well done.  I\'ll try to post some pics in a bit.   '),(360,3671580,'17,June,2004','2004-06-17 00:00:00','    As I work on a photoshop project here at the U, I\'m taking advantage of the office\'s subscription to ITunes.  It\'s great, I\'ve got a station in the UK, and I\'m listening to the 80\'s at 8 0\'Clock at 1:15 in the afternoon.  Life is good :)   '),(361,3671580,'17,June,2004','2004-06-17 00:00:00','    Today I bought an invitation for Gmail on Ebay.  It was the greatest $.99 I ever spent.  I don\'t know why yet, but I just know that getting in on the ground floor of this new project will pay off down the road.  When I find out how I will let you know.   '),(362,3671580,'17,June,2004','2004-06-17 00:00:00','    After weeks of badgering, I\'ve finally moved my blog to blogspot.  I feel like such a tool. :)   '),(363,3671580,'31,July,2004','2004-07-31 00:00:00','    urlLink    Harmon Killebrew returns and he has Four Arms!&nbsp; urlLink   '),(364,3671580,'28,July,2004','2004-07-28 00:00:00','    urlLink    Fred is so enthused by the game!&nbsp; urlLink   '),(365,3671580,'28,July,2004','2004-07-28 00:00:00','    urlLink    When the Twins fell behind 7-2 I decided to have some fun with my camera.&nbsp; urlLink   '),(366,3671580,'28,July,2004','2004-07-28 00:00:00','    urlLink    Random Metrodome Shot.&nbsp; urlLink   '),(367,3671580,'28,July,2004','2004-07-28 00:00:00','    urlLink    Me with the big Twins jersey outside the Metrodome!&nbsp; urlLink   '),(368,3671580,'28,July,2004','2004-07-28 00:00:00','    urlLink    Water Fowl!&nbsp; urlLink   '),(369,3671580,'28,July,2004','2004-07-28 00:00:00','    urlLink    Fred is endorsing C2.&nbsp; urlLink   '),(370,3671580,'28,July,2004','2004-07-28 00:00:00','    urlLink    I hiked up Fred\'s shorts to document his bug wounds.  Now tell me this doesn\'t need neosporin!&nbsp; urlLink   '),(371,3671580,'28,July,2004','2004-07-28 00:00:00','    urlLink    Oh no I\'ve been spotted!&nbsp; urlLink   '),(372,3671580,'28,July,2004','2004-07-28 00:00:00','    urlLink    Master photographers at work!&nbsp; urlLink   '),(373,3671580,'28,July,2004','2004-07-28 00:00:00','    urlLink    Fred\'s mug shot at Pizza Luce.&nbsp; urlLink   '),(374,3671580,'28,July,2004','2004-07-28 00:00:00','    urlLink    The happy couple :)&nbsp; urlLink   '),(375,3671580,'28,July,2004','2004-07-28 00:00:00','    urlLink    Me and my good buddy Jango.&nbsp; urlLink   '),(376,3671580,'28,July,2004','2004-07-28 00:00:00','    urlLink    Us at Grandma\'s.  Grandma\'s cooking is as good as ever, but since when did she start charging so much?!&nbsp; urlLink   '),(377,3671580,'28,July,2004','2004-07-28 00:00:00','    urlLink    The gang just got here, and Jeff could not resist showing off his motley assortment of facial cremes and deoderants.  &nbsp; urlLink   '),(378,3671580,'25,July,2004','2004-07-25 00:00:00','    Yesterday I went to a party for my friends Alec and Elma, who just got married a few months ago.  I met Alec and Elma in Dortmund, so a bunch of friends who were in Dortmund then also made the trip up!  It was great seeing old friends again, and nice how we all seemed to pick up right where we left off.   '),(379,3671580,'21,July,2004','2004-07-21 00:00:00','    urlLink    Mr. and Mrs. Alec Johnson!&nbsp; urlLink   '),(380,3671580,'21,July,2004','2004-07-21 00:00:00','    urlLink    The Stammtisch Reunion! :)&nbsp; urlLink   '),(381,3671580,'21,July,2004','2004-07-21 00:00:00','    urlLink    Deep philosophical conversation&nbsp; urlLink   '),(382,3671580,'21,July,2004','2004-07-21 00:00:00','    urlLink    Elma and John&nbsp; urlLink   '),(383,3671580,'21,July,2004','2004-07-21 00:00:00','    urlLink    Alec Rocking Out!!!&nbsp; urlLink   '),(384,3671580,'21,July,2004','2004-07-21 00:00:00','    urlLink    I was awarded the prestigious honor of tapping the Keg.  I\'ve never seen Mark so giddy before :)&nbsp; urlLink   '),(385,3671580,'21,July,2004','2004-07-21 00:00:00','    urlLink    Debating how to move the Keg.&nbsp; urlLink   '),(386,3671580,'21,July,2004','2004-07-21 00:00:00','    urlLink    Mark and Elma\'s brother playing more fooseball.&nbsp; urlLink   '),(387,3671580,'21,July,2004','2004-07-21 00:00:00','    urlLink    The building where the party was held.&nbsp; urlLink   '),(388,3671580,'21,July,2004','2004-07-21 00:00:00','    urlLink    Alec kicking butt in Fooseball.&nbsp; urlLink   '),(389,3671580,'21,July,2004','2004-07-21 00:00:00','    urlLink    Random action shot of Ted&nbsp; urlLink   '),(390,3671580,'21,July,2004','2004-07-21 00:00:00','    urlLink    \"Keith\" trying really hard to be not-so-photogenic.&nbsp; urlLink   '),(391,3671580,'21,July,2004','2004-07-21 00:00:00','    urlLink    John being photogenic.&nbsp; urlLink   '),(392,3671580,'21,July,2004','2004-07-21 00:00:00','    urlLink    Mark is looking for the missing Fooseball.&nbsp; urlLink   '),(393,3671580,'21,July,2004','2004-07-21 00:00:00','    urlLink    The Wedding Party.&nbsp; urlLink   '),(394,3671580,'21,July,2004','2004-07-21 00:00:00','    urlLink    This was a sign posted on the gate surrounding the pool.  John was not allowed in :(  As I asked him though, what about children with Huggies or Luvs?  &nbsp; urlLink   '),(395,3671580,'21,July,2004','2004-07-21 00:00:00','    So I finally posted my pictures from Milwaukee.  The Stadium was ok, but it was like watching a baseball game in a hockey arena.  Miller Park did not have the same feel as the parks in Cincinnati or St. Louis.  Nevertheless it is a decent building.  The high point though was the Bratwurst.  Miller Park had the closest Bratwurst to authentic German Bratwurst.  A must have for anyone who goes to a game there.   '),(396,3671580,'21,July,2004','2004-07-21 00:00:00','    urlLink    The Stadium after the Game.&nbsp; urlLink   '),(397,3671580,'21,July,2004','2004-07-21 00:00:00','    urlLink    No baseball game would be complete without the Sausage race.  Go Bratwurst!&nbsp; urlLink   '),(398,3671580,'21,July,2004','2004-07-21 00:00:00','    urlLink    This was my seat.  It was in a good position behind the plate, but that railing on the left really sucked.&nbsp; urlLink   '),(399,3671580,'21,July,2004','2004-07-21 00:00:00','    urlLink    The Stadium was filled with Cubs Faithful&nbsp; urlLink   '),(400,3671580,'21,July,2004','2004-07-21 00:00:00','    urlLink    Miller Park as I was rushing into the stadium.&nbsp; urlLink   '),(401,3671580,'21,July,2004','2004-07-21 00:00:00','    urlLink    The Obligatory Field Shot&nbsp; urlLink   '),(402,3671580,'21,July,2004','2004-07-21 00:00:00','    Recently, I have read a significant amount of material on homosexuality and how it relates to the Bible.  I also recently found a new church here in Minneapolis, (www.bethanyinseward.com) and much to my surprise discovered that they are currently on the verge of breaking boundaries by ordaining an openly gay pastor.  I personally could not be more supportive of this.    Let me return to my original thought though, the position of homosexuals in the bible.  Citing passages in Romans and 2 Corinthians, many people have argued that homosexuality is a sin, and that therefore ideas like gay marriage should not be legalized.  In a recent conversation with my friend Fred, he stated that, barring a reinterpretation of these passages (something he is not opposed to) homosexuality appears to be sinful.  I hesitate to use Fred as an example, but I believe his statement is emblematic of the problematic nature of the biblical debate over homosexuality.  If I understand this statement correctly, a literal interpretation of the Bible forbids homosexuality, however if biblical scholars chose to provide an acceptable reinterpretation, then it might be ok.    There are several problems with approaching the question from this angle.  First, if one only relies on a literal interpretation, what does that say about other aspects of life the Bible discusses.  For example, when discussing sexual perversion, and self abuse, all of the passages argue that it is unaccepatable for a man to abuse himself, and lust after other women.  Clearly this could be interpreted to understand that masturbation is frowned upon.  However, since it only mentions men, does that mean it is ok for a woman to masturbate?  Based solely on a literal interpretation of the Bible, it would appear so.    I believe it is extremely problematic to take the Bible literally.  The scriptures so often cited have been translated numerous times through different cultures, periods, and by different types of people.  Therefore, how are we to know that this is actually the original text?  For all I know, the passage that explicitly forbids homosexuality could have been purposely altered by a 12th century monk who had a personal disdain for the practice.  As a historian, I am constantly tought that one must be sceptical when researching, that one cannot always take a document at face value.  I believe the same must be said for understanding the Bible.  Its quite possible these passages were altered repeatedly to reflect the general norms of society at the time it was translated.  I think its safe to say that 13 century Gaul is not the same as 21 Century Western society.   So how then is one to understand homosexuality?  I do not believe it is necessary for one to rely on other people to interpret the scriptures for us.  A central tennant of Lutheranism is that every person is his own intperpreter.  Therefore, when trying to understand the Bible\'s position on homosexuality, it is necessary to properly contextualize the passages.  For the purposes of this post, I will rely on Romans 1: 26- 2:6.  In this passage it appears that Paul explicitly rejects homosexual acts.  I find it interesting to note on a tangent, that the verse before states that women should be condemned for having sex in \"unnatural ways.\"  (Does this mean Missionary style is blessed, but that Doggie Style is a sin?  :) )  However, I arrive at a different understanding, once it is contextualized.  As Paul condemns homosexuality, he is also very critical in his letter of Idol worshiping, and emphasizes that salvation can be found in faith and faith alone.  It appears, that in his condemnation of idol worshiping, he uses sexual perversion and homosexuality to emphasize the selfish nature of these people.  He indicates that these people are sinful because they place more emphasis on self-gratification than on worshiping God.  Could it therefore also mean that homosexuality itself is not sinful, but is only sinful when one places sexual gratification above God?  It\'s now 3:18 am, and I really should get to bed, so let me conclude by saying, that we do not need to rely on scholars, or priests to interpret the Bible for us.  If you want to believe homosexuality is a sin, thats fine, its your interpretation.  However, I believe, through personal interpretation, its also possible to accept homosexuality.   '),(403,3671580,'13,July,2004','2004-07-13 00:00:00','    urlLink Chicago Tribune | New spin for Rock \'n\' Roll McDonald\'s   Heresy!  Wesley is spinning in his grave!   '),(404,3671580,'12,July,2004','2004-07-12 00:00:00','    urlLink FOXNews.com - Foxlife - \'Jeffersons\' Actress Isabel Sanford Dies at 86     \"Isabel \'Weezie\' Sanford expires of natural causes in Los Angeles\"  This was the subtext of the fox news headline.  Could this possibly be phrased any worse?   '),(405,3671580,'11,July,2004','2004-07-11 00:00:00','    urlLink ESPN.com - MLB - MLB PREVIEW    Jeff and I have recently been arguing about which team is better, the cubs or the cardinals.  Although the Cardinals are well beyond the rest of the division, one can see that the two teams are virtually equal in all the major statistical categories.  There is one glaring difference however.  The Cubs have a slew of players still out with injuries.  I think that once these players come back, and the Cubs are whole again, they will once again be one of the best teams in the league.  A complete roster will make them very dangerous in October.   '),(406,3671580,'08,July,2004','2004-07-08 00:00:00','    urlLink   '),(407,3671580,'05,July,2004','2004-07-05 00:00:00','    So, here are the pics from St. Louis on Tuesday.  I have several photos of Miller Park in Milwaukee to post as well.  I will do that perhaps tomorrow. Perhaps I will also post more closing comments.  In any event, enjoy these pictures of St. Louis.  It was by far the best city we went to, and the stadium had the most character.  In conclusion, Go CUBS!   '),(408,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    Metro Link on the way home&nbsp; urlLink   '),(409,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    Batting practice&nbsp; urlLink   '),(410,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    Go Cardinals!&nbsp; urlLink   '),(411,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    Construction on the new stadium next door&nbsp; urlLink   '),(412,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    The Stadium from Center Field&nbsp; urlLink   '),(413,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    Big Mac Land!&nbsp; urlLink   '),(414,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    there\'s so much character in Busch stadium, you can\'t find it anywhere else.&nbsp; urlLink   '),(415,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    Outside of Busch Stadium&nbsp; urlLink   '),(416,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    The Savvis Center.&nbsp; urlLink   '),(417,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    Inventive advertising in St. Louis&nbsp; urlLink   '),(418,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    The fan shop in the mall&nbsp; urlLink   '),(419,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    Art Deco!&nbsp; urlLink   '),(420,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    Union Central Station in St. Louis&nbsp; urlLink   '),(421,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    One of the many court houses in St. Louis.&nbsp; urlLink   '),(422,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    The Arch behind the Jefferson Memorial.&nbsp; urlLink   '),(423,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    The Arch behind the Jefferson memorial.&nbsp; urlLink   '),(424,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    The entrance to the Horse\'s Anus ride, operated by the plutonians.&nbsp; urlLink   '),(425,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    Underneath the Arch.&nbsp; urlLink   '),(426,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    East St. Louis, where we parked the car and took the train in.  We were later accosted by a Crack Whore on the way back afterwards.&nbsp; urlLink   '),(427,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    Busch Stadium from the car.&nbsp; urlLink   '),(428,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    Carshot of Downtown St. Louis.  Jeff wanted to kill me afterwards because he was trying to change lanes while I was taking pictures.&nbsp; urlLink   '),(429,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    Jeff in the train.  &nbsp; urlLink   '),(430,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    St. Louis, keep Left!&nbsp; urlLink   '),(431,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    Welcome to Louisville.&nbsp; urlLink   '),(432,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    Welcome to Illinois.&nbsp; urlLink   '),(433,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    Car shot of downtown Louisville.&nbsp; urlLink   '),(434,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    Here\'s why.&nbsp; urlLink   '),(435,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    We wanted Dairy Queen, but it was closed :(&nbsp; urlLink   '),(436,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    Another Random Nashville shot.&nbsp; urlLink   '),(437,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    This is the Chemistry building.  We took this photo for Joe.&nbsp; urlLink   '),(438,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    Nashville, Indiana.  The wierdest place ever.&nbsp; urlLink   '),(439,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    This is my Schrockesque action shot at the U.  I\'m reading an alternative paper.  :)&nbsp; urlLink   '),(440,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    University of Indiana&nbsp; urlLink   '),(441,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    Entrance to the University of Indiana&nbsp; urlLink   '),(442,3671580,'05,July,2004','2004-07-05 00:00:00','    urlLink    Jeff after using the facilities in Bloomington, Indiana.  This photo was taken in the same style as previous photos of Fred and Joe in Europe.&nbsp; urlLink   '),(443,3671580,'05,July,2004','2004-07-05 00:00:00','    Yesterday we went to Cincinnati to see the Reds play the Indians on the fourth of July.  The game was good, even though the Indians lost, and I got really sunburned.  After the game we drove to Dayton, and decided there to cancel the next day of the road trip for an off day.  Jeff bought a baseball glove at Meijer while I purchased Aloe Vera Gel for my sun burns.    When we got home, we watched Meatball and Milkshake, but Jeff was unimpressed.  This morning we played catch on the rolling green pastures of Terre Haute.  Tommorrow:  St. Louis!   '),(444,3671580,'03,July,2004','2004-07-03 00:00:00','    urlLink    We ate dinner at a Big Boy restaurant.  Jeff on the ride from Cincinnati:  \"I really want a Big Boy!\"&nbsp; urlLink   '),(445,3671580,'03,July,2004','2004-07-03 00:00:00','    urlLink    On the way there.&nbsp; urlLink   '),(446,3671580,'03,July,2004','2004-07-03 00:00:00','    urlLink    We stopped at Dunkin Donuts to get coffee and Bavarian Creme Donuts.&nbsp; urlLink   '),(447,3671580,'03,July,2004','2004-07-03 00:00:00','    urlLink    Welcome to Ohio!&nbsp; urlLink   '),(448,3671580,'03,July,2004','2004-07-03 00:00:00','    urlLink    The Pepsi Powerstacks.&nbsp; urlLink   '),(449,3671580,'03,July,2004','2004-07-03 00:00:00','    urlLink    The Cleveland Indians, and Dunkin Donuts Coffee.  Could life get better?&nbsp; urlLink   '),(450,3671580,'03,July,2004','2004-07-03 00:00:00','    urlLink    The Ballpark after the Tribe blew it.&nbsp; urlLink   '),(451,3671580,'03,July,2004','2004-07-03 00:00:00','    urlLink    Paul Brown Stadium&nbsp; urlLink   '),(452,3671580,'03,July,2004','2004-07-03 00:00:00','    urlLink    It\'s the lawn mower/bath tub.  Fred, we thought about getting this for you.  It seemed like something you would be interested in.  :)&nbsp; urlLink   '),(453,3671580,'03,July,2004','2004-07-03 00:00:00','    urlLink    Downtown Cincinnati&nbsp; urlLink   '),(454,3671580,'03,July,2004','2004-07-03 00:00:00','    urlLink    I almost missed this shot.  Jeff sped up so I couldn\'t get it, and threatened to swerve the car.&nbsp; urlLink   '),(455,3671580,'03,July,2004','2004-07-03 00:00:00','    urlLink    Back to Terre Haute!  We decided we needed an off day on the Baseball Odyssey.&nbsp; urlLink   '),(456,3671580,'03,July,2004','2004-07-03 00:00:00','    Today the Odyssey almost encountered a setback with severe thunder storms.  However, with the help of Fred and Megan\'s skilled weather analysis, the storms passed by and allowed for just enough time for the Indians and Bats to play ball!    Before the game, Jeff once again failed in his quest to find a renewable energy source, which led us to the Indianapolis convention center to make a pit stop. Apparently today the African Methodist church was electing its bishop for the year, which led to massive last minute campaigns.  As Jeff said upon entering the building:  \"Do you think they\'ll know we aren\'t here for the conference?\"    Tomorrow, on to Cincinnati!   '),(457,3671580,'02,July,2004','2004-07-02 00:00:00','    urlLink    Jeff and I took pictures of Squirrels today.  This one approached us at ISU this morning as we tried to Blog in front of the student union.&nbsp; urlLink   '),(458,3671580,'02,July,2004','2004-07-02 00:00:00','    urlLink    If only Vinnie Cook could have dorms like these. . . .&nbsp; urlLink   '),(459,3671580,'02,July,2004','2004-07-02 00:00:00','    urlLink    This is Depaw College, a really rich, aristocratic, plantation-esque liberal arts school outside Indianapolis.  Believe it or not, this is the Math and Science building.&nbsp; urlLink   '),(460,3671580,'02,July,2004','2004-07-02 00:00:00','    urlLink    Aloe Vera, the underground cult punk rockers of Indianapolis play their most successful show to date!&nbsp; urlLink   '),(461,3671580,'02,July,2004','2004-07-02 00:00:00','    urlLink    Art shot! :)&nbsp; urlLink   '),(462,3671580,'02,July,2004','2004-07-02 00:00:00','    urlLink    I support Winston Hill for the Bishop of the African Methodist/Episcopal Church. . . . &nbsp; urlLink   '),(463,3671580,'08,August,2004','2004-08-08 00:00:00','    After Jeff and Fred unceremoniously blew out of my apartment like bats out of Hell at 6:36 AM, Megan and I followed in close pursuit at 6:48 as we headed to New Ulm, MN.  New Ulm was a German utopian community founded in the 1800s by a group of German immigrants.  Everything was in German, and the atmosphere of the town was a quaint cross between Europe, 1920\'s Architecture, and the old west.  The high point was definitely Schell\'s Brewery, the second oldest family owned brewery in the country.  All in all, a great day!   '),(464,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Megan is trying to sexually arouse the peacock so he will show us his feathers.&nbsp; urlLink   '),(465,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Another Peacock!&nbsp; urlLink   '),(466,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    The wide assortment of beers&nbsp; urlLink   '),(467,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Nice, but not as many kegs as in Bremen.&nbsp; urlLink   '),(468,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    A peacock with his tail feathers!&nbsp; urlLink   '),(469,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Guten Tag!&nbsp; urlLink   '),(470,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    A fountain in the garden&nbsp; urlLink   '),(471,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    A garden at the Brewery.&nbsp; urlLink   '),(472,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Schell\'s Brewery&nbsp; urlLink   '),(473,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    My Gosh there are Peacocks!&nbsp; urlLink   '),(474,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Can you find the Chipmunk?&nbsp; urlLink   '),(475,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    One of the scenes&nbsp; urlLink   '),(476,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    This is the hill for the Way of the Cross.  It was organized by a group of German Nuns at the turn of the Twentieth Century, and features 12 scenes from Jesus\'s crucifixion.&nbsp; urlLink   '),(477,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    my skirt shot of Herman&nbsp; urlLink   '),(478,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Wilkommen!&nbsp; urlLink   '),(479,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Herman\'s box&nbsp; urlLink   '),(480,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    The view from atop the dome&nbsp; urlLink   '),(481,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    After lunch we went to visit Herman the German, but unfortunately he was taken off his dome for cleaning.&nbsp; urlLink   '),(482,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Megan descending from the top of the dome&nbsp; urlLink   '),(483,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    an art-deco gas station.  Actually, its late art deco because it has curves!&nbsp; urlLink   '),(484,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Fred and Jeff opted for more traditional German fare&nbsp; urlLink   '),(485,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Megan and I ate lunch at the Kaiserhoff restaurant.  Sadly the Spaetzle were sub-par. :(&nbsp; urlLink   '),(486,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    The Glockenspiel&nbsp; urlLink   '),(487,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Can you spot the Butter from Tops?&nbsp; urlLink   '),(488,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Musst du deine Kleidungen waschen?  Denn bitte gehst du rein!&nbsp; urlLink   '),(489,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    New Ulm   urlLink   '),(490,3671580,'08,August,2004','2004-08-08 00:00:00','    Last Sunday we went to play Golf at the Walker Sculpture Garden after another rousing Twins Game.  After the game we went to Stub and Herbs for dinner while we watched the Cardinals game on TV.   '),(491,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Before this picture, Jeff told me:  \"You don\'t have respect for anything. . . get inside that sculpture so I can take your picture!\"&nbsp; urlLink   '),(492,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    I was just so affectionate after Golf!  Here not even a hug from me can stop Fred from taking pictures!&nbsp; urlLink   '),(493,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    I wanted to hug Jeff after the game, but he kept running away from me!  Almost like a homophobe. . . :)&nbsp; urlLink   '),(494,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    This bird started some serious crap with me after the game.  It was just asking for it.&nbsp; urlLink   '),(495,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    White ball go down da hole!&nbsp; urlLink   '),(496,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Megan putting, and Photog at work!&nbsp; urlLink   '),(497,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Megan is thuroughly perturbed that I continuously shoot pictures of her.&nbsp; urlLink   '),(498,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    This hole was constructed out of the old city hall bricks.&nbsp; urlLink   '),(499,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Fred and Jeff \"shacking up\" on one of the golf holes.  urlLink   '),(500,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    The Walker Sculpture Garden!&nbsp; urlLink   '),(501,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    The Twins congratulate eachother after a well deserved victory over the Red Sox.&nbsp; urlLink   '),(502,3671580,'08,August,2004','2004-08-08 00:00:00','    Ok, I\'m a little behind in my blogging.  These are pictures from the Twins REd Sox game last Friday, as well as pics from our trip to Downtown St. Paul on Saturday.  Both were awesome excursions, eventhough we were almost killed by a never ending pack of bike riders in front of the capitol building.   '),(503,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    I am fascinated by Fred\'s random neck hairs in St. Paul!  I can\'t remember if he decided to document my random facial hairs at the same time. . . &nbsp; urlLink   '),(504,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Snoopy is everywhere!  I think Megan is hiding behind the green one up front.&nbsp; urlLink   '),(505,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Downtown St. Paul from Cathedral hill&nbsp; urlLink   '),(506,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Church Altar&nbsp; urlLink   '),(507,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    My Voyeuristic up the dome shot of the Cathedral ;)&nbsp; urlLink   '),(508,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Unfortunately Charlie Conway\'s mom wasn\'t there to serve us :(&nbsp; urlLink   '),(509,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    The Facade of the St. Paul Cathedral.&nbsp; urlLink   '),(510,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    HA!  And you didn\'t believe me!&nbsp; urlLink   '),(511,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Second largest free standing marble dome in the world!&nbsp; urlLink   '),(512,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Megan is so enthusasitic about our tour!&nbsp; urlLink   '),(513,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Jeff is mad at me because I made fun of his shirt.  I think he\'s threatening me here :)&nbsp; urlLink   '),(514,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Fred thinks he\'s Rocky!&nbsp; urlLink   '),(515,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Our wonderful State Capitol building!&nbsp; urlLink   '),(516,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    St. Paul is falling apart!&nbsp; urlLink   '),(517,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    This was my \"surprise\" that was hyped all week.&nbsp; urlLink   '),(518,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    The Twins were getting blown out so I decided to have some fun with my camera. . . &nbsp; urlLink   '),(519,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    AAAAAHHHHHH!!  Harmon Killebrew has grown 4 arms!&nbsp; urlLink   '),(520,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Standard candid action shot.&nbsp; urlLink   '),(521,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Fred is so into the game!&nbsp; urlLink   '),(522,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    This was almost as good as the Expos!&nbsp; urlLink   '),(523,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Obligatory water fowl shot&nbsp; urlLink   '),(524,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Fred endorses Coke 2&nbsp; urlLink   '),(525,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Oh no!  I\'ve been spotted!&nbsp; urlLink   '),(526,3671580,'08,August,2004','2004-08-08 00:00:00','    urlLink    Fred at work!&nbsp; urlLink   '),(527,3155970,'29,April,2004','2004-04-29 00:00:00','   From what I see on Oprah, being a mom is a tough job.  I\'m sure my own mother would agree with that.  She served as cook, chauffer, tutor, maid, butler, self-help guru, bread winner, and referee to both my sister and for many years and even now still.  But when my momther entered into this contract of never ending servitude she was in her twenties and already married.  Her struggles began after she left my father and had to raise us on her own.  Congrats to her because she did a great job (obviously I have no issues).  But should the same congratulations be given to all women that sacrifice their lives for motherhood?  Should we put teen moms on a pedestal and admire their sacrifice?  Hell no!!!  Yes motherhood is difficult, fatherhood is difficult.  Raising children is tough and that should be beat into the heads of every man women and child.  Honoring a 19 year-old girl for having a 3 year-old kid is the stupidest thing I have ever heard of.  Instead she should be told what a dumb fuck she is for getting knocked up and use her as an example for all little boys and girls for what not to become when you are 16...a mom or a dad.  At 19 you are still a baby and shouldn\'t be raising one.  Yes accidents happen everyday and what is fun and exciting for 10 minutes can lead to a lifetime of hard work and servitude.    So you got knocked up at 16 and decided to have the baby, sacrificing whatever sort of life you have led in order to be a mom.  Good for you, here\'s a cookie.  But don\'t parade your ass on TV and say shit like \"I\'m a mom so I deserve a break.\"  No you don\'t deserve anything, what you should have gotten was a kick upside the head for being stupid at 16.   '),(528,3155970,'27,April,2004','2004-04-27 00:00:00','   This past Sunday I got stuck in downtown DC due to the pro-choice march down Pennsylvania Ave.  I\'m not really passionate about any causes enough to march around a city block for hours but I don\'t mind watching.  Its great to see people come together to fight for a cause, to see unity among men and women all of different backgrounds, colors, ages, and sexual preferences.  To see people peacefully expressing their opinions.  What is not good to see is flabby guts spilling out of belly shirts, fat chicks dressed as belly dancers shaking all over the place, and ugly guys with full beards dressed as women.  This wasn\'t a peaceful protest march it was a cry for help!!!  \"Please help me because I look like I fell off the ugly truck.\"  If I were a guy looking to get laid I would definitely had made to go to a pro-choice march because you know all the chicks there would soooooooooo put out.  I\'m glad however I am not a guy because I would have been sooooooooooooo turned off by all the uggies in the crowd.  I have never seen a larger gathering of nasty, fat, disgusting looking women in all my life.   I know, I know, I am a women and I should be behind all of my \"sisters\" and feel proud to see so many of them not giving a damn what society thinks and letting their big flappy guts fly.  Fuck that!!!  It was nasty and if I had eaten something before going to the march I would have heaved it all over them.   '),(529,3155970,'23,April,2004','2004-04-23 00:00:00','   A good question that I ask myself all the time.  \"What is wrong with my life?\"  Nothing really, I have a job that challenges me everyday, a job that I have wanted and now I have.  Not a job but a career, how many people fresh out of college get their dream job?  Not many but I was among the lucky few, so why am I unhappy.   I want more with my life, but what is that \"more\" that I want?  I can\'t put my finger on it and it is killing me.  I compare myself to too many other people.  Somedays I look at things that I have done and the places I have traveled and I feel satisified and happy until I remember that others have traveled more than me and have seen more than I have seen and then I am back to being unhappy.  I have married friends, engaged friends, actively dating friends, and single friends.  I don\'t want to be married or engaged but I feel lonely without someone there to cuddle with or to laugh with.  I date even though it is hard at times but still its fun.  I would prefer something more permanent.  My mom has a social phobia, so does my sister.  I believe that I have this but to a lesser degree.  I get nervous when I think about going out alone.  I think everyone is laughing at me.  \"Look at the fat girl sitting alone,\" they all say in my head.  \"She is pathetic and had no friends.\"  My head tells me that I am full of shit for thinking this but still I sit in my apartment alone.  I have my own apartment, car, dog, mailbox key, phone and answering machine.  I have a coffee maker with lots of yummy coffee.  I can get up in the morning and make coffee naked if I want and sometimes do.  I love thinking about the things that I have, it makes me happy, but something is missing and I don\'t know what it is.  I have narrowed it down to loneliness, but in what form.    I have a friend that says that I am not lonely because I have my family nearby.  We have different meanings of the term lonely.  My family is nearby and his isn\'t.  He is lonely without them.  He is outgoing and fun and has lots of friends and girlfriends.  I am shy and somewhat of a introvert.  I don\'t have lots of friend or boyfriends.  I am lonely without that.  Most nights I just go to sleep and forget about it.  I don\'t know how much longer I will do this.  I need to get over this social phobia thing I have I guess.  Its killing me.   '),(530,3155970,'22,April,2004','2004-04-22 00:00:00','   I don\'t like people, actually I hate 90% of the people that I come in contact with.  So is it horrible to hate most people?  No, and I will tell you why.    Everyone has issues, whether it is because your father abused your mother when you were younger, or your mom abandoned you as a child, you are stuck in a career that makes you feel empty, or you married for money.  Whatever the reason you, me, the guy next door, and the chick banging dudes alley we all have issues.  How you decide to handle the issues you have will determine whether or not you deserve to be hated by me or anyone else in the world.  So you are stuck in a shit job so you are rude and nasty and lazy all day at work.  I hate you.    You lost 100 pounds by starving yourself and over exercising but you still feel fat on the inside and all day long you beat yourself up over how horrible you look and how you will never be beautiful.  I hate you.  Sometimes I hate for silly superficial reasons but that is only after the person has opened their mouth and said something stupid.  And I bet you are wondering now, who don\'t you hate?   Well I don\'t hate intelligent people, so there you go.   '),(531,3155970,'18,April,2004','2004-04-18 00:00:00','   I need sleep.  I just got home from work and I am tired.  Work as in my part-time job at a coffeehouse is DC.  All the coffee I drank isn\'t going to keep me awake at this point.  Some man came into the coffee shop and gave me and my co-worker Kasey some advice.  \"Never get married and have lots of anonymous sex with bike messengers.\"  I can\'t make this stuff up I swear.  Lots of weird people come in and out of the coffeehouse and say the strangest things, it makes work more fun, that and all the free coffee.  So now I sleep while watching You\'ve Got Mail for the 100th time and maybe later I will workout OR clean-up.  Not both and maybe neither.   '),(532,3155970,'16,April,2004','2004-04-16 00:00:00','   So I tried to get a blog up and running at Tripod but I swear to goodness Tripod is on crack.  How could this be you ask?  How the hell should I know, I didn\'t get it hooked!!!  When I grow up I want to become a chain smoking alcoholic.  Sure its not the best life to live but that is what I want so kiss off you judgemental fucks.   '),(533,3155970,'26,May,2004','2004-05-26 00:00:00','   What do you do when you don\'t know what to do?  When you don\'t know whether to eat or not?  To exercise or not?  To go out or not?  To stay home or not?   How do you know if you are giving the right answers for your questions or asking the right questions to be answered?    Tonight is another night where I don\'t know what to do because doing something will result in something being done or a need being satisfied and that is just to overwhelming for me.  So many choices that I don\'t know which to pick and if I don\'t pick I will just end up sitting here all night and nothing will get done and I will kick myself.  So if not picking a choice results in nothing getting done because having too many choices is so overwhelming to pick then why is having only one choice so frustrating?  (Oh god I hate myself sometimes.)  I don\'t know what to do because anything I do will come out with the same result: I\'ll still be bored and talking to a dog.  I would so much rather being bored and talking to a human and hanging out with a human and making fun of other odd looking humans with another human.  Ahhhhhhh so I guess I need human friends!!  Well DUH I knew that all along.   I\'m too independent.  If I were more needy and pathetic then I would have more friends, or if the stupid friends that I have weren\'t so stupid then I would have someone to hang out with. Stupid fucking humans.   '),(534,3155970,'21,May,2004','2004-05-21 00:00:00','   I am going to buy a gigantic bag of cheetos and eat the entire thing.   '),(535,3155970,'15,May,2004','2004-05-15 00:00:00','   I have decided to start running again.    But you are so lazy how can this be???   Well I would like to lose some weight and this is the best and quickest way to do so.   Yeah you are pretty fat.   Fuck you.  I used to run 3 miles 3 times a week.  Sure to those psycho marathon runners that is nothing but for me that is something.   Whatever.   Lately I have been working out to a work out video and its been pretty cool.  I have all these muscles that no one can see because they are buried under layers of fat.  Of course I would like to see the fruits of my labor so I will run off all the fat.  I bought a cute stop watch to time myself when running and eventually I will buy myself some new running clothes so that I will look cute while running (that is very important). Also exercise is suppose to make you happy and I guess that is a plus seeing as how I tend to...sometimes...on occassion...not be so happy.    So in conclusion, I am fat, will run to lose fat and gain happiness through the release of neurotransmitters. :)   '),(536,3155970,'11,May,2004','2004-05-11 00:00:00','   I have no interests in anything.  I care about nothing.  Right now my only interests are eating and drinking but I\'m not even interested enough to go to the store to buy food and booze.  I exercise because its what I\'m suppose to do.  I go to work because I have no choice.  I pay bills for fear of what will happen if I don\'t.  I walk the dog because if I don\'t she\'ll piss on my clothes.  However I do put a lot of effort into budgeting my finances.  I shouldn\'t even be able to afford to live where I do but with the right kind tweaking I get by just fine AND still have money to buy my booze and food when I\'m not being lazy and depressed on the bed.  Sometimes I think life would be simpler if I had a sugar daddy as long as I didn\'t have to kiss him or sleep with him.  If all I had to do was to be his arm candy for a few nights a week it would be perfect.  But eventually I am sure that he would pop a few Viagra and want to do it so or else I wouldn\'t get my money.  To avoid that senario I would rather just not involve myself with that sort of lifestyle.  So I am back to being disinterested with life, at least for the rest of this week.   '),(537,3155970,'03,May,2004','2004-05-03 00:00:00','   So on Sundays I work in a \"gourmet\" coffeeshop preparing over priced coffee for yuppies in gang infested neighborhood in DC (Its called gentrification but we\'ll save that for another blog).  Its amazing what how addictive coffee is and how those of us that are addictive (yeah that includes me) feel we can\'t begin our day without our morning cup of personality.  However myself and my bank account have evolved to reject the other priced coffee drinks because they are a waste of money but not to these yuppies.  And in addition to taking their $4 for an espresso drink my colleagues and I also have the nerve to set out a tip cup to further bleed our customers dry.  I feel bad taking charging a ton of money for coffee but do I feel bad about taking thier extra change as a tip OF COURSE NOT!!!!  My first day I got $21 in tips.  The next day I only got $14.  Why such a decrease in my earnings???  Same number of people came in, they all bought the same drinks as the weekend before, so why less money.  The next week was the same, $14, and the week after, $14.  WHERE IS MY EXTRA $7????  I GOT A DOG TO FEED HERE PEOPLE!!!!!  And then it came to me.  The proverbial lightbulb went off in my head.  My first day I wore a tight shirt and smiled a lot, and the following weeks I wore loose fitting clothes, I still smiled but not as much.  So I decided to do what any money strapped woman with a dog to feed and DD sized breasts would do, show cleavage.  I wore a low-cut shirt and stuck my chest up for all the customers to see and it worked great.  Guys were smiling at me and shoving money in the tip cup like there was no tomorrow, even the women customers were digging in their wallets for change.  My tips for yesterday...$16.50, not great but up from $14.  The only problem was that I wasn\'t at the cash register the entire time, part of the time I had to make the espresso drinks which made me disconnected from my customers and them disconnected from my breasts.  Also the girl that I was working with was very flat chested and spacey, customers don\'t like that.   So what are the key points here???  When working for tips you must A) smile and have a cheery attitude B) show some cleavage and C) don\'t be a space cadet.  End of lesson.   '),(538,3155970,'26,June,2004','2004-06-26 00:00:00','   I think that I have a problem, an addiction you might say to romantic comedies.  I want to punch myself over this.  Every weekend I pop in one of my DVDs and watch these overly, unrealistic movies with perfect couples saying the perfect thing to win each other over. One might think that I am living in a fantasy land where I hope that one day a man will sweep me off my feet with the perfect line, gift, or with his quirky behavior.  Ugh I am not that naive, I just like seeing people happy and laughing.  All my DVDs really are comedies but I am partial to the romantic ones.   '),(539,3155970,'22,June,2004','2004-06-22 00:00:00','   Tarvor decided during his trip to Vegas that life was no longer worth living so he jumped to his death. Luckily someone had the forethought to tie a rope to his feet just in case something like this was going to happen.    urlLink Tarvor Jumps   '),(540,3155970,'18,June,2004','2004-06-18 00:00:00','      '),(541,3155970,'18,June,2004','2004-06-18 00:00:00','   The resume is being sent out and in my head the going away party is happening as we speak.  I\'m looking for a new job.    Apparently the powers that be here at my place of employment want me to be underpaid and overqualified babysitter for a 60-year-old idiot woman.  I have put up with a lot at this job.  I have given up weekends, come everyday for a month, stayed late, deal with being paid less than people that I do much better work than, and work under the most annoying, stupid, and dim-witted man on the planet and now this.  I will not rest until I have found another job.  I don\'t care that I will be losing this beautiful notebook, which I have grown very fond of, nor do I care that I will be leaving behind friends that I have made and enjoy laughing with.  Somethings are more important like my sanity and my bank account.  Boys and girls of the world hear me now.  Never let someone tell you that you do not have a choice.  You always have a choice.  I could choose to stay at my job and work with these miserable excuses for human beings or I can choose to take on new challenges and meet new miserable excuses for human beings.  Right now all I know is that any movement is better than no movement at all.   '),(542,3155970,'15,June,2004','2004-06-15 00:00:00','   I\'m at work and I\'m tired. My hair is soft and smells great. I\'m totally bored. My apartment is messy. I think that I will go out and drink margaritas tonight. I have to read a philosophy paper. I want to start knitting clothes. I sleep in a drawer.   '),(543,3155970,'08,June,2004','2004-06-08 00:00:00','   Have fun today, [GIRL]. Put your worries aside and live this day as if it were your last on Earth. Keep a light and cheery outlook on life and don\'t dwell on minor mishaps. This is a day to enjoy the sunshine and keep smiling no matter what circumstances may arise. You will find that your good mood is contagious, and that pretty soon you will be the pied piper of good cheer and a positive attitude.   This is the dumbass horoscope that I received in my e-mail today.  Perfect day for stupid ass advice seeing as how I am back to hating everyone I come in contact with, especially men...wait let me correct myself...BOYS!  Every man is a boy and every boy is just a smaller, younger, and stupider version of what he will grow up to be.  Sometimes I think that I would be better off if I were a lesbian but then I remember that women are over-emotional crybabies and drive people crazy.  The real problem is is that men think that they are right in their dumbass thinking and so do women.  We are both right and need to find a common ground with each other.  If you don\'t want to compromise your thinking at all then don\'t fucking talk to me because I am going to stop listening after you express your first narrow-minded, dumbfuck thought.  What might be surprising to all is that I am not the sort of person that is going to tell you after the first of talking with you exactly how I feel about you, I just won\'t talk to you anymore.  Why should I waste my breath explaining why I don\'t like you when I could just walk away and pretend as if I don\'t know?  The latter is less aggrevating for myself.  Let someone else beat themselves over the head trying to knock sense into yours.  So as of this moment I am through with people.  You are all stupid, self-centered fucks with bad taste and horrible table manners.   '),(544,3155970,'03,June,2004','2004-06-03 00:00:00','   choice   (n)  - 1.The act of choosing; selection. 2.The power, right, or liberty to choose; option.   Having a choice suggests that there is an opportunity or privilege of choosing freely.  We as humans make choices everyday whether they be bad or good, right or wrong, a choice is a choice.  Last night I made a choice to take a walk with my dog to Starbucks and buy a fattening yet tasty drink.  On the way back an opossum crossed our path.  I\'m not sure if it saw us or not but I did call to it not to walk into the street and when it did WHAM! it was hit by a car and then a second car.  I stood there watching it struggle for life for about a minute then it died.  For some reason I felt the urge to cry but I didn\'t because it was a stupid rodent that made a choice to walk out into the middle of the street and get hit by a car, or did it?  Seeing the blood ooze from the mouth of this larger and uglier version of a rat made me glad that I walk my dog on a leash so that something like that won\'t happen to her because animals don\'t seem to make logical choices they just seem to act on instinct.  But maybe I caused the opossum to run into the street.  Could it have made the choice to either a) face an enemy (me and my dog) or b) run away in the other direction?  Maybe if I didn\'t go to Starbucks last night the opossum would still be alive now and having opossum-style sex with a mate.  But alas the creature is dead and life goes on and in the world of streets with big cars opossums cannot compete but this is where Darwin steps in with his theories.  Dumb, slow moving opossums will be killed by cars but smart and quick opossums won\'t and so in years to come downtown Alexandria will be full of the faster smarter breed.  Or the tiny opossums with thin, weak skulls will die but the bigger ones with thick skulls will hold up to the weight of cars and soon the species will the size of big dogs.  All I know is that because I was struggling with the choice to either pay my rent increase or move, caused me to choose talk on the phone all night instead or make dinner, which then caused me to make a choice to walk to Starbucks with the dog to buy a fattening drink, which caused the opossum to make a choice to walk in the street and get hit by a car.   '),(545,3155970,'01,June,2004','2004-06-01 00:00:00','   I\'m going back to school!!!!  Its not grad school but its a step to grad school.  At the end of June I will be taking Calculus because I need the class in order to go back to school.    I am such a nerd.  I had to take a placement test at the Community College I will be attending to see what math level I could take and naturally I placed in all levels so I can take anything I want.  (Save the ooohs and ahhhs I know that I am smart).  I\'m so excited about about going to class and studying but what I am most excited about is buying school supplies.  I LOVE school supplies!!!!  But what is even better than buying my own school supplies is stealing them from work!!!!!!  Pens, pencils, highlighter, stick pads, 3-ring binders, paper, notebooks, oh god I am getting sooooooooo excited.  I need a cool drink.    I\'m sure that after the first week of class I will be crying and complaining about how I wish I was bored again with nothing to do after work, but for now I will be excited and drinking cool drinks until then.   '),(546,3155970,'23,July,2004','2004-07-23 00:00:00','   I have but a few moments to write a bit about my life for the past couple of weeks in a short paragraph.&nbsp; I wanted to jot down my feelings about people.&nbsp; People are stupid, especially when those people are your friends.&nbsp; Your \"friends\" (I put friends in quotation marks because if these people were really my friends then they wouldn\'t act like jerks) expect you to always be there for them when their car dies, when they need a couple of bucks, when they want to go to the movies etc.&nbsp; Or when they are bored at work they call you expecting you to answer because you are always so happy to be their interest when they have nothing else to.&nbsp; They don\'t call you when they have free time to romp and frolick.&nbsp; They don\'t call you when they know that you are on your last dollar and really bummed out about your shit job and could use a laugh.&nbsp; No these people only think about themselves and what you can do or can\'t do for them and they call themselves your \"friend.\" So instead of sitting around the house waiting to be someone\'s last resort you dust off your dancing shoes and hit the town, ALONE.&nbsp; You make new friends, discover new interest and make your own plans.&nbsp; Then your \"friends\" start to call you again since now their new friends are boring to them and they want to return to good \'ol reliable you, but now you have a life and are too busy having fun to be bothered by someone who didn\'t have the time to be bothered with you.&nbsp;  So what does your \"friend\" do?&nbsp; Accuse you of being the type of person that would ignore their friends for some guy, which is what that person has done to you.&nbsp; Well with friends like that who needs them at all.   '),(547,3155970,'07,July,2004','2004-07-07 00:00:00','   I\'m tired My belly is full I need a vacation Shoot me in the head   '),(548,3155970,'01,July,2004','2004-07-01 00:00:00','   Today will be my second day of calculus class.  I was so happy last night as I did my math homework.  Its been awhile since I have sat down in front on the TV to do math and it felt great.  Working out problems with my little pencil and my little calculator.  I felt at peace with the world. Last night I bought some yarn and some ribbon so that I can knit my new project (I just completed a furry scarf) and when I got home I had received my knitting needles from ebay, but I didn\'t dare knit a single stitch because I had calculus homework and school comes first!!!!!  I LOVE SCHOOL!!!!!  I want to be the best student ever!!!!!  I will crush my classmates by receiving perfect on all exams and being the teacher\'s pet.  Laaa laaa dee da!!!   '),(549,1636406,'24,March,2004','2004-03-24 00:00:00','    Happy Birthday  urlLink OS X , I hope we get to enjoy many more ;)   '),(550,1636406,'14,March,2004','2004-03-14 00:00:00','    That\'s about all I have to say for today.  Tomorrow I have Japanese, I\'m SO unexcited about the whole deal-- let\'s hope I feel good tomorrow and I can stay awake.  So um... yeah that\'s about it--   ...OH!  I watched  urlLink Party Monster  last night, and I have to admit I really liked it.  It was completely dystopian, but so is life.  I wish my life was more exciting these days-- perhaps that\'s what\'s killing my motivation.  So much for the weekend.  -Ry   '),(551,1636406,'13,March,2004','2004-03-13 00:00:00','    Update: The Tickets were sold out!!! so I saw the dead sea scrolls instead (BORING!).  I wanted to see Taiko badly, but life enjoys making things unfair.     The Dead Sea Scrolls were so boring, there was no translation available, and it just was... pointless.  Some dirty pieces of paper under some dim strobing lights-- those dirty pieces of paper are so ridiculously pointless.  I\'m sorry to say that to all the religious nuts-- but think of all the people suffering around the world, then thinking of all the money going into preserving and showcasing that nonsense.  Even if that money went to giving 3 people fresh water it would be worth the loss.  So my 2 cents again- but it\'s my weblog :)   '),(552,1636406,'12,March,2004','2004-03-12 00:00:00','    Today I\'m hoping to see the  urlLink Kiyoshi Nagata Ensemble.       I\'m actually pretty excited about the whole thing-- while I\'m there I\'m hoping to see the dead sea scrolls which are showcased in the same building.  I will post more later after I\'ve seen the performance.   -Ry   '),(553,1636406,'12,March,2004','2004-03-12 00:00:00','    Let me just start off by saying: I hate  urlLink Old Navy .  The clothes may be poor quality, and often ugly- the stores might have the atmosphere of a Chinese sweat-shop, but my real problem with Old Navy is the television advertisements blaring 24 hours a day.        An orgy of colour, beauty, and retardation   These ads seem to be formulaic.  You take a couple beautiful people, video tape them doing stupid things and make them smile.  You then throw in some weird progression of events and add a few unintelligible lines.  Worst of all Old Navy could probably show a video of a child cracking it\'s head off pavement and they would probably still see sales of their slave labour wares (yes, little hands do make little stitches).    Every-time I see one of these ads I change the channel- they are eerie, they make me loose faith in humanity-- and I think could even make me suicidal if my exposure was increased.  Old Navy commercials are pornography for christian-american-hillbillies, they project images of american stupefaction- and of the homogenization of culture in America.  There is also some real racial stereotyping going on-- you have to appeal to every \"visible minority\" and the formula for that is: 1 mulatto person, 1 asian person, with a variety of white people to interact with.  So to sum it up.... I hate these fucking ads!   '),(554,1636406,'12,March,2004','2004-03-12 00:00:00','    Welcome to my new Weblog. I hope you enjoy it.  I should be posting on a regular basis.  Today I had a  urlLink Japanese  test which I totally failed-- I was half awake, feeling like crap and I didn\'t have enough time to finish it.  It was horrible. Saturday, there is a  urlLink Taiko  performance at the museum of civilization.  I hope I get to see it- I really enjoy the sound.  Post more soon,  -Ryan   '),(555,1636406,'25,April,2004','2004-04-25 00:00:00','    urlLink Gmail  is  urlLink Google\'s  new web-based e-mail service.  It offers 1 gig of storage, junk-mail filters and a few other perks-- it\'s really nice!  Thank-you Google!  The service so far is great, I hope someone fixes the Safari compatibility issues and I look forward to using this service more frequently.  Thanks!   '),(556,1636406,'17,April,2004','2004-04-17 00:00:00','    some believe in the year one double odd three...     I saw  urlLink Kill Bill v.2  last night.  It was great, I really enjoyed it-- despite the fact a bunch of losers cackled at nearly everything shown on screen, including music, and a ski mask.   At one point during the movie I almost felt as if one of the characters was talking to me indirectly, it was kinda strange.  I can see that some people will have a hard time seeing the two halves as a whole, and I think that is an obstacle for this movie-- the one plus is that volume 1 was recently released, so my suggestion is: go rent or buy volume 1 before you see volume 2, you\'ll appreciate it more.  It was a great movie, I\'d recommend it for sure-- I look forward to seeing it again in the future.   Thanks Quentin!   '),(557,1636406,'16,April,2004','2004-04-16 00:00:00','    America needs a regime change, such an evil, backwards place.  What ever happened to liberty? freedom?  It\'s one thing to sell your soul for something real- but America has sold it\'s soul for money--   '),(558,1636406,'12,April,2004','2004-04-12 00:00:00','    Missy Elliot is so awesome, and has really had a positive impact on hip hop.  I love Missy, I think she is the greatest.  yeah.. this is a bit of a sad update to my blog, but missy is just too kakoii to ignore.   '),(559,1636406,'03,April,2004','2004-04-03 00:00:00','    I\'ve done a lot of thinking and I\'ve decided I don\'t want to come back to Ottawa.  What I have decided though is to try to get into  urlLink McMaster.   I\'ve decided that I want to make Japanese my major, so that I am able to give it the attention and effort it demands, as well as gaining a skill that will last my whole life.  I really think this had to happen, I just hope that I am able to get in and that I can continue with my studies instead of being out of the loop for a couple years.  Wish me luck!   '),(560,1636406,'27,May,2004','2004-05-27 00:00:00','    Today I practiced the devil\'s tounge...  And I got through about 110 kanji and about 440-500 compounds. Mostly new compounds too- but none of that set in.  It was hard sitting in the heat figuring out how to pronounce compounds I\'ve never seen, plus a few kanji I never knew... like bugs.  so now I\'m about ready to hit the reset button with my trusty ice pick- my brain is too full, and I think I\'m risking seizures looking at all these black little lines.  So tip of the day: Don\'t learn Kanji in the sun!   '),(561,1636406,'26,May,2004','2004-05-26 00:00:00','    So big bad Paul called an election for June... joy  This election is the worst election I\'ve been involved with. You have Paul Martin, an out of touch, quasi-liberal wannabe conservative. You have Steven Harper, a deranged, quasi-conservative, right wing extremist-nut. Lastly you have Jack Layton, a true NDP leader- completely out of touch with the sad realities of the country.  All of the men running for prime-minister are pathetic. The NDP has a great platform, but how stupid do you think people are? You cannot promise the world and have it all work out magically.  Sadly you can\'t raise corporate taxes without a serious negative effect.   Steven Harper is the worst of the three, he appeals to the intellectually disadvantaged by spreading fear, anger and lies.   Paul Martin is pompous and completely out of touch with reality, but he is the least harmless of the big three.  In all honesty I want to vote for the NDP... but we\'ll see.  I may vote Liberal, It\'s fairly confusing.  I will write more as the stupidity develop.   '),(562,1636406,'25,May,2004','2004-05-25 00:00:00','    You Are....          A Swan!!      This was the show of the century for me. A bunch of fat, depressed, confused housewives give their bodies over to the  urlLink Fox Broadcasting Company  in order to be hacked up, abused, and caked with hooker makeup in order to compete with each other for some crappy-ass prizes!  It is sheer marketing genius.  There is nothing more enjoyable than seeing stupid people do very stupid things.  Now for my little rant: The finale was so rigged it wasn\'t even funny- it was stinkingly obvious. I knew Rachel would win after seeing the first episode- why? Because she was the prettiest so they put her on first to get people to watch the show. Rachel was also much taller than all the other girls, which for whatever reason goes far towards winning a modeling contract.  So the show was rigged! A shock to everyone I\'m sure. I am keeping my fingers crossed that next seasons girls are fatter and uglier and require even more hacking and slashing.  Hell I have an even better idea: take a bunch of men that want to be women, give them a sex change and hormone therapy, then throw them in a beauty contest facing them off against The Swans... that would be my ultimate show.  FOX Execs if you read this and you like the idea give me a call-- I\'d make a great producer.   '),(563,1636406,'24,May,2004','2004-05-24 00:00:00','    Supersize Me   Almost everyone I\'ve talked to has at least heard about this movie.  It\'s a Michael Mooresque--sensationalist documentary, and as enjoyable as it was I felt this movie was lacking a little.  Don\'t get me wrong, I really did enjoy the movie. I find the topic to be fascinating, and I have to give kudos to the star/director, Morgan Spurlock, for coming up with the ingenious idea of capitalizing off one of the strangest aspects of modern culture. Despite the topic, the movie lacked the speed and coherence of most of Moore\'s documentaries and I think that as his debut movie Spurlock almost nearly hits the nail on the head.  Yes, this movie does have a message: eat mcdonalds and you\'ll die-- but it lacks the planning, and information to really get the point across. The movie ends up being a 2 hour circus show-- with some white guy scarfing and puking up quarter pounders. It disappointed me because Spurlock ends up meeting all these really interesting characters like Don Gorske, a John Lennon look-alike who is addicted to Big Macs due to his Obsessive Compulsive disorder. The most interesting thing about Gorske is he is in great shape despite the 19,500 Big Mac Attack.  You also meet some insanely overweight guy who is about to get his tummy tucked- and in some amazing cinematography you get to see the procedure accomplished. The problem is Spurlock goes nowhere with these interesting, and bizarre people. You don\'t get to see how Gorske lives- and let me tell you where there is smoke there is fire... I am willing to bet Gorkse\'s house was insane, talk to his wife for 5 minutes and I bet you\'ll hear some good stories. The tummy tuck guy dissaperes after the surgery- you don\'t see or hear from him again- what would have been interesting is if Spurlock came back to see the man and offered him a supersized McDonalds Combo-- just to see how much of it the guy could eat. What would have been even better is if Spurlock took all his doctors and some of these weird people in a bus, and drove down to McDonalds headquarters to not only make a scene,  but also to announce the results of Spurlock\'s evil diet.   Ultimately the movie was boring, there were too many scenes showing him going to the doctor, and too many scenes of him doing nothing. The movie didnt have enough weight behind it, it deviates off topic a bit, and the interesting characters are not put to good use. I suggest people watch the movie, but don\'t expect much-- the movie falls short of being a classic, but that doesn\'t mean it\'s not enjoyable. So cheers to Spurlock, the movie may have taken 10 years off your life, but I\'m sure it was worth it.   '),(564,1636406,'24,May,2004','2004-05-24 00:00:00','    Queen Victoria   '),(565,1636406,'23,May,2004','2004-05-23 00:00:00','    Not one god-damned thing happened.   '),(566,1636406,'19,May,2004','2004-05-19 00:00:00','    So I got this job, and get this it pays   $$$8.75$$$  . Wow, now I know how benjamin franklin felt when he received his pay- it\'s just too bad I\'m not living in the 1700\'s.  To be honest I am going to keep looking for other jobs where employers understand the concept of inflation, and fair pay.  I also have been thinking about friendship a lot, and trying to figure out why it\'s so hard for me to meet people I get along with.  So again I\'m tinkering with my life so I can increase happiness, and decrease the bullshit.  I guess things could get worse, and no doubt they will.  I just hope I can get into McMaster.   '),(567,1636406,'18,May,2004','2004-05-18 00:00:00','    Tarot: XVI - The Tower   '),(568,1636406,'15,May,2004','2004-05-15 00:00:00','    I hate these things, but I have one on Monday.  I\'m pretty excited.    Wish me luck!   '),(569,1636406,'14,May,2004','2004-05-14 00:00:00','    I was thinking about the little creatures today, and realized that Mosquitoes are one of natures most benevolent creations.  They don\'t survive on dead organisms, and in order for them to live they require that their meals are living.  The males drink the nectar of plants to survive, while the females drink the blood of mammals in order to gain enough energy to lay eggs. Both of their meals exist outside of the insect realm, they do not eat other insects at all- which is interesting when you start to think of the diets of other creatures on the planet.  Mosquitoes are very successful at what they do- although interestingly enough their only real enemies are the elements and human beings.   Why is it the most powerful creatures on earth hate these little bugs so much? Because the allergic reaction to their bites are inconvenient and in the past Mosquitoes have been a source of disease.  The most interesting thing is that in order to kill mosquitoes people use really nasty chemicals, which in the long run could harm people more. Personally I would rather take my chances with a mosquito bite then to be inhaling or absorbing some toxic chemical thats bound to give me cancer.  So in a strange way, I\'ve found a respect for the biology of the mosquito, and although they lead very simple lives- they are very kind to the world around them.   '),(570,1636406,'12,May,2004','2004-05-12 00:00:00','    ...I will have no money in 2004!   urlLink      Video Games were once a passion of mine. I saw them as the purest form of art with so much potential. But the Internet boom happened, and games were a casualty. The Industry started to receive attention, reports circulated on the profitability of games,  urlLink Mario 64  selling in excess of 10 million cartridges- pulling in way more than any hollywood blockbuster.  In fact the Video Game industry was everything the movie industry wanted to be: consoles changed every 3-5 years, there were competing standards that many families welcomed, and of course there were rumors of price fixing by the bigger companies like  urlLink Sony  and  urlLink Nintendo --the DVD business model was ripped right from the Video Game industry. The cow that broke the camel\'s back was  urlLink Microsoft  getting into the biz- first by working with  urlLink Sega  on their  urlLink Dreamcast , and eventually by releasing their  urlLink Xbox.  The mentality for any market it seems is- if it\'s good enough for  urlLink Bill Gates,  it\'s good enough for me to invest in.   When Sony entered the market in  urlLink 1994  things changed-- the old rules didn\'t quite fit there was heavy competition, and since then it has been maturing at a growing pace. Everyday there are more poor quality games hitting the market and many more movie tie-ins (which ultimately are not as bad as their forefathers see  urlLink ET ).  Nintendo reigned supreme since the \'70\'s but was ousted by Sony. Nintendo survived by taking the stance they didn\'t care if they were number #1, they just planned to always turn a profit.  The strategy has paid off- Nintendo products may not be on the bleeding edge, but they always make a nice profit- unlike Microsoft who has lost a  urlLink fortune  over their Xbox.  So... Needless to say, seeing an industry grow up before my eyes was fun, I just had lost interest...  Until E3 of this year. The games coming from Nintendo  urlLink look great.   urlLink Zelda,   urlLink Metroid,   urlLink Mario,   urlLink Starfox  and most of all the  urlLink Nintendo DS  have really sparked my interests again-- the games are starting to get better, they do look good but may end up being total duds like most of the games i\'ve played recently. (I\'m keeping my fingers crossed).  I will buy these games, and a  urlLink Nintendo DS -- and I will probably love them. Sony now has its very own handheld called the  urlLink PSP - but I sneaking suspicion it will be overpriced and will not be worth the money.  I have a PS2, and I have to admit it isn\'t half as much fun as my GameCube.  So Kudos go to  urlLink Nintendo - you\'ve survived  urlLink 115 years  and I suspect you\'ll be around for at least 115 more.   '),(571,1636406,'11,May,2004','2004-05-11 00:00:00','    ...two slugs trying to meet each other half way on hot pavement.   '),(572,1636406,'10,May,2004','2004-05-10 00:00:00','    My parents are crazy.  I have trouble figuring out just what is wrong with them.  Today I was actually writing a post about improving transportation problems in the world but I was sidetracked, by a rabid, menopausal, bitch of a mother who is on my case because I do not have a job. Whenever the economy is doing poorly it seems like it\'s my fault I can never find work. My resume is decent, I have a high-school  urlLink diploma  and a college  urlLink diploma  and two years of  urlLink university  on paper.  Finding a job is easy- I could work in any fast food joint instantly, or some miserable telemarketing job, so I try to hold out, I try to apply to different jobs, but because of insane pressure from two unstable adults who try to exert control over me as much as they can, I end up getting some ridiculous job that ends up making me unhappy.  Do they care?  No, I\'ve learnt that no matter what I do they benefit in the long run-- and the only reason they are placing pressure on me is because they are miserable unreasonable people.  I can\'t take this nonsense from them anymore, it makes me unproductive, and I lose focus. I am going to keep applying and hopefully I manage to find a good job.  wish me luck   '),(573,1636406,'10,May,2004','2004-05-10 00:00:00','    I\'ve decided that along with the updated design I shall also update what I write, and the way I write. Over time I will offer insight into my mind, how I feel, and provide opinions on issues I believe are important to me. Also...    Every Monday I will write a Haiku. For those of you who do not know what Haiku is I offer you a  urlLink primer.   Here is my Haiku for today:    Jogging    in the heat of sun I find myself seeking truth as life passes by  -Ry   '),(574,1636406,'03,May,2004','2004-05-03 00:00:00','    I hope you like the new web log style.   I Love It!!   I will be posting more often now,  so enjoy!   '),(575,1636406,'03,May,2004','2004-05-03 00:00:00','    Today I\'m going to go nuts on Kanji, see if I can go through about 200 today.  I\'m keeping my fingers crossed!   '),(576,1636406,'27,June,2004','2004-06-27 00:00:00','    I have decided to go back to  urlLink Carleton University  come September. I am looking forward to school, and I think that it is going to be a great year. I have learnt from my mistakes which were:  1. Roomates 2. Lack of transportation 3. some lack of confidence  Like anyone else I am entitled to my mistakes, but thankfully I have been granted the luxury to try again and to do things right.  It is anyone\'s guess how things will unfold this year- I believe I will succeed, and become a better person for it.  It has been a crazy weekend, and my head is still spinning, I have learned a great deal- and really gone through a lot of emotions. I am nervous and excited about the opportunities that lay before me, we shall see how things turn out.   '),(577,1636406,'22,June,2004','2004-06-22 00:00:00','    Red Bull and Vodka = Mana from heaven   '),(578,1636406,'21,June,2004','2004-06-21 00:00:00','    Right now I am torn up inside...  ... do I want to go back to Carleton? To the city I hate so much? Does everyone(thing) deserve more than one chance? Do I try to start over? or do I try to correct a mistake?  I don\'t know what to do...  I really loved some things about Ottawa, and hated others. I just am confused as to what I am going to do next year. The most certain thing is that I am willing to try again- and this time try 100 times harder, not to let myself fall into the same mistakes that plagued me before.  we will see what fate has planned, all I know is the stresses and problems with last year gave me grey hairs... I am not going to let that happened again in a million.   '),(579,1636406,'17,June,2004','2004-06-17 00:00:00','    3 million years and this is the best evolution came up with...   '),(580,1636406,'15,June,2004','2004-06-15 00:00:00','    I just watched the political debate and was so excited I had to run online... and well since I don\'t have anyone to talk to about the outcome, I am left with no choice but to talk to myself.  So first a summary, the election is charged with a handful of hot yet mostly irrelevant issues only because they are debatable and make good news. First we have the sponsorship scandal-- Let me sum this up by saying it is over-blown nonsense. 100 million dollars was spent on a sponsorship program, it is merely some pork barreling that went to liberal friendly agencies. So they spent the equivalent of 748 thousand us dollars-  big deal more money is spent in iraq in week. The second hot topic is privatization of health care. The only reason this is an issue is because of the popularity of the conservative party that will introduce private health care elements to our economy. There was also talk of other issues like gay marriage, funding for cities- but all were really minor points.  The majority of the debate consisted of Paul Martin getting picked on by Jack Layton, with Stephen Harper throwing in his 2 slimy cents and Duceppe asking a few questions. Layton was the most eloquent, and really came off as Trudeauish, and by that I mean fruity. Harper was mostly eloquent, but didn\'t say much, he came off as weak (which I believe he is) and childish, if I had to place a bet he\'s either the youngest or only child in that family. Martin resembled a goat attempting to climbing a muddy hill, he was all over the place, out of control, arrogant, and very ineloquent. Duccepe shouldn\'t have even been there- he was such a waste of time but was probably considered due to cbcs marketing purposes towards the Quebec market. It would have been more interesting to have thrown in the Green Party- but the CBC wants it\'s little turn at trying to stir a big pot so they threw in some french onions.  So before the debate I was going to vote liberal, but now I am most likely going to vote NDP. Layton is the strongest, and best suited for the job, Martin is an out of touch old man who disgusts me to be honest- here is a man with all the tools of government at his disposal, and instead of showing them off like a kid with his daddie\'s gun he makes promises!? It\'s insane, he deserves to lose.   I think I\'ve decided to vote NDP. I don\'t care if Harper gets in- and if he lasts his 4 years without getting a bullet in his head it will be a miracle! Lest we forget the spartan army was comprised of homos- if Harper pisses them off who knows what they will do... and if he takes away the right to choose beware of pissed feminists.  Its going to be tough but I have about 13 days to think about it. Good luck to all those involved- I can\'t say I\'m very enthused to see these results... but this is politics at in its most interesting state.   '),(581,1636406,'13,June,2004','2004-06-13 00:00:00','    Last night I got to see my cousin Lisa aka \"Darling Starling\" play in her band the Flamingo-gos. It was my first time seeing her play, and I have to admit she was really good. She had a little bass solo that sounded pretty funky. I had a good time and although I didn\'t say much It was really great. I left kind-of early because I was tired, but they were finished playing anyway.  Right now I am reading Against All Enemies by Richard A. Clarke, and Sleeping with the Devil by Robert Baer. Both relate to one another and both are excellent in my opinion. I am looking forward to reading further into the oil industry. Oil is something seemingly innocuous- yet we fail to realize the entire modern world has been built with and on oil. The economy is oil, the cities are oil, technology is oil- everything you  see around you required oil at one stage of it\'s production, whether that was in it\'s transportation or an energy source during fabrication. About 290 kg of fossil fuels goes into the production of one desktop computer not including transportation.   Most people think the oil industry is all about cars- but it\'s so much more, everything relies on fossil fuels, that is why the economy goes crazy when oil prices fluctuate.   Many people think that it\'s simple to convert to a hydrogen fuel based economy- but its near impossible. And I read somewhere it is enough of a problem to bankrupt every government worldwide for a century.  Ironically, because it is rich in hydrocarbons, and the infrastructure is in place- the best substances to use for hydrogen based engines are fossil fuels.   '),(582,1636406,'06,June,2004','2004-06-06 00:00:00','    So I am now working as a bill collector for NCO Financial. It is a little easier than I expected, but its hard when you have to deal with some very scary, strange people.  I am nice to everyone on the phones, because to be honest, as Ryan, I don\'t care if you pay your bill- I just leave messages and give you a chance to set your credit history right. Cause let me tell you- I have debts too, and when I\'m working I do what they ask of me but I don\'t get nasty or over emotional because for what they pay me: it just isn\'t worth it.  So I\'ve been there for a week and all is average.  I also had a fantastic weekend, props to Jarod!  So I will try to keep posting, but work eats my day away.   '),(583,1636406,'29,July,2004','2004-07-29 00:00:00','    Finally! My iPod arrived- and it is amazing- better than I could have expected.  I am very impressed with my nice new 4g ipod.  I would suggest everyone get one.  Well thats all for now  Ry   '),(584,1636406,'22,July,2004','2004-07-22 00:00:00','    Okay, I will admit I have to be crazy to be thinking this... but I drove someone home from work today and this huge urban sprawl had appeared while I was in Ottawa. I really had some mixed reactions. I have only been gone for a year, and it\'s like a whole world appeared in what was once farm-land.  When I was little my grandparents would take me to feed horses there... it just was so strange to see that all disappear and be replaced with houses and roads. It\'s so depressing.  I\'m not sure if what I am feeling is natural or the result of a really tough day at work.  So lately all I do is work and rip cds-- My CD collection is so pathetically small I am scrounging to find cds to rip.  I can\'t wait till I go back to school.   '),(585,1636406,'20,July,2004','2004-07-20 00:00:00','    urlLink'),(586,1636406,'19,July,2004','2004-07-19 00:00:00','    So after this weekend I\'ve decided alcohol has been detrimental to my life, and because of that me and the bottle are breaking up for good.  I\'ve never been much of a drinker (except for perhaps last year), and after the killer hangover I had on cheap vodka this weekend- I don\'t think I will ever drink again.  I really do appreciate fine drinks, and I just think it\'s unfortunate that they have to be laced with liquid poison. I guess that is part of the excitement of any drink... When you combine fine ingredients to make a tasty yet deadly drink it makes you feel good inside.  It\'s why people make shots with tequila and tabasco sauce, it\'s why people spend $300 for a bottle of wine. Sadly I will take a risk and live without CH3CH2OH floating through my brain.  Again, it may not be the perfect solution, but I believe it is a step in the right direction.  Perhaps if you\'re a casual drinker you might take a similar stance.   '),(587,1636406,'16,July,2004','2004-07-16 00:00:00','    So yesterday my registration package arrived, and today the courses were uploaded to Carleton Central.   So far everything looks good for a fairly clean schedule... that is if I can get into the courses I want to.  I have something like 34 days left at work!   woohoo!   '),(588,1636406,'14,July,2004','2004-07-14 00:00:00','    So last night I was laying in my bed thinking about life when I decided to open a book I had bought called   urlLink Fast Food Nation   by Eric Schlosser. I opened it up and just started to read, by chance I opened the book to the chapter on Mr. Schlosser\'s trip to a Slaughter-House-- The only way I can describe it is that it was shocking, disturbing, and more than enough to repulse me from eating meat for the rest of my days.  So I am a vegetarian, sort-of... I\'ve decided I\'m still going to eat fish occasionally- I don\'t think I could give up sushi.  Now I don\'t think it is hypocritical, I am a \"vegetarian\" because I don\'t agree with the practices of the meat industry, the fishing industry isn\'t great but they still haven\'t come up with a way to completely torture the lives of fish.  I think this is a step in the right direction.   '),(589,1636406,'12,July,2004','2004-07-12 00:00:00','    urlLink'),(590,1636406,'05,July,2004','2004-07-05 00:00:00','       '),(591,1636406,'03,July,2004','2004-07-03 00:00:00','    urlLink   '),(592,1636406,'02,July,2004','2004-07-02 00:00:00','    So I wrote this eloquent, long post yesterday... only to close the window by mistake and lose everything I had done (stupid tabs)...  I\'m not going to get into what I wrote though, I guess it\'s just between Safari and me.  Lately I\'ve been looking for an Apartment in Ottawa, it isn\'t easy- but it isn\'t that hard either.  I have my little eyes set on one place- it seems decent and isn\'t all that far from school. Next thing on the agenda is to get a laptop for school... it\'s expensive, but things are working out.  So now I just sit and wait (and try and save money) and study.  There just is not enough time in a day to accomplish everything I want to do.  So cheers to July!   '),(593,1636406,'07,August,2004','2004-08-07 00:00:00','    Ok I have to admit, MSN Messenger and my \"so called\" buddies are driving me insane.  First I need to calm down, take a deep breath and remember that this isn\'t 1996 anymore.  People have forgot how to communicate- Instant messaging has brought into the world a new era of asinine conversation and the ability for people to ignore you mid-sentence without having to endure the social pressures perfected over the course of our evolution as a species.  So I am thinking again that it is about time that I abandoned this cursed, monopoly driven technology and move to more staid forms of communication for example e-mail, and telephone.  I will mull over the details tonight, and perhaps compile a list of people that I would like to keep in touch with.  So if anyone wants to know why I no longer use MSN or AIM- it\'s because no one really ever chats with me...  bye bye IM- I wont miss you.   '),(594,2935790,'28,February,2004','2004-02-28 00:00:00','   Hey, it\'s Sean,     I just, uh, decided that to start of a blog with comedic roots, it would be cool if I mentioned a few.  Mitch Hedberg is the world\'s greatest comedian, and I\'m not even joking.  If you don\'t know him*, you should get to know him.  I\'m not going to go on much more about him because I can tell by the look on your face that blah blah blah here\'s a quote;  \"I have a feeling that a duck\'s opinion of me is strongly influenced by whether or not I have bread.\"  -Mitch  Until next time, Sean  * \"How many of you people know me?  Well, if you don\'t know me why are you coming to my special?\"  '),(595,2935790,'27,February,2004','2004-02-27 00:00:00','   Welcome! Testing out the comments bar.  '),(596,2935790,'27,March,2004','2004-03-27 00:00:00','   *Sigh*  I\'ve made some stuff. Here it is. For you. Enjoy.   urlLink Useless souvenir-type image of awesomeness.   urlLink A display picture.   urlLink Another display picture.  '),(597,2935790,'13,March,2004','2004-03-13 00:00:00','   Sean.......aaaagain.   Rosie O\'Donnell is a useless, ignorant, fat, useless, fat, ignorant, fat, useless, ignorant fat ass, with no use. She shouldn\'t be paid for anything she does, ever. Lucifer has a special seat to his right waiting for her, crafted with thousands of souls (and we\'re still not sure if she\'ll fit). The only thing bigger than Rosie is her big, useless ego. She\'s a stupid lesbian who\'s ruining the world with her fat cries for attention. If I owned a tabloid, I\'d publish \"Rosie O\'Donnell does something else while looking fat and hideous and STILL no one cares.\" Where are those crazed gunmen I\'m always hearing about? You guys should get off your asses, there\'s work to be done. Let\'s not remove the blame from our unamiable target. She\'s a fat ass. She\'s useless. No one likes her. Cry Rosie, cry prz! God bless her \'children\' for being around knives and ropes and still being alive today. Fatty. Rosie O\'Donnell does not deserve to have rights! GRR!  '),(598,2935790,'12,March,2004','2004-03-12 00:00:00','   Sean here.   I\'ve gathered some proof that the Careers course is stupid. I\'ll keep it short and sweet. Huzzah!      urlLink Healthful diet.    urlLink Trends don\'t come from a vacuum.    urlLink Words of wisedom.    urlLink The other side of the ping-pong table.    urlLink Smile.    urlLink Ahmed never learned about the proper use of conjunctions.      Nelson text books are made by 3-year-olds.  Case closed.  '),(599,2935790,'07,March,2004','2004-03-07 00:00:00','   Sean, yet again.  I heard about this new game Manhunt for PS2, and well every system I guess...It looks really really awesome, needless violence  and  somewhat deep gameplay.  I haven\'t played video games in a  looooong  time, but this is definately worth my time...I think.  Anyways, check it out at  urlLink Rockstar Games .  The Manhunt site drove me nuts, but I finished it!  On a completely unrelated note, I found this Personality Disorder test and it seems to be decent, although I\'m basing that on the fact that there are no \"did you ever poop your pants when you were little\" type questions.   urlLink Here\'s the test  if you wanna take it, and here are my results...I don\'t suggest believing these results at all, I mean, what are shrinks for if this \'amazing\' test could tell me what I am.       Disorder      Rating        urlLink Paranoid :    High       urlLink Schizoid :    Low       urlLink Schizotypal :    Moderate       urlLink Antisocial :    High       urlLink Borderline :    Low       urlLink Histrionic :    Very High       urlLink Narcissistic :    High       urlLink Avoidant :    High       urlLink Dependent :    Moderate       urlLink Obsessive-Compulsive :    High  '),(600,2935790,'06,March,2004','2004-03-06 00:00:00','   Sean, again.  Hope Dain\'s having an awesome trip :D.  I would just like to shoot down that stupid \"Test the Nation\" show. The \'on the spot\' questions with contestants are more fake than wrestling, and worse acting than \'orno. I answered every question within like, the first couple seconds, and I used the rest of the time to think about how stupid Americans are, sitting at home crappin\' themselves over the right answer. I\'ve taken real I.Q. tests, the kind you pay money for, the real kind, and some people portray the show as if it were a \'real I.Q. test\'. It\'s not, it\'s just not. Don\'t be gay, don\'t watch the show, unless you feel like barfing all over your TV.   Dissapointed in the human race,  Sean  '),(601,2935790,'05,March,2004','2004-03-05 00:00:00','   \'Tis Sean here.  A quick thought on the Careers/Civics course.   Civics  = important, depending on your future  career , which you don\'t need an  ENTIRE  half a semester to stress over.  Remember those people, down in the office?  The ones who you can make appointments with to help direct your future?  Yes people, they still exist, and trust me, they\'re helpful.  Careers, however, is a load of hooey.  We watched a movie in \"careers\" about the top ten jobs, I\'m soooo glad we did!  If it wasn\'t for that movie I would be completely oblivious to the world of  Sand Sculpting , and lets not forget  Food Critic !  What would we do without amazing educational videos like that?  With 90% of the cource being common knowledge, the boredom sets in.  This class must have dropped my I.Q. by like a trillion [ joke, I\'m not stupid ].  Oh, lets not forget our E.Q.!  \'Cause your future doesn\'t depend on a silly thing like math, no no, it depends on your Emotional Quotient.  The only think that class is good for is reminding smart kids they are smart, and giving leaving dumb kids open to ridicule.  This class is easy, unless of course your teacher is a gym teacher.  Sure, he is a wicked guy, I bet he\'s very smart too, but why can\'t you just let the guy teach gym?  I complain about careers SO much that the teacher is just failing me on purpose by now, an excellent opportunity for me to grab some evidence that I\'m wasting my time in this class, I bet my mom would love to see a teacher who is ACTUALLY doing their job wrong.  For once, the \'dumb 15-year-old\' is right.  To quote Rob, \"Careers should be optional, and civics should be a full course, and you should have a choice, too.\"  Of course I cleared up some grammar :P but he has a sweet-ass point.  Good luck getting through the course, try to ignore the inevitable anhedonia you will soon be fond of.  OMG CIAO LOLZ!!!!111, Sean  '),(602,2935790,'03,March,2004','2004-03-03 00:00:00','   Sean here.  If you are easily offended, don\'t read this...  assuming you can read...    The majority of rap artists  these days don\'t make their own music.  Now, lets just say, hypothetically, that the music industry was about  music , they wouldn\'t have jobs, n\'est pas? So, really, what are these guys getting paid for?  Oh yeah, they can rhyme.  You see, back when I was in  kindergarten  I also learned how to rhyme.  Why am  I  not making millions? I must have used proper grammar or real words, or maybe it\'s because I did not mention my \'gat\' enough. So in pseudo-conclusion, they are getting payed to flash their bling, and shake their tail feathers.  What the hell?!  Millions of people work hard, get educations, get jobs that help today\'s work force, and by the time they die they will not have made a third of what these  artists  are making.  Problem number 1:  Rap artists + Money = More Rap Artists  You see, a rap artist can coast his way through life.  After making billions, his brother says, \"Yo, G, can you loan me a record deal or two, bro?\"  Problem number 2:  This means the more rap artists we have now, the more dumbass rap artists we will have in the future.  I call it the  Parabola of Brothas  effect.  y = x^2  .\'. Sean goes on a rampage and destroys the earth. And because it\'s in popular culture, kids will eat out of these turds\' hands.  Problem number 3:  The future, a generation of illiterate gang members.  If you\'re not in a gang, ask yourself this, how many people do you know who are in gangs?  Now how many of them can spell \"intelligence\", \"employment\", or \"literacy\"?  Exactly. The only problem is, anyone who complains will get a \'cap\' in their ass.   Thank god for skaters.   Don\'t get me wrong I\'m not super pro-punk, but they are the positive to the Thug negative.  I\'d like to see a huge gangs vs. skater war, no guns, unless you\'re a big pussy.  Jackass rap dildos.  Grr.  \"Payce\" [ GRR! ], Sean  '),(603,2935790,'27,April,2004','2004-04-27 00:00:00','   Sean here.  For  les record , let\'s call it  80\'s DEVO!  Ahhh.  Alright, we need less fat dumbass thugs, and less suicidal goths making the world so serious.  Be a kid!  Be whoever the hell you want!  We need the 80\'s!  Screw the 70\'s, no one wants hippies, and screw the future, who wants neo-techno-hydraulic-pneumatic pancakes?  I want pancakes with ripped jeans, and self esteem through the roof!  Lets synthesize our ways into equality people!  (Except for the gays.)  We need an   80\'s DEVO  , and we need it now!  If you see the world as an awesome place, you\'re trying too hard; we have to make the world a place first!  This won\'t happen without you people.  Screw world vision, all Africa needs is more lasers!  Lasers and synthesizers.  Alright, take all your money that you\'re giving to dying kids in Africa, the freezing in Antartica, the homeless in Toronto, the paralyzed, the handicapped, the deaf, the blind, the impotent, the foreign, the obese, take all that money, and send it to ME!  A vote for  devo  isn\'t a vote for progress!  Let\'s make the world how it was before a gun was a gat, and it was cool to strut and hold your groin while abusing countless substances!  That\'s stupid, you don\'t need that!  Come on!  Let\'s go!  WTF ARE U WAitING 4?!!!1  LOLZ!!11 ROFLMAO WATTTTT!!!!  LET\'S GO GO GO,   DEEEEEEEEEEEEEVOOOOOOOOOOOO .  Yes, I had a lot of sugar and coke tonight.  The lonely days of summer.  '),(604,2935790,'20,April,2004','2004-04-20 00:00:00','   Sean.  Again.  Because Dain isn\'t posting.  So I have to post  all the time .  Heheh.  Okay, so I had this crazy dream where I was black, and Eddie Murphy lived on my street.  He only wore a yellow jacket and grey pants every day (yes, the dream was so long there were \'days\').  Anyways, I kept going down the street every \'day\' and breaking down Eddie Murphy\'s door, and then he would greet me and hand me a new Celine Dion (or someone like that) CD that he burned for me.  I did this a bunch of times until eventually, I became a black  woman .  Weird.  So I went down to Eddie\'s house as a woman, and this time my friend Andrew was with me.  I broke down the door as usual and Eddie got really mad, I think he didn\'t recognize me now that I was female. He pulled out an automatic something, like an Uzi, and started spraying bullets everywhere, and then Andrew disappeared.  I ran back up the street and went home and as soon as I stepped in the door I was me again, and Eddie Murphy was making out with the black chick that  I was  a couple minutes ago, and it wasn\'t me anymore.  I looked in the mirror and I was kind of shocked, that\'s when I woke up. I then had another dream with James celebrating a certain day, that day being today, but I\'m hungry and I\'m gonna go eat so I won\'t get into details.  G\'day!  '),(605,2935790,'10,April,2004','2004-04-10 00:00:00','   Hey, it\'s Sean.  A quick thought on the show Cops;  isn\'t it aaaaawesome?  ^_^ Don\'t get me wrong, I dislike *, but so do the cops on the show, and if you can\'t see that you\'re missing the point of filming the great duty of busting *.  I\'ve seen  Bowling for Columbine  and yes, there is too much violence on TV, but pinning a guy down and sending him to jail, and then listening to a cop explain the technical components of his lunch break is insignificant when compared to the exploding heads in Dawn of the Dead, or something along those lines.   Which reminds me, Maddox is a doodie sometimes.   Anyways I\'m watching Cops right now, and a squad car just got a flat tire  on  some train tracks.  Why were they there?  Some nut job dirt biker was shooting down the tracks.  What could make this better?  A train\'s coming.  So these cops are wiggin\' out trying to get it off the tracks, and then the train takes a left on these other tracks; the train\'s original route.  Funny?  It happens twice in a row.  Could this be better?  Nope, the cop who got his car stuck has a lisp.  I agree, Charlton Heston is a dink, but I think Cops is one of the greatest low-budget shows out there.   * Drug dealers, \'thugs\', addicts, criminals, anyone over 200 lbs., anyone with an I.Q. under 100, any man with a tattoo anywhere below the belt, anyone with over 20 piercings, those \'goths\' with black nails and clothes, \'thugs\', anyone who has bling, anyone over the age of 30 missing 2+ teeth, and anyone who loved Hellboy, \'thugs\', idiots who stand in circles free styling, and Brad Taylor, the world\'s largest animate boulder.  '),(606,2935790,'09,April,2004','2004-04-09 00:00:00','   Before I write this post, which I assume will be long, I\'d like to thank [insert god] for helping make this video, and get it to a theatre near me. For those of you who didn\'t recognize the title of this post, I\'m talking about Jim Carrey\'s new movie  Eternal Sunshine .  I\'ll start by rating it.  4 trillion trillion trillion TRILLION stars / 5.  It is truly the work of geniuses, namely;  Charlie Kaufman  - Writer/Producer of  \'Confessions of a Dangerous Mind\', \'Adaptation\', and \'Human Nature\'.   The guy is 130% genius.  Michel Gondry  - Writer/Director of  \'The Science of Sleep\' and \'One Day...\'.   Awesome guy.  Pierre Bismuth  - Helped out with  \'I\'ve Been Twelve Forever\' , which all three men worked on...   My crappy synopsis:  Lacuna Inc. has this technology that erases bad memories, and all it really takes is one night of being hooked up to a machine where they track down the memories in your brain and erase them one by one.  Joel Barish (Jim Carrey) is an astoundingly complex introverted character, kinda shy, kinda funny.  He meets Clementine Kruczynski (Kate Winslet) in Montauk (sp?) and they kinda flirt, even though she\'s a nut.  \"Why does he fall in love with every girl that shows him the least bit of attention?\"  Anyways, they go out, and get in fights because Clem is a h00r, so Clem erases Joel from her memory.  Joel accidentally finds out and gets pissed, so he erases goes to Lacuna to erase her from  his  memory.  Still with me?  The majority of the movie is after all this; it takes place in Joel\'s mind where it goes back memory by memory, erasing Clem step by step from his mind, and the whole time Joel is living out his mind, watching the memories get erased.  As they are being erased, Joel realizes that he really loves Clem, and changes his mind about wanting her erased.  Twist?  He can\'t get out; he\'s hooked up to this crazy contraption all night and can\'t escape (or can he?).  So in his mind, he tries to avoid the Lacuna workers and protect his memories of Clem by burying her deeper and deeper in his memories, and they travel all the way to his childhood and back.  That\'s about all I can say without spoilers, all though I\'m sure that wasn\'t exactly spoiler-free.  The movie is so good.  When we walked out of the dark theatre back into the hall, Ryan kinda looked shocked, and I sure as hell was.  I don\'t think we were capable of mentally classifying what we just watched, we just knew it kicked ass (Ashworth and Spence seemed a bit too confuzzled...).  The movie flicks back and forth between scenes quickly sometimes, so confusion is understandable, but when the rush dies down a bit the scenes blend a lot more, and there\'s definitely a visual flow that goes with it.   If you haven\'t seen the movie yet, go see it before you read the following.   I love how they do things like the store blending into the apartment room, and the barbed-wire type fence in his room to switch from one scene to another closely related scene, it\'s pure genius.  One of my favorite scenes overall is the Kitchen scene (in his mind).  It\'s so complex because Clem is acting \'normal\', trying not to be noticed, but  really , that is Joel, because she is just composed of his memories.  I find it amazing how his childhood memories lull him into repeating actions like getting the cookie and reaching for ice cream, yet his mind manages to recreate Clem\'s character on a completely different train of thought, all entirely possible and the multitasking type stuff like that is just cool.  It\'s like a form abstraction; you can see something from a couple viewpoints at a time.  If something like that were pulled off in any other movie, you wouldn\'t feel the characters\' emotions at all, that\'s one of this movie\'s strongest features.  The funniest scene would definitely have to be the childhood memory of the bullies, the one with the hammer and he\'s wearing a cape and stuff.  It was refreshing to see Joel\'s character played by a boy for a second, and after it switches back to him wearing the cape and he tries to fight the kid, he remembers the guy as a bully so he is still tossed down, even though he\'s like thrice his size.  It\'s done so well, and I can\'t picture watching it being filmed on set without pissing myself.  I also love the scene in town where his memories get mixed up, if I remember correctly it\'s about the first time you hear the Lacuna workers, Patrick and Stan, talk and their voice is subliminally heard and Joel hears it blasted throughout the streets of his mind.  Crazy concepts in this movie, I\'d like to see anyone who owns a gat wrap their mind around it, or any comic book loving 4 year-old try to sit through it without reading his latest issue of Inquest.   I love this movie.  '),(607,2935790,'03,April,2004','2004-04-03 00:00:00','   Sean here.  I was MUDding it up when I came across this girl who had this joke, and it\'s really funny.  Maybe not, maybe I\'m just really tired and I\'ll laugh at anything.  Either way, this man is a genius:   If you don\'t know the difference between exothermic and endothermic, stop reading now.  The following was a question given on a University of Washington chemistry mid-term exam paper: \"Is Hell exothermic , or endothermic? Support your answer with proof.\"  Most of the students wrote proofs of their beliefs using Boyle\'s Law (gas cools off when it expands and heats up when it is compressed) or some variant thereof.  One student, however, wrote the following:  First, we need to know how the mass of Hell is changing in time.  So, we need to know the rate at which souls are moving into Hell and the rate at which they are leaving. I think that we can safely assume that once a soul gets to Hell, it will not leave. Therefore, no souls are leaving.  As for how many souls are entering Hell, let\'s look at the different religions that exist in the world today. Some of these religions state that if you are not a member of their religion, you will go to Hell. Since, there are more than one of these religions, and since people do not belong to more than one religion, we can project that all people and all souls go to Hell.  With birth and death rates as they are, we can expect the number of souls in Hell to increase exponentially. Now, we look at the rate of change of the volume in Hell, because Boyle\'s Law states that in order for the temperature and pressure in Hell to stay the same, the volume of Hell has to expand as souls are added. This gives two possibilities:  If Hell is expanding at a slower rate than the rate at which souls enter Hell, then the temperature and pressure in Hell will increase until all Hell breaks loose. Of course, if Hell is expanding at a rate faster than the increase of souls in Hell, then the temperature and pressure will drop until Hell freezes over.  So which is it?  If we accept the postulate given to me by Ms. Krissy Jones during my Freshman year that \"it will be a cold night in Hell before I sleep with you\" and take into account the fact that I still have not succeeded in having sexual relations with her, then (2) cannot be true, and thus I am sure that Hell is exothermic.   I\'d give my life to come up with stuff like that.  Sir, I salute you!  '),(608,2935790,'01,April,2004','2004-04-01 00:00:00','   HYPOCRITICAL!  It\'s amazing the jello, it\'s so FREAKIN\' AMAZING! I could step in it all day.  If you find my megaman, please grow an ornate garden.  '),(609,2935790,'31,May,2004','2004-05-31 00:00:00','   W E L C O M E  T O  T H E   L A S T  T W O  W E E K S  O F  S C H O O L ,  A S S H O L E ! ! ! ! !  SERIOUSLY, why is all the homework crammed into the last two weeks of school? If you\'re going to be educating future workers of society, you have to do a damn good job. For any of you incredibly smart teachers designing the curriculum, I\'ll put it this way. Managing Time == Good Cramming Work != Good WHAT THE HELL IS WRONG WITH YOU!? HOW DO I READ A BOOK WHILE PRACTICALLY WRITING ONE?! HOW DO I PRESENT THE DAY AFTER I\'M ASSIGNED A PRESENTATION?! HOW DO I STUDY FOR 3 EXAMS, WITH 4 UNITS IN EACH?! HOW DO I DO ALL THAT WHILE MAKING AN ART PROJECT ON HOW WRONG EDUCATION IS HANDLED?! HUH!?! H U H!?!?!?!?!?!?!?!?!?!?! I HATE YOU!!!!!!  Ahhh. All better. Teachers, get smart you fucking idiots, if you\'re going to teach us do it right. If anything, the last couple weeks of school should be EMPTY so we can just study in order to pass! YOU IDIOTS! YOU STUPID GODDAMN IDIOTS YOU ARE SO STUPID I WANT TO SHOVE SALAD IN YOUR EYES AND SHAVE YOUR HEADS. UGH! I want to do good, but I just can\'t decide what to do good on first! I\'ll put my 10 pieces of work due into a hat, and I\'ll draw maybe 4, and do those. Joking. I will get it done. You\'ll see. I\'ll do it for all the slackers out there.  Love, as you can clearly see, Sean  '),(610,2935790,'23,May,2004','2004-05-23 00:00:00','   \"So what\'s in the book?\" If you watch the movie and ask that question, thinking it will pull the plot together on your head, I think you need a new approach to it. Let me start over.  Clutch  is a good movie. Somewhat indie, Canadian, 1998, Tom Green (not-so-good) cameo, makes you tick. It\'s a dark comedy in the greatest sense of the term, well...A good sense of the term. The only thing in the movie that unfolds to be crazier than the plot itself are the characters in it. The writer/director Chris Grismer seems to be kind of an amateur, very talented, obscure kind of guy, at least to the extent that he projects himself in the movie...I mean, I thought at times my sense of humor was morbid, but I\'d  never  be able to pull it off like he did in the movie. I\'m gonna go ahead and use the line  if you haven\'t seen the movie, go rent it before you read the rest of this . Actually, I don\'t want to review it, I\'m not good enough to put it into words, so let me rephrase that to  go rent the movie , and I hope you\'ll like it.  That\'s it for Clutch...Other than that, this has been a decent fake holiday weekend. I think my dog has done more for Canada by fertilizing grass than the Queen has in her whole life (all what, six-hundred and thirty years?). Yes, that is something to be proud of Queen Victoria, you were born...But wait! I was born too! We all were! Well most of us. Yes, the Queen is about as useful as all the bullcrap traditions that take place in Canadian government. Useless. Wanna cut back taxes? Use some tax money to buy a noose for the Queen and all other \'royalty\'. Where do they get money for diamonds, nice dresses, and being the only people in that entire area with good teeth? I don\'t remember seeing the Queen flipping burgers when she was young. I suddenly understand why other countries, ie USA, make fun of Canada. We\'re faggots! Okay, so Canada isn\'t that bad, but the overseas government and everything ever related to it is incredibly stupid, and should die. The end. No more Queen.  Goodnight!  Sean  '),(611,2935790,'20,May,2004','2004-05-20 00:00:00','   Nananananananana-Sean hereeeee.  Feminists suck.  How would good ol\' Rob put it?  Check out his thoughts here:  urlLink Rob\'s thoughts on feminism.  I think this may be one pivotal issue where Rob has a more conservative stand than I do. It\'s great through, read it...Just ignore the irony of him spelling \'neanderthal\' wrong, among other words, he just didn\'t spell check. :P As for my thoughts?  Right here:  Feminism is competing for the top of the list of stupid causes for the  underemployed . People, you\'re losing your edge. Women don\'t have rights? Here\'s a right for you: shut up about being underpaid and all this incredibly bullshitty bullshit! My Mom works, she gets paid just as much as anyone else there. It doesn\'t even have anything to do with equal pay anymore, it\'s unemployed activist ugly fat women who can\'t shut up and be happy!   If you weigh 300 pounds don\'t complain that there is pressure on female appearance, be an  activist  and get off the goddamn couch! Feminism is an excuse for being ugly and lazy.  So, in order to clear up the jargon the dikes use, I\'ll write a quick  3-point feminism translation guide . It\'s all you need to know.  \"We\'re underpaid.\" == \"Give us money, we\'re lazy.\" \"Women in the workplace.\" == \"We don\'t want anyone to be more successful than us.\" \"Too much pressure on female appearance.\" == \"I\'m ugly/fat/disfigured.\"  Gay pride, feminism, racism, sexism, and all other \'causes\' don\'t need to be paraded, they need to be held up in a court of law! If someone punches you because you\'re black, good for you, sue their asses off and play the racism card! If someone punches you because you\'re white, good for you, sue their asses off and play the racism card! Yes, racism and all this stuff exist, and they\'re wrong, but they need to be dealt with by a man (or   woman  ) in a court of law, not some ugly fat woman with a shaved head waddling down the street holding a sign, a sign she had to get her son to write because she never went to school, she was too busy being paid for doing  shit , just complaining all day.  Damn you! Love, Sean  '),(612,2935790,'19,May,2004','2004-05-19 00:00:00','   S here.  I trimmed my hair.  Erica likes it.  Rob likes it.  I like it.  It\'s good. WOOOOOOOOOOOOOOOOOOOO  '),(613,2935790,'15,May,2004','2004-05-15 00:00:00','   LEGAL, READ THIS:  By reading anything beyond this point you agree that you will not take any of this opinion as fact or seriously in any way, and will not use it as motivation of any kind for any action. You also agree that the person who wrote this is in no way responsible for anything you do.     If you\'re going to slit your wrists, how about instead of talking about it you  finish the job ? Cutting your wrists = suicide! What a novel idea! I\'d go out on a limb and say the world is better off without you taking up our air, and our black paint, and our ugliness. Nevermind, keep the ugliness, just GO, or in the words of my outspoken, unnamed internet buddy, GTFO! I\'ll  pay  you to die. PLEASE. JUST GTFO, I DON\'T WANT YOU HERE ANYMORE. LEAAAVE. Jackasses, get a friggin\' I.Q.. Here\'s some truth to cry about; no one loves you. That\'s why you\'re a Goth in the first place right? Same with \'hardcore anarchists\'. Anarchy, as far as I know, comes from the French  laissez-faire , which means \'let it be\'. I\'m pretty sure it was a pseudo-neutral political party. Anarchy doesn\'t mean rebel, it means leave shit alone! Quite the opposite if you ask me.     A new guy at school, I think his name is brain-dead moron, once said to me, \"You know, total anarchy can\'t exist because there will always be one organization to take down all the others.\" Hey, dildo, do you know what anarchy means? No? Eat my rubber. You know the evolution of the human race is slowed every day by people like you? No? Eat my plastic. I actually think I\'d prefer looking at feces over looking at an \'anarchist\' because that way I don\'t have to brace myself for inane remarks. Everyone on earth exists because there is some agreement between the people that live on earth.  If you don\'t agree, then don\'t live. It\'s that simple.     In conclusion, the more you brag about being an anarchist or a Goth or a shit head, the more you deserve to be shot, and when I become president of the world, you will see just how right I am. Want to have a cause? Want to be against a cause? NO! NO NO NO! Maddox put it best in the heading \'Looking for a safe stance on abortion? Me neither\', and I\'d say it\'s safe to replace \'abortion\' with anarchy, atheism (or religion in general), feminism, or any other stupid cause that these inbreds can come up with. I\'m starting a cause. It\'s called   logicism  . Get smart or GTFO. I urge you to be  logicist , too. You know why I can force people to join my cause? Because technically the people who aren\'t  logicist  are stupid, and I\'m right. Do you want to be stupid? Then don\'t join my cause. If you are smart, feel free to call yourself  logicist , and go enforce your extremely correct opinions on others, and keep working at it until one day you\'re almost as right as I am.  Love, Sean  '),(614,2935790,'13,May,2004','2004-05-13 00:00:00','   Alright, Sean here, thanks to Evan I\'m now officially a cantaloupe. You see, Evan and I were at the local coffee shop sipping our cappucinos when a guy wearing a scarf rollerbladed by and quickly said, \"If you glue your balls to the carpet you turn into a cantaloupe!\"  So I did.  Evan and I discussed which type of glue would work best, and I decided on rubber cement.  I quickly rubber cemented my genitalia to the carpet and BAM! I was a cantaloupe.  Just like that.  It was crazy!  Now that I\'m a cantaloupe I have so many career choices...like...Cantaloupe!  Yeah!  That\'s what I\'ll do!  I\'ll be a cantaloupe, and I\'ll be a cantaloupe.  Think of how much money I\'ll make...I can\'t wait.  Thanks evan, I owe you a  brass monkey .  '),(615,2935790,'09,May,2004','2004-05-09 00:00:00','   Sean posting. Oh Jehovah, you really know how to push my buttons. Imagine you\'re about 3 years old and you\'re playing in a park, when suddenly this fat kid comes over and throws sand in your eyes.  Alright, now let\'s say the fat kid\'s name was Jehovah, and replace the sand with witnesses.  A couple weeks back some  j-ho\'s  came to my door.  Two young men fully dressed in ties, suits, and fancy black shoes.  There\'s one thing they forgot to wear: an I.Q..  I quickly said, \"Oh, you\'re not Jehovah\'s witnesses are you?!\"  They barely responded, just kind of walked away.  Score one for not enforcing religion!  Now today, on this wonderful Mothers\' Day, big fat Jehovah decided to send  two  very attractive female  pornstars  witnesses to my house.  Radical!  How do you tell a pornstar to get the hell away from your house without going impotent?  It went something like this:  Pornstars:   \"Hello, we\'re doing a something something for  better living .\"  Me:             \"Better living?  It says Watchtower, so aren\'t you Jehovah\'s witnesses?\"  Pornstars:   \"Well yes but...\"  Me:             \"I\'m pretty busy right now actually.\"  Pornstars:   \"Oh, well then let us give you this handy dandy guide to our cult, so that instead of spending Mothers\' Day with your Mom, you can sit alone in your room and read it.  Everyone\'s doing it, it\'s all the rage, homeboy.\"  Me:             \"I am  actually  going to kill you.\"  Pornstars:   \"Eat my shit, clear thinking child!\" We then began to throw fecal matter at each other but Jehovah\'s witnesses are incapable of anything other than looking stupid, so imagine if you will two ladies in suits on my front lawn trying to pick up feces with their wrists, while yelling, \"Ahhh me so stupid can\'t get poopie throw at boy.\"    Okay  , so it didn\'t go exactly like that.  Use your imagination, people, you get the point though.  I can\'t make Jehovah\'s witnesses look any stupider than they already are.  So, in conclusion, next time you see one scream, \"Eat my crap, you pornstars from hell!\"  Then throw some poop at them even if they don\'t immediately react, trust me they know what you\'re talking about, they\'ll get it.  Hugs not drugs, Sean  '),(616,2935790,'06,May,2004','2004-05-06 00:00:00','   Sean, again.  So soon? So soon.  Angelfire was cool, they hosted our images, and our images were cool.  In a matter of seconds Angelfire went from cool to tool, and we don\'t have images anymore.  Down with Angelfire.com!!!  Poop.  '),(617,2935790,'04,May,2004','2004-05-04 00:00:00','   I\'ve had too much sugar in the last 30 minutes.  Like, too much. I\'m watching a Cops marathon. I just wanted to clarify the rules of being a cop, aka the rules of being awesome: 1.  Whenever you enter a room traveling above walking speed, you must hold an object (preferably a firearm) at a potent 90Â° \'strong-arm\' angle, and scream, \"Get on the ground!\" or, \"Put your hands above your head!\" at least once, and it has to be at least 4.6 dB (decibels) louder than your normal voice. 2.  Anytime you witness an awry or corrupt occurrence you  must  intervene by means of physical, verbal, or mental force, but don\'t waste your time trying to shoot mind rays, only I can do that.  Trust me, I\'m awesome. 3.  If someone is really stupid, like  really  stupid (ie \"The Cheese Dentures Incedent\" hahah) you  must  kill them.  What?  Oh, right, there are laws against that. Well...Kill them, then kill anyone who objects, and just to be sure you\'re safe, kill yourself. 4. Buy me a digital video camera, I\'ll love you for life and I\'m serious.  Aww hell, I can\'t resist that smile, I\'ll always love  you .  It would be appreciated though.  There you have it, so now it is up to you.  I leave you with these words of wisdom passed down onto me by my good friend glucose;  Be awesome, kick some ass, and don\'t be stupid.  Practice every day, you\'ll know you\'ve achieved awesomeness when you can shoot mind rays like me.  G\'night!  '),(618,2935790,'04,May,2004','2004-05-04 00:00:00','   This post brought to you in part by  SEAN-O-VISION .    \"Hell yes, we have color, what are you, stupid?\"     Imagine videos.  Really good videos.  Really  great  videos.  No, I\'m not talking about Tom Green on DVD or even porn.  I\'m talking about the  Sean Show *.  Yeah, I have some great ideas cooking in my little head and they need to be filmed, but this is where I run into a problem.  I do not own a video camera, and people who do are incredible tight asses!  Yes, you guessed it, I, Sean Berwick, the one and only abuser of commas, am asking  YOU , yes  YOU  for help.  Please lend me your video camera every once in a while or...   Send any donations to the Sean Needs a Goddamn Video Camera fund! 1084 Queen Street West - Unit 32b Mississauga, Ontario, Canada L5H 4K4 (People who know me can always donate personally)   Not only will you get a sweet ass hug (or the female variant known as an ass grab), but people who donate will receive  a FREE secret prize !**  SWEET!  Oh, and I am also currently accepting souls if you have an extra you could spare.   * Show name subject to change. ** You only get a prize if A) I feel like it. B) You\'re liked by me (see the COPS post).  '),(619,2935790,'02,May,2004','2004-05-02 00:00:00','   Alright, Sean here, and I have a wonderful treat for everyone to enjoy. First, anyone who knows me knows I hate the United Kingdom, and anyone who knows I hate the UK knows I hate Stacey Mellish.  Sure, I don\'t actually  know  Stacey, but that doesn\'t stop me from hating her does it?  Besides, she\'s always bugging me, and it seems no one from the UK can spell.  My good old friend Shum (*tips hat*) introduced me to Stacey  Faggish  a while back, and hilarity ensued.  Today was different.  Why?  Because I went to hotmail, looked at her secret question and then simply said, \"Stacey, what\'s your dad\'s middle name?\"...A couple seconds later, \"Anthony.\"  I proceeded to change her password to \' stupidf \' (yes, stupidf) and logged onto her account.  I had nothing planned, it just happened.  Anyways, a bunch of her friends started talking to me, and I saved a couple conversations, and well, they\'re funny.  I am here today to present to you  the Mellish archives :     urlLink Eine ,  urlLink Zwei ,  urlLink Drei ,  urlLink Vier  The messages in  blue  are from me. The information at the top of each convo is stuff I got from their profile.  '),(620,2935790,'13,June,2004','2004-06-13 00:00:00','   Sorry I haven\'t posted in a while...I was working on the server and now it\'s exam week.  School\'s pretty much over though, and I present to you  one of my favourite posts so far.  =D  Do you know Andy Rooney? If so, I feel your pain. Usually as a person gets older, they get uglier, however in this case Andy got stupider. He must have been smelling the fumes off his 200 plastic candles at his last birthday party. I just don\'t understand how anyone can be as brain dead as sir Rooney.  For those of you that don\'t know Andy Rooney, he is the rotting corpse on 60 Minutes who delivers the news that all of us are interested in. An example? Sure, I happen to have one right here;   \"You can\'t tell much about a store from the store windows.   I\'m not a window-shopper, anyway. If I go to a store, I don\'t stand outside looking at their windows. As a matter of fact, most people don\'t seem to pay much attention to the windows.   That must bother the people who do them. Stores go to a lot of trouble making their windows attractive, but they do some strange things. They try to make the mannequins look real, but then, half the time they ruin it by leaving the heads off.\"   - Andy Rooney (read: overpaid ass)'),(621,2935790,'09,June,2004','2004-06-09 00:00:00','   AAAAAAAAAAAAAA WHY WON\'T EXAMS END?!?! WHY WHY WHY I HATE THE LAST TWO WEEKS OF SCHOOL WHYYYYYYYYYYYY?!  AARGH! SEAN, DAMMIT  '),(622,2935790,'05,June,2004','2004-06-05 00:00:00','   Everyone who reads this will hate it. But,  I do give Lord of the Rings credit for being the  least  bit interesting. You know whyyyyy? Because Harry Potter and Lord of the Rings are very well made movies. They\'re made so well that even if you hate the movie, after you see it you will enjoy it. I hate both, because;  a)  The Neverending Story  was shorter. b) Wizards, talking trees, massive computer generated wars, there is only so much I can take. c) Harry Potter, written by some homeless dyke mom who is a good writer, was created to be read to her little kids.  The books got published, and somehow it was not only little kids who read the little kids books. Now she\'s really rich.  If Harry Potter is your thing, then, whatever, but something\'s not right when you try to act uber mature and then you go put your wizard hat on to catch the first release at 1:30 AM. It\'s just not possible.   I would rather see someone get stabbed through their temples, or better yet,  I  would rather stab  myself  through the temples than watch someone who hasn\'t hit puberty go to a fictional school (HP). Also, I don\'t know much about Harry Potter, but they all talk like they\'re from the UK, and I detest the UK. It\'s just not possible.   The only thing I hate more than Harry Potter are the retarded parents who tried to  BAN  the book! OMG! He\'s casting a spell, call the pope.  The whole thing is messed up, you have university graduates reaching climax at the thought of seeing the new movie, and you have parents trying to ban a childrens book!  It\'s just not not not possible.  Lord of the Rings on the other hand is okay. I finally watched the first one, and it was not bad. Yes, that\'s right, everyone who tried to get me to see the movie but I refused, there you have it. I might actually consider it a great movie if a third of the characters weren\'t  the gayest things anyone could come up with, ever . Overall though, LOTR is okay, but Harry Potter...Come on...  Salut, Sean  '),(623,2935790,'04,June,2004','2004-06-04 00:00:00','   Guess whaaaaaaaaaaaat... =D  Sean Berwick, yes me, the awesomest of all that is awesome, has finally set up his home server!  Hosting hooray!  I will divide this introduction into two sections;    Basically,   This means I will host Audio, Video, Art, Text, and other stuff all one one   urlLink big huge server   known as my computer.  You can see all my stuff, and if you want, send me your stuff over MSN Messenger and I might post it! =D    But technically,   My 802.11b router had to be disconnected for the site to work.  Unfortunately, we need to have the router up so my mom can use the other computer that is networked to it, so until I find out how to negociate the site and the router to work together, there will be a problem.  If you have any information on how I could solve that, please inform me at stringtheory@gmail.com.  If you need details, it\'s a microsoft wireless router, 802.11b as I said before, and my server is running on Faststream NETFile 6.6.  I\'m guessing the quickest solution is to use different hosting software, but this is free and yeah.  Well, I hope you guys like the changes! Keep checking for updates.  Love, Sean  '),(624,2935790,'03,June,2004','2004-06-03 00:00:00','   Welcome to 2004, the year where every girl is different! Now there is a wonderful diversity for everyone\'s niche and fetish. Every person has a flaw. You know what you shouldn\'t do? Compete with your friends over who has the most flaws! Welcome to 2004, the year where every girl is totally W00T!  I\'m the fattest! I\'m the thinnest! I\'m the prettiest! I\'m the most popular! I\'m the most laid back girl ever! I\'m the girl that wears unique clothing! I\'m the twisted girl with the mental problems! I\'m the evil girl and I will cast a totally W00T spell on j00! I listen to classical music because it makes me feel superior! I listen to emo because I am the one with the broken heart! I am the only girl with hormones! I am the only girl with a vagina and therefor I will have sex with every boy! I have the biggest jugs in the world which compensates for my I.Q., or lack there of! LOLZ!!11 I\'m totally hardcore punk roxorz! I\'m totally the girl who is always on drugs! I am the girl who wears the tightest clothing! I am the girl who is the cutiest pootiest wittwe ting evew! I am the girl with a cause, down with labour or something! I am the girl who is misunderstood! I am the only girl who can give tips to all my girlfriends! I am the funniest girl every because this one time I shot milk out of my nose! OMG!!11 I am totally depressed and I have herpes or something so I have my reasons, leave me alone! I am the most extravigant female who is more than intillectual, here is pi to the 30th decimal off the top of my head; 3.14159265358979323846264338327950288419716939937510! Love me! No wait, hate me! Alright just give me attention, but leave me the hell alone!  Now how did the divorce rate move from one divorce every 1000 marriages to 40.5? Could it possibly be related to every guy becoming a brain dead gangsta\'? Every girl becoming miss totally w00t mental illness? Hey, people, get smart dammit. How about instead of having a bling collection and telling your latest nursery rhyme to your homies you get an education, get a job, make money, and if not, get castrated because we don\'t want you anymore! Hey laaaaay-dieeeeeeeees, try opening your mouth and closing your legs, well, start at closing your legs, because if you\'re a super goth/slut/anything aformentioned no one cares what you have to say.  WURD!  Sean  '),(625,2935790,'02,June,2004','2004-06-02 00:00:00','   Howdy-do!  I\'ve come to the conclusion that everyone has something they want to share, and if possible, I\'d like to make that possible.  I might be hosting it with my ISP or buying some cheap web space, but this means I need money.  I don\'t have a job yet, so if you have music/art/videos that you would like to contribute, I might have like a sign up page and members etc. and I\'ll host it all somewhere.  You get the idea.  No guarantees here, this is all hypothetical.  I will probably end up creating a large FTP on my hard drive and just using my PC as a server.  If you have any ideas or can help me with this, I\'d gladly give you credit so drop me an IM at stringtheory@gmail.com on MSN Messenger.  Hooray!  Love, Sean  '),(626,2935790,'01,June,2004','2004-06-01 00:00:00','   Hello everybody!  I just want to mention some changes...Dain left, he seems very busy, so now it\'s just me =).  Also, I changed the colors a bit I guess, and the title.  If you want to see a full picture of that check it out  urlLink here . That\'s pretty much it for now. I won\'t be posting for a while because it\'s exam week and stuff, but as soon as that\'s over I promise I\'ll have some good fresh stuff.  Hooray!  Sean  '),(627,4263309,'22,August,2004','2004-08-22 00:00:00','   Well it is sunday morning and I\'m out of cream for my coffee.  Sunday mornings are not suppose to be like that. In fact I should have been served coffee and breaky in bed by some tall good looking stud! That way the stud would have to get dressed to go for cream and not me.  At least that\'s the way I see it in my world. The weeks events took me to work and a funeral. A close friend of mine lost her great grandmother and the service was held on Friday.  I grew up with Shannon and have always been concidered one of the family. Shannon has a large family who are all very close and know how to celebrate life. It was wonderful to see them all and great to take a little trip down memory lane. So that is four weddings and a funeral this summer, I bet that would make a great movie! It\'s cold in my apartment this morning and I\'m just the picture of loveliness with my flannel nighty, housecoat and sweater on. Shouldn\'t it be summer for pete\'s sake?! After I run for coffee cream I will sit and try to fiqure out how to put pictures up on this thing so stay tuned...  '),(628,4263309,'18,August,2004','2004-08-18 00:00:00','   Well let\'s see, what did I do today?  I did laundry.  Yup a productive day off. Tonight I\'m going out to dinner with my son and daughter in-law in celebration of their birthdays. Paul turned 27 (that\'s what happens when you have children at the age of 12 and your so young and your son is so old) on the 8th of this month and Dana turns 25 on the 25th. We are going to a quaint little resturant called \"Le Speggett\". There are only about 8 tables in the place and you have to make a reservation weeks in advance. The food is to die for! Ok I better run downstairs and get my clothes out of the dryer before someone empties it and sees my large cotton panties. The last time that happened I caught the girl with them in her hand, it wasn\'t pretty and when I was done she was brighter pink then the colour of my briefs! As if it isn\'t bad enough to have to put your clean clothes into a dryer which has been filled with god knows what, you run the risk of someones paws on your nickers!  Have a great day all  '),(629,4263309,'17,August,2004','2004-08-17 00:00:00','   Well today is the begining of my postings.      I found this wonderful site through a dear freind of mine who enlightened me of a daily posting being done by mutual friend. This friend (and I use that word loosely lol ) happens to be a tad brain dead in his faith of my beloved Tiger Cats.     I look forward to writing and keeping you informed on my well being.  '),(630,4220141,'17,August,2004','2004-08-17 00:00:00','    i had a sleepover yesterday! lacey came and it was simple, yet entertaining. i taught her some stuff on psp and she made a shitload of lecavalier icons, lol. lacey lacey lacey. lacey and her lecavalier. i just think it sucks that vincent lecavalier plays for TAMPA BAY. godddd. he\'s canadian, play for a canadian team damnit! frvgforjfwe it\'s not fair... the only reason is because the usa offers a lot more money. >:[  i got a betta fish and i named him andy.  supposedly my friends DID call on my birthday, by the way. my brother just never told me. T______________T he\'s such an idiot!!! if he had just told me it could\'ve saved some trouble!!!! jesus christ he\'s annoying...  meh. '),(631,4220141,'13,August,2004','2004-08-13 00:00:00','    lol, some guys on hotornot want to meet me. i just joined that site for shits and giggles and being the idiot i am, i also joined the meetme thing on there for shits and giggles too. now about 8 guys want to meet me and they\'re all in their 20s it seems like. haha, maybe i\'ll start a friendship with one of them but that\'s it. tha\'d be wrong if i started a relationship with one.  they\'re making a movie out of the hitchhiker\'s guide to the galaxy!! i already knew this but i felt like mentioning it because it\'s so damn exciting. ^_______^ i love the book and i can\'t wait to see the movie whenever it comes out (which will probably be sometime next year). i just hope they do a good job with the movie. annnnd sam rockwell will be playing my favourite character, zaphod beeblebrox.  i\'ve been experimenting with brushes recently on account of i just downloaded a lot on psp 7 and i just figured out how to use them. ^^; here are some things i\'ve made:  rachel mcadams:  urlLink http://pictures.greatestjournal.com/userimg/1744504/339596  casey:  urlLink http://pictures.greatestjournal.com/userimg/1745618/339596  casey again:  urlLink http://pictures.greatestjournal.com/userimg/1745619/339596  me:  urlLink http://pictures.greatestjournal.com/userimg/1745860/339596  me again:  urlLink http://pictures.greatestjournal.com/userimg/1745861/339596  me againnn:  urlLink http://pictures.greatestjournal.com/userimg/1745867/339596  me once more:  urlLink http://pictures.greatestjournal.com/userimg/1745868/339596   this was my first attempt with brushes:  urlLink http://pictures.greatestjournal.com/userimg/1744541/339596 . lol. i love grover! '),(632,4220141,'13,August,2004','2004-08-13 00:00:00','    wow, i just got a new blog... cool... i have another online blog/journal too, and you can find that one at  urlLink http://slingshot.greatestjournal.com . as you\'ll notice, everything on there is more polished than it\'ll probably be here. that\'s just the environment on gj though.  yesterday (the 12th) was my birthday. i am now 15 years old. i dunno, i don\'t feel much different and i don\'t feel a year older. the past year has gone by so quickly! it seems like just yesterday i turned 14, not 15. crazyyyy. but i\'m happy to be 15. :]  i didn\'t do anything for my birthday, because practically none of my friends could show for what i had planned. :\\ i only wanted my close friends to come, so only 4 people were originally going to come but 3 of them couldn\'t. so, what was the point in the other person coming? heh. i guess i\'ll just postpone it, but whatever. mehhhh. my friends didn\'t even call. well, supposedly one did but the others didn\'t. that kind of ticks me off. i mean, seriously... they were home \'n all (just couldn\'t make it to the thing i had planned because they couldn\'t show up at the specific times and it was kind of too late to change the times), couldn\'t they have at least called? particularly because they couldn\'t show up for my birthday? grrrr. my friends are weird like that sometimes. it\'s annoying but you get used to it.  buuut i did get very nice birthday signs from my online friends! i   yeah, nothing else to talk about... so i\'m out. '),(633,2300555,'28,November,2003','2003-11-28 00:00:00','   today is better than yesterday... the annoyin customer came back!... *hehehe* and I did the sale... eventhough i had to run around and around and around..... aniweis... have u ever learned how to fly? Me my self... i believe i can fly.. coz all my dream is always about flying.. but today.. i watched a lil birdie - which was found by Daniel - learned to fly... *we really dunno if she or he is a gurl or a boy* we named her or him UWI... so cute...  aniweissss... found out sumthin... that it aint easy to learn to fly... even for a bird... needs lotsa practices.. and guts as well.... i guess that is sumthin that we can copy... any new things we try to do... no matter how hard it is.. if we put our effort into it and keep on tryin ... we will be succeed! * thx for UWI*  '),(634,2300555,'26,November,2003','2003-11-26 00:00:00','   There is only one kind of successful negotiation......it is the kind where both parties come out ahead..... I really do not like using the word \"win\" because that implies a game of some sort where there are winners and losers......  negotiation occurs everywhere in life and life is not a game....2 be a successful negotiator....you must have the other person\'s interests in mind as well as your own.....  '),(635,2300555,'26,November,2003','2003-11-26 00:00:00','   To all my dearest friend..... when it comes to the other person\'s character and personal habits, make sure you can live with these as they are now..and if u cant be happy with the person the way he or she is now, dont pretend that ur happy with him or her... one of the biggest crime in this world is being cruel to ur own feeling...!  '),(636,2300555,'26,November,2003','2003-11-26 00:00:00','   feel like cryin....!!!!....being so nice to people sumtimes can be really hectic.. coz what happen is.. they will easily muck around with ur time.. and feeling...... and after a whole day dealin with all this matter... and what u wanna do is just take a deep breath and relax for a while.... they still expect u to give them the best u can give without knowing that what they do to u is a waste of time...  and CORRUPTION OF EMOTION!... i dunno what the hell is that.. but it\'s just dat.. i feel like im being used..... *huaaaaaaaaaa* hiks hiks...  but this wont stop me from being nice... i know sumtimes it feels fake when u have to be nice while u\'re actually feel sick of being nice... but at least by being nice.. will make the world be friendly back to me...!  '),(637,2300555,'26,November,2003','2003-11-26 00:00:00','   Happy Eid!!!!!.....The winning day for us!!!!.... Just like what i have forwarded to every number in my add book... Let\'s push rewind, reverse and drive away and see all the mistakes disappear in our rearview... Let\'s start forgivin b4 we\'ll be running outta time..  but wait a second... is it only bout forgiving and being forgiven..??? it\'s more than that.. coz that is sumthin that we can do everyday without waiting for this holy day...It\'s about EATING! *LOL* naaah not that one as well... what i like about today... the gathering... laughter and joy.. all blends togetha... forget all the sadness for a moment...  well it\'s sumthin that u can also do everyday though... is just that this time has a different meaning.. i cant really explain what sort of feelin i have that make this day is so special to me... maybe tomorrow... too tired to think..... to selfish to share? hehehhe ;P  '),(638,2300555,'24,November,2003','2003-11-24 00:00:00','   love ur friends while u can still love them! *comes from the deepest of my heart*  '),(639,2300555,'18,November,2003','2003-11-18 00:00:00','   been workin hard lately.. take ova the shop... bit hectic... lotsa fun.. especially if u reach the target... aniweis.. satisfaction and happiness... which one comes first? and which one u after? to me... they should come togetha.. but i know that sumtimes i have to be realistic... that sumtimes happiness and satisfaction do not come togetha.... and sumtimes they dont even get along!.... ok u feel happy because u pass the exam.. but u\'re not satisfied with the grade... that\'s one of the example..or... u can feel that ur body is shakin bcoz u\'re experiencing an enormous satisfaction... but then at the end u dont feel happy at all...  well for me happiness and satisfaction comes togetha is like when i feel so bad about sumthin that i say or i do to sumone and i feel so bad so that i have to ask for an apologize.. and then when u say sorry from the deepest of your heart.. and the person that u feel bad to... trully forgive you and say \"I forgive u\" with a smile... and that smile warms ur heart.. that\'s when i feel both satisfaction and happiness.. that\'s one of a small example...there\'s a lot more.. Afterall, life is all about risks and it requires you to jump. No matter which one should come first...don\'t be a person who has to look back and wonder what they would have done, or could have had *hehehe nyambung ga sih*.... btw topic taken from the convey between Mr. SokTau and Mr.TakeiTEasy... hehehehhe thx guyz....  '),(640,2300555,'18,November,2003','2003-11-18 00:00:00','   just finished all the exam! finally! it\'s time for holideiiiii... :) yippee *alangkah senangnyah* hehehehe.. but wait.. lotsa people think that this is not the right time to think about holidei.. it\'s too sooon they said.. it\'s actually the moment of waitin... waitin for sumthin that doesnt even exist.. ? hehehe... they\'ll kill me if they know i make fun of this... i guess what they\'re tryin to say is that... we\'d better prepare ourself for the worst thing that might happen.... which is FAIL from the exam!!!!... hmm i kinda agree though... but it\'s just that what\'s wrong for having some fun after exam week... it\'s sort of the way of releasing stress... let\'s not forget the minimum sleep hours during the exam week..red eyes.. numb brain.. COME ON!!!... i need to have some fun!... we did what we could... we did the best we could do.. just dont forget to pray... but dat doesnt forbid us from getting pampered? rite? live the life today as maximum as u can... u\'re free to do anything!....AND \"fuck the pesimist!\" (quoted from pearl jam).. excuse ma language  '),(641,2300555,'17,November,2003','2003-11-17 00:00:00','   have u ever felt really uncomfortable in front of the person u just met? it\'s not an okay feeling i guess... no matter how far u try to get to know her/him... it just doesnt feel right.. then it makes u feel unsecure..  *sigh*... so the question is... how d\'ya deal with it? just do what u have to do and dont take everythin she or he does personal?.. or take her/him to the arena? and do the one on one battle.. let some of your kungfu styles out.. :)  i choose the last one.. the worst likely scenario... there will be an argument.. but then u just dont wanna be a hypocrite...and again all i want is just tryin to be honest by sayin what i really feel.. and hope that it will sort everything out...  '),(642,2300555,'16,November,2003','2003-11-16 00:00:00','   hello world?:) sumone\'s bday today.. gettin old.. is that what everybody wants... old = mature? i dun think so...it has got nuthin to do wit age i suppose.. it\'s bout experiences and backgrounds... pasca divorce.. tends to build some internal strenght... a resistance of reality might as well occur but at the end.. it\'s a process of being ourself... we tend to develop ourself... infected by the condition that surround us..  '),(643,2300555,'15,November,2003','2003-11-15 00:00:00','   woke up wit a smile.. today is a beautiful day.. what is a beautiful day? and why sumtimes we have a bad day? a beautiful day exists when we say to ourself that today is gonna be a beautiful day..  '),(644,2300555,'14,November,2003','2003-11-14 00:00:00','   damn my exam.. cmon.. keep blamin on my exam.. dat\'s right! it\'s always the pattern we choose.. blaming on sumthin else while actually the mistake is on us..  '),(645,2300555,'13,November,2003','2003-11-13 00:00:00','   Laughin at people.. that\'s easy and probably sumthin that we do everyday.. if sumthin\'s funny we laugh right? i suppose we should laugh.. but laughin at ur stoopidity.. YOURS not theirs.. sumtimes isnt funny anymore.. admit it.. if u do sumthin funny (accidentally).. ppl will laugh at u.. and u\'ll feel ashamed.. angry probably? Somehow, it\'s hard to laugh at ur own mistake... but i found out that laughin at urself means u will be able to see urself from other ppl\'s view.. and that helps u to build a positive energy..  '),(646,2300555,'13,November,2003','2003-11-13 00:00:00','   Struggling hard to sign out.. i have to .. but what has to be done usually never be done.. that\'s the case!  '),(647,2300555,'13,November,2003','2003-11-13 00:00:00','   The first post is always -not always- the freshest one.. u reckon? i dun think so.. see it\'s all about mind games and about time management... why is that so? dunno.. well one thing i know u will spend lots of time typing all these words if u have enough time.. but if u are controlled by time... then goodbye...  '),(648,2300555,'31,December,2003','2003-12-31 00:00:00','   It\'s nearly nu yer now..  Hope u guys spend it with those u love!  The New Year has always been a time for looking back to the past, and more importantly, forward to the coming year. It\'s time to reflect on the changes we want, or need, if we\'re to have the motivation move forward.  It is having sumthin old.. and making it a lil bit better.. or way so much better.. or to have sumthin new.. and decide to keep it and believe in it..   if u have nu yer\'s resolution, try to stay with it.. or if u have to break it, break it with pride..  aniweis.. stop talkin too much yacko! heuheuehhe HAPPEEEE NU YEEEEEER!!!!!!!  '),(649,2300555,'30,December,2003','2003-12-30 00:00:00','   finally back to shitney.. which i think still the most beautiful city i love... :) but many thx to melbern and indi! *just be creative always ok bro!* :) and glad to see adi again.. on the way back to sydney.. starboy has a bit of problem with his tummy.. so we had to open the window again and again and again.. ! heuheuheuhehe...  guyz thx for the road trip.. for everything!  '),(650,2300555,'30,December,2003','2003-12-30 00:00:00','   finally i have decided... wanna get a tattoo! it\'s a big decision i know.. coz this is sumthin that i will keep foreva... and i dont want to regret it later on.. and one thing i know I LOVE MY MUM! SOO MUCH... u hear me mum? heuheuhuehuhe *ga nyambung yah* i luv her for being her! u rock mum! aniweis.. back to tattoo again.. still deciding what design to put.. and where should i place it...  '),(651,2300555,'28,December,2003','2003-12-28 00:00:00','   still thinkin.....  '),(652,2300555,'27,December,2003','2003-12-27 00:00:00','   still thinkin... and browsing for jewels and what sort of piercin...  '),(653,2300555,'27,December,2003','2003-12-27 00:00:00','   thinkin to get anotha piercin...  '),(654,2300555,'27,December,2003','2003-12-27 00:00:00','   im starviiiiiiiiiiiiiiiiiing.............. but too lazy to go out ! the boyz in da hood are still sleepin... with numerous position.. some weird.. some so sweet.. or pretend to be so sweet? *alias sok imut*.... what\'s gud for boys for being sok imut aniweis? hmm they can use it as a way to attract gurl\'s attention.. gurl usually say *oooh so sweeeeeettt * or *lutunnaaaahh* but beware guyz.. after a while jadinya hiperbola banget!!! and the gurl will say... *ooh u\'re so gay!* heuheuheuheue.... either u wanna take that as a compliment or else .. is up to ya... heuheuhuehe oiss out!  '),(655,2300555,'27,December,2003','2003-12-27 00:00:00','   last nite rite at about 2400 hours went for Lord of The Rings: The return of da king... tried so hard to stay awake.. the first 15 minutes was struggling so hard though..  fortunately didnt miss the battle at all.. the ghost army was way sooo kewl...!!! but i know if i see them face to face i\'ll be scared to death... !!!! hmm one of the gud ending movies....  '),(656,2300555,'27,December,2003','2003-12-27 00:00:00','   im in melbourne now.. after 10 hours road trip... with bayu, daniel and miji.... it was along way down!... of a roller coaster! heuheuhe well actually arrived yesterday mornin.. then went ot while the boys were sleepin.. joined the massive crowd of the boxing day... found sumthin for my mum... and then.. went to chapel st... was about to queue the lord of the ring.. but ever show was sold out! *oooh shit* ... akhirnya ngupi2 di Capelise.. tattoo photos! heuheuheuh i will upload them sumtimes tomoro maybe.. then heading back to Indi\'s apt.. SING-A-LONG! heuheuheuheuhe... *seruuuw* it was full house man! ape lageeeee... hihihi more comin up  '),(657,2300555,'20,December,2003','2003-12-20 00:00:00','   gosh it\'s 1100 hours already.... i can see that im stucked inside my laundy bag in the next 15 minutes...  '),(658,2300555,'12,December,2003','2003-12-12 00:00:00','   life is dinamically changin all the time.. three of my good friends are about to tie the knot soon... two other good friends just recently gained their master degree....other two good friends are nearly finished.... another good friend of mine is a diplomat now... and the other one is busy behind the computer.... just wanna say to all of them... wish u luck in every way... and success always ..............miss u a looooot!!!!...  '),(659,2300555,'12,December,2003','2003-12-12 00:00:00','   got new piercings!... just like a human being... piercings also have ups and downs.. i\'ve been through the downs for the last couple of days... and i think they\'re in a healing process now... hopefully :) aniweis... new year is coming soooooonn!!!!!!............  '),(660,2300555,'12,December,2003','2003-12-12 00:00:00','   at the moment im still thinkin of a good reason to make up our own mind.. dunno why... it feels really difficult to make up our own mind.. ok here\'s the case... we have promised to ourself that we wont do this or do that.. but at the end there\'s always sum excuse to break the promises and act like nuthin happens.. but after a while ... we begin to regret what we have done...*ooh shit*...  hehehhe... sooo... still tryin to find a gud reason.... get back to ya soon... piss out!  '),(661,2300555,'01,December,2003','2003-12-01 00:00:00','   just got back from GoldCoast! it was whole lot of fun! but forget bout the weather! damn the weather! hiks!  '),(662,2300555,'01,December,2003','2003-12-01 00:00:00','   REVIEW In The CuT: totally made myself feels like being cut into pieces! mmmm what\'s dat supposed to mean?  '),(663,2300555,'01,December,2003','2003-12-01 00:00:00','   Chacha belzde!!!! and also AIDS day! Happy Chacha belzde and Happy AIDS day then! :).. aniweis just woke up and smell the fresh air... had a gud weekend... spent ma nite with ma mates... Struggling to keep my eyes open wide.. Finding nemo was still okay.. but Dumb & Dumber fell asleep! :) hehehe how come i fell asleep while the movie was played.. felt sorry for Mr.SoToy for the *Toasted Sandwich*...  '),(664,2300555,'31,January,2004','2004-01-31 00:00:00','   face it! no matter how hard it is...  '),(665,2300555,'30,January,2004','2004-01-30 00:00:00','   sakit jiwa nih gw!  '),(666,2300555,'30,January,2004','2004-01-30 00:00:00','   how can u prove dat u\'re really in love? how can u deny dat u\'re not a playa? how can u make them believe? how can u?.....  '),(667,2300555,'29,January,2004','2004-01-29 00:00:00','   to prove dat u\'re not a playa.. is not to be a playa.. but the question now... will they believe u? will they accept it just like dat? damn! trust is sumthin dat is very hard to get...  '),(668,2300555,'29,January,2004','2004-01-29 00:00:00','   being a playa..  create lotsa betrayal....  '),(669,2300555,'29,January,2004','2004-01-29 00:00:00','   suppose we make up our own ending.. would it be ended like dat? i dont think so...  '),(670,2300555,'26,January,2004','2004-01-26 00:00:00','   i believe dat everything happens for a reason but one thing i dont know.. what\'s the reason behind this? the people that we luv the most... is actually the best position to hurt us so much  '),(671,2300555,'26,January,2004','2004-01-26 00:00:00','   ready to explore a nu life.. ready to move into my nu house.. my nu house for the next 6 mths togetha with miji;daniel;adir...  '),(672,2300555,'26,January,2004','2004-01-26 00:00:00','   i got so confused... i got so mixed up... i got so messed up.. i got so complicated.. at the moment...  '),(673,2300555,'26,January,2004','2004-01-26 00:00:00','   is it okay? if i pretend to be like dat? is it okay? if i let it all happened? is it okay? if i didn\'t do anythin about it? would it change a bit? ...........................  '),(674,2300555,'25,January,2004','2004-01-25 00:00:00','   maybe..  im too happy maybe... im to excited... maybe... im exaggerating too much... maybe..  '),(675,2300555,'25,January,2004','2004-01-25 00:00:00','   never been like dis before... never been happy like dis before.. never been so excited like dis before... never been... never....  '),(676,2300555,'24,January,2004','2004-01-24 00:00:00','   yum cha.... yummy!  '),(677,2300555,'24,January,2004','2004-01-24 00:00:00','   hmmmm.......... new tattoos..... made by bi and di... didnt hurt at all...  '),(678,2300555,'24,January,2004','2004-01-24 00:00:00','   so unpredictable!.... so unbelievable... now i dont know wat to react... fuct!...  '),(679,2300555,'22,January,2004','2004-01-22 00:00:00','   drawer bookshelf chair mattres could it be more than that?  '),(680,2300555,'21,January,2004','2004-01-21 00:00:00','   go to the flow... sumtimes aint easy........ take it easy ........  why not take sumthin seriously...  '),(681,2300555,'21,January,2004','2004-01-21 00:00:00','   it\'s late at nite it\'s way late at nite... it\'s time to hav a dream dat is not real dat is just... a dream....  '),(682,2300555,'21,January,2004','2004-01-21 00:00:00','   keep ur eyes wide open to every elements... the elements  that keep the thinds suck... and b4 u check da whole system ... go check urself..........  '),(683,2300555,'21,January,2004','2004-01-21 00:00:00','   being a fool sumtimes is okay.... i mean fool in terms of freshman...coz at the start of our trip, u feel like a newborn - fresh, open and spontaneous and that we are ready to conquer the world.... however we have to be aware with lots of obtacles out there...  '),(684,2300555,'20,January,2004','2004-01-20 00:00:00','   only with ur heart u would truly see coz the most important things are not visible to the eyes...  '),(685,2300555,'20,January,2004','2004-01-20 00:00:00','   worried.... kinda worried.... pls no no.... i hope it wont happen....  '),(686,2300555,'20,January,2004','2004-01-20 00:00:00','   for my friend who likes to be alone at the moment... i\'ll give u sum time to ease ur mind... for my friend who\'s feelin down at the moment... i\'ll still be here incase u need me...  for my friend during the darkest day..... i\'ll bright ur day...  '),(687,2300555,'18,January,2004','2004-01-18 00:00:00','   adi.... get well soon... sorry that this deadly virus is so contagious... im sad... hiks...  '),(688,2300555,'18,January,2004','2004-01-18 00:00:00','   feel a demon about to explode at anytime... can u beat the demon inside ur body? everybody has a demon inside... and that this demon can come out whenever they want... turn u into a nightmare... i guess it\'s okay for us to sumtimes release this demon... let their anger burst out...  '),(689,2300555,'18,January,2004','2004-01-18 00:00:00','   ....................................  '),(690,2300555,'16,January,2004','2004-01-16 00:00:00','   went for a movie.. love actually... is all around? i dont think it\'s around me....anymo..  '),(691,2300555,'16,January,2004','2004-01-16 00:00:00','   she was blurry... or was it my vision?  '),(692,2300555,'16,January,2004','2004-01-16 00:00:00','   just saw a good looking gurl sumwhere.... and i think i am a lesbian... am i?  '),(693,2300555,'16,January,2004','2004-01-16 00:00:00','   hei shit happens!... all the time... u just have to get ready for it... but the thing is dat... u cannot be ready for it... u cannot! at least be prepared perhaps? but naaaah...  im outtie....  '),(694,2300555,'16,January,2004','2004-01-16 00:00:00','   is it fair? that u know sum people... and they know u but they actually dont wanna get to know u.. while u actually wanna get to know them... replacement happens!...  '),(695,2300555,'16,January,2004','2004-01-16 00:00:00','   have i ever done this..... have i ever done that? did i really do this.... did i really do that? U can never tell!....  '),(696,2300555,'16,January,2004','2004-01-16 00:00:00','   to learn is to read... to learn is to practice... to learn is to make mistake... to learn is to tell... to learn is to discuss... u live and u learn....  '),(697,2300555,'16,January,2004','2004-01-16 00:00:00','   kinda hard to explain the feeling i hav at the moment.... during the darkest moment u hav... generally people claim that they dont see any lights... but i see them... they\'re just there!.. all u do is just open ur eyes...and run... run to them... dont just stare at them... do sumthin!  '),(698,2300555,'15,January,2004','2004-01-15 00:00:00','   all i need is just that sumthin to hold onto...  '),(699,2300555,'13,January,2004','2004-01-13 00:00:00','   now and dreams we run.....  '),(700,2300555,'13,January,2004','2004-01-13 00:00:00','   believe or not to believe in sumone.... u never know whether they\'re saying is right or not...  u never know if they lie about sumthin... u never know.... but never stop believin ur self!... if u can\'t believe in sumone... the least u can do is believin urself..  '),(701,2300555,'13,January,2004','2004-01-13 00:00:00','   log in.... and log out.. like u ain\'t got anythin better to do...  '),(702,2300555,'13,January,2004','2004-01-13 00:00:00','   yesterday... a deep convey about sum theory of relationships...  everyday we learn sumthin bout love... everyday we\'re tryin to figure it out.. to know what love is...  is it like life and death experience? why use the word life instead of birth? birth is the opposite of death... not life... life has no opposite!... life is life... and live ur life to the max... !  '),(703,2300555,'13,January,2004','2004-01-13 00:00:00','   is it true? that what they say is true....  '),(704,2300555,'11,January,2004','2004-01-11 00:00:00','   is it true? that ur heart is beating faster when u meet sumone u like...  '),(705,2300555,'11,January,2004','2004-01-11 00:00:00','   is it true? that if u need sum help then u hav to offer them a reward?......  '),(706,2300555,'11,January,2004','2004-01-11 00:00:00','   i need sum light.... i need sum lampu!.... huaaaaaaaaaaaaa  '),(707,2300555,'11,January,2004','2004-01-11 00:00:00','   here i am again... in front of the monitor.... wondering that it\'s funny how people sumtimes can easily change their mind without thinkin any further... that any other person involve in their decision can be really affected... in a positive way and in a negative way...  '),(708,2300555,'11,January,2004','2004-01-11 00:00:00','   happpeee betsssdeiii shiiiennntaaaahhh!!!!!....... :) :) :)  '),(709,2300555,'11,January,2004','2004-01-11 00:00:00','   Hello there, the angel from my nightmare..The shadow in the background of the morgue...The unsuspecting victim of darkness in the valley....We can live like Jack and Sally if we want ....Where you can always find me....We\'ll have Halloween on Christmas....And in the night we\'ll wish this never ends....We\'ll wish this never ends.....Where are you and I\'m so sorry....I cannot sleep I cannot dream tonight.....I need somebody and always.....This sick strange darkness comes creeping on so haunting every time......And as I stared I counted webs from all the spiders.....Catching things and eating their insides....Like indecision to call you and hear your voice of treason....Will you come home and stop this pain tonight.....Stop this pain tonight.....Don\'t waste your time on me you\'re already The voice inside my head (miss you miss you)  '),(710,2300555,'10,January,2004','2004-01-10 00:00:00','   it\'s really hot in here... !!!!! i need some fresh air!  '),(711,2300555,'10,January,2004','2004-01-10 00:00:00','   today i spent ma day watching movies... movies and movies... she\'s all that, bad company and meet joe black... plus with ice cream and bread, m&ms, doritos, pad thai...  and all other junk food... whatta day :) thx for makin ma day boyz!  '),(712,2300555,'10,January,2004','2004-01-10 00:00:00','   it\'s funny how people have set up a plan on their mind.... but at the end all they can say that \"hey it\'s just a plan!\" and there\'s no realization at all for the plan... but hang on... it\'s not becoz of them... it\'s sumthin that we actually dont think about... it\'s what u call destiny...*i guess?* no matter how good the plan u have but if ur destiny says that the plan doesnt work... then it doesnt work... it\'s unavoidable!... it\'s sumthin that\'s just not meant to be... it\'s not that we choose not to be... but it\'s just have not to be!... so not be it!... and that we can not do anythin about it...  '),(713,2300555,'10,January,2004','2004-01-10 00:00:00','   anotha weekend... still cant spent it outside... still need sumtimeto heal this chicken pox on ma body and ma back.. evry chick pox on ma face has already dried... leavin me with lots of black spot.. which i hope will be gone sooon...  '),(714,2300555,'09,January,2004','2004-01-09 00:00:00','   btw Miji bday yesterday... he blew the candle last nite while he was here... sorry ji.. no proper cake.. :) wish u all the happiness in this life! go miji go miji! :)  '),(715,2300555,'09,January,2004','2004-01-09 00:00:00','   I thank God for the beautiful friends i have.... thx for taking care of me while im in my downs.. :) and for Lampu thx for being so patient...  '),(716,2300555,'09,January,2004','2004-01-09 00:00:00','   *sigh* nevermind... there\'s always a good reason behind all of this.. it\'s just that try to find a positive side while u\'re in the middle of negative things is not always easy... perhaps this is the way that God reminds me of sumthin... to get sum rest coz last year was a bit hectic for me.. to realize that we are not always surrounded by happiness...  and the important thing is that we are nuthin in front of God...  '),(717,2300555,'01,January,2004','2004-01-01 00:00:00','   it\'s been 9 days now.. since i got this stoopid chicken pox which ruin the whole plan i made! I was supposed to be already in Jakarta now, but at this particular time i\'m still stucked in my bloddy apt. I was supposed to attend my best friends wedding ceremony and wear the dress i designed, but at this particular time im still wearing my pyjamas. I was supposed to be in the hairdresser to do ma hairdo, but at this particular time, i even let my hair dry without even comb it....  '),(718,2300555,'01,January,2004','2004-01-01 00:00:00','   huaaaaaaaaaaaa im suffering from chicken pox! damn chicken pox!  '),(719,2300555,'28,February,2004','2004-02-28 00:00:00','   pain hurts but it\'s a part of life must not flee it must not ignore it must not negate it  pain is real and true sumtimes we can only wait for the pain to stop in the worst and most terrible way and tears help ...........  '),(720,2300555,'26,February,2004','2004-02-26 00:00:00','   i cried... my emotions are blocked need to know what\'s really happening.. interior vision that needs to be opened no hope an abandoned dream abandoned my soul coz of my duty  '),(721,2300555,'26,February,2004','2004-02-26 00:00:00','   11 o\'clock in da mornin.. still can\'t close my eyes.. seatin in a cup that\'s half broken observin the world around me and it\'s like joy in loss but is it really joy? or is it like much sorrow and perhaps much remorse as well?  '),(722,2300555,'26,February,2004','2004-02-26 00:00:00','   The pain dat we don\'t feel in that moment from the wounds inflicted and received, hurt a hundred times more later on when we have time to think over all that occured  '),(723,2300555,'12,February,2004','2004-02-12 00:00:00','   I dont love war and battle eventhough i am skilful in both  '),(724,2300555,'12,February,2004','2004-02-12 00:00:00','   aku dan semua yang terluka karena cinta...  '),(725,2300555,'12,February,2004','2004-02-12 00:00:00','   full of emotions rite now..  '),(726,2300555,'10,February,2004','2004-02-10 00:00:00','   how can i forget... the siluet in every dreams of mine.. always there  '),(727,2300555,'10,February,2004','2004-02-10 00:00:00','   10 days after... standing still.. no rewind.. no forward.. no pause.. in the middle of it..  '),(728,2300555,'22,March,2004','2004-03-22 00:00:00','   miss him and her in my own way  '),(729,2300555,'17,March,2004','2004-03-17 00:00:00','   this is not supposed to be like dis... this is supposed to be like dat..  '),(730,2300555,'17,March,2004','2004-03-17 00:00:00','   so many stories to tell so many happiness to share so many black dots to admit so many hypocritism to point out so many euphoria to remember so many sadness to forget so many best friends to count on so many thangz  '),(731,2300555,'05,March,2004','2004-03-05 00:00:00','   back in sydney! sydney is always fun! but.. i miss jakarta.. never felt like dis before .... why oh  why?  '),(732,2300555,'05,March,2004','2004-03-05 00:00:00','   hard to say goodbye...  '),(733,2300555,'02,March,2004','2004-03-02 00:00:00','   time to be happy  and  time to feel sad at the same time...  '),(734,2300555,'02,March,2004','2004-03-02 00:00:00','   dan sakit...! di saat matek pun rasa sakit itu masi terasa!  '),(735,2300555,'01,March,2004','2004-03-01 00:00:00','   matek!  '),(736,2300555,'01,March,2004','2004-03-01 00:00:00','   terbunuh rindu! tapi ga bisa apa apa..  '),(737,2300555,'01,March,2004','2004-03-01 00:00:00','   In death there is transformation a passgae the end of present moment ,now past,  is the beginning of a future moment , now present it arrives suddenly ..inevitably and it\'s not the death of the body nor of the spirit it is an obligatory passage a metamorphosis  '),(738,2300555,'01,March,2004','2004-03-01 00:00:00','   laughter against silence i cry my self to sleep...  '),(739,2300555,'01,March,2004','2004-03-01 00:00:00','   feels like to stop to compromise coz everyday i learn more and more pain..  '),(740,2300555,'01,March,2004','2004-03-01 00:00:00','   my heart is breaking in front of her  or him?  '),(741,2300555,'30,April,2004','2004-04-30 00:00:00','   hav a nice weekend... hav fun go mad...  '),(742,2300555,'26,April,2004','2004-04-26 00:00:00','   is it true? or is it not? is it real? or is it not?  '),(743,2300555,'26,April,2004','2004-04-26 00:00:00','   the only thing that matters is just following ur heart......  '),(744,2300555,'26,April,2004','2004-04-26 00:00:00','   dont take it away from me ..........  '),(745,2300555,'26,April,2004','2004-04-26 00:00:00','   If only i had more time I\'d take you where you wanted to go dis place isn\'t the same without you here If only I had one wish I\'d want a million trillion lifetimes that I could spend with you .... ..... try to understand u .... ....  '),(746,2300555,'26,April,2004','2004-04-26 00:00:00','   :( but i have to smile... :( but i have to laugh :( but i have to understand :( but i have to....  '),(747,2300555,'26,April,2004','2004-04-26 00:00:00','   A: there\'s nuttin to worry about.. B: yes but... A: I said stop complainin.. B: im not complainin.. i just dont understand A: but u gotta understand ... B: hmm i smell sumthin.. A: what? B: nuttin.. A: what? B: i said NUTTIN... A: then stop mumbling and get back to work! B: ...........  '),(748,2300555,'26,April,2004','2004-04-26 00:00:00','   im not as bad as i say i am ................................  '),(749,2300555,'26,April,2004','2004-04-26 00:00:00','   A: im jokin B: im not... A: cmon im only jokin B: yeah rite.. A: dont u see dat im just jokin? B: nope...  goodbye...  '),(750,2300555,'26,April,2004','2004-04-26 00:00:00','   da moment u think u believe in sumthin is actually da moment that u r being denied da moment u think u put ur faith in sumthin is actually da moment u r being betrayed .... denial and  betrayal ... pls dont hurt me now .......  '),(751,2300555,'26,April,2004','2004-04-26 00:00:00','   what if the situation u predict turns out to be unpredictable... what would u do?  '),(752,2300555,'24,April,2004','2004-04-24 00:00:00','   im in love wif the world today im in love wif the view today im in love wif the smells of coffee today eventhough i dunt like coffee...  '),(753,2300555,'24,April,2004','2004-04-24 00:00:00','   dear stars... today i feel a lil bit lonely today i feel like i got confused today i feel like vomiting dear stars i feel like u are everywhere...  '),(754,2300555,'07,April,2004','2004-04-07 00:00:00','   sumtimes i wanna be like water sumtimes i wanna be like rocks sumtimes i wanna be like rainbow sumtimes i wanna be like stars but in the end... i just wanna be me....  '),(755,2300555,'07,April,2004','2004-04-07 00:00:00','   i hate to wake up in da morning knowin dat i hav to.. (what???) i think everybody has got to finish their story....  '),(756,2300555,'07,April,2004','2004-04-07 00:00:00','   i hate this damn dial up thing! geblek! i hate that damn corporate thing! geblek!  '),(757,2300555,'15,May,2004','2004-05-15 00:00:00','   I\'m not applyin no pressure just wanna let you know dat I dont wanna let you go  and I dont wanna let you leave  '),(758,2300555,'12,May,2004','2004-05-12 00:00:00','   If u\'re playin me pls keep it on the low coz my heart cant take it anymo and if u\'re creepin pls dont let it show  '),(759,2300555,'12,May,2004','2004-05-12 00:00:00','   skool, work, recording, rehearsal, assignment work, recording, rehearsal, assignment, skool recording, rehearsal, assignment, skool, work rehearsal, assignment, skool, work, recording assignment, skool, work, recording, rehearsal which one comes first? which one is my top prior?  '),(760,2300555,'11,May,2004','2004-05-11 00:00:00','   unbelievable! he is smart but he\'s not smart enough to control his heart..  '),(761,2300555,'09,May,2004','2004-05-09 00:00:00','   just got back from work.. a bit tired....  '),(762,2300555,'04,May,2004','2004-05-04 00:00:00','   givin up on someone.. means  smashed him or her into pieces..  '),(763,2300555,'03,May,2004','2004-05-03 00:00:00','   if u enjoy these too-good-to-be-true tales (which are not fairy tales) then u wont givin up on them  '),(764,2300555,'03,May,2004','2004-05-03 00:00:00','   why too good to be true? is there such thing like dat? coz to me it has a sarcastic meanin.. nuttin is perfect in this world.. nuttin.. anythin on the surface can be so perfect but nobody knows what\'s in the inside.. no one knows.. unless they try to find out... and learn... only wif ur heart would u truly see coz da most important things are not visible to the eyes...  '),(765,2300555,'03,May,2004','2004-05-03 00:00:00','   i know that it seems ridicullous  or maybe im just too sensitif... but i feel like i got rejected today im just tryin to love them and i know dat it\'s too soon to have a feelin like dat i dunt even know them i just feel like i wanna share the way i feel and begin to learn to love someone again  '),(766,2300555,'02,May,2004','2004-05-02 00:00:00','   i guess im just not alrite now i think i need some time off  '),(767,2300555,'02,May,2004','2004-05-02 00:00:00','   it hurts so bad.. when u\'re not ready to hear the news it drives u mad when u\'re not ready to read the news so be prepared but u can never prepare it  '),(768,2300555,'06,July,2004','2004-07-06 00:00:00','   it was almost a month ago since ma last post feels like 24 hours not enuff for me to do everythin but ,still, (thx God) i managed to do everythin rite  and im happy to live ma life now and i dont fake it  '),(769,3498277,'31,May,2004','2004-05-31 00:00:00','    OK Aaron, the boy with condoms from the play that i mentioned earlier-- i thought he was gonna leave me alone now but today his friend spencer gave me this note--  Julia, As you know i like you and well... i was hoping maybe you had maybe got over your ex-boyfriend. And i know you cant just get over your ex but i just wanted to know.  because if you were i was hoping me and you could go do something together. and if you are ready maybe we could go out. So what do you think? -Aaron  so i didnt respond, right? but i showed the note to Dani and Christine and Madison. then during notes for ACT II,and spencer had seen me so he gave me this other note that aaron wrote to him b/c he thought it would make it better:  Spencer, i like julia but i dont know if she likes me. I know she said that she thought i was a great guy but she also said that she needed time to get over her ex boyfriend but i dont know when she will be over him. I know it takes time but it has been a month plus days. i want her so bad but i do not want to make a move until she is ready. so im hoping she will tell if she likes me. -Aaron  the getting over my ex boyfriend thing was, by the way, an excuse because i mean he is nice and he is kind of hot but HE CARRIES CONDOMS AROUND!!! HE\'S AND EIGHTH GRADER!!!and also, he seemed like really over sensitive. so i figured he was either really sincere or really fake and oh such a jerk. jack saw that i had a note and he asked me about it (he\'s lookin out for me he\'s like a big brother)and i let him read both of them. after reading the first one he asked me if i responded and i said no and he said smart move. after he read the second one, he said   \"No. no. see, i know aaron, and he only wants two things. he wants a relationship, and he wants sex. he\'s playing this game.this is the way it works at his school.the girls at his school...this works with them. see once he gets you...\" \"you need an eighth grade version of me. i know a couple of those...i think i was like that for a while, but i realized it wasnt working. its alot better this way\" jack is a good guy. jack is my foster (couldnt figure out how to spell surregit) brother. and he needs to go ahead and call my sister!!! for those of you who read MY LEG IS ASLEEP (my sister\'s blog), jack is, infact, the \"hot play guy\" her blog is at http://my-leg-is-asleep.blogspot.com  ttyl   '),(770,3498277,'30,May,2004','2004-05-30 00:00:00','    ha! i recruited bunnsy!  PrincessHules: mike skidude759: yeah PrincessHules: you should get a blog. random people would like to get inside your strange strange mind. that would be interesting skidude759: ok skidude759: yd u say mike though skidude759: it scared me PrincessHules: i dont know PrincessHules: i felt like scaring you skidude759: i thought i was in trouble PrincessHules: that was my aim  everybody go to his blog. its http://bunnsazzleisgangster.blogspot.com he\'s a cool kid. that should be an interesting one.   '),(771,3498277,'30,May,2004','2004-05-30 00:00:00','    ok yeah i said the time issue got fixed, but no. no it didnt.   '),(772,3498277,'30,May,2004','2004-05-30 00:00:00','    school really needs to end. i mean like, now. we have 3 more days goddammit! screw that hank battle.  skidude759: i dont want to go 2 school 2morrow bohemianbabe713: why not skidude759: i dont know skidude759: just not in the mood bohemianbabe713: people rarely are  that about sums it up and that was like...a looooong loooooong time ago school is just like icky  yeah atleast the rest of the school days arent real tomorrow we sign yearbooks, have the awards ceremony, and spring fling and then wednesday (how the fuck do you spell that???) is emerald pointe. and then thursday is the moving up ceremony and friday is commencement which i cant play at anyway we\'re playing at the moving up ceremony though so its all good. Pachelbel\'s Kanon is the one i really wanted to play, and ill get to so its yayish. but this way i dont have to play pomp and circumstance or chorale, both of which i dont like, so its all good. i would like pomp and circumstance if i wasnt on second. first is pretty. chorale is just boring all around.  but thats ok. hey, i fixed the time. exellent. im on the east coast like im supposed to be now.  the flaky biscuits were very good as promised.  we\'re not having people over anymore today. i dont have to move my ass off this couch until five oclock if i dont want to. oops...i just did...teehee   '),(773,3498277,'30,May,2004','2004-05-30 00:00:00','    aaarghhh!!.!. It\'s SO EARLY and i CANNOT MAKE MYSELF GO BACK TO SLEEP!!! no matter how hard i try i always end up waking up early! i cannot sleep in to save my own life!! well atleast we dont have any school today becuase its memorial day. thats a good thing. tomorrow we get to sign yearbooks and good crap like that. i look suggestive in my yearbook picture. hehe. he made me like raise my eyebrow and stuff. either that or i look drugged out. and by the way when it says the time of my posts, its wrong because i is on da eastern seabord yo. wow the ghetto and the navy talk dont really mix. imagine that... so right now its about nine o\'clock. darn that mosquito...biting me.... my knee has been like killing me for the past like 5 days and i dont know why but thats ok nick was not at rehersal AGAIN and its gettin pretty darn wierd. we have rehersal again tonight...this play is soooo much fun. its a huge cast, theres like 20 of us. aaron is finally leaving me alone. but you dont know about aaron actually... well he\'s this boy who carries a pack of condoms around \"incase he gets caught up in the heat of the moment\" and he kept hitting on me for a while. but he stopped now. he\'s really nice and stuff and he\'s kinda hot but i dont really want to go out with someone im in a play with...who carries condoms around cause \"you never know when someone\'s gonna get raped\" (when he said that, Dani\'s response was- \"who are you planning on raping??\" ive gotta work on that last monologue in ACT II.  haha look at maddie\'s away message guh..    movie then bed.. hopefully i wont wake up till 2marro after noon..hehe (cos its summer & thats what you\'re supposed to do...:-P) but yeah .. call//text the cell i wont be asleep for a while;-) *******..mwahhhhh.    it\'s just this thing the seasons do.  lol its just this thing the seasons do?? i dont get it, does madison not like summer??awwww. thats ok, maddie is great. maybe she is referring to the summer heat making julian all crazy like  well ill ttyl my moms making flaky biscuits for breakfast. joy.   '),(774,3498277,'30,May,2004','2004-05-30 00:00:00','    \"okay, father...get out your holy water and cast satan from this foosball machine!\"                                      -Cheers  heh   '),(775,3498277,'30,May,2004','2004-05-30 00:00:00','    *weeping* spongebob mac and cheese just tastes so much better than the regular kind!!! goin 2 rehersal soooooon   '),(776,3498277,'30,May,2004','2004-05-30 00:00:00','    Today is sunday. I woke up and like laid in bed for like an hour. And then i had a doughnut. And then i started watching the Cheers 48 hour marathon. then i did work outside on the pool because we\'re having some people over tomorrow. *sigh*. this has been like an amazingly boring day. the only good thing will be rehersal tonight. yay! i\'ll get to see maddie. Adrienne (the director keeps telling me im going into story mode when im telling that     STORY in the third scene. IMAGINE THAT. i dont know how to stop though. but she said its getting better. Nick wasnt there yesterday so the kiss scene was me like standing there. lol. i   think he\'ll be there today though. i need to relax with that. Dani and Christine are like obsessing over nick. or atleast they were. dani keeps asking me if i think he\'s cute. i cant really think about him that way though. its wierd. being forced to kiss him so much kind of makes it impossible to do. hmm...opposite of what you\'d expect, huh? no school tomorrow!! and then only three more days. whoa. im OLD.     urlLink      I am Rapunzel!  urlLink  Find your fairy tale character  at  urlLink kelly.moranweb.com .    thanx, andriia ;)   '),(777,3498277,'13,June,2004','2004-06-13 00:00:00','    Your Homicidal Rampage! by  urlLink  crash_and_burn        Your name:           Weapon of Choice:    Hairspray      Your Favorite Target:    Necrophiliacs      Your Kill Count:    1,336,282,091      Your Battle Cry:    \"My kidneys tingle with pleasure!\"      Years You Spend in Jail:    22      How Much Money In Damages You Cause:    $82,229,805,016,771        Your Homocidal Insanity Level:: 46%                                                                                             Created with the ORIGINAL  urlLink  MemeGen  !   '),(778,3498277,'12,June,2004','2004-06-12 00:00:00','    im back from the beach, and tomorrow im going to camp. ill be back on the 25th.   '),(779,3498277,'06,June,2004','2004-06-06 00:00:00','    we\'re going to the beach today after strike. my parents are coming to the show today...they havent seen it yet. last night we had an awesome audience, in some ways i feel like it went better than last night and in someways it didnt (such as jack twisting his ankle when he jumped on the bench and it fell over). the last show is today...i dont want it to end! noooo... beach then camp YES   '),(780,3498277,'05,June,2004','2004-06-05 00:00:00','    yay the show went really well. we have another one tonight, and then tomorrow afternoon. there were no more notes this time, but when we went out to la carretta afterwards, when we were about to leave kelly told aaron not to do anything else or else they will find out about it, and he says \"i love her\" and kelly goes \"i know\". arrrgh. creepy.   '),(781,3498277,'04,June,2004','2004-06-04 00:00:00','    Guess who called me again last night? AARON! THATS RIGHT! AARON CALLED ME AGAIN!! AAAAAAAAAAAAAAAAARRRRRRRRRRRRRRRRRRRRRGGGGGGGGGGGGGGGGGGGGGGGHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHH i have caller ID so i looked at the phone and braced myself. he asked me why we hadnt talked in like two days. i said i was just in a really bad mood yesterday, and i didnt really think he wanted to talk to me. and then he asked me if there was anything wrong today and i was like well not really. and then we talked about a bunch of random stuff, including where we\'re going to highschool. i said i would go to public school but my parents wont let me. theyre overprotective. and he\'s like well i can understand that, youre a beautiful girl and they dont want you to get hurt. i was like...ok... and then he got around to \"we should do something after the show\" and i said \"yeah i think everybodys goin to la carretta\" and hes like \"no i mean after the run of the show we should do something, like go see other shows or something\" and im like \"yeah we should get a group of people together and do that, that would be fun\" aaron: \"i knew you were gonna say that...i was waiting for you to say that\" me: \"what do you   expect   me to say, aaron?\" aaron: \"You\'re freaking out\" me: \"and   how   do you want me to respond to that???\" aaron: \"chill\" me: \"what is the purpose of saying that? is it to get me to calm down and...(do what you want me to do)\" aaron: \"julia- chill. calm down. ive gotta go.bye\"  NO I WILL NOT CHILL AARON!!!!!!AAAARRRGHH!!!!! Opening night tonight! yayayayayay! i think im ready. i think the cast is ready. its a pretty good show, i think. tonight\'s show is sold out, so thats good. before the show i have the luncheon thing at the country club before the TIP recognition thing. and then i go to Wait Chapel, and then i go home and relax and call is at six thirty! ooo im starting to get nervous...eeeeeeeeeeee   '),(782,3498277,'03,June,2004','2004-06-03 00:00:00','    i thought i was off the hook after aaron called me last night...i was wrong, i guess! ugh. I came into rehersal yesterday absolutely exhausted and in a really bad mood. i was feeling really overwhelmed by the whole aaron thing, i had been at emerald pointe all day, and i hadnt slept well the night before. so i really didnt feel that great. Jack came in from eating at La Carretta, and he was talking to aaron and alexis and nick and everybody. i was like jaaack...and shot an evil glance at aaron. he was like come here, and we went over and sat on the stacks of chairs and i told him about what i said to aaron and him calling me. Jack said that he was trying to cover his tracks. Or if he\'s mature, he was calling me for closure (or what he sees as closure). I tried to calm down a little bit, and we started running the show. that made me feel a little better. it made me feel a little better until i was sitting with Dani and Christine, and they gave me another note. it was folded up and it said Cristin (hehe he cant spell christine right)in aarons handwriting. my stomach just dropped. i was like not again!!! They told me i absolutely could not let aaron know that they had given me that note. dani actually said \"put it in your pocket! do not read it until you are home safe in your bedroom and you have locked the door!!!\" so of course, at break i went in the bathroom and read it. even after jack had told me not to. i couldnt help myself. Dani and Christine came in the bathroom with me. As i read it, my eyes widened and i could barely stifle a scream. Dani and Christine were trying to shush me...it wasnt working too well. i counted to ten, went back into the rehersal hall, and sat down next to Jack. I thought i was saying \"Jack, i read the note even though you told me not to\" but i guess he couldnt understand me that well. \"whats the matter, did he say something?\" i told him that i read the note again. \"Why\'d you read it, jules?\" i didnt know why. i said i had to read it, but i think i was too exhausted to feel suspense. i probubly would have if i was feeling normal. i gave the note to jack and told him to read it but told him to make sure aaron didnt see him. we did ACT II. when that was over jack beckoned to me. he said \"is this really bothering you\" and i nodded my head. then he asked me what i want. i asked him what he meant. he just said \"what do you want?\" and i said that i dont know. Then we had to go over to the trailors b/c Annie Get Your Gun had to rehearse, and they\'re a mainstage show so they have the right of way if you will. we sat down in the trailor and jack told me that it looks to him like aaron is talkin about all this depressing stuff like how bad his life is, but then he sees him laughing and acting normal, when im the one whose upset and putting emotion into it. he said he\'s trying to put me on a guilt trip and i need to just not think about it. then we ran the last scene and worked out curtain call. im inbetween jack and nick, and when the other line is bowing jack keeps using me as an arm rest. lol. he\'s great. maybe nothing will happen with aaron today...maybe... today we have the closing ceremony at school. next year we r gonna b eighth graders, and rule the school...hehehe. revenge! yes!   '),(783,2678611,'17,April,2004','2004-04-17 00:00:00','   If you follow the path you must not follow the path itself, but rather, follow your goal, because if you focus on the path and not your ultimate goal, you may fall astray and turn back.    You can make yourself believe anything. If you want to hate someone you can, if you want to love someone you can. It\'s actually a very simple concept, its how people fool lie detectors, all you have to do is believe the lie. So if you can believe a lie, then you can believe a feeling, maybe thats how people in the middle ages learned to love the spouse they were forced to marry, either that or they hated them and eventually murdered them in their sleep...one of the two.  '),(784,2678611,'12,April,2004','2004-04-12 00:00:00','   Notes to self:  Get Zima  Work out EARLIER  Go and find a new easily editable template, or step shit up a notch....  Get school work caught up..  Learn to wake up a half hour earlier for school work catching up purposes.  '),(785,2678611,'18,April,2004','2004-04-18 00:00:00','   Word plays get you vanished Like a shadow in the day, demolished Another skull to add to my collection Another life intersected like a contradiction Off the chain, Delivered to your brain Drove a man mentally insane Chewin\' myself out the straight jacket 180 degrees fall accross the wratchet 6 AM off to go break another back Pull 9-5 without an ounce of respect Every day I do my best to chill Everytime tryin to find another cheap thrill I could use a little more structure in my life Check all the possible prospects not pertaining strife Tryin\' not follow everything along the same circuit Doin\' the best not to become another name in the obituaries Spread across the page flows the deepest of my brain quaries Too many people willing to risk their soul for solom vindication Willing to spend a 2-week spoil on premeditated medication I\'m straight forward with the words, just letting you know Too many two-faced people are they friend or foe Just keep your gun cocked up Fuck around though and find your bitch knocked up Keep callin me cuzzin and find your chin bone buzzin I aint a nice person, your head is about to be waht im submersin Pull the chrome out, too late, im pullin ur spine out  '),(786,2678611,'12,May,2004','2004-05-12 00:00:00','   Oh shit!  On the news some kid from my school was booted out cuz he had a hit list with like 10 dudes and some teachers, and shit about how he was gonna kill them and then kill himself, fuckin Columbine wannabes, shit, what the fucks wrong with people. Then i started thinkin\', was I on that list? Nahh, prolly not, dude was 17, makin him a junior, or a pretty stupid sopohomore, so i was most likely NOT on the list, they didnt say who that kid was though, and I wanted to find out when I went to school today, but im sittin at home cuz my fucking ride is 2 stupid to give me a call to see if im on time, they stupid fucker always leaves late to begin with, an extra minute isnt going to fuck up his schedule, god damn, he forgot me yesterday 2.  I also watched that vid of the dude gettin his head cut off, on the time stamps 11 hours lapse over the whole thing, one time lapse was inbetween when they assaulted him, and cut his head off, so i was wonderin what the hell kinda fucked up shit they were doin do that dude in god knows how much time, i\'m going to just assume anally raped. I also noticed dead peoples faces look really weird, really creepy, but before i get TOO sadistic lets move on.  My mom gave me an extra 10 bucks this morning, i dont know why, but she did, so thaz cool, maybe ill go buy a CD or something with my other money OR!!! ill put it in my fund to get my Klipsch 5.1 Ultra speaker system for my comp, after i get that i wanna get another 2 extra subs i can link up 2 it, so it will have the dual side firing subs, and then 2 extra 8 inch subs, 100 watts a piece, my current 6-inch sub is enuf to make parts of the house rattle, so i figure with my next set up ill be shakin shit off the walls.    Sometimes i just wanna grab somebody by their hair and start punching the crap out of their face, i figure before they can fight back ill have 6 good hits to their nose. at that point they might be in a state of shock allowing me to smash em the eyes 2. Dont ask.  I started a new book, its BAD ASS, yeah thats right, books can be cool 2 you mudda puckers. It\'s The Face by Dean Koontz. I haven\'t gotten very far, but its about this ex LAPD cop workin for the biggest dude in hollywood as a security manager or something, and some dude has been like sending in \'disturbing\' packages, and the dude is all tryin to solve the puzzle. Except so far its been being solved really fast and theres like 350 pages left so i figure the plot will thicken.  Indeed it shall.  Who said that?  I\'m listening to Atmosphere, i\'ve been listening to that a lot lately, Slug is a bad ass rapper and he generally has pretty good beats layed on his trax. oh yeh that reminds me I burned a copy of Aesop Rock - Labor Days for my russian teacher Mr. Plichta, he asked me 2 so i figured why not, I doubt he will like it though, oh well, doesnt hurt anything.  Also in russian we were discussing about how people need to be able to share their opinions freely, cuz fuckin Kim got pissed at mr. mac for saying there was no god and only science so she and her uppity bitch friends met for discussion after school, fuckin bitches like that get me pissed, who gives a flying fuck what somebody said, especially in science class, by a FUCKING SCIENCE TEACHER, what the hell do you think hez going to say? get real you stupid fuckers, i swear to god, people are so ignorant, its like they are appauled that somebodys opinion conflicts their one so throw a fit.  Who lit the fuse on your fuckin tampon?  good god people, seriosuly, lighten up  now if he was inforcing his beliefs and saying that his way was right, thats a whole nother story, but thats not what happened, atleast to my knowlege.  Oh yeah i was watching a news video about the raising gas prices and it informed me well. there is no shortage of gas, the prices ballooned cuz the demand on wholesale is so high they cant keep up with production so they jack their prices causing the other prices to soar also. And compared to inflation, this what our grandparents payed in the 1950s. (think ratios you stupid fuckers)  '),(787,2678611,'10,May,2004','2004-05-10 00:00:00','   Like a mexican i drop my ride on metal, no shocks I stole these rims, find your bros car up on blocks I pack bump in the truck, soundin like bags of pepsi cans I \'found\' sparco seats and 4 vtech enjans 3 tires and a doughnut, i smoke the rubber who the fuck u think i am? i dont dress in fine leather i dress in kmart blue light special, light on the wallet like a feather im a cheap son of a bitch congratulations, mr. hood rich fuck you and your baby is about to become a orphan ugly ass mother fucker looks like he got Marfans is it really dark, or are you just really blind you better lock me in a padded room cuz ive lost my mind  '),(788,2678611,'09,May,2004','2004-05-09 00:00:00','   Some Pics for you fools                  and VIDS! (if you are mentally retarded, you need to use the copy and paste method if you havent figured that our yet)  http://www.ebaumsworld.com/dontloot.html  http://www.ebaumsworld.com/airhornfun.html  http://www.apple.com/trailers/universal/thechroniclesofriddick/trailermedium.html  http://www.apple.com/trailers/fox/dayaftertomorrow/trailer2/medium.html  http://www.apple.com/trailers/fox/dodgeball/trailmedium.html  '),(789,2678611,'09,May,2004','2004-05-09 00:00:00','   If this doesnt work, there will be hell to pay.  '),(790,2678611,'09,May,2004','2004-05-09 00:00:00','   TESTING AGAIN!!!  work damnit work!  '),(791,2678611,'09,May,2004','2004-05-09 00:00:00','   This is a test   damn skippy!  '),(792,2678611,'19,May,2004','2004-05-19 00:00:00','   i have to pee  :-\\  '),(793,2678611,'18,May,2004','2004-05-18 00:00:00','   Ok so where was I, yelling cynthias name to no avail right?? ok anyways.....  I fumbled my way through the endless sheets of rain, and soaked ground, trying to get my ass out of here. I figure I\'ll head back to the car and wait out the storm and that\'s where I will probably be the safest, except I didn\'t remember which way I came from, any foot prints I had layed were wased away 5 steps back.  What was that? I could\'ve sworn I saw Cynthias figure running in the distance, I took running off after her, yelling her name, every few minutes I would catch a glimpse of her, never being able to reach her. I thought the rain was drowing out my voice which would explain how she didn\'t hear me.  (i really dont feel like finishing this story)  I felt like I was running accorss something hard and I looked down to notice I was running accross the road once again, I looked down at at it, and turned around to see a car flying at me, breaks screeaching, it swereved and fish tailed, the rear end of the car collided with me, launching me back.    I woke up 3 days later in the hospital.  ~Fin     Bitches  '),(794,2678611,'18,May,2004','2004-05-18 00:00:00','   Since when did love for a night in a hotel room cost 5 dollars?  Since when did mexicans name their kids Jesus?  What\'s with all these kids ridding 22s?  Why the fuck do people do crack?  Where the fuck is Osama Bin Hatin?  Why the fuck is everybody so negative?  \"This is gau, that sucks, i\'m better than you cuz you suck so bad, everything is shit, you arent shit, go to hell, everything stinks\" Ya know what? Fuck you and fuck off, nobody wants to listen to a baby whine all day long, unless ur name is michael.  Why the fuck do people think i wear expensive clothing? Its not my fucking fault they\'re poor! good god!  Why the fuck is everybody so fucking uptight?! i manage to live without being uptight, why cant they!?  Why do people talk about how they would only date hot chicks and shit, when they always have to talk to the biggest losers of chicks who are ugly as fuck?!?! dont be so fuckin picky, you stupid retards are going to be virigins forever, and paying for it doesnt count!  thats enuf ranting for tonight  *takes a deep breath*  '),(795,2678611,'17,May,2004','2004-05-17 00:00:00','   AAAHHH MY FUCKING TOE!   itll only cost like 370 bucks for a 12 gauge mossberg pump action shot gun with pistol grip, thats cheap, i want one bad, talk about personal defense.  Speaking of personal defense!  today in both health and russian class we went to this one crazy asian karate dude, and learned some TEQ, he does this  kemp bhudist style thing. and i was fucking around and jacked up my toe good, thats wat i get for being a idiot, oh well.   im bored, cant think of anything worth righting, oh well fuck it.  '),(796,2678611,'17,May,2004','2004-05-17 00:00:00','     '),(797,2678611,'17,May,2004','2004-05-17 00:00:00','   I stepped of the train at about 9 o\'clock at night, the ride was what should be expected from any train ride, fairly calm with no interuptions, it wasn\'t a full a train and i didnt have to sit next to anyone, i pittered about on my laptop for most of the duration.  My ridr was late like always, I was expecting my brother Ryan to pick me up, i was expecting him to be late, so all was on time. I noticed a man drinking tea and feeding his dog a cookie, this intrigued me for some reason. Checking my watch i notice Ryan is later then usual so i figure id give him a call, right when im about to call him i see his car pull up, his wife is driving, he explains Ryan was too busy to come pick me up so she had to come instead. It was another 2 hours before we got to where we were headed, I watched out along the cold landscape, as it passed me by. Cynthia (ryans wife) and I make little conversation on the ride home, a gentle rain starts hitting against the windshield.  As we put miles behind us the rain is picking up, the visibilty worsens and the night is turning into a spiraling staircase of rain drops and windy hell. I never liked the way Cynthia drove, keeping her consistant 80 mph as we barreled down the narrow highway, only about another hour to go.  This type of weather was uncustomary to this dry terrain. A coyote darts accross the highway barely missing us, we serve out of its way and back into our lane, Cynthia looks at me, close one huh? Her distraction leads to something more unpleasant as we hit a section of washed out road at 80 miles per hour, the car veers to the left sharply and it flips off the road. I woke up with the most god awful head ache ive ever experienced in my life, and im laying in about 5 inches of water.   I go to check on Cynthia and she\'s no where to be seen, I crawl out of the wrecked car and call for her, we mustve been thrown for about 300 feet off the road because I can\'t even see the highway anymore, and my cell phone is ruined from all the water.  I find disturbed earth, or atleast what i think is disturbed earth, its still raining, and im having a hard time finding where the car impacted along the ground, and i feel highly disoriented, i touch the sore spot on my head and check my hand to see that im bleeding a good amount. I call cynthias name to no avail and i start to become very worried...  Would you like to know what happens next?  Then stay updated on my blog, I may post again tonight.  '),(798,2678611,'17,May,2004','2004-05-17 00:00:00','   Wut\'s up bitches?  I think i might have started a blog post like that in the past before.....hmmm, oh well  So i fuckin wake up at the time o 745 hours, and its like fuuuuck, no time to get ready, half asleep, ive been fuckin attending class lately, so i say to my mom \' can i stay home from school, me so tired\' she ok, and im like score   so its like fuckin 1045 hours and i roll outta bed and check the laundry and hit it for like another 30 minutes cuz that pos drier needs to be replaced by one that would be greater.  this monring i had another feeding frenzy, ive be hungry a lot latley, i normally i dont eat that much.  so its like fuckin 1200 hours and im pretty full and ive fucked around on my computer and i walk outside to take my dog for a walk, and my momz all like \'why dont u weed eat\' so i mow the lawn instead.  so its like fuckin 1400 hours and ive mowed the lawn, well maybe its not this late but i didnt wear my watch today so i dont know what tha fuck is goin on, so i walk inside, the weather is going to piss and i just doint feel like bein outside anymore, oh yeah i walked my dog around some more and played ball with him.  so from that point until now has been spent sitting instead of this fuckin box, and its like fuckin 1800 hours.  '),(799,2678611,'30,May,2004','2004-05-30 00:00:00','   nah i wish i could write something like that, couldnt shoot close if i arced it at the sky.  if u want to know any other good hip hop u can contact me on aim: icewatterr    hip hop is all about the real stuff in day to day life, how many people can relate to ferarri enzos and enuf gold to be called Midas?  new rappers like 50 cent and lil jon cant rhyme and they wont battle the tight mc\'s because they are still being challenged by reading, so why rap when they can sweat over a good paperback  if u go into a chat room geared twords hip hop ull prolly find a bunch of kids \'batteling\'; their rhymes consist of the stereotypical shooting up bitches and \'that gangsta shit\', sad isnt it?  Talib Kweli - Good Mourning  \"Good morning, Brook-nam\" Another stop... on the train (wake up... wake up... wake up...) We come to a stop that everybody got to make...  Whether you local or express  [Talib Kweli] What\'s the meanin of ghettofabulous Not ridin the back of the bus I\'m a revolutionary antagonist Some playas is mad at us for just doin our music out of love Some underground heads is hatin cause we have fun at clubs I\'m probably on some government list for my rhymin You a fool if you don\'t think they already tapped your line Medicine is big business so my remedies is herbal It\'s music is for the people so we Reflection Eternal Listen, you hear the difference between science and science fiction We blow it out like if you leave on every appliance in the kitchen  at once; still rolling kind bud in Cuban blunts On the corner watchin how kids comin to Brooklyn for they fronts Niggas run past what they need chasing after what they want Fuckin chumps, you walk down the street and get jumped Brooklyn cats like to bubble out of town no lookin back When you a ghetto chef you mastered the art of cookin crack Some get caught sleepin on the Mother City so when they go They come back as tales of niggas we used to know Never looked up to see the stars in all they heavenly glory Just straight ahead cause the peripheral is buildings with mad stories Not floors but dramas is played out, shorties get laid out like respect and fade out like TV sets into the banks of our memories (let it be) we\'ll never forget you Lyin on your deathbed askin for God to bless you  [Hook: Talib Kweli] Good mourning, good afternoon, good night What have you done with your life? Everybody time comes to be embraced by the light You only scared to die when you ain\'t livin right, man I\'m puttin up a hellafied fight {*Hi Tek scratch: \"Stay awake to the ways of the world\"*}  [Talib Kweli] I need you all to be clear on exactly what I\'m sayin With your attention span I understand that I ain\'t playin You mistaken if you somehow think it\'s just me you facin Starin me down while your enemy is standin adjacent My heart is racin but I know just what I stand for We chasin death carelessly like +Jessica+, I +Care Moore+ Who said, \"Just because no one can understand how you speak don\'t necessarily mean that what you be sayin is deep\" In case you die in your sleep you ask the Lord for a blessin Sometimes they sneak up so quiet that the silence is deafenin You\'ll never know who the assassin is until it\'s your time to go Your life is flashin, askin for forgiveness but you move too slow Now the people that you love bear the pain that you once harbored You was livin for yourself so you could never be a martyr Life is hard, death is harder; you somebody baby father Someone\'s lover, son of your mother, somebody brother Somebody nigga, now your spirit in the air like a whisper Hearin your name mentioned when we pourin out some liquor The days go by quicker and the nights don\'t seem to differ It\'s gettin cold, so I shivered and asked my soul to be delivered  [Hook: Talib Kweli] Good mourning, good afternoon, good night What have you done with your life? Everybody time comes to be embraced by the light You only scared to die when you ain\'t livin right, man I\'m puttin up a hellafied fight {*Hi Tek scratch: \"Stay awake to the ways of the world\"*}  [Talib Kweli] Yo, the time come for everybody.. It ain\'t somethin you can really prepare for  Yo, yo, Mad Duke, rock rock on and Curtis Mayfield, rock rock on and Grover Washington, rock rock on and My Aunt Hazel, rock rock on and Big L, rock rock on and Freaky Tah rock rock on and Jerome Green, rock rock on and Slang Ton, rock rock on and We celebrate life.. {*echoes*}   ^ thats that straight real shit, no talk about spinners and fuckin bitches in a club.   more lif and ak greatness  Inhuman Capabilities Lyrics by Akrobatik  feat. Mr.Lif   Akrobatik [Mr.Lif in brackets] Akrobatik brings the closest thing to lyrical perfection When I extent my tecs send spiritual conections Never catch me lounging in the ignorant section Kill emcees for my profession, with the lyrical [invection?] injection [Does akrobatik burn?] No question! I\'m vexed trends got ladies drooling over the next men But I flex pens and bench-press thoughts to build my strength Cuz the battle will be long and I plan to go the length You can search latitude and longitude and never find a stronger dude My attitudes define emcees to prey on for food My minds like hannibal lectur\'s Cuz when it comes to eating niggas up I give cannibals lectures [It\'s time to peep some artists of a different texture] Uplifting the motivated, mentally slaying the extras [They look gorgous as soon as rigor mortis sets in. This is when Lif steps in. Just in time to bust another lyrical headspin. and windmill... Hes been ill. since way back! He came into the studio laid back Then he proceded to blaze a hole in your track Hurry up and get sum medical aid, the beat starting to fade This dj\'s cutting you on a blade And if you reach for the board your fingers sliced the niggas nice. You got caught up in a lyrical heist And got physically diced You can try to splice him back together] Will he be the same? [Never]  -scratching- Inhuman capabilites Is the god human Inhuman capabilities We the inhuman  Yo! We stand miles above ur highest point 93 million styles above ur flyest joint  Welcome to the brain you about to get the grand tour Bringing rappers to they demise like Gigantor Speaking of demises me and Lif about to plan yours The incompetent we stomping them at the damn door [Cuz he\'ll never comprehend what we stand for We\'ll f**k up your Air Force and start up a land war. Running down the line A to Z like transor Go the whole line like 5 and 4 Combined after the poeple signed I heard the people find me...to be...mad shockin when im rocking Check how our voices are] Interlocking! [To crush the mental toxin, thats poisening] You and yuor offspring [So were launching] Bombs! [In fact, a full scale f**king attack] On anything wack [Ak hold back] Naw never that! [I\'ll sever that thought completely] You are now getting sleepy [Hypnosis] Submerged [I\'m under] Implant elements [Lightening] Thunder [I\'m surfacing] Prepare to [Start murdering those verballize but really aint wise Give them several tries they still meet they demise Ak and Lif defy] Aries [Large] Got everything you need [Hammers] Claws [Stitches] Bars [Flamethrowers and saws] Lif and Akrobatik always blowing the spot [Cuz we do the type of shit the average human cannot] Span the globe on foot [In a day an a half] Go back in time and tell Confucious that [He don\'t know the math] Control the elements like a male X-man Storm with a staff Navigate the 7 seas on a white water raft Confiscate all your rounds even before you pull your gun out [Put some spit on your index finger] And put the sun out!    duo artist compilations like that kick serious ass, just like Cannibal Ox  http://www.seeklyrics.com/lyrics/Cannibal-Ox/  '),(800,2678611,'30,May,2004','2004-05-30 00:00:00','   godly lyrics anyone?  Mr. Lif - Avengers Lyrics   {Akrobatik + [Mr. Lif]} It\'s Akrobatik With the final fantasy tactics Eliminatin\' wack shit [With strategy that\'s sick] Teaming with my accomplice who helped me on the Fat Shit We\'ll f**k your tower up With multiple power-ups My appetite for pain will eat a meteor shower up Crash a thousand crews [Grab a bite] and get showered up  {Mr. Lif + [Akrobatik]} You feel refreshed Now let\'s continue the quest I use a towel to Catch innards of those we disembowel for Foul play [Lif what did Colin Powel say?] Who knows? I had flood waters wash his ass away He was just another rabble-rouser from Babel\'s tower Living in the glorious realms of false power  {Akrobatik + [Mr. Lif]} I make you shake ya ass I make you nod ya head Plus give you the motivation To get your ass out of bed Cause to me, MC means Motivate the Crowd And eliminate the evil cause I know they hate the proud I blow the fakers out of the box with Hip Hop [They oughtta stop or watch they frame get chopped Or better yet they suffer from paralysis and shot] So if you feeling this pump ya fist when it drops yo  [Chorus]: Mr. Lif + [Akrobatik] The avengers Sent to tear down the terrorists Perceptionists [Who\'s that?] Ak and Lif [And you know that we\'re Attacking with the raw for sure] The earth [yo check it] get another realm to explore  {Mr. Lif + [Akrobatik]} At this point the predator blinks Knowledge rifles past the head of the sphinx The nation cringes as a terrorist thinks How to spike his next target But we\'re all americans here so our place is the market   Death to whoever tries to harm it We\'ll demonize you Provide adversity you\'ll never rise through We upgrade to downsize you [Who says we can\'t buy you?] Here\'s a dollar that claims faith in Allah Currency currently funding a holy war You\'ll only get a little cut So you know there\'s more We assure you that we\'re heavily in debt [I\'m sick] Sorry we don\'t have the remedy yet  {Akrobatik + [Mr. Lif]} Yo I shower fat freeze On the powers that be So when they come to me with bullshit I devour that steeze Like tryin to charge a hundred grand for Bachelor\'s degrees And creating filthy air that be stifling the breeze And filling our TV\'s With these gold tooth wearing MC\'s Who probably don\'t even know their ABC\'s [But got cheese] But we all know that cheese goes bad and gets molded So my priority is wisdom for my head to hold it  {Mr. Lif + [Akrobatik]} The wicked get scolded [Their soul] They sold it They got a million ways to make sure that we\'re molded Along the same lines as those with no spines But the power of these rhymes is changing the times  [Chorus]: Akrobatik + [Mr. Lif] The avengers Sent to tear down the terrorists Perceptionists [Who\'s that?] Ak and Lif [And you know that we\'re Attacking with the raw for sure] The earth get another realm to explore  [Chorus]: Mr. Lif + [Akrobatik] The avengers Sent to tear down the terrorists Perceptionists [Who\'s that?] Ak and Lif [And you know that we\'re Attacking with the raw for sure] The earth get another realm to explore  '),(801,2678611,'30,May,2004','2004-05-30 00:00:00','   \"Gotta be somethin for me to write this\"   If Shakespere was alive today he would be a rapper.   i cant think of anything to post, my mind is dead  '),(802,1234212,'07,April,2003','2003-04-07 00:00:00','   blaa blaa blaa  just in a need to sleep mood.  Last night I tried to download the shitload of disks for RedHat...and that kinda failed...I have .NET Server 2003 RC2....so screw it I will tinker with that...or Win2k for my serving needs...   anyways I just need sleep...  will make a longer post tomorrow  '),(803,1234212,'07,April,2003','2003-04-07 00:00:00','   Oh by the way....I am going to download Red Hat 8.0 and install it on my spare computer...for use as a simple server...I want to host my own shit again...those were the days...  '),(804,1234212,'07,April,2003','2003-04-07 00:00:00','   Today was a pretty good day...  I got a date....I am happy.  I found out that I know more advanced shit in Word 97, Excel 97, and Access 97 than I know basic skills...it should be the other way around...  I found out that my car is being a whore! I think I am going to need to replace the fuel pump...  I found out that Area 51 is indeed still active... www.desertsecrets.com  '),(805,1234212,'07,April,2003','2003-04-07 00:00:00','   Well this is the modest start of my blog...or restart...I have had several blogs...one which was hosted on my own server...then one on /. Now I have one here...maybe I will keep it up to date.  My purpose is simple...to let people into my life...seeing that I am pretty reclusive as it is...  So lets start this off with a little joke I heard...which made me laugh my ass off.  \"So I am dating this homeless chick right now...now don\'t worry...really it is great, because after we have sex and she needs to get going...I can just drop her off anywhere.\"  Which brings me to this point. As probably many people out there in cyber land...or at least most of the geeks out there...I have no female. I did have one...and it was a relatively long relationship...but things obviously didn\'t work out. I\'m not mad at her...I am dissapointed though because just after we broke up she finally developed a sex drive...which would have been nice while we were dating...but hey not everything can go well.  Now I am looking for a female to let into my life...or my life as such...I do have one that I am interested in...and we are going out to see Jim Breuer on Sat. evening. I am looking forward to it because I have known her for quite some time now...and we are pretty compatible. At least I think we are. So hopefully I will have a female to call my own.  Speaking on females...can anyone tell me why females always seem to gravitate towards males that are just not right for them...or are asses, dickheads, or all around pricks...I have several of these types in my life and it always confuses me. I am not a bad looking guy...I\'m reasonably intelligent (at least by IQ standards) and I always treat women right...but when it comes down to it I always fall in that friend catagory. Guys you know what I am talking about here.  It just baffles me...I will obviously write more on this topic...because I think that it needs an analytical approach to figure it out...and I am just that ass to do it.  Till next time...  The world sux and there is nothing that can be done to fix it...so sit back...pop some pills...and watch the pretty colours.  '),(806,1234212,'07,April,2003','2003-04-07 00:00:00','   this has been a rough few days...family stuff...things I won\'t go into right now.  I would love to rant right now...cause god only knows I have a shit load to rant about...especially after these past few days...  oh and from now on...all my references to \"god\" (personal ones, not quotes) will be changed too blue glowy thing...cause for all we know...that is god...oops I mean blue glowy thing  I am just a tired puppy...lonely and looking for some cuddling...anyone want to cuddle me?  '),(807,1234212,'07,April,2003','2003-04-07 00:00:00','   what shall I say  This weekend I had a good time with my date...Jonelle...  We ate, we laughed, we drank...fun was had by all  I am still wondering what I will do for a server...I need some RAM for my little server box...I have a gig floating around in Indiana somewhere...I need to get at least 512 of that back to pop into the little server...and then I will be golden...right now it has 160mb in ram and I just don\'t think it will cut it...  Anyways...I am off...I will update later :-)  '),(808,1234212,'07,April,2003','2003-04-07 00:00:00','   today has been a shit day   I have had to reinstall XP and everything...I hosed my box trying to move to Longhorn...that\'ll teach me...heehhe  Well off to do more installing  '),(809,1234212,'07,April,2003','2003-04-07 00:00:00','   hey its been a few days...  nothing really to report that is why I don\'t update as often as I would like to.  I did go to the Dentist again today...that is becoming one of my favorite places...hehehe Just call me Mr. Swisscheese Teeth...I have more fillings in there than I thought metal could go in the mouth...so I think I will chew on some Aluminum Foil later this evening for shits and giggles...  hopefully, blue glowy thing willing, I will have a job soon...then something interesting in my life might actually happen...until then I will just have to be happy downloading porn at insanely fast speeds, and drinking lots of fluids....  \"My Idea of work is masterbating.\" \"My Idea of hard work is masterbation that takes more than five minutes.\" \"My Idea of being creative is masterbation without porn.\" \"My Idea of co-ordination is masterbating while having your hand on the mouse and getting that all important put the dick away/close website/not look like you are doing anything look down to a millisecond affair.\" \"I have now resorted back to plain old TV for masterbation...no longer do I need a blond with her ankles at her neck screeming \"harder you stud\"...that is my pitiful life.\"  Quotes by me.  '),(810,1234212,'07,April,2003','2003-04-07 00:00:00','   oh yeah and my car is a fucking whore...it is breaking i think....after over 600 dollars in repairs it is breaking...and it is the same problem as before...so the mechanics fucked it up royally...now they will have to fix it properly at a discount...because its the same problem...  I hate my life...and I can\'t wait to see what next month will bring...if it is what I think it is...then this blog will not be updated for a long long time  '),(811,1234212,'07,April,2003','2003-04-07 00:00:00','   Today my mom and I went to see  Tear of The Sun  or whatever that movie is...with Bruce Willis...now I love to watch Bruce...I think that he has had some fine acting...of course Sixth Sense and 12 Monkeys...and a few others...but this movie was just kinda a movie I expected more out of. It was kinda cool...but certainly not worth 8 bucks to see. If it had been me I would have drugged the people needing to be extracted and got the fuck out...  outside of that...it is a boring night...  '),(812,1234212,'07,April,2003','2003-04-07 00:00:00','   I have been in a pit of despair as of late. A lot of shit things have been happening and all of them were of my own doing. I don\'t think at this point I will make it the rest of the year...and frankly I could care less.  No I\'m not gonna off myself or any stupid shit like that...instead I think that I will br off\'d by our government...if they get their way I won\'t need to be on their rolls anymore...which is good cause I am kinda tired of my life anyway...why should I waste the oxygen  '),(813,1234212,'07,April,2003','2003-04-07 00:00:00','   today I should have gotten my car back...but nooooo! why...cause apparently they have trouble with cars that have electronics in them.  my question is...why are they having such a freakin hard time replacing the fuel pump...you find it...unplug and remove the old one...put the new one in and plug it in...then prime the system...bingo its done...or maybe it might be that they had the thing misdiagnosed from the get go and all it may be is a faulty fuel pressure regulator that costs around 40 bucks to replace and I can do that myself...if it has been misdiagnosed...then I will be demanding a rather large refund...if they couldn\'t fix it in the first place they should have just told me so...I would have been happy to take it to another shop.  sometimes I just want to cry...and today has been one of those days  '),(814,1234212,'07,April,2003','2003-04-07 00:00:00','   today is no better...yesterday was far worse...and tomorrow will be the worst it has been in a long time.  that is my motto  '),(815,1234212,'07,April,2003','2003-04-07 00:00:00','   well i moved into the back room today...and after I got in here my grandmother got pissed off and wants me to move back into the middle room...  the hell I live in is a very complicated one. You see everything revolves around her and how she percieves things. I can answer a simple question with a yes or no and that would be enough to set her off in a snit. Now mind you I am a prick bastard...I know this...so when I am not in the mood to deal with any shit I make myself scarce and when I do have to speak it is simple and to the point...but it is rooted in the fact that I cannot escape doing something wrong even when I do nothing at all.  Lately she hasn\'t been feeling well. You see she had colon cancer something like 5 years ago...had a huge tumor removed...and it was very serious...any longer of a wait and she would have most likely died because of it. However the cancer could have been diagnosed a lot earlier if she would have gone to the doctor. She had been having gut pain for something like a year and a half...then she went. So hearing she had cancer in the first place really didn\'t shock me...really it didn\'t shock anyone...we were worried...but not shocked. Cancer is more or less the flu...it strikes a lot of people and kills a lot of people...but if properly treated with the right medicine it doesn\'t have to kill you...but to cancer \"victims\" (not to belitle anyone who has or who has had cancer...but they all think it is the worst thing you can get) it is some sort of horrific happening. My mom told me today that it is maybe the fear that they will linger and suffer with pain...well my answer back is they have wonderful pain meds...and if that doesn\'t work...a bottle of drano or a gun will (suicide is never the solution...but stop being such babies about dieing and face it with dignity).   Back to the story...so you see for the past five years she has been this \"victim\" or \"survivor\"...every conversation she has with someone who may not know (which there are few that don\'t by now) goes something like this...\"well I had cancer...it was nearly N-Stage....I still have pain but I manage...and I watch my cancer markers...plus I have a restricted diet\"...now she says it out of the need for someone to be shocked that she had something happen to her...she has to be a victim of something or other and has to have people in a tizzy over her whatever it may be... Most people simply go \"sorry...glad your better\" (while biting their tongue)...but some indulge her and because some people still indulge her it feeds this vicous cycle of telling more people. A friend of the family had cancer...I didn\'t know it...because she didn\'t tell anyone.   Because she craves such attention all the time there is only so much you can handle...even mother teresa would tell her to blow it out her ass. So I don\'t indulge it...that makes her mad. Now she went in for her monthly blood test and her markers were up...and yes they are at the \"you should be concerned\" level...so the doctor ordered a PET Scan (Positron Emitions Topography Scan...or something like that...they inject you with a radioactive dye and give you a full body CAT scan...close enough description) well they had originally scheduled it at this one place but the earliest they could get her in would be towards the end of the month. Then she heard about another place that could do it and they would be able to take her earlier (last Friday)...well for almost a week and a half all I heard about was this bloody test and how the two different places have two different types of prep (I told her it may be one place has a better Technician who will read the scan and may need less prep or one machine is newer than the other...but not to worry)...well it turned out she went to a newer machine so the prep was less. Now for the past few years she has had pain in her stomach and that is what she was worried about...but then she develops a pain in her shoulder...all this weekend she has been laying down and complaining about being in pain...which ok I am a pussy about certain pains (when my body doesn\'t want to work right and gives me hell...I lay around too...but at least I don\'t make noises for attention as she does...how do I know its for attention because she doesn\'t make them when other aren\'t in the room...I have sat in the next room and listened for them)...well she got the results of her test today (tuesday) and they found some sort of mass in her shoulder...now they have to do a CAT scan on that area to pinpoint what it is...so as a result she has been a real pain for the past few hours...very \"look at me the sickly old woman I need to have all attention payed to me\" role...now I have run out of patience...the only thing I am going on is the fact that I live here rent free...and really don\'t have to pay bills and she helps me a lot (and I mean a lot)  My whole bitch is this...for those who know me they kinda know what it is...for those who don\'t...I have Cystic Fibrosis. It is a genetic disease and it can get quit nasty and will eventually kill me. Now I have no problems with this...I have resigned myself to the idea...I am so comfortable with the idea that I invite it and embrace it (I am a little dark at this point in my life)...while most of the people around me seem to have a problem with accepting it...especially my family (the want to try and make me think I am normal now after having kinda pounded into me for years that I wasn\'t). Now maybe it is me...but I know I am going to die...and have a pretty good clue of how it will happen baring any accident or mishap. Being fairly young I should be pissed that I will not live long...but I am not...so why are so many old people or those who are \"victims\" of disease afraid to die? By \"victims\" I mean people getting cancer or coming down with menengitis or a myriad of other ailments that kill human animals. Death is yet another stage of life. For old people especially...why are they afraid? They have had a run (most good...some bad) so they have lived their lives. Just accept the fact you will die and nothing can be done with it. I can see giving a heart to a younger person or replacing a kidney in a 20 yr old...but in some 60 plus person...that is rediculous. A man who is 73 in the middle of renal failure and congestive heart failure (which my grandpa is dealing with...a whole other story) is hoping someone will give him a kidney...fuck that...I would give one to my sister...or Billy down the street if we were a match...but not to my grandpa or any older person who has lived their life.   Now I don\'t go around anouncing \"hi I\'m Mike...I have CF...it will kill me\" that is information that people really don\'t need to know...if they want to know and have found out and ask me...then yeah I will share...but my disease isn\'t me...I\'m not my disease. I want science to advance and cure diseases...I want them to use Stem Cells and do some cloning...because I know it will help future generations.  But...  People when your time is up...don\'t struggle or fight...just go on and die...it will be better anyway.  Well my rant is over...not very cohesive...but maybe you get the idea...  oh yeah my car is in the shop...fucking fuel pump  '),(816,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I went to that fucking job fair...it was worthless...why would I expect that I would be any different from the masses...a job fair for technical people and business persons really means Wal-Mart need associates and Notel needs some phone pimps...only two tech jobs there...  I mean really at this point I think I will get some rare form of cancer before I get a job...  oh and I got a hospital bill for almost a grand...and when it was supposed to be covered by insurance but they say I wasn\'t covered then...then when I tried to tell the billing person I spoke to that I had tertiary insurance (third level) and gave her the name of it she said it didnt exhist...and that if I wanted them to bill that insurance I would have to call the insurance company (which happens to be a state agency) and have them contact the hospital...well they ain\'t gonna do that...and the hospital won\'t call them to get the billing procedure on as how to bill them...so I say fuck it...they can\'t get blood from a stone  My grandmother is having her lung tumor removed later this morning...now I feel bad for her...but here is what she has done so far. Later this morning I have to have her there by 6:15 AM...ok I can do that...I would normally wait until they drug her and leave, come home, catch a couple of hours of sleep, and then when the hospital calls me and tells me she is out of surgery I would go and talk with the doc....no need for me to sit for five hours and wait....  So she planned to have a friend come up at 7:30 and I am going to breakfast with her...and then we will come back and sit and wait...when I elluded that I wanted to come home and get a little sleep that made her mad...and she then said well if you don\'t want to take me then my friend will take me...I never said I didn\'t want to take her...shit I could care less...but I don\'t want to wait with my thumb up my ass in an uncomfortable waiting room...  besides when she comes out of surgery I wouldn\'t be allowed to see her anyway...and will probably be restricted...cause I went to the docs the other day and found out I have a serious lung infection...serious enough that they would normally have pulled me into the hospital but seeing that I have become allergic to pennicilian based drugs dragging me in would be the last thing they would do...cause while I was in they would desensitize me to Pennicillian...or at least try to...and that could take a long time. So with this serious lung infection I have been put on the last oral antibiotics that my bugs respond to. After the bugs become resistant to them...then I get to head to the hospital every time I get sick for major IV anti-biotics (which happen to be pennicillian based)...the doc told me point blanc that he doesn\'t have many options left to treat me...and that once my bugs get resistant over the next few months it will come down to how fast they can get me non-alergic to pennicillian to treat me...in other words I might have but a few months of relative health left...  Now I could have let him go ahead and pull me in now and start the process...but I didn\'t because my grandmother needs me around...if I don\'t start seeing results in the next couple of weeks from the anti-biotics then I will be in the hospital...  I am glad of all of this...to die will be nice...peacefull and less stressfull than my life as such...  Yesterday was bad, today was worse, and tomorrow will be harder than today.  '),(817,1234212,'07,April,2003','2003-04-07 00:00:00','   well later this morning I have to get up and go to a job fair...I hope it goes well...but before I do I must write a cover letter with additional experience I have including some of the rudementary programming experience I have...grrrr  I am tired and today has been long and well just plain bad  nothing bad happened today as such...but it is still a bad day...hopefully soon I will have a job...and then I can have bad days at work and then come home and have a reason to drink  '),(818,1234212,'07,April,2003','2003-04-07 00:00:00','   today I got a new keyboard...and I like it. It is the microsoft multimedia keyboard...it is pretty jizm...now I just need to mod the LEDs to a blue colour and all will be well  why did I get a new keyboard you ask...well I wore the old one out...actually I wore two of them out...I really dont know how I managed to do that...but I did...the one (MS Natural Pro) was pretty good but the keystrokes were sticking and some keys failed to register when I would key them...and my logitech cordless keyboard just plain out stopped working (and yes I changed the batteries)...the next thing I have to get is a new mouse....I think I am going to get another microsoft mouse...the one I have my eye on is the Blue one...it looks pretty good...now I just have to find it at a reasonable price (I am using the intellimouse optical...) The current mouse I have hurts my hand after using it to long...I just need to try mice until I find something that will work  I also need to do some upgrading on my computer (actually get some parts...)  but all of this is superceded by the need for a job...and sanity  '),(819,1234212,'07,April,2003','2003-04-07 00:00:00','   well I got my car back the other day and something was already wrong with it...which makes me mad. It seems that they might have engaged in a little bit of sabotage. When they dropped the fuel tank (which they did not have to do) they removed part of the exhaust system...well when they put it all back together they put a shield on improperly and that little shield was jammed behind a sensor. The way it was jammed would eventually cause the sensor to break off...now I am sure this little sensor would be expensive...especially seeing the way it would break off requiring the replacement of most of the exhaust system. They also let down one of the mufler retaining systems...and when they put it back they only screwed it in maybe two turns...so it was making a racket and would have eventually fallen off allowing the muffler to hit the ground.  Well needless to say I fixed it all. I am not all that pleased...and when I confronted them about it they only said \"wow you have had nothing but problems with that car\"...well it seems that more of my problems have been caused by them...I don\'t think I will be taking the car back there.  My grandmother has a tumor in her lung and it has to come out...and then they will know whether it is cancerous or not. They seem to have to remove the entire lobe. So over the next month my normal hell will go into overdrive...I just hope they give her good drugs to keep her out of pain...and I hope she does alright...  Well off to look for a job  '),(820,1234212,'07,April,2003','2003-04-07 00:00:00','   today I got out of the house for an hour...it was so nice.  My brain is gew, my guts are smart, my ass is hairy...my penis needs lovin  I want to be an angry comedian...my mom says that I have a very angry informed comedy about politics...that my sense of humor is very strange and without any moral or politically correct compass...and that the voices in my head really are only meant to be heard by me.  You see I get pictures in my head...now you say \"everyone gets pictures in their heads\"....well I think mine are a tad different. I dream in colour. I daydream all the time. I can have an entire seven year relationship with a banana in five seconds...the things that are funny to me are funny. I love humor that makes fun of people and disabilities. There is nothing funnier to me that a person with some twitchy disease like Timmy on Southpark belting out the National Anthem on a computer and me envisioning the batteries going dead and the machine kinda dying off...I find that hilarious...  I have a disability...no I\'m not retarded...although I think retards are funny. I have no problems with people making fun of the fact that I am a gimp...  Sometimes I do feel bad about my sense of humor because I do realize that some people couldn\'t take it...but at the same time I feel that their is nothing sacred in this forsaken world...so make fun of everything and everyone...  If you can\'t laugh then you can\'t live...and lets face it...some shit is just funny  '),(821,1234212,'07,April,2003','2003-04-07 00:00:00','   I am just trying not to get to depressed right now  The reason is because I am pretty ill and I have to take care of someone else. The doc didn\'t want me to come in. I see his reasoning...there is nothing he can really do that I can\'t already do here at home. That doesn\'t comfort me in any way. I just want to be healthy...but I am not afforded that luxury  The one thing that really sux ass is that in addition to the coughing up blood thing...I am in pain. My left leg has kinda gone numb...which sux...my lower back on the left side hurts like a bitch...my headaches are more intense and frequent...and my chest feels like it has been hit repeatedly by a baseball bat which makes coughing the shit out of my lungs a tad difficult.  Unfortunately the doctor won\'t prescribe anything for pain...which also sux ass. However if I had something to take the edge off I would be even lazier than now...cause I would just want to sit and enjoy the good numbness that pain meds would bring.  Maybe soon they will give me something that will let me get back to something that remotely resembles life...but until then I will soldier it out like I always have...with a grin and a \"I\'m fine\" to the family...knowing that in the end the pain will stop and all will be well.  I wish I were an optimist...because if I were it would mean that I live in a world of denial and stupidity...stupid people are happy...smart people are forced to live in reality and suffer for having to deal with the stupid people and their happiness  '),(822,1234212,'07,April,2003','2003-04-07 00:00:00','   hey everyone....tried to post a sec ago  granny may be home tomorrow...  started coughing up blood today. my chest hurts and I am bleeding in my lung...this sucks ass...I guess the infection is getting worse...if I am still coughing blood up tomorrow then I will call the doc...  I hate this all  '),(823,1234212,'07,April,2003','2003-04-07 00:00:00','   hey everyone...  granny should be coming home tomorrow...  today I started coughing up mucho blood....that isnt kewl in the least...that makes me a tad worried...the infection must definitely be worse....and now on top of it my chest hurts pretty badly  Oh I love being sick  '),(824,1234212,'07,April,2003','2003-04-07 00:00:00','   Working on a presentation...and it sux ass...I don\'t think it will turn out good...but then again I could be surprised...I just have to do it...  ok I will start on it in the morning  oh yeah I am still hella horny....anyone wanna bone?  '),(825,1234212,'07,April,2003','2003-04-07 00:00:00','   its been a long couple of days...  I just need laid to get the stress out....  any takers?  '),(826,1234212,'07,April,2003','2003-04-07 00:00:00','   today was a long ass day...and I still need to get some sleep.  I went to the Doctors and I have to go into the hospital tomorrow...which sux major ass...but hey I will look at it as a vacation. I just hope that I get a hot nurse or something.  I got my laptop up and running once again. I don\'t think that I will be able to post while there...but I will try...  so till I get out of there I bid you all a fond fairwell...  '),(827,1234212,'07,April,2003','2003-04-07 00:00:00','   the meeting went well...it was like a job interview...only for jail...long story that I will go into only when I have finished up with the whole procedure.  I had to give a hundred bucks to the court today...that sucked...but hey it is what happens when you run afoul of the law.  I rented and watched One Hour Photo...it was well...interesting and sucky all in one...I love Robin Williams...he could take a shit on film (which he did in this movie) and I would find it entertaining...all I can really say about this movie is that it was disturbing...and makes me glad I have a digital camera.  Later this morning I have a phone interview for a Job...I really need something to come along...cause I need a job...hopefully it will go well and I will get something through these people. That would be money yo.  Then I have a lunch date with a friend Amanda...then a Doctors appointment at which they will tell me whether or not I should feel good...I don\'t, but I have to wait for definite word as to whether or not I am supposed to feel like shit.  Well I am going to get a little shut eye...peace  '),(828,1234212,'07,April,2003','2003-04-07 00:00:00','   Today was a busy beaver day.  Lets see...first after I got up I made myself three scrambled eggs...now the only way to make scrambled eggs is with milk mixed in...so I did that and they were great.  After that I did a couple of breathing treatments and spanked the monkey.  Then I went out. I went and got some dog chews for Rufus as well as a dog tag with his name and my number on it...and I got him doggy breath fresheners...cause he has stank nasty breath. Maybe it is from him licking his twig and berry (he has only one dropped ball)...or maybe his ass...but it stinks, his breath I mean.  I also got my snake a mouse to eat and fed it to him...that is always kewl to see...my snake eating another living critter.  I got my hair cut...it is a little shorter than I asked for...but hey hair grows...and before long I will look like a shaggy dog again.  Also I had the oil in my car changed...that was needed badly seeing as it was way way overdue....  Not much else happened today...  Oh my grandma and I watched the Red Dragon...the Prequil to Silence of The Lambs...well I have now seen all three movies and have been totally unimpressed with all of them...ok Silence was creepy...the other one was gross (where they ate the brain soup)...and Red Dragon had only one saving grace and that was Edward Norton...and even with him in...the damn thing was just plain boring.  Well I am off to bed now...I have a meeting at 1:30 this afternoon...and I don\'t want to be to shot for it...so I must sleep some....  Peace ya\'ll  This me tryin to cyber...well not really me...but close enough  urlLink Cybersex h0n3z Style  '),(829,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I was sprung on Monday...but I still feel crappy. I had a port put in...it looks like ass and feels like a rod in my chest...hurts to boot.  Looking around at people\'s IM info has just made me really depressed. A lot of the people that I met when they were freshman are graduating...and I have yet to achieve that goal. I suck. I have no job. I am sick. I am pretty reclusive. I am also a felon now...which really sucks balls. More on that when all is said and done. I don\'t know where to go or what to do. I am at a complete loss as to what I can achieve now. Really all I want to do is crawl in a hole and die...and that may happen soon enough given my poor state of health...but what have I achieved while on this planet.   Not a whole hell of a lot. I had hopes and dreams as most do. I wanted to go places and see things. I just wanted to be happy, and I would still love for that to happen, but it has yet to happen, and the more I go on the more I feel that it simply won\'t happen.   I feel completely disenfranchised by the whole deal. We have all be raised to be rock star, sport god, multi millionaires...yet no one really ever stopped to say. Hey get real buddy...you are like 99% of the rest of the world...you will be poor and live what will amount to being an unhappy un-fulfilled life. Yet people keep buying into the dream. People keep on thinking that :in America you can do anything, be a millionaire, and have freedoms only others dream about\".  Well I am sorry to say that in America we are all running towards the same goal. Money and Power. Yet the simple fact is...there isn\'t enough money or power to go around. The money and power lies with but a few select people in our country...sure you maybe can achieve that through our system...but most likely you will fail and only end up afowl of the law if you try to achieve those things with any great vigor.   America is but a dream. It isn\'t real. What we are fed on TV is what the people in power want us to see. It has nothing to do with the common Joe (hell even they took that from us Joe Millionaire). Whenever there is a good idea it is almost always be corrupted by big business and stolen from the average man.   The one thing that distiguishes a citizen from a non-entity in this country is a Vote. As a felon I no longer have the \"right\" to vote. It was removed from me permanantly. Chances are I won\'t be able to get a Visa so that I can leave this country and start anew somewhere else. As a non-entity trapped in this nation I am essentially a slave to the state. I no longer have the right to vote in elections and help determin the way I think the country is run...but I still have to pay for it through taxes (which makes me even madder).   So fuck America (land of the free indeed) Fuck George Bush and his corrupt cabinet and little lackies. Fuck the Legal System which puts many under and essentially kills their voice. Fuck everything. It has all been a lie.  '),(830,1234212,'07,April,2003','2003-04-07 00:00:00','   I got my scanner working...now I have a few things to post on Ebay...some sports cards...one is worth over 5 grand...hoping to get 4700 or more out of it...we shall see...this may be a way for me to make some money...all in all I have around 100 grand worth of crap to sell online...all in sports cards...I would get a percentage of the proceeds which I need...cause I am poor  nothing much else went on today...I had my access needle replaced on my port...when the nurse jammed it in I came very close to passing out...the pain was that horrendous...arrghhhh. They say the pain will subside with time...well I am impatient and it hurts like a fucker...I hope it stops hurting soon...I am glad I have some access to pain meds...otherwise I would be gouging out my eyes to get relief...I will be getting a small prescription of Lortabs to help out in the next day or so...thank blue glowy orb...  '),(831,1234212,'07,April,2003','2003-04-07 00:00:00','   So I got my camera working on my system and I went looking for an online camera site...but alas I haven\'t really found a free site yet...maybe soon...I want to broadcast my sad sad life  '),(832,1234212,'07,April,2003','2003-04-07 00:00:00','   just chillin like a villian...  not much to report...the needle comes out of the port later this morning...that will make me happy cause I will be able to sleep however I want without my body resting on a needle jabbed into my small boob...  got the pain meds...they help and will help with the rest of the healing...that is good...they aren\'t lortabs...they are vicodan...same thing different name  lots to do this week...found a possible job...only pays 7.50 an hour...but it would be something interesting...assembling circuit boards...something different and the place is only five minutes from my house...so hopefully I may have a job. It will be good to get working even if it doesn\'t pay much.  also have a student lesson this week...20 bucks for an hour...not bad at all...teaching email basics and such.  baaaa  '),(833,1234212,'07,April,2003','2003-04-07 00:00:00','   wow two days in a row...  last night after ranting about voting rights and gun rights as a felon I found some interesting information  I will be allowed to vote once I have finished probation/parole/whatever...only seven states take away the right permanantly....thank blue glowy orb I live in one of the others that you get the right back....so at least I will be able to vote again  As for the right to buy and own a weapon that issue is a little cloudy. The ATF has the authority to restore that right on a case by case basis...however the congress has tried to stop that and the Supreme Court has said that the congress can\'t do that...but there is no real clear way yet for me to own a weapon after this crap is all done with.  Also I can get a passport...as the INS and various agencies that administer that program review on a case by case basis as to whether or not to give out passports to felons...so I will more than likely be able to travel one day...  Basically I will recieve most of my rights back...except that I will have a mark on my record and it will hinder me from getting security clearances...but that can be dealt with as well.  So all in all last night I found out good news...but still I want to own a gun...I have owned guns for some time now...and have been an avid weapons person...as I know a lot about them and love to go to the shooting range and kill imaginary bad guys...  Ahh well that fight will be later in the life...when all is said and done.  '),(834,1234212,'07,April,2003','2003-04-07 00:00:00','   Well a milestone has been reached...as of about 20 minutes ago I turned 26. Why is this a milestone...well lemme tell ya. For the past 26 years I have only been able to \"see\" or \"foresee\" my life up to this point. I simply have not been able to see past this point. I can remember when I was much much younger being able to see things in my future from the most mundane of things as eating a hot dog at Disney World (which I did at the very place I saw in my dreams...which is very weird as have not ever being to Disney before) to more important things as the first person I would \"fall in love\" with.  I thought my life would end at 26...I would no longer exist. Honestly in some ways that has very much come true.  Here is what has been going on for the last six months.   In December I recieved an official mailing from the Eastern District court of PA for Federal court. For you see back in 2000 I was involved in what is called computer fraud...however it is prosecuted under the title mail fraud. I knew better yet I chose poorly. I have no one to blame but myself for the trouble I got into.  Flashback to that period in my life. I went down to a suburb in Pittsburgh to pick up a shipment of fraudulently obtained computer hardware. After myself and my girlfriend at the time and another person got there and loaded all this crap into my car out of nowhere came the FBI and Postal Police with guns drawn and what seemed to me at the time as a wish that I do something stupid and allow them to pump some rounds into my body. At that time I made a statement and turned over evidence of the crimes that had occured. I was told that something would happen but I did not know what would go down. They released me that evening and I went back to my dorm room and cried and worried. Fortunately for me at the time I had someone in my life that helped me get through the initial shock. I didn\'t tell my parents, I didn\'t tell anyone in my family. I kept quiet about it all and pretty much put it in the back of my mind with the knowledge that I probably wouldn\'t be punished to harshly and if all else failed I wouldn\'t be around for much longer seeing that I couldn\'t \"see\" past 25.  Forward to December of 2002 I recieve the letter...or better yet the endictment of the crimes involved. Well of course being the government they made the charges out to be much worse that what actually happened making the crime seem like I had killed someone and that I should go to jail without even so much as a trial.   I obtained a court appointed attorney which has been very good. He has literaly told me not to worry...not because he would get me off, but in the sense of \"you didn\'t kill anyone, really this is a blue collar crime\" sense. He has not made me any promises outside of that he would be the best advocate he can be...and he certainly has been.  He got the inital 11 counts down to 1 count through simply contesting the counts in the idictment. The remaining count is a charge of Mail Fraud which I have pled guilty to. It is a felony and I am now a felon.   While this whole ordeal is far from over...as I am still looking at possibly having to go to jail, paying up to 30 grand in fines and restitution to the company of up to 96 grand as well as up to 5 year probation. I have come to some resolution in my life.   With the way we are arguing...and rightfully so seeing that I have Cystic Fibrosis and I am quite expensive I most likely will not see jail time. There are some guidlines on the books about how if it is considerably cheaper to keep an offender out of the klink, then that step should be taken as opposed to burdening the system with another inmate. Looking at it from this point over the last 8 months I have racked up approximately 120 grand worth of health care costs I don\'t think they will put me in jail seeing that in that environment I will most likely become even more ill than I am.  I have mixed emotions about this all. I have never been one to use my illness as an excuse for anything. I don\'t want to be known for that. The only thing I have used it for has been for federal benefits such as insurance so that I can at least have health care. Barring that I have no gone around saying \"I have CF treat me differently\". Hell even when in college I only used it as an excuse if I went into the hospital...not if I just skipped a class. If I was honestly out because I was ill I always provided proof that was the case. Even then it didn\'t always get me the results I had wanted.  Being a person with CF has really sucked ass overall. I wanted to go into the military. I wanted to be a doctor. I wanted to do a lot of things that CF simply will not allow for. Most of all I would love to gain about 40 pounds so that my 5\'9\" frame would look a little more natural...and instead of hovering around the 120 pound range I would be a more healthy 160 pound person.  As for the not seeing past 25 thing I am almost right...barring something happening in the next few months (and I am a pessimist so hell I may croak soon and with the way I feel it would be welcomed). I am already dead.  I am dead as a citizen of the US. I can no longer vote. I no longer have a voice in the government. I cannot run for an elected office. I am dead. For me the right to vote is what this country was founded on. The equal right to representation. Now we all know that there have been hickups in this idea (blacks, women, and initially those who didn\'t own land) it is still the foundation. I have voted in every national election since I turned 18 in 1995. I voted for Dole then Bush (which I regret by the way). While in MS I voted for Lott (he really isn\'t as bad as everyone makes him out to be). Hell I have put my contribution (albeit small) into the system, yet I will no longer be able to do that because of my conviction.  In the end I am left with a feeling of loss. I am left with a feeling of disalussionment. I am left with a feeling of complete and utter confussion.  I have confessed to my crime, I am willing to do the time if necesary. However I am not willing to give up my basic right to vote, or the rights afforded to me in the constitution after I have paid my debt to society. I am not violent, nor would I be, but this does bring up thoughts of unrest in my head. I have been stripped of all but the basic rights to exist and breath (which with CF have also been almost completely removed through disease).  What is my future? I don\'t think it will be much. To this point I haven\'t found my place. I haven\'t found lasting love. I haven\'t found happiness no matter what I have done (drugs, working out, trying to live right, trying to live wrong, god, anything). The only hope that I have at this point is that in my last days (which I feel are probably sooner rather than later) I will be comforted by a morphien drip or something along those lines, and by the knowledge that as I loose the ability to bring oxygen into my body that my brain will slow down and loose some of the potential that it once had. That when I do slow down I will be able to not worry about everything and all the wrong that is in the world (from what is wrong here in the US (and there is a lot of that) to the wrongs in the world).  The most comforting thought to me is this.  That in the end there will be silence.  '),(835,1234212,'07,April,2003','2003-04-07 00:00:00','   ok so I haven\'t updated in a few days...well there was a reason.  I was once again in the hospital...oh the lovely hospital...how can I explain my joy at visiting the place. The people are so lovely, the treatment is great...I just love it there...outside of being couped up it isn\'t so bad...  Next time I go I am going to make a concerted effort to get online service while there...I have DSL here at the house and I know that they have a dial-up backup for when it goes down...I just need to find out a number for it...and I also need to see what kind of restrictions are on it...cause I don\'t really want to have to pay a lot for that...if worse comes to worse I can get some sort of access for 10 bucks a month...at this rate it might be worth it...or I will get a wireless card for my computer (which actually makes more sense cause I have unlimited internet access on my SprintPCS phone and they have cards for that...I wonder if I can have a card and a phone on the same account without running up a huge ass bill). Another thing is my poor ole laptop is giving up the ghost yet again...the CD-Rom drive is just plain broke down...it doesn\'t want to even read clean scratchless disks...I guess it is about time...the thing is only 4 years old or something...a long life for a part like it. So I have to think about getting a new laptop...something that is certainly more powerfull than the one I have now...maybe close to my home system (which even now isn\'t anything to sneeze at)...I personally have my eye on a newer Sony Centrino Laptop...and with a Sprint card that would work great for a few years...but alas I need to make money first...which I am trying to do...  The only reason that I would want to go to a slightly higher model laptop is because it will last longer...it has better video subsystem on it that would allow me to play some SOF2 or something while in the hospital...in other words I could game a little...I could opt out and get a whole hog gaming laptop and spend 3 grand...but that is rediculious for something that will be outdated in a few months...the Sony I have my eye on is pretty fast...has a decent display and bits to run it (good enough for the basic gaming that I do) and plenty of hard drive space...plus a DVD/CD-R drive...so I can watch movies and burn CD\'s...plus the battery life is up to 8 hours...and it has a lot of other built in goodies...I would just have to find a good deal on one...I find that I would do more if the system I had to plod along with me could do more...it is just old and worn out :-(  But that is in the future...  Again I was in the hospital...only out a record 2 weeks...and back in...it only took 10 days to get something back in my lungs...ohhhh I love having little critters in my chest...they bring me such joy. I was in only 7 days...so that wasn\'t so bad...  All in all I like going there...mostly because almost all of the people I have dealt with there have been great. From all the nurses aids (John, Katy, and others...although those two are the most amusing)...the nurses have been great...Andy sticks out in my mind...nicest guy...think of very tall guy with very germanic angular features...someone that you might not want to meet in an alley under bad circumstances...but damn he is the nicest guy...really goes out of his way...  I also would like to think they like me as a patient...simply because I am autopilot really...hey bring in the meds (even if they need to be hung and run), leave some flushes...and bam just check in with me every once in a while and I am all good...and the food is pretty damn good overall...they have a few things that I wouldn\'t even feed to my dog...but for the most part it is good food...  Really I can\'t think of anything bad to say about the hospital outside of the 6 or 7 dollar a day charge for a TV and Phone...which the TV get only a few chanels...that needs to be fixed...but overall that is a great hospital and I am glad that I get to go there...one thing they could use would be some sort of diversion for the younger folks there...video games, VCR carts, something...but hey only in a perfect world (and in a perfect world they would have Wi-Fi and you could use a fat internet connection)...  Oh well...lots to do...lots to research and get ready for the innevitable return to Shadyside floor 6....it will happen...I know it...they know it...so at least we are all in the know :-)  Thanx to the people that read this...  I am still having problems posting my email for people to contact me on the site...so here it is again...as always any input would be great... m.p.honse@verizon.net  email me and we can start a pen pal relationship...or I could just verbally abuse you...or you me...anyway it will all be in good fun  '),(836,1234212,'07,April,2003','2003-04-07 00:00:00','   Tonight I watched a show again that I had seen a while back. It was on the Discovery chanel. It was on the differences between the sexes. When it got to one point in the show it made me feel really fucking jipped.  Ok so aparently those little things we call pherimones play some role in reproduction. Well we all know they do. Women and men are helped out by these chemicals as they signal compatitbility and whether or not you have a strong imune system.   Well I guess I am screwed in the sex game. My system is frail and weak and certainly not meant to go on. Pherimones may well be the reason that I have such poor luck with the ladies (as with some other things). Simply I signal to them that I am not good mating material. Who am I to argue with millions of years of evolution and science based fact.   As far back as I can remember I have never been good with the oposite sex...especially in trying to get it. Sure I have gotten lucky more than a few times...but that was usually through the excessive consumption of alcohol on both parties side. Women are just not attracted to me...or my smell. I don\'t smell funky except when I am on anti-biotics (which I currently am)..and I have always kept myself clean...so it must be my unoticed but important pherimone smell.   Because of my illness I send it out in chemicals that women pick up on...and while they may think I am nice, whitty, intelligent, maybe even cute...I am never the material for mating...  So where does this leave me...in the shitter...  I haven\'t had sex in quite some time...partly because of my lack of going out and being social...but I also think that my chemistry has something to do with it all. So who knows...  At this point I would love to get laid by a hottie...everyone wants too...but at the same time I really don\'t care if I get laid or not...I have a hand and a large collection of high quality porn...so I can keep the primal urges at bay. Which is another point...I don\'t have many primal urges...and that comes from my dad who has a very low libido...I mean he might as well be a monk...  Ahh well...thus is life.  '),(837,1234212,'07,April,2003','2003-04-07 00:00:00','   I went to see the Matrix Reloaded movie...ahhh that was a kick ass film...I have a feeling this year will be a panicopia of films. All sequels...but so far all of the sequels have kicked ass!  So lemme tell you about Reloaded. Reloaded and Revolutions were filmed at the same time for around 300 million...of that roughly half of the budget went into \"wow\" effects...in other words CG and Special Effects...and lemme tell you they certainly got their bang for the buck! It simply kicked much ass. Some people have complained that there is no story and that it is just a long action sequence...and well there is a story and it is bolstered by the action most definitely.   There is one scene in the movie where the \"evils\" of man is being explained and they flash Hitler, the A-Bomb and some other things...then they flash across the screen both of the President Bushs\'...to call them evil is going a little far...but I will say neither man has lived up to \"good\" standards...so I can see where they are coming from (even though I did vote for this Bush in the last election).  The story is simple...after thinking about ten minutes about it I got it. You have to wait till the end of the movie before they really \"bend your noodle\" but it is well worth the wait. So here is the story in a short and simple fashion.  Mankind goes in cycles. History goes in cycles. Everything that was will be again and everything that is will be destroyed. Throughout history mankind has built up many great civilizations. The Greeks, Romans, Egyptions, Syrians, English, Spanish, French, Russians, Japanese, and even Germans. They all built empires that eventually imploded and left the regions decimated and in tatters. People may say \"well the Americans haven\'t done that and we will not repeat the fate of those other great nations and civilizations.\" Well folks we will repeat it...over and over again. Sure once America per se is gone another great power in the world will grow and take over, but it too will fall in time. Humans are doomed to repeat history...its as simple as that.  However. There is one way that we can defeat history and move forward and continue on our journey through time. The only way to do that is to shed the old, get rid of the old way of doing things such as building wealth, building power, and rid ourselves of our fudal past. How you ask. First we need to stop looking at everything as a competition. We all compete on a daily basis. We compete for a better lane position at a red light, we compete for the hottest chick on the dancefloor, we compete for world resources. Sure competition drives technology and drives humans to do better things, but at the same time we take the competition to far in many cases. We compete out of fear and intolerance.   History has taught us that this competition only drives us to failure in the end because no one can have everything. Not everyone can have unlimited resources and expect the rest of the world or people of the world to go with this ideal willingly. We are in that same situation right now. America has become resource needy. We need massive amounts of everything to just survive as we see it. We need billions of barrels of oil to function anually. We need billions of metric tons of natural gas to heat our homes. We simply need to much to get by. Now if it were only us in the world that might not be so bad. But we are but a small portion of the worlds population yet we consume the most resources by far.   How can we stop it. How can we not repeat history. Well I really don\'t have all the answers, but I do know this. We need to be more tolerant of other peoples needs in the world. We are the richest nation on the Earth yet we have massive amounts of people without even basic health care coverage, yet don\'t seem to mind sending billions of dollars to Africa to \"stop\" AIDS. Is that a bad thing trying to stop AIDS in Africa...no it isn\'t, but the money would be better spent on developing cures and vaccines for the AIDS virus...then once we have those drugs we could easily erradicate AIDS everywhere. A worldwide innoculation campaign to stop it. We spend more money on learning about the male erection than we do on many diseases...in all our priorities are fucked up.  Another step would be to start giving services away. No I am not talking about making everything for free...that wouldn\'t work either because to many people would simply choose not to do anything and sooner or later that type of system would crash...but basic things like food, water, drugs, health care...make those things a non-issue...a happy well fed people are generally healthier, and less violent. If we fix things here first...and we make really significant changes here then we can export those changes and guess what...those people who have less then us all of a sudden don\'t have less then us. They have food, health care...they have the means in which to live.   We need to stomp out materialism. I know I know...I love my toys and gadgets and fancy things as much as everyone else. I love sitting in front of a computer with two Flat Panel Screens and having my desk cluttered with electronic gadgets...do I need them all. No. However they do make life more bearable. Does everyone need these things...yes. We all need some level of materialism. It helps us...makes us feel better about ourselves...are these things nesecary for life...nope...but they make life a little more fun.  We live in an unprecidented time folks. We have a lot of choices to make. We can choose to get along and bring everyone to a reasonable level of \"consumerism\" and \"materialism\" yet function as a planet. But that is a choice that has to made. It has to made by people, it has to made by the leaders.   We fight amongst ourselves like five years olds fighting over a swing. We can choose to make things better. We can choose to drive more efficient vehicles and demand that the government mandate that...we can choose that car companies make more efficient vehicles and move towards alternative fuels...we can choose that here because we have a system that is based on market demand. We can choose with our dollars.   We can choose with our votes politicians or people with better world views. People that see the problems and choose to say something about it and try to do something to correct them.   Face it folks if we don\'t choose to go down the path that hasn\'t been traveled before then we choose to repeat history. The old well worn path is easy to follow. It is the new untested path that we have never chosen to go down before that will lead to our salvation as a nation and a species. There are so many bigger problems out there. We could be hit by an asteroid tomorrow...we have no plans for that...hell we only scan 2 percent of the sky for rogue rocks headed our way. We will run out of oil soon. Our entire world that we know is oil based. Many synthetic products start their lives as oil. We can choose to move to alternative energy sources and spend billions on learning how to \"do\" fusion instead of building new nuclear weapons. All of these things we have the power to do. But because it is so much more difficult to do that than to keep the status quo we always choose the easier path.   And that my friends is the quandry we face. Do we choose to go the easy way. Or do we choose to go the harder way.   Red Pill or Blue Pill. The choice is yours.  '),(838,1234212,'07,April,2003','2003-04-07 00:00:00','   So today has not been a complete waste of a day...I was able to get a copy of T.a.T.u.\'s 200kph in the wrong direction...I like the CD...it is pretty bad ass...and you know I just love pop tarts...especially when they are foreign...cause with foreign chicks they don\'t speak english real well (along with myself...not) so a relationship would be fun...  Scenario: Chick: something spoken in a foreign tongue and then some broken english Me: \"shhh...don\'t ruin the moment\" as I put my finger over her lips in a Jim Carrey\'esk way (think Dumb and Dumber) Chick: \"awww...we really click\" Me: \"yeah babe...now shhh\"  heheheh I could see it happening...with any foreign hottie...like Shakira, DJ Rap (even though she speaks english), both the girls from TaTu...ahh that would be the life...  What I would do to those women...uuuhhh jizz...  I am finding places on my body that don\'t hurt as much when giving insulin...which is great...cause shots that you give yourself suck ass...but I will get used to it...also on the health front I go to the docs this Wednesday and he will tell me how I should feel...I don\'t feel bad...but I don\'t feel good. So we will see...I can tell now that my lung functions really haven\'t changed at all...they are still the same. When I go back to the hospital I am going to hunt down the physical fitness center and get in there and get my endurance up. When I was doing a lot of physical activity not all that long ago I felt better...and I know that helped my heart out a great deal giving it extra strength...now it feels weak...so I need to build it up. The only parts on me that work relatively well are my heart and my eyes...and with this diabetes crap I only hope that my eyesite doesn\'t go in the shitter...cause that has always been my strong point physically....like having excellent night vision and great vision overall (considering the rest of my family needs heavy prescriptive glasses)...  Other shit...well I have been doing some research for a friend on his Baseball card collection and it would seem he has a relative fortune in them...which is kewl because I know he will share some of the winnings from them with me for helping him out...  On a sad note that same friend is in a bad way...his grandmother is sick and isn\'t in great shape...now I don\'t pray often...but I did say a little prayer for him and her...having been sick and having had my own grandmother been sick (she had Colon Cancer five years back and just recently had a Tumor removed from her lung) I know how stressful it is...especially when you have a family that is going to end up fighting over the assets after the passing...families are supposed to stick together and get close in times like those...but for him it would seem that his family is already swooping in like vultures...as I know it will be here when my GM goes...(I have an uncle who can be an ass and will be when that time comes...when our Great Grandmother died he and his wife at the time swooped in and collected up some keepsakes that maybe shouldn\'t have gone to them at the time but that is over with)...  So all in all it has been eventfull few days...my mom and I went to Prospect PA to visit with old friends from way back in the day to have a picnic...it was very relaxing because I didn\'t have people breathing down my neck...and with me not being in the best of health it was needed...I really hate when people are harping over me when I feel like shit...ok I feel like shit just leave me alone please...I don\'t enough energy to deal with all that and heal...my energy supplies are very very limited...  Wednesday I am going to see The Matrix Reloaded if I don\'t have to head back to the hospital...and even if I do I will tell them I will come in on Thursday...hahahaha...I need to see that movie! I will be going with the husband of one of my grandmother voice students...I am looking forward to it because I will be hanging out with someone my age for once....and plus he is sooo low maintenance as far as friends go...I call him the \"budda\" because he has the patience of a Buddhist monk...nothing bothers him...I wish I had his brain chemistry...that would make my life easier...  Well enough for now...If I have to go in I will let everyone know...well everyone who reads this blog...which I feel is probably like three people! hahaha  I really want to know who reads this...so if you do read this please drop me a quick email...I am just curious...please! oh please! m.p.honse@verizon.net  '),(839,1234212,'07,April,2003','2003-04-07 00:00:00','   ok so I have been playing a lot of games lately...like really a lot of computer games...shit I have downloaded several demos...Will Rock and Breed...and I intend to download more.   I am a demo guy...I don\'t own many games...simply cause I haven\'t really run across that killer game yet...but I know I will sooner or later.  I have been hunting for a laptop...a dream laptop...cause that is all it is at this point...a dream..hehehehe...but I think I have found a worthy candidate for a moderate kinda do a little of all things...an Acer Travelmate 800CLi...Centrino 1.4, 512MB RAM, 40GB HD, Radeon 9000 64MB, and some other goodies...would suit me fine and is a lot less expensive than a Sony or most other notebooks...and it has a DVD/CD-R on it...so hey it would fill all the needs most certainly.  Not much else to report...my uncle and his wife and kid are here...so right now I have the stress level of a CEO...only without the perks...like free money, hookers, good attorneys...but I did clean my car from head to toe...with Orange Cleaner...yeah...that did the trick...now it is clean and smells good...didn\'t give it a coat of Armor All yet...don\'t think I will do it this go round...I also washed and waxed it...after all that cleaning I was tired...and that is what I am now...so off to bed...  Peace all  '),(840,1234212,'07,April,2003','2003-04-07 00:00:00','   oh yeah and found a great site that you all need to visit.  www.tuckermax.com  Be sure to find his stories...reminds me of some of the drunken and other intoxicated dibouchery that I have seen in my days...hopefully soon I will be able to get up to those standards though...regardless he is funny as hell and articulate about it all!  '),(841,1234212,'07,April,2003','2003-04-07 00:00:00','   Later today out to Hershey to visit with Jesse...then down to my Ex\'s house to work on her puter...  Gotta get sleep...hopefully I will have stories when I come back...we\'ll see!  '),(842,1234212,'07,April,2003','2003-04-07 00:00:00','   people who have lyrics in their AIM profiles show no originality whatsoever...that is why I have put some lyrics in mine...closer by NIN...I put it in there for the ladies...hahahaha  Well today has pretty much been a crap day. Not much happened...I didn\'t get out to sell my last train bits...but I will do that tomorrow...hopefully.  Tonight I may go see my friend Jesse...he has come back to Indiana for the next few days and I haven\'t seen him in a while...be good to hang out with someone my age...actually he is younger...but that makes no difference yo.  Got a call from one of my friends named Teresa..haven\'t heard from her in months. She is the former roommate and friend of my friend Jonelle...she is pretty funny. She has a good look on life...she is pretty religious...which I don\'t talk to her bout that cause I am very anti-religion...not anti-higher power...but anti-religion. I have to give her a shout back to see what is up and make sure all is well with her...  Oh well I am off to surf for some porn then off to do some other shit...  Peace out Yo!  '),(843,1234212,'07,April,2003','2003-04-07 00:00:00','   I dunno where I am right now...I mean literally I am in my room...but figuratively in my life I have no clue  I would love to be a millionaire with hot babes wanting to get in my pants...cause I am pretty...but that hasn\'t happened just yet.   There are a few ideas on the chopping block on how to make that happen...but nothing firm yet...shit I\'d settle for just some female and a couple of bucks.  Tomorrow I will sell the last of my Lionel Train bits hopefully and close out one phase of my life...mainly a fine in MS...then onto bigger and badder things...like what kind of fine will I have with all this legal crap facing me...will I be able to leave the state...will I come out of this one...who knows but I hope that it will all work out for the better.  I don\'t have a job right now...and really I haven\'t looked all that hard this past month. I have been sick and just not up to it all...but soon I will be well again and ready to face that challenge. So we will see...  On a side note my porn collection has now topped 10 gigs...almost time to get new system with bigger hard drives...the next system I build will be better than this one...but that may be a year or so off...or whenever I get a job...because right now I need a new laptop as I have said before...if I cram up my hard drives then I will get an encloser for a firewire or usb2 port and a big ole HD...like 180 or more gig HD....plenty of space for my growing porn and mp3 collection. In total I have 20 gigs in mp3\'s (mostly complete albums) and the aforementioned 10 gigs of porn...all hq and very tasty...  Well off to kill a few people on SOF2 and then some rest...  peace  '),(844,1234212,'07,April,2003','2003-04-07 00:00:00','   sex is a wonderful thing...but i get none...I need some...anyone want to volunteer and give me some lovin  m.p.honse@verizon.net  '),(845,1234212,'07,April,2003','2003-04-07 00:00:00','   I wanna bang you ;-)  So anytime anywhere...no strings attached...just some good hard fucking...or soft...you won\'t break me...I am the Ironman in bed...as in Triathlon....I can go for a hell of lot longer than ten minutes I can tell you that.  I\'ll even buy you a pizza!  This message sponsored by ragin\' boner  '),(846,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I am back from court and the good news it that I am not going to jail per-se  Here\'s the wrap. I got six months home detention, four years probation, and 96 odd thousand in restitution. So for six months I can\'t really leave the house but to go to the store, hospital, doctor, probation office, or job. Plus I get a k-rad ankle bracelet and base station...which will be in place next week. Four years of \"don\'t fuck up time\" in which I can\'t fuck up with anything more serious than a traffic ticket (even then they could technically nail me). During this time I am not allowed to drink or do drugs at all (which drinking pisses me off...but I don\'t drink that much now). For the first six months I am stuck here in the house...after that I can come and go as I please as long as I report to the office every so often (probably once a month) and let them know if I am going out of the Western District of PA. In addition during the home thing they can come visit whenever they please...plus I get the honor of pissing in a cup...most likely once a month for the next four years.  The 96 thousand in restitution will have to paid at a rate of 10% of my gross monthly income which right now would come out to about 53 dollars...and if I make more of course it will go up. Also if I want to get a new car, house, line of credit I have to report it and get approval for it (and they said as long as I am up to date with my restitution that wouldn\'t be a problem). I cannot own a gun either.  Technically I can\'t vote either...however they cannot prevent you from voting. All they would be able to do is challenge my vote and have it nullified. So a note to fellons out there...go ahead and vote with the knowledge that while technically you cannot the constitution provides that you can. The only thing they can do is challenge it and have it tossed...which in the end would make a great case to go in front of a court if they did challenge your vote. If nothing else the ACLU might take that one up. I plan to continue voting.  The head hancho of this whole mess finally plead guilty but is still trying to weasel out of responsibility and blame it on me, but the prosecutors and investigators aren\'t buying it and might have me and my ICQ log at his hearing. That kinda pisses me off. Trying to blame me for the entire deal. That is just obsurd, and apparently the prosecutors think that way as well. They seem to be out for blood and seeing the way they work...they\'ll get it.  All in all I didn\'t do so bad...and I am glad that it is over with, because now I know my fate. I am saddened that I have to be under house arrest, but I am glad to not be in jail. Because I would be a prison bitch fo sho! At least one of my co-comspirators will likely see jail time...and he will probably have to face that...and I feel somewhat bad for him. However instead of being a man and just facing up to it, he has seemed to choose to be a child and try and get out of it. Not a good way to go. There is no dignity in that at all.  I plan to spend the time learning a lot about web-programming and getting C# under my belt. Cause shit I\'ll have the time to do it. So expect to see a proper web site soon with a countdown time for all this shit.  Till next time...peace out.  \"The truth is so precious that it must be protected by lies in the time of war\" - Winston Churchill  '),(847,1234212,'07,April,2003','2003-04-07 00:00:00','   With friday looming closer and closer I am getting less and less sleep. It is hard to sleep when you know that there is a possibility that you may go to jail for the crime you commited. Plus add to that I have been using sleeping pills for the past five months and my prescription ran out at the beginning of this week forcing me back to my standby yet less effective alternative Benadryl.  I am trying not to worry...but the thoughts of me being a prison bitch just keep looming in my mind. It is hard to envision some large black man named \"Big Rob\" nicknamed that for his large unit and the offense that he was incarcerated for. It is hard to imagine that might be my next roommate. That is hard to deal with. Again I reiterate that I probably won\'t be headed to the Fed. Pen. but it is still within the realms of possibility and with my luck you never know.  We had good news from the Prosecutor as they have decided that I should recieve a downward departure from my current offense level (13 if anyone is interested) for being cooperative. However what this means is that it is up to the judge to make that decision. Plus with the health documents we have submitted that will also allow for another downward departure and the prosecution has not objected which is also a good thing...making a possible double downward departure a thing that I can hope for.  Through all of this I have had the support of my mom and grandmother. More than could ever be expected. They have both been thankful that I haven\'t lied about what was going on and that I have handled this the right way. Both have told me so.   The plain truth of it all is I screwed up. I let them down, I let my friends down, but more importantly I have let myself down. Through my friendliness and willingness to always be a help to people I have constantly gotten myself into cituation which really I shouldn\'t have gotten into in the first place. This has helped me learn that when it comes to friends that I do have to be guarded and cautious when they are asking for help. It does not mean that I won\'t help out when someone is in a pickle, but it means that I will not put myself into a situation where I have to compromise what I know to be right. I knew in this instance that my actions were indeed against the law and would place me smack in the middle of something that perhaps I shouldn\'t have been in. In my own greed and hopes of expanding friendships I compromised what I know to be right and wrong.   I have done many things wrong in my life. I have done many many drugs, and have been caught (yet kept up the behaviour...although I have been pretty clean for some time). It also means that I recognize the things that I have done wrong. Which is more than I can say for some of my friends/former friends.  My plans regardless are to try and not run afoul of the law in the future, but I will not be afraid of the law. It does not mean that I won\'t break it from time to time, but I will be better suited to deal with the consequences of said actions. I can\'t promise anything as no one can, but from now on I will try and be a good little doobie. Although I feel some laws are unfair and should be stricken down.  I have but a few regrets in my life. I can count this as part of one of my regrets. That regret is trying to be something I am not, trying to please people that I cannot, trying to make myself more attractive to others when they should accept me the way I am. I will still be a good friend...but a more cautious one, and certainly not the guy to call if you need a car to take a ride in or the guy you give a call when you think you have this great idea that is indeed illegal. Because I won\'t be your friend in those circumstances. I will be your enemy. I will not be there to help in those cituation.  And with that I will leave you with a qoute I came up with today. It pretty much sums up a lot of things I have been feeling lately.  To those that say \"Have Faith\"...well faith needs to be validated from time to time, kinda like parking. If you don\'t get validated sooner or later you will move on to another if not just for the free parking. Me 2003  '),(848,1234212,'07,April,2003','2003-04-07 00:00:00','   ok so it has been a few days since I posted last...bad me.  Well the port was de-accessed a week later than expected so I had to lug all my meds out to hershey with me. That was a pain in the ass, but all is well and things are relatively good now. I am feeling a bit better...altough my cough is worse than ever and has made me puke for the last two days. When I went to the docs this last time I really had impressive lung function scores so that was a bonus...but the drawback is even though I have good scores the important scores which are the small airway passages (the airways that have Alveoli in them) have continued to drop and are now down to 9 percent of what they should be. That score alarms me and now I know why I feel so out of breath during even light excersize (like walking up a flight of stairs).  On Friday I have to go to court to find out what my punishment is...hopefully it won\'t be to harsh. It really comes down to what the judge orders. The Prosectutor filed a really important document stating that I have been very helpful in the investigation and that they would not oppose a downward departure from my current offense level...all that means is that they have no problem with me getting probation. Again though it comes down to what the judge orders.  While out in Hershey on the way back I stopped at my ex\'s place in Hanover...she has gotten hotter since I dated her no shit. She trimmed off twenty pounds or so and is rockin...any guy that bags her would be lucky. We had a good visit and she even bought me dinner...which was hella kewl.  Well that is all I have to tell at this point...I will most likely post again after Fridays proceedings to let people know how I faired.  Wish me luck!  '),(849,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok so now I am getting pretty sure about this whole gut thing. I think that it is plugging up fo sho now. It will be a few more days until I am 100% sure but right now I am 90% sure that is the cause of all the problems.  Let me put it this way...imagine constipation times five...but regular things don\'t work to get rid of this because more than likely it isn\'t poo that is blocking the way...it is a mucus plug...usually they get about the size of a small fist. Plus it doesn\'t usually strike in the colon (where it would be easy to get at with a chemical enema) but it strike up way far in the gut where they can only get at it with strong drano like chemicals that either go in through a tube down your nose or another chemical that you can chug. If they can\'t get it with chemicals they have to cut it out...  Well some of the signs of this happening are extreme gastric pain, constipation for two or more days, fevers, and throwing up what looks like poo from you gut (basically your gut backs up and the only opening available to it is your mouth). Now I have most of the symptoms but not all are presenting as normal...I am constipated but I still pass a little poo now and then...I haven\'t had a fever (until tonight that is), and the gastric pain has been more in my back than stomach, and I have been throwing up but right now I contribute that to coughing (although what I threw up a few minutes ago would be considered bowel crap).   So over the next few days I will see...if everything I put down tries to come up...then I know that it is blocked...if I double over in screaming pain and my stomach gets distended...then I know it is blocked...if my guts explode and stomach turns green...then I know I am fucked royally.  Night all...and thus ends another glorious day in my life.  '),(850,1234212,'07,April,2003','2003-04-07 00:00:00','   Well Blogger has changed a bit...but it is nicer now.  For the past week I have been having pretty severe pains in my side...they are getting worse day by day. I am really starting to get tired of it. It is almost hindering my moving about the house. So I have spent an inordinate amount of time in bed and on the couch. I know this pisses GM off to no end...because I am being lazy, but hell it hurts like hell!   Tonight it has moved into a new realm. What I was thinking it was may just be and that sucks ass. See previous blogs to see what I thought it was/maybe is. The reason I am leaning more that way is because now the gut it binding up and hurting even more and I haven\'t gone to the bathroom really well all day. I know it is more info than is needed...but this is my blog damnit!  It would seem that the RIAA is now gone completely nuts. Yes completely nuts. Gone, out to lunch, they all need to be admited to a loony bin. While I have engaged in my fair share of music piracy I still hold the truth that artists need to be paid for their work. Under the current system we buy a CD and they get a small percent of the sale (maybe like a dollar or so from each CD if they are lucky). I would not begrudge them that money at all. I feel that they should get more. However I get really pissed when I spend 15-20 bucks for a CD to get just one or two decent songs. Rarely have I bought Artist CDs that have more than that average. That is why I have changed my listening habits a lot.  First I am not really into the top 20 artists. Not because they don\'t have catchy tunes. But because to own the right to play that tune anytime I have to buy a CD with maybe a few decent tracks on it. So I have moved over to Techno and Rap and other types of music. With those genres I know what I am getting. I am getting a CD full of something that I will probably like if I like the few catchy tunes that are on the Radio. When there is a top 20 song that I want I wait for a compilation CD to come out with several songs on it that I want. I simply don\'t bother with buying many artists lone CDs.  Second if I do get a bunk CD with only a couple of decent tunes on it then I burn my own compilation CDs. However the RIAA would not like this practice. Hell if they looked in my car they would find it chock full of cheaply burned compilation CDs that I have made. While this is seen as cheating by the music labels I see it as smart consumerism. I don\'t want to have to listen to a crap CD or scroll through a CD to get to a song I like. This is partly because I am lazy, and partly because I realize that CD players wear out. Yes folks they do wear out. Hell I wore out three CD players over the years at least. I have one of those old Sony Discmans that cost an arm and a leg yet it can no longer track and play any CD that is put in it. I wore out a Car CD deck. I have worn out CD players in home stereo systems. If I can avoid wearing out any more I would love to do that and they wear out when they have to scroll through CDs full of crap.  Third I try and listen to as many non-signed artists as possible. I would love to pay them all for the music they have given me over the years, but alas I am still poor. So this isn\'t a reality just yet.  The music industry is corrupt and rife with money grubbing idiots that like to sign crap artists to huge deals so they can get a bonus. Just think of crap artists like Mariah Carey and Michael Jackson. So they have had great songs in the past...but does it justify contracts of 90 million for Mariah or 100 million advances for Michael. Hell no...especially seeing that both are bordering on being insane...if not completely so.  What does this bring me to. It brings me to having to break the law just to get some music in most cases. That sucks.  So what has the RIAA done to go absolutely crazy. Well they have decided since they have sued just about every P2P company out of existance and are now facing stiffer resistance in the courts (recently a lower court held that Kazaa and others of its ilk aren\'t illigal and are not responsible for the possible infractions going on...kind of like tape players and VCRs in the early days). Instead now the RIAA will be going after the end user. After a recent court victory in favor of producing the identity of a user on Verizon\'s network they have decided to start suing you and I. That means that if you are a user of any of the current unprotected and un-encrypted networks out there that you may be a target of a lawsuit for sharing music. Right now they are going after the big users sharing the most files. If I were on one of those networks I would certainly fall within that catagory (as I have around 3000 mp3s...most of which I own a CD that has the songs on them). Sooner or later they would get around to suing just the average user...especially seeing that just recently they had rulings in their favor against four college students that were hosting sharing services or searching services on Campus Networks.  The RIAA has essentially declared war on the consumers. Now granted their are a lot of people that use these services and never buy CDs anymore. As I have said that isn\'t fair and those users should be scared and should have some sort of action against them. However for people like me that do buy music (usually around 12-20 cds a year) would be the target of the very orginization that I help support. That simply isn\'t right at all.  So what have I done. Well first of all I have moved from the open networks over to a program called WASTE. It is an encrypted file sharing system that only allows for so many users per network and makes it so that you have to authorize anyone who you want to share files with. While this doesn\'t mean a mole can\'t wiggle his way into your inner circle it does make it a little more secure. Also with the encrypting that the program does they may be able to monitor your ports and say you are using a lot of bandwidth but hell they can\'t readily crack the encrypted packets flying out of your machine. The system isn\'t foolproof and they will eventually find a way to track it...but in the meantime it is better than the current systems. It gives you the user more control over who has access to your files.   Also I have pledged to not buy anymore CDs or at least try not and buy anymore CDs. While my measly CD purchasing contribution each year means nothing to them, it is still that many less dollars that they have to fight this war. I would urge others to do the same. Copy a CD from a friend. Have a few friend pitch in and buy one CD and all of you copy it, tape and then burn songs off the radio, but try and not buy as many copies of CDs that you currently do.   Also get in touch with your representatives and voice your opinions against measures that the RIAA and MPIAA would have introduced into congress.  They are all out of touch with the consumer and it is time that we start telling them that with our dollars and our votes in congress.  '),(851,1234212,'07,April,2003','2003-04-07 00:00:00','   If she reads this blog...which maybe she does...she knows who she is...and I am not ashamed :-) Because after all sweetheart you are quite the woman.  '),(852,1234212,'07,April,2003','2003-04-07 00:00:00','   Well...I had one of the most interesting dreams that I have had in a long time last night. It probably was the hottest and most bothered I have gotten in a long time over anyone or anything. Wow!  I had a dream about someone for which I have had an attraction to for quite some time. I am sure she knows about it...its not like I hide it very well. Everytime I get around her (which isn\'t at all very often...as a matter of fact last time I saw her she was in a coat and there was close to a foot of snow on the ground and I saw her for maybe three minutes) I get all gooey inside. Well I think dirty thoughts about her from time to time....but good christ last night. If what happened in my dream ever ever comes true it will be the most wild excellent thing that ever happens to me period. After that my life would be complete. The only thing that would be better is if we were to go out...but I doubt that will ever happen. For you see I would have to be a better man. I mean by a long shot. I would need to do a lot in my life to elevate it from where it is I fear.   Its not that she would be high maintenance or anything...but I would want to better so that I could be better around her. If you have ever had a person in your life like that then you know what I mean.  Anyway the dream was wild...and I would hope that I could have that dream more often...if only I knew what set it off then I would recreate those circumstances over and over...kinda like a groundhog day of dreams.  '),(853,1234212,'07,April,2003','2003-04-07 00:00:00','   ohh by the way...I have been playing a lot of SOF2  if ya wanna play me look for h0n3z when you fire it up for online play...I may be around  '),(854,1234212,'07,April,2003','2003-04-07 00:00:00','   so the last twenty four hours has kinda been a shit day...  For the past couple of days I have had this pain in my side...left side right about where the kidney would be only a little forward of that...it has caused me some severe stress...I am thinking with the recent spreading of the uncomfortable feeling that is currently spreading round the entire belly that maybe I am experiencing another partial or full blockage of the bowel.  Now normally this isn\'t such a huge deal...but for me I always get the nurse or doctor who thinks that they can get it with an enima...well they can\'t as these reside far north of where an enima can reach. So what they end up doing is having to give me some sort of turbo lax for days on end to get me flushed through...one good thing about this...is this...if it is a bowel blockage then I will get a hospital stay for it...and if it is a bowel blockage I will get good pain meds...cause they are the only thing that will relieve anyone of the excruciating pain that they cause...  If it is that...I hope they can find the broke piece of bowel and get it out...I know that it would originate from the same area that the four or five prior blockages originated from...so maybe they can rip that piece out and just get rid of the problem area...who knows...  other than that I am me...later today I have to run to the Drug Store and the Bank to get money to pay for drugs...so I will actually get to leave the house justifyably...wooohooo!  '),(855,1234212,'07,April,2003','2003-04-07 00:00:00','   Well yesterday I found out that one of the other people that was in this whole mess with me was also sentenced on Friday. Now I haven\'t seen him online since then. I don\'t know what to think about that...I hope he didn\'t go away...while he may have been more involved in it all (seeing he was the others/ringleaders roommate) I don\'t think he was as much at fault. Today I read in the paper that he recieved a similar sentence to what I got...but if that is so then why isn\'t he online...this poses an interesting question. Did he indeed go up the river for this so to speak, and if so for how long?  All of these questions I have and no way to answer them just yet.   Well everything is getting inline here at the house for this six months of sollitude...I have my room arranged a little more comfortably...no the only thing will to bring my TV up here and get it some cable or something. I have a lot of DVD\'s and Movies to watch so I will be at least somewhat ammused for a bit. Plus with all the money I will be saving by not having to drive around and do shit I will be able to buy so many more DVD\'s. I can see that Amazon.com and others will become part of my monthly purchases.  I just might sign up for another DVD Club membership to get some movies....hehehehe  '),(856,1234212,'07,April,2003','2003-04-07 00:00:00','   BIG POST ERROR, POST ID 105713457918546720  urlLink REPORT IT   This post that died on its way to the server was shall I say one of my finest pieces of literary work ever. I will try and re-create it in Word and then post it here....what a shame. Now I have cramped hands for nothing...BLAH!!!  But Blogger is still DA-BOMB (even though randomly it drops things).  '),(857,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I got my room a little cleaner than it has been for a while. It is still a pretty big mess...and I think there is a midget made of funk living somewhere in the room...when I poked a whole in the mess that covered the floor I heard the distinct yell and banshee cry of the \"one\".   Today has just been a waste of day. Right now I am watching Super Troopers and just changed my keyboard over to my Logitech Cordless Optical that I have had laying around. So we will see how this goes for a while.   Well I am going to clean up a little more and watch this movie.  Peace out...and no I didn\'t get to meditate. Tomorrow if I get everything done, like moving my TV and Entertainment center and get all that shit set up and connected to the computer of doom. That is what I will call my puter from now on..the computer of doom!  Again...peace out!  '),(858,1234212,'07,April,2003','2003-04-07 00:00:00','   Oh and I haven\'t started the experiment with the Brain thing yet...I am charging bateries for my CD Walkman now and will give it a shot later tonight...Wish me luck cause I need all the luck I can get when it comes to women...hahahah  '),(859,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I spent most of the day in bed...I am a lazy ass yes I know.  Once I got up and moving I had a bite to eat then did my treatments, and now am cleaning my pigstye I call a room...It is far from done...but it is a little better than it was...  I just have to finish it up, move my TV up here from the Garage and a few other things and I will be all set for my sollitude over the next 6 months...oh joy!  '),(860,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I went to another graduation party today and the really hot girl that I had formerly overlooked was there. All I can say is Ka-POW!  Tonight I read part of a guide to get women really hot and bothered over you. It is based on doing it the new-fangled new-aged way of sending out brainwaves (or vibes as I like to call it). Now people might think that this is rediculous...but I would beg to differ based on just some common sense working of the mental powers that some have.   First what is the most common trait of guys that have girls falling all over them? They are all relatively stupid. Think about it for a moment. Yes let it seep into your cranium. Of all the hot guys that have females fawning over them almost all of them have a single trait in common. They are dumb as a brick. I mean come on these guys wouldn\'t be rocket scientists ever. You have met them. Then think of all the guys that are super smart, even some of the good looking guys. They rarely have females fawning over them. They instead are just kind of introverts with a great personality and whit but rarely get to use it on a girl. Why? Because they are nerds...and classically females avoid nerds at all costs. Now I am not saying that you few females out there aren\'t into brainiac guys, some of you are, but the vast majority of you don\'t go for the nerd. Also look at the reverse. Almost all of the good looking girls that have guys fawning over them are pretty dumb as well. The smart girls almost never get any play. Of course there are exceptions to the rule as I call it...but those are few and far between.  Second why is it this way? Well maybe they tap into something more primal or primitave...or maybe it is voodoo. Well not in my universe there has to be something to these rules. Remember rules are meant to be broken.  Now this little read that I did basically states that you need to slow your brain down to a lower operating level, something around 7.83 hertz. The brain natively runs at around 14-30 hertz or so...Now using this logic you see that the average person runs at this speed. Using the old concept of faster is faster and thus more productive I would be that smarter people run at a slightly higher speed (maybe in the 21-30 hertz or maybe higher range) and dumber people run at a slightly lower speed (maybe around 14-21 hertz). This would give them a lower total number to reduce their brains down to this \"majical\" 7hz rating.   What is this \"majical\" 7.83hz. Well it is where the natural frequency of the earth resides. We know as fact that the earth hums along at this rate. Now every person has the ability and indeed changes their frequency with normal function around 14-30hz (or Beta Stage). When we enter into sleep we go through several stages (which are well documented). The stage that rests around the 7.83hz mark is considered (which is 7-13hz or Alpha Stage). The other stages are Theta 3.5-7hz and Delta Stage .5-3.5hz.. These stages correspond with sleep Alpha-being around that time when you are just starting to fall asleep (or daydream or meditation), Theta-sleep, and Alpha-deep sleep and dream.  Now if we can get our brains down to that \"majical\" level we can essentially tap into the earths frequency by just using logic. If we are running at that speed we are running at the speed that the earth does on the eltromagnetic scale. Remember our brains run of electricity. So to tap into this natural energy might not be so far fetched.  Think about this. When you have a short nap you wake up refreshed and ready to go usually. Meaning you have been recharged or maybe your brain is just cooled of a bit (remember that heat is generated by electricity and when you get shocked or struck by lightning it isn\'t nesecarily the electricity that kills you, it is the heat damage that kills).   At this \"majical\" number we are essentially plugged into energy that is greater than our own by itself. Now what can be done with this energy? Just about anything!  Now this book tells you to funnel that energy when in that state to do some good...like getting a girl all hot and bothered for you. There are many steps to achieving this, but all in all the technique is relatively simple overall.   We hear about Martial Arts masters being able to do all sorts of things that seem superhuman, and what do all Martial Arts masters have in common. They all do some sort of meditating. Plus you always hear them say \"center yourself\" fight not in anger but from a centered place. The same could be said about \"Christian Healers\" or those that pray for others, they are practicing a form of meditation (by praying and concentrating on what they are praying for). The Bible even mentions meditating for God\'s glory or something like that. Through mediation we esentially lower our brain wave activities to these lower levels and some people get to the \"majical\" number easily through this technic.   While the book claims that you can essentially heal and do all sorts of ill shit when we are at that level (some of which I debate to be valid until proven) it would make sense to some degree. When does the body heal itself? When we sleep!  So bringing it around to the argument that dumber people are more attractive to the opposite sex we see that they have a lower level to drop and smarter people have to drop much further to get to the level that a lot of dumb people are at even on a regular day to day basis.   When a dumb guy gets drunk he is essentially bringing his brain waves down to a semi-concious dream state buy default and taps into that power, because lets face it what happens to most when they get drunk? They get horny to boot. So when a dumbie is drunk and horny for a particular person they are sending out these signals fairly easily. That is why it is almost always easier to hook up when drunk. Think about it...it is true. The smarter person has still to make it to this level whenever they are drunk as they naturally function faster. Which when trying to bring down the frequency with alcohol you run into barriers such as alcohol poisoning.   So through meditation we may be able to get our levels down to this \"majic\" number.   You might say that once you are at that level then you are essentially asleep. Well I beg to differ. Ever asked someone a question when they are daydreaming or napping. Sometimes you get a responce that may be way off the wall...but you always get some sort of responce...even if it is only a look or nod. Try it sometime...you may be surprised.  Well I am going to try this out and see what happens...because you know I need all the help I can get. I have several people that I will try this on. One will be easy to get the info out of...the other almost as easy...and the third one will be blind to the attempt so we will have to see what happens.   This will delight one of my friends that I am trying it...because also at the same time I will try and think healing thoughts and see what happens. Again I will be able to tell this easily because I am me...and I know my body well.  I will of coarse post my results as time goes on. In the meantime I need to find a mp3 of the sounds that help you get into this mode as you can change your brain mode by just listening to noise or music (we all know this to be true...listen to a slow relaxing song and you will be relaxed, listen to something faster and heavier and you will notice that you may be more alert...something many of us have done when on our way home late at night sleepy).  Well I am off to find that mp3 to help...and to all of you non-believers out there hopefully if I see results you might start to believe. Because as humans we know less than we think and their are many things out there we can\'t explain but accept.  '),(861,1234212,'07,April,2003','2003-04-07 00:00:00','   So today I went with my grandmother to a graduation party for one of her students. I had to take her because she wasn\'t feeling so great and wasn\'t up to driving her.   Imagine this...pale skinny white ass me in amongst a gaggle of very good looking young nubile 18 year old women...and a few that were younger. Woah momma! I was kinda shocked because mostly I see these students of my grandmother and they are not always made up and just kinda schleppin along (which even then most all of them are ultra hot), but I got to see them today all dolled up. If I had been feeling just a little better I would have had a very large hard on simply because I have not been around that much fine ass in a long long time...maybe a frat party once, but it has been a long ass time.   The student graduating was Cara and she is a fox admittedly...plus she is super super talented. I tell you this woman is just a damn artistic genious. She paints, sings, acts...all of it and she is just absolutely amazing at it all. She has brains along with it. I tell you she will definitely do well in whatever venture she decides on doing in the future. I have always been impressed by her talent and today I was able to see all the art that she has worked on over the years. She is talented...and whoever snags her in the future will be one lucky mofo....put simply she is a great human being.  Yesterday was another girls graduation which I sadly missed...I wasn\'t feeling up to it, however looking back I should have gone. I have seen her in here before but never really paid attention to her. God damn is she fine. She is cute as a button and has a nice ass. You know the kind of ass that you look at then look away then look at again to make sure it is really there...wow! It was nice to see her to say the least. I guess at some point when I went to get my grandmother a drink she asked her if my tongue was pierced...my grandmother didn\'t answer I don\'t think but told her to ask or something like that. Well lemme tell ya little youngin it is...and it has many many good uses...I can be skilled with it when needed. I sure am glad all these girls are 18 now...because damn I would feel just soooo wrong if they weren\'t.  Tomorrow I have to take her to another graduation party. This girl I don\'t know but all of the others will be there...which is kewl because I won\'t be the biggest dork in the world...or something like that. I guess they all see me as an inigma or something because they have some sort of strange reverence towards me...or something like that...if they only knew that I was just a lazy ass twenty something...sick and lonely...hahahaha...and if only they knew some of the things I will be thinking about them ;-)  Anyways not to brag to much or anything but I was one of the best looking guys there...I mean sure there were larger hulkier guys there...but they were all just goofy looking and some were plain out fugly...so it gave me some confidence...at least I knew I was one of the better looking guys and if we were all forced to pick partners to be with if we were the last people on the earth I wouldn\'t be last this time! YAY!  Anyways I am tired and I need to rest...even though I sleep on average 12-14 hours a day now...not because I want to but because I have to...it sucks ass so bad...if only I had a little energy left...I would chase those girls in the bushes...or something like that ;-)  If there are any little green alien guys out there that would like to abduct me and fix all my bad parts, add 30 pounds to my body, and insert a vast amount of computer knowledge into my brain...well then go ahead and take my ass...you can probe anything you want anytime!  '),(862,1234212,'07,April,2003','2003-04-07 00:00:00','   Retraction...the qoute I posted from that persons info was actually from a Beastie Boys song...so it was not his original work. And the point of this website is to be about me...and me alone...thus the name \"A Day in the life of...\" Blog. The only thing this blog is designed to do is simply let people into my relatively small little world such as it is at this point as I don\'t go out much (right now because of house arrest) and even before because of deteriorating health.   I don\'t drink very often and now haven\'t drank in months...drugs have been cut out of my life for a long time as well. As a matter of fact I have few vices outside of Pepsi and Spaghetti O\'s. So this is my lifeline to the outside...excuse me if I use it as a platform to tell the truth.   I still haven\'t sat down to write that thing...maybe I will drag my laptop downstairs and write it out...jeesh I have a lot to do...  But I still stand by my earlier statements...I may be a dick...but at least I am not a weasel.  '),(863,1234212,'07,April,2003','2003-04-07 00:00:00','   Today I finally found out that SprintPCS has added AIM to its list of services...this makes me mucho happio...and when I signed on to an old name that I still keep active from time to time I found my ex-roommate and soon to be federal inmate online (well I presume he will be a federal inmate as he is pretty screwed)...so then I signed on to that account not from my phone but computer to read the info that he has posted up...and I qoute  \"Can\'t stand it I know you planned it Ima set it straight This Watergate\" What a poet...hahahahaha...anyways I have an idea who it is aimed at...or rather the parties it is aimed at...my simple response to it is this...I planned nothing, there is nothing to set straight you are guilty as sin...and whats with the Watergate reference...everyone knows that Nixon was involved with that and it was a leading factor in him leaving office...so are you saying that your guilty...  Anyways enough of that...  I have had my fill of laughs for the evening...peace ya\'ll  '),(864,1234212,'07,April,2003','2003-04-07 00:00:00','   Well the pills last night finally did kick in...and when they did they made me sooo dizzy that I thought I was gonna throw up hard...I didn\'t...instead my lung started to bleed a lot...and looking at that maybe it was the slight blood loss that made me really dizzy (my blood count is always low as I leak like a siv).  Today the US Probation office came and put my leash on. The guy is really nice and I am thinking of cronicleing the next six months...maybe he won\'t mind if I mention his name. However I have to run that by him to see if he cares...plus obviously it would be transparent...if I talk shit then he would most certainly know. But again he is really nice and laid out a few ground rules...plus he told me that I am not gonna be a hermit cause he will be somewhat flexible as long as I be straight with him...and I do plan to be straight up with them all...cause I ain\'t gonna lie I don\'t wanna revoke this and have my ass end up in jail.  Well I dunno what now...maybe I will find some amusing shit to get myself into on the internet over the next few months...I have a lot of time and some energy to get rid of...plus in the next month or so I will be really starting to get my Ebay thing going...I have a lot of stuff to get rid of...baseball, basketball, football collectables for my mom and friend...so that will take up some time and I will make some money wooohoooooo!  Well peace out...tomorrow I promise I will write that thing again that was a masterpiece...plus this time it will have all the correct spelling and grammer and will contain more than I had originally wrote.  Peace  '),(865,1234212,'07,April,2003','2003-04-07 00:00:00','   While I wait for it to kick in I will kill people on the internet  '),(866,1234212,'07,April,2003','2003-04-07 00:00:00','   Well sleeping pills would be great if they worked as prescribed...but they don\'t of coarse...they just kinda make you tired...so I am going to take another type and try and sleep...maybe that will work in connection with the other one I took a few hours ago.  '),(867,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok so later today the US Probation office is coming to put the leash on me...aww this is gonna suck big time. Being stuck in the house with my grandmother for the next 6 months is sure to drive me absolutely mad.  Funny thing is I am getting a lung infection and I may have to take the leash off to go to the hospital soon...life is Ironic  '),(868,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I haven\'t yet rewritten my literary work of art yet...maybe later tonight I will concentrate and try to write it.  Not to much has been happening around here.   My sister did finally get up here. She brought with her new toys for me to play with. A Sony Camera (very swanky indeed) and a PocketPC by Viewsonic. She needs some apps for the PocketPC so I downloaded a shitload of development apps for it. Overall the downloads totaled around 500megs from Microsoft (they are all free development apps surprisingly enough). The only thing that I wasn\'t able to download was Visual Studio .NET which I already have the Architect Edition...so I am in the butter zone so to speak. Now I just have to sit down and do a little coding and try to get what she needs up and running.  The apps are basic calculators for Vet Medicine. One is for figuring out dosages for animals based on weight and the other is a subset, dosage dilluted into another substance by weight. They are all in the microgram level...so that will be interesting to get all the calculations written down then coded. I am looking forward to it because really when it comes to programming a calculator is probably one of the easiest things to figure out...the only difficult part will be GUI as I have never done any GUI programming...everything I have done was text based. So it will be a learning experience to say the least but I am looking forward to coding it out.   Plus it will give me some coding experience on PocketPC\'s which are fast becoming the standard for handheld devices. If I get semi-good at doing it on PocketPC then I will learn about PalmOS programming...which is a little different and with much different programming apps.  Well I am off to dinner...something down there must be for me to eat...must eat I am...  '),(869,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok so I just haven\'t hooked the lappy up...I am not feeling real hot right now. Whenever I get up tomorrow I will go ahead and hook it all back up and get it plugged in. Then hopefully online.  Right now I am watching Robin Williams Live on Broadway...it is funny shit. I love watching the comedy master at work.  '),(870,1234212,'07,April,2003','2003-04-07 00:00:00','   So I finally wrote out the masterpiece as I will call it...well it isn\'t nearly as good as the original because this time I was trying to write something Earth shattering...I don\'t think it will have the same effect as the first version.   Basically it is my religion as I call it. My views and beliefs. I finished it up yesterday as my sister was screaching into the driveway making my a sound that I never like to hear my car do.  I will put the laptop back on the network later today and get it posted. It was something like three pages in MS Word.  Well something happened to me that I certainly didn\'t expect this last Sunday. My grandmother planned a surprise birthday party for my sis and I. Now I had known about it for a while so it wasn\'t in and of itself a surprise, but what was surprising were all the people that showed up. Even more surprising is that she was able to get consent from the Probation Office for me to go to it at a local resteraunt. Now I would never have expected that because the whole point of being on house arrest is to keep you in the house, but as I stated before my Probation Officer is really nice and was kind enough to let me go.   I got all sorts of gifts (which I hate that all those people spent or gave money to me) and all of them are greatly appreciated. I recieved 95 bucks in cash and 70 bucks worth of Gift cards to Sheetz and Hollywood Video, plus some other assorted stuff including a T-Shirt that is da\' bomb and a really unique gift of a bath plug.   A bath plug? Well you see that is from my neighbor. When I was over fixing their family computer about two months ago he was teasing me that if I lost anymore weight that I would fall through my asshole. He kept this up for the two hours I was there. He packed it in multiple boxes that were harder to get into as I got further in the package, making it a challenge for me. On top of it he gave me his dull pen knife to cut into it making it all that much more difficult.   I was really shocked by all of it. It was a really nice thing to have seeing that I have only had three other birthday parties ever. One was my own, the other was a combined party that really was my sisters party, and the third was my 21st which I had at school. Now I have this one to add. Even though it did not include copious amounts of alcohol it was probably the most fun I have had in a while.   I also recieved an instant message from one of my friends that was also involved in this whole mess we are dealing with. It was nice to hear from him and his message was also very nice in that he told me to keep my chin up and hoped that I would start feeling better (or something to that effect). As I have always been towards him and this is true I hope all is well in his circle, and I hope that he continues to do well. After all this mess I think that we may have learned a pretty valuable lesson. First is to stand up and take the blame for what we had done. Second that friends are only friends when they don\'t try and screw you for things that you haven\'t done (as one other party involved in this all seemingly tried to blame everyone else and took a damn long time to stand up and take it on the chin...although I see his point of view seeing that he may be facing serious jail time.)  As has always been my schtick in the past. I certainly don\'t hold grudges. It is to time consuming and if you hold them sooner or later they come back and bite you in the ass. Even after all this shit I don\'t blame anyone but myself and I freely and willingly stood up and admitted what I had done, but no more than what I had done.   Friends help you out when you are in a pickle, but never ever expect friends to take a fall for you in situations like this especially if they aren\'t as involved. Also don\'t ever expect to friends to take even a minor fall for you if you treat them like shit several times over, and then talk shit about them. (Now I do not know for a fact nor would I claim that one person talked shit about me. However saying that I know this party well enough to know that he has probably said some things about me that weren\'t very nice. The problem is that for the people that know both myself and this other person they know that I am not in anyway the way what he would describe me as. I am a dick, I am a prick, I am many things, but I don\'t screw people over to make my cituation better. I have done many things that I am not proud of. Shit there are things that I would change, but I emphasize that I don\'t screw people over for my betterment. It just doesn\'t pay.)  Enough of that for now...I am reactionary so things do tend to excite me...and occasionally I like to try and stir shit up just for reactions...but again that is me being a dick, but you won\'t find me claiming anything outragious or outright lieing about people.  As for my health. That is a mixed bag. My lung infection is getting worse...very slowly but surely. I don\'t know when it will reach the \"I need a hospital\" point right now as this lung infection is very different from the ones in the past. It seems to be eating very little at a time, but it is feeding. As for that pain in my side it has gone away for the most part. I don\'t know what it was at all. I thought it might be a bowel obstruction but it wasn\'t. Maybe it was a fluke thing, or maybe a sign of something to come. Either way it wasn\'t pleasant and I hope not to be in that kind of discomfort again for some time. If it does re-occur then I will definitely call the Docs and get something for the pain as it certainly interfered with my life.   Till later today...I\'m outie  '),(871,1234212,'07,April,2003','2003-04-07 00:00:00','   Tomorrow laptop will be charged so that I can carry it downstairs to write on...  '),(872,1234212,'07,April,2003','2003-04-07 00:00:00','   Well all of today has sucked serious donkey balls...I am getting sick and there is nothing I can do about it except wait for it to get bad enough to head to the hospital. For you see they are almost out of drugs to treat my little bugs with...the stuff that is left that works better than the stuff now is penecillian based...and well I am allergic to Penicillian...so when I get sick this time it has to be bad because they will have to drag me in to treat me for penicillian allergies (try to reduce the effects of my reaction) and then treat me with the drugs...  Well trying to get me to tolerate Penicillian drugs will be dangerous to say the least...I swell like and Oscar Mayer weiner...or a Ball Park Frank...plus I get all itchy and it just isn\'t any fun at all...last time I had a reaction I was swollen and had hives for a week and a half before the stuff worked its way out of the system...so this time I dunno what will happen. I just have to wait and see...  Tomorrow will be an interesting day as I will be getting even more ill...  Illin\' like a villian ya\'ll  '),(873,1234212,'07,April,2003','2003-04-07 00:00:00','   spent most of the day in bed...at least until 3PM...tonight I will actually get to bed at a reasonable time as I need to be up for tomorrow...my mom will be back from MS and has a truck load of collectables to be unloaded, semi-cataloged, and packed away again...  Today I almost finished cleaning up my room...I at least got my TV, Stereo, VCR up here and set up. That was a freakin chore...now I have it all hooked up and I have my computer tied to the TV which allows me to watch all sorts of videos from my PC on my TV...that is so freakin money yo!   Plus I can play games on the TV screen as well...and that is kewl as hell! SOF2 actually looks good on my TV...all I would need to do is stretch my keyboard and mouse over to my bed and play from there so I would be in front of the TV...I may give myself the ability to do this once I get some extension cords for both my mouse and keyboard...I have one I just need the second one.  Right now I am updating this blog, looking for a cheap price on Ebay for a Nikon Coolpix SQ (neat little camera), and getting ready to hit the sack (my meds are kicking in).  I need a new camera and the Coolpix SQ would fit the bill 3.1Megapixel Camera, teeny tiny in size, and has a lot of cool features...I think I will try and get one this week as I have a lot of photos to take of all these collectables...already have my scanner set up, just need the camera to really get all those things moved...  It looks like there will be many items that will sell for close to a grand or more...and plenty of other things to sell...  That makes me happy cause my mom could use the cash, I can use the cash and get some of the things I need to...like new laptop and a digital camcorder, and maybe a TabletPC...maybe I will be able to fullfill all my gadget wishes by the end of the year! YAY!  '),(874,1234212,'07,April,2003','2003-04-07 00:00:00','   Also I would like to quickly mention that Dean is the first candidate to really cater to the online world...the rest have pages...but Dean\'s team has really made it their base of operations...  Think about it...each month they may spend 50grand on Bandwidth but get huge penetration into the market.  To put a thirty second spot on TV it would cost many times that just to get it in the primetime slots...while I am sure we will see comercials for him...I really think that he is being fiscally responsible by going the major web route...he is able to just simply get more bang for the buck this way...so look at his site often and inform your friends and family.  This blog will start to become a podium of sorts to get Bush out and Dean in...it is time for a change...stand up, take charge, grow a pair, and help get this Tyrant out of office...it is the Patriotic thing to do.  '),(875,1234212,'07,April,2003','2003-04-07 00:00:00','   And by the way I will be giving this next month when I get my next check... :-)  '),(876,1234212,'07,April,2003','2003-04-07 00:00:00','   I set up a contributions page for Howard Dean. My personal goal is to raise 100 bucks for the man. Help me out!   Give Money to Dean  '),(877,1234212,'07,April,2003','2003-04-07 00:00:00','   Today I cleaned the garage, cleaned my room, and nearly died of a coughing fit...  I hate cleaning but I love clean...I need to be extremely rich or really lucky...if I was rich I would have maids that would keep things clean (plus I would treat them good and pay them well...they would really be part of my family)...or if I was lucky I would have a female that would clean up for me and then lay me...  That\'s right I am horny again...well kinda. Wanna have sex but don\'t want to expend the energy to actually get it or have it...I want a new drug like Huey Lewis...I wanna sex pill that gives all the physical pleasures without the actual act having been made...it would be like this.  Step one. Pop the pill. Step two. Wait ten minutes. Step three. Look some porn. Step four. Get off (without any physical activity or even mess) Step five. Smoke a cigarette and go to bed.  That would be the best drug ever.  I also want a drug that makes me feel better. They have it now...it is called Vicodan. Now I just need to get some.  Maybe when all this house arrest shit is over I will go and find a woman to bang...and then hopefully bring home.  Think about it.  '),(878,1234212,'07,April,2003','2003-04-07 00:00:00','   Oh and a just yesterday I saw a show on either FSTV (Dish Network 9415) or Worldlink TV (Dish 9410) that was about the Raeleons or whatever they are called...they seem to have similar beliefs to the higher power as I do...however I don\'t go as far as saying that aliens are our gods...because I don\'t know for sure...but whatever...call me a crackpot but I don\'t care...cause I tolerate your beliefs...  also a couple more links for those of you who believe that the government is up to no good.   Conspiracy? You Decide!   A Little Weirder and off the Wall  '),(879,1234212,'07,April,2003','2003-04-07 00:00:00','   Again today I wasted most of the day away...as is usual with everyday...  Now off to kill a few people on the internet and then a little sleep  '),(880,1234212,'07,April,2003','2003-04-07 00:00:00','   Today I was able to get a almost two hour reprive from House Arrest as I was able to go out and get a few things. I didn\'t get everything done that I needed to but I did accomplish a few things.  The only comfort about this whole house arrest thing has been that I have internet access...outside of that it has been pretty torturous...I mean I am locked in the house with my grandmother 24 hours a day, seven days a week. Now as some of you that have known me for a long time may recall that I go pretty nuts when I am confined with her. Well it is happening again.  Not having had sex in well over two years is also playing a factor in that. Last time I was confined like this at least it was only a month since I had last had sex...but now with all those pent up tensions it is starting to wear on me a lot more than I initially factored. I think that I will go ahead and seek counseling again as I have in the past...not really to work things out, but to just get out of this friggin house more often...also I have a lot of medical appointments that I have to make so that will also get me out a little more often. Maybe I can swing it that I get out once a week to do these things.   On a brighter note though my grandmother is looking into buying another house here in Springdale PA. This house would be for me as I qualify for Section 8 housing assistance and that would essentially pay for her to get another place...it would be an investment for her, and a way out for me. I would finally be able to get out of this place and maybe get myself a little more established. I would finally have my own domain...that would be very nice indeed...so on the off chance that I would meet a suitable woman then I can have her over without being a total dork living in my grandmothers house.   Also this coming week my mom will be getting back from her jaunt to MS. When she comes back I will have the monumental task of selling a ton of collectibles like baseball cards, action figures, bobble head thingies, basketballs signed by teams. This would help her out by giving her some money that she needs...and it would also be a way for me to make some money as I would get a percentage of everything sold.   Today outside of going out was uneventfull as usual. I just laid around and listened to my lungs gurgle as they fill with sticky mucus that is surely getting infected and will soon land me in the hospital.  The other evening I saw one of those news shows like Dateline or something like that. They had a feature on a girl who just recently died of the disease that I am fighting. She had a double lung transplant and things were looking up for her until her body started to reject the organs...While I am comfortable with the fact that I will die soon I found myself eirily moved by her story as it was really similar to mine. She also attempted college but was eventually forced to quit because of health (and that is one of the many reasons I quit as well...maintaining a decent GPA and being very ill and in the hospital is very challenging). While I am sure that she took much better care of herself than I ever have it brought me to a conclusion.  I will not have any organ transplants. Even if the outcome would look bright I simply would not want one. I already wasted my lungs to many things...like smoking pot and other things over the years. Now I know this seems suicidal to do this considering that I have a lung disease...but really I don\'t care what people think about it. Marijuana experiences are among the best memories that I have in my life. It brought many good things into it...mainly through the bonding that I did with people while high.   Some have questioned why I would do such a stupid thing knowing that it would make me sick and land me in the hospital if I did those things. The only answer that I have for that is that I was looking for something fun and something that brought me a little happiness. Drugs did just that. They gave me some great memories (as well as a few I would love to forget), but more importantly I was happy when I was doing them. When I cut them out of my life for a period of six months I saw a very sharp downward spiral in my mental health. It was not that I was addicted to anything because I did not crave any of the drugs that I have done...it was the feeling they gave me that I longed for. They were like a majical happy pill. Overall they were much better than any anti-depressant that they ever put me on. If I could only smoke pot I would be in much better mental state...or even pain meds...those are truely miracle drugs in my mind.   Of all the drugs that I have done I still am in love with pain meds...the legal kind prescribed by physicians. Not only did they take away the daily pain that I am in, but they also elevated my mood to levels that would be considered normal by comparisons...In other words I wasn\'t depressed at all when I would take pain meds.   People may say that I could fall into a pit of being dependant on them or addicted to them...however I beg to differ. First of all I didn\'t do them as a recreation...like I wouldn\'t pop a few pills and go out to socialize or party on. I would take two Vicodans in the morning and the mood that they induced would last all day and maybe I would take one at dinner time to round out the evening. I did this regularly when the drugs were around for me to use. Not once did I experience a loss of effect even when on them for extended periods of time. When I ran out I did not find myself physically craving them either...instead it was more like my body was taking a break...but my mood certainly suffered when I didn\'t have access to them.  During my experimentation with prescription drugs I found that the perfect combination was 20mg of Valium and two 500/5mg Vicodan in the morning with another 10 of Valium and 500/5 of Vicodan in the evening. This gave me the ability to make it through a day without being depressed and gave me the ability to function as a somewhat normal person...alleviating pains that I have everyday...but also giving me that ability to just deal with things in a more rational matter...things didn\'t upset me and I was calmer...plus my brain slowed down sufficiantly to the point where it was not always racing.  Another plus was that all of my screaming migraines stopped when on this combo...Migraines I might add that were brought on by a seizure that was caused by an Anti-Depressant that I was on (Wellbutrin). I have a headache everyday when I wake up and it usually goes to a migraine by mid-day that will continue for several hours...while they do not completely bring me down...they certainly cause a lot of stress in my brain.  I may have to request that I go on this coctail so that I can function a little better in the world...and I would definitely need it if I were to actually obtain a job.  Speaking of jobs the rest of this week I may go over my resume\' and then start hitting up companies in a vain attempt to find employement...but that will depend on my health over the next few weeks...  I see a hospital stay of a week or more in my near future...my lungs are bleeding pretty badly now...and I definitely have seen a decrease in lung function overall...as I can barely make it up the steps of this house without huffing and puffing and coughing after the climb...and that is after I do my treatments...  Well I am out for the eve...got some shit to do tomorrow...like clean the garage slowly to allow for all these collectibles to be close to me for Ebay Auctions...  Peace  '),(881,1234212,'07,April,2003','2003-04-07 00:00:00','   So I wrote some more stuff today on the economy and maybe why we aren\'t seeing a recovery in the Job market even though the economy is techincally growing.   It is a very difficult subject to get into really. What I would need to do to make it a little better of a read would be to do some actual research on all of it to get together numbers so it would look a little more official.   I wonder how many of you out there actually read through the discourse I wrote about religion. I would love to hear from some of you out there in CyberLand...  You may have noticed that I changed the layout ever so slightly....in the right hand corner area you will find a couple of actual links...I will update them as I come across things that are interesting...also I have my email addy posted...just click on it and it will launch your default email client so you can email me....feel free to email anytime. I will usually respond quickly...  Also the first link over there is a link to the Howard Dean Campaign page...After having looked through his page and over his topics I find that he seems to  be pretty balanced out there...I would only hope that he gets the Democratic Nomination as he would be the best to actually take on Bush and his Neo-Con friends...  I used to be a Republican...until this administration got in there and started yanking away many civil liberties, lowering taxes on the people that can afford to pay a few extra bucks, leading us into a war that we cannot win (the war on terrorism....just think have we won the war on drugs yet?), and the seemingly endless amount of nepotism towards companies affiliated with the Bush administration. It brought me to relize that the Republican party is not a good thing to be part of at this point.   Are the Democrats any better? Well I don\'t really know that they are. They have been in the past all for Gun Banning and many other things that I am most certainly against...but right now they would seem to be the party that is the lesser of two evils...so here is a note to the Democrats...get your shit together or you may find yourselves on the lists that the Ashcroft Justice department is building that considers people who don\'t fall in line with the administrations ideals as enemies...and if you are on that list you may find yourself loosing your citizenship because of being a Democrat...sounds harsh right? Well believe me if the Administration gets everything it wants then that scenario would be very easy to imagine...  Whatever you vote for in the coming presidential election just make yourself more informed as to who these people are affiliated with...remember that both parties absolutely hate when a voter is actually informed about the issues and what their candidates stand for....instead both major parties would like you to believe what the PR departments put together for them...with a little research anyone can become what I will call a Dangerous Voter...  New Definition: Dangerous Voter: adj, verb: A voter that does his/her homework and becomes informed on what candidates have stood for in the past, a voter that is willing to vote against party lines as long as it is for the beterment of the nation, a voter that uses their vote as currency in our democracy forcing politicians to trade in nothing but the truth and do the voters bidding, a voter that is not affraid to confront any elected official and call them on campaign processes even if the elected official is in the middle of a speecha and that speech may hold information that is contrary to the actions of the official.  '),(882,1234212,'07,April,2003','2003-04-07 00:00:00','   My theory on Religion and what my resulting religion is.  So here it is.  Q: Do you believe in God, if not what do you believe in? A: Yes and no. I believe in a higher power. Some religions call this thing God, Allah, Jehovah, Buddha, whatever. I call it a higher power. I don’t really call it any of those things. By labeling it anything other than the “higher power” would be limiting that power into someone else’s predefined small little black box description of it. By doing that we limit the power and what it can and cannot do. For me to limit this unknown power seems a little silly.   Q: Do you believe your higher power is directly involved as most religions believe? A: No. For me to believe that any higher power is directly involved in every little thing that goes on here on Earth would be silly. It really is silly in my opinion to believe in that. Maybe I am wrong but I really doubt there is this big ole dude up there sitting and moving pieces around a chess board. I would hope that we have a little more power over our own lives then that.  Q: What is your credo? A: Well I would like to think that I have started to live by a simple code. That you try and treat others as they would treat you. I do believe that the Ten Commandments that are in the Christian Bible are a good roadmap. They lay out a simple code that people should aspire to live to. Now mind you not everyone can do that and live exactly to that code. I wish that we all could because if we did live up to those ideals then we would all be living in a much better world.   Which brings me to this ideal. From what I have read and heard from all the scholars out there most religions preach peace and tolerance. However how much peace and tolerance do you see between all the various religions now a days? Look at the middle east for example. What is that all about? It is simply about religion. The Jews hate the Muslims and the Muslims hate the Jews, and the Christians don’t like any of them. Hell there is infighting in the religions. You have the Sunni’s, the Kurds, Shiites. We have Baptists, Catholics, Protestants, Lutherans. We have all these different sub versions of the same religions that don’t get along. Yet all of the religions preach one simple thing, TOLERANCE. If they would all have a little tolerance then maybe we could get a little further.   Q: So who is your “higher power”? A: Well I really don’t know or really care exactly what it/he/she/they may be. I just know that there is a higher power out there. What or who it is really isn’t something that I can really even give an idea without sounding like a crackpot.   Q: Do you believe in evolution? A: Yes, but not to the extent that we all came from a puddle of goo. Evolution is pretty much a fact. We can see throughout time how things have evolved and to think that it is not true or to deny it is really like trying to deny that the Earth revolves around the Sun or some other thing that we know as fact.   So if I don’t believe that we evolved all the way from goo then where do I think we came from? Well I do believe that the higher power did indeed have a hand in us coming about. However I believe it is much different then it going “poof” your there. I believe that it was more of a seeding than anything else. He/she/it/they dropped off a bunch of DNA samples or actual critters and then let it go. Almost like a giant science experiment. That the higher power is more or less a scientist would make sense to me. In other words as a good scientist the higher power only really intervenes whenever absolutely necessary otherwise it just lets the experiment run and it will to its conclusion.   Q: Do you believe in Christ or “the son of God”? A: Yes as a matter of fact I do. However I don’t believe in it as deeply as religion would like me to believe it. I believe that in fact these people that pop up throughout various religions who are “sons of god” or “gods” themselves; that they are emissaries to us from the higher power that is out there. Indeed they may have powers that we may not have such as healing and the ability to walk on water. Who really knows? I believe though that they were sent as emissaries to kind of guide us and give us some pointers along the way. Unfortunately we haven’t listened to them and have pretty much gone about our own way. These are the times that I believe that the higher power has actually tried to get involved not so much to control, but more to guide and see what happens. Again unfortunately we haven’t really heeded those ideals brought forth by those emissaries. It is rather unfortunate for us really.  So where does this leave me? For me really it gives me the option of expanding my ideals as my personal knowledge grows. It gives me the ability to adapt to new ideals and situations quickly and without much religious fervor. It gives me the ability to see many different points of view on religion yet be able to look at them all and see the remarkable similarities between them all.   I am open to all peoples’ views as none of them really interfere with my own in that we all believe in very similar things except maybe we just call it different things. This also gives me the remarkable ability to worship anywhere. Although I really don’t worship.   In the end I just think that we can believe in whatever we want to believe in. If it makes you feel more comfortable to believe in some all knowing all powerful being than so be it. However if you are going to look at it like that then also look at the underlying religion and ascribe to the basics of it. Which in most of them are Love and Tolerance of all your fellow human beings. In the end that is all my religion asks of you. It doesn’t care about the knitpicky things. What it desires is that you treat those around you with some respect, but also acknowledges that it may not be possible to be this way all the time.  The last thing that I need to address is death. What do I believe in?  This is really hard to say. I don’t think we just die and that’s it. I think we do move on to something else but I am not really sure what it is.  There is an idea that involves quantum physics and some really advanced science stuff that basically says that things are mirrored throughout the universe and this may be the way it is with our brains. That when we die essentially our consciousness moves on even though we are technically dead. I remember seeing a program on Discovery or TLC about it and thinking that it was a reasonable explanation. Maybe the energy that makes up our soul our consciousness stays together even after we die and it moves onto something else, another plain. This kind of fits in with my “everything must have an explanation” way of thinking. If we move on after words that means we are still together somewhere out there. Now that leaves open the possibility that we can still communicate in some form or fashion. This doesn’t mean I believe in Psychics whole heartedly, but it does leave open the possibility that they aren’t all full of shit.  I consider these ideals to be well balanced and something that I can carry through on a daily basis. It does not make me hate anyone because of their beliefs as some religions do. It helps me want to understand the world and universe around me instead of just accepting that some all knowing being has it all planned out for me. It makes me want to be engaged in life and learn from my mistakes. After all why were we given free will if that great power already had everything planned out for us? If free will wasn’t in us then most certainly I would believe that everything was planned out and would be fear stricken if it weren’t all planned out for me.   Maybe that is the human condition or sickness. Maybe we really wish we didn’t have free will, because if we didn’t have it then all would be planned and easy for us. Since we have it we have made our lives hugely more difficult to execute. Personally I like having free will, because it gives me license to question things, make fun of stuff, and really live freely.  '),(883,1234212,'07,April,2003','2003-04-07 00:00:00','   I am booting up the laptop as I write this. So I will finally be able to say that I have my thesis on religion up and online. Damn passwords...laptop...typing here...there...jeesh it would be nice to have one keyboard, mouse, and monitor station for all your computers...this may signify the idea that I might need to soon invest in a KVM switch...  here are my plans for the near future...hopefully over the rest of this year.  First: Get a laptop...nice big desktop replacement type for the road. Sager system with Dual Channel DDR 2.4+ P4 HT ATi Graphics, gig o ram Second: Build a new mini-system like I have now based around a Shuttle XPC with Dual Channel DDR and 3.0+ P4 HT and dual 17 inch flat panels ATi Graphics, gig o ram Third: Wireless Gear 802.11G based or better...a system that supports all the subsets of 802.11 Fourth: Cheapy Tablet PC (some will be coming out this year that are under the 1 grand mark) Fifth: Build from Scrap parts that I have a server up for serving media using drives out of current main machine (think sophisticated NAS) and maybe basic Web-Serving capabilities (but file serving comes first). Sixth: Get a cheap PocketPC Seventh: Get a decent Digital Camera (at least better than what I have now as it eats 10 batteries every twenty minutes) Eigth: Get a decent but cheap Digital Camcorder Nineth: Get a backpacky type thing that will hold all the portable things...and still be somewhat comfortable (hopefully loaded weight will be less than 20 pounds with the laptop, tablet pc, pocket pc, camera, and camcorder, and all the extra cords and crap for them all plus a wireless bridge for when I get to a place that has broadband but no wireless....roll my own network on the fly)  The laptop will be in the 2 grand range, the Tablet in the 1 grand range, the new PC, in the 1500 range, Pocket PC in the 400 range, Camera 350, and camcorder 600 or so...gimme bout eight grand and I could have them all...now mind you I have a lot of work coming up this fall...and I should be able to swing at least a few of these things...I don\'t need them all...but they would be nice...this is just a wish list...  Oh and I wanna get a new Cell Phone later this fall...I am going to continue with Sprint but I want to get a Digital Camera Phone...and a pittsburgh number...my current number is an Indiana PA number...  ok...ok...so now for what some have been waiting for...lemme post this and then I will post the writing I did.  '),(884,1234212,'07,April,2003','2003-04-07 00:00:00','   Ode to the Lung.  Bloody lung I feel your pain Bloody lung you drain my veins In the course of the many years I have treated you shitty Now I shed a tear Bloody lung you teach me at last The you are just a Bloody lung with a spotted past.  Ok so I am no poet...but that is what is up today...bloody lungs are giving me hell...everything I cough up today has friggin blood in it. I may start to make some art out of it or something...I will cough all this crap onto a canvas then laquer it so the colours stay...then I will sell it on Ebay...  I am going to try and get some sleep...sleep is a fleeting thing for me...while I sleep a lot it never seems to be enough...I could sleep 18 hours a day and still be tired. My body is weak and it tires my soul. Yes I do have a soul.   I am just hoping that soon all of this will be past...  *Done waxing philosophically*  '),(885,1234212,'07,April,2003','2003-04-07 00:00:00','   Well at least I know one person reads this blog. Thank you for messaging me :-)  In the course of one day I went from happy, placated, farely well adjusted about being me to a complete shambles.   Of coarse my brain has everything to do with that. It kicked into overdrive after writing this afternoon about my dreams. Everything about being spent and not having the energy or real drive to expend energy of coarse got me thinking. What shall I do then?  I really wish somebody could give me the answer because it really is tormenting me right now.   Tomorrow I am going to play some more Go (the game that John Nash created)...or at least the game is based on his idea...and I will try and solve the problem behind it. That problem is a strategy that consistantly wins. If I solve that problem then I solve one of the most difficult problems in Math the \"P = NP\"...if you don\'t get it then head to this website  \"P=NP\" Problem   I certainly don\'t understand all of it but maybe that can kill some time...learning to understand it.  '),(886,1234212,'07,April,2003','2003-04-07 00:00:00','   Is anyone out there? Is anyone reading this. I am pleading for anyone to write me back.  '),(887,1234212,'07,April,2003','2003-04-07 00:00:00','   I just woke up from the strangest set of dreams I think I have had in a long while.  It was a series in which I would get into a dream, wake up then fall back asleep and continue the previous dream. Strange I tell you.  What was the subject of this dream. Its more like who was the subject of it.   These dreams were about my ex Beth. So here goes the entire set of dreams.  In my dreams she was about to marry some dude that I never got to see in the dream. I wasn\'t going to have that and I was going to stop the wedding. Almost \"Freshman\" style without the church and banging on the glass. Anyways I ran into her at a mall and she was surrounded by security people almost like police. When I got close I just said to her \"does he make you laugh like I do\" (a similar line of that in Oceans Eleven). We get into a short conversation and she then alarmedly tells me to scram cause he was on his way back and wouldn\'t be happy if I was there. Pretty soon the security people see me and start chasing me through the mall. I give good chase (here is a little note for you out there. I would love to be able to run and it is little dream of mine but my lungs are so fucked that I can\'t run or be physical more than a mear minute before collapsing from oxygen deprivation...I need to work on that but I doubt I will ever be able to be physical again...it is really frustrating.).   I get to a store and now hear that the police are chasing me as well because the security must have called them and said something, because the police thought I had a gun and was on the hunt for me. I get to the end of the mall and it is some superstore type place with all kinds of goodies and of coarse I stop and look at the electronics because I am a nerd. The police spot me but I vanish into the back of the store, the storage area and make with ninja stealth technics and make my way past several of them back into the store itself. I then get to the front through the super long dairy food isle that bottomed out on the liqour center. I walked past the counters and got outside. Then all hell breaks loose again.  I see from the one side a police vehicle screaming towards me with sirens blasting. So what do I do, run like a bullet. I mean in this dream I streaked across the parking lot (remember this is a super mall type place and it is in suburban hell). I make with the ninja crouching tiger hidden dragon skills and hop up onto the top of a mini-strip mall thing to avoid the cops. I make my way across the roof and get to a stand of trees. I then hop in them. Now the trees are wet like after a rain so I was getting soaked by their leaves. The first ones I jump into are pine trees I make my way through them and into a small stand of older hardwoods. From there I hop down and run across a neighborhood with the police chasing me all the way. I hop back onto a roof and then to another which is another store. I then hope from the store into a courtyard that has this large porn sign above it notifying me that I had jumped into something of a strip club parking lot only I got the backyard place of here. I jumped in the middle of several stripper types and said \"excuse me ladies, if cops come looking say that I wasn\'t here.\' They all nodded and said they wouldn\'t say I was there. I then lept out of there. Now in my dream I must have been even better looking that I am in real life because shit to get any woman to do something like that is nearly inpossible without a large diamond or lots of cash and I had neither in the dream.  Now the next dream in the series. I guess the wedding was called off because I went over to her house only it wasn\'t here real life house. It was this mansion type place that was gorgeous. I did a ninja stealth approach but I soon learned it wouldn\'t be nesecary because they were expecting me. Especially Beth. In this dream something was slightly different than in person. She had a boob job. Her boobs really weren\'t any bigger but they were fuller and just different than in real life I guess. (I haven\'t seen them unclothed in a long time so I don\'t know what they look like) Anyways I go to her room and I just crawl into bed with her. We just lay there and are close to each other.  At this point I woke up. I tried going back to sleep to get more and I did. However when I woke up it was the most peacefull fealing I have ever had. I was just so calm...I don\'t know how to explain it really.  I got back to sleep and the dream kinda continued. This time I went off to get something and came back and she was taking a mud bath. I played with her while she was muddy it was fun in the dream. Of coarse we had sex in my dreams what good dreams are there for a guy without sex attached. However it wasn\'t an illicit sex dream. I knew that we were going to have sex but it was like a movie where we got started then it would flash to the end when we were in post coital cuddle. Strange indeed. Again I was waking up with this sense of calm.  I don\'t know what it was all about and the only thing that I can come up with is that I am lonely. Lonely beyond all belief. However even if their was someone I don\'t think that I could open up to them.  The closest I got to opening up was with Beth, I have told her that and I don\'t know if she believes me on it. Just simply becuase we ended so weirdly I don\'t know what she thinks about that period. I know she was mad about it all becuase she wasn\'t in the loop of my mental illness at that time.   I am starting to ramble because I don\'t know what I need. I can\'t really invest the energy in getting a new girlfriend and I really don\'t know if thats what I want. By investing energy I mean that it was so much mentally from me the first time around that it exhausted me, nearly drove me nuts, and now my health is much worse. That is another point I am not healthy, really I don\'t know how much time I have and there is no one out there who wants to have anything to do with that.   There is no woman out there strong enough to deal with that. To deal with someone they care about or love knowing they will die soon. I don\'t think there is any female that can cope with that. I have seen guys do it, but women don\'t seem to be able to handle that. Besides I really wouldn\'t want to do that to a woman. Fall in love then leave.   This leaves me with a dilema. I am lonely and need that, but how do I get it? Plus someone already has my heart. I have told her before and she tells me to \"stop that\" or move on essentially, but I am not built that way unfortunately.   I don\'t really want sex per-se. I want intimacy.   Jeesh...I am all bound up now...talk amongst yourselves I\'ll give you a topic. What\'s wrong with me?  '),(888,1234212,'07,April,2003','2003-04-07 00:00:00','   Not much to report today...just another day of not much happening.  I got to go out briefly and pick up a money order and some pepsi and that was about it.  I had Tequito\'s for breakfast...they are the perfect food.  Finished watching A Beautiful Mind...that movie brings me to tears it is so well done and you feel for the character. There is most certainly a fine line between genious and insanity. I think John Nash struggles to stradle the line, but in the end he has done it well.  Now I am watchin Monty Python and The Holy Grail. Monty Python was and is ground breaking in comedy. Each one of the troupe had very specific talents and they were all able to showcase them together. If they ever ever tour again I will go to a show. Even if it is overseas or whatever. I will be there just because they are legends.  Tomorrow I will seriously look for a camera. As tomorrow my mom comes back with my credit card. I need that in order to place an order online. I think by tomorrow evening I will be able to get one.  Till tomorrow when something may happen...peace love hapiness and pie  '),(889,1234212,'07,April,2003','2003-04-07 00:00:00','   Hey out there in Cyber Land...another thrilling installment into my life.   Yeah right. My life is exciting as a shoe that happens to be on fire with a village of fleas living in it. Well hell not even that exciting. In that scenario you would have fleas jumping about saying in flea speak \"help me help me, my baby is in there, stop touching me there chief flippet I already have a man!\" Or something like that.   My world consists literally of this. I wake up round a bouts 4PM go downstairs and get some sort of food. I then veg for a bit in front of the TV in the kitchen while eating whatever I grabbed from the cupboard or freezer. Whenever my grandmother comes into the kitchen and starts rattling around (and she comes in there to simply make noise I am quite sure of that) I then gather my Pepsi and head out to the living room to continue watching TV. If something interests me enough to head upstairs and hop on the net I do so, otherwise I just veg in front of the TV watching either CNN, CNNHN, FSTV, Worldlink, Comedy Central, History Channel, or maybe another channel if something catches my eye, but those are my main flipping channels. I then veg until 11PM when I faithfully flip over to Comedy Central to watch Daily Show with John Stewart, because it is probably the best news source in the world. Ok maybe not but it is funny as hell. Then I watch Tough Crowd with Colin Quinn...that show makes me mad because all they do is yell at each other and call each other stupid when indeed they are stupid because they can\'t form sentences most of the time. Then at Midnight I switch to MTV for The New Tom Green Show. It has its ups and downs...overall it is pretty funny.   I veg maybe another hour downstairs before heading back to my room. I grab a few Pepsi\'s and make my way upstairs to my room. Sometimes I watch a movie and surf. Like tonight I am watching A Beautiful Mind (which rocks ass people). The way he saw things is a lot of what I see sometimes...it is weird. Anyways I surf for a bit, watch a little porn, head to my favorite sites which by the way are listed here (Highradiation (Cloud10), HardOCP, Howard Dean Campaign, and then to here to update and rant!). Then I head to here to update this pathetic blog with my absolutely boring life.   After a bit of staring off into space I head to bed. Sometimes early in the morning, sometimes late at night...depending on what you call which.  Well today I have decided that trying to get a Coolpix SQ for the price I want is useless, but I can get a Coolpix 3100 for the better price. Plus it is a little better camera I would guess and it fits into my sexy little camera catagory.  So I will try for that model.  Now off to some porn.  '),(890,1234212,'07,April,2003','2003-04-07 00:00:00','   I am really hating this bidding thing...simply because I am on a budget kinda...I mean I could get an uber camera and spend a bit more than what I am bidding at right now...but I would never hear the end of it because my mom and grandmother are cheap and wanna get the cheapest thing around...even though they wouldn\'t even have a clue as to what is worth the money and what isn\'t.  They would go to wal-mart and pick up the cheapest digital camera and then bring it home like a prize without realizing that first it is a piece of crap, and second they overpaid for it.   I need to be able to go to about 320 for what I am looking for...and I may have to do that. I will just have to hear the heat...but the cost will be readily made up in the ensuing auctions that I will be having.   They wanna get started tomorrow selling all this shit, but the stuff they want to sell is stuff that needs to be photographed first. It is just a big ass mess and I am bought at wits ends with them both. They want it done tomorrow and aren\'t willing to help with the legwork.  Well over the next few days between trying to get a camera and figuring out what I should start my auctions for some of this crap at it will be a busy few days. Plus I need to go out and get a money order and a few other things....like pepsi...I have a pepsi shortage, I am down to a dozen cans and it will be critical soon...hehehe  Its hard to believe that almost a month of this house arrest shit is already over...I wish it would go faster...and hey it may well go faster partly because I have auctions to hold, a virus in my lungs that isn\'t letting go and will land me in the hospital soon, and money to be made...as well as a lot of learning to do!  Piss on me...  '),(891,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I am trying to get a digital camera on Ebay and it is just extremely frustrating...I may have to up what I will pay for one.  I am still trying to get a Nikon CoolPix SQ...small sleek and sexy. Something that I need in my life, cause I am have none of that :-)  I am in this strange place that rarely is tread by any person. To tread here is dangerous for the soul. To be in this place is toxic for life and all its joys. This is the place that I call home. The place that comforts me. It is the place where I know what is coming next is not anything unexpected. To live here would kill others, but for me I thrive. If you think my life is easy, walk a while in my place. If you think that I like living this life then you may be wrong. I don\'t like my life, but I accept what it is and have decided to thrive within the parameters set forth by whomever or whatever. To thrive is to live, to live is to die, to die is to enter unto peace.  '),(892,1234212,'07,April,2003','2003-04-07 00:00:00','   Well there are two sites that I love to peruse every now and again...one is the infamous Tucker Max and the other is TardBlog (also from the creator of Tucker Max)...there really is a guy named Tucker Max and his web site is hilarious...he has the most insane stories out there...he is one of my hero\'s. The Tardblog is a recalling of a special ed teachers daily life. Yes she teaches tards.  Well Tucker Max has been involved in a nasty court battle that has actually kinda caught the establishment by its balls. He has stories posted on their of his many conquests and one of them happened to be someone semi-famous Katy Johnson Miss Vermont 2001 + 1999. They have both been featured on MTV...Tucker on Sex2K and Katy on Made \"I wanna be a beauty queen\".   Tuckers story of Katy was pretty harsh but from what I have read and what I believe it is true...she sounds like one hell of a character herself that is in need of a little therapy. Who knows...maybe she does.   Anyway she sued to have the stories of her pulled from the site as well as any mention of her name as well as any mention of her title as Miss Vermont. Well this simply flies in the face of a little thing called the First Amendment. You know the one about free speech. If you never have heard of it the amendment basically says you can say anything that you want anytime that it is your right. Now there have been limitations on it like yelling fire in a crowded theatre and such, but never has anyones right to tell a story been pulled. Until this case was brought. Basically a judge in Florida (yeah that state with all the old people who aparently can\'t figure out how to vote, a bunch of Cubans, and a dearth of pretty much mindless beings) told Tucker to pull the story down and ruled that he couldn\'t tell that story...against the First Amendment.  Well to kinda cut it down the order has been struck down, Katy lost her motion, Tucker got his groove back, and well we all get to laugh at the pitiful attempt at yet another poor hapless soul who needs to get a friggin life....  Oh and one other thing you can\'t say is this \"I want to Kill The President of The United States of America\". That gets the Secret Services cockles up...that don\'t like that...even though it really is protected speech.   I will probably be hauled off in a few days for saying that...but I would still have one hell of a lawsuit...its free speech baby...I have never made plans nor would I to kill the current president...even though he is a lying sack of cow shit that has sent many to war to die all in the name of power and money...I hope he enjoys killing people...because he really is a sack of shit.   Anyone who has graduated with a master\'s degree should have even the most basic command of the english language...but our president doesn\'t...hell a Japanese teenager that goes to \"engrish\" class probably has a better grasp of it than that horses ass!  Can I say how much I dislike George Bush...I hope is defeated in this election, then I hope he goes home to Texas and is run over by some of his cows....and then whoever is in office at the time will award those cows a medal.   George W. Bush is no better than any other dictator in history and should be brought up on war crimes trials as well as he should be impeached!  '),(893,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I got into a semi fight with my mom and grandma today...  Here was the scenario...we were all sitting in the living room with our neighbor...now my neighbor has a retarded daughter that is in her thirties...and we were talking about my dog and I said \"yeah he really is a retard\"...it never once stuck in my head about my neighbors kid...and really it wasn\'t meant to offend...and I think that the neighbor didn\'t take offence (and if she did...if she can\'t take a joke, then fuck her). This comment was not meant with any malice because my dog is indeed a retard and dumb...does dumb shit all the time...  well she left to go over to her house to show her husband a signed football that we have and no sooner does she get out the door than I am hit with from either my mom or grandmother (it happened so fast it was a blur) I was hit with something along the lines of \"you have to watch what you say around her because she doesn\'t like the word retard\".  Now instead of leaving it there my grandmother continued on about it for another minute or so. Now it would have been fine leaving it at the initial comment as I fealt bad about it,  instead they continued on...  I got mad and said that I didn\'t mean any harm and they should let it drop...the next door neighbor is a big girl and can handle a battle if she so chooses...again I felt bad about it but now I was mad because they kept on.  My mom called me insensitive...which I am about certain things...like handicaps and such...hell if someone wants to call me a handicapped broken gimp I wouldn\'t take offence...I would just fire it right back at them.  It just kept on happening and happening and my grandmother was getting more rude about it...finally I snapped and told her to be quiet about it...it was over and the point was moot...that made my mom even more pissed...I just wanted to drop it.  It finally got down to my mom and I after my grandmother stormed off...and I told my mom had they both let it drop after the initial blasting it would have been fine but instead my grandmother chose to take it to the next level...now she does this all the time...it is like a little dog nipping at your heels...sooner or later your gonna kick the dog...and tonight apparently I kicked the dog.  I love both my grandmother and mom a lot...they have done more for me that I could have ever hoped for and have been great about all my flaws...but I just can\'t stand being attacked like that for something so insignificant as that.  That is my hell...it is knitpicking and heel biting...and it gets old.  I am dark and moody but I don\'t let it spill onto others as much as possible....when I get like that I stay out of the way so as to not pick fights like this. Because I avoid people and can generally be a crabass I am the bad guy in it all.  Well tonight I put in a bid on a Coolpix SQ...hopefully I will get this one cause I need a new camera for all of this...and I think that would be a good camera for that and for just toting around as it is teeny tiny...plus it looks K-Rad...well hopefully I will win the auction cause that would be hella sweet...if I don\'t get the auction there are a couple of packages that are reasonable for around 500 bucks...and that would be kewl as hell any way you look at it...the packages are worth it seeing the camera sells for about 400 in the stores...and with the extras it would push the price over the 500 mark...and the package includes a larger memory card, extra batery pack, and some other goodies...  Well wish me luck on the auctions...I will need it as I have a lot of crap to sell...like this one football that is signed by a bunch of hall of famers (namath and many others) that would bring around a grand or more depending...plus I have a Kobe Bryant jersey that is signed by him...so that might make a few bucks...lots of rookie cards (shaq and more)...this whole thing will bring in the big bucks...maybe will allow for a lot of goodies plus a lot of cash to invest.  I will get into really selling the shit out of this stuff as soon as I can get a working camera.  Peace out...   Check out the Howard Dean site!  '),(894,1234212,'07,April,2003','2003-04-07 00:00:00','   My mom  got here today and we unloaded the truck...jesus there was a lot of collectables...I would say way more that a hundred grand worth if all were sold for retail...hell maybe even more...  We have 50 or so Basketballs that have been signed by all the teammates on that team...I dunno what all teams but there are a shit load of basketballs...tons of cards, many starting lineup figures, lots of other collectables...I mean just tons of shit...  plus she brought up several boxes of CDs and DVDs...so I will go through those tomorrow...and tomorrow I am going to look for a camera on Ebay...I will definitely need one....and should be able to recoup that cost with the sales of even the first two items...  Well I need to sack out as my med are definitely working now and I am tired...so lets see if I am able to get an uninterupted nights sleep shall we....  Peace out yo!  '),(895,1234212,'07,April,2003','2003-04-07 00:00:00','   I finally did get back to the person that wrote me yesterday...I wrote a long ole\' email.   I also got my picture hosting problem fixed for the time being. Lets see how long it lasts...  Not much to really tell today. I finally took a bunch of pictures of the jerseys we have. I have a Kobe Bryant signed, Shaq signed, Gretsky signed...and more...tomorrow I have to figure out how to take good pics of the bats...  I am having to resort to interesting ways of photographing all this stuff. Tomorrow will be more interesting as I have more unusual shapes to try and accurately capture with a camera.   Tomorrow I should have some of these auctions up and I will most likely pimp some links from here...because after all I am a consumer whore...  Which reminds me I need to see how much bandwidth per month I have on my host because I have a funny video that people should see. It isn\'t bad or gross...it really is damn funny. It is called \"Rejected\" and is a cartoon. Do a google search for the video and see if you can come up with it...really it is funny and if you can find it watch it, definitely worth the time to download.   I will check on my monthly transfer and see if I can host it...or if I even have enough webspace to do so... :-)  Till tomorrow when I may have some links for ya...PEACE!  And BTW if you link to me would you be so kind as to let me know...I will be more than willing to give you a link :-)  '),(896,1234212,'07,April,2003','2003-04-07 00:00:00','   Hello there folks out there in cyber land.  My picture place has for some reason stopped allowing me to link to it. I am working on that now. I was hoping this would all be easy but of coarse it can never be that way.  I recieved a letter from someone this morning that read my blog. Thank you for the kind words :-) I will be writing you later today.  As I have said before anyone who writes me will always get a reply. As I am not a stuck up dude.   Peace  '),(897,1234212,'07,April,2003','2003-04-07 00:00:00','   I am thrilled that I can post pics on here...that will make for some interesting things to come! YAY!  Plus with the new cam I can take pics of things interesting and then share them with you all!  I am also setting up a VoiceMonkey account so people can call and leave me messages and then they will be posted here!  This interweb thing is great!  '),(898,1234212,'07,April,2003','2003-04-07 00:00:00','   This is me...I look like hell right now because I am underweight by about 30 pounds...and I am getting sick right now. However I thought I would post a pic for all you out there who want to know what I look like.  The facial hair growth is bout a week and half old.  Peace  '),(899,1234212,'07,April,2003','2003-04-07 00:00:00','   Just took an IQ test over at  Emode.com  and I scored a 122. At least I know now that I am not a dummy...and considering that I am groggy from sleeping pills I think I did well.  Here is the page they give you when you score those ranges.  Scores  '),(900,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok so I got all of Office Beta 2003 installed and it is da bomb baby. When I upgraded from Office 2000 to Office XP the changes were small and not really noticable. However this new version really is worthy of the Upgrade status for those who are back at Office 97 or 2000. This will be the new Office suite to get a hold of.   Outlook 2003 is really slick and powerful and has some great Junk Mail filtering built in already even in this Beta 2 build. It catches a lot of it from Hotmail and that is great. That is something also that they have fixed in this version. Office XP didn\'t play well with multiple Hotmail accounts, Oulook Express was much better in that sense. Now this Beta fixed many of the bugs and is very powerful.   Today I got my camera. The Nikon Coolpix 3100. It is also a great little camera. I think I will have a lot of fun with it.   Right now I am watching the Al Gore speech that he gave earlier today from  MoveOne.org . So far it is a great speech. Here is a direct link to it.   urlLink The Speech  You will need Realplayer and probably broadband to view it. While Gore is somewhat of a bore...the speech is still good and brings out the idea that we have been mislead by someone who has nothing to gain as he isn\'t going to run for office...at least now.  Watch it and let me know what you think.  Peace (ps there may be another update in a bit once I finish with the speech)  '),(901,1234212,'07,April,2003','2003-04-07 00:00:00','   I also just finished installing MS Office 2003 Beta and I am currently downloading the updates...that may take a while.  From what I see so far the biggest change is in Outlook 2003. The changes look pretty nifty and it will be interesting to see how it stacks up against what I think is one of the best email programs out there...good ole Outlook Express.  So we will see tomorrow when I will be using it more vigorously.  '),(902,1234212,'07,April,2003','2003-04-07 00:00:00','   What is funny about it is that I am watching what it being typed and done remotely from both my screen on my main computer and then on the laptop!  This will be a boon for my little business...as I have people asking questions all the time and now I can remotely log onto their computer and show them without having to make a service call to the site...  I have a guy that I will be going to his house this next week and I will install this so in the future if he has minor problems I can remotely look at them and possibly fix them! Woohaaa!  '),(903,1234212,'07,April,2003','2003-04-07 00:00:00','   Now this is damn cool. I have installed VNC on both my main system and my laptop. I am writing this post via a remote connection to my laptop and it will update view my laptop.   I love technology...it is a little laggy but overall it is a damn neat little program...plus it is free!  Get it  here.   Give it a shot...this may be the way I can update things remotely and stuff...nonetheless it is kewl to play with!  Peace!  '),(904,1234212,'07,April,2003','2003-04-07 00:00:00','   I got an email from one of my friends down south that reads the blog...big shout out to Missy! Thanx for reading it...hope it doesn\'t bore you to much, and I hope that I can make someone laugh from time to time while looking into my life.  Here is a little info for you Missy...and I will email you back tomorrow as I have to be up to recieve my camera that I ordered from  Newegg .  So here is the little update in a convoluted sort of way. I haven\'t been laid in over a year and half and my balls are starting to turn a nice shade of blue...maybe it is because of no sex or my lungs not currently delivering the needed oxygen to them.  I am still sort of looking for work although with my health being on a roller coaster from hell it is hard to secure regular employement when you don\'t know how things will be week to week, but do not worry I do happen to have on of the best doctors for CF in the Nation and he is really dedicated although he is an ass sometimes.  Mom was down there not but a couple weeks ago to clear out the trailer and storage box of all that shit Doug got over the years as he is now in jail...and will remain in jail for some time. He was busted about a month ago for guess what....more credit card fraud. He bought tickets to the Memphis Invitational with a stolen card and then they busted him for it, he also bought a butt load of movie tickets as well. They also got him for having weapons in his car, as he is a convicted felon he is not supposed to have any weapons. The weapons where this knife he got me and a pellet pistol he got me...ironic huh?  So she went down and got all the collectables and it is now up to me to sell them all on Ebay...hence the need for a digital camera that should be here tomorrow.  On the health front I have been in and out of the hospital over the past six months or so several times and I fear I will be headed there again because of a nasty lung infection that I can\'t get rid of...it is becoming a chore to breath because of it, however I have been putting it off for some time because I am now allergic to Pennicillian based drugs and the rest that they can treat me with my bugs don\'t respond well to them...so the next time I go in I will have to be desensitized to Penn. drugs which will entail days of hives and uncomfort...that sucks ass! The weird thing is I am taking decent care of my self and have cut out many things in my life to lengthen the life of my lungs...I still slip from time to time...but hey I am doing better than ever on that front.  I am finally finished with all my legal battles and just serving out my sentence in my little room here in PA. I was busted several years ago as the driver for what turned out to be a large computer hardware theft ring. I really was just the driver in it all...you can Google my name and see what all they said. In the end I came out alright because I was straight with the prosecution on it and gave them the information that really made the case for them as it turns out (because I am anal retentave and kept logs of what was said on ICQ). Right now I am serving a six month house arrest and I am down to five months now or a little less...I will be off of that by just after Christmas...then I have four years probation...plus I have a whopper of restitution (96 grand in conjunction with the other people also convicted). Funny thing is I didn\'t get anything out of it...everything that I got was (which wasn\'t much) was taken away (of coarse!) and the rest was smokin pot...but no where near what would be worth all this crap. So that is that story.  Michelle came up during July to do an externship...she has about a year to go. Also she is still engaged to Arkansas boy...he is a real treat I tell ya. However she is well, although they are making her pull a lot of night duty and you know how she likes to be in bed by 11...so that is pissing her off. She is also worried about the debt she has amassed during school...my philosophy on that is they can\'t get blood from a stone...so they can suck it. She will make herself clear of that problem soon enough.  Mom is ok...but when she went down she found a bunch of letters from women that Doug had been \"carrying on\" with over the internet and through the mail. She also found a bottle of that herbal stay hard stuff...so he was tryin to get some action even while dissin\' my mom. That made her pretty depressed but then it made her mad.  As for Doug he will have to serve out at least his remaining probation sentence down at Parchment...which is three years, plus the prosecutor is going for ten years on the other charges...there are so many of them. They went all Macgyver on him and had him under watch for months...so they have plenty to get him with. YAY!!!!  My blog here is just to keep people up to date because I really don\'t have many real life in the flesh friends right now...I still have people I talk to but I can\'t see them now because of this mess. However that will end and I will definitely get out to see people more often after I am done with it.  Hmmm what else...oh yeah I am going nuts being here at my grandmothers...I daily have to remind myself not to stab myself or her with a fork...fun times!  Once I get done with house arrest I will most likely do a little travelling with some of the proceeds out of the selling I do on this collectables...no shit there is like 100 grand worth or more just laying around here...it is a big task but I will get it all sold just to make my mom and I moola!  Well I will talk to ya\'ll later...and I will email you missy tomorrow...  Peace and thanks for reading!  '),(905,1234212,'07,April,2003','2003-04-07 00:00:00','   Corrections to a past post...Bush will raise in excess of a quater of a billion dollars...that is a shitload of money, and we saw last time that he was able to buy the presidency with nearly that same amount.  Also I would like for some people to consider this. What if Dean gets the nomination nod from the Dems. Which I think he will. Who would be a great VP running mate?  After about two minutes of thought I came to this simple conclusion. I think that if he wins the primary he should give good ole Sen. McCain a call and have him sign up. Now for it to work he would have to keep his party affiliation.   Think about it...Dean if he won the primary would get most of the Democratic vote in the election. If he has McCain he could then garner the support of us Republicans that are disgusted with out party and current Administration but wouldn\'t normally vote for a Democrat (even though I will regardless of the running mate he chooses). Also McCain could bring in the support from the Independants that he held many of until he lost the primary in 2000. He could better bring in those Nader supporters and independants easier than just a fully Democratic ticket.  It may well be \"the perfect political storm\" for Bush and Cheney. As far as I know there isn\'t a law that says you can\'t pick a running mate from another party. The only thing would be is that both parties would be upset because of the choice....but it may be what the people want. Two fiscal conservatives both running on the platform that they will do what is in the best interest of the people, not what may be in the best interest of Washington and the Lobbyists.  Now that is a ticket that many could get behind and support I think. What do you think? Gimme a  urlLink Shout out!  '),(906,1234212,'07,April,2003','2003-04-07 00:00:00','   PS I am getting used to putting links into the text on the page and will continue to do so from now on...HTML is becoming less of a mystery to me now because of Blogspot.  Also I want to let people know that last night I saw  Dean  on Larry King live...I must say his smile is a little creepy...but I can grow to like it.  The only critisizms that I have about the interview are this.  First. Dean needs to recognize that not all his supporters are Democrats and that other factions may well be the ones that help in getting elected. Instead of classifying people as Dems and Repubs he needs to just say \"the people\" as he is the people\'s candidate that just happens to be a Democrat.  Second: He tiptoed way far around the \"gay marriage\" issue. I wish that he would have been more forcefull on it. Marraige liscences are issued by the government and are not a religious thing at all in my book. It is a piece of documentation that comes from the government that is the same as Civil Union documents. Marriage has to long been associated with religion and in our society it is a legal contract, not a religous one. Again the Church and State should be seperate. However in codifying a law to allow \"gay marriage\" it needs to made clear that Religious institutions do not have to accept them but the law should. I am for a homosexuals right to marriage...we let far worse people enter into the legal contracts so I think it only fair that anyone who is willing (or stupid enough) to enter into marriage should be allowed.  Also I read the Time writeup on Dean. It shed some light on his past and his upbringing. Something that people really need to know is that he really isn\'t as far left as most people make him out to be...in reality he is more centrist in the Democratic party and more like a classic Democrat than what is parading around Washington posing as Democrats. The Dems in DC are more like Republicans of old, and the Republicans of the now are more like Nazi\'s or Stalinist\'s then their humble beginnings as they are so far right it would take a few more right turns and then be in 1930\'s era Germany in many ways.  Another thing to consider when you cast your vote. Republicans claim fiscal responsibility, but no Republican president in the last 30 years has managed to even come remotely to a balanced budget. Instead they like to run huge deficits that only favor the wealthy in the end and will bankrupt the government if left unchecked over the next few years.  To balance the budget there will be some things that have to be done and they won\'t be pretty or even popular, like tax increases to levels that were in place before Bush took office. With his tax cuts have we seen new growth? With his tax cuts have you seen your salary increase? Most likely you have lost your job or seen a pay cut (to be fair however not all of that can be blamed on Bushenomics), but instead of keeping taxes at a point where the government could pay down its huge debt they added to it with unesecary cuts that overwhelmingly favor the wealthy.  The only thing that anyone can ask another to do is vote their heart...but before voting with your heart become familiar with what each and every official running for office stands for.  On a side note it was much publisized that Colin Powell will not be back if Bush is re-elected. I am glad to hear that...not because I want him gone but because he is to good of a man to be working in this administration. I would bet a buck that he writes a book after he is out telling of the abuses and plain out manipulation of the Administration during his time as Secretary of State. He is not a Bush goon at all, but unfortunately he has to go with what the administration wants or risks being labled a traitor to his nation. Had Powell been given lattitude on the Iraq bungle he would have most likely upped sanctions and let the country implode without a single American soldier dieing in that hell hole.  '),(907,1234212,'07,April,2003','2003-04-07 00:00:00','   I have been frenetically checking to see what the status is on the camera that I purchased the other day is...well folks it has made it to Pittsburgh and may well be here later today...YAY  For those of you that don\'t know I got myself a Nikon CoolPix 3100 and a 128mb Compact Flash card to go along with it for mad picture taking...now the camera does have an official purpose and use...for taking pictures of goods to be sold on  Ebay . So it will be of great use. I am hoping that I will have some auctions up later this week...I have lots of stuff to sell and it will be great to get rid of some of it all...plus it will pay for the camera!  Yay...just the thought of something coming in for me! woohooo!  One of my former computer students sent me this  book  yesterday and I plan to read it while I am in the hospital next...which I have a feeling will be this month as I am getting pretty damned sick...  O Well...such is life.  '),(908,1234212,'07,April,2003','2003-04-07 00:00:00','   Last night I had this bomb diggity of a dream.  It had to do with campaigning or my idea of campaigning. Grass roots of sorts using new ideas of marketing. What needs to happen for someone to beat Bush is to get to the root of the problem. There is not a great deal of grass roots campaigning going on. With the exception of one campaign and that is the  Dean  camp.   We need to see a return to that to get the people that have been alienated by the whole process back into the process. After all this is our America. Regardless of what you think we all can actually make a difference. It is true that lots of money can buy the presidency as we have seen it in the last election and may well see it again (think about it Bush will spend a quarter of a million on his campaign against whomever runs against him).  How can anyone dream of beating that kind of money. Well we can start by using technology. We can use it to make politicians accountable to us. We can demand to be heard at the debates through Instant Messaging technology. We can communicate with them and have them answer questions important to us and debate the issues that matter to us. We can demand through that a higher standard of politician. If people fire off questions and topics at them in real time then they have to be able to answer us, if they can\'t then we obviously we know that they aren\'t capable of thinking on the fly or are groomed by higher powers.   How can we accomplish getting into these debates. By getting behind someone who will do it for us. Maybe  Dean  will help us in this. Insist debates that include real people not just a selected few, and not just reporters.  I don\'t know whether or not we can do it...but it is worth a try.  '),(909,1234212,'07,April,2003','2003-04-07 00:00:00','   I finally had a question answered today by someone I care about. Of course the quetsion was a loaded one. Really how was it going to be answered. I won\'t tell you all what I asked because I still have a little dignity. Needless to say the answer was no and not again.  I feel a little weird about it because I don\'t even know what I wanted the answer to be.   I think it is time to move on in my life, and I will because it is about time. Moving on is hard to do. Maybe when I get done with all this crap that I am going through I may just go out and try to go buck wild if my health allows me to.  I may just become the next Tucker Max...only of Pittsburgh. I would need do a lot to do that...but I think I may be able to do it. After all what do I have to live long for...certainly not love.  Which brings me to this idea...Love is a farse, something created by man. An emotion that is flawed in every sense of the word. To unconditionally love someone is impossible because we are all capable of stopping that love is certain conditions are met.  So for me I will no longer love as I cannot achieve that emotion. Instead I will lust.  '),(910,1234212,'07,April,2003','2003-04-07 00:00:00','   Today I went and got a camera off of  NewEgg . I got a Nikon Coolpix 3100 and a 128MB Compact flash card for 332.99. Not a bad deal considering just the camera alone retails for 349.00.  Also tonight I went surfing around to read a little more about my man  Howard Dean . The Democratic establishment doesn\'t seem to like this guy because he is not really a true blue Dem.   Now lemme try and set something strait about him in my own words. The reason he is not a true blue Dem. and the reason that I will be voting for him when it is time to vote for President if he is nominated is because he isn\'t a true blue Dem. He has many policies and ideas that don\'t jive with the rank and file.  First he is much more liberal (or conservative) than most Dems. He believes in gun owners rights. It doesn\'t seem like he wants to take any more of those rights away which goes against the Dogma of the Dems. He also is a supporter of Gay Rights openly. While the Dems support Gay Rights in their meetings and thoughts they don\'t really stand up for them when the time comes.  It was also critisized that he really doesn\'t have a true plan to fix things that he sees wrong. My question is this. Should he this early in the game? I mean things change as times move on. A true representative of the people has to be flexable when it comes to things like this in my view and not set on his own agenda for things like it is the bible of their lives.  Sure he may not have a budget plan right now, but how many Dems do have a budget plan that is detailed right now? I bet none of them. The budget is something that can be worked on and proposed whenever he gets into debates with the Republican candidate which will be George Bush.   His ideas are sound. Greater spending on health care, education, and security. Now all of these things can be achieved with a balanced budget. If a guy like Clinton (who really was bright in his own right but not a genious) can do it than someone who is a Doctor certainly can.   What we have right now is a crisis here in America. The health care system is crumbling. It really is in shambles. Partly due to unlimited lawsuits towards doctors. Now I am not for limiting these lawsuits, I really don\'t want to see caps put on because there are times when a Doctor really screws up and messes someone up due to negligence, those lawsuits are justified. However a vast majority of them aren\'t and those are the lawsuits that need to be capped. Maybe a good system would be to put in place a national board that is comprised of both lawyers and doctors who collect facts about a particular case. They then make a ruling as to whether or not the suit has merrit or should be capped. While this would certainly cost a lot of money to do simply because of the shear volume of lawsuits it would be a more efficient way of clearing out the courts of these frivilous lawsuits and then letting the ones that have merrit go onto a jury trial to get \"justice\".   Another thing that is clear is that the Education system is in collapse. The \"No Child Left Behind\" initiative has effectively dumbed down the students of many districts so that districts can stay off of a list of places that need improvement. In doing this they lower the effective testing measures to allow for students that aren\'t proficiant to pass therefor helping the disrtict out. That is simply wrong. Education needs to be overhauled from top to bottom.   My education plan is fairly simple in concept. First we need to get rid of Teachers unions as we know them. They have a lot of resposibility in helping create this crisis by keeping teachers that aren\'t up to teaching in a job because they are in a union. Second we need to make it so that starting salaries of teachers are 100% greater than what they are now. There is absolutely no reason that teachers are living below the poverty level in many places. These people are teaching our future. If they aren\'t able to make ends meet then something needs to be done to alleviate that. One way would be through pay. Third schools need to be thoroughly modernized. There should be a computer for every child. Not because they need toys but because we could cut costs by handing out E-Books instead of real books. The technology is there in TabletPCs now. They would have a tabletpc that would allow them to scribble notes on the book pages, take notes digitally, and in the end test with certainty of lower cheating. It would also give children the basics on computer operation that everyone needs. Fourth the whole book issue needs to be laid to rest. Standardize the texts. Using ebooks we could do this easily. If errors are found in the books then you could release point updates to the students via wireless connections in the classes as they become available, plus it would cut down on publishing costs tremendously in the long run. Fifth schools need to be broken up into specialty schools in many more cases. Test the students and find out what their accuments are. If they are strong in science and math fast track them with more difficult studies in those areas, send them to schools that teach that. If a student is strong in Literary arts and Liberal arts (like art and music) weed them out and send them to special schools to cultivate that talent. We need to start playing off of the talents of our young people and start to realize that we can\'t possibly teach everything to everyone and instead focus on producing better students through specialization. Sixth and finally we need to revamp the entire State College System. Everyone goes through a state test that ranks them by states. Ultimately they can be ranked nationally. If a student has a high enough score to enter a State School then the tuition is automatically provided in total with the expectation that they keep a certain GPA. If you want to go to a private school then the government will pony up the number that it would cost for tuition normally and the student and parents can make up the difference. A note to this though, state school would have to provide education for anyone who wants in. If you are below the minimum ranking then you pay on a sliding scale as to how much the tuition would be. So say if you were one slot down from being accepted (one point) then you would pick up a percentage based on how close your score is to the minimum for full tuition coverage. This would allow anyone who wants to go to school go, but would make the less well adapted pay for some of it. There would be acceptions to the rule for handicapped students that may have disabilities that prevent them from making normal scores. These things need to be done to fix the system. This will cost a lot of money for sure. However seeing that we spend so much on the Military budget I am sure we can find some ways to reduce it or simply cover the extra cost on a federal level. Education is the key to higher paying jobs, higher paying jobs create higher tax bases from which to tax.  The military needs to be fixed. We must have a strong military, but that doesn\'t mean that it should cost as much as it does currently for the level of protection they give us. We could spend more on technology now and in the future the military would be more of a scalpel than a sword and precision strikes would be far more effective than a huge military force. The threats have changed. No longer do we have any vast armies to go against (with the exception of North Korea, China, and Israel). Fighting a large scale battle of old is a waste anymore because we can do more with less and be more precise through technology.   Another thing that is desperately needed is some great public works projects. There are several that come to mind here.  First we have a drought in most of the country. However we are surrounded by seas. The government could conceivably build huge desalinization plants and pipe water wherever it is needed through existing infrastructure (the oil and gas line currently in use...I\'ll get to those in a minute). With these huge plants not only would we need huge projects to construct them (thus creating short term labor jobs as well as engineering jobs) but they would just simply be good for the country. No longer would we have droughts that threaten our existance.  Second there is a plan to build a massive world highway with a bridge across the Bering Straite. Now this project alone would cost about 100 Billion dollars to do. Considering that we have spent about 85 billion for a war and will continue spending billions in Iraq we can afford this project over several years. This project would literaly create so many jobs that we would not be able to fill all of them from existing labor pools. The needs for this project would encompass almost every industry. If you madated that at minimum 50% of the material be bought from american companies you would see an overnight turnaround in the steel industry and construction industry. We can\'t spend it all here because we can\'t meet the demand for material internally. Plus this would be a boon to the world economy which we depend on to be strong so we can flourish. This project would be huge and really what better way could the goverment help out the economy other than start a huge public works project like this.  Third the government needs to spend massively on alternative energy sources like nuclear, hydrogen, solar, wind, and geo-thermal. For the hydrogen market to actually come into existance would require a mandate from the government and massive investment from them as well. The estimated costs to overhaul the infrastructure to create enough hydrogen have been projected in the 10\'s of billions, but no private company will go out on a limb and spend that kind of moola on something that hasn\'t even been standardized. If we could move to hydrogen for vehicles and small power facilities we could elliminate the need for the massive oil pipelines that are currently used for moving gas around the coutry freeing them up for one of the precursers to hydrogen fuel, water distribution from the massive desalinization plants. By the time the plants would come online and ready to flow water we could have most of the parts in place for hydrogen powered vehicles. That would also need to be done, massive breaks and subsidies on hydrogen vehicles for the first four years so that people could afford to move to them. Have kind of like the Beetle of the hydrogen cars to get people started. Nuclear needs to have the stigma taken away from it. It is safe, clean, and affordable if done in the French or British model for nuclear.  Fourth we need a massive restructuring of the mass transit systems. This would help with hydrogen powered vehicles because you can mandate that they be hydrogen powered. This would jump start that whole deal. We need a national light rail system for passengers that works, bullet trains, Maglev systems, whatever to get things in order.   These are a few of my ideas...lemme know what you think. I will continue this tomorrow evening with my ideas because I have a lot of them.  '),(911,1234212,'07,April,2003','2003-04-07 00:00:00','   Today was another long day of nothing. My mom is over and she got some shelves for some of the cards that I have to get rid of. Jesus there is a lot of crap there to go through.   I shit you not there may well be over 100,000 dollars worth of just cards! We even have a Britney Speares card with a bit of cloth on it that was from one of her concerts or something. We have several things from Kobe Bryant...those may sell well. There is just so much to go through.  Tomorrow we should know about a camera hopefully. May be able to finally get something cemented in...cause really I got to get at it with selling this stuff.  Well I am going to get a bit of sleep...well more like tons of sleep. Even with as much sleep as I get I am still tired all the time. Plus I am in the middle of a screaming migraine!  Peace  '),(912,1234212,'07,April,2003','2003-04-07 00:00:00','   I tried to get some sleep indeed I did get a few hours but then I woke up and haven\'t been able to get back to sleep again...and that sucks balls. I am tired but I can\'t get to sleep...I just toss and turn. All this while on sleeping pills...yeah those work great.  Well I am going to try and toss and turn my way to sleep again cause I am tired and need rest.  '),(913,1234212,'07,April,2003','2003-04-07 00:00:00','   Got the links up...now I am going to bed...good night to all...  Blaa blaa blaa...on yeah and a couple funny flash links before I recuse myself to a night of sleep.  Gas-E-Pop   Rant   Night all.  '),(914,1234212,'07,April,2003','2003-04-07 00:00:00','   The desktop is a little different now...the bar is at the bottom and the XP start bar is at the top...also the desktop Icons are off so the desktop is clean all the time.   Well not much to report today. I kinda feel crappy...  Oh I found Devon\'s site finally...I had been a loyal follower of her blog...she was one of the first people\'s blogs I got addicted to. Now I found her new address and I will link her up in a bit.  I also have to put a link over there for Brian\'s film...he wrote, directed, and produced his own film project. It looks to be pretty good and I am looking forward to it being finished and shown. By then I may be actually able to go to the premier! Wooohoooo!  Well let me get those links for ya\'ll.  Have a good night. Since I lost my HD the other day I lost all my scans and pics of collectables and now have to redo all that work as well as listings that I hadn\'t posted....jeesh...  Peace.  '),(915,1234212,'07,April,2003','2003-04-07 00:00:00','   I finally got it all running...yay!   Here is a image of the desktop.     I think it looks A-OK.  Notice the Bar at the top...it acts exactly like the bar in OSX on Apple Computers.  Bedtime now.  '),(916,1234212,'07,April,2003','2003-04-07 00:00:00','   I just got this K-Rad program that emulates the Docbar in OSX...it is very Jizm indeed...now I am trying to make the rest of the os look a little more OSX like...because they have a great look  '),(917,1234212,'07,April,2003','2003-04-07 00:00:00','   To make a clarification to some people out there who think I am a bleeding heart liberal...I have this to say. I am a registered Republican...yes folks I am. I just don\'t think that the Republican party has moved in the correct way. I believe that is has moved so far right to accomodate a small fraction of people and then the rest of Republican (or many of them at least) have followed blindly into the valley without questioning what is right.   Democrats are no better in many senses...they have many of the same problems and have moved away from their ideas.   Here is a simple list of things to clarify my ideals. Gun Control:  I believe in smart gun control. Yes we should have checks, but no we don\'t need databases. We need to enforce the laws that are already in place. I believe that any non-violent American should be able to own a weapon. Even Felons...as long as they haven\'t been violent.  Abortion: I believe that it is up to the woman to do what she wants. I am not a woman and I would never want to impose my morals or views on one (they are already pissed enough). It doesn\'t mean that I think that abortion is right or wrong, it means that no law should be put into place that usurps the womans right to choose.  Military: I believe we need a strong military, but a smarter one at that. It should not cost us this much per year for a military and not be able to send them into war without costing more. Their budget is over 400 billion...yet they need more to go to war. Why?  Education: It should be free and effective. Anyone that displays the talent and the willingness to be a good student should have their way paid for them all the way through college. Schools should be more efficiant...but I think they need a lot more money from the government. The education system gets somewhere around 40 billion from the government (including colleges)...that certainly isn\'t enough.  Health Care: Universal all the way baby. No American should have to worry about a hospital bill. Healthy citizens cost less than unhealthy ones. The way to promote health is to take the dollar factor out of it. If it doesn\'t cost anyone anything then people will go to the doctors and take their meds. Everything should be covered including Plastic Surgery for those who need it (like burn victims and such). Now even Doctors are believing it would be the more economical way to go...7000 of them have banded together to promote the idea here in the States. We are the only first world nation that doesn\'t have Universal Coverage.  Environment: A strong healthy environment leads to a strong healthy economy. Our economy needs the environment. Without it no one lives and business dies. Strengthen the EPA and other agencies giving them enough teeth to even shut down companies that are killing the planet. Energy conservation and alternative energy resources need to become center stage in the environmental issue. We should be the leader in the world not the laughing stock.  Any others that I forgot. Email me from the bar on the right there and ask...I will answer.  '),(918,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok so New York, part of Ohio, Detroit, parts of Canada all are having a power outage.  In New York City we have yet again seen the best of humanity and Americans...and I don\'t say that litely. We are seeing it because people are banding together and acting calmy and rationaly. However I don\'t think that calm will make its way to Washington. I think that Washington and the Bush Administration will have a knee jerk reaction and end up causing more problems then it even comes close to solving.  Tonight on Larry King Live California Governer Gray Davis, Hilary Clinton, and some other people plus one Bush energy special interest lacky were on. Davis and Clinton both made great points. That the Bush administration has consistantly thrown up roadblocks to progress on the road to making our utilities better. One of the things that the administration would like to see is national deregulation of the Electric Grid. We saw that in California when this act happens that energy companies and traders come in and make shortages to boost profits and generally make things worse for consumers. California is looking into a lawsuit agains Enron because it seems that it created an artificial shortage so it could boost prices and make money.  Now both Clinton and Davis said that the key is a truely balanced energy plan that doesn\'t nesecarily include tapping new resource but instead focuses on conservation and alternative energy sources like wind, solar, and other types of energy. Of coarse the administration doesn\'t want that because it would mean that their friends in Big Oil and Big Coal and Big Gas would all see their types of energy start to fall by the wayside causing the prices to drop.  The Bush lackey was saying that what is needed is more energy plants (which he is half correct) and that energy companies need to be able to tap additional resources and that deregulation would help out. Well most Americans know that big business is not in the business of making life better or cheaper for Americans...instead most big business is in the business of making money. The Bush White House is hand in hand with them all.  This is why we need them out. Electing someone outside of the Energy Loop would be a relief. We may see from that person a renewed interest and government money for energy sources like Hydrogen.   On another note did anyone see 60 Minutes 2 the other night? They had a piece on GM and their efforts in Hydrogen Powered vehicles. Some environmentalists critisize them saying that GM is just doing a little dog and pony show to placate us all. I think that if GM wanted to Placate us all they could do it for a lot less money as they have spent billions in research on Hydrogen power. No company spends that much (especially GM) without a plan to make it back from the research. The reason they are spending the money on it all is because when they get all the pieces together they stand to make Trillions off the technology...as do other companies.  Also here is an interesting story of sorts...our military may cause another Nuclear Arms race with some new fangled Gamma Bombs. Read the story  here.   We are in a scary time. I just hope that calmer heads will prevail and we will make it out alive.  '),(919,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok now he is added to the side panel as well. :-)  '),(920,1234212,'07,April,2003','2003-04-07 00:00:00','   Hey there people my friend has started up his blog...he is a great dude and has a lot to say. Brian is very intelligent and quite possibly one of the most gifted people I know...although as he admits he has a little ADD...hahahaha  Well go check it...also he is a big  Dean  supporter...check it.   Brian\'s Blog  '),(921,1234212,'07,April,2003','2003-04-07 00:00:00','   Its early and I haven\'t been to bed yet...but now I can go to bed finally...everything is in order.  I am just glad that nothing is really dead on my computer. I was dreading a dead hard drive but fortunately it isn\'t. Over the next few days it will eventually get back to the original configuration with all the software loaded...  Did I mention that Office 2003 Beta 2 rocks? Yes it does. It is the best version yet. Smaller than before yet more functional. Also Outlook 2003 is just da bomb when it comes to email clients. I am still in awe over it...talk about shock and awe...MS may have it with this package. I plan on getting the full version when it comes out soon as it would be worth the money for what it does and the abilities they have added. The entire Pro package fits on One (1!) CD right now...that is a feat I tell you. OfficeXP had many more disks for all the goodies...and Office 2000 had four discs that you had to load to get all the goodies. While I know that 2003 will probably have 3 or 4 discs when finalized with all the goodies...all of the core components will still fit on one disc which the others could only attempt to do...  Anyways that\'s nuff\' bout\' Office 2003...  Now off to bed...finally!  '),(922,1234212,'07,April,2003','2003-04-07 00:00:00','   yay! I got it up and running...now I just have to respond to emails and stuff  '),(923,1234212,'07,April,2003','2003-04-07 00:00:00','   I am glad that I have DSL...because it is downloading the MS updates at an insanely fast speed!  Woohooo 85MB in around 20 minutes...  '),(924,1234212,'07,April,2003','2003-04-07 00:00:00','   Getting closer now...it only took about an hour to reinstall all I need for basic operation...now I am downloading updates to protect this computer from the Blaster Worm and many others that are out there lurking...  '),(925,1234212,'07,April,2003','2003-04-07 00:00:00','   Although at least my laptop is still running well enough for me to work on....  *crosses fingers*  '),(926,1234212,'07,April,2003','2003-04-07 00:00:00','   Hey there cyber land...hope all of you are making it with this nasty ass worm going round.  Well I have made it so far. I am behind a firewall and I have patched my systems long ago. So all in all I am pretty safe from that...however I am not safe from simple old hardware problems!  The main drive on my system decided that it wanted to drop the boot sector as well as the NTFS tables leaving my hard drive as a big lump of nothing just hanging out in my syste...in other words my computer took a big ole shit.  Fortunately I am well prepared for things like this and I am not phased when this kind of stuff happens. I have all my disks just layin around and I am in the process of repairing it now...but I will kinda be down on my main box for the time being.   What makes me maddest of all is that I had a lot of stuff saved on that hard drive and not all of it was backed up. Fortunately not much was really lost overall...just really my emails that I had recieved and my documtents...as well as my favorites...outside of that I didn\'t loose much and whatever is gone I can live without really.   Oh well...such is life.  Yesterday was bad, today was worse, and tomorrow it will get even more evil.  Peace  '),(927,1234212,'07,April,2003','2003-04-07 00:00:00','   I\'m sorry that I haven\'t responded to emails recently sent to me. I have been pretty busy and my brain is mulling over many things right now...I promise I will respond.  Yes folks I have a full blown lung infection with all the trappings...however I don\'t have a Doctors appointment for another two weeks...originally I thought I had one this week but apparently I didn\'t have one this month at all...sometimes I am retarded about appointments...I can never remember them and apparantly my PDA doesn\'t remember them either...hmmm my PDA...I thought it was supposed to help me be organized.  I got out a bit today. I had a lesson with a student. He had really good questions and they have caused my brain to actually have to process information. I mean these are questions I have never been asked before so I have to actually learn how to answer them...oh joy I get to learn something new.   I posted a couple of things on Ebay...and I have bunches more to put up...I may get some more pics taken tomorrow and get some more stuff up.  Lately I have been doing a lot of online reading...like the post below about the current state of the economy, taxes, and the federal reserve system. It would seem that the whole system is a fraud as it is supposed to be owned by the people of the US...however the federal reserve is actually a corporation that is not owned by the government...it is a federation of banks that are owned by many companies. They buy currency from the government at about 2 cents on the bill or something like that and turn around and loan it to the government at face value...the value of the currency is supposed to be backed by precious metals and by the loans that it makes to the government...however there isn\'t enough precious metal to actually cover the currency in circulation and in addition the loans that the government has taken out over the years could never actually be paid back to the system. They could recall all currency in circulation and not actually be able to pay off the loans.  Also it seems that the government has been bankrupt since the 40\'s or so...and that since then the government has been actually operating under bankruptcy laws of some sort or another. All of it is really interesting.  While I cannot vouch for the validity of this site...it does raise interesting questions that may need to be looked into.  Check it!   It is all very interesting that we the people may not actually own the currency we have and that we may never actually be out of debt. It also raises the question why would the congress try to give away control over currency that is strictly forbidden by the Constitution. Strange indeed...  I am in no doubt that the entire system will colapse. It may not happen anytime soon but it eventually will and our way of life, our nation will fall by the wayside at some point. It is the natural life cycle of things. All great empires or nations go through a period of time where they are at the top then fall or are revamped (after all Brittian while still a monarchy has actually become quite democratic and the monarchy is more of a figurehead then anything). I just hope that when we fall it isn\'t at the hands of foreigners....I hope that we can rebuild our system internally that ends up being stronger and better than before based off of the very documents that were written over two hundred years ago...of course with some changes...but the basic framework can work if protected and executed faithfully.  Me...  '),(928,1234212,'07,April,2003','2003-04-07 00:00:00','   There may just be a new debate over taxes after this little footnote in Legal History...read it for yourself.  Yes folks Taxes are still voluntary and there is no law explicitely saying you have to pay taxes...or at least as far as the Federal Courts in Memphis are concerned there are no laws saying that.  So stop sending your money to our corrupt government and take it back...   Are Taxes Legal...Do we have to pay them?  '),(929,1234212,'07,April,2003','2003-04-07 00:00:00','   Nothing to report today.  Tomorrow will be an action packed day of phone calls and such. I have a lot to do...or something like that.  Plus I need to get a hold of Jesse...if you read the blog Jesse I have a slightly different configuration for you and it is a little better in some aspects...more practical as well.  Well I am going to play a short game of SOF2 then head to bed.  '),(930,1234212,'07,April,2003','2003-04-07 00:00:00','   Here is a link that is open to the video of Rejected. Right click and save as to get it. :-)   urlLink \"Rejected\"   I would like to thank the owner of that site. He has allowed me to use the link.  '),(931,1234212,'07,April,2003','2003-04-07 00:00:00','   Nevermind...got it fixed in a moment of genious!  Included with the link that you will recieve will be a page with instructions on how to download it and play it. I will figure out the easier way later...but this is the down and dirty way right now...may not be pretty but it works!  '),(932,1234212,'07,April,2003','2003-04-07 00:00:00','   I am having problems with it...but I am working them out...I will finish with it tomorrow...   GRRR  '),(933,1234212,'07,April,2003','2003-04-07 00:00:00','   I am working on a way to get that cartoon short posted for you folks. I can host it on my personal webspace with Verizon, however I don\'t know how much monthy transfers they will let me do. I am trying to find that out...if it is a big ole number or unlimited seeing that it is my webspace then I will provide a link for everyone for a limited time...or give out a link via email for those who want it.  Actually you can request that link now if you like. As I don\'t think there are but a few readers as of now, just email me to request it and you will have your link within hours.   urlLink Email me for the link.   Peace  '),(934,1234212,'07,April,2003','2003-04-07 00:00:00','   We did a lot of pricing today...and still there is a long way to go.  Got through a few things. We now have an account over at eBay...so that is good. Now I just need to get a few things posted.  This week I have a lot of things going on. First I have a computer that a bud of mine is wanting to build. I revised the plans to save a few bucks and now I need to get a hold of his ass so he can see the lower price yet still kick ass computer setup we are looking at. (Saved around 600 bucks I think...he wants an Uber Box).  Second I have a job on Tuesday. A student of mine has requested that I come out because he has some questions about some things. That is at 1PM. As far as I know I can work. I really don\'t know what to tell my guy as I don\'t know how long I will be. I might be there and hour, but with this guy in the past it has drug on to maybe two or more.  Third I think I have a Doctor appointment this week on Wednesday. If things are as rotten as I suspect (and how I feel) then I will most likely get a short vacation from the ankle bracelet (I can\'t have it on when in the hospital...hehehe). So that is coming up.  Fourth I also need to talk to my jailor because before all this crap was going on I was volunteering my time at the local library teaching a computer class at the rate of one or two classes a month depending on demand. Once I started getting sick I had to stop it. Now that I have got a pattern down for this crap I can make a night a month to teach I am sure. I think I may be able to get that through.  Also I got a tentative job offer. While it isn\'t in my field...and isn\'t something I really have this huge interest in...a family friend needs help maybe two days out of the week, and I need to be able to get out of the house and make a few dollars without blowing my Government Benefits. So this may be a way to do that. It would be running a kiosk in a mall. So hey...that might just be kewlio.  Well I am tired (as I slept only about an hour last night before I woke up and tossed and turned trying to get back to sleep)...so I think I will forgo the normal routine and just get to sleep.  Peace all!  '),(935,1234212,'07,April,2003','2003-04-07 00:00:00','   I finally got all of the boxes working right. We have to send back Jesse\'s Nvidia 5900 Ultra 256MB card because it draws to much power from the XPC power supply. We are going to change it to a 5600...it will be almost as fast.  Had to format his drives and get everything to play nice...then I worked on his old computer which he is giving to his dad. All in all I have spent way to much time in front of different computer screens over the past 48 hours...but getting it working was well worth the time and energy. There is nothing like completing a project.   He built up a nice box...here are the specs. Shuttle XPC Base 2.4GHz P4 at 2.52GHz 1 Gig o\' fast ass RAM 2 * 200GB Hard drives Sound Blaster Audigy 2 EX (live drive couldn\'t fit but the card works) Nvidia GF 4200 64mb RAM to be upgraded to a 5600 128MB RAM when it comes in.  Overall it is a fast ass machine. Plus it looks good to. The clear panel on it kinda glows with the power LED (which is blue)...it is badass! I will try and get a pic of it tomorrow if I remember and post it.  Got to get out today...that was kewl. Spent two hours with Jesse just bullshitting and doing things...plus had a Primanti Brothers sandwich...that was good (note to self look up spelling of sandwich just in case).  Tomorrow have a dude coming to look at some of these mini-helmets we have...hopefully he will take some of them off our hands. We sure as hell don\'t need them. From those funds I would build my mom a system similar to mine...cause her box is dead.  Well I am sick and tired so I am going to crash for a bit.  Love to all out there in Cyber Land.  Peace  '),(936,1234212,'07,April,2003','2003-04-07 00:00:00','   Well today has been a long day...we have been fighting with Jesse\'s new computer...got it all running and a power supply that was included seemed to be dead...grooowwwlll! Then the haus of a video card he ordered up just won\'t play nice in these XPC boxes...just not enough wattage from the power supply to keep the beast fed.  Tomorrow will hold interesting times as I try to complete it and have it tweeked and tuned...  Stay tuned for yet another saga of...\"Building a computer\" cue echo!  Peace  '),(937,1234212,'07,April,2003','2003-04-07 00:00:00','   Jesse will be here tomorrow. I am building his computer and I am actually looking forward to it. He got some hot stuff and I wanna see how fast it is and get that puppy purring. It will be k-rad.  To everyone out there worried I want to re-assure you that I am not ignoring my health. I am doing everything that I can to stay as well as I can until this coming Wednesday when I have an appointment. When I go I will be headed to the hospital that day or maybe the next depending on when they can get a bed for me.  So until then I will soldier on...plus when I go in I will be more than likely going into the ICU for a few days because I have to be de-sensitised again to Pennicillian drugs...or at least the one they will put me on. It isn\'t so bad because in the ICU they wait on you hand and foot and are glad when they have a semi-concious patient...and I am always at least that.  Well I am going to kinda head to bed as I need to get a little rest...I will post a little longer tomorrow as I will be showing Jesse around his new box as well as showing him more about this blog service. I am also thinking about making a custom layout and see how that goes...I may just fool with the colours though...who knows.  Till tomorrow...peace out everyone.  Again thanks to those who read the site, and even more thanks to those who write me. I enjoy getting email and I would love to get more.  '),(938,1234212,'07,April,2003','2003-04-07 00:00:00','   Added a new link over there...my bro Jesse...got his link up. He just started his blog up and it will be interesting to see what pops up in it...hahaha  '),(939,1234212,'07,April,2003','2003-04-07 00:00:00','   I got this email from Bryan today...  Bryan: FYI, my name is BrYan, not the faggy spelling Brian.  hahaha...sorry BrYan...I won\'t mispell BrYan again...jeesh I love that guy...he\'s the kind of guy that will make someone a great dad someday...hehehee  '),(940,1234212,'07,April,2003','2003-04-07 00:00:00','   Well over the last two days I have seen several appearances by Jim Hightower on his book \"Thieves In High Places\". What he says makes sense and he certainly has no love for our \"president\". Check out his site...the link is on the right.  I haven\'t been to bed yet...didn\'t get any sleep last night as I was working on my mothers computer only to find out there is just no hope for it.   So now I am going to bed after I update myself on todays events.  Goodnight  '),(941,1234212,'07,April,2003','2003-04-07 00:00:00','   Well this is strange...lemme see if this helps...  HiiYAAAAA!  '),(942,1234212,'07,April,2003','2003-04-07 00:00:00','   And here it is...my little corner of the world.  Pics removed for bandwidth reasons...  It may be small but hey...I am a small person!  '),(943,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok so no sleep for me...I ended up trying to fix my moms computer...and boy was it a task to figure out there is just nothing that could be done for it.  I tried swapping motherboards, video cards, hard drives, ram...tried all sorts of things...just everything I have for her is old and broke down...what I built for her this last time was pretty much the last of my spare parts that would work together so it is no surprise that the computer finally gave out. Hell the processor alone was from back in the days (1998)...so it was about time that it all died.  Now I am trying to get my laptop up and running at least half assed so she can have something if she needs it. Updating it as I type.   I also got some pics to share with you all...they are now uploading and I will have them posted shortly.  '),(944,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok so I cleaned my room yet again and fixed it up a little better...I will post a pic tomorrow of it as my camera is downstairs right now and I don\'t want to go get it...did I mention that I am lazy?  Well not much to tell. I ordered a twenty dollar part for my computer...it was five bucks to ship it...jeesh...but I need that part badly  What else...I dunno.  Jess will be coming in this week for me to build his computer so that will be kewlio...I will get at least one visitor while I am on house arrest...plus he makes me laugh on the inside...hahaha  I am also planning to get a day while he is here to go out and at least grab lunch with him...have to run it by my officer but I don\'t think he will mind awfully much as he wants me to get out more often so I don\'t go berserk...he is a nice guy.  Well I am going to watch the rest of A Guy Thing and then I am getting some sleep.  Oh yeah did anyone see the new Man Show last night...Rogaine Blows donkey balls...he just isn\'t funny at all. Stanhope shows hope as he is just as funny as Adam and Jimmy...Rogan had a game show spoof of identify the ass crack or something like that. The funny thing about that is it shows just how much of a tool he really is. He is nothing but a brownose little bastard with bad hair. He really had to kiss ass to get this job.  Also did anyone see tonights Daily Show with John Stewart...well he is funny as hell and I think that show is probably one of the finest pieces of television programming anyone ever conceved. I am just so glad they are there to make fun of it all because if they weren\'t then the news would be soooo depressing.  Peace  '),(945,1234212,'07,April,2003','2003-04-07 00:00:00','   Again I must say RealVNC kicks ass...  Wow...today I was laying on the couch and realized how sick I am...and that sucks. I have to make it through another week and a half then I will be in the hospital...guaranteed...that will be a nice vaction.  There is no doubt about me going in...and I am not blowing the horns for nothing...I can barely breath...but I can\'t go in just yet as I have to build a computer at the end of the week for my friend and plus I don\'t have an appointment until the 27th...  Oh the joys of lung disease...  Eesh...  Ann Richards (former Dem. Gov. of Texas before Bush) was on Larry King live a few days back and they ran her interview a couple of times this weekend. While I never really liked her she made some great points...and after having seen her interview I gained new respect for her because she was very outspoken and well spoken about Bush and his policies inside of Texas...she also gave a psuedo endorsment to Dean which was cool.    Dean for Pres!   And Brian...did you get mom\'s message...she is coming around to Dean...finally! She especially liked his views and ideas about health care and insurance. I think she may just vote for him...and for those who don\'t know my mom...she is a staunch republican and has done a lot of defending up until I started opening her eyes to his policies and what he has done to hurt america (bush)...now she is coming around especially after I told her this next thing.   Abuse of power   Cnet write up on it   This websites author and proprietor was convicted of a Federal Felony for providing what would be construed as sketchy instructions for making a molitove coctail. Now this information is widely available around the internet and much worse information is available elsewhere. What makes this different is the political intervention to convict him. He was convicted of some obscure federal law that basically says that if you hand out instructions on how to make bombs and such with the explicit intention that followers follow through and make and use them you are breaking the law.  Nowhere on his site have I read that people should go out and violently take over the government or use weapons to wage war against it. However he was charged under this law.   What is really interesting is this. Under the USA Patriot Act he was eligible for a 20 year increase in sentence because of certain domestic terrorist wording contained within it. The Federal Prosecutor made this very clear and thus forced him into making a deal instead of going to trial because he was facing a 20 year sentence if convicted. The twist is they initially offered him 4 months or something like that if he would just plead guilty. Sounds strange.   What got me was Ashcroft put out a memo to keep track of Federal Judges who go under the recomended sentencing guidelines with things like co-operation and such. He wants to put together a list of Judges who will give the maximum all the time so that Prosecutors can use those courts instead of more liberal or level headed courts. Also this list would be used to figure out who would be advanced based on who gives out the hardest sentences.   Now what he was doing was excercising his First Amendment Right in saying that people should protest the government...there is no doubt about that yet he still got a year in prison (the judge called him a threat to security) for essentially exercising the right. Kind of scary if you think about it. Hopefully someone in the ACLU will take notice and take the case to get him out and have his record expunged plus give him reimbursment for his rights being denied.   The site isn\'t friendly. The site may not be right in what it preaches...but regardless the idea of free speech and the right to it is absolute and should never be taken away...regardless of how disgusting or disturbing it may be.   I watched a show on Nazi\'s in America and they had a similar thing happening in a little town outside of Chicago. A Nazi party wanted to have a rally and a march down to the courthouse of this little town (which had at the time the largest per capita Jewish Community in the US). The town banned it and then tried to regulate what kind of rallies could be held. The organiser took the town to court and eventually it came before the Supreme Court and they upheld the Parties right to congregate and have the march saying it was protected under the First Amendment. The case was even fought by the ACLU and a Jewish Lawyer who basically said that he may not agree with them, he may think that it is hate speech, but regardless no one should be able to trample their right to say what they want to say...even if he didn\'t agree with it.   Free speech protects all speech...it is there to protect the ugly side of humanity. It allows them to be able to voice their opinions regardless of what others think and allows them to be able to put it out there into the public...because that is what our country is founded on. Being able to say what we want when we want about whom we want as long as it isn\'t libelous or slanderous...and that even protects Charicatures of people (as seen in Flint v. That hate spewing preacher).  Something that might be taken into thought is this (mind you I hate no group of people based on Race, Religion, or Orientation because I hate everyone) but the Nazi\'s had one thing almost right (it wasn\'t at the time but it is true now). The Jewish community controls governments throughout the world through its control of the Banking system. While I don\'t believe openly that it is a Jewish conspiracy...it is still a fact that most of the major money in the world through banks is controlled by Jewish businessman. They don\'t openly bring their political views into the the business of banking, but I bet if the Palistinian Authority (which is the ritefull Palistinian governing body) went to get any substantial loans from the World Bank to help them set up a viable government they wouldn\'t get the loans. Think about it...fire back at me if you think differently.  One thing I do believe in...the US Government most certainly protects Jewish interests in the Middle East at the cost of Billions to the US taxpayer and by doing so also puts us in greater danger than any other arrangement we have made with governments around the world. With our protection Israel has become the 6th largest Nuclear State secretely...and single handedly caused the Middle East Arms race that was taken on by Iraq and Iran. In addition to being protected and unacknowledged the nation of Israel has also commited horrible environmental damage because of this program and have kept the one person who spoke out against it jail for 18 years (Venunu...look him up...what happened to him is a crime).   Now I put my soapbox away...and with that I will leave you with a final thought that has been expressed by Jewish people in Israel...it is not Anti-Semetic to speak out against the Israeli government and its policies, instead it is the attempt to bring them into the world community and hold them as accountable for their actions as the World Jewish population wanted the Nazi party leaders to be held accountable after WW2.  '),(946,1234212,'07,April,2003','2003-04-07 00:00:00','   So I am going to try and get some sleep now. Who knows maybe it will actually work this time round...  Not much to report on today. I remotely connected to my friends computer so that I could use his system to order exactly what he wanted from  NewEgg  using a little program called RealVNC. I tell you that is a damn neat little program to play with. I think that will be of great use in the future when helping people out.  Anyways off to bed.  '),(947,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I am most certainly sick.  I will be in the hospital for a few days or more depending on how things go.  I won\'t be updating regularly if at all while I am away.  Don\'t worry folks I will be back.  Till next time. Love one another as you love yourself...and if you hate yourself then just pretend.  '),(948,1234212,'07,April,2003','2003-04-07 00:00:00','   Jumpin jiminy...tomorrow I go to the doctor and no time is better cause now I am really feelin\' the pinch.  Well I am tired and today although filled with sleep and a lot of nothing however I am still freakin tired.  I may post in a bit...I may not. I will however post what happens at the doc\'s tomorrow.   Blah...  '),(949,1234212,'07,April,2003','2003-04-07 00:00:00','   Take a gander a Devon\'s blog...she addresses something that I have spent much time thinking about personally. Why are smart people so inept at social interaction, and why are dumb people so good at it.  It\'s worth thinking about.   Smart people are pent up all day with other smart people and have little time to pratice their social graces therefore letting their social graces slip. Dumb people (usually pretty people) spend all day talking to each other making fun of the smart people and therefore are always honing their skills at socializing.   My take on it.  '),(950,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok so today was a washout. I slept most of the day. This weekend took a lot out of me. First Jesse was here. Second we had some guy come in and look at our collectables. Now it wouldn\'t have been so bad but he literally went through it like a tornado and didn\'t put anything back...plus he was rough with some of it because he wasn\'t interested in it (like star wars figures and spawn stuff)...  He wants a lot of the shit...but I don\'t think he will come back with a price that will be reasonable. He was going through the mini-helmets and wants them all but I think he thinks that we were going to leave the most valuable in the group and he would just pay 9 or 10 per helmet even for the autographed ones (which we have a few that are valued at around 150 bucks). Well that ain\'t gonna fly at all. Also he just wants all the cards (even the graded cards we have) but again I think he thinks that he will get it all for nothing...but he won\'t. He will have to come up with at least 40 percent of book value.   So that is why I am tired. Plus I am also pretty ill. I go Wednesday to see the doc...so I expect I will be headed to the hospital that evening or Thursday. So in the meantime I have to wash clothes and get ready just in case. If I do go it will be a relief because it will be a vacation from my house as I see it...plus I will not have to really cook...all I have to do is order food.  Well I am going to try and get some sleep again...peace out ya\'ll.  '),(951,1234212,'07,April,2003','2003-04-07 00:00:00','   Did I already mention I love RealVNC....I just logged into Jesse\'s computer and tried a registry hack to fix one of his problems...I think it may have worked...we will see when his machine reboots...hmmm  Ahh well...  '),(952,1234212,'07,April,2003','2003-04-07 00:00:00','   After much fighting with my USB 2.0 controller on my XPC it is now working up to spec. I just had to get the right driver and now it is giving me full speed burning love. I moved my TDK 48x into an external drive setup on USB 2.0 and it wasn\'t wanting to burn at anything greater than 2x...now with the proper driver it burns at full 48x....yay!!!!  I moved my DVD into the case in the free drive bay vacated by the burner. It blends more smoothly with the entire system now that I have the face painted silver. I also did the TDK\'s face silver but it still has a blue translucent drive bay cover. It looked out of place on the silver system. Now it all looks good.  I am just glad I was able to get something working tonight.  Peace outie yo!  '),(953,1234212,'07,April,2003','2003-04-07 00:00:00','   Well Jesse got home and then his computer started acting up again....little fucker. Mostly it was stuff that I had installed but now didn\'t want to work. Great...so I logged in over the network and looked at it.  Seems the video card is acting up...it is sending distorted information out...and I saw it. I think the card just doesn\'t like it in there and when he gets his new card it will be fixed.  The Processor is registering funny...it is running a 2.4GHz but registering as something else...great...  Also he can\'t find his original driver disk for his Audigy 2 Platinum card and you can\'t just download the drivers off of Creative\'s Website (which is strange in itself)...this is just all fubar...I have never had this much trouble with a system...plus I can\'t just sit in front of it and monkey with it till it works...it is a couple hundred miles from here.  I should hopefully be able to figure it all out...given a little time...but this is just frustrating because again I have never had so much trouble...with the exception of my mom\'s pc.  What else...well I painted my CD-Burner and DVD Drive to match my case and external drive case...now it all matches up nicely. That went pretty easily once I figured out how to get the right texture to match.  Well I am tired and I think I am going to get some rest. I am searching for solutions to the problems posed above...hopefully I will find them easily.  Peace out...and never fear I am still a computer guru...I just need time to figure out the problem.  '),(954,1234212,'07,April,2003','2003-04-07 00:00:00','   Wow am I tired...you wanna know how tired...juwanna know....juwannow...well I am more tired than an overworked hooker in Bangkok...do you know how hard they work down there? Pretty damn hard.  Well I saw Jesse off today...got his box up and running after we encountered another problem. I went to turn reboot it today and when I did...damn thing just wouldn\'t....so I had to rebuild the software...had to do the same to mine as well...it sucked ass.  I think it was A: A virus because both computers exhibited the same thing, B: A bad patch from Microsoft, C: Really bad luck. I just don\'t know...but I did get his up and running and mine was a little easier to do.  Right now my body is so tired that my face is cold...I\'m not physically cold...but my nose and upper cheeks are. It is like they can\'t get enough blood to them to get em warm...my eyes are droopy as well.  Well I think I am going to hit the sack...I am tired.  Night all  '),(955,1234212,'07,April,2003','2003-04-07 00:00:00','   Just chillin out in da hospital. I may get out on Monday...or at least that is when they think they will get me out of here...  I have lots to upload to the blog once I get home.  Till then Peace  '),(956,1234212,'07,April,2003','2003-04-07 00:00:00','   Sorry that I haven\'t been in to post before now but it has been kinda hectic here in the hospital. I have been keeping a log so that everyone will know what has gone on while I have been in here.   It has been a decent visit so far and I can\'t complaign to much but I am ready to go home that is fo sho.  Well I will hopefully be out in a few days time. If I find out when I will let you all know.  Peace  '),(957,1234212,'07,April,2003','2003-04-07 00:00:00','   More of the same from outside sources.   Jeesh...the more I read the less convinced I am about all this \"War On Terror\" and shit   Story   I will try and encode the \"Unanswered Questions\" Video tonight depending on what is going on...  Tomorrow I start working for a friend of the family...so I will be out most of the day.  I  will update later on...as right now I am doing five things at once...plus I am getting Jesse\'s computer back up and running like a little kitten.  Peace  '),(958,1234212,'07,April,2003','2003-04-07 00:00:00','   Oh and I got a comments thing up and running so you can fire back directly from the site if you don\'t want to email me.  Although I love getting email...which reminds me I have a few to write and send from while I was in the hospital.  '),(959,1234212,'07,April,2003','2003-04-07 00:00:00','   Don\'t have much to say really...  Just gonna get some sleep...peace all.  '),(960,1234212,'07,April,2003','2003-04-07 00:00:00','   Well yesterday the 11th passed with little much more than a lot of rememberance shows.   Jesse updated his blog. It is a good read from a different perspective.   Well I don\'t have much to say except that I was able to tape the GNN Unanswered Questions show. So I will digitize it soon and may offer it up somehow or another. It really is a good watch. I will just have to check and see what restrictions are on it. Hopefully I can post it.  I am still waiting on Jesse\'s Video Card to come in. It should have been here today but I expect I will get it tomorrow. As soon as I get it I can get it installed and running in his box and get the rest of the niggling problems fixed up as well.   I have a job now. It is a little kiosk at a local mall selling boutique pet stuff. So we will see how that goes. It starts this Sunday.  Also I have to get my resume\' typed up again so I can send it off to a company that I know I would be qualified to do. It would be great experience if nothing else.   Well I am tired and today has been taxing. So I think I will hit the sack...  Peace  '),(961,1234212,'07,April,2003','2003-04-07 00:00:00','   Read this one article...it states many things that I have said...and it came from a senior British Parliament official that is in the same party as Blair. The site is a reprint of a BBC News report.   Interesting Article    A Timeline of that morning.   Timeline    Enjoy!  '),(962,1234212,'07,April,2003','2003-04-07 00:00:00','   Another year has passed since that fateful day in 2001. Another year of events has come and gone. Another year of fear and agony. Another year of lies and deciet.  I morn for those who lost loved ones. I cry for the children who lost fathers and mothers. I ache for the pain that we have enflicted on the world around us.   The people that died, and have died since 9-11-2001 are all heroes in my eyes. Many have paid with their lives for the intolerance of but a few on both sides of the fence. It is not the people in power who have anything to worry about as we have seen.   Bush, Cheney, Blair, Ashcroft, Rumsfeld, bin Laden, and Hussein are all still alive and well, yet they are the architects of the new world we live in. A world of fear and hate towards each other that has boiled into war that has only hurt the poor, hungry, and those less fortunate than others.  Out of this hate will come a new world, and with a little luck that new world will be better. I don\'t hold much hope for that because out of every great tragedy throughout history there have been those that have hoped for our world to be better, and their hope is lost.  The only thing that I can ask out of this period of time is that maybe, just maybe the little people will end up coming out on top.  There are many questions surrounding what went on and what is going on now. Should we be in Iraq? Have we been lied to? Who is benefiting from all this bloodletting? Could it all have been prevented? The answers aren\'t all that simple, but one thing is to be sure. We won\'t know for many years what we have reaped through our policy now.  Inform yourselves, ask questions, question everything, expect our leaders to live up to promises, expect our leaders to lead us, lead ourselves, set policy, become understanding.  In all we all need to sit back and take a breather then come back to the table and start to ration things out. In the end the mistakes of today will be reaped by our children and our grandchildren so if we are to make the mistakes at least let them be made out of the idea that we are trying to do the best thing.  Some readers out there may be pro-war, some anti-war. Either way it doesn\'t matter to me because you can\'t change my mind, but I will never discourage you from voicing your view and opinion as I would expect you would do the same for me.   We could all use a lesson in tolerance, because without that we are lost.   Unanswered Questions   GNN   Infowars   Al Jazeera   BBC New   CNN   These are links to different places around the net. I have included one conspiracy theory site, two outside or alternative news sites, one site for their side, one neutral site (BBC News has become one of the most balanced news organizations in the world), and one site that is firmly American (although maybe a little to the left sometimes).  Visit them all and take in the sights and sounds. Each gives you information. Draw your own conclusions from the information provided. Once you have then at least they will be informed unlike so many others out there.  Remeber to practice tolerance and moderation. Love your fellow human as your religion may dictate, and if you aren\'t religious then respect your fellow human as you would want to be respected.  I\'ll leave it with this much over used by drugged out Raver\'s Motto.  P.L.U.R. (Peace, Love, Unity, Respect)  '),(963,1234212,'07,April,2003','2003-04-07 00:00:00','   Playin with the counter...  '),(964,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok so I added a hitcounter because I would really love to know how many people read the page...hope no one minds. It has been set to not collect any information outside of hits...it doesn\'t log anything but that...  Peace  '),(965,1234212,'07,April,2003','2003-04-07 00:00:00','   Last night I had the most detailed dream I think that I have ever had and it scared the piss out of me. Basically I was an experiment or am turned into a living experiment which goes in conjunction with other dreams that I have had over the past six months or so.   In this strange dream I had sex with this girl and then it was found out that I carry a strain of a very lethal new type of disease that had been bread in me through all of the hospital stays that I have had over the past years. It is a competely new strain of disease that seems to have no cure outside of the fact that I am still living despite it wiping out major amounts of population. I get blamed for infecting all these people and the government is after me because of it, but the kicker is that I was used as a breeding ground for the disease by the government for a new Bio-Weapon that they don\'t want people to know about. Something that if people knew the existance of then anarchy and overthrow of the government would happen.   You could say that I was the Nexus of the disease...but not that I was the only one, many other CF patients are also infected with the similar agents in the attempt to create a new superbug type thing that could be used.   The name of the bug is E. Plura, E. Pulma, or E. Pulmo or something like that. It is basically a nasty bug that acts in a similar way to HIV/AIDS as it can be spread sexually or even through the air when the bug is active and not dormant. The difference about the bug is that its effects are fast acting and people just stop dropping dead all over the place with no real explanation as to what it is. The CDC is brought in but can\'t figure it out and they are scrambling for answers, but they are not really looking for a cure or treatment but to get me and a few others that were infected and seemingly show no signs of illness. At first they think it is a new strain of especially viralant HIV/AIDS because it kills in a similar manor by attacking the immune system then giving them a severe lung infection that breaks down the lungs eventually leading to death. The mortality rate of the new superbug is in the high 90% range.  The government is chasing after me after having identified me as one of the nexus to the new disease, while most of my friends are captured I am able to ellude them through shrewd cunningness on my part and link up with a group of scientists that are trying to figure out what is going on.   At one point I am in what could be considered a bubble shack that keeps me isolated from the people looking to stop it and I start coughing up this mess that is just disgusting and bloody...but not like anything that I have ever seen come out of my body. I play an integral role in figuring out what it is and how to stop it and I along with the scientists that I am working with are working to stop it but the government has other ideas.   They have the city cordoned off so that people can\'t get in or out and have it locked down pretty tight with the military. Instead of trying to stop it they decide to expand the experiment and start spraying the raw bio-weapon into the air telling people that it is an agent that will help kill off the bug. We go through some adventures to stop that and really piss them off. The government really wants to get a hold of me and get me back into a lab to culture and bring this back under wraps.  I along with others are able to get the information of the new bug out to the world so they can get into gear to stop it. Meanwhile the government is blaming terrorists for the new bug even though they created it. To keep us quiet about what we know they go to all lengths to kill us off or capture me as I am not seemingly sick from it...just a super carrier of the bug.   Now this dream I think stems from the idea that I was thinking about the other day in the hospital and pondering while I was there. It has to do with the fact that CF patients are breeding grounds for very nasty and resiliant little bugs. Over time with out lung infections and the responding coarses of anti-biotics have indeed made newer more robust versions of bugs that are becoming harder and harder to treat. Basically we are living bio-weapons labs in the sense that we bread these bugs as a side effect of treating ourselves with modern drugs. I know for a fact that my little bugs have become resistant to all but a few drugs left out there and with the current rate of reocurring infections pretty soon they will be out of drugs to treat me with that would be effective. We would make for a good ground to try out new drugs and also breed resistant strain of psuedomonis strains simply because even though we get treated with all the latest and greatest drugs you simply can\'t kill all of them when you treat for a lung infection. So the survivors in the lungs become stronger and mutate over several times of treatment quickly becoming resistant to the drugs used to originally treat them with.   This isn\'t so far fetched an idea. We could be being used to breed new dangerous strains of bugs that could very well be weaponized and if ever used would wipe out massive amounts of population quickly especially in second and third world nations that don\'t have access to the latest and greatest of drugs. It makes me wonder what happens to the mucus that I am constantly giving to them to send off to test for resistance of the bugs. They have to isolate the bugs and then culture them, then try and kill them with anti-biotics. It is not so far fetched that you could develop a plan of resistance that would keep the patient relatively healthy for a period of time all the while using them us as living incubators (becuase ultimately you would need to be able to get it into humans to see the effects and we also make the process go faster...our bugs may mutake on the order of twice as fast as they can do it in the lab).   This scares me greatly because I don\'t want to be that at all. However I could most certainly see our government using already sick people who are chronically sick to test new things on and develop weapons out of us. Why not...it makes sense to me. It is a hell of a lot cheaper to incubate a live virus or bug in a living host and build it up to be ruthlessly efficient at making people deathly ill.  When I woke from my dream I was shaking and literally scared that this may be happening...and why not? Again we already know that America has a very active Bio-Weapons research program even though we claim to have none. It makes sense to keep the project under wraps because we signed treaties stating we wouldn\'t develop those weapons...but they can culture them in the medical field and call it a natural progression of disease without raising much suspicion.  It is sad that I have so little trust in the government. I think they know a lot more about things than we know and yet we pay for them to keep it hidden from us.  I have often thought what would happen if you got a group of about a thousand people to storm Area51 and other top secret we don\'t have these bases they don\'t exist places...if you had a large enough group with the balls to storm the compound in a peaceful manor they simply can\'t get rid of a thousand people without someone noticing. They just can\'t kill us all and make us disapear...so why not do it sometime and find out really what our tax dollars are developing.   Again I feel that we are being lied to on a much grander scale than we even can fathom.   So what does this mean all in all...yes folks I am a firm believer that we have been visited by something other than humans here on Earth. There is simply to much credible information out there to deny that something is going on. I believe we need the truth because if we are up against something much bigger and bad ass as ourselves don\'t you think it would be better to know so that the world could band together and be ready for whatever may come out way? I think so.  Lemme know what you think...  I am also going to write up the full story and maybe do some illustrations as to what I saw in the dream. It scared the piss out of me.  You heard it hear first folks. E. Plura, E. Pulma, or E. Pulmo...if those come out sometime I thought of them first...and all through a lousy dream. I tried to google the terms and came up with nothing of significance...if you find anything remotely resembling this email me and let me know...as always you can get me over on the side bar there under \"Email Me\" link.  I hope I am wrong...but the last time I had a dream that vivid and detailed the dream came true in many ways...that is why it scares me.  '),(966,1234212,'07,April,2003','2003-04-07 00:00:00','   One thing that I didn\'t tell all earlier is about my shipment of drugs that came from Chartwell.  As usual they kind of screwed things up. You see Chartwell loves to send out way to many supplies, but they never can seem to send the supplies I might need. Today they called to find out what supplies I needed and we went through a list and sure enough they didn\'t send the proper stuff, like dressing kits for my port. They also have the uncanning ability to screw up dosage orders and such as well. On my Benedryl order they have me taking 50mg IV push of the stuff...however I only need about 25mg to get by on. At the 50mg dose I probably would barely get it into me before I would go into a mild coma for an hour or so, or at least what I like to call the \"fog of concentrated benedryl\". I wouldn\'t even get to hooking myself up to the anti-biotic it is supposed to keep me from having a reaction too. So I am having to adjust dosages.   Well that is all. I finally posted the hospital blog. It is a little different format with the oldest post being first and the latest last...my bad. Oh well I am sure the people reading this are intelligent enough to figure it out seeing that I did happen to put a Date and Time stamp on the posts.  Peace.  Did I mention I am severely upset by the way things are going in our country and I am glad that at least if Bush isn\'t beaten in this upcoming election that people will most likely be awakened to the idea that he and his people (except Powell...he is still clean in my book. I don\'t believe that he believes the crap he is instructed and scripted to say) don\'t have what\'s best for the country in mind, only that which is best for them and their people. I just hope people wake up and see that he is pulling us down a path that will only lead to our ultimate untimely demise at the hands of the very people he claims he is fighting.   To fight terror with terror only begets terror. The Israeli\'s have written the book on that. If only the Government would read it.  '),(967,1234212,'07,April,2003','2003-04-07 00:00:00','   To get a full cold boot on my laptop takes forever...never noticed it till now. Well it is old and stuffed full of an oversized OS for its little Celeron 366 Processor so I guess I can\'t complain to much, at least it runs...sort of. Plus it has been off network for a few weeks so it is now searching for updates of coarse!   Here we go!  ---------------------------------------------------------------------  8.27.2003  11:45PM Well today I came into the hospital. I know that I won’t be able to update regularly for people out there if at all so I am going to go ahead and keep a blog of while I am in. It will be kind of a play by play as to what happens.  So today I got here around 5:30 or so in the evening. As usual when I get here things are initially unorganized. Just kind of getting things going is a little anticlimactic. When you think of hospitals you think that things get done boom boom boom…but not really. It takes a while for orders to come through and you have to be asked the same questions at least 10 times just in case someone missed it the first time around.   That is probably the most frustrating part of coming into the hospital or even seeing your doctor. They literally ask the same questions over and over. Six people see you and ask you the same battery of questions and you just think do these people ever talk to each other or write notes…or hell if someone does write a not do the rest of the doctors not bother to read the notes. I mean come on…I am patient about that with residents and interns and such but I still would think it would be nice if they read each others notes.   As for things getting going my port wasn’t accessed until later this evening. Fortunately it didn’t hurt so bad to get it accessed as it has in the past. However now it is just sitting here accessed but no meds are going in just yet. That brings me to the boom boom boom part. Well again it takes time to get all the orders through. So I am waiting until midnight until my first med will come up…looking up at the clock it reads 12:10AM and still no meds…doh!  So this is where I am at right now. I am just waiting for everything to get started. I guess it will all trickle to a full firehose of activity over the next 12 hours.   Of coarse I will be bored over the next few days or however long this will take so I will updating this thing over and over again so there will be one big entry when I get back home and then it will broken into the date and then times as to when it was wrote. That will at least keep my mind preoccupied, which will certainly be needed over the next few days.  Something that is kinda kewlio is that I am getting a break from the house and the oh so k-rad anklet so that is a bonus. I will also try and take a few pics while I am here. I brought the camera so I will snap some shots off as I see something that might be good to get shots of.  Peace out for the eve.  8.28.2003  6:47PM They finally got my medication to me at about 1:15AM so that is when I actually started the IVs. I had a lot of trouble getting to sleep last night I just kept tossing and turning and when I finally did get to sleep it certainly wasn’t good sleep because when I was woken up this morning for the normal ritual of taking temps and blood pressure I was running a fever of 101F. So needless to say I burned off a few brain cells last night. Also when they got me up it felt like someone had worked me over with a baseball bat and I was extremely uncomfortable.   All day today I have felt pretty bad overall. I have had a screaming headache and it just won’t go away. I have had on and off sweats and chills. My body still feels like someone is beating me up. My joints are just balls of pain and that sucks. When I was last in one of the drugs that I was on made my face, lips, and tongue numb, well it is happening again.   This is the bad thing about coming into the hospital because when I am at home things don’t really hurt that badly. When I actually get into the hospital parts just stop working and I end up on pretty bad pain and the doctors are reluctant to put me on anything for the pain because the meds I am on will put an even greater strain on my already overworked kidneys.   So I sit here feeling just disconnected and uncomfortable. I hope tonight that I can get some good rest because I am tired (even though I have slept throughout the day).  Till the next installment…woah!  8.29.2003  2:45PM  Last nights sleep was no better than the night before. It turns out that I am having a slight allergic reaction to the drugs I am on.  I am allergic to penicillin which sucks because all my little bugs are only responsive to penicillin and not so responsive to other drug bases (like sulfa drugs). So I am having an allergic reaction to the penicillin base in Zosyn which is causing me to have these hellacious fevers. This morning I woke up with a 102F fever…yeah brain stew is my gew that plants in the skull.  Also something of note is that my oxygen saturation seems to be falling off at night. When they wake me up in the morning they usually take my temp, blood pressure, and then O2 sats. Well yesterday I woke up with my O2 sat at around 88 and this morning it was around 87 or so. My lowest level yesterday was 84. Now these sats should normally be up around 98 or 99 or 100 for a healthy person and normally I run around 96, but now I am down to a norm of 94 or 93. It just sucks that I have to know all of this stuff.  One thing that being on oxygen is it only helps you feel a little better overall. See the problem is that you get the good stuff in of coarse because you have a great deal of it in your lungs so it can’t help but get into your body, however your body can’t get the bad stuff out of your blood fast enough. So while you add the good stuff you don’t necessarily get the bad stuff out fast enough. That knocks people all out of whack and that really I would guess could be a problem that would eventually affect the chemistry of your blood.   Overall though my stay so far has been pretty good and I have actually gotten a lot of rest so I am looking forward to a few more days of this. Plus you can’t complain when you are almost being waited on hand and foot…that is never a bad thing.  Oh well. I may update a little later.   3:25PM  ELMO TEACHES INVESTING!!! Ok ok…it isn’t bad enough that we already teach young children today to be consumer whores, but now some folks up at Wall Street are going to introduce a video featuring Elmo and talks about investing. Really it is aimed at helping parents to talk with their children about the value of investing. Will this video also cover why mommy and daddy don’t have a job right now? Will the video cover how Mr. Lay has untold millions while all the people who invested in what was pushed as a great investment are now broke with some being destitute as well? Oh oh…wait will the video cover what is acceptable business practices?   If only people could see that absurdity of it all. We are now hawking our children’s children future…our government has borrowed so much money and has such a large debt that we could never ever pay off the debt. However it is ok to tell little Billy to put his nickels in a piggy bank and when he has enough saved up he can go ahead and buy a share of Halliburton.   I am baffled by it all really.   8.31.2003  4:06PM  I am disappointed in myself because I didn’t write yesterday. Really I didn’t want to write today but I decided that I better just get on and write.   In my visit with Dr. Weinberg today I learned that indeed according to blood tests I am allergic to Penicillin and that really sucks. Why does is suck? Well my little bugs are all responsive to penicillin and not so responsive to other classes of drugs. So now Dr. Weinberg is waiting for my sputum cultures to come back so he can see what exactly he has left to treat me with. Ohhh boy I am looking forward to seeing that because last time I saw one of those reports it basically said that the bugs were only sensitive to combinations of penicillin drugs and not much else.   Today was kind of nice because I had this hot little number beating on me for my P.T. which was kewl…I really wanted to say you could beat me anytime…but I wasn’t really ready to be really beaten up. I imagine she would have a very large guy in her life that would easily beat me to within a mere smudge of my life. However she was really cute and I would have done some dirty dirrty things with her.   She did make one comment that I almost seized upon because it would have been pretty good. She said that her face was itchy and it was because of her allergies. I countered with what are you allergic too? She replied that her allergies really kick up in August. My reply was almost this…”its ok you can say your allergic to me. I have that effect on women.” Now that would have been sadly true. Women that I am even remotely attracted to I have this effect of driving them away as if I had a helmet on my head, drool running down my chin, a swastika carved in my forehead, and a knife in my hand screaming out “I love you, I love you!” Yeah it can be that bad sometimes. So you can imagine if you saw that you’d run. Fortunately I save that look for the weekends and I only go hunting in the middle of the week. So you can say that I am repellent to women in general. I always fall into that friend category…which isn’t so bad but it has the unfortunate position of being sexless.   I also have a cute little nurse’s aid…actually several, one who is my age maybe a little younger. She is cute but has made clear (even though I never made any kind of “I wanna hump you” vibes towards her…maybe I just come off desperate all the time) that she has a boyfriend. Hey that’s kewlio because while sure I wouldn’t kick ya out of bed for eating crackers I haven’t even invited ya into my bed yet, so slow down girly. The other aid is a little older than I am…but she is very attractive and I can tell she has something going on upstairs.   Speaking of nurses I have what may be, well I don’t know how to explain this guy really. Something isn’t right with him. He seems normal but I don’t think he is. He is slow or shy or scared. Jeesh he was just here doing his “checkup” very strange. Other nurses just are like check, check done…he draws it out like he is scared to do something wrong. I don’t really have any strange tubes coming out of me (well with the exception of my port in the chest) so he doesn’t need to make sure they attached or anything. I just don’t know. He must have had a nervous breakdown or something or maybe he is a special needs placement (hire the handicapped their fun to watch campaign).   Well my dinner is here way early. I am expecting a call from my dad. That was weird he came up to PA to visit his family in Greensburg and he gave me a call to see if I would run out to visit him, he didn’t know I was in the hospital. So he said he would call me today and try to get down here. Now I am not complaining or anything because hell I don’t really like people visiting me while I am this way (maybe that is why I am soo low maintenance with the staff here because even when I need something I don’t ask unless it is an emergency) but since he made the trip all the way from MS to visit his family do you think that he could swing by the hospital to see his son. I dunno but I guess I would if I were him. He drives back to MS tomorrow (he drove up yesterday so he has only Sunday in town)…I understand that he has only a little time up here but I would still like to see him. After all it has been over a year since I last saw him.  Also last night I puked again which wasn’t pretty. The stuff in my stomach had so much acid in it that it burned my tongue badly as it was coming out. As a result it burned my esophagus and my vocal chords a bit. So I am horse and have throat pain (which sucks ass). My mouth is burned and my teeth feel like I have run steel wool over them until they were raw. Very strange indeed. When my Russian attending Doc saw me this morning he was like “why you not tell last night and get somezing for that. You need tell us when you breaking so we can fix you with medicine.” I wanted to reply I am not a whiney assed little American pussy but I don’t know if that would have made him laugh or offended because he came here (and I presume has chosen our great nation as his home over wherever he is from). However that is what it is. The nurse twit that I had last night came in once to hook me up to my meds. She never came back until about 7AM this morning. So while I can take off my own IV hose hell it would have been nice for her to offer or even do it because after all it was her job. So she was never in to check on me or ask me how I was doing. I would have volunteered the info but I wasn’t going to call for her, for all I know someone else down the hall was doing much worse.   Well enough of this writing for now. My mom is on her way down and my food is here (early…way early). If I don’t eat it she will…hehehehe…  PLUR  11:45PM  Well I learned that Daniel is in. I met her officially last week at clinic and I have known about her for a while. So I went over to visit with her and see how she is doing and we had a conversation for about 20 minutes. It was nice to talk to someone else my age who also talks…that is certainly different. Plus I guess it is just different to talk to what is your equal. I have friends who I talk to but I don’t feel their equal or do I feel they are my equal in many ways, but tonight it was a little different because it was equal in most ways. It was almost like verbal diarrhea between the two of us and we kind of went over what each others boat is…it was nice that there is someone else that is there too, but it is also kinda depressing because she is 27, I am 26, yet we both have to live in this terrible situation that only a few people get to experience first hand and no one would ever want to be in our shoes. I would never wish this on anyone anyways.  I found out a lot about her. Of coarse there is still much more to learn but I can pretty much tell that we have pretty similar experiences when it comes to life. She is a little worse off than I am in some ways as she is bad enough to be on the transplant list (as for myself when it comes time for that I won’t be put on the list) but she isn’t allergic to penicillin. Six one way, a half dozen the other. I imagine that I will be on her health level soon as far as the lungs are concerned because we can’t use the big guns against the bugs right now...so I think I will be headed down that path soon.   So all in all it was nice to talk with someone else in the same type of boat that I am in…what type of boat is that? It is a boat that is quickly sinking that has run out of tar to plug holes with and is overcapacity. It’s only a matter of time…but will we make the time matter.  9.03.2003  2:34AM  The past few days have just seen me sleep a lot. That is all I have really done. I have been pretty boring.  Yesterday (or earlier today because I am still on yesterday as today) my cultures finally came back and Dr. Weinberg was able to see what my bugs are sensitive to. Well as he suspected they are pretty much only responsive to penicillin drugs. So he put me on Fortaz which on the bag it says that I am allergic to it although I don’t know if I am allergic to it or not. The nurse thinks that it is a penicillin based drug so if it is I will be getting hives sooner or later, if it isn’t then we may have found a drug they can use for the next few infections.   Also today I went down to get PFT’s done (pulmonary function test). I had a set done last week which helped bring me into the hospital. There are several important numbers in these tests but the most important ones are the overall capacity of the lung and the small airway clearance or capacity of the lung. Well when I came in my overall capacity was 50% and my small airways number was 6% of normal function. My numbers have increased to a whopping 51% and 12%. The small airways really is the important one because it indicates the amount of scarring in the lung as well as that is where the work is really done on getting O2 and CO2 in and out of the body. If you have dismal scores on the small airways then you are going to feel like crap. Well I feel like crap but I feel better than I did when I first came in. So the low score explains all the resulting problems of low blood oxygen, headaches, and other things.  Outside of that not much else happened over the past few days. I saw Daniel again and we are going to be sure to exchange email addresses so that we can keep in touch. It will be good to have someone to talk to about this crap.  I also informed my mom today that I wouldn’t seek out the transplant list when it comes time. I don’t think that I really want to go through all of that. Partially because I don’t want to endure the torture of waiting for something that I may not get, but also if I do get the lungs and heart that may be necessary then there is the possibility of my body rejecting them and it also doesn’t make me any less penicillin allergic. My mom was a little shocked and wonders why “I don’t want to live?” It isn’t so much that I don’t want to live, but more of a “I have already lived enough let someone else have a shot” kind of thing. Plus I am ready to move on whenever it is time. I don’t want to hasten that arrival but I won’t be mad when it gets here. If only we could all go quietly.  One thing that did come up today was the abortion doctor murderer in FL. is coming up for execution tomorrow. Today on CNN they had his “spiritual guide” dude on. Lemme tell you that the rhetoric that he was espousing was just as bad as any Taliban or Muslim Extremist out there. I pretty much decided that the spiritual advisor should be bumped off as well for good measure. Let him meet his maker so his maker can damn him to eternity for the hate that he espouses. I firmly believe that if you are one of those people in any religion that carries it twelve step beyond the edge than you are going to the place that you don’t want to go. You will most likely end up in your religions hell. By the time you have stepped into that extremist realm it is a good bet you have already thrown away most of the doctrines and have abandoned for all intents and purposes what your religion teaches and therefore have abandoned your god.  Well it is late and I am hungry. I am going to get a bite to eat and then hit the sack and sleep for another day or so…  Peace out.    9.4.2003  3:52PM  Well I did get on the internet the other night so I updated the site to let ya’ll know what is up. It was a short update I know but it is something at least. When I get home I will upload this entire blog so you can get an idea of what’s going on.  Well last night after having a few doses of Fortaz one of the drugs I am allergic too I finally started to see some allergic reaction. I turned pink and started to itch so they had to stop it. So today when Dr. Weinberg came in we came up with a semi game plan to go with. They will first dope me up with benedryl and then they will hit me with the anti-biotic and go from there. He basically said that they really couldn’t desensitize me because I am having to little of a reaction to the drugs. If it was that I was swelling up and loosing an airway then they could work with that…me just turning red, pink, or other colour and itchy they really can’t do much for that except treat the symptoms of that. So I will just have to deal. Benedryl kinda sucks because as soon as I am hit with it I sleep like a baby but without the rest part. Last night they hit me with 25mg of it and I was out in less than three minutes and they came back later to unhook me from the IV and I have no recollection of them coming in and doing that at all. So I go into a coma like vegetable state. So I guess I will just have to deal with that. Great…  Today I got sick, just puked out of nowhere and that sucked ass. They had to give me some stuff for my stomach and that made me feel a little better. This whole hospital thing really does blow. I am getting to the point were I am ready to be heading home even though when I am at home I am in another prison of sorts. At least at home I can sit in my room naked without the danger of someone barging in. Also it will be nice to be back to my fat internet pipe so that when I sit nekkid I can watch some goddess screaming out “faster, harder you big stud”.   Well I think that I am going to lay back down. One thing I am looking forward to is the debate tonight amongst the Democratic hopefuls on PBS. It will be an interesting thing to see for sure and I am hopeful that Dean will be able to pull even further ahead.  Peace.  9.5.2003  5:37PM  Well they started me back on the regiment of Fortaz and benadryl…oh joy! I spent most of last night passed out and slept a lot. It was kind of nice but at the same time it was really what I call fake sleep. I didn’t dream until early this morning so it wasn’t really restful sleep.   I was told today that I might get out this coming Monday if all goes well and I continue to improve. That would be great because I am really starting to jones for a fat internet pipe. Whenever I get a new laptop I plan on getting it with a an 802.11G card so that I can have wireless internet and I think that I will also look into getting a cell phone internet card from someone like Sprint or something and maybe tie the bill onto my Cell phone plan. I just need some internet access anywhere I go cause I am such a geek and need to get my fix every few hours…I guess I am truly addicted to the internet. I just had a thought…I wonder if Pocket PC devices can share their internet connections? Hmmm…because if the can then I could possibly get this one phone that is really phat (pretty hot and tempting). It is a combination phone for Sprint Network. It has a phone, and small keyboard, and is a pocket pc device all in one, and it is a Vision phone so it links into the unlimited Sprint PCS vision plans that are out there. Now you can dock pocket pc devices up to your computer to trade information back and forth and the computer and the pocket pc see each other pretty equally. Now since this pocket pc device is connected to the internet I wonder if it will let you share the internet connection with the laptop or computer that you link it too. The reason I ponder this is because the device itself is pretty expensive but it would be cheaper to have it than it would be to get a separate PC card and plan for a laptop over a years time. So if it can share its internet connection seamlessly than that would be the way to go…plus the phone is just k-rad!  Unfortunately PBS here in Pittsburgh didn’t air the democratic debate and I am pretty pissed about that because I wanted to hear it all instead of just sound bites here and there. Hopefully they will play it soon because I am interested in seeing it all. If they don’t then I will just have to hunt for the next debate.   Well I am going to wrap up this edition of my blog with this thought or opinion.  When watching Crossfire the other night on CNN the question is being raised should we have the UN help us in Iraq. Should we go to them to get support, troop, and money. Well here is my take on it. We went against the wishes of the world by invading Iraq. Instead we had a president come on tv and spout some catchy phrases and drum up just enough support here in the states. So here I go with my own catchy phrase using plain language that GW uses as he is such a great linguist…”Ya bombed it, Ya bought it”. We shouldn’t be crying now that we can’t afford it. The administration should have taken that into account a long time ago before they even really started kicking around invasion plans. Once they figured out a number or even tentative numbers they should have come to the American people with them saying it is going to cost X number of hundred billion dollars, which is what the CAO (Congressional Acounting Office) came up with a whopping 300 Billion dollars at the low end of the spectrum to rebuild them. That really doesn’t even take into account how much it is costing for the military portion which if you look at how the dollars have fallen so far the ratio of military to reconstructing costs so far has been about 6 or 7 to 1. While this ratio will most certainly fall over time it will still cost us a lot of money to keep our troops deployed for the five years that it is estimated it will take to rebuild and get Iraq back up to steam. Now here is the important part of this argument…who gets the money? Well we know that Halliburton will get a large chunk of that over time as they have an already large contract with the military, but the real robbery has yet to come. Many companies will of coarse get in on the action. Already many companies have announced that they have contracts for parts of the rebuilding, while many dollars will flow into companies not many of those dollars will flow to the end worker. As we have seen the economic recovery that everyone is talking about still has yet to really become real, we are still seeing people loosing jobs, but at the same time companies are making record profits. So those companies getting the contracts and the people at the top are getting rich but the common person is seeing less and less out of the money being spent however it is their dollar that is being spent to build Iraq back. It comes down to this all of the work has to be done in Iraq so it will employ mainly Iraqis not Americans. So Mr. Bush thanks for the big mess, and thanks for lining the pockets of the corporate crooks that you so prominently promised to clean up and send away, also thanks for this economy in which the rich get richer and the poor…well fuck the poor the rich man doesn’t see them behind his large fences.   I can’t wait for Bush to be unemployed. It will be a great day when we finally get him and his crooks out of office.  Peace  -------------------------------------------------------------------  There it is folks...enjoy.  '),(968,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok...so I haven\'t uploaded it yet. I am lazy and I haven\'t unpacked the notebook yet. Believe me it is there I just have to dig it all out. All those cords and stuff. Jeesh...ok ok stop whining Mike...just do it now. Ok Mike lemme get it out...  Post forthcoming in a few moments.  '),(969,1234212,'07,April,2003','2003-04-07 00:00:00','   Finally home...I will be uploading my full blog later on this evening...it is good to be back to my room.  '),(970,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok...now I am just perplexed...just saw the Presidential Address to the minions..well here it is folks...  87 BILLION DOLLARS! Muahahaha....Muahahahaha....Muahahahahahahahahahahaha!   With 66 Billion of it going to the pentagon and the war machine.  Ok now don\'t get me wrong...but didn\'t we just cut taxes three times now, go to war in Afghanistan (a place where the Russians ultimately lost 500,000 troops), and then set out on a crusade thru the middle east....where is this money going to come from.  Hell he never even said that we would be able to top our gas tanks off for less than a ten spot. Not only is the American public paying lots of taxes (less taxes the richer you get)...but we are also suffering through what can be called at best a dismal economy...and to top it off gasolene and natural gas prices have risen through the roof...so in the end the common man gets it...in the end.  To quickly rattle off a few better uses of 87 Billion dollars...roads, sewage, water supplies, electricity grid, schools, health care, alternative energy research, AIDs Research, bridges, damns, hell even a billion gallon vat of chocolate pudding would be a better use of the money...  I propose that we all tell Bush and his cronnies in Washington to shove it...or get crackin on a huge vat of pudding...cause we need to see something for that 87 Billion dollars.  '),(971,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I think I may be getting out tomorrow...so that would be really nice.   I have gotten Jesse\'s box in the mail so when I get back home I have work to do which is also great...gives me something to do when I get there and keep my mind off of having to do all the treatments and having to dope myself with Benadryl. I just am not a big fan of drugging myself with that stuff because it does make me really groggy.   Well I am going to get a little surfing out of my system for today...I really think that I am addicted to the internet big time...when I got on for even just a few minutes it was like cocaine for the fingers and the eyes...it was like my addiction came right back hardcore...so yes folks I have found something that I have been able to get addicted to...and that is the internet....heheheh  One good thing is that while I am addicted I don\'t feel the need to knock over a convenience store to get money to get a fix...because in the end the internet is pretty much free for anyone who seeks it out...because we can still freely access it at out local libraries and even hospitals...so I guess while it is like any other addiction that can ruin lives...the internet is a little different in that at least you don\'t have to pay a seedy dude to get your fix...  Enough of that for the day...  '),(972,1234212,'07,April,2003','2003-04-07 00:00:00','   This is something I ran across while on my limited time in the computer lab here. I feel it sums many things up very well...   Assholes get the girls   Jeesh I hope I remembered how to post links...it has been after all a few days since I have done this...oh yeah and I hope to have a few hospital pics to post when I get back as well...I may try to get some of my blog that I have kept posted tomorrow night if I can find a floppy disk to transfer the information over...if not I may be getting out Monday so it won\'t be to much of a problem.  Anyways peace out my niglets....  '),(973,1234212,'07,April,2003','2003-04-07 00:00:00','   Today I got my case painted...so I accomplished something finally...the paint job isn\'t too bad...it needs some touch ups to finish it off...but first I will have to buy another can of paint...I ran out.  Not much else to tell right now...it looks as if Tuesday will be the day that I get my leash put back on...but on the bright side there isn\'t to much time left to go.  I added the comments feature for peeps to use...it is there to interact with others who read the page and to give me quick input...you can even suggest topics for me to write about.  Well I am going to hit the sac and get some shuteye...till next time!  PLUR  '),(974,1234212,'07,April,2003','2003-04-07 00:00:00','   My card is supposed to be able to capture video but it is being a bitch...I have this video that I want to capture and haven\'t even been able to get the card to recognize that it indeed has this capability...and all the software I have right now is just so bad...time to look for something that will let me get that video in to my computer...grrr!!!!  '),(975,1234212,'07,April,2003','2003-04-07 00:00:00','   Well Izzy is finally starting to blow here...I was outside just a few minutes ago and the winds are picking up and it has started to rain a little bit more than what it was earlier.  I love inclement weather. I would really love to go drivin right now...but it is late and I am tired...and I am still on house arrest. Tomorrow I will get to go out and play in it...I have to work a few hours tommorow evening.  I started my mods on my box today and almost (I say almost because I am able to fix it all) screwed it all up. I was trying to make nice curvy windows...but it is very difficult when you just don\'t have the right tools! Tomorrow when I head out I am stopping by a Wal-Mart to pick up some spray paint...I am going to spray paint the entire thing (with the execption of the front plate) black...that means I have to strip out all the guts and go to work on getting it painted. I already got the shell painted with primer...and it will look nice when I get it all worked out. I also have to stop by a radio shack and pick up some LEDs...preferably blue ones. My next purchase will be a new keyboard mouse combo...wireless of course (I had a set but gave it to Jesse...the Logitech setup with the optical wireless mouse was to slow for gaming) and it will be from Microsoft. I still have to cut the windows from the plexi I have (the first set of windows was an utter failure) and then mount them on the case shell. When I get it all done I will post a few pics of it all. I would also like to do an OC mod to the mobo to get a little more juice out of it. I think it will look good when it is all done.  Maybe this  keyboard-mouse combo  if I can find it.  Well it is late and I have to get up for the nurse in the morning...she is replacing my port needle...that is gonna hurt a little, but it is time this set has been in for a couple of weeks.  PLUR  Oh and isn\'t it interesting that Clark is now a candidate...jeesh.  '),(976,1234212,'07,April,2003','2003-04-07 00:00:00','   It looks as if Blogger has fixed the coding problem they had and now my entire page can be viewed...a big ups to them.  So maybe we won\'t see anymore problems with the page...also I need to get my counter fixed so that it tracks multiple pages and doesn\'t register each page view of the archives as a new hit...so that will be a task for later today.  PLUR  '),(977,1234212,'07,April,2003','2003-04-07 00:00:00','   Today I actually got a decent amount of things done...  First was the big bang of the day. I went to the doctor to see how I am doing...and as expected and told in the numbers I am getting better but I am not completely healthy yet. The way I described it to the doc he was amazed to see it in the numbers. So it was an eye-opener for him that I don\'t really complain that much and when I say I am sick or well that they should listen even more closely because my description was dead on.  One good thing came out of the appointment. I have gained an amazing six pounds...so now I am back to around 121-122 pounds...I need to gain more thought, but it is a start.  I got my oil changed today in my poor car...it was way overdue in both miles and time. That was good and while I was there I met some pretty and nice female...unfortunately she is currently in a relationship or I would have humped her right then and there...she was cute. Interesting thing is she is a MRI tech at the hospital I go to often.  While I was there at Wal-Mart I was able to get the tools I need to mod my box. That is certainly kewlio...then I had to go to a hardware store to pick up the correct screws so I could mount the mods. All the equipment came to about 35 bucks...not to bad plus I have the drill bits for another project that my grandmother wants me to do...she wants to have some plexi cut and put into a China Cabinet to keep the dust from accumulating on the china pieces...so I will be able to do that now.  I also went to get my prescriptions filled. If I would have had to pay for my meds they would have easily added up to over a grand. Speaking of health care costs I got my bill for my recent hospital stay. It was over 60,000 dollars! With about 30,000 grand going into medication! Fortunately for me my various state and federal insurance packages cover the entire bill...thank you to all the taxpayers out there. Because of your sacrifices in tax money you are keeping me alive with your money. Thank you a lot!  Well I don\'t know what else there is to tell. Tomorrow I will start my mod project...hopefully I will have pics of it if I can complete it tomorrow...  Till then....PLUR  '),(978,1234212,'07,April,2003','2003-04-07 00:00:00','   Today I slept in late...when I got up I worked on Jesse\'s computer a little more. Got it a hell of a lot more tweaked out now. It is still registering some stuff that isn\'t right...but I guess that is what you get when you tweak the hell out of it.  The video card doesn\'t recognize as the right model...but it is recognized by games and benchmarks correctly so it really isn\'t a problem. It runs fast as ballz too so that is a plus. I found out the the FX5600 cards are actually slower than the card series they were ment to replace...so I tweaked it out with a custom BIOS and now it runs faster and harder than before...so I made up the speed difference in tweaks. Downside it doesn\'t recognize as the correct card to Windows XP...a little odd but hey all the features of the card do work as far as I have seen. Also the processor is still registering as something else...but hey it is running at stock speed like it should. I was hoping that I could overclock it like I was able to with my chip...but the chip we got is really finnicky and just causes errors anytime you bump even a little over the rated speed...but again it runs properly so as long as it does what it is supposed to do I don\'t have a problem with it.  When I first set it up Windows Setup wasn\'t able to recognize the full 200gig drive and would allow me to only partition out 136gigs of it...but if you plug the drives into a system that is already running and format them from within Windows it recognizes the entire drive and lets you format it without having to split them into partitions...so that is what I did...I basically merged the partitions and now he has two big ass drives instead of a 136gig, 64gig, and 200gig drive...they now both show as 200gig single partition drives instead of one being split into two partitions...so that was kewlio.  Now all that is left is for me to match the faceplate up with color of the system...I may change the faceplate altogether to get a better look...but that will have to wait until I take a look at it tomorrow...after I get back from the hospital (I have an appointment).  Well not a lot happened outside of that...I am planning to mod my box and add windows to it...I got some plexy and now all I have to do is get a couple of cutters, some torx screws (to match how the faceplate is connected), then cut out the windows and tap a few screw holes for the plexi...I am doing it commando style...forget about laying the window into the metal...I am going to lay it over the top of the hole and secure it down with torx screws...plus with the limited space on the inside of the box actually trying to lay the plastic inside (or inlayed with the metal in the traditional method) it just wont work...I also have an idea of adding a LED lights to light the inside of the case...and perhaps lay an LED into the plexi to make it light at the edges...I would work with of coarse blue and red LEDs. I can\'t get a cold cathode tube into it (again size is a factor) so I will guerilla style it...also I would add additional cooling holes because my little box is tweaked close to the max so it gets warm (I may add another fan somewhere inside it...depending on what happens). I have a few ideas floating in my head that may get tried out...inluding some wirewrap to tie up a few loose wires.  Well I am fairly tired and I need some sleep. Peace...  '),(979,1234212,'07,April,2003','2003-04-07 00:00:00','   Went to work today...not a lot to tell...I am sooo tired from the past few days and now I am starting to get itchy from my meds...I upped my Benadryl dose and it makes me twitchy...almost like I felt before I had a seizure...  I am just worn thin from the past couple of days...especially with my patience...it is truly a virtue and this one whore who is \"the guru\" of mall kiosks almost got an earfull of hateful words that nearly came spilling out today.  If I have to see that old, fat, retarded, foolish, cow one more time I will snap...I will. She just needs to go away and die!  I will elaborate when I have more energy and when I am not so doped up on benadryl...because right now just about everything is a blur.  Peace  '),(980,1234212,'07,April,2003','2003-04-07 00:00:00','   Started work today and I am now really really tired...I am going to bed...  Jeesh...I have to go in to work tomorrow morning...eeek!  '),(981,1234212,'07,April,2003','2003-04-07 00:00:00','   I have not been able to get the video digitized onto the computer yet...but rest assured once I have the time I will do it and find some place to host it...the video raises great questions that should be being asked by everyone in the Nation.  I was happy today because I recieved an email from one of my favorite \"quirky\" bands called Automusik. They are based out of Memphis. I first saw them about a year and a half ago at a show where they were the openers for The Cory Feldman band whatever it is called...their show rocked ass and really it was quite amusing and different. Almost like Devo on crack with much the same sound and idea with it being really a quirky A/V experience that you have to see as well as hear.  They were nice enough to send me a link to where a recording and pics from some of their concerts is located. So here it is...click the link and experience  Automusik !  Well I am tired and I have to work tomorrow so I am going to get some rest...also of note I got Jesse\'s box up and running so that is a good thing...it is fast as ballz.  PLUR  '),(982,1234212,'07,April,2003','2003-04-07 00:00:00','   I wish that I were healthier...because I would join the military in a heartbeat.  You hear me talk a lot that this war isn\'t just and that we shouldn\'t be there...and I firmly believe that with my whole heart. However that does not relieve me of the great pride that I feel for our troops and the great sorrow that I feel for them as it seems that they will be over there for a long time.   I am one of those people that is not affected easily. Meaning things can happen around me, bad things and I don\'t mind. Seeing terrible things does not bother me nor would it as far as I have been led to believe with my extremely warped mind. I have killed animals (hunting and poaching dogs and cats that were eating our livestock in truely brutal ways yet never felt even the slightest bit of pain or remourse....) I am one of that rare breed that comes along that would see it as my duty, whether right or wrong, to serve my country even if the actions aren\'t just....as long as I was not serving in a capacity that would bring me into conflict with US Citizens...as the US Military Forces are never to be used against the public in any way...which is why Waco pisses me off as well as Ruby Ridge...not in a mad wanna bomb something way...but in the way \"why did the officers accept the orders to deploy against US citizens even if they were breaking the law\"...that is what I am pissed about...law enforcement should have handled it. Not the military.  But that brings me back to my main point. If I could I would serve as a front line troop now. I would sign up. Learn the language (at least functionally enough to get by) and go and show the Iraqi people that there are decent folks here in the US that think it is wrong...but at the same time treat them with respect...we are hearing more and more stories come out that individual units are turning over control to Iraqi\'s even though it is against the edicts handed down...but these commanders and troops are taking the point and making things better...in a humane way...instead of the Bush policy of degradation of the people and steal their resourses....  I would love to be there. I would love to be part of history...I would love to be with some of the best Americans that we have. They put their asses on the line for us whether we like it or not. Even if the action they are in is just or isn\'t...they are still out there with their balls on the line...and for that they deserve the ultimate respect...and the best way that we can show our respect is to do this....  Bring our sons and daughters, sisters, brothers, mothers, and fathers home...now!   More Troops will be needed regardless of what Cheney and Bush think.  '),(983,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I have one more evening with my PC...on Sunday morning I have to rip the ram out and send it on its way to be returned so they can send me a new pair...I am just glad that I have a lifetime waranty...that is lucky ducky for me little rubber mucky...ok enough rhyming for tonight  I am just glad this week is almost over...I look forward to another week...as that will be the first week that I haven\'t had a needle in my chest...yay! However sad to say folks it looks as though I will have to get it shoved back in before to long.   Literally I have been ok over the past few weeks since coming home. I have had a minor cough but nothing productive at all. The maybe four hours after having the port de-acessed, after having not being able to bring anything up for two weeks, I started coughing blood...blood for christs sake...jesus h mary baby jesus fetus christ...out of no where blood! Believe in a higher power I do...I believe that he/she/it fucking hates my guts...and hey no love lost there buddy...hell you smote Joel in the old testament...killed his family and took away all that he had...all because you are a prick.  Anyways enough of my daily cursing of christ...I am going to get some rest...damning god to hell tuckers me out...  Peace all...  '),(984,1234212,'07,April,2003','2003-04-07 00:00:00','   Today I got to see the Democratic nominees debate...well I wasn\'t so impressed with it all. Maybe I wanted something more. I think it is because of the full ticket that they have that none can really get a word in edge wise...that is probably why I wasn\'t so impressed.  I will say that out of all the candidates there the most charasmatic was by far Al Sharpton...whom I despise utterly. First he should be running on the \"Snowballs Chance in Hell\" ticket...but he does bring something that is seriously lacking in all aspects of public service. Character...plain simple character. I love when I see real people...not some stuffed suit. I want someone who is at least from the human species as my leader.  Lieberman is just a lost cause. He needs to go ahead and drop out now and just endorse someone else. He really is just there as stage filler it looks like. It is really sad that a man can be that boring. Just listening to him made me want to go put on some Nirvana and kill myself.  Kerry is scary looking. I think it is because his head looks like a rectangle...did they torture him in Vietnam by sqeezing his head into a cardboard box or something. Again he is just another stuffed shirt up there.  We have Moseley Braun. Another one that should really just get out of it...not because I don\'t think a woman or a black woman could win...but she has yet to bring up any real issues and take to task what is going on in this nation. She is wanting to wage a war for womens lib...when that war has already been fought and mostly won. If she had something more to say I would listen...but so far her running is her running because of her ego.  Gephardt...this is the candidate that the Dems would love to win...either him or Kerry...because they are both part of the Washington crowd...I will make a comment on that in a minute. This guy really needs to come up with something better than saying Dean is like Gingrich...(in some circles that is an honor)...also he has played \"the miserable failure\" line a little too long...once hey it worked...twice ehhh ok we\'ll aplaud..any more and we\'re gonna drag ya out back and beat ya with a dictionary so you may get some new words in you vocabulary.  Kucinich...call me crazy but that little dude has a real bug up his ass...and he actually had a good answer to \"what would be the most unpopular thing you would persue in your presidency\"...his answer was cutting the military and forcing it to be more efficient...then taking that money and putting it into education. The rest of the field had answers but most centered around repealing the tax cuts. I like this little guy because he is more of a Democrat than the rest...not that I am particularly fond of Democrats...but at least his view is strong and towards the left...he doesn\'t seem to want to bend on changing the way money is spent and fixing the environment.  Graham in short is an old fart that just needs to retire to his state...go head out and lounge around the pool with the rest of Florida...I don\'t want an old man running what has become a world in which you have to think fast and be at least somewhat up to date with your thinking...we need solutions not \"well this is the way we did it\"...the way you did it wasn\'t all that effective the first time around or we wouldn\'t be facing the same problems you jackasses put together for us now. So in short....piss off.  This leaves us with three remaining candidates. Clark, Dean, and Edwards.  Clark: First of all I am not impressed all that much by him up to this point simply because he has wavered back and forth on things. I understand the world is not black and white...but sometimes just go ahead and make a stand and take your lumps like a man. If I close my eyes when he is speaking I hear Bush in many respects (at least he sounds like him)...I would hope that people will see through the curtain and just see that the Clintons are propping him up as a way for Hilary to run for VP...with no real chance of winning in 04\' and then setting her up for a strong run in 08\'. I can see if he gets the nod that the Dems won\'t run him hard and if he pics Hilary there would be somewhat of a backlash...as she is not the most pleasant person. Her softer makeover hasn\'t taken complete effect yet (just think Tankles). If they don\'t run him hard then she will be set up for almost an assured shoe in as the Dem nominee for 08\' and possibly the Presidency...I will get to her in a few minutes.  I just want for Clark to stand on his own and fight the battles...I would love for him to run on a ticket with Dean...a Dean-Clark or Clark-Dean ticket would be unstopable period.  Dean: I was slightly dissapointed in his performance today...not because he did anything wrong...but that he wasn\'t as animated as I have seen him in some of his personal appearances that he has made round the country. I want to see the man that takes off the jacket and loosens the tie and gets down to brass tax...I want to see that he isn\'t polished...that he is a man not a politician even though he is a politician. He did very well deflecting the blows that came his way from other hopefuls in that he didn\'t really attack them when they came after him...he told they were wrong and then tried to focus them back onto the bigger issue of getting Bush out of office...very grown up of him...a quality that is sorely needed when it comes to politics...because right now you see on the world stage politicians acting like children yet in the meantime we have our men and women dieing in Iraq...  He needs a little more in the way of tacking down some of his policies and he will be golden...yes we need to repeal those tax cuts...but what money will that bring in...and then what will you do with that money. How much do you think it will cost to give us all healthcare....etc...get those numbers together and show us how it will be paid for and then more people will take this movement as more than a movement and more of a real thing.  Edwards: Why do I put him in this group when Kerry or Gephardt would be better as front runners. Well because I like him. He kept a promise and officially anounced his candidacy on The Daily Show with John Stewart.   Keeping a promise...even as mundane as that is a pretty big thing. So he got extra credit for that. He really doesn\'t have a shot in this field right now...but his ideas are sound and he is charasmatic. He is a young guy that can have the pull with younger voters if he can get across to them.   He will be one to watch in a few years...just a little more under his belt and he may become a good leader in the party.   Now on the Hilary question. She has already promised that she wouldn\'t run for President in 04\'. Which I take her at her word (even though I shouldn\'t). However she has not said that she wouldn\'t stump for VP in 04\'. Now she has stated that she wants to make a run for President in 08\'. Now the numbers don\'t quite work out here. If she gets the nod for VP this year then one of three things will happen...with one only really being a possibility.   First the two out there scenarios. The first is her and running mate win the election. Once in there would be a tragic assasination. She would be in the number one spot. The second is that her running mate makes it all the way through but decides not to run for a second term (it happens but rarely).  The other scenario which is very plausible is this. That her and running mate X run a poor campaign and don\'t go at it hard and loose to Bushco. In 08\' the Republicans won\'t have a good candidate for Pres...come on Cheney (my ass)...and she would have a relatively easy time of a campaign because of another four years of Bushco in charge...by then even hard core hillbilly republicans will be crying for anything different.  I want a woman to win the office I really do...but I don\'t want her specifically. I don\'t think she would be a good one...I would have to see how she talks and see some more of her ideas...also see if she has changed some of her views. Then maybe I would vote for her...but if it were today I wouldn\'t.  Coming to my last point of the evening or this post is this. I am tired of seeing stuffed shirts. I want someone who is entertaining. I want my candidate to be human in every way...hey if the F-bomb were dropped every once in a while like in this context \"what the fuck was he thinking? Come on people!\"...I think that could be tolerated today. I don\'t want our politicians to be potty mouths...but in real life people swear...they should too and not be ashamed of it. That shows they are human and prone to slipups...also sometimes those words are really the best to describe some things.  I want to see my candidate in jeans and a decent shirt...just like I dress in my day to day life. I don\'t want them to make me feel inferior because I can\'t afford a 2100 dollar suit. Hell go to a debate in that for all I care...actually do that because then it shows that the pomp and circumstance is just that...it has nothing to do with the issues...hell wear PJ\'s and bunny slippers...but for God\'s sake just be a real person.  Scripting is great...if you are an actor (Ahnold) but not if your a politician. If you are truely informed you should be able to give a speach about things. Sure you write down some notes...but damn if you can\'t take a few questions and answer them (without having known what they were before hand...no more of this scripted Q&A sessions bullshit) without the aid of a teleprompter or hours of preperation...then you certainly haven\'t been doing your job. If it is something that you don\'t know...it isn\'t a crime to say I don\'t know the answer to that...I will get back to you on it (and then in a novel idea...call the reporter or person who asked you the question after you have researched it and give them the answer...that would endear you to the people as a person that takes time to listen and then think).  Ok...I am tired and it is late...I need some sleep!  PLUR  Oh yeah later today I have to take my box offline to send back my RAM kit that is bad...it will be down I would imagine through to the beggining of next week...I will still update..but I don\'t know how good they will be :-)  '),(985,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok so I had to change my counters around a bit....over there they represent the data a little better and the peace hippie thing is actually a counter that ya\'ll can take a look at...gives a break down of the info...  Enjoy and maybe this will help straighten some of the conflicting numbers I have been getting...jeesh  The first one just counts how many total visits to the site...the second counter actually keeps a little better track of who, where, when...but it doesn\'t display on my main page the number of people that have visited and it won\'t let me start it at the number that I have gotten too...if I can figure that out then I will use it exclusively...but till then...blah!  '),(986,1234212,'07,April,2003','2003-04-07 00:00:00','   I recieved an email from the makers of my ram today and they are going to replace it for me. I am glad they have a lifetime warranty on them. Actually they couldn\'t have been more helpfull about it all. The just gave me an RMA number and said send in the pair and we\'ll get you a new set sent out. So far it looks painless. The only problem is that I will have to not be on the computer for a couple of days...but I can handle that. Also it means that I have to use my laptop for those few days.   As I think more and more about what I want to get next when I get some money for computer stuff. I think I am going to upgrade and get a new laptop...actually a desktop replacement laptop. They are larger and heavier than the smaller and thinner ones...but they come with top of the line components and run right in line with desktop PCs.  I would get something along this line. A laptop from  Powernotebooks .  Something like that would do...I will wait and see what other notbooks come out with the ATi Mobility 9600 because that is the biggie right there. To have that piece of hardware in your laptop means that you have the high end of graphics that will be very good for gaming on the go. The rest of the computer is almost irrelevant...I would love to find a Centrino Based system that has the 9600 in and 1GB of RAM as well as a 5400RPM drive...but if I cant find it I would head to Powernotebooks to get going...as they have had great reviews and are the cutting edge of technology without being way out expensive.   This is the same company that supplies notebooks for  Alienware . With the one exception...Alienware goes with the Geforce FX 5600GO video chipset in it which is far inferiour to the ATi chipset.  It will all depend on getting some money russled together here soon. I would like to try and get my act together and have something by Christmas...that would be soooo K-RAD!  Well I am off to get a bit of sleep...I get the needle taken out of my chest tomorrow which makes me a happy boy...and I also completed the final work on the computer case...now all I have to do is add the few BLUE LEDs to the inside of the case to light it up a bit...overall it looks ok...I can see that while I am waiting for the RAM to be returned I am going to dissasemble it again and sand it down a bit and paint it with a flat black paint solution instead of the somewhat shiny gloss black I have on it...the sanding will be a bitch...but then again I don\'t have much to do...so I might as well make it look really good...  Well off to bed!   Peace....and keep an eye on the cost of war to the right over there...the American Taxpayers dollars at work and not getting us anything for it all...and OPEC just announced that it was going to cut back by 900 thousand (I think that is right) barrels a day to bring the prices up slightly more...so as you can see...we have invaded an oil rich nation...we are paying american corporations to go in and fic it up (on the taxpayers tab) they then sell the oil to themselves...and pass along higher gas prices to us in the end...so we are paying for Iraqii oil in taxes to rebuild the industry (which mind you the industry has plenty of money to build it up without having to ask for a dollar as they will make money on it which ever way they go)...then the consumer pays again at the pump because OPEC has decided to create another shortage as they did in the late 70\'s.  Oh yeah one more thing...since we legally became the owners/operators/occupiers of Iraq we have under international law assumed all of the debt that the Iraqi government accrued over its years of operations...which is in the 200 Billion dollar range...that debt is now on the American taxpayer...also any and all contracts that were signed under the old regime still have to be honored because the contracts were for the Iraqi people...so really Bechtell and Halliburton shouldn\'t even be involved in the oil fields because Russian and French contractors had already signed contracts to do the work on the systems...  Part of the gripe about all of this was that when America was going to go in they would attempt to void all debt and contracts...but under current international law the American leadership can\'t do that...so it will be interesting to see how this plays out over the next few months...because a lot of these lendors and contractors haven\'t been paid yet...and they have legal claim to it all...  So in the end 87 Billion (which in reality only about 20 Billion goes towards rebuilding the rest is for military operations and upgrades) and the current totals that have been spent (around 76 Billion with about 10 billion going towards rebuilding if even that much and the rest to the military to wage the war) we could be looking at having to pay the 200 billion dollars worth of debt that is owed to the Russians (they sold them weapons and heavy equipment), the French (they sold them oild field technology among other things), and have to live up to contracts that are worth billions more with the likes of BP/Shell oil and many others...so it could cost us over 500 billion when all is said and done...and when we do leave it will only take one election and the entire state will be what we certainly wouldn\'t want...a fundamentalist Muslim state.  Wow...just think folks...while the threat of terror weighs heavily on our minds...and we have an urational fear of it really (considered to what people think around the world)...we have set ourselves up as the main target for some truely bad things that maybe we should be scared of...  Leave it to some idiots to fuck up the world for us...and this time the idiots are from Texas...and they run under the newly NYSE company name of Bushco (not really...but when they get there 250 million for a campaign they will be a rather large company).  BLAH!!!!  '),(987,1234212,'07,April,2003','2003-04-07 00:00:00','   Check out the added feature on the sidebar...it is a cost of war tally...and if you go to the website it shows you the cost in perspective to things here at home...like housing, or teachers, or health care...and they tell you how they get their numbers. Sure looks to me that this thing is going to cost way more than 80 billion plus over the next year...  That is really insane considering that here at home people are starving and homeless, children at all levels of primary education are deficiant in many of the basic skills, that 25 million or so children have no health care, and numerous other things are in the shitter...  Vote no on anymore money to Iraq...tell the administration to speed up the building of a government...tell them to make the money that will be spent into a loan...tell them to bring the troops home and spend the money here.  '),(988,1234212,'07,April,2003','2003-04-07 00:00:00','   I am convinced now that someone who is in power hates me...hates my guts. I just had a stick of RAM go bad! I wrote the company that made the RAM because it has a lifetime warranty on it and the place where I got it reffered me to them so hopefully they will make good on the warranty.  Jeesh just as things might have started to look up. This has to come and smack me in the nose. Not to mention all of the other shit that we are all having to deal with in the family.  Oh yeah and an update on my grandpa...it seems he has a bad staph infection...yup good ole staph...the real problem that I see is that he has massive kidney failure and his liver isn\'t in the best shape and all of the good drugs that could be used to fight it off would be very damaging to both those organs...so what are they gonna do? I wonder.  I guess it really doesn\'t matter in the end because he is pretty dieing anyway...and the sooner the better because there is nothing worse than having a sick family member linger on for a long time...it is easier to deal with when the person just dies because then you don\'t have to deal with the illness day in and day out...because for anyone to deal with that it eventually fosters some feelings of disdain. I say that especially in this case because he was never a good man to my grandmother or mother for that matter and since my mom moved up here to be there just in case he has really taken advantage of her.  She moved up under the impression that things financially were ok and that he had all of it arranged so that if he did get sick or go and die suddenly it would be relatively painless to figure out what he owes and pay it off and then distribute any assets as he wanted...well that wasn\'t the case. Instead she came up to a mess and when she got things in order they also took out extra money against the house to have a cushion for a bit if things got messy...well instead of not touching that money he has basically spent a good chunk of it and we don\'t know were he spent it.  We really aren\'t talking about that much money in the grand scheme of things...but it still is a burden because if he does live for another year or two my mom will end up having to pay all the bills because he won\'t have the money or manage what money he does get from Social Security in any sensible way.  He gets about 700 a month from the government. That would be enough under the way she got everything consolidated to pay the majority of the bills. However he cashes that check for running round money and has taken out of the account additional money to do god knows what with it. The only thing that she is asking him to pay is the morgage on the house...she is taking care of the utilities and such. She then finds out just in the last week that he had some bills outstanding from his business that are basically leveraged against his property...the amount is only about 500 dollars...but still it hasn\'t been paid and instead of paying that bill when he got extra money after refinancing he just ignored it. So she has to straighten that out...and god only knows what else will crop up.  In addition to all that he is just a piss ant and really is one of those people that you would not want to know. He has alienated all of his family (it is a miracle that he has two people in his life that don\'t really care as my mom and I don\'t really let bad behaviour get in the way of family) and he has no real friends per se.  It baffles me really. I have been irresponsible with money for a good part of my life. I have never really been good at budgeting but after having seen this and other cases in which people live way beyond their means I am glad that I have started to live within mine and have seriously curbed my ways (which I have my grandmother and mother to thank for that at this point)...  Right now I get 530 bucks from the government every month to live on. I have only four bills per month at this point. I have a car note that is 230 bucks...I wish that I didn\'t have that but I needed a car and in reality it isn\'t that bad considering the car I got is a luxury car (a 1998 Oldsmobile Aurora...I got it for a song and a dance and it is helping rebuild my shitty credit)...I have a cell phone which eats about 50 bucks a month if I stay off of it during the day (which is why I rarely use it during the week before 9PM because I only have 300 minutes on peak and unlimited off peak and weekend minutes...plus I have unlimited PCS Vision internet on it). I have to give back to the government 53 bucks a month for fines and then I give my grandmother money each month to help offset some of the bills that I encure as well as pay on the credit card I have that she gave me...total I owe about 300 on it...and I try not to buy anything on it except gas...I do however owe for the camera I got on it which is 300 bucks but that is a business expense that I will be able to make back hopefully...I also owe her more than that but she is good to me and you won\'t ever hear me complain that I don\'t have nice things because she has done so much for me...so in a given month I have maybe 100 bucks of spending money which right now is limited because I am on house arrest...so in the end I might have some money saved up.  I just got a job that will bring in I hope about 300 a month...which hey that would be great because I can pay my grandmother off and not have to worry about the credit card getting out of hand. Oh I forgot I spent 80 bucks the other day on an oil change and a couple of bits of stuff for my case mod...but that I will give her this next month...usually I giver her about 125 to 150 a month depending.  Here is my point. I am going to try my damndest to not get into a situation anytime soon that makes it hard for me to pay my bills...I know that when my grandmother passes away (if she goes before I do) that I will have to make big changes in my already low budget lifestyle and will have to reduce costs as well as find something that will bring more money in.  I had wanted to move away after I got done with this house arrest but at this point I don\'t think that I can afford it. I don\'t have the income to go moving off anywhere yet. With my illness it is also hard to get a regular job seeing that at anytime I could end up in the hospital and have no income whatsoever...because of the rules if I get a good paying job I loose even my little check from the goverment...and to qualify to get it again I would have to go a year without income..and if I have bills I can\'t do that and live somewhere else...you see the math doesn\'t add up. I would have to have thousands in the bank and if I have that I might not qualify to get aid as well.  The idea that is currently floating around the family is this. My grandmother is taking a hit in the stock market right now. Her portfolio is shrinking because of the stock markets spotty performance. So she is thinking of taking some of that money out now and putting it into property. She wants to help finance a house for my mom and I.  Now I have really mixed feelings about this all. It comes down to two things. First and foremost I wouldn\'t want to burden my grandmother with that even though I would qualify for section 8 housing and they would basically pay the mortgage payment for me...well rent to my grandmother which in turn would pay the mortgage...or if we got the loan with me as a co-owner it would be paid for me...  Second problem is that I don\'t really want to live with my mom. I love my mother greatly but I don\'t know if I could live with her again. I am no clean freak, and I am not one to cast stones...but we are really two different people. I am messy and unorganized...she is dirty and unorganized. Two very different things. I try and keep things clean where as she hasn\'t ever been into keeping things clean...and minor things like leaving butter out. Small things yes...but also I am 26 years old and I live with my grandmother and then would be living with my mother. I am already a pathetic person but I don\'t want to move in with her and jeapordize the good relationship we have. In addition I would also feel even worse than I already do because it would be her supporting much to much.   I already am torn in the situation I am in now...I don\'t really contribute a substantial amount to my own upkeep...I don\'t know folks...I just wish I had a bunch of money because I could then just say this is how I want to do it.   I would buy a little loft on the outskirts of Pittsburgh and have spartan but nice furnishings from Ikea and one of their kick ass beds, a simple sofa set, a nice entertainment area with a few nice toys to say...a nicely equiped kitchen that is spotless and has nice cookware (because I do love to cook if I have all the tools and nice cookware to work with)...and that would be all I need...all of it would be open space because I like that concept of a one room living space that is open. One side would be the bedroom area and the other would be the living area, and another area the kitchen dining area...with minimal walls sectioning it all off...more like fake walls to give some privacy but not completely closed off. Basically you would be able to see the entire space from any point in the place...with the exception of the bathroom...that has to be closed up somewhat...it would be an ideal living working space for me. Partly because I am a nerd and need to have all the toys together to get them all working in concert with each other. The entertainment setup would consist of a Plasma TV from Gateway and a computer hooked up to it with big flat panel connected to the computer for work and the plasma screen for playing video and TV and games on...all run through a speaker system that puts out a lot of sound run out of the computer...it would be a convergance dream in which I could make it do things like display art on the Plasma screen for kicks and get togethers...and play visualizations for winamp when playing music...and double as the gaming screen...but the main flat panel would be big and work for gaming on the computer side of things...nice appliances but nothing special...just a nice trendy but spartan environment designed for modern efficient living.  That is my dream place to habitate in.  It really isn\'t much to ask for...and I would love to be able to have a job that would allow me to do this...which I am working on. I have stuff to sell on Ebay that would get my mom money but also get me some seed money to get the things to put together this dream...also I have seen a few jobs that I can actually qualify for without a degree but would probably pay well enough to live like that...but get to that level over time.  Hey maybe in that environment I could even afford to have someone in my life that would want to be around me.   And that my friend is what I want the most. I want someone who is of the opposite sex who would want to be loved and would love me. Someone who will understand that I am really a very complicated yet shallow emotionally person (shallow in emotion because I feel that emotions complicate things in an irrational way...as I don\'t respond emotionally in a normal way to things)...someone who will feed my intellectual side yet be able to be dumb at the same time...so that when I don\'t have earth shatering things to say will be alright if I make farting noises with my mouth. Whether or not that person exists is highly debatable.  Most of all that person has to be willing to force change on me in the right way...and someone who will be willing to loose me and at the same time assure me that loosing me won\'t kill their soul. That has been my biggest stumbling block to date for me to be close to anyone...becuase of emotion. In my knowledge and critical thought I have come to the conclusion that I don\'t want to inflict pain on anyone when I go...I want it to be celebrated as I would celebrate their life if they were to pass. I would mourn my passing but I would also know of be assured that in my passing that I as a person had something that I enjoyed and loved and would want to go on...  Such deep thought is always hard...and yet I am always compelled to think this way. It is a blessing and curse. It is the problem with having a brain that is always going a million miles an hour thinking in so many branches and working every angle...trying to find the best solution to a problem that has no solution. It is a curse because I always come to the conclusion that I can\'t have those things based on past experience.  What I fear most is that no one will come along to proove me wrong...and that is what I desperately want. Someone to proove me wrong.  Yet after having written this I could ramble on for another twenty hours because my brain has kicked in again and is analysing every word as I write it...  Please someone respond to this...someone proove me wrong.  '),(989,1234212,'07,April,2003','2003-04-07 00:00:00','   So finaly this week of sheer hell is over for me now.  Lets start by saying that I really don\'t like the job I am doing because it is just such BS. This Kiosk that I work at sells ritzy pet things...all of it way overpriced and nothing that I personally would ever buy...but people do buy it. Yesterday I sold 33 dollars worth of stuff from 11AM to 6PM...today was even better...I sold 0.00 from 10AM to 5PM. I can tell that the woman who is going in on this is dissapointed that I am not selling more...or rather that I am not pushing people into sales. My simple philosophy is this...you either want it or you don\'t. I really can\'t make you mind up for you in the end. So my schtick is this. I walk up to the customer...but I don\'t interfere with their personal bubble (because I hate when people do that to me when I am in stores) I say \"Hello...how are you today\"...they say a few words back and I usually reply with \"Oh things are good...if you need anything I will be here to answer questions for you.\" Then I leave them pretty much alone unless they engage me in a conversation which has happened a few times. I have a no pressure approach because of my philosphy...you\'ll either want it or not...I can\'t take your wallet out and make the purchase for you.  Couple that with prices that are really to high (and she went back and raised a few prices during the week) and you have for slow days. It really isn\'t the holidays yet and that is when we hope it will get busy.  In the end she really can\'t afford me to just stand there with a goofy look on my face...and frankly I feel bad because this is supposed to be a vehicle to help get some of her debt paid down.   Here are the numbers. 5500 for the Kiosk through Christmas (I don\'t know if it is per month or a lump sum). Her inventory which she paid wholesale through a middle man when she could have gotten it much cheaper from the factories...and that is going to kill her there.  Plus she has me who comes in for two days a week...a total of 16 hours at what she wants to pay me (8 bucks an hour) and she also has another person who is working for her and I would imagine she has a similar deal too. Well all that adds up and she may not make it back...hell she will be lucky to break even in the end.  She has left it \"up to GOD\"...and we all know that he hasn\'t been a real good manager.  Hell his churches blow through money left and right and are constantly begging for more.  So that is that story.  Also in this hell week my grandfather has gotten really ill. Now I am of the impression he will die real soon because he is pretty bad from what I here. I haven\'t been to visit him yet in the hospital. My mom came home one evening and found him collapsed on the kitchen floor. Not but a few days earlier he was backing out of his driveway and backed into a parked truck. He has been on a rapid decline over the past few weeks. He is on Dialisys and from everything I have heard you get a max of five years on that and then people die off like flies after that...and you only get to that point if you take excellent care of yourself.   Well back to the colapse. First thing you do is call the ambulance right? Well my mom has never once in her life made a sound health choice regardless of who she was having to make a decision for. I went for over a week one time with a bowel blockage when I was younger even though I knew something was wrong...but she wouldn\'t take me until it got to the point where I couldn\'t stand...even then she resented the fact that she had to take me to a doctor. Also when I was younger I never got the proper drug treatments ever...and that could have helped me now...partly because of her...partly because my dad wouldn\'t spend the money on the drugs (and this was before they had the really expensive stuff)...she never contacted state agencies to see if I qualified for some assistance or anything. She is just horrible when it comes to medical stuff. When my sister was in a severe wreck my mom talked her out of going to the hospital in the ambulance...turned out my sister did need to go. When it came round to getting money for the accident (it was someone else\'s fault) because she hadn\'t taken the ride she lost out on a substantial amount of money (like 50 grand or more).  So when she found him collapsed she first tried to administer aid herself. Got him up to his room because he wanted to pee. Then she finally got around to calling his PCP. Now what were they gonna say \"give him an aspirin\"...no they said to call the ambulance immediately. So she finally did more than an hour after she found him on the floor. Compounding all of this was a fever that was rising and eventually is topping out (he still has it) at 104F.   She said he was very jaundice (yellow skin tone) and that is a sign that things aren\'t working right. With his bad kidneys it is probably cause by that because once they shut down other things shut down as well like the liver. He is essentially having organ failures...and on top of it he has a heart condition...smokes, drinks, doesn\'t eat right the whole nine yards...so I told her that she has always made bad decisions when it comes to health and just listen to me on this one...I ended up being right.  She got upset when I told her that...and it may have hurt her feelings...but I said it because it was the truth and in this situation she could face charges if she didn\'t act properly...and she didn\'t. She flat out didn\'t. Fortunately her ass is covered now...he excuse was that she needed to get the PCP to get things arranged so he could go strait in...well ER have doctors too and they would be just as well versed with medical conditions...and they could have called his PCP to get full details.  On top of it all she has had Jesse\'s box to send back to him...in all the rush and hub-ub she forgot about it which I can understand that...she is able to get a big ass discount on sending it through Fed-Ex so she said she would get it in to them in the morning and have it delivered overnight service...she was going to try today but got detained at the hospital with all this shit.  Heres the kick in the balls...my lung infection is not over...and I know as soon as I stop the meds on Wednesday within two weeks tops I will be back in the hospital...because these bugs I have are just not responding to the meds they have me on...  Sometimes I just want to say fuck it all and quit.  What I have felt for a long time and my friend Danielle (who is also a CF patient) elloquently said...\"We are so old yet so young. It is a crime.\"  That is what it is alright. A crime. I have friends who are out there and in relative health yet they complain about meaningless bullshit...oh my back hurts, I have this ailment, or that one...shut the fuck up people.  You aren\'t sick. You haven\'t even begun to experience what some unfortunate people have had to deal with. For the older folks out there...I understand that it is a bummer to start to have your health fail knowing that sooner or later you are going to die...but you have lived a full life. Those of us who are young and have to deal with all this shit seem to have a much better grasp on it and accept what will come of us.   I would never wish this on anyone...but if some people could walk a day in my shoes to see they shit I have to do...and when that shit stops working is when you truely become fucked...and for me things are no longer working.  Before I exit this planet I hope to make a difference...I want to write a couple of papers...and maybe a book. I have material that I have scribbled all over the place and in that material I am sure is some great stories and possibly some interesting ideas.  I just can hope that I can realize some of the things in my head before they die with me.  Oh and I figured out what I want done with my body when I die...first I want it to be a research corpse...something so students can disect me and play with my guts...but then I want to be cremated and then have the ashes taken to one of the new labs that can take a persons ashes and turn them into diamonds. So that is what I want done with me. I want to become eternal...and what better way than through a diamond. In addition I want to have some genetic samples drawn and preserved (preferrably encased in the diamond or the ring that will hold the diamond) so that maybe in a future time they will be able to clone me...at least it will be a historical record that will say to someone in the distant future...he lies Michael Honse in stone, DNA attached for reference.  PLUR  '),(990,1234212,'07,April,2003','2003-04-07 00:00:00','   Well today I had my first full day of work...I can tell you now with all honesty that the business this woman is running will fail. The market just isn\'t there and she pays way to much for everything causing her to jack her prices up to a point where things are just unreasonable.  I hope for her sake that thing will pick up and that we can sell the hell out of the items she has...its not that the items in and of themselves are crappy...they are just to expensive in this economy. You see this lady is in deep debt and she thought it might be good to start her own business. Now for me I would have chosen to go into something that lands contracts so you have some sense of security as to what you are working for and selling. You can then build in profit from the beginning and you only need to get what you need and not carry any type of stock. This thing is costing way to much money and frankly it probably won\'t even return her investment...all she could really hope is to break even...and that might be a stretch. She has left it in \"GODs\" hands...and well you know that I don\'t particularly put much faith in an entity that has been corrupted by man.  One thing about this job and I really noticed it today is that the mall is just a place of absolute absurdity. You have maybe four or five major types of things sold in a mall...the main thing being clothes. Well all the stores sell pretty much the same thing at huge markups considering all their clothes are made in third world countries for pennies and people here plop down 50 bucks for a pair of jeans...it is just mind boggling to me.  I watched people aimlessly amble by with these blank looks on their faces like they had been put into a trance and were consuming for the sake of consumption. I saw teens walking around with bags and bags of stuff with a cell phone attached to their head talking way to loudly and about such meaningless things.   When I get done at the end of the day I feel dirty...I literaly walk out of that mall and feel like I have just been through some sort of evil ritual that Americans call commerce.   My philosophy is simple...I need maybe 7-10 pairs of pants and a couple pair of \"dress up\" pants, about a dozen shirts or so, a few hoodies or sweatshirts, and maybe a few button down shirts...all of that would add up to maybe 300 dollars worth of goods. That is all I need and I can make those clothes last over a year (seeing as most of my clothes now are several years old). The only thing that I would even consider buying on a regular basis is shoes...and that is because I like funky shoes...but because of my current income I don\'t even bother to do that because I don\'t have the money.  Of coarse there are things that I would like to have...but there is nothing out there that would make me want to go out and blow thousands of dollars on...a three hundred dollar watch...gimme a break...I would loose it. 200 bucks for sunglasses...fuck that...I can get a decent pair from walmart for 10 bucks...sure it doesn\'t have a \"label\" on the side of them...but it blocks UV rays and such.  Which does remind me I do have to go get some clothes because I just don\'t have much in the way of normal clothing...most of my clothes consist of RAVE clothes that I don\'t really wear all that much anymore.  I sit across from Abercrombie and Bitch...and I see little people going in and out all day and wonder...why would you spend all that money on something that has a real value of maybe twenty bucks for several things.   The only thing that I really have a vice with is computer equipment...and I get that because it serves a function and makes my life easier...like I want a new keyboard and mouse combo because I go through them fairly quickly (I wear them out). The set I have my eye on is really to much money...but it would be something that I can use on a daily basis. I want gadgets...cause I love them. However I don\'t indulge until I have the money to get them because I don\'t want to go into massive debt to get something....the only thing that may change that soon is to get a new cell phone...I have my eye on several but the dream one is pretty damned expensive...but it would be almost like a laptop setup. What I may do instead of that is buy a cheaper model, buy a cell card for my laptop, and in time get a new laptop that will last me a long long time...and really I would love to get a Tablet PC instead of a laptop truth be told...but a laptop would be better suited to what I want to do with it...such as coding on the road, building sites on the road...and of course game on the road.  Again these are things that I want...but at this point I am not going to go into massive debt to get one...at least not yet.  Well I am going to get some shuteye because tomorrow I have to face the throngs of people yet again...jeesh!  Peace out yo!  '),(991,1234212,'07,April,2003','2003-04-07 00:00:00','   This is new to me...but it shows Dean\'s commitment to using new technology to get the message out in a big way....and it probably saves money allowing him to reach more people for less cash.  I present to you  Dean TV . You will need broadband to see it...but it is a great way to get the message out. You have to install some simple software and from what I have seen of it, it does nothing as far as spying on you. A worthy download in my opinion.  On to other things...not much got done today. I have to work tomorrow...woohooo!  I may update a little later.  PLUR  '),(992,1234212,'07,April,2003','2003-04-07 00:00:00','   Oh yeah and Rush Limbaugh...I hope if the allegations are false that he is vindicated...but if they are true...by God I want to see that fat, racist, bigot nailed to a cross...  It would be irony at its sweetest...  '),(993,1234212,'07,April,2003','2003-04-07 00:00:00','   I just took a blogger poll...thing is the polls that they never allow you to write in things or give you enough choices to really get a poll sample down correct...  For instance one of the questions is how long have you blogged with blogger...without asking how much you have blogged in the past...and then one of the questions is were you one of the first in your group and did you introduce it to other or did others introduce you to it..  Well a fair question would have been to find out how long people have been blogging...for me it has basically been on and off for around 4 years now...with my original blog type forum on a site called Camarades...when it was a free web cam site...I had typically 20 users on my webcam all the time...and I updated the page almost daily with what was going on...so that could certainly be considered blogging...then I laid off of it for a while until I moved too my apartment in MS where I set up my own web server and hosted my own blog...then when I moved back up here I started blogging on /. for almost six months before moving over to blogger...and I have been with them ever since because they have been very reliable and easy to use...  Over time my blog has evolved somewhat and I would love to continue to use this service but in the future I will probably move to a movable type based blog...I would love to get the pay version of blogger as well...and I may do that just to keep it simple...with that package you get photo hosting and expanded ability to move things and customize...and that is what I need...so maybe that is what I will do in the end...  Well enough for tonight...I have a nasty sinus infection that started yesterday and has been kicking my ass all day today and is worse tonight...so if it isn\'t any better tomorrow I will have to call the doctor and figure out what we want to do...because the damn thing is migrating south for the winter...and south is the lungs and I don\'t need that again.  Well I am off to bed...night all!  '),(994,1234212,'07,April,2003','2003-04-07 00:00:00','   Not much to post about today seeing that it has just been another day of the same old same old...waiting to see if anything will be made of the election problems looming in the national media...  Once I have my main box up I plan to put together a presentation as well as search out the source code for the voting machines so that I can put together something that the media could take off from.   Then I will pimp it around to the media...of coarse I will post it hear first or link to where the presentation will be hosted (probably on my webspace with Verizon).  I think that I can be annoying enough to get it somewhere with the mainstream media. As it is really important for them to investigate the claims that there may be problems...I am thinking Dateline mainly and then CNN...or maybe 60 Minutes as I know that both Dateline and 60 Minutes take on hard things...with 60 Minutes being the premier in journalistic integrity for many years...hopefully they would pick up on the story.  I plan to contact people who have been involved in the process up to this point...and I plan to offer my own information once I have obtained results when I get the software...I also plan to have programmers look at the code and bring those programmers in who have looked at it thusfar so that we collectively could present a complete story with much of the research already done so that networks would only have to validate and air the story...it is about time we get together...  This all depends on when I get RAM back for my box...I plan to spend around 6 hours a day on the story until I figure it out for myself and to hound everyone involved so that maybe something constructive can be done.  I urge you to all help if possible in some of the tracking of company leaders...who they are tied too and who they have dealt with. That would be of great help...and of coarse those who come to me with information will be credited if they so desire...or remain anonomous if they so choose (as this may cause quite a stir and get some high ups pissed off).  Also I plan to actually crack the voting machines if they are deployed here in my precinct during the next vote and then publisize that I did...in the hopes that they would bring prosecution on me so that we could push it out into the mainstream if that is what becomes nesecary...I don\'t want to do it...but if I have to then I will...and I will make the information on how to crack it once I have it available to all so that others can engage in civil disobediance through what is essentially voter fraud...but for a good cause...to remove the idea of Blackbox Voting.  Send me your thoughts...I need to hear them as I am embarking on what will prove to be a hard and long journey that I am compelled to do. It is time people take action and I have said that many times without having done much myself...so I will take my own words to heart and bring my experience and knowledge (albeit limited) to the table and offer it up to the idea of Democracy.  '),(995,1234212,'07,April,2003','2003-04-07 00:00:00','   In the news as of late there has been a lot of hoopla about the leaking of Amb. Wilson\'s wife as a CIA operative from the Administration.  I watched a bit of the debate back and forth between the various parties and watched the press secretary come out and give a statement.  All of this back and forth leads me to believe that first and foremost it is most definitely a partisan move on the Democrats behalf simply because of what is at stake here. If it can be proven that the Administration did leak the name and Robert Novak (a CNN analyst amongst many other things) reported it in a piece he wrote for the newspaper then indeed their would be indictments of the people involved in the leak.  The leak itself was intended to send a warning out to anyone who might speak out against the Administration and its policies so that people may think twice and not speak out fearing that the Administration would use its considerable pull within the government to find anything to attack the reporter or person who might speak out. This has been a tactic that has been used repeatedly by all administrations and it must be pointed out that both parties are guilty of using such tactics whenever the opposing party is in power. It is nothing new to the overall scheme of things in Washington.  The bigger question here is did it happen, if so what will be done, also who knew inside the administration, and who authorized such a leak of sensitive material.   Also this brings into question the credibility of our secret dealings around the globe. If informants and operatives aren\'t protected and are to be used as pawns in political games then we may loose our best assests around the world in fear of them being found out then prosecuted and quite possibly put to death by the people they are informing on.  Another point that needs to be made is when something of this nature comes about is it the duty of the reporter to reveal his/her source if indeed the rumors are true. The answer is an unqualified yes to that question. Not because we want to discourage people from being informants, but more simply because we want to discourage blatent abuse of the reporters and the weight that their voices carry in modern media.   Leaks happen on a daily basis, but no leak like this has ever come up. It truely is of national security and pride in our foriegn intelligence services that comes to question. If we can\'t bring in the rogues that try and bring others down because they don\'t like political views or because someone doesn\'t like what is being said why not just go ahead and kill people off instead of tarnishing reputations...it is in the end much more human and a more forward and to the point action. Subtlety has never worked in the past, but actions do speak louder than words.  As for the administration it is claimed that the leak didn\'t come from the white house, but they can\'t confirm if it came from the administration. Well come on folks, the administration is the white house...all edicts come down from there to the various foot soldiers to be carried out. So the fact of the matter is that regardless of whether it was someone directly in the white house or someone in the broad administration is of no relevance...because they are inextricably tied to each other.  Some say that it was Karl Rove who leaked the information. If it was shame on him and he should be prosecuted to the fullest extent of the law. Besides what value is it that he would hold such sensitive documents as he is really just a political advisor and not much of anything else. To think that he would have access to such sensitive information is rediculous...and there is no doubt that he does have access to documents that really he shouldn\'t.  The President share cuplability in this whole mess because of what he has said in the past. He stated that he would bring back integrity to the office and that the buck would stop with him. If that is so then regardless of where the leak came from and whether or not he knew explicitely shouldn\'t matter because he pledged to be responsible for all actions taken in the name of his Administration.  In this search for answers I have gone looking for ways to deal with the problem seeing that it really is a problem that extends up to the President regardless of whether he tacitly knew of the events...and it brought me to the impeachment process.   In looking up the impeachment process it revealed that impeachment is really a difficult thing to do and rightfully so...but would be impossible to actually implement in this case because of the makeup of the government and how it is controlled at this point in time.  Here is the actual law that calls for and outlines what is impeachable activities.  \"The President, Vice President and all civil officers of the United States, shall be removed from office on impeachment for, and conviction of, treason, bribery, or other high crimes and misdemeanors.  --U.S. Constitution, Article 2, Section 4\"  The process to do so is quite convoluted and requires something that can\'t be achieved even if all the allegations prove true because of the makup of the congress in general.  Taken from a law website which I will give the link to at the end of this post.  \"The Impeachment Process in a Nutshell  The House Judiciary Committee deliberates over whether to initiate an impeachment inquiry.   The Judiciary Committee adopts a resolution seeking authority from the entire House of Representatives to conduct an inquiry. Before voting, the House debates and considers the resolution. Approval requires a majority vote.   The Judiciary Committee conducts an impeachment inquiry, possibly through public hearings. At the conclusion of the inquiry, articles of impeachment are prepared. They must be approved by a majority of the Committee.   The House of Representatives considers and debates the articles of impeachment. A majority vote of the entire House is required to pass each article. Once an article is approved, the President is, technically speaking, \"impeached\" -- that is subject to trial in the Senate.   The Senate holds trial on the articles of impeachment approved by the House. The Senate sits as a jury while the Chief Justice of the Supreme Court presides over the trial.   At the conclusion of the trial, the Senate votes on whether to remove the President from office. A two-thirds vote by the Members present in the Senate is required for removal.   If the President is removed, the Vice-President assumes the Presidency under the chain of succession established by Amendment XXV.\"   As you can see it has to go through many stages and all of those stages at this point are controlled by Republicans...so even if there were clear and obvious wrong doings it would be impossible to actually proceed with an impeachment no matter what the public outcry calls for.  F. James Sensenbrenner, Jr., (Jim), is a Republican Member of Congress is the head of the Judiciary Committee so you can see that even if it got to this stage it would be shot down. The committee is stocked with more Republicans than Democrats. Here is the list.  Republican Members                Mr. Hyde (Illinois)                     Mr. Coble (North Carolina)         Mr. Smith (Texas)                       Mr. Gallegly (California)             Mr. Goodlatte (Virginia)              Mr. Chabot (Ohio)                      Mr. Jenkins (Tennessee)            Mr. Cannon (Utah)                     Mr. Bachus (Alabama)                Mr. Hostettler (Indiana)           Mr. Green (Wisconsin)               Mr. Keller (Florida)                    Ms. Hart  (Pennsylvania)           Mr. Flake (Arizona)                  Mr. Pence (Indiana)                   Mr. Forbes (Virginia)                Mr. King (Iowa)     Mr. Carter (Texas)     Mr. Feeney (Florida)   Mrs. Blackburn (Tennessee)    Democratic Members    Mr. Conyers (Michigan) Ranking Mr. Berman (California) Mr. Boucher (Virginia) Mr. Nadler (New York)  Mr. Scott (Virginia) Mr. Watt (North Carolina) Ms. Lofgren (California) Ms. Jackson Lee (Texas) Ms. Waters (California) Mr. Meehan (Massachusetts)   Mr. Delahunt (Massachusetts) Mr. Wexler (Florida)  Ms. Baldwin (Wisconsin) Mr. Weiner (New York)   Mr. Schiff (California) Ms. Sanchez (California)     So as you can see the votes would go pretty much along party lines (20 Republicans and 16 Democrats) and the vote to get it out of committee would be shot down before it would even be considered for consideration that an impeachment might be nesecary.  If the allegations prove true than he should definitely have the impeachment process started as he would have broken the law...in addition it would include VP Cheney as well as others in the Administration.  The plain fact is that Mr. Robert Novak should indeed reveal his sources if nothing else in private to the appropriate authorities so an investigation could proceed. Without that key bit of evidence the investigation will die on the vine so to speak.  Remember that we actually impeached President Clinton and the vote was against removing him from office. The vote wasn\'t agaisnt what was at the time the Whitewater scandal as they could not find the documentation to actually use in the process but instead the moved ahead anyway with the fact that he lied about getting oral sex from an intern. Basically they impeached on the fact that he lied.In the end it wasn\'t enough to impeach him.  That brings up special prosecutors. This strange phenominom was implemented in the seventies as a way to get independant investigators into the mix as it was seen that the Justice department often comes into conflict on matters of this importance seeing that the head of the Justice Department is appointed by the sitting President...and often the head of the Justice Department is reluctant to do anything against the very person that gave him/her the job. This bit of legislation had a sunset clause put on it and it lapsed in the late nineties (or thereabouts) and was not reinstated because it was seen that independant investigations often proved fruitless in the matters they persued partly in responce to the fact that they were often denied the very information they needed to proceed.  So getting an independant investigation into this matter just simply won\'t happen because there is no law requiring it.  On another note it has been well documented and seen in Washington over the past few years that outside investigations by any body have also wained and in most cases simply not happened because of the often used \"if you question the policies and intentions of the Administration then you are a traitor\" view. So many things have happened that warrent investigations regardless of who is in power that just simply haven\'t happened because of the stigma of being labled an unpatriotic american is simply to much to fight against seeing that the White House has on several occassions used its executive powers to launch probes into investigators and them leaked the information to the media so as to discredit the investigators.  While at this point we do not know who leaked the information, my money is on a high admistration official that wanted to attempt to intimidate and discredit Amb. Wilson...a purely political move that if his investigation into the Niger uranium deals had become common public knowledge that discredited the claims of the Bush Administration that Hussein was trying to get Uranium from them would most certainly lead to the questioning of the policy of war in Iraq.  Also I read my new issue of Time today and they brought out some pretty interesting information to the effect of because of the widespread corruption within the Iraqi government Hussein may well of thought he had weapons when indeed he didn\'t because of the lies and misreports given to him on those weapons programs. It would seem that the scientists and people in charge of those programs knew better than to screw around with the UN on this...their only mistake (which can be excused) is not having documented the destruction of said weapons. If they had documented the destruction of the weapons surely they would have been put to death if the documents made it to Hussein...as we all know he aspired to have all sorts of weapons. The deception that was perpitrated was to protect the lives of the scientists in charge of the programs from the very dictator that was telling them to \"build them or die\" so they had great reason to lie and not leave a paper trail about the lies.  It comes down to this folks...Hussein had aspirations to achieve nuclear weapons and terrible other weapons...but in reality his own scientists, who were just that and not party fanatics, lied to him to save their own asses while giving him the documentation that they were making progress even though it was all a big hoax. They did admit however that they kept a few weapons back to test the shelf life of the weapons but they were eventually turned over to the UN inspectors who then mislabled them and lost them in the clutter...so even the small amount that we have found (something like maybe a couple dozen warheads of chemical weapons) were in fact already documented and known of by the UN but through snafu\'s in their procedures they we not kept track of and eventually destroyed.  When all of this washes out in the end the Presidents rational excuse for pre-emptive strikes will prove to be a bad policy based on the fact that people can simply lie about the extent of programs...and it goes to show that not even dictators can know everything that is going on in their own country....  I am pretty sure that Bush has no clue of our own chemical and biological research that is thriving even though he touts the official position that we are destroying out stockpiles. That gives him deniability when it comes down to it...he didn\'t know the scientists were developing new weapons...because our policy is officially to dismantle those projects.  We will all have to come to the understanding that our leaders are not omnipotant people and can\'t have detailed knowledge of everything that goes on in government even though they claim to have intimate knowledge of what all is actually going on. Maybe that is a good thing, maybe it is a bad thing...but that is just simply the way that all governments do business.  The link that I promised to the articles of impeachment and the process in which it goes through. I plagarized huge chunks of the writings and am not ashamed as they state very clearly the law and how it actually works.   Impeachment Process in detail.    House Judiciary Committee Website.  '),(996,1234212,'07,April,2003','2003-04-07 00:00:00','   I am taking down the comments part of the blog for the time being because the servers that serve it up have been infected with a nasty virus that can spread to your machine if you choose to post a comment. Hopefully the comments will be back up soon as the provider of this service is working hard to figure the problem out and fix it. I would like to send a word of encouragment out to him...all of us that use the service are greatful and hope that the issue is resolved soon because we all really like having this service made available to us at a cost that no one can complaign about at all! So we hope that things are resolved soon as to let the pressure you must feel up...again thanks for the service :-)  '),(997,1234212,'07,April,2003','2003-04-07 00:00:00','   Having read through the links all of them seem to be of worthy consideration considering the possible bearing it might have on the voting of Americans.  Also of note is that some of the farther left leaning links as well as the links to the source code bear no fruit as the servers are either down, or the sites are removed which would indicate one of two things bearing on the credibility of the debate at hand.   First is that the information and opinions are true and that the sites have been \"removed\" in the light that they damage the quest for power that people in power so love...this would be a blow to us all because they are in essense supressing free speach. That would be a crime. I find it wholey possible that is the case because if people and investigators were to get a hold of the source code for these machines and it was indeed found to be poorly programmed with the ability to change election results at a mear whim then it seems that we have all lost...and in that we should all be screaming in one voice \"release the code, let us judge if it is adequate and protects our rights\".   The other possibility (although I tend to think that the first possibility is a better one seeing the actions of Bushco and the far right republican party) is that the claims hold no water and if the information was let out then their claims would be proved to be wrong. They would simply not hold up to public scrutiny and that would lead people to discredit the authors of said stories...and we all know that discrediting is the way to go when you want to dismiss \"conspiracy theory\" over just denying it.  I think it is the first possibility because it is a matter of law. All Diebold has to do is claim DMCA over its work and that shuts down all research of said code because to actually investigate and critisize said code would leave an independant researcher open to litigation. To simply post code that is a product of a large company is strictly frowned upon by the courts even if said posting of code is for the benefit of the people.  Remember that one of the main arguments for the case against Microsoft in the 90\'s was so they would release their source to operating systems for other software writers to be able to look into making their products integrate more effectively into the OS...well Microsoft was ruled against in that regard with the government finding to not post or release the code to credible sources or companies was a violation of anti-trust (RICO) laws. So they had to release the code in question (which happens to be a discontinued OS...Windows 95. Apparently the lawyers for the government didn\'t specify that newer OS code be released although MS has been a little more willing to let people peek into the guts of newer OS\'s...although they are still fiercly protective, and rightfully so, of the inner workings of Windows XP).  This all smacks of something scandalous...and we should demand an investigation into it all.  One other scandalous thing is the exit polling. It used to be independant of the government election process...now it isn\'t by law. All of the exit pollers basically have to adhere to the same rules now and report findings to the same place which then distributes that information to the news outlets...so in this next national election it can be expected that we will see delays in exit polling numbers so that things could possibly be changed on the fly without making it seem that someone is rigging the election...all very tidy as to keep the public in the dark.  I have been, and am all in favor of Electronic Voting as long as you get a reciept of the vote in paper so if results are questioned then there is still a paper trail somewhere (and it would be better if the machines held a reciept as do cash registers for keeping a paper trail). However reading some of the election law it seems as though using electronic voting machines as they are in current form violates the voting laws specifically designed to keep fraud at a minimum...and the voting laws also cast doubt on even current voting systems because votes are not hand counted (as made so painfully obvious by the Supreme Court blocking a Florida hand recount).  On another note I watched a program on Worldink TV the other evening about the Floorida debacle...and the voting process in general and it would seem after having been two years that the ballots would be counted by now by the government...but they have yet to be counted...they were counted by the media and it was found that Gore actually had more votes so in reality the Electoral Votes of Florida should have indeed gone to Gore...but because a hand recount was blocked by the Supreme Court (which by the way has five conservative judges on it out of the nine...all of which were appointed by Republican Presidents) the true and accurate count was never done.  Also of note was the issue of \"overvoting\"...in which some of the ballots were not clear on how to vote (especially if you are an idiot and can\'t read directions). There were thousands of these overvotes which machines counting the votes just simply didn\'t count, however the vote for who the person was voting for was painfully clear...i.e. People punched or filled in a blank for Gore then wrote his name in where it says \"Write in the name of the candidate you are voting for\"...for some that would mean a double check...but for informed people it would be for the write in ballot (which we still have...and I wonder if it will still be there in the electronic ballot)...  In addition the voting laws also state that the vote must be made in handwriting...so machine votes don\'t count in the first place.  It is all very dis-organized and we need to clarify the law to allow for machine voting in the first place...but at the same time the process in which a person votes needs to be transparent so that fraud will not occur...  It is sad to think that it may not be even worth voting knowing that your vote on one of these electronic ballot machines may not be counted or it might be changed when the votes are phoned into the counting centers...  It is all pretty scary on how we have lost our power over the government.  One other thing...I hate typing on this laptop keyboard so for me to do all this typing is pretty impressive...it goes to show that I have to give my opinion on this because it is sooo important...  Go Vote...and demand a reciept of who you voted for so that in the end if the votes don\'t add up then we can demand the correct victor to be put into office whomever it may be.  '),(998,1234212,'07,April,2003','2003-04-07 00:00:00','   Hey before I head to bed I have a few things to post...from a reputable news source in the Tech world...as a matter of fact the inquirer has been dead on in its reviews and breaking news in the tech world for several years now announcing new products, previously unkown problems in software, and genrerally being dead on with predictions \"from reliable inside sources\"...so the following links I give you with the complete knowledge and full faith that they would not knowingly ever report anything that was questionable or false...when it comes down to it I think they are a hell of a lot more accurate than even CNN when reporting news outside of the Tech world...so here are the links....   Interesting story number 1    Interesting story number 1   Again...I trust them as a news source...I ask you to visit the pages and then follow the links...I haven\'t read all the stories yet but I am working on it...considering I am reading all of this on a shitty laptop screen it is taking me a while...you will notice that one link in one of the stories is already dead...as it has been killed (probably by higher powers because they don\'t want the info to get out...that link holds the source code which has not been independantly verified by computer security experts for the new Diebold voting machines...which Diebold happens to be run by a huge Bush supporter...an independent review would allow for security flaws to be found and for possible ways to cheat also be found...remember that recently the government sued Microsoft and one of the aims was to open the source code for independant review so that companies would be able to better write software...but I guess in a case as important as voter trust it is more important to keep things totally secret!)  So go read...read people...get off your asses and click the links!!! It is very important...is has to do with the fact that your votes may not be counted if the party in control doesn\'t like the way the vote is coming in....it is important to you whether you are a Republican, Democrat, Green, Independant, Nazi, any party voter...it is important to us all...it is important to voters period...  Damn and I thought that machines counting votes in a totally digital way would at least be built with good protection in them...so that votes would be counted and no one would be left out of the vote if they voted...  I am soooo mad about this whole deal...Democracy...ha for the wealthy...I guess the addage is true...\"The Best Democracy Money can buy\".  '),(999,1234212,'07,April,2003','2003-04-07 00:00:00','   Today I sold 48 bucks worth w00t!  Anyways I am spreading the Dean message...even though my employer is a Republican and thinks the President is doing a great job...yeah and pigs are flying out of my ass as I type....  Dean Dean Dean!!!!  I am a Dean monkey...and I am proud of it...I need to get some literature to spread the word...and do the good work.  '),(1000,1234212,'07,April,2003','2003-04-07 00:00:00','   Hello all from my laptop...yay it works!!!!  hehehe  anyways I won\'t be doing huge updates because this keyboard is a little small to type on and I hate typing on it...but at least I will be able to update each day...  Well today I sold 88.10 worth of stuff...that is good...at least I know that it wasn\'t my charming personality keeping people away last weekend when I sold pratically nothing.  I hate my job but I love working...make sense to anyone out there?  I have to get up at the butt crack of dawn so you will have to excuse me if this is short but I am tired and I my hips hurt a lot from standing and walking all day long...so I am going to get some rest.  Hopefully tomorrow there will be a slightly longer update...  PLUR and stuff...  '),(1001,1234212,'07,April,2003','2003-04-07 00:00:00','   Yay! I get to work tomorrow...yay!  I don\'t really like my job oustide of the fact that I sit in front of Abercrombie and Fitch and they have hotties in there all the damn time...I really need to wear loose clothing to hide my \"happiness\" at the fact we are located where we are...hehehe  Outside of that I leave out of the place and I just feel dirty...I really do. I walk out of the mall at the end of the day and I feel filthy...I just want to wash the filth from my body. I guess I just hate the consumerism that is in there...sure I love going into stores and buying things...when I need something...but christ I think that malls are just filthy consumer shrines that we have built to appease the gods of corporate america...where we spend ourselves into such debt that we could never even imagine a life without it...  Hey I guess that is life...but I am glad that I don\'t have outrageous debts...and would most likely not want to get under any big ones here soon.  Till whenever I get back online...cause tomorrow morning I take out my RAM...yeah baby...  '),(1002,1234212,'07,April,2003','2003-04-07 00:00:00','   I was up for a few hours really just to do treatments and put food in my body and now I am off to bed again as tomorrow I have to work. It will be good to get out of the house tomorrow for a bit. It will give me a break.  I got to talk to Jonelle the other day as she was able to catch me while I had my phone on and on me...so we talked for a bit and she is coming out in December for Christmas and we plan to meet up then and have a good time...and my sister may be here as well so she can give me an outside opinion as to what she thinks of Jonelle.  I would seriously date Jonelle if she we here or I were there (in Callifornia where she is)..but that isn\'t happening right now. I am planning to go out that way once I get a lot of my crap out of the way and maybe get a little healthier. I have cousins out there that I have never really seen since they we were all tots so I could see them and then see Jonelle as well. While I would be out there hey I can put in my Resume\' out there and see if I get any nibbles. You never know.  Well I am going to get back to bed...my sleepytime meds are kicking in and the screen is getting blurry...  Peace, Love, Unity, and Respect to all.  '),(1003,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I got back from dropping my mom off at the airport not all that long ago...and I am tired as hell. I am going to get some more shuteye.  '),(1004,1234212,'07,April,2003','2003-04-07 00:00:00','   Its official...I am sick again. I can\'t seem to get rid of this cough and green stuff coming out of my nose, throat, and lungs...at least I\'m not peeing green cause that would suck balls.  I have to run mom to the airport in a few hours so I am going to hit the sack for a bit.  '),(1005,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I looked at my bosses computer today and boy is it fucked up in all sorts of ways. She has spyware, worms, bugs, virii, and just plain crapped up software. However being the guru I am I will be able to sort it all out with some time and patience.  On my way back from her place I stopped and picked up LOTR: The Two Towers and The Italian Job (which I got the wrong version...got the full screen instead of the widescreen version). I am watching LOTR now and enjoying it greatly. I just wish the sound was a little more even as it gets loud and quiet quickly. Its just meant to be watched loud.  Well I watched part of the debate on CNN tonight and Dean did much better and was allowed much more time to voice his views. In this debate he won hands down as being real and telling how he would do things. It was a nice thing to see as the other candidates (especially the ones out of Washington) all pretty much sounded the same. I was waiting for him to say a couple of times \"shut the fuck up fucktards...you haven\'t done anything to stand up to Bushco and voted for everything he pushed through.\" Alas he didn\'t, it would have made for a more interesting debate if he had.   They are still lacking in some ways though. They need to stand up and be different, and out of the pack only two are doing just that. Dean and *gasp* Sharpton. Again I don\'t like Sharpton but he sure is fun to watch because he brings some life to the debate where as the rest (except for Dean) look like stiffs that have been coached on how to be more lifelike.  Well I am going to watch the rest of this and hit the sack. If you can catch the reruns as you will see what I mean.  '),(1006,1234212,'07,April,2003','2003-04-07 00:00:00','   Now I am morally oposed to downloading music because it can deprive an artist of a few cents on a CD sale, but I am more moraly oposed to support a corrupt organization such as the RIAA.  The RIAA has seen sales drop at around 10% over the past several quarters and has blamed the continuing downloading of pirated CDs on the internet, but according to their own numbers downloading of CDs has dropped about 40% in the last quarter so something doesn\'t jive here.  I think what has happened is this and it happened before and they blamed it on another reason similar to the current CD piracy over the internet. It happened when they moved from a primarily Vinyl based system over to Cassette tapes. They blamed copying of cassettes towards the end life of cassettes as the reason for the dropping sales in cassettes. Yet when you look back at the actually numbers you see that people simply updated their collections and those sales were just slipping because many had completed their updating. So in the initial years people bought a lot in the way of cassettes to replace vinyl they already owned and as it got towards the end people had already bought all they were going to buy. Even though the record industry at that time blamed it on cassette copying as the reason they saw sales drop. However they had a replacement ready in the wings...the CD-ROM or CDs as we know them. They offered superior sound to all previous formats (although they lacked the warmth or depth that vinyl often had) but the initial high cost of CDs kept sales slow but steady until the early 90\'s when it really took off.   I remember that I got my first CD player in 91 and I started buying CDs then as I started to work. I bought about 12 a year. Once I started making more money I was up to around 40 cds a year. Now that I am out of college and broke (partly because of shitty economy and higher costs of living) I no longer buy that many cds. Instead I do download most of them now and burn them off.   Now as I have said sharing is bad, but what is worse is the RIAA selling a product for anywhere from 10 to 23 dollars for something that clearly costs less. Even during the era of narrowband and hourly rates you could still download a CD and burn it for about the same cost as that...now we have broadband and cheap cds and a person can do it for a fraction of the cost.  As far as the artists are concerned you will find most of them publicly support the RIAA as they control much of the media content that is seen out there as well as the retail channel that gets their music into stores. In private many aren\'t that worried about it because in reality they see very little from their cd sales in the first place. A band or musician may sell 10 million copies and only recieve a small pittance (maybe 100 thousand dollars if lucky) for those sales while the music companies and RIAA see millions upon millions. Again many of the artists in private don\'t mind it that much from what I have gathered...and in addition I don\'t think those artists that don\'t benefit from the marketing machines in the industry would really mind at all because they may see nothing for their work at all. With the spreading of their music online they may see increased ticket sales at the box office which the artists directly bennefit from instead of the labels.  There are some new venues to sell music and this is one of them  this company may be changing the paradigm.  They give 50% of sales to their artists and allow free downloading of music of their artists for sampling. Now I am sure as more bands move over to this business model we will see ever decreasing costs for music and artists that see more of their work going into their pocket. Check out the site and see what you think. If you think it is a good idea go ahead and patronize their service.  Also I downloaded the new Bubba Sparxxx album...and it is very good and laid back. The problem is that outside of Deliverance you may never hear the other tracks off the album even though they are all pretty good so far. I will be adding this to my car collection to bump...as it is good R&B/Hip-Hop that isn\'t trash. Good work Bubba and I am looking forward to hearing more. I am sorry for downloading it without purchasing it but if I knew of a place to send you the 10-15 buck directly I would be more than willing to do that as your music is pure gold.  I don\'t have an aversion to paying for music as long as I get to test it first and then as long as the artist gets the lions share of the money paid.  Peace  '),(1007,1234212,'07,April,2003','2003-04-07 00:00:00','   I finally got my RAM back today and I am overjoyed...yay! I was having the same problem with this set however after jiggling the RAM a bit I was able to get it to run...YAY!  '),(1008,1234212,'07,April,2003','2003-04-07 00:00:00','   Well today I spent most of my day asleep with the exception of a few hours due to whatever is grabbing a hold of me and making me ill. When I first came back upstairs to head to bed and before writing this I puked up what was a good dinner brought over by the neighbors...I feel so bad because it was good but I was coughing so hard it couldn\'t stay in.  I was up for The Daily Show with John Stewert as it was live tonight. I guess I expected more out of them because it wasn\'t as funny as it could have been. As you may have guessed it was a live show covering the Recall Election in California.  I also watched a little of the coverage on CNN and something really struck me as odd. As early as 9% of the polls reporting back Wolf Blitzer was calling for an expected Davis concession of the race to Ahhnold (I can\'t spell his last name right now). Now this seems kind of funny to me...but don\'t you think you would at least wait until 50% have reported in to see what you would do? I guess counting votes really isn\'t as important as air time and we all like for our news to be quick and without any real merrit or substance. CNN is also calling the election now with only 33% reporting back that the recall went through and Ahhnold wins. Again...33% is only 33% that leaves (to overstate the obvious) 67% of the votes not counted.  Which brings me to another question...there were supposedly 13 million or so that came out for the election today. By my calculations if 33% have voted 4.2 million votes would have been counted...but on CNN it shows that only 2.8 million have been counted. If there were only 10 million or so voting that would still be 3.3 million votes counted...so where are the extra votes or are the initial reports incorrect about the number of voters that would come out. Also how do we know exactly how many actually came out.  I am actually for Ahhhnold in all this mess believe it or not. He is a moderate Republican with liberal social views...now the real test will be what he does when he finds out there is a mess to fix and he either has to cut programs or raise taxes...and it will be interesting to see if he settles the lawsuit with the energy companies because the state really needs the 9 billion or so from them back.  I am all just a little leary of it all after having read reports on the new systems that are in place in some places out there...but to be fair they are mostly on the old system that has proven ok in their state for some time.  Well I am a sick puppy and I need to rest for this coming weekend. Goodnight all.  '),(1009,1234212,'07,April,2003','2003-04-07 00:00:00','   I was informed today by one of the people that I \"work\" with that all of the Israeli women that I have met so far have thought that I was cute...and there was one that she is definitely hot enough to not toss out of bed for eating crackers....after hearing that news today I figured out why she was acting the way she was acting towards me yesterday...she was very giggly and silly (like a little school girl)...but she is no school girl rest assured...so it was nice to find out that someone out there still thinks I am attractive in some way.  Today was long and tedious and in the end I still fealt dirty. I am just glad that this day is over and I can get some sleep...because over the last few days I have barely gotten any with this sinus infection and now a dimension thrown in for good measure...I have a sore throat now. Great. Now I go into the winter cycle of sick, well, sick, and then well, all the while my overall health looses ground. I love the winter but I hate how it makes me sick.  '),(1010,1234212,'07,April,2003','2003-04-07 00:00:00','   Interesting details coming out about Ahhnold and Enron...if proven to be true (which we might only see if he is elected governor) it could mean that lawsuits filed against energy companies and their flagrant pilaging of California consumers during the supposed \"evergy shortage\" could be in jepoardy...to be sure they were short on energy...but the reality is that they were overcharged for the energy they did get by the energy companies...  Read and learn more about it.   Scandal looming?   Why do I bring these things before the public? Because I want people to think differently, question everything, and form opinions. Without that all hope is lost. Take or leave it...only the people have the power to change things and if we routinely allow our power to be usurped then when all is said and done we will have no one to blame but ourselves for the mess we are in.  '),(1011,1234212,'07,April,2003','2003-04-07 00:00:00','   I feel like a dirty dirty whore...without the pleasures...  That is all I have to say  '),(1012,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I got into another verbal sparring match with my mom today over the issues facing all of us.  She is still blind when it comes down to it...in the past I have considered her to be an informed and therefore educatable person...but I retract that idea. She still cannot see that the policies both my grandmothers and her generation have tried and used have failed and therefore have sown the seeds of discontent in the world. She cannot understand that the reason we are under attack from so many sides is because of the policies of prior generations and her own towards both Africa, The Middle East, and Asia and her subcontinent have put us directly in the crosshairs. She cannot bring herself to the idea that we aren\'t the good guys when it comes to  those parts of the world...that indeed we are the enemy and it isn\'t because we are \"the infadel\" or some other novel idea...but instead simply because we have opressed and killed many of their fellow countryman.  When you have the policies that the United States has had since the end of WW2 (1945 for those of you who have no clue of history) you are bound to make enemies.   We have consistantly had policies that bring us wealth and power while leaving the rest of the world with the tab. You see wealth is a finite thing. You can think of it in the simplest of terms...and here they are.  The world is made up of three disinct things...people, space, resources. Each disting thing you can break down into tenths. For the United States we have the following pieces of the pie...in people we have one fourth of one tenth of a percent in algebraic terms 1/4(1/10)% (we have about 280 million citizens the world has roughly 6.5 billion people)...in space we take up about the same amount...in resources we consume about 2.5 tenths (or 25 percent)...and if you want to boggle the mind we pump out about 25% of the total polution in the world.  Now our population is growing at the average world rate...with that growth it would not be long before we would require the entire power output of the world...so as it is now.  Now take into mind that our current power suppliers (oil, natural gas, and leastly coal) are all fossil fuels that are rapidly declining in supply...you end up with a population that is eating more and more power...with less and less available...and the rest of the world...well we Americans don\'t care about that do we?  That is where the resentment starts to come into play...not so much that in these little shit countries around the world even really need tons of energy like we do...but that we are taking their share of it without really paying them for it...so we are stealing it...and don\'t give me the line \"we pay 30 dollars a barrel of oil\"...yeah our oil companies pay that to themselves...and the money eventually heads off to regimes that don\'t like us...but like our money, and in turn those regimes opress their people and torture their citizens...when if they didn\'t have that money in the first place everyone would be on equal footing.  It really is so complex a problem that it is hard to lay it out in a way that makes sense to the Average person. It\'s not that people can\'t understand it...but rather they would prefer to think of America as this big benevolant provider of goodness throughout the world. When in fact that our government has long been involved in things that would never happen here...we funded a shaw in Iran for years that killed untold numbers of his own people...then he was overthrown by Radical Muslims (who then turned their hatred towards us because we backed the shaw)...then we paid for and helped the Iraqi regime of Sadaam Hussein develop WMD\'s to kill Iraniians (who for them confirmed we were out to get them so they hated us even more) not to mention that the people of Iraq knew where the weapons we coming from and the money was coming from (America)...the people that were propping up that evil regime...then when we looked like would fix it all and called on the people of Iraq to rise up with us against him in 1991 we left them flat...don\'t think they forgot that (so now we are targets there).  Afghanistan...we have an even worse story there. Nobody in the sixties or seventies could tell you where that country was...hell most people here now couldn\'t find it on a map (so much for our great educational system...leave no child behind...). Then in the eighties the USSR (our great evil foe...that is a whole \'nother story) invaded because it was on the lower end of the great Caspian basin of oil (they tried to consolodate the entire capsian region for oil purposes knowing that if they controlled oil they controlled the world)...so what did we do. Well we just got out of Vietnam so we weren\'t going to go rushing in there...plus not to mention it would have been directly Russians against Americans so we sent money....and lots of it. Untold billions actually. We funded a little known dude at that time named Usama Bin Laden (or Osama if you spell it the way americans do) and his Al Queda network (not sure on the spelling of that) and the mujohadine (again spelling) that fought against the Russian. We even sent advisors (as we did in Vietnam in the beginning) to teach them to fight. Well we taught them well as the Russians taught the Vietnamese well how to resist an invading Army...the Russians suffered a million or so casualties...but at our dollar.  When they pulled out in 88 everyone sang and danced it was a great victory for all but the Afghani\'s...because once the Russians left...so did our dollars (even though many people were of the mind to keep dollars flowing in to stabilize the political future as to a pro-western government...but we got burned on Iran so we decided not too)...when our dollars left the country fell into civil war for the next oh....lets say 10 years (our own was a brutal 4 years long) from which emerged the Taliban. Lo and behold we started funding them. The money was to obstensively go towards the destruction of poppy fields in our war against drugs...and they were effective at that as well as killing people, cutting clitorises off, stoning...all sorts of fun. When we asked for a little favor in return (an oil pipeline secured by troops from the Caspian bowl to India where an get ready for it....Enron power plant was) we were flatly denied...this coincided with the Planes into the towers...  Don\'t get me wrong. We should have gone after the people who perpetrated that. However if you will all recall it happened once before. We knew at the time (or soon after) that it was Al-Queda as we caught the dude behind it and he pointed his finger at bin Laden. Now why did we not go then? We could have (some people would say we didn\'t have the balls or Pres Clinton was weak...well he wasn\'t he actually promoted an unpopular position in the Baltics and sent troops to somolia which he then abandoned) tehn but didn\'t. People died then as well and they struck out our guts...but this time was a little different. And besides we knew who it was but we still haven\'t found him...why did we need to invade when all we could have used was special forces to surgically weed him out. Because we had a larger goal.  A lot of this is round the board on the \"making sense scale\"...but from what I have read, put together on my own, and been shown is this. That America really isn\'t as great as we think we are...sure we have all these great things (which I readily partake in)...but it all comes at a cost to the rest of the world. We steal, polute, and rape everyone else in the world so we can live the way we do...using the current technology we have.  How can we change it all. Well first of all we need more nuclear power....hey it is unpopular and has had its fair share of problems...but overall it is still fairly safe and can give us the go juice in a few years time rather than having to curb our appetites while waiting for better technology. The other thing we can do is research research research...and a little research on the side. We can innovate our way out of this...we can invent our way out...we just need to spend the money to do so. Under the current system we never will...but maybe just maybe the system will change a little this next election and we will see real change and maybe we can get ourselves out of this mess...  As for cleaning up the human collateral we have spilled around the world...we can\'t. All we can do is sit back and take the lumps on the chin. It doesn\'t mean we have to take it laying down...but it means that we have to live up to the mistakes we have made and make a good faith effort to fix them.  Point out one country that we haven\'t screwed and you get a gold star...cause we have screwed them all...if you look back far enough.  '),(1013,1234212,'07,April,2003','2003-04-07 00:00:00','   Have I told you lately that I love you...have I told you lately how much I care...you fill my heart with gladness, take away all the sadness...ease my troubles thats what you do  '),(1014,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok so now I am finally going to bed...finally. Tomorrow I have to pick up my grandfather who went into the hospital today because he nearly died at his dialysis...oh joy. I get to hear him bitch when I take him home...great.  Then Sunday I get to return to the crass commercialism of the mall to sit there and laugh at people and read a book.  I finally got that download in and installed the software and actually started a little site. Now all I need is the stuff from my friend and I can spiff it up quite a bit and add some stuff.  I think it will turn out nicely because it already looks good even though I am using one of the built in templates...as time goes on I will create my own for her but for right now a default template will do. There is a lot that I have to add to it.  '),(1015,1234212,'07,April,2003','2003-04-07 00:00:00','   I was contacted by a long time friend of mine (well I consider her a friend even though at this point we are more of aqaintances (SP)) and she would like me to help her build (read build) a website for her resume and modeling resume...lemme tell you this thrills me to no end. I of coarse agreed whole heartedly and want to do a good job...so what am I doing right now? Downloading some software to make a kick ass website.  There are things that I would like to do for it and I have some ideas...I just need to get the relevant info from her so I can start working on it...I will build for her at first something to get started (read good) and then add features to it as I learn them. It will be a great learning experience for me forcing me to get some things done.  I am giddy over it all really. I haven\'t done a real website for anyone but I have done a few for myself (but never been happy with how they turned out) so it will be nice to get something going in that arena...plus if I get something put together that is the bomb diggity I could get other work from it. That is always a good thing.  I took a day off from reading today...I really should have sat down to read the rest of the book but I didn\'t want too today. I want to save it for Sunday so I can read the rest of it. Note to self...get some highliters to highlite some of the good material in it...  After I get done with it I will probably get some other books to read...My next may be Jim Hightowers book. I also want to get one titled \"Bushwacked, Living in Bush\'s America\" I think that is the full title...anyways I think it would be a good read.  Well right now my DSL is bogged down crunching away at a 450 meg download at an avg. of 80kB/s...really fast but not as fast as a Cable modem (but DSL is much better for gaming.)  I also have to get together an updated price on a web server (basic type) to build one out for a friend and see what kind of fast DSL I can get to where (1.5 Mbps each way I think is 300 a month...not bad for a T1 linespeed).  I talked to my mom today as well and I got what I consider is one of the nicest compliments from one of my hero\'s. Bryan told my mom that he loves when I comment on his site and that I am very well spoken (or written) and have original ideas...my mom also told me she loves to read what I write (I know she is supposed to hehehe) but it was nice nonetheless because someone is actually noticing.  I would also love to point out that Bryan is probably the most articulate person I know in real life as he is able to pull together disparaging topics and roll out a final product that is easy, amusing, and informative to read...that is a real gift that he has. As always I check his site daily and try to glean a little of the genious from there and incorporate it into my life. (ohh ego stroke ego stroke) But seriously he is one of the best writers that I know personally and one of the brightest bulbs on the grand christmas tree of life. Kudos B.  Peace...keep reading and learning. Keep on keepin on.  '),(1016,1234212,'07,April,2003','2003-04-07 00:00:00','   Yeah one of the Jewish girls has a little thing for me I am certain...and I dunno what to do. Perhaps it is because I have been out of the \"game\" so long. I really have no clue where to go from where I am at (silly banter) to where I want to be (freaky circus sex).   Maybe I should explain the ladder theory to her...although she doesn\'t speak english very well so I would have to use even more simple terms to get the point across...then I would know for sure where I fall on the ladder.  What is the ladder? What is the ladder theory? Oh my you are missing out. Let me link it for you to all behold. I hold it as self evident as the forefathers held certain truths to be self evident. This is truly a master of work that needs to be taught in college...hell even in high school as it would save a lot of energy and perhaps keep a few kids from blowing their brains out or shooting up the high schools.  Well maybe not....but it certainly is worth a good read and from personal experience it is dead on.  I present you with the  ladder  theory.  Enjoy...and oh yeah I feel like a dirty whore.  '),(1017,1234212,'07,April,2003','2003-04-07 00:00:00','   As I read more and more of Al Franken\'s book the more angry I get at the right of America...it just infurates me to no end to recall some of there efforts now as a liberal (or moderate liberal) and see them all in a new light as I continue to educate myself in what is really going on.  To think it is even legal to be a Republican is beyond belief...because they have done just terrible things and whenever they get caught they downplay it all...but by God when a dude gets a blow job he has to be impeached (remember impeachment is actually for any federally elected person).  The point is this folks. Republicans blame Democrats for all that is wrong and accuse them of class warfare over and over again.  The truth is that the majority of our current problems with terrorism are spawned out of the eighties and the regimes and terrorists we funded (al-queda being one of them from Afghanistan). As for class warfare I would say that the Republicans have waged it ever so much more effectively. Essentially they have given a pass to all rich folks on taxes (and those rich folks who cheat on the taxes that they should pay rarely ever see any repercussions when a person that makes fifty grand a year would likely see his or her livelyhood destroyed by the IRS due to penalties imposed for a minor indescretion that a rich person makes yearly on their tax forms).  The fact is that the rich have gotten richer to the tune of 110% of \"real\" money or value since the eighties while the lower class has actually lost about 10% of \"real\" money or value in that same time...so the poor get poorer. If you question these statistics you are called a traitor or said to be conducting class warfare because you blame the rich for getting richer...no I blame them for the poor getting poorer...I don\'t care if they get richer but only if at the same time we see the same numbers of growth across the board...because as the rich get richer prices go up and those that are actually loosing money value suffer because of that and are forced into ever deeper debt just to keep their heads above water on actually doing things like feeding their families or making sure they have heat.  Class warfare can also be seen in the educational system as this...public education is a mess. We all pretty much agree on that. Private education is the only way to go in many places to have your kid achieve something...but if you can\'t afford the tuition you don\'t get the better education. So only the rich can afford to be educated (when in reality they usually are the among the dumbest of people out there because of their \"selective\" inbreeding to keep wealth at the top.)  Another thing that irks me is this...the debate of the 87 billion dollars needed for Iraq.  Well actually it is the 67 billion needed for the military portion. Yup thats right I am questioning \"supporting the troops\".  The question should be this...what are we getting for the 67 billion dollars? My thinking is this. Ok so it is going to the troops right? Well is it going to pay the salary and combat pay of the troops? Well some of it will...but the point is this. The regular pay should come out of the existing Military budget because they would have to pay them that salary anyway if they weren\'t at war. The combat pay should be paid out in the 67 billion because that wouldn\'t and shouldn\'t be allocated in the regular budget. The same goes for all the reservists...subtract the normal monthly costs from the Pentagon\'s normal budget and the extra to have them deployed come out of the 67 billion.  Surely some of that money is going to go towards replacing all the HUMVEE\'s and APC\'s being blown up. My question is this. Don\'t we already have a buttload of these vehicles? I mean since the eighties the budget for the military has always been over 200 billion a year...during that time you would think that they would have amassed hundreds of thousands of these vehicles...and seeing that the HUMVEE is supposed to have a service life of 12 years or so then it would make sense that a lot of them would still be around. Sure we need to replace them as the get destroyed...but can\'t we replace them over the next few years instead of all now...and where are all the rest of them?  Ammunition...ok this is a given...they need it...we don\'t always have huge reserves of it so it needs to be bought. But does the budget cover new M-16\'s, M-224\'s (the squad weapon), M-60\'s...if it does we already have millions of M-16\'s and tons of the rest of them...so do they need new ones? When in fact the majority are still using M-16 A1\'s...I can understand updating them all to M-16 A2 or M-4 status...but that should come from the regular military budget as they would have to do that anyways.  Missiles and bombs...hey again we have been spending all this money for all these years and we don\'t have a buttload laying around? Yes we need to resupply...but why are we running short in the first place.  Fuel...yup you need that and you really can\'t make huge stores of it...so I can understand that a large chunk would go for fuel...but at the same time we just invaded an oil rich nation...I\'m sure they are getting there gas at less than 1.50 a gallon.  The questions should be \"what is the 67 billion paying for?\"...is it covering expenses that would normally be covered in the regular budget...and if it is what is the normal budget going for?  I would challenge them to give an accounting for that money...the 20 billion I can actually understand. We bombed it we bought it...and I bet if you found what looks to me to be the fraudulant overcharging of the military to the taxpayers we could come up with that 20 billion in savings from this appropriations bill.  If it is all straight up...then so be it then it needs to be spent...but if it isn\'t on the up and up...then withold the money until they can explain the new expenses.  '),(1018,1234212,'07,April,2003','2003-04-07 00:00:00','   I am getting restless and am getting close to wanting a new phone...if I get a better job here soon I will be getting a top of the line Pocket PC phone with built in Qwerty Keyboard for Sprint...the phone is the bomb diggity but costs over 600 bucks right now...but that would get me a new phone, a camera on the phone, and the pocket pc that I have wanted, plus the ability to hack it and let me go online from a laptop anywhere...by getting or making connectivity software that allows me to tap Vision services over a USB cable...  Ahh just need the money first...I need to get Ebay\'ing  '),(1019,1234212,'07,April,2003','2003-04-07 00:00:00','   I started reading Al Franken\'s book today...and it is as good as Michael Moore\'s but a little different as it goes specifically after certain targets like Ann Coulter, Bill Oreilly, and others...thankfully instead of just spewing lies as those people do he backs up all of his statements about them with facts that are readily available.  When you are faced with bald faced lies like he has done with all of the people he has attacked (so far as I have read 110 pages into the book) it would seem that these people would at least tone down the rhetoric...but instead they say that people are lying about them or they lie about lying about lying about something (no that isn\'t typo...it is meant that way for a reason read the book and you will see that some of the people lie about lying about a lie they told in the first place).  Again I haven\'t ever had so much fun reading a book as I have had with these two...and they don\'t even have naked women in them...that is totaly awesome! Well they could add the naked women and the point would still be driven home that unfortunately America has been lied to and too many people buy the lie.  '),(1020,1234212,'07,April,2003','2003-04-07 00:00:00','   For the first time ever I bought a book and read it all the way through in one day.  What is this marvel of writing brilliance you may ask?  \"Dude, Where\'s My Country\" by Michael Moore.  Pure genious...get it and read it...then reread it. The book is just excellent and he is able to bring together many of the thoughts all of us left-wingers have been feeling for quite some time. In addition he basically says in some places \"suck it up and stop being such whiney little twits\". Those are my words...and they are great.  It is a great book with bits like this on pg 72 \"Ethiopia (nothing like sending a squad of starving children to help!)\" and this one on pg 71 \"Afghanistan? What exactly was their contribution going to be? Horses? Ten sticks and a stone? Don\'t they have their own problems at the moment?\" both on the Coalition of the Willing.  Or this one on the left admitting it has made some mistakes a dealing with them \"Why are you still bitching about right-wing author Ann Coulter? Sure, she\'s as crazy as a loon, but she\'s got more balls than the entire Democratic Leadership Council. You\'re just jealous we don\'t have an Ann Coulter. And stop looking at her damn legs! That\'s why she\'s got you so jumbled up!\"  And this one \"Animals don\'t have rights. Yes, they should be treated \"humanely.\" Yes, Tyson Foods and all the others that \"harvast\" chickens are disgusting. But \"freeing\" chickens from their factory farms is idiotic. They don\'t know how to survive in the wild and they\'re just going to get hit by a truck...This shit just makes me wanna go kick my dog.\" (a bit was left out of that one about milk and whatnot)  I really like this book.  Today I bought it and Al Franken\'s \"Lies and the lying liars that tell them\" and I will start that tomorrow.  Well I am going to be...as it has been a long ass day so far!  '),(1021,1234212,'07,April,2003','2003-04-07 00:00:00','   Gotta love Bushco and their endless peddling of this war. We are in the thick of it and they say all is rosey and going well. Americans dieing at the rate of about 6-8 a week right now doesn\'t seem like things are going well. As well as the prospect of our troops being deployed for a year or more (just think about it folks they are already strained with 140 thousand troops deployed with there being few reserves to send so to think that they will be back after a year is a little optimistic at this rate of \"recovery\"...I personally think troops may be deployed longer than a year possibly up to two years with little R&R.).  While I want to say to the troops good job and way to go and keep your asses safe I would also like to say this...quit bitchin\' you signed up for it...for service and while it certainly isn\'t fair that your there and I am here you all certainly knew when you signed your enlistment papers that the government owns your ass for the next few years and they can do with you what they please.  If it were up to me I would bring them home now...the Iraqi\'s seem ready to take over...and if they are in the position of holding an election but that can\'t happen until all our companies secure contracts and we pacify the entire nation. I would say get out and let them handle it...let them rebuild off of oil revenue and let oil companies actually compete for position over there like they would elsewhere.   However I think another reason Bushco doesn\'t want to leave them to their own devices yet is a pretty good argument...they would likely duely elect a very Religious Right government -ala- Taliban style. That probably wouldn\'t be good for us...but hey there is a lot of medicine that we are due. We are Americans we can take whatever they dish out.  On another note I started writing something of a book. It will certainly contain better grammar and punctuation than this blog (or at least I hope it will). The running title is \"I\'m Not Coughing, I\'m Laughing. A life with C.F.\"  So far I have a prologue already penned out. It needs a little more but I will write more as it comes to me. I have the first chapter penned out completely as it is the shortest that basically starts out my life...and hey what can I say other than \"I was born...then diagnosed two years later. All was normal for the first five years of my life.\" That is pretty much what the first chapter says but in a little more wording and with a little background to build consiquent chapters on.  I don\'t really expect it to be published but you never know. If it all comes out as something really short I may enlist some of my CF friends here in PA to write a little about their lives and compile it into a book together with the title still being \"I\'m Not Coughing, I\'m Laughing. A look into X lives with C.F.\"  It is a novel idea (heheh forgive the pun)...and I think it is about time I pen out some of my ideas. Plus I have two works of fiction that I really need to get to work on. I have rough draft outlines for them all I need to do is flesh them out and make the worlds that I dreamt up reality in words. The one story I already sort of told here on this blog a few weeks back. It was the dream that scared me half to death...but the dream was so real and the visuals are still there. I just need to find the right words to put it down.  I love working but I hate my job...and that is about it for this update. Till next time. Peace out yo!  '),(1022,1234212,'07,April,2003','2003-04-07 00:00:00','   argh...har har har me maties...hard to starbord and lay down the plank  Off to the hospital for a couple of days...this visit should be over Saturday.  Peace...hey think of it as a vacation.  '),(1023,1234212,'07,April,2003','2003-04-07 00:00:00','   It will take much bloody war and people to throw themselves into the abyss of the system...but we will win.  Civil Disobedience over the Diebold Voter Machine problems that threaten to take away our right to an actually elected official.   Students take on da\' man   Go and download the documents...read them and be frightened...plus it keeps them out in the world instead of them dieing a digital death at the hands of the powers that be.  '),(1024,1234212,'07,April,2003','2003-04-07 00:00:00','   Well not much really to post or rant about today except that some new law has passed that makes illegal certain Abortions.  While personally I am pro-life my political view is pro-choice. What? What the hell Mike? What kind of wishy-washy position is that?  It is simple for me really. I am a man...or at least last time I checked I had all the male parts. My personal beliefs and views are my own and not to be laid on anyone else. Should I stop a woman from having a procedure that is completely up to her deciding? She will be the one that has to live with it both good and bad...so for me to impose my views on her is rediculous...and for me to support any law that would restrict that basic right (to choose to do something) goes totally against even the Bible.  Yes even the Bible. For all you religious nuts out there your God says one thing...and really it is his only basic tennent (even though a larger tennent would be to love each other) but that basically says...You have the right to choose to believe or not. You alone are the sole decision maker to make decisions that go with him or against him...but in the end you will be held accountable for those decisions.  A life is a life...but a bundle of cells that is essentially treated as a virus for a period of time in a womans body does not take precedent over that woman...as she is complete...the bundle of cells is just that...a bundle of living cells. We have no problem attempting to eradicate Anthrax (unless you are the US gov...then you try and develop it into a weapon), trying to stop the flu bug, or fighting AIDs...all of those are living beings...but they are considered bad so we have no problem in trying to kill them.  Again I am distressed at the new law and it would seem that we are on the road towards outlawing abortions all together which would be a bad thing...a bad thing because it takes yet another right away from the people.   The right to have power over your own life.  Think about it and fire back at me if you think I am wrong. I am willing to debate this one to the hilt...and unlike many out there I see both sides...but I choose to side with freedom.  Tomorrow I go to the docs..and I fear that I will be heading to the hospital tomorrow...we will see. As always I will update and let people know what is going on.  Peace  '),(1025,1234212,'07,April,2003','2003-04-07 00:00:00','   I got a PM from a friend over at HR...that was kewl.  He suggested that I get a few pictures to give as an example of \"twinkies\"...and hey that is a damn fine idea. This next weekend I will get some pics of \"twinkies\" so as to illustrate what I mean.  Peace  '),(1026,1234212,'07,April,2003','2003-04-07 00:00:00','   I added something new...see if you can see what it is :-)  '),(1027,1234212,'07,April,2003','2003-04-07 00:00:00','   I am tired and I still have to go get my mom in a couple of hours from the airport...I am tired...hey can I say that again I am tired.  Today was yet another day in corporate consumerism hell...the mall is possibly one of the worst places to work. The \"twinkie\" factor was really out today.  For those of you who don\'t know what a twinkie is. Here is the definition. Twinkie: Noun, pronoun: 1. A tasty sugar flavored and cream filled snack cake made by Hostess. 2. A derogitive description of the female of the species that looks like everyone else i.e. dressed alike with same cell phone attached to ear. Usage. Wow! That twinkie tastes good. or Wow! Look at all the twinkies walking by...I think I just got diabetes...please god kill me now.  Yes a twinkie in this context is the way girls dress at the moment...about a year and a half ago or two years ago black sorostitute slut pants were all the rage...in and of themselves they don\'t make you a twinkie...but if you are blond and in a pack and all have similar shirts on....then those are twinkies. Now it is Abercrouchy and Scratch...almost like scratch and sniff.  The only delight of my day is lunch in which I put my headphones on and escape into a realm of music (not the continuous loop crap they run at AF which we are across from) head upstairs and grab a slice of pizza or some other tasty morsal of mall food.  I must rest before I head out to the airport...so I will probably not update again tonight unless something comes up.   Till tomorrow...rage against the machine, throw a monkey wrench into the works, and break it all down. Because until we build a new one we will be stuck with this outdated piece of shit we call \"the system\". Funny in a culture that loves to aquire new shit all the time we haven\'t gotten a new system.  '),(1028,1234212,'07,April,2003','2003-04-07 00:00:00','   I added a new site that people need to see...we are taking it to the right in the way that they have taken it to us.  Time to change from tree hugging, owl loving, liberals into junkyard dawg, bar brawling, tree loving liberals. That is the only way to fight.  We can be nice and finish last...or we can use their tactics (only we will use the truth in addition) and take it to the mat.   Bob Beckel on Politics  '),(1029,1234212,'07,April,2003','2003-04-07 00:00:00','   Go read some of The Nation...it is a good read as it attacks policy with an utter viciousness that only the right has been able to accomplish...only this is a leftist news source.  They make some great points...and having been in the business of disource for over a hundred years they sure have a good reputation.   The Nation   Time for bed...I have to work in the morning...oh yeah and I got my grumpy grandfather from the hospital today. At least it was amusing for me because he told me some war stories (he was a member of the Navy in the Korean War).  '),(1030,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I went down to see Danielle this evening...and I must say she is the most bomb ass person I have met in a long time. She is truly a funny person...its good to see a sense of humor in someone who has just gotten the shittiest hand of cards dealt to them.  We took a few pics and here is the funny one...I am very self consious about my teeth...out of everything my teeth are the thing that I would change about me first...I would love to have perfect white teeth.    She has a piece of cardboard in her mouth that is yellow/orange to make me feel better about my teeth. It was funny as hell.  She really boosted my spirits tonight. I dunno but she is damn kewl...and while I was there I got lots of comments about my hair. Everyone loves it...yay me!  Well I am tired and wiped out from the meds...I am going to get some sleep.  '),(1031,1234212,'07,April,2003','2003-04-07 00:00:00','   Hey folks I am writing from my laptop in the living room.  How? You may ask.  Well I got my data cable and new phone yesterday...and let me tell you it is the bomb diggity to be able to walk around a surf the web.  It clocks in at 115Kbps here at the house...and can connect at up to 160Kbps...so wooohooo!  I will be taking this to the mall on Sunday to update from there...this is sooo kewl!  '),(1032,1234212,'07,April,2003','2003-04-07 00:00:00','   Here is a good read on why I think that in their current form TouchScreen voting machines are a bad idea   Voting machine problems.   I would be behind a voting machine that was designed by our best and brightest at MIT, CIT, and CMU...otherwise if it is built and maintained by corporate interests then the vote will always be suspicious.  This artical is well thought out and written with grace and poise...yet is able to get to the problems in simple elegant language that even a moron could understand.  Read it now! Or don\'t...but be forewarned our vote is no longer our vote it is now controled by big business!  '),(1033,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok so today I got my new phone...coming on the 4th of November I will be getting a Pittsburgh number as well. Time to get it out of the Indiana area.  So far the phone is da bomb...I am just waiting for my data cable to arrive. It should be here Monday...which will be great!  Today was long and slow...like a good hardcore fucking...only without the money shot.  Anyways...I am out for the evening  Peace  '),(1034,1234212,'07,April,2003','2003-04-07 00:00:00','   An interview with Michael Moore and Amy Goodman from Democracy NOW!   Interview page   Watching it now....so far so good.  '),(1035,1234212,'07,April,2003','2003-04-07 00:00:00','   I am ever more amazed at the overclocking ability of my little P4...now getting 3.00Ghz on a 250FSB at 1.65 Volts...damn this thing is just screaming!  I was having problems getting the P4 to ramp up higher than 2.7Ghz but I figured out why it was having problems...the RAM was way overclocked and wasn\'t agreeing with that speed...so I clocked it down a little and it is showing me some love...it is running at slightly over 200Mhz (or 400 DDR rates)...  I love this little Shuttle box!  '),(1036,1234212,'07,April,2003','2003-04-07 00:00:00','   Well that was strange...  Ok...the phone I am getting is a SprintPCS Vision phone that supports data rates of up to 144Kbps...that is quasi-broadband. It will be fast enough for internet on the road and hospital trips...so that will be great. Now I just need to save for a new laptop.  The laptop I figure will run about 1700 bucks. 2.4Ghz or higher, a gig of RAM, Radeon Mobility 9600 128mb RAM, and a smallish 40 gig hard drive, plus Wi-Fi.  That would be good.  Well I am tired and I need my beauty sleep.  '),(1037,1234212,'07,April,2003','2003-04-07 00:00:00','   This will be my new phone as I have already ordered a Data Cable for it...   Sanyo SCP8100   Now I just need a few more bucks (another paycheck) and I will have it :-)  '),(1038,1234212,'07,April,2003','2003-04-07 00:00:00','   No pics of the case yet but they will come. I need to finish off the damn thing so you can see the full cool blue effect that I am going for. I need to order a couple of blue LED fans and some other crap...once I have all that up and on it I will get you pics :-)  The hospital wasn\'t so bad...although I had a rough start. They came in to access my port and only had a 19 gauge needles. That is big when I am used to a 22 gauge. When the dude first tapped it he didn\'t prime the needle with saline so he would actually be able to get a flow...so he had to take it out and go at it with another needle. When he drew the first one out it bled a little and he slapped on some alcohol just for good measure. If you want to get that feeling cut yourself with some paper and dunk the cut in alcohol...that would be a close feeling. Only this was on my boob...so it hurt a lot!  I was able to spend a lot of time with my new friend Danielle...she has been in the hospital for around 2 months now. She is currently number one on the Lung Transplant list...all things considered she looks pretty good for only having about 10 percent of a lung working for her. I would imagine that is how bad it is but I don\'t know what her PFTs are.  I try to tell people what I live in and some may see me as complaining and saying that I have it worse than many others...well I do. This was confirmed by Danielle. We both are in the same boat (while I am not nearly as ill as she is though). She has run into the same problem in conveying this to others who aren\'t CF patients and looking down the barrel of death all her life (as I do). She has friends and family surrounding her but yet she is alone in the struggle. Only CF patients can sympathize with each other as to our illness.   I guess part of it is that we want to desperately be independant and healthy but we know we can\'t have that. I have often felt almost disgusted when people offer to do even the most simple of tasks for me and generally treating me like a gimp. Now its not that I don\'t need or want the help, but instead if I need or want it I will ask. Danielle is in the same boat. Sure she can\'t do everything and needs a lot of help, but I think she feels the same way I do in that we don\'t want to put upon people to do the simple things. I know that people aren\'t put apon by helping out if they offer but it is still just the pride issue...CF patients pride themselves in being as independant as possible (from my experience at least) and it is just hard to ask for help.  I think that is why her and I are getting along so far. I will help when she asks but I won\'t jump whenever she coughs or try and shuffle her things around to help out unless she asks. That is the way I want it in my personal life. I think she sees that and respects that. Because it is sooooo frustrating to have all that attention placed on you when you cough, or when you poop out, or for anything. We want to be normal and conduct ourselves as best we can. The best thing I think I can say to people is treat us like a normal person, if we need to rest or need help we will ask. Don\'t ask me repeatedly if I need a break...if I need one I will take one.   I took my lucky buddha that was given to me by a friend John and Amanda and gave it to Danielle, I also got her some bamboo. They have helped me and I think they will help her.   Here is something that I am going to ask my friends out there in cyberland. I don\'t do it with ease because of my own personal beliefs...but if you could do it I would be greatful. If you all would say a little prayer, or whatever, and keep her in your thoughts I would be greatful.   Also I remind you that you need to fill out your organ donor cards or check the box on your driver\'s license to be an organ donor. It would be a great help for all. You never know you may need one yourself and if people would just do it we would all be better off.  Here are some links.   Official Organ Donor Card   Another one   Remember fill them out and keep them with you...also tell your family that you would like to be an organ donor so they know as well.  '),(1039,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I just got home a few minutes ago...I will be making a large update later this evening :-)  Oh yeah and my new case cover came...I will try and get a few pics of it :-)  '),(1040,1234212,'07,April,2003','2003-04-07 00:00:00','   I will be going back to XP this next because I will be getting a new box. Longhorn is really nice and it runs ok...but there are somethings that I am not ready to give up just yet...and that is Norton System works. Without any kind of protection from Virii and many other things I am afeared (yes I meant to spell it that way) that I will run into problems.   Once a Beta comes out...or the final product I will stick with XP.  Until next week I will run it and see if I get used to it...but I will most likely switch back.  '),(1041,1234212,'07,April,2003','2003-04-07 00:00:00','   Done a little tweeking to it and so far it looks a little better...the speed is a little slower for some things then XP...but overall it ain\'t to shabby  '),(1042,1234212,'07,April,2003','2003-04-07 00:00:00','   There are some programs that don\'t play well with Longhorn...Norton SystemWorks 2003, DataPilot (phone software)...so there are some things that need to be worked on...outside of that it is ok  The boot up time is long, the shutdown time is a little long, and the browser renders a tad slower than IE 6...the installed version here is 6.05.  Looks good overall...but needs a little work yet. I will keep it around for the time being to see how it works.  '),(1043,1234212,'07,April,2003','2003-04-07 00:00:00','   I got it installed and I am on it now...the graphics are creamy. It also has a built in download manager that is really k-rad!  More as I see more :-)  '),(1044,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I am installing Longhorn Alpha on my desktop box...so far creamy. I did an upgrade...but I would rather have a fresh install so I am re-installing it on a clean HD.   What I saw of it while I ran it for a few minutes I liked.  I tried a few months back to run it...but the build was way to early and buggy. Now it is more mature and easier to work with.  Also all the drivers I threw to the new OS it liked so I think I will run it for the time being.  Well I am going to monitor it and give updates as things go on.  Oh yeah I figured out what is going on with the desktop system...it just doesn\'t have enough juice to run at the speeds it should. So I am going to be replacing it and give my mom my box...with the specs that I posted before.  The new box will be a Soltek EQ3401M...that will be creamy as it has room in the system for two optical drives and two hard drives. Plus it has a mega 250watt power supply.  The power supply is key here. Becuase that current little box has a 200watt power supply and it just can\'t keep up with the requirements of the system that I am running. To many whats are being drawn off.  Plus the box just looks creamy.  What it looks like.   So I should have all the things upgraded by mid-week next week. That will be hella cool. The drive bays are stealthed on it so it looks hella sweet.   I am happy that I will be able to get everything ready for an extended stay in the hospital. In the near future I fear that I will be spending a lot of time there and if I am going to have a system and I can\'t afford a good laptop for some time then I may as well go ahead and make sure that my system is portable yet powerfull.  I am off to monitor the install now. Peace.  '),(1045,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok...so I think I figured out what is going on...finally. Well maybe. We will see after some gaming what happens with this machine of poo...yes it is on my shit list right now.  Working fine at high speeds for a while then just doesn\'t want to co-operate. It is still OC\'d but only slightly now...and that cheeses me off! (although I am getting a free boost in speed anyways!)  So I went shopping. You see my mom wants to get a computer and wants to spend less than 400 bucks on hers. Hell I can do that. Well I can do it will shifting things around :-)  First of all since my system will not run at full speed because of problems with power (yes I think the Power Supply and Mobo just can\'t handle the wattage demands) I can run it safely at up to 200FSB (=800 FSB with QDR on P4)...  I shopped for a Celeron 2.4 on a 400Mhz FSB (which can safely be overclocked to a 533FSB and the mobo will do it well withing its 800Mhz limit), two sticks of RAM for a total of 512MB, and the requisit drives that would be needed to get it up and purring like a kitten. The total came to 387...now lemme remind you it will be a 3.1Ghz Celeron after the overclock, CD Burner, DVD Drive, and 512MB of Dual Channel DDR (at 333Mhz), and over 50 gigs of HD space. So that is a kick ass system for the price. But I would have to get a new Mini-System to replace mine....which adds 260 dollars to the price.   That 260 is actually less than what I spent on my Mini-System barebones in the first place (doh!) and as her processing needs grow I can add a faster processor. The only drawback that I can see is the built in graphics...but that can be remedied as well if the time comes.  So the total comes too 648 bucks (rounded up) to get me back at full speed and get her a kick ass system! Plus my current system has a pretty nifty case! Also it would be powerfull enough to run Longhorn when it finally makes it to market! YAY!  So I am going to talk to mom about it...I could pay her back the 260 bucks over time...or it could be a Christmas gift from her and GM...Either way it gets us both up and running.  As for other things...well I got a new music reporting thing for my site as you may see here.  My health is shitty shitty bang bang...I hate being sick. The steroids are really fucking with me right now. They make my blood sugar go all goofy and when I take the prescribed dose of Insulin to get it back down it crashes down instead of just easing down (I nearly passed out today even though I took less Insulin than was prescribed for my Blood sugar level of 287).  Also the roids are making me very edgy and mean...I am on a very short fuse right now. I mean short. I nearly yelled at GM and Eileen today...even though they didn\'t ask anything taxing of me...it was just so close. I could barely control it. So I appologized for being short and warned GM that the roids are making me looney. I told her that I was on a short fuse and that if I snap at her I don\'t mean too and she needs to call me on it and tell me to calm down. While that will make me madder I can distinguish between the roid anger and true anger and cool off...or at least I hope. I am going to call the Docs tomorrow and see if we can do something about that. I am already a prick...I just don\'t need this putting me over into the Asshole catagory.   On another note I finally got my system to recognize that it can capture video and I got GNN\'s \"Unanswered Questions\" ripped onto my hard drive...it is a large file and I encoded it twice. Once for broadband and once for High Quality. I will try and find a place that has big storage as they are big ass files and will require mucho bandwidth allowances to download. However I think I will be able to get it out there. It is a good video.  Halliburton had its extension of a contract cancelled today. YAY! Partly because the Pentagon came out and said that it was being charged twice as much per gallon of fuel than it should pay (it pays around 1.20 a gallon now and Halliburton was charging 2.49 or something like that for fuel to Iraq...the pentagon says they can do it cheaper with their own equipment and even shipping it from the US if nesecary)  Bectels contracts are coming under scrutiny finally as well.  Well I am going to play a little bit of SOF2 and then hit the sack...I am tired.  Peace  Warning...I have roid rage!  '),(1046,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I don\'t have to go to the hospital again...well at least yet.  Instead I am going on a heavy diet of steroids and insulin. We will see how that goes. One good note is that the steroid will help me gain a few much needed pounds.  Computer is still acting like a whore. I contacted the manufacturer and we will see what can be done. I am hoping that we can just exchange it for a different product. I just think these sticks are not suited for this board as they give off much RF noise...as can actually be heard if you hold your head close to the case and scroll a page.  The next step is if I can\'t get that fixed then I will have to get different RAM. I would go with Corsair XMS Pro...which has crispy LEDs that indicate activity. They will be expensive but they will be a good investment as they can be moved to a newer system when the time comes.  I may in the end have to build a new system. DAMNIT! I hate that!  '),(1047,1234212,'07,April,2003','2003-04-07 00:00:00','   I take back the crashing thing...I think I may have found the cause...indeed the hard drives are part of the problem...but the RAM is giving off a strange buzzing noise as I scroll or do other things...when I wiggle a RAM stick and get the noise to stop it does stop...but sometime the computer freezes...so I am thinking that part of the problem may be the way the RAM is connecting. That would really suck...but it might be ok because there is some new exciting RAM out on the market that would make for a good replacement...and it has pretty LED active meter lights built in.  I watched the Debate tonight on CNN. Rock The Vote.  Overall this was certainly a better forum than I have seen in the other \"debates\". I will break it down for ya\'ll as I see it.   Dean  Dean got trounced in the first few minutes. However let me explain why. He made some statements lately saying to southern voters that he wanted those guys in the pickups with Confederate Flags voting for him and that he was the Candidate for them.  Immediately a dude in the audience got up and called him on it...trying to bait him into a corner and then let the others pounce. Which it worked well. He said essentially that he wanted to extend his campaign to everyone white, black, brown...everyone and you can\'t simply label people as racist  and others as not. We need all people to vote to get Bush out.   I think Kerry was the first to pounce demanding an appoligy. Dean tap danced a bit.  Sharpton took exception to that and attacked...and did so effectively. He called for him to appologize. Dean said that he wouldn\'t and stood firm. Saying a few things but really danced around a lot.   He was attacked again and then nailed it home. This isn\'t about race...this is about including people and you have to get through to them on their level...even if they fly those flags. It is time to stop being goaded into a Race debate and focus on just people.   My personal thoughts on it are the Republicans have been the ones to make Race an Issue and play on the fears of both sides...and have gotten away with it because they deal with the extremes at either end of the spectrum which are never likely to get together and compare notes...so they would never see the lies told to them.  He got some of his views across on GLBT issues...which at least he stood firm on civil unions.   Kerry  Kerry...well he came off ok...but he was a stuffed shirt and unfortunately wasn\'t interesting. He didn\'t really say anything of note. Except I think he didn\'t get the meaning of one of the audiences questions which was framed in a baseball analogy. It was whether or not he would switch out people if they become innefective as Ashcroft, Rumsfeld, Wolfowiz, and Rice have. He eventually said he would...but instead of taking it one directly he used analogies...don\'t appologize for having a view. Just make sure it is informed and you will never go wrong. He was also asked about hunting and didn\'t really impress me on his gun policy, but he did say that if we enforced the current laws then we wouldn\'t need new ones...he took a pot shot at Dean because he has gotten an NRA endorcement in the past.   Clark  Clark...well he fell flat on his face. He simply wasn\'t able to articulate his ideas in the least. The question that you could tell he had problems with was GLBT issues (Gays, Lesbian, Bi-Sexual, Transgender)...he just didn\'t make a good impression.  In the end he is really a political puppet for the Clintons...and frankly once people realize that he will loose what support he has. Not that people don\'t love the Clintons...but it would be easy to make a case of hipocracy against him because we are making a big deal out of Bush being a puppet for Cheney or Rove. The knife cuts both ways and people need to realize that.   Sharpton  He continues to impress me on his ability to connect. I don\'t like him...but he at least is a real life person and not something mysterious. Not much to note although he was much more aggressive tonight than he has been.   Braun  She was very demured...very stately. Her answers were well thought out and easy to go with. I think she realizes she can\'t win but wants to bring her issues to the table anyways and maybe the candidate that does win will carry on the ideas.   Edwards  Not much to tell. Motivational...but that is about it.   Lieberman  He was at least more lifelike...less like a corpse. He needs to pull out. He really is to nice of a guy to be doing this. Also he is more like a republican than most republicans.   Kucinich  Ahh behold the dark horse. This little guy is the wildcard. He would be a Green party member but has the smarts to be a Dem.  His website.   Look over his policies. He is excitable and doesn\'t speak well but he has all the concrete ideas and has maps and plans to back them up. I have to give it to him. If candidacies were actually run on platforms and ideas this guy would get the nomination. He has so many ideas and goes after many things that don\'t make sense and would be better off changed for all of us. Of particular note is his \"water for everyone\" ideas, full gay marriage (he said that in the debate tonight and got perhaps the loudest set of cheers and applause), all sorts of things that would help us out. He has laid down the research for you to look at.  By far the best candidate for issues and ideas...but the worst because he is so far left of center that he almost crosses round the globe to opposite side...but I bet he is there flippin them the bird and telling them to pick up their styrofoam burger boxes.  Just check out the site.  For me I still support Dean, but in a close second for me is Kucinich...  Well that is it...  '),(1048,1234212,'07,April,2003','2003-04-07 00:00:00','   I found a couple of 160GB SATA Hitachi drives...currently the top rated for 158 a piece. 8MB Cache, 7200RPM...Hitachi bought out IBM storage a while back. IBM got a bad rap because of some faulty drives...but out of all the ones I have used they have been the most reliable...although I did not have that particular model. Plus the drives offer a three year warranty...hard to come by. Also the shipping is included. They have 250GB versions...but they are much more expensive. Like 315 bucks...now I ain\'t gonna pay that for another 100 gigs of space...nope.  I would need to get a couple of external cages for the drives I already have...as they are faulty but not dead yet....so I could use them for extra storage...which would mean that I would need another power strip and possibly a small nuclear plant to power it all.  If I get the external housing units for those and some of my other drives I would have about 510GB of storage total...more that enough to rip my entire CD collection plus some.  2 - 160GB = 320GB (7200RPM SATA) 2 - 60GB = 120GB (currently in my box to be put into external units 7200RPM PATA) 1 - 40GB = 40GB (7200RPM PATA) 1 - 40GB = 40GB (5400RPM PATA)  = 510GB of total storage  That would give me plenty of room to play with I would think. I have some movie stuff that I need to get on here and convert to digital format so I would need that kind of storage overall.  Well I am extremely tired and I am going to pass out for a while...  peace all  oh yeah and it hasn\'t crashed since I rolled back my Nvidia Drivers...  '),(1049,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok so my box isn\'t running as stable as I would like it to be...I think that I am suffering from HD failure...on both HD\'s nonetheless..crap!  We will see after this next reboot...I am going to have Norton Scan them to see if it can come up with something.  It really is a shame because most new hd\'s come with only a year warranty...they used to all be three years, but in these days of fast moving technology and cost cutting they dropped most of their coverage.  I am going to go ahead and shop for some new drives just in case...this just sucks ass!  Oh yeah and I feel like shit on top of it all...for the past three days I have been coughing up lots of blood! I hate that! I go to the docs on Wednesday and if I am still coughing up blood I will probably be headed back to the hospital...I know that my PFTs have dropped since I left because it feels like it.  I just hate this time of year when it comes to health. Over the past two years during this period of time I have lost over 20 percent of my lungs...so I expect I will loose another 10 percent this year to infection. That will bring me down into the 40% capacity level...and bring my already low O2 sats down to around 92 or 90 on average (they are at 93 to 94 average now)...and the most important indicator, the small lung passageways are already down to 10-12% of normal...and only bound to get lower (that is where you actually get your oxygen)...I think that I will have to seriously start to think about transplants soon...it won\'t be for a year or more...but I had better start to prepare myself...and make the decision as to whether or not I want to get transplants.  It just boggles the mind.  '),(1050,1234212,'07,April,2003','2003-04-07 00:00:00','   Wow am I bored or what...oh dear baby jesus fetus.  Soon this day will be over.  Have I said what a life saver this phat internet connection is...well it is!  '),(1051,1234212,'07,April,2003','2003-04-07 00:00:00','   Well my day here at the mall is nearly over thank blue glowy orb (you thought I forgot my promise of not using God and instead using Blue Glowy Orb)...well like I said it is almost over.  Today wasn\'t to terribly busy overall because therer wasn\'t much traffic in the mall...although we did do a pretty brisk business today. So I don\'t feel to bad just sitting on my ass all day long.  I also got a pay cut. Well she at least asked me before she cut it. Actually I am relieved because she was paying me way to much for what I actually do which consists of me getting here in the morning, opening the cart up, dusting (if I feel so inclined), and then just pretty much sitting here to people watch. Sometimes I sell something...but that isn\'t a guarantee.   It really is just kind of a dull around here. Although as I have said before I work with interesting people. The Israeli\'s next to me are pretty damn kewlio...and the Indians to the other side are great too.  It never fails to amaze me how many Israeli women this dude has working for him...I mean it is like a veritable eye candy sugar overload....so I guess I do enjoy myself a tad.  Also working in the mall is evil...but I do get to see all the hotties wandering around here...theree are some that I would just do very dirty things too...on Thursday there was this obious bit of jailbait strumpet...but damn did she have some legs on her. I wanted the to be wrapped around me...but that is illegal (to be fair I am pretty sure she was at least 17...maybe even legal...either way DAMN!!!!).  Well as I sit here and ponder what I will have for dinner (my mom is coming to take me out) I must reflect on some sad news out of Iraq.  16 US soldiers were killed today in a helicopter crash...well it was actually shot down. Another 20 or so were wounded.  The more time goes by the more I get angry about that whole mess. It is simply terrible that all thise guys up there in power (except Powell) basically did everything in their power to get out of Vietnam...and did so quite well. Yet they are the \"patriots\".  I can\'t wait for the 04\' elections...so at least we all may have the ability to get this train wreck of a country back on track.  Once that happens maybe we can get back some of the good will the world once held towards us...but if we have another 4 years of this crap...we\'re screwed, and there ain\'t gonna be a kiss afterwards.  '),(1052,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok so I got the laptop up and running...it purrs like a little itty bitty kitten.  I am now at the mall and now I am updating...ain\'t technology grand?  I will probably post throughout the day on and off. I am also going to try and get some pictures of twinkies while I am here.  '),(1053,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I nearly have my laptop ready to go for tomorrow...all I am waiting for is a few updates and all will be swimingly well.  I am almost completely convinced it is time for a new laptop. This one is just damn near wore out. She has served me well. Almost time to move on and find a better one...preferably one that I could even game on if I felt the need.  However I am thinking that I will go with the most portable system out there...something small and lightweight yet powerfull enough for basic computing. Maybe something with a Centrino (or Pentium M processor)...I just don\'t know yet. We will see.  '),(1054,1234212,'07,April,2003','2003-04-07 00:00:00','   it has been a long ass day...but I think I have got all the computer crap taken care of for now...  I am really tired yet I have a lot to say...I wish that I could keep my eyes open long enough to tell you all about it...it kinda took me by surprise...eeesh  I promise I will have a better update tomorrow...well hopefully :-)  Till then,  Peace Out  '),(1055,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok moms computer is pretty much finished...all I need to do is transfer over old data from one of her hard drives and I will be all set...but I think I will do that later today before I install the DVD burner...as I will have my case open already.  Wow...it has been a long nite...so now I must rest!  But on the plus side...the FSB pn this beast os 1078MHz...or in CPU terms my processor is 3.25GHz....and the RAM is running at a cool 437MHz at CAS 2.3.3.6...very tight memory settings.  The temp on the processor runs around 45C which is damn good for the overclock that I have running now. I also got a little more out of my video card...it is running at 46oMHz core and 640MHz RAM...so that is also fast.  A lot has gone on...and by god it is all working.  Expect that the posts later today will be more political as the 30 hours of Senate Debate should be wrapped up by then...all ready for me to comment on that whole mess.  Till then...peace!  '),(1056,1234212,'07,April,2003','2003-04-07 00:00:00','   Now I got the phone recognized....yippeee!  '),(1057,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I have got moms computer installing everything now...that should go ok now...earlier it was sticking but now it seems to be just fine :-)  A lot of work accomplished today.  One strange note...my Cell Phone isn\'t recognized by this system...I need to fix that...hmmmm  '),(1058,1234212,'07,April,2003','2003-04-07 00:00:00','   Well it is all up and running back where it was...oh yeah.  Now this new system is running at much higher speeds than I would have ever imagined before. I mean really this thing is screaming along without a hitch at all.  Now what am I gonna do with all this power? I dunno...but I know I am gonna have some fun with it.  Right now I am waiting on a Download for Nero 6.0 and I am still waiting for a drive that should come tomorrow...TDK Indio DVD+-RW 8X and 40X CD Burner...woah nelly!  So I guess I just need to get moms up and running  She has a Shuttle Mini system now with a Celeron 2.0GHz that will be overclocked...512 Megs of RAM, 56K modem, and 40+ gigs of HD space...16X TDK Burner and a DVD Drive...now I just need to get it up and running for her.... :-)  w00t!!!  '),(1059,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I got the new stuff in today with the exception of a drive...that should be here tomorrow...I am getting all the boxes up and running...I will update through the night as I get things done.  :-)  '),(1060,1234212,'07,April,2003','2003-04-07 00:00:00','   Well this system has been actually running for almost a week straight on Windows Longhorn. Now for peeps running XP out there thay may not seem like such a long time...hell I had XP up for two week stretches at a time without much incident many times. However...Longhorn in its current form is still very very beta...as in they just got some of the features implemented...and not all of them work, but for it being an early beta it has been rock solid stable for the basics. I just simply haven\'t needed to reboot the system.  Now having said that it is starting to slow down even further...a result of one of many memory leaks I am sure. All OS\'s have memory leaks...it is just how well each OS deals with those leaks is how we end up with a good or bad experience. In the 9X platforms things would grind to a halt almost daily...you would need to restart and refresh the system. About every six months you would need to reformat and just do a general house cleaning on a system  NT was a little better in its various forms...the memory leaks it experienced were easy to deal with as the memory managers they built into the OS were a hell of a lot better at keeping things tidy. So if a program went rogue it was easy to terminate that process and free the memory it was using. That carried over to Windows 2000...and probably 2000 has been the best OS to come out of Microsoft. With NT systems it was and is common for them not to need rebooting but maybe every week or two...unless you got a virus or just liked messing around with tweaking your systems out...then you may be in for reboots daily regardless of the OS you choose.  XP is by logic an extension of 2000 and it inherited quite a stable of features...mainly reliability and usability...but it also inherited some of the nastiness of 9X series of OS\'s...that is Virii...that has been the bain of a one OS environment that Microsoft is quickly learning. To make a system that is easy for users to use...you have to ratchet down many of the native security features that XP Pro has and they ended up with Home...now the big problem is they are the same in almost every way but because you have the home version out there you essentially dumb down the people and make them lazy as to keeping their systems safe. If they were all getting the heavy corporate or pro edition with all the security turned on we could possibly cut down on the Virus plague.  With Longhorn Microsoft will attempt to do that...and really I hope they can do it...sure Microsoft is big, bad, evil, world dominating, and all that is wrong...but in the end they have brought computers to the masses. They have made it easy for grandma to get her email in many aspects.  Other os\'s are trying to get there too...and to be sure OSX is a great OS...really I will give Apple Credit where credit is due...they have great products...but because they are expensive they have marketed themselves out of the 400 dollar computer arena...which is where the growth is being seen.  Linux...great...another player. It is a great OS as well...but overall if something goes wrong there aren\'t many resources to go to and get help. The Linux community is its best asset and worst foe...most of the people in the Linux movement who are academics or are genuinely computer people are willing to help people out...then there are the toads that are elitist assholes who act as if you are a retard because you couldn\'t get your sound card to work. I have run into both..and that is why I run Windows...not because I am retarded but because if I need help I am not always sure that I can deal with ignorant pricks that instead of helping out will instead call me everything but white and call into question my experience with computers.  So what will the future hold...who knows...I hope that Microsoft can reform and play a little nicer...but I doubt it. They are competitors, they know what they want and they have the cash to get it. In being that they have been ruthless in the business...but also they have driven the competition to be better. If you failed in a war with them it was because you fell behind. Adobe Photoshop is the tops in photo editing...not because Microsoft doesn\'t have photo editing software...but because Adobe\'s product is superior...by far.  For me when Longhorn hit I will install it...I will run it...I will learn it. It is the future...and the future looks good even from this very early operational beta.  What the fuck was that post about?  ');
INSERT INTO `msgs` VALUES (1061,1234212,'07,April,2003','2003-04-07 00:00:00','   Had a long talk with Jesse tonight...I must say dude got his shit together real fast.  I make no bones about being extremely reserved in my early thinking about what all he wants to do. I still have things that I need to resolve for myself...not about his plans or about how is going about getting things together...but they are for me to resolve in my mind.  One of them is just plain fear of having not ever done what we are planning. It is this fear that I have of failing at things so I often don\'t even try...just to avoid the possibility of things failing.  Jesse put into me tonight on the phone the fire and desire to see this through...regardless of what happens.  For a long time I have been a believer in one simple tennant for personal success...and that is you have to make it for yourself. Most of the people that have gone places have done so with nothing more than a wing and a prayer...and so many have come out on top.  Well I am going to step in to uncharted territory and throw my energy behind his vision...not because I am just saying it...or paying lip service...but indeed I do believe in what he...and we will be attempting to do.  Sure there is a risk of failure in it all...there are always risks...risk comes with rewards...without risk we would have nothing.  So over the next few days I will have to write a bio on myself for some of the plans...and I will have to get down to business in trying to make lucrative contacts around abouts for support and ideas.  Jesse has laid out his plan...and it is good. From where he has it now I believe that we can make the vision a reality...and what a reality it will be.  So the next few days I have to get into a mode that will allow me to build up my personality...or rather my personality on paper. The next few weeks I will have to gather information to help me prepare for the next steps. The next few months I have to prepare myself to make some personal changes that will allow me to see this all through...and part of those choices have to do with doing my damndest to get healthy as this venture will most assuredly be hard on the body...not so much in a bad way...but because I will have to be on my A game all the time to do what we need to do.  I have a doctors appointment tomorrow and I will basically be like...look I have a lot coming up in the next few months...I need every drug that I can get my hands on to make myself healthy and still be able to function completely....give me the Anti-Biotic that makes me itchy...but in doing that get me pumped on the steroids and antihistamines as well...also get me something for the stiffness and general lack of energy that I have...get me performing at my peak so that I can throw all my energies at this venture.  This also means that I am going to have to hussle to get money together...which means I have to get off of my lazy ass and get some of this shit sold...save money...and just be ready to sprint into action when things are ready.  Jesse this goes out to you bro...I want to thank you for considering me in all this. I know that recently I have been dodgy about all of this. Yes it was because I had reservations...but I now believe in this with all my mind and heart...I am sure this is going to work. I am sure that we will be able to make this a go project. From this point on I am going to be that dude that is real, but is ready to run out into traffic and make things happen if it is needed. I am ready for this...you are ready for this. We have finally gotten to the points in our lives that we are ready to make things happen for ourselves and move on from our current positions.  Here\'s to the future...I tip my vodka tumbler. We are ready to go. Lets make this happen.  '),(1062,1234212,'07,April,2003','2003-04-07 00:00:00','   Its on folks...got some big things comin down the road.  '),(1063,1234212,'07,April,2003','2003-04-07 00:00:00','   I watched an Al Gore Speech on the Patriot Act. It was a great speech...here is the link. You will need Realplayer to see it.   Al Gore on Patriot Act.   It is an hour and eleven minutes long...and well worth the watch. Go for it :-)   It is the first link on the page there.  '),(1064,1234212,'07,April,2003','2003-04-07 00:00:00','   It looks like the parts will be here on the 13th. YAY!  Tonight I saw my friend Carrie on Elimidate. Tonight was her nite to shine...and shine she did. Well sort of.  I love Carrie in the greatest way...I see her as the virginal girl that we all lust after...but with that side that you want...a dirty little secret side that only comes out in private...or at least that is what I think.  Well her Elimidate was just all wrong. Having been in the Pittsburgh area on and off over the past 6 years (all adult years...how weird) I rarely run across guys out there that are even somewhat intelligent.   Now I am not tooting my own horn...but I score very high on the IQ test...some tests I score in the 130\'s and in others up to 150. So I am by no means a dummy. However it seems that Pittsburgh is full of guys that are pretty much dumb as bricks...or celery...or whatever veggie you want to compare them too.  There were four guys on the show to win Carrie. The first was a Wrestling dude. He seemed a little different from the typical sort...he may not have been the brightest bulb on the tree...but he was at least somewhat coherant. Now as a guy that is comfortable in his sexuality this dude wasn\'t any great shakes in the looks department.  They had another dude on there that is in a Rockabilly Band here in the Burg. Unfortunately he may have been the most intelligent of the group...but the way the show was cut you maybe heard him say maybe a sentence or two. Plus my little Carrie didn\'t know what Rockabilly was.   Carrie...think of the Stray Cats...or rather think of Brian Setzer Orchestra (sp?) of \"Jump, Jive, and Wail\" fame...Brian was a founding member of the revival of Rockabilly in the Stray Cats. It is pretty good music that is catchy...a little bit nitchy...but hey it is musically better than most pop music. Download some of the Stray Cats...and you will get the jist of the beats that they use.  Now the next two guys were the \"winners\" of the pack. And I don\'t mean winner in a good way. First they are twins and between the two of them they may have enough brain matter to actually guide a rat through a maze...wait no strike that....they may have the combined brain power of your average orchid. I will give it to them in one way...they looked ok. Think of Oz from American Pie fame and you will get the idea.  Having known Carrie since she was a youngin...16...I have kinda seen her grow up over the years. She has turned out to be a beautiful woman and a great person. The show really didn\'t do her justice because of the editing.   Now here is where I need to get nasty about this whole affair. This is where I will go into a rant about women, men, and young people in general (including people in my own age group the mid-twenties).  I am getting rather tired of seeing pretty boys...(I\'m no great shakes...but hell I ain\'t a dog). Pretty boys give all decent looking guys...and guys less than decent a real hard task to get anyone that we may desire. However what really gauls me is that these guys were just your typical wannabe frat boys...idiotic, dumb, retarded, and pack animals...not hunters but more like a gang of heinias (sp?). They really have no skill of talent of any sorts. Sure they will get through college (and finish unlike myself)..but it won\'t be because they are really all that bright...but because of the educational machine.   Now I will cast stones here. I am smart. I have spent a lot of time researching various topics and consider myself fairly informed on many subjects...and if you are lucky enough to meet me in person on a good day you will see that I am also a great orrator. I can take a subject that is hard to discuss and make it easy...not by dumbing it down to pander...but by bringing in lingo and jargon and explaining it to people...and I can add some humor to even the most boring or scary of topics. I take pride in the fact that I can get up in front of a group and give speeches that border on comedy...but it comes with a message. Very similar to Lewis Black (save the anger) and George Carlin (albeit without his vast knowledge of semantics and great command of the English language). I can take a subject and make it understandable. That is why I teach at my local library computer courses for the older people of the community. I always get compliments and have ended up with several clients from those sessions. Everytime I get up to be \"the center of attention\" I always make sure to include everyone and try to be accomodating to peoples needs. I need to please...that is my purpose.  I am also very good at debate...maybe that is where my other skills come from. I pride myself in being able to debate anything as long as I have some leadin to what the subject is...if it something I know a lot about then it will be a breeze to debate on that topic...if it is something I don\'t know much about then I attempt to educate myself about the topic. In doing this I take no position in the beginning. I look at facts and roll them into my own conclusions...that usually strikes a balance.  In having spent time honing these skills I have largely seperated myself from the social norms of my age range. Most of the people I know at my age are fairly settled...yet they are rather uninformed on many things. I have chosen to be a \"geek\" in the sense that I yearn for knowledge on things. I want to know all that I can and am always willing to listen to something new.  What has it gotten me? Well it hasn\'t made me the most popular person because it is hard to debate against people of my own age when I share little in common with them. Yeah sure I can kick back a few beers...or dozen...but I am not good a small talk. I have not honed that capability. Instead I have chosen to hone my skills in a more meaningful way. So I am rather dull at a party...unless I shred off my grown up attitude and become a kid and just have fun...which is hard for me to do right now with all that is wrong in the world.  Now wrapping it around to this Elimidate I found it disheartening that the future Men of America is not interested in anything further than they ends of their dicks. These boys (and no they weren\'t men) were out for ass...plain and simple. Is that wrong...hell no, but if you are gonna go out just to get ass do it with class at least. Don\'t be players...that is wrong. This crop of guys they chose were nothing short of the failings of our system.  Now I must focus on women. It is often said by women that they want: 1. A sensitive guy 2. A good looking guy 3. A smart guy 4. A good guy that treats them with respect 5. A non-player.  Now I know this because I hear it over and over again from my female friends...and they are all the ones who choose to go out with exactly the opposite of what they want. They end up getting treated like shit, then they cry about it to people like me. Don\'t get me wrong; I enjoy being there for my friends when they need support, but I am starting  to really get tired of the cliche\'.  There are nice, funny, decent looking, smart, and considerate guys out there. Usually you are talking to one of them when things fall to shit. Those are the guys who are the catches out there...yet they are constantly overlooked for the pretty boy that will just use them for sex...or have the female as an accessory...much like a cool cell phone (I have one...and it is cool...but I don\'t need another cool thing to be hanging from me).  I have often said that I would give my left nut to be with some of the women that I console...and I would. Not because I want to bang them...(although being a guy and being programmed to breed through millenia of evolution sex would be great...but it is never nesecary to have that right off the bat)...instead I would just like to be that guy that they will look back on and say \"jeesh he really was good to me and actually cared about me wholely and entirely.\"  That is what I want. I want someone who is my equal...a woman that can be intelligent, smart, beautiful, and sexy all rolled into one (as everyguy proffesses he wants..but in my case this is what I want)...I don\'t need eye candy...that is what internet porn is for when it comes down to it. I want someone who will complete me and fill that gap that I have.   Does this mean I am picky...yes to a degree...not really that much in the looks or sexy department...but I am extremely picky when it comes to the smarts of the person I want to spend time with.  I want to fawn over them...be there at their beck and call...yet I want the favor returned...I need for someone to bring me chicken soup when I am sick. I want someone who will be there when I am blue, but will let me be in that mood and not get mad because I don\'t snap out of it right away.  I am in a unique position I feel. I have CF...you can see it in the title. I will most likely die because of it. Actually I may die in a year...two years...who knows but I will die sooner rather than later at my current health decline rate. That puts me in a dilema of sorts...because I don\'t want for someone to become to attached because it would kill me to know that when I pass that person will have to mourn for me. That is worse than death to me.  That is part of why I don\'t actively seek a woman to share my life with.  Another reason and it is as plain as the nose on your face is that I don\'t have the energy to really persue anyone. I did that once and expent a lot of energy. I simply cannot do that..that puts my health at risk.  I want a comfortable relationship that when I pass it won\'t be so huge that anyone would be devastated...I would want to be remembered...but remembered in a good way as that I was good to people...and good to the people that I love and care for.  There is still someone who has my heart...and I don\'t give it freely...and it is only because she was my first love...and we all know what a bitch that is...you never can get rid of that first love.  I have room for another...and I would always be willing to entertain a new relationship with someone at anytime as long as I know what is expected up front.  I am not perfect...but I am not flawed...I am average and I just want to share myself and my life with someone.  I have one person that I would gladly invite back into my life at anytime and she knows it...I hope that would happen but I doubt that it would because I hurt her badly in the past...and yes I actually admit that it was my fault.  So Carrie...you know I love ya lots...I hope that you find someone that will be your equal...and know that I am always here.  Beth...you know that you are still engrained into me...you were there in the worst that I have been...and I greatly appreciate that. I would love to be together again...and of coarse you know that as well.  Will either of these things happen. I doubt it...but I can hope and dream can\'t I?  I just wanted to share this all and get it out there...because sometimes I find myself at a loss for words when in person...but I would ask that people look past the somewhat shy outer shell and see what is inside...because what is inside is beautiful...just like what I see in ya\'ll.  No this isn\'t a weirdo stalking thing...because while I of coarse care what you think...my life doesn\'t hinge on it...and regardless I will always be the same guy I have always been. A good decent young man.  On the finishing school thing..I would love too...and most people know that. Most people know that really all I lack are really about 30 credits to have several degrees...but I would hope that people would understand that at this point in my life I can\'t finish...my health won\'t allow for that no matter what happens...I would end up failing not on concept or knowledge...but on being there. That is why I haven\'t committed to finishing my degrees.  Here is what I could have if I applied my credits wisely and took some extra courses...BS in Computer Science, BA in History, BA in English, minors in GIS and Math. Its not that I don\'t know how to achieve these things...but it is because I physically can\'t do it. It would literally kill me.   This doesn\'t mean that I won\'t go places...I am working on that and have been working more and more. I am becoming rather succesfull in my own right. I work now, I have clients, and I have a growing base of knowledge from which I can make money. So while I failed at standard education...I have succeeded in life in my own small way.  '),(1065,1234212,'07,April,2003','2003-04-07 00:00:00','   I don\'t know if I mentioned it...but I got those parts ordered so I will have my moms computer built whenever all that stuff gets in. I am expecting it on Thursday or Friday...yay!  '),(1066,1234212,'07,April,2003','2003-04-07 00:00:00','   Dean chose to go with Private financing in his campaign by turning down 19 Million in public funds. So it is now for all of us (his supporters) to give a little to the campaign.  He made a very eloquant speech that can be seen on my link to DeanTV.   He sent out Emails to his supporters (about 30,000 votes were counted) to ask what he should do...and most people chose to go with Private funding of his campaign.  Now I am for campaign finance reform, but right now we are in a state of emergency. We need Bush out and Dean (or any democrat for that matter...and this coming from a registered Republican).   Bush turned down the public financing so that he could build a war-chest of about 200 million bucks to fight off a non-existant primary foe. So what will that money actually go towars...well he will be running smear campaigns to try and discredit any Democrat that may pose opposition to his re-election bid.  Right now Dean is at around 20 million for the primaries...if he chose public financing his total cap would have been 49 million or so to run his campaign on...that is hugely unfair.  all we need is 2 million people to contribute 100 bucks to the campaign to help him out. I will be giving 100 bucks for the primary and another 100 bucks for the general election if Dean is chosen to run.  You can see a link on my page here to go and give him some money. My total right now is 0...but I will be adding to it soon (like this month).  You can go give money and you will be credited with giving the money...so go do it...we can all spare 100 bucks over the next 10 months...or less. So I encourage you to vote with your dollars and help Dean become the next President of the USA.  Go ahead you know you want too :-)  '),(1067,1234212,'07,April,2003','2003-04-07 00:00:00','   The internet is shit!   Funny site...well funny in that it is true. Oh so true.  What is the internet to you.  For me it is a vast expanse with little knowledge and lots of rumor flying around. I find it a great tool, yet a great waste of time. For me I cannot get off the internet because I do use it for much of my information gathering...but I also take whatever I read on it with a grain of salt just like I take any printed material...or any published material for that fact.  The truth is in the eye of the beholder.  '),(1068,1234212,'07,April,2003','2003-04-07 00:00:00','   Well today has been pretty slow overall...we have done about a hundred in sales...give or take a few bucks...  We need to be making more...but ya know you can only trip so many people in a day...and you can only force them into wanting something that they don\'t need if they are in the mood to whip out the cash or credit card.  What really astounds me is how easily people pull out there credit cards to get shit. I mean whenever I pull out my credit card I feel guilty...usually because it is something that I don\'t have enough cash on me to afford...or something that I couldn\'t afford in the first place.  Like I placed my new phone on the credit card...240 bucks...no biggie really. I can pay that off in two weeks...and I have the money for it...but still I fealt guilty about putting the money on the card.  I paid my personal credit card today online...and that is good...just needed to get that out of the way...I have plenty left on it...not because it has a high limit but because I really don\'t put much on it. I put gas and bought some stuff online with it...but I got it paid down to only a few dollars of being paid off (in other words I will have it paid off in a month).  I guess I am lucky because I went through a really rough spot with credit cards...I had one that I ran up too about a thousand dollars...and I had family that helped me out...and in doing that I cut up all my cards and haven\'t had one in my own name for years...only about a month ago did I get one in my own name.  Now I have had a credit card attached to my grandmothers card...and that has a huge limit on it...but I haven\'t put anything on it (save for the phone) and last night we put about 640 bucks on it for computer stuff...of which my mom already gave her 500 towards it...now I just have to give her some money (for the phone and previous balance that I owe on it)...I have been pretty good with credit for a long time...in that I don\'t go out and spend mad amounts of money.   Everytime I want something I wait for a few days and debate it...usually I don\'t get it if I don\'t have the cash money for it...because I already know what it is to be in debt.  Overall not including my student loans (which will never get paid back) I owe around 6300 total (on my car and credit)..so I am in pretty good shape...oh not to mention my monthly phone bill which this month may be a little high.  I have made do with around 530 dollars a month for a long time now without really blowing loads of cash...granted I don\'t go anywhere...but at least I know that if I want to get something or go somewhere I have the credit available if I would like too...  Oh yeah and my house arrest...funny thing is that he didn\'t come back to put the anklet back on...instead he took the kit and cabootle with him...I just have to be in by 6PM (and really I am still technically on house arrest just without the anklet). He asked me if he could trust that I won\'t go running off or break the rules...which I won\'t run off...and as for rule breaking...I can never not do that...but I can adhere closely to them.  I think it came down to that I had it off almost a month because of going into the hospital...I think he just got tired of having to come out and do it every couple of weeks...plus it adds to their costs.  I should still be off of it by mid-December...only a few short weeks away.  Well my boss came by and I have a few things to do before the end of the evening...  Peace out ya\'ll.  '),(1069,1234212,'07,April,2003','2003-04-07 00:00:00','   I really hate the holiday seasons...why?  Because there is a plethora of abnoxious twits walking around breaking shit...fondling things...some of these kids are the prime reasons that women should be allowed to have abortions...they didn\'t need to be here in the first place...  What is that addage...you will get children that are tens times worse than you when you were a kid...well if your were one of those children that people didn\'t like...then if you get knocked up...do humanity a favor...vacuum that puppy out of there. We don\'t need a bunch of little souless kids walking around making the world a more miserable place.  Fuck I hate kids...  '),(1070,1234212,'07,April,2003','2003-04-07 00:00:00','   I got work in the morning...I will be taking my laptop with me to post from there.  Got everything ordered tonight. YAY! Oh yeah and I cleaned my room...shock me shock me...I can see the floor!  Well I am tired and I am going to get some sleep now...peace out!  '),(1071,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I am at work now...and so far it has been pretty busy.  I just dunno bout this billy...it is pretty strange :)  My mom brought me more Benadryl. I needed it.  Arghhh....I will try and update later this afternoon.  '),(1072,1234212,'07,April,2003','2003-04-07 00:00:00','   I am always looking at my hit counter statistics...I need more people to read this thing...  So tell your friends and get them to come here...of coarse this is up to you...I don\'t mind writing for a few people..but if you come across someone who thinks similarly to myself then send em over. If they don\'t agree with me...send them over...its all about the love.  Also if you have links that you would like me to add I will be more than happy to do that to increase trafic to those sites as long as they link back to me.  Peace  '),(1073,1234212,'07,April,2003','2003-04-07 00:00:00','   New obscene laws about to be passed by the government...this is why I constantly harp on Free Speech and the rest of the bill of rights.   Why oh why?   Op/Ed piece...scary stuff   I believe in the bill of rights as it was originally drafted...and I agree with most of the amendments to it as well.  We have the right to free speech...it is there to protect speech that some of us may find questionable. That is the beauty of this bit of law. It is also why it is first. Without it there can be no public discourse or debate on anything...the founding fathers realized that and made sure that we can say what we want when we want. It also makes the separation of church and state sacrosanct. Their fear was that if the government pimped a particular religion than that would effectively close down free speech, free press, and all of the things that they were fighting for.  We have the right to keep and bear arms and maintain a well regulated militia. The idea behind this was to say that the people could organize in active opposition to the government if the government was to get out of control. They could not foresee the problems that people have with this. Nonetheless they have it in there. So I believe that if you are a citizen of the US then you are allowed to own any type of weapon you want...even a howitzer. They made it so other laws would be enforced if you chose to abuse this right by indiscriminate killing people...hence the oenal code. If you want to hunt a deer with an M-16 then so be it...just don\'t shoot up the local school yard. Banning all weapons for private ownership and use only hurts us...and here is why...because the people must have the option of rising against the government in times when the government runs amok (much like it is now).  No soldier shall be quartered in a private residence unless in times of war and only with good reason. This stems from the fact that the British military would often come in and take your house and use it as a base of operations in an area. It applies now moreso then ever in my opinion. We have seen the military come into areas and taking over in the name of \"the law\"...examples include Waco and Ruby Ridge...having the military being able to take your house is saying that the government can turn our forces against us the people.  Search and Seizure...basically laid out to make sure that a police or military force can\'t just come into your house and look for stuff. This right has been nearly stripped from us...moreso than the right to bear arms...cops nowadays simply pull you over and they have probable cause and search your car. Whenever I have been pulled over I have almost always refused them a search of my car and call them on it when they shine a flashlite into my backseat saying that they can\'t do that. This right has been the most eviscerate of them all. Also it covers re-imbursement for property that has been commandeered...when was the last time you got a check for Los Alamos? They took the land and never re-imbursed the people for that land.  The Fifth amendment takes a big bite of many things...basically saying that you cannot be charged with a crime by one person or on that persons authority...that they have to get an indictment...it too has been eviscerate through the use of closed door grand juries. It also makes it a right to \"plead the fifth\" so that you are protected from implicating yourself...again it has become distorted by many in the corporate world...what do you think when you hear someone take the fifth...you think \"GUILTY AS CHARGED\"...it is ashame because it really only means that you don\'t want to tell anyone what you know or don\'t know.  Six is the right to a speedy trial and due process. Pretty obvious here...but lately it has gotten out of hand. Everyone who steps foot in this country is protected by it...but under the current administration and Patriot Act we can now be thrown in jail without due process or a trial...it was to try and prevent one person from becoming to powerful and locking up opponents.  Double jeopardy is the next one basically. This right has been upheld all the time...and it is one of the few that still works well. It only works well because of the rich constantly asserting it for crimes they have done...if it were only poor people being dragged through the system then this right would surely have been rescinded. Trial by jury is also included to make sure that the people have a say in the system.  Fair Punishment and fair Bail...well this one is obviously gone...ask any young black man who sells pot on the corner. They get long sentences and huge fines, plus they rarely get fair bail. This is one of the ones that applies conversely to your wealth...the richer you are the less you have to pay for your crimes.  Nine touches on the fact that the rights cannot be denied to anyone...and was later strengthened by including minorities and women. It says that the law can\'t be changed willy nilly.  Ten asserts states rights. Basically ceding federal law to state and local law...the foresaw that they couldn\'t make plans for all contingencies...so they gave the states flexibility to shape laws how they saw fit as long as those laws didn\'t interfere with the Bill of Rights. However notice lately in the debate over medical marijuana the Federal Government is trying to enforce its views over the states...it was put in the bill of rights to allow citizens to make up their own law for their communities without being bound to a slower moving federal government.  That is your quick lesson in the Bill of Rights.  Hopefully you learned something. You should because these rights are under attack from all sides by the Bush Administration. Fight for them.  For the people who want to take away gun rights I challenge you to prove that if private ownership was banned tomorrow crime would go down. The laws on the books already make it illegal for felons to own weapons. The laws that are in place now only interfere with a person who is obtaining weapons legally...the criminals will still get them regardless...plus if you are for banning weapons then you are probably liberal...and you should be fighting to keep the right because of the current way things are going...it may come down to an armed conflict between the government and the people...if we don\'t have the tools to fight them then we would loose and things would be much worse...a little revolution now and again isn\'t always a bad thing.  So read up and become informed of your rights...Exercise them daily. Use them often. Make them your own. If you don\'t then in a few years when they are all gone you will find yourself in a world of hurt.  Does this mean Mike that you oppose gun control? No actually I don\'t oppose it...I think some controls are fine. You should have to wait for a weapon so that checks can be done...but when it comes down to brasstacks you see that the current controls really don\'t filter out the people that go buy a gun then shoot others up...because obviously if they obtained the gun legally the person was found to not be a threat. You can\'t regulate anger or behavior...so don\'t try to. We have to accept that there will be gun deaths...but we can make sure that the most egregious acts that are committed are punished to the fullest extent of the law.  That is all for tonight...I will step down from my stump now :-)  '),(1074,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I am trying to get a hold of Outkasts album...there are some good songs in it...I hope I can get it all.  Now I am finally going to bed  '),(1075,1234212,'07,April,2003','2003-04-07 00:00:00','   As you can see I updated the look of the blog...and I got everything moved into this template...finally.  I am learning HTML and stuff on the fly here. I am glad that I can figure it all out.  Just got off the phone with Jesse...and he will be coming out this next week...hella sweet. We have a lot of paperwork to go over and get done.  I promised more about Danielle...but I am very tired. I will be getting some rest now.   Peace, Love, Unity, Respect.  '),(1076,1234212,'07,April,2003','2003-04-07 00:00:00','   I am glad that I am home...it is good to be back in my little room with the fat internet connection...hehehehe  Well I got the heatsink/fan unit for my system....got a thermaltake and it is a little whiney...but it keeps the system cool. I also got some Arctic Silver 3...and I got a new keyboard mouse combo that I have had my eye on for a while...the MS Cordless Elite Pro Optical or whatever the hell it is...  Normally it is 99 bucks...but CompUSA had an instant rebate on it...well they did for a few days but neglected to take the sign down...so I got the rebate four days after it expired...and I got the setup for a total of 79 bucks...the keyboard is really responsive and the mouse is as well...it has the second axis on the wheel so you can scroll horizontally as well..that is nifty and I will use that a lot in gaming (assign it as straif or something)....overall I would rate them both together as pretty good...better than the Logitech setup I had before (which was wireless) and more convieniant than the wired setup I ripped off of here to upgrade to this setup...  Only time will tell.  I got to visit a lot with Danielle today and it was just the bomb. I really enjoy talking to her as she is very bright...and seeing that we are in the same boat (well hers has more water in it right now...my boat just got a new bucket to bail out with it would seem) it is nice to be able to talk to someone else and be actually able to relate to them. It makes me miss Melinda a lot...I wish she were still around...she really did help me during that time that I knew her.  I am honored to get to know a bit about Danielle...we talked a lot about relationships today and how they apply to us. We have come to similar conclusions on it all...We would really love to be in a healthy relationship...but it helps if you are healthy. We both want them...but at the same time shy away because of what goes on in our lives...her reasoning I suspect is similar to mine...and mine is that I really don\'t want to drag anyone down my path with me...because that when I do die I don\'t want to have that person mourn me...that is simple. It would be selfish to enter into a relationship for me at this point knowing that in all probability I will be dead or very near dead in a few years time. That is selfish...to me at least.  Well I am tired...and I will give a better account when I get some sleep.  I will post more of my time with Danielle tomorrow. Again it was really fun and I enjoyed talking to her. We also talked a little bit about politics...and she is also a Registered Republican...after I get done with her she will be most decidedly opposing the Republicans because it will be easy to show what all is wrong there...I may even wrestle her away from the Dems as well because all in all they aren\'t much better than the republicans...  Fight the power...learn the jargon...fight them on their level because you can never elevate them to our levels...its time to get dirty and take it to the street!  '),(1077,1234212,'07,April,2003','2003-04-07 00:00:00','   Woohoooo!!! Never a dull moment over here. I am getting out today. So never mind that I thought I would be here another day or so...Dr. Weinberg is kicken me out....that is so good cause I gots a lot to do. I have to run to the bank, run to the computer store to pick up a HSF for my P4 at home...and get some heat sink gew...I will head over to Comp-USA to get that stuff...they have gotten quite a few things in that are of decent quality.   My computer runs ok...but it does run a little warm considering that the processor is hella overclocked right now...I mean the sucker is running at what almost a gigahertz over stock speed...but the temps have risen only a little bit...normally the processor was running around 45C in this mini boxes...now it is running up to 52C...also the voltage is only slightly up since I moved from my old Shuttle to the new system that I got...actually the voltage is down from the Shuttle (it was 1.62V in that) and now is running at 1.575V in the current system. That makes me happy...and now I just need to shed a little heat off of the processor...so maybe the fans can spin a little quieter...and the box run a little cooler.  Over all the system is currently running at ballz out blazing speeds...the only thing that I really need to do is change out to SATA hard drives and maybe a new Video Card...and that thing will be screaming...I want to get one of the SATA Raptor 10,000RPM 72GB hard drives as my primary drive...and then get a 250GB plus SATA drive for my backup and data drive...then I will need to get a few external HD cases to hook into the USB ports so I can keep the 120gigs of space (plus another 40gig drive that has no home)...My goal is to have around 500gigs of space on my box available at anytime...cause I have a friend that has about a thousand CDs that I want to rip over to MP3...she has already given me the approval to do that...plus if I get a new laptop and were to bring that mega-box up here with 500+ gigs of space I could rip all those CD\'s two or three at a time and get through the CDs in just a few days...  Oh yeah in my main box I have a TDK Indio 8X DVD+-RW Drive...and a TDK 48X CD-RW Burner (when I get those SATA drives in there I will be able to put both drives on independant PATA channels allowing for independant burning and ripping without slowdowns...heheheh). The DVD+-RW Drive also burns CD at 40X...so it is a bad ass burner to be sure!  Well I have some food here and I need to get my stank ass cleaned up so I can jet out....  Love to all that sent love to me while I was away....I will be back home later this evening. The only thing that I need to do really is just get a few things in order and all will be will.  Peace out...fight the power. Listen to Democracy NOW! With Amy Goodman...link in my political links section.  '),(1078,1234212,'07,April,2003','2003-04-07 00:00:00','   Well this morning has aleady been busy. First I woke up to the lovely Missy and her pipes of perfection....well by that I mean that she had all my treatments for me. That went well.  Then I went down to get my PFT\'s done. Well the first two were predictably bad but the third one was very good...but in doing that I heard something pop in my lung and now it kinda hurts over there...oh well. One good number for the road at least.  Yesterday I went down to the Vascular Lab where they do Angioplasties and such in this hospital (I think the Department is ENG Labs). Well they fiddled with my port and got it to work without much in the way of pain...they didn\'t need to dig into my chest (thank goodness)...I was really worried that they would have to replace the damn thing. However the Queen of Ports (related to the Queen of Hearts as she works on hearts too ;-) got in there and did something and got me all fixed up. She basically jiggled it around in the port and whatever clot essentially broke free and went whizzing down the tube into me...that felt weird. Then they ended up replacing the port needle...or reaccessed it. That was good...and they now it works just peachy keen.  I love being in the hospital...not because I like being here but because I just like being able to have what I need when I need it. They come in and make me do the treatments and then beat me...it just is so easy. When I am at home it is hard to keep up with all the things that I need to do because there is just so much of it.   Well I have a CT scan that they are going to do later on to take a look and see what my lungs look like and see if they can figure out why my heart is all goofy. I am just thinking that basically it is because I have all these anti-biotics in me and they are essentially hurting my heart. All the meds go directly into the heart through the Port. So I would imagine after a bit the heart would get a little tired of the insult of cold meds hitting it and with some of the meds (like Azactam and Zosyn, and others) that basically I am allergic too I would imagine that the heart is just not happy with those things going into it. It is becoming damaged to a degree.  Some of the drugs that I take basically make me numb and a little goofy. So I think that may be almost retarding my heart rythm a little...and that may be why my heart feels heavy.  So....I just don\'t know what is up.  I am going to try and get pics of the people here today because I may be leaving tomorrow with any luck...I say tomorrow because they still have to do a CT and get that looked at by a pro...then I have to wait for the Doc to look at it and see what he thinks.   I just love the people up here. I really enjoy being around these people...because they are nice and just do what they are supposed to. Plus a lot of them I have grown fond of because I see them a lot.  Well I am going to get a little nap then I am going to be doing a little paper work. I have a lot to write today and some charts and graphs...hopefully I have all the software on here.   Oh yeah and there is a female here that is a hottie...woah momma...she was here the last time I was here...and she is even hotter this time. Perhaps it is because I am getting on up there in the celebacy level....not chosen mind you...forced because of circumstances (like not really having gotten out to meet the locals and perhaps just not really wanting another distraction in my life)...but now the urges that we all have are coming back (perhaps because I am feeling a little better) but regardless I am feeling a little better.  Although I just stood up and got itchy...eek!!!  Well lemme get some benedryl out of these people. Then I can get things on the road...  Talk to you later. Remember yesterdays posts powered by Sleep Deprevation and lunacy...todays may be even better as I go over what all I said.  Oh lord I have a feeling today may be an interesting day....and oh yeah I have decided that somehow over the next month or so once I get this thing under way I am going to have to get a real laptop...cause this thing is just ghetto...really....It is four years old and beaten all to hell. I am sure that many parts are worn out...and parts of my screen are not as bright as they should be....and somewhat pixelated. Yeah I think I am going to be shopping for a new notebook....I want to get a Centrino laptop with a Radeon 9600 and some extras....that is what I want...and somehow I will get that hooked up.  Peace...remember change the world...make bold step. resist the powers that be.  '),(1079,1234212,'07,April,2003','2003-04-07 00:00:00','   Well today was just a long day. I really didn\'t get all that much sleep overall. I maybe have had a combined 4 hours of sleep over the past 60 hours or so...because I really didn\'t sleep at all on Sunday night...my chest was hurting pretty bad and that was the beginning of this all. Monday I didn\'t sleep but maybe an hour throughout the day...just enough for quick nap.  I of course didn\'t sleep last night...well because I wasn\'t in a place that I could really sleep...now I have a room and I am pretty tired so I am most sure to be dying here pretty soon....actually I am so close to falling asleep right here at the keyboard....jessh!  Well maybe tomorrow update will be a little bit more up to par...I mean lemme just say that these posts today were pretty far out there...not because I don\'t think that way...indeed I do...but normally I have a little better filter on my brain. Today that filter was forceably removed by drugs and sleep deprevation.  goodnight all and I will talk to you tommorrow what I am a little bit better off....  Peace love and change the world....fight the power!  '),(1080,1234212,'07,April,2003','2003-04-07 00:00:00','   Here is something that caught my wrath in the last few minutes. The hacker of Al-Jazeera basically got a slap on the wrist for his stunt.   CNN Story on Al-Jazeera Hacker   Now riddle me this....if I had taken down CNN and redirected all their traffic to a porn site...or worse just took it off the air for a few days...what would happen to me? Would I get a slap on the wrist? Or would the corporate lawyers at AOL-TimeWarner grind my frail body into the ground after the judge sentenced me to at least a year in jail with possibly an extension for the Patriot Act of up to ten years.  So I imagine some of you are wondering what this is all about.  I don\'t know yet...but I feel a fucking revolution coming on...even if I have to go out and make it happen.  Revolt, revolve, re-constitute, change it for the better...  Oh yeah and I fucking hate being in this room right now...I just want to get the fuck out of here.  '),(1081,1234212,'07,April,2003','2003-04-07 00:00:00','   It is all fucked up right now. They have me in a room with another person and I can\'t even be nice.   I want to be in a private fucking room right the fuck now. CF patients are almost always kept away from the other patients partly because we tend to be foul tempered...but mostly because we are already walking a thin fucking line if we are in the hospital in the first place. I am on a razor thin line...my bugs are already getting the best of the doctors and  I know that I can\'t handle another infection of anything right now.  My port is most definitely stopped up. They will have to surgically remove it. Then they will have to surgically put a new one in. So I have two major things right off the bat that I have to deal with. I bet they won\'t get to it tomorrow because if they were smart and planning anything they would have me fasting now...instead it is just all fucked up.  I dunno what I am gonna do. I am pissed. I am tired. I am sick of being sick...and now I am in danger...this isn\'t right!  '),(1082,1234212,'07,April,2003','2003-04-07 00:00:00','   Hey I had a post here...where did it go...damnit!!!!  Well I am down at the hospital now...I had to wait almost four hours before I could get into the emergency room...I hate having to come in through here.   I hate having to type all this again.  The ER really is broken in most places...here is a fine example. Really did I need to come into the ER...no...but there should be a better way for me to get into the hospital in the middle of the night.   The ER is for emergencies...but here it is used as a primary care facility. That is what is wrong here with our system. We have people that should be going to doctors...but instead they can\'t because they have no insurance...instead they have to fall onto the ER...so it is broken. The time has long since passed to move us over into a national health care system...it is for the good of all to have one.  The only people who benefit from this mess we call health care right now are the insurance companies and their managers...they are the only one making any money on it all...the doctors aren\'t...why because they have to pay insurance companies to protect them from malpratice lawsuits...while on the other end you have insurance companies forcing doctors to not treat with the most effective treatments causing harm to people...and what do those patients do...they turn around and sue...  The only people making any money are the insurance companies...and that is just as clear as all the blood on their hands.  I call for everyone to rise up and defeat this monster...it is so easy...it really is right there for us to just do. Instead we let Democrats and Republicans divide us on the issue as to which will be better...instead we should be giving the system over to the doctors and asking them what would be best. They would know...they are doctors...their primary goal is to treat patients.  The whole system is a mess...we can fix it. Republicans won\'t fix it because they are making money off of it being broken, the Democrats won\'t fix it because they use it to strike fear into people and keep us coming back to them for a fix...when indeed most of them are just nothing more than pushers of a quick fix.  I fear that no one has the balls to stand up and say this simple thing. We need to fix it. We need to Socialize it! We need to demand as taxpayers and as humans that Medical Care is a Human Right!  Fight the power...change the world...start a movement...  Oh yeah and I love having this wireless internet on my phone :-)  I get 315kbps over it...well above ISDN...about half that of a medium DSL line...not bad at all :-)  Fight fight fight fight...make the world better. Only we can do it!  '),(1083,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok so I am off to the hospital...I am sure I will update later on this evening...  Great...chest pain, lung bleeds, and now a blocked port!  As always I will be checking my email...so send me love...  Off into the breach once more.  '),(1084,1234212,'07,April,2003','2003-04-07 00:00:00','   Hey all...well I should be at work right now, however I am not. After I went to bed early this morning and while I was asleep my port decided that it wanted to go ahead and clog up. Oh the joy!  Right now I am experiencing much discomfort from it. It feels like I have a large weight on my heart...my heart feels like it is being squeezed into a little box and every heartbeat is beating against the box. Breathing is hard...and very labored. If feels like I have something in the left side of my chest over there that is sitting on the lung and heart...it feels almost fluid like because it doesn\'t compress at all...instead it just exerts much pressure.  The nurse will be here by 4PM to try and get the line unblocked...I hope she can do it because if she can\'t get it opened up I will have to go to the hospital and have this port removed and another put in...oh joy!  I just feel like shit...my kidneys hurt, my liver is hard, it feels like I am breathing with a large weight on the one side of my chest...something is flat our wrong folks...very wrong.  '),(1085,1234212,'07,April,2003','2003-04-07 00:00:00','   oh and I fixed the spelling before I sent it....  '),(1086,1234212,'07,April,2003','2003-04-07 00:00:00','   A letter that I have sent to my Senators from PA.  ------------------------------------------------ With the recent bruhaha over a Prescription Drug Program being offered by Medicare I have found myself in a dilehma of sorts.   I do think that health care for all should be enacted immediatly. A national health care initiative is the only solution to the current crisis, and to be sure it is a crises when we have 40-60 million Americans without any type of Health Care coverage.   The plain truth is that while they can get health care it usually ends up on the taxpayers bill regardless and with the recent push to \"privatize\" medicare and prescription drug programs I feel that this will definitely bankrupt the system and leave millions more with less than competative coverage or no coverage at all.  Why am I concerned about this issue? Because I benefit directly from it. I am 26 year old with Cystic Fibrosis and I depend on Medicare and state programs to cover my medical care. I am not alone as most CF patients are in a similar situation. As the private system now stands if I were to even be covered by any insurer I would have an exclusionary period of a minimum of two to a maximum of five years before any private insurer would even consider picking up the health care cost.  If you are not familiar with CF it is a very expensive condition to treat. In this past year alone I have racked up around 300,000 in hospitalization and drug costs. Under the current system I am covered and could want for nothing. Under the new system that is being pushed by your coleagues in the Senate I fear that I would be left without any coverage as you want to push people off of the public system onto a private system. If this goes through I, like many others, will end up with no coverage because of the exclusionary periods.   I can\'t wait for up to five years before any benefits would kick in from a private insurer.  As a Christian you should be fighting for a national health care system period.   It plainly states in the Bible that we are to take care of the ederly and infirmed. There is no question on this. If you are opposed to these ideals than you may need to re-think your stance on being a Christian...as no self respecting Christion would even dare to suggest that we let out old and sick just die off instead of extending a helping hand.  From the initial sneek peek of this program there would be large gaps in coverage and many people would simply not be able to afford health care. Which will make them have to choose between their medication or food or a roof over their heads. That is simply un-Christianlike.   So I call for you to make a moral stand on the principles that you have publicly spoken. I call for you to propose a sweeping national health care initiative that will cover every American...regardless of their socio-economic status.  With CF I have had to make hard choices. One of those choices has to forgo trying to get a well paying job to help and pay taxes. If I were to get a high paying job, or even a decent paying job, the benefits that I depend on simply vanish because I am working. I would love to work and add to the common good, however under the current system that is not possible.   Again I call for you to take a moral stand against this prescription drug benefit plan and instead call for a national health care plan that covers every person in the United States.  It is the Christian thing to do, it is the Patriotic thing to do. It is the duty of a moral and just society to provide health care to all. ------------------------------------------------  I challenge all of you out there to write similar emails and letters to your representatives. It is the only way to make your views known!  '),(1087,1234212,'07,April,2003','2003-04-07 00:00:00','   Today was an interesting day I will tell you. Actually it wasn\'t a very interesting day...I lied.  I woke up, did some drugs (not any fun ones), drove to work (that was OK cause I rocked Fatboy Slim all the way), got some Mcdees lovin (mmm steak bagel), got to the mall, opened the kiosk, and started selling almost immediately...today it was pretty busy...we did about 350 in sales (which is better than almost all the Sundays I have worked combined...but people are out buying for Christmas now), my boss was there (and I love her dearly but she makes me insane with her nervousness), I did more drugs, I didn\'t get to eat lunch because it was busy, I did more drugs, I stood around for about a total of an hour (not actually doing anything), talked to the Jews (remember I work next to Israeli\'s).  I am still trying to figure the one girl out...her name is Galit (I remember her name because it is so close to Galil an Israeli weapons manufacturer). I would definitely bang her...I mean she is certainly hot...but I don\'t know how we would communicate outside of random grunts and moans...her English isn\'t the best...actually the Indians (India Indians) speak better English...no surprise there as that is India\'s primary business language...  The last couple of days I have felt particularly sexy...I dunno what it is...probably the steroids...but I feel sexy. I am still self conscious but not nearly as bad as I usually am....its almost as if a switch was flipped and I am no longer a little boy...strange. So now all I need to do is just get some game and I am sure that I could have most any woman that I wanted...  I am finally comfortable (or reasonably so) in my skin...I am skinny...which most people wanna be (I would love to gain weight), I know I am smart...not because I am arogant but because I am indeed smart (at least I have something going for me), and I don\'t think I am dog shit ugly soooo...hey I am not in a bad shape. I think women are noticing me a little more. There are several A&F females that stare at me like they see me as a piece of meat...hey thats kewl...and then I have Galit...I know she is attracted to me it is painfully obvious...but I don\'t know about how to make a move.  Overall I am starting to get happy in my life...well not so much happy but content that I am me and I can\'t change it.  The next thing that I have to do in my life outside of getting ready for this project with Jesse is that I have to do some Biblical, Tora, Quran studying...I want to talk to the three religions that are in a horn lock right now and try to make some sense of it all.   I have so many views that really do come from a place of desperation as a believer in a higher power. I don\'t expect anyone to convert to my views...I just want them to analyze their own views and pratice what they teach/preach/spew...It is time for the moderates to come out and take back religion. The big three have been stolen from people by radical fundamentalists. Christianity has the Baptists and many other wacko forms, Judaism has been hijacked for years by Zionists, and Islam has been taken by force from the far right in their factions to spew hate towards all others (even themselves).  Why do I say this and why do I believe it to be true? Look at the recent comments by Pat Robertson on the Christian side...he was asking for people to pray that some of our most respected Supreme Court Judges be struck down by illness because they support a womans right to choose...that is absolutely wrong. Look at Judge Moore (well former Judge)...he was bringing religion into the courtroom asking for people to pass judgement on people from a religious stance (not unlike many of those supposedly evil nations that have Muslim governments and law)...  You have Sheron...he is a fucking lunatic...and all of the Zionists are just evil...and yes they are evil. They are literally killing their BROTHERS AND SISTERS!!!! The Palestinians are decendants of the Cainanites...which come from Cain...of Cain and Able fame (brothers...Able went on to be a founding Jew)...and recent archeological finds show that Cainanites and Early JEWS lived together...the Jews left the cities for the countryside to get back to a simpler life a couple of thousand years ago...mitochondrial DNA proves that they are related...every damn one of them! They are killing their brothers and sisters. That is just plain wrong.  You have the Mullahs and their flocks in Islam...that spew nothing even remotely masked as fear and hatred towards others. It is all rather rediculous.  All of the tenants of those religions are exactly the same...(the Muslims and Jews actually are better at taking care of each other and stay closer to some of the tenants of their religions as where the Christians have pretty much all but left taking care of their fellow man behind as a novel Idea)...  What are the Tenants? Love one another as you would love yourself, take care of your infirmed and elderly, and show love towards the other religions in the hopes that you would bring each other over to the other side...oh yeah and that God loves everyone.  Do we hear that anymore...well we hear God Loves us...but then we hear right after that...God hates the others and only if you prescribe to one religion will you be let into heaven...  I just don\'t know...I want to put together a presentation that I can give to anyone who is a religious zealot and at least plant the seed that what they are doing is actually against their religion...and that they need to return to the basics if they ever want to move forward.  I have a lofty goal of bringing peace...but in a more realistic reach I would just want them to critically analyze where they are at...and have them stand up to the bullies that have pushed these religions to polar opposites of what they are really about. That is all.  Well I have to get up in the morning and I am tired...I have a lot of reading and writing to do...  Love each other...be tolerant...and drop the hate, forgive each other, drop the hate.  '),(1088,1234212,'07,April,2003','2003-04-07 00:00:00','   Another long day...I worked today...my boss needed for me to come in and help out so I did so willingly. Hey I was getting paid....and cash is good. Especially seeing that I need some to buy things like gifts for Christmas and shit.  I talked to my sis a long time tonight...I love talking to her...sure we don\'t agree on many things...but we are brother/sister so we get along pretty well overall. Actually we get a long a lot better than most of the brother/sister pairs out there. She is really an amazing woman...she is about to graduate from Vet School...and she will have completed the entire body of learning early (although she has to wait to graduate with her class...in the meantime she will be doing externships)...not only did she really get finished early she was also accepted to MSU\'s College of Veterinary Medicine on the first shot (it is not uncommon for people to apply two or three time before they get accepted)...so as that she started into vet school pretty early.  She really makes me proud...and I love my sister a lot.  Well I am tired and I am not feeling that red hot...I guess I am just worn down. My kidneys and liver hurt a lot...but I would imagine that is because I am on FIVE seperate Anti-Biotics and a coarse of steroids...not to mention the other things that I have to take each day...so I know why they hurt...I just wish they would stop hurting...probably the worst part about all of this is that my eyes are watering...like I am having some sort of alergic reaction....but I can\'t be because I am on the Steroids and high doses of Benedryl...so it can\'t be....it sucks because it makes my eyes really sensitive...I am sitting here squinting at the screen...because it is so damn bright!!!!  Well I am going to get some sleep....got all the computers up and running which is a good thing. I also got Norton Systemworks 2004 Pro today and got it all installed on my machine...crispy indeed.  Goodnight all :-)  '),(1089,1234212,'07,April,2003','2003-04-07 00:00:00','   Finally hitting the sac...benadryn and Restoril seems to work...ahhh sleeo  '),(1090,1234212,'07,April,2003','2003-04-07 00:00:00','   Today was an extremely busy day...we did 1299.81 in sales (not including tax)...so that wasn\'t to shabby at all. In not being shabby though you end up pretty much getting to the point of strangling the next innocent passerby.  It was a long hard slog at work today...I\'ll just put it at that.  On my way home (I was taking my time as the roads are starting to slick up...and I just needed to release some tension) I get a call from my grandmother. Apparently I am not old enough to be able to handle situations or something because god forbid I am late in getting home...or there is traffic...there must have been an accident...oh god he\'s dead in a ditch...my cancer markers are up...blah blah blah...  Well she called to check up on me on the way home because it was like 10:50...mind you it takes an hour to get out to where I work so I wasn\'t even late...jeesh.  I get here and I am barely in the door and she has crawled right up my ass with rapid fire question and answers session...I just answered simply and plainly as I am fucking tired...tired!  I get my food heated up...and she has her nose in it asking what it was...but she looked at it...what does it look like...hmm chicken...and it smells like chicken...wow it must be chicken. The other half was in the refigerator in a styrofoam box...and jee I guess it could be something I brought home...I dunno...she no longer has the capacity to do even simple problem solving...or maybe she is just dumb about that kind of thing.  I am tired and I want to relax...but instead she is crawling up my ass for all sorts of stories or whatever the hell she expects...  There was a message on the phone for me from the nurse...nothing big. I already knew what it said...so it wasn\'t a big deal...while I am just starting to eat she starts handing me the phone to get the message...so I stop what I am doing and I go out and get the cordless and get the friggin message...lo and behold it was nothing important.  I sit down again to try and eat...she tells me about Springdale not winning in football tonight...ok I don\'t care. It doesn\'t interest me...so when I just shrug she gets huffy and starts in on me...about she should just be quiet and this and that...yes you should I have been working in a loud place all day...I have a headache...and I am sick. It doesn\'t take a friggin genious to figure out that possibly I would want a few minutes without being harrased by anyone...including and especially her!  Then she gets up and starts tromping out saying something and I just wring my hands and shake...because if I didn\'t do that I would have punched her...I would have...she took offense to that...and started screeming at me...at that point I know that I probably shouldn\'t have wrung my hand and shook like that...but what the fuck...leave me in peace for just three minutes...please. She screemed and stomped off...I still don\'t give a flying fuck. I hope she sleeps poorly tonight realizing that she was acting like a three year old going \"hey hey ...wanna see what I can do...hey hey...what about a bunny\"...  What do you do? I mean I just don\'t know...I can\'t say anything because it will be wrong...I can\'t not say anything and just be quiet and that is wrong. She pulled that old lady crap yesterday when my mom was down for thanksgiving.  Moms back is screwed up...she was thinking out loud and was saying to the dog...\"I wish you would walk on my back Rufus\"...GM replies blythly with\"I can do it\"...mom just said to her \"no that would probably not be good...I could just see you falling off and breaking a hip...imagine how I would explain that to the doctors...naaa better not mom\"  I got a laugh out of it...my mom got a laugh out of it...it wasn\'t meant to offend...it was funny. GM took offense to it. Then made it plainly clear that she did...and attempted to belittle my mom.  Lately things like this have been cropping up more and more with her. It is getting to the point where she has reverted in many ways to a child like state.   At thanksgiving dinner at our friends..close friends that know how she is...she didn\'t have much nice to say to anyone...they shrug it off as her being \"Betty Lou\"...and that isn\'t right. I was always taught by my mother...which my mother was taught by her mother...that if you don\'t have anything nice to say...don\'t say anything. Simple...plain...to the point. However GM has forgotten that lesson that she ultimately bestowed on me.  Part of the problem right now is her cancer markers have jumped up again yet they can\'t seem to find any cancer in her. Now cancer markers aren\'t the be-all and end-all of cancer detection...they are like pointers and other things can actually trigger the same rise in marker levels...in other words the cancer markers aren\'t the most reliable test...but a good place to start.  Now she has already been through cancer twice now. Bad times indeed...but she made it...and frankly she didn\'t suffer much from it...she has a little pain now and again but instead of just getting something for it she would rather complaign and make herself out to be the person who needs pity or something.  This time around...it could possibly be bone cancer...and you can\'t do anything about that...simple. I know she is afraid...I know she doesn\'t want to die...but shit come on...there are somethings that you just can\'t control. Cancer is one of the millions of things that you can\'t control.  I feel bad for her...but the pity well dried up long ago. I will help her in anyway possible...I will be there as I have always been. I will do everything that is expected of me plus some...but don\'t demand pity. I am not in a giving mood...and pity is for the weak.  Face up to your disease with honor and pride. Own it...make it yours. Do not try and pass it off to others so they can share the load...  Just shut up and deal. Ok..simple!  I am sure tomorrow will be an interesting day as my mom will be up...and I will hear the same line \"just be patient with her\"...well I am being that...really patient...with the meds that I am on...it is nearly impossible to be patient because the steroids make me into a raving lunatic...but I have been handling her with kid gloves. I have been as careful as I possibly can be so as not to offend her. Aparently I need to work on it...because me just being quiet and out of the way is wrong...as usual.  I just need a break. She wonders why I don\'t want her to come to the hospital when I am in...because when she does it is all about her being there...and god forbid anyone ask her how she is doing...because you know she is a cancer victim...not a survivor...a victim.  Walk one day in my shoes there GM...you would gladly go back to your own pair...because at least you have lived a life...life before cancer...I have never had that chance. I have always had CF...and I always will. CF will kill me...and probably more likely to kill me than cancer will get her.   I love her...I really do...but she is just irritating the hell out of me...and there is nothing that I can do right...so it makes it worse.  Also I did warn her...I told her that the meds make me nuts...please be considerate enough to understand that incesant questioning and crawling up my ass about a letter on the table puts me that much closer to snapping.  Boy I love my \"zero to rage in four hours\" pills...also known as prednisone.  Goodnight all.  '),(1091,1234212,'07,April,2003','2003-04-07 00:00:00','   I am tired and I am going to bed....read my posts. I am a genious!  '),(1092,1234212,'07,April,2003','2003-04-07 00:00:00','   Ummm what was that previous post? I sure as hell don\'t know. I was punchy this morning...but I don\'t think I was that bad off....oh dear baby jesus fetus....  I wanna say this...I love C-Span...I mean it really is a national treasure that unfortunately most people don\'t know about...or really know what it is. Find it on your cable dial...and watch it sometime. You will find that it really is the only \"free\" speech out there...it is unscripted...no news anchors...instead you get a feed from the House or the Senate...or you get a taping of a national book club meeting...no topic to dull for them to cover...their motto should be \"More information than we know what to do with\"...or...\"Need a nap...don\'t watch us...cause we\'re interesting\".  Ok well its not like that, but the channels cover so much.  Tonight I watched on C-Span1 a debate that was held on Oct. 25, 2003 at Regent University...the topic \"Constitutional Judicial Activism\".  The panelists ranged from well qualified to downright insipid. Here was the lineup.   Nadine Strossen - ALCU President Barry Lynn - Americans United for the Seperation of Church & State. Executive Director. Alan Dershowitz - Havard Law Proffessor Jay Sekulow - American Center for Law & Justice Chief Counsil David Limbaugh - Writer, pundit, retard Ann Coulter - Writer (HA!), pundit (Double HA!), whore, retard, skeletor, slut, bitch, cuntrag, oh what else...uneducated talking head.  Ok...so now that we have introduced out cast of characters (some woman from CourtTV was the moderator...no it wasn\'t that other cunt that is on Larry King Live all the time...she is a bitch of another caliber) let us get down to brass tax here.  Basically the debate was over whether or not the Supreme Court has been meddling in politics and taking political stances and not necesarilly legal views. Well the answer to this is really loaded...and it all depends on which side of the fence you sit on with a particular issue.   Take for instance abortion. The constitution (and I have to give the conservatives a little credit on this one) says nothing about it. It doesn\'t protect nor deny it...it just never mentions it...probably because it was before it really was an issue.  The argument on the side of the conservative now (it has changed so many times from murder, to morally wrong, to plain out religious hatred) is that the Judges should never have ruled on Roe v Wade in the first place because the court (which I happen to agree with) is only supposed to rule or even hear constitutional arguments...in other words it is only there for the constitution and nothing else. I agree with that.  However in the case of Roe v. Wade it was a constitutional issue as it was framed in a lawsuit because it specifically challenged a state law that limited private choice and freedom. To me that is simple. The court did need to weigh in on the issue.   By the way...if you want to know. I am PRO CHOICE!!!! It is in line with my beliefs very easily...and here is why. We all have the choice to do good or bad. I personally wouldn\'t have an abortion if I was a female...but I am not so I don\'t have to make the choice...but that does not mean that I should be able to impose my morals or views on someone as long as their activity is not directly harming my person or property. I can think that it is disgusting and horrible...but that doesn\'t mean that I can puch that view on someone else. I would just hope that they would see it as that. In my eyes...do whatever you want just don\'t impose it on me.  I took a few notes of this debate...it really was sad to see that the people in that place were actually applauding the retards on the right. I call them that because they are indeed morally and socially retarded. It would seem that they would love to live under a system that really doesn\'t give them even the choice of what they want for dinner...because to argue against the seperation of church and state (which they all were arguing that they should be combined!!!!) is really to argue for a system that has been proven time and time again to fail. So to argue this point...or from this side is indeed retarded. Like sticking your hand back into the fire five minutes after you stuck it in before and got burned.  David Limbaugh...who the fuck is he? I dunno...but I will say this. He is about as smart as horseshit in your refrigerator. He actually argued that the congress should limit the Supreme Court under Article 3 of the constitution. I looked and I sure couldn\'t find any limiting factors there. Actually it affirms that the court is pretty much above the fray of political shit that floats in our bowl of democracy.  He argues that (what in reality has been a mostly conservative president appointed panel) the court is very liberal. They are out of hand and need to be replaced.  I dunno...but I think this guy needs to have some sense beaten into him.  Jay Sekulow. He is a piece of work I tell ya. Imagine what a failed televangilist looks like. The hair, plastic skin look, tanned, with a bad almost pimp suit on. Umm yeah. That is this dude. He stated that President Bush is \"the comander and chief of the constitution\". What? I literally shook my head at that. I really don\'t know what that means. It was probably one of the dumbest things I have heard...almost like \"if it hadn\'t have been for my horse I wouldn\'t have made it through college that year.\" A famous Louis Black joke...about anuerisms and how they are caused...  This guy is a walking anuerism. I can\'t think of one thing intelligent coming out of this dude. Really he just didn\'t make sense. He would say something one way and then turn it around and completely disprove his prior statement only a few minutes later.   Ann Coulter is the argument for abortion...ow my chest got a little pain on that one...I guess I do have some sort of soul...and I ripped a little of it out for that one.   I have seen her on CNN, Fox, ABC, CBS, NBC, now C-Span...she is a living, breathing, barely intelligible, cave woman. What she argues for...the positions that she takes would literally take women back to being nothing more that walking wombs...they would have no rights...hell under her ideals she wouldn\'t even exist. I can not honestly understand how she can be such a walking talking conflict...  Not only does she flat out lie...Read Al Frankens book and he actually has the good ones on her...but she is also not that intelligent. Now...I know I call a lot of people stupid...not always because I think they are genuinely stupid...but becasue they act that way. I think she is the real deal. I am sure that my intelligence would be sucked out if I were ever to be in the near presence of this woman. She really needs to get back to what she must do well and that is suck...there is no way that she could get published unless she was blowing people...really she has no platform...she has no intelligible ideas...I mean the woman is just useless. The only thing that the Republicans need her for is a pretty face (and hell she looks like skeletor...even more so than Maria Shriver....Ahhhnolds wife). She is the Atypical Republican woman...or at least what they would love to have all over the place. Someone who really isn\'t informed or intelligent and will seed to the man whenever the man needs to take \"charge\"...she is a Barbie...nice and pretty to play with...but you put Barbie away whenever you need to actually take care of business.  I hope she reads this...or this gets on to her. I think you are quite possibly the most useless person on the face of this earth...hell at least even our current president is good at something...he is good at making war on everything and everyone. You on the other hand probably are good at consuming air. I don\'t know. I would spit on your shoes. After that I would simply ignore you...because for me to give you anymore acknowledgment of a person would be a waste of time and energy. I don\'t have time for stupid people.   Alan Dershowitz...he is an interesting bird...I don\'t actually agree with his positions on Israel...but hey we can dissagree on some things yet be united on others...  He put it very well and this argument can hold up to any litmus test of history. Try and prove it wrong.   Religion thrives only when the government keeps its hands out of it. Religion is endangered only in places where a state sponsored religion is promoted.   The final most damning one...the one that really is good.   If you need to ask the government for help in your religion (monetary or promotional) then you don\'t have faith in your own religion.  It states that if your religion is so whacko and out there then probably you are wrong (well at least according to market economics...which ironically is what most Republicans ascribe too...their religion is to whacky so people don\'t go and give to the church...hahaha...ironic)  Again I don\'t always agree with him. But when he is right...he is really right.  Barry Lynn is a preacher. He sees the need for church and state to be seperate. Again because once you get government involved in religion in any way...then you will be promoting a state religion. He goes as far as I do and says that \"under god\" should be struck from the pledge...\"in god we trust\" from our money...the only place where he does not promote the stripping of references to \"god\" is from the supreme court...and not because he thinks it should be there...but because the court has already ruled on it...saying basically that in the introduction for the justices and the ceremony that goes along with it are basically untouchable.   I agree...it was and has always been a part of the ceremony...blah blah blah...god save the country, blah blah blah...it really is a holdover from the British \"god save the queen\"....  I have the strong belief that he holds a similar view as I do on abortion. You cannot be as seemingly enlightened as he is without having come to a similar conclusion on abortion that I have. I feel I am somewhat enlightened on things...that is why I have that view.  Being a preacher gives him street cred...and he calls for Christians to defend the constitution because if they/we don\'t then sooner or later our religion will be banned...it just happens that way. History shows us that...our nation was born out of that.  Nadine Strossen is the President of The ACLU...while I am not always in agreement with them...they really aren\'t politically scewed...they only want to fight for the constitution...which makes for some strange bedfellows.   She brought up a few myths...one was that Christians here are oppresed...she makes the point that in reality you could never be elected as president without being a christian...just wouldn\'t happen and may never happen.   The other is that in reality some of the most devout christians...the really good godly ones...are actually for the seperation of church and state. The believe that religion is private and should have no mettling in it from the government...  Hell the ACLU has even argued a case for Jerry Fallwell...and are currently on a case in VA for a city councelman that was giving out fliers in his campaign that essentially argued for the merging of Church and State...something about the election laws there prevent that...but the ACLU has stepped in on his behalf because they view his right to free speech and free religion are paramount over what bad he might be saying.  Overall the Liberals...or the liberal side of the panel really made its case. Not just because I am liberal or want them to win. They made it because they fired everything back at the other side...and to an open minded person made them look like alarmists.  The conservatives were just plain pitiful. They play the fear card all the time. They attempt to frame this entire issue as a moral issue.   I am tired of the right wingers out there arguing for church and state to be together and faith based initiatives...when you get into that you know that your government is no longer working for the people...all of them. Instead it has become a quasi-taliban like government that places religious ideals over all else...  Now would that be bad if the people who were in power were actually following what the \"good\" book says? I dunno...but I know that if they actually lived up to the ideals described in the Bible...we would really see something quite different from what we see today.  These aren\'t hard concepts at all people. Really they aren\'t.  '),(1093,1234212,'07,April,2003','2003-04-07 00:00:00','   oh and all my crap is almost over...20thc Dec. Babay!  U have been hitting up the net for music...  This is what I found...and I will rate them.  Britney Sprears In The Zone...  I am sorry that I like it...but the engineers and sound man get mad props for that...and She acttually poer  (not all singe Well) \"It is time to back out country...its time  '),(1094,1234212,'07,April,2003','2003-04-07 00:00:00','   I get off the Anti-Biotics tomorrow...only because I have been on them for five weeks now...and frankly we are crossing into kidney/liver damage territory if I keep on them any longer.  It will be good to be off them for a bit...cause my body is just completely wrecked. I am so tired and completely wrecked emotionally and physically.  I am going to enjoy my body righting itself...well I won\'t but at least I can try and get back to normal for a bit.  I go in this next week and hopefully things will be well...but I doubt it because I really do feel like shit most of the time.  Today I had work...I worked about 8 hours...and I feel totally wiped out. It feels like someone has been beating on me with a bat...all over.  I have this nasty rash on my back...it is itchy and has been putting out pustules and crap for the past two weeks. They look like blackheads of zits...but they aren\'t...it is a reaction to the high levels of drugs in my body...most of the itchy bumps dont have anything in them...but are red and irritated...it is certainly weird...the nurse caught it today and she thinks it is an alergic reaction to something (probably the IV Azactam...as it is a derrivative of a pennicylin)...so we will see if it clears a little after a while.  I dunno folks...I am just worn out.  Well I am going to get a few more things done and then head to bed...  Oh yeah HAPPY THANKSGIVING...remember to be thankful for even the most inate things. I sure am...for me I am thankful for those few days that I wake up and actually feel half way decent.  '),(1095,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I have found out so far that the generic benadryl is better than the name brand...strange huh...I guess it has less fillers in it that my body would need to process.  So I am gald I got them instead of splurging on full Benadryl....the brand name add some strange stuff that is different from generic.  So I will stick with that...oh yeah I am getting the video recoded for Unaswered Questions and there is a new DVD that I wanna get a hold of because it promises to be a good watch.  Man I am so disapointed in what we have become. I love America...I think that we have the potential to be what we once were...thought well of and somewhat respected across the globe...as it stands now we arent...and we need that back.  Mike  '),(1096,1234212,'07,April,2003','2003-04-07 00:00:00','   Well today I did finally get some sleep...well yesterday.  It wasn\'t good sleep. I hate that.  Tonight I am going to say it loud and clear and fully support Howard Dean. I have been a supporter of his for a little bit now, however after watching the DeanTV and some of the talks and \"propoganda\" (I call it that because propoganda really is information...and not always a bed thing)...  The First one that I tackled watching was a 23 minute Speech in front of the AAPI Town Hall Meeting. He nailed everything. Really he did. He hit on all the weak spots and then went on to tackle the morals of those spots. He speech is rough, his message is clear to those who are even moderately informed...and his remarks on Race Issues were absolutely spot on (no doubt because of his earlier speeches that were right...but wrongly worded). He is tackling all these things head on.   At the end he finally said what I needed to hear for me...really it was the words that he said that I have been wanting any candidate to come out and say. Basically he said \"I can\'t do it for you...politicians promise many things and in reality deliver nothing...I can only effect a little change on my own...but with all of us we can make the changes.\"  For too long all politicians have said was \"vote for me and I will make things better\"...and all to often people have fallen for it (most likely because we have placed false belief in them with the hopes that maybe, just maybe they will pay lip service to the causes that they preach.  He sad he can\'t do that. That takes balls..big brass balls.   He hasn\'t hid his past in shame. He was a teacher for some time...he is a MD and could choose to go back to that (and probably make obscene amounts of money because if he brings his ethics of campaigning into a Practice then he would most certainly make a great living).   He is calling for us to return to the high moral purpose so that we can once again have the hope that we are the best...and that we can be fair to the rest of our world citizens.  If you haven\'t seen the clips or films...go and watch...you will be converted.  He is a scappy northeasterner most well known for their aggresive attitudes and straight forward speak...  It is time to take it to the streets and change things.  It is time to fight fire with fire...fire back at them on the morality issues. The Republicans weakness as of late has been Moral Virtues and Iraq as well as the economy.  He is gruff and many times doesn\'t get the \"politically correct\" answers...  Just go watch...because it really is the right words and the right message...  He has it...he has the best balances. We need Dr. Howard Dean.  Go get involved...you will not regret it.   Do it now! Please  '),(1097,1234212,'07,April,2003','2003-04-07 00:00:00','   Oh god damnit...mother fucking son of a bitch ass shit stab motherfucker....damnit!!!!  I had a large post that I had already written when I get a message to go here...and lo and behold I lost my post....shit!!!!  Ok...lets try this again...only this time it is worse. Ok..ok...here we go.  I feel like shit. My heart hurts really badly. It hurts on the left side of my chest right on top of where my heart is at. It feels like a stabbing pain each time I breath in and out. It hurts like a motherfucker.  In between each and every breath my lower back hurts like someone is hitting me there with a fucking baseball bat.   So between breathing, heart pain, back pain, and a really big fucking headache I am not a very happy camper.  Oh yeah and I have gone into another bout with Insomnia...for the past few nights I haven\'t slept hardly at all...plus the medication isn\'t working to make me sleep. When I do get to sleep I wake up from something going into a spasm...usually a lung.  All this shit is happening right when I should be doing much better. The doctors don\'t know, the X-Rays don\'t show anything, the bloodwork shows I am a little out of whack, and obviously the CAT Scan I had wasn\'t to off otherwise someone would have called me.  On top of it Jesse is here and we need to be concentrating on this project we have going. I have a stack of paper over here that I have to scan in and run through OCR so that I won\'t have to retype everything.  I will finally get to see the plan as I get it re-entered into the computer and hopefully get a sense of what is going on. I am really excited about it...but obviously I can\'t show it which I am afraid is making Jesse nervous. I really wanna do this and I am hoping that I physically can hold out and get things together.  So today will be spent doing a lot of office type work.  I also got to look at the problem on his system...he wasn\'t able to log on using the Ethernet port in the system...but it worked fine when I plugged it in here...no problems. However between his house in Hershey and here Evercrack died on him. It just won\'t start. I dunno what is going on with that...I haven\'t ever played Evercrack nor do I want to...I think it just needs to be re-installed and things should be fine.  I am hoping that I can get over all this soon because I need to be healthy for more than a few days at a time. I hate this part of it...but honestly I cannot complain to much because in the grand scheme of things I am not all that bad off. Really I am just frustrated that I have gone from where I was to here in such a short time. I want it to stop...I want to feel better. I don\'t want to be this way.  I am also very lonely. There are people around me. I have made a great new friend in Danielle and I really enjoy talking to her. I have a great family all things considered. Yet I am still alone. Maybe Danielle has felt the same way.  The suffering that goes on is very personal. It is something that even though I know others are, and have gone through it; I still feel alone.  I have tried turning to religion and never once has it made that go away. I tried other things (wink wink) and it didn\'t help...in the end it only hurt and now makes me long for those things (because I may have been alone, my body may have been failing, but it didn\'t matter cause mentally I felt well). I don\'t know how to help the feeling out. I don\'t know that there is anything that can remedy it. I feel only a close connection to something for short pieces of time...like that first millisecond that I pet my dog or give him a hug...or the short thrill when I see a porn clip, or the laughter that I get out of Aqua Teen Hunger Force...but it is all fleeting and the happiness that comes with those things doesn\'t last long...it lasts but mere moments and then is gone.  I don\'t think there is anything for that. I don\'t think there is a cure for that...and really I don\'t know that even if I was healthy that it would be any different. I would like to think so...but I don\'t think that for sure.  Well this post is getting long...and sitting here has really put a strain on my back...and the pain in my chest is getting worse. There is something wrong and I don\'t know how to fix it...so in all likelyhood I will remain in bad pain.   Not to mention that the Medicare \"reform\" bill will probably pass and become law...so that means I will not have insurance again soon. Just wait...people watch what happens. So many people will be hurt by this...many seniors and many more sick people will go untreated. The system will fail. People will die...and the rich will get richer.  If I could make it all right by killing people...if I could be all powerful for even a week...all I would need is a week...things would just be done.  One of the first things I would do is just open peoples eyes...and guess what. I bet I wouldn\'t need to do much more. Because if people were to actually see and experience the wrongs...then they would in all likelyhood make things right. I would also bannish all of the rich and selfish people to a small island that they could not escape...let them try and survive without all the things that they have now. It would be required that they live in the dirtiest conditions that they have wrought upon us...live without micron filters in the Lexus Autos to help filter out the polution that they have pumped into our atmosphere.   The rich selfish people would simply be banned. What would be required of all people is compassion and understanding. If they were to walk a mile in our shoes they would see how unfair it all really is.   I am just so angry....so sick...and I really wish that things would just be right.  In a perfect world there would be an all knowing and all powerfull god...but we don\'t live in that illusion or fantasy...at least not the people that are suffering at the hands of others.  Just fight people...fight while you can. If you don\'t fight pretty soon you won\'t be able to live. By Live I mean literally...the people that have our country in a strangle hold aren\'t in some far flung corner of the world...they are in power here...they are the real evil. The only difference between them and Bin Laden and Hussein is that they were obviously evil...these guys wrap themselves in a \"cloak of christ\" and pass themselves off as good people...when clearly if you strip away the god robe you would see that they are the same evil people...and I do mean evil. Evil in the real sense of the world...if there is a hell...they probably hail from it.  Only the ungodly need to prove to each other that they are godly. A godly person doesn\'t need prove it to people...they just know.  So the next time you see someone on TV talking about god, or espousing their godly purpose...remember only those that need to prove, brag.  '),(1098,1234212,'07,April,2003','2003-04-07 00:00:00','   Well today was pretty much boring...it was semi busy which made the day go by pretty fast.  I am so tired out...it was just a long tedious day. I am glad that I don\'t have to work tomorrow because if I had I would certainly harm inocent bystanders...  Now much to report...outside of the fact that I am tired.  I have tried out the new keyboard and mouse in gaming...and it works well. It will take a bit of time to get things down again...but I know I can do it :-)  Goodnight all...tomorrow Jesse is coming out from Hershey and we have lots of stuff to go over...yipee! We have a lot of work to do :-)...at least it is work.  The closer we get to having things together the better it looks and the more excited I get about it all...yes!  Alright...now bedtime for me.  '),(1099,1234212,'07,April,2003','2003-04-07 00:00:00','   Oh yeah and I have found the IndyMedia in the burg...I may be joining to help out...lord only knows they need more people to get the word out...  Also the ACLU is looking for people as well...I might also zip something off to them...  What will be nice is that these jobs are mainly done behind desks and paper pushing for a good cause...leaving a mojority of my time to devote to other things.   PGH Indy Media Center  '),(1100,1234212,'07,April,2003','2003-04-07 00:00:00','   Well...today I met a hottie. Her name is Jen. I\'m sure she likes to bang...she is maybe a hair taller than I am...she has nice tits and ass...I wanna bang her. Is there anything wrong with that.  I think I should change the name of this blog to \"psychotics am i\" because I am so schizo about my topics...no coherance...no direction...just plain out ramblin from my head...  yup  today was one of those days that if my brain could have come oozing out of my nose...it probably would have.  Balu...the Indian that runs one of the stands in our little group has been having problems with the people that run the stand next to his. Basically when you have a kiosk you are only allowed to accost people that are within two feet of your stand...so you don\'t step on retailers feet, or other kiosks...  Well the Israeli\'s have consistantly day in and day out ignored that...I have seen them running after people to stop them and have them come look at their products....nothing wrong with that in my mind...I wouldn\'t do it but hey more power to ya if you do...but the problem is really poaching...  People will stop in front of Balu\'s stand and look and decide...hell its bamboo...tranquil...it can take a few minutes to sink in. The Israeli\'s don\'t give Balu anytime to react and are all over people in front of his cart...looking at his stuff...well lets say they are like flies on shit.  Balu has repeatedly asked them nicely to leave people alone when they are obviously in front of his cart and looking...and repeatedly they have ignored it...well the Jews that work there are great and they are only doing the bidding of their boss...plus they work on straight commission...almost indentured servitude...only on Jews. (strange how history repeats itself...)  Well Balu finally got fed up and got a hold of the guy who actually runs that cart...and before you could say Jihad he was down there yelling at Balu.  Balu...picture your average 28-35 Indian male from a lower caste...hard worker, intelligent, and really nice guy. He wouldn\'t say shit if he had a mouthfull...he probably thank Sheva or something...  Back to it...this little Jew was up in his face screeming bloody murder (I\'m sure he called Balu an anti-semite) about this and that and how Balu\'s stuff was crap...blaa blaa blaa blaa blaa...now Balu never once raised his voice...he is a good Hindu (I think he is Hindu) and maintained his calm.  Well the thing just kept going and going...I was starting to get a little pissed for Balu...because obviously how do you stop an attack when you are a passivist in nature...you can\'t. I am not really a passivist...I am more of passive aggressive...I let things slide to a point...then I attack.  Well after the verbal tiff was over Balu turned around and I have never seen such controlled rage in a man before. I could tell that he wanted to beat the little Jew to death...but he maintained his cool.  Once the Jews were slightly out of earshot I looked at him and said bluntly...\"That is why they are hated the world over\"...and they are. It isn\'t because of religion or whatever differences Jewish people bring up about faith and this and that...the simple fact is that I have yet to meet a Jewish person that is polite.  Polite.  Polite...the key word. Yes you are trying to run a business...but when you shit on your employees (who are Jewish as well) and you shit on the community around you...and you shit on people outside your faith as the Jewish people tend to do...at some point people say enough is enough...it isn\'t about hating the Jew...it is about hating the attitude.   I am glad that I kept my mouth out of it...and I am glad that I haven\'t been asked my opinion...because this is what I would say. Harsh coming...  \"You don\'t want my opinion...because the last time my people had an idea we killed over 8 million of you.\" Yes I said it.   Look I really did say it...its right there.  Now this comes from humor first and foremost...but it also comes from a total space of frustration with a people that has been whining and bitching throughout their history no matter how good or bad it was...they have always whined and put others down and treated others with less respect than a dog.  I am not saying that I hate Jews...I am saying that I hate when they live up to the stereotype...the Israeli\'s that I work with are certainly nice people...but several of them have the whining gene that must be ingrained into them all...they are really nice...but they still whine and moan...and bitch.  Am I wrong? I would love to be wrong. Really I would. I feel that I am not.   The world doesn\'t have it out for the Jewish people...the world has it out for their attitude and culture. They do not treat anyone with the slightest bit of respect and constantly dehride people that aren\'t of the Jewish faith...and then they internalize it and put each other down...  Basically they are an unhappy people. They all need to take massive amounts of Prozac and just let it go...if the Jewish people as a whole continue down the path (a path they have travelled before I might add) that they are on...we will see another holocost...and this time it may be a complete whiping out of their people.  History repeats itself...always. The only way to keep that from happening is to learn from history and change the course before it is to late.  So to you Jews out there...start praticing what you preach...and spread it out...because if you don\'t you could very well find yourselves in the Warsaw Ghetto...or a death camp...and this time around most of the world will remember the havoc you have wrought across the globe.  Now back to that Jen girl....she made me all tingly...very tingly indeed...I could have fun with that. I hope I get to see her again ;-) Nekkid!  Goodnight...I have work in the morning :-)  '),(1101,1234212,'07,April,2003','2003-04-07 00:00:00','   All day today my mind was consumed with thoughts of my impending decision. I keep waffling back and forth.   Of coarse I want to live longer, everybody does, but I don\'t want to continue living under the gun or on a time table...I have already done that for all my life and frankly it is old now.  Again I haven\'t made any real decisions...of coarse when I do I will post them.  Tomorrow I should hopefully be back to spunky old self with a lot of anger over something political.  The only thing that I will mention today will actually be nice about the administration and a hope that they will do something.  It is about NASA...it really is a national asset and Americans have just kind of taken it for granted. There is currently word circling round Washington that a call for returning to the Moon and setting up a base is in order and that the Administration will look at the options once a plan is submitted to them.  I know politcally and strategically why they want to do it...because the Chinese have committed to a ten year plan to get to the moon and set up a base...we are mulling it over because that can\'t happen of course.  Well I would hope that the president would call for us to go back and set up shop...it is about damn time. However I would suggest that he do something slightly different.  I make the call that we combine our resources with Russia, China, Japan, and the EU and make Mars the target and no later than 2015.  We can do it...together we can get there. The technology gained from that trip or rather preparing for it would be a boon for the world.  Peace  '),(1102,1234212,'07,April,2003','2003-04-07 00:00:00','   Before I went to bed last night I was finally able to find the PanjabI MC CD \"Beware\" and was able to get it downloaded last night. It is so hard to find anything other than the same old drull in music stores nowadays. I mean seriously I go to a music store and I see rows and rows of pure shit but to actually come across something halfway decent or even original.  I guess I heard the MC Panjabi track \"Mundian to Back Ka\" back a few months ago and I thought damn that is cool. I downloaded it in one of my late night fits looking for something new. I listened to it and was enchanted because of the Knight Rider theme and interesting lyrics (which I don\'t understand as it is one of the Indian Languages...I think Hindi..I have a True Indian friend and I will ask him what dialect it is). Then I saw the video...wow beautiful women. The people in the video were different and colourful. Again not the same old drull that is forced upon us here in the states. We live in a homogonized world where everything is edited and pasturized for our listening pleasure. We never see any colour outside of black and white...which is sad.  Well after a long quest for the full CD I finally found it and pirated it. Yes I pirated it...but try finding it in a local store! It is imposible even though he won coveted awards for the CD and best dance track in Europe! On MTV no less!!!!! The CD is absolutely amazing! While I don\'t understand the lyrics because I don\'t speak the language the music translates into any language. The beats are just there. He is able to blend together the best of western style beats (dance mostly, house flavoured) and Eastern sounds like the sitar, some sort of flute (pan perhaps), and the beautiful vocals the accompany Indian music (things that I can\'t even imagine making my voice do).  As you listen to the CD you yearn to know what is being sung, and you yearn to see that part of the world.  While I never advocate piracy of any sort...you have gotta get this Album...it really is different and opens up different types and styles of music without being so far that western ears can\'t handle it. It gives you a taste without the bitter taste of Curry. Give this album a shot. You won\'t be dissapointed.  To give you a taste here is the video  Panjabi MC \"Mundian to Bach Ke\" Video  '),(1103,1234212,'07,April,2003','2003-04-07 00:00:00','   Today was pretty long...the nurse came out early today to gather some blood. She had to draw from my port and stick me in an arm because this was a full workup...or rather a workup to see if my port is infected.  I have basically spent the day trying not to think about everything. I have been fairly successful at that...until I come upstairs to write in my blog.  Funny thing ain\'t it.  I have not made any decisions yet...but I do think I will go ahead with the evaluation and get that process started. So if I decide to go ahead with it then I will already be started...if I decide not too...then really nothing is lost but maybe a week.  Tariq Ali...a name that you all should become familiar with. He is one of the most intelligent voices out there that has critism for the US and its policies...but he holds no illusions. One quote that he gave tonight was pretty good \"the though was consumerism and fornication...and I hope that is what we all are doing...but it isn\'t enough\" Brilliant....saying that hey its ok to be that way but at the same time saying that a person really needs more than that to get by.  His voice is one that I have come to trust because I hear nothing behind it sinister. He speaks well and just seems to hang it all out there and lets people decide.  Go get a hold of some of his material and judge for yourselves.  Peace  '),(1104,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I went to the doctor today and it was an interesting visit to say the least.  They simply can\'t figure out why I am not responding to the drugs, they don\'t know what the rash is, they can\'t pinpoint why my lungs, heart, and back hurt...basically they are at a loss overall. I hate complaigning to them because it makes me sound like I am whining all the time...and sometimes I think I am crazy because they keep on coming up empty handed.  The jist of it is...if I continue to have fevers over the next few days, and things just don\'t get better I will be spending a lot of time in the hospital until they figure out what in the hell is wrong.  Also I got the news today that I need to actually think about transplant. It comes down to they know something indeed is going on in me and since they can\'t find it they are going under the assumption that it is tearing me up. My lung functions have no come back at all...I have remained pretty low overall...and now my oxygen saturation is finally dropping...that means my lungs have finally sustained to much damage and they are just going to continue breaking down.  I knew this was coming and in reality I have been thinking a lot about it. It is hard to have to think about it because I just don\'t know what to do.  Dr. Pelewski laid it out for me...he told me that we should go ahead and start the paperwork and evaluation process now considering that we haven\'t gotten the bugs. I will have to go in for a full batery of tests that will last a week or so depending. I am sure there will be psych evals and all.  He gave me some numbers to chew on. One is depending on my blood type I could wait up to three years before organs become available...it is all the luck of the draw. Also that while there are many people living past five years now the general rule is that 50% don\'t make it longer than 5 years after their transplants. Usually because of infection or rejection.  A lot to ponder.   What this means for me now is that I have to literally put my life on hold until I know what I am going to do. Really it means that I have to stop everything in its tracks...because it is only going to get harder from this point out and I will need all my energy just to fight this war.  I am releaved and saddened at the same time. I guess I am in the home stretch now...and I just have to do what is best for me...and try and make my dent on the world now...because I am almost out of time.  I hate all of this. I hate that I have to face it. I hate that anyone has to face it.   So I guess I am off to ponder...  '),(1105,1234212,'07,April,2003','2003-04-07 00:00:00','   Today as you can see by my first post below started out pretty late. Below is about the time I got up today. When I got up I felt like hell...I mean I really felt like hell. I sat around for a bit feeling like shit and wondering why...finally I realized that I was a little warm. So I broke out the trusty thermometer...lo and behold I was running a temp of 103! I kept an eye on it throughout the day and it wavered a little...but right now it is at around 102F. I wondered why I felt bad and had a raging headache all day...could it have been because I was cooking a few braincells...hmmm.  I got a message back from Indymedia Center here in Pittsburgh. I guess I need to get together a little resume\' of things I have done and what I can do.   It is really hard to type right now because my fingers actually hurt...they started bleeding...yes bleeding the other day and haven\'t really stopped. I think my pinky has an infection in it...it is swollen and feels like it has needles in it...the other pinky is also swollen...but I don\'t rely on it as much to type.  So tonight I will be brief...as you can see from the posts below things are coming together quite nicely for our side...the peoples side. Like I have said before we need to include the mainstream...woo them and bring them into the fold. They are the ones with money and influence...they are the ones...the only ones that can effect change on the scale that we want. We must not alienate them...we must court them and show them we mean no harm...that we mean to help and bring about a just and fair society that will benefit all people of the planet.  Peace  '),(1106,1234212,'07,April,2003','2003-04-07 00:00:00','   Score one for the good guys! Diebold has backed off of its DMCA prosecution against people posting internal emails about flaws in its voting software!!!  Guess which plucky little presidential candidate jumped on the bandwagon to help in the fight. None other than Dennis Kucinich!  Go little D!   Cnet Writeup   I call for all the other Democratic Nominees to also come out against Diebold and help expose the looming problem that many in the computer world have known about. We have been saying for some time that this is a problem. Finally it is getting a little attention. Mainstream media has dropped the ball on this one, but the people have not.  '),(1107,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok...I let my anger somewhat slide over the past few days over Bush\'s visit to Iraq. Now that I have calmed down I have finally come to place where I think that I can talk a little about it.  How do I feel about it?   First I am glad that he did visit Iraq and the troops that he sent there to probably be shot up or killed. He should go. He should be faced with the decision that he and his politico\'s came too well before even 9/11. I wanted him to go...perhaps stick around for a bombing or something...but it was right that he go.  Do I agree that he should have gone?  No...actually I think it was foolish and obviously a political play. It will be plastered all over his upcoming campaign, it will be used as ammo...it will be used as a tool to \"show\" that he cares.  His visit to Iraq was in all fairness not really a visit to Iraq...it was a visit to a US military base...where the security could be assured and the possibility that he might actually face an angry public non-existant.  As many international news agencies have bluntly put...this was the opening for his campaign.   I only wish that he would have taken a helicopter ride around Bahgdad...to really get the feel of the situation. Well he is after all a pilot and supposedly is trained in how to survive...you know all that loyal service to the Texas Air National Guard. Or he could have taken a drive around the city in the unprotected HUMVEES that our troops trek round in. Of course this was not to be because after all he is the most powerful man in the world...  Thinking of it he is...yet he cannot go anywhere without facing the angry mod. What differentiats this mob from others is that it truly is a cross section of the people...these aren\'t hippies, these aren\'t revolutionaries, these aren\'t perhaps some of our seedier members of society...these are college students, homemakers, servicemen, veterans, politicians, people like you and I. The people that are against this are the people who are hurt by it the most...and that is everyone outside of the select few at the upper eschelons of power.   We are rapidly approaching critical mass here in the States. Our economy is so badly damaged that we have around 9 million unemployed (around 6 percent officially unemployed...not including those who have just stopped looking for work and not recieving benefits). We have upwards of 40 million people uninsured. No longer are the poorest amongst us without...it is now eating into the suburbanites lifestyle. People across the nation are reeling at what is going on. Quietly sucking their thumbs, rocking back and forth, hoping that next week things will be better. Some are hoping that the economy has a quick rebound, but most know that it won\'t happen. These are the educated amongst us that are starting to see things slip away. They see that even if they are employed that their take home pay...what they see each month is less and less.   Where will this lead us in the future? I haven\'t a clue...but I do know that if things don\'t improve people generally tend to get a little restless. Social norms change, people look to other solutions, people start to stray from the idea that capitalism is the best system.   Capitalism and Democracy don\'t mix. They never have because capitalism requires a willing public to go along with making a mere few very rich while delegating the rest to a life of survival from paycheck to paycheck. I know this because of my neighbors. I see them with a mortgage, car notes, monthly bills, and if they were to loose one paycheck the entire system would fall to its knees. I know this because while we have the richest people in the world living here we also have the highest level of bankruptcies in the world. I know this because when you can walk from one house in a suburban neighborhood where food is plenty to just three houses down there is a family that scraped together enough money to purchase a turkey for Thanksgiving.   In Ohio 2 million of the 11 million citizens have to go to food pantries so they can eat. In Ohio! Where 1 in 6 lost their jobs over the last three years (mostly manufacturing jobs which pay well). In some places college tuition is up 25% over the past couple of years...what does this all mean?  It means simply that the capitalist idea does not work well...and especially here it doesn\'t work well. We see it in policy. We see it in the new rules where Corporations no longer have to tell the community when a chemical that could harm the community has been spilled...not because of National Security (I don\'t think PPG here in Springdale makes much outside of house paint...or dyes). We see it in the policies of our nation to curb free speech...not by banning it but by making it not profitable to the person making the speech. Where we can essentially curb free speech through threats of forclosure or causing a person to loose their job.  Now that Mr. and Mrs John Doe on Cherry Hill Dr. see that all they have planned, saved, and worked hard for can go because some Corporation decides that it would be a few pennies cheaper to outsource Mr. Doe\'s job to a foreign nation...not just because it is cheaper but because then the company doesn\'t have to have environmental plans.  as Mr. Doe sees this...sees all the problems that he faces he will start to look outward...not inward because he knows that he has done everything \"right\" to be a success...he went to school, got married, had 2.3 kids and a dog, bought a house, a new car, and the big screen TV...he did all the right things but he can\'t put food on his table for his children...and not to mention that he can\'t even imagine being able to afford sending little Billy and Sue to college...even though he started a fund for that.  When Mr. Doe sees this and realizes why...that is when things will change. Because there isn\'t just a few Mr. Does\' out there. There are millions of them...from every background, every ethnicity. Once they all come together something will have to happen. Not because someone planted it in his ear...but because he knows it is right...he will take action and things will change.  Often people refer to all these problems as numbers. That is all people see...numbers. They are easy to swallow, easy to digest, easy to forget. Putting a human face on these numbers is the priority...or rather should be the priority of us who want to effect change.  That will be our weapon. We may need guns to forcibly change things...we may need other things. However the most important thing that we could ever have...weapon wise is 40 million Mr. and Mrs. John Does\'. We musn\'t forget them, we need to include them. We need to bring them into the fold.  We can\'t scare them with and overload them with all this information that is indeed needed for them to survive...but instead we need to introduce them to new thoughts. We need to show them that we really don\'t mean them any harm. Right now we are the left wing conspiracy nuts. They can\'t identify with us because we want so many different things...our voice has become muffled by a frayed front. When we unify, when we pacify, when we pull it all together and slap a happy face on ourselves and become very serious and business like in how we deal with things...then we will have those 40 million strong middle americans who want for nothing more than to live comfortably without having to worry about health care or where their next meal is. These people are mainsteam for a reason. Not because they don\'t believe what we say...but they can\'t identify with what we say. However they are coming. They are our great saviour.   So reach out to the conservatives...you may find in the end that you hold many things in common. We must not comprimise, but we must not intimidate them in following our lead. Show them that what we want is the best for all and identify with them on a personal level...and they will gladly follow. That is how the right won their stunning victories. They scared the people over to there side by speaking half truths and outrite lies. We cannot do that because they are already disallusioned. Face them with facts and numbers...then show them that they are what we are fighting for...not just for our crazy ideas but for their welfare.  Once we do that we will win everytime...regardless of who is in power. So continue to be as left or right as you want...it doesn\'t matter...however instead of threatening the other side to move with us we must invite them.  Once we offer a better deal they will take it. They will take it because it is best for themselves and family and they are smart enough to realize that.  It is simple in concept hard in practice because knowing how strongly I feel about it...I am sure there are others out there that are a lot better at this than I am. We can do it if we pull together and speak as one...then once we have made it the topic then we can get through better law, better ideals, and in the end a better healthier planet.  '),(1108,1234212,'07,April,2003','2003-04-07 00:00:00','   I just noticed on my page there is a little banner add. I have known that there was one there...mainly because this is a free site that allows me to blog...I sometimes knew that it was advertising for Google or other places...  Today I noticed when I brought up my page that it had ads for David Limbaughs book \"Persecution\"...I am almost mad because I would never ever promote a book like his...that right wing nazi. He is one of the bad people in the world and should die.  In smaller print you see Jim Hightowers link...in really really small print...and it is just a link...not really a tag line or anything.  Also a link for political discussion in an open forum...I really haven\'t followed the link so I wouldn\'t know about it.  I wonder if my blog has garnered the attention of anyone yet. I know the Government has little web crawlers plodding through the internet all the time to kinda keep an index of things...I wonder if it has made it to my corner of the web yet...It will it is only a matter of time...  '),(1109,1234212,'07,April,2003','2003-04-07 00:00:00','   Wow do I feel like shit. I mean I really feel like shit. I didn\'t sleep well last night yet I still had to get up early to come to work. If Denise didn\'t have to depend on me today I would have called in sick...because I am seriously feeling like shit.  Did I mention that I feel like shit? Yeah shit...the stuff that you poop out each day...  The rash is worse today...there seems to be no end to the hell that I go through on a daily basis. I have pretty much given up hope that I could have a normal day at this point.   I really wish they could pin this infection down and get rid of it quickly because all the meds are doing more damage than the infection itself.   I hate bitching and moaning about things I really do...but right now that is about all the energy I can muster...because bitching is soooo easy.  I get to visit the doctor on Wenesday so maybe he will have a new plan of attack...who knows? I don\'t.  What I think I need is a solid week of sleep....like no waking up to eat or pee...just sleep. My bateries are run down and my nerves are frayed.  Well I am off to look for some political stuff...cause you know that gets me fired up...maybe it will wake me up a bit so that I am not sitting here almost falling asleep.  Baaahhh  Oh yeah and did I tell you that I feel like shit!!!  '),(1110,1234212,'07,April,2003','2003-04-07 00:00:00','   Today was just a long day...that is all I can really say positive about it.  I am still sick...stuff in my lungs is getting thicker and nastier in colour. My body is a mess and their ain\'t much more they can do. I see the doc on Wednesday so we will see if he has anything...anything up his sleeve.   I am breaking out in a rash...I dunno what it is from. My back looks like I have really bad achne and it is spreading to my face...but it isn\'t achne or zits...it is just a rash...with little blisters popping up. It all is painful and kinda gross...so when I see the doc on Wednesday I will ask for something...maybe a referral to a Dermatologist to see if there is anything that can be done...my skin just feels like it is breaking down. I hope they can figure that out because it is not something I really need right now for another part to go bad.  So here is the summary. I still have a lung infection, my kidneys and liver hurt, my gut is constantly upset, I am constantly on the verge of being out of breath, my fingers (cuticles) are peeling back, migraines, CFRD (CF Related Diabetes) is working a number on me, and I am tired...  Outside of that...well my world is peachy.  I ask for no sympathy or pity, all I ask is that if you pray or meditate, please devote a little time thinking nicely of me.   This week promises to be hectic...I am sure because I have a lot to get done...and just seems that I don\'t have all the energy to get it all done with...  Love out to the world. Mike  '),(1111,1234212,'07,April,2003','2003-04-07 00:00:00','   Today really was a wasted day. I slept until 4:30PM...then I really just zoned out...now I am getting ready for bed.  Tomorrow will be busy I am sure. I hope it will be busy because I really don\'t want to just sit around and do nothing.  Oh well...maybe something tomorrow will be interesting.  Oh...I saw Bill Moyers speech from Madison WI...it was a forum on media reform. His speech was dead on. He is the host of PBS\'s Now! With Bill Moyer. He is one of the few out there on a national level that has called to task all of what Bush and others have done in government to kill free speach and ultimately Democracy.  It wasn\'t all Bush bashing...he mentions democrats in with it as well...but more along the lines of the democrats have failed yet again to stand up to Bush\'s policy...  One line...a quote of his really got me and drove home the point \"virtue is not inherited, even if your father was the president\".  That drives it home. Bush Sr...while not the greatest President we ever had...definitely no friend to the liberal or people wanting to save the earth was at least someone you could look at and know that maybe he wasn\'t outright lying to you. Also he actually served his country...no so with Dubya...  I am still a Registered Republican...I still believe in many Republican platforms...but I have found myself at odds with that because of the overwhelming things I don\'t agree with them on. I would say I am a liberal republican...the only Republican that I can even remotely identify with is Sen. John McCain of Arizona...and I feel that the people Arizona should be proud of him. He is the last of the middle or liberal Republicans left. He is still what I would call the moral fiber holding the Republicans together.   I hope for his sake that he moves to become an independant...I am sure it will be much harder for him to be elected...but if the People of Arizona were to look he really has been very good for the State...and the Nation. Only fools would vote in a replacement at this point.   I find that I most identify with Kucinich on many many issues...but I approach it from a slightly different angle...like bringing back \"public\" utilities...I don\'t come from the aspect that it would be good...which it would...but I come from a cost perspective. Energy companies and producers get the largest tax breaks.   TVA (Tennessee Valley Authority) was a public utillitie that sold electricity to people of the South...The TVA is credited with electrifying the south. Their kilowatt/hour charges were a little on the high side...but not really beaten out to badly with other provite power companies...  When TVA went private to be sure the rates dropped...probably by half....however no one seems to have taken into account for the massive tax breaks and incentives that were given to TVA by the Federal and State Governments...  Break it down like this.   TVA Public = 13 cents/kilowatt hour close to the cost to produce each watt TVA Private = 6 cents/kilowatt hour actual loss on each kilowatt sold...but with tax breaks and incentives then they break even and make money.  TVA Private actual cost = 6 cents from consumer + 8 or 9 cents from government tax breaks (which the people pay for by paying taxes) = 15 or 16 cents/kilowatt hour  It costs more to the consumer for power now...however it now costs less for big businesses to run using that electricity...because they are probably paying 4 or 5 cents/kilowatt hour...because they buy in bulk...the electricity actually still costs 12 cents to produce...who picks up the difference...not the businesses...but the taxpayers ultimately subsidize tons of industries just through direct subsities and tax breaks to Power companies....  not to mention corporate taxes have gone down steadily...so you see only the end consumer gets screwed on the deal. We are subsidizing people who are making money hand over fist...  Now the numbers I have given are only relative examples...they don\'t actually jive with current rates...but the concept is there....and that is why I chose to use numbers...numbers people can understand...not just words.  If you were to take away the tax breaks and incentives then you would quickly find out that all Electricity Providers in the nation would fail if they didn\'t raise their rates...because as it stands now the actual selling of the energy is not enough to cover the costs of producing it...  Why move them back public? Well under public systems the prices may be higher...but in the end are lower because Public means only that they have to provide power...and if something is going wrong at a plant they can pull it offline without fear of loosing profit...they just do it to make sure nothing happens...plus it is just plain cheaper to do things...  I hate the fact that no one has really looked into it to get all the numbers right. That may be something I do and zip off to the networks...because if people knew that they would cry for public systems to be re-established...oh yeah and the power shortages in California would have never happened because plants would be constantly updated instead of trying to run them to death and not fix anything...  I hope we can all come to the conclusion at some point that privatization is not always a good choice...privatization really only works on small scale things...not on large scale social needs or programs...never has, never will.  '),(1112,1234212,'07,April,2003','2003-04-07 00:00:00','   Yesterday our ruler GW Bush declared that those surrender monkey nations would get none of the booty from Iraq saying that they could not participate in primary bids for contracts in Iraq. So I guess that means that Halliburton, Carlisle Group, and Bectel will get all the good no bid contracts (to be fair Carlisle Group is more of an investment group).  What wasn\'t said is that companies from those countries would be allowed to get secondary sub-contracts that spring forth from primary contracts...a good example is Seimens and Erricson are the actual companies supplying hardware for the new cellular network. So Bush misleads yet again.  Now what is so funny about this all is that he also appointed James Baker as a special envoy to go to these very countries to ask them to forgive the debts against Iraq....again that is a little misleading because in international courts...or rather law. When the US invaded and took over we actually took on all responsibilities of the government there from providing water to paying off debts...so in reality when we invaded we took on the 120-130 billion worth of debt that Saddam had racked up...and legally we are obligated to pay it.  Wooohooo! Now our national debt is really going to sore!  Now actually I agree with the president on this one...really by all rights it is the US taxpayers footing the bill right or wrong...so the money should stick to our companies...or be given to the Iraqi\'s (preferably) and let them contract out the work. However I don\'t that the latest \"fuck you world\" statement out of the White House is going to help us one bit in the longrun.  Whoever wins the White House is going to have to be making a lot of phone calls appologizing for our rogue government and what it is doing now.  Another great thing has happened. A week ago the EPA and White House were calling for a market based solution to Mercury Emissions that would allow plants to buy and sell pollution credits for Mercury. It is estimated that it would reduce the mercury emissions by about 70 percent over the next fifteen years. The prior mandate was that they clean up their acts and the mercury emissions would be dropped by 90 percent in the next four years. At that time they were arguing that mercury poisoning was a little overblown and such.  This week the EPA issued a warning for children and pregnant women to have them eat less tuna because of elevated mercury levels in the tuna...that could lead to mercury poisoning in people, especially those two groups.  Does anyone else see the irony here?  I think the country has gone mad...from all the mercury in the food...that could be the only explanation!  Peace..don\'t believe the lies.  '),(1113,1234212,'07,April,2003','2003-04-07 00:00:00','   Not much to say today as I am tired from work...there was this christian lady that stopped today to talk to Denise...but before she got to talk to Denise this woman was trying to convert Hamza (I think that is her name) the Indian lady...I am pretty sure they are Hindu\'s  The woman left her with a flyer...and there was a word that Hamza didn\'t recognize...Satan. So I told Hamza what it meant. She then asked me about the christians and we talked for a few minutes and I gave her a quick lesson in christianity. Hamza said the greatest thing back to me \"what is the big deal we all worship the same god you just call him by a different name\".  Right on...even a Hindu gets it! But christians, muslims, and jews just don\'t...and really those religions are cousins of each other.  As I have come to my conclusion about it all and I told Hamza that you don\'t have to live a godly life...just a moral life and you will A-OK.  So stop the hate, forgive each other.  Peace  '),(1114,1234212,'07,April,2003','2003-04-07 00:00:00','   Cloud10/Highradiation is holding a raffle for a Sony TRV900 3CCD.  If you would like a shot at is the tickets are 20 bucks each..not a bad deal to get in on a swanky camera. It is used as he was using it for his business and needs to upgrade his camera setup...and the new setup will run about 4 grand.  I personally is am going to enter twice...maybe more because it would be sweet to get a camera like that. So far the pool is somewhat low for the raffle so the odds are good.  If anyone is interested here is the   Link!   Go get you a ticket peeps.  Mike  '),(1115,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I watched the new Battlestar Galactica mini-series on Monday and Tuesday and I must say that I was thouroghly impressed and overjoyed at the treatment SciFi gave it.  The classic BSG was only 24 episodes for a total of 24 hours of program. The original became somewhat of a cult classic mostly because of the characters and technology they used which was good for the day. The largest complaint that I had with the original was it left a lot of things unexplained leaving gaping holes in the story line.  The new BSG mini-series tackled most of those holes in a new way. It explained the Cylons and went on to explain that the Cylons and Humans got into a bloody war and then there was an armistace. They built a far flung station for the Cylons and Humans to meet yearly to keep the peace. For 40 years the humans sent someone and for 40 years the Cylons sent no one.  That is how the series starts out with an ambassador from the humans at the station and he is startled when the opposing door opens and in march two Cylon centurions followed by a blond woman...who is a Cylon. The space station is destroyed and the war is on.  The series ends with a major showdown in which the Galactica protects the fleet as they jump with their Faster Than Light drives out of the hotzone and th Galactica following suit before they themselves are destroyed. It left open the door for a weekly series which I would bet good money that they will do...after all the put Tremors on...  For hardcore fans they may be mad at the results as it really changed the story and really only kept character names and only basic elements of the original. However this remake will bring a new generation into the BSG galaxy which is a good thing.  Overall I loved it all and even for me a fan of the original I came to grips with the new look and feel as well as story because it was well done.  I went to the doctor today and I don\'t have to go in yet. My numbers are getting a little better but I still feel like crap. So right now we are in a wait and see holding patern to see what shakes out over the next few weeks. I go back next week to see what is up and see if I have continued to improve.  Well I am out for the evening...Peace!  '),(1116,1234212,'07,April,2003','2003-04-07 00:00:00','   look I re-added the coments interface.  now just give me some feedback...See B. I told you I would get it back  '),(1117,1234212,'07,April,2003','2003-04-07 00:00:00','   I see that I have a loyal readership of around five people and I would like to increase that number...because I want to be popular damnit!  Gore endorsed Dean today officially...and he did it right \'round the corner from Bill\'s office in Harlem. Clark revealed today that he had been in touch with the former president indicating Bill and Hilary would be endorsing him.  The problem is that no one can deny the Clintons became the undisputed leaders in the DNC when in 92 they energized the party...however now they are both the problem. They have continued to move the party more centrist in their views and have resisted change...such as bringing Dean in to the inner sanctom or at least to recognize that he is the new face of the party and brings with him the opportunity to effect change not only in the DNC but the nation.  I am sure the powers that be in the DNC are running scared...as well they should be because they have single handedly taken the party from power and moved them to the outside.  Over all this flap the DNC is close to breaking and is having an internal war over what to do next. That is their biggest problem and that Dean isn\'t an insider and has so far refused to become one of them because that may alienate his core...not because he moved mainstream but means that he would become exactly one of the people he is targeting.  Dean has made mistakes but I stated this on Brian\'s blog in his comments that I would rather see a flawed man in office that would begrudginly reconize mistakes and appologise (sp) Bush on the other hand has been become Republicans golden calf in which he can do no wrong and is as close to god as ditka fans portray him as god...and that is their greatest weekness...refusing to admit their wrongs. So this is where the Dems have a great shot...showing that the White house says one thing and then does another.  Kerry got in trouble for saying fuck...\"I didn\'t realize he (bush) would fuck up iraq so badly.\" He said this in an interview with Rolling Stone magazine...the white house responded that they would hope he would retract that saying that cursing is not polite...  Well I got news for you kids...we should be screaming obsenities...it is far past the point of being polite and time to use real language.   I think it a lot of hubbub over nothing and really that is common language and really swearing is just fucking words...it isn\'t a sniper rifle aimed at them...curse words are just normal so why try and not use more colourful language. These words are used by most everyone and really only offends a small portion of the population because they to use that word.  The Daily Show with John Stewart got Bush good over this showing him at a rally with Cheney saying some swear word to him without realizing the mike was still hot...the hipocracy is just mind boggling....  Well I am going to see the doc tomorrow and I expect I will be going in soon. Perhaps tomorrow...and who knows how long I will stay.  Until next time kids.  Mike  '),(1118,1234212,'07,April,2003','2003-04-07 00:00:00','   I want to send a quick thank you out to Pam. Thank you for your kind words...and I still think about you a lot and hope all is well over in your neck of the woods.  The love that has been shown over the past few days by my friends and family has really touched me deeply. I am greatful for all the experiences that I have had and the journey that has brought me to this point in my life.  I have but a few regrets in my life and really none of them are that huge. Mostly it is regrets on how I treated some people or what I have said to them...but I regret none of my actions.  Some of you have been through thick and thin with me and I want to express my deepest gratitude towards those who have been there. I love you all and I am thankful that I have people like ya\'ll in my life.  I still haven\'t made any decisions on what I am going to do yet. Partly because I am very torn between what I believe and what I know is logical. So I will break it down for ya\'ll here and I would appreciate any opinions that you have. Everything that I am going to write comes from my view and is what I would hope be a critical look at my views and my feelings.  There is a large part of me that does not want to go through a transplant procedure and it comes from several different places.   I will start with the easiest of the two arguments as I see it. Part of it has to do with just numbers.  Depending on blood type (I have never known my blood type but my mom is B and my dad is O I think) and tissue type I could end up waiting for up to three years (less or more depending on really someone else dieing that is compatible). That part scares me because really for all of my life I have been living under a time allotment. When I was much younger basically the guidance was that I would live for 16 years (as most CF patients at the time lived to about 16). As time has passed the treatment for the disease has gotten better thus increasing average lifespans for CF patients like myself to the mid 20\'s and early 30\'s without much illness. So I have been pretty dead on overall with the average.  Of course there are people who live longer (there are several that are alive now in their 40\'s and 50\'s) and there are some who die very quickly (die by the age of 10 or less). It all has come down to what your particular Genotype is. There are several hundred genotypes of CF. With the least affected people (with a very mild genetic anomaly) that live for a very long time, and those with very severe genetic anomalies that don\'t live long at all. This has led to new treatments and courses of action depending on your genotype. Well right now the CF foundation and the doctors are forming real treatment for the different genotypes. Why they haven\'t done this until now is beyond me but they are now seeing that they can more effectively target treatment based on what type of CF you have. Knowing the genotype in the future will give doctors a better gauge of how long people will live and what treatments are most effective.  The other part of the numbers side is that only 50% of transplantees live past five years. Some live longer, some die shortly after the transplant. Again they are starting into studies with genotypes to figure out who will do best and what treatment course would be needed for the different types. Also the rejection drugs get better each year so the percentages will most definitely change over time. There are now transplantees that are 10 and 12 years out with no signs of illness.  The transplant itself does not cure you. It merely replaces the defective part with a normal functioning part. Whatever part you replace no longer has CF. So you get your lungs and heart replaced and they are normal but the rest of the body still has CF. So you could die of other complications from CF such as your intestines breaking down completely or liver or kidneys. It is all a crap shoot. I think that in the future we will see more and more CF patients coming in for kidney and liver transplants because of the treatment of the disease. We use very heavy anti-biotics routinely and those anti-biotics kill of both organs when used at the dosages and frequency that we have to use them.  From a numbers aspect it is about even. So I cannot base my decision on that alone. If they could tell me that I was 80% likely to be around for more than five years than it would be easier because that is more of a guarantee than a warning and I would definitely go ahead with a transplant.  The second reason and perhaps the more difficult of the two has to do with how I have lived my life thusfar and what I have to offer down the road.  I cannot predict with any certainty that I will go on to do great things. Really no one can so that becomes somewhat of a moot point but it is something that I take into consideration. More from a tools and what I have already done aspect.  There is no doubt to those who know me that I have always lived in the now and done many things that others would consider endangering my life.   I have done lots of drugs. I love drugs. I really do. I have done so many that the list is longer than the ones I haven\'t done (and those that I haven\'t done I would love to try). The drug use that I have participated in was always recreational and never ruled my life (as many can tell you that have known me). I was always the one pushing the limits and going well past the point of common sense. An eigth was good but a quarter is better, and half an ounce is even better. I rarely bowed out because of health problems (and when I did it was just because I knew I needed to heal). Not once did any of the drugs rule my life...partly because I not a hugely addictive person. Its not to say that I didn\'t like them or I couldn\'t become addicted...but as many of my friends know I was always cheap. I never wanted to spend money on things alone...I would often pool in with people...or if I bought something I almost always shared with my friends (and strangers).  I pissed away a college education. I can explain why but it isn\'t a good explanation. I always held the belief that I would be dead by now (by age 25 in my mind) so I never threw myself into education full force. I did enough to eek by and then it turned into \"well fuck it I won\'t be here much longer\". This reasoning actually has led me almost always into trouble. Whether trouble with the law, family, or health.  Its not to say that I didn\'t learn anything. Actually I learned a lot and I consider myself an educated person (ableit without a degree). Really I don\'t have but maybe 20 core credits to get done. I have around 110 completed...but most are non-specific. Partly because I just wanted as much as possible but also as diverse as possible. I took a GIS class (geogralhic infomation studies/systems)...not because I needed it or it would count for anything. I took it because I liked the teacher and I though it would be interesting and it was. I loved that class. The computer knowledge that I have gained over the years paid off in spades in that class...not only for myself but for other students (like being able to look up census data and formating it for presentation on the computer).  I never took college seriously and that is why I was in for such a long time. I would love to finish it just to say I finished it, but it honestly doesn\'t bug me that much that I don\'t have a degree. I take my grandmothers position on it...you go to college not for a degree but to learn how to learn. I accomplished that goal. I accomplished the development of my mind to think of things critically and on what steps I need to take to accomplish any task that is presented to me. I take pride in that because I know many people who went to college (some finished, some didn\'t) that really never developed that skill. I have friends who just can\'t problem solve and when something comes up they panic and make poor decisions or no decisions prefering to let others make them or do the work.  This ties into my view or feelings on transplant. I have lived a good life overall. Sure it hasn\'t been roses, but overall I made great friends, had a few good times, and expanded my knowledge, as well as lived in relative comfort. So there is not much lacking in my life (except maybe hot monkey sex).  If I were 12 or 16 then I would go ahead with it because I hadn\'t lived life up to that point. However I am 26 and I have done more things than most accomplish in a lifetime (normal one that is). I have traveled (not as much as I would like), I have lived (not as comfortable as I would have liked), and done many things (not all that I wanted but close). So continuing on would only diminish what I have done...because it may be downhill from here. What will I do if the transplant succeeds? Where will I go? How will I live?  Which brings me to a rant...I cannot see why anyone in their 50\'s or 60\'s would expect someone to give a kidney. You have lived your life folks...exit stage left gracefully. Accept the fact that you are older and worn out...so be thankful for the time you had on this Earth.  So that is what is going into the thought process. Again I haven\'t ruled anything in or out. I just am mulling over what I want to do.   Well I love you all and I hope to get to see some of you before things get bad...as they will and I want people to remember me (whatever the outcome) as I was...a fun person that loves to have a good time...even if it isn\'t the best thing for my health.  Peace!  '),(1119,1234212,'07,April,2003','2003-04-07 00:00:00','   Today was a slow day for all...not much occured...I am up way past my bedtime and I think I will go to bed.  Peace  '),(1120,1234212,'07,April,2003','2003-04-07 00:00:00','   Today was pretty eventfull...I took my sister out to where my mom works at Fed-Ex the Pittsburgh Ground Headquarters...she was going to sit with her and spend some time. While I was out in the Robinson Towne Centre (they actually use those spellings which are olde english spellings...but nonetheless good spellings) I got some of my gift shoping done. I still have more to do but overall I am nearly done...and so far I have only spent about a hundred bucks total. Did I ever mention I am a tight bastard? Well I am...I have saved a little money but I really don\'t want to spend it so I am not going too.  The anti-biotics I am on are seriously taking a toll...the one med smells like cat piss and I have to infuse that...eeesh! My urine has turned a lovely shade of canary yellow...almost orange and it smells horrible...I know way to much info.  I just back upstairs because my poor poor Rufus is sick and has been for a few days. Yesterday GM woke up to a pile of gew that my dog had shit out on the floor. Now Rufus has never before shit in the house...he peed one time but never this. All of today he has been puking and just a little bit ago had those runny bloody stool thing. Good thing my sister the Vet. is here...she thinks he has something seriously wrong. I think that whatever it is...if it is a bug he got it from this little toy poodle my sister brought with her here to get a home for. Now mind you that the two of them didn\'t play or really get near each other so much...but still he could have got something...but most likely not there.  My sister thinks that he is bleeding in his gut...and if that is indeed what the problem is then we may have to put him down...and I don\'t want to do that at all...because I do love my dog a lot...I just hope it is some bug that we can kill easily and get him back up and running.  This sucks that my dog is really sick. We are taking him to the vet that my sister did an externship with...partly because she can keep and eye on him and they will treat us right down there...so I am thankful to have that otherwise the bill would be huge!  Ok...now where was I. Oh yeah the Seatle Washington WTO protests from 1999. I have seen several documentaries that have been put together by Seatle IndyMedia...they were informing but disturbing.  The WTO is short is where government heads of state and corporations from all over the world get togethe and basically discuss where things are to be done...how to get the people to go along with it....and of coarse make the US Companies the only ones who would really profit from this...well not the companies so much but rather the CEO\'s and board members for those companies...as a matter of fact this is where deals are struck for getting another \"garment mill\" (sweat shop) from some far flung place in the world...think upper third world, and all middle world nations. So the business man from Gap goes in and negotiates a deal with a Phillipino contracter to make a pair of jeans for 4 US dollars...the worked of those factories rarely recieve more than a few dollars a day in wages while the boss\'s make all the money. Gap then turns around and sell the same jeans for 30 dollars...that is a hell of a markup I tell you.   The WTO is designed obstensively to promote free/fair (I added fair) trade between nations. As it stands now we take advantage of smaller nations, plunder, loot, and pollute them then leave when someone else offers a better per unit cost.  In 1999 The WTO was meeting in Seatle WA was nearly shut down and because of the protests (people flew in from Korea and many other nations to protest). The shut it down. With this civil unrest many of the third world leaders pretty much walked away and said (even brutal dictators) we can\'t force this down our people\'s throats and have them like it at all...  During the protests and marches there were up to 50 Thousand people...and of coarse you have the dumb \"anti-everything\" people who go around causing havoc and breaking windows at a Starbucks (in Seatle you dumb fucks...Starbucks home!) and several other places. The vast majority were peacefull and did nothing but Civil Disobediance excersizes. Again such a minority (maybe 20-50 kids max out of 50 thousand protesters) that the rally and march was by all intents peaceful.  After they stopped the first day of real negotiations (many delegates had to make their way through the protestors (where many of the third world leaders say the unrest and changed position on the WTO.).  After that day the mayor made a 25 block zone in the city that there could be no protests, no signs, no buttons, stickers, or anything that would talk against the WTO...that is when things started going wrong. The police repeatedly fired their tear gas (even though they annouced that they weren\'t deploying Pepper Spray of Tear Gas) cannisters into the crowd almost alway deliberatly at people. Now those things come out pretty fast so getting one in the gut or chest would knock a good size person down...then the cannister explodes on targer and out pours gas...now getting hit by one of these cannisters is dangerous no doubt...and they could be used in a manor that would kill people (thank god the police no longer know how to shoot and properly handle a firearm.  The third day by mayoral decree no gas masks were allowed to be sold to protesters, a protester with one on can be arrested. So they basically said you haven\'t a fightin chance here, give up, and go home. But the video footage of the protest was very informative.  80% of Police department here in the states have a paramilitary (swat) team. Their overall sucess rate for all of them is in the fifty percent region. Sure I believe we need SWAT teams but to be used very sparringly and only when they are needed.  The banning of getting gas masks pushes a little to far.  Big labor unions have been becoming more buddy buddy and it is about time. As one put it \"we are fighting the same fight, we may not agree with them on everything (as they don\'t us) the basic concept is there and once all the little groups get put together they will be unbeatable.  The steelworkers came up for the Seatle protest and marched the first day...and in trying to get more people to walk from the union many of them said \"what the fuck good will that do?). Once they saw their colleagues being gassed and kids being kicked around those holdouts came out in force and again locked the downtown core up again.  A similar thing happened this year The WTO protesters shut it down (it was in Miami I think).  For us the good guys to win we need to embrace the unions and once elected push some of their agendas because they are not politicaly tilted...but facts. We don\'t have to agree on everything but we need to make the first steps to start on the road...so that is good news for us all (and for a little history on me...for a long time I didn\'t like Unions. Mainly because of the Teamsters because they were and are crooked as hell (although they are straightening out). Once we join our forces and talents we can effect change.  Well I have about three pages more to write about this but I won\'t...so!  Peace! (oh yeah and happy holidays...for me I will be glad to get my shopping done).  '),(1121,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok one last thing unless something really interesting pops up.  Dean has released his new domestic plan calling it \"A New Social Contract for America\'s Working Families\". I will provide a link to the entire speech/plan.  Reading it is like a dream come true and in the correct wording so that no real American can look at it and say \"hey that isn\'t good\".  He proposes health care for all based off of the insurance government officials get right now...a nationalised health care. Only one other candidate has proposed what in our country is a radical idea (even though all industrialised nations have national health care save us).  Having thought about this a lot in the past year or so I realize that it will be difficult to get people to buy into it and companies and insurers will most certainly denounce it as socialism or communist...and maybe it is socialism...but we already live in a quasi-socialist state already.  Dean would repeal the Bush Tax cuts (oh dear raising taxes...but in reality it isn\'t a raise in taxes for most people as they didn\'t make enough for the Tax cuts to help them anymore than maybe a couple hundred a year...which the average middle class cuts amounted to around 300 a year...that is not much...and not really a stimulus for the economy because for many this money went to pay bills down...so most of it wasn\'t spent on products.  To make it work the Tax Cuts would have to be repealed...and companies would have to pay into the system what they were previously paying into private insurers...instead that money goes to the government for health care (some might call it a tax but in reality it is shifting funds that go to private insurers to a federal agency that we all benefit from). People would also pay in as well but their contribution may only be a hundred or two per month far below the average of 680 a month for private health care that is flakey at best.  The money can be raised and used properly...but only with our, The American people, lobbying and really making a stand from it.  In this document or speech Dean again stresses that almost all politicians say they will do this or that for the public when they rarely carry out what they say...and Dean says basically I can get to the white house but I can\'t effect change without continuing support from the public.  He just said he can\'t promise anything but will instead inlist our voice in the debate for such large projects...and that is the right thing to say and do.   Dean promise and plan.  '),(1122,1234212,'07,April,2003','2003-04-07 00:00:00','   Today I spent a few hours at the hospital getting my first dose of a anti-biotic I have never had...they needed to monitor me and see if it would kill me...it didn\'t.  Really nothing much happened...I informed my sister of what is going on around the globe in our name...and she was kinda surprised.  Basically I gave her the simplified version of why we went to Iraq and Afghanistan seeing that the major media has never really taken on what is really the motive behind it all...and they probably never will.  Again nothing much happened so that is all I can write about...oh one thing I can go see Daniel because we are both negative for the new nasty little bug...but to be sure we have to keep three foot between us...which is good news. I am going to try for early this next week to see her.  The new meds are already taking their toll...so I don\'t know what is going to happen for work. Really I think that I am almost spent there because now I have new meds and they are harder on my system overall. I will call my boss tomorrow and inform her and tell her that I can work...but I can\'t be there for a long time...plus I have to avoid the public as much as possible because of the Flu...even if I wanted to get a flu shot I can\'t because I am already on a razor edge with my health and the shot would most certainly shove me right off the cliff.  LOTR: The Return of the King...rocked...go see it you won\'t be dissapointed for sure. The complete series (the three films) will most definitely go into many people\'s top 25 movies of all time...Peter Jackson did a superd job on them...even if he doesn\'t end up getting best picture or Director...the film is literaly...all three films are literally some of the best films of all time. Everything was done with great detail and was dead on...he stayed as close to the book as possible without having the films run four or five hours each. These films are long films...and LOTR: Return of the King is very long...but you do not notice it at all because you watch it and yearn for the next scene to see how things play out. If I would have been alone I would have cried like a baby because it was just so moving and so well performed...so go see it.   Peter Jackson...you are most definitely one of the best directors ever and the masterpiece that you have created with this series will stand the test of time...people fifty or a hundred years from now will look back and watch the movies and they will hold up due to your hard work...and the hard work of the talented people you surrounded yourself with from the actors down to the grips and special effects people. There is no other series of movies that is better...kudos!  Peace out  '),(1123,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok so I did want to post something else tonight.  A lot of speculatio has been floating around on what to do with Saddam. Should we execute him, should we turn him over the Internation War Court, or should we let the Iraqi\'s decide?  This is one area where I stray from the liberal view...sure the Death Sentence doesn\'t really prevent crime as the criminal is to caught up in the moment to comprehend the consequences...and many crimes happen in the heat of passion or for simple retaliation. I belive in the Death penalty nonetheless if the trials were fair and the verdict returned without the specter of any kind of impropriaty by the prosecuter...however I am not for killing of \"slow\" (retards) no matter how bad the crime...and I disagree with courts mandating a juvenile stand trial as an adult (my logic is...they can\'t join the army, buy a pack of smokes, or go to the local bar...so they are still children).  So what would I do with Saddam? Well I wouldn\'t kill him that is for sure as he would become a martyr....and that would be bad, very bad.   Also I want to make the disclaimer that I also do not believe in torture of any kind...but I will be flexable on this one in this case.  I submit that Saddam be put through the worst of torture however we keep him alive to we can continue to do that for a long while...then we execute him. Let him feel the pain and terror he inflicted on the people of Iraq...and I also submit that he be tried by the Iranians as well as he committed horible terror on those people.  I hate that I have to think like that...but ya\' know sometimes you just gotta get midevil on some people...people that are obviously guilty and they deserve the worst punishment possible.  Again swooping in to talk about/for Dean. The Dems keep on attacking him and have great what would seem like armour peircing ammo...it really isn\'t.   Dean never said we would be better off with Saddam in power...he did say and is correct that we should have never gone in the first place. The evidence at best was sketchy, at worst it was all forged and manipulated.  He was pleased with the capture but felt that we should have never gone in to get him. I guess stupidity goes a long way in this world because they are going against Dean with the lame less than lethal weapons they have in their arsenal.  So to the rest of the Dems...fight Bush, not each other...there is way to much riding on this election.  '),(1124,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok so todays Doctors appointment confirmed my worst fears...I am most definitely very ill indeed. My lung functions dropped almost 20% across the board...oh yeah talk about breathin\'!  He wanted to really bring me in but he has no room at the inn...the hospital is full and what they would do I can do at home...so he didn\'t press to get me in...also the holidays and he wanted me to be with family and save my patience...  They are starting me on an old school anti-biotic...something that my doctor hasn\'t prescribed in many years...at one point it was the most popular drug out there...then because the drug companies weren\'t making money off of it they stopped producing it...which when that happens the FDA drops its certification of the drug so it becomes no longer safe for humans (even though it was used for years)...currently the only place making it is a company in Mexico and its use in the US is limited to Vets...they use it on horses and other animals...so I am going to be taking a drug that the FDA no longer sees as safe...yay!  The only bad side effect (and it is so remote that it really isn\'t even a problem or worry) is that it can kill your bone marrow or rather shut it off so it doesn\'t work.  I am very weary of it all folks. Very.....  '),(1125,1234212,'07,April,2003','2003-04-07 00:00:00','   Well today has just been a shitty day. First of all I woke up and was sick as a friggin dog. I mean I got out of bed and started hacking and couldn\'t stop for nearly an hour. Blood and phlegm was coming out and there is still more down there sitting waiting to come up and cause problems.  I get to go to the docs tomorrow for a scheduled appointment...I think I will end up in the hospital because I can\'t breath...I am so behind and feel so starved for oxygen that it has resulted in a pounding headache and just a general level of discomfort.  The crowing of the Republicans is now becomming deafening...and the Democratic candidate Kerry, Lieberman, and Gephart have all decided to go after Dean.  Has Dean been hurt by this? Nope...the train keeps on rollin\' on. Why? Because Dean was against the war....he was against it because it was unprovoked. Some would say \"well Saddam was in violation of many Security Council Resolution from the UN...someone had to do it\". Ok...it that is your argument then why haven\'t we done something in Israel? Why haven\'t we ridden in on our white horses to stop the bloodshed in Monrovia or The Congo?   Why? Because Iraq has vast oil reserves. Yeah I know it is a cliche\' and every liberal has said that...but it is the truth...or at least the closest thing anyone can come up with since there aren\'t WMDs (even if there are...them being as hidden as they would have to be would mean that they wouldn\'t be that threatening to us now would they...because they would have to go diggin\' for them just to use them)...Democracy...HAHAHAHA!!!! The Iraqi council submitted a plan to implement voting by next summer...it was rejected because \"there isn\'t a good census reference point\"....well they had a census done not all that long ago so they have pretty good numbers to work with...  Everytime the Administration or some Right Wing Nutbar Hawk says it is about something new not but a few days later is that dispelled and a new reason is formulated.  What really gets me is that there are people who think we are safer!!! What planet are they living on? Not this one for sure!  Ok now what the military and law enforcement agencies around the world should be doing is surgical, precision targeting of terrorists and cells...not invading sovereign nations in the middle east where we will happen to piss off a lot more Arabs and push them to be terrorists because of our actions.  Someone made a good point that we won\'t know how these policies will play out ten years or more from now...well we kind of do in all fairness. The war on terrorism will become like the war on drugs...something we can never win because there will always be demand (or with terrorism wrongs that need to be righted in some people eyes). We will spend billions upon billions compared to their relatively small resource pool....yet they will continue to win just like the drug lords. All the while we will be racking up debt and that is what will break us...not attacks or civilians dieing...but a crushing debt.  Many people feel safer...buy I don\'t...not because I don\'t like Bush...or that in all honesty once he gets his teeth into something he is like a pitbull...but in the end he is easily distracted and his policies are fracturing at the seems...  Another thing that I saw today was that people think the economy is getting better. Well according to some economic indicators it may well be getting better...but all of those economic indicators only indicate how those in power or wealth are doing...not your average Joe taxpayer.  Sure the DOW is up...but let me remind people the DOW is like pretend money when you make it...it is only real money when you loose it. The DOW is like going to the races and placing a bet on how companies (or horses) did in the past hoping they will do well or better in the future.  How are these companies getting their stocks higher? Well for one they have been streamlining operations for three years now, laying people off, and reducing back inventory. Worker productivity is still rising...but can\'t continue because people break. Right now it looks like one person is going the job of three...and in many cases it is that simple. So in order to add a job to this economy an employer has to actually create the work for three persons to justify hiring someone to do those jobs. However it doesn\'t mean they will hire more people...actually they will shove that extra work onto it employees and the work will get done and it will appear that worker productivity will go up...it is a visous cycle that will not get any better anytime soon.  People blame our woes on outsourcing. I don\'t think that is so much the case. Sure we have lost many jobs to that...but we must not forget that companies aren\'t beholding to the employees...but rather the stockholders and anything that can be done to reduce costs will be taken so that profits will rise. I can\'t fault that logic at all. It is a pain in the ass and hurts us...but it is a fact of life...we can\'t compete on some of the industries and only remained at the top as long as we did because of shear luck.  I will save \"illegal imigrants\" for later...and I will counter Lou Dobbs on this one...because he is for the most part wrong.  Well that is all for the evening I am pooped and sick...so I am going to get some rest.  Peace  '),(1126,1234212,'07,April,2003','2003-04-07 00:00:00','   Sis is finally here...and I am going to bed cause I feel like shit...  Not much happened today outside of the nonstop coverage of our great \"victory\". Grep Proops who was on Tough Crowd with Colin Quinn tonight put it bluntly \"how perfectly timed to have him seeing that Halliburton was on the cover yesterday and now we catch Saddam...how conveniant.\" My thoughts exactly.  It was a great day for the world...but it would have been better had he been captured by a truly international force. This was an invasion...the same kind as that of Poland by Germany in the beginning of WW2. That was later tried in court as an unprovoked attack and they were found guilty.  I wonder since we are a society of law (ha! what BS!)...I wonder if our leaders would submit themselves to the Hague courts for a similar trial. I doubt it...cause after all god called on us to do this...or at least that is what Bush heard in his head.  The war was unjustified...but even from trajedy we can find some good...and the only good is that Saddam won\'t be returning to power ever. That is a plus...but not something worth going to war over for sure.  Peace  '),(1127,1234212,'07,April,2003','2003-04-07 00:00:00','   My sis is flying in today...yay!  '),(1128,1234212,'07,April,2003','2003-04-07 00:00:00','   Well the past 24 hours has been a real rollercoaster of a ride.  We caught Saddam Hussein. Rather the 4 ID (Fourth Infantry Division out of Fort Hood Texas) caught him...and they caught him without so much as a shot being fired.  When I first got up this morning my mom asked if I had heard...well no I hadn\'t so I guess I woke up to a pleasant surprise.  Yes it was a pleasant surprise. I have been against this from the very beginning, but at the same time I was really rooting for the troops. I have a great deal of admiration and respect for our armed services. They tirelessly protect us and in general the military is very honorable (and would be more so if it didn\'t have the civilian leadership that it currently has).  Back when Uday and Qusay were killed I was a little miffed about that. Not that they were killed mind you, but they could have been taken alive and why the military didn\'t is beyond me. So for them to catch Saddam without any shooting was really a great task and they should be praised because they didn\'t have itchy trigger fingers (as most do now because of the daily onslaught of carnage there).  At first I thought it might be a double...but then they said DNA samples prove it is him (not being one to trust the government I am still a little hinky....but I will take this one at face value).  It is a little upsetting because the timing is perfect for Bush and crew...what better way to shore up your support and approval ratings with his capture (leaving me with a what if question...like did they put off actually capturing him until Bush needed a boost?).  That is why I am a little miffed because he will crow about it and tout it as a victory brought on by his policy.  Some have said that Dean is the loser in this...but I find that contradictory because Dean wasn\'t for Saddam...he was against a war that had little merrit and has turned out to be a bloody mess.  All of this amounts to a hill of beans really when you start to analyze what is actually happening on the ground there.  First is the fact that there are a great many resistors (freedom fighters if you will) that hated Saddam and for far better reasons than we did. This capture will only insure that if the resistors gain control they won\'t be overthrown and killed by Saddam and his cronies. Now they have something to really fight for...their country without the spector of Saddam. So attacks by this group probably will not be lessened, but perhaps increased.  Second are the insurgents. These are basically right wing Muslim nuts that are there to kill Americans. These people come from all over the region. They are fundamentalists and they had no love for Saddam (outside of the fact that he would send money to Palistinian families whos sons or daughters committed a suicide bombing against Israel). Actually Saddam hated those folks and killed many of them whenever he got the chance (he did more to fight the \"war on terror\" in that respect). His capture doesn\'t affect these people...it may perhaps embolden them to do more.  Third is the Saddam Fadayeen (sp). These were the Saddam supporters. They were hard core killers in the regime and continue to be now as they fight the Americans. Now this group could go two ways as I see it. They could lay down there arms and stop the violence (as I hope they will), but they could also step up attacks in the hope they can get their leader back in power. So really this group is truly unknown...I can\'t tell what they will do at all because there are two ways to go.  Now I wish that it would be as simple as catching him and the attacks stop alltogether...but we all know that isn\'t going to happen. The public is behind its troops and we hope for the best and that they will come home safely. Again they all deserve a round of beer on the public\'s tab as they are performing the best they can in the situation considering that they have really limited support from the Civilian leadership in the Pentagon. They need more troops but Rumsfeld simply won\'t have any talk of that.  Again it is marvelous we caught him alive...but we all know the war is still going on and we will see more American (and Allied forces) death before it is all over.  So to our troops...that\'s the way boys and we hope to see you home soon...preferrably in one piece.  Peace (and we love you armed forces...you keep us safe and you make us proud).  '),(1129,1234212,'07,April,2003','2003-04-07 00:00:00','   I almost didn\'t post tonight because I am tired...but I will anyway.  We were busy...did 1600 in business...so today she made a little money. Now we just need lots of those days.  That is really all there is to tell outside of we are bracing for a bitchin\' snow storm...which I am looking forward too...I love snow.  Peace  '),(1130,1234212,'07,April,2003','2003-04-07 00:00:00','   Well at least I can breath right now. Barely but I can breath. All my parts are numb though. I am going to do some treatments and then wash clothes if I can...  '),(1131,1234212,'07,April,2003','2003-04-07 00:00:00','   I bought Noam Chomsky\'s \"Hegemony Or Survival, America\'s Quest For Global Domination\"...of the chapter I have read it is a great book. It reads exactly as he speaks...very deliberately and cautiously to bring you to a level of anger yet not have you boil over.  I think I will like the book.   I am sick. I had to leave work early today...I couldn\'t breath. I would walk maybe ten foot and start coughing and hacking and was totally out of breath. I hate that. I sounded like I had the flu or TB...people were giving me a wide berth.  When I got home I hit the pipe trying to catch up...but it has been slow to recover. Another really strange thing is that my feet and hands are numb...I barely can feel them. I am calling the doctor tomorrow and will probably be going in tomorrow evening.  So I am just worn out. I am going to stop writing now because I can barely string together three words without passing out. I am tired and ill...so!!  Goodnight :-)  '),(1132,1234212,'07,April,2003','2003-04-07 00:00:00','   Lately I have been hearing more and more from Noam Chomsky. Many consider him the premier intellectual of our time. I happen to agree. I don\'t agree with all his thinking partly because I guess I am still young and stupid (nieve\'), however he speaks the truth and frankly if you are a flag wavin\' American with the matching lapel pin what he says may frighten you or drive you into denial.  I am by no means anywhere near the level of understanding of Chomsky but I am getting there. I would really love to see him speak. I have caught a few of his speeches on FSTV and I must say that outside of being monotone he is a wonderful speaker. The way he presents things is not in an alarmest way, or angry...but rather he just presents the information and allows you to choose to believe or not. I think his main goal is to get people thinking. He certainly makes me think. Again he just presents the information pretty much without emotion (I think and get the feeling that he deliberately leaves out passion and emotion so that those feelings will not cloud the mind of the listener. To cloud the mind of someone is easy. It is all in intonation and passion).  Years ago Chomsky recieved much flak for supporting the right to free speech of a man who has the theory that the Halocost never happened (which we all know isn\'t true...it did indeed happen). He wrote a short bit that was put on the flap of the book that basically said that people should be free to speak, and that unless that is protected then debate doesn\'t happen, ideas are stiffled, and solutions may be missed. He never said he agreed with the man or his ideas but regardless fought for his right to speak.  Now the gentleman he wrote this for was a Frenchman (ahhh those evil Frenchy bastards!) and he was sued by many different organizations for his book and ideas. I am sure that I would find the book and the ideas invalid but I still recognize that he should have the right to say it regardless of how tasteless it may be.  From that point on Chomsky (who is Jewish I believe) was seen as an anti-semite when indeed he isn\'t. This kind of blackballed him and set him on the sidelines for TV and as an expert (even though he is the predominant expert on these things). Even under these circumstances he has still been a voice in the wilderness for many who are dissalussioned.  The reason he doesn\'t get on TV all that often in the mainstream is because he carefully words his responces and instead of simplifying things down to a soundbite he actually answers questions in great detail. We all know that the press loves to keep things dumbed down and as simple as possible because god forbid that people actually stop to think and ponder things. It just doesn\'t jive with making simple answers...instead he answers things and simple questions he may answer with an eight minute responce to answer the question and then qualifies it with evidence or stories that he has put together.  If you ever want to read something and learn go get some Chomsky...he has written many books that deconstruct what we see and hear daily. I know I will be looking for some of his material tomorrow at the book store.  Here is a GNN  interview  with him. It is a good read and gives you an introduction to him and his thoughts. Again you may not agree with it all but it will get you to thinking.  That is something that I have been finding of great value to myself. Reading as many sources as possible and coming up with my own ideas...and those ideals are pulled from as far right as the  Heritage Foundation  and  NRA  to the left leaning  GNN  and  The Nation  as well as  Commondreams.   Another great source of daily info that I get in my email is this   The Daily Mislead (Misleader)  and   Bushgreenwatch.   All these are good reads...and yes it is strange that I would have the NRA and The Heritage Foundation. These groups are far right leaning (and by the way I am an NRA member...because that for all their wackiness I believe that we should have the right to bear arms...regardless of what they may be...even howitzers.). Also to know your enemy is good so reading some of the rights ideas from Heritage Foundation gives the average Joe the opportunity to look into what they are preaching...crazy as it may be.  I suggest people read and form their own opinions. It is the only way for people to become truly informed.  Peace  '),(1133,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I am well into downloading Black Adder...a britcom that was absolutely hilarious...it starred Rowin Atkinson also known as Mr. Bean. Adder was his first big gig and he wrote and starred in it...a good comedy.  I have Season 1 and most of Season 2...I am working on Season 3 now. I am getting them for my mom...she loves Black Adder.  Well I am going to read my news and then hit bed because tomorrow I am meeting with dad at lonestar steak house and then I have to be at work by 2...so tomorrow will be a long day.  Peace  '),(1134,1234212,'07,April,2003','2003-04-07 00:00:00','   Just got back from the extended family\'s place. Damn I didn\'t know I had that many relatives...good thing too because of the high chance for genetic matches for body parts...well not perfect but better than the public at large.  All this was on my moms side...I met third/fourth cousins and all sorts of people...very strange.  Tomorrow before work my sister and I are meeting dad for lunch at Lonestar. It will be good to see him.  Well I may update later...who knows.  Peace to you and yours.  '),(1135,1234212,'07,April,2003','2003-04-07 00:00:00','   Well colour me suprised. I got a bunch of neat things for Christmas. I got a new pair of jeans and a pair of Sponge Bob housepants. My mom got me the Alien Quadrilogy (all four movies, two versions of each, and 50+ hours of additional footage) and my grandmother got me a portable DVD player with a 7\" widescreen so that when I go to the hospital I can at least have something to watch. Mom also got me an office chair...very comfy.  Of course I got the obligatory money...some from GM, Grandpa, and my dad sent us something (but it hasn\'t gotten here yet).  All in all it has been a good holiday...and so far not to much stress. We just have tonight to go and if we make it through that all will be well.  Hope everyone had a good holiday.  Peace  '),(1136,1234212,'07,April,2003','2003-04-07 00:00:00','   Merry Christmas to everyone out there...it feels like just yesterday I was a little kid enjoying Christmas with my family in relative peace. Not this year...har har har.  As I get older I find myself becoming ever more cynical to this holiday and I guess it is because I live in the real world compared to most people. I am not belittling their world, but it is this sacharin coated fruit cake of life that people have gotten used to. I don\'t know if that made sense...well it really didn\'t to me but lets run with it.  I really hate the holidays and it wasn\'t because I didn\'t get the G.I. Joe aircraft carrier when I was a kid. I hate them because of what they have come to symbolize.  The holidays have basically been hijacked by one of two groups...the capitalists and the Christians...both equally evil in my book. Both groups attempt to monopolize this one time wonderful holiday and it is sad.  From my years of learning (yes I have learned some things) I know that Jesus wasn\'t born in December. From the shows that have been on lately it is doubtfull he was even born in Bethlehem. The Bible goes a long way to explain why he would have been born there but in reality the system at the time wasn\'t so backwards that people had to travel to their birthplaces to be counted. The Romans were nothing if they weren\'t ruthlessly efficient. So having all these people all of a sudden moving about to get to their birthplaces would have been a logistical nightmare...and again the Romans were efficient.  Another tidbit...people were named (especially in the Mid-East) something like Tobias of Galilee or Pud of someplace. They went by their names and birthplaces...Jesus was known as Jesus of Nazareen (sp) making it likely that he was born there and not Bethlehem. The Bible moves his birthplace basically because earlier in the Bible it states that the son of god would be born in Bethlehem. It is all really interesting history because if you strip religion from it you find that the Bible has many holes in its stories and frankly doesn\'t add up. That points to the fact that facts and events were changed to fit a concept that the early church had.  Capitalism has hijacked the holiday as well demanding that we shower each other with gifts. Who doesn\'t like gifts? I sure love them...but must we save it all for one or two days of the year (Christmas and Birthdays)? The answer of coarse is no. We can give to each other anytime we please but for the sake of relatively new tradition we save it all for those days.  The holidays to me are a time where you get together with friends and family and just enjoy each other. Unfortunately because of all the stress of giving gifts and what Christianity places on people we often loose site of that and just end up treating each other like shit.  The above was all I could come up with as I don\'t have the energy to really talk all that much about it.  We went to the Cirino\'s tonight and had a nice meal and enjoyed each others company. It is nice to have friends like that. I consider them my extended family as they do me.  I am tired already and today offers to be just plain crazy...but I hold out hope for some sanity in it all. We will see.  Just a quick note to the person who responded in my comments section below. I never have wished anything bad on anyone nor would I...in fact I have very few people that I flat out hate and Brandon surprisingly isn\'t one of them. I feel indifferent towards him not really caring one way or the other. I do wish him the best in his life...because he will need it. I do of coarse think that the punishment should have been more equitable but I can\'t change that nor would I have made stories up to get him in any jam. That isn\'t my style, never has been, never will be...even for the people I dislike or flat out hate. Why? Because I know that it will come back to bite me in the ass...plus I don\'t have the energy to let things bother me all that much.  I also wanted to quickly respond to one other thing. Brandon may have a great job where \"he makes more in a day that I see in two weeks\". Good...again I hope it is true because when he wants to be he can be very talented and driven (something he wouldn\'t regonize in myself...or any of his enemies) and with the right breaks may become someone of importance in the future. It is probably true that he will make more than I do partly because of a degree but mostly because outside of minor ailments he can work. I couldn\'t hold any job of merrit not because I am stupid or undeducated...but because I am often in the hospital...so working is difficult when you can\'t be at your job. So if it true great if it isn\'t I hope that it does become true because lord only knows that he is an unhappy individual (similar to myself in many regards).  Side note...I have written a lot for my book. I don\'t know how it will end or how it will play...but it will certainly be interesting to finish it. I even have a working title...so it looks like I can make it fly at some point. The hardest part of writing is that things flood in all at once and it is sometimes hard to get it all down, but I am working on that. I currently have two seperate things I am writing, one is on me, the other is fiction. So we will see.  Anyways people. Merry Christmas and I hope your holiday is a happy one. Don\'t forget to sign your organ donor cards because god forbid something happen, if it did then you can spread the greatest gift of all. The gift of life. Look for the link and get your card made up and ready.  Peace  '),(1137,1234212,'07,April,2003','2003-04-07 00:00:00','   Today has been an extremely suck ass day as I spent most of it sick on the couch on the verge of puking. The meds are really working a number on me now.  My cough has turned from a dry unproductive one to this TB sounding death rattle...and I am weezing a lot. If you were to listen I am sure you would hear my lungs gurgling with shit in there. It really sucks ass.  What really sucks is that I still have to wrap presents and on Christmas day instead of it being calm it will be absolute madness. We have dinner here, then at grandpa\'s, then Cirino\'s, then cousin Max\'s (gandmothers cousin), then we have to try and catch dad...but I think that will be on Friday...which I have to work.  This holiday really blows because not only is it shitloads of driving...but I am also supremely ill and the only one who can get everyone anywhere....as Michelle simply doesn\'t know the area so I am her driver everywhere.  I will be glad when all this shit is over and I can go to the hospital to heal...my body is in desperate need of that.  I am afraid I won\'t make it through the next few days and by the end of it all I will be flat on my back, barely breathing with a headache.  Oh yeah...christmas time is fun...time to rejoice....time to pay homage to the great gods of capitalism.  '),(1138,1234212,'07,April,2003','2003-04-07 00:00:00','   Well today was a long fucking day. My sis and I went out to finish some shopping. I hate that...the malls were madhouses and the people were nuts.  Driving could have been easily summed up as bumper cars...lots of little accidents and many idiots out driving.  I am still pissed.  '),(1139,1234212,'07,April,2003','2003-04-07 00:00:00','   Well fuck me. I found out what Brandon got today...same thing that I got. 6 months home detention and five years probation.   Article   Now I am a bit miffed about that...you see he was far more involved in it all than I was and frankly I am pissed that I got as hard a sentence as I did seeing that I was just transportation for this whole mess.  I didn\'t get one single thing out of this meanwhile he recieved cash, tons of hardware....hell he probably still has the SCSI hard drives he scammed.  I was expecting him to actually see jail...and while I would never wish it on anyone I am pissed that he got the same sentence I did...I mean come on.  Oh well my home detention is over now and I am pretty much free again...its not like I can really do anything anyway...I am too sick now.  I guess that is what I am most pissed about...that the past six months that I have been on home detention has really been the last of my good health left and I have slid into the downward spiral all CF patients get into at some point. I would like to have had some more good times but seeing that I was on home detention I couldn\'t.  In the grand scheme of things I know that Brandon didn\'t learn his lesson at all. I am sure that he is all pissed that he got caught and that he has to be punished for it all. He will blame others, namely me, for him getting any time at all.   I would state here and now that I would have testified against him without reservation. It is just a shame that instead of the prosecutors actually doling out punishment in a fair and equitable way they just slapped him on the wrist.  The only thing that I wonder about is the fines and restitution he has to pay...because I have been paying on it all this time...which it hasn\'t been much but I still have paid on it. I doubt he will if he even was sentenced to pay.  I still wonder if I am to pay a third of the 96 grand that he profited from or do I pay all 96 grand. Basically that is what the prosecution said he stole and I transported with John as the guy who found drop points. I think that it is meant to be split three ways...but I don\'t know. However even on that I am pissed because everything that I got I handed over. Brandon didn\'t...plus he actually profited from it by selling Sound Blaster cards, RAM, modems, and I think video cards at one point. He got all that stuff in this deal and then sold it to people on Mack 3. So he made money...plus he ebayed some things off.  I am just miffed because he was the hub here and yet he got the same punishment I did as being the transportation.  Oh well at least I know what I did was wrong and won\'t be doing that again. I just wish that the punishment was more equitable.  '),(1140,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok I officially hate Christmas...actually it took me longer this year to get to the point of hating it...so I have done well.  I hate it because of the all the fake people who pretend that it has some great religious meaning when in fact the \"birth of christ\" probably didn\'t happen in December at all...instead Rome worked a census in October and November for tax purposes. Mary and Joseph travelled to Bethlehem for the census. So unless they got really lost (which hell Moses wandered around in the desert fourty years...can you imagine all the whining of the women for him to pull it over and get directions...and these women were Jewish so the whining must have been on the level of \"oh my god I want to stab my ears because the pain would be a welcome relief to all this bitching\") and missed the census then \"jesus\" was born in October or November.  Also these semi-religous retards almost always are the ones who are the most evil towards others in the Christmas holiday...so go figure.  The second reason is if they are going to celebrate it as \"the birth of christ\" then they need to stop the crass comercialism that goes along with this time of year...or leave religion out of it and just have it be a day that we all give gifts to each other.  At my little so called job as you folks may well know I work with Indians (India) and Jews from Israel so neither group really celebrates the holiday as most Americans do. Balu, the guy who runs the stand next to us (from India), is always asking questions about the holiday. Now I guess he wants to understand it from a Christian perspective and frankly I am not really good at explaining it in normal people terms because for years I have been against the holiday season. Also I am not a fanatic christian so I really can\'t speak with authority on the subject outside of my own beliefs.  I am a horrible consumer. Companies hate people like me. I don\'t shop and buy...I don\'t consume unless I need too. I know what I want so I get it when I need it (or can afford it) and the christmas holiday is the worst time of year for me because I don\'t shop.  This year though I did get gifts...I hate that but I know I should so I did...got most everything I needed too however I have a few odds and ends to pick up. Overall I will have spent maybe 150 total. Not bad for getting gifts for 12 or so friends and family.  I had to resist the urge to get things for myself...it was hard but I managed it well.  I wanna say something about healthcare...and actually it is from Ted Kennedy and a speech he gave after the Medicare thing happened. Here is the qoute as closely as I can recall. \"Since the sixties the republicans have been trying to destroy medicare and privatize it. Well if privatization works then why do we need health care from the government in the first place?\"  Right on the money there...why if privatization works so well that we have 40+ million without health care? The fact is you can\'t privatize this because it goes against what a profit driven company needs to be doing...any company.  A private insurer has to make a profit for its investors, and this means that they need to pay out as little as possible for health care. To do that they deny claims, they deny clients, the deny anything or anyone that costs them money.  As a person with CF I have run into this time and time again at jobs where sure I can get the insurance but I have to wait for up to five years to be covered on CF (and the can still deny it to you) and I am paying...and many insurers exclude pre-existing conditions...so in the private system I cannot be covered. That doesn\'t work does it?  It is the same for millions out there as well...so sure there may be many that don\'t have insurance because of no job...but the number is growing in the population that would have health care problems that are considered pre-existing.   So what is the answer? Nationalized health care. Period. All current moneys that companies kick into insurance bennefits for their employees plus their portion or copay per month goes into a nation pot. That pot would also include Medicaid/Medicare moneys that are allocated in the budgets. Only a minor adjustment would need to be made in the budget which would include probably more money from the Gov...but hey thats what we pay taxes for right?  While like every other American I hate taxes...partly because we don\'t see the money at work. However is I was guaranteed health care, better roads, and schools I would gladly pay up to fifty percent of my wages in taxes knowing that I will see some of that money at some point in forms of service to me.  Kucinich has an interesting view on it all...very similar to mine, and much better thought out really.  While on politics I must say that it would appear that the DNC is shitting big bricks over Dean. I firmly believe the DNC does not want to run any viable candidate clearing the road for Hilary in the 08\' election...as it stands the opposition is weak (although growing grass roots style, offering a peace treaty the other candidates).  The only person I fear they might even run hard would be Wes Clark...and frankly he isn\'t that impressive. He would have a harder time winning because he really isn\'t all that eloqent in his speeches. Plus this is the candidate that Gillespi (DNC National Director), and the Clintons want.  Dean is the outsider in this whole thing. No one really nows what will happen frankly.  Good news though Dean is on his way and looks as if he will get the nod from Iowa and New Hampshire soon enough.  The DNC has to figure out what they will do...should they throw the money at Dean and give him a fighting chance. Also they need to imbrace this new Democrat. He balanced budjets in Vermont and 97 percent have health care as of now. The man is no fool.  On one side you have Gillespie, Bill, and Hillary Clinton. and these are but one piece of the pie. On the other side is Dean and Gore...both with much clout...so I thing it will be intersting snough.  If the DNC rejects Dean they will have made a huge mess even worse.  For running mates I can up with some names for it all. McCain...sure he is a republican but is old school and knows the system.  The other lark I could say would be Kucinich...in doing this it would get someone semi-honest in governent. By doing we would see green canidates make a good showing.  Either pick would be ballsy. McCain probably wouldn\'t take the job...but it would be great if he did (after all McCain recieved many honors in Vietnam don\'t know which ones off the top of my head...but they are there).  With McCain you will catch those Repubicans that are swing voters...and Dean needs them.  For Kucinich...while he many not bring big money he will stand by his policeics. His nomination would bring out the geens that would vote for Nader as even Nader recognizes the little guy and supportomg any of his environment causes and fly  Kucinch is a tad scary from time to tme...he can scowl with the best of the. I have no doubt he will always be more green (newer party for the environment.  Either way he chooses it will be wise because he is a smart/tough cookie. Kucinich gets the green vote back into the fray, and McCain is beloved by the people here. So to me either pick he may be great. He has the attention of the liberal mess up there.  McCain would help in bringing a moderate view to the state deparment. He has the chops to do it...so we should persue a change in foriegn policy.  Whomever gets the job, will be woried.  So that combined efforts can get us back to normal.  Rember Bush 1, Dems don\'t like a long term stay.  Vote for the Democrats at all costs just to get that madman out of controk here.  Peace  '),(1141,1234212,'07,April,2003','2003-04-07 00:00:00','   Lemme say this was a long day. A very long day. A busy day.  My dog got sick so my sister (who will graduate in May from Vet school) took him to the office that she did her externship with here...he got something they don\'t know for sure what it is but all the treatments are the same...so he will be ok.  My meds are all finally in and they just kick my ass. I have never been so worn thin. I will be glad when the holidays are over...because I hate them. People are just so mean to each other. Plus I don\'t like Christmas.  Well I have to get up on the morning so goodnight all.  '),(1142,1234212,'07,April,2003','2003-04-07 00:00:00','   Well over at Brassmask\'s site we have been having a spirited debate on whether or not we supported Saddam and Al-Queda and Bin-Laden...which we have supported all of the above. It can be proven and frankly anyone who would deny that we helped them out would have to be foolish.  It has been interesting because there is someone over there that has basically said it is all liberal lies that we supported them. I really don\'t know how someone could make that claim seeing that the government has even said it supported them both and made a mistake.  Some people I guess just have a really short memory.  Something that I posted over there and I have not posted here yet...I believe is this.  We have a real problem developing over in Pakistan. President Musarif (don\'t know the exact spelling) took the country in a coup in 1999. He did it very wisely because Pakistan had by then become a nuclear nation. He waited until they had them developed and then took over. Imediately the nation was put on our terrorist state list but we didn\'t really do anything to stop it all because they were a nuclear state.  We did filter much money through Pakistan and their ISSA (similar to our CIA) to the mujahideen in Afghanistan and the Taliban. We did this even after we placed them on our lists. We continued to put money through there because it was seen as an asset to us.  At the same time we were in direct negotiations with the Taliban government in Afghanistan to try and get a pipeline through the nation that would eventually lead down to a power station in India owned by Enron. This pipeline was turned down by the Taliban not all that earlier than 9/11/01.   However we still pumped in millions a year in aid to help them destroy poppy fields that supplied opium. We did this as part of our war on drugs. The money of coarse was used for other things but we will never know for sure where all of it went.   Well during all of this the Taliban and Al-Queda were relatively neutral towards Pakistan and Musarif...they didn\'t like them but didn\'t mind them...they just sort of left each other alone and it was a good arrangement.   After 9/11/01 Musarif aligned himself with the US and that basically knocked him off the fence for Bin-Laden and crew. He is now seen as a threat and they will do whatever they can to take him out. In fact he has had many attempts made on his life over the past few weeks. All attempts believed to be being made by Radical Islamic groups.  Now this is a problem because sooner or later he will be killed and the country will swing very violently over to a Islamic Fundamentalist government. This government will be in step with Bin-Laden and since Pakistan is a nuclear state Bin-Laden will at long last have access to nuclear weapons.  Now with our current precarious position in the middle east we really are not in any shape to take on this new emerging threat as well as continue our operations in Iraq and elsewhere. Our military is extremely overstretched and we would have to re-instate a draft to make up the numbers we need to fight this new fight.  Also we have a long standing tradition of not actually attacking nations that could possibly pose a threat to us as Pakistan may be actually able to with its nuclear arsenal. So we have that dimension to it all.  Now once Bin-Laden has nuclear weapons he will strike at three targets and you can quote me on it. These three targets will be Riyadh, Moscow, and Washington DC.  Riyadh because Bin-Laden holds that Saudi royals responsible for allowing the infadels to remain in the holy land. He is royally pissed off about that and has been stepping up efforts in Saudi Arabia to overturn the government and move it towards fundamentalism. They are already close to getting their way as the Saudi Royals have at best a precarious hold on power. So we can safely assume that is one of his prime targets. If he gets them the whole of the middle east will essentially fall into Muslim Fundamentalism.  Moscow he will attack because it is a relatively soft target and frankly Bin-Laden has even much more deeply rooted hatred of them than even the US. It will be revenge there. Plus the Chechans need the help and they are brothers in arms with the Muslims.  Washington DC is obvious. If he were detonate a nuke here we would surely fall into great disarray. We may not fall...but guaranteed our way of life will be over for a long long time and we will have serious problems for a lot of time.  Now the Bush administration will say that we can detect these weapons coming in, or hell any administration would say they could stop it. Well the plane fact is we can\'t detect nuclear weapons coming into the nation whether by cargo plane, ship, container, really anyway. The reason is because when you take the material and make it suitable for a weapon you essentially make it very stable (you don\'t want it being jossled and just blowing up) so it is really stable, which when stable like that it doesn\'t give off much free radiation. Also the weapons are also somewhat shielded for the protection of the people that have to handle them.  You can\'t detect them because they don\'t give off much radiation and as such attempting to detect them is pretty much impossible.   So if they were to get their hands on nukes there is no way we can actually protect against them getting here unless we lock down tight and frankly that is not in the same interest of freedom.   So we have some choices to make. Mark my words it is going to be hard over the next few years not withstanding whomever as president. All we can hope for is that we will not have a reactionary president that will make things worse and instead have one that will keep his head and make good sound decisions.  When the attack happens how will we respond. Under Bush we would declare nuclear war most definitely. We cannot do that because of the protections that are built into the Russian and our system of nuclear arsenals. We launch one off they launch one, basically locking us into a nuclear war that may not be able to be stopped. It is left over MAD (mutually assured destruction) protections built into the system. We may not intend to target Russian targets but their systems are not designed well enough to determine that for sure and they might launch on the fear that they are being attacked.  Regardless if we do attack the middle east with Nuclear Weapons and Russia does not repond we will still be screwing ourselves hardcore because like it or not we need the oil from the region for a time before we can transform into a hyrdrogen economy.   All of this is speculation of coarse...but I feel it is the most likely scenario that we are looking at and however you look at it it just looks bad for everyday americans and the common people of the world.   I hope I am wrong but I fear I am not.  May we all come out ok in the end...may we elect someone other than Bush so hopefully we won\'t retaliate with nuclear weapons making the situation worse for us all.  '),(1143,1234212,'07,April,2003','2003-04-07 00:00:00','   I have watched four DVDs over the past 12 hours or so. Van Wilder, Dude Where\'s My Car, Planet of The Apes (the Tim Burton Remake), and right now I am watching The New Guy.  This little DVD player has been a godsent. I have been watching so many movies it is rediculous. I have watched almost all of my movies that I have. I have also gotten a few more and watched them as well.  My mom also got me a subscription to the Wal-Mart mail rental service for four months I am just waiting for the package of stuff to come in so I can start getting movies.  I think I will also subscribe to another movie club like I have in the past. Just so I can get some more movies at decent prices. It will be good to get into that again because I did that once and got a lot of movies that way.  Well I have a list to do for Jesse so we can build his new system. So peace out for now.  '),(1144,1234212,'07,April,2003','2003-04-07 00:00:00','   Yesterday I actually had written out this big ass post but it was lost to the internet info monster...the post didn\'t take. It was about lies that the administration has told and then I was going to document them one by one.  You know what I am not going to do that because if you read the blog here I am sure that you are already aware enough to do your own research and see the lies for yourself. I am always willing to help research if there is something particular that you are interested in...but I am going to give ya\'ll enough credit to do your own research.  I have however gotten to thinking about something I said. It goes back to something that I have wanted to do for some time now but never gotten the balls to do it.   When all this crap started happening I wanted to start a website and basically do what I do here a lot. I wanted to bitch about the administration. I may still do that site or spin out the idea because it is universal and works for any administration really and can be used as a forum for change.  However the other night I stumbled across an idea. You all read it. So I started to look into it today to see what I would need to do to make it fly. Well I have found out some information and really it will be a little work but I can do it. Especially seeing that I have wireless internet access from anywhere now...and pretty soon the new rooms here at the hospital will have fast internet access so that will also facilitate the idea.  I have some things to do but I can make it all happen. It will take a few dollars to get things rolling but I have credit cards and I will have some income here soon. Hopefully I will have an anouncment here soon about it all. It is something I think can fly...it will take some time and a lot of work...well a little work each day...but it might help out.  Well I am going to look into some more things and try and get in contact with some people.  I will be calling on ya\'ll when it all gets started...I promise it won\'t be anything hard or time consuming. It will require only about ten minutes and no more per day.   Peace  '),(1145,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok Happy NEW YEAR! Enough of that joy crap.   My mom just left for the evening. We had a really good talk about journalists and how shitty they have become in general.   Look at all the national journalist leaders. Tom Brokaw, Ted Koppell, and the others that are lower than the \"heads\" but certainly should be competant journalists. Maybe Charlie Rose is decent and we know that (or at least I think) Bill Moyer is probably one of the few left out there with balls.  However the vast majority of the \"journalists\" on TV are not even worthless, worthlessness denotes that they lack value and for them...these people on TV they don\'t lack worth...actually they are very valuable to the cause of mis-information. Their job has now become that of mis-informers and has great value when it comes to keeping the masses out of the loop on what is real.   Many people on the outside and in indy-media have a real problem with the \"Joe Millionaire\" shows, or the other reality TV shows...I think they suck and shouldn\'t be on TV but I understand that you have to entertain people from time to time and frankly if the News Departments are doing their jobs then let the people have their schlock TV.  However TV News isn\'t doing its job...not even nearly. They do even worse than not do their job they mis-inform people.   On mis-information I can pull one thing right from the headlines is the \"drop in jobless claims\". Personally I find this offensive. Why? Because it is mis-information. Its not a lie so much...because lieing would be unethical but mis-information is ok. They report that unemployement is down to 5.4 or so percent. That isn\'t correct it is actually closer to 9.4 or so percent.  Read Between the Lines and read where long term claims have actually grown.  What the reports neglect to tell is that many have stopped looking for work and are no longer counted by the system, and some go so far as to say \"employers may be thinking happier thoughts as less people are being laid off\". Well if they are thinking rosier than I would hope someone would tell that to the people still loosing their decent paying jobs.  In addition Congress and the White House (both Republican Controlled) failed to renew unemployement benefits before they let out for the holidays. So literally millions of people were to loose their benefits around the 21st of December.  One source ...good luck on finding other newer sources of info for it. Now this isn\'t the first time the Republicans have balked at extending the benefits the let them lapse last year at this time only to renew them around the 15th to 20th of January after initial economic bright spots turned out to not translate into real jobs. It happened in May as well. It is a trend. I am sure that the benefits will be demanded and eventually won, but the point is that in our nation...the best country in the world according to our leaders this kind of thing should never be a problem. Especially when it is known that the economy is truely having problems like it is now and has been for a little over three years now.  So we know they don\'t really tell us the whole story. Now what about not doing their job.  Well the big guys haven\'t been asking real questions for a long time of any of our \"leaders\". I mean right now there is so much to ask yet no one seems to be asking the questions. Why? The common reason that some people would say is that if they asked the questions then they would be not invited back or denied entrance to the press briefings again.   My question or comment back to them is why don\'t you all drill the person at the front with the same question over and over and over again until you are all banned or they answer the question. Hey...novell idea huh? One reported did it at Centcom during the Iraq war and was pushed out of the briefings by the command...and instead of the reporters and news people sticking up for each other they let it slide like a bunch of spineless wimps. In reality they should have all stuck together and just left the place because they weren\'t getting real info anyway.   That is what we talked about. I took a bigger picture than what we actually talked about...we talked about a story a little closer to home about why things happened they way they did and why reporters haven\'t followed in any real way what happened. You would think they would at least make a phone call to someone involved in the case to get a comment...but no.  Anyway...regardless of how or who it applies too...I think that reporters have simply done a shitty job...and it is now being taught in schools and universities. I mean tell me one halfway decent University newspaper that is actually doing its job...or a faculty that will stand up to administration pressures should they come down on a report that may be unfriendly. Now I know all of them will say \"oh yeah sure we would\" but when has it actually happened in recent times. If it has great...but I bet it hasn\'t happened all that often. So students aren\'t really being taught to do what is right and that is fight for the truth.   I worked at a school paper back at my community college...and while I would love to say that if I had ever gotten a really juicy story I would have been backed up...I know that I wouldn\'t have been in reality. Partly because of the reality of the world. My teachers had families that partly depended on their income from that job. So to choose me over their family is dumb and I would expect them to choose family over me or my story but I would find a way to get it printed. So I understand why but I think it should be different.   So I would hope that people out there would bombard these guys up on the air with questions. Maybe just one question a day. All the same question, one a day.  Thanks for spending some time reading tonight. I want to wish everyone a happy new year and I hope that better things await us come next year. We have had a loosing streak that is about to break one way or the other I can feel it.  Goodnight. May your night be wonderful and safe.  '),(1146,1234212,'07,April,2003','2003-04-07 00:00:00','   I noticed that I has spelled something wrong over in my description. The word One...jeesh.  Well I changed the description anyway...cause that is what this blog is becoming...an Indymedia spokespiece...or at least I would like it to be.  I have to learn how to spell...maybe I should get HOP...what\'s HOP...well that is Hooked On Phonics new brand or branding name...just saw the commercial for the first time early this morning.  Well I signed up for a bunch of newsletters from some more places so I will be getting maybe 10 or 15 newsletters a day...that should give me plenty to write about as I get into them.  I am going to rework my links over there to reflect more catagories...as I now have political really absorbing what would be activism and newsletters catagories when I really only have a couple of truly political links...I may give links to the candidates over there...as I truly believe in being informed so I may give them all links (although many of the sites are Ra-Ra sites like Bush\'s and Lieberman\'s as well)...but in fairness I need to give links to all...although I have not made a final decision on that...give me your feedback on that to let me know what you think  nevertheless there will be a shakeup over in that area to make it a little easier to get around in...  Peace  and oh yeah...Happy New Year.  '),(1147,1234212,'07,April,2003','2003-04-07 00:00:00','   Blogger is still screwed up I can\'t get to my own page right now...maybe the rest of the world can. I can obviously post but I can\'t pull my page up. They had a DNS problem but have it fixed now and should have everything restored soon.  GM was here and brought me my drugs that I needed and the visit went well and at least I wasn\'t stressed by it all. I like not being stressed....I wish that I could enjoy that all the time.  Well I am off to surf a little more. Something of note...I got all my bills paid freakin\' early this month...credit card, phone, and car all paid early...my account is nearly empty (although I have a Christmas Check at the house for 250...) but my monthly check should be credited by the 4th or sooner...plus I have a last paycheck and taxes that are being returned to me...hahaha...so I have a few bucks coming my way.  Oh yeah I also have to get over to the SS office to get that code fixed on my account because I am paying for Medicare when in reality there is a program here in PA to pay the premium for me...I have been paying it for over a year now and since it is going up I think I should get that fixed...once I have that fixed I will get back the premiums that I have paid over the past year...that will be about 650 back to me.  Basically I will be able to be out of debt (with the exception of a car note) in a matter of a month and still have plenty of money (which I could be out of debt this month if I so choose...but I am \"rebuilding\" my credit by leaving small balances and paying them each month).  I will post again later...after my mom stops or I eat...and perhaps after I visit with Danielle...I need to get down the hall to visit.  Peace  '),(1148,1234212,'07,April,2003','2003-04-07 00:00:00','   Oh yeah...forgot to give the link to the Commondreams article.   CommonDreams article on Hawks new book.   There you go...sometimes I forget the basics. ;-)  '),(1149,1234212,'07,April,2003','2003-04-07 00:00:00','   Well Blogger or Blogspot was acting a little goofy here for a while but now seems to be fixed.  Some interesting stuff over at Common Dreams. It seems that Perle and Frum have written a book on \"how to win the war on terror\". Oh yeah they have it all figured out. Just about as well as Perle had the war on Iraq figured out.   It calls for some radical stuff like taking out Syria and Iran as well as blockading North Korea and preparing for war against North Korea.  Now even though Powell is not always on the right side of the issues he is still the only one in the cabinet that at least understands things aren\'t as simple as attacking this and that. He realizes from time to time that you have to different things. He has been the only one who has pretty much stood in the war of an all out war by the Bush administration on the rest of the world. He has again stepped into the breech and is protecting all of us again.  Perle and Frum have written this book and sent it out into the public sphere just as Powell is at his weakest, remember he just had surgery for prostate cancer a few days ago. This is a tactic to force the administration to act.  I have a feeling that this coming year that we are facing is going to be very rough and perhaps we may see some real violence.  If you look around the world (even to our neighbors to the north Canada) no one else is really this frightened and on edge as we are. We have been whipped into a frenzy and many who are misinformed are really scared. So who knows.  Something I wanted to mention last night but just ran out of steam for was this new requirement by the US Homeland Defense thing that air carriers from foreign countries now have to have armed air marshals on any aircraft coming into the states. Now if they don\'t comply they will not be allowed to enter our airspace.   Ok I kind of understand the rational on this one but I have to again disagree with it all. We are saying to the rest of the world \"hey you fuckers can\'t get airport security right, and since we are the best at it (ha ha) comply with our mandate or fuck off\". Well that is my interpretation of what they have said. It is a little crude but it gets the point across.  Some places are already doing this and it will not be hard to do, but some places are barely even able to put fuel in planes so the costs will be prohibitave as well as take some time to come up with air marshall programs (which during that time we would provide our own to help out).   I find this all very offensive because we are again forcing something on the world based on poor information at best and especially seeing that not one of the hijackers started on a plane from overseas....they started from Boston so international air marshall wouldn\'t have done squat.  Also of issue is that British pilots are objecting because they think it is a bad idea for anyone to be armed in a pressurized tube flying miles above the earth. Bullets and planes don\'t mix well. They just don\'t think it is a good idea.  Another thing that I am cheesed off about is that we are saying that we are better than everyone else at security. Oh yeah that works! We are the masters of air security. Shit we couldn\'t secure a bag of dog shit shot from a catapult. Well maybe not that bad...but close. The Europeans have been doing security better and longer than us and have a pretty decent record as of late. So to tell them what to do should be insulting to them.  Another thing that is being placed out there is new Bio-metrics systems that record pictures and finger prints and compiles data-bases of all these things and allows us to see who can fly and who can\'t...well a good idea if it applied to everyone (and even then you have to consider privacy issues)...but it only applies to certain countries and it so happens that all of the nations that the hijackers started from (France and other european cities) then they would not have had to go through the system at all. So it wouldn\'t have even caught them. As it is being implemented it is only meant to record and catalog people of the mid-east. For me it seems that it is the best way of figuring out how to number everyone in the mideast...or at least get a start.  There is so much happening before the new year even gets started. Jeesh...I can feel it will be busy.  '),(1150,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok I have a lot of things that I want to tackle tonight so bear with me as while I am in the hospital I have nothing much else to do.  The first thing that I want to discuss and give my opinion on is this new law (well it is a law that has been on the books and been available for use since 1967. It is only for Ohio but I can imagine other states will be watching to see how it works out.   News Story   Well in Ohio some judges are going to start using this. If you didn\'t read the article it is a special license plate for people convicted of DUI. Now this special tag is different from all the other tags in that it is yellow with red lettering.   Now from a completely common sense view I see that \"hey this might be good and might cut down on accidents or DUI\'s\"...but then I get into my civil libertarian mode and I can see that there is a real problem with this rational.  Ok lemme just state a few things. Most DUI drivers probably have done it more than once and really don\'t see much of a problem with it all. Even when caught these people continue to do it inspite of the law.   Now I can honestly say that I have done it a few times but out of those few times there was only once that I was truly a menace and should not have done it...and looking back I would not have done it. Actually happened recently when I was living in Indiana and my roommates and I were at a house party and I was the only one that was barely standing at the end of the evening. Fortunately we only had to go a few blocks and it was very late at night and I have always been a good driver...so we made it without incident. I will say I will never do that again because when I got back to the apartment I got out of the car and screamed \"I was way to drunk for that!!! I will never ever do that again\"...then I profusely thanked the gods above because I really should not have done that.  Ok with that little story out of the way I will go back to this law. It gives the judge the power to convict DUI people to have this special license plate. Again not wholely unreasonable...but when you consider that if you are someone with one of those tags on you will attract attention not just from your fellow motorists (which hey it is supposed to warn them that you have DUI\'s in your past) but it tips off police agencies.  Ok so what it tips off the police? Well think of it this way. Police when they pull you over and want to search your vehicle they have to have probable cause to carry out the search. With this tag not only would the vehicle garner additional attention from law enforcement but also set up automatic probable cause for police to just search your vehicle.   Alarmest am I? Yes on this definitely.  I think that in reality you could make a good constitutional case against this law using privacy rights and 5th ammendment rights. One is privacy and if you have this tag you stick out like a sore thumb and people know your private problems...but the worse violation is against the Fifth Amemdment. Basically it says that you have the right to refuse to incriminate yourself and cannot be compelled to. This tag could be seen (from my eyes) that you have just tossed that sucker out the window. Having the tag on the back automatically prejudices any officer who pulls you over to go through a whole batery of tests (sobriety mainly) and then the officer would want to search your vehicle which because of your tag they have probable cause so you can\'t stop them.  Here is the perfect storm example of this. John Doe got a DUI conviction and still has his license but has the new plates. John Doe also runs drugs from Cincinnatti to Columbus. You are cruising along and not breaking any apparent laws. You aren\'t speeding nor driving irratically but you get pulled over regardless because of your tag. Now they run the tests and find nothing but continue to search the car \"because you never know\"...and even if you were to say \"don\'t search my car\" they can say not only from that statement but also because of the tag that they have probable cause to do it. Of course they tear the car apart and find the twenty pounds of pot.  That is the scenaro I see. To me that is a great problem.  Another thing is more dissendence in the Rank and File Israeli armed services.   More troop refuse party lines.   Basically these guys are the uber elite fighters (more secretive than SEALS, Green Berrets, SAS...and likely better trained) of the Israeli Defense Force. The IDF has had some problems as just recently 27 pilots have decided against flying these assasination mission which are illegal under international law. So this uber force has said no we aren\'t going to continue to kill palestinians and deprive them of basic rights.  This is not only a blow to the extreme right wingers in Israel but also here. America has to come to grips with how we are going to deal with our child over there.   Now it is not anti-semetic to talk critically of Israel\'s issues as long as you acknowledge their right to a state. While I have different feelings on that (mostly because they are essentially setting up a racist gov. that doesn\'t want anyone outside of their ethnic background to be a part of...that is a whole \'nother can of worms) I do recognize they are indeed a state and are soveriegn. They have the right to exist is as plainly as I can put it.  However to blindly support Israel is just plane dumb. It is like saying to your kid \"its ok you killed you cousin, and we won\'t mind if it happens again\". What has been going on over there has been a systematic attempt from both side to wipe each other off the planet. That is the key...both sides.  You cannot deny that the Palestinians have done some horrible things that are simply barbaric, the same goes for the Israeli\'s as they have done horrible things to their brothers.  So why do we as Americans blindly support Israel and all its policies when clearly the support for those policies is surely waining in their own nation? I can\'t give you an answer outside of that we here in the states are afraid to question it for fear of having the Anti-Defimation League crawl all over anyone who talks out of turn. They did it recently to Dean on the mere suggestion that we may want to take a more even handed approach to Israel/Palestine.  Today on my way down to the hospital I stopped to get a pepsi and some three musketeer bars at sheetz in our little burb. Low and behold I saw a H2 there. If I had been in better shape I would have confronted the owner and asked him how he felt about the 400+ lives that have been lost so he can run his vehicle. I also wanted to ask him if he was compensating for a small penis. What is sickening is all the commercials on tv for them lately. I think what is happening is that people are getting in there and looking at them and then they are told of the Tax Credit for those vehicles if they register them under a business. If a person buys a big SUV over a certain weight there is a tax credit that you can get if you register it under your business or phoney business (its not hard to \"start\" a business). You can actually recieve the total cost of an H2 back in one year as a tax credit...basically the Government pays you to buy the vehicle.  That really pissed me off. I tried to get my camera out so I could get a picture of me flipping the H2 off and the owner as well. I believe this guy lives in Springdale and I may track him down and tell him what I think of his hummer and maybe insult the size of his obviously small dick.  Oh I have more don\'t you worry...I have a lot to say about the Mad Cow thing and many other issues and I will get to them over the next few days...you just wait. I guess I have been waiting for the right time to get to it all...  Peace...and if you see people driving a H2 get a picture of you flipping them off and lemme know and I will give you a site that you can have it posted on. A site dedicated to the hatred of the H2 and its complete lack of anything remotely needed.  '),(1151,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I am here and now I am waiting for the fun to begin.  I have something to rant about later on. I will do that once I get started with all the fun.  Peace  '),(1152,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok so I am headed to the hospital in a bit. I am just waiting for clothes to be dry, get a quck shower, then out the door.  I also cleaned up my room a bit. Had a run in with my grandmother in that I just needed for her to open the doors so I could take trash out and she started to say \"I\'ll get it\"...well the bags were leaking and I asked her again and she snippily did it with some parting words...then when I came back in and saw the drips out of the bag she was bitching about that...ok if you had just opened the door then there wouldn\'t be this big mess.  I have learned over the years that I am never right when it comes to how I deal or don\'t deal with her. No matter what I do or say it will be wrong so I really don\'t care anymore. I know that even if I do it the way she wants I will still get yelled at. So I have just gotten to the point of just not being around or just ignoring her when she gets moody...because again no matter what I say or do I am wrong..ahh well such is life.  I will update from the hospital. I will have my laptop and phone and my little DVD player...so at least I will have something to keep me busy.  '),(1153,1234212,'07,April,2003','2003-04-07 00:00:00','   I just watched \"Freedom Downtime\" a documentary about Kevin Mitnick. It was a good watch well worth putting food of to watch it.  Get a copy and watch it...definitely interesting and with my own dealings in the Federal Criminal System it gave me something to reference my own experience which was considerably \"more friendly\".  As I see it now true hackers, those who are in a quest for knowledge not for power, are the true freedom fighters left in this nation. Those that have stood up to Diebold are true patriots.  If I was more skilled (which I guess I could become in a matter of time) I would fight the fight seeing that I really don\'t care about laws all that much. I see the law as something that has be perverted into something to protect the rich against the poor, not about justice.  I guess what we need now is a true orginazation that would fight for us and help protect us...and I feel the time is coming when it will be a reality because of the many abuses of the government towards our rights.  How cool would it be to gain some sort of control of the vast military systems and deny those in power from using them to fight unjust wars, take away freedoms, and ultimately stop the people in power from endangering us all.  When that day comes there will be great rejoicing.  '),(1154,1234212,'07,April,2003','2003-04-07 00:00:00','   Hey I also have thought of something for ya\'ll to do. Visit FSTV (link on side Free Speech TV) and look for videos. They have many Chomsky video\'s. Also Google for him and you find a derth of info and many links to speeches and recordings oh Chomsky.  While Al Franken and Michael Moore are great writers that incorporate humor and a tongue in cheek attack against the right. However while Franken\'s book is well researched...again he is no Chomsky.  Chomsky says many of the same things but in a more mature way which is what is needed considering that the Republicans come of better in interviews because of the way they carry themselves (the only thing that bothers me with them is the trail of slime that accompanies slugs as they move across a surface).  Moore and Franken are very educated (perhaps not formally I really don\'t know) their information has been carefully pieced together and made to be bulletproof on their facts (where as Anne Coulter and Rush are ill informed on most issues and frankly I would call attention to that and perhaps between their two brains there is maybe the IQ of 120 which shows how low they are in my book...plus Rush has that pill problem).  Chomsky is an academic and is very believable in his message. If any news orginazation were to actually pit Chomsky against either Coulter or Rush he would win hands down because he is just simply smarter than both of them combined. He would be the guy that would look at and say \"holy shit...this dude is smart\" and it is proven time and time again with his books and intimate knowledge of whats going on as he must read a lot to catch the lies.  '),(1155,1234212,'07,April,2003','2003-04-07 00:00:00','   On blowise my rankings are growing steadily but surely and on the one hit counter I am now in the 400\'s for popularity so I must be doing something right...now I just need to get some more visitors.   Maybe I should post some hardcore porn to get my ratings even higher?  '),(1156,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok..well I got the counter fixed and it reflects the number from my other hit counter...so maybe all will be ok now.  '),(1157,1234212,'07,April,2003','2003-04-07 00:00:00','   Oh yeah and the counter screwed up...the service I use had someone hacking it and they had to restore from scratch...so I am going to enter my numbers back soon and get the correct number put in.  '),(1158,1234212,'07,April,2003','2003-04-07 00:00:00','   I spent a little time reading Chomsky\'s book...the more I read the more I see what I already know...not from being spoon fed the info but from me drawing my own independant conclusions. Chomsky certainly has the ideas better polished that I did but I had the basics already.  I suggest for anyone who wants to read something just for information and expanse of knowledge then go get Chomsky\'s book. It will be a worthy purchase for the 23 bucks it commands.  The next books on my list to get are Jim Hightower\'s book and Molly Ivan\'s book.  I am watching Alien Resurection Special Edition on my little DVD player...the screen is very good on it and extremely clear. The only fault that I can find is that maybe the hardware is just barely adequate as sometimes you see jaggies (anti-aliasing) and just a little rough...but really it is certainly acceptable...they got me an Initial 7 inch model that includes a battery pack, home adapter, car adapter, and remote. A great little package.  Well I am going to watch it and then get some sleep. Peace...oh yeah tomorrow I call the doc so I will know what is going on soon. I know I will be headed to the hospital seeing that I can barely walk 20 foot before I get winding and start hacking...so I surely will be headed in.  '),(1159,1234212,'07,April,2003','2003-04-07 00:00:00','   I was sitting downstairs and I thought of something disturbing and it will be my little prediction.  Haily Barbour (I think that is the spelling...really I don\'t care) the former RNC leader was recently elected governor of Mississippi. Really I can\'t figure out why but here are my predictions for his stint in office.  Mississippi has by far the worst health care in the US...I can vouch for it firsthand. Mainly because social programs such as Medicaid and Child Heath Care have long wanted for more money in the state to provide even basic health care. The system down there is by far very regressive and has needed help.   With the recent budget crisis in almost all states Mississippi was again bound for greatness in its lack of health care.   So the first of my predictions is that we will see the almost entire erosion of all social programs there. While they have never been anything to brag about, MS has provided at least some benefits. I think that we will see bad things in MS that will be a harbinger model for the rest of the nation.  MS was number one in one area and that was tort lawsuits. Almost any lawyer worth his salt knew that if you wanted to sue for big bucks against Corporate America then you held your suit there. MS was the first state to sue bog tobacco and the state won  billions that were to go, oddly enough, towards health care. MS has had a extemely effective Attorney General and he was even featured in a movie called \"The Insider\" starring Russell Crowe and Al Pachino.  My second prediction is that under Barbour we will see the death of tort...we will see massive tort reform that will allow the rolling back of at least a decade of work on everything from tobacco and chemical plants. The goal for the Republicans has been to severely limit tort lawsuits because their contributors have long suffered crippling settlements. The reform will be the model for the rest of the nation and as a result we will see sweeping changes that will allow corporations to effectively be imune to lawsuits that might have an impact that would result in positive change to business as usual.  No one until this point has really seemed to grasp this...or at least that I have heard about. So if we see people and media picking up on the idea remember that you read it hear first.  So there you have it. My predictions.  On the health front I am pretty sick...today I spent a lot of time asleep...and really that felt good but it didn\'t help.  Today I went to a record store here in town and found a unique gem. This store has just about everything you could want or dream for. On the second floor they have around 35,000 used CD\'s and what I would guess would be around that number in vinyl. They also have an extensive collection of used video tapes. On the first floor is the new selections but the majority is not the popular schlock you would find at any record store (they have it too), but you will find many groups from the outside. DVD\'s are also on the first floor.  I got four CDs and two DVDs for 51 dollars. Can\'t beat it.  Well goodnight folks. Peace.  '),(1160,1234212,'07,April,2003','2003-04-07 00:00:00','   Lou Dobbs has a poll question up tonight...are you better off today than you were three years ago?   Go Vote on the poll!   Hahahah when I voted 92% said they weren\'t.  '),(1161,1234212,'07,April,2003','2003-04-07 00:00:00','   Walter Cronkite is on Wolf Blitzer right now and he really does a good job for an old coot.  Asked if the media has done a good job of covering the war he replied we have done satisfactory job after it started but we missed the lead up and didn\'t cover the debate in the Senate for giving the blanked authority to the president to wage war against Iraq.   He basically said they failed the people and they should be doing more now to make up for that.  Right on Walter...there is a reason that he has been voted the most trusted man out there...because he has been straight shooting since 1950 and was the first to say the war in Vietnam was a stalemate. Which he called correctly long before anyone realized it.  When asked if Iraq reminded him of Vietnam he said it does and it is striking the similarities of how the wars were underestimated and poorly prepared for.  I would be honored to meet that man. He is the last of his kind and today pound for pound is the best reporter or journalist in the world even well after he has retired.  Again it would be an honor to meet and talk to him because he is a national treasure and we should be recording as much of the man as we can...we will never be able to replace him.  '),(1162,1234212,'07,April,2003','2003-04-07 00:00:00','   Harken plans to endorse Dean...the official statement should be forthcoming.   Another endorsement, another hit by the media...CNN is now riding even harder saying he can\'t be elected.  Dean is to be on Inside Politics in about an hour so we will see what happens.  Don\'t doubt Dean...  '),(1163,1234212,'07,April,2003','2003-04-07 00:00:00','   Last night I tried to make a post about someone whom I have great respect for and that was Noam Chomsky.  There are many video and audio clips of his speeches over at  Freespeech.org .  I would highly suggest that you head over and take advantage of the stuff they have up. A good primer for listening and getting into his style would have to be this video  \"Chomsky: The World After Iraq Invasion (Realmedia Video)\" .  This is a newer speech of his and very telling. It really is a culmination of many years of work and he is optimistic in this particular speech saying that there are things we can do to correct the coarse of action...in many of his other speeches he tends to shy away from making suggestions on what to do and generally lets the listener, viewer, or reader decide what course to take...but this speech he did give a few more suggestions of actions to take so it seems that he has finally gotten to his critical mass point and now feels compelled to really say something outside of just outlining the problems.  Again it is a great primer for the current Chomsky...be warned as he says he is not a charasmatic speaker...you think Gore is a bore...than this guy may lull you into sleep if you don\'t listen to what he is saying...but if you do listen and actively process the message it really causes great amounts of thoughts to start bouncing around in your bean.   Health Update Extra! Extra!  Well if you have been reading even semi regularly you have seen far more daily posts than in the past...particularly over the past week or so. The last few days the post have come even more frequently and they tend to be a little longer.  Well if you have noticed that then know that it is still because I am stuck in the hospital and this has become my \"lifeline\" to the outside world. People want to know what state of aggitation I am in...just read. The angrier the posts the worse I fell....just a little guide to the levels.  Yesterday was a really really shitty day. My kindey functions started going up the day before so they took me off of Tobramycin thinking that may be the cause of high kindey readings and dropped my dose of Collistan because that may have also been the culprit. Well to deal with that they just generally flush you with a lot of saline and liquid and usually your body gets straight...  Over about a 24 hour period they pumped about 3 liters into me, normally when you get that fluid it is because you are leaking somewhere along the line and you need the fluids to raise blood pressures and boost fluid amount to you can have a pressure. Well I am not really leaking from any spot so most of the liquid retained and started to bloat me...I\'m not talking about you know womans time of month bloat...I am talking probably 20 pounds of liquid in 24 hours. Normally kideys and liver and bladder step up getting out the fluid (plus I have had the runs...eww gross I know) but I haven\'t expelled nearly enough fluids to keep balance...so now I sit here with probably 20 or so pounds of fluid on me...thirsty life a bugger, I am bloated and having a harder time of breathing because of the fluid.  I imagine that this is what kidney failure patients feel like often (I know my grandpa feels much worse than this daily and has because he has been in full blown kidney failure for a while now...when he gets dialisys they regularly remove 15-25 pounds of toxins and water from his system).  Now I sit here and I am really thirsty but taking in the fluid is hard because it has nowhere to go...and I haven\'t peed nearly enough to evacuate the fluid....so something is up doc and it isn\'t pleasant.  Surprisingly I am ok with it all because there is now way that any doctor could have caught this earlier and Dr. Weinberg came in and apologized for this snafu and that while he was off this weekend he would make sure to check up on me through Dr. Yee (one of his talented associates).   The biggest problems that are presenting themselves right now is that I am breathing much harder and more shallowly than normal (result of liquid) and while last night before bed my pulse oxygen level was 97-98% on room air and a heart rate of around 100bpm my sats dropped in maybe five hours of rest to 90-91%, heart rate of 120-130, and ever more shalow laboured breath. So somthing is up.   If I were really sick and kind of out of it than I wouldn\'t feel so bad I guess or at least not care that I feel so bad...but I am not that ill...I am just so uncomfortable almost like the plum girl in Willy Wonka and The Chocolate Factory once she started blowing up. I am still getting bigger and the fluids are just coming off....it is really shitty. Plus with all this liquid I have a hard time standing up and getting moving because my joints are packed with fluid...you can see it because you can \"jiggle it, just a lil bit, woah!\"  So I am just not feeling well and it sucks balls.  Well it is time for my morning beating...hopefully they can knock loose some of the shit.  Peace  Oh joy now I have a nosebleed...christ on a cross....  '),(1164,1234212,'07,April,2003','2003-04-07 00:00:00','   I posted at some vanity blog today...something I rarely do because I find it a waste of time trying to educate people that have no desire to learn anything of importance...but it was on the matter of the MTv Show \"Rich Girls\". I feel bad for them because of their basic lack of knowledge on anything. They don\'t look happy, they look like they are stuck...they look like cannon fodder.  I tried to give a few words...I am sure it will sail far above their collective IQ of 90 over there...but one thing I have learned in the years...just because someone is stupid doesn\'t mean that they can\'t learn or become informed.  So I dropped a little nugget off and challenged them to come in for some education h0n3z style...straight laced and eager to let you come to your own conclusion.  Learning can be fun if you make it fun. Thinking is fun when you let it be. Intelligence is the gift of the ages that seperates us from most other species...that gift is whimsical thought and desire.   Rich Girls Fluff blog...normally I wouldn\'t but I felt compelled to leave them with something.   Peace...and I still have more reading and possibly posting to do yet...but first off to get a sandwich from the cafeteria...I am slightly hungry...oh and have I got a shitty day to tell ya\'ll bout....  Blah!  '),(1165,1234212,'07,April,2003','2003-04-07 00:00:00','   I am watching Colin Powell\'s press thing today.  I must say that the journalists that are there seem to have a little more teeth with him than they have shown with Scott McLellan, Rumsfeld, and other political idoits in the Administration.  Why is that...perhaps because Powell actually attempts to answer questions...sure there is doublespeak and dancing as anybody who looks to be in some sort of conflict with the leaders over basic policy stances as I think Powell has with the hawkish nature. Powell is a hawk to be sure, but he is a more cautious hawk than the rest. I think it is because he has been shot at he realizes that even the highest flying hawk can be shot out of the sky with the right mix of powder and bullets.  One question that was really good and I just had to laugh at his responce was when a reporter asked him how he felt about the just released Carnegie Endowment for International Peace report on how he has been perhaps to hawkish about foriegn policy and how they trumped up the weapons claimed to have been held by Iraq. She then said how Perle and Flum have in their latest book called him soft on policy and that he is the weak link in the administration over how to deal with terrorists. Powell just got this look on his face and said \"I don\'t do book reviews.\"  He then took the next question but quickly put that answer on hold and said \"seriously I don\'t do book reviews. If I am soft on these things there is only one person that I have to answer to over there and the president doesn\'t seem to bothered by my actions.\" That is to the best of my recolection. Not the exact words but a very close paraphrase.  All I could do was laugh because he at once showed that he was indeed the better man, as well as in my view bellitled them because in part they have never been in the military and have no clue as to how to handle things correctly. That is what I read out of it...the only thing that would have made it better for me is if he would have chuckled and said that they were as worthless as tits on a boar. That would have been perfect...unfortunately he still has to have a little tact.  Now I am not a huge fan anymore of Powell because I believe that he could have stopped or at least spoken out more loudly pointing out the shortcomings of the administrations case for war. I understand why he didn\'t...frankly if he would have he would have had his pretty good name and record absolutely smeared by these people who have no real integrity. Sometimes it takes you sitting in an uncomfortable position for long periods of time before something good will come from it. Powell looks like he has been sitting waiting for the right moment and I believe that the moment is close at hand...especially with this whole CIA leak thing in the White House...I think that may be what Powell is waiting for to really let loose and call these people on their statements and lies.  I hope that is what he is doing. Because I feel deeply that Powell is essentially a good man at heart and in the end he will be vindicated...but he could only do it through taking all this shit first.  I just hope in the coming months he proves to be the man that I think he is...even with his flaws without him having been there to really stand in the breech I feel that we would actually be a lot worse for wear without him and what has seemed to be his very good sense of limits and balance that he has ever so expertly infuenced policy just enough from making this a blowout against us.  Just my views...  '),(1166,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok I have decided that I need to stop reading at around midnight because if I don\'t I get all fired up and have a hard time getting to sleep because of all the stuff that bounces around in my head.  So many things I read last night had me nearly up on my bed jumping around screaming yelling \"yeah you go you go...damn more people need to read that, or the ever dreaded hip grind into the face of the right\"...it was almost if I had that false profit Pat Robertson in here so that I could dress him down with the most intelligent and hard hitting bible study that I could muster and all he would be able to do is fall to his knees and pray for forgiveness for his sins.  You see I am not a religious person. I believe in something but it certainly falls outside of what would considered main stream religion. I believe in a higher power but I refuse to name it/he/she/they because I simply do not know what or who it is that is our higher power. I also give my higher power the ability to be autonomous and really be like any good benevolant leader should be, objective and willing to stay out of the fray considering that when he seemed to meddle in the past affairs of this planet and its people it has always been at great detrement and large loss of life. I give this power the credit to think that it is capable of not only great passion for its children but also even greater wisdom to truly learn from the mistakes it has obviously made in the past.  All of the current religions tend to dumb down their god. Make it some inflexable being that is not capable of rational thought or coherant ideas. I find that heresy considering that all of the religious texts seem to back up my theory of this really truly great being or beings.  Now I fully admit that I don\'t know what we are dealing with because I have not met or directly spoken to this being. I cannot possibly fathom what it wants out of us whether it be good little subserviant doobies or instead real thinking animals that test and probe the limits of our understanding left and right.   There just isn\'t simply any text that I know of that gives us direction, instead all the religious zealots have been able to say (whether Islamic, Jewish, Christian, Buddhist, all the religions) is have faith and trust us when we say jump, you jump. For me I find that insulting that my higher power doesn\'t have the trust in me to possibly come to some conclusion on what or how that power is...and if it just wanted us to believe whole heartedly in the descriptions given by the religious leaders of the world than it would have been hard wired into our heads long ago and we would never have debate on religion...it would just be right and we would all be doing it. However that power thats up there (as I am sure something is) instead gave us minds that are powerful enough to not only master tool building, metal working, and advanced science reasoning but the ability for truly creative thought. So that would lead me to believe that in its infinite wisdom they gave us the ultimate power to shape and form outcomes.  I also believe that at some points it/they/he/she did intervene albiet with not a mighty calvary of winged beasties but instead through men that have dotted our history that if you look a little outside the \"written gospels\" on them you will see that they really were different thinking from all the \"religious heads\" of the day and truly sought to teach things that other considered heresy. Many of these men and women met with rather nasty endings after they challenged one to many of the rules that the \"men of god\" laid down.   Breaking News From Iraq  A Blackhawk Helicopter went down early today and now nine have been confirmed by CNN. This in addition to the one fallen soldier from the Mortar Attack on a logistics base outside of Bahgdad. Around 34 soldiers were injured in that attack.  This is horrible that we have lost more soldiers. This also unfortunately prooves Dean\'s possition that our forces are indeed no safer even with Saddam Hussein captured and on ice.   As a person that wants to see our current administration removed from power I would much rather they be removed without having to have the daily bloodshed of people that I can identify with because many of them are of a similar age to myself and many of them really do believe in why they are there (even though I think they are mislead on it all... but I am of the firm belief that many more are coming to the conclusion that they have been indeed used for corporate profit taking considering that they are short of supplies, they know their institution is paying way more than market value for fuel, and many of them plod around in Flak Vests that offer not much more than minimal margin of safet if not making them less safe with the reduced mobiliity and the very design of the material to stop more blunt trama and not necesarily designed to stop bullets). If truth were to be told and I was healthy I would join the military...not because I want to kill me some people, but because I believe that the military needs thinking people not killer robots. I would protest openly in the military against this invasion and now occupation, but it would not stop me from doing my duty to my nation. The thing is you can be honorable and disagree with the policy but still have the moral high ground by truly defending your nation by treating the people that we are occupying fairly and humanely. Being nice to people is far better than kicking them around and belittling them, taking away the basict dignity that seems to be lacking in many areas.  So we need to make it a point to hire \"softer\" troops that are a little more tolerant of others. We don\'t need pansies in the military but we certainly need an injection of humanity.  So my heart goes out to the fallen men of the military on this day...I would love to hear from men and women in the military and know their opinion for myself. I hope we can get an administration in place this year that actually cares for you and if it is decided that you all have to stay then the least that we can do is make sure you are well equiped and protected as much as possible so that your chances of coming home are heightened and the exposure to danger relieved as soon as possible.  Good luck and godspeed to you all. Remember that even though many of us do not agree with you being there you can rest asured that you will not come home as a national disgrace. The disgrace belongs solely to those in power and they will be the ones to face the heat...not you.  Take pride in our troops.  '),(1167,1234212,'07,April,2003','2003-04-07 00:00:00','   What I have decidedly avoided saying not because I fear the flak but because I lacked the ability to even apply simple rational thought without irrationally knee jerking to the conclusion that I needed to go around and just slap some sense into everybody that I could get my hands on.  I am not a genious...I am of rather normal intelligence, maybe slightly smarter than your average bear in search of a picnic basket...but I do take pride in the fact that I spend part of my day just thinking...taking a break from mind dulling activities such as watching TV or movies (even though I spend considerable time in front of the boob toob and the wholely addictive internet)...I just sit and ponder while I doodle or stare at the wall of ceiling.  This plucky Seatle native (they do tend to be snobs and that has been the largest reason people tend to recoil from the liberal ideal...because liberals do sometimes or almost always when they come from the northwest come off as snobs...and your average Chris B in MS just takes offense at snobs even though he believes that everyone should have health care and that people should have food)...so I am glad someone has said it...I just wish it weren\'t so true that it was ugly.   The \"S\" Factor explains Bush\'s popularity at Commondreams.org  '),(1168,1234212,'07,April,2003','2003-04-07 00:00:00','   Stop the presses...stop em quick. The little firebrand Arianna Huffington has spoken...ok so its no big deal...we all just like her accent (well at least I do).   Arianna Huffington telling the establishment Democrats the frightening truth.   Wow...another person that some know of has gotten behind Dean...and for all the right reasons and put it clearly in their intent of support without resorting to lighthearted \"hehehe he\'s great\" rhetoric...instead choosing to make the point through direct language and straight thought.  '),(1169,1234212,'07,April,2003','2003-04-07 00:00:00','   All this hoopla over labor?  Got off the phone with mom about it all and she was a little blown away at how I wrote about it. In the middle of talking to her the respiritory dude came in and was larguing with me on what the labor department has done and that it would never happen and they have been trying to do that for years.  My response was \"well of coarse employers have been trying to do it for years, but until now there hasn\'t been an administration of body of pro-corporate interests that have actually filled enough of the government to actually change policy and effect real world change...so up till now it has been rhetoric, and for all we know people will get wise to it and stop it dead, but based on how mislead and how cowed people really are then the changes will go through and be implemented quietly and with little cause for alarm over a five or ten year term. People won\'t see the change because it will be so slow and with us being so used to fast change when something like that grows slowly we largely ignore it and accept the new reality because it didn\'t come all at once and there was no \"a-ha\" moment to it all.\"   The ice age didn\'t end overnight most likely, but as the ice packs retreated life flourished in places that it couldn\'t before and most certainly the life that pervaded into what was once hostile land had no collective memory of how bad it was but a few years or decades earlier. The same goes with advanced life forms...the evolve and adapt just as the simpler forms of life do making small incremental changes over time to accomodate new realities (as long as they are largely non-toxic to life as a whole)...so in that is what these new labor policies rely upon...the inability of the human animal being able to track small changes or changes in norms over long periods of time. Time has proven this to be the way things go.   A natural example would be of the oil crisis in the 70\'s. Fuel costs sored at once and people cried foul...as the crisis let up the costs went down but not down to their previous lows if weighted against inflation (a normal thing in a capitalist system) and slowly over time they have creeped back up to what if properly weighted against the current dollar would likely prove to be the same per-gallon average as seen during the oil crisis...but because it incrementally went from 50 to 60 cents then maybe a bump to 80 cents a gallon then a slight easing...now we are all pretty used to paying 1.40-1.80 for fuel...but if the price had jumped up there in the early nineties and remained people would have demanded some sort of explanation...but the prices didn\'t go up dramatically (until this last Iraq war and prices are still high from it all). It is just in our human nature and our biologic inability to track very fine movements over any period of time. It was not the way we were bread.  So here is a link to the new policy that is currently working its way into law that got me all fired up at the dude (I feel almost bad because I went right on the attack and flat out told him he was wrong...and he may not be the brightest bulb on the tree as having seen with dealings with him in the past)...I came off strong and aggressive and he was like \"woah! your a radical\"...my flat responce was this (and it is because I still am) I am a registered rupublican and I have mainly voted republican in the past, but I no longer take the even moderate republican view now because it has shifted way past center into whacky right field with no eminant return to sanity to be seen for as long as the eye can cast its gaze. Which yes I am still a registered republican (I try to think of myself as a liberal McCain supporter...may not agree with him on all things but I find myself agreeing with him more often than not because of his seemingly no BS aproach to it all)...   Commondreams.org writeup on new proposed laws and enforcement for the labor department.   Fact Sheet #23: Overtime Pay Requirements of the FLSA.   Minimum Wage from the horses mouth.   That site is hard to get through but it has all of the pertinant information on new proposals....if someone stumbles across the link before I do with the new recomendations then please send it my way and I will get it up...and you will get the credit for finding it. Hell I\'ll send you a dollar if you find it.  This is important stuff people...very important indeed seeing that it will happen but it will happen at glacial speeds so that no one notices or protests before the change has already happened.  Peace  '),(1170,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok so I didn\'t tackle the whole slavery thing I thought I would...but I will add it as an afterthought...under the system that Shrub has proposed would basically allow US businesses to take advantage of this cheap labor for up to three years then cast them back no better than they were...to me that kind of equates to slavery because the worker benefits minimumly while the business reaps maximum reward from the sweat of their brow.  So ok it isn\'t the best argument that it will bring back slavery...but hey its all I could come up with this fast...so I am not the best truthsayer...but I do sometimes get it right ;-)  '),(1171,1234212,'07,April,2003','2003-04-07 00:00:00','   Well today Shrub anounced his new sweeping illegal alien program. Unfortunately it won\'t catch any of those pesky et\'s but it will manage to create an underclass here in the US one that can only be matched by our past in a system much like slavery.  Ok now hold on folks...that\'s a lot of rhetoric there and I promise I will try and back up what I say...while the numbers on this subject are hard to get a hold of because really there is no central position on it...you are either anti-imigrant or pro-imigrant and that really leaves the vast majority of us out of the debate forcing us to one side or the other when really I believe if people looked at it as I do then you would see a pretty sensible approach to all of this.  First I actually have to commend Shrub on this because the last president to take this hot button issue on headlong was Reagan...and he got it wrong just as Shrub is getting it wrong again...but it does take some moxy to step up to the plate and for that he gets maybe a half a golf clap out of me.  Here is what I think people overlook a lot in this debate. Currently illegal aliens are filling positions that US citizens would not and will not work for the money that is being offered to do those jobs. Notice the inclusion of money in that sentence...because money is rarely used in its proper context in this debate. People who say \"illegals take away good jobs from Americans\" actually they aren\'t taking good jobs, often these jobs are cooks, dish washers, yard boys, field workers....and the catch is that they are being paid less than minimum wage in many cases. So yes in a sense they are taking jobs, but these aren\'t the jobs that Americans would work for the wages offered so businesses do what businesses do and hire the next best thing, a warm body and don\'t even bother to think \"maybe if we raised the pay we could find an American citizen to work the job\"...but we also have to face the fact that most Americans are snobs and would turn down 7 or 8 dollars an hour to flip burgers while a foreign worker would be perfectly happy to make 3 dollars an hour to do the same job...they see a more to work than perhaps pay.  So what do I suggest we do about the \"massive illegal alien workers\" situation.  Well I suggest we give out all the worker visas we can possibly print. Hand them over with no regard to where or who the people getting them are (I don\'t mean just hand them out willy nilly but nonetheless hand many many out). Then in addition to handing them out we extend the labour laws to cover these workers and indeed tax them on their earnings and tax their employers just like a California business is taxed on a legal citizen and in the process raise the minimum wage or move towards a living wage system.  Now while that may not attract more US workers into these less than best jobs we can at least guarantee that if we are to employ foriegners it won\'t be at the expense of lower paid immigrant workers. Also if we document these people and tax them then we are adding revenue streams to our own system that essentially nullifies any fears of us loosing money in social programs to help out the sick that will surely happen amongst them. Most of these workers come over here now (illegally) and work for months or years rarely actually hitting the health system (mainly because they fear if they do go into the health system they will be deported...which is almost always the case under our current system)...so when they do hit the system it is usually because they are very sick...and we all know that preventative medicine is always cheaper but the current system makes that impossible.  So we are generating taxes off of these people that essentially zero\'s their impact on our social services plus we fill jobs that Americans classically won\'t work.   Now the big catch is that we have to force businesses to pay these higher wages and if they don\'t then they need to be fined or shut down if they do not comply with fair labor practices. That is the hard sell because right now they have no real incentive to pay these workers more because they can always find more (usually for less as time goes on) and the fines that they may incur for employing these illegals really ads up to less than a slap on the wrist and certainly not even close to what it would actually cost to hire a legal alien or American to do the job...so the cost benefit ratio (big business term there) is actually in favor of hiring illegals and occasionally having to pay fines for employing them because the fines will most assuredly be less than the cost to employ a legal or American citizen. It is as simple as that.  You change that to where it is vastly more costly to employ illegals at these low wages and make it truly a hassle then you will see real wages rise for these immigrants (who in the end only want the same thing as us and that is happiness and prosperity and enough to feed their families) as well as we may see an incentive for American citizens to go back to what they think are the low class jobs. They can\'t be low class if the guy next to you (who is foreign) is making the same money that you are...you are the same in the eyes of the IRS so that is fair.  Now how do you deal with the inevitable beauracracy that this will add. Well first we make these work permits just that...work permits they are not an entitlement to SS (and as such would be exempted from SS taxes, but not federal wage and state wage taxes...and you would still pay maybe an insurance tax for medicaid as many of these immigrants would probably need state insurance because as many of us already know private insurance does not work at all...it has failed). These temporary permits would allow the free travel of the holder of the permit and his/her immediate family (spouse and children only not extended relatives). These relative would be allowed to stay in country for maybe 4-6 months out of the year and must return to their native country for the rest of the time. The actual  documented worker must be here for 8 months minimum and at a maximum of three years (similar to Shrubs proposal)...after that time is up they would be allowed to re-apply just like getting a drivers license and if the person has been a good doobie (not broken any major laws...we can let em slide on drinking and minor offenses but not major felonies just as a play to keep law and order intact) we let them automatically renew. That process can go on forever and it will benefit us...we will have the labor but we won\'t carry the SS burden unless they choose to go through the process and are selected for citizenship then at that time they are a citizen and they pay their fair share.  This is a sensible idea...the idea of deporting all these illegals does us a double diservice...first it means that we have to spend billions upon billions on border defense and as we have seen for all these years and all the money we have spent we have not made a dent in even making illegal immigration slow, and we also deprive ourselves of willing laborers (in the process we may inspire a few workers of our own to realize that flipping burgers is just that...its not a life statement...Emeril essentially flips burgers but no one thinks of him as some wetback...because he takes pride in his work).  Also to consider is that borders are arbitrary lines drawn on a map and have never actually once represented what or how a particular population may actually use or travel land in a natural way. Politicians draw lines and then expect indigenous people to adhere to those lines when just a few years before they could travel openly with maybe paying a small tribute to a local chieftan or through exchange of labor or ideals amongst people.  So instead of knee jerking it like the Republicans do and saying \"ship em all out\" or instead of sugar coating it like the Dems do...we need to make it an honest debate. The Republicans make money through the tough anti-immigration laws...all of the law enforcement supply agencies are overwhelmingly Republican backers...it makes sense that the Republicans would want to promote anti-immigrant policies so they could keep their corporate sponsors happy. The Dems aren\'t much better because they never come straight with the truth...they sugar coat it.  Here is a first for me...and you may see more of it from me in the future...finally Labor has got the right track on this one. Finally they have gotten their collective heads out of their asses and realize that regardless of whether or not a laborer is in a union that the union still has a responsibility to the workers and should fight for all workers...because if all workers are promoted and paid fairly then the entire tide rises instead of splitting people up, forcing people into anti-union zealots because of past union misdeeds. Fight for them all win them all...fight for the few you will surely die on the vine.  So here\'s from me to the Labor Unions...while I may have disagreed with many of your practices (mainly because of lack of flexability with employers to work towards equitable solutions...but lately we know that businesses are making much more now than before so I have reversed my previous opinions and now mostly support Unions as long as they continue to keep their heads) I now know that you have landed on the right side of the fence and I am glad to finally be proud to say that while I may not always agree with you if I ever do come to a place that is a union shop I will join and if there is a great inequity in business practices and management practices towards workers than I will be the first on the bus to help unionize.  So you see people can change...I have over the past few year (as once I was staunchly anti-union because of their past practices) because now I am Pro-Union even with the problems...at least they are there banging the drums and fighting for worker rights.  The same must be done for workers around the world and the current illegals as our unions attempt to do for our own. We all work together and we all rise...if we neglect certain peoples than it only hurts the many...and weakens us all.  Peace...and open your minds.  I wonder what Mr. Chomsky would think of this line of thought? Hmmmm....  '),(1172,1234212,'07,April,2003','2003-04-07 00:00:00','   Later today I will be making a post most likely a little more towards the themes that I was working on yesterday...if you enjoyed yesterdays post I hope you will enjoy what seems to be brewing in my head right now.  Oh yeah go get you fix of  Democracynow!  The todays show has yet to be posted but it will be up soon.  Peace  '),(1173,1234212,'07,April,2003','2003-04-07 00:00:00','   Well last night I read a little more and today I spoke with one person already about our current situation and like me he was really alarmed...not scared about terrorism mind you, scared of what is being made of it all and how many freedoms he has seen lost like when he drove his brother to Pittsburgh Airport and his SUV was searched, not with probably cause or because of any real scare, but because the Department of Homeland Defense or Security is just hell bent on violating as many of our civil rights as possible.  So for me if I have to go out to the airport and pick people up I will refuse the search...hell let them ticket me or detain me but they do not have the authority to unlawfully search my car. It is one of my rights to unlawful search and seizure and if they have a problem with it then we can take it to court...I think that all Americans should start this move towards that ideal period.  Face it folks on one hand we play down how \"dumb\" the terrorists are that they would even try to smuggle a car bomb nice and close to an airport recieving area (not the terminal itself moind you, but ticket counters and baggage claim areas). Thinking that they would be dumb enough to try it is really stupid, they know the security is tighter and while they may attempt to probe they would not seriously attempt an attack like that. On the other hand we give common thugs (the Taliban and Al-Queda could be classified as nothing more than street thugs not all that much unlike the early years of the Itallian Mafia) this elevated sense of power over us by saying that they are so bright that they are outsmarting us left and right. Now they are either really bright or really dumb folks, can\'t be both ways. However by making them out to be a real true threat we give them far more credit than they deserve and more attention than is due for them.  If we treat them like criminals and move this \"war on terrorism\" back to a police and intelligence/investigative mode we would certainly see far more progress than sending our troops in to occupy nations that don\'t want us, and we could save ourselves billions of dollars and maybe perhaps return some of the freedoms that we have so willingly given up in the name of fighting terror.  If we don\'t do this then they have won. They will have with one bold, yet trajic event will have changed our system and our nation for the worse and they have won if they get that much out of us.  Freedom isn\'t free, freedom lives through the shed blood of our country men/women and to ask me or any other true patriot to give up rights just because we did loose people is wholely Un-American. We lay our lives at the alter of freedom and from time to time we loose people to the alter of freedom, but through those trajedies, through those losses we should always choose to come out stronger in our will and desire to respectfully defend our freedom...and by respectfully I mean leave the system alone and tweak it to help in defense but never at the expense of liberty.  Fight the police and Homeland Security Defensive military, don\'t allow them to search you. If you have nothing to hide they have no grounds to search you. It is one of your most basic rights guaranteed under our Civil Rights declaration.  '),(1174,1234212,'07,April,2003','2003-04-07 00:00:00','   I would like to welcome Anne\'s blog to my links over under blogs.  My Imago  She has an easygoing style and is insightful. Worth a read.  Thanks Anne for the link and here\'s one back at ya.  '),(1175,1234212,'07,April,2003','2003-04-07 00:00:00','   There were some good reads on  CommonDreams.org . Some stuff that is much needed right now to help liberals who may feel a little blue at what is going on out there especially seeing that the Republicans have seemingly got the better of us on the whole Bush/Hitler gaffe which I hold was a plant to do exactly that...give the right some cheap ammunition to fire off randomly hoping to hit and wound a few foes...for some it has worked I am sure because people are geniunely not that well informed on the behind the scenes actions that the RNC regularly takes to discredit others while propping itself up on the very thing they use to tear down their competitors (like horrible advertising campaigns like that of the Max Cleland campaign in GA where they morphed him into Osama...even though the man had lost three limbs in Vietnam.)  Another place to head to is  Truthout.org . Head there for some hard hitting straight shooting towards the Republicans...while they say they are able to attack both sides they really are much better at getting under the skin of the Republicans (because after all they are morally bankrupt and slimey creatures through years of inbreeding). They have a great op-ed on Bush/Hitler that brings the brutal truth and in another op-ed singlehandedly gave me hope that those of us who are true moderates have a lot to look forward to because of exactly what GW Shrub has pledged and delivered on. He brought unity...most definitely...but to the left, a unity that has not ever been seen before on the left side of the fence. The entire left while still having very different opinions on what should be done all have one common goal and that is to get GW out of office.  So there is hope and I believe we will all be free of the evil that has embraced our nation even though the big thing will be for corporations to try and stop it...they still can\'t stop a true vote...and if we make it fair and demand accountability then guess what...we will win.  Chins up...and keep the faith.  '),(1176,1234212,'07,April,2003','2003-04-07 00:00:00','   There are new polls out showing that Dean trails Bush by about 15-20 points and say that this will be a blowout for the President.  I would like to offer up an opinion on this. Dean is currently running against 10 candidates. All the Democrats and Bush himself. So he has a lot of different points that he has to cover. Really I am sure Dean is a little tired because he really is having to fight away in what has become one of the worst primaries ever with all these guys not talking about issues instead trying to sling mud and only have made a Democratic President more difficult to get to.  I figure once it is a head to head fight with true national exposure we will see the numbers indeed tighten up.  I like how Dean has said that we have to move campaigns away from God, Guns, and Gays and instead should be dealing with issues like education, health care, social responsibilities, and other things that are way more important. Unfortunately the Democrats have been weak in the three G\'s as far as most people say. In reality I believe that democrats are pretty even on those issues with Republicans but the Republicans have been more willing to wear they\'re faith on their sleave like a swastika. Unfortunately people have consistantly fallen for that argument and as a result have in reality been lead away from some of the most basic tennants of Christianity. Free Market economics and capitalism don\'t actually jive with Christian tennants but unfortunately few christians realize that.  So...  We will have to wait for them to go head to head for real issues to surface.  My mom just left and I wish that I could record the things I say and the conversations I get into because I have all of this pent up frustration about things.  My main thrust of thinking right now is the whole system that we live in and work with is downright wrong. We have people that want to farm and enjoy that throughout the world but they are not aloud to farm, we have people that are healers yet they are not offered the opportunity to do that. We have more than enough talented people to do all of the things that need to be done to live in a comfortable world for all and not have even a need for an economic system or currency outside of basic work and labor.   I often think to how nice it would be if we all did what we had talent for. Some would be computer geniouses, some artists, some doctors, ditch diggers, and lawn mowers...but there are plenty of people that are happy with even something so simple yet they are pushed into working long hard hours for money which is really just a few cents worth of material. We have people that are miserable because they don\'t have enough to buy their kids food, or have access to even basic health care or clean water.  The whole system that we are working under and that is being pushed by those in power is not even remotely natural. Humans (pre-historic ones and even middle aged humans) lived in a co-operative structure that allowed people to contribute what they could and everyone had food, water, shelter, and they lived what we would consider simple but happy lives. The system that we toil under now doesn\'t even remotely take into consideration that health, happiness, and prosperity don\'t exactly revolve around the dollar, yet day in and day out it is beaten into our heads that we have to have more money to be happy, and more stuff to be truly prosperous.  Now mind you I love toys and things. I love my car, and there are cars that if I could afford them I would buy...I love to have gizmos and such and if I could afford more I would certainly love to have them, but it does not mean that I am incomplete without them.  I wish that we could live in a world that we would truly seek happiness not through material wealth but instead find it enough to be in the company of nice people, or at our jobs doing whatever they might be but really enjoying them.  We could still have a reasonable version of what we have today. We would still have cars and planes and skilled people to do particular tasks, but instead of forcing people into positions that aren\'t where they may be best suited we move more towards the natural things that bring us happiness. There are people out there that love to work on cars, fix planes, dig ditches...they may not really enjoy it now because they are having to base their lives off of it, but if they didn\'t have to worry about making x number of dollars a month just to keep a roof over their head we would see a outpouring of kindness and happiness that would truly make the system we are toiling under look absolutely barbaric.  It won\'t happen overnight and may never happen at all, but I just wonder where it went. When we were children we did things that were fun and during those early years you can see the primitive nature come back out in us all as we do things for joy (some of us who grew up on farms made it a game of picking beans, those of us in more urban dwellings maybe put our play time towards computers, yet we enjoyed what we were doing). At what point do we as a society start to reprogram that most basic of all instincts, the instinct to be happy or content. That would be interesting to study.  Maybe many people have it wrong and those adults who act childish may have it more correct then us serious adults. Those adults live for fun (although in todays day and age probably at the expense of others around them) but they have a more organic view of life that includes playtime.  I know some will probably think \"what is this kid smoking\" well right now it is recombined hamster ovaries in my Alpha-Dornase (Pulmozyne a DNA med for CF patients)...but I think that we may need to think about some of these things.  I wish sometimes that I could tame the beast in my head because it really does run at what feels like a million miles a second and it can be at times quite tiring. I can say that in the past my attempts to numb that were fruitless because once I became un-numb things just moved faster and more fluidly and I had a harder time keeping up because of essentially the backlog.  I just feel that perhaps a whole series of thought has been seriously overlooked here in our nation and perhaps most of the first world over. I won\'t say that I would leave and live in a jungle, because I love modern things like air-conditioning, computers, the internet, and DVDs...but I think that in first world people\'s have gotten caught up in it all and have ignored for the most part what may be the most rewarding thing on the earth and that is simple time to think, time to contemplate, time to just spend around friends and family. Instead we pack as much diversion into a day as we can jst so we can avoid those most basic things.  I am not saying that some farmer Zapatista in rural Mexico spends his whole day working his land and thinking about astrophysics...but in a way he may be more able to come to more unique conclusions than the guy who has to drag himself day after day to his lab just so he make his payment on his chevy and house.   If this makes any sense to anyone please let me know...because having been the way I have been in the last few months (fairly sick) I have been re-aligning my thoughts and many new things that I though would never come to me have. I now espouse ideals that not all that long ago I would have looked down on, but now I after having spent a little time thinking about some of those things they make sense now.  I would just wish that people not get sick and have to think about this stuff under these circumstances but it would be nice for people to take maybe 20 minutes a day and think outside the box, read something that they make think is whacko, and concentrate on not thinking about the status quo. If we could all make that little concession I bet the world would be a better place twenty minutes at a time.  Well to all my fellow thinkers out there...I am glad if you have read this far and I hope that perhaps you too will come to think a little like I do (I don\'t want you to be little parrots marching around repeating the chant....instead I want only to inspire you to put some time into thinking and not having something spoon fed to you). I would love to be adored for my views and seen as the missionary of hope...but that is pride. Instead all I ask is that you take that twenty minutes a day a do something you might not do with it...go look at clover, watch the sky (you\'d be surprised how quickly you can loose track of time and thought when you watch a blue sky...often coming to some rather organic and natural thoughts that if you hadn\'t watched the sky you would never have thought the thoughts). It is just about centering yourself a little each day...and as you do that I am fairly certain more people would see the insanity in it all maybe strive for better things.  I wonder what Noam thinks of that?  '),(1177,1234212,'07,April,2003','2003-04-07 00:00:00','   I got my some of my newsletters today and some of them are pretty alarming.  The first comes from bushgreenwatch.org and it has to do with changes on how sewage treatment plants would be allowed to dump raw sewage if they have just minor overages caused by even modest rain.  We had about 2 inches of rain over the past few days but that hasn\'t been enough to really overstress the system but under the new regulations the sewage plants would be able to dump the sewage without treatment. Now they can already do this dumping under severe conditions of flood or when the system is truly in danger from flooding and would cause a greater public health threat then just dumping the sewage...while that is far from a good thing at least there is some protection ensuring that they can\'t dump willy-nilly. The new changes would allow for them to dump untreated sewage without having to wait for things to get bad. So it drizzles and then they dump.  Here is the action alert target link.  New EPA rules changes.   It is a perma link.  I also recieved some stuff from Jim Hightower the populast dude. I\'ll let the articles do the talking for me as my fingers are tired.   To many secrets, so many that there is a problem with things looking like coverups, even the Republicans find this strange as well.    Koppel is an ass and Kucinich points it out. Since when has Koppel held that much sway...and we all know not to fuck with Kucinich because he looks like he would knaw your arm off if it was in his way.    Gotta love privatization of everything....including our water. Not their water, our water. Time to speak out.   Thats all from Jim Hightower..but there is another tidbit that is starting to creep into the mainstream media and it has to do with Cheney, the French, Shell Oil, and bribery.   Will Cheney be indicted?    This is interesting because Cheney has no legal protections from overseas prosecution. He can be prosecuted and found guilty and even jailed under international extradition treaties. However this all plays out will be an interesting test of what has been long standing practice in international law adherance (although we rarely adhere to the laws)...but it will make for interesting debate as to whether or not the current regime will actually co-operate and move towards international laws and norms. I doubt it...but it would be nice to see one of our leaders essentially held hostage here in the states because if he were to travel to other nations that adhere to the laws would be bound to take Cheney into custody...how will that play. It will be interesting for sure.  I would love to see GW tried under War Crimes tribunals and be subject to arrest if he steps foot in a nation that adheres to the international laws...that would be an interesting incident.  Peace  '),(1178,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I had two posts hosed today by blogger...that sucks balls. Both were on the same thing...I tried to rewrite what I had written even more simply but it just didn\'t work out at all.  Now I know when I finally put together my analysys of Hilter up for people to see many will initially find it offensive that I would give that horrible man even some recognition for what he accomplished and how he single handedly remade the 20th century through his brutal wars and war crimes.  Being as such I think that if we go back and attempt to critically analyze what was happening over there and the dynamics of the situation we may find ourselves no in this day and age in a similar boat. While I would never say Bush is as bad as Hitler but I must concede that Bush is pretty close on many of his ideals.  I will have to do some research and make head to head comparisons...and I bet many will be stunned at just how closely played our game is now to how it was played then. If people were to see some of the frightening similarities to Nazi Germany and todays USA as well as the Former Soviet Union and the Stazzi Police in East Germany it may make people think differently about the whole thing.   The key will be to remove the emotional attachment to that war and the pride we rightfully won by soundly defeating our foes. In that though is where we started to steer wrong (although the invasions of the Phillipine\'s, Haiti, and Cuba we also early warning signs for people and our intensions.  I guess one of the things that I will have to stress as Chomsky has so well done, although in tough langauge for even more advandced readers (I consider myself a decent reader and can comprehand many things better that you average guy....yet I have some difficulty in reading and translating it into language that is easy to understand and then pass it off to people in more of a primer form.)  Now this will take time as I am only into the third chapter of Chomsky and already I feel a little overwhelmed by some of it all.   To fix our problems we must face our past and in my humble opinion we need to get our fingers out of everyones pies. The stem of hatred towards us has been through supporting very terrible men over the years and that has always come back to bite us in the ass. We should be worried.  We can make it work but first we need to bury some of our \"We\'re Number One Pride\" we have and get back to basics. We need to immediately cut all military funding to foriegn nations immediately. Especially when it comes to Turkey, Pakistan, Columbia (which now receives huge amounts only outdone by this last mention for money), and Isreal.  If we were to withdraw all military aid tensions would sure rise for a short time in those areas but the ancient tribal wars of civil wars would be solved finally. If we did that we will piss some off but it would seem that as our weapons dry up in those country people will be less likely to attribute all the crap to the USA because at least they won\'t be staring down the barrel of a US made M-16. That I am sure would make a world of difference.  Any aid that we must give (and yes I feel we have the moral responsibility to help the world out as much as possible) should be in the form of food and medical aid. Something that we are surprisingly good at when we out our minds to it.  We can help bring hunger under control and with what is considered drop in the bucket money for other things like making sure people have basic health care (not first world mind you right off the bat, but at least them up on vaccinations and help prevent the spread of AIDS. Also we could for a small amount of money make sure that every person should have access to clean water.  These humanitarian things would help our image elsewhere.   I also propose that while we are helping the world we help ourselves and bring our derelict equipment up to par and our roads rebuilt. These would be massive pupblic works projects that would gaurantee labor here in the states for a long time to come.  We need to cut the Military Budget...and this is something that I get into a fight with my mom and dad who insist that there are baddies all over the place and the army is our last line of defense. I know that right now we could rid ourselves of nuclear weapons and chemical weapons and still have enough conventional weapons to destroy us all.   Also my rationing is this. We know very well how to build these weapons. If we were to come under sustained attack there is nothing to say that we can\'t literally slap together nuclear weapons should we need them to defend ourselves. Although we would likely never have to use them.  There is a lot to talk about and I haven\'t gotten to Hitler. Believe me...it may anger some of you on how I think of him. I think he was evil but at the same time psychopaths can be extremely intelligent people as I have no doubt he was...maybe not a genious in the grand sense of the word...but I also give Bush some of the traits because he has similar command over what is going on often to the detrement of the country.  So once I get my post on that together I will make it available.  Oh yeah I will be moving this into a pay format cause I can get good solid deals that won\'t cost an arm and a leg. So now I just jave to secure a domain name and a hosting package...which could be done for around 70 a year with 2GB month transfer, and oodles of storage as well as unlimited POP3 Email boxes. That will be parlaid into what I want to do as well.  Please if you people could take just a bit of time and send me email addies for press people, reporters, TV talking heads, edititorian page moderators. I need them all so that I can make a database with them all and effectively start my plan.  Also I would like to see \"question of the day questions\" soon so that I can get people voting on them.  Well love to you all.   urlLink Journalists and news anchors here     urlLink Your help here     All your help will be greatly appreciated...and you may become part of the grass roots movement that will take the country back.  '),(1179,1234212,'07,April,2003','2003-04-07 00:00:00','   Damnit I lost another post but this time due to my fat fingers hitting the wrong combination of keys. So I will repeat even more quickly what I tried to say.  The ad that has people in a tizzy seems a little off to me. The ads that Moveon.org has chosen to be finalists are all excellent and very well done but some lack perhaps the shine of proffessionalism...although again they are great and very slick in their own right.  This Bush/Hitler ad though takes it to a slightly higher production level and looks a little more pollished than the others...had it not had the message that it did and was going on technical merrit it would certainly be one of the better ads submitted because of its production value.  Now I am not saying that the RNC planted the ad, however if it came out down the road that it was a plant it would not surprise me. That kind of tactic is well withing the realms of possibility for the RNC as they have nothing better than that. The lie and manipulate and the only hopes for them to retain control requires them to deligitimise their opposition and this ad certainly helps tarnish Moveon.org\'s sheen up to this point.  Moveon.org should activily persue the producer of the video and if indeed it was a plant then they need to raise holy hell over it all. If it was a genuine ad from someone they need to know that as well and perhaps have the author defend his position or offer his view. Moveon.org just needs to do it to help wipe the oil off so that they don\'t get bogged down and labled as another \"bunch of crackpots\". The only way that I can see that they can do it is to out the author and then react appropriately.  Now I read an interesting article over at Commondreams.org on the DLC and the DNC showing the internal power struggle at the national level DNC.  It is a good read and you should all give it a few minutes of your time.  Interesting article that is worth the read to gain an insider view.   So today has been a whirlwind for everyone. It is just good that debate is pushing forward and we are all starting to see some honest discussion come out of it.  Also I saw Nader on Crossfire and he really stuck it to Novak shoving shit right back at his fat ass. I don\'t think Novak knew what to do...which was just nice because he didn\'t let Novak run him over and indeed challenged him to debate or discussion. As usual Novak backed down once Nader got started.  Nader basically came out and said while he supports most closely Kucinich but ceeded that he would support any of the Dems to come out on top and fight for them because the goal is simply to get George W. Bush out of office regardless of who it is that replaces him even if it is a wishy washy Dem unlike Kucinich of Dean.  So it was interesting to see.  Well I am going to rest my head for a bit...I will probably write a little later.  Peace  '),(1180,1234212,'07,April,2003','2003-04-07 00:00:00','   Unfortunately the ad I just saw says that it was sponsored by Moveon.org but I don\'t believe that at all. Almost all of these ads which were independantly produced has a \"sponsored by\" so that it would be broadcast ready.  The ad was dead on and didn\'t tell any lies. All it did was compare what happened then under a leader and what is happening now under ours. Not an unfair comparison, but it will be hard to deal with and will cost a little political capital...but all will be well in time because we would be able to prove what the ad says.  '),(1181,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I have seen all the ads that are up for the grand prize and really they are all good as I have said.  There is currently a big hoopla about two ads so far that have been submitted that compares Bush to Hitler. The republicans are all up in arms and the media is in a tizzy as well.  Now I would hate to have to think that we have to compare any one of our leaders to someone who was so obviously evil like Hitler (and I have no doubt he was an evil man), but I do happen to think that Bush is skirting right on the line along with Hitler. While Bush has yet to open gas chambers or locking people off in ghettos here (yet Gauntanimo Bay is pretty close to a warsaw ghetto situation), he has certainly infringed on our rights and helped put us into a position of not being that safe.  Again it gives the republicans ammo because they can say look at how unreasonable the liberals are...blah blah blah all the while they have been attacking the liberals calling them essentially commie pinkos for wanting to have health coverage or a decent educational system. It really is a pain in the ass when you have to fight with both hands tied behind your back while they take pot shots at you and no one calls them on it.  So do I support the ads? Well I haven\'t seen them and I will see them tonight so I will be able to see. Do I support the idea of the ads...sure as shit I do because they have accused us of much worse so it goes to show they can dish it out but not take it. It will hurt somewhat but again as I have said in the past while the liberals and the elite educated have attempted to remain above the fray and be honorable (or at least try to appear to be) all the while you have the republicans down in the real world meddling and making things a real mess for the rest of us. So I think it is time to step off the stump and get into the mud and expect it to be ugly...we have to fight them on their turf because the people that have been voting for them all these years has been only subjected to that so when you have a softer approach the republicans have a great line in \"look how soft they are\". We have to be harder than them, we have to be harsher than them, we have to fight their fight and win it against them...we are smarter and shrewder...so it is time to take the fight to them and hit them where it hurts.  They are going to play the ads on Wolf-Blitzer and his tag line was that republican leaders are upset about the comparison, but remember when they had Dean and Bin-Laden in the same ad....I guess that is ok, but comparing who is undeniably not good for the country with one of the worlds worst leaders...nope you can\'t do that.  I have something to say about Hitler that I will voice at some point soon.  Oh here we go.  Well they weren\'t that bad...it is just a bunch of hoopla yet again. I have the other to watch. The Republicans can dish it out but can\'t take it when it comes to deeply political and personal attacks.  The Republicans can attack Max Cleland who lost three limbs in Vietnam calling him un-patriotic. They can attack one of their own John McCain saying that he fathered a child out of wedlock (which is a flat lie...he adopted an Asian child with his wife)...so they can dish it out but can\'t stand it when the fight is brought to them because that is their tactic and when it is used against them they have to become creative to retaliate...and republicans aren\'t that creative.  So just remember what they are all about. The are chickens and can\'t really stand in a fight as shown by how many of them or their children actually have served in active duty military roles...yeah. Try to find them in the Republican party and you won\'t find any...  '),(1182,1234212,'07,April,2003','2003-04-07 00:00:00','   Today I recieved an email from Moveon.org and their campaign Bushin30seconds.org.  Well the finalists that they have chosen are really the bomb. The several clips that they have chosen are all the cream of the crop and several of them are really hard hitting and get to the point of the whole campaign easily and with great impact.  I will go ahead and post the links so you all can take a look at them. Although I am sure you would be able to get them on your own I will just make it a little easier for ya\'ll.   Without further ado, the finalists are:       CHILD\'S PAY by Charlie Fisher of Denver, CO   High-Bandwidth Version   Low-Bandwidth Version       IN MY COUNTRY  by Harry Katatsakis, Derek Rittenhouse, Chris Wight of New York, NY   High-Bandwidth Version   Low-Bandwidth Version      POLYGRAPH  by Rich Garella, Adam Feinstein of New York, NY   High-Bandwidth Version   Low-Bandwidth Version      BRING \'EM ON  by Mike Cuenca of Lawrence, KS   High-Bandwidth Version   Low-Bandwidth Version      WHAT ARE WE TEACHING OUR CHILDREN?  by Fred Surr, Ted Page, Janet Tashjian of Needham, MA   High-Bandwidth Version   Low-Bandwidth Version      IMAGINE  by Mark Vicente of Los Angeles, CA   High-Bandwidth Version   Low-Bandwidth Version      HUMAN COST OF WAR  by Brian Wilkinson of White Plains, NY   High-Bandwidth Version   Low-Bandwidth Version      WAKE UP AMERICA  by Lisa M. Rowe of Hollywood, CA   High-Bandwidth Version   Low-Bandwidth Version      DESKTOP  by David Haynes of Dallas, TX   High-Bandwidth Version   Low-Bandwidth Version      ARMY OF ONE  by Penny Little, Nick Green, Michael Stinson, Julie Stigwart of Isla Vista, CA   High-Bandwidth Version   Low-Bandwidth Version      BANKRUPT  by Adam Klugman, Dave Adams of West Linn, OR   High-Bandwidth Version   Low-Bandwidth Version      HOOD ROBBIN\'  by Nathania Vishnevsky of Foster City, CA   High-Bandwidth Version   Low-Bandwidth Version      LEAVE NO BILLIONAIRE BEHIND  by Andrew Boyd of Brooklyn, NY   High-Bandwidth Version   Low-Bandwidth Version      BUSH\'S REPAIR SHOP  by Eric Martin of Silver Spring, MD   High-Bandwidth Version   Low-Bandwidth Version      GONE IN 30 SECONDS  by Eric Blumrich of Montclair, NJ   High-Bandwidth Version   Low-Bandwidth Version     You can also view all of the ads on the Bush in 30 Seconds website:  www.bushin30seconds.org   Now those all should work hopefully. They are all in Apply Quicktime format so you may need the player and here is a link to get it  Apple Quicktime Download Page.   Enjoy and thanks for reading.  Peace  '),(1183,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I was extremely thrilled today to get a email responce from Mr. Noam Chomsky on an email I had sent to him voicing my concerns over Pakistan.  It was short and to the point and while it held little information from him it is still nice to think that he took a look at my argument if even only for a minute or two. I would never pretend to even be on his level politically or with the vast amount of information that I know he has amased over the years.  I really don\'t have much to say tonight outside of I caught literally the tail end of the Democratic Debate where they were all giving their 45 second spiels. I hate that any candidate be mandated to limiting his/her time to get a sound bite. That is the downfall of our democracy and our system. We have become to dependant on sound bites and tid-bits of information. People have seemed to loose the ability to have an attention span for anything more than five minutes or so...even when it is of such national  urgency.  I will say it now and I will continue to say it in the future. I personally believe that all \"politicians\" are worthless because of their very design. The problem is that under our system or the way we have been brought up to think is that we need these guys to run things for us. So seeing that we are trapped at the moment with the flawed system the best we can do is just try and figure out who will be the least threatening to us and the rest of the world.  Now I have made my choice of whom I support...and you can clearly see it in my links. I support Dean. I don\'t blindly support him and there are many things that I see differently from him on...but in the grand scheme of things he holds the best balance of things for me personally at this point in time.  If I were to have my way I would vote Kucinich and I may do that in the primaries not because I want to sabotage Dean...but because I believe in my vote and would at least want it to count for something at some point. I suspect Dean will get the nod and it looks to be that way so using my vote for Kucinich wouldn\'t be a waste as he unfortunately stands no chance even in the Primaries (which pound for pound his ideas and policy statements are better than any of the other candidates combined...as well as his record for action. For years he was critisized for taking the stands he did as mayor of Cleveland or wherever but it has come round to hey he did the right thing even in the face of the corporate monster.)  I wish more people could see these men stripped of their ambitions and see them for what they really are. All of them are pretty equal in their drive to control (Kucinich at least seems to want to use the power for good, Dean to not do any more harm) things.   Maybe one day once our system falls into disarray (as it is rapidly doing) we can move back towards a more organic form of government and something that will actually serve the public and be in its interest. However I fear that we will have to wait a while for that to happen (and really I have no clue what kind of government that would be because I have not studied it at all...but I am sure it would be more fudal perhaps more anarchistic with less top heavy structure as it is now).  Well again I was thrilled to get a short Email from Mr. Chomsky and I look forward to finishing his book...and again as I have said my impressions on Mr. Chomsky is that he doesn\'t seem to tell you how to think...but to just think and come up with your own conclusions...and in todays academic world that is almost like heresy...thinking isn\'t allowed when it comes to social consiousness.  So go be good little doobies and think for yourselves. Read, amass knowledge, decide, and act for yourselves. As long as you allow yourself to be lead you will always be last and the least of the leaders concerns.  '),(1184,1234212,'07,April,2003','2003-04-07 00:00:00','   I hope some of you will heed my call to arms below. It would be greatly appriciated and will be rewarded somehow.  '),(1185,1234212,'07,April,2003','2003-04-07 00:00:00','   I need a flavor from my readers. I need for you all to help me gather Email addresses of as many news outlets that you can find. ABC, CBS, NBC, Fox News, CNN, CNNHN. I am looking for specific email addresses for on air talent as well as behind the scenes people.   This is the first step in what I want to do and if you could find it in your hearts to help me compile lists of emails addresses it would greatly be of help.  I will be giving out gifts to the top ten or so contributors. I have not decided what that will be yet but I am willing to take ideas on what I should give away for you hard work. It could be gift cards, or something else (remember I am still pretty broke so the gifts will have to be reasonable...maybe something like 20 dollar main prize and then steps down from there.)  The hosting package that I am looking at will also include unlimited POP3 email addresses which could also be up for grabs for top contributors.  I think I have found a host that will give me decent bandwidth and space for around 8 bucks a month. It will be an interesting thing to get started.  you can send the emails here   urlLink Send Email Addies here   Just take a little time and help me out. This is the first step that I can figure out on how to get on my journey.  I will also be calling for volunteers at some point to help me with some of the site coding as I am not the bomb diggity in that yet....so I will need to inlist outside help.  As I get replies that are serious I will outlie my ideals even further to those people. The idea is to get co-operation on an unprecedented level to get out media back. It may not work at first but the idea that I have can and will gain steam as well ratched up pressure.  Please help out if you can. Right now I need no money, just time and research time. Later I will need help maintaining for which anyone who helps on that end would be rewarded off of any profit we may make or attention we gather. My whole idea is to make this about the people, not myself, or my ideals, but the people and their ideas.  So help is greatly appreciated....and I do welcome all comers no matter what your party affiliation because ultimately the ideal will extend to confront both sides of the isle and with dogged determination we can at least hold them accountable.  Peace...and don\'t forget to send me those email addresses (CNN, CBS, FOX, ABC, NBC bigwigs...all of them).  Mike  '),(1186,1234212,'07,April,2003','2003-04-07 00:00:00','   Well Nasa has made it with their little probe to Mars...after some of their more devestating catastrophies it is about time that the boys out there get some good news.  Congrats guys and I hope that we are able to find some interesting things.  Love to all of you...you rocket scientists.  '),(1187,1234212,'07,April,2003','2003-04-07 00:00:00','   Holy shit....fuck fuck fuck fuck fuck fuck fuck.....I am really friggin sick right now! Oh my god this really sucks balls.   My throat is full on sore and red, my sinuses are clogged and draining, my mouth is raw, swallowing is hard, my hips and back hurt, and hey how bout those lungs....  Ya know it goes to show that sometimes you just get worn out...and I am worn out right now. I am very tired...and itchy...I just need to get a little better please...just something?  Throw me a bone here people...just a little bone.  '),(1188,1234212,'07,April,2003','2003-04-07 00:00:00','   I had said that I was going to write about Bush and his call for the moon and Mars.  I have been struggling with what to do or think about this all. I am really torn in that I personally think we should already be well on the road towards those goals and should have never abandoned our space program as we did back in the mid eighties. Yes we abandoned it. After Challenger we pretty much just let it go and have never really made any attempt to get back in the \"race\".   Once we lost Werner Von Braun and his true vision we lost our way. Von Braun was undoubtedly the father of our space program in its current form (along with Goddard to a lesser degree). Braun had planned to have us on Mars already with semi-permanent moon bases as the launching pad for the cosmos. Many would say that he was a horrible man that helped enable Hitler to strike terror into the Brits and also helped allow for the killing of the Jews. Werner was never that political. He was first and foremost a scientist hell bent on one thing. Building rockets to go to the moon.  At the end of WW2 we were lucky enough for him to know which side to join (ours and not the Soviets...he dissapeared and sought out our forces for surrender). We got him and brought him initially to Alabama (the reason was because Alabama at the time suprisingly is reminicent of Germany without the extreme cold weather. So he was there and started to work. Initially he was working on a modified version of his V2 to deliver weapons for the military. Again he still wanted to go to the moon...but alas he was still stuck with developing weapons.  Soon Kennedy issued his challenge to the US and the World. Von Braun now had a mission. Get a man to the moon and back.  Well we all know how it ended...we got there.  Now we lost the initiative in the late seventies because space was expensive and the Rusians had given up on the moon and were working on space stations and such. So we put modest effort into a space station. We also designed and built the Space Shuttle (or STS).  The Space Shuttle was at first thought of that it would \"make money\". They would be able to turn shuttles around every couple of weeks and they would carry payloads to space for the military and others...essentially they would rent out their cargo capacity and pay for the program. Well that never worked really well..  We have now had two of them go down. One was because of gross negligence on the part of NASA bean counters, the other was a fluke that had been foreseen but nothing done about it.  NASA has an anual budget of around 15-20 billion...and this is practically nothing in the budget...the numbers have been pretty consistant over the years...while not technically the budget hasn\'t been cut all these years, but as a percentage or value the budget has shrunk.  Bush has proposed we head back to the moon and soon, then head off to Mars. What he didn\'t do is actually figure out how much it would cost or how he would pay for it. He asked for modest increases in funding (a billion over several years with massive fund redirection within NASA itself).  Now one thing we should have learned coming out of the eighties is that you can not have a stingy budget on time or money. When you do that you get accidents and you actually see costs rise.  The perverse thing about it all is for the relatively low amount of money we have put into the space program since its inception we have seen (I heard the figure somewhere...if I can find it I will post it) over 20 trillion dollars of economic activity from the small investment. It has been simply the best investment that the government has ever made.  We do need to go. We need to be back on the moon. We need to go to Mars...we need to do it because we are human and we have the urge to explore. It is in us.  The administration wants to make this into an evil thing though...they want to do it unilaterally so they can militarize. That is wrong.  I would propose that instead of competing against the Chinese or any other nation we need to involve them for the same goal. We need all of the world on this because it will truely be a huge task. The moon is easy. In flight times it is insugnificant. Mars is a tad different.   We need the Russians because they have the long term space exposure experience. With them and the drive of the Chinese, Japanese Space Agency, European Space Agency...all the rest we could get there easier and with less risk overall. If we involve all the people of the planet in this goal than we can make the world better.  Bush doesn\'t want that...co-operation means that we have to loose...we have to loose our \"need\" to be number one. We can do it...but it will take time.  I wish that this idea wasn\'t so politicized because really going to the moon and mars would be great, and be good for the world...unfortunately it likely isn\'t going happen under the current administration.  I could ramble on about this for a long long time...we need to do it...not because we need to be number one...but because we need to make a move back towards that pioneering spirit and the drive to get things done right.  I dunno...we could just do so much better.  '),(1189,1234212,'07,April,2003','2003-04-07 00:00:00','   I am getting sicker I think...that sucks balls.  My mouth is now pretty much raw and my chest hurts a lot now...and now my throat is sore and the glands are swollen.  I have Jesse\'s computer to build...I just have to figure out when he can get out here to do that because I need to get back to my house so I can get all the tools and extra bits that I will need to make his system run.  So maybe tomorrow I can escape for the few hours that I will need to get what I need...it would take me maybe two or three hours to get together what all I would need...it is just getting there and really going at it hard for a few hours and then get back here and relax for a bit...that would also give me the oportunity to get the right discs I need to rebuild this laptop...I wanna move it to Window XP Pro Corp...ya know cause it is better than XP Home.  I just need to get the bits together and get it all running...probably the only thing that I won\'t be able to do here at the hospital is do the updates to the operating systems...just because they are large downloads and so far my ethernet port here in the room is not yet active...  I will have to bring my box down just to get software off of it...but then I can send it home after I get Jesse up and running because I know I have all the software on my main box...so I can get somethings updated...and others won\'t...  Baaaa....well it is time for me to get more drugs...benadryl and percocet...oh yeah....get to shudder for a while then maybe sleep...  Peace...this next few days is gonna get hectic...real fast like.  '),(1190,1234212,'07,April,2003','2003-04-07 00:00:00','   The corruption just stinks...we all know it is happening because we can just feel it in our guts.   Ashcroft not only looses to a dead man but had to cheat and still lost!   We all know we need health care...just another call for it based off of accounting... We can save the system only by going one payer.   In the darkest of places even light shines in. Workers are rising slowly but surely and will soon be able to fix the mess. What is funny is that Labor unions and other unions have finally figured out that their struggle is not only in the factories where they are based but also around the globe with others that do the same job. You protect all your sector workers and the collective boat risis, you protect but your factory it can be isolated and cut off thereby making yourself the greatest obsticle to making it work. It has taken the unions way to long to figure it out...but hey better late than never.  Nafta Ten Years on.   A good commentary on kitchen politics...very well written and to the point on everything.  Kitchen-Table State of the Union.   We all know that they will never get to the leaker or Joe Wilson\'s wife.  Here is why.   We should all be mad...we should all be screaming bloody murder...we should all be out in the streets with signs and then into the voting booth to get this absolutely horrible regime out of the White House.  Dean has been slipping in Iowa...or so it would seem. I would like to know the method of testing. Is he truly loosing votes or is each poll a little larger net casting than the old with more people giving voice to their opinion. If it is more people signing up and voicing their opinions than his numbers aren\'t shrinking but still growing however not at the same pace as the others.  I was watching on CNN interviews with Caucus goers and they were surprisingly engaged. More so than most Americans and it made me wonder...what if we all had caucus style elections then maybe it would engage people, not just politicians, but people in the process. Really this contest is a form of debate amongst people...not politicians. This is great news because people are taking back the process.  Regardless of who you support be engaged at the \"grass roots\" level. We also need to rename \"grass roots\" because it is outmoded and doesn\'t give credit to those who are truly involved...I am grass roots by definition, but it doesn\'t mean I am out with a sign and buttons (although if I were healthy enough I would be)...maybe we should call this group \"The People\".   When it comes down to it we have to interact with each other...I interact with you all through this blog...some interact in person...some just put a sticker on their car...maybe we need to have several classes of \"The People\"...not classes like you are = x classes...but instead classes or specialists on things. For me I would be a speechgiver, someone else would be the bumper sticker sticker, someone else with another job. We all know that campaigns are already broken down that way...but call of the people in that office \"grass roots\" when they really aren\'t. They are the people.  Call them what they are...exhault them for their work...recognize that every cog in the wheel is important from the guy that goes on the news for his candidate down to the 15 year old in the office licking stamps (hey I did that...and got a healthy liking for stamp glue...hehehehe).   Well I am going to read some more...the day is still shitty...but after I got that post out of the way below...I have felt a little better...  Peace  '),(1191,1234212,'07,April,2003','2003-04-07 00:00:00','   Today is shit. Really it is.  I feel like hell...my mouth has nearly gone completely raw and I have these little blister type things in there...skin is peeling...and overall everything tastes like metal. It is really annoying and now my glands are getting swollen and my throat is getting a tad sore.  I also talked with my grandmother and she isn\'t well right now. She has been going through a lot of shit lately. She has had this pain in her side and cancer markers have continued to rise. I know she is really worried.   What makes me upset is that they haven\'t really done more...they have needed to stop with the CATs and MRIs and Bone Density Scans and just went in and get a damn sample of it so they know for sure what it is...but what worries me is it isn\'t where they thought it was...it is in her tailbone not her ribs like where they have been concentrating. So it has gone almost ignored completely until now.  I don\'t know what it is...but I know this I can\'t loose her right now. I really can\'t. I love her dearly and I am so worried about this. I have gained great strength in knowing she was there...always has been and always would be.   I rely on her for so much that no one could ever know. I can make money, I can make my way, I can do all that...what I cannot do is see her in pain again or have to see her go through all that again. I don\'t know if I have the strength to do it...I will of coarse do what it takes in the end...I just don\'t want to have to...  With all that has gone on lately in and around all of my families lives...we are just in a big mess. My grandfather is sick and will die soon (we think, hope, and pray...no that isn\'t mean it is merciful), my grandmother is fighting whatever this is and has been dicked around since March of last year to figure out what is wrong. My mom has pnuemonia I am sure...she was in here hacking and coughing. She sounds worse than I do....no matter what anyone tells her she is right and she won\'t take the time to go to a doctor and just get something for it. It is something that could be fixed in a matter of days...but by delaying it she has not only put her health in risk but mine and my grandmothers. (my grandmother has the same cough that my mom has now...she got whatever it was from mom and my cough has changed from a sharp cough to a gurgle just like my moms....so she has helped us all get sicker...yet doesn\'t seem to get it even though it has been pointed out...she just acts repugnant about it all...if you are reading this...just go to the damn doctors and get the meds that you need...I don\'t want to have to tell you not to come because you are making me worse or GM worse...it is simple, you are stubborn...and through that stubborness throughout your life you have always done this and always made things worse by not even doing anything about it).  I am done with that...she knows what she needs to do...will she do it. No she won\'t because after all she knows better than a doctor.  My shit has gotten worse and I don\'t know what is going on...I know but I don\'t. My mouth is raw (as I said), my kidneys are screwed up again (they hurt), my hips and knees are all buggered up (they hurt)...it is all just a mess and my cough has changed from a sharp hacking cough to a more gurgling slick cough...There is stuff there but I can\'t get to it...plus I just feel like shit.  I hate to complaign...I really do...but it is all I can muster right now.  Peace  And send good thoughts towards my grandmother. She needs them.  '),(1192,1234212,'07,April,2003','2003-04-07 00:00:00','   Well my mom and grandmother got me a new laptop...a really nice one. It is a Toshiba Satellite M35-S359  Intel Pentium M 1.4Ghz 512 MB RAM 60GB HD Nvidia FX GO5200 32MB RAM Cenrino wireless setup as well.  This damn thing screams along pretty fast I must say...the keyboard is a little bigger...well a lot bigger and I am having to get used to it. Strange....  Also the screen is crispy clear...very nice...no ghosting at all (my old Tohiba suffered horribly from ghosting)....the only thing that I could gripe about is the keyboard flexes a little much...so I will just have to re-learn typing...but damn I am impressed.  I am going to return the Portable DVD player I got for christmas to put that money towards this..because this has a DVD player (as well as CD-RW on it) so I don\'t need the extra DVD player.  I am just thrilled....giddy as a little school girl...look at me oh so pretty as I prance round in my catholic school girl outfit....weee weee!  Ok...well I am tired and I do actually need some sleep...and I have a lot to read before I get to sleep....  Oh yeah the connections here in the room aren\'t active yet...so I am on my Cell Phone which is still kewl as hell...this thing has all the connections Ethernet, Modem, Cell Phone, 802.11B...just missing Bluetooth (but I have no use for it)...it is just damn fine :-)  Thanks mom and GM...love ya\'ll lots.  '),(1193,1234212,'07,April,2003','2003-04-07 00:00:00','   Well today I had a trajedy occur. I had asked for a heat blanket and when they finally got it to me and they were hooking it up water sprayed out of the damn thing and soaked my laptop. Killed it deader than an upsidedown armadillo...now don\'t that just suck.  So I am without computer...and really I can\'t do much about it right now...I think I am going to look for a new laptop...something that isn\'t uber but instead spartan with enough oomph to get by....  I have a couple of that I am looking at...or rather I have kept my eye on...so maybe one of those will do. I could bring my big computer down...but that is a hassle and frankly if that were to get broken I would really be ass out.  So...updates over the next few days will suck...even if I update simply because I am using a semi-public lab and I don\'t want to broadcast everything out there....especially over a network that I am not completely confident of...when I get another computer and get hooked up to the network here at Shadyside Hospital (we move tomorrow to the redone floor and I should end up getting a suite with internet access) I will have to see what it is...and how private it will be...it it ain\'t private enough than I still have my SprintPCS Vision service...and I am a little more confident in the security of that than a UPMC network...  Well I have some shopping to do...peace ya\'ll.  '),(1194,1234212,'07,April,2003','2003-04-07 00:00:00','   Just saw this on CNN and it is so touching. If we could all aspire to be half as good as this man seemed to be then we would be so better off.  I plan to try and do something on this front...because it is the right thing to do.   SSSSH   That concludes the good deed work today.  Later after the President does his address on Space I will give my view....and I have one :-)  '),(1195,1234212,'07,April,2003','2003-04-07 00:00:00','   Just got back from my PFT\'s (Pulmonary Function Tests) and they still suck. They have not improved at all.  An interesting thing did happen though right before I went down...I started coughing up blood. Woohoooo! So something is mad in there.  I just want to feel somewhat better....and I don\'t I just feel like I am standing still sliding backwards.  I have figured out how to fight back against the Neo-Conservative movement by taking a look at it.   You have to give them one thing...they are very well orchestrated, organized, and compartmentalized. Each person has a specific function. You have the politicos, the mud slingers, the misinformers...each person has a function and they do that function only. You can see it all over Washington right now. While they sometimes infight they mainly stick on message.  So to combat that we need to be as organized and make sure that we have a person at every forum they are at. Each person we send will become an expert on the person and that persons area in the Neo-Con movement. In other words shock troops. Are people would have to be aggresive and on point all the time. Everytime Perle shows up on a show we have Mr. X there to combat them on air backed with facts.   They say one thing but are doing something different...then we show it on paper or in graphics. Now the problem is that we will be chasing them all the time so we will have to at some point go on the attack and change the debate to our positions.  It will take many resources and we will have to give chase but at this point it is what we can do. We just need to systematically take them on as they did us years ago. As we get stronger we can then change the debate and make the changes that we all need.  Ponder....  '),(1196,1234212,'07,April,2003','2003-04-07 00:00:00','   Kim Alfano...the new RNC cheerleader...Coulter has been sidelined it seems. I have seen Kim on a lot lately. I guess maybe finally people have realized that Coulter is just some idiot that got a law degree and then styled herself as a person of the people all the while wearing Cartier and Prada...she can\'t speak for more that three seconds without a lie coming out of her mouth...not to mention her whole age thing...she doesn\'t even know how old she is...or at least officially she doesn\'t know.  Now I will throw Kim a bone here cause she is new and really inexperienced...so here is my advice dear. Start talking crazy...oh wait you\'ve got that already...hell go do some guest spots on Rush and maybe Hannity.   She is fresh faced and so far hasn\'t really made an ass out of herself to much except that she doesn\'t know the truth and still does lie...but I don\'t think she has had time to wrestle with that nasty little voice in her head that has to be telling her that she is lieing...lets hope we can save just one from the evil clutches of the Republicans.  Well I am off to surf and look at notebooks....I think I have decided on something...it won\'t be uber box...but smaller and more portable for guerilla work on the road...just needs to be reasonably fast but long on battery life and small in size.  I am leaning towards a Sony sub-notebook....a little more expensive but more useful and stylish...it would be good because then I can have it with me all the time. I would like to get something at around 3-4 pounds ready weight...so I will be cutting myself out of bigger laptops and higher power graphics...but if it is small and with a 12 or 13 inch screen I won\'t want to game all that much on it anyway. Just as long as the graphics are reasonably snappy.  Peace...off to see if I can find something to fit the bill.  '),(1197,1234212,'07,April,2003','2003-04-07 00:00:00','   Well today they have a bunch of little twinkies running around pretending to be nurses...it is almost funny.  Well today I go down for a flow loop to see how my lungs are doing. I know that they haven\'t come back because I can feel it. So we will see what happens a little later.  I am just now tired and weary...but hey...I will be getting out sooner or later.  I will be back on later I am sure.  Moveon.org anounced the winner for the bushin30seconds.org award. Without further ado  The winners .  The real war begins now...and Bush is all out of ammo. Go ahead pull Osama out of a hidey hole...it still doesn\'t make us safer...and if it does are you going to repeal all the laws that you have placed on us that strips us of our rights...I doubt it. I think that even if we find the \"boogy man\" our rights will be clamped down even tighter than before.  In the end we will win. They are but a few...we are the masses.  '),(1198,1234212,'07,April,2003','2003-04-07 00:00:00','   You know its bad when the Swedes are after ya.  I think this is a great idea and I hope they can find a way to send money to help us get Bush out of office.   Democracy Aid 04\'   We unfortunately need a little lesson in how things are run...and again I hope they can find a way to get the money here so we can fight Bush.  The writing is on the wall...now all we have to do is read it!  I think you\'ll like the Idea B. It really is original and very well thought out. They make no recomendation on who to vote for...only a recomondation of who not to vote for.  They have a link to some law on it all at  Findlaw .  Not only do they make safe cars...but they want to make the world safe...they are just such a great people...  '),(1199,1234212,'07,April,2003','2003-04-07 00:00:00','   I got a nice short message from an old friend today. I won\'t mention her name right now just simply because I want things to move forward and all be well.  Thank you for you kind words and always know that I have never held any ill will towards you or anyone around you. I hope that things are going well over at your place...and I hope your family is well.  I am doing ok myself...things are rough but I always know that I will land right side up cause they just can\'t keep me down.  Well I will leave it at that...and just know that I do really care for you and I am sorry for all the crap that passed between us all...I had never intended for things to get so bad or out of hand...and I just hope that you can forgive me for any harsh words that I have pointed at you.  Peace :-)  '),(1200,1234212,'07,April,2003','2003-04-07 00:00:00','   I thought I lost that post below because of this piece of shit laptop....I am so tempted to just hop onto that one site and order a laptop and then just face the music for me putting 1600 on a credit card...  But because I am now \"responsible\" I won\'t...but it is soooo tempting. If I were working more regularly than I wouldn\'t hesitate...but I am not...so I will wait.  Oh well I will let this little thing crawl along until it truely dies. Then I will get something newer that will work.  blah...  '),(1201,1234212,'07,April,2003','2003-04-07 00:00:00','   I am now in ICU because of my alergic reaction like it is something new. I have had an alergic reaction since the first day I have been on the cefepime...it is in the same family og Zosyn abd Miripenam...all pennicillian family members...so I am just alergic to them.  Just pump me full of benadryl and steroids and all will be well...I am pretty good at screaming when something is wrong...so just lemme be and lemme be drugged.  I missed the O\'Niel interview on 60 min. tonight...I really wanted to see it just to see what he was going to say...oh well I am sure I can find it somewhere on the net.  Well they don\'t know I am connected using my Cell Phone...hehehe I am a sneaky bastard...and they don\'t need to know.   I have to get over and rebut something I read on another forum and make my mark....actually it will be pretty easy to do seeing the argument the guy made was just retarded. I have a new place to play...and a new place to try and spread the love.  I have said it before and I will say it over and over and over again. I don\'t want to make up your mind for you...I don\'t want to be your \"guru\" or leader...instead I want you to learn to think for yourself...regardless of where it takes you even if it is polar opposite to what I think or feel.  I just want to make sure that people get to their conclusions through serious thought and through good fact based sources. Don\'t just recite jingles or tow the party line. Fifteen second sound bites are no longer acceptable in these times. If you choose to go with these little sound bites or choose to over simplify a problem then you should be relegated to obscurity.  People like Rush Limbaugh, Pat Robertson, Karl Rove, Sean Hannity, Anne Coulter, and all those people on that side really aren\'t all that smart when it comes down to it. Sure they have been successful in spreading their lies and puffing out their chests and making \"the liberals fear them\"...but in reality we don\'t fear them, at least I don\'t...and it is because I know to my bones that these people are not moral people. If they were moral people then they would not attack in the manor they do. They rarely even attempt to debate on issues because all they know is the overly simplified explanation of most issues that have been spoon fed to them by whomever was their \"guru\" during their formative years. They rarely can stand without each other and you will rarely see them entering into the fray unless they have backup.  They have a shark pack mentality and as soon as they smell blood they attack. I call them sharks instead of wolves because sharks lack spine...they are really a rather fragile animal once you get past their tough skin (a good example can be seen in Rush and how he is now the \"target\" or the liberal law enforcement and courts for his illegal actions...I have something to say on that later) they really are nothing but a bunch of mixed up diseased parts that in reality would simply die off if they didn\'t have the tough skin.  We don\'t need to be afraid of them...instead we need to dismiss them. Their views are not valid, their ideas are not original (because they all keep clucking around the chicken pen repeating the same things over and over), and they are just morally bankrupt.  We as people...not liberals or conservatives need to rise up against these people and just tell them to head back to the looney farm with their ideas.  One of these days I will outline how and why I think the way I do about our system and then show that even communism isn\'t valid in the end because it was taken back to the rich.  I want people to be prosperous, I want for us all to live in some form or mode of modernality that allows people to really be free...however as long as we have radicals on only one side of the fence (and that is the conservative side in my view) then we will never be able to break the shackles that bind us to this horrible fate that awaits. We have to make a consious effort to figure out why things are happening and then remedy them. To oversimplify now only condems people to continue to repeat the mistakes of the past.  Challenge them on their own turf and I bet you they will crumble and fall off into the sea. Once that happens we can then move on.  '),(1202,1234212,'07,April,2003','2003-04-07 00:00:00','   bugger me...I lost my post because my connection took a shit....   I think that I will be going to a new place for blogging...all I have to do is make sure I get all my previous posts and stuff so that I am still on point.  Blah........  '),(1203,1234212,'07,April,2003','2003-04-07 00:00:00','   Sorry no real posts today...I feel like shit.  This one bothers me...and it makes me frightened...   Does this mean I loose too? I hope not...and I thank the taxpayer for keeping me alive.   Bush has got to go...if he isn\'t deposed we all loose...and we will all suffer. Believe me health care is of the utmost importance. We as a nation cannot afford to ignore and not have national health care now. If we ignore it and continue gleafully down the path we are on we will assure ourselves our own extinction. This is not alarmest of fear mongering...our health care system is in such a state of shambles that we cannot even get enough flu-vaccine and children die.   With how the virii and the bacteria are mutating ever faster in the face of our drugs...and the fact that many people are breeding these little super bugs and not getting adequate care or sub-standard care and only half runs of suitable anti-biotics that don\'t kill but ultimately strengthen the little buggers. Those little buggers propogate and soon all we have are very stubborn bugs and no drugs to treat them. Now all bugs will become resistant to anti-biotics over a long enough timeline.   We only shorten that timeline by not having adequate care for those who are sick.  We can only beat them or have a shot at beating them if we go national and have a concentrated effort towards genetic drugs, therapy, and modified anti-biotics...maybe even creating our own super-bio killer bug that kills all the bad ones yet allows people to be safe.  That is the only way we can even attempt to survive. We must do this now before we all die.  Down with Bush, up with Dean or Kucinich...up with anyone who is at least a good man.  '),(1204,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I have my laptop all tuned out and it is hella fast. I got most of that done last night. There are just some niggling things that I wonder about.  Such as...  Centrino technology. With this new processor the Pentium M it has an uber cache on it. 1MB of L3 cache along with the smaller L1 and L2 cache. Now they have for this on the restore disk a more aggressive setting for L3...I wonder what the registry entry is? What does it do. They also have an .inf file for the Pentium M that you install and from what I can tell it does indeed seem to speed things up, but what is it doing.  Also there are a other niggling bits here and there but overall it is performing well above what I would expect. I have a pretty fast computer at home and it runs really well and fast. Going from that to my old laptop was like torture. The old laptop was a 366Mhz Celeron with 192MB RAM. So it was a huge step back (well at least since I really upgraded my systems...when I had my Gateway it was a 333Mhz Celeron as well...then it wasn\'t really a speed difference)...but going from P4 to Celeron was like someone had just kicked me in the groin. Now going from my P4 to this Pentium M I must say that I don\'t see much of a difference even though the actual clock difference is around 1Ghz from 1.4 on this laptop to 2.4 on my desktop (its overclocked). My laptop has 512MB RAM (yeah hella sweet) and it runs fast but not much slower than the 1GB of faster RAM in my desktop (which is also dual channel unlike the single channel on the laptop). I have a decent Video card in the desktop...good middle of the road nVidia gear...and on the laptop I have an FX5200GO with only 32MB RAM...it isn\'t balls out fast and probably will lag greatly on Half Life 2...but in SOF2: Double Helix it runs well and I don\'t see many frames skipped or mis-rendered..it runs really well.  So what does this mean...have laptops finally come up to the point that we really don\'t need desktop boxes anymore? I am going to venture and say for most people yes...laptops are finally at the nexus where you can get a decent laptop that is perfectly respectable and able to do most of the tasks that a full size computer can do. Sure they have a slight price premium...usually around 200-300 for comporable setups...and sure you can always get the latest and greatest bleeding edge graphics and sound for your desktop...but that is changing in laptops. Right now you have good video in many of the higher dollar machines. You can get a Radeon 9600Pro 128MB for laptops on the high end...not balls out blazing like the latest greatest 9800XT 256MB...but the 9600PRO chip for mobiles is damn fast and really remember that for everything most people do that card will be perfectly capable for several years at least until the next Windows OS (Longhorn) comes out. Also there has been a push to make things in laptops truly upgradeable...we now have some platforms that you can swap in and out a nVidia FX5700GO 128MB to a ATI 9600Pro 128MB with little fuss.  So things are changing...and I believe that I will probably be investing more and more into laptops and mobile systems simply because they have finally hit the point where even though desktops will always have the latest and greatest and fastest...but the laptop is just behind them giving the option of being able to carry a world class computer with you without having to tote a big heavy box, monitor, and the rest...when carrying a laptop is a relatively simple affair...one case, one power brick (if you plan to charge or use your laptop more than a few hours), a mouse if you must (I carry one because integrated pointers anymore are great but I love the mouse), and then other things that you may want...all in less than 20 pounds.  So I will probably be swinging that way...maybe getting several notebooks each for specific missions. Maybe someday a Sony ultra portable (those small 3.5 pound babies with small 10-12\" screens and a Centrino 1Ghz) and then an uber P4-3.20Ghz 17\" Ati 9600Pro 128MB, 2*60GB 7200RPM Raid 0 HDs, 1GB of PC4000 RAM, DVD+-RW/CR-RW and a host of other goodies...then get a Tablet PC.  Each for its own use...each different for specific tasks...a lot of money and only a dream...but certainly would still weigh less that one uber-box...and infinitely more portable...so that is my new mandate...no more desktops...only laptops or portables for this guy.  I may actually write something of value later tonight...the past couple of days have been rough and I haven\'t felt all that good. All my meds have been readjusted upwards. Now I am on more of everything and I am having to cope with it all. However at least they have upped my pain meds as well because that is the hardest thing...and now at least I am not having as much pain as I have had over the past few days (it came with higher doses of steroids)...so as I get to feeling better I am sure I will write more...  '),(1205,1234212,'07,April,2003','2003-04-07 00:00:00','   I am watching the 6th Day and I must say it is an interesting movie. Visually the graphics are a little cheesy but I love those heli-planes things...the concept is sound...it would just need a lot of work to get it to that point. We could do it.  Today I got a good two hour nap and I feel a little better overall. I just wish I could get some good sleep so far though...it ain\'t happening.   I got my wireless router and got it all set up...now I am waiting on my port to be activated. I even set up the security on it. Once I read a little about it the thing was easy to setup. It has some pretty amazing features on it. I am impressed. Now I just need to get the rebates printed out and sent off so I can get my money back on it. Basically I got it for 39 bucks...and it has a four port router on it as well...that is k-rad...and the box itself is attractive. I may have to get some more wireless equipment at some point.  Also Intel is introducing Centrino 802.11G now...so I will probably upgrade my wireless setup to that...they will offer the mini-pci card plugins and you will be able to replace it and still have full Centrino certification. It will upgrade it from 11mbps to 52mbps wireless. Plus it is backwards compatible.  I am going to get a candy bar...and did I mention I love this laptop...I mean it is really k-rad..and even with me having to access through my cell phone...it is still damn fast!  Baaa  '),(1206,1234212,'07,April,2003','2003-04-07 00:00:00','   I have seen the future of the OS and it is good.  Sun has a new idea...well not new but a very good implementation of an idea thats time is due. They call it \"Project Looking Glass\".   Sun\'s Looking Glass   It looks interesting.  Got my wireless hub...just need the connection to be active.  Tired and barely able to see...going to get some sleep.  '),(1207,1234212,'07,April,2003','2003-04-07 00:00:00','   My new home for political Debate.  Go give it a look...a good place to wet your appetite for some good ole\' political debate.   Debategate  '),(1208,1234212,'07,April,2003','2003-04-07 00:00:00','   The whole world is overwhelming today...I just don\'t know.  '),(1209,1234212,'07,April,2003','2003-04-07 00:00:00','   Some good points on Commondreams.org   The new Eugenics of Education.    Dean gets some credit...Bush looks like an idiot.   It is nice to read well written articles...it is hard to find people who write well.  '),(1210,1234212,'07,April,2003','2003-04-07 00:00:00','   Well JPL lost control of the plucky little Mars Rover...they describe the problem as that the craft has gone into \"safe mode\"...so I guess even Microsoft can screw up a glorified remote control atv. I can just see it...there is this mutli-million dollar machine sitting out there probably with a little LCD screen glowing a glorius blue with a cryptic message of numbers, periods, letters, and the message if this is the first time you are seeing this screen....  The Rover has BSOD\'d....that is so funny! Well I dunno if it did or not...but I guess you could compare its safe mode with that of Windows...hehehehe..and we all know that safe mode is preceeded by a BSOD.  Oh boy...I am sure those rocket scientists will be able to get it back up and running.  '),(1211,1234212,'07,April,2003','2003-04-07 00:00:00','   This movie or documentary or whatever it is about Noam Chomsky and his book Manufacturing Consent is really damn good. It really goes a long way to explain what is going on...especially now in the media.  It is a long thing to watch...but it is great and tells a great story. It is something that people should burn a few brain cells and think on...he really is a great mind and what is great is that he offers them for free...he gives them to you freely. All he asks in return is you think about it...and that is all I want for people to do when I post my thoughts or ideas...just think about them. You don\'t have to agree with them and can even be combative with your view...I support that because I support free speech and I support your right to be whatever or whomever you want as long as you afford me the same right.  All this is in jeapordy right now because the Republicans are trying to clamp down on free speech and dissent of anything the president or the cabinet espouses. If you dissent you are un-patriotic...calling people those names isn\'t appropriate because how do you defend against that? You can\'t just say \"well I am not a Nazi\"...the damage is already done. The person who the charge has been levied against simply cannot dispute a charge like that.  Watch the movie. You will see.  '),(1212,1234212,'07,April,2003','2003-04-07 00:00:00','   Bush in Arizona I think and is stumping for the Patriot Act...it is so transparent it isn\'t even worth ranting about.  He is of course in front of a friendly audience (mainly military)...but what I don\'t understand is why those guys are clapping because it is their asses that are being put on the line and all the while the Administration is cutting funding for pay, housing, health care...all for the military.  I just don\'t get it...I really don\'t...the military should be the most upset about all this policy and crap...but they seem to like dieing without cause and going to war under tha false pretense of WMD\'s and now Liberty.  I just don\'t get it.  '),(1213,1234212,'07,April,2003','2003-04-07 00:00:00','   I want to clear something up real quicklike...what I was saying in this previous post is that we changed the debate and we forced the rest of the candidates to take on our issues...we also forced the President to do the same...as we watched in the SOTU speech last night he has had to put his grand plans on hold so he can come back and fight the issues that we are pushing, not what Rove, Rummy, RNC, DNC, and DLC were trying to push on us. Instead we made them all step back and have to deal with the issues as we see them.  For that our movement is defined even more so tha Dean himself. We are behind him of coarse but as he likes to put it without our support he and the issues are dead and we seed our authority as voters to special interests. We have taken it back and made it ours once again.  '),(1214,1234212,'07,April,2003','2003-04-07 00:00:00','   Thank you Michael Moore...we Dean supporters needed to hear that from someone outside of our camp.  While you support Clark and we support Dean we just needed some validation that what we are doing is resonating within the community at large. I wasn\'t so much shocked that we lost Iowa as I was dissapointed that we didn\'t carry a larger percentage overall.  While it was a loss there were some bright spots and Michael Moore points them out quite well. We didn\'t loose at all we actually won...all of us won big. We proved that we could bring out the vote through our new tactics of online activism and through our Drive. While we didn\'t carry the caucus we got the vote out and we nearly doubled the turnout from four years ago. That is an accomplishment because we came in and shocked the party back to life.  All of the Issues that Dean has brought to the front have become the issues. We are no longer tap dancing around things and we have finally started to get \"god, guns, and gays\" out of the debate...or at least we are trying. The real needs are being debated such as education (which would have been pushed to the back burner yet again if it hadn\'t been for some brave people to stand up and say \"no child left behind\" has failed), health care.  Why do we know they are on the table...because Bush put them back on the table last night. By doing that whomever the democratic candidate is that runs will now have the onus to point out the Presidents failure to even fund his pet projects instead choosing to give out huge subsidies to his health care buddies.  So Dean supporters we need not dispair because regardless of how the chips fall, our message has been sent out and received loud and clear. We want change and we will make it change period. No more of the same. No more mediocrity. We have already taken back the process...now we just need to take back the country.  Thank you Mr. Moore for giving credit where credit is due.    Michael Moore\'s Encouragement to Deniacs.  '),(1215,1234212,'07,April,2003','2003-04-07 00:00:00','   SOTU Day plus one.  All the media seems to be able to come up with is that Teddy scowled and Bush threw daggers.  The Republicans are now saying it was inappropriate for the Democrats (especially Kennedy) should not have been so openly disapproving of the President how Kennedy seems to be challenging the President.  Now this is when I really have to start to get very very angry. Disapproval and dissent is part of our Democracy. Calling into question any policy by any politician on any side is never hate speech, it is never questioning someone?s patriotism, it is never inappropriate to have open public debate on any policy that faces our nation.   If you are one of the pundants or people who have said or say \"well he commands respect because he is the president and he shouldn\'t have been doing that, it was disrespectful\"...you need to shut up and stop that.   Yes shut up and stop that. If you attempt to curb debate on an issue because you think it is hateful to question your position than you are no better than a hardliner Communist or a hard line Nazi. They both curbed public discourse, called their opponents un-patriotic, everything that the Republicans are saying about Democrats lately.   There has been a lot of fuss about some people comparing Bush to Hitler and many groups have come out and said that is anti-Semitic, wrong, and hateful. I am sorry I don\'t agree with any of those arguments especially considering that the Republicans have been using the same tactic against Democrats for years calling them Socialists and Communists...saying that the Democratic way is the way towards destruction and away from freedom.  It is ok when the Republicans engage in that tactic but it isn\'t ok when Democrats even say \"hey that isn\'t right!?  What do I have to base this on...how do I come this way. Well I read and watch TV and I see what they do.  They compared Max Cleland of GA, a veteran of Vietnam who lost three limbs (he is a democrat), to Osama in a commercial because Sen. Cleland openly questioned the Presidents policy on the \"War on Terrorism\" pondering whether or not a different approach could achieve the same goals and whatnot. He also questioned tax cuts and other Bush policy. So the Republican Party put together a commercial that had Cleland and Osama in the same frames (albeit only for a second or two...but enough for even a slow brain to register) and said that because Cleland wasn\'t supporting the President and his policies then Cleland might as well have been for the terrorists.  This didn\'t raise the ire of the press or the Republicans because it was in their interest...the Dems didn\'t make enough of a stink of it. It was just passed over and left behind. The campaign was successful and Cleland lost his seat and a Republican was installed.  In the New York Post just recently one of their columnists wrote a piece on Dean and his supporters comparing their flannel shirts to Nazi Brown shirts. The republicans didn\'t make a stink about it then...nor did any of the Pro-Jewish groups out there (like the Anti-Defamation League) make a peep.  But when you suggest that some of the Bush tactics are similar to how the Nazi party took utter control (not even directly comparing Bush to Hitler) of Germany in the 30\'s. Well that can\'t happen. Fox news jumps all over it, the anti-defamation league comes out in droves and says that you can\'t say something so insensitive, CNN who is not all that right leaning even makes a fuss out of it...then there is something wrong.  You can actually trace or compare some of Bush\'s policies and laws passed under and by Bush very aptly and evenly to laws that were passed under the Nazi party.  Some of the first things to go under the Nazi party and the Communists were basically the right of dissent. Our right to speak out against our leaders without fear of being called un-patriotic or the fear of being jailed (we are not at the point of jailing dissenters here yet...but under some of the laws that have been passed it is legally possible for the president to jail a dissenter without so much as a trial by going to the extreme with the laws...again it hasn\'t happened that we know of yet...but it is possible.).  The laws include the Patriot Act, the Security Enhancement Act of 2002, and the soon to be introduced (although it probably will die) Patriot 2. Under the extremes of these laws you can be detained without cause, without access to an attorney, access to a judge, placed in isolation, your family will not be notified that you have been arrested (you disappear), called an Enemy Combatant (which the president declares...no court or jury...but the president has lone discretion a huge violation of the tenants that not one man can imprison another without a jury or consensus of people...making GW a pseudo-dictator), then after being declared an Enemy combatant you can have your citizenship revoked (regardless if you are a natural born citizen or immigrant), and ultimately with the Security Enhancement Act you can be brought before a secret court (again without a lawyer) and found guilty and then sentenced to death...all this and you have no legal recourse at all.  Now that is the extreme of how the laws can be used. Has it happened? No. Is it likely happening...probably not? Is it possible and could it be hidden...yes yes and yes...it may never be applied in such a manor but that doesn\'t really matter because the law is there and it can be used.  Those laws are similar to how the Communists purged dissenters; how the Nazi\'s took control of the government...not that the policies or aims are that same...it is the tactics that are the same.  If some Bush monkey saw this they would accuse me of comparing Bush to Hitler or to the Nazi\'s, or to Stalin...sure we don\'t have death camps burping out the gasses of the burned corpses in their furnaces out onto the landscape.  So to those who say it is wrong to question, to take to task something you don\'t agree with, to openly dissent...then you need to read a history book because you are committing a crime. You are repeating history...a history that many better men than yourselves gave their life for.  One thing that we can say...and it is pretty clear is that the United States in collusion with Israel is committing mass crimes against Muslims. I am not on the Muslim side...they have some crazies over there...but we also have our own. By us declaring war on Terrorism we are actually declaring a holy war on Muslims...perhaps a Jihad against them.  I have heard over and over again that it is the Muslims doing the suicide bombings and killing each other as well as our troops or Israeli citizens.  I think if we were to step back and ask this simple question....why are they doing it, what drives them to want to kill themselves to kill us. There is a reason because it is not in human nature to willingly without great cause put us in harms way like that. Something has to be happening so bad in their world that drives them to do that. What is it and can we help stop it?  Instead we have stated that they are \"evil\" and go off and kill them wholesale. We are committing crimes against humanity. We have to be held accountable sooner or later.  We just need to withdrawal from the world as a military power because we have not used it responsibly and instead abused the power for far too long.  By withdrawing from the world stage militarily and instead engaging them through diplomatic means and just accepting that we will not always get what we want is the only way forward. Right now our military is the primary diplomatic arm or hand that we are presenting to the world. That isn\'t a safe or wise way to get things done.  There is a lot to do. There is a lot we can do. There is a lot that we must change. We won\'t be weakened by pulling our military back to a true defensive posture, we won\'t be seen as weak if we negotiate with other nations, instead we will be seen as strong and confident able to stand on merit and not on military might.  When you have to use the military to push your foreign policy then your policy or diplomacy has failed. Once it has failed then we have failed, and when we have failed then the nation falters.  We can no longer afford to fail on the international diplomacy field. We can only start to win again if we make a move back to diplomacy and international co-operation. We cannot survive without it.  Think about it.  '),(1216,1234212,'07,April,2003','2003-04-07 00:00:00','   By the way I love moveon.org and their commercial. I actually gave 20 bucks to help run it. That is the first political contribution I have ever given in cash...I have volunteered before...but now I have given money.  Go see it if you haven\'t already...it really is powerful and true.   Childs Pay   High Bandwidth Version   Low Bandwidth Version  '),(1217,1234212,'07,April,2003','2003-04-07 00:00:00','   I am downloading a video of Noam Chomsky called \"Manufacturing Consent\". I don\'t know how he would feel about that seeing that I will essentially be getting the work for free and I would be glad to send him money for the video...  That is my major dilema anymore with downloading things off the internet. I don\'t want to buy music from Record Stores or record clubs...not because I don\'t want to see the musician paid...but because I don\'t want to see the record company paid...simple as that.  So when I have to have a CD...I will try and download it and send a check to the artist or musician directly if possible, or I go up to my local record store where they sell used CDs.  So there it is...I would love to send Mr. Chomsky money for his work and I may well just do that.  '),(1218,1234212,'07,April,2003','2003-04-07 00:00:00','   Well he finished.  I\'m scared...we are in trouble. When will the Democrats challenge him...when will they openly \"boo!\" him. Why aren\'t we doing that?  The president\'s path that he has laid out is one of destruction. Partiot renewal, the call for making his tax cuts permanent, calling for a possible constitutional ban on gay marriage...all of this is wrong.  Never seeking a permision slip to defend America...um well we can\'t go it alone. Rome, France, Germany, England, the Greeks, Egytpians, and numerous other states or world powers have tried to go it alone with little or no co-operation have always failed and fallen into ruin.  Bush just laid out...the path to ruin. We go forward on anyone of his plans and guess what...we will fail.  '),(1219,1234212,'07,April,2003','2003-04-07 00:00:00','   It is telling when you look at your president and you only see the empty face of evil staring back at you. His eyes filled with hate, his brow furrowed with lies, his mouth moving telling me why. He tells nothing but lies.  Soon he will be gone...one way or another...he will be gone.  '),(1220,1234212,'07,April,2003','2003-04-07 00:00:00','   Oh yeah...   Commondreams State of The Union  '),(1221,1234212,'07,April,2003','2003-04-07 00:00:00','   Tonight is the state of the union speech. It will be interesting to hear what whoppers he tells tonight. I am sure there will be some great points to work with.  One thing that I have heard that he wants to do and I actually agree with is removing the down payment requirements on FHA financing for houses. It is about time...the down payments have always been the most difficult thing for many to get together. While FHA financing has always had a low down payment percentage it has always still been difficult and the main obstacle for people trying to buy a home. So that I agree with and maybe at some point soon if I ever get well enough I may even be able to afford a house of my own.   However the rest of what has been hinted at is just plain bad. Bush wants to extend health care out to more people but it will most likely be another weakening of Medicare and Medicaid with huge handouts to the drug and insurance companies.  We all have to keep our eyes and ears open and listen for those buzzwords...because the speech will be packed with them.   It is going to be interesting and I am sure I will have something to say about it all. We shall see.  I just hope Dean doesn\'t stumble anymore...because he really is a better candidate than the rest (except Kucinich) and would at least be able to take it to Bush and hold him accountable for what he has done.  '),(1222,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I had to think about Dean\'s performance last night when he was giving his speech after he came in third. After having thought about it I have to say I was a taken by surprise. I know he is angry, I just didn\'t know he was mad.   I understand exactly where he was coming from...and frankly I have made the same noises and expressed the same frustration many times over. Until now being angry was the right thing but now he needs to shift from sheer madness to a little bit softer approach. He can still be mad, indignant, and pugnacious, but he can no longer run on just anger (because we see Edwards and Kerry moving forward because of their \"nice guy\" persona). It just needs some tweaking and he will be in good shape.  The performance was just over the edge last night. I understand the grunt and the \"wraaarrr!\" feeling because I do the same thing when frustrated and angry (its a better responce than just punching a random wall or person).  It was just over the top and I feel that he may have to pay a little on that for a while...but I am sure of my candidate. I still support him...and in the end I feel that he will come out on top.  The rest of the field (except Kucinich) are just DINO\'s...they haven\'t stood up to the Republicans and have been in colusion with them...  We need our country back.  '),(1223,1234212,'07,April,2003','2003-04-07 00:00:00','   We have the State of The Union Coming up tonight...that will be fun to watch what the Boy King will call for...I can only imagine.  Permanent Tax Cuts, War is Good, We Are Safe, 9/11, 9/11, Patriot, Saddam a bad man, \"Naaaahhh guunnnaaa do itttt\", \"its baad, itsss badddd\", and probably the inevitable outlining of the new \"Axis of Evil\" that will include Iran, Zimbabwe, and France..cause you know them Frenchies are giving the starving children so anthrax tampons to lob at the Iranians who will then retaliate against Bush\'s butt buddy Sharon (that crazy little fat jew...really look at him...he is as wide as he is tall...reminds me of Violet in Willy Wonka...and yes he is a jew...not meant to be derogatory in this sense...although he is also that way as well). Then we be releasin a rocket...from out da pocket...oh yeah heh ho heh ho!  This is all nuts...its all upside. All we can see is GW moonin\' us as he pockets the wealth of the nation. If we don\'t vote someone else in...then we damn well deserve it (well we don\'t and those of us who don\'t vote for him should be allowed to make our own little better nation). I just wish more people could see through all the Bullshit on both sides of the fence...  I would characterize what we have now as this. The Republican Side that is nothing but death and destruction...it is then seperated by a middle ground that is lush and lively, and then another fence with the Democrats on the other side...and guess what it looks the same...but it has been painted green. The middle ground is where most on either sides of the fences want to get too...it is beautiful and peaceful and healthy, but unfortunately these fences are much more secure than any other fence in time...as it has held us all collectively back...and kept us from that dream.  Both sides have fences...and neither fence is a good choice...but at least right now the Dems side of the fence won\'t kill you in a few years...it will take many years to bleed you dry...the republicans shoot on site just like they would like to do on the US/Mexico border.  Bahh...bed time.  '),(1224,1234212,'07,April,2003','2003-04-07 00:00:00','   Kerry got the nod of the Iowa Caucus...big deal.   I will say now that regardless of who gets the nod (unless it is Lieberman) I will vote for them. I just think that it is kind of odd that the majority of people in Iowa are against the war in Iraq yet the essentially said they were for a man who voted for the war.  At least they didn\'t go with Gephardt...he was co-sponsor of the Iraq War Legislation Bill...so at least they were somewhat awake.  What is surprising is that Kucinich would have his voters caucus for Edwards in the end if Kucinich didn\'t look like he would get any delegates.   Edwards placed second...and I can\'t figure out why. He is affiable and he is nice...he is Clinton Lite...reminds me of him. Clinton was great in his period...he is no longer relevant to the party and anyone who would willingly follow him down his meandering pathway of life is just plain foolish.  The Democrats need to take back their party from the DLC...and Edwards is a high ranking member of the DLC.  Kerry talkin about special interests is also kind of funny.  Dean has the only \"clean\" money so far (and it is pretty much assured that later the money may get a little dirtier but at least the seed was planted with small donations and not 2000 a plate dinners).  The only reservation that I would have with Dean v Bush is this very plain thing. Dean is much smarter than Bush. The powers that be have been grooming Bush lately to be this kind of lovable frat boy oaf that people kinda feel bad for when he is picked on (although they should just go all out and give him some foam headgear and some novacaine so he drools on command0...and when Dean would \"debate\" Bush, Dean would look like a meanie to our poor little boy king. That is the only thing that may be worrisome...because Dean would literally run Bush over on a debate of anything. Dean thinks fast and talks fast...and when he shoots from the hip he at least has the right target in site (unlike Bush who fires off half cocked like the pre-mature ejaculator that he must be).  When it comes down to it...Bush has got to go...no matter how it is done...whether it be through us voting him out, kicking him out, or just plain getting rid of him...I say we throw our best pretzels at the problem...because apparently that snack food has it out for him.  Ok Kerry...enough...you won. There...Heinz money will only go so far...and you did vote for the war. You and the others from Washington (except Kucinich) all cowed to Bush and the Republicans on every major issue. You voted for his policies, you supported his resolutions, you helped put the country on the fast track to hell...the only clean ones out of the bunch are Dean and Kucinich.  People are blind.  '),(1225,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I got the laptop all fixed up and ready to rumble...well as ready as I am ready for it too...now I just need to head out and warchalk and find me some places to connect up to wirelessly.   I have changed the look a little around on it so it resembles something that I like...the laptop is sleek so I made the interface a little slicker as well.  I am downloading Real Player so that I can watch my show in a bit Democracy Now!  Lately I have slacked off on my postings around the internet at all the normal haunts...and I have a pretty good reason. I am not feeling to well at all.   Today the doc came in and told me I would be here for a while and not to look forward to heading anywhere until they figure out what is actually wrong...which they don\'t know at all. They have me on all the right stuff but I am not getting any better I am getting worse for some reason.  On top of everything else I got a nice sinus infection...it took a couple of days to hit me...I know where I got it from. It sucks ass...  Hopefully over the next few days I will be back to fighting shape...I am probably going to spend a lot of time sleeping over the next couple of days which will be nice :-)  If there is anything that really catches my eye as always I will report. They have a couple of really good reads over at  CommonDreams.  Go give em a look...you might learn something.  It seems that my comments have gone away again...who know I will try and get that fixed soon.  Peace.  '),(1226,1234212,'07,April,2003','2003-04-07 00:00:00','   Oh yeah...and updating at 115Kbps takes forever!!!  '),(1227,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok...so I got Jesse off yesterday...luck bro...that box is mad fast.  I also got my laptop re-installed with XP Pro...and it is crispy sweet now...I am just waiting on my wireless router/hub.  I will be looking forward to that.  Got some bad news from the doctor today. There is no end in sight for me. He is completely baffled as to why I am not responding but instead getting worse. So he will keep me until he figures it all out.  I imagine that I will be here a month soon....who knows but this is a hell of a stay.  Well I am very very tired and extremely overspent on my energies...so I am going to take a nap.  Peace  '),(1228,1234212,'07,April,2003','2003-04-07 00:00:00','   Oh yeah I feel like shit...also got all of Jesse\'s box put together overnight here...now I just need software to load onto it...got the wrong stuff here...so I will have to have my mom bring the right stuff...  I also need a couple of cables that I don\'t have here...but I have them at home....  It should only take me a few hours to get all the software installed and I will be able to let the little bird fly away...far far away.  The parts we got are primo and they all appear to be working...I just need to get the software fixed so it will actually do something other than sit there like a lump of silicon and plastic.  Once I get his boxes done I will sleep for a while and then I will work on my systems. I am going to install XP Corporate Pro on this laptop (it has XP Home installed)...and my main box can stand a refresh...so I will do it as well. It will give me something to do over the next few days...keep me pre-occupied.  I ordered a wireless router/hub yesterday so that I could use this laptop wirelessly here and at home. I will need to get GM a wireless card...but that won\'t be expensive...I dunno exactly what I will do yet...but I am looking forward to having this thing wireless...especially when they get my port turned on in here...cause then I can walk around and compute...w00t! Oh yeah and I got that wireless and 4 port hub/router for 29 bucks...oh yeah gotta love rebates :-)  Oh yeah...try and catch on CNN Presents on Muslim Radicalism and that whole story. It was very well done. A lady from Singapore (where the real war on Islam is...) said that the rhetoric from both sides is from the pulpit...Bush is a fundamentalist just as Osama is...which if you were to take just the words of the speeches they each give and compared them without face or voice...they are the same. Both preaching radical views of the world and casting people as good or evil.  It is really interesting.  I just hope this year we will change and get back on track. There is to much at stake now...and we can\'t afford to be wrong (we can afford to loose, but not wrong).  23 people so far dead...but that is a low number.  '),(1229,1234212,'07,April,2003','2003-04-07 00:00:00','   Well another day another mass bombing in the middle east.  When will we say \"enough is enough...get our troops out of there\"?  I understand that if we leave now it will fall into chaos and many more people will die and we would possibly have to face an even uglier government there than what was in Saddam\'s government.  I just don\'t even know people...it is just so bad and so many people here are absolutely oblivious to how bad things are getting over there...all we hear is \"ra ra...the economy is going good, the war on terrorism is going well, we are winning in Iraq\"...when in reality there is nothing of the sort.  The economy is in the shitter...sure the stock market is rising and corporate profits are rising...because they have been firing people and forcing their workers into longer harder hours...that is how productivity goes up.  The stock market isn\'t real money...remember that folks...you are basically saying when you buy stock...I bet this company will be around in a few years and making money...nothing more. It isn\'t real. The stock that you own entitles you to nothing (even less than what a dollar bill is worth) if the company goes belly up its not like if you have 50,000 shares and you trade it in for a colour copier from the office...its not real.  Of coarse stocks are rising...because taxes on corporations is low and over the past few years companies have fired people left and right and have been able to drive their current employees into doing more work...again not real...  Terrorism...ok if you haven\'t figured it out by now...than go ahead and get some duct tape and tape yourself up and suffocate...do us all a favor. We are still in yellow...or orange...whatever...but we aren\'t any safer now than a few years back (and any monkey with some maps and a few numbers could make the point that we are less safe).  Iraq...again...if you think things are going well then I got a bridge to sell you along with some nice real-estate out in this little swamp. It is all a mess and we know it...but we are going to go on and pretend (as we did in Vietnam) that we can make it better...we all know how that turned out.  We are in the middle of a crisis right now. We have cast ourselves as the \"good guys\" and those who aren\'t with us as the \"evildoers\"...problem is there are a hell of a lot more of them than us...and if majority rules (as we like to say with our democracy) then we are not the good guys.  We just don\'t know...and frankly I am getting confused at who is who...because I can see it from both sides...and neither is right or wrong.  What it all ends up coming down to is who\'s imaginary friend is bigger than the others.  I think we need to grow up a little and realize that even if we have an imaginary friend...I don\'t think it wants us going around killing each other over it...so harvey...I let you go...hop through the forrest.  '),(1230,1234212,'07,April,2003','2003-04-07 00:00:00','   arggghhhhhh  maybe later  '),(1231,1234212,'07,April,2003','2003-04-07 00:00:00','   ohh its been an interesting morning I tell you...maybe later when I can think straight I will.  '),(1232,1234212,'07,April,2003','2003-04-07 00:00:00','   Here is something for all you folks out there that like to think. Read it...good points and certainly illegal. Hmmmm...   Not just illegal but unconstitutional...    Commondreams.org  is a great wearhouse of information. They basically have people all over read what is being printed in the news media and then they compile it into one big news edition. Things do get reported but usually it is under-reported or put so far back in a paper that no one really sees the stories and they rarely get airplay or discussion.  I read it everyday just to keep up. It is like having many different newspapers at my disposal at anyone time. They sometimes ask for money and I have considered giving a few bucks over because the service is so worth it.   Anyways...make it a daily read and I bet you will become informed very quickly of how little you actually hear and know...and then think of all the news that we still don\'t get even through these non-mainstream sources...it is very similar to a method that Chomsky uses to gather data...it is out there...but in tidbits and such. He and his coleagues swap folders of clippings back and forth and then weave together the news as it really is.  Good way to learn...you have to learn to learn.  '),(1233,1234212,'07,April,2003','2003-04-07 00:00:00','   Well there have been some shakeups over in Dean country...  Joe Trippi has left the campaign...I just hope this doesn\'t hurt to much. He is credited with the movement but not the poll numbers. He most certainly organized the overall movement but so far from what I have seen that doesn\'t always translate into votes in the vote box.   It is sad to see him go...but maybe we can refocus and move forward.  The only thing that I have a little bit of a problem with is that the campaign may be having money problems. Now we all pretty much understand that running a campaign is expensive but my question is 40million already spent? I dunno...but that is a lot of money to go so fast.  Well I donated regardless...and I think he is the best man for the job.  '),(1234,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I gave to Dean today. I really feel like he is the best candidate out there. I was only able to give 25 bucks...but it is better than nothing at all.  I would be updating more frequently like in the beginning but with all the drugs they have me on right now I am barely able to read the screen and corolate what is being typed to what my fingers are doing...it is very strange and very frustrating.  Looks like Shrub and his little patches of weeds are about to have to ride through a Briar patch with this whole David Kay \"there were no WMD\'s before we went in\"...Now let me get this straight Bush is calling for more time to let the inspectors do their job and find them all...but when it was the UN he wanted to get them ouit of there as fast as possible so he could ride to battle.  Time is ok for us but not for the UN? Whathefuck!  Oh yeah and the 9/11 commission has now said it will need a couple of extra months to complete its work as the original deadline is in May...they need until July or something like that. The Republican \"leadership\" in the house, senate, and white house have all said that they won\'t get more time.  The claim is that they have had enough time and we should prepare to move on...well tell it to Mr. Shrub when almost everyone of his major speeches includes references to 9/11 and the \"Evil Doers\".  The whole \"they have had plenty of time\" thing is really moot because the Administration has drug its feet in coming forward with any documentation about 9/11. Basically the Administration is at least hiding something...what it is we don\'t know...but they are hiding something...and because they drug their feet they ran out the shot clock leaving the people without answers.  We just all need to step back and take a good look at the situation...once you do if you are not enraged at the sheer volume of lies and well just plain lies.  They are lieing to us...they have been and are continuing too.  It is time for us to rise up and take it all back...this is horrible. If I could make sweeping changes right this minute I would make such sweeping changes and so many people would be hanging from the Gallows...we need to hold our \"leaders\" accountable and string them up.  It all makes me sick.  '),(1235,1234212,'07,April,2003','2003-04-07 00:00:00','   I have a lot to say...really I do...but I can\'t really see well enough to type. My vision is blurry and my hands are kinda shakey...it just is a pain in the ass.  I am glad the first primarey is done...but you know Joe will continue...he just needs to step down. This is all rediculous.  You have Kerry, Dean, and maybe Clark/Edwards...  Well it is getting close...and unfortunately none of them up there in the bottom of the pack is realizing that they would be better served if they quite and support another candidate...but some of them (Joe) are dillusional...and to say Kerry is the comeback kid is misleading...  Also it will be 200,000 plus that showed up...that is a tribute to Dean....got huge numbers...and he is sitting pretty....  Here\'s Howard...  '),(1236,1234212,'07,April,2003','2003-04-07 00:00:00','   Oh yeah I had my mom bring down my old laptop so I could strip it down. I did and I am going to see if it will even bother to start up...we\'ll see...if it does then I will install linux on it and make it run...oh yeah I scrapped the case...so I just have the guts ripped out of it...if it will start up then I will make a getto setup for it and make it industrial looking like Matrix style computing...interesting indeed.  '),(1237,1234212,'07,April,2003','2003-04-07 00:00:00','   Today I got a little sleep finally and they have bumped up my meds so I am mostly numb. My blood sugar is way out of whack...my lungs hurt (and are bleeding right now cause I can taste it in my breath)...and I am just worn down. Hopefully I will have full internet access tomorrow.  My mom brought the speakers I got out of the deal for building Jesse\'s computer. They are the Logitech Z-680 set and they are jizzmistical. I have it all set up on one cart with the laptop and then the speakers and the sub setting on the railing under the cart. The whole set is just primo. I got a few pics of it...and I will get them posted soon.  I also got it tweeked out so it is running smooth. The only things that I really want to get it an SB Extagy and maybe a pocket pc...I found a great one for 199 with a little camera built in running on Pocket PC 2003...it would be an interesting toy...but I don\'t need anything like that right now...although the Extagy would come in handy.  So we will see what the techno gods have in store for me.  looking out my window I see a beautiful site...snow everywhere. I love it...  With the big speakers, fast laptop, and all the other things I have around here...I am almost like at home here.  Tomorrow I may have something important to say...who knows...but right now I have no fight in me...  Peace  '),(1238,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I checked my Blog Stats and hit counter...I am now 889 in entertainment catagory and 3485 overall. Not to shabby for this thing being dodgy at best.  Well I am tired and I haven\'t slept in a while so I am looking forward to a nap...peace outside.  '),(1239,1234212,'07,April,2003','2003-04-07 00:00:00','   Watchin Seabiscuit with my nurse Ammie...she had to step out for a minute but will be back...getting used to this whole hospital family thing...it is nice to be in a place that is like home where the people treat you well and understand that you don\'t always feel good and can let when you get a little mean just roll of their backs...  That is always the thing that I get the most worried about is that I am mean or snippy...I worry with it in my daily life because I want to be as good as I can. I don\'t like to treat people like shit...I just don\'t. I try to be as nice as possible because hey you never know when you will need that person that you are treating like shit.   I don\'t like being in the hospital as far as it is a hospital...but I am getting a little used to it because it is nice that I don\'t have to go get things. Everything is just brought to me...and that is K-Rad...  Well I have some reading to do and I am going to watch 2Fast 2Furious...its a suck ass movie...but I have only seen it once...so I need to watch it again.  Peace out homeskillets...  Oh yeah I might have to back off my utter disgust and hatred of Barbera Streisand...she is evil...mecha-streisand...only because she is a Dean supporter...so in her black heart there is something that resembles life...so I no longer hate her outright (I don\'t know why I hate her...it is just this knee jerk irrational feeling I have towards her...she just makes me wanna toss my cookies...but I can swallow the hate to be on the same side...and if you all only knew how hard it is to just even think about her for me...then you would get it)...  Oh well...  '),(1240,1234212,'07,April,2003','2003-04-07 00:00:00','   Today has been a rather bizarre day. My mouth is really bloody raw from the meds I am on and I think I know which one so I will stop taking for a day our too just to see if I am correct. I bet I am right.  Watched Dreamcatcher Tonight...weird movie indeed. A little off thr beaten path.  Well I am tired and I am waiting on pain meds then off to la la land....  Peace Mike.....  '),(1241,1234212,'07,April,2003','2003-04-07 00:00:00','   Well lemme start out by saying that I love this laptop...it is very jizz.  Today however has been a pretty shitty day over all. First I woke up to my mouth getting sores and stuff making it hurt like a son of a bitch. Really it is a pain in the mouth. However I figured out what drug is causing it and it is Cipro...one of the many possible side effects are mouth sores which I have an abundance of.  Well the power just went out here at the hospital...very strange indeed...glad I am on a lapop and can limp through it...also glad that my entire setup is a laptop and can survive without wall juice.   I may post later..who knows....  '),(1242,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok so the network access that the hospital was supposed to provide isn\'t really network access after all. It is an extra phone line for Dial-Up access to the outside.  Now don\'t get me wrong...wow, gee, um thanks for the extra phone line, really I need an extra one. I already have one to the oustide world and it costs 7 bucks a day regardless if I use it or not...I don\'t and haven\'t had a dial-up account in over 2 years now and wouldn\'t consider even bothering to pick up a small account for even 10 bucks a month just to have access at the hospital for a land line instead of wireless line to SprintPCS.  Again it isn\'t what they gave out that was the problem it was how they came to the decision and how long it took them to make the decision that cheeses me off. It took them all this time to figure it out what they were going to offer and then come in and bodly anounce it like they had just created fire.  Why they didn\'t give us just a port into the system is because of security. Ok...security. I understand that...security is security and this is a hospital...they want to make sure that their system is secure. However without any passwords how do you get it? Before the move they had computer labs open to people...really anyone who wandered in. To get onto their archaic network system to get records or x-rays and such you have to navigate through a program that looks like it was made by a four year old and have the proper access codes and passwords...and it doesn\'t even work most of the time. You could still go to the internet or use their pipes out to the internet...it was filtered of coarse but it wasn\'t filtered totaly out.  Now I know I am making a mountain out of a molehill but when you are told that you are going to have network access you would think that you indeed would have network access. So...  What I am going to do is there is a Starbucks across the street...I need to find out their range....if it is enough I may try and get some access through them...but I still have Cell Phone Access so I can use that...but hey I wanted network access.  I also have a wireless router here and I may bridge the connection so that if Danielle has wifi on her Sony laptop...which I think she has...because Sony has always been ahead of the pack on connectivity features...if she has WiFi then she is more than welcome to share the 115Kbps connection I have...and when she heads online I\'ll have her dial-up also bridged and if both setups are on than we will have combined about 178Kbps or so....it won\'t be the fastest in the world but we can make it work I am sure.  If I can get a good connection to Starbucks (which we have one close by...and maybe I can get their wireless network) then I will definitely share the connection.  I am also half contemplating hacking the network and putting my access point out there and just going online...who knows...  I am just pissed off about it all because they could make guest accounts that limit us to just internet access and not access to other network resources, they could do it the right way in the first place and have real security in place...there are so many things they could do and make it all work easily for us and them...but instead of actually doing something about it they just said it can\'t be done and moved on.  Oh well...my kung fu is strong.  '),(1243,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok just because I am nice doesn\'t mean I actually like you...I don\'t think this respiratory guy gets it. He is new and slow...and I helped train him yesterday and was showing him some tricks and stuff to make it easier.  Well he is back again today and I tell you just won\'t shut up, won\'t leave...I was sitting here eating and he watched me for ten minutes until I told him that I was eating and wasn\'t ready for a treatment...he went away for a few minutes then came back and stared at me some more doing that \"retard lick lips thing\"...and mouth breathing...oh hes back and wandering around...well I went to close my laptop when he came over...I think he gets the message...if I need him then I will call for him...you don\'t need to hear what is going on in my lungs...it hasn\'t changed from about five minutes ago dipshit.  Sometimes I worry for humanity....actually I worry a lot about humans...I think we are already extinct and we just haven\'t realized it yet. The planet will be rid of us soon...and it will be better for it.  I just wish we could outlaw stupid people...that would make the world a better place. Ban them, oulaw the breeding of mentally inferior kids (im not talking retards here...I am taking those kids who walk into walls and are too dumb or lazy to just go to the damn bathroom before squeezing out a fat turd in their underoos and that kid happens to be 30 years old...)  I guess I am just in a fould mood today.  '),(1244,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok...so the big Half Time shocker was apparently that Justine Timberlake expose Janet Jackson\'s tittie...ok no biggie...it wasn\'t a close up shot...but it looks like she might have some nice titties and it looked pierced...  Fuck the FCC....fuck fuck fuck the FCC and those republican prudes...hehehehe...betcha ol GW Shrub has seen a few of his servants titties in his days around...hell he probably had a black wet nurse.  Although no self respecting black woman would ever nurse him or any of his racist evil clan...you know they ran with the Nazi\'s back in the day...Shrub\'s grandaddy had his bank taken away from him by the Government because it was doing business (and a good one too) with Nazi Germany and making money off of slave labour in some of the death camps...  Well at least that is what history has told us.  '),(1245,1234212,'07,April,2003','2003-04-07 00:00:00','   This week has been probably one of the more bizarre weeks in a while.  Lets see if we can hit the major points.  David Kay the chief US weapons inspector quit his job and has said that we were all wrong in our assumptions on Weapons of Mass Destruction. At the beginning he was convinced that Iraq had massive stockpiles and now he sais there aren\'t any at all.  Cheney still believes in them being there. Bush has thought maybe they weren\'t there but we still have to give the US weapons inspectors more time (didn\'t he argue against that when it was the UN?) before we would know the truth.  Cheney may still yet be prosecuted in France in a bribery charge against his company Halliburton. Cheney also has had a good hunting trip with Supreme Court Justice Scalia, who if you recall is hearing the case brought by some groups (Sierra Club for one) to open up who came to the Secret Energy Meetings that Cheney had. Scalia has refused to recuse himself from the case almost assuring whatever the outcome (especially if the court keeps the meetings information closed to the groups) look fishy.  Bush appointed Pickering...the supposed racist Judge from MS (which he isn\'t...but nonetheless shouldn\'t be appointed any higher because he holds views that are for me to far to the right). He backdoor appointed him and now the Dems have to figure out what to do.  The Democratic ticket is all screwed up.  Dean has had some shakeups in his camp. First Trippi is gone...Trippi the main man behind the orginization of the Internet portion of the Dean Machine. Obviously in the Iowa Caucus we lost pretty badly overall, Gephardt dropped out then because he just wasn\'t getting the vote. Dean then had his scream...which anybody who has actually enhabited the what many of us refer to as \"the real world\" (not MTv Style...and condolences to the MTv family over the loss of Bennam or whatever her name is who was one of the creators and producers of \"The Real World\" series on MTv she died from a a cancer that she has suffered with for many years). You see out in the \"real world\" there are these things called emotions and even the most rational of us suffer from them.  He was venting his anger, passion, revving up the troops, getting attention, and yes howling...so what? Who cares? At least he is literate unlike Bush and hasn\'t nearly been taken out by a pretzel eating mishap.  Dean went on and lost NH...he came in second to Kerry. He showed very well and really he didn\'t loose share to Kerry instead Kerry just picked up people that came into the voting process and was able to turn out his base at the poles.  Now Dean looks like he will come in Second or Third in the upcoming Tuesday primaries.  Now he had arounr 40 some odd million...or raised that much. Now he is down to 5 million or thereabouts. Most of his staff will have to give up a few weeks pay and he has had to pull all ads from TV for the time being to conserve funds.   My only question outside of who he replaced Trippi with is how did they blow through that money so fast? That is probably the biggest shocker right there...The only thing is that he was running ads much earlier than competitors and running ads in states that aren\'t even close to primary. I understand why he did...really it is marketing 101...but unfortunately marketing 202 is where they teach you how to spend wisely and better target your ads.  He is still raising money...so that is good. Dean still has pull with his core...but his core is tiring because we have spent much time (or at least I have spent a great deal explaining the good and bad...trying to be balanced instead of \"ra-ra-ra\" about him and all) explaining what is going on and trying to put a good spin on things.  It is no secret that the media has it out for Dean. They bash at every opportunity and it is almost sickening to watch Crossfire and watch the left come out to attack him. They have replayed that damn clip to death and never give up the chance to take a cheap shot at him.  Kerry has moved into the head slot for campaigning...but if you look behind the scenes you see nothing. Kerry has less Deligates devoted to the primary than Dean (Dean has something like 113 Kerry 98 or so...Dean is still on top). His policy chest is empty...he took Deans talking points and shamelessly plagarized them and makes himself out to be anti-special interest even though he has recieved the most money out of all the people running in special interest money for senators over the past 15 years...so that is baseless...he can be bought.  Kerry looks good...he looks electable. Well he isn\'t as electable and I will try and tell you why. Look at his voting record over this past three years and pick a random republican senator. Guess what they voted almost the same down the board. So there is no difference between him and the republicans at large.  The only reason Kerry is 1 right now is simply because he is a better politician right now. He has been able to ride a wave but soon that wave will crash.  Dean has a problem there too. Roy Neel...the new co-ordinator was a lobbyist. He was Gore\'s manager as well. However Dean is an outsider and frankly it is bold enough to work...but it is so far out there that it may crash and burn.  If this all works the system may well just be changed...and we may well get our government back...I can say one thing though is that even if Dean doesn\'t win (oh please we need a moderate in the white house not another cowboy) the system will have to change to accomodate us. They may be able to placate us for a while...but eventually we will win.   This particular election cycle is pretty unique because of all the outside Dynamics coming into play this time around. We have all been forced to look at what is going on. While many people see it one way as being great...about 98% of the rest of the world looks at it and sees a mess. So I think just shear dynamics is going to change things.  Ok now this post was rambling and not very well written...and frankly I am ashamed of it...but I had about an hour of a clear head.  I am on  50mg of Benadryl every four hours. 40mg Oxyconton every 12 hours 500mg (5/250*2) Percocet every four hours. (10mg effective oxycodone) 2mg Dilaudid every five hours.  So I am pretty doped up almost all the time. I love drugs...but damn...  The dilaudid isn\'t the best out there...although it is powerfull...for me it lasts around 2 hours and then the pain is back. The percocets last longer on me...and oxycontin doesn\'t matter much...it just is longer lasting...you don\'t get dopey fast off of it....  Well it is late and I need to have them bring me some percocets for these hips and knees....they do hurt...not to mention that my chest hurts from almost constant coughing. That sucks....  Well whenever I get a clear head I will write again I am sure...  Peace....  '),(1246,1234212,'07,April,2003','2003-04-07 00:00:00','   It has been a few days...and I am starting to jonez for my laptop...eesh I wish it was finished already.  I did finally get my new toy...got it at a steal...you can\'t beat getting a top of the line (of yesterday...or rather a year old design) Pocket PC 2002 for around a 100 bucks. The thing has a 400Mhz processor, 4\" screen, both a SD and CF slot...I just ordered a 256MB memory card for the SD and an 802.11B card for the CF slot...not to mention I also have a 128MB and two 16MB cards for the CF slot...this little thing is damn fine. It also has a unique capability in that you can get a patch type cable and plug USB devices directly into it...so it is concievable to plug my cell phone in using the USB cable, whipping up a quick set of drivers for PPC2002 and then surf the net from my handheld.  I had been planning on getting one for a while...and since this was such a deal I didn\'t want to pass it by...really I couldn\'t. Hell its even a Toshiba (made by Audiovox)...  Anyways...as soon as I get my laptop back I will be online to give all the details of what has happened for the past few days.  Peace  '),(1247,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I had a slight problem with my laptop...I was entering a nice entry last night and the Keyboard died on it...just up and died!  I had lot to say in that post as well...  I will update a little better once things calm down around here...my friend Danielle got a pair of lungs last night and I am still waiting for word back as to how she is....so keep a good thought for her...  anyways that laptop is off to the shop and hopefully will be back in a few days...that just blows!  Without laptop I have no phone...and that sucks ass as well...my phone charges off the laptop...and my home charger that I brought with me was destroyed in one of the moves about...  Well I will have a better update at some point....damn laptops and their funky keyboards...also this was a floor model PC...so it was banged on pretty good....  Off for food then bed...oh boy have a got stories to tell....rants to voice....and a lot of shit I gotta get off my chest!  '),(1248,1234212,'07,April,2003','2003-04-07 00:00:00','   Sorry Jess that I didn\'t call...Today was a huge setback day...and I mean huge!  I had gotten home from running GM to Monroeville so that we could get a rain check filled on 100 blank DVD+R\'s (got all the discounts! So the hundred only cost with tax 42 bucks and some change!)...We ate and then I dropped her off here at the house...Took a few digital pics for her and went to have them delivered and to get a replacement wireless card for GM\'s computer (because the other one wasn\'t working properly)...Well I then just putzed around for a while because it was a nice day.  I ended up at my local pharmacy to pick up some benadryl and then something happened. I started coughing in the pharmacy...No biggie I walked a half a block and I was out of breath...I got outside and couldn\'t stop and I was bringing shit up so I spit it out...  Well what I spit out was not phlegm or mucus...It was pure blood. I kept coughing up blood and more blood...It wasn\'t like mucus and blood...This was all straight blood...and it would stop.  I got to my car and spent almost five minutes coughing up a lot of blood...I coughed up maybe a cup or cup and a half of blood. There is blood on my steering wheel, on my door...on some CDs...all over the friggin place. I was finally able to get a little handle on it and headed home...I live only four blocks from the pharmacy...I had to stop four times to cough up and spit out several mouthful of blood...It really scared me...What scared me most was I could feel where the burst had occurred...and I could here it gurgling as it was filling that bit of space with blood. It was the most scary thing that I have gone through so far....and it terrified me...  The end result is that I am now short of breath...A tad weak...and scared that this is gonna happen again tonight while I sleep...Because I didn\'t do anything to trigger the lungs like that and make them bleed.  Well that was my day...oh yeah and my grandmothers computer is fighting me...up, down, left, and right...it just doesn\'t want to go wireless...so I put an ethernet card in it...moved the modem and router/access point down here to the first floor and I will install the wireless card in my box...I am sure it will behave.  While I was exchanging the bad USB Wireless 802.11b adapter that I had previously tried I was able to get 802.11g gear for the same price (and they are internal cards...I got two at I think 39 bucks a piece...that is if I send the rebate in for 20 bucks on each). I also ordered up 2 Logitech Webcam NX\'s and they will be 19.99 after I send in the rebates, and the crown jewel was that I got an 802.11g Router Access point for 59 bucks (again with a rebate)...I was able to get a balls out fast wireless system for not a lot of dough...and what is nice is that 802.11g is backwards compatible with 802.11b so my Centrino setup will still be good and my 802.11b card for my Pocket PC is also still good!  Got great deals...then got slammed with a bleeding lung...damn you sweet irony!  Peace  '),(1249,1234212,'07,April,2003','2003-04-07 00:00:00','   Oh yeah...I am glad that the Mayor of San Francisco has ordered his city to allow Gay Marriage. It is very brave of him and I commend him on it.  To take the point that he made for it...Let me repeat it. At one point interracial marriages were prohibited all across the country. Now most people don\'t make a big fuss over it. So to take the step and allowing gays to marry is great and a bold step forward.  Kudos out to the mayor...  '),(1250,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I would have given a lot of updates today had I not been in all sorts of pain.  I got a prescription for Oxycontin but my doc wrote it for a non-existant dosage so the pharmisist had to get a different prescription for a proper dosage...and that took until earlier today.   So once I got the meds I was ok...I had to take one additional pill over what I should have taken just to get the levels back up. Now I am on a regular schedule of 40mg every eight hours. Not to shabby.  I am of coarse dissapointed that Howard Dean has essentially dropped out...but after the race he has run and all the crap that he alone has had to go through with the media and several of the other candidates (with the exception of Edwards, Braun, and Kucinich)...I am still a Dean supporter and will carry on with the ideas...but now I will be throwing my vote towards Edwards. At first I had a little trouble with him...I got a gut feeling and I still have it about him...but I cannot support a candidate (Kerry) who has blatently stolen others platforms...Kerry stole Dean\'s and now he is trying to steal Edwards.  I call for all Democrats to realize that Kerry is a political oportunist and will jump on any bandwagon whenever it is politically expediant or favorable. Kerry has yet to come out with an original idea. You will see as this grinds on the Kerry is indeed very much like Bush.  However if he does get the nomination I will vote for him. I just hate being put into the position of having to choose the less of two evils.  Well that is all...hopefully I will be able to get those updates for you...and I am sure much will be devoted to politics.  Peace  '),(1251,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I finally got my laptop back today...and I got out of the hospital. I at home now writing this on the laptop in my living room...gotta love wireless internet.  Well I have a lot to update people about what has happened over the past two weeks or so since I have been offline.  I will do that in time...so just bear with me.  That is all for now. I am doing a treatment right now and have several to go...then I go upstairs to set up my main box and get it up and running again.  So much to do...so little energy.  Peace  '),(1252,1234212,'07,April,2003','2003-04-07 00:00:00','   Well the good news is that I get my laptop back on Wednesday....yay! It turns out I was right about what was wrong...and you doubt me...come on people.  The got the new keyboard in today and got it installed and all should be well. They actually should have sent it out today and if they did and got it out today there is the remote possibility that it will be here tomorrow but it will definetly be in by Wednesday. Also on Wednesday I have the parts that I need for my pocket PC to go wireless that day as well....so it will just be a big ole love fest of hardware and goodies. What is nice is that I got a SD memory card that I can use in both my Pocket PC and my Laptop (it has a SD slot built in)...it is just neat all this technology.  Well a lot has happened over the time that I haven\'t had real unfettered access to the internet and as soon as I get the laptop back I will most definitely making a rather large update as there is a lot to tell.   There are good things and bad things...but at least there are things.  We have a new CF patient up here on the floor she has come over from Childrens she is a little frightened. I may go visit her tomorrow and see how she is doing seeing that I am the old man on the floor now. There are CF patients left that are sicker than I am...but I am fast rising to the top of the list of the sick ones. I am also now officially the longest inpatient right now here at Shadyside seeing that Danielle is no longer over here....she is now over at Presby...she got her transplant finally...I will tell all about that when I have more time to update properly.  I also have a lot to say about the current political climate and what seems to be almost a given now...the Kerry nomination. I have a lot to say about that....and boy oh boy it will be good to get it off my chest.  Well there isn\'t much else I want to write about right now...so I will leave it at that...  Remember hopefully by Wednesday you will see a much larger update and a lot of information. I will probably do multiple posts that day as I get one thing tackled after the other.  I also have a lot to do online that day as well...like pay credit card bills and such...plus download to outrageous stuff for my Pocket PC.  Peace...and until then have a happy time......  '),(1253,1234212,'07,April,2003','2003-04-07 00:00:00','   Oh yeah and I just took 20mg of Oxycontin...so that should help out soon...  All these pain killers? Yes...all those pain killers...being couped up and sedentary makes things hurt that you never thought you had...plus it helps get rid of the pain of coughing (when you are coughing day in and day out things get sore real fast).  So I wish I didn\'t need all of it...but at the same time I am thankful that I have it.  Peace  '),(1254,1234212,'07,April,2003','2003-04-07 00:00:00','   Well earlier I went downstairs to get some food from the cafeteria and I thought that I wouldn\'t need any oxygen...well was I wrong or what...by the time I got back up here to my room I my fingers were turning blue and I was definitely feeling like I needed oxygen.  I guess that will be my reality for a while. I am going to have to have O2 on me...that sucks ass.  My energy reserves are so tapped you have no clue out there.  I got to talk to Danielle today and it was so good. She sounded great. Her mouth was going a mile a minute and she never got winded...she is having some minor problems but nothing that is extremely serious. It was just very good to hear from her...plus she has offered to answer the myriad of questions that I have about transplant. Sure I can read the books and I can watch videos, hell even the docs can answer some things...but it isn\'t first hand experience and the questions that I have can only be answered by someone who has been through it. So I am very greaful that she has offered to answer questions.  Well it is extremely early in the morning and I just had a dose of Benadryl and Dilaudid...so I am fading fast folks...  Peace  '),(1255,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I know that I should be updating this blog more often but to be honest I just haven\'t had the energy. I spent most of yesterday in a hazy half sleep state and just not feeling well.   Today so far I have seen two docs...both Indian...now mind you I don\'t mind foreigners...but I mind when they are trying to tell me something and they have a bare grasp of our language. Plus I don\'t like dealing with doctors outside of my own because chances are they don\'t know how to treat CF patients and end up screwing things up.  Well I am already worn out and I think I will take another nap.  Peace  '),(1256,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I am back in the hospital again. This time it is because I have had two episodes of coughing up plain good ole blood. There wasn\'t any type of mucus or gunk...just straight dark dark red (probably from an artery...oh joy) and my oxygen level on room air are down to between 80-85% when I normally run around 95-96% and that is what I was at when I left the hospital last week.  When they first got the pulse ox monitor on me I was all the way down to the high 60s range...so to same I\'m sick would ve an understatement.  We will see how this all plaus our...so updates will probably be updating a little more frequent. So hang in there folks...he blog will slowly come back to life.  Peace  '),(1257,1234212,'07,April,2003','2003-04-07 00:00:00','   baaaa!! Thats what I have to say to it all.  '),(1258,1234212,'07,April,2003','2003-04-07 00:00:00','   so I haven\'t been to sleep yet but I have something to say about our favorite serial lier...Ann Coulter (that neo-nazi, neo-con whore wannabe).  You have heard me talk about Max Cleland before. He was a Senator from Georgia and he lost his campaign to a Republican partly because of a massive smear campaign that the Republicans are well known for. They ran an add with him and then pictures of Osama and Hussein. They called him \"un-patriotic\" for voting against several of Bush\'s bills that are now proving to be quite troublesome...mainly Iraq, the war on terrorism...  Well Mr. Cleland lost both legs and an arm in Veitnam...now the story goes is that he was being sent out into the field to fix some radio equipment and when he got off his chopper he noticed a grenade on the ground. He thought that it had fallen off of his kit (as had happened many times before...they had a problem falling off of kit) so he reached down to pick it up and it exploded. He lost both legs and an arm. This all happened at Kei San...the worst battle that we fought in Veitnam...many young men died in the battle and Max recieved several high commidations for his work...the incident happened as the battle was winding down and he was flown to a hilltop to establish a radio tower so that the guys in the rear could better communicate with forward positions (although the guys in the rear were being shelled and had a great battle themselves). Needless to say he was in a combat zone....or hot zone.  He came back and became rather succesful despite his obvious disability.  Our beloved whore...wait I mean Ann...well she makes the case that this all happened not in a battle or combat zone but in a relatively safe place and while he was drinking with his buds. She basically said the whole \"war hero\" thing was a hoax and that he had lost his limbs out of drunken stupidity.  Well the facts tell quite a different story...the military even responded and his commanding officers responded as well.  Here is a link to a better telling of the  story.   I find it absolutely dispicable what this woman has done and I would love to challenge her to a debate and go on TV with her to just humiliate her. I would have no problems whatsoever calling her a lieing whore who doesn\'t have the foggiest of idea of what is really going on in the world.  It makes me mad that she would question his patriotism...and it all came about because of the recent Democratic push for Bush\'s national guard records. Yeah those...where the major questions still lay unanswered.  It is about time that liberals and centrists stand up and collectively tell her she is no longer welcome here. She needs to go live on an island somewhere along with all the other conservative hate mongers out there...let them destroy themselves....because I am tired of them raping our nation.  To Ann...you need to get a clue, you need to stop lieing, you need to leave...you are no longer welcome. Anytime you question someones patriotism by attacking them for an accident where a man or woman gave their all in service to the American population you have then become what you hate most. It is time for you to exit stage left and for us to never hear from you again.  '),(1259,1234212,'07,April,2003','2003-04-07 00:00:00','   Well today I got some new antenna\'s for the wireless setup and now all is peachy. It looks like it will be ok now.  My grandpa is in the hospital...he has been having problems with his cental lines that they do his dialysis through. Over the past few days his one decomissioned line started to leak...well today he almost bled out from it. At some point the surgeons that he has seen put a patch on the artery in his arm and that patch blew today and it was spurting blood all over the place...part of the problem is that he has been picking at the wound for months and it has been infected for a long time.  He went into surgery tonight and got that fixed and they had to take veins from his foot to save the arm. When they got the vein out they noticed that it was completely lined with cholesterol and plaque deposits...now normally you only see that close to the heart and in the groin...but this was all the way down at his hand...so he is really in bad shape.  For me today it wasn\'t so bad...I got out a little bit and it was non-stressful.  Right now I am about to play with my DVD burner and try and copy a movie...one that I already own.  Now that the network is all fixed I will definitely be updating more often and I will get some of the hospital backlog up.  Peace  '),(1260,1234212,'07,April,2003','2003-04-07 00:00:00','   Yet another crap day...but at least I got my grandmothers computer up and purring like a kitten.  The wireless setup is really flaky and I don\'t know why. I have tried so many different things. Every once in a while the AP just craps out and all the cards go looking for a network. The only thing that is good at switching like that is the damn pocket pc setup I have...and it works great!  It won\'t matter much longer because I have a Netgear 802.11g router/ap on its way and that should clear uo the problem right there.  Tomorrow I have to go to Radio Shack to pick up some raw cat five cable to make a longer one so that I can keep GM\'s computer hard wired and have the AP at a central location in the house. While I am there I am going to look for better antenna equipment for both the AP and the Reciever cards...they both have dinky 5 cent jobs...that will help out on the signal issues.   Right now my PDA is showing  an almost flawless conection to the AP...while the card that is in my box is showing only about a quarter of the PDA\'s signal. Not only to mention I have been getting a great \"link quality\"...so I think I have a bit to do before the system is really fixed up.  Well kids its time for me to try and sleep...ttyl.  BTW: I will get pics of my new laptop and desktop and PDA when I get all settled back in...because I am sure a few of you are wondering what kind of stuff this techno weeny is using...so stay tuned.  '),(1261,1234212,'07,April,2003','2003-04-07 00:00:00','   Well Bush has unleashed his first round of re-election ads and so far they are going over worse than a fart in church...actually I would imagine Satan could walk into a Catholic Church and molest some alter boys and people would be less offended at that than Bush\'s ads.  I have seen them and frankly I am offended not because of the images but because Bush said that he and Cheney would not use 9/11 as a political chess piece. Of course no one believed him at the time but we all hoped that they wouldn\'t.   Bush is having to use 9/11 because really he has nothing else to run on. The economy is shit, the so called security that was supposed to be provided by the war on terror has turned out to have made us less secure, and to top it all of the Bush administration has nearly demolished our Civil Rights through various laws.  We haven\'t even gotten to his Tax cuts for the middle class (I didn\'t know the middle class started at several hundred thousand a year and up).   This administration is the worst in the history of the US. They have almost single handedly destroyed what America is.  Today Martha Stewart was found guilty on all charges. While I agree that she was guilty as hell, I have to say that she is the big white elephant in the system to keep our attention off of Ken Lay and all those people in Enron and even WorldCom (although Bernie Ebers has been charged), Global Crossing...she was a diversion. It disgusts me that a big deal was made out of her crime which in the grand scheme was minor. She profitted all of 60 grand and since the government has decided to go after her the government has actually cost investors millions through prosecution of her...causing greater damage overall than if they had just fined her and suspended something or other.  Today was also the day that John Ashcroft was admitted for severe pain to a hospital. He has something called Pancreatic Gall Stones or something. Painful and not pleasant for sure. I would never wish pain or illness on anyone (unlike Pat Roberts who called for his flock of sheep to pray that some of the Supreme Court Justices to get sick and die)...however maybe after he gets well he will see that his policies and the things he has done have been bad for the Union and maybe he will reverse his stance on some things...however I doubt he will change his tune. I wish him well and a speedy recovery so that the nation can bash his brains in during this election season and send him back to podunk USA where he lost a senators seat to a dead man.  That is all for now....Peace  '),(1262,1234212,'07,April,2003','2003-04-07 00:00:00','   Sorry for the lack of updates lately...I just don\'t feel like posting all that much. I am not well and really I have spent almost this entire stay in the hospital asleep.  I went to meet the surgeon today and it was good that I got that out of the way. It was about time. I don\'t know how I feel about it all really. I don\'t want to be in pain and really I don\'t want to feel like shit. I am basically scared about getting sick and getting to the point where I have to carry oxygen with me and all that crap...as I have to now.  I had to take O2 with me because I am not getting enough from my lungs out of the air...I really don\'t want to go through this all. I am afraid and I am worried. I can\'t really talk to my family about it because I don\'t know how. I don\'t want to talk to a therapist because I have done that before and it never helped. I don\'t want medication to feel mentally better because it comes with a lot of side-effects. I just don\'t know. As for friends...yeah like I have many that I can talk to about this.   Sure I will probably get some email here and they will all be like \"dude you can talk to me about anything\"...well it isn\'t that easy and no I can\'t. There is no way you can understand what I am feeling and what I am going through. Really you can\'t help and I am not mad about it or do I want to not be friends...it just isn\'t something that I can talk with them about.  I just don\'t want to be sick and I certainly don\'t want to go through all of this.  I guess it is that time that I just have to buckle down and deal...and I hope I can.  Another thing that is bothering me is that I have not had sex in over three years now. So? You say....well it isn\'t so much the sex or the intimate contact that I miss...it is that piece of the puzzle that you feel needed or wanted if only for a few minutes or hours...that is what I miss. I miss just being loved by someone other than my family and I don\'t mean \"cutesy I love you, I love you\", I mean it for real. To feel that connection again would be great. That is probably the biggest fear that I have...I will go through all of this alone and if I make it out the other side what will be there waiting...at this point nothing. On the other side of all of this is nothing...not a family, not a girlfriend or lover...just me.   I wish that I could go out and play the field now...but even if I could I am horrible at it...sure I can be funny and witty...but I need an audience for that to work...one on one sometimes it doesn\'t relate.  I would just like to know what is in store so I could make a more informed decision for myself instead of walking blindly into the great unkown. If I knew what was up...then I could make a decision and be content with it. Since I don\'t know or have a clue, making a decision is hard and I am not happy with it because I don\'t know.  Peace  '),(1263,1234212,'07,April,2003','2003-04-07 00:00:00','   I just saw a commercial for Viagra that uses Queen\'s \"We Are The Champions\"...now we all know that Freddy Mercury loved a good stiff cock from time to time...but I would think he would have a problem with his song being used in that commercial.  Have they no shame.  heheheheh...stiff cock.  '),(1264,1234212,'07,April,2003','2003-04-07 00:00:00','   As we get closer to finalizing Kerry as the Democratic nominee for President, the more I feel like the American Public has lost control of our own government and that we are no longer needed except to rubber stamp things.  I was thinking last night and I came to some interesting ideas and conclusions.  Under this administration we have seen the budget for education slip...While it has grown it has not grown at the pace that is needed to just keep at baseline. With the \"No Child Left Behind\" act we are seeing that in essence the Federal Government wants to get out of the Education Business because after all if your school or district is failing then you first get put on a probation of sorts then if (which reduces Federal Funding...When it should increase to fix the problems) your district or school continues to not meet the standard then you will loose all Federal Funding.  That is a classic case of breaking an already broken system. If you cut funding to a struggling school then obviously it won\'t have the resources to better itself and will continue to fail.  Now what happens...Well your child is left without many alternatives...While I won\'t say that public schools will go the way of the Dodo but I will say that the public schools will produce even less educated people that it already does.  It won\'t be a problem if you can afford to send your kids to private schools...and if you think that \"well they will make vouchers available\"...Don\'t count on an educating voucher to cover the entire cost of a private school...You the parent will have to make up the difference. So basically the lower classes here in America will be denied a decent education.  Also if by some miracle little Sue or Johnny make it through primary education college will be a much more difficult beast to conquer. We have seen at State Schools a rise in Tuition average about 25% over the last three years (at least here in Western PA college tuition has gone up by about 25% over the last 3 years) and it is becoming more and more difficult for even an average family to send their kids to school. Don\'t even get me started on those young adults who are paying their own way....They now have to go into even more debt than before.   The reason that College Tuition is going up...Well lack of funding from the Feds...But mostly because states are struggling to pay their bills. The economy is so rough right now that states simply have to cut budgets and pass the cost on to students and their parents.  Yes the economy is improving according to the numbers...But those numbers don\'t tell you that the people benefiting from the slowly growing economy are the wealthy people. I saw recently that the average extremely wealthy person pays about 17% of their income to the government in taxes...While the middle class and the lower classes pay on average of 24% of their income to the government.  The people that are the engine of the economy and the main payers of taxes haven\'t yet seen real growth or pay increases.  Now if you take all of this into account and then look at what happens because your lower classes of people are less educated or completely un-educated you start to see some really bad things come to pass.  We all know that as long as we have a Federal Government we will always have the Military and Security forces and they will always receive the lions share of the federal budget.  What we don\'t see is that for many of the positions in both areas you need a college degree or high level of education. Take an Air Force Pilot...They won\'t talk to you if you don\'t have a college degree...  Now look at who typically goes into the military (even as pilots because you would think higher classes would sign up for that...But I argue differently)...The people who drive the military are in general lower class citizens...You won\'t find many rich kids in the military.  So what happens when these lower class people no longer have access to decent education...Well our military is going to suffer...Perhaps that is why the Air Force and other branches of the military are pouring money into  UAV\'s and such...If your plane doesn\'t need a pilot only a minder...Then hey who gives a rats ass if the recruits are well educated. The same goes for other technical aspects in the military...They are slowly becoming automated and need only for human minders to just keep an eye on things.  What is a real mind blower for you...Well it blew my mind. The people who are in the military...mostly from lower classes, they are also the ones who are paying for wars.   Enlisted men and women are essentially paying for the privilege to get you ass shot up in some foreign land...Isn\'t that a bite.  I would also argue that we no longer have a middle class. We have the upper classes which includes several layers...Those who never have to worry about money again, and then those who could loose it all in a blink of an eye but generally don\'t because they are \"well leveraged\".  The you have the lower classes which can then be broken down into several layers. There are the poor, those who are under or near the poverty line. You have the working poor, those who have a job or two but can\'t really make ends meet but manage month to month. Then you have what would be the formerly middle classes down here.   The middle class of the poor are those that have decent jobs, maybe college educated, and generally living beyond their means but they make it.  However any bump in the road may derail these people.  The upper middle class are those that are so close to being upper class they can nearly taste it...But they will most likely never make it. These people are college educated, have great jobs, and live lavishly, but again they don\'t have the money to just go on vacation out of the blue or spend a buttload of money on a purse shopping spree. These people are well off but as time moves on higher prices and taxes will slowly push these people to the brink and they will then have even less disposable income than they did before.  The key to rating classes for me is Disposable Income. This is the excess income after you have paid your bills and put money away for college tuition and retirement. The money that you can afford to blow on frivolous activities. The richer you are the more money you have to blow. To me that is pretty simple, and the most reliable way to figure out who is rich and who isn\'t.  So this whole post is about how the majority of the public is being screwed and on top of it all we are paying to screw ourselves.   I hope that something gives soon because I am damn tired of looking like a war monger in foreign lands, I am tired of being feared to the point where people will go out of their ways to kill us in the hopes keeping us away. I am tired of the overall good people here in America paying for others to suffer and put us in danger.  Peace  '),(1265,1234212,'07,April,2003','2003-04-07 00:00:00','   Lord Of The Rings: Return of the King swept the Oscars tonight. Hoorah!  Other than that...nothing to report.  Peace  '),(1266,1234212,'07,April,2003','2003-04-07 00:00:00','   I feel like shit. It is early in the morning and I just had all my drugs...but before I got them I puked my guts out. So hard that my nose bled a little. What really sucks is that I haven\'t eaten all that much in the past few days because everything has just tasted off to me. Since I came back to the hospital I haven\'t really had one good full meal. Like I say...nothing tastes right.  I muscled down some food tonight but a little later it decided that it was going to come back up for a visit.  I don\'t enjoy being sick like this. All of my energy is gone and while nothing is majorly off...something is still not ticking over correctly so you know I just have to feel like shit.  All I know is that when I went to get food the other night from the cafeteria I didn\'t take an O2 tank with me and by the time I staggered back up here I was almost at the point of passing out.  I just hope that I can get to a point where things are all semi-stable...because I don\'t like this feeling at all...even doped up it isn\'t pleasant...while I love all the drugs they are useless in the quest to make me feel level...only good health can do that.  '),(1267,1234212,'07,April,2003','2003-04-07 00:00:00','   Oh yeah and I think I will be doing a little work on the blog here to spruce it up and rid it of some links. I am also going to seriously look into getting my own domain name and private host.  I feel once I get the majority of this illness out of the way I will be freed to persue what I want which is simple really.  I want a nice trendy (but not yuppie) loft apartment (preferably I would like to buy a building with a loft and either apartment units or a downstairs that can be a storefront or business). I will have it nicely furnished with lots of electronics and computers everywhere controlling all sorts of things. I would also set up a small hosting service of my own so that I can subsidize having my own fat internet pipe into the place...and on that I can do what I kinda have been floating in my head.  I want one of those floor beds...you know IKEA bed where it is low to the floor in a big king or queen size and a big down blanket. I want the apartment to be an open floor plan so that you can see everything in the place with the exception of two bedrooms and the bathroom (which will be the bomb) and the rest is just living space with a nice kitchen and living area...and the entertainment center will be a center all right. All around a Plasma or LCD TV screen (depending on if LCDs come down as they have been...oh and has to be 40+ inches)...THX certified surround, and a media computer that controls it all with HDTV, gaming, and lots of goodies that will be tied into my apartments network (both wireless and wired). There is the living and kitchen on the one side and the other side will be a home office with another kick ass computer with dual LCDs and uber specs for work...and a nice desk...I have that kinda picked out already because the one I want is glass topped and just look beautiful.  The bedroom will have a media center as well and a medium size LCD or Plasma screen and computer that is powerful enough...and for sleeping I could play mp3 with a kewl visualization on the Plasma or LCD...plus the sound system in there will be small but powerfull and clear. The bed and open closet design (and little storage things so it looks like I am anal).  The second bedroom will be a little more basic...a bed, dresser, and few other things...but it will be comfortable.  The kitchen will have a big ass fridge with a glass door so that I can see what is in it at anytime...kinda like a deli cooler. An island with the stove and oven and microwave...hanging above it is the pot rack (stainless steel all around)...just a tight looking kitchen.  The floors would be wooden...with possibly radiant heat built in (saw it on this old house and it looked neat as hell and keeps things warm and is efficient)...there will be area and throw rugs...but the floor has to be wooden...I have always loved a good wooden floor.  The place will be for both living and entertaining...not huge in space...but big enough that I can have people over and we won\'t be stepping over each other during a party (hence the big ass entertainment center that is  THX plus it will have speakers wired around the house...and there will be a place for wanna be DJs to jack in and spin if they please)...  I want these things because I could live there easily and comfortably for a long time. Plus if I get my ass going I could run a business or have rental units that will pay for the building itself. That would be the key to it all. The building must be able to sustain itself financially either through renting apartments or business space or by me running a business out of it.  The goal is set so that I have something to aim for. Once I have something along those lines and my life is moving towards the achievment of that goal...then I will be ready for a relationship with some hot super model...hehehe  And if not...all those LCDs and computers can put on porn throughout the house....  Oh yeah and I want the computers to be tied together and voice activated and for them to have a voice...like wakeup alarms...and I can say \"set temp 70 degrees\" and the system sets the thermostat, or I can say \"tv history channel\" and the tv I am addressing will turn onto the history channel.  Kinda like HAL without the homicidal tendencies.  '),(1268,1234212,'07,April,2003','2003-04-07 00:00:00','   I wanted to give a quick update to the story that I posted yesterday early in the AM. Well the death toll in Madrid is now up to 190 (and expected to rise) and over 1400 injured.  It was three seperate attacks on three trains in three stations using a total of 10 bombs. At least that is what is known right now.  At first the Spaniards insisted that it was ETA, a freedom fighter group for a small bit of land in the Northern part of Spain right along the border with France.  Of coarse we chimed in and said it was Al-Queda. Some convieniant evidence shows up and it looks like it.  Nobody really knows who did it yet and we all need not jump to conclusions until we know more.  It was a trajedy and generally the people of the US send our best and we hope that you will not give in and your government will not turn counterproductive and revoke many of your rights unlike our own.  Well I am still in the hospital. I talked to the doc today (well technically yesterday) and he is going to try and get me out of the hospital. We are both a little leary because of the infection and we have not been able to isolate it. If he springs me I am afraid that I will bounce back into the hospital in a short time.  Whatever the decision is I will be happy and make the most of it because that is all I can do...and believe me I am living one day at a time.  Peace  '),(1269,1234212,'07,April,2003','2003-04-07 00:00:00','   I have been thinking more and more about this whole morality thing that the nation is going through. How so many people were shocked by Janet\'s Nipple, and how the sex on the tele has just gotten out of control (the only time I would consider it out of control is if you go tumbling off of it while in the act...that is wild sex)...but regardless the entire nation...or rather many people in the nation have gone completely nuts and started making mountains out of molehills.  What it comes down to is that the Far-Right fundamentalists of the christian coalition and other radical groups have decided that the nation is filled with heathens and we all need saving.  For a person like me it scare me to think that these wacko\'s are now censoring what I can see on TV...not just the old standards and practices kind of censorship but rather the fuvor of censoring even minor things.   The first strike has already been done. We can\'t really do much about it right now...it can be fixed but it will take time. That first strike was the consolidation of the media giants. There are now but 7 major media giants and they literally control what we see on Tv and the Theatre...and even what we hear on the radio (with Clear-Channel or Infinity being the two companies who own the majority of radio stations).  With this consolidation it becomes very easy to control the message that is being spread. Simple and easy. If you own all the media you can tell the people whatever you want and the public have no other place to get news. The interesting thing is that most people will take what they are being told as the truth because it is coming from repsected sources. Hey its NBC...they tell the truth most of the time.  You see...easy.  Then you start with your second wave. This is the morality police. They come swooping in and start the argument that children are seeing to much on the TV and that things are out of control and it is about time for some restraint to be shown by the media giants...ok no problem. That happens.  Then once you have started getting little things done here and there...you start to go for larger prey and the outright dropping or pushing out of shows or people who do not share your views.  No this predicates on the idea that the media giants have given buttloads of money to one party and that party has happened to deliver them the ability to become even larger...so they are allowed to grow and they only have to take orders every once in a while...and most of those orders are rather benign compared to boycots that would start if the radical right isn\'t listened too.  So they target people and shows and get them fined and eventually thrown off the air.  This could never happen right?  It is and has already happened. Howard Stern is the most famous example as he has been pulled from several Clear Channel radio stations (including my own Pittsburgh station...and I enjoyed catching a show now and then) and will soon loose his job because the FCC will be fining him and his station soon...and rather than stick up for him they will fire him (he works for an Infinity station).  They claim they are pulling his shows due to his crude humor and him stepping over the decency line. Well he has been doing this for years and years. He has not recieved many repremands in the past from the FCC. If anything his show has gotten a little more tame over the years. So the decency thing doesn\'t really wash.  What changed is that he started talking out about Bush. Stern was a former Bush supporter as he is a liberaltarian and they vote mainly republican and he voted for Bush. Well he has changed his mind and decided to talk out about how things have gone whacky and that we are loosing our country of moderate people to a group of right wing nutters that carry crosses on their mini-vans and SUVs. Stern finds that frightening (as I do) and now he is targeted for removal. To be sure he will be removed.  They also went after Bubba the Love Pump down in Florida...for similar reasons.  Now compare it to some of the right wing morons like Hannity, Coulter, Rush, and many others and if you listen to right wing radio talk shows you will here more foul language and disgusting things out of them...and they are doing it to be funny or point out fact...they do it with a mean spirit and with the intention of riling people up.  Those guys say horrible things...but because they pull the party line they are fined or dropped from the air...so it has nothing to do with language or decency...it is about politics and silencing political dissodent and dissodents.  May we see this trick and fend it off for our own sake...because before to long we are going to have to clear our thoughts even through a government agency lest we want to run afoul of someone or something.  peace....  '),(1270,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I forgot to mention that I had to get a blood transfusion on Tuesday. 2 Units total. It seems that I am leaking from somewhere because during my last visit I was also having similar problems, mainly that my blood count was low, iron levels low, and a high white cell count. Basically I was sick, had a few vitamin and mineral problems with the levels being low, and I was low on blood most of the time.  This time it was enough to have to get a transfusion. I felt better later in the evening and yesterday all day I felt ok...so far today I feel so-so but not great. Hopefully after I get a little sleep I may feel a little better...which is important because I have to go start my car and run it up and down the garage to keep it from having problems from sitting to long.  Also this morning there were a series of bombs that went off in Madrid, Spain. Many people are dead (62 and rising as of 5:15AM Eastern time). It is sad because right now because the world have basically become unstable. Around the world you here of at least one bombing per day and many deaths.  All of the bombings have something in common that should be noted. All of the groups that are setting of the bombs are groups that are seeking greater rights or the ability to set up a nation of their own. All of the bombings have targeted the popular or predominant parties of nations. Basically most of them are Freedom Fighters.  Let me stress that I do not think bombing inocent citizens is a valid tool for fighting a power that is opposed to you. The only place that it has worked has been in Ireland...and frankly all of the bombings that the IRA committed didn\'t do much except to get moderates to the table. In the end the IRA had to stop bombing places and killing civilians and then they were able to make progress towards independance and peace. Today still there are tensions over there and the peace can be called shakey at best.  So it bombing civilians doesn\'t really get you far...but it does strike fear into the public at large. Sometimes drastic measures are needed to get negotiations going, and for me I can understand suicide bombers and bombings in general. When a people or group have to get to that point of dispair I can understand why things that are normally off limits become fair game.  All of the conflicts that are happening now would not be happening if we still had two superpowers pointing missiles at each other (we still do but they are no longer \"those commie bastards\"). During the cold war all of the little tensions around the world were essentially supressed because people were always at threat of nuclear war and local spats were simply not worth the energy.   Now that the cold war is over all of these little groups that have come out may well have the right to set up their own nations and govern themselves, but as always through history, the bigger governments won\'t let the little guys go (partly because the big guys that won\'t let them breakaway usually are not letting them break off because the little guys in that land have natural resources that the big guy needs to exploit) because of arrogance and ignorance.  So we all need to buckle up because it is only going to get worse until the governments of the world realize that they cannot run roughshod over their people, and that some groups it is just better to let them have their land or new nation then fighting it out with them.  Anyways...time for a nap...peace ya\'ll  '),(1271,1234212,'07,April,2003','2003-04-07 00:00:00','   There is one Bush ad that is playing that is really pretty damn funny if you consider the Presidents past business dealings.  The ad talks about Americans having a strong entrepreneurial spirit and how they need a President who understands that.  Now the funny thing is that over 50% of small businesses that are opened each year will likely be closed within five years.  Another funny thing is Bush saying he understands business. Um yeah...that would explain his many failed companies and the millions that people threw into them just disappearing. Oh yeah and much of that money came from Saudi Royalty and Bin-Laden family.  It is just funny.  '),(1272,1234212,'07,April,2003','2003-04-07 00:00:00','   Double post...oops.  '),(1273,1234212,'07,April,2003','2003-04-07 00:00:00','   Gitmo.  Our interment camp. Where there is no law, where there are no rights, where you go if you are A: possible terrorists B: Possibly political dissidents C: Or just plane baddies.  Ok so the last two we don\'t know about...and it is probably just conjecture, but how do we know for sure? We don\'t. As an American who has read the Constitution and The Bill of Rights I find it absolutely criminal, and I find it even worse that normal everyday American citizens have not questioned it at all. These people in charge are breaking not only our own laws to due process and speedy trials, but they are breaking International Law as well as the Geneva Convention which all nations essentially play by (well at least the \"Western\" Nations...most \"Eastern\" Nations don\'t play it by the letter..they still do some pretty nasty stuff to POWs...however I am sure that we do similar things).  We need to stop it plain and simple. We need to release all of the \"detainees\" and take them home and find some way of paying restitution to them..whether it be millions of dollars or help their country with giving aid, whatever they want we need to give it to them.  I understand the fears of the Pentagon and White House, they think that if they release them that these men and children will re-join Al-Queda or the Talliban, if they were part of those organizations in the first place. The problem is we don\'t know for sure who these people are and what their true affiliated were. It is very possible that we picked up 50% of these \"detainees\" and they have no affiliation with either group, they were just in the wrong place at the wrong time with a long beard. Some of them are indeed part of those organizations and may well be a threat if they are let go.  Now here is the real problem. We have held these people for two years or more without rights and in isolation. The problem is they are probably pretty pissed off at America and Americans. We may very well release them and they will join up with the first group that comes along and says \"lets go destroy the great satan\". That is our real problem.   We have bred hatred in them. Before we came along and picked these poor people up they may have disliked America but many of them (at least the ones who weren\'t Al-Queda and just simple Afghan peasants) remember that we helped them rid their country of the Russians in the 80\'s. Again they had no love for us, but they certainly weren\'t going to out of their way to kill us. Now those people (with a minor minority who actually practice the love of Islam and decide to forgive us) are definitely candidates that may well go way out of the way to exact some vengeance on us.  We can not avoid it sadly. There is no way that we can make it better. We stepped in the pile of shit and we are going to smell like it for a long time. We certainly can not keep these people until they die natural, over the years people will certainly get a bug up their ass about the whole thing and find some way to free them. We can not just execute them, that would lead to true International Outrage and may well trigger foreign countries to execute prisoners that are US citizens for some sort of revenge.   The only option is to release them, try and make it right, and just watch our asses because they are going to be there ready to do something should we slip up for even a minute.  Many people forget however the even larger picture to this all. Sure it is bad that we are holding these people without cause, without charge, without any legal rights, but it is what could set a precedent for the rest of the world.  If we can do that to foreign nationals, then any other nation can do the same to us under the reasoning of the Bush administration (or rather rule of tyranny). Imagine you are over in lets say Egypt...not the best place for Americans, but a hell of a lot better than say Iraq or Iran. Say the Egyptians get a bug up their ass and they come and pick you up at one of the pyramids. Now mind you they don\'t call the embassy and tell them \"hey we have John Q. Public and he is being charged with goat molestation\"...no they just snatch you up and lock you away. You cannot get letters out or make a phone call. Not only that but they hold you indefinitely and in addition the people holding you are not familiar with the English language (hey do you think the average troop has even a passing bit of Arabic language...sure there are some that can speak it well for interrogation, but the average \"grunt\" knows only a few commands and not much else).   Hopefully in the above example you can see that it is a real problem. If you think that it won\'t happen because you are an \"American\" and if you were held special forces would swoop in and rescue you. That only happens in the movies, very rarely does it happen in real life. Besides the Government doesn\'t even know you have been detained. If we really did those kind of things then we would be making many missions into South American nations to rescue Americans...but they don\'t, unless you are politically connected or very wealthy.  Being an American at this point in history is not a particularly good thing in the rest of the world. We have no special rights in the rest of the world, and we are not special.  Hopefully we can unseat our current \"leader\" (ha...leader...he\'s leading us alright...leading us to destruction) and get someone into the White House that will be willing to approach this all very humbly and ready with a lot of \"I\'m Sorry\" cards from American Greetings. Unfortunately the person running against Shrub is Berry (I mean Kerry) and they are not to much different...both went to Yale and were members of the Skull and Bones club. Kerry went to Vietnam and won some medals so we know he isn\'t a complete chickenshit where Bush went AWOL from a reserves position and the most danger faced was stubbing his toe on a bench or jamming a finger when flying an obsolete jet.  We can hope that things will change just enough so that Americans can at least start to hold our heads level...but it will take many years of groveling before we can hold them high.  Peace...and to foreign readers. I assure you that not all Americans a trolls or evil people. Many of us are disgusted at what is going on. Many of us are mobilizing to save our nation, and hopefully to save some lives around the world. We aren\'t all beer guzzling, wrestling watching, inbred, and married to our cousins uncles sister in law who is the mother of Billy\'s kid your best friend. We aren\'t all like that. Many of us want and will change our nation for the better...and you can help us by supporting the non-cretins amongst us and recognize that some of us are trying...it is just hard because we have to live it day in and day out...and sometimes we tire.  Peace...  '),(1274,1234212,'07,April,2003','2003-04-07 00:00:00','   For an interesting video go check this out. It\'s about Bush and how he got where he is.   Good little video   It may take a few minutes for you to load and it is 56 minutes long...but certainly worth the time.  '),(1275,1234212,'07,April,2003','2003-04-07 00:00:00','   Well the United States finally claim the number one spot in the world for at least one thing.    We have the highest level of STD\'s in all of the industrialized world. 1/3 of all Americans will get some sort of STD at some point during their lives. Of the 19 million or so new infections of various STD\'s almost half (49%) of those people are in the age bracket of 15-24.  You gotta love this country...maybe that will be a deterant to people wanting to come here. Maybe we can have a whole new slogan for the nation. \"Come to America, get a free STD\" or \"STD\'s...an all-American tradition\"...or maybe we should give some tips to people \"Avoid the clap\" (especially effective on resteraunt bills where is says \"Tip\"...waitresses love that...especially the ones that look like they have been ridden hard and put up wet...and those ladies are often found at truck stop food dives).  So we need to be proud of something.  Peace...  '),(1276,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok...we now know that Republicans want to rule the world and they will attempt it in any way possible..and that the media consolidation of late is a prime example of how they will do it. Through stifling any dissent against them or Premiere Bush...or Dictator Bush...whatever the hell he is because no matter what the supreme court says...he didn\'t actually win the election just like he never really finished his reserve duty.  The RNC has warned CBS stations not to air the Moveon.org commercials saying that they are illegally funded and break the rules of 527\'s (a new class of PACs). A 527 is allowed to collect as much money as they want from donors and they can publish ads that knock candidates but they cannot explicitly endorse any candidate.   Well if you have been watching TV lately you have probably seen a few moveon.org ads. They do a pretty effective and truthful job of attacking Shrub. Notice in the commercials they never once endorse a candidate, they just attack Bush. Well they are allowed to do that.  The RNC states that they can\'t do that. Well the election commission sees it otherwise and says that they can as they are following the rules.  So if you stand for stifling free speech, are for saying anyone who opposes your point of view, or anyone who may remotely disagree with you on any subject...than the Republican party is for you. All they want is world domination and all the money.  It is yet more proof that the Neo-Cons want to run everything and are willing to do anything to get their way. It lends greatly to the idea of shipping them all off along with anyone in Israel that identifies themselves with the Lakud party, or any person that is born out of Taliban or Al-Queda thinking, to a remote island and arm them to the teeth. Let them destroy each other...and once one party is left standing the rest of the civilized world take a joint action and bomb the island back into the sea to make sure that we can be rid of all this nonsense.  This makes me want to take what is left of my life and go start killing key people that are causing problems for America...and believe me there are both Republicans and Democrats on the list....no one in power would be safe.  Alas I am to sick to do it though...but a guy can dream can\'t he.  '),(1277,1234212,'07,April,2003','2003-04-07 00:00:00','   Today, well more like yesterday Dean launched his \"new\" idea...so far I am underwhelmed, but hey it is just getting started.    Democracy for America, Dean\'s \"new\" idea.   What it looks like he wants to do, and really it is a good idea, is keep the grassroots movement that he and Trippi started and give it direction for not just the Presidential race, but all races from local to state to national. In addition he wants to try and get the disparate groups together and pool resources and unify them all into a single entity with many specialties.   I would imagine part of that would be getting people along the line that can go on to a News show or talk show to counter the Right\'s onslaught of pundants, experts, and policy wonks. Everyone would be under a unified banner while keeping their individual ideals and direction. Kind of like how our nation is set up. We have states that make their own laws and go about enforcing them and take action one way, but all of the States fall under the Republic of America or know as The U.S.A.  That so far as I can tell, and coming out of various Republican campaigns that I had worked in over the years, is the weakest part of the Liberal movement is lack of coherence and direction. You have a thousand groups pulling in different directions at once. It is not a wonder that the Liberal point of view has largely been lost in the din of the Republican machine. If they were to get together and pool resources the movement could go a lot farther and have more of an impact on policy.  Another thing that has been a big mark against liberals is that people think that is a dirty word...well many Americans at least. The Republicans have won a major coup by redefining the word and turning Liberals into these Ivy League pricks that all they want to do is raise taxes and then go out and spend the money on the National Endowment of the Arts so some starving no hack artist can carve a penis out of prison soap.  The reality is that Liberals are far from that...well ok maybe one or two want to do that, but almost all of them to a tee want to see the Government returning services for the massive amounts of money it collects each year. The majority of that money they currently collect go towards the Military, and hey when is the last time the Air Force gave you a call asking if you wanted to take the new F-22 Raptor out for a spin, or the Army drops by and lets you run the M1-A1 Abrams through its paces and fire off a few practice rounds. It doesn\'t happen.  Liberals want to see that needs are met for the neediest, and that our infrastructure doesn\'t fall into disrepair (which it already has).   What is funny is that some have been calling for National Single Payer Health Insurance. In other words health care for everyone or universal health care. Republicans have been saying \"it will never work, it will cost to much, sounds a little bit like a communist idea\". The Republicans also paint it that everyone is against it. Well that really isn\'t the case. Currently GM has billions of dollars of yearly outlay for health insurance for current and retired workers, they estimate that 1200 dollars from each vehicle they sell goes directly to health care. Other companies are seeing similar problems. As health insurance prices go up, the harder it is for a company to afford them. They have to raise the employee contribution portion, or lay off people, raise prices on the product or service they provide, or in some cases they move plants overseas or to Canada where either Health care isn\'t thought about (China, India...have health care but state sponsored and not really high tech) and to Canada where they have great health care and it doesn\'t cost the operator/owners of businesses a lot of money.   It doesn\'t help that the worlds largest company Wal-Mart doesn\'t really offer good health care, and most of its workers don\'t carry the company insurance unless they are management, most of the \"associates\" don\'t get the insurance. With that huge group not having insurance when they get sick and need to go to the hospital, then the entire system pays and prices have to go up to cover the uninsured.  The health care system is in a mess. So many of the larger companies are coming around to universal health care. 60% of Doctors are willing to take minor pay cuts if it lessons paperwork, and along with it there is some sort of malpractice lawsuit reform. I am not saying tort reform, but rather a new system that vettes out the bogus claims and allows the real ones to go on to court where the person can get a bijjilion dollars or whatever.  It is estimated that if we moved to single payer we would see a saving in administrative costs of almost 300 Billion (yes Billion) dollars and that would nearly cover all of the currently uninsured.  It is all a big mess that can be fixed...and with this grassroots idea it could be used as a lobbying tool, but also a public education tool that informs people of the misconceptions that have often been floated by the Republicans.  So that is where I think it will be headed...it is going to take some time to get rolling, but once a stone starts rolling downhill sooner or later that stone picks up speed and by the time it is at your doorstep it have enough force to blow right through your door.   Peace  Oh yeah and I am going to work on links and stuff when I feel well enough to code it out.  '),(1278,1234212,'07,April,2003','2003-04-07 00:00:00','   Now I know these poll results are more than likely scewed...because we all know that the Presidential Race will be much closer than what this poll shows...but it is interesting nonetheless.  I wish that the site had a place where you could write in your vote...I would have given Dean a vote. Suprisingly even Nader scores higher than Bush....I could only wish and hope that this really was how things are going...but we live in the real world.   Here is the poll that I am talking about.   Maybe it will come true...we can only hope.  Again the poll is by no means really representative of what is going on (although more and more people are pissed off now towards Bush). The poll is being held by a liberal group so...take it with a grain of salt.  Finally a journalist has done his/her job and came prepared for an interview. Who was the interview with? None other than our pal Donald Rumsfeld or Rummy for short (because he must be swilling gallons of the stuff if he really believes what he says...you could only believe it all if you were high or drunk or of some sort of altered mental state). I don\'t know who the reporter is or news guy...but I would love to buy him a beer because he came prepared and actually called Rummy on one of his many lies and in doing so set ole\' Rummy on the defense...when he is confronted with the truth he breaks down and starts backpeddling as fast as he can. Watch the video from the link below.   Classic Video that call Rummy on his lies...priceless!   Peace  Probably larger update later...  '),(1279,1234212,'07,April,2003','2003-04-07 00:00:00','   As many of you who read my blog with any regularity know that I am not a fan of Shrub\'s Administration at all. Many of you know that I voted for him in 2000...and I regret that vote all the way. Every chance that I get on my blog I like to go after topics and issues that Shrub has screwed up or lies about.  Now don\'t take me as just a simple Bush hater. I hate him for sure, but my hatred is of equal opportunity. I really just hate overall politicians and the system overall. I think that it is no longer for the people at all and has been corrupted by money for not even \"special interests\" anymore...it is now just the government works for the corporations.  Unfortunately the average citizen is poorly informed of the past. Most people simply don\'t know history. So lemme give you a quick little American History lesson. Really I will only focus on the latter part of the 1800\'s and then this past century the 1900\'s.  In the late 1800\'s the nation saw the rise of Robber Barons. These men were \"capitalists\" of first class ratings...something that many current businessmen would love to emulate and move towards once again. These Robber Barons exploited labor, exploited the land all for the mighty dollar. There was very little regulation of business at that point in time and the Robber Barons spent vast amounts of money on politicians to keep it that way.  These \"men\" made their wealth on the backs of other real men and never gave back or rarely gave back.   The 1930\'s saw some change in the air. The depression was in full effect and the average citizen was suffering either working many hard long hours for almost nothing (with the fear that if they tried to stand up to their employers that they would be fired...and if you had a job you were lucky and did nothing to rock the boat). The wealthy continued to do well. As a matter of fact their wealth soared while the average American barely and rarely put food on their tables.  Some people that had both money and a conscious saw that it wasn\'t really right and with their political positions started to do something that had never happened before. The government started passing laws that limited what businesses could do to employees and started to give the employees sort of sharing in wealth. The government also embarked on a public works building spree. During this time we saw thousands of miles of roadways paved, the Hoover Dam...you get the idea.  Also Social Security was also passed during this revolutionary period.  Basically what was happening was that the government wasn\'t taking from the rich or redistributing wealth...instead it leveled the playing field as best as possible and with the leveling that allowed for ordinary Americans to move forward and earn more money.  These laws and regulations continued for years. New things happening all the time. The minimum wage laws, child labor laws, the EPA, and many other things came into being. These changes were happening all the way into the 70\'s. Nixon was the President that signed the EPA into law and existence.   Sure Democrats were usually at the center of the social programs and the ones who pushed hard...but if we look back we can see that many prominent Republicans actually proposed and passed many of the social programs.  In the eighties things really changed...do we need to say why.  The middle class is not an actual entity. It is not natural in a capitalist system. There are two levels period. You are either wealthy or poor...no in between. The side effect or symptom of many of the social programs and laws that were passed to protect workers and the environment actually produced the middle class. The middle class was a byproduct of these laws.  During the eighties the Reagan administration started to dismantle many of those laws...well no they would take a bolt out here and a nut there...and while the laws remained they were weakened. Same thing happened under Bush 1. Clinton reversed some of the trend but in many other ways helped it along.  This Bush...GW...he has really dismantled a lot of those laws and programs. Not quietly but with glee and in the open.  Now that many of these things have been taken apart guess what has happened.  The Middle Class has almost completely disappeared. The class will be extinct in a short time.  Which brings me to what I really wanted to talk about...and that is jobs.  Right now our economy is in full recovery...or so the administration line goes. Well we really aren\'t if you are on the street or in public. Hey the stock market has been doing well (up until this past week and this week), but as I have said many time...Wall-Street has nothing to do with the real world. Sure they can be an indicator of how weak or strong the economy is...but it is by no means a predictor of job or economic growth.   Sure we are seeing growth...but much of it is credit card economic growth. The government is spending money like it grows on a tree and that helps grow the economy because they are buying stuff. Sooner or later that has to end because they will run out of credit.  Some companies are doing great...but their bottom lines have improved through layoffs, increased productivity from remaining employees, and lastly (but growing quickly as a reason to worry) through outsourcing and offshoring (outsourcing is being used incorrectly most of the time...that means to part out work to a contractor...offshoring is the correct term...it means that jobs are actually going overseas or out of the country).  Jobs have not been created. Sure the unemployment numbers have been falling, they are falling because so many jobless people have used up assistance, and stopped looking for work altogether. This means that our workforce is actually shrinking...and that is bad news.  Last month there were 21 thousand new jobs. However none of them were created in the private sector. All the jobs were created by government at the state and local levels.  Really this all wraps around to that right now we are looking at an unprecedented time in our history. We once again are seeing the rise of the Robber Barons just like in the beginning of the 20th century. I just hope that we are not doomed to repeat history...Unfortunately the adage stays true and never fails. If you don\'t know where you have been you won\'t know where you are going...or something like that.  The parallels are frightening...only this time the stakes are higher with abilities of the armies of the world to destroy the world many times over...and on top of it all our Environment is so battle scared that whether or not the planet can support us if we don\'t do something to conserve what is left is up in the air.   I hope we can fix things...but I hold little hope because the ordinary American has such a lack of knowledge about history...and this all isn\'t even ancient history...rather it is recent and you would think we could at least remember back 20, 30, 50 years.  Read up folks...then act to change.  '),(1280,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I am finally back home. You have no idea how different it is to go from a sleek little laptop or notebook back to a clunky desktop...even though my desktop is far from clunky.  The main thing is relearning typing. Of course the notebooks keyboard is smaller...and this is much bigger...but you wouldn\'t really notice unless you go from using a notebook exclusively back to a keyboard...I guess...maybe...I dunno...if you have ever had to make the adjustment you would know.  Well there isn\'t much to say outside of that I got a feeding tube placed. YAY! Well small case yay...it hurts like a bitch and will do so for a few more days I am sure...and all I have is really a pea shooter for elephant pain. Oxycontin is great for long term pain, Oxycodone is great for minor, dull, easily numbed pain, but this is elephant hurts like a bitch needs morphine or dilaudid pain killer level. It really hurts like a bitch. Not that those pain killers take away all the pain, but they take away the constant nag pain and leave only the sharp pains that accompany coughs and sudden movements. So right now I am in pretty bad pain.  Now that I am home I hope I can get a hold of people over the next few days depending on I am feeling. If I am in a shitload of pain then fuck it...I am going to be a hermit. If the pain is manageable then I will call people...because even breathing deep hurts like a bitch...  Well goodnight all...I am going to try and get a little sleep and another 1000 calories while I attempt to sleep....yippie!  '),(1281,1234212,'07,April,2003','2003-04-07 00:00:00','   This update is going to be short as all I can muster is shortness right now.  My laptop is going to have to go back in to the shop because the DVD/CD-R/RW is on the fritz. It still burns CD, and sometimes it will read CD\'s, but it will not read DVD\'s. I have tried several and two were new DVD\'s...so we know it isn\'t the media.  This frustrates me because my last Toshiba was reliable up until the end when it started to not read CD\'s.  We are going to see what we can do because frankly if the laptop is going to have to keep going into the shop then I don\'t want the hassle. So we are going to see if we can just get a refund...this make two serious problems in three months and I don\'t want to risk anymore.  If we get it returned then I will go completely insane and get an E-Machines laptop. M6805 Model, Mobile Athlon 64 3000+ (1.8Ghz), 512MB RAM, 60GB Hard Drive, ATI 9600 with 64MB graphics, 15.4\" screen, DVD/CD-R/RW. The package is 1500 or something close to that and a 100 dollar rebate.   I hope that they will take this piece of crap back so I can give something else a go. I just don\'t want to have fight with it anymore...and I foresee that I will be returning this thing back to the shop a lot...and if that is the case then it certainly not worth the time or energy.  Jeesh...everythings gotta fight.  '),(1282,1234212,'07,April,2003','2003-04-07 00:00:00','   I haven\'t updated in a few days again...I am getting lazy about this whole blogger thing. I need to get my ass back on the ball and get back to my political speeches and general hate towards all sorts of things conservative and crap that is even remotely associated with the Neo-Con right and their hate filled group.  Ok so what new is happening in my little world...outside of getting a little better each day not much. I am trying to get myself up to full speed and then find something to do until I get a transplant...if I get a transplant.  My sister got a job offer...essentially a little over 50 thousand a year...that makes me feel like a complete and total loser...which I am. I am never going to be employable...it is as simple as that. No one or company will employ me for any length of time because I am not healthy...and to even think about insurance...hahahah forget about it.  Getting the transplant will change all that right? Nope...not in the least. First of all a transplant is not a cure, nor is it a permenant solution in all reality. A transplantee has a 50% chance of survival out to 5 years...odds haven\'t been given to me for after that time. So in reality...or bringing the elation back down to Earth and actually thinking critically about it, at best a transplant adds five years to your life, or life as such. You can figure that the first year even if things go swimingly well you will still spend most of it recovering and getting up to speed. The remaining years can be broken up into another two phases as I see it. The first phase is good health overall as far as the lungs are concerned and not many problems. The second phase is probably the last year or so of that time (or a persons life if you want to be real about it) you see your health decline and problems start popping up and you start to get sick and feel like shit. You then die.  Hey...I am just being real about it.   So what does this have to do with me and a job? I want to be productive in some way. Earn money so I can do things and get a Mini Cooper. I just want to be normal as possible. I know that I cannot be.  I have so much in my head and so much that I feel that I have to say...but I have no credentials, I have no voice, I am a mute in a world full of people with bullhorns. Maybe one day someone will see what I feel, but that day must come soon as I don\'t know what I am going to do. I haven\'t decided to definitely get the transplant, I haven\'t decided on anything so who knows.  I am tired...and Taken is on. I am going to lay down for a bit and watch that...and read Michael Moore\'s \"Stupid White Men\"...so far a pretty vicious book...in a way that makes me wanna go and take money directly from rich whitey.  Peace...  '),(1283,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok well it again has been a few days and that was because I went to the hospital and while I got my laptop back I didn\'t have the proper cable to sign on and update...although I wasn\'t really all that much in the mood to do much of anything seeing that my ass was broken. They put me on GoLytly (I am not sure of the spelling)...that stuff works and works well. They should have given that to me a while back because it has been the only thing that has worked repeatedly and regularly. Well my ass is working again and that is good. I only spent two days there this time.  I am hungry but I can\'t figure out what I want to eat and frankly I have little desire to cook anything.  Well I am going to stop my bitching and go to bed. Love to all...peace out.  '),(1284,1234212,'07,April,2003','2003-04-07 00:00:00','   I haven\'t written to this thing in a few days because my ass is broken. Yes it is broken. I haven\'t been able to go to the bathroom since I left the hospital. That hurts and sucks...so now I have to try and get it fixed. Hopefully I won\'t have to go back in to get it fixed...hopefully it can be fixed with something from the drugstore...because I don\'t want to go back in for a broken ass.  '),(1285,1234212,'07,April,2003','2003-04-07 00:00:00','   Sick boy here...hehehehe...Well at least I have my sense of humor. May not have all my health but I still have my humor.  Today has been a long day even though I have slept (YAY!) through most of it.  Nothing else to say really...just out of words and thought right now.  Peace...  '),(1286,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I am back in the hospital this evening...yay for me. Its ok because I am getting sick and I need to be here to get things well again...or at least keep things from getting to far out of hand. It will already be hard enough to get these bugs under control so might as well get started as early as possible before a deep infection gets itself entrenched in there.  Outside of that I am still damn tired and I am hoping that while here in the hospital I will get a little sleep which would be nice...especially if I can get a cocktail of things that will knock me out.  Well I am tired and I stink from my PEG (feeding) tube...it smells south of cheese.  Oh by the way folks if you get a chance go see Ladykillers...damn funny movie that ends very differently than what you may think. While many critics have panned it calling it schlock essentially. I beg to differ. It is a Cohen brothers film which by default is quirky comedy, it also inherits a great use of the English language (and coming out of Tom Hanks he just does it justice even with the weird accent...almost correct for the place in MS that he says he is from), but the comedy is intelligent and you have to think about some things.  Let me help a few of you out there with one thing from the movie. When Hanks is introducing all the cohorts to each other and he gets to the General and describes him as a tunnelling expert from French Indochina...that means Vietnam, rather seeing that he was a General and a tunnelling expert then he is North Veitnamese. I am sure that many people don\'t know that...especially if you were raised during that period as it is not really a thing that they teach you in school...and many college profs. don\'t tell you that either. It is just a little bit of Jeopardy Knowledge...but it is accurate.  Also one other thing. For those who are looking at the screen and wondering who Mr. Pancake is. He is the Psychiatrist from the Original \"Law and Order\"...he replaced Dr. Olivet somewhere around 96-97 I believe. Anyways that is who he is so you don\'t rack your brains to hard.  I am tired and worn out...so I will rest now.  Peace...  '),(1287,1234212,'07,April,2003','2003-04-07 00:00:00','   Oh and I have been sneezing like sneezy dude...all the damn time. Plus I can\'t keep warm or cold either way. I end up uncomfortable so that sucks and that is why I can\'t get a good nights sleep...I get hot and sweat then I get cold and shiver...then it happens all over again.  I have tried all sorts of things including turning the heat up so I wouldn\'t need blankets...buff dat\'...didn\'t work. Tried turning the heat down and cracking the windows to get it cool in there...nope...so I might need to get to the hospital to get some damn sleep.  Gaa!!!!  Peace again....  '),(1288,1234212,'07,April,2003','2003-04-07 00:00:00','   Sorry for the lack of posts. I am sick yet again and my PEG (feeding tube) is infected. I went to the doc today and he thinks it is and put me on yet another anti-biotic.  I am coughing a lot more and shit...tomorrow morning one of the nurses from the clinic will be calling me and if I am not feeling the slightest better she will get Dr. P and see what he wants, which will be head to hospital. Do not pass go, do not collect 200 dollars. Which I guess will be ok because they will be able to get me back on drugs that will knock my ass out for a bit so I can get some sleep. I haven\'t been sleeping hardly at all...and it sucks ass!  Oh and speaking of 200 dollars I went to the Social Security Office and got some things straightened out and the money that has been deducted from my monthly SSD money for Medicare A+B will be refunded to me...now the problem is for how long back...I have been paying it since Oct 2002. PA picks up the cost of that when you are in a position like myself, disabled and not doing well...and while they should have been paying all this time something and some people didn\'t do there jobs...of course. So now we have it straightened out. I will be looking at between 600-900 back from them if they go all the way back to Oct 2002...normally they wouldn\'t go far that back (the state) but because they screwed up in the first place they are at least going to review the case and it looks like I will get some sort of chunk of change back. I will know on May 25 what the news is and how much I guess....nonetheless is that they will be picking the cost up from now on.  Well I am headed to bed...I feel like shit and I think that I will end up in the hospital tomorrow.  Peace...  '),(1289,1234212,'07,April,2003','2003-04-07 00:00:00','   I may be going home on Monday if nothing goes majorly wrong between now and then.  Well dinner is here and I am a little hungry I guess.  Peace  '),(1290,1234212,'07,April,2003','2003-04-07 00:00:00','   Well my port is getting infected again. Whoopie!  Outside of that I am headed down to physical therapy to walk on a treadmill and work with weights. Does anyone know how much I dispise physical activity, I avoid it at all costs because I am pretty lazy when I am aloud.  Well I am going to finish this lunch and then go to therapy....  '),(1291,1234212,'07,April,2003','2003-04-07 00:00:00','   I watched the Bush\'s little press conference and he yet again has confirmed that he is one dumb ape.  I find it hard to take seriously any person who says \"terrists\" and \"my fello merricans\". He seems to not be able to read very well because when he was reading his speech he had trouble sounding out some of the words.  The Q&A was a joke. It was so staged that Andrew Loyd Webber would have been proud. When he was asked questions about whether or not he felt that he was in some part responsible for letting our guard down, or simply to appologize he said \"I think back to that day and it makes me sick\". What did he have some bad crab cakes that day? The correct responce should have been of sorrow and looking for absolution because he and his staff have clearly to many now, fucked up.  Even more now then ever I feel that the people in power of this nation at that time knew exactly or knew well what was going to happen and to be able to get their way they let it happen.  Since then we have invaded a nation that had nothing to do with terrorists. A nation that happens to be sitting on massive amounts of oil. Now I people may say \"yeah the oil thing\" in the classic way thinking that it is for our nation. Actually we get very little of our current oil needs out of the middle east. However Europe\'s oil and natural gas suplies come from that area.   This is giong to be an economic card that we can use down the road to either hinder the European\'s economy or ability to make war. So what I see is that we just bought ourselves a little insurance because down the road the next major power will be coming out of Europe both economically and militarily. Imagine if the Europeans joined with Russia. All of a sudden you have a massive power that cannot be stopped. With our control over at least Europe\'s supply of oil and fuel we can dissuade them from challenging us. The Russians have their own oil fields...but we are making a play for that as well. Afghanistan, Kazakstan, and several other \"stans\" and those nations around the Ural Sea are sitting on the same oil that supplies the Russian oil fields. While we can\'t turn off the tap to them, we can certainly reduce the overall supplies available.  Of course at the same time that oil that we now kinda own is great collateral towards issuing bonds and such.  Bush was lieing on stage there as he was lieing back in 2003.  I feel for the people that don\'t understand the larger picture here. To get what the Neo-Cons want they had to have an initial attack against us. So when this came down the pipe instead of preventing it, or even feining that they were doing something, they let it happen and then had their reason for starting this war against others. The war on terror is not a valid war. War by definition is against a state actor and cannot be against a group or idea.  I feel for those parents of soldiers that are so misled and still support what is going on. If I were a parent today and a child of mine was faced with having to go into a war like this then I would have to say no and protest loudly because as I said....war can only be waged against a State (Nation), not ideas or a group.  For those who have lost loved ones in Iraq...I am sorry. I am sorry that I cast my vote this last time for Bush. I am sorry for the party that I used to belong to. I am sorry that your children are having to go over to Iraq in my stay for something that is absolutely wrong.  The problem with being there now is that if we pull out cold turkey then the place will fall into shambles and many will suffer and die, if we stay there we will loose many more people and our nation will once again be forced into a draft.   Yes folks a draft. It will be the only way that the military will be able to get enough warm bodies to put over there if the plan even to deploy 100,000 for the time being (we already have gotten the hint that we will need around 140-150 thousand at all times for a long time...and that doesn\'t give us a sure victory). Right now our military is really stretching thin. How can we tell this? Because we are calling up even National Guardsman because they already have deployed most of the reservists and all of the active duty. Now you have many soldiers over there being told they will have to stay longer than the anticipated year they were initially told.  Pretty soon they will run through all of the troops they can rotate through and then they will have to draft people. The draft boards just recently have been calling on local delegates to come in for training and other stuff. The official story is that the Draft Board people are at the end of their 20 year commitments and that new people are needed. That doesn\'t was though considering the draft was ended in the early to mid 70\'s then the newest of the board members would have been done in the ealry to mid 90\'s. Now we are remanning the Draft Boards. Sounds fishy, and considering that the Administration has been really good at lieing (not covering up though).  Another small note I would like to say. When one senior government employee comes out and says that the administration is doing this or that. You can say \"well maybe they have sour grapes\". The second employee, well he was fired so he has a bone to pick. Third person the same story about this and that. By about the dozen or more employees and some being very high up like Clarke and O\'neill with no reason to lie (becuase their asses are on the line as well) then you have to start to think and believe that all of these people can\'t be making it up. The conclusion that a person has to come to is that the Administration really was doing all of the things they were accused of (especially when it was confirmed over and over again by different people) then they must have been doing it. When the White House then launches a smear campaign against the people speaking out, that should also be a hint that the White House has much to cover up. If Clarke and O\'neill were really as bad as the administration says they were then why didn\'t they fire them earlier and why didn\'t they come right out when the person was gone and level the accusations against them before they spoke out. It is all too convieniant that after these people blow the whistle is when the smear starts and not when they left or before they started to talk. This is just rational people. I will take all of those supposed \"sour grapes eatin\" people\'s word over just a few like Bush, Cheney, Ashcroft, and Rumsfeld. We all know what those \"men\" have done, they have drug us into wars and taken away many of our freedoms, so I don\'t trust any of them. I trust the recently unemployed from the Bush white house.  Peace  '),(1292,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok so Ashcroft is a moron and a lier.  Notable Quotes: \"Evil Chemistry\" \"Evil Biology\"  Ben-Venista - \"So you had no experience in dealing with Washington beauracracy\"  Ashcroft on his private jetting about because of \"threats\". \"I still fly on commercial airlines for personal business, as a matter of fact my wife just flew on a commercial carrier.\"  And the excuse as far as I could tell for not using commercial airlines all the time was because of his security detail carrying weapons and that poses problems on airliners.  He just needs to be bumped off so he can sit by his father...Satan. He is evil...very evil. Listen to him sing sometime and you will know that he is evil. He sings like a bad televangelist. Now that is evil.  Oh he wants more power for Patriot Act and finds no fault in holding people secretly and indefinitely.  EVIL I TELL YOU...I think even Satan would shove pinapples up his ass.  '),(1293,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok so I got my computer rebuilt in about 3 hours with all the basics. The only thing that is really left to do is install a few media players and codecs, and update Windows. The updating is a long process because while the Cell phone connection is faster than dial-up, it is still slow for big files.  I am hoping that maybe I can get up and running here soon and get moved on with my life. It is time. I don\'t know what I will do for money or a job as I don\'t have a degree (and frankly my health even if I get a transplant may still be an issue...the transplant doesn\'t cure you...it only \"fixes\" the problem).  I am hoping that when I get moved on I can live comfortably and with niceties...but nothing fancy. I just want to finally be free from everything and on my own as best as I can.  Realistically I know this can\'t and won\'t happen unless I go through with transplant. I was dead set against one for years, then I decided to go ahead, now I am just having doubts and fear and not feeling right about it all. I mean I have done stupid things and wasted my life up to this point and not accomplished much of anything...why do I deserve another shot at it all? Will I do it differently? I really can\'t answer that. I wish I could. I am just frightened and frankly I still have no one to talk too.  When I was in school still I went to the counseling center every week for a little over a year (well at least while I was there and it was open). I had a great counselor...but while it helped for a few hours or maybe a day afterwords I would crash right back down into my self loathing state. I don\'t want pity or anything like that...I may have sought it at one time but it never did any good. The problems always remained the same. I am self destructive in many ways...all of them attributable to the fact that I will die young if I don\'t get a transplant, and if I did I could only look forward to a few extra years realistically.  It really is a hard position to be in. My family seems to be pushing for the transplant and just assumes that I want to do it and will do it because I have gone through the procedures to get listed. I got listed simply because I would rather be listed and not accept, then not and all of sudden find myself in the position that only a transplant will save me and me not being on the list would basically be the death nell.   I don\'t know how to approach it all. This is new and there are no answers.   I feel really alone. I have people around me, my family supports me, and I have a few people that I consider my friends (but don\'t ever see them because of my poor ass health). There was a brief period where I didn\'t feel alone but the person that was in my life at the time felt alone with me because I couldn\'t open up (which I couldn\'t without looking like I was seeking pity). I have never felt that way again and frankly I won\'t. Simply because I don\'t have the energy or drive to even consider getting into another relationship unless I literally fell into one (in other words I don\'t even have to work at it outside of being nice...and that ain\'t happening anytime soon). So what to do, what to do.  I hate to post these words because it leaves me open...but I promised myself that I would never self censor unless it was for legal reasons...otherwise everything and everyone was fair game. It has gotten me into trouble with some friends and family because I have pretty much stuck with that.   That is what this blog is about. The complete lack of censorship, and just plain out stupidity on my part (because I have a lot of stupid thoughts and many people read them here). This blog is not my ultimate chronicle or anything, but it is a telling of my attitude from day to day and my political leanings because it all ties together for me. My life is thinking to much, writing jibberish, being outraged at the latest things coming out of D.C. (from all sides because most of our law maker I am sure are just plain retarded...or crooked as hell...or both). Hell it is all tied together...and for me I feel that it is my welfare that is at stake when those morons get a good idea. When they want to reform Medicare, that effects me. When they cut back on funding for health care funding to the states, well that effects me. The cut the budget for education. That effects me because students get dumbers, and sooner or later those students become workers and with most of them being dumb (even ones that get a college degree) by thinking that the American way is avoiding their duty...and the Christian way is not helping people (that one I will never get...where in the bible does it not say to help your fellow man...the Bible says over and over again to help them)...but damn when workers wages go down they pay less into the system that helps keep me alive.  When we go to war I get worried. Not because my ass will be on the imediate front...but because when we go to war things at home go to shit. With this war that our \"leaders\" have gotten us into could very well involve us all (and I believe that the essentially good...but stupid people of the USA will reap the whirlwind of how or leaders have abused and run our good names into the ground).   So everything that goes on I think about. I am selfish I guess...but it is because the welfare of society that it effects me. If society as a whole is doing well, people are working, kids are learning something outside of pigskins and sports stats, and we are at peace with our neighbors on this tiny planet then things will be good for me (and everyone else). If things are shitty...sooner or later that shit falls on us here.  My biggest fear is that we are already to late to fix things. We could start doing things correctly tomorrow but that won\'t guaruntee that things will be ok...it is still a role of the dice for that. Yet our \"leaders\" and many of my fellow Americans think that screwing up our planet (in any number of ways) is ok and that we still have time to fix it...let the technology mature a little and the cost come down and then we will implement things. Well fuck people...things don\'t get less complicated unless you are implementing them, prices don\'t fall if the products aren\'t ever offered (even novely in the case of the GM electric car).  Then we have the people that do want things to change sabotaging things as well. I have the solution to how to get \"free\" hydrogen without using nuclear reactors or stripping it from hydrocarbons (fossil fuels...which kinda defeats the purpose if you still need gases to fuel your hydrogen vehicle). The answer is simple...but it is scary because we have to become godlike. We have algea that goes around making the vast majority of our oxygen on this planet...we know how to tinker with the basics of genetics. I am sure with enough funding and trial and error we could come up with a bacteria or algea that instead of belching out oxygen it spews hydrogen. Hey the new bugs and stuff you create that don\'t work...well destroy them and then they can\'t slip into the wild. As for the bacteria or algea that does work you control it carefully because you don\'t want it taking over and making our atmosphere toxic (and we could even engineer a way to kill it directly without harming other algeas or bacteria). However we have those nuts who see any form of genetic engineering for any purpose as bad and they forget that man has been playing with genetics for thousands of years. Man has been for all of agricultural time been crosspollenating and creating new species or types of corn, wheat, and all sorts of plants. The have orange trees that spit out grapefruit as well. So we have been doing it for a long time now...only now we understand what we are doing and have better tools to work at the genetic level than ever before. Sure its scary...but we have to go down the road or someone else will.  That is the problem here in our nation. Once we were adventurers. Once we took chances. We made many mistakes and screwed up a lot as well as being the cause of most of the environmental problems. However now instead of being bold we are being timid and not facing issues. Instead of venturing out we are shuttering ourselves in just like any other fundamentalist society. We are slipping back into another dark ages because we are being told that progress is evil...and because we are so afraid they may be right we defer to their judgement when in reality those are the very people that have caused much of the untold death and suffering throughout the ages. Yet they accuse anyone who wants to move on and progress as people who want to destroy life.  Wake up people. Those are all the things I ponder daily...and for many of those reasons they are why I have doubts about a transplant...because if I can puss out without pussing out before all the shit hits the fan...I would rather do that then try to survive through what I feel are impending horribly dark and dangerous times.  The only way we can even think of avoiding those bad times is by changing things completely...and rebuilding almost from the ground up. I am also thinking about this whole capitalist system and think that it is a model that cannot work. Like Dr. Nash has pondered if we all are attempting to attain the same goal then all of us will loose. So we need to rethink what we are doing.  Do that...if nothing else think about these things. We won\'t suffer a loss of quality of life if we no longer \"make\" money but instead earn a living. Changing from fossil fuels to hydrogen and other power sources will not cost us as much as we think...it will cost less than having to invent some sort of technology to scrub the atmosphere or running it so far into the ground that only the rich survive because they can afford to rocket off this rock that they screwed up through poor leadership. All leaders are rich...all of the rich are leaders. It isn\'t a coincidence...it is the sad fact. Those few that are born leaders rise to be rich. Those rich that aren\'t leaders loose it all.  Just think...  Peace  '),(1294,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I did it today. I pissed my mom off. Over something stupid too.  I have been watching Band of Brothers on The History Channel and have been really enjoying it. The couple of nights that my mom has been here during the show she starts to knatter on about one thing or another...tonight it got to me.  She told some story tonight about a skit on MadTV...she thinks I didn\'t listen but I did...heard every word of it. It just wasn\'t funny and when she was telling me in the middle of the show I was less able to pay attention. The respiratory therapist was in here to beating on me so they were chatting back and forth making it very difficult for me to hear what was going on.  Tonight I just let her know...obviously it was the wrong way. I just said \"I will just catch it later tonight\"...she then just went off.  Now I know I am an ass sometimes and I haven\'t been in the best mood for a while now but I really didn\'t mean for her to take it that way. I meant it literally that I would just catch what I missed later on. No biggie.  It makes no difference what I say or do most of the time. I just have this knack at pissing my family off.  It will blow over sooner or later I just wish that things like this wouldn\'t happen...I guess I will just have to muzzle myself...but even then I will catch flack because of not talking enough...six one way, half dozen the other.  I know that I said it wrong and I know she overeacted...I just wish that wouldn\'t happen.  Ah well what can I do about it...not much she is already offended and she will take it out on me for a few days.  '),(1295,1234212,'07,April,2003','2003-04-07 00:00:00','   Just dropping a line in on the blog so people won\'t think I\'m dead. I just don\'t have the energy to do a lot of typing and thinking right now.  I have some things to say but I want to get this page updated before I do another taxing post. So I am going to work on the layout (or style) tomorrow and hopefully it will be up and running tomorrow as well.  I am still thinking about getting a private domain and hosting...I just haven\'t decided what I would do with it all.  Well until later...peace.  '),(1296,1234212,'07,April,2003','2003-04-07 00:00:00','   Well my room is painted and I have gotten it all re-arranged and stuff. It is in its summer configuration. It really is hard getting used to a full size keyboard...jeesh.  Well I am going to hit the sack because I need to finish cleaning tomorrow and then I have to help my mom with the collectables down at this fleatique thingy...I can go sit for a few hours and be bored or something....  Well night all.  Peace  '),(1297,1234212,'07,April,2003','2003-04-07 00:00:00','   I forgot to tell ya\'ll what I did the other day when I went out for trouble. Well I went to a movie.  I went to see Man on Fire with Denzel Washington, Dakota Fanning, Christopher Walken, Mickey Rourke, and Mark Anthony. I will say this now. Washington is a serious contender for another Oscar for this movie, and I would also go out on a limb and say that also filmography or whatever the catagory is that Black Hawk Down won in the film department. This film was not only brilliant...and it most certainly is not the same schlock that just keeps getting pumped out over and over again. I have been thinking about what to say about the movie and first I can say that Dakota Fanning is going to be around for a while because she not only held her own with Washington but also in some cases took the scene and made it hers. I could see a supporting Actress nomination but I doubt the Academy would have the balls.  Walken delivers probably the best monologue that I have heard in a long time. It start with a little banter back and forth between himself (an ex assasin for some sort of government agency) and a Mexican equivalent of an FBI agent. The Agent asks something to the effect of \"what did she mean to him\", Walken \"she showed him how to live again\". Walken \"you see every man has his art something they are very good at. Creasy\'s art is death. He...is about to paint his masterpiece.\" Ok so while it is not the exact lines from the script or movie...it is damn close and frankly dead on. No one but Walken could have delivered that line...no one. Not one single other actor known could have made that line as stunning, chilling, fascinating, and riveting as he did is less than 30 seconds of speaking. If you were to hear that in real live you would shit yourself because the way he delivered it, the line was absolutely perfect, cold, and telling all in one.  So if you get a chance go see it. Warning that some parts are rought...but the rough parts are few and far between...and not shown in great lengthy detail.  One other memorable line that was really good.  Pita (Fanning) : \"Creasy. What was the name of your first girlfriend?\" Creasy (Washington) : \"Nanya\" Pita: \"Nanya?\" Creasy : \"Nanya Business now get back to work\"  One other note on Man on Fire...much of the music was done by none other than our good old northern boy Trent Reznor of NIN....The stuff he did for this is just mind boggling. As soon as you hear the first few measures you know it is him if you know his former work.  Go buy Underworld as well. Damn fine movie....and Kate just looks hot in Leather and with guns....she could turn me into a vampire anyday.   Grrrooowwwwwlll!  And yes there will be some changes soon. I am working on something here.  Peace  '),(1298,1234212,'07,April,2003','2003-04-07 00:00:00','   Updated a few things...  '),(1299,1234212,'07,April,2003','2003-04-07 00:00:00','   My friend...B over at  Brassmask.com  and his wife just had their baby about a week and a few days ago.  I can remember when they had just met like it was just yesterday...and now they are a family.   That is the part of time that I do like when it passes when a great pair of people get together, make a life, and then create a life, and I know that they will raise their child with love and enlightenment.   With all the grumbling that I do about things passing me by and life moving every once in a while I get word of news of something wonderful and something that I can look at and just be happy about. This is one of those times.  This update is dedicated to Benjamin. My love out to you and your parents.  Peace  '),(1300,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok...it has been well ten days since I last updated and I am sorry.  The last few days that I was in the hospital they had to remove the port in my chest and go back to regular old arm IVs. Well the problem with that is that my arms don\'t really have good veins anymore due to all the prior IVs that I have had. Really I have maybe, I stress maybe four veins they can hit. The rest...shot. The veins they can hit are all in places that prevents me from using my hands...two on the wrists and two on the underside of the forearms. All of those locations are not ideally suited for typing or using mice...or really anything. Plus on top of it the veins are weak and when you put oh say benadryl into them...it burns the piss out of them. So to get my meds I have to dilute them down greatly and it takes forever and a day to get everything in...so it is incoveniant.   The reason they had to remove my port was because of a persistant fungal infection. I have had so many anti-biotics they have killed off all the good little bacteria on my skin that eat the fungus that we all have. So I got a fungal infection that stuck to the actual rubber in the port and it just wouldn\'t let go so they ripped it out.  Next time I go in (oh say in a week or two) I hope that my fungal infection will be under control to a point where they can go ahead and put another port in me...because when you need the anti-biotics and all the other things that I get when I am in, you really need to have a port. So hopefully I will get one put back in. If that doesn\'t happen then I will get a PICC line which is in the arm...it is uncomfortable but it can take a lot of abuse and a lot of meds and it doesn\'t burn. It is just a pain in the ass.  I did get out of the hospital...on Sunday to be exact. I can honestly say that I don\'t feel all that well but I don\'t feel to terrible. I just want a week, maybe two of good times and decent health, and maybe decent sleep. I can dream.  Well I talked to one of my old friends yesterday. She is getting married in September (or Septembre\' if French...if I remember correctly). It came as a shocker to me...but hey everything is anymore. I have been left in the brokedown duck lane of lifes highway. I can\'t swim, I can\'t waddle, and I sure as hell can\'t fly...so I just have to watch everything and everyone go by. My sister is done with a Doctoral level degree, my \"friends\" are mostly done and gone, many on the road to the rest of their lives, while I just sit here sick and not able to move forward. I want to but I can\'t.  Well enough wallowing for right now...I am going to get cleaned up and do something today. It is a nice day and I am free, white, and over 21...so I can do what I want...well within reason and my credit limit. So off I go to see what kind of trouble I can get into without getting into trouble.  Maybe a movie or two, food...and soon I will be able to get my PA license and then nights out drinking till I\'m blind will come...because hey with my liver functions anymore that would be oh say a beer or two...hehehehe...  Off I go...peace out peeps...  Oh and something that I wanted to say that struck me as oxymoronic.  A few weeks back we had the NRA national convention here in Pittsburgh. Vice President Dick was coming into town to speak to some of the attendees. The ironic part of it was that they had security out the ass and all of the attendees had to go through security checks and metal detectors. Now call me silly but this dinner and speaking engagement put on by the NRA was on gun rights and such (I must mention that I am a lifetime member...a gift from a family member and I happen to believe in the Second Ammendment...I believe that you can own a howitzer and as long as you never use it inapropriately and for a crime then go right ahead...but I do also believe that some form of double blind registration should be in place so that you can\'t locate a person through records but you can tie a person to a gun kinda thing for some measure of accountability). Back to what I was saying being a dinner and speech by the NRA it strikes me funny that they would have metal detectors for all of the \"law abiding\" people to catch guns. If their was no danger from even this group of normally avid Republican supporters why would they need all that security for the VP.  Very oxymoronic.  Peace.  '),(1301,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I pretty much have things straightened out with the laptops. It is taking a bit to become accustomed to the smaller keyboard.  I am also lucky because I got a G-Mail account today. 1GB of online storage. I can get huge emails!   Well I feel like hell because my kidneys hurt like a bitch. If they don\'t stop hurting soon I am will call the doctor and see what to do. I took some Ibuprofin and they don\'t hurt as much, but I can\'t take a lot of it because it can hurt your kidneys.   Well I am off for now.   Peace  '),(1302,1234212,'07,April,2003','2003-04-07 00:00:00','   Well tomorrow I have to finish up the laptop game so it is ready for my sis when she comes. I got her a card and will have it on the laptop when she gets here.  Tomorrow I just have to get data off of it and then toss in the restore disk. After the restore disk runs I will then install a full version of Office 2003, Norton Systemworks 2004, Norton Anti-Virus Pro 2004, and Nero 6.3.1.6 (latest version and the version that is working with my DVD burner). Then I will tidy it up and get rid of some of the crap the restore disk will install and make it a little leaner then I will update it and get it ready to walk out the door.  The laptop that I got today is indeed slower but not by much in processing, however I know that I will definitely have to get a RAM upgrade as soon as I can afford it. The graphics subsystem grabs memory from RAM and I have it turned down to just grab 16MB. Once I get more RAM I will let the buffer increase to 64MB. The only problem is that I will have to by a single big stick instead of just adding RAM. The laptop has only one RAM slot and it is used by the RAM that is in it now. Outside of the limited RAM it runs fine. I am surprised at how much the Athlon XP-M fluctuates in speed. It runs from down at 400Mhz when not doing much all the way up too 1.66Ghz when really busy, and it fluctuates fast.  Well I am tired and I am having some sort of acid reflux thing that is making me want to puke...but I am holding it.  Peace  '),(1303,1234212,'07,April,2003','2003-04-07 00:00:00','   I have been trying to figure out what I want to do for my sisters graduation. Well I figured it out. I am downgrading. I am going to give her my laptop which will keep her for several years. It is practically new and has plent of horsepower for all she needs to do.  I will get another laptop that suits my needs a little better. The Toshiba is just a little big for me to tote around. It weighs in at around 7 or 8 pounds with the power supply and for me to lug that in and out of the hospital is just getting to be a bit much. I found a good deal on a laptop for around 800 bucks. It is most certainly a downgrade from what I have now. The graphics subsystem is much less (S3 something to the Toshiba\'s Nvidia FX5200), the hard drive smaller by half (30gb to 60gb), and the processor less powerfull (Athlon XP-M 2000+ 1.67ghz to a Pentium M 1.4ghz...the Pentium M smokes everything in its class), but the big thing is that it weighs in at about 4.5 pounds with power supply. It is much smaller and will be easier to carry with me.  Why would I do it? Because frankly I don\'t need the power of the bigger system. It will be a good replacement for her giving enough perfomance and durability for a couple of years at least. If I could afford a better system then my own I would get it for her but I can\'t so I will give her the more powerful system. I have my home computer which is more than enough for gaming when I game (which isn\'t often anymore) and the smaller system while not suited for games would be enough for what I need and that is surfing the net, writing, and watching DVD\'s while at the hospital.  Well I am tired and I need some sleep...so I will update some tomorrow...I am going to call my mom and tell her what I am going to do and see what she thinks.  Peace.  '),(1304,1234212,'07,April,2003','2003-04-07 00:00:00','   Well good news today...well kinda. I am indeed sick again but not sick enough to go in the hospital. Doc is putting me on some oral antibiotics Cipro and Zithromax. So we will see. I don\'t react well to Cipro, my mouth gets sores and very raw. It is a painful side effect. I will only be on a ten day course so hopefully I won\'t get a raw mouth.  I am on the Go list for a double lung transplant. I was talking with Brian online today from  Brassmask . He asked the very questions that I have about all of this hoopla.  The major questions that I have and if I got negative answers before a transplant may influence my decision are. 1. What happens when I fully reject the organs? Do I go back on the transplant list or is this a one shot deal? 2. Will they save some tissue for possible future growth of new perfect lungs once that technology is available?  I have one answer already. They will be saving some tissue for research by one of my docs, but will it be available for making a new set once the technology is available.  As for the other question...I really don\'t know the answer. The answers to the questions will influence my final decision. However if they call before I have the answers then I will go ahead with transplant and deal with all of that later (although I will ask for additional samples to be preserved).  Am I ready for this all? Not really. I am not fully emotionally ready but I know that if thrust into it I would be ok. It just is something that I have to do.  Well I am tired and I need some sleep.  Peace  '),(1305,1234212,'07,April,2003','2003-04-07 00:00:00','   I forgot to mention in other posts that I got my b-day present early this year (as almost every year for the past 6). My mom got me tickets to one of the best comics out there and I can\'t wait.  I am going to see with my sister none other than George Carlin. He is probably the satirical comic out there. I cannot wait to see what he has ready for the masses in this tour.  My PocketPC was having screen then synching problems that I was able to fix when I opened the puppy up and found a loose cable and a little resistor that had come loose. Once I fixed those problems my PocketPC is all ready for a date. I am going to take it along and hopefully will be able to get a good recording of the show. If it works I will be so happy. If he has CD\'s there then I will buy a couple (would rather do that).  I am so happy...and even if I am in the hospital (which I probably will) I will drag the IV pole to the show...I ain\'t missin\' it for nothing (short of transplant).  Peace  '),(1306,1234212,'07,April,2003','2003-04-07 00:00:00','   So today I was thinkin...yet again about Stem Cell research and the future of medicine in this nation. As I thought more and more I started to get pissed off about what has happened to Stem Cell research and how severely limited it has become in this country as far as government funding is concerned.  One of the first big things that Shrub did when he got into office was to place a ban on harvesting new Stem Cell lines because of the misconception that all Stem Cells come from aborted fetuses. Because Bush is a pro-lifer (which should also mean that he wants to save lives but his other policies show otherwise such as pro death penaly and this horrible war in Iraq) he banned the harvesting of new lines because people are misinformed about the topic. Now I am not as informed as many out there but I do know that you can get Stem Cells from other sources outside of aborted fetuses (although they are the easiest to harvest these cells from).  Quick science lesson. Stem Cells are basically blank slate cells that can develop into any kind of tissue for living animals. Stem cells need only DNA to tell them what to turn into (I don\'t think the scientists yet know how they know to become lung tissue where the lungs are or kidneys where they are). The promise of these cells is that you can program them to turn into anything that you need grown. They hold very little information from the start...they do contain DNA from the parents of the Stem Cells but that can be extracted or overwritten.  As I said before Pro-Lifers wanted a ban on Stem Cell research altogether because the most common supply would come from aborted fetuses, however you can harvest some cells from the embilical cord right after birth.  The ban at this point only includes retrieving new lines if government grant money is used. These grants basically say that you have to use one of the already existing lines harvested pre-new harvest ban. At the time Shrub gave some number and said that is enough...in reality there are only about a dozen or so viable Stem Cell lines to do research on. So you can see the scope of research will be limited. Outside foundations such as the Hughes Foundation (Howard Hughes) are allowed to harvest new lines at will, but I fear that may come to an end as well.  A Hughes Foundation grantie has already grown new kidney cells from his research on these cells. Pretty soon they will be able to make filters for dialysis machines that use real himan kidneys cells and in doing so they could extend life expectancy on dialysis by years or more. The researcher hopes to soon be able to grow actualy kidneys for transplant within the next ten years. I believe he can do it.  This opens up a realm of possibilities because if he is succesful the research could then be expanded to heart tissue, lung tissue, any type of tissue needed. It also holds promise of being able to make blood at some point (stem cells could even be taught to be blood cells or marrow that creates blood cells).  Again the problem is an image problem nothing more.  We need to change that image and make it human so that John Q. Public can understand and here is how I propose we do it.  We go to old John there and ask him if little Bobby or little Suzy were dieing from some acute illness that effects the heart and their child has maybe at best a few months. Now they will also tell John that their child could be waiting years for a heart to become available but there is this new technology that can get the heart the child needs and put it in without fear of rejection. What do you think any sane or rational person would say? Of course they would want that procedure done.  In bringing it that way and leaving how it is accomplished out of the picture you remove the horrible image of aborted fetuses from their minds.   Now why do I think they need more lines than the few they have now since these cells are so versatile. I think that way because of the diversity of humanity. Lets say there are 7 billion people on the planet and maybe 20 million actual genetic trees (basically huge swaths of people in family tree like structures). In those 20 million or so trees there is great diversity. We only have a sampling of about a dozen of them now maybe not even that many. Down the road we may stumble into a situation where we really need to have that many lines just to get it right or get Stem Cells that develop tissue close enough to transplant without rejection into a patient. They may also find down the line that certain trees develop certain tissues better than others...like Stem Cells out of Arnold Swartzenegger\'s (sp) tree may be better at growing into muscle tissue and Stem Cell out of Einstein\'s (or Hawking\'s) would be better suited for nueral tissues. We don\'t know and we have already limited ourselves before we even got off the block.  We need the research, we need the additional lines, we need to be at the forefront of this field. If we chose to do that then we could guarantee new jobs and better health.  The possibilities are endless but we must first get over the negative image problem Stem Cell research has.  For the Pro-Lifers\' it is better to save something or get something good out of trajedy.  All of this is self serving because for me the issue is clear. I need new lungs. Scientists one day in the future could take DNA from my lungs, fix the CF defect, then grow a new pair for me. This will happen in the next 10-15 years. It would be better to be on the edge of the wave for this or we will loose it to the world.  Think about it...then educate yourselves.  Peace  '),(1307,1234212,'07,April,2003','2003-04-07 00:00:00','   I want to continue on the track that I started last night about our environment and the pending changes that are possible.  The biggest problem that scientists have is that they cannot give definite timeframes all they can do is estimate what will happen and when it will happen. The reason they can\'t say for sure is because this is all new to us. There is no recorded history of how fast or when a huge climate change will occur. They just know what causes it and that it happens.  There are several ways that our planet can be thrown into another ice age. We could have a meteor hit and have what would be essentially a nuclear winter where the sun is blocked out and the planet cools causing an ice age. We could have a nuclear war which would cause the death of probably 99% of the planets living things and then throw the planet into a nuclear winter and result in an ice age. The third way is how we are actually doing it...and frankly we don\'t know if its happened before or not.  We are pumping out hundreds of million of tons of pollution each year easily. This polution then hangs around causing a thickening of the atmosphere which helps trap heat and warms our planet beyond an optimal temperature. This warming cause the polar caps to melt. We are already seeing that happen greatly. The polar ice is retreating more and more each year basically giving us a warning that something is wrong. The ice caps are the place you go to find out the health of the planet because even the slightest shifts in planet chemistry or balance show up there glaringly. You don\'t go to New Jersey to get a gauge on how polluted the Earth is, you go to the ice caps.  You ask \"how does global warming cause an ice age?\". Well it is complicated yet very simple. In our oceans there are vast currents both shallow and deep. Sailors used to use them to get around and that is why for hundreds of years trade routes were pretty much the same because sailors were catching these currents. These currents actually serve more of a purpose than just getting boats around the ocean, they actually help keep the planets climate temperate and the oceans warm. As the polar ice caps melt they infiltrate these currents that are warmer than the surrounding ocean and cool them down. The effect of the cooling is that the currents first begin to shift and if the melting is extensive enough they could possibly stop alltogether. Now you can look at old sailing trade routes from hundreds of years ago. We have pretty good records from that period and they are fairly accurate. Now try and sail then and you will not have the same luck. The trade currents have shifted somewhat. Even if they have shifted a hundred foot you have to realize that a one hundred foot shift of a thing that has been happening for thousands of years is a major deal. To move billions of hectars of water one hundred foot requires huge energy being released. The engine for that energy is the melting ice caps. Not only do we shift currents the water level rises, and can rise drastically rather quickly. Look at Florence Italy. Sure it has always been the city of canals, and it has been sinking at a relatively constant rate for hundreds of years. Now it is sinking faster or so it would seem. Most likely the sinking is still at a constant rate but the water level is actually rising at an every increasing rate.  The currents that are in the sea are much like the jetstream above in the skies. As the jetstream dictates weather, the ocean currents effect the weather of the ocean. As it shifts around and weakens the sea cools. Seeing that 97 percent of the planet is water, what happens when all that water cools? It eventually freezes.  There is a movie coming out this summer called \"The Day After Tomorrow\". It will show what will happen in graphic detail. You may think it is just movie majic and movie hoopla but scientists have recently found that contrary to what was believed before that ice ages took many hundreds of years to happen and advance ice across land masses. The scientists have found that the changes from temperate climates to ice age climates took but a mear two to five years. In other words as far as geological terms a blink of an eye.  The scary part about all of this is that we are already seeing the changes happening now. El-Nino and La-Nina (I don\'t have the proper accents at hand) have become much stronger and severe making there presence known. How many people can think back oh say fifty years and honestly say they knew those terms. Now most people have at least heard of them and many can desribe what they are. Well to scientists who study climate and climate shift they are a warning. Hey look here something is wrong. All of a sudden these things actually become a problem. The past few years we have seen much harsher winters than what we have become used to over the past several decades. When global warming first starts sure things get a little warmer but soon the temps drop and become colder.  This year we have really seen the harshness of the new climate that we are facing. Right now in PA it is 50 degrees F. Normally or in the past the temps would be in the mid 70\'s or low 80\'s. The climate changes that we will see from now on will be more drastic and dramatic.   This all ties back to ocean health and currents. If the largest percentage of our planet is unhealthy then it is bound to effect the rest of the planet.  Another problem is that we don\'t know if we even change our ways today that it will make a difference at all. The damage may already be to extensive to avoid the catastophy that we are headed for. The only thing that we can say is that if we change things now we can curb some of the more catastrophic parts of what will happen.  The military was even worried about it enough to write a report on how we should tailor our forces and what conflicts of the future will look like. So it is real. It is not just hippies and tree huggers that think this way. Respected scientists are seeing it, the military is worried, and some companies are making preperations. So the administrations and Republican line that it is voodoo science is crap. The line of science that they buy into is voodoo. They believe that there is a big guy in the sky that will make things all better should we screw it up bad enough, but they don\'t even believe that we are screwing things up. Conservatives just think they are entitled to 25% of the worlds resources and a huge chunk of the worlds pollution output.  This all has to change. However it has to change from the top and the bottom at the same time. The top needs to lead and the bottom demand so the soft center will have to change and meet the challenge.  I\'m spent.  '),(1308,1234212,'07,April,2003','2003-04-07 00:00:00','   Tonight I starting thinking more and more as I was driving back from dropping off videos about what this country needs and what Americans need especially seeing that this coming election could very well hold the very future of our nation as we know it at stake...not to mention that the world hinges on what America does environmentally in the next decade (if we even have that much time).  If we stick with the way we have done things in the past and stick with the old ways then the Earth is destined to change. This change will kill a good portion of the population off. Now in reality...and if we keep doing things the way we are doing things then that won\'t really be a bad thing because if there are less resources then there has to be less population. It is a simple biological and ecological fact.  If we stick with Bush we die...simple as that. He will not change things. If we choose Kerry we still may die...simple as that. The same people that fund Bush fund Kerry. There is no difference, or at least as far as the money is concerned there is no difference.  What is different is the difference in possible environmental and ecological policies between the two. Bush as we know doens\'t give a flying fuck about the Environment. Most of the so called \"good\" environmental policies that will \"help\" are really just corporate give aways that enable companies to polute more while getting tax breaks. The \"Clear Skies\" Initiative, or so it is called, is really a cover that allows big polluters like coal fired power plants to pollute more. They accomplish this by creating a \"credit\" system. Each company that wants to pollute gets so many pollution credits (for various things like CO2 and Mercury). They use X number of each credit when they do what they do, like produce power. The problem is that these credits are transferable from company to company. What will happen (and it will even though they say they will have rules to prevent front companies to create bogus credits) is that big polluters will create front companies who\'s mission is to recieve the credits because they are going to make something, they then make the products but (albeit) efficiently (because remember it is cheaper now to build things to be efficient than inefficient and dirty) the extra pollution credits will then be either transfered or sold to the big polluters like coal power plants who don\'t really want to upgrade to clean technology. It is a system to replace the current clean air policies.   Hey what is wrong with this system? Well look at it. I could theoretically set up a fake company...or even a real one that produces something. Then I could apply to get the polluter credits. I use 5 of my 50 credits. I then sell the other 45 for millions a year to a polluter. I get free money. Not a bad gig. Now imagine this happening tens of thousands of times over and over again. Big polluters could then collect more than enough credits to not have to worry about how much pollution it pumps out. The main problem with this new system is that there is no total credit cap for the nation. It is not like they are going to say that there are only 1 million credits available per year. If there was a total cap that was lower than what we are producing now (which under old law each year we are supposed to pump out less toxic and harmfull emissions) than the system would work. However the system allows for unlimited polluting as long as you have the money to purchase other peoples pollution credits.  That is one of Bush\'s policies. So we know that he doesn\'t give a shit. He is only in it for big polluters who give him millions to run and make these new policies (which by the way don\'t have to pass through congress because it isn\'t law...just policy a big distinction). Whenever he either is voted out or gets through both terms (if Americans are stupid enough to vote him back in...hey I\'m all American baby I can say we\'re stupid) he will get large endowments to his \"library\" which is funny because I am sure the longest book he has read was \"Cat in the Hat\". So...  Kerry will at least keep the status quo and not do more damage and would roll it back to the old way.  The big problem with Kerry is that he has no vision outside of \"anything other than Bush\" and frankly being a hoitie toitie he will loose against Bush unless he makes himself different and make a stand for a bigger picture, a big thing, something that will save us in both the real sense, but will save us economically by putting us on the forefront of a new world.  This is what I propose Kerry do. Kerry needs to pull a JFK...as he so often notes that his initials are. He needs to say this and make it his motto.   Hydrogen Powered Economy in ten years.  Then he needs to outline how we get there, and here is how we do it. We are already using genetics and engineering stuff, we put money into it and tell our universities to engineer us an algie that burps out hydrogen instead of oxygen or methane (hell we can even strip hydrogen from methane), but coming directly out as pure hydrogen from these genetically engineered critters. Now I know the \"purists\" and those who are against genetic engineering would get all up in arms over it all but tell them \"come up with a better system and we\'ll use it\". All the mistakes that don\'t work simply destroy them, the one that does work we keep it locked up so it doesn\'t spread to the oceans. That takes care of hydrogen. We can do it in ten years.  For hydrogen production we are going to need massive amounts of water. We either convert the massive oil pipelines we have running all over the nation or we build a new system, but we pipe water in out of the oceans. We then clean it and de-salinize it. Not only do you get the fuel for our fuel, but we could also have no shortage of potable water and irrigation water. The salt we take out we use as salt or we add it back to the sea in the correct amounts for the water that will drain back into the sea (seeing that there will be greater water runoff to the sea.  So in this vision we not only get off of oil but we innovate and create jobs at the same time.  If Kerry does that then guess what...he may have a chance. He has to think big...not just for the presidency but for the world. He needs it, we need it, and the world needs it.  I hope he has the balls. Only two people in this race had this vision or something similar and had the balls to fight for it. Kucinich which he is a great man and Dean, who while isn\'t really a huge environmentalist he is a doctor and is educated well enough to know that we have to do it...not because it would be environmental...but because we have to. To keep on living we have to change our ways.  Until next time...Peace  '),(1309,1234212,'07,April,2003','2003-04-07 00:00:00','   Well my sister and I went to dinner with my mother and grandfather. Let me just put it this way. He is a two dimensional character not all that unlike Fred Flinstone. The difference is that he swears, is racist, smokes like a chimney, and is in the advanced stages of kidnay failure but still drinks like a fish and then heads out for a nice drive in his Town Car.  We went to this place that I hate because the food is overpriced and the service is shitty, and they have lounge singers. I have a fixed amount of energy for each day. I was worn out within minutes of being there.  One thing that happens is everytime I am driving with my mother and sister together I end up lost. Where this place is I don\'t know my way around. I got there fine with just my sister. When it was her and my mom with me driving shit just fucks up. My mother had said something about parts of the highway shutting down. I heard her say it during dinner but paid no mind because she said nothing when we stopped at a donut shop for some coffee for my sister. So I go to go back the way I came...on the highway. Well the way I need to go ends up being closed down. So I have to get onto the proper highway by going up an exit then planned on turning around. Well if I had turned around the traffic was detoured back to almost where we started. I ended up taking some bumwood road and it was a sheer guess as to where it lead...but we got to the detour route to get back on the highway. Now mind you this is my moms neck of the woods but she doesn\'t know North from South and got me turned the wrong way heading the opposite direction until I see a sign that told me we were headed the wrong way.  Now I love my mother...but she has a poor sense of direction. You could give her a map, a compass, and a GPS unit with NAV points programmed in and she would still go the wrong way. She thinks and acts like she knows, but she hasn\'t a clue. Fortunately this time I was doing the thinking and we didn\'t end up lost in some suburban hell like the last time she was \"navigating\". The last time the three of us ended up going thirty miles out of the way to get back home because she thought she knew where she was. I know better now and pay more attention to my compass and keen sense of direction (and photographic memory...if I drive it once it will always be in my head).  I have a lot of Indiana PAs area roads memorized. I can get from Indiana to Greensburg without using major highways, and from Greensburg to Latrobe using a highway for about three miles...the rest is back roads that a friend showed me.  I can navigate most of North Mississippi above Batesville without using major highways...I often thought that if ever in a car chase I could simply loose the cops because I was so familiar with all of the service, dirt, minor, and dinky roads in MS. I had it to the point where I could blast down them at speeds that would seem suicidal if you didn\'t know the roads and where you were going. I have tried for this area and I am getting better at it...but west of Pittsburgh is a mystery to me until Robinson Town Center...  So I am going to bed and I am going to sleep as much as possible.  Goodnight...I am working on more stuff to rant about...I am sure I will come up with something soon.  '),(1310,1234212,'07,April,2003','2003-04-07 00:00:00','   I wanted to say something else about religious zealots and right wing nutters like Rush Limbaugh, Pat Robertson, and Jerry Falwell. They are all pretty strictly anti-homosexual and are pretty homophobic on top of it all. Now I ask Rush and I doubt he would answer, if he were stacked in a naked man pyramid and had pictures taken and then distributed do you think he would be embarrased? Well it may happen...and I hope it does. I hope they convict his fat ass and he ends up a prison bitch because he already is a little bitch...well a big fat little bitch.  The problem with the left is that it does have way to many pansies in it. For years it has been weak in down to earth, salt of the earth language speaking people. Well not only am I intelligent but I can say fuck with the best of them and wrap sentences and misinformation that they spew around and hang them with it all. In other words I filled with such rage towards these cowards who pick on little people that I would have no compunction about getting in their faces and confronting them with the facts and numbers (if someone would compile them for me) and tell them that they are fucking liars. Call them out. Not in a funny Al Franken way (which is amusing but to deal with pigs you have to get dirty and root around with them). I would not take the \"moral and intellectual\" high ground. Instead I would spew the same venom back at them only with the truth from the numbers...and I wouldn\'t stop at the right wing pieces of shit either. I would call out the left wing pussies who have for to long tried to appease and make nice with these vile humans that deserve prison treatment no better than what Saddam afforded to his prisoners.   Fuck principles and moral decensy. These people have stolen the political landscape and molded it for their own gain and political raping.  I love Amy Goodwin, I think that Al Franken is marvelous, but unfortunately these people cannot reach the masses. Not because they are to left wing or boring but because they speak a different language. It is the language of reasonable and intelligent people. They don\'t speak NASCAR dad speech. I understand them and frankly anyone who gave them an honest listen would be amazed that they can understand what these people are saying, but in the world of venom spewing vipers like Limbaugh and his ilk you have to do the same. Not to be like him, but to beat him at his own game. We can\'t make them play our game, and reasonably we shouldn\'t have to play their game, but face it people it is time we start and we get good at it. There is no reason why smart people can\'t beat these idiots at this. We can do it better. Once we have won then we change the tone. Then we shift the country left or at least more center.  I advocate an all out attack on these people. If we can\'t play their game or we can\'t break into it then we break the game. All of these morons use technology they don\'t understand. Technology that peace loving people invented. Technology that smart people know how to manipulate. If we can\'t get into the game we shut it down. We take it off the air. If we can\'t play then there can be no game.  We need to have a multi front war on these people and their ideas. We can beat them if we put our minds to it. We can beat them on their own turf, and if they won\'t let us play on their fields then we pave their fields over and take the game away. We invented it, we gave it to them, we can take it back.  Call it the biggest Indian Give in the history of the US. (no offense to native americans...in my world you would have you land and nation inside ours if you so chose).  So Peace...and if that doesn\'t work then all out war. They are rich and fat and out of shape...we are poor and hungry and lean...we know what it is to struggle and fight. One or two good black eyes and they will back down like the pansies they are!  Well fuck me swingin\' (thank you Russell Crow for that saying) where did that come from?  '),(1311,1234212,'07,April,2003','2003-04-07 00:00:00','   I may have gotten myself kicked off of the CF mailing list today that I suscribe to because I simply cannot be an optimist and I must put before people the simple fact that under the current administration it doesn\'t matter what disease you have they want you to die. That is the simplest explanation that I can come up with. Harsh? Yes, but not as harsh as what they are doing to so many people.  People with cancer, alzheimers, cf, and most of all the inherited and even other diseases are being denied the best research by the government because of a moral opposition to stem cell research and a fiscal allergy to spending money on actually helping people instead of killing people.  Again this is all self serving because I need for billions of dollars of public funds to be spent on stem cell and genetic research. I need that money so that I can live longer (even though I don\'t have much that I really want to do outside of DANCE!...no write if I ever get my shit together and actually write down where I collect my info...I just hate documenting stuff, not research, but documenting because I love reading but not writing out all that crap).  Well I made it through today and made out with a couple of cards, a balloon, pretzels, swedish fish, an ice cream cake (I don\'t eat cake but my sister loves cakes and ice cream...she got it for \"me\" but it was nice and it tasted good so I am thankful). My mom got me Carlin tickets for Saturday (I have already mentioned them) and my sister is going to get me Scary Movie 3 tomorrow (they were out of it today). So that was my B-Day...not bad and thankfully it was small and I didn\'t have to expend a lot of energy on excitement. Last year I had a big bash (along with my sisters b-day) and it was very taxing because I was sick, on meds, and just had my leash put on at that point (glad that is over!). I had to expend a lot of energy on being chipper and I didn\'t want for people to think I was ungreatful because I was...so I am glad that this year it was much less involved and only my family.  Quarter plus two, two more than I thought, five years to the median, waiting for lungs, and average wait of two years. I have about 40% of my lungs left at a burn rate of 20% this last year...and it increases as time goes on. I am in for a rough time but I can handle it whatever it is. Not because I am that good...but because I just don\'t ever do anything like I am supposed too...so I will trudge on and blow the bell.  Peace  '),(1312,1234212,'07,April,2003','2003-04-07 00:00:00','   A quarter plus 2. 2 more than I thought.  '),(1313,1234212,'07,April,2003','2003-04-07 00:00:00','   Well my doggie had his balls removed today finally. My sister took him down to the place where she did an externship in Cheswick and did the deed. He had one dropped and one in his abdomen. It was a hard surgery for Rufus because they have to cut him open and dig for the testicle. It is ok though he has pain meds. She also took Xrays of his hips to see how bad they are...and they are pretty bad. She figures about a year at most before something has to be done, but it could be as short as 6 months. So I will have him at most another year and that makes me sad cause I love my dog. We could replace the hips but it is a huge surgery and is a lot of pain for the dog and I cannot put him through it. I wouldn\'t want it I would rather be put down.  It sounds ruthless but I see it as humane. I may cry I may not but I know I will have done the right thing because it will relieve his suffering.  Well I am staying up with him because he is pretty gimpy and because of the invasiveness of this surgery (I couldn\'t imagine hip surgery) he is pretty gimpy and I just want to keep an eye on him cause he is my baby.  Well off to get him his pain med for the night.  Peace  '),(1314,1234212,'07,April,2003','2003-04-07 00:00:00','   Well today I had a lung transplant support/informational meeting. It was really depressing for me to go to.  First off I am glad I took my notebook because the computer the docs had there to use wouldn\'t recognize the USB flash drive the speaker had...it was an older Dell model with Windows 98 on it so it would\'t run without the proper drivers. So the laptop came in handy.  The information I recieved was really a bummer. First out of the thousand or so patients that came in for evaluation for transplant only about 48% percent were accepted. I am through that phase.  There are a bunch of other numbers as well but it boils down to that only 13% of the original thousand make it through transplant.  CF patients have about a 48% transplant rate in all. If I had a different, less virilant lung disease I would have a much greater chance of making it through it all.   It is just a little depressing. I know that I have as good a chance as anyone but the numbers are daunting and frightening.  Well tonights discourse is about Social Security and the push by people on both sides of the isle, as well as many banker lobbiests\', to privatize Social Security and its benefits.  Now the rational is that the private market can make more money for the retirees in the stock market or even the most conservative of money market accounts. Yes it is true that if a person were to take the equal amount of money that Social Security witholds from the paycheck and put into a money market or riskier stock fund that person may make more money than what the returns would be on just the straight payment to the SS system. However there are two problems to that whole way of looking at it.  First is that Social Security was never intended to be a retirees sole income but it has become a safety net for many many people. The deposits that many workers make into the system really have no value. Sure you can go look up potential benefits that you may get it really isn\'t there. You cannot reliably calculate annual percentage returns on Social Security because of how it works. If when you retire you have substantial savings or holdings of value your SS benefit will be less, the less you have the greater your benefit will be. It has very little to do with what you put into the system. However you are guaranteed to get out of the system more than what you put it. Who know the total \"percentage return\" you will see but it is nice to know that you will have something.  The second big problem even larger than the first is that if you invest money you had better be prepared to loose it. Investing is almost as risky as gambling. In a casino if you have a large bank roll and play blackjack you have about an equal chance of walking out with at least as much money as you walk in with. Playing the stock market is like taking 100 bucks and hitting the poker tables. Sure you may win a lot of money, but more than likely you will loose money. The stock market is for the rich. If you have tons of money you can diversify and hedge your bets (investments) and you will always be ahead. If you are a small player then you will probably break even in the end, but the risk is great that you could loose it all. Look at the stock market bubble the burst in 01\'. How many people lost everything. Tons...if it weren\'t for SS for some of those people who did all the right things (saved, invested, and are frugal in retirement) lost their savings and now rely on SS just to buy food.  Now the bankers and investment firms would love to get a hold of the literally trillions of dollars worth of SS money. It would be trillions because they would have to transfer people who have been paying for years but aren\'t retired to these new accounts, plus all of the people currently paying in and new workers each year. For the government to privatize it all would cost the taxpayers trillions.  Now for the real reason why they want to privatize the whole Social Security system. Or at least this is what I see. All of the ultra rich have vast holdings. Much of it in the stock markets. This money is tied up, if they were to start selling off these vast holdings it would drive the stock prices down, hence driving what they can get out of the stocks down, thus leaving them less wealthy then they were before. Laws of supply and demand work here. Their very own motto and creed kicks in free market economics. If there is a huge supply and little demand the prices are low, if there is a large demand and small supply the prices remain stable or rise. Over the past years the stock market hasn\'t performed as well as many would like so some legitimately want to get out and save their money, but imagine what would happen if say Bill Gates said that he was going to sell all of his Microsoft shares off. The price of Microsoft shares would drop, not because the company is unsound or bad, but because the market will be flooded with openly available shares of Microsoft (think about it now as a bidding war...20 million shares available we\'ll say, 18 million owned, 2 million floating around, and someone outside of the 2 million wants to sell 1 million shares...price drops slightly).  Now what will all these people do with the money they would be able to take out of the stock market if public money (social security savings accounts as Shrub would say) became available. They would move to good ole\' commodities. The rich are both smart and dumb. In the near future there will be famine and drought (we are seeing it now in its begining phase) and if you own commodities like wheat or corn not only does your wealth grow but you also own so much of this or that. With that you can go and get it physically to feed your family. You can eat wheat and corn, you can\'t eat stocks.  Another problem is that these \"savings accounts\" won\'t be insured so if someone screws up and your money vanishes (as it tends to do on wall street) you have no recourse. Since there will no longer be Social Security as we know it you don\'t get anything. It is a play to get the poor mans money and give it to the rich man. If they get this infusion of cash many wealthy people will cash out of the stock market and many others will simply steal the money as we have seen at Enron or Tyco, and many others.  The \"experts\" say that we need to do it now. The system is broken! It will crumble tomorrow. Well even they admit the system will be able to pay out until 2042 at current growth rates and that Social Security will remain solvent until then. With minor tweaks it will work until 207?. The main tweak is that people who make over 88 thousand a year pay more in SS tax. That is the big thing about SS right now a person who make 79,999 dollars a year pays the same as a person who make 4 million a year. So as you go lower on the pay scale the percentage of income paid to the system goes up, and the more you are paid the lower the percentage you pay into the system. Bill Gates pays that same amount of Social Security Tax as a person who make 88 thousand a year, yet he is paid far more money. The tweaks would extend the number upwards a little. What is unfair is that Bill Gates will be able to draw Social Security of some amount and will get something similar to my grandmother per month yet he hold vastly more in property and stocks than she does...but he will still get the same money. She has paid more of her wages in percentage over her lifetime than Bill has percentage wise. The Social Security system is really the poor helping the poor and the rich riding the poor for more cash as well. Social Security is run on the backs of the poor...but to privatize it would mean its demise and you will only be able to be able to afford Social Security because it will no longer be an entitlement. You will have to be able to afford to retire and afford to be part of the Social Security system if it changes.  So when politico\'s come before you and say \"the social security system is broken! it costs to much, we need to raise the retirement age, and if we don\'t privatize it the system will fail\". Just reply \"BULLSHIT!\" The Social Security Administration is the most efficient according to Jim Hightower with only 1 percent overhead. There is no investment house or bank that has lower overhead. So don\'t kill a system that works already...and will continue to work for some time.  Think...plan...act.  Fight the power, buck the trend, be smart and confuse the leaders, use their words and studies to make them look like the fools they are.  Peace  '),(1315,1234212,'07,April,2003','2003-04-07 00:00:00','   I want to appologize to some people out there for not having kept in touch lately. It is not that I am being totaly anti-social it is because I just don\'t feel much like talking lately mainly because I am constantly bothered by a group of head people over this pending transplant. By the time I get done with dealing with them all I am all talked out and just ready to collapse.  That is also another point. I don\'t feel much like talking because I am worn out and I get short of breath when I talk to long (like more than 5 minutes...yes it is getting that bad). Unless I get fired up about something and then I can talk for a long time and after I finish I pass out for a few hours. I sleep constantly and I am still pretty tired all the time. People tell me to sleep less and I will have more energy. Well I tried that for a week (slept 8 hours a night) and by the end of the week I was sicker than when I started the week and I ended up having to catch up on more sleep...causing the vicious cycle to go round again.   I am not ignoring people...I am just not talking to anyone. Not because I am don\'t want too but because in many cases I can\'t.   Later this morning I have a \"support group\" for lung transplant people. I don\'t want to go but I need too so that I can meet more of the transplant team.  So until later...peace.  And by the way the best way to get me is not on AIM but email...I am on AIM only occasionally...rather around my computer when messages come in. Email I check five times a day religiously and that is the best way to get me. For some people I don\'t have your email addies so email me and I can get back at ya!  Night all  '),(1316,1234212,'07,April,2003','2003-04-07 00:00:00','   I changed my email to my gmail account...so I will get mail there.  Peace  '),(1317,1234212,'07,April,2003','2003-04-07 00:00:00','   This new interface is really K-Rad.  Well I said that I had more to write about and really it is more of a rant about health care and the military.  If you add for the past century in this nation alone you would probably come up with 100 trillion in todays dollars that have been spent on researching and aquiring weapons. We have literally figured out a thousand ways to Sunday to kill people. Now if you add up the money spent on health care in the public sector (remember all military research and spending is on public money ultimately...Boeing develops a new missile and sell it to the US all of the costs are passed on to the Gov.) anyway you add up all the public money spent on research and implementing these life saving ideas it would probably be less than a tenth of the money spent on the military. To me that is uterly insane.  We spend all this money on killing each other and we are right in the middle of a health care crisis. We have superbugs being made each day through the use (and overuse) or antibiotics. Instead of investing in new meds for healing our pharmicuticals are spending tens or hundreds of millions on penis pills. Don\'t get me wrong I love my dick...but when it comes down to it would I rather have a better antibiotic or stiffy...I\'ll take the antibiotic.   Sure money is being spent on good research but it is not enough. Public money has all but stopped to help develop drugs. All of the current life saving drugs were developed years ago on the public dime in Universities then turned over to private companies in the hopes that they would not only develop the current class of drugs but also future drugs at the same pace. Well the pace has fallen off. We have hundreds of new drugs a year however many of them are not life saving drugs more like life enhancing drugs.  Meanwhile the military needs more armor, needs more bullets. I am for a strong defensive military, but what have is not a defensive military. It is an offensive army. It can\'t defend anything, it can only attack. We have seen it in Iraq. The military was great at the war part and invasion because they have the tools to do that, but in the peace and defense of Iraq it has failed uterly. Canada has a defensive military. Sure they can attack but the majority of their military is defensive in nature. Our military is never in the country to defend. It is always elsewhere. We need a strong defensive military but not this monster we have now. A B2 Stealth bomber is not a defensive piece of hardware, it is for attacking nations secretly. Nuclear weapons may have had some defensive position at one time in a MAD (mutually assured destruction) scenario but we can\'t use them now. If we were invaded by Mexico its not like we can drop a few on Brownsville TX to stem the invasion. They are for attacking another nation. They serve no defensive purpose whatsoever. If you look and take into account the size and configuration of our military it is very easy to come to the conclusion that it is only for attacking and empire building...not defense. Department of Defense my ass...the old title would be much more appropriate The War Department. That is what it is simple. A Department of Defense would never keep a military as large and aggressive as what we have, a War Department would.  That brings me to my next point. We are failing at all the things people are saying we are doing. If we are bringing peace and democracy to other oppressed nations like Iraq as the Administration would have us believe then we are failing. If we are empire building as most of the rest of the world says we are doing then we are also failing because all of our \"colonies\" are in revolution already. They all want us out. Sure the Romans, British, French, Spannish, Germans all suffered from the same problems we now suffer in building our empire they at least had a few years of peace (more peace the further back you go in time...things happen faster now due to better communications).   So anyway we look at it we are failing on all fronts. So instead of wasting lives and money we might as well pull back and evaluate what is going on. The world will not fall into some evil hands if we leave it alone. Many would argue we are the evil hands because of what we do, how we do it, and then how we empose it on others. We have failed and we might as well face it.   Bring all of our troops from around the world home. De-activate 80% of the military and change its configuration to a defensive posture only. Concentrate on creating the technology to make our nation cleaner and better. Finally we need to spend money on figuring out and making people healthy. Healthy people are happy people.  That is what the rich have still not figured out after all these years. If you keep the people happy and healthy then they work harder and do not revolt. When will they figure that out. Who knows? But if they don\'t Americans will revolt and bring the nation back to where it is supposed to be.  Peace...  I have another one in me for another day...as always thanks for reading and if you have any suggestions email me and let me know.  '),(1318,1234212,'07,April,2003','2003-04-07 00:00:00','   Wow...blogger has changed how you enter a post. I likes!  Well today I was surfin the channels and came across a Chomsky speech from somewhere. It was recent, I want to say in March or April. He said exactly what I have been saying for months now to my family and I may have mentioned...actually I know I have mentioned here.  While the reason for invading Iraq seems as simple as getting our hands on Iraqi oil for our use it is a tad more complicated...and I firmly believe in what I am going to write here.  The real reason I feel is because most of the middle eastern fossil fuel reserves actually go to Europe or rather the EU. There are three large players in the world right now and frankly they will probably be that way for some time. I am talking mostly economically but also militarily.   You have of coarse the US and North America (USA, Canada, Mexico) and together we are a large economic block and with the USA in this block we also happen to be the strongest militarily at this point. The second power is most of Asia or the Asian nations. China, North/South Korea, Japan, and then Sub-Asia (Cambodia, Vietnam, Malaysia, and many others). The third region is the EU. Both the EU and the Asian nations combined are much larger than we are and combined they are militarily stronger than North America (or as Michael Moore would put it \"The Big One\").  Asia is pretty much self contained. They have vast resources at their fingertips and they keep tight control over it all. They have huge populations that can afford armies of tens of millions. They also have something else we often forget. They have technology that is on par with our own because most of our technology is built there. We invent it (not so true anymore) and they build it. If they ever decided to exersize their power we would be helpless unless we wanted to destroy the world with Nukes. Also China owns huge chunks of our national debt. When our government issue bonds it is like a loan from someone (an investor) to the nation, at the end of the term the nation then has to repay the loan plus interest (however many people roll their bonds into new ones). The largest buyer lately has been China. They are able to do this because of the huge trade balance deficit we run with them. With the billions in difference that we see now that money goes into buying US bonds. So essentially China could call the debt tomorrow and we would probably not be able to pay it. Asia again has vast resources there. So they could isolate themselves and be fine because most of their materials they need are already there.  This leaves the EU, particularly France and Germany. They both have the largest economies in the EU. Brittian is third and not a full member of the EU (they still haven\'t gotten rid of their pound for the Euro). However the EU must import most of its energy resources and they come out of the mideast and have for some time. Whoever holds the mideast holds their economies and can crush them if they were inclined to do so.  Sure we need oil, but we still get most of it from here or South America (hell we even import from Canada!). Only some of it comes from the mid-east.  With our hands on the tap we can turn it off anytime we want. That is also why we are in most of the Caspian Sea nations to control the natural resources. Not so much to bring here but to keep others from reaching full economic and military power, or rather to become a threat to the US.  Chomsky of course put it better...but I am not as learned as he is, but these ideas truly have been bouncing around in my head after I saw a map somewhere of where we are in the world militarily. It so happends where we are all happens to coincide with where oil or gas or coal deposits are.  When I told my grandmother about Chomsky\'s speech she said \"hey you have been going on and on about that for a long time\". My reply was well at least I am not the only one who sees this trend or patern and if Chomsky sees it as well then I may not be to far off base as to the thinking I aspire to do.  She also suggest that I write something about it all and submit it. I told her its not that I don\'t enjoy to read but I hate to research because I research by grabbing bits of info here and there and cataloging them in my head then stringing them together how a Capitalist would and then put a twist on it. The sad part is my predictions have been pretty close so far as to what is happening. I would write for real if I could ever remember to keep pen and pad handy so that I can get quotes or write down sources and notes. Again I am lazy about that.  Within a day or so I will write what I think of our \"defensive military\" and health. I will also talk about our noble goals of Democracy and spreading the love...and how we have failed at every angle you can look at it (including emperialism).  Well Peace...my sis will be here tomorrow and I have to clean a room for her to sleep in.  '),(1319,1234212,'07,April,2003','2003-04-07 00:00:00','   My sisters new laptop is nearly done and almost ready for her. It is finishing updating now downstairs close to the access point.  Goodnight...I\'m spent and my kidneys hurt worse now. Alas doctors are in bed and I get to suffer for a few more days. If it doesn\'t straighten itself out in a few days (I have drank almost a full bottle of cranberry juice to flush them out of any toxins possibly built up) then I will get a hold of the docs...but not until after Carlin on the 11th. Tuesday baby!!!! Then Thursday I turn a quarter century plus two. Longer than I thought or planned for.  I\'m out!  '),(1320,1234212,'07,April,2003','2003-04-07 00:00:00','   yeat another post...drunken speech is the best!  '),(1321,1234212,'07,April,2003','2003-04-07 00:00:00','   urlLink'),(1322,1234212,'07,April,2003','2003-04-07 00:00:00','   I now have audio blogging as well...and I will say my voice sounds kind of girlie...I hate that. I always get on the phone \"yes ma\'am\" and shit like that.  I would pay attention to it from time to time because I talk to myself a lot and now that I can record it for posterity sakes I will be recording a lot I figure...so tune in from time to time.   I tend to bounce \"crazy\" ideas off of my brain and see how it goes. It may interest you in how I think.  I would like to thank Blogger.com for all these great services. I not only get a free blog, but also the audio blogging now, and also I got my 1 gig gmail account because of being a longer standing blogger person.  Of you don\'t have a blog get one...really it is cathardic in many ways. Gmail may seem harmful because they scan email but really they do it robotically and no person reads it...and frankly I have nothing to hide so scan the email bitches!   My life is an open book...and I am sure that the audio blog portion will come in handy when I spend time in the hospital because it will be easier to do...  Remember! PEACE!  So much tonight...damn!  '),(1323,1234212,'07,April,2003','2003-04-07 00:00:00','   urlLink'),(1324,1234212,'07,April,2003','2003-04-07 00:00:00','   I have to figure out a few things....I have to figure out the links portion and get the counter back up and running.  I will add about 14 hits to it considering I have been getting about 20 a day if I get it done later today I will be close on my hits.  These new layouts don\'t have places for links but I will make one ;-)  '),(1325,1234212,'07,April,2003','2003-04-07 00:00:00','   Changed the layout...now I have to re-add all the links.  It gives me a change to update and get a fresh start on links.  If you would like a link email me at michael.honse@gmail.com  '),(1326,1234212,'07,April,2003','2003-04-07 00:00:00','   Oh...it is getting ready to storm again.  I love a little thunder and lightning and rain...lets me sleep well.  '),(1327,1234212,'07,April,2003','2003-04-07 00:00:00','   I just realized that I spelled wine as whine, as in \"I want some whine\" or \"you are a whiney bitch\", not \"wow that bottle of wine was good\".  Damn this spelling thing!  '),(1328,1234212,'07,April,2003','2003-04-07 00:00:00','   Well today has been a long ass day. I spent most of it seeing my sister off, visiting, and then just spending time out of the house.   I got my license the other day so now I can go drinking which is a good thing. I went and got a couple bottles of whine and a bottle of Absolute Vodka...I have drank about half a bottle of red whine and half a bottle of white, now I will get a little vodka in me.  Tomorrow hopefully will be easy. I am going to try and catch up with some friends. I am definitely going out to a bar tomorrow night.  I figure what the hell...I need it. I need to get out and have a little fun. Might as well because who knows how long much longer good time I have left, might as well spend it having some fun.  Heard something interesting today. It seems the Justice Scalia\'s son was a lawyer for Bush when he sued to stop the vote recount in Florida so he should have recused himself. Also Justice Thomas\'s wife was a consultant for the Bush campaign that was gathering information for staff positions and appointments for Bush and his administration while the whole recount thing was going on, he should have recused himself as well. With those two justices recused the count would have been 4-3 in favor of letting the recount go on and that would have given Gore the victory.  It sickens me worse each day as I learn about new things that this administration has done and is looking to do. Each day I become a little more convinced that anyone willing to vote for Bush again has a serious mental disorder.  Peace  '),(1329,1234212,'07,April,2003','2003-04-07 00:00:00','   OH this is priceless! This only further proves that the Republicans are no friend to the working stiff.   Republicans offshore even their fundraising operations!  Those fuckers...can\'t even raise money for their evil plans using Americans.  Damn.  '),(1330,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I made it through today with no fever until a bit ago. It is mild and not bad.  I got my birth certificate and MS Department of Licensing and Public safety drivers record clearance so I was able to go get my PA drivers license...even with my wild hair! Something bit me on the nose the other day and the end of it is all scabbed up on one side and I was thinking \"oh great I am gonna look like shit on my photo\"...well the picture is a great one, one of the better photos I have ever had of myself anywhere. Usually license photos are horrible...but this one is good. I had my head slightly turned and you can\'t really see the huge chunk of nose missing. I am pleased with it, and now that I have a valid ID (I had a valid MS drivers license but I didn\'t physically have it...or have it for almost two years!) I went to the liqour store and got a couple of bottles of whine for tonights dinner with grandpa. I picked good bottles too!  So now I can go drink without a hassle!!! w00t!  Well I am going to watch Aqua Teen Hungerforce and then hit the sack..  peace  '),(1331,1234212,'07,April,2003','2003-04-07 00:00:00','   So I was ok today then I get home and low and behold I now have a fever of 102...pretty damn high.  So I am going to pass out and sleep for as long as I can. Maybe whatever this is will pass.  '),(1332,1234212,'07,April,2003','2003-04-07 00:00:00','   Doctor appointment tomorrow at 2.  Lost of stuff floating around in my head. I will get it all formed into something semi-coherant sooner or later.  peace  '),(1333,1234212,'07,April,2003','2003-04-07 00:00:00','   Nothing to report today outside of that a shell was exploded over in Iraq that seems to have some sort of Nerve Gas in it. The gas was old and fairly weak it would seem as none of our soldiers has been reported to be having serious effects from it.  WMD\'s seem at the moment to have been found in Iraq, this must make Bush pleased. I feel however that in a good analysis of the toxin they will find that it is old and was likely to have been produced pre-1990 as the potency was very low.  I was hoping that no WMDs would be found but now that one has been discovered I hope that no more are detonated for the sake of our troops and for the sake of the civilians because if a truely potent WMD was detonated then there could be many casualties. That is the last thing our troops need to deal with on top of explosives. It will now be infinitely harder for our troops and much more hazardess because of this discovery. It will raise the \"terror\" factor with our troops because now they have to be worried about being gassed.  As I have stated previously on this blog and in real life I still feel that any weapons that may be uncovered will most likely be old weapons with little potency. I do not believe that they had active and productive programs. If they did then someone was selling it to them, and that someone will definitely lead back to us when it washes out.  If we find many weapons and evidence that a full fledged program was in effect that was producing weapons then I will eat crow and admit that they had it. However I want to make sure that we are safe and that the zeal to find weapons and programs will not overshadow reality.  Regardless of the outcome on WMDs the war was still not justified as there are no apparent ties to terrorism or orginizations that have targeted us. Again if we can prove it beyond a reasonable doubt then I will eat crow. I will however diffuse one possible piece of \"proof\" in the training camp that was known to be in Northern Iraq. We knew it was there for a long time and because of where it was located it fell under our protection in the no fly zones so Saddam could not have actually destroyed the camp...even if he had marched some military up there we would have bombed it. Also the fact remains that the military had requested several times to destroy it and three seperate times to kill Al-Zarqowi yet the approval was never given by the Bush Administration (which leads me to believe even more so than ever that he is a CIA asset or that they need a boogy man in Iraq).  Goodnight...and I send my thoughts out to our troops. I hope that we bring you home soon because you don\'t deserve to be over there fighting for economic and corporate interests that have little to do with national security. There are people here fighting to bring you home safely and soon.  Peace  '),(1334,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok todays update will be short and sweet. I hope.  My sis, mum, and I went to see Troy. Go see it. Not a bad epic. Pitt was very good in it and Peter O\'toole was excellent and may see a nod for it. Pitt may also see a nod but it will be on the back of one scene between him and O\'toole. It is a big movie and a long one, and a much better movie then Van Helsing (although I love Kate Bekinsale...I might bear a child for her...hehehe).  I caught my Insult-A-Day calender up today and read it for the days I missed which included my birthday and it is like a prophecy for me. It makes me even more determined to become one of those people who speak out and spread the truth as best I can.   Here was the insult.  \"A toothache of a man.\" - Radio Host (and former Secretary of Agriculture in Texas? I think) Jim Hightower on George H.W. Bush  My calling may have come to me from an insult a day calender. Either it is destiny or pathetic. Either way I\'m goin\' big or goin\' home.   Night all.  '),(1335,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I went to see Carlin and damn was it good. I took the PDA but it just wasn\'t having anything close to recording all of the set.  I just want to say that is really was worth all the walking up and down steps that I had to do. He had great new material and starts the show out with a bang.  I am pretty tired from it all and I am still processing much of it...but this tour has less political satire than I was expecting. Seeing that there is so much cannon fodder right now I would think that he would have a lot of political stuff going on. This isn\'t the case but he does put Bush down twice and the second time it is \"he lays his big fat empty head on his little pillow\". There is a story behind it but it really is pretty funny.  Just got this in...take it with a grain of salt as with everything on the interweb...but the AP picked up the story it seems.  Rummy may have authorized the torture after all with the presidents knowledge.   Well that hit me up pretty good.  It is sad when we as Americans think and act like we have the moral high ground yet we are the ones killing and torturing on a scale that far exceeds what the so called \"animals\" in Iraq are and have done to us. Falujah...I am sure we killed not only a few true targets that were shooting at our men, but we killed hundreds if not thousands of women and children. The argument that has been made is that they had the chance to turn in the insurgents or leave the city. Well first many weren\'t allowed to go because their husbands couldn\'t leave...and last time I checked it is still a male dominated society over there so women can\'t really travel without men. Ok so back to the horror of it all. Second did many of our founding people turn in Washington, Jefferson, Franklin and others? Nope...why? Because they were revolutionaries. They were the insurgents of their day. No matter what we do or how we do it we will be kicked out of Iraq. No nation or empire on the face of the planet has been able to hold the middle east. Every nation or empire that tried was driven out by the local peoples, including the local Jewish people. (A dirty little secret is that before Israel was a state the Jews and Arabs that inhabited Palestine lived in relative peace outside of the occasional squable over goats drinking from the wrong part of the stream or eating in the wrong field. Only after Israel became a state and started expelling the local Arabs or Palestinians is when all hell started to break loose. The other Arab nations didn\'t help matters much because they propugated the myth that Jewish settlers would wholesale slaughter palestinians and that they should run or kill every jew they saw. Well that led to more violence. Also since the time of Muhammed the keys to the church or temple or mosque in Jerusilem, a religious site for all middle eastern religions including Judaism, Christianity, and Islam, the keys were entrusted to a Muslim family that there sole purpose was to make sure that the site remained open to all faiths and that the site was clean and protected. Muhammed foresight actually protected the site from the invading armies of Europe during the crusades because they were burning villages that wouldn\'t convert to catholisism. So they burned jew and muslim alike. Both groups fought side by side albeit contentiously but fought the European crusaders. The family to my knowledge still has the keys and still is in charge...it may have changed after Israel took Gaza and the West Bank which includes Jerusilem. But it was the Muslims who protected and kept up the most sacred sights for Jews and Christians alike because Islam was and still has hope of being a religion of peace.)  Again we cannot win in Iraq. We have already lost. The thing that we must do now is find a way to retreat and let the most stable power come to power without looking like we placed him in power. Sistani may be the answer...let him \"beat\" the United States and feel his oats. That is fine. As long as he promises not to wholesale slaughter other factions and have some sort of representative government as part of our \"surrender\" or \"cease fire\" treaty. The only thing we can do is pull out strategically and hope that the region settles into some sort of stable peace.   So we must pull out...otherwise we risk the attitude that started breeding (and it is starting even earlier this time) like that of Vietnam were we supposedly can\'t loose or the region will fall apart and become this huge communist domino and more nations will fall after that. Well it fell yes...but not to many fell after it. Now the region is seeing modest but steady economic gains and the people aren\'t attempting to export anything but local goods. For the the communist system has helped move them out of tribal and clan like systems that even the French were unable to do. It may have been through ruthless re-education but that doesn\'t concern us and doesn\'t affect me at all. So why should we have been concerned.  Iraq is a strategic asset, but only if we can completely control it and we can\'t. If we continue and fight harder and harsher then it will galvenize the Arab nations against us. They hate us now, but soon they will despise us and when that happens you will see people from Saudi Arabia, Syria, Jordan, Iran, and Iraq come together and fight us as they have always done with past invaders. It won\'t be radicals that will be fighting us...as the powers that be will say. It will be average people who just want us out. It is already happening in Iraq and soon it will spread. That is what the politicos should be worried that the domino effect will be cause by our continued presence instead of our leaving the region both militarily but also economically. When we leave the region will have some upheaval as power bases shift but it will settle out quickly. If we don\'t leave soon we will end up fighting every Arab nation and incurring tens of thousands of deaths on our side militarily and increase the risk for civilian deaths here at home in terrorism, not to mention the millions that we will kill in retribution.  We can save a lot of money and a lot of blood if we admit failure and pull out now. Pulling out is a great American tradition, only this time it is sure to work because we will remove the substance that is giving the enemy all of its gas to fuel the fires of hatred.  Peace  '),(1336,1234212,'07,April,2003','2003-04-07 00:00:00','   urlLink'),(1337,1234212,'07,April,2003','2003-04-07 00:00:00','   Played again tonight...although so far a lot less eventfull.  I was mistaken in my thoughts the other night that I had banged most of the good looking women here in PA. I found a few tonight that matched well.  I have this problem...it is called \"no game\". I am a severely stricken person. When even the worst are spittin\' there best I am still only at \"shutes and ladders\" (mispelled on purpose that time). I really have no game. I would imagine that it would be like a young Noam Chomsky (my hero) trying to pick up women. It just ain\'t happenin\' outside of a classroom setting with me being the teacher (and then it might be a felony).   It takes a few sittings for me to warm up to people. I have just this impulse of that I hate people but I am amused by them and therefor need to be around them (plus when you drink alone you are an alcoholic...as I sip a glass of fine wine...really I am!).   I need a good looking woman with some brains to keep me even remotely interested. For me at this point it is hard to find (and chances are I won\'t find her in a bar).   I want a relationship but I don\'t. I can\'t, from my altruistic and rational side, see putting anyone through all the shit that will happen over the next months or years. On the other hand I would love to have someone to lean on from time to time when I am feeling blue (and of course the sex!).  I am watching Fight Club and Cloe is saying she wants to get laid for the last time...that is how I feel right about now.  I dunno what I want but I know I will probably not find it. I want closeness without the pain that goes with it. I want to be close to a woman for all the carnal reasons without the pain that will come with my passing. I know that I am going to die, sooner rather than later, but I still want someone there, yet I cannot bring myself to put anyone through that.  I am on this CF mailing list and many of them are in relationships and I cannot understand why for the life of me. I cannot understand why someone would want to put another completely unrelated person through the pain of this disgusting disease.  I need a \"Marla\" in my life...someone who doesn\'t give a fuck, a person who has realized the futility of it all, and yet has enough panache to live life (I used panache)...I know that person is out there...but I don\'t know if I will find her in time.  Peace  '),(1338,1234212,'07,April,2003','2003-04-07 00:00:00','   I wanted to say oops...I didn\'t update last night...there is a reason, Goldschlager, beer, and loud music all played a role in causing me to pass the fuck out and at points see parts of my stomach and bowel expelled into a trash can (thank god for plastics!).  I will be paying for a bit but that is ok...I needed it.  I will update with something a little more interesting later today...after I get some sleep, clean my car (it is truly filthy!!!), and get some more alcohol in me because it is Memorial Day Weekend baby!!!  My deepest and greatest admiration and respect goes out to all of the men and women who have served this nation with honor and respect, and to all those who have died in the name of freedom on whatever battlefield. I salute you and I hope for a swift return home for the troops that are now deployed.  Peace  '),(1339,1234212,'07,April,2003','2003-04-07 00:00:00','   I feel rough. I feel rough. I feel rough.  I feel like I have but limited time before I need a tune up. That sucks.  Peace  '),(1340,1234212,'07,April,2003','2003-04-07 00:00:00','   urlLink'),(1341,1234212,'07,April,2003','2003-04-07 00:00:00','   So I was going to wait and let things settle in my head and then put them to words but I couldn\'t wait any longer.  Tonight my mother and I were talking over dinner and she said that I was rude in my defence of Arabs and intimated that I was mislead in my sympathies to their ways and way of life. She said that I was the same as Nick Berg who went there obstensively to bring new thought and that it would be my undoing as it was his.  I am not a Nick Berg. Actaully I feel that Arabs are barbarians much of the time and their capacity for barbaric actions is almost limitless. Many of their actions shock me and disgust me in many ways, but at the same time the actions that westerners and especially we Americans have done are just as shocking if not more so. Let me explain.  We go places to bring peace and unity and a new way of life under the banner of democracy. The problem is we rarely practice it here as it was meant to be practiced by our founders. Had they been able to foresee what we would do in the name of our way of life they would have most likely been opposed to it and probably would have written into our constitution and civil liberty documents specific measures to keep us from doing what is blatently obvious today.  We went to Iraq under the pretence that Iraqi WMD\'s were an iminant threat to the US as well as suposed ties between Saddam and true terrorists that target us. A WMD has been found and there are sure to be others as well, however it is likely that these weapons are old and not as big a threat as they once were (being less potent), and they were never a threat to the United States because of sheer distance. Saddam held malice towards us as most Arab states do but not because they just hate us for hates sake, they hate us because of actions we have taken or actions we have sponsored in the region.  The fact is that we have been medling in middle eastern politics for over fifty years with little success. We put into place horrible regimes that killed many people and we looked the other way as long as those regimes remained friendly to us, as soon as they turned on us all of a sudden they were a problem, but not before. Iraq, Iran both prime examples of shit we did to piss off the local populations.  The big difference between them and us is that they accept death and welcome it in many ways because the afterlife is sure to be better than the current one. Suicide is a time honored tradition for them (or so it would seem) and what they fight for is perfectly legitimate, their methods are not though.  I rail against the notion or idea that they just hate us and hate our way of life. Sure they see us as decadent, but so do our own fundamentalists here in this nation.  Islam is no different from Christianity or Judaism in that all three religions require conversion or death, but also all three religions have a loving god figure.  While many would love for this to be about religion (and for many on all sides it is) this is simply a war of failed policy on all sides.  The scary part is that we intiated the failed policies and we are still doing them today in other countries. The more we fail the harder it becomes on us.  We cannot underestimate how bad things are going to get, not from just Arabs, but also South Americans and much of the rest of the world. All of the worlds anger seems to be aimed at the United States. Not so much the people but the government that has done so many misdeeds in our name.  Every tax dollar that is used to buy weapons, support dictators, and send death overseas will be revisited on us in ways that will literally tear us apart.  To get out of this mess we have gotten ourselves into over the last century will not be easy, and we would be niave to think that even if we changed tomorrow that terrist acts would stop here. Instead we will see more of them and they would taper off over time. We can fix it but it does not stop the orphans that were made by US bombs, it does not ease the grief stricken mothers and fathers of lands that have been devistated by our policies. The thing we can hope for and change is that we no longer create more of these people.  '),(1342,1234212,'07,April,2003','2003-04-07 00:00:00','   urlLink'),(1343,1234212,'07,April,2003','2003-04-07 00:00:00','   Nothing to report today...I slept most of it away  '),(1344,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I did more damage to the body tonight. Went and sat in a smokey cellar and listened to deafeningly loud music while sipping some hooch. Oh was it good. I have enjoyed going out and having a bit of fun from time to time. Unfortunately this may have been the last night I can do it in a while unless whatever is brewing starts to behave again.  Not much happened today but I have to give a big \"ups\" to the Pope. I am not a big fan of Catholisism but I do actually admire this pope because he has remained pretty steady throughout his time as pope. I have to give it to him for admonishing Bush and this war. When the pope is against you...damn.  A good bit that is coming out of the Catholic church lately is that they seem to be cleaning house out in Boston and here in Pittsburgh they are seeking people who were molested so that they can remove priests and take the appropriate actions. While many (and a younger me) would call for excommunication I do not feel that would be appropriate simply because the church and the teachings bear out forgiveness in God\'s eye...but it does not mean that I approve or think it is right and that the church should do everything short of that including seeking legal reperations against these rogue priests. Again I do not think kindly of the \"fondling\" priests I feel that the church and people should forgive if they truly believe the teachings of Christ. Remember if you subscribe to the teachings these men will have to answer for what they have done.  That is all for tonight....  Oh by the way there were a couple of females there that were smokin\'. Damn I wish I had game.  Peace  '),(1345,1234212,'07,April,2003','2003-04-07 00:00:00','   urlLink    Hospital tray art...a very time consuming thing :-)&nbsp; urlLink  '),(1346,1234212,'07,April,2003','2003-04-07 00:00:00','   urlLink    Something nice from the hospital...taken a while back. Gotta love Sterile water and a nice flower.&nbsp; urlLink  '),(1347,1234212,'07,April,2003','2003-04-07 00:00:00','   Well tomorrow I will going to see my friend Danielle it is her birthday. I am very excited about this because I have been wondering how she is doing and it seems that she is doing well.  Another CIA dude quit today...although this resignation has been planned. I think that Tennet\'s was planned for a while because they need a fall guy and he makes a good one.  I feel kinda crappy because I have something big brewing in my lungs and I know it for sure. First all of the stuff coming out is green...and I mean green. The second is that I have hives a lot which means something is going out of balance somewhere. The third reason is because I am sleeping a lot lately. I mean a lot because again something is out of whack.  Well my grandmothers student the ever beautiful Cara. She is back for the summer and let me tell you that if I were a little younger and in better shape that I would go chasing her through the weeds so fast.  Peace.  '),(1348,1234212,'07,April,2003','2003-04-07 00:00:00','   Interesting little piece of writing.  '),(1349,1234212,'07,April,2003','2003-04-07 00:00:00','   My friend Danielle that had her transplant in January is having a birthday this weekend. Hmmm....yay!  Well today on the political scene things keep on getting worse and falling apart. It is fun to watch.  Peace  '),(1350,1234212,'07,April,2003','2003-04-07 00:00:00','   I hate getting up late...because it puts me behind the mark so many times on news of the day.   Blaaah!  '),(1351,1234212,'07,April,2003','2003-04-07 00:00:00','   CIA Tenent resigns.   I am happy that it has happened but this is an obvious falling on the sword. It doesn\'t get shrub out of the crossfire.   The snake Chalabi said he was glad Tenent was gone and that Tenent was the source of the misleading information also that Tenent has it in for him. Well Chalabi...you sir are retarded. Many of us remember that you and your I.N.C. were the source for much of the faulty information and it doesn\'t change the fact that you are a thief.   It looks like these dudes are sitting in a barrel of gasoline and ole Bush is lighting a match.  hehehehehehe...  Peace  '),(1352,1234212,'07,April,2003','2003-04-07 00:00:00','   Interesting tale of Bush Sr\'s bad luck in a plane during WW2  crashing and letting his flyboys die .  There is always another account...and we know the Bushs\' lie a lot...so I will give the benefit of the doubt. Even CNN didn\'t pick this one up.  Peace  '),(1353,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok I finally found Howard Stern\'s site. I looked a while back and couldn\'t find it. Now I have.  Regardless of if you think he is a mysonganist pig you have to defend even his right to free speech. I will say however that I love to listen to him because he has no censor in his head...he is probably the most honest radio host on the air because he has no agenda outside of shock and he is able to say what many think. Unfortunately he is under attack from all fronts and really it is all for nothing. Plus he hates Bush (not bush)...I am glad that we have him and I readily support him.  He may say obscene things from time to time but it is not the states job to keep things clean...it is up to the parents.  Go Howard! Get this...he is a conservative on many fronts as I am. He is a constitutionalist as I see myself, yet has a progressive attitude towards many things.  Remember he is pretty well off in the end...Bush\'s tax cuts help him yet he sees something wrong with it although he still says he likes the tax cut.  Peace  '),(1354,1234212,'07,April,2003','2003-04-07 00:00:00','   Well today the President was reported to be seeking outside council on the whole Valerie Plaime thing...the CIA agent that was exposed by someone in the administration. Bush could be called to testify and most likely he knows how did it. It was one of three people I am betting Cheney, Rove, or Bush himself.  We will see how that one shakes out. I must say CBS is really hitting the stride now. They also reported yesterday on the Enron tapes and again today more was revealed.   Hmmm what else. Oh yeah Bush denied having talk with Chalabi and said he didn\'t know him well although Chalabi sat behind his wife at this years State of The Union and was the handpicked leader at first for the CPA council. Like he thinks the rest of us are stupid.   It is just all very silly how they are trying to get out of all the shit they did. If it turns out that Chalabi was working for the Iranians as well and one of the higher ups spilled the intelligence bean on knowing how to decode Iranian intelligence than that may be high treason. Treason is punishable by death...so if these people are so pro-death penalty then Rove may have to see the gas chamber or see the needle...because we must remember that Rove is the nucleus of this warm little moist chewy administration and everything that happens inside the White House Rove knows about and probably had a lot to do with moving pieces around the board.  I hope that we nail these people and break the neo-con back for good...or at least for now.  Peace  '),(1355,1234212,'07,April,2003','2003-04-07 00:00:00','   Well things are falling apart in the Administration even faster now. While many have know what is going on and saying over and over again to seemingly deaf ears we may this time have caught them with their pants down.  We have to remember that these people aren\'t smatter than many of the people they rule. These are business people and business people are not intellectuals by trade. Engineers, scientists, historians, language professors are, but not business people.  Business people essentially are the ruling class they don\'t know how things work but they are supposed to know how to manage and sell things. This administration has been pretty good at selling to people here in America as many fell for the lies. Again though they are not bright people, but are smart enough to attempt to control the flow of information (which this administration has done well admitedly), sooner or later those in the lower tiers of power get their toes stepped on and that is when we have leaks and new information come out. That is what has happened here with this administration over and over again.  The administration has consistently tried to marginalize even Republican darlings like O\'Neile. He is not a good guy, he never claimed to be one, but he does realize that you can\'t just lie like they do and get away with it forever. Generals who have four stars have come out against many of the policies, not because they don\'t want or like war, but because the administration stepped on their toes. The administration attempted to manage the government without much knowledge of what it does or how it does it. When you try to manage without having enough knowledge of how to work a system you find bad managers at all levels that have been appointed, and the people they piss off won\'t stand by forever and watch their hard work go to waste regardless of what it is from great big bombs that go boom, to a frog specialist at the EPA who is fighting a company that wants to dump waste in an area that will kill all the frogs (I know of now direct case...it is just a point).  Sure the administration brought in some smart people (very few mind you) but they didn\'t listen to them and when those people weren\'t listened to they left and now they are talking about what they advised. Remember these are people they picked so if you choose to believe in the administration then you have to believe their appointees.  Instead of listening to these people they have instead chosen to block the flow of information in every way they can and when something does get out they attack it (all of the radical right) and say it is \"sour grapes\" or make it out like the person has a political agenda. Sure some do but most again were appointees that Bush and Cheney tapped for these jobs.  The information being leaked out in general is done by good honest hard working Americans within the government because what they see as wrong they believe the people have at least the right to know. Many have given up careers and put themselves into the fire for this and instead of praising them many Americans simply choose to ignore them just on the administrations word.  I have no doubt that much is politisized in these days, but it always has been from Kennedy to Bush 2. Another thing to remember is that these people who are in charge are the same group out of the Reagan administration who helped fun the Iran/Contra deals, killed tens of thousands or more in Central America by supporting dictators as bad or worse than Hussein, and indeed supported Hussein himself even know he was a menace. These men are liars of the worst kind because their lies get innocent people killed en-masse. Not only have we lost 800 troops of our own there is little doubt that about 10,000 Iraqi\'s have been also killed. These men lie to kill, kill to lie, and then hand the bill to the taxpayer and then say \"we are doing great\" all the while the Nation is falling apart because of budget shortfalls.  I have resisted the idea to Impeach Bush because while we know he lied about many things the lies can be manipulated to a point where it was misinformation. If this latest bit about Enron and the tapes leads to the administration then I say it is time we start the impeachment process because of their deeds and high crimes. Treason it is not but it is certainly worse than Whitewater or even Watergate.  Peace  '),(1356,1234212,'07,April,2003','2003-04-07 00:00:00','   Today CBS reported on some Enron tapes that show that there was market manipulation in California.   Linky link...   These tapes have been in the custody of the Department of Justice and were witheld from the public by both the Administration and the Department of Justice. I wonder who would such a thing. Maybe Bush, Cheney, Ashcroft and other oil men?  Well it also is going to get Arnold in trouble as well because it was reported a while ago and I put here last year that Arnold had made a deal with Enron reps to settle the lawsuits that California had launched to recover about 6-10 billion dollars from Enron because of price gouging. Arnold agreed that once he took office the lawsuits would be settled and prosecution would stop for a settlement of just mere millions. In exchange he got great support from the energy lobby and recieved money as well.  In 2000 Enron gave millions to Bush/Cheney/Republicans for the election all the while fleecing the California consumer. In return they got a President that witheld federal oversight into the market manipulations and also held the line and would not impose price caps of energy. In this climate Enron was able to do what it did.  The roosters are coming home and it looks like they are shitting on Bush\'s desk this time.  Ahh the sweet irony of being right :-)  Peace  '),(1357,1234212,'07,April,2003','2003-04-07 00:00:00','   Well I went to see my dad off and then to see \"The Day After Tomorrow\". I will say that it is indeed a pretty damn good movie. We will see however a trotting out of \"scientists\" that will say the science is hokey. Well taken as a whole it may seem a little hokey until you break it down like I did.  First the shelf of ice that breaks off. There is currently a very large chunk of Antarctica that has broken off or is in the process of breaking off and will soon be the largest iceburg that has been recorded in modern times. It is huge. Once that much ice melts into the ocean it is sure to cause a slight cooling because it will drift and melt. Once it melts the ocean will cool maybe a degree or so and also change the salinity of the ocean. Again these changes will be slight in human terms.  The ocean currents that I have spoken of before do indeed exist and we do have the knowledge of if you shift these currents you will cool the ocean and thus the planet also cooling the air temps.  Here is the big if question. Scientists don\'t know how much ocean cooling or salinity shift will cause these currents to shift or stop all together. But we do know something will happen but not the extent.  We know that changing paterns in the air flows and masses change our weather paterns. We know it because we can now see it pretty clearly. If the jetstream is changed drastically (or even slightly for all we know) we will see many problems. Tornadoes in Hollywood, blizzards in New Delhi. We just don\'t know. But we do know that the Gulfstream (ocean current) is linked with the Jetstream somehow.  Also we know for fact that melting ice raises water levels and we have seen the ocean slightly rise (and it is rising at ever increasing rates) over the past years.   Ok so now comes the hokey science. Or the part that can be passed of as hokey but hopefully I can convince a few people that it is probable. The Ice Age that starts in hours, days, or even a few months.  We know it is possible. It has happened in the past. We have wooly mamouths that were frozen in place with food in their mouths and stomachs as if they were eating and then flash frozen. It is possible and it has already happened. It happened about ten thousand (10,000) years ago. A blink of a geological eye. This is where the movie kind of makes up some science...well sort of.  The make the case that the shifting and eventual stopping of the Gulfstream effects the Jetstream while the water level is rising. As the Jetstream changes huge storms start to form similar to hurricanes only they form over the northern land masses toward the north pole. I think that the storms would be effected by the season in some way as to where they will form but that is my opinion. If the northern hemisphere is in fall winter and the southern is spring summer then the storms form north and vice versa.   These huge storms reach the troposphere where the air is very very cold. Those storms happen now but they are small and the air that is sucked down as it is in a hurricane is warmed as it reaches the surface by water and air because it is over the water. Over land there is only air to warm it. A small storm would cause cooling of a few degrees. A large storm over land would suck down to much air at a time and allow it to not warm in the surrounding air. It would be a -150F blast of air to the ground. That would flash freeze almost anything and anyone.  We know hurricanes do this and that is how they form and get strength along with warm water. Warm water, cold air and you get hurricanes. Well over land we don\'t know for sure and that is where the movie really speculates.  However seeing that it has happened before it is plausible that it will happen again and we will be the cause of it. It was cause before by a buildup of greenhouse gasses that took tens of thousands of years to do naturally through volcanoes and other things. This time we have volcanoes and the rampant burning of fossil fuels that only adds to the problem.  Now we can dismiss it all as hokey science but it will only be hokey till it happens. I believe it will happen.  As was made evident over and over again in the movie there is no timetable we can follow because we didn\'t really record it the last time it happened. This time we will.  You can slowly warm a frog in a bucket of water until he boils without him jumping out. When you toss a frog into boiling water he will try to jump out.  I believe the water is already boiling but because we are all so used to these slight changes we will only see that the water is boiling around us when things actually freeze.  Go watch the movie...it is impressive and entertaining all at once...and the science if you dig into it a bit isn\'t far off at all.  Peace.  '),(1358,1234212,'07,April,2003','2003-04-07 00:00:00','   Finally I will be able to add some pics easily to my blog. Thanks Kates! Anyways I have some pics on my laptop and camera that I wan\'t to put up...also a couple on my phone. I have pics of my sis (who I am so proud of now that she is a practicing vetrinarian) and my family. Now I can post them easily. Only a few more things and my blog will be put together to my liking (mainly adding some links and a full profile of myself).  Well I am going to get some sleep. I have to get up in a bit to call my dad and possibly meet with him before he flies home, plus I wanna go to a movie.  Peace.  '),(1359,1234212,'07,April,2003','2003-04-07 00:00:00','   urlLink    I want to thank Kates from over at \"behind the teacher\'s desk\" and her blog for pointing me to this service...this is a picture that I actually like of myself!&nbsp; urlLink  '),(1360,1234212,'07,April,2003','2003-04-07 00:00:00','   So I ended up in the hospital at 4:30 yesterday (Sunday) morning because I had an ongoing allergic reaction to something. I woke up Sunday and my lip was very swollen due to some allergy. I had taken about 150 mg of benadryl throughout the day and it just continued on to hives then swelling in my neck and throat to the point I lost my voice for three hours or so.  It turns out to be some freak thing that I have had happen in the past but in the past I was able to stave it off with benadryl, it was a little different this time because it didn\'t respond. They gave me 125 mg of solumedrol and IV zantac. It brought the itching down but the hives were still there. It is all pretty much done I am still a bit itchy but at least the hive are gone for now.  Outside of that not much is going on. I am going to try and see \"The Day After Tomorrow\" later today (I almost said tomorrow). I am also going to look for a little film called \"Saved\" or something like that. It is a film about a Christian School that Falwell was just on CNN blasting even though he admitted he hadn\'t seen the movie.  As much as I can remember about my times in Christian schools I always found they were filled with horrid children and teachers. We went because my parents had just moved to the area and heard the public school were bad. Well the christian schools were worse because they tended to collect those who had been thrown out of public schools and those who\'s parents were rich sent their spoiled children to school. You had the worst of both socio-economic classes...and that always ends bad if you are in the middle of it all.  I wanna say a little about the evil that is Jerry Fallwell. He is an evil horrid man that if there is a \"god\" I would hope that \"god\" will send him straight to hell because of what he preaches. He plainly preaches hate. There is nothing christian about his views outside of him framing himself as a christian. He is a neo-nazi\'ish bastard of the evangelical movement. He has capitalized on the fears of what I feel are mostly good people. I do not agree with most evangelicals because they teach what is essentially no different from what the Taliban or right wing of any religion teaches. Convert or die and go to hell. The difference basically for the evangelicals is that they are mostly here (with the exception of those in Africa and they do what I am about to describe) and don\'t have the opportunity at killing people who don\'t.  In Africa, the Middle East, and many parts of second and third world nations religious radicals or extremists do go around beheading and raping villages. We constantly talk about how the radical religious extremists, and many times comentators don\'t even bother to differentiate them form normal muslims. We talk about how horrible they are and how truely disgusting the acts that the extremists do when they burn some people or behead a person. It is no different from what happens in many other nations with and equal ratio of christian to muslim atrocities. With extremists in both camps constantly going back and forth killing literally thousands of people.  Extremists always go to the most extreme that a population will tolerate, that is what makes them extremists. If the extremists could get away with killing people here we would see a lot more often. We would see more Tim McVeys, we would see more serial killers. We would also see more religious extremists killing people off. It happens elsewhere, the only reason it doesn\'t happen here is because it is not acceptable. Many of us are just run of the mill people that do not acsribe to the extremists of any religion or political view and when we see those people we tend to wince and turn our heads.   On that note I will say that we have seen a marked shift in politics and religion in this nation towards more radical conservatism (which actually means bigger government and more wars) and more fundamentalist religious views. In moving that direction we have seen a degredation of intelectualism. The religious right come out in droves to fight anyone who questions their ideals or puts to task any of their teachings on history whether ancient or recent. That has been largely seen as acceptable and many mainstream christians have fallen prey to those extremists because it is framed as attacks of christianity and their ideals. When all that is asked is whether or not the historical accounts are correct when they are posed by \"christian leaders\".   Mainstream christians, jews, and muslims need to drive these fundamentalists out of their ranks because those are the people that drive wedges between the moderates and always (always as in throughout history) drives those same moderates to war. It is always the moderates who suffer at the hands of extremists because by the time the moderates are driven into war it is because the extremists have come to power. The extremists are therefor through position protected from being one of those who have to fight. You can see it here right now. Ask any of the right wing hawks of the republican wing (or hawks of the democratic party) if they would be willing to fight or send their children as front line troops. The universally known answer is no. Instead they send pretty much moderate americans to go fight for their extremist views.  All I ask is that we move back towards education and to hold that educating and intellectualism isn\'t a bad thing. We see week in and week out what football or baseball and basketball teams do at tournaments and in games yet we hear only once a year about some great students who have nothing to do with sports. We have university sports programs that have many of their students attending on partial or full scholarships, yet an english department may have only one or two students who get scholarships if lucky. As long as we reward the lack of intellectualism we will continue to see the nation move to the right. The greatest period of enlightenment in this nation of recent times came during the 60\'s. Not that we had a lot of great things but we had artists, thinkers, scholars who came out of this period and thrived during it as well. We saw some of the greatest laws passed during this period and many programs that have been overall good for the republic. Yet many people look back on it and think \"those damn hippies\". Well out of those hippies we had many of the greatest artists and thinkers of recent history. It all came to us via a educational culture that had influence at the time. Many kids were forced into college because of Vietnam and out of that we ended up getting excellent engineers and technology that has changed the world. While it is true we have more college students as percentage now than then the difference is that at that point in time students were encouraged to be students and thinkers and were not looked down on for being such. Now you go to college and you are looked down on in class by others if you attempt to be good at what you are doing. When you do well you are then expected to help others cheat by those who are poor students by them sitting next to you and copying (I have had that happen many times...I get a 90 or better on a test and everyone around me would get the same score!).  We just need to promote thinking again...unfortunately we need to see it from the top and right now the top is a C average student and someone who proclaims proudly that he doesn\'t read often. If you don\'t read you don\'t know...do they draw cartoons like they did for Reagan (they actually did do cartoons for him so he could learn about a weapon system as claimed by Dr. Helen Cladicut who was the only no-nuke advocate to meet with him and she summed him up as being of low IQ...she is a doc so I believe her...she is a pediatrician).  We need educated people who are willing to stand up and be educated and then spread the idea that the fundamentalists are anti-intellectual and therefor should be dismissed as uneducated.  Peace  '),(1361,1234212,'07,April,2003','2003-04-07 00:00:00','   Ahh there we go after many attempts to sign in.  Tonight was good. I drank a six pack and had some fun. I then went to Eat-N-Park for some food.  Not much else to report.  Peace  '),(1362,1234212,'07,April,2003','2003-04-07 00:00:00','   urlLink'),(1363,1234212,'07,April,2003','2003-04-07 00:00:00','   I wanted to write something else and it took me until now to remember...how pleasant.  My mom and I get into what I would call heated debates or talks over Bush, Reagan, the Republican neo-cons, and the radical right, not to mention religions.  She has often expressed her dismay that I am so disallusioned and so negative against my country. In some sense I am negative against our nation...but not so much the people or the physical land, but our government.  I do talk negatively against it and I talk negatively against many of my fellow Americans who are not educated as to what is going on, and I really dispise the people who don\'t know and don\'t care because those are likely the people helping to make matters worse for the rest of us.  I have railed against many of the actions that we have undertaken especially since 1980. As I become more informed about things the more I am enraged that people allowed it to happen. That adults of this nation supported these actions either directly through voting for the leaders or through inaction by not caring at all.  My basic ideal, what I think we get right in all the wrong things that happen are the few acts of blind kindness that do actually happen. The recent mudslides and floods in Haiti where our military flew fresh water and supplies to the people, when we sent tons and tons of aid to Iran after the devastating earthquake last year. Those are acts of blind kindness. I don\'t consider foreign aid as acts of kindness because it always has some sort of caveat attached. When we send that kind of aid it usually is required that the aid money be spent here to get the goods to ship over there. Why can\'t those government spend the aid inside their own nation or in a closer nation on goods that are produced locally. That kind of aid actually would drive economies in those aid nations and help them out of the situation and build allies without restrictions. The fact is that you can get wheat cheaper in other nations where it is needed than you can if you buy it from us.   I see a real disconnect between the rhetoric and the actions of this nation over and over again despite the fact that most of the people here really do mean well. We spend 400 billion on our military to fight against imaginary foes. Our military is simply designed to go toe to toe with other militaries, not against nationless drifters.   I often think what would happen if we dropped bread or supplies instead of bombs. We bomb places back into the stoneage then we bring aid and expect the locals to say \"our hero!\" when we come rolling in. In reality we usually bombed them out of their sustinance living and thus making them reliant on our aid.  I look at all humans as whicked, horrible beings. I hate all of my fellow humans. The reason is because any human has the ability to become a Saddam or Uday, but they also have the capacity to become the Dhali Laama. The problem I have with our government is that we support the creation of the horrible people instead of supporting the development of the good people. The development of bad people is easy because we do it everyday, it is much more difficult to create or grow good people because it means that we have to set aside personal differences and motives. We have to set aside our drive to have more than others, instead it means that we have to work with people and be willing to share the pie with the other 5.7 billion people on this planet.  So change the world. It really is that simple. We just need to give up our primitive drive to have more stuff than the rest of the world. It should be quality over quantity.  Peace  '),(1364,1234212,'07,April,2003','2003-04-07 00:00:00','   Well the hair is now purplish..I had a lot of blue left and only a little purple so I mixed the two and the resulting color looks a lot more purple than anything. I have to wait about 30 minutes then I can wash it out and see the results.   This is probably my favorite thing to do...I love changing my hair color to many shades just for the hell of it all.  I am going to try posting pics again and hopefully it will all work out...if it does I will post the purple hair and the toenail of power.  Peace  '),(1365,1234212,'07,April,2003','2003-04-07 00:00:00','   The service I use for posting pictures is acting up. Well rather the software that is used with it is having some problems. I really want to put up that huge toenail for all the world to see. \"Behold the power of the toenail!\"  I am still working on my Reagan rant...millions dead, tripling the defecit, economic policies that still haunt us today, AIDs \"they deserved it\" epidemic, and the law that required troops to pay for meals when in the hospital because no one in that administration could add or do basic accounting (law was signed in 1981...yeah republicans love the service men and women...see deficits), and just so many other things.  People have been saying a lot that Reagan will be judged as a good president by the future. I hope that the future likes facts better than our current society who instead loves believing that we are great. History would hopefully be well read enough to say that he was a horrible president who presided over many wrongs. History will judge alright, but only if we have a future to judge from.  The biggest thing that offends me is that the right wants to put him on the ten dollar bill and they want to build a monument to him on the National Mall, and they want his face on Rushmore. If they get any of it through, especially the Ten spot I will not use them. I will use only fives and explain to people why I don\'t have tens. Because Reagan was horrible.  We can find fault in almost all of our leaders throughout time, but Reagan is recent and he ultimately was the person responsible for the terrorism hoopla that we are living with now. The root, the seed, stems from his administration and the people who ran it, and the same men run the government now. So it looks like the people will be the ones screwed for years and years. Not only did those people get richer at the expense of the rest of the nation and the world, but they also managed to help create the radical muslim movement that haunts us today. What a crack bunch of fuckers.  Peace  '),(1366,1234212,'07,April,2003','2003-04-07 00:00:00','   I didn\'t get the pics of the huge nail up yet but I will.  I wish I had something in me to write. I am going to work up a Reagan piece here soon. It won\'t be nice because he was a horrible leader with horrible ideas. He was a simple man with below average IQ according to Dr. Caldicot. So I am sure it will be easy together a hell of a piece on him.  Peace  '),(1367,1234212,'07,April,2003','2003-04-07 00:00:00','   I finally was able to dig out this nagging ingrown portion of a toenail. I am so happy! My foot feels sleak and sexy, much more aerodynamic and lighter. It has taken three months to get to it!   Bwahahahahahaha! Behold the ingrown toenail!  I will have pics up of it later :-)  Peace  '),(1368,1234212,'07,April,2003','2003-04-07 00:00:00','   I want to stress how much I love to hear Chomsky on topics. Everytime I hear him it is like music to my ears because of how knowledgeable he is on so much.  I would like to ask that people forward me any and all links that they come across for speeches he has made. I would love video, but just audio would be ok as well. I will probably start another page that has just Noam Chomsky links. So help me out here folks. I am going to start getting it ready here soon because he really is great to listen to his musings on what is going on.  If you can get a copy of the Video/Movie/Documentary of \"Manufacturing Consent\" I would say do it. If you can find it I would be willing to send out a few copies to people as long as they agree to send out copies of it as well. So if you are interested then let me know and I will send out a few.  Peace  Essentially I want to create a Chomsky media library. I have the disk space, the DVD burner, and plenty of DVD\'s and CD\'s so I can get stuff together. I think it would be great inspiration to those who want to learn about ourselves and what we can do to change things.  '),(1369,1234212,'07,April,2003','2003-04-07 00:00:00','   Cities react to Patriot Act.   I have been watching a series on Free Speech TV called \"Water The Drop of Life\". It is a good series and I would recommend it for some educational watching.  I thought I had a Doctors appointment later today but I don\'t it isn\'t until the 16th. I hope I make it that long but the good news is that if I stay out the extra week I will definitely showing the signs of the infection I know I have so they won\'t argue over a hospital stay. It sounds nutter for me to stay out the extra week but it is really for the best because then I can get a new port and all and just get the drugs I need instead of dicking around with oral meds.  Reagan has been dead since Saturday and the media blitz of his \"excellent presidency\" is really wearing. I will be glad when it is over.  Peace  Off to surf the news.  '),(1370,1234212,'07,April,2003','2003-04-07 00:00:00','   urlLink    Pissed&nbsp; urlLink  '),(1371,1234212,'07,April,2003','2003-04-07 00:00:00','   urlLink    Beauty&nbsp; urlLink  '),(1372,1234212,'07,April,2003','2003-04-07 00:00:00','   urlLink    Rufus&nbsp; urlLink  '),(1373,1234212,'07,April,2003','2003-04-07 00:00:00','   No one has any comments?  Come on people comment!  '),(1374,1234212,'07,April,2003','2003-04-07 00:00:00','   I don\'t know what all to write because not much has happened.  Well a few things happened. I am refinancing my car for a hell of a lot less per month. It will only cost me 150.90 a month for it now. I can actually pay it off in about a year and a half. Once I do that I am going to save for a bit then I will probably trade it in and get a newer vehicle that is either 1. more fun for the fuel economy or 2. much more fuel efficient.  If I could find a true sports car for around 10-15 thousand (like a Vette 00-01 model) I would get it. Otherwise I will look into a smaller sportier car that is more efficient. I want something sporty and efficient if I can\'t get a Vette. Probably will get a small two door...unless I find a hell of a deal on a hybrid.  Well I think that will be all.  Peace  PS Is anyone else tired of Reagan shit? I am.  '),(1375,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok so I tried to sleep a bit ago but instead hives came on stronger. I have taken enough benadryl to kill a moose so maybe now they will go down.  Peace  '),(1376,1234212,'07,April,2003','2003-04-07 00:00:00','   I don\'t know what all to write tonight I am pretty much at a loss. I went back and read through a few of my old posts and all I can see are the errors. Spelling and others.  This whole Reagan thing is out of hand. He was loved and loathed. He was good and bad. He was alive and is now dead. I will not attack the man because he is dead but I will attack the policies and the ideas because they were wrong in many cases.  So with that I will be heading to bed or something...  Peace  Oh yeah...I will dig up a few more pictures to put up and I will get with the audio blogging :-)  '),(1377,1234212,'07,April,2003','2003-04-07 00:00:00','   urlLink    My friend Danille at her party...she looks great. When I last saw her before her transplant she looked so run down now she looks a lot more perky. Now we just need to get a few other things worked out and she will be ready to rock!&nbsp; urlLink  '),(1378,1234212,'07,April,2003','2003-04-07 00:00:00','   I wanted to drop a quick note before I head to bed. I went to Danielle\'s birthday/transplant party and she really looks great. She is having troubles and things aren\'t perfect by any means, mostly other things CF related that are getting to her, but it looks like the lungs are ok so far and that is the best anyone could hope for at all. I sat there and talked with her and Joel (Joe...J something), and we talked and talked and talked. It was nice to sit with people who are in the same sinking boat. Joe (l) is in the best shape right now out of the three of us, Danielle and I tie for second/third, she has new lungs that work but guts that aren\'t real well and my lungs are heading down the shitter and my guts still work ok...pain she certainly has far more than I do...so I have nothing to complaign about at all...although my pain is getting past the 6000mg of daily aspirin, tylonol level (yes I take that much). 2000mg two to three times a day depending on pain. It would be easier to take pain meds but really I haven\'t the energy to fight for them at this point. Plus I figure I will doped up soon enough.  Well I will leave it at that...I am just glad that I had a great time...and I got to hang out with people my own age, and I met some really nice people (and a few odd ducks as well).  Peace  Oh yeah Ronald Reagan died today. While I mourn his passing as I do any person, and it is a day of sorrow for our nation. We all must remember that his proposals and his budgets (with help from congress) ran up huge deficits, was responsible for at least several million dead from military actions he supported around the world, presided over the spread of AIDs and said \"they deserve it\", and managed to run many cities into the slums through decreasing federal tax dollars (it all went to \"Star Wars\"...loved the film hated the policy). Again it is a day of sorrow, but also a day to not forget all of the man...the good and bad. Rest in Peace Gipper.  '),(1379,1234212,'07,April,2003','2003-04-07 00:00:00','   Right now as I sit here contemplating my navel and watching \"Duplex\". So far it is pretty funny.  I to Danielle the other evening and we just had a merry old chat. It was nice. I am getting more of my story written and so far it looks like comic gold.  There is so much that I have to get out onto paper and maybe just maybe I will get them all down on paper or on the computer. It will be interesting and fun.  Peace  '),(1380,1234212,'07,April,2003','2003-04-07 00:00:00','   They changed out my big ole feeding tube for a button. It hurt like a mother fucker getting the tube out. The dude just grabbed a hold of and yanked it out, as a result now the incision is bleeding. This was the same fucktard that said my feeding tube wasn\'t infected or having any problems. It just fucking hurt...and if hurts still.  I am going to try and get a nap.  Peace  '),(1381,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok so I hate PICC lines. They hurt like a bitch going in and they are inconvenient as well as a pain in the ass because it limits mobility.  I just hope we can get everything straightened out quickly with meds and all because I need benadryl and if I am going to be couped up I will need something for my damn hips (I take 3000mg of Tylonol three times a day for it).  Well we will see.  I just hate all this mess. Now the numbing is coming off the PICC line insertion point (right now it is only a half hour old) and it is hurting a lot!  Peace  '),(1382,1234212,'07,April,2003','2003-04-07 00:00:00','   You\'re a Speak & Spell!!  You nerd, you.  Just because you were disguised as a toy doesn\'t mean you weren\'t educational, you sneaky bastard.  I am a speak and spell.   What 80\'s toy are you?  '),(1383,1234212,'07,April,2003','2003-04-07 00:00:00','   So lets put it this way. I get a stay at club med Shadyside. I lost 11% of lung function in one month. Wooohooo for me!  I went out tonight (yes in spite of illness) and there was this one elephant there wearing a zebra pattern dress that was way to small. All I could think was \"that dress ain\'t hiding you from anything or anybody\"...things that make you go brrrr...  I will update later today when I get settled in.  Peace.  '),(1384,1234212,'07,April,2003','2003-04-07 00:00:00','   Well as you can see below I finally got my picture service to work properly and you now have the TOENAIL OF POWER!!!!  I also got all the kinks worked out of my SprintPCS connection and got some software to work for my phone finally.  Later today I have a doctors appointment and as of right now I think there is a good infection in there. Last time I thought that I was wrong...but hey this time it feels a bit different and most definitely there is something growing like a little weed in there.  I talked today with the counselor for the transplant center. She asked a bunch of questions and is sending out a packet for me to get started with. When she asked how I deal with stress I was seriously tempted to say \"I run puppies over with a lawnmower\" but thought better of it. No I would never do that.   What kind of question is that? Really can someone tell me? There are two types of people, those who get stressed and those who don\'t. If you get stressed you fall into one of two catagories. You either cope and deal or you don\'t.  First I rarely get stressed about things. I have found it is a waste of time and energy to stress about things that are ultimately out of my control so why worry. When I do get stressed I just tend to shrug most things off to fellow human stupidity.   Fuck it...ya know. Just don\'t let shit get to you. You either have control or you don\'t and if you do then do the best you can and be happy with that. Look you can\'t be right all the time, you can\'t be perfect, and you sure as hell can\'t please everyone, so fuck it. Do your own thing to the best of your ability and enjoy it. Do what you love, love what you do. It is a simple philosophy.  When things get really bad I go for a drive...because I have at least partial control over what is going on.  Well that is it for the evening. Enjoy the toenail...I sure did. The little fucker was in my damn toe for months and months.  Peace  '),(1385,1234212,'07,April,2003','2003-04-07 00:00:00','   Google ads are funny. Because I have \"toenail\" in my posts it is trying to hawk fungus cream and toenail care products.  Freakin\' hilarious!  Peace  '),(1386,1234212,'07,April,2003','2003-04-07 00:00:00','   urlLink    Behold! The Ingrown Toenail of Power! Fear it deeply. Bow to its merciless encroachment into the flesh. The TOENAIL OF POWER!&nbsp; urlLink  '),(1387,1234212,'07,April,2003','2003-04-07 00:00:00','   urlLink    Galaxies Colliding&nbsp; urlLink  '),(1388,1234212,'07,April,2003','2003-04-07 00:00:00','   urlLink    Beautiful yet disturbing&nbsp; urlLink  '),(1389,1234212,'07,April,2003','2003-04-07 00:00:00','   Went to see saved today. The best way that I can put it is if you are a good christian, in other words not a fundamentalist then you would get it, if you weren\'t then you would be offended. Culken was great in it.  Not much else to report. We did have a hell of a storm that knocked power out to the theatre but it was soon restored.  Peace  '),(1390,1234212,'07,April,2003','2003-04-07 00:00:00','   Up late yet again. All of my parts hurt, my brain aches, and my mind is numb.  Looks as if we are going to have a gully washer here so it will be nice. Maybe the bird shit will get washed off my car.  I am writing again. I think I have box office gold on my hands...really grade A classic box office gold. It came to me when my mom and I were walking to the movie theatre from the chinese place we ate the other night. These young girls passed us by and it was like a pack of small dogs yipping at each other and they were no further than five foot in front of us when I said out loud \"I really fear for our future\"...it spun out from there.  I will leave people with this tidbit...it will be one part that will be a takeoff of the WWJD bracelets, only it will be \"What Would Jessica Do?\"...as in Jessica Simpson.  Really it will be box office gold.  Peace  '),(1391,1234212,'07,April,2003','2003-04-07 00:00:00','   Went to see Riddic...or however it spelled. Decent movie, story choppy, slow middle, awesome fight scenes, Diesel good action star, leaves room for sequels or prequels. However go see it at a matinee price, certainly not worth 8 or 9 bucks to see.  Also watched for the first time \"The Last Samurai\"...definitely an excellent movie. See it if you haven\'t. Very well done, great performances, and show the beauty in other cultures. Visually stunning movie and an excellent story. Worth the rental for most, a must buy for some.  Peace  '),(1392,1234212,'07,April,2003','2003-04-07 00:00:00','   Bush left the country today for Ireland and then Turkey. Last night I saw an interview that he had with an Irish reporter for Irish TV. She was a tough questioner, probably giving Bush his first real questions that he had no script to answer.  At one point Bush said that most of Europe was behind the US and the war in Iraq. She looked at him as if he had just run over kittens with a lawnmower. The truth is that overwhelmingly Europeans were against us and this war. Even in the nations that supported us like Spain, Italy, and England, the majority of the people in those nations were against the war. Although their \"representative\" governments failed in listening to their people and took their orders from Crawford TX.  The interview offended me in that our President, elected, appointed, whatever you want to call him it doesn\'t change the fact that he was rude to this woman. I wish that someone would have just slapped him across the mouth because of his behavior.  I mentioned that last night but I forgot to add in the crap with Cheney. He told Sen. Lehey to \"go fuck yourself\" on the senate floor. There are rules in the Senate and profanity is prohibited. What was kind of funny was a Republican backed morality bill that effectively quadrupled fines for indecent things played or said on the radio or tv. The same day Cheney is telling his opponents to fuck off.  Now normally that wouldn\'t bother me all that much. Actually I wish our politicians cursed more and yelled more often at each other. It would certainly liven things up a bit.  As I said normally it wouldn\'t get me upset but this did. A few months back Kerry was interviewed for Rolling Stone and said something to the effect of \"I didn\'t think he\'d fuck it up this badly\" in reference to his vote to authorize Bush to declare war. When it was printed immediately the right wing whackos came out all over and the chief of staff for the White House called for Kerry to immediately apologize for using that kind of language yet when Cheney says it on the Senate Floor in public (and probably broadcast on C-Span which would qualify him for the new fines they enacted..I should look for it on C-Span and then submit a complaint to the FCC and have him fined...that would be a bite).   Again I have no problem with foul language but if you are going to expect one side to live up to a certain standard then they should do just the same. If you criticize one side then do the same for the other side.  What I have noticed is that republicans have become the party of \"do what we say but not as we do\". Just like almost all right wing religious nut preachers and christians.  I don\'t have much hope for our nation for the next four or five years frankly. If we re-elect king moron then we will surely be dragged into more conflicts, loose almost all funding for domestic programs (education, healthcare, EPA) and see military spending increase greatly. Taxes will continue to drop on the ultra wealthy and corporations while the tax burden is shifted to the middle and low class.   If we elect the corpse we will still be in trouble because he proposes tax cuts for companies (although some rollbacks on the taxes on the rich). He will probably drag us into more war and other nasty things. Whatever happens it will be doubtful that he would be re-elected because so many drastic things will have to happen while he is president that no one would vote again for him.  2008 and we see some glimmer of hope if our nation is still intact. We may see Dean run again..and if he does he should tell people exactly the truth. If they had nominated him in 04\' he would have had things fixed up by now or at least on the road to recovery.  I just hope that regardless over the next few years that our fellow humans have patience with us until the people can take back our country.  Just give us time people...we will get things back on the right track sooner or later.  Peace  '),(1393,1234212,'07,April,2003','2003-04-07 00:00:00','   Tomorrow is Ed\'s wedding and I am going to escape on the down low for a few hours to attend it. I am looking forward to it...at the very least I will be able to have a couple of beers and say cheers to the couple.  On the health front there is not to much to report outside of my PFT\'s are going the wrong way. When you come in and do all the treatment and have all the IV antibiotics you can possibly stand the whole idea to the torture is that your lung functions may improve. Not me...no sir. Instead I come in, get tortured (and like it mostly), and loose lung function percentages. I went from an FEV1 or 39% a month and a half ago to 28% roughly two weeks ago, and then just yesterday my FEV1 was down to 23%. Overall capacity dropped from high 50%\'s to the low 40%\'s yesterday. I hate that this is happening but there isn\'t much I can do about it but hold on for the ride and hope for a new pair of lungs.  I saw an interview of Bush by an Irish reporter for his trip over there today. He is pretty much hated all over Europe equally. He was saying that the majority of Europeans were in agreement with his war in Iraq. I don\'t know where he got that idea but almost all of Europe was against it with the exception of the governments of England, Italy, and Spain. However even those nations who\'s governments support the war the people of those nations are overwhelmingly against the war. She was a good reporter because when Bush would tell a lie (which was pretty much as soon as he opened his mouth) she would try and tell him otherwise. When he said that most of Europe was with him she tried to point out that the majority of Europeans were against the war he would interrupt her and tell her she was wrong. It was absolutely baffling because he almost looked like he believed what he was saying. I mean I know he has admitted with pride (he was proud of this?!?) that he doesn\'t read. I never could understand what he doesn\'t read. Was it the newspaper, or books, or reports from the intelligentsia apparatus? What exactly does he not read? I have a feeling that he doesn\'t read anything except maybe his speeches. How do I know he reads his speeches, because he often mispronounces words and in most of the cases of him screwing up words I doubt he actually knows what they mean.  I have no respect for this President. He is not worthy of any respect and if I were a soldier I would not salute him or any of his administration. In all likelihood I would be attempting to find a legal way for the military to remove him from office and allow elections sooner. I would also be looking for a way to root out the Neo-Cons because their ideas and visions are truly dangerous for not only the people of the US but also the rest of the people on this planet.  I cannot respect any group that elevates people who are uneducated, illiterate twits to their top ranks so they can manipulate that person. I cannot respect any group of people who say one thing, mean another, and enact something totally different. The people wear christianity as a badge yet have very little knowledge of how or what a christian is supposed to live. Hell many christians don\'t even know how they are supposed to live, but they all know they must live in fear.  The people on the right are in fear all the time. They are in fear about loosing their jobs, in fear of a god of vengeance that is taught every Sunday and Wednesday. These people are scared of people in the far flung corners of the planet not because there is a genuine threat but rather an irrational reaction because their leaders (the right wingers), and indeed the moderates leaders as well. The far radical right operates on and out of fear as well. The fear that they will be caught, they fear loss of power, they fear poverty, they fear people wising up. That is why the people on the right are often the most combative people around. They are constantly trying to defend against imaginary enemies while at the same time trying to bring more people into the ranks of the right.  The right is afraid. Afraid of what I don\'t know, but they are afraid. It is not just fear here in the good ole USA, if you look around the world and pinpoint all of the radical or fundamentalist right groups or religions, they are all afraid. Afraid of something that is not real. The way they have been able to gain power is by brokering in that fear. As long as we allow ourselves to be afraid then the right will always run things and keep us in fear and at war with each other while they sit back, watch the show, and make money off of our deaths and our fears.  Stop being afraid. There is nothing out there to be afraid of. Don\'t even be afraid of death, it is yet another stage of life. There is no need for fear anymore.  Peace  '),(1394,1234212,'07,April,2003','2003-04-07 00:00:00','   Did I mention that I look like shit. I am having a breakout on my face...I hate being 27 and still having problems with zits and stuff. It is part and parcel of the meds that I am on for CF. I wish that it would all clear up for good because I hate having these breakouts.   Maybe I will get lucky and it will stop...but I doubt it.  I just want it to clear up. I am self conscious about my weight (I want to weigh more than I do), my back because is is curved and I wish that when I sat up straight it would be straight to, I wish that my teeth were white...they are yellow and spotty because of the meds I am on...and I wish my skin would clear up. I have had only a few points in my life where it was clear...and again it is the meds that cause all the problems.  Outside of those few things I am pretty confident of myself. I like the way I look and there is nothing I would drastically change, but the few things I wish were different would make me feel all that much better about myself if they were fixed.  Peace  '),(1395,1234212,'07,April,2003','2003-04-07 00:00:00','   I am so happy. Danielle stopped by today and brought me a cinnamon and sugar bagel and a cup of chai tea.  She looks good but she feels like shit. Her blood tests were all out of whack and she stopped by today because she was on her way to an appointment to have her lungs looked at. I hope that it all goes well and that she starts to perk up more soon because what is the point of getting new parts if you can\'t go out, have fun, and enjoy them.  Something that has been floating around on this CF gimp board that I subscribe to (actually a mailing list) and it has to do with how people perceive CF patients based on looks. To look at most CF patients we look pretty normal on the outside, we may be thin in many cases, but mostly you can\'t tell. Now I have a gimp pass so that I can park closer. It has helped me many a time when I haven\'t felt all that good and allowed me to get out and be self sufficient when going to the store. I don\'t have to rely on someone else to help me from my car.  Here is the problem with it all. People look at me and often confront me about having the gimp tag and using it. It doesn\'t help that I drive a big white Oldsmobile Aurora (Oldsmobile\'s are usually old people cars...or had that reputation). People look at me and assume that I am relatively healthy just from looks but I am not.   I don\'t know how to put it and sometimes it is really frustrating to deal with people that don\'t know. When I am confronted about it there are times that I just want to go off and make them feel through my words how shitty things are. Sometimes I wish that I had this power that I could look at them and muster all the pain that I have gone through and the shortness of breath then touch that person and have them experience it if only for a few seconds. After they crumble to the ground crying I could then ask them if they now have any questions.  I hate being sick, I hate knowing myself so well, I hate wanting to be so involved yet sidelined because of my health.  Not to much sooner and I will be able to move forward...I can hardly wait because there is a lot out there left for me to do.  Peace  '),(1396,1234212,'07,April,2003','2003-04-07 00:00:00','   I am sitting here watching \"The Simple Life 2\" and it is so damn funny. I mean those two are just boobs...well they have boobs but I mean they are boobs in the \"fool\" category.  I think I am going to enjoy watching it this season.  Don\'t have much else to talk about right at the moment. If something comes up I will write.  Peace  '),(1397,1234212,'07,April,2003','2003-04-07 00:00:00','   Early this morning I was watching CNN and on the Early Morning show, not American Morning, but the show before it the man on there derided Michael Moore and his movie and called it propaganda and said it was filled with lies. His cohost looked at him as if he had just farted and called him on it. I wish I could remember the whole exchange but it was just absolutely uncalled for. I cannot believe that an anchorman would say something like that on air. That show in the morning is not an opinion show it is a news show where they report the news. It is fine if he has his opinion and its good that he does but that particular show and that particular time were not the appropriate time to be voicing it, especially in saying that the film lies. Did he see it? If he hasn\'t then he is relying on either uninformed speculation or has only read the talking points of Hannity or Rush.  From the previews that I have seen and the articles I have read this particular movie is not the typical Michael Moore movie like Roger & Me or Bowling for Columbine. Both movies contained fact but twisted and distorted pieces of video to make points that could not have been made without manipulating the situation. For this movie the articles by the people who have seen it basically say that even if it is much propaganda that it still raises good questions and asks things that haven\'t been asked by mainstream media at all.  Again I am mad about how the anchorman handled it this morning and I am going to try and get his name and then send it off to the films crack team that have been put together to counter all the little right wingnut attacks. So we will see. It was wrong and not appropriate and the guy should apologize to the viewers and state that while he believes his ideas but he should have reserved them for an opinion show or other forum, not the news.  Peace  '),(1398,1234212,'07,April,2003','2003-04-07 00:00:00','   Things are finally getting all straightened out. My guts still hurt a bit but much less as time goes on.  I was able to get the meds increased. I now can get oxycodone every four hours, and the oxycontin was increased to 20mg every 12 hours. Also the Benadryl has been increased to 50mg every fours hours. That takes care of most of the med problems.  As for the hole in my stomach hopefully it will be healed up soon. I think they may scope me to see if it is healed, fortunately they knock you out for that...  Talked with the Third World Assasin aka The doctor who fucked up and he tried to explain and weasel his way out of responsibility. I told him that he screwed up and when he didn\'t listen to me when I said it wasn\'t placed correctly that at that point the prudent thing that should have been done was get a CT to see if the tube was placed correctly. Since he didn\'t do that he neglected the complaint and let it go. From there on it was a series of failures from that point on that were not the fault of the staff. All of the errors stemmed from his poor decision and his lack of interest in making sure everything was ok.  He continued to try and get out of it all and I finally just told him that it was unacceptable what he did and ultimately it was his responsibility. He then told me that he was going to check on me tomorrow (later today) and I said that I preffered not to see him and that I would rather see one of the other doctors from that practice. I then thanked him for taking the time to come down and dismissed him.  If he had come down and made a point of appologizing on Saturday and being the guy who did the follow up care and attempted to fix the problem I probably would let this all slide. Instead he avoided me until today (Tuesday...I\'m still on Tuesday) and only came after he got a call from patient advocate department to get records and paperwork from him. He rushed down after that to try and cover his ass...but it is to late.  I am going to sue him...not for money so much. Of coarse I want something for the pain abd additional time down, not to mention because he nearly killed me. I do want for him to pay the difference of bills for me having to go down to the CCU (ICU) and pick up those extra costs and reimburse Medicare because it could have been avoided with a simple CT scan.  I am not asking for much, but I do want it to stick in his mind that in the future he should listen to his patients. I know my body and when I say something is wrong I am usually right.  Well things are coming along...I don\'t know how much longer at this point but I know it may be a bit more time.  Until later...  Peace  '),(1399,1234212,'07,April,2003','2003-04-07 00:00:00','   So here is what happened. I came into the hospital because I dropped my FEV1 from 39% to 28% in about four weeks. It is time for a tune up and hopefully roll back a little of the damage, and if nothing else get the infection under control.  Back in March I had a PEG tube placed and I thought while I am here I might as well have the cap thingy put in instead of having that huge ass tube that I could lasso my dog with.  I should have known better when I saw the GI guy looking at the packaging and the directions on the Button package. It got me a little worried but I just thought “this guy surely has done this and this may be just a tad different from ones he has done before”. Well I still don’t know if he had done others but from my experience I am pretty sure he had little experience.  First he told me “this is gonna hurt a little bit but it will be ok”…he then grabs hold of the G tube and yanks the damn thing out. Of coarse it hurt like a mofo and I was crying by then but I still had a bit to go. We had to get the button in yet.  Still studying the packaging he decides to go ahead. He takes the button and pushes it into my stomach and then inflated the retaining balloon. Well immediately I could tell that something wasn’t right at all. I told him that I didn’t think it was in properly but he dismissed it and said it was fine. Well throughout the day on Friday I was having some pretty hellacious pain in my guts. I kept telling everyone but was pretty much ignored.  Meanwhile I am trying to not have to take shitloads of narcotics because the last thing I want stamped on my chart is “med seeker” that is the death bell for people in true pain. You have to walk a tightrope getting enough to take the edge off but not so much that the docs just think you like getting high.  Well Friday night rolls on and the pain was worse and worse. The kept giving me Dilauded and that was helping but in the end I didn’t know how much I had. They started the tube feeding and within a few minutes I was screaming bloody murder. The came down and by that time I had unhooked myself from the feeding equipment. They soon came down to see what I was screaming about and I was still in pain and writhing around apparently. So the shot me up with more Dilauded. Now I was on schedule of 2mg every four hours which was keeping most of the pain at bay. However apparently I had way more than that over the course of just a few hours. All of the feeding fluid was going into my body cavity instead of my stomach.  Next thing I know I am coming around in ICU.  I found out later from one of my friends who is a nurse here on my floor that I had coded and they had to bring in the crash cart and all. The didn’t need to shock me but my O2 sats were down to 23%...frankly I should have been dead. My friend the nurse came in that morning and noticed all the bustle around my room so she came to investigate. When she got in here I apparently waved at her and said something but for the life of me I don’t remember.  All of this was a cascade of bad events that ended up in me coding.  The problem that I have is what if anything am I going to do about it. If the doctor had listened to me in the first place this chain of events would have never occurred. I don’t know what I want to about it and I am looking for some advice. Pretty much I am forgiving dude and few things bother me with the exception of negligence, laziness, or just plain ignorance, as well as not knowing your job.  The even worse part is my doctor has pulled the dilauded out of my schedule. Which really sucks. I could understand if I was a heroin junky that scored a bad bag, but this was not my fault in the least. The OD was cause by a cascade of events…and now I get to pay for them both in pain and in the battle it will be to get the meds back on the list of things I need.  Again I don’t know what I should do because I was just getting to the point where I knew if I needed meds to control pain I could get them without sounding like a drug seeker. Now that is shot to hell because now I have a history of OD.  So what do you folks suggest…I want to cut the GI guys heart out with a spoon…why a spoon? Because it will hurt more.  Peace  '),(1400,1234212,'07,April,2003','2003-04-07 00:00:00','   Where have I been that I haven\'t updated my blog in a day and a half. Well let me tell you. I have been in ICU...yup ICU, now let me explain how I got here as best as I can piece it together.  It started when Dr. Aziz came to replace my PEG tube (feeding tube) with a Button (much smaller device that is less intrusive and easier to live with). Well when Aziz yanked out the tube it hurt like a bitch and I was starting to think that something went wrong. I should have known when he was looking at the directions on the box on how to insert the damn thing that he probably wasn\'t qualified to do it. Well he jammed the new one in there and I knew it wasn\'t in my stomach and I said something, plus it was really hurting like a bitch. I told him it didn\'t feel right and he dismissed it and said it was fine.  Later that evening I was getting ready for bed and they came to hook me up for tube feeding. We probably got only a few CC\'s in but I don\'t really know for sure because I was pretty out of it all. I started screaming because it hurt really really bad. Now we get to the point where I have very little recollection of what happened.  The must have given me more Dilauded because of the pain...or at least one of the residents did. I remember puking all over the place, then the next thing I know I am coming to in the ICU with a re-breather oxygen mask on.  So basically I think I od\'d without any direct action from me to make it happen. Go figure.  Yesterday I went and had to get a CT scan to see where the tube was and they had to put some contrast in it, the gave me something down there for pain and it worked somewhat well because I didn\'t scream out in pain as I had earlier when my nurse was just trying to push a little air into it to see if she could hear it in my stomach (I about jumped off the bed when she did that). They got the contrast in and it turns out I was right. It was just dangling in my gut and anything that was put through it went into my body cavity and not the stomach like it was supposed to.  Later another GI doc came up and removed it...it hurt but at least it was out.  Now everything is hurting worse than before and it hurst to breath even because of the gut pain. I think there is something in there that shouldn\'t be. The kicker is though that Dr. Weinburg doesn\'t want me to have anymore IV pain meds because I od\'d. Even though I had no control over it, and it wasn\'t like I scored some bad heroin, they OD\'d me and now I have to pay the price in pain for it.  He has me on Oxycodone and Oxycontin, the codone is for \"breakout\" pain, the contin is longer lasting. Problem is that neither is actually enough to numb the pain. I would need a lot more of it more frequently to get the same relief as a small dose of just about any IV pain med.  Again let me stess I am in pain because someone fucked up and they won\'t give me pain meds that work because someone else fucked up. I had been fighting for a long time to get any kind of pain med and then just like that, something that I did not do, I get the relief taken away.  They are going to do another CT scan to see if something else is damaged. I have this feeling that something else is broken down there. This kind of pain isn\'t just normal belly pain, this is \"something is broke you dumb fuck\" kinda pain. Even if they don\'t find anything it won\'t matter, they aren\'t going to get me on the right pain meds anyway.  I had been trying to keep myself from looking like a pain med seeker so I have been very judicious about asking for meds even though I am in a lot of pain, now it is going to be even worse because now I have an OD on my charts. Again someone else fucks up and I pay.  Fuckin a, fuckin a...  '),(1401,1234212,'07,April,2003','2003-04-07 00:00:00','   Damnit...dropped that last post I wrote.  Well I won\'t repeat it because I am to lazy to type it out again.  The jist of it was that I am going to start getting together a lot of Health Care info and I am going to find out how much state health care bennefits have been cut and where they have been cut. In the past four days I have heard of three seperate states that are cutting their health care programs for chidlren and disabled adults. Ohio, Tennessee, and Texas are all cutting programs either completely or so deeply that they might as well cut the programs altogether. That is in addition to the Federal Government or rather the Bush Administration has removed all CHIPs funding in the 05 budget that is now being shoved through the congress.  I am going to compile all the changes in the coverage and then I am going to figure out the costs of a national program, put together proposals for enactment, and shop them out to politicians.    In the process I will also be looking at other nations and how they implement their national health care systems. I think with a close look at how other nations do it we can then build an awesome health care system that serves all who need it here in the states.  It means that I will have to learn a lot and then I will have to help explain it all to people.  It will be my job I imagine and then pimp it out to people. It is going to be interesting.  Peace  If you run across articles or online blurbs about public health care cuts in your area or state please send it on to me I would be very greatful for the help.   michael.honse@gmail.com  Send it all there.  '),(1402,1234212,'07,April,2003','2003-04-07 00:00:00','   I have been pondering over the last day of what I am going to do and how am I going to do it.  I have a pretty good idea of what I want to do with my time. I have a lot of research to do over the next month or so. I need to compile a database of health care for the world. I want to know how much each nation spends on health care per capita and then I need to compare systems.  It is a lot to do and I know that it will take me some time to get it all together but once I get it together I will be able to approach the topic of healthcare for our nation from a more educated standpoint.  I would appreciate any articles that people run across about health care cuts to state programs. Please email them to me at michael.honse@gmail.com that would help me out. I would love to see online articles, newspaper bits, anything that talks about cutting bennefits to the needy. I need to know all of this so that I can see how much damage is being done by all the cuts.  I am not out of the hospital just yet. I am still fighting all of this crap in my lungs and a fight it is.   Maybe I can get myself healthy and then make getting health care for us all my job. I have been needing to find my nitch and I know that it has always been there for me. I am going to be the best I can at that and I will also become a pundant. I will spare no mercy on those who do not agree. I will be the best. I will be the voice of all of us. I watch those people on TV and they talk about shit that really doesn\'t matter to the majority of the population. I think there is a better way to do it all. I will find it and I will make the system come to me somehow...I don\'t know why or how but I know that something is going to happen. Maybe it will be my calling.  Peace  '),(1403,1234212,'07,April,2003','2003-04-07 00:00:00','   I didn\'t get out yet. My PFT\'s are no higher or lower than when I came in so I get to stay a bit longer.  Yesterday I was interviewed by a sociologist from a little college in Carlisle PA called Dickinson. It was my honor to talk to Dan and Maggie for their project. I am not sure what all they were looking for. I know they were wanting to know how people live with CF, not medically but outside of that. What our lives are like, how does it involve our personal relationships with other people, romantic relationships, etc...  I enjoyed talking with them because as I have said before my life is pretty much an open book for those who want to read it. That is part of why I have this blog. I try to keep it as open as I possibly can so that when people read it they can see that I am a person that has much conflict of policy and idealism. There are many things that I believe in such as the death penalty. It appeals to my primal gotta-get-even center of my brain. I often find myself thinking about certain people or criminals that they are great candidates for the death penalty. However I know that the death penalty does not deter crime, it accomplishes nothing in the grand scheme of things so why bother. It also is a sign of how barbaric a society is. If you are still wanting or craving this basic act of revenge it only helps to show that maybe you aren\'t up to speed with social ideals.  These are the conflicts I face. I love guns and I think that any person who is not a violent person should be aloud to own whatever they want. A howitzer, fine just don\'t blow anyone or anyone\'s property up. I think you should be able to own an arsenal but I do believe in some forms of gun control. I believe that it can be helpful but I also know that most criminals don\'t get their guns from legal dealers. They may go through a straw purchase (having a friend or relative get the weapon for someone who can\'t), they can purchase it from a private owner, or just steal a gun. So gun control ultimately only really tracks those who purchase weapons legally. Again though I don\'t think that it is a bad idea to have some sort of controls on weapons.  That is what this blog is for, that is what my life is like. I love politics, I love conflict, I love learning and then teaching.   I beg of the people that read my blog to give input because how am I to know what others think. That is why I have the comments section. Please send me input and let me know how I am doing.  I am going to go ahead and announce here that this blog may suffer a slowdown of posts possibly over the next quarter or more time. Basically there have been a lot of bad things happening in health care and cuts to state budgets that are threatening to start taking away health care that is needed for many people including myself. What it comes down to is that I have to do something and become active somehow.   The past couple of days I have read several news articles announcing cuts to heath and welfare programs in combination with me pirating (which I don\'t do often because I don\'t have a problem with ripping off big media, but I feel bad for taking money away from directors and actors) and downloading \"Fahrenheit 9/11\" by Michael Moore. I knew all of the things he stated in the movie, the sheer theft of tax payers money, the welfare for corporations that is always forthcoming, yet for example CHiPs (Child Health Insurance Program is what I think it stands for and covers children with no health care for both poor and many middle class kids that have severe illnesses or lack of basic care) is being completely cut out of the Bush Administrations budget proposal for 05\'. That money would not be provided to the states which in turn find their CHIPS programs. The states will have to do one of two things. They could gut or eliminate the CHIPs program for their state or the state could cut funding from other health related programs to try and make up the shortfall from lack of Federal Dollars. It also places the state into the uncomfortable position to have to increase taxes or run large deficits. In the states where you have Republican control it is even worse. In those states they first cut programs relentlessly, they then cut taxes where they can, third they run deficit spending as high as possible. The reason is because they want to bankrupt the governments so they can privatized as much of the services as possible.  It is a very dangerous thing they are doing. They essentially want to take as much money from the taxpayers as possible and transfer that money into private companies for services because they believe that private companies can do things more efficiently than the government can. They use the argument that government wastes a lot of money of projects for example the 400 dollar hammer or other outrageous prices for things we can pick up at Wal-Mart or your local hardware store for a few bucks. However they never tell you that it was a private contractor that charged them 400 bucks for a hammer, or where other gross overcharging occurs. Every instance can be traced back to a private contractor overcharging the government so what do you think will happen when private companies take over even more of the duties of the government.   They want to stop social spending altogether so that money can flow into other areas where companies can justify taking more than they could than if the money is supposed to be going to the needy or ill.  I have come to a place where I have to stand up and fight for what I know is right. I can not stand on the sidelines and watch these people destroy my nation, take away programs that many good people depend on. I can not allow that to happen.   Over the next few months I will be gathering data and reports and try to put together a movement if you please to try and fight this. That is why posts here may drop from time to time. I am limited in what I can do because of my health, I am limited in my resources, but I do have a computer, access to the internet from anywhere, and a mind so I am going to fight. I will keep people updated on what is going on.  Mind you what I plant to do. What I hope to do is put together a platform of sound health care plans and options that will be farmed to all politicians. Whomever decides that they like our position then we will support that politician or candidate. If they are in office and claim to be in support of our platform and do nothing to promote it then we will confront them. I want to be the pitt bull of medical care reform for a single payer public health care system. I want to be a pain in the politicians ass.  So hopefully we can get this going. My only cry to people is that they have the same view as me on health care and I will not care what their political leanings they have as long as they will help me fight for real substantive change.  Peace  '),(1404,1234212,'07,April,2003','2003-04-07 00:00:00','   I did a bad thing and downloaded \"Fahrenheit 9/11\" by Michael Moore. The copy that I downloaded was unfortunately not a complete or finished copy.   It was the worst movie I have ever seen. The movie stirred such deep emotions that all I could do was cry while I watched it. To watch a movie about things I have been reading about, to see the pieces put together in front of my eyes, to see it visualized and put together. This movie isn\'t the worst I have ever seen...the content is the worst subject to cover because it is true.  My mother has told me over and over again that the way I talk is anti-American and she feels that I hate America. I love my country, I love the idea of my nation, I love many of my fellow countrymen and women, but I detest my government and all of the leaders within in. I detest the fact that many of my fellow Americans cannot look past the end of their nose and see that the real war, the real battle is not between us and some little shit nation on the other side of the world, it is not with some man in a cave, our war is with other human beings. We bomb these people back past the stone age most of the time. It is a war that is not even valid or remotely real in the sense of that these people are a threat to me or any of my fellow countrymen and women. The real war, the real battle is between the vast majority of people on the bottom end of the economic scale and those very few who hold power and riches.   We are in a war and we the people are loosing. There is no difference between my family and an average Iraqi, Indian, Iranian, Israeli, French, British, or Mexican family. We all struggle for the same thing, our families strive for the same thing, we strive to have food on the table, happiness in our homes, and the hope that our families will be healthy. We are all the same in that respect, we all strive for basically the same thing. We all have a common enemy and it is the upper crust of our societies, the powerful in our societies. Not all rich people are bad, not all powerful people are rich, and not all powerful people are bad, but the vast majority are. People with money and or power will do anything to keep and gather more of it and the prime way to do it is to lead us against each other. Powerful people only remain powerful when they are allowed to pit us against each other and in the process it makes wealthy men who have absolutely no compassion for anything or anyone outside of themselves.  I don\'t know how I can explain it any better but to say that all people are the same. If you want to go and avenge the beheading of Nick Berg or any of those poor people who have been brutally killed...if you feel that we should go and kill those barbarians and behead them then you are no better than those people, you are no better than the people that want to drive the masses against each other. You are a barbarian just like the rich and powerful.  Crime of all sorts and kinds comes out of a need for control, power, and wealth or some combination of the three. A rapist is someone who wants control and to dominate someone else, a crook is someone who wants control and to gain wealth. All crime is based in those three basic things. All crime comes out of it corporate, personal, or national. Our nation is nothing but an international criminal in its search and war to gain control, power, and wealth. It is that simple. If we spent as much time and energy as well as resources that we do on trying to gain those things we could solve almost any problem from world hunger to split ends on your hair. There is no problem that cannot be solved that if we put our minds to it in an equal and equitable fashion then we could really start to heal ourselves and our planet.  The movie was not a typical Moore film at all. He only pulled his schtick in small portions of the film. Fortunately Moore allowed facts to tell the story and only a couple of times did I find where people might find things confusing because of how he sequenced things. Otherwise he put together a horribly wonderful film. I am saddened that a film like this has the facts to support it. It is sad that the story is indeed true. I hope that more people see the film and it sparks them to go do there own research. We should never take one persons word on faith as the absolute truth because that is how we enable powerful people the ability to turn us against ourselves.   The even sadder thing is that Bush may be defeated in November but in reality little will change. Kerry won\'t stop the war or calling for wars. He will be in power and it will corrupt him. We would be better off with Kerry but not by much. The simple fact is that our government no longer serves the interest of the people it only serves the interests of the rich and powerful. I think it is getting close to revolution time, for the people to rise and speak as one around the globe that we are no longer going to kill each other for the powerful and wealthy people\'s amusement or benefit. If Bush has a grudge against Hussein let them fight it out man to man.   That movie has made me mad and made me want to change things...I cannot stand by and let these things happen. To see the brutality of our weapons ripping apart children and women, people burned by napalm, for what? For what? Because someone needs more money or power. This war does not benefit me or the average soldier. No more personal freedom will be seen by the average Iraqi person all they continue to see is oppression and fear that they had under Saddam...and to say that at least we aren\'t killing them by the thousands and we aren\'t torturing them then I beg to differ. We drop bombs on them daily and whether we like it or not those bombs kill more innocent people than \"terrorists\", and we have seen the pictures of torture. We haven\'t seen all of those pics so I can\'t and most of you can\'t say \"at least we didn\'t cut of limbs or fingers\" because we don\'t and probably won\'t know all the facts and details of the torture that is worse than we saw in those pictures.  All of mankind is created equal in that we all have the capacity to be Hitlers, Hussein, Stalin, Napoleon, Bush, Reagan, or any other number of equally despicable dictators, rulers, or criminals. It is every one of us. We also have the ability to be Mother Teresa, Gandhi, or other humanitarians. The goal should be for all of us to try and foster the growth of good people.  First we must wake up to the cold plain truth of it all. Second we need to acknowledge those problems. Third we need inspire and lift up good deeds to the level of attention and adulation that we give sports stars and business people. Fourth we need to forgive those people that have kept us at war for their benefit and profit. Fifth we must make sure that we do not allow ourselves be ruled by fear and the same forces that control us now. We must resist that urge to follow blindly and resist following on sheer faith alone. Doing it that way has allowed humans to continue to be bad people.  Go see the movie and I am pretty sure you many of you will cry as I did.  Peace  '),(1405,1234212,'07,April,2003','2003-04-07 00:00:00','   This is a continuation of the incident I had during my first couple of days in the hospital. I had all those problems on the 19th where Dr. Aziz had removed my PEG tube so that he could put in a button instead. His missed getting the button into the stomach so when I had a feeding that night the fluid went directly into my abdominal cavity cause great pain. My belly was distended and was limiting my ability to breath and then because of the pain my nurse was able to get my normal 2mg of dilauded upped to 4mg to catch the pain. Well they didn’t know what all was wrong and didn’t realize my belly was distended and once the dilauded was on board it naturally slowed my breathing down.  Long story short I coded and we traced it back to the button not being properly placed that led to the chain of events that caused the code.  Today I had my PEG tube replaced and went in for the procedure and all went well there, they kept me a little longer in recovery just to keep an eye on me with all the meds they had in me for twilight sedation and pain meds. Just to make sure. Once I was pretty much aware (which they hit me with a reversal very quickly after the procedure was finished) they brought me up to my room and immediately I needed some pain medication because the stuff they gave me down there was wearing off.  Well I already have prescribed meds 5mg oxycodone every four hours as needed and 20mg oxycontin every 12 hours. The added meds that were approved was another 5mg of oxycodone every four hours. I knew it was low but I wasn’t going to fuss until I knew for sure that 10mg wouldn’t work. Took the 10mg oxycodone and no dice, the final bits of the bigger meds wore off from downstairs and the oxycodone wasn’t cutting it. I requested that we see if we can get something stronger just to take the edge off, not to rid me completely of the pain just get the edge off.  My nurse called my docs service and the doc on call said I could have 15mg of oxycodone once and then back to 10mg every four hours. She brought it in and I took it because at that point I was crying even though I knew it wouldn’t work. I explained to my nurse what had happened at the beginning of my stay and said that it wasn’t supposed to be a problem in getting IV pain meds when needed for bad pain. The doc had seen that I had the “OD” which wasn’t an “OD” but rather a bunch of crap that just dominoed out of control.  So my mom called the service and then talked to the doc on call and he was telling her he can’t give me more because of my OD a bit back and oxycodone was all he was going to give me…she crawled all up his ass and he ended up calling my regular doc. He called in 1mg of dilauded every three hours as needed and I am now finding out if that is all the pain meds I get or if I still have the oxycodone every four hours plus oxycontin twice a day.  The problem was that I should have been able to get IV meds but was unable to because of my chart and record (mind you my first PEG tube I was ordered 1mg dilauded every four hours and vicoden 7/500 every two as needed). I had talked with my doc because I was afraid that this exact thing would happen and it did.  Before I wasn’t to worried about what all happened and would have accepted an apology from Dr. Aziz for his fuck up, he never did and instead tried to weasel his way out of it saying things like “He couldn’t tell” or “it happens every 1 in 100” and other crap, I replied well when it doesn’t feel right or you aren’t confident that it is placed correctly what do you do? He said we get a CT scan. Well he didn’t that day because I was the last task of the day for him so he blew out at the end of his shift even with me saying it wasn’t right.  After that exchange I wanted for him to pay the difference to the insurance company as well as Medicare for the additional costs of his screw up like ICU costs, extra CT scans and such…nothing for me but reimbursement for the extra costs.  Now that I have had this problem with getting meds because of his fuckup and me in pain and having to track down my doc to get something that would work and knowing that this will happen again in the future now I want blood.  What would you folks do? I have already had my family look for a malpractice lawyer and as luck would have it we know one and his daughter studied voice under my grandmother. I feel bad in one way because I don’t want to bring a frivolous lawsuit and the system needs reformed so huge settlements for silly ass things would never fly…I am torn on what to do.  What would you do? What would you ask for? It wasn’t about money for me in the beginning, but now having dealt with what I feared no more than a week later and knowing that if it happened this time it won’t change anytime soon and if my doc is out of town then other docs who aren’t in the loop to what happened then I will be in pain until I get a hold of my normal doc. What would you ask for? What would you demand? Am I overreacting to this all? Please let me know what you would do…  ________________________________________________________________________  That was a letter I wrote to my CF group and it pretty much sums up my suck ass day.  Peace  '),(1406,1234212,'07,April,2003','2003-04-07 00:00:00','   WEll today I am going to be getting another PEG tube put in...oh joy!  It is better that I have one because this whole eating thing is just not working. I will also try weight gain formulas instead of that crap they give you. First the aftertaste in the morning is just bleech...and secondly if I am going to be filling my stomach and not eating during the day I might as well put something in there that has more calories than what that crap they give you.  The good thing about surgery is you go in, they give you the drugs, then you wake up later on with everything done. Another good thing is that they will be giving me better pain meds, or rather stronger ones. Maybe I can get the Dilaudid back because that shit works miracles. Oxycodone products (Vicodan, oxycodone, Oxycontin, and others) work when you get enough in me...but they don\'t work well at the doses they give me which are too low considering my high tolerance to pain meds.  Well that is all for this update...I will probably update while I am still drugged when I get back so the post may be interesting...  Peace  '),(1407,1234212,'07,April,2003','2003-04-07 00:00:00','   We turned Iraq back over to the Iraqis yesterday and it looks like there is going to be much more bloodshed over the next year or so.  CNN has a story about what people on the streets over there thought about it all. The wanted us to go...blaa blaa blaa. They are leary of the new \"governors\" and government that is in place. Then they got to a newspaper publisher and he is running a headline today which was pretty interesting and I hadn\'t thought of it all this way. The headline goes something like this \"America keeps its promise\" and goes on to tell what happened.  I am glad that we did hand soverienty back over to them, but what we handed them is a bit of legal mumbo jumbo that really sais that if they (Iraqi government) want to change parts of the laws Bremer and his crack team of corporate leaders and lawyers they have to come up with big votes and the way the structure is set up it may be impossible in practice to do.  The prime minister Ayad Allawi was a loyal Ba\'athist leader until he left the nation a while back. He has little experience and I believe he comes out of the same school of thought as Hussein. Allawi is a brutish thug that thinks he has great ideas, not unlike Hussein. If some people are correct in their assesments we may have just installed the next Hussein.  I hope that things calm down for a couple of simple reasons. 1st I want for our troops to get home as quickly and safely as possible. I just wish we were leaving now and let the nation drag itself into a civil war and then deal with the regime that comes out of it. 2nd I want it to end soon because the Iraqi people have been brutalized for the past 35 years and we have only continued the pratice. I fear that we may not know how to deal with them and their culture differences, but it is clear that we are bumbling about making things harder. 3rd That whatever Iraqi state that emerges will be a secular state and reject fundamentalist religions.  That is my wish list. I don\'t want Bush to get the credit at all. I want for there to be a debate of what we did and how badly we damaged our relationships with other nations around the world.  I will say something positive about what has happened and what is going on. We were wrong to do what we did but we have done a few good things. They now have a less limited TV channel and mor papers have cropped up spurring free speech. We have been attempting to get schools open and hospital equipet (why that one is so hard is beyond me...they could do it and fully stock drugs for hospital in mere months if the money is invested). There are things that are getting better, but it is going to continue to be hard and we will continue bleeding. I just hope we can get a pressure dressing on it soon or we may loose it all and be dead in the water.  Personal stuff. I am still in the hospital. My PFT (lung function test) was bad a few days ago (worse than what they admitted me for), the second one I had on Sunday was back to where I was when I was brought in. Overall no real improvement. So Dr. Weinburg wants to keep me a bit longer.  My pre-albunim levels (protien measurement) was 18.9 when I came in. 20 is normal so I was not to terribly low. The last set of blood they got yesterday moring shows that my levels have dropped to 8-9 which is not good. The best reason for this happening is the whole ICU adventure. First I am a skinny little dude, I have few resources to go digging into when things go bad. Well after the ICU visit I was on a clear liquid diet then was bumped to dark liquid diet (grits and other stuff like that). I finally got back on a regular menu and my appetite is pretty much shot. I will put something in my mouth and chew and chew and chew and then try to swallow. The top portion of the swallowing mechanism (back of mouth) works somewhat ok. It is the middle bits that are causing problem. I have to fight to get basically mush into my stomach. It is something physical because I am having trouble with getting even pills down (considering I could take 10-15 at a time) has me worried. I need for those parts to work.  The nutritionist (which she is nice here and helps out as much as she can) stopped by and told me about my albunin levels and said to me that she thinks I need the tube badly to get my minerals in order. I know it doesn\'t help me gain weight. Doctor Weinburg also agreed and first we are going to look and see if a fungus is causing problems with swallowing and if that is the case then we will treat that and then swallowing may come back on its own. However they are also going to evaluate me to have another PEG tube in. I think that I won\'t escape not having to get one. They suck bacause they just hang all balls out. I am going to ask if I can go directly to the button. My stomach hole is almost healed totally and they may be able to cut back through and save the one hole and just get it into my stomach. We will see...I would use it more often (every day and night) if it was easier and less intrusive to me, and less balls out hanging of hoses and such.  With the lungs not perking up he is going to run a coarse of Solumdrol (a steriod) to get them working maybe a bit better. It should also stimulate hunger. I am glad I didn\'t leave today because no more than an hour after Weinburg leaving I had a treatment and went for CPT (chest physical therapy -aka- a beating). I didn\'t just but barely get comfortable and then after two lite taps I knew something popped. I didn\'t even bother to cough as it just slid out of my mouth. I had the moron PFT dude here and once it started I asked him to get me a contianer to cough in or anything that will hold sloopy  bloody gew. It took him almost five mintuse to find the bucket I asked for and just grabbed the whole thin and brought it over. Now in this bucket is assorted goodies you might need incliding a spit cup. So she\'s handing me the entire bucket with all the shit and I just reached in grabed the kindey like bowl to cough into it. I then just gave him a dirty look. I don\'t like him because he is just a poor respiratory therapist and probably the worst here in this hospital.  After Weinburg got back he lookes at it and mumbled \"guess I am glad I didn\'t let him go\". This will give him a chance to figure out what is up. so...  Well it is very early in the morn and I want to get some sleep.  Peace out  '),(1408,1234212,'07,April,2003','2003-04-07 00:00:00','   Nevermind about Wednesday...we already transferred sovereignty of Iraq back to them. It changes nothing, we still will be occupying, we still be making most of the choices, we will still be the \"security\" for them, and if they ask us to leave we will laugh at them and tell them to get back to pretending that they are running things.  We know that the right-wing nutters will all be out on the talk shows and news shows talking about how smooth it went and how they now have control. Bush will once again be the \"hero\". It will look like we caught Hussein in the polls. Bush will receive a bump in the poles but once people realize that there will be continued violence against Iraqis, our troops will still be in the crossfire and many more will die (unfortunately), the beheading will continue. Nothing will change and we will still pretty much be in control...the difference is they have a home grown \"hero\" that has done a great job in the past working for the CIA.  Oh something that I don\'t know if I had mentioned before is the situation with the detainees from Afghanistan, and prisoners from Iraq. The Red Cross and several other groups have come out and said that first they should have more access to these folks, but more importantly they need to charge them with a crime, release them, or break international law.  The administration line is that Afghan detainees do not fall under the Geneva Convention because the Taliban never signed onto the treaty, and that Al-Qaida forces are not under them either because they are \"enemy combatants\" that have practically no rights for anything. Legally they may have a case in some weird way but they painted themselves into a corner on these folks legally. First when a case was brought by a lawyer who could not gain access to his client there as well as take sworn testimony from others. The military and justice department claimed that because they could not be questioned because of their status and that the supreme court has no jurisdiction over them because they are not in the country. Then they claimed that Gitmo was part of sovereign US territory so certain laws applied. Basically the Administration tried to play both sides...so it is confusing.  As for Iraqi prisoners the administration considers the top guys from the Hussein regime as prisoners of war. Well the good ole\' Geneva Convention says that once the war is over and all prisoners captured during the war, the winning party or the ones who have custody of the prisoners have to bring charges on these people if they did anything within 90 days or they have to set them loose, or they can break the rules of the treaty which is what they have done.  Hussein has fallen into a limbo situation. He has been in custody for over 90 days, the war is over, and he hasn\'t been charged or released so keeping him in custody is against international treaties and law. It is claimed that we will not turn him over the  Iraqi courts until we can guarantee that he will get a fair trial (hah! yeah right!) so we still have him. The US government will not turn him over to the international court because during the trial they will be able to indict other big wigs who helped him get weapons or facilitated war crimes (gassing of Kurds and Iranians). So who would that be...hmm Donald Rumsfeld could be charged with crimes against humanity, the late Reagan could have been charged, and many many others high in the government that was in place during the 80\'s (Papa Bush would also be charged). So world court is a no no...so we have to turn him over the new government there or we have to deal with him.  Oh things are just so fucked up...and if we could actually get Hussein a fair trial it could take down the all mighty neo-cons.  Oh and on another note. Hussein, Bin-Laden, Scott Peterson, Kobe Bryant...in our court system, in our law, they are all innocent until proven guilty by a court of law. Sure we know Hussein was a bad dude and killed many of his own people...and yes he is guilty as hell, but he still deserves a fair trial. The same with Bin-Laden...we pretty much know he did it and planned it but under our system he is innocent until proven guilty in a court of law. I hate all those people who attack people like me or Michael Moore when we state this point of fact. It doesn\'t mean that I think they didn\'t do it, but to serve justice, to be a true democracy, and to follow the law that has been practiced for many many years we must insure they get a fair trial. It is the American way...if you are out there saying why bother with a trial for these dudes just hang em high...well I hope that you are never caught up in some false charges over killing a little retarded girl. The evidence may point to you but in reality someone else did it, but if you go with the \"screw it he\'s guilty\" ideology then the system might go away and you won\'t get a trial. So as repugnant as it seems these men are still due there basic rights.  Peace  '),(1409,1234212,'07,April,2003','2003-04-07 00:00:00','   I may be going home a little later today if I am lucky but I am thinking that there is something brewing, while not fully out the bugs are certainly partying up in my lungs.  After I coded last Friday evening I went on O2 and have been on it since. Without the O2 my sats drop to between 82-89% on room air. I require some O2 to get back into the norm. Right now I am on 2 litres of O2 and my sats hover around 90-94%, on more oxygen the better they get. The room air sat was at rest to, and that scares me. I me going on O2 no.  Right now I am watching \"Enemy at the Gates\"...an excellent movie.  Well I am going to try and sleep.  Peace  '),(1410,1234212,'07,April,2003','2003-04-07 00:00:00','   urlLink    Proud to be a Moore fan. Seeing it as soon as I am out of the hospital.&nbsp; urlLink  '),(1411,1234212,'07,April,2003','2003-04-07 00:00:00','   Wednesday is the day that we turn \"power\" back over to the Iraqi puppet regime. The violence is getting worse by the day, more Iraqis are dying each day, US troops are in the crossfire, the resistors are getting bolder, and all the while the administation line is that things are better.  Inflation supposedly hasn\'t risen but for those of us who aren\'t wealthy inflation is a bitch. Inflation is based off of a rise in price of durable goods (car, dishwashers, I don\'t know about housing but I think it is counted as well) in addition to other goods and services like food, gas, and a whole host of other things.  Milk in some places is over 4 gallon, gas is hovering right around the 2 dollar mark (although we still pay less than most of the world for gas), foodstuffs have gone up because it costs more to get food to market, housing costs are up, energy costs are up...all of these things are up yet inflation is still low. The problem is that wages haven\'t risen to keep pace with these basic goods. Sure a washing machine cost the same as last year, but you can\'t eat a washing machine.  Inflation has happened only to the poor...or rather it is the poor that are suffering the most. The pay check that has remained the same for the past year or more isn\'t making it now when the basics command a higher price.  It is all just a mess. So what the economy gained 100,000 jobs in a month. Chances are that job is lower paying for many of the workers getting jobs now then their last job paid. So not only is Joe making less in dollar amount per pay cycle, but that money is also worth less than just a year ago.  The nation is groaning under its debt. Trillions in federal debt, tens of trillions in private debt and Bush says things are getting better. Well Mr Bush it may be getting better for you and your circle of men, but for the rest of us it isn\'t...  Things are going to break soon. I just hope it breaks for better and we come out with something better because we certainly cannot keep doing it the way we are doing.  Peace  '),(1412,1234212,'07,April,2003','2003-04-07 00:00:00','   I got an email today from the person who runs the CysticL mailing list. He wants me to stay on topic and stop saying other shit. I sent out a reply to a reply of a reply...basically I floated a porn movie idea...it wasn\'t dirty or descriptive, it was funny. He also wants me to tone down the political discussions that I often have/do/start.   I kinda told him that I would tone down the off topic stuff (basically not spin things out as far as I do...my brain will spin things to the n-th degree and I often write it down or say it out loud...my little censor dude gave up a long time ago and said fuck it...let him make an ass out of himself) but I told him that he could pretty much shove it up his ass for other things. I will hold political discussions about health care and what our government does because it does directly effect CF patients.   Also my kidneys feel a bit better...so I am not as gimpy as I was.  Off to read some news and other stuff.  Peace  '),(1413,1234212,'07,April,2003','2003-04-07 00:00:00','   Today I have been pretty miserable. I hope that the rest of this day goes better.  I am rehydrating myself because I think that I am very dehydrated. That is probably why my parts hurt (kidneys), and might be interfering with me just being able to breath well.  Well I am going to get some rest after I get my next shot of dilauded.  Goodnight.  '),(1414,1234212,'07,April,2003','2003-04-07 00:00:00','   One of the nurses here at the hospital brought me a couple of old issues of Time. Many of the things, and people involved, are the same cast of characters now.  The issue is from December 13, 1971. In the reader comment section of Time some people posed or nominated a couple of characters that are now on the front lines.  \"Sir: My nomination for Man of the Year: John Kerry. His eloquent, haunting protest against the war may well mark a turning point in our country\'s \'long day\'s journey into the night.\'\"  By: Jero Magon of Miami Beach  \"Sir: For Man of the Year: the uncaped crusader, Ralph Nader.\"  By: Mrs. M.E. Valentine of Dearing, Kans.  It is interesting to note that the players that were around then are still around 33 years later. I find that sad because it means that politics and the whole machine is definitely not geared to bring in new ideas and leadership.  Health Matters.  I went down for a CT earlier today to find out what in the hell is going on that has been drawing down my O2 sats. I had to get another IV because they can\'t put the contrast right into the PICC line because of where it dumps out at. The contrast, had it gone through the PICC line and dumped out where the PICC line ends, could have caused a serious embolism.   The CT scan showed that there aren\'t any huge mucus plugs or blood clots, but it does show some sort of diffusion in there that could be a new virus that we haven\'t seen in me yet.   My kidneys started hurting earlier today, not just uncomfortable hurt, rather please stab me in the heart with a rusty spoon, salt, and lemon drops for sterilization.   The biggest problem is that my O2 sats are dropping more each day and that is the scary part.  So maybe if we can figure out what that stuff is on the scan and treat it, suck it out, or yell at it, then I think things will get better.  Peace  '),(1415,1234212,'07,April,2003','2003-04-07 00:00:00','   Ok I am going to bitch a little. I am having a lot of problems breathing and my sats keep on dropping.  With 8 litres of oxygen I am at 91%, at night worse I guess because in the morning the past few mornings my oxygen has been bumped up another litre or two.  Right now I can barely breath it feels like so I am going to relax some more and just remain calm.  Peace from the purple boy.  '),(1416,1234212,'07,April,2003','2003-04-07 00:00:00','   Instead of sneakily changing my previous post I need to report that the news about the Bush documents didn\'t hit the light of day until 6:47 AM. I reported that it hit the world at around 6:30. That was the last time that I had looked at the clock before I laid back and relaxed (I also got benadryl and dilauded at that time so time moves funky on those drugs).  So there is the correction. I don\'t want people to say \"YOU LIED!!!\"...I just got the times mixed up by about 17 minutes.  Peace  '),(1417,1234212,'07,April,2003','2003-04-07 00:00:00','   If I had felt a little better when I first heard this early this morning around 6:30 AM I would have posted it immediately. What would posses me to want to post something that early considering I had been asleep (I was woken up for Benadryl, Dilauded, and Fortaz) and would have slipped back into the warm embrace of unconsciousness.  There was a very short bit of blurb or press release from the White House. It seems that three months of El-Presidente\' Bushs\' military record had been destroyed. The real shocking part is that it was the months of service records that would show where he had been when missing and why; possibly, he was decertified for flight. I am sure that is was this big ole\' mistake and all of the records from that three month period for his Texas Air National Guard unit were destroyed...so if you wanted to oh say, find someone else\'s record it would be impossible. The New York Times is the one who is reporting this mess. It seems that there was an effort to preserve microfitch films of payroll records and they were destroyed.  I don\'t know how you screw that kind of restoration and preservation up. They have been practicing for many years in keeping them fresh and usable. It is just mighty damn convenient.  I just wanted to get that up. I heard it first on CNN and they were reporting on a NY Times report.  Some linx...how clever of me linx/links (wink wink)   NY Times links may require registration   Reuters report at NY Times.     The same article different place on NY Times.     Campaign 2004 Coverage on NY Times.    You may have to register to get the articles...but registration is free.  Regardless of whether it was truly an accident or it was a deliberate destruction of files there is something else that comes into play. When the records were finally pried from the White House some time back at the beginning of the year they claimed that those were all the files and they had done thorough research trying to find the paperwork. They released all of it with some of it blacked out for privacy reasons (which I understand, but it looks like he is hiding something he is still entitled to some privacy) they made a very big deal of the fact that they were releasing all of the files and that they missed nothing. Remember they released them late on a Friday evening and didn\'t let the press remove them from the office so they had to scour through them late in the evening and were not allowed to take them out to thoroughly vet them. So they lied again on this...while it may have taken them by surprise, it really should not have if they had truly looked under every rock, looked through all the records and such. So it was a lie of omission.  Peace  '),(1418,1234212,'07,April,2003','2003-04-07 00:00:00','   There is so much coming to head today. We have terrorist alerts at such a strange time. Even given the benefit of the doubt and they may be planning a large attack, the timing is just way to damn convenient.  Stories of Pakistani security forces being given deadlines to catch or kill the top leaders like Bin-Laden and others.   However the congress, specifically the Senate is locked up with legislation about a gay marriage constitutional ban, and other worthless pieces of legislation instead of bringing up the bill that funds our intelligence machine and money that is needed to make preparations for a possible terrorist attack.  Of coarse the Republicans say that the Democrats are holding things up because they refuse to vote on the other bills, the Dems blame the Republicans because of bumping up the legislation for funding they would rather have things go in sequence and before that bill comes other much less important bills that will probably not be passed anyway.  I have an idea...dump all non-essential bills. Get to security bills, get to bills that send money to police and firefighters. Who cares about a fruit marrying a fruit. If you are more concerned about gay marriage and think that it should take precedence over anti-terrorism legislation and funding...then go for it. You know what you deserve to die.   Fucking retards.  '),(1419,1234212,'07,April,2003','2003-04-07 00:00:00','   I have been looking over the past few days of posts and I have noticed something. Actually I noticed it earlier this morning when I was being given my morning doses of IV meds. Let me explain.  Over the past few days I have been breathing a little faster and shallower. I haven\'t thought to much about it until this morning when the pieces just fell into place.  This morning I had ripped my oxygen off at some point during the night and when my early morning IV treatment which this morning included benadryl, dilaudid, fortaz, and then added this morning was a blood draw. When the nurse first woke me up I guess I was kinda going along with it all then all of a sudden I was aware but blank and I asked what I was supposed to be doing when she handed me the blood collection stuff. She looked at me strangely and then she said she was going to get the other meds. It came to me but it took a good minute for me to realize what I was supposed to do. Once it clicked in my head I drew out the waste blood and then got the blood sample they needed and just sat there pretty much dead headed. It was a strange and uncomfortable feeling. I put the oxygen back on and pretty soon I started to get back in the game and started processing things more completely.   When I am not on oxygen I go goofy. My lungs have finally gotten to the point where I think I may need to have oxygen all the time. I was completely disoriented and it only got better once my O2 levels came up. When I mentioned what happened to my nurse that is taking care of me now he said that the night shift nurse had said something about me being disoriented when I didn\'t have the O2 on. So it was like a revelation out of the blue.  Damn that sucks because I really don\'t want to have to drag around an O2 tank and shit. I was hoping that I could stave off carrying around an O2 tank altogether and just get to transplant or whatever was going to happen. I guess I will learn to deal with it as I have to with other things but I really didn\'t want to have to. Side effects of low oxygen is that my fingers peel. The skin up to the nail starts peeling back and that has really kicked in the past few days.  Shit.  I also wanted to do an early post because of some things I have seen over the past few hours. Particularly concerning the Kerry - Edwards campaign already. I saw their first speech together and it was almost like night and day. Kerry is a shitty speech giver. He is able to say a lot without making a point. He can tap dance around an issue, give a twenty minute speech, and all you asked is \"how was your day\". When Edwards got up it was lightning had hit the crowd. While the applauded and cheered Kerry, the might as well have built an altar and sacrificed some virgins for Edwards because he was just that much better. He picked up where he left off during the primaries and just kicked ass.  Now I will give Kerry a little bit of a bone. He has started to get better at giving rousing speeches and I believe that one on one with Bush he will definitely look smarter, sound smarter, but I fear that he may do to much tap dancing. I fear that his intelligence may work against him. I understand what he says, many people do, but we all just want to yell at him \"GET TO THE FRIGGIN\' POINT\". There is nothing wrong with the way I feel or many others. Our hopes are on the line here. Many of us feel that the very future of humanity may be on the line. I feel personally that if we choose Bush we choose war. We will be at war for some period of time and it will cost us the good will of the world and in the end we will be defeated. If we choose Kerry I cannot say for sure that there will not be war, but I do have the feeling that Kerry won\'t go marching off on premature or pre-emptive wars. Will Kerry stop the bloodshed in Iraq or will he remove us from that mess. I doubt it. I would hope that he would listen to the will of smarter people and just get us the hell out of there, but I doubt he will. We do need to find a way to help them, maybe we leave and just send food and equipment, let them fight out a new government. I dunno the answer because it has now become infinitely worse.  Back to the campaign. Edwards is a help. He is able to identify with normal people. I don\'t care that he is worth something in the neighborhood of 30 million, and that Kerry through his wife Teresa is worth between 500 million to a billion bucks. I don\'t care about that. What I care about is what they are going to do to help me, or rather are they going to make my life harder or easier. Harder is me worrying that I may loose my health insurance, easier is me not having to worry about that. I don\'t expect gold to rain down and the economy to miraculously open up and honey to flow freely. We just need for it not to have bad fiscal policy that is driving our nation into ever deeper depths increasing the likelihood of our nation foundering and put into a position where it will be ripe for the armies of \"privatizers\" to come in and destroy what so many people have given their lives for over our history.  This campaign is a turning point, I am not so naive to think that if we elect Kerry - Edwards all will be magically better, rather if we vote them in then we at least will have a few more years on this our experiment.   So forget about Kerry being dull (he talks all the right things that are emotional...but rarely does emotion seep through the tough veneer), forget that he blusters on for interminable periods of time, but in the end he is the better choice and the best hope that we can keep on rockin\' in the free world.  Peace  One last thing. This is an article that I stumbled across that says what I have tried to say about Michael Moore\'s  \"Fahrenheit 9/11\" . Give it a read as it really is dead on.  '),(1420,1234212,'07,April,2003','2003-04-07 00:00:00','   Woo...I thought that they had monkeyed around with Blogger some more. I am glad they didn\'t because I am used to this format.  I cleaned the room pretty nicely because there was just odds and ends hanging around. So I got it cleaned up and the room looks nice...a side note is that I got a bit hungry and I have been biting the heads off of sour patch kids (you see when you bite their heads off you need to make little screaming noises so that people with think you are eating babies....mmm babies. I want my baby back baby back baby back ribs...chilis baby back ribs....git in my belly!  Ok enough of that. I have this bumper sticker that I can fill in. It has a \"I\" then a heart, and then you fill in the rest with whatever you want. I was going to go with WHORES! But lets face it...I don\'t want a whore (unless she is a dirty dirty girl like Christina).  I got two compliments today. One was from a family friend. I had wrote her a letter about what all happened since I had gotten into the hospital. I really don\'t remember much of the letter because I scribbled it down so soon. Apparently what I wrote was a very detailed description and she liked it. This lady is a retired librarian so she knows good from bad. She showed the letter to her daughter (or daughter in law) who teaches an English class and she raved over it. My friend called to tell me that she loved the letter and to ask me if I would allow for her daughter (in law?) to use it in her English class to drive home the descriptions. Hey so something I have done is going to live on and be taught with.  Also I had a newcomer to my blog site last night. She is a nurse from the hospital here. Normally I do not invite people from the hospital because I am weird and if it got out how weird my care might get whacky...naa but there are some things that are part of my sordid past. Some things of coarse I would change, but overall I am proud of the work I have done. This nurse complimented me on the content and told me to continue to be true to me. Woah!  I have things to work out...and right now I am dead ass tired. I am about to get a booster or benadryl and then dilauded.  I will be back tomorrow....till then  Peace  '),(1421,1234212,'07,April,2003','2003-04-07 00:00:00','   I am currently doing a complete backup of my Toshiba Pocket PC GenioE 550. It is about time just so I have it all if for example it does something goofy and I have to restore the system.  My friend Kate stopped by last night (7/5 and early 7/6) and we chatted for a bit. She is pretty damn cool and easy to talk with because she has shit to say.  Today a friend of the family just popped into town. Carol Lee is a great person and I love her dearly. She is the lady that has offered to give me a place to say, a car, and all sorts of things to just make me happy and help her out. She certainly is a great person and would bend over backwards if there was anything she could ever do for me. I feel really lucky because in reality I have a great set of people around me in my family and they are very supportive.  She gave me a great compliment today and it is actually something that I have been thinking about it so getting the compliment made me feel like I wasn\'t just wasting time. She said that she thinks that I should go back to school and get a political science degree with a history minor (anyone who has read this blog would understand that I would easily pass these courses without much effort). Then with that teach and write books. She knows my teaching style and I will toot my horn a little...I can do a good job at teaching and you will walk out with the heavier duty words because I will use simple language and attach the more difficult terms to it all and associate it so that people get what it means. I am like Edwards and what they are saying where he can break down complex directives and speech and break it down to what people can understand. I do that well.  I do want to write and I am practicing and trying to work out some of my bad habits. I will then have to pick a topic and run with it. Like I have said I have ideas in the works for things I want to do and from that I can teach and become active. I want to be on the national show and challenge these dufuses about ideas they supposedly know all about what middle America wants.  Well I am going to get myself ready for another shot of benadryl and dilauded then I need to get some food from the cafeteria and then hit the sack.  Thank you all for sticking with my blog and know that you will see the changes as they happen. Hopefully you will all become somewhat active with me....and if you all give input into my work you will all be credited.   Peace...  Oh yeah and I have a few more blogs to add and some other stuff. I am also going to be working to make this place a place where I will feel no shame about letting the nursing and hospital staff read....so I will have to not forget and slip and say I would chase her in the weeds kinda thing.  Night all.  '),(1422,1234212,'07,April,2003','2003-04-07 00:00:00','   urlLink    A picture of my doc...nobody believed me when I said I could a pic of him in a tutu...so here it is :-)&nbsp; urlLink  '),(1423,1234212,'07,April,2003','2003-04-07 00:00:00','   urlLink    Here I am at the beginning of the year visiting with my dog Rufus when I was in the hospital. As soon as I have a better pic to post of myself I will....but this will have to do for the profile pic at this time.&nbsp; urlLink  '),(1424,1234212,'07,April,2003','2003-04-07 00:00:00','   urlLink'),(1425,1234212,'07,April,2003','2003-04-07 00:00:00','   urlLink'),(1426,1234212,'07,April,2003','2003-04-07 00:00:00','   I am going to drop a quick note that this mailing list that I am on for CF patients is just a lot of fun.  Many people on there are prudes and seem to get their knickers in a knot over really silly stuff. However most of the people are really nice and can take a joke well, and it does have some great things going for it because they are all there only for support and questions from each other. It is a good place to get things out, but it also a good place to have fun.  TTYL  Peace  '),(1427,1234212,'07,April,2003','2003-04-07 00:00:00','   Well obviously those posts I wrote did go through but I wasn\'t able to tell that they had actually made it to the blog.  I am watching \"Fahrenheit 9/11\" again just so I can stay enraged on this; our Nations Birthday, travesty of what is going on in the world.  Our Nation is now 228 years old. We are still a young nation in looking at others around the world. Our democratic republic has been the most ambitious attempt at creating a society that allows great personal freedoms to go after any goal one could ever imagine. The great part is that even with all of our national faults people have proven time and time again that any man or woman can lift themselves from one socio-economic class to another. The promise of this nation is that it offers the opportunity to go as far as you want. Our nation is great and nearly alone in this idea or practice. The promise that our nation can be a beacon to the rest of the world of personal freedom and political progressiveness. The sad part of all this is that our nation is faultering in these ideals.   Our nation is in grave danger from a very determined group of individuals that preach the virtues of our system yet continuously subvert them through money, political clout, and downright theft in many cases.   Some of us know who I am talking about some of you don\'t. I will give you a clue as to whom these people are although I cannot tell you outright to continue my thought or spread the word because that actually subverts the very idea of our own nation. The enemies of our nation are the very people who wear their patriotism and religion on their sleeves. They are the people who tell you to listen and follow blindly without questioning what we are doing or where we are going, much less what it will cost in lives, dollars, and above all else our own personal freedoms. You have seen these people on TV lieing about current world events, although many people don\'t see the lies or recognize them, but if you don\'t believe them then we can look at these same people who say \"You are with us, or you are with them, the terrorists.\" Those are the people who are the true threat to our nation. It crosses party lines, there is no specific race behind it all. This is a large group of people from all sorts of walks of life.  I also speak of Pastors, Ministors, Rabbis, Emans, and all religious leaders that have that same mentality. You are with them or you are with the damned. All of these people are the same. One group may be running our nation, the others running our moral lives. All of them are the same, they are all terrorists of the worst kind. They are the terrorists of thought, the terrorists that take away our freedoms, the terrorists are those people that seek to keep us in fear all the time regardless of what is going on around us.  9/11 changed things for us for sure. The biggest thing that it changed was our blindness to the rest of the world. We lost 3000+ lives, we launched two wars that have had tens of thousands of civilian dead, we have lost around 900 of our finest in military actions. However probably the most devistating loss to us has been the loss of the illusion. Finally with the blinders off we now see what the rest of the world has lived with for many years. Our nation has been lucky on the terrorism front, we have remained relatively untouched considering the role we have played in causing terror around the world.   The loss of the blinders, the opening of our eyes to even a cleansed, sanitized, and largely censored view of the rest of the world has cost us our national sanity.  Sanity? Yes. Sanity. Since we were not attacked but rather the target of a terrorist strike or event whatever you want to call it we have lost our minds. We had people on TV selling parachutes for people in skyscrapers to use in case of fire or terrorist attack to use as a means of escape. There has been a marked increase in sales of so called \"safe rooms\"...hell we were told that we could survive a biologic or chemical weapon attack if we simply sealed off a room in our house with duct tape and plastic. I wonder just how many people actually fell for that. I wonder how many people honestly felt that with some plastic sheets and duct tap they could avoid anthrax, or worse. How many people really fell for that. The simple fact is if you sealed yourself into a room with plastic and duct tape when an attack was happening it probably wouldn\'t stop the little baddies getting through that minimal barrier, but if you did manage to seal yourself up well enough you would run out of air to breath.   We lost our sanity and we have willingly given into fear. When we started to fear we gave up willingly many rights that our men and women of the military have fought for. What our forefathers struggled for, and what we continue to preach that we have. It is ashame that at the very time we are supposedly bringing \"democracy and freedom\" to other nations, offering up our military for that ideal in other nations, at the same time many of the very rights we fight for are being taken away here at home. They aren\'t fighting to take them away, we are handing them our rights.  It is time to stop the madness. It is time to stand up and say that we are more affraid of loosing our rights than being the victim of a terrorist attack. It is time that we actually honor those who died in the name of our freedom on that trajic day in 2001. We must honor them by taking back what we have lost.   So on this 4th of July 2004 I mourn for my nation. I mourn for the men and women who have died for my nation, a nation that has been hijacked by fundamentalists no better than those who occupy caves in Afghanistan, or the deserts of Iraq, or those who are here, now plotting their next strike. I mourn that we have become largely a nation of scared people who no longer rely on themselves for their truth but rely on those who have lied to them over and over again.  On this day I also see that our nation is on the verge of rebirth. I see all of the various groups and interests coming around, heading directly for each other in a crash that will shake the world. Out of this crash I feel that our nation will be better for it. I feel that on the other side we will be that beacon to others. At no other time in our history have we seen such forces at work, while many have said that before I truly feel that we are in a time of change.   So on this, the 228th birthday of our nation, I see that she is coming to a rebirth.   Long live the United States of America and her promise to be the best that humans can be.   Let us focus our attention and energies of shaking the tree of freedom and loosening the bonds of fear.   I don\'t know about all of you, but I am ready to have my nation back.  Celebrate that our nation is indeed a great experiment and we are currently adding many ingredients and pretty soon we will have something new.  Peace...love your fellow man. Love those around you who dispise you. If you live that way and pratice this every day and you will be better than all those out there who would drive us around in fear.  '),(1428,1234212,'07,April,2003','2003-04-07 00:00:00','   Lets try this post again.  I am going to compile a database of each state\'s Public Health Care plans and systems. In that database I am going to look at the number of people it serves from year to year, the amount of money spent year to year, and the budget year to year. Basically what I am going to do is see how much the systems have been cut.  I will need some help in getting the numbers together but mainly I will need to have people help me collect and compile news articles on state health care budget cuts.  The whole idea is to find out how fast the system is shrinking and then find out through calculations how much it would cost to insure the uninsured. Ultimately I want to see how much it would cost to have a national single payer health care system.  Now here is one of the things that I think should stay the same. I think that hospital groups like UPMC (University of Pittsburgh Medical Center), or other University hospitals, and all hospitals (except VA facilities and those need to be upgraded and fully funded immediately) will remain under their current control structures. In other words the Government doesn\'t appropriate them, instead they will bill one entity for all medical bills. You can, and should, keep records of procedures and costs per patient but the hospital won\'t nesecarily won\'t bill individually. The hospital will submit a monthly bill to the insurance system and it pays it out to the hospital. Auditing will be done every so many years.  Adding new technology to a hospital will be a cyclical thing. A new type of scanner is on the market that is usefull then the hospitals buy them just like they always do and then pass the bills on to the insurer as part of what would be a yearly upgrade budget for hospitals. There will be of coarse funds that will allow for hospitals to expand and renovate when needed but again because the hospitals are run privately or as they are now the billing only goes to one place. They still have to come up with money for new wings or equipment. There will be some money available each year from the Government but the rest will come from fund raisng drives or through submitting grants. Also they could by the equipment and expense it over a period of time.  The idea is to allow hospitals to run as they do now but instead of having to bill up to 250 different plans just bill one place. Overhead goes down radically and the waste of beaurocracy is almost gone immediately.  Well that is all tonight....maybe this one will post afterall.  Peace  '),(1429,1234212,'07,April,2003','2003-04-07 00:00:00','   urlLink'),(1430,1234212,'07,April,2003','2003-04-07 00:00:00','   Hmmm..they have changed the input section yet again. It is pleasing to the eye and it looks like they have added a few things. &nbsp; Anyways Steven Hawkings has solved his own paradox or so he says. He will be presenting it soon. Basically from what I understand he said that everything that goes into a black hole get gobbled up and just goes \"poof\" and all the information of the matter that went in there is lost. It is complex quantum physics stuff that I certainly have little grasp on. I just know something isn\'t right about the whole \"poof\" its gone. He now thinks that black holes essentially leak information over time so it destroys his \"poof\" idea. Here is a  Newscientist.com writeup on it.  &nbsp; I think that theoretically any black hole can gobble up all of the matter in this universe over an infinite period of time. When it finishes gobbling up the universe or matter here that matter isn\'t just lost it goes \"bang\" into a new universe and we have the big bang all over again. Now the limiter is time because it becomes meaningless. So there can be billions of black holes in this universe and they are all creating new universes elsewhere on another plane. Because time is the limiting factor or rather not a factor you essentially have this continual big bang happening over and over again as a new black hole forms. So once one forms a new universe is started because time doesn\'t matter. That is what I think. That is the only way I can wrap my mind around it because I always thought you cannot destroy matter, you can move it, you can change it from one form to another, but you cannot outright destroy or create it so it has to come from or go somewhere.  &nbsp; Wow my head hurts after that one.&nbsp;I love the theoretical stuff. I wish that I was smarter so that I could do that.  &nbsp; Well if Hawkings comes out and says this in his speech then guess what I got it down first. I have thought this over and over again.  &nbsp; Anyways...I am going to surf a bit more then probably sleep. &nbsp; Peace  '),(1431,1234212,'07,April,2003','2003-04-07 00:00:00','   They finally got me back to a normal room and I can\'t tell you how much better it is. The rooms are about the same size but they are better configured and plus I have a mini-fridge here in the room so I can get to keep extra food and drinks with me.  I had another little lung bleed again this morning. It happens, I am not going to get to upset with it all. As long as it stops like it did then it isn\'t a big deal.   I don\'t know what else to say this morning except that I am mad because they have on Golfing on TNT instead of my normal morning TV schedule. 10AM ER, 11AM ER, 12PM Judging Amy, 1PM Law and Order. It is all just screwed up so I have to float and I am actually watching ESPN and on right now is the trade of Shaq to the Miami Heat. It is actually the most interesting thing on right now. Plus I like Shaq. I like how he lives, I like that he says that he is no role model so put him up as one, yet he lives a pretty decent and private life. I just see a good man who wants to get paid for his play and he plays well. I just hope that he can continue to be a good role model because even though he says he doesn\'t to be, it seems almost out of question that he would get caught up in any kind of a scandal. He has kept his lifestyle on the downlow and if he has diddled outside of his marriage is seems that he does it right. I doubt that he does that completely though after seeing him in various shows with his wife and those kids he has. You can just tell that he is love with them and would never do something to jeapordize that and I have a great deal of respect for that.  I also like the fact that he credits his parents for where he is and that his drill seargent dad kept him inline. I believe that is why he is the way he is. If he were to screw up it would disapoint his parents and I think that he holds them in high regard.  Well I am going to surf around for a bit. Dr. Weinberg said that he thinks that these ports in here are supposed to be setup for Ethernet. If they are then I will need to get wireless router so that I can get all my stuff working. I love my internet on the phone but if I can get uber speed through the ethernet port then I am all for it. That will make things easier for me down the road. As I get sicker (which is going to happen I have realized that) I know that I will be spending more time here. So I might as well have all the right stuff on hand so that I can play and have some fun. There are things that I want to do.  I am looking to build a better SFF box (small form factor). The new stuff that is getting ready to come out is driving costs down. I just want to get a system that has everything built in and allows for the possibility of RAID drive in the future because I would love to get two WD Raptor 10,000RPM 74GB SATA drives and RAID them together and have my third drive be just a big ole SATA drive (I have 120GB right now and I love that drive). Also I have a decent processor and good RAM and an excellent vid card. I can also get a full fledged double layer DVD+-R/RW burner for about 90 bucks. If I can put this new system together for a few hundred bucks that means I can pass down a system to my mom then my grandmother and I would get her old box so I could set up a storage server and get to my Terabyte of storage that I want to get too. I think I can do it.   Well of to read this review of a new SFF and see if it fits. I will go to a better one after a year or two (I think I want to go to a Dual Opteron SFF that is coming out, but only if it has PCI Express for the new cards, enough room for a RAID drive, I can live with the extra storage out on the network...so I have a lot of plans).  Off to read...  Will update later.  Peace  '),(1432,1234212,'07,April,2003','2003-04-07 00:00:00','   I have been reading around a bit and I was right many people have started writing about the whole idea of postponing or canceling the election on Nov.2 and how bad an idea it is.  Unfortunately my brief rant pales in comparison. My was a little more anger filled and fact light, but I did get the way it would happen right. President Bush is the only one who can declare a State of Emergency, once he does that FEMA becomes the default government which basically cancels out all of the rest of it.  FEMA runs the nation and ultimately at the head of FEMA is the president. They literally take over everything including your own car and property. Everything falls under them and then they declare martial law to basically enforce it all. This swing essentially is for six months if the President chooses so and the congress can not take up the State of Emergency issue for at least 6 months.  Now in this article that I am going to give the link to gives all the solid code and law behind it showing that it is indeed a real problem that I would expect Bush and crew would seriously consider doing to remain in power. The reason that he could and would is because he has seriously made 9/11 into a political weapon.   Basically if there is a threat, a real one (and I happen to believe that there is the good possibility that something may happen) and the Administration didn\'t do anything they would be knocked for it. Also if the Democrats raise to much of a stink and essentially make it politically impossible for him to do it and an attack did happen the Democratic Party would literally crumble because they will be portrayed as having let it happen which the Republicans would say \"we could have stopped this if the president had the power\" kind of argument. However if it is allowed to happen because we are in fear then the terrorists have won and we might as well stop being us. Plus all of the worst things that have been happening under this administration will come on full bore and you just know that is dangerous for all of us.  Well I will not longer bore you with my paltry words. I know turn you over to  William Rivers Pitt and \"Only Cowards Cancel Elections\".   Very well written and how if puts it is poetic.   Peace  '),(1433,1234212,'07,April,2003','2003-04-07 00:00:00','   This was a letter that I sent to my CysticL support group and I thought I would post it on my blog for tonight.  --------------------------------------------------------That is an absolutely astute observation that doesn\'t just apply to our little group but that should apply to our day to day routine.  I think that there has been a large attack on intellectualism in this country and around the world...but especially here in the United States and I have been able to see it in just my short life.   What does this nation hold up as that standard bearers of \"wealth\" or good living. They hold up sports hero\'s, models, actors, and all sorts of people that aren\'t always the brightest bulbs on the trees.  We see every night on our evening local new the scores for all the local football, basketball, baseball (not even always girls mostly just guys) scores and who did better. You have MVP\'s and others. You may see once a year the spelling bee national conference, and then maybe one or two local stories of those kids who do exceptionally well in their class.   It continues on to college in a big way. How many dollars does the university give to have English or History full scholarships. Now look over at the sport department and you will see that many if not all students there have some sort of scholarship and many of those scholarships are full tuition scholarships.   Now many people say that sports programs bring in lucrative contracts to the universities that infuse millions into them. The problem is the money almost always goes without fail to the sports programs for stadiums and many other things. All of the endorsement deals put shoes on their feet, uniforms on their backs, and helmets on their heads. The TV deals go to support scholarships and great TV studios (that rarely if you have AV tracks at your university get to use those facilities and often find themselves using older equipment). So all of the money that the University derives from the sports program goes right back into it and then the University still has a budget for them that grows each year, meanwhile they cut science funding, or they decided to put off a computer lab upgrade for a year.   During all of this the students tuitions are going up. Now how does this fall into the placating and dumbing down of kids. Well used to we would get our education at schools and university level but now you can\'t expect to get it unless you fight for it.  We hold these people up as hero\'s, we hold up the all mighty dollar as the absolute measurement of success and happiness. Unfortunately that puts most of the world at a disadvantage. There is simply not enough for us all to be multi millionaires.   In doing that...or rather in figuring that out both political parties have joined together or rather an entire class, the rich, have loosely banded together to keep the rest of us placated and hungry enough to work hard for the possibility to move up, but only let a few people through so you keep the illusion that everyone will do well.  I watch it all the time. When you look at trying to engage in debate with a person outside of your socioeconomic class you are automatically called a class warrior, and your argument may have nothing to go with class outside of that there is a definite class system right or wrong.   You want to debate what is good for our nation, you not a patriot if you don\'t tow the President (dictators) line. Without question, yet our forefathers were great fans of debate. They would make it a sport almost. A speech that would be given back in that day would be several hours long and it would engage the audience at hand. People would fire questions at the candidates and then the candidates would actually answer the questions directly. An interesting thing.  Politics is verbal war. Many people have forgotten that politics is the practice of democracy, and politicians and the powerful people behind the scenes don\'t like for ordinary stiffs to be involved.  So well. I tell people be irreverent towards power. The powerful only derive power if you give it to them. I don\'t say be reckless. I don\'t say become an anarchist, but I do say become educated on something you like or an aspect of life you love because you can be an activist for anything. For CF\'ers the best place to be an activist may actually be in health care and politics. We know what is going on and we have to deal with the system so might as well become good at it.  Whatever you do enjoy it...and it also means that you have to learn how to debate. Learn to play devils advocate on issues you might deplore. The reason is because you will be able to argue a better case against your foes.  I did that for a while on abortion rights. I am pro-choice in my legal standing and pro-life in my personal standing. It is not a weak position. I can comfortably tell people that I think it is horrible and they are destroying some sort of life, but I would not presume to attach my morals or ethics to anyone much less force them on someone. The law should be to allow safe reasonable abortions that follow good sound medical (not right wing partial birth and ugly term that doesn’t even apply terms) terms and procedures. So let people choose. Do whatever you have to in helping clear your own conscience if it bothers you. Go down to health services and just engage, not nastily (remember you always catch more flies with honey) and talk to some of them. You may save a few, but chances are none. Remember that if you are the religious type then you should know that those people will have to answer for their own decisions, not you.  If you are pro-life, take the pro-choice side a couple of times and be open to it. Really research it and see that some of these people fighting for that right are fighting for some basic women\'s rights of having choice over there own bodies. It is hard but anyone with reasonable intelligence can do it. That is how I became pro-choice...and I also became that way while I was in my really religious phase because I kept on reading that god gives choices to all his creatures and we can choose to go his way or another way, but either way he loves ya.  Anyways...I tell people to question things and be ready to argue. It helps out. Drag your families into these discussions. I was always taught to listen twice as much as you talk (makes you look smart in front of strangers, but also you have two ear and one mouth so you should listen twice as much as you talk) but when at home keep on engaging people. If they won\'t, just keep on and keep on.  Well that is all I am going to get some sleep...  Peace -------------------------------------------------------------  Hope ya\'ll like it.  Tonight I started on the BYPAP machine. It is this device that kinda breaths for you...no tube and not invasive. It is a big ass mask that just has positive pressure on it and blows in when you inhale and draws air out when you exhale. I had it on last night and it brought me back quick like...so tonightI am giving it s shot again to see if tomorrow is even better.  Gnight\'  '),(1434,1234212,'07,April,2003','2003-04-07 00:00:00','   Well here is my \"oh shit oh shit\" update for the day.  Last night I coded...again.  I am not to sure of the events as usual but I remember flushing my IV because it was done and getting a head rush, I think I pushed for help, and I think I remember calling out for help...I think.  Last night Lori was looking after me and I know she was right in there helping me out...I\'m sorry Lori I didn\'t mean to do that...I dunno what happened. Maybe I can get clued in.  I woke up kinda down here in the unit...again. They were taking a blood gas and I remember telling the guy it would be easier to get it out of my left side. They tried to get it a couple more times...anyway I have six holes that would be blood gas sticks.  It is all very odd. We will get to the bottoms of it sooner or later.  Peace  And no I am not dead yet...  '),(1435,1234212,'07,April,2003','2003-04-07 00:00:00','   There was something else that I wanted to comment on but I just forgot about it.  Bush has turned down the NAACP for the fourth year to come and speak at their national convention.  He stated the reason he has turned down the invitation is because of \"what those people have said about me\"...umm so he is not going to give a speech because some of the NAACP top officials have been critical about the administrations handling of affirmative action legislation.  It is just a stupid thing. Bush is nothing but a racist pig. Every time they have to change Air Force one out and get one of the older birds to fly because of mechanical problems (seems to be a lot of \"mechanical problems\" throughout this administration) on the main Boeing 747 that jets him around the report is always \"the president was not in any kind of danger\"...all I do is snap my fingers, nod, and say well damn. It would be fitting if he went to Iraq on one of his secret fly ins and for the bird to be downed by RPG fire...I would cheer all day long. Bush is a piece of shit and we will be better off without him.  The more it goes on and the more often I have to see him, Cheney, Ashcroft, Rumsfeld, Ridge, or any of that bunch of Nazis make me more and more angry and hoping and wishing that something will happen to knock all of them off at once...cut the entire head off of the republican party (don\'t forget Tom \"show me the money\" DeLay...he needs to be picked off as well).  Pieces of shit!  Peace  '),(1436,1234212,'07,April,2003','2003-04-07 00:00:00','   I feel a bit better today and at least I didn\'t wake up completely short of breath, I hope it lasts through the day.  There has been more coverage of the possibility of election postponement. I hope I run across some op-ed pieces on it. As I do I will link them for you all.  Peace  '),(1437,1234212,'07,April,2003','2003-04-07 00:00:00','   I have more links about this whole delaying the vote thing.    Commondreams.org Coverage     Truthout.org Coverage     MSNBC.com Coverage    Now these links are all based off of the same Newsweek Story and Reuters coverage. In each of the articles you will find links to those various sources as well.  It is all very scary. As I got the info up, many others were getting their info up. Most people simply got the story and reprinted it, so far I have yet to run across any opinions outside of my own. I had a link to the story but I also shared my thoughts about it all and my fears.  I am sure that over the next couple of days we will see more opinion pieces about it, but so far I am one of the first people to get an opinion up.  Health matters:  Today has been a shitty shitty day. All day I have been cycling between not so bad periods where I could breath easily to stretches where I couldn\'t catch my breath at all. There was not rhym or reason to it all and that tends to be very frustrating. It is also painful as well. The pain that you experience is very different from other pain. I really don\'t know how to explain it. Maybe if you were to take a deep breath and then swim from one side of a large pool to the other underwater without coming up for a breath and then turning around and getting half way back. When you come up and you at first are huffing for air, well take that first two huffs and puffs and extend that out for ten or twenty minutes. You have all the same muscle pain because you entire body can\'t get the carbon dioxide out so it is sitting in there giving you horrible pain. Well it is something like that. Oh yeah don\'t forget the headache that comes with it. It sears through you and it hurts horribly.  When it gets that bad there isn\'t much I can do but keep breathing. It hurts horribly but you do it. I have found that if I get a shot of pain meds and benadryl it calms me down real fast and I stop hyperventilating. That is probably the worst part of it is the hyperventilating.  I just hope we can figure out why it is happening so maybe we can keep it from happening more.  Well I am tired and I do need some sleep.  Peace  '),(1438,1234212,'07,April,2003','2003-04-07 00:00:00','   Hi kids...and welcome to the new Fascist States of America, where a constitution and bill of rights watched over our land to keep us safe from the tyranny of evil men. These beacons of civilization held on high as if they were the documents of some radical religion, guarded zealously with the blood of millions of people from our own nation and protected from the blood of the tens of millions who dare challenge this most basic document set.  I despair that today may have been the warning shot, an alert, to the new norms in our once proud nation.  Today it has been found out that the Bush Administration has started to look at if a terrorist event were to happen before the November Election (or maybe even if it happens after the election up to the installment of the President on Jan. 20, 2005) whether or not it would be legal to declare a state of emergency and post pone the election for a set period of time or whether they could suspend the election for an unknown period of time.   There is precedence of this happening and it was relatively recent too. In World War 2 President Roosevelt was retained as President because he was already up to speed on the issues and the war and to replace him would be difficult at best, so they kept him for a third term until he died and Truman took over. The idea is that Roosevelt could have remained president as long as the war was on simply because of the logistical problems of getting a new guy up to speed.  That is the precedent. Today we have already had our \"leader\" say that this war will be a long one and a very dangerous struggle and that shifting courses mid-stream could be devastating for the nation so we should vote him in (that is his thinking at least).   You could say that well why wasn\'t it done in the Vietnam era, and there is a simple answer to that, the Vietnamese posed zero threat to the USA, they could not pull of any type of attack he on the homefront. If the Soviets had invaded then surely elections would have been postponed.  Some people will argue, and it is a valid argument (if your an idiot), that Bush would be doing it for the best of the nation and a vote would be held swiftly. The action would be a terrorist strike here, perhaps NYC or Washington, maybe LA, Bush would declare a state of emergency and martial law. All of the government would switch into a protective or secure mode which means the government becomes airtight and we cannot elect in new officials, actually it could give Bush even greater power to appoint more of his people to positions of national importance. Since the action happened during the \"War on Terror\" it would enforce the idea that our nation is in a state of war and give him the imperative to keep us from voting. Under a state of emergency and martial law all civil rights are removed and you have to listen to the government or you can be jailed or killed.  What about the courts. Well that beloved Supreme Court of ours is already leaning right and we have several judges that are on the verge of retirement. If the state of emergency was still in effect at that point Bush could and would install whomever he wanted to the vacated seat/s thereby tipping the court over to radical right. Also if the people were to bring suit to try and get voting reinstated the courts are nearly powerless to do anything. So if you managed to get a case heard it wouldn\'t likely go far.  This is scary. I talked about this over a year ago as it was something that I was frightened he would do. Scary scary scary stuff.  So welcome to the new Fascists States of America.  Ummm :-/  Linky Links    CNN\'s Coverage    Remember they are asking the Justice department, the very same group of crack lawyers who said torture is legal! and the Geneva convention was \"Quaint\". Somebody help us.  '),(1439,1234212,'07,April,2003','2003-04-07 00:00:00','   Yes I do have a lot to say about the whole Sandy Berger thing...there is a lot more to it than meets the eye.  Peace  '),(1440,1234212,'07,April,2003','2003-04-07 00:00:00','   Today has been another crap day but I have gotten a few more things in order. My grandmosther wants me to get antourh place. Partly because I am after all 27 and I should be out of the house by now. So I guess I will have to start thinking about that.  As for the Averatec I think that I am laying the groundwork for one and possibly another. One ultra-portable (my current system), another micro system with MS XP Tablet Edition (a kick ass spec\'ed convertable Tablet PC...having one of those would just be fun). The third is the desktop replacement system. It has 3.0Ghz P4 HT Enabled, ATi 9600 64MB RAM, 512MB RAM, 60-80GB of HD Space, DVD+-R/RW/CD-R/RW combo drive. It seems that Averatec has decided to enter each of the mainstream markets, ultra-portable with the 3200 Series that I have (AMD XP-M 2000+, 256MB RAM, 40GB HD, S3 Omnichrome Graphics, DVD/CD-R/RW, 802.11G, and much more for a grand), the afformentioned 6100 Series for Desktop Replacement, and now the 3500 Series for TabletPC\'s based on a AMD XP-M 2200+, 512MB RAM, 80GB HD, DVD/CD-R/RW drive, 802.11G, and other goodies...  They seem to want to offer products for any occassion.  Well I am tired and tired and tired...so g\'night.  Peace  '),(1441,1234212,'07,April,2003','2003-04-07 00:00:00','   I must update...I must update...I must update. Hey I am reminding myself to update...and soon!!!  So here it goes.  I found out what the specs&nbsp;are in&nbsp;the New Averatec C3500 TabletPC Model.  AMD Athlon XP-M 2200+ (1.7Ghz) 512MB RAM up to 1GB 60-80 GB HD DVD/CD-R/RW drive Full keyboard (covertable style) 12.1 Inch Screen 4 USB 2.0 Ports 802.11G Microsoft Windows Tablet PC Edition Up to 4 and 1/2 hours on a charge.  Very decent specs...and it will start at $1350...so a kick ass deal! With the one I have and if it is any indication that these little units are good (and so far this one is kick ass) then I think Averatec just helped make TabletPC a viable alternative for students finally. If a product would have been available when I first headed off to class for that price I would have jumped all over it.  I am going to jump on this one as soon as I can because it is just kick ass. Averatec is also looking into doing a Centrino Laptop with high end graphics (NVidia 5700U Mobile or ATi 9700 Mobile both with 128MB RAM, a mid range Dothan Processor, all crammed into a small laptop like I have...or in a more traditional 15.4\" setup. They need to do a compact with HDTV setings 16*9.&nbsp; That way there will be no wasted pixels for playing movies on these little machines.)  I will say something about the performance of this little beast. It runs warm...you can tell. It isn\'t super hot but could burn you if you are not careful. The battery life that I see is generally a little better than what is usually reviewed. I get around 3.5 hours of medium use with a Cell Phone attached drawing power for internet service. The keyboard is fairly responsive although a little cramped. I like the click feel for them indeed. It has an annoying clunk that comes from the HD from time to time...it is the thermocoupling fluid that is re-adjusting...or whatever it is called.  Well I passed out for a about half an hour and it is getting close to the next set of drugs...so I am going to zoink out some more.  Love to all...and hopefully&nbsp;a handy addition to my Laptop stable...and then after tablet from Averatec...possibly a gamer from Averated Desktop P4 @ 3.0 HT Enabled, 512MB RAM, 15.4\" screen, 60-80GB HD, DVD-+R/RW/CD+RW, ATi 9600 64MB RAM, and some other bits...unless they get the Dothan System out....at and at that price point with higher graphics Ati 9700 128MB or Nvidia 5700U 128MB....if I knew for sure they were going to do that then I would go save for that and get it when it hits because it will be a hot portable, gaming ready. and just overall goodness.  Peace  Let me drop this old RAVER saying on ya\'ll.  P.L.U.R. (Peace, Love, Unity, Respect)...if we all showed it and praticed it daily the world would be a better place...a lot more polite at the very least.  Blaaaaa  '),(1442,1234212,'07,April,2003','2003-04-07 00:00:00','   Today I had the most horrific experience that I think a person can have...well maybe not the worst but it damn well ranks up there.  &nbsp;  I went down to the GI lab today to have my PEG tube placed a little further down in the GI tract...well out of the stomach and into my intestines. Now to do this they have to scope you and look down there to guide things. Now to do this generally they put you into Twilight sleep which is not fully out but not awake, and certainly not aware.  &nbsp;  Well I already had a bad experience with the one asshole, now I have another bad experience. This is the doc (dick) that came to look at my tube over the weekend. It has been leaking since I got the damn thing...or rather stomach contents have been leaking out from the incision where all the mess goes into the belly.  &nbsp;  Well we decided to put the tube a little deeper out of the stomach into the intestines to do two things. First to get the food from the tube into the intestines hopefully bypassing the \"full\" stomach trigger so I can eat and tube at the same time so that I can get full nourishment. Second I can tube feed for a few days to let the incision heal up properly and hopefully keep the food in my stomach from dribbling out around the incision.  &nbsp;  Well to do all this they have to twilight and tube you...well we got part of it right.  &nbsp;  I got down there with this asshole who worked on me today and he knew that I was nervous and I told him to make sure that I was pretty much out and all would go well. I also had told him I have higher tolerances to most narcotics and other drugs...which I do.  &nbsp;  Well I was wheeled into the little procedure room and he started to give me the sedative. Starts with 100 of Fentinol...well it comes in these carpi-jet tube things that you have to crack and setup...well if you do it right all the med makes it into the IV...if you do it like he does it 50% of it ends up on the patient and the other makes it into the IV...so for 100 units he only got in maybe 50 at best...so I was underdrugged. He was in a hurry and not paying attention.  &nbsp;  At some point I wake up during all of this...now it might have been when he yanked my old tube out and I started screaming bloody murder...because it hurt. I remember them tubing me...and I couldn\'t breath and had a coughing fit plus a panic attack...the manhandled me...or rather he did...it was just a big fucking mess and I remember balling because it hurt so bad. When I came a little more too in the recovery area I was already sobbing because it hurt so bad so I woke up or came too crying. I don\'t cry from pain often...but when I do it is because something really hurts.  &nbsp;  While I was down there I was wiped out and when one of the nurses came over to me I said to her \"did what I think happened really happen to me?\", she told me \"if you ever come back down here make sure you only have Weinberg or Dubner, never the other two.\" I had two other nurses down there tell me the same thing. Basically I woke up in the middle of the procedure and instead of stopping or getting the situation under control before continuing he just plowed ahead and basically beat me up.  &nbsp;  When I got back to my room my mom was here and she asked me what happened because I was bruised and beaten looking. I was beaten up pretty bad. I recovered quickly and I am sore and swollen but I am not to much worse for wear except that I cannot believe that is the second doctor from this service that I have had that have caused me harm. There is no reason for all of this. The only thing that can be on this one was that I was the last patient (which I was) and he was in a hurry...plus he had someone observing. So he might have been showing off and in a rush. Hey...slow down and do it right.  &nbsp;  So when I see my Doc tomorrow I am going to bring it up because I don\'t want others to go through that.  &nbsp;  I try not to complaign to much about things because I don\'t want to be a whiney bitch about things. Unfortunately this time I am going to have to be a whiney bitch.  &nbsp;  Well I have some drugs coming to me now so I am going to do them...veg. and hopefully enter tomorrow with the hopes that it will be better.  &nbsp;  Peace  '),(1443,1234212,'07,April,2003','2003-04-07 00:00:00','   I am sitting here watching \"Secret Window\". One of the respiratory care people brought it in for me to see. That is probably the biggest saving grace of all this mess is that the people here make me feel as part of their extended family and help as much as they can. They constantly offer to bring me food or movies, always asking if there is anything they can run and get for me. It makes me feel good that at least I am doing something right somewhere.  &nbsp;  Of coarse I rarely take them up on the offers because I don\'t want to put them out, although I know that it wouldn\'t be a bother for them at really. The thing that I do take people up on is the movie thing. If people have a DVD that I haven\'t seen and they offer I always take them up on it because I love watching movies. I will say that there is one disadvantage to this laptop and the smaller screen size. It is harder to watch a movie and do other things on it at the same time. It can be done and I am doing it now but it is cramped. The other thing that I really need to do is to get this thing upgraded to 512MB of RAM. The 256 works well but the graphics subsystem does pull from the main memory of the unit. So right now with 256MB of RAM installed I pull 16MB for the Graphics system and that leaves me with 240MB of RAM available for program use. It is an odd number and some programs that I would love to use have to have at least 256MB. The biggest problem with being short of RAM is that the hard disk spins more and the computer just has to work harder. So 512MB of RAM is in the near future. I also have to save for the Averatec Tablet PC...that looks like it will be a good system. I still have yet to find out the specs for sure on it but for 1300 bucks you cannot go wrong at all. Pretty soon I may be sitting here writing my blog entries and watching a DVD on this machine. If I do get the Tablet PC then I will not upgrade this one...I will instead make sure the Tablet PC is maxed out with RAM as it would be more usefull there.  &nbsp;  Oh so much to see and do...plus I would love to build a new desktop just to build one.  &nbsp;  I would use the majority of the components that I already have like the high end P4 running over 3Ghz, 1024MB DDR500 RAM, 120GB SATA HD, 60GB PATA HD, good video card, DVD Burners (I have two...w00t! but I want to move to a Dual Layer DVD+-R/RW and I can do that for 90 bucks), WiFi 802.11G card. I would roll all those into a newer system maybe based off of the Aria case and a Micro-ATX motherboard. I would love to add a HDTV tuner card from ATi, a much better sound card (perhaps a high end SoundBlaster Audigy2).  &nbsp;  I would love to add 2*74GB Western Digital Raptor 10,000RPM in a RAID0 Array and that would be hella fast, plus the 120GB SATA as a backup drive for videos and music. I do have my eye on a system from Shuttle, it is a new one that has the new P4 packaging, new chipset, and it will accommodate 3HD\'s, one optical drive, PCI Express graphics and a HDTV card. However that would mean I would have to get the two HD\'s, new Proc. (at a slower speed than what I have now because to get where I am now in Proc under the new packaging would be expensive, plus I would have to get an ATi X800 video card for the PCIe...that is major moola...that is why I am leaning towards the Aria case, and a Micro-ATX motherboard, and just add bits as I need them. Down the road I could use the case and upgrade to a board with the newer features and such.)  &nbsp;  That is all off in dream land of what I want to do...if I did all that&nbsp;I would love to save and get a big LCD flat panel (like around 30inches HDTV) so I could compute and work on it but it would be good for TV here in the hospital and watching movies...not to mention kick ass gaming. I would love to do all that because once I get through all this shit I could use it in an apartment and have a hell of a setup.  &nbsp;  I have so much I want to do...and I need to get moola to get there. Its just a lot of dreaming right now. However I will get somehow, somehow the Tablet PC...that would just be cool as hell...also Averatec offers a good mainstream PC 15.4 widescreen, P4 3.0Ghz (Hyperthreading enabled), 512MB RAM, 60GB HD, DVD+-R/RW drive, ATi 9600 Mobility 64MB RAM...Basically a good solid portable gaming machine. That one costs about 1300 as well. So if I could I would get two new laptops, the gaming edition and the TabletPC one...if I did that I would ignore altogether the desktop system with maybe a tansplant of cases to something a little more roomy with a better MOBO room for my DVD Burners, lots of HDs...and have it on a network so that I can use it as storage and brawn setup or something...I dunno I am rambling...really really rambling.  &nbsp;  This is a shit post I am sorry. I am bored...so much I want to do but not the energy or seemingly at the moment the mental acuity to do anything...  &nbsp;  BLAH!!!  '),(1444,1234212,'07,April,2003','2003-04-07 00:00:00','   Today has been just a crap day and that is all I want to say about it. &nbsp; I hope I can have some nice days where I feel good and have the energy to feel good. &nbsp; That is all...I sat there for a few minutes to see if anything would come to me and low and behold not a damn thing did. &nbsp; Shit.... &nbsp;  '),(1445,3808913,'30,June,2004','2004-06-30 00:00:00','   mic check 1 2 3...  '),(1446,3808913,'31,July,2004','2004-07-31 00:00:00','   Whiskeytango  wow. last night\'s gig went great. but my guitar has weird stage fright issues. i tune it right before i go on, and the second we hit the stage, it falls out of tune like \'whoa.\' and i\'m like, why are you doing this? yeah, that\'s the picture... me soothing my guitar, giving it a pep talk before we perform... lol. good stuff. i tried out some new songs and was very happy that they were received well. i was even approached by the girl who runs the thing and she invited me to be a part of some compilation cd they\'re putting together. so that\'s vedy vedy cool. just another thing to put in my press kit. woot woot.   Calypso Joe\'s of Vestavia  wow. my sticker is officially on a girl\'s VW car. how cool is that? so is she considered my first fan? tonight\'s gig was nice, but i was strangely calm... or was it lazy? the songs lacked the energy i usually put forth... who knows. maybe i was too comfortable. but the venue was nice... it was outdoors and you couldn\'t ask for a more beautiful night.   22nd Street Jazz Cafe  i am exhausted. i didn\'t get home til 1:30am from last night\'s gig, and had to be at work at 6am. blech. i feel like poo. thank god today is friday. i\'ll have the weekend to recoop. last night went really well... i met an  amazing  guitarist, eddie warren ( urlLink www.eddiewarren.com ). it\'s rare that you\'re in the presence of an actual phenomenon. i sat there, jaw on the floor, during his entire set. jesus christ, i\'ve never heard anyone play the guitar quite like him. if there was ever any one person who is  one  with a guitar, it would be him. he approached me after i stepped off the stage and handed me his CD. \"i was wondering if you\'d be interested in collaborating with me...\" \'\'me? what do you need me for? you\'re a one-man show. you\'re brilliant. you\'re the  word ,  brilliant .\"   *mental note: don\'t be such a blubbering idiot when in the presence of a force of nature*  '),(1447,3808913,'26,July,2004','2004-07-26 00:00:00','   say you\'ll come?   Tuesday, July 27, 2004  Whiskeytango 2738 18th Street South Homewood/Birmingham, AL 35209 205.868.4645 8pm   Wednesday, July 28, 2004  Calypso Joe\'s of Vestavia 700 Montgomery Hwy Birmingham, AL 35216 205.978.6761 7pm   Thursday, July 29, 2004  22nd Street Jazz Cafe 710 22nd Street South Birmingham, AL 205.252.0407 9pm  '),(1448,3808913,'19,July,2004','2004-07-19 00:00:00','   Well the Atlantis Music Conference came and went... it was loaded with helpful information and invaluable networking opportunities. But it was hard not to feel overwhelmed. There is so much to think about, so many different routes to take, to decide on, to choose from, as an independent artist. It was overwhelming, hard not to feel lost in a sea of hopefuls. But invigorating to know there are so many options out there.  I\'ll be in the studio the 2nd week of August, which I am very very excited about. I can\'t tell you how excited I am to finally start working on this project. I can\'t wait to see what comes about.  '),(1449,3808913,'19,July,2004','2004-07-19 00:00:00','   Come out and see me play my guitar and sing and shtuff... &nbsp; Friday, July 23rd, 2004 11pm Smith\'s Olde Bar (The Atlanta Room) 1578 Piedmont Ave Atlanta, GA (404) 875-1522 &nbsp; say hello and I\'ll give ya a neat gift!  '),(1450,3808913,'08,July,2004','2004-07-08 00:00:00','   I will be playing the infamous open mic at Eddie\'s Attic in Decatur, GA. Please come by and say hello! More info can be found at www.eddiesattic.com  I\'m posted to play around 8:30pm on Monday, July 12th. Hope to see you there!  '),(1451,3808913,'07,July,2004','2004-07-07 00:00:00','   will be back in the studio faster than you can say \'house of wood.\' I plan on starting late July into August. I hope to release the record into the world late August/early September. Can you say \'Jen is thrilled\'? Yes yes yes, makes me vedy vedy happy. And, in addition to the original tunes, I plan on adding two bonus tracks for your lovely listening pleasure. Now tell me, who loves ya baby?  '),(1452,3808913,'01,July,2004','2004-07-01 00:00:00','   this past weekend was bliss wherein i lived music. eat-sleep-breathe. we are very happy. to the Nth degree. yes we are. which makes the daily grind much more jarring to return to.   met with both producers/engineers in \'nashvegas\' and we hit it off. musically, creatively, socially. it\'s inspiring and emotionally thrilling to find people who share the same creative ideas. they both quickly grasped the vision of the record, which makes me doubly excited to start rolling tape. (or rather, \'spinning silver\' - is that what one would call it? since we\'re in the digital age?) both were so kind and hospitable, and seemed eager to start working. having people who are passionately involved in the same creative product is just... god. it\'s such a humbling and awe-inspiring feeling. i am just so thankful. so thankful. it really makes me wonder, why? why would anyone want to work with me? it\'s amazing. it\'s the word, amazing. and i can\'t can\'t can\'t wait.  '),(1453,3808913,'02,August,2004','2004-08-02 00:00:00','   more dates to catch me live...  08/02/2004 07:30 PM -  Moonlight Music Cafe  628 Montgomery Hwy, Birmingham, AL 35216 -  08/03/2004 09:00 PM -  Innisfree   727 29th St South, Birmingham, AL 35216 -  08/04/2004 08:00 PM -  Calypso Joe\'s  700 Montgomery Hwy, Birmingham, AL 35216 -  08/05/2004 07:00 PM -  Highland Coffee Company  2255 Highland Ave South, Birmingham, AL 35216 -  08/09/2004 09:30 PM -  Eddie\'s Attic  Decatur, GA 30030  '),(1454,3983686,'29,July,2004','2004-07-29 00:00:00','   THE 3\'S COMPANY QUIZ!     &nbsp;   Thats right, I made a 3\'s company quiz..a) Im on the verge of being an expert on the show and not afraid to say it..b)Im an insomniac and always getting to brush up on my info..and c) everyone needs a hobby..    so if youre interested and youd like to play along for the fun it.. join in! If youre here to poke fun at me or anyone taking it please excuse yourself out the back door..Ill post the answers in a few days. Copy and paste this to an email with your answers and send them to  urlLink Jaime_sher@yahoo.com  and Ill post them right away.. unless of course you dont want me to.. but come on its for fun..enjoy!     &nbsp;   1) Please list the Full names of the following characters..Jack, Janet, Chrissy, Terry, Cindy, Mr and Mrs Roper, and of course.. Mr Furley..     &nbsp;   2) Name the same characters\' occupations..(bonus points for any extra info provided on all questions..for example Im only asking for the main jobs but any extras you think of is great)     &nbsp;   3)What is Chrissy\'s full First name... bonus points for remembering how the public was given that fact..     &nbsp;   4) Who played \'Greedy Gretchen\' one of Jacks many female interests..     &nbsp;   5)How did Jack meet Terry?     &nbsp;   6)When Chrissy \'left\' the apartment, what was the reason on the show that she was gone?     &nbsp;   7) Who came first ..the Furley..or the Roper?     &nbsp;   8)Who had spin off shows and bonus points for the names of those programs?     &nbsp;   9)What was Mr Roper arrested for?     &nbsp;   10)What was the name of the local bar?     &nbsp;   11)What tenant in the building (not roommate) was obsessed with Jack?     &nbsp;   12) Who in turn was obsessed with that person?     &nbsp;   13) Who broke their leg on the stairs in front of the apartment?   (left or right leg for bonus points)   14) Who took that person to the hospital?   15)In the episode where the 3 took&nbsp; in a homeless elderly man, try and give what his name was.. first or first and last is sufficient..   16) What did he do in Jack\'s bedroom?   17) What was the name of Janets fiance at the end of the show?   18) What was the name of the building that they were offered a penthouse in, but it ended up not existing cause the man was a conartist?   19) Which of the roommates, for a short time, assisted in singing the theme song during the opening credits?   20) What was the first name of Chrissy\'s father?   How did ya do? Thanks for filling it out!!   COMING SOON:  FULL HOUSE QUIZ  COSBY SHOW QUIZ  WHO\'S THE BOSS? QUIZ..  Also taking requests for quizzes youd like to see done!!   '),(1455,3983686,'27,July,2004','2004-07-27 00:00:00','   Hi Everyone, I think Im in the mood to ramble.. so if no one minds I think I will. Well no one (except beth as of recent) comes in here anyways.. so what the hey.. Here are some things that have been on my mind...  For one thing.. is it ever going to stop raining?? I mean I saw what appeared to be construction of an ark in my neighbors yard.. not to mention I thought I heard some goats and horses. Last night my son came into my room swearing he saw a unicorn prancing around the backyard.  Why do we always look fat on video? They say the camera adds x amount of pounds..oook..why? Is it jealous of us for being in the limelight so its like.. *insert deep scary voice and cryptic music* mwuhahahaha NOW I take my revenge..* Why cant they make the food thats bad for you taste gross and the healthy stuff better? Am I the only one this seems backwards to? Last night, a man decided he wanted to be on&nbsp; my icq list and not knowing if it was the nickname of a friend that I didnt recognize...I went ahead and added him..we chatted briefly and I decided I wasnt comfortable with the conversation and frankly told him Im quite gunshy with the internet and dont volunteer personal information or anything of the like.. he was annoyed with my defensiveness and wanted to know why I had the internet.... wtf.. am I supposed to be content with him telling me hes not nuts to just take it at that..OOOO EVERYONE IM A %FT$ TALKING AND DANCING SQUIRREL!!CALL ME SANDY CHEEKS!!!!!!!! *dances around and shakes tail* PLUS, did I miss the memo that said computers were invented to go with a worldwide web solely created for chat with strangers to make the lives of psychos easier as they hunt their victims... Ill glady&nbsp; dust off the welcome mat on my *front door* as I continue to live in the dark ages.. thank you very much.   Jaime Besides that.. did I miss   '),(1456,3983686,'23,July,2004','2004-07-23 00:00:00','   Well, mine isnt and I think Im a little more tired of it today than usual.   *thinks* ahh whats the point of posting.. no one comes in here anyway...   '),(1457,3983686,'22,July,2004','2004-07-22 00:00:00','   Im bored. NO other way to put it. I mean dont get me wrong, Ive been finding things to do.. like taking care of my sister after her operation yesterday.. But I feel like something substantial is missing. My mind moves at 1000 miles per hour but lately its filled with trivial bullshit.. know what I mean?      For example...Ive seen my share of tv lately, and on vh1 I caught this show explaining meanings to songs...     Remember that song.. you oughta know..by alanis morisette? ya know the one! going down on whoever in the theatre and i hate to bug you in the middle of dinner.. and each time i run my nails down someone elses back I hope you feel it.. get it now? I think weve all smirked once or twice to that song.. right? Well theyre claiming that she wrote that song personally about an ex.. (you might have even seen this episode already..) they said its about Dave Coulier.. (Joey From Fullhouse) I was like !!!!!!!!!!!!!!I find this hysterically funny.. am I the only one? I mean COME ON.. snarly intimidating Alanis \'dont fuck with me\' Morisette... goin down on Joey in a movie theater?&nbsp; Was he laughing like Popeye while she tickled his nuts???&nbsp; I mean thats like picturing my rockstar brother hooking up with that black lady with the shower curtain on her head that does the commercials for Honey Bunches Of Oats!&nbsp; Funny shit.. Ill leave it at that.. give me your take on it!   '),(1458,3983686,'21,July,2004','2004-07-21 00:00:00','   DAY 2....    &nbsp;  Hey again.. if youre back for more than THANK YOU for taking the time to come check things out and congratulations to the mysterious Nicholas for being the first one here and to post! You win 1 dozen chocolate donuts! .oO(thats what elves do in their off time...!) This is just going to be a short message today cause its been a long day.. Im still welcoming posts and comments.. you can really rant about anything you want.. and if you think youll use it that often Ill just send ya the password to post it yourself.. or just go ahead and email me at  urlLink Jaime_Sher@yahoo..AHEM  SIS!!    &nbsp;  Love yas and blessings..    &nbsp;  Jaime   '),(1459,3983686,'20,July,2004','2004-07-20 00:00:00','   Hey everyone! So glad to see you find the place.. It seems everyone these days is having a blog, and its true I have more than enough things to say.. so why&nbsp; not jump on the band wagon? As most people know, I reign from the message board days, where my friends and I frequented the tub clubs and various hot spots, and it was a real blast. I certainly hopes this catches on, cause Im basically trying to recreate some of the laughs we had back in the day.. Im gonna invite anyone to toss in their 2 cents about whatever, or funny stories.. just please join in..leave your comments..good or bad.. and come back often.. who knows this may even turn into a journal of sorts for the lot of us.. whatever it is.. lets take advantage of it...  &nbsp;  Jaime  &nbsp;  p.s. Happy Birthday, Kendra!!!!!   '),(1460,3983686,'15,August,2004','2004-08-15 00:00:00','   Well howdy all... been awhile since I rambled.. the quiz was a major flop it seems.. sorry to take up good blog space.. maybe I can come up with something more appealing to any of my loyal bloggers...but not tonight... tonight is a night for sad dribble..  Im sure most of you all know that my brothers band, seventh hour, has decided to call it quits.. and its just symbolic to me.. When this band started WAY back when they were Allstar it was such a new thing to me for so many reasons.. its was a time of my life when so many things were changing.. it was a like a door was being blown open.. for those of you that know me well Im sure you remember. Anyway it was totally exciting and I wanted to be a part of it in any way I could and I was really thrilled when I got to \"come along for the ride\". I was living with my brother at the time, who was also managing them, and they were practicing in his basement, and it was there that not only did I start many long lasting and cherished friendships, but I met my husband. We became fast friends and got to know each other through being a part of the Allstar experience.. him being a drummer for Searching For Gladys at the time, and had some great laughs and great shows til we finally got around to dating and the rest is history.. we\'re married with kids and tried to make as many shows as we could.. and usually were put out when we did miss something! I feel as though this band is a part of me cause it represents something so personal that I cant even begin to describe.. and now another time in my life has come where things are doing that blow the door off the hinges thing  again.. but it just isnt in a positive sense.. (one of the biggest things is losing my mother in September) and now seventh hour is coming to a close too. It all ties in together, see? Like life is a big roller coaster.  I could go into more detail cause its not just a chic whining about a band breaking up. Im not a groupie.. its FAMILY.  Look at it this way.. where else can I go to see all these same faces in the same place at the same time.. not to mention their originals are honest to God favorites of mine.. and Im not being partial cause of my brother..  To everyone in and a part of Seventh Hour and even the days of Allstar.. thank you for the memories.. they will be cherished and the days will be missed.. love yas!   '),(1461,3983686,'10,August,2004','2004-08-10 00:00:00','   Heres the answers to the quiz.. not that anyone was really interested...    1) Please list the Full names of the following characters..Jack, Janet, Chrissy, Terry, Cindy, Mr and Mrs Roper, and of course.. Mr Furley..  Jack Tripper, Janet Wood, Christmas Snow, Terry Alden, Cindy Snow, Helen and Stanley Roper, Ralph Furley   2) Name the same characters\' occupations..(bonus points for any extra info provided on all questions..for example Im only asking for the main jobs but any extras you think of is great)   Jack- chef, Janet- florist, Chrissy and Cindy -secretaries, Terry- Nurse, Helen and Stanley- Landlords, and Mr Furley was building manager  3)What is Chrissy\'s full First name... bonus points for remembering how the public was given that fact..   Christmas Snow.. she was asked her full name in the hospital when she got a concussion  4) Who played \'Greedy Gretchen\' one of Jacks many female interests..   Loni Anderson  5)How did Jack meet Terry?   he cut his finger and went to the hospital   6)When Chrissy \'left\' the apartment, what was the reason on the show that she was gone?  she was taking care of her mother..   7) Who came first ..the Furley..or the Roper? the ropers     8)Who had spin off shows and bonus points for the names of those programs?    both the ropers and jack (john ritter) the ropers\' show was self titled.. and jack\'s was called 3\'s a crowd.. referring to himself, his fiance, and her father  9)What was Mr Roper arrested for? he was accused of soliciting a prostitute    10)What was the name of the local bar?  reagle beagle   11)What tenant in the building (not roommate) was obsessed with Jack? Lana   12) Who in turn was obsessed with that person? Mr Furley    13) Who broke their leg on the stairs in front of the apartment? (left or right leg for bonus points) Jack broke his left leg   14) Who took that person to the hospital? Chrissy   15)In the episode where the 3 took  in a homeless elderly man, try and give what his name was.. first or first and last is sufficient.. Leo Moran  16) What did he do in Jack\'s bedroom? He caught Jacks blanket on fire from his cigarette   17) What was the name of Janets fiance at the end of the show? Philip 18) What was the name of the building that they were offered a penthouse in, but it ended up not existing cause the man was a conartist? Paradise Towers  19) Which of the roommates, for a short time, assisted in singing the theme song during the opening credits? Chrissy (Suzanne Somers)  20) What was the first name of Chrissy\'s father?  Luther  thanks for playing   '),(1462,3838040,'31,July,2004','2004-07-31 00:00:00','   The  aphorism  â€œQualities reveal your soul and mind to the worldâ€, though being forged by me, suits every step of every manâ€™s life. Itâ€™s  quality  but not quantity that matters.  For every step one takes or for every action one does, one needs a quality plan and master - mind execution.  He/She may posses a huge amount of wealth, but he/she is said to be wise and rightly characterized, only if he fosters in him, good qualities. The quality of life one leads (or â€˜livesâ€™, may be) depends on the quality of his/her thoughts.  If thoughts are good, then heâ€™ll attain fruition and vice versa.  urlLink  James Allen   has stressed, in his famous book,  urlLink  As a man thinketh   (rechristened for simplicity, as â€œAs you thinkâ€) that :   A dishonest person is not totally corrupt and the honest one is not entirely virtuous.  I go with him. A dishonest person might have many other good qualities. But this one, single bad quality makes him squalid. So, bad qualities buffet you down very much. So, try coveting more good qualities than bad ones. So, quality speaks much of oneâ€™s real way of life, than quantity. Quality excels quantity. One must be qualified enough in the league stage to advance into the super league stage of life. In order to make progress in life, you need good qualities and wise actions rather than showing vice. Good thoughts can provide you a helping hand in  shaping  good qualities. (I know I am  iterating  this statement). One must shirk deceptive attitude and proceed with an  unerring  mind.  Iâ€™d like to complete with James Allenâ€™s :   As we think in our hearts, so are we.   I completely agree with him. How about you?  '),(1463,3838040,'29,July,2004','2004-07-29 00:00:00','   Rolling my eyelids towards sky,  did I see an a logy round object,  lacking exuberance, twas todayâ€™s  moon. Like ours, itâ€™s character amends  everyday. One day â€¦gloomy,  dull, the other dayâ€¦bright someday.  Itâ€™s moon, we admireâ€¦.  adds beauty to the sky, reaving  sunâ€™s light, brightens the slobbish,  night sky, proliferates the beauty  of the nature. Neither a profligate,  in luminosity, like the sun, nor a  miser alike darkness. Itâ€™s something  neutrally forged by Him, may be.  Resembling our life style, may be, not to some..  Fortnight later, it disappearsâ€¦.  Fortnight later, becomes effulgentâ€¦.  Like, some days very gloomy and,  dormant, in oneâ€™s lifeâ€¦.  some days bright and galeâ€¦  Life is nothing but a cycle of  eventful happenings, may be,  similar to moon, a cycle of  brightness and darkness..  '),(1464,3838040,'27,July,2004','2004-07-27 00:00:00','   It was well said by Rousseau â€œMan is born free but everywhere he is in chainsâ€. What Rousseau meant, according to me, is that man is leading a free and unrestricted life.   But from the moment man has entered the civil society all his freedoms and liberty have been finished.   Yes. Manâ€™s life in the civil society is an artificial life. It is a false and untrue life. Man in the society becomes highly formal and sophisticated. He becomes greedy and wicked. The life in the civil society is not natural. It is a pretentious life. Here man follows double standards. He becomes civilized as wells as a hypocrite. He says one thing but does something else. Is it not?   In the society man is everywhere in chains. I think, everywhere there are a set of doâ€™s and dont\'s. The social, martial and governmental laws bind man. Thus a man cannot be free in the society. These restrictions, restraints and limitations are the real obstacles in the path of his progress and development of his talents.   Man should try to shed the artificial aspects of his life and try to live a life in harmony with nature, which is more natural and genuine.   Your opinionsâ€¦â€¦. ?  '),(1465,3838040,'25,July,2004','2004-07-25 00:00:00','   Mind becomes dull at times....  No thoughts.. No ideas..   For the past three months, I am away from \'studies\'. That\'s forms the main reason.  Yes. Liveliness has gone out of my mind. Very dull... indolent....  No progress being made. Still standing in the same stage of life...   \'Inhaling\' and \'exhaling\' are the two principle jobs in my daily life.. No proper itinerary for a day\'s work.   Yes. I think I should amend it. I should activate my mind otherwise it will be prone to decay. I think I\'ll have to be peart and exuberant.   One\'s mind should never go idle. It has to be effulgent enough. Otherwise life would become a dull stroll towards death...  '),(1466,3838040,'23,July,2004','2004-07-23 00:00:00','   He has attained whatever he requires, whatever he craves for, &nbsp;uncaring  about the consequences. He is someone extra-ordinary. Yes.  He discoversâ€¦. inventsâ€¦  But still insatiable. Wants more. But â€˜no problemsâ€™ does he feelâ€¦  Coz, heâ€™s bestowed with something unique, compared to others, to achieve and attain his longings... He is overreachingâ€¦ Never stopsâ€¦That insatiability prevents him from ceasing. He hurtles in the field of science especially. Very fastâ€¦.  He goes on and on and onâ€¦â€¦  He ideates&nbsp; â€˜itâ€™s possibleâ€™. He is imbued with optimsm.  He has outflanked all others. He is a maven in all events. Yes. He rules every peck of this globe. Itâ€™s his regime.  He is â€œmanâ€. Man is great.  '),(1467,3838040,'21,July,2004','2004-07-21 00:00:00','   What is life? Oh! Atman  answer me! what is life?  is it a process? or is it  a creation by anybody?  Or is it the palpitation of  something? or is it a  world of tranquility? or  is it a cavern? or is it  a volcano of indignance?  Or is it a pitch of  gratification? or is it a  progeny of something? or is it  a pool of savageness and  viciousness? or is it a vial for  bloodsheds? or is it an aura of  superstitions? or is it a residence  of serenity and wisdom?  '),(1468,3838040,'19,July,2004','2004-07-19 00:00:00','   \"Hm.. What to do? Fate wills the other way.\" \"Have no ways, but to got by fate.\"   Yes. These words pull down your good attitude, your fine brain, your brave mind and your courageous heart. Huh!   \'Fate\' is something that doesnt exist (It\'s my opinion). It has been forged by us. It\'s fictuous after all. Every walk of Life depends on you and your mind and not on your fate. Every action you do, every step u keep, is the decision of your mind.   Ya. We\'ve seen people blaming FATE as the first two lines suggest you. I, too confess, I do that. But I feel, it\'s not right.   Try to be antagonistic to fate..  coz.. It would depress your will. Would degrade your performence.  Try achieving, what you think. Forget about fate. Don\'t let fate govern you.  Believing that fictuous fate, will levigate your ambitions, your goals etc. Try overcoming fate.  You have the power within you to change according to the world, to adjust according to the circumstances. Bring that out...  Have positive attitude. Focus on your goal. DENY following fate.  Go on... Attain what your inner soul wishes and your dreams can come true... You\'ll attain fruition...  '),(1469,3838040,'17,July,2004','2004-07-17 00:00:00','   Hmm...  Life is boring..  No plan. No work. No mind. No effort. No activity. Nothing. What to do?? Can any one suggest me?   Yes. I have finished my +2. And from April 1, I am free. I was elated for somedays that burden was all over. But nah......   It\'s 3 months since I worked something genuine. Life has become a bullock cart.. Doesnt flee. But crawls....   Very slow..   Mmm.. Yes. Had a friend here, who used to regale me. Gone..  he has gone somewhere..   Days are spent sleeping...  Tried to refurbish my house. Had no interest, but.   College might start somewhere in the middle of august. Still 40 days more..   I have become indecisive. Sometimes, I get ireful &amp; irascible, thinking how to spend days. Every hour crawls.. but no idea, do I get on how to spend.   I am living a rural area. No entertainment. No amusement. Nothing.  Hm.. Even to get a nice pen, got to travel down 15 kms. Huh!   No fine friends in this proximity.  I am helpless, desperate..  Trying to find a solution to quell this inactivity...  Huh!  '),(1470,3838040,'15,July,2004','2004-07-15 00:00:00','   Reclining on my easy chair, I  was relishing the beauty of,  the mauve Daphne, Ah!  what a fragrance does it emit!!!  As sun was giving way for,  the dusk to rise up, it was  welcomed by the plashing  leaves of the rangy eucalyptus,  whose pliable branches ostracized,  some of the leaves, it owned, to the,  mud. Why? Why? Was, enmity, the  reason? Or is that tree atrocious?  Or, are the leaves ragged? â€¦â€¦â€¦  ..a plethora of thoughts and doubts,  traveled through my mind. So I,  got riled and tried to cease,  probing on that. I rolled my eyelids  towards the sky and glared at the  stars. Inconspicuous.. and at the  moon, very jejune and melancholic  was it.. Bowing down my head, I caught  sight of a cat, with its bright eyes,  staring at me from the facade of,  my bungalow. Trying to hide my,  Fear, I replied it, with my fervid,  Sight on it. Raising from my seat,  I stepped on a grasshopper unknowingly.  Oh ! Oh! What a crime was that ?  Oh! What a mistake was that? Why?  Why should I do that ? What made me  do that? Why did I disturb the eirenic  nature? Oh! Why did I disturb your  creation? Tell me! You divine deity,  Tell me! Tell me!, I cried desperately.  '),(1471,3838040,'13,July,2004','2004-07-13 00:00:00','   Easily pulverizable will  be your imperils when,  your helping hands extend  on to the unaffordable...  Serene will be your,  days and happy will  your nature be, when  life is an unerring light,  and joy, its own security.  â€¦.Eternally eirenic will  be your life, if you are  benevolently, amiably  characterized. Eccentricity,  abounds when and where  strength of character abounds.  Life is a machine and  character its input,  output relies upon  character. Remedy for  atrociousness and  irascibleness is serenity,  which gives you a joyous,  plight less and peaceful  life to breatheâ€¦.  '),(1472,3838040,'11,July,2004','2004-07-11 00:00:00','   You people know, he\'s great.. No question arises there.  He is really great.. Vow.How melodious will they(his music) be? He loves music, Very boisterous personality. Works hard..Day and Night...  Dusk & Dawn(unlike me..). Never does he get annoyed. Very calm &amp; serene. Very Natural person..   My admiration.. Yes. So many films & albums. All touching your soul and my soul too.  You\'ll feel every bit of his music. I felt. One in one lakh might be a flop. That too is rare.  Is that Music, which he composes??  Nah, it\'s something superior than that.  He is divine.  Indolent? Does he know that word. I don\'t think so..  Very hard working.. No easy goings.. Huh!  He improvises every minute rather than getting satisfied with the first attempt.  Very eirenic will his music be. Yes. Can even make the SUN rise in the west.. This is no exaggerations.. Mind it.   Today, I was hearing some of his old tracks. I was lost u know..  Hey, Try quelling your goose - pimples, if you can, while you hear his compositions. Right?  His music gives us a taste of his superiority &amp; love in Music.  Hey, Music comes from his heart and not from his mind.  But, u know, he is a simple person, but very realistic in his profession. Really god - gifted huh!   Ya. He is none other than  urlLink A.R.RAHMAN , the man who can make you imbued with joy with his m..u..s..i..c.  '),(1473,3838040,'10,July,2004','2004-07-10 00:00:00','   A buffer for want of death, Climbed an arÃªte with deep breath, For he had the first heart attack That shook him with a shock. A rivulet, under the morning sun, Had itâ€™s serene run, By the wayside producing a fetid Smell creating a sense of trance, for It had many odd smelling fishes Ascending and descending Down the rivulet. The smell made the old man vomit, Made him lie down the summit. Rising up after a deep sleep, Continued, the old man, his creep. As the dusk rose up itâ€™s head The buffer had him fed, As the temperature had itâ€™s descent The old man erected his tent. As the dusk changed to dawn He made his journey on, Reached the edge breathtakingly And glanced at the height wistfully. The old man, getting terrified, Hurried down the arÃªte pacified. As he descended down the arÃªte in peace He overcame the smell of the rivulet with ease. Realizing his foolishness to die, He traveled his way back home.  '),(1474,3838040,'09,July,2004','2004-07-09 00:00:00','   Every man in this world is a critic. If a critic is dull-witted, he naturally exalts the value of the work of literature. He magnifies the value of work. In the same manner some glaikit people admire what wise men approve. A Critic must possess a vast knowledge of worldly affairs without which he cannot criticize. If we do not posses the necessary intellectual equipment we will not be able to estimate the true worth of anything. In order to make a true assessment of anything we should posses sense, wisdom and mental balance. So we should not plainly admire anything. We should probe on it to ensure if it is true to be admired. We should be able to see a thing as in itself it really is. There is the need for some sense and a set of values to measure the worth of anything. We should develop in ourselves, a practice of looking deep into matters, probing on it, trying to understand it without any flaw and then admire it if it is right or rise out your opinion on that if you feel that it doesnâ€™t seem to be correct, instead of blind admiration. There is a good criticâ€¦  '),(1475,3838040,'08,July,2004','2004-07-08 00:00:00','   Plodding through a very thick, forest, a boy was relaxing his mind, for he had been hustled out from his dwell for something. Trying to forget his loneliness, he did sing songs of various kindsâ€¦ which made him feel a bit relieving. Striding, through the thicket and  through the gusty gale for a  considerable length of time, he got exhausted and sat down relaxingly. What am I going to do for the night? How am I  going to manage the day? What to doâ€¦â€¦.he was confused. Juddering with intimidance,  he searched for water to have a gulp of it. Oh! Water, Where are you? Arenâ€™t you humanitarian? Arenâ€™t you benevolent?.. After a  desperate search, fate gifted him with a small pond alongside. Noticing  that, he quickened to that pond and gulped the impure amrita and forgot the world, till he was awakened by the plashing shrubs and cuckoo birds which hinted him that, it was dawn. Continued, the boy his plod to infinity, suddenly he was haunted by devils. Oh! Help Me, I shouted, that brought my mom and dad from their rooms, to my enquiry. What made me go to the forest? I questionedâ€¦. Oh!â€¦ What a nasty dream was it?  Very incredible, Oh divinity, What made you haunt me like this?  May such dreams discontinue  in my life..Oh! Creator.  '),(1476,3838040,'07,July,2004','2004-07-07 00:00:00','   This was the topic given to me for the state level essay competition which was hosted by THE HINDU jointly with THE UNO,  which was held a few months back, and I was positioned third.  ------------------------------------------------------------------------------------   Thoothukudi â€“ An Introduction:      Tuticorin   or more recently rechristened as Thoothukudi to match with its Tamil equivalent is the city, I would like to write about being quite close to the village by. This city is also known as the   Pearl City   (or MuthuNagar As we call it in Tamil) being the hub center for collection of pearls from the deep sea. This city has played a major role in the freedom movement against the British rule. From the port of Tuticorin was the first ship, which sailed with an Indian flag atop under the captainship of great freedom fighter V.O.Chidambaram Pillai. The Kattabomman brothers, Vanchinathan who have created history with their valor have lived around this port city. Tiruchendur, one of the six abodes of the Hindu deity is about 40 kilometres from Tuticorin. Since 1960â€™s the new port of Tuticorin has come into being and Tuticorin now is considered one of the major ports in our country.     How liveable is my city today? What can I do to make it more liveable?:    Well now let me come to the moot question â€œHow livable is my city?â€ Before that let us perceive the criteria that make a city livable. First things first. One of the basic needs namely shelter should be available for its citizens at affordable prices. In Thoothukudi almost everybody has a shelter, even the poorest having holed up in hutments.   The city is free of sky scrappers  , the houses are little cramped in the city proper, but well laid out around the suburbs where quite a number of colonies having spring up during recent years. The cost of purchasing a land and constructing houses is definitely on the increase but still reachable to middle classes. The Port Trust colony, the Thermal Nagar colony, the SPIC Nagar colony, the Heavy Water Plant colony, the DCW ltd. Colony, the Tac Nagar colony are townships built by the concerned organizations with all necessary infrastructure. One thing about which Thoothukudi can be proud of is the almost   non-existence of slum dwelling  . The educational facilities in and around Thoothukudi meet the needs of the children in this place what with a plethora of primary, secondary and Higher Secondary schools. Catering to both Tamil and English medium aspirants the colleges around dishing out degrees in Arts, Science, Engineering, Medicines, Fisheries, Law, Home Science etc. and thus Thoothukudi is firmly entrenched in the educational map of Tamilnadu. Thoothukudi is a Railway Terminus from where we have access to all corners of our country, it has a non-functional airport (I for one am not able to appreciate the lethargy on the part of civil aviation authorities in providing fuelling facilities for the aircraft, the availability of which should make this airport agog with activities), buses both government and private run ply to the nook and corner of the city and we also have the waterways, the passenger ferry service to Sri Lanka. Though the major roads are well maintained the ones inside the city are full of potholes and need the attention of the civic authorities for carrying out repairs. Traffic congestions do take place despite the best efforts by the police department to divert the lorry traffic to the periphery.  There is no proper drainage system in place and the Buckle Canal needs to be thoroughly desilted for allowing proper flow into the sea. Drinking water to the people is supplied through municipal taps and tanker lorries but the demand far outweighs the supply and here is one area where efforts should be made for conserving water by digging out ponds and diverting the Tambiraparani river water, rain water harvesting etc..   With growing population of both men and motor vehicles, problems of pollution have started cropping up. Though the levels of oxides of Nitrogen and oxides of sulphur in the air may not be as high as in the metros or bigger cities, still I feel the switch over to CNG (Compressed Natural Gas in Delhi should apply to all cities and use of diesel for vehicles should be banned. Littering of plastic wastes is a menace in itself (for heaven sake do not brand me as a campaigner of Anti plastic lobbyists and I firmly believe the plastics are very much need for the sake of improving our national economy and no effort should be made to ban them. Let the Green Peace movement look for different pastures!) and the two-bucket garbage disposal scheme namely one for the plastic and the other for non-plastic must be in place. Here again the municipalities have to be woken up from their deep slumber and made to do what they are supposed to do â€“ keep the city clean.  Job opportunities arenâ€™t many given the kind of recession the country is undergoing through even though we have a few big industries around. However these are enough opportunities to star oneâ€™s own industry especially in the processing of seafoods, fishing, salt and related products. The SIPCOT complex at Milavittan provides the necessary infrastructure.  Thoothukudi is a place of varies culture and where communal harmony is quite prevalent. The annual festival of the church of Our Lady of Snows, where a golden chariot is hauled up attracts everyone the Christians, the Hindus and the Muslims alike.  Regarding health care, apart from the Government Hospitals, there are quite a few private ones which have the latest facilities and one need not venture out of town for medical attention except for major surgeries or problems connected with heart. Almost all major banks have their branches around the city and the latest facilities like ATM (Automated Teller Machine) is being introduced. Cell phone facilities are available (The city police would do well to stop its use while driving) at affordable rates.    Conclusion:    In conclusion I would like to sum up by saying that the major problem confronting Thoothukudi is the supply of drinking water and as already mentioned earlier rain water harvesting must be made mandatory in all buildings. Digging of reservoirs for storing water (the big industries like SPIC, Sterlite, DCW, the Lions clubs, the Rotary Clubs can help in this regard) and diverting the Tambiraparani water before it meets the sea can be thought of. The two-bucket scheme for garbage disposal should be implemented and the schools around should take a lead in educating the people. The Buckle Canal should be disilted and encroaches around this area should be relocated elsewhere. A citizens committee can be formed to monitor the functioning of civic bodies and the areas of concern like poor road maintainence, sanitation can be addressed to. Should these be done, I believe, Thoothukudi will be a paradise on earth for one to live.  I conclude by quoting Robert Frost:    â€œThe woods are lovely dark and deep, But I have promises to keep, And miles to go before I sleep, And miles to go before I sleepâ€.  '),(1477,3838040,'06,July,2004','2004-07-06 00:00:00','   Renowned are some men,  Who revolve round the sun,  Egocentrically and uncaringly,   Renowned are some other men,  Who are not self-centered,  And breathe for others too.   Significant for their illegal acts,  Are some goons and glaikit people,   Dandiprats are most.   Some of them,   eirenic  ,  Some, frolicsome,  Some of them serene,  And some yearn to be renowned.   Lepers and handicapped are some men,  Who wish for a flawless life,  Desire to be significant,  And will achieve that feat too.   On the contrary, flawless men,  Try to go by fate,  And will blame it, when under trouble,  Difficulty or inconvenience.  '),(1478,3838040,'05,July,2004','2004-07-05 00:00:00','   Trying to slake her thirst, on a hot, sultry summer afternoon, from a  wayside cistern, a girl tried to open the gauzed door, which provided aegis, to that cistern. â€œIs it very resilient to  unfasten? or am I not robust enough? why should you hurt me, trounce me? why shouldnâ€™t you allow me to quench my proliferating thirst? Tell me you divinity! Am I not fit to have a gulp from this cistern? Am I not gifted enough to breathe in this globe? What is spurring you to deny me from taking it? What is walling you to grant me allowance? Are you atrocious? Arenâ€™t you â€¦. humanitarian?â€ The girl began to question the divine as she was reaching the  apex of her dying thirst. Then ultimately, she, out of desperation, tried to open the door and tried it with great diligence, when the door trashed open and peeping in through she could find no water in it. â€All is my fateâ€, she uttered  desperately and went off in tranquility.  '),(1479,3838040,'04,July,2004','2004-07-04 00:00:00','   Started a blog... But hoping and wishing it a long life...Huh!!  EIRENE - my favourite \"word\".. Hope all of you wud like it.. Hope this blog lasts for long..  '),(1480,3838040,'14,August,2004','2004-08-14 00:00:00','   Alas.........The day has come. It\'s time for me to bid good bye to my equanimous  H..O..M..E ..and enter a college environment.. I am leaving my home once for all.. Huh! I am not very happy, nor am I in dismay. But I am not  emotionally  touched. I am as normal as...... I hope and wish that the \" Atmospheric change \" suits me. I wish for an  intrepid  life..Not much of it.. Hope I enjoy the rest of my life towards....  mm.. I\'ve had many friends, be it  e-friends  or direct friends.. All of them have been very mellow towards me. I wish them all \" Happy Life \".. I\'ll still try to keep in touch with all of \'em...   My friends : LIVE A LIFE TILL YOU REACH ...........'),(1481,3838040,'12,August,2004','2004-08-12 00:00:00','   Funny? Sleep well..... I do that very much. \" Sleep Early. Get up early \" - is right. I do the opposite very fine. But I think one must not be restricted from his sound sleep. Right..?  Children must not be forced on a compulsory night study until and unless it\'s necessary.   Insomnia  - Which I sometimes have... Not often..  Somewhere did I read :     The quality of your sleep will improve because of  YOGA . coz, Yoga increases the elimination of toxins from the body.. I am not sure of it, but..... Anywayz.. try avoiding naps and dozes at night... Have a sound sleep..  '),(1482,3838040,'10,August,2004','2004-08-10 00:00:00','   It wasnâ€™t so exciting when I entered the college, which hosted the  counselling . (Here, after 12th class, we use to have couselling for admission in B.E/B.Tech degree for choosing our desired colleges according to our  state rank . It commenced  today .) I wasnâ€™t so enthusiastic. But the excitement was geared up when I entered the hall and it was at its apex when I stepped into the couselling centre.  My all-time favourite use to be  Mechanical Engineering . But according to last yearâ€™s trend, for my rank, I was not in a position to get it (Mechanical Engineering) in the most  top-notch  college. And therefore taking Mechanical Engineering was unpremeditated and was almost ruled out of my choices, I harbored. But to my surprise, when I was reclining at the hall, I noticed quite a number of seats left for Mechanical Engineering in my long desired college ( urlLink College of Engineering, Anna University , Chennai, Tamilnadu) the best one in Tamilnadu. So I made that my first option. But seats were getting into paucity when time progressed. I was becoming restless and petulant and started to be malevolent. Twas my turn. I had just one seat left for B.E (Mechanical) in Anna University (above mentioned college) and   I selected itâ€¦â€¦â€¦â€¦â€¦â€¦â€¦â€¦.   I was very happy and satisfied. College is going to commence on 18th of this month.So, I have got into a very good college. I think Iâ€™ll have to work hard and study well and do my  best  there.  '),(1483,3838040,'08,August,2004','2004-08-08 00:00:00','   Mind plays the most inevitable and important part of everybodyâ€™s life. It fosters \'one\' and oneâ€™s way of life. It shapes  â€˜youâ€™.  One gets, what his mind ideates, what his mind loves. Mind controls oneâ€™s life.   Here comes the theme of  James Allenâ€™s  â€˜As a man thinkethâ€™:     Mind is the master power that molds and makes, And we are Mind, and evermore we take The tool of thought, and shaping what we will, Bring forth a thousand joys, a thousand ills. We think in secret, and it comes to pass â€“ Our world is but our looking glass.  '),(1484,3838040,'06,August,2004','2004-08-06 00:00:00','   Environment will steer it\'s way and we will have to steer towards it....     We have seen people changing their dwellings. We have noticed students getting into hostel atmospheres from their equanimous residences...  \"Environment doesn\'t suit\" - This is a comment that buffets our ears often from the \'shifters\'. But I think environment has nothing wrong to do with them. Has it got any malice of them? Nah. Their thoughts may be erring...  If one\'s mind is abject, it\'s arduous to adapt to the environment. Every thing depends on one\'s \'mind - set\'. One\'s thoughts need to pure enough to sychronise with the environment, one dwells. Then one will find the environment automatically turning towards him/her and benifitting him/her. Make thoughts and mind fit to adapt yourself to the environment. One has to change his attitude and soul in favour of the environment, then you can surely expect the environment to hold you, support you, benifit you and bring bliss in you.....  '),(1485,3838040,'04,August,2004','2004-08-04 00:00:00','   I remember the  4th of July  when I conceived to start a blog and was searching for a good template for it.  Yes. My blog is a month old....  Though I attempted many a times to foster and sustain my blogs, I was thwarted by my mind. None of my previous blogs were enduring and long-lasting. This is the one, which has lasted for two full fortnights and is yearning to last for long. Though the zest, I had at blogging, 31 days back, is missing in me now, I still have the interest in me to sustain this serene and eirenic blog till....................  '),(1486,3838040,'03,August,2004','2004-08-03 00:00:00','   Twas yesterday did I ideate and realise that the way of life one leads is all the result of his/her  â€˜thoughtsâ€™  and the way he/she executes the thoughts. Whatever one does, itâ€™s all because of his thoughts, be it, good or bad.   From now on, I think I have to renovate and refurbish myself. I have to manifest my thoughts and convert them into useful actions and attain fruitful results and live a blissful lifeâ€¦.      \'Good thoughts can bring you good results and bad thoughts bad results\'  â€“  urlLink  James Allen  .  '),(1487,3838040,'02,August,2004','2004-08-02 00:00:00','   People do have interests in many acts of life. â€œInterestsâ€ â€“ Ok, one can have. But that interest must not go uncontrolled.  Got to confess. If I get some new games, I get addicted to it. My whole dayâ€™s work gets spoiled and marred. But now I could realize my folly and am in to my effort to control addiction on any act I do.  \'Interests\' sometimes may give a defiled way to ADDICTION. Addiction results in bad consequences, not good results. Addiction, in every way, affects one and can make him bewail for it. A person may love â€˜eatingâ€™. But once he has turned into a glutton, then his health may spoil.  Reiteration â€“ Have â€˜interestsâ€™ but not greed. If you have â€˜greedâ€™ avoid over-greed.  Addiction can impair oneâ€™s plans for work.  Alcohol addiction - A very familiar one. â€œAvoid drugs and alcoholâ€ â€“ a recurring statement. But people shirk this statement and then realise while they suffer. Late realisation has become common these days.. Addiction can mar you down. It can have direct influence on oneâ€™s health. It can affect oneâ€™s character.  A mellow minded can turn bestial, ignoble and sulky. Such an effect it has got. Oneâ€™s whole itinerary gets spoiled.So, try avoiding addiction or else it can make you a tyrant of your mind and body.  '),(1488,3838040,'01,August,2004','2004-08-01 00:00:00','   World comprises of good and evil. â€˜Goodâ€™ and â€˜Badâ€™ friends. This condition arises only because of the character, those friends have fostered. Had they fostered good character, they join the \"Good Friends\" group and vice versa.  Friends do play a vital role in one\'s life. They too have a hand in the shaping of our character.  Good Friends will help you in your empowerment and development and you know how bad friends would be. Yes. Theyâ€™ll be cynical, malevolent and will mar down your plans for life. There is a saying â€“ Unsinkable ship is friendship. But it has to be amended as â€œUnsinkable ship is a good friendshipâ€   Try to befriend a good friend, but never tease or revile the bad ones . Thatâ€™s the motto for â€œGood Friendsâ€ group.  Good friends make world, a mansion to live in....  '),(1489,4142807,'20,August,2004','2004-08-20 00:00:00','    CHANGES       I still don\'t know what I was waiting for     And my time was running wild     A million dead-end streets     Every time I thought I\'d got it made     It seemed the taste      was not so sweet     So I turned myself to face me     But I\'ve never caught a glimpse     Of how the others must see the faker     I\'m much too fast to take that test          Ch-ch-ch-ch-Changes     (Turn and face the stranger)     Ch-ch-Changes     Don\'t want to be a richer man     Ch-ch-ch-ch-Changes     (Turn and face the stranger)     Ch-ch-Changes     Just gonna have to be a different man     Time may change me     But I can\'t trace time          I watch the ripples change their size     But never leave the stream     Of warm impermanence and     So the days float through my eyes     But still the days seem the same     And these children that you spit on     As they try to change their worlds     Are immune to your consultations     They\'re quite aware      of what they\'re going through          Ch-ch-ch-ch-Changes     (Turn and face the stranger)     Ch-ch-Changes     Don\'t tell them to grow up and out of it     Ch-ch-ch-ch-Changes     (Turn and face the stranger)     Ch-ch-ChangesWhere\'s your shame     You\'ve left us up to our necks in it     Time may change me     But you can\'t trace time          Strange fascination, fascinating me     Changes are taking the pace     I\'m going through          Ch-ch-ch-ch-Changes     (Turn and face the stranger)     Ch-ch-Changes     Oh, look out you rock \'n rollers     Ch-ch-ch-ch-Changes     (Turn and face the stranger)     Ch-ch-Changes     Pretty soon you\'re gonna get      a little older     Time may change me     But I can\'t trace time     I said that time may change me     But I can\'t trace time     The above lyrics go out to Shawzies and pooh bear love ya\'s      Urlzie '),(1490,4142807,'16,August,2004','2004-08-16 00:00:00','    Within You    How you turn my world You precious thing. You starve and near exhaust me.  Everything I\'ve done, I\'ve done for you. I move the stars for no one.  You\'ve run so long. You\'ve run so far. Your eyes can be so cruel, Just as I can be so cruel, Oh I do believe in you. Yes I do.  Live without your sunlight. Love without your heartbeat. (Falsetto) I... I... Can\'t.. live.. within.. you... '),(1491,4142807,'11,August,2004','2004-08-11 00:00:00','    Hello to all my loyal subjects!!!  how is everyone doing? hope ya\'s all had a great day!!!  personally i had a good day, got to go to school late and even played soccer with da hot male teachers (al and melk) but then was told i lit sac the next day!!!  Why spring shit like that on me... life is so unfair!!!! and j. dickhead is still pissing me off, when i direct a qn specifically 2 someone u dont jump and answer it.. its rude!!!  oh well i guess i should stop havin a whing about a lil insignificant spec like her and just enjoy life....   well must be off, im never really sure what to write in these diaries- everything just seems so trivial when i read it back....  love urlzie '),(1492,4142807,'11,August,2004','2004-08-11 00:00:00','    Hey dudes,  Just completed 2 sac a math and an english one !!!! My brain is completely fried. Oh Well I think i did alrite in the math one, but the english well lets just see....  I\'ve decided to have a h/w and work free nite... im going to pamper myself later...  oh well hope everyone is as well as i am!!!  LOve Ya All  Urleen!!!!  PS: i hope shaw is feeling better soon!!! '),(1493,4142807,'07,August,2004','2004-08-07 00:00:00','    Hey,  Didn\'t do much today, just went to few uni open days, they were all really basic. So now im back and really bored!!!!  I guess i\'ll lets ya\'s all know a little about me, Im 17, irish and just love to have fun. I got great friends and fantastic enemies. What really ticks me off are ppl who lie, who are snobby and just really selfish ppl... but then again I love to be around funny and intellectually stimulating ppl.  My favourite colours is green/yellow Lucky number is 38 favourite music is a mix of robbie william, matchbox 20 and a twist of old style - elton, bowie, elvis etc.  Currently in year 12 and loving school, altho i can\'t wait to get out, i wanna travel- see the whole world. Especially Eygpt, Peru, Guatemala, Rome, Paris, Germany all those types of places- i cant wait   Oh well lunch is ready  must go  love Urleen '),(1494,4142807,'04,August,2004','2004-08-04 00:00:00','    Hello Everybody,  Hope ya like wat i write and can respond to it easily....  love ya  URLEEN....  P.s this is was inspired by my good friend shaw moore, from her entry entitle Reuniting!!!! LOVE YA SHAW '),(1495,4075308,'30,July,2004','2004-07-30 00:00:00','   Hi guys!!     &nbsp;'),(1496,4075308,'11,August,2004','2004-08-11 00:00:00','   see also:    urlLink  http://mostlyfiction.com/contemp/haddon.htm       I donâ€™t do this with every book I read. Itâ€™s what you do in primary school; choose a book and take it home. Bring it back the day after and write a review. This isnâ€™t really a review; More an account.  Iâ€™m writing about this one though, because although fictional, there is a great sense of triumph at the end of the book. Written by Mark Haddon and winner of the Whitbread Book of the Year Award (JUNE 2003), this book tells the story of fifteen year old Christopher Boone, and his struggle to find out the truth about Wellington, and his mother.  Christopher suffers from Asperserâ€™s syndrome. He loved lists, patterns, numbers and the truth. He hated things that are brown or yellow, crowds, changes, and being touched. He believed his mother died of a heart attack, which is what he was told by his father.  He lived a happy, settled life, until he discovered his neighbourâ€™s dog, Wellington, dead. He began a search to find the murderer of the dog, and began to write a book about it. In his book, he includes mathematical facts, and his view of other peopleâ€™s minds, and the way the brain works.  This book stands out for me, because there are so many times when you feel truly sorry for him, and heâ€™s the kind of person youâ€™d want to meetâ€¦ if he were real. His disability causes him to think like an infant; innocent and immature. Yet he is able to cope in most situations, so long as certain rules are applied. His food must not be mixed. People must not touch him. Even his own parents arenâ€™t allowed to hug him.  I think our lives are like that sometimes. We donâ€™t want everything to be complicated, and get annoyed when people are clingy. And yet some of us are bored so easily.  For me, Haddonâ€™s book is a story of a childâ€™s personality, and the way it works in his favour, to achieve the goals he sets for himself.  Thank you Mr. Haddon. Looking forward to the next.'),(1497,4075308,'02,August,2004','2004-08-02 00:00:00','   \"On the day that you were born the angels got together, and decided to create a dream come true. â€™   Since I can remember I have associated this song with my happy childhood. As a child I felt closer to my Godmother, my fatherâ€™s sister Aunt Rosemary, than I was my own mother.  We used to sit in the back room, the nicest room of the house, and listen to â€˜ Carpentersâ€™  records. I canâ€™t recall a time when I couldnâ€™t sing those songs. Even now I can picture us sat on the pink fuzzy carpet, warm and cosy, singing happily away. â€˜ Itâ€™s yesterday once more .â€™  The house was huge, and I had my own room, blue, with a small window above the door to let light in. Every night I stayed at the house, my Aunt would remind me that she would never leave the landing off over night. She knew how much darkness did and still does scare me.  I was a weak child, scared of everything! The Hunchback of Notre Dame was incredibly frightening, and one night I woke after having a childish nightmare. I ran, crying, to my Aunts room, and blubbed that I couldnâ€™t sleep; I was having bad dreams of Quasimodo. â€œWell, you better sleep in here then.â€ After getting me settled beside her she told me a joke, one which I never have found funny: â€˜Mrs. Modo, is Quasi coming out to play? â€™ â€˜No, heâ€™s got a bad back!â€™ Still it was the way she handled things; always looking on the bright side, always knowing how to make it better.  The house was, in my opinion, the most beautiful house in the world. The park outside it was â€˜mineâ€™ I was convinced, and I loved the fact that it was three floors. The attic was huge, and I can remember my aunt making curtains on the big machine table up-stairs.  Downstairs had three living rooms, and a kitchen. The second floor had four bedrooms and a bathroom. My older cousin Robert, once locked me in the hall, and being young, I was too small to reach both the door handle and the light switch. I bawled my eyes out, and hated him for years.  Then, sometime in 2002, I went to tea at my Aunt Rosemaryâ€™s. By this time, weâ€™d moved away from Wortley, and were living in Drighlington Bradford. Aunt Rosemaryâ€™s new house was smaller, but it had two bathrooms, four bedrooms, a sitting room, dining room, kitchen and conservatory. My mother and I sat in the room with my aunt, while she told us that Rob had been sent away to Afghanistan, to fight the war on terrorism. He would be gone four months.  I wrote to him. I told him I was sorry for hating him all those years, and how I was scared for him. Since leaving Leeds our family had drifted apart. Today we never sing happy songs.  As you grow up, you realise changes you never saw before. All of a sudden the world is a scary place, and the people you trust and respect stop lying to you, and you learn to face up to what the world is really like. Suddenly I could really relate to poetry I never understood before. Especially \' Out   Walking With My Father\' , by Paul Bitrill. It made sense. Nothing stays the same.  When I was about 12, I asked Aunt Rosemary why we never sang those songs by the Carpenters. She simply laughed. \"Karen Carpenter starved herself.\" What did she mean by that? She died so that\'s the end of it? It never stopped us when I was a child. But then I guess I wasn\'t old enough to know really.  Ever wish you was a kid again? Yeah Karen. I really  do  wish it was yesterday once more...'),(1498,3421075,'29,May,2004','2004-05-29 00:00:00','   One well known guy has this to say about another even famous guy.  \"....The first version of Linux was like a time machine. It went back to a system worse than what he already had on his desk. Of course, he was just a kid and didn\'t know better (although if he had paid better attention in class he should have), but producing a system that was fundamentally different from the base he started with seems pretty good proof that it was a redesign.....\"   Guessed right?   Andrew Tanenbaum on Linus Torvalds.  More at the article on his homepage: http://www.cs.vu.nl/~ast/brown/   '),(1499,3421075,'28,May,2004','2004-05-28 00:00:00','   There are 10 kinds of people in the world... Those that understand binary, and those who do not   '),(1500,3421075,'26,May,2004','2004-05-26 00:00:00','   Check out the  urlLink Wired 40 . Infy moved from 34 to 11. Kudos! Guys, thats too good and a real India Shining Campaign. Anyone willing to discuss!  Shed the Comments   '),(1501,3421075,'24,May,2004','2004-05-24 00:00:00','   urlLink JoG  Writes \" Ijust got back from a road trip through the UK and India. I got to meet up with lots of Java developers. Lots of interesting folks doing lots of cool projects. The high point was at the end of one talk in Hyderabad one of the conference crew brought a birthday cake on stage (it was my birthday) and they had 3000 people sing happy birthday to me. Quite overwhelming. Thanks, everyone!  The peace movement is quite strong everywhere. I took this photograph of a chalk drawing on a street in London. Folks aren\'t exactly anti-American: they\'re anti-war. Everyone was hugely excited about Michael Moore\'s new movie. I\'m a very liberal Canadian living in America: I got lots of questions. \"   Reza, shared a very interesting note on his Sun Tech Days Experience. The post made me feel the missing out of the event. :(  JoG\'s Birthday on that day is a news. and BTW, Michael Moore\'s Fahrenheit 9/11 seems to have wom the Cannes Film award.   '),(1502,3421075,'23,May,2004','2004-05-23 00:00:00','   If the builders built building the way programmers wrote programs, then the first woodpecker to come along would destroy civilization.   urlLink Phoe6   '),(1503,3421075,'23,May,2004','2004-05-23 00:00:00','   Hello World!    This is the blog version of our Yahoo Group  Start Tech Blogginn   Adios!   '),(1504,3421075,'24,June,2004','2004-06-24 00:00:00','   More you dig deeper,more you are amazed by what was it like... Was like it Google, or more , or lesser? Was like the real turner? Yeah it was.  \"Tim Berners Lee created the WWW, but who put it on the Boards of the Trucks and Lorries?It was Netscape\"  Scott (from Mozilla) urlLink SSL    urlLink Javascript  (Enjoying Gmail??)   urlLink Mozilla  come from Netscape and also this guy  urlLink jwz  , urlLink this guy   and lets add more ... when come across. There seems lotta hackers and geeks at Netscape during its yore.   '),(1505,3421075,'23,June,2004','2004-06-23 00:00:00','   A Simple One!    Longfellow\'s Bees       The Poet Longfellow in his novel Kavanagh, introduced several clever mathematical problems from an ancient Sanskrit work. One of them is as below :  If one fifth of a hive of bees flew to the ladambaflower, one third flew to the slandbara, three times the difference of these two numbers flew to an arbor, and one bee continued to fly about, attracted on each side by the fragrant ketaki and the malati, what was the number of bees?   '),(1506,3421075,'22,June,2004','2004-06-22 00:00:00','   The most intriguing figure in 19th century cryptanalysis is Charles Babbage, the eccentric British Genius best known for developing the blueprint for the modern computer. He was born in 1791, the son of Benjamin Babbage, a wealthy London banker. When Charles married without his father\'s permission, he no longer had access to the Babbage fortune, but he still had enough money to be financially secure, and he pursued the life of a roving scholar, applying his mind to whatever problem tickled his fancy. His inventions include the speedometer and the cowcatcher, a device that could be fixed to the front of steam locomotives to clear cattle from railway tracks. In terms of scientific breakthroughs, he was the first to realize that the width of a tree ring depended on that year\'s weather, and he deduced that it was possible to determine past climates by studying ancient trees. He was also intrigued by statistics, and as a diversion he drew up a set of mortality tables, a basic tool for today\'s insurance industry.   '),(1507,3421075,'17,June,2004','2004-06-17 00:00:00','   A Ride into the World of Philosophy. Some intro to  urlLink Sophies World . This is a book that proceeds along with intresting dialogues between a young Norwegian girl Sophie and a fifty year old philosopher monk Alberto.   '),(1508,3421075,'10,June,2004','2004-06-10 00:00:00','   urlLink http://www.iit.edu/~shartan/tagore/crescent.html    Observe the above url clicking twice or thrice. Add ur comments if any.   '),(1509,3421075,'09,June,2004','2004-06-09 00:00:00','   A number puzzle from the Guardian Newspaper U.K. - (As seen on a website)  This puzzle was given to my class in a Functional Programming lecture at uni. It is listed as originating from the Guardian but I\'m sure I\'ve also heard it on a BBC Radio 4 programme - Puzzle Corner I think.  A 14 digit number is to be constructed in the following way.   It starts with the digit 4 and ends with the digit 3.   It contains two 7\'s, separated by 7 digits, two 6\'s, separated by 6 digits, etc.  What is the number??  (Guardian,7 December 2002)   '),(1510,3421075,'08,June,2004','2004-06-08 00:00:00','   Just now was glancing through a few pages of our president\'s book \"ignited minds\".  In this passage he categorically states that Aryabhatta was THE first to give an approximation to Pi. The passage goes this way    \"Aryabhatta, born in 476 AD in Kusumapura(now called Patna), was an astronomer and mathematician. He was reputed to be a repository of all the mathematical knowledge known at that point of time. He was only twenty-three years old when he wrote Aryabhatiyam in two parts. The text covers arithmetic, algebra and trignometry and of course, astronomy. He gave a formula for the areas of a triangle and a circle and attempted to give the volumes of a sphere and a pyramid. He was the first to give an approximation to pi as the ratio of a circle\'s circumference and diameter, arriving at the value of 3.1416. to celebrate this great astronomer, India named its first satellite launched in 1975 Aryabhata\"   '),(1511,3421075,'06,June,2004','2004-06-06 00:00:00','   Synopsis of the latest book from Arthur.C.Clarke goes this way       Based on the recent sensational proof of Fermat\'s Theorem 350 years later by a young British mathematician, Andrew Wiles, THE LAST THEOREM charts the story of Ranjit Subramanian, a man fascinated by Fermat\'s Last Theorem - so simple that anyone can understand it, yet not proved for more than three centuries. Ranjit learns about the Indian mathematical genius Ramanujan (1887-1920) and discovers a three-page proof of the Last Theorem: this might even be Fermat\'s own proof. The discovery of the Theorem wins Ranjit the Fields Medal - and the attention of the NSA cryptography branch. However, Ranjit soon finds himself drawn by physics rather than cryptography, as there have been some spectacular advances in fusion technology. And these in turn lead to a plasma drive that can open up the Solar System ...   '),(1512,3421075,'01,June,2004','2004-06-01 00:00:00','   :::GTray:::    GTray  , a software to access GMail account.  urlLink http://torrez.us/archives/2004/05/23/gtray.zip  (52.9kb)   :::Gmail -> POP3 converter:::    PGtGM  sits between you GMail account and your email client, converting messages from the web based mailbox into POP3 messages that a program such as Outlook Express or Firebird can understand.   POP3 server now fully works. You are able to download gmail emails through any email client. SMTP server appears to work.  More info:  urlLink http://www.neowin.net/forum/index.php?showtopic=169789    Download Link:  urlLink http://jaybe.org/pgtgm/pgtgm.zip (122.9kb)   NOTE: You need to have the .net runtimes installed   '),(1513,4045197,'29,July,2004','2004-07-29 00:00:00','    Wow, what day, today was awesome..i didnt have to get up at 6:20 i got up at 11. then got bored and took a nap. haha. i just hung around the house in pain. my stomachs been hurtin since i went to the dr and he told me i had a stomach virus. and my back just started hurtin after i helped my friend britney move into her g-mas house. i can tell you know i dont know how to pick up heavy things without hurtin my self, instead of me pickin it up with my legs i picked everything up with my back. wow....big mistake there. so now i have to suffer with the damn pain. it makes my whole body hurt by sendin sharp pains up and down my back, legs stomach chest...everywhere. its horrible. =(    &nbsp;    tomorrow im goin get my grade for summer school, then prolly gonna get my phone. i came up with a plan cuz my rents werent payin $300 for a phone, which i dont blame em, but this is what were gonna do...my moms gonna upgrade her phone to the phone i want, and then were gonna switch phones. hopefully this will work. if not ima be really pissed off. byrons at demolay this evening for pratice and im stuck at home with nothin to do.     im kinda bored. hopefully he\'ll get home early so we can have some time to talk before i go to bed since i dont feel good.       &nbsp;  mom and i started tannin again, we were startin to turn white again, not tan like we want. hehe. this weekend britney my mom and dad and i are goin to kemah texas for a little get away, byron cant come cuz hes got a carwash to go to...but ima be seein my baby hopefully b4 conclave. =) brit and i are gonna have a busy week, but its gonna be fun. hehe. brit broke up with her boyfriend for good, but i gotta make her stop talkin to him cuz if she talks 2 him shes gonna get back with him and she doesnt want that. im like the mom to alla my friends. haha its kinda fun.     &nbsp;  well, thats enough for now, ill update tomorrow night or something. toddles* '),(1514,4045197,'27,July,2004','2004-07-27 00:00:00','    Hey there, &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Today was a blahh day. i went to school for my last full day it was alright, tomorrows my last day we only go for the test and then we get to leave...so thats gonna be fun. I havent been feeling well still, i still have some of that damn virus in me, and i think i have indigestion to, or at least thats what my mom told me. when i helped britney moved i pulled muscles all in my back and now i have a huge knot in my back and its makin my whole back hurt. =( it sucks big time, so thats been hurtin me a lot...and i go to the dr. 2morrow and this dr. ive never been to before so im kinda nervous. dont know what im supposed to do and all that good junk. but i need to go. i need to get some medicince for this stupid stomach virus and indigestion also. its makin me feel like crap. my immune system really sucks balls...it makes me mad. but theres nothin i can do about that. i mean i never get sick durin the school year, but i get sick in the damn summer time when its hot as a mutha outside. i mean gah, what are the odds.   were still goin to Kemah this weekend, were leaving early saturday morning and comin home sunday sometime. i asked britney to come with me but Jareds comin into town, so i dont shes gonna come, so i might ask nicole. Brians not goin cause hes gotta work...he works as a bouncer @ Yesterdays with his friend Erik. Well, i dont know what else to say so ima go call byron and go to bed since im not feeling well and all. Good Night and Sweet Dreams Everyone.   &nbsp; '),(1515,4045197,'26,July,2004','2004-07-26 00:00:00','    Hey there guys,   &nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Hows everything going. Today after class I went and helped my Britney Drew pack alla her things into boxes and put them in her dads truck and then we went to her new house (her grand-mas house) and unpacked everything that we brought the first trip. Lets just say we started at like 11ish and didnt get done loading,unloading, and unpacking till about 3ish. Then we went to get some lunch cause we were like starvin, and went back to her house and ate it. And helped her dad bring some other stuff to the truck of theirs that was gonna go to the storage room they had to buy. Britney and I stayed at her apartment and started packin some more things in her room. We almost finished but not quite. The only thing is they gotta be out of there apartment by the end of the week (saturday).   Other than that, nothings really happened. I have 2 more days of school left. Tomorrow and Wednesday, im excited. This weekend were going to Kemah, Texas for one night. Were leaving early saturday morning and coming back sunday night, and then the 3rd is Jr. Registration, and then on the 5th we leave for Baton Rouge till the 8th for Conclave, i get to see Byron this weekend but not for most of the time. It sucks, but i gotta suck it up. lol.  Byron and I have new songs their:   baby when i look at you you know it breaks my heart in two how beautiful you are seen you in a million dreams now your finally here with me we will never be apart i wanna hold you forever thats all i\'ll ever need chorus you are my love you are my life my heart and soul the truest friend i\'ll ever know all of my dreams, my fantasy, my reality i love everything you are yes i do everytime i close my eyes it hits me so deep inside how real this feeling is i\'m intoxicated by your touch its a sweet sweet rush i\'m in love with your kiss your the one that i trust the most you changed me chorus 2 times i love everything you are...You Are by Jimmy Wayne.  She thinks I walk on water She thinks I hung the moon She tells me every morning \"They just don\'t make men like you\" She thinks I\'ve got it together She swears I\'m as tough as nails But I don\'t have the heart to tell her That she don\'t know me that well Chorus: She don\'t know how much I need her She don\'t know I\'d fall apart Without her kiss, without her touch Without her faithful loving arms She don\'t know that it\'s all about her She don\'t know I can\'t live without her She\'s my world, she\'s my everything And she thinks she needs me Sometimes she cries on my shoulder When she\'s lying next to me But she don\'t know that when I hold her That she\'s really holding me Holding me Chorus Yeah, now the funny thing is She thinks she\'s the lucky one Chorus She thinks I walk on water She thinks I hung the moon....She Thinks She Needs Me by Andy Griggs.  Well, thats enough for now. ill update this new journal later. '),(1516,3913126,'22,July,2004','2004-07-22 00:00:00','   Absolutism    There are no absolutes for something so relative as a human life.  -Hugh Prather, Notes to Myself  Funny how it seems true when you read it in a book. When I say, at the age of 40, I want to write a book. When I say I want to be a hair dresser for 2 years. When I say tomorrow I want to meet this person, go this place, finish this much work, wake up at this time, and wear this. So how often have things gone the way I wanted them to? And how often have the results been exactly the way I wanted them to be? And how often, when the result has been the way I wanted it to be, have I been as happy as I thought I would be?   Has it happened to you that you wore this outfit that you never imagined yourself wearing 6 months ago, just because you had nothing better to wear, and got more compliments for it than on a day you meticulously planned your dress code? What are the chances of you making more close friends in 6 months of working than you did in 4 years of graduation? Has someone else\'s question ever given you answers and shown you realities that were always in your face? How often have things been perfect? When you do come across things that ARE perfect, how often is it that you planned them? Have you ever been bitterly disappointed that perfection was too perfect and lacked spontaneity, and hence, wasn\'t perfect after all?   So is symmetry perfect? Is asymmetry not beautiful? Isn\'t the now more important? Aren\'t mistakes more important than failures? So would you really like a spate of winnings and no losses at all? Is what I dream to be tomorrow, stopping me from being me right now? Is the life I am leading what I want? Am I living or existing? Is it wrong to want to live? For that matter, is it wrong when at times, you want to give up and don\'t want to live? Is it wrong that you don\'t want to be strong? But if you are, does that mean you are superior? Does that make you a better person? So who is a good person? If a social worker, who helps unknown people improve their lives, beats up his wife, is he a good person or a bad person? Are you a good person? Are you bad? Are you never bad? Does bad mean wrong? If you are being bad in self defense, is that wrong or correct? Did God put you/us on this Earth to make a point? Does he have to? Is there a point to everything? Should there be a reason for living? If you don\'t achieve all that you dream of achieving and are still content, are you a loser because you didn\'t get what you wanted or are you a winner because you don\'t want no more?So are dreams really important? Am I really important? 200 years later there will be no trace of this magazine or the person who wrote it. But the ideas will persist. That makes the thought important, more like significant, than I, the creator or thinker of these thoughts.   I may get two minutes of fame in the future, when I am 40 and dying of lung cancer. Is that worth wasting my youth doing what I don\'t want to do then? Is doing what I don\'t want to do really a waste? If not, am I gaining something here?Should you gain something from everything you do? If you gain nothing, is that an experiment that went wrong? So how can an experiment, where you cannot really predict an outcome, ever go wrong?   If you gain a lot, and you think you gained from a mistake, will you repeat the mistake, even if you know making it the second time would be harmful/hurtful/stupid? But if you do gain, isn\'t the mistake much better than all the corrects you have done in your life?So can you really control your life? Can you say with certainty that at 9 am tomorrow you will wear a red T-shirt and eat eggs for breakfast? So what is definite? Is anything? Does definite give you a sense of security? Is security tantamount to safe? Do all these questions merit answers? Does anyone have the answers to them? Are these questions really questions or statements with a question mark at the end?Does that matter at all?  ************************************************************************************* I am a question mark reincarnate. This is my mind being me. As for the time, I wrote this after I read the book Notes to Myself by Hugh Prather, at a particularly complex and crossroad-ical juncture of my life.  ] d e e p a [ '),(1517,3913126,'22,July,2004','2004-07-22 00:00:00','   Silence of the Crowd  It\'s not enough. Or is it?Flowing water. Curbed thoughts. Expression suspended. Spiraling smoke.An amulet encircling my thumb. Uncomfortably numb.I want more. Or not. Will I stop looking?This is the beginning.Or the end?No I am not stoned.Blinking cell. Blue and white.Disgusting noise. Eclectic psycho-babble.Long hair. Unfinished looks. Bald man; doesn\'t smoke.Spiraling smoke.Crystal ashtray. Pen moving. Thoughts astray. Crumpled tissue..Non blank cheque, enclosed inside. Uncashed yet.Trying to hide or reveal.Noisy water. Crowded train. Beaded nose ring.Not too bored. No motion, of thoughts or feet, or things, living or unmanifest, just water.Silence in words. Words of silence. So who defines natural? Conservative courts? Closed minds?Filthy, rotten books, wormed into by aging decay.I don\'t want to be Jesus. I don\'t want to be an angel.I don\'t want to be anything.  Being John Malkovich. I don\'t want to be him.  ********************************************************************************** I wrote it partly on a crumpled tissue paper and partly on the backside of an ICICI credit card bill envelope. I put down the first thing that came to mind when i saw&nbsp;OR felt anything.   ] d e e p a [ '),(1518,3913126,'22,July,2004','2004-07-22 00:00:00','   The Birthday Speech  I donâ€™t know why Shona has asked me to speak at her birthday. Iâ€™m not good at these things. Donâ€™t like giving birthday speeches. Iâ€™m afraid that Iâ€™ll come out to be too critical or my jokes will fall flat. Iâ€™ve known Shona for such a long time, and there are times I feel that I donâ€™t know her at all.   Shona likes her space. Iâ€™ve never known anyone who likes quiet time so much. She likes her quiet in the morning just when she gets up. She reads a lot. She likes interesting people. She has her mood swings. Does one talk about such things at a birthday party? Remember, I will be competing with the wine for attention.   Shonaâ€™s friends are a mixed bag. There are those that have 9-5 jobs in financial corporations. I met another group some time ago. They were second generation Asians discussing the body as a space for cultural identity. I know that Shonaâ€™s interested in culture. She did a degree in it. She grew up in three different countries. I donâ€™t know what she calls home. She says that sheâ€™s a cultural anomaly. She smiles when she says this. Must be hard not fitting in. Making friends knowing you will have to let them go in a few years. Having to explain your life over and over again to people who nod their heads but donâ€™t quite understand.  I donâ€™t understand these discussions on space and identity. For me space is full of stars and planets. I know that Shona wanted to be an astronaut when she was small. She was one of those chubby Indian kids in England. Her mum used to dress her up in frilly dresses with white knee high socks. If you ask her now, she will tell you that the 1980s was the era of bad fashion. A well-meaning family friend told Shona that she was too fat to be an astronaut when she was 7. Shona lost interest in space after that. Exercise is something Shona caught on to since she started university. Now, sheâ€™s in India and her relatives tell her that sheâ€™s too thin (and too dark) to bag a husband.   Life is full of contradictions. I never thought that Shona would end up in India. I know that she was scared when her parents decided to move to Kolkata into the arms of a conservative family. She moved with them. A concrete jungle for a girl who grew up in a small English town, a tiny rain forested equatorial country, and ended up in the leafy suburbs of Sydney. She misses her green space. She misses her independence. Her snot has turned black because of the pollution. She has also grown up.   There are a lot of things about Shona I could say. I donâ€™t know how appropriate they would be for a birthday. I know that she wishes her clever/quirky younger sister or that her best friend/boyfriend were here. Neither of them will be. One remains in London and the other in Sydney. Maybe thatâ€™s why Shona asked me to give a speech. I sit here racking my brains. Iâ€™ve never written a birthday speech before. What does one say?&nbsp; &nbsp;  ~shona~ '),(1519,3913126,'28,July,2004','2004-07-28 00:00:00','   mr.x: well  mr. yyy: who fell? mr.x: asho mr.x: he fell down some stairs mr. yyy: did he see stars? mr.x: and now he has to sit at an angle mr.x: no idea mr. yyy: oh he died?! angel and all mr.x: hehe mr.x: angles tell no tails mr. yyy: hidden behind their veils mr.x: of sms and emails mr. yyy: from friends and females mr.x: blue blazing express rails mr.x: flying foxes and swimming quails mr. yyy: the runover leaving grissly trails mr.x: crime masters in so many jails mr. yyy: now weeping into their pails mr.x: well watered glass pails mr.x: oye mr.x: pail coincidence mr. yyy: bing. invalid mr.x: people studying at yale\'s mr.x: yale\'s wot u say? mr. yyy: its ok mr.x: chardonnay mr.x: donkeys bray mr.x: fathers pray mr. yyy: fingers in the clay mr. yyy: demi what do u say mr.x: pay day mr.x: yayyyy! mr.x: make hay mr. yyy: drinks in a tray working on a cray mr.x: child\'s play mr.x: blue day mr. yyy: result of the quickie lay mr.x: spank spank mr.x: about blank mr.x: fill \'er say mr. yyy: run through the left flant mr. yyy: flank mr. yyy: lest they say he got afraid and sank mr.x: what\'s the way mr.x: the gun misfired and went \"bank\" mr. yyy: better get a tank mr.x: damp wet drippy dank mr. yyy: ah like castaway\'s tom hanks? mr.x: and that model chick tyra banks mr. yyy: oh she is just a crank  ~j~ '),(1520,3913126,'28,July,2004','2004-07-28 00:00:00','   Thirst - AIC  \"I\'m thirsty. Can I have something to drink?\' he said.  \'Hmmm. Come on in.\'  A minute ago, I had been sitting in my rocking chair listening to Bob Dylan, when I heard the doorbell ring.  I pointed him towards the sofa and made my way into the kitchen. Should I give him water or should I offer him something else? Tea, nimbu pani, litchi juice, tang..? After all, there are different kinds of thirst.   I turned back towards the living room.  \'What would you like to..\'  He was rifling through the top drawer. He swung around.  We looked at each other for a moment before he pulled out a switchblade. Click.   \'What do you..\'  I sat down into the rocking chair, watching my blood gush out of my neck.  Yes. there are different kinds of thirst.  ~j~ '),(1521,3913126,'28,July,2004','2004-07-28 00:00:00','   Frame of mind - AIC  The frame of a train window, a train stalled at a lonely point. A farmer, alone with his ox, tilling the cracked land- a series of electricity grids- the lone signs of evolution, the sky downcast , the air still , the land quiet, streaked by the screeches of two crows.   The bars of the window cast a painful doubt-is this man trapped in a life, with no hope, no future, no happiness, a life of strife and struggle, against the land, the landlords, his will?  The sweat drips off his brow as he stops and scans the horizon. I follow his gaze. In the distance, a woman sprints her way over bunds and bushes, a cloth package in her hand. Soon she is with him and he opens the package, while the woman wipes the sweat of her brow with the sari. The man sits down and begins to eat the food and talks to the woman.   I move towards the door for a better view.  The man smiles at the woman and the train whistle blows and the train begins to move.  ~j~ '),(1522,3913126,'28,July,2004','2004-07-28 00:00:00','   the story of the emaciateddy. - AIC  Long long ago in bethlehem, the holy book says so, there was once a head so bright, could only be brought by the bald head shine, then all the other reindeers, used to laugh and call him names. and then there was the reindeer fodder scam and rudolph was banished to earth. And one lovely monsoon day, I told rudolph, will you ride my sleigh today. And then he rode the auto with the headlight on. And with this story, he will go into history. ~j~ notes: based on an thin auto guy who was almost bald and had his headlight on at 9 am today. '),(1523,3913126,'28,July,2004','2004-07-28 00:00:00','   THE MAN wrote this. I hope this makes your day.  \'Thou hast made me endless, such is thy pleasure. This frail vessel thou emptiest again and again, and fillest it ever with fresh life.  This little flute of a reed thou hast carried over hills and dales, and hast breathed through it melodies eternally new.  At the immortal touch of thy hands my little heart loses its limits in joy and gives birth to utterance ineffable.  Thy infinite gifts come to me only on these very small hands of mine. Ages pass, and still thou pourest, and still there is room to fill.\'  -RT '),(1524,3913126,'28,July,2004','2004-07-28 00:00:00','   I think people write long posts or say long-drawn things really dont know what they are talking about. '),(1525,3913126,'28,July,2004','2004-07-28 00:00:00','   Just a vague line of thought: Stars:Suns::Entertainers:Businessmen  Stars are like these pretty, good to look at things- cheer people up, entrall people and all that, but no real necessity value as as such.  So also are entertainers.  For example: a filmmaker cannot trade films for wheat in a barter system.  Now the suns (ok now please dont start with the \'the sun is also a star\' theory) are like these sources of energy, providing light and sustenance to life, generating more activity in other things. So are businessmen.For example: A business starts a business and provides livelihood for like hazaar people.  Now so would u want to have stars in ur eyes or a sun in ur eyes? '),(1526,3913126,'28,July,2004','2004-07-28 00:00:00','   shine shine shine.  6 am 7 am 8 am  its time to get out of line. are running out of time. make the best of life mine. nay lay in bed sighn,  get up shine shine shine '),(1527,3913126,'28,July,2004','2004-07-28 00:00:00','   j e d i says: yaay j e d i says: i have money in my account j e d i says: Interest also. 0.06 rupees j e d i says: total balance 91.08 INR  H A R I says: you are rich man j e d i says: lets go deposit 9 rupees in the bank j e d i says: then we can remove 100 rupees  H A R I says: hahaha....wait for an hour, spoke to manish he just told me that within one hour it will happen j e d i says: let us await the event with bated breath j e d i says: every moment spent with visions of squandered pleasure and limitless quanities of little joys j e d i says: today let us be thankful of our dear Lord God H A R I says: Mr Alfred has lost it j e d i says: for all the wonderful blessings he has bestowed on us. j e d i says: aye. thou sinner. hast thou forgotten thy past, of hungry days and hungrier nights? of forlorn days, spent toiling and slaving, with brief flitting dreams of rest and comfort H A R I says: what language are you speaking in My Lord?? j e d i says: Archache english. H A R I says: heart ache english j e d i says: ah. that would be best served by the language of love, sanskrit H A R I says: ||astho ma sadga _ma mayar|| j e d i says: erm. Om bhu bhav svaha H A R I says: ||thamsoma jodiga mayar|| H A R I says: ||Mirthormaya amarthamg gamayar|| j e d i says: OOOOOOMMMMMMMMmmmmmmmHHHHa j e d i says: what did all that mean? H A R I says: god is omnipotent,omnipresent and all knowing  j e d i says: other two lines? H A R I says: so i bow my head in respect H A R I says: to that most the all knowing j e d i says: good good. i like ur attitude H A R I says: what attitude j e d i says: H A R I says: so i bow my head in respect '),(1528,3913126,'28,July,2004','2004-07-28 00:00:00','   The parody of life - Ah. Ha. The mystery of life -Eh The astonishment of life -OMG The shock of life -Gasp The birth of life -I\'m cuming The ecstasy of life -I love you The individuality of life -I The duality of life -me too The resignation of life -Amen The melody of life -hahaha The affront of life -Uh The laziness of life - '),(1529,3913126,'28,July,2004','2004-07-28 00:00:00','   why does my hand tremble?  Maybe its the fear that one day someday someplace sometime somehow, ok now enough, write now, right now, the sky will turn. And maybe wink and give a flying kiss. Miss me, it will whisper. For NOW is an neverending tale of hope, soap and grope. Misery will keep you miserly with yourself. So you too will turn. OH, you say. So does the rest of the world as it sits to drink its blues away. I meet the sea and the land, within your reach. The land on which you exist, trying for life to soothe you, soar you into the sky, but you hope, for a better tomorrow, a better future, a better everything,- turning your back to now, and how. Just adding to the tears in the sea. Global warming they call it-if only they knew. One day it will be an earth of tears, of unfulfilled desires, of unchased dreams - terra firma to teary karma.   Reclaim your soul, reclaim your mind, the heart pumps blood, no more, no less.  Energy will stay constant, some this now this, some that, then that. Dreams are like clouds, some fluffy, pretty, some wispy, useless, some nimbus, savour them-for they renew your life. You live, you make good, you make mistakes, you are not wrong, you are not right, these are just conventions. Let them bring structure to your life, but let them not be bounds.   You are who you are, when you are, what you are, how you are. Break your heart, break free of your heart. Read fairy tales, dont believe them, there are no fairies, ever seen one? Dont live on hope. Hope never ends.  Dreams define, life refines. Hold on to the things you believe in, but dont make them your anchors. Anchors dont let you move, they ground you. You were wrong, now make it right.  Will you turn? '),(1530,3913126,'28,July,2004','2004-07-28 00:00:00','   Yellow - bright, happy, cheerful. Make it yellower and it becomes white. White - blank, potential, nothing. Make it whiter and you are an advertiser of detergent soap which is blue. Blue - blue, numb, nice. Make it bluer and you go dead, bloodless which is red. Red - whoa, achtung, rusky. Make it redder and you like a woman author who wears pink. Pink - girly, healthy, rockstaress. Make it pinker. '),(1531,3913126,'28,July,2004','2004-07-28 00:00:00','   I brought silk flowers tonight. Also known as rhodesias/rhodesia/rhodesiums? A walk along the coast on a cool breezy night after it has just drizzled can do wonders for your spirits. They are now in a green wine bottle in water. Tomorrow i have to buy salt and put a little of it in the water.  Apparently it makes them live longer. But arent they already dead.  Salt to get the experience of watching the flowers a bit longer.  50 bucks and a pinch of salt. '),(1532,3913126,'28,July,2004','2004-07-28 00:00:00','   the purpose is lost in its search the objective of an explanatory end insane the paths throughout vitality crowding vision the lack of the knowledge of the ultimate feeling, a pointless exercise in attempting the seclusion of an unknown. life unclear, out of focus, truth based, nirvana ended, a wake of consciousness, the vigour of motion,  the delectable platter of options, the sourness of consequences, the futility of trying to control it, rather than assimilating the outcomes, and savouring the moments. the rush for the next one, trampelling over now '),(1533,3913126,'28,July,2004','2004-07-28 00:00:00','   I realized something about myself today.  Maybe it applies to everyone as well. I dont know and I dont care.  The act of giving. Amrita, a long long time ago, told me to hook up with someone.  She said I had a lot to give; whatever that meant.   I have always given freely; but been awful when it comes to receiving.  I try to be graceful about it, but it never goes down well. Maybe its some need within me to be the role of a giver.  maybe it provides fuel for my big ego.  Recent events where I gave and I received; events that brought about a thought that something wasnt right, not balanced. I could not give as wanted and expected. And i couldnt receive at the same time. Because I wasnt giving. There was no balance between the giving and the receiving.  I often committed to something due to my inability to say no. I eventually ended up not doing those things, leading to various altercations and so on.  Now thinking about it, my sense of giving is one where I give, on my own accord, at my own time. The moment something is asked from me, it loses its purity.  Something similar to my intense mental closure to orders,  especially orders that do not make sense to me. I just clam up.  when the act is original, infinite and boundaryless, I can give and give and give.   just a thought into my nature, the way i think i function. '),(1534,3913126,'28,July,2004','2004-07-28 00:00:00','   Santa Cruz to Malad, NH4, S.V. Road - 070320041430  As you know I am currently trying to live a simpler life. I also need a cheaper mode of transport. The auto fares are killing, and no, I do not wish to take a bus or train.  Also I dont have much money on me to buy a car or even a motorcycle. Plus I desperately need some form of exercise.  4 reasons. 4 very good reasons. So I decide to buy a bicycle- the simplest form of urban self-transport.  A bicycle- I want to ride my bicycle I want to ride my bike.  So I set forth to purchase one. Am still not sure whether to go in for those fancy 5-geared jobs (never ridden them before), a bmx type funky bicyle, or my good old Hero Ranger types (4 years of hardcore riding experience in Bombay and Vellore Roads).  A simple life. Simple transport. A bicycle.  Now those who know me well, or even somewhat, know I am hmm how do I put this- not simple? Me being me, I end up buying a Hercules Topgear- the best of its kinds in India.  Obvious by its name, its a geared bicycle. But 18 gears- 18 different fucking gears.  With two Japanese rotory handgrips that I need to turn back or forward to set the right gear setting for that particular speed, incline or slope.  And life gets complicated all over again. '),(1535,3913126,'28,July,2004','2004-07-28 00:00:00','   someone asked what I would do if I had all the money i wanted in life. here was my answer.  \'travel the world. stay atlleast 3 months in each place. meet people. write stories and features and not care whether the mag/newspapers pay or not. commission short movies based on concepts that i like.  throw some money from Dadar overbridge and watch what happens down. maybe take this to other places in the world. get invitations from world leaders to do the same.  open and run the sound studio, club, coffee shop, publishing house, radio station, street children and old age homes, hire assasins to kill people who fuck the world. make a statue on myself and look at it from time to time and feel humbled. hire people who will make more money for me. give money to all my relatives and snigger and push them around when they are trying hard to be nice to me. clean the rivers and streams and sea shores.   make a movie about myself and pay people to go watch the movie and break all records. watch people go purple with envy and jealousy and haughty indifference and drop a 1Lak bundle with my name on it near them and watch them secretly. if they take it and push, videotape them and show it on national news. if they return it, make them part of my trustees for the trust.  invest in aids and other diseases research. buy 2-3 countries. buy a 5\'5\" by 2\' plot for my remains. write a will with 5 unwritten names as beneficiaries. watch everyone try to be one of the 5.  tie money all along a string that runs around the world.  start my own mint and become a standard against which currencies gain or lose.\'  tbc.. '),(1536,3913126,'28,July,2004','2004-07-28 00:00:00','   the more u learn, the less u feel u know,  the more u want to learn or give up,  the more u go out into the world and into urself,  the more u begin trying to understand the meaning of it all, the more u realize the answer is within u,  the more u withdraw into urself,  the more u want to get away from the world u are trying to understand,  the more u realize that the root of all confusions is within u,  within ur perceptions,  within ur understanding,  the more u withdraw into urself,  and then u die. '),(1537,3913126,'28,July,2004','2004-07-28 00:00:00','   Once there was this young bright mal girl who her father named Jennifer. Because, at birth, her nose resembled a conifer. Her father married her mother. Now her mother too called her Jennifer. Because of all that she had to suffer. We also call her Jennifer. Because, now, she is a duffer (and often a bluffer).   In a distant planet, once runover by apes, now trucks runover apes, a bright exhibitionist studied how to exhibit. Its now an ex-habit. He disliked fools and liked tools and was named Atool. Due to strong eddy currents in the solar plexus, he left his coop of a house and bought a car, a coupe. It was blue. Sadly, one day when everyone had the flu and into their hankies blew, his heart chanced on something new.His heart beat his lungs and he had a thunder stroke. He saw the one, noe, not neo, but the one who was known by many names and by one name all at the same time. The name will be asked for at the end of the Ramayan.   In the same distant planet, once runover by apes, now trucks runover apes, someone had aped an ape using some tape. That someone had flat ears and was vice. He was the jedi. One day, a solar plexus flared and he became Atool\'s best fiend.   There was an old lady. She was a land lady. She wanted to meet her friend, a bag lady. The bag lady didnt want to the old lady who was a land lady. But she didnt know that the old lady wanted to meet her. To make things worse, the old lady didnt know that the bag lady did not want to meet her. There was no power in Eastern America. This is the story of the two ladys.  Jennifer came from the south and stood at the bus stop on Street helm. She was waiting for a bus. A bus was late. She stood at the bus stop. She continued stooding at the bus stop. Freeze frame 1.  Jedi was mad. His wallet was misplaced. He had some change in his pocket. Jedi wanted to go home. He came from the east and stood at the bus stop on Street helm. He was waiting for a bus. A bus was late. He stood at the bus stop. He continued stooding at the bus stop. He ate bananas till he was ready to split. Freeze frame 2.  The old lady wanted to meet the bag lady. She came from the west and stood at the bus stop on Street helm. She was waiting for a bus. A bus was late. She stood at the bus stop. She continued stooding at the bus stop. Freeze frame 3.  A man came in from the north. He saw Jennifer. He saw the long sexy legs. Then he saw the nose. He frowned and looked at the old lady. Freeze frame 4.   Krash, zoop, crackle (tree burning), singe (tree gets wet and has an organism. For some, this is the representation of one stormy night. For others, this is the representation of stormy night. This is known as herd mentality. What they heard from the grape vine. Too much of wine makes people talk. Some grapes, also known as kajoors, whine.   Okay, there was a storm. The man ran across the road to the bus stop on Street helm. He slipped on a banana peel. He banged into the old lady. The old lady fell. Her ankle broke. Jedi threw away the banana bag. Freeze frame 5.  Far far away, there lived a princess. The bag lady smiled for no reason. In the nearby temple, bells rang and tolled. There was no tax. The government was lax. Clinton just played his sax. Monika also played Clinton\'s sax.   Rapidly zoom in and zoom out frame 1 to 5. Show frame 1,3 and 5 within hearts and let the hearts rotate around the scene. Loop background music track of Mein pisal gaya from Koi Mil Gaya. Thaw frames 2 and 4.   Pan to south south east. Atool cruises into view. He glances into the frame. He flicks his cigarrette aside and parks the car to the side. He reverses and parallel parks. He goes ahead and parallel parks. He reverses and parallel parks. Now he is in parallel universes. he comes out of the car and steps towards the bus stop on Street Helm. Behind him, the petrol station explodes. He flicks his coat to a side and comes to the bus stop.   What happens next? Who will Atool take home? Will he drive into the mignight masala? Will he be the saviour of the old lady? Will he be the best fiend? Will he be Romeo naam mera, chori hain kaam mera? What? Why? Who? When? Where?  ~j~ '),(1538,3913126,'28,July,2004','2004-07-28 00:00:00','   The old man was dead. Baba was dead. I stood there, a throbbing ache resounding in my heart. Abdul walked away, a vague smile on his face. Abdul and I hated each other, years of dislike now cemented by my changed fortunes. It didnâ€™t matter anymore. This place didnâ€™t matter anymore. Did anything matter anymore?  I was 8 then. 10 years ago, when the zoo had started in the locality I worked as a shoe-shine boy. To encourage more people and more money coming into the zoo, the management had made the zoo in a unique manner. The ticket counters were situated after the zoo had started. It was meant to entice the potential patrons. A pair of young elephants was caged on each side of the entrance. It was an amazing sight. Two elephants running amok in their huge cages, tearing at the bamboo trees and rolling in the little man-made pools. Once the public had passed this spectacle, the fairly high entrance fee didnâ€™t seem much.   I had run away from my village when I was 5. My mother had run away when I was 4. My drunk of a father ensured that we were fed well with beatings and shoutings. Yes, beatings first, shoutings later. Why? I would never know. All I would know was that I was evil and my mother was a whore. Did washing vessels and clothes at peopleâ€™s houses make someone a whore? I didnâ€™t know and I didnâ€™t care. My back burned. One day in the midst of burning tears, I just walked out and held onto the back of a bus.   The big city was dark, dirty and noisy. A week ago, I spent my evenings at the pond at the edge of my village and the forest, watching the animals drink and play around. Now I was watching others drink and play around. I was hungry, ragingly so. I watched other children beg at the red lights and others polish shoes and others work at tea shops. I tried asking for work at tea shops, but the boys already working there would shoo me away. One of them beat me up. I then mixed with the shoeshine boys. They were a weird lot. One of them, Afzal, the friendliest of the lot, used to smoke bidis and sometimes sniff a white powder. But he was the one who fed me and we shared the same pavement. One morning, I woke up to see that he was gone. That white powder smeared on his nose and he lay there, a serene smile on his face, dead. That is how I became a shoe-shine boy. I had inherited a shoe-shine box from my friend.  The zoo was super. A year ago, the area was this waste dump. I had watched the animals being taken into the zoo. Elephants, tigers, birds, rhinos, peacocks, a giraffeâ€¦ Visions of my village sprang into my dreams ago. But no more nightmares, happy dreams. I slept properly after a year of nightmares. The next day I ran to the zoo. The elephants were up and trumpeting away. I oh so wished I had something to feed the elephants. I reached the entrance and saw that it was 20 rupees. I checked my pockets to find that I had 12 rupees. Dejected, I stood there, and looked back at the elephants. The smaller one stood at the edge of the cage and looked at me, its trunk swaying to and fro. And it smiled at me. Baba used to laugh when I said that to him, but it did. I ran to the fruit vendor on the road and bought 3 bananas. I sped back to the cage and threw two of them over the cage and sat across the road and watched the cage. In a few seconds, I saw the bananas being scooped up by the trunk. Thatâ€™s when I first met Adbul.  tbc.. ~j~ '),(1539,3913126,'28,July,2004','2004-07-28 00:00:00','   A friend asked a question which goes as \' Why do people fall in love?\' This is the theory I came up with: \'because of gravity.  It all started when the apple fell on Newton\'s eye. Thence, the apple of the eye has always been associated with that someone u love. And u must have one of them, because thats how you can keep the doctor away. (Now if the apple of the eye is itself a doctor, well, u\'ve dug your own grave. Keep away from doctors. Who knows, they might be practising on you too!) Now the doctor is important because usually without them, man can kick the bucket. According to studies, man does not like kicking buckets; it hurts, especially if your not wearing any footwear.  Your sole can save you in such a case. Similarily. your soul can save you too. Now, since sex is a human being\'s second basic instinct (the first being self-defense-for the unoriented and non-orientals), your soul too needs a mate.  To sum it all up, if you want to be saved, your soul needs to mate, thereby requiring a soulmate.  Hence you fall in love.\' '),(1540,3913126,'28,July,2004','2004-07-28 00:00:00','   I finally saw it. The heavenly body rested on a shiny platform. The rays of its sun shimmered from the surface. A world rich in minerals and oil, it has been the quest of millions. I approached it in my tri-rotored space transporter. At the same time, a different life form swooped in and landed on a ridge on the body. I made a mental note of remembering to avoid that land. God only knows what other life forms have infested the globe. I got off my transporter. Now I could see the other heavenly bodies all clustered in a divine platter. I raised my index finger and pointed towards the skies. A minute later, I bit into my vada pav. '),(1541,3913126,'28,July,2004','2004-07-28 00:00:00','   If there is a God, good. I need to ask God some questions. If there is no God, good. I dont see life changing anyways. If I am God, I dont have any answers, esp. not for you. Faith is a parallel universe one lives in, a mental link from the known reality to the hoped for unknown. Organized religion is a deal.  Philosophy is a study of the Ideal that ends with tired opinions and compromises on the Ideal. The Ideal is an unknown. Different things for different people, something never really clear. The answers are all there. Learn to frame the questions properly. Do not ask questions unless you can handle the answers.  Nothing is constant. To ask a person \'dont ever change!\' is a stupid thing to ask. everyone changes, whether they or others like it or not.  Life is crazy. And so are you, if you try to make sense of it. '),(1542,3913126,'28,July,2004','2004-07-28 00:00:00','   Sand and a Sunset â€“ AIC  The waves lapped relentlessly over the rocky protrusions of the reef. I stood there, still, the zephyr swirling my curls and a mild, salty mist layering me with the smell of the ocean. The sun was on its way down; another hour and it would be gone. The sky was taking the colors of a psychedelic trip, flushed with hues of purple, blue, and orange. I took a swig at the bottle of mineral water and then let the remaining water flow into the ocean. Maybe the ocean will be less salty and my eyes would not burn as much. I smiled at my own stupid reasoning, and began hopping across the rocks. The sun slid behind a cloud, bringing a sudden darkness around me. I paused, sighed, and walked towards the shore. The recent happenings in my life were drilling a hole into my heart, a constant throbbing pain in my head. He had been pursuing me relentlessly for the last 6 months. I was now at a stage in life where companionship was getting important, but I wanted companionship that came out of love. He believed in it too, but other events in my life were buffeting me.   I often wondered whether what I felt for him was love or something else. How would I know? I knew he loves me from his mind, heart, and soul; and he was ready to wait. He was a beautiful person, but would his love last? Would it be like the beautiful sunset that went away when the sun went down? Would it just be like the way rocks are battered by time and water, and become sand? I had reached the shore; my now-bare feet leaving footprints on the wet sand. A corncob seller was selling his goods at a corner of the beach. I walked up to him and asked for one-with salt and lime. I watched the way he made the corncob and put it in a corn leaf before giving it to me. Some things in life were just beautiful. I turned around, walked to a spot under a coconut tree, and sat. I watched the sunset while relishing the corn. I wish he were here to share the sunset with me. ~j~ '),(1543,3913126,'28,July,2004','2004-07-28 00:00:00','   Yo-Yo â€“ AIC  A spear of hurt and pain streaked through my being. I sat down on the kerb and sobbed, my head a dead weight between my arms. Warm bitter tears wet my sweatshirt burning a path. A passerby slaps me on my back and says something cheering I guess.   Francescaâ€™s betrayal was the last straw in the late decline of my life. All those wonderful movements we had shared together during our loving and passionate romance through our years together at college was marred by the sight of her kissing a man at the cafÃ©.  Francesca had been an exchange student at my university. The moment we locked eyes, we both knew we were meant for each other. Two years of pleasant courtship finally culminated at the airport where I proposed to her before she left back for LA. The most delirious moment of my life, that expansive, eye-twinkled smile saying yes. I had rushed back home to convey the tidings to my rather orthodox parents, wondering how they would react, but hopeful that all would go well.   I remember that night like a horror flick poster. In my exigent desire to share the good news with my parents, I failed to notice the intense gloom pervading the house. My slow, barely composed speech was met without any reaction, if not for a slight drop in my parentsâ€™ countenance. I had felt it would be better in the morning, so I continued walking upstairs on cloud nine.  The next morning, I was startled from my slumber by the incessant drone of an alarm far away. I woke up to find that it was 6.15. I walked down to my parents room, to find them sound asleep. Something wasnâ€™t right. Imagine my dad breaking decades of strict regimen of waking up at sharp 6. Confused, I shook my mother. She was cold and stiff. I spied an empty bottle of sleeping pills on the bedside table. I ran over to my fatherâ€™s side and felt him. His face was frozen in a contorted frown. I rushed to the phone and called up our family doctor who later proclaimed that I was an orphan hereafter.  A strange buzz filled the air and everything begun moving slowly. I slid into settee and stared at the newspaper on the teapoy. A million questions were whirling in my head. I had killed my parents. The people who loved me most on this earth, fed me, taught me, cried with me, cried for me. And I had killed them. It was long before I focused to find my fatherâ€™s photo on the front-page. I grabbed the paper and read. A horrible report about a raid being conducted at my home yesterday by the police and tax authorities, with a libel of public fraud and tax-evasion. The other accused, my fatherâ€™s best friend and long-time partner was absconding.   The next few days were a hazy chain of unsavory events. I had lost my parents, our business, our reputation, everything that had ever mattered to me. My only saving grace was my beloved Francesca. Her smiling face peered at me through countless people and her cool calm hands held my hands through the ordeal. She and her sister called up frequently. One day screaming from a nightmare where I push my parents over a cliff, and I decided I had to be with Francesca. I emptied my meager bank account, and borrowed some more and booked passage for me on the next available LA flight.   My spirits rose with the airplane. Soon I would be in the arms of Francesca. I hadnâ€™t told her I was coming. To see her face break into that smile of hers was worth anything. It was late evening when I stepped out of La Guardia airport. I picked up a bouquet of her favorite tulips and walked down to the cafÃ© where she spent most of her evenings with her friends.   I reached the cafÃ© and scanned through the crowd, my heart pounding through my ears. Through the fumes and the smoke of the bustling cafÃ©, I saw a couple in the far corner, twinned in each others arms. I smiled, finally happy with the world. The waiter approached them and they broke up, still clinging onto each other. My balloon deflated ad my head exploded. It was Francesca in that corner. The tulips slid off and I walked away, dazed and defeated, dropping along the remaining pieces of my heart.  Here I am now, sitting at a kerb in a new land, with nothing more left to live for. Just guilt, pain, misery and loneliness. All my dreams revolved around my parents and Francesca. Now nothing. I might as well go with my parents.  I walk into a drug store and ask for the same sleeping pills my parents used as a one-way ticket. I wander aimlessly, popping in pills one after the other. A little pup cuddles into his motherâ€™s belly. I make my dying wish to go away in the arms of Francesca. I trudge to her condo, slowly feeling the pills taking effect on my system. I approach the lobby manager and ask for Francesca. He replies that Ms. Francesca isnâ€™t in town, though her sister was in. He confirms Elaineâ€™s knowledge of my existence and I waddle into the elevator. What the hell was happening?   The door opens and Elaine peers out. Iâ€™m stunned. Itâ€™s the same girl at the cafÃ©, who I had presumed to be Francesca. I had forgotten in my delirium that Francesca and Elaine were twins. I meekly follow her into the chamber impervious to Elaineâ€™s condolences. Elaine explains that Francesca had left a couple of days ago to India to be with me, and reached there to find out that I had left for the U.S. SO she was returning soon.   Shame descended on me as a pack of hyenas on dead carcass. How could I? I interrupt Elaine and tell her that I needed to be admitted to a hospital immediately before falling into a stupor. I awake to find myself in a cozy hospital room with tubes sticking out of me, and Popeye punching the daylights out of Blutto on the television. Elaine smiles and tells me that I was going to be alright and that Francesca had just called from Boston airport to say that she would soon be boarding a flight back home.   Every passing moment inflates my spirit and longing to see Francesca again. I listen to Elaine noncommittally and watch the television, my mind far away with Francesca. Elaine goes to fetch breakfast and I begin to watch the morning news. Suddenly, a live feed of a burning building was shown. I recognize it as the WTC. Just as I figure out what is happening, the second plane slams into the tower.. ~j~ '),(1544,3913126,'28,July,2004','2004-07-28 00:00:00','   Loneliness    A new moon, millions of stars twinkling above so many dead loved ones living on... the road below begins to feel cold, Get up and start walking down the long straight road. cold, black, leading onto darkness  the huge mountains coated with a murky dark hue giving a backdrop to the single tree a lone leaf hanging on to precious life  a chilly gust sweeps through the wastelands the leaf bids adieu to the living world and begins its march towards fertiliserhood  Is there any life here... A crunch beckons an ant lays crushed under my feet. it writhes and lays still  the wind ceases clouds cover the sky Spin around surveying my world Close my eyes, spread my arms out breathe, feel... ~j~ '),(1545,3913126,'28,July,2004','2004-07-28 00:00:00','   In the beginning â€“ AIC  The TTP â€“ OEM â€“ 01003 was vibrating with an ear- aching resonance as I snuggled into the anti-matter chair. My uncle was due any moment now. This was his prototype time travel portal, the one he had been secretly working on in his basement. Off late he had been upbeat and had a twinkle in his eyes, a sign that invariably meant that he had had a breakthrough. I was his sole confidant and friend a relationship balanced by my enormous patience and his need to validate his work.  I was supposed to wait for him and he was already 3 hours late and I was hungry. It was way past my lunchtime. So I began fiddling around with the knobs which reminded me of salami. I was breaking a cardinal rule with my uncle. NEVER EVER TOUCH ANYTHING. But then he had broken one of his own â€œNEVER EVER BE LATE, TIME IS EVERYTHINGâ€. This fascination rather his fanatic obsession with time was a well documented affair. He had won the Nobel price for his research in time but now his continuous public research was marred by the recent global ruling by the UNTCC, the time control wing of the UN, to restrict time research to the future, a dictum which my uncle failed to meet eye-to-eye. For him everything was a circle, much like his head now. â€œwithout a past there can be no futureâ€ he said. This had lead to the basement experiments. I called it the â€œDexterâ€™s laboratoryâ€. His father having named him after the cartoon character. Fitting actually Sir Dexter Dietrich.  I woke up with a start. I must have dozed off in the chair. Where the hell was uncle Dexter? My stomach was like a raging furnace, panting to be fed. Frustrated, I banged my hand on the console. The TTP started vibrating and resonating. It sounded like a foghorn being buffeted in the wind. Atleast something was finally working. The computer screen was lit up and was showing the message â€œPlease provide the destination timeâ€ I laughed aloud and said â€˜Please take me to the beginning of time.â€™ Suddenly the arm clasps slammed on and the harness tightened and started to oscillate. The vibration had increased and a blue band of light fluxed its way in front of me.  I opened my eyes and found myself sitting on a rather damp rock amidst the most fabulous sight I have ever seen. Lush green foliage with animals running and jumping all around. The air, clean and fragrant. Birdcalls rang out, interspersed with the whistle of air breezing through the bamboo clumps. I jumped off the rock and tried to orient myself. I began walking around trying to recollect what had happened. One thing I was sure that I was amazingly hungry.   So I headed off in the direction of the trees which had fruits dropping off them like leaves during autumn. On the far corner I spied this stark nude couple walking hand in hand. I reached the trees and in the middle saw this rather strange apple tree. It had just one fruit, while all the others had so many. Any way the apple being my favorite, I plucked it off and bit into it.  ** END 1 **   I was well into the apple, when the nude couple walked up-to me. What was I to say? â€œ whateverlings, I come in peaceâ€? I offered the half eaten apple to the lady who took a bite and offered it to the man.  ** END2 ** '),(1546,3913126,'28,July,2004','2004-07-28 00:00:00','   Scarface â€“ AIC  Her cascades of hair flailed around her, the sunset glistening off her strands of glossy, black silk. She unclasps her arms, and walks along the shore, the waves kissing her feet. I stood, mesmerized, soaking up her exquisiteness. Her smoldering dark eyes, with that perpetual sparkle which belied her look of sad aloofness. The aquiline nose perched over the ruby bow, which suggested a smile, which never came. The small, sinister scar that ran alongside her nose, bringing forth-untold mysteries of scuffles unknown. The long elegant neck that rest upon a divine body resplendently draped in a crimson red sari with molten gold trimming.  The sun slid behind and I rose from the park bench, biding the image of my dream woman to cleanse itself from my existence. Something I knew would never happen. I trudged my way along to the jewelers, past the restaurant where I had made the reservations for the night and dispensed instructions and â€˜presentsâ€™ to the waiters and the musicians. We had been a regular sight at the joint for the three years we have been seeing each other. They, congratulated me in unison, warm gleeful smiles lighting up their faces. The musicians promised to play â€˜heavenâ€™, which was our song promptly on the dot of 9, followed by â€˜will you marry me?â€™  I made a mental note to call Antra after picking up the ring. She had mumbled something about having to take her mother to the hospital. Thoughts caromed around my head. My cherished bachelor status was soon to be history, not that I had played the field, but still.. Antra, ever since having lost her job a month ago, had been rather persistent on the M word. It wasnâ€™t like I had any encumbrances. I was a successful investment banker, having tasted success all through my life, except for the one fact of not having met my dream woman. Antra was the closest person I had found to her and she though perpetually bitter about it, alright with it. She called herself the â€˜dreamgirl without the scarâ€™. Often she would crack up with the suggestion that she would go get the scar done up on her surgically.   So this was it. The end of my dream and maybe the beginning of a long, happy life together. I wouldnâ€™t even bother with such elaborate proposals for which Yes was an obvious answer, but Antra was always one for elaborate dos. She even wanted a Christian wedding with the veils and the flowers. She being a Hindu and I, a Muslim.   I walked into TBZ and looked around the resplendent store, full of gold ornaments and crimson sari clad saleswomen, feeling my wallet. Having seen the diamond section, I went up to it and gazed down at the counter, showcasing marvelous creations with diamonds glittering and enticing. â€˜May I help you?â€™, a voice chirped. I raised my head and saw the scar. ** I stood, aghast, staring at her as if I had seen a ghost. Oh my God! Finally, it was her, this time without the usual sad face, but a rather bemused smile. A few moments later, she looked at me worried and inquired â€œSir, is everything alright?â€. I just shook my head and fled from the store.  I reached home, put the phones off the hook and took a long, long cold shower. I fixed myself a stiff drink and gulped it down in the recesses of my hall, staring at the intricate patterns of the rug. I looked up to find myself with the now half-empty bottle in my hand and night set in. I let a roar out in frustration in the dark room as the front door clicked open and Antra popped her head in. I mumbled a sign of welcome.  She came up-to me and gave me a tight hug and a peck on my cheek, sniffing up the air around me. She sat beside me as caressed my hair and finally asked â€˜Anything I can do?â€™ I remained silent. She hmmed and prised the bottle from my limp hand. â€˜I had gone to the hospital todayâ€™ she said. As I turned to look at her, the doorbell rang.  Antra walked upto the switchboard, turned the lights on and opened the door. â€˜Hmm, I believe this is the residence of a Mr. Saeed M. He had left his wallet back at our store. Is he in? I would like to hand it over.â€™ Consciousness pummeled its way into me, as Antra simply moved aside and let her in. She walked up to me, still in her work uniform, the scar subdued in the dull fluorescent lighting and handed my wallet to me. â€œPlease check it , Sir, before I leave.â€™ She said. I opened the wallet, pulled out a 500, and handed it to her saying â€˜thank youâ€™. She vigorously shook her head and said â€˜ No Sir, that would not be necessary. But please do visit our store soon. We have some marvelous new selections coming in tomorrow. Hmm and after you are feeling better.â€™, she said cheekily and left, closing the door softly.   I watched her leave, my heart still missing beats after seeing her smile. I heard a soft moan from the corner. I turned to see Antra on the floor, huddled up against the wall, moaning softly. â€˜Antraâ€™, I called out. After a while, she lifted her face and looked at me, with a look of despair, tears rolling from her red-shot eyes. The tears that flowed over a new surgically made scar. ~j~ '),(1547,3913126,'28,July,2004','2004-07-28 00:00:00','   You â€“ AIC  Events occur in our distilled lives, transforming out perspectives, disfiguring our sensibilities and blurring the edges between rhyme and reason. One hesitates to wonder about the casual existence of daily life. We breathe, we eat, we sleep; we occupy the passing hours with jobs, hobbies, duties, pleasure and slumber, some more than the unfortunate other.  Momentarily we halt at infernal signals, with sudden pangs of compassion or scorn, zipping their emotional variations through the being. The haggard filthy-looking boy, dressed only in an extra large discolored shirt, its frayed ends nesting a bevy of colorful flies; gazing eagerly at you, with the grubby outstretched palm, beseeching your benevolence, the glazed eyes flicking incessantly to the next beseechable in view.  You grope around for some loose change, a thought immediately discarded with the lights turning get-going-green; accelerating ahead, casting a quick glance at the rapidly diminishing forlorn image in the mirror.   The image lingering, slipping through your voluminous stream of thoughts of potholes, rash bikers and that flashy blue Merc SL500, winging its path across, morphing the pity for the kid into a desire and hunger to achieve a Merc-ised lifestyle.   Panic sways your senses as your eyes return to the road, in time to avoid a collision with a swerving idiotic cyclist. Calm and order returns, with the exception of to the knees; a mental renewal of visual priorities being structured in the consideration of an assured longevity of your life.  You sweep into the graciously vacant parking space and rummage through your wallet for the parking fee, your probing fingers jingling against some coins, refreshing your dormant memory with the reminiscence of that beggar boy.  You traipse into the cafÃ©, eyeing your watch, contemplating that for once you have reached early. A bleak smile breaks upon your face, spiraling into a light-hearted grin on sighting a vacant table, the remnants of the previous partyâ€™s evening tea yet to be cleared. Disregarding that, you park yourself on a chair, facing the entrance and snuggle to achieve the preferred comfort level. Your pour yourself a disposable tumbler of cool water. Why order a coffee right away when you would have to order another later anyway, when she arrived.  A cigarette lit, the smoke caroming off the table and throwing a blue hue to the sunlight streaming through the potted plants. The table, now being cleared by a smiling, rather buffoon-like attender. You gulp down the water, its refreshing qualities instilling a rejuvenated feeling towards life in general, constantly eyeing the anatomy of the busty, cute bomb in orange, just leaving the place. A despondent glance at the watch later, you settle down to watch the football match between two vague teams, being telecast on the television.  Another one of those beggar-boys comes scampering across and reaches out for the empty tumbler, his eyes asking the question. You nod and he retreats with his newest possession cradled in his hand, hidden from the accusing eyes of the cafÃ©â€™s employees. You follow the movements of the dirty kid with a detached aloofness and lean forward as he empties the contents of the half-drunk cup of the orange bombâ€™s coffee into the tumbler. Your eyes lock and he breaks into an embarrassed, sheepish what-else-can-I-do-boss smile. That troublesome feeling of pity floods you again, slowly snowballing into an amazement of the ways and means by which some people have to fend for themselves.  She arrives, hurrying across the cobbled path, a fixed apology on her face. She sneaks a suspicious glance at her watch and then with a relieved grin on her face leans across, pecks you on the cheek, a chirpy â€˜Hiiieeâ€™ on her lips. You reply in kind and the usual pleasantries of a meeting follow. â€˜What will you have?, you enquire. â€˜A chocolate excess and a iced cafÃ© mocha, please.â€™ You beckon the waiter and place your order, adding a cold coffee and a pound cake to the list, mentally calculating the total cost of this escapade.   Within a short minute, replete with a verbal recollection of the dayâ€™s events so far, the tray laden with your evening refreshment arrives. She digs in, apparently not having had her lunch, while you sip on your coffee, gingerly attacking the warm cake with the fork.  A flurry of movement, caught from the corner of your eyes, makes you turn around. A bustle of beggar-kids squat huddled around the erstwhile can-I-take-it-away eyes beggar-boy, frantically dipping chunks of bread into the purloined tumbler of coffee. ~j~ '),(1548,3913126,'28,July,2004','2004-07-28 00:00:00','   Cheers  Wondering away the years,  Wandering with our fears,  Weeping over our beers,  Wishing away the tears,  Waiting to be hers,  Hey how about saying cheers '),(1549,3913126,'28,July,2004','2004-07-28 00:00:00','   nobody - aic   A commercial break sprung up just when the identity of the bad guy in the movie was about to be revealed. A dumb-looking guy who claimed to be a handsome, sexy man just because he bought a cell phone was telling us how. Surrounded by a bevy of hot women, he waved his sleek phone around. That reminded me, I was due to attend a party tonight-Yet another Saturday night, yet another party. I somehow didnâ€™t feel like going. I surfed a few channels to pass the time before I could get back to the unmasking of the bad guy. It had to be the sweet-talking lawyer dude. Everyone else was just being thrown at your face as bad.   Channel after channel flickered past my eyes when suddenly I heard the refrains of that song again. â€˜You are nobody until somebody loves youâ€™. I switched back to that channel. Yeah, it was the same movie they had shown a month ago â€“ Payback starring Mel Gibson. I had laughed at that time. Yea right. My love-life was non-existent, and I was still someone. I had a job I liked, friends who were caring, cool, and fun, and a decent life going places. I was nobody it seems. Bah. Some cranky heart-broken bastard trying to make a fast buck out of his washed out love life.  But today the same song stung. I hadnâ€™t known her last month. But today, I wanted to know her for the rest of my life. Life had become a mixture of happiness, yearning, hope, and patience. I had met her in the oddest of ways. A place one least expects to find oneâ€™s complement, an ENT specialist. I had gone there to get my throat infection checked and she some problem with her ears. Never had I connected so instantly or deeply with someone; and it didnâ€™t help that she was as pretty as a sunflower on a stormy day.  I had kept in touch with her after that first meeting at the doctorâ€™s, and we had become good friends. The odd coffee-shop rendezvous and the phone calls became a source of intense joy and pleasure. Hours of witty, intelligent and stimulating conversation in tandem with her exquisite beauty, both inner and outer, made my life one continuous happy thought process. Not a minute passed where she didnâ€™t sneak into my thoughts.   Never one for bold declarations of my feelings I had nevertheless got my feelings across in a roundabout fashion, only to be shown a closed door. One thing I had learnt in life was that just because you felt something for someone always didnâ€™t mean that the other should and would reciprocate in kind. With that lesson in mind, I took the rebuff in my stride and continued as usual. Someday my love will prevail and my persistence will pay off, if not, well, life would still go on, though I somehow feel rather miserably. But hey, to love is special, and to be loved is divine, and like someone said, you canâ€™t hurry love.  I&nbsp;joined along, singing the song with Dean Martin, suddenly feeling like a nobody.  ~j~ '),(1550,3913126,'28,July,2004','2004-07-28 00:00:00','   Melody to a Memory â€“ Alex. I. Conway  The night was settling into a crisp, quiet affair. I stacked up the papers and dumped them in the briefcase. Ahhh. Thatâ€™s it for the day. I stretched out in the settee and looked around. I liked my house; spacious, bright, cheerful. The tall wrought-iron lamps casting a shine on the mauve walls; the black leather couch lay as a smiling beckoning maitreâ€™d. I smiled. â€˜timeâ€™s treated you good, dudeâ€™ Well, time for a drink.   I sauntered along to the bar and poured three fingers of scotch in my favorite crystal tumbler. On a second thought, I picked up the bottle of Jack Daniels and placed them on the teapoy. There. Hmm what could make it better? Music, of course. I grabbed the Audio system remote and pressed play. The 30-CD changer started whirring and a gentle voice started whispering into the night.  I settled into the couch and began sipping my drink. The liquid slid down, spreading a warm feeling through my being. I munched a peanut and looked at the view visible through the curtains of the french windows. My apartment overlooked the sea; the beam of the lighthouse periodically illuminating a portion of my living room with a pleasant pale blue light. The audio player whirred loudly and changed a CD. I sat there, marveling at the technological advances that made our lives so easy. And then it happened.   That song. That song that always brought a rush of emotions spiraling through my body. Cliff Richardâ€™s crooning that lovely song of his. Our song. The song we danced to, the song we listened to when we were one, the song we built our dreams on. It had been so long since Iâ€™d met her. So many moments away from her. Maybe I should call her. What would she say? I sighed.   I took a sip of my drink, ambled on to my cupboard, and pulled out my shoebox. I smelt it. A faint fragrance still lingered. She used to dab a bit of her perfume in her letters and cards. To make me want to be near her, she used to say. How true. I took the box to the settee. Letter after letter, card after card; little pink, mauve, blue green satin ribbons; each of them a different place, a different time, a different memory.   I remembered the day I decided to say the words to her. My heart wanting to, my heart needing to, my mind fiercely blocking my words; it was this song that made me listen to my heart and tell her. And so I had done, not to much avail. Well so much time had passed since then, so much water under the bridge, so many memories; though memories are not what the heart desires.   Cliff Richards finished singing our song. I hit the repeat button and pondered â€˜I should tell it to her againâ€™. A theme for a dream and what a wonderful dream it was. A perfect melody for my memories. The doorbell rang. I shoved the letters back into the box and slid it under the table.I plodded to the door, and opened it. There she was, my wife, my princess, my theme, my dream, returning from work. â€˜You naughty boy, listening to our song alone!â€™, she smiled and walked in to my outstretched arms.  ~j~ '),(1551,3913126,'28,July,2004','2004-07-28 00:00:00','   me now           Last Cigarette:     wills navy cut           Last Alcoholic Drink:     beer           Last Car Ride:     drive to marine drive           Last Kiss:     :) on the train           Last Good Cry:     while watching grave of the fireflies           Last Library Book:     hmm cant remember           Last book bought:     complete works of shakespeare           Last Book Read:     the curious incident of the dog in the night time           Last Movie Seen in Theatres:     spiderman-2           Last Movie Rented:     tomb raider 2           Last Cuss Word Uttered:     fuck?           Last Beverage Drank:     tea           Last Food Consumed:     mutton dum biryani           Last Crush:     anjelina           Last Phone Call:     last night :)           Last TV Show Watched:     friends           Last Time Showered:     yesterday           Last Shoes Worn:     nike\'s           Last CD Played:     rammstein           Last Item Bought:     umbrella           Last Download:     limewire           Last Annoyance:     PM sends mails with no attachments           Last Disappointment:     saeed hasnt called            Last Soda Drank:     na           Last Thing Written:     mail to ananya           Last Key Used:     tab           Last Words Spoken:     yawwwn           Last Sleep:     huh?           Last Ice Cream Eaten:     butterscotch           Last Chair Sat In:     na-not the chairman yet           Last Webpage Visited:     http://deviltail.blogspot.com               urlLink  CREATE YOUR OWN!    - or -   urlLink  GET PAID TO TAKE SURVEYS! '),(1552,3913126,'28,July,2004','2004-07-28 00:00:00','   finally, after all the group things, a place where i will attempt to generate the \'je ne sais quoi\' feeling from my jeus d\'esprit.  this will be me, em, m e, my, my oh my, i, i, iii. '),(1553,3913126,'06,August,2004','2004-08-06 00:00:00','   jedidiyah, your belief system is best suited to religions that value open mindedness   How do we know? While you were taking this test, we compared your religious beliefs against 10 of the world\'s most common religions. Your score shows that you share core beliefs with religions that encourage you to find your own spiritual path.  You are attracted to a religion that tolerates mixed beliefs about the existence of God and upholds the idea that there is something to be learned from every religion. You are open to a wide variety of religious and spiritual ideas. You are attracted to spiritual groups that are composed of typically open-minded and intellectual people who actively engage in individual exploration of many different spiritual truths.   http://web.tickle.com/tests/religion/ '),(1554,3913126,'05,August,2004','2004-08-05 00:00:00','   urlLink Censors ban Final Solution   how the fuck can anyone ban a film? '),(1555,3913126,'05,August,2004','2004-08-05 00:00:00','   Jedidiyah, your true talent is Business and Marketing  If there\'s one thing you want from your future, it\'s to be in charge of it. While you\'re happy to hear what others think, you usually have a clear plan for all of the things you want. And once you\'ve made up your mind, it\'s not easy to get you off that path. You\'re a smooth operator who gets along well with people, but you also make sure that people know exactly what you think.   Always strategizing about the next best thing, you enjoy making decisions and directing the course of your projects. It\'s easy for you to see the big picture in situations, and with your cool head and keen determination, it\'s a safe bet that you\'ll be leading the pack in no time. '),(1556,3913126,'04,August,2004','2004-08-04 00:00:00','   jedidiyah, your signature city is Seattle   Sleepless in Seattle? We don\'t think so. You may be laidback but that doesn\'t mean you\'re lacking in energy. Someone as adventurous and active as you is the perfect pairing with a city that\'s filled with fresh air, beautiful waterfront, snow-capped mountains, and plenty to do.  Whether you\'re hiking, skiing, or biking, you lead the way to new heights and lots of fun. And what better place to reach new heights of fun than in your signature city? How about from the top of the Space Needle or Mt. Rainier? You can quench your need for speed in the Great Outdoors, check out a jazz club or a museum, and wash it all down with a fresh cup of java at one of the city\'s countless cafÃ©s â€” all in one day. Now that\'s something we can all toast to! '),(1557,3913126,'04,August,2004','2004-08-04 00:00:00','   jedidiyah, you\'re a Station Wagon!   You swore you\'d never get one, but deep down you knew it was inevitable. Now it\'s finally happened. The car for you is a sensible, reliable, and oh-so-traditional Station Wagon. But don\'t panic: you haven\'t totally become your parents! We\'re not thinking a Chrysler Town & Country or even a Ford Taurus. The car for you is a hip, updated version of the classic. You\'ve inherited everything that was cool about your folks, and taken it to the next level. Gone are the wood panels and cheesy colors, and way in are the leather interior, CD player, and gear racks for your weekend adventures. How about a Subaru Outback, Volvo V40, or a Saab Aero Wagonâ€”in a classic navy or a firebrick red? You\'re so happening; no one would ever know that you\'ve got those sensible airbags tucked away for a rainy day. http://web.tickle.com/tests/car '),(1558,3913126,'04,August,2004','2004-08-04 00:00:00','   jedidiyah, you\'re a Raccoon in the wild world of love.  Aren\'t you a crafty creature, Raccoon? There\'s a reason for that mask on your face â€” it makes it easier to sneak out the back door unnoticed after a late-night prank or two. You\'re a nocturnal prowler and a fun-loving jokester with a definite social streak. There\'s never a dull moment when you\'re up for a party or a wild night on the town. After all, who can resist a free-spirited flirt with a great sense of humor? Your abundance of charm doesn\'t hurt your act either. You subtly move in on your prey, sneaking out of the shadows where you plot your next punch line. Before they even know what hit them, you\'ve left them howling with laughter and disappeared off again into the night.  It\'s a wonder people never tire of your rabble-rousing antics, you tease. It must be something about your mysterious allure that keeps you so fiendishly intriguing. Or maybe it\'s your approach to living in the moment. Life\'s a party and you live it and that\'s what makes you so irresistible!   http://web.tickle.com/tests/animal/ '),(1559,3913126,'03,August,2004','2004-08-03 00:00:00','   i am Barnabas. i am on my way to the office. it is raining. it has been raining for a week now, almost non-stop. the city is coolish cold and wet. there is water everywhere. water on the road, water in the auto, water on the leaves, water in the gutters, water on windshields, water on raincoats. everyone walking has an umbrella. everyone! except me. so i continue my walk in the rain looking for the elusive auto.   ah. there it is. a white car splashes water on me. fucker. get into the auto and sit in the middle of the blackish grey seat. what a color! look out of the window. nice thoughts swirl through my head. i think on. tossing the idea around in my head, shaping it, seeing it through different perspectives and lights, and hues. i think arrey wah. nice one. and then i try to remember it-will write it down.   and i dont remember it. i spend 2 minutes thinking about what i had been thinking. retracing my thought patterns. a sick feeling sets in my stomach. what is going on. the nice ideas that i get, i dont remember. so where does this leave me. do i just sit and write about the experiences around the idea rather than about the idea itself. is this the first sign of the end.  ~j~ '),(1560,3913126,'03,August,2004','2004-08-03 00:00:00','   1. werbs 2. memorial 3. battles won, war lost 4. tales from the south 5. tellmeastory '),(1561,3913126,'03,August,2004','2004-08-03 00:00:00','   \'You could feel a little moody or thin-skinned today, Jedidiyah Alfred. Don\'t worry, you\'re not going crazy! And no, the world isn\'t out to get you! Your nerves could be a bit heightened because of some intense planetary energies. It\'s best to take frequent breaks throughout the day. You\'ll want to clear your head with a nice walk or some distracting coffee breaks. Don\'t allow other people\'s moods to bring you down, either!\'  now i know why :| '),(1562,3913126,'12,August,2004','2004-08-12 00:00:00','   a bookshop. any bookshop. do u pick a book out (a) or does the book pick u out? (b)    lets negate the sometimes both and the recommended by someone options. look at your books, how many are of type a and how many b. '),(1563,3913126,'10,August,2004','2004-08-10 00:00:00','   hollywood films aside, i like brit movies. why? maybe its the language. the accent. now i\'ll just piss off.  edit: oh yea. how could i forget! the amazing sense of humor '),(1564,3913126,'10,August,2004','2004-08-10 00:00:00','   4 days of crappy metallic taste in my mouth. am fed up of not being able to feed myself. drinking electral isnt my idea of nourishment! '),(1565,3913126,'10,August,2004','2004-08-10 00:00:00','   i want to see a tamil movie in a movie hall now. '),(1566,3913126,'10,August,2004','2004-08-10 00:00:00','   http://creativecommons.org  hmm. read it. '),(1567,3913126,'16,August,2004','2004-08-16 00:00:00','   urlLink UrbanDictionary.com: Define Your World   whack '),(1568,3913126,'15,August,2004','2004-08-15 00:00:00','   urlLink Dhananjoy\'s last wish -- sweets and curd   for all that i despice and propose  a. public televised castration b. one year imprisonment in a woman\'s jail c. public hanging or the guillotine for rapists.. i feel sad for the loss of a life. '),(1569,3396638,'22,May,2004','2004-05-22 00:00:00','    If, in the minds of the writers, Tina and Bette are destined to get back together at some point, what role would you like to see Candace continue to play in the show?   '),(1570,3396638,'14,May,2004','2004-05-14 00:00:00','    BLOGBUZZ!  Just the place to find general info, updates and etc relating to the Ion Overman Unofficial Website, Message Board and Chat.   '),(1571,3396638,'11,May,2004','2004-05-11 00:00:00','    Join us for I.O.U. CHAT in the Chat Room for scheduled meet and greets! We will be serving your favorite cocktails and spinning some old school tunes.  The usual format for this chat is \"loose and informal\".  Keep your eyes open for an upcoming I.O.U. Chat schedule that will include more structured and moderated chat sessions, as well as \"theme-based\" chats for fun!   '),(1572,3396638,'11,May,2004','2004-05-11 00:00:00','    Let us know why Candace is IMPORTANT to you as an L Word character.  What does she mean to you and why...   '),(1573,3396638,'10,June,2004','2004-06-10 00:00:00','    Here\'s a place to post those little things that suddenly (or not so suddenly) occur to us about Ion, the character Candace, the L Word or life that we\'d like to share.   '),(1574,3396638,'05,June,2004','2004-06-05 00:00:00','    *  UPDATED  *   WITH RESPONSES RECEIVED FROM ION ON JUNE 22, 2004!!  THIS LADY IS TRULY A GEM THAT APPRECIATES HER FANS.  SIX HANDWRITTEN PAGES OF RESPONSES TO  YOUR  QUESTIONS!   Finally compiled and sent out the list of questions for Ion.  Hopefully she\'ll have the time to answer them while she\'s FILMING EPISODES FOR SEASON TWO OF THE L WORD!  ;-P  Actually we dont know that this is the case but we\'re hopeful.  Thanks to the \"Candygirls\" for helping to come up with an outstanding set of questions, and a SPECIAL thanks to PRicanesq, Sistah4U, Mo and Storm for their valued input during the selection/editing process.        Ion Overman Questionnaire    1.	Are you currently working on any TV project or films other than â€œThe L-Wordâ€ [â€œTLWâ€]?  Not at the moment   2.	If you were not an actor, what would you be doing?  I probably would have moved to New York and gotten a job with a fashion magazine.  Something that would allow me to shop, travel and play with make-up.   3.	Since your college major was English Literature, who is your favorite author and/or favorite book?  I have so many favorites, I canâ€™t pick one.  However, one of my favorite books in high school was The Stranger by Albert Camus.   4.	You have become an instant success and a role model for all women, particularly women of color.  What advice would you give to those women of color who would like to break into the industry?  Assuming the person has talent, the most important thing is hunger.  You have to want it so badly, you canâ€™t breathe.  All the stories youâ€™ve heard about the emotional and financial stress and rejection are true.  But if you put your mind over matter and commit 150%, you have a shot.   5.	What was it about acting that made you decide to pursue it as a career, and what is the most rewarding experience you have had as an actor?  Iâ€™ve always been a loner and a bit of an introvert.  I love acting because itâ€™s a way to be an extrovert for a short period of time, without really revealing anything about myself.  Also if I can successfully pretend to be another person, it validates what I like to believe is my understanding of human behavior.  The most rewarding experience as an actor is always having someone tell you that something you did on screen has affected him or her.  I love acting so much, I have a blast whenever Iâ€™m working.  And when someone is moved by something Iâ€™ve done, itâ€™s a responsibility on one hand, but for me itâ€™s the biggest thrill.    6.	Do you have a daily\\weekly exercise regimen? If so, what does it consist of?  I actually hate working out.  I do it because I have to, but I should be more consistent.  However, right now I do about 45 min of jog\\walking in my neighborhood, followed by about 30min of lunges, squats, sit-ups etc.  When Iâ€™m feeling really crazy, Iâ€™ll do pilates.  I do all of this in various combinations about 4 or 5 times a week.   7.	Your skin and hair are absolutely beautiful.  What is you beauty regimen? What are your favorite beauty products?  A gentle fragrance-free cleanser.  Prescription Differin cream at night.  In the morning I use an antioxidant by Epicuran called Alpha Lipoic Omega Complex, and of course sunscreen.  On my hair I use modern organic products (M.O.P.).  First I spray on their styling tonic (on wet hair), then big blobs of their leave-in conditioner and light hold gel.  Then dry with a diffuser.   8.	Other than collecting antique furniture, what other hobbies do you have?  I love to watch movies.  I go to the movies pretty much every week, and Turner Classic Movies is my favorite channel.  Also I love to shop, read fiction, hang out with friends and cook.   9.	What is your favorite film and why?  As with books, there are so many I canâ€™t pick one.  But some of them are:  - The Man Who Came To Dinner - Sunset Blvd - All About Eve - Dark Victory - A Place in the Sun - A Streetcar Named Desire - Grease - Last Tango in Paris - Breakfast at Tiffanyâ€™s - Like Water for Chocolate - Godfather 1 & 2 - Pulp Fiction - Kalifornia - The Age of Innocence and tons more.  There are lots of reasons why a film becomes one of my favorites, but usually itâ€™s the performances.  I love watching other actors.   10.	What is your ethnic background? What is your family like?  My mother is Black with a little American Indian (Sioux and Cherokee), and my father is a blond haired, blue-eyed Englishman.  Iâ€™m an only child, and my parents and I are close.   11.	Are you single, engaged or married?  Not married, but my boyfriend and I have been together for 4 years.   12.	What kind of animals do you have, if any? What are their names?  I have a fish named Sushi (trust me, he doesnâ€™t mind), and I have shared custody of my parentsâ€™ Golden Retriever.  Heâ€™s 8 yrs old and his name is Bogie.   13.	What types of music do you like? Who are your favorite performers?  I love classical and Jazz (esp. Miles Davis, John Coltrane, Ella, Louis and Billie).  I also love Frank Sinatra.  I also love contemporary R&B and Hip-Hop like Jill Scott, Alicia Keys, Lauren Hill, Snoop, Ja Rule.  I also love Motown and the Beatles.  Again, too much to name.   14.	What year were you born?  Now, now, I do declare Miss  {webmistress\' last name} .  You know it is impolite to ask a lady her age.   15.	Regarding your character in â€œThe L-Wordâ€, how are you like Candace and how do you differ?  Moreover, what do you feel is the most important aspect of her presence on the show?  Well our differences are superficial ones, such as occupation and sexual orientation.  But I think we are similar in determination and belief in ourselves.  Also in being true to ourselves.  I think Candace showed us that even strong women are human and can be ruled by their emotions.  That combination of strength and vulnerability is part of what makes us complex and interesting.   16.	How does it feel to know that after playing the role of Candace, so many women are attracted to you?  How does it feel to be a â€œlesbian sex symbolâ€?  I had no idea I was a â€œlesbian sex symbolâ€ until I read the question.  But I am incredibly flattered.  Thank you.   17.	As Candace, your ex-girlfriend (Yolanda), your carpenter crew and the young lady introduced as Mikela were all women of color.  Did you have any say in this or to your knowledge, was this done intentionally/unintentionally by the producers?  I had absolutely no say in it, and I donâ€™t know what the producers meant by it â€“ if anything.   18.	â€œNew York Post: The L Word - Gay â€˜Sex and the Cityâ€ By DANIEL RENZI Beals was so upset by Betteâ€™s unfaithfulness towards Tina, she took out her frustration on the actress playing the part of the other woman. \"That was incredibly distressing,\" (Beals) said gravely. \"I couldnâ€™t be with her on the set for a couple of days. I couldnâ€™t talk to her. I would just look at her like, â€˜Youâ€™re breaking us up!â€™\"  Did Jennifer Beals really act this way or was she just kidding?  What is your reaction to her comments?  If you will, please tell us about your experience working with Jennifer Beals [â€œJBâ€]?  Until this Questionnaire I wasnâ€™t aware of the article.  I donâ€™t know if she was kidding or not.  But if she really did feel that way, I didnâ€™t know it.  She was professional and we got our work done.   19.	The chemistry and sex scenes between you and JB were phenomenal!  Are the creators of the show aware of this?  Further, how did you feel about performing those scenes?   While we were shooting, the Powers that Be told us that it was working, and that they were pleased.  I was grateful for the challenge of performing the scenes.  I had never had a sex scene with a woman before, so as an actor I was determined to make people believe me.   20.	Of your fellow actors on The L Word, with whom did you spend more time with outside of the set?  I didnâ€™t hang out with anyone off the set.  I did a lot of sightseeing, shopping and reading while I was in Vancouver.   21.	Are you aware of the negative responses towards you as \"the other woman\"? If so, how do you feel about that and would it have any effect on your return to TLW?  I wasnâ€™t aware of the negative responses.  But I take it as a compliment.  It means I did my job.   No, that would have no effect on my return to the show.   22.	Since it seems a lot of people feel you really nailed the body language of a more tomboyish or butch lesbian, can you tell us if there is anything or anyone that you drew specific inspiration from, in your portrayal of Candaceâ€™s presence and mannerisms?  The first thing I did was think like a man.  And one day on the set, I asked one of the male crew members to walk back and forth for me so I could study his gait.  That was helpful.  It helped me get a grip on my center of gravity.   23.	To your knowledge, will Candace return next season and for how many episodes?  If so, can you share the direction in which the show plans to take the Candace character and/or in what direction would you like to see the character go?  I donâ€™t know if Candace is coming back.  I havenâ€™t been contacted.  I donâ€™t know what Candaceâ€™s place could be since she has caused so much turmoil.  But I suppose anything could happen.   24.	What were your immediate thoughts and comments when you saw the website for the first time and have you shared it with any friends and\\or family?  Well, I was and continue to be humbled, thrilled and grateful.  Yes, Iâ€™ve told everyone I know about it.   25.	Anything else you would like to tell us?  Iâ€™d like to thank everyone again for visiting the site.  Iâ€™m so glad you enjoy the show, and I hope these answers are satisfactory.       Thank you for your time and cooperation!   '),(1575,1305545,'16,May,2004','2004-05-16 00:00:00','   there are different types of email clients. on the pc side of the world you can maybe differentiate between outlook-like types (easy and unflexible) and netscape mail-like types (the classical 3 panel setup - folders, messages, message preview - multiple accounts, filters and so on..).  i\'ve always been in the latter camp with using, certainly, netscape,  urlLink pmmail/2  and for the last years  urlLink the bat .  the bat has always been the closest to pmmail, that\'s why it has been chosen at the time (some 4 years ago) and  it was fast as hell even with giant message bases at that time of some 10.000 messages.   but - this was then and now is now. the world of email changed radically, first with the heavy increase of spam, than of virii and worms and later with the demand for increased mobility and the need for synchronisation or online access to you mailbox by IMAP. today, message bases of power users count in the gigabytes and 100.000 messages.   using an email client seems to be a bit like driving a car of a certain brand.  when you need a new car you take the successor of your old one. so, when a new version of my mail client is released, i certainly use that.  with the bat it was the same, i shelled out the 45 EUR for the business version, i use it partly for that purpose and appreciate the work on a program of that complexity.    but then, things went ugly.  the IMAP support in the bat\'s 1.x branch was crude but usable, it basically provided the same functionality like you had with POP3. the 2.0 release of the bat, expected and announced for years, promised high and mighty \"full IMAPrev1 support\" - but what an exaggeration!  from the developers point of view this statement may even have been right - client and server could exchange commands and data according to IMAPrev1 - but the client side was so lousy! basic UI considerations like instant feedback was not provided. filtering mechanisms - the proud of the POP3 side of the bat just didn\'t work.  but not enough, after waiting some months until the arrival of a maintenance version which, after some testing, provided at least an again  usable  IMAP support i bought my license in the hope that  urlLink ritlabs  (the developers of the bat) are working on fullfilling their promises. a week later they suddenly removed the \"full IMAP-support\" advertising. great.  i bought a new version with a lot of new but mostly questionable usefull features minus one usefull. thanks.   it was time to look thoroughly for an alternative.   finding an email client with full IMAP support turned out to be hopeless  - if i didn\'t want to change to, cough, outlook. yes, outlook is the only email client on the market which uses the strong points of IMAP to an extent, that you could call \"innovative\". it hurts a bit to say such things abut the big evil from redmond ;), but hey, honour where honour belongs to.. IMAP offers multiple connections to servers, server side folders and searching, server side implemented instant notification about new messages without the need to check \"every x minutes\" (not all servers support all features, but there are enough out there who do). but these features are supported only by outlook and to some extent by  urlLink mulberry  (which has a lot of drawbacks in other areas, IMHO)  ok, i forgot about my IMAP dreams and concentrated on the client\'s capabilities when it comes to managing my mail. i have multiple accounts, i have a lot of contacts, receive some 100 mails a day and never throw away anything - i keep even spam for some time - you never know what\'s gonna happen to your penis ;)  after installing and uninstalling some ten or fifteen different demo versions of email clients i sat back and tried to tackle the problem differently - by thinking about it.  my first question was \" what is email? \" email is an exchange of information - and more and more a tool for archiving it. you can at best compare it with the letters and documents you find in your (real life) mailbox. and suddenly i realized something strange: in the last years i had developed two completely different approaches of handling email on the one side and dead tree documents on the other.  in RL i try to spend as few time as possible on  organizing  my working life. i get the bills, enter them into the tax program, put them in a file and that\'s it. quarterly i do my turnover tax statement and put the file with the bills into a box. all letters which require a reaction go into a todo file - everything else gets into a different file and into a big box at the end of the year - which goes into the basement later. that\'s it. this way, all i have lying around in my office are two files and a box. when i need a letter from a year ago, i know, it\'s somewhat deeper in the box and search for it. this takes less time then organizing everything in different files for different senders, months or whatever - considered how seldom i need a year old letter..   but - when it comes to email i suddenly behave like 65 year old prussian bureaucrat.  i have some 200 folders and subfolders in the bat, write extensive filters to move incoming mail into them and look proudly at this manic work of a psychopath. why do i do this? i have no clue! all i usually do is looking into the inbox for unread mail, read them, answer them or put a red flag on them if i want to answer them later. sometimes i need to refer to an older message, let\'s say a seller\'s message from ebay and scroll thru my 2000 pixel long folders tree until i come to business|ebay|sellers|unfinished deals. or i just press f7 and wait 10 seconds until i get 20 results which i scroll thru to find the mail from yesterday. or i take some 10 seconds more time to refine the filter to include only mails not older then 3 days old.  but hell, why did i filter the mail in the first place when i now do the good old search/eye ball search thing?!  obviously, something went wrong in my routine in the last years.  and then i remembered the \"revolutionary\" new mail client in my browser of choice,  urlLink opera . incidently at the same time opera entered the 7.5er beta cycle which brought improvements for  urlLink the mail client  - and so i gave it a try.   what shall i say - someone in norway had the same thoughts like me. opera mail fits perfectly to the routine i describe above. it\'s approach is to put everything into one big box, a database obviously. a database you can index, and everyone who knows a bit about databases knows what this means to search speed. nowadays i have some filters for reoccuring types of emails (status mails from my servers,  one  folder for ebay and folders for mailing lists) and everything else i do with incoming mail is giving them labels (mainly todo and mail back).  all my mail but the above mentioned regulars arrives into the inbox, i read them, give them a label or just mark them read. a nice thing about the inboxis, that you can make it to \"show only unread mails\" so that all mails i dealed with just disappear - which is very rewarding somehow ;) (if i pressed to fast \"mark as read\" - no problem, ctrl-z works as it should)  when i need a mail from some time ago, i go into the big \"received\" folder and type a detail from the mail into the quicksearch box. with every letter i type the amount of mail shown in that windows shrinks until i find the mail i was looking for. with a drop down box with options like \"show mails from this week\" or \"show mails from today\" i can  quickly  refine my search.  all this usually doesn\'t take longer than 2 seconds and i found the mail i was looking for. HEAVEN!   in the end, opera mail sacrificed a little bit of flexibility for speed and usability - that\'s a price  I  am willing to pay.   the problem with the bat and it\'s developers, ritlabs, is, that they never listened to their user base, nor thought about the changes in the world of email in general and, instead, implemented features without plan and vision. meanwhile, they get more and more autistic by shutting down the user forum on their website without announcement and explanation, and beside announcements of new betas almost noone is talking to their beta testers anymore. (not that they did very much in the last years anyway..)   somehow i have to say \"thank you\" to ritlabs - for giving me a decent mail client over years - and forcing me to think about what went wrong lately ;)  '),(1576,3560320,'06,June,2004','2004-06-06 00:00:00','    The Y Factor  After my playtime with the kids on Saturday morning, I thought I would send you a weekend â€œrantâ€!  I thought a lot about what I would write about, world events, like the Middle East, nah, or jobless statistics in the US, nah, a nonevent.  So I guess I will discuss in my own dogmatic way â€œLove in a Relationshipâ€.  Itâ€™s funny how many teenage girls hear the words â€œI love youâ€ in the back seat of a Chevy just as the bra strap is cautiously being removed.  And then the girl feels the stick shift that the girl thinks is poking her in the side of the hip until she remembers she is in the â€œback seatâ€ â€¦ dah!  Of course this teenage boy will forget he ever said these words 2 milliseconds after the words â€œno, stopâ€ hit his ears.  This kind of â€œeventâ€ is happening every weekend in every small town and metropolitan city in America, and has been going on for years.    Of course even earlier in Egyptian days, if the guys got too frisky Cleopatra just had the guyâ€™s hands, and / or other parts, chopped off .  These poor fellows were known as â€œeunuchsâ€  and the king had them working the queens quarters (for obvious â€œsecurityâ€ reasons) for the rest of their lives â€¦ what a boring life!  Think about it, you watch and help the queen bath everyday, and you canâ€™t do anything about it, to yourself or others.  These eunuchs would have paid $1,000 for one pill of Viagra!   Well, back to the 21st century.  So what do these three little words mean, â€œI love you.â€ They are easily the most inappropriately overused words in the English language, and the most manipulative and misused (see above teenage example) words.  But what do they mean?  Are they the same for everybody, like the color â€œredâ€ is always â€œredâ€ to everybody?  But love is different in its interpretation by each individual â€¦ I think anyway!  I am not sure even the most prominent psychologists have discovered the precise meaning of the words, correct me if I am wrong here!  I have often felt the â€œin loveâ€ part of love was dramatically different from the just â€œlovingâ€ someone, like your sister or brother, or even husbands and wives in some situations (these couples keep therapist in business!).  I believe the â€œin loveâ€ part of the term refers to â€œthe romantic intimacy combined with a pinch of lustâ€.  In other words the â€œstrong desire for the physical sensuality with a heavy dose of respectâ€ part of the relationship enduring in the couple to some degree due to the letâ€™s call it â€œthe Y Factorâ€ still being present in the relationship.  The â€œY Factorâ€ is very elusive and a strangely mind-boggling item and the subject of about 500 million books.  And due to the fact these â€œY Factorâ€ (my term and I get the royalties to the term if it becomes famous!) books with the varied subject of love are still being written, and becoming best sellers, it is obvious that the answer to the question of definition and how to get the â€œY Factorâ€ remains a vague (what do I call it, thing, item, feeling, event â€¦ what?) term!  Well, not being a genius, accomplished clinical psychologist, or â€œunusually enlightened by heavenâ€ I will not be able to truly answer this question â€¦ today, but maybe next week!  So I will not bore you any longer â€¦ till next week!   '),(1577,3560320,'06,June,2004','2004-06-06 00:00:00','    Well, yesterday afternoon, I thought to myself itâ€™s my Friday night of true â€œpersonal unstressed freedomâ€, therefore I should venture out into the mysterious world of the â€œUrban Kansas Cityâ€, and explore the world. Hey, I have been watching â€œSex and the Cityâ€ for months, now it is my turn to see for myself the realities.  Of course I was taking one of my employees and his wife to dinner first, but that just an innocent looking â€œdodgeâ€, my true adventure was a quest to witness the bar scene antics of â€œSex and the City - \"Kansas City style\"!!  After my employee and his wife said their thank yous and good-byes, I sat down at the stressfully populated (the crowd looked like a 250lb woman in a size 6 polyester pants suit from the 1960s) main bar of the Capital Grille on the KC Plaza to start my â€œvisual explorationâ€!  The following are the items I found â€œinteresting & laughableâ€:  1) Why does a 65+ year old â€œladyâ€ dye her hair bleach blonde, cruise the bar with a totally fake Julia Roberts grin, with plastered Tammy Baker eye liner, looking like the Saturday night â€œblue lightâ€ special from Billy Bobâ€™s Retirement Village & Escort Service. This was not appetizing for the rest of the patrons who were waiting for a table to dine in the restaurant!  2) Why do bald, or almost bald men, in the 55 to 65 year old category insist on combing over the three hairs left on their head so it appears like they have more and better hair than Brad Pitt. You know, those 3 hairs with all that lacquered hair spray are more nervously starched on the top of these menâ€™s heads than George W. Bush as the guest speaker at a Sierra Club dinner.  I say please guys, take a hint from Bruce Willis and Andre Agassi, just cut it short and BE YOURSELF. You are baldâ€¦. Get over it!  3) Of course I still canâ€™t figure out these wealthy 55 year old male divorcees with their bold Tommy Bahama shirts (of course I love the less bold Tommy Bahama clothing) unbuttoned to their beer bellies with more gold chains than Mr. T from â€œRocky IIâ€.  I thought this went out in the 70s, or maybe they are just now catching up to the 70s.  Guys it is 2002 alreadyâ€¦. Grow up your not 25 any more!   It is always interesting to witness the mating dance of the high end restaurant martini bar, all the men are so bold flashing themselves to the audience like Jesse Jackson looking for a TV camera, actually any camera within 100 miles will do for Jesse.  And, of course the women with their cool, collected demeanors like they have no interest in anything, except maybe something attached to a Mercedes SE series or better preferred, an Armani suit (not off the rack), and a Rolex will do!  And these women think that with their bodies by Dow Corning that they deserve this attachment to the Mercedesâ€¦. Of course if they spent $10,000 on boob size and cleavage the guy should â€œappreciate herâ€!  This was too much for me, I chuckled as I walked out the door and to the valet to retrieve my company provided Chevy Malibu. Dressed in my subtle but smart Tommy Bahama shirt (buttoned all the way up) and off the rack jacket and slacks. Then go home to watch the New York version of this situation on HBO, because I can be anything I want to be in my own home, in front of my own TV â€¦. Thank you very much!   '),(1578,3753830,'25,June,2004','2004-06-25 00:00:00','   hi. this is my first post.....yay!!! anyway, i\'m a HUGE fan of Linkin Park and i just went to their show on the 22nd of June. best day of my life! i\'m in the lpu so i got to meet the band which was the coolest thing ever!!! anyway, here\'s what happened during the day. first i met up with the other lpuers at cityhall mrt station at 4 or so then walked to the venue which was the padang. there were already so many people there so we quickly joined the que. we were sitting on the stinkin\' grass for so long i managed to make a new friend - jolenne. then, i noticed that the lpuers who got into the meet and greet session were not in the que but were standing somewhere in the front so jolenne and i joined them there. i met some malaysian lpuers and we all quickly became friends. we were standing for quite a while when some of the girls said they want to go to the bathroom so they went to look for one and i was left with platinumscorpion(forgot your real name!) and alan. we waited for a really long time before they came back and we went to the meet and greet counter to sign in.  waited and waited for a very long time and it even started to drizzle a little but luckily did not rain. finally, a lady told us to go que up in another place while she started checking our names off a list. we waited some more and i was entertained my none other than magic man syafiq. showed me some tricks and then who could have guessed who i saw? mtv vj denise keller!!!!(for those who are not in singapore you might not know who she is!) she came over to us we were filmed for mtv use. here\'s what was said: denise: what are we here for? us: LINKIN PARK denise: what do we wanna watch? us: BREAKING THE HABIT!!!!! it was so cool then after that, i managed to get a picture with her. then somemore news people came over and we started shouting take picture! take picture! then mark from lp newscaster came over and picked alanvoon and ipporek to do the newscaster episode with the rest of us in the backgrond. he kept on taking picures and a few days ago, i checked the lpu website and found my picture there! so cool! then, it got dark and we were finally taken to meet lp. it was so cool! they are really cute in real life. especially mike! anyway, first in line was brad so here\'s i put my poster on the table to get signed and then i said \"this is for you\" coz i got him a b\'day keychain of his b\'day, 1st december. he said \"thanks\" and read it. then he said \"i didn\'t know queen mary of scotts died on my b\'day!\" i just laughed and took a picture of him before going on to mike. he\'s so cute!!! and he was smiling such a nice smile i almost melted. anyway, he signed my poster and i gave him the keychain. his b\'day\'s on 11th february. he said \"thank you\" and stared at it a while before putting it in his bag and smiling at me. then was phoenix. he said \"what\'s up?\" and i just stared at him, totally awestruck that he was talking to me. he looked at me and then said \"nothing?\" and i just nodded. so stupid!!!! anyway he signed my poster and i gave him the keychain of his b\'day. then he said \"thanks. how did you know my birthday?\" i just laughed and he laughed. then was joe. he signed my poster and then i gave him the froggie stuffed animal since he likes froggies so much. he looked really surprised and said \"thanks\". then was rob. he was really quiet. he signed my poster and i gave him the keychain. he said \"thanks\". that\'s when i remembered : I DIDN\'T TAKE A PIC OF MIKE, PHOENIX, JOE AND ROB!!! OMG!!! i jsut turned around and snapped a shot of brad, mike, mark(newscaster guy who was filming the meet and greet), phoenix and joe. then i tured to rob and took a pic of him! phew, that was close. then was chazzy. he looked really tired. he signed my poster and i took a pic of him and left. OMG!!!!! oh shit. my sis wants to use the com so i\'ll continue the rest another time. for now, bye! '),(1579,3753830,'04,July,2004','2004-07-04 00:00:00','   so after i left i went to meet the rest of the lpuers and guess what? the security guards didn\'t lead us directly to the venue as stated, they led us to the ENTRANCE of the Padang. holy shit they made us rejoin the que so we were last in line. but that didn\'t stop us. once we got into the venue, the tallest guy from the lpu pushed his way through and we followed him behind. we were stepping on peoples feet and stuff  and they were cursing at us but we didn\'t care. in the end, we got somewhere in the middle. oh shit. i\'m telling you, my sis needs to stop using the com. peace. '),(1580,3659117,'23,June,2004','2004-06-23 00:00:00','   Well, fellow Fleetwood Mac fans, what did you think of the concert last night? I played their cds incessantly before going, fought traffic getting to the Coor\'s Ampitheater with my gf in tow, we got there just in time. M* and I walked quickly to our seats so that we wouldnt miss a single golden second...only to wait for them to start (no warm-up band!) and look around our EMPTY section! Too funny. We were in the special \"upper levels\", of seating, and gradually inched our way down, until finally, right before the fIRST encore, we were practically near the stage level!   Did you see that fan try to touch her royal shawl-ness? And did you hear her reaction? I thought that was too cute: \"His touch was very gentle folks, almost like he was a part of the band\".  Highlights of the show: They sang their new song \"Peacekeeper\", and at the end Lindsay did a little dramatic \"thing\", that made me cackle and snort. Then Stevie (or should we call her Stephanie?), changed Shawls (from Black, to Gold, to Red), and sang a song, then we got two songs from Lindsay, then a song from Stevie, then another two songs from Lindsay....Is this Fleetwood Mac featuring Lindsay Buckingham?  FM, we will always love you.  Stevie did a Stevie song (at the very end!), \"Stand back\". As we were being carried along in the undulating, yet-not-dancing crowd, my friend M* was crushed by an eager fan, just as the song \"Stand back\" was bursting full throttle.  Please...Stand back Stand back!  And she did her little \"twirl\" wit her skirts and shawls flaring out. Yup, we got our STevie fix. Although at first we were wondering when we\'d get to actually HEAR her voice...she sang into the microphone but nothing was coming through.  However, for the most part, the crowd was mellow. I guess its not a band you dance to? Anyone dance to Fleetwood? It was strange not hearing Christine though, I did not realise how many songs she is actually featured on. She was not there...and I grew curious as to why.  urlLink This website  has the answer, I think..  All in all, I wish I had read about the band before going...the\'ve had a tumutluous time of it, and I would have appreciated \"The mac\" a little bit more, with his bulging eyes, big grin, and electric vest.  We loved loved loved the bongo-playing chinese guy...what was his name? His rythms and Linsay\'s guitar riffs were pretty great...  And the smell of pot and kettle corn wafted through the air, filling our nostrils. :)  Thanks Mac! '),(1581,3659117,'21,June,2004','2004-06-21 00:00:00','   This seems to be the week where I am feeling happy for everyone else\'s success! Which doesnt bother me at all, because there are plenty of times when I am commiserating with things that are going wrong. My friend Steve (I can mention his full name because Im plugging for him...) just announced his honorable mentions at the fair via his \"click-n-pay\" web site, chock full of frameable photos. Some from Hawaii (whales and beaches), and some from China (people scenes, temples), they are all extremely sharp and colorful, but for that extra kick, the composition is EXCELLENT! I think my favorites are from Bora Bora. Steve doesnt know this, but for about a month I used his Bali H\'ai pic as a desktop wall paper. :)  My new favorite is the Chinese ladies on a bicycle. ( urlLink William de Kooning , watch out!)  So, \'nuff said, go visit  urlLink Steve Azer\'s website  already and see for yourself. Oh. And he won FOUR honorable mentions at the San Diego County Fair (did I mention that?), as well as his pics appear on the cover of San Diego Weekly. This is one eye to watch... '),(1582,3659117,'20,June,2004','2004-06-20 00:00:00','   I caught this link on a fellow Blogger site, and wanted to share it with y\'all: I feel  urlLink Great with Nutri grain !  Bleh, my narcissistic tendencies shall not be satisfied today - Blogspot is not showing me my published Blog. \"Oh, my, just HOW am I to edit my masterpiece of genius musings?\" (hand on face, eyes wide, puckered mouth) Could be the internet connection.  Lesson Learned: when in doubt, check yer connections  So, the plan today - clean house, play with my MAN, Jack, eat (love those carnitas burritos with chips and guacamole) watch DVD \"Along came Polly\", try to foget last nights showing of \"Mystic River\" by Clint Eastwood (speaking of Narcissistic tendencies...anytime a 70-year old gets jiggy with a 36-year old, SOMETHING  is wrong when said 7?-year old does a film about Child molesting. Wait, what does that have to do with Narcissistic tendencies? Dont know, but it sounded good)  Have a great Sunday, San Diego! '),(1583,3659117,'18,June,2004','2004-06-18 00:00:00','   Hi,  Well this weekend is almost over and here it is - Father\'s day!  Friday I took off to Downtown for a night of Bitter-sweet conversation, Bitter Apple martinis and a bit of Bitter dancing at the Bitter End. That was fun; I met a few nice people, a few interesting people, lost my disposable camera after having taken pictures of my cousin G* and her friend C*, dancing up a storm with some li\'l cuties. Well what happens Downtown, stays Downtown I guess. ;)  Saturday was recovery day (on my advanced years, I cant do all-nighters anymore, darn!) and really got nothing done except read Maxim, and ate some tacos. While watching a DVD of \"Carrie\" by stephen king, (yikes I was so out of it I got the made-for-tv version instead of the really fabulous one with sissy spacek) anyways, while watching \"Carrie\", at about 10:00 on Saturday night, I get a phone call. From T*y buttons. He has some really great news, folks. The film company that he works with just won Two Emmys!!!! He was calling me from the San Diego Emmy Awards Party that was being held at the USS Grant Hotel. He had no one else but moi to share it with. I was at home with my boy-toy Jack, on the couch watching some girl get doused with pig blood. What a great evening!!!!  So check out the  urlLink Ortiz Film Works  website, for a list of their credits and accomplishments. And if you are a director with lots of cash who wants to make a movie or music video, hire them! They are hot!  Well, thats all for now. Im am very happy for them, and very very proud of T*y buttons. Congratulations, everyone at Ortiz Film Works!!!! '),(1584,3659117,'18,June,2004','2004-06-18 00:00:00','   I cant believe this! I am clueless in San Diego, thats why Im single in San Diego!!!!!  I have no idea where to go! Can someone out there help me? My cousins want a fun place to go, and they want lots of cute MEN. Not the gay ones, although they can sure dance well and smell divine. So Montage is out. Not on broadway AGAIN!!! please!!!!  Perhaps the \"red circle\"? I think the bitter end has lost its fun edge...these days. Any help!!!  Thanks, for helping a couple of depraved females... '),(1585,3659117,'17,June,2004','2004-06-17 00:00:00','   Im sorry, but I used to love that song...those were the only lines to it, it had a catchy beat, you blast it as you are driving down 5th avenue downtown San Diego ready to get down and funky...now with today\'s news, I will be turning it off when I hear it. ugh.  So, ok, I used to be a member of match.com. I thought Id found my match - he was a bit funky and different (Scientologist notwithstanding) and quirky, but turns out that he has this really dark side. Anyways, I blocked him from my work email, I circulated his contact info and picture to all the members of my company security department (I work for a government contractor. We have gates to keep the riff-raff like YOU out), I called the front desk person B*B, and gave her his info, I told my boss, I talked to the Office manager...Ive done EVERYTHING that I could. Yet, he emailed from a saved message I had sent him...he emailed through to my match.com id. I had forgotten to block that! Darn!  So girls and boys, if someone is annoying you, (or in my case threatening) from match.com, you have to block that too!   Well, its Friday, and my best friend canceled on this weekend, my cousin hasnt called or written and my other cousins want to \"go out\" tonight, and I just found out. How come I always find out at the last minute? Well anyways, it should be entertaining - Ill bring my camera this time. Last time...all the fun stuff was OFF camera. :) hehe.  I have no idea where we are going, but knowing those two itll be fun. '),(1586,3659117,'17,June,2004','2004-06-17 00:00:00','   Well, typically I would be out an about celebrating the end of the week, but here it is - Thursday evening and I STILL dont know what is going on this weekend. Not a single persona has decided to get back to me...ok, what the heck?  So I will pop on over to the Sporting Club one and SPA, do some weights, spin the cycle a bit, and then slink home to do some cleaning and watch the rest of the \"Station Agent\". Ill let you know what I think. So far I really really like it. '),(1587,3659117,'16,June,2004','2004-06-16 00:00:00','   We can write about anything here online, and so I thought that it would be great to expunge all my bitterness regarding my previous attempt at a normal relationship, here.  I read a quote from one of my groups, and it spoke volumes to me:    \"People who want to share their religious views with you almost never want you to share yours with them.\"  This seemed to be the ENTIRE description of what went wrong between K* and I. Besides the \"bitter betty syndrome\", he was fanatical about Scientology.  So I decided to do some research on good \'ole Scientology and L. Ron Hubbard. Such as - what the heck does the \"L\" stand for? (Lafayette Ronald Hubbard) So I took myself to the  urlLink Church of Scientology  website, first of all...  Here is what I found out: There are different levels of \"enlightenment\", called \"clearing\". Its like Regression Therapy, where you go back into time and throughout your past, and \"clear\" away all the bad experiences that are holding you back from attaining all your true power. One of those true powers are telekenisis. Didnt they ever read \"Carrie\"? Only adolescent girls are allowed to move things with just their minds.  So anyways, you pay through the nose for the different courses and learn all these new words. You are required to understand all the different words, and armed with a dictionary, you learn the definition of \'knowing\' and \'knowingness\'. \'Being\' and \'beingness\'. Isnt that nice? Just what I always needed to know. I needed to have beingness. (And at the end of the relationship, i was required to learn the definition of the words whore and prostitute. Girls, stay away from this one - email me and I will give you his name...)I learned also that the church of Scientology has been taken over by the American Medical Association, and now it is necessary to create your own \"group\", to study and follow Hubbards precepts. All his ideas of course are based on Science. I dont know what KIND of science, but well, it all seemed like new-age hokey sh!t to moi. So, I looked for the TRUTH about this religion, and found some interesting tidbits on   urlLink Operation Clambake    An interesting section that I found, was that focused on HOW they obtain reasonable people and turn them into veggies; titled  urlLink Thought Control and Scientology , detailing how they gradually take over your mind through sensory deprivation, and \"trance-inducing\" repetition.  HA! Sounds like any other religion to me! Just kidding. '),(1588,3659117,'16,June,2004','2004-06-16 00:00:00','   WEll, I guess to be fair, I should write about Gram\'s 75th birthday last weekend (June 12th).  So...reaching back...  I arrived Friday morning after getting up BEFORE the crack of dawn at aproximately 3:40 am. I woke up in a panic having gone out the night before and neglected my packing. Whew, plenty of time. My buddy-girl C* picked me up in a haze, tried to dump me off at the commuter terminal before finding out that to qualify as a \"commuter\" at Lindberg field, you must be going to LA. Well, Im way on my way to San Jose! Do you know the way to San Jose? (sorry HAD to do that)...  So I spent the entire day Friday hauling. moving, pushing and shoving, running around. Then saturday spent the rest of the say setting up speakers for Karioke, putting up balloons, stressing ebcause the caterers werent there (they finally got everything set up 3 hours after the party started!!!!)...lets face it - it was allotta work! Oh yah, Dad was miserable about the whole thing, and if he criticises me one more time Im going to scream...  But then, as soon as it was time to \"turn on\", I took a nice, deliciously hot shower, squeezed into my lime green dress (strapless -  my first one!) with a fringed bottom and lacy overlay, splashed on some makeup, and varooom! out I go...  It was a blast. We all got toasted (including my 16-year-old cousin who was currently taking a \"vacation\" from school, in other words he was kicked out) and sang karioke while DJ Daddy spun some tunes. Items of interest: JB, the black sheep of the family got accepted to Yale, and is on his way to becoming a doctor. Curious George was all over the girls, and finally drove me away, running and screaming...Darrins baby is exceptionally cute...who is she related to?  So the evening was fun for all. There was not a cloud in the sky, the garden had butterflies flitting in and out, the peace and tranquility of the Prunedale (that just makes your lips want to pucker up dont it?) countryside seemed to imbue all with a sense of calm and cheer.  The location was perfect - Carlos Perez, his house is perched on a hill, off of Garlen Drive in what is affectionately knowns as \'Prune-tucky\", outside of Salinas California.  All in all it was in a word, perfect.  The next day we all piled into the dining room of the Hilton Hotel\'s restaruant, and binged at the all-you-can-eat buffet, chatting over eggs and mimosa. The three girls that had been out on the town all night the night before, they were in a world of hurt, but otherwise none-the-worse for wear. You know the saying, \"what happens in Monterey, stays in Monterey....!\"  Update: Gram was admitted to the emergency room of the monterey community hospital,  I called yesterday and she was all doped up. Also, mom was given drugs for a nasty viral infection. Otherwise everything is fine. '),(1589,3659117,'16,June,2004','2004-06-16 00:00:00','   People are always asking me why I am still single. My mother, Dad, Grandmother, new propects, etc. I always tell them that its because in San Diego, the men my age all like the 20-year olds. For proof, look on the dating site  urlLink match.com . Most, if not all men who are aged 35, want a woman aged 18-25. What can I do about that? And for further proof, look at the  urlLink Belgian pedophile ring  currently under investigation. Turns out that the corruption goes all the way to the top, through the entire legal system of cops and judges. And they wonder why the guy wasnt investigated properly? And they wonder why I am still single?  I mean, why go out with a 35-year old, educated, fit, attractive, strong-willed American female, when you can go across the border and get a hot little mexican 18-year old tamale?   Dont get me started. So, I am still just hanging out with my cat, listening to Sara McLachlan singing about being stupid. Thats moi...stupid choices by smart women. Isnt there a book out there? Is there a book called smart women, smart choices? Is there a mail-order groom site out there? :)  Lately Ive been thinking about enhancements...its been preying on my mind...\"everyone who has them, loves them\" I hear. Its not fair!  Speaking of fair, I havent been yet. Im thinking that I should go, but well, a night of battling the crowds as well as my weakness for fried oreos doesnt sound all that appealing. Besides, Ive already missed seeing the b-52\'s and journey perform, so whats the point? '),(1590,3659117,'16,June,2004','2004-06-16 00:00:00','   Did I mention that I am practically exploding from my sugar high? I just finagled an M&M cookie from hot computer guy, who also brought me one yesterday. :P   I guess that I will need to go to the gym after I pick up my car (which is STILL not ready, hello). While the car guys were replacing the darn battery, they decided to try to force my window up. Its sort of sticky and Ive got a major callous on my finger from having to press really really super hard to close it. So, they broke it and decided that I needed a new window closer, along with the new window. What started as a simple \"minor tune and battery\" turned into an all day car-orgy!     Thank you Brake Company and Moe from Eastgate towing!!!! Just kidding. Things will work out ok. Did i just say that? I must be a descendant of Anne Frank. You know who she is dont you? She discovered hot dogs. :) '),(1591,3659117,'16,June,2004','2004-06-16 00:00:00','   I just set this up TODAY, and already Im listed on another Blogger site, http://sandiegobloggers.com/, created by Joe Crawford at artlung.com. Check out the other San Diego bloggers!!!  ............................................  Ive got the biggest crush on the guy/boy at Starbucks. He blushes when I talk to him, its so cute! Hes like waaaaaaay too young for me (you know I like the older dudes in their 40\'s and 50\'s) but you know, he\'d be fun to hang out with. No baggage, no \"bitter betty\" bullsh!t as my cousin Gwen would say, and no kids or divorces. Ok I am assuming this, we havent gone beyond \"I\'ll have a tall split and a vente ice water. With a STRAW. And a LID. I jogged to work today, stopped by my starbucks at the Costa Verde Center and BOTH the Starbucks Cuties (as they shall be called henceforth as I have no idea of their given names) were vying for my attention. Ok, so my ego is getting big, they were being nice and chatted with me.  Soooo, anyhoo...I jogged to work today because my car is at the shop. The battery died and I thought it would be a great idea. It sounded good at the time. OK. Well my Lungs are now black, but heck, I burned off a few hundred aclories. The hill at Genesee killed me. The cars whizzing by (and you know who you are) almost killed me. I should change this BLOG to \"single and uninjured in San Diego\". :) haha.  ..........................................  Whoo hoo. I get to go to the Fleetwood mac concert on tuesday!!! Whoo hooo!!!!  ..........................................  Ive also got a huge crush on my friends friend, and the computer guy at General Atomics. They are all pretty cute. And single. Any single gals out there? I could set you up. Computer guy has a dog, so you cant be allergic and have a cat. And mind his roomates/best friends. friend-of-friend is lots of fun and likes to go out to clubs. His travel schedule might be annoying, but he likes to give massages so I guess it balances out. :) '),(1592,3659117,'16,June,2004','2004-06-16 00:00:00','   Wow, its hard to believe , Dear Friends, that I am suddenly single again. It seems like it should be a lot easier to find a mate in a city that boasts over 3 million people. And one Scientologist. :)   ..............................................................   Seriously though, I am going through the depressed-then-pissed-off phase, you know the one...its descibed as a loss that is like death. So Ive been lethargically completing my 30-minute workouts (hey 30 minutes is a long time to be standing up), pull myself home somehow, flop onto the couch with a bag of trailmix in hand and grab the \"clicker\". In between wonderful movies such as Footloose, and Love, actually, I engage in long, meaningful conversations with my cat. He has such interesting viewpoints...Id like to share.   ................................................................   #1 Playtime is important and superscedes everything else in life - even licking ones butt. #2 Feet, paper, string, fingers, hair - everything qualifies as a mousy-toy. #3 Politicians dont feed you tuna, they feed you bull. Vote for those that promise to feed you tuna and deliver, because they will get tired of you crying for tuna. Eventually. #4 Cats DO like fruit. #5 If you have to ask if it looks good, or if it makes you look fat, then it probably does. I will look at you as if Im interested, but its the same way I stare out the window or look at the wall - it means nothing except that you seem like you want to feed me Tuna. #6 Miss Elephant really likes it.   ............................................................................   Suddely single means never having to say you are sorry - and never having to go for that Brazilian torture device they call a wax. I found this out, one pain-filled day, tears pricking my eyes. If you do decide to \'go brazilian\',  Ask for Stephanie at Brazilia Skincare. She will lie to you when you ask if it will hurt, which will make your heart stop pounding in your chest for a little bit. And then she will tell you how much she hated the corporate world and how much she loves her new career ripping peoples hair out. :)   ...........................................................................   Speaking of hair - ever have one of those really, truly bad days, where you come home and hack off all your hair with a pair of cheap drugstore scissors? Well I did recently and boy was THAT liberation at its finest. Talk about washing the man out of your hair? Well I CUT him out of my hair, and now I feel a true lightness. Yes, folks, a total of 7 inches came off somewhere (ok, dont ask ME, IM not a stylist, but it will be a few months before I can face Jennifer. I know she will take one look at what Ive done and grimace like shes in pain. Like she is getting a Brazilian.) Its really a co-dependent relationship we develope with our stylists, you know. They depend on us for cash, and we need them to talk to, make us feel good and help us improve our self-esteem. Like a therapist that cant prescribe anything harder than a shiner or shampoo. As I was cutting, I thought to myself....\'wow, Im really going off the deep end arent I? What will everyone think? Wow, how can I DO this?\' And then I just said \'F&^%k it, who cares?\' The ends were dead, so I put them in their little white porcelain coffin and flushed them down! :)   ..........................................................................   By the way did everyone read the article about dogs knowing more than 200 words? Anyone who has ever had a pupster knew this a long time ago. Everyone has had to spell out w-a-l-k sometime in their lives...I mean ...did they really need to do a study on this? Does this mean they are sending German Shorthairs and Golden Retrievers to Iraq now? Sorry Harlan and Reba, you know Im grinning as I say this...   ........................................................................   Speaking of dogs, I was playing fetch with Jack the cat the other evening after a hefty spin class, and was thinking, \'what if I could train him to go to the bathroom in the toilet?\' Wouldnt that be a cool cat trick? So I looked it up on the net and sure enough I found step-by-step instructions on potty-training your cat! So I thought I would share this priceless picture with everyone.  urlLink http://www.karawynnnet/mishacat/toilet.html    Isnt that cute?   ...........................................................................   Speaking of cute, has everyone seen the memorial pictures of Ronald Regan? Poor guy, I think it was time- I mean he WAS 93. And EVERYONE wants to see his poor lifeless body. WHY? The guy was our president, but WHY? High ick factor there.   ..........................................................................   So what has this Suddenly Single been up to lately? Opening day at the Polo field in Del Mar for starters. This is where I was able to watch the horses frolic around a field, stomp the divets and follow the chuckers. If you dont know what that is, youll have to go and find out for yourself. I was also able to learn more about Restraining orders than I ever cared to know. And darn it, but I took notes. You know, just in case the security guards at the front gate, south end and north end as well as the front desk receptionist fail me. :)   ........................................................................   Then there was the Beach Boys concert that same evening - the first concert of the summer at the famed Humphreys. The sky was overcast but it was warm, with a friendly easy atmosphere. They were out of tune, but I doubt anyone cared. I saw 20-year olds singing along to all the favorites like \"I get around\", and 60-year olds dancing like they were at the sock-hop-by-the-sea. So imagine my surprise when, during a  bathroom break after the concert, I wandered off and met Mike Love!!!! For those that dont know who that is...thats the LEAD freaking singer, man. And for those that dont know who the Beach boys are...and you know who you are...they are most famous for the FIRST version of \"California Girls\". I guess that means me. But let me remind you, I am a Central Coast California girl. What does that mean exactly? It means that when I smell cow doo dee, I think of broccoli. It means no plastic...er, \"enhancements\". It means \"pale face\", knowing the various brands of chewing tobacco, knowing that cold is 40 degrees not 60, and it means trees that arent black and still smoking from the last fire.   Hope every has a great weekend. This Suddenly Single Central Coast \'California Girl\', will be flying back to a place called \"Prunedale\" for her Grandmothers 75th birthday '),(1593,3659117,'16,June,2004','2004-06-16 00:00:00','   Whats new in the land of the Suddenly single? Well, how about a review of the best places to goggle the other supposedly singles, suddenly singles and permanently singles, and see if we can tell the difference between them. Our favorite places are in the La Jolla/Del Mar area, but there are lots more yet to discover, say Downtown San Diego, and Mission Valley.   Cafe Japengos - known for its BUSTling Thursday nights, this crowd is a little more relaxed these days. Here you will see mostly blonde and gray hair - blonde women and gray-haired men chasing them! You can get the best sushi in san diego here as well as the best apple martini. All this after working out at the club/spa sporting club one across the circle.   Epazotes/Cafe Pacifica - Del mar is not only famous for its horses, but its long-legged FILLIES!!! Here you get a much mellower scene, where you can start at Epazotes, stroll across to Cafe Pacifica, then wine-down at the wine bar at Il Forniao. This is my personal favorite, but watch out - here is where you are most likely to be taken in by a beautiful sunset and romanced into thinking a permanently single is just a regular old single (You see there are differences). Here you will be lulled by the sound of the fountian tinkling as you sip a special port or smoke a special cigar. Dont fall asleep, or like a midsummers night dream, you will wake up to find yourself with bruises and not know where they came from!   what else? There is Dobsons Downtown, where I hear the lawyers lurk, and then Gordon Beirsh, the younger business/networking crowd.   This suddenly single has decided that she will not be going to these places anymore, due to an unfortunate incident involving too much, too soon, and not enough from the wrong person! You figure that one out and you win the \"suddenly single sweepstakes\". :)  When suddenly single, it is generally a good idea to call all the people that you forgot about while not single, and re-connect, apologising and supplicating yourself until it becomes embarrassing for said friends. They will forgive you, but you must try harder next time you are not single to maintain contact or they may forget about you.  Places to take those that you are kissing up to: Art - The Vatican is coming out of the closet - er - showing itself - er - displaying its treasures (oh! I give UP) at the San Diego Museum of art, in a rare never before seen in public appearance. music - the yeah yeah yeahs are playing at Soma, soon. Very high cool factor. duh. For the rest of you yahoos, Friday is the \"your 5\" concert with Jessica simpson and a bunch of un-altered MUSICIANS. Sports - The rock and roll marathon is coming up - running, drinking and bands - a great event to spend all our suddenly single time preparing for!   When suddenly single, I found that the best therapy was window-shopping for things that any of your exes would not have approved of - open-toed shoes, low-rise jeans, new peircings- this time in your tongue.   Keep to your friends, friends, and you\'ll never go wrong. '),(1594,3659117,'04,August,2004','2004-08-04 00:00:00','   So the Sara concert (you know, THE Sarah...) was excellent and full of warm vibes all around. But lets start at the beginning shall we?  The weekends always start on Friday, and so naturally I left early and went straight to my apartment to pack. You see, I\'d gone out the night before to celebrate the weekend. Thursdays are always fun at Japengos. Friday evning, I dressed in my new outfit, a sporty white denim mini-skirt and a white Guess? tanktop with lace that showed *Gasp* my belly-button. I guess skin is no longer in, but in Vegas I figured it was a requirement. I arrived in a hot, muggy Vegas airport and saw my love, J*, who greeted me with \"guess what, I just won some money\". Ok, how much I ask. \"about 1000 dollars at blackjack\". Ok, so our trip has been paid for already!  He whisks me to the hotel room and we wash our faces and pop into the Hard Rock Hotel lobby where there are clothes from the Temptations and Elvis in glass cases.  Then off to a very nice dinner with steak and Champaigne. His two favorites apparently.  Saturday: A day at the spa where I was massaged and trimmed, then worked out (ok, so Ive gained a few pounds since then....) and then off for an impromptou shopping trip at the Bellagio. I am now the proud new owner of a small, black leather chanel purse! Id card and everything! The concert was in the Mandalay Bay Hotel complex, and at the time there were 3 concerts going on simultaneously. Wow. Snoop, Sarah, and someone else...cannot remember who. Well, we traipsed on over in the Hard Rock car, and entered the theater excitedly. The crowd was a mix of older/younger, straight/lesbian/gay. Hushed we listened as she belted/growled/whispered the words of all our favorite songs. J* actually enjoyed her music, and made the comment \"her music sounds sort of country\"...   :)  Sunday: A day by the poool in the middle of a huge pool-party called \"rehab\". We had a hosted cabana that ordinarily goes for about 250US$ per day, ordered some Vueve Cliquote Champaigne in a  cold bucket, some mexican food from \"Pink Taco\",  and enjoyed ourselves as we both obained deep-tissue massages from Michelle - a very talented massuesse.  Crowds of beautiful, tanned, oiled, outrageous people thronged around the pool, a circle of blue water flanked with sand.  At various junctures, waterfalls spilled into this pool, and in the center was a bar and \"wet blackjack\" tables. In secret corners of this pool, small hot tubs and wading pools gave the illusion of privacy.  The hot, sunny Vegas day, turned cloudy and muggy, and suddenly a lightning rainstorm crashed overhead. It was amazing, as florescent white light cracked the blue vegas facade. The lightning made J* a little nervous, but otherwise we hung out the window like kids at the carnival.  That, folks, was the \"softer side\" of Vegas. :) '),(1595,3558382,'27,June,2004','2004-06-27 00:00:00','    Since I will not be here for a few days, you could amuse yourself by making a comment or 2. This is supposed to be an interactive medium but I have only seen a few brave comments so far.  You can comment about the blog, my writing skills, the weather, or anything you like. Part of the purpose of this trip is for me to get ideas about what to do when I get back to Geneva. I have already received many suggestions and will probably write about them in here in the future. I keep writing the suggestions down in a special blue book with planets on the outside.  If you\'d like to think about this for a bit and give me suggestions, you too are welcome to participate in this game by commenting in here (or, if you prefer to do it anonymously, you can send me an email).  Thank you very much in advance.   '),(1596,3558382,'25,June,2004','2004-06-25 00:00:00','    For those of you who know me well, I am VERY allergic to horses. Therefore it wouldn\'t seem like a very good idea to go to a horse farm where they have 16 horses, 5 cats, and one dog. The amazing and unbelievable thing is that I haven\'t taken a single anti-allergy pill since I arrived here on Friday morning!  Susanne, aka Sushi, and I have known each other since we were 10. We used to be so much alike that my mother couldn\'t tell us apart from behind. We used to have the same clothes, sneakers, things, used to like the same colours, you name it. We realised that our fabulous connection still exists...we have the same blue mobile phone. Not only that, we have chosen the same ring tone as well. Susanne has her right ear pierced twice, just like me, and she is also wearing a glittery thing in the second hole, just like me. Fun fun fun. Today we bought the same kind of fleece jacket just like we used to do as kids.  Midsummernight\'s Eve rained away. It was pouring cats and dogs when I arrived on the train from Copenhagen. We agreed to stay in that evening instead of celebrating. And that was ok because we only wanted to chat and chat all evening. It was 4 years ago since I visited Susanne last, and then she was living in Wimbledon living a slightly different life. Since then she has made a dream of hers come true: Susanne bought Lynga horsefarm 18 months ago and rents out boxes to 16 horseowners. There are plenty of girls helping with the work on the farm, and Susanne\'s and Han\'s daughter, Samanta 9, gets to ride her pony Cello on a daily basis. Han is also happy living on the farm because it enables him to do things with his hands, which he enjoys (like building horse boxes, for example).  Never in my wildest dreams would I have imagined that I would be able to live on a horse farm with all my allergies. But I guess The Cleanse I went through has done me a LOT of good (if you haven\'t yet heard about The Cleanse I have done, I\'ll be more than willing to tell you).  My early departure from Edinburgh gave me a glimpse of the city from above on a nice sunny morning. I wish I could publish my pictures from there in here, but realise I have to learn the Picture Blog before being able to do that. I\'ll try to learn that later during my trip.  It is late and my train leaves at 8 in the morning. I\'d better go and re-pack my bags. The coming week I will be with my aunt in her summer house. Hence, no computer or internet access until Thursday (at the earliest). Therefore, I wish you a good week and see you soon. (If you don\'t see anything from me by Friday, it means I will be busy at a friend\'s wedding until Monday 5th.)   '),(1597,3558382,'25,June,2004','2004-06-25 00:00:00','    Today I finally had some Scotch Whisky. I visited The Scotch Whisky Heritage Centre. A couple of days ago I was approached in the street by someone who wanted to do a survey of tourists coming to Edinburgh. If I showed up at a hotel the following day and promised to participate in a focus group I was going to be given Â£40. Pretty good pay for an hour so I went. There were 8 of us, 3 Americans, 2 Canadians, 1 Lett, 1 Chinese, and me. We received the money immediately and then they served us coffee and tea. The focus group discussion was taped and basically the Scotch Whisky Heritage Centre wanted to know why we had not visited their attraction since it was right near the Castle. We all gave our reasons why and left forty pounds richer. We had a nice time actually, and were given brochures of the whisky centre which made us all decide to go there now that we knew about it.  Yesterday the weather was absolutely horrible. Windy and rainy and cold. At least today the rain didn\'t fall sideways... I went to Holyrood Palace yesterday, only to find out it was closed 23 June to 5 July. Next I went to an interactive exhibition called Dynamic Earth--only to find out they had closed at 15:15. Typical. I went to the hotel to wait for my 40 pounds instead. After the market research Margita and I had a really nice girls\' dinner, just the 2 of us. We caught up and reminisced about old times (\"yes, remember that one who, at 17, said he had always liked older women, and I was only 21!).  Today I finally got to both Dynamic Earth and the Whisky Centre. I really enjoyed getting to know how whisky is made and the distinctions between malt whisky and blended whisky. Whisky is the Gaelic word for \'water of life\' in case you didn\'t know.  In the evening went with Margita to the children\'s weekly swimming lesson. Great fun. Freja is crazy in the water. She extra crazy today because of her birthday. She told every person she met: \"I\'m five today.\" Cute. I gave her a t-shirt (pink of course!) where it says: \"I\'m a wee angel\".  Early tomorrow morning I take off for Copenhagen. From there I will hop on the train and cross over to Sweden to visit Susanne (aka Sushi), an old school friend who lives in Halmstad. Hopefully we will be able to celebrate midsummer in style, but the weather doesn\'t look that promising. Oh well, it\'s the being together that counts.  Ok friends, until the next time I get my hand on a keyboard.   '),(1598,3558382,'22,June,2004','2004-06-22 00:00:00','    Holistic Quotient. This is the combination of IQ, BQ, EQ and SQ, where the letters I B E and S stand for intelligence, body, emotional and soul respectively. I love this interpretation of the human being and of humanity.  Susanna and Martin Ehdin are the Swedish authors who have written about this very interesting perspective in the book with the same title as this blog-entry (Original title: HQ - den manskliga helhetssynen). As far as I know, the book is yet to be published in English, but at least you can take a look at their websit www.ehdin.com for more information. Susanna is an MD in Immunology and Martin is a consultant who lectures at business forums.  I have been reading this book for the past month. It\'s less than 200 pages so it\'s not very long but there is a lot to think about in there. I have been re-reading many passages which resound with my thinking and values. Throughout my reading I have wished for this book to be translated into English so that I will be able share it with my friends. I even considered translating it myself in order to get it done. But by now I have checked out my namesake\'s website and realise that she will is probably in the process of getting it done already.  The book describes how you and I can create a better future for ourselves and our world by giving our HQ more prominence in our lives. We need to create a balance whereby we reduce our reliance on IQ, which is the dominant way of thinking in today\'s world, and giving more space to our kinestetic and spiritual abilities and senses.  The reading of this book gives me hope. I very much belive in its approach, particularly since it resonates with all the cleansing that I have been doing lately. The book talks about the intelligence of the heart and the body, the value of our own personality, our inner and outer environment. It presents a new view on health and integrated medicine, and then goes on to discuss wisdom, spirituality and identity. The tinal chapter pulls it all together by bringing HQ into our relationships, our workplace, into parenting, and into society. One important message is that we are all connected and what each and every one of us does had an effect somewhere else. Just imagine our hearts beating with the same frequency. What power!  For me, this book represents a holistic view for humanity. I highly recommend it.   '),(1599,3558382,'21,June,2004','2004-06-21 00:00:00','    Today is Siri\'s birthday. We will celebrate with a mini-party this afternoon. She will have a big party on 3 July and is very much looking forward to it.  On Thursday we went to see Jesper perform on his recorder (blockflojt in Swedish). He played his solo by heart (he was the only one who did). There were 16 performers all in all. At the end 10 of them played a piece all together, which was very nice.  Friday I got the privilege of picking up the girls from school. We had to walk to Jesper\'s school to catch up with Margita and to watch Jesper during his sports day. Unfortunately it started raining and was pretty miserable so we had to go home and change before heading off. Once we were better equipped we set off on our walk. Freja, with slight exageration, said \"I don\'t like this rain. This is the worstest day of my life!\" Once we finally got to Jesper\'s school we had missed all the action. Jesper had received 8 first places and one second place so he was very pleased.  Friday evening Margita had her girlfriends over. They all went to pre-natal class together when they were expecting their now 8-year olds. And they still get together every 2 months for dinner at each other\'s houses. We had fun and chatted a lot. I had made falafel which turned out to be a very popular dish.  Saturday afternoon, after kid activities like Gymini and football, I took the 3 kids to see the new Harry Potter movie. Jesper and Siri had already seen it. My darling goddaughter Freja sat on my lap the whole movie through and was only scared twice. Good girl! She loves Harry Potter.  Saturday night Nick took Margita and I to the party of an Irish friend, who had just moved into a penthouse apartment down in the harbour. Wow, what a view of the city from there! We left when it was past midnight so we were pretty tired on Sunday after two late nights in a row of partying.  Today I will go into town for a while to have a look around.   '),(1600,3558382,'16,June,2004','2004-06-16 00:00:00','    Monday was a great day. Margita and I started off with a breakfast at POSH with Nick. Then we went to change the battery of my Baume&Mercier which had died during the night. Next stop was Boots to get some suncream. I also bought a nail file which was a casualty in my suitcase during the first leg of my trip. Next stop was the mobile shop to check out a new mobile phone for Margita. She was in need of a free upgrade. She ended up getting a nice little light oyster handset from Siemens. By the time we had done all this it was still only ten o\'clock.  The tourist bus took us around the city and down to the port during the next 2,5 hours. The guide we had was very funny and kept us amused and interested. It was very windy and cold out there on the roof of the double-decker bus, but it was the best place from which to see this beautiful city. The sandstone from which the majority of the buildings are made from were once upon a time dug up here in Craigleith, right next to where Nick and Margita live. We had lunch at a very quaint little cafe called Clarinda\'s. Clarinda was the muse and love interest of the writer Robert Burns and his grave is located just beside this little cafe in the Old Town. Margita had a nicer reason for going there, as she and Nick went there to celebrate each time after she had her scan to confirm her pregnancies.  Tuesday I had a nice time at home fiddling with my digital camera. I have signed up for an on-line course and I went through the first lesson, as well as the user guide for my camera. It was nice to have the time to do only this for as long as I wanted. The next lesson is tomorrow. In the evening on Tuesday we went to see Siri in her ballet class end-of-year show. She was the grandmother in Peter and The Wolf. Before her came the smaller girls, i.e. the snowflakes and the flowers. They were all very cute to watch. Thursday night it is Jesper\'s turn to perform on his flute.  Today was a very lazy day. I woke up at 10:30 which I haven\'t done since I was on holiday last summer. It was wonderful.  BTW, I forgot to tell you that the Swedish football star Henrik Larsson, who scored twice in the game against Bulgaria on Monday, he is a much loved local hero in Scotland. He has just spent the last 7 years playing for Celtic, one of the popular local teams. Hence, there are many scots wanting Sweden to win the Euro2004 in Portugal because their hero is on the Swedish team. As a curiosity, a Celtic\'s home game draws an audience of 60\'000 people.... A bit different to poor Servette Football Club in Geneva.   '),(1601,3558382,'14,June,2004','2004-06-14 00:00:00','    On Saturday I went for a lovely walk along the water of Leith, a small river running through a section of Edinburgh. In the evening the kids insisted that i watch Bruce Almighty on DVD with them before going to bed. During the whole movie Freja was cuddling up beside me. Every 10 minutes or so she turned around and stroked me on the cheek, just to make sure i was still there. Cute.    Bill, a Canadian friend of Nick and Margita, came over for leftovers (he is affectionately known as the Hoover from Vancouver!). He had brought a video of when his band played in a pub recently. Unfortunately we could never get the cable to work between the video camera and the TV so we ended up listening to the Amazing Greys instead. The Amazing Greys is my brother\'s band. I received a copy of their first CD with 8 songs just before I left on my trip. very good stuff! Bill and Nick were very impressed and loved the sound.    On Sunday the whole Price Family and I went to an animal park in the country side. The weather was nice but very windy. There was a funny looking feasant which Siri thought looked like Elvis. He really did with his large tuft of pitch black feathers on his head and the white suite of silver-spotted body feathers.  We got to hold baby chicks in the hatchery, and when we went to the falconery I got to hold a big European Eagle Owl called Honey. She was beautiful! brve Jesper held her too, plus 2 other birds of prey. Freja got to hold a very cute Barn Owl, whereas Siri fed a buzzard on her arm. After an excitig day we stopped for Fish&Chips on the way back home.    For the past 2 days I am drug free, i.e. I haven\'t taken any paracetamol for my back pain. I now only take a herbal remedy called Kawa Kawa root. Keep your fingers crossed that I will be able to keep this up. But isn\'t it funny though that I am able to do this less than a week after leaving on this trip? I think I was very keen to get away.  Have a good week! Today Margita and I are going on the tourist bus around the city.   '),(1602,3558382,'12,June,2004','2004-06-12 00:00:00','    Last night, Friday, Margita had invited 5 of her colleagues for dinner: Alan, Simon, John, Roger and Robert. They all had hamburgers and sausages, while the 2 vegetarians (Robert & I) had oven roasted vegetables, pasta salad and green salad. Jesper gave us a brief flute concert [I will try to post his photo here eventually] before we went downtown to visit a few pubs.    First we went to the Berlin Bierhaus, which was located just past Alva Street (for those of you who don\'t know, my niece is called Alva). Then Alan insisted we go to Harry\'s Bar. There were a lot of people, a hen party with a dancing blow-up doll, there was a lot of smoke and extremely loud music. I immediately had to put in bits of kleenex in my ears in order to stay sane. Just goes to show that I have outgrown the previously joyful experience of hanging out in bars... The energy field is so different in bars and after all the cleansing I have done lately I need calmer and more harmonious surroundings. Nevertheless, we had a good time and Margita and I came home at a very respectful time around eleven.  Ah, I have just found where I can change the timing of my posting. This time it will be Edinburgh time instead of California time.   '),(1603,3558382,'10,June,2004','2004-06-10 00:00:00','    My trip has started!!! On Tuesday the 8th I landed in Edinburgh in great spirits. It was really nice to see the Price Family again: Margita & Nick, Jesper 9, Siri soon to be 7, and my godchild Freja who will be 5 later this month.     The departure from Geneva was a nice suprise. I had proudly packed one suitcase which weighed 20.4kgs. I also had one carry-on with a beauty case on wheels and my back-pack. The lady at the check-in counter said she had never seen such a thick packet of tickets. When she heard about my adventure over the coming six months she smiled and gave me Priority tags on my luggage, \"For a lucky start to the trip\"\', she said.    The surprise was that I am allowed to carry 2 pieces of luggage with 32kgs each on the overseas legs of the trip. Luckily I only packed the one suitcase because with my back I\'d never be able to lug around as much as 64 kilos, forget it. But it gives me a bit of room if I\'d like to buy something later on. For now what I have is enough.    Yesterday I rested after many late nights last week preparing the trip. Today I went for a bicycle ride around the neighbourhood. The kids are eager to meet me and play with me after school as of 14:15 and onwards every day. Right now they are all at their swimming lesson which means I can get some time to type this. Margita and I will do a tour of the city on Monday. Hopefully I can take some nice photos.   '),(1604,3558382,'06,June,2004','2004-06-06 00:00:00','    I\'m glad I found out about Blogging. This way I will be able to share a brief diary about my trip around the world to see friends.  Only 42 more hours before I take off on my World Tour 2004. I\'m very eager to get going but still have a lot to do. Must get back to my papers.... See you soon on this Blog!   '),(1605,3558382,'31,July,2004','2004-07-31 00:00:00','    What an amazing show! All the artists are doing amazing things. Not your average circus.... I missed them when they came by Geneva last year so I am very happy that I got to see them here in Calgary. If you get a chance to go see the Cirque du Soleil, GO.   Sharilyn and I were just talking this morning about how wonderful human potential is. If you train enough your body can do fantastic things.      In a short while Sharilyn will take me up to her new Cabin at Buck Lake, about 2,5 hours north of Calgary. From there we will to a road trip, a kind of Thelma and Louise trip but without the illegal stuff, with two kids in tow, and a different ending. We are heading for Vancouver on the west coast. We will see the Rocky Mountains and the wine districts. We will spend 3 days on Vancouver Island. After that we head back in-country to Sharilyn\'s parents\' cabin. Also by a beautiful lake.      On 26 August I will return to Calgary for a last few days before heading off to California...   '),(1606,3558382,'30,July,2004','2004-07-30 00:00:00','    Today I went to the coolest and biggest Harley Davidson shop I have ever been to. Too bad I didn\'t bring my camera. Above the entrance there is a huge fork with a wheel. Very cool. The salespeople introduce themselves by name and shake your hand when you come in. I was very impressed. They had a large remaining stock of the 2003 Centenary version bikes. Hard to believe when what we heard over in Europe last summer was that there were no more to be had last year...      Sharilyn, a very good friend from Red Cross during my Asia posting, brought me there along with her two kids Kieran and Laurel. After admiring the bikes in the shop we went to the front into the Harley diner which was decorated in Harley orange, black and crem white. Very nice place. Today\'s special was a \"HOG Wash, Vodka Paralizer.\" I didn\'t have one, because I have been drinking way too much over the past few days. Kieran, who will be 5 next month, had fish and chips and managed to finish one of the HUGE pieces of fish on his plate. It was a real Harley driver meal alright. Sharilyn and I had Chef Salads, while little Laurel (19 months) had bits and pieces of fish and ham.      This past week in Calgary has been great. Lots of shopping, more shopping, quite a lot of drinking and then a bit more drinking. As you can see I\'ve been having so much fun that I haven\'t even had time to write in here. Actually, Sharilyn and I are going off for drinks right now, she tells me, so I\'d better write off here and continue later. After the drinks we are going for dinner and then to see the Cirque Du Soleil show Quidam. I\'m very much looking forward too it. More later...   '),(1607,3558382,'25,July,2004','2004-07-25 00:00:00','    Before telling you about Ottawa I have something very important to say:     &nbsp;   26/7 -- HAPPY BIRTHDAY ROBERTA!!!     &nbsp;   My arrival to Ottawa was my very first visit to Canada. I instantly liked it. Roberta brought me first to the farmers\' market in the middle of town and then&nbsp;home to her beautiful house about 25min north of the city on the Quebec side. (I realise I should probably invite Roberta to do some interior decorating for me, I liked&nbsp;the way she has made up the house so much.)     &nbsp;   Although Roberta had arranged for me to go to a dinner that evening, I was exhausted from an early start in Washington that morning. It was also&nbsp;nice just to have&nbsp;personal logistics&nbsp;time all by myself that evening.&nbsp;I have to schedule in more of that during this trip.&nbsp;I unpacked everything in my suitcase, did lots of laundry and ironing. Yes, I did the ironing&nbsp;even if the clothes were going straight back into the suitcase again. (For those of you who don\'t travel as much, I can tell you that the fit in the suitcase is much better if you&nbsp;iron your clothes first.)&nbsp;   I also took time out to&nbsp;write a couple of blog entries, which&nbsp;I posted the following morning.     &nbsp;   Roberta and her husband&nbsp;Sandy have a very nice cottage by a lake 1,5 hour north of Ottawa.&nbsp;Our trip there&nbsp;the next day&nbsp;I saw as&nbsp;a great opportunity to get a first experience of the Canadian wilderness. On the way there Roberta got the use of Sandy\'s BMW M roadster (if you don\'t know it, check out the BMW website--it is one cool car!!! Slightly more sophisticated than my 1987 325ix.). It felt great to swish through the countryside with our&nbsp;caps on. Sandy graciously took all the luggage and the groceries in Roberta\'s Subaru.   The weather was wonderful that day.&nbsp;     &nbsp;   The cottage is on a rocky incline down towards the lakeside. We&nbsp;jumped in&nbsp;immediately upon arrival. It was so beautiful and very similar landscape and nature to Sweden. Just some differences in&nbsp;flora and fauna. We took a boat ride around the lake in the sunset and watched the loons which were calling to us. We went passed some friends and received a very precious gift: chanterelle mushrooms. Mmmmm... We had them with butter, cream and a bit of white wine. Divine. After dinner I eagerly watched the humming birds flying back and forth to the feeder filled&nbsp;with sugar water. Very cute. There was a bird which had a very prominent whistle. Given its persistent song, we later we found out it was a Singing Sparrow.     &nbsp;   The next morning looked kind of grey and the rain started around noon. And did it ever rain. Roberta said in the 5 summers they have been there they have never seen such a downpour. The screened-in porch flooded with rain blowing in and the dinner table got soaking wet. We had a laugh and took a nap to recover from the thunder and lightning. Fortunately it cleared up by the next evening so we had another nice tour around the lake and saw beavers swimming right by the boat.     &nbsp;   Sandy had to go back to town, leaving&nbsp;us girls&nbsp;to talk and talk like girls do. We know each other from the Red Cross. Although she was doing consultancies even before,&nbsp;we are both thinking about what is next in our lives.&nbsp;We had a very nice time while discussing. Roberta had&nbsp;brought a book called  Energy Medicine by Donna Eden . My nose was stuck in this book&nbsp;a lot&nbsp;of the time at the cottage. We did The Three Thumps, The Wayne Cook Posture, and Spinal Flush and felt all the better for it. We experienced some additional torrential rains and the drive way flooded sand all the way down the hill to the cottage deck. A ditch was dug to remedy the disaster so hopefully this will address the problem in the future.    &nbsp;   Although we did get some more sunshine, the rain was mostly on the agenda and Roberta had to wipe the dinner table on a couple of more occasions before we left. Unfortunately the top had to be on the M roadster Thursday on the way back to Ottawa. Thusday night we all went to see Spiderman 2 and had a great Thai meal. Friday was a last chance to&nbsp;drive around&nbsp;some of the city but I did no sightseeing at all. It wasn\'t that kind of visit and I thoroughly enjoyed it that way.   '),(1608,3558382,'18,July,2004','2004-07-18 00:00:00','    Lance Armstrong just won the Tour De France for the 6th consecutive time!!! I have a tremendous amount of respect for that man. What an athlete! Last summer I went on my Harley into the mountains  down  a road where all the T de F cyclists had gone UP. Wow...and to think they go up hills at between 22-25km/h.&nbsp;I even have great respect for Jimmy Casper,&nbsp;&nbsp;the poor guy who came last for the 2nd time in a row, nearly 4 hours behind Lance.    &nbsp;  On Friday night I arrived in Calgary. I received certain comments that I must start posting blog entries again after my (off-line) cottage visit in Ottawa. Well, it\'s encouraging to know that people read what I write in here and look forward to what\'s next. Makes me want to write more.    &nbsp;  Ok, I will post a couple of things today so keep reading folks. But first I must call my nephew who turns 18 today    &nbsp;    HAPPY 18th BIRTHDAY MAX!!!!!!      &nbsp;   '),(1609,3558382,'18,July,2004','2004-07-18 00:00:00','    Noreen lives in Maryland in Darnestown just north of Washington DC. I arrived on 5 July after the wedding and stayed until the morning of the 17th. Noreen and I have known each other for 13 years. We met in Geneva when she worked for the WHO. At the moment she is between projects and is recovering from suffering a severely broken foot last summer.  &nbsp;  Noreen introduced me to her astrologer in 1994. She did a reading for me when I was based in Cambodia, and I have done about 4 since then. We went to see Eleanor the day following my arrival so that we could actually meet face to face for the first time. We had a great time and then I received my first face-to-face astrology reading, as opposed to on a cassette. It is amazing how the planets influence you. My Uranus opposition is having all sorts of influence on me right now.  [Should you be interested in a reading please send Eleanor an email at    urlLink   rolphkibrick@mindspring.com      and tell her you are a friend of mine.]   &nbsp;  Given that I have specifically told people that I am not on a tourist trip, Noreen took me on a slightly different tour around WDC. She took me around to a couple of the sights seen in the movie Farenheit 9/11 by Michael Moore. Since I didnâ€™t feel like stopping anywhere, I took pictures around the city by doing what we jokingly called â€˜drive-by-shootingâ€™. Last Sunday I finally got to see the movie as well. It was great to hear the response from the audience and I\'m really glad i got to see it in the USA. &nbsp;  We did go to the Eastern Market just behind the Capitol. I love markets and always enjoy a stroll around the stands. Next we went to a place called Adamâ€™s Morgan where they have a pub called â€œMadamâ€™s Organâ€. You can imagine the painting on the wall of the pubâ€¦ The neon sign outside the door says: Sorry, weâ€™re open.â€  &nbsp;  Around town we met various anti-Bush campaigners. In Adamâ€™s Morgan we saw a group called Run Against Bush. The organize runs around the White House every weekâ€¦  &nbsp;  Anyway, there is more to be said about WDC but I have to stop here as Iâ€™m heading out to a cottage in the Canadian wilderness for a few days. There is no electricity and no internet. Wonderful! I hope I get to see some bears.   ');
INSERT INTO `msgs` VALUES (1610,3558382,'14,July,2004','2004-07-14 00:00:00','    As I travel and stay with people I notice things, both about them and about myself. Have you noticed how people have different energy? How we sometimes are more energetic and sometimes more laid back? Some people bounce around life, while others are nervous or calm in their general nature. Of course I notice these things in the people I meet. But what I realize while making this trip is that I am totally exposed to and made aware of the different energies in the homes of the people I stay with. What is even more interesting for me to observe is my own reaction to these different energies. Sometimes what I notice are habits, sometimes rituals, or even ways of speaking, all of which give indications of the energy present in the home Iâ€™m visiting. Food habits also have an impact on the energy present as we all take in the energy of the food we consume.  &nbsp;  When you travel on the way I do this phenomenon that I am describing becomes even more apparent. It is very different to when you travel and then go back home and then go on another trip. During this trip itâ€™s not possible for me to go home and dwell around in my own energy and surroundings. I go from one country to another (yet a very different energy already there!) and from one home to the next. Itâ€™s a very funny and unusual experience. I try my best to simply â€œbeâ€ in and with the energy and the habits that surround me and not make any judgment. Observing is much more powerful than judging. I have found this so interesting that I will try to describe a couple of situations for you.  &nbsp;  When visiting my friends I try to be respectful of their habits and space, while at the same time trying to help out as best I can. While in Edinburgh I supported Margita in the kitchen. I helped putting the dirty dishes in the dishwasher and replacing the clean plates in their cupboard. With 3 kids it tended to fill up rather easily. As Per and Eva, my brother and sister-in-law, will know, â€œdishwasher-dutyâ€ is one of the things I like to help out with when I visit their home in Stockholm. Margitaâ€™s husband, Nick, also has a habit of taking care of the dirty dishes in the kitchen. Except his approach is quite different from mineâ€¦  &nbsp;  How do you operate when putting your dirty dishes in your dishwasher? Are you a rinsing kind of person or do you just stick it in there with food still visible? Do you have a system for putting the plates and glasses in? Everything in their place or do you let the cups stash where they fit in? Iâ€™m only asking so that you can start to analyse your own behaviour a bit.  &nbsp;  Nick definitely has a system for putting in his dishes. He rinses the dirty dishes thoroughly under the running tap before putting them in the washer. In fact, he nearly does the dishes before putting the dishes in the washer.&nbsp; When stacking the plates he places them in order of size and nearly in a colour coordinated manner. The cups and glasses are arranged by precision on the top shelf. According to his standards he is acting totally logically. According to my standards he is a compulsive dish-control freak. If I would have put in the dirty dishes during the day and Nick would continue with this duty in the evening, he would actually take out all the dishes I had put in, rinse them, and put them back according to whatever system he follows.  &nbsp;  The first time I noticed this behaviour, I observed Nick in total disbelief. How on earth could a person go through with all this work when having a dishwasher that is supposed to do the work for you? Margita laughed with me and said: â€œI gave up commenting about Nickâ€™s dishwasher activities years ago!â€ Towards the end of my 2,5 week stay I started to actively make fun of Nick. I said the dishes still got washed when he wasnâ€™t around, and I asked if he could notice the difference on the plate taken out of the cupboard. I asked him why he went through all the trouble. His stern comment was: â€œThings have to be done properly.â€ Thatâ€™s when I realised how attached to this behaviour he is.  &nbsp;  One night Margita and I sneaked into the kitchen and started messing up the order of the dishes Nick had just rinsed and put in the washer. We moved plates and glasses around, making it look really messy by putting plates diagonally and in the â€˜wrongâ€™ place etc. We expected him to comment but Margita suspected he didnâ€™t even notice but just went through his regular motions of putting things back in order.  &nbsp;  Although my mind really rebelled against Nick compulsive behaviour, I also tried to see things his way and decided to see if I could learn from it. I tried to see where I might need more order in my life and where I might benefit from â€˜doing things properlyâ€™. Perhaps we might all benefit from examining our â€˜dishwasher behaviourâ€™ and bring it to a different are of our life to make a difference?  &nbsp;  At the same time I really started to wonder whether Nick bring this same behaviour to his office, and whether he keeps doing other peopleâ€™s work over again until things are â€˜done properlyâ€™. Dear Nick, I can only hope&nbsp;that as&nbsp;you get older and wiser that&nbsp;you relax&nbsp;your attitude to dishes a bit.  &nbsp;  By the way, there are other ways of using your dishwasher. Noreen (who I visited outside Washington DC) uses her dishwasher as the rack to dry the dishes she has just washed by hand. The reason for this is&nbsp;that the house she lives in is so old that she doesnâ€™t dare to use the dishwasher anymore. Roberta here in Ottawa has the most old-fashioned dishwasher I have ever seen. I can&nbsp;presume it works very well since I discovered dirty dishes in there--non-rinsed, I might addâ€”when I put in my own plate&nbsp;last night.'),(1611,3558382,'14,July,2004','2004-07-14 00:00:00','    The wedding between Vidar and Malin took place on an island off the west coast of Sweden. You get to Styrso island by taking the tram from central Gothenburg to Saltholmen and then you can use the same ticket for the half hour boat ride out to the island. Amazing. This was a full 3-day wedding party, and for me it started already on Thursday.  I arrived to the car-free island in the afternoon. I quickly realised that the preferred means of transport on the island was the ever so Swedish \"flak-moppe\", or a three-wheel moped with a flat carrier top in front. I trotted up the island to the 100-year old house which was to be my accommodation. It was right next to the church. I was to share the bedroom with 2 other girls but for now I was alone. 12 people in all were to live in this beautiful Herlitz house, which had a terrace in the garden with a fabulous view of the island. The communal breakfast for all of us was provided by the bride and groom, who were very keen that their guests were feeling comfortable.  Thursday night we all had dinner at the Herlitz house. When I arrived only the family and people helping with the wedding had arrived. They were all very busy with the preparations. Hence, I helped out with the dinner arrangements for the whole gang. We had a BBQ on the terrace with the view. It was quite windy and cool, but after a day with a bit of rain it was nice to be able to sit outside in the evening.  Friday was nice and sunny and as I left the house I took a peek inside the church. A funeral had just been held, which only goes to show that it is a living church. For the next four hours I helped Bjarne, a friend of Vidar\'s and the volunteer cook, to cut vegetables for the 90-people dinner taking place that evening. The three of us helping out with the various chores of fileting fish, cutting vegetables, and peeling potatoes, had a lot of fun. Bjarne made sure to keep us served with nice cool white wine while he made the artistic work of preparing delicisous paella and two kinds of fish soup. After a small mishap of a blown fuse (imagine cooking for 90 people!), Bjarne could coninue his cooking. The bride\'s brother, Erik, was busy driving around on a \"Flak-moppe\" helping the arrivng guests to their various accommodations. It had proved quite a challenge to house 103 guests on an island with only one hotel!  Friday night drinks and dinner were held by the water in a kind of club house. Bjarne\'s cooking was disappearingly good. Norwegian and Swedish guests continued to arrive during a day which saw a lot of movement on the roads due to vacation travel. Once everyone had settled some speeches were held for Malin and Vidar. Among other items, some friends had arranged a singing match between Vidar and Malin which was extremely funny. Malin\'s stepfather gave Vidar a parcel containing all sorts of Swedish goods from blue and yellow t-shirts with crowns on them, to ditto boxer shorts, socks etc. This evening was a nice informal occasion to get to know some of the guests. There was an awful lot of wine available. 18 bottles of red, even more white, lots of beer. By the time the last party people had left at 4 in the morning only one bottle was left.... Luckily I decided to go home at midnight.  Saturday morning I was woken up by the birds at 4am. I could just see the church from my window on a day which looked very promising indeed. The next time I woke up at 9 it was pouring down with rain. Wow, was this the way it was going to turn out after we had come all this way to this beautiful island? It kept raining for a coule of hours, but cleared up in time for the wedding in the afternoon.  By Saturday morning the Herlitz house had filled up with all the other people staying over, among them 2 cute babies of 4 and 6 months. The bride\'s hairdresser, Hans, stayed at Herlitz house as well. He took care of fixing the hair and makeup for the bride and her sister. It was nice to see them get ready, while they were carefully hiding their dresses under big black plastic bags. The time was getting closer...  By 15:15 everyone had gathered in the warm sun in front of the church and we were asked to get seated. A lovely bridal march was played on violin as the couple walked up the isle. The ceremony was very personal and beautiful. Many of us cried, as did the wedding couple. After the ceremony it was back to the pension where the dinner was to be held. The wedding coule took the only maxi-taxi on the island to the helicopter pad, where they changed vehicle. Vidar drove the obligatory \"flak-moppe\" up to the pension with his wife on board. I caught a very funny 15-second video clip of their arrival. I hope I manage to find a way to post it somewhere so you can see it. It was an arrival in style with all the guests laughing out loud!  After champagne and strawberries it was time for dinner. I was extremely lucky to be seated at the long table next to the one with the wedding couple and I had a great view. I had Danish Rasmus to my left, Swedish Nisse to my right, and Norwegian Oystein opposite. A truly Scandinavian mix. Each of us had a small leaflet beside our plates. Normally this would be the menu you would guess? Not in this case. It was a complete list of the names of every guest present and a brief presentation. What a splendid idea! Beside my name it said something like:  Humanitarian globetrotter. Patient ski teacher to Vidar.  This list kept conversations going and as the evening went on it felt like I knew more and more people.  The pension where the dinner was held was very nervous because it was the biggest wedding party they had ever cooked for. They need not have worried because the food was absolutely sublime. The starter was a nice plate of various exotic mixtures of herring. The one with basilica was very tasty. This was followed by a fantastic flat fish (halibut?) and a white wine which was divine. I don\'t particularly like white wine but this Alsace was fantastic. Then followed a chicken with lime rice, but given that I don\'t eat meat I was served fresh cod instead. What a treat! The red wine was in a class of its own as well and everyone was marvelling at the food.  The toast masters made themselves known and started off the speeches. Every single speech made us laugh and cry. Every speech was very well prepared and highly interesting. And there were quite a few. The amusement factor was on top and the boredom nowhere to be seen. The one speehc which had us all in stitches was when Vidar\'s brother read the essay Vidar wrote about the cow when he was 8 years old. \" ...The cow has four legs. They are called cow legs. The cow has hair. It is called cow hair. It is brown. That\'s why it\'s called cow-lour. \" And so on... I think 8-year old Vidar had been improved along the line a bit, but it sure made us laugh!  After 5 hours of sitting down it was time for coffee and a chance to move around a bit. The tables were removed so that the band could put their instruments in place. By midnight everyone was dancing and kept at it for many hours. By 3 it was time for a typical Swedish snack: Jansson\'s Temptation. Yummy! By 4 the band had gone to bed and the staff kicked the 30 or 40 of us still partying out of the main dining room. We continued to party in a different part of the building. In fact, this part of the building was right outside the room of the priest. Poor woman and her husban... It was raining outside so I used this as an excuse that I couldn\'t go back home. By 5 the bride and groom finally decided that they should get some sleep and left the brave dozen or so still dancing. Malin really wanted people to enjoy themselves and party all night. Well, she got what she asked for!  Hans the hairdresser and I borrowed bicycles from the pension and made our way home in the drizzle. It was a lovely morning and we saw fishing boats going out to sea. We had so many giggles on the way back home we probably woke up half the island. Our bicycles didn\'t have any protection on the back tyre. Hence, the water spurted up our back making us very wet all from the saddle and up. Major giggle-enhancer. Hans sang opera on the way back. I managed to take a picture of him while riding. Once back at Herlitz house we decided to have some cheese and a beer. By 6:30 we were ready to go to bed and decided to make it up in enough time to make the brunch between ending at 13:30.  By 10 I was already awake. I decided to shower while there was still hot water (yes, all 11 of us shared one shower). Sabina who shared my room was also waking up and we chatted about what a nice party it had been. We were full of admiration and I decided I wanted to spend more time with all the nice people and decided to go to the brunch early. Needless to say, there were only a few brave people up at that hour at the pension. The priest and her husband were off to attend service at the church. They had slept little but said they enjoyed the comfortable beds while listening to the party people singing and dancing right outside their door until 05:30. Malin and Vidar were of course up in style and were very happy with the successful party. During brunch it was nice to catch up with the people from the night before and to get a chance to praise the staff in the kitchen.  During the day guests left on the boats back to Gothenburg. I left with the majority of the people at 15:00. We had had a great time and many of us were tired but happy. Poor Erik had been so busy during 3 days on the \"flak-moppe\" that he fell asleep on the boat back. Most of us felt we could do the same.  During all these days I took many photos. I am experimenting with a new site for posting photos. Please take a look at: http://ssoderstrom.photosite.com It would take me a too long to download all of them but hopefully you will be able to view a nice selection of them soon.   '),(1612,3558382,'09,July,2004','2004-07-09 00:00:00','    After Lynga Horse Farm I went to see my maternal aunt at her cabin in the countryside. Her little house is in the middle of the woods, as the crow flies quite near Alingsas (which is east of Gothenburg, for those of you who might still be lost).  I love going to Maj-britt\'s house because of the calm and quiet. The fire from the iron stove in the kitchen always smells so nice, and we take lovely walks in the surrounding woods. During this visit she showed me a phenomenon which I have never witnessed before. She took me to a big bush of burning nettles and said: \"Look at the babies, aren\'t they cute?!!!\" She meant the little brown and black worms crawling around. Well, that\'s what I thought, anyway. They were actually the larvae for a well-known Swedish butterfly.  Once I had spotted the larvae, which after a while did look quite cute, I went to check on them at least twice a day. They changed skin and went from brown to hairy black. They ate their home and moved on to the next nettle plant. They split up into two small gangs, etc etc. There was always something happening on the nettle bush while I was visiting. I don\'t know how long it will take them to reach butterfly stage but they were doing great by the time I left.  While at the cabin I finished a wonderful book which Susanne gave me. It took only a day or so to read.  The Alchemist by Paulo Cohelo . Brazilian author. The book was number one on the Swedish best seller list in the book shops. A very wise story about a goat herder going off to find a treasure. A MUST READ with lost of nice messages!  Before that I had already started (and by the time I\'m writing this also finished) a book called  Intuition by Osho . Since doing the cleanse I find myself being much more intuitive and even telepathic comapred to before. While I was in Edinburgh I just went into a book store and bought this book. It turned out to be very interesting reading. Here are a few quotes: Talking about Barriers to Knowing, Osho says: \"Millions of people carry their office home with them. Your intellect is needed. Your head has its own function, its own beauty, but it should be in its place. Theree are far greater things that are beyond the reach of the head, and when you are moving into those realms, you should put the head aside. You should be capable of that. That\'s flexibility. That is intelligence.  And remember never to get confused between intellect and intelligence. Intellect is only part of intelligence. Intelligence is a far bigger phenomenon; it contains much more than intellect, because life is not only intellectual, life is intuitive too. Intelligence contains intuition. So many great discoveries have been made not by intellect but by intuition. In fact, all the great discoveries have been made by the intuition.\"  From the chapter on Strategies: \"The heart is feminine. You miss much in your life because the head goes on talking; it does not allow the heart. And the only quality in the head is that it is more articulate, cunning, dangerous, violent.\"  This from the same chapter calls to me in relation to this trip I\'m making: \"The beauty of facing life unprepared is tremendous. Then life has a newness, a youth, then life has a flow and freshness. Then life has so many surprises. And when life has so many surprises, boredom never settles in you.\"  Strategies again: \"...the intelligence of the head is not intelligence at all; it is knowledgeability. The intelligence of the heart  is  the intelligence, the only intelligence there is. The head is simply an accumulator. It is always old, it is never new, it is never original. It is good for certain purposes; for filing it is perfectly good. And in life one needs this--many things have to be remembered. The mind, the head is a biocomputer. You can go on accumulating knowledge in it, and whenever you need you can take it out. It is good for mathematics, good for calculation, good for the day-to-day life, the marketplace. But if you think this is your whole life, then you will remain stupid. You will never know the beauty of feeling, and you will never know the blessing of the heart.\"   And then from the afterword:  The distinction is subtle, but it is the same distinction as between the mind and the heart, between logic and love...between prose and poetry. A destination is a clear-cut thing; direction is intuitive. A destination is something outside you, more like a thing. A direction is an inner feeling; not an object, but your very subjectivity. You can feel direction, you can not know it. You can know the destination, you can not feel it... ...Destination is fixed by the mind; direction is earned by living. ...You cannot decide direction, you can only live this moment that is available to you. By living it, direction arises.\"  This trip that I am making has a lot to do with direction and living. Although there are quite a few destinations on the trip, while at the desintations the trip has much more to do with direction and inner feeling than anything else. This is what I am exploring with the friends that I visit.   '),(1613,3558382,'07,July,2004','2004-07-07 00:00:00','    Yes, I am back on-line again, after an unusually long absence. But it so happens that it is 10:30pm here in Washington and I\'m still jetlagged.... Therefore, no lenghty posting tonight. But if you are waiting for something good.... :-)  I have a lot to write about, a couple of books, a nice visit to my aunt, the wonderful wedding on the west coast of Sweden, my arrival and past couple of days in Darnestown outside Washington DC.  See you soon!   '),(1614,3558382,'09,August,2004','2004-08-09 00:00:00','    Well, we left Ravelstoke Thursday morning and drove all day in the rain all the way to Abbotsford, Sharilyn\'s home town (and yes, we did stop in Kamloops on the way there!). That evening I was feeling so tired and could feel my throat getting sore. I have spent the past 3 days in bed with fever and probably a strep throat or something. Yesterday I was feeling a lot better and finally today I dared to venture out with Sharilyn and the kids to buy yummy blueberries.      Here in Abbotsford we have been staying at the (non-air conditioned) house of Sharilyn\'s sister and her family. They are all up at the cabin by the lake until Sunday. Right now I am sitting in the nice air-conditioned house of Sharilyn\'s parents. It feels great, especially since it is +38C outside. Luckily it rained the first 2 days when I was sick... We will have dinner here in a nice cool temperature before heading off again.      Imagine my surpsrise when we went to the blueberry farm this morning. I was so mad I didn\'t bring my camera. They had a huge blueberry-picker which went through blueberry fileds the size of vineyards!!! The bushes don\'t look anything at all like what I\'m used to seeing in the forests back in Sweden. We bought a 5kg box and really stuffed ourselves with blueberries for lunch. Yummy! Some of the berries are as big as cherries. Amazing.      In Ravelstoke I didn\'t even have time to write about the trip. I had to buy a time card to log on to the internet and although I bought 30 minutes it ran out after 15. She gave me another 30min card and that one ran out after 15 too. Hence my hurried typing. We had a really great day that first day of driving. THe stop at Lake Louise was really nice. There is a huge hotel up there and there were so many tourists heading up there that we had to wait in a queue to get to a parking. Sharilyn (aka Samy) kept asking me to look out for wildlife along the road during the drive, but it was sparse. We saw tame gofers begging for food at one of our quick stops. No bears... boo hoo...      Tonight we do the one hour drive to Vancouver to visit Samy\'s in-laws. They have kindly offered to babysit for 3 days while we go to visit Vancouver Island. That promises to be a great trip. Samy has been planning this since February when I told her I would be coming over. I\'m so glad that I got well again so that we can do it. If you look at a map, we will be going to places like Victoria and Tofino. The ferry lands in Namtua. I hope it is a bit cooler there with a bit of a sea breeze.      As for the subservient chicken, I tried to get it plucked and grilled, but it didn\'t want to obey... Did anyone else try except my Whales&amp;Wolves friend?   '),(1615,3558382,'04,August,2004','2004-08-04 00:00:00','    Buck Lake is a small lake SW of Edmonton in Alberta, Canada. That is where Sharilyn and Bruce\'s cabin is. Sharilyn and I arrived there on Saturday afternoon after our lovely spectacle at Cirque du Soleil. The cabin is great, an orange house with red roof and HUGE logs inside holding up the ceiling. It takes 30s to walk down to the lake. They have a hot tub on the deck outside the kitchen, where we relaxed every evening.      Bruce and little Kieran took me out fishing in their paddle boat (they haven\'t had time to buy a proper boat just yet). I spit in the water three times, like any good Swede would do, to brign me good luck during out morning fishing trip. The boys were very good to me, giving me the best lures. Well, I caught 2 pike in two different spots. Unfortunately, neither pike met the required 63cm for brigning it back home for dinner, but I sure had fun fishing all the same. The poor boys didn\'t get anything that day so I thanked the spitting!      The next afternoon Sharilyn went fishing with the boys. Kieran caught 2 Walleye. Sharilyn, who took my advice and spit 3 times, caught none other than 5 Pike and 4 Walleye!!! She was extatic. Poor Bruce didn\'t catch anything, until the very last minute when a little pike felt pity for him and got hooked. Sharilyn was so proud all evening, while the boys were sulking a bit.      In the evening last night there was a big thurnderstorm at the cabin. Bruce had gone back to Calgary to work and Sharilyn and I were out on the deck looking at the lightning show. She told me about the tornadoes that sometimes hit, and how it is best to hide in the bathtub if you don\'t have a cellar (which their cabin doesn\'t). As we were still giggling from this and the lightshow and just about to go to bed at 22:30, there is someone at the door. By this time the rain had started to pour down quite heavily and there was a strong wind. Diane Love, one of the neighbours, told us that there a tornado had hit Rimbey about 35km away and was heading to Buck Lake. They had the truck outside out house, rady to evacuate us to the neighbour\'s basement! We ran to put some clothes on, got the kids out of bed and hurried outside. Phew... What a suprise! We spent the next hour and a half at the neighbour\'s basement watching the Weather Channel. It soon became clear that the tornado had died down and had gone on the other side of the lake. The kids were having a grand time playing with toys and we were given drinks to calm down.      At midnight they drove us back without anything having happened, but it seemed the bad weather from Ottawa and the cottage finally caught up with me! :-) We had a shorter night\'s sleep than planned before our road trip, but it felt good to be back in our beds again.      This morning we left the cabin to go past the Rocky Mountains. We left at 08:10 and stopped an hour and a half later in Nordegg, a tiny sleep little settlement in the middle of the woods. We had our breakfast at the Coalminsers\' Cafe together with surprisingly many other guests (there were not many other places around that area to stop). The room smelled of nice home cooking, but mostly oily fried dishes. The funny thing was that all the chairs in the room were old office chairs with wheels and armrests. I\'ve never seen a cafe like that before. The kids were very well-behaved and once done played outside in the little coal train before we took off again.      We got to Lake Louise for lunch. We ate at the Deer Lodge which was a nice oldfashioned place.  Later we went for a walk around the lake and watched the wonderful glacier-green water. The rain caught up with us again on the way down. Tonight we are sleeping in Ravelstoke at  urlLink www.regentinn.com        My time is running out on this thing so I\'ll write more tomorrow from Kamloops...   '),(1616,3779062,'28,June,2004','2004-06-28 00:00:00','   So, I am doing this because I have been told by a friend I need to rant to the world, and I think he is right. Although, today is ok and I don\'t seem to have anything to rant about. Oh well, the day is young, I am sure that will change.   '),(1617,3779062,'30,July,2004','2004-07-30 00:00:00','   I am bored. I am bored. I am bored. I am bored. I am bored. I am bored.   '),(1618,3779062,'24,July,2004','2004-07-24 00:00:00','   So, I just found out my grandmother died last night. Apparently people paged my at 2 am but I didn\'t hear it, so I just found out. So, why don\'t I feel anything? Is it because it was expected? I think she is in a better place now. She is no longer suffering, which makes me happy. And now I have to go to a wedding today. Not really what I feel like doing, but oh well. At least I will be with friends, and they will understand. And they will make sure I get nice and pissed tonight, which I think I need.   '),(1619,3779062,'19,July,2004','2004-07-19 00:00:00','   O  k, so you ever have those days where you just feel like shit? I am having a bad one today. My Grandmother is in the hospital dying and has been for the last few days. But because of my work I haven\'t been able to be there much. So, I went to work this morning and just felt horrible being there. It was actually making me ill. So, I left at lunch, and now I feel guilty about leaving. Oh well, I think a day will help me. I need some time to process things. And I will do a shift at the hospital tonight with my mom. I just really wish there was something I could do. I hate being helpless. Fuck, this sucks.   '),(1620,3779062,'13,August,2004','2004-08-13 00:00:00','   So, Friday the 13. Hmmmmm. Either something good will happen or something bad will happen. Or maybe it will be like the rest of my life and nothing at all will happen. I have rediscovered why I don\'t stage manage any more. I am so bored watching this show. And I would feel bad about sitting there and reading. Grrrr. Well, only 4 more. And if one more actor comes to me with a brilliant idea, I am going to have to kill them. Maybe if they knew all there lines or all there songs or didn\'t goof things up every night, I would listen to them. But I wonder why they are worried about other peoples blocking, but don\'t seem to care that they have no idea what they are doing.   '),(1621,3734069,'30,June,2004','2004-06-30 00:00:00','   yay.. power came back.. got pple to chat with.. hahaa.. yah.. going to choir pract later.. with fish & cat.. long time that the three of us sit together dUring pract...  hmmm Vicky jUst pointed Out somthing while we chatted.. hmm wat a coincidence that the collapse of the nicoll highway.. the previous major black out and the back out thAt happened cOuple hours ago.. all fall on tuedays.. hmm??  hmmmm.. was watching Guess3.. hahha.. they were picking the gal with the Most-teh voice.. hmmm all i can say.. their voice not equals = their age/ looks...  hey my vOice teh meh? hmmm i knOw its kinda high pitch.. hmmm.. :)  hmmm tiREd.... sleeping sOOn... haha said tata to long.. liao.. ..hahha but i continued blogging.. can\'t gO to ktv with my cell pple..got chOir prac.. :) hmmm see hOw.. ah muS call pris tmrw.. to check on cuiling... & her frien..   yay.. next week gOing book of Is part tWo..  ~~*~~** verse for the day~~*~~** :  Mt7:11~\"If you then, being evil, know how to give good gifts to your children, how much more will your Father who is in heaven give what is good to those who ask Him!\"   I\'ve asked and yups... He will briNg it to pAss.. :)  gOt an email to ask to volunteer to prefOrm for national day and/or D&D .. hmmm wOnder got time or not.... aiyohs... prAy abt it... :)   '),(1622,3734069,'29,June,2004','2004-06-29 00:00:00','   hmmmm majOr poWer failUre in SingapOre... bored ah... noOne online... :( all my talking kakis kanna pOwer failUre.. haha.. think I\'m the oni one that\'s left standing... sighsigh... hmmm am really fortunate ah??       hmmm haha... rachel called me & complained abt the screaming pple.. then after i put down the fOne.. saw joyce online.. haha she oso kanna hit but she was using her laptop.. now offline le... cuz bat low .. then long sms me.. haha... his place oso in daRkness... cat oso... but her area okie le...hmm thank God my area not hit.. bUt hmmm now bOred ah...think i\'m a chatterbUg... ahhh...   '),(1623,3734069,'28,June,2004','2004-06-28 00:00:00','   Hmmm took me quitE sOme tiMe to reVamp.. phEw.. thInk nOw it lOoks bEttEr... hmmm am nOw @ epiSode 8 of mo nu de tiao jian.. (yeah. i noe that channel8 showed this loooooong time ago....) hmmm still it\'s a gOOd shOw..  hmmm like to waTch jap dRama... tHe mosT reCent oNe i caught was gOOD luck... cUZ that\'s the lAtest thingy that my btother bought...   tired? nope.. stAyed for seRvice 4 after chOir... hmmm prEsence of God much thicker for service 3 though.. (fish & her dar thot so too..)  sAng for chinese cHurch.. haha.. Father of creation kinda diff tO sing in CHinese.. but then again.. i\'m qUite ok with it.. :)   kk it\'s late... gotta WoRk tmrw... Gotta complete another 20+ caSes by 30th.. (left 3 days.. pRay hArd for God for prRoduCtivity!)   '),(1624,3734069,'27,June,2004','2004-06-27 00:00:00','   Hmmm am really refreshed.. :)  His prescence really tangible.. let me meditate then I blog ah.. haha.. just gave this bloggie to dear vicky & long..   bear with me... btw i love this song... cuz of mo nu de tiao jian... & the lyrics as well.. (do understand a bit.. the chorus.. by the ear.. jap still okie la.. know abitsy.. rather rusty though...)   '),(1625,3734069,'26,June,2004','2004-06-26 00:00:00','   Hmmm am really blessed by today\'s prac( i mean fri nite)... truly felt ministered... Love the new song that we learnt.. the Source! Truly.. He\'s Our SOurce!! This song was written by our own members! :) learnt another 2 songs...   wow... power..somehow what brother jae said during worship really ministered to me... almost all the words that he spoke was like wham bang... His presence was truly tangible... GOd you\'re amazing... Lord I will go wait upon YOu now.. :)  really chatty tonite.. prior to calling pris.. was like heehee.. airtime still got spare.. yay.. on budget this month.. but deep down in me.. i knew that I did not do what I should have done... hmm called pris.. ai.. she went service with cuiling.. hmm talked to cuiling for abt 5-10 min.. talked to pris for another 5 min.. then called lian... hmmm dunno why just came with this thought that our group need to sit down ans discuss a stragety.. If YOu fail to plan, you plan to fail... so talked to lian on fone... really chatty mood.. talked with for close to 40 min at viod deck..cuz i knew that mom would not be plasese.. so ya... had a good talk with lian.. then came home.. chatted online with vicks... ( was alredy done that while smsing her whil talkiig to culiing... )  am glad to be out of the san jiao le..  okie.. gotta go.. Ja ne !   '),(1626,3734069,'25,June,2004','2004-06-25 00:00:00','   Hmm am very tired now.. haha.. Robin.. U get a preView to the Other side of me.. hmm actually U\'re the 1st person that I tell of this site.. hmmm what a coincidence that we both created a blog on the sAme day.. :)  Cell groUp was great.. well.. haha.. when darren was summarising of the bad habits to be changed during cell grp discussion.. he was talking abt.. bad habit to sleep late at nite like 3-4 am.. haha.. though i know that the guys prolly stay late at nite to watch euro? this crazy lass typing here.. just stay up to?? haha my inspiration window is after 12 midnite.. and ideas just flow while the time ticked into daylight.. hmmm time is now 1:40am.. gotta sleep am wOrking later.. and yay! got choir prac later.. :)  hmmm am still exploring blogspot.. different from xanga.. haha managed to write a link to have some music in the background.. okie.. i need to pray.. and sleep.. haha just now i nearly made a roundabout trip to boonlay... i doze off @ bedok and was @ pasir ris for 5 to 10 min before i woke up.. and Praise the Lord that I woke up in time!!   would i forsake my xanga? haha..i dun think so! :)   '),(1627,3734069,'24,June,2004','2004-06-24 00:00:00','   Yo!! Created my sECond blog... :) kinda bored with xanga... see how it goes.. :) am doing this @ the reSource centre of my organisation.. haha.. just realised that I\'d forgotten the lyrics.. so popped down to surf guitar4christ.. :)  Just finished a sharing.. I think i was speaking in tOunges.. ah..phew.. people still understand...   okies.. back tO woRk..   '),(1628,3734069,'30,July,2004','2004-07-30 00:00:00','   Hellos... Just hang up with Lian.. heehee.. she sms me&nbsp;@ ard 12:10 to ask if I\'m sleeping.. (something she doesn\'t know about me... )&nbsp; so I called her.. so vick and long if you happen to have long pauses while chatting with me... pardon me....     &nbsp;   btw... yili... thanks la... i know when i sms last wed.. u had your ups & downs too... but like what Brother ryan said in cell today... it\'s feels really good to turn ard and stll see ur fellow bros and sis around... i turned and saw u and vick... :) i think 3 years from now... we will definitely still be around and definitely another level...&nbsp; I was kinda&nbsp;not too&nbsp;good&nbsp;when still over min stuff when Brother ryan was preaching... hey let me back track..&nbsp;     &nbsp;   actually my day @ work was pretty yucky cuz i was way behind schedule... so ya..with the \"break\" from min.. and the horrifying national day thingy next week... i\'m really stressed...&nbsp; (it was soo last min that I did Chris & jacelyn card ar 5+) and managed to do the songsheet @ 2+...&nbsp; so was kinda worn out by the time I got to Ziv\'s @ 7.. nope i didn\'t have dinner... was practising in the room with sis siew choo & brother ryan...and gutiarist&nbsp;long as well...&nbsp;suddenly i feel so \"extra\" that why am i dOing here... what i lacked... ya.. a douse of self-esteem. .so times it\'s like i will demean myself.. negative thots... so it\'s like what my \"daughter\" Claire &&nbsp;&nbsp;sis&nbsp;Gwen&nbsp;shared with me.. (reinforced by bro ryan) counter them with God\'s promises...And man.. during p&w.. i felt my hair stood up... (it did not stand up literally) can feel His presence in the room.. then when we prayed in tongues... man... His Presence was sooo strong that all my hair stood up (somewhat like being in a&nbsp; magnetic field...) and even my neck.. i was trying hard not to feel tickled.. and giggle.. cuz.. man.. but it was good.. times of refreshing... :)     &nbsp;   then during the sermon.. bro Ryan shared this... that when God called me... He has already prepared me... i.e.. if i fail to perform/deliver... it\'s not that I\'m not called.. but rather... I neglected what I was to do... cuz i was rather :( due to the break... but... Holy Spirit spoke to me... that I was indeed walking in my calling... just that I slacked and took a detour.... what i yearned for will not happen overnite but... surely it willl....&nbsp; thoughts of dropping choir over cell came to mind.. cuz last week.. i was off.. and with the cell... somehow... I was more happy serving the cell...&nbsp; cuz not doing choir let me interact with the members, talk to the new friends.. to really serve.. I may not know when, how.. but surely, it will be sooon... i need to buck up... and step Up & OUT!     &nbsp;   hee.. i managed to talk to lian..&nbsp; to \"psychoed\" her to jOin NTU min... (cuz she wasn\'t that willing.. \" heehee&nbsp; PTL...&nbsp; am glas.. that I have the oppty to serve cell again this weekend... but am very :)&nbsp; to see that I\'ve left my mark (hahaha) in vick and lian... cuz sometimes when i give them bs.. I lack the faith to deliver... but all u need is not power sermon.. but a good and open heart... heehee    going o\'nite prayer meeting later.. yay!&nbsp; but gving&nbsp;VIcky bs first... am very happy to see pple growing.. : ) i need to too... ttataat!!   '),(1629,3734069,'29,July,2004','2004-07-29 00:00:00','   heehee..    was supposed to be the make-up model for Cat.. ended up being the model for ivan (eh.. my lips had a brief encounter with his palm... not very nice feeling... ) anyway.. the theme was 1930s.. aniwae cat did my foundation and eyebrow (1930s look ~ Starlet eyebrow ~ aka one line shaped like a C turned 90 degrees clcokwise) then she became model for another make-up artiste.. then me became the model for ivan who was i learnt from cuiqi\'s friend.. that it was his 1st time too.. 1 ) man... I\'d a bitsy of shock when he told sis irene that he doesnt know how to use the eyebrow curler.. (man.. i was thinking. there goes my eyelid.. phew, he was okie la.. )... poOr guy doesn\'t own make-up.. he only got make-up set.. can\'t blame.. he\'s a guy.. and new..&nbsp;ghmmm&nbsp;despite the color&nbsp;used on my lips was blood red (that\'s&nbsp;the IN color for&nbsp;1930s.. ) i like the&nbsp;taste of the lip color (think it was&nbsp;a&nbsp;chanel palatte).. I think brother ben like the red color, he saw and was telling the make-up artiste that he was tempted to use that red on one of the BV this weekend.. but dun tink the BV might be willing.. heehee.. it was bright red.. very 30s feel... and I won;t wear it normally...&nbsp;and ivan was&nbsp;of so close proximity with me.. i can smell his hands & breathe,, ehh .. advice to him.. eat mints ( but didn tell&nbsp;him la).. he was pretty yOung and&nbsp;kinda shy.. but got the cool look...&nbsp;&nbsp;&nbsp;iit was fun.. heehee cuiqi was sitting opp me.. heehee...&nbsp;     &nbsp;  took cab hOme with cuiqi and her friend.. somthink-shan cuz we all are easterners, cuiqi stays @ meville park ( whoops,, wOnder why never bump in to her before..&nbsp; when we were at eugene;s godfather;s), her friend tp..     &nbsp;  reached home.. tired... and ya i still need to birth 2 cards... and i\'m very tired.. I QT first then see how,,, :P    &nbsp;  yili.. tell me wat songs ok.. hmm i\'ve faith that the original praise can kick la.. cuz when i saw his msg when i got home.&nbsp; got a shock cuz only saw that song once in svc.. for emerge ppl sligihtly more times...     &nbsp;  ahhh tired... zzzzz     &nbsp;   &nbsp;   '),(1630,3734069,'27,July,2004','2004-07-27 00:00:00','   man.. today I went to office but I didn\'t wOrk... It\'s not that I dOn\'t want to.. but i was involved in the the ticketing of the national day celebration thingy.. spent morning consolidating orders.. then deposit $ ... tried&nbsp;calling couselee&nbsp;then makan lunch @ food Junction Novena Sq.. hey. the claypot chicken rice not bad .. the rice is actually bUrnt! yumz..     then spent the whole afternoon from 12:50 to 5:30 to \"fill 35 branches\' order\" totalling to 16K .. yah. felt like a production worker... (all by hand.. and gotta re-perforate the sheets cuz sloppy wOrk done by the asst. social sec.. the hon social sec ahh... love the chinese martial arts tai chi... she got great eye poWer! ya.. i \'ll stop lamenting.. ) Interesting to learn that my catholic friend\'s hubby is a free thinker.. she asked me and my friend (both of us are Christians) that CHristians have to marry Christians right (cuz I asked her doesn\'t her hubby need to be converted as well)... she said she doesn\'t want to impose her religion onto her hubby.. cuz all religions all meant abt doing good.. so long that the right moral values imparted to her children, she doesn\'t want to force her children (in future) to be catholics despite she would want to instil the catholic teachings to them.. I was like HUH??? Told her that it\'s not that we have to marry fellow Christians.. rather.. we believe that we should do so... (the other sister in Christ did not challenge her.. ) I would definitely want my loved ones to have eternal life... man... how can U accept the fact that you\'re going heaven and hubby going hell.. man... she felt that all religions boil down to the same.. doing good.. manman.. WOrks salvation.. man oh man... Argh.. felt that arguing against that but it ain\'t doing much good cuz I still need to work with her on the national day tingy &DnD...&nbsp;     &nbsp;   ahh left office @ 2230..haha max called me.. ask me if I\'m interested in getting ink pads & stamps&nbsp;(ehh actually i oredi have soMe.. ).. told him nope.. and ..etc.. then finish up my national day tingy packed and left.. man.. Bro Ryan called me 6+ to follow up on Mary.. btw she\'s my counselee that received salvation last sat.. (man I procrastinated in follow-up..resulting in my last min call this morning.. ) and give her reassurance on salvation.. man.. had the notes oredi.. never got the chance to use it.. except the mock up bs that Brother Andrew made me, rach & raymond to give him last year (b4 the reshufflement) hahaha.. had proceeded long way from there i think... (vick, lian rite?)     &nbsp;   okies.. so i was left with no choice... finally when the train came over the tunnel @ bedok.. (it was already 9.. I called Mary.. was actually praying silently for wisdom to talk to her cuz (a) I didn\'t get the chance to talk to her after service last sat (b) she is 20 years my senior (c) i really dunnot what to talk to her on (d) i was trully clueless on how to start the cOnversation... man i was nervous when i talked to her, cuz like talking to someone that can be my mom... took me a couple of min to warm up... hmmm the flow got better when the train progressed nearer to home.... Thank You Holy Spirit for annoiting and wisdom to answer her questions.. and when i reached pasir ris..walked home cuz can\'t talk on bus.. noisy la... man.. I think she was really touched .. heehee.. and asked me where i stayed, when she realised (she stayed at bukit batok) it was that far, told me that if i stayed nearer can come to her place and chat (I was like hahahahha.. not bad la) heehee... i amazed myself... I talke to her for 20 mins!! hahaha...thank You Holy Spirit for Your annoiting!!&nbsp;yay! I\'m bAck!    when i was in n181.. i talked to xiao meimei.. now in w110 i talked to aUnty..&nbsp; LessOn learnt: I need to be able to relate to pple of different age grp.. dOn\'t let my age be the limits... heehee...     tmrw need to go office early cuz i hope myday wont be bUrnt.. need to clear my Workitems by Fri.. and tmrw I\'m gOOna be cat\'s model for the 1930\'s theme.. she gave me this link   urlLink  http://www.sensibility.com/vintageimages/1930s/images/simplicitybeltdress.jpg   &nbsp;   hahah, how can.. hmmm if it looks good.. I post online hahaha... :P     &nbsp;   tmrw i will call my girls up... and the uncontactables... aHHh.. pHew. me not the main shepherd yet..     &nbsp;   '),(1631,3734069,'26,July,2004','2004-07-26 00:00:00','   heehee... just finish toking to rachel.. 2nr round of idle talk.. first was with yili then he zzzzzzz then rachel...     heehee rachel... heeeheee.. thanks fOr ya compliments! hmmmm let me see...  u\'re oNe of my clOset friend... heehee like wat i say.. my dearest friend wOuld be thOse that have sOrta some kind of misunderstanding heehee.. dunnOe why ...     wellwell... just now had a fantastic celebration @ siew choo\'s...&nbsp; hmmm on dunday..ming and i was planning.. hahha though ultimately he came knowing that we\'re celebrating for him.. i can see that he\'s tOuched... :)&nbsp; very blessed.. emmy drove me home.. after dropping ryan, then vick@ hall15.. then me before heading back to bukit batok with yili... THank U.. emmy.. U\'re sooo sweet!! :)  let me give thanks to God for giving me the creativity & productivity.. left office @ 530... reached Spotlight @ 550.. man.. still dOn\'t know how to make the card.. ahh all i have in mind is normal card.. thot of mounting board... can\'t find in spotlight.. hahha in the end thought of buying paper only then get the board from popular... then i saw the placemat of hardened paper mache material... then saw the heart shape one.. then the idea just came to mind.. hmmm.. then thot of using the cellgrp fotos.. (sad ah.. i sacrificed my 2 cell foto &amp; oso the 2003 emerge fotos.. haha.. gotta reprint...) really love the final product... too bad.. ain\'t mine...     hmmm somehow it\'s the feeling... i mean if the inspiration comes..it comes... I have no plan.. ya.that\'s me... hmmm certain pple would have nicer cards.. I\'m not pian xin.. but.. that\'s the flow .... so can\'t mass produce or make \"mass cards\" in advance.. that\'s why none of my cards are alike (frankly, i dun recall wat I\'d done... maybe i should have taken a foto.. heehee)     aniwae.. we got ryan a 2 piece suit from X2... man.. way exceed budget ($200) that munhon gave.. dun disclose price la... but I think he deserves it... :) ard 15 pple turned up.. heeeheee... more bros than sis.. (but that\'s representative of my cell.. ) but good fellowship... hmmm funny.. just recall the 1st time I joined w110, i jOined them for the monday prayer meeting.. on Aug 18... hahah it\'s siewchoo\'s bday ( and i din even know, so no cards. ) we had&nbsp;a mini celebration!! &nbsp;heehee i started the celebrating bday tradition there by cards... hmmm Ryan was rather moved when some of us tOld him of our thots of him... hahha, as usual.. i was silent.. heehee silence speaks a thousand words.. But really I\'m indeed very blessed by Brother Ryan... perhaps I\'m under him while I was at a more matured stage... (but this one year in w110 was my most eventful year of my almOSt 3 years of Christian Walk... )... aniway, I really learnt lots from him.. and i really wanna thank him for the gentle shakes he gave... (hope there won\'t be 3rd time... ) and it\'s a real privillege to serve the cellgrp alongside him... as well as the rest of the helpers...     AS for my the other cgl.. bRother Andrew... he\'s my inspiration as well... personally, i must really thank him for his painsaking effort for making me to do my QT.. and him being there when i needed help... and he\'s really annoited.. can\'t hide much from him... and his pretty wife ..gwen.. heehee... one of the person whom i grew rather close... hahha sOmetimes like to tease me abt her pri sch classmate.. but she sOmeone that I will turn to when i got my pROblems.. veryvery easy to talk to...  hmmmm latelate le.. I gOtta zzzzzzzzzz but pray 1st...     later will be another hectic day for me... gotta go fulfill my duty as a social sub-commer... the coupons sale for our national day celebration.. then afternoon gotta learn the ketchup dance cuz they will do a MV filming for D&D... and ah... D&amp;D.. (fortunately.. i need not perform again for D&D.. cuz in the sub-com.. hOpe Oct no big day.. if not.. ahhh) and on top of that I still have not churn out my letters to Singtel &amp; pwc ahhh.. very technical... and i\'m dry on technical knowledge... and my backlogs... ahhh ..things are soooo contradictory.. .the less time you have, the more you need to achieve.... and on top of that.. my&nbsp;spiritual walk recently is kinda spiral... up UP Up.. then down down down.. and Up.. esp last week... but thanks to my dotter to help me see some light... had scribbled this on my blackboard @ work&nbsp;:&nbsp; \"A winner is one not one who never fails, but one who never quits\" Dr Edward Louis Cole ...&nbsp; many times i wanna give up cuz it\'s just sooooooo tiring to go on... cuz I\'m like the car with flat tyre... & w/o fuel... I need more of Holy Spirit.. more of his power/ strength/ annoiting...       To achieve more with less!   I need to clear 7 cases with 30 workitems by end of this week.. and i can\'t OT that much cuz mum not happy i not at home... and i\'ve to help cat on wed.. and thurs got cg..and tues \'s gone... ahhhh&nbsp; why i agreed to help in the sub-com knowing it\'s DnD year... ehhh... hmm it\'s all in His plan..   To achieve more with less!  &nbsp; ooo! I haven\'t cOntact the new friend that responded to altar call last sat... result of procrastination.. gotta get her by 12nn... I need to priortize!!   To achieve more with less!&nbsp;   ahhh boss just passed me something to read b4 i left with regards to the research i need to do but shelved it for 2 weeks!! Holy Spirit.. help me be an efficient &amp; effective persOn later...   '),(1632,3734069,'26,July,2004','2004-07-26 00:00:00','   Yo!.. My eyes are like slit... I wanna sleep soon.. haha...later will be fun... why? cuz we\'re planing a sUrprise for our cgl.. man.. everything was such a rush rush... the concrete planning started @ 21:00 .. started with idle chat with ming... then.. wellwell.. cut the long story short.. decided to plan a surprise for dear cgl.. orginal plan was to squat at his flat till he come back.. heehee.. then chris came to mind..    hmmm am tired.. i continue when i can... man... i need sleep... generally happy cuz.. heeheee...Don\'t tell u why.. heeheee..btw i changed my music again.. no particular reason.. :p   '),(1633,3734069,'25,July,2004','2004-07-25 00:00:00','   Saturday.. 24/07/2004...  Today.. was the day that I smiled the most... And the day I met a no. of  U4s, U3s...   okie.. brief updates on me day..  10:45  woke up... gee i need to meet nancy @ 12nn@ Clementi..can\'t made it in time.. sms her.. meet her 1245 instead..  11:00 prayed..did what daughter suggested.. man.. it was good..  11:35 showered..and did slight make-up.. cuz of my incredibly dark rings under me eyes..  12:00 left house.. gee i was late.. took a cabbie down...  12:35 reached clementi.. man.. nancy not here yet... waited.. and waited.. 10 min later.. she arrived.. lunch with her.. catch up... trying to encourage her... man... she going thru\' this phase of \"no motivation\"... man tried to encourage her.. and invited her for service.. Straight rejection... NOpe.. church is not for me....treated to tang yuan.. hey it\'s yummy.. i like the peanuts one.. the seaseme one was yummy too.. but the kopitiam very ulu... hoPe to instill more hOpe & jest into her... hmmm she asked me whether wanna go wildwildwet.. told her.. after aug.. hey wonder my cg wanna organise outing there. can get her integrated.. afterall she is of vick, lian, joyce & long\'s age..   15:30 boarded the Clementi bus.. called lian.. cuz she was late.. I offered to pay cab la.. (cuz her friend almost fly plane).. Praise the LOrd... they were coming down.. man.. sms... vick wondered how to book 40 seats..(cuz she & joyce, the usual seat bookers were bringing their friends) told her that no worries we would settle it.. man.. I haven\'t been doing this for so long.. miss doing cell grp hounding and meeting friends.. haven\'t been able to do it since serving in choir.. am getting the hang again..  15:50 reached chUrch.. met lian & yeeling.. friendly girl.. chatted a lot with her.. ( I\'m suRprised at the degree of friendliness of mOi.. haha... stePping out & up.) took her ard the building.. and then down to service..hmm Praise God.. I managed to find 3 seats on floor (actually it\'s one whole row with only two sisters by the aisle.. so grabbed the seat... drama was funny... noted that some changes were made... hilarious... (but no drama for s3... ).. was ultra-hyper... (ahh saw msg from siew choo to take counsellor packs..at 645)... hmmm yeeling didnt respond.. :O took yeeling to lvl 4 aft shc.. hotdog was yummy (shared half with lian)..went riveria took ice cream for her.. then walked her to busstop to take bus.. man... I wondered how i managed to find sooo much to talk to her... hahha   1850: made a mad rush to riveia... hmmm not late.. dexing was there with siew choo.. hmmm ryan asked me i not singing ah.. haha told him no... didn\'t tell him why... i think i should la.. will do so tmrw... gee... reached b4 at 1900 ahh yili only booked 2 rows.. ai yos.. but then later managed to get another 1 at the creche area...  haha joshua, yushan, wanling, fish,steph..and many others asked me why i not singing.. haha didn\'t know that I\'m rather prominent.. heehee... but gotta do what i gotta do... :) stood next to sis val during p&w.. hmmm surprising dun have the weird feeling that I thought I might have... :) have a feeling that this might not be that long..  ahh went back to creche area after pnw.. gee they did not carry my bag up.. ah... man.. surprisingly i wasn\'t angry or kick up a fuss (am amazed at the patience & calmness me got) haha joked with max.. I feel like a new friend.. hahaha.. during the altar call.. I just had the feeling that some in my row will... man.. 3 did.. PTL!!   after service... did my mingling.. hahaha.. once again.. am sUrprised @ how chatty I was.. hahaha.. got the flow again....  hmmmm walked vick\'s friend weiling to busstop.. she very U3.. but ok la...   then vick met her sec friend.. it happened that he is in brother Justin cell... (heehee but he rather new.. and his attitude just remind me of how vick was when she just joined us..he was rather uncomfy abt the size of his cg.. think it\'s rather large.. but am sure he will move from glory to glory..)  22:10 went back to church.. karen came over to talk to me.. ahh..I didn\'t know that ziyang, shaoyang & monique are under me permanently... (man!) know that she\'s not happy that I\'m not doing a good follow-up on them.. but her words rather ok.. tone stern.... (i think if it was brother andrew...sure kanna scolding one.. ) taught me the \"techniques\" required to deal with them.... ( Till now.. dear monique has not returned my calls and smses.. and I even called ziyang\'s home.. eh.. he was out.. his housemate ask me to call the hp... Man... I called the hp like... ah.. still no reply... )shaoyang not feeling well... so maybe that\'s why he no reply... PTL brother ryan giving him to yili..  but I got the more challenging one.. ( to think that they are the very 1st friends i got to know from chc.. and now I need to shepherd them.. ehhh... kinda weird feeling...)  ~~** now dear hermione gotta sleep... yay! next week still on \"choir vacation\".. yay can do more cellgrp stuff.. i really missed doing those little things like cellgrp.. actually the reason for my problem was solved.. but let time take its coUrse..glad that I can be there with the members when they bring friends.. (i told my gals.. if u bring friends, let me know.. I\'ll go dowm with you) cuz if I don\'t get to talk to them or let them see us.. how to integrate.. hahha.. 0211 now I need to wake up @ 6 cuz my ezlink is at defict $0.30.. need to walk to mrt.. taking bus 1 @ 7:20.. meeting susan & friend ; emmy &  cousin ~~**  yay.. I\'ve found it again...  Praise the Lord!!   '),(1634,3734069,'24,July,2004','2004-07-24 00:00:00','   hmmmm I was talking to claire for the past 2 hOurs over the msn.. :)   man... Thank You claire... for \"whipping\" me back into shape/ back to focus...   haven\'t talk to aniOne abOut everything... (hmmm i cut the story short so.. she didn\'t get the chance to hear abt my family wOes.... esp the histOrical faCts.) come to think of it.. she\'s the One that hear the whole stOry abt my salvation.. (hmmm ehhh andrew knows oso.. cuz remember there\'s one time he ask us to write down our testimony ah during cell but.. I think no one did but me.. cuz he saw mah.. \"privilege\" of travelling home with cgl..) and the most recent downturn..   anIwae.. thank God and thanks cLaire... :)   actually i oredi noe wat i need to do... if i still failed to after 3 reminders... I think I will get the red card.. :)  like wat my dotter says.. to seek \"Daddy\" now..   '),(1635,3734069,'23,July,2004','2004-07-23 00:00:00','   å“­äº†ã€‚ã€‚ã€‚æ³ªç»ˆäºŽè½äº†ã€‚ã€‚ã€‚ ç´¯äº†ã€‚ã€‚ã€‚æœ€ç»ˆé¢†æ‚Ÿäº†ã€‚ã€‚ã€‚  ç”Ÿå‘½çš„æºå¤´å¹¶éžè‡ªå·±ã€‚ã€‚ã€‚ èŒ«ç„¶çš„é è‡ªå·±çš„åŠ›é‡ï¼Œ æ‰€æˆå°±çš„å¤šåŠä¸é•¿ä¹…ã€‚ã€‚ã€‚  ä¸€æ—¥é—´ã€‚ã€‚ã€‚èƒ½å‘ç”Ÿä¸‡äº‹ã€‚ã€‚ã€‚ ä¸€çž¬é—´ã€‚ã€‚ã€‚ä½ å¯ä»Žé›€è·ƒä¸‡åˆ†çš„å¿ƒæ€ è½¬å˜åˆ°æ²®ä¸§ç»æœ›ã€‚ã€‚ã€‚  å°±å½“ä»¥ä¸ºä»¥æ”€ä¸Šå±±é¡¶ã€‚ã€‚ã€‚ ä½†å› ä¸€æ—¶çš„è‡ªæ»¡ç‹‚å¦„è‡ªå¤§ã€‚ã€‚ã€‚ å¿˜äº†ç›®çš„ã€‚ã€‚ã€‚åˆè¢«ç»Šå€’äº†ã€‚ã€‚ã€‚  è‡ªå‚²ç»ˆè¢«ç ´ç¢Žã€‚ã€‚ã€‚ç»ˆäºŽç¢Žäº†ã€‚ã€‚ã€‚  é†’äº†ã€‚ã€‚çœŸçš„é†’äº†ã€‚ã€‚ã€‚ æŒ¯ä½œäº†ã€‚ã€‚ã€‚å› ä»–æ‰¶æŒäº†æˆ‘ã€‚ã€‚ã€‚  å› çˆ±ã€‚ã€‚ã€‚æ‰€ä»¥ç»Šå€’äº†æˆ‘ã€‚ã€‚ã€‚ ä¸ç„¶ä»æ²‰è¿·åœ¨è¿™è‡ªç¼–çš„æ¢¦å¢ƒã€‚ã€‚ã€‚  å—¯ã€‚ã€‚ã€‚åˆå›žåˆ°èµ·ç‚¹äº†ã€‚ã€‚ã€‚  å—¯ã€‚ã€‚ã€‚ä¸ï¼æ˜¯åŠå±±è…°ã€‚ã€‚ã€‚   '),(1636,3734069,'21,July,2004','2004-07-21 00:00:00','   hmmm today was more than a blessed day.. rejection again.. never mind I still have one more... straight in the face (ehh no la.. over the fone... hahaha but she has a crush on a Christian friend.. haha she asked me.. can a christian marry a non-christian.. \"unequally-yoked\" came to mind.. but i can\'t tell her that right (she\'s should be U4.. )told her that it\'s against his/her belief.. should not even began the relationship... haha she was disappointed..hearing that she wanna intro him to me.. hahaa. meeting her next week.. sianz.. was kanna scolded by mum for cooking porridge..   man.. it was only 10:30pm and i\'d only just talked to her for less than 5min..)hmmm actually started hounding only at 9:30.. ended 10:30... i think i talked most to lian cuz using hOusephOne..then called simin loh...always kanna persecuted for doing things... but then again.. i supposed I was made the chu qi tong ba.. cuz..dad met with accident this morning.. blocks of bricks fell on him @ the worksite this morning... and he was buried underneath them... Praise the Lord.. his head only sufferd scratches.. but his uppper body & back got lotsa scratches and his right arm & right leg was injured.. all bundled up.. the miraculous thing was that he actually managed to drive himself back from woodlands to home... Praise the Lord for His grace upon my family... I think he\'s really blessed cuz this is the 2nd time he got into accident.. 1st time, he got minor stroke after jumping down from his lorry..he recovered..not completely but 95%..then now this... but thank God.. I still have the chance to lead him to salvation before it\'s too late... man oh man...   Ahh.. on a positive note.. I have a good feeling abt nancy.. i almost gave up on her.. hmm she called me again to ask me abt tax matters.. let me try inviting her tmrw (cuz she just came home from hospital & was rather tired..) Yay!! I found the flow again... hahha simin was wondering why i so persistent in inviting her.. i wanna witness but was stopped cuz mom was mumbling at the background.. no choice...   ehhh.. tired.. yay! got chOir prac later... and cellgrp on thurs.. must ask sis siew choo if she needs any help... geee next week got many bday celebrations...  hmmm funny right.. people might be wondering why i still sound so cheery... the joy of the Lord is my strength!! It might seems like an endless chase but... I must press on...    \'Therefore do not cast away your confidence,which has great reward. For you have need of endurance, so that after you have done the will of God, you may receive the promise: \"For yet a little while,  And He who is coming will come and will not tarry.   Now the just shall live by faith;  But if anyone draws back, My soul has no pleasure in him.\" But we are not of those who draw back to perdition, but those who believes to the saving of the soul.\"  ~ Hebrews 10:35 to 39   '),(1637,3734069,'20,July,2004','2004-07-20 00:00:00','   ehhh just spent 1 hoUr revamping a bitsy... had a shock of my life when i went to my primary school webbie.. hahaha.. the school sOng.. i supposed sang by my lil juniOrs in chOir... heehee.. haha if u click on the links for moi schools.. U realise how cheena all my schools is..safe for ntu..  hmmm some greviances.. ahh.. went to tampines mall afterwOrk.. hmmm on the way..decided to sms nancy (a-U4).. not sure if her no still in use.. cuz vivian (friend in chc, used to work in iras.. now left.. think she\'s working with sis eve now..)said she had lost contact with her... so aniwae.. i sms her the other no.. haha she called me.. asking who am i.. hmm chatted with her.. sigh she\'s now full time loafer.. well she\'s said that she\'s now waiting for results.. her colon not doing that well.. well..she was rather harden.. she came before.. she was one of the first few person that I reached out to.. hmm she came to hollywood before.. still rather ok.. then invited her to a normal service after we shifted to jurong west.. well ..she said that she was very irritated by the sermon.. (hmmm i wonder why? this is the most extreme response..) then she went on to seek other help... the funny idols.. and sigh her life got bad to worse.. hmmmm.. I guessed Holy Spirit prompted me to call her cuz perhaPs the time is rite to call on her again... well.. will try to reach out to her.. perhaps with vivian? :) hmmm perhaps invite her to 15th anniversary if she reject me this week (haven\'t ask her cuz maomao ran called her out of the blue..& talked to her oni 10min.. contact her again perhaps on wed.. )meanwhile praying for her for healing..   actually wanted to buy an mp3 player.. hmmm out of budget..cuz hahha my bOnus is used up & I\'m still short of 1/3 of my mission fund.. God.. I\'m sure You will provide!! hmmm bought a tape recorder instead cuz md broke down, and i wanna record choir prac..aniwae.. tape recorder is cheaper.. bought the sony one @ $79..  my blusher had ran out.. so went to espirit to replenish.. hahha people who knows me know i detest make-up.. heehee I guess my cellmembers can see the vast difference liyi in cgm & liyi during svc.. hmm oso bought lipgloss & eyeshadow cuz my fave shade running out.. and i have to use it ah... heehee didn\'t know that espirit card could be used for red earth products as well.. haha 20% disc.. happy..  I was still carrying the red earth bag in my hand when i board the train bound for pasir ris.. this aunty/lady in her late 30 was looking at me and my red earth bag..~ staring to be exact! (btw I was dressed in tee shirt & jeans cuz today org dressed down cuz there\'s a sharing session)and i gave her a \"YES?? can i help u look?\"  and started to tell the guy next to her ( i supposed that should be her brother/friend/relative?) that \" next time ah, if you have any girlfriend, better don\'t find those that spend on make-up... \" I was like \"what the %#$#$$?\" hey.. what\'s the items that I\'d bought has anything to do with U.. moreover.. i did not have a huge bag of comestics and hello! I\'m already 20+ (hahha maybe she thot i was still in school.. from the way i dressed) and i bought the make-up cuz I need to make-up for choir.. (need not really make-up for work except big meetings to look professional) and aniwae, to look pro.. ladies do wear make-up.. my-oh-my she must have came from tim-bak-to!  Can\'t be bothered to heAr what she wanna say (I whipped out my headphones)but she was still glaring at me.. i gave her a dirty look (God i repent) and she was still talking to the male next to her and looking at me.. I took out my mobile.. started to msg \"hello.. there\'s this stupid woman commenting on how much I\'d spent on comestics..\" and while i was typing this.. she was standing next to me..and i sorta sense she was looking at my fone screen... hahha boy i\'m evil but i feel good... hey.. why she picked on me... and i resisted at barking at her.. she was talking so loudly in the mrt.. i chose to ignore by blasting my discman (not really blasting.. just at vol6)  hmmmm i know Holy Spirit wants me to develop the fruit of patience and long-suffering.. lotsa funny peOple came thrOugh my pathways for the past few weeks.. hmmm endure.. ! yay!!   '),(1638,3734069,'19,July,2004','2004-07-19 00:00:00','   ä¸»æˆ‘æ¥åˆ°ä½ é¢å‰ å°†ä¸€åˆ‡çŒ®ä¸Šç»™ä½  ä½ æ˜¯æˆ‘ç¥žæˆ‘æ•‘ä¸» å°†ä¸€åˆ‡æ›´æ–°  æˆ‘çŒ®ä¸Šç”Ÿå‘½ä¸ºç¥­ å…¨å¿ƒäº¤æ‰˜äºŽä½  å› ä½ æˆ‘æ‰¾åˆ°æ„ä¹‰ ä¸€ç”Ÿå¹´æ—¥å…¨éƒ½ä¸ºä½   æˆ‘çš„æ•‘ä¸»ç”Ÿå‘½çš„æºå¤´ æ— äººä¸Žä½ ç›¸æ¯” æˆ‘æ„¿ä¸€ç”Ÿèƒ½æ›´å¤šè®¤è¯†ä½  åœ£çµæ±‚ä½ æµ‡çŒæˆ‘ç”Ÿå‘½ æˆ‘æ¸´æ…•ä½ åŒåœ¨ ç”Ÿå‘½æ°”æ¯æˆ‘æ„¿å…¨å±žäºŽä½   å”¯æœ‰ä½ æ°¸ä¸æ”¹å˜ æ‹¥æœ‰ä½ æˆ‘å·²åˆ«æ— æ‰€æ±‚ æ„¿æ”¾ä¸‹æ‰€æœ‰æ¢¦æƒ³ åœ¨ä½ è£è€€åº§å‰  æˆ‘çŒ®ä¸Šç”Ÿå‘½ä¸ºç¥­ å…¨å¿ƒäº¤æ‰˜äºŽä½  å› ä½ æˆ‘æ‰¾åˆ°æ„ä¹‰ ä¸€ç”Ÿå¹´æ—¥å…¨éƒ½ä¸ºä½   æˆ‘çš„æ•‘ä¸»ç”Ÿå‘½çš„æºå¤´ æ— äººä¸Žä½ ç›¸æ¯” æˆ‘æ„¿ä¸€ç”Ÿèƒ½æ›´å¤šè®¤è¯†ä½  åœ£çµæ±‚ä½ æµ‡çŒæˆ‘ç”Ÿå‘½ æˆ‘æ¸´æ…•ä½ åŒåœ¨ ç”Ÿå‘½æ°”æ¯æˆ‘æ„¿å…¨å±žäºŽä½   è€¶ç¨£æ¥æŽŒç®¡æˆ‘ ä¸€ç”Ÿåœ¨ä½ æ‰‹ä¸­  hmmm. this is my source in chinese.... sang this on sunday.. realised that they did some minor refining to version 1... i love the chinese translation!  hmmm gotta do something more important.. my QT...will blog later.. btw.. I gOt rejected by my U2... the oni U2s.. left all the U3s.. yay.. mum almost confirmed on Fri.. yay.. wonder need to sing or not.. shall see.. at most ask one of my member down.. :)   '),(1639,3734069,'16,July,2004','2004-07-16 00:00:00','   å—¨ã€‚ã€‚ åˆæ˜¯æ˜ŸæœŸä¸€ã€‚ã€‚ åˆšè¢«æ‹’ç»äº†ï¼Œ å› å¥¹æ˜ŸæœŸäº”å’Œå…­å¾—å‚åŠ æœ‹å‹çš„ç”Ÿæ—¥celebrationã€‚ã€‚ å—¯ã€‚ã€‚ ä½†è¿˜æœ‰ä¸€çº¿å¸Œæœ›ã€‚ã€‚ å› ä¸ºè¿˜æ²¡é‚€è¯·å‡ ä½åŒäº‹ã€‚ã€‚ ä»Šå¤©ä¼¼ä¹Žå¥½æ¼«é•¿ã€‚ã€‚ã€‚ä¸½ä»ªå•Šã€‚ã€‚ã€‚ æ‹¿å‡ºæ–—å¿—å•Šã€‚ã€‚ã€‚ä¸èƒ½å†æ²‰ç¡äº†ã€‚ã€‚ã€‚å—¯ã€‚ã€‚æ˜¨å¤©è¢«mom æ”¾é¸½å­ã€‚ã€‚ã€‚å“Žã€‚ã€‚æ˜ŸæœŸå…­çš„service åŽï¼Œ è·Ÿå°ç»„supperã€‚ã€‚ã€‚ æ²¡æƒ³åˆ°ç«Ÿç„¶å‹¿äº†æœ€åŽä¸€ç­çš„åˆ—è½¦ã€‚ã€‚æ²¡åŠžæ³•ï¼Œ ä¸ºäº†çœé’±ï¼Œ å°±æ­å·´å£«åˆ°å‹¿æ´›ï¼Œ å†æ­å¾·å£«ã€‚ã€‚  &nbsp;  å—¯ã€‚ã€‚ åˆæœ‰ä¸€ä½åŒäº‹é€’ä¸Šè¾žèŒä¿¡ã€‚ã€‚ å“ˆå“ˆå¥½ç¾¡æ…•ä»–å“Ÿã€‚ã€‚ æˆ–è®¸åŠå¹´åŽå§ã€‚ã€‚å…¶å®žï¼Œ åœ¨å¥½å‡ ä¸ªæœˆå‰æˆ‘ä»¥å¤±åŽ»äº†å¹²åŠ²äº†ã€‚ã€‚ä¹Ÿæ²¡ä»»ä½•ç‰¹åˆ«ç†ç”±ã€‚ã€‚ä½†è¿™å§‹ç»ˆå¹¶éžæˆ‘çš„ç†æƒ³çš„å·¥ä½œã€‚ã€‚çœŸæƒ³é£žå“Ÿï¼ä½†æ—¢ç„¶æˆ‘è¿˜åœ¨è¿™å„¿ï¼Œ å°±å¾—å€¾å…¨åŠ›å§ã€‚ã€‚ã€‚   ä¸½ä»ªã€‚ã€‚åŠ æ²¹ï¼ï¼   '),(1640,3734069,'16,July,2004','2004-07-16 00:00:00','   in line with the retro-background.. let me change the background mUsic... you should be hearing radiohead\'s fake plastic tree in the backgrOund..(the music track is only the 1st 1min,, i like the last part...most...interesting arrangement... build up then soften... hhaha like praIsE Song...) :) heehee maybe sOme might not have heard this before.. this is one of my fave tracks on \"the bends\" the slower Ones.. but i like high & dry.. like the music arrangement. hahaha.. hmmm this is certainly my fave album in sec 3/4.. was very into.. loud & noisy music...and indie/alt music... and britpop... heehee... to name a few: blur.. oasis.. greenday.. radiohead..suede.. there this soNg by weezer called buddy holly & supergrass\'s We\'re young.. hahaha i like that song..cuz the lyrics & the song\'s funny... heehee.. and the muSic is happyhappy musIc..   sUddenly am bROught baCk to this mOde... I loVe muSic alot.. just that I dOn\'t the prvillege to pick up a musical instRument when yOung.. Hmmm I\'m not that well-Off when yOung.. bUt like mUsic... JiOned the chOir when in P2 @ TNS.. hahha... got into chOir by chance.. cuz my p2 teacher asked me & my friend whether we\'re interested to join the choir.. (hahah we became the youngest member in choir... hahha) Pri sch was fun... hmmm remember the SYF.. hahaha our choir uniform was like a sailor\'s outfit.. hahha u may think chOir lobo eca hor.. heehee agree la.. but i oso rAther active ya.. was roadcrossing mOniTor (hhaha i jaywalk... but i do look to the left, to the right, and to the left again,, just that i DOn\'t raise up my hand when crossing the road.. )hmm was oso in the prefectorial board.. just a nOrmal prefect..hahaha.. quite malu... was made to stand in class.. and kanna \"suan\" by my p6 form teacher for talking too much in class with my partner.. somemore still wearing my tie.. hahah... hmmm sec sch.. Once again... was in CHoir.. was lead to join choir by one of my senior.. heehee.. ironically i was in alto.. (haha cuz i could reach really low notes.. despite of my naturally sharp vOice).. haha felt really honoured to be one of the few sec 2s in the SYF choir when we entered the competition in my sec 2 year.. audition stress ah.. one from each part and to sing our own parts to the 3 choir teachers who were sitted across the room ~ quite a distance~ and both songs were acapella!! heehee we came in dinstinction.. (haha and i managed to siam one chinese test cuz of syf competition!! )  hmmm stOpped singing in JC.. hahha cuz thOt time to change la.. (hmmm ahahh..but still came back to singing...to use the voice that my Heavenly Father has given me..)was rather xianz le.. cuz doing the same for 8years(4 yrs @ tns, 4 @ dhs).. i guessed the sado nature of sweetie emerging... hahaa... jOined chinese & drama society cUz wanna tAKe part in the maJor drama prdtn..huacheng yeyun (was in the prOps team 4 2years... fun days.. stayed till like 10+ to do prOps.. and hahha gOt home 11+ to 12?).. but my major eca (haha now cca) was vnr.. hmmm it was real interesting.. haha suit my sado nature to be \"tortured\" haha.. always kanna \"knock it dOwn\".. heehee and the interesting traIning camp @ Ubin.. out of civilisation for 5 whole days.. hahha.. never got sooo dirty & smelly in my then 16 years of life... hmmm fun days... and went on a 4days3nites round island canoeing expedition.. hahha that\'s was a real acheivement for me still... (haha never realise sweetie so garang one rite??) hmmm heehee... then i quieten dOwn in ntu.. sorta waste my uni life away.. cuz i was reading a cOurse that i dOn;t really like (that\'s the resUlts of not chOsing wth the wisdOm of the Holy Spirit.. I should have chosedn archi.. hahha too late la.. ) then i oso regret not staying in hall... bUt i think i\'ll becOme mOre rebelliOus if i stay in hall... heehee cuz mei ren guan ma... and i can get very wild if i want to la.. (now i\'ve learnt the art of self-cOntrol..hmmm ocCasionally sOme might see the gIla side of me... but @ the right time & place..)  hmmm eugene sms that he just had whipping cream, flour, ribena, eggs on his hair.. my reply.. heehee how enviOus i\'m... hahaha.. btw it\'s his bday today.. so hence kanna sabo.. hahaha i think i\'m abnormal.. haha i\'m a good sport.. u wanna sabo me... can la... i wOnt mind.. hahha if u dOn;t mind my scrEams!! (the mOst malu sabo i ever had was the apple game!! @ the farewell party for seniors for vnr.. clue apple hanging by the stalk.. each pair is supposed to take a bite off the apple blindfolded... heee c\'est tout i\'ll say.. hhaha but fond memories...heehee)  hmmmm no cell tonite.. but i met huiming for dinner.. we went shOpshOp.. then makan.. heehee... thanks ming for taking me hOme and that U managed to find ur way back on PIE...  Oh... i just told my mom abt prayer meeting later.. and that I might stay o\'nite at ming\'s... she was like hmmm okies.. what u guys do? sing & pray... yup. aNother MAJOR BREAKTHROUGH!!!... and hahaha i drafted a simple/graphical manual for mom (in chinese characters.. yay i still remember how to write chinese)for her to use her 1st mobile phone... heehee (great patience man.. i nearly flared up at her for holding me back to work when she asked me how to operate.. i showed her.. then left but i din raise my voice.. cuz AV la.. so hahahha.. )  hmmmm it\'s 2:40 am.. i need to rest..Long day.. helping s1 cuz sis val sms for helpers... then i\'ll meet vick to get eugene & susan\'s present..  ahhhh need to reply to that \"interesting appeal letter\" that i left unfinished earlier this evening... hmmm some tax agents appeal for the sake of appealing..hhhhmm hope i dOn;t have to do this 4 long... 3 years @ RH le.. time for a change?? back to books?   '),(1641,3734069,'15,July,2004','2004-07-15 00:00:00','   am now in my orgn resource centre to find lyrics for songs later.. yili oni sms @ 10am.. not his fault la... think he should also just receive the info ba.. :)   sad case.. my reliable URS played out on me... the left strap came out.. phew i just finished a meeting..(was informed at 955 that i should sit in cuz one of my trader is coming.. ah..phew it was okie...) later gotta buy new pair of heels..:(  had a great time yesterday with my mom..treated her to lunch.. then went shopping with her... and it\'s amzing that a few hours of quality time works wonder!!   went to part b of church w/o walls.. poor victor was \"invited\" by Pst Kong a numerous times... ehhh caught myself in the tv screen a couple of times.. man.. i looked pale... gotta change the color of my powder..   Holy Spirit gave me a vision during end part of seminar:my mOm responding to altar call ! PTL!! am sure it will come to pass... fresh look on the U4, U3.. etc i\'m reaching out... perserve!! if i can persist in solving thoSe rOutine mathematics problems despite many failed attempts.. mission accomplished after a no. of tries.. (haha i think i must have finished solving the maths C & fm 10 yrs series back to back at least once..)i sure can soften the hardened and move them up the rainer scale... yay! exciting time ahead!  reached home ard 12 midnite..caught buffy! gee spike looked awful with floppy hair (though his usual hairstyle resembles mattel\'s ken\'s plastic hair!) but i still like spike... heeheehee.. dOn\'t know why...  okie.. got my songs now... gotta go back to work... and to do up the songsheets..   (woke up rather crappy today cuz it\'s AV season.. hahaha..this is up for me to know.. if u guess wat it meant.. shhhhhhhh!)   '),(1642,3734069,'14,July,2004','2004-07-14 00:00:00','   hmmm actually intended to log off when i saw r8ch & yili online.. hmm r8ch no stamina... she didn\'t talk muchie to me..then zzzzzzzz.. as for yili ah... i\'m amazed that he\'s still awake after running ard ntu...   hmmm o yes.. this just came to my mind.. i went to a talk by one of the big shots on investigation work... i was drifting to zzzzzzzzz.. i was awakened by \"the harvest is plentiful, but the labourers are few\" ~~ I was shocked!  for a moment i thought i was in serivce... ehhh he was using this in the context of potential of tax revenue leakage (aka harvest) and the no. of staff in his division (labourers)... hmmm taking the biblical verse out of context?hmmm i dOn\'t know... but i was reminded of the need to reach out.. (i wonder if mr Ong is also a brother in Christ..)  okie... i need to zzzzzzzzzzzz.. but gotta seek souRce 1st.. Ops.. i haven\'t play with brownie...   '),(1643,3734069,'13,July,2004','2004-07-13 00:00:00','   hmmmm once again.. hermiOne left the office late.. haven\'t been doing soo for a long time... hate the haze..irriated my left eye and throAt... my left inner lid kinda swollen.. since sunday.. at fiirst I thought was lack of sleep... then i realised that the air ain\'t very fresh these cOuple of days.. i actually smell bbq smell last nite when i walked hOme..   caught the webcast the last part from how to reach the U2.. yups.. hermione is silent no more... some of my friends are U5 (simin), U2 (ting), U3 (jenny), U3 (aifen), violet (U4? U3?) why no U1??? hahha  hmmm me ah...i think i was a U2 or U1.. more U1 la... just waiting sOmeone to pop the questiOn... and to ask... (heehee the person who invited me didn\'t personally escort me from the mrt station to SIS.. i took the chartered bus from kallang mrt alone.. and called her after i reached.~ aiya how i know she\'s usher ic then.. i\'m not complaining abt ming la... in fact.. love her lotsa.. actually i was kinda complaining on the inside that aiyah.. why i come.. so pai se... dOn\'t know aniOne and eveRyone ard me seEms to know everyOne.. yUps..felt so lost.. bUt impOrtant that (a) monique & meiWen really made me feel @ hOme.. :).. & (b) i do really want to be churched.. ( ehhh one of the rare few... one of the 11%)...  just waiting for people to ask... ;) but no one did... hahaha until ming... :) cuz City Harvest is where GOd wants me to be..   hmmm jUS spOke to lian.. PriAse the Lord.. her stUdy loAn is Sercured... am really very :) that bOth lian and vicks have grOwn.. time fOr them to be on their oWn? i just have the feeling that I mihgt not be in the same cell as the ntu pple.. but nevertheless if so be it.. hmmm I know that I\'d help them to gRow.. & especially Vicky... she is sooooo  different now... am so happy for her!! :) now i guess i should spend time with Joyce le... heeheeheee...   what Pst said during the conference is so trUe... one of reason i sorta stop reaching out is.. tooo busy... i need to strike a balance... Qn: Am i really over-serving? hmmm i really love to help choIr as and when i can.. (i think i fill like one of the perm fixture on chinese church chOir) and s 1& s4 if i can.. i think i will need to priortise..cuz what\'s good is the enemy of the best...  hmm overnite praYer meeting this fri.. hmmm wOnder yili shifu would be teaching us on sat or will he be toooo tired?   yay.. tmrw spend time with mum in noon (it;s her bday!) then go conference!! rite now need healing.. on my eye... & thrOat.. and now.. gOtta sEEk my sOurce..   '),(1644,3734069,'12,July,2004','2004-07-12 00:00:00','   heeeheee...never had i had the opportunity to sleep till 11:30 for the past few months... luxury man.. but then again.. i slept @ 5am.. hmmm not that many houRs of sleep...   hmmm after i blogged last nite.. I played with my darling.. haha maybe i should give \"him\" a name.. haha.. it\'s brown in color.. hahaha.. brownie.. duh,, hmmm sounds like the eca in primary school.. heehee.. haha.. can\'t call him scout rite? hahha.. getting colder if i continue in this line of thought... hmmm miss the days of JC.. hahaha never imagine myself to jOin Ventures & Rovers.. (for the uninitiated..it\'s an scouts group.. hahha I still can remember the first line of the scout motto.. Be prepared!!)but if i didn\'t.. wOn\'t befriend ming and heeheee.. won\'t be in chc? hmmm hard to say.. cUz i saw sOme of my pri sch (Jean.. think she\'s in S1 but wonder if she remeber me..) and a couple of DHS pple.. think one of them is kelvin..he should be one of the usher IC.. saw a cOuple of guides.. hmmm think one of them is a youth cgl.. (she was in the same SOT batch as jolene..)hmmm think i saw my sec sch classmate zicao on board on chartered bus the other time..? not sure if it was him or not..but it was quite some time back,, and i seems to have lost contact with bulk of the 4C pple... hmmm except for Sandi & Shu\'en.. hahha heehee liyun, who was?my classmate for upper sec & same eca in JC)?was the one who invited huiming to chc.. but hmmm she backslided oredi when ming invited me 3 years ago... haha wonder why she never invite me that time leh.. hahha maybe i looked like the not so likely to respond pple.. hahha hmmm me U1 leh... just need pple to invite me hahha....   Opps digress... was kinda late for chinese prac.. reached ard 245..Ops.. phew.. they were just practising open up the gates.. hmmm sang a old worship song.. ehh i hardly recall ever singing it.. but guess that it\'s an inborne ability to catch the tune or foreguess the tune.. so..got the tune easily and the words ah... ehh.. chorus okie lohz.. hmm give me a couple of weeks can liaos.. went to g4c just now.. hmmmm chinese lyrics sorta direct translation... hmmm really meant a lot when i sang...cuz was practically left with not much when i pledged (hmmm the bonus is now deposited in my acc.. yay!! I can fulfill a huge portion?my pledge!!).. it will be memorial offering to God!..   hmmm last nite msg Sis val that I should be singing for Tues BS (but knowing how much Wed bs need pple, i told sis val that in case tues too many pple.. can put me on wed..) so sis val sms me ard 16:30 to ask me to go wed instead.. so told her okie.. wellwell.. initially thought of going tues cuz wed\'s my mom\'s bday..so was praying that mom will be okie if i told her i\'d spend the afternoon with her then go church in the evening (me to take half a day on wed lohz...) well amazingly.. when i met her @ whitesands while waiting for dad to fetch us for dinner.. i asked her abt wed..whether she was working.. she told me she would be working half a day and so me told her abt my plan la.. she was cool with it.. Amen!! and the more joyous thing was that... after dinner.. she told me that she will be taking leave this coming sat, sun & mon and she told me that this sunday i go chinese church with you!! Praise the Lord!! I was like.. i don\'t even need to ask and she wanted to attend the service.... Thank You Holy Spirit for softening her heart and making me a light in the darkness.. cuz sweetie hermione was always the angel shining in the situation.. ( actually dad & mom sorta get into a shouting match again the car?cuz of miscommunication..and dad wanting to catch the 7pm drama on channel U..? but thank God that i said the right things at the right time...)  was watching the NKF fund raising show for the children fund.. hmmmm heart really goes out to these children and their parents.. the question kept repeating in my head... \"have they heard about Jesus..have they experience the goodness of God\".. I\'m not trying to act holy.. but the words they said just expressed their helplessness and hopelessness in their situations.. \"What would their life be if they had Christ in their life...would they be feeling the way they do\" ~ nope.. \"so what are you doing abt it.. \" ~ ehhh yah.. sin of silence...the wealth of knowledge is in me.. but i\'m just like a sponge absorbing.. time to squeeze the sponge..   need to OT tmrw (i.e mon) cuz gotta finish lotsa stuff...  hmmm gotta go now... gotta play with brownie &?take my spiritual bath.. will see if i can sing and play.. tried yesterday but.. hmmmm ehhh not very smooth.. ended up accapella... :) will keep pract.. \"the more hours you put into it, the better you are at it..\" ~ the bookmark in my birthday present from cell.. (ehh i haven\'t read it yet.. hmmm cuz more suitable for a matured woman..ie married.. cuz i flipped to chap 2.. ehh husband?? hmmm my cgl very interesting.. but am sure it\'s a good book but hmmm i would start reading after i finish my current book.. ) my left fingers don\'t really like me now..cuz i put them thru chambers of torture.. ai yoh.. last nite when?joshua hear lili that i was learning the guitar.. he told me that I need to use waterproof sandpaper to \"rough it out\"..\" yao tuo pi le..?cai bu hui tong\"?ehhh?? hahahaha..phew he\'s not my shifu... hmmmm must ask yili if that\'s true.. though i don\'t see the logic behind...   geee wonder how to celebrate lil bro\'s bday... vick & yili...any ideas? perhaps sUrprise him with a cAke dUring break of o\'nite prayer meeting? he should be going la...   '),(1645,3734069,'11,July,2004','2004-07-11 00:00:00','   hmmmm just a brief note before i log off.. currently lacking the insipiration to blog... just polish off a bowl of lychees...sweetlike me!!! aM glad that I\'d gave all that I could... YAy!! all of my 1/2 month bonus and above... am looking forward to September where i complete the pledge.. truly my dear bank acCOunt @ current state is bleeding really bad...the lowEst ever!!!(it\'s really veryvery unpleasant to sight...)but despite that am believing that I can fulfill it... hmmm am now resorting to keep all the plastic card at home (except my identity card, my warrant pass, my staff pass & ezlink... yups.. leave all cards behind so i can stick to my budget..) haha not living in a poVerty-mindset.. definitely gave wHAt Holy Spirit prompted me... thoUgh i very muchie wanted to save this $$ for either SOT or mission trips.. bUt.. hmmm am giving it to the fund 1st.. (btw my bonus only accounts for 2/3 of my pledge and I\'m still paying my dear ntu loan.. bUt God is my Jehova Jireh.. so I believed that THis can be done... )  God is a God who hears and answers prayers.. two years ago.. I used to even have difficulty finding a place to do my QT...so skipped at times... bad example...(was sharing room with my mum.. was staying at a 3room flat.. brother got his own room.. dad took the living room at nite.. why ah? long story lah...but now not the case oredi la...) hmmm i was desperate to the point that i prayed in the tiolet.. needless to say.. my best QT was when the whole household is out and i can sing all i can.. hence i always love walking from bedok interchange to home... (abt 10 - 15 min) cuz.. hahaha i will be singing to Him.. conversing to the Holy Spirit aloud.. (haha i still do now.. so i love walking home.. ) and always almost kanna questioned by Brother Andrew Wai abt my spiritual life... particularly QT... (hated it when he aasked me that when we shared cab home after cg @ estella\'s.. cuz didn\'t do mah.. hahaha)  so asked God for breakthrough in my spiritual dryness... and a room to myself so that I can seek Him... :)heehee.. and Praise the Lord! Family decided to shift.. to my cuRrent place adjacent to Downtown east.. and me need to pay a cent.. (my brother co-owner with dad) (side track.. ahhaah 2nd stupid msg from guys wanting to chat.. ) so my blessing not monetary in return la.. hahaha but i dun mind... cuz 2003 was the year of growth from the gal who eats from the spoon and needs a shoving in the back from dear ivory... & brother Andrew..  hmmm not that I\'m complaining abt not being able to spend my bonus... in fact.. i\'d been giving my bonuses to building fund and now the mission fund.. (haha when my collegues ask me when i was going to hols.. and i definitely had lotsa $$ hahha.. yups in God\'s storehouse..)and yups... but God is good... I guessed i was pretty blessed.. despite i din get $$.. my desktop, my MD player (sad.. retired now..), my hi-fi set.. my TV.. and my T610 was sponsored by mum .. (wat i gave for my 2 building funds.. i sure had reaped more.. )  had always been a spendthrift.. since sec... haha used to spend $40+ monthly on CDs & singles. (hahaha i think i should have spend at least a thousand on CDs & mags like tOPS.. smashhits..big.. seventeen.. her world..cleo.. elle (not the local edition) (haha i was a groUpie...I simply adore bsb.. and i have 4 versions of the CD single of I\'ll never break your heart..)Ops..hahah mei xing xiang le..   started work.. splurge on bags (think that I got one of my side cabinet filled).. clothes (hmmm not veryvery x la.. but think most X was this british india skirt that cost $108.. )...shoes.. now i shou lian le.. boh bian la.. cannot lead this kind of hahha waste $$ lifestlye.. but then i never fail to tithe la.. am :) that I tithed the very month i was saved... heehee and to my 1st building fund two months later... hmmm cuz i believe that whatever we have is not ours... hmmm if we can trust Jesus for our salvation.. why can\'t we trust Him for the 10% that was in fact the full 100% from God...   hmmm think i need to take up tuition... :) cuz i wanna go Israel next year...hmmmm and guangxi too.. :)  hahaha i\'m a natural long-winded gal... This is not a short-update.. i guessed u just gimme any topic.. i can just blabber & blabber...  but truly.. am really :) that I gave all that I could to the point I can sing mariah\'s my all (hmmm perhaps parts of the lyrics not suitable?/ ha change the sub will do the job.. ;p) and cuz I know He will make my desires and dream come thru\'!!   Ohh btw the song in background.. i love this song.. It\'s Sammi\'s æˆ‘å€‘çš„ä¸»é¡Œæ›².. i like sammi.. hmmm eversince Feel 100%.. hmmm i\'m not cantonese but haha that;s abt the dialect I\'m semi-fluent...(not bad la.. the time i went HK.. natives shocked that Singaporean can speak cantonese..they thought only M\'sian can...) hahaha i\'m very aunty in sense that I love watching TVB\'s drama... still adore å¦™ æ‰‹ ä» å¿ƒ I & II (canto version)... and è«‡ åˆ¤ å°ˆ å®¶ (hmmm especially like this series cuz chilam is soooo cool... ) remember that I watched the whole series in record time.. the moment that my bro took home the VCD.. i finished it that nite.. (the discs came in portion..cuz the show just arrived in S\'pore.. think not for rental yet..)okie...i  better stop.. i can go on yakking.. ooohhh btw i should be working today (at least half a day) but was talking too late to yili.. hmm not yili\'s fault.. i then read read my bible.. then hahha fell zzzzzzzzzz... and woke ard 6 to off my nite light..then at 10... geee to late to work..so i take urgent leave ba.. and prepare my bs and for my guitar lesson... hmmm speaking of which let me go play with him... let me try to add music to my worship.. had been singing accapella for past 2 years.. :) time flies..  gonna be 3 years @ CHC... came @ 12th ani. sat beside monique & meiwen.. hahah now in same cell as monique again.. ;p  the long-winded gal..going to play with her hubby...   '),(1646,3734069,'10,July,2004','2004-07-10 00:00:00','   Heellos...  haven\'t been bloggin in this blog.. hahaha..cuz was playing with my xanga...both has their pros and cons.. I like blogspot cuz can play with the look without having to pay..  tired.. went gym after work.. hahaha stress-reliever.. eric was playing with my sonyericsson to terrorise me.. wanting to take snapshots of me.... but let him be la.. cuz i could sense that he\'s got things on his mind but hmmm me not that close to him la... well at least i bring joy to him...   still gotta work later... but still chatting with long.. ah... gotta tell me me gotta go soon...   yay..taking haf a day later.. cuz got guitar lessons & bs to vicks.. nOpe.. i\'m beri eng.. infact.. i\'m up to neck with Work & corporate event... am in the publicity com for D&D.. and i\'ve to print 25K coupons for orgn national day carnival.. (cut cost.. so self-print.. ) :P   '),(1647,3734069,'06,July,2004','2004-07-06 00:00:00','   let me see.. time is now 3 a.m..  did a lot of things today.. Once again.. i slept with the TV on again... fell zzzzzz after watching the euro finals half way when the score is still o-o.. bOrin...  didn\'t blog on sundae cuz got hOme rather late..then stArted chatting with vicks...let me see what happened on sunday...  went gym in the am.. hahha was sUrprised that my pt (i won\'t have the luxury of this soon...gotta skim my spending habits) actually confided in me.. ehh he was stUck in a situation.. wanna seek my opinion (btw he\'s a free thinker.. )(actually heard the Inner Voice.. saying.. reach out to him, i was like.. huh... okies.. ) then he popped the question.. what would you do if you were me.. somehow.. i just blurt out.. I would pray.. if i\'d the peace.. then i\'ll proceed with the decision.. (cuz he was actually @ the crossroad of his career..)was kinda sUrprise that I actually managed to sow sOme sense into him.. seems that he was escaping.. so hahha.. told him that if we don\'t face up to reality &  solve the prob.. , you will always end up facing the same problem again.. like ta-baoing tests..  In a way, i know that I myself gotta face whatever I\'m facing...running away will just loop me back to the starting point...   sang for chinsese afterward.. learn 2 new sources.. my source & opne up the gates...  hmmm.. went dinner with my dad & mom.. dunno who mentioned Blk 88 bak chok mee earlier that week.. suddenly thot of eating that.. :) so we went.. :) then went grocery shopping with mom & dad at the cheapo stall near my previoUs area.. bedok nth.. ehh.. haven\'t spend much quality time with parents for long.. Peace.. never take this for granted... came a looong way and hard way to appreciate that.. am ever so thankful that He always come thru\' for me at that moment..am believeing that at least 50% of my household will be on fire for GOd by year end... :)   ehhh song:\"that\'s what friends are for\" is on air (am listening to class)am really blessed.. :)with friends at wOrk (unsaved yet.. but will soon be.. )cOme to think of it.. friends that know me lOtsa aRE thOse who I\'d Some sort of MisUndersTanding with.. my longest-friend sandi.. I known her since P6..used to sleepover at her place.. go to her place after sch.. and the paperback that was wrapped up in brown paper (i think was laura\'s) that we hahaha read... Ops!! was in the same class from sec 1 to 4.. and we had a COLD war from Aug 94 till Dec 94.. And it was really freezing cold war.. sub-zero temp.. but God is good... She invited me to X\'mas mass @ a church in kembangan (so adpt.. a song that reminds me of my sec sch years just came on air.. sometimes love just ain\'t enough)and yups.. it really awkward (btw we\'re in the same eca too.. yeah.. Choir)cuz we had not been speaking to each other for half a year.. well...somehow.. we just hugged and reconciled... ya.. she tried to jio me to church but i was just plain lazy to go sunday svc consistently cuz i think blossom was showing 130@ channel 5.. (haha.. i like the show cuz of joey lawrence..)so me was a bigday kinda person...  Then my JC friend..shumin.. haha..during JC we sorta distant ourselves suddenly.. haha but we patched up via.. haha apology letter.. and violet.. think during FYP.. (o, we studied same course @ ntu as well).. things got kinda tense at times.. hmmm but thru\' misunderstanding, r/s got closer...  most recently... the WOrse ever..cuz she\'s really someOne that I trust... she knows more abt me (hahaha a whole lot more than most/all?.. else except God)than hahah sandi.. despite knowing her only 3 years.. i think less than that.. and man.. it\'s so haRd to act nOrmal when i still feel the hURt.. despite that she sOrta apologise.. then...  I guessed cuz I dUnno what was going On.. and she\'d proMised me not to divulge..but.. (but i know she meant well cuz ...)am soooo liberated after that afternoon after the \"trashing-out\".... everything was so divine.. \"sorta-suan\" u via blog then msn prv nite.. then woke up late next mornin.. went doc.. given mc.. and went to ur place.. and hahaha..  ger.. if u reading this.. It\'s not that I\'ve not forgive u.. or i want to bring this up as a reminder.. cuz.. let me tell u.. I really treasuRe u as my friend.. One of my best friend ah.. and the other party who is in this dRama (the unknowing party)... U\'re one of my best kakis as well.. :) proTEct ur identity dun want to name u two la.. u noe who u are... :)  okie.. thanksgiving aside.. monday was ahhhh  meeting from 10 to 11:30 for the national day celebration committee (me in social sub-com this year..)gee gotta come up with ticketing design then print 25K coupons.. by end of the week.. another meeting from 14:30 to 16:00 with my AM & D.. (nearly zzzz in the meeting.. was actually in a switch off mode when D was talkin..and came to mind.. why are we so harp up on the treatment.. what\'s the puRpOse...)geee...remember wat i read abt the Fish! philosophy in Streats today.. if you don\'t like what you\'re doing.. quit.. after the meeting..saw a horrendous reply from pwc.. ahhah now gottta seek legal opinion.. ah... gotta write submission.. and also the case discussed during the meeting and my backlogs (60++) to cleared by end of August... AM knows that I\'m very tied up... ask me to prioritise.. and the national day thingy...   but Praise the Lord.. managed to come up with the design with one hour.. and found a way to create the serially no. coupons of the desired dimension.. :)  another of my fave song.. (SClub7\'s ehh what\'s the title.. is it the day i let you go?)hahha.. can tell how long i\'d beem typing.. already.. franky.. i don\'t know why but i\'m especially awake now.. the nice songs help..   yay.. watching spidey 2 later with vicks, her friend and long.. (bought tix oredi!!)  hmmmmhmmm must zhu gen jiu di later when i see her... heeehee..  then just now was chatting to pple..hahaha i think i can go play darts.. somehow two of my friends just went offline when i sorta \"grill them\"... eh.. make it three.. haha.. was rather bored just now... so i did a loong winded questionaire on fri3endster.. duh..   hmmm can\'t sing later cuz cuiqi appealed for pple to sing wed.. so me go tues seminar than sing for wed.. see can leave haf way or not...   ah... 430 leh... gotta zzz.. at lst wanna go chen yun but ehh think i\'ll end up sleeping.. :) later spend quality time with mom.. then meet vick & long.. yay.. my brother coming back from jp tonite.. wonder wat did he bOught for me... the other time he bOught me this cute miffy when he went to aussie...   okie... c\'est tOut!   '),(1648,3734069,'04,July,2004','2004-07-04 00:00:00','   am still overwhElmed by my enCOunter with God..   was reading my Bible on bed @ 3pm+.. wAs rather tiRed.. set my alarm @16:30.. befoRe I knew it... I Opened my eys and it was 5 minuTEs to 6!! AAAHHHH I\'m late big time!!!  jUmped out of bed, grab my clOthes and slappped some make-up on my face and geeee realised that I had only $16 left.. ahhh it will cost at least $20+ to get to chURch.. ah,,, just did half of my face and flew out of the hOuse.. time is 18:05... reached downtown east at 18:10 ah.. the atm machIne was Down.. ahh..  no Choice.. flew to the busstop & flagged a cab.. God is good.. got a very nice cabbie unCle.. asked him to take me to the atm @ Blk 445.. Praise God.. No Q.. got in cab again @ 18:20.. did my cheeks & lips.. and prayed that I\'ll be in time to catch the tags... Phew!! reached chUrch @18:50... flew across the Road.. and Praise the Lord.. Frankie was still giving Out tags.. flew dOwn.. Joshua was sweet.. saw that i was rather hot.. fanned me with his songsheet..  never felt so joyful dUring praise & wOrhsip..fel really close to Him... ah.. during the item song was sandwiched between Adrian & Frankie.. (i really mean sandwiched.. )but the atmosphere was thick.. different from service One.. despite hearing the message on Fridae.. The words of Holy sPirit was mUch clearer.. and half way when Pst Kong was preaching.. man.. I could really sense His presence.. . bUrden for the lost really weighs on me.. Ultimate pUrpose of a Christian.. find mOre empty vessel to fill... and in tUrn fill more...   Holy Spirit was telling me that.. ger.. where\'s your empty vessels?? tRue.. I\'ve been discipling vick & lian... bUt they are not my seeds.. I have yet bRought any of my friends to Christ again.. except for Tien\'s re-dedication in 2003 but she\'s someone\'s elses seed.... (geee one year le!!) hERmione.. U need to stARt wORking again....  Demonstration of our love for God via loving His love.. God love all nations.. hence, i love God, hence I will love people..   Am really toUched by Brother Calvin\'s testimony of how he brought his mum to Christ.. via email... his & his wife..   Truly, I will not be an ubin residence ever again..  okie.. i need to sleep le.. bte was rathet bewildered why apple see me before.. she was sooo surpRised that lOng & i in same cg... apple.. i was all along in A320.. :) okie.. sleepy now... good nitex!!   '),(1649,3734069,'03,July,2004','2004-07-03 00:00:00','   am having suCh jOy... remnant of what I\'d went thRu\' dUring prayer meeting few hOurs ago...   thank God that I was able to make it...oNi told my mUm @1745...   saNg for sErvice One.. PrAise the Lord for divine hEaling.. my calf muscles weRe kindA over-stretched (I gUessed it was dUe to me ovEr-hYper on tueS.. (was actually haVing tRouBle climbing dOwn staiRs..for a cOuple of days.. ) if not... can\'t be a \"jUmping jack\" last nite...   OvernIte prAyer meeting was aweSome... nevEr sTayed thRoughOut the fUll session... until last nite.. and bReakThru was that theRe are more than the usUal faCes i see... Six HouRs..  :) fiNally rEceive cOnfiRmaTion to over ceRtain arEas of my life thAt\'s like so mind blOggering and rather difficult to tRust in Him... And the other vOice in my head trying to sOw thOughts of dOubt & deFeat in me... sAw thaT i was oNe of the sand pArticle that He picked up.. (ya.. that\'s hOw small i thouGht i was... small and insignificant.. inSecUre...) despite of how i depict myself to be... I was who God has creAted... maNy a times.. I dOubted whether cerTain things are frOm Him...I was faithless BUt He saw what was in me...   saw that I was in this hOle of daRkness where I was leading mysELf into... A hand that strEtches in and pull me Out...   acTually was On the vErge losing my vOice whiLe praying in pairs.. sOrry long... I know I sound like a frOg.. my thRoat was simply toooo dry... (from the kopis that I\'d) and thROat was constricting... and Jaws wEre aching... yah.. i was shOuting.. badbad for my vocal cords.. bUt i regained my vOice back after the breaK..  LASt paRt of the Prayer meeting was ... wOrds can\'t expREss whAt i felt... we\'re just taking time to meet Father God and to wOrship Him...saw myself singing to Father God up in the heavenlies....Just to Him alone...He never left me.. I was drying Up cUz I was shUtting Him Out...His presence in the Auditorium was soooo thick... we cOuld just continue to wOrship on & on...  Members who were in church before 2000 was laid hands on... Many received the Holy Laughter... Pastor Ming came over to lay hands on us.. DEspite that pastor was at the other end, the moment that Pastor lay hands on Max... I saw light.. then.. i was DOwn... (later I realised i caused a domino effect.. cuz i pulled emmy down... poor Chris was trying hard to sUpport us...) i just feel that the oppression was lifted... Liberated..   was still lingering in His presence when i was off the mrt.. so i took a detOur in the Pasir Ris Park..  actually dOzed off when i got hOme and stArted to finish mo niu de taio jian.. (finally i did).. Gwen called me @11 cuz she wanna loan formal dress from me.. (but i do not have la except my prom dress hahaha.. i think tooo long for her oso...) Suddenly just knew that I need to seek Him.. started singing and just went on & on... and saw this: the back view of a man & a little girl... Yup.. that\'s me & my Father...He was holding my hand.. walking with me... I was this little girl with two pony tails..  He is always there in all my situations..He never let me go of His sight.. yes.. i was already His before i heArd His name.. (realised that i was actually rather envious of my friends duRing my primary School days that they attend sUnday school... now I know why... I was not a mistake... i\'d was handpicked...)  never will i wanna lose touch again... Lord I TruSt in You...Visions you given me... yes..they have not come to pass yet..but it will.. it\'s not my timing.. it Yours! And yes.. I know what He wants me to do next year... it\'s getting clearer & clearer.. Obedience is better than sacrifice!  verse of the day: \"Now may He who supplies seed to the sower, and bread for food, supply and mulitply the seed you have sown and increase the fruits of your righteousness, while you are enriched in everyting for all liberality which causes thanksgiving through us to God\" ~~2 Cor 9:10-11.  P.s. another friend decided to blog.. :)   '),(1650,3734069,'18,August,2004','2004-08-18 00:00:00','   Hmmm I haven\'t been online for days..was rather tired and realized that I was easily irritable cuz.. hmm cuz the cycle goes again...(din have any major outburst...but was tested by two incredible sisters..?one from POM/ one from my cell...     hmmm many thoughts flowed thrU my mind for the past few days...but more or less.. they revolved around the same topic...well.. I shall attempt to summarize...      for the past few months...I\'ve been thru...some of the most turbulent times of my spiritual walk.. though the deep valley age might not been long...it did made deep impact...ya.. the plunge was rather deep..every time I edged towards some sort of victory, my complacency would result the slip to be even deeper.. well.. to the extent that I would have lost 2 of my very good friends cuz of my \'not-very-discerning-antenna\'... and as a result given some mixed signals but glad that al\'s histOry now and that friendship with them have gone stronger...        During these 3 weeks break from choir...I realized that :  (a) I really missed serving in choir... (b) I didn\'t get into this ministry by mistake.. I\'m not void of talents to serve in this ministry... singing in parts and harmony not a difficult feat to me.. (though i\'ll prefer if beansprouts are given...) (c) I can bring friends even if I\'m involved in the production...(used to use serving as an excuse to not bringing friends...bad ah...certainly!! it\'s all in the mind... ) but glad that I\'d a breakthrough of the drought since eh...X\'mas 03...oh...nope...did brought mOm to service sometime this year...but ya.. c\'est tout! hmmm ya...violet came for service 7 last week! yay!! hmmmm she didn\'t respond to the altar call...at least I can upgrade her from U4 to U3... she just find Pst Kong\'s sermon very inspiring ...but she see that religion ain\'t a priority for her yet...hmmmm I\'ll continue to work on her... preserve! (incidentally, that\'s part of the school motto for my primary school.. sincerity and perseverance)  was burnt out by the POM.. yah, I know that all i did was the cheerleading...but as a result of the \'heartbeat\' move... think I stomped my right feet too hard...(hmmmm my right ankle kinda hurting...my old injury...) can\'t wear heels now...  after service 7\'s Pom.. went cafe.... heehee saw mei.. my youngest daughter (haha.. sijia was shocked to learn that I\'m her mum!! heehee.. ) took a neo-print with her...J then ate with alvina...hmmm she said that he\'s è›®å¸…çš„.. (haha... guess who is he??)  Watched fireworks at marina with mOm.. glad I did that.. (though the fireworks were not bad...but glad that I managed to pacify her bitsy by spending time with her..)    I,Robot     Took the day off on Monday...( partly to recuperate & partly of my cramps.. ) Went PS to catch I,Robot... before that I went to buy my medicine at bedok nth...then went Ps...I was like...so glad to be free..from work...I-Robot was great.. (notwithstanding that the male lead was Will Smith...yups...he\'s one of the actors that I\'d like since young...heehee.. since fresh prince of bel-air..) well.. as the movie progressed...the question that the robot (which I recalled was named sonny by the creator, which he fondly named as my father) \'What am I created for?\' plagued my mind... hmmmm prior before to hermione\'s dark ages, I\'d been plagued by this question...which was what led me to buying \"A purpose driven life\"... but still due to the dark ages...I\'d lost focus...lost direction.. and most importantly.. forgotten the ultimate fact.. I was created uniquely..by my Heavenly Father even before my birth...Was brought to mind of the bare fact that everyone was created by God for a reason at the scene where by Sonny was being examined by susan- a post mortem... and that he was told that he was different from other NS-5, eg.. his composite was stronger...than others...(sonny was built for a purpose..but he was unaware of his purpose.. He has many secrets hidden in him.. like his composite...he only realised his ultimate purpose towards the end of the movie...) and once again... whereby he related his dream to susan and spooner... that \"I believe my father has made me for a purpose...We all have a purpose\"... Everyone will be used by God.. to serve God in the way in his unique manner... each of us has a purpose in life! and ya.. I was immensely touched by Holy Spirit especially towards the end of the show...heehee...haven\'t teared in movie for a long time... Cuz...I felt the Holy Spirit reassuring me.. that I will reach that dream that He has placed in me...I may not see it happening...but it will.. in due time... The last scene of the moive.. I depicted myself to be sonny...standing on top of the hill... and... :P      \"For the gifts and the calling of God are irrevocable.\" Rom11:29       lian called me (for once she called me.. heehee.. ) she sounded rather excited abt her soccer.. yeah.. her 1st session in female soccer... hmmm spent 40+ minutes loitering at my voiddeck.. ya! cuz not convenient for me to call at home... ya.. PO... even though it\'s my own hp... hmmm glad to see her breaking thru\' in some areas... :) Jia you! sister!!! told her that if she has probe with her maths.. can come find me... (haha... that\'s used to be my stronghold... ya. my strength is mathematics... if only I\'d been more diligent in my F.maths... hahaha.. then I\'ll have 3 As... but I\'m contented with my B... heehee.. cuz it was an \"E\" in my prelim! haha!! btw my grades in As really a miracle... cuz it was a COEE to AABC... ( in this order) heehee... :)     okie... gotta zzzzzz.. may sing soon.. yay!! :) break might be over sooon.... :) sis val just sms me.... :)   '),(1651,3734069,'14,August,2004','2004-08-14 00:00:00','   Amen!! despite after the last run of the pom at 730.. that we still not sure of wat to be done.. amen! turned out great! Atmosphere was good... btw we did the TRP cheer for POS (cuz bulk of them from TP... and the overall i/c brother melvin as well...)  One thing I learnt from brother melvin.. he sees the good side of everyone.. slow to reproach, quick to praise... to boost our confidence cuz we were like kinda discouraged that we can\'t get the steps right... revelation:  you need not harp on what went wrong to get things right... treat pple lovingly... the greatest of these is love... 1 Cor 13:13  love changes changes everything... (one line out of the odie that i like...  love changes everything... )  gotta zzzzzzzzzz... meeting 12:30 at church... stunts pple even earlier... :)   '),(1652,3734069,'13,August,2004','2004-08-13 00:00:00','   hmmmm I\'m fine... just tired... hmmm i think i\'ll blog later... or tmrw...  but i did have a good time tonite.. and made some new friends!! like michelle..qiuyun... Jo.. carmen... roy, sis xuemin.. btw I think TP pple rocks!! (i know quite a no oredi..)  i didn\'t noe amanda is still studyg! (in TP final year ithink..)  okioki.. gotta zzzzzzzzzzz :) watch this space for more!!!   '),(1653,3734069,'12,August,2004','2004-08-12 00:00:00','   heeheehee... I\'m in the POM for YOUTH ministry... heehee.. !  received a fone call from ryan at 1815... when i was about to leave office.. to visit dad at ttsh... events was a whirlwind... but dun wish to elaborate now... Just pray that as i serve... He will change the way...  btw YAY! Eunice might be coming for service1.. I\'d asked her half-heartedly after gym during lunch today... Praise the Lord.. she seems excited abt this... now with POM wonder how...  heehee... so.. went to church via cab (ya, i pretty broke now but was running late... so took a cab... )met sis yah lan... she asked us to join a brother when she realised that we are the helpers for POM... turned out that the brother was in youth min.. so hahaha.. lian and i became part of youth min... hey saw amanda.. wow...she dance sudsud.. happy to see familar face.. ( amanda was standing next to me during easter prdn..) Brother Melvin was our leader... man I salute him.. he just came back from mission trip and he came up with the cheer and steps today!  hmmm the dance, if done properly would be stylo... haha... me not dancing... at first tried learning the steps.. but hahaha.. too short a time (the rest has been doing this since 1730... )so hahaha...  Praise the Lord, I\'m just doing the clapping (hahha.. My area of testing... I guess after this maybe break will be over?? ) did great... hmmm shall not elaborate more... heeheehee... o.. saw michelle (from dance &amp; choir) too... :)  turned out that lian not involved, her usher head dun want to let her go... hmmm we shall see... man.. tmrw prac is at 6pm.. gotta take cab down again... today i left at 11+ (still got one more rd of prac when i left cuz Pst Kong coming down) but i gotta take 2 ypung girls home .. they\'re 16.. got curfew... hmmm estella\'s dar looks good in the army uniform...  o... i gotta wear my bread uniform... funny ah? but they dun have enuff JC ppl... so.. heeeheehee... okie... gotta zzzzzzzzz ... :)   '),(1654,3734069,'11,August,2004','2004-08-11 00:00:00','   1. Robin : for divine healing in his ankle and leg.. that absolute no harm will come to him during his camp.  2. Mary\'s friend\'s son, Jonanthan: divine healing in his left arm that has been feeling numb.. for him to experience the love of God again... that he will be well enough to attend our 15th anniversary this weekend! :)  3. For a fruitful harvest this weekend, that all in w110, e206, w285 not just talk the talk, but walk the walk! perserve on brothers and sisters in christ!!  ~~~ anything else?? ~~ ~   '),(1655,3734069,'10,August,2004','2004-08-10 00:00:00','   hmmm I have not blogged for daysâ€¦ Many things have occurredâ€¦ good and bad.. happy and sadâ€¦ encouraging and frustratingâ€¦ being a long-winded person by nature, I would try to keep this entry as short as possibleâ€¦ (heeheeâ€¦ my dear pals would know that Iâ€™m rather lengthy when it comes to writingâ€¦)      Okieâ€¦ Saturdayâ€¦ Service dayâ€¦ I think I did bloggedâ€¦ hmmm sermon message very applicableâ€¦      Sunday      Had a good day at service 3 choir outing.. not many turned up as the one at marina bayâ€¦ hmmm got to know some more ppleâ€¦ (seems that most of the youngsters didnâ€™t come, except Peiyuâ€¦) hmmmmâ€¦ I didnâ€™t know that my presence significantâ€¦ Peiyu was asking me when my break will be overâ€¦ it seems different without me.. heeheeâ€¦ dear girlâ€¦ soon la.. I hopeâ€¦. yayâ€¦ my group came in topsâ€¦ heeh  ee.. got a bookmark for prize.. thatâ€™s so adapt to me currentlyâ€¦    â€œSome people see things as they are and ask why. Others dream things that never were and ask why not. What your mind can conceive and believe. It can achieve.â€       The games ended 12++ â€¦ we went to lagoon food center to makanâ€¦ hmmm ate my fave duck porridge.. (have not ate it for apprx 1 year cuz the food center undergone renovationâ€¦) hmmm after lunch went causeway point with sis Jocelyn, Jo & Catâ€¦    J    hmmm cat had a good time at the array of cosmetic at the JL thereâ€¦    J    sis Jocelyn always on phoneâ€¦ looking for pple for the parade of ministryâ€¦ sigh.. I wonâ€™t be able to help despite I should be able to go all 7 servicesâ€¦ cuz still break mahâ€¦ hmmm some were rejected cuz of their countenance not goodâ€¦ man.. heng.. I donâ€™t fall into that categoryâ€¦ hmmm neverthelessâ€¦ there will always be next year!    J      left causeway point ard 5:15..   took 969 to tampines.. heehee.. it was fastâ€¦ reached tampines in less than 50min! hmm went to a D.Y.I. Shop to get my shower head (ya, it went on strike on satâ€¦ hmmm to prevent my mom from nagging.. decided to fix it myself!) Itâ€™s very easy to replace.. unscrewed the old one.. and screwed on the new!!    J    hmmm wat else did I do? canâ€™t recallâ€¦ slept very earlyâ€¦ ard 11â€¦ and woke up 10+ on Monday!      Monday ~ August 9.. Yay! Happy 39 th  Birthday Singapore!!       Had a good QT! was reading Roms.. then was x-reference to Habâ€¦ Hab2:2-3  \"Then the Lord answered me and said: \"Write the vision And make it plain on tablets, That he may run who reads it. For the vision is yet for an appointed time; But at the end it will speak, and it will not lie. Though it tarries, wait for it; Because it will surely come, It will not tarry\"    Amen! Despite that the vision He has given me has not come to pass... it doesn\'t mean it\'s not real.. just that the time ain\'t ripe! Gotta wait....      met simin @ 5:30+ hmmm had a quick dinner at moss (which was good cuz the money bag kinda emptyâ€¦.and my phonebill for August (combined) is just a few dollars short of $200!!) she lamented abt the unmarried colleague whoâ€™s kinda good catch but als.. a Christian, so impossible (I broke the news that usually believers will & should date believers).. hmmm she tried to intro me to her â€œshort, fat, uglyâ€ colleague to meâ€¦ haha cuz he too was a Christian.. manâ€¦ hmmm heâ€™s two years our senior. also from hwa chongâ€¦ Ekâ€¦ hahaâ€¦ I know who I fancy and He knows when heâ€™ll comeâ€¦    J      saw mei, my colleague, @ Isetan Orchard, with her sisters..    J    hmmmm we jalan-jalan a bitsy at takaâ€¦ then went to far eastâ€¦ simin was complaining abt her doctors colleagues didnâ€™t want to gop far east.. cuz very lian & cheapoâ€¦ (ehâ€¦. hmmmm)      Well.. I bought a pair of ballerina pumps at 29.90.. very sweetâ€¦ itâ€™s white with rainbow stripes across.. (hahaâ€¦ cannot buy pink stuff anymore.. I got a lot of pink stuff recentlyâ€¦) man this is the story of sowing and reapingâ€¦ I elaborate laterâ€¦ saw Adrian when we were abt to leave far east.. (cuz 50% of the shops had pulled the shuttersâ€¦by then.. ard 9pm)     tried to invite simin.. but she\'s on call this weekend... (meaning she\'s gotta work round the clock! btw she\'s a doc, currently working at imh...) hmmm... to make myself happy... i went to Mango.. bought myself a sweet baby pink turtle neck.. (ya.. I know it\'s pink!)   not that ex-just 35... :) but i guess that will be the only item for the month...      Tuesday: Today!!      hectic day at work... (hmmm i can be pretty nasty to nasty pple... not my colleagues..but the TPs... hmmm) hmmm invited Cherie.. she has yet declined... gotta pray that the chalet she\'s going will be a drag and she will go service on sun! (she sounds rather keen when i gave her the invitation! )     yay! Yeeling (lian\'s friend) coming to cg on thurs!! :) just spoke to her for 50 min.. (feel bad.. i called her hp... she\'s at hall... ) hmmm wonder why eugene and joyce said that I\'m too busy to give her bs... Hey.. i enjoy giving my gals bs.. I get to know them better, and they get to know me better!! :) (that\'s how I got to know Vick & lian a whole lot better!! ) maybe that\'s why i dunno joyce that well... never gave her bs... giving BS is not merely acting as an information provider.. it\'s impartation of knowledge.. and for me , I never fail to learn something new each time i give my gals bs... :) (yay! Lian just sms me.. huiling coming for cellgrp this thurs.. we were just talkin abt her.. nah, not ms banana.. :p)     In shortâ€¦ am glad that I had not been living on stale manna for the past few daysâ€¦ bread must be fresh!! hahahaâ€¦ brought to mind how my dear pal violet and I coded the JCs via their uniformâ€¦ let me see if I still recall?     here it goes:     * TJC= Tree/ mushroom (their badge VJC= butter NJC= Stone HCJC= bread NYJC= burnt bread! CJC= SBS bus driver (males) & factory workers (females) TPJC = Nurses AJC = blueberry ACJC= hmmm this one we didnâ€™t name but we like the uniform the best!.. eh the other JCs,,, hmmm canâ€™t remember the restâ€¦.     ps sorry if u feel insulted.. heehee... i was rather notty then.. but violet and i pride ourself to be bread!!      talking abt Pink... hmmm in uni.. there was this gal that we like to label her as pink friend.. cuz of her penchant for pink.. man... little do i realise that recently.. esp 2004... i had became a Pink friend.. heehee... vicky commented that i like to look at pink stuff... and so did simin... heehee... confession.. ya... :) maybe its a PINK year for me... heehee... :) but please dun call me pink friend ya?      yay! my left fingers went a-peeling-again!!'),(1656,3734069,'08,August,2004','2004-08-08 00:00:00','   heeheee... wOke up at arOund 9.. had my 2 slices of cranberry bread and 3-in-1 coffee... dumped my clothes into the landuary... had decided to clean up my Room.. (which is rather... hmmm heehee... only i find it comfortable...) and re-organised the layout... (actually just shifted my table to the side.. shifted the TV to the side... and my beanbag beside my bed... heeheee... but..  took me 1 1/2 hours... to swipe/mop the room.. and yay! I found my transaparent guitar pick! ( but my fossil watch is still MIA... wonder did i left it in office... Ops!)      11:45.. showered... called mary (my counselee) at 12 nn when i saw siew choo\'s msg.. man.. she sounded soo different from last week when i called her...so positive... God is good.. He has truly delivered her from her past hurt and pains... was encouraged! to think i\'d thought that she won\'t come... man.. I\'m of little faith.. but i did a short prayer b4 calling her... :) 12:05 started my QT... man... joy & peace prevalent..  was sUre that the sabbath ending soon... aniwAe... hmmm did my make-up (man..  a little foundation and blusher did wonder! hahaha! man.. i think my usual make-up must be real heavy... Jacelyn was telling me that my make-up lighter when i\'m not serving...       gee.. it was 1pm already.. got guitar lesson at 2... sooo went downtown east to draw $$ and took a cab to NTU.. hmmm am tight on budget but... dun want to be late... 1400.. reached Canteen 14... was wondering if that\'s the correct canteen.. hahha saw yili..  phew.. correct venue. after lunch..  we went to vick\'s room... hmmm did made some progress... hmmm yili shifu... thanks 4 ya greAT patience! heehee ur shi wan ge wei she me lil brother did trained ur patience ah?? after prac, we nau at vicks\' room till 6:15... then went to hall 14 to meet yili\'s friend.. shiliang... 18:55 reached church.. man late..  joyce managed to get only 2 rOws...but darren got another row on the side...service was great.. sermon was like a double-edged sword to me as well... hmmm truly, u need the Word of God to be victorius in life...was brought to mind how i had let the spirits of negative sow seed in me...devil got the better of me for the past 2 months ... level of spirituality went towards the gravity like an avalanche.. but.. I was found again...  Pastor Kong asked who were sick to respond after the sermon.. wellwell... I was having this cough.. since monday.. kinda phelgmy.. but i felt much better after PnW.. then...Pst Kong asked members to lay hands on those who were sick... hmm Ivory prayed for me.. :) (Thanks Ivory!  and Tracy... thanks Tracy!! see ya at eastcoast later... ) and truly.. the phelgm was like GONe.. after that... (heheeehee... one sister from another cell sitting behind me.. asked me curiously.. DID God really heal U? told her yup...she was totally amazed! HAha.. hOpe this will strengthen her conviction!! kekeke..     hmmm actually during the ministry time after sermon.. while we\'re singing amazing... guess what...God spoke to me... there isn\'t any gutiarist in ur cellgrp...U\'re now learning the guitar... why not rise up and play for the cell... I was like hmmm.. but I\'m a gal le... and my guitar skills is so amatuer... then He brought to mind a couple of gals that played for their cell.. so.... hmmm let see how... but i\'m determined to learn... heeheehee... there will be improvement....    saw mario and jolene after service... hmmm missed jolene... hope can catch up with her more tmrw... heehee had a good time teasing  mario...  kept calling him brother mario.. he\'s gonna to be cgl...  yay!       glad that shiliang enjoyed his time..long.. keep up the good work...    heheee.. had a good time teasing JAcelyn... kekeke.. dunno why i started to be more notty towards her since thurs.. haha... hmmm brother max.. hahhaa.. i shall stop calling U ma... hahha... U gotta maintain ur image... heehee..but it;s ur choice to expose ur identity in cellgrp... aiyo.. i hope the members didn get the wrong ideas...since u\'d told so ambigiously how u earned the nick... hmmm going for s3 chOir outing @ east coast later.. haha.. they meeting at 9am at bedok... kekeke.. can leave home at ard 8:30... hahhaha...     hmmm hOpe that monday can get simin & anne to the 15th anniversary... maybe jio huiming to see wanna join us tooo... :)    okie.. let me fiddle with hubbie.. &amp; zzzzz.....   '),(1657,3734069,'05,August,2004','2004-08-05 00:00:00','   hahha... think my grey matter ain\'t functioning that well...       had been working late for past 2 days... will reach home 11+ if i din share cabbie with colleagues... if not for cell grp later.. i guess i\'ll have to OT as well... man... this week was like a circles...everything is sucha mad rush... ahhhh next week gotta help in the dnd ticketing... man life was easier last year as a hon secr.. (haha I call the shots ma.. but i seldom delegate.. ) life as a subcoommer--- hmmmm shall take it in my stride... man.... Productivity! that\'s wat i need....  would spend tmrw evening deco the mulit=purpose hall...       Ah.. the center of my face kinda red and raw now.. had been rubbing my nosey... cuz it\'s more active than my grey matter... hahah.. long & rachel.. if I sounded crappy.. prolly cuz i was tired... am i avoiding  anything... eh... no la... brilly dear.. sorry wasn\'t much of a help..      perhaps I\'d made an incorrect choice 6 years ago... man... but despite choosing the not my ideal option... the other option seems to be open still... hmmmm PS37:4.. Delight yourself also in the Lord, And He shall give you the desires of your heart... Amen!       ya.. am speaking in parables again... heehee...       hmmm lian hOpe u\'ll make a wise decision b/w make-up econs or make-up cell...   '),(1658,3734069,'04,August,2004','2004-08-04 00:00:00','   man... am right now in training.. having my 20 min break... ah... no chance to work today cuz whole day would at training.. man.. imagine system training w/o letting us hands on...   hmm.. the flu season\'s here... hopefully wOn\'t get into boughs of coughing again... yeah! It\'s wed today... hmmmm i must start inviting friends soon... :P  brrr it\'s cold here...   '),(1659,3734069,'03,August,2004','2004-08-03 00:00:00','   The talkative mOde again... Arggh... left office around 21:15.. reached home around 22:10.. man... tired.. spent 2 hours generating the coupons... (2100 of them).. printing.. cutting.. perforating.. and sorting... then the letterings for the stage.. and I still need to cough up my cases to boss tmrw.. and I\'m on FULL DAY course tmw (wed) ... man.. i\'d a very scumptious dinner... think i only have the opportunity to eat a good dinner once a week.. i like the soup that my mom made... like what they says, Cantonese are expert in double*boiled soup! :)    hmmmm got a shock when i checked my email..saw a friendster message from a mattew?? man.. I\'d no connection with him at all.. he\'s a friend of my choir friend.. and this man is like 38!! hmmm ?? so told him that sorry to decline his request and that I was not in the \"looking for partner stage\".. ( I was rather stOned.. cAn\'t think of the words to tell him that I\'m not interested.. he ain\'t my type... ) oo... I was chatting with max online then.. so i told him this la.. and I was bewildered at his reply...  i was totally confused so..bo chup... anyway.. continued the chat with him.. ard 1.. when my dotter logged off.. my dotter blogged too.. the middle one... :) man... ended up that haha.. I got myself a mummy.. (he\'s trying to freak me out so.. he will go with any title... ! hahaa!! )     RAchel.. thanks for ur sms... hmmmm very extremely busy so i didn\'t reply...  revelation? I was reading the book of Ecclesiastes... Ecc7:8 to 9 \"The end of a thing is better than its beginning; the patient in spirit is better than the proud in the spirit. Do not hasten in your spirit to be angry, For anger rests in bosom of fools..\" I was encouraged... that despite of my tumblies.. my backwoods manner... how i started. ain\'t that crucial.. cuz it\'s how i end the race... do i give up halfway and throw the baton on the ground... Nope... I\'m must be patient and endure... gettin out of the tunnel... ! man.. it\'s getting more an more exciting every year! This year is the most eventful by far!!  yay! I\'ll be 3 years in chc soon!  okies.. late... gotta be in office by 8.. (cuz lesson starts at 8:30.. and i need to do my cases and to send something out.. ah.... THank God for his wisdom...!  okie... ot\'s 230am .. late... ciao!   '),(1660,3734069,'03,August,2004','2004-08-03 00:00:00','   Hmmm I was too tired on Sunday to blog...   hmmm reached home around 9:45 pm...gee, I was out for more than half a day...   let me recall the events...(it was sure an eventful day) woke up at 08:30.. very tired.. went back zzz...Gee...9am.. late! sms lian.. told her I\'d not meet her at bedok...9:15am left house...reached church at 1050? long arrived ard same time.. Eugene was already there waiting for Qingyun (what a sweet sounding name ah? My impression of this person should be someone of a sweet composure.. demure?man I was wrong! )    Meanwhile.. Eugene was telling me how fitting?the shirt the cell-group got for him was.. Vicky, we under-estimated his chest size! haha!    QY arrived at 11:05...(Fortunately, I\'ve got lian to book seat for us?) hall only opened around 11:20.. well.. between that, tried to make small talk with her.. Real conversation with her was not easy cuz she has ATTITUDE! Heeehee?saw bimei @ staircase.. haven\'t seen her for sucha long time.. (ooo?mei is my youngest dotter)? I missed her? guessed that it not the length of time we know each other but.. the chemistry we have? :) okie.. in B4.. man.. QY just sank into her seat and played with her 8250...hmmmm tried to talk to her but to no avail.. long and I tried to get her to stand during p&w but she refused...wellwell.. lian came over and asked if she could get her to stand up for p&amp;w.. Lian asked her \"do you mind if you stand up ~ as a sign of respect?? guess what\'s her reply YES, I mind!\"\" Ohmigosh! After p&w, she asked if she could leave now...Eugene asked her to stay a little longer ...but she left after we stood up to pray for the mission team.. MAN! (If she stayed, am positive that many demons would get delivered!) wellwell.. Eugene was rather disappointed that she left even before the preaching of the word...    Hmmm atmosphere of service 5 was rather ok..compared to service 3..but I perceived more..cuz.. hahah..the condition of my heart.. hmmm the matter of truth was that I don\'t even feel like to open my mouth on sat nite...(yah, I got a very bad countenance.. I faked my smiles.. Ops.. but.. Sunday was different... heeheehee.. it\'s really a release and relief and free)  Saw jason &amp; linda & son after service... man they have all grown in sizes... Jason look sooo rOund!! think he\'s attending service 6.. hmmm made some small talk with them.. :)  makan lunch at kopitiam with long and lian...hmmmm one kid approached long to buy lucky no. from him (he looked rather pitiful..and gestured that he didn\'t have $$ for food.. ) kind-hearted long asked if I\'d change.. in all.. we gave him $3.. heehee.. not sure if the no came up in the 4d last nite.. think it was 1490? hahaha  after lunch.. went with lian to her hall...hmm on way to bus-stop...saw my darling dotter from afar outside church...man.. I was just telling lian that I saw my dotter with her friends when she shouted MOMMY!?~ not once.. twice! heeheee? so cute.. she got a cutsy voice.. hahaha just like me!! hahaha talked about sowing and reaping?I became lian\'s \"Maria\"for the day? heeheehee...and I was wearing a skirt...room was rather dirty..perhaps her side of the room was not occupied for sometime...and the blinds were sticky to touch!! started off by praying for the room and her roommate (by her timetable, conclude that mostly like she\'s a third year eng student?) Was rather inspired by the cleaning-process...should clean up my room also.. The brand of detergent we concluded to be very effective in cleaning up the blinds.. Dynamo! (we \'bOrrowed\" from her roommate cuz she forgot to bring hers.. ) then it started pouring at ard 6+...hmmm we finally completed the cleaning process around 8pm+ and it was still pouring...Thank God for sheltered walkway.. went to hall 4 bus-stop to take 179 out..  Man..was rather tired by the time I got home.. fell asleep on the couch while watching some info-ed program...dragged myself to room and plonked onto bed at 11:15... man...at 12:05 my fone rang...it was the same no. that called me last nite when I was fellowshipping with cellgrp.. I guessed getting long to tell him its the wrong no. was not effective enough.. if he called again..I\'m gonna tell him to stop calling or else I will file a report to the authorities..  Got an irritating call from barley. her name sounds like barley...got barked by the wild barley for no reason... and some very inquisitive staff to enquire about the D&amp;D .. (cuz I just sent out the 3rd teaser to staff...ahh.. am not in the mood to entertain cuz my throat rather itchy...maybe cuz I shared the green tea with lian yesterday? cuz lian got cough?or maybe it\'s the rain? )  later gotta meet violet.. don\'t really feel like meeting anyone.. cuz I got chunks of work to do.. but gotta meet her...cuz I really wanna invite her to church? hopefully for the 15th anniversary?Not sure by then would I come out of retirement mode?I hope so?  10:30 pm.. (monday) reached home.. hmmm din exactly invite violet but..at least.. gOt in touch with her again.. (she rejected me straight in face last week... let me give her some breathing space 1st! she\'s a U3.5 to U4) went shoppng with her... hmmm saw this light pink corduroy @ mango... hmmm wanted to buy but...hmmm out of budget... :( still short short of my mission pledge...) guess that I\'d changed quite a bitsy.. used to be very super materialistic.. hmm my dear friend was like trying to find the perfect bag (note: must be designer-brand).. i was like...if i were u.. let me give unto mission... ( i used to be a sucker 4 brands... remembered that the 1st item i bought with my part-time pay from POSB after JC was a JPG long wallet (hahaha.. and all my pay from working 2 weks in december went into it... !! )  hmm now gotta go shower and start cutting alpahabet for the national day celebration... man this gonna be a superbusy week... am on course tmrw and wed.. and I gotta print another 2000 coupons (serially-generated by hand &amp; excel...) and to proferate by hand....) ..... and my cases... ahhhhaaaaaaa!! think i might have to work this stt morning again!! :)  I hope i\'m not on ministry \"sabbath\" for long... dun think soooo.... aniwae... ja ne!   '),(1661,3734069,'01,August,2004','2004-08-01 00:00:00','   Man.. i must look real bad today.. man... i was not least in the mood to praise and worship.. and yes.. i dressed very down today..  ah.. i need to stop doing that... man... even if i dun serve next week.. you won\'t see me in tees and jeans... Claire saw me and said.. i look very tired and dun have the countenace... man...       ah...going service again tmrw... cuz lil bro called me and ask me to acc qingyun.. so.. i agreed and ask long as well.. cuz qingyun from ntu mah... and man... a disgusting man called me.. while eugene talked to me... so i shoved the fone to long and ask him to answer on my behalf.. This guy called me twice at 11pm+ on friday, again at 11pm.. Long.. xiexie ni....       need to go... to recharge and recharge... wonder can help lian shift later...       ah... i need my spiritual bath... and thanks to the kind sOt sister i met just now... hope to serve alongside with u soon too!   '),(1662,3646978,'23,June,2004','2004-06-23 00:00:00','    See I knew it would be hard to keep it up every day.  I really have been busy though!  Let\'s see...baseball Saturday and Monday. My youngest son had his last game in his minor league. He will move up to MAJOR league in the Fall where the boys pitch harder and faster. He is a pitcher now as opposed to a catcher and he is so much happier. He does really well and enjoys it. Then my older son played his last game Monday night. They played 13 games and won one. That was their last regular season game.  Then my dryer finally bit the dust so my DH got a new one last night that matches the washer we got last year. Man does it make a difference! I spent the whole day yesterday cleaning the garage to make way for the new one.  Then last night we took a party over to my grandmother\'s house for her 80th birthday! I can hardly believe that she is 80. I lived with her quite a bit growing up and it is really hard to watch her grow old. She has had a lot of health problems lately and it is harder for her to get around. She also seems to be losing her memory functions somewhat and she is afraid she is getting Alzheimers.  But we had a great night eating and celebrating so that is good.  Anyway....off to do some stuff.  :) '),(1663,3646978,'18,June,2004','2004-06-18 00:00:00','    My baby is the most beautiful baby in the world. It\'s true! One look at her and it is obvious. But she is a handful! She has about a fifteen minute limit on time spent in any type of constraint. Exersaucer 15, carrier 15, high chair 15, etc. She either wants to be held or be put on the floor where she can freely explore and pick up loose change, marbles, small bits of paper and legos. She has fully developed what someone called the cheerio pinch with her forefinger and thumb. So now she is able to fetch the smallest item and deftly aim it toward her mouth. I usually watch her like a hawk and catch anything she sneaks in but I did find a piece of paper in her ummm...diaper remains one day. She is not really crawling yet but she can sure move. I am afraid that the pack and play will have to come downstairs soon to contain her. Have you ever noticed that toys hold little appeal to an infant as opposed to smallish dangerous items? My baby likes to play with the strangest things. I gave her a toilet supply line (clean of course) and she spent more time with that than any of her toys! Maybe all toys look alike? We as parents choose the cute bright (expensive) toys that WE like and they take one lick and throw them on the floor.  Kenzie now is the proud host of TWO teeth. They have popped through in less than two weeks after all that time teething! Nothing gives her greater joy than biting my finger with her little sharp teeth.  Aren\'t babies neat?    mary '),(1664,3646978,'17,June,2004','2004-06-17 00:00:00','    It is late but I wanted to try to stick with making a new post each day.  My son had a game tonight and it was a great game. He pitched very well and only gave up one run. The problem is that we lost. :( But the bigger problem is that my husband fell into the role of coach dad. He was picking out the things my son didn\'t do instead of what he did. I think I will have to remind my DH that it really is just a game and the bigger lesson is good sportsmanship and trying your best.  I was thinking about that little boy who was all over the news because the jerk man got the foul ball and wouldn\'t let the kid have it. The kid is like 4 and has no clue what has been happening. He was on GMA and couldn\'t even sit up on the couch. It was apparent that he didn\'t know what the cameras were and that millions of people were looking at him. (or maybe he didnt care) Anyway, the guy was a jerk. He should have given the kid the ball. BUT, the kid had no interest in the ball. His mother flipped out and started hitting the guy with her cardboard. (It was very hot that day) I think the guy has given in and said he will give the kid the ball now. All that for a foul ball from the TEXAS RANGERS for pete\'s sake! It\'s not like it was Red Sox/Yankees World Series or anything! He is a jerk and the mom was a typical mom and the kid is just another goofy preschooler who probably didn\'t watch one second of the game.  Have you ever been to a game with a 4 year old? They act like they have a whole MOUND of ants in their pants!   Nuff said Time to go to bed    mary '),(1665,3646978,'16,June,2004','2004-06-16 00:00:00','    One of the things I despise in life is laundry!     I have 6 kids and a husband so my laundry pile just grows and grows. I do laundry like a mad woman and get the pile down way far and then I say \"whew, I can slack off a bit now\" and the next trip out to the garage is like BAM! This huge mountain of dirty laundry! The funny thing is that I am in this group that is supposed to help you get control of your house and chores and all that. I have been in it for years,,,,it hasn\'t done much! I guess I am just so worn down that when I have a free 15 minutes all I want to do is sit down and zone out!  My youngest child is the most precious 8 month old in the world. BUT she won\'t nap for more than 15 minutes without me beside her. This severely limits what I can do. She wants to be held or be crawling on the floor. Well, kind of crawling....more like the army training slide.  Anyway, Life is kind of tight right now as far as time.  My boys are playing baseball day and night, my daughter is in a training camp at the military base for her JROTC unit, and then my 5 and 3 year olds are on a constant mission to destroy my house.  Back to the laundry....what is the deal with being able to fold it but not put it away? I have these piles everywhere! Piles on my dresser, piles on my chest, piles on the kitchen table, piles on all the kid\'s dressers. I had the bright idea to use the dining room table as a folding area since we rarely use it. Do you know how many piles you can fit onto a 5 foot oval dining table? A LOT! So I scrapped that idea. Right now I have two piles on the sofa in this room and instead of putting them away I came over to check the voicemail and then decided to check my email. That\'s how my life runs. One distraction after another!   Well, the demons just walked in and they look suspicious so I better run. Yesterday they threw toilet paper wads onto the mirror in the bathroom. Fun.     mary '),(1666,3307733,'11,May,2004','2004-05-11 00:00:00','   Cyndi is gone for a week in the Bahamas so I\'m doing her job as well as mine. The dynamics of the office are a lot different without her around and while it\'s stressful in someways, it\'s better in others. I am doing the work she does, but doing it right and trying (albeit only moderately successfully) to not panic when I have so much on my plate. She\'s leaving in June for a career change since she just graduated and although I\'m looking forward to it, I\'m also kind of freaked out about it. That means a new person to train, and I think I\'ll be the one doing some of the training.   Office politics are also interesting. It seems that everyone has a beef with everyone else on some day during the week. I\'m just generally bitchy, though I don\'t mean to be. I\'m bringing my day to day life with me to work and I need to stop doing that. Stress in life needs to be seperate from stress at work. That way it won\'t compound things. Today I started out with a nasty attitude...but talked with Mary and felt better immediately. Tomorrow there\'s some sort of meeting at 8:30 for the whole company. It seems that we\'re having a change in the way that insurance is being taken out of our checks. Pre tax instead of Post tax I think. In other news, there should be raises coming, if what I percieve is correct. Hopefully that\'s just a boon to me with another raise on top of that for a change in position.   Boring stuff..but it\'s my life.   Went out to take care of Erik tonight and he read a book to me. That was a wonderful thing to hear. :) I\'m so proud of this little boy. He\'s working so hard at school and is struggling through the desire for independence and the need for his parents. At the tender age of 7 he\'s already been through so much. I hope to make his life a little easier in the years to come.   '),(1667,3307733,'11,May,2004','2004-05-11 00:00:00','   Just testing this thing out! Woohoo! I\'m blogging.   '),(1668,3307733,'20,June,2004','2004-06-20 00:00:00','   Father\'s Day 2004. I haven\'t spoken to my father in 3.5 to 4 years now. My ex Step Father is recovering from a day out at the ballgame yesterday. My current Step Father is with my mom 4 hours away at a family reunion.  On Mother\'s Day, my mom was out of town. Easter, mom was with my Step Father\'s Family.   So, seems like it\'s a crappy set of holidays this year. ERik will be with his father on Christmas. So I\'m sure I\'ll be here in St. Louis instead of with my family. I don\'t want to be around them any way.   Today is a bitter day. Ugh.   Blogging is entertaining, I guess. I\'m going to start writing in here more. I\'m going to write about specific things having to do with my family. Maybe this will be the catylist to begin the book I know is there.   '),(1669,3307733,'11,July,2004','2004-07-11 00:00:00','   urlLink CNN.com - For fall, NBC will rely on reality, aim for laughs - Jul 11, 2004   Goooood God!!!!! Not more reality television. Kill me now.   '),(1670,3307733,'09,July,2004','2004-07-09 00:00:00','   urlLink Disinfopedia - Disinfopedia   Just something interesting to read   '),(1671,3307733,'05,July,2004','2004-07-05 00:00:00','   urlLink Fundrace.org   This is a little creepy, but cool. Not only can you search out city maps, but your neighbors, and find out how much someone has donated to which candidate.   '),(1672,3307733,'05,July,2004','2004-07-05 00:00:00','   urlLink NEWSMEAT - Hall of Fame   Stuff that\'s interesting in an election year.   '),(1673,3307733,'05,July,2004','2004-07-05 00:00:00','   I remember dreaming about sitting in a swing as a very young girl. It was the fall and my father was pushing me back and forth, back and forth. I remember being happy to be with him.   Now, as I sit here at age 26, I feel that I will never have another happy dream about my father. I will no longer have happy memories of him. There aren\'t too many to remember anyway.   There was a day when I was 4 years old, sitting on the couch and crying as I watched my father take his suitcase and walk out the door of our home. When I was young, I wanted nothing more than his approval and his love and I suppose things haven\'t changed much in that way. I would still like his approval and love. I\'m not going to get it.   This week we exchanged e-mails. They were nasty and hateful. I haven\'t seen him for years...hadn\'t spoken to him in over 3 years. He\'s lived in Arizona since I was seven years old and I have never seen him much.   I sent an e-mail out to everyone on my list and ended up sending it to him. There was a question asking who would be least likely to respond and I put he and my stepmother. A few days later, I recieved a response criticizing me for that comment. I was speaking nothing but the truth and thought it would be glossed over. Apparently emotions are raw in that household and I struck a nerve.  Over the course of two days, he called me names, told me I was just like my mother ( the biggest insult he can weild) and told me I shirked my responsibilities. He said I was still angry with him about not giving me money at a time when I needed it. If I hadn\'t been desperate, I never would have asked. He said that he needed to help my step-sister with the purchase of a house and I said as much and said I understood. He wrote back stating that she had never needed their help and took nothing from them, so I should leave her the hell out of it. Caught in a lie apparently. I left my boyfriend and left him with custody of our son because I couldn\'t have handled being a single parent. My son would not have been in a good environment. My ex and I both new that and instead of going through the courts for a nasty battle, I conceded and left my son with him. My father did the same when he left my mother and apparently holds anger and guilt over that situation. He acts as though it was an easy decision for me. Acts as though I just left and went galavanting off on a childless vacation. In reality, I left and started to rebuild my life and myself.   Now my son\'s father and I are closer than ever and we all have a wonderful relationship. Given the circumstances, that\'s the best thing that could have come out of everything.   Anyhow, nasty things said by both sides.   I decided to have no further contact with him last week. Ever. Maybe when I\'m older that will change. But for now, he and I are both too stubborn to apologize. He hasn\'t been a father to me for years and I can\'t say that anything much has changed. Only the reality that he\'s the man I had always heard about and had always hoped everyone was wrong about. I still can\'t believe he acted in the way he did.   He attacked my aunt as well. My closest friend and ally in the family. Apparently he thinks I\'ve been poisoned against him. The two of them have had a fairly long standing feud. He said to me that family is the most important thing to him. When his parents are pissed at him, his sisters will no longer speak to him and he attacks his only biological child.   I\'ve never seen a father act in this way toward a child.   Now, all of this is my perception. A one-sided arguement, I guess. I will not post e-mails here because they are simply too much to handle at this point. I want to sweep it all under the rug and forget about it. I haven\'t needed him for years and will continue to get along without him. It\'s the finality of the decision. I had always hoped he and I could be close. Apparently that\'s not going to be the case.   Could I have done something different? Probably. But I\'m not the parent here. I\'m his daughter. I shouldn\'t have to clamor for the approval of an apparition. If he had wanted a relationship with me for the last few years, if he could have seen past whatever anger he held toward me, if things could have been worked out enough in his head to have a conversation with me....then maybe the situation would have been different.   I have an ex stepfather whom I have called \"dad\" since I was fairly young. He\'s the one who was there and continues to be. I have a stepfather now who would take care of me if push came to shove. I have a great family with lots of loving support. It took me until last week to realize what a support network I have. Randy (the bio father) seems to be the only one who doesn\'t want to be a part of it. He wants to get to know his grandson, that\'s evident. But he will not get to know my son until he gets to know me.   I\'ve already checked out the legal issues governing grandparental rights and have found out that he doesn\'t have any rights. He hasn\'t seen Erik, hasn\'t contacted him and it\'s not like he\'s been seeing him since he was an infant. So, that\'s that. If they want to take it to court, then so be it. I wouldn\'t put it past them.   Anyhow, that has been the main part of my week.   Happy Fourth of July to me.   '),(1674,3146636,'29,April,2004','2004-04-29 00:00:00','   urlLink  I adopted a cute lil\' death fetus from Fetusmart! Hooray fetus!  '),(1675,3146636,'27,April,2004','2004-04-27 00:00:00','   just another play with colours in PSP8  '),(1676,3146636,'26,April,2004','2004-04-26 00:00:00','   My grandaughter Lily, adorable to me at least :)  '),(1677,3146636,'24,April,2004','2004-04-24 00:00:00','   another of my \'playtime\' pics from paintshop pro - same filters  '),(1678,3146636,'20,April,2004','2004-04-20 00:00:00','   Perhaps what I wish to do isn\'t quite as adventerous as  Tom Neale , but I also want to leave.  Maybe I\'m just tired. Maybe I\'m drunk. But maybe, just maybe, someone put an idea into my head that I can\'t shake.   Granted, this person didn\'t say Utopia, I did. As a matter of fact, this person stated that Utopia was \'unattainable\' due to \'them\' putting the idea into our heads that it isn\'t. I however feel that utopia is a mindset, and that it is attainable as long as  I  do not lose site of the goal.  All I wanna do is have some fun.   My idea of fun may not be the same as yours. I want to build my house out of adobe and straw bales. I want to grow my own food, harvest my own food, hunt my own food and live my own way.  I want to create a new \'civilization\' based on not the most \'technological\', not the most \'hip\', not the most \'current\' life styles. I want to draw from past civilizations that have prospered for a while.  I may or may not want someone there to do it with me, it all depends on my mood....  '),(1679,3146636,'19,April,2004','2004-04-19 00:00:00','   The  Special Olympics   has always been a charity I\'ve supported.   In the past few years,  truck drivers  have adopted this charity as their own, donating time as well as money for these wonderful people.  This year the excitement mounts as one trucker has issued a challenge to raise even more money for the Special Olympics.  My very own brother-in-law  , Rusty, has challenged the truckers of America to donate a total of $1,500.00. In return for this he will have his head and beard SHAVED off, and wear a 3-piece suit while driving his truck around the country making his deliveries.   To bid on the privilege to be the person to cut his hair, or just to donate money to this cause, please go  here    If not, (shrug) send the money to me! (no link, just kidding)  '),(1680,3146636,'19,April,2004','2004-04-19 00:00:00','   This pond (created by busy little beavers) is about a mile away from my house.     I go here to relflect, fish and skip flat stones.  '),(1681,3146636,'19,April,2004','2004-04-19 00:00:00','   I have survived the weekend with 1.4 year old twins and their 3 year old sister.  I will now make the t-shirt \" I survived the SMRs\"  Their initials are ALL S.M.R. - that\'s gonna get confusing, later in life.  '),(1682,3146636,'15,April,2004','2004-04-15 00:00:00','   I tinker too much. I try to fix things that are fine the way they are.  I was once told by a very dear old man, that if you were inspired to begin with it doesn\'t need tinkering. If you weren\'t inspired, it\'s not worth tinkering with. He was referring to writing at the time, but I\'ve seen that it works in all aspects of creativity/life.  So why then do I tinker?  My dad told me it was in my genes, so I washed them.  And yet I tinker.  Who needs to know this? No-one in particular.  I have found out things great and small, that have added to the texture of the being that is me.  For instance, changing just 2 ingredients in a long island iced tea will turn it into a kick-ass frou-frou drink that I call a Baja-sunset. It was well received at the No-Frills Grill in Denver/Aurora, sadly shut down now due to poor selling practices. :(  And yet I tinker.   I tinker to learn, I tinker to improve, I tinker to destroy.  It\'s in my genes.  '),(1683,3146636,'14,April,2004','2004-04-14 00:00:00','   just a quick venture into fiddling with a new plug-in for PSP8  '),(1684,3146636,'14,April,2004','2004-04-14 00:00:00','     '),(1685,3146636,'14,April,2004','2004-04-14 00:00:00','   Cilantro, even though part of the carrot family, is evil. The seeds (coriander) are divine in many applications, but the leaves leave a lot to be desired for many.  A friend of mine in Denver has such a bad reaction she has to go to the hospital if she happens to ingest any.  Just why do so many have such a bitter (pun intended) reaction to this herb?  A quote from The Gourmet Sleuth may give some insight:  \"The essential oils of the cilantro leaves contain antibacterial properties and can be used as a fungicide.\"  Personally, I think anything that can be used as a fungicide should not ingested.   just sayin..  '),(1686,3146636,'14,April,2004','2004-04-14 00:00:00','   horrible pic, but the phone doesn\'t have flash ;c)  '),(1687,3146636,'14,April,2004','2004-04-14 00:00:00','   the lawn doesn\'t want to be mowed right now but the tomatoes and lilac bushes want to be planted  plus  I\'m tired of hearing mom say \"Plant these things!\"  '),(1688,3146636,'22,May,2004','2004-05-22 00:00:00','   Do NOT piss me off!  '),(1689,3146636,'13,May,2004','2004-05-13 00:00:00','   Went out with friends and friends of friends tonight, so much fun, I\'d forgotten what it was like!  One of my very best friends in the whorle* gave me this:    *whorle = whole world...  '),(1690,3146636,'01,May,2004','2004-05-01 00:00:00','   Camera wasn\'t working so I popped it in the scanner... played with it and suddenly I was channeling Mr Warhol...  '),(1691,3146636,'01,May,2004','2004-05-01 00:00:00','   A dark pink, heavy on the scent!   Sam bush as above, faded out.   Ramblin\' rose, close up   Ramblin\' rose  '),(1692,3146636,'07,June,2004','2004-06-07 00:00:00','   I\'ll be heading back south and east tomorrow. To make sure I\'m not getting myself into any nasty weather on the way, I\'ll check this  site  before I leave.  Across the midwest in June, yeah I\'m insane. But you already knew that.  Off for now, see you when I get back to lower alabama.  '),(1693,3146636,'31,July,2004','2004-07-31 00:00:00','   urlLink     Happy Birthday Aaron, Love you  MORE  !      urlLink  '),(1694,3146636,'30,July,2004','2004-07-30 00:00:00','   You\'re gonna have crust left over.  Make cinnamon and sugar nibbles!  Roll out extra crust.  Brush with 1 to 2 tbsp melted butter.  Sprinkle cinnamon-sugar mixture over the butter.*  Cut into pieces about 2\" square. (Here\'s where you can let the kids help, with cookie cutters)  Use a large, flexible spatula to move the treats to a cookie sheet.  Bake at 450 for about 10 minutes.  Allow to cool on the baking sheet for a few minutes then remove to a plate.  These can be stored in a cookie jar for later nibbling.  What?  You turned your back, and they\'re all gone?  You didn\'t set the intruder alert, did you?  * cinnamon-sugar mixture  1/2 cup sugar 1 to 2 tbsp cinnamon (I love cinnamon, adjust to your own taste)  '),(1695,3146636,'29,July,2004','2004-07-29 00:00:00','   I love pie.  Especially fruit pie.  Today I made a mango and peach pie.  It\'s divine, with a scoop of good vanilla ice cream.  Mango Peach Pie:  2 mangos cubed (about 1 1/2 to 2 cups) 5 peaches sliced very thin (2 to 3 cups) 1/2 cup sugar 1 tbsp lemon juice 3 tbsp flour (mix with the sugar before adding, trust me)  Mix all ingredients together in a largish bowl, cover and refrigerate while you make the crust.  Yes, I said  MAKE  the crust.  Pie Crust:  3 cups flour 1 1/2 tsp salt 1 1/2 cups cold shortening, lard or butter (refrigerate it, I\'m not kidding!) 3/4 cup ice water  Mix the flour and salt, cut in the shortening until it looks like crumbs. Add 1/2 cup of the ice water and gently stir in. Add additional water 1 tbsp at a time until the dough ball is formed. Don\'t make it too wet, that will make a tough crust!  Form into 2 balls, cover with plastic wrap and refrigerate for about an hour. (You DID put the filling into the fridge, right?)  Set your oven to 450Â°.  Take the dough out of the fridge, one ball at a time.  Knead gently, about 10 times. Roll out on a floured surface to a round big enough for a 10\" pie pan. Place into pie pan gently, easing it down the sides. There should be about an inch overhang outside the pie pan.  Take the filling out of the fridge, stir and pour into the crust.  Now roll out the top, according to the above directions. Fold into quarters and cut vents into the middle.  Wet the edges of the crust in the pan with water, egg white or juices from the filling.  Unfold the top over the filling and trim to 1/2 inch from the outside edge.  Fold the bottom crust up over the top crust and finish by crimping or pressing with a fork all the way around.  Cover edges with foil and place pie on center rack in the pre-heated oven.  Bake for 15 minutes at 450Â°, then turn temperature down to 350Â°.  Cook for an additional 30 to 45 minutes, or until crust is golden.  Remove from oven and protect from others by saying how nasty it smells.  Cool for at least 3 seconds.  Dig in.  Seriously, let it cool a while (30 minutes or so) or the filling will be runny and your pie will look as yucky as you said it smelled earlier.  '),(1696,3146636,'29,July,2004','2004-07-29 00:00:00','   urlLink    A darker view, not sure about this one.&nbsp; urlLink  '),(1697,3146636,'28,July,2004','2004-07-28 00:00:00','   urlLink Apathy   as I see it.    When you don\'t care about something, you\'re said to be  urlLink apathetic .     I don\'t completely agree with this.    I don\'t care what your sister\'s husband\'s cousin\'s fiance had for dinner last Tuesday, does that make me apathetic?    No, it just means I\'m not interested.    I don\'t care how the election turns out,  does that make me apathetic? (hypothetical case)    Yes, it does. It means that I don\'t care how the future of my country (as well as yours) turns out. It means that no matter how hard YOU might try, things may not change for the better or the worse, because you can\'t make ME see my own apathy.    Every person counts. Every fire lit under some apathetic ass helps. Every spark ignited in some boob-tube-dulled eye makes a difference.    But apathy isn\'t just for people who don\'t vote. I see it all the time in people I personally believe in.    Friends who make grand plans, then don\'t work towards the goal, end up whining because it didn\'t \"happen\" to them. Things don\'t just \"happen\" to you. You have to MAKE the event, whatever it may be.     You have to put that one foot in front of the other, put pencil to paper, put oils to canvas, put music to the dreams you had last night, or you\'ll always be wondering why.     Why didn\'t that happen to me?     Help fight apathy today, go DO something that means something to you.   Find something that you turn a blind eye, or deaf ear to and learn about it.  Volunteer, learn something new, taste a different flavor.    DO SOMETHING!  '),(1698,3146636,'28,July,2004','2004-07-28 00:00:00','   urlLink      Ijaha, now 3 months old.  urlLink       She weighs  almost  as much as my  small  baby did (he was 9lbs at birth, she weighed 2.7)  '),(1699,3146636,'28,July,2004','2004-07-28 00:00:00','   urlLink    revision&nbsp; urlLink  '),(1700,3146636,'28,July,2004','2004-07-28 00:00:00','   urlLink    not happy with it yet, but it\'s getting there&nbsp; urlLink  '),(1701,3146636,'28,July,2004','2004-07-28 00:00:00','   urlLink    Buy your tshirt, support caitlinb\'s tour 2004!&nbsp; urlLink  '),(1702,3146636,'27,July,2004','2004-07-27 00:00:00','   urlLink    work in progress: glacial upheave&nbsp; urlLink  '),(1703,3146636,'27,July,2004','2004-07-27 00:00:00','   Sad but true, I\'ve sold out for a pittance.    Well, it is ala-fuckin-bama and I wasn\'t expecting to make 20 bucks an hour.  '),(1704,3146636,'25,July,2004','2004-07-25 00:00:00','   the lawn hasnt been mowed in two weeks  the mower died a temporary death and has now been resurrected  unfortunately  so  im out there mowing  fifteen inch swaths with a thirty inch deck  this sucks  but if i mow a full swath the mower cuts out  one acre takes two hours to mow this way  and  the grass is up to my ass  granted im only five three  but thats tall grass  and thick  and full of bugs  and the trees have spiders in them big enough to carry off a small woman  and the fireant mounds are tall enough that  when i run over them  they get on my feet  that hurts  the toes on my right foot are numb from the bites  '),(1705,3146636,'23,July,2004','2004-07-23 00:00:00','   nothing else to report  '),(1706,3146636,'23,July,2004','2004-07-23 00:00:00','   Last night I had the dubious pleasure of stocking the cookie aisle. What is UP with this? I counted no less than 30 brands of cookies, not including the varieties!              Why are you so willing to eat these sugary pieces of cardboard rather than make your own? It canâ€™t really be the time factor, some of the most complicated cookies I make take less than 2 hours.       2 hours folks, please! You spend more time than that watching the tube in a stupor after eating your fast food.              Nothing conveys \'love\' (or \'like\') like freshly baked cookies.     When my sons visit, the first thing they want to know is \"Where is the cookie dough?â€    Grandnieces (I seem to have a million) all go straight to the cookie jar.    Granddaughter is a bit young yet, but she knows, she  knows ...    My personal favorite is a banana-walnut-oatmeal-raisin that I created one day when everyone at work    was asking for \'something new\'.     The boss (major yum factor of 37) liked them so much; I had to make them at least twice a month, just for him! Of course, he did share.    In this chapter, however we will explore the universal favorite, chocolate chip.    Now, I don\'t like crisp or hard cookies.  I like them soft and/or chewy, so I\'ve adjusted recipes to make them  my own , and to make my cookies soft. To me, soft makes them seem fresher somehow.          Now for the recipe;               5 cups all-purpose flour      2 teaspoons baking soda      2 cups butter, softened      1 1/2 cups packed brown sugar      1/2 cup white sugar        4 eggs      2 teaspoons vanilla extract      4 cups semisweet chocolate chips      2 cups chopped walnuts (not an option in my mind)               Sift together the flour and baking soda, set aside.      Cream together the butter and sugars in a large bowl.      Beat in the baking soda until blended.      Stir in the eggs and vanilla.      Mix in the flour.      Finally, stir in the chocolate chips and nuts.      Drop cookies by rounded spoonfuls onto ungreased cookie sheets.     Bake at 375Â° for 10 to 12 minutes.      Edges should be golden brown.     Remove from pan and set to cool on a rack.              Set the guard dogs up around the table.                Make SURE you have plenty of milk on hand, as these will be eaten as fast as your back is turned, even while youâ€™re putting the next pan into the oven.               Thatâ€™s all for today.  '),(1707,3146636,'22,July,2004','2004-07-22 00:00:00','   First, we\'re not using any convenience stuff here. You\'re gonna make the teriyaki sauce yourself, no little bottles of pre-made marinade from the grocery.    Remember, this is about the satisfaction of the end product and the bits of yourself that you put into the meal. It\'s not about getting the meal ready fast because you have at LEAST a dozen other things to get done before you plop down in front of the tube.      Teriyaki sauce:    1/2 cup soy sauce  1/2 cup mirin (sweet rice wine)  2 tblsp sugar  (or to taste, yes TASTE it as you\'re making it, otherwise you\'re not gonna know if it\'s worthy to serve to yourself)    Mix all above in a saucepan and warm just to bubbling on the stove. (Stove: that thing in your kitchen that you leave the plates from frozen microwaves meals on until you take the garbage out)    Now you have the sauce.    Time for the chicken.    2 lbs chicken breast, preferably boneless    You can cut the breasts into strips or bite size pieces if you like. I usually cut them into 4 pieces, so you have to cut them while eating.     This forces you to eat slower and enjoy your food. Really, it does.    Dip the chicken pieces into the sauce and place in asingle layer in a buttered baking pan. Pour the rest of the sauce over  the chicken and put into a 350Â° oven (that thing behind the door on your stove, you remember the stove, right?) for 30 to 45 minutes.    After 30 minutes, baste the chicken every 5 minutes until done. Turning it over doesn\'t hurt either.    Done!     Serve with fresh bread, a nice salad and maybe some kind of mango-chili salsa* or fresh fruit of your choice.    Now, don\'t you feel better about the meal you just created? Honestly, 45 minutes is worth it.    Trust me on this.    *mango-chili salsa can be as simple as combining fresh mango cubes with the chili pepper of your choice, roasted chilis impart a yum factor of at least 7 to this.    1 cup mango cubes, bits and pieces (mangos are  so  much fun to remove from the seed)  1/4 cup roasted, peeled and seeded chili pepper (of your choice, although I recommend poblano or anaheim for those not attuned to the need for heat)    Mix the mango and chili together in a bowl. That simple.    You can complicate it a bit by adding some sweet onion, a bit of garlic, whatever.  '),(1708,3146636,'21,July,2004','2004-07-21 00:00:00','   Everyone must eat, this is a given.    Man (human, not just the guys) cannot live on bread alone. Regardless of current dietetic fads, fresh bread is a major yum factor.    Baking bread takes time.    You say you don\'t  have  time?    Get over it. Take time. Make time.    The very ritual of making and baking bread is, in itself a stress reducer.     Choosing the recipe.    The gathering of the ingredients.     The melding of these ingredients to the mass of lifeless dough, so soon to be a loaf, a batch of rolls, a breakfast delight.    The waiting time. For the dough to rise takes that precious commodity, time.    Hey, you probably need to vacuum and clean the bathroom anyway, right? (Please wash your hands after.)    Once the first rising is done, you get to punch the dough down! Fun times!    Kneading is very good for your stress, I suggest making Italian bread if you\'re especially stressed as too much kneading will make for a tougher crust.    Leave the cinnamon rolls for non-stress-filled times, trust me on this.    On that:    My recipe for cinnamon rolls (Doesn\'t include a trip to the mall)    First gather your ingredients:    2 pkgs of fast-rising yeast  1 1/2 cups milk, heated to 110 degrees  1/4 cup salted butter (don\'t substitute margarine, please)  4 tblsp sugar  1 egg room temperature, beaten slightly  5 cups flour, sifted (or not, most flour is pre-sifted anyway)    Melt the butter in the hot milk.  Mix together 4 cups of the flour with the yeast and sugar,  Stir in the milk/butter mixture and the egg.    Add additional flour until you have a soft, but not sticky dough.    Turn the dough out onto a floured surface and knead for 5 minutes, adding flour as you go from the remaining cup, or until the dough is smooth and no longer sticky.    Cover and let let rise until doubled in size.     Punch the dough down, literally, punch it hard to let out the built up gasses.    Knead for a minute or so then form into a ball and roll out into a rectangle, approximately 9x13.    Filling:    1 stick butter, very soft, but not melted  1/2 cup sugar   2 tbsps ground cinnamon.    Spread the butter over the rolled out dough to the edges.    Mix together cinnamon and sugar and sprinkle heavily all over.    Don\'t worry too much about getting it perfectly even. Or at all even.    Roll the dough up from the long side, pinching the edges together to seal.    Cut into 1 inch sections with a VERY sharp knife.    Place rolls into a 9x13 well buttered pan, about an inch apart.    Make sure there\'s room for the rolls to rise and spread out.     Cover and let rise until double then bake at 400 degrees for 15 to 20 minutes or until golden brown.    Remove from pan and drip a glaze of 1/2 cup confectioner\'s sugar and  a tablespoon or two, more or less of milk, over the tops if desired.    You can also substitute orange juice for the milk in the glaze. (YUM factor!)    Feast!    Variation:     Melt 1 stick butter and add 3/4 cup packed brown sugar, plus 1/4 cup REAL maple syrup for maple sticky buns.     Pour into the bottom of the pan before placing rolls in to rise.    When finished baking, place a baking sheet ( cookie sheet works well for this) over the pan and turn it over.     The rolls and mapley-brown sugary goop will NOT all come out, so you may need a helper to clean out the pan. Kids work well for this task, but men seem to like all that finger-licking too.    Again, Feast!  '),(1709,3146636,'20,July,2004','2004-07-20 00:00:00','   got my lawnmower back    I fought the lawn    and the lawn won  '),(1710,3146636,'20,July,2004','2004-07-20 00:00:00','   1.   Uninterrupted     sleep.  '),(1711,3146636,'20,July,2004','2004-07-20 00:00:00','   As I work at my job, in a semi-major southern grocery store, I am amazed.          Amazed the choices.          Amazed that the choices are there.            Amazed that the American public has become so fucking lazy, they can\'t combine ingredients other than water and \'this bag of stuff\' to make a meal anymore.        I mean, really.         Whatever  happened to COOKING?         You know, simple stuff.         Red beans and rice, for example (this is what started me on this line of thinking).            Now don\'t get me wrong, that certain brand of rb&r from Louisiana is a great quick meal, combined with some yummy         urlLink  andouille  sausage, however they now offer a variety - with cheese?       Please.        How FUCKING hard is it to add some cheese to the mixture just before it\'s done?        How have we Americans become so lazy?        Another item: Rice Krispie treats.        I\'m sorry, it takes 10 minutes, TOPS to make this silly sweet. HOW LAZY ARE YOU?       See, to me cooking is a ceremony.      I equate it with love.      To me, cooking is love.      Love of the meal you\'re preparing.      Love of the people you\'re preparing it for.      Love of the individual ingredients.      Love of self.      I know the world is a fast-paced place these days, but you can\'t live on these convenience foods and be even remotely healthy.     You can\'t convince me that a meal made up of say, precooked meat, cook-it-in-a-minute rice/noodles and some canned veggies proves that you love yourself or anyone else you may be cooking it for!      more on this later...      I hear my bed calling.      Who\'d a thunk a part time job in a grocery store would light a fire under one?  '),(1712,3146636,'20,July,2004','2004-07-20 00:00:00','   urlLink    side, short&nbsp; urlLink  '),(1713,3146636,'19,July,2004','2004-07-19 00:00:00','   urlLink    bangs&nbsp; urlLink  '),(1714,3146636,'19,July,2004','2004-07-19 00:00:00','   still writing in the style of archy    my favorite cock    roach  '),(1715,3146636,'19,July,2004','2004-07-19 00:00:00','   i just turned down a job offer from the post office    they wanted me to work one day a week and be on call    in case of emergency    i can\'t do that and the job i currently have    why is it so complicated  '),(1716,3146636,'19,July,2004','2004-07-19 00:00:00','   after just 2 nights of work i have come to the conclusion that   americans are lazy and they get bored too easily    not a brilliant insight i know     but    how many different flavors of popcorn do you need  whats wrong with the original flavor  i do not like microwave popcorn  i prefer to cook it the old fashioned way  its the ceremony that makes it taste good  that  and real butter  oh  and cooking it in coconut oil  mmmmm  cholesterol be damned    why are there 37 different types or flavors of spaghetti sauce  what happened to making it yourself  or adding your own ingredients    just how many shapes can pasta come in to entice your child  who will only eat mac and cheese anyway  to eat it    after that  and having a horribe lunch from krystal  the greasy filmy feeling will take weeks to leave my mouth  i just know it    i have decided  to pack my lunch from now on      also  i weighed myself  and i weigh 10 pounds more than i thought  not happy    also  im cutting my hair off  1 to 2 inches should be long enough to be socially acceptable  and it wont be all dripping with sweat all night    well    thats about it for now  oh wait    my feet and back dont hurt as much  i wore my comfy sneakers instead of the not comfy black shoes  i will buy comfy black sneakers after i get my frst check    ok    thats it    bubye  '),(1717,3146636,'19,July,2004','2004-07-19 00:00:00','   its taking me some time to get used to being fully nocturnal again  but im sure i can get there within a week    or three  '),(1718,3146636,'18,July,2004','2004-07-18 00:00:00','   Ronald M. Whitcomb, Jr.  passed away    died    kicked the bucket      threw in the towel      started leaving me alone  '),(1719,3146636,'17,July,2004','2004-07-17 00:00:00','   the jumble of my life 		 		 		 			 urlLink     			 				 			 urlLink city life bw 			 			Originally uploaded by 			 urlLink kamylyon .  '),(1720,3146636,'17,July,2004','2004-07-17 00:00:00','   long first night at work....    i haven\'t stood on my feet that long in one day since 1987, when i left my last retail job for the wonder-filled world of computers and the internet.    i\'m too exhausted to sleep, and man do my feet and back hurt!    on the upside, it\'s a lot of bending and stretching, which will, in the long run, be very good for me, yes?    later, message, sleep.....    zzzzzzzzzzzzzzzzzzz  '),(1721,3146636,'17,July,2004','2004-07-17 00:00:00','   one of my favorite poets has always been  urlLink archy  a cockroach who wrote a poem each night on  urlLink Don Marquis\'  typewriter.  archy was too short to use the shift key so everything was lower case and unpunctuated  i will be writing this way for a while to see if i can stand to be a cockroach in my next life  words to live by  live so that you can stick out your tongue at the insurance doctor  archy  '),(1722,3146636,'16,July,2004','2004-07-16 00:00:00','   No, not THAT night (more\'s the pity).  Tonight I get to go back to that thing people refer to as \'work\'.  Now don\'t get me wrong here, but I\'ve been looking for employment for close to a year now, and suddenly I have not 1 but 2 jobs, within a week.   WTF?  Did my southern accent improve that much?  '),(1723,3146636,'16,July,2004','2004-07-16 00:00:00','   urlLink    Who do these lips belong to?&nbsp; urlLink  '),(1724,3146636,'15,July,2004','2004-07-15 00:00:00','   and counting  '),(1725,3146636,'15,July,2004','2004-07-15 00:00:00','   Before the storms hit this evening I happened to be in town and saw this cloud.  '),(1726,3146636,'15,July,2004','2004-07-15 00:00:00','   Hasn\'t it already been a week?  Beer tastes bad without cigarettes. This means I will not be drinking much beer for a while.   Much better for me.  I can do this.  Yes I can!  '),(1727,3146636,'15,July,2004','2004-07-15 00:00:00','   6+ hours into it and I\'m not panicking toooo much, I might make it through 24...  '),(1728,3146636,'14,July,2004','2004-07-14 00:00:00','   Doesn\'t one have to have some sort of definition first?  Maybe if I define myself as a needy, clingy ditz, I\'d get more attention.  I\'d shoot myself first.   Granted, I\'m clingy, I need, and I can ditz with the best of them, but not on a regular basis.    1st step of definition taken: last cigarette smoke 3 hours ago.  '),(1729,3146636,'14,July,2004','2004-07-14 00:00:00','   2 packs and it\'s done, no more smoking for me.  I finally found something (again)that\'s more important.  More on that later.  For now,  wish me luck!   Thank goodness I like  urlLink cold turkey...  '),(1730,3146636,'13,July,2004','2004-07-13 00:00:00','   512 MB of RAM If found, please send home  :(  '),(1731,3146636,'12,July,2004','2004-07-12 00:00:00','   Sometimes I just don\'t wanna know.  There are things I can\'t say out loud, for fear of hurting someone else. The things I don\'t understand, and things I don\'t wanna understand. The things I do understand, but don\'t know why. Things that are beyond understanding are placed at my feet.   On silver platters.   \"Here, fix this.\"  No  please  or  thank you , just  do  it.  Somewhere, at some point in time, I was \'defined\' as the listener. Is there a listener out there for me?  '),(1732,3146636,'11,July,2004','2004-07-11 00:00:00','   My mess  '),(1733,3146636,'10,July,2004','2004-07-10 00:00:00','   I was mowing the lawn, and noticed that I needed air in the tires. So I get out my handy-dandy portable air compressor and as I\'m filling the rear tire I notice this noxious smell.   Looking under the gas tank, I see not a drip, but a  STREAM  of gasoline leaking out.   Needless to say, I did  NOT  start it up again.  I\'m sure it can\'t take a full   2 weeks   to fix it....  '),(1734,3146636,'10,July,2004','2004-07-10 00:00:00','   Muses come and go for most. The lucky few are able to entice them to stay around.   Lately  urlLink Euterpe , who has been my muse from birth - according to my dear father, has decided to revisit.  Now legend has it (well this is MY legend)that I was born singing. As far back as I can remember I have heard music. Not on a radio, or a record player but in my head.   It may well be that I trilled out a little tune when I slide down the chute of delivery, after all who wouldn\'t rejoice at being freed from the confines?  All I know is, there is ALWAYS a tune of some sort in my head and sometimes, if they don\'t already belong to someone else, I can even write them down.  The most likely place for me to \'find\' a new tune is in someone else\'s writings. My own words hold little in the way of music for me. I don\'t know why, that\'s just the way it is.  But give me a nice bit of poetry and the words start dancing in my head, jotting the notes down as my eyes skim over them on the page.   These notes will stay stuck there until I either get them written down, or they die from lack of attention.  I mourn the notes lost, dissipated into the mist of my mind.  I rejoice in the notes I am able to save to paper, or these days, to computer.  I thank you muse, for coming back to me!  '),(1735,3146636,'09,July,2004','2004-07-09 00:00:00','   Small, but gorgeous!   Close-up :)  I know, it\'s called a Christmas Cactus, but it\'s a whole lot closer to  urlLink Bastille Day.    I got this plant from a former boss, who felt guilty for not being at work when I was taken from there into the hospital due to \'aggravated stress\'.  '),(1736,3146636,'08,July,2004','2004-07-08 00:00:00','   It hurts my fingers to type this.  I haven\'t played in 8 years (or thereabouts) and had to restring it after having retrieved my dear, beloved guitar from a sister. In so doing I discovered an almost stripped tuning peg :(  That\'s gonna have to be fixed.  It\'s playable, to a degree, but not fully tuneable. Right now it\'s about a note and half off. I can\'t live with that.  Note to self: Never, ever, ever, ever notinamillionyears lend ANYTHING to ANYONE EVER again.   And so, my life, such as it is.  My fingers hurt.  '),(1737,3146636,'02,July,2004','2004-07-02 00:00:00','   when you\'re surrounded by people who can call you by name why do you feel so alone?  when it\'s eating you up and your stomach\'s in knots and they all want you  now  and they take and they take...  who\'s there to give when  you  need it?  '),(1738,3146636,'06,August,2004','2004-08-06 00:00:00','   urlLink          urlLink sea dreams at desert sunset      Originally uploaded by  urlLink kamylyon .'),(1739,3146636,'05,August,2004','2004-08-05 00:00:00','   urlLink    touched up in psp&nbsp; urlLink  '),(1740,3146636,'05,August,2004','2004-08-05 00:00:00','   urlLink    sketch&nbsp; urlLink  '),(1741,3146636,'05,August,2004','2004-08-05 00:00:00','   There are few things that I absolutely hate.  urlLink Fireants  top the list, now that my late  ex-husband is, um, late.  They\'ve invaded the house now, so sleeping is less than optimal.  They seem to think I am a delicacy. Granted, I\'m the sweetest person on earth, now that Mother Teresa has gone to the Great Beyond, but really, WHY ME!?!?!  They sting, and it hurts. They get infected in a matter of minutes. They leave scars. Did I mention, they  HURT !!!?  Time to bomb the house inside and out and hope I get all the queens...  '),(1742,3146636,'04,August,2004','2004-08-04 00:00:00','   urlLink    c\'est fini!&nbsp; urlLink  '),(1743,3146636,'04,August,2004','2004-08-04 00:00:00','   urlLink    rainbows in a western sky&nbsp; urlLink  '),(1744,3146636,'03,August,2004','2004-08-03 00:00:00','   They suck.  Having a cold in the winter is expected. You  PLAN  for winter colds.  You have plenty of medicinals on hand, having planned.  Summertime colds hit you like a ton of bricks and leave you feeling like the bricks got lodged in your brain as well as various other spots of your poor, aching body.  I have a summer cold.  I\'m miserable.  I ache.  At least I\'m not puking anymore.  Anyone know why summer colds make you puke?  ::sneezes 7 times in a row::  OK, back to the steam and vicks.  I hate summer colds.  '),(1745,3146636,'02,August,2004','2004-08-02 00:00:00','   urlLink    still working on this one, layers make the work much more fun&nbsp; urlLink  '),(1746,3146636,'01,August,2004','2004-08-01 00:00:00','   ... was my baby\'s 17th birthday.  My baby. 17. Impossible!  He was just 3 feet tall a couple of years ago, wasn\'t he?  Now he\'s close to 6 feet tall and can grow a full beard.  How did this happen?  I know. I understand the inexorable march of time and all that.  But really.  Wasn\'t it just a couple of years ago, that I made his  devil  dinosaur   costume for Halloween?  The year after that was an Indian costume and the advent of the  stupid barbie-like doll  in the spider web on the porch. (He and his older brother borrowed it from a little girl down the street, only Barbie I ever had to buy was to replace the one they  borrowed )  That march of time always seems to creep up on me. I just don\'t know how I got to be 48!  How the hell did I survive?  Better yet, why?  Somewhere out there, there\'s a reason.  And so I trudge along in the march ( urlLink John Phillip Sousa  playing in the background), dazed and confused. Lonely at times. Once in a while, scared.  That I\'ll never find that reason. That one person. That other half. That someone to grow old with. Laugh with. Cry with.  Most times I\'m apathetic about it, but once in a while...    I\'m scared.  '),(1747,3785626,'28,June,2004','2004-06-28 00:00:00','    This is lovely. A new blog, a new life, just \"a new.\" Well today was extremely boring. My brother wouldn\'t take me to Starkville so I couldn\'t get my Blue Converse. Darn him. I went to my Aunt\'s house for supper. She can cook good. I  NO . Because that\'s my friend, and plus I go with Tregory. So yeeeah, that will just make the tension worse. Oh well. Anyways that\'s my day. I will try to update alot. Later. <3 Cherry '),(1748,3785626,'29,June,2004','2004-06-29 00:00:00','    What an interesting day today. I woke up around 10 to get ready for my dentist appointment at 11:30. Yes, it takes me an hour to get ready. Haha. Tregory called, YAY, I didn\'t get to talk to him yesterday. Poopy. But anyways...I have no cavities, BUT I have to get braces so I can have space for my wisdom teeth. I don\'t want braces though. :( MAYBE I can talk my dad into getting those invisiline braces. :D Happy Me. Anyways we ate at Bumper\'s. Yummay. Then I mowed the front yard, took a shower, and I was going to get my haircut but they closed. My dad said that alot of stores close on Tuesday afternoon. *rolls eyes* STUPID HICK TOWN. Anyways now I think I\'m going to take a nap. I\'ll mow the rest of the grass tomorrow maybe. Yeeah. Anyways holla. Love, Cherry '),(1749,3785626,'02,July,2004','2004-07-02 00:00:00','    Woot Woot. Okay lets sees here, I woke up at 9:30, from a really bad dream, which sucked. Then I mowed the grass which sucked even harder. Then I talked to Tregory for a while, then I took a shower, which makes every thing better. What a shower can do to yaz? Anyways later today I\'m going to the lake and watch the perty fireworks go Boom! I hope it doesn\'t rain, but then again it might be fun for me...to run around in the rain. Teeheehee...everybody else will be runnin\' for their lives and I\'ll just take my sweet lil time. So anyways I think I might wanna be a counselor or something when I grow up, or a psychiatrist, cause lots of people come to me with their problems or I just like to give advice...Teeheehee...But I might wanna be a neurologist and work on the brain. Wooohooo...Lol. Well I\'ll be bloggin\' later tonight. Teeheehee..BLOG is a cool word. Blog Blog Blog. Okay yeah I\'m hyper. Laters...Cherry<3 '),(1750,3785626,'02,July,2004','2004-07-02 00:00:00','    Today was good. I woke up around 10:00, Tregory came over at around 12, left about 3. Then dad had to go to Grandmother\'s so I had to watch mom. Dad came back at around 4:30, Chris took me to Starkville. Shoe Dept. didn\'t have my Chuck\'s so I\'m probably going to buy them off the internet. Should have checked at Hibbet\'s though. Then we went to JC Penney, I bought 2 shirts and 1 for my mommy. Then we went to Wally World. Got me a SpongeBob Rocker notebook and some Sharpies and some Candy. Wheeee...Then we went to China Garden and got food. This was my first time ever to eat with chopsticks. And I\'m Asian!!! Oh well, it was fun...Tomorrow is the Fireworks Over The Water at Choctaw Lake. I\'m going and gonna have fun I guess. Lol. Spuddy came over for a few to chill, he told me what Javors and Mary told him to do...>Try to go with me instead of Amber, because of what happened between him and Tregory. He said he wasn\'t gonna do it, because that is disrespectful and all. He thought I was going to be mad, I would only be mad if he would have tried to do it. Plus, I\'m helping him and Amber \"talk\"...So yeeah, he wouldn\'t wanna upset me. Lol. Anyways Tregory left his pullover at my house, so now I\'m wearing it. Teeheehee...*sniff sniff* :D  Well layder people. Hyper! Hyper! (Off Skittles n Gummi Bears) <3 Cherry <3\'s Tregory <3 '),(1751,3785626,'03,July,2004','2004-07-03 00:00:00','    Today was too boring. Woke up the lovely sound of an egg beater. My dad was making a cake for my mom\'s birthday tomorrow. JULY 4th. I did nothing, then about noon I went to the store, got some stuff. I saw Claire. Then I went home and did nothing. Called Tregory about 3. Wheee! Fun. Then I did nothing. Now I\'m doing nothing. Tregory called a few minutes ago. I\'m so Happee!! Lol He got to go shopping. Lucky him...Anyways I think I\'m going to stay up ALL night. I\'m bored. I have nothing better to do. I don\'t wanna go to sleep really. I might clean the whole frikkin\' house, who knows. Anyways I\'ll holla. <3 Cherry//Boredom Has Struck '),(1752,3785626,'03,July,2004','2004-07-03 00:00:00','    Things That Make Me Happy...No Particular Order Just As They Come To Mind   .God.n.Jesus. .Sharpie Markers. .Duct Tape. .Talking On The Phone. .Music. .Food. .Friends. .Family. .Jervon Tregory. .Good Dreams. .CKY. .Driving. .Blogging. .Peace. .Love. .New Things. .Chuck Taylor\'s. .Smelly Good Stuff. .Getting Mail.  I\'m bored as you can tell...TEEHEEHEE '),(1753,3785626,'03,July,2004','2004-07-03 00:00:00','    Okimidokies...Tregory came over and we chilled. Then about 6 me, Tregory, and my brother went to Choctaw Lake for Fireworks Over The Water, which really doesn\'t start till like 9:30, but they tell you to get there early. So we get there, nothing to do but walk around and eat. Tregory\'s cousins weren\'t there so he had to spend all day with me. Lucky Him. Teeheehee. Anyways then about 9:40 the fireworks started. It was very cool especially over the water, and with Tregory <3Cherry<3\'sTregory<3 '),(1754,3785626,'04,July,2004','2004-07-04 00:00:00','    My life is being sucked into the blackhole of boredom. I don\'t think last year was as bad as this maybe it\'s because I\'m in Ackerman. The only person I hang out with is Tregory, or Spuddy or Mary if they come over. It\'s kind of sad, like 8th grade summer was off the hizzy forshizzy, I guess because I was such good friends with Krista and she lived like 5 minutes away. And I went to Oneighty alot so I could see my buddies. And now I move to this redneck hick of a town and I get excited when I go to Wal*Mart. *sigh...* Nobody gets online anymore, I guess it\'s because they actually have LIVES TO LIVE!!! I have like this insignificant little speck of what you call a life. I wish I had church to go to...I can\'t wait till I get to drive by myself to church. That\'ll be cool. I have a boring life. A crazy one but boring. The fact that my mom and grandmother have Alzheimer\'s disease doesn\'t really help. Adds lots of drama, but doesn\'t help. I\'m trying to work on an Alzheimer\'s Disease website for mostly teens but I just haven\'t felt up to it. Maybe I\'ll work on it later tonight. My grandmother is getting worse, but at least she is old enough for the regular stereotypical alzheimer\'s patient, unlike momma. I really wish she didn\'t have it...Just irks me to see her, because I know she isn\'t the same. Enough about that...I think I will work on that site. Later Peeps. <3 Cherry '),(1755,3785626,'05,July,2004','2004-07-05 00:00:00','    Okay so last night I didn\'t exactly stay up...I went to bed at around 2. So I was more tired than I thought I was. I can\'t help that. Anyways I have nothing to do today. I wish Tregory would come over, I have nothing to do. I\'m so bored...I\'m playing video games. Which I haven\'t done for like a long long long time. Hahaha. So anyways I\'m really bored....I\'ll be bloggin\' laterz. Peace out. Cherry '),(1756,3785626,'05,July,2004','2004-07-05 00:00:00','    Okay lets see...last nite I really did stay up till about 7:30, then I took a sweet little nap, really by accident. Woke back up to the soothing sound of the doorbell, guess who? Aunt Faye, Uncle Joe, and The Triplets. The triple torment. They love to get into everything, ask everything, remember everything, and just be annoying. I love them to death but they are just annoying. All in my business. Lol. My room is really for teenagers, I don\'t think its kids approved, cause I have like stuff that they shouldn\'t see until they understand. Teeheehee. Like I have this picture of this dude from a magazine, CosmoG!rl, that my friend gave me and he has his shirt off, and they were like why do you have that? I\'m just thinking, ONE DAY you will start liking guys. I don\'t remember a day I didn\'t like guys. In kindergarten I had a major crush on um...somebody. I think it was because I watched soap operas...anyways then after they left I went to sleep. Woke back up for cake. Happy Birthday Mommy! Then called Tregory Holla Cherry...Aka Night Owl '),(1757,3785626,'09,July,2004','2004-07-09 00:00:00','    So today I woke up groggily at around 10:30. I cleaned my room a little bit. Um..I mostly stayed on the computer. You know what bugs me? RACISM. STEREOTYPES. and IGNORANT PEOPLE. Those are the 3 things that if we got rid of, the world would be a better place. Anyways I got to call Tregory at like 12 or something, I don\'t know why. Anyways I found this nice little poem site. www.creative-poems.com Kristian told me about it. I\'m glad she is finally online, now I can talk to her. She is one of my many best friends. Anyways maybe if I mow the grass, and stuff, maybe my brother just might take me shopping. I want school clothes. NOW. Poopay. <3Cherry '),(1758,3785626,'09,July,2004','2004-07-09 00:00:00','    Okay so I haven\'t written in this blog because I thought it was screwed. It would show up my other new posts so Oh well. I\'m testing. Testing 1-2-3...Testing 3-2-1... CAN YOU SEE ME!!/!?!?? '),(1759,3785626,'10,July,2004','2004-07-10 00:00:00','    Okay so I\'m think of baby names. So soon you think? Heck no, I\'m not having kids right now, maybe when I get OLDER. Not grown up, OLDER. I\'m just thinking. I want a boy and a girl. So that way I have both. Lol  Girl Name-Paris, Chevelle, Alexandria Boy Name-Devin, Devon, Devontae  Tregory came up with Paris. :) Don\'t think it\'s like Paris Hilton or anything. I like the name Chevelle even if it is a car. People name their child\'s Alexis or Mercedes. So why can\'t I? I\'ve always loved the name Alexandria.  I like Devin and Devon, I think I heard if off of That\'s So Raven. Not sure. I like Devontae because this cute little boy in my Karate class was named that. And he was just so cute even though he was much younger and shorter than me. He was just SO cute. So were his little brothers...Awww  Anyways I just got off the phone with Tregory.   Laters...Cherry...Happy...Girl... '),(1760,3785626,'10,July,2004','2004-07-10 00:00:00','    Something that cures boredom...Chocolate Chip Ice Cream, Usher: Confessions CD, and Dr. Mario. Isn\'t that just dandy? Anyways I didn\'t flat NOTHING today...Isn\'t that just the it? I woke up around 11 o\'clock cause I really didn\'t want to mow the grass, but I should have. I wanted to go to Starkville but my brother likes being a butt and would not take me. Maybe he\'ll take me tomorrow?! Hopefully...Anyways...life is pretty boring for me. So um..guess that\'s it. Hollarz. <3Cherry<3Tregory<3 '),(1761,3785626,'12,July,2004','2004-07-12 00:00:00','    Tregory wrote me a poem it is so sweet. I \'bout cried but I didn\'t because I don\'t cry really when I\'m happy. I cry when I\'m sad or really po-ed. Anyways I thought I would post it.   \"Remember\"  I remember our first kiss And how much I miss it Remember when I first held your hand You knew I was your man (I mean boy)    Remember me in your heart And lets never fall apart Remember how much I love you And that you love me too Remember that we\'re not through Not now or not ever Remember me forever Remember when you sleep My love is what you keep Remember when you lay down at night That I held you tight Tightly in my arms To keep you warm Just remember I care To lie to you is a dare Will you just remember that you\'ll forever be my boo If not just remember I love you... '),(1762,3785626,'12,July,2004','2004-07-12 00:00:00','    Well today I really thought was going to suck, but turned out kind of okay. I woke up at 6 because I had to watch my mom, while my dad and Chris went to Holmes to help register for his college schedule and what not. I\'m stuck at home making sure mom doesn\'t wander off or get too upset or whatever. Anyways I mowed the grass too, which wasn\'t that hard. It was at like 8:30 in the morning. So anywho, I took a shower and then called Aunt Betty because Mom was getting a little too upset. So she came over, we went by her house got some stuff, went by the store and got some stuff. The bagboy tried to hit on me. I walked into the store and he goes \"Hey\" or whatever so I do the whole \"Whaddup nod\" thing because I have no clue who the heck he is...I don\'t talk to strangers. Anyways walking around the store I see Beauty (Tregory\'s lil niece) and his sister. Beauty is so cute and loves Tregory. Anyways I was walking out of the store and the dude goes,\"Hey um...I was wondering...\" And see I thought he knew my brother or something...then all of a sudden...\"Would you mind if I get your number?\" I said, \"Uhhh...YEEEAH!?\" And I thought I think Tregory would mind too....Anyways we got home made lunch and dinner, then Tregory came over about 11. Then Aunt Betty left around 11:45, and I told Tregory to leave but come back. Lol, so he did. Yay. Now I\'m really tired. Oh well. This day was fun because I got out of the house. So laters. Hoola instead of Holla. '),(1763,3785626,'14,July,2004','2004-07-14 00:00:00','    So um...today was like any other boring day. Tregory came over, we played around. NOT LIKE THAT...we kept throwing those sticky gumball things at each other. Then after awhile we just talked. I really love spending time with him, it just makes me feel so HAPPY. Then after he left I went to sleep. He called around 4:30 or so. Then after awhile Kendra called. Then I ate. Then um..that was like it. Lol. I can\'t go to Jackson which sucks. I\'m thinking if I really want to go I could ask Jamie if she wants me over or something. Anywho, hopefully Friday I can go to see I, Robot. Saturday I get to shop till I drop. I just hate going shopping with my brother cause he follows me around and tells me what I am buying is slutty. *rolls eyes* ANYWAYS I might buy some shorts for Band Camp cause it\'ll be hot and all. So yeeah. I might have to get me some real cheap tennis shoes for that, cause I don\'t wanna mess up my Chuck\'s ya know what I\'m sayn. Go to Wally World or summin\'. It don\'t matter. Anywho, I\'m bored. Later. <3Cherry<3sTregory<3 '),(1764,3785626,'14,July,2004','2004-07-14 00:00:00','    Okidokies, so today was fun. Most interesting...Tregory came over and stuff. Then I went to Fred\'s to get some stuff. I had 15 dollars and left with 3. Strange? I bought page \'tectors for my binder. I think I need a new binder now. I printed out all the lyrics to all the Meteora songs by Linkin Park. :) One happy camper. I haven\'t heard from Sharla yet, so I don\'t know if I\'ll be able to go to Jackson or not. I really don\'t feel up to it, but I really want to see Sharla before she leaves. She is like my sister, and I\'d also see Kristian, and that is one of my sisters for real. I only seen \'em once this summer. Anyways, I\'m so ready for school to start. I don\'t know if I\'m going to act all crazy like I did at Whitten on my 2nd year, but who knows. Alot has to do with the people that go to that school. I don\'t like \'em. Ashley, Shirreka, Tregory, Adriano, Sam, and maybe some others are my friends. Salaste is moving, which kind of sucks. Well I\'m now making pretty little band pictures. I made H.I.M. and CKY, now I\'m going to do Linkin Park and a few others. Hm...who knows. Well laters. Cherry '),(1765,3785626,'16,July,2004','2004-07-16 00:00:00','    Today was wicked.&nbsp; I woke up early to mow the grass to make a little more money. Then I played on the computer and played Dr. Mario on SNES(Old Skool Baby) And then, Tregory called and then it stormed, and so the power went out for like 10 seconds. Then he called back. Um...then I went to bed.&nbsp; Then afterwhile he came over I was asleep. Lol, Then anyways we went to see I, Robot, which is a really good movie. I say A-. 4 1/2 Stars...Anyways that\'s it.&nbsp; I\'m bored. I\'m tired. I\'m going shopping!!! Later. Love Cherry '),(1766,3785626,'16,July,2004','2004-07-16 00:00:00','    So yesterday, which really feels like today for me, was okay. Tregory didn\'t come over, which kinda sucked, but I slept late. He called me though around 12:15 and we talked till about 3 something.&nbsp; So that made up for not seeing him. I got a letter from band, band camp starts August 2. Which really sucks, but I\'ll be okay. Anyways I get to see I, Robot. Yayness. I\'m going to wake up early tomorrow to mow the front yard, so I can get some money. Then Saturday I\'m going shopping. I wonder if we can eat lunch there. I want to go to China Garden. Eating with chopsticks rocks my socks. It\'s fun LOL.        Anyways I\'ve been thinking about this and this is my opinion. I think we make guys cheat. Seriously, this is what you see sometimes. A guy that has a girlfriend, then another girl comes along and wants to hook up with this guy. Then the guy is like I have a girlfriend, and then the girl says, So? What\'s that got to do? I just think that is plain nutty. Cause I think every girl wants a trustworthy boyfriend and honest one, so when you do that, you screw with their train of thought. Guys are simple yet complex. Anyways if you make that dude cheat on his girlfriend with you, what makes you think he won\'t just cheat on you to? Seriously now. I\'ve seen this posted on gurl.com message boards, and it just makes me sick. I know Tregory won\'t do anything to hurt me or meant to at least. (You know guys) I\'m just helplessly, completely, undeniabley, in love with the boy. We\'ve already talked about marriage and kids. I mean seriously, I really do not think will get married, but that would be cool. I just want this to last as long as possible. It\'s been 3 months and 3 weeks and 3 days. HAHA, thats funny. Anyways...I LOVE HIM!!!! Teeheehee...can\'t wait to see him tomorrow.      Love Cherry Loves Tregory Love '),(1767,3785626,'18,July,2004','2004-07-18 00:00:00','    I love Jervon Tregory Esters. &nbsp; Well today was okay. I basically did nothing. I got to drive today, which was fun, I did eh-okay. Then I went to church, which was good. Now I\'m listening to my H.I.M. CD.&nbsp; I think next time I\'m going to get Breaking Benjamin. Costs 13.99 at Sam Goody. And then maybe CKY, but I have most of the songs burnt on CD\'s. Lol. Talked to Tregory today, I\'m happy. I wish I could just tell him how I felt. I wrote a poem, it sucks, but tells exactly how I feel I guess.... &nbsp; Words don\'t express enough for me When I try to tell you how I feel It just comes to these 3 words: I Love You I don\'t know any other way to let you know this Except stay by your side and always be true Do you really know how much you mean to me? I think about you all the time I just can\'t get you off my mind I wish I could tell you everything I feel But all it comes down to is... I LOVE YOU &nbsp; I believe I started to cry. And that\'s never really happened before, I\'ve never cried over a boy, and especially for no good reason, or is that a good reason? Anyways *sigh* And because I\'m 15, it just scares the living bageeba\'s out of me, I might get hurt. But that\'s just a chance I\'ll have to take, right? RIGHT!!! &nbsp; Later, Cherry &nbsp; I Love Tregory 4 Ever N Alwayz....3/22/04 '),(1768,3785626,'18,July,2004','2004-07-18 00:00:00','    Okidokies, today was a good day for me. Dad woke me up early saying I needed to help clean up Grandmother\'s yard, because yesterday the storm blew in alot of sticks and stuff. So me and Chris helped picked them up. I got $20 for it. Yay. Then after that we went to Wally World. I bought me 2 binders, Spongebob Rocker notebook, and white out. Lol, I\'m going to get more school supplies later. Then after that we went to Sam Goody. I bought me H.I.M.\'s cd: Razorblade Romance. It\'s really good. Then we went to JC Penney. I bought me some jeans, they fit really good, and they are long. Just how I like them. Then I bought a black shirt and it has a cherry on it. I got a Snoopy shirt, it looks old skool like. I also got me a blue shirt with some stuff on it. Anyways after that I took a nap. Then Tregory called me. Happy Me. We are going on 4 months, which is my longest relationship at all. I\'m so happy. I\'m going to buy him some boxers for his birthday, and then give him like $20. Right now I\'m watching Star Wars Epi II, Hayden Christensen looks good in this movie. LoL. Samuel L. Jackson, is one of my favorite actors. Fav Actors: Samuel L. Jackson, Denzel Washington, Richard Gere, George Clooney...Fav Actresses: Julia Roberts, Halle Berry, Uma Thurman, I don\'t have that many actresses. Lol, Anyways I don\'t know what else to say, but I just feel like blogging. It\'s fun. I don\'t want to go to Band Camp, I don\'t want to see some people, and I don\'t want to wake up early. At least I\'ll have something to do. I haven\'t played my clarinet for awhile, so I really suck. Poopy. I can\'t wait till school starts, these are the classes I am going to take. Spanish, Family Dynamics, Band, Geometry, Chemistry, English II, and History. Adriano\'s mom teaches spanish through fibro-net. \"Hola! Como Estas?\" Then Mrs. Moncreif teaches Family Dy. I think. Mr. Mac teaches band. I don\'t know if he\'ll let us play the good&nbsp;music. I don\'t know who will teach me Geometry, because Mrs. Burton left to Holmes. :( She was one of my favorites teachers. Then I don\'t know who teache Chemistry...Its not Mrs. Nail. She was cool. Then English II, might be Mrs. Reed, the evil teacher...And History. I don\'t know. Some teachers left...Anywho, I don\'t know where we will have the football games, because they tore up the field and everything. I\'m definately going to the basketball games because Tregory is on the team. :)&nbsp; Happy Me. Yes Yes Yes. Happy me. Tomorrow I\'m going to church, so yay. I haven\'t been to church for a few weeks cause my ride has been in Florida.&nbsp;Lol, Anyways my brother wants me to get off, so later. '),(1769,3785626,'20,July,2004','2004-07-20 00:00:00','    So I woke up at 9 today to watch my mom because my dad had to go out and help Grandmother. Then at like 10:30 Tregory came over. :) I was looking at him, like I always do, but I noticed he looks alot like Usher. Or Usher looks alot like him. LOL, I wonder if he has the body of Usher?...Something we\'ll have to find out later.  ANYWAYS  then after he left I went to Fred\'s, and um...then slept from 5 to 6. After supper slept until 8. I\'m still tired. *Yawn* Oh, Tregory didn\'t bring his dog over here, because he couldn\'t get one cause his little brothe got the last one. :( And I wanted to name it. PooP. Well Sharla called today, it\'s her last day, she is moving to Arkansas. :( I\'m so bored. Well um...later.      I love Tregory 4 Ever N Always '),(1770,3785626,'20,July,2004','2004-07-20 00:00:00','    Okay well, Tregory came over today.  :) Um...Then after he left I fell asleep.   Then woke up for supper.   Watched WWE RAW.   Called Tregory.   He kind of made me mad,   so I quit talking,   but he wouldn\'t talk because I wasn\'t,   and I don\'t think he quite knew why I was mad,   but anyways I hung up cause nobody was talking,   but asked first, \"What are you doing?\"   So I thought he might had fallen asleep.   Then he called back,   after like 5 minutes,   he said I\'ll call ya back in a few minutes.   So yeeeah, he hasn\'t called back yet,      and I don\'t think he is,    which is pissin\' me off.   But I\'m sure he has a good reason.   Anyways I really want to see him  ...*sigh* I hate to love,    really I do. '),(1771,3785626,'21,July,2004','2004-07-21 00:00:00','    I am deciding on my future career. It seems like people with problems seem to gravitate towards me, I\'m not saying my friends are problems, I\'m just saying they seem to come to me with help. I\'m thinking of becoming a psychiatrist or psychologist, but I don\'t want to start analyzing my family and friends. And they say the highest suicide rate is caused by psychiatrists, because they hear all those problems. But maybe a counselor, they aren\'t as bad? Or maybe just a sociologist? Study human behavior. I really like listening to peoples problems I don\'t know why....It\'s strange.    Anyways I\'ve talked to Tregory almost all frikkin\' day. YAY. Last night he said he would call but he didn\'t he fell asleep. He\'s getting bad at that. I hate it when all those \"what if\'s\" go through your head...BLACH.   well later. '),(1772,3785626,'23,July,2004','2004-07-23 00:00:00','    I believe my insecurities are stupid. Stupid and no good. I talked to Tregory today. :) And I think I trust him with everything I have, and that may be a stupid gamble, but I\'ll take it. I have to understand that my past relationship was the PAST and the present is a gift to me, especially because it\'s not my past! Is that confusing? Anywho, I was talking to him awhile go and this is like the conversation or close to it. Me: \"Have you thought of a boys name yet?\"    Him: \"No.\" Me: \"Why?\" (I\'m always asking Why because I like too.) Him: \"Because I haven\'t been thinking about it.\" Me: \"Why?\"   Him: \"Because what I am thinking about is more important, I\'m thinking about the person I love.\" Me: \"Ohhh...\" Him: \"Do you know who that is?\"   Me: \"No...\" Him: \"You don\'t know...?\" Me: \"It could be your mom...\"    &nbsp;  Okay and then we got on a whole different subject. Anyways Saturday, hopefully we can go see Catwoman. Yay! Today he is supposed to come over...Yes!!! I    I love Tregory.   Yes I do.   No question about that.   My trust in him is 100%.   Yes it is.   If I get hurt. Oh dam* well, I\'ll get over it. Won\'t I?    &nbsp;  Love Cherry Loving Tregory Love '),(1773,3785626,'25,July,2004','2004-07-25 00:00:00','    *I Love Jervon Tregory Esters*'),(1774,3785626,'27,July,2004','2004-07-27 00:00:00','    Okay I just need to talk. For the last few days I guess, me and Tregory just don\'t seem to have the same phone conversations. They are, well, boring. He doesn\'t talk like he used to. Is it me or am I just boring? Maybe it\'s just a phase...maybe our \"spark\" is dying? I mean it doesn\'t really happen when we are together, because well durh, we\'re together. I just wish I wasn\'t so boring...or what not. I don\'t know...I think I\'m just over reacting. Yeah thats it. '),(1775,3785626,'29,July,2004','2004-07-29 00:00:00','    So I just now recently thought about this...I use this mostly to talk about Tregory. Hrmm...I don\'t know why...Anyways yesterday I went hiking up Little Mountain. Then I went to French Camp and looked at all the historical stuff. Then ate at French Camp\'s park and played. Reminded me so much of recess as a little kid. Then after that we went to Ackerman\'s park and raced around the track. I made it 2nd. Then me and Kristina played batmitton&nbsp;against Adriano and Jacqui, we won. Champs!! Then after that we went to Kristina\'s house and stuff. It was lots of fun. Then we went to church. Then today Tregory came over. :) Happy Cherry. Then after that me and my dad drove around Ackerman. I\'m really getting better. I didn\'t stall as much. Hehehe...'),(1776,3785626,'30,July,2004','2004-07-30 00:00:00','    Okay yesterday I posted but my computer decided to make me mad so this is the post from my LiveJournal.  Okay, Yesterday(July 21) was wayyy awesome. I woke up at 6:30 got ready for the hike. They picked me up around 8:00. Then we went to Little Mountain. It was a nice day for a walk too. I got my exercise. The whole time I was thinking about Tregory. =D Then after that we went to French Camp, got there at like 10 something I guess. Then we walked around the historical places, it was very interesting, I think my mom would like it because she likes that kind of stuff, but I don\'t know if she could take it. We went to the little gift shop thing, it was awesome. Then after that, we went to the French Camp park and ate and played on the Merry-Go-Round and the swings and the zip cord. It was like recess. Then after all that we went to Ackerman park. Me and Christina played batmitton(Sp?) and we kept winning. Yay. We rock. We also raced around the \"track\". I got 2nd place but that\'s just cause I was getting tired. Maybe next time, eh? Then after that we went to Kristina\'s and Jacqui\'s house. I got to see their kittens and we played Speed and War (Card Games). After that we had hamburgers. They took me home to take a shower and change. And then I went to church. I had fun. Did you have fun? Anyways then today I did nothing. Tregory came over. Really happy!!! And then I got to practice driving. I saw Derek, Tregory\'s cousin, and I saw Cortez and Cortez. Lol Then I went to sleep after that. Saturday I am going shopping with Ashley and Monday after band camp I\'m going shopping. Yayness! We\'re going to have lots of fun. Wheeee!!! I Love To Shop. I just wish I had a job for money. Yameen? Anyways I get my contacts the day before school starts. I\'m happy. I don\'t know when I\'m going to get my hair cut. Just whenever...Blah...Well later.Love Cherry Loving Tregory Love  Anyways now yesterday I didn\'t do much but e-mail Ashley and see Tregory. I drove around a bit. I think I\'m getting kind of better. Hm...Anyways then today I saw Tregory. :) Happy Me. Tomorrow I\'m going shopping with Ashley! Woot! If I mow the grass I get $10, and maybe my dad will fork over $20 more. So then I\'ll have $30 to go shopping with. Cause you know these days you can\'t buy anything good with $20. Lol, Anyways I\'m off. Later. Love Cherry '),(1777,3785626,'01,August,2004','2004-08-01 00:00:00','    So I don\'t feel like retyping what I did...I\'ll do it the lazy way. LiveJournal. So yesterday I went shopping with Ashley, they came and picked me up at around 3. We headed to Starkville and went to JC Penney to meet up with her aunt and cousin who is 10 and talkative. Anyways we went their, she bought a lot of stuff, and I got 1 black shirt that says Pixie with Tinkerbell on the front and back. REALLY CUTE. Then we headed to Columbus and went into Burlington Shoe store, and looked for Ashley some shoes. I already bought mine. She didn\'t find anything and then we went to the mall. Went into Shoe Dept. she found some cute blue and yellow sketchers. I liked the pink and black though. ^_^ Then after that we went into Sears. I bought another shirt that says DOTS like the candy. Then we went to Claire\'s I bought me a Happy Bunny Pillow and a ring that has dice in it. SO CUTE. Ashley bought me a pair of earrings for my bday. (April 30) Then we went into 4ever Young which didn\'t have much. Then after that we went to Pizza Hut, and ate. Yum. Then we went to McRae\'s, which I think I LOVE!!! They had the most cutest shirts and pants. It was alot like JC Penney \'cept with more variety and better sales. I bought a shirt on the sales rack for 1.99 ROCKIN!!! Then I bought this blue and baby blue cap. Matches my chuck\'s. Big Pimpin. Anyways I had lots of fun. We were going to go bowling but her little cousin aggravated the heck out of his momma so we couldn\'t go. Pooey. Oh well.  >>today tregory came over. woohoo. we played basketball. i won. cause i played by my rules (lots of fouling). hehehe, anyways i\'ve seen the biggest trend this year is Pink and Black. and it doesn\'t have to be PUNK its on EVERYTHING!!! i swear...some pink and black is okay but i see it everywhere. ugh. and also \"vintage\" shirts that have stuff on \'em. Like DOTS candy or CHARMS suckers. yameen? they are cute but its such a trend. BLAH...anyways tomorrow is band camp. woohoo lucky me. '),(1778,3785626,'03,August,2004','2004-08-03 00:00:00','    i think my insecurities about tregory cheating are stupid. i believe he wouldn\'t cheat on me but then again my mind continues to think about it. sometimes i wonder...but i have no doubt that i love him. yeee band camp not. '),(1779,3785626,'03,August,2004','2004-08-03 00:00:00','    .band camp. yay. i love tregory. wheee. goodbye. love.cherry. '),(1780,3785626,'05,August,2004','2004-08-05 00:00:00','    Today was okay. I went to band camp like usual. At break Tregory was at the park so he came over to see me. I told him last night he was scared to come up there, I just wanted him to come and see me since I haven\'t seen him in 3 days. Heheh, Sneaky. Band was okay, we got alot of stuff accomplished but Mr. Mac got to see our bad side, cause alot of people were talking. Then after band Tregory was still there. Yay. And I went home for lunch. Then Ashley called me and told me we weren\'t having practice. So yay. After that I went back to the band hall and got my instrument. Then I went to Piggly Wiggly. I\'m never going there again. I know I said that before but now I\'m for real. I\'m buying a cake for my dad so I\'m looking at cake mixes. (Did you know they make caramel cake mix?) This dude walks by and says..Dude: Hey How Ya Doing?Me: Um..Okay?Dude: Baking a cake?Me: Uh..Yeah? *Stupid! Why else would I be in the cake mix section.*Dude: Gonna bake me one?Me: Uh..NO!? (I should have said HELL FUCKING NO)Dude: Aw, man.Perverts must crowd around Piggly Wiggly or something. Ugh Anyways I\'m so happy I got to see Tregory. Let me tell you my dream.Okay I don\'t really like this girl Jackie because she\'s kind of nosy and loud but I\'m cool with her. Anyways last night I had a dream I was stabbing her with a fork. Weird huh? Not stabbing her in the heart or nothing just on her hand because she had something of mine. I guess I don\'t like her because she used to go with Tregory in the 8th grade, even though I wasn\'t here. Anyways it was kind of cool. Haha. Well thats my day. Today is my dad\'s birthday. Yay!!Later Love, Cherry I LOVE TREGORY!!! '),(1781,4031020,'31,July,2004','2004-07-31 00:00:00','   Weight: 290  Food: 2 cans of Cheez Balls 1 20 oz Diet Mountain Dew Bowl of grapes Foot long Subway sandwich 32 oz. pink lemonade Cookie 2 Fat free Italian Ice snacks  Maybe I\'ll end this week with a 2 lb. weight loss. Maybe not. Either way, I\'m very glad that I finally put it out in the open (at least to myself, since I think I\'m the only one reading this blog), and I can see with my own eyes that I must make a lot of changes if I\'m going to lose a ton of weight in the next 5 months.  Tomorrow is the start of a new month. Don\'t you just love fresh beginnings? I think God gave us new beginnings for a reason. New minutes start every 60 seconds, new hours every 60 minutes, new days every 24 hours, new weeks every 7 days, new months every 30 days or so, and a new year every 12 months. That was no mistake. Every day is a day to to start over.  Tomorrow is August 1. I\'m so glad I started this little journey of mine exactly 7 days ago. I got to see a week of how I am when I call myself trying, when in reality I\'m barely making the effort. That changes now.  I finally made it to the grocery store. Wow! Who knew it would actually happen? Now I have healthy food to eat and a determination to do it. Tonight I\'m going to pray for help and follow a more calculated type of plan.  My teeth hurt. I need to go to the dentist desperately (like for the past 6 months), but I can\'t afford it. Bleh.  '),(1782,4031020,'30,July,2004','2004-07-30 00:00:00','   Weight: 290  Food: I don\'t even want to GO there.  I always seem to have my most poignant revelations after I\'ve eaten an insanely huge meal. It\'s also very easy to promise myself all the things I\'m going to change when my stomach is full. When I\'m hungry, I could care less about changing my life -  I just want to eat .  I don\'t hate thin people. I also don\'t admire them. I truly think people are just people, and everyone has their quirks. I have to admit though, that sometimes I wish my \"quirk\" wasn\'t so damn visible. I mean, you\'ve got your anorexics, your alcoholics, your nymphomaniacs, your drug addicts and your fattys. For the most part, these people can pretty much hide their problems and either work them through personally or literally kill themselves over it.  Except the fattys.  No, we have to not only eat ourselves miserably into oblivion, but the whole damn world knows it. And then the world makes fun of us for it. Even Paris Hilton would have to admit that sucks. Except she doesn\'t know what it feels like so she could probably care less. Hell, I could care less about her problems, so who can blame her?  I walk around every day advertising that I have a problem. So what do I go and do? I start a blog, so I can talk about my own big fat ass. What a genius thing to do.  Actually, jokes aside, the point of this blog is to help put my weight back in my face. A few years ago, I moved to a separate city after graduating from college. A year and a half after starting my first job, I started losing a bunch of weight. I started out at 263 lbs., and then I lost 55 of them. Yep, I held steady right around 208, and I stayed there for a while, waiting for my next \"losing spurt\" because it happens in cycles of determination. Actually, even at 208 I was quite fit. You could see my abs, I was in a size 18, which is just 4 sizes larger than the average woman, and I could run a few miles per day. I was large, but NO ONE would guess that I was over 200 simply because my level of fitness was so high.  Then my dad got sick and I moved back to Chattanooga to run the family business. My hometown.  The city that made me want to eat in the first place because all my problems started here.     It was a good decision. I think I accomplished what I set out to do. My father died, and I was here for my family when it happened. But here it is, exactly 2 years and 6 weeks after I moved back home. Not only did I eat my 55 lbs. back, but I added 29 more. The fog is lifting and realizing what I did to myself is very devastating.  That\'s why I\'m more determined than ever to beat this thing. As you can see from the first week, I\'m a serious creature of habit, and it\'s very hard for me to change my patterns.  But once I do, there\'s no turning back.  I did it before.  I can do it again .  In my heart, I know I\'ve already won this battle. But the journey is the fun part, so I hope you can join me.  '),(1783,4031020,'29,July,2004','2004-07-29 00:00:00','   I\'m fat. I\'m so freaking fat.   Weight: 292   Food:  3 ribs, Baked Beans, Potato Salad  Chick-Fil-A Sandwich, Waffle Fries, Diet Lemonade  20 oz Smoothie  1 Pint White Turtle   So here\'s the deal. On the Chick-Fil-A, I actually ordered the biggie size, but later reduced it. That was my only victory for the day. Hoo-ra, Hoo-ra. Actually, I\'m just saying that to make myself feel better for being such a fatty. Writing this&nbsp;blog has helped me realize one thing. I like being a fatty. Who cares if I bury all my emotions into food instead of booze, drugs, sex, work or relationships? Nobody could give a damn.   Except me.   That\'s the one problem. Me. I truly hate being like this, but look, it\'s been 5 days and I\'m still resisting making permanent changes. How on earth do I interpret that? Creature of habit? Self-sabotage? Who gives a damn what the psychoanalysis is? I\'m HUMONGOUS.   I had a job interview today. I really need a job because I\'m getting married in April and I have no money. The family business went out of business and that was where I worked, so now I\'m SOL. I had this job interview, and whereas I\'m usually overly confident in my abilities, this time I felt like I was faking it the whole way through.   The head conversation went kind of like this (italics = thoughts, plain letters = actual spoken words).   Interviewer: \"So, tell me about yourself. \"  There\'s no desk to hide behind! How huge do I look to this little tiny lady?    Well, I\'ve been in marketing for 5 years now...    I wonder if my humongous cheeks are shiny.    I have great experience...    Are they looking for someone who has to be camera friendly? Because they should have just turned me away at the door if that\'s the case!    My past jobs have given me so much experience...    If you hire me, I swear I\'ll lose weight! I\'m getting bigger simply because I need this job so much!    Marketing is exciting to me...    Please, be mesmerized by my lilting voice and outstanding work history&nbsp;and not my shiny face and jiggling jowls.    I am a perfect fit for your company.    This is agony. If I start sweating, I\'m getting the hell out of here.    It\'s highly irritating trying to interview when you can\'t get past your own looks. Never before have I been so concerned with how my weight affects my interviews. I\'m quite capable with an outstanding resume and I\'ve landed jobs much more often than I\'ve been passed up for them, so this is unnatural to me. This is the first time I\'ve ever had an interview above 250 lbs. It truly is agony, and I hate it. Big, humongous reason to lose this weight NOW.   Notice that I\'m back to my original weight (from Sunday!) today. Don\'t be surprised, this happens every time because of my out of control eating. Every time I try to lose weight, I end up right back where I started. This time, I\'m just going to work through it, and eventually past it.   Yep, I still plan on getting to that (*)@#$@# grocery store and following the rules - that *I* made up for myself no less! I can\'t believe how hard it is to get motivated to do this. It\'s such an overwhelming task, that I don\'t even want to start it, lest it never end. Then again, I haven\'t stopped thinking about my weight in about 4 years now, so what makes me think NOT doing anything about it will make it better? I would love to end this journey as soon as possible. I\'m working on it folks.   Really good bit of news: Today I went walking with a friend. We were out for about an hour and a half, 60 minutes of which was walking (with a stroller, not too brisk). It was so difficult for me, which is ridiculous because I used to be able to run this distance without blinking. But it\'s my first exercise since I started the blog so I\'ll be sure to keep it up. Our next adventure is to go bike riding, and I\'m looking forward to that. I just feel VERY sorry for my bicycle tires. I hope they make it!   My unemployment check will post to my account tomorrow. Once again, my primary goal will be to hit the grocery store. Only this time, I mean it.   Wish me luck.  '),(1784,4031020,'28,July,2004','2004-07-28 00:00:00','   Today\'s stats:   Weight: 288 (Another pound gone! Woohoo!)  Food: 14 pieces sushi - tuna, shrimp, eel Miso Soup Salad Chick-Fil-A Combo #1 (sandwich, fries, diet lemonade) 6-piece chicken strips  You know what? Discussing my food choices every day could get awesomely boring, so I think I may just make light comments and move on to lighter fare. Don\'t ask me about the Chick-Fil-A. Notice I  still  haven\'t been to the grocery store yet. Actually, I had very healthy sushi today, just 2 rolls and 2 nigiri with soup and salad, and Chick-Fil-A\'s chicken sandwiches meet the 30% fat requirements. Plus, I got the diet lemonade. It was the supersized fries and chicken strips that put me over the top. But since it was an improvement over yesterday, I think I\'ll make note of the fact that each day I make better choices. Maybe by the end of the week, I will be cooking for myself instead of depending on strangers to feed me.  To all you \"skinnys\" out there, listen up. For someone at my weight, 288 now, going to the grocery store is a harrowing experience. No, the food is not calling to me - I could care less about the Entenmann\'s and the Krispy Kreme. It\'s just the experience in general. You see, fat people know more about dieting, weight loss, health and food than any normal person could ever know. We\'ve read it all. We know it all. Losing weight has nothing to do with knowledge.  However, it\'s that knowledge that is my downfall. Let\'s say I want to buy spinach. Here\'s what a typical thought process is like:  Spinach sounds good. What an excellent choice! I can get it canned, but actually that\'s not as healthy as fresh. How does one cook fresh spinach?  I bet you have to use olive oil. Hmmm, maybe I won\'t buy it fresh - I can\'t spare all that olive oil.  Do I even have olive oil at the house? How many calories does spinach have again? 20. 20 per serving.  Wow, I could eat 20 servings and still not eat as many calories as I had for lunch! So should I get it in a bag, like lettuce? No, I think I\'ll buy it loose. Organic is even better than bagged. Damn, the organic costs twice as much! The last time I ate spinach, I was on the pooper for an hour. OK, I\'m deciding on the fresh, non-organic spinach. Maybe I\'ll eat it like a salad! Oh crap, what about dressing?  Should I use lemon juice? Ugh. They\'ve got some lowfat stuff.  Remember, that stuff tastes like urine. How about the low carb? Argh! Forget the damn spinach!  Every time I step foot in a grocery store, I end up with this dialog that drives me absolutely insane. I really would like freedom from that. I bet skinnys can just walk into a grocery store and buy spinach and then figure out what to do with it later. Me, I have to consider calories, carbs,&nbsp;servings, how it has to be cooked, what affect it will have on my body and last of all, whether I even think it tastes good. It\'s a real pain.  And it\'s like that for every choice I make.   Health knowledge isn\'t always an asset.  &nbsp;  &nbsp;  '),(1785,4031020,'27,July,2004','2004-07-27 00:00:00','   Man, I really suck at changing my life for the better.  Today my diet has been:  Arby\'s Turkey, Ranch Bacon wrap (I figured as long as I\'m eating fast food, I may as well go low carb. Stupid, stupid logic). Small order of cheesesticks Large Sweet Tea Caramel Cashew Waffle Cone Pint of White Turtle Ice Cream  All in all, I think I did better than yesterday, but not much better. I never did go to the grocery store like&nbsp;I said I would. Hell, I didn\'t even shower today. That is a HUGE first step for me. I think I will make it my #1 goal for the morning.  Don\'t worry - I\'m still going to make it! Changing my fat mentality is really going to be a pain, I can see that already. My mind WILL submit!!!!!!!!!!  '),(1786,4031020,'25,July,2004','2004-07-25 00:00:00','   Weight: 289 lbs.   Wow! Yesterday was a challenging day.    Day One Food:   1 apple  3 roma tomatoes  2 large scoops of chunky peanut butter    Day Two Food:   Large baked potato with butter, sour cream, cheese, barbecued pork and bbq sauce  1/2 homemade Pecan Toffee Bar  Wendy\'s Bacon Cheeseburger, Great Biggie Fries, Chicken Strips and Great Biggie Sprite   Day Two is really a typical day for me. You do NOT get to&nbsp;292 lbs. without a concentrated effort at eating and&nbsp;not exercising.&nbsp;The 3 lb. weight drop occurred because of the small amount I ate on Day One. Day One is not a typical healthy day for me, I just ended up sleeping an awful lot and I didn\'t do much so I wasn\'t hungry. I might have been a bit fatigued. Probably because I weighed 292 lbs.!   One notable thing that happened yesterday was my niece was with me when I ordered that baked potato and she ordered the EXACT same thing. Keep in mind that she\' s only 14, but she\'s still growing and she\'s thin as a rail. Her whole family (my sister\'s husband/children) is naturally thin, so she never overeats. While I almost cleaned my plate of that baked potato (even *I* couldn\'t finish that monster), she ate approximately 1/4 of the whole thing. And even more, she had no problem throwing away the remainder of it!   Initially she said, \"I\'m going to eat this whole thing.\" Yet, she didn\'t charge into it ravenously like I did. She took her time eating it, she mixed it all together until it was perfect, then she ate it with relish. When she was done, she was done. We both laughed at how much she had left on her plate, but secretly I wish I could have done such a thing. There\'s something about that amount of food put in front of me that presents some type of challenge. There\'s a voice in my head that says, \"You can eat it... just dive on in... wait, don\'t leave that piece out just because it doesn\'t taste as good... you paid for this, so finish it!\"   Tapping into this knowledge again is going to help me continue toward my goal. What I need to&nbsp;do is go grocery shopping. After reading a small mountain of dieting books throughout the course of my life, and losing more than 50 lbs. a few years ago (only to gain back almost twice as much) &nbsp;I already know what my eating plan will be. It will be lenient in some places, strict in others. But with my aggressive goals, I have to have a specific plan or I\'ll eat about 100 more stuffed baked potatoes between now and then!   My eating plan is as follows:  Any kind of white fish, broiled, baked or steamed  Any kind of fruit (preferably raw, fresh)  Any kind of vegetable (raw or steamed)  Any kind of herb or spice  Limited amount of extra virgin olive oil  Basmati Rice (Is it just me or is it extremely difficult to stomach brown rice?)  Brummel & Brown non-transfatty margarine-like product  Guilty pleasure: ONE Bruster\'s ice cream cone per week (yep, it can be single, double or even triple!)   See? My eating plan includes a lot of different types of flavors, combinations and allowances. I basically just took out refined products, processed products and other stuff that I don\'t really like anyway. And my guilty pleasure... well... we all scream for ice cream!&nbsp;&nbsp;*smile* Since there is so little sugar in my diet, this may disappear after a while, but as for now, it is a happy part of my list.   My first step? I\'m off to the grocery store!   P.S. I\'ll be posting pictures of me and all kinds of good stuff in the near future! I just don\'t have many pictures of me at my top weight yet, but we recently went on vacation in the Bahamas so when I get them back, I\'ll post them.  '),(1787,4031020,'25,July,2004','2004-07-25 00:00:00','   To start off, here are my stats as of today.   Weight: 292 lbs.   Measurements:  Arm - 17  Neck - 17  Under Bust - 44.5  Over Bust - 48  Waist - 46 (above belly button)  Mid&nbsp;- 57 (below belly button)  Hips - 55  Thigh - 29  '),(1788,4031020,'18,August,2004','2004-08-18 00:00:00','   Weight: ??? Didn\'t check  Food: I ate well - fish, a baked potato with just Brummel & Brown, and some rice. The only bad thing was an enormous amount of ice cream.  Today\'s thoughts are about something simple: Burying Emotions. I don\'t plan to censor myself in this blog for obvious reasons - for one, it would be horribly boring and for two, that would defeat the entire purpose for writing it. But recent online conversations had me thinking about how much we fattys bury behind food. Hell, I eat sex, drink pain and gobble down frustration like a Thanksgiving feast. My fiance and I are waiting for marriage until we have sex, which is fabulous IMO, but in the meantime, I\'m still a woman, and I still have needs! Whoops - I\'m getting off topic. hehe ;)  Anyway, haven\'t we had enough of that? I mean, what has food ever done for us that warrants that we use it as a cure-all for every event or emotion, whether it\'s positive or negative?  If anything, I would love for my fellow fattys to just get downright nasty with what they have to say, and stop worrying about who \'appreciates\' what you are saying and who thinks you\'re an ass for speaking your mind. I would love to hear fattys just admit their problems. Remember \"What Women Want,\" that movie starring Mel Gibson where he could hear women\'s thoughts? I imagine that if that movie was entitled, \"What Fattys Want\" it would sound something like this:  \"Food tastes better than sex feels sometimes.\" \"I love my children, but they are awfully annoying. I\'d secretly prefer that they watch TV than whine in my ear about their homework.\" \"I make a load of money, but I would totally trade in my job for a smaller house, an older car and zero debt if that were socially acceptable.\" \"Why do all these trainers keep trying to force me to love exercise? I just do it because I have to, and no other reason.\" \"How come you never see a jogger smiling?\" \"Are people looking at me? Is my ass jiggling? I wish I could *will* my ass to stop jiggling!\" \"I\'m tired. I\'m so very tired.\" \"If anyone could see what I\'m eating right now, I\'d be horribly embarassed.\" \"Maybe I should just give up. No, I can\'t give up, I\'ll be 500 lbs.!\" \"I resent her for giving me dieting advice. Does she realize I have read exactly 17 low carb books, 58 lowfat books, subscribed to 6 fitness magazines and I even accidentally bought TWO copies of the South Beach Diet? Does she think I don\'t  know  how to lose weight? I\'m the master and commander of weight loss knowledge!\" \"If I gain one more dress size, I\'m going to have to buy a whole new wardrobe.\" \"Isn\'t being fat punishment enough? Do I have to be broke and ugly, too?\"  I can\'t say that I\'ve felt all of those things... or even most of those things. But I\'ve read them or heard them from friends, ranging from 10 lbs. overweight to 100 lbs. or more. Yet, we still feel a need to hide from people, to cater to how others feel about we might say, or how we might feel. Well, you know what I say? Screw \'em.  I\'ve been catering to other people long enough. I\'ve kept my mouth shut and politely obliged, apologized or catered to others when someone told me what to think, how to act, what to eat, what to wear, and where to go. And you know where it got me? 290 lbs., in the unemployment line, with enough debt to feed an entire city and less than a week\'s worth of clothes to wear. I\'m  over  caring about other people\'s opinions of me.  I\'m starting my new job on Monday. I have the opportunity to get away from my hypercritical family and make new friends. I finally have the chance to tackle my weight without being embarassed about who I might see at the gym that I knew when I was smaller. I don\'t have to worry about my sister (who owns a catering business) manipulating me to try her latest concoction. Or eating Sunday dinner with soul food, which I hate. It\'s like starting over!  Anyway, I bet if more fattys let their emotions out instead of keeping them buried down, there\'d be less fat people in the world. Or at least less unhappy ones!  '),(1789,4031020,'17,August,2004','2004-08-17 00:00:00','   Weight: 290  Man, I\'m starting to realize that I really suck at this. Last night I was watching the Olympics and I\'m just fascinated with the physical feats these kiddies can accomplish. I say kiddies because I heard one of the announcers call a swimmer \"old\" and he was 23. And in gymnastics, 23 is ancient. So kiddies it is. But damn impressive ones!  These kids train rigorously, which means they probably have to eat prescribed meals, sleep very little, work out constantly, and strive for perfection. What looks flawless to my eyes may look horrible to the eyes of a critical judge.  I saw an Olympics commercial that had a bunch of people saying \"I can.\" \"I will.\" \"Get ready.\" to themselves to get their adrenalin flowing before a competition. I think that attitude is the one thing I lack in trying to lose all this weight.  Losing 140-150 lbs. *is* an Olympic feat. It requires the following: - Abnormal determination to beat the odds - Special diet to achieve maximum performance - Daily workouts and rigorous training - The ability to ignore the naysayers and critics - Critiques of goals and performances - The ability to accept wins  and  losses gracefully - Comfortable with strangers watching you, waiting for you to mess up, and proving them wrong  Seems the biggest difference is, I\'m not representing my country and I don\'t get a gold medal if I make it to the top. And I get no recognition if I actually achieve this feat, besides societal acceptance and oh yeah... health. I think the general consensus in America is that if you lose weight, it\'s no big deal, you should have never let yourself get so fat in the first place.  So the Olympics have given me a new reason to shine today. No more Greek desserts, heavy pastas and fast food for a while. I am back to my fish, rice, fruit and veggies (and I\'m glad about it!).  '),(1790,4031020,'16,August,2004','2004-08-16 00:00:00','   Weight: 290, 288  Sunday Food: Greek Shrimp Tourkemanelolo (incorrect spelling) Tortiyava (incorrect spelling) Leftover angel hair pasta from Saturday 1/2 sugar free Greek cake Popcorn King Size Almond Joy Movie-sized Sprite  Monday Food: McDonald\'s McChicken Sandwich Large Fries Large Sprite Apple Pie Greek Oreo Cake Greek Sugar Free Cake  In the interest of celebrating the Olympics, and its return to Greece, my friend and I went to The Acropolis, a local Greek restaurant here in town. Of course, after my diet in Atlanta AND my Greek special interest day, PLUS movie-sized snacks at Bourne Supremacy, the scale lashed out at me this morning like an abusive husband. The red glow of the digital numbers glared at me like someone just recovering from an alcoholic binge. 290! Fatass! What the hell were you thinking??  Know what I was thinking? Mmm... buttery croissants (delectable!)... pasta baked in mozzarella (sinful!)... Greek (it\'s a celebration!)... movies (a movie isn\'t a movie without a jumbo snack...or three!)... and more cake (because my scale yelled at me).  And I was doing so well for the past couple of weeks, too! Well, all is NOT lost. I\'m still looking forward to moving to Atlanta and making good use of the fitness center and swimming pool on my complex, Shamu bathing suit and all.  After all, if I\'m going to develop a secret competition with a pregnant woman in my imagination, I should get on the stick right away, right? hehe...  '),(1791,4031020,'15,August,2004','2004-08-15 00:00:00','   Weight: 288  Food: Was in Atlanta today - went to an English tea for lunch and Italian for dinner.  Tomorrow is weigh-in, but I wouldn\'t be surprised if I broke even. I had Italian for dinner in Atlanta. I only ate 1/2  my entree, but I feel \'heavy.\'   Other news:  I found my apt. in Atlanta today! I did a ton of research but I really believe God led me to the right one today. It was the LAST ONE I saw, and it\'s also by far the cheapest one available in Buckhead. Yet I have all new carpet, walls and appliances, with  new  hardwood floors in the bathroom! NICE! I would tell you guys where it is, but you\'d just drive the price up when you flocked out to rent a place. hehe... Btw, this place has an AWESOME fitness center that I will most definitely be using often!  What I want to talk about today is a friend of mine that lives in Atlanta, and who, by the way, is excited about my imminent return to the city! I\'m also excited to see her more often. This friend successfully lost 140 lbs. a couple of years ago and she is now pregnant. We even started our own weight loss company together (I dropped out when I had to take over the family business... then I promptly gained 80 lbs.!). Now that she is pregnant, however, she\'s gained quite a bit of weight and she\'s over 200 lbs. And she still has 5 months to go.  Is it  so bad  to say that, in a way, I\'m glad she\'s gaining? It\'s not that I want her to be fat, or that I want her to suffer. It\'s that  I feel as if I can catch up ! When we were losing together, she was always approximately 25 lbs. ahead of me. It was obvious to me (at all times) that she received a certain amount of smug pleasure out of that fact. Whenever I tried to give her weight loss advice, she would say something along the lines of, \"Well, when  you  get below XXX lbs., then maybe you\'ll understand.\" or \"Would you like my size 18s? I can\'t fit into them anymore.\" That irritated me to no end! I mean, what is really the difference between 240 lbs. and 215 lbs.???  Anyway, she\'s about 208 lbs. right now. I\'m 288. If I can lose about 40 lbs. before she\'s due, and she gains about 20 lbs. or so in the next 5 months, then we\'ll once again be 20-25 lbs. apart. Now, I\'m sure she\'ll be much too involved with her new baby to be concerned about such a trivial thing, and I would expect her to drop the weight as quickly as she gained it, but I must admit...  I long for the day that my weight is lower than hers. I crave it like a fish craves water.     Of course, I\'m being quite unreasonable. I mean, come on,  I\'m trying to compete with a pregnant lady.  There\'s no honor in that! But there is most definitely a guilty pleasure. A few weeks ago she asked me for the clothes that I can no longer wear because I\'ve gained so much weight. I can\'t tell you the horrible, selfish joy I felt in handing those same size 18s that she gave me right back to her.  Actually, she\'s a wonderful friend and I know that she fears the weight gain. I know that she offered her old clothes to me because it\'s very difficult to update your wardrobe when your size decreases once per month. I know that her weight loss journey has been long and arduous, and I only wish the best for her - a trim, svelte body and a long, rewarding life.  However, this blog is not about how she feels or what she\'s going through. It\'s about my journey, and for the record, I must confess -  if I could see the look on her face the day I\'m just one pound lighter than her, it would make all this worthwhile.    Think of me what you will. I\'m just a squirrel, trying to get a nut.  '),(1792,4031020,'12,August,2004','2004-08-12 00:00:00','   Weight: 286  Food: Oatmeal Chick-Fil-A Chicken Sandwich Large fry AND medium fry Large diet lemonade Pint of Chocolate Oreo Ice cream  Yep, you read it right, I had fries and ice cream yesterday. Two of my biggest weaknesses! The reason I dropped a pound, however, is because I essentially ate one large meal yesterday, with a 100 calorie breakfast. Sometimes I wonder if the \"One Meal a Day\" theory is valid in its own way, but I like breakfast, lunch and dinner equally so that\'s something that wouldn\'t work for me. I know a lot of men have done well doing that, but I don\'t care.  Anyway, I want to talk about the Beautiful People. I\'ve mentioned the Beautiful People (BP\'s) once before, but I wanted to expand on my thoughts about them. Let me first reiterate that I neither despire nor admire the BP\'s, though I do sometimes find them fascinating to observe. Rather like zebras in the wild. You could care less if a lion eats them or if they live for another millennium, but they\'re interesting to watch.  I\'m a big fan of reality television. I think most people, whether they admit it or not, are fans of  some  type of reality television, whether it\'s Survivor, Celebrity Poker, Last Comic Standing or Fear Factor. Watching people interact with each other is fascinating.  Remember the first Survivor? It wasn\'t the world\'s first reality show, but it IS the show that was responsible for sparking the reality insanity that we experience on television today. There was Fat (and Gay) Richard, Old Rudy, Pretty Colleen, Weird Greg, Single Mom Jenna, Athletic Gervase and other types of people. Out of 16 people, there were maybe 4 BPs - I don\'t remember anyone being especially spectacular. They had a good Gilligan\'s Island type of mix - which in its own way, is kind of like the real world.  But notice what we have today. Even the OLD people on Survivor are fairly attractive, and really not even that old. Old Rudy was a leathery ex-military guy, abrasive, but human. But now since so many people from those first years of reality success have made it big, we\'ve got a different type of person auditioning for these shows.  Everyone  who applies to be contestants on the new reality shows are aspiring actors/singers/artists who are currently bartending/secretarying/hairstyling. They even choose the roles they want to play - the Bitch, the Drunk, the Slut, the Cassanova. I\'m not so sure if the cash prizes are even the motivation behind the show anymore. They want to make a career out of it these days. The direct result is a disproportionate amount of BPs kidnapping our televisions even for the  reality  shows.  Now, I don\'t know about you, but I really don\'t see that many BPs in  my  reality. That\'s why I was so astounded to see so many when I went to Miami. Maybe the producers of these television shows have lost touch. Don\'t they all live in LA, New York, Miami and Hollywood? Where almost everyone has bleached/highlighted/frosted hair, abs to die for, and perfect teeth? You know what I mean - the kind of person who has never thought twice about \"whether that skimpy bikini is appropiate\" or \"if I should wear a t-shirt in the pool.\" If this is  their  reality, then how can they cast a TRUE reality show?  I realized a long time ago that I actually feel lucky not to have been a BP. I\'m fairly attractive (more so when my weight is down), and also sociable when I want to be, so I\'ve never had a problem meeting friends and having a life. But I did grow up with a messed up image of myself. As a teenager, I thought the BPs I saw on TV were the  standard  and not the exception, so I had this really low self-image. I was the last one in my group of friends to get kissed (meaning: my first kiss was only because everyone  else  had done it and it sucked!) and the last one to start dating (which I stopped shortly after I started).  But that now irrelevant low self-image actually helped me. The people in my life were genuine. I was rarely approached by a man whose sole motivation was to get in my pants (well, just a couple of times, but it didn\'t work!). I never had to prove to people that I had inner beauty as well as outer beauty. I remember a few of my cheerleader friends always trying to do well in school so that people didn\'t think they were dingbats. I felt sorry for them. I, on the other hand, always had people assuming that I was far more intelligent than I actually am, simply because I was average and a bit on the shy side. There were even rumors going around in my old high school that I was secretly wealthy!  I read an article the other day where Halle Berry was asked what it was like to be so beautiful. Paraphrased, she said something along the lines of, \"Beauty never did anything for me. It didn\'t spare me pain or hardship and I still have to work for everything I get. And since it doesn\'t last, I don\'t rely on it.\" I thought that was pretty wise, and taken from a true BP, we should all listen up. She\'s known as one of the most beautiful people in America, yet as far as life goes, her life has been hardly different from the average woman. She has still had the pain of two divorces, she hasn\'t found someone with which she can trust to have children with, she still has to work hard to keep the momentum of her career going, and right now, she\'s alone. I\'m not a BP, and I don\'t even have half those troubles. Who needs it? Life is hard enough for the average ugly person, but contrary to what I used to believe, it can probaby be harder for a BP.  So here we are, people like me, the fattys of the world. Though we typically have genuine friendships, spouses who love us and nice careers, we look at the BPs in the magazines and billboards and visualize ourselves looking like them. But why? I think that generally we set ourselves up to think that when we are that thin, we can do  anything.  We can run on private beaches with a hunk of guy, falling into the surf and kissing him, or we can not only be thin, but also have the perfect makeup, skin, teeth, hair and nose. And strangers on the street will want to take pictures of us or hang out with us or just admire us for our wonderful thin-ness.  What?  There\'s one thing I realized when I lost 55 lbs. 2 years ago. I kept getting thinner, but  nothing else was changing.  Somehow I had imagined myself with a great body and hair and makeup, when in reality I was just getting thinner. I still had a scar on my forehead, a pockmark on my nose, big feet and leg hair. I kept looking in the mirror and seeing the same person. In fact, the only way I noticed I was losing the weight was I kept putting on pants that there\'s no way I could fit my big behind into and they would be loose on me. Why? Because I woke up and looked into the same face every day.  I think one of the biggest things I learned that I would like to remind myself of throughout this journey is that when I look in the mirror, I\'m not necessarily looking at my body. Not always. Sometimes I\'m looking at ME. Through my eyes. My eyes will never change. I think that\'s why my body perception seemed so skewed when I lost weight before - I felt like I hadn\'t lost a pound even though my body had completely changed. It was because when I looked in the mirror, I saw me, myself, and not my fat, or my skin or my blemishes.  There were times when I would consciously look at the differences that were happening on the outside, but if it hadn\'t been for my constant need to buy new clothing, you couldn\'t have convinced me I had lost a pound!  The question then was, if the person I looked at in the mirror wasn\'t changing, would it matter if I was thin? Suddenly, thin-ness took on less of a weight and I realized what it really was. Just improvement. That\'s all. All I\'ll be at the end of this journey is thinner. I\'ll look into the same eyes every morning no matter what.  All that \"love yourself\" mumbo jumbo has never worked for me. I think the simple fact that a person gets up, washes, brushes their teeth and combs their hair in the morning is proof that you love yourself. Getting a pedicure or a new hair style has never made me feel better or motivated me to lose weight. Mostly it just gives me something pretty to look at for a little while.  So now I\'ve lost 6 lbs in approximately 2.5 weeks. I think that\'s pretty darn good. I think I can clear 10 for this first month. Woohoo! Now, Do I think that makes me valuable? Nope. I think it\'s just damn cool! Do I think I\'m going to be a BP anytime soon? I hope not. The only person I really care to notice my beauty are my future husband and my children. And being fat didn\'t prevent me from finding the man of my dreams and it certainly didn\'t keep me from achieving new heights in my career (I\'m making the highest salary I\'ve ever made at my new job that I start 1.5 weeks). It didn\'t keep me from going on a cruise last month, or wearing a bathing suit in public. My weight actually hasn\'t kept me from doing anything.  However, my life is about making improvements spiritually, mentally and physically. The simple fact that  I don\'t like  my weight is enough reason for me to want to change it. And now that I\'ve dropped those first few lbs., I want to keep up the momentum.  Life is good right now. My debt is minimal, my weight is diminishing, my marriage is looming and my career is thriving. Who could ask for anything more?  '),(1793,4031020,'11,August,2004','2004-08-11 00:00:00','   Weight: 287  Yesterday\'s food was basically one large slice of pizza and then some fish, rice and olives later in the day. I don\'t drink hardly anything besides water, by the way. Not worth it.  Anyway, remember that debt that I mentioned yesterday that was weighing on me? Well, let me tell you, prayer CHANGES things! My mom has agreed to take care of it! It\'s really a miracle - my mom is extremely tight with money. I know that she will get it back ten times over, but she may not be able to see that right now. The last two years have been tough times for my family, but we\'ll overcome all of it, I know.  Now that I\'m moving back to Atlanta it\'s like I\'m being given a second chance. A second chance at weight loss (last time I was there, I lost 55 lbs, from 263 to 208!), a second chance at giving back to the community, and a second chance of proving myself. I think living here in Tennessee was like a training ground. God had to get me in a place where I couldn\'t look left or right, but just up. And I am hardheaded and stubborn so that\'s no easy task!  So now I can focus on my career, my marriage next year, and my spiritual growth. I think that when life falls into place just so, the weight comes off easier. I\'m at 287 today, so I\'ve officially lost 5 lbs in 2.5 weeks! Woohoo! I\'ll take it! It would be really nice if I could shed at least 1 or 2 more by Sunday.  I ate the last of my food in the fridge last night, so back to the grocery store for me today! (This time I won\'t avoid it for a week! hehe)  '),(1794,4031020,'10,August,2004','2004-08-10 00:00:00','   I got the job! I\'m moving to Atlanta in TWO WEEKS! I\'m so excited!  Yesterday was a bad day - I found out I\'m being sued for $25,000 because the old business credit cards were in my name and there\'s no money to pay them. I\'m a christian so this type of stuff doesn\'t worry me (when eternity is involved, this is definitely small potatoes), but it was quite a stressful experience!  Finding out I got that job today really cheered me up. No more depressing unemployment checks! No more wondering where my next step will be! No more small town angst from a girl that was made for the Big City!  As far as this blog is concerned, I guess I can just say that weight does NOT keep me from achieving my dreams. Weight is nothing - it\'s time for this to GO.  '),(1795,4031020,'08,August,2004','2004-08-08 00:00:00','   urlLink    This is me on my cruise to the Bahamas in July 2004. In all my 290 lb. glory. I\'m still cute, but check out that huge head and those arms! Well, at least you didn\'t get a view of my behind.... you\'d be blind by now. hehe  Actually, I don\'t think this pic is so bad, besides the fact that it\'s blurry and not very clear. I\'ll try posting another one that comes out nicer later. It took me a half hour just to get this one up on blogger so for now I will just admire my effort.   urlLink  '),(1796,4031020,'07,August,2004','2004-08-07 00:00:00','   Today is the beginning of Week 3. As promised, my blogger headings aren\'t going to be based on the week or day anymore. Mostly because by the time I get to like, Week 50, it\'s going to seem like a helluva long time.  I weighed in today at  288 !  So in reality I lost 5 lbs. since last week, but really 4 lbs. since I started the whole thing. YES! I\'m going to make Sundays my \"free\" day. I will still watch my food, but I won\'t record it. Just so you have an example, today I ate an appetizer and a chicken sandwich + fries with a mimosa at O\'Charleys, and I\'ll probably have some ice cream later tonight. So it\'s not so bad, but I just need one day of freedom, and since Sunday is my \"day of rest\" it seems the perfect lazy day to do it! I guess you will see the results of it on Monday\'s weigh-in.  I don\'t have any poignant stories to tell today. Maybe I\'ll look to other people\'s bloggers for inspiration, like journal de jude\'s (see Links, on the left). Ciao!  '),(1797,4031020,'07,August,2004','2004-08-07 00:00:00','   Weight: 289  Food: Oatmeal w/ tsp of Brummel & Brown Shrimp, Spinach, Mushroom and Rice concoction Orange Fish and Rice Another Orange  Lazy Saturday here. I\'m proud of myself - I had a hankering for some fast food today, but instead I just went to Google and typed in what I had in the fridge (shrimp and spinach) and found a recipe to make something with it! It turned out terribly - too much lemon juice in the recipe - but you can never go  too  wrong with shrimp.  Tomorrow is my weekly weigh-in. Hopefully, I can maintain the 3 lb. loss for the week OR shed another pound or so. But I ate like a pound of that shrimp, so who knows? We\'ll see.  I can\'t believe I\'m eating normally and losing weight. It\'s not that much (yet), but I\'m thankful for every little bit I get. Sometimes I\'m tempted to starve myself but then I decide against it, and it makes things SO much better. Lesson #1: It\'s much easier to eat well when you\'re not STARVING!  '),(1798,4031020,'06,August,2004','2004-08-06 00:00:00','   Sorry, had to combine two days here! Stats for the last two days:  August 5  Food: Baked potato w salt, pepper, Brummel & Brown spread Baked fish with diced tomatoes, peppers, olives and lemon Basmati Rice SIX \"Lite\" White Chocolate Almond Bars (yep, that says 6!)  August 6  Weight: 289  Food: Sushi Same fish as above Basmati Rice The other six White Chocolate Almond Bars  Well, I\'m living proof that you don\'t have to starve yourself to lose weight! I\'ve lost a grand total of 3 lbs. so far and it\'s been just under two weeks. The first week was horrible, I\'ll admit, but this week all I\'ve really done is cut out the junk. I pretty much just told myself that I won\'t eat anything out - I\'ll cook everything. I achieved that for the most part - I was in Atlanta and HAD to stop by my favorite sushi restaurant on Wednesday and today. Nothing abnormal about that.  I\'ll let you in on a secret. That visualization stuff works. I\'ve ALWAYS seen it as such a crock. Visualizing myself thin made me want to eat, just like everything else I couldn\'t instantly have. But I\'ve been using it differently lately and it\'s highly effective. Whenever I tell myself no to food, it\'s very easy to accomplish. I just imagine the undesirable food as my own fat. As if, someone came and liposuctioned the fat rom my gut and slapped it in an ice cream cone, and I\'m supposed to lick it. I don\'t know about you, but that makes me squeamish.  This week I\'ve visualized yellow fat McDonald\'s hamburgers on a thick, white fat bun, globs of cellulite in a Bruster\'s waffle cone, and breaded, fried fat on a Chick-Fil-A biscuit. It was really horrible. I actually made myself ill a couple of times - I have an insane imagination. Now that I\'ve put it to use \"for the cause,\" I think I\'ll keep it up. I think the clencher is that with each gross food I crave, I imagine exactly where the fat on my body came from that I visualize. And whenever I think of my huge behind, that always sends me over the top. Gross.  I\'m glad I\'ve lost a few real pounds - that jump to 288 from last week was totally fake. It was totally because I starved myself the first day like an idiot. I\'m losing weight even better by eating as much as I want right now. I may ACTUALLY prove that a fatty can lose a massive amount of weight in a healthy way AND a short period of time before this is all over. That\'s the goal anyway, right?  Today I completed my second interview down in Atlanta. I find out if I got the job on Monday. Wish me luck!!!  '),(1799,4031020,'05,August,2004','2004-08-05 00:00:00','   I couldn\'t post to blogger last night for some reason, so I\'m just going to put yesterday\'s stats:  Weight: 291  Food: Banana Sushi Miso Soup 2 pieces lemon pepper tilapia Basmati Rice FF Italian Ice Big bowl of grapes  '),(1800,4031020,'04,August,2004','2004-08-04 00:00:00','   Weight: 291  Food: Banana Sushi Miso Soup 2 pieces fish Basmati rice FF Italian Ice Big Bowl of Grapes  I\'m still eating quite well - it\'s a real shocker, huh! I guess I just had to get through Hell Week, and then I start doing fine for myself. My weight finally reflected my little tiny changes. In the morning, I\'m going to go walking again with a friend. It\'s probably going to kill me - my legs are so sore from yesterday!  I got a 40-hour contract job today that will probably spread over a couple of weeks. Thank goodness - takes some of the edge off unemployment which REALLY SUCKS. I seriously need some dinero and quick!  Not much to say today. Lots of work to do.  Thanks Jennifer for your comments on my previous posts! I\'ve been trying to post replies to them but for some reason my board won\'t let ME post comments. Weird. Anyway, you guys can feel free to email me too at  urlLink evian321@hotmail.com .  '),(1801,4031020,'03,August,2004','2004-08-03 00:00:00','   I think I\'m going to start changing the titles to my post to be more interesting. Maybe on Sunday - the first two weeks are my intro phase.  Weight: 293 (still? grrr!)  Exercise: 35 minutes, walking (wow!)  Food: Lemon/Strawberry Smoothie 3 lowfat egg rolls Grapes 3 pieces fish, 2 servings rice Fat free Italian Ice  Today was the first day in a while that I went to intentionally exercise. Woohoo! Believe it or not, I love exercising. I am a really active person and always have been, overweight or not. I like kickboxing, bicycling, aerobics, dancing, and pretty much any other activity there is to be done out there. I went hiking on some trail I\'ve never been on. I brought my bicycle too, but after hiking the trail I realized I need to get in significantly better shape to continue.  I had an interview today. I also got a second invitation for that interview that I mentioned the other day. Wow, I guess they really DID see my qualifications instead of my fat. One more interview and we\'ll see the results. First interviews are never a problem for me, but the second interviews are nervewracking because then you can start TASTING the job. I\'ve put much more time and research into it to prepare (I\'m meeting the CEO) and I really want it. I think I will imagine that I already have the position, and just go in there and claim it.  I\'m in marketing, so jobs are really cutthroat right now. But I\'m particularly interested in getting this position because I have a wedding to pay for in April, I have to travel overseas in January, and a vacation in October. I need to WORK. Mostly to save up money, but also to keep my mind off of it. My fiance is long distance so I don\'t like to have so much time on my hands - I get obsessed with thoughts of him.  I had an interview for another job today, too, but they are not sure if it\'s even solid yet. Bleh. We\'ll see about that.  Now, let\'s just hope the CEO sees why I am the best candidate for the position! And not just the largest.  '),(1802,4031020,'02,August,2004','2004-08-02 00:00:00','   Weight: 293 (Ack!)  Food: 3 LF Egg Rolls (baked, not fried) 2 packets Duck Sauce Bowl of grapes Baked Lemon Pepper fish Baked potato FF Italian Ice  Great news! First of all, today was my BEST FOOD DAY SO FAR. I\'m very excited about this. I guess when you actually buy food you like and keep it at home it\'s not so tempting to stop at Arby\'s or Wendy\'s or McDonald\'s or...well, you get what I mean.  Yeah, but what\'s messed up about today is that I weighed in at my HIGHEST WEIGHT EVER. Yep, I\'ve never ever seen the numbers 2.9.3. on my scale until today. Just lovely.  I got a random appointment for another job interview today. In the same industry I\'m trying desperately to get out of. I\'ll go, but I\'ll probably just end up recommending one of my former employees. You know, I guess I should count my blessings with this whole job interview thing. I mean, I may be large, but I\'m not a slob. And I live in the deep south, where there\'s a lot of fattys. People don\'t look at me and say, \"She\'s fat.\" They look at me and say, \"She\'s fatter than me.\" Southerners eat more than any single group of people on earth, I\'m sure of it.  It could be worse... I could live in L.A. or New York or Miami, land of the Beautiful People. I would never even get an interview there. I went to Miami last month before my vacation to the Bahamas. I was astounded at how many Beautiful People live there! They\'re all tan and thin or muscular and model-like. I could have spent my entire vacation there just staring at people. I went to South Beach and I felt like a whale in a fish bowl. It was still fun, though.  Sometimes though, the Beautiful People seem confused to me. I saw this one promotional girl, she was naturally gorgeous, but she was wearing leopard print pants that showed the crack of her ass, a halter top and so much makeup you could have raked it off with a serving spoon. Why do the BPs do this? There are times when I see a pretty woman wearing an outrageous outfit, like a catsuit or a miniskirt, and I think, \"Work it girl! If you got it, flaunt it!\" But for this girl, I wanted to say, \"You need to fix whatever part of your mind that told you that outfit looked good.\"  Anyway, I now understand why Will Smith had to write a song about Miami. I look forward to going back there!  '),(1803,4031020,'01,August,2004','2004-08-01 00:00:00','   Weight: 292, 290, 291  Food: 3 oz. ham Collard greens Rice Black Beans Fresh corn Kool-aid (Sunday Dinner at my sister\'s)  I actually did very well on food today - I just had one spoonful of each, I didn\'t overeat anything, and that\'s difficult for \"soul food\" meals. I wanted to start August off right.  If you\'re wondering about the weigh-in, I put all those numbers up there because occasionally my scale does that. I weighed once, then again, then again, and all three of those numbers came up. I didn\'t want to choose one, so I chose all 3. I think that just sort of happens when you\'re this large. My scale is confused... \"Heavy load! Tell her anything, just get her the heck off me!\"  Anyway, as I lose weight that will happen less and less. And yes, I do plan to lose weight.  '),(1804,4298922,'25,August,2004','2004-08-25 00:00:00','   You work your butt off all day You stress yourself to get your bossâ€™ O K You go home wrecked and tired Meeting deadlines to avoid getting fired.  You forget your family &amp; friends You prioritize your colleagueâ€™s comments You work hard to get your stuff done You work until late, â€˜til everyoneâ€™s gone.  You think youâ€™ve achieved something You think people care a thing Theyâ€™d laugh with you and thank a bit But thatâ€™s all there is and nothing more to it.     â€¦ a petal from a rose               there was a time that i felt like my work and everybody else was squeezing the vigor out of me. sucking up all my energy, giving me fake smiles and meaningless \'how are yous\'.. such a stage of crapness!       just penned my thoughts one morning soon after i woke up - t\'was just 6.00 am. then, i read this out at work during the morning briefing, everybody was there. some thought it was funny... perhaps sarcastic too!     oh! life...  '),(1805,4298922,'24,August,2004','2004-08-24 00:00:00','   i wrote this poem while i was in a relationship that couldn\'t give     me an emotional security. i loved the guy, but i had always felt      and known that \'parting\' will come. so i wrote this poem to      capture that moment where i was capable of saying to him -      \'i love you and i always will\'.  knowing that, \' that\'  very moment     will also soon pass once the parting is done.      Time will passâ€¦ Far away you may sail Thereâ€™s a chance that your memory will fail Distance and people might close our door, Then we might not be one anymore.  So, I wrote this poem for you to remember. That you and I once had times together That there was once a lass who loved you truly There was once a girl to you did things that are silly.  You will recall that once upon a time youâ€™ve been lucky You were given, you were blessed with a girl like me And might wonder once more about things that came our way Scrutinize from each scene, â€œ IF â€ you may say.  In time a smile in your face will also be seen And thatâ€™s when youâ€™re thinking of the places weâ€™ve been And the things we did - laugh and tease each other. â€˜Hope these memories shall serve as a reminder.  But before you reminisce things of our past, Think of me, my presence from somewhere donâ€™t cast. Never forget and be reminded still That once I told you, â€œ I love you and I always will â€.    ... a petal from a rose  '),(1806,4298922,'23,August,2004','2004-08-23 00:00:00','   I used to frown during the setting of the sun Iâ€™d always feel cold, empty and glum. I hated to see the sun coming down the sky I never liked the sun saying goodbye.  It seemed to me that Iâ€™d be more lonely, It made me feel alone with nobody I associated it with tears and goodbyes The reason why I hated to see the shine dies.  That feeling had stopped when I finally set In my mind I entrenched something good Iâ€™ll get From the setting of the sun today becomes past Tomorrow is another day at last!  Tomorrow sun will rise, hope shall it bring Everythingâ€™s new just like the coming of spring. Life we can start anew, new dreams to build Everythingâ€™s new; new fruits are yield.  I frown no longer during the setting of the sun I just smile with new hopes that tomorrow will hum For deep inside I know, after the sunset Hopeful sunrise the next day I will get.      ... a petal from a rose  '),(1807,4298922,'23,August,2004','2004-08-23 00:00:00','   Iâ€™d like to let you know, I enjoyed the times with you The talking, sharing Had set my heart aglow.  Iâ€™ve loved you that moment I know you did as well. I felt it so Iâ€™ll treasure In my heart it shall dwell.  Though it will come to pass Itâ€™s something special still It was a momentâ€™s gift To cherish and so I will.  If it was but a fleeting flame We found in each other I know the message that touched In our hearts will linger.  If tomorrow would choose For feelings to wane Let today ourselves admit That our hearts feel much the same.  If this may be the last That Iâ€™d be able to tell you Allow my heart to say That I really cared for you.  For now Iâ€™d leave this sad thing To tomorrow which can tell How everything will end And when the wound will mend.        ... a petal from a rose  '),(1808,4298922,'23,August,2004','2004-08-23 00:00:00','   I wrote this poem almost 8 years ago when i was feeling so down and felt like have been through a lot of ups and downs...felt pain, felt happiness, hoped for hopes and had to cling unto something to remain standing against all odds.      Iâ€™ve been through a long and winding road I have traveled so far â€˜til my mind became broad.  Iâ€™ve shed tears after the bliss, I cried after laughter. I have answered questions but continued to think and wonder.  I have traveled life, left a trail of joy and sadness Iâ€™ll never forget the pain I experienced after gladness.  If I would count the wounds, if I would feel the pain I would be so proud of myself to be able to survive the rain.  If anyone could only feel the harshness and the blow He or she too will surely grow.  This is my sad story This is my fruitful journey.     ... a petal from a rose  '),(1809,4298922,'23,August,2004','2004-08-23 00:00:00','   When the star we follow had gone into hiding As the rain cloud appears all smiling When crooked path seems the one guiding Itâ€™s the time that you mustnâ€™t stop trying.  When the dream may seem unreachable As go we go our way in chasing the ball, When things we planned wonâ€™t shape up This is the time to push hard, not give up.  When the dream you dream becomes impossible. Start again but donâ€™t you fall Dream good and dream anew For there isnâ€™t a dream you canâ€™t renew.  Hold on to your dream firmly and steadily For often it drifts apart unknowingly Youâ€™ll learn too late when time would come The dreams you made have all been gone.  Whatever may come let your dreams stand Donâ€™t ever join the failureâ€™s band Dreams may remain dreams for a while When they come true, truly worthwhile.       ...a petal from a rose  '),(1810,4298922,'25,August,2004','2004-08-25 00:00:00','   This poem goes to a gorgeous lass Whoâ€™s really funny when she gives out gas Too often than not she belches out Sheâ€™s one of a kind - without a doubt!   In the morning she comes wearing a smile Owing to the night that was worth her while. If you see her with a morning frown I guess because she was never out to town.   When she starts her work for the day â€¦. Better be quiet or out her way She could be an angel she could be otherwise. Sheâ€™s a real person, witty and wise.  Thereâ€™s so much to like about her â€“ Her singing, humming `and swearing like no other. In the office we could have hours of bliss Because of her and her phrase â€˜What the fuck is thisâ€™.  So there you go my partner and friendâ€¦. Hope youâ€™re okay with this, my life you wonâ€™t end. You are liked and loved the way you are. Stay that way â€“ lively &amp; bright like a shining star.     ... a petal from a rose '),(1811,4298922,'24,August,2004','2004-08-24 00:00:00','   loving every moment ...     never thought that someone could love me this much.... never thought that i could love and give so much of myself to someone.  after going through tough times of that \'getting to know each other\' stage,  sometimes we realize that no matter how loving, thoughtful, intelligent & fun that person can be, still he\'s not the one. then, when you u finally found that special someone, it\'s bliss! it\'s like finding that last missing piece of the puzzle! '),(1812,4298922,'23,August,2004','2004-08-23 00:00:00','   click on the title and get more out of this link!!! Happy reading folks! '),(1813,4298922,'23,August,2004','2004-08-23 00:00:00','   the bliss is contagious.... click on the link to see our very own wedding magazine! '),(1814,4298922,'23,August,2004','2004-08-23 00:00:00','   this poem was especially created for my one &amp; only partner in life who has given me all the love in this world - Jeffrey Ximo Ramos...the man who showed me what real love is!       Everything I have, youâ€™ve got for me. The bliss, the truth that you made me see. The love you give - the same you keep Makes me alive, my heart upbeat.  My world you coloured, enchants me well. For I see the rainbow and there I dwell. My life when you entered is made complete, Full of joy, contentment replete.  The man you are is a man of word Every promise fulfilled, I thank the Lord! The way you care, the most I adore The reason for going on, to have this amour.  My days are blessed ever since you came I seriously journeyed and played no game You helped me get through lifeâ€™s burden Offered a paradise, a Garden of Eden.  Myself so proud for I am with you The kind whoâ€™ll make me happy, never blue. This feeling for you never will I doubt, â€˜Coz you are the person I cannot live without!      ...a petal from a rose '),(1815,1969237,'28,June,2004','2004-06-28 00:00:00','   I was very upset about something the other day, so I went and bought a slingshot.I have no idea about the slingshot..I can\'t hit fuck all with it. I can hit whatever I want with a gun though..which makes me immeasurable happy. I so thought I had lost that skill. Honestly, it\'s about as happy as I am about getting back the ability to play guitar- like I found a peice that has been missing.     Anyway, I got to thinking..it was that damn \"Frontier House\" program that started me down this road to wanting a house. There is a scene where a guy is building a fence and bitching that he isn\'t getting enough food- and it turns out he needs to drink more water. Oddly enough, I have had a desire to build a fence or anything ever since that scene. It feels like everything in life that should have hit me at 33 is hitting me now. Damn wreck.    So I am digging into how to build a house, reading books and the internet. I have come to realize that the Jeep can\'t haul enough of a load to be of use, and that as much as I want some city car..I need another truck. Funny thing, I seem to have a line on a Chevy as of tonight...for about half of what I hope to sell the Jeep for. It\'s in perfect shape too..a retired air force guy here owns it. Weird stuff. It\'s like what you need happens if you let it.    Speaking of which, I have been all over the internet lookinbg at alternative buildings..meaning, for the most part, anything that was done for centuries before the lumber barons of the US decided to say it all sucked. I have looked into Solar, Straw Bale, Cob, Adobe, Earth Plaster, Rammed Earth, Log, Slip-form Masonary....pretty much anything that\'s been done. I am looking for cheap but effiecent with little upkeep.     So Adobe is cool, good for thermal issues, but lots of upkeep. I have stumbled onto Straw Bale construction. It is simply bales of straw, stacked as walls..and then plastered to protect it. You can stucco the outside and make very fluid shapes. Weirdly, a freind mentioned that a friend of his back east built a house this way..and now..it seems like every damn magazine I pick up has straw bale write ups in it!   I mean...mags on living green, mags on home power...its like the world is saying \"LOOK HERE DUMBASS!\". I have also found colored stuccos that remind me of old Italian villas...so much cooler than normal paint. Anyway....  ...today I was going to drive to Forest Falls..but went past it and made a grand loop all the way through Big Bear. During the drive I was thinking of how much I really don\'t like the whole \"Southwest\" type building. It\'s OK..but not for me. Then it struck me...most of the construction is very much like old Italian villas...! So I began to surmise that for a floorplan and exterior look I could do a straw bale house for many modern reasons, with a villa look to it. Then, I drove past a lot- and had a major deja vu flash from years and years ago..I pictured a villa on this lot near Mentone I think it was- and here I was driving around thinking about how that would fit in to the area well....weirdness. It was like the general area, the lighting, everything, had crawled out of the dream like 15 years ago. The rest of the dream had a wife and a young mistress living under the same roof and about 4 kids. So I doubt it is actually a sign of a true thing that will happen...;<>!    So, as I study, it seems like log construction, while useful in this area, is still more expensive and can\'t have the overall effect I like best...while straw bale can.I suppose I should work towards a villa...as I could probably build the first section to live in, and then build the other wings later. In the court yard I would simply HAVE to have a mosiac layout of the Italian Raiper circle invented by Thibault.....which, of course, would force me to get off my ass and actually learn the system....     But building something like that HERE? I am not sure. It\'s where my income is at now. I want a place by the beach in the end. Maybe this is just a step I need to take.     Most of all..it seemed that no matter where I went looking for eco-friendly stuff...sooner or later I got linked back to Findhorn...! How weird. I am much in turmoil tonight.....  '),(1816,1969237,'26,June,2004','2004-06-26 00:00:00','   I have lived in apartments for far too long. I am in Southern California, where I moved some 20 years ago to pursue music ....which, after a car wreck in which I broke my neck, vanished as a choice in my life. I am stuck with chronic pain and memory problems, including speech problems. Most people never notice the speech problems, and the condition itself is known as Aphasia. I was a very intelligent person before the wreck, if a bit young and unfocused (hey..everyone deserves to be in their youth!). The wreck was in December of 1988 and took away any ability to play guitar I had worked so hard at. It makes it pretty hard to be a musician when you can\'t make music anymore.  Overall, I am in better shape mentaly now than I have been in many years. I have figured out the limitations of the injuries and begun to work around them. I have had about 15 years of my life ripped away by the problems it caused, and many of the issues I need to address are normally solved by people about 15 years younger than me. This is more than frustrating.  I can\'t seem to find work that is worth my time, so have become self-employed. At least I don\'t throw up every morning from the stress of submitting myself to a job I hate anymore.   I want a house of my own as a base to work from and get whatever I can done in this life. Cost of buying anything at all has always scared me away. I have come to the line of thought that I should do everything ,myself...so I am planning on purchasing land and building my own house. I don\'t mean paying contractors to build a house for me, I mean building it myself. In the end, the entire project should yeild an investment I can take loans against and the like...enabling me to pursue other interests.  I want to live near the base of the mountains nearby, as I seem to have a decent business started. I would rather live near the beach, but I have to start somewhere. I also am working on wrting novels, so once published look forward to something of an income from that. I am enough of a whore to be thinking in terms of the fact that its rather useless to write a book for arts sake, it IS a product after all, if I intend anyone else to see it.  I\'m single, actually divorced, my ex is a friend who I stay in touch with and worry about as I do all my friends. She is a good person, timing and life were wrong for us to have been together. I don\'t like being single and want to find a woman who is artistic and has life goals. Maybe a writer, musician, artchitec or whatever. As for looks, I like a lot of people for who they are, but I would be hopeless to find a woman who looks like PJ Harvey or Joan Jett...someone skinny with dark hair about 30ish. Someone with drive like these two women would be a gift from the gods.  As I figure out how to work this blog I\'ll add links to useful sites about weird building techniques I study. In the end, hopefully it will become a record of how I did it, and I can turn it into a book for others who are a bit lost and don\'t know how to ever afford the concept of a house- or who want something outside the norm to live in.  '),(1817,1969237,'11,July,2004','2004-07-11 00:00:00','   I haven\'t been avoiding this blog, I have been out doing research. Not exactly what I wanted to find in the end.  In short, alternative construction isn\'t really all that alternative. Its mostly based on stuff that has been done for centuries, but has fallen out of favor in the USA because of the cheap and easy availability of wood.  One straw bale book said \"If you are looking for cheap housing ASAP, then straw bale is not for you. Buy a mobile home and put IT on a peice of land, it\'s the lowest debt method of owning housing.\" OK, I can see where that is correct.  A student turns out to be a contractor. I love talking to him, he\'s so cool. He told me land costs have gone up to like 5 times what they were even a year ago. What that means is I missed the boat. I can\'t AFFORD to live anywhere at all in Southern California now. I am thinking that the high desert may be the only option..before the building craze wipes it out too. I don\'t like deserts...but have to admit the air is cleaner, and with no mountains to lock in smog it will probably stay cleaner even with a zillion people there.   Alternative housing means trying to do the best for preserving the area, so if you have to commute a 100 miles a day, it becomes rather pointless anyway. It looks like my optons to do something good AND to have a comfortable home are slimming quickly. I beleive, in the end, that I need to build up my business, get it frnacised so I can travel and do set up, and live where ever I feel like. Probably somewhere in Northern California. I am not comforatbale in leaving the area business-wise yet. There is a great potential here for income. If I can get a bubble of decent income for the first time in my life I can get ahead enough to see the tunnel there is supposed to be light at the end of.  '),(1818,1969237,'26,February,2004','2004-02-26 00:00:00','   I have been using the Palm Pilot I got to set taks and appointemnts. Much of the stuff I had to do is melting away. I am off to projects to make better my business. I placed an ad in a local business journal today, I was worried about spending the money, but you can\'t tell until you try....    I went to get my haircut, I love my hairdresser, she does such a good job. I also love my hairdressor because she is hot. Skimmy, sexy in just the way I like and intelligent in one pckage. She also drives like a maniac, having totaled a car after I last saw her..and she was recovering from the wreck she had just had. Tos ay I want to sleeep with her is a start. This is one tough woman, she could handle my crap I do believe.   One of my best friends called an hour ago to say he totaled his sports car. My other close friend did his a few months ago..my hairdressor has done it twice..I don\'t want to join this club any time soon.  I have been doing repairs to my Jeep so I can sell it. I am stuck for one major tool I need right now, but I hope to get things rolling soon on it. I have decided to do airbrushd graphics on the Jeep t raise the value before selling it. I don\'t know why I never thought of it before....  Need to get back to the book...  '),(1819,1969237,'19,February,2004','2004-02-19 00:00:00','   ...since I put anything here. What have I been doing?  I have decided my 9-5 job is a waste of time, and started the wheels turing to get another one. I am aiming at being a writer and fencing coach (I am a coach to pay the bills now), with a minor interest in music. The job I am looking at has been paying a friend who used to be at my current job quite well. He is happy. I certainly could stand enough money to survive again.  I have decided all women all annoying as hell. I have also decided maybe I should never have a family. I have always held women to be equals, but have received nothing of the same sort of respect back. I am not dating until I have money in the bank to do what I want to in life. I\'ll be too old to date then anyway. I will be able to travel. Perhaps it\'s American women that are such pathetic losers. I don\'t want women as friends anymore either. I am tired of being father confessor to them. Maybe I should eb as abusive towards them as I see most men. I have been told to never change, that someone right will come along. It isn\'t working. To hell with all of them.   It\'s all about money and that\'s it. Money can buy happiness, because Donald Trump never seems to be depressed.  I am writing a book.  I am completly repairing my Jeep so I can sell it.  I am compleltly repairing my Escort so I can sell it also. Proceeds will go into business advertising and the Miata. Oh, and travel.  I am reformating the tower so I can get the music project going. I just want to complete my goals, I care not is anyone else approves of them.  I need to go to another country. or two or three.  '),(1820,1969237,'05,February,2004','2004-02-05 00:00:00','   Looked at last year\'s blog and it was nothing but pathetic whining. I got rid of it, and with it all chance to llok at all that whining. Now no one sees my blog or has the address to it..ever again...unless they meet me THROUGH the blog.   It\'ll be sort of a daily journal of what I did with some related thoughts- as the last one just went too far off..and was whining.  There are books to write, lesson plans and tests to write and 9-5 stuff to get done for real. Let\'s make getting LIFE done the job...  '),(1821,1969237,'31,March,2004','2004-03-31 00:00:00','   .     Well..I finally got the title to the car I bought in december. What a nightmare. I went to smog it this morning and thought it was going to fail. It finally passed..which is good, otherwise I can\'t get the title transfered into my name...     That done I blasted to DMV, who managed to find even more problems this time. Every time I go they find something else that wasn\'t done, can\'t be done or whatever. I just stood there and looked pathetic until they sorted it all out and I got my tags! Once on the car, it finally felt like MINE..even though I paid for it months ago. It was like having a lover who is fucking someone else besides you..you never know where you satnd. That\'s the worst kind of stress for me, and why I got sick and gained weight. Now all that\'s left is the tax mess from last year.    I traded in my cell phone today also, poor old warrior is retired. I got a new flip up one that is a bit smaller- and I can feel more like Captain Kirk with it. The ring goes off and it says \"You have an incoming phone call\"...I have to figure out how to change that, or begin refering to it as the HAL 9000. It\'s too erie.    I did something else major today but have forgotten what it is already. I am too busy. I wish to hell I had a wife or girlfriend to lean on sometimes. Just someone to chat with and snuggle up to on the sofa. Trouble is, women are afraid of commitment. They don\'t know how to do it. They all want a puppy that comes when they call, not an equal who has thoughts in their head. Maybe I should finish up the Jeep and the other car and sell them both and move to wherever....    I need to get money to a friend to get his old header and muffler from him. That will boost the HP of the car about 15% or so. I need all the power I can get. I had fun last night when some college geek in his VW Turbo something or other thought he was gonna leave me behind. Whenever he sped up, I stayed the same exact distance behind him. He chickened out at 110 mph. Loser, daddy must be paying for it all, because no one that young is driving a car like that and paying for their own college. If I had that kind of money, I wouldn\'t be going to college, I would be starting the production comapny I want to have.....  Sex.  '),(1822,1969237,'30,March,2004','2004-03-30 00:00:00','   .     It seems like the week or weird problems already. I had a claim denied that seemed perfectly legit to me. I\'ll have to call and see if I can push it through. I also had a call that the karate class is pulling out of the Riverside location, so I have been offered additional nights. I am thinking I should move the failing Epee class to Wed nights, and in the second hour add a Sabre class. I have to be able to balance this against the other classes...and not undercut them. I guess we will see what we can do. I had considered burning down th Redlands class, but it DOES bring in SOME income..no matter how frustrating it is. Then again, giving up Wed might fill out Riverside to the point thatiI CAN close Redlands.    Awoke to mad dreams of my girl in LA. Don\'t know what to think of it all. Do I have time to chase her? I doubt it...then again, she hasn\'t worked in a year or more- living off money from her mother\'s passing away..and the sale of her house I think. I need my life to move forward. One thing though, I haven\'t scared her away like every other female I ever meet. Maybe that means something too.    Mad Racket....the words keep popping back into my head. A song by Krokus back in the early \'80\'s.    My title for the car should be here today- if so I get to pick up my tags also. To have this all done will be a godsend, I hope it works out soon!!!     Cross my fingers for change...  '),(1823,1969237,'29,March,2004','2004-03-29 00:00:00','   .    Well, it\'s time to get moving again. the weekend was a diaster. As ever, the divsion took forever to run a most simple competition which destroyed my weekend timewise.     I drove some today. I almost went to LA, as my mind has been full of a certain woman there. She is a friend I have always want to by with. I would die for a kiss...people die for much less most of the time. The dreams that wake me about her are back, so I better call her and make sure she is OK.     Time to get organzied again and get rolling. Need to FINISH the ever-increasing Jeep Project and move on it also. The apartment is begging to be finished, and I have hit on a design direction after much watching of the Home Version of Extreme Make-Over. I love that show. It finally struck me how much I need a house...to have something to work on as therapy.....  I would if I can write a house off as a medical expense?  '),(1824,1969237,'27,March,2004','2004-03-27 00:00:00','   .  Well...it appears this is working again. I have no idea about html, but figured it out by reading it line by line, hitting preveiw and seeing what changed what. Life is too busy....     I have blown off working at RenFaire this year. I just don\'t care to tell the truth. It\'s too tiring, and I already teach fencing 4 night s week. The last thing I would do to relax is fence two more days during the week. The Long Beach Grand Prix is the same weekend as Faire opens, guess where I\'ll be?     Today is an air show, the Redlands Bike Classic and a Competition I need to coach at. It was also the yearly booth meeting. What kind of fool runs a fencing booth and then schedules a mandatory meeting on the same day as two differant divsions yearly qualifing events and then expects people to show up..on pain of not getting the yearly pass? I decided that was the last straw. How stupid....the people that work the booth, by and large, are non-fencers...with an inflated sense of their own abilites. One of the current head people..who has less than half the years I do in...has never fenced a competition..yet thinks he is too good to go to most clubs around the state. What an idiot.      I don\'t even care about chasing women there this year. The older I get, the more I only find women with 4 kids by 4 differant guys- who want someone to pay all their bills so they can sit home with their drug drenched kids and bitch and complain. Then they feel like never having sex because \"They have rights\" or something..??!?!?!?!? I think maybe I should buy a useful and friendly younger woman from over seas somewhere..someone who would apprieciate just being cared for as a person...    The overbearing stillness that felt like death has moved on. It could be the Overbearing Noise of Freeway Construction that has shattered all calm for weeks. Or the noise of the high school next door. Now that they put up the tarps blocking veiw of the feild next door I can no longer lust after the dykey looking girls coaches and their short hair and hot bodies...oh well...the Crazy Cat Woman loves it because she thinks the police are not spying on her anymore. Dumb bitch. Why can\'t I EVER have sexy and horny neighbors?     Jeep projects went well until I burned the brand new paint off the sides with the brand new bufer. Oh well, time to start over.  '),(1825,1969237,'19,March,2004','2004-03-19 00:00:00','   . I have managed to screw up the html on this page. Don\'t know if I can save it. I do a preveiw and everything looks fine...I jump the link to this page and the writing is overlapping itself....maybe if I post something the template will be updated???  '),(1826,1969237,'19,March,2004','2004-03-19 00:00:00','   .           Morning pulled the veil awake and Death has slide back to the dark of tunnels, of pathways, of behind trees and under things. Death is not resting, Death is just waiting. Death peers out from where He sits. Death peers out from where She sits. Will cats wander in and be crushed? Other animals more aware stay away.   Yes....Death is STILL Close At Hand.  '),(1827,1969237,'18,March,2004','2004-03-18 00:00:00','   Death stalks about tonight. He pranced in at nightfall and I felt him smile over me in Riverside. I felt him spread out over the city hiding behind the veil we call darkness, and nobody saw him but me. I am not depressed in that way, I do not long for my own end or others- but I know his face, I know his talk and I feel him breathing on me again.    Something big is about, something bigger than most times. Something that is a misbalance in the very force of life, allowing death his waltz. I have no words for it, and facing it, find myself much more calm than I could ever imagine myself to be. Maybe it is come for me. Maybe not, but it has come.   There is so much I have yet to do, so much I have yet to see, but instead we must dance yet again. I am well past hate with this mental exhaustion. I know Him too well. I drove his old hunting grounds and he was looking at me, puzzling against me to scramble my thoughts.He does well doesn\'t he?  It has been so still, not just the freeway which has been exploded in workmen\'s noise all night for a week or two...but the very earth itself is still, waiting for more death to drift towards it and refresh it with bodies to digest. I hope someone survives this. I wonder if things come to soon because something has been set akimbo somewhere else?  I can look and almost see everytime I turn at night. Like some translucent monster with no face but a hulking body Death sits there, ready to twirl away and make me wonder. Ready to make me search the papers for his night journeys. Werewolvian in his pursuits. Hidden during the day and breathing and alive at night. Making the peacefulness wicked..no...clear. Something to fall through on the way to somewhere else.  I feel like I need to be in India or some place balanced with beliefs. Someplace spinning madly out of time with this world, where radio is the only contact to keep you alive.  I feel so lost. But I feel...Death..closer again at hand.  '),(1828,1969237,'18,March,2004','2004-03-18 00:00:00','   I feel death close at hand. I myself was struggling for breathe for the last many weeks. I had another blog, actually two, that I got sick of.In one way I am glad I dumped them both. They seemed to revolve around meeting a woman, falling madly in love with her and then being rejected by them. I am everybody\'s best friend. I need to be more of a snake in the grass. Women think they want a nice guy, but they really want a devious bastard who would fuck their sister in an instant. Then thye have a goal of keeping him at home with their charms. If they know he is going to stay there already, there is no challenge.    Why can\'t there be women who want something more than Jerry Springer Trailer Trash Lives? I used to think they existed. The nutcase next door is letting her cats out all the time and acting like they are not hers, she has a constant stream of middle age black men going to her place (she is skinny and white) at all hours of the night. Just try and tell me she isn\'t dealing drugs again....    So much for that. Something bigger is in the air. The whole earth feels like it, like some grand event is unfolding and the earth is breathing litghtly to be ready. Perhaps the ascent of mankind to a new higher level? Perhaps the end of mankind? All I know is that the only words I have is that I feel Death Close At Hand at every instant.    I am working away on many projects, I am down to like 4 important ones. Oncethey are done I can breathe deeper myself. Travel would be good, my restless roots are bound in a pot lived in too long. I may blow off this weekend and drive as far north as I can to see how far I can get. I may not. I don\'t know.    Faire is coming, and I really don\'t care this year at all. I said yes to working, maybe I\'ll quit half way through. That would be a change, taking the control of my working back away from the booth owner....    I have airbrushing to do, a Jeep to finish repairs on (it\'s getting close to being only small repairs..YEAH!!!), paperwork and taxes to do (oh crap...), and testing to write. Sex doesn\'t even enter into it right now. I don\'t know why.    My stomach feels better for the first time in months with the addition of acidopholis to my diet, and a severe reduction in the amount and types of food I eat. I am eating one meal a day is all. I was eating too much crap.    Didn\'t try and pick up a bimbo, instead I started reading a book by Arundhati Roy. It is densely written and confusing, giving glimpses of things and then thrown al out of chronological order. All the things they tell you to never do in fiction. Screw \"Them\". I write much more straightforward than I like to. I write like Hemingway, like a newspaper man who wants more. I don\'t really like my writing ...yet.   Write On.  '),(1829,1969237,'07,March,2004','2004-03-07 00:00:00','   I see everything and I want to paint it...not black..but a touch wouldn\'t be bad. I am doing much bodywork on my vintage Jeep so that I may seel it. I intend to go to Scotland with some of the money. I also think maybe I should go to like Thailand and sleeze around with any girl who breathes. Why not? The president can blow up half the planet and act like it don\'t matter? Sex sure doesn\'t in the long run...!  Until then, I study and learn about airbrushes and come up with mad schemes to paint everything I own.  I went with a friend who wanted to flirt with a waitress. I find her attractive, but heard she had a BF. She said something nice to him, and he choked completly..leaving ME to carry the conversation. Now the woman and I are exchanging CD\'s and the like....it\'s not like I planned this....she is, however, highly attractive. I have found she lives with mommy, and she is a bot old for that. She doesn\'t seem to want to do so but is. That tells me Mr BF ain\'t all that much. Maybe I should go along farther. Friday I was going to ask her to coffee, but timing was bad. I will yet. I have to finish so many projects quickly! I need time and space to move forward.  I have discovered \"Beaming\" on my Palm. I now am a danger to the free world.  Wonder if I could pick up a bimbo in a pub on a Sunday...?  '),(1830,1969237,'29,April,2004','2004-04-29 00:00:00','   .    OK..pet peeve. I am SICK to death of hearing, \"Well so and so has a DEGREE in (Insert something uselss here...Biz Management, Admin, Logistics, Accounting, etc) so they are more qualified for any job than (Insert name of someone with years on the job who KNOWS what they are doing..).\"     The last manager I had was actually pretty damn cool as a person, but after years as a salemen, he got a degree in Logistics and decided he knew how to run warehouses and production lines. When presented with real world supply issues control by companies other than our own, he simply expected them to change to suit our disorganziation....    Worse is the College Student themselves. The NEXT time I hear some wet behind the ears child going on at length about how everyone on the planet deserves as much as everyone else I am going to level them. I am going to jump into their loudly put so everyone can hear them conversation and make them the ass they are.    WHY is it that these rich kids..I mean, going to school instead of working is not a RIGHT..and MOST people never get a chance to do so....think that from their rich little back grounds they can tell everyone else what to do? I will, and mark my words..NEVER hire a college grad at any of my companies. I want someone who was working, fighting, clawing and learning in the REAL world to learn what my company does..while THEY were going to frat parties, getting bombed or getting preggars.    Another good one: \"There should be a cap on what EVERYBODY makes..so money can be spread around to take care of everyone..\" Great..no problem, let\'s set it at about 30k a year..what? Can\'t afford college now? Too bad, somebody with 6 kids is sure to have the seventh child never find itself hungry again...yu don\'t NEED college anyway. Want a BMW and can\'t afford it now? So...someone else doesn\'t have to worry about paying for health insurance because YOU did it for them. What do you mean you don\'t feel like its worth going to work? What about everybody else? They need YOUR money..get off your ass and get to work!    Darwin was right- the struggle for survival is SUPPOSED to weed out the loser from the gene pool. When you upset the way of the world you end up with morons like GWB at the TOP of the food chain somehow.     A few years ago a guy walks into my Fencing Studio and announces he wants to take Fencing so he can beat everyone in his SCA group. Then he tells me that only the SCA really knows what they are doing with edged weapons. THEN, he asks if I can give a poor, struggling college student a discount so he can afford to take lessons from me.       So I looked into the empty parking lot at his BRAND NEW BMW, looked at him as he kept going on and on about what HE needed, and how he was going to leave as soon as he beat his SCA\'ers at their thing and said, \'SCA? So if THEY are the only ones who know anything about swordpaly, then WHY the hell are you here? Doesn\'t that logic rather negate your last staement..and as for  being a POOR college student...if you are going to college at ALL you are a fucking rich kid and won\'t admit it. Next..you\'re driving a brand new BMW worth more than everything I won put together...how DARE you come in here and and try and tell my you are either POOR OR STRUGGLING. Sell your fucking Kraut car if you ahven\'t got money...and DON\'T EVER walk back in my fucking door..now LEAVE before I throw you\'re two bit ass out in the parking lot...NOW...!\"    Of course, the business partner at the time, with her BA in Management (who couldn\'t produce an Income/ Expense sheet in 5 years..didn\'t understand advertising, hadn\'t a clue about building  a set of functioning lesson plans around our rented time schedule..and spent extra money on an 800 phone number we didn\'t need...) was VERY upset with me.    It\'s simple..public or private, schools are in business to sell an education. If not enough people make it to the end..they lose money from drop outs. If the person comes out at the end with no clue about their chosen field, no problem..it\'s brushed off as some people just don;t make it in their chosen feilds..BUT..student laons can never be bankrup[ted out of....very nice for them...!     A few centruires back, if someone needed a skill that required special study, there were Guilds to educate and certify them. They simply didn\'t LET people who didn\'t know what they were doing get cetified, because it made them ALL look bad. Somewhere along the line it got mutated into a College System...when Guildsmen discovered that the selling of an education was a business in itself.  Ban Colleges, bring back Guilds!  '),(1831,1969237,'26,April,2004','2004-04-26 00:00:00','   .     Just barely. All it seems I have done is sleep. I have to go to work tomorrow during the day, what a drag. Maybe I\'ll feel better by then. Got word my father is in the hospital, but doing OK. Gee, maybe when he went in 3 days ago it would have been a good time to call me.....     I am going to Faire this weekend with an ex fuck buddy. I think she wants sex again, but I just don\'t care one way or another. I am so sick of all the crap that it takes to be involved with someone...I wonder if I\'ll ever run into the \"right\" person. As for now..I have no hope of it at all.     Patti Smith was just on TV. That\'s the kind of woman I always hoped to meet, a smart, artistic and productive woman. A tough woman...not one of those fluffy doing crafts at home and calling herself an artist type that seem to abound.  Oh well, need to get rolling on life for the day.  '),(1832,1969237,'26,April,2004','2004-04-26 00:00:00','   .       This has been a pretty bad bout. I get grumpy when I am sick, I just want to get past it. If it gets bad enough, I feel crappy so I get wholesale angry at everything- as if to drive away the sickness. I have been throwing up since Thursday, couldn\'t sleep at all. Then I woke up Sunday morning bright and early wired like a speed freak with a couple of extra doses.     I got so much done on Sunday it was bizarre. I even got to sorting out 25 years worth of handwritten short stories, poems, novel drafts and god only knows what else. years ago I finished a novel I thouyght was a good idea, but written rather clunky. I had a woman (the same bitch I was angered about the other day) who INSISTED that she take the handwritten drafts and enter them into a Word format. This way I could edit/ publish..do something with them. SIX YEARS later when I got sick of getting screwed under (I may be slow, but a major head injury does that to people) and stopped putting my dick in her, I DEMANDED the work back. All the while she had been saying how she got a little time in to work on it here and there. I had expected a file with some 150 pages entered. She returned only FOUR of FIVE notebooks, and then INSISTED I was wrong about the count. Last night, I finally got a program to read whatever bizarre program she had used to make the floppy of all that hard work she did in SIX YEARS and discovered a TOTAL of NINE PAGES had been entered..and..OF COURSE, the missing notebook is the FIRST one...with all the set-up for the whole book. Lucky enough, it\'s the only major part of the novel I remember by heart and can rebuild from memory.    The OTHER problem is, when I ran out of paper during another writing project, I wrote one third of a long short story on the BACK side of the pages from this story...so this stupid, selffish bitch lost the intro to one novel, and a third of a short story..and had the nerve to tell ME I was wrong on the count. Christ, how stupid would I have to be to write FIVE binders full of stuff and think it was only FOUR....?!?!?!?!?     Iam SO tired of seeing TV shows where every woman got the short end of the stick, and every man is a loser wioth no brain. I used to hate Alannis Morrisette because of \"You Ought to Know\"...I know a goodly number of fairly straight forward guys who got major screwed by women. You know, married 27 years, she goes back to school on his dime, meets some younger doper/ fake artist type who beds her..so she divorces and takes half of his income...even though she hasn\'t WORKED in 20 or more years. I will NEVER marry a woman that doesn\'t work. If she doesn\'t, she is no better than a domestic whore. She better be ready to be on her back at a moment\'s notice...because she is selling HERSELF short to be a \"Stay at home\" mother. Before we got to this place in the world, mothers had to work all the time too...forging for food, making shelter..doing more than banging the pool boy- eating too much and driving a wasteful car around with a number of bad tempered, ill natured children they aren\'t really guiding.    I want to see a TV show where some stay at home mother who never did anything LOSES the custody of the kids because she is no better than a vampire, and then gets NO alimony, because HE needs it to raise the kids. I could do a damn site better job of raising kids than most of the druggies I see in this area.    Yeah- maybe I AM angry. Maybe having two years of work lost, and then getting told i was full of shit makes me this pissed. One thing I know- I never thought this novel was good enough to publish- but I NOW understand how to edit it and fix the issues in it...and I WILL get it published to spite this bitch among other things........I will make good on my promise to myself to be a writer.  '),(1833,1969237,'25,April,2004','2004-04-25 00:00:00','   .  I need music. I need something to listen to. I need something new, that doesn\'t sound like something old. I need something olkd only as a matter of record and archives. I need music. I can walk into Amoeba and in a place THAT big still find NOTHING of interest to listen to. I think I need to go to Austrailia or something. Or maybe ..no...I won\'t say it. It would just curse me from ever having it happen.  I wonder what my friend who is in Cebu is doing now...probably another nurse. I wonder if he will come back, or if he will move there and work his job via the internet as he probably could.  '),(1834,1969237,'24,April,2004','2004-04-24 00:00:00','   .   ...I just don\'t know what I want anymoire. All I ever wanted was to play guitar, make music, tour and live a simple life saying nice things overall. Instead, I got a lifetime of jobs I hated, more injuries to my body than half a football team and nobody to lean on next to me on the sofa at night. Everyone who knows me goes on about how nice it must be to be single. Great....I look that happy?!?!?!?  The grass is always greener on the other side. When I was married I got a number of the injuries, so it never really had a chance. My wife is very intelligent, but in many ways simply not my \"type\". I hate that fact. She is a really cool person overall- and the best thing I managed to do this life so far was to leave her...she finally had to work and stand up for herself when I did..and she is exactly as solid as I wish she had been when we were married. I wish she would find someone and get married again...then again, she is from a big family and all she ever wanted to do was be left alone. Doesn\'t work too well for an only kid who just wants a touch of attention......life sucks.  Life would be much better if I didn\'t hurt from head to toe for god only knows what reason. I wish I were in the south pacific or someplace...with warm salt water all around..and I could just swim and soak these pains away. I miss not living near the beach.   really should be working on writing a novel..not a blog...  '),(1835,1969237,'24,April,2004','2004-04-24 00:00:00','   .    It\'s like 2 in the morning, I live less than an 1/8 of a mile from a freeway- and for WEEKS they have been blocking it off at 10:00 pm, just in time to screw up my drive home after work...and then fucking working on it all night. NO ONE can tell what the hell they are doing. They are cutting huge slabs out of it, and then filling them in with concrete before morning.     CalTrans has got to be the biggest bunch of idiots on the planet. EVERY project they do takes FOREVER. I am really wondering why the hell I ever left Florida...oh yeah. before I broke my neck I played guitar and sang and was working on a band...DUH...figures something on a road like a car wreck would destroy my dreams. Probably CalTrans fault for the way the road crowned so high in the middle and then almost curved off into the curb...other wise the damn other care would have hit us BEHIND the front door I was sitting behind.    Ever break your neck? No..not front to back the EASY way...sideays, by smashing your head through the sidewindow of a car, and then bouncing the same head off the hood of the OTHER car as it crumples up and throws your carcass back into the car you were in...Yeah. It fucking sucks.    I need more. I need a house. If I can\'t find a woman and a family, I need a fucking place to build cars. Maybe I\'ll drift one off a cliff somewhere and go out in a blaze of glory. Am I 20 and full of shit and talking  tough? No, I am 42 and tired of trying so hard at things I don\'t give a crap about- which leaves me too tired to do what I want to do....crew the USA. I wish to hell I had been born in Europe..I always have.    I spent most of last night throwing up, even when nothing was left. It\'s fatal, but I keep hoping something would just burst, and I would slowly fade out and be found blood-free and gone at some point after I hadn\'t paid my rent. My stomach hurts all the time and has for 6 months. I can\'t eat anything without getting sick, and I have no insurance to go get it checked out. It makes no sense at all. It\'s what\'s putting me in such an angrey and frustarted mindset. Haven\'t I been sick and injuried ENOUGH in this life...???When do I get a break at anything?    Wish I had something high caliber with a scope so I could shut fucking Caltrans down. Assholes. I wish I knew where the project manager lived- I\'d go mow all his neighbors lawns for free...non-stop- until the project by MY home where I can\'t sleep is done. Beat THAT would speed it up....  '),(1836,1969237,'23,April,2004','2004-04-23 00:00:00','   .      God..I wonder what looked at me. I have felt like creeping death for days. Drove to Palm Springs yesterday for work, did 15 minutes, drove back and went to bed until I HAD to get up and go teach. Later, I was awoken by the overwhelming need to be sick.      There is no time lonleier than when you are sick. The last...bitch...because I can\'t call her a woman, I was involved with ignored my pleas for her to come over and help me when I was sick one time. Not a little sick, that never bothers me...but sick enough to barely be able to get out of bed. She gave me a list of all these important things she had to do and why she couldn\'t possibly come over. I nearly died that night ..I was puyking blood and just laid on the bathroom floor for hours, wrapped in a blanket and tring to get warm. I couldn\'t have possibly driven to the hospital. I still hate this bitch\'s guts.    It\'s alway men who are painted to look so bad. I am so sick of the USA, I don\'t care if I was born here. The women are mostly evil and heartless. They all want to be equal with men, but don\'t want the responsiblity of actually being equal. In other words, they want the benefits without the work. I hated working 9-5 jobs. The women were CONSTANTLY gone or LATE. It was always \"My kid this or that..\" Since I had no kids, OF COURSE, I was expected to work late, carry the work load and do what they didn\'t do. Then one of the bitches made such a fuss I got fired. That\'s fine, I hated the place anyway. As I walked out the door she actually said, \"They fired YOU..crap, I wanted them to fire RAY...\" who..of course, was sitting 8 feet from her.....     I have begun to think that women don\'t belong in the workplace. The boss at the last place had everything running great, and then we got forced into a room with a department full of women. His first words were \"This is the beginning of the end...\" and he began to look for other work. Good deal, as he was so right. He got fired...mostly because of a loud mouthed, over-weight, sub intelligent girl who thougt she knew everything..she NEVER stopped talking. he had screwed exactly one weedy looking ittle boy, but thought she was some kind of goddess. She managed to get him to marry her....what a fucking loser. Too stupid to even go whoring instead of getting latched onto by this vampire.  ...she was SSOOO proud of her...HONDA! She thought it was a race car because it had 17\" wheels. Gee..I wonder if she realizes that that threw off the suspension geometry so bad that it now handled worse than new. Srings, sway bars and shocks...? What are those...?!?!?!?!    Last time I checked, she was some sort of Jr manager at the place, and everyone was scared of her. She have a lifetime job there, because her fat size intemidates everyone. She even throws punches at people and shoves them around when she feels like. She started to shove me once and I told her her she laid a finger on me they would never find her body in no uncertain terms. Of course, I got in trouble, so instead of signing the sheet that I was being written up I told the boss\'s that if they insisted on writting me up, I was going legal for jobplace harrassement and suing the shit out of them. I included that if she came within 20 feet of me, or ever tried to talk to me again I would consider it grounds- then I mentioned the name of the law firm I use. They just didn\'t get it that I had been running my own business on the side for 8 years already and didn\'t think like a victum at work..I thought like a business owner..you threaten me and I\'ll kick your ass.    I am so sick of all of this. There are NO good women to find anywhere. I used to enjoy the company of women a lot more than men, until I realized that I had become Father Confessor to all of them. Why do women pick a nice guy and tell him all their problems and look to him for answers...then ignore the fact that HE may like them, and desire to treat them with respect..and instead get involved with morons who want nothing more than a blow job or a quick fuck? To hell with all of them. To hell with even being alive.  The blog is called \"No More Pathetic Crap\"..but I feel like SHIT right now so I\'ll write what I want to.  '),(1837,1969237,'21,April,2004','2004-04-21 00:00:00','   .    It\'s always too late when I get here to post. One of my locations is a complete pain in the ass. A tiny town with the exact mentallity I left behind in a tiny town in Florida. I\'ll pay them their stupid money and bail out if I can. Fuck their attitude. Fuck all the politics I get sucked into.  '),(1838,1969237,'17,April,2004','2004-04-17 00:00:00','   .    Yep. It finally happened. Bush said that we are in a \"war\". Gee, didn\'t he declare that was over like a year ago? How is this buffon anyway, and how did he buy his way into office...?     I can\'t seem to drum up new students for the next classes. Things are pretty dire- might ven have to go back to a 9-5 job. Oh well. I have an appointment some 2 weeks off during  the day, and 2 weeks after that to follow up. I really don\'t give a crap about that 9-5 job anymore. Need a change.    Been working on cars and computers for weeks. I have nothing else to focus/ live for. Trying to get all the old projects out of the way so I an move on soon. Back to the current projects. Rain today prevents any type of painting. Sure glad I am not working fair and stuck in the rain with my boots and swords....last year it took a week to clean them all up....  Need to drive........  '),(1839,1969237,'12,April,2004','2004-04-12 00:00:00','   .      Last weekend my dear friend came from up north, spouting starry eyed crap about how nice it would be if no one was rich, and everyone had to share everything. Strong words from a woman whose family lives on top of a major prime piece of real estate, in a wonderful house, who has never been at want for anything. I hope I didn\'t scare her off. I treasure her like she was my own, but just have a hard time with all the \"Love and Peace\" ideals coming out of Northern California..when you see it propped up by strong commerce they don\'t wanna talk about. It\'s simple..it takes MONEY to do things. Money isn\'t evil, misuse of the power it can buy is....I felt beat up, grumpy and old. Her child has major hair now, so I felt even older. He weighs as much as a good bag of grass seed now. I have no idea why that thought sticks with me.     The woman I seem to be pining for in LA may not be in good shape. The parents of a friend who she used to date were found dead in their home. I can\'t raise her by phone, don\'t know if she\'s in the state or not but really want to check in with her.      I have been aghast at what happens in Iraq. Four \"Contractors\" were killed and a mob abused their bodies. The news doesn\'t bother to say what KIND of contractors..so the USA is up in arms with the thought that a couple of carpenters or electricians working hard to rebuild bombed out housing were killed. You track it down, they were some sort of legal mercenaries. Former Seals and stuff. I have no doubts they were running around like John Wayne, flashing their firepower and acting touch and got the crowd pissed at them. There are many fine people in special forces...but it is a trade where you KILL people. It is a trade where you have a mojor chance of getting killed yourself...I can\'t feel sorry for people that get out of it and then GO PRIVATE doing the same thing! I have a friend in the know...won\'t say why, that says pay is in the thousands of dollars per day to these kinds of people. I knew mercs in the early \'80\'s and have no doubt of it.     So the US military is puffing out their chest and saying, \"This is no Vietnam, the people fighting here have no Superpower backing them up...\". Oh god....does every one killed at Pearl Harbor roll over in their graves every time the military completely misses the obvious? The way the USA has missed the point on global issues for the last 105 years is amazing. Now..they are tromping on Iraq for revenge...killing woman and children..and people say, \"Well, if they kill ours, we kill theirs and show \'em whose in charge! They won\'t try THAT  again...\" Christ! Does anybody remember that 200 years ago or so there were w bunch of people living HERE that were rally sick of the British running around in what we considered OURS..telling us what to do? Why was the War of 1776 \"Right\" and when Iraq\'s stand up for themselves it\'s \"Wrong\"...?!?!?!       I can only hope that that hell rains down on those who start wars like this. that their souls never get rest, and their families suffer with them at least as much as the suffering of the families they destroy. When my parents, so conservative that even the 1950\'s were a little to sexy think the USA is being lead by a madman then something major is wrong..something very, very major.     I saw Condleeza Rice speak in front of some jury. She said nothing what so ever in her whole sitting. She sounds just like the president. They ALL sound like the plague on this land, middle management...those educated with degrees folks who couldn\'t figure out how to turn on their own computers and use the company programs if their children were being held hostage by their wives yoga instructors. Middle Management. What a farce. People with degrees who have NEVER works a day in their lives. When the rest of us were out running business\'s, their only thoughts were when the next party was...and they get paid 4 times what we make...  ...that said, maybe this country being torn apart by Islam wouldn\'t be such a bad thing. Have these middle managers from hell done the math? No..it\'s not a lonely Superpower \"Backing Up\" Iraq..this is a Holy War these morons are starting...I see it, if a stupid American sees it then I\'ll be damned sure that Islamic folks tired of being crapped on see it! Yeah..Vietnam, two Superpowers fighting over a tiny country...but now, the USA, with 5% of the world\'s population is pissing off more than 50% of the globe...the Islamic folks. Guess in the end it WILL be the Book Of Revelations being played out. Maybe it\'s the end of the world as it was to be known to Christians...the end of the world doesn\'t need to mean the end of the actual world..it can mean the end of everything as it\'s known...just ask the Greeks or Romans about it.   Pray for someone..anyone else...to be \"Elected\" at the next US elections...stop the madness or pray for Assassing...  '),(1840,1969237,'10,April,2004','2004-04-10 00:00:00','   .      Bad day. They did something to the field next door again, maybe sprayed it with something. I was so sick yesterday I could not function. I barely got out of bed by 3:00 pm. It felt like I had poisen in my blood. Wasted day......    I have to get things done today...painting..need to finish all major repairs on Jeep TODAY.  '),(1841,1969237,'06,April,2004','2004-04-06 00:00:00','   .      Well..gee..I got a lot done so far today and I have yet to leave to go teach my classes tonight. I managed to get the marking of all my jackets done today. I still need to look at masks and see what I need to do for and with them. I guess I\'ll mark them also....     I must feel a little better, having gotten some stuff done. It would be stupid, but it feels like my body is already starting to re-propotion itself a touch. Clothing is chaning in the way it fits. Good deal..soon I will be back into my good clothing.    I expect to have all my projects done in 10 days. Then I will only have on-going projects like Fencing, Writing and Music to deal with. Miata projects will all be easy or quick at this point. In a year I plan to got an engine from Fling Miata and build it to be able to do a supercharger. I will start scouting for a 6 speed transmisson also. When I go to do it all, I will have it all sitting there, brand new and ready to swap into place. Then Project Miata will be done.   Time to go teach....  '),(1842,1969237,'02,April,2004','2004-04-02 00:00:00','   .    I am feeling somewhat better after the weekend. My system woke and and seems to have purged out the toxins and stuff. The result was a massive headache all weekend long. Today I finally awoke feeling...*gasp*...\"Good\". That would be the first time I call myself feeling good in the morning in a year.     The horrible side effect of all of this is that I have begun to notice females again. BUT..only fit ones. I am noticing only females half my age, and while in public they all seem to say they don\'t want an \"old Guy\"...it funny how many get naked ASAP when the offer comes along. The problem is, everyone is just a person. Every person has a certain sex drive and every person can be triggered. It\'s gotten to the point of being a big so what to me now. Maybe I need a TJ weekend of sleezy and easy sex with these females half my age. A few years ago when I did so it was easy enough, and these females half my age never batted an eye when I hit on them. And the said \"yes\".     OK..instead of dreaming of banging half the town (but no one in my neighborhood, as they are all old and worn out chain smokers for the most part...) that is half my age, I have projects to complete today...  '),(1843,1969237,'02,April,2004','2004-04-02 00:00:00','   .    Well..it seems time for change. Two years ago I lost 80 pounds, because I was badly out of shape and hated myself for it. All the TV shows with surgery to lose weight...idiots. It takes WILLPOWER to do so. You have to care about feeling better. People don\'t get that fat withou a lot of work at it. I sure didn\'t...     So two years ago I was at the best fitness level of my adult life, and then I injured myself in a demo of fencing..in front of 400 people. I couldn\'t let them see the pain, so I didn\'t. I thought I was going to die- but worse, the injury to my legs kept me from my pre-appointed schedule of walking, with which I had fallen in love. Top it off with having been fired from my 9-5 the week before, and no insurance to cover a doctor visit. Bummer...     So I kept off most of the weight until about 6 months ago when I got sick. A total of 20 or so pounds came back and it\'s too much. A close friend lost about what I did two years ago and kept it off with herbalife products. So....here I go trying them. His claims of pretty instant raise in energy levels seem to be true. Otherwise I would be asleep now! It\'s close to the way I was eating on my own, with a lot less hassle. I figure taking myself down to a good level this way, and changing overall to keep fit. I will lose 50 total pounds, and I know I can do it. THAT will put me at a lower weight than in high school 25 years ago. I hope like hell there are reunions planned somtime soon after that!     A friend from out of town visited yesterday, she is a sweetheart but so young. She wants to save the world, but is so naive at how to do so. \"Nobody needs to be rich! That\'s wrong!\" Geez....start with your parents girl...they have a very expensive home, anything they want, fancy cars...not everybody\'s parents get a new Benz every few years....I love her dearly, but she hasn\'t yet figured out that saving the world is done one person at a time, and that means yourself first. I certianly have tried to save enough people who didn;t want to be saved. Christ, the USA is trying to do that to Iraq right now....what a waste.    Anyway, gotta get healthy. I have a lot to acheive in the next 4 months.  '),(1844,1969237,'02,April,2004','2004-04-02 00:00:00','   .    Geez...I love cheese. I love Ice Cream. I don\'t love my lactise intolerance that makes everything go bad in my system. I don\'t like the way I feel after last nights Mexican food- as tasty as it was. Too much Cheese. Damn it.    Someone IM\'ed me yesterday. Someone I was screwing a year ago who dumped me for another guy. I was lead to beleive that soon after, the other guy thought he had \"won\" some unstated contest, so he then dumped her. That kinda sucks. it was OK for me, as we didn\'t have a lot in common other than getting naked and screwing ourselves silly. Sounds like maybe it\'s time for a yearly hump weekend or something. Whatever, I can\'t waste too much effort of realtionships right now. Too many things to do.    I was hoping a friend would have written a program for me by now to track all my issues with the business, but he has discovereda girlfriend and has done nothing ever since. Oh well, it\'s cool he is happy. I guess that gives me some hope in the big picture.      All the rqin the last few days has driven all the plants into blooming. I finally figured it out. In Florida, it rains all the time and that keeps the pollen count way down, even though more stuff blooms. Here, it rains and drys up- so the pollen goes crazy for weeks. I am dying. Today sucks on a pyshical level so far.  '),(1845,1969237,'01,April,2004','2004-04-01 00:00:00','   .    I have spent all day trying to sort out my data enty for last years taxes. Usually I turn it all in to my accountant by the end of January...but that\'s because they send me a notice for an appointment. Last year, the day after I picked up my taxes..he died. The new comapny didn\'t send out a notice with a time- just a generic notice months ago. I forgot I had to get them together.    I awoke in the middle of a dream about my girl in LA. I have known her so long, and find her very attractive. She., like me, is an only child, which means she\'s all fucked up in the head too. I have always wanted to seduce her, and last time we got together told her I have fallen madly in and out of love with her for years. It\'s true. I honestly think a good week of shagging each other silly would be all the world for both of us. I am debating following through and finallt trying to seduce her. Hard part...she\'s now in her early 30\'s and has never been with anyone before. That so sucks. I so wanna change that.    Her mother died a couple of years ago, and she doesn\'t talk to her dad.  I know she is lonely, it comes off her like a wave. She has told me before she wanted a family, but I think she is too awkward to go do it. Maybe Ihave been too. She hasn\'t worked in a few years, so if I can break down her walls, I will drag her home for an extended shag afterwards....     It started raining, and I couldn\'t see it because I was wearing my reading specs. That sucks, wet car inside now. I have to finish the taxes tonight, no matter what. I am missing the bills for the computer I bought to get the business running better..and the police report of all my equipment being stolen. Great- I guess I\'ll be up all night again.  I wish she were here taking my mind of things.  '),(1846,1969237,'31,May,2004','2004-05-31 00:00:00','   Not a single person called or e-mailed me today. They were all out with their friends I would guess, or families. The Crazy Cat Woman that attacked the neighbor had visitors all day. Some cool, some who looked like maybe they were drug dealers. She did get busted a few years ago....     Seems everytime I get in place to work from home and make a lot fo phone calls, she is outsiude screaming for her cats or running some sort of machine to refinish some sort of broken crap she has pulled from the trash. It sounds really professional, and I have even had people ask what all the weird yelling is about. I figure that if she is such a pest during my important times, I\'ll return the favor. As soon as I see a group showing up, I run out to the garage and open the door and turn on the air compressor or power sander for the next couple of hours.     Whn they come outside for a smoke, I begin to airbrush with nasty smelling paint which happens to blow their way. You get the picture. Return the favor to them, these kinds of people keep acting like the whole world is out to get them, maybe it\'s time to put some teeth in their paranoia.    Still, I got a major amout of work done on the jeep. The hood has been wet sanded and readied for a next color coat. I thinhk it will come out looking good. I then painted 90% of the undercarriage on the thing. It was cleaned some time back and needed to be shot. The front of a jeep has a small flat area between the frame rails. Mine was rusted badly under the cheap paint thrown on years ago by somebody. Now it has been sanded to bare metal, primed and color coated. Tomorrow morning I will top it with a clear coat and it will be like it rolled off the show room floor...actually better. They didn\'t pay attention to this area much in 1977, just ugly flat black back then, I mean, it was a jeep? Who ever looked at them close anyway?    I need a haircut badly, I have come to the polace of considering myslef a writer, not a musician, so I can afford to shorten and make things easier to care for. Thanks Courtney Love, for pointing out that among the artists, writers of novels get paid best of all. Maybe I can take you to dinner sometime after I get my first sale of a work. Yeah, the kid could come to, looks like you do a great job with her, no matter what everyone else thinks.    OK, shower time and then write for some hours. I passed 68,000 words last night, need to make it to 75,000 tonight if I can. I will have to go back and compeltly rewrite the first 60,000 words, I have learned a lot about plots and character interplay in the last two weeks. I can see how to make my characters stronger and it will help the story line. Why am I writing? Because I CAN.  '),(1847,1969237,'30,May,2004','2004-05-30 00:00:00','   The trouble with writing a novel based lossely on your life as a time frame is that it digs up lots of painful old memories you thought you had buried. The story really has gotten away from being a true telling of things, but I had to dig deep and try to remember how I felt when I was young and invicible. The memories of it hurt like hell.  I wondered where I lost my way, but I know, it was the damn car wreck that just fucked me beyond belief. I finally got back to my own head enough oto become something of who I was, to find I had changed so much I didn\'t know myself at all anymore. Reading the pages written during the heady days of NaMo I don\'t even remember half of what I wrote. It all works, but I will have to go back twice to turn it all into something of merit.  I used to know what I wanted, a band that made good musc, increasing skills in production and a wife and family to make it all worthwhile. Its those realizations that cut like hell to have float to the surface as I dredge deeper than I expected to.   A least I am into an acion based section, talk about giving a hero something to overcome in your story. God, how am I still alive? There is so much I have lived through, since people don\'t beleive me, I might as well recast it and use it as ficton. Maybe the payoff will be travel and the next part of my life.  Death in the shadows will wound you, he\'ll maim you, he\'ll KILL you...for long forgotten cause. I remember that as the man who wrote it has had enough and run off to be an actor instead of a musician anymore. God I\'ll miss the way Fish writes. Best of luck to him as an actor, I think he\'ll be great at that too.  Sleep please please take me to you tonight!!  '),(1848,1969237,'30,May,2004','2004-05-30 00:00:00','   I found an interesting book by an editor about writers, it offers advice as to what editors need from said writers. Its called \"The Forest for the Trees\".   It starts with a number of descriptions of writers and how insane and lonely they are. Hmm..I already figured that part out. Still, its been a fun read so far. It\'s by a woman who was an editor, so, from reading it, I had a mental picture of a woman who had spent some 40 or more years in the trade, and who dashed into things in the early 60\'s. I figured she would be rather old and quaint, with sort of a granny look to her. Nope, I looked at her website, she\'s younger than me and she\'s a babe. Of course, she has two books out now herself, which means she was too damn shy to just get out there and do it in the first place. She had to get a degree and work the trade first, before she loosened up enough to actually do what she started at as a child.     She makes a lot of good points, but paints all authors as New Yorkers with mental issues. I wish she could have met Martin Caiden. In some ways he was a hack, but what a wonderful hack! He wrote and published well over 100 books. He was larger than life in real life. I used to see him all the time at air shows, flying his restroed Junkers JU-52. HE is a good example of the kind of writer I aim to be. A working one....!    Yeah, I love the Beats and all that, but I have a southern voice myself, even though its hard to hear accents now, the sensabilites behind it remain the same. I figure to just write and then edit it into a decent form. Its the way I always work.    I see novel writing just like painting a car. First you do all the body work (outline) and get things in place where they should be. Then you sand it and blow a coat of primer on it (rough draft) and let it dry. Then wet sand it with paper down to 800 grit (First Edit). Then comes spraying a color coat (Second edit and read through start to finish). Color sanding and spot touch-up comes next (Second edit), and finally clear coat and polishing (Final edit and text made ready to send to editor. If the editor can find any flaws in the paint, its because they are better body workers than me, and thats why they get paid the tall dollar to do so. In the end, when the car rolls out the door we have money in our pockets and I look like the artists while they look like the manager.     Speaking of which. I got the entire hood of the Jeep painted today before I ran out of color. The next batch should be here on Tuesday or so. I can\'t beleive I have covered 90% of the thing with a quart of paint, and it looks good. I have to color sand and shoot a couple more coats on the hood to make it perfect, and then debate the whole custom graphic world. Of course I\'ve never done that before, but thenm again, I\'ve only written one novel and that isn\'t stopping me there either. I will be the author with the shinest cars of all.  Write on.  '),(1849,1969237,'29,May,2004','2004-05-29 00:00:00','   Its true, I am caught in a blog. All of them. Mine, theirs, all of them. I am stuck, hiding from the reality of writing mnovels. I can pund out an impressive word count with little effort, it just takes moving the entire universe to actually get me started.  I am sick and tired of sitting in front of this screen doing nothing but blogging and surfing. I lost interest in most things a year ago.  I have been avoiding writing since the end of November, and become a great fencing teacher, a business man of sorts and a drop out to the world of 9-5 jobs. Problem is, I am not writing enough.   I put all my work that was handwritten (except the five notebook novel) into Word to wrangle into shape. I need to get up and ride them.  I can\'t sleep, and am getting out of shape. It\'s all stress, and it all started at the end of November. I need to complete what I started then.  I can\'t sleep tonight and need to go to LA for inspiration tomorrow. I need to go to Anaheim for supplies also. One and then the other I guess. Maybe I can shower and force sleep upon myself. The weird weather has screwed up my internal clock something awful!  Write on.  '),(1850,1969237,'28,May,2004','2004-05-28 00:00:00','   The drive to Riverside, which even Mapquest says should take 18 minutes...is now taking AT LEAST an hour and a half, no matter what time of the day or night. Facts and figures say Riverside is the third fastest growing city in the US, so it\'s time to move back there and build the business located there. The other two locations need to be killed by December of this year.  Too damn cold and humid to do any painting without heat lamps and professional stuff, so today is a loss at that so far. The last of the demos is today, so I need to leave soon to make it to that. Need to do 9-5 job stuff but hate it so I have been putting it off. I need a new 9-5 job. Even part time mornings would be fine for now. Actually, that\'s not a bad idea either.....  I am thinking about tackling the concept of putting flames on the Jeep when I finish the basic repairs. It had factory bird decals that I had to take off (They were cracked, worn and faded out anyway). It looks too plain like it is...so maybe flames would be a good idea...?!?!?!?!?If I do get stupid and sell it, that would help some too I suppose...having a custom paint job.  Oh well, too much to do today better start on it. The Stout last night kicked my ass and I actually slept in with deep weird dreams most of the time.  '),(1851,1969237,'26,May,2004','2004-05-26 00:00:00','   What a day, maybe I should be more angry more often. I blew the clear coat onto the jeep project, made one tiny mistake (I can sand it off and do it over anyway...) and later sanded the vintage (and worn out and cracked ) 1977 decal off the hood. It\'s done to 320 grit, so I need to cut it down to 1000 grit and primer it. That\'s tomorrow mornings project. I think I need a trip to Coast Airbrush for my soul....and other spiffy things I don\'t have now.    Got the testing for the first of my Level Two classes done- have to make the practical exam cards now. That shouldn\'t be too hard, I will cut/ paste them into what already exists for the first level. Maybe I can get the other two weapons written this weekend?    Mayhaps I can even write sometime soon. It does feel good to make progress. I am so easily bored, but need a challenge.....     Seems like a good thing to swear off dating per say right now. Oh wait, tomorrow I have demos to do.,..better focus on them....not that that has anything to do with dating...just thinking about managing my time better...  '),(1852,1969237,'25,May,2004','2004-05-25 00:00:00','   Got a ticket last night, tired of feeling like the only person who ever gets caught for anything. Make it work, driving, anything for that matter. I see all sorts of broken laws of all sorts every day and nobody does a damn thing about them. The law won\'t make the slum lord who owns my apartment do anything about it...why should I respect the portion about traffic laws? Why should I respect ANY other portion of any law if they all are not applied?  I got robbed six months ago- took the cops 45 minutes to get there, tell me I should never defend the things I own because somebody could get hurt (Hell yes...THEM...) and then do their best to leave before taking a report. Detectives never followed up or did a damn thing, no one ever does anything when I am the victum. I suppoise next time I am wronged I will simply take it all out on the person doing so....it works for everyone else it seems. Pity the fool that next tries to take advantage of me or steal from me.  Got some stuff done today so far- too cold to do most of what I wanted to do.  '),(1853,1969237,'25,May,2004','2004-05-25 00:00:00','   Crap.  Got a ticket for speeding. I haven\'t had a ticket in almost 20 years. Couldn\'t get away from the yuppie in the SUV so I took the pace up. Of course, they never saw the SUV pass us after I got nailed....    I drove trucks and jeeps for almost 25 years. I know how they handle and how they don\'t handle. I am sick to death of being tailgated at 80 mph by idiots who never drove anything bigger than a VW before. Its not a tragedy when these soccer moms drive off a clif and kill the family beause they were on the cell phone and ignoring the road.     There should be both a special exam, and extra insurance costs for anyone who owns an SUV or a motorhome.  '),(1854,1969237,'24,May,2004','2004-05-24 00:00:00','   OK, so the day has gone well so far. I reshot the Jeep where I made a major mistake in the painting process last week. Its too cold to do much this week, so I had to wait for the exact perfect time before hitting it hard. I think I corrected things well enough.    Laundry now stands defeated also, thats a lot of things so far- have to call about taxes later and complete that stuff.    I need to get an hour of writing in on testing and get out to the remore location for teaching. I hate the drive to the location. If the five in the beginners class do well, I will keep it rolling, if not...bye- bye...     Got the second broken Palm fixed last night, hopefully it will last a bit longer. Off to the next issue....now that I have a Palm back and can tell what it is!!!  '),(1855,1969237,'24,May,2004','2004-05-24 00:00:00','   ..good thing too, I hate mornings, so I\'m GLAD its BROKEN. Maybe it won\'t come on time tomorrow. Night is so much better than day- then again, if there was no day I suppose night would be ruined by all the lost daytime people wandering around during it and acting like they own it.    Got to get rolling on too much shite. taxes yet again, cars, writing, testing. It never ends...  '),(1856,1969237,'23,May,2004','2004-05-23 00:00:00','   It\'s five months away until this years National Novemeber Novel Wrting Month contest. I have picked a satory name and idea, with the intention of turing it into a graphic novel afterwards.  I have been plunking away at las years story beefing it out to prepare for the publishing aspect of things. I wanted 130,000 words, but think it probably won\'t have enough beef in it for that. I DID resolve a major plot point so I can write towards a target now. I hate the way everything in the USA HAS to be written in one of three forms that people in this country expect to be spoonfed to them. I love stuff like Arundati Roy, and her \"God of Small Things\". It has all the elements of greatness, but would NEVER get past a college professor.    Speaking ofhich, I hve an adult friend who is getting six kinds of grief over her format of a major paper from some dumb fuck professor who never did anything in his life. I so HATE these kinds of goofass losers. Life is for those who DO things, not those who didn\'t do them. True, I became a fencing teacher,but I still compete when my kness let me!   I don\'t see professors taking over the world.    The TRUE reason its an aspect of war to kill all the scholars and students at first hint of war is so nobdoy has to listen to their uninformed WHINING all the time. Everything can be handled like a Monty Phython act..\"This requires IMMEDIATE discussion!\".   So the trick, as I see it, is to write some 6000 words a day between all the stroy lines I have, and there are plenty. Maybe be November I can have a clean slate to write with. SO..,I am living every day like its NaMo for a while! Rock On...Write on!  '),(1857,1969237,'23,May,2004','2004-05-23 00:00:00','   That\'s the problem, in a proactive, PC 90\'s sort of fashion. I have no mission statement.    I used to have one, I wanted to change the world, save it from itself. Then I found out I was just as lost as it and didn\'t know how to save it.    So I gave up. Things got better. Sometimes, things get worse, when I forget that I gave up and start to worry about things.    For two years I have held to the credo \"Whatever needs to happen, let it happen, let it be.\" I have no idea where the money for the next rent will come from, our how to pay the bills. It just happens. I may get fucked soon by taxes, but if so, I guess it needed to happen. I think too much about women, I can\'t turn that part off for some reason. That\'s why, I suppose, nothing ever happens there.    Maybe my mission statement is that I have no mission statement...?????    Oh well, need to go do something, but I don\'t know what yet.  '),(1858,1969237,'22,May,2004','2004-05-22 00:00:00','   I seem to be getting back to something of a baseline. I got sick as hell in Ocotber and have no idea why yet. Whenever I get sick, I gain weight. This just didn\'t go away and it was a steady gain in weight. I tried to change my diet last month to very ill effects- bringing back the very stomach sickness that wiped me out in the first place.    I have been eating like crap and sleeping too much, not excercising at all...and finally feel like I am getting somewhat recovered. Now all I have to do is kick the weight back off, which sucks.    A person I had entrusted much friendship to blew me off in January, and it really hurt my feelings. She was female, and though we were not dating, for some reason it hit me really bad. I am so sick of being blown off by people, women in particular. It seems everyone is so self-absorbed they have no time for anything except their own little plans. It also seems like people in general have become much more selfish. Everyone says they have no time, but they all know exactly what is going on with a huge set of TV shows. Everyone wastes too much time in front of the TV. I wish I hadn\'t gotten a TV a few months back. I need to regulate its use better. I does put me to sleep well if I can\'t sleep it seems. I guess its got something of a use.    Today I head about 80 miles away to a special car club event to get my windsheild replaced. I have never had a windsheild replaced before, as much as I have wanted to do so. I am taking extra money to see what else I can have done to the car while there. If I get enough time, I hope to come do more paint/ touch up work on my 9-5 car of all these years. I have done a few things and she just gets better and better looking. Not so worn and frayed around the edges.     Car number two, the 9-5 car, needs struts and a front oil seal. I am trying to figure out how/ when to afford this. I have a job interveiw lined up after my parents come visit next month. Even one sale a week for this company would be a major influx of cash. I have avoided it because I wanted to sleep until 10 am every day. I need to get at least one novel finished and get this job. I need to move to Riverside and get away from this town. I moved here to fuck somebody who said they wanted to be with me. Now, all I think about is how bad I got screwed under by them. I don\'t want to live in this stuck up little town anymore. I also intend to change my website address this coming year to get away from the ref to the town. I plan to drop the program I teach here and be down with the place completly.    Oh well, 15 minutes until I need to be rolling down the freeway. Better Go.  '),(1859,1969237,'22,May,2004','2004-05-22 00:00:00','   God, I don\'t know what my mind is doing sometimes. Yes, I had a Tramatic Brain Injury, but god this sucks.....    So I go to blast a clear coat on the last part of Jeep paint repairs..looked close at the rattle car high grade clear to make sure what it was and started to shoot. Bad move..I didn\'t roll it out into the sunlight...its clear..I can always polish out the bad parts as I go anyway. Except. I twasn\'t clear. It was high fill primer.   I KNOW I read the label that said clear- I guess its just part of the on-going brain malfunction I deal with.    So now I am stuck with my most recently painted areas coated with a nice think covering of primer I couldn\'t see in the dark garage. Actually, it seems like it may have surface filled things a bit, so in the end it might not be a problem. I\'ll let it dry, sand it down and blow the last of the color I have onto it. It should make things turn out really smooth...which is a plus.    Guess I need to get off my ass and order more paint. I would have been fine had I not tried to paint the entire jeep with a quart of paint. Oh wel, I\'m getting pretty good. Think I\'ll sand the last of the factory decals off the hood and repaint it to match also.    N othing like the smell of paint to ruin my breathing. I still wonder where my belt went. I last wore it Tuesday, I have the PANTS I wore...how the hell does one lose a belt?!?!?!?!?  '),(1860,1969237,'21,May,2004','2004-05-21 00:00:00','   I just can\'t get this Pig Slop thing to leave me alone. I don\'t know why. Perhaps there are some things we just don\'t need to know.     What a long week, now that I added a fifth day to my schedule of teaching, I can\'t stay home and watch Star Trek anymore on Wed. I am rather bummed. I also can\'t go to bed early and wake up feeling great on Thursdays. Oh well, something will work out.     I have been thinking two ways. I could get a job like I really mean it, or I could get a job in a bookstore or something for fun. I mean, my life is teaching fencing and plnking away keys towards the novels anyway. Why not take on jobs I always wanted to do for fun?    Oh great, most of the USA is terrified from one paycheck to the next..and here I am joking about work. Life has been so much better since I gave up hope. No more major stress over dumb things.    So much to do this weekend, getting a brand spanking new windhsheild for Molly the Miata on Sunday at a special event. Why can\'t I find a girl like my car? Nice lines, fun to be with, silver...well...maybe not silver...  Oh Pig Slop, who knows?  '),(1861,1969237,'20,May,2004','2004-05-20 00:00:00','   OK...I have no idea what the Pig Slop deal is. I just like the sound of it together. Its silly, and reminds me of a friend who was mostly silly.    I helped another friend move today, I no longer have a pick up truck, but it must be a genetic alteration that happens when you own one....whenever someone moves you get invited to help. I didn\'t mind, we only moved the heavy stuff, so it was OK by me.    I am what we in the professional world call \"Tired\". I haven\'t gotten a damn thing done today and should be in bed. My Palm took a crap and lost a bunch of data that I hadn\'t Hot Synced out of it. Oh well, tomorrow morning I guess I\'ll be trying to reconstruct my life.....  Wish I knew where my belt went...  '),(1862,1969237,'18,May,2004','2004-05-18 00:00:00','   There was a really attractive, somewhat shy, but basic solid and nice young woman of 18 murdered here in town about 8 months ago. I bought hamburgers from her at the drive-in she worked at. She had problems, it showed on her face..but she was a teen..and who doesn\'t have problems then?     Still, when your two best friends take you out in an orange grove and put a bullet through the back of you head and then lie and say it was an accident...poor, poor dear lost child. My heart goes out to her.  The gun that killed her made its way to the killers through a former student of mine. I saw him at midnight with his friends in a Denny\'s a few towns over a month or so ago. He is going to a fancy college now.    Other friends of mine went to school with her and went to the last party she attended the night before she was murdered. They are more than a little freaked out still.     The last time a teen murdered another teen around here was 10 years ago, and he was the brother of one of my dearest friends. OJ had just happened, and ANYBODY that did ANYTHING got convicted as they tried to keep a lid on things. I have yet to meet the poor fellow convicted, but I can see second degree manslaughter...but not murder. He killed a known drug dealer working at the high school he was attending..while he was still underage...yet they tried him as an adult. Did anybody notice..HE KILLED A KNOWN DRUG DEALER SELLING DRUGS AT THE HIGH SCHOOL....?!?!?!?!!?    I have had far too much death in my life and wish it would go away and leave me alone.    Well..off to something else equally light hearted I suppose...  I have no use for druggies and drug dealers. Instant justice for the cost of a bullet is too good for them. Slow, painful death should be meted out to them. Instead, we jail the ones who stood up and said , \"No!\". From all I heard, they went out to trade guns for drugs and a scuffle happened..a gun went off..and a dumb fuck was dead. Yeah- that sounds premeditated to ME.  '),(1863,1969237,'17,May,2004','2004-05-17 00:00:00','   All of a sudden photos of naked Iraqi prisnors have vanished from the news. Shallow female soldiers are not seen, and no one at all is talking about the actual events. The Bush admin has spun it as far as the eye can see. Rumsfeild has gone back to Iraq, and the so called resistance there should have KNOWN he was coming....yet they missed yet another chance to take him out. In a purely tactical sense this is sheer idiots in mass at work. If you see a trap you set, or someone else set....and the prey is climbing in...spring it!  I am so sick of war. I am so sick of the so called leadership of the USA. I might even vote this time, just to get Bush out of office. A bullet or a couple of rounds would be so much more fitting. Remember...live by the sword, die by the sword.....might as well modernize it. I just want to see PROOF when the draft comes back that Bush\'s daughters are signed up like good little americakans....  '),(1864,1969237,'17,May,2004','2004-05-17 00:00:00','   Well, maybe I needed to change the name of the blog- so I did. I am supposed to be working on a novel, but ain\'t been able to get it rolling again. I am directly influenced by Beatnik. It gave me an idea of how to get things out of my head.  I was surfing blogs, and as ever, at this time of the evening, found a girl from those islands that my buddy just got back from. Her blog was just called Beatnik. It had lots of refs to The Flaming Lips. Then, The Flaming Lips came on TV in the background. They seem OK overall, but two of my best musical friends have gone INSANE over them.  They remind me of my best friend ever, Kevin, who died what seems like a million years ago.A good sense of humor, a good sense of melody.God I miss Kevin, he would be having so much fun now.  Now Oprah is on TV going on about living like 400 years ago for some TV show. Give it to me, I would THRIVE away from this century.....  Need to write.  '),(1865,1969237,'17,May,2004','2004-05-17 00:00:00','   Class was a bust. They still show up...but they seem to hate working. Why does anyone pay for, and show up for a class they don\'t care about? There were five people total. What a waste of time I seem to have created.     Sat in a pub afterwards and listened to two guys lying to each other about how many women they had screwed. I wanted to do something..anything...to interrupt, but just didn\'t care enough. I used to care about everything, and for the last five or six months have cared about nothing.      I just don\'t feel well overall. Now the problem is this..is it mental or pyshical? I feel more like it is my body telling me things are almost done for me. I just hurt all over a lot of the time and have no energy. It has been suggested I get a real job...which I am beginning to actually want. I have begun to think in terms of getting a job, killing my business, so I can write when I want to at night.    I should be writing now, but I feel rather unsettled like I ate something bad. This feeling has plauged me for months it seems. It can\'t just be in my head.    Got to get going on writing more than anything.  '),(1866,1969237,'17,May,2004','2004-05-17 00:00:00','   Drove to Palm Springs for the 9-5 today. The traffic was terrible. Where the hell is everyone going...and why the hell are they wandering all over the road with no regard to anyone else?   Did my three stops and was exhausted by the time I got home. Now I have to go teach at the soul draining location where nobdoy really wants to get off their ass and work. I so wish I could get like 15 more people in Riverside steady..so I could kill the Redlands location...  '),(1867,1969237,'16,May,2004','2004-05-16 00:00:00','   Went to the Chino Air Show. I have been going to air shows since I was 4, and I have never seen such a huge collection of vintage restored aircarft in my life. I like thme better restored than hacked up to be racers..by far.    Most of the show was fly-bys of incredibly rare aircraft. No stunting in them to speak of, jusy get \'em flying and let everybody look at them. They tossed thr Flying Wing out into the skies for a while also. What a spiffy aircraft.    Air shows seem to have this \"Us against the World\" and \"We win because we are bigger\" attitude about them. I saw Pappy Boyington\'s wingman, he looked like anice fella...not out to take the world, just doing his job. I also met Richard Bong\'s sister. She was a normal sweetheart of a lady...hard to tell her age. Looked like her brother was a nice fella too...what a shame to lose him so young. The P-38 Association was there in FORCE..and there were actually THREE P-38\'s THERE...including one that got flown a bit!    In addition, there was a group of WWII Japanese Vet Pilots as a group. I was amazed. I am sure many of therm were finally getting to see close up the aircraft they had battled against. Evrytime ANY aircraft went by their eyes were riveted to it. They were still pilots in their heads I am sure. Some had their families with them. I wanted to talk to them, but had no idea what to say...    There was a young Japanese fellow in FULL WWII flight gear posing on a Zero for the group..it was COOL..he looked really good and dashing....  '),(1868,1969237,'14,May,2004','2004-05-14 00:00:00','   Still don\'t know where the term \"Pig Slop\" came drifting in from- but it\'s stuck in my head. Sort of like the theme music from \"If You Had Wings\" at D World......      Did a demo in front of a room full of 7-8th graders yesterday, have another one today. I somehow expended a great deal of energy at it and very nearly tweaked my knees in the course of it. I was exhausted last night and woke up so very late, with less than an hour to get to Palm Springs...which..of course, after 4 hours of waiting...made me not one single cent.     Got another demo in a few hours, tried to nap to feel more alive. I hope I kick into gear. Intro class tonight, and I am keeping my eyes open for missing students in my group. It\'s that weird , looking for an excuse to quit time again....  '),(1869,1969237,'13,May,2004','2004-05-13 00:00:00','   I\'ve got to admit I love the internet. I have three cars right now, a 1994 Ford Escort LX I bought new...now has 245,000 miles on it. A 1977 Jeep CJ-5 with 170,000 miles on it and almost fully restored and looking really sweet...and my new baby, a 1999 Miata with 150,000 miles on it.    The Escort has many new parts, needs many more- and on the way home I got the idea to do all sorts of stuff to it. Then I hit the internet. I have figured out that I have an actual sports car, and the money it would cost to barely improve the performance of the Escort could BUY a same year Miata...which is a different body style than mine. HHHmmmm...econo car acting like its quick..or a purpose designed car that already kicks the crap out of it?    Problem is, I do need something to haul things around in on occasion, and another sports car is too small for that. Perhaps selling the Jeep and the Escort could net me a half decent BMW M3 or something.     I get far too attached to cars, but its really getting close to a time to change. I need a place to live that I don\'t rent either...really, really soon. In the end, I am guessing I will seel the Jeep, get a place to live by using the money as a down payment, and keep the Escort to be my \"truck\" and haul things in.  Only time will tell....  '),(1870,1969237,'13,May,2004','2004-05-13 00:00:00','   Well..today the news comes that Pvt. England had photos taken having group sex with the captives. Rumsfeild is in Ira repeating the \"It\'s only a few people who did this...all on their own\" lie.  Back to observation number one...the woman simply isn\'t bright enough to come up with things like this. Besides, if a woman wants group sex..all she has to do is find a bunch of soldiers and they\'ll do anything she wants....and she had a bunch of them around her all the time....   God only knows what we\'ll be asked to believe next...  '),(1871,1969237,'12,May,2004','2004-05-12 00:00:00','   Salman Pax is having his blog turned into a movie. It got kinda boring when he vanished off to London for a long time...and never really had a bite to it afterwards.  Riverbend is always amazing. Such an intelligent person and still, everyday, in fear for her life and her family and friends. If I prayed I would pray for her. I will burn a candle and think of her and her people, it\'s more fitting. God I hope she makes it through......with as little pain as possible.   In the 1970\'s a woman named Patti Smith changed everything about my thought process. In 2003 a woman I know only as \"Riverbend\" has had as profound an effect.   Communication is everything, the internet is the next generation to carry the truth. Pray for the end of Baghdad Burning and the start of its healing.  Why don\'t they make a movie of Riverbend\'s blog? Maybe the truth doesn\'t play too well. I would love to know a fat Hollywood deal landed on her, allowing her to get her family to safety.  '),(1872,1969237,'12,May,2004','2004-05-12 00:00:00','   Don\'t know why that came to the forefront of my mind- maybe because they are going on and on about \"American Idol\" on TV right now. I have seen it twice. The fellow everyone hates isn\'t bad at all. He NAILS it with a solid TRUTH to each comment, which are seen as insults.     The biggest problem I have is that NONE of these people are really talented. they are not artists in the least, they are simply entertainers- able to play back other peoples work and stay in tune. I don\'t prefer this sort of stuff. I want to hear people signing what they FEEL. Hence why Micheal Jackson is such an utter waste of time. He NEVER writes anything. Prince, on the other hand...well..if you can\'t see the contrasts, it\'s probably a result of your local high school cutting out art programs to fund yet another losing football or baseball team in your town.    As for football teams, they teach that only brutes who follow orders ever get ahead. When brutes follow orders we get things like Iraq.....10-1 all the soldiers who commit such nasty crimes could name their favorite teams in less than a heartbeat and tell you everything about them.    Why don\'t we ever hear about ones like the woman...that Hopi woman...who was killed doing her job? I think I heard she was the first Native American Woman killed for the US Military in the line of duty. Poor thing, what a waste.    All the brain waves seem to be going back to the Iraq Diaster. I find it so funny that GWB\'s people are saying its a low blow to bring up something as unimportant as  HIS RECORD IN OFFICE AND HOW HE HANDLES ISSUES LIKE THIS STUPID WAR THAT HE STARTED!!!!!!! Death to them all and their families. I hope to hell when the draft comes back the Bush girls (ALLL of them...Jeb\'s too...) get drafted and dumped into the unit that was tormenting people in Iraq.    In fact, I DEMAND to see their sign up papers (black out the personal stuff..) on line and in public veiw. Best not be in the Texas Air National Guard either!!! I wonder how long until they would all come back home preggars anyway...if they are as bright as their parents.     No more pathetic crap...anger baby, anger...spew it out...piss on those whow would piss on you before they get the chance...  '),(1873,1969237,'11,May,2004','2004-05-11 00:00:00','   Just saw the woman from all the early Iraqi abuse photos on TV. Not to be mean, be she\'s not the brightest candle in the room. She\'s also awfully young. By her accent I would say she is from some back water of a town with no chance in life at all....and the army was her only chance at something more....than just a boring existance. So, in that respect, she\'s not  completly  stupid.    As we ALL suspect, she is saying that officers told them what to do. Here\'s another reason I suspect she is telling the truth: I don\'t think she is bright enough to come up with the types of things that were going on in the pictures. Sure, she could stand by and follow orders as they went on....but she doesn\'t strike me as someone who would have the slightest clue about what would really bother someone of profound (or een average)Islamic faith....!    Clinton had a young woman that the republican party expressed outrage over him fucking. I couldn\'t care less. (Hell, the majority of the leaders in the world ALL have other women they are fucking, it comes with the power. It\'s one of the things Clinton did that made other leaders figure he WAS in control.) I would rather know the president was fucking someone than ordering them to abuse people who had no control of their situation.  '),(1874,1969237,'11,May,2004','2004-05-11 00:00:00','   It\'s 1:00 in the morning for gods sake. Giant Stone Eaters have moved to this are of the freeway and are pounding it so hard it feels like a 5.0 earthquake that won\'t stop. Just what the hell are these Caltrans morons DOING for the last TWO MONTHS..ALL NIGHT LONG....?!?!?!?!!??  '),(1875,1969237,'11,May,2004','2004-05-11 00:00:00','   ....That the Iraqi prison nightmare is \"just a few out of control individuals\"...?    Christ, the US Government can\'t figure out this is going on, and humble little me has been hearing almost exactly the same details as this from blogs for MONTHS! There are hundreds, maybe more, stories of someone\'s cousin, brother, father being taken away for torment. The stories say the women are being raped- and I have NO doubt they are telling the truth.    Mostly, I suppose I feel sorry for my own parents. They lived through WWII, when the US seemed to have dignity. Dad was in the Korean Conflict, and they saw the world go mad in Vietnam. Now it has gone to the point where its obviuos we are just as big bastards as everyone else- with photos to prove it. A few weeks ago I argued with my father that exactly this kind of thing was going on. I don\'t think he denied it, he just didn\'t want to believe it.    My earliest politcal memory? I was young, probably 7 years old. The family did the only vacation I remember, going to Ft Lauderdale to some army buddy of my dad\'s. I got pawned off on the poor teenage boy of the family. He wanted to go chase girls I am sure. He had the Democratic National Convention on TV and was watching it with rapt attention. That was fine by me, I found it interesting too. He was stuck with me, so he taught me how to play chess and I did OK. I think I surprised him, but it made the next few days go a lot easier.    On the TV was a row of vets in wheelchairs and then they were getting called all sorts of names and even being attacked. They had already given to their country, and in my young mind- of ALL the folks in the country, felt THEIR word should carry a little more weight than others. I remember going into the front room and asking why the dasabled vets were getting beat up by the police officers. I was told they didn\'t love their country and were traitors by the man whose house we were staying at. I asked how that could be with so many missing arms and legs they gave in the war.    The teen boy seemed to sense a storm and just sort of left. The man couldn\'t really yell at me because I wasn\'t his son, but every adult there was angry at me. It was like the Calvin and Hobbs comic I saw years later...Calvin walks up to his dad and asks \"How do soldiers killing each other solve the world\'s problems?\" The dad is just dumbfounded. I wish I had an answer for that one still. I cut it out (long before PC\'s and scanning) and enlarged it on a copier, and still have a copy of it.    We all know the truth, they put bottom of the barrel troops in the prison and forced them to do the evil things they did. No direct orders mind you...just that terrible implication that bad kids use to get shy kids to break the rules, \"Wouldn\'t it be cool if those guys were naked and scared their balls were gonna be ripped off by a dog....?\" Trying to get ahead, get a promotion, get more money, be cool, be a soldier or whatever the mindfuck against them was, they did all these evil things.    I sent far too many years with my hair half way down my back, getting harrassed for NO reason to NOT feel sorry for these men. mayhaps some of them knew something? For sure they are ALL part of the armed resistance as soon as they get out. If it were done to people in the US by outsiders, you can damn sure bet that amricans would be taking up arms against the oppressors...why is everyone so surprised the Iraqis are bent?    The truth is this: The President is in charge, he is responsible at every level for every public employee, including the military. In the end, it\'s HIS fault for allowing this to happen. My his family go through what he has allowed others to go through......and remember what has been happening to the women....    If I believed in god I\'d beg him to save us all- but we see what that hath wrought already......  '),(1876,1969237,'11,May,2004','2004-05-11 00:00:00','   Woke up at 5:30 am for no reason. Well, maybe because the traffic on the freeway is so bad anymore no living human can sleep once the stupid rush to LA starts every morning.      What kind of moron lives 90 miles from LA and commutes there every day to work? No matter how much money you make, the IE is a dump and not worth it. Families? WHEN do they actually see them if they leave at 5:30 am and get home at 9:30 pm? These people all have SUV\'s that drink gas like no tomorrow and are working themselves to death for the almighty dollar to pay for it. What was that Harry Chapin song? Cat\'s in the Cradle...maybe they need to listen to the words to that one.    Today the secretery for the owner of the apartments comes to \"inspect\" my place, since I refused to pay the rent increase until they effect all the repairs they need to. I fully expect to be thrown out instead of getting things taken care of. If so, I will milk this for every moment of free rent I can through my lawyers. I happen to understand I can get 90 days from protesting this set of issues, and another 90 to actuallu get evicted. Most people just don\'t know how to legaly respond in such a way that they can take advantage of it.    It\'s time to buy something to live in, and I can\'t stomach house costs. It goes against my nature and beleifs to spend some $250,000.00 for a simple house. I am going for a mobile home instead. Maybe a small house at some later date, but I practice \"K.I.S.S.\" on issues like this. That\'s why I am not in debt to the tune of the average $500,000.00 like most americans. I suppose my debt is around 400 dollars all told....    Time to go iron clothing and put it away before the loser doing the inspection gets here....  '),(1877,1969237,'10,May,2004','2004-05-10 00:00:00','   Very cool, Blogger has changed a lot. I set up a new template which lost all my links but I don\'t care. They weren\'t doing anything for me anyway.     The world is a mess, I will probably be getting thrown out of my apartment for refusing to pay the latest rent increase...I mean, I\'ve lived here since 1996 and they still have fixed only two thing the whole time...it was 350 bucks when I moved in, and the current owner expects 725 a month...Nope, ain\'t gonna pay until it\'s all fixed up...      Probably time to move anyway, so I am checking on mobile homes tomorrow after the appointment in the morning. Work is a pain, the company keeps losing students, I have no idea why- mainly that Southern Californians NEVER commit to anything, so why should I expect students to commit....?    Need to get more Jeep paint- hope I remember tomorrow- I do have it listed on the Palm. I am so modern now.  '),(1878,1969237,'09,May,2004','2004-05-09 00:00:00','   ..    Here\'s a simple thought for all the asswipes who think the Bush brigade is doing a great job:   Would you want Rumsfeild, Bush or Ashcroft as members of your family? I mean, yes, they are old...but would you want to know any of your sisters were dating them or married to them? Think about the evil spawn they would create....  '),(1879,1969237,'09,May,2004','2004-05-09 00:00:00','   .    Between three locations, 400 dollars spent, and I have only enough nes steudents to make a single class. OK..whatever. Did a picnic yetserday afternoon that went well so that\'s good. Gotta get the demos put together and get the lists of people made up. What a chore. I needed a database a year ago, thought one was being made and never got it. I have a stack of paperwork a foot high to try and make sense out of. Bummer.   High speed run on twisty roads tomorrow. Just found out it\'s a picnic at the end.,....damn it...oh well...I\'ll figure something out I guess...  '),(1880,1969237,'05,May,2004','2004-05-05 00:00:00','   .  My 9-5 job is insurance. Last Friday I went to my main account to present to a host of new employees they hired. I was there all day, repeated myself countless times. My partner and I signed up a fair amount of people. I just checked my bank account , and it yeilded me a total of only $70.00. Adding together the time it took...8 hours...the drive it took 140 miles round trip...the drive time about 2 hours...and the fact it almost made me late to my teaching gig...and I believe the job is a waste of time. In addition, I just hate doing it now. There have been complex and wasteful politics the whole time I have been involved, with only my direct manager doing any damn thing at all to ever support me or help me learn.    Looks like it may be cool enough, no wind and no lawn care going on here today. I might get the next step of paitning done on the Jeep project. Its been two weeks since I was able to touch it with all the trouble around here.     Noticed the girl of my hopes hasn\'t callled. I don\'t think she will...but then why does she seem interested and even give me her number? I don\'t understand. Is it the pace of life? I mean, a lot of people seem to like me in person, but then never return my calls to get together and do things. I\'m not overbearing, don\'t quit get it. Maybe I need to transform into the slut chasing TJ visiting madman I was a few years ago. Funny, in TJ one can pick up and bang a US woman that wouldn\'t even look at them in the US. I mean, I nailed a string of college age girls who wouldn\'t give me the time of day here. I wanted someone to hang out with, get to know...and they just got naked ASAP. I\'m not complaining, if some female young enough to be my daughter wants to get wild, lets go...but why do they only do that down there? I mean, I found interest from a wide range of the local woman, and even the whores had more dignity to them than the girls from college. What\'s the deal with that?   Next year I am going to the Phillipines with my friend for sure. If I have no woman, I will be looking for one there. His inviatation means I\'ll follow through. I wanted to check it out alone, but what the hell. I need a skinny girl. Simple enough...I really wanted one with a good mind but have come to wonder if its worth the trouble. Then again, I think maybe its american women in general that are the problem. They have no idea what they want. They want to crush every inch of a man\'s personaility and individuality out of him...and then expect him to be special and exciting. Then they leave with the first man that is exactly what they say they don\'t want in the one they are with. Worse yet, who do they come crawling back to when they get shit on by the losers? They guy they left....at least I am good about never going back into a realtionship with someone who screwed me under. If I\'m going to be that pathetic I might as well end it all.   Guess it always comes back to girls. Mostly, its nothing to do with sex, I just want someone to travel with and to share life with. Someone who is smart enough to notice the interesting things and enjoy them.  Oh well, probably best to go breathe toxic paint fumes now.....  '),(1881,1969237,'04,May,2004','2004-05-04 00:00:00','   .  Well..the Winchester location is doing its best to recycle. Damn near everybody has quit excpet for the newbies. All have some dramatic stroy to justify themselves...none had goals or a method to reaching them, so I am not surprised.   Was chatting a friend a bit ago, and his girlfriend blew him off. She blew me off as a friend some time ago, quite the little drama queen. I hope she grows up because I like her overall. If not, she\'s gonna end up bad, bad, bad....   Asked out the most attractive woman I know. She hasn\'t answered yet. Last two times I asked, she never called back. BOTH times she totaled her car. Come on, if you don\'t wanna go out you don\'t have to KILL yourself over it.....just say what every other girl I found attractive says , \" I would, but my grilfriend wouldn\'t like that....\"   Wonder what is posed to destroy itself next? I just sit and smile at the world, its going to keep on going around no matter what happens.....  Turn hard and fast very often...or drive off the mountain....Miatas Rule.  '),(1882,1969237,'03,May,2004','2004-05-03 00:00:00','   .     Because RenFaire is in full swing. I visited Saturday, as I am not working this year. The booth I worked ta is majorly short of workers, but before the run started it was like all favorites to get a pass form the owner, as he kept saying he had too many people. I didn\'t feel near as bad when I heard that, so decided to stay out of the heat and not work it. I am glad I didn\'t work, it\'s been HELL out there. My friends there have finally seen all the bad things I kept talking about, but at least were able to stay clear of them because of my heads up.    I was supposed to go to my friend\'s babie\'s B Day party today but got really sick waiting for it to start. I rushed home and threw up a number of times and then passed uncomfortably out. I do believe I am allergic to pine nuts..and badly so. Last week when I felt so bad I had eaten pine nuts, same this week.    I had woken up feeling great and got a lot done around my palce..so it was weird to have gotten so very sick...seeing how I had no contact with anyone other human until the party...and I don\'t think the same thing as last week would kick in some unexpected exactly a week later.    So..I am exhausted and only got two more short stories entered into my system. I have 20 years of writing to do the data entry for, might as well get things in saleable order finally. I long for the day when it\'s all burned to a disc (or 5 or 6...) so it is safely saved. I\'ll take the disc down to be run on a high speed printer somewhere to get draft copies to work on.  It was just far too hot today- I dread tomorrow and have new classes to teach....  '),(1883,1969237,'28,June,2004','2004-06-28 00:00:00','   I have gotten much done the last couple of days. I am tired for it. Bills have been paid, e-mails sent, phone calls made...it seems endless.  I don\'t know if/ when I can jump ahead to building a house, so I might have to take a step of a crappy mobile home in between the two. At least I would have a place to live.   People I know aren\'t answering my phone calls, so now I am worried what the hell is going on again. The summer drop out for class hurts my income a great deal. I hope summer ends soon...  '),(1884,1969237,'26,June,2004','2004-06-26 00:00:00','   I started a blog to record my efforts to buld my own house. This blog will remian my frustrated place to blow off steam, and that will become a record of acheivement on that one subject. I am so tired all the time. I need forward motion.  I will get it soon, I\'m going to be thrown out of the apartment I have lived in for 8 years so they can build a football field for the high school next door. It sucks that the effort is being lead by the father of one of the people I care most about on the planet. I really didn\'t need the stress. Maybe I can play it for all it\'s worth and get some conections out of it and use them to get my goal of a house.  '),(1885,1969237,'25,June,2004','2004-06-25 00:00:00','   There is a little area in the local mountains..right near the bottom, not too far from everything else to be cool. On the way out everyoine kept pulling over and letting me pass. It was cool, and it was weird. I wasn\'t pshing anyone at all..I was taking in the sites.  I need to move, maybe that will give me peace of mind. I went to bed early last night and slept deep enough to have nightmares. I dreamed of Kevin, Steve and Tina...and the fact that they died some 16 years ago now and I am probably the only one who even remembers they were alive outside their immediate family. I dreamed so much I woke up in tears. I am majorly depressed and know it. I put on a work face for teaching..but it gets down to lonely mostly. I am more than rather tired of seeing asshole men abusing women..and the women being too stupid/ shy/ dependant / whatever to take care of themselves.  When did this country go bad? I think in the last 3 years it paked out. All sorts of women have children without even trying to be married. That\'s so damn selfish...having a child because you want one and to hell with having a family around it. I have a friend who is dating a woman with THREE kids who still lives with her parents. What the fuck is the deal with that? She is sexy, smart-witted..and expects someone/ anyone else to pick up the bill for all of what SHE wants. The problem is, this is the only kind of woman I ever see anymore. They want \"equal rights\", but don\'t want to give up being put on a pedestal and worshipped. It\'s all so very selfish.  I have always hated being alive at this age in this time. I wish it was sooner or later for me always. I am always out of time. Time, of which I am a year older today also. Why me?  '),(1886,1969237,'23,June,2004','2004-06-23 00:00:00','   I am exhausted and nearly sick from the weekend. I overdid things so badly that my legs are cramping and my knees won\'t support weight for any length of time.  The launch was simply amazingly cool. It was nice to be surrounded by intelligent people of all sorts, from many countries. Spaceship One made its goal and the flight itself was impressive. The staging left a lot to be desired. I wish I could have set it up. I guess they don\'t quite understand how playing to the crowd and including them could be very good for them long term.  I can\'t even explain the details now. My parents are gone, this is over, and I have nearly a month of catching up to do on life. Even though I have already begun to search for a better place to live...I have found a front page write up that this land is to become a football field for the school. It\'s official now so I better get moving.  I want to build my own house. Not have a contractor do it. I hate when someone says \"I built my own house, I had so much trouble with this contractor and that contractor...\" I want to BUILD my own house. It looks like money is the biggest issue, I need to start getting things in line for it all. I\'ll go from there afterwards.  The focused idea is an alternative house of some kind. Maybe log, straw bale, slipform or something else like rammed earth. I don\'t want a \"normal\" house. I might have to live in a tiny trailer fr some time, but if I make progress it will all be fine by me.  One of my friends ran over something in the road last night and unched a hole in his oil pan. Bummer city. Oh well, everyone needs to work on their cars I guess. I have to get sanding and clear coating like a madman. All my cars have to be smogged in the next two months. Great...I needed that.  Oh well, got to get to Santa Ana and have my windsheild replaced. The new one has a delaminating problem. Got to describe the Space Flight better when I can.  '),(1887,1969237,'20,June,2004','2004-06-20 00:00:00','   Well, it\'s now 7.5 hours until launch tyime in Mohave and I am so fucking wired I can\'t sleep. I need sleep. I need it before the drive. I\'m going to meet up with a friend, maybe he\'ll car pool us there..it is rather pointless to take 3 cars for 4 people and all. At 10 bucks a car one of us could buy breakfast for two....get the picture?     It was Midsummer today- and the first day of summer began a few hours ago. I celebrated like I haven\'t in like 25 years..I went ...SHOOTING...?!?!?!? It was fun, and after all these years I was hitting targets at 100 yards with a .22 and no scope. I was DAMNED happy. I was worried a sometimes useful skill was gone. If I were young I could at least handle it if I were drafted. It was a bummer being out there with a group of kids young enough to be the children I never had. I was sad inside wishing that somewhere a child of mine was running around discovering life.     So in my mind I feel EXACTLY like I did before the first Shuttle Launch. It got delayed from a Friday to a Sunday and I was able to go along. A friend had a special pass being that his mum worked at Rockwell. The company I worked for had components on the Shuttle, but we were not big enough to get passes. I didn;t sleep all weekend, and afterwards died. A eek later as the engineers I worked under huddled together in the lunhroom to watch the landing, I became much respected when word got out I had been at the Launch. I miss that world. Maybe I should have tired harder...? But I was going to go into Music.....**sigh**....    So here I am a quater of a century later, shooting guns, fixing the neighbor kids model rocket launcher, going to High Power Rocketry events and playing guitar. Writing books sometimes too. 25 years ago I wasn\'t working a real job either. I was cutting down cypress trees that were sick in the swamps. Then record stores. Then more record stores..and then a very long time in warehouses at all levels....NOW a Fencing Coach. A COACH...?!?!?!? For an anti- athlete this is STILL too weird. My younger ones go off to impress at colleges, and there seems to be a batch going to the AF Acadamy now. What\'s THAT all about?    Next I have decided I need to build a house by myself. The biggest investment most people make is a house. If I build it myself I can save a great deal. I am thinking up a ways in the mountains. Not too far, but far enough to have toys (Read that Cars and Trucks) laying around where no one cares. I guess the american dream is to live away from other americans.     Half an hour until the alarm goes off for me to get up now. I can\'t hear fuck all, as the shooting did a number on me. I feel like Pete Townshend after a tour. Oh well, I hope it passes. Christ I wanna gun, I want rockets, I want guitars, I want cars, I want a house, I want a sexy younger woman ...maybe desire returning is a good sign.    It gets down to this. A young woman, whom I was not involved with, but trusted a great deal, blew me off and vanished. I valued her enough to NOT try and get in her pants or treat her like prey. It hurt my feelings more than I can say, and turned me off to many things. I am hoping the feeling passes soon. Its been about 6 months. If ever I run into her again, I do want a whole bunch of cool things to have happened in my life, so she understands the world doesn\'t orbit her, and the rest of us keep on living....through the pain of it all.    OK, Guess I\'ll go shave and hit the road a little early.....    Space, The Final Frontier. In hours the whole world I dreamed of as a child BEGINS!!!!!!  '),(1888,1969237,'18,June,2004','2004-06-18 00:00:00','   Today will be too long. I go shooting in the afternoon for the first time in years. I go to a Midsummer\'s Party later. Then I have to try and get a few hours sleep. At midnight I leave for Mohave Airport for tomorrow mornings Spaceship One launch.  Tomorrow at this time I will have witnessed the first private launch into Space of an astronaut. I saw so many NASA launches, I may well have seen more launches than anyone at the site. I must be there at any expense.  '),(1889,1969237,'18,June,2004','2004-06-18 00:00:00','   So I realized it was 20 years ago that Live Aid happened. Where has the time gone? Where has my life gone?  A major portion counted in years vanished into a veil of haze caused by a car wreck.   20 years ago. All my dear friends were alive. Kevin, Steve, Tina. I have tried over and over to let you rest in peace, but I miss you everyday as it you died only yesterday. 20 years ago was the biggest event in history, when lowly musicians did what governments refused to do. They tried to help others.  I was stuck in SF showing someone Fisherman\'s Wharf when I was DYING to sit and watch what was going on. Nobody would tape it for me. I wanted to see every damn minute of it. I regret it to this day.  I saw snips of things. Of course, I saw Queen....it was destined to be that a group that changed my life was the only one I saw the whole part of I guess. I think I was hiding in a little shop behind a door looking at a 8\" TV screen someone brought to work whil they were caling my name to come to them.  There are photos of my that day, young, in fringe and ready to change my world. I s wish I had lived the life I wanted to live. In the wings were Marillion, waiting if others couldn\'t play. Cat Stevens with his new name was refused entry on the tage but still left the million pound check the Islamic community gathered for the event.  Tonight I can\'t stop crying. I think of all the gay people I knew in Florida. I think of how every time I wrote or called anyone I got a list of who had died. I lost touch with everyone that mattered. Then my three best friends in life died within a year of each other. I want to know why the fuck I had to live on?!??!?!?!?  It has been at least 17 years since I had THEM to talk to. To help me sort things out when I needed to. To help me keep my sanity.  Noticing Live Aid has broken my heart in some unexplainable way.  I am mired in the middle of a novel I am writing that I have spent 20 or more years thinking about. Now I don\'t give a rat\'s ass if I ever finish it. I just think I could sell it now.  How long ago was it that I saw the SAS band and sat next to Fish as he drank and talked...? The one night of my life I didn\'t say much of anything...because I was having a flood of memories the whole time...about what had mattered to me more than anything. About how I believed. Its been so hard being alive since that night.  Just when I get an idea I get sick, or chronic pain nails me to the ground again, or memory won;t work right again. I\'d pull the trigger and end it tonight but I am just too vacant to figure out why. I think I\'ll wait until my parents are gone to do that anyway. Why hurt them that bad? It won\'t be long for that either..and then I will truely be compltletly and utterly alone on this planet.  Tonight, if I could give my life, and back to earth would come my friends instead of me..I would. They were doing something more with their lives than taking up space and resources.  All I wanted to do was play guitar and have a family. Was that too much to ask?  I don\'t dream anymore. I haven\'t dreamed since the wreck. My mind feels dead.  Old music doesn\'t bring hope and strength to change the world, it brings tears and suffocation.  What was I supposed to ask for out of life?  '),(1890,1969237,'17,June,2004','2004-06-17 00:00:00','   My parents are gone, this part of the trip is over. I realize I have been increasingly stressed for about 8 months knowing they were coming. I just wanted everything to go OK for them while here. But I dreaded all the \"Well so and so\'s GRANDCHILDREN...\" talks. They seem to understand that it costs too much to do anything in life now a days. I have never made more than what my father made in 1973, and THEIR house payment was 185 bucks a month..while my RENT is 725....They are starting to understand the modern world.  It feels like something of a weight has been lifted, but I am so worn out that I slept in today. I need to finish the Jeep project and get rid of stuff I own. I have too many \"Things\". A lack of \"Things\" gives one freedom. You don\'t need a big place to live, or all sorts of other things that go with it. Actually, besides 3 cars, I own less things now than probably ever before. I think maybe in 1984 I had about the same amount of ownership.  Speaking to a friend last night outside on the curb. A woman came out of her place to see if we were OK. Offered us wine..like an idiot I declined! Damn- I thought later she probably wanted someone to talk to! I was worried our talking had upset her so felt hurried to leave. I should have talked to HER. She was both young and attractive, and was listening to Enya in her living room. She had a dog...geez, am I blind, stupid or BOTH...?!?!!!?  '),(1891,1969237,'15,June,2004','2004-06-15 00:00:00','   God..I stayed up all night fixing a hundred year old Cavalry Sword. The guard had been crushed pretty bad in some time far, far past. You could tell by the way the crap was built up on the surface of it. I found an interesting way to massage it back into shape- it\'s brass or bronze so it isn\'t exactly pliable. I got it 99% repaired and its about 5% away from where it should be. I managed to sand out a lot of surface detail issues and it looks like a dremel and a lot of polishing compound will make this poor soldier rise to most of it\'s former glory. I beleive I can replate the blade if I can figure out how it\'s attached, and it needs to have a new handle carved.It looks like cherrywood. I need to see if the jewlery place I know can braze two places on the guard..then I can detail it out the rest of the way. The scabbard isn\'t in very good shape- but its great t find a way to take the injuries and age off this old friend of mine.....  So I didn\'t get much done today then, went to get the city payment and as ever, it isn\'t ready. Thursday they say..bastards. I got a lot of data entry done for the business, which is hurting with the stupid vacation time that has hit us. I need to do more, but at least I have 95% o the paperwork is sorted and organized stacks now, a vast improvement on the last few months. I should go walk- as I can\'t do laundry, but need to get to sleep also. If I can get up at 6:00 am I can do laundry then- and then the day will be spent driving my parents through the mountains in my car and my friends...!  c ya...  '),(1892,1969237,'14,June,2004','2004-06-14 00:00:00','   Carry my digital camera more. Take more Walks. Ignore getting a real job for 8 more weeks. Fuck the hell out of the GF who dumped me for another guy last year (and then she got dumped). Fuck the hell out of the woman I take walks with. Fuck the hell out of some young asian woman I have yet to meet. Tune up the cars. Finish things I started. Fuck the hell out of the neighbor\'s cat. Oh..I mean...shoot the nasty little bastard and hang its limp, mangled body from their front door. Throw rocks at a few police cars in the name of repressed people world-wide. Declare myself a planet and seek membership in some non-existant organization for planets. Fuck the hell out of some young chick who likes Anime. Pick up a stripper, discover she was abused as a child and fuck the hell out of her anyway. Steal a car I don\'t like and fuck it up badly. Wear Gloves. Mail Roadkill to the assholes that got me fired from my last job (now that they have been fired after having been caught as the ACTUAL department fuck ups.). Fuck the hell out of just one of those girls who walks by every afternoon. Fuck the hell out of someone too old for me, her daughter and then her grandaughter. Again. Fuck the hell out of that cute sheep. At least in Photoshop... Discover a purpose for living, and fuck the hell out of it. Fuck it all.  '),(1893,1969237,'10,June,2004','2004-06-10 00:00:00','   The truth about Ronald\'s death is this:    He died nearly three years ago, and it was never announced to the world.Instead, he was taken to Disneyland Anaheim and placed in deep freeze storage next to Walt Disney (Better known as \"The Waltsicle\") in Cinderella\'s Castle. (The upper rooms are locked and no one can ever go in there...so why won\'t they open them to verify the rumors of Walt being there?)Ron passed away some time ago, and at Bush the Elder\'s guidance (as former head of the CIA) the body was placed there for later use. They knew even then they were headed towards war someday- and would need major distraction of the press at some point or another. When they did, they simple roll out Ron\'s body and proclaim his death. They effectively took the world\'s attention off the mistakes of Bush The Lessor for nearly two weeks.  For a number of years, a character actor has played the part for doctor visits, and all files have been altered, as his blood type isn\'t the same. He was paid well, and finally got to undergo reverse plastic surgery to return his looks to his own now that he is no longer needed.  The first hint of this was a bad joke nearly three years ago where they mentioned the Waltsicle and the Ron-Pop in the same sentence. No one has actually seen him since, other than Nancy, who obviously had many interests of her own to protect and therefore played along.  There is a chance that Ron\'s daughter who passed away was dosed with something to give her cancer, so she wouldn\'t interfere at the time. While she may have been misguided in her little girl like undying love for daddy and the way she proclaimed him to the world, it\'s a shame she was killed for it. Then again, maybe that\'s a bonus, when just one stuck up little rich Republican girl is shut up too I guess.  In the last few years since his interment at Disney, that organization has bought up major chunks of land with little resistance to their advances. They own nearly everything around the Magic Kingdom, except the few fields from the fellow who refuses to sell to them right next door. Everything has fallen into their grasp, and it\'s all now patrolled by the Anaheim Police, even though leaving out the actual public roads, it\'s all private Disney property. This is because they then have jurisdiction over everything and can use the police force to back them up.   The recent flair up of the rape trial of an underage girl and the son of a high ranking cop in Orange County is simply a smoke screen in that area that kept attention away from Disney\'s issues. Disney itself is not evil (Well, other than maybe the Country Bear Review) and has simply been doing its civic duty as requested by the US Government. Their payoff was the expansion they were supported in. Its been said that M. E. got wind of the whole Reagan thing and was going to have the body removed, but it couldn\'t be put in the public storage place in Riverside that Timothy Leary is at for security concerns. M.E. was trying to keep the nose of the Disney organization clean, but the power play to remove him is all at the hands of ranking key people in the world of Disney who are deeply tied to the Republican Party.  There are probably a bunch of other things I forgot that the guy said, but its a little weird to run into a scared biker who thought people were out to get him in a bar on the Ortega Highway, only to see he\'s driven off the edge of the mountain less than an hour later. Most people who own a Ducati know how to handle them. Most people who work security at Disney can\'t afford one either I guess. All in all, it certainly added up to make the other day Number One in the Weird file......  '),(1894,1969237,'10,June,2004','2004-06-10 00:00:00','   It\'s hard to be positive after breaking your neck and having a major head injury that affects every aspect of your memory, your thoughts and your speech. It\'s hard covering it so well that no one would guess it. It\'s hard knowing who I WAS and the potential I had before the wreck, and knowing it will never happen now because of it.     People in danger get uncontrollably sexual. They must mate endlessly. When your head is injured it puts you into an endless cycle that can\'t be broken. It\'s a survival instinct is all, that tells the animal part to reproduce before dying. The closer you get to dying during youth, the stronger the urge.     A lack of control over the things you find important drives you the same way. It\'s pretty simple, like a drunk has to get that woolen feeling to relax enough to fnction, with sex you need that release and hormone dump to bring you down.    It truely sucks living in this uncontrolable mode for years on end.It sucks having so much sex you injury yourself and can\'t then function at other levels. I hate cronic pain induced by my broken neck, I hate the blur of my mine when the day is too noisy, or I am too tired, or I got too much sleep. I hate daytime, because the extra amount of images one can see overwhelmes my mind and forces me into a near coma of inactivity. I live for the night, where the sights and noise go away and I can focus on anything at all.     Only then do I get insights into conditions overall.      More women have been my sexual mates than the number of people in my graduating high school class...by far. If its not triple digits you are a rank amatuer at it. If you\'ve never had five woman in 4 hours you haven\'t a clue. If you\'ve never paid for it, or gotten paid for it you don\'t count.    I never bragged about it, in fact, never told anyone about it because I was so embarrassed. the other night I heard a college boy lying through his teeth that he had been with 30 women. It was everything I could do to not stop him cold and call him a liar. 100 to 1 he has been with maybe two girls and a whore that ripped him off big time. No more than that.     Even ants mate, otherwise there wouldn\'t be so many ants, so the big score card of sexual interaction is beyond me. I treasure more a conversation that means something than a fuck. I don\'t care anymore.    Screwing everything on two or more legs simply blows a certain kind of life force out of you without using it in a creative way. Its a sure sign that something in your life is not what yopu want. Maybe you make a lot of money, maybe you are very good looking, maybe you are entirely too talented in some fashion- but the overactive sex drive is a sure sign that a person is making up for a lack of control somewhere else. Mybe the rich one wants nothing more than an extremely simple life- but has demons of childhood to pass ( I HAVE to get farther than my drunken/ drugged/ stupid/ missing/ whatever parents did in life- and I\'ll prove myself by owning more than them). maybe they have regrest that they chose to do something they hate rather than want to do, so no matter how much money they have, they feel useless anyway. Maybe they haven\'t a dime and never wanted more than a small deal that was taken care of....but they drain the potential out of their balls or twat and will never amount to anything they care about.    The wreck made me that way for almost 15 years. Its rare nights like this I have gotten so tired I can see it and work it out in electrons enough to say it in a linear fashion. I\'m not weird or sick, I WAS injured, I am not better or worse than anyone..I am just me. Whatever needs to happen, let it happen...let it be.  '),(1895,1969237,'09,June,2004','2004-06-09 00:00:00','   Long day entertaining parental units here on Vacation. Longer dayteaching tonight. My Sabre babies are doing well..I am happy for them. They will rip the local competition a new asshole next season. No one will be expecting this I am quite sure. Evil grins abound. Too much to do and sleep is needed. Aloha.  '),(1896,1969237,'08,June,2004','2004-06-08 00:00:00','   OK..I have to say it. I am  fucking sick and tired  of \"SUV\'s\". (I drove a truck for like 20 years non-stop. Corvette engine I rebuilt, Good gearing and I did all the work myself. Replaced it with a 77 Jeep that is still out in the garage and I am painting myself. I know more about trucks/ 4 x 4\'s than 95% of the population..including most people at \"Specialty Shops\". I was doing it when they were in diapers.) SUV\'s not trucks mind you, but in this area they have become SUV\'s. Most of the time what makes something an SUV is one of two things:  1.  A dumbfuck guy around 30, shaved head, tats, goatee and peircings blazing along at 100 mph in something that shouldn\'t go over 80 and  tailgating me ...with no idea how to actual drive fast. (Probably listening to \"Korn\" or some other such crap rap/ rock outfit that was made by the record company).  2.  A soccer mom with two or three brats, lunch and a cell phone all taking more of her attention than driving...while  tailgating me . (Probably listening to J Lo or some such other crap that was made by the record compainies.)     Two weeks ago coming home late I was getting ridden by a Cadillac Escalade. (Like Caddy has a fucking clue how to build ANYTHING at all..?!?!?What a CRAP company. I like the old name \"Jew Canoe\" for them...). Take it up to 80, he\'s right on my ass, slow down to 45, he\'s blinding me with his highbeams. With all his \"American Muscle\" does he PASS me..NO..DUMBFUCK!!! Long straight road with no curves so I got nothing to play for handling wise. Kick it down to second from fifth and leave his pansy ass behind by the time I topped out of second gear. Pegged it at 135 for a ways and in the distance I see him getting his speed up. That\'s fine, dead end light ahead about a half a mile, I am looking forward to seeing him blow through it into the field across the way.     I can see the Caddy rolling over bumps like the pig it is, so I slow to make him think he\'s catching me. I WANT him on my ass when we get around the next two corners and go for the mountain pass.....I\'ll put him into a guard rail there for sure. ...(Nice trick for sports cars like mine. Letting them keep catching up in the mountains...tap the brakes to \"slow down\" for the corners...when you know somethingt nice is ahead, speed up to sucker them in, don\'t touch the brakes as you apex the corner and watch them fling their piles of shit either off the side of the mountain or into a waiting guardrail. ...yeah..I am getting good at it. The Ortega is a nice place to pull it for sure. If they can\'t drive and kill themselves..TFB.)     I let him catch almost back up when I hit a stoplight. I can see by the bodyaction he has just begun to understand he is doing 110 mph and won\'t be stopping at the light, so he blows the light without stopping at all..ends up in the wrong lane and off the road on that side of the road anyway....I had ducked around the corner and nailed it...I\'m still in 3rd gear when the CHP sitting with his lights off (facing me in the other lane) turns on everything and fucking blinds me as on-coming traffic with his spot before spinning around behind me to pull me over. Just behind him, the Escalade slams brakes to try and not run over him.     So I get pulled over...and HE just drives right by! The cop never saw him at all and I get a fucking ticket for speeding! I asked what to do when it happens next time and the guy said \"Just slow down and let him pass.\" OK brightboy..I tried that and he didn\'t want to...what next? No answer. (How the FUCK do you miss a pile of shit raised that high in the wrong lane..with high beams and an extra set of driving lights (Illegal in this state to have uncovered while on any public road) and fishtailing back and forth while you see the damn sports car that is less than a third the size of this bastard...?!?!?!?!)     So my lawyer is going to take it to court and get it thrown out. Screw them all. Since two weeks ago I have done Mr Speed Limit everywhere I go. So far since then, 7 cops have blazed past me while not on call..all majorly breaking the speed limit. What the fuck are the laws FOR?!?!?!? I have even been passed by a church bus, overloaded with kids..because there were a bunch STANDING in the aisles...while a cop was riding along just behind it.!!!Oh yeah....give the SPORTS CAR a ticker for SPEEDING..but an overloaded bus is FINE becuase it says \"Jesus Loves You\" on the side....?!?!?!?!?!?!!?!?!     One fun note..in the last three weeks, I have seen three reports on TV where stupid soccer moms were on the phone and wrecked their SUV\'s. The first one drove off a cliff on Mullholand (What kind of MORON speeds there AND tries to grab a phone? Hell, in a SPORTS CAR I won\'t even begin to think of the cell when I\'m out there!!!!!!!) while talking on the phone...kiled herself and all her kids..DUMB BITCH! The next hit a pole and damn near kiled both her kids. Poor bastards will fight chronic pain for life..I really feel for the kids......of course..looking for the phone... and the last rolled her SUV and killed herself and everyone as it dove off the freeway.     Don\'t these idiots get it? They drove a damn honda or VW for so long and now they think something HUGE handles the same. NO...DUH!  Then the compaines make all these claims of how safe they are. Well..yes...compared to cars..they are safer in many ways..BUT...the ratio of acciedents in SUV\'s is going through the roof. I read that more people were killed in SUV rollovers last year than any other kind of traffic wreck. Good...kill more yuppies...!    Another good thing...SUV\'s are so hot, the bottom has fallen out of the rest of the used car market. I saw a 4 year old BMW M3 for like 6 grand...less than half of what its really worth. There are some DEALS to be made on great Off Shore sports cars and tourers. Fuck american cars..I have had so many and had so many break down. They bled me dry. Support the american worker in Detroit? Where the fuck was THEIR support when I got screwed under in MY industry? Fuck no...they just wanna build shit cars and have everybody buy one every three years. I hope they all have to work in 7-11 like the rest of us poor bastards had to do...! The little jap car takes 6 times the beating ANY of the muscle cars I had ever did and never misses a beat. Takes a corner like a pissed off whore chasing a trick that snuck off without paying. I have friends who put a blower on their cars at 200,000 miles and are at 410,000 miles now without a problem. Over 200 hp at the rear wheels in something that had 110 from the factory..and nothing else has worn out or broken. Christ, I couldn\'t even keep U-joints in the damn chevies when I built the engines...Drag races are for losers..if it ain\'t in a mountain at high speeds..it ain\'t a race...    I have found that holding the digital camera up and snapping a photo of how close they are seems to make 95% of them back right off. I have decided to go the next step, write the times and places down and report these assholes to the police. I mean, it shows exactly how close they are running to me..has their plate number and their face.     My dad used to bitch that teenagers in the 60\'s ruined it for him when they all started wearing jeans and he couldn\'t get a decent pair for a reasonable price for work anymore. I just got what he meant I suppose.  '),(1897,1969237,'07,June,2004','2004-06-07 00:00:00','   Weller words pop into mind. He was the patron saint of saving Mod until he actually made progress and imploded. Now he is a grumpy looking fellow wimped out playing R & B...who used to have the nerve to call Pete Townshend a sellout! Christ, Ya gotta hand it to Pete to keep playing his own tunes, even if sometimes they have bad timing. If Pete sold out, Weller burned out, and that\'s worse because you don\'t know that flogging the dead horse ain\'t gonna make  it get up and move. Poor bastard. (Weller that is...).  So I give great advice to everyone, I have such insight. Maybe the fuck I should take my own advice sometime, it sure seems to work for my students and they love me for it. Got to fence an old friend tonight, he is a touch older than me, so we took breaks as we whined about our age. I really needed it. I did things wrong, but at least I did things.   Feel a little better, even though the visit from an alien land started weird as usual. God save me from somebody.  Feel more like getting something done for the first time in ages. Hope it sticks. Wish to fuck I had a woman to play with. Maybe after I finish this novel.  '),(1898,1969237,'07,June,2004','2004-06-07 00:00:00','   I was up until 4:00 am again last night. I started getting into the kitchen stuff about 5:00 pm and never stopped. The dunpster was about half filled by me this morning, I have once again gotten rid of so much stuff I can\'t believe it. I don\'t buy a lot of things, so how is it I always seem to have so much to toss out?  Well, a lot of it is stuff from 15-20 years ago that is broken or no longer of any use I suppose.  I got jeep painting down today much to my surprise, before the terrible lawn workers who ignore my place got here. I got the garage area picked up and cleaned out a lot also. I am putting more and more into bins as I empty useless stuff out of the same bins. It\'s all towards moving at some point hopefully soon.  Parents get here in like 18 hours or something, and I am stressed. I never look forward to these trips at all. It\'s so weird, we only had one vacation in 17 years of my growing up, and maybe 4 trips to the beach total. Now the people who wouldn\'t travel have become travelers...and yeah, I feel left out.If I ever had kids they wouldn\'t know which one was our home....just so they get OUT there to see the world instead of just be born to work and die.  Yeah..like that will ever happen.  '),(1899,1969237,'06,June,2004','2004-06-06 00:00:00','   Too much stress is coming down on me. I have an idea I can really make it big teaching at my own place, but am self-sabotaging. Got to get rolling.  I reassembled the kitchen I took apart months ago with grand ideas of making it nice. I painted the walls and lower cabinets and refinished the doors. It\'s pointless to do anything more for an owner who keeps raising the rent and fixes nothing himself. Its half done in my eyes, and HE can try and match the stain colors and paints I started with. I know for a fact he\'ll just have some under the table illegal alien blast the whole place with white paint froma spray gun when I\'m gone anyway. What ever happeed to taking care of what you own?  I went looking for a new place to live. I have to admit I don\'t want a mobile home in a park at all. I don\'t like the looks of the so called neighbors. I really would rather move to the mountain and hid from everyone. I mean, no one ever comes by now anyway, might as well completly become a hermit and have seasons to see at least.   Got my hair cut and wasn\'t feeling well. The woman thanked me for calling with an invite out, I felt so awkward because I called like 6 times. At least she acknowledged me. I don\'t think she\'ed go out with me anyway. She too is self-defeating. I don\'t think I looks so good with the weight that came back and all that. Oh well.  I shredded 20 years worth of useless stuff I had saved tonight. Including every single peice of paper the bitch who screwed me under ever gave me. I am working to get rid of every single thing she ever gave me also.She used to buy things all the time, it made me feel like I wasn\'t good enough for her, like I was a pet or something. I asked her to stop but she wanted to keep doing it. I finally figured out it was all about how she felt about everything, and she wanted to feel like she was doing some big favor to me by buying things instead of commiting to me as she indicated she had been willing to do.  Moving out of this town will be hard. I\'ve never lived anywhere as long as I lived here. I dislike it for many reasons, but I am scared all the same. I don\'t have the option of up and just leaving the area like I wish I did. The internet has made functioning as a writer much easier than in the past. I could effectively live anywhere at this point it seems. Oh well, got to get done and published before I worry about that anymore.  I realized I have to do things one step at a time. Finish the jeep and sell it. Sell my collections, I don\'t need material things holding me down. My parents gave everything I left behind to the fucking church school, for the little spoiled rich kids to destroy. Some of the toys I had expected to bring out here someday for the sake of my childhood memories. Some of them are worth incredible amounts of money now also. I knew this too would happen. I didn\'t know it was so much of a bother for parents to hold on to one small closets worth of toys for an only child.  I don\'t know what I want at all again. I wanted a woman, but they all seem some vindictive and vampire-like now. I wanted a nice car, got one, and found I don\'t care anyway.I don\'t EVEN feel like touching any of my musicial instruments, after spending so much time learning them and writing music.I write because I can\'t help it. I seem to be on track to never have the family I wanted so bad, so my only legacy wll have to be self-made.  My students parents tell me I am amazing with kids, they all love me. That is cool. Still, I am quickly becoming Mr Chips- the only other book besides the Great Gatsby I wanted to avoid living. I seem to be weaving the two stroies together. At least Chips HAD the woman of his dreams for a time, and Gatsby had cash.....its like a sick joke that I don\'t have any of the strengths they had. Somebody just shoot me.  I must be severly depressed, and have no idea how to break the chain right now. This isn\'t teenage angst either, if I didn\'t feel so sick I would be worried..if that makes any sense at all.  June 6, 1979 was the last time I smoked pot or did any other thing that was bad for my body like that. I wonder if it was worth the gesture?  '),(1900,1969237,'05,June,2004','2004-06-05 00:00:00','   I knew it would happen this way. Raygun finally dies and everyone is again giving him credit for things he didn\'t do. It was bad enough THEN, do we HAVE to hear it all again....?!?!?!?!? They also are rosy eyed not remembering all the damage he did to the world.     They remember him making the biggest tax cut ever..They don\'t remember it screwing things up royal and MAKING the biggest INCREASE to cover it the next year.    They complain about crappy products made overseas, but don\'t remember him busting the unions (Starting with the air traffic controllers) and ruining US Industry...Thus FORCING things overseas.    They remember a sad eyed speech when Challenger blew up..They don\'t remember his DIRECT order that it be launched against all operating procedures because of the tempature. They also don\'t remember his MURDER of NASA, when he took so much money from the program to refloat the god damned Battleship Missouri. They don\'t remember that the only way NASA got Shuttle in the air was to go to solid boosters instead of liquid ones....which is the SECOND way Raygun is directly responsible for murdering the crew.     They don\'t remember that the hostages had their release secured by tireless efforts of CARTER behind the scenes..who didn\'t showboat everything. Carter showed TRUE dignity by not saying anything about it at all. Raygun beat his chest and acted like somehow HE was responsible for it.    They don\'t remember it was a Republican Congress that killed the B2 program, and CARTER made sure all the parts of the program were moth-balled so it could be built later....so Ronnie could get credit.      They don\'t remember he froze Min Wage for eveyone and took away cost of living wage increases so that everyone at my young level could never get ahead.     They don\'t remember that health care went to crap because he backed big business getting into HMO\'s.    They don\'t remember he traded ILLEGAL DRUGS for MONEY to illegally arm people trying to overthrow a government..and then he LIED under oath about it!    They don\'t remember that terror first hit us because of him. They don\'t remember that the US was looked at as a place of compassion with Carter, and a nation of brutes telling everyone what to do under Raygun.    He destroyed all my dreams when he destroyed NASA. I was THERE. Everyone in my hometown had someone who worked there. I heard it right from the people at NASA, NOT from TV. All I had dreamed of doing since I was 5 was working for NASA, the days I tried to get a job at ANY contractor were met with anger as companies were laying HUNDREDS of people off. Bastard.    They don\'t remember the birth of the whole PC world with fucking christians telling everyone else what to do.     They don\'t remember this was the start of the holy war Bush carries on now.    They don\'t remember censored lyrics and bands dropped by labels instead of facing the white house.     They don\'t remember that he so angered Europe that they were able to unite into the EU and exculde anyone else from their nations who might not want to live in the US.    They don\'t know all those things about what his policies were doing world-wide that my friend Steve told me before he was murdered, probably by his governmental loonies. Steve, I miss you so and have to admit you were right.     These are the same people that put the Bush cult in office. These are the same people who don\'t see the currnet cult leaders faults. These are the same poeple who had to CREATE a probelm with Clinton the second he got into office (Whitewater) and yet REFUSE to hear anything bad about a moron who lead us into a war that wasn\'t needed and is driving us into even MORE debt and wants more money to build prisons in Iraq.    The USA will die early. It will die because there are too many resources here, and 90% of the people have never faced TRUE adversity and won\'t know what to do if it comes without warning.     They don\'t remember cuts to art funding to expand the military.    They don\'t remember the repressing horror that was the 1980\'s for anyone that wasn\'t a bible beating goody two shoes that wanted MORE out of their life.    They don\'t remember how the entire media was in his back pocket with cheery versions of everything he did to confuse the ones who never looked at more than the surface.    THEY...run this country now. God get me out......get me out....I need to leave this...place.  '),(1901,1969237,'03,June,2004','2004-06-03 00:00:00','   So most of my friends are single, through no choice of their own. One will probably be single soon because the woman of the last three years just won\'t stand on her own two feet for damn near every day to day issue.    Women are out there. They say they want love and respect from men. Fromn what I can tell in the US, what they really want is a bank account they don\'t work for. At the same time, men have dropped into wanting a free whore.    I used to work with a really sexy woman in a record store. After about the first week, we got to talking prtty easy, and her pet peeve was what she called \"Washing Machine Whores\". They were the tennis club wives wearing the little dresses everywhere they went. They didn\'t give two fucks about their husbands or really their kids. They cared if they had the best car of all their girl friends, the best house, the best washing machine...and they would only fuck their husbands if they wanted something (like a washing machine) that had caught their eye.    In the meantime, they were gladly fucking the living shit out of the younger tennis pro at the club. Word spread like wildfire anytime a new one started, it was like a contest to see who could get laid first, who could give him a BJ first, who could get nailed anal first...whatever. The pros were quiting because they couldn\'t keep up with the wives. All known to be true becuase we both knew tennis pros in the area!    Anyway, this woman was engaged, and I finally asked her how much she loved her husband. We were alone in the store and she said she adored him, etc. But the longer the day went she sadly admitted how much she was realizing that she was looking for the security of money herself, and wasn\'t feeling love towards him.    I suggested that she was becoming her own pet peeve and should just become a syraight forward whore instead. She was taken aback a bit, until I added a unchline that at least it was all down to business when she wanted something, no lying to a mate about it. The conversation came and went until she asked , \"But who would pay me for sex?\" in all honesty. I answered back with, \"I would in a heartbeat!\" with a laugh. Her eyes widened and I tossed out , \"I\'ve got 50 bucks in my pocket, I\'ll give it to you to fuck me in the storeroom right now.\"      The door was locked in an instant and I fucked the shit out of her. I gave her the 50 bucks and she was amazed. I told her I had no problem with it, because as she had pointed out, most of the time men are paying for sex even if they didn\'t want to admit it. She asked if I had paid before and I said yes, and that 99% of guys had been whoring before even if they didn\'t admit it. I just told her I didn\'t find it to be a big deal.    The next days work I joked that I couldn\'t afford more than 20 a shift for a quicky with her if she was game. I fucked her every day we worked together for months. After she married I kept fucking her, and it got a lot hotter when I was \"the other man\". She brought her husband in at one point and he was a complete fucking dork. A moron with some sort of high paying job but no brain cells to actually hold a conversation with her. Thereafter I asked her to leave him everytime I was in her. She couldn\'t make up her mind, I moved out of the state. Maybe I should have tried harder?    So my male friends can\'t find a woman, yet proudly proclaim they won\'t pay for sex. YET, they will dump 300 dollars oon a stupid dinner hoping to \"Get Some\" afterwards. I just can\'t do that. I don\'t give a fuck if I get sex from a woman on a date or not. I would like a nice conversation and something in coomon with a woman. Sex is always nice, but if its a problem, there are plenty of real and imagined whores to get your glands on anyway.    I\'m thinking my one friend really wants a steady woman but is making sure it won\'t happen. He never falls for anyone who doesn\'t have a BF, husband or something else. Then, in fits of stupidity can\'t talk to them, and does things like text message his undying love to them at 4 in the am on a worknight. Then he wonders why they never talk to him again.    I am tired of the whole chase right now for myself and haven\'t had the heart to care about anything. Maybe I should go on a grand Oscar Wilde 1890\'s whoring adeventure in TJ or something.Maybe I should get a certain 19 year old friend drunk and naked. People all just wanna get naked anyway, I don\'t know why all the headgames about age, commitment and all that. When you find the right one  it may happen. It\'s the damn churches that complicatd it all so much. I am supposing cave men/ women rather just fucked until a kid showed up, and that bonded them together.    Oh well, I need to not be bored today.  '),(1902,1969237,'03,June,2004','2004-06-03 00:00:00','   Went to DMV to renew my DL this morning and drove the Miata. I have been driving the Escort everywhere, as its a tax write off as my business car. There was a parking spot along the road in front of DMV, but something told me to park out in the dirt field across the street.     Two hours later, after they took my money and a photo, I went back out to find a woman who had been taing her driver\'s test had run into like 5 cars...exactly where I would have been parked! I\'d love to see the comments on her test sheet....;<>!    They were so busy that they didn\'t make me take the eye exam, which is good because my eyes are tired from the computer work I have been facing. I guess they saw glasses and figured I was current.    Been waking up with dreams about a young woman I know. We have mutual friends that would kill me if I slept with her. Still, she has the goofiest sense of humor and is a flake exactly the same way as I am. She has seemed rather intriugied at my dating other females her age...so who knows what will happen. If I get the chance...why say no?  '),(1903,1969237,'31,July,2004','2004-07-31 00:00:00','   Well..there is a fellow in his 80\'s who comes to fence when his health allows him to. He picks up every mistake you make and nails you for it. I can beat him, but don\'t always want to. Its nice to get behind a few points and see if I can pull it out of the hat and do what is needed to win. He plays tactics perfect everytime...he is like a 20 year old who has 80 years of strip time. Probably going to be one of the best things about fencing I remember in my own twilight years. Last night, when we began to fence the entire room stopped to watch. He digs it, he is a ham and a showoff as much as me..I like it too. I almost rendered a knee..it seems solid today with a little pain...but not destroyed.  I spent a lot of time doing quicky private lessons one on one also. Bringing the bar up a notch at a time. The students are all great at this point- some may be sloppy, but all WANT to get better.  In the middle of the evening I was handed a complement that really sank in. One of my women student\'s ( A working professional with kids..and a damned intelligent woman one can tell in conversation) told me I\'d be not only a good father, but an amazing one. She can\'t believe the way I teach and make everyone feel like they are doing increasingly better. Two other women jumped in to agree. What struck me is that I have been told this about being a father by women with kids who are professionals...in other words, intelligent women I would trust as having made a logical deduction. I laugfhed to myself that I wish the ones without kids could figure this out...! Maybe then I could find someone to hitch the bowline to and create a crew. Yeah, I was married at the wrong time, under all the wrong condidtions...so it wasn;t time in my life yet for it. NOW...I KNOW I am in the place to be the world\'s best husband and father. But..like everything else..it will happen when it happens!  What was it I read years ago? You have to treat a woman like your business partner at a fancy dinner, you\'re only daughter when she cries, her patron when she creates and your whore in the bedroom...or something like that. Having a range like that of things to do sounds good to me. I need an adventure.....  I saw a very interesting study about animal nature about a year ago. The researcher had used any number of animal types to test the thorey..and then added a none official part backing up the observations with what they saw in local pubs...:    They placed virgin female animals in cages with young, healthy males, and with older healthy males. They also placed non-virgin females in the same cage. Without fail, the virgins ALL went for the young, rippling males to mate with. The older, sexual active females ALL avoided them like a plauge and went onlky for the older males. The researcher noted that the younger males mated and left the females. Thy also were aggresive and usually hurt the females. It seems that the older males provided sex and no pain, and also interacted with the females after mating. These were the mated couples that went on to produce offspring. Most of the couples that stuck together with a younger male did not...but always had both parts of the couple in fights or roughed up and injured.  I came out of it thinking that maybe I should try and get a grant to study something extremly obvious also! I mean, last year- a young woman I knew (19) was attracting the attention of a number of men I knew. She ignored all of them, and persisted in going on-line to meet these...well...marines. You know the knid, big muscles in their arms and nothing in their heads. She, of course, wanted to be appreciated, taken to intelligent places and treated with respect. She just couldn\'t get it through her head these guys were keeping score cards of their mating prowess and didn\'t give a rats ass about a woman for more than something to fuck...once. I fear she will finally give herself to some idiot who is going to just fuck her heart up. Then again, I think that\'s what its going to take for her to gain some self-respect! Then, of course, all the guys that wanted her so bad last year will be spoken for (as I believe all of them are..). Why do I see yet another enlistment into the military coming with this one...? Oh well, I always advocate brusining ones own knees!  So..the GIG. How very funny. Don\'t get me wrong, I think this group of folks is very nice, and admire them for NOT being yet anopther Korn or Sublime copy band. ..BUT...they are very self-important. They think they know everything already. I do beleive this is there first gig. If not, it sure seemed like it.  I mean, when you have three fronts in the band and a drummer..it stands to reason that one shouldn\'t out TWO keyboards across the width of the stage so nobody can SEE the band! Then, in a very, very dark room with black lights..one doesn\'t DRESS the band in all dark clothing.....once again..BECAUSE NO ONE CAN SEE THE FUCKING BAND! Then, one doesn\'t turn their back to the crowd everytime they take a solo...because NO ONE CAN SEE WHAT\'S GOING ON!!! Yep- these pups have never studied anything about actually being in a band. They are doing a great job impressing their close circle of friends..but it isn;t going to go any farther than this until they do it right.  My friend is their George Martin. He understands all aspects of the whole deal...and can\'t get them to do so. I am going to advocate that he begin to move past them as a group and build his own skills. My guess, if he does so, is that in five years the band itself will be working at their parents busniess\'s, and draming about the past, while my friend will be working his way through an amazing carrer with really interesting bands and music. Sort of a Todd Rundgren deal.  I invited someone I knew to show at the last minute- that god she didn;t make the drive! I mean...they play FOUR tunes and stayed on stage like 26 minutes at the MOST. It was SO funny that they finished the last song and immediatly began ripping down the equipment....I mean..I wasn\'t sure they were even done yet!  I mean..even PROG bands learn tunes from other bands..because if a fan likes what you do, one can bet they\'ed like older things like \"Muscial Box\" by Genesis, or maybe the odd PFM track or something. It gives one room to stech out and play and better connect with the fans!  I am, now, rather dumbfounded as to WHY I never got anywhere in music. I have always seen these things. I suppose the choice to not go to the recording school I wanted in the 80\'s was a key moment. That...by the way..was cut short by the car wreck like so many other things. At 13 my best friend Kevin was playing better than any of the folks last night. He could, no lie..probably play every song by Genesis on the first 5 or 6 records by that point. Those are not easy songs to learn...its not a bunch of 3 minute pop like Van Halen or something...THAT\'S the enviroment I STARTED in....yeah..I think maybe I need to get the Cubase SX program working and get to writing on it.  I have just begun to wonder if the keyboard I use, which has a rather weird sizing and spacing is the main reason for the onslaught of typos as of late...and the pain in my hands when I type too much.  Oh well, I need to do something/ anything now.. (Like the clever ref to Todd Rundgren?) Oh well..no one else caught it either...I can see for miles and miles...  smiles for miles  '),(1904,1969237,'30,July,2004','2004-07-30 00:00:00','   So what\'s going on? I can\'t tell. I awoke so sick and shivering and full of visions I couldn\'t sleep again and now its light outside. I am promised to attend a musical show tonight of friends half my age- and long to do so.  See-saws balance right in the middle if built correctly. If you try and push up or down on the end of the board to force the issue, it throws the balance off compeltly and the entire deal tilts wildly to the OTHER direction. Children enjoy all this tilting back and forth and all I ever wanted to do was see if I could balance the damn things. Was that a shape of things to come?  I may be completly wrong, but I believe Cygnus X1 is the nearest Black Hole to us on this planet...in a week when the experts have changed their minds on everything important about Black Holes. Funny, the celstial SeeSaw is out of balance again.  Everything I seem to eat makes me sicker the last 24 hours- so I abide it now with a huge chunk of chocolate cake. If it was good enough for the Aztecs to kill for, its good enough for me to do the same. Chocolate is god.Dark bitter Chocolate is even closer to the Truth, the Cygnus balance of life itself.  I hve been trying and trying so very hard to \"Let it Happen\", which, from this mornings sick and twisted perspective seems to indicate I have tilted the SeeSaw yet again the other way from trying too hard.  There is so much to be said for all the Asian perspectives, something from an older people, that still have their ancient souls. I have one such friend whom I can\'t explain. I met her for massage when the pain was unlivable, we ended up in sex...which, in the end, seems to have been the most healing of all things I needed. There is something ancient about this woman, spite not her 2 children or modern life. She seems to genuinely enjoy my comapny as I hers. She is so very honest that she has no time for another husband because she is on her way to the duties of a grandmother...yet she has repeatedly encouraged me to \"Go find a really young girl who giggles and enjoys life still..\" Age, to her, seems not to matter. She just wants to know when I\'ve found someone, so she can stop being worried for me. I dont see her enough, but suppose I see her exactly as much as I am supposed to.  I can\'t tell if the answer is in running off to Asia to do whatever, or to stay ...here. I can\'t tell if the answer is in an Asian woman\'s arms, or just a young woman\'s. I can\'t tell what large chunk of land ruled by whomeever I should be on.  I can\'t figure out this Tom Robbins thing either. Sickness shivered me all night so I read \"Villa Incognito\". I probably missed a damnsite of it all, but I know I didn\'t overall. There seems to be a hell of a vibe going \"LOOK OVER HERE YOU FUCKING BLIND IDIOT!!!\" as of late. I mean, I take a borrowed Robbins book to a woman because it just seems to be exactly what I need to do, then I meet another woman who advocates reading Robbins and so I do. The themes of the book sound like a laundry list of my questions in life.  I may never write like this, I am closer to hemingway (AAAHHH!!! A heavy handed newspaper hack...but hey..the world needs them too...). I am the fellow who sees the brillant but hasn\'t yet acheived it. I have also been told by a goodly number of people that I am an amazing teacher of fencing. Last night a person with many degrees, who has yet to lose herself because of it said it in a most wonderful way that touched my soul...but I can\'t remember. She was thinking if I moved to New Zealand or Australia I could do well in the gig of fencing coach.   I suppose I need to get writing again, before the money taps out. Thats a true path that needs to get followed also. I have been, as ever, trying to hard at it.  I am no better or worse than anyone. I am just myself and not them. they are no better or worse than me, as I am to they.  I remember that night, as I tried to impress Allen Ginesberg, because thats what it was come to think of it. I was sick at the thought of war- so I pulled out 70\'s bellbottoms and red, white and blue Sly Stone cap to wear to the reading. I got his autograph on my amazing copy of \"Howl\", and a sly smile which I caught as he exclaimed \"Oh! You\'re so bright! So patriotic!\" so my whole attempt to buffon the patriotic war thing backfired completly, but he smile at me. So it\'s been years as I wonder at the words, that he spoke so relxed and quick, that I always wondered did they have deeper meaning? Did he, where he was in life, see so much more and it just came out his mouth? Probably. Certainly, that night got me to thinking a lot again. I may not have impressed, but I was noticed for sure. I am not better than anyone, just different. I love different.  The new Morrisey disc has a tune that goes on about how he \"Likes\" a girl because she\'s not right in the head. This I understand too. I agree wholeheartedly with like girls not right in the head.  God, its light again outside. I hate that. I wish the bedroom door hadn\'t fallen off, so I could close the light out and pull the room saving shades down and sleep.My own cocoon from the world of daylight and mankind. Maybe I should go to a movie or something. Or work on the Jeep as ever.  \"You can choose from phantom fears or some celestial voice..if you choose not to decide you still have made a choice...\" I like that thought...I wonder if I should answer the door naked for the next round of JW\'s that come through.....?  Thats more like it...god I am sick still...  '),(1905,1969237,'27,July,2004','2004-07-27 00:00:00','   I began feeling realy bad yesterday at about noon or so. Like something I ate had still been alive or something. It got more and more painful, yet I still had to go teach. I have no idea how I didn\'t get sick at it all.  I came home to read a book suggested by a friend, or at least the author, and fell asleep on the futon as my illegal alien neighbors seemed to be moving out. There is a trailer out front where I usually park, and my second car is blocked in.  I just awoke with chills and a turned stomach to a most affecting nightmare. I was in a house fairly out in the middle of nowhere with a lot of people I didn\'t know. Two folks I didn\'t now had died and for some reason this group was tearing the place apart and taking whatever they wanted. There was a huge four car garage filled with all manner of junk, but many useful and valuable things also. The people were picking the weirdest things to take and ignoring great stuff.  I began to feel sick at it all in the dream. They were tossing off things that probably had great meaning to these people..an Andrew\'s Sisters record, golf clubs, other smallish things. I saw a guy&nbsp; I used to work with who had vanished a few years back..I called him \"The Last Pirate\" to his face..because he WAS a pirate, and looked EXACTLY like that one woodcut of Barthlemew Roberts..and was named Roberts. ...he was looking at an Indian motorcycle from the early 70\'s. It was a dirt bike and basiclly complete, but he was just knocking it over and saying it wasn\'t worth saving. People out back were in a small canyon breaking things seemingly annoyed at not finding things they wanted.  Nothing ever said WHO the people had been, it was like a time capsule to 1973 in the place, excpet for modern stuff like DVD layers and all. My uncle was there being very annoying. (My uncle is dead and has been for some time. He was always clowning an annoying people and didn\' know when to stop. One day, he was at it again and annoyed the hell out of his third wife, who was trying to cook dinner. He kept acting like he was choking and then would grab her when she came to see what had happened. Except this time..she got tired of it and told him to shut the hell up and get ready for lunch. He didn;t come, and she told him she was tired of working so hard to make lunch and stuff to have his clowning ruin things all the time..so she ignored his clowning. Some hours later she went back into the bedroom and found the LAST gurgeled clwning he had been pulling was him being unable to actually speak because he was overweight and had a heart attack and died. He had always been so annoying to me- always going out of his way to be the center of attention, stealing it away from even us little kids like he was in competition for attention. He was a source of constant grief to me.)  I felt bad in the dream, getting sicker and sicker until I was shivering with sickness, the place was being torn apart. In an old bag I found an Epee in perfect shape that was from like 1950 or something. I was weird, had an interchangable tip so it could be dry or electric (nothing like this actually exists) and I took it to my car, along with the DVD player and some stuff like that eveyone was ignoring.  I saw an amazing headbaord that was pure Dean Martin, only to have my father show up and start tearing into it. He was throwing away parts of it...acting like he only wanted the wood for his preciuos stack of wood (Thats another blog entry...). I was heartbroken trying to keep it together. I looked like Palm Springs in 1962 or something.  I woke up for real to find my stomach churning and my body shivering, I feel awful right now and am prety much still mentally asleep and stuck in the nightmare. I hate this feeling, stuck between awake, asleep...and dead. I don\'t know what thise eans, and when I try and think of it feel like I am going to throw up. I wish I could and just get over it.  Sometimes there is a most ravishing pain at existing between the worlds of life and death that chronic pain that weakens your system brings.My mind feels sick, like its on its last legs in some important way. I feel like I need O2 in it to chase away the feeling of death.  Its nights like these when dreams come back- I have had so very few since my car wreck in 1988, when they used to color my mind every night until then..and now they leave me worried at my own demise in some well known and unexplainable way.  I\'d be sared as hell if I didn\'t feel like throwing up. My neck injuries are raising hell on me and my hands hurt like hell too. Every joint seems to be screaming at me right now. Sickness or something between worlds reaching out? I just don\'t know.  Sometimes I wonder at death and wonder if it would be so bad, I have so many friends who died already, I feel like I would actually know more people. Right now this sick feeling seems a struggle against just fading off into death myself. Its not a thing I ever take lightly. I hate it so very much and can\'t accept it, even if it is part of life itself.  I need blankets. The spelling must be awful in this....I smell skunks....  '),(1906,1969237,'27,July,2004','2004-07-27 00:00:00','   ..that by now Boozehound is deeply inside the Republican network in Riverside, prepared to do whatever the hell it is she intends to do. I hope that she displays the behavior she is so fond of showing around here...because then maybe they\'ll lock her up for real!  Dreams...the thoughts..the cats would rip the apartment apart before anyone discovered it. Maybe they would run off to someplace else. Her stuff would get thrown out and a couple of 25 year old Japanese girlchild twins with a fetish for PuffyAmiYumi, Anime and fencing would move in to the place....  oh wait..I\'m working on moving...yeah..so the twins move in and fall on hard times finiancially and have to move with me into my second bedroom.....  Oh..it\'ll never happen. She\'ll get arrested again, tossed into the nutbin for months, and everyone will feel sorry for her doing stupid things. Rent will be ignored, and she will be allowed to catch up just like last time. In the mean time, the local courst will go insane destroying my life over a data entry error on my speeding ticket.....  I need a dog. Dogs are cool...something named \"K-9\" with a laser in his nose....  '),(1907,1969237,'26,July,2004','2004-07-26 00:00:00','   Lord.   Sometimes things just happen. The boozehound cornered me to drag me to her place to see some anti-government website she discovered. She has been drinking wine again and is loud. That, as&nbsp;a matter of fact, scares me. She\'s a nutcase. Sometimes they are lovable, but I have seen the \"Dark Side\" of her everyone else ignores.   SO..she finally admits that the reason she was in prison so long was they declared her a nutcase.    SEE..I KNEW IT!!!           &nbsp;     I am afraid I have been evil...and wonder why I can\'t get the text looking normal again either.     Oh, that\'s better. You see, BH sends e-mails to everyone from the president on down speaking her mind. Usually, now that I\'ve been subjected to a barrage of them as she showed off..they say highly pointed and focused things like \"YOU ARE AN ASSHOLE AND I HOPE YOUR CHILDREN DIE.\" Which, of course, is the exact way to effect great positive change in this world and the next few.    BH is constantly saying all sorts of lightweight Michael Moore type things..so (and this is the possibly evil part), I took control of her computer and sent her to all sorts of sites she never heard of before. Threw out names of important Seal Team Six type people and the lot. I am supposing by tomorrow morning, in her drunken haze, she will have e-mailed them to everyone from the DNC on down to the local college radio station. Sooner or later I am going to see the shadowy dark figures of Mr Bush\'s Ninja\'s as they race away across the rooftops having dispatched her. Oh wait, COMIC CON flare up again..sorry world...   I am sorely depressed that I never found the \"Bob the Angry Flower\" booth last weekend. As far as I can tell, \"Sinfest\" was not there either..Megatokyo, however, was rocking...   I\'m sorry, I dig the whole Japanese dealio. I LOVE PuffyAmiYumi unreservedly. I think Shonen Knife are groovy as hell. I love the art, the attitude..as little as I understand of it. I love the people..the long hair rock guys striking out (Loudness was so cool all those years ago...and friends of the first musicians I met here.). the chicks who rule the world and the old people who look so good at any age. I so want to go there. I might not make it back, I think I could really fall into the place like a sailor in a whorehouse after a 2 year tour with no leave.....   Anyway..the shout out goes to a new on-line friend with a soul living under a poet\'s moon. I\'m wondering if she got my e-mails I sent direct to her account and not just comments I sent to her. I wonder if she ever heard \"The Lamb Lies Down on Broadway\" from back when Genesis was fucking awesome..? I\'m dying to exchange e-mails or IM this groovio person. It ain\'t often you stumble into someone cool. So I make the effort..I try...to say Hello ! Send me messages..!   OK..time to do soemthing, anything creative...see ya Cats and Clones....  '),(1908,1969237,'25,July,2004','2004-07-25 00:00:00','   Well, its about time to address the issue of bad spelling and sentence structure in this here little old blog. I\'ll try and list reasons why they end up as far as the actual published page:   1.&nbsp; Its the car wreck thing, gave me a case of dyslexia that is damned annoying. I only see the problems some days later when I go back and think, \"Fuck all, how did THAT get past me?\"   2.&nbsp; I am practicing to be a published author. Its been said that one must leave mistakes in a submission, or the editors will have nothing to do. If they have nothing to fix, then they start tearing into your story, just to prove their worth and keep their jobs.   3.&nbsp; I am trying to spell more phonetically so as to be able to better communicate with my younger students who live on AOL IM...OMG!!1!!!!!1111!!! WTF!!!!!!1!!   4.&nbsp; I am usually exhausted (did you catch it? Just testing...) beyond actual words description and can\'t tell if I am posting to a blog or writing somewhere else.   5.&nbsp; I don\'t care anyway. I just don\'t care.   6.&nbsp; I am trying to regain the glory of youth by having a bad attitude and acting like I don\'t care.   7.&nbsp; There is no number seven. I am the new Number Two. Who is number One? YOU are Number Six!   8.&nbsp; If I did things too well Rover would bounce out of the stream behind my place, chase me down and kill me for it.   9.&nbsp; My brain seems firmly stuck at Comic Con and distracted.   10.&nbsp; What was this list about again?   Wgae Paeec  '),(1909,1969237,'25,July,2004','2004-07-25 00:00:00','   OK world, its the Comic Con thing..are the Klingon refs. But it is funny what a subculture exists around this work of fiction! Writers, in the end, rule the world!   I had a thought I forgot to record. I noticed a great number of people into the whole fantasy world are like me. That means..an only kid. Now this world of theirs can be relaxed and a part of things, or can be pushed way too far. My reach into this world doesn\'t make me want to make costumes, it make me want to write novels and movies. Oh the thinly vieled cry for acceptance!!!!;<>!!!!  Yeah, because what it gets down to is that being an only child sucks. I have no close relatives at all. I have nothing really...so being an only kid has given me all sorts of \"Issues\" as Dr Phil would say...! People think THEY are awkward...HA..Welcome to my Nightmare... (Sing it to that old Alice Cooper tune!).  I never quite know when I have talked too much. You see, I always took silence as \"Crap..I said something WRONG again..! Now I MUST TALK until I have smoothed it out and made it all work!\" (Funny, I have come to find that what I ACTUALLY did was so overwhelm everyone with such a complex and on-going line of talk that I INTIMIDATED them instead of talked to them. Thats fucking FUNNY!)  Then there is the realtionship thing. You see, I don\'t know how to be human. I don\'t know when a touch on the shoulder is the right thing, or a smack on the butt in jest. Or whatever. Because..well..there was no one else around growing and learning with me! So, I read books instead. I never got into a fight with a brother or sister, never chased a cousin around (There are 2 cousins, but we were never close...barely saw them). Never punched someone in friendship, or pushed them down in appreciation...all those human things folks learn as kids.  That, of course, is the root of the problem with my on-going (but now at least understood) complete mistaking of SEX for LOVE. Come on, its terrifiying to TOUCH someone when you haven\'t had that learning curve spelled out in the last paragraph. I also fell in love with every one who ever had sex with me. Which..sometimes...people it seems just have sex with each other because it is there. Damn...next thing you know folks will be off climbing mountains for so much of a reason!  So I had this idea some years ago. I was in Arizona for a big competitoion, and the hotel was OVERRUN with twins and more. They had their own special club..you couldn\'t be in it without a twin. Funny thing was, most of these people were hugely more dorky than I\'ve ever been.,...but they had this safety in numbers thing going down...so nobody laughed at them. Wel, I did, I was in a mood. OK, &nbsp;I was drunk..and I had to ask as many as possible what the hell the deal was with dressing alike..and maybe a different HAIRCUT might give them some INDIVIDUALITY..but I was roundly shunned.   Oh yeah, the Idea..I was thinking I should form an \"Only Child\" club! You know, we\'ed rent hotels, plan events mingle and get to know others. We\'d be THERE for our brothers and sisters in solitude! We could take on the world together..a lone family!  Oh course, truth is, we would all probably dress well, show up to the dinner the first night..and the first one in would pick a table with no one at it. Then, the same thing would happen until all the tables were taken. Those who came in afterwards would either search around, pretend they left something in their room and leave to avoid having to sit near a stranger, or act like they had walked into the wrong room. Others would get loud and long winded in their attempt to cover up their sheer terror.  The music for the evening, of course, would be all solo acts. Just one person, twisting in shyness under the spotlight, trying to express themselves. Tori Amos, Nelly MacKay, those sorts of things. We would all apallaud and nod without making eye contact with others. I guess that leave the reformed Trapp Family Singers out of the program...oh well, no one else would confess their love for \"The Sound of Music\" because it was the only taste of a family they ever had either!  If anyone asked another a question, we\'d answer, and then excuse ourselves to the restroom to make sure we didn\'t say something that offended anyone. In there, the sheer horror of all those people avoiding the main room would hit us, driving us back to the dinner. HOWEVER..I was thnking..I would put booze in everyone\'s drinks...that should do the trick...not mine though..god knows everyone hates a sincere drunk searching for acceptance. Then again, no one would actually SAY anything about it.  Let\'s see, what else....well it would actually be funny to a degree to watch a room full of people who have no idea how to fit in trying to do that. I noticed once that a number of times I ran into people with big families..and found out both parents had been only kids. I found that so interesting. Maybe it was just nice to have someone to touch without fear, maybe it was a blind ambition to populate the world, or maybe it was booze. Who can tell?&nbsp; (Sidenote: It was never going to work with my ex. Nice person from a big family. Her response to EVERYTHING was to run into another room and slam the door so she could be left alone. NOPE..now I\'VE done something wrong and have to beg forgiveness for it!..which don\'t play well to someone who always got out of facing issues through solitude in a mass setting).  So Comic Con was rather fun. I think I have been so damned shy, so damned whatever that I pushed everyone away or drove them away by talking too much. I used the weekend to just be myself *shudder!*. I found that people actually seem to like me. What a concept!  Funny, I meet a porn star and everybody else wants to fuck her. I found her eyes and the small wrinkles around them interesting. I found her accent from the east coast, so barely covered intriguing. I want coffee with her and conversation. I want to know how she got to where she is, and if she\'s actually OK as she acts. Sex..whatever...but for ONCE&nbsp; didn\'t fall in love with her. I did smile though. Smiles are nice.  Its so funny how all these people try to be loners- I wish I could give them my life for a few months. There would be a lot less Goths running around out there. Probably lead to a lot more Trapp Family Singers though.  Makes me think of my friend, the msuician in LA I have known for like 12 years. She\'s an only child and still a virgin. I have always found here so attractive and been terrified of touching her. There\'s a paradox \"I want to sex you until..well..I don\'t think we need to talk about the end point yet..but I\'m too shy to do so. Whats your favorite band again..?\"  Funny thing, mostly it gives one no baseline to set against in most matters..this only child deal. You drift and never feel grounded. You have no confidence because you don\'t know what it feels like, you overacheive because you don\'t know when to stop..or maybe you kill yourself thinking you are a failure. Mostly, for me, the word \"Awkward\" sums it up. Mostly, for me, I have NO self-confidence (which is the mother of this maddening procrastination). How very funny that everything in this life has failed, except teaching fencing. Something NOBODY trhinks they can do. I find that the actual teaching is no big deal, it just takes effort, but here am I, scared, talking too much, awkward and under confident...instilling confidence into others...for a living! Irony is a funny motherfucker!  Then again, no one prowls the night like only kids. We make cats look loud and cumbersome. I love teh sneak of the night and the protection it gives, the warmth and the way its arms hold me. I need a nice long drive somewhere. maybe later..maybe later....;<>!!  Until then, I just think that maybe we should all acknowledge Bender as God.  '),(1910,1969237,'23,July,2004','2004-07-23 00:00:00','   Well..so much for the best lasid plands of mice and men. I awoke very late today, and in a great degree of pain from the old war wounds. I think I stood too much yesterday without a break- the knees are not in working order. Thus..all my plans of conquest have been ..sidetracked, as I missed today\'s Comic Con.  I have laid in plans for next year. I\'ll go during the week for the first few days when nobody is there, and do the seminars that seem interesting on the weekend, so I can see everything. Makes sense, at least I got a feel for the deal now.  So as for planning, I got to work on the Jeep yet again. Its getting down to little details now..reinstall bumpers, tow hitches, trim, latches, catches and bottles, switches and wihoses and wires...and I THINK all under the hood has been restored to a 1977 Glory. Its very complete and clean under the hood. Its 98% on the outside. I have discovered a few more things that need replacement, but since they are details they really need to be replaced to complete the projrct. maybe 30 dollars more of expense to go.  Do I want to sell it now? No, the sentimental side of me that never wants to get rid of anything wants to store it away forever...but then again, its like a prson to store a car. The more you drive the better they run, the less you run them, the more they rust...and I like to keep mine all Shiney-Shiney. It needs a new home. Rather like any other object of value, the more someone pays for it, the more someone an be expected to take care of it.  Th other side of me is going nuts from boozehounds, illegal alien neighbor children who are known to steal, cats, and just generally weirdness. I also relaize I don\'t want my friends over her ever because I am so ashamed of the dump. That sucks. No matter how hard one tries, you can NEVER polish a turd. I have found this to be true once again.  So I have tests to grade and comments to write by e-mail to students. I need a shower now and to get to work, so I am clear to go out tomorrow morning early and get some stuff done. Perhaps by this time tomorrow the Jeep will be officially done?  Wage Peace...!  '),(1911,1969237,'23,July,2004','2004-07-23 00:00:00','   What else can one say? I met the person most responsible for change in my life..Roger Dean. He did the album cover art for the band \"Yes\" all the way back. His artwork caught my eye and turned me onto Prog Rock...something outside of Jazz I was so into. My life has never been the same since...  We spoke at some length, and I discovered much to my amazement and joy he also weilds steel. He used to fence Sabre but has long since gotten into Kendo. I wouldn\'t expose the details of the conversation we had- but it was, I can already tell...another key point in my life!  Then the \"Networking\" button switched on, and I saw openings everywhere. I talked to so many cool people. I got cards and asked abour certain projects that had been sitting at the back of my mind for some time. I got a lot of leads onto how to get somewhere with them.  There is an actress of no small acclaim I have gravitated towards a number of times at conventions, as she is going in a very Eileen Caddy sort of direction. I gave her a book by Eileen she is yet to read. I want to ask her out to coffee or something just to have time to tell her about some things I know of that would play out very well for her...rather then her reinventing the wheel. She may be moving back to her home country as a result of some business issues, so I guess I need to drive back to San Diego tomorrow and ask if I can.  I spoke at no small length to a woman porn star also. Now, one would think I want to fuck her, but not really. (I mean if she asked I wouldn;t say no..DUH!), but I stumbled onto her deal, didn\'t know her and was ever curiuos about who people are and what they do.  Not THAT part! No..I wanted to know if she was going to work it and go into production someday..to find she already owns her own production company. It got me to thinking about an old idea..about a girl I used to love who was murdered..and wondering if somehow the story could be filmed with an \"X: rating..but not be directly a porn film. There is something about a \"Super Realism\" or an almost Pre-Rapehlelite Brotherhood depection of the events that would serve the story well. I have thought for years that a film of true grit and truth, with pure and actual sex scenes in it, and also those kind of horrible things that happen in real life might be an advance in the film making world. But in another way just need this out of my head....yet another reason to go back.  God I want to create something that matters..if not children then art.  Time for sleep so I can drive back- this time alone to move at my own pace. There is so much to be done for one short life.  Wage Peace!  '),(1912,1969237,'23,July,2004','2004-07-23 00:00:00','   This would be that post, that famous #100 in the series, later to be read by scholrs and critics alike to chart the path wrought upon me for my life, fame, fortune and everything that goes with it (I thank you all..) that may happen hereafter.  So what does one talk about in the 100 post one asks? Well, I think it srather obviuos that one speaks of the graces of the more amazing members of the opposing sex. Thusly we begin.  My hairdressor is godawful hot. She\'s about 30, total little skinny rock and roll chick. ALWAYS dressed rock and roll, cool shit done to her hair and stuff. I LOVE talking to her and have asked her out a number of times. Twice, she has destroyed her car in a wreck (No...is a simplier way of avoiding the dreaded date if its an issue.) She is sexy, intelligent, and all her friends are married and she\'s bored. I want her. I guess she could be \"MHD\". My Hair Dressor.  There is a chick at the best record store in the area, she is also skinny as a rail and sexy as hell. Of course she is all rock and roll working there. She has an amazing amount of tattoos all over her body. I want to see her naked to see all of them, I also want to see her naked so I can have wild and non-stop sex with her. She is damned intelligent and I am dying to talk to her at length. I beleive she has a dumbass marine boyfriend who has no idea the prize he has in her. I want to get closer to her, so when he bloews it (marines always do...) I can dive in for the kill. I so want her. I guess she could be\"RSG\" Record Store Girl.  There is another chick that I am unsure of her age, but I am guessing it would send me to jail to kiss her. She isn\'t as skinny, has nice curves and is the funniest, and cutest package I may have yet seen in a woman. I have woken up with dreams about her if I wanted them or not. In the aspct of what I like about females, she has it all partly because of her race, and partly because of who she is. I can\'t have her and it makes me crazy I tell you, crazy... I guess she could be \"TOISN\" The One I Shouldn\'t Notice.  10 years ago I would have bedded all of them already. Now, I want someone for who they are, and don\'t wanna screw it up by just bedding them. I am in no need of notches.  But, if I have to die soon, I\'d like to do so after having had a night of long, hot, sweaty and fulfilling sex and conversation with just these three woman...OK? Thank you god....  '),(1913,1969237,'22,July,2004','2004-07-22 00:00:00','   So its this time of the night when I have finally worn myself out enough to have only one or two thoughts going at once. I have been trying to rather get at the root of my problems in general. The biggest of which is the annoying habit of having 11 things going at once. I mean, I can\'t go work on the car...I have to strip the hood with a sander, primer it, while it is drying wash dishes, while they are drying put in a load of laundry, go back and check the hood, weetsand it, finish putting away the dishes, run and throw the clothing in the dryer, check my e-mails and phone messages, toss a topcoat on the hood, fold the laundry and push the car bck into the garage. That about handles 3 hours.  Is it any wonder I get so little done? At a 9-5 job I was so limited in what I had to do that I always got it done before anyone..better and faster in every way. I suppose this burn of energy is why most bosses are terrified of me. They always think I am out to take their job...when its the last thing in the world I care about. I just want a steadt income to funs my mad adventures..my life is always away from work- but those 9-5 jobs took a heavy toll on me with all the backstabbing and jerking around. I guess I am sti;; trying to break the time wasting habits of that portion of them with all this screwing around.  I used to think I wanted like 3 women in my life at any given time...as lovers and more. I know now that isn\'t true. Trying to actually build a harem all those years ago (for all due purposes) is enough to kill oneself. Everytime I hear some guy spouting off about all the women he has I can tell he is lying...if they really had all that many they would be more menatlly exhausted, and the last thing they would want to do is talk about it. All these so called \"Players\" are fakes for the most part. Going to a whore is whoring, not having more women.  I\'m like a race car, I have a high performance chassis, but I need a track to dump it on and run the shit out of. I used to have some 50 or so issues/ projects that needed to be done and am down to like three now. That\'s progress. Some went away because they were a waste of energy to even want..I am happier without the complexity of that desire or what I thought was a need (a big house/ an overly fancy car/ etc). Funny, the things left are simple enough: Turn all 12 story outlines into novels, buy an older mobile home to live in so I can reduce expense, sleep until the crack of noon, go to a movie during the day when everyone else is slaving away, read a book....a focus on the much more simple. I am actually nearing a quiting point on the Jeep and its a weird feeling knowing I may actually sell it, and that I would have a certain freedom from the possision of it.  Years ago Val loaded her Miniture Dobie named Sid Viciuos into her 1967 Thunderbird, her clothes and that was it. She declared there are only two ways to be truely free: To have enough money to have everything, or to have so little money you have nothing. Either way, you have the freedom to move and live as you like. I thought then I still wanted everything. I know now she was a bodhivatsa showing me the way. I so wish I could have found her in Vegas to chat with. I always fall hard for the lesbians don\'t I? I still am amazed that she was 20 with an 90 year old soul.   Guess I need to get to the money part- maybe even get seriuos about a 9-5 job for a while. Sell the Jeep, move, get some income to boost marketing for the business and do the two major things that I need moeny for...go to the Findhorn Foundation in Scotland and Lasik surgery so I can get rid of these ever annoying glasses. Then again, whatever needs to happen will do just that.  '),(1914,1969237,'22,July,2004','2004-07-22 00:00:00','   Almost forgot the best way to have fun at work.  &nbsp;First, take a look at the Blue Screen of Death and get an idea of what it looks like. Then, go into Paint or something and make something that looks as close to that as possible. Now, toss it on a floppy disc, and go around to every computer in the office and place it on the computers as a wallpaper. (Be sure you didn\'t make it at work, unless you used someone else\'s computer.) Now, send a message to IT that your computer is frozen and you can\'t do your job...from every computer you put it on. Since its the wallpaper, if nothing else is running they\'ll spend countless amounts of time trying to figure out why they get the message the second the thing boots up every time.  Sit back and watch IT come back from work.  OR..just do it late one weekend after everyone is gone and see what happens Monday as IT wanders in two hours later than everyone to see a wall of moniters with the BSOD on them....  Or....just steal the balls out of everybody\'s mouse.....when someone freaks just&nbsp;say, \"Oh, you must have a girl mouse...\"  &nbsp;  '),(1915,1969237,'21,July,2004','2004-07-21 00:00:00','   Well, for me at least. I have to sign up as many folks for testing as possible tonight. I need to get this program implemented and rolling for everyoine\'s good. I have nearly finished the second level in this weapon..and certain students in a third weapon are going to be running up against the limits of what I\'ve written as a lesson plan very soon. Time to get back to work! Once all six levels are worked out, with pre-reqs and all, it will be easier. I have come to the grand idea that once past a certaain level, a requirement will be to become a class assistant. Why not, its how eveyone else does it..and it burns the details into their minds like no other fashion I can do.  Been thinking a lot about women as of late. Not bad things. No, I\'m not saying that I don\'t think women are bad things. No..I mean, I am saying I am thinking of them (women) without a negitive context! There..that\'s better. Women are good for many things, one can lean against them on the sofa while reading a book. Every so often they\'ll buy you a coffee, and sometimes they will allow you to buy them one. I have even heard tales of more mutual sharing of life itself, but squint my eyes and wonder who makes these wild stories up...? Next thing you know they\'ll be telling us that stuff that goes into a Black Hole CAN actually escape it in some fashion.....  &nbsp;&nbsp; Women, they certainly are nice to look at. I just read this rant in a magazine by a woman going on and on about how her best friend was so sexy and got all the attention. She described herself as rather plain, dark hair and thoughtful, while her friend was a big boobed ditz that all the guys stared at. I was wondering if this woman was yet another that ignored me at some point when I TRIED to go shower her with attention...because..damn it...not EVERYBODY thinks the same things are attractive! I would think she is, until I saw she\'s never been out of North Carolina. Great...and trapped in boredom too. Poor thing, she needs a lost weekend in TJ with far too much dancing, far to much drinking, far to much worshiping the moon and finding someone has left an ear and a a tail in her purse (obscure bullfighting ref no one will get) when she digs through it at work on Monday.  Life is supposed to be for living. Living can sometimes hurt... many things cause pain, death of someone dear, death in general, getting dumped, getting trapped, getting the idea you allowed yourself to get screwed (not in the fun weekend noted above sense), countless things.   &nbsp; Pain however, tempers the blade you forge by your actions. Too much pain and the blade is brittle, too little and its too soft to weild. It takes just the right amount of both to get a perfect balance, and everyone is different. I wonder who I am telling all this too...;<>..?!?!? Thats the cool thing about a Blog, you can just spew it out, and delete it later when you figure out you had too much to drink.  Like Christmas Day at a friend\'s house. I swore never again, I get sincire when drunk, I also get inquizitive..great, like a big wasted cat wanting to leap upon the table of life I guess. I really don\'t want to know what I said that day, but haven\'t been drunk sense. A stout here and there, but no need to get drunk. I think maybe drinking to that point is some part of a great pagan ritual that I haven\'t had occasion to taste yet.  Speaking of Pagan, my dear friend, former student with the yearling boy child should be visiting mommy and daddy soon, so we\'ll get to do coffee and I can hold the child again. He is a mamma\'s boy already, but I am sure he will grow into a woman chasing terror by his teens..I know..I met his dad. Funny, his dad is losing his hair. Such a good looking fellow too, oh well..maybe there IS a god!  If so, he lead me away from booze and into coffee. Coffee seems so much nicer on the soul somehow. Being wired is good. I can\'t stand these bookstores of late..the smell of books, coffee and women all mingeled into one. See, somehow we loop right back to girls again. You know, they are all girls as long as there is that devilish twinkle in their eyes. Sometimes it just needs a nudge to awaken.  My friends from Mexico City are working too hard, we never get to have a glass of wine together anymore. Maybe thats good, if I go over there I won\'t be able to take my eyes off the younger one....who is much of a dear friend as her mom is. But geez...she\'s a wacky cool chick. We don\'t really talk when we have been alone, we start sentences and disolve into laughter knowing full well the answer.  &nbsp;The FUNNIEST thing in the last five years was mom sending her and me to the store for something for dinner she was making. We got there, and both of us completly forgot what we were after. She asked, \"Great, what should we do?\" I had the sly answer, \"We walk around the store with a cart and buy anything that looks good to us, maybe we\'ll get what we were sent for, and then we lie and say we just felt like picking a few other things up.\" We couldn\'t see through the tears of agreement...and mom certainly rolled her eyes at the four bags of groceries we came back with. She had that Latin Woman Look going on...\"WHAT in the HELL did you tw DO...?!\" Finally, rather than go on with the lie- I told mom. She just shook her head and said , \" I should have known...\" Daughter said...,\" I TOLD you not to send the TWO of us!\" Private joke between us...rather like when we were having dinner and talking about the Ent in the second LOTRs film that had burning hair..only to have daughter swing her head in impersonation, and flop her hair fully into the gravy boat....and only I saw it...and they wonder why we just start laughing...damn..why can\'t SHE be a couple years older? oh well,....she will be in a few years....what ever that means...  &nbsp; 51 Out..for now...  '),(1916,1969237,'21,July,2004','2004-07-21 00:00:00','   Taught tonight..Sabre..two students remain and they are working at it. One has finally found his place in fencing..its so cool to see.  Got home to answer e-mails and the such, worked on the Jeep for hours and hours doing stupid me tricks as I went. Got much done. I may actually see the end soon. Got the white spoke wheels touched up amazingly well among other things...  &nbsp;&nbsp; Busted knuckles a great deal also..not too good. I wonder if I meet John for breakfast tomorrow? He never let me know....I wonder if I should just stay up..?  &nbsp; Got a cool e-mail from a new friend...they seem really intelligent and focused..and nice. The world is a better place when you meet nice people....  &nbsp; My back and legs are killing me, too much of a workout Sunday, but that\'s fine, it was worth it. Spoke to an old, dear friend today- she is getting married and I am happy for her. Wish I could go drinking with her one more time before then though. It\'ll be hard to have open conversations like we can now once she is living wth someone again. I think she is the only one that understand the depth of the pain I feel at having lost my three best friends so long ago, so close together. I probably would have lost it without her to cry to for a while. She has confided back to me, which is good.   OK...off to shower and sleep and start running again..I may require an afternoon nap to function tomorrow night in class at this point.  Wage..PEACE!!  '),(1917,1969237,'21,July,2004','2004-07-21 00:00:00','   Nothing too big today, just got a fair amount of repair stuff done on the never ending Jeep project. I tried to remember all the annoying things about the Jeep, so I would want to sell it...but realized I have repaired pretty much every annoying thing that it ever did. Not the end result I was looking for.  Found out that if I were buying a house, what I pay in rent would purchase something worth about 175,000 bucks. More than I ever want to spend for sure...but it all becomes invetment and capitol to get loans against in the end. Oh the scheming I could do with loans....BUT..I have to focus on old/ used/ small/ outlying and try to spend no more than 15,000 bucks total..so I can pay it off pretty quick.  Got an e-mail from someone I posted a comment to on a blog- sounds like a really cool person, and I hope I can get a dialouge going with them...sort of an electronic pen pal dealy. That would be cool. Its so hard to meet new people these days. Other than discovering the place for Boba Tea, internet dating has proven a diaster in more ways than one to me!  Speaking of which..a woman who I met on-line, in a racy chat room called \"Intimate Encounters\" last year (who dumped me big time...for another guy who then dumped her...) has been in touch. We got together for the sole purpose of getting naked, then she acted like she wanted more- then freaked and left. Anyway- she has, out of the blue, been back in touch saying things like \"We should get together some time...\". Finally, she tells me she has \"a friend\" she spends most of her waking hours with..who, after 9 months, made a pass at her. Now she is freaked out. She only wants a friend but is horny..which, I guess, is why she got ahold of me again. GREAT..am I destined to always be \"The Other Man\"? I think not. I have been civil, but never want to touch her again. I don;t wanna be second best to anyone. If she wants sex..maybe I could charge her...I mean..at least it would benefit me then...;<>!!!  Oh well, I am very tired, and ts earlier than usual..need sleep...!  Goodnight world.....  &nbsp;  '),(1918,1969237,'20,July,2004','2004-07-20 00:00:00','   Oh wait, I\'ve seen a number of their films..they already have it!!! Chances are, I will be unable to get the phrase \"Pig Slop\" out of my mind for years and will therefore resign myself to flogging it well past its own decent death.  &nbsp;  I hurt like hell after class tonight, but a rush of not hurting came into play after a handful of pain killers. This is good- other than the fatc I am wired and can\'t sleep (what else is new..?) things seem OK overall.  &nbsp;  So I am killing the location I live nearest too. I can\'t deal with the stress of things is the biggest reason. Next, with all the building/ not building/ rent increase/ lack of service/ etc...it does rather seem like time to move along. I am learning more and more about housing costs. I am determined to spend as little as possible to get about the same square footage I have now- if only for an exercise. From what I can tell, I should be able to knock a few hundred off my monthly expense this way. That would be good, as I can dump it into advertising.  &nbsp;  Today I got 4 calls in 20 minutes about classes. EVERY one of them had like a 5 year old. CHRIST..does NO ONE actually READ the website? It has \"Lower age limit of 12, No Exceptions\" listed in like every page except one. I guess everyone thinks their child is the only one who would be excused. Funny, when I challenge them to join the class and then take the lessons home to the child until the child turns 12..I get a solid \"NO..I couldn\'t possibly do  that!! \" So if I understand right, their child is the most special one on the planet, I am supposed to break every rule I have to accommodate them, but they will stand by and do nothing. ...yeah..EXACTLY the kind of parents I want around my place. Also exactly the kind that freak out when little Johnny shows no skill what so ever after finding out fencing is WORK..and not just waving his plastic ninja sword at the cat....which, of course, will be MY fault for not keeping him happy. &nbsp; I was going to say something positive but can\'t remember what it is. Oh yeah- if I can clear the Monday night slot I hope to get back to LA every week to work under my Mastro once again.Perhaps I could then get time to date the girl I like in town I have known for far too long. I could, in effect, stay Monday nights in LA after class, and then get back in time for later class. &nbsp; Yep. Need to get this business rolling so I have time to do as I please. I got more stuff done today on tests and lessons, what a chore. Once its all down on paper I can relax and just steer the ship a little better. Then again, I have three wepaons and three more levels to write to stay ahead of the bunch. I think maybe a \"Competitive\" class will be a lot more loosely based then the current levels..so at any point I can toss out...\"Everybody just fence tonight...\" when I hurt as bad as I did tonight. &nbsp; The moon is just a sliver tonight. I like it dark.  '),(1919,1969237,'19,July,2004','2004-07-19 00:00:00','   Wow...looks like   Blogger      has gone nuts  with  features. That\'s pretty cool. I\'ll have to get something to load as a photo. I so remember the world before computers, all the paperwork, the issues, how many people it took to do things. People bitch that jobs are going away...they are just shifting. Years ago there were specialized jobs for everything...now they have become data entry type jobs that require a lot less skill...yet people STILL can\'t figure them out ...weirdness.  &nbsp; &nbsp;&nbsp; Monthly it looks like things never get done for me until about the 20th, a fact I\'ve discovered by reading my own blog entries. This is yet another useful internet/ computer tool.  &nbsp; &nbsp;&nbsp;&nbsp; I am way behind in answering e-mails, and I didn\'t have the \"Lost in Las Vegas\" weekend I had last weekend. I have slept through the hottest part of the days, until the swamp coller just can\'t keep up and I awake. Then I work late into the night. I have so many plans/ programs/ ideas to sort out and complete. &nbsp; &nbsp;&nbsp; I have completly revamped my private group for my students. Instead of printing pages that I find tossed on the floor just after class, I\'m treating it like every other modern business in the world. They have access to a website that stores all the document sthey may need. If they go to the competition level outside the club, they\'ll HAVE to access all the information over the internet anyway, so I see it as a training tool to get them used to using the technology. &nbsp; &nbsp; Months ago I rewrote my lesson plans to make them 8 week courses in the same format, instead of 12 week courses. I had printed copies, but handed some out to people who actually needed them...only to find I couldn\'t find them ANYWHERE between three computers. Tonight I stumbled onto them in the strangest place I can\'t even describe, and spent all night renaming them to make them clearer, setting up folders that make sense, and tranfering files back and forth until it all is easy to access. In each weapon I have a testing section, and each section has duplicates of forms needed for testing...just in case! &nbsp; &nbsp; I dumped the data on a disc from a friend into my system, to find he had brought me exactly everything I already HAD from him instead of the ONE file I NEED from him. That\'s OK, he brought a hard copy, so I\'ll probably just re-enter the whole 4 pages to get it done. &nbsp; &nbsp; I got a fair amount of auto repair done also, with some of the final repainting issues looking like I may finally get them right this time. That lead into much driving and looking around the area for a place to move to. I am beginning to see an opening just the next town over. They have a bunch of mobile home parks, and while not my dream house..yet...its a step towards it. &nbsp; &nbsp; I need to get all these projects/ testing/ issues/ auto repair work just finished, so I can work 9-5 on my novel. I must finish it soon. November is fast upon us, and I have a solid goal for it this year. That is if I can find my notes....  '),(1920,1969237,'14,July,2004','2004-07-14 00:00:00','   So I have come to understand that compared to 95% of the people with blogs I am a perfectly happy individual. I have day to day crap...but I\'m not worried about the sex life of a sister-in-law/ sister/ cousin/ ex-wife/ neighbor/ hen/ zoo animal/ boy/ man/ duck/ aircraft/ etc I want to bang, or what it really means when the neighbors look at me funny or any other bizarre shit.  They look at me funny because they are druggies and drunks, that\'s pretty simple. The other one is an illegal alien with offspring who touch things they don\'t own...repeatedly. One is kicking drugs, good luck and more power to him. One has too many fucking cats, which I didn\'t used to HATE, but am becoming increasingly annoyed with.  I have decided its normal to be annoyed when you open your front door to be greeted with a boozehag sticking a nasty little beast in your face and drooling on about how it thinks its a person.  I have decided its normal to be pissed off when people don\'t watch their kids and then expect to blame everyone else when they get injured or killed.  I have decided its normal to be sick and tired of everyone around me acting like the world is out to get them when their problems are the result of their own choices and actions.  I have decided I need to live anywhere else than here, and probably the vast hinterlands of the boonies would suffice nicely. I\'d rather deal with bears than illegal alien children. Bears I can shoot and no one wants to put me in jail for making a terorist threat for it.  I have decided to vote for Kerry, just as a sign of solidarity to get GWB OUT of the White House, no matter how good or bad Kerry might be. I don\'t care. I want Bush and the havoc he has wreaked on this country dead and gone. It\'s a simple matter of numbers, I might prefer to vote for someone else...but hell, when Ralph Nader says \"Vote Kerry instead of me to get GWB out of the White House..\" I think I\'ll listen. Thanks Ralph...didn\'t like \"Unsafe at Any Speed\" but at least you cared enough to try. Bet you wish you had bought a bunch of Corvairs now don\'t ya?  I have decided it was so cool to actually quit a job again rather than hang on until the bitter end.  I have decided that cheap marketing is best and hope to make my fortune with it. Signs, Signs...everywhere signs...upping my income all of the time....  I have decided that women in the USA are 99% out for a wallet no matter what they say.  I have decided that if and when a female should enter my life again is a pretty good time frame.  I have decided to live on island time (as Jimmy Buffet calls it) and seem all of a sudden to be happier.  I have decided that once I get the immediate projects done, I\'ll have no excuse to get in the way of the REAL projects.  I have decided that I feel better again and even some women look OK again.  I have decided that I will do as I damn well please at this apartment, and if I am told not to I will pitch a fit and \"Stand up for myself\" in the same annoying/ drunken/ incoherent fashion as the boozehound.  I have decided that I am annoyed that I work hard to keep any and all autos both running well and looking good while boozehound has left a destroyed auto sit without attentiuon for over a year, three other cars sit abandoned and sad and yet I look like a probelm to boozehound.  I have decided I like guns of all shapes and sizes. Except pistoils, which I will abide by but never really like.  I have decied a cross bow would be a good project to construct from scratch once the major projects are gone. God save the car whose leaf springs appear to be the correct size for my bow.  I have decided that I need to know everything about black powder guns.  I have decided that swords are still as cool as ever, and once I work the kinks in my arms out I need to drill with actual weapons in the front yard to keep away cats, children and boozehounds.  I have decided that music or all sorts, except that canned mexican crap from next door, is good enough.  I have decided that I don\'t care about most things, because they are an immense waste of my time.  I have decided that I shouldn\'t bother being worried if I can do nothing to change it (whatever IT is..), well, short of explosives..which COULD change most anything. Hey..wait a minute...  I hve decided that if people promise to meet up for excercise and don\'t, I won\'t call again for days and just go on my own.  I have decided its kinda late and I need sleep.  I have decided to finish this and log off.  Really.  '),(1921,1969237,'12,July,2004','2004-07-12 00:00:00','   I sent an e-mail today quitting my 9-5 job, it hasn\'t been any fun anyway so screw it. I spent the rest of the day repairing equipment, cleaning equipment and making phjone calls. I found an e-mail I missed from my 9-5 from a month ago and replied to it. Oh well...I don\'t care anymore. If any of you want to get screwed under by goody two shoes born again christian bitches with too many kids by lots of men she can\'t ID...feel free to call her for it.  I have thrown boxes of shit out today, I have no idea where all this useless shit comes from. Lots of it is junk mail I never bother to open until I finally toss it out. Then its just as useless as I thought it was. I am so tired now..and still have to go teach tonight...bummer.  Oh well, signs are ready tomorrow and maybe a couple new students in the hell location.  '),(1922,1969237,'11,July,2004','2004-07-11 00:00:00','   Maybe that\'s what I actually need. It\'s been now 9 months or stomach hell for me and my digestion. I finally feel a touch better and seem to be rolling in a little of a direction.  Lets hope the signs I ordered make a change in the cash flow issue around the business. Its getting the word out thats hardest. I need to really establish myself before someone else blows into town and reaps the rewards of all my efforts of the last ten years. That could be read as anyone with money....  Took a ride tonight and saw a lot of mobile homes in Yucaipa, looks like it may be a lot nicer place to live than I figured. It was like 9:00 pm and actually really quite quiet. Might just work out for now.  '),(1923,1969237,'11,July,2004','2004-07-11 00:00:00','   It\'s funny sometimes when you try and run away from everything to get a break- that you end up right back on the front door step of where you left.  I went to the desert to watch High Power Rockets (Large and Dangerous Rocketships) being flown. I got there a touch later than I wanted and was almost hit by one landing when I drove up. I had to take care not to run over another as I went to park. Soon after, a small one attempted to get airbourne only to go wildly out of control in my immediate direction. Instead of running, I stood watching it loop and loop close and closer towards me. Fencing taught me it wasn\'t close enough to be a threat, so don\'t go wastng energy running around until you need to do so. It hit the ground nearly 20 feet away...layed there until it burned out...and then in an anti-climatic \"POP\" the nose cone popped off from the ejection charge. Poor little thing. It was, however, in perfect shape so that was cool. Who knows why it flew that way?  So I saw a few decent launches..right up into the clouds, and they had to radio track the rockets to find them as everyone lost sight of them. I was bored when the wind closed it down for the day at about 2:30 pm. I had never been north on the 247, so decided to see how far away from State Line it was. I figured that NOT getting hit by 3 rockets meant it was my lucky day...I needed to capitailize on this luck.  It wasn\'t too far away, so I went to all the casinos there...Whiskey Pete\'s, Buffalo Bills, Pimm something or other.....and then figure I might as well go to Vegas...being only like 40 more miles away. I had no idea what I was looking for anyway.  Just outside of town was a billboard...of all things.Lynyrd Skynyrd was playing! OK..time for a Skynyrd show I guess.  The show was very good, the ciggie smoke annoying as hell and the slots played OK overall. I got home at 5 this morning thinking about how much I miss about Florida and wondering why the hell I ever moved to California. Then again, I would have been miserable if I hadn\'t tried it I guess. I need a place to live away from th annoying mexican neighbors and all the women and children who aren\'t on the lease and aren\'t supposed to be here. Another \"Laws for YOU..not US\" kind of deal. It\'s really annoying how EVERYTHING can be covered with \"Well, they are children..you have to expect that with children.\" OK..I asked if children were allowed here when I moved in and was told no...so I moved in...now I live on a fucking playground it seems like. I put forth the effort to ask in advance.  I suppose what this means is I actually should stop being worried that the neighbors won\'t like me working on cars or painting or whatever...and just do it. They seem unconecrned that their loud music wakes me up every morning at 8...and I work nights. I figure maybe I need to bring things to a head and live exactly like I want to..and when they don;t like it..I\'ll \"Stand up for myself \" and bitch and whine like the nutcase women around that everyoine seems to respect sop muc for being bitches...  Maybe I should list life as \"My Turn!\"  '),(1924,1969237,'09,July,2004','2004-07-09 00:00:00','   So I was driving to the police run community center today in Redlands. There is a school zone, with all manner of lights flashing to make one aware that children are present...so I slowed to 20 mph, as I saw no signs that said to go slower. 100 yards later, the fat black woman in the clapped out Toyota behind me began blowing her horn and waving her arms madly...she wanted me to speed up!  Did she not see the signs? How could she miss the huge yellow strips and markings of \"SCHOOL ZONE-SLOW\" on the road for a few hundred yards? It\'s a mainly black school..and I do suspect that if I were speeding along, all sorts of hell would be brought to rain down on me..but it\'s perfectly accpetable for a BLACK woman to become annoyed and aggressive because she couldn\'t leave in time to observe speed laws.  My response was to immediatly slow down to 7 mph. Made sure she say the brake lights too. there was traffic coming the other way, so there was no way to pass. Had she tried, I would have blown her econobox doors off and not let her past me. Damn it, if I have to get tickets..the hell if anyone else is gonna blow the laws around me! Besides, school zones are there for a REASON..because it sucks when a kid is killed because of a stupid driver.  Last I saw her she was pulled into a left turn lane with about 15 cars in front of her. If she had half a brain she would have gone one more block, made a left, and been faced with a back road without a single car on it.Dumbass bitch.  Once at the center, I found two of the three black teens that caused all sorts of trouble a few weeks ago milling about at the front desk causing trouble. I walked up and gave them a \"Fuck you all to hell\" look and they immediatly vanished. I simply know they are going to pull something else. You can tell they are fucking assholes. Mommy and Daddy certainly seem to have enough money to drive brand new cars- I don\'t seem to think they are actually underpriveleged. Sort of like the assholes in \"NWA\" all those years ago...the TV show going from house to house with them..and ALL of the houses were a damn site better than anything MY parents ever had...and the driveways were filled with cars no more than two years old. Oh yeah \"Fight the power\" and all that lame ass lying through their teeth shit. If they lay a hand on me I will respond in kind. I don\'t want to risk re-injury of my neck because some smart ass kid wants to show his ass. They touch me, I\'ll take them down hard and fast. THEN I have every intention of suing their families for any damages.  I am tired of giving to assholes. If they want to show their asses and be like this, then they don\'t deserve respect back. I hope to hell nothing happens with this Intro Class so I have reason to close in two months. That\'s OK, they need another raquetball court to throw basketballs around in and make out in....soon enough they\'ll have a brand new football field to mess up. I hope to move out of town ASAP. Let this place rot.  '),(1925,1969237,'04,July,2004','2004-07-04 00:00:00','   Christ.  I got the ticket in the mail for 75 when an asshole was chasing me. It will cost me 300 dollars. I now have absoultly no resepct left for traffic laws at all. If someone does stupid things to me again, it\'s war time...and their vehicle or life is part of the bargin if they fuck with me. If THEY can endanger my life, I can return the favor. I\'m am tired of getting blamed and the people who caused the problem wandering off scott free. This has been a pattern in my life..the idiot losers with no idea get the promotions, never get the punishment when needed.  For now I\'ll stop short of the urge to own a pistol and open fire on these idiots...but I am getting rid of two cars with every intention of a scabby old truck that doesn\'t matter if I hit things in it. Pipe bumpers here I come. Maybe its about time white men declared something instead of being blamed for everything.  I declare if you fuck with me on the highway again I\'ll put you in a ditch, off the road or into a tree before I get another ticket caused by anyone else.  I hate living in the USA, I don\'t care what anyone says about it. If,perhaps, I was an illegal alien, then maybe I\'d have some rights here. Maybe I can declare myself hispanic or soemthing.  I want out.  '),(1926,1969237,'01,July,2004','2004-07-01 00:00:00','   Well, it is all happening yet again. I happen to be an individual by definition, and a group of people want something that takes things away from me. The outright lie about the USA is that individuals matter in the least. What matters is money and nothing else.  So some goofass salesman that made stacks of cash down by the beach is feeling his age and has decided to donate a great deal of money to build a footbal stadium, a soccer field and two baseball fields for the school next door. The city morons over the years have all opposed any form of building, so never set aside zones for growth for anything. It happens the apartments I have lived in for 9 years is right in the gunsights of the planned building project..so ..of course..fuck ME, I\'ll have to move because they want me to.  First, football, baseball and basketball are all losing steam on the national level, second, its really stupid to build the hopes of teens up that they ever have a snowballs chance in hell of becoming a pro ball player (in a town where 90% of this school is white...) and last..they don\'t even have money to buy paper for student assignments...what the FUCK do they need ball fields for if they can\'t even doing the teaching part of their job?  I am searching for a way to move anyway, but it looks like all this will catch me money-short and fuck me under. I expect to raise a public stink about it demanding cash for me having to move.  Probably what will need to happen is a quick sale of a car, a purchase of another with a second car used as a trade-in..only to establish credit for all my problems caused by debt when I broke my neck. No one gave a crap about me then either..as I now find the settlement for that injury should have been between 250,000-500,000 dollars...and I got a stunning 15,000...which was all gone after paying the basic doctor bills...leaving me to charge my credit cards to the limits to survive in the months I was without work.  Most of all, I am sick and tired of being forced to follow every law, be a great person and never get a touch of respect from anyone for it. I should rob banks, rape girls, drive like a fucking moron and attack anyone I want..then I would have established RIGHTs which couldn\'t be violated.  I am sick of being an \"American\" and STILL wish to hell I wasn\' born here. There is no \"American Dream\" if you aren\'t a minority or a criminal. The dream is to work your ass off so those other two groups can whine and take ever more money from you in taxes so they don\'t ave to pay for anything. Otherwise, why would all the best public parks be in the worst parts of town?  Fuck everyone. In Redlands. When I move I will close the location here with a letter to all major papers (not the rag in tow that promotes all the old money) on the closing as a protest against the town itself.  '),(1927,1969237,'13,August,2004','2004-08-13 00:00:00','   ...last night the font size in the browser changed to something so small I could barely read it. Just now, it switched back to some sort of older, bigger and easy to read font. I like this...but can\'t understand why its doing it...oh well, I care not anyway.  I pulled out an older keyboard than the one that came with the current tower set-up to see if it was any different typing on it. Right off, I found it had a slight dishing out to the set of the keys on the board itself. It seems a whole lot more comfortable than the other one. The other one had all sorts fof bells and whistles and buttons and crap on it...but it was slower overall. Yes, I still fight the dyslexic fits for sure..but seem to be getting less typos. This is the keyboard I pounded out 50,000 words on in 5.5 days last November. I can see why I wasn\'t getting as much done as of late. It\'s like a guitar or something- you get used to the neck. Yeah, you can play other guitars, but there is a certain radius and scale length that works best for you always....Also, it seems like my hands hurt a lot less using this one..and my pace is at least twice as fast even with a few mistakes. Anyway- enough technical blissery....     I hate clothing of this century. Men\'s clothing hasn\'t really been cool since the 1850\'s, but that last time a man got to dress like a man was the 1940\'s. Then came the 50\'s and all the rock and roll and bluejean bullcrap. Men\'s clothing lines got softer and softer and I hate it. I have always been much prone to very old clothing or WWII type military cuts of things. In the early 80\'s I took the sleeves off a Korean War heavy army shirt...stiched on all black buttons and then dyed it black. Like a moron I wore it when I met Bono of U2 that tour..and I swear..like a month later he was wearing his custom made sleeveless black military cut shirt all over the world. It pissed me off because everyone thought I was the one who stole the idea. Mike Peters of The Alarm has a great story about Bono, Dave Sharp\'s most treasured expensive new hat and a night the newspapers caught \"IT\" on film.....  I figured that Bono didn\'t directly steal it..but I am sure that the influence was there in the back of his memory somewhere. Funny...The Alarm was already heading in that direction with all the black cowboy stuff and later black Dickies stuff.  So...I most regret loaning my favorite peice of clothing ever to the bitch who fucked up my life. One night it was cold, and as always, she hadn\'t brought anything except the shirt she was wearing. I had my always present French Army Issue Field Jacket with me. The REAL field jackets..not these crap things the Air Force is issuing now....so I loaned it to her. I have never seen it since. I asked for it back a number of times to be told \"I gave it back to you already!\" Yeah...right..I totally wouldn\'t notice having received back my favorite peice of clothing ever.  So I have been on the hunt for one like it ever since. I have found a few close, but nothing that has that \"Zing\" to it this one had. It has to have four pockets on the front, epualets on the shoulders, long sleeves, expanding pockets for the lowr..and a belt WITH only two loops on the back...and the ever popular expanding back in a true 1938 style.  I found but one so far- and it\'s exactly the cut, color and even BUTTONS I want...BUT...it\'s $300.00...! It now has become a goal, as I find what I actually had was a damned collectable vintage jacket from the 1940\'s made by a company that supplied many militaries and the public..and they went out of business in the 1980\'s. Funny, I probably got one of the last original ones. Until then, I have found something that is probably about 85% correct and I plan to get it. I am on the way to losing weight again and once I get to where I was a few months ago before getting so sick..I will take it to my tailor to have it tucked and tapered at the waist..to have a more MALE look to it! I HATE the soft, dumpy, mall=crap feminized clothing they \"allow\" men now! I am revolting against it..now!  So I have a number of fedoras..and I yanked the fancy iron with the fancy steam setting out of the closet this evening and re-shaped the crush on the crown of one, and reset the brim on the other. This has lead me to a mad plan for winter. I am going to get a pattern I found and then find a sewing machine...I have every intention of making a black field jacket with some extra trim I have seen..around the wrists and the such. Just a few slight changes. Of course, the more I think about it, I\'ll probably end up with something looking like an Officer\'s Jacket from Babylon Five! That, however, would be cool by me..I would love it. Anyway, I can\'t see what the big deal about sewing is. I mean, I have built operating and flying models that had to be covered with japanese tissue paper that you had to apply with twizzers, I have rebuilt all manner of cars and machines...done body work, and can figure out anything that is three dimensional. I am rather tired of hearing how hard sewing is (I used to have a good friend years ago who was a tailor, as was his family..he said it was easy as hell to him after building a hot rod), so once I get the Jeep sold I will buy a really hot sewing machine and dive in. Maybe I\'ll get a Husquavana...I can\'t believe they make fancy sewing machines..I like their chain saws..! I saw a computer controlled machine a while back..you program your computer, interface it to the sewing machine, and it changes sticthes, threads...all sorts of things..in mid-stitch. That\'s not a woman\'s thing..that\'s a MACHINE that a MAN must own. I wonder if I can put headers and duals on it...?  Speaking of the coolest clothing on the planet..I saw..IN PERSON..the clothing Sean Connery wore in the first Highlander movie. It was Oh So Cool..and for sale..only $35,000.00! Crap...lottery anyone?  So, I guess I must be feeling healthier if I am THINKING about cool clothing again for the first ime in almost a year. Gotta knock the tonnage back off now. It will happen, it did before....  Oh for a nice Waistcoat and Frock Coat to cover it......knee-high boots....a dress sabre hanging low (hidden knives and psitols..)     Men used to get to look cool! Damn it...! Time for the Revoultion!'),(1928,1969237,'12,August,2004','2004-08-12 00:00:00','   The weird thing is..for some unkonown reason the words just popped into mind one night. The other day I did a Yahoo search on them...and found that it has some vast biblical meaning behind it. Many folks are all twisted up over Pig Slop. How fucking weird is that?  Today I have to eat a decent meal- because whatever I ate yesterday kept me awake all night and feeling sick. I\'m none to well right now as a matter of fact. Ick. I was searching the internet for information about the whole West Nile thing and have come to think that maybe my problems starting in October were a case of that. It isn\'t always fatal it seems...but it\'s much more widespread than the public is being lead to believe.  Oh well..I wish the headache would go away..I need to solder some tiny wires very soon...then work on the car door....later...world...  '),(1929,1969237,'11,August,2004','2004-08-11 00:00:00','   So...the night hasn\'t been a total waste. I managed to find the missing outlines of some of the novels I have in mind. I also found a Word file of the titles and a blurb for each one. This gave me the ability to see that I need to not only lay out the basic elements such as characters, but also that I need to have the series build as they go..start with a simple plot type and move forward to more complex with each book. One of the things many writers seem to forget is that in between each book people grow older, hopoefully wiser and change. I intend to make each book in each series an increasingly complex type of story line. This doesn\'t mean it can\'t be clearly written, it just means I challenege myself more with each one. To just write the same book over and over again just isn\'t correct. I could name some folks of great acclaim that did just that. I figure they were trapped into it by publishers. My deal will be to have the first novel of each series written before trying any submissions, and the rest of the series already outlined..so all I have to do is write. I wonder if there is a record for finishing books? I remember talking to Martin Caiden (the old dog!) when he had like 85 books done..he said, \"What am I going to write NEXT? I have the next SIX DONE already!\" Then again, interveiwing WWII vets and typing out their stories isn\'t nearly as hard as novels...but he did have a fair number of them also come to think of it...I wonder if I can ever afford a real Junkers JU-52?  Part of the writing problem is this keyboard..which is sized weird compared to all my other ones..ever. I need to dig something else out and try it...maybe the spacing will allow me to work better with less errors.    Novel Overveiw:           There are five series of novels waiting to be written, with a total of  at least 16 books that I have at least a basic idea for. Currently, three novels are partly completed in Word, while a fourth is completed in handwriting.  In the  Novel I  series, the novel has a total of 68,590 words done. In the  Historical Fiction  Series, the novel has a total of 62,723 words done. In the  SF  series, the novel has a total of 12,791 words done. The  Handwritten Novel  is guessed to have about 110,000 words total.   Novel I  needs to have the first 50,000 words rewritten to change the point of veiw of the character telling the story and toi increase interaction in the proper fashion between characters.  The  HF  novel needs to be finished in rough draft, and then rewritten to smooth out the annoying problems between characters. I need to settle on the triangles I want to use. I expect it to be a shifting triangle of charcters surrounding the main chacrter...each time changing in each of the main three acts.  The  SF  Novel needs to be outlined before going any farther, to clarify what the hell the intent of the thing is. Pretty pictures of things don\'t make a novel..that\'s a travel guide.  The  Handwritten  Novel needs to be entered into Word, the missing notebook reconstructed and then the story line ironed out so it is crisp and shiney. I expect to read it all, then outline it before entering it into the Word format. This shoul make the process better.  Overall, I have the outline of basic ideas to scribble on and put notes on tomorrow. Things like which plot type to use. Once this is set up, I can load stuff into the outlining program and get cracking. That way, if I am not in a writing mood, I can toss ideas into the program and still be making progress.  I wondered  WHY  my mind seemed to be working a touch better and more creatively tonight and just realized...nobody is working on the freeway that\'s 1/8 of a mile a way!  No explosions, no heavy equipment beeping and grinding and shaking all night long. No flashing lights and beams of insanely bright light spilling out over everything...and no mass of over 1000 idling cars backed up both ways for at least a mile..with grumpy, stupid drivers blowing their horns at each other.  There is actually   AIR   to   BREATHE   tonight..thats not laden with exhaust fumes so think it sets my allergies off. It HAS to be bad for my health to live here...when something as simple as this break makes a profound change in my feeling and output.   OK..sleep seems to be NEXT in the outline of life...tomorrow we work on doors and then words.  Wage Peace!  '),(1930,1969237,'11,August,2004','2004-08-11 00:00:00','   So I have read much of the Beats output...traveled to the hallowed grounds they once tramped about on and sat and thought about all of it.  Most of the stories are written in that 1950\'s pacing I like so much. It worked for Beats, it was great in SF and many others seemed to catch that certain...tempo and cadence back then.  So many get wrapped up in the wild stories from the road. I\'m sure there were many never told that topped all that was written anyway. I seem to have far outpaced any of the stories yet read...and...have found....  ...that the novels were written many years after the events upon loving and kind refelction- with the clarity of distance afforded by time passing. I could never write what went down because it just went down...now I look back and can.  I can think of the women I SHOULD have fucked, or loved or married or whatever. I can think of the chances and educations I should have sought and pursued. I can think of the things I should not have done- and there is enough distance to make some sense of them now.  Jack K sat on those early TV shows and people just never got it that he wasn\'t the hero of the story anymore. That was then..he was now. He was put off a bit too I suppose...everyoine wanting to ask questions about years and years ago...just didn\'t quite fathom it I suppose. Why should he, he was the first in history to go through this sort of thing?  So one of my problems is losing complete interest in anything that has happened to me thus far in life. Makes it hard to write out novels using things I saw out of context. I suppose, like a confession, I need to get it all down and done on paper anyway.Maybe then my mind will be free enough to move to other forms.  I intend to write a number of formats- not just stick to one. Fiction, Historical Fiction and SF about covers it. Why get pinned down?   I love it when a woman slaps me...I always respond with \"Wow...cool..usually I have to pay EXTRA for that...!\". It never fails to lock them in their tracks.....now..to find a woman who hasn\'t slapped me....;<>!  Boy I got home late...oh well..such is life..who am I to question it?  '),(1931,1969237,'11,August,2004','2004-08-11 00:00:00','   All day...20 minute segemnts so far. It sucks. Grind, dust off...hit with wax and grease remover..dry for 5 minutes...primer...let dry for 20 minutes...wet sand with 600 grit....let dry for 20 minutes...spot putty...let dry for 20 minutes.....wet sand...let dry for 20 minutes...primer....etc. the whole day has been a set of 20 minute segments.  I hope to get to bed early, so I can awake by 7:00 am and start on things. I am working on actual damage..so it needs to be finished before going on to smaller spot repairs. I want this damn thing painted and clear coated by this time tomorrow.  Oh well...trying to get used to not getting to class an hour early..students don\'t show until 5 til anyway. I am going to cut back Tuesday and Wed to getting there 15 minutes early instead of half an hour. I can\'t stand to be late though.....  If possible I will wet sand in the garage tonight to be better prepped for tomorrow morning.  Off to work....  '),(1932,1969237,'09,August,2004','2004-08-09 00:00:00','   Considering how much I am moving my feet and how little distance I have covered...the old one step forward and two steps back seems to apply.  The repair on the escort door went bad..very slowly. It was perfect to the point I was surprised at it. Days later it had a bubble appear...I thought in the paint..but sanding revelaed it wasn\'t a paint issue...it was a bad bond between the bare metal and the body material. 80% stuck just fine, so this is just weird. Maybe it wasn\'t mixed compltly...who can tell? It\'s ground off now, I\'m waiting to clean the area up an hour to see if anything lifts at the edges. Back to square one for things. I need to finish all the car projects. I don\'t even care about the minor repeirs on the Miata anymore.  I HATE summer here. Its too damned hot, and all the students vanish on vacation or whatever. Many never come back it seems. I was looking over the last few years of things. I need to get a postcard sent out to counter this habit- but that, of course, requires having money.  Money is in very short supply, having not enough to pay the main locations rent- much less bills after that is paid. I sent a message to everyone to pay for their lessons ASAP...because I am too broke to cover things another slow month. It gets into a speeding ticket, and TWO cars to be smogged in the same month. The cost of keeping the Jeep current knocked the hell out of my finances for sure. Time to sell it to just survive I suppose. Then there are all the benefits of actually having advertising money to spend. In the end I suppose it will put me far ahead anyway.  I sort of have a mental plan to finish the car issues before trying to do any other 9-5 job so I can actually complete things for once. I find that finishing all the damn projects gives me more peice of mind.  OK...back to grinding and sanding and patching stuff.  '),(1933,1969237,'09,August,2004','2004-08-09 00:00:00','   Well..the door on the escort has become a Problem, in each attempt to make it better, has, of course, made it worse. that means grind it all down to bare metal and do it again. Oh well. I guess I need the practice?  The repair of the 500 dollar scoring reels doesn\'t work. The replacement wire is too big and won\'t move through the roller system..so I\'m out 100 bucks for nothing..and very bad timing for that also. Maybe I can rebuild one of the old godawful French devices instead....  The swamp cooler is doing less than nothing today- making it sticky in here instead of cool. Such is life..I got laundry done at least...!  Time to send an e-mail to the  urlLink $@!#%^$  City and cancel the program there. Too much hassle, not any income..too many annoying \"Underprivileged Youth\". They can have their dump back..but don\'t ever think of moving to MY neighborhood with that attitude..I\'ll give them the treatment they have afforded me. What goes around comes around..Instant Karmas gonna get them someday. There is a reason some people stay at the BOTTOM of the heap..they belong there....(oh my their tattoos impress me..what\'s that one? Your name? How creative! Your gang, oh that\'s nice..you SHOULD be treated as an equal because you cut class, dropped out of school, got somebody else preggars and now my taxes are paying to raise the child, you\'ve attacked people, abused drugs and generally don\'t give a fuck about anyone but yourself...!)  Let Steel Sing.  '),(1934,1969237,'08,August,2004','2004-08-08 00:00:00','   I find as I get older things change. In my youth I\'d give anybody my body, and never my heart. Now, I\'ll withhold the body part until I am sure about the heart. In some galaxies that\'s known as progress. Fancy That...  Anyway, I\'ve gotten fuck all done all weekend. Oh sure, I worked on the Car and its paint issue, and made decent progress. I read a book..but mostly, I have been watching two seasons of a fine SF TV show I never got to see before. I find it a great show and consider myself a fan now. That, actually..IS progress. I slowed down a little.  I as of late have been looking for the simplest answers to all my problems. You know \"I need a place to live\"...what\'s the cheapest immediate answer...Mobile Home..sell a car to get a down payment and move.  \"I don\'t have enough space to live in\" - Hmm..there are still like 600 LP\'s and a rack from a record store to hold them...plus tapes and CD\'s. Answer ..MP3\'s. I got a program to record and clean up all the LP\'s. I\'ll sell them off as they are mostly collectable ones that are left- and keep only the complete collections and the autographed ones...that\'ll clear out a LOT of space. Also, I can then get a lot smaller player...as I won\'t need both a CD player, a tape player and a way to plug in a turntable anymore. All this space savings means a small place becomes a bigger space in the long run. Players seem to be about 100 bucks. I\'ll wait until I get a bunch of stuff converted before investing in that part.  \"Its too hot living here\"....well..I was going to spend a few huindred dollars I couldn\'t afford on an A/C unit..instead..I spent a few hours just shaving the damn hair off my arms, legs and neck...like when I was avidly into bicycling. I feel 30 degrees cooler instantly..and it should help on working outside the rest of the week also! That\'s CHEAP. I got the idea when realizing that Peacekeepers in a film couldn\'t take the heat...art becomes life.  Simplcity really is the best idea..go directly at the issue and be done with it. I need to knock the diet back in shape. It\'s amazing how much just one decent meal last week made me feel better. Funny..I have no desire at all for beer or wine at this point..but crave a hot tea or maybe a coffee. How very funny. Then again, I am sick of the bar enviroment. Its hard whenever you go to meet friends at a pub and they already have a Stout waiting when you walk in. Geez..the image got larger than life it seems.  I need to go to the beach and swim in the salt water fairly soon. Maybe I should do it one day when the rest of the world is at work...like early..so I am like a local. Salt water is good for many ills. I wish I had a pool to swim laps in around here. That needs to be something to look for when I move.  Funny, tonight I find out my place is next on the list to finally get repaired..after I have finally let go the idea of staying here.  I find that when I practice peace, peace surrounds me. This could mean that people remain calmer, or that bad things just don\'t happen at all. This weekend has been an experiment in peacefulness and has worked well enough. I feel like I got a base line reading on balance again somehow.  Wage Peace  '),(1935,1969237,'08,August,2004','2004-08-08 00:00:00','   Well..there\'s nothing like a good horn section to spice life up..is there? You know Gabriel plays trumpet..after all....  So...a student loaned me a TV series on disc because I had a strong interest in one of the people playing a character in it. I came to a point of wanting to know more about this person not because of their movie work, but because of the rest of the things in their life. Now I find I shouldn\'t have been avoiding the series after all. The short part is, its well written and well made. Like little else I\'ve seen in years. The long part is- how close this person\'s character seems to be to the part of the person I stumbled onto and was interested in.  I\'ve met the person a number of times, and become hopelessly toungut-tied in their presence. This has only happened one other time in my LIFE..and that would be when meeting Eric Clapton many years ago. It simply doesn\'t happen to me...usually, when I get nervous..I talk too much. Now..I say just a little..and NONE of it is what my MIND wants to say. I am completly disconnected from what I intend to express...its maddening.  So..they have a passion for an area of my life I wish I knew more about- but I happen to have stumbled onto something I think would be of life-changing importance to them and have tried to make them aware of it...except for the total lack of comminucation thing I have going on. I have exchanged books that would point them in a certain direction...Findhorn..and now have an increasing draw to get my sorry butt there as soon as I can. The Jeep is named Pan...perhaps releasing Pan to wander with a new owner guides me to his mountains sooner?  I am trying to write a letter to the person I need to express the thoughts to as they may leave the country for good at some point in the near future...so I am trying to swallow my....fear..and act. Life is again teaching me the need for a simplicity of action, a complexity of thought, a saving grace of hope and a creavite desire for the future.  When I was very young, before I discovered music, the only thing that mattered was Space. It was right next door with all the NASA launches, and I still defend and beleive in the need for mankind to explore. I have noticed that there are only two times mankind makes great advances in understanding...war time or times of great exploration. Granted, there is much conflict in the way we have explored thus far, but there is a changing understanding as the world grows closer together.  Today, in Anaheim I explained that I had made a rather long drive to get to the shop I was at and they asked \"Where are you from?\"..without thinking I gave the BEST answer I ever gave..I simply said, \"Earth...\". Then I moved on before I could be questioned. I LOVE that thought. I mean, in the end..if giant space goats with tentacles were attacking the planet...EVERYBODY from Earth would be friends...right? Why wait for a war to push us into that understanding?  What, he\'s going hippy on us again? No...something far more ancient and responsible. Back past Crop Circles and Standing Stones I suspect. Back to something older, deeper and richer in texture. No..there are no words for it yet.  In youth, Space was the dream. Life has been a drifting horror without that dream to look forward to. The governemtn gutted my beloved NASA and caused destruction of so much of what made them great. Now...out in the desert..a couple of modern day Wright Brothers hitch on a knowing smile and toss it out there to see if they can make it work. I saw the first time they did. The next goal is the Ansari X Prize. I will be there to see that step into Space taken also....then...I have a goal.  All the writing, all the working, all the teaching, all the planning...is towards a ride on Spaceship One. First guesses are at a price less than that of a house. If I have to choose a house in my lifetime, or a trip into Space...fuck mowing lawns!  I read so much SF in my youth it was amazing. I remember many things that were expected when I see them happen. Its amazing that a Palm has more computing power than ALL the computers on the first shuttle....god how amazing!!! I seem to have reached aplace where hoping for Space and trying to inspire the kids around me to be a part of making it happen are the most important things to me...Maybe someday the old coach will smile behind the pilot of a Space Craft taking him for a ride...and remember their first stumbling steps as a fencer...and their journey towards themselves.  Yes..more and more ..sopmething of a place. If I can\'t be the kids in \"October Skies\", I CAN be the teacher who believed in them! Of course, I am far too stubborn to stop there. Now I really need a few more lifetimes to reach my dreams again.....I need to rage.  Living one step at a time..putting one foot in front of the other..healing on my mind...been a long way....a long way..to the light. (Words stolen but better than my own..!)  '),(1936,1969237,'07,August,2004','2004-08-07 00:00:00','   Back to earth for a check in it would seem. Yesterday was a LONG, LONG day. I shuttled back and forth between smog stations and finally the most beloved of all California creations...The Department of Motor Vehicles. The unknown secert aboiut DMV is this: If you join AAA for like 40 bucks a year, you get discounts, towing...and can do all tag and title related stuff at AAA instead of DMV. The folks there are a LOT nicer because nobody gets in their face the way they do at DMV. Now.... DMV could also stand to take a lesson from them about how to run an organization with intelligent and nice people also!  So...finally..the Jeep has been smogged, insured, tagged and is ready to go for another two years. It will need one more smogging EVER...in two years, and then is clear of the system. Still, as much fun as it is to drive around, the clutch has left my left knee in agony, and I seem to be dreaming of something...anything else.  It has felt like a prison sentance trying to get this whole deal wrapped up. Actually, it\'s been a trial. Had I failed it, I would have been sentanced to a lot of hassle correcting it all. I spent a great deal more money on all this than I had a budget to do. Having two cars come due the same month, needing smog the same years is too much. That in itself is reason enough for some fundimental major change.  Class is thinning yet again, but LOTS of people are on vacation. I guess I\'ll need to mail something to recover people back into the fold. The next session starts in about 3 weeks..so next week I should send postcards or something. I think I can pull attendence from my excel files now and see how to market to folks. I am bummed..a while bacj somebody\'s brother broke something at the facility and they made a big deal about it...the kid in class left and never came back because I am sure he was embarrassed over it all.  I hate it when somebdoy else causes a student to quit...its hard enough keeping people as it is.  AND..a student on vacation in Orange County was refered to a group where I know they have quality teaching and organization. Somehow, they stumbled onto a group I thought was gone. This group has a person who claims to have been coaching for like 30 years- has NO pedegodgy at all..and their vast staff of coaches has no training as coaches what so ever. Well..it gets better.  The student\'s mother, being upset that the local location for her is closing, has given this group the idea that I don\'t know what I am doing and need help. They now figure on moving into the area and setting up shop because they think it is ripe for picking! Christ...why do people get it all so wrong and then never bother to ask? In addition, this group brings in lots of young kids- they are a playground-NOT a training facility...so this student did well against them. Somehow because they had all these kids it equates to THEM really knowing what they were doing!?!?!?!? NONE of their students even place in local competitions...even their coaches don\'t fair well. So I may have to get ready for a turf fight. I believe mostly, that no one from their group will make the drive in traffic out here to actually do anything. If I find they are planning to do so then I will level them with pointed questions..and counteract their infleunce with a section on my own website before they get rooted here. If other places open in the are I have given up my free time, marriage and most of my life to build..they need to belong to ME. End of Story. I bring the Shake.  Moving forward, the intial repair on the Escort has gone south for the winter. I got an adhesion problem with the color coat- so it was a goodthing I waited before clear coating..otherwise it would have been even more of a waste...Time tos and it all off and start over again. I know what I did wrong also- I rushed the process and painted over something that probably wasn\'t cured enough...then, the color dried faster than the stuff below it...casuing a big old bubble to pop up....OPPS Live and Learn...  In the end, the Escort shouldn\'t take nearly as much to get done as the Jeep has..so I need to knock it out quickly. I\'m not going to do Northern this year- why bother? I may GO one day- but not work. Hmm..if I lst weight I could do the Dickens festival at the Palace in SF this year too..I have the clothing..all I need to complete it is a nice percussion cap pistol..maybe a revolver would work...go as a Scot who has moved to the new country of America....I have the perfect sword actually from taht time- not a repro...swords AND guns...that would be cool. Now all I need to find is a young woman to dress wild for the times and play the part of a fallen angel I\'ve taken a shining too....  Hmm..maybe I should lose weight sooner and go to the Exotic Erotic Halloween Ball. I had a freind who went some years ago- and any number of people offered to fuck the hell out of her while there. She was to die for- a Bondage Fairy. I have a picture somewhere. I would have offered too if I was there...;<>! (She has since gained like 50 pounds..it sorta ruins the effect of the fairy costume I expect- oh well, she\'s getting married now anyway- the third time is supposed to be the charm..I hope it is for her...)  OKOKOKOKOKOKOK  get rolling...need a blast to Anaheim for clear coat I think...see me later!  '),(1937,1969237,'06,August,2004','2004-08-06 00:00:00','   I had nightmares, very graphic and violent all night long- or at least in the span where I slept. Sort of end of the world by virus stuff. Then, a woman I know only slightly showed up in the dream to be like a mate or something...not my type at all....so weird. I can\'t figure what any of it means.  So far today got the jeep back and need to set an appointment for a recheck now the repair work has been done. Guess I need to get to it...  '),(1938,1969237,'05,August,2004','2004-08-05 00:00:00','   Imagination is the waking line between dreams and reality. Reality is a thin facade that covers the infinite in a way the knowingmind can not grasp. The sleeping mind, the mind of drugs, the drunken mad mind of Byron, the sickened sad mind of Shelly\'s Mary reach beyond what is seen and touch what is. Sometimes one can feel what is long enough to pull a taste of it bacl to this world that seems to be.  There was atime when I was laways altered, before I was injured, and my mind spun madly endlessly and I am sick with too many thoughts. I prayed that I would lose this torment and then I did in a tragedy. Now I dread the roaring silence that brews in my head....  But something stirs increasingly ...growing..perhaps madness to some, but to me a taste of the mind once that I had to hold. A place that was not of my own making that I was blessed to veiw and qoute from and make sense of to herion runied whores, little lost angels and runners in the night that sought death but found hope.  I have retreated from the normal world without normal employ yet seem stronger than yet ever. I have a place and it seems to need the space to find me and be....seen.  I dread the coming daylight..and all the noise it brings with it....  Tonight I have had enough chocolate to kill an army.  '),(1939,1969237,'05,August,2004','2004-08-05 00:00:00','   Ghosts are supposed to be trapped between life and death with a weight on the side of death. Are there white ghosts trapped on the side of life..still breathing but dead and trying for the transmutation that releases them into whatever will be or not be?  Do those ghosts we see..that I see actually want me to follow them to wherever? Is there a hereafter? I have no idea..but don\'t believe the christian pissings of a final revenge of an angry god against a mere mortal as I.  No..the other great religons have it more right...from the hot breath of Islam and its whirling and dervish understanding of the circle of life, to the hot temples of some Asian land wet with humidity, mildew, life and sex.  Christians believe in ruling the world we see..and submitting to the ones we don\'t as if they are all a great evil. I doubt this is so. Because...I have demanded answers of unseen gods and received them.  In all this I still feel like I am lost in time away from a century ago when I would have been fitted for the times. How I wish I had traveled with Burton and Speck through their trials....into glory...into knowledge...Darwin and The Beagle..abiding and respecting Fitzroy....or maybe as late with Carter and his amazing dig to the tombs. When you light a candle in the grave of a pharoh does it illuminate the way...or destroy the guidence that was actually there...?  1988 my most dear friend and confidant of youth, Tina was killed in a car wreck..the last thing we shared on this earth was my seeing her as the last of her friends before the coffin was locked tight...just after I slipped in the only token I could think of to show how much I still loved her and always would...a guitar pick...she, of all understood.  I felt something as we carried her to her grave, a brushing by or feeling I can\'t explain..hiding behind my long hair and fighting the tears she deserved the whole world see. Why couldn\'t I cry as I wanted to...let everyone know I had loved her so very much? Because her husband was there? Because I wished she had married me and never met this wreck on a bad road in Georgia..or because the gods would have played as cruel a trick on her if she ended up with me anyway? No..for I have found the one sure way to prolong a Bohdi\'s visit to this earth is to have a child borne by them. Then they have yet more to learn and must stay. Thank god for Annie\'s little Quiron....But alas, Tina had no child and was fast leaving this , this...plane? Was it her brushing by? Probably. I remember staring at her face..I still love her image so....and seeing the brusing covered with makeup to concel her death. My dear Tina...and knowing she had lived long enough to actually bruise. Did anyone else realize this? Oh the pain of that moment- and the burning question as to what EXACTLY is it that dwells in the human or any other body...which in can be knocked out with a hard blow so that the body lives no more?  And I got my answer some 2 weeks later in a rendering explosion of metal and glass smashing the person I used to be surely to death as its never come back. I know what it is that leaves. Words can\'t be spoken for a powerful feeling. But its been years that I have been trapped here..in between. Sometimes seeming all the world alive..sometimes a ghost of something that was all wrapped up into one.  I miss Kevin so. I feel him around me at all times...like now, whne the traffic is gone, the lights out around the area and its quiet enough to feel him here. I tried to say goodbye to all them to not trap them here by my pain of losing them. I lit candles and burned them all evening...each one was a friend..Kevin, Tina...Steve. I thought, in the morning I had let them go...but there is more- something primordal and ancient that still is here. I am not scared by it anymore- I accept it. Somehow. If I am to feel it, then its meant to be.  Tonight I wonder if it is myself who is trapped..and that THEY are trying to HELP me reach wherever they are. I\'m not in the mind of suicide. I am in the mind of wonderment at the way it all works. Just how have I made it so long..or have I?  I always swore to myself I\'d be dead by 27 for some reason. I wasn\'t..but my wreck was when I was 28..and it took some long years to understand that I had died..for I was not as I had ever been.  If one can call them angels..do they have trouble getting used to the change? Is THAT a ghost? Final retribution in hell..I think not.....it doesn\'t fit the way the material world flows around itself like water over rocks in a stream. Hell stops you in your tracks...even mountains melt in time.  What am I trying to say..I don\'t know. I just need to say it. It is what keeps me awake until all hours of the night- this...feeling ...of knowing....more. I am scared to death to go to Findhorn. I am scared to death to sell my jeep and have the money to go there...but it seems like a great wooden door hiding the rst of my understanding from me until I walk through it. Haunted when the minutes drag...  Perhaps its not even Findhorn itself..its something about getting there...I don\'t know. Late at night when faries dance and spirits take to wing... the harsh wind blows the moss about as gators start to sing.....the bullfrogs all have bedded down as crickets loudly sing..and kneeling low I touch the earth.... the soul of mother earth...its just another decade lost I tell the midnight sky...late at night.....  where did that song go......can I sing it or will beseeching it bring some of this change...?  Am I the one who is Trapped? I still don\'t know...  '),(1940,1969237,'05,August,2004','2004-08-05 00:00:00','   So..the jeep went for it\'s much vaulted smog testing today...and failed as I expected...I believe it is the carb I put on a year ago. The biggest problem is its running rich at highway speeds..which means they gave me the wrong carb as its jetted too rich...GGRREEAAATTT....I have to PAY somebody to fix it to prove I had it fixed. SCAM!  So its was a long ass day dealing with all the stuff on this issue..plus teaching.  Sadly, one of my students has a problem paying and was gonna have to quit. I need every penny I make..so..I bartered...! Now..I will have a series of radio ads on a small radio station in the area! I rock! I have to write a script now and practice the acting.... . Class went well tonight, with the students being driven into a frenzy while fencing to the likes of Jet, Julian Cope and PuffyAmiYumi....I love pulling this out on them..they don\'t know what to do...but they are exhausted at the end of the evening. You know, teaching is just like making mad love to a stunning woman..if they are exhausted at the end of a night ..they are happy. Man, what a fuck I am....(dumb, stupid, silly, goofy, loud..etc..).  Well...there is nobody to go with me to see rockets (Large and Dangerous Rocketships...) on the 14th and I am bumming....its so cool to be able to see them..its hard to find a group a third this size anywhere else in the US...I think everybody hates me. Why does no one want to drive out to the middle of fuck all nowhere and see giant fucking rockets being launched? I mean, some even blow up if that\'s your gig...I will go anyway. Maybe I am already becoming a rocket geek since no one likes me.....maybe this is the initiation or something...oh well.I still play better guitar than most of them I imagine....  What else..I may have a chance at some great advertising for 150 bucks..I have to see where the bucks stand tomorrow..I don\'t think I can afford it....maybe more reason to sell the jeep? Probably sell it, the ads go nuts..and I would have made enough to keep it. Oh well, the Miata would look good with the competition rollbar in it. And I could totally Rice the Escort...big wing on the back- grind off half the paint and primer it..put \"Eibach\" stickers all over it...no..wait..I LIKE the Escort....  Guess we\'ll see where life goes won\'t we?  \"The drummer thinks that he is all the rage....if you can\'t rock me...somebody will....if you can\'t rock me....\"  '),(1941,1969237,'04,August,2004','2004-08-04 00:00:00','   Which is yet another reason to not believe TOO strongly in this god fellow. I mean, his followers are off dashing about folding, spindling and mutalating the heathen unbelievers over....there..  So it\'s rather obvious he\'s just out to fuck with everyone, because he hasn\'t created anything worth a fuck since the dawn of time. ..so he wants to pick on people and keep them distracted until he figures out something that will really impress us like adding gills to all human babies for no reason, or maybe allowing Lance whateverhisnameis from the wanker boy band whoyamacallit go into space so he can find a GIANT monolith and then freak out when he flies over the end of it and tell the world \"It\'s full of STARS!\".....which will freak out all the people that own an IBM 8000 or something...  ...speaking of which..ever notice the name of the computer in \"2001\" is the \"HAL 9000\"? Yeah..shift those letters by just ONE PLACE and it\'s \"IBM!!!!!\" Damn! We\'re all doomed to live on the set of a place that looks like IKEA took over! Which, by the way, would be perfectly fucking fine with me as long as the women wore short skirts and those Go-Go Boots everywhere.  So I was the kid that read EVERY BOOK in the children\'s section of the local public library by the time I got out of 6th grade. I mean it. I read what I liked, then I just started at one side and read the rest as I came across them. It was Orlando, and its a big fucking kid\'s section. I had no idea what to do next, so I discovered microfiche and annoyed the hell out the librarians asking for fiche so I could keep up with the reading of reports for WWII submarines and the dawn of flight and maybe shit about Marconi or Tesla or something....until they told me to get into something else...so I found the \"Fine Arts\" section and began to check out books on antique dressors and stuff. that ruled. No other 7th grader had a clue when they invented machine made screws..but I DID! (I forgot now though..but I can tell a handmade screw when I see one...which...at one point, rather got me in trouble at The Getty for getting too close to a certain valuable peice of fluff..but the security guard dug the reason I was looking too..and we both got caught on our knees staring underneath this fine relic...so it was worth it).  So..when the Internet was invented, it was simply to curse ME, because I believe I have to visit every site on it before I can stop. I can\'t remember what it was like having time to do things in the evening anymore..like walks, bike rides, chasing women, drinking, street racing...damn..what the HELL am I doing?  BUT..I DO know the benefits of building a Jet Engine out of an old turbocharger so as to cool your beer faster and have a reason to use your shed in Australia. I DO know who makes the smallest gas powered turbine engine in production (Wren in England...). I DO know the exact time of the Atomic Clock in Boulder, Colorado without reaching for my Shortwave Radio...I do  know  a lot of things....  Like...\"Whatever happened to Zodiac Mindwarp and the Love Reaction\"? Or maybe what are the on-going adventures of \"The Arch-Druid Julian Cope\"? I even found out what\'s going on with \"Dumpy\'s Rusty Nuts\"...!  Hmm..that reminds me.....What was the name of that hot chick that played \"Holly\" in the original \"Land of the Lost\"? She was such a babe....I gotta go....I think I hear a Sleestack....  '),(1942,1969237,'04,August,2004','2004-08-04 00:00:00','   So, anyway, enough whining. What I really can\'t explain is why I have had songs from the almighty Hanoi Rocks just pounding in my head for about two months. What a great band they were..I am still bummed out and annoyed that they didn\'t try and get a new drummer after Razzle was killed (FUCK Vince Neil FOREVER for killing Raz...no forgiveness....he didn\'t even get serious jail time!).  I played a couple of live things I found on disc, but need \"Two Steps from the Move\" on disc..the production was amazing as were the songs. I did get the new studio one of the reformed group, it\'s pretty good. I need to hear it again..tried to listen to it in the car on the way to Comic Con but the friend my buddy picked up started in about \"All those crappy hair bands from the 80\'s that he never got into..\" Funny words for someone in their late 30\'s who has probably never had a date, lives with like 3 other guys that are the same age and have focus on the things 14 year olds do...gaming, computers, etc. Guess he didn\'t know what a hair band was anyway..because the Rocks certainly were NOT just a show- they were a fuck of a band.  Yes, The Hanoi Rocks were a great band. The four most amazing bands of the 1980\'s were The Alarm, Marillion, The Hanoi Rocks and The Rainmakers. Each band was great on recorded output, amazing live and never got their proper respect or recognition, The Alarm beat the path every \"Alt acoustic rocker\" band followed..even taught U2 how to play \"Maggie\'s Farm\" ( which they did at LiveAid and the whole world freaked out over it...). Marillion just never got noticed....that is a sin. The Rocks started to make it in the USA until that fuck of a loser from Motley Crap, Vince Neil murdered their drummer in a drunk driving accident. (How many people are aware it was VN\'s second drunk driving wreck with injuries, Raz was killed, the other passenger was badly hurt and a woman in one of the other cars he hit was nearly paralyzed..? For this he got like 15 days and 300 hours of public service..must be nice to have the record company BUY YOUR WAY OUT OF RESPONSIBILITY.) As for The Rainmakers..every fucking so-called \"Alternative Rocker\" owes their living to them directly.  BUT..Rap (so-called) music came along and the focus on rock went away immediately. Was it because Rap was better or cooler? Hell no..it was because MTV bought into the disposable music scam that Berry Gordy invented with Motown Records. Nobody remembers the class action lawsuit against him for unpaid royalties by like 90% of the people that were ever on his label..even his own SON sued him.  It\'s simple..get songwriters on board writing for a flat fee. Find people wth an avearge amount of talent, toss them a great tune with great production and promote the hell out of it. Then send them out to tour until they run into utter exhaustion and get a drug/ drinking habit. By then, the next sucker is recording..so you let the first one fail or gice them crap for the next album and then drop them. Before they ever figure out the business they are in they are gone. It worked like a charm for him, it works even better for Rappers. Why do you think only a tiny percentage of rap ever has a SECOND big hit record? Because they are getting used by the people they put their trust in is why....  I had a very enlighteneing conversation with a very famous singer in a very famous heavy Metal band a few years ago. He had a reason to spend a lot of time at MTV a number of years ago, and overheard two exec\'s talking on and on about how rap was the next big thing, and that they were getting rid of all the Metal and Rock to make room. he finally got annoyed at them and told them Metal wouldn\'t die that easy...etc. A year later, his major, major band couldn\'t get a vid on MTV for any amount of trying..and gigs dried up as younger people all went over to the rap that was flooding ther airwaves. Since then, through personal marketing this band has rebuilt a solid business and tours all on their own..while 99% of their buddies have vanished to work at Marks and Sparks or someplace.  The truth is, MOST people in the USA blindly follow whatever is fed to them..be it music, fashion, sports or going to war. They just don\'t think for themselves. Well, damn it, if everybody else is making money of this blind worship then I should as well.  Once again, FUCK Vince Neil...he deserves whatever evil befalls him.  '),(1943,1969237,'03,August,2004','2004-08-03 00:00:00','   Yes..It\'s ON! Scaled Composites has made it official, the dates for the launch of Spaceship One to win the X Prize are Sept 29, 2004 and Oct 4, 2004...which happens to be the 47th anniversary of Sputnik!  I was at the launch when they made it to claim the prize as the first civilian astronaut in space..I have to be part of this also.....I am SO THERE!  Funny, \"October Skies\" was on last weekend on TV and I saw it. I miss the way NASA was before the reagan government screwed it under. People can say what they want to about my comment..I lived in Florida and 60% of my friends had immedaite family at KSC. I tried to get hired there when I got my degree in Computer Electronis...but nobody was hiring as reagan had cut the budget so badly...the bastard. So unless you lived within 80 miles of KSC..your comments to the negitive don\'t matter..because YOU didn\'t LIVE it.  Now..to see if I can inspire any of the teens from the fencing group into going....  Details as they come out are on the website for Scaled Composites at  urlLink www.scaled.com  .     The sky is no longer the limit!  '),(1944,1969237,'02,August,2004','2004-08-02 00:00:00','   Spent so far today working on the door damage on the car. Got most of the actual bodywork done yesterday..but as ever I am having issues with pinholes in the primer and a couple of spots that are just being a bitch in general. I seem to have worked them out by filling, sanding and primering with different types of supplies. I beleive the main problem is the lack of a pro facility to work in with all the toys and tools they have to work with. I can do the work, it just takes longer.  I\'ve just shot the color coat and figure to give it two days drying time before wetsanding it. I know the label says I can do it now..in 20 minutes..but I found better results when it has a couple of days. I intend to shot 3 color coats to give good, even coverage and probably at least 4 clear coats to ensure as shiney a finish as the rest of the car.  The front bumper has one crack in it from hitting a dog on the way to a fencing competition years ago. I believe it shouldn\'t be as much of an issue to patch as the door..as the door was cruched and rippled and metal. The bumper is plastic..grind it away and fill it in with kitty hair...sand it to shape. I should have enough color paint to do all the details I need to do. I do need another can od clear though. Once I get the bills paid and some income from this monet maybe I can just run to Coast AB and pick up supplies there. In all, the painting part is the easiest part- the bodywork is the bitch. I look forward to finishing the repair projects so I am free to experiment on skulls and other figures on things. I am starting to get the hang of using the airbrush. I might even be able to do a Roger Dean inspired logo of my own...for a display item at the main location. I think if I can get it as nice as I want they\'ll put it up for sure.   Life seems somewhat balanced..yes, bills are lacking and there are a couple of things hanging like that famous fellow\'s sword over my head- but that\'s OK. Tomorrow I need to get active on the smog cert for the Jeep. That buys a lot of time once done...and some relaxation overall.  I can\'t work on the rest of the car for a few days, so it\'s back inside tomorrow to work on business stuff like testing and time permitting get into the novel yet again. I figure one 7 day week, 8 hours  a day would complete the rough draft of it. THEN..I need to do a seriuos outline to be ready for Nanowrimo in November...I can work the SF novel in the meantime. So..finish last year\'s NaNo novel, fill the space until Novemeber with the SF novel- and if I finish that pull the first one out and do the data entry and rebuilding for the stolen notebook for it. There is a possiblity of having three novels done by December at this point. If so, THEN I can shop an agent and a publisher. I find so many authors DON\'T actually work on their job of writing, I want to show a back log of completed work to potential publishers. I also want to surprise them in my willingness to get the editing done and the work complete ASAP. Why should it be any different from any other job? I\'m not pissy enough to be a whining little artist type..that\'s too much self-importance and not enough balls to get the job done.  I need coffee, hot tea and Boba Tea very soon...(not at the same time mind you..just over the next few days..)  Rock on..  '),(1945,1969237,'02,August,2004','2004-08-02 00:00:00','   So..I handled the Tax dealy this morning and say the thing that caused all the problems in its current;y fatted and disgusting shape last night..I have moved on as of this morning.   I have just received word that the woman I was transferred under in my 9-5 (which I quit last week..) has stepped down from her manager position, and the little group I am in has been transfered to someone else..who...rumor has it, is the top seller in the area. My friend and manager will go check it out sometime this week. He\'s going to push again on the big account I lined up a year and a half ago that hasn\'t been closed yet. I was under the impression it was being \"Saved\" for the woman who quit.  So much for her anyway. A goodie two shoes woman with like 4 kids..by 4 men, and she was only like 28 or something. THEN, of course, she found a rich guy twice her age and with a ton of money...who was all into god..so now she is too. She lied flat out to us about being on appointments, when she was giving church birthday parties, skipping days and doing all sorts of things. Yet this is the woman who was writing bad reports on people like me when a car broke down on the way to work and we\'d spend all night fixing it to get there the next day. FUCK HER!  I wonder what will happen next...?!?!?!/  '),(1946,1969237,'01,August,2004','2004-08-01 00:00:00','   Drove to Riverside to the State Tax Board. It took all of 20 minutes to cancel the permit that showed up under my name for no good reason. Its been years- what was so hard about that for my ex-partner? She truely is a self-centered, selfish little...er big flake of monumental porportions.  It is interesting, I seem to be the white person in the State that is in business. Everyone working there and waiting was Mexican. What can that mean?  '),(1947,1969237,'01,August,2004','2004-08-01 00:00:00','   in my KNEE! Geez! Woke up..or should I say..failed to sleep because my left knee began to hurt like hell last night. This would be a result of fencing on Friday. I am getting closer to considering my competitive days as a thing of the past..I can\'t deal with this kind of pain much longer- esp if I don\'t have to...guess I shoulda kept up on the Glucosumne and MSM..opps...   Off to resolve the state tax issue this morning with a big old southern grin. That gets farther than threats.....we\'ll see at least...  '),(1948,1969237,'01,August,2004','2004-08-01 00:00:00','   WOW...sometimes timing is everything. Well, timing and razor fast reactions such as mine. I was on the huint just moments ago, for meat and food for the tribe when suddenly...    Actually, I was at the store buying bread when I almost walked down an aisle. There was this short but HUGE woman in about her mid 50\'s reaching for something on a shelf. She was buck ugly in a \"Let yourself go sugar, you\'re white trash and should be proud of it...\" sort of fashion. The clothing hung loosely off the short but undulating body. Funny thing that...loose clothing DOESN\'T hid the fact someone is grossly out of shape...   Then, in an instant, I realized..it was my ex-business partner that fucked me under so badly! OH MY GOD! What a wreck! What a PIG! There is no way she can make the normal fat girl claim of \"I have big bones..\" as she didn\'t used to look like that. Her hair looked like crap, was too bushy and colored for the roundness of her fat face. Besides honey, that style went out in about 1987!    How fat do I mean...well..she is as wide as a shopping cart..and nearly 5\'4\" tall. That should paint a ...well...ickers...picture! Yes, I put my dick in her years and years ago, but I don\'t know how she still has sex at this point, and by the looks of her expression don\'t think she really does. It was so disgusting. So VERY disgusting..no one should ever let themselves go like that.    So, I made my way through the express line by dropping a cuple of things off to cut my count..I DID pick up a huge peice of chocoalte cake which I have now wolfed down...for god sakes I look GREAT! I DESERVE to splurge. I am thinking the incredibly fat begger I ignored on the way out was her HUSBAND!!!!!!  Thank you Universe, I needed a boost and a laugh. I have WORK to do, a life to lead and a whole world to save....! We March!  '),(1949,1969237,'01,August,2004','2004-08-01 00:00:00','   Went to the mailbox. It never has anything but junk mail and bills anyway- so why bother staying current. Got a form from the \"State Board of Equalization\" looking for payment of taxes from the company my ex partner embezzeled from, stole from, ignored, left and I closed officialy the instant she said she quit. Funny, there was a sticker over her name with my home address on it and my name...how did this happen? I would guess SHE doesn\'t want to pay back taxes and is trying to pin it on me. I wish she and her husband and their families were quite simply dead. Drunks, whiners, losers and theives all from what I can tell. A waste of human space and resources.  Seems she wanted to sell equipment. I kind of entertained the idea, but she was all into it. HOWEVER, in six years she couldn\'t bring herself to show me ANY of the accounting..I never took an order for products or equipment, believe she did, have no idea if anyone paid for them or who her source was. Last I saw when I checked the permit was listed on-line in HER name..not mine..as it should be. I had come to the conclusion she was still suppling her buddies with stuff and blowing off the state. Seeing how she doesn\'t return any calls or anything..who can tell?  Monday I will call the board and explain to them that as far as I know, the taxes WERE paid the last year as a business together, and I have had no farher dealings with her since. Time to throw it back in HER lap.  I have decided to not follow any news about the so-called \"war\" or the election. I already know this country is a diaster, and don\'t need any more negitive stuff in my own life. I went through the blogs I had bookmarked and deleated all the negitive ones I was following. You know, 350 pound guys whining about why their wife hates them, angry ones about politics..all that. I have retained those of passion and value..and, of course, poetry...as something of great importance to me.  This tax thing seals it. I need to move OUT of this town so the bitch from hell can\'t find me.  She is pure evil and I don\'t want to be within 30 miles of her for any reason...ever again. Funny, I think I finished the Jeep tonight..I may have some tune-up stuff..but as far as all the work..it looks done. It took this long to right all the wrongs she caused with this vehicle when I loaned it to her for a few short months. Fuck her all to hell..in a redneck kinda way...!  '),(1950,4128156,'05,August,2004','2004-08-05 00:00:00','   BOGUS BLOOD BAN   Jehovah\'s Witnesses adhere to an unscriptural position for Christians ( kosher law)              How risky is a Blood transfusion?( low)  A comparison of transfusion risks to the risks of dying from other common causes:  Serious disease or death from transfusion: 1 chance in 200,000 / year Death from electrocution: 1 chance in 200,000 / year Death from drowning: 1 chance in 35,000 / year Death from a car accident: 1 chance in 7,000 / year Death from pneumonia or influenza: 1 chance in 3,000 / year Death from cancer: 1 chance in 500 / year Death from stroke or heart disease: 1 chance in 300 / year Source: Canadian Hemophilia Society  Check out these links:       [   cut and paste to browser   ]   http://www.ajwrb.org/   http://www.towertotruth.net/Articles/blood_transfusions.htm  http://tlc.discovery.com/fansites/trauma/case/case_07.html [Discovery Channel Video]  http://www.pennhealth.com/health_info/bloodless/blood_jchart.html           The Watchtower\'s Bogus Blood ban.It is so twisted, ( to obstruct lawsuits ). Selected, JW elders receive special,(  esoteric  ) indoctrination just to try to make sense of it.          Remember,always do that,  \'regression analysis\' .Just where the heck did it all come from ?   ........Well,Watchtower Leader \'detox\' Joe Rutherford,got a \'burr\' on his butt,back in the 1940\'s about blood. They knew it was bogus back then .         It\'s an overextension of the old testament kosher law,which DOESN\'T apply to New testament Christians..           They thought that the end of the world was a coming back then,(1940ish).Sooo, there wouldn\'t be much of a body count causality.           Now with Blood on their hands,and the murder of thousands of innocent minor children,how can they account for the body count, if they repeal it outright now?          More importantly,  to them ,the WT$ would be sued,for mega-bucks,for all the wrongful death lawsuits.          So,as you can see before your very own eyes,they are eroding away gradually the dogma,eventually the day will come,when \" new light \" will \" flash up \" from Jehovah.and it will be dissolved.                   What about all the dead kids,who died for NOTHING?          By the way the chief symptom of my own 28 year horror show with Bleeding ulcerative colitis was, profuse bloody bowel movements on the order of 20-30 x\'s a day.It created enormous stress for me to know I could bleed to death at any moment.Yet,as a devout Jehovah\'s Witness be forbidden to have a blood transfusion!                        I am bearing witness to the   Watchtower holocaust  .          Watchtower Whistle Blower  www.DannyHaszard.com Bangor Maine USA       Â» Theologically,Jehovah\'s Witnesses  are a cult of Christianity.The oppressive organization does not represent historical, Biblical Christianity in any way. Sociologically, it is a destructive cult whose false teachings frequently result in spiritual and psychological abuse, as well as needless deaths.  '),(1951,4128156,'04,August,2004','2004-08-04 00:00:00','   The Watchtower leadership and a number of JW elder perps,are well aware of what I want...   Undaunted Danny wants  EVERYTHING  promised to me by the WT$.    Actually,I would be willing to settle out of court for all my money back,and they know it well.   Repay my money that you extorted from me and I Will go away   Meanwhile,I am giving them a run for their money........           MY Mission Profile and Biography:  Cult Buster 47 y.o. Ex JW X 33 years 3rd generation.Wrongfully disfellowshipped [1-05-92] on perjured testimony from  flunky  Elders:    (1) \" Mealymouth\"  George Marsden,       (2) \" Flunky \"Mark Henderson,       (3) \" A.O.P.\" Richard Hall ,    (4)  \" Punkhole \" David Mcgrath,      (5)  \" Cesspool mouth\"  Robert B. North    (6)  and \" Pedophile protector\"  William T.Darche,     All perps,from the Kingdom Hall of Jehovah\'s Witnesses Summer street Rockland Massachusetts USA 02370. TELEPHONE # (781) 878 - 6773.      My mission profile is to get the above family wrecking,pedophile protecting,and blasphemous bastard  Elders all Disfellowshipped themselves .       I also demand a compensatory damage payment of $100,000,000 from the WT$ real estate empire and ripoff racket.       Watchtower,do the right thing!        Hold these perpetrators accountable immediately!     My motto/credo :\"well living is the best revenge\" and \"kill the enemy besides\"           [ Footnote:  MADD\'s motto;[Mothers Against Drunk Driving] \"..Say nothing,....do nothing,...  STOP   nothing..\"     God hates a coward NO REST FOR THE WICKED  '),(1952,4128156,'03,August,2004','2004-08-03 00:00:00','   Before you debate me read this:      http://www.freeminds.org/psych/beel.htm       A Study of the Persuasion Techniques Used by Jehovahâ€™s Witnesses and the Watchtower  A research report submitted to Tabor College, Adelaide, by NATHAN CHARLES BEEL as the Directed Study Project component for the degree of Bachelor of Arts in Christian Counselling--- November, 1997.   -Introduction  -Liftonâ€™s Criteria for Thought Reform  -Milieu Control   -Mystical Manipulation  -Loading the Language  -Doctrine over Person  -The Sacred Science  -The Cult of Confession  -The Demand for Purity  -The Dispensing of Existence  -Hassanâ€™s Criteria for Mind Control  Behaviour Control  -Information Control  -Thought Control  -Emotional Control  -Conclusion  -Propaganda  -Cognitive Dissonance    http://www.nizkor.org/features/fallacies/  Examples of Fallacies : AKA:\"DON\'T PISS DOWN MY BACK AND TELL ME IT\'S RAINING \"  Boloney Detection Kit      Inductive Argument    Premise 1: Most American cats are domestic house cats.  Premise 2: Bill is an American cat.  Conclusion: Bill is domestic house cat.    Factual Error    Columbus( discovered America) is the capital of the United States.    Deductive Fallacy    Premise 1: If Portland is the capital of Maine, then it is in Maine.  Premise 2: Portland is in Maine.  Conclusion: Portland is the capital of Maine.  (Portland is in Maine, but Augusta is the capital. Portland is the largest city in Maine, though.)    Inductive Fallacy    Premise 1: Having just arrived in Ohio, I saw a white squirrel.  Conclusion: All Ohio Squirrels are white.  (While there are many, many squirrels in Ohio, the white ones are very rare).   Index    Ad Hominem  Ad Hominem Tu Quoque  Appeal to Authority  Appeal to Belief  Appeal to Common Practice  Appeal to Consequences of a Belief  Appeal to Emotion  Appeal to Fear  Appeal to Flattery  Appeal to Novelty  Appeal to Pity  Appeal to Popularity  Appeal to Ridicule  Appeal to Spite  Appeal to Tradition  Bandwagon  Begging the Question  Biased Sample  Burden of Proof  Circumstantial Ad Hominem  Composition  Confusing Cause and Effect  Division  False Dilemma  Gambler\'s Fallacy  Genetic Fallacy  Guilt By Association  Hasty Generalization  Ignoring A Common Cause  Middle Ground  Misleading Vividness  Personal Attack  Poisoning the Well  Post Hoc  Questionable Cause  Red Herring  Relativist Fallacy  Slippery Slope  Special Pleading  Spotlight  Straw Man  Two Wrongs Make A Right      \"SELF-TERMINATING CLICHES ------- all apostates are \'full of hate\' and \'hate the truth\'.     OCCAM\'S RAZOR aka the K.I.S.S. rule \"Keep it simple silly\"              Cognitive dissonance  : When Prophecies Fail A Sociological Perspective on Failed Expectation in the Watchtower Society by Randall Watters:   http://www.freeminds.org/psych/propfail.htm        DANNY CALLS THIS CULT PHENOMENON (cognitive dissonance); \"JUSTIFICATION OF SOUR GRAPES\". e.g.The car of my dreams was sold to another party just as i got there,ahhh,it was probably a lemon anyway...or,armageddon didn\'t come in 1975,just think,\"now apostates will have time to make it back\".       Last but not least dispargement/mudslinging e.g.:\"apostate opposers R full of bleep cause they like to,\"\"Throw around   LARGE   words\" to show that they R know-it-alls\"   ....... or \"Your post is silly. JWs aren\'t like you depict them. You seem very hateful and bitter. Seek help.       Watchtower\'s canon of ethics [Rules] Also known as; \'approved operating procedure\' [ A.O.P.]Jehovah\'speak = doublespeak. Carefully crafted cult rhetoric. JWessse             It\'s plain old: mealy mouthing, innuendo / deception / flim-flam / hoodwink / bamboozle / fraud / mind games / confidence game / blowing smoke and bluster......  It\'s the,  deflection \"JEOPARDY GAME\" Framing false \"profitcy$\" in the form of a QUESTION ??  '),(1953,4128156,'03,August,2004','2004-08-03 00:00:00','   Blowing the Whistle on the Watchtower re: sales tax fraud             The following letter was mailed out in quantity to the 50 states (Sales Tax Fraud Divisions) in early 1997.         The collecting of sales tax by peddlers from customers is required in 49 of the 50 states [except,N.H.]       For a full report on the fraudulent nature of the Watchtower in this regard, read:  The Watchtower Way of Laundering Money @ my home page.                Earlier the Watchtower organization suffered a defeat along with Jimmy Swaggart. The previously-mentioned article says,  \"Just one month before the new policy was announced, the Society watched how the Supreme Court ruled in California Board of Equalization vs. Jimmy Swaggart Ministries.       California wanted to assess sales taxes on the sale of books and tapes and other items by Swaggart\'s ministry.  The Watch Tower Society filed amicus curiae  (a friend of the court legal brief)  in support of Swaggart\'s position,  that a religious organization should be exempt from such taxation.         On Jan. 17, 1990, just a little over a month before the Society\'s change in policy, the Supreme Court ruled against Swaggart and permitted taxation. The Society\'s new policy avoided any liability for taxation by taking the financial transaction out of the picture (donations cannot be taxed).\"       While the Watchtower\'s \"new policy\" shielded them from a lawsuit by the State of California (and subsequent others) , they continued to peddle the magazines and books from door-to-door (and to this day), mentioning contributions and although not saying a specific amount to the householder to give, the Witness peddler himself has to come up with the \"value amount\" of the literature and give it back to the Kingdom Hall to send to Brooklyn. If not enough funds come in according to their evaluation of the \"value\" of the literature, the Watchtower\'s literature supply desk will halt the flow of all literature to the local Kingdom Hall until the \"bill\" has been paid.                 This is still currently the hypocritical practice as of March 2003 .  I also included in the mailings a copy of David Reed\'s book Blood On the Altar. Some states wrote back; their reply was basically \"Thank you,\" or that they will process the request of the Watchtower\'s abuse.  Here is a facsimile of the letter I write in 1997:      Danny Haszard P. O. Box 15449 Boston, MA 02215     ATTN:  Certified Mail to Criminal Investigation Tax Fraud Division To all 50 states and all I.R.S. Offices  January 15, 1997  Dear Tax Office,       I am writing to   blow the whistle   on an organization I used to belong to, an organization that is selling millions of dollars of merchandise in this state without collecting sales tax. The Watchtower Bible and Tract Society of New York ships truckloads of books, videos, tapes and assorted other merchandise to their Kingdom Halls across the state. The members take these items from there and sell them house-to-house, returning the money collected to the Kingdom Hall, so that it can be mailed back to Brooklyn headquarters each month. They talk about \"donations\" but this is a thinly veiled tax-dodging sales operation. Dunn & Bradstreet reports for recent years show their sales in the $1 billion range, so the Jehovah\'s Witness are certainly selling $--millions in this state.  The religious publishing operations and bookstores connected with other churches are collecting and paying sales tax.  As a taxpayer, I am writing to blow the whistle and to ask what is going to be done about the Watchtower\'s tax evasion.   Yours truly,      Danny Haszard Bangor Maine USA ( presently )          Â» Theologically,Jehovah\'s Witnesses  are a cult of Christianity.The oppressive organization does not represent historical, Biblical Christianity in any way. Sociologically, it is a destructive cult whose false teachings frequently result in spiritual and psychological abuse, as well as needless deaths.  '),(1954,4128156,'03,August,2004','2004-08-03 00:00:00','   LOVE BOMBING      HOW DO CULTS SEDUCE THEIR PREY ?    \"\" ......A tactic used most often is the \"LOVE BOMBING\"...They swarm over you in a sort of ,\" Cult   COCOONMENT  \".   All of a sudden,,you have,,INSTANT FAMILY,,INSTANT PURPOSE,,INSTANT COMMUNITY,,INSTANT FRIENDS..... And you don\'t have to look inside yourself  for answers anymore,,.. because cult leaders, or their designated high disciples,,dispense all your spiritual needs.   A mass delusion,,mass hysteria,,operating under the illusion of a master plan.      In military jargon this is known as ,\"Cluster F**k\".     Beware of Love Bombing:      Leonard Brenner Love-bombing is characteristic of most cults. Prospects, recruits and members are drowned in a sea of love and caring. Recently in an evangelical church I heard the pastor describe his visit to two cultic groups in which he praised their love-bombing and urged that his church adopt the same loving attitude towards visitors and members.         Should the evangelical church practice love-bombing? Is this what Christ meant when he said, \"By this will all men know that you are My disciples, if you have love for one another,\" (John 13:35)?       I\'ve heard cult members say, \"Of course we practice love-bombing: Who\'d want to be in a group or church that practiced hate-bombing?\"       This attitude highlights a common misconception.  >>>Hate-bombing is not the opposite of love-bombing.       The opposite of love-bombing is unconditional love. Love-bombing is highly conditional. The cults will love you to death while you represent a prospective convert to their group.          As a member a tight family love will surround you as you faithfully promote their cause. However, when it is clear that a prospect will not join the group or a member voices doubts, create waves, or leaves the group, all love ceases.        Indeed scorn is immediately heaped on these individuals and remaining members are told not to have any contact with them. All time, effort and love-bombing is then directed towards new prospects and the faithful members. Is this the love evangelical churches should practice?  Unconditional love  is what God practiced when he sent his Son to die for us?  '),(1955,4128156,'03,August,2004','2004-08-03 00:00:00','   OF INTEREST TO: LAW ENFORCEMENT:    DIFFERENTIAL COMPARISON,OF APOCALYPTIC CULTS vs. COMMON CRIMINALS:       Credo of common criminals;..\" there is no honor among thieves\",\"every man for himself\"   Credo of apocalyptic cults ..\"death before dishonor\" , \"all for one, and one for all \".    A  BIG  DIFFERENCE!   Common street thugs will betray one another at the drop of a hat..This is pretty much a foregone conclusion.   The,imminent \'end of the world\' , apocalyptic concept creed , is much more formidable.     This was the lesson learned with the David Koresh disciples at the Waco Texas disaster..These fanatics are pre-programed for martyrdom..The more force that you apply against them, the more it feeds into their self destructive mindset..   This phenomenon has been called,\"  suicide by cop\".     They think if you kill them, they are going to a [BETTER} place !   This delusion is the underlying doctrine of global terrorism..   Question ? ? The 9/11 suicide perpetrators,, where they sincere ????   YOU BET !!   We have a saying about them,..\"they are sincere , but they are sincerely WRONG\".      They are not in an eternal paradise of bliss,being waited on by a dozen virgins....They are just dead..     What about the convictions of the Jehovah\'s Witness cult , and their apocalyptic infatuation ?   My personal experience is some are sincerely devout regarding their convictions of faith.     I know I was.       And, many are driven by the age old credo,of the  trifecta of 3 p\'s  :Power , Prestige , and Pu$$y....same as the street thugs.More carnal than spiritual.   Is their an effective tactic against entrenched fanaticism ?   Simple....       The tried and tested divide and conquer technique.No matter how much cult leaders would have it otherwise, cult disciples can be split and fractioned.It\'s possible to dissolve their suicidal resolve.      Appeal to the failed fallibility of their headstrong leaders,and their failed promises,e.g.   \"...the emperor has no clothes...\"    Appeal to reason and ration.   This may draw some moderate members out , especially if they have small children..   They,don\'t all want to die.They, are not all ready and eager to be  glorified.    But,if they are belligerent , [about going to paradise] ..You pretty much have no recourse except to ,  \"punch their ticket for them \".       Undaunted Danny  http://dannyhaszard.com/ Bangor Maine USA         Only Prisons have  \'Watchtowers \'  '),(1956,4128156,'03,August,2004','2004-08-03 00:00:00','   What a waste! 33 years of my life in a destructive religious cult     My catchphrase:\"the anguish and the agony of having to renounce the life-long convictions of my heart\".      It ain\'t hyperbole it is a surreal horror show.        So what am i now?I like to consider myself an Honorable Humanitarian.Just be a good person for now.This is kinda like, a throw back to basic Christian altruism.       KNOW THIS!  If you go into the woods a mile,it will take the same to get out....       For us especially for it is a   crusade not a cakewalk.            Undaunted Danny; Watchtower Whistle Blower www.DannyHaszard.com       Bangor Maine USA former member of the abusive Kingdom Hall of Jehovah\'s Witnesses Rockland Massachusetts USA           Â» Theologically,Jehovah\'s Witnesses  are a cult of Christianity.The oppressive organization does not represent historical, Biblical Christianity in any way. Sociologically, it is a destructive cult whose false teachings frequently result in spiritual and psychological abuse, as well as needless deaths.  '),(1957,4128156,'03,August,2004','2004-08-03 00:00:00','   Dedicated to my family members, nieces and nephew Lauren, Ashley and Kevin Haszard that have gone missing in the Jehovah\'s Witnesses.     ? CLOSURE ?    PAN AM FLIGHT 103  DOWNED BY TERRORIST in 1988  Why they can\'t,\"..JUST FORGET ABOUT IT\"  SOUL RAPE OF MY FAMILY:  Why I can\'t, \"..JUST FORGET ABOUT IT\"  \"CLOSURE\", is a phrase for people who probably have never been through a horrific tragedy..When you have losses that are permanent..It\'s forever!  WE ARE THE INCONSOLABLE !  It is not being ,\'retro\' for us to demand justice, (Proverbs 6:33-35 ) There can be NO appeasement,for vicious atrocities .  259 souls of loved ones perished in the murderous actions of global terrorist that downed that doomed airliner..Family members who lost even a single loved one, did not rest until a measure of justice and retribution was achieved..  HELLO !..I LOST MY ENTIRE FAMILY !!  They are being  held hostage  (cocooned) by the Jehovah\'s Witness cult..  There is a time honored taboo..In a fair fight,you don\'t assail and assault your opponent\'s kin..Family\'s are off limits..You don\'t,\'hit below the belt\'.  Like the,next of kin, of the victims of terrorist..I am positioned front and center , to seek retribution and justice..   NO REST FOR THE WICKED  '),(1958,4128156,'03,August,2004','2004-08-03 00:00:00','   And another.... and another.... and another....      \"Throughout history countless thousands of people have predicted the end of the world, and they all have one thing in common... they were all wrong\"                  http://www.sullivan-county.com/nf0/y2k/bible_ca.htm     Date setters, making irresponsible predictions!  Misleading the ignorant masses for higher TV ratings and book sales!    1843-1844   GROUND ZERO: William Miller was the founder of an end-times movement that was so prominent it received its own name-- Millerism. From his studies of the Bible, Miller determined that the second coming would happen sometime between 1843-1844. A spectacular meteor shower in 1833 gave the movement a good push forward. The build up of anticipation continued until March 21, 1844, when Miller\'s one year time table ran out. Some followers set another date of Oct 22, 1844. This too failed, collapsing the movement. One follower described the days after the failed predictions, \"The world made merry over the old Prophet\'s predicament. The taunts and jeers of the \'scoffers\' were well-nigh unbearable.\"           1874  invisible return of Christ; Original position of the Watchtower. Then Charles Taze Russell predicted the Rapture in 1910, followed by End of the World in 1914--later interpreted as invisible return of Christ. (99 Reasons Why No One Knows When Christ Will Return, by B J Oropeza, Foreward by Hank Hanegraaff, IVP publishing, 1994)       1914   was one of the more important estimates of the start of the war of Armageddon by the Jehovah\'s Witnesses (Watchtower Bible and Tract Society). They computed 1914 from prophecy in the book of Daniel, Chapter 4. The writings referred to \"seven times\". The WTS interpreted each \"time\" as equal to 360 days, giving a total of 2520 days. This was further interpreted as representing 2520 years, measured from the starting date of 607 BCE. This gave 1914 as the target date.    1914  Charles Russell, after being exposed to the teachings of William Miller, founded his own organization - the Jehovah\'s Witnesses. In 1914 Russell predicted the return of Jesus Christ.    1914, 1918, 1920, 1925, 1941, 1975 and 1994   Jwâ€™s: Here are other dates that the Watchtower Society predicted. 1975 looked likely as it was computed as the 6000th anniversary of the creation of Adam in the Garden of Eden in 4026 BCE. They interpreted Psalms 90:10 as defining the length of a generation to be 80 years. Since 1914 plus 80 equals 1994, they predicted Armageddon would occur around that year. The latest estimate was 6000 years after the creation of Eve, for which no date can be determined with any accuracy.    1914  End of World; Charles T. Russel, Jehovah\'s Witnesses; later explained that Michael (ie Jesus) had defeated Satan in heaven    1914  ; Jesus invisible and quiet return to the Earth. (Jehovah\'s Witnesses, Book: This Means Everlasting Life, page 221) \"So A.D. 1914 marks the time of Christ\'s invisible return in spirit.\" Explained after they said he would return in 1914          1918  In 1918, new math didn\'t help the Jehovah\'s Witnesses from striking out again.           1925  The Witnesses had no better luck in 1925. They already possessed the title of most wrong predictions. They would expand upon it in the years to come.    1953-AUG:   David Davidson wrote a book titled \"The Great Pyramid, Its Divine Message\". In it, he predicted that the world would end in 1953-AUG.    1957-APR:  The Watchtower magazine, quoted (1) a pastor from California, Mihran Ask, as saying in 1957-JAN that \"Sometime between April 16 and 23, 1957, Armageddon will sweep the world! Millions of persons will perish in its flames and the land will be scorched.\'       1975  The Jehovah\'s Witnesses were back at it in 1975. The failure of the forecast did not affect the growth of the movement. The Watchtower magazine, a major Witness periodical, has over 13 million subscribers.    1975  Armageddon, 1975; Jehovah\'s Witnesses; (Various publication, cited in Newsweek Oct. 15, 1984)     1984  The Jehovah\'s Witnesses made sure in 1984 that no one else would be able to top their record of most wrong doomsday predictions. The Witnesses record is currently holding at nine. The years are: 1874, 1878, 1881, 1910, 1914, 1918, 1925, 1975, and 1984. Lately, the JW\'s are claiming they\'re out of the prediction business, but it\'s hard to teach an old dog new tricks. They\'ll be back.         In a moment of clarity, I peeked out of the WT$ cult cocoon and come to my senses.I came to the realization that they are just like the Moonies and all the other cults.     Eight Marks of a deadly Mind Control Cult applied to Jehovah\'s Witnesses  [ click my homepage: www.DannyHaszard.com  ]      The end of the world has been postponed so that doesn\'t mean it\'s not coming!      Rather it,means we have to go out door to door all the more now.       It\'s called:      cognitive dissonance  : When Prophecies Fail A Sociological Perspective on Failed Expectation in the Watchtower Society by Randall Watters:    http://www.freeminds.org/psych/propfail.htm [ also at my home page; www.DannyHaszard.com  ]                   Watchtower Apocalyptic Infatuation  '),(1959,4128156,'06,August,2004','2004-08-06 00:00:00','    What a waste! 33 years of my life in a destructive religious cult   My catchphrase:\" the anguish and the agony of having to renounce the life-long convictions of my heart\". It ain\'t hyperbole it is a surreal horror show.  So what am i now?I like to consider myself an Honorable Humanitarian.Just be a good person for now.This is kinda like, a throw back to basic Christian altruism.  KNOW THIS! If you go into the woods a mile,it will take the same to get out....  For us especially for it is a  crusade not a cakewalk.  Undaunted Danny; Watchtower Whistle Blower www.DannyHaszard.com   Bangor Maine USA former member of the abusive Kingdom Hall of Jehovah\'s Witnesses Rockland Massachusetts USA   Â»  Theologically,Jehovah\'s Witnesses are a cult of Christianity.The oppressive organization does not represent historical, Biblical Christianity in any way. Sociologically, it is a destructive cult whose false teachings frequently result in spiritual and psychological abuse, as well as needless deaths.   '),(1960,3418455,'30,May,2004','2004-05-30 00:00:00','   Well my sisters birthday party was pretty cool.  My niece was VERY uncertain of things and wouldn\'t let go of me until her mom (my sister) came.  After all the sneaking around Juli was very surprised.  I had planned on just hanging out for an hour or so, but my brother in-law\'s brother and girlfriend are just like me so we hung out all night.  We made fun of people drank and all of a sudden 3 hours had passed.  Well @ 10 they left so I got in on a conversation with my sis and her husband and some of her close friends I\'ve known since she\'s known them.  I got home @ 2 in one piece so all-in-all it was a good night.  Now Juli says she\'s gotta throw Curtis a party or his feelings will get hurt.  They\'re stupid!  Well until the next time.  ~opphold negativ  '),(1961,3418455,'29,May,2004','2004-05-29 00:00:00','   Well it\'s that time of year again...my sisters birthday.  Last year, I remeber very vividly, I was in the emergency room for food poisoning.  Very dreadful experience.  I have yet to have crawfish since then, I feel the gurgling of stomach juices just thinking about it.  What does this have to to with Juli\'s b-day?  Well I missed going out with everyone b/c I had been @ the hostpital and whatnot.  Only good thing that came out of it was that I lost 20 lbs; but it came back...who cares.    Well now Curtis has decided to throw Juli a surprise party (I think she\'s going to be aggrevated) and I\'m going to have to go hang out with people I don\'t know for who knows how long.  FUN FUN FUN.  ~stay negativ  '),(1962,3418455,'26,May,2004','2004-05-26 00:00:00','   I really am at my wits end with trying to get 3d to work under gnu/linux.  It has sucked up all my time today.  I guess it\'s good I have no responsibilities.  Until the next time I want to talk to myself...  '),(1963,3418455,'25,May,2004','2004-05-25 00:00:00','   I hate it.  I also hate going to sleep knowing the next thing I do will be to wake up.  Dirty vicious cycle!  '),(1964,3418455,'25,May,2004','2004-05-25 00:00:00','   There\'s nothing worse than waiting for UPS to deliver something.  '),(1965,3418455,'22,May,2004','2004-05-22 00:00:00','   Yes, that\'s me.  I was who knows how old...maybe 4 or 5, yet the anger is so evident even at an innocent age as that.  I guess I was born mad at the world, or maybe just jaded by this whole live your life to the fullest but don\'t be bad or you\'ll go to hell.  Well I say Uff Da!  remember ~ stay negativ  '),(1966,3418455,'22,May,2004','2004-05-22 00:00:00','   I have WAY too many things I want yet no possible way to ever get them.  I am getting some of the \"bestest\" education possible but somehow doubt that my future career in digital forensics/security will be able to pay the bills.  The one good thing I do have going for me, other than the monthly supply of Klonnopin, is that I have lost interest in relationships.  My reasons: they are too expensive, at least they have been for me, and I normally come out of a relationship a little closer to psychosis.  As always: all your oil are belong to U.S. / stay negativ  '),(1967,3418455,'22,May,2004','2004-05-22 00:00:00','   urlLink Z/n Blog  Well, my first post.  Sitting here in bed @ 2am doing what I am always doing...watching the X-files. I\'m about to have some cocacola.  Just had some, very refreshing at this time of night.  '),(1968,3418455,'27,June,2004','2004-06-27 00:00:00','   I FINALLY saw all of \"To Catch A Thief\" with Cary Grant and the so beautiful, so classy Grace Kelly.  I have been in awe of her pure beauty since my mother and I watched \"Rear Window\" as a child.  Great movie...I really hope there are people out there my age who appreciate these great movies.  '),(1969,3418455,'25,June,2004','2004-06-25 00:00:00','   Friday in class I was thinking about the tattoo that I have wanted since high school.  Well I finally scratched it into my skin (not psycho-ly) and it looked awesome!  Though it was red and not green it still looked awesome.  I think I need to get it.  '),(1970,3418455,'25,June,2004','2004-06-25 00:00:00','   Well that test I was \"worried\" about has past and I ended up getting a low b/ high c.  The jigha next to me got 1 point higher than an F...po\' fool.  I really do like my teacher.  He\'s crazy smart, a lot about the historical foundations of mathematics, which I like.  Well, obviousl I am still not dead which Andree will be happy about, thanks for caring!  Ohwella...off to go play with the laptop (putting SuSE on it...maybe).  ~stay negativ  '),(1971,3418455,'25,June,2004','2004-06-25 00:00:00','   I am not dead yet  '),(1972,3418455,'20,June,2004','2004-06-20 00:00:00','   Well, I\'m 26 now.  It will be interesting to see whether my \"preminission\" I had years ago comes true.  I\'m talking about me dying when I am 26 years old.  Who knows...  '),(1973,3418455,'18,June,2004','2004-06-18 00:00:00','   Well I\'m about to take a test, which I think I may do \"o.k.\" on, but that\'s 1/2 of what\'s on my mind.  I have heard about the search for the largest Mersenne Prime quite a few times and yesterday in class we discussed it.  Well theres a program you are able to run on your own machines, very similar to the SETI project, but there\'s a catch!   If you are the person who finds the first Mersenne Prime with 10 million digits you win $100,000.  Well knowing that i\'ve already started!  i can\'t imagine what the current number is up to since p = 7.1 megs in text format!  p, as in 2^p-1.  It\'s a long friggin # but hopefully i\'ll get it and run off to Sweden!  '),(1974,3418455,'13,June,2004','2004-06-13 00:00:00','   Well it\'s Sunday it rained a bit and was overcast so it was okay.  I\'m bumbed about the upcoming week since I have a test Friday in the strangest class I believe I have ever taken.  The teacher is quite the Nazi when it comes to the subject.  I am learning it, I think, but who knows how I\'ll do on the exam...  STAY NEGATIV  '),(1975,3418455,'08,June,2004','2004-06-08 00:00:00','   The 3rd day of class has come and gone, but I didn\'t.  Alas here\'s the 4th day and I\'m wearing f\'ing bells.  I\'m kinda uff-da\'d b/c a guy is downloading something from me @ 8k/s  ...and the file size is just over 4 gigs.  You do the math, I don\'t wanna.  That means I have to live with this highspeed fan wizzing in my dreams....ohwella it\'s for a good cause; The propogation of SuSE!  '),(1976,3418455,'03,June,2004','2004-06-03 00:00:00','   Well another summer; some more school.  I am taking only 1 class since I am poor and could only afford 3 hours plus they weren\'t offering much else that I needed.  The teacher seems to be a good guy.  I hope the class is like his last name, Holladay.  The most gruesome aspect is the price of the text...$100, not like I am surprised.  But the real bad thing is that it\'s a new book so I am forced to buy new!  I hate that!  Ohwell, just ate dinner, going to go take some Klonnopin and go to bed while watching Scream2.    Preview for tomorrow:  New Computer Arrives!!!!  '),(1977,3418455,'01,June,2004','2004-06-01 00:00:00','   urlLink    Well last night brought back some very painful memories for me. My MiMi is in the hostpital and just seeing her lying in the bed brought back all the memories from 3 years ago of my PaPa. It was like I saw him there for a moment. Whenever MiMi coughed and winced at the pain in her chest I saw my PaPa react the same way to coughing.  Three years ago was pretty much the time I lost all faith in God, and love for him as well.  There was no need to let my poor PaPa suffer for as long and as hard as he did...yet God allowed it.  I have not gotten over that and don\'t think I will anytime soon.  PaPa I miss you so much and I hope MiMi comes to you with as little pain as possible.  '),(1978,3418455,'01,June,2004','2004-06-01 00:00:00','   urlLink    Definitely my girl! I am slowly rubbing off on my niece, not sure that\'s a good thing but at least she\'s the cutest bumb!  Gabby has spent the past 3 days over here, just during the day, and it\'s been a blast!  The summer will definitely be one to remember.  '),(1979,3418455,'27,July,2004','2004-07-27 00:00:00','   On the phone with Rob, he\'s getting a #4.  talk to myself later.    p.s.  he said some chick @ Wendy\'s was hot  '),(1980,3418455,'24,July,2004','2004-07-24 00:00:00','   I can\'t believe it, but I did; I forgot to mention Teresa -*-Heinz-Kerry.&nbsp; That bitch has as many names as her husband has sides to an issue.&nbsp; Sorry about forgetting your hypocritical ass, but at least I remebered.&nbsp; Funk you!  ~bli negativ  '),(1981,3418455,'22,July,2004','2004-07-22 00:00:00','   I hate Spam, John Kerry, Susan Sarandon, Madonna, Michael Moore, Linda Rondstadt, Johnny Depp, Sean Penn, The Clintons....ARG!!!!!!!!!!!!!!  Send them to \"Internment Camps\" or something, PLEASE!  '),(1982,3418455,'22,July,2004','2004-07-22 00:00:00','   \"The 5th Element\" is on!!!  For four hours straight....I love TNT!!!!!!!!!!!!!  '),(1983,3418455,'19,July,2004','2004-07-19 00:00:00','   Wow!  \"The Grid\", a new TNT drama series was great.  It finally is going all out on the damn muslims.  I could type forever on my feelings on religion, epecially muslim.  The simple fact that they degrade and limit the freedom of 1/2 of their population (women, that is) is outrageous!  These poor women need to be free to choose what they do...not some crazy funadamentalist clerics.  F THEM!    ~bli negativ  '),(1984,3418455,'18,July,2004','2004-07-18 00:00:00','   I want to know where the hell my boxer shorts are going!  Are the damn underpants gnomes visiting my house???  That\'s about it.    ~bli negativ  '),(1985,3418455,'15,July,2004','2004-07-15 00:00:00','   Yeah well I get a test back tommorrow morning, hoping for the best.  I\'m also hoping Rob gets his puter shit in too so we can build a couple of machines over the weekend.  Well I need to get more hard drive space.  I\'m constantly deleting data to make room for newer data but that is getting old.  Well blah blah blah...I\'m going to go watch Howard Stern so bye.  ~bli negativ  '),(1986,3418455,'12,July,2004','2004-07-12 00:00:00','   Well everything can\'t last forever... After a day of having a 2004 Dodge Ram 1500 SLT I had to bring it back :(  It was the best friggin thing I have driven!  It was so high...I sat higher than when driving an Explorer or Expedition.  It was a nice treat...thanks Shell!  '),(1987,3418455,'11,July,2004','2004-07-11 00:00:00','   I am having a lot of luck getting all this great music lately.  I have gotten multiple gigs of music, which isn\'t strange, the strange thing is that it\'s all REALLY good.  All I ask is that this keeps up...and the guy whose posting the \"My Dying Bride\" flood keeps it up.  Update...Still not dead and Lacuna Coil is still my favorite.  '),(1988,3418455,'08,July,2004','2004-07-08 00:00:00','   Possibly this may be the illness which kills me @ the age of 26!  ...See Zorg finds the good in the bad, negativ finds everything bad b/c it just so happens to be that everything IS bad.  Just to remind myself, Lacuna Coil is my new favorite band of the moment.  '),(1989,3418455,'06,July,2004','2004-07-06 00:00:00','   I hate being sick when there is stuff to be done.  I feel like I have the flu, but it\'s friggin July!  What tha dilly?  What\'s worse is that I have class in 30 minutes and am going to get an assignment that is due tommorrow.  Doesn\'t he know CSI is on!!!!  I can\'t do shit and I don\'t feel like it either.  F it!  ~stay negativ  '),(1990,3418455,'06,July,2004','2004-07-06 00:00:00','   Right now I am watching \"The Saint\" with Val Kilmer and the beautiful Elisabeth Shue.  Well one of my old roommates sister looks just like her, or close enough for me.  Well we were always dating someone at different times while the other was single.  Mostly she was constantly getting engaged to different guys and I was with this beast.  Well one night when she broke up with her b/f she came over and chilled with one of my other roommates.  Well they ended up getting together a couple of times...all the while I was right on the other side of the wall without any beast!!!      Well we talked a few weeks later about it all, she had already gotten engaged to someone else, and she said she thought I was still with the Amazon and that\'s why she didn\'t call me!  For years we were always \"talking\" about it but there was the chance and poof!      I regret not cheating on the Amazon when I could have b/c she did it to me, but unfortunately I am not like that.  Now I don\'t believe in that crap anymore....it\'s too much trouble.        So the moral is \"When temptation tempts....give in for you\'ll probably never get the chance to do it again, plus there will be time to \'repent\'\".  '),(1991,3418455,'03,July,2004','2004-07-03 00:00:00','   A long time ago I was dating this Amazonian \"girl\" who turned out to be psycho.  Well for her birthday I gave her a laptop, kinda thinkig we were still dating.  Well unknown to me, basically because she never told me, she \"stopped\" dating me and was doing her thing all over the town while still sleeping over at my place, that is when she wasn\'t over \"somewhere else\".  Anyway, time went by and I stopped seeing her all together, THANK GOD!  Well she still had the laptop and my roommate had wanted to borrow it and told him that the psycho still had it.  He got mad and went and got it!!!  He\'s the shit!  Well just b/c of him I had that computer to be able to let my neice play with and now it\'s over at sistas house getting used...Thanks Charlie!  '),(1992,3418455,'02,July,2004','2004-07-02 00:00:00','   I HATE my sisters computer.  It\'s my old one from, I believe, 1997-1999.  It\'s a 400 MHz AMD, Slow as shit and useless!  I can\'t even put a *nix system on it b/c it can\'t recognize a hard drive larger than 40 gigs!  USELESS.  So instead of constantly trying to fix it I let em have my old laptop...works pretty well, at least for their needs.  Of course Curtis wants a 3GHz machine but doesn\'t wanna have to pay for it.  It\'s funny, my sis has always had a puter; always one of my old ones.  Well F that crap box....i think I am going to ask curtis to take me to a shooting range and we\'ll take his assault rifle to it.  That should be fun.  Until next time... ~stay negativ  '),(1993,3418455,'10,August,2004','2004-08-10 00:00:00','   urlLink    Hahahahaha.  I found this old pic of my sister and her boyfriend from high school.  He was the biggest dork and we all knew it except her.  That\'s why I hate taking pictures.  10 years from now I don\'t want to look at a picture of me and some beast...FUNK!  '),(1994,3418455,'07,August,2004','2004-08-07 00:00:00','   One of the best movies with one of the hottest women is on tonight!  Resident Evil!  I am looking forward to going to see the 2nd one next month.  Well, Rick James is dead...so I\'m feeling kinda \"funky\".  Talk to myself later.  ~bli negativ  '),(1995,3418455,'04,August,2004','2004-08-04 00:00:00','   I\'ve been looking over some of my posts, my god...do I have poor spelling or what!  \"No Soup For You!\"  ~bli negativ  '),(1996,3418455,'04,August,2004','2004-08-04 00:00:00','   If anyone out there is wondering where the terror starts it\'s in Switzerland.  The Swiss take great pride in their shady \"no questions asked\" mentality to global banking.  I think a contemporary and professional global bussiness should practice more legitimate methods in their trade.  All a terrorist would have to do is create a numbered, anonymous account filled with millions of dollars and their little friends take what is needed.  Thanks to friends like the Saudi royal family and every damn African country, these people have free run to create fear.  I wish the American people would unite, regardless of differences (not unlike post Pearl Harbor) and crush the enemy before they move again.  Time is on their side, not ours!  It is definitely not this administrations, nor the sex-filled admin priors fault for anything happening, they are bound by our own laws protecting the terrorists.  We should seriouslly re-think our form of law, at least for the time being to protect our future.  Yeah, rights will be lost, but they are only \"rights\" not \"lives\".  Which is more important to us?  I would gladly suffer any inconvieniences imposed just knowing that those people are feeling the same squeeze, making it harder to do their \"jihad\"    Fuck your jihad, I\'m Rick James bitch!  '),(1997,3418455,'03,August,2004','2004-08-03 00:00:00','   Damnit.  The best \"New\" show on during the summer, namely \"The Grid\", has only got 1 episode left!  That sux.  I really hope that TNT makes it a continous series.  '),(1998,4166693,'07,August,2004','2004-08-07 00:00:00','    Post number 30! Everyone dance, now.  But really, I just wanted to wish everyone a great weekend.  I leased an apartment today. My move-in date is set for Sept. 1st. It\'s not R.\'s parents\' place, but more on that later.  He was a wonderful, loving boy today and helped me beyond belief. These are the times that I realize how lucky I am. And they are the times that erase anything negative that has come before.  I\'ll update soon. I have to go to class (ugh!)  Love to all! '),(1999,4166693,'07,August,2004','2004-08-07 00:00:00','    It hasn\'t been updated in a while, I apologize. I\'ve been spending my time on my blog, really. A lot of my friends stopped using livejournal, so it\'s not as much fun anymore. However, if you want to read about my life prior to my appearance on blog the address is:   urlLink www.livejournal.com/users/jessic   It had its moments. '),(2000,4166693,'07,August,2004','2004-08-07 00:00:00','    I figure that after last night\'s post of irritation, I should probably follow up with some form of an update.  Things are better.  I should probably describe my anger management style. It\'s not as effective as I would like; however, I\'ve been doing it for so long that I don\'t know any other way to do it. I think it\'s something I have to work on. My \'style,\' as it may be, is very ordinary for a lot of people. I have a high tolerance for people and their bullshit. I can take a lot before I really snap. And usually it takes a while for it to build up to the point where I\'m really angry. Such was last night. Since R.\'s roomies have moved back in and there have been four people in the house instead of the two of us, I\'ve been slowly fighting the tide of irritation. Last night, after waiting around for an hour and a half for R. to show back up at the house when he was supposed to be home in 30 minutes, I finally snapped. I snapped because my keys were in his pocket, and I was stranded at the house. I snapped because I get tired of babysitting 22 year olds. I snapped because I like to keep things a certain way--clean--and not everyone can do the same. I snapped because after living with my sister for 14 months I expected that briefly living with a bunch of people my age would be an improvement. Mostly, I snapped because when I asked for my keys I was given a smart ass answer, and, honestly, I wasn\'t in the mood for that crap. So, as you can see, my irritation with not only R., but everyone else in the situation, took a while to grow--about a two weeks. And finally, BOOM!  I did tell him what the problem was, although in a very not happy Jessica voice. It was more like a mid-level yell. I have a hot temper once it gets going, but it only lasts about 10 minutes before puckering out on me. Long fuse, short span of temper, cool down. It\'s a strange scenario.  Somehow, he always manages to say something that makes me forgive him even if I am still hurt by what he\'s been doing. Even if it takes about 45 minutes for him to say something like that because he\'s too busy being defensive and trying to deny things. He gets a lot like his father when someone brings up a problem in a way that doesn\'t fit the way he wants it done. He gets very stoney and very uppity. He gets a holier than thou attitude that makes me want to kick him in the shins. I point it out to him, and he gets mad about my pointing it out. I\'m not the only one that needs to work on things, that\'s for sure.  We\'ll make it through the bullshit, but I agree that it\'s a good idea to still have our own space. I can\'t live with him right now because living with him would entail living with him and his two roommates. I think what makes that impossible is this:  When it\'s just the two of us, I know that I am an equal part in the messiness of our house. I know what is mine or what we used those five spoons for, and it doesn\'t bother me to have a hand in helping to clean things up even though, dispropotionately, I clean much much move often. However, when I come home from work, walk into the kitchen and find someone else\'s dishes stuffed into every crevice of the sink and their papers all over the table and the refrigerator crammed almost to the point of bursting because it\'s too hard to organize the shelves so things atually  fit , I don\'t feel like I should  have  to deal with those things. However, if I don\'t deal with them (i.e. clean them up, straighten, do dishes) they will sit there and sit there and sit there, untouched and undone until we\'re out of paper plates and they have to break down and finally lift a finger to wash them. Or until the kitchen smells so much like rotten food that I have to drag my ass out to the trash can at 11 p.m. to stuff it full with the crap that they\'ve left in the house. That\'s not my job. It\'s the whole being  their  mom that really bothers me. It\'s not the helping R. to have a clean house, it\'s the creating a clean house for my new \"family\" that\'s more the issue. Does that make sense? I think, if it had just been the two of us, it would have been a different story.  Last night he said one of those things that made me forgive him in a heartbeat. Well, first, he did the dishes without me saying anything about it. That made me happy. But then, when I was lying on the bed with my head upside down getting that little kid buzz, he came into the room and gave me a funny look and then said the thing that made me forgive him. And now it\'s getting better. '),(2001,4166693,'07,August,2004','2004-08-07 00:00:00','    Today is my long, long, long, long, long, long, long, long day. I\'m in the office from 8-2:45 and then I\'m off to the library from 3-8 for class.  Soooo long.  The real excitement was that my class was posted as starting at two different times, both of them listed in the syllabus, and I had no clue when to be there. So, luckily, it\'s the later time and I don\'t have to be in class from 12 p.m. - 8 p.m. That would have sucked.  I have to make some phone calls to agencies today to find out what their problems are with the contracts that we submitted.  Otherwise, I am hoping for a liesurely day in the office. Friday\'s are typically pretty quiet. I hope it stays that way. '),(2002,4166693,'07,August,2004','2004-08-07 00:00:00','    I was going to post something long here, but I am suddenly doing the potty dance in my desk chair.  Damned diet juice.....  Stay tuned..... '),(2003,4166693,'07,August,2004','2004-08-07 00:00:00','    I am  trying  really hard to be okay with living with R. and his roomies. Sometimes, I\'m mostly just trying really hard to be okay with living with R. It\'s difficult to move in with someone, even temporarily, and realize that you aren\'t really sure anymore if you can see yourself being with them forever because they are doing/saying/acting in a way that really makes you pseudo-hate them from time to time.  Such is tonight. Tonight is a night when I really want to yell \"dammit, I want my own place now!\" And I really would mean it tonight. I really want my own place. I really want to be away from him right now. Thus, I am sitting in my office at work at 10:12 p.m. on a Thursday night updating my blog instead of spending time reading or relaxing or watching TV at home with R. Reason, you ask? Because I feel underappreciated. Because I\'m working my ass off at a new job (or at least working as much as I can with the new students not being here and the programs just getting started) and then I\'m coming home to a messy house, nothing taken out for dinner, and the presumption that I am going to do everything that needs to be done there because nobody else seems to want to do it. It\'s like having a family already, except they\'re all 20 + years old and they are know better than to leave their milk glasses half-full in the sink so that it curdles and makes me want to yack when I have to wash the dishes (no dishwasher) two days later because no one else has done it. I\'m not ready to be someone\'s mom, nor do I think that I should have to be my boyfriend\'s mother.  This is the first night this week that I\'ve been able to come home and relax. Or so I thought. Instead, we went out to dinner with his roomies, which was fine, instead of spending time alone. Then, I come home and he goes out with one roomie (the male roomie) to buy a TV quickly at Wal-Mart--quickly being a 1 1/2 hour trip to look at a TV he already knew he was going to buy. So what am I doing? Our laundry. Alone. When they got back, I said I wanted my car keys, which he had taken with him in the roomie\'s car, so that I could go do something. So what does he do? Stomps down the stairs and in his more arrogant voice tells me that here are my keys, and the place I want to go is open until 11 so I can just go now instead since it took them longer. Pardon? I\'m doing his fucking laundry and cleaning up after his mess and their messes, and I\'m supposed to be happy that I\'m stuck at home without my car? Where in the hell does it turn into a Jessica is a fucker night just because I say I want my keys back?  We had talked about going for ice cream 3 hours earlier. I wanted to go to Dairy Queen for a blizzard because it sounded sooooo good. Instead, they went without me because I am too damned pissed to really want to be near him. I\'m on the edge right now. I think, if I see him, I might come undone. And not in a good way. Probably in that going to kill him slowly by sawing through a minor artery with a rusty spoon that would disable him but not kill him kind of way.  And, on top of that, no one I call will answer their phone. I give up with the phone. Apparently, it is an ungrateful apperatus. '),(2004,4166693,'07,August,2004','2004-08-07 00:00:00','    Enough said. '),(2005,4166693,'07,August,2004','2004-08-07 00:00:00','    My computer is fixed. Now I have email at work again. Yay. '),(2006,4166693,'07,August,2004','2004-08-07 00:00:00','    Ohhhh.....my office supplies showed up today! Hooray, hooray! I was so excited. I didn\'t get the really awesome desk organizer because we\'re supposed to be getting those when we move to the new Student Union in October. Aside from that, I got everything I wanted. My tape dispenser came in in \"putty\" which is a blah off-white instead of black like I requested, but I think I might Sharpie it to make it pretty...or putt fabric on it....ohhhh....I might do that! I think I will.  Last day with the freshmen was today. No more \"orientation week\" activities. I\'m a little sad, but I\'m glad not to be on my feet in heels for 4 hours straight. That\'s always nice.  Tonight is laundry night! The only reason I am really excited is that I need the panties desperately. I\'m almost out! Ewwww......  Tonight is also easy dinner night. I am either making R. take me out for dinner or we are making sammiches.  Hmmm.....I\'m a little sammiched out. Let me explain....  The last three days I have had freshman orientation seminars. All three days they have provided lunch. All three days that \"lunch\" has been the same sub sammiches from Wal-Mart. Tasty the first day; not so tasty the second day; disappointing the third day. Maybe dinner out would be a good thing for me.  Classes start tomorrow. Hooooo-raaaay! '),(2007,4166693,'07,August,2004','2004-08-07 00:00:00','    Sometimes, it\'s just best not to title something. So, I am being a rebel.  I watched  The Alamo  last night. Generally, I wasn\'t too impressed since it was sort of slow and plodding, and a bit predictable as far as some of the interactions between the characters. They had the typical \"rally round the captain\" moment and the \"peace between the conflicting main characters\" moment. And that \"sentimental because you know they are all going to die\" moment. However, the fighting scenes disturbed me a lot.  They didn\'t disturb me like the scenes in  Saving Private Ryan . That movie disturbed me on a level I never knew existed. I cried the entire first 10 minutes and then proceded to be upset periodically throughout the rest of the film due to certain characters\' deaths. It disturbed me in a way where the war was a backdrop to the struggle of a man to survive. Which, I know, is probably wrong, but it was the way in which it was all presented that gave it that feel. It was a man triumphing over adversity sort of thing--surving all odds. That sort of disturbing.  The Alamo  disturbed me in that it brought to light how power hungry we are as humans. Never can we just settle for what we have. Instead, we must run after each other with guns drawn and flags waving trying to dominate someone we find to be inferior. We always think we   deserve   something better than good. Sometimes, we think we deserve something better than great just because. And while the vast majority of the movie was speculation since everyone in the real Alamo died, it was difficult to watch a handful of ragtag men be trampled by a large army simply because the rest of the people were too arrogant to believe that Americans could be defeated.  Perhaps it is most disturbing because I believe that arrogance exists now. I feel as if we, as a country, or at least as a country led by certain powers, believe we are both infallable and invincible. Even after being attacked and wounded on our own turf, we still have the unshakable belief that we are right in the face of all adversity and we are the ones that should be in control of everyone. I look back at our history, and I realize that we\'ve had the same roughian view for more than 200 years. We never learn from our mistakes. We perpetuate them. I\'m not a pacifist who believes in being walked all over, but I do believe in taking care of your own. Until we can provide job security, health care, and competitive education for all, we need to focus ourselves a little more on our own reflection in the mirror and a little less on the shadows we claim to see everywhere else.  I figure we\'ll probably end up killing ourselves with our own stupidity someday. I don\'t think humans are as smart as they think they are. In fact, judging by every other creature on the planet that lives and breathes peacefully and contentedly with what they have, I think we have a lot more to learn. '),(2008,4166693,'07,August,2004','2004-08-07 00:00:00','    There\'s a little Yoda figure on my desk. He reminds me that \"the force is with you,\" and that I am a wise and special leader.  Actually, he also makes me wish I was shorter and painted green, but we\'ll not talk about that.  :) '),(2009,4166693,'07,August,2004','2004-08-07 00:00:00','    Today, I realized the meaning of the term melt-down. My account at work has officially done the melt-down thing. Yesterday, at 2:28 p.m., my account at the university where I work suddenly disappeared. The reason? I am a student at another university in our system, and, with the new software they are using, you are only allowed to have   one   email address in the system. When my internet-based course went through yesterday, the system went \"huh?\" and switched all my account info over to my grad school info. Poof! No more work account. This means: no more work-related email, no way to get in touch with me unless you call me, and no connection to my calendar! I can\'t check what I\'m supposed to be doing! Ugh.  Thankfully, the IT people are working like mad to get things taken care of. This morning they switched everything back over to my work account, but then the system decided to change my user name. Um, no. And my real name is wrong suddenly. Suddenly, my name is Jennifer. Quoi?  I think computers may already be able to think, and we just don\'t realize it. I think it is switching my name on purpose. I think it hates me. '),(2010,4166693,'07,August,2004','2004-08-07 00:00:00','    This came about while we were making burritos the other night:  R. shredding a lot of cheese into his burrito, \"If I had a restaurant, I would call them constipation burritos because there\'s so much cheese in them. And no one would disagree with me because they would all be constipated. And I would just sit back and laugh, because, no, I don\'t have any toilets.\"  It was the damned funniest thing I think he\'s ever said. Maybe only surpassed by the time he made me shoot milk out my nose. But I don\'t remember what comment made me do that, so this is his funniest line to date. '),(2011,4166693,'07,August,2004','2004-08-07 00:00:00','    I love when that happens. Even if you don\'t really know the friend because you haven\'t met, it\'s nice to know that people care about what you do and what you think.  I love my friends. They are all so wonderful.  And I want to learn how to post pictures so that B. doesn\'t feel as if he needs to say a lot. Pictures are just as important and words. I\'d like to share some of my pictures with people. I\'ll work on that. '),(2012,4166693,'07,August,2004','2004-08-07 00:00:00','    I was able to actually answer some questions for people today. I got to play with the freshmen. I have lots of meetings coming up in the next few days. Classes start on Friday for me (at UMSL). Classes start for the rest of the world on Monday (UMR students).  The shower is fixed.  The washer is fixed.  I have $0.91 in my bank account.  I have a check for $56 on the way.  I\'m hungry.  A guy named Dwight posted on my blog entry about sports. Right on, Dwight that I do not know.  I would like to write like Dr. Seuss one day. Maybe. I can\'t decide either way.  My neck is sore from sleeping funky on the bed to avoid rolling over on Mr. Lance Kitty.  I made my first call to an agency that represents a large, musical act and got....the voicemail. I left a message. It\'s all good. I feel important at work every once in a while.  Need to dress smart tomorrow--have to be at the directors\' meeting to meet lots of new people.  Um......that\'s all....for now.... '),(2013,4166693,'07,August,2004','2004-08-07 00:00:00','    Is an indication that I should be done posting for the time being.  I think I\'ll heed it\'s advice. The forgetting\'s advice that is. '),(2014,4166693,'07,August,2004','2004-08-07 00:00:00','    I was thumbing through various blogs, and I came across the blog of a girl who is waaaay into sports. Which I initially think is really wonderful and unconventional. I am into sports in that I watch them, some religiously, and I yell at the TV and hate it when my team loses. I am into sports in that I follow team scores, but I can\'t tell you who won the World Series in 1956. This girl strikes me as the type of person to be able to name who won the World Series in 1956. This is this super-cool.  Then I notice that she has established her own criteria for what a sport truly is, and in doing so has thrown herself into that whole world of sports snobbery, something I think she was trying to avoid. Her criteria are legitimate, however, she does pose a problem in the following through of the logic. A necessity is a measurable means of winning--faster, stronger, etc. Yet then she mentions that ice dancing is not a sport. I think it is. Unless you\'ve danced--and I mean danced in the form of actual dancing with steps and movement and grace, and not in that boom-chica way where you flail your arms like a monkey on speed on a dance floor--you have no idea how strenuous it can be. It is an amazing form of exercise, and it takes a lot of practice to get good at it. Besides, if you dance well enough, you\'ll be sore the next day from using all the muscles you never thought you had. Ever seen any of the guys that swing dance well? Yeah, muscles. It\'s a hard, hard thing to lift your partner and swing them around. It\'s even harder to effectively lead your partner. So take dancing, any kind you want that does not fall into the fake-tribal dancing genre, and put it on skates! That, my friend, is a sport. That is freakin\' hard. I think it matters if you\'re faster, stronger, and more precise than other ice dancers. It\'s a need to be those things if you want to win. I think that declaring it a non-sport is snobbery. Especially, if you\'re going to place curling on your list of sports, but negate ice dancing or synchronized diving. Curling--a sport in which you slide a heavy barbell down an ice lane by using small brooms to brush snow chunks out of the way.   Last I checked, you needed to be an athlete to jump off a 25-foot platform and do four twists and spins in the air before landing straight up and down like a straw in a giant pool of water. I know I couldn\'t do it. If it isn\'t a sport shouldn\'t everyone have some ability to do it? Hmmm..... '),(2015,4166693,'07,August,2004','2004-08-07 00:00:00','    I plan on using it to call people. Last night was crazy with the phone. Well, first of all, my parents called which is always a detailed conversation about what I\'m up to and what work has been like and if I have a place yet and etc. So, I had one of those. Then, A. called to check up on me, talk about current job opportunities, and relay wedding plans for the upcoming week. Then, J. called to talk about school, discipline issues, and lesson plans. I liked talking to everyone, but I was about wore out! Heehee. I think it was just the mellowness of Chinese food and the Olympics that truly wore me out. That and the fact that I took the opportunity of watching the Olympics to pull out some crazy kung-fu moves on R. I don\'t think he appreciated being \"chopped\" across the arm or having me pummel his legs. I bruised my arm. I think it\'s some form of cosmic retribution for rough housing too much.  But anyway.....  I will call you if I haven\'t already, I promise! '),(2016,4166693,'07,August,2004','2004-08-07 00:00:00','    I forgot about this until I burped. I know that is sooooo gross, but it was nasty then so the recurring nastiness just makes it worse. I had a grilled cheese and salad from the cafeteria for lunch. The grilled cheese? Oh, so good! I was in Heaven. The salad? Not so good. And here is where we get to the nasty cheese.  The salad had what appeared to be nice, yellowish orange cheese on top. I\'m thinking, \"Mmmm cheddar\" because I am such a cheese freak. R. says I was a mouse in a former life. Anyway. I take a nibble, and, oh God, the horror! It had this powdery texture on the outside that reminded me of how latex gloves feel on the inside. Ewww. Then, to make it worse, the cheese tasted faintly of cheddar but more like really rotten bacon flavoring. Gross! I wanted to yack. For real.  To make it worse, the tomato and the cucumber both had hints of nastiness to them as well. I can\'t place it, but it was almost a meaty flavor that frightened me since they are veggies.  Oh...I\'m shivering just thinking of how gross it was! '),(2017,4166693,'07,August,2004','2004-08-07 00:00:00','    I had a conversation with a friend the other day where she really bashed blogging. She thinks it\'s a stupid thing to do because it allows people to psuedo-stalk other people. Her reasoning makes sense, I guess, since people can read what you write and never post. However, when you write it you know people are going to read it that you don\'t necessarily know, so it\'s really more about publishing your words without the copyright. You want people to read about your life if you post. Or you want to let specific people read about your life, at least. Either way, it\'s like a diary that you can share without fear of retaliation or mockery. It\'s something I sort of have come to enjoy.  I know that she is bitter because of personal reasons which incorporate blogging but aren\'t necessarily typical of it. However, I found it somewhat offensive that she insinuated that all people who blog are stalkers. I\'m not stalking anyone! Besides, the people that I communicate through the writing I do here have all had wonderful things to say. I enjoy their comments and their insights. I appreciate that they communicate back to me. It\'s fun. And, let\'s face it, one of them makes my friend light up like I\'ve never seen before, so that\'s just phenomenal.  Blah to being negative.... '),(2018,4166693,'07,August,2004','2004-08-07 00:00:00','    That is what the band Emerson Drive (they\'re country if you aren\'t familiar with them) is considered according to the tax forms we received as part of their contract with UMR for the fall concert. Okay, that\'s fine.  Here\'s the problem--the way the UM System pays people who are non-resident aliens is much different since they still have to pay taxes, but aren\'t really citizens.  Emerson Drive doesn\'t want to pay taxes. Thus, our options right now are this:  1. Every band member fills out a bunch of paperwork and gives us a bunch of documentation regarding their legal citizenship status, and then the University figures out tax payment from there. Effectively: lots of legal paperwork to end up not paying taxes since they aren\'t residents.  2. We, the Student Union Board, mark up the amount we are paying them to cover their taxes. Effectively: SUB pays an extra 30% on their contracted amount, which is a lot, to pay their taxes for them.  Here\'s the hitch. The group doesn\'t want to do the paperwork mumbo jumbo because they said they never had to do it before. The students don\'t want to pay another 30% for the show that they have already contracted in for. Which leads us to the last option.  The band pays their taxes and then files their income taxes in April and gets it all back.  Why do I tell you all this? Because I am in charge of figuring out what the solution is going to be.  Ugh!  But non-resident alien is fun to say..... '),(2019,4166693,'07,August,2004','2004-08-07 00:00:00','    Today, I have to be a decision maker. This frightens me a little because typically I am a very indecisive person when it comes to minor choices. Major choices I\'m only slightly better with. Mostly, I like to avoid making the decision until the last possible second, or until the party that is vested in the event with me flips out and forces me to choose. This is not to say that I don\'t have a strong opinion about things. I think it just highlights my incessant need to be a procrastinator. However, today, I have no option to do so.  A little run down on Rolla, MO. First of all, it has a population of about 14,000 people. This includes the fact that Rolla is a college town. If there weren\'t a college here, I would estimate the population to be roughly that of my home town, Cedar Hill, at about 6-7,000. It\'s quaint. But when you take the quaintness and add into it the fact that UMR has reduced the number of credit hours required to move off campus to 45, you have a problem. I moved to Rolla about a week ago. I\'ve been to two realty places in the last weekend. The prognosis is not good.  Investment Realty has 3 apartments left in all of Rolla. One is a total shit hole, and I wouldn\'t wish living there upon a snake. And I hate snakies, so go figure how much I like the idea of a person living there. It was really dark, very super small, and had the type of people living around it that remind you of drug dealers or mental patients. Plus, there was the lingering smell of food in the apartment and outside of the buildings. The kind of food that smells old and rotten. The kind of food that makes me think of nursing homes. The kind of food that people shouldn\'t force other people to smell, let alone, eat. I immediately thought, \"Oh, hell no.\" Did I mention that the \"driveway\" to get to it is really nothing more than an alley between the bowling alley and McDonalds? Yeah....not going to live there.  The second apartment and third apartment are very close to each other in make/model. They were nice. One was set up much like a duplex would be except it only had one floor. It was spacey for a 1 bedroom. It was nice. I liked how it was bright and had vaulted ceilings. The only thing that made me cringe was that the back door was in the bedroom. Made me a little nervous. The second place was much like the first, except it was upstairs and had no vaulted ceilings. I liked that there was no back door, especially in the bedroom, and it had a nice view of a lake with ducks on it. I liked that a lot. I wouldn\'t mind either of them.  Here\'s the dum dum dum part. Investment Realty is leasing these places starting in September, however, they are requiring the deposit immediately. They said that they expected the two to go within a day or so. Additionally, they said that they couldn\'t hold it for me even if I thought I would maybe take it. Either I put money down then or they rented it out to someone else. Made me sad.  I also called Century 21 Realty. They have one and only one apartment left in Rolla. It\'s a 1 bedroom, $250 a month, NO pets. Once again, money now or no move in.  There\'s only one other realty company in town. This poses some serious problems, or so I believe.  I thought maybe the \"moving expenses\" that UMR offered to pay would include a deposit on an apartment. Sounds like a moving expense to me! Nope. It only covers mileage and gas and a U-Haul. No meals. No deposit on an apartment. If you don\'t have the cash, too bad. Sucks. I\'m a poor, poor girl right now. I think I have about $75 to my name. I was living all summer off the fruits of my springtime labor--i.e. Substitute Teaching--and that started to run out long before I actually had this job. I get paid once a month since I am salaried, so the 31st can\'t get here fast enough.  Here\'s the light in the dark with a little shadow thrown in. There is an apartment that is available that I\'m almost 95% sure that I can have. R.\'s parents own two houses down here. One, he lives in with two roomies. The other is split into an upstairs apartment and a basement apartment. Guess which one is open? The basement, of course. That is the \"little shadow\" part. It\'s very spacious--a full basement turned into an apartment--so that\'s great. A little too warm for me since the people upstairs don\'t like to keep it quite as cool as I\'d like, but I can invest in fans. I was totally against it at first because it wasn\'t my dream apartment. However, I have to be practical. It has a lot of space, and Stitch can come live with me comfortably without extra rent. R.\'s parents said I can paint it however I want to paint it, which allows me to dream about Trading Spaces-esque paint techniques and color palates. It\'d be easy to move my stuff in since it\'s on a ground level. I could do a lot with hologen lamps and candles for mood. And, above all, it\'s cheaper than what I would get elsewhere because they know me. So, I have to make my decision tonight or else R.\'s parents are putting it up on the market this Wednesday and I\'ll have no place to live.  It\'s a little stressful to be looking for housing in a college town at the beginning of their school year when you have no money. I\'ve been freaking out, but, on the flip side, if I can get the apartment for September, I can spend the next few weeks/weekends over there painting and scrubbing and making it look the way I want it to look. That would be nice. I\'m a little more relaxed. I\'ll be even more relaxed if they say \"Yes!,\" and I know that it is mine.  Wish me luck. '),(2020,4166693,'07,August,2004','2004-08-07 00:00:00','    I suppose that I should start this post off with a disclaimer.   Disclaimer:   I still love my boyfriend even if he is the most forgetful idiot sometimes. I\'m sure that I also render myself a fool from time to time (an instance with a particular brand of donut comes to mind), so I cannot harbor ill-will toward him for long.      Now, here\'s the reality of things.  My morning was a morning from hell. And not just in that woke up late and feel hung over sort of way. No, no, no. It was much more in that nothing can go right way that makes you feel like laughing and crying at the exact same time much to your own dismay. It was one of those mornings that starts off with a bad night that just circles its way into the rest of your day. Let\'s start from the beginning.  Crawled into bed at about 11 p.m. last night after a long, hard night of playing  Mario Kart .   I was ready to hit the sack something awful. I pull back the covers, turn on the fan and jump into bed while R. putzes around the house turning off lights and avoiding tripping over the cat. He crawls into bed a few minutes later and hits the lights after setting the alarm clock to 6:45 a.m. Kat, his new roomie, has moved in and has to be at work at 8 a.m., so I need to be up a little earlier to get in the shower and get ready for my 8 a.m. day. Hence, the alarm being set earlier than usual. I pass out cold within ten minutes. I am woken up about ten minutes after that to R. yelling at the cat to get off the desk. He jumps up and throws him on the floor. I mumble nasty things about stupid cat and go back to sleep.  I wake up at 2:00 having to pee like you wouldn\'t believe. I attribute it to the diet Dr. Pepper and stumble to the bathroom, all the way cursing my need to drink soda that late at night. I flush, the toilet decides not to wash the toilet paper down, as it often does if don\'t hold the handle for precisely the right amount of time before washing your hands, and I say \"screw it\" and shut the lid. Whatever. I\'m gross, sue me. I crawl back into bed, arrange myself next to R. and around the cat, and fall back asleep.  Train rumbles through about 3 or so and Lance, our cat, decides that this is his cue to be bitchy. Mroew...meoroow.... Scratch on the bed. Turn in circles. Jump on me. He tries to physically push his way in between my hip and R.\'s even though there is only enough space for maybe one of his legs. I grumble and push him away. He licks my leg and lays back down. Fall back asleep.  Lance decides it\'s 5 a.m., so he better wake us up again. This time, he scratches at the door and meows loudly. He\'s being ten different kinds of obnoxious. I want to kick him. R. throws him out the door and shuts it, but it\'s too late. I have to pee again. More cursing of the Dr. Pepper and some cursing of the cat. Back to bed. Sleep. Lance pushes the door open and runs into the room to start his crap again. Meowing and jumping around and nuzzling your face to wake you up. It\'s 5:15 a.m.! I don\'t have time for this. R. throws him out again and pushes the door closed then puts a chair in front of it. Sometimes it latches, sometimes not. Hence, Lance getting in when we shut it earlier. I mumble something about setting the alarm for 6:55 because I need the extra ten minutes. I hear the beep beep beep of the alarm buttons and R. rolls over to tell me it\'s changed.  Sunshine on my face. What time is it? HONK HONK HONK. The neighbor gets his morning wake up from the guy out front. He always honks at..... dammit, R. what time is it? Um, 7:05 a.m. What happened to the alarm; it didn\'t go off! Oh....I, um, I set it for 7:55 (I have to be at work at 8 a.m., remember?) Jump out of bed, throw back the covers on his side so he is cold and has to get up. Stomp to the bathroom. I grab my toothbrush and brush like a mad woman. Then head upstairs to Clay\'s apartment to shower since the bathroom downstairs is still in disrepair a week after we got there. The toilet up there has no water. What the hell? Oh, R. turned it off to fix something. I try to get in the shower. No water there. R. turned it off when they fixed the pipe yesterday afternoon. He turns it back on. I turn the water on. White chunks fly out of the faucet and the room suddenly smells like acetone. They used acetone to fix it, but it should be dry. I\'m going to smell either like dirty person or acetone. Hmmm...I let it run for a minute more and then just jump in. No time to shave the legs, I\'ll wear pants.  Hop out of the shower and head downstairs in my towel. Meet Kat on the stairs going up to shower and apologize quickly for the tardiness. Start to throw on clothes. Blow dry hair. Style. Put on make-up. Realize that I\'m still a little bit on time. Run down to the basement to grab a few decorations for my desk. Grab three picture frames (w/pictures), my diplomas from MU, and my Yoda figurine. Throw them into a bag. Grab my insurance paperwork. R.\'s made me a bagel with cream cheese. Yum. He\'s also given me $6 for lunch since I had no time to make my lunch this morning. Yay.  I realize I can\'t find my keys. R. had them last for the run for chili powder yesterday evening. I ask him where they are. He can\'t find them in his jeans\' pockets or on the desk. I assume he locked them in the car, but that\'s okay because my driver side window will go down if you push on it a little. I run outside. Door is unlocked--that\'s not good. Keys are in the ignition--um, really not good unless you want your car stolen. I show him my most steely look. I throw the stuff in the car, retrieve my bagel from him, and give him a quick peck. I turn the key in the ignition and realize it doesn\'t have to turn very far. Quoi? I look down. Yeah, it\'s been on all night long. The key was facing in the forward \"on\" place that isn\'t really on, but lets your radio work anyway. What the hell? My battery is dead! Dammit, R.! What were you doing? He frantically calls everyone in Rolla that he knows, all three of them right now. Kat\'s left already. No one is answering. I look at my cell for the time. 8:13 a.m. I was supposed to be there at 8 a.m. It\'s only my 5th day. Growl. Growl. Growl. I grab my stuff and start marching up the sidewalk--I have to walk, blister on my foot from yesterday\'s shoes and all. (A string of nasty curse words is going through my head at this time, but I\'m too much of a lady to share them all). R. chases after me. I turn around and give him my most deadly look. He puts out his bottom lip and gets the puppy eyes. I crack a smile. Argh!!! And I wanted to be mad.  I got to work at 8:23. Not too late. Still, it sucked. I have officially been offered a nice dinner/ movie night for my trouble this morning. And it\'s all because of Robo. At least that\'s what he said. It was just one of those mornings that you expect on a Monday but not on a Friday. Although it is Friday the 13th. Maybe there is such a thing as bad luck. Or, as R. likes to say, it was a miracle. God made me late this morning on purpose because he didn\'t want anything bad to happen to me and it might have if I had actually gotten out of the house on time. Whatever it was, it was irritating at the time. Although it has the funny angle now. Hmmm....  I need to do some work. I\'m outta here for now. '),(2021,4166693,'07,August,2004','2004-08-07 00:00:00','    I\'m going to stay a few minutes late today so that I can actually post a few things that are intriguing. Or maybe just a few things that are....  Spent the day in meetings and meeting people. Worked with a few of my SUB (Student Union Board) students today which was awesome. I think I\'ll have a good relationship with them as long as I clarify that I am learning, too, and that we need to work together. They\'re all really nice, somewhat quiet, but nice and very smart. I\'m glad I have a great group to lead me through the ins and outs of SUB for a while.  I didn\'t decorate my desk at all. I was going to, really, but I got so sidetracked with working on other things and going to meetings that it all got lost in the mix. Tomorrow is my work \"me\" day, so I will do it then. I have things on the schedule, but they aren\'t too much, so I should be able to do a little of everything I feel I need to do. More work with the students tomorrow. Some student meetings. Good things.  I have apartment hunting on my list of \"to dos\" for Saturday. I looked at R.\'s parents\' place, and though it\'s okay, it\'s totally my back up. I\'d rather live in an apartment apartment. Not in the basement. Does that make sense? It\'s very dark and warm (which is weird for a basement) and reminded me of many creepy movies I have seen in some ways. I can\'t describe it really. I wish I could. So looking at places is on my list of things that must get done. I hope I can find a nice place that accepts pets. I have to bring my cat down here, so I have to find a pet friendly place that isn\'t a shit hole. Good luck, right?  I received my first graduate class A- and B+. They hurt a little. I worked really hard in those classes and thought I deserved a little better. Somehow, UMSL is still calculating my GPA at 4.0. I don\'t know how that works, but I can\'t complain. I hope that it\'s not just fostering false hope. The jerks. I\'ll survive without the 4.0, but it will be hard on my fragile ego.  I miss my friends and my family. I really miss Amanda and Jessica. They seem too far away. My family is more of just a longing for a hug from my mom and not really a want to move back in sort of feeling. I\'m okay not living there. I want to see them soon. And I do miss Rachel even though she was driving me nuts before I moved out. I miss her quirkiness and her cracking me up with her strangeness. It\'s a weird combination of love and hate with us sometimes, but I love her a lot and I miss her. And I also miss talking to Bruce. I used to talk to him every day, and now, with no internet access at home, I can\'t. It\'s sad. I should call, but I\'m bad with the phone and I don\'t like to interrupt. That\'s probably why I don\'t really call. I should, though, only with permission.  An office supply club is in the works. Keep watching for details.  Hmmm..what else? Staff shirt day tomorrow. And chili for dinner tonight. I\'m outta here! '),(2022,4166693,'07,August,2004','2004-08-07 00:00:00','    Once again, I am sitting here at work contemplating things to do. I know that tomorrow will be really hectic. Tomorrow I learn PeopleSoft, a new database program, and I go over how to reconcile my university credit card. I also have my first staff meeting. I also learn how to edit contracts with our campus performers. All in all, there is a lot going on tomorrow. It makes me only  slightly  nervous.  Friday is my day to run my own day. I know that I have a meeting with Dustin, a rep for the BSU, to go over plans for their upcoming concert. I\'m not sure if they realize that using the Rec Center actually requires spending $3000 on set up and removal. I\'m not sure how they are paying for their event. I\'m not even sure if they are planning on going through the box office for tickets (another $1000). I do know that this is my first non-SUB event. This is also the first event that I get to handle all on my own. It\'s really quite unnerving. I would like to meet with my SUB director who is handling the Emerson Drive concert in October so that I can get some specifics. I might see what I can scrounge up for a meeting with him on Friday. Just so that I am in the know. I have my first event to attend on Sunday. It\'s a big week with lots of big events.  Next week is freshman orientation week. I have three solid days of academic events in the mid-morning/early afternoon. In addition, I have staff meetings, CPAS meetings, Homecoming meetings, etc. Now, not all of those will necessarily take place next week, but a good portion of them may. Then I have evening events that I have to attend. And weekend events. The only nice part of that is that I get to come in late/take time off during the week to balance everything out. I wouldn\'t want to work 60-75 hours a week! No way!  I think tomorrow is also decorate my desk day. I\'m bringing in some supplies from home. Pictures, pens and pencils, etc. Then my office supplies come in on Friday or Monday! Hooray. I put down about 11 things that I wanted to decorate my side of the room. I can\'t wait until I get my real office in October. A real room with four walls and a door. This is a real room, too, but it\'s much more open and not private. Not that I am a super private person, but to have some escape space would be nice.  Wal-Mart run is on the agenda for tonight. Hopefully, by the time I get home at 4:30, the shower will be fixed. That\'d be nice. Right now it is really not working at all. Neither is the washing machine. Or the dryer really. It was a list of many problems when we arrived on Sunday afternoon. Thankfully, R. has handled everything splendidly and allowed me to be mostly mellow this week. I get really crabby at the cat in the morning. Mostly because he\'s an ass this week. Meowing all the time in a bitchy-cat way. He likes to meow loudly at me at night when I move even the slightest bit. I feel like saying, \"Screw you cat! You sleep all day while I work. I don\'t care if you\'re comfortable.\" Maybe I should say it.  Life is good. I wish I could comment on conversations with people or on my social life, but other than R. and I talking about cleaning or making dinner, there\'s not been much social about my life. Mostly just work. But that is what the first week is all about. I like it a lot. The day goes by really fast..... really fast.  Okay, I\'m out. I\'ll post something decent here eventually. '),(2023,4166693,'07,August,2004','2004-08-07 00:00:00','    Sitting here at work killing time while my boss, Paul, is off doing something highly important in the Student Life field. I feel the need to be doing some important, however, I am posting here on my blog. Maybe it\'s important to some people. I did get to pick out office supplies today! Hooray for that. I love,   LOVE  , office supplies. I love looking at them. I love touching them. I adore buying them.  I am such a nerd.  I need to go home and grab some lunch, actually. It\'s a good plan. Whether or not I follow through right now has yet to be determined.  Hmmm....  My boss has a meeting at 1 p.m., and Katie, the support staff, will be downstairs at the front desk, so I\'ll be all alone in my little room of doom. I wish I\'d brought my desk decorations. I could have been doing something constructive with those. Note for tomorrow--bring desk decoration and files. Yeah.  I\'m turning into a boring girl. I\'ll post something important soon. I promise. '),(2024,4166693,'07,August,2004','2004-08-07 00:00:00','    Second day of work--spelndiforous. I\'m meeting waaaay too many new people, but at least I feel all official about it. The people are tremendous. I love them already. So much fun. I\'m really excited about my SUB students; they are so independent and cool so far. I\'m really pumped about my first activity as Student Advisor. I have to be at the comedy night Sunday night. That means I have to supervise so that our comedy guy doesn\'t light himself on fire. Fun, eh?  I\'m being placed in a position with a lot of responsibility. I serve on 10 committees. That\'s just a fraction of what I have to do. However, it\'s going to be such a challenge that I\'m going to love it. I know I will. I always have loved being stretched to the limit.  I need to get out of the office. It\'s already almost 5 p.m. That means, I\'m staying a 1/2 hour later than I need to. I\'m supposed to be looking at an apartment soon. I\'ll post more later. '),(2025,4166693,'07,August,2004','2004-08-07 00:00:00','    And that is all there is to say about my days of packing. Finished. Hooray.  Now, all I need to do is crawl onto the couch and watch some TV before hitting the sack with Monty. Monty would be my stuffed moose, ya\'ll.  How many people can say they have a ginormous stuffed moose from Colorado? Hmmm...if there are a lot of you, then I suddenly feel in good company. If there aren\'t, then I feel like an individual.  Note to self--do not forget your Gamecube when you leave tomorrow. Bad idea.  G\'night all. I\'m outta here.... '),(2026,4166693,'07,August,2004','2004-08-07 00:00:00','    I just  had  to come back and post again. Why in the world is my advertisement for brain tumor diagnosis? And I thought I\'d seen it all. I was so wrong. '),(2027,4166693,'07,August,2004','2004-08-07 00:00:00','    I\'m not quite sure if I\'m going to be keeping with this blog or not. Ideally, I just want to have it so that I can comment on all of your blogs without having to hit anonymous and add length to my post by signing my name. It\'s just a hassle. Although, my livejournal doesn\'t get read that often, so I might just connect my posts or consolidate down to one.  I\'m not sure.  I should probably figure that out.  I\'m avoiding packing. Let\'s be honest. I just don\'t want to pack anything else. And the big packing issue at the moment is that I need to pack my shoes. Ugh. Too many. Too much work. I want to sleep or watch TV. I\'m a lazy ass.  I hope everyone is having a good weekend. Especially, two people that I know (or know of).  Okay, enough of this. I\'m off to pack. '),(2028,3415194,'26,May,2004','2004-05-26 00:00:00','   Sleep.  Soft, wonderful, dream-filled, night-passenging sleep.  How elusive.  How estranged from this mortal being.  There are nights when even pills don\'t help.  '),(2029,3415194,'23,May,2004','2004-05-23 00:00:00','   If it isn\'t one thing it\'s another.  First I start having trouble with my lower back and now my right shoulder is giving me pains.  My diet isn\'t working, the cholesterol medication the doctor gave me has painful side-effects, it\'s allergy season and my landlady is being a bitch this weekend.  I can\'t get a minutes peace even when I leave the house.  '),(2030,3415194,'21,May,2004','2004-05-21 00:00:00','   Well, here it is another Friday and I feel like this was another useless week.  It doesn\'t matter how hard we work in this office it is like there is never any progress.  Kids still get hurt and parents still lie.  Other states ask for help, we ask for help, everyone asks for help.  And the help never comes.  So you pick up another doughnut or soda or candy bar or . . .   And such shining examples of denial.  We have a father that come in to visit with his two year old daughter and screams for dear life during the entire time and he tells the Judge that the visits are going fine.  Kind of hard to keep that \"professional attitude\" going when even the psychologist is laughing at the guy.  '),(2031,3415194,'10,June,2004','2004-06-10 00:00:00','   A sense of humor is very important in this job.    They want me to do more work.  BUT they aren\'t paying me more.  AND I am supposed to be happy that I have a job.  My benefits are decreasing.  But my needs are increasing as I get older.    Next the insurance premium will increase.  But I won\'t get a cost of living increase.  That\'s always good for a laugh.  '),(2032,3415194,'08,June,2004','2004-06-08 00:00:00','   Well?  Why not?  I don\'t feel good.  And I will probably feel worse tomorrow.  Can\'t this just go away and stop?  '),(2033,3415194,'03,June,2004','2004-06-03 00:00:00','   I know, reliving childhood nightmares is supposed to be an OPPORTUNITY.  I know, I should be addressing the \"inner demons\" in order to make a full recovery.  I know, Susan can\'t read my mind and cannot possibly know what kind of nightmare I was living through.  But gee, does this kind of event have to be so brutally EFFECTIVE?  Oh wait, you mean if  I  wasn\'t in such deep denial then this wouldn\'t have been quite so bad?  You mean if  I  were taking care of my problems this might not have even happened?  Damnit!  Why does everything always have to be my fault?  '),(2034,3415194,'02,June,2004','2004-06-02 00:00:00','   1) I am not perfect. 2) I am not well physically. 3) Money issues are forcing me off of my Prozac. 4) The leasing company is fucking with me. 5) Showering in the morning is becoming an \"issue.\" 6) I am regressing again. 7) Where is Danni anyway?   8) Last night Susan called it her computer. Of all the horrible images that raised in my mind.  My childhood came rushing back at me and I saw my parents taking away my best-loved toys because someone else wanted them.  I was so devastated.  If you haven\'t experienced the horror of being an unwanted child just try and get a clue.  They always took away the things I liked and used.  They would say \"You\'re the child we never wanted.\"  To this day it still hurts.  '),(2035,3415194,'02,June,2004','2004-06-02 00:00:00','   Here it is June already and I am just now finding Christmas ideas.  I am usually more organized than this.  But then this has been a very chaotic year.  Overwhelming debt,  moving, theft and puppies and God only knows what else is yet to come.  Does this mean giving up the Internet?  I don\'t know but I can no longer afford to pay for it.  So I borrow access from others.  '),(2036,3415194,'13,July,2004','2004-07-13 00:00:00','   I see myself in a house and I am surrounded by friends.  The house has a brown tile floor with plush area rugs.  I am walking toward the sliding glass door to go outside.  There is a small pool and a large grass area.  Perfect for playing.  There are trees and a small garden in the yard.  It is never too hot or too cold and the rain never lasts too long nor gets us too wet.  My friends are the pets that I have known over the years.  Omar, Twinkle, Bear, Buddy, Princess, Samantha, Jasper, Muffy, Sampson, Tinkerbell, Bruiser, Romeo, Honey Bear, Chase, Critter, Mattie, Augie, Barney, Wit and Charm.  We are walking out to the backyard so that we can play.  Bear gives me a kiss and I thank him.    This is a happy visual that I can use to combat depression.  '),(2037,3415194,'12,July,2004','2004-07-12 00:00:00','   A month goes by and suddenly everything is different.  Well, . . . maybe not so suddenly.  I have been reading Dr. Ornish\'s book about reversing heart disease and had a comeuppance or two.  Or three.    I am forced to admit that I am the unwanted child.  From infancy I have had to force myself on others to be accepted.  Other people never like me they just tolerate me.  The only reason I am tolerated is because I am useful in some way.  It is necessary that I admit this is my problem and that I need to accept this reality and learn to live with it.  I am forced to admit that there is an invisible wall surrounding my heart that will not open.  It is protecting my heart from me.  Nothing has changed.  '),(2038,1232458,'06,August,2004','2004-08-06 00:00:00','    Hey kimmer let me know if you get this so i know how to give comments. dad has a lot of info on greece for you. did you know the apostle paul attended the first olympics   '),(2039,1232458,'02,August,2004','2004-08-02 00:00:00','    Our first \"unofficial\" unit study will be on the 2004 Summer Olympic Games.  We will be studying every subject in regards to the olympics, Greece, gymnastics etc..  We have about two dozen books from the library on such subjects as the summer games, the Parthenon, Mary Lou Retton (winner of the gold in the 1984 summer games), Bruce Jenner (decathalon winner), Greek Mythology, Marion Jones, Olympic history.  We will also be looking at the significance of the Torch, Olga Korbut (medal winner of the 1972 summer games), Nadia Comaneci (7 perfect 10\'s and medal winner of the 1976 summer games), we will be listening to Nadia\'s song, we will explore greek food and desserts and study the globe.  While researching all of these subjects in our books we will incorporate spelling, vocabulary, math, science, history and crafts into the study.  We have made some really fun crafts to decorate around the t.v.  We made two olympic torches to go on top of the television, we have American flags hanging on the wall, we made the Olympic rings out  of colored pipe cleaners and we are hanging olympic articles on the wall.  We took the girls and two of their friends to a Greek cafe for lunch on Saturday and we made them eat Greek food (ok, so it was only a grilled cheese pita),  Michael and I were a little braver, we tried gyros.  While we were waiting for our food the girls all worked on an olympic word search puzzle that I made out of  grid paper.  There is a restaurant close by called \"The Partenon\" so we will take our Parthenon book with us when we go to eat another Greek meal.  I told all the girls on Saturday while we were all having a blast that this is homeschool.  We had our Tivo unit shipped to us so that we could record all the events and pause and rewind whenever we wanted to discuss something or see it again.  We are going to eat Greek desserts while we watch the games.  This is basically an introduction to the girls on what homeschooling is going to be like, and how our lives have changed.  We are having a blast.   '),(2040,1232458,'01,August,2004','2004-08-01 00:00:00','    I am also expecting that God will be my guide. I have always prayed that if we were supposed to homeschool that God first would give me the desire and then the ability. Well, here we are. He has definitely given me the desire and a million creative ideas and I now expect that He will give me the ability to do this. When we left Florida I asked God to change me. If I had to go through this painful and emotional move then I wanted to be changed for His glory. One of the things that I have always despised within myself was the lack of discipline and diligence in my life. My heart\'s desire is to be stronger in both those areas. Guess what? He is changing me . The only way I will be able to successfully homeschool my daughters is to exercise more discipline (being in the Word and prayer on a more consistent basis) and diligence (sticking to what I have chosen to do). I\'m sure that I will still \"fly by the seat of my pants\" just as I always have, but that will be the exception and not the rule to my life. I don\'t plan on giving up my spontaneity or spirit of fun, it will now be incorporated into our homeschool life. I can\'t wait to just up and announce to the girls that we are doing school in our pajamas, or guess what, we are going to do school at the pool, or who wants to blow off school today and go shopping, etc.....Please pray for me that God would direct my paths and give me discernment.   '),(2041,1232458,'01,August,2004','2004-08-01 00:00:00','    My plan for Hannah is similar to Sarah\'s only in the math curriculum. Hannah will be using hands on manipulatives called \"fraction layovers\" and will also share the blocks with her sister. This program is designed to teach the child more critical thinking, and to give them a clear understanding of the math that they are learning and now to use it and apply it in life. I am sure she will do great. Now her other curriculum is called \"Total Language Plus\". This is based on the reading of a lot of literature and similarly to the math curriculum, will teach her critical thinking, make her think outside the box and come up with her own answers and ideas. Hannah\'s spelling, grammar, and vocabulary will also come from this language program (plus). Once again we will be doing a lot of unit studies to incorporate every area of their education. We as a family are looking forward to a lot of field trips in our RV for these unit studies (Washington D.C., Mt. Rushmore, Grand Canyon, Williamsburg, Underground RailRoad, etc.) I think Hannah is going to group spiritually and intellectually over the next year.   '),(2042,1232458,'01,August,2004','2004-08-01 00:00:00','    My plan for Sarah is to really concentrate on her comprehension of phonics. I am starting her on a program called \"Explode the Code\". This should give her a firm foundation to build her reading skills on, hopefully boost her confidence so that she will fall in love with reading. It has been suggested that I don\'t force her to read books right now, just concentrate on the Code and give her a chance to develop her reading skills. I was told that this is basically called \"delayed academics\" and if we give our children a little time to get ready naturally then they will soar as soon as it clicks....kind of like potty training a toddler....you can\'t force them until they are ready. I also ordered her a math curriculum with manipulative blocks to teach her to understand the math that she is learning....not just memorize it. It looks fun. Those are my main two focus points for Sarah for at least the next four or five months. Then of course I will be using unit studies on all kinds of fun topics to incorporate the spelling, history, science, crafts etc....   '),(2043,3782048,'25,July,2004','2004-07-25 00:00:00','   I think I understand what Cobain meant by the famous line in Aneurysm.  I told Charlie how weird I am. When ever she texts me or signs on to msn, or i see her, i get excited.Its sad. What\'s more scary is sometimes I feel pain when I see her. Theres nothign wrong with her, I love her. But I sometimes feel pain coz i love her so much.  Oh well. '),(2044,3782048,'23,July,2004','2004-07-23 00:00:00','   Do I care.  I have more important things on my mind at the moment. I don\'t care if people think I fancy Charlie. People think war\'s wrong. People think it\'s right. People think the Audi\'s a better car then the Lotus, people don\'t.  It\'s all opinions to me.  And anyway, as long as I know it, and Charlie knows it, that\'s all that matters. Lana says I\'m just as bad as those people who accuse me, for saying I don\'t care. I know I\'m bad. I\'m no good, but I don\'t really care what people think of me anymore.  There\'s nothing wrong with the idea of fancying Charlie. It\'s just people seem to have a big problem with me having a friend.  &nbsp; '),(2045,3782048,'22,July,2004','2004-07-22 00:00:00','   This always happens. I have a bad day, and then I go and ruin it all by asking Charlie a really stupid question.   Today\'s topic was \"do you trust me?\"   Now she\'s angry at me I think, she won\'t text back, I only asked her because I feel like she doesn\'t trust me sometimes.   THere are times where I tell her anything, she tells me anything, and it\'s great. I\'ve never felt so open (mentally) to anyone before.   But there are times, when she and Ellen/Andy will just abandon me to have a private talk. Pisses me off. I\'m sure there are things between Andy and Charlie that no one else should know, of course. But when Andy and Ellen walk off or Charlie and Ellen walk off, I feel like no one trusts me.   It sucks. And charlie thinks I don\'t trust her. :( '),(2046,3782048,'22,July,2004','2004-07-22 00:00:00','   I don\'t feel important. Simple as that.   So many things prove it. Small things. Like people not trying to find me on a train. Like not going after me to see if I\'m ok&nbsp;when I\'m pissed off. Like not asking me how I am. Like not askin me how I am as much as you ask other people. Like rarely being invited out places.   There\'s more.   &nbsp;  I feel like my head is going to explode if I don\'t talk to someone. It hurts.   &nbsp;  Things keep me going on. Somehow. I call these things, memories.   I remember, when we went to the cinema on Monday, to see Spiderman 2. Charlie and me were in Tescos, and she told me how she was going to jump on me and hug me, because we hadn\'t seen each other for a while. That would of been cool (the hug), but she was worried what Andy might think.   There isn\'t anything to think. Nothing\'s going on, nothing ever will. We\'re best friends.   She jumped on me later, and punched me. But i got a jump hug thing after which was cool. I want another one.   &nbsp;  Don\'t really know what\'s making me sad. I can name things, but I can\'t pinpoint them.   Just want to spend some time alone with my best friend. I like it like that \'cause we talk about anything. Also, she must perform surgery on my beloved kurty!!!!   Poor guitar. I love him. Love charlie too.   Sometimes I wish I was an emotophobic. I have to complete that poem. It\'s basically done.   I hugged Charlie from behind. That\'s the first time I\'ve hugged someoone like that.   Why am I jealous of people being happy?   One painful thin is when you know someone\'s lying, just because they know the truth will hurt you.   DOnt know whether or not to hug them and thank them for caring, or to hate them for lying. '),(2047,3782048,'22,July,2004','2004-07-22 00:00:00','   Does anyone? '),(2048,3782048,'22,July,2004','2004-07-22 00:00:00','   Today     wasn\'t too bad.    &nbsp;  We didn\'t stay in town for long. At first I was REALLY pissed off. When I was on the train, I was bored. Got to Wool. I thought \"yay! my friends are here!\". I thought I may as well just wait for them to find me. We always do that. If me and andy get on the train we find Ellen and Charlie. Andy wasn\'t here today. So, I waited. They never turned up. Got off the train, they got off a few carriages down. Ellen said they just saw an empty table and sat there. Didn\'t even tell me to go find them. Now, I\'m sure if it was Andy, Charlie, or Ellen in my place, we would of gone to find them. I was then beginning to wonder if they even wanted me on this trip.    &nbsp;  I guess I was pissed off&nbsp; about before too. I asked Charlie if we could go to her house or something. She agreed, and said she\'d have to ask her mum, because they have guests (some very cool guests). That\'s cool. Our plan is overwritten, and Andy sleeps over the night before, I think. I\'m not going to get angry at her or ignore her, because I love her, and plus I don\'t have any solid facts. Nothing. I called Andy\'s house, to talk to him about what train he ws getting, and his dad said he was at Charlies, probably.    &nbsp;  Still, it pisses me off.    &nbsp;  I was kinda hoping the secrecy would go away today. Didn\'t happen. Charlie and Ellen would run off from time to time to have a chat. At one point we went into a shop, Lemon Tree or something, after they went in. When we went in, they left, quickly, and walked off.     &nbsp;  I really have to ask if they trust me or not. Doesn\'t seem like it sometimes.    &nbsp;  I brought Charlie a ring today. I\'m happy about that. I made her smile. Something I\'ve tried to do for so long. I don\'t mean like saying \"smile charlie!\" but I mean making her happy that she smiled.     &nbsp;  I just wish I made her smile. I wish I was more important to her too. Too much to ask.    &nbsp;  But she deserved that ring. If I could I\'d buy her a better ring, don\'t really care about the price. But, it would be seen as a \"mo fancies charlie!\" thing. I wouldn\'t care, I love her, lots, and I don\'t give a fuck what people think. I don\'t think she does either.    &nbsp;  She wouldn\'t want the ring anyway. I\'m sure Andy will buy her one.    &nbsp;  Good news, She can come round my house. My parent\'s don\'t mind. Maybe, just maybe, we\'ll be alone for some time. I don\'t want to sound possessive.    &nbsp;   &nbsp;  Anyhow, peaceeeee '),(2049,3782048,'21,July,2004','2004-07-21 00:00:00','   Fuck it feels like I\'m writing a suicide note. I guess I can call it that because everything will be fucked up more if anyone else reads this.  Where to start. I have no idea.    Dad says I can invite Charlie round my house. His image of girls, was gone long ago. I\'m glad.  Today, I felt ignored. Ellen and Charlie ignored me sometimes. Like when they went off to some shop whilst we were in Poole, they asked Andy to look after Tom, said goodbye to Andy and Tom, and thanked Andy for looking after Tom. THey just kinda forgot I was there too. Oh well. Also, the secrecy was pissing me off loads. Hiding a few things is ok, but Andy/Charlie/Ellen kepy going off to have little privacy chats. And here I am, thinking they can trust me.  Charlie thinks I don\'t trust her. I do. I just don\'t want to tell her what\'s wrong, because she\'ll hate me. I trust her more then myself, and when she thinks I don\'t trust her, it hurts. It\'s like saying I don\'t trust myself.   I try not to hide things from friends. Yet they just seem to want to hide what they can from me. I like being kind, but it\'s not done much. I remember reading a notice on a message board in St Osmunds, that read \"treat others how you would like to be treated\". At first I laughed. It was on a RE Board, and I didn\'t believe it. Yet it stuck with me, and I soon believed it. Does it work? Hell no. I\'m still going to try and be as kind as possible.  We\'re going to Poole tomorrow. Don\'t really know if I want to go. I want to go because these are great people I\'d be spending my time with. I don\'t want to go because I\'m fed up of the privacy. I just walk with the group, for a while, only to notice them slow down, so they can talk. Tom continues walking, so do I. But it feels like they have no trust in me. I\'m also fed up of basically being ignored half the time.   I\'ll go, and see what happens.  I was meant to be going to Charlie\'s house tomorrow anyway. She just overbooked it with this visit to Poole. For some odd reason that pisses me off. It shouldn\'t. I like these people. Maybe I just want to be alone with her for a while. Everyone else spends time with her alone, I don\'t. And I can barely tell her that.  As I write this, I have a gut feeling. I want to share this with Charlie, yet I feel she\'ll hate me after reading it.  She only ever hurt me with things she said because it was her, Charlie, my best friend, who said it. Anyone can say those words. They don\'t bother me much. But when my best friend, who I love so much, says it, I start to wonder.  &nbsp;  Stomach problems are scarily overwhelming my mind. When I\'m not thinking about life, the stomach pains come out to play like usual. I am seriously fucked. I can barely go a good hour without feeling a huge increase in pain. Things were meant to get better. I\'m going to hospital.  Seeing a psychatrist soon too. August 2nd 3:00. The date I pour my feelings out in front of a stranger. Hopefully a friend will be there.   Home life is shit. Don\'t want to think about that.  Sometimes I want to give up and die.  Sometimes I give up wanting to die.  Sometimes I feel dead.  Sometimes I wonder why.  I really need to talk to someone.  &nbsp;  Peace,  M.&nbsp; '),(2050,3782048,'10,July,2004','2004-07-10 00:00:00','   I was kinda hoping that epople would remember I was going LOL. Tonights been hectic...my parents have been fighting loads, ym dad thinks I\'m so irresponsible I#ll go missing in Iceland. Maybe I\'ll fulfill his dreams. I\'m tired as hell. Can\'t taking any sleeping pills coz I wont wwake up early.   Mona came into my room and gave me a big hug, then asked me to put her to bed \"one last time\". All this \"last\" stuff is scary. AM I going to die? To be honest I am scard. I don\'t want to go righht now. Things are still very insecure.   Damaged things can be fixed with a touch. Damaged things that arent touched, slowyl crumble.  Oh well.  Will an yone miss me? Mona.  Ilove Mona I love Charlie I love Andy I love Ellen. '),(2051,3782048,'07,July,2004','2004-07-07 00:00:00','   I don\'t want your sympathy, Just someone to re-assure me, That safety in numbers is the policy.  I don\'t want your sympathy, Just someone to understand me, A promising friendship, A friend to help when I\'m sad a bit.  I don\'t want your sympathy, Just a check of reality, That I\'m not the only one, Here.  I don\'t want your sympathy, I just want your curiosity, To know that you want to help me.  I don\'t want your sympathy, I just want you really, To tell me there\'s no fear, Because you\'re here.  But I have your sympathy. '),(2052,3782048,'07,July,2004','2004-07-07 00:00:00','   Love is between \'him\' and \'her\', There\'s no place for a follower, It gives me a headache, The fact that I cant take, Is that my best bests are leaving me.  People in general, Want me to avoid my best friends, And their opinions are meaningful  I think i should saty away, Come back someday, When it\'s alirght, And I\'ve seen the lgiht.  But maybe people would let me, Be with who I want to be.  He misses her, She misses him But no one wll miss me, Hopes wearing thin.  People tell me to leave them alone too, I just want friends like anyone would do.  I want those friends. '),(2053,3782048,'07,July,2004','2004-07-07 00:00:00','   This is that poem i spoke of earlier. I think it\'s a good poem. Right now, things are shit between me and Charlie I think.   She... She makes it alright  When it\'s all gone wrong She\'ll soothe your ears With her soft song,  She\'s the only thing That holds me back, From killing myself, It\'s just hope i lack.  But she fills me with hope, Tells me I can cope, And I believe her too, Without her, I can\'t imagine what I\'d do.  I could talk about her all day, But I\'d know what you\'d say, â€œMo you love herâ€, So what if I do?  She sees my life as a blur, And cleans it with a hug, So gentle yet meaningful, A gift from her is special, However big or small.  She makes everything bad Seem like the best miracle I ever had, She taunts the hateful loneliness, With her kindness, She tells it to leave me alone, ANd she changes my tone.  She makes me happy, What I thought wasn\'t a possibility, Became reality.  And here I am, crying, Into the crying pool, I\'m such a fool, I\'m lying, Waiting, Missing her.   If she read this she\'d blush too, But it\'s the least I could do. She saved me.  But it doesnt matter what I think of her, Life\'s just a murmur, Without her.  I learned what she though of me, Recently, And to be honest, I\'ve lost count of the nights I couldn\'t sleep easy.  She keeps me awake, When she\'s not even there, Makes me feel at my best, In moments of despair.     I just wish I made her happy.     Crap poem I know. '),(2054,3782048,'07,July,2004','2004-07-07 00:00:00','   I guess I am going to hospital then. Doctor says if the meds I\'m on now don\'t work, he\'ll have to refer me to hospital. So far they aren\'t working, and they\'re ny 5th prescription for my stomach.   To be honest I have it all visualled out. Me in hospital, with only 2 or 3 visitors. My mum and dad arguing, and Mona, just talking like the lovely gal that she is. But that\'s it. Don\'t imagine anyone else coming to see me. I have a fear that if I eevr need an operation, I may wake up during it. Scary idea.    Anyhow, 24 days left till I find out if I need hospital treatment or not. :( '),(2055,3782048,'07,July,2004','2004-07-07 00:00:00','   I am off Iceland tomorrow.  Scary stuff.   Why am I scared? Because I feel insecure about things right now in my life. And I\'m just worried if I leave them for a short while, they\'ll collapse. I also have to sort a few things out with people. NOt looking forward to that.   Will anyone miss me? I don\'t really know. I\'m not going to pretend I know how everyone feels and say no one will miss me, but I don\'t expect it much. I read Andy\'s messages today (at his will of course lol) and I guess, I was kinda...shite. All this time I\'d been thinking whether or not I should text Charlie or Ellen. I left it at the end because I thougght they didn;t want a fool like em bothering them. THey probably  didn\'t, I don\'t know. It\'s so odd, I was there when Charlie text Andy telling him she\'d be home. I don\'t know why but I was expecting a text any minute, from her, telling me she\'d be coming back. Not in the same words coz she doesn\'t love me, but I dunno.  Apparently Charlie\'s pissed me off. I don\'t know :S . '),(2056,3782048,'07,July,2004','2004-07-07 00:00:00','   This is weird. I don\'t liek the idea of talking to myself but then again I\'m currently on my 14th journal....so....yeah.   I\'m scaring myself. Lately my poetry has been Kurt Cobain Stylee. He was a great guy, and he mentioned in his journal something like                                                  \"all my lyrics are a bunch of contradictions\"  I don\'t want that to happen to my poetry but it may happen. He also didn\'t make much sense in songs because he just focused on rhyming. ONce again I don\'t like that. Scary.  I am going to continue writing poems although I am getting a book published. The final number of poems is 177. I will be cutting this number down to 100-150. This is odd. I didn\'t want a publisher who would be selling a book. I just wanted a publisher who wouldn\'t ask for much money, who\'d just make a copy for me. But now I have an author\'s contract to sign, 8% royalties (not bad considering I didn\'t want anything), two free copies of the book, and, well, an achievement. Getting a book published at the age of 16. Wow. I am currently writing a poem called Perfect Girl. Yeah it\'s a poem about charlie but I don\'t really care what people think. There is nothing going on  between her. I was going to text her sayign I missed her, but I thought nah. Ellen would see it, and probably tell Andy. Then everyone would think I fancied her. CHarlie said a while ago about when I say I love her, I immediately say something funny afterwards, as if I\'m ashamed of telling her I love her. Same goes for the text. I\'d text her just sayng \"I miss you\" but I\'d be scared everyone would take it the wrong way. If I said \"I miss you *sob* lol\", then there\'s humour, and I feel safe. But love is no joke.  I truly do miss her. She has changed the world for me. Or changed me for the world. I don\'t know.  Pills. All I can say is they suck. After my 4th prescription, the Doctor prescribed me a similar drug, but with less side effects.   It\'s kinda working. I took my first one yesterday at 9am-ish. All went well till about an hour, then i got pains for about 3 hours non stop. Had a break from them for a few hours, they came back.  I woke up today (WELL DUH! lol) and, um, I missed the dose by about 30 minutes. On any previous prescription I wouldn\'t be in much pain by then, only 30 minutes late. But all I got today was a sharp pain which lasted even after I took the pill. Just for being 30 minutes late.   The sleeping pills don\'t work. All I can say on the matter.  Am I becoming dependant on meds?  Peace for now, IlovekurtyIlovecharlieIloveandyIlovemona '),(2057,3782048,'03,August,2004','2004-08-03 00:00:00','   I love making songs on la guitar. Calms me down, and I cna also be proud for a few seconds. I know it\'s not long, but still it\'s cool. To anyone a song i make up is nothing but to me, it\'s like a huge collection of emotions. My psychatrist reckons my emotions control me too much. When I play a home made song, lol, it\'s like I control them, like everyone else does.  I\'m proud. :D '),(2058,3782048,'02,August,2004','2004-08-02 00:00:00','   I love my new unoffical random  job.  Seriously, it\'s great. I don\'t mind working all day. Just incase readders think I am being sarcastc I am not. I have to wait up at 7:30 tomorrow, go to Harmony Music, move some piano\'s, then I work all day. For some reason I love it. Great people to work with, and I do everything. The till, the card machine (ugh...just gotten used to it), doing demo\'s, explaining stuff, and telling people how old Paul is.   But tis cool. I also plan to wake up early 2moro, maybe 5 or 6, have a shower, go out with Black Saphire (can\'t get a girl, get a guitar :S), go somewhere non-urba, and play for a bit. Then go to work. The morning is beautiful. '),(2059,3782048,'02,August,2004','2004-08-02 00:00:00','   She\'s my friend, And I can\'t pretend, That I don\'t want anymore, Than her friendship.  But I must lie to her now, I wish somehow, I could tell her I love her, But I can\'t. She has the life, A perfect life too, I\'d just ruin it, Something I always seem to do. These two were made for each other, The perfect lovers, And I\'m trying to fit in somewhere, But I haven\'t realised, Lovers don\'t come in threes, but pairs, Maybe I should keep this rule memorised. There\'s no point in trying, She\'s with someone, And there\'s no point in lying, Her love for him is never gone, And who am I to get in the way? She tells me she loves him almost everyday. She\'s defined her love for him, And maybe I\'m just too dim, To realise, she\'s happy, She doesn\'t want me. It\'s nothing new. Maybe I should take some good advice, Something I\'ve heard more than twice, I should just quit while I\'m ahead, Continue this and I\'ll be dead. '),(2060,3694314,'30,June,2004','2004-06-30 00:00:00','   I\'m not an extreemly religious person. Call me curious. But i have to thank Juilian for recomending this \"forgotten Gospel\" to me. These are the secret sayings that the living Jesus spoke and Didymos Judas Thomas recorded.  1 And he said, \"Whoever discovers the interpretation of these sayings will not taste death.\"  2 Jesus said, \"Those who seek should not stop seeking until they find. When they find, they will be disturbed. When they are disturbed, they will marvel, and will reign over all. [And after they have reigned they will rest.]\"  3 Jesus said, \"If your leaders say to you, \'Look, the (Father\'s) kingdom is in the sky,\' then the birds of the sky will precede you. If they say to you, \'It is in the sea,\' then the fish will precede you. Rather, the kingdom is within you and it is outside you.  When you know yourselves, then you will be known, and you will understand that you are children of the living Father. But if you do not know yourselves, then you live in poverty, and you are the poverty.\"  4 Jesus said, \"The person old in days won\'t hesitate to ask a little child seven days old about the place of life, and that person will live.  For many of the first will be last, and will become a single one.\"  5 Jesus said, \"Know what is in front of your face, and what is hidden from you will be disclosed to you.  For there is nothing hidden that will not be revealed. [And there is nothing buried that will not be raised.\"]  6 His disciples asked him and said to him, \"Do you want us to fast? How should we pray? Should we give to charity? What diet should we observe?\"  Jesus said, \"Don\'t lie, and don\'t do what you hate, because all things are disclosed before heaven. After all, there is nothing hidden that will not be revealed, and there is nothing covered up that will remain undisclosed.\"  7 Jesus said, \"Lucky is the lion that the human will eat, so that the lion becomes human. And foul is the human that the lion will eat, and the lion still will become human.\"  8 And he said, The person is like a wise fisherman who cast his net into the sea and drew it up from the sea full of little fish. Among them the wise fisherman discovered a fine large fish. He threw all the little fish back into the sea, and easily chose the large fish. Anyone here with two good ears had better listen!  9 Jesus said, Look, the sower went out, took a handful (of seeds), and scattered (them). Some fell on the road, and the birds came and gathered them. Others fell on rock, and they didn\'t take root in the soil and didn\'t produce heads of grain. Others fell on thorns, and they choked the seeds and worms ate them. And others fell on good soil, and it produced a good crop: it yielded sixty per measure and one hundred twenty per measure.  10 Jesus said, \"I have cast fire upon the world, and look, I\'m guarding it until it blazes.\"  11 Jesus said, \"This heaven will pass away, and the one above it will pass away.  The dead are not alive, and the living will not die. During the days when you ate what is dead, you made it come alive. When you are in the light, what will you do? On the day when you were one, you became two. But when you become two, what will you do?\"  12 The disciples said to Jesus, \"We know that you are going to leave us. Who will be our leader?\"  Jesus said to them, \"No matter where you are you are to go to James the Just, for whose sake heaven and earth came into being.\"  13 Jesus said to his disciples, \"Compare me to something and tell me what I am like.\"  Simon Peter said to him, \"You are like a just messenger.\"  Matthew said to him, \"You are like a wise philosopher.\"  Thomas said to him, \"Teacher, my mouth is utterly unable to say what you are like.\"  Jesus said, \"I am not your teacher. Because you have drunk, you have become intoxicated from the bubbling spring that I have tended.\"  And he took him, and withdrew, and spoke three sayings to him. When Thomas came back to his friends they asked him, \"What did Jesus say to you?\"  Thomas said to them, \"If I tell you one of the sayings he spoke to me, you will pick up rocks and stone me, and fire will come from the rocks and devour you.\"  14 Jesus said to them, \"If you fast, you will bring sin upon yourselves, and if you pray, you will be condemned, and if you give to charity, you will harm your spirits.  When you go into any region and walk about in the countryside, when people take you in, eat what they serve you and heal the sick among them.  After all, what goes into your mouth will not defile you; rather, it\'s what comes out of your mouth that will defile you.\"  15 Jesus said, \"When you see one who was not born of woman, fall on your faces and worship. That one is your Father.\"  16 Jesus said, \"Perhaps people think that I have come to casy peace upon the world. They do not know that I have come to cast conflicts upon the earth: fire, sword, war.  For there will be five in a house: there\'ll be three against two and two against three, father against son and son against father, and they will stand alone.  17 Jesus said, \"I will give you what no eye has seen, what no ear has heard, what no hand has touched, what has not arisen in the human heart.\"  18 The disciples said to Jesus, \"Tell us, how will our end come?\"  Jesus said, \"Have you found the beginning, then, that you are looking for the end? You see, the end will be where the beginning is.  Congratulations to the one who stands at the beginning: that one will know the end and will not taste death.\"  19 Jesus said, \"Congratulations to the one who came into being before coming into being.  If you become my disciples and pay attention to my sayings, these stones will serve you.  For there are five trees in Paradise for you; they do not change, summer or winter, and their leaves do not fall. Whoever knows them will not taste death.\"  20 The disciples said to Jesus, \"Tell us what Heaven\'s kingdom is like.\"  He said to them, It\'s like a mustard seed, the smallest of all seeds, but when it falls on prepared soil, it produces a large plant and becomes a shelter for birds of the sky.  21 Mary said to Jesus, \"What are your disciples like?\"  He said, They are like little children living in a field that is not theirs. when the owners of the field come, they will say, \"Give us back our field.\" They take off their clothes in front of them in order to give it back to them, and they return their field to them.  For this reason I say, if the owners of a house know that a thief is coming, they will be on guard before the thief arrives and will not let the thief break into their house (their domain) and steal their possessions.  As for you, then, be on guard against the world. Prepare yourselves with great strength, so the robbers can\'t find a way to get to you, for the trouble you expect will come.  Let there be among you a person who understands.  When the crop ripened, he came quickly carrying a sickle and harvested it. Anyone here with two good ears had better listen!  22 Jesus saw some babies nursing. He said to his disciples, \"These nursing babies are like those who enter the kingdom.\"  They said to him, \"Then shall we enter the kingdom as babies?\"  Jesus said to them, \"When you make the two into one, and when you make the inner like the outer and the outer like the inner, and the upper like the lower, and when you make male and female into a single one, so that the male will not be male nor the female be female, when you make eyes in place of an eye, a hand in place of a hand, a foot in place of a foot, an image in place of an image, then you will enter [the kingdom].\"  \"It will not come by watching for it. It will not be said, \'Look, here!\' or \'Look, there!\' Rather, the Father\'s kingdom is spread out upon the earth, and people don\'t see it.\"   I know thats long and theirs more if you want to read it fully just type in the gospel of thomas in yahoo,ask.com,goggle,or which ever you prefer. Once again thanx Juilian even though your probably not reading this. Later Dayz  Be Safe DarkPoet Danielle Martin DM Peace '),(2061,3694314,'29,June,2004','2004-06-29 00:00:00','   \"Facts are twisted,  illusions fostered,  truth destroyed to prove that the perverted is normal,  the sordid noble,  the brutal beautiful,  the guilty innocent,  the coward a hero,  disaster a victory-and the reverse of all these things. \"(Pgs 277-278). The books called \"Transvaal Episode\", by Harry Bloom, and its fictional. Although I have not read the book Its on my to do list You should pick it up  If you want Later dayz Ciao Bella Dark88Poet Danielle DM '),(2062,3694314,'27,June,2004','2004-06-27 00:00:00','   hey guys sup.  sorry i haven\'t posted in a while.  I had computer problems. Anyway, since i\'ve been gone i\'ve been workin on some new and old music.  i just made a song  from one of my poems Its called  City Of Broken Dreams . I  won a contest a while back with it and i really love the song.  i\'ve also been searchin the web for vocal techniques and song/poetry contest\'s.  any ideas and or links are welcomed and very much apreciated.  Later Dayz Be safe '),(2063,3694314,'20,June,2004','2004-06-20 00:00:00','   Just wanted to say hi to my friend, Kathleen.  She rocks and is an awsome writer.  Hey girl. Remember what I said in the yearbook cause i meant it. Luv ya Dark88Poet Danielle Martin  DM '),(2064,3694314,'20,June,2004','2004-06-20 00:00:00','   for some godforsaken reason i can\'t find my blog. i can still post and edit my profile but i don\'t see what u\'reeing. im at blogger.com. helllp '),(2065,3694314,'20,June,2004','2004-06-20 00:00:00','   Hello i am the dark poet a.k.a danielle.  This is my first time having a blog.  I thought i\'d try something new.  hope you like it.  If u don\'t o\' well get out.  Later Dayz  Dark88Poet DM '),(2066,3694314,'31,July,2004','2004-07-31 00:00:00','   My mercy has limits  You\'re still my enemy tonight, but even enemies can show respect  Honor the gods, love your woman and defend your country. Before my time is done I will look down on your corpse and smile    *My Three Favorites     Before you came to Sparta, I was a ghost. I walked and I ate and I swam in the sea... I was just a ghost.    Men will hunt us, the gods will curse us, but I\'ll love you. Until the day they burn my body, I\'ll love you.    If killing is your only talent, that\'s your curse. I don\'t want anyone dying for me.   May the Gods keep the wolves in the hills and the women in our beds  Men are haunted by the vastness of eternity. and so we ask ourselves: will our actions echo across centuries?  I can\'t imagine life without you '),(2067,3694314,'30,July,2004','2004-07-30 00:00:00','   urlLink http://vintagetears.proboards31.com/   urlLink http://www.jasonmraz.com/journal.html   urlLink http://www.fictionpress.com/~theblackenedheartofmalice  http:// urlLink ww urlLink w.soundclick.com/bands/2/brittneywalters.htm    urlLink http://www.emmamusicnj.com/home.htm  ( keep A LOOK OUT ON THESE GUYS THEYRE HUGELY TALENTED, its just a matter of time)  urlLink http://www.gurl.com/play/games/pages/0,,626950,00.html   urlLink http://www.gurl.com/findout/guides/pages/0,,625067,00.html   urlLink http://www.gurl.com/findout/guides/pages/0,,626692,00.htm  urlLink    urlLink http://p204.ezboard.com/bshadowsofthenight.showUserPublicProfile?gid=iamnumb   urlLink http://inocencia.4.forumer.com/index.php ?  urlLink http://www.geocities.com/mi0tic/   urlLink http://www.darkness-embraced.com/   urlLink http://www.majandra.com/   urlLink http://www.elitelit.com/ '),(2068,3694314,'30,July,2004','2004-07-30 00:00:00','   MOod  : lonley    Current Music  : In the shadows by the Rasmus, Somewhere Only We Know by Keane    Book  : mY NEW Record, i read a whole trilogy in three weeks.  The Sea Dragon Air Trilogy  By Storm Constantine (great Reads) '),(2069,3694314,'30,July,2004','2004-07-30 00:00:00','   Ok I promise you no more quizess. that was the last one for quite a while. i just wanted to find out who i was. weird i know, but whatever. No more school . I am free for a month then back to torture and ridicule. yeaaaah sarcasm is a great thing  Later dAYS Dm DarkPoet Be Safe '),(2070,3694314,'30,July,2004','2004-07-30 00:00:00','   schizoid   urlLink  Which Personality Disorder Do You Have?    brought to you by '),(2071,3694314,'30,July,2004','2004-07-30 00:00:00','    '),(2072,3694314,'30,July,2004','2004-07-30 00:00:00','   The Tradegy...   urlLink  Whats your story?    brought to you by '),(2073,3694314,'30,July,2004','2004-07-30 00:00:00','   Your an ice spirit. Your very deep in your own ways, and very intellectual. Although some people may define you as crazy. You usually like to be alone and probably only have a few close freinds.   urlLink  What kind of myth should you be?(for girls only)    brought to you by '),(2074,3694314,'30,July,2004','2004-07-30 00:00:00','   Dream, the third of The Endless, you are in charge of the Dreaming, all imagination and creativity, everyone knows your beautiful realm, but none truly understand it. You are dark and brooding, creative, and spend a lot of time by yourself, just thinking. You are almost as serious as Destiny, but not quite. Everyone is enchanted by you, but you keep them all at a distance, even when you shouldn\'t.   urlLink  Which Endless are you?    brought to you by '),(2075,3694314,'30,July,2004','2004-07-30 00:00:00','   ~Imaginary~ You\'re creative and often daydream. You build your own little world whether it\'s a bad or a good day, you can always escape to. You believe dreams are the best thing reality has to offer. You prefer the art of illusion over what drones have made the real world out to be. Life can be tough and you know this. Somedays you hate the world and somedays you can tolerate it. You love imagination and the mind, for this is where you can mold your idea of how things should be. \"In my field of paper flowers and candy clouds of lullaby. I lie inside myself for hours and watch my purple sky fly over me.\" *Please Rate*   urlLink  Which Evanescence \'Origin\' Song Are You? (With Pics!)    brought to you by '),(2076,3694314,'30,July,2004','2004-07-30 00:00:00','   \'Dreams are the best thing reality has to offer\' \'A positive attitude will not solve all your problems, but it will annoy enough people to make it worth the effort.\' \'I always knew I would look back on my crying and laugh, but I never thought I would look back on my laughter and cry.\' \'Love isn\'t finding the perfect person, it\'s finding the imperfect person perfect\' \'People are like stained glass windows: they sparkle and shine when the sun is out, but when the darkness sets in their true beauty is revealed only if there is a light within\' \'The rain of my soul are the thoughts that drown within me, tears are the raindrops to let the world see.\' \' Sticks and Stones are hard on bones, aimed with angry art. Words can sting like anything, but silence breaks the heart.\' \'A brighter light casts a darker shadow\' \'A girl and a guy can be just friends, but at one point or another they will fall for each other. Maybe temporarily, maybe at the wrong time, maybe too late, or maybe forever...\' \'Instead of complaining that rose bushes have thorns, rejoice that thorn bushes bloom\' '),(2077,3694314,'30,July,2004','2004-07-30 00:00:00','   Night   urlLink  ?? Which Angel Or Demon Are You ??    brought to you by '),(2078,3694314,'30,July,2004','2004-07-30 00:00:00','   Seahorse   urlLink  ?? Which Creature Of The Sea Are You??    brought to you by '),(2079,3694314,'30,July,2004','2004-07-30 00:00:00','   Hecate   urlLink  ?? Which Of The Greek Gods Are You ??    brought to you by '),(2080,3694314,'30,July,2004','2004-07-30 00:00:00','   Lost Emily   urlLink  Which Emily Strange are you?    brought to you by '),(2081,3694314,'30,July,2004','2004-07-30 00:00:00','   `Broken Angel`- Once an Angel of Light. You were hurt beyond repare. You drown in your sadness, sinking deeper with each passing day. People pity you, they see your sadness as beauty and all want to save you. People try and pull you from your misery but each time you slip back. You know the only one who can save you is you, you just cant bring yourself to do it. You dwell in the darkness all the while giving off a faint light from your Angelic nature. People are drawn to your misery.   urlLink  good......dark`~Angelz~`light......evil    brought to you by '),(2082,3694314,'30,July,2004','2004-07-30 00:00:00','   Wow......You really know your\'re evanescence!I bet you have the cd and no matter how many times you listen to it.......You never get tired of it! Rock On!   urlLink  Evanessence Lyrics..........Fallen    brought to you by '),(2083,3694314,'30,July,2004','2004-07-30 00:00:00','   Cancer is symbolized by the Crab. You\'re deeply emotional and intuitive, and your top concern is family. Something of a homebody, you prefer to stay in rather than go out on the town. You can be moody, drawing into your protective shell if things get too intense around you.   Element: Water      The Element of Water is associated with the Signs   urlLink  Cancer   ,   urlLink  Scorpio    and   urlLink  Pisces   , and it also rules the   urlLink  Fourth   ,   urlLink  Eighth    and   urlLink  Twelfth    Houses. The watery words flow with this Element: fluid, flowing, wavering. These terms can easily apply to an individual\'s emotions, which are exactly the realm in which Water exerts its influence. Water Signs are intuitive, sensitive and feel more intensely than the rest. They are emotional and nurturing and, like a river, run deep. How things feel is what matters to these folks, and they base their actions on sense, rather than on logic or intellect. Water is also about compassion and understanding, since the receptive talents of these individuals are remarkably high. Water Sign people can take in the feelings of others, process them, and put forth a plan or solution that will be conscientious to all. Those graced by Water love to take their sense of the world and translate it into the artistic. They desire aesthetic beauty -- and for everyone to be happy as a result of that beauty. Much like still water can become stagnant, however, an inactive Water Sign person is not a Water Sign person at their best. Water Signs feel most fulfilled when they are helping others, and they do so in an enchanting, considerate and even romantic way. The flip side of this dreamy Water world is the tendency of these individuals to brood. Water Signs are susceptible to mood swings, and, at their worst, they could become self-indulgent, controlling and hostage to a fantasy world. This also brings up the ability of these folks to see things more clearly than others. Call it psychic, if you will. Water Signs are emotional, empathetic, receptive and feel things deeply. Alternately calm like the sea or possessing the force of a torrential rain, the emotions brought forth by this Element are plentiful, indeed.    Quality: Cardinal      A Cardinal Quality is attached to the Signs   urlLink  Aries   ,   urlLink  Cancer   ,   urlLink  Libra    and   urlLink  Capricorn   . The Cardinal Houses are the   urlLink  First   ,   urlLink  Fourth   ,   urlLink  Seventh    and   urlLink  Tenth   ; if an astrological chart shows a preponderance of Planets in these Houses, its energy is weighted toward cardinality. Cardinal Signs are the initiators of the Zodiac. They are also found at key jumping-off points on the chart wheel, specifically the   urlLink  Ascendant   , Medium Coeli (M.C. or Midheaven), Descendant and Imum Coeli (I.C.). Individuals possessing a Cardinal Quality like to get things going. They are active, quick and ambitious. Many projects get started, thanks to Cardinal initiative, although a good deal of them are never finished. That\'s because Cardinal folks are much fonder of starting things than finishing them. You won\'t find a Cardinal person slacking off. These people are full of vim and vigor and possess a drive and ambition that is unmistakable. Enthusiasm and a zest for life fill the Cardinal individual. Some might perceive this rampant energy as domineering, and, at times, it can be. Cardinal people can easily forget about the rest of the pack when they are busily focusing on their own endeavors. Even so, their energetic spirit often wins the day. Cardinal folks are clever and want to win. They love to start things, and whether they finish them or not, there\'s always a lot going on. Nay sayers who find them to be too self-centered will simply have to watch (and marvel) as they speed by!   Ruling House:Fourth      House of Family and HomeThe Fourth House is commonly referred to as the House of Home. When we think of home, we think of that place where we put down our roots. We lay our foundation and plant ourselves firmly into the Earth, as it were. One day, we will return to that very same Earth. The Fourth House brings things full circle by also addressing old age, endings and our final resting place. Much of the emphasis of the Fourth House, however, is on the concept of home. By laying down roots, we make a home for ourselves, or more specifically, the self. It\'s worth noting that in addition to the external home (all the bricks and mortar around us), we have really brought the essential self home. \'I\'m home.\' The words themselves have a peaceful ring to them. The self is now centered, grounded, one and at peace with the Earth. We seek to come home both physically and psychically, for ourselves and for those we love. By creating home, we create a meeting place, a sanctuary, a sacred place for ourselves and for others. In our home, we integrate the self with all that has come before us and helped to shape what we are today. We create a domestic space which comforts and nurtures us and serves to keep safe those we love. Also important here are family history, cultural and societal norms, and ways of being. All of these are ruled by the Fourth House, as are our ancestry, roots and heritage. These qualities are brought \'home\' through us and integrated into the place we call home. Helping to create the home we make are our parents, for it is they who greatly nurture and shape our being. Therefore, the Fourth House can also be thought of as the House of the Mother, the Parent or the Nurturer. Looking at things from a strictly tangible point of view, we can see that the Fourth House also encompasses physical structures (houses) and real estate. The Fourth House represents family, history and traditions. All of these contribute to the process of becoming a true, actualized and individualized self. This is how we come home. The Fourth House is ruled by the Sign Cancer and the Moon.   Ruling Planets: The Moon     The Moon: Ruler of the Emotions Since the Moon is the closest \'Planet\' to Earth, this satellite literally zips around the Zodiac, completing its circuit in less than a month. It also touches us more deeply than most Planets. As the ruler of the tides, it is fitting that the Moon should be the ruler of our emotions. Still waters run deep? Making waves? A wellspring of emotion? Yes, our emotions have long been portrayed in terms of the sea: fluid, momentous, churning from within. Mood swings, instinct, how we feel about things and how our feelings affect others are all influenced by the Moon. Whereas the Sun gives us our spirit, it\'s the Moon which gives us our soul. The Moon is goddess-like in that it symbolizes mother and the relationship between woman and child. This Planet (also known as a luminary) speaks to the women in one\'s life and their role as nurturer. Fertility, pregnancy and childbirth are also governed by the Moon. We see the Moon casting its silvery glow from our earliest moments, when we were stroked and caressed by our mother and felt her tender touch. Our emotions manifest themselves through our being and set the tone for our daily lives. The Moon is party to this continuum, rendering us vital one moment and fragile the next. We\'re up and down, naughty or nice, and may laugh wildly or cry at will. Through the Moon\'s energy, we endeavor to reconcile these varied emotions in order to make ourselves complete and one with the world. The Moon also helps us see that which we want, and to use memory and the past as part of this process. The Moon spends roughly 2 1/2 days in each sign and takes 28 days to circumnavigate the Zodiac. It is feminine energy and rules   urlLink  Cancer    and the   urlLink  Fourth House   .        Symbol: the Crab Ruling Planet: the Moon Ruling House: Fourth House Element: Water Quality: Cardinal Body Parts: breasts, digestion Keyword: HOME Date with destiny: Scorpio, Pisces Run for the hills: Aries, Libra Where you glow: hosting a party What makes you tick: nostalgia Fitness forecast: swimming Play date: redecorating Perfect jobs: chef, nursery school teacher Best accessory: hand-knitted scarf A sure thing: best dish at the potluck Destination: Orlando Pleasure: baking, comfort, children Pain: snide remarks, forgetfulness, crab jokes What\'s my line? Home is where the heart is. '),(2084,3694314,'30,July,2004','2004-07-30 00:00:00','   One, two, Freddyâ€™s coming for you, three, four, you better lock your doors, five, six, grab your crucifix, seven, eight, you better stay up late, nine, ten, Freddyâ€™s back again. '),(2085,3694314,'27,July,2004','2004-07-27 00:00:00','   baise, merde, rien, chienneâ€™ is French. â€˜Bumsen, ScheiÃŸe, Fluch, Weibchen.â€™ German. â€˜cogida, mierda, maldiciÃ³n, perra.â€™ Spanish. â€˜scopata, merda, maledizione, femmina.â€™ Italian. â€˜foda, merda, nada, cadela.â€™ Portuguese. '),(2086,3694314,'26,July,2004','2004-07-26 00:00:00','   Little Bunny Foo Foo, skipping on the highway, along came a semi-truck and OOPS! Bunny\'s dead '),(2087,3694314,'26,July,2004','2004-07-26 00:00:00','   I\'m tired of being sad but i don\'t no how to stop '),(2088,3694314,'26,July,2004','2004-07-26 00:00:00','   When you walk to the edge of all the light you have and take that first step into the darkness of the unknown you must believe one of two things will happen:  There will be something solid for you to stand upon, or, you will be taught how to fly.  -Patrick Overton '),(2089,3694314,'26,July,2004','2004-07-26 00:00:00','   I\'m so happy. school was as depressing as usual but of course i don\'t expect that to change anytime soon. that would be like me sprouting wings and those idiots at school aquiring brains, but aywhoo thats not what i\'m happy about. i finished 2 songs today and Im starting another one. golly i\'m on a bagle today call me cream cheese.(yum)  What i\"m Listening to right now:   Clocks by coldplay   What i\'m planning on listening to afterwards:   Lithium by Nirvana       &nbsp; Anyways  Quote of the Day:&nbsp;    A thing of beauty is a joy forever: Its loveliness increases; it will never Pass into nothingness; but still will keep A bower quiet for us, and a sleep Full of sweet dreams, and health, and quiet breathing.  John Keats       &nbsp; DarkPoet Dm Later Days  Bon Voyage Arivada Derchi   &nbsp;'),(2090,3694314,'26,July,2004','2004-07-26 00:00:00','   You are Xuan Wu!  Mythological background: Because the turtle has a thick, solid shell that serves as protection -  this animal is associated with stability. You enjoy intellectual pursuits.&nbsp;Also, in Feng Shui (the Chinese myths behind choosing a house), the black turtle\'s solidity is used to protect from cold northern winds.   urlLink  Which Chinese Mythological&nbsp; Being&nbsp; Are You?     brought to you by '),(2091,3694314,'26,July,2004','2004-07-26 00:00:00','   Modern rock!&nbsp; You\'re very good!&nbsp; Your lyrics have lots to say, but you can go crazy sometimes... Careful now...&nbsp; Just keep making that music that keeps the rock world watching!   urlLink  What genre of rock are you?    brought to you by '),(2092,3694314,'26,July,2004','2004-07-26 00:00:00','   You have an angry soul! Angry Souls arent always angry, but they cannot easily forgive and hold grudges. You probably often get in fights with your friends and family, and its difficult for you to understand. When someone makes a mistake, you dont let go easily and hold on to those memories. Your very stubborn and your rage is known to everyone. Though you never actually mean it, you can say mean things in a fight and go over board. Many people are sometimes intimidated by your anger. But you have many redeeming qualities and those are that you are quite intelligent and smart. You would make a good businesswoman or lawyer because you know how to prove your point. You cherish the ones around you, and appreciate life, even though you can complain or throw a tantrum now and then. The good things is, you keep your emotions very outspoken, and are normally a very happy person because all your rage is let on the outside. Anger is simply a state, but you, yourself as a person, are great.   urlLink  What Kind of SOUL do you posses? (For Girls only) Incredible Anime Pictures!    brought to you by '),(2093,3694314,'26,July,2004','2004-07-26 00:00:00','   Your Heart is Grey   urlLink  What Color is Your Heart?     brought to you by '),(2094,3694314,'26,July,2004','2004-07-26 00:00:00','   You can give her a rose every single day of her life, give her the most beautiful dresses in the world, shower her with kisses, read poems, and she\'d greatly smile at you. But at the end of the day, nothing really matters, because at the end of the day she\'s all alone. '),(2095,3694314,'24,July,2004','2004-07-24 00:00:00','   You are the spirit of wisdom. You are wise beyond your years and know what others don\'t. You are what others need in a freind. You might appear shy but are strong.   urlLink  What kind of element fey are you? (PRETTY PICS)    brought to you by '),(2096,3694314,'24,July,2004','2004-07-24 00:00:00','   what your listening to right now: &nbsp;Lover I don\'t have to love by Bright Eyes '),(2097,3694314,'24,July,2004','2004-07-24 00:00:00','   Lonliness dominates you. You can hide it well, but its there, and your friends can see it. You constantly feel alone, and need to do things to fill your time. Your afraid to tell people this, but sooner or later it gets out in a bad way, and you think you screwed up everything. And when you are in love is when you are sad the most. (Please Vote)   urlLink  What Emotion Dominates you?     brought to you by '),(2098,3694314,'24,July,2004','2004-07-24 00:00:00','   Nihilist Bear   urlLink  Which Dysfunctional Care Bear Are You?    brought to you by '),(2099,3694314,'24,July,2004','2004-07-24 00:00:00','   I hate&nbsp;school with a major passion. it is an unrealistic attempt at making humans out of idiotic baffoons. (not u goys of course).  This asswipe at my school called me a rasist. i\'m like yeah i\'m black and i\'m racist. people ar constantly treating me like shit and the teacher doesn\'t even bother to correct them, but i guess i should be used to it. its been going on for more than half my life. i mean last year this motherfucking asswhole drug dealin little imbicile called me a nigger and all my principle did was tell his parents. God i just wish that i could move to london where nobody knows me and write music and poetry music in some obscure little coffe shop.   i never knew how dramatic i was until read over that paragraph. but seriously somtimes i can\'t take my life i get treated like insignifigant&nbsp;little pion&nbsp;at school and like i\'m shit at home. 2 more years then i\'m free from all this drama and hurt and pain and blah.   its exaughsting being so gosh darn depressed. i wish i were happy,   &nbsp;  any way my quote for the day is  Life is like a box a chocolates you never no which one your gonna get(forrest gump, which i have not seen)  no seriously its:   Overcome all boundries to become the person you know you can be. Its all up to u don\'t let jealousy, pain, or lazyness bring you down. In the end its all worth it. (Meeeeee!)   Later dAYS&nbsp;  darkpoet  dm  danielle '),(2100,3694314,'24,July,2004','2004-07-24 00:00:00','   &nbsp;-Deprive me of everything...It is you who cannot deprive yourself from your own humility-~Layla Greenleaf~- He who breaks a thing to find out what it is, has left the path of wisdom-~Gandalf~  -Few can foresee whither their road will lead them, till they come to its end-~Legolas~ -Faithless is he that says farewell when the road darkens-~J.R.R. Tolkien~ -The measure of a person\'s life is the well spending of it, not the length- ~Plutarch~ -First, learn the meaning of what you say, and then speak- ~Epictetus~ -The most wasted of all days is one without laughter- ~E.E. Cummings~ -Life\'s greatest happiness is to be convinced we are loved- ~Victor Hugo~ -It is the mysterious equations of love that any logical equation can be found- ~John Nash~-Love is not blind - it sees more, not less. But because it sees more, it is willing to see less- ~Rabbi Julius Gordon~ -When a woman is talking to you, listen to what she says with her eyes- ~Victor Hugo~ -Gossip is when you hear something you like about someone you don\'t- ~Earl Wilson~-Happiness is as a butterfly which, when pursued, is always beyond our grasp, but which if you will sit down quietly, may alight upon you- ~Nathaniel Hawthorne~ -It is never wise to seek or wish for another\'s misfortune. If malice or envy were tangible and had a shape, it would be the shape of a boomerang- ~Charley Reese~ -Everything that is faced will not change, but nothing will change unless it is faced- ~James Baldwin~ -One man practicing sportsmanship is far better than 50 preaching it- ~Knute K. Rockne~- Your worst humiliation is only someone else\'s momentary entertainment- ~Karen Crockett~ -I don\'t know half of you half as well as I should like; and I like less than half of you half as well as you deserve- ~J. R. R. Tolkien~ -The best things in life are not things- ~Anon~ -We can easily forgive a child who is afraid of the dark; the real tragedy of life is when men are afraid of the light-  ~Plato~-Education is not received. It is achieved- ~Unknown~ -Do not let what you cannot do interfere with what you can do- ~John Wooden~ -Knowledge speaks, but wisdom listens- ~Jimi Hendrix~  -Success usually comes to those who are too busy to be looking for it- ~Thoreau~ -Love is an exploding cigar we willingly smoke-  ~Lynda Barry~ &nbsp;-The meeting of two personalities is like the contact of two chemical substances: if there is any reaction, both are transformed- ~Carl Jung~  -A real friend is one who walks in when the rest of the world walks out- ~Unknown~  -Everyone hears what you say. Friends listen to what you say. Best friends listen to what you don\'t say- ~Unknown~  -Have a heart that never hardens, and a temper that never tires, and a touch that never hurts- ~Charles Dickens~  -When someone does something good, Applaud! You will make two people happy- ~Samuel Goldwyn~ -Reputation continues to be made by many acts and lost by one- ~Anon~  -A warm smile is the universal language of kindness- ~William Arthur Ward~ -The bitterest tears shed over graves are for words left unsaid and deeds left undone-  ~Harriet Beecher Stowe~  -If more of us valued food and cheer and song above hoarded gold, it would be a merrier world- ~J. R. R. Tolkien~  -All that is gold does not glitter; not all those that wander are lost-~J. R. R. Tolkien~ - One Ring to rule them all, One Ring to find them, One Ring to bring them all and in the darkness bind them- ~J. R. R. Tolkien~  -The beginning of learning is silence; then comes hearing; then writing; then success- ~Japanese Proverb~  -To be upset over what you don\'t have is to waste what you do have- ~Unknown~ &nbsp;-A great man is he who does not lose his childlike heart- ~Mencius~  -Wise men learn more from fools, than fools from the wise- ~Unkown~ -Success is not the place at which one arrives but rather...the spirit with which one undertakes and continues the journey- ~Alex Noble~  -You don\'t have to be great to start, but you have to start to be great- ~Joe Sabah~ -If there is no struggle there can be no progress- ~Frederick Douglas~  -You will never find time for anything. If you want time you must make it- ~Charles Buxton~ -The men who try to do something and fail are infinitely better than those who try nothing and succeed- ~Lloyd Jones~  -Opportunity is missed by most people because it is dressed in overalls and looks like work- ~Thomas Edison~  -It is hard to fail, but it is worse to never have tried to succeed. In this life we get nothing, save by effort- ~Teddy Roosevelt~  -If you wait until you can do everything for everybody, instead of something for somebody, you\'ll end up not doing anything for anybody- ~Malcolm Bane~  -Obstacles are those terrible things you see when you take your eyes off your goal- ~Jan Claire US Gymnastics Federation~&nbsp; '),(2101,3694314,'24,July,2004','2004-07-24 00:00:00','    '),(2102,3694314,'21,July,2004','2004-07-21 00:00:00','   Music is the universal language. It knows no boundaries, colorlines, ages, backrounds. Its forever  Unfortunatly this astoundining insight did not come from me  /teardrops are falling/  It came from this amazing JT ffsite  anyways   music is life   it is as essental to me as breathing   i cannot do without it   and i hope i will never know a day without music  for i would truly cry  that would be the day i would feel the true extent of pain    wOw talk about a teenage drama-qUeen (watch out lindsey lohan)    anyways that quote did come from me  i\'m having a fabulous weekend(music wise. i finished a song called  rise above  its kind-off soulful rock  i\'m workin on like 5 other songs(wish me luCk)  What i\'m really excited about is that i have a great idea for a screen play and i\'m workin with my friend, Jen C, in a book tentivley (sp)?) called the black fores, its based on a german myth   i\'m so excited but extremly scared i have no clue where to begin o even have to make up my own language which is so up juilien\'s alley  and on top of alllllll of that i have to finish the last 2 books of the sea dragon air trilogy by thursday or i will diiieeeee    Ok now that i\'m done spasing remember  don\'t conform to what people say or want or expect from you, remember your dreams and your goals. remember yourself cause at the end of the day thats who u have to live with.    such an Oprah moment    aaaaaannnnnnnnnywaysssss    Later dayz  (Good Luck Ashley simpson, I believe, rock on Sister, go On with your Bad self)  God Luck to all my friends  with impossibly possible dreams  Anthony M  Jen C  Gorgeous  Teresa  julien  Felicia  Me/*/  Jenna  kathleen  Shamilia  My cousins  Cory  kirsten  maryanne  maria  You guys are awsome in  All  sense of the word  Be Safe every one  Peace   DarkPoet (yeah i turned sixteen july 15) '),(2103,3694314,'21,July,2004','2004-07-21 00:00:00','    '),(2104,3694314,'19,July,2004','2004-07-19 00:00:00','   anthonywinter.com  vocalist.org*******  guitarists.net  briangilbertson.com.au  musictheory.net '),(2105,3694314,'19,July,2004','2004-07-19 00:00:00','   Hey j luv ya like a fat kid(me) luvs cake  dark88poet  be safe '),(2106,3694314,'12,July,2004','2004-07-12 00:00:00','   love is something that everyone needs.  Family supposedly has an obligation to love one another, and one might argue a wife must love her husband.   â€œYou love your family in a different way than you do your husband, because there are more layers to romantic love,â€.  â€œLoyalty, protectiveness, and affection are all present amongst family.  But with a husband, there can be stirring amounts of possessiveness, lust, and even hate.â€  Hate is a strong emotion, and though it might be considered the opposite of love, it is one of the strongest feelings a person has.  Romantic love is not always happiness.  Love can cause you to end your own life.  Love can tear friends, family, and even countries apart.  Love can bring out the absolute worst in people.â€   Even though love is fear, anxiety, jealousy, confusion, frustration, and confusion, it is strong enough to tie up every other emotion in its path,\".  â€œLove is complete and utter chaos.  It means a thousand different things to a thousand different people, but when you love someone you invest yourself emotionally in their feelings and that never changes.â€ '),(2107,3694314,'11,July,2004','2004-07-11 00:00:00','   Writing in the \"dark\" is cathartic for the mind.\"  \"I feel theres an angel inside of me that Im constantly shocking.\"  \"You laugh because Im different, I laugh because your ugly.\"  \"Better to reign in hell than to serve in heaven.\" -John Milton  \"Heresy is only another word for freedom of thought.\" -Graham Greene  \"Insane people are always sure that they are fine. It is only the sane people who are willing to admit that they are crazy.\" -Nora Ephron  \"Only love will make your cold heart melt\"  \"Remember youth as you pass by, as you was once, so once was I, as I am now, you soon will be, prepare to die and follow me.\" ~Epitaph on a warlock grave  \"Death is a mystery and burial is a secret.\" -Stephen King  \"People are like stained glass windows: they sparkle and shine when the sun is out, but when the darkness sets in their true beauty is revealed only if there is a light within.\" - Elizabeth Kubler-Ross  \"Damaged people are dangerous, they know they can survive.\" -Charles Evans Hughes  \"In the real world as in dreams, nothing is quite what it seems.\" -The Book of Counted Sorrows  \"Be different and be damned.\"  \"There\'s poetry in despair, and we sang with unrivaled beauty, bitter elegies of savagery and eloquence of blue and grey.\" -AFI [A Fire Inside]  \"When in this life you find darkness, the light of madness shines through it.\" ~Yami Bakura  \"History may be on your side, but destiny is on mine!\" ~Yami Malik   \"Death is only the beginning, a mere transgression. A life without hope, now that\'s a living hell.\" ~Me  \"Character is what you are in the dark. ~Dwight L. Moody '),(2108,3694314,'10,July,2004','2004-07-10 00:00:00','   i had the weirdest day on friday. i still don\'t know why i did what i did. i had just come from the library after summer school and was waiting or the bus as usual it was taking its own sweet time. when it finally arrived this drunk man came off the bus. he was wasted couldn\'t even stand up straight and for some reason i asked if he was ok. he said no his girl friend had just dumped him the night before. i helped him over to a chair in front of some convenience store. all alomg his arms were slashes.i was obvious he had done them to himself and for some reason i asked why he said i helps him. i asked whre he needed to go he answered 116 ridge road toi was like i\'ll help you get there. he asked me why and i said i don\'t know.   his name was chris and he was 35 years old. his mother lived in florida. ki felt so bad for him. he was drunk hadn\'t taken a bath in 3 days. and  to make matters worst the house on 116 ridgeroad where he said he was meeting his pastor was abandoned. time we had walked about 5 blocks. i told him i\'d take him to church which of course was being renovated. lucky me. so i took him to the police station. which was of course 2 more blocks away and he had to pee.   2 days later i still don\'t know why i did what i did. i don\'t know why i put up with this drunk man who was practically tryin to molest me. telling me he loved me and asking me if the reason why i wasn\'t feelin him was cuz he was white.   it was an odd xperience that i will sure as hell never forget. i feel so bad for this guy. when i took him to the station. he hit his head on the side of the door and feel on the floor. to makle mee feel even worse they said thaat they had had problems with him before and the officers seemed amazed at the fact that i did all that i had for a guy i had never met before.  so am i any way i\'m tired later dayz be safe lalala '),(2109,3694314,'07,July,2004','2004-07-07 00:00:00','   \"Consistency is the last refuge of the unimaginative.\" '),(2110,3694314,'05,July,2004','2004-07-05 00:00:00','   -- \"I don\'t beleive in racism in any way, shape, or form. I think there are idiots in every colour, race, or religion..\" -- Al Brown, Rock 95 (Radio station)  -- Join the army! Travel the world, Meet interesting people, and kill them.  -- When women are depressed they either eat or go shopping. Men invade another country.   -- You can\'t scare me. I drive a school bus!  -- My family puts the \"FUN\" back in dysFUNctional.  -- Roses are red, violets are blue, I once thought I was ugly, until I saw you!  -- We have enough youth, how about a fountain of SMART?  -- A bartender is just a pharmacist with a limited inventory.  -- Some people are only alive because it\'s illegal to kill.  -- I don\'t suffer from insanity; I enjoy every minute of it!  -- Stop Reading My Bumper Stickers and DRIVE!  -- I know Karate, Kung Fu, and 47 other dangerous words.  -- Not all men are fools...some are dead!  --No matter where you go, there you are. Huh?  --My opinions may have changed, but not the fact that I am right!  -- I\'ve gone to find myself. If I should return before I get back, please ask me to wait!  --\"Those who live by the sword get shot by those who don\'t.\".  --\"The decision is maybe and that\'s final!\"  --\"Teamwork is essential -- it allows you to blame someone else.\".  --\"Work is accomplished by those employees who have not reached their level of incompetence.\".  --\"The average woman would rather have beauty than brains, because the average man can see better than he can think.\"  -- A balanced diet is a cookie in each hand.  -- FACT: No husband has ever been shot while doing the dishes.  -- If you look like your passport picture, you probably need the trip.  -- For every action, there is an equal and opposite government program..  -- Car sickness is the feeling you get when the monthly car payment is due..  -- Families are like fudge .. mostly sweet with a few nuts..  -- Age is a very high price to pay for maturity..  -- Don\'t judge what your small mind cannot comprehend  -- The road is full of flat squirrels, who couldn\'t make up their mind!  -- My imagionary friend thinks you have serious mental problems..   -- Your jealousy is my energy, ever wonder why I am so hyper?  -- My stepdad says the only way around dealing with idiots is to become a lesbian, but I\'m not that desperate.... yet.  -- Never settle with words what you can accomplish with a flame thrower.  -- \"I can see you with a candle, burning orange lady bugs with fangs in one hand, and writing greusome deaths for people named Tessa on paper with the other hand\" -- Tessa  -- \"I\'m not scared... but my inner child has run away in terror and is threatening to jump off of the coffee maker... funny thing is, we don\'t have a coffee maker.\" -- me, on an insane day!!  -- \"Bachelorette Number One is a mentally abused shut-in from a kingdom far, far away...\" -- the magic mirror, in Shrek (You just gotta love that! Really!!)  -- \"And then people look at you weird and say \'what the hells wrong with YOU?\' and I\'ll say \'nothing\'s wrong with ME, what the hells wrong with YOU!!\" -- Jenni, who else??  -- \"If less is more, think about how much more more will be...\" Jezmine, whom has undoubtably said this with a possessed, glazed look in her eyes while discussing CHEESE with me on the phone. Oh, yes, and I enjoy spelling her name wrong.  -- \"I was trying to drive my Car! Those... Things, were in the way!\" \"Um... Those things are pedestrians...\" -- Me on a hyper day  -- \"Beauty is only skin deep... Attitude runs to the bone.\" -- I forget.  -- \"It\'s not school! Winnie the Pooh was right! It\'s SKULL!!\" -- me on an insane day.  -- \"I don\'t wanna get hit by a car! I wanna get hit by a TRUCK!!\" -- me, on (yup, you guessed it,) yet another insane day.  -- \"Never underestimate the power of stupid people, especially in large groups.\" -- everyone  -- \"Never underestimate the power of a blonde, especially if she\'s smiling..\" -- Jenni\'s version  -- \"I\'ll stab your back if you\'ll stab mine...\" -- Tessa  -- \"I saw a leprechaun! He told me to burn things!!\" -- Ralph, The Simpsons  -- \"The truth is.. I sold my soul for a truck\" -- Gun, Angel  -- \"What, your soul wasn\'t worth air conditioning?\" Fred, [Winifred] Angel (after Gun\'s revelation)  -- \"I\'m a bitch? Yeah, I know. I say it, my friends say it, my mom says it, my teachers say it. So stop wasting my time and get over yourself, okay?\" -- me, in a good mood.  -- \"i was sitting there, wondering why a frisbee gets bigger as it gets closer to you. then all of a sudden, it hit me.\" -- Jasmine (The Oreo of Love and allmighty CHEESE lord)  -- Britney Spears: Slutty bitch whore American muggle pop singer.  -- \"If practice makes perfect, and nobody\'s perfect, why practice?\" -- my little sister.  -- Like all plots to take over the world, Barbie was hatched, not born.  -- \"I\'m going to jail, lookit me!\"   -- Nobody\'s perfect, but I\'m so close, it\'s scary!  -- Oh, I\'m sorry, did I look interested?? Let me try again... -- I\'m not a bitch, I just don\'t like you  -- I do what the voices in my head tell me to.  -- I kill the ugly people! You\'re next!  -- Save a tree -- eat a beaver!  -- Karate is a form of martial arts in which people who have had years and years of training can, using only their hands and feet, make some of the worst movies in the history of the world.  -- Never knock on Death\'s door -- ring the doorbell and run. He hates that.   -- Order is for idiots, genius can handle chaos   -- My young brother asked me what happens after we die. I told him we get buried under a bunch of dirt and worms eat our bodies. I guess I should have told him the truth -- that most of us go to Hell and burn eternally -- but I didn\'t want to upset him.  -- Love is a perky little elf dancing a merry little jig and then, suddenly, he turns on you with a miniature machine gun.  -- One by One the penguins steal my sanity....  -- I plan to live forever... So far, so good.  -- One night, I was lying in bed looking up at the stars, when suddenly I said to myself, \"Hey, where the hell did my ceiling go?\"   -- I think all the stars are salt, and God\'s gonna\' eat us all...   -- Who\'s cruel idea was it for the word \"lisp\" to have a \"s\" in it?   -- We have reason to believe that man first walked upright to free his hands for masturbation.  -- Profanity is the crutch of inarticulate motherf*ckers.  -- whoever told you that your not good enough... is a fucking liar and should be struck down by a fast moving truck in some freak accident where it comes through their bedroom wall and runs them over in their own room while in the middle of having almost amazing sex   -- If you\'re naughty, go to your room, if you want to be naughty, go to mine....  -- Friends don\'t let friends drink, and bring home ugly men.  -- I like to spell marshmellow with an \"e.\" It makes them seem docile and demure, when actually they are flaming infernos of death.   -- Silence is golden...but shouting is fun!   -- Scientists say 1 out of every 4 people is crazy. Think of 3 friends. if they are okay, you\'re it.  -- An answer to that nagging question............... I let the dogs out!!   -- I know you think im cute....i kno you think im fine....but like the rest of the guyz....take a number and wait in line..   -- We are the people your parents warned you about!   -- EVIL is just LIVE spelled backwords  -- nobody\'s perfect therefore i must be a nobody  -- You always stop to stare, Why not just take a picture?Â¿?   -- I\'m not a COMPLETE IDIOT, some parts are missing  -- WHAT YOU SEE IS NEVER WHAT YOU GET.   -- Insanity is just anger put to good use!   -- Love can conquer anything - Except for my fist -So before i take a chance on you - I gotta make sure you aren\'t here - just to get me pissed - So if your gonna get with me dont break my heart - So then I wont have to tear yours apart  -- I don\'t lie, my mouth does it for me.  -- i told my dad i stopped raising hell and he called me a quitter !!!!   -- In life, only one person gets to shine, so move outta my way, you\'re blocking my light!  -- BACK OFF! You\'re standing in my aura!  -- I think that butterfly is looking at me...  -- The early bird gets the worm, but the second mouse gets the cheese.  -- Before you live by the \"eary bird\" policy, figure out whether you\'re the bird or the worm.  -- Fact: If you ever hurt me...you get it back 10 times worse...   -- Can i get directions to you heart, I\'m kinda lost in ur eyes.   -- Boys are bad, throw rocks at them all!  -- I love rainbows, I love rain, I love you and... um, i forget the rest, but I think you\'re hott...   these are not mine i found them on some one\'s profile at faNFICTION. NET. you guys should check her out nayanya, she\'s a great writer. so mall the credit goes to nayanya for finding them. i\'f she\'s reading this i love your homepage and your an awsome writer.  Later Dayz Be Safe '),(2111,3694314,'05,July,2004','2004-07-05 00:00:00','   i was just randomly searching the net for some coldplay fanfiction when i happened upon someone\'s online diary. i have to tell u i have never met someone so like me and in all actuality i have never met them before in my life. Although i\'m so blind it could have happened. I love their poetry its beautiful, sad, raw, truthful. I\'m actually jealous seeing as how i haven\'t been wriing poetry in a long while, i\'ve been to busy concentrating on writing music.  you guys should check it out  http://offence.diaryland.com/bio.html  Later Dayz Be safe '),(2112,3694314,'14,August,2004','2004-08-14 00:00:00','    '),(2113,3694314,'13,August,2004','2004-08-13 00:00:00','   Dead so I\'ve played to merely hurt him hard Don\'t let greed take too much of usHe\'ll rivet hard to suffer rotten wrong The prince\'s mood darkens me strong C\'est lui,Mon prince bleu, mon soleil Chantant dans mon sommeil Des mots, des monts et merveilles Attendant que je m\'eveille C\'est toiMon prince bleu, mon soleilJe vais chanter pareil Des mots, des monts et merveilles Que jamais je ne m\'eveille Le Prince Bleu Je brulerai ton absence J\'erigeraiUn trone blanc Il reviendraSur son fier alezan Le Prince BleuComme je l\'attends C\'est lui,Mon prince bleu, mon soleil Chantant dans mon sommeil Des mots, des monts et merveilles Attendant que je m\'eveille C\'est toiMon prince bleu, mon soleil Je vais chanter pareil Des mots, des monts et merveilles Que jamais je ne m\'eveille Je donnerais Tous les croissants de lune Tous les bijouxDe la fortune Sent you to tears when pride first seized my womb Post-morten\'s tune, my prince he croons Sickly my prince\'s mood, so sullen shutdown the door today The more no more is relayed after dark confirms his bane His fall my prince is bruised, he\'s so lame his fairness ricochets\'till all his soars are self slain and demand a brother\'s reign C\'est lui,Mon prince bleu, mon soleil Chantant dans mon sommeil Des mots, des monts et merveillesAttendant que je m\'eveille His fall my prince is bruised, he\'s so lame his fairness ricochets\'till all his soars are self slain and demand a brother\'s reign Sickly my prince\'s mood, so sullen shutdown the door today The more no more is relayed after dark confirms his bane C\'est lui,Mon prince bleu, mon soleil Chantant dans mon sommeil Des mots, des monts et merveilles Attendant que je m\'eveille '),(2114,3694314,'08,August,2004','2004-08-08 00:00:00','   Don\'t label me! Don\'t look at me and blankly place me into a category like I\'m an inanimate, insignificant item on a store shelf. I am not a Goth or Punk, a Raver or Freak. I\'m just a unique human being with enough mental toughness and ambition to feel no inhibition in being myself to the fullest. Don\'t judge me on my choice and style of clothing. I am not your verbal pincushion for you to vent your frustrations on. You look and see weakness in profound individuality and it eats at you. Look away if it scares you. Look away if you see someone that\'s broken the mold to normality and you feel distress concerning where you would like to categorically place me.   Most of the time the people labeling are those who openly call themselves, \"Preps.\" Maybe it goes with the territory? \"Look at me in my Abercrombie and Fitch, check out my new Tommy Hilfiger gear. I am like, so Prep now, tee-hee!\" And to each their own. The corporate machine is what decides the in-fashion and most desirable name brand clothing to own. All the beauty and glamour magazines that are aimed at teens to 20 and 30-somethings are guilty as well. Making it seem as if it\'s a sin not to be a size 1 or less, making people feel socially unacceptable if they don\'t buy and use the newest trendy makeup brand. Followers feed it, it will only grow bigger and bigger. The demand for new, more expensive brands because there is always someone that\'ll pay 100 dollars for a T-shirt. Stop being sheep, stop letting the wolves in Giorgio Armani clothing push their styles on you.  To those who have ever look at another human being and judged them, there\'s always hope. There\'s always hope for you to change your ways of thinking. There\'s always tomorrow and with every new day is a chance to be more open-minded about others. It\'s no ones problem but yours and it\'s up to you to find a way to deal with it. People aren\'t going to change, it\'s human nature to want to be happy. Next time you look and feel the need to cast judgment upon another, take a look at yourself first. We\'re all pretty much the same deep down inside. We\'re all made up of flesh, blood and bone. It\'s a personal decision to be optimistic or pessimistic, letting something negative influence who and what you are is just a waste of time. We only get one life, I intend to live mine to the fullest. I don\'t know if the glass is half full or half empty because I\'m not stopping to look  NOT MY WORDS...but they describe how i feel so well '),(2115,3694314,'08,August,2004','2004-08-08 00:00:00','   I dont believe in labels. They\'re just false names and identity\'s created by the masses to classify some one as diffferent.      Who are you to judge.    To Thyne own self be true                                           dark88poet '),(2116,3694314,'08,August,2004','2004-08-08 00:00:00','   My music is the most important thing to me in the world. If u ever wanted to know who i was all you had to do was listen to my words. they\'re me in a nutshell. but sometimes i get worried that i\'ll never make it. do what i\'ve wanted to do since i was four, and that scares me cause then what will i be. i have nothing else, at least nothing that i enjoy and expresses me to the fullest.   Darkpoet dm be safe '),(2117,3694314,'08,August,2004','2004-08-08 00:00:00','   \"If they ever tell my story let them say that I walked with giants.Men rise and fall like the winter wheat, but these names will neverdie. Let them say I lived in the time of Hector. Let them say I livedin the time of Achilles.\"-- Odysseus (as played by Sean Bean in the movie \"Troy\") '),(2118,3694314,'07,August,2004','2004-08-07 00:00:00','   For all you wanna-be female and or male musicians, like myself who want to write songs about what you know and jam with your friends a must read is Guitar Girl by Sarra Manning. I know it may seem childish to all you to cool school kids but its quite deep in a sarcasticely nuerotic way if you ignore most of the main character\'s whining. it tackless subjects like betrayel, parental dissaprovment, interband relationships, managment misshaps, and so much more. I Just read it today and i learned alot even though its meant as a fiction read. One might wonder if the auther herself might have been writing from personal experiences, but what ever its all good and so is the book. so READ IT.  Later Days Dark88Poet DM BE SaFe much love '),(2119,3694314,'06,August,2004','2004-08-06 00:00:00','   If yall are wonderin why i\'m all of a sudden putting my poetry up its cause i\'m tryin to learn to write off the top of my head. usually i write in in 1 of my three poetry/music books and they turn out pretty good. Sorry that they\'re a little ruff around the edges, i promise you that they are not ussually like this. its just that i\'m just getting back into the swing of things seeing as how for this entire year with the acception of a few poems here and there i have been mostly concentrating on my music. so hopefully you\'ll bear with me  Much love Dark88Poet DM Danielle '),(2120,3694314,'03,August,2004','2004-08-03 00:00:00','   i walk a fine line between the shadows and death stray but a little and i\'ll be gone in the vastness those who are hungry wait for me to stop so that they may pray on my weaknesses it is but a game to them But For me it is a fight for control control over others power over my emotions  I will not stop until i win this game Tho childish it may seem I shall look upon it as my  first test my true test To see if I have what it takes And Then we shall see '),(2121,3694314,'02,August,2004','2004-08-02 00:00:00','   1. \"Hello. My name is Inigo Montoya. You killed my father. Prepare to die.\"--Inigo Montoya (Mandy Patinkin) in \"The Princess Bride\" 2. \"Inconceivable!\" \"You keep using that word. I do not think it means what you think it means.\"--Inigo Montoya (Mandy Patinkin) in \"The Princess Bride\" 3. \"Have fun storming the castle.\"--Miracle Max (Bill Crystal) in \"The Princess Bride\" 4. \"As you wish.\"--Westley (Cary Elwes) in \"The Princess Bride\" 5. \"I\'ll have what she\'s having.\"--Female diner (Estelle Reiner) in \"When Harry Met Sally\" 6. \"You had me at hello.\"--Dorothy Boyd (ReneÃ© Zellweger) in \"Jerry Maguire\" 7. \"Fasten your seat belts--It\'s going to be a bumpy night.\"--Margo Channing (Bette Davis) in \"All About Eve\" 8. \"You know how to whistle, don\'t you, Steve? You just put your lips together...and blow.\"--Marie Browning (Lauren Bacall) in \"To Have and Have Not\" 9. \"If I\'d been a ranch, they would\'ve named me Bar Nothin\'.\"--Gilda (Rita Hayworth) in \"Gilda\" 10. \"Veda\'s convinced me that alligators have the right idea: They eat their young.\"--Ida Corwin (Eve Arden) in \"Mildred Pierce\" 11. \"You\'re not too smart, are you? I like that in a man.--Matty Walker (Kathleen Turner) in \"Body Heat\" 12. \"Get outta here! And don\'t come back for five to seven days!\"--Vada Sultenfuss (Anna Chlumsky) in \"My Girl\"   AND.....   1. \"I coulda had class. I coulda been a contender. I coulda been somebody, instead of a bum, which is what I am, let\'s face it.\"--Terry Malloy (Marlon Brando) in \"On the Waterfront\" 2. \"We all go a little mad sometimes.\"--Norman Bates (Anthony Perkins) in \"Psycho\" 3. \"Mrs. Robinson, you\'re trying to seduce me.\"--Ben Braddock (Dustin Hoffman) in \"The Graduate\" 4. \"Hey, don\'t knock masturbation. It\'s sex with someone I love.\"--Alvy Singer (Woody Allen) in \"Annie Hall\" 5. \"She\'s my daughter! She\'s my sister! She\'s my daughter! My sister, my daughter. She\'s my sister and my daughter.\"--Evelyn Cross Mulwray (Faye Dunaway) in \"Chinatown\" '),(2122,3694314,'02,August,2004','2004-08-02 00:00:00','   Dark Poets Society  Tracy Ivy Star  Okay, so here you have it: metal with yet another twist. This band reminded me of Coal Chamber but without the budget (or makeup artist). Funked-up bass from the wonderful Tracy; nu metal stylings on guitar from Star; and a decidedly sinister backbeat from Ivy. Tonight at the show, be sure to listen for \"Teenage Angst,\" my favorite cut on the trio\'s debut album \"Dark Poet\'s Society.\" It\'s a classic in the making.And prepare yourself for the larger-than-life voice that singer Star never even knew she had. --reba rock   Dark Poets Society  Tracy   Ivy  Star   Dark Poets Society become known in the underground metal scene as that band with \"wicked mosh parts\"--the heavy instrumental segments of their Teenage Angsts. When Marilyn Manson hears \"Teenage Angst\" and wants to use it for his soundtrack to the Sara Michelle Gellar vehicle \"Buffy, Destroyer of Worlds,\" the band refuses on principle. They go on to rock the underground, becoming virtual royalty at Swedish Death Metal fests. fun fact:Dark Poets Society inspired a new genre of rock music called neopseudo-. '),(2123,3694314,'01,August,2004','2004-08-01 00:00:00','   anyways, just wanted to wish lots of luck to the band emma on their  demo\'s. they are awsome, and give hope to all us up and coming musicians   i feel a tear coming on  i\'m such a freak,  you guys should che them out, the link to theyre site is in my  \"Great Sites\"  section  Dark88Poet Dm Later Dayz Be Safe '),(2124,3694314,'01,August,2004','2004-08-01 00:00:00','   As you all know by now i have a thing  for quotes and with that said here are some more:  You have to be willing to look bad in order to get good. - Jack Canfield There are 2 kinds of people. Those who generalize and those who donï¿½t. - David Waisglass The test of a first-rate intelligence is the ability to hold two opposing ideas in mind at the same time and still retain the ability to function. - F. Scott Fitzgerald The belief that oneï¿½s own view of reality is the only reality is the most dangerous of all delusions. - Paul Watzlawick When there is no wind, row. - Old Japanese proverb An emotion without social rules of containment and expression is like an egg without a shell: a gooey mess. - Carl Tavris I cant choose how I feel, but I can choose what I do about it - Andy Rooney Anyone can become angry. That is easy. - Aristotle There is nothing good or bad, but thinking makes it so. - Shakespeare I f I go crazy, then will you still call me Superman? - 3 Doors Down The mind is its own place and in itself can make a Heaven of Hell and a Hell of Heaven - John Milton A man said to the universe: ï¿½Sir, I exist!ï¿½ - Stephen Crane ï¿½No one can make you feel inferior unless you agree to it.ï¿½ - Eleanor Roosevelt I believe that Courage is all too often mistakenly seen as the absence of fear - Leonard Zunin A smooth sea never made a skilled sailor - English proverb When you point the accusing finger at someone, three of your own fingers point back at you - Louis Nizer What you are speaks so loudly, I cannot hear what you say - Ralph Waldo Emerson Experience is the hardest teacher. You get the test first and the lesson later. I can feel it when someone smiles. - Harold Grace If you think communication is all talk, you havenï¿½t been listening. - Ashleigh Brilliant All speech is a dead language until it finds a willing and prepared hearer - Robert Louis Stevenson Television is another example of how people will look at anything instead of each other. - Anne Landers Not comprehending, they hear like the deaf - Heraclitus Iï¿½ve learned that being kind is more important than being right. - Andy Rooney The need to be right, the sign of a vulgar mind - Albert Camus Let us begin anew, remembering that civility is not a sign of weakness - John F. Kennedy No one said it would be easy, but no one said it would be this hard - Sheryl Crow Only the gentle are ever really strong - James Dean '),(2125,3694314,'01,August,2004','2004-08-01 00:00:00','   In my place, in my place Were lines that I couldn\'t change I was lost, oh yeah   I was lost, I was lost Crossed lines I shouldn\'t have crossed I was lost, oh yeah   And yeah How long must you wait for it? Yeah, how long must you pay for it? Yeah, how long must you wait for it? For it   I was scared, I was scared Tired and under prepared But I\'ll wait for it   And if you go, if you go And leave me down here on my own Then I\'ll wait for you, yeah   And yeah How long must you wait for it? Yeah, how long must you pay for it? Yeah, how long must you wait for it?   Singing please, please, please Come back and sing to me, to me, me Come on and sing it out, now, now Come on and sing it out to me, me Come back and sing   In my place, in my place Were lines that I couldn\'t change And I was lost, oh yeah Oh yeah '),(2126,3694314,'01,August,2004','2004-08-01 00:00:00','    '),(2127,3694314,'01,August,2004','2004-08-01 00:00:00','   We live in an abominable world where hate is accepted and baisically encouraged we all Turn a blind i to what we wish not to see face it its real as real as you or me  Keep an i on your children lest they be corrupted by someone else why give them the satisfaction when we can do it ourselves what i say may seem harsh but to quote Oscar Wilde \" The Truth is rarelly Pure and Never Simple\" And it does not help that silence rules our tounges at least when it benefits ourselfes or on the off chance someone we actually give a damn about  We are selfish creatures, human beings mostly thinking of ways to get away with things we ought not have done in the first place ah but now i\'m rambling or am i just saying what needs to be said who really knows anymore  I feel so tainted and insecure so corrupted and vain i long to rid myself all my sins and petty greivances  Maybe that is what heaven is a cleansing of all that is wrong in our bodies the same bodies used to kill and rape But now i have gone off the topic If ever i really was on it  I want so much more than predjudice and greed Starvation children dying on the streets wars being faught out bordom and pride And of Course Jealousy The root of all evil  I\'ve had enough of rambling although i have not said all that i wanted to say i guess That I shall save it for another time Another Day  Dark Poet Danielle DM Later Dayz Be Safe '),(2128,3694314,'01,August,2004','2004-08-01 00:00:00','    '),(2129,3694314,'01,August,2004','2004-08-01 00:00:00','    '),(2130,3694314,'01,August,2004','2004-08-01 00:00:00','   The chaos theory states that out of great confusion comes a simple and small solution '),(2131,3823631,'31,May,2004','2004-05-31 00:00:00','   Although the relationship was short lived, I had a great date at La Vista.  We went to pick up some wine before and sat in the parking lot drinking while we waited. Although that doesn\'t sound all that romantic, its a cute tradition at La Vista. Friends of mine met there two years and I haven\'t been able to pry them apart since then.  The mood is very festive and the food is wonderful. If you need a good first date place La Vista is a great suggestion.   '),(2132,3823631,'26,May,2004','2004-05-26 00:00:00','   Anonymous said...      Wow, this is great. I have something I have wanted to share with all the single men and women out there about my worst dating experience, but have never had a place to begin my tale.      When I turned 37, I was pretty determined to find the man for me and get married. One day out of nowhere, I recieved a call from a dating service telling me they wanted to survey my experience. I had previously responded to a personal ad I found in a publication I picked up at Cafe Artiste and thought they were referring to that. So I agreed to go in and talk to them.      It was actually Together Dating Service. I was put into a room and given a hard sale by one of their reps. I decided that since it was a goal of mine to be married, it might be worth the incredible sum of money to join.      The gist of their program is that you fill out a thing about yourself and what you are looking for in a mate so they can find you the perfect match. Sounded good to me. How substantial can a relationship be if it is founded only on physical attraction? Doesn\'t this process help you get to the chase?      Well first of all, I quickly discovered that there were no qualifications for the people making matches. As a matter of fact, I don\'t even think they read my profile. It said, clear as day that I\'m an atheist. They set me up with several born again Christian men. I had also indicated that I didn\'t want to date anyone outside the beltway and that I preferred meeting people close to where I lived or out towards the galleria. I guess they thought Conroe, Katy and Pasadena fit in that area.      All told, I went out with 23 men. By about number 19, I was screening them myself over the phone before even agreeing to meet them in person. They were sending me 2 and 3 intros a week. I was discovering that very few women were in this organization. While I was getting 2 - 3 intros a week, these guys are getting 1 - 2 intros a month. There was a lot of pressure on them to make something click since the opportunities were so few and far between.      Bottom line is I was absolutely miserable. I am so angry by the lack of care and respect they held for their patrons. It is a money scam from day one.      If anyone wants to join a dating service such as Together, DO NOT FOCUS ON THE NUMBER OF MEMBERS in their membership. That is what they want you to see. YOU GET THEM TO SHOW YOU IN WRITING WHAT THEIR SUCCESS RATE IS IN CREATING SUCCESSFUL MATCHES. Because I promise you, you mix enough men and women together, you\'re going to get some matches. But they gave me the impression that they did more than just throw people together.      It isn\'t worth the investment.   '),(2133,3823631,'26,May,2004','2004-05-26 00:00:00','   Nice blog, Angie!   I\'ll be filling some of the webmaster duties for this blog, but Angie is your fun and willing administrator.  If you would like to post items on this blog, please contact:  angie.jeter [at] gmail dot com (I didn\'t list the actual email address to prevent spam, but sound it out).  Even though I\'m not thirtysomething nor really in the single crowd anymore, I can\'t wait to see what kind of wild stories get posted up here.   Have fun and happy blogging!   '),(2134,3823631,'25,May,2004','2004-05-25 00:00:00','   Hello,  This blog is for singles in the thirty plus age group.  My friend Cynthia has a blog and it\'s kinda fun to see what\'s going on in her life.  I thought that this would be a good forum to tell the good and the bad of dating.    The following questions are examples of what I feel might be helpful to other daters.  Where\'s a good place for a first date?  What are your dating nightmare stories?  Do you have good ideas for meeting new and fun people?  Do you use an online service?  What are the best ones?  Do you have any tips as to the best way to fill out the online profiles?  How do you transition from the cyberworld, to the real world?  There are probably others I haven\'t thought of.  I\'d like to add links, but I am waiting to get feedback from others before adding them.  Thanks for taking the time to read my blog, I look forward to hearing your thoughts!  Angie   '),(2135,3823631,'30,June,2004','2004-06-30 00:00:00','   In a comment to my last post somebody asked why would I want to get dates from the internet instead of traditional ways of meeting people.  He wasn\'t asking in a judgy kind of way, he\'s done it too, but it really is a good question.  The first reason that I use the internet is that I am busy.  I am a struggling realtor just trying to get as much business as possible.  Sometimes I am not home until late and don\'t want to go out.  With a traditional job you can go to happy hour, but that is not often in the cards for me.    The next reason is something that may be difficult to understand.  I am an extremely social person.  I have friends all over this town and am invited to really cool events.  You\'re probably thinking well what\'s the problem?  I think that the problem with that is I am so unapproachable when I\'m out and about because I\'m always engaged in conversation with some friend I haven\'t seen in a while.   Finally, I just can\'t get into that meat market groove.  When I see somebody that I\'d like to meet, I am too shy to approach.  When somebody starts talking to me, I am often clueless to the fact that they would like to get to know me.    I am a very nice person, but sometimes appear to be aloof to new people.  I think that an internet date gets alot of that out of the way.  It\'s like I\'d like to meet somebody, you\'d like to meet somebody, so we are here to see if we click.    What do you think?   '),(2136,3823631,'29,June,2004','2004-06-29 00:00:00','   I don\'t know if anyone reads this, but I just thought I\'d share.  I am going out for dinner and drinks with another fine gentleman I\'ve met on the internet.  I think he\'s a nice guy, but I\'m not sure we click.  I\'m going out with him anyway because my friends tell me I don\'t give people enough of a chance.  So, I\'m not too sure about this guy.  I guess if the date turns out well, I can never show him my blog.  hee! hee!  I have had some very sweet comments on my last dating disaster.  I know nice guys are out there!  I even think that the guy I\'m going out with tonight is nice.  We\'ll see what happens.   '),(2137,3823631,'25,June,2004','2004-06-25 00:00:00','   I just read Monica\'s response to Bill Clinton\'s new book \"My Life\".  She is upset about the way that Clinton described their affair.  She then told the British Press that \"Clinton\'s book made her feel \"like an insignificant piece of dirt.\" Well when you are having an affair with a married man, you often feel like a piece of dirt.    Maybe the new publicity will give her a forum, but a forum for what, whining?  Read for yourself and see what you think.  http://www.msnbc.msn.com/id/5295636/?GT1=3584  I try not to be judgmental.  I know that people make mistakes, and that although an affair is a bad thing to do, it does not make you a horrible person.  Bill Clinton did not \"ruin\" Monica\'s life.  Monica was not forced into the affair.  Any account of this affair will not show Ms. Lewinsky\'s good qualities.  In her defense, an affair with a man that powerful might be difficult to resist.    I hope that this ugly wrinkle in our history one day goes away, but in the meantime, take responsibility.  Say I did a bad thing and then \"Suck it up\"  Oh, I guess she already did!!!   '),(2138,3823631,'21,June,2004','2004-06-21 00:00:00','   This past weekend, I just didn\'t have the energy to go through another internet match up.  I had discussed meeting one gentleman, I just had to put it off.  Too many bad dates in a row.  I just needed a break.  No pushy guys, no whiny guys.  I just needed to hang out with my friends!  I am not saying that I\'m discouraged, or don\'t want to do this anymore, but just needed to back away.  This internet thing can be fun, but I think I might need to take a step back.  My friend told me that I should do one of the more involved sites like E-Harmony.  I am considering that too!  I\'m also trying to get out more and meet people the old fashioned way,in person.  Let\'s see how that works.   '),(2139,3823631,'14,June,2004','2004-06-14 00:00:00','   I started chatting with a guy from one of the personals sites and we decided to meet.  We went to a bar, and hung out for a bit.  As time went by the hints that the night should end shall we say at breakfast increased.  Not that this is not flattering, but I got the impression that if he were with any female, she would be getting the same \"offer.\"    This guy told me how if he didn\'t like the profiles that he read, he would sent them rude messages.  The reason he didn\'t like the profiles was that he felt that the women put in too many expectations.  I guess he couldn\'t measure up.  The rude messages would be like hey, you wanna get laid and stuff like that.  I knew at that time, his match made in heaven wasn\'t me.  I told him that I realized that he was a \"sure thing\" but that I\'d have to pass.  Made up some stuff about hope to see you again, and bolted.     I don\'t know that there is a moral to this story, just that there are people out there that don\'t have them.  It\'s not that this guy made clear what he wanted, just that I didn\'t expect such a bottom feeder.    I am looking forward to a new and better experience.   '),(2140,3823631,'13,June,2004','2004-06-13 00:00:00','   I found  urlLink this article  on CNN that gives some incite on how to watch out for bad online dating experiences. Something I learned while reading this article is, men lie about being in a relationship, and women mostly lie about their looks. Some online dating sites want to offer a certified, dated photo on your profile.   '),(2141,3823631,'10,June,2004','2004-06-10 00:00:00','   Good date We were heading from West Houston to a friend\'s house in Galveston. Julie needed to pick up a computer from her friend. We seemed to spend time in the truck quite a bit. We liked the time spent traveling together. We would talk while she drove. I cannot drive, so it is a good thing she enjoyed being the one behind the wheel. it was a beautiful day. The drive was nice. When we got to the friend\'s house we decided to go for a day at the Aquarium Restaurant. We made it down to the front and waited inside the lobby for a while. We were sitting down on a couch. Julie laid her head on my shoulder and fell asleep. We got that \"cute couple\" look from just about every one who walked by. When we were seated, we got a table right by the window overlooking the sea. It was a breezy day with a bright sun shining high. I don\'t remember what we had, but it was good. I remember looking at the stingray (I think it was a stingray) swimming up the glass. I remember looking at Julie and wanting so much to love her. When we left the restaurant, we went to a couple of shops then we went to some one\'s house and hung out for a while. Julie and I broke up a couple of years ago. When I think of Julie, I remember her smile that day.   '),(2142,3823631,'30,July,2004','2004-07-30 00:00:00','   Slug boy got kicked to the curb again last night.&nbsp;  I had a semi-bad day and was going to skip swimming (maybe).&nbsp; Charles came over like 10 minutes after I got home without calling, empty handed except for the Sonic cup he was drinking (where the hell is mine).&nbsp; Aside from a hug and kiss when he came in (initiated by me of course), no other affection was shown and I really didn\'t know what the hell he was doing there.&nbsp;  He asked what I was doing and I said I had just got home and was watching&nbsp;Green Acres.&nbsp;&nbsp;Duh, I\'m&nbsp;still in a suit and the TV is on.&nbsp; He didn\'t say anything else so I went back to the couch and he followed, sitting a few feet away.  No hand holding, no kissing, nada.&nbsp; He would just ask every few minutes what I was doing.&nbsp; (Duh!!)  I finally asked why he was there and he said I had a bad day so he came over.&nbsp; (And????&nbsp; Did he offer a massage, going out to dinner, anything????).  I was bored.&nbsp; I definitely wasn\'t going to entertain him.&nbsp; Apparetnly that\'s what\'s expected.  He fell asleep.&nbsp; I got up and did laundry and the dishes. him snoring the whole time.&nbsp;  I finally went over, kicked him awake and told him he needed to go.   Haven\'t heard anything.&nbsp; definitely not calling him.&nbsp; Will he contact me?&nbsp; Do I care?  Let the wagering begin!   '),(2143,3823631,'27,July,2004','2004-07-27 00:00:00','   I stole this from A friend who stole it from a friend. I think somebody else even stole it from the person I stole it from. Feel free to steal it from me.         I AM 32% WHITE TRASH!   urlLink     The white trash in my blood will not keep me from becoming a doctor or a lawyer, but it will keep me from a good haircut and any sort of fashion sense.'),(2144,3823631,'26,July,2004','2004-07-26 00:00:00','   Throughout the past couple of weeks, I\'ve been pondering the question of what exactly is love -- does anyone know?   (I\'m just talking about romantic love, not family or platonic love.)   Of course, the question arose because I\'m dating someone new, but also things have been happening to people I know that contrast each other and have weird timings.   Two weeks ago, one friend that is in a rather idyllic relationship for the past two years (July 2002) with a real sweetheart mentioned not&nbsp;having a shiny bobble on her left hand.&nbsp; (Note: she wasn\'t complaining, just mentioning as part of another conversation.)&nbsp; These two really do seem \"perfect\" for each other.&nbsp; In the past two years, both have really grown as individuals and at the same time grown as a couple.&nbsp; Pretty awesome.&nbsp; Yet still a naked third finger.   The next day I went to lunch with a&nbsp;friend that has barely known&nbsp;her husband for two years.&nbsp; They met in February/March 2002, dated 6 months (he\'s a pilot, so has a weird travel schedule), he got sent to Kuwait for 6 months with the Army Reserves but 1/2 way through came back and they got married (October 2002).&nbsp; So in Feb/Mar 2003, about a year after they first met, her husband returned.   What a contrast.&nbsp; Not that one relationship is better or stronger than the other.&nbsp;&nbsp;It\'s just odd how two different relationships move at such drastically different paces.&nbsp; I hadn\'t thought about it before, but the timing of the two conversations made the timing differences stand out.   I would definitely say the love is there between&nbsp;both couples.&nbsp; But that takes me back to the question of what is love......   Background info on my neighbors -- they\'ve been living together for a year.&nbsp; The guy in the relationship has been more the guiding force in the relationship.&nbsp; He was the one that wanted them to move in, he\'s been the one talking marriage and&nbsp;asking about different ring styles, etc.&nbsp; She defiantly reciprocates the emotion.&nbsp;&nbsp;They had know each other for years before the relationship developed and it seemed very strong.&nbsp;  Last week, she came by to tell me they were moving.&nbsp; The prior Sunday he confessed that he had been seeing someone else since February.&nbsp; She said he came home from work (a paramedic/fireman) kissed her and said he loved her and took a nap and when he got up told her it was over.&nbsp; He said he wanted it to work and that\'s why he wanted them to get a dog together (so they would be a family), but decided he would rather be on his own.  Damn.   When I went back to work I was e-mailing with a former co-worker who has been married for TWELVE years and has for kids (but seems like it\'s nothing to have that many kids, she seems like a kid herself), who said she\'s leaving her husband.&nbsp; He knows and they are still living together while she saves money to get her own place.&nbsp; Once she does, they\'ll each take two kids and keep trading off (she laughed too).&nbsp; The kids are like 10, 8, 5 and 2 (the 2 y.o. was after a tubal ligation).&nbsp; She said it had been coming on for 8 years.&nbsp; She loves him, but not in that way.&nbsp; That seemed like a tight relationship.&nbsp; I remember one time he came to pick her up at work and we were talking about him running off with someone famous (can\'t remember who) and he said he\'s rather keep his wife and wouldn\'t want to trade her for anyone else.&nbsp; He didn\'t say it in a sappy way, but a content, happy way.&nbsp; Every now and then they would do super sweet things for each other -- little gifts, flowers, etc. That made it seem like a really awesome relationship.&nbsp; And now it\'s ending.....   I guess there is no guarantees in the world.&nbsp; And feelings/emotions change.&nbsp; So what\'s the point?    In Steel Magnolias, when Shelby was talking about her illness she said \"I\'d&nbsp;rather have 15 minutes of wonderful than a whole lifetime of nothing special.\"&nbsp;  She\'s got a point there....   A former co-worker that was in her late 30s and not married had the attitude of \"I\'d rather never be married than be divorced.\"  She\'s got a point there.....   '),(2145,3823631,'22,July,2004','2004-07-22 00:00:00','   I\'ve got a future blog entry in my head, but no time to transcribe.&nbsp;  In the meantime, here\'s&nbsp;a song to get stuck in your head, as it has been in mine.&nbsp;   It\'ll be a little appetizer to the blog enrty.....     WHAT IS LOVE &nbsp;&nbsp;     CHORUS:    What is love    Baby don\'t hurt me    Don\'t hurt me    no more&nbsp;     &nbsp;   Baby don\'t hurt me    Don\'t hurt me    no more    What is love    Yeah     &nbsp;Oh I don\'t know    why you\'re not fair    I give you my love    but you don\'t care    So what is right    and what is wrong    gimme a sign&nbsp;     &nbsp;   (chorus x2)&nbsp;     &nbsp;   uoh oh...    Oh I don\'t know    what can I do    what else can I say    it\'s up to you    I know we\'re one    just me and you    I can\'t go on&nbsp;     &nbsp;   (chorus x2)&nbsp;     &nbsp;   uoh oh..What is love    What is love    What is love    Baby don\'t hurt me    Don\'t hurt me    no more     &nbsp;   Don\'t hurt me    Don\'t hurt me&nbsp;     &nbsp;   I want no other    No other love    This is your life    our time    When we are together    I need you forever     &nbsp;   Is it love&nbsp;     &nbsp;   (chorus x2)&nbsp;     &nbsp;   uoh oh..&nbsp;     &nbsp;   (chorus x2)&nbsp;     &nbsp;   Baby don\'t hurt me    Don\'t hurt me    no more     &nbsp;   Baby don\'t hurt me    Don\'t hurt me    no more     &nbsp;   Baby don\'t hurt me    Don\'t hurt me    no more     &nbsp;   Baby don\'t hurt me    Don\'t hurt me    no more     &nbsp;    what is love?!...  &nbsp;   '),(2146,3823631,'20,July,2004','2004-07-20 00:00:00','   I thought with two very different personalies, I\'d ad some variety to this blog.&nbsp; Stay tuned for Anna Bannanas spicy serial she calls her life.&nbsp; Will the red bull work?&nbsp; Should she kick the sug to the curb? Stay tuned!   '),(2147,3823631,'19,July,2004','2004-07-19 00:00:00','   Howdy happy and adoring fans~  &nbsp;  I recently wrote a series of e-mails updating friends on my adventures going out with someone I met through 8 minute dating.&nbsp; The updates were quite long and one was rather risque. I would be more than happy to e-mail the entire text at your request, but in the interest of saving time, space and some electrons I\'ll just post the brief synopsis included at the beginning of each update:  &nbsp;  #1  \"Anna Finally Goes on a Date\"  The short version: Yea!  It was a good date. Not quite sure when it\'s going, but definitely worth pursuing. In fact, pursuit continued at lunch and will tonight also.....  &nbsp;  #2  \"Anna Finally Goes on a Date AND is Actually Still Seeing Him a Week Later\"&nbsp;  The short version:  I almost ditched him earlier in the week, but I didn\'t and am now starting to look forward to spending time with him.  &nbsp;  #3  \"Anna Finally Goes on a Date and Yeah, He\'s Still Around Here Somewhere\"  The short version:  I guess now he should be classified as a boyfriend. Whatever.  &nbsp;  I never got around to writing a #4, but in summary I\'m kinda sick of the&nbsp;lazy-ass slug boy, but regular sex is nice so I may continue to keep him around.  &nbsp;  My future bitching on this site will probably revolve around this relationship, so please make yourself familiar with the situation....  &nbsp;  Wahoo!  Anna   '),(2148,3823631,'19,July,2004','2004-07-19 00:00:00','   &nbsp; &nbsp; Following is some information on a single\'s event TOMORROW that was on the Houston Single Source weekly e-mail.&nbsp; Sounds like it could be a fun event, and if it isn\'t at least there is a bar there. I\'ll be at swimming class (I\'m trainng for a sprint (mini) triathlon), but want to hear some good stories about what I miss. &nbsp; Wahoo! Anna &nbsp;  Tuesday, July 20  Fast Dater presents Singles Fast-Tag Party at The River CafÃ©, 3615  Montrose Blvd. A creative way to meet other Houston singles. Just  bring a picture of yourself, or have one taken at the party, and have  fun. Let others discreetly know you would like to meet them via the  Picture Wall of Fame, and others will do the same. Bring your  friends...you never know who you will meet. There will be door prizes,  palm readings and a lot of fun. R.S.V.P.\'s are mandatory.  rkratz@fastdater.com or  http://houstonupcomingeventslist7.c.topica.com/maacsKfaa8wZ6bfNV4VbafpKht/  7 p.m.   '),(2149,3823631,'16,July,2004','2004-07-16 00:00:00','   Ouch!&nbsp;&nbsp; My arm has finally been twisted enough to break into the world of bloggers.&nbsp; I don\'t see myself posting enough to maintain a fan base of happy, adoring fans but Angie has been kind enough to grant me priviledges here.  &nbsp;  Most of my posts will probably center around the trials and tribulations of dating someone new.&nbsp; I\'ve been seeing him an entire month and in Annaville, that\'s about 3 weeks too long.&nbsp; But to make an attempt at keeping&nbsp;happy, adoring fans I\'ll keep him around a while longer to entertain y\'all.  &nbsp;  Be prepared for the first three posts that will be made shortly.&nbsp; It\'s a cut and paste from some e-mails that went out regarding the first month with the aforementioned dude-de-jour.&nbsp;&nbsp; I blew a couple of people away with my candor in post #2, so I\'ll try to keep comments more PG in future posts.&nbsp; Must keep the fans happy and adoring......  &nbsp;  Wahoo!  Anna   '),(2150,3823631,'16,July,2004','2004-07-16 00:00:00','   I was with a group of people last night debating whether or not the United States actually went to the moon.&nbsp; To be honest I never really thought about it.&nbsp; When asked if I thought we had ever gone to the moon, I, without hesitation replied that I believe we went to the moon.&nbsp; I am not naive, I know that many events in our history did not occur exactly the way that our history books said they did.&nbsp; I know that it really wasn\'t the tooth fairy that left money under my pillow.&nbsp; I don\'t think that our government, even if they wanted to, faked us out and staged the moon landing.  &nbsp;  This is a link to the article that refutes the hoax claims on the NASA website: urlLink http://liftoff.msfc.nasa.gov/News/2001/News-MoonLanding.asp &nbsp;  I am only looking at one side of this issue today.&nbsp; Please let me know what you think!   '),(2151,3823631,'14,July,2004','2004-07-14 00:00:00','   My friend sent me information about an upcoming singles event that I wanted to share with you.&nbsp;  &nbsp;  There were 140 young adults--mostly singles, a handful of couples--from  all  over Houston at St. Cyrils Church on Monday night.&nbsp;   Won\'t you join us this week for the next episode of CAFE CATHOLICA? It  will surely be engaging as Fr. Michael Barrett speaks to us about dating  and sex. He will then take questions from the floor.&nbsp;   Newcomers are welcome to join our carpool from St. Bernadette church  this  Monday, July 19th. We\'ll meet near the playground between 5:30 and 5:45.  We\'ll leave at 5:45 sharp. You can call or email me if you plan to come so  we\'ll be looking out for you.&nbsp;   Mass begins at 6:30 and will be celebrated especially for us by Bishop  Joe  Vasquez.&nbsp;   7:00 Music and Food&nbsp;   7:30-9:00 Fr. Michael Barrett \"Bachelors and Bachelorettes\" - dating  and sex&nbsp;   If you\'ll be meeting us at St. Cyrils (10501 Westheimer Rd @ beltway  8),  I\'m looking for someone willing to take 2-3 passengers back to the carpool  site at St. B\'s after the event. Please call or email me if you\'d be  willing and able.&nbsp;   There is a $5 suggested donation per night.  In past years, this has been a very popular topic.  Hope you can join us!  Call me for more information at 281 488-0251 or, on Monday at 281  455-1141  cel&nbsp;   God bless,  Donna D.   '),(2152,3823631,'14,July,2004','2004-07-14 00:00:00','   This question is just as annoying as \"Why are you still single?\" - if not more so.   Translation:  You\'ve been dating your boyfriend for a while.  You talk about him a lot.  Shut up and make it official.   The big problem with this is that people ask the woman in the relationship.  If she doesn\'t have a ring on her finger, chances are that she doesn\'t know either!  I have been asked this question three times today.  Is it because I talked about my boyfriend or brought it up in any way?  No.  All three came from people I haven\'t talked to in a long time who wanted to make conversation.  I graciously referred all three of them to my boyfriend if they really want to know.  I should be used to it, after getting the question all the time, but I\'m not.  At first, I thought it was a sign that someone cared and was interested in how I was doing.  Now, I feel a bit insulted when people ask.  I don\'t know why.  It must be my own fault; I have pictures of us together on my desk, many times I answer the \"What did you do this weekend?\" question with talk about what we did together.  Perhaps people think that having a boyfriend is all I have going on in my life.  On the other hand, between school, Toastmasters, and work, I know I don\'t give off that impression, either.    I wonder if he gets the question as often as I do.  Chances are, no.  Why do people ask this, when they clearly know the answer?  Any thoughts?  Thanks for letting me rant.   '),(2153,3823631,'12,July,2004','2004-07-12 00:00:00','   I was watching  Live With Regis and Kelly  this morning and saw the new 1 million dollar winner.  What a great thing, but this define who he is?  No, but a statement he made on the show did.  He told Regis and Kelly that he has received many letters of encouragement during this process, and told about one in particular.  A girl wrote him and said that she was considering dropping out of high school, but something about his success on the show made her change her mind.  There is nothing sensational about this guy.  No headlines about him saving a baby from a burning building or other acts that are generally considered heroic, but something about him helped somebody choose the right path.  I don\'t think that this girl thinks that graduating from high school will ensure her fifteen minutes of fame like Jennings, but who he is, made her want the best for herself.    I know this isn\'t with the my normal blog theme, but I was inspired by this story and wanted to share.    Remember to be the good person you are, and good things can come from that.  I think that Jennings place on Jeopardy is not to win a million, that\'s just a bonus.  He is there to provide a good example, maybe just for that one person.   '),(2154,3823631,'12,July,2004','2004-07-12 00:00:00','   I am currently in a training program for the Houston Marathon.  Yes, I am gearing up for the big event in January.  I wont be training alone.  I\'m working with Houston Fit you can find them  urlLink here.    This seems to be a positive way to meet people.  I signed up this past weekend and they are taking registrations through next Saturday.   If nothing else, what a confidence builder.  At the end of this program I will be able to say that I can complete a marathon.  Cool!   '),(2155,3823631,'09,July,2004','2004-07-09 00:00:00','   I just read the following and would like to know your take on this.   Dear Rachel, I am so tired of being asked why Iâ€™m still single! I am 42 years old, and would love to meet someone special. Itâ€™s hard enough for me not to get discouraged about dating without everyone asking me about it, as if there is some deep, dark secret to my plight. From my parents, to my co-workers, to my hairstylist, to my dentist, I am constantly asked this question. How can I answer without sounding annoyed, defensive, or depressed? (All of which I am actually feeling, by the way!)  â€” Bridget in New Hampshire  Dear Bridget, Ah, the eternally annoying question! It should be illegal in 50 states to ask that question, donâ€™t you think? It ranks up there with asking a woman â€œWhen is the baby due?â€ when sheâ€™s not actually pregnant.   First of all, donâ€™t assume that someone is asking â€œWhy are you single?â€ because he or she is looking for your faults, but rather always tell yourself that the question is really meant to imply, â€œYouâ€™re so fantastic! I canâ€™t believe a great guy hasnâ€™t snapped you up yet!â€ Edit your perception of the question first.   Then you need to answer the question directly with this single sentence: â€œI just havenâ€™t met the right person yet.â€ This is a complete sentence. There is no â€œbecauseâ€¦â€ that trails off the end here. This is not the time to divulge all your issues, quirks, and broken romances.   No matter who asks you, and how annoyed you really are inside, remember to always make lemonade out of lemons. Instead of acting irritated, use this opportunity to probe whether the questioner has someone wonderful to introduce to you. You are speaking to someone who is likely to have a single brother, cousin, neighbor, college roommate, or co-worker, so you need to convey an upbeat, lovely impression of yourself. Who knows what may ensue during the course of a conversation about you being single?    This is my take on the whole thing.  I think you shouldn\'t \"make lemonade out of lemons\".  I think you should be real.  People who have no business asking that question ask it.  I\'ve been asked at family gatherings in front of several people.  I agree you shouldn\'t have a fit because people ask you this, but it is a good idea to let them know that the questioning makes you feel bad.  Be nice, but be real.    What\'s your take?   '),(2156,3823631,'08,July,2004','2004-07-08 00:00:00','   6 billion people are on Earth at the moment.  At least, thats what I heard, somewhere.   3 billion of them are women , roughly.   1 billion are about the right age.  18 to say 45 or so. It\'s funny that the 18 will never go up. I\'m not so sure about the 45.   500 million are healthy.  Drug attics, starving, disease, ... I wonder if this number should be much, much smaller. With the international AIDS epidemic, I bet the number is smaller.   250 million are of the right level of looks . Lets face it, I\'m week I cannot fool myself into thinking that I&nbsp; will find just any one attractive. Also, a very pretty woman is out of my league. She would just leave me. =[   20 million are not involved at the moment.  These are good catches so far. Healthy, cute, of proper age, that\'s a formula for success.   40 thousand are in the U. S.  The U. S. has about 2% of the world\'s population. Yet, we use 40% of the worlds energy. I bet that is an old number. I\'ve heard China is about to take over as the fastest growing energy consumer.   2 thousand are in the great state of Texas.  I\'m guessing 5% of the U. S. population live in Texas. I don\'t care if that pretty far off. It gets the point across.   1 thousand will eventually be in the same city as I.  I don\'t travel much. Hopefully these women do. Optimistically, about half of the women and I will be in the same city at the same time I suppose.   50 of these near perfect matches I will meet.  I don\'t get out much. I suppose I will only actually run into a precious few of the women who have met the above criteria. I define running into as \"hi.\" ... \"hi.\"   10 women will notice me, and I will notice them.  Ah, this is the one that almost counts. This is where our eyes meet. Our breath grows short. The world stops just for us. That is not love. There is more to it.   1 woman and I will fall in love.   '),(2157,3823631,'01,July,2004','2004-07-01 00:00:00','   I thought that this article (see link below) made a lot of sense.  You can use this etiquette advice with you potential dates, or your regular IAM friends.  The key to etiquette is being sensitive and putting yourself in the shoes of others.  So many times I\'ll tell somebody that I\'m busy, or have somebody at my desk.  They keep sending me messages.    Do you have any other IAM advice?  http://msn.match.com/msn/article.aspx?articleid=2389&TrackingID=516164&BannerID=541888   '),(2158,3823631,'10,August,2004','2004-08-10 00:00:00','   Jennifer Crisp and I are putting together a new Theater Under The Stars (TUTS) group for the upcoming season. The season is SIX shows for    ONLY $95    per person (less than $16 per show!!). Your tickets can be upgraded (just pay the difference) and/or changed to a different date for a $3 fee (this stuff is between you and TUTS).  Let me know if your interested so we can make sure we have the minimum number of 15 people to get this special price - otherwise tickets would be $120..  All shows will be on a  SATURDAY night at 8pm . The shows and the date of our tickets (with the dates of the run in parenthesis) are:   Hair , September 25 (September 9-26)  Chorus Line , November 6 (October 21 - November 7)  Singing In The Rain , December 18 (December 2-19)  Anything Goes , February 12 (January 27-February 13)  The King and I  ( with Stephanie Powers ), April 2 (March 17-April 3)  Peter Pan  ( with Cathy Rigby ), May 21 (May 5-May 22)   '),(2159,3823631,'08,August,2004','2004-08-08 00:00:00','   Girls are evil and I can prove it. I found this somewhere.   '),(2160,3823631,'04,August,2004','2004-08-04 00:00:00','   for when they don\'t bring flowers.....   urlLink http://www.davidandgoliathtees.com/games/boygame5.swf   '),(2161,3823631,'02,August,2004','2004-08-02 00:00:00','   Tuesday Charles brought me two dozen caramel colored roses.  Makes up for some slug activity....not sure how much though....   '),(2162,3823631,'02,August,2004','2004-08-02 00:00:00','   I\'ve heard stories about parents at little league games, but this one was so dissappointing!  I think that Clemens could have definately provided a better example for parents.  Clemens asked to leave after arguing call Story  --> Houston  Astros  pitcher  Roger Clemens  was asked to leave a youth baseball game over the weekend for arguing a close call that went against his son\'s team. Photo gallery urlLink Week in baseball   urlLink   Clemens was at the game Saturday watching his son, Kacy, compete in a 10-and-under game organized by Triple Crown Sports when Clemens contested a call at second base that went against the Katy Cowboys. He spit sunflower seeds at an umpire\'s leg and was asked to leave, said Jim Carpenter, a field supervisor with Triple Crown. \"I supported the umpire\'s decision and he (Clemens) respectfully left,\" Carpenter told the Craig Daily Press. Katy lost the game to the Bakersfield Curve, 11-5. Triple Crown Sports features a franchise system aimed at pitting top teams from across the country against each other. Clemens\' agent, Randall Hendricks, did not immediately return a call Monday. Clemens has racked up 322 wins and 4,240 strikeouts in his 21-year major league career.   '),(2163,3823631,'02,August,2004','2004-08-02 00:00:00','   Slug boy slithered back in.  He called Friday asking me to dinner.  I had already planned to meet Wyatt at Brian Oâ€™Neillâ€™s.  Slug boy came along.  Whatever. On Saturday after scuba, I was ravenous.  Slug came over to go with me to Tacqueria la Tapetia (which is not as good now that theyâ€™re in the fancy new digs).  Then he was stuck running errands with me all day â€“ back over to my place to take my car in to get inspected and the oil changed then back over to the scuba store to get the tanks refilled and for me to get a skin suit, then back over to pick up my car, then back over to Academy and the dollar store, the back to pick up a prescription.  We didnâ€™t plan well, so Â½ the stuff was by me and the other Â½ was in the Westheimer/Shephard area.  Anyway â€“ the point of all this is that running errands is more bearable when someone else is along.   Especially when itâ€™s a slug boy that knows heâ€™s about to have a ton of salt poured on top of him.   I donâ€™t usually do relationships.  So I figure slug boy is a good practice relationship until the real guy comes along.   '),(2164,3374253,'30,June,2004','2004-06-30 00:00:00','   After about one year of experience in just maintaining the existing piece of software by providing some fixes and new features, I was given an task to write a new code right from scratch.  First where to start is the biggest problem we faced.  Actually you learn a lot in maintenance, but it was all in the negative. Every problem you faced was caused by a failure of the old code to prevent problems during maintenance. But this doesn\'t always teach you how to write good, new code that avoids such issues. Your entire thought process only reacts to the events it becomes difficult to visualize the big picture in just one short and come up with some design architecture.  But for new development, we need to sit down and consider -how- we intends to build the application, rather than focusing in on how to structure specific features. '),(2165,3374253,'18,May,2004','2004-05-18 00:00:00','    Just trying what this blogging is all about....   '),(2166,3374253,'24,June,2004','2004-06-24 00:00:00','    The flash mind reader http://www.dslextreme.com/users/exstatica/psychic.swf  But how come this is possible man ?  Whats the trick ?   '),(2167,3374253,'20,July,2004','2004-07-20 00:00:00','    DIAMOND is just another piece of COAL that did well under pressure   '),(2168,3374253,'01,July,2004','2004-07-01 00:00:00','    Somebody just forwarded this mail...  10 Ways of Marrying The Wrong Personby Rabbi Dov Heller, M.A. 1. You pick the wrong person because you expect him/her to change after you\'re married. The classic mistake: Never marry potential. The golden rule is, if you can\'t be happy with the person the way he or she is now, don\'t get married. As a colleague of mine so wisely put it, \"You actually can expect people to change after they\'re married... for the worst!\" So when it comes to the other person\'s spirituality, character, personal hygiene, communication skills, and personal habits, make sure you can live with these as they are now.  2. You pick the wrong person because you focus more on chemistry than on character. Chemistry ignites the fire, but good character keeps it burning. Beware of the \"I\'m in love\" syndrome. \"I\'m in love\" often means, \"I\'m in lust.\" Attraction is there, but have you carefully checked out this person\'s character? Here are four character traits to definitely check for:  Humility: Does this person believe that \"doing the right thing\" is more important than personal comfort? Do I want to be more like this person? Would I like my child to turn out like him or her?  Kindness: Does this person enjoy giving pleasure to other people? How does s/he treat people s/he doesn\'t have to be nice to? Does s/he do volunteer work? Give charity?  Responsibility: Can I depend on this person to do what s/he says s/he\'s going to do?  Happiness: Does this person like himself? Does s/he enjoy life? Is s/he emotionally stable? Ask yourself: Do I want to be more like this person? Do I want to have a child with this person? Would I like my child to turn out like him or her?  3. You pick the wrong person because the man doesn\'t understand what a woman needs most. Men and women have unique emotional needs, and more often than not, it is the man who just doesn\'t \"get it.\" Jewish tradition places the onus on the man to understand the emotional needs of a woman and to satisfy them. The unique need of a woman is to be loved -- to feel that she is the most important person in her husband\'s life. The husband needs to give her consistent, quality attention. This is most apparent in Judaism\'s approach to sexual intimacy. The Torah obligates the husband to meet the sexual needs of his wife. Sexual intimacy is always on the woman\'s terms. Men are goal-oriented, especially when it comes to this area. As a wise woman once pointed out, \"Men have two speeds: on and off.\" Women are experience-oriented. When a man is able to switch gears and become more experience-oriented, he will discover what makes his wife very happy. When the man forgets about his own needs and focuses on giving his wife pleasure, amazing things happen.  4. You choose the wrong person because you do not share a common life goals and priorities. There are three basic ways we connect with another person:  -chemistry and compatibility -share common interests -share common life goal  Make sure you share the deeper level of connection that sharing life goals provide. After marriage, the two of you will either grow together or grow apart. To avoid growing apart, you must figure out what you\'re \"living for,\" while you\'re single -- and then find someone who has come to the same conclusion as you. This is the true definition of a \"soul mate.\" A soul mate is a goal mate -- two people who ultimately share the same understanding of life\'s purpose and therefore share the same priorities, values and goals.  5. You choose the wrong person because you get involved sexually too quickly. Sexual involvement before the commitment of marriage can be a big problem because it often precludes a fully honest exploration of important issues. Sexual involvement tends to cloud one\'s mind. And a clouded mind is not inclined to make good decisions. Of all the studies done on divorce, sexual incompatibility is never cited as a main factor. It is not necessary to take a \"test drive\" in order to find out if a couple is sexually compatible. If you do your homework and make sure you are intellectually and emotionally compatible, you don\'t have to worry about sexualcompatibility. Of all the studies done on divorce, sexual incompatibility is never citedas a main reason why people divorce.  6. You pick the wrong person because you do not have a deeper emotional connection with this person. To evaluate whether you have a deeper emotional connection or not, ask: \"Do I respect and admire this person?\" This does not mean, \"Am I impressed by this person?\" We are impressed by a Mercedes. We do not respect someone because they own a Mercedes. You should be impressed by qualities of creativity, loyalty, determination, etc. Also ask: \"Do I trust this person?\" This also means, \"Is he/she emotionally stable? Do I feel I can rely on him/her?  7. You pick the wrong person because you choose someone with whom you don\'t feel emotionally safe. Ask yourself the following questions: Do I feel calm, peaceful and relaxed with this person? Can I fully express myself with this person? Does this person make me feel good about myself? Do you have a really close friend who does make you feel this way? Make sure the person you marry makes you feel the same way! Are you afraid of this person in any way? You should not feel you need to monitor what you say because you are afraid of how the other person will view it. If you\'re afraid to express your feelings and opinions openly, there\'s a problem with the relationship. Be on the look out for someone who is always trying to change you. Another aspect of feeling safe is that you don\'t feel the other person is trying to control you. Controlling behaviors are a sign of an abusive person. Be on the look out for someone who is always trying to change you. There\'s a big difference between \"controlling\" and \"making suggestions.\" A suggestion is made for your benefit; a control statement is made for their benefit.  8. You pick the wrong person because you don\'t put everything on the table. Anything that bothers you about the relationship must be brought up for discussion. Bringing up the uncomfortable stuff is the only way. To evaluate how well the two of you communicate, negotiate, and work together. Over the course of a lifetime, difficulties will inevitably arise. You need to know now, before making a commitment: Can you resolve your differences and find compromises that work for both of you? Never be afraid to let the person know what bothers you. This is also a way for you to test how vulnerable you can be with this person. If you can\'t be vulnerable, then you can\'t be intimate. The two go hand in hand.  9. You pick the wrong person because you use the relationship to escape from personal problems and unhappiness. If you are unhappy and single,you\'ll probably be unhappy and married, too. Marriage does not fix personal, psychological and emotional problems. If anything, marriage will exacerbate them. If you are not happy with yourself and your life, take responsibility to fix it now while you are single. You\'ll feel better, and your future spouse will thank you.  10. You pick the wrong person because he/she is involved in a triangle. To be \"triangulated\" means a person is emotionally dependent on someone or something else while trying to develop another relationship. A person who hasn\'t separated from his or her parents is the classic example of triangulation. People can also be triangulated with things as well, such as work, drugs, Internet, hobbies, sports or money. Be careful that you and your partner are free of triangles. The person caught in the triangle cannot be fully emotionally available to you. You will not be their number one priority. And that\'s no basis for a marriage.   '),(2169,3509206,'31,May,2004','2004-05-31 00:00:00','   Well, I think this is going to be interesting. I hate hand writing in journals, and I recently just found this amazing site that all I have to do is type, and well, out comes this journal. It even has spell check, wow, can it get even easier.   Some may be wondering, \"why make a journal, what\'s so important for you to say and for us to listen to\". Exactly my point. Let\'s just say it\'s a fresh start to my self realization. Yes, I haven\'t been exactly finding the real me in others, so I though, why, a good start is to stop taking what other people say too seriously and just focus on me. I have much to say, and I hope that my procrastination doesn\'t result in this journal not being written in for days on end.   Well, what can I start with. School is killing me at the moment. I hate work. It\'s a pet peeve of mine. But I\'ve got to live with it. The only thing that\'s keeping me going is knowing that it\'s all over in two weeks, excluding exam days, and summer is oh so very near. Cheerleading, what can I say about that. It\'s fun. The people there are great. It\'s an over all great experience. But what if you wake up in the morning dreading going to practice that day. That\'s exactly what I\'m going through. I haven\'t found an activity that I could stick with for a longer period of time. I\'ve tried ballet, ballroom, polish dancing, track, cheerleading, and I feel that I\'m still missing out on something, and that none of these is really my passion. So I\'ve decided that two years in cheerleading is enough for me. Maybe I\'ll go back to track next year. We\'ll see how it goes.   Well that\'s enough writing for tonight. Don\'t worry there will be more to come.  Take Care. Don\'t drink and drive. :) Tootles                                   -ela '),(2170,3509206,'24,June,2004','2004-06-24 00:00:00','   Welcome to my summer.........What is it exactly that I\'m doing right now. NOTHING. LOL. Well relaxing before camp. Camp is in a week. I\'m excited. It\'s real fun. You don\'t have to believe me, and you may think as if going to camp is a little odd. It\'s not. It\'s way more than that. Well I\'m not going to spend all this time explaining about it. So I\'ll change subjects.  My birthday is in 6 days. I\'m turning 17. Wow, how slow getting older is. Anyway.....I don\'t mind. More time to have a different kind of fun. I don\'t like birthdays, well my birthdays. It\'s a normal day. I like going to other peoples b-day parties but not mine. I don\'t now why. But I just like it that way. And I\'m not sad, I\'m happy, so it\'s all good.  Well, i have to get back to doing nothing, I think the laundry is done. Got to check on that.  Tootles           -Ela '),(2171,3509206,'21,June,2004','2004-06-21 00:00:00','   Well yesterday and today are much better days. I guess you get one of those days once in a while where u think you\'re going to lose it and nothing can help.   I got a nice t-shirt tan on one side of my arms at the picnic at our church. I wore my new skirt that I made myself. It looked sooooooooo pretty. LOL. I liked it.   well I think it\'s about time to shave my legs. They\'ve been hairy for too long. But it\'s not like they\'re so visible and all. They aren\'t. LOL. Well its a nice day. I\'m off to do stuff!!!!!!!!!!!!!!!!!!!  Tootles. '),(2172,3509206,'19,June,2004','2004-06-19 00:00:00','   Holding it in, going to lose control any second, got to seem strong, got to seem tough.  How much longer will this shit last, how much more of this am I going to have to hold back.  Have to pretend it doesn\'t hurt me, while inside I weep.  At times it surfaces, I let out my weaknesses, but no, never again, I am not like that.  I am not who you think I am. I am not what you expect me to be. I will not give in to your schemes. I will not become just like you. I forgot who I really am. I forgot what it feels like to be happy. I don\'t know what I like anymore. I don\'t know what makes my happy.  I want to change me, I don\'t like who I am. I want to become a different person, different pleases me. I will hold in what I can. I will never forget how you did me wrong. I am who I am. I don\'t care what you think. '),(2173,3509206,'18,June,2004','2004-06-18 00:00:00','   My father and I ate dinner together alone at the kitchen table. A very awkward moment in both our lives. You see my mother hadn\'t get come home from work and my brother hadn\'t come home from his residence either. They\'re both late. And I always make dinner on time so well, my father and I ate. It\'s not that my father and I don\'t get along it\'s just that we really don\'t have anything to talk about. It\'s either him complaining at me for watching TV when he wants to, or me complaining to him why he leaves the kitchen a mess. Well, we sat in silence. In uttermost silence. He usually takes his dinner plate and ate in front of the TV, but no, not today. So sat looking at my plate, him at his. Eventually he sat beside the TV. It was the longest 10 min of my life. I ate almost nothing so slowly that it took forever. Well my mother just arrived, I have to talk to her about her day, and o crap! I was supposed to make a list of groceries, o well. I\'ll do it quickly now.  Tootles '),(2174,3509206,'16,June,2004','2004-06-16 00:00:00','   Well today I woke up nice and early for no apparent reason. There is no school today, no early appointments, no flippin\' reason at all. Yet I am totally awake and functioning well. I woke up at 5 am thought, \" it\'s my first day of summer vacation go back to sleep\". Then every half hour I woke up, and made myself crawl back into bed. At 7 I finally decided there is no point in this all and got up. My father was still sleeping my mother already left to work. With nothing else to do I went on msn. Then checked out the forum. I found myself watching the college years of saved by the bell. Even though I\'ve seen all the episodes it still interest me. Anywho, then I started watching this movie on the superstation, don\'t remember what it\'s called but it\'s an ok movie. I started watching Live with Regis and Kelly. The shows been better before. Now it\'s so repetitive, it bores me. They should get like Brad Pitt to go on the show. Then I\'d watch it.   Well, I have a few things on my agenda today. I have to wash my hair, do some grocery shopping for tonight\'s dinner which I am making. Then possibly do some laundry, vacuum, clan my mess, wash my bathroom, and the list goes on. I\'ll be happy if I get the first two done. LOL.   Well I think it\'s time for that shower. O man I hate walking to the store. There\'s so much construction going on and all the guys working there start screaming something in Spanish to me. I just ignore them. I feel so violated. LOL. I need to learn Spanish so I can yell something mean back to them. LOL. And they\'re all ugly so yeah.  well I\'m off...............Tootles '),(2175,3509206,'16,June,2004','2004-06-16 00:00:00','   I\'m tired of all of this, you know when you get into one of those hopeless moods where you don\'t want to do anything just sit at home, but you also don\'t want to sit at home. I hate it, despise it. LOL. All I can do is laugh. Crying is for the weak. I don\'t think I have reached that point yet. I am lost. I don\'t know who to listen to. Who to believe. And I have lost friends because of that before, but they don\'t understand that I care about them, that I really don\'t mean what I did. I cared so much about them that when I heard something about them I started to doubt that that friend was really my friend, maybe it was a game they were playing, just to make fun of me in the end. I think like that a lot. I think low of me. But that will never change. You need to understand that. That that\'s who I am. I\'m sorry about everything. But I feel that I\'m the one always saying sorry. I have feelings too. But you don\'t seem to see that. What can I do, but think, I need to stop thinking. I always tell everyone that but I should start listening to myself for once. I need to take under consideration about my feelings. But I\'m scared that others will think I\'m conceded. When really I\'m not. I know I shouldn\'t care what others think. And I also tell other people that too. But it\'s hard. It\'s hard not to care. Very hard not to care. '),(2176,3509206,'15,June,2004','2004-06-15 00:00:00','   Sweet Surrender - Sarah McLachlan   it doesn\'t mean much it doesn\'t mean anything at all the life I\'ve left behind me is a cold room I\'ve crossed the last line from where I can\'t return where every step I took in faith betrayed me and led me from my home  and sweet sweet surrender is all that I have to give  you take me in no questions asked you strip away the ugliness that surrounds me are you an angel am I already that gone I only hope that I won\'t disappoint you when I\'m down here on my knees  and sweet sweet sweet surrender is all that I have to give  sweet sweet sweet surrender is all that I have to give  and I don\'t understand by the touch of your hand I would be the one to fall  I miss the little things oh I miss everything  it doesn\'t mean much it doesn\'t mean anything at all the life I left behind me is a cold room '),(2177,3509206,'14,June,2004','2004-06-14 00:00:00','   I need you right now.  Right this minute.  I have no idea who you are.  I don\'t know if I already have met you.  Maybe we\'re friends.  Maybe there\'s something more between us.  I just know that I\'m alone.  Some thing is missing and at this exact moment I\'m falling apart inside.   Who are you? Where can I find you? Help me out please? I know you are out there. You\'re watching right this minute.  And as you read this, you know there\'s nothing you can do. Time will explain everything. Well, I want it to explain now!  Why is this taking longer than I hoped it would.  Why can\'t you see that I like you. I don;t want to be blunt about it. But I wish you\'d give some kind of hint.  Maybe you\'re not the one.  If not, then I\'m happy that we\'re just they way we are. But if there is more..... I want to know.  But only, when the time is right.  And it seems as if I can\'t wait any longer, time for me is right now. And if not for you.  Sorry. I waited out as long as I could. But there\'s still a chance.  I\'ll be waiting for you. For the one, who I don\'t know. '),(2178,3509206,'13,June,2004','2004-06-13 00:00:00','   The sound, the known anger, the blinding dramatic light. Last night was scary as hell.   Ok so it may sound a little too over dramatic, but I was surely scared as hell. I dislike thunderstorms. To begin with a quick flash of light is never good. It\'s like turning around to see a train a second before it hits you. Or like they say, when u die you\'ll see a bright light. You see that lighting is not a good thing. And thunder. It\'s like a loud crashing noise, and that means destruction. As you can see this whole thunderstorm thing is so awful. I like rain. I love rain. But just Rain. Alone, with no extras.   Ok, well I\'m sleeping when suddenly I wake up because of the storm. I look at my clock it\'s around 12:30 ish. The rain is falling in my room, so I go across to the window and close my window. The second I am doing that I see a flash of light, then the sound of thunder and then all the street lights go out. I glance at my clock and it\'s flashing. My heart is beating at it\'s fastest. I\'m a little chicken when it comes to these things. I have a fear of getting struck by lightning. I try to convince my self that, if I do get struck then it\'s ok. It\'s not my fault I couldn\'t prevent it. And dying is ok. But last night I was scared to death. I considered going to my parents room and crawling into bed with them as I had done many many years ago. But I was too afraid of opening my door and crossing the hallway. So I jumped into bed, put the covers over my face and ears, and a while later found myself fast asleep. I woke up in the morning tired as hell and oh so hot. I shouldn\'t have covered myself with all the blankets and closing the window wasn\'t a good idea.   Well, that\'s all for today folks. Now you all know one of my fears. And I tell ya, there\'s way more to come.   Tootles         -Ela '),(2179,3509206,'13,June,2004','2004-06-13 00:00:00','   So I lied, I said I\'d write tomorrow but I didn\'t. Too much to do, and I don\'t always feel like writing. So Thursdays athletic banquet was awesome. I had heaps of fun. Wish it would never end. Our table started out being Andrea, Liz and me. Then a group of guys joined us, because well, we\'re just so great to get along with, LOL.   Saturday was also really fun. We had our bieg for scouts. Then the picnic afterwards, where I got to see the guy that came from Australia that I haven\'t seen for almost a year now. It was great to hear his Australian accent, but he\'s lost some of it. That\'s too bad. The guys played soccer, us girls had a water fight, then we played volleyball. My arms are nice and burning at this moment, but they\'re not as bad as my face. It won\'t be that bad tomorrow, I hope. Hehehehe. Then I slept over at Gosia and Ania\'s house. We rented American Beauty. It\'s a twisted movie. I guess if I saw it again maybe I\'d understand it a bit more. But the young crazy guy who films everything in the movie is hot. In a mysterious kind of way.   Today we had Boze Cialo at our church. So more sun and walking around the neighborhood in our scouts uniform. It not that bad. After that my family and I went to my cousin\'s house for her birthday get together. She turned 5 last week. And see gets hyper and more hyper by the day. I ended up taking a nap on my aunts couch. I don\'t know why I was so tired. I got like 10 hours of sleep last night. I\'m going to blame it on all the sun. hehehe. Yes, that\'s a good excuse.   And now I\'m trying to study for my first exam. Drama. It\'s a written exam and well, I am not good at that. Oh well, I\'ll try my best. But I hate studying. We\'ll wee how it goes.   I\'ll write more when I got time, or something interesting happens.   Tootles     -Ela '),(2180,3509206,'10,June,2004','2004-06-10 00:00:00','   Kind of busy now..............I have lots to say..........Will write more tomorrow...........................  Tootles               -Ela '),(2181,3509206,'09,June,2004','2004-06-09 00:00:00','   Well, before this post I was improvising a poem. I got fed up with it because it sounded corny, even though it was from the heart. I gave up writing poems for a while. And now I remember why. hahah. oooooo.....Good times.  Well the heat has been awful today. It leaves me weak, unable to breathe at times (just like the song by Maroon 5 - harder to breathe) it\'s a good song. But the sun always leaves a smile on my face. We\'re learning about different religions this semester. Mostly about spirituality. I feel that in my life, especially at this moment, I am lacking in that section. I don\'t have time. I really don\'t. Does that make me a bad person. I sure hope not. It\'s just that I really don\'t think I can calm myself down to the point where I am totally silent, with no thoughts in my head, or anything to bother me for that matter. Something always comes up with is urgent, and well that is my first priority. Maybe later it will change. We\'ll just have to see.   I am currently memorizing a monologue for drama class. It\'s from the movie and book \"white oleander\". It\'s real good. But it\'s deep stuff. hahah. I\'d like to end up seeing this movie. Maybe with my mom. It seems like one of those mother and daughter movies. oooooooooooo, u know what I haven\'t had in a long time. A movie night. I need one of those. Hmmmmmmmm, what movies, so many that I want to see but I am completely blank. Oh well. My memory isn\'t as good as it used to be. Wait, a sec. It was never that good to begin with. hahaha.   Ok well I got to get back to memorizing it, and also the French project and studying for the science and fashion in class exam. yay!!!!!!!  Tootles     -Ela '),(2182,3509206,'08,June,2004','2004-06-08 00:00:00','   Yeah I found this funny quote today:  New!  Improved!  Instant ass hole... Just add alcohol!  Ooooooooo, that was good. It\'s not always true. But it can give you a good laugh.   Well today was a long day. It has been, and will continue to be the longest week of school. And the reason behind it is, that it\'s the last week of school before exams and well, it just drags on. Today was the last day of scouts. It\'s about time it\'s over. It\'s been going on for too long. But it surely has been worth while.   The heat is coming in. And life with out A.C is going to kill me. Oh well at least I got a nice and cool basement to turn to. Except for the fact that it\'s unfinished, dusty, and scares the heck out of me. I\'ll get used to it. Not to worry.  Well, I officially need a break. Just to get away for one day, or longer. Maybe go to the beach, or somewhere nice. Away from my family. Somewhere with friends. After exams and before camp. That\'s the perfect time. I need time off from thinking. Yet, there will always continue to be a million things going on at the same time and I\'ll get so lazy that\'s I\'ll probably end up finishing none of them. Well, that\'s me for you. One big procrastinator. I try not to be. But it gets the best of us, sometimes.   Ok, I have so much to do and maybe I\'ll end up doing something.   Tootles        -Ela '),(2183,3509206,'07,June,2004','2004-06-07 00:00:00','   It\'s been a rough two days.   Do some people not understand that if they act a certain way while they are drunk that they should not drink as much or be more aware of their actions. I know they mean well, and they aren\'t really proud of what they have done, but if it repeats itself more than once then you seriously need to rethink a few things. You are not only hurting yourself but others as well. If you were in your right minds then maybe you wouldn\'t have done those things again. Think twice people. I tried to help as much as I could that night. I did my best, but you still screwed things up. And it\'s up to those who you hurt to make things right again. It\'s only those people who can forgive you for what you\'ve done. I am disappointed that you could\'ve acted those ways. And I know you didn\'t mean for it to happen, but you did partake in it so it is your faults. But that is all I\'m going to say. All I know is that if I were you, I\'d think things over way more than twice next time when you\'re in the same situation.   And man, I just wish my family would see things my way for once, They just like to see things their way which makes them look right all the time. Well, they\'re killing me inside. They don\'t understand that what they say, what they do, their little \"innocent\" comments really beat me up. They can\'t see that it hurts me but make up poor excuses like, you\'re just tired and don\'t know what you\'re talking about, or, don\'t let your situation with friends right this moment interfere with your relationships at home. When in reality there is nothing wrong with me. It\'s them. They\'re against me or something. Then a second later they act as if everything is ok, well it\'s not. And I will not act as if everything is ok. I will not be mean, but I surely do not forget when someone is being unfair to me. It makes me mad, and really I am lost. I don\'t know what to do. I\'d like to run away but there\'s no where to go, and when I\'d come back they\'d just get mad at me for trying to runaway from my problems when really I\'m trying to runaway way from the people who create them.   Well, I feel better now that I\'ve vented.   Tootles         -Ela '),(2184,3509206,'06,June,2004','2004-06-06 00:00:00','   So I haven\'t written in here for a few days now, I couldn\'t find the time.  We\'ll start with  Friday: Garbage picking day. It was fun. Well except for the whole picking up garbage part. Nah. It was good. Found a lot of interesting stuff. Tires, clothes, etc. And Trojan did win against Durex. Hahaha. My body hurt so much from picking up all that stuff then walking to and from church. Yup, it just was a fun filled day.   Saturday: Started off by walking to school for the cheerleading competition. We got 2nd place, which is really good. Came home. Got all ready for Lukes party. Went to Lukes party. And well what can I say. Maybe I can put into words everything that I want to say later. I can\'t seem to find the words right this instant.   Sunday: Came home at 2:30 am. Woke up at 6:50 am. Went to church nice and early. Came home and watched TV until 1pm. I just finished watching an old movie on TV. It\'s called \"Three came home\". A very, very good movie. I\'m fond of old, black and white films. And this was exactly just it. About how the Japanese took over. A ladies story of the whole thing. The rough times they had to go through. Trying to keep herself and her little boy alive. Then at the end Finally being able to see her husband after many years, and thinking this whole time that he was dead. So emotional. I usually can hold it in. But this time I just bawled. Couldn\'t hold it in. I don\'t like being so emotional. People assume you are weak, and fragile. While I\'d rather be strong and tough.   Well, I\'ll write more later.   Tootles     -Ela '),(2185,3509206,'03,June,2004','2004-06-03 00:00:00','   Well, it sure had been a crazy filled day. I\'ve lived it to the max. And not even a sec. Was I bored.  It started off with waking up after like the longest sleep ever, and I was SO tired. It was crazy. Then came school. Fire drill. Lunch outside in the nice sunshine. Then lounging around in fashion class. And then came THE LAST DAY OF CHEERLEADING PRACTICE. O man. I haven\'t worked my butt off that much in so long. Even though I dreaded coming to practice that day, it sure was not that bad. But my mind is made up, I\'m not going to change my decision.  Well at the end of practice our couch is like there\'s a dinner thing tonight at 6. We want like the 5 of you to come. So we said sure. I got home at 5:30, dressed up nicely, got to the hall around 6:15. And what a fun night it was. So many jokes.....Well there are always a lot of jokes. First on the drive there these construction guys tried to pick us up in our car. Ewwwwwwwwww. They were old. Then at the hall an old guy wanted yola to buy him a drink. And he kept stalking her. It was a dinner for all these athletes in mississauga. It was cool. Free food. That\'s all I can say.  And here I am. At 11:38 pm. Having a bunch of stuff due tomorrow. I don\'t think I\'ll ever finish. And then I got to shower. Bah! hahahaha. All I can do is laugh, because I\'m so tired and I don\'t think I can stay awake any longer.   I haven\'t had time to think at all today. So now is my thinking time before I go to bed. Because if I don\'t think now, I\'ll go to bed and I won\'t fall asleep because I\'ll be thinking too much. So first of all, I have to remember to live my life for me. I have to plan things out and not worry if someone else is doing something that same day and I might want to go there, even if it\'s not for sure. Because if I\'ll do that I\'ll end up staying home doing nothing. So I need to plan things according to me. You may not understand, but I don\'t know how to explain, it\'s all clear up in my head, but when it comes to typing it out, yeah, not so good at that.   Okej, I need to get other stuff done, so I\'m done writing for tonight.  tootles           -ela '),(2186,3509206,'02,June,2004','2004-06-02 00:00:00','   I had the best nap ever! As I lay in my bed the sun shone threw my window and warmed me up. I soon found myself in deep sleep, and had no worries at all. Not having a certain time to wake up or anything. Then Waking up after a few hours, feeling more awake than ever. And I sit here now, in front of my computer slowly coming back into reality, and still in my pajamas.   Today has been a very tiring day. Not that I did so much today, just the fact that every little thing tired me out. And some people noticed and knew to stay away from me and give me some space. I was also very emotional. I found myself during science class tear eyed for no reason at all. Possibly because of everything right now is overwhelming me, maybe.   But I really appreciate those friends who you know can always be there to joke things up. They don\'t do it on purpose. They\'re just naturally funny. And that\'s the best you can get. A few jokes before heading home, made my trip home more lively.   This is one thing I hate. When you know that someone wants to talk to you, but they made a promise to themselves that they will wait until you speak up first, but clearly, when you do talk to them they hardly seem interested to talk to you. But you know that they\'re just trying to be the better person and stick to what they say, when the truth is, that the better person would just confront the other person and talk things through.   Well, i just needed to get that out. And i feel much better. I\'m finished writing for now. if more comes to mind, you\'ll be the first to know.   Tootles -ela '),(2187,3509206,'01,June,2004','2004-06-01 00:00:00','   All these things to do, in such a short period of time. I doubt if I\'ll get the energy to get anything done. The weather sure plays a big role in everything. Not only is it gloomy outside, but I get these massive headaches at the back of my head that a mandatory nap on the bed is in order. If not, I think I\'d die from the pain. Well not literally.   Bah........This is when I\'d really like to crawl into a ball and disappear, just until everything gets sorted out. So many things running through my head. Holy, I never knew I could store that much stuff up there.   I need to do something fun, and soon. Well maybe this weekend will fulfill all of that. Being with friends is always fun. A road trip would be cool. Now if only I\'d get my G2 before the summer. But no, I\'m going to have to wait until October.   The time is coming when soon I\'ll be another year older. So much has happened in the sixteenth year of my life. I\'ve learned so much. I went through a lot. It surely has been the best year of my life so far. When I come to think of it, I am real young. I haven\'t lived half of my life yet. I\'m really excited to get older and see what more knowledge I can acquire.   Well as for now, I\'m going to go with the flow. And try to not lose my temper, be positive, and smile. Got to keep a damn smile on my face. Why? At least I\'ll make others feel happy. But that\'s going to be be tough. But I can do it! Hehehe :)  Tootles.....Until I write again..........ela '),(2188,3509206,'01,June,2004','2004-06-01 00:00:00','   That fresh smell of falling rain. The sound it makes on the rooftops. Man I love the feeling that comes along with it. It makes me just want to curl up and sleep. Well, I do need my sleep. And being tired is more common to me. I don\'t know why. I don\'t have an explanation for the way I\'ve been feeling for the past few days.    >  Now back to what I was talking about. I\'m not stressed out, I\'m not going into depression. Am I talking crazy here. I\'ve been thinking a lot. But it doesn\'t seem a lot. I don\'t know how to make sense of it all. I need a vacation to get away from this all for a while. Meet new people. Experience new things. That\'s what I strive for. Yet, my shyness always interferes. I can never get the courage to do things I\'d love to do. Maybe one day that\'ll all change.   It\'s thundering right now. I hate thunder. Makes me feel all helpless inside. Won\'t it just go away. That\'s one of my fears. I also hate the dark. I can\'t see anything, and that scares me. Wow this thunder is loud. Make it stop! Sorry just getting a little paranoid right now. You know being home alone and all. Not knowing where my father is at the moment. Loneliness is filling up my house along with me inside. But that\'s what MSN is for. To talk to friends and hope they\'ll be there by your side.   Well, now that it\'s raining there goes the whole rollerblading with scouts idea. Seems like it was never meant to be.   Well, what more can I write. It seems that every guy I like either likes someone else, or just isn\'t interested, And most guys that like me, I just don\'t feel anything there. Or, maybe I\'m just scared to feel. I\'m not to sure. But I\'ve learned not to dwell too much in that area, for it only takes away for all the happy times I could be missing out on. If there really are happy times in between, I\'m not sure. But that\'s what people have been telling me.   Well, I might write some more later on tonight. After scouts. Maybe something interesting will happen.   Tootles        -ela:) '),(2189,3596192,'10,June,2004','2004-06-10 00:00:00','    Plans for the garden are taking shape and growing! I have spent hours working in the space - getting to know the space, really. Now I have committed my ideas to paper, staked out the areas, and know the features I am going to install. The place will be pretty \'raw\' the rest of this year but the construction will be in place and ready for next year to be started on time.   I am adding a small pond for effect and for the wildlife. The herb garden will be distinct from the vegetable garden. The back bed of the vegetable garden (4\'x40\') will be set up for cutting flowers. Despite the mission of Frog Pond to be \"native\" in its plantings, a little fancy flowers make the place look nice!   New animals and birds were seen this week. Cedar wax-wings and killdeer are new to my sightings here. The hummingbirds are in full action at the feeder and the bird boxes are fledging nests. Summer is upon us at Frog Pond!   The first tomatoes are set on the vines in the garden and the beans are blossomed. The pumpkin vines are growing well and all in all things seem to be doing fine. A little rain this week would be nice.   '),(2190,3596192,'10,June,2004','2004-06-10 00:00:00','    Summer is here and is off to a roaring start. Everywhere there are plants growing and blooming, birds singing, and wildlife getting on with the business of life. The Father\'s Day hummingbird feeder went up yesterday and within 10 minutes, the Ruby-throated hummingbirds had found it.   There is not enough time to blog everything that goes through my mind about Frog Pond. Five minutes of sitting on the porch swing provides a lifetime of material to think about and write about. Fortunately, it is a lot more fun to get out in to the land and work with it as opposed to write about it! Maybe when the land sleeps this winter I can catch up on my writing...   Every day I fall more in love with this place. The pace and the connection with the land is so much more interesting that life in the city. Maybe I will never go back...   '),(2191,3596192,'10,June,2004','2004-06-10 00:00:00','    There are 15 bluebird boxes up at Frog Pond and today we took a long walk surveying them all. We found a little bit of everything; wrens in a couple boxes, tree swallows in others, and even some bluebirds! Most of the nests had eggs or chicks in them - it is the prime nesting season for most song birds.   One of the tree swallows was calm enough that we were able to pick her up off the nest and hold her. She was not thrilled with being handled but we were successful in getting her back on the nest without incident.   Need to create a little log of the boxes, conditions, and nests. One more thing for the web site!   '),(2192,3596192,'10,June,2004','2004-06-10 00:00:00','    One of the primary chores in front of us is to catalog the current flora and fauna at Frog Pond. We need to create some basic bio-survey in order to have a \"baseline\" of data. There is no need for this to be a scientific survey, but I think we will collect our notes in a variety of ways and keep reasonably accurate records.   Organizing the data will be a challenge. I believe we can start small databases; one for flora; one for birds; one for other fauna. The bird information is different enough from other data that it might merit its own database. On the other hand, if we begin to catalog insect life, that might create a new database need. All of this needs put on to the web site.   '),(2193,3596192,'10,June,2004','2004-06-10 00:00:00','    This is the opening post to this web log. It is our hope that the ideas here are of some value to the reader and that interested readers will share their experiences, questions, and feelings. Frog Pond is part of the world we all live in and this blog just might be a way to extend the community of Frog Pond beyond the traditional boundaries of geography.   '),(2194,3816397,'12,July,2004','2004-07-12 00:00:00','    The sad saga of the broken treadmill has led me to some self-discovery.    First, I\'m very pleased that I don\'t have to pay someone to install the parts into it.  Sears would charge a pretty penny to send a repairman over here, and I feel confident that they would not hesitate to charge more than once in this saga.  So that\'s a victory in itself.  Second, I\'m proud of myself and my persistence in pursuing the repair project through all these many calls.  It\'s annoying but I\'m not letting it stop me.  Third, I\'ve discovered that I really miss my treadmill.  I miss being able to walk after work each day.  Sure, I could walk outside.  Of course, it\'s presently 90 degrees out there.  And that\'s been the temperature every day lately.  So I *could* walk outside but I\'d get a better walk on the treadmill.  Fourth, in an effort not to be stalled in my fitness goals I persuaded my partner to join the local rec center with me.  I think she\'s shocked that I would want to do that, since I\'ve never been very athletic.  The thing is, I know that if I\'m to become healthy I have to become more active.  And the local rec center offers a lot of options for that.  Plus it\'s reasonably priced and what we pay helps support the center for everyone.  And finally, I got creative today and looked into my college catalog.  We live very near a community college campus and I\'ve been taking classes there for a couple of years.  It turns out that they have a fitness center on campus.  For the cost of one credit hour I can use the facility from August 23 through December 12, and taking a Physical Education class includes what amounts to guidance by a physical trainer!   So GO ME!!   '),(2195,3816397,'01,July,2004','2004-07-01 00:00:00','    I\'ve been struggling with my treadmill for the past three weeks.  I love my treadmill, don\'t get me wrong.  But it\'s broken.  I\'ve called and called for help.  The treadmill is a NordicTrack C1800.  Not very top of the line, but not entry-level.  Including delivery it was around $900 when we bought it, but I digress. We bought it in March last year.  So it\'s out of warranty for labor, but parts are still covered.  First call to tech support I was pleased to learn they\'d let me install the replacement part myself and save the labor charges.  The tech put me through the diagnostics and concluded that the problem was a bad motor controller board.  This seemed reasonable based on the symptoms I had observed.  He ordered one to be sent to me, which I duly received 3 business days later.  This was on the weekend of the fourth of July. I installed the part.  To my dismay the problem persisted.  I called again on July 6.  Next diagnosis was a failed power board.  One was ordered.  I received it within 3 business days.  Installed it.  And the problem persists.   I called again this afternoon.  At this point we have diagnosed the likely problem to be a failed motor, which may be feeding current back to the motor controller board and zorching something.  I am now awaiting the delivery of a motor and a motor controller board.  And today\'s tech had the courtesy to upgrade the shipping to UPS Air so I should have it in hand on Wednesday.  And then we shall see if the problem is fixed.   '),(2196,3816397,'01,July,2004','2004-07-01 00:00:00','    through.    Someone recently said \"my family is so whitebread\", meaning \"middle-class and normal\".  Mine\'s more bananas and nuts.   '),(2197,3816397,'01,July,2004','2004-07-01 00:00:00','    urlLink    Rear wheel of bike mounted on Mag Trainer.  Notice the slick tire...very quiet on the trainer.&nbsp; urlLink   '),(2198,3816397,'01,July,2004','2004-07-01 00:00:00','    urlLink    Bike on Blackburn Mag Trainer with block under front wheel.&nbsp; urlLink   '),(2199,3816397,'01,July,2004','2004-07-01 00:00:00','    Ok, got the picture to work in my profile. Consider that to be a \"before\" shot.  \"After\" shot will be coming later.  Meanwhile, what\'s to tell? Well, let\'s see... Oh, yeah, I officially have a home on the internet now!  Woohoo!   '),(2200,3816397,'01,July,2004','2004-07-01 00:00:00','    the first blast of hot air from the whoopicush.  Just a brief note to get this started.  I\'ll have more to say later.   '),(2201,2680773,'22,September,2003','2003-09-22 00:00:00','   My short story  urlLink \"Beggar\'s Banquet\"  is over at Dmitry\'s  urlLink XPEHb  journal.  The story is one of my lit-fic in sci-fi pieces but is more of an experiment in broken imagery than anything plot-wise.  During my junior and senior years of college, I was playing a lot with poetic flow in long prose settings.  Ultimately, the stories became  too  removed from setting and structure to work well (\"Journey to the Heavens\", a meditation on Eastern philosophy through the American \"space monkey\" launch is probably a good example of its failure), but the gradual collapse of stream in \"Beggar\'s Banquet\" probably helps pull a reader along instead of throwing him/her into a spastic firing of meta-electrons.  Anyway, check it out.  Dmitry is running a nice show over there, and there\'s plenty of substance if you get fed up with me.  I mean in no way to demean my fellow XPEHb contributors, but Dmitry\'s photos make any pain you might experience from our rhetoric worthwhile.  urlLink Full Post and Comments    '),(2202,2680773,'27,October,2003','2003-10-27 00:00:00','   Less volatile indeed!  One has to wonder whether Sunday\'s Bears win could have been achieved with Kordell Stewart, but one needn\'t mill over it, because a win is a win is a win.  The new Shins CD,  Chutes Too Narrow , is excellent by the way.  Critics are claiming  O Inverted World  marked the beginning of concious thought for humanity, though, and that this is just a nice follow-up.  I don\'t remember the universe getting knocked to the floor by the last album, but the \"sophomore slump\" is like juicy red meat in criticism.  I don\'t think there\'s any REAL \"sophomore slump\", the same kind of approaches used following a successful first are the same as those used following a second, third, fourth, etc...  One magazine I read said, essentially, \"some bands clone their original sound, some bands rebel against their original sound, and others mature/evolve\".  Who are we trying to kid here?    It is no great feat nor surprise that bands must have their second record sound like  something .  There is no mysterious trend in how bands make their decision, they just do it and the results are as numerous as second albums themselves.  The whole idea of the slump needs to be abandoned as a viable introduction/structure for a piece of criticism.  It\'s an introduction much too steeped in the business of music.  Though we\'ve been trained to care about the marketability of music, I don\'t think the album itself cares too much.  urlLink Full Post and Comments    '),(2203,2680773,'23,October,2003','2003-10-23 00:00:00','   urlLink Networks have noticed a mysterious disappearance of 18-34 year old males watching television this season .  Maybe because we are downloading all their shows off the internet.  urlLink Full Post and Comments    '),(2204,2680773,'22,October,2003','2003-10-22 00:00:00','   The question for the day is:   Is a mall a good place to hole up in case of a zombie outbreak?  George Romero\'s  Dawn of the Dead  and Max Brooks\' seminal  The Zombie Survival Guide  seem to give us contradictory answers.  In Romero\'s film, a group of four survivors find great safety in a shopping mall after securing the structure.  All of their needs (aside from social relations) are accounted for and the mall is a large, relatively controllable environment...  once it is secure.  That is the important thing to emphasize.  In  Dawn of the Dead  only two mall residents survive the process of shutting down outsider access, a sign that the task is extremely dangerous.  Of course, one can argue that if early situations had been handled more wisely, the deaths could have been avoided.  Romero\'s mall is strangely vacant aside from a spattering of zombies.  Brooks\' however sets a more realistic stage of the possible overcrowding of a mall.  In the early stages of an outbreak (and continuing indefinitely), malls will be points of attraction for both zombies and people.  THIS MEANS DANGER!  Even if you can arrange the initial arrivals into an efficient military, constant influx of outsiders and zombies would present continual problems.  In order for Romero\'s mall to be as vacant as it was when the heroes arrived, how many regular people would have already died at the location?  This certainly doesn\'t say much for your odds of survival.  Still...  Are the potential benefits greater than the dangers?  Large parking lots give you a good scope of approaching undead, there is a large quantity variety of inherent supplies and the attraction of people means an ability to establish new trade, society and culture.  If you\'ve got the answer, weigh in at the  urlLink forums .  urlLink Full Post and Comments    '),(2205,2680773,'21,October,2003','2003-10-21 00:00:00','   No more Tuesday Morning Quarterback for awhile     beeFymojo:   espn.com just ousted Gregg Easterbrook (one of their best non-sports-people-but-still-serious-about-sports columnists) presumably because of something he miswrote about Jews off-site  Moogle:  Something he mis-wrote?   beeFymojo:   [Edited for length: text is here in the  urlLink   10.16.2003 entry .   beeFymojo:   as he explains it, he didn\'t mean to bring in the jews-love-the-money stereotype   beeFymojo:   meant more to go for, how can the victims of horrible violence also market it  Moogle:  I got more of a \"Jews are corrupting us\" sort of thing.  Moogle:  Yeah, I can see the glimmer of that argument lying in there.   beeFymojo:  obviously, what he wrote was bad  Moogle:  But, yeah, this is why you have people look over your work.   beeFymojo:   but that is part of the problem with the quick-update philosophy of the web right now   beeFymojo:   the blog mentality, so to speak  Moogle:  Oh well, he\'ll get his foot surgically removed from his mouth, and get a job somewhere else.   beeFymojo:   yeh   beeFymojo:   i just hope the football column gets picked up quickly  Moogle: Uh huh   beeFymojo:   i\'d paste some quotes in here, but espn eradicated the whole archive   beeFymojo:   it is a good thing that RF has taken a year to get its first issue out  Moogle:  Yeah, otherwise you guys would totally have come across as the New Hitler Youth of the critical world, or something.   beeFymojo:   and our publisher would fire us all  Moogle:  Hey, speaking of insta-update and blog mentality, I\'ve been reading Transmetropolitan...   beeFymojo:   o?   beeFymojo:   how is that?  Moogle:  Oh, it\'s good.  Moogle:  Especially if you have delusions of being a writer or journalist.  Moogle:  But the \"Feedsite\" thing Ellis put in there is remarkably similar to blog-news  Moogle:  Basically, people sign up to be \"listeners,\" and run around equipped with backpacks that feed everything they see and hear into these massive website/tv station things.  Moogle:  These are then filtered and turned into television.   beeFymojo:   that is interesting  Moogle:  A bit more centralized than modern blogs, but I could see centralized \"Plastic\" style hubs of \"professional bloggers\" who actually get good at writing or investigative journalism, or editorials, for that matter.   beeFymojo:   that may be where things ultimately go   beeFymojo:   it is kind of moving that way  Moogle:  Especially as professional writers are starting to keep blogs of a sort...   beeFymojo:   Gregg Easterbrook, incidently has one   beeFymojo:   which brings us full circle to the problem of that kind of system   beeFymojo:   pretty soon, we will just have rediscovered the old media publication environment   beeFymojo:   i mean, if you think about it as a frontier  Moogle:  Yeah, but for a few moments, right now, it can be remarkably simple to thrust yourself into literati.   beeFymojo:   in the literal frontier you had people just going and living independently off the land...   beeFymojo:   and here we are   beeFymojo:   yeah.  it\'s nice to have a moment  Moogle:  Yup. Pay attention. This is history.  urlLink Full Post and Comments    '),(2206,2680773,'20,October,2003','2003-10-20 00:00:00','   Well, the Bears with Chris Chandler look a lot less volatile than with Kordell Stewart.  I\'ve been trying to figure this Kordell guy out and my basic assumption is that he\'s essentially a rookie who has been in the league a long time.  With some good and possibly difficult coaching, he could turn his career back around and at least be considered a decent quarterback.  He just makes too many rookie mistakes on a Bears team whose veterans have the injury bug.  Chandler has plenty of experience and can use his limited abilities more effectively than Stewart can use his versatility.  I don\'t like to see Chandler start, but I feel fairly calm when he does.  Maybe Kordell Stewart needs to stop trying to prove he is capable of getting out of his slump and just focus on the situation at hand.  He\'s given the opponent the ball at the end of each first half this year, all in situations where the game was close and we could have made it, at least, closer.  These aren\'t life-or-death moments, so don\'t treat them that way.  Honestly, I have no ideal.  But my initial statement is probably the most true thing...  The Bears with Kordell, for some reason -- whatever reason you decide, are an unstable concoction.  Super Bowl!  urlLink Full Post and Comments    '),(2207,2680773,'19,October,2003','2003-10-19 00:00:00','   Kill Bill was tonight\'s attraction...  Bought some comp tickets off a guy for $5 a pop for a sold out show.  The story is funny considering a dream I had about a week ago...  My old friend Chris and I decided to go to the big Chicago premiere of Kill Bill, only we didn\'t have tickets.  Somehow we ran into a guy from my high school who had stacks of comp tickets he was giving away.  Chris, being the astute kinda guy he is, figured out the tickets were counterfit though.  My response, \"If they find out, the find out... but it\'s worth taking our chances.\"  Soon after, I learned ticket holders would get to meet Quentin Tarantino before the film.  Now, I don\'t know any HUGE Tarantino fans, but my imagination created every type from handwritten t-shirt-clad teeny-boppers to film geeks with scraggly hair and tiny specs.  And not only was Quentin meeting his fans, but he was staging a seance\' gone awry for groups of a dozen or so.  I finally got into the room where the seances were being staged and while waiting for the scene to reset itself, Quentin looks at my jacket and says, \"That\'s a good coat.  It\'s got that classic, vintage look... That\'s a great coat.\"  I humbly say thank you and in the process notice he is wearing a pair of sneakers like the ones I own.  I point and say, \"Same shoes too.\"  He laughs and points out we\'re like evil twins and all of a sudden he\'s schocked.  \"Your shirt,\" he says and goes to his bag and pulls out a book and holds it up to my chest to confirm this strange occurence.  I am wearing a Forlorn Funnies t-shirt, he has an issue of Forlorn Funnies.  \"WHOA!\" is what we both say, and I am about to suggest he make a movie of Forlorn Funnies when I realize I can use such a point to get a creative \"in\" with Mr. Tarantino once we get a chance to talk more.  Well, in the back of the room is a fanboy type guy I went to college with, who is upset by the parrallels between me and the director.  He starts teasing me about how I\'m taking the attention away from everyone else for no reason; he steps back and begins falling.  Quentin Tarantino, invoking Gene Wilders when a child falls into his river of chocolate, says matter-of-factly, \"Watch out.  This may be a fake seance, but that\'s a real bottomless pit with acid at the bottom.\"  I wake up before flesh begins to be eaten away.  By the way, Kill Bill is loads of fun.  I\'ll keep anymore comments about it for the review and move on here in the blog.  urlLink Full Post and Comments    '),(2208,2680773,'17,October,2003','2003-10-17 00:00:00','   Went to go see  urlLink Neal Pollack  read at  urlLink Quimby\'s  last night.  We were a dumb-founded (maybe just dumb) crowd who couldn\'t think up any questions in the question and answer part.  We tried.  Oh yes, we tried.  But trying to think equals silence, and silence is not conductive in a question-answer session.    In a public venue, I find it difficult to ask questions of a underground/alternative artist.  For this reason:  No one knows much of the background details of the artist.  When you\'re in these book reading crowds, you have the tendency, though, to assume everyone else knows many more details than you do.  You don\'t want to ask a question that will get you old women staring at you over the tops of their glasses while shaking their heads at your baffling ignorance; this makes you feel like the illiterate scum of the known universe; this is bad.  At  urlLink Quimby\'s  last night, the only information the group universally knew was that Pollack is from Chicago, is now living somewhere else, and writes as/about hilariously ultra-important alter-ego characters.  We got through the first two facts with a certain amount of cautious prodding...  \"How is your return to Chicago?\"  \"So, are you living in Philly or Austin right now?\"  \"Do/did you like it?\"...   Neal spared us the grief of stumbling around the third and read an Anthology piece on Columbia.  If you haven\'t read/heard Neal Pollack yet, you should write it down in your planner for next Tuesday at 10am.  Stop into a Borders or your local underground zine shop, and find the  urlLink Neal Pollack Anthology of American Literature  and read the introduction.  That\'s all you\'ll need to understand how important a piece of work it is.  Look for an interview with Mr. Pollack in the second full RF issue.  urlLink Full Post and Comments    '),(2209,2680773,'14,October,2003','2003-10-14 00:00:00','   Football is a game of gods.  Baseball is a game of curses   For the 7 out of 8 American households not tuned into the Cubs ill-fortuned drops of games 5, 6 and 7 of the National League Championship Series...  Shame on you.  Most people don\'t believe in magic anymore because they don\'t bother looking.  A good portion of the world, though, got to witness some lingering force from a more mystical time incinerate the bolstering hope (and egos) of Wrigley parishoners.  Sure, you can look at each moment in the Cubs\' demise and explain it all logically, but when you take the tragedy in whole...  There is something far more powerful than you, I, Alex Gonzalez or Steve Bartman at work here.  And all of baseball are its puppets.  Red Sox lost as well, by the way.  It is very satisfying for all whose allegiances lie in the south side of Chicago, that Cubs fans were partly responsible for game 6.  They, moreso than the Cubs, are the bane of our baseball existence.  But, more on this in an upcoming RF essay.  Time to work.  urlLink Full Post and Comments    '),(2210,2680773,'14,October,2003','2003-10-14 00:00:00','   The  urlLink Kill Bill  soundtrack has a 10 minute Latin-ish cover of the Animals\' \"Don\'t Let Me Be Misunderstood\" by Santa Esmerelda.  It\'s my feeling that covers should never be recorded straight (concerts are a different beast altogether).  It\'s disrespectful and shameless to peddle another musician\'s wares as one\'s own - even if you claim it\'s in \"tribute\".  Can you imagine if Picasso had painted a copy of Van Gogh\'s Starry Night?  Well, the inbred patrons of art would probably have clapped very loudly and pulled some bars of gold out of their cavernous belly buttons to pay Picasso with.  But Picasso... Picasso would have laughed all day in  urlLink those shorts  he wore in some of the David Douglas Duncan photos.  On the other hand, if Picasso had fractured Starry Night, we\'d have an actually interesting piece of work.  It\'s how Dada and Pop Art fed their extended families.  So, my advice to all the ambitious bands out there that would like to do a cover or two...  Listen to the original and add something to it.  Change its focus, change its flavor, find something inside of the song that the current emphases hide.  People will love you more for it.  So, remember.  Faithful covers are for chumps and concerts.  Required listening: Cat Power  The Covers Record , Red House Painters \"Silly Love Songs\", Yo La Tengo \"Little Honda\", the \"Roxanne\" song in  Moulin Rouge , and the above-mentioned track from Kill Bill.  urlLink Full Post and Comments    '),(2211,2680773,'13,October,2003','2003-10-13 00:00:00','   As you might have already noticed, RF proper is functional.  The last touch I\'d really like to get working before we start pulling people in is to get The Coco Cabana Club\'s beautiful eulogy of Captain Ahab \"Moby Dick\".  The band is struggling to give me the track without giving me a new incarnation of the blaster worm, but hopefully we can get things figured out in a short amount of time.  Regardless, we\'ll probably start publicizing the magazine and hope word-of-mouth extends our coverage.  And I\'ll be writing about a bunch of different pieces here so keep coming back.  urlLink Full Post and Comments    '),(2212,2680773,'28,November,2003','2003-11-28 00:00:00','   Turkey is pretty good    Pretty  good!?  You fool!!  Spent yesterday enjoying chunks of fresh turkey and generally being refreshed by hangin with the fam\' at my parents\'.    I hope everyone else had similar experiences.  Here\'s to keeping it together in the next few weeks before the second part of the holiday season.  Prost!  urlLink Full Post and Comments    '),(2213,2680773,'26,November,2003','2003-11-26 00:00:00','   Missing Links   There are some updates to the \"Stuff to Read\" section of the blog.  Tuesday Morning Quarterback has found a new home at NFL.com (ye gods!), so the TMQ link now takes you to TMQ.  There is no archive yet, so I will be linking directly to the articles.  I\'ve also added  urlLink Spinsanity , which is a very nice site that cuts down rhetoric and tries to give its readers a clear idea of what\'s  actually  going on in America\'s political system.  They criticize the truth-twisting of Ann Coulter and Michael Moore with equal disdain...  Now, that\'s refreshing.  urlLink Full Post and Comments    '),(2214,2680773,'24,November,2003','2003-11-24 00:00:00','   Missing Legs   As a symbol of hard-lucked perserverance, the three-legged dog has always been a favorite thing of mine.  So, it is with great sadness and amusement that I give you this story:  urlLink Five-legged Dog has Two Legs Removed   See to believe, folks.  urlLink Full Post and Comments    '),(2215,2680773,'24,November,2003','2003-11-24 00:00:00','   Skinned   On Friday, I happened to flip-and-stick (a term I made up and hope to use on a regular basis) on HBO\'s  Naked in America  documentary on a global tour of the artist  urlLink Spencer Tunick , who has made a name for himself by orchestrating seas of naked individuals across natural landscapes.  When Kristi and I saw the original ad promoting the program, we both kind-of winced at Spencer\'s... exploitation in the name of art, i guess.  That is...  here was another artist peddling mediocre nude compositions for their cliched shock appeal and \'taking the art world by storm.\'  Bleh.  It\'s not that I won\'t eat it, it\'s that I really have no taste for it.  But what engrossed my attention were the motivations behind the exploitation; the documentary could assess these things where Tunick\'s photographs could not.     Before getting too far into this point, though, I\'d like to say that the director/editors did an excellent job of presenting the finished photographs of Spencer Tunick; detail crops were handled with a good understanding of composition and  the  composition itself, and I got an excellent idea of the presence of Tunick\'s work through it.  I\'ve seen this botched plenty of times, so it was nice to see it done correctly.    Anyway, back to more pressing matters...  In documenting Spencer\'s struggles and triumphs at convincing random individuals to pose for him and in interviewing those who did agree,  Naked in America  explores the national relation of global citizens to their bodies.  Probably the prime example is the director of the Russian Museum\'s argument with Spenser over the implied erotocism of his work.  Tunick\'s goal is to create very stoic nude photographs, and from an American perspective this is a success.  But to a Russian, coming from a recent cultural heritage of communism, the simple presence of a naked body has an erotic value to it because of how new such an individual act is.  As one of the models (the director was the centerpiece of a photo, incidentally) states in explaining his drive for posing, under the former cultural/political system one\'s body did belong to oneself but to the greater society.  It is these relations that the documentary explores so splendidly.  Japan\'s corporate mentality, France\'s openness to nude on a wall but prudent opposition to naked on the street. the world-travelling youth of Ireland stepping away from the reserved ideals of the church...  There is also some handling of individuals who are coping with the scarring of their physical bodies (through HIV or early-life wounds), but for whatever reason the director chose not to go into this type of relation too deeply.  The other thing we don\'t see is Spencer Tunick approaching locations where his art might actually be deemed dangerous by national institutions.  For the time being, if you want to get to why Tunick\'s work is exploitation you can simply address the absence of, for instance, the Mid-East.  Tunick arrives at scenes where the fundamentalist control structure has been laid to ruins, but never to where it still stands or is beginning to detioriate.  For what purpose does Spencer avoid locales where his art might have real impact?  There are plenty of answers, many justifiable, but the question is worth addressing for this artist who has turned the globe into his canvas and its people into his paint.  Overall though, an excellent tv special worth stopping for.  urlLink Full Post and Comments    '),(2216,2680773,'19,November,2003','2003-11-19 00:00:00','   Have Intercourse With Our Friends and Beat the Crap Out of Our Neighbors   At least, that\'s how this George W quote from a speech in Britain is bound to get picked up:  \"It really is about time we started to realize who are allies are, who our enemies are, stick with the one and fight the other.\"  It must have sounded good in his head at the time.    Aside from when he really gets his mind set on something, Bush tends to be much more a grayscale thinker than he expresses himself to be.  This has been part of his PR problem since... well, since forever.  I imagine him seeing this kind of statement as grand and heroic rhetoric, a head coach in the locker room at halftime about to stage the biggest come-from-behind victory the world has ever seen.  After the show, he gives Condoleeza Rice a high five and like college buddies they point out the kick-ass-ed-ness of the sound bite.    I sincerely doubt Bush literally meant we\'re going to send the armed forces of all allied nations to massacre all non-allied nations.  But you can\'t use global struggle as a metaphor when global struggle is what you\'re addressing.  On a related note, it\'s times like these that we should read British papers like  urlLink The Guardian .  It allows us to see how the people being addressed are viewing how they\'re being addressed.  Reading them to get the most honest outlook (which I know many who do) is just as immature as trusting mainstream US media as the end-all be-all.  More on that later (if I can get myself riled up about it).  urlLink Full Post and Comments    '),(2217,2680773,'16,November,2003','2003-11-16 00:00:00','   Eat, Drink   Last Saturday, Kristi, some friends and I went downtown to eat at Lalo\'s, one of the River North food machines that caters with huge atmosphere and overpriced dinner plates.  Reservations for 9pm unwittingly turned into 9:45 and then 10:30 and could have become later if it weren\'t for the diligent pestering of Kris and Ben.  In a restaurant whose overall floor area is half-devoted to the bar, it is no wonder that reservations get bumped to a later spot.  Drinks from bars are already grossly inflated but tourist-gobbling enterprises like Lalo\'s jack it up another notch; so why not pocket a few extra dollars for a nudge of the reservation time?  In their defense, resaurants also don\'t want tables left open because of reservations where the diners are late and justifiably have to cushion what would be a loss.  But praying to the gods of chaos eventually incurs their arrival.  Not that Lalo\'s (or any busy restaraunt) really loses any money when things go awry.  For our wait, we were negotiated free drinks and some free appetizers, which have a high menu value but extremely low real cost.  The only person affected is the tangentially related waiter/ress whose guests might not tip for the menu value of the free items (we did).  So, I guess what I\'m saying is to not specifically fear Lalo\'s but to hold a general wariness for any dinner factory and most restaurants.  The food, though, was just okay, so my feelings won\'t be hurt if you decide to never go there... except you\'ll be missing out on a fantastic paintings of boats sailing by a cross against a clouded map sky.  urlLink Full Post and Comments    '),(2218,2680773,'16,November,2003','2003-11-16 00:00:00','   Respect for Amnesty International   I would probably join the American Civil Liberties Union if it were not for a few pet issues which I feel have become politically motivated and are counter-productive to the key principles of the organization.  In conversation with people on the behaviors of the ACLU, I have often cited Amnesty International as having the kind of ideal apolitical atmosphere the ACLU  should  be striving for.  That said, I find  urlLink this Salon interview with Amnesty\'s US executive director William Schulz  to be very interesting.  Salon has chosen to sauce the tease (a term I made up and hope to use more frequently) to fire up its audience, but at its root the interview addresses the non-stance Amnesty International has taken on what the organization interprets to be political decisions.  Some interesting implication of the ethics involved in such decisions raise some debatable questions; for instance, should Amnesty International take a lesser-of-two-evils approach in certain cases? Of course, the edges are glittered up with sparkles to get the \"antiwar left\" dazzled in one way or another.  urlLink Full Post and Comments    '),(2219,2680773,'15,November,2003','2003-11-15 00:00:00','   Princes of the Universe   At  urlLink Penny Arcade , something has been written about the new Prince of Persia that makes me want to jump on my bike and ride down the frozen tundra of Clybourn and trade in all my old Game Cube games to purchase it.  O...  Since you are so curious, here is a little portion of what was written.  \"If your game doesn\'t look as good, play as good, or just feel as good against my bare skin maybe it should never see the retail shelf.\"  So, a reminder.  I don\'t necessarily hate this type of criticism (although I don\'t think Tycho necessarily sees himself as a critic persay); in fact, I enjoy it.  Good article in  urlLink The Reader  this week about  urlLink Bookslut  that ever so subtly touches on the debate over the purpose and approach of literary criticism (Why couldn\'t it just go in that direction a tad more!?).  And it reminds me why I think we\'re so cool for trying to do what we\'re trying to do at RF (and what other underground zines are also stabbing at)...  Most normal places... Whether the criticism is too harsh, too soft, too true; it\'s all geared at the sale.  I love the fact that such criticism exists (this might go against RF 1.1\'s theme) and that I can hear/see beautiful praise or scorn laid at the altar of the works of man.  But there are times when I KNOW what I\'ll think of something and what I really want to learn about is what makes a work work.  I am at a point in my life where having my furnace and water heater explode at the same time seems like less strain than reading plot summaries and back catalogs in reviews.  Fluff.  Let\'s trim it so we can see the glorious beast underneath.  But yes.  Accolades and attacks are great fun, I eat them like they are ice cream.  I\'ll even eat tiramisu ice cream because I love ice cream so much.  urlLink Full Post and Comments    '),(2220,2680773,'11,November,2003','2003-11-11 00:00:00','   Elephant and a Drawbridge   Alchemists pay attention: Viva La Bam is GOLD.  Take 2 parts Jackass/Camp Kill Yourself, mix in 1 part story, then remove consequences: the mystery has been solved.  But now the question of the hour, will the increase of gold in the marketplace radically devalue it?  We can only wait and see.  urlLink Full Post and Comments    '),(2221,2680773,'06,November,2003','2003-11-06 00:00:00','   But the Water Main\'s Busted   New American Ninja Underground strip is up.  \"Water Main\" is probably my favorite ANU strip so far and I think is the best portrayal thus far of what I\'m trying to do with the comic.  ANU is meant to be a humanized  urlLink Xiao Xiao .  Although the stick figure fights will be cool and stylish, my interest is more in telling a comic story about a bunch of ninjas trying to be ninjas in a world not made for ninjas.  I hate punchlines (gag frames and what-not) in comic strips.  Alright.  Let me correct that with a convoluted story.  In the first strip of my short-lived series,  In Pictorium , I wrote three frames of solid text about how the comic was never going to have a punchline; the fourth frame said, \"Oh.  How about a picture?\"  It was a stick figure sketch of a drunk man and a crow and a chicken on some grass.  This is to say:  I\'m not necessarily opposed to gag frames or punchlines, but I don\'t think a comic strip should read like a joke.  A regular-running comic strip depends on the notion of each strip being just a piece of a large body of work.  This is true even if there are not regular characters persay, because the narrator and the themes become the continuity.  With  In Pictorium , the worst strip I ever did was one where the intrusive self-analyzing writing of the narrator was replaced with thought bubbles of the comic\'s central figure.  Although the structure of the humor and the visual style were exactly the same as in previous outings, I had lost the strip by thinking each  In Pictorium  could depend on those two principles and stand individually.  I was wrong.  In a continuing comic, there is assumed story around a strip and to be most effective, the author needs to depend on this story in structuring these individual parts.   Calvin and Hobbes  is the example I keep thinking of...  Bill Watterson used the continuity of his strips to ground his humor. i.e.  He didn\'t have to force a joke on the surface; instead he depended on his characters\' natural responses to create the humor.  The way he structured strips could change dramatically day-to-day, but each strip acted as just a part of something bigger.  To put it most clearly, though...  A strip didn\'t make sense just at it\'s conclusion; it made sense in every frame.  And that\'s what I like about \"Water Main\".  The final frame is the climax of a joke and the strip is at some root level a build-up to that gag.  Even if you haven\'t read any other ANU strip though, you know there is more than \"Water Main\" going on.  The apex feels good, but all the other frames are doing more than leading-up to that point.  We are getting a better idea of who the characters are, as well...  Cyd\'s as-of-yet-unamed wife, continues to be confrontational towards the ANU\'s boys club behavior, Cyd himself continues to be a guy getting pushed around by these two powerful halves of his life, and we meet Peter Travers for the second time...  Cool and deadly serious about his craft.  It just has a great sense continuity while pushing the story critically forward.  What Gary Larson did in the  Far Side  is a different creature altogether though.  The major difference being that  The Far Side   was  gag strips.  Larson, however, managed to establish strip continuity in ways which I have not thought about hard enough to fully explain.  But, his strips were not simply jokes, and anyone who has read the strip can probably attest to each piece really feeling like a piece of the entire body of work.  If you have any thoughts on this, please post in the  urlLink forum  and I\'ll be sure to respond.   urlLink A Modest Destiny  is a neat strip that I think is structured very well; if you\'re ever looking for something fun to read on a boring day go back to day one and read the catalog thus far.  Anyone who would enjoy it has probably heard of it, but  urlLink Penny Arcade  is another online exclusive that generally works well at deemphasizing the final frame gag.  urlLink Full Post and Comments    '),(2222,2680773,'06,November,2003','2003-11-06 00:00:00','   Gatorade\'s Cool Blue (it\'s in a silver bottle, it is NOT the Frost blue but is colored the same) has an electric aftertaste.  I am not even joking.  urlLink Full Post and Comments    '),(2223,2680773,'05,November,2003','2003-11-05 00:00:00','   The Spirit of Jackass Alive and Kickin?   I preface by saying, I\'ve been away from MTV for a long, long time...  Although Johnny Knoxville has left the hijinx of Jackass, MTV is gratefully reluctant to let go of the weird assortment of talent they had surrounded the turn-of-the-century underground legend with.  Yes, Jackass spin-offs have hit the tube and it\'s a funny thing considering the original show was the exact opposite.  Ashes to ashes, dust to dust.  Breaking up the circus has allowed for some things that worked well on the original series to be fleshed out.  The adventures of Steve-O and Chris \"Party Boy\" Pontius with Manny, a kind of wildlife guide from the So Florida episodes, are turned into a half-assed, tongue-in-cheek nature show.  I\'m sure this will be said of it ten million times, but it\'s the Crocodile Hunter if the Aussie gator guru were an actual idiot.  There\'s a bad catch-22 working against the show though, the \"Wild Boyz\" can\'t hurt the animals but the animals are totally willing to hurt and/or kill them.  So, there\'s not the same kind of idiotic danger as in Jackass...  but, sometimes, it still manages to be funny.  When Steve-O had eggs from a dead salmon squirted in his mouth and then salmon semen squirted all over his face, I laughed.  Pretty loudly too.  The other show is Viva La Bam.  Which I have not witnessed yet, but which presumably follows the Margera family\'s psychological torture at the hands of Bam Margera.  Sundays, I believe, if anyone needs another show to clog up their Sunday evening viewing.  urlLink Full Post and Comments    '),(2224,2680773,'03,November,2003','2003-11-03 00:00:00','   All the Prophets in the House   Carnivale has opened a mighty big metaphysical can of worms on its viewers as the battle to inevitably be waged between good and evil begins to unveil itself.  This is a show with an almost painful patience at revealing its characters, as if it\'s saying \"You are about to witness one of the most collossal conflicts in the history of the universe.  Let\'s make sure you fully understand what brought every piece into place for it to happen.\"  The allure of, say, the Sopranos is in waiting to see what will become of Tony Soprano...  Sure, hints have been dropped that the show will end in Big T\'s demise, but we have no idea where that will come from.  Also.  Tony has been working towards remedying his feeling of the end being nigh both practically and psychologically.  So, there are no guarantees.  Instead we watch to see where the characters will bring themselves.  Carnivale is different in that we have the feeling that we are seeing a history of the story; we don\'t have a full character sketch for any of the major players at this point, but revelations to who they are gradually given.  Even within an episode we see this structure; we don\'t understand the relevance of things until the conclusion (even then, we oftentimes are only assured that items will have relevance to the grand conclusion).  Anyway.  We\'ve hit a point where we know enough that we can actually talk about what\'s going on in the show and not just stare at each blankly and say, \"What the hell just happened?\"  urlLink Full Post and Comments    '),(2225,2680773,'03,November,2003','2003-11-03 00:00:00','   Critical Mass   Friday night I participated in the October edition of  urlLink Critical Mass .  For those not in the know--basically everyone--Critical Mass is a bike trek through Chicago\'s busy streets during Friday rush hour with many hundreds of bicycles.  It is one of the strangest (or most frustrating) things you will ever see and honestly one of the most fun to participate in.  The line of bikes, once the mass gets moving, stretches at least five city blocks, and although the front line stops appropriately at traffic lights the rest will keep moving through red intersections in order to keep the line together (done partly for safety, partly for order, partly for the fun of it).  Arguably the best part of the ride is the reaction to it.  The majority of witnesses fall into the puzzled and/or awestruck category... Bar patrons plaster the windows or spill out into the sidewalk, pedestrians step to the curve to get a closer look or to question riders breaking to the side, car passengers think they\'ve lost their minds.  After all, the streets, which we associate with the dangerous, accelerating iron-grid of automobiles has been pacified by an endless stream of pedal-powered skeletons.  Part of the wonder of even participating in the event is this dramatic reversal of authority.  Even if the corking of traffic frustrates you to no end, there is little to say or do aside from just watching it all go by.  Howerver, some people are, I suppose, in denial.  They will honk (which is greeted by the \"woo\"-ing of passing cyclists), try to inch their way through the mass (which is greeted by one excited cyclists pulling to a stop directly in front of them, thus ceasing the car\'s forward movement), or in extreme cases, get out of their car and try to pick a fight with their halting rider (which is greeted by swarming bicycles demanding for the individual to get back into their car).  Normally, routes are planned ahead of time and pretty much stuck to, but a few \"tricks\" in this month\'s map ended up bringing the ride into a benevolent chaos as the ride was given its own life on a wandering exodus from Daley Plaza.  The group I was with for most of the ride ended up experiencing a radically different conclusion to the ride than I did after we were separated --DUE TO THE INTENSE SPEED AT WHICH I RIDE!!! -- but both concluded with a practical party social in the middle of triangle intersections.  They got bagels though.  (There were probably 3-4 different packs by the end of the ride.)  For more on the Mass, check out the website :  urlLink Chicago Critical Mass  or drop me an e-mail.  urlLink Full Post and Comments    '),(2226,2680773,'22,December,2003','2003-12-22 00:00:00','   Dorks in Sweaters   Headline has nothing to do with anything.  But.  I\'m going to be writing for the comics section of the pretty major web-zine  urlLink Pop Matters .  It\'s a very exciting prospect for me, and I\'m currently weighing some other neat options outside of Robots Fighting.  Again.  RF is going to become a more experimental publication and I think we\'ll all be better because of it.  urlLink Full Post and Comments    '),(2227,2680773,'20,December,2003','2003-12-20 00:00:00','   Sick   I\'m sick.  I\'m either not sick ever or just slightly sick all the time, but either way I don\'t get like this so often.  But when I do I can I feel my body fighting itself, which oftentimes gets personified in my dreams with whatever strategy game I\'m playing.  I remember having some horrible nights reliving Starcraft and Command and Conquer scenarios.  Most ended up with me dashing to the bathroom to puke.  Warcraft has been much nicer to me and has held the diseases in check better.  Am I just better at Warcraft and so can use it to command my white blood cells more effetively?  Not likely.  urlLink Full Post and Comments    '),(2228,2680773,'20,December,2003','2003-12-20 00:00:00','   Shake it like a Polaroid picture!!!   \"This is my happy song,\" my cousin Tim turned to me and said as I did about as much getting down as is possible for a bundled up white boy constrained to a seat in a freezing Soldier Field.  \"It\'s probably everyone\'s happy song right now,\" I responded like a jerk on the subject of Outkast\'s \"Hey Ya!\"  Even though the song has become love-it-or-tired-of-it party driving vernacular and Tim Martens\' official Happy Song, I\'ve been wondering about how really happy the song is.  Somewhere between the body moving hand claps and peppy guitar strumming and the get-unmoving-bodies-moving call and response, there\'s real despair in the song.  The lyrics, about the seeming impossibility of love being able to hold two people together for long, tends to show that.  Even if we aren\'t paying attention to (or making out) the lyrics the first hundred thousand times, the song still reaks of a melancholy darkness.  Just listen to the wilting delivery of the title phrase, or the eerie lonely echoes of the chorus chimes and Andre\'s calls.  But it\'s a song that overwhelming makes us feel happy.  Do we find joy in this celebration of our mortality?  Do we just ignore the song\'s mood of impending disaster so that we can clap along?  What does either say about our relationship with music as it relates to ourselves?  Is it even worth thinking about or should I shut up and dance?  Hey ya!  urlLink Full Post and Comments    '),(2229,2680773,'15,December,2003','2003-12-15 00:00:00','   The Times They are a Changing   A little forewarning.  I\'m making the move to fold the creative writing of Robots Fighting into Dmitry\'s xpehb.com .  This is the first part of a number of moves which will be enhancing the experimental aspects of RF, and will be a major move in establishing the magazine.  Regular features will be bolstered but organization will be less formal, making more room for informal experiments in presentation while keeping a focus on interesting style.  I want to create a sort of frontier of writing and visual art that can be explored and consistently be filled with surprises.  I don\'t want people to say, \"Who the hell is Tim Howard?\" and never read about his approach to songwriting.  I want readers to stumble upon things and make connections and most importantly feel like RF is a place, that it is as intended a sandbox in a warzone.  I also want more psychological freedom for myself.  A magazine is measured in issues and volumes and for a personal project like RF, I don\'t want to get caught up in an expectation of linear production.  As I\'ve come to realize, I dig myself into holes in order to dig myself out of holes.  If I can produce laterally...  spend 8 hours working on a cryptic flash animation or the script to a comic book and not feel that what i needed to do is finish that Shattered Glass review.  The way I work, I feel the refocused concept for the RF website will actually help me improve production.  I know that may sound odd.  But that\'s my thinking.  Anway, Dmitry is being given a great amount of access to my creative writing, so a lot of poems which were only ever publicly read or printed up for classes will get an expanded audience.  I am excited to be able to share a lot of these pieces to a bigger audience, and I continue to commend Dmitry for the exquisite site he produces.  urlLink Full Post and Comments    '),(2230,2680773,'11,December,2003','2003-12-11 00:00:00','   The Google War Begins   Do a Google search for \"Miserable Failure\" and look what pops up as the #1 response.  The battle for America\'s #1 Miserable Failure is beginning as folks like  urlLink Glenn Beck  are looking to get Michael Moore\'s name to be king.  The process of creating this is called Google Bombing and has been a realized Google weakness for about a year or two.  Now that it\'s hitting the mainstream and with political debate heating up with the election year, you may notice it more regularly.  For a full explanation of Google Bombing :  urlLink check here .  urlLink Full Post and Comments    '),(2231,2680773,'09,December,2003','2003-12-09 00:00:00','   You and Me Were Never Meant to be Part of the Future   The great grandmother of a gaming clanmate passed away several days ago.  Although he\'s not one that I have gotten to really know yet, I\'ve been trying to think of some encouraging words to pass on to him.  At funerals we often end up mourning the inability of a person to continue with us, instead of accepting the ceasing of that person\'s mortal walk and seeing their place here for what it was in the past.  Maybe I am fond of the latter idea because of the solace I find in the theology of and surrounding predestination, but I think there\'s a certain truth to it no matter what your doctrine or creed is.  That said.  I would like to have the Flaming Lips play at my funeral.  I am officially abandoning my prior request to be trebuchet\'ed into the sunset, but I do with great sincerity thank those who kept this knowledge with them.  You have done a great service to me and I am indebted to you.  I abandon that plan because looking back it seems less romantic and more a giant logistical hassle for everyone involved in planning.  I mean, if you can make it happen... great, I appreciate it.  But you can move it out of the special request category.  Anyway.  The Flaming Lips.  Really, if they could just play Do You Realize I think I\'d be happy enough (knowing this would happen), because that song captures in a moment the way I feel death should be universally dealt with better than many others.  \"Do you realize - that everyone you know someday will die - and instead of saying all your goodbyes - let them know that life goes fast it\'s hard to make the good times last.\"  That is, to me, you can\'t get caught up in the finality of mortal things, you make the best of the moments you share and you share the understanding that there\'s no reason to waste mortal time bidding farewell...  \"The sun doesn\'t go down, it\'s just an illusion caused by the world spinning \'round.\"  So that\'s that.  If they want to play a couple other songs, I\'m cool with that but see if they can somehow not make my traditional relatives walk away saying, \"That was weird.\"  Because it\'s for them as much as it is for anyone else.  As much as I might like them to appreciate that weirdness, it\'s not necessarily a forum they should be doused with it in.  A bit like how Christianity is handled in the same situation; it\'s laid out for anyone who wants to understand the calm it brings those who embrace it, but you don\'t have altar calls and trash like that.  I\'m still hanging onto having \"Be Thou My Vision\" played, but I wouldn\'t want the Lips to be forced into it unless they were really feeling it.  If they are, word.  If not, you can try to pick up Pedro the Lion or just whoever has the song in their blood.  But don\'t have Kirk Jackson or Michelle LaGroue sing \"Do You Realize?\"...  They have wonderful voices, but I really wouldn\'t feel right about it not being The Flaming Lips.  Of course, what do you care?  I\'m dead.  urlLink Full Post and Comments    '),(2232,2680773,'09,December,2003','2003-12-09 00:00:00','   Digital Millenium    urlLink Yet another story  of a guardian getting the lawsuit for a child\'s Kazaa downloading.  Intellectual property lawyers who have the upbringing and understanding of digital issues may still be too young to make a difference in the courtroom, so it is disheartening to see the media (generally considered younger and hipper) failing so miserably to alter perception of file-sharing.  Stories of people too stupid to make a plausible argument only make the broader battle more difficult.  The lady in the story above, for instance, has decided that her best vocal defense is to ignore the issue altogether...  As if pretending she doesn\'t have to settle or fight will make the lawsuit disappear.  How zen.  Grandparents, working class moms, 12 year old teeny-boppers... these faces are meant to elicit sympathy against the big bad wolf of the RIAA... but it\'s not what we need.  Watching straw and stick huts fall isn\'t going to spark people.  Make a story out of a house made of brick; somebody who\'s standing their ground and might actually have a shot at holding their own.  There\'s gotta be such a case somewhere...  The anti-share defense is so simple and so ignorant of legal sharing that I find it hard believe that someone in the position to do so is not closing in on not just a loophole but a solid legal stance for peer-to-peer.  urlLink Full Post and Comments    '),(2233,2680773,'03,December,2003','2003-12-03 00:00:00','   Killing in the Name of   A number of gaming sites have been reporting with a not-so-subtle flair of irony the release of  Medal of Honor: Rising Sun  in Japan.  The game, developed in America, places the player in the role of American soldiers during World War II\'s Pacific Campaign...  You know, the one where the US fought the Japanese.  Game Girl Advance\'s write-up and discussion on the release is  urlLink here .  The basic questions stem from Japanese players simulating the slaying of what could in real-life have been their fathers, uncles, grandfathers, etc.  There is generally this type of pondering when a game based on recent historical conflict is released, but this debate is possibly greater because it relates to a work in the birthing place of the art form.  Now, I don\'t want to say this is a non-issue, because there are without a doubt a number of very interesting issues involved with this release of  Rising Sun .  But.  When a person plays a video game, they are knowingly entering a fiction...  No matter how detailed and realistic it may be, it is a known fiction.  This is what allows a gamer to \"kill\" in the video game world and never give even a thought to killing someone in the physical world.  We play  Grand Theft Auto  or even  Midtown Madness  with an understanding that oncoming cars could--in the real world--potentially be us or our families...   There is a definitive line between fact and fiction; although we may try to blur it, although we may astonish audiences by nudging them closer to the line than they have previously witnessed anything go, it is there.  Our maturity is in some ways measured by our ability to discern each side of this line, and it is definitively immature to claim that a Japanese player is killing a relative while playing a somewhat accurate portrayal of World War II\'s Pacific Campaign.  Or... to claim that there is betrayal against the nation in fighting against Japan within the game.  The problem is in a critic\'s application of the player\'s reality onto the fiction of the game.  The avatar and the player are two separate entities and such projection ignores this.  While I eat a bean burrito from Taco Bell, my character could be starving in Siberia.  As much as the game or I may try to trick my mind in order to make the game\'s fiction more absorbing, any bond between character and player is an artificial one.  While my character dies from a gruesome head wound, I sit in an upholstered chair at a computer desk sipping a glass of milk.  While an enemy may be killed in the game world, its code still exists just as it did before and can be resurrected by simply reseting a level or the game entirely (and arguably, just by coming across an identical enemy).  If I kill in the real world, I wipe out the code for that being; there will be no revival or cloning.  It\'s gone.  By nature, the deletion of a crappy game from your harddrive is closer to real-world killing than in-game slayings.  Shouldn\'t it be considered more morally absurd to kill the avatar of a friend down the street in multiplay, than to halt an autonomous character?  Well, maybe not... because both events are equally part of an entirely fictitious universe.  Again, our eyes may be fooled by pretty particle effects, but the game is still fiction.  Only in the most tremendous of  urlLink Philip K. Dick  stories, does the line between the real and the artificial become blurred to an extent that there is a problem of morality.  The key to these stories, though, is the erasure of an individual\'s memory of entry into the fiction.  But the  Rising Sun  issue is one based not on morality but on business (specifically, marketing).  How do you get Japanese players interested in the  Rising Sun  story?  Once they decide to play the game, they are knowingly entering a fiction.  It\'s getting there that\'s the tough part.  If we try to turn fiction into reality and force ourselves to put real constraints upon the fiction, we destroy the purpose of it ever existing in the first place.  No longer could we explore the  un real.  And that\'s the real danger.  urlLink Full Post and Comments    '),(2234,2680773,'01,December,2003','2003-12-01 00:00:00','   Whoops   My apologies to ANU readers.  The book which contained the strips is lost.  If I can\'t find it by the end of the week, I will start anew.  Thanks for hanging with me until it gets back on track.  urlLink Full Post and Comments    '),(2235,2680773,'29,January,2004','2004-01-29 00:00:00','   Howard the Duck   And we\'re going to [pull all of our ads in] Missouri!  We\'re going to [pull all of our ads in] South Carolina!  And we\'re going to [pull all of our ads in] New Mexico!  And we\'re going to [pull all of our ads in] Arizona!  Yeeeeeeeeeeeaaaaah!!!!!   urlLink Dean Skips Airing Ads in 7 Primary States  urlLink Full Post and Comments    '),(2236,2680773,'27,January,2004','2004-01-27 00:00:00','   Just Be In Love When You Scream That Song, I    Some songs I\'ve been enjoying lately:   \"Staring at the Sun\" - TV on the Radio  : like The Polyphonic Spree if they were stuck in a factory machine   \"Now It\'s On\" - Grandaddy  : possibly  the  single of 2k3 (after Hey Ya), also possibly the only band able to make geek-tech themed music that isn\'t geeky   \"Be Thou My Vision\" - Pedro the Lion  : a little awkward at first, but a nice reworking of the hymm to emphasize the human frailty of the narrator   \"Last Night It Snowed\" - Ass Ponys  : mainly because I\'ve been able to mimic the titular line perfectly, but it\'s a good song anyway   \"Last Call\" - Outkast  : a tough guy hip hop track that\'s still suave as hell   \"Sinking Ship\" - Soltero  : at the end of a long night of painful drinking, lift your last pint of Guiness and sing unintelligbly along with this song.  skip the \"woos\" and stumble out of the bar.  this is how you make an exit.   \"Cinematic\" - Erin McKeown  : i\'m a sucker for ms mckeown\'s music, and that\'s the real issue here.  a fast, light, maybe even agile song.   \"Kingpin\" - Wilco  : i wanna be... your king pin...   livin in...  PEKIN.   \"Anything For Love\" - Robbie Fulks  :  how this song is even possible beats me.  clean-cut noise folk with soul.  at least, i think that\'s what i hear.  urlLink Full Post and Comments    '),(2237,2680773,'26,January,2004','2004-01-26 00:00:00','   A Man of My Word   So, I followed a link from the Windows Media Player media guide today.  I swear, it isn\'t something I usually do but how do you resist a link that reads \"Bustin\' a Move for the Pope?\"  You don\'t.  It is your duty as a good Catholic to find out who is bustin\' what moves and how the Pope reacted.  (where and when are also important but to a lesser extent)  The thirty second clip features some Polish guy breakin\' on the palace floor in front of the Pope throne.  He wasn\'t bad but I\'m not about to adapt it into a full length screen play.    The real reason to watch the movie is to see the aging John Paul himself.  There he is with his Pope cap on his Pope throne watching some jackass spin on his head when he reaches up his hand and does the Pope wave.  Done from a balcony or Pope-mobile, this gesture is a pious expression of welcome.  Done towards someone in the middle of popping and locking, it looks like an especially God-Fearing \"I feel you dawg.\"    The Windows Media Player wouldn\'t let me snag a direct link to the content, but searching for the terms \"pope break dance\" on the media guide will take you to the good stuff.  urlLink Full Post and Comments    '),(2238,2680773,'26,January,2004','2004-01-26 00:00:00','   A more complete piece on the Daily Show town hall meeting in New Hampshire is  urlLink here .  urlLink Full Post and Comments    '),(2239,2680773,'25,January,2004','2004-01-25 00:00:00','   Google Searches That Will Bring You Here   Apologies to anyone looking for something in specific who are totally disappointed to find no archive system as of yet.  I\'m trying to come up with a way to organize some of our more thought-provoking posts on the RF mag site, but that is on the back burner until the rest of RF gets back up.  But.  People are finding their searches end up here.  So, it being the end of the month, I\'ve decided to present the more quixotic search strings to find their way here.  (There aren\'t many this month, but maybe in time there will be more.)  \"Encouraging Words for the Sick\" \"Flaming Lips Songs that Make People Puke\"  Tune in next month for another exciting edition of Google Searches That Will Bring You Here.  urlLink Full Post and Comments    '),(2240,2680773,'25,January,2004','2004-01-25 00:00:00','   All Your Base...   Usually, I\'m not into disseminating links to inane internet quirks, but every once in a while I find something great.  So, in the tradition of Super Mario Bros. 3 in 11 minutes and Terry Tate: Office Linebacker, here is  urlLink The End of the World  urlLink Full Post and Comments    '),(2241,2680773,'25,January,2004','2004-01-25 00:00:00','   The Dailies   Just discovered, thanks to my RealVid install (wait, am I thanking Real?) that  urlLink Comedycentral.com  has been doing a very nice job of archiving Daily Show headline segments.  I always end up missing the actual show (Comedy Central is in what we call the \"fuzzy high channels\" on my TV), so it\'s a nice resource.  Also, I\'d like more details on this, but apparently The Daily Show pulled together a panel of media higher-ups for  urlLink spirited discussion .  The CNN story I\'ve linked is only good for a couple chuckles, no actual information on the event.  urlLink Full Post and Comments    '),(2242,2680773,'23,January,2004','2004-01-23 00:00:00','   Yes, politics is an inbred pile of skin tissue, focussed more on \"our party, our country\" (a term I just made up and hope to use more frequently in the future) rhetoric than on actual issues (they are getting worked out somewhere inside the epidermal lump, but it\'s pretty well hidden from us).  But sometimes, rhetorical analysis handled properly can be insightful.   urlLink James Fallows\' notes on the State of the Union   By the way, pick up the current \"State of the Union\" issue of The Atlantic while it is still on newsstands; by far, the best magazine being produced today.  urlLink Full Post and Comments    '),(2243,2680773,'23,January,2004','2004-01-23 00:00:00','   The RealMark of the Beast   Well, it\'s happened.  I\'ve installed Real Player again after avoiding it like the armageddon for the past year and a half.  I can already hear my computer crying about having its will broken by the software program...  Real Player does what it wants no matter what you tell it.  Want WinAmp to keep control of mp3\'s?  TOO BAD, REAL IS RUNNING THE SHOW NOW!  To cap it off, the introductory audio clip tells you, \"YOU\'RE IN CONTROL of the most powerful media player...\"  And you have to ask yourself...  If it\'s so powerful, why does it let a fragile user like me control it?  The answer:  IT LIES TO YOU.  This is the true sign of power.  urlLink Full Post and Comments    '),(2244,2680773,'23,January,2004','2004-01-23 00:00:00','   Water, Water Everywhere   Just when you thought the Mars craze had passed, you get contact reestablished with Spirit (see Andy\'s post below) and the Europeans have apparently  urlLink confirmed water on Mars  (the last time you heard about it, it was just a suspiscion presumably).  My hope is that we can blame water on Mars on an impact-created exchange of molecules with Earth.  Let me tell you why.  As someone who writes and enjoys science fiction in a variety of forms, my fantasies are reliant on the idea that we are, as far as we know, alone.  For one thing, this allows a region of mystery which one can fill with fiction.  If we meet a bunch of lizard-chimps on Mars who tell us \"You know what, we\'ve traveled f-in everywhere and there are no f-in wookies,\" well, our ability to dream that there are wookies is hampered.  For another, our entire relationship with the universe radically changes.  Even if all we find are microbes or maybe tuna fish, they become an integral part of the story of us and the universe.  It\'s like if you were to write stories about life in your apartment and suddenly your dying cousin Larry moves into your bedroom...  Larry now figures into your stories; even if you never have meaningful conversations with Larry and generally despise him, you\'re stuck with him being part of your apartment life.  You can\'t just write him out of the story, because, well, he\'s dying and that\'s a rude thing to do to someone terminally ill.  This is how it will be with the microbes.  You\'ll look up at the stars at night and think about the millions and millions of miles of nothing but burning gas and floating rocks and, oh yeah, a bunch of parasites next door.  And when we finally discover some hot, super intelligent aliens half way across the universe, we\'re going to have to bring them home to our solar system and introduce them to the microbes, who will do nothing but swim in circles and drool.  What are we going to do?  Ignore them?  They\'re life!  You can ignore Jupiter, you can ignore Pluto... but you can\'t ignore life.  It just doesn\'t happen too often.  And it happened twice in the same solar system and you want to forget about it?    No.  Impossible.  Life on Mars is going to be a part of our day-to-day interstellar realities whether we ask for it or not.  We could hope that it won\'t be found, even though we would really like the initial thrill of discovery...  Which is why I can only root for life on Mars to be some kind of backwash of the cosmos.  And now you know.  urlLink Full Post and Comments    '),(2245,2680773,'21,January,2004','2004-01-21 00:00:00','   The General vs. the President   I watched the entire Democratic New Hampshire debate last night because I\'m a nerd and was a poltical science major.  Anyway, all throughout the debate I kept thinking, \"Is it possible for anyone to do  worse  in a debate than Wesley Clark is doing tonight?\"  It was as if he didn\'t expect to actually be asked questions.  He had no answer for numerous inqueries, embarassed himself in trying to explain away Michael Moore\'s comments on Bush, and relied on stupid cliches all night long.  (My favorite?  \"This election is about the future, and that\'s why we\'re here tonight.\"  Thanks, General.)  I was shocked at his answer to Peter Jennings question about Michael Moore calling George Bush a deserter while onstage with Clark.  The General basically said he didn\'t think it was relevant, didn\'t bother to check the facts, and thought Michael Moore had a right to say whatever he wants (Don\'t infringe on our free speech!  Even if it is factually wrong!!)  Clark is a General in the United States Army and someone is calling his Commander In Chief a deserter right in front of him while campaining for his presidency.   Clarks says he doesn\'t care and it\'s not even relevant to him.  Probably because Moore is such a great \"leader.\"  His word, not mine.  Anyway, the  National Review\'s  Byron York has a great column on this same subject.   urlLink Byron York on Wesley Clark in the debate  urlLink Full Post and Comments    '),(2246,2680773,'21,January,2004','2004-01-21 00:00:00','   The Red Menace   Good news urlLink   from the Gusev Crater: after a scary day and a half, Spirit, NASA\'s robotic Mars exploration rover, has reestablished contact with an eager flight team here on Earth.  The contact came early this morning in the form of a 20 minute data burst broadcast from the red planet.  For the scores of scientists involved in the project, this means the hundreds of millions of dollars and countless man hours put into landing the rover on Mars won\'t be wasted after only a few weeks of exploration.  More importantly, for me, this means more mind blowing, high quality images of Mars.  Since Spirit touched down on the third, I have spent about half an hour of every morning studying the latest pictures and reading press releases about the minutiae of the Martian landscape.  I have never spent this much time on a .gov website before.  In fact, before the Mars project, I can\'t say I\'ve ever spent any serious amount of time on a US government endorsed website.    This has major implications for our president.  His administration is in charge of one of the most popular websites on the Internet (as of today it is ranked 302 on Alexa) and there isn\'t anything on the site that connects him to it.  Think about it, every day thousands of nerds visit this site and try to found out as much about Mars as they can.  What if they also found out that George W. Bush supports the search for signs of extra-terrestrial life and is dedicated to furthering our understanding of our universe through blah, blah, blah, vote for Bush.  It doesn\'t matter that he\'s taking credit for something he doesn\'t have a whole lot to do with.  Politicians do it all the time.  It worked for Gore, didn\'t it?  I can\'t say that I voted for W. his first time around or that I plan to vote for him in November, but if he were to change the focus of his campaign from the War on Terror to the War on Space, I might at least watch one of his debates.  urlLink Full Post and Comments    '),(2247,2680773,'21,January,2004','2004-01-21 00:00:00','   The Ladder Device   Possibly the easiest way for Palestinians to convince Israel to stop with its damned wall is to threaten an arms race of the wall vs. Palestinian ladders.  As the new security fence, priced at $1.9 billion dollars has been  urlLink breached .  urlLink Full Post and Comments    '),(2248,2680773,'20,January,2004','2004-01-20 00:00:00','   Some Random Fun   Got this link from NOL Mauler and thought I would share.   urlLink uber guitar playing  urlLink Full Post and Comments    '),(2249,2680773,'19,January,2004','2004-01-19 00:00:00','   OFFICIAL: Family Guy About to Go Back into Production   Family Guy is about to go back into production any day now despite there being no official word from Fox, according to an article in the New York Post last week. In the article Seth MacFarlane says: \"We\'re sort of waiting on an official word, but to the best of my knowledge, the plan is to start producing a batch of new episodes without necessarily knowing where they\'re going to wind up.\" Several options are open for the eventual home of the new episodes including Fox, or the Cartoon Network, but eventually they will end up on DVD.  Click for  urlLink the New York Post article .  urlLink Full Post and Comments    '),(2250,2680773,'19,January,2004','2004-01-19 00:00:00','   Let\'s Get Ready to Rummmmmble!!!   I don\'t know why this makes me laugh so much...  But it does.   urlLink Howard Dean after Iowa   I think it\'s the pro-wrestler mode (if Jesse Ventura\'s election means anything, it might just be that Howard Dean should do this more often) spiced with a maddened \"YeeeeaAAAH\" at the end.  urlLink Full Post and Comments    '),(2251,2680773,'19,January,2004','2004-01-19 00:00:00','   Remember the Vernacular   Or sometimes the best laid plans ignore the actual geography.  I at once cringed and laughed a few years ago when Malcom X College here in Chicago had the slogan \"Capping Off the Millenium\" on banners around their urban campus, somehow missing the double entendre in the street definition of  urlLink cap .  On Saturday, it happened again.  Marshall Field\'s ran an ad in honor of Martin Luther King Jr, which was filled with excruciating iconography. The worst of which read in big wood-cut Soviet-style lettering:  Celebrate  urlLink THE MAN  (and then smaller text which cleared up that the man was Martin Luther King Jr)  I know the phrase is somewhat common, but the ad\'s large and equal emphasis on THE MAN seems to allude more to  urlLink Urban Dictionary\'s definition  than to the traditional honorary phrase, especially when coupled with a typeface reminiscent of a cruel and faceless regime.  Perhaps it\'s unfair of me, but in both these cases I\'m puzzled as to how such copy/design escapes the various levels of the editorial process...  Someone somewhere along the line must recognize that the wording or design emphasis could be changed to avoid embarassing puns in the vernacular of the target audience.  Do they knowingly reject such interpretation, or is it just a mistake of circumstance?  I have a hard time fully believing either.  urlLink Full Post and Comments    '),(2252,2680773,'15,January,2004','2004-01-15 00:00:00','   Trebuchet Snobbery and the Return of the King   A common misuse of the  urlLink trebuchet  in cinema is the siege weapon\'s combative use against ground troops.  As any trebuchet hobbyist will tell you, the easiest way to explain a treb to someone is to refer to it as a catapult and yet the two have very little in common aside from their basic hurling purpose.  A catapult (onager, mangonel, what have you), generally perceived properly, is a tension mounted device used to hurl heavy objects for moderate distances... It\'s target is pretty much up to chance, so a catapult is most effective against large groupings of enemy forces and large or light structures.  A trebuchet, however, is a long-range sling-based weapon known for its precision; it can not fire as quickly (changing trajectory and distance are also a more involved process) as a catapult nor can it carry multiple ammunitions in one firing.  The trebuchet was/is used against heavy, preferably stationary fortifications; it can hit these with more force and accuracy than a catapult would ever be capable of.  While a trebuchet is certainly capable of firing on ground forces, the excess power and precision is superfluous (a giant boulder does not have to be moving too fast to flatten a human), the sacrifice of speed and area-of-effect of shots would be too great to be effective.  That said, one has to wonder what the defenders of Middle Earth\'s Minas Tirith (P. Jackson version) were thinking when installing trebuchets along its walls for city defense.  Ye gods!  The only way such a thing would be worth while is if Sauron\'s tower sprouted hen\'s legs and began marching on the city itself.  In  Return of the King , we watch as a few boulders crush a few targets but against half the army of Mordor such scattered shots are inconsequential.  If only onagers had been installed along the front and secondary walls!  Or ballistae along the fortress\' gates!  So many lives would have been saved.  Case in point, it took legions of the dead devouring all in their path to tilt the battle towards the human side in the movie version of  Return of the King , while the books\' army of man (presumably, armed with proper knowledge of siege warfare) needed the dead only to tangentially capture Mordorian ships in order to succeed.  What a difference a siege makes.  urlLink Full Post and Comments    '),(2253,2680773,'14,January,2004','2004-01-14 00:00:00','   Memo on More (Moore?) Blogging   Scot never got the memo about RF being apolitical, but that\'s okay.  Readers will need to get used to his incendiary rhetoric at some point or another, and it\'s best to start with something simple like Michael Moore.  Scot\'s a Michael Moore defector; he used to swear by  Roger and Me  and  Awful Truth .  But then, even long before  Bowling for Columbine , Scot saw Moore start to go a little loopy (maybe he\'ll tell us the story).  And here we are.  For more on Moore (pun of the day?), check out  urlLink Spinsanity ; there you will find some of the most independent-minded indictments of Moore playing fast and loose with the truth (a phrase which is now forbidden for the rest of 2004).  To be a dork for a second, I would like to explain each of the three current blog team member\'s in Dungeons and Dragons alignment terms (the idea was given to me by my manager at work, who is a serious but respectable D&D player).  I am replacing Good and Evil with right and left or left and right (it is inconsequential really), but Chaotic, Lawful and Neutral will remain the same.  Scot is Lawful Right, Andy is probably somewhere around Lawful Left and True Neutral and I like to think I am somewhere between True Neutral and Neutral Right.  I think there\'s a lot we can learn about politics through Dungeons and Dragons, but I\'m not really qualified to make that post.  urlLink Full Post and Comments    '),(2254,2680773,'14,January,2004','2004-01-14 00:00:00','   Scot on Moore on Clark   Apparently, I will be posting here now too, introduction by Mike or not.    This is a test post.  This is only a test.  But to make it worthwhile, a few notes about crazy Michael Moore\'s endorsement of Wesley Clark.  First of all, if I were Clark, I\'d decline the endorsement, but that\'s not going to happen.  Moore sent out an e-mail explaining his reasons for support.  A few of the highlights:   1. Clark has committed to ensuring that every family of four who makes under $50,000 a year pays NO federal income tax. None. Zip. This is the most incredible helping hand offered by a major party presidential candidate to the working class and the working poor in my lifetime. He will make up the difference by socking it to the rich with a 5% tax increase on anything they make over a million bucks.    This is Moore\'s main thrust in his current incarnation (besides not responding when terrorists attack us): class warfare.  The left has no new ideas, no passion, no drive, no real support in working class America for their radical anti-war stance.  So, it comes down to taxing the rich.  \"Socking it to them,\" as Mike says.  What have the rich done to deserve this?  Well, they\'ve been successful in life.  They\'ve worked hard to accumulate wealth.  They are the success stories in this country.  And the reward is paying 5% extra in income taxes.  What about the family of three making under $50,000?  Are they less deserving of a tax break than a family of four?      2. He is 100% opposed to the draft. If you are 18-25 years old and reading this right now, I have news for you -- if Bush wins, he\'s going to bring back the draft. He will be forced to. Because, thanks to his crazy war, recruitment is going to be at an all-time low. And many of the troops stuck over there are NOT going to re-enlist. The only way Bush is going to be able to staff the military is to draft you and your friends. Parents, make no mistake about it -- Bush\'s second term will see your sons taken from you and sent to fight wars for the oily rich. Only an ex-general who knows first-hand that a draft is a sure-fire way to wreck an army will be able to avert the inevitable.   Well, there\'s a lot of lies in that paragraph.  Bush will not bring back the draft.  Recruitment is up right now.  And in fact, re-enlistment is supassing the Army\'s goals.  That\'s right.  More troops in Iraq are re-enlisting than even the Army expected.  DOn\'t believe me?  Unlike what Moore would do, here\'s an actual link to facts:   urlLink http://www.estripes.com/article.asp?section=104&article=17852&archive=true   Another of Mr. Moore\'s reasons:    3. He is anti-war. Have you heard his latest attacks on Bush over the Iraq War? They are stunning and brilliant. I want to see him on that stage in a debate with Bush -- the General vs. the Deserter!    Excuse me, (cough, cough) Mike. You might want to read this:   http://www.drudgereport.com/mattwc.htm   Those attacks for the war are also stunning and brilliant!   Mike\'s never been afraid to play fast and loose with the truth though.  That feels about right for a first post.  Let\'s see if this works.  urlLink Full Post and Comments    '),(2255,2680773,'14,January,2004','2004-01-14 00:00:00','   More Bloggers to Blog Things   Andy Buchenot is going to be joining the The Machine\'s Still On crew as RF moves its critical writings into this forum.  Another voice and one as interesting as Andy\'s brings variety and reduces my feeling of preaching (even if I\'m right about everything).  Posting will probably be picking up as a result of this and things will be more regular as we continue the major overhaul of Robots Fighting magazine.  If you\'re getting anxious to see those changes check out  urlLink the testing page .  As soon as I finish the illustrations and basic structure, this page will be taking over the old main page.  Anyway, Andy can feel free to write here now, and you can feel free to read.  urlLink Full Post and Comments    '),(2256,2680773,'13,January,2004','2004-01-13 00:00:00','   The Way Things Ought (Not?) To Be   Today my grandmoms and regularmom came out for lunch and visit.  At lunch, our crossing conversation had me and my mom discussing a new addition to her sewing machine (it is, to my credit, a very high tech sewing machine) while my grandmas, one in a pink sweatshirt the other wearing a soft white sweater, discussed the weekend\'s playoff football action.   Palindrome Speaking   Two guys sitting next to us were talking about home furnishing and one said to the other, \"They are the nicest wood floors I have ever seen.  And I said that to him, I said, \'These are the nicest wood floors I have ever seen.\'\"  I do this occasionally but it made me smile to know others fall into the trap as well.  Palindrome speaking (a term which I just invented and hope to use regularly) is an awkward, dull and clumsy part of rhetoric, but there is something...   neat  about it.  urlLink Full Post and Comments    '),(2257,2680773,'13,January,2004','2004-01-13 00:00:00','   For Those Who Have Rocked   Thought I\'d pass on an excerpt from this week\'s Tuesday Morning Quarterback (see link to the left, or simply look on nfl.com) which shows the kind of thing commenators  should  be pointing out to you.  It\'s these guys\' JOB to watch football and tell you what\'s happening.  DEMAND MORE!   The Panthers called a three-man rush. Marshall Faulk cut-blocked Carolina DE Mike Rucker, diving at him low; Rucker responded by jumping on top of Faulk. Brilliant play! Why? Watch Rams film and you will note Faulk often cut-blocks a DE to get underneath him while selling pass-blocking, then jumps up to catch a screen pass. Rucker knew this, and simply sat on Faulk so he couldn\'t get up. Marc Bulger just stood there holding the ball, wondering where Faulk has disappeared to. Carolina ended up with a sack that effectively ended the first half -- although with Rucker seated atop Faulk, at that point the Panthers had only two gentlemen rushing against five offensive linemen.  urlLink Full Post and Comments    '),(2258,2680773,'04,January,2004','2004-01-04 00:00:00','   Hold your breath and prepare to be shocked...   urlLink It\'s Mars!  urlLink Full Post and Comments    '),(2259,2680773,'02,January,2004','2004-01-02 00:00:00','   For Those About to Rock   Maybe it\'s mostly chance, but there\'s got to be some kind of skilled craftwork in FOX\'s coverage of the Green Bay vs. Seattle game.  One astounding replay showed a closeup of Brett Favre in the center of the screen with D.D. Lewis\' back taking up the bottom right corner...  You see Lewis jump a little early and Brett Favre look directly at him and realize the threat.  Favre stops counting and makes a change up.  Now, the chances of a FOX camera man catching this moment perfectly composed like it was are slim.  The chances of a producer catching the moment in enough time to kick it up to the announcers to specifically comment on are even slimmer.  But there it was on my TV screen.  Yes, this is the number one coverage team and this kind of thing should be expected, but they still deserve a hats off for this shot.  I\'ve been wishing that football commentary for the playoffs (and in general) was a little bit more like what FOX did with the baseball playoffs (and world series), where the network brought in a pitcher for the Mets (I can\'t remember his name; apologies to him and his family).  Said pitcher was detailed in explaining the strategy and method of pitching throughout the game, leaving anyone listening to FOX\'s coverage a little more learned.  Many (including me) are under the impression that football is all emotion and drive, because that\'s all announcers can talk about.  Their ability to explain what quarterbacks are seeing, how blocks are working or why defensive schemes are working is virtually non-existent.  So few actually understand the complexities of the strategies of the game and believe solely in hype.  Chris Collinsworth.... Don\'t tell us that \"angels are still floating above this Green Bay Packers team\".  I have seen you on Inside the NFL and I know you know more about what\'s happened than that.   Clarify-cation   I guess I should put the real emphasis in my Soap entry on drug stores.  I generally find soap in grocery stores (but I do think it takes longer than it should).  urlLink Full Post and Comments    '),(2260,2680773,'02,January,2004','2004-01-02 00:00:00','   Soap   Does anyone else find it incredibly difficult to find a bar of soap in a drug or grocery store?  If not, I have to doubt that you\'ve ever tried.  I spent a half-hour at Osco today looking for some soap.  What I ended up with is a Neutrogena POWER SCRUB or some sort of thing.  It\'s basically tiny soap in a dish sponge.  But it\'s the closest not-hotel-size thing I could find.  How can there be ten thousand types of shampoo for every \"type\" of hair and there be no basic rudimentary soap?  I don\'t understand.  There\'s not even that much body wash (which is a joke... more perfumes and less actual soap to actually get you clean... this is WHY you need to have 13 stages of cleansing substances to open your pores....  because you\'re coating yourself in flowery feces).  ANGER!  urlLink Full Post and Comments    '),(2261,2680773,'29,February,2004','2004-02-29 00:00:00','   The Funny Thing About Putin   The funny thing about Putin is the way he says, \"I had nothing to do with this,\" with a clear and present wink.  The ex-KGB colonel just  urlLink pounds the opposition with every tactic he\'s got , but expects (and gets) people to just nod along.  In the last election, he answered the question \"What\'s your policy?\" with \"I won\'t tell.\"  Does the fact that he wins with only murmurs from the opposition mean that he\'s a genius or a bully or maybe just an ex-KGB colonel in politics?  urlLink Full Post and Comments    '),(2262,2680773,'27,February,2004','2004-02-27 00:00:00','   A: You must have to be Catholic.   Q: What was Mom Martens\' response to Mel Gibson\'s  urlLink The Passion ?  urlLink Full Post and Comments    '),(2263,2680773,'25,February,2004','2004-02-25 00:00:00','   Please don\'t sing us a melody   A fairly big pet peeve of mine is when when people sing along to pop songs (whether they be rock, hip-hop, country... whatever) in their \"pretty voices.\"  Yesterday, one of my coworkers was singing very melodically along with the Beatles \"I\'m Looking Through You\" in this soft, passionate way like it was \"Yesterday\" or \"Amazing Grace\".  The problem is that the glory of the imperfections is lost; \"You\'re not the same,\" instead of being this coarse, assertive holler that -ahem- fits the music turns into this loungey whine.  Blegh.  My favorite part of \"Build Me Up Buttercup\" is the \"WhhhhYYYYY do you...\"  Specifically, the \"WhhhhYYYY\" because it\'s a dry. strained exhale that just builds all the tension for the chorus.  This part gets mauled by sing-alongs all the time.  It\'s like the entire song is getting sold out in one second.  And these are the most fun parts to sing.  So, I just cannot stand it when someone sings over them with their crappy, trying not to sound crappy crap crap voices.  I mean, what fun is singing along with Bob Dylan if you don\'t contort your voice like a Cirque De Soleil clown...  What\'s the use of a rendition of \"Beast of Burden\" without an accentuated imitation of Mick Jagger singing \"Pre Ty Pre Ty Pre Ty Pre Ty.\"  There are people out there wasting the best moments of pop music with their classically-trained for karaoke voices;  IT NEEDS TO STOP!  urlLink Full Post and Comments    '),(2264,2680773,'25,February,2004','2004-02-25 00:00:00','   Things that make you go hmmmmm   While watching Dennis Miller, I saw a new commercial for Viagra featuring \"We Are the Champions\" by Queen.  I\'m not sure if the good people at Pfizer knew  urlLink the song is a major gay anthem  or if they\'re just trying to appeal to a new set of consumers.  After all, homosexuals have erectile dysfunction too, right?  urlLink Full Post and Comments    '),(2265,2680773,'24,February,2004','2004-02-24 00:00:00','   Fool me once, shame on you.  Fool me twice...   William Ligue, Jr., the man who attacked Kansas City Royals coach Tom Gamboa two years ago,  urlLink violated his probabtion over the weekend.   What a surprise.  Ligue allegedly tested positive for marijuana three times during the month of December, and will now, god willing, head to prison.  Prosecutors urged the judge in the case to give Ligue jail time to begin with, but the judge gave him probabation.  Since then, the state of Illinois has passed a law making it a felony to illegally enter restricted areas, such as a playing field, basketball court, locker room or stage. A violation carries up to three years in prison and a fine of at least $1,000.  Ligue used the death of his infant daughter to explain away his actions from that night, saying he turned to drugs after her death.  (No word if Ligue had another infant daughter die in November of this year)  He was under the influence of five drugs that night, including cocaine, marijuana and Valium.  While chaperoning his nephew\'s 13th birthday party.    \"That was not me that night,\" Ligue told the Daily Southtown. later. \"I was going through so much stress. ... I see on the videotape (that) I was out of my mind. I had to let anger out and it came out that way.\"  Sure.  The real him is just a loser druggie, not a violent lunatic.  I never quite understood why White Sox fans got such a bad rap over this case.  The criminal element in our society will almost always find a way to commit a crime if he really wants to.  This was not a reflection on Sox fans or basball fans.  This was a criminal committing an act of violence.  Where the system failed was in the sentencing.  As I said, criminals will find a way to commit their activities -- it\'s up to the legal system to punish the criminals for their actions and deter others from doing the same thing.  After Ligue another fan did the same thing to an umpire at Comiskey (fine, U.S. Cellular Field) last year.  Determent failed.  Ligue now has found himself on the wrong side of the law yet again.  Punishment failed.  There is good news -- the man who attacked the umpire was sentenced to jail time.  I believe he got three months.  We\'ll see if he decides to violate his probation.  urlLink Full Post and Comments    '),(2266,2680773,'23,February,2004','2004-02-23 00:00:00','   Grey Tuesday   So, if you haven\'t heard...  DJ Danger Mouse (a guy who reportedly has done gigs wearing a large mouse costume) has mixed a cut-up of Jay-Z\'s  urlLink Black Album  and the Beatles\'  urlLink White Album  to create...   urlLink The Grey Album .  On a purely critical level, the Grey Album is an interesting reenvisioning of Jay-Z\'s album but shouldn\'t be considered a tribute to the Beatles.  While the mixes are smart, they are frequently sloppy with the White Album samples.  As a an album, it pales to a great number of official releases and really doesn\'t deserve some of the frothing accolades it is receiving.  But.  It\'s free.  And.  It\'s become the centerpiece of a copyright litigation protest called  urlLink Grey Tuesday , which is occuring right now as I type this.  EMI has threatened lawsuits on sites distributing the free musical experiment, and so these sites and many others (maybe on principle, maybe caught up in the excitement) are distributing the album anyway on this day.  However, as neat as it is, I don\'t think the Grey Album is strong enough a piece of art to become a rallying banner for actual change.  And, as I\'ve said before on other issues, the problem is that the protest groups are stirring the muck, but they\'re not guiding the solution.  Grass-roots groups like  urlLink Downhill Battle  do an excellent job of piquing public interest, but they end up losing the long term debate by not engaging law-makers and officials with sensible solutions or interpretations.  But.  Hats off to Downhill Battle (and all other Grey Tuesday participants)for standing up to the internet-infamous Cease and Desist letters.  Even if there could be better long-term strategies, their short-term efforts are a noble attempt to seize back the power on the internet from corporate powers that are claiming it as their own.  urlLink Full Post and Comments    '),(2267,2680773,'23,February,2004','2004-02-23 00:00:00','   Save the Pinata   Turnpike Films has a series of very funny mock-up ads for various brands.  Check out urlLink  this one  and then click on next or previous to work your way through the cycle.  urlLink Full Post and Comments    '),(2268,2680773,'23,February,2004','2004-02-23 00:00:00','   Technologically Proficient   Got to spend a good chunk of today working on the site as you  urlLink may  or  urlLink may not  notice.  As those links show, the graphic menu system on the main page is starting to take form and promises endless fun for everyone involved.  Also, there is now a working archive section to the blog.  So, if you missed my post in December about Outkast\'s \"Hey Ya\" and are now wondering what I would have to say about it... well, you can try and find it.  Good luck, chump.  I\'ve updated the link bar once again taking the out-of-season Carnivale link off and switching Tuesday Morning Quarterback to Gregg Easterbrook\'s non-football column/blog on the New Republic.  Good stuff, highly recommended for frequent reading.  I\'m learning more and more about Blogger script and honestly, enjoy the power and customizability of it.  I\'ve been thinking about switching to a MySQL based server blog script, but I think this gets us a little more exposure and certainly does the job in a simple fashion.  Finally, comments are now available and I encourage you to use them.  Follow up on a post with your own witty remarks, ask the author important questions, just waste everyone\'s time with random statements... the power is yours!  urlLink Full Post and Comments    '),(2269,2680773,'22,February,2004','2004-02-22 00:00:00','   New Merchandise from Oak Street Park   Over the summer I decided to try and design some Abercrombie-esque t-shirts publicizing fake locations or events.  The t-shirts would be ridiculous but (like their A&F brethren) would have just enough realism that people might have to give them a second glance.  The only idea that really came to fruition was the Oak Street Park Hippie Burn...  An annual event where, presumably, hippies are burned to the great entertainment of spectators.  I think this event may occur with the hippies being willing participants as it cleans them of pachouli stink and pachouli dirt like a prairie fire rejuvinated the plains of the New World.  Of course, no hippies are forcibly ignited. They do, however, run like antelopes out of control once ablaze.  Oak Street Park Hippie Burn T-Shirts are now avaliable for a limited time at the  urlLink RF Store .  The Hippie Burn logo is vibrantly colored and captures the true essence of this fictional event.  Future incarnations of the shirt will be one-color printings, so this is a good chance for hippies  and  not-hippies to get their hands on this celebratory memorabilia.  urlLink Full Post and Comments    '),(2270,2680773,'20,February,2004','2004-02-20 00:00:00','   Fall of the House of Mouse   Saw Michael Eisner on Larry King Live the other day.  Ouch.  As well as he did trying to hide his worry about the upcoming board meeting (which might result in his ousting) with descriptions of attractions as \"awesome\", it was painfully obvious the kind of stress he is under internally.  The guy obviously loves what he does, but his tragic flaw may be that he believes too strongly in the power of the brand.  With attacks from all sides, it is becoming increasingly apparent that Disney isn\'t a media superpower; it can\'t expect creative talent (like the folks at Pixar) to come groveling to its feet anymore.  Disney needs to work to regain its mythic status and its going to require an about-face turn around.  Because of his love for his place and the company itself, Eisner, if he continues to be obsessed with brand power, will ride the company into the ground.  urlLink Full Post and Comments    '),(2271,2680773,'20,February,2004','2004-02-20 00:00:00','   Victims of Freshness   The words \"more fresher\" were uttered many times to me today.  Some answers regarding the current Chipotle crisis based on my conversation with the manager of my workplace Chipotle.  The recipes haven\'t changed, I was told; apparently, though, the stores will be making the salsas from fresh ingredients on site.  This means your salsa will be fresher.  It\'s funny then, that your salsa tastes more like  not  fresh salsa.  Still, it may be necessary to plead the Cheese and Macaroni defense.  It\'s true that fresh ingredients often do equal better food, but occasionally they don\'t.  Macaroni and Cheese is horrid trash if made with real cheese, but dehydrated cheese is enough to make kids worship the stuff.  FREE GREEN CHILI!  urlLink Full Post and Comments    '),(2272,2680773,'19,February,2004','2004-02-19 00:00:00','   If you see Beatles albums go on sale for incredible prices, DON\'T buy them!!! It\'s all a ruse.   Picked up the reissue of Bob Dylan\'s  Blonde on Blonde  and continue to be floored by SACD hybrid reissues (hybrid discs can be played on either SACD or CD players).  The SACD format allows audio to be stored in distinct channels corresponding to a 5.1 speaker set up, so sound engineers are able to emulate the actual space which musicians recorded in.  From the  urlLink Crutchfield Advisor\'s  interview with Dylan reissues producer  urlLink Steve Borkowitz :   The panning presents Bruce Langhorne (guitar) on the left and the bass on the right, with Bob in the middle â€” exactly like the stereo. But now, you have the actually palpable sound of the musicians playing in a room â€” in a real space. Itâ€™s three-dimensional, just like it was when he recorded it. The only thing you hear at the back is the stuff that you would naturally hear behind you in the room.    For those without a 5.1 SACD system (I\'m still not quite to that point yet either), it\'s important to note that the work put into the SACD tracks is so intensive that it comes across remarkably even on a simpler system.  Extensive work is done working with original masters in an effort to pull out all the detail and sound they contain.  You can hear instruments as individual units of a song, giving the music a much more real feel.  The Rolling Stones were the first to receive an extensive hybrid reissue discography and now we have Dylan.  I\'ve stopped buying Beatles albums  hoping  that they\'ll get similar treatment but there doesn\'t seem to be anything definite on the horizon.  urlLink Full Post and Comments    '),(2273,2680773,'19,February,2004','2004-02-19 00:00:00','   If I were to make you a tape right now...   ... it would probably have these songs.   Modest Mouse   urlLink Float On       I picked up this song after opening a lucky bottle of Pepsi yesterday and I have listened to it about 15 times since then.  The Mousers continue to emulate the sound of the Talking Heads and now are picking up a little of their sunny disposition as well.  Also, new drummer Benjamin Weikel looks funny when he plays.     David Byrne   You and Eye      Speaking of the Talking Heads, I revisited this used bin pick up earlier this week and found out it wasn\'t as bad as I remembered.  Sure, it can\'t touch 90% of Talking Heads songs, Byrne\'s warm burbling guitars and conspicuous rhythm section work together to make a pretty charming song.   Funky Four plus 1  That\'s the Joint      Best known for being sampled by the Beastie Boys, this nearly eight minute jam shows just how amazing hip hop was before Yo! MTV Raps made it commercially viable.  Featuring four funky DJs plus female DJ, this track has everything you could possibly want.   OutKast  Happy Valentines Day      This song came into the playlist last Saturday and refused to leave.  Hey, when arrows do not penetrate, Cupid reaches gets his pistol.  urlLink Full Post and Comments    '),(2274,2680773,'19,February,2004','2004-02-19 00:00:00','   Money For Nothing   If I ever get rich and famous (of course I mean even richer and even more famous than I am now) and I need to negociate a contract, I\'m hiring one man and one man only: Scott Boras.  At least three times in the past three years, Boras has gotten his clients more money than they should have been able to get.  Boras, somehow, convinces teams to bid against themselves in the persuit of his free agents.  Yesterday,  urlLink Boras\' client Greg Maddux signed a 3-year, $24 million contract with the Chicago Cubs.     This despite the fact no other team offered Maddux a contract.  The Cubs originally offered 2 years and $10 million.  Boras flirted with the San Francisco Giants.  The Cubs upped the offer to 2 years and $14 million.  Boras was in New York to announce the Alex Rodrigues trade (more on \"A-Rod\" in a bit).  Rumors flew that Maddux would sign with the Yankees.  And guess what?  The Cubs increased their offer to the final numbers.  So, without anyone else bidding for Greg Maddux\'s services, Scott Boras got $14 million more out of the Cubs.  Brilliant.  As I said, this isn\'t the first time.  When Alex Rodriguez signed with the Rangers, no other team in baseball bid more than about $16 million per season.  Boras got the Rangers to give A-Rod almost $23 million per season.  This offseason, Ivan Rodriguez signed with the Tigers after once again Boras got the team to raise their offer even though nobody else offered him a contract for more than two seasons.  The result?  Four years, $40 million.  Ka-ching.  Years from now when the big radio stations come after me, begging me to work for them I\'ll direct them to my agent -- the frozen head of Scott Boras.  urlLink Full Post and Comments    '),(2275,2680773,'18,February,2004','2004-02-18 00:00:00','   Which is the Better Pun: Dog Days or Dog Daze?    urlLink The Guardian\'s  L.A. dispatch reporter, Dan Glaister has brought up the claim that America is an  urlLink awards-obsessed culture, especially when it comes to canines .  I\'d expect this from Canadian parliament, but our friends in England\'s press?  Tsk tsk.  I think the big problem with the column itself is that Glaister says that the big story in the US was not a.) primaries b.) janet jackson\'s nipple c.) bush\'s tv interview, but was a d.) dog show.  He then mentions that he found out about Westminster in a news service e-mail where it was the  third  story listed.  Out of the four options, I\'d hope that the dog show would be number 3.  But Glaister also points out that nearly 4.6 million people watched the show last year... Astonishing for a dog show or for the UK, possibly, but for prime-time US television...  it\'s a wet newspaper.  Also amazing to Glaister is that there are 58 million dog owners in the US.  But the key is that there are 293 million people in the United States.  There are 60 million people in the UK.  If you multiply all of Glaister\'s US figures by five, you can see why he has deemed dogs an abnormal craze...  But as it is, it\'s nothing special.  What\'s obvious, though, is that there may be a small cultural rift between the US and the UK when it comes to awards and canines and awards for canines.  We\'re not batty because of it, as Glaister might suggest; we\'re just different (I feel like I\'m in a third world country defending myself against  US  reporters and tourists).  It\'s an issue of size and population.  We have space to keep large pets, so a sixth of the population does.  As far as awards go... we have a large population of creative work and a somewhat similar coast-to-coast culture, so awards help our culture to pinpoint and document what the greater works are over time.  And we enjoy giving awards for much the same reason; it helps to distinguish our favorite pieces.  Personal rankings work similarly (Nick Hornby\'s cast of characters in  High Fidelity  were certainly willing participants).  The system is, of course, filled with flaws, but anyone will tell you that; there aren\'t many highly regarded awards whose awarding is highly regarded.  I know a few people who are obsessed with awards, and the nominees can often get very worked up about them, but most people are only casually interested in them if at all.  But look at me... preaching to the choir.  Just remember folks, Canadia and Dan Glaister are in cahoots (sp?).  urlLink Full Post and Comments    '),(2276,2680773,'18,February,2004','2004-02-18 00:00:00','   Two Things Everyone Should Know   Give me a couple days to confirm the universality of this, but it appears  urlLink Chipotle  has changed its salsa recipes.  I grew suspicious on Monday when my meal was provided \"on the house\" after i asked for chips and salsa (maybe i should explain that I have a handful of \"usual\" items which half the employees have memorized... ).  Why am I being offered free food from people who know I will be back every other day until I change jobs or they change their salsa recipes?  A-HA!  Sure enough...  The sacred has been defiled.  Green Chili (aka crack-cocaine aka medium-hot) has been turned into a horrid parody of itself, a tomato-y edge gives it the flavor of a watered-down generic jar salsa.  Chili-Corn (aka medium aka THE CORN) supposedly has a stronger onion flavor.  Mild (aka tomatoes aka mild) is indistinguishable to me so far but may have a little less cilantro.  Red Chili  (aka chiles de arbol aka ---> HOT >  Give out a holler if you\'re hearing me on this, but there was the Word of God and then there was Green Chili salsa from Chipotle.  The divine spoke through us through these things and now we\'re stuck with the NIV of both.  Yeah, it\'s all there...  but where\'s the poetry?  Do you feel it?  Crisp, clean salsa taste found nowhere else...  Gone.  urlLink Full Post and Comments    '),(2277,2680773,'17,February,2004','2004-02-17 00:00:00','   Je t\'aime, Triomphe le chien comique d\'insulte!   Unlike other commentators on this board, I think of Canada as a sweet, soft spoken older brother looking down lovingly from the North.  Or at least I did until they started taking  urlLink petty shots at one of Americas Favorite Sons.    Fortunately, Conan was able to  urlLink rise above  but I\'m not sure I\'ll be able to drink a Labatts without thinking twice for a long time.  urlLink Full Post and Comments    '),(2278,2680773,'17,February,2004','2004-02-17 00:00:00','   Reviewing Makes the World Go \'Round   Article over at the New York Times looking at  urlLink anonymous reviews on Amazon.com  after the anonymity was negated on the Canadian version of the site.  (OH CANADIA, WOT TUMULT THOU HAST INVOGUED!!!! (a phrase which I have just made up and hope to use more frequently when citing the wretches of Canada)).  Regular readers of this blog might find it a nice tie of New Media political anarchy and user review databases (from a few days ago).  urlLink Full Post and Comments    '),(2279,2680773,'16,February,2004','2004-02-16 00:00:00','   \"Nothing Else Behaves Like Me\"   From Thomas Oliphant of the Boston Globe (which has been used as a primary source in recent media quagmire):   \"However, the impression fostered more in the press than in politics that the inevitable existence of questions and gaps permits  urlLink inferences about the answers  and about what might be in the gaps is... pernicious.\"   I can\'t promise that I am not going to lament New/Old Media relations over the next year, but I\'ll at least promise to possibly keep it limited.  urlLink Full Post and Comments    '),(2280,2680773,'16,February,2004','2004-02-16 00:00:00','   Only In America...   ...and perhaps Canada and England and maybe Japan.   urlLink A single line from an overrated song requires a company to clarify the correct way to use their product.   I just yearn for the days of Kodachrome.  Which, of course, was about 12 years before I was born.  Hey ya, indeed.  urlLink Full Post and Comments    '),(2281,2680773,'16,February,2004','2004-02-16 00:00:00','   i realize that i would be a more entertaining person if i were six inches shorter  urlLink Full Post and Comments    '),(2282,2680773,'13,February,2004','2004-02-13 00:00:00','   It\'s not so much an issue of who started what, but that  urlLink it\'s already started .  I\'ve been flipping through alternative news sources over the past couple days and have been appalled at the coverage.  Watch them bitch about the coverage, instead of actually providing ideal coverage. This is what you have to look forward to.    Get ready for eight months of democracy eating itself.  urlLink Full Post and Comments    '),(2283,2680773,'13,February,2004','2004-02-13 00:00:00','   Good News for People Who Love Carson   After rocking out with  urlLink Modest Mouse  on Wednesday,  urlLink Carson Daily  continues his string of surprisingly good musical guests with  urlLink Death Cab For Cutie  tonight.  Past musical guests have also included  urlLink Sonic Youth  and  urlLink Candy Butchers  and many others whose homepage addresses I am too tired to type in right now.  The point is that the one time terrible Last Call has nearly become almost watchable by catering to the snotty rock hipster set - something MTV, MTV2, and just about every other music venue on TV has given up on.  Not to make a tired argument, but putting interesting music (even if its not popular) on TV can actually draw a sizeable audience.   And they\'ll even put up with Carson Daily.  urlLink Full Post and Comments    '),(2284,2680773,'10,February,2004','2004-02-10 00:00:00','   Let\'s Get Political   If just for a second.  Techno popster Moby (who created the very stellar album  Play  and many par-for-the-course electronic records)  urlLink suggests lying on political-oriented message boards  to incite dissent in  Bush\'s  presidential campaign.  The can says \"Disgusting Worms\" but Moby is too busy looking for the perfect sample that he never notices what he\'s opening.  I can only hoping he was joking....  but if this were to catch on, you could basically write-off the relevancy of democracy.  I am not shitting you.  I\'ve seen Moby get tied up in a few political strangles recently, and what he absolutely does not understand is that your own weapons can be use against you.  Basically, this kind of thing would start a war of disinformation in the New Media, and the Old Media would be too busy salivating over the maggots to actually  urlLink fact-check .  Is it true that the ends can justify the means?  Yes.  Absolutely.  But we cannot actively endorse unsavory means to get our dessert.  Just can\'t.  urlLink Full Post and Comments    '),(2285,2680773,'10,February,2004','2004-02-10 00:00:00','   Record Round-Up   After spending one of my morning walks to work planning an elaborate independent shop review / database site, I discovered that one has  urlLink already been made .  At least for record stores.  It\'s a neat resource and I\'d like to see it expanded.  The problem is that by focussing on user reviews, the whole thing could turn south in a second.  We\'ve seen it happen on practically every other major database; individual reviews would be less than meaningless if it weren\'t already a suffix.  So, the major sites generally either favor the Top Reviewer awards to those who consistently write well (or, well, write consistently) or just create a giant database giving average ratings and demographic stats.  Smaller sites don\'t have the traffic to do the latter effectively, so they must rely on the former but don\'t have the traffic to effectively run \"review the reviewer\" scripts.  Even on larger sites, the most interesting subjects and reviews get ignored or thrown into a giant   Maybe the large question is how do you effectively filter democracy?  I haven\'t yet seen a system work extremely well at sorting out cheese from the deli meat (our own government included).  This, incidentally, has been the subjct of talks between politicians and video game designers in the past week, as the two discovered their problems and solutions are almost identical.  (I will link as soon as I can find it, but the story after being everywhere yesterday is buried today).   A site I thought would filter well was  urlLink Plastic .  Plastic automatically assigns moderator status to frequent, registered readers which allows them to endorse the quality of individual posts.  Two major problems have befallen Plastic.   A.  The left-leaning readership majority mods up liberal bitch-fest sessions, while conservative participants struggle to get \"karma\" points for even the most well thought out posts... (A well-recognized issue which generally frustrates the community\'s most active members.)  B.  As posts build up in popular conversations, readers use karma points to decide what to read, so fresh posts--no matter how astute--are ignored.  But.  Plastic is a pretty fair system and is one of the best, if not the best, mass public message boards on the internet.    I\'m realizing I\'m getting way too long with this post, so I\'m going to stop for now and continue on specifically how an early era Plastic model could help sites like RSR from drowning in their own democracy.  urlLink Full Post and Comments    '),(2286,2680773,'10,February,2004','2004-02-10 00:00:00','   Cheese Blog   I think of a lot of good ideas for blog entries on my walks to and from work, but with all the crazy stuff that happens on those walks.... I generally forget them.  I did remember this item though.  Someday if I ever get steenking rich and I feel like being mean, I\'m going to go into specialty stores and ask for a clerk\'s advice on purchasing...  say, a new ottoman or digi-cam.  In order for this to work, I will also need them to suggest some really expensive luxury item.  Then.  I will mock it.  But the trick is, I will demand to buy it  because  it is so hideously awful.  \"I never want to forget how horrible a thing this is,\" I might say as I chortle up the twenty dollar bills that will line my stomach.  Pause.  Unpause.  This would completely crush the self esteem of the sales clerk because of their role in the exchange.  If I call something \"ugly\" and leave, I\'ve created a barrier of conflict.  The clerk will think I\'m a jerk and be on with his/her life.  But if I patronize the clerk while insulting him, I force him to in some way agree with my statements in order to make the sale.  In sales, we often become convinced by our own sales; at my store, we generally distrust Canon cameras because Canon pays us the least amount of money for selling their goods.  Clerks often buy the things they sell because they\'ve convinced themselves that it\'s something they need.  It\'s something that can be turned back on them/us.  And with it you can demolish their/our aesthetic taste.  urlLink Full Post and Comments    '),(2287,2680773,'09,February,2004','2004-02-09 00:00:00','   Segway into Disney   I\'m categorically against the Segway scooter, that crazy gyroscopic upright personal transport, as a means of transportation, especially to be used around pedestrians.  In the words of Gregg Easterbrook, getting hit by one of those things would be like getting tackled by an NFL linebacker at full speed.  I don\'t trust people\'s focus on navigating it enough to feel at all comfortable with them zipping around me.  Maybe I should...  But I know the way people are in crowded and/or interesting places, and motor vehicles aren\'t good to add into the mix.  Regardless.  Disney World employees use them; I saw them a lot around Epcot when I went last spring, and I\'m pretty sure I saw one or two at the other parks.  Disney, I also seem to recall, was one of the major buyers of the Segways when they were first introduced.  I am not certain if  urlLink Disney\'s ban on patron usage of the scooters  instead of wheelchairs is straight out hypocrisy.  But I do know that this really doesn\'t say much about Disney\'s confidence in the product.  Another sign of the tremendous flop the Segway was/is.  I remember that in the expo portion of Epcot, a staff member was showing it off... and you could tell everyone gathered around was waiting for a chance to try it.  No such luck.  Why?  Because.  They\'re friggin\' dangerous around groups of pedestrians.  So, in Disney\'s defense, if someone came up to the front gate in a golf cart (which staff do use inside the park) I probably wouldn\'t let them in either.  Those complaining about their rights to the Segways use could just as well be saying, \"Not letting me drive around Disney World in my pick-up truck makes me feel more crippled\".  urlLink Full Post and Comments    '),(2288,2680773,'08,February,2004','2004-02-08 00:00:00','   Sometimes, news from Mars is  urlLink exciting and stupid  at the same time.  urlLink Full Post and Comments    '),(2289,2680773,'05,February,2004','2004-02-05 00:00:00','   Announcements   Work is continuing on RF Proper; the Dogmi gallery is now working and you will see that American Ninja Underground is also on the front page of the site.  (I will always post here when there is new content.)  There is a small problem with the Dogmi gallery system that I am looking into fixing.  If you are experiencing any other difficulties with the site -- things not laid out properly, etc -- contact me via e-mail ( mike at robotsfighting.com ) with what browser you are using and exactly what the problem is (a screenshot is helpful if you have the time / capability).  In other news, former Kindler Ben Johnson is at EA doing some work on the Ultima X morpg project Odyssey.  Also, if you buy the new Dead or Alive game in Japan, you will get  urlLink this .  O Dead or Alive... solid games turned legendary by a little jiggling and a lot of cheesecake.  urlLink Full Post and Comments    '),(2290,2680773,'03,February,2004','2004-02-03 00:00:00','   Doing My Part   The following is a message to anyone who can\'t resist opening up an e-mail attachment because it might be something important...  It\'s not.  From CNN :  MyDoom.A, also known as Novarg or Shimgapi, emerged nearly a week ago in the form of a spam e-mail message that contained a well-disguised virus attachment.   NOOOOOO!  It\'s not well-disguised!  It\'s a freaking e-mail attachment!  You can\'t disguise an e-mail attachment!!!  That\'s like saying that the bomb outside your front door was cleverly disguised as a bomb.  Well, maybe, even though it looks like a bomb it\'s actually a package from a long lost lover.  Let\'s open it up and screw the entire world up the polar ice cap.  Also from CNN:  Yes, so basically it disguises itself so there\'s no clear thing to look for.    Wait, wait...  Even MyDoom\'s arrival as a no-body-text e-mail attachment (basically, saying to us plainly \"I am a f**king virus\"), is not a clear enough sign to tell there\'s a virus?  TRULY... TRULY THERE IS NOTHING WE CAN DO!  QUICKLY!  TO THE PANIC ROOM!  (Why do these explanations get buttered up in the news media?)  Honestly, I can be a little forgiving to people who are just coming to the internet and don\'t understand that it\'s not Disneyland...  they don\'t understand that they cohabitate this space with other people who are more savvy and more exploitave than them.  They need to learn this quickly, though.  Perhaps a tour of the world, where they can see metropolises like CNN or AOL.com but also be driven by the vast frontiers of porn and piracy.  Whatever the case, they\'re naive and the community needs to do its best to make them less so.  My problem is people who consistently get burned by viruses and scams.  It\'s been explained to them, they\'ve seen the results... but still... they just can\'t not click the freakin\' e-mail attachment.  They look at it and maybe they think for a second, \"Hey...  This is a little suspicious, maybe I should just delete it.\"  But fantasies of Disney erotica and de-light-ful forwards override reason, and suddenly their computer is effectively disabled and Microsoft has lost $250 million.  I hate these people, and I wish we had a way to make them go watch TV and leave the internet alone.  Seriously, don\'t be that guy.  And if you know that guy, BASH HIS BRAINS IN WITH THE CLOSEST SOLID OBJECT!  DO IT NOW!  urlLink Full Post and Comments    '),(2291,2680773,'02,February,2004','2004-02-02 00:00:00','   Beyonce\'s a Survivor   The girl\'s a class act, but a guy who isn\'t is Richard Hatch.  Honestly, I know why they picked who they did from the Borneo (I) cast, but the Sue and Jenna thing was over by the time the first season reunion special occurred.  There\'s a good possibility that the oppurtunity for tension may never even arise (and there\'s no guarantee it will even then).  I guess they selected the four fiercest players from season 1, but I would have rather seen Gervase or Colleen come back for a showing.  You don\'t have to be in constant competition to win Survivor (Tina from Australia is good enough evidence of this), so why not bring back a few more interesting/fun characters?  Anyway, the downfall of many would-be contenders is that they get to playing \"the game\" too early.  In its early stages, Survivor  is  about surviving; your group needs to stay as united as is possible in order to lay the groundwork for a strong post-merger showing.  If you make too many enemies too early, if you let your tribe get picked off in immunity challenges, you\'re not going to last.  You have no immunity protection from your own tribe nor can you make actual strategic or trust alliances based on immunity.  In these first stages it is about tribal unity, not individual strength.  Richard, having been gone for half a dozen seasons, is way too anxious to play.  He is playing as if he were only 5 people away from getting another million dollars already, pretending to hold all the secrets and promising to muckrake.  He\'ll have to be on his toes if his actions cause his group to shatter.  Even if he\'s around just to have fun screwing people over, he\'s not going to be doing it for long.  Pride is his hubris, and he appears to be blinded by it right now.  I\'m rooting for any member of the Africa cast, although I wouldn\'t mind if Colby got his dues for losing so graciously to Tina in the Outback.  urlLink Full Post and Comments    '),(2292,2680773,'01,February,2004','2004-02-01 00:00:00','   You Can\'t Coach That   So why watch a Superbowl between two semi-mediocre teams that you don\'t really care about?  Okay, you might catch one or two   urlLink good commercials  but mostly you\'re going to see another championship game that doesn\'t feature  your  team.  And maybe, just maybe, you\'ll see  urlLink a celebrity boob.   Viva la Superbowl!  urlLink Full Post and Comments    '),(2293,2680773,'31,March,2004','2004-03-31 00:00:00','   On File-Sharing and Music   The RIAA may be nearing the point where they will have to pick  urlLink another scapegoat  for their failing business model.  urlLink Full Post and Comments    '),(2294,2680773,'30,March,2004','2004-03-30 00:00:00','   Apologies   I deeply apologize to everyone for the near incoherency of the previous post.  I was writing it on the brink of sleep, much like I used to write all my columns in college...  which reminds me that is time to apologize for the incoherency of those pieces.  Sometimes, just because words can be fit together, does not mean they should be.  I also apologize to everyone who I have to send e-mails to... This time my excuse is still lame, but it is better than most times as my laptop has inexplicably lost function of its G, H, Backspace, and \' keys.  I generally prefer writing on my laptop, and it is psychologically difficult for me to use my tower for that sort of thing these days.  We (and by we, I mean I, but maybe you will empathize) have been spoiled by an abundance of technology.  I will get my act together very soon, but I have two other things distracting me right now.  1.  I received a pass for the Chicago preview screening of the  urlLink movie version  of  urlLink Hellboy the comic .  I like Hellboy a lot.  To me, it\'s a readable superhero comic, one that works well in a short story format along with the drawn-out serial plotting.  The characters are clear-cut and consistent.  The art is superb; bold, dark and brooding with elaborate simplicity.  And it\'s got everything... Demons, Nazis, mythological monstrosities, fishmen, frogmen, wolfmen, Lobster Johnson...   I\'m excited to see how well it will transfer onto celluloid.  I\'ll throw my thoughts up here.  2.  Also for free, I signed on for the  urlLink Lineage 2  open beta.  It\'s my first experience with a Massive Multiplayer Online Role Playing Game since  urlLink Gemstone 3  and Federation (which I cannot find even a tangential link to) were centerpieces of AOL\'s online gaming aspect.  Both were text-based if it helps you put things in perspective.  Unfortunately, Lineage 2 has been  packed  and the servers seem unpredictable as to their ability to keep people on, so I haven\'t been able to get a decent session going.  I have, however, killed my fair share of wolves on the open plains of the game\'s starting towns.  It\'s a funny sight to walk over a hill and see hordes of individual players chasing wolves around and slaughtering them.  I don\'t understand why MMORPGs insist on having players mass execute in order to get to a point where they can actually begin playing.  Give players a quick tutorial era where they can get a chunk of abilities and develop their own style within a few fights.  After about ten kills, I should be choosing if I want to be on my way or if I want to screw around with brutality.  No other genre could get away with the tedium that\'s acceptable in MMORPGs.  But, as I said, I haven\'t really gotten to play for more than a few hours, so my criticism may not be entirely fair.  Anyhow.  I\'m going to try and ride my bike for a bit before the movie, so I\'ll be taking off now.  urlLink Full Post and Comments    '),(2295,2680773,'27,March,2004','2004-03-27 00:00:00','   Dawn with the Sickness   Saw  Dawn of the Dead  (the remake) last night, and I think I agree with the widespread critical analysis of it as the original without the underlying social commentary.  Not necessarily a bad thing.  It\'s a crazy zombie action movie, and it\'s definitely fun.  What  Dawn of the Dead  (2004) does well is tell a lot of mini zombie-stories through the basic structure of survival in a mall.  As if you were shopping it, the movie has a lot to choose from...  And you\'re not going to buy everything.  There are scenes and plot twists in  Dawn of the Dead  (neuvo) that are hopelessly ridiculous and would likely spoil the movie if you were allowed to ponder them as in the film\'s inspiration.  But walk away with one shopping bag of savvy finds, and you\'ll be delighted.  One of my favorite moments was the addition of \"Andy\" a gun shop owner alone on his rooftop across the parking lot from the mall that the central characters dwell in, whose communication with Ving Rhames\' removed cop is the centerpiece of a montage scored with a lounge cover of Rage Against the Machine\'s \"Down With the Sickness\".  Other added characters don\'t work out as well; the head mall security guard, for instance, undergoes a major attitude shift almost instananeously.  Some moments work (the communication breakdown during the rescue of a truck full of \"refugees\"), others don\'t (the escape from the mall is generally silly).  But everything gels well.  Zombie fiction, however, is in a state of crisis with the continuing presence of fast zombies (28 Days Later) which goes against traditional zombie lore.  Directors seem to feel that fast zombies will help make zombies more terrifying, and they\'re right.  Fast zombies are much more frightening than true, slow zombies.  But.  They are made so overpowering that human survival becomes completely implausible.  Directors create a situation that  requires  a deus ex machina in order to drive the story.  This breaks any added suspense having a fast zombie creates.  We need slow zombies back.  Now, more than ever.  urlLink Full Post and Comments    '),(2296,2680773,'23,March,2004','2004-03-23 00:00:00','   \"Not Cool at All\"   A great  urlLink interview  for anyone who loves/hates/has heard of Ryan Adams.  urlLink Full Post and Comments    '),(2297,2680773,'22,March,2004','2004-03-22 00:00:00','   Desktop of the Pops   Links to some of the best new/nearly new music that clutters my Windows Desktop.  Of Montreal -  urlLink Disconnect the Dots  New Wave plus Vocal Pop plus Kevin Barnes Style Psychedelia - sounds like a train wreck but works as well or better than anything to come out of the Athens twee scene in years.  From the forthcoming  Satanic Panic in the Attic   TV on the Radio -  urlLink Dreams  Unbelivably heavy and gloriously uplifting at the same time, TV on the Radio is the flying buttress of the indie rock world.  From  Desperate Youth, Blood Thirsty Babes   Out Hud -  urlLink  Dad,There\'s A Little Phrase Called Too Much Information  How could you possibly dance to this?  From last year\'s  S.T.R.E.E.T. D.A.D.   The Decemberists -  urlLink Shiny  This straight forward country/folk single from The Decemberists proves that they can write good songs that don\'t involve pirates.  From  The Tain.   Xiu Xiu -  urlLink Clowne Towne  Far closer to art than entertainment, Xiu Xiu are destined to dominate the mix tapes of stuck up collectors for years to come.  From  Fabulous Muscles   Kanye West -  urlLink Through the Wire  At first, rapping through a broken jaw seems like an obnoxious gimmick, but West\'s nimble lyrics and solid flow turn it into a inventive disruption of an increasingly predictable genre.  From  The College Dropout.  urlLink Full Post and Comments    '),(2298,2680773,'21,March,2004','2004-03-21 00:00:00','   Errol Morris\' Fog of War   In the same way that people apply Sun Tzu\'s  Art of War  to  urlLink practical use , I feel like the eleven lessons that make up the structure of   urlLink Fog of War   have similar application but for a culture where combat is generally avoided.  In the film, McNamara looks back on his role in major points of 20th Century conflict and grapples with his successes (the Cuban Missile Crisis) and his failures (Vietnam, the fire-bombing of Japan) and attempts to explain what is important to learn from those events in hindsight.  As he meets with former foreign opponents to openly discuss history, he discovers that they had very different national interests than the Cabinets McNamara was part of had assumed.  Instead of Sun Tzu\'s \"Burn the bridges\", Morris and McNamara present \"Empathize with your enemy\", referring to how nuclear war was avoided in the Cuban Missile Crisis and ultimately the United States\' failrue to control the situation in Vietnam.  While McNamara is not saying that conflict is always avoidable, his thesis in the film is that the time we spend in war can be minimized with practical values, with an understanding of the flaws of human nature, the flaws of reason.   Fog of War  is masterfully shot, edited and scored and is an incredible insight into conflict in the 20th Century.  Definitely a documentary deserving of its Academy Award.  urlLink Full Post and Comments    '),(2299,2680773,'21,March,2004','2004-03-21 00:00:00','   Proof that  urlLink God has a sense of humor.  urlLink Full Post and Comments    '),(2300,2680773,'20,March,2004','2004-03-20 00:00:00','   Live Action Anime   What would happen if you made a live action Anime?   urlLink Casshern  is what.  Can you call it Anime?  It sure looks like Anime...  But the actors aren\'t animated.  (Are the  Mary Poppins  subset of films considered cartoons?)  You might argue  The Matrix  is already an example of this.  But I\'d respond that what  Matrix  does is take a lot of Anime style and paste it onto an American action movie.  It\'s kind of like the Chipotle of movies.  Yeah, it\'s a burrito, but too many things have been tweaked to make it a  real  burrito.  urlLink Full Post and Comments    '),(2301,2680773,'20,March,2004','2004-03-20 00:00:00','   RIAA Realizes... There is a Better Way   From the  urlLink Chi Trib  today.  Apparently, the new boom in SACD / DVD-Audio formats (Elton John\'s  Goodbye Yellow Brick Road  has been re-released as a hybrid disc, by the way) is a way to discourage piracy by presenting better audio quality and more features in a physical format.  Brilliant.  This is good news all around.  Hybrid discs are still rippable to the mp3 format, still unlimitedly burnable to new CDs, but the original discs have quality that is uncapturable.  Sharers rejoice because they can have the freedom to share, to test drive music and to become more informed on what\'s out there without having their minds flogged by radio.  Record companies hold onto sales even if people have downloaded an album.  Audiophiles and music lovers reap the benefits of having a high quality audio format being pushed in the mainstream (and will have to replace their old Elton John CD\'s).  And if you\'re Sony right now...  Keep crying about album sales and let no one realize the kind of money you\'ve been minting with the sales of mp3-playing devices and now SACD-compatible players.  You are so money and you know it.  urlLink Full Post and Comments    '),(2302,2680773,'19,March,2004','2004-03-19 00:00:00','   I guess Jack Kelley\'s involvement with WJI is a big was, because they have (in the very least) taken his name off the listing.  Here is the  urlLink google cache  so that you can be sure I am not lying to you.  urlLink Full Post and Comments    '),(2303,2680773,'19,March,2004','2004-03-19 00:00:00','   The Real Humor is in the Layers   During the State of the Union address, W was caught in an unforseen rhetorical problem when anti-Patriot Act politicians applauded after the statement, \"Key provisions of the Patriot Act are set to expire next year.\"  Bush continued to say that it should be renewed, but his point was lost in the shuffle.   urlLink Easterblogg  notes today that there was another layer of humorous applause in the segment.  urlLink Full Post and Comments    '),(2304,2680773,'18,March,2004','2004-03-18 00:00:00','   Proof That You Don\'t Have to Be Young and Black...   ...all you need is  urlLink no sense of ethical responsibility .  Here\'s the cut-and-dry \"we\'re gonna sink this bastard\" version of Jack Kelley\'s ouster for plagiarism and fabrication from  urlLink USA Today .  Here\'s my you-heard-it-here-first moment of the century, discovered by my tireless research (Google)...  Jack Kelley is (was?)  urlLink faculty  at the  urlLink World Journalism Institute .  The reason I\'m familiar with WJI is because flyers for it are frequently passed on to me by my pastor.  That\'s right, WJI is a journalism school focused on Christian ethics.  From their mission statement (linked above):   There is one primary reason why the World Journalism Institute should be committed to the education of young journalists: it comes directly from the need to be faithful to the Christian example of accurately reporting (e.g., being reliable eyewitnesses) the work of God in today\'s world.    We at WJI believe it is now time to implement a new phase of this statement in order to help turn out journalists capable of presuppositional reporting.   O I love the smell of irony in the morning.  urlLink Full Post and Comments    '),(2305,2680773,'18,March,2004','2004-03-18 00:00:00','   I\'m Just Like All These Other Collector Losers, I   Thought I\'d diversify the listing of things a little bit by adding some of the better comics I\'ve been reading.   urlLink  Hell   - (preview available)  Gorgeously drawn and colored.  Hell of the title is actually an island of genetic experimentation (Eden) gone horribly awry.  The coloring is lush; there is probably no better word for how well its details and pallettes work.  And the drawing... Done in a graffiti-ish style, with sharp, jagged emphasis along curving and twisting forms, captures action and motion as well as (if not better than) anything I\'ve seen.  I\'m only through issue 1 (there are four on stands), but if the story holds up, Hell could assert itself as one of the best action comics in production.   urlLink  B.P.R.D.   - ( urlLink preview ) With the Hellboy movie coming out, it seems providential that I discover B.P.R.D., a Hellboy spin-off as sure as they come.  This comic makes me very curious about the film.  Yeah, the first issue ends with a lot of screaming and shooting and punching and monster noises.  But you gotta have some of that when you\'ve compiled a task force of tortured superfreaks.  There\'s an air of \"League of Extraordinairy Gentlemen\", but I get the feeling that much more has been invested into the story than with that book.  We will see when the second issue comes out in April.   urlLink  The Pogostick   - Our hero is a mixture of Milton from Office Space and Jimmy Corrigan of Smartest Kid on Earth fame.  As such, Pogostick is a dark comedy of office interrelation hijinx and tragedy.  Al Columbia is an amazing comics artist, yet (like many underground comix) his work is difficult to come by when it\'s out of its run.  It\'s good to be reading it during the cycle.  Some quick notes as well...   urlLink  Dead@17   is essentially Buffy the Zombie Slayer and a little disappointing because of it...  urlLink  The Clockmaker   folds out to tabloid size and wins some art points because of it, but the story itself is plodding (I\'m on issue 3) especially for a short-run comic...  the new  Optic Nerve  is on par for Adrian Tomine; I\'d reccommend the previous issue (#8) if you\'d like to read an exemplary story of his.  urlLink Full Post and Comments    '),(2306,2680773,'17,March,2004','2004-03-17 00:00:00','   Ripped from the headlines...   Kudos to  urlLink Eric Zorn\'s amazing blog  for this one.  If you\'ve ever seen even one episode of \"Law and Order\" or \"Law and Order: SVU\" or \"Law and Order: Criminal Intent\" or \"Law and Order: CSI\", you must check out the  urlLink Random Law and Order Plot Generator .  My favorite so far:   A headless corpse is discovered in a park by out-of-towners. Lenny and Curtis initially pin the crime on the victim\'s lover, but after insurance fraud is revealed, they arrest a down-on-his-luck executive. McCoy and Kincaid prosecute, but McCoy must skirt the law to win. The old DA takes a drink and says \"Advice is like castor oil, easy to give but dreadful to take.\" Tony Danza guest stars.    These are the actual TV Guide recaps of tonight\'s episodes on NBC:   After a sweatshop owner is beaten to death, suspicion falls on the husband of a worker he impregnated.     The discovery of a woman\'s body in the Hudson River leads to an investigation that uncovers a scandal involving attorneys and a judge.    I wonder if McCoy will get his man.  urlLink Full Post and Comments    '),(2307,2680773,'16,March,2004','2004-03-16 00:00:00','   My Michael the Car   \"Michael Jackson wanted to star in a movie about  urlLink a man who turns into a car  thatâ€™s driven around by a little boy.\"  urlLink Full Post and Comments    '),(2308,2680773,'15,March,2004','2004-03-15 00:00:00','   Get an Axe...    urlLink ...it\'s a trap.  urlLink Full Post and Comments    '),(2309,2680773,'15,March,2004','2004-03-15 00:00:00','   Spanish Bombs   It\'s impossible to tell just what the swing in the Spanish vote was caused by. Any of the reasons I\'ve heard (from the polls being wrong, to the Iraq war, to the heavy-handed accusations at ETA) have validity on a micro level with individual voters. It\'s silly to think that the entire swing was motivated by one particular response.  Democracy is wild and diverse...  The issue here, in my mind, is that what happened went  urlLink exactly as predicted .  From a piece on Iraqi jihad that the Guardian cites being published in December:   \"If [Spain\'s] troops remain in Iraq after the attacks, a Socialist victory is practically guaranteed and withdrawal of the troops will feature in its election manifesto.\"   This is what makes the election results frightening however fair and justified they may be.  There is a plan in action that we (again) did not foresee.  urlLink Full Post and Comments    '),(2310,2680773,'15,March,2004','2004-03-15 00:00:00','   The [Start] and Fall of Matte   Noticed today that alt.culture magazine Matte has given up its domain and is no longer even in Google.  A sad sight (site) because the folks at Matte were putting out one of the best magazines on the shelves.  What did Matte do well?  Well, for one, its design balanced text with art in interesting ways (like a black and white Print almost).   urlLink Here\'s a sample .  Tied to that, it was terrificly readable.  The writing was handled maturely without being too heavy-headed, and the design was enough to draw you into an article you might otherwise skip over.  Plus, there were just a  lot  of in depth articles.  Reading 30% of the magazine was like reading everything in two issues of Rolling Stone.  Even aside from these basic foundation points, Matte had a lot going for it.  A large advertising base with pricey rates (for a small press publication) was in place from the very first issue.  They had a publisher belaying, at least, part of the up-front cost.  And distributors.  A large staff and what seemed to be a good support structure...  So what went wrong?  Well, I have guesses...  The change in printing between issue 2 and 3, moving from a thick, unfinished heavy stock like The Believer to a more traditional paper quality, seems to hint at the economic plan of the magazine falling apart.  I\'d gander that the magazine wasn\'t able to pull in subcribers with its $20 for 4 issues plan.  I also imagine its abstract covers weren\'t able to attract the eye of buyers on a newstand end.  The third side of effective distribution for a print pub these days is the web, but Matte\'s website was static and fairly minimalist.  The readership for Matte  should  have been much larger, but its most basic draws for building that readership were fairly impotent.  In my dreams, I imagine co-editor Anne Elizabeth Moore\'s seeping-through-the-pages pretension (which may or may not be fair to her outside Matte) to have turned enough people off to lead to the withering of Matte.  But, I\'ll place my bet in reality on the business model collapsing because of the mag consistently turning its back on effective marketing techniques.  I don\'t know who is at fault for that.  Still, Matte\'s collapse has an important lesson in it for the small press...  Even if everything goes right, it can still be blown.  urlLink Full Post and Comments    '),(2311,2680773,'14,March,2004','2004-03-14 00:00:00','   Delegates in the Dark   The two man against 193 men parliamentary Double Dragon style fisticuffs reported in a prior post apparently looked more like  urlLink a musical  than River City Ransom.  I don\'t know if this is good, bad, or in between.  But, yes, that\'s a flying toupee.  urlLink Full Post and Comments    '),(2312,2680773,'11,March,2004','2004-03-11 00:00:00','   Does  urlLink this story  imply that 2 guys physically fought everyone else in the room?  urlLink Full Post and Comments    '),(2313,2680773,'07,March,2004','2004-03-07 00:00:00','   SHOCKED   Jayson Blair seemed shocked and unprepared to deal with hardball questions on MSNBC\'s Hardball earlier today.  Hm.  Nobody ever said the guy was a genius.  O wait!  They did!  To make matters even zanier....  The guy who once pretended to be a journalist is now pretending to be a  urlLink superhero .  Not enough pain for you?  In his book, Blair blames his problem on \"drug and alcohol addiction, undiagnosed manic-depressive disorder and the pressure of being a young, black reporter at a large, competitive newspaper.\"  On Hardball, he said he was at home sick when he was first lured into fabricating stories.  If he would mention the  urlLink depravity of man  he\'d have a straight flush.  Blair ducked and weaved through the interview attributing all he could on the \"old\" Jayson Blair, even looking like he had never thought about how he had shown a total lack of respect for the NYT readership.  The moment of the Hardball interview when he shook Chris Matthews hand and said with a smile, \"Thanks for having me here,\" made me physically nauseous.  If the New York Times has done one thing right, ever, it\'s currently refusing to acknowledge that Blair exists as a media figure.   Update!    urlLink Bookslut  caught a tidbit on Slate with another Blair excuse \"He was do-gooding. He had visited a man in prison, taught at an inner-city school, and visited his girlfriend\'s childhood weekend home in Morris, N.Y. Also, he was tired.\"  urlLink Full Post and Comments    '),(2314,2680773,'07,March,2004','2004-03-07 00:00:00','   I\'ll Be Your Friend If You Give Me a Dollar a Day   Excited to be joining the Chicago Blogs web ring.  You\'ll see the link added to the left in the links column... Click on the arrows in either direction to move through the blogs in the ring or hit  urlLink \"chicago blogs\"  to go to the central site.  Initially, the other bloggers will probably view us as a pack of tools without the concise, poetic wit of their own pubs.  But, over time, they will come to realize that our prose is so poetic, our subtlety so concise, our wittiness so witty that they were fools to ever think elsewise.  They will shower us with comments like rose petals from the city walls, paving roads where we are thinking about walking (but will probably change our minds), reinforcing corroded infrastructure, asking for our autographs and hollering \"Yo da man!\" while we try and sink a putt for birdie on the 7th hole.  And then we will quit.  The world will fall into the black void of silence once again... Until they post in their own blogs.  At which point, we\'ll be nothing more than the  urlLink Suck.com  of the 21st century.  urlLink Full Post and Comments    '),(2315,2680773,'07,March,2004','2004-03-07 00:00:00','   Maybe I\'m Weak   But  urlLink these guys  make me laugh.  urlLink Full Post and Comments    '),(2316,2680773,'05,March,2004','2004-03-05 00:00:00','   This is the Washington Post we\'re talking about, right?   From the Washington Post:  a comprehensive analysis and explanation of their  urlLink fairness and accuracy policies , including a discussion of the use of anonymous sources.  I, like many people, despise anonymous sources.  It\'s nice to hear someone apologize for having to use them.  But much of the problem comes from protection of people whose information is not ni fact sensitive while skewering up those who are in actual danger.  There isn\'t a consistency in the standards.  What that means is that the press gets manipulated by its interviews, especially in Washington.  urlLink Full Post and Comments    '),(2317,2680773,'05,March,2004','2004-03-05 00:00:00','   Just be in Love when You Scream that Song II   Here we go.   Violent Femmes  -  Johnny  : I\'ve been appreciating a bit more how good the Violent Femmes are since seeing  urlLink Guster  cover their self-titled debut on MTV2 last Friday.  My first exposure to the band was during high school, sitting in a car with two other not-theater-type theater friends while they with great anticipation popped in an album.  Like the whole environment of the evening and my relation with those friends, I remember the music being roughly defined, dark and still somehow poppy as hell.  So, I\'ve always had a considerable warm spot for the Femmes, but I don\'t think I ever delved too deeply with them.  \"Johnny\" is a track I had never listened to.  It has all the angry variations of emotion that Femmes are known for but in a song that pulls itself back into a simple eulogy every time it starts to break away.   Guided by Voices  -  Echos Myron  : Honestly, I like it just because it\'s a catchy little lo-fi number; a toying melody against an electric wall of noise.   Elvis  -  (Marie\'s the Name) His Latest Flame  : I\'ve spoken my love for \"In the Ghetto\" before for its pure bizarrity.  But before Elvis was overblown (or  urlLink switched places  with an impersonator), he was singing simple numbers that - so much more than his later efforts - delighted and shocked the world.   Josh Rouse  -  Love Vibrations  : I\'m saying \"Love Vibration\" even though I probably listen to \"Slaveship\" a bit more, because I feel like if you want to get to know  urlLink Josh Rouse , \"Love Vibration\" is a very good place to start.  Think  urlLink Conor Oberst  if he was obsessed with soul and could feel the human emotion of love.  The album is appropriately titled  1972  , right before the 70s turned sour.  (I think.)   Scott Miller and the Commonwealth  - Any Song off Thus Always to Tyrants. : Alright.  I know I broke the rules, here, but...  If you have any love for rock, if you have any love for country, if you have any more love in you than Conor Oberst, you need to stop by your local music shoppe and nab a copy of  Thus Always to Tyrants .  Steal it if you have to.  Scott Miller is not Radiohead.  He\'s not obsessed with form and complexity.  What he does is write a song that will hijack your brain and make you believe there is a guitar in your hands and a rock star in your heart.  Perfectly crafted, perfectly delivered, perfectly captivating.  Nothing more, nothing less.  Yum!  urlLink Full Post and Comments    '),(2318,2680773,'05,March,2004','2004-03-05 00:00:00','   Picture Pages, Picture Pages   Draw simple line art in  urlLink The Scribbler  and generative lines give the piece an inky textured look.  I spent about 4 hours playing with the program this evening and made some fun stuff.  Share your own art in the comments section.   urlLink Gollum!   urlLink Homeless Wolverine finds pot of gold!!!   This is where they stop being lame as Corey Wilson joined in the fun:   urlLink Imp  by Corey  urlLink Old Dude  by Corey  urlLink Flower Picker  by me!  urlLink Dude  by Corey  urlLink Flowers  by Corey  urlLink Flower Pot  by me!  At this point, I couldn\'t think of anything else to make... Corey had one more in him   urlLink Lamp  by Corey  Happy Weekend!  urlLink Full Post and Comments    '),(2319,2680773,'05,March,2004','2004-03-05 00:00:00','   At the Movies   Today in the Sun-Times (Still 35 cents!), Roger Ebert writes  urlLink one of the best movie reviews I\'ve read in a long time .  What movie does he review?  It doesn\'t matter.  And that\'s the point.  A movie review should be interesting and hold your attention even if you don\'t know a thing about the film.  The lead paragraph in this review is superb.  urlLink Full Post and Comments    '),(2320,2680773,'03,March,2004','2004-03-03 00:00:00','   urlLink Holy golden nuggets!  urlLink Full Post and Comments    '),(2321,2680773,'02,March,2004','2004-03-02 00:00:00','   Bug me not   A neat way to get around registering for access to online newspapers.   urlLink Bugmenot .  urlLink Full Post and Comments    '),(2322,2680773,'02,March,2004','2004-03-02 00:00:00','   Air America Radio, The Left AM   So, it appears the much-ado, thankfully-not-named-Indie-Network, politically-left radio network  urlLink Air America  is about to dive into the pool with air-time locked in New York, here in Chicago and a handful of other cities.  I\'ve decided to create a short list of both reasons it could succeed and reasons it could fail; this way, I\'ll be right regardless of what happens.   Reasons Air America Could Succeed   1.  Al Franken - He\'s funny.  The guy is good at sifting through rhetoric and has become the classic antithesis of Rush Limbaugh because of it.  Accordingly, he\'s been able to hold his ground in debate against conservative rhetoricians like Sean Hannity.  A name I\'ve seen for his program is the O\'Franken Factor. Cute.  2.  Jeaneane Garofalo - She\'s funny!  Another talented comedi person and the Hannity killer for the network... Because Sean Hannity is so Rah-Rah-publican and so so unentertaining to listen to if you disagree with him, Garofalo stands an excellent chance of picking up stray listeners (even conservatives) who are tired of that idiotic Hannity theme song.  3.  Political Satire - It\'s funny!!  Liberal, conservative, it doesn\'t matter... Satire is unassuming and coy.  Basically, AAR has assembled representatives from every quality poltical satire entity in the past decade.  Franken is from Saturday Night Live, Garofalo was involved with Michael Moore\'s TV projects, Lizz Winstead is somewhere in the mix and a co-creator of the Daily Show.  Is that gold or is that platinum?  You can decide.  4.  There is a Void - Liberals who are tired of repetitive music stations have escaped to talk radio along with their conservative brethren.  They listen to conservative programming and end up calling in to complain, at which point they are obliterated by well-trained conservative hosts.  Yes, there are liberal talk radio hosts even on otherwise conservative networks, but the conservatives are the ones moving up in ranks (for whatever reason).  Somewhere there  is  a market, and you need only listen to Rush Limbaugh for a few minutes to hear voices from it.   Reasons Air America Could Fail   1.  Al Franken - We\'ll start with the easy points here...  Franken comments on commentators.  Now that he has his own show, he\'s going to be unable to listen to and study those commentators (**cough**Rush**cough**) and direct commentary has to be the center of the show.  While we\'ve seen Franken push policy, it\'s generally been in the wings.  On a side note, how many weeks before trying to call in to Limbaugh\'s show becomes a regular staple?  2.  Garofalo - Her fame is from generic comedy.   Yes, she\'s been vocal about political issues, and being vocal is the basic prerequisite for radio.   But is she going to be able to spit out dates, cases and records like her counterparts whose mainstream identity was born from political debate?  Maybe.  If she\'s not at the start, it\'s going to take her some time to get into the swing of things...  Is she going to be interested enough in this program to stick with it that long?  As much as she is an asset, she is also a tremendously large question mark.  3.  NPR - In the radio world, NPR is like a badge of courage for the liberal educated and political talk is unbearable schlock.  Is it going to be necessary for AAR to pull liberal listeners from NPR (where they\'ve found refuge from the likes of Limbaugh)?  Is it possible given the social status of NPR?  Not saying NPR  is  liberal, just saying it is something of a haven.  4.  What Good\'s the Radio? - Franken, for one, admits he is in this deal partly as an effort to get Bush not-elected come November.  But add a lack of radio experience with talk radio\'s traditional impotence when it comes to presidential elections, and you could get frustration... especially for investors.  Ask yourself.  What good did Limbaugh ever do?  The AAR website itself places the foundation of conservative talk radio in 1988 when a (R) was in the Oval Office, which boomed through 1992... when Bill Clinton (D) won the presidency.  It continued to steadily grow through 1996 when...  Bill Clinton (D) won again.  After the 2000 election results, certain individuals (including candidate Al Gore) began partly blaming talk radio for Bush\'s win (robbery, of course, was the major complaint).  Thus, the polterzeitgeist was born.  But it\'s never proved to be anything substantial; it\'s obsessing over gnats while you\'re getting stung by bees.  Here\'s a column on this issue from  urlLink a liberal perspective  in an effort to prove I\'m not being ridiculous with this point.  Anyway, I think I\'ve covered my bases.  Back to you.  urlLink Full Post and Comments    '),(2323,2680773,'02,March,2004','2004-03-02 00:00:00','   Traditions I Don\'t Understand   Sorry.  Feeling a little brain-fried today so no diatribes.  But here\'s a link to a Mid-East tradition I don\'t understand yet find fascinating...   urlLink The aforementioned link.   If anyone\'s interested, I changed the splash page over at the  urlLink Nation of Legacy  sub-site as part of my ongoing effort to beat the worldly excesses from RF leaving a much purer, better site.  urlLink Full Post and Comments    '),(2324,2680773,'01,March,2004','2004-03-01 00:00:00','   Dispatches from the Commonwealth   Been a fan of  urlLink Scott Miller  since he assembled the Commonwealth and released  urlLink  Thus Always To Tyrants   so I was excited to read about a recent whistle-stop tour on the  urlLink Amtrack Crescent .  Jack Neely of Knoxville\'s MetroPulse stayed along for the ride and crafted  urlLink \"This Train Goes On\"  a long form piece of rock journalism that blends remnants of the Civil War, public transportation and rock and roll into a portrayal of region-by-region perceptions of Miller on this interesting tour.  urlLink Full Post and Comments    '),(2325,2680773,'01,March,2004','2004-03-01 00:00:00','   Porn on Beta   A couple weeks ago, I was casting out the question, \"Was porn ever made on  urlLink beta ?\"  Not because of a particular desire or need for it, but because of the  urlLink popular   urlLink legend  that the porn industry\'s exclusive support for VHS was a major force in the demise of Betamax.  You most commonly hear the story in conjunction with a defense of  urlLink porn\'s importance in driving new technology  particularly the internet.  If the accounts are correct, then porn on beta is something of legendary status in the annals of antiquated technology.  One band from Vancouver, at least, has  urlLink acknowledged this  (although, their bio avoids it).  Unfortunately, the answer I fished up wasn\'t nearly as exciting as the mythos of the question.  One respondent attested to owning a fairly large collection of porn on beta.... which ruins my hope that such an item is a monumental rarity.  He also claims that the fall of beta mostly had to do with aggressive marketing of VHS by JVC, including some ultra-discounted distribution in the professional market.  (This may be where the porn myth arises.)  The claims seem backed up by the  urlLink Urban Legend archive , which outlines the marketing battle and resulting price war.  As much as I hate to abandon this novel story, it appears the facts are forcing me to do so.  urlLink Full Post and Comments    '),(2326,2680773,'30,April,2004','2004-04-30 00:00:00','   How the Lion Lost Its Pride   There are new comics reviews up at  urlLink Pop Matters  including my review of Paul Hornschemeier\'s  urlLink  Mother Come Home  .  My goal was to reflect on the central symbol of the comic, hopefully broadening the possibility of its meaning for the reader.  The Pop Matters style is interesting because at a root level it is searching for the same type of criticism that we have been with Robots Fighting, yet its large readership also looks to the site for clear opinions.  It is taking me a few reviews to be able to balance things in a way I feel is appropriate without seeming too constructed or too vague.  I feel like I\'m getting there with the  Mother Come Home  review, although my attempt to move beyond the symbolism to actually rate the book is a tad weak.  urlLink Full Post and Comments    '),(2327,2680773,'26,April,2004','2004-04-26 00:00:00','   These Are the Vistas   Any two things moving with different rhythms will at some point for some time reach a synchronicity.  If you\'ve ever ridden a swing next to another person you\'ve seen the effect in practice, when without effort you swing  with  someone you had been swinging opposite of.  But the synchronous rhythm does not last perpetually and soon your swings are disjointed again.  Enter the joy and the frustration and the joy in frustration of post-bop jazzers  urlLink The Bad Plus .  Their compositions soar over jagged and scattered beats until they rise over a particular peak and reveal some perfect, hidden vista.  But just as you begin to acquiant yourself with the surroundings maybe even settling in, the music\'s wings are taking you away to some other secret majesty.  That said, their contemporary rock covers have warmed to the populous, giving the listener a thread of familiarity with which to trevail the rocky and glorious heights of The Bad Plus\' breed of jazz.  These tracks are like Sound Picassos, breaking their sources apart and reforming them so you can hear every angle at once...  But despite the brilliance of these moments, does The Bad Plus risk becoming a novelty act because of them?  It\'s a question the band must be aware of and an awareness that can only heighten the danger.  As their repertoire of covers builds, will anyone care to listen to their original compositions?  It\'s true there are a number of pieces that reach places more beautiful than anything in The Bad Plus\' explorations of common locales, but for average listeners the trails to get to those places are foreboding.  \"Iron Man\" is possibly the cover least coy about its inspiration and as a result the least graceful, yet this is what the audience of last week\'s show at Martyr\'s demanded to hear.  I don\'t think it was because of a particular contempt for the original compositions - the crowd grooved to all of them - but because, given the choice we\'d rather discover an antique pocketwatch in the floor boards than to face hunger and cold as we traverse the Andes to find  urlLink the clock the world spins by .  urlLink Full Post and Comments    '),(2328,2680773,'26,April,2004','2004-04-26 00:00:00','   Fox Brings All the Boys to the Yard   I try not to say this too often but,  urlLink whoa-zay .  urlLink Full Post and Comments    '),(2329,2680773,'22,April,2004','2004-04-22 00:00:00','   Volume 2   If  urlLink  Kill Bill vol. 1   had been what  urlLink  Kill Bill vol.2   is, I probably would have been apathetic to the entire series.  For certain, I didn\'t want talking.  I told Kristi back in the day, \"If there is no talking in the entire movie, I will be happy.\"  And in the end (or middle?), I was impressed with Quentin Tarrantino\'s ability to embed complex themes into the two hour fight sequence with as little talking as possible  and  to still keep it constantly fresh with an amalgam of international action movie styles.  (Was just thinking, incidentally, about Cottonmouth\'s board room scene being Tarrantino\'s way of suggestively defending his creation as a legitimate fight movie.)  That said, at the cliffhanger conclusion of Volume 1 what I wanted for Volume 2 was not non-stop action, I wanted talking.  I didn\'t have any desire to see the Bride storm through dojos for another 2 hours.  Yeah, some swordplay was going to be necessary, but I understood the world the characters lived in and wanted to, obviously, the plot to reach a full-fledged fruition but also to be given a deeper look into the mythology of the characters.  And it\'s what we get.  What seems clear to me is that Tarrantino is one of the best character developers in cinema.   Kill Bill  (and arguably the entire filmography) is a handful of interesting characters tossed together in a basic framework.  They invent their own story and Quentin just tweaks the knobs and dials...  \"What happens if we have Uma Thurman, Darryl Hannah, and David Carradine all get their training from the same guy?\"  \"Let\'s set it so that David and Uma and Michael Madesn have the same one-of-a-kind sword.\"  Most of the intricacy of the character\'s relationships to one another stays in the subtext but feels nevertheless more true because of it.  The story is controlled by the characters, and as such allows the actors to dazzle.  So, how Tarrantino approaches some scenes is, in some sense, dependent upon how the actors control their characters.  In other cases, Tarrantino adjusts the controls... camera angles, lighting / color design, exposure, whatever...  and the characters play through.  There\'s a symbiosis to the whole thing.  A few years ago, I winced at the idea of Tarrantino doing a  Dirty Dozen  style World War II flick; I couldn\'t see him being able to make the transition from Resevoir Dogs or Pulp Fiction into that...  But, here in  Kill Bill , he  proves  his proficiency at character design and his intent to let that design drive the story.  urlLink Full Post and Comments    '),(2330,2680773,'20,April,2004','2004-04-20 00:00:00','   iRecycle   If you\'ve got free iTunes songs from Pepsi\'s big promotion and have no real interest in using them (I\'m in the boat, where unless I have a full cd, I have no interest in having an ugly file format on my comp), check out  urlLink Tune Recycler  from the folks at  urlLink Downhill Battle .  Send them your codes and they will use them to repetitively purchase the albums of various independent artists (including  urlLink Pretty Girls Make Graves ).  Pepsi only expects 10-20% of all codes to be redeemed by the April 30 deadline, so this is a way to get the codes that you never intended to use active and make some indie artists a little bit happier.   Creative Use of iTunes #2:   I was on iTunes looking for something to spend my songs on (I ended up donating them via TuneRecycler but may end up buying an album or two in the future from Apple) and discovered that the service has high quality, quickly streaming samples for every song in the database...  What\'s more is that the clips seem to be smartly produced, picking highlight moments of the songs.  It\'s exploitable for  anyone who would rather buy a physical copy.   Weird iTunes iTem   I signed on looking for tracks from the Guster-covers-Femmes show and was disappointed at the limited number of songs from the event (five altogether, with one being a Guster original)...  But I did find the Meow Mix of Guster\'s Keep It Together, which in the spirit of the Goldfly bonus track replaces all the lyrics for the entire album with meowing.  If the sugar-coated lyrics of Keep It Together weren\'t your thing but you loved the music, the Meow Mix album is $2 less than the actual album.  Save money!  Love Guster!  urlLink Full Post and Comments    '),(2331,2680773,'17,April,2004','2004-04-17 00:00:00','   Musical Chairs : Your Weekly Challenge   I\'ve decided to post a weekly challenge (every Saturday)  for the writers and readers of this blog; no prizes aside from the satisfaction of beating your peers in competition.  Entries must be received by the following Friday, in the form of comment, e-mail or blog post/link.  All very simple.  This weeks challenge is \"Musical Chairs\" and here is your task.  The new Broadway cash-in is to take a washed-up artist\'s discography and reenvision it as a heart-warming story with ubiquitous singing and dancing.  Abba, Billy Joel, Boy George... They all have their own musicals and people other than themselves are making money off of them!!!  I would like to make money too.  So, you are challenged to take any non-concept album (no Yoshimi, no Sgt. Pepper) from any artist (including the relentless V.A.) and come up with a plot synopsis with song cues.  Entries should include a list of characters and should utilize no fewer than 5 songs from the given album.  Several independent judges will be pitched each of the musicals, and their opinions will triangulate the winner.  Degree of difficulty, creativity, and marketability will factor into the final decision.  Now is your chance to give Jon Bon Jovi (or the like) immortality!  urlLink Full Post and Comments    '),(2332,2680773,'15,April,2004','2004-04-15 00:00:00','   It\'s Winter Time for Spring!   A co-worker and I were walking home from work and turned onto a street that I had seen blocked off with a squad car and police officers several times during the course of the day.  As we walked down the street, and I related my curiousity over the blocking, we realized we were walking through snow.  Not a lot, just along the edges of the sidewalk.  Still it\'s been warm in Chicago the past week, so any snow that  might  still be lingering in the shade would certainly be gone by now.  We soon discovered that the snow was not snow but actually suds, like the lather of a bikini car wash.  Fumigating?  Street washing?  We listed through the possibilites unable to accept any of them... soon noticing white plastic sheeting bolted over any grass along the street.  And as we returned to our surprise at finding \"snow\" we suddenly came across real snow, and more extensive sheathing of the ground around us.  We were baffled.  Suddenly, it all clicked in my mind.  El tracks, brown stones, low traffic side street, manned road block, real snow on one side, fake snow on the other...  It must have been a movie shoot.  Sure enough,  urlLink The Weather Man  had been filming scenes that day.  urlLink Full Post and Comments    '),(2333,2680773,'15,April,2004','2004-04-15 00:00:00','   Soundtrack of Our Lives   One of the things I miss most about living in a dorm is the free exchange of new music.  Hearing a song wafting in from the other room and walking in to find out who and what it is.  If I were still living in the dorms, the song you\'d be hearing more than any other right now is a great pop piece called \"Artists Are Boring\" by the  urlLink Kingdom Flying Club .  Seemingly owing equal debts to the low-fi sound of Guided By Voices and Pavement and the polished power pop of bands like Material Issue and Weezer, Kingdom Flying Club comes to us from Columbia, MO.  For you Chicago area kiddies out there, the band actually reminds me a lot of Kleenex Girl Wonder.  Great pop sensibilities, but not afraid to get down and dirty (and fuzzy with their guitar sound).  \" urlLink Artists Are Boring \" is a fun slice of the band at their most polished.  Think Kinks circa \"Village Green Preservation Society.\"  The song starts with an irresistable piano intro and holds on to your ears \'til the very end.  The band even writes a great bridge, a lost art these days, with what I believe to be mandolin carrying the melody.  \"Hate Shirt,\" the first track of the album, goes the other direction with crunching guitar riffs and a thundering bass line.  Production is fairly clean and complements the band\'s sound nicely -- kind of like Ric Ocasek if he didn\'t suck.  If you order the album, Emergency Umbrella will even throw in free buttons and stickers. Can\'t beat a deal like that.  urlLink Full Post and Comments    '),(2334,2680773,'14,April,2004','2004-04-14 00:00:00','   Taking Seasame Back to the Streets   Thanks to the rise of audience specific cable channels and direct market videos, children\'s television has gone from a doe-eyed attempt to enlighten our kids to a brittle, plastic effort to sell Yu-Gi-Oh! cards.    Fortunately, all is not lost.  In an effort to rekindle the spirit of the Children\'s Television Workshop and, uhm, Rumpus Room, a group of DC artsy-types have created a kids show that tries something different.   urlLink Pancake Mountain  applies a rough around the edges punk rock aesthetic to the puppets and songs approach to children\'s television.  The web only pilot featured appearances from a host of indie celebrities including  urlLink Bob Mould   (as an evil executive)  urlLink Vic Chesnut  (singing a classic kids number) and  urlLink Thievery Corporation  (showing the kids how to get down).  Best of all, Fugazi\'s Ian MacKaye performs as the Evens in the more clever than it sounds  urlLink Vowel Movement .  Chances are Pancake Mountain\'s unique take on getting through to the kids will never make it over to the small screen, but if it does, it might just be a beautiful day in the neighborhood.  urlLink Full Post and Comments    '),(2335,2680773,'13,April,2004','2004-04-13 00:00:00','   More like \"Mr. Suck!\"   If you\'re like me, you\'ve probably been tempted to check out the Mr. Show inspired  Run Ronnie Run  but always decide against it at the last minute, afraid that it might not live up to Bob and David\'s other work.  Your instincts were right.   urlLink Here\'s why .  urlLink Full Post and Comments    '),(2336,2680773,'05,April,2004','2004-04-05 00:00:00','   Apparently, the parents of song parody emperor Weird Al Yankovic were  urlLink found dead  in their home from carbon monoxide poisoning.  I\'ve always felt Weird Al has lived in a sort of comedy dream world where the most horrible thing to happen to him was the surreal event of Coolio threatening to throw down over  Amish Paradies ; so, it couldl be interesting to see where he goes from here with his music.  Will he retreat further into his wacky novelty world or will we, perhaps, see a darker era of his music?   Or will he make none at all?  Whatever the case, it\'s a stunning loss for anyone and a reminder that our bodies are fragile to the most unexpected things.  urlLink Full Post and Comments    '),(2337,2680773,'05,April,2004','2004-04-05 00:00:00','   Dreams Do Come True...    urlLink WXRT  now is (finally) steaming over the internet.  Trust me, no matter how long and hard people in Chicago complain XRT is not what is used to be, it\'s still at least 10 times better than almost anything else around the country, and especcially here in Rockford.  Is it what is used to be?  No, of course not.  But the strength of the station is the music  and  the DJs.  People listeners have known for more than a decade.  I would have paid to hear Bobby Skafish and Lin Brehmer and Tom Marker again.  Now I get to hear them for free.  In the first of a series, I present songs I probably would not have known about if not for XRT Streaming on the net:  Jack Johnson --  Taylor :  I liked  Flake , and I do think Johnson is more than Dave Matthews lite.  The songs I\'ve heard from him do a great job of creating mood and feeling.   Taylor  does this perfectly and is a great acoustic guitar workout too.  At the very least it beats hearing Britney Spears for the 100th time on the Top 40 station out here or more Jackson Browne on the classic rock station.  urlLink Full Post and Comments    '),(2338,2680773,'02,April,2004','2004-04-02 00:00:00','   Forget You   Then you see a movie like  Eternal Sunshine of the Spotless Mind  and everything gets put into perspective.  A perfect piece of humanistic science fiction.  Aside from Charlie Kaufman\'s fairly regular brilliance, you also must remove your hat and then hand it over with a family heirloom to Michel Gondry for being able to pull this off so eloquently.  I could say more, but it is probably best that you don\'t know anything about the movie going in...  Yeah, it\'s one of those.  But like I said, it\'s eloquent about it.  urlLink Full Post and Comments    '),(2339,2680773,'02,April,2004','2004-04-02 00:00:00','   Red Right Hand    urlLink Hellboy Movie  is bad.  Even if Guillermo Del Toro and company were able to  urlLink fight off requests  for a Blue Hellboy or an Incredible Hulkboy, they seem to have let the soul of the character slip through their red, stoney hand.  On the outside, Hellboy is a bulky, crimson demon with shaved down horn and a collossal rock of a right hand, which as far as he has figured out is good for smashing things.  Movie Hellboy is very insecure with this form, having been raised by humans.  All Movie Hellboy really wants is to be a real boy, to freely associate with people.  He is driven by this desire.  In an effort to close the gap between him and the social life of man, he raises cats, watches cartoons, falls in love and, in general, becomes the quintessential, stereotypical, run-of-the-mill freak.  On the other hand, comics Hellboy (the real Hellboy) doesn\'t really care.  He sees himself as being part of humanity, even if a giant, abnormal part.  In fact, humanity sees him as part of humanity (the United Nations gave him Honorary Human status in 1952).  The central conflict for the source material Hellboy is his stuggle to eradicate his relation with demons/evil despite his Nazi-funded inception as the key to armageddon (something which plays out quite literally in the film).  While the difference between the two is subtle, it\'s integral to the identity and actions of Hellboy.  You see, Hellboy\'s superpower is not shooting flames out of his eyes or eating your soul; he can\'t lift objects with his mind, he can\'t fly, he doesn\'t command the weather and he certainly has no psychic ability.  What Hellboy does do is get the crap beat out of him.  As he hunts down what are essentially his own kind, Hellboy is pummelled to the edge of death, his strength and durability lasting him just long enough for heaven and fate to sign an agreement on how to get him out of the mess.  And the movie captures this (and, off the subject, the color pallette) fairly well.  But again, the movie misses the essence of his beatings.  Hellboy is humanity\'s scapegoat, taking the brunt of Hell\'s active attempts to seize the world, buying time for providence to patiently reveal itself.  He isn\'t Clifford the Big, Red Dog, he\'s the friggin\' hinge of of the apocalypse.  He has to fight to prevent it, but his encounters put him closer to the thing he is really meant to be by Hell.  If the spirit is gone, the shell is empty.  The plot that the movie is based on, \"Seed of Destruction\", certainly needed some filling in order to make a quality, full-length feature film.  But instead of adding additional plot or remixing the Hellboy series (like what we saw with  X-Men 2 ), Hellboy Movie\'s creators made scattered tweaks that are more destructive to the essence of the story than removing Hellboy\'s horns would have been.  For instance.  The frogmen of the comics have become hellhounds...  who lay frog eggs, are amphibious and have long tongues for grabbing prey.  Why!?  Why!?  It\'s a classic and pointless Hollywood tweak that helps to wipe away the myth and lore Hellboy stories are drawn from.  Another glaring alteration not directly related to the Hellboy character is Kroenen.  In the comics, the guy is a masked, laboratory-dwelling rogue among the Nazi collective that summons Hellboy; in Hellboy Movie he\'s a particularly noisy and particularly showy (in Star Wars Kid fashion) assassin.  While it adds a mini-boss to the film, it does nothing to build upon the source story in any meaningful way.  Again, we have a ineffectual, standard Hollywood \"solution\".  The gilled assistant to Hellboy, Abe Sapien has also been altered, this time into a nerdy jester, the intended comic relief.  I always felt like Abe played the straight man to Hellboy\'s smarmy attitude in the comics; it may just be the rebel in me, but I have a hard time believing that a comic relief character was necessary especially when he gets written out of the story in Act 2.  Also an ill-advised move since in the book Abe kills the Rasputin \"head boss\" (see the paragraph on Hellboy in action above for an understanding of why it works this way).   The film\'s creators sell-out minor elements left-and-right, while defending their ability to create Hellboy exactly as he appeared in the comics.  There are moments in Hellboy that work extremely well, the fight scenes, a scene where Hellboy spies on his love interest and another agent out on a friendly date, the general adherence to the source.  But the movie takes a slightly problematic adaptation and makes it more scatter-brained and incoherent while attempting the exact opposite.  The bar for action comic book movies is pretty low, though, and I really feel the movie is no worse than most of that crop.   Spiderman  was maybe a tad better, and  X2  and the early  Batman  movies look down from the tops of Gotham City skyscrapers at  Hellboy .  But for the most part, it\'s on par for the course.  The larger issue is that the filmmaking community still has no idea how to adapt a comic book, an absolutely ridiculous point since the storyboards are basically laid out for them ahead of time.  These aren\'t 400-page novels with internal narrative; they\'re very simple projects...  Few lessons have been learned from the exceptions (the action movies above plus  Ghost World  and  American Splendor ).  This saddens me because the  Watchmen  movie  is  going to get made some day, and I\'m afraid one of the great pinnacles of comics is going to be merely satisfactory.  urlLink Full Post and Comments    '),(2340,2680773,'31,May,2004','2004-05-31 00:00:00','   I haven\'t done this in a while, because honestly... there hasn\'t been anything interesting.  But here\'s some search strings that lead to Robots Fighting or The Machine\'s Still On.  -Slipknot Orange Jumpsuit -Pictures of People Physically Fighting Each Other -The Erotic Accordion -60 Dollars Robot  We are your source for all of these things and more.  urlLink Full Post and Comments    ');
INSERT INTO `msgs` VALUES (2341,2680773,'29,May,2004','2004-05-29 00:00:00','   I am finding myself believing very strongly in this idea, that any government that commits genocide should be deemed by the international community as commiting an act of war against global society.  Eventually, if the world can hang onto its gonads, the fuckers  urlLink will learn  that they\'re not going to get anywhere by slaughtering our fellow citizens.  That we sit around  urlLink feigning blindness to mass killings  is abhorrent.  It should be a given that any leader who abuses his power will be taken from it by the entire civilized world.  And it should be made explicit what is going on, to leave no question to the people of any region that those in power over them (elected or by force) have committed the single act that the world will not stand by to.  Say it like that; let there be no argument.  The second part of this is that the UN should have a substantial branch devoted to nation-building whether post-war or no-war.  Specialists should be brought into (or called upon by) wavering nations to assist in the planning and establishment of a solid economy and government.  If a military attack is the House of Representatives, with the biggest guns bringing their biggest guns, nation-building should be the Senate, with equal participation by all members.  This is the most crucial step to cutting off oppressive regimes: eliminating the desperation that allows for them.  Iraq is important to this in its effect.  If the war and nation-building is successful, we bring strength to a global battle for human rights.  It is a strength that could effectively  stop  the need for an attack; we are already seeing this in our dealings with certain nations, we will see it more if we can complete our goals.  (Honestly, if John Kerry is going to be the one to best finish that effort, give him the presidency.  But, I want to know that he\'s going to be able to overcome Bush\'s shortcomings.  Open the can or keep it on the shelf, John Kerry.)  The  most  important thing in this type of policy, though, is universal support for it.  Can you get that?  I dunno.  How many countries celebrate Memorial Day?  (links from  urlLink Instapundit )  urlLink Full Post and Comments    '),(2342,2680773,'29,May,2004','2004-05-29 00:00:00','   Sorry about the lack of posts this week; I\'ve been spending some good time with friends and haven\'t had the oppurtunity to sit and type up everything I\'ve been wanting to fill this space with.  Also, tremendous apologies to Chicago Blogs individuals.  Again, I haven\'t had time to figure out how to get a links section working in this site\'s new layout template.  I\'m finishing up the first season of American Ninja Underground and will post the entire final Act in about a week.  I like what ANU has become, but it\'s very different from the direction I had originally envisioned it going.  Really, it\'s become an action version of my short-lived newspaper strip \"In Pictorium\", the problem being that my long-term interest in the ANU world was dependent on the human story I wanted to tell - not something IP was built to do.  As I\'ve said before, I have a plan to drastically change the presentation of the comic and I feel it will create a better environment for the type of thing I want ANU to be as it matures.  Before that, I\'ll be playing with a series tentatively called Alternate Realities, which will be a sort of freak show of classic superheroes living very different lives.  Homeless Wolverine will be a regular.  Also news to everyone except myself...  Realtime Record Reviews will be launching by the end of June.  This\'ll be one of those side-projects I promised back at the point that I brokedown the old RF structure.  I\'m anticipating an expansive update of Dogmi as well.  And \"Tom Waits Sings the Hits\"...  ...you\'ll just have to wait to hear about that, though.  Yes, it will hopefully all come together soon.  urlLink Full Post and Comments    '),(2343,2680773,'24,May,2004','2004-05-24 00:00:00','   Bad news for Cubs fans.   urlLink Kerry Wood will have a bone scan performed today  after shutting down a throwing session yesterday after eight pitches.  This after he shut down a throwing session last week after nine pitches.   The bad news (or worse news) is this will not be a short-term thing.  Going by what I have observed with this injury and other like it, Wood may be done for the season.  At the very least, I would expect him to be out for two months or so.  With the way the Cubs have tried to cover up their players\' injuries this year, the news shouldn\'t come as much of a surprise.  Wood\'s injury will be a long-term thing.  Mark my words.  Without Wood, the Cubs are still dangerous.  When Mark Prior returns, a top end of the rotation of Prior/Clement/Zambrano is still one most teams would love to have.  Maddux as a #4 isn\'t the worst thing in the world either.  But judging by the way Prior and Wood have succumbed to injuries this year after being abused like a rented mule last year by Dusty Baker, it wouldn\'t surprise me to see Zambrano miss some time this year as well.  Are Cubs fans happy Baker left Wood in for 131 pitches the second week of the season to try to get him a complete game victory?  That\'s the game Wood was ejected from for arguing balls and strike in the ninth.  While that alone may not have caused an injury, it\'s still a great example of how Baker treats his pitchers.  It\'s OK though.  The White Sox are in first place and are playing tomorrow night at home.  Half-price night.  Just take the red line in the opposite direction.  You can\'t miss it.  urlLink Full Post and Comments    '),(2344,2680773,'23,May,2004','2004-05-23 00:00:00','   O Wait.  It is.   Edited by underground comic lord Chris Ware,  urlLink McSweeney\'s 13  promises a treasure trove of other alternative artistry from all the majors, Daniel Clowes, Joe Sacco, Richard Sala, Charles Burns, Adrian Tomine, Joe Matt, Art Spiegelman, Crumb, Seth, Kaz...  with late sketches from Charles Schulz and George Herriman.  The list is a remarkable roster of heavy-hitting Raw alums and a nice cross section of talent across the alternative comics realm.    The volume is exquisite, and there\'s really few other words to describe it.  It\'s got all the quirky attention to detail that people love in both Chris Ware and McSweeney\'s.  The hardcover volume, reminiscent of the Acme Novelty Datebook, is wrapped in a funny-pages-of-yore mock-up of Chris Ware \"round guy\" comics (its reverse side is the contributor list).  Hidden within the folds of the paper are a pair of mini-comics, rounding out a full representation of underground comics formats.  So, that\'s why I was ecstatic as I left  urlLink Chicago Comics  with the heavy volume at my side.  And maybe I had gotten my hopes up, I\'ll give you that but...  Blegh.  The volume is certainly nice, and anyone who doesn\'t frequent the work of the artists I\'ve listed above will probably find it rewarding.  It\'s also a great collector\'s piece... your grandkids (or grandkids\' grandkids for those not in my personal demographic) will look at it and yearn for the days when art was on paper.  The most upsetting thing is that the volume is for the most part, excerpts from recent longer works.  Sacco\'s piece is from his tremendous  The Fixer  released last fall, Tomine brings his latest Optic Nerve to the table, Art Spiegelman reiterates the mess that is \"In the Shadow of No Towers\", Ware\'s work will be no stranger to those who read the regular Acme Novelty series in alt-weeklies, and it continues to the point that you\'re startled when you see somethign without the words \"excerpt from\" in front of it.   Bless Jeffrey Brown for his unpublished pile of sketchbooks that are always releasing fresh pieces.  If this wasn\'t comics... if this was, for instance, an anthology of \"short stories\", this thing would be torn apart for what it\'s trying to present itself as.  There is little new here for those that would be most interested in it.  I\'d think McSweeney\'s would have the clout to get some original or rare work from this line-up.  I\'d be wrong.  urlLink Full Post and Comments    '),(2345,2680773,'19,May,2004','2004-05-19 00:00:00','   After a liberal selection of occupational possibilities ranging from \"Writing and Editing\" to \"Women\'s Studies\", non-profit-centric  urlLink Idealist.org  had the following exciting news to tell me:  urlLink Full Post and Comments    '),(2346,2680773,'18,May,2004','2004-05-18 00:00:00','   My Chipotle is back to normal.  Pax Romaine!  Brood X, the largest army of Cicadas, is  urlLink on its way up .  Stupid, stupid bugs.  The East is already dealing with swarms of the idiot insects, and David Letterman points out that Brood X has only shown their beady, red eyes when Republicans have been President.  Book criticism at  urlLink The Atlantic  is awesome; more an article in response than a traditional review.  Kristi(n) sends me  urlLink this piece  on Andrew Horton\'s  Screenwriting for a Global Market .  Good stuff.  Blog archives are functional again. Thank you for your patience.  urlLink Full Post and Comments    '),(2347,2680773,'18,May,2004','2004-05-18 00:00:00','   An interesting piece from  urlLink Reason Online  that tries to decipher the point where hairs are split when comparing our preemptive entry into Kosovo and the same into Iraq.  Thanks to  urlLink InstaPundit  for the link.  urlLink Full Post and Comments    '),(2348,2680773,'14,May,2004','2004-05-14 00:00:00','   Remember the golden days of puzzle games?  Nintendo threw every mixture of colored blocks they could at us and we let it all build up until we got the really long straight piece and then...  everything disappeared.  I mean, there was a point when the characters from Street Fighter had their own  urlLink puzzle game , when you\'d have a couple editions of Bust-A-Move a year (there have been at least 26 iterations since the game debuted in 1993), when people shared their Tetris high scores over a bottle of wine...  Tetris is still a friggin great game, and I still haven\'t gotten the  urlLink space shuttle to launch  in the regular mode...  Snood - which you might consider a Bust-A-Move clone - hit the  urlLink nation\'s universities  for a while, and ultimately turned from shareware into a full console release...  But it was  urlLink ugly as sin .  What would happen if someone made a simple, interesting,  new  puzzle game that utilized the 3d processing of even base-level PCs - or at least took some advantage of the tremendous color depth of contemporary displays?  Probably, the gorgeous  urlLink Zuma .  The rule set is basic, the construction is elegant, and the result is a game that demands creativity and coordination.  Colored balls slowly roll down a winding track towards a consuming hole.  Your goal is to stop them, often by stringing palindrome combos so stunning that they temporarily reverse gravity\'s downward pull.  It\'s not going to be a phenom (not because of any fault of the game), but if your heart is ever warmed thinking about the proper techniques for placing a z-block or the time you thread the needle to topple three-quarters of a Snood board... Reawaken that slumbering puzzle gamer and try out Zuma.  urlLink Full Post and Comments    '),(2349,2680773,'12,May,2004','2004-05-12 00:00:00','   Frequent readers of this blog will recall our previous reporting that trendy burrito-for-white-people place  urlLink Chipotle  changed their salsas for \"more freshness\".  I\'ve been shocked the past few days to discover that Romaine lettuce may be going the way of Ye Ole Green Chili Salsa.  Lettuce has always been a tangential ingredient at Chipotle; at both the stores I frequent, you will be asked if you want every ingredient short of the lettuce.  I\'ll admit I don\'t get asked about the guacamole all that often, but it\'s a lot more often than lettuce.  So, it comes as no surprise that Romaine takes the hit, but I\'m in mourning nevertheless.  The last tin in the line at my current Chipotle is now filled with Iceberg, so-called and so-identified due to its pale color.  Apparently, I will use more nutrients eating Iceberg than I will gain from it.  And I am currently paying almost a dollar more for a Chipotle burrito now than when they came with Romain lettuce  and  Green Greatest-Salsa-of-All-Time Chili Salsa.  *sigh  Edit: Must have been an emergency thing.  We\'re back to Romaine.  Woot.  urlLink Full Post and Comments    '),(2350,2680773,'12,May,2004','2004-05-12 00:00:00','   Been feeling the fallout of  urlLink a happy escape  from work.  urlLink Full Post and Comments    '),(2351,2680773,'11,May,2004','2004-05-11 00:00:00','   In order to provide a more efficient blog that will take advantage of a number of the new components of Blogger, TMSO will be going through a design change.  This  may  effect some things negatively in the short term, but over time these issues will be resolved and you should find the new setup to be much friendlier in your (and our) usage of the blog.  The basic setup will probably remain the same from the time I post this, but you will see some drastic changes in the visual appearance of the blog as I modify the template in the next few days.  Here is your update list:  -Permalinks : You can now link specific posts to send to friends via e-mail or posting on your own blog / website.  -Monthly Archives : As a result of permalinks, I\'ve decided to archive monthly instead of weekly; this will allow the archives to be integrated into the blog\'s main page.  -Integrated Comments : Comments are no longer a plug-in, which means comments will be stored in the archives and should be easier to use.  -Photos!!! : Thanks to the new design, you\'ll be seeing photos and artwork posted in the blog regularly.  -Better XML Support : We are settling what is essentially a stylesheet for posts so that those using news feeder programs will see a cleaner presentation of posts.  urlLink Full Post and Comments    '),(2352,2680773,'11,May,2004','2004-05-11 00:00:00','   NO MO MILES   I first heard about the overestimated efficiency of hybrid cars from Kristi after she attended a lecture on \"green\" products at the Field Museum.  Apparently, the cars use a similar amount of fuel as regular cars across for the first few miles of driving.  But, Wired reports that even the long trip  urlLink fuel efficiency is bogus .  Consumers have found the cars, which were reported to get nearly 50 miles per gallon, seldom top 30 miles per gallon.  30!!!  The Toyota Prius is on the high end at a still disappointing 35, while the Honda Civic Hybrid gets only 26!  For comparison purposes, the  urlLink Corporate Average Fuel Economy  (CAFE) for passenger cars is 27.5 mpg; it is 20.7 mpg for light trucks (SUVs, etc).  The problem - and you can read about this in the Wired article - is that the  urlLink EPA standards  measure waste substances from fuel consumption not the actual use of fuel.  I\'m no physicist nor automotive expert, so I can\'t tell you why that throws off the tests for hybrids... but needless to say, it apparently does.  urlLink Full Post and Comments    '),(2353,2680773,'10,May,2004','2004-05-10 00:00:00','   Saw  urlLink Okkervil River  (aka Coner Oberst With a Soul) at  urlLink Schuba\'s  on Saturday night.  I can\'t say enough about this band, but that\'s not going to keep me from trying.  Seeing a band that\'s emotionally involved in their music at a live show after months of their being on the road is exhilirating, and that\'s what the Okkervils provided to an - to be polite - eclectic post-midnight crowd.  As tired as the image of a rocker breaking crap on stage is, you can\'t deny that you love seeing it, especially when there\'s some honesty to it.  The amazing thing about the outbursts of raw musical emotion for the Okkervils is that their music is heady and introspective.  The centerpiece of the new album is a song about  urlLink empathizing with war criminals  for instance.  Yet there we were, rocking - by the true sense of the word - to the lyric \"While you gasp at my bloody crimes, / please take the time / to make your heart my own.\"  And we could feel it; there was nothing ironic about the moment.  This is a band that in another life would be overly pretentious crap; indeed, opener and spin-off project  urlLink Shearwater  pretty much is.  But somehow, it works right through a balance of tranquil calms and torrid rapids.  urlLink Full Post and Comments    '),(2354,2680773,'07,May,2004','2004-05-07 00:00:00','   Neal Pollack has got the  urlLink \"rough copy\"  of Michael Moore\'s  urlLink lastest message  to his  urlLink adoring fans .  For those not ambitious enough to click that final link, moveon.org is calling for people to call their local Disney Store to voice their concern.  urlLink Full Post and Comments    '),(2355,2680773,'06,May,2004','2004-05-06 00:00:00','   iTool   iAlso realiZe the iPuns are tiRed   It seems obvious that the major record labels are playing iTunes, and not in the good way.  Apple\'s music service has agreed to  urlLink hike its prices  by 26 cents per song and in many cases seven dollars per album.  The move seems intent to chase iTunes users (i.e. legitimate music downloaders) back into record stores.  In other words, don\'t believe all the hype; we\'re far from any kind of distribution revolution.  urlLink Full Post and Comments    '),(2356,2680773,'06,May,2004','2004-05-06 00:00:00','   The Library of Congress is developing a former Cold War bunker (for currency with which to jumpstart the US economy in case of nuclear attack) into a protective  urlLink storehouse for arts artifacts .  I don\'t know what use it is, but I like the fact that the LoC is preserving both the spirit (through digital means) and the flesh (the physical recordings) of our cultural heritage.  urlLink Full Post and Comments    '),(2357,2680773,'05,May,2004','2004-05-05 00:00:00','   Someday, Endzone Markers Will Be Mickey Mouse Figurines   But for right now, the bases and pitching mound in all Major League Baseball games will be printed with  urlLink spider webs to help promote Spiderman 2 .  ...and there is a resounding \"Ugh\" from left field.  urlLink Full Post and Comments    '),(2358,2680773,'03,May,2004','2004-05-03 00:00:00','   Overrated   There\'s a new issue of  urlLink Chunklet  on the prowl.  It\'s the punk rock magazine that hates your favorite band.  In the language of Chappelle\'s Show (you all speak), it\'s Keepin\' It Real Gone Right.  urlLink Full Post and Comments    '),(2359,2680773,'03,May,2004','2004-05-03 00:00:00','   Your Life Is About To Be Torn Asunder    urlLink And it will never be the same.  urlLink Full Post and Comments    '),(2360,2680773,'03,May,2004','2004-05-03 00:00:00','   Right Now   The ACLU is blowing all your money by placing full-page ads in high volume magazines in order to convince more people to send them money to spend on advertising.  The road may never get anywhere, but at least it\'s paved with gold.  urlLink Full Post and Comments    '),(2361,2680773,'03,May,2004','2004-05-03 00:00:00','   Weekly Challenge : Past Masters and the First Third   The Musical Chairs challenge is still open although it will be closed by the next time I have a day off of work (just two more out of the nine I\'ve had to work left to go).  But here is your new challenge; something much easier and also music related.  Everyone can participate with little craft-work, so everyone is encouraged to.  The new challenge is called \"Past Masters and the First Third.\"   Part One   Every year we do this thing called the Lost Sock Collections which is an exchanging of mix-cd\'s featuring each participant\'s top ten albums of the year.  What I find is that a few months later, due in part to the LSC and in part to ambling through record stores, my list would have looked quite different.  However, the calendar-year-centered environment we create with \"Top Ten of **\" lists prevents late-comers from being garnished with attention.  The artists still make money off old albums, and the listeners can still enjoy the albums though, so maybe we should give a little more praise to what we missed.  Thus, the first part of this challenge is to name a double-plus good 2003 album(s) that you first heard in 2004.   Part Two   Albums from the early months of any calendar year generally lose ground in annual top-lists due to the added competition, the freshness of that competition and the flaws of human memory.  Arguably, a truly great album should be remembered over the course of a year, but is it fair that late-comers don\'t have to stand up to the same test?  In order to give the underdogs their own place, TMSO issues this simple challenge.  Name your favorite 2004 release(s) thus far.  Fulfill your challenge as a comment below or as a separate  urlLink e-mail  and all entries will be put together for a future post.  urlLink Full Post and Comments    '),(2362,2680773,'03,May,2004','2004-05-03 00:00:00','   Wired Notes   Came across a few great quotes from the May issue of Wired.   Cult Life for the Atari 2600   (page 74)   urlLink Homestar Runner  is getting console game treatment with an RPG-style game for the barbaric Atari 2600.  Do you remember when we banged rocks together in order to make Pitfall Harry leap upon crocodile heads?  Other people do, and Atari gaming has been making a kitschy comeback because of it.  You can purchase joysticks whose bases are bundled with a small museum of these ancestral video games.  There is also a blossoming homebrew scene using actual cartridges.  And now this.  A \"commercial\" release.  Great quote from the Wired write-up on cameos by classic Atari characters, \"If it\'s 8 pixels wide and a single color, it\'ll be in there.\"   Fear and Loathing of Information  (page 193)  John Poindexter is what Wired refers to as \"the Pentagon\'s Big Brother in chief\", a guy trying to organize intelligence so that it can be better used to capture would-be-terrorists.  In Wired\'s interview, Poindexter talks theory about what he is trying to do in broadening intelligence powers.  To the statement \"Your critics never relented on privacy questions,\" Poindexter responds...  \"Advocacy groups want to stay in business, so it\'s in their interest to paint a dire picture.\"  A point I\'ve been becoming increasingly aware of recently...  If the problems are solved, the solvers are out of a job.  I\'ll probably post a lot more on this in coming days/months.   The Facts Get Clearer the Further We Get From Them  (page 135)  In an article on the MPAA\'s in-school anti-piracy lesson capsules, the reporter says, \"The real point, of course, is to protect Hollywood from the fate of the record industry.\"  What\'s telling about the quote is how the RIAA\'s claim (that MP3 swapping has devoured their business) has become a given, despite no clear evidence of its truth (just because the animal has four legs doesn\'t mean that it\'s had its business destroyed by peer-to-peer).  Is piracy a problem?  Sure.  Is piracy crippling the RIAA?  No.  An idiotic business model is.  The MPAA continues to strengthen itself  despite  a dramatic increase in movie piracy as faster connections and more efficient file-sharing progs are released...  \"Hollywood has seen revenue rise 27 percent in the same four-year period that the recording industry went into free fall.\"  Look at what Hollywood has been offering customers...  big, fat collector\'s editions of even the worst movies on high quality discs.  They\'re reaping what they sow.  urlLink Full Post and Comments    '),(2363,2680773,'29,June,2004','2004-06-29 00:00:00','   On Sunday,  urlLink the White Sox traded  starting catcher Miguel Olivo, top prospect outfielder Jeremy Reed, and AA shortstop Michael Morse for pitcher Freddy Garcia and catcher Ben Davis.  In short.  White Sox general manager Kenny Williams was taken for a ride.  Yet again.  Before I say anything else, I must say that Freddy Garcia is a very good pitcher.  He will help the Sox this year and is better than what the team was throwing out there.  But...   *He\'s a free agent after this season.  The Sox have rented him for three months in exchange for a great haul of prospects for Seattle.  There\'s no guarantee the team will be able to re-sign him, and with this organization\'s history with pitchers, I don\'t feel confident he\'ll be back.   *Garcia is moving from a severe pitcher\'s park to a severe hitter\'s park.  Plus he\'s a flyball pitcher coming to Coors East in Comiskey.  His numbers will almost undoubtedly suffer.   *He\'s had ERAs of 4.39 and 4.50 the past two seasons.  Yes, in 100 IP this year, he\'s been better.  But I\'m not going to discount the past two seasons just because he\'s off to a nice start.   Now the White Sox have a giant, gaping hole at catcher.  I believe at this point the Sox have the worst catching situation in the majors.  Alomar is old, not productive, and probably just should have been hired as the bench coach instead of Baines.  Let\'s not let 30 at bats this year cloud the fact Jamie Burke is a 31-year-old career minor leaguer.  And Ben Davis has been a huge disappointment.  A .670 career OPS.  Striking out 1 out of ever 3.9 at bats in his career. OBPs routinely less than .300.  SLG less than .400.  Add to the catching hole the fact that Crede has been struggling all season long and may or may not turn it on.  Willie Harris was just in a huge 0-fer slump.  Juan Uribe is 4 for his last 31.  Aaron Rowand, great against LHP, still can\'t hit right-handers.  And now the team\'s one productive catcher is gone.   Compare what the Royals got for Carlos Beltran, a superior player, to what the Sox had to give up for Garcia.  KC got 3 decent prospects, the best being Mark Teahen.  The Sox gave up a starting ML catcher, one of the best prospects in all of baseball in Jeremy Reed, and another Top Ten prospect in the organization in Mike Morse.  Seem equal?  Of course not.  Not to mention that just two years ago all the booty KW could get for All-Star Ray Durham was Jon freaking Adkins.   I know the Sox have organizational depth in the outfield, but Reed was the one \"untouchable\" guy in the organization.  He\'s a left-handed bat with great patience, decent power, good speed, and the ability to play center field.  Reed is the best hitting prospect this organization has had since Frank Thomas.  He walks, something most players in the Sox oganization are afraid to do.  And if the rumors are true about Bavasi wanting Crede, but the Sox forcing him to take Reed -- well, that\'s ridiculous.  Would you rather see a hole at catcher and Crede at 3B or Olivo catching, Uribe at third, and a Rowand/Perez platoon in CF?  Especially since the Sox just drafted Josh Fields in the first round who, god willing, will supplant Crede at third in two years anyway.   \"Going for the jugular\" and \"being aggressive\" are great things.  But it didn\'t work in 2001 with David Wells.  It didn\'t work in 2002 with Todd Ritchie.  It didn\'t work in 2003 with Bartolo Colon and Billy Koch.  It didn\'t work in 2003 with Carl Everett and Roberto Alomar.  And there is NO guarantee it will work this year either.  The difference being never has so much of this team\'s future been put on the line for potential present success.  Olivo still had 5 years under Sox control.  Reed hadn\'t played a game in the majors yet and is just about as sure of a thing as there is in the minors.   If the White Sox make the World Series this year, I\'ll take it all back.  I\'ll never question KW again.  But, I still don\'t think this team is strong enough to do that.  I hope I\'m wrong.  Nothing will make me happier.   My bottom line?  Garcia will help this year, but the only way the trade will make sense is if the Sox make the World Series.  If not, we have just mortgaged a HUGE part of the future for the right to finish with 84 wins instead of 82.  urlLink Full Post and Comments    '),(2364,2680773,'26,June,2004','2004-06-26 00:00:00','   urlLink Pitchfork Media  has compiled the Top 100 Albums of the 1970\'s, which apparently consisted of listing every album Brian Eno had even a partial fingerprint on.  There were mistakes, ovbiously, but better to discuss are the complete misses.  What follows are the Top Nine albums to be ignored by Pitchfork on their list.  I\'d place each of these in my personal Top 20, if not maybe 30 of that decade.  If any of these bands had just hired Brian Eno to turn a knob in the studio...  1.   urlLink  Elvis Costello -- Armed Forces   (1979, Columbia)   If there\'s room for 16 Brian Eno albums and 25 David Bowie discs, there\'s gotta be room for the third release of the decase from the best songwriter of his generation.  \"This Year\'s Model\" popped in the list at #52 with \"My Aim Is True\" at #37.    2.   urlLink  Rolling Stones -- Sticky Fingers    (1971, Virgin)   Are you kidding?  One Stones album on the entire list?  While \"Exile\" has the reputation, it\'s \"Sticky Fingers\" that is the Stones at their absolute apex.  \"Let It Bleed\" still held onto some of the styles of the \'60s, but \"Sticky Fingers\" had the band roaring into the new decade with the underrated Mick Taylor helping to shape the band\'s sound.   3.   urlLink  Joe Jackson -- Look Sharp!    (1979, A&M)   The list ignores a lot (or all) of the late-70\'s Angry-Young-Man/New Wave/pub rock, which is surprising since this stuff isn\'t very far away from The Clash (ranked #2) or bands like the Ramones.  Joe Jackson\'s debut is a masterpiece, bouncing through different genres like two drunks on the bumper cars at the circus.  Of all the albums on my list, I\'m probably most surprised this didn\'t crack Pitchfork\'s 100 or EVEN GET ONE SINGLE FREAKING VOTE.  \"I\'m The Man\" appeared on one list.  (sarcastic clapping begins)   4.   urlLink  Gram Parsons -- GP     (1972, Reprise)   The list ignores most (or all) of country music in the 70s, even groundbreaking stuff like Gram Parsons\' two albums.  Apparently, listening to less popular genres of music is only OK if it makes you look cool.  Dance, disco, funk, jazz and Kraftwerk are all over Pitchfork\'s list.  But the man who basically invented country-rock and paved the way for the 90s alt.country movement gets no credit.   5.  urlLink  Graham Parker -- Squeezing Out Sparks    (1979, Arista)   See comments for #3.  Parker was more focused with his effort, however.  This is the sound of a man with complete confidence in his craft.     6.   urlLink  Flamin\' Groovies -- Flamingo    (1970, Kama Sutra)   One of the latest, greatest forgotten bands.  The Groovies smoked on three early 70\'s releases.  Keith Richards once said the Groovies album \"Tennage Head\" is what he wished \"Sticky Fingers\" turned out to be.  I prefer one release before that album, \"Flamingo.\"  Check out the guitar outro solo on \"Heading For the Texas Border.\"    7.   urlLink  AC/DC -- Highway To Hell    (1979, Atco)   \"Highway To Hell\" was AC/DC\'s finest moment.  Bon Scott\'s last album with the group.  If Van Halen\'s debut album cracked the list, this also needed to have a spot.    8   urlLink  The Kinks -- Lola vs. the Powerman & the Money-Go-Round, Pt. 1    (1970, Reprise)   The Kinks had finished off back-to-back slices of British life with \"Village Green Preservation Society\" and \"Arthur.\"  \"Lola...\" is a step below the high-water marks of those two albums, but still a tremendous effort from one of the most underrated bands of all-time.   9.  urlLink  New York Dolls -- New York Dolls    (1973, Mercury)   Put simply, this is better than anything the Stooges ever released.  And both \"Funhouse\" and \"Raw Power\" made the list.  Make some room for the band with the guy who would later become Buster Poindexter and record \"Hot, Hot, Hot!\"   I\'ll take comments.  I\'ll even respond.  urlLink Full Post and Comments    '),(2365,2680773,'24,June,2004','2004-06-24 00:00:00','   The female lead of  Dodgeball  and  Zoolander ,  urlLink Christine Taylor , who I feel has been in 85% of similar comedies (she hasn\'t, but does embody the archetype), was Melody on Nickelodeon\'s  urlLink Hey Dude .  She\'s also Ben Stiller\'s wife.  Sadly, she\'s the only  Hey Dude  alum to successfully move beyond tv (or beyond  Hey Dude  at all...).  Ted was in  White Squall  and Mr. Ernst was an ER regular.  And that\'s it.  On a related note for all you non-Trekkers (and all you Trekkers as of yet unaware), the ex-wife at the heart of Jack Ryan\'s (Republican running for Senate)  nail-in-the-coffin scandal  is none other than Jeri Ryan, Seven-of-Nine from Star Trek : Voyager.  urlLink Full Post and Comments    '),(2366,2680773,'23,June,2004','2004-06-23 00:00:00','   I really enjoyed  urlLink Coffee and Cigarettes , the recently released film of black-and-white vignettes featuring mismatches of musicians and actors.  There\'s enough change in style between segments that each sketch is fresh in its cute-and-awkwardness.  But one thing I found tiring was director/writer Jim Jarmusch\'s absolutist treatment of silence in dialogue as a rift in communication.  Whenever characters are quiet they appear to be at odds with their counterparts, realizing their train is somewhere in Albuquerque they are plainly contemplating catching a ride back to Buffalo.  Iggy Pop contemplates leaving Tom Waits, whom he has inadvertently offended.  Cate Blanchett restrategizes her conversation with a black sheep cousin (Cate Blanchett).  Pauses are never comfortable.  In real dialogue, though, the pause can imply an exact understanding: a way for conversers to think the same thought without speaking over each other, to conversely observe an event or item, to actually communicate via body language or simply through the lack of speech.  We don\'t find such silence in  Coffee and Cigarettes ; a pause is a certain sign of a derailed conversation.  True, the movie is largely about uncomfortable combinations (why else pair Bill Murray with RZA and GZA).  But there are mutual moments in the positive spaces, so why not the negative space?  In a movie so intimately focussed on  unpaced , diner conversation, heavily one-sided use of momentary pauses is a definite weakness. Moments of silence are tense because of it.  A tempo has been set; breaks feel unnatural and can leave the audience feeling as awkward as the characters on the screen.  Of course, maybe that\'s the point.  urlLink Full Post and Comments    '),(2367,2680773,'23,June,2004','2004-06-23 00:00:00','   How come my not-internet-enabled discman will display track titles on a cd (original, copy, or mix cd) and Winamp or Media Player can\'t do it without contacting CDDB?  urlLink Full Post and Comments    '),(2368,2680773,'23,June,2004','2004-06-23 00:00:00','   A good explanation about why a lot of us get so worked up about You-Know-Who\'s recent diatribes (from Christopher Hitchens on  urlLink Slate ):  \"So I know, thanks, before you tell me, that a documentary must have a \"POV\" or point of view and that it must also impose a narrative line. But if you leave out absolutely everything that might give your \"narrative\" a problem and throw in any old rubbish that might support it, and you don\'t even care that one bit of that rubbish flatly contradicts the next bit, and you give no chance to those who might differ, then  you have betrayed your craft . If you flatter and fawn upon your potential audience, I might add, you are patronizing them and insulting them. By the same token, if I write an article and I quote somebody and for space reasons put in an ellipsis like this (â€¦), I swear on my children that I am not leaving out anything that, if quoted in full, would alter the original meaning or its significance. Those who violate this pact with readers or viewers are to be despised.\"  urlLink Full Post and Comments    '),(2369,2680773,'18,June,2004','2004-06-18 00:00:00','   urlLink'),(2370,2680773,'15,June,2004','2004-06-15 00:00:00','   urlLink'),(2371,2680773,'15,June,2004','2004-06-15 00:00:00','   urlLink'),(2372,2680773,'12,June,2004','2004-06-12 00:00:00','   urlLink'),(2373,2680773,'12,June,2004','2004-06-12 00:00:00','   I\'m constantly on a soapbox about how getting free MP3s online from bands is pretty much the only way to be exposed to new music.  So why don\'t more bands do it?  Here\'s a great piece on  urlLink indie artists and their websites   from John Scalzi\'s blog.  This is the bottom line: I want to hear new music.  I can\'t hear it on the radio.  I live in Rockford.  No one plays here.  The only way I can hear it is on the web.  It\'s the only way I can hear about bands like the  urlLink Kingdom Flying Club  and  urlLink Runner and the Thermodynamics .  I\'ve bought CDs from both bands in the last few months based solely upon MP3s I\'ve downloaded online.  This is the way things should work.  Next time I\'m ambitious (hopefully tomorrow), another post on the glut of great music that will be released in the next few weeks/months.  Someone\'s got to pick up the slack while Mike\'s on vacation.  urlLink Full Post and Comments    '),(2374,2680773,'12,June,2004','2004-06-12 00:00:00','   I\'m on vacation right now, but in an exciting development will be posting audio posts during my trip.  Stay tuned!  urlLink Full Post and Comments    '),(2375,2680773,'02,June,2004','2004-06-02 00:00:00','   Since we\'re on the subject of digital cameras, here\'s an interesting tidbit from CNN on a World War II aerial reconaissance camera converted to a  urlLink digital behemoth.  urlLink Full Post and Comments    '),(2376,2680773,'01,June,2004','2004-06-01 00:00:00','   Small-time publisher  urlLink Ben Brown  is taking duties over for the week at the  urlLink Bookslut blog  and has been doing a nice job.  He\'s not a link sniffer like full-timer Jessa Crispin, but he\'s a bit better at expounding on the links he does have.  It\'s a One Week Only kind of thing and should be interesting to anyone who likes books but doesn\'t screw every one he/she sees.  urlLink Full Post and Comments    '),(2377,2680773,'01,June,2004','2004-06-01 00:00:00','   I\'m inclined to tell you that the Nikon Coolpix 5200 (and 4200) is the real deal.  Pulled one out of the box today and was floored.  The camera is hardly larger than its younger sisters (2200, 3200), but is almost half as thin with a metal face plate.  Lithium rechargeable battey, sexy menu system, and what may be the fastest shutter/autofocus response time I\'ve found on a digicam so far - which, as I tell most customers, has been the number one complaint from digital camera buyers.  People should go nuts over this thing...  As long as the image quality is there, the camera is going to be hot hot hot.  I\'ll try and take some pics and post them here.  In the very least, I\'ll describe the results to you.  A disappointing three day weekend ends today.  urlLink Full Post and Comments    '),(2378,2680773,'01,June,2004','2004-06-01 00:00:00','   I was beginning to think that a mythic Saturday afternoon where my friend John and I discovered (upon my dad\'s recommendation) the Star Wars knock-off  Message from Space  was just a dream.  As I would attest to others, unable to provide evidence, the story is centered around a galactic struggle over magic walnuts.  Not mystic globs resembling walnuts...  Magic walnuts, literal, period.   urlLink Vic Morrow  is a professional drunk.  Wood sailing ships float around space at full mast.  Things happen for no explicable reason.  Film Threat proves that it exists and others have seen it (click title for link), and  urlLink this site  has a full synopsis and sound clips.  urlLink Full Post and Comments    '),(2379,2680773,'01,June,2004','2004-06-01 00:00:00','   Sometimes I think it would be fun to buy a barrel of oil just so that you could show it off at parties or walk around telling people, \"I\'m not worried about no three dollar gas prices as long as I gots me this barrel of oil.\"   urlLink Only $42 ... although I\'m not sure if the barrel is included.... or shipping.  urlLink Full Post and Comments    '),(2380,2680773,'30,July,2004','2004-07-30 00:00:00','   Here is some free punditry for you based on my scattershot viewings of the Democratic National Convention, the highlight of which was figuring out the different time delays between networks and then switching between them to hear the same word up to four times.    Bill Clinton: &nbsp; Clinton proved why he\'s remembered for his charisma and speaking skills; far and away the strongest speaker at the convention.&nbsp; Well-written jokes that were equally well delivered.&nbsp; He\'s smooth and at ease and almost conversational.   Al Gore: &nbsp; Gore spoke to his audience like they were stupid, stupid children.&nbsp; Make no mistake, he should have won in 2000...&nbsp; it was handed to him and he blew it.&nbsp; This speech showed why.   Obama: &nbsp; The man with no one to run against is the new star of the Democratic party.&nbsp; I imagine Democrats are thinking what sportswriters have been thinking about Michael Vick since watching him do that-crazy-thing-he-does...&nbsp; Please don\'t let anything happen to him.   John Edwards: &nbsp; For all the praise he gets about being a great orator...&nbsp; Bleh.&nbsp; Not a bad speech, but the kind of speech&nbsp;that would be considered&nbsp;quality&nbsp;from a Cheney or a Gephardt or a not-so-good-speaker.&nbsp; Edwards reputation sets him on another plane, where speaking actually turns votes.&nbsp; And it\'s on that plane that he fails.&nbsp; BONUS NOTE:&nbsp; Very socialist undertones.&nbsp; SUPER EXTRA BONUS NOTE:&nbsp; Edwards did demonstrate that if you don\'t mention the opposition, you can say whatever the hell you want.&nbsp; This could actually be an effective rhetorical strategy in the long haul to November.   John Kerry: &nbsp; Didn\'t see it.&nbsp; I\'m sorry.&nbsp; But I wanted to mention that Kerry\'s delivery oftentimes reminds me of the Simpsons Treehouse of Horror with Bill Clinton and Bob Dole being replaced with aliens.&nbsp; From Kerry stumping today (and not an unusual thing):&nbsp; \"We need to reach, reach, reach for the future.\"&nbsp; From Kang (or was it Kodos?): \"Always twirling, twirling, twirling towards the future.\"  urlLink Full Post and Comments    '),(2381,2680773,'26,July,2004','2004-07-26 00:00:00','   \"They have a total ability to block out anything they don\'t want to get through... that\'s what makes this animal so scary.\"  ...in an advertising sense.  Wired has  urlLink this story  on what advertisers are doing to recapture a generation that doesn\'t listen to their messages.  There\'s a suggestion here that we are on the brink of a marketing revolution.  I don\'t know if it\'s very clear  what\'s  going to happen but something is, and advertisers and consumers are aware of it.  urlLink Full Post and Comments    '),(2382,2680773,'26,July,2004','2004-07-26 00:00:00','   While internet forums and Amazon.com reader reviews can be a good jumping point for your own interviewing, they do not replace it.  The next time I read the words, \"Fishiekuj23 says, this book sucks,\" or any other variation thereof is the time I cut off your internet access permanently.  urlLink Full Post and Comments    '),(2383,2680773,'26,July,2004','2004-07-26 00:00:00','   It\'s no secret that once you\'ve gotten on  urlLink Pitchfork \'s bad side that you ain\'t comin back.&nbsp; The only thing the magazine loves more than holding a grudge is pretending to struggle not to.   Certain bands are given a protection clause, which gives them a few albums to muck up before being thrown to the fire.&nbsp; Perennial favorites like Radiohead or The White Stripes are included here.&nbsp; But most receive the termination of contract notice in the mail and will never regain their former status.&nbsp; The Eels, Ryan Adams and Pedro the Lion have all produced slump-ceasing records which have&nbsp;regained for them some respect elsewhere while they continue to dig through dumpsters at Pitchfork.&nbsp;&nbsp;It\'s a bit like an individual who has gone through a difficult and public divorce...&nbsp; They may rebuild their lives into something truly joyful, but they\'ll never fully discard their painful past, and those who saw the painful moments will never be able to look at them the same.&nbsp; Damien Gough, the Badly Drawn Boy, is&nbsp;a member of the Pitchfork Pit of Eternal Despair&nbsp;and proves so with his latest album   urlLink One Plus One is One  .   But what\'s interesting in the  urlLink Pitchfork review of the album &nbsp;is the introduction\'s focus on Gough\'s weird tie to another on Pitchfork\'s blacklist...&nbsp; Nick Hornby.   Hornby is part of this weird&nbsp;group of fringe mainstream writers who, upon receiving commercial success, began attempts to&nbsp;yank increasingly popular underground artists into the mainstream light.&nbsp; These artists became Lottery millionaires who really could have earned their own way to wealth&nbsp;but now&nbsp;stand without&nbsp;legitimacy to their riches.&nbsp; It\'s the classic story of the one-hit wonders and the sell-outs.&nbsp; It\'s just that instead of a faceless money-hungry corporation, the source is a creator who wants to share their achievements with other artists they respect.&nbsp; What makes it even weirder is that the benefactors enter and set up residence in the world of those they benefact.&nbsp; Nick Hornby\'s venture in music criticism and Michael Chabon\'s work with comics are probably the two most prevalent, current&nbsp;examples.&nbsp; In both these cases, neither has earned their right to the new land they\'ve staked out, they\'ve just gone in and taken it (as a provision of the Fair Housing Act?).   The result of it all is&nbsp;hollow wish fulfillment that depletes both parties.&nbsp; I\'m not saying that it\'s a bad thing to donate one\'s fame or wealth to those below you.&nbsp; The detrimental thing is to graft one\'s self to a championed cause as opposed to feeding it so that it can grow.  urlLink Full Post and Comments    '),(2384,2680773,'23,July,2004','2004-07-23 00:00:00','   Gorgeous, cool day here in Chicago (perfect football weather a few months too early) following a brutally hot and ugly yesterday.&nbsp; Took some pictures which I\'ll share as soon as the new comp is assembled.  Guts for my desktop computer are on order, so hopefully I\'ll be back to actually working on the site in a week or so.  Kristi and I went to go see Tim Howard of Soltero at a coffee shop up by Wrigley Field.&nbsp; The act after him was his complete antithesis.&nbsp; In a world&nbsp;where the events of movies like  Drumline ,  Bring It On  and&nbsp; You Got Served  are real, Tim and Jeff&nbsp;(whatever his last name was) would&nbsp;have to have had (!!!) a guitar fight.&nbsp; (Tim would have to be given bad choreography just to make the fight look fair.&nbsp;)&nbsp; Howard is a scrawny, pale, straight-haired, smartass cynicist from Boston.&nbsp; His nemesis was a fit, tan, curly-locked, sincere romantic from California.&nbsp; Strum skirmish!&nbsp; Chord combat!&nbsp; A borrowed guitar pick&nbsp;unfortunately kept things civil.  Eightball 23 is a nice comic from Daniel Clowes, but honestly, the best thing I picked up from the comics shop last week was Infinite Kung Fu.&nbsp;&nbsp;omg.&nbsp; If&nbsp;Kill Bill gave you any joy, run to your local comics shop and buy some&nbsp;IKF immediately.&nbsp; Review is upcoming.&nbsp;  urlLink Full Post and Comments    '),(2385,2680773,'19,July,2004','2004-07-19 00:00:00','   I saw  Napoleon Dynamite  for the second time last night.  It makes me laugh.  Stylistically, it\'s an MTV Films clone of Wes Anderson.  As a hybrid of the two, it works.  Only occasionally is  Napoleon Dynamite  detrimentally reminiscent, and even then the moments are too hilarious to pay too much attention to.  Honestly, there are plenty of reasons not to like the movie.  If you don\'t think the whole Napoleon Dynamite character is funny, you\'re probably going to want to walk out.  But check  urlLink the trailer  and if you enjoy it, I swear you\'ll love the film.  urlLink Full Post and Comments    '),(2386,2680773,'18,July,2004','2004-07-18 00:00:00','   The defining point of Fahrenheit 9-11 occurs when Michael Moore meets the film\'s poster-mother, Lila Lipscomb - the Michigan mom of a fallen soldier - in Washington D.C. On her way to see the White House, Lipscomb steps aside at a poster booth set up on the sidewalk... While the camera never lingers on the booth for long (why doesn\'t Moore want us to see it? -to use one of his own techniques), it appears to be the kind of conspiracy hut, common in DC and not unheard of elsewhere, where the names are changed to fit the times and particular argument but the pictures remain the same. As Lipscomb\'s attention pulls deeper into the booth (and, really, her own grievance) and mutters about President Bush killing her son, another woman steps in front of the camera flapping her arms and sqwuaking, \'This is all a set up... it\'s all staged.\' Lila Lipscomb confronts this second woman to express the reality she is living, to which the woman allegedly responds, \"You should blame Al Qaeda\" - obviously, not even a factually correct statement.   And so, our middle-road Mid-America heroine, emotionally torn and confused, is stuck between two extremes: the first a paranoid distortion, the second a cold ignorance.   This is how it\'s been in the national debate. On the news networks, pundits are paired off like a dysfunctional Noah\'s ark to give \"perspective\" on cookie-cutter issues where the audience already knows what they\'re going to say. Pick your side because it\'s not the other side, they announce. The middleground is still the territory being fought over and is still where most politicians dwell, but the parties have learned it\'s a dangerous territory to wage the battle within.  urlLink So they don\'t . When you\'re standing right next to the dividing line, it\'s easy to lose people to the other side. But if you wage war from the extreme corners of the map, you can easily pull folks in to your purported safety; even better, they have a long run if they want to switch sides. So, the extremes have hijacked the debate and are tugging at the wheel for control of the country\'s direction.   Michael Moore yanks hard in  Fahrenheit 9-11 .   I have a few major issues with the phenomenon of the film, although I will not pretend to abhor every ounce of it.    First : Moore the Magician    If you can suspend your disbelief, maybe the first act of the movie will captivate you. When you see this movie (again), I encourage you to watch for the sleight of hand, though. You\'ll notice it.   Moore\'s narration (through humor and rhetorical question) is intended to guide your thought pattern while he takes the reasoned explanation and hides it in his hat. Same with the editing. Little is total fabrication; just as magic tricks are not performed thanks to mystical power, Moore\'s film relies (mostly) on entirely true bits of information. But pay attention to how Moore  needs  you to believe certain assertions laid carefully on top of those facts, though, or how he bends the timetable of events (dates occasionally appear on the screen, making it easier for a viewer to keep track). The whole trick will collapse in front of your eyes. Check out Dave Kopel\'s (a 2000 Nader endorser)  urlLink extensive write-up on the film  if you would prefer to have some of the method blatantly explained.   At other times, Moore constructs more elaborate devices; take, for instance, the recruiting of congressmen\'s children. No congressman has the actual right to sign their child up for the military (17-year-olds require parental  consent ). So, the exercise of having congressmen sign up their sons and daughters is guaranteed to get the results Moore is seeking. Considering this \"failure\" is bona fide, how pitiful are the actual results? Moore gets people to ignore him on their way to other things or to be - at least, pretend to be - thoughtful with his laughably outrageous documentation.   If Moore\'s case is really convincing, why does he feel the need to resort to parlor tricks and crafted illusions to express it?   Because that is what he is... A birthday party magician. Having discovered powerful footage outside of the public eye (particularly, the \"7 Minutes\" and Gore\'s silencing of the 2000 Election opponents), he  must  turn these pieces into props for his act, unable to let them stand for themselves. Somehow, the footage overcomes Moore\'s murky plot; in nearly every major review of the film, these are the highlighted moments.   The bottom-line: Moore might succeed as a comedian, but he\'s made a piss poor documentary out of blatant misdirections.    Second: Critics, Accolades and the Media    I find it shocking that given the extensive reporting of Moore\'s fibs, exaggerations and stagings in  Bowling for Columbine  that an overwhelming majority of critics so quickly garnished  Fahrenheit  with unrelenting praise. Alright, okay, so they might have relented long enough to say that the film was pretty one-sided. In fact, stories which critically analyzed Moore\'s construction and content have been labelled as \"critiques\" distinctly setting them apart from the salivary traditional reviews of the Justice League of Film Critics (JLFC). The craft is dead.   The following is addressed to the JLFC, although the rest of you may read:       Critics, you may think you can hedge your bets by saying something along the lines of, \"It is worth seeing, debating and thinking about, regardless of your political allegiances.\" You are wrong.   To say that a hastily fabricated conspiracy plot is worth listening to simply because you are unsure of its legitimacy is to betray your readership and to cede your piece\'s entire point for existing.   Your purpose is to head poor arguments - whether in their construction or content - off at the pass. Don\'t tell me to do your job for you.   Next time, send your paycheck to someone who knows what they\'re talking about. Let them give us some actual criticism, because a claim that is blatantly untrue is not worth listening to, it is not worth thinking about, it is not worth debating. It is a waste of time.   P.S. When it comes to fueling debate on foreign policy, tell people to fucking pick up the newspaper you write for (or.. Hell. Turn to another section).     Needless to say, I also disagree strongly with this year\'s judges&nbsp;at Cannes.&nbsp; It seems to me, they were swept away in the emotional intensity surrounding  Fahrenheit \'s premiere and were unable to objectively step aside and - like a good chunk of critics - do their job correctly.    Third: The True American Patriot &nbsp;   Something I\'ve heard from&nbsp;a lot of people&nbsp;who liked&nbsp; Fahrenheit 9-11  is, \"I agree with his goal, so I mostly enjoyed it.\" It\'s my opinion that we shouldn\'t endorse extremists simply because we see them being on our side. If we, conservative or liberal, simply smile quietly while the fringes spout ridiculous claims, we effectively disenfranchise ourselves.  One of Moore\'s recurring&nbsp;commentaries on the so-called \"War on&nbsp;Terror\" &nbsp; is that the United States funded/supported Osama bin Laden in order to oust Russia from Afghanistan.&nbsp; Whatever the validity of the claim, it is the criticism itself that I want to focus on here; that is, the U.S. tolerated extremists when such a course was favorable but now witnesses&nbsp;its brutal repercussions.&nbsp; See what I\'m getting at?  Since 9-11, there has been a screwball cultural war over who is more patriotic: those who unfailingly support the government\'s actions or those who unfailingly question them.&nbsp; Of course, the middleground has been largely eliminated as we who are skeptically-faithful of any mixture have polarized ourselves - magnetically attracted to our sense of \"team\".&nbsp;  We  are largely to blame for disinformation wherever it comes from, because we are too tied to our partisan \"patriotism\" to challenge that which aids our greater efforts.  I recently came across this piece from George Orwell, which I think speaks strongly to this issue and with which I will conclude this response.&nbsp; Incidentally, Moore misfocuses  Fahrenheit \'s quoting of Orwell and, in effect, misinterprets it.&nbsp; I hope to fair better with this on the inherency of dangerous nationalism:     The most one can say is that people can be fairly good prophets when their wishes are realizable. But a truly objective approach is almost impossible, because in one form or another almost everyone is a nationalist  ... The most intelligent people seem capable of holding schizophrenic beliefs, or disregarding plain facts, of evading serious questions with debating-society repartees, or swallowing baseless rumours and of looking on indifferently while history is falsified. All these mental vices spring ultimately from the nationalistic habit of mind, which is itself, I suppose, the product of fear and of the ghastly emptiness of machine civilization....     I believe that it is possible to be more objective than most of us are, but that it involves a moral effort. One cannot get away from one\'s own subjective feelings, but at least one can know what they are and make allowance for them      urlLink from the  Partisan Review , Winter 1945  urlLink Full Post and Comments    '),(2387,2680773,'16,July,2004','2004-07-16 00:00:00','   From  urlLink The Atlantic :   In early February I sat in a Starbucks in downtown Washington with Dan Feldman, who is helping to organize Senator John Kerry\'s foreign-policy team.   If you are going to mention where you met for your interview, you better damn well be sure it\'s an interesting location or, in the very least, has some relevance to your story.  The local Starbucks doesn\'t cut it.  Sorry.  This and the \"light streamed through the window\" are possibly the goofiest magazine writing cliches around.  urlLink Full Post and Comments    '),(2388,2680773,'14,July,2004','2004-07-14 00:00:00','   It\'s no secret that we\'ve been bouncing around a Rock Criticism Dictionary for some time.  Allegedly, Andy has a pretty sizeable portion of it written and in a secret vault somewhere.  But what happens when math dorks try to dissect the puzzling language of pop music critics?   urlLink Pitchformula  is what.  A database analysis of the reviews on Pitchfork Media, Pitchformula is intended to discover the perfect song, at least when it\'s pitched to the Pitchforks.  Little is particularly startling, but it\'s funny to see a mathematical proof that the indie rock elite think \"yelling\" in a song is a bad move.  urlLink Full Post and Comments    '),(2389,2680773,'14,July,2004','2004-07-14 00:00:00','   Gregg Easterbrook, a brillant man first, a brillant writer second, and a man with liberal ideas third,  urlLink tears into Michael Moore  at the New Republic online.  The New Republic is a fairly left-wing publication, so I\'m even happier to see the man torn down at a place where he expects to be revered.  The money lines?     \"Canadian Bacon has a 20-second scene that presages the Michael Moore to come...at one point, a [U.S] commando falls and injures himself. The nearest other commando turns and, without hesitation, shoots his comrade repeatedly through the head. It\'s an ugly, sadistic scene totally out of place in what is otherwise a slapstick farce. Though it gets the message across: Michael Moore deeply, deeply hates the United States.\"   Also, the fine people at  urlLink Cox and Forkum  have a wonderul political cartoon today. If you\'re checking the link at a later date, this would be the cartoon from July 13.  urlLink Full Post and Comments    '),(2390,2680773,'12,July,2004','2004-07-12 00:00:00','   urlLink Ken Jennings  has been winning on Jeopardy since June 2 and is less than $80,000 away from a million.  Apparently, Jeopardy lifted its five-day limit for contestants at the beginning of this season (another guy had a 7 day run in January) and Jennings has been making the most of it.  If you remember the early run of  Who Wants to Be a Millionaire , it\'s my opinion that people were fascinated because of the show\'s ability to perpetually beat its contestants.   So, when that first million-dollar contestant used his final lifeline simply to announce to his father that he had beaten the machine, the show reached its unsurmountable apex.  Ken Jennings\' Jeopardy run has the potential to rekindle the phenom of Jeopardy in a very similar yet fundamentally opposite way.  Here is a person the machine (and other people) simply cannot beat.  As remarkable as the success will be, the climax of this story is going to be his defeat.   That  is what we are watching for however much we respect/adore Jennings himself.  As such, his humble and human approach to his winnings (telling his wife the money would be going solely to DVDs, for instance) is probably the healthiest thing for him.  ...I may be renting  urlLink Quiz Show  this week.  urlLink Full Post and Comments    '),(2391,2680773,'07,July,2004','2004-07-07 00:00:00','   Will Ferrell just showed up and went crazy on CNBC\'s Power Lunch, not where I would expect him to be and - it seemed - not where the hosts of that particular show expected him to be.  He yelled a bit, threw a chair on a desk, and then pounded away at a tri-LCD-screen computer and screamed \"This computer isn\'t real!\"  urlLink Full Post and Comments    '),(2392,2680773,'07,July,2004','2004-07-07 00:00:00','   In case you didn\'t notice,  urlLink Michael Chabon  has a story credit for Spider-Man 2.  So, he\'s written a terrific novel about comic books, a pretty decent movie based on a comic book but has yet to write a good comic (this is where the link to the Pop Matters review of the comic book spin-off of Kavalier and Clay will go, but they\'re running sluggish today).  urlLink Full Post and Comments    '),(2393,2680773,'07,July,2004','2004-07-07 00:00:00','   Spiderman 2  is not the  urlLink best of a bad genre  (link from  urlLink Bookslut )...  I reserve that prize for  X-Men 2  (capturing the spirit of the serials without apologizing for being a big budget action flick) or  American Splendor  (a comic that transcends being a comic book, a movie that transcends being an adaptation).   S2  is manically disjointed and the series\' characters have turned into caricatures of what they were in the first film.  Overall, the movie seems uncertain how it wants to spend its time when there are not collossal battles going on.   What makes  S2  better than the first and a pretty decent comic book film, then, is this: the villain.  Let\'s be honest... Most Spiderman villains are small-time.  Sure, they can do some real damage, but these guys aren\'t the devour-the-world, wipe-out-the-human-race, gruesomely-unstable types.  True, somebody has got to stop   urlLink Lizard  or  urlLink Rhino  and stop them quickly, but on a national or even global scale their potential to harm is low.  They are little buckets of rage, which makes for intense dueling with other superhumans and a few dozen homicides but no real danger to society as a whole.  They are the status quo of villainy - they punch really hard and are good at blowing things up.  There\'s nothing particular captivating about their method of destruction, they\'re just hell bent on it.  There are really only a few Spiderman villains you are interested in seeing in action: only one is human.  He is  urlLink Doctor Octopus .  Four bionic tendrils attached to his back reach, grab, suspend, crush and delicately manipulate in ways that a two-ton muscle man or a special suit on a hovering glider could not even attempt.  Sure, the real spider is Peter Parker, but when Otto Octavius sinisterly looms above another character he is as frightening as a prowling arachnid. The wild strands of chaos are spectacular when these cyborg arms claw and climb against the web-slinger.  Where Spidey\'s fights with the Green Goblin were like tennis, a concentrated volley of lethal strikes, his fights with the Doc Ock are like football, multiple fronts on each side attempting to outposition the opposition.  These scenes and this character mask the faults of the film so exceptionally well that you don\'t mind the fact that the rest of the movie has regressed.  Of course, the problem for the foreseeable future is that Sam Raimi will send us back to the lackluster world of the Green Goblin and astronaut John Jameson as Manwolf (maybe even Dr. Connors as Lizard?) before we get another not-dull, maybe even spectacular villain (in the form of Venom).  urlLink Full Post and Comments    '),(2394,2680773,'07,July,2004','2004-07-07 00:00:00','   A former Kindler has some thoughts on  urlLink  video games and addiction  in response to the  urlLink Everquest Daily Grind , a collection of anecdotes from the associates of Evercrack junkies where you can read stories like this:   \"I\'ve tried tricking myself into letting my own baggage about the game go, for instance by pretending to myself that he\'s a Ph.D. candidate, say, or a medical student who needs to study all the time and thus cannot fairly be expected to do his share of cooking, shopping, or cleaning. I find myself hoping against hope that I\'m not simply being used (allowing myself to be used), and that our friendship is in fact one-sided and wonderfully convenient for him.\"   As I write in the GGA coments section, I recently read that video games release dopamine in the brain, inducing pleasure and concentration and also making games highly addicting.  With gaming\'s mainstream popularity rapidly growing, we are seeing more cases like those on EDG appear without the tag \"he just really likes video games\" alongside.  As much as those in the gaming culture want to protect the hobby, there are negative issues that may need addressing.  urlLink Full Post and Comments    '),(2395,2680773,'06,August,2004','2004-08-06 00:00:00','   Went to Millenium Park today.  Click below for pictures; thoughts will follow.       This is the big face fountain section of the park; needless to say, the chids love it.     Here is the inside of the Cloudgate, known by Chicagoans as the Bean.  urlLink Full Post and Comments    '),(2396,2680773,'04,August,2004','2004-08-04 00:00:00','   The meta-art project  urlLink \"Library of Unwritten Books\"  takes people\'s ideas for books they\'d like-to-write-but-probably-never-will and collects them in a published form.  Now, I realize I am probably sucking the entire life and vibrancy out of the project by writing what I am about to (for that I apologize greatly), but...  Who owns the copyright?  urlLink Full Post and Comments    '),(2397,2680773,'04,August,2004','2004-08-04 00:00:00','   ::Gasp::  The Jim Henson Company is behind a film called   urlLink Mirrormask   written by Neil Gaiman and directed by Dave McKean.  Has the Sandman taken hold of me or is this real?  Well, judging from the linked article\'s emphasis on the business of the film, this is happening in our actual lives.  While it seems short on puppetry largely negating the Henson corner of this awe-striking triumverate, I\'m ecstatic to see a Gaiman/McKean work hit the screen.  This \"adaptation\" will be interesting as we see if McKean\'s distinct page style will effectively transfer to celluloid.  I feel like JHC (ha!) has lost its spirit, bleeding the Muppets dry (they don\'t even have blood!) and now - it is reported - trying to tap into the financial butter that is  Labyrinth  and  Dark Crystal .  ::Sigh::  urlLink Full Post and Comments    '),(2398,2680773,'04,August,2004','2004-08-04 00:00:00','   Reporters across the city are visibly shocked and appalled that Chicago is not on the list of most recent probable targets for Al Qaeda.  You can almost hear the underlying stream of obscenities in their reports (vowels removed for the innocent), \"Motherf*cking Newark?  New-*ss-f*cking-ark?  What the f*ck Al Qaeda?  We\'re the motherf*cking sh*t dispensers for commodities f*cking exchange!  You don\'t get your commodities sh*t lest you come to our f*cking door, b*tches.\"  Of course, the news has to remove more than the vowels.  So, the reports are desperate pleas for Al Qaeda to consider Chicago as a target but because of our awesome security measures to not-even-think-about-it.  A person should, when discovering they are not the targets of a deadly and largely indefensible suicide attack, say to themselves some brief praise to a deity of their choice and/or persuasion.  If they are a potential target, they might even avoid advertising to their enemies all the viable and important strategies and locations for such an attack.  But this is the \"Second City\".  Chicagoans take great pride in our place on the ladder of national socio-political influence, and if we can\'t even have that... well, there\'s gonna be hell to pay.  It\'s just a question of who\'s going to be paying.  urlLink Full Post and Comments    '),(2399,2680773,'03,August,2004','2004-08-03 00:00:00','   One thing I have learned in my old age is to not eat dog treats.  This very funny experiment (challenge, dare, what-have-you) almost turned horrible.  But since it didn\'t, I\'ll post the video clip here for the world to enjoy.  urlLink Full Post and Comments    '),(2400,2680773,'02,August,2004','2004-08-02 00:00:00','   The big news from the zoo was that I found elephant book ends... something I\'ve been hunting down for a while.  But here\'s a couple photos I took:     Some fish there.  Here\'s a building!  urlLink Full Post and Comments    '),(2401,2680773,'02,August,2004','2004-08-02 00:00:00','   Finally got to play with some stuff that is long overdue.  -Snazzy New Design.  Really needed to pull the visual of the blog closer to where the site is going, so edited one of the blogger templates a bit for the desired effect.  -Snazzy New T-Shirt.  Right now, there\'s just a female version of  urlLink \"The Comma Joke Ringer\"  aka \"Area Hippies...\", the second in the Hippie series of RF shirts.  I hope to have a few more fun designs up by the end of the week; the Studs merchandise will be coming down.  urlLink Full Post and Comments    '),(2402,2680773,'02,August,2004','2004-08-02 00:00:00','   There\'s two kinds of fun / there\'s things and there\'s stuff / and they just cannot get along / which side you on?   New  urlLink Danielson Familie  song rocks mock caucus.  urlLink Full Post and Comments    '),(2403,4040811,'30,July,2004','2004-07-30 00:00:00','    Hello,  First here is some information:  1.)I am in the process of trying to get some pictures up here but it is difficult; my computer does not want to cooperate. But I am trying so hopefully pics to come!  2.)Also for all of those who do not know this Sunday is the Alpha Marchdown. It is the biggest stepshow in St. Louis. Other Greeks have tried to hold a show, but none have been as successful as this. It is always a great show. We are the reigning champs! That\'s right Blue and White Sweep last year and this year to big baby! All of this is to say if you have nothing better to do and are in the St. Louis area check it out. Heck if you are out of town..ROAD TRIP! Tickets are on sale at From One Greek To Another on Delmar and they sell them at the door. My advice is to just get the VIP ticket. It includes the show, dinner and the after party. The whole event takes place on the campus of University of Missouri-St. Louis (which is my alma mater shout out to UMSL). If you need more info go to my profile and email me:)  3.)I have recieved feedback that when you try to comment on the page it tries to make you sign up to have your own blog page! Sorry it does that but to bypass all of that just leave a message anonymously and put your name in the body of your post! You can still just always go to my profile and email me...  Now to the real deal,  So my last detailed post was about the club. In which post I stated that I was going to stay out of the club circuit for about six months. (about the needed healing time) Well this morning I heard an advertisment about grown folks Friday at Spruills. It is 23 and up! What! No young bucks in Air Force Ones! We might have something here. So one day I will go there. Not today though because it is raining. The next time I go anywhere it has to be clear and sunny skies.  The funny thing about Spruills is that it is a banquet hall. When did the banquet hall turn into the club? Some of us know how to pull a hustle... During the day it is Rolanda\'s beauty salon but at 8pm it becomes the new concert venue!  The last time I was at Spruills I was, oh about 16 years old and I was there for my ROTC military banquet! This is just too ironic for me.  So if you have ever been to Spruills let me know...Is it even worth the gas that I will burn to get there?  God Bless, Raquel   '),(2404,4040811,'27,July,2004','2004-07-27 00:00:00','    I have gotten a lot of feedback on my post, Where the heck am I, but I have yet to get a comment here. Leave your comments here or, email them to me. I will ask for permission to post them then I will post them right here for everyone to read! Exciting I know!   Now my co-worker Alandrea went to the club last week with me as well and also had quite an experience. She valet her car at&nbsp;the club because we got the hook up on it.  (Hook-up: can be a noun or verb-to get free or get a discount on. In this case it was free)  I could have gotten&nbsp;valet but I had already parked and just decided to forget it because my friend and I parked right around the corner.   So when Al decides to leave, the guy who valet her car could not find her keys! He pulled a whole bunch of keys out of his pocket and asked her which keys belonged to her. She ended up having to go outside in the rain to help him look for them. Now I will admit that I do not know a lot about valet but from what I see on TV there should be a box, a cabinet of sort, to house the keys that you have acquired. Your pocket, not very professional.   We need to improve people, we need to improve.   God Bless,  Raquel   '),(2405,4040811,'26,July,2004','2004-07-26 00:00:00','    Ok,   So now for my second post I will get to it. I will get the party started like the Black Eye Peas. I will talk about a subject that I have expressed to a few of my friends in St. Louis and we all agree.   When did the club life in ST. LOUIS MO turn in to New York.   I mean come on. I know we trying to come up but my Goodness! This is getting out of hand and needs to be addressed.   Let me just tell you about how I remember club life: (note: never a big clubber per say, but I did attend a few B.C.* when I turned 18 and thought I was too grown)&nbsp;&nbsp;  9:00am decide that I am going to go to the club with the girls  12:00 noon: go to mall to get outfit or get matching pieces to items that I already own&nbsp;in an attempt to make something funky happen. I have generally seen this outfit in my mind and now I will try to bring it to life.  6:00 pm go to sleep  9:00 pm wake up and prepare for the club by showering, doing hair, makeup, and purse change (can be done in car if your are running out of time)  10:30pm leave for the club or friends house if I am driving  11:00-11:15pm arrive at club  Now pay close attention here:  After arrival at the club I exit vehicle and walk to the door. Show my ID go to the cashier, pay and go into the club and commence getting my party on!   Clubs in 2004 post Nelly, post Chingy, have gotten out of control! These club owners and bouncers think they are God or something and tend to treat you how the hell ever.   Now when you go to the club, you can get to the club after 10:30 if you bad, but you will be waiting in line. I ABSOLUTELY HATE LINES AT THE CLUB. This comes from a bad experience on my friend Candice\'s birthday where we waited in line&nbsp;FOR ONE HOUR and did not get in!&nbsp;I did not get dressed up in heals to stand outside, sometimes in the rain. Oh no! You got me confused with that other girl over there. Call me stuck up if ya want to! I am just stating my opinion.   Ok, let me calm down. But last week was the worst.  I go to the club right. I have to get there at 10pm-10:15 to avoid the line. Even if you get there and there is no line that make you stand there like Boo boo the fool just because they can. Then they let you in but you are there all extra early and then the music don\'t get right until about 12:00. I don\'t drink anymore so I usually just sit there with a water and look at what people have on.   My second issue with the club is how crazy crowded it is! It is scary&nbsp;crowded.&nbsp;It is if&nbsp;some dropped a cigarette we would all die crowded because no one would have a prayer of making it to the exit. Quit letting all these people in! If you want all those people in here expand, open the back door to the alley or&nbsp;something! I go to the club to dance, I have no room to do it, I paid to get in, and&nbsp;I am not a satisfied customer.   Let me speak to the manager.....  Hello. Because I did not have adequate room to shake my behind, I need a refund.   Then when I am ready to leave, it is pouring down rain.&nbsp;Now there are all these bouncers at the door and they all have umbrellas but are not offering to escort anyone to the car! SO I got completely drenched so much so I had to wash my hair. It looked like I had been the shower. I was SO not happy.   Well that is it for me and the club for I would guess somewhere around another six months. It will take a while to forget about this one. I think I will be hanging out at the A\'s: AMC and AMF.   Thank you for allowing me to vent, more stories about my three times to the club this month to come!   God Bless,  Raquel   * B.C. stands for Before Christ was a big part of my life.   '),(2406,4040811,'26,July,2004','2004-07-26 00:00:00','    Well!&nbsp;&nbsp;To all of you I say welcome to my world!   To all that don\'t know me my name is Raquel. I am a 23 year old college graduate currently residing in St. Louis Missouri.   University of Missouri-St. Louis!   I am also a proud member of Zeta Phi Beta Sorority Inc!   I have to state why I finally decided to do this:   Well I read a lot of these things and I find the stories absolutely hilarious. I question myself am I witty and creative enough to maintain something like this? Is my life exciting enough? Do people give a care what I think about things in the world? Well there is only one way to find out.   Please feel free to comment on any post you chose. I welcome all suggestions and opinions!   I think that this is enough for one post, expect another latter today or tomorrow!   God Bless,  Raquel   '),(2407,4040811,'11,August,2004','2004-08-11 00:00:00','    Hello Everybody, I have to take this time to issue a disclaimer:  In this here post I will make many comments towards men that may be construed as a little angry. I am not a bitter Black woman; this post is meant for entertainment purposes and not to offend anyone. Reading this post may cause women to yell \"thatâ€™s right\" and \"uh huh girl, you better say that\". It also may cause men to feel picked on or targeted. Please read w/caution.   Now maybe about a week ago I ran into one of my old boyfriends. I don\'t even know if you can call it boyfriend or girlfriend given that I was all of 13 years old in the eight grade at the time. The relationship lasted exactly one week. I found out he was cheating on me so I started this rumor that I was talking to someone else too so he would break up with me. I used to have a serious problem with breaking up with people so I would do things to get them to dump me. I know, I needed professional help, I know. So after this break up, we just went back to being friends.  So I had not saw this guy for about 3 years or so until about a week ago. I am sitting on my porch talking to my boy Justin (Hi Justin!) on the phone and he walks up and says Raquel? Now it is pretty dark out at this time so I say, \"Who is that dar?\" And he says his name and I get off the phone and give him a hug. He is like, \"dang you look good, and you lost weight?\" He then bends over to tie his shoe and says, \"Dang you got some pretty feet.\" So I am like ok, you are quite the little charmer. Not that I did not like it; every girl likes a compliment... several is a bonus. So he asked me if I was single and I said yes, and he asked could he call me and I was like cool, and I gave him my number.  He called me about a week latter and was like can I come over. I said cool. Now when it is someone coming over and I don\'t have to leave the house, I don\'t put on clothes until they get there. So I eventually fall asleep and he never shows up.  So maybe a couple days after that he calls and is like can I come over tonight I am sorry about last night. I say cool and eventually fall asleep and he never shows.  So last Monday he calls and says sorry I have been busy, In the studio (I won\'t even get into this), can I come over latter and bring you a CD, I say cool and end up leaving to go somewhere and he never showed because he never called.  So last night, he calls when I am on my way from Bible study and was likeâ€ Where you at\"...\"I am in the car\".....\"Where in the car\".....\"on 367\".....\"Oh man I am at your house now\".....\"Well wait, I will be there in about 10 minutes\".....\"I can\'t I be on the move man! I have a stop to make a run in Berkley and I will be right back\".... I start to laugh and say sure. Of course he never called back.  Now Even though I could care less what he does, one phrase did kind of make me upset. \"I be on the move man! What the heck is that? Like I am not on the move. I have a job, a workout schedule, a church I attend regularly on Sunday and during the week, plus family, friends, sorority business, etc. Now if I can be where I say I am going to be when I say I am going to be there, surly you who has no job, church, sorority, workout schedule, just family and friends can be there. I am just telling it like it TI is.  MEN, yes I am talking to some of you, wants us to always understand. But you don\'t understand. We have lives too. I know it might not seem like it because we answer the phone when you call and do things like come over your house to visit, when you want us to, when we are dead tired. Because we are available to take you to work when your car is down, there to rub your back when you are sick, angry and frustrated and accompany you to parties, family reunions, funerals, and always manage to look great on your arm. You want to know why we do this men, because when we care about you, WE MAKE YOU A PRIORITY IN OUR LIVES. Thatâ€™s right! We move you to the top of the list. All I am asking for is the same treatment that you expect...thatâ€™s all.  All I am saying is think about it. Do you right now, and this can go for men and women, have someone who just thinks you are the cats meow and you play them to the left and off the cliff?Shame isn\'t it.  SO then I am talking to my friend Justin again about this and I said\" I am not really worried about it because it is not like I could ever have a relationship w/ dude we are on two different pages. He would only be good for talking to, hearing him compliment me and rub my feet.\" He made it sound like I was wrong. What do you think people based on what I have told you about him? Let me know if I am evil.  God Bless Raquel   '),(2408,4040811,'03,August,2004','2004-08-03 00:00:00','    Hello All!  Yes, Yes We won again. The Zeta Ladies are just BADDDD! Oh my goodness! No Blue and White Sweep though, the Sigmas did not win, but I am still ecstatic about my sorors though. Just beware that this is going to be a long post, but very very entertaining:)  So, this years show was by far the best show that I have been to, and I counted this week and this is my 6th year going! I am getting so old. Everyone stepped except for the Omega\'s which kind of shocked me because they always step. The Sigma Gamma Rho\'s opened the show. They had a 70\'s... old school theme, you get the picture. They had a good show, very high energy and very difficult steps. They did really well.  Then it was time for the main event! Here is what the people came to see! Ha Ha! (In case you do not know me, I talk much trash about sports, my sorority, my church, just about anything I am passionate about.) Our theme was a court case. Pure genius if you ask me. It started with a powerpoint show where it stated that the ladies of Zeta Phi Beta Sorority, Inc. have been accused of theft of stepshows and strolls, something we deny and tonight, you will be the judge. Then the powerpoint introduced the judge and the prosecutor Ivy D. Poodle. Then it introduced us. I can\'t even remember the song that we came out on but it was hype! I was so excited! I was in the VIP section with a whole lot of sorors and we were screaming! It was great. There show was so hot, not one mistake, steps were difficult, they held the theme the entire show and it made it interesting. I was so proud.  Then it was the Delta\'s turn. Now at first I was a little frightened because it was like 30 of them stepping and they came out hype too. But then they had no theme they were just out there doing there thing. It got kind of old to look at. They wore casual boots w/no heel or tap so it made their steps sound too clunky and not precise and crisp. At the end I was no longer worried:)  The AKA\'s, well again they came out hype. They had no theme as well, made about two or three mistakes....We had this puppy in the bag! I knew we won by the intermission. No need to even wait on the judges.  Now here is why I think that we always win. WE ALWAYS HAVE A THEME. A theme is essential to keeping your show interesting. Stepshows are about entertainment, and a plot of some kind entertains. It shows creativity and that you put in a lot of effort. Anyone can put a step together and perform it but, to come up with a theme, get clothes, costumes and props to fit your theme, takes effort and work. Plus your want to maximize all your point categories and those creativity and originality points can go a long way. If you have a theme you have to carry it through the whole show. You can\'t just have an intro theme; it will make your show start out hot but get cold at the end like so many other shows did.  The men I will sum up here: The Kappa\'s won, there show was good. The Iota\'s stepped they did this step with knives and two of the guys threw the knives out at the judges instead of down and one of the knives hit the Delta judge. (She was not injured though)  Now during the show I sat on the 1st row of the balcony in the center (I got there an hour early for those seats). I sat with my friend Candice and her sister Sydney. I was also next to my former protege Kenyatta on my left. At the intermission people started to sit in other peoples seats. So these 3 girls sat in Kenyatta and her friends seats and refused to move. The Alphas tried to be nice but she refused to move. We told her that those seats were taken but she sat there anyway. That kind of thing makes me so mad. I know I should not care, no one was crazy enough to sit in my seat, but I just hate ignorant people! If you get there half way though the show, you need to sit in the back at the top, behind that pole. These people are crazy these days, I guess morals are out the window.  The afterparty was cool, I had fun. I stayed till about 1am and I was off to the bed!  All in all, one of the best dang marchdowns period.  God Bless, Raquel  p.s. I will post latter about what people had on. That deserves special attention.   '),(2409,1401816,'02,August,2004','2004-08-02 00:00:00','   Today I can take a calmer look at things. I want to write a jazzy piece of music called \'Return of the Panty Cats\' and maybe do a gig with a specially made band. So, the song is underway (drums, piano melody but not the changes yet) and the band would depend on the instrumentation of the tune - so that\'ll be a while.  '),(2410,1401816,'02,August,2004','2004-08-02 00:00:00','   This is my first ever Blog entry so I had better make it something special. OK, hello to anyone out there. I\'m a bass player and a singer (and a teacher too). This should take you to one of my bands... www.soulbeaver.com  In the coming days, weeks, years I\'ll write up what is going on and tell you about my life, and the interesting things going on around me. \'Interesting\' is a word we\'ll have to redefine as we go along.  '),(2411,1401816,'02,August,2004','2004-08-02 00:00:00','   A further note!  Just to make this worthwhile for you to read, I\'m going to add dates for the various bands I play in. Hope that makes things more interesting. No? I can see you shaking your head. Just wait and see.  '),(2412,1401816,'02,August,2004','2004-08-02 00:00:00','   Return of the Panty Cats is taking shape in my head - this will only make sense if you read the last post - and this is not an excuse for not having written anything. The tune is swirling around, slowly taking shape, losing the more cliched elements it began with and trying hard not to sound too contrived. All this in my head - no wonder I can\'t sleep.  '),(2413,1401816,'02,August,2004','2004-08-02 00:00:00','   OK, Return of the Panty Cats is a step closer to completion. Quite a small step though. I\'ve become the proud owner of a 16 track digital studio (it needs a name) and have the drums, piano and organ parts for the tune down.  In the last entry I promised details of gigs. Here they are: Sat 21 June - The Chester Arms, Chester Street, off Iffley Road, Oxford. FREE ENTRY. Playing from 8:45   Sat 5th July - Iffleystock @ The Prince Of Wales Iffley, Oxford.   Sat 12 July - The Barn @ The Red Lion - Witney.  £5 entry on the door, doors open @ 10.30pm with a bar \'till 1am.  See http://www.redlionwitney.co.uk  '),(2414,1401816,'02,August,2004','2004-08-02 00:00:00','   urlLink          BBC NEWS | Entertainment | Music | R&B music legend Ray Charles dies   Ray Charles has just died. I\'ll remember him for some great songs, a truly dodgy country album and that wonderful appearance in the Blues Brothers movie.  '),(2415,1401816,'02,August,2004','2004-08-02 00:00:00','   urlLink A Blog\'s Life - Ole Gal Fibs  I\'m not sure how sad it is for an unread blog like mine to be suggesting a much better blog that is still, well, a blog. Is anybody out there?  '),(2416,1401816,'02,August,2004','2004-08-02 00:00:00','   So, for those of you that want to know, my fab other half and I have  booked our honeymoon but not the wedding. Yet. Since we can\'t really  call it a honeymoon until we have a wedding I\'ve decided to call it the  holimoon. Part holiday, part honeymoon. Oh, you got that.  '),(2417,1401816,'02,August,2004','2004-08-02 00:00:00','   My dad bought an iBook and wanted a BT account like I have. Easy!  Um, no.  BTW. For those of you already on the net (surely all of you) this is a pointless post.  BT and their internet people are a friendly, lovely bunch of folk I\'m sure. However, try signing up for an account if you are a Mac user. Especially if you use OSX and browse with Safari.  Bloody nightmare is the phrase that comes to mind. Blocked pop up screens, faulty system problems and limited advice from the rather expensive help desk. Eventually all sorted out but how many times did I consider going with Demon?? Quite a few, that\'s how many.  Anyway, my beloved dad is now online with BT and everyone is happy.  '),(2418,1401816,'02,August,2004','2004-08-02 00:00:00','   As some of you may have noticed the last post and this one are the  first for a very long time. \'Return of the Pantycats\' has been on hold  for a while. But news of other musical goings on will be posted here  shortly.  '),(2419,1401816,'02,August,2004','2004-08-02 00:00:00','   urlLink Tony Levin\'s Road Diary  Went to see Peter Gabriel in concert last night. Truly inspiring. I recommend the great bass player Tony Levin\'s site too.  '),(2420,1401816,'02,August,2004','2004-08-02 00:00:00','   Eight weeks to the wedding. I\'ve decided to turn this blog into a wedding countdown area. No wait, that shouldn\'t be too bad as it is not supposed to be a conventional wedding. There shouldn\'t be any tales of picking out colours for themed clothing, or who to put on the guest list. We picked only the very closest friends and family, thereby alienating almost everyone else we know. The good news is that we\'re having a party for the others so they can drop hints about not being invited to the wedding. Anyway, the countdown begins, I still need a job and somewhere for the cat to live while we\'re away.  Non-wedding news Has anyone else been driven nuts by \'Petals around the rose\' recently? The are brain teasers and then there are things like this.  The new project is putting all my Cds onto a separate hard drive to produce Will\'s Enormous Jukebox. This is a jukebox that, statistically, will probably reward you with some Zappa or Miles Davis. My iTunes library is now up to eight days of music.  OK, I know this is getting nerdy and self-indulgent - but it\'s my blog for goodness sake!  Current listening: Butch & Butch by the Keith Jarrett trio  '),(2421,1401816,'02,August,2004','2004-08-02 00:00:00','   The holimoon saga - booked a holiday which will become a honeymoon when we book the wedding - continues. Received wisdom is that you must book a registry office or church at least a year in advance to have any hope of getting the place, county or indeed country that you want. This is doubly true if you want to get married in the middle of August. Well, fate, blessing or fantastic good luck has intervened and Jenna and I will be getting married in around eight weeks.  Phew!  '),(2422,1401816,'02,August,2004','2004-08-02 00:00:00','   Date: 2004-06-13 20:08:33 +0100 Topic: Bob Dylan  I\'m slowly getting into Bob Dylan\'s music. A friend at university had  recommended \'Blood on the tracks\' as a starting point. I bought the CD,  listened, didn\'t like it, put the CD away, turned forty, started  listening to Neil Young, had a chat with a friend about Dylan and  thought \"hey, what about \'Blood on the tracks\', maybe it\'s time to  listen again.\" It was. I love almost all of it. I\'ll try to find my old vinyl copy of  \'Bringing it all back home\'.  What worries me is that there might be something genetic about turning  forty and then doing the washing up to Dylan and Neil Young. I\'ll put  the Wiretapper 11 CD on from this month\'s Wire (www.thewire.co.uk) magazine to balance  things out.  '),(2423,1401816,'02,August,2004','2004-08-02 00:00:00','   &#183;  Since the last post about things sporty Tim Henman has won the first and second sets of his game. In honour of this I am posting some fact about strawberries (taken from the  urlLink BBC Sports site  to give us all a sense of being at Wimbledon and eating overpriced fruit.      &#183;  &nbsp;  Roughly 27,000 kilos of strawberries are eaten during the Wimbledon Championships, together with 7,000 litres of cream. New bellies please!        &#183;  &nbsp; Wimbledon may have 64 seeds in the men\'s and women\'s competition, but did you know there are about 200 seeds in every strawberry?       &#183;  &nbsp; A 100g serving of strawberries will serve you for just 50 calories - that\'s a whooping 0g of fat - Another punnet anyone?       &#183;  &nbsp; We\'re all smiles - did you know strawberries can help whiten your teeth? The acids in the fruit help to remove stains.       &#183;  &nbsp;    Strawberries are full of a special substance called ellagic acid which can help fight cancers.           &#183;  &nbsp; They are also full of    flavonoids   , which form part of the strawberry\'s red colour. They are known to help reduce cholesterol from clogging up the heart\'s arteries.       &#183;  &nbsp;    Eight strawberries have more vitamin C than an orange, that\'s 140% of our daily recommended amount.           &#183;  &nbsp; The word strawberry comes from the Old English words &quot;   streowberie   &quot; or &quot;   streawbelige   &quot;. There are a few theories about how they got their name. The &quot;straw&quot; bit could come from the straw that was used to keep the strawberries fresh, or it could come from &quot;strewed&quot;, which means to spread wide.       &#183;  &nbsp;    The Latin name for the famous everyday strawberry is  Fragaria Ananassa . We\'ll just stick to plain old strawberries for now.          &#183;  &nbsp; You cannot be serious! Strawberries can also be white or yellow and some can even taste like pineapples.        &#183;  &nbsp;    The strawberry plant, along with its good mate the raspberry plant, are members of the rose family.           &#183;  &nbsp; The Romans loved their wild strawberries - and not only for their taste. They believed strawberries could cure a number of illnesses.      &#183;  &nbsp;    If you\'re watching your waistline, why not sprinkle a little black pepper over your strawberries instead of the usual calorific cream? It doesn\'t sound that appetising, but you\'ll be pleasantly surprised...  '),(2424,1401816,'02,August,2004','2004-08-02 00:00:00','   A terrible night last night. As soon as it became obvious that the England football team were going to face a penalty shootout I should have turned off the television and got back to adding Beatles albums to my enormous jukebox project.  It was certainly a night of contrasts. I sat mute for the last half hour, biting my hand and rocking backwards and forwards like a vegetating animal in a badly maintained zoo. Jenna, in order to bring balance, got in touch with her inner hooligan and enjoyed periods of swearing and dancing around the room in a lively but dangerous fashion.  The cat took great umbrage at our lack of attention in her, even resorting to bringing in a mouse during the penalties. Both survived.  In the non-football world it seems that cake makers and flower arranging people are being readied. I, as you might imagine, have had nothing to do with this. Although, as the countdown continues I am starting to think about how to spend the hours before the ceremony. Squash games, coastal walks and full on panic counseling have been features of friends\' pre-wedding hours.  The last thing to comment on today is that the England rugby side are playing Australia and Tim Henman faces a chap called Ivo Heuberger. As I\'m writing this the match is in progress, with Tim one game up in the first set (it\'s 6-5). It seems Tim lost his serve after double faulting in the first set. No tension there then.  Time to get back to work.  '),(2425,1401816,'02,August,2004','2004-08-02 00:00:00','   Well obviously I\'m not going to be able to keep up a day by day running commentary on how the preparations are going. Who would want to read that? Who (even more importantly) would want to type it?      But, having said that, the meeting with the registrar is booked and the cat has a home for the short time we\'re away.     Next? A power meeting (serious coffee, find space to sit down in the front room, no distractions) with Jenna on Saturday to work out what we have to do next. I suspect buying the rings will be problematic. Do I really want to spend a lot of money on a ring I\'m never going to wear? Can I get away with some rolled up shiny paper on the day?      Current listening: Someone teaching in the room next door.  '),(2426,1401816,'02,August,2004','2004-08-02 00:00:00','   Don\'t panic - it\'s back! The Hitchhiker\'s Guide to the Galaxy is back on Radio 4 - 25 years after it first started. The BBC has officially announced that the sci-fi comedy will return on 21 September for a new six-part series. Most of the original cast - including Simon Jones, Geoffrey McGivern and Mark Wing-Davey - are reunited for the new episodes, which will air as a Hollywood film of the book nears completion.  Even author Douglas Adams will feature in the new radio episodes: he made recordings for a character called Agrajag a few years before his death in 2001. The programmes are based on Adams\'s last three books: Life, the Universe and Everything; So Long and Thanks for All the Fish; and Mostly Harmless. Another new series is scheduled for broadcast next spring.  A preview can be heard at: www.bbc.co.uk/radio4/hitchhikers  '),(2427,1401816,'02,August,2004','2004-08-02 00:00:00','   Less than eight weeks to go to the wedding and nothing is planned. Jenna is trying to arrange a meeting with the registrar... and that\'s about it. We have the honeymoon booked, that\'s true. But no home for the cat, no idea what we\'re wearing, how we get to the place (or even where the place is in my case). Confusion reigns.      What makes all this amusing (to me, if no one else) is that nobody seems to believe we are this badly organised/laid back or that we hadn\'t secretly arranged things months ago and are only now choosing to tell people.      England play Croatia tonight, that should relieve the stress.       Current listening: the buzz of computer hard drives in the college work room.  '),(2428,1401816,'02,August,2004','2004-08-02 00:00:00','   LONDON (Reuters) - A major investment bank is advising clients to have sex, get more sleep and stop equating happiness with money -- turning the industry image of hard-nosed dealmakers on its head.  German-owned Dresdner Kleinwort Wasserstein offers the advice in a note to clients by its strategist James Montier.  \"I thought it was time that I reminded people there was more to life than watching screens every day,\" he told Reuters.  The note recommends clients have sex, ideally with someone they love, reflect on the good things in life, give their bodies enough sleep and exercise regularly.  But they shouldn\'t get too carried away.  \"I still need a little bit of money just to keep me happy,\" said Montier.  (From Dave Barry)  '),(2429,1401816,'02,August,2004','2004-08-02 00:00:00','   Two fetes in two days. England is gripped with sports fever this week (and next, and the one after that) but this weekend fete-mania has broken out. I went to one in Coleshill, Oxfordshire yesterday. It was at the home George Martin (the nice man who produced some Peter Sellers records) and enjoyed home made produce as well as winning \'The Little Book of Calm\'. Today I visited Stanford in the Vale Summer festival (it\'s a fete really). Fab stalls, won a bottle of fizz and some lovely looking French cider from the town twinning stall. On the way there I heard hymn singing from a small local church and enjoyed the sight of a Ghandi lookalike cooking veggy burgers on a grill in the rain. Someone in Paris said they thought I looked like Ghandi. It\'s good to return the favour (must change my glassess too).  OK, this does not sound like the blog of an, up-all-night bassplayer/net-nerd. But I love cakes, fetes and... oh, back to the sport...  Current listening: \'Take your clothes off when you dance\' - Frank Zappa (from the Lost Episodes CD)  '),(2430,1401816,'02,August,2004','2004-08-02 00:00:00','   Only six weeks to go to the wedding. I\'ve decided to abandon the apologetic tone of this blog and will plough on with wedding news safe in the knowledge that some of you find it interesting.  So, today we bought the rings. Mine does not make me invisible (although I can dream) and now resides by the computer. Jenna\'s is being sized. We also have the post wedding stay over sorted. A posh hotel for the two nights before we get whisked off to Polenta in Italy. OK, it\'s Positano, but that always make me think of Postman Pat, or Il Postino. Polenta makes me think of food (mmm, food) which is never a bad thing.  I got a little freaked last night watching \'Will & Grace\' when Grace got married and discovered she didn\'t even know her new husband\'s real name. If Jenna turns out to be a Doris I may take a while to get over it.  Now, the sad tale. Ach! Excuse the pun. Yesterday I saw a fox in front of my car. Obviously injured and unable to move its back legs I was debated whether to get out and wring its neck, or drive over it (I certainly couldn\'t leave it in this state). Just as I decided to take the cowardly way out and run it over it looked up at me, eyes directly to mine, full of hopelessness and frustration (yes I\'m anthropomorphizing). I went to put the car in gear but at that moment the fox dragged itself off the road, using its front legs, and propelled itself into the undergrowth of a field. Looking over the road I saw a woman in her car who had stopped to watch, tears streaming down her face.  Well, I thought, this is a great omen.  Current listening: Back Home Blues - Charlie Parker  '),(2431,1401816,'02,August,2004','2004-08-02 00:00:00','   A miracle. On Saturday we sat down and planned the whole wedding. Everything is sorted out. Nothing else needs to be done.      Truthfully though. On Saturday we wrote out a list of everything that needs to be done and it is now on the fridge door. Step one: remember to look at the list every morning.  '),(2432,1401816,'02,August,2004','2004-08-02 00:00:00','   We planned to have our own mini Glastonbury yesterday. Nora Jones was performing at Althorp Park so we bought tickets, made up a picnic and locked ourselves out of the house and car before we even got going.      Thanks to understanding neighbours and help from our landlord\'s son we were off again after only an hour or two\'s delay. As it transpired this worked in our favour, as we did not have to spend those hours stuck in a queue of traffic.      Heavy wind and rain tried to put us off, well alright, tried to put me off. The corporate nature of the \'shops in tents\' was rather galling too. However, the support - Amos Lee - and Nora with her Handsome Band were excellent.      I was pleased that the live versions of her songs brought out an edge that I hadn\'t heard on record. Great playing from all concerned and our little candle collection burned brightly through the night.      Getting out was, of course, a nightmare. Hundreds of cars all trying to get out of one exit and an otherwise cheerful and positive crowd showing their grumpy, pushy side. After what could only been a day or two we were back on the road and heading home.      Zzzzzzzzzzzzzzzzzzzzz.  '),(2433,1401816,'02,August,2004','2004-08-02 00:00:00','   The dead or dying animal omen theme continues. I don\'t belive in signs though, so the dead mouse, dead baby bat and (most freakish of all) frog with half its face missing mean nothing to me.      I keep telling myself.  '),(2434,1401816,'02,August,2004','2004-08-02 00:00:00','   Today seemed to be the point in the marriage countdown where there guest list threatened to explode in a scary, out of control manner. Jenna and I agonised over this list, keeping it small but feeling pain at excluding so many people we love. However, \'recent developments\' have caused the list to be slightly tweezed. Luckily the size of the reception venue has helped us stick to our original plan without offending anyone. There will be at least one post hooneymoon gathering with food, drink and music. Emails will fly when we\'ve sorted that out.  Today I played music in front of a live audience for the first time in months. Great fun. It always is fun to stand in for someone else. You get to experience different arrangements of songs, different songs for that matter, new people and ideas too. Today was no exception. We played blues/rock/soul to a mixed audience, chilling in a field under a mostly blue sky. Two DJ tents, men playing drums, a barbecue with a veggy option and the sound of the Muppet Show album mixed with Zappa\'s \'Road Ladies\'. I only screwed up a few times (playing \'Brick\' a semitone out for a whole verse and thinking \"boy, someone is really out of tune\" untill the horrible truth dawned). The sleeping dogs in the title belonged to the man whose home we were playing at. They slept head to... er, tail during soundcheck then came for a wander about on stage during the otherwise intense climax of the second set. Everyone\'s a critic.  Home now, it\'s still light (wierd after a gig) about to have pizza and wine. No stress.  Current listening: One Man Guy - Loudon Wainwright III  '),(2435,1401816,'02,August,2004','2004-08-02 00:00:00','   Look at all the changes. Wedding lists have gone out, there isn\'t much left to do except buy clothes - oh, and find the money to buy them with. Luckily I have two gigs this week: one singing, one as a bassist.  I managed to get back into swimming today. It makes no sense but I already feel as if I\'ve been exercising for weeks, not scarfing low-level comfort food on an all too regular basis.  Next week I face the horrors of a job interview. This time I fully intend to ask plenty of questions to make sure it is the right job for me. Last time I was so pleased to get the job I forgot to check it made sense to take it. I met plenty of great people but ultimately couldn\'t keep on with it. Rash and stupid are the words that spring to mind when I think of my resigning without a secure job to go to, at 40, just before getting married. Brave is what my lovely friends all say.  And finally, the weather. Today has been one of those typically untypical British days. Cold and cloudy in the morning, with the radio reporting rain that wasn\'t falling here. I went to the local market town for more paper and diabetic chocolates (for a wedding guest). By the time I was home the sun was out and by four I was roasting. As I look out of the window, Jenna is catching up on sleep and a calm night is slowly descending on the village.  Current listening: Hairless Handle - Piore Z  '),(2436,1401816,'02,August,2004','2004-08-02 00:00:00','   Time seems to be speeding up. We had a long shopping trip to a major metropolitan area and returned with some chocolates and a Goodies DVD. So, the wedding clothes are not in our hands, the invitations are not out and the reception still isn\'t officially booked. Today we went to more local places and sorted out a variety of low level wedding stuff and one very important thing - the wedding ring. Or, to be more specific, Jenna\'s wedding ring. Mine is still safely sandwiched here on the desk between a Bach box set and a Kent College prospectus. This should also be the week that I get back to playing some bass in front of an audience. I\'m sitting in for a nice man called Carl as the bassist in his blues/rock/soul band. I am very keen to get out there and play.  Current listening: Surfacing - Orin Ambarchi & Martin Ng  '),(2437,1401816,'02,August,2004','2004-08-02 00:00:00','   The most amazing rainstorm today. I stood in the door of the outhouse with the kitchen CD player outclassed by what felt like rain with something to prove. I was hypnotised by it.  Also, sad news tonight that Sacha Distel had died. My mum was a fan more than I was, but his recent Desert Island Discs was entertaining. Like Johnny Haliday without the embarrassing wardrobe as someone once described him. Spookily, considering I came online to talk about the rain, his big hit here was \'rain drops keep falling on my head.\'  '),(2438,1401816,'02,August,2004','2004-08-02 00:00:00','   Thirty five years ago I was staying up late, watching telly with my trusty Action Man in his space suit by my side. I was just old enough to make out the blurry shapes on the screen and to know that history was being made. They could have sent a poet though. A great selection of pictures to browse through if you follow the link above.  '),(2439,1401816,'02,August,2004','2004-08-02 00:00:00','   Something I hadn\'t mentioned before was the music and readings for the ceremony. Or is it a service? Either way civil marriages have come on a bit since the old cash and carry days. We have been asked to provide suitable music and readings for the day. Suitable in this case means \'with no religious connotations whatsoever.\' This is harder than it sounds. My best man suggested \'Whole lot of love\' by Led Zeppelin. I\'ve been tempted by \'You know my name\' by the Beatles. Just a dream. On the way home today I heard a half hour radio programme about Bach\'s double violin concerto. First movement is all formal communication. Second movement: sexy intimacy. Third movement: Rude interruptions, apologies and talking over each other. Could there be a more perfect piece. When I finally got home I played it through and thought this is it. Then I went back to the Beatles.  Current listening: You know my name... - The Beatles.  '),(2440,1401816,'02,August,2004','2004-08-02 00:00:00','   Those of you sick of me going on about wedding preparations might find this amusing. Click on the title for the story. I can sense air crews the world over thinking \'it\'s about time.\'  '),(2441,1401816,'02,August,2004','2004-08-02 00:00:00','   Well thanks to flickr.com I can now share some photographs with you. These two were part of yesterday\'s gig. 		 		 		 			 urlLink     			 				 			 urlLink Two dogs 			 			Originally uploaded by 			 urlLink Roostar .  '),(2442,1401816,'02,August,2004','2004-08-02 00:00:00','   Had a great night out last night with David, the best man. The only thing we didn\'t sort out was what we were going to talk about in our speeches. He had a great idea that we should write each other\'s. This is certainly a less worrying concept than we cooked up the last time we met. Then there was a suggestion that our speeches should be written (and presented) as if we were preparing them for work. This would mean the best man\'s speech would be delivered as an account executive\' s presentation to potential clients and the groom\'s would be given as an English lesson. 	  Writing each other\'s could still be problematic, and anyway, we need to talk about content. Or, more importantly, what to leave out. 	   Last night was fun because we got to go out for Italian food and talk a lot. Then we went on to what his wife (they aren\'t married but were and look like they might be again so, as Jonathan Richmond once said \"I call her wife, because it stops all talk right away.\") calls our Nightclub. To me Borders is my second home. Once we\'ve walked the store for interesting books and magazines we settled in for coffee and chess. Three very, very bad games which I am in no way proud of. 	  They did help fuel my project to make a short film though. It\'s scripted already and I\'ll make a start after the honeymoon. If nothing else it should give me something to write about once the \'wedding month\' has passed and the new job has kicked in. It\'s going to be about things that remain as constants in your life. One of my constants is this ability to play chess at a frighteningly low level despite all my efforts to read books on improving my game and regularly playing David, the computer and strangers on the internet. 	   BTW, I listened closely to Bob Dylan\'s \'The Highlands\' from \'Time out of Mind\' the other day. I admit I was cleaning the bath at the time so may not have been concentrating as fully as I might, but all the same... what a spectacularly unfocussed and dull song. It gave me the impression that he really was making it up as he went along. Maybe I haven\'t got the hang of Dylan yet. 	  It pails into insignificance when compared to the almost satanic dreck that Borders were playing late last night. Maybe they were trying to get rid of their customers before they had to lock the doors. Maybe a member of staff had had a truly awful day and wanted to take it out on as many people as he or she could. The weapon of choice, for whatever reason, was an early David Bowie track called \'London Boy\'. There were several tracks from the same album actually, but they merged into one another. The song(s) sounded as if they came from the \'Laughing Gnome\' period but lacked the charm and cunning structure of that now almost forgotten \'hit\'. It also reminded me of why so many English people choose to sing as if they are Americans. Bowie\'s Anthony Newleyesque accent made me want to buy a sweetshop full of lollipops and fire them at the man until he stopped. 	   Current listening: Wild Cat - Joe Venuti &amp; Eddie Lang  '),(2443,1401816,'02,August,2004','2004-08-02 00:00:00','   I don\'t know how many of you are Mac people but, for those of you that are, see if this seems familiar. 	   http://www.misterbg.org/AppleProductCycle/  '),(2444,1401816,'02,August,2004','2004-08-02 00:00:00','   A long night of listening and discussing (polite euphemism) brought us to choosing a combination of Bach, Mussorgsky and Vaughan Williams for the wedding music. Reading supplied by Will Shakespeare and Percy Shelley. 	   Current listening: The most annoying car alarm in the world.  '),(2445,1401816,'02,August,2004','2004-08-02 00:00:00','   It\'s all getting worryingly close now. The big panic this week was when Jenna and I were talking about our two nights in a Gatwick hotel and suddenly questioned whether we were actually flying from Gatwick. It sounds silly but we\'ve been here before, two years ago. I had to make a quick phone call just before we caught a coach to a hotel we, technically, weren\'t booked into. So,as I said, panic.  While Jenna was away with her sister I had a night on my own. I rented a copy of \'Lost in Translation\' which I loved. Direction, cinematography, acting, settings, all came together to make me forget I was watching a film. I\'d put it right up there with Three Colours Blue (and White and Red) in the \'if someone held a gun to my head and, for a reason not made clear asked for my favourite films of all time\' list.  I am also pleased to report that both the bands I play in rehearsed this week and, even better than that, the \'new song\' has been finished and sent off to Radio 2\'s song-writing contest. This song was a combination of a tune Richard (guitarist, engineer and country gent) had written some time ago and a lyric I supplied. The two went together well but we spent nearly four hours working on the chorus, trying to get just the write combination and sound of words.  Also on a music theme, Jenna and I are getting very close to choosing the wedding music. this is good because it has to be selected and checked by the registrar by ... let me see ... this Saturday. I\'d better get back to that.  Current listening: In My Place - Bridge Street.  '),(2446,1401816,'02,August,2004','2004-08-02 00:00:00','   urlLink          urlLink BB      Originally uploaded by  urlLink Roostar .'),(2447,1401816,'02,August,2004','2004-08-02 00:00:00','   It\'s  urlLink true . What next? No Christmas.  '),(2448,1401816,'02,August,2004','2004-08-02 00:00:00','   What do stag nights mean to you? To me they mean trips to Dublin, waking up tied to lampposts, strippers, waaaaay too much alcohol and excuse for excessive sexism and then paintballing the next day. Very good friends, people I love and admire have done this but somehow gigs or previous appointments have kept me away from the paintballing. Shame. The last stag night I went on was great. I spent hours in a pub discussing Art Deco architecture (I was bluffing, having just been to the Art Deco exhibition at the V&A) and drinking, maybe, at a push two whole pints. The next day was paintballing but I had to be back in Oxford for two gigs. I\'ve been invited to the paintballing party that is Dave (trombonist and sound wizard) from Soul Beaver\'s but, hey, look at that! I\'ll be in Italy.  A theme of this wedding is the last minute and, dare I say, half arsed nature of some of the celebrating. I blame only myself for this but the reality is that tomorrow my \'stag do\' will take the form of meeting few friends to see a band, have a few drinks and then skulk home.  Jenna\'s hen night is yet to be sorted out.  Now, on a techy note - despite my strong allegiance to all things  urlLink Apple , I am terrifically impressed by the  urlLink Mozilla Firefox  browser, which allows me to make the blogging experience much less stressful and complicated. It\'s not as slick looking as  urlLink Safari  but seems more web friendly by several leagues.  Expect pictures from the stag night? Hmmm.  Current listening: End of the Game - Sting  '),(2449,1401816,'02,August,2004','2004-08-02 00:00:00','   Time is speeding up, that\'s for sure.  Had a wonderful rehearsal with  urlLink Soul Beaver  on Tuesday night. We played through new ideas for the set list and throughout the hot and heavy evening played some Led Zep, Joss Stone and who ever did the original version of \'Tainted Love.\'  Bridge Street (the other band) has completed its entry for the  urlLink Radio Two songwriting competition . So listen out for \'In My Place\' which we\'re all very proud of.  On the wedding front the pub have sent us a sample menu, which looks wonderful. This is what we\'re offered...  Crispy Duck Confit with Fig Compote & Crazy Salad Or Asparagus Vinaigrette with Shaved parmesan Or Brochette of Scallops with saffron & Tomato ~~~ Roasted Plum tomato Tatin with Melted Goats Cheese & Avocado Salsa Or Fillet of Scotch Beef with Rosti , Wild Mushrooms  & Two Sauces Or Pastry Puff of Newlyn Fishes with Tomato Butter Sauce ~~~ Terrine of Summer puddings with Clotted Cream Ice Cream Or A Fondant of Valrohna Chocolate with pistachio Ice Cream Or Selection of Farmhouse cheeses ~~~ Coffee with Champagne Truffle   Current listening: You\'ll see glimpses - Ian Dury and the Blockheads  '),(2450,1401816,'02,August,2004','2004-08-02 00:00:00','   urlLink BBC - Proms - Prom Events by Day  Great Prom last night. So much emotion, especially in the Tchaikovsky. I was knocked out by Gidon Kremer\'s encore after the Shostakovich Violin Concerto. It was, apparently, \'Aurora\' by Eugene Ysaye.  '),(2451,1401816,'02,August,2004','2004-08-02 00:00:00','   Three pints of Guiness, lame rapping and some energetic ska, not to mention the company of friends. That was the \'stag night\'. The lame rapping wasn\'t me by the way, it was the support to the Drug Squad, who played original ska music until the early hours. The best man wasn\'t feeling well so he missed out. I very nearly missed out too after a long day\'s shopping in Oxford (very hot, so many people). I wish I\'d had the camera with me during the day as town was buzzing with music, colour and every accent from around the world.  For anyone interested, I can\'t seem to make the \'Government Bans Smiling\' link work. The story was about new regulations on British Passport photographs. No more smiling, it seems. Have you known anyone who did smile in their passport picture though?  '),(2452,3155540,'30,April,2004','2004-04-30 00:00:00','    The Myth of TINYPANTS   Here\'s the story that Zaq and I created.  It\'s very strange, but that\'s the way we like it....  Sometimes we get bored.  *  = Zaq  ~  = Tara   *  In a village not far from the horizon, TinyPants was near death.  While working in his famous ketchup mine, he fell asleep and a starving donkey ate away his entire midsection.  Jo-junia county, in a sincere act of appreciation, put on a festival for TinyPants so that he may die with a smile on his face.  Unfortunately, TinyPants wouldn\'t be able to eat anything because his digestive system was in the belly of a burro.  ~ Tinypants mentioned his concerns of having no midsection to the nearest festival participant.  \"I have no stomach.\"  The small man looked at TinyPants, blinked and murmured, \"Yes.  It is true. Would you like the midsection of a donkey instead?\" TinyPants paused, and reflected on the thought. \"Yes, I think I would.  Please pass the donkey.\" And he started to munch on some donkey liver.  * What a strange circumstance!  After a donkey took TinyPants\' stomach, he would, in turn, take the belly of a donkey.  This was going to take some special grammar work.  \"But TinyPants,\" the town\'s cute girl murmured, \"How will you be able to get the operation done before your impending death?\"  ~  \"Oh my dear Towns-Cute-Girl,  Fret not!  The great God-of-stomachs-Eaten-By-Donkey\'s (SEBD) will save me!\" gleamed TinyPants while staring at the poor girls chest.  The girl looked at TinyPants with a very puzzled expression, and choked \"Have you gone mad? There is no such G...\"  Just then a gail of flatulent gases rushed through the air around them and a prepubescent sounding voice echoed around them.  \"I AM THE GREAT GOD-OF-STOMACHS-EATEN-BY-DONKEY\'S!  I AM HERE FOR....uh, what does this say? Tiny what?!  giggle  What an unfortunate name,\" he mutters under his breath, \"His parents must have really...  cough   I mean...I AM HERE FOR TINYPANTS!\"   *  \"And where is your membership card, Mr. Pants?\"  The God roared. \"Well, I haven\'t paid my dues in awhile...\"   \"Stupid bastard!\"    \"I\'ve got $4 in my sock.  How\'s that?\" \"Yeah, that about covers it.\"  ~ Tinypants gave the God-of-stomachs-eaten-by-donkeys his $4 and closed his eyes, (the pain was becoming unbearable at this point and Tinypants was having a hard time breathing.  Some doctors might have said that it was because he\'d been so long without his mid-section, but this is my story and those doctors can kiss my grits.)  The God of SEBD raised his hand, balled it up, and hit Tinypants squarely in the jaw knocking him completely unconscious.  The Towns-Cute-Girl gasped in horror at the scene, yet was mysterious attracted to the God of SEBD.  She\'d never before witnessed such power, not to mention he was immortal, which we all know is a huge turn on.  She started to blush and fiddle with her hair to try and get the God\'s attention.  The God being completely oblivious rubbed his knuckles and muttered \"Well, I guess that is that.\"  And he vanished in the same flatulent smelling gust of wind that he came in on. The-Towns-Cute-Girl dejectedly walked home.  She was so wrapped up in her rejection by a God that she forgot about TinyPants altogether.  TinyPants awoke to find his mid-section completely intact and a note pinned to his shirt.  It read \"the membership fee is actually $6, so I kept one of your kidneys as payment. Signed by the God of SEBD.\"  TinyPants slowly got up off the ground, and walked back to his ketchup mine.   '),(2453,3155540,'29,April,2004','2004-04-29 00:00:00','    Who the hell do I have to sleep with to get some hate mail around here?!   And where\'s Gretchen?  I was certain that she\'d make an appearance to trash my writing, or spelling, or something.  I think that I\'m going to have to start picking fights with my friends here on this blog just so I can encourage some nasty behavior.     Hey Kiddies , Check out  urlLink  this article   '),(2454,3155540,'28,April,2004','2004-04-28 00:00:00','    I\'ve decided that I need to develop some sort of rating system for each entry.  I am thinking something like \" Skipable \", \" Crazed Rant \", and \" Sheer Lunacy \".  That way a reader can monitor their own reading material, and we can give the FCC the finger.   \"#%@$ YOU, FCC!  And the horse you rode in on!\"    (Plus, I\'ll have to do a little censoring until I\'m absolutely certain that my mom is actually working at work, and not reading my bloggy blog.) I also think this rating system will allow me to be a bit more liberal with my writing, because lets face it, I have a lot to say but very little worth listening to.  So this should help alleviate both issues.  I also want to quickly inform everyone that my ability to spell correctly is going to drastically decline.  *SOB*  Because my bestest walking Spell Check friend, Fhay aka \"Scenecrasher\" is leaving for San Francisco this weekend!  Leaving me all alone in a uncaring world of vowels, consonants, and random punctuation.   '),(2455,3155540,'28,April,2004','2004-04-28 00:00:00','    Ok, here\'s a quick list of the things we\'d like to  NOT  hear Professor Selby say again.  Revisions and additions welcome. 1) \"This kid pee\'d on my head once....\" 2) \"Come on, Tim.  I\'m a gentle lover...\" 3) \"She @%&#ed better than My Uncle Tom.\" 4) \"B.J.\'s right to the top for her...\" 5) \"NO!  Please don\'t Hump me!\" 6) \"Um...Puppies taste good...\" 7) \"I was going to cook the turkey, but some bastard puked on the stove...\" 8) \"Oh, I have to pee.  Thats why I keep standing up.\" 9) \"I Luuv da cock!\" 10)\"My boyfriend ain\'t yo babay\'s daddy!\"  The list goes on, but the FCC is strictly monitoring this site.  Please help me with the stuff I\'ve forgotten.   '),(2456,3155540,'17,April,2004','2004-04-17 00:00:00','    I think I just caught the security guard here at work pilfering food from the freezer.  Granted, the poor guy should have some sort of perk to a lousy job.  (Very few people like to work the graveyard shift, and I can vouch for the fact that there is nothing pretty about 3:30am.) But I think that thereâ€™s a bit of an ethics problem with stealing peoples food from the workplace refrigerator.   I, myself, am not all too worried about it, but thatâ€™s probably because the food he was taking was some rank chicken wings that I dumped off on Professor Selby a few  months  ago.  Chances are that that poultry is going to make that security guard very ill.  I guess it was a risk he was willing to take.  Maybe later on Iâ€™ll traipse around this place listening for screams of agony but for now I\'m going to enjoy this semi chilly cup of coffee and watch  urlLink this  .   '),(2457,3155540,'16,April,2004','2004-04-16 00:00:00','    Oh, and you should always,...ALWAYS keep in mind that I view spelling to be completely optional.  I think that correct spelling really limits my creativity.  I\'m convinced that as long as my point is getting across, and the words  look   pretty, then I\'m doing very well.  And for some reason I think that  urlLink this  is funny...   '),(2458,3155540,'16,April,2004','2004-04-16 00:00:00','    Wow, who would have ever thought that I would step out of the shadows and into the spotlight?  I apparently decided that it was time to quit posting crazy rants on everyone else\'s blog and take credit for my actions.  Let me tell you in all confidence that it\'s not going to be pretty.  It\'s all down hill from here, kiddies.  (It\'ll only get worse when my ability to post pictures gains momentum.) So buckle up, and grip that yak-bag with all your might.   '),(2459,3155540,'28,May,2004','2004-05-28 00:00:00','    1)Dancing around to the Violent Femmes.   1a)\"Can I put in something like \'This is American music Take One\'?\" 2)Watching \"Two Guys and a Girl\" reruns   2a)any movie with Ryan Reynolds 3)Chocolate Milk.   3a)M&M\'s 4)Watching Jeb growl and twitch in his sleep like he\'s chasing rabbits.   4a)cuddly puppies and kittens  5)Sitting on the dock at Twin Lakes listening to the waves.   5a)sun on my back 6)Not fucking up one of my paintings half way through the process. 7)Going to really bad movies with my wussy brother, and laughing at how bad they are.   7a)listening to his crazy SN stories. 8)Having my expectations exceeded. 9)Paying less for something than I thought I would have to.   9a)getting stuff for free. 10)Cliff Jumping.   10a)...Ok, thats more scary than happy, but it\'s a great rush. 11)Fibbing to my Grandma about my cousin Chuck.   11a)happy family gatherings 12)Drawing witty sketches on the white boards at work. 13)British Comedians that dress like women. 14)Ben Folds albums. 15)Spider solitaire. 16)Waking up to sunshine on my face.   '),(2460,3155540,'27,May,2004','2004-05-27 00:00:00','    The air is biting cold, and as I walk down the path the pine trees all around me shuffle in the cold breeze.  I follow the path, winding just a few feet above the choppy December water.  I know where this path is going.  I\'ve been here hundreds of times before.  As I round a corner in the path, I see him sitting on the cliff overlooking the early morning grey sky.  I quickly sit down next to him, and simply stare.  I watch the wind sift through his chocolate brown hair and wonder how after 50 years he still thinks that a checked flannel jacket is the warmest garment he could wear.  I can\'t bring myself to speak, because I\'m afraid that I will wake up, and he won\'t be here anymore.  He continues to stare out at the lake.  Then as he raises his palm to the sky, one large perfectly formed snowflake falls into his hand.  I lift my own palm to the sky to cradle another fat flake.  He turns to me and smiles.  \"Happy Birthday, Princess.\"  As I lean my head on his shoulder, I say \"I miss you Dad.\"    The phone rings.  Damn.  I roll over in bed to look at the clock.  7:23am.  She\'s right on time.  Every year within a few minutes.  \"hellloo?\" \"Haaaaaa ppy Birthday TOOO YOOOU!\" \"Hi mom.\" \"Happy Birthday, Sweetheart.\" \"hee hee.  Happy birthday to me.  Am I grown up yet?\" \"I doubt it, honey.\" \"Hey.  You\'re my mom.  You\'re supposed to be nice to me on my birthday.\" \"...I love you, sweetheart.\" \"Love you. ....I miss him, Mom.\" \"I know, sweetheart.  We all do.\" \"So what are you doing today, honey?\"  I sit up in bed, \"Uh...I don\'t know yet....Holy Shit!\" \"What is it?\" \"It\'s snowing!  It never snows this early....thats really weird.\"   '),(2461,3155540,'26,May,2004','2004-05-26 00:00:00','    After my dad died I discovered that there is actually a limit to the things a person can deal with at any one time.  This was a bit of a surprise to me, and it took some time before I could accept it to be true.  So, I found that it was imperative for me to decide which things, events and people would receive my attention, and which would not.  I wasn\'t being hurtful or judgmental, I just didn\'t have the capacity to deal with everything.    The priorities fell into: grieving, being available for my family, putting the estate/possessions in order, getting enough sleep (very important), and school.    This kinda threw my friends and associates to the bottom of the list.  If a friendship can\'t easily exist without constant management and maintenance, then it was doomed to be ignored.  I just didn\'t have the ability to give a shit.  Several long-term friendships have bitten the dust over the last year and half.  Some for the better, and some were unfortunate casualties of changing times.  Tabitha, gone.  Dave, bye.  Michelle, don\'t let the door hit ya\' on your way out.  Siobhan, hope all goes well.  Abby, later gator.  Angie, get out.   (There\'s more, but I dont\' care enough to list them.)  The thing is, that I also have several friendships that have gone completely unphased.  These are my favorites, because there is nothing expected of either party.  Even if a long period of time passes between interactions, we just pick up where we left off.   Jenny, Drooler, Kinki-lee, Chris, Smelley, Emi, Ali-Cat, McFatass, Tina, Fhazzer, Chistina, Zaq, and of course Wallace.   '),(2462,3155540,'25,May,2004','2004-05-25 00:00:00','    For the most part, I really don\'t venture out of my preferred genre of music.  I like equal doses of Classic Rock, Indie, EMO, and Alternative(ish).  I\'m pretty open minded to closely related genres, but I just can never seem to get into country, or Rap/R&B.   After taking an example from Professor Selby, (I don\'t recommend you try it,) I\'ve spent the morning listening to the R&B/Rap Sampler on Rhapsody.  My preferences are as follows:  * = sucks ** = wouldn\'t play it in my car *** = can pretend it\'s not playing **** = would hum along ***** = will probably stay in the club playing it.   ARTISTS  (If you can call them that...) -Missy \"Misdemeanor\" Elliot :  ***  -Eryka Badu :  ***  -50 Cent :  ****  -Ludacris :  ***** + **   (so funny!) -Busta Rhymes :  ****  (WHOOHA!) -OutKast :  *****  -DMX :  *  -Eve :  ***  -Jay Z :  **  -Puffy Puffer-ness, or whatever:  Negative ****  -Jurassic 5 :  **  -Black Eyed Peas :  **  -Lil Kim :  ***  -Sean Paul :  ****  (hee hee. He talks funny.) -Nelly :  ****   All in all you\'d be hard pressed to find any of this in my CD collection.  But then again my brother is a little Frattie Brattie, and likes to burn me CD\'s that force you to turn your hat slightly skewed, and bob your head.    I feel so ashamed.  I\'m going for a hot shower.   '),(2463,3155540,'25,May,2004','2004-05-25 00:00:00','    Consider this a warning to all participants in the male gender.  I, Millty, being of sound mind and body do hereby solemly swear to kick the teeth in of the next man I encounter making a sexist, and or obscene sex statements.  Examples of these kinds of statements are as follows:  -\"I need a girlfriend so that I can have regular sex.\" -\"She\'s perfect in every way, but she\'s a little fat.  Do you think I can ask her to loose some weight?\" -\"Her friend wears tight clothes.  I bet she\'d be in for a threesome.\" -\"I really enjoy talking to you.  Can you get your ankles behind your head?\" -\"She\'s nice, but at the end of the day she\'s just another fat girl, now isn\'t she.\" -\"Nah, she is good on her knees.\" -\"If only she had a flat top head, no teeth and big ears, she\'d be perfect.\" -\"Yeah, she\'s got those great wide hips.  Bet she can pop out babies like no ones business.\" -\"You\'re great.  And I bet you\'ve got a mouth like a hoover.\" -\"Its all about the punani.\"  (which is the stupidest phrase I\'ve ever heard.)  I understand that I have no power, and no right to ask this walking wastes of skin to not say these things.  I have no power to change the way they think.  But I do have the power to eliminate myself from the situation.  So, If you can see yourself making comments similar to, but not exclusive to the phrases above, then understand fully why it is I look at you as if you\'re lower than pig shit. And if you could just be so kind as to leave me the fuck alone.   '),(2464,3155540,'23,May,2004','2004-05-23 00:00:00','    Dear Tara, I know what you\'re thinking.  I know because I\'ve been riding shotgun to your conscience, and have been watching through your eyes for some time.  A unique team we are.  And I want to tell you that I know. I know that sometimes its too deep to objectively get out of.  That self-evaluation that is real and uncompromising.   There are no foofy song birds to tell you it will all work out perfectly.   There is no ray of sunshine to split through the murky waters of your mood.   The truth of the matter is that you are not important.   There is no one who needs you.   There is no good that you provide that the world will miss.  For the most part you are just another competitor for some breathable air and livable space.    You can wake up every morning and tell youself that it\'s important you get out of bed and shower.  It\'s important you make that cup of coffee.  Its important that you drive into work to produce something that no one sees and no one cares about.  You can sit at a computer for 10 hours a day convincing yourself that you need to be there.  That what you are doing is important.  Its not.   The only person who benefits from your day to day existence is you.  The only person who finds you important is you.   Your neighbors won\'t notice if you don\'t wake up one morning. Your boss will just fire you and find some other stupid shit to do your mindless work. You rarely interact with your friends, so they won\'t notice if you don\'t answer your phone.  Once this realization takes hold it is damn hard to shake. I know. But it is imperative that you do. Because without this meaningless day to day life, there will be No. More. YOU.  And that\'s important.  TO YOU.  And now you know.   '),(2465,3155540,'22,May,2004','2004-05-22 00:00:00','    You know that little voice in the back of your mind that tells you  \"It is detrimental to our wellbeing that you stop talking about this topic.\"   Or at more important times it says  \"SHUT UP, YOU RAVING IDIOT!\"  Yeah.  Well I don\'t have one of those.  I\'ve been meaning to purchase one, but they never go on sale.  All I\'ve got to say for myself is \" Umm, my feet taste goooood .\"   '),(2466,3155540,'21,May,2004','2004-05-21 00:00:00','    -the security guards here at work are  all  creepy.  One followed me around the building this morning (3am) to say \"hi\"  -i wish my blender would arrive in the mail today -i think the people of redmond are the left over in-breds of Seattle -i have somehow created a very secluded life with minimal interactions outside of work -i am convinced that my work schedule is killing me slowly -i like \'the vines\'.  Sorry.  Broke Matts rules on that one, and Fhay\'s going to beat me -i wish i owned a time machine, because there are events in my life which i wish had gone very differently -i have to pee, because i\'ve drank much too much coffee -my little brother is one of the most amazing people i know.  Be sure not to tell him -somehow, i think i own the softest most cuddly beagle in the world -i can\'t eat peas.  they remind me of rabbit poop. -i hope that \'arpel arpie\' has her baby today -i hope to own my own home by january 2005 -i want my \'68 firebird to be the most beautiful tribute to my dad that it possibly can be -i might be half asleep right now -i\'ll hack all of matt\'s shit if he threatens to rub himself on anything in the office again.  i\'ve got your passwords, you little snot.  it\'s war -i think i need to go see a doctor.  and a dentist for sure -i actually like this job, but i like to complain about it even more   '),(2467,3155540,'20,May,2004','2004-05-20 00:00:00','    After an excruciating conversation with the notorious \"he-man\" last night, I discovered that he has a soft spot for Guns n Roses, (which makes him one handlebar mustache away from living in a trailer)  thats for the \'creed\' comment, asshole!   and has never heard any Led Zepplin songs other than what I\'ve forced him to listen to.  His comment was \"Led Zepplin is folky crap.\"    Excuse me?!  What did you say?!  So in the name of all things holy... i.e.  Rock and Roll  ! I\'ve listed a few musicians that I used to like, but now wish would just wither up and die.   DIE!   1)  REM.   Ohmygod,go away    2)  Everclear.  Who gives a shit about volvo driving soccer moms.   3)  Smashing Pumpkins  Tell that whiney Billy kid to SHUT UP!   4)  Blink 182  All Pop Punk can go get F*#&ed!   5)  The White Stripes  Take a shower, brush your hair and call me when you have a new catchy piece of shit.   unfortunately, The list of musicians that still make me dance around my livingroom like a dork, but am too ashamed to mention is much longer.  (I.E. Matchbox 20) I shouldn\'t have told you that.   The CD\'s in constant rotation seem to be a mix of \'sublime\', \'incubus\', \'fuel\', \'ben harper\', \'jack johnson\', and \'radiohead\'.  I find that these are always good for me.   '),(2468,3155540,'19,May,2004','2004-05-19 00:00:00','    I am a hypocrite.  A huge hypocrite.  A ginormous (for lack of a real word) hypocrite.  Everytime I\'ve seriously dated someone I change, or morph a part of my personality to better suit the relationship.  My track record is as such:    hypocrisy 1  = I attended church, youth groups, listened to Christian rock. (Not as much by choice as that\'s just what was continually playing.) I even sat in a few little prayer circles with my little eyes closed, looking oh-so-intent, holding hands with other praying kiddies.  (You can be guaranteed that I was most certainly NOT praying.)  I was usually thinking about what I needed to pick up from the store on my way home.  The boy I was dating at this time was overly religious in a \'hyper-christian\' sorta way.  And I guess because I was interested in him that somewhere in my mind it was important that I pretended to be interested in this.  You all should know that I am not now, nor have I ever been religious.  The point is that I would go to school, smoke, swear, cheat, and generally just be a snotty little kid.  Then I would visit this boy and become a saint.  All of the foul lingo erased, and the cigarettes hidden.     hypocrisy 2  = I attended motorcross.  Need I say more?  I mean, who really is interested in motorcross?!  Honestly! It\'s certainly not me.  But, later I pretended to be, in a milder version of hypocrisy 1. (I am much too ashamed to go into detail.)   hypocrisy 3  = I became super interested in money, and looking it.  It was perfect hair.  Perfect makeup.  Perfect skirts, heels, nails, pressed blouses.  Those of you who know me must be thinking  \"What?!  I\'ve never seen Tara in a skirt. When was this?!\"   It sounds like a fabrication, but it\'s true.  If left uninfluenced I revert back to my natural state of soft t-shirts, flip flops, and loose jeans.  Make up is something I don\'t have time for.  It\'s been one new guy with one \"new hobby\" after another.  And as of this point I just have a great big list of things that suck.  BLAH!      How depressing.   But entirely True.   '),(2469,3155540,'15,May,2004','2004-05-15 00:00:00','    Remember those stupid posters and \'feel good\' books that had such titles as  \"Things I learned in kindergarten.\"   Or even more ridiculous  \"Things I learned from my cat.\"   These half-witted writings included such sayings as \'be sure to share\', or \'don\'t lick yourself in front of guests,\' etc. etc.  As ludicrous as these stupid things were, I\'m too brain-dead to think of something clever.  So, in the same theme, I give you:   THINGS I LEARNED FROM \'SOGGY BOOTS\'   ~If you wear big cowboy boots to a bar with the intention to smuggle a bottle of beer out, be sure that the bottle mouth is not facing down, thus filling your shoe with beer.  ~If cutting down large pine tree\'s for firewood, be sure that the trees path to the ground is not obscured by your truck.  ~Flannel, (apparently,) goes with everything.  ~There was nothing good about disco.  ~If you need your kids to wake up and help you with the yard work, 115 decibels of  Iron Butterfly  works really well.  (FYI, I believe the ear drum bursts at 140 decibels.)  ~Surprising a sleeping dog is funny.  Surprising a sleeping cat is painful.  ~It is not only your responsibility, but also your DUTY to harass your siblings.  It builds character, and who are you to deny them character building moments?  (You can be sure that they will return the favor.)  ~Rolling a \'72 Blazer is actually quite easy.  Telling your parents about it is not.  ~Try not to brag to your kids about how you out-ran a cop in your old \'68 firebird.  (You\'ll never win another argument from that point on.)  ~Always sound like you know what you\'re talking about, especially if you don\'t have a clue.  ~If you\'re working on a car of foreign make, it helps to swear at it in its native language.    ~Reruns of \'Startrek\' and old western are always on TV.  (Unfortunately.)  ~If you\'ve burnt the nights meal, you can impress your guests by telling them that it is \'Cajun\'.  ~One of your biggest responsibilities is to watch out for your family.  Especially if they tend to be lunatics.  Hey, Wussy & Chumpzilla.  What did I forget?   '),(2470,3155540,'14,May,2004','2004-05-14 00:00:00','    Every so often I run into some momentary tid bit of disturbing information, or a slightly disturbing encounter.  Am I simply disturbed? You bet your sweet ass I am.   But this just happened:  Get up.  Walk to bathroom.  Quietly open door.    What the hell was that?!  Did that woman just grunt?  In the bathroom?! Oh my god.    Silence quickly ensues, and the embarrassment is as obvious as a 30ft neon sign flashing over the stall saying \"Guess what I\'m doing in here!\" Hold breath. Turn on heel.  Walk away as fast as possible.   Yeah, that\'s right.  There\'s a grunter here at work, and this time it wasn\'t Fhay.  Unfortunately, this isn\'t the first time I\'ve been unlucky enough to be present for \'the grunters\' demonstration. Blech.  I wish there was whiskey in the coffee.  Instead,I\'ll just sit back and watch  urlLink this   '),(2471,3155540,'13,May,2004','2004-05-13 00:00:00','    I think that we\'ve exceeded the maximum capacity for comments under one post, so I\'ll create another baby post for today.  But I\'ve had two beers, so don\'t expect anything witty.    The greatest thing I\'ve ever heard a co-worker say:  \" Jerry Cantrell got me hooked on dope .\"  discuss   '),(2472,3155540,'12,May,2004','2004-05-12 00:00:00','    Well, the votes are in.  First I\'d like to thank all of the parties involved for their expertise and their input.  (And will ya\' just look at the number of comments!)  There were a lot of good points provided for us to ponder...well except the crap about the deserted island.  That was bogus.  But all in all points well made. Secondly I\'d like to just quickly point out that the club He-Man was trying to get into was probably a Chuckie Cheese, where a kid can be a kid and a creepy old guy can leer at them in public.  After yesterday\'s discussion the conclusion I\'ve come to is thus:  Q:    Is a platonic friendship possible between two people of the opposite sex?   A:   Yes.  It\'s entirely possible when neither party is attracted to the other.  I\'m convinced that pheromones play a huge role in attracting people.  If there isn\'t that attraction, then both parties are free to grunt, slap each others asses and/or go shopping together for matching pink sockies.  It can and does happen, but like I said, it all comes down to pheromones.  Q:    Does a purely platonic friendship between two people of the opposite sex happen very often?   A:   Not from what I can tell.  It seems that a large portion of the time one person has an interest in the other that, perhaps isn\'t reciprocated.  I spent the last several seconds dissecting my own seemingly platonic friendships and what I\'ve come up with is this: 2 = admitted they had had crushes on me at some point. 1 = told me that a platonic friendship was flat out impossible between two people.  Which was a little awkward at that time, but the good news is I\'m really talented when it comes to acting ignorant. 2 = still maintain that we\'re just platonic friends, and always have been.  And I would use this for the basis for my argument if it weren\'t for those silly conversations they initiated: \"Do you think that two people who are really great friends have the potential to be really great in a relationship as well?\"  Also awkward.  (If you\'re wondering, my answer was \'no\'.) 3 = I had itty-bitty crushes on at some point, (which quickly wore off, but whose to decide the best way to meet a friend?!)   Q:    Is a platonic friendship less likely if the two people in question are attractive people?   A:   I think that only a guy would assume that to be the case.  Being very visual creatures you guys tend to base all of your judgments on what can be seen.  I can say with certainty, that most of my guy friends are  very  attractive guys, and get a lot of girly attention...Just not from me.  Some should even be models. That doesn\'t mean that I\'ve been attracted to them.  (For the most part, I\'d say that I am pretty oblivious to what\'s going on around me.)   The important thing to note here is that regardless of underlying issues and motivations, guys and gals DO, in fact have friendships without romance.  Happens all the time, Folks.   NOW, the real question is:   Can two people who dated seriously, become platonic friends afterwards?    Tara says:  Probably not.  But we\'ll see.   '),(2473,3155540,'12,May,2004','2004-05-12 00:00:00','    Question:  Can Girls and Guys be completely platonic friends?     If you\'re the only person alive who doesn\'t know what I mean, then I\'ll define \"platonic\" for you, and maybe I\'ll even learn a thing or two.   1 capitalized : of, relating to, or characteristic of Plato or Platonism 2 a : relating to or based on platonic love; also : experiencing or professing platonic love b : of, relating to, or being a relationship marked by the absence of romance or sex.  Did ya\' get that?  That means no romance between one guy and one girl who are friends, and actually spend time together.  (For the sake of my argument, both are heterosexual.)  I\'ve always maintained that platonic friendship is not only possible, but the common state of being for two individuals of the opposite sex.  I say this because I have always had close guy friendships that have never been romantic.  The reason I bring the topic up is that I\'ve recently been accused of being a bit naive on the subject.  My brother maintains that (generally) a guy wouldn\'t just be friends with a girl unless there was romantic interest or potential.    What are your thoughts?   And what the hell happened to Blogger?!  I spend two days away from the pc, sitting in front of my tv watching \"the office\" and they go changin\' the whole damn thing!  Not to mention someone has taken off with my comments section!   '),(2474,3155540,'08,May,2004','2004-05-08 00:00:00','    In a rare occurrence I felt a pang of guilt yesterday.  It transpired after I used gorilla warfare tactics to get more of my friends to pay attention to me.  The pang was only momentary though and I realized that it was a Brilliant move!  And conducted with great success!  As I mentioned, I like to refer to this as \"developing my audience.\"  I\'ll have to be sure to do it again, but maybe next time with different people.  Perhaps my old ex\'s?  Or even Selby\'s friends?  I may have to wait a bit to harass the masses again.  Mostly because I think the village people would become restless, and chase me with torches and pitch forks...  Again.   But to kill the time between mockings I decided to go to  urlLink thespark.com  and check to see how dateable I am.  I wanted to see how much of a Wench I tend to be, but they wanted a blood sample, and I never donate before 9am.  But as it turns out, I\'m  43% Dateable.  No big surprise there.  I discovered that I\'m attracted to prawns, and the best place to meet people is \"at the airport, going the wrong way on the escalator\".  Also no big surprise.  I highly recommend that you check out the link on my sidebar labeled \'quizipoo\'s\' and find out more about yourself than endurable.   '),(2475,3155540,'07,May,2004','2004-05-07 00:00:00','    Here is the official posting of friends,â€¦. Ahem, I mean â€˜minionsâ€™ that think its funny to visit my site for a free chuckle and not make a single damn comment.  These Bastards are about to be appropriately punished.  The list is as follows:   The names and faces have been changed slightly in order to protectâ€¦well, me.      1. Wussy  is a frattie brattie at Sigma Nu best known for his expertise in drinking games, catchy T-shirt slogans, and how to spend no less than $3000 to NOT fix his Jeep.  Not to mention he is extremely skilled in pissing off his older (and much more attractive) sister.  2. Chuck  is a slackass who eats too much soy and granola, being the hippie lover that he is.  Spends most of his time not progressing on his schoolwork, shaving his legs, and trying not to be mowed down in traffic while on his bike.  3. McChunky  is a closeted exhibitionist whoâ€™s favorite past time is flashing his ass at unsuspecting motorists, and newly acquired associates.  Also enjoys rock climbing, but not after alcoholic beverages.  4. Emi  likes to imitate a cute koala bear by sitting high up in trees, and eating eucalyptus leaves.  Oh, and she makes beer.  Ok, Ok, so I canâ€™t find anything demeaning about Emi, but give me time.   5. Ali-Cat  is stuck in the middle of $#&ing Nowhere claiming to study, (which we all know is bull$hit,) and mistreating her poor cell phone.  Iâ€™ve heard she still carries the mobile device around even though its service has been turned off.  6. Gas  is a computer geek, who youâ€™d THINK would visit this damn site once in awhile, but apparently is too special to bother.  What a wanker.  7. Zaq  is the funniest bugger Iâ€™ve known. (sorry selby,)  Now that that has been established, you should know to never ask him what it is he does for a living.  Its top secret.  He could have told me, but I didnâ€™t want to die.  8. Whipple  is a really nice guy who will probably scream at the top of his voice â€œI am NOT a nice guy!  Iâ€™m an a$$hole!  Why canâ€™t you people see that?!â€  9. Mundo  is a newly established transvestite.  Lets just say he watched too much Eddie Izzard, and decided that stiletto heals look great on guys.  Too bad he makes such an ugly woman.  10. Drooler  is trouble, and itâ€™s just better that he not visit this site.  As a matter of fact I may have to keep him quarantined from the rest of the minions.  Heâ€™ll pervert their little minds with a simple look.  Consider yourselves punished! (oh, and Crawford, Smelley, Jerome-Rome, Hosers, and Kinki don\'t think that you\'re going to get out of being reprimanded.  I have to have something to do tomorrow.)  Now I\'m going to shut up, and watch  urlLink this   '),(2476,3155540,'05,May,2004','2004-05-05 00:00:00','    17 posts?  Wow, that\'s a new record for a bloggy blog that isn\'t even slightly political, religious, or even interesting.  This isn\'t even a college blog where I would share all the answers to our most hated professor\'s final, although I kinda\' wish it was.  I\'m always up for something deviously cunning.  But! back to my point. I hope you know what this means, kiddies.  This makes you all my little minions.  (The matching minion uniforms have already been ordered.)  I\'m very excited about this.  I\'ve always wanted minions to do my dirty work.  Well not all of my dirty work, but most of it.  I certainly wouldn\'t think of making one of my wonderful minions clean Waffles fish bowl (see comments for \"waffles\" reference.)  But I might encourage one of you to shop for my Mom on mothers day.  (Psst!  She likes anything that smells like Plumeria flowers.)  Here is some good news.  We can now discuss openly what we\'ll get Mom now that she\'s assured me that she won\'t be tuning into the Playground anymore.  I believe her exact words were, \"It was kinda cute, but I got confused when everyone started talking back and forth to each other in some sort of under 30 kid-code.\"  So it\'s now safe, my little minions.  Go about your minioning!    Stay tuned for my next chore.  I plan to locate the Minions that have gone AWOL, and rip them apart.   '),(2477,3155540,'05,May,2004','2004-05-05 00:00:00','    After spending the weekend at my Alma Matter I have decided to do an experiment on whether attending classes actually makes a student dumber.  I am fairly certain that there is positive proof that this is the case, but further research is in order.  I\'ll need a heart monitor, six pints of Mac & Jacks, and a set of jumper cables.  But first I\'d like to introduce my three test subjects.   Subject #1 ,    \"Chuckie\",    spends hours upon hours at his work station finishing homework, trying to simply pass the classes required for his degree in Forestry Management,...or some ridiculous  major like that. Claim: \"I swear I must have A.D.D., or something.  I can\'t seem to focus enough to get my shit done.\"   Subject: #2 ,   \"Kinky-Lee\",  Works two jobs, spends nights hammering out homework so that she can graduate this summer with her degree in Journalism/Radio broadcasting. Claim: \"I must have A.D.D.  I just hate sitting here and working on this crap.\"   Subject #3 ,   \" Wussy \", parties like its\' 1999, furiously uses few remaining brain cells to further degree in Mechanical engineering. Claim: \"I think I have A.D.D. or something.  I seriously can\'t remember shit.  Damn.  School sucks.\"  Let us notice that all three subjects seem to be convinced that the fault lies within themselves, and are certain that they suffer from a very serious disorder such as Attention Deficit Disorder because they have no interest in working on homework.  (Also notice that NONE of the test subjects mentioned that their regular consumption of alcoholic beverages, and erratic sleep patterns may be effecting their work performance.)    But, I ask you, is the fault really theirs?  Or is it a by-product of an overzealous educational system which overloads each subject with ridiculous mind numbing amount of useless busywork?     Discuss amongst yourselves...  Thus far the evidence is inconclusive, and such research demands more time, but as we all know I don\'t have the attention span for that kind of crap.  Now, where the hell did I set my beer?  I should also mention that my previous experiment has produced very interesting results.  Apparently, plants actually  don\'t     like a strict diet of cherry coke, and will wither up and die.   '),(2478,3155540,'03,May,2004','2004-05-03 00:00:00','    I\'m visiting my friends and family this weekend in Idaho, and this morning I was pleasantly surprised to hear that the state of my well being was being discussed on the local radio station.  This would mostly be because my good friend Kari works on the morning show for 104.3 The Border in Pullman.  I feel like I\'m famous!  hee hee!  So other than dealing with the idiocy that is my family, I am having a great weekend.  Who would have guessed?  I was a little disappointed to see flames shoot out from under my firebirds hood, but it\'s always good to get your pulse racing everyonce in awhile.   Speaking of racing pulses, has anyone heard out of the Scenecrasher kid?  I\'m dyeing to know how the move went.   By the way I\'ve added links to the sidebar for my pictures from Hawaii.  Check \'em out.   '),(2479,3155540,'01,May,2004','2004-05-01 00:00:00','    Itâ€™s always there.  I can feel it lurk around the fringes of my consciousness.  Iâ€™m completely aware of It and can, for the most part ignore It.  But some days, like today, It gets enough courage to sneak up and slap my face.  Iâ€™m powerless when it breaks loose.  Sometimes Its escape hatch is triggered by a song, a smell of cologne, or perhaps a picture.  But one location sets it free every God Damn Time.  I hate this particular place and try to stay away from it as much as I can.  I wouldnâ€™t admit it openly, but I pretty much moved to a completely different state so that I wouldnâ€™t ever have to be forced to drive by of this location.  Today I will drive by this location.  Nothing I can do about it, I just have to.  And I can tell you exactly what Iâ€™ll do.  Three blocks before, I will change the CD in my car to something obnoxious, and I will turn that bastard stereo up as loud as I can tolerate with minimal bleeding from my ears.  Then Iâ€™ll pound on the steering wheel in time to whatever hateful, screaming song is playing.  Iâ€™ll race around other cars, and be horridly reckless in general until I am safely past this place.  The place is like a magnet, sometimes It whispers in my ear and forces me against my better judgment to pull into the parking lot.  Usually armed with flowers, and kleenex.  The first time I ever went there was the only reason to never go back.  I wasnâ€™t alone.  I was supported by my brother Bret on one side, who would have looked oh-so-dashing in his new suit if it hadnâ€™t been for the look of pain and sorrow on his face, and my mother, wearing a similar expression.  As a matter of fact, everyone was wearing that damn expression.  Youâ€™d have thought it was on sale that day, because I know mine was a perfect mirror image of theirs.  Everyone was there wearing that look.  Cousins: Charlie, Keith, Justin, and Nathan.  Uncle Pat, and Aunt Kathleen.  Grandmaâ€™s expression was a bit more intense, but Grandpa made up for it by being a little more wooden faced and staring at the sky.  We were all there holding each other up, because it seemed that the person, whose presence normally would have allowed us all to be strong on our own, was gone.  â€¦Well, technically he was there.  But we could say that he was the only one not quite as vertical.  Howâ€™s that.  I hate the way this damn place always smells like moss and dirt, because there is always fresh dirt around.  I really hate that the first time I went I would sit so quietly.  Listening.  For any little sound.  But they donâ€™t put bells in pine boxes anymore.  Listening to something other than my own whimpering.    So, today Iâ€™ll go.  Iâ€™ll bring the most unusual looking flowers that I can find at the near by grocery store, and youâ€™ll find me kneeling in the grass.  Iâ€™ll be the one crying.  Youâ€™ll know itâ€™s me, because Iâ€™ll be the one whispering, â€œI miss you, Dad.â€ And Bret will eventually make me walk back to the car.  I know this, because weâ€™ve danced to this tune before.  In case you were wondering, this morning It was set free by James Taylorâ€™s â€œFire and Rainâ€, (and perhaps a good dose of anticipation towards this afternoon.)  That song is a guaranteed ticket out for It.  So is the smell of Old Spice aftershave.    It has several names.  â€œSorrowâ€. â€œGrievingâ€.  â€œLossâ€.  â€œLonelinessâ€.   But I call it â€œLeave me be!â€ and that usually works.   '),(2480,3155540,'29,June,2004','2004-06-29 00:00:00','    And, Yes they were.  Yowza!  They were definitely a sight to behold.  This weekend, the annual Iron Man Triathalon took place in Coeur d\' Alene, Idaho where I spent my weekend.   These poor bastards start with a 2.4 mile swim in the frigid waters of Lake Coeur d\' Alene.  They then emerge looking like grape smugglers just in time to have some oh-so-lucky volunteer strip their wet suit off and dress them for the  112 MILE  biking portion. Followed closely by a  26.2 mile  marathon!   Impressed?  Well, you will be once I mention that the winner does all of the above mentioned forms of torture in  8 hours, 42 minutes, and 56 seconds .  Jesus!  I spent my weekend trolling around CDA just admiring the scenery.  And by \'scenery\', I mean all of the hot men.  But you probably knew that.   '),(2481,3155540,'25,June,2004','2004-06-25 00:00:00','    Oh, kiddies I wish I could say it was over, but alas.  This is not to be the case.  I feel that I interviewed very well today, but my very last interview had to be postponed until Wednesday.    Poop.  I\'m going to keep holding my breath, though.  I hope that I don\'t kill off too many brain cells from lack of oxygen.   '),(2482,3155540,'24,June,2004','2004-06-24 00:00:00','    I\'ve gone way past \'nervous\' this morning, and now have a comfortable seat in \'twitchy\'.  My interviews are tomorrow, and my stomach is performing its own interpretive dance to the tune of seven cups of coffee.      I say \'interviews\' because there is more than one.   There are six ! But to be fair with myself, I have been offered every job that I have interviewed for.  I\'ve turned some down, however.  You\'d think that would comfort me a bit, wouldn\'t you.  But no.  Not even the knowledge that I\'m pretty much applying for a job that I\'ve been doing for the last year soothes my nerves.  I need a beer.  What\'s the rule for drinking, Bret?  \"Not before 5?\"  That\'s A.M., right?   '),(2483,3155540,'23,June,2004','2004-06-23 00:00:00','    It was the shriek that woke me up.  The woman next door screamed at the top of her lungs, and then continued to yell in a tone more suitable for a WWF arena.  I could hear her scream the same phrase over and over, but it was indecipherable.  I couldn\'t figure out what language she was speaking.  In a lower, calmer tone I could hear her husband/boyfriend/whoever the hell he is, saying something as well.    I rolled over to check out the clock.  Great.  A lovers dispute just three hours before I have to get up.  Thanks guys.  The woman is frantic, but not crying.  Not yet anyway.  She\'s shrieking, but try as I might, I cannot figure out what language she\'s speaking.  From the sounds of it, he\'s trying to console her, but it\'s certainly not working.  I remember thinking that he\'s got a nice tone to his voice, but it sounds like he\'s just making the situation worse.  After 15min of staring at the clock and listening to the dispute there is a loud crash against the wall near my bed.  Whatever it was it was big, and now it is very broken.  There is no way these people are getting their damage deposit back.  The woman starts to scream again.  Not saying anything, just screaming.  This scares Jeb awake, and he howls.  The howl must have startled the fighters, because they are silent for a few brief moments.  Then there is a thud, like someone being hit, and an even louder thud against the wall which was followed by a sob.  Was someone just slammed against the wall?  Now I\'m fully awake, and my heart is pounding.  Shit.  They\'re beating each other over there.    The door slams, shaking the pictures on my wall, and an engine starts outside.  I am to assume that she took off in his car, because he then stood in the parking lot yelling and chasing after it.  The funny part is that she tagged a cement post as she was backing out of the parking space, and it sounds as though she gauged the side of his late 80\'s BMW.  All of this took about 45 min to play itself out.  Unfortunately, it was just enough to fully wake me up.  And I spend the next 2 hours thinking about all of the fights I\'ve been in, and how I never want to experience one like that I\'ve just heard.  Then I thought of all the possible scenario\'s that would cause a couple to treat each other that way.  Did he cheat on her?  I\'d be surprised.  He\'s nothing special, but it\'s probably most likely.  Did she cheat on him?  Possibly, but I can\'t imagine him staying that calm.  I\'m betting it was something much more mundane, and stupid.  That seems to be the way it always works out.  Either way, I doubt that there is any amount of make-up sex that will fix this conflict.   '),(2484,3155540,'22,June,2004','2004-06-22 00:00:00','    Well, since none of you skunksucking bastards want to hear about the random shit that I was going to share, then you\'re going to be forced to listen to whatever the hell I feel like talking about this morning.   The good news is that I want to talk about my breasts. The bad news is, they\'re smaller. I realize that that still doesn\'t put me in the \'average chest size\' zone, and a significantly smaller chest is exactly what I was trying to achieve when I started getting serious about training. But this is kinda freakin me out. I figured that with the genetics that I suffer through that my chest would be the last area to get smaller. When in reality it\'s among the first. (The ass is being stubborn, but the ankles look great.) I know, some of you girls are saying \"Ooh, poor Tara. Whaaa. Do you have to suffer with a D-cup instead of your normal DD? Poor little baby.\" But it\'s a scary thought to think of having no chest and a big ass. Things have to stay balanced, here. It was fine when it was big ass, big chest. But now I\'m getting worried. I\'m going to look like a female verstion of the sloth off of \'The Ice Age\'....only with less hair and better teeth. Same lisp, though.   Oh, well. I guess I could always become a contestant on \'The Swan\' and maybe they\'d fix \'em for me. That\'s pretty ironic, though. That\'s not exactly the procedure I had envisioned myself needing...ever.   Good thing Victoria\'s Secrets caters more to the smaller of chests, and not so much the larger.   LONG LIVE THE PUSH-UP BRA!  ~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~   '),(2485,3155540,'21,June,2004','2004-06-21 00:00:00','    What\'s the opposite of \'writers block\'? Is it \'writers inspiration\'?  Professor Selby, being the wordsmith that he is, offers the term \'cranial diarrhea\'.  (He also really liked the term \'skunksucking bastard\', but I\'m not sure how to put that into context so it\'ll just have to be happy as the title.)  I can\'t decide what to talk about.  So, I figured that I\'d list the possibilities and see what people are interested in hearing.  I\'m also open to suggestions.  1)The girl standing in line at lunch with a spider crawling up her ear. 2)Getting MY car fixed for FREE!  You heard it kids!  Being a cutie helps you to get your car fixed for FREE. 3)Seeing HIM again. (Its a gamble for the mystery, but I won\'t elaborate today.) 4)The importance of having missed \'Car d\'Lane\', with its largest number of late \'60\'s firebirds ever.   5)My plans for the rest of the summer if I don\'t get this job. 6)The white hairs on Jebs face as indicators of his mortality. 7)The beat down that Matt will receive if he threatens to rub himself on anything in the office again. 8)How a person could never meet the love of their life while drinking a PiÃ±a Colada. 9)The hazards of taking the person you\'ve been dating to a friends wedding. 10)And if no one wants to hear about any of this random shit, then I\'ll be more than happy to talk about my lack of sex life.  Cast your vote, or piss off.   '),(2486,3155540,'20,June,2004','2004-06-20 00:00:00','    It\'s Fathers Day.  Go do something fun with your dad.  I\'m going to have a rootbeer float in honor of mine.  I\'m not super fond of rootbeer floats, but he was.  And some things just have to be done a certain way.  It\'s tradition.   '),(2487,3155540,'19,June,2004','2004-06-19 00:00:00','    \"See the lacing around the top of the glass?\" \"Yup.\" \"That is the hydrophobic....( beer related micro biology information ensues )...and that\'s why certain beers have more foam than others.  Hey.  Tara! Are you listening to me?\" \"Huh? Oh, yeah.  You\'re talking about the hydro-phillies playing in the foam.  But I can\'t see how sea horses ever come into contact with beer in their natural environment.\"  *THE LOOK*  \"Come on, Emi.  That was funny.  Admit it.\" \"Dammit Tara!\"  \"I thought it was funny.  Look, Emi, you can\'t possibly expect me to retain that information.   I am a VIS. COMM. major .  I make things look  pretty .  That\'s what I do.  I have only the most passing interest in micro biology.\" \"Oh, no.  I don\'t expect you to retain any of this, as a matter of fact I hope you don\'t, because then I get to look forward to telling you all about it again.  And again.  And again.\" \"Shit.\"   '),(2488,3155540,'17,June,2004','2004-06-17 00:00:00','    It seems to me that my life likes to develop in cycles, and correct me if I\'m wrong, but I think that the term \'Culmination\' applies.  A new opportunity arises.  I pursue it to it\'s furthest extent.  Then it ends leaving just enough room for the new opportunity to come waltzing in.  Am I the only one who has noticed this type of pattern?   That seems to be the point I\'m at now.  The endeavors that I\'ve been following for the last several months are gently coming to an end, and just beyond my tangible perception are little sprouts of opportunities to come.  All I really have to say is:  \"It\'s about damn time.\"   '),(2489,3155540,'16,June,2004','2004-06-16 00:00:00','    \"Miller, why do you put up with that crap?\" \"I don\'t know.  I think, perhaps the \'crap index\' was a lot lower before.\" \"Bullshit.  You should have given \'em the finger months and months ago.\" \"I know, I know, but sometimes its just easier to go with the flow then to rock the boat. Ya\' know?\" \"Well, you know I don\'t really have room to talk, but you know who\'s to blame, right?\" \"Saturday Morning Cartoons?\"  *The Look*  \"No.\" \"Right, I know.  It\'s me.  And you\'d be the second jackass to point that idea out.\" \"So what are you going to do?\" \"I\'m going to sit back and see what happens.  Good things happen when you\'re not busy ignoring them.  That\'s my philosophy.\" \"I thought your philosophy was \'No Regrets, No Bullshit, and Everybody Wang Chung Tonight\'.\" \"Right.  That too.\"   Now, Shut the hell up, and watch  urlLink this.   '),(2490,3155540,'14,June,2004','2004-06-14 00:00:00','    Somehow, you have to believe that:  1. Jesus loves you, and shares your hatred of homosexuals and Hillary Clinton.  2. The United States should get out of the United Nations, and our highest national priority is enforcing U. N. resolutions against Iraq.  3. \"Standing Tall for America\" means firing your workers and moving their jobs to India.  4. A woman can\'t be trusted with decisions about her own body, but multi-national corporations can make decisions affecting all mankind without regulation.  5. Being a drug addict is a moral failing and a crime, unless you\'re a conservative radio host. Then it\'s an illness and you need our prayers for your recovery.  6. The best way to improve military morale is to praise the troops in speeches while slashing veterans\' benefits and combat pay.  7. Group sex and drug use are degenerate sins unless you someday run for governor of California as a Republican.  8. If condoms are kept out of schools, adolescents won\'t have sex.  9. A good way to fight terrorism is to belittle our long-time allies, then demand their cooperation and money.  10. HMOs and insurance companies have the interest of the public at heart.  11. Providing health care to all Iraqis is sound policy. Providing health care to all Americans is socialism.  12. Global warming and tobacco\'s link to cancer are junk science, but creationism should be taught in schools.  13. It is okay that the Bush family has done $ millions of business with the Bin Laden family.  14. Saddam was a good guy when Reagan armed him, a bad guy when Bush\'s daddy made war on him, a good guy when Cheney did business with him and a bad guy when Bush needed a \"we can\'t find Bin Laden\" diversion.  15. A president lying about an extramarital affair is an impeachable offense. A president lying to enlist support for a war in which thousands die is solid defense policy.  16. Government should limit itself to the powers named in the Constitution, which include banning gay marriages and censoring the Internet.  17. The public has a right to know about Hillary\'s cattle trades, but George Bush\'s Harken Oil stock trade are none of our business.  18. You support states\' rights, which means Attorney General John Ashcroft can tell states what local voter initiatives they have a right to adopt.  19. What Bill Clinton did in the 1960s is of vital national interest, but what Bush did in the \'80s is irrelevant.  20. Trade with Cuba is wrong because the country is communist, but trade with China and Vietnam is vital to a spirit of international harmony.  21. Affirmative Action is wrong but that it is ok for your Daddy and his friends to get you into Yale, the Texas Air National Guard, Harvard Business School, part owner-ship of Harken Oil, part ownership of the Texas Rangers, the Governorship of Texas and then have the Supreme Court Appoint you President of the USA!   '),(2491,3155540,'11,June,2004','2004-06-11 00:00:00','    Hey kids, Sit back, cuddle your mocha, and giggle while you watch  urlLink This   '),(2492,3155540,'09,June,2004','2004-06-09 00:00:00','    Today\'s topic is \"The Look\".  \"The Look\" conveys a clear message to the recipient.  That particular message is as such:   ~\" That was possibly the stupidest thing I\'ve ever heard.  And I\'ve heard you say some pretty stupid things.\"    Also an acceptable message:   ~\"How do you function with such a severe case of stupidity.  I\'m surprised you have enough intelligence to breath.\"   I have been the recipient of this gaze on several occasions.  Most of the time its distributed without discrimination from my wussy brother.  Actually, my little brother is so skilled at distributing this look, that he deserves some sort of acknowledgment.  (He\'s not going to get it here, however, because he\'s a little shit.)  But, one can recognize \"The Look\" by watching for the following:   1) the eye lids drop to half mast, as if they can\'t be bothered with really looking at you.   2) frequently, the head won\'t even turn to face the recipient, but instead just look out the corner of the eye.  (This tactic is for those truly skilled at \"The Look\", and should only be attempted after dedicated practice.)  3) the corners of the mouth pull back in a slight grimace, as if the actual words produced are the cause of the most wretched and foul odor.  4) long moment of silence as the distributor of \"The Look\" gives you the opportunity to realize your own stupidity.   '),(2493,3155540,'08,June,2004','2004-06-08 00:00:00','    While visiting family out of town this weekend, some dumbshit (who shall remain clueless,) sent me this series of text messages.  \"u back frm bar hoser?\"   what are you talking about?   \"stupid drunk\"   I don\'t know who you were talking to, but it wasn\'t me.  I\'m in Spokane.   Needless to say, the dumbass didn\'t respond to that last one, and must have realized that he had spent the night drunk off his ass, talking to some other blonde.    Idiot.   '),(2494,3155540,'04,June,2004','2004-06-04 00:00:00','    I walked out of the restaurant closely followed by my brother and mother.  We laughed as we strolled to the car. \"Tara!\"  shit! I\'ve been spotted.  This town is too small.  I turn to see the person so intent on having my attention.  Who is tha...awe damn.  \"You\'re just going to walk by and not even say anything?\" \"I\'m sorry, I didn\'t see you. How are you?\" He doesn\'t look much like the image I have of him sketched in my mind.  He looks better, in fact.  A little stronger, a little older, a few more lines around his eyes indicate that he\'s seen a little more hard times, and a little more laughter.  My mother and brother turn and continue to head to the car. \"So you living in town?\" \"Oh, no no.  I\'m just visiting my brother. I\'m living in Seattle these days.  How about you, are you back in town?\" \"You were here and you didn\'t look me up?  Are we a little anti-social these days?\"  Who\'s he kidding.  It\'s been almost 4 years since we stopped talking, why the hell would I look him up. I think not.  Not after all the guilt and the grief from his wife, well, actually she\'d be his ex-wife now... But still...  \"I suppose I am a bit anti-social, but I didn\'t actually know you were in town.\"   Because I would have gone somewhere else for lunch.  He just stands there and looks at me with a half smile on his face.   He\'s trying to think of something witty to say, isn\'t he.  Then he\'s going to say we should keep in touch, and offer me his cell number for when I\'m in town next.  He\'ll suggest we go out for dinner sometime, or a beer.  That is possibly the worst idea.    \"Well, it was nice seeing you.  Hope things are going well these days...\" \"Yeah, next time you\'re around visiting, you should....\"  Shit! He\'s going to hug me goodbye!  I quickly stick out my hand to shake his, and brace myself for the worst.  He takes my hand, and stares at me.  The smile ebbing from his face.    Dammit!  I wish his eyes weren\'t so easy to read.  Is he quickly thinking of all the things he\'s wanted to say?   I pull my hand free of his, and I realize that I felt nothing.  Finally after all these years, I feel nothing. \"Have fun, kay?\" He stares at his hand, \"Yeah.  You too.  See you around.\" I turn away, and quickly walk to the car where my family is waiting.   \"He\'s still standing there looking at you, Wussy,\" my brother tells me as he peers over my shoulder. \"I know, I know.  Lets just drive.\" \"I saw him the other night at the bar.\" My mom looks inquisitive, \"Did you go say, \'hi\' to him, Bret?\" \"No.\"  My brother looks at me out of the corner of his eye.  \"It\'s more like we just know of each other.\"   '),(2495,3155540,'04,June,2004','2004-06-04 00:00:00','    I think it\'s a requirement for membership to the human race to create one of these lists at least once.  I\'d hate to be accused of not doing anything with my life, so I\'ve composed a list of things that I\'d truly like to do, and their proposed dates of actualization.   1)Play in a competitive doubles volleyball tournament.  Currently in training.  2)Open a fine art gallery featuring my own Paintings, and Photo\'s.   Retirement Project  3)Mint the \'68 Firebird.   Also a current project  4)Create a short video documenting the progress on the firebird\'s restoration.  Set to Steppen Wolf\'s \"Born to Wild\" and \"Magic Carpet Ride.\"   Summer 2004/2005  5)Meet Ryan Reynolds.   As soon as he lifts the restraining order . 6)Play the djembe outside of a coffee shop, and be accompanied by my brother on the guitar.   3am 7-04-2004 after approx. 5 beers each.  7)Compose a piano piece that doesnt suck.   Winter 2005/2006  8)Create a letter for Dan Savage\'s \"Savage Love\", and have him publish it.   Fall 2004  9)Take a class on glass blowing.   2005?  10)Swim with Dolphins.   2006  11)Do a back handspring without scraping my face on the ground/take a gymnastics class.   2005  12)Tour Europe with my brothers.   Summer 2006  13)Fly a plane (Cessna).   My 40th Birthday present to myself . 14)Live on a boat for at least one summer.   Before age 30.  15)Write my life story.   What the hell do you think this is?!   '),(2496,3155540,'03,June,2004','2004-06-03 00:00:00','    This morning, I actually woke up to the warmth of little bitty rays of sunshine on face.  That puts me into such a nice mood.  Its like the sun can\'t be stopped by the shades on the window and will seep in to greet me any way that it can. (Eleven hours of sleep does great things for my disposition as well.)  Ah.  Now if only I had a Starbucks barista in my kitchen making me my non-fat, sugar-free vanilla latte, the world truly would be a perfect place.  Well, for me anyway. I can\'t really speak for you guys, but it\'d be damn good for me.   '),(2497,3155540,'02,June,2004','2004-06-02 00:00:00','    6:05am    Wake up.  Shower. Dress for classes this morning.   7:08am    No emails from Dad.  Hmm.  That\'s strange.  He must be out running errands for work.   1:24pm    Phone call from Lisa  (friend/intern at grandfathers law firm.)   \"Tara!  Are you ok?!\" \"Yeah, why wouldn\'t I be?  I\'m just getting ready to go to my last class.\" \"...Hang on....* audible whisper * ...Do me a big favor and call your grandmother right now.\" \"I\'d prefer not too, if it\'s all the same to you.  Tell me whats going on.\" \"...* whisper *...it\'s nothing, just call your grandma.\" \"Lisa, knock it off.  What\'s going on?\" * conversation continues. Sastifactory lie is fabricated to veil actual reason for call .  I never call my grandmother.  Inuitive alarm to the quiet wrongness of the world never goes off.   2:05pm    I leave a few minutes early for late afternoon Law and Ethics class.   2:25pm    Try not to daydream about weekends plans in Seattle.  Try to pay attention to lecture.  Nap through class instead.     4:45pm    Walk home.    5:02pm    Door to landing is open.  I never leave it open. Heart pounds.   Burglary?   Race up steps to apartment. \"Hello?!  Wuss?  Who\'s There?!\"   5:03pm   Apartment door is also open, and I burst into kitchen.  Greeted by a red-faced, very somber Uncle, who stands in kitchen dressed in suit and tie, obviously coming right from work. \"Oh, Hi Uncle Pat.\"  Relief.  \"Did Bret let you in?\" \"Sweetheart...It\'s your dad.\" \"What are you talking about?\" Blood rushes to head.  Ears feel full of water, can\'t hear.  Pulse is galloping towards heartattack.  Legs tremble.  \"Whatshappened?Whereishe?!Ishehurt?\"  Try to push past uncle towards rest of apartment. Representative for CDA sheriffs department steps in my way.   5:07pm   Tears run down uncle\'s cheek. \"Honey, your dad died this morning...* continues to speak *  Can\'t hear anything more.  Breathing quick.  Headache coming on.  Push past uncle towards living room.   5:09pm   Manage to swim through the thick air to living room.  Bret sits in recliner with head in hands, shoulders shaking.  Sounds are still murky at best. Voice breaks: \"Wussy?...\"   5:10pm   Bret rushes from chair, weeping stongly.   But Bret never cries.  He throws arms around my neck.  Hot tears are burning like acid through my t-shirt and scalding my skin.  I have all 6 foot, and 200lbs of my little brother hanging from my shoulders.  I don\'t think his feet are supporting any of his weight. I slowly lean forward and place my little brother back in the chair.  Hearing comes back in a tsunami of sound.   \"Oh God...\"  The thunder of my own words rips through my skull.   5:20pm   Like the Miller that I\'ve been groomed to be, I spin away and dash to bedroom.  Things need to be taken care of.  There is more sense in movement.  Keys?   Check.   Phone?  Check.   Wallet?  Check.   Contacts?  Check.  Clothes? ... quickly fill backpack .   Check.   Toothbrush?  Check.   I storm past; Uncle, cousin, random guy, and brother.     5:31pm   I run down the stairs, throw backpack into car, and jam keys in ignition.  It takes 7 attempts.   I have got to get to Dad .  Car starts. What\'s that noise? Oh.  Driver door is still open.  Uncle puts hand on my shoulder.   \"Oh, honey. I\'m so so sorry....\"   5:35pm   My world officially shatters.     5:36pm   My eyes burst like grapes, in my eye sockets.  Or at least thats the way it feels.  Possibly because I\'ve managed to get all ten fingerpads to jam inbetween brow bone, and cheek bone.  The high pitched whine is coming from somewhere in my own throat.  And I\'m not sure why my car isn\'t running,  and where the hell did my keys go?      Sometime after 5:35   All time looses its value.  And we all take the longest trip up to CDA that I\'ve ever known.  Keith drives my car, as we follow Grandpa\'s white Cadi through the mountains of Highway 95 N.  We get passed by a logging truck.  That\'s never happened before. But there has to be a first time.  There is always a first time for every single event.   '),(2498,3155540,'01,June,2004','2004-06-01 00:00:00','    \"Maslow\'s hierarchy of needs was an alternative to the depressing determinism of Freud and Skinner. He felt that people are basically trustworthy, self-protecting, and self-governing. Humans tend toward growth and love. Although there is a continuous cycle of human wars, murder, deceit, etc., he believed that violence is not what human nature is meant to be like. Violence and other evils occur when human needs are thwarted. In other words, people who are deprived of lower needs such as safety may defend themselves by violent means. He did not believe that humans are violent because they enjoy violence. Or that they lie, cheat, and steal because they enjoy doing it.   According to Maslow, there are general types of needs (physiological, safety, love, and esteem) that must be satisfied before a person can act unselfishly. He called these needs \"deficiency needs.\" As long as we are motivated to satisfy these cravings, we are moving towards growth, toward self-actualization. Satisfying needs is healthy, blocking gratification makes us sick or evil. In other words, we are all \'needs junkies\' with cravings that must be satisfied and should be satisfied. Else, we become sick.\"   http://web.utk.edu/~gwynne/maslow.HTM   And that is all I will say about that.   '),(2499,3155540,'01,June,2004','2004-06-01 00:00:00','    \"Why do you expect anything from anyone?  Why did you expect a company to offer you a job? Why do you expect someone to be around to \"care for you\"? Why are you looking for someone else to make you happy?  I think you will find there are many people in your life who DO care for you. Maybe not in the relationship/sexual way you want, but in a very strong and caring way nonetheless.  Nothing in life is guaranteed or secure and no one has been promised anything. You are in a much better situation than others in the world right now and have every opportunity to make what you want of your life.   You have full use of your body. You have full use of your mind. You have incredible artistic and creative talents that many people work hard at trying to achieve and NEVER get there. You are living in a free country. In real terms,  you have no actual problems.  Stop sitting around waiting for someone or something. Stop crying. Realise what you have and make an effort to get what you want.  No one owes you anything at all. Anything given to you from anyone is a bonus.  And if someone is holding you back from moving on, cut them the fuck off.\"    some philosopher   '),(2500,3155540,'30,July,2004','2004-07-30 00:00:00','    Today when I woke up, I realized that it was just going to be one of  those  days. And I\'ve been proven right... A couple of times.  Anyway, in a worthless attempt to add some humor to a barf-bag of a day I offer you a documentation of the drunken stupidness of my brother. (Sorry, Wuss. But it\'s for a worthy cause.)  I should preface by telling you that my brother had to have all of this explained to him the next day.  My brother, being a professional consumer of fine alcoholic beverages, was escorted by his frattie brattie chums to the local pub. He and the \'Sigma-Nerds\' proceeded to get wasted.  Really  wasted.  I assume that they only left the bar because they ran out of money, but who really knows. A few feet outside of the pubs entrance, my brother finds that a simple stroll home is too boring for a savvy adventurer like himself. So he runs. Actually, he sprints towards home.  Unfortunately, like most drunks, his coordination was a bit off. So, when a parked Van found itself directly in his path, Super Wussy had his wits about him enough to know that he wouldn\'t be able to dodge such a crafty adversary.  He ran over it. Bumper. Hood. Windshield. Then down the roof of the van. He finished his routine with a jump off of the back of the van with arms windmilling around his sides.  To the complete surprise of his chummy chums, (who were desperately trying to catch him,) he landed the jump. Apparently it received very high marks. \"10. 10. 9.5. 8.0.\" As his expensively cushioned sneakers hit the ground, he took off running again.  He was reaching speeds more associated with jet planes than a pedestrian, and clouds of burnt sneaker sole were pluming into the air behind him. But faster he ran.  Sprinting down the sidewalk he quickly found a low hanging tree branch posing an immediate threat to his well being. He ducked. He dodged. He fell. Wuss found himself completely parallel with the ground. Or lying on it, if you will.  Laughter peeled through the night behind him as his chummies doubled over, red in the face, and trying not to wet themselves. When they composed themselves enough to catch up to him, they found him still lying on the ground.  He looked up at them, and with a look of complete sincerity he said  \"Whoa. Gravity Storm.\"  More laughing ensued.  The next morning all of this was re-enacted for him. Mostly so that he could understand why he had giant road-rash scrapes on his legs and shoulder.   Oh, and Cheers , Wussy!   '),(2501,3155540,'29,July,2004','2004-07-29 00:00:00','    When I was in Hawaii some woman gave me a free sample of perfume.&nbsp; I just smiled, stuffed it in my purse, and kept walking.&nbsp; It\'s been just sitting around for months, but yesterday I tested it out, and found I REALLY like it.&nbsp;   The funny part is, I don\'t like it as a perfume that I would wear.&nbsp;&nbsp;Oh, no.&nbsp; &nbsp;It\'s not a scent that I want people to associate with me.&nbsp;&nbsp;&nbsp; It\'s more of a masculine cologne smell.&nbsp; It happens to smell very similar to the cologne that my dad used to wear.&nbsp; Which is why I like it.&nbsp; Very comforting.&nbsp;   Strange, yeah?   '),(2502,3155540,'28,July,2004','2004-07-28 00:00:00','    \"It is a really cute hampster.&nbsp; Nice choice.\"  \"Thanks.&nbsp; I think I should name it \'Mokie\'.&nbsp; Wasn\'t that the name of a fraggle off of that old muppet show?\"  \"Sure.&nbsp; Why not.\"  \"Ok, you hold the box with the rat while I drive.\"  \"Hey, do you want to grab some lunch?\"  \"Uh, yeah, but it\'d have to be drive-through.\"  \"Oh,...I think you\'re going to change your mind about that when you see the guy that just walked inside...\"  \"Damn.&nbsp; Ok, we\'ll go inside, but it has to be quick.&nbsp; I don\'t want this rodent roaming around my jetta.\"  \"Don\'t worry about it.&nbsp;&nbsp; I\'ll put my jacket over it.\"    20 minutes later and&nbsp;without the phone number of the really cute guy...    \"Well, I would have gotten&nbsp;his number&nbsp;if I was wearing a low-cut shirt.\"  \"Whatever.&nbsp; The Tara I know would have just walked up, smiled, and had him begging to give you his number.&nbsp; You\'ve obviously lost your touch.\"  \"Piss off.\"  \"Uh,...Tara?\"  \"Yeah.\"  \"I got some good news, and I got some bad news.\"  \"How about I give you 30 seconds to give me all the news, or I make you walk your happy ass home.\"  \"Ok, well the mouse is not in the box any longer.&nbsp; I can tell, you see, because of this mouse size hole that\'s been chewed out of the side of the box.&nbsp; But the good news is that it\'s probably still in the car somewhere.\"  \"Are you shitting me?\"  \"I shit you not.\"  \"Well, shit.\"    Quick search turns up nothing.    \"Where the hell could a little mouse possibly have gone?\"  \"SHH!&nbsp; Be quiet...\"  \"Oh my god.&nbsp; It got into the vents.\"  \"Well it needs to get the hell out, because I am not having a rat carcass decreasing the value of my car.\"  \"Ok, well...My dad might still be at work.&nbsp; We could go to his&nbsp; car shop, and see if he can take it out.\"  \"Ok.&nbsp; But lets just be sure to turn the heat off.&nbsp;&nbsp; No cooked rodents, please.\"    3 min later, while driving down a very busy Division St...    \"Where do I turn?\"  \"Its right on the rig...\"  \"ARGH!!!\"  \"What?!&nbsp; What?!\"  \"ITS ON MY FOOT!&nbsp; It just flopped out of the dash onto my foot!&nbsp; Jesus, that scared the hell out of me!&nbsp; Grab it before it has a chance to get into anything else!\"  \"I can\'t reach!\"  \"Ok, hold the wheel and I\'ll grab it....\"  \"Car.&nbsp; Car!&nbsp; CAR!\"  \"Dammit, Tina!&nbsp; Hold the wheel!&nbsp; ...Ok, I got it.&nbsp; Here take this...\"  \"Don\'t throw the rat at me!\"    Two days later that demon possessed rat had babies.   '),(2503,3155540,'25,July,2004','2004-07-25 00:00:00','    God, this week has been so nice.&nbsp; Sleeping in everyday has been oh-so-beneficial.&nbsp;&nbsp; I\'ve spent all weekend just cleaning my apartment, and lounging around.&nbsp;   Wait,...well, I did go out to dinner with an admirer earlier, but I got to be honest.&nbsp; It wasn\'t all that much fun.&nbsp; I mean, he\'s really cute, but I was seriously bored by the time our drinks arrived at the table.&nbsp; I think sometimes there\'s that connection, and sometimes there\'s not.&nbsp; What can I say.&nbsp; My batting average isn\'t that great.&nbsp; As a matter of fact, I strike out more than anyone I know.   Anyway, my brother \"Super-Wuss\" came over for a visit a couple of days ago.&nbsp; When I say, \"Visit\" I mean that he came here because he found a car he wanted to buy, and I just happen to own a convenient place for him to crash for the night.&nbsp; But I had a really fun time hanging out with&nbsp;him.&nbsp;&nbsp; AND I nicked a couple of his humorous&nbsp;T-shirts.&nbsp; None of which are really appropriate for work.... not like thatâ€™s ever stopped me before.&nbsp;   The Wussy did bring me an&nbsp;unassembled model of a \'68 Firebird, so I\'ve been diligently working on that today.&nbsp; This car project is going to be a massively emotional&nbsp;one, I can tell.&nbsp;&nbsp;I get teary even as this model is taking shape.&nbsp; I just keep thinking about how much&nbsp;Dad would have loved working on the Firebird with us.&nbsp;&nbsp;&nbsp;Hopefully,&nbsp;he still&nbsp;is, even if I can\'t see&nbsp;him.&nbsp; Today I painted the chassis, body, and added Dad\'s initials to the tail of the Bird.&nbsp; I also assembled the engine block.&nbsp; Seems kinda stupid to be gluing all of these little pieces in place that people will never be able to see, but if I\'m going to do it, I\'m going to fucking do it right.   Oh shit.&nbsp; The Simpsonâ€™s are on.&nbsp; &nbsp;Later   '),(2504,3155540,'22,July,2004','2004-07-22 00:00:00','    I feel so damn stupid.&nbsp; I\'m about a week into this job and I don\'t think I\'ve ever felt so unqualified for the challenges that life has thrown at me.   Ok, I\'m going to take a \'time out\'&nbsp; and hide under my desk for a couple minutes then I\'ll get back to work.   '),(2505,3155540,'20,July,2004','2004-07-20 00:00:00','    Dating is a lot like... Little Red Ridding Hood ? &nbsp; Hmm.&nbsp; &nbsp;This metaphor might take some fancy foot work, but we\'ll see where it goes.  &nbsp;  I\'m a lot like Little Red.&nbsp; &nbsp;Only, the term \'little\' doesn\'t describe anything about me.&nbsp; &nbsp;And I would  never  make such an atrocious mistake as to wear a red hooded cape. &nbsp; (Honestly, what was she thinking.)&nbsp;&nbsp; But it\'s the innocence that is the key similarity here.&nbsp;&nbsp; ...Hey, it could be true!&nbsp; Stop laughing!... Anyway, like I was saying, we\'re both&nbsp;very inno...&nbsp; Fine.&nbsp;&nbsp; Innocent-ISH.  &nbsp;  I was just skipping along through the forest to visit Grandma, (who in this story will be played by Fhay, because I really only visit my Grandma if I absolutely have to.)&nbsp; Skip, skip, skipping along through the parking lot,...I mean Forest, all the while being stalked by the Big.&nbsp; Bad.&nbsp; Wolf.&nbsp; (Played by the boy in question.)&nbsp; The Boy, I mean \'Wolf\"&nbsp;spends half of the story following me through the forest, trying to woo me by whispering in my cute little ears, but I\'m oblivious, because the wind is whistling through the 8 rings in my ears.&nbsp; (That actually happens on occasion.&nbsp; Its very unnerving.)&nbsp; So, I continue to skip,...well, not so much skip, because I\' probably wearing flip flops,... So let\'s say \'flopping\' through the forest... Singing to myself.  &nbsp;  \"I don\'t like Regae...Oh no!&nbsp; I love it!&nbsp; Yeah!\"&nbsp; (What?&nbsp; You were expecting a sweet little \"Tra la la la\"?&nbsp;&nbsp;Me thinks not.)  &nbsp;  So the Sneaky Wolf, decides that he really likes a challenge,&nbsp;and he quickly trots to Grandma Fhay\'s house.&nbsp; But the good news is that Fhay wasn\'t home...&nbsp; No.&nbsp; I don\'t know where she was, because she rarely leaves a note.&nbsp; Anyway, Mr. Wolf lets himself in and quickly adjusts his appearance so that \"Mr. Wolf\" becomes \"Mr. Hottie Pants\",&nbsp;and he&nbsp;sits quietly waiting for semi-innocent Red to come to him.&nbsp;&nbsp;&nbsp; ...You\'ve got to admit, it\'s a damn good move.&nbsp; I\'ve never been able to really get it to work myself.&nbsp; Mostly because I have a pretty short attention span, but a damn good move nontheless.  &nbsp;  \"Stupid flip flops.&nbsp; Can\'t skip at all in these&nbsp;damn shoes.\"&nbsp; And I,&nbsp;(playing &nbsp;\'Little Red\' in this fabrication)&nbsp;flip and flop up to the door.&nbsp;&nbsp; I quickly knock twice.&nbsp; The door swings open, and there is Mr. Wolf,&nbsp;looking&nbsp;oh-so-sexy, holding a dozen roses&nbsp;and whispering the sweetest things.&nbsp;&nbsp;  &nbsp;  It\'s at this point that the fable is going to need a bit of a tweak or two.&nbsp;&nbsp;  &nbsp;  \"My, what big eyes you have... And a really nice butt.&nbsp; Yowza!. \"  &nbsp;  Blah, blah, blah.&nbsp;&nbsp; Anyway, the conversation goes on, because it\'s so entertaining.&nbsp; And I completely neglect to notice the great big teeth Mr. Wolf is sporting.&nbsp; I walk in, and BAM!&nbsp; The sweet talk ends, and I find myself on the receiving end of Wolfish sarcasm.&nbsp;&nbsp;  &nbsp;  Happens every time.&nbsp; &nbsp;&nbsp; &nbsp;&nbsp; &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; ~*~*~*~*~*~*~*~&nbsp;  ...Wow, that turned out much better than I thought it would.&nbsp;&nbsp; That metaphor&nbsp;worked perfectly, didn\'t it.&nbsp; As a matter of fact, those two stories are almost identical.&nbsp; Damn.&nbsp; I\'m good at this!&nbsp; Tomorrow we discuss \"Hanzel and Gretel\" as a metaphor for the American work place.   '),(2506,3155540,'19,July,2004','2004-07-19 00:00:00','    I don\'t have much to tell you about, Kiddies.&nbsp;  &nbsp;  -I really like waking up AFTER the sun has risen.&nbsp;   -I took a boat trip to Victoria B.C. with my parents this weekend, and saw about 20 orcas flopping around making a spectacle of themselves.&nbsp;   '),(2507,3155540,'16,July,2004','2004-07-16 00:00:00','    What is it that is really attractive in another person?&nbsp;  &nbsp;  Some girls really subscribe to the \"bad boy\" stereotype.&nbsp; The motorcycles and fast cars. The tattoos and the piercings.&nbsp; The wild hair.&nbsp; The \"I don\'t give a fuck what you think\" attitude.&nbsp; I\'ll admit, I\'ve fallen for the occasional bad boy, because let\'s face it.&nbsp; Confidence, (even if misplaced) is sexy.  But it never works out.&nbsp; Because that \"I don\'t give a fuck what you think\" attitude usually means that they don\'t, in fact, give a fuck what you think.  &nbsp;  There\'s also the \'Mr. Sensitive\' stereotype, which in my humble opinion is over used bullshit.&nbsp; I think that one of two things happen with the \"Mr. Sensitive\".&nbsp; 1) They\'re just quite or&nbsp;shy, thus being mistaken for mysterious, or a sensitive thinker.&nbsp; Or 2) They\'re going out of their way to act sensitive and caring, because someone told them that that\'s what girls want.&nbsp; It\'s crap either way.&nbsp; I doubt that most people develop&nbsp;sensitivity, and a caring attitude&nbsp;towards another until they\'ve&nbsp;known them for awhile.&nbsp;  &nbsp;  I have found myself being drawn more to a guys voice than his appearance.&nbsp; Mostly because a person can dress to portray whatever they want, but they can\'t hide in their voice.&nbsp; I\'m not talking about the tone, or accent behind a voice, because everyone&nbsp;has a different opinion on that.&nbsp; But If you listen carefully you can pick up on the genuine person.&nbsp; You can hear their pride, intelligence, ego, sincerity, affection, insecurities, motivation, ignorance, sarcasm, irritations, restlessness, and their desire.&nbsp;     &nbsp;   SO, if a guy is a cocky son of a bitch, you\'ll hear it.&nbsp; If he\'s laid back with a good sense of humor, you\'ll hear it.&nbsp; If he\'s happy with life, you\'ll hear it.&nbsp; If he\'s lying, you\'ll hear it.&nbsp; If&nbsp;he\'s comfortably being himself,&nbsp;you\'ll hear it.&nbsp;&nbsp;&nbsp;If he\'s just saying what&nbsp;he thinks you\'d want to hear, you\'ll hear it in his voice.&nbsp; &nbsp;If he\'s a whiney bastard, you\'ll definitely hear it.&nbsp; If you need help figuring it out, then simply listen to him laugh.&nbsp;  &nbsp;  A really good voice is the audio equivalent to walking in from the rain, and curling up in front of the fireplace with a cup of cocoa.&nbsp; It should be able to peel you off the ceiling with just a couple of syllables when you\'re freaking out.&nbsp;&nbsp; Its whisper should give you goose bumps.&nbsp; &nbsp;It should be one of&nbsp; the most comforting sounds.&nbsp;&nbsp; It should be able to make you smile like a muppet with just a \'Hello\'.&nbsp; &nbsp;It should be the biggest turn on, and the sweetest kiss.&nbsp; It definitely should never make you cringe.  &nbsp;  Oh, and guys,&nbsp;this all goes for girls as well.&nbsp; Only girly&nbsp;voices are even&nbsp;even easier to read.   '),(2508,3155540,'13,July,2004','2004-07-13 00:00:00','    After the New Employee Orientation that the Professor and I suffered through I promptly picked up my new little blue ID badge, gassed up the Jetta and headed to meet my mom for a late afternoon lunch. My mom was the president of the Pacific Northwest Human Resources chapter, so I was sure to arm myself with the black box containing all of my benefits, 401k, and stock information.  Oh, I also toted along the complimentary coffee mug with the company logo.  (If anything would prove that I was now a valued employee, the mug surely will.  They don\'t just hand those out to anybody, you know.)  I was ready.  ...Or at least I thought I was, but you can bet that there were questions beyond my ability to answer.  \"Do they match everything you make in the 401k?\" \"I uh...\" \"Did they provide you a list of doctors if you decide to go with Group Health?\" \"It\'s umm...\" \"What about vision?  Does Group offer the vision, or is it through someone else? Is the coverage any good?\" \"Well, it\'s uh...\" \"When are your shares vested?\" \"I think it\'s um... Hey.  Did I show you my coffee mug?\" \"That\'s nice, Sweetheart.  Well, from what I can see they really take care of their employees.  Some of this stuff is better than what we offer our employees.\"  The analytical look starts to slide away from her eyes, and the corners of her mouth turn up. \"You did it, Honey.\" \"Am I a success?\" \"Of course, Sweetheart.\" \"Am I a \'Grown Up\'?\" \"Let\'s not push it.\"   '),(2509,3155540,'10,July,2004','2004-07-10 00:00:00','    This exclamation was screamed at me as my car winded through the extensive apartment complex that I live in.  After being mildly surprised that someone other than myself was a native English speaker, I scoffed at the idea that someone would go out of their way to embarrass themselves like that.  As I continued to drive past, I said what you may be thinking right now.  \"Trash.\"  Normally I would just continue with my judging ways and assume that I am, of course, superior to all.  But I\'ve been thinking about my Firebird dilemma quite a bit lately, and it occurred to me... \"Wait.  My  \'White Trash Index\' could be considered quite high.\"  So, please allow me to get off my pedestal for a moment, and display for you, faithful reader, my white trash qualities. Hopefully, this irresponsible display will alleviate some of my shame.  -I was born in the Snake River Valley, where teenagers live it up by country swing dancing, spitting tobacco, and cow-tipping.  -I own a \'68 firebird, which won\'t loose it\'s Flaming Chicken status until its pretty close to being mint.    -My brother and I also own a \'74 Ford F-250.  If you\'re wondering, its \'Aztec Gold\' and liberally spattered with rust.  Not to mention, it happens to be sporting some mighty glass packs, and used to have a gun rack.  I kid you not.  -There is a pair of cowboy boots in my closet right now.  -I have nine rings in my ears, although I think that makes me a different kind of trash...  -Within the last week, I\'ve used the phrase \"It was funnier than ALL GET OUT\" as well as the word \"Rig.\"   -I\'m listening to \"Lynyrd Skynyrd\" this very moment.  -I\'ve crushed a MGD beer can on my forehead. (Ya\' see, there\'s a trick to it.  You\'ve got to dent the side of the can just slightly with your thumb before you smush it to your head.)  -I also spent a few summers as a kid hanging out at a horse ranch, where I learned to ride, and chew on hay.  -I\'ve been to a bluegrass concert. Where they didn\'t have anything as fanshy shmanshy as \'Dr. Pepper.\'  -I was proud to meet a bull rider.    I\'m not trying to imply that all cowboy behavior indicates \'white trash\' because there are a lot of other qualities not mentioned.  It\'s just a generalization, if you will.  Dammit.   '),(2510,3155540,'09,July,2004','2004-07-09 00:00:00','    Ok, so I may have a couple of sneaky skills up my sleeve when it comes to finding a persons phone number, or address.  But today, I found myself at a complete loss.    The gentleman that I bought the firebird from neglected to sign the title before sending it with the car. I\'m assuming that he\'s never bought or sold a car before and wouldn\'t know better.    Or   possibly, space aliens have been doing a lot more than anal probes to the red necks down in Texas, and have left him with no intelligence to speak of.  Either scenario is plausible.    In order to get the car titled I have to send him a few forms to sign.  No big deal, right?  Wrong.  This little bastard has disappeared off the face of the earth.  None of the contact information I have for him is correct any longer.  His business?  Gone.  His phone numbers?  Disconnected.  Is there even a person with the same name in that city?  Nope.  Not to mention, I was told by another victim of the aliens evil plot for butt information, (logically employed by the DMV) that without some sort of signature I cannot title the car.  She said \"There is nothing you can do.\"  Damn aliens should have just let her drop back to earth without a parachute.  So, at this point I\'m at a bit of a loss.  I have this doodah\'s old address, but have found no way to find his forwarding address.  My only idea is to call the post office near his old address and see if they\'ll give me a new address.  Any other suggestions from you little hackers?   '),(2511,3155540,'07,July,2004','2004-07-07 00:00:00','    The Islanders have voted, and I\'m  NOT  being cast off the island! ...I mean office.    \"Well, I got to tell ya\'.  The race was pretty intense there towards the end.  I thought I was a goner in the last stretch of the interviews.  The competition can be pretty charismatic.  But, I said to myself, \'SO AM I!\' And I just dug down deep and gave it all I had.  And I won.  So I\'m feelin\' pretty good about that.\"  I\'ve only seen \'Survivor\' once, so I can\'t remember how those idiots talk.  But anyway, I GOT THE JOB!  And if you don\'t know what I\'m talking about THEN YOU CAN PISS OFF!   '),(2512,3155540,'03,July,2004','2004-07-03 00:00:00','    I have a couple little treats for you cute wittle Minions.  I bring these wonderful gifts mostly because I\'ve been moping around because of my estranged job status.  So in an offering of peace I bring you:  The musical stylings of  urlLink   Chickiboom  .  I believe it is referred to as \"Musical Poetry.\"  And it is oh-so-nice when you\'ve been self-medicating, if you will.  The second gift is for you  urlLink MoveOn.org  followers out there interested in wearing your world views on your body.  For you I have  urlLink  Sovereingntee .   Yeah!  T-shirts Gallore!  I personally like the \'Mad Cowboy Disease\' T-shirt.  (Caleb promises that the site will be up by sometime today, so if you are having troubles browsing through the catalogue, then keep checking back!)  You guys will really love these.   '),(2513,3155540,'01,July,2004','2004-07-01 00:00:00','    urlLink This  is funny.  I\'m not going to copy and paste the whole thing for you, but I highly suggest you check it out.    Oh, and I totally agree in regards to thongs.  The stupid things are pointless.  Girls, if you\'re worried about panty lines, then don\'t bother wearing anything at all.  It saves the rest of us from having to watch you pick your wedgie, as well as limiting the view of your t-backs hanging out the top of your low-rider pants.  It\'s not sexy.  It makes you look seriously masochistic.  Not that I\'m an advocate for granny panties, or anything...  Oh, nevermind.  I\'ve had my say.   '),(2514,3155540,'05,August,2004','2004-08-05 00:00:00','    I have a confession. I\'m developing a little bit of a crush here at work. I\'ve seen this cutie wandering around the place, and being the gutless coward that I am, I\'ve only managed to follow him around from a distance. What can I say! I don\'t like asking guys out. I\'ve done it a few times, and frankly that was enough for me thankyouverymuch.  The first time I asked a guy out, we damn near ended up getting married.  The second time we went out for a few meals, a movie, and then I couldn\'t get rid of him.  The third time I got rejected outright. I know, I know. I couldn\'t believe it either, but it\'s true.  So I\'m not doing it anymore! I think I\'m just going to slink around corners and try to look like I\'m on an important errand to see if I run into Mr. QT again. Who knows. Maybe I\'ll even talk to him someday.  Right. And Ryan Reynolds is stalking ME now. Sure.   '),(2515,3155540,'04,August,2004','2004-08-04 00:00:00','    Well Kids, it\'s done.  The monstrous Ford f-250, (affectionately referred to as \"Old Gold\") that was my dad\'s prized vehicle has been sold. This truck symbolizes my dad and everything he was in this life. Dependable. Rough around the edges. Determined. And a bit growly. He had it since before I was born, and happily abused it for the next 27 years. Even Jebzilla (monster dawg) recognizes the sound of the glass packs rumbling down the road.  I don\'t know how many times I\'ve gone careening out of the house to jump into the back of the truck as it was about to drive away. ...What can I say? It was bound to be an adventure if Dad was taking the truck. Not to mention that learning how to drive the Truck was the most accomplished I have ever felt in my life... thus far.  \"How do you know so much about cars, Dad?\" \"The Truck used to break down a lot.\" \"Why did the truck break down so much?\" \"Because I used to worked on it a lot.\" \"Ah. Gotcha.\"  -It\'s been destroyed and rebuilt so many times, that it openly defies the idea of mortality. -It\'s had very large trees dropped on it, and was tough enough to drive home without crying. -It\'s pulled many unlucky cars out of ditches. -It\'s laughed at perilous snow storms as it plowed through, completely unphased. -It\'s hauled more furniture then any stupid U-haul van. (Most of which was mine.) -It helped to convince the world of its superiority in the \"Chevy\'s vs. Fords\" competition. -It\'s been stuffed with so much wood, and yard debris that the leaf springs have seen their lives pass before their eyes on numerous occasions.   IT. WAS. UNSTOPPABLE.   And now its just rusts out at our lake cabin, developing its own little collection of pine-needles. I certainly don\'t want to get rid of such an important piece of memorabilia, but I don\'t have much choice. I really don\'t have a use for it, and I kinda think Dad would want someone else to beat the shit out of it, instead of just letting it sit.  So off it goes. Hopefully to a good home. On a farm. With lots of hauling to do.   '),(2516,3155540,'03,August,2004','2004-08-03 00:00:00','    I\'ve got nothing really to tell you wittle minions about, so I\'ll give you a savory little taste of my favorite poet.   View with a Grain of Sand   By Wislawa Szymborska, Selected Poems (Nobel Prize for Literature)     Love At First Sight       They\'re both convinced    that a sudden passion joined them.   Such certainty is beautiful,   but uncertainty is more beautiful still.      Since they\'d never met before, they\'re sure   that there\'d been nothing between them.   But what\'s the word from the streets, staircases, hallways--      I want to ask them   if they don\'t remember--   a moment face to face   in some revolving door?   perhaps a \"sorry\" muttered in a crowd?   a curt \"wrong number\" caught in the receiver?   but I know the answer.   No, they don\'t remember.      They\'d be amazed to hear    that Chance has been toying with them    now for years.      Not quite ready yet   to become their Destiny,   it pushed them close, drove them apart,   it barred their path,   stifling a laugh,    and then leaped aside.      There were signs and signals,   even if they couldn\'t read them yet.   Perhaps three years ago   or just last Tuesday   a certain leaf fluttered from one shoulder to another?   Something was dropped and then picked up.   Who knows, maybe the ball that vanished    into childhood\'s thicket?    There were doorknobs and doorbells   where one touch had covered another   beforehand.   Suitcases checked and standing side by side.   One night perhpas, the same dream,    grown hazy by morning.    Every beginning   is only a sequel, after all,   and the book of events,   is always open halfway through.   '),(2517,3155540,'01,August,2004','2004-08-01 00:00:00','    She\'s got horribly cute puppy pictures on her site today, Kids. I highly suggest you give the \'fhayzzer\' link a whirl.  I finished the Firebird model this morning. I say that with an underlying tone of accomplishment, but now that I read back I realize how pathetic I am. I skipped out on going to the International Brew Fest with friends yesterday because all I wanted to do was to glue my fingers together, and get paint on my shirt while I piece together a model of a \'68 Firebird. Something\'s just not right with that. Either way, I\'m horribly proud of this wad of plastic posing as a car. It doesn\'t look that great. Some dastardly glue snuck it\'s way onto one of the windows, and there\'s a fingerprint or two on the body\'s paint job, but I love it. It makes me want to start work on the lonely Firebird that is begging to be driven, but can\'t until it gets titled. I\'m seriously considering forging the Mr. Dillhole\'s signature so that \"Firebird Restoration 2004\" can commence. In case you\'re wondering  urlLink this  is what a well loved Firebird should look like.  I\'m getting the impression that my boss think\'s I\'m retarded. Ok,... To be fair, he\'s only about half wrong. I think a couple of my highly sarcastic comments have been taken literally. Which is SUCH a bad thing. Matt\'s constantly telling him that I was dropped on my head as a child, which really doesn\'t help matters much either. All\'s fair in Pranks and Taunts, but I really would like to do some cool things with this job, or at the very least somehow manage to keep it. All I have to do is try not to run with scissors, and avoid programming a VCR, and I\'ll be money, Baby.   '),(2518,4248100,'23,August,2004','2004-08-23 00:00:00','  sad..jiawei could have won!!! hope she wasnt thinking abt the $500,000 while she was playing.. she was winning 3-1 then in the end lost 3-4..nvm tmrw hope she\'ll win bronze! then history will repeat itself! medal dream...the last round was so tense.its lyk they were tie all along..7-7. 9-9 then 11-9 sigh..  wearing contacts now. went to lensmen at parkway this mrng. took half and hour to wear the contacts! the optician was lyk waiting for me..poor thing.but who ask him to do this jobb! :P anyways earlier took another half hour to wear it.. nvm soon ill learn the art of wearing contact lenses! sigh..later have to spend another half hour in the toilet taking it out.. ):  mrng trng was so boring butterfly drills..butterfly drills..butterfly drills..my dad talked to peter after tt. im gonna train under old man frm tmrw onwards! yipee! hahahs finally in silver grp! can slack.. heard mel c resigned! woohoo!  kk i gtg take off my contacts now~'),(2519,4248100,'23,August,2004','2004-08-23 00:00:00','  thanks alot bird flu! now the malaysia experience thing is postponed..cant skip school! why of all time this period? sigh..hafta go for boring lessons..  6 ppl absent today..plus nic who went for matchh.  had geog test today.. im just feeling to sian abt this! sigh... ):'),(2520,4248100,'23,August,2004','2004-08-23 00:00:00','  just checked out kluang.. its an industrial estate i guess.. coal field, oil palm plantation and RUBBER PLANTATION!!! i guess i got what i wanted hahahs! and its a super ooh-loo place! its lyk we\'re tribal ppl..no contact with the outside world and the main event taking place - olympic games. and we cant sleep in our cosy bed.. sad ): kk i really gtg now..!'),(2521,4248100,'23,August,2004','2004-08-23 00:00:00','  hehh! is our class setting some new record or something? 4 more ppl went off after recess.. yh, lizzie and jugs sick.. nic comp.. tts 12 ppl absent already.. ppl who are sick are junwei, huan bin, clement and me!! recovering are kevin, kaibin, prem, joshua, jayanth and hanson! 21 ppl either sick or recovering!! hahas nice one 2h!      anyways english was dead boring except for all the jokes abt quah and yanxiang drawing her :P pon second period of comm today! had a slight headache..whoo! went to dining hall history watched video.. i slept.headache.      now? still a little.. is this period the sick season? whye is everyone sick during this period? hope i\'ll be even sicker so can\'t go for the malaysia challenge!'),(2522,4248100,'23,August,2004','2004-08-23 00:00:00','  whoo! it was raining on me last nite!!! the raindrops woke me up.. damn it! couldnt sleep after tt. now i\'m soooo sleepy! and feeling a little sick!      in comm lab now! hahahs.. 8 ppl are sick today!!! lang, aaron, boyi, chris, shailesh, jingping, kelvin and leon!!! bingo! the return of the sick ppl! have maths pop quiz later..gng to mug now! bye!'),(2523,4248100,'23,August,2004','2004-08-23 00:00:00','  woke up feeling unwell this mrng! hahahs skipped chinese((: pe)): and english went to school after recess so tt our grp could present re. but guess what?! chris was sick so he left for home! bingo!! waste of time gng to school.. had malaysian experience 2004 briefing.. slack cool. yyea hope tt this time wld be better than sec1 orientation camp. i want a nightwalk! harharhar.. hope we can stay between rubber and banana plantation.. psycho me. the synopsis of our camp is lame -.-\' lyk some abstract frm some book. we\'ve gotta find 4 necklaces to save the world frm some mad scientist.. something lyk tt..      anyway after tt went back to class.. helped mel did some survey.. then some ppl went to change the buttons of our class comm\'s keyboard hahahas! fun?! cool?! great?! probably.. quah and chiang came back. leon first grp to present.. he found the \'weird\' keyboard.. said i did it. fine! whatever.. he wasnt even there to see what happened and who were the ones changing the keyboard. then cheryl yap came in.. she was pissed. quah is so irri. cant stand her in class. so she insisted to go to 2g to do the presentation. yipee! we have a whole classroom to out self. it was fun. played badminton in class. cheryl yap walked past again.. peeked into our class..luckily jev told us lyk few secs b4.. if not we\'re dead..might hafta write essay. found more \'quah\'s around the class room. again..i swear its not me.. hmm.. whatever lahhs if they still dont believe me then let them say what they want.      went to kallang kfc for tea! hahas saw a company set up after me! woohoo! its \'bert lighting company\'.. wonderful ehhs? its lyk i seldom see my name around cos its so not common lyk benjamin or nicholas or john. anyways there\'s a street called \'filbert street\' in san fran.. was there last year. felt weird though. anyways i gtg mug geog now! have test this friday.. cyyers soon!'),(2524,4248100,'23,August,2004','2004-08-23 00:00:00','  slept frm 4.20 to 7 pm..wasted 2hrs 40 mins of mugging time..!'),(2525,4248100,'23,August,2004','2004-08-23 00:00:00','  ppl suspecting its me who wrote \'quah\' on others worksheets! i swear it ain\'t me! how many times do i have to repeat? i think its chris! he has almost the same handwriting as me..how could i have been the one cos i left class at 10.40 and came back only at 2.15? was in the canteen throughout recess! saw this smiley on aaron\'s worksheet and it looked so much like chris..and also elton saw him did it..however he keeps denying it..why?! and now who\'s the one \'addicted\' or \'crazy\' over quah? chris. who is the one who composed the pledge of quahlaland? chris.. hmm..i just dont understand.  chinese was first period..hahas! forgot my chinese speech! damn i haf gotta do 2! luckyy me. fong said that my name came from \'DILDO\'? funnyy..donnoe whye he said tt. the only link is..philbert-philbo-dhildo-dildo..hmm..weird! then was english presentation! ppt was damn funny! couldnt help laughing. recess time quah agreed to let me take photo of her. she\'s very weird out of class though..totally different during lesson when she keeps gng \'focus\'..watched some video on DNA during bio..boring! nigel and jev fell asleep! me? had to prepare for oral comm speech. lucky me the another sg guy did the same topic as me. decided to do the topic i wanted to do last week..failed bingo! have to redo..somehow i donnoe how i should feel..lucky? sad? glad? relieved? i really cant describe it. now i\'ve gotta go mug chinese. tmrw have ting xie and i\'ve not touched anything yet..gonna die! ):'),(2526,4248100,'23,August,2004','2004-08-23 00:00:00','  sigh..today was promethium day..lim geenee (genie) had to scold us!!! ((: quite unluckyy day for me though. tried to cling on over the toilet wall and fell in the toilet bowl..one leg wet :P and then came photo-taking...hahahs junwei said that i looked weird in my \'new\' specs..actually i had it for quite some time but just did not wear it..some say i look more lyk a MUGGER..do i?! came back to class today and found everyone\'s worksheet with \'quah\' on it..?! others think i did it..! but i swear its not me!! i\'m not so childish and...stupid! boring ne lunch talk next..almost slept. the 2g guy beside more was sick!!! mixed milk with asparagus, tomato, green pepper and pineapple rice?! yucks! then squash them!.. went over to the other table and guess what? gabriel flicked food all over me... left yellow stains on my uniform... -.-\' quah somehow said that i was DEFYING her?! just had to go and buy uniform tts all! what\'s her prob? hafta pick 9 ppl out of 23 ppl to participate for house carn badminton..HEADACHE!!! did english project today..hahas great fun though. junwei and kennedy looked so spastic in the ppt! doing persuasive speech now..damn tired!!! cut on arm - aching elbow - urine-wet feet - yellow-stained uniform..and oh ya! have to carry books for quah another day..pissed really want to watch the full length version of the opening ceremony of athens 2004 somehow..i really dunnoe whye though. gonna sleep soon..hope tmrw would be a better dayy! zzzzzzzzzzzzzzzzz..'),(2527,4248100,'23,August,2004','2004-08-23 00:00:00','  watched the short repeat version of the olympic games opening ceremony today..it was fantastic!!. the lighting of the olympic rings was superb! singapore\'s team was so small. compared to others..and couldn\'t find indonesia though.hahahs! and there one country with only one girl..so poor thing ): and i like the fireworks! esp the colourful ones. the lighting of the olympic flame was real cool. haha getting better every four years!'),(2528,4248100,'23,August,2004','2004-08-23 00:00:00','  sigh...jiawei lost!!! think she was frustrated cos the south korea kept chopping the ball! hahahs.. anyway the koreans are lyk dogs!!! kept barking all the way.! do koreans bark often? how nice if both the koreans played against each other..the it\'ll be a barkin match lohhs!  spent the whole day mugging today..now\'s my only free time. but after tmrw chinese test, i\'ll be free!!! :P'),(2529,3483901,'28,June,2004','2004-06-28 00:00:00','   Like my PG rated title? Never know when a 13 year old my stumble across my blog. Little fuckers. Oops. Um....anyway, nothing much has happened lately. My life is very uneventful, which is sad since I live in Las Vegas. I do see some very interesting things here. Things you would never see in Colorado. Like the following: 1.A dancing Santa on the corner of major intersection. 2.A Walgreen\'s shopping cart right outside my door (my future home) 3.Being able to hire a Mexican to wash my car at the do-it-yourself carwash (they don\'t have drive thru car washes here, crazy stuff). 4.A cracked out, hitch-hiking couple screaming at each other on the side of the road. 5.A homeless man with a \"Why lie? I need a beer\" sign on the same corner everyday. (I always give him money) 6.A homeless man in a wheelchair on the opposite corner of the beer man with some sign asking for money (I never give him any)  There are others, but I\'m too lazy to think what they are.  God, I have no motivation to do this right now. I\'m only doing to amuse Carissa. However, I\'m putting myself to sleep right now, so I don\'t think I could amuse anyway at the moment. I have to go to Hell (also known as work) here in awhile and I\'m dreading every second of it. I work with a bunch of fire-breathing retards and because of that, I\'m on the verge of insanity. Don\'t be surprised if my next blog is written inside a jailcell. Do they get computers in jail? Probably. I bet they have nice Sony Vaio laptops. Fuckers.  A friend of a friend gave me a kitten a few days ago. I named him Jagermeister. All he does is eat, sleep, meow, and bite at my philanges. Philanges are toes and fingers right? Correct me if I\'m wrong.  I really need money for school. If I was still a virgin, I would sell my virginity on Ebay. Didn\'t someone do that? Uh, yeah....I\'m gonna go now. My hair is wet and the AC is on, so I\'m starting to feel like an ice-haired medusa. Adios ya\'ll.   '),(2530,3483901,'13,June,2004','2004-06-13 00:00:00','   For the past 96 hours I have been in Phoenix visiting my parents (I actually had to get a calculator out to figure that out-so much for college). I love seeing them and all, but damn am I bored. I think it\'s because I am stuck here and I don\'t have the choice of being social. From what I hear Phoenix is a happening place. Actually I didn\'t hear that, I\'ve just seen too many Elima-date episodes. Anyway, I flew so I don\'t have a vehicle. I could borrow my moms, but where would I go? I would probably end up at a bar, but I couldn\'t drink because I would have to drive home. Drinking and driving in my mom\'s Accord definetely don\'t mix (I would be disowned). So I\'d go to a bar, have a glass of water (maybe tea if I\'m feeling frisky), probably get hit on by some dirty homeless man, go to the bathroom, vomit, and proceed to go home. Sounds fun, let me find her keys.  I shouldn\'t bitch. My mom has done more than enough to entertain me. Pedicures, massages, facials, nice dinners, aloe vera for my wicked-one sided sunburn, and plenty of coffee. I appreciate it all, but I can only spend so many nights on the internet before I go crazy. What the hell am I talking about? I remember countless nights of staying up until 4am talking to god knows who. And I loved it. I wonder what happened? Let\'s see:(1)My computer gets shittier and shittier by the day (2)I turned 21 (3)I slowly lost touch with all my fellow internet geeks (4)I met a boy, fell in love, and moved in with him. I think that last one is the kicker. It\'s amazing what having a boyfriend will do to you, but that\'s a whole other story.  At least my parents have the fast internet connection. On my computer I can view maybe 10 websites, then the beast freezes up on me and I\'m forced to kick the shit out of it(note: Compaq\'s keyboards are the Tuff-Sheds of all keyboards). Because of all this, I have been looking to buy a new computer. I have been approved for a $1500 Dell credit card, I just don\'t know if I can really afford the additional monthly payments. I\'d probably have to spend $30-$40 a month, that means I would have to give up buying 2 t-shirts or about 13 fast food meals a month(hello Mr. Calculator again). I need to go on a diet anyway, so maybe this will be a little incentive.  Holy shit! I have just added up the total amount I have spent on fast food/restaurants since April 12 is $280.67. That\'s like $30 a day and that doesn\'t even include what I have paid for in cash. Screw the diet, I need to stop eating that crap to save money. So maybe I\'ll buy the computer and minimize the eating out. I hate Jack in the Box. I just reminded myself of a series of books I just read: Confessions of a Shopaholic and the two books after that, which I won\'t waste your time with the names. It\'s about a girl that has a serious shopping problem and finds ways to justify everything she buys. I feel like I just did that with the computer. I think I have the determination to actually do as I say. I hope so anyway. Actually, I just really want the computer. Maybe I\'ll get rid of my $30/month water cooler also. Maybe I\'ll get a second job passing out the hooker ads on the strip like I said I was going to. We shall see.   '),(2531,3483901,'11,June,2004','2004-06-11 00:00:00','   Okay, so since I am bored to death, I decided to try this blog thing. Normally I wouldn\'t do this, because I hate to write, but I had a stoke of inspiration...from an away message on AIM.  Not that anyone cares (I know I don\'t) and maybe I\'m just totally out of date, but it seems that the word of the year is Couture. I don\'t know what the hell it means, but I have an idea that it deals with fashion. I think this because I\'ve heard/seen it from two fashion-y people: Ashley (where I got my inspiration) and from that \"everything on me has been surgically enhanced\" old model biatch from that America\'s Next Top Model show. Anyway, I could look it up in the dictionary to see what it means, but during the summer I try to stay away from anything educational, so screw that.  Damnit, I forgot what the whole point of my point is. I need to lay off the wine. Um..........let me think. I hate myself right now for ever trying pot. It\'s true kids--marijuana effects the memory. My mom told me today that she thinks they should legalize all drugs, damn hippie. Just kidding, she\'s too young to be a hippie. I just realized that maybe its not the pot, maybe I have that adult ADHD. I saw a commercial about it and it just seemed to call to me. I bet thats how Ashley and face-lift lady feels when they talk about couture.  Aha!!!!Attention all potheads:::::Are you trying to remember something but can\'t because you smoke too much? Well, I have discovered the remedy:year old Carlo Rossi wine!!! I took a sip and instantly remembered the whole point of my point. Actually, this probably will only work for me because my point involved alcohol. Here goes....  So, why do I care what couture is? Well, as easy-going and non-shallow I try to be, deep down I am a money hungry bitch. Why the hell should I care what couture is? I know I\'m not it (or dress like it?), but I\'m curious because it sounds expensive. This goes along with other things in my life, like: I want a Louis Vuitton bag, not because I think they are cute, but because they\'re the thing to have; I always say that someday I\'ll own designer clothes, even though now I cringe at buying a $15 Roxy tank top; I want an Escalade, even though it is a gas-guzzling POS; I want to learn how to golf, even though in junior high we had a golfing unit and I always hit the ball backwards; and last but not least, I want to start hanging out at martini bars.   What\'s wrong with that, you may ask. Well, I hate any and all martinis. I hate gin. It reminds me of a party I went to when I lived in Steamboat Springs. There was nothing but a keg and a bottle of gin in the house. I was desperate to get drunk (a true sign of alcoholism) and I hate beer, so I bust out the gin. There was nothing to chase it with, so I poured some in this little toothpick holder that was shaped like a boot, bit the two ends off a Twizzler, put it in my boot and started drinking the gin through the twizzler. It tasted like pure ass. Needless to say, I got hammered and I think I ended up hooking up with some dirty dude with the nickname of Sloth (like from The Goonies). I\'m going to go throw up now.  So now that I have resurfaced some horrible memories, I leave on this note: I am a rich, bitch stuck in a poor girls body. Maybe some day I\'ll have all those materialitic things, but I\'ll also have my white-trash past to beat me back into reality.   It is time for a refill....   '),(2532,3483901,'30,July,2004','2004-07-30 00:00:00','   Summer is almost over and it\'s very depressing. I\'m not looking forward to spending all my time off from work at school again. I\'m really enjoy my lazy days off. Last night, I stayed up until 5:30am reading a book, then I woke up at 8, went to work for 1/2 an hour, came back home and passed out until about 2:30. It was great. I could spend my life doing that. Then I would get way fat. I need to go work out. Carissa, you need to come out here and make me motivated. You\'re all I have (tear!). I\'m just kidding. Also, you need to call me whore. I\'ve been trying to call for a couple of weeks now.  Anyway, it\'s a day later now. It\' s hard for me to focus on this damn thing. Like right now, I just woke up from a nice little nap, but I\'m still all exhausted and all I want to do is sleep. My dumb blood sugar has been really low now for a few days and its not getting better. I basically feel nervous, and every once in awhile I feel the urge to puke and then I can\'t breath very well and then I feel like I\'m gonna faint. I know I need to eat, but just looking at food makes me want to hurl. Maybe I\'m really just becoming anorexic and its all subconsious. Nah, I couldn\'t live without my chesseburgers (hurl now). Maybe all that eating has officially turned me diabetic. God, that would really suck.  Boo.....so Jess has been gone all week and I\'m sad to say, it has been nice. Hopefully he\'ll never see this, but I\'m really thinking its time we take a little break. We\'ve had a really good run, but I don\'t think either one of us is ready for the relation ship we have. I just don\'t want to break up with him because I know I would regret it for the rest of my life. He\'s so good to me, but of course my dumbass has to ruin everything. We\'ll see. We have one more month of living together, maybe things will get better.  Happier topic, the Devil (also know as my district manager) is stepping down from her position. Thank you Jesus! That woman is seriously the worst person I have ever met and I really wouldn\'t be all that upset if she got stranded on some deserted island (with plenty of food and water of course). I\'m hoping that work won\'t suck so bad now. Finding a new job is not easy. I\'m smart and semi-educated, but no one likes to hire college kids. Apparently, working around a school schedule is a really difficult task. Well potential employers, you can all fuck off because someday, I\'ll show you all. When I have my own super bank, I won\'t let you open an account with me. That\'ll show them. Yeah right.  So I have some eye candy at work finally. I would never date one of my employees or anything, but this guy in unbelievable. Totally Jamie. He\'s got spikey hair, gauged ears, and a labret piercing (which normally I don\'t like, but he looks good in it). He\'s a little odd, but fun to talk to. At least he could make work a little more interesting, nothing wrong with that. It will be like in high school. Most of the time, my only motivation to get out of bed and actually go to school was to see whoever I was pursuing at the time. It\'s like a reason to do my hair in the morning. That\'s horrible that I say this, it should be my boyfriend that gets me up in the morning, but no, it\'s not. I think that might be a good sign it\'s over. That and I was almost tempted to meet a guy I have been talking to on the internet for awhile. He was totally cool and I actually saw like a friendship developing, but I pissed him off (I\'m good at that) and he doesn\'t want to talk to me anymore. Oh well, life goes on. I\'m super cold, so I\'m gonna go. Bya.   '),(2533,3483901,'28,July,2004','2004-07-28 00:00:00','   Ugh....I\'m never drinking again. Last night I went to this club called the Beach. It\'s kinda shitty, but ladies get in free and drink free until 11, so I was stoked. Anyway, we got there at 10, so I had one hour to drink as many free drinks as I could. I ended up chugging four, then later had a couple shots. One of the shots was from those hot guys that walk around all buff like with a bottle of some sort of tasty poison(Ashley would uh, for lack of a better term, cream her pants). I had to get on my knees and he made me touch his ass. I felt like a white trash, 30 year old that couldn\'t get any action. I don\'t know why exactly, but the whole experience was traumatizing. Anyway, we danced for a few hours and my ass was still furshnickered when it was time to leave, so Karen took me home and let me pass out on her couch. The next morning I woke up and had no idea where I was for a few seconds. I must say, that is a pretty scary feeling.  So it wasn\'t a very exciting night, however I did have to suffer through the day with a really nasty hangover. It\'s 1 in the morning and I still feel like poo poo. We\'re supposed to try this new club at treasure island tomorrow, but my ass is dirt poor and I\'m not sure if I\'m ready to drink like that again. I guess I could play sober, but where is the fun in that? Sniff, sniff....i smell weed. My roommate must be home, maybe I\'ll go play. Nah, I\'d rather hole up in my room and not be social right now. I think it\'s actually my bedtime. I keep looking at the book I\'m reading and it\'s&nbsp;way more appealling than this blog. So live long and prosper (or something). Nighty night night.   '),(2534,3483901,'22,July,2004','2004-07-22 00:00:00','   I am such a loser. Everyone is at Gabe\'s tonight playing pool and drinking and stuff, but because I\'m wore out from work and I have to be back there at 8 in the morning, I stayed at home and got on the internet. Which isn\'t too bad. I am experimenting with my new computer&nbsp; (yea!!!!). Either way, getting old sucks ass.   Yesterday was Jamie gets cool stuff day. I got my computer and my camera in the mail and Jess finally got a security system and my CD player put in my car. I felt rich for like 2 hours yesterday. Then I looked at my checkbook and saw the whopping $25 in there and that forced me back into reality. Hmm...I\'m gonna be a nerd and upload some pictures of my kitty. If I can figure out how.   Ugh...nevermind. You have to download some program and I don\'t feel like doing that right now. I\'m gonna save this and go eat. I shall return soon....   ...2 Days Later  So much for returning soon. Jess just left for Colorado for a week. This is gonna be the first time he\'s left me alone for more than a few days in a long time. We\'ll see how that goes. I\'m gonna party it up man. It will be like being single, but not really. Like I can go dancing and not have to come home feeling guilty just because I danced with some dude in a drunken stupor. I can be freeeeeeeeeeeeeeeee.   Hey Ashley, I need to get your your blog site again. My dumbass forgot it again. Thanks.   Random though, yes. I hate my butt. I don\'t care how many black guys like it (and white guys too apparently), but having a ghetto ass is nothing but trouble. Seriously, like everyday at work, the subject of my ass comes up in some sort of way at least twice. My boss (who is black) might say something about me looking like a \"sista\" or my sales associate\'s roommate who says he would like to work at my store so he can watch my \"rump-a dump\" (or something) all day and the worst: my sales associate\'s boyfriend who wants me to come to their hottub so he can see my ass naked (sick!). I can\'t walk through the mall with some sort of dirty remark. \"Damn shorty, what are they feeding you?\" \"Pssshhhh girl, are you mix?\" \"Mumble mumble girl mumble mumble mumble ass.\" Or my ever favorite:\"Jamie, do you like chicken?\"   The very worst part about it is buying jeans. I can\'t even explaing the issues I have. I am 4\'11\'\' with a size nine ass and a size five waist. So when I buy my jeans, I have to buy a size nine. That means the waist is going to stick out like 3 inches in the back. Hello ass crack!!!! Even if I put a belt on, I can\'t catch a break. The waist gets all bunched together and plops out from under the belt. There are pants called Applebottle jeans by Nelly that are made for girls with big asses and little wasits. I checked them out and they weren\'t too bad. However, I don\'t mean to stereotype, but they look like jeans that black girls wear. I would like a tool wearing most of them. When I went into the store that carries them, the black girl at the counter looked at me all funny, then looked down to my ass, nodded and went back to watching the news (they have TV in their store, I\'m jealous). Either way, they are like $80 a pair. I\'m too poor for that shiat. I\'ll just let my crack hang out.   I saw another crazy person in Las Vegas today. We were all driving to Subway and as we were leaving our complex, there was this old lady with a shopping cart crossing the apartment exit. We decided to wait for her thinking that she was almost done crossing. Soon enough, we realized the lady was just inching along VERY slowly. It took her at least a minute just to cross the exit. As she was passing by our car, I could tell she was probably in her fifities. She was so skinny that she looked like one of those starving Ethiopian babies and her skin was burnt to a crisp. AJ said it best: \"She looks like the fucken cryptkeeper.\" It was quite entertaining.   So that\'s all for now. I\'m sure I\'ll be back \"soon\"   '),(2535,3483901,'20,July,2004','2004-07-20 00:00:00','   Dumby dum...burp! Nothing new lately, surprise surprise! Work is shitty and hot. The damn AC has been broken, so I\'ve been sweating up&nbsp; a storm. I\'m gonna have to start putting deoderant on my entire body just so I don\'t offend. Hmm....last weekend, Matt and Shanna came from Colorado to visit for a day or so and we all went to the strip to ride the monorail. I parked my car at the&nbsp;MGM, planning on taking the monorail to the Sahara, then walking to the Stratosphere, then taking the monorail to Imperial Palace, grabbing a yard margarita, and then walking back to the MGM. We each bought two train passes (back and forth) and started our journey.   We did the Statosphere thing and got back to the monorail station at 11:40 pm, just in time for a security guard to tell us that the last train had left because the place shut down at midnight. At first I thought he was kidding, but no, he wasn\'t. We were all pretty broke, so we couldn\'t take a taxi. The busses are nasty as all hell, so were forced to walk. It was late at night and we were in the ghetto. It was hot and I had been on my feet all day. So basically, we were completely miserable the entire 3 mile walk home.&nbsp; F-ing monorail.  However, some interesting things did happen. I found a cool little liquor store that had unique gift baskets. A crackhead lasy told me I looked like I was on meth because of my eyes. \"You look like you\'re in the Twilight Zone,\" says crackwhore. She then preceeded to ask Jess if he had papers and showed him a huge bud of pot wrapped up in a towel. There was another crackhead screaming at himself and a drunk man that tripped over the curb.   I\'m sooooo bored. My computer is supposed to come today, I just hope it comes soon. It will give me something to do. In the meantime, I\'m going to go play Simpson\'s Road Rage. That game is the shiat.   '),(2536,3483901,'13,July,2004','2004-07-13 00:00:00','   I am the laziest POS that I know. For the past three days I have holed up in my room playing Super Nintendo and shopping for a computer on the internet (I finally found one, it comes Tuesday!). I guess I did go to the strip yesterday. Jess and I went to Cheeseburger Oasis in the Aladdin(i think thats what it is called). Anyway, we each had a drink there and then we walked down to some ghetto hotel with those huge margaritas. We each got one and walked around Caeser\'s Palace. We were both wasted by the time we got home. We passed out, Jess puked like always, and later that night I was hungover. It was weird.  Hmmm, I get a free digital camera with my new computer. It\'ll probably be cheap and shitty, but who cares, at least I\'ll have one. I can take pictures of my super exciting life and post them for the world to see. I\'ll spice it up by throwing some S&M pictures in there (eww-what?).  Moving on. I\'ve decided that when I become rich, I\'m going to give out a scholarship to middle-class, average, white girls and whoever writes the worst essay wins the most money. I hate the fact that because I to write, I can\'t get any scholarships. The system sucks. Damn the man! Save the empire!    m     d  I need something to do. I\'m gonna look for a job. Peace out.   '),(2537,3483901,'08,July,2004','2004-07-08 00:00:00','   Today is a half day and I\'m excited. I don\'t have to go to work until 6pm. Granted, I would much rather never have to go, but a 4 hour shift is nothing compared to an 8 hour shift. Woo hoo. So my 4th of July was very interesting. I had my BBQ. It was fairly uneventful, except my future roommates new man got all psycho about his theory on why people do drugs. He thinks that people try drugs because they are lost and/or depressed. Fuck no, they try it because of peer pressure (cough cough Carissa-brings back good times in the smelly Pontiac). Whatever.   Anyway, after lighting some fireworks, we headed off to this bar called The Whiskey. It was the nicest bar I\'ve been to. The inside was like a typcial fancy smancy bar, but the outside was the coolest. There were beds and couches to sit on and some of the beds even had a little TV built into them. Most of the outside area was surrounded by one of those eternity pools. I had to wonder if any drunkass had ever fallen in. It was free to get in, but the drinks started at $6 a piece. Ouch.  So not long after we got there, Jess claimed that he wasn\'t drunk enough, so he went to get a double shot of 151. He was beligerant just a few minutes later. He started wandering around, bumping into things. Then he got angry and said I was the worst girlfriend ever and that I\'m always cheating on him. I just laughed it off because I knew he was drunk and talking nonsense. I was having a good old time with everyone else there, when all the sudden...dun, dun, dunnn....Jess blows chunks all over the sidewalk in front of everyone. It wasn\'t even a minute before security was surrounding us giving us the go home evil eye. So we packed our bags and dragged Jess\'s drunk ass out of there (which wasn\'t easy, the bar was inside of a huge casino).  I learned my lesson: never take my underage boyfriend to a nice bar until he learns to moderate his drinking. I would have to say this was the worst 4th of July ever. I\'ll get over it.  I need to eat and I have some coffee on the pot needing to be drank (I bet your jealous now, eh Carissa?) Peace out ya\'ll.   '),(2538,3483901,'04,July,2004','2004-07-04 00:00:00','   My cat is trying to kill me. My arm, my eye, and everywhere else on my body has been sliced to all hell. I\'m just one big, walking scab right now. It\'s sick. I forgot how much fun it was to have a kitten. He keeps trying to eat my pizza too. No way fucker, it\'s mine! I can\'t wait until the little shit gets declawed. My life will be complete again.  Today is a sad day. Instead of primping myself for a night of hardcore partying at Ashley\'s, I\'m trying to get motivation to clean my house for a little BBQ I\'m having tonight. Fuck cleaning, I want to fly out to Colorado. I wonder how much tickets are? Anyway, back to reality. After the food tonight, I\'m going to the Whiskey Bar in Henderson (I think that\'s what it is called). I hear it\'s nice, but drinks are expensive. I\'m gonna have to wear something skanky so I can get some free drinks. Oh wait, I think Jess is coming. Screw that idea. Owwwww....cat just bit my toe. I\'m going to go do my dishes right now. Maybe I\'ll think of something else to write while I do that.   Nope, nothing. I did find a giant cockroach though. I hate the damn things. They are huge and they refuse to die. You can step on one and they won\'t have a scratch. Rumor has it that if you chop of their head, they won\'t die. Thats some shit isn\'t it? Ugh...well I have cleaning to do and showers to take. Peace out ya\'ll.   '),(2539,3634853,'27,June,2004','2004-06-27 00:00:00','   Day 11 Recap    $1/$2 Limit HE        UltimateBet    -$11.00  PokerStars     -$15.00   $1/$2 Limit STUD    UltimateBet    -$27.00   $50 NL HE    UltimateBet    +$4.00    Total   -$49.00  -$59.10 for days projection -$14.84 for years projection  Summary  Well, I said yesterday that I\'d probably regret bitching about being ahead and now the poker gods have reached up and slapped me square in the face!  I truly can\'t get any cards right now.  While playing Stud, trying ot mix it up a little bit, I never stayed in a hand that I shouldn\'t have and only lost one on the river.  I just never got good starting cards.  As far as the limit Hold \'Em tables it\'s more of the same.  I vowed to myself when I began this that I wouldn\'t use it as a forum to complain about bad beats or bad players.  My soul intent was to evaluate myself and my feelings, and to keep a record of how my earnings (or lack thereof) were going.  In reviewing what I\'ve writtne lately it seems like all I do is cry about the cards I\'m getting.  I think it may be time to re-evaluate my starting hands, do some work in PokerTracker and see if I could be playing hands a little differently and find a way to make this venture more profitable.  I don\'t feel like I\'ve dug a hole that I can\'t get out of, but I think if I continue with the attitude I have now I can do no better than to play scared and lose any effectiveness I may have gained through study and analysis.  Now to stick to my guns...... '),(2540,3634853,'26,June,2004','2004-06-26 00:00:00','   Day 10 Recap   Sit N Goes    PokerStars  (1) $10 SnG     13/18        -$11.00   $1/$2 Limit HE        PlanetPoker    -$8.25  UltimateBet    +$47.00  PokerStars     +$3.75   Total   +$31.50  +$21.40 for days projection +$44.62 for years projection  Summary  It still seems as though I can\'t get any cards. Although I had a decent sessio at UB all in all I got very few cards.  Had I not caught quad 10\'s on the limit table on UB I would have been struggling there as well.  As it turned out, I flopped the quads and had 2 guys fighting over the pot so I just followed until the river where I made sure to cap and they both were playing an off pair from the flopped 10\'s.  Lucky me!  I made almost my entire result on that hand.  I think I\'m playing a little too tight in these limit games but will continue to try and make some adjustments and hopefully can begin seeing better results soon.  I guess I shouldn\'t complain as I am at least ahead of projections, but I feel a little underwhelmed none the less.  One day soon I\'ll wish I wasn\'t complaining.  Oh well, Until next time......  Enough for now.  Until next time..... '),(2541,3634853,'26,June,2004','2004-06-26 00:00:00','   Day 9 Recap   Sit N Goes     PartyPoker  (1) $10 SnG    4th/10        -$11.00    MTT   UltimateBet   (1)  $5 MTT NLHE    184/2402     -$5.50   $1/$2 Limit HE        PlanetPoker    +$34.88   $25 Pot Limit Hold Em   PartyPoker     -$3.30  Total   +$15.08  +$4.98 for days projection +$23.22 for years projection  Summary  I seem to be catching better cards at PlanetPoker right now and should probably be spending all my time there until things change, but I\'m not real hot on that interface and haven\'t figured out their tourney schedules yet.  All in all at least it was a positive day although I\'m giving away a lot on UB.  I played a large MTT tourney there last night and probably should have hung in until the money since they were paying 150 and I broke out at 184, but I was really playing for the win and I needed to make a stand before the blinds started eating me up and caught a bad break.  I played very well in the tourney though.  Maybe as aggressive as ever and just making sure all my good starting hands either paid well or didn\'t get sucked out by making sure people paid to draw against them.  I may try to play a little more limit tomorrow on all the sites and see if it isn\'t a bit more profitable than the SnG\'s.  Overall, I like the SnG\'s because I limit my losses, but if I\'m not catching the cards, as has been proven of late, I can still end up posting losses for the day and I just can\'t accept that.  The other thing is that I think I may have to vary my game selection a bit more as sometimes I catch myself feeling like I\'m just going through the motions and that just doesn\'t have a very positive expectation.  Enough for now.  Until next time..... '),(2542,3634853,'25,June,2004','2004-06-25 00:00:00','   Day 8 Recap   Sit N Goes    PokerStars  (1) $6 SnG    10th/18        -$6.50  PartyPoker  (1) $10 SnG    3rd/10        +$9.00    MTT   (1)  $5 MTT NLHE    39/733     -$5.50   $1/$2 Limit HE        PlanetPoker    +$57.04  Total   +$54.04  +$43.94 for days projection +$18.24 for years projection  Summary  I decided to give PlanetPoker a shot tonight to see if maybe a change of scenery would do me some good and apparently it did.  I caught some nice hands early and made them pay.  Then I stuck around until I felt like the cards were turning and got out booking a nice positive session.  My play hadn\'t really changed but I seemed to get some better starting hands and it worked out.  I will continue moving around and see if I can\'t maximize the good sites when I catch them while trying to keep my losing sessions a little more under control.  On to the weekend, let\'s see what happens. '),(2543,3634853,'23,June,2004','2004-06-23 00:00:00','   Day 7 Recap   Sit N Goes    PokerStars  (1) $10 SnG    10th/18        -$11.00    MTT   (1)  $3 MTT NLHE  PS     98/1448     +$3.48  Total   -$7.52  -$17.62 for days projection -$43.32 for years projection  Summary  Not much to say.  I have changed up a bit but am being careful at the same time and thinking about some other things I can do to shake things up.  I haven\'t played Stud in a while, maybe that would be the ticket. '),(2544,3634853,'23,June,2004','2004-06-23 00:00:00','   Day 6 Recap   Sit N Goes    PokerStars  (1) $10 SnG    15th/18        -$11.00   PartyPoker  (1) $10 SnG    4th/10        -$11.00   UltimateBet  (1) $5 SnG    4th/10        -$5.50     MTT   (1)  $5 MTT NLHE    388/738     -$5.50  Total   -$33.00  -$43.10 for days projection -25.70 for years projection  Summary  Things aren\'t getting any better as you can see.  I just don\'t think I\'m getting any starting cards and now when I do I\'m either already so far behind that no one is bothered by my play or they come over the top with little or no fear.  I think I may be a bit to patient and may try opening up a bit to see what happens.  I just hate changing things when I don\'t have a lot of bankroll room to absorb things if it doesn\'t work out.    Guess I\'ll just give it some thought and see how things go.  Until next time..... '),(2545,3634853,'23,June,2004','2004-06-23 00:00:00','   Day 5 Recap   Sit N Goes     PokerStars  (1) $5 SnG    6th/18        -$5.50 (1) $6 SnG    4th/18        +$4.80  Total   -$.70  -$9.40 for days projection +$17.40 for years projection  Summary  Seems that the cards have dried up.  I wasn\'t able to get anything to start with in either tournament and only lasted as long as I did in the $6 one because I was ruthless in trying to steal blinds.  Oh well, mom said there would be days like this :) '),(2546,3634853,'23,June,2004','2004-06-23 00:00:00','   I haven\'t updated my results in 2 days, but will most definitely do that tonight when I get home from work.  Although I know better, my thought processes really suck right now as I haven\'t been getting much for starting cards and when I do have a hand they aren\'t holding up.    At one point I showed my wife where for 7 hands in a row I had 2xos with nothing higher than an 8.  Right after that I asked her \" now at some point don\'t you think I should be getting AX suited 5 or 6 times in a row?\"  She laughed out loud.  It kind of made me think how frustrated I was sounding, and probably silly too, but at the same time, I haven\'t seen yet where you get a good streak that lasts anywhere near as long as the bad ones.  I\'m hoping that soon I catch one of those \"hot\" streaks that you\'re able to sustain, but it hasn\'t happened yet.  Anyway, I plan on going home this afternoon and doing some analysis in PokerTracker.  I want to make sure there aren\'t any loops I\'m opening up as a result of how things have gone.  Also, I want to make sure that my study and reflection habits remain in place regardless of  my short term results.  I still think that the goal I set is attainable and look forward to the progression of my game.  I also still believe that with diligence, commitment, patience, and discipline I can bring my game to at least another level.  That next level should get me beyond the casual or undedicated player and help me to supplement my income as I am hoping.  Enough crying for now '),(2547,3634853,'22,June,2004','2004-06-22 00:00:00','   Unfortunately, I wasn\'t able to pull myself to the computer last night to do my normal recap post.  Maybe that\'s a good thing, but I felt like I was letting at least myself down as a result.  Anyway, I played 2 Sit N Goes on PokerStars.  One was a normal $5 SnG and the other was what they call a Turbo SnG with the blinds escalating at a much faster pace, every 5 minutes.  That one cost me $6.  I bombed out of the first and ended up 4th in the turbo for a total loss of roughly $3 for the day.  I will post accurate numbers this evening which will reflect where I am towards my goals.  In addition to those I played a freeroll on PS for an entry into a tourney on Sunday with a guaranteed $300K prize pool.  This cost me 500 frequent player points and there were 206 entries which meant that they would be giving 4 entries into the big tourney.  I ended up in 15th place, but was really never a factor after the first hour as I was simply playing catch up the rest of the time.  When it was all over I had looked at 18 flops in 187 hands and had won 15 of them.  The cards simply weren\'t there.  Then at the end I had KK cracked by a flush.  Oh well, it was interesting and a nice change of pace.  I really just wasn\'t very interested in working my game yesterday.  I probably would have been better off reading a little bit and giving myself a break, but sometimes find myself feeling obligated to at least giving it a shot.  In the long run this attitude would get me in a lot of trouble, I\'m sure, but I generally only sit down if I\'m focused on building my bankroll and playing good cards.  It was another good lesson and fortunately not an expensive one as most are.  I will just need to keep this in mind the next time I sit down to play when I don\'t really feel like it and pull myself away and find other ways to occupy my time that can help me on my quest.  That\'s enough rambling for now.  Until next time..... '),(2548,3634853,'20,June,2004','2004-06-20 00:00:00','   Day 4 Recap   Sit N Goes    PokerStars  (2) $10 SnG\'s  1st place  +$61.00                                                      7th place  -$11.00   MTT\'s    UltimateBet  (1) $5 NLHE   53/656 for  +$2.70  Total +$52.70  Total for the week (4 days since I started mid week)  +$67.20 --> projected for $40.40   +$42.60  for days projection  +26.80  for years projection  Summary  I had a good start to the day by winning a $10 SnG on PokerStars.  It\'s been a long time since I had the feeling, but from the very start, although I didn\'t see my first hand for a while I felt like I was dominating the tables I was on.  I don\'t think after I made my first move with KK to a rag flop that I was ever truly threatened.  I came at any pot I wanted to play strongly and normally didn\'t get much action but it allowed me to steal some blinds and get to a position where I stayed just as aggressive as I thought I needed to stay in control.  The only real trouble I had was with the guy I ended up Heads up with as he was very careful and wouldn\'t put his chips at risk and he was only behind me by about T2500 chips.  I decided the only way I was going to get him was to try and trap, and eventually got my chance, slow played a flopped straight, and knocked him out.  After that win I found another tourney and registered but really never caught any cards.  I ended up in the middle of the pack, but did so only by being very patient, but finally got to a point where I had to make a stand and pushed all in with QQ and got caught by AA.  I took quite a break through the middle of the afternoon. Partly because of Father\'s Day but as I started thinking about it, also because I wanted to protect my win for the day.  That\'s probably not the best thinking all the time, but at this early stage of the building of my bankroll, I was able to justify it.  Anyway, after dinner I decided to try a Multi tourney at UltimnateBet.  The entry was only $5 so I wasn\'t risking a lot, but had potential to make a big hit if I got some cards.  On top of that, UB hasn\'t been very good to me lately and I\'m just not running very confident there, so I thought this might be a good time to try to shake the cobwebs.  I ended up placing 53rd out of 656.  It only netted me $2.70 so my BB/hour rate won\'t be skyrocketing, but it was nice to at least finish in the money.  I truly have no idea how I lasted as long as I did.  At one point I had been moved to a table and only saw one flop in 48 hands!  That was the last before they moved me and it turned out to be a blind steal, but it certainly tested my patience.  Anyway, I remained patient and eventually got a couple of hands, was able to make them stand up and doubled up at least enough to get into the money.  I ended up getting busted out with AKs and a Kxx flop, when pocket QQ\'s caught a straight with runner runner.  Such is life :)  Anyway, I\'ve gotten myself back over the hump and am feeling somewhat confident, although I want to guard against being cocky.  I may have also shook the UB feelings of late so this day could have had more positives than at first glance.  I guess we\'ll see.  Until tomorrow...... '),(2549,3634853,'19,June,2004','2004-06-19 00:00:00','   I\'m posting this one a little early as I\'m hosting a small home tourney this afternoon so I believe this will be all I can play online today.  Day 3 Recap   Sit N Goes    PokerStars    (1) $5 SnG no cash   -$5.50  18th place :(   PartyPoker    (1) $10 SnG  1st place  +$39.00    Limit Games    PokerStars    $1/$2 Game   -$22.75   UltimateBet    .25/.50 Game  +$1.25  Total  +$12.00   +$1.90  for days projection  -$25.90  for years projection  Summary  I had a rough start to the day when I played the SnG at PokerStars and ran into KK when I had QQ.  I raised coming into the pot about 5xBB and was just called by 2 people.  I didn\'t have any notes on anyone and no real reads so I probably jumped the gun, especially that early in the tourney, but when the flop brought rags but 2 of a suit I thought I wanted to make someone pay to draw to the flush.  Oh well, another lesson learned I hope.  After that I played about an hour of limit at UB and was up and down the entire time and when I had a small positive balance after that hour I decided to take a break.  I came back a bit later and played limit on Stars and never caught a break.  With AQ suited I had a flop of AAK, raised the pot and another guy capped.  Turned out he was on KK and never raised pre-flop and caught me off guard.  That was most of my losses there.  When I couldn\'t get anywhere after about another 1 1/2 hours I moved on again.  This is where I played the SnG at Party and early on I wasn\'t catching any starting cards at all.  I managed to steal a couple of blinds and just tried to stay patient.  Soon my cards changed.  I was getting great starting hands and any time I had a chance to put someone all in I did it and all but one of those hands held up.  I ended up knocking out 4 of the last 6 players and had such a large chip lead on the last 2 players that it was just a matter of me playing smart heads up and I won.  Overall, I didn\'t make a lot of money but was at least able to post another positive outcome and knocked my years deficit down a little bit. I think if I continue being patient, studying when I have time, and catch my share of cards that I should be able to have a realistic chance of maintaining my goals for year end.  Now if I can just do something good tonight in our home game.  If it\'s not too late and I\'m not wiped out, I\'ll try to post a report later.  Until tomorrow..... '),(2550,3634853,'18,June,2004','2004-06-18 00:00:00','   Day 2 Recap   Sit N Goes    PokerStars    (1) $10 SnG        3rd place   +$25   PartyPoker   (1) $5 SnG           1st place   +$19   Limit games    UltimateBet     -$13  Total   +$31.00   +$20.90  for days projection  -$27.80   for years projection  Summary  I had a better day of cards today.  Overall it still seemed as though I wasn\'t getting many cards.  When I looked it over in PokerTracker I had only played 11% of the flops and that kind of proved what I was thinking.  Anyway, I was able to make what I did get pay off and thought I maximized hands fairly well.  One thing much different today was that when I needed to push all in and I had a good hand, generally they stood up.  Much different than yesterday.   Anyway, I was able to post a positive day.  It wasn\'t huge by any means, but if I can continue chipping away then soon I\'ll be back on track.  Until tomorrow :) '),(2551,3634853,'17,June,2004','2004-06-17 00:00:00','   Day 1 Recap   PokerStars    1 $5 SnG\'s  no cashes  -$5.50                               1 $10 SnG   no cash     -$11   UltimateBet   1 $5 SnG   no Cash       -$6   PartyPoker    1 $5 SnG   no cash       -$6   Total    -$28.50   -$38.60  for days projection  -$38.60  for years projection  I was stung by the bad cards all night.  I never had a run of good cards and as we neared the end, I never was fortunate enough to catch when it was time to try to defend to stay in the tourney.  On UB & Party both tourneys I finished 4th, one spot from the $$, On PS I made it to the final table each time, since the SnG\'s I play there are 2 table, but ended 6th & 7th, with cash at the 4th spot.  Although my first day was less than spectacular, I basically played what I thought was good cards, remained patient as long as I could and didn\'t overplay hands and get caught.  I was able to make a few blind steals in position and never played a weak or suspect hand out of position.  I\'m sure I\'ll catch this up when the cards turn but it gets frustrating sometimes waiting for that ot happen.  It should also be noted that I didn\'t play these tournies back to back.  I played 2 after coming home from work, ate dinner, helped my sister-in-law try to buy a car, and then played 2 more.  Tomorrow is another day! '),(2552,3634853,'16,June,2004','2004-06-16 00:00:00','   I\'ve decided that to be successful at this I must first set a goal.   If I\'m totally honest with myself, and log each and every days results, I should be able to see how I\'m doing toward my goal, but should also be able to see where I\'m most succesful and where my play needs work.  Some of this may be difficult to track at first, as my interests are so varied and my ultimate goal is to be at least proficient in many different games, but I will do my best to keep this journal understandable, easy to read, and most importantly, functional to me in my assessment of my play and progress.  Since I am starting with such a limited bankroll, I will primarily be playing Sit N Goes (SnG\'s), some $1/$2 Limit tables, and an occasional $25 NL table.  These are all Hold \'Em.  Additionally, on rare occasions, which I use to primarily change the pace, I will play $1/$2 Stud games.  Each of these will be reported seperately with a days recap at the end.  Most of the SnG\'s will be either $5  or $10, and occasionally I will venture into a MTT, but only as a change of pace and to give myself some experince in that arena.  Now, to the nuts & bolts!  If I begin this quest today there are 198 days left in 2004.  Since I don\'t want to put too much pressure on my initial phase, I am going to shoot for a $2000 BR at year end.  This would mean that my daily average would need to be $10.10.  If I \'m not able to play for a couple of days that doesn\'t matter as this is a daily thing and I have figured in a low expectation so that I can absorb difficulties in getting to the table as well as the totally expected down times.  The quest begins! '),(2553,3634853,'16,June,2004','2004-06-16 00:00:00','   I played in a $75 + $15 NLHE tourney last night at Holywood Park.  This is a continuation of their Endless Summer series but with a different format.  It used to be a lower buy in with re-buys and now they\'ve increased the entry and eliminated the re-buys.  Last nights turn out was a little lower than I\'ve seen in the past but we still had 156 people.  I caught a couple of unexpected hands early.  In the BB with no raies in front of me I had 36s. The flop came 663!  I slowed played the flop and the other guy in came at me with T100.  since we start this with T800, I thought if I could keep him thinking he had the lead I may be able to pick up almost 25% more chips here.  He came at me again on the Turn when a K hit the board and I smooth called one more time.  A rag hit the river, there was no flush or straight draws and once again he came at me with T100.  This time I raised to T300 hoping he would think I was trying to bluff, but he folded there and I took the pot.  I managed to steal a couple of blinds and basically played tight through the first hour. Then remembering what I had read about trying to make something happen just before levels increased or going on a break, with about 5 minute left before the 1st break, I caught AKs from a late position.  I came in with a raise of 3XBB and was called by one guy. The flop brought KQ8 rainbow.  The other player checked, I came in for about T250 and he smooth called me.  The turn brought a rag so I thought if he checked again I was going to put him to a decision since he only had about T400 left, and I was at about T2100.  sure enough, he checked and I pushed all in!  He was on QQ and I got no help on the river.  This put me in a postion where I had about T600 left and just a couple minutes before the break.    On the very next hand I get QQ in the hole, the flop comes 789 rainbow and once again I\'m in a raised pot with the same guy I had just lost the last hand to.  I figured he thought I was probably tilting a bit ( I don\'t really think so though) and also that since he had just popped me, I was sure he was probably playing this hand a bit looser.  I pushed all in again, expecting him to fold to the straight possibility and he calls!  He has TT and sure enough the river gives him a 6 and I\'m out.  I didn\'t check the board but I think I was right about 100th.  Not a very good showing, but overall I think I played my hands as well as I could and caught a tough break. '),(2554,3634853,'14,June,2004','2004-06-14 00:00:00','   I have just opened this Blog and am not sure at this point how things are going to work, but thought it would be fun and maybe interesting to give it a shot.  I am hoping to share the site with friends and family and hope to hear from others when something I\'ve posted is of interest or particularly funny to them.  Anyway, I hope you enjoy and be easy on me. '),(2555,3634853,'03,July,2004','2004-07-03 00:00:00','   Last 5 days have been an up and down affair as usual. It just doesn\'t seem that I can sustain an upward swing, but am working the numbers and doing some reading and studying and hopefully can keep my head above water until I get this figured out.  Monday 6/28        -$3.00 @ Planet PLHE table  Tuesday 6/29 PokerStars       -$11.00 SnG +$10.25 $1/$2 Limit HE -$.75 total  -$15 PLHE  PartyPoker       Day total  -$15.75   Wed. 6/30 Stars  - $11.00 SnG -$3  Multi  Planet  +$23.75  $1/$2 Limit      Day total  +$9.75   Thursday  7/1      Party -$3.75 PLHE  UB   -$11 SnG  Day total  -$14.75   Friday 7/2         Planet   $1/$2 Limit HE   +$15.25 UB       $10 SnG          +$28.00 Stars    $1/$2 Limit HE   +$13.25 Party     PLHE            -$3.85               Day total  +$52.65    Total +$28.90   -$21.60 for the 5 days projections -$36.44 for the years projections  Summary  I am becoming more and more patient but also looking for opportunities to open things up a bit in the Limit games.  I think it\'s working, but will hold off until I have a few more sessions results to evaluate.  Playing the PLHE tables is nice for the variey as it is a much different game than Limit and I seem to be doing ok there.    I will trying to play a few more SnG\'s at UB this month as they have what looks to be a very cool promotion where they are giving back 1/2 your SnG buyins in Bonus points.  It takes forever ot release them, but free is free!  This format is still being worked out and I hope it becomes cleaner, and more interesting as we go along.  I will probably being playing off and on over the long 4th of July weekend and hope to once again get on the plus side of my projections and start seeing some real progress.  Until next time...... '),(2556,3634853,'01,July,2004','2004-07-01 00:00:00','   I\'ve decided that instead of daily updates on what I\'m doing playing poker that I will probably just do weekly updates and then maybe a month end recap.  I keep pretty concise logs in other ways of everything I\'m doing and it gets a bit redundant to re-hash it every day.  Besides, for anyone looking at this thing, it would get pretty dry if I continueed, I think.  So, it looks like I\'m going to go see Fahrenheit 9/11 this wekend with Dave.  I\'m not really into Michael Moore, because I think that although he brings up some great points, he also distorts the facts terribly and the genreal public doesn\'t always see through the BS.  Then I have to come to work and listen to these people talk about what a great movie it is, because it shows how mislead we\'ve all been and what a terrible job Bush and his cronies have done.  It took this movie to point this out?  Where the hell have you been?  I can\'t remember the last time I didn\'t vote Republican across the board, but I can truly say that won\'t be happening this November!  If I can see it , then those that are looking for issues shouldn\'t need Michael Moore to point it out for them.  I got a kick out of the news today about Saddam Hussein being brought into court.  Almost everything I read said something about it being a ploy by Bush to boost his ratings.  What about the people this man has kept in fear all of these years?  Any chance that may have something to do with it?  I did see one person come forth and say that they were looking forward to their chance to testify against him, ONE!  I\'m really getting tired of living my life in fear.  Every day another person is shot down or run over or beaten to death, all becasue we no longer remember how to deal with adversity.  If something doesn\'t go our way, hell, we\'ll just make someone else hurt even more.  What kind of logic is that?  Then I keep getting these letters in the mail from the support people back in NY.  If the god damned governement spent as much time trying to break down the gangs and putting criminals away, as they do pestering me about whether or not my kids all have sufficient medical insurance, this country would be back on it\'s feet in no time.  Each and every one of my kids receive support, have medical insurance, and are as well taken care of as possible, yet I still have to go to my benefits people at work and resubmit tons of paperwork so that they can be sure.  What a bunch of bullshit! '),(2557,3634853,'23,July,2004','2004-07-23 00:00:00','   Well, I was having trouble posting thigs to my Poker Journal because I\'ve been in such a funk lately, so I thought I\'d try a different format here and see how things go.  Sometimes all I need is an opportunity to vent.&nbsp; This blog won\'t be very specific, just things going on around me and thoughts I\'m dealing with.&nbsp; One of the first things you should probably do when putting one of these together is consider your audience, and I\'m here to tell you that I have done that one bit.&nbsp; :))  Anyway, let\'s see where this takes us.   '),(2558,4184832,'21,July,2004','2004-07-21 00:00:00','   ... is Charles William Wikle III   '),(2559,4184832,'21,July,2004','2004-07-21 00:00:00','   From NatalieLife, Will\'s Friend:  \"The tattoo is the the initials MP. They are the initials of one of his best friends that was killed a couple of years ago.\"   '),(2560,4184832,'21,July,2004','2004-07-21 00:00:00','   Rumour has it that Will and Holly are cousins... could it be true?   '),(2561,4184832,'30,July,2004','2004-07-30 00:00:00','   **This post contains spoilers**  Will\'s alliance was spared a major blow on Thursday night when, by a vote of 7-1, Holly was evicted from the house.&nbsp; An endurance Head of Household competition ensued.&nbsp; Will made it to the final three, but after getting promises of safety from Jase and Diane, he voluntarily stepped down.&nbsp; Diane went on to win the competition many hours later, which means Will really is safe this week (along with Adria/Natalie, Nakomis, and Karen).   '),(2562,4184832,'27,July,2004','2004-07-27 00:00:00','   Want to join up and post on this fan site? Just  urlLink email me  and I\'ll send you the instructions on signing up! :)   '),(2563,4184832,'07,August,2004','2004-08-07 00:00:00','   It\'s genius! Nominate 2 decoys from the GPA, then only GPA get to play in the PoV. One of them wins, takes off one decoy, and up goes Jase - who is then easily voted out. Smart thinking!   '),(2564,4184832,'13,August,2004','2004-08-13 00:00:00','   Yes yes, it\'s a sad day for Will fans everywhere as he has been nominated, even though as just a pawn... And to add on to what the wonderful Spamgirl (Hey Spams! :D) has previously said, everyone send good thoughts Will\'s way that he will be able to either pull himself off the block with veto or avoid eviction...   '),(2565,4184832,'10,August,2004','2004-08-10 00:00:00','   Will is nominated :( He\'s supposed to be a pawn, but who knows what will happen...   '),(2566,4184832,'10,August,2004','2004-08-10 00:00:00','   It\'s looking as though Will may be next week\'s \"Pawn\" Nomination...  cross your fingers!   '),(2567,4184832,'10,August,2004','2004-08-10 00:00:00','   There seems to be another twist heading towards the Big Brother house- is it going to involve Will?  Only time will tell, but for now, it sure is fun to speculate :)   '),(2568,4184832,'20,August,2004','2004-08-20 00:00:00','   Do you like Will?  You\'re definitely not alone!  According to CBS.com\'s \"Love \'em or Leave \'em\" poll, Will is now the most liked Houseguest in the house, far ahead of all the other houseguests, with a rating of 9.0.  You rock, Will!!!   '),(2569,4184832,'19,August,2004','2004-08-19 00:00:00','   The other night I was watching the live feeds and Will and Nakomis were talking about their dream guys.  A clip of it was even shown on Thursday night\'s show.  What is Will\'s dream guy?  He is skinny, with tattoos and chin length greasy, brown hair.  Nakomis agrees that they like the same type of guys.   '),(2570,4184832,'19,August,2004','2004-08-19 00:00:00','   Well well well... there sure are many strong opinions on Will\'s grand exit this evening!  Many people feel his parting words were childish and assinine, etc etc...  I personally commend Will for the way he went out.  He didn\'t cover his feelings with a false \'lovey-dovey\' type goodbye- He was feeling raw towards both Adria and Natalie and had no wish to tell either of them goodbye...  I\'ve read that Natalie didn\'t deserve the cold shoulder she got from Will; however, remember back to when Adria and Natalie were discussing who to nominate for eviction...  it was Natalie who first brought up Will\'s name...  I said it once, I\'ll say it again- I commend Will for an honest exit from the Big Brother house. :)   '),(2571,4184832,'17,August,2004','2004-08-17 00:00:00','   In a stunning (that is, if you don\'t have the live feeds) turn of events, Will has been evicted by Adria\'s tie breaking vote.  Having previously assured Will time and again that he was safe, this vote did little to help Adria\'s falling stock in the house...  Shortly after telling Will of her vote, Will refused to look at Adria and would not hug Natalie.  His last words in the Big Brother house were to Adria:  If karma is a boomerang, I\'ll be seeing you sooner rather than later.   During his interview with Julie Chen, Will refused to even look at Adria\'s (Natalie\'s?) goodbye video, before bursting into tears at Karen\'s.  He wished Karen and Nakomis the best of luck before heading out to the Jury House.  Adria\'s vote that evicted Will did very little to help her plunging stock in the house.  In a \'sweet-revenge-esque\' turn of events, Nakomis went on the win the Head of Household for this week...  Who  ever  will she nominate??   '),(2572,4184832,'17,August,2004','2004-08-17 00:00:00','   It seems the winds of change may be coming back to the Big Brother house... Diane has apparently realized that it is in her  alliance \'s best interests to evict Will this week, however, it is in  her \'s (and consequently, Drew\'s) best interests to evict Marvin this week, citing that the twins will only get strong if Will is gone... Diane didn\'t feel it right to blindside Will with an eviction on Thursday night has (graciously) given Will a warning about the plan to evict him, which is not making Adria, Natalie, and Cowboy too happy... Diane has said she plans to try and talk to Drew and sway him to vote for Marvin, also... it\'s true things can change on a dime in this house, but for now, at least there is some hope that Will is going to stay...  Keep those fingers crossed!  (And your toes, too, just for good measure ;-D )   '),(2573,4184832,'16,August,2004','2004-08-16 00:00:00','   As it seems more and more likely that Will is going to be evicted on Thursday night, I\'ve been thinking about Big Brother, the game, etc etc...  And here\'s a warning- I\'m probably gonna get pretty sappy in this post...  Very rarely do the \'good ones\' make it to the end and win, (Case in point: Alison and Jun- the house as a whole did not care for either of them) so really, it very well may be Will\'s time to go.  He\'s made the jury, anyway, so he will be sticking around for the rest of the game in the jury house.  (And with  Project: DNA , who knows what will happen?)  Really, I\'ll be sad if he\'s evicted, but Nakomis and Karen will still be in the house to \'avenge\' his eviction- Will won\'t have to be in the house when things start to get really ugly...  All in all, I\'ll be okay if Will is the one to go this week (I know other fans of BB5 and Will that have said they will be wrecked if he\'s gone this week and may stop watching the show, even!)- regardless, I\'m glad he was on the show.  Watching him, and his commentary, on BB5 has been a blast and I\'m glad to be a fan of his.  That said, Best of Luck to Will on Thursday- who knows?  Maybe it will finally be Marvin\'s time?  This  is  his third time in the eviction chair...   stranger things have happened...   '),(2574,4184832,'15,August,2004','2004-08-15 00:00:00','   Highlight to read...   Adria did not use the veto.  And sadly, Will still seems to be the target of the \'Twisted Twins\' Alliance...  Things are not looking good folks, things are not looking good...  but then again, who would have thought Adria would have stayed over Holly, especially when the Horsemen were in power?  Stranger things have happened- keep those fingers crossed!  -  - End of Spoiler   '),(2575,4184832,'15,August,2004','2004-08-15 00:00:00','   Here\'s a luxury competition spoiler-  Will (along with Drew and Diane) won a screening of  Without a Paddle  last night.   -End of spoiler.  Sadly, what Spamgirl has reported is indeed true- it\'s starting to look like Will\'s days are numbered... :(  Granted, they (The Twin Alliance) are still discussing possibly putting up Karen or Nakomis in place of Marvin to really break up the alliance, but for now, things aren\'t looking too bright for Will- as things stand now, he could pull off a 3/3 tie (with Karen, Nakomis, and Diane voting for him to stay), but that would leave the tie breaker with Adria...  (The twins know Diane is playing Nak/Karen/Will and have accepted that she will vote with them)  It\'s interesting to note that Will seems to think Drew will never vote him out- let\'s hope that holds true!  Stranger things  have  happend (Remember Adria vs. Holly?  Whodathunkit? :D)   '),(2576,4184832,'15,August,2004','2004-08-15 00:00:00','   Rumour in the house is that the Twin Alliance, Adria/Natalie/Drew/Diane, is gunning to get Will voted out of the house! Let\'s hope Will can pull something to save himself!  (Hi Roomie :) )   '),(2577,3038828,'29,March,2004','2004-03-29 00:00:00','    I am so woozy right now. we had a night out at viper room. Had beer + chivas coke = overkill!   Yesterdayâ€™s dinner at chiliâ€™s (with kak rose and pretty baby Isaac). We all passed around plates and shared the love:-PPP  Bbq beef ribs Rib eye steak with mushroom sauce Classic nachos (with sour cream and guacamole) Pints  At viper, I was recruited by amir as his unofficial spokesperson/cheerleader/assistant in ass-kissing. We sat with his boss gary and we both beguiled him with charming conversation (sigh). Kiss ass, ass kiss, kiss ass. Gary got pretty expansive and starting blowing his own horn pretty hard (why are drunks who get buttered up get really full of themselves?? Happens all the time! Sheesh!) as we left, khairul and his brother came. Pulled him to a corner and proceeded to shred his ex to pieces.   I still feel bad about that. Putting ann above him and berating him for leaving her. Boy, now I know why he did.   '),(2578,3038828,'26,March,2004','2004-03-26 00:00:00','    Amir and I hardly have time for each other these past few days. After his shower and dinner and joint, he would (literally) pass out in bed. And Iâ€™m left all all alone with wati: I might as well be all alone! I miss my baby so much. But when I see him asleep with his mouth open and tousled hair all over the bed, my heart goes out to him and I make a secret promise to not complain.  Evidently itâ€™s working: heâ€™s been calling me like 5 times a day at work every day now:-D   Last nights dinner: Mixed salad (butterhead, aragula and mustard greens) with thousand island Carbonara with spaghetti (Iâ€™m sick of fettuccine!)  Mango juice 3 plump little joints  hehehe. A pothead I am. I am a pothead.   '),(2579,3038828,'24,March,2004','2004-03-24 00:00:00','    dinner with mama yesterday:  starter: baked mussels with toppings (cheese, basil and bacon. Mmmm.) me: margherita pizza mama: roasted lamb shanks (oh gawd) amir: homemade fettuccine with cream marinara sauce   I ended up eating only two slices and ate up half of mamaâ€™s and amirâ€™s dinner. It was so good! Was there room for dessert? Naw, but there was room for the sticky icky icky when we got back.  Iâ€™ll be manually checking each and every page of my company site revamp. Oh gawd. Iâ€™m so sleepy.   '),(2580,3038828,'23,March,2004','2004-03-23 00:00:00','    blogg? ok!  i\'m hangovering at 9.49 am and just registered to blogspot. i\'m dying for a couple more hours of sleep, but what\'s a devoted girlfriend to do. amir needs to start his new job at 9, so i\'m here by 8.30. dayum!  will be working on the kemuning hospital medical card brochure in a minute. in a minute!   '),(2581,3038828,'30,April,2004','2004-04-30 00:00:00','    anak kucing misses jebon BAD.   '),(2582,3038828,'29,April,2004','2004-04-29 00:00:00','    diana was lost.   she frantically looked left and right, right and left. she took a few hesitant steps forward and then froze with fear. she heard stories about lost people walking about and complicating things further.  goddamn fuck it! she swore to herself. pukimak harrrraaaam. what the fuck do i do now?   she quickly reached into her little guess handbag and fished out her tiny nokia. the batteries? full. she exhaled a little. the connection? zilch.  feeling her anger grow steadily into a fullblown hissyfit, she threw the goddamn phone to the ground. the item split in two and she stepped on it, twisting her ankle to mash the poor phone to pieces.   she suddenly heard a familiar voice next to her.   \"sayang! where were you?\"   \"where were  you ?\" she spat to her stunned boy. giving him a withering look that lost its power because she constantly used it. she gave a little sniff of contempt and walked off.  \"hey, hey, yang! nanti sesat nanti!\"  \"see if i give a flying fuck!\" she yelled.   the boy looked at her disappearing form, and sighed deeply.   he went back to the girl waiting for him at a close distance. she had watched the whole thing while sucking on a coke from mcd. her other hand held a cigarette.  \"so? apa cerita?\"   \"dia tak nak jumpa you la, yang. my sister memang macam tu... tak tahu apa problem dia. i asked her to wait for us for just 5 minutes and she stormed off just like that.\"   ***  and that, my friends, is my lame attempt at irony fiction. yucccch.   '),(2583,3038828,'29,April,2004','2004-04-29 00:00:00','    mmm... and yeah. what about the paintings? ahaha.   i\'ve been plotting and planning to draw up a pretty underwater painting. i don\'t know how the hell it\'s gonna turn out.   sleepy.   '),(2584,3038828,'27,April,2004','2004-04-27 00:00:00','    it\'s only thursday. 1 more goddamn weekday to go and it\'s all good! yay cuti!  i\'ve been planning oh lots of things to do for the 4 day break. ima cook up a storm, buy nice wine, sleep a lot, drag amir to my favorite bookstores a lot, kiss a lot and hang out a lot:-)  cooking itinerary saturday : mee kari sunday: steak frites / barbeque chicken monday: rice & rendang & bok choy tuesday: fajitas (i wish)   my mother will be going through some scanning in putrajaya tomorrow. please God, do be kind.   '),(2585,3038828,'27,April,2004','2004-04-27 00:00:00','    i have finally decided what i want to do with my life!   i want to paint.   huge, acryllic abstracts of everything i love: the sea, cats, ciggies, a glass of coke.   i\'m sick of copywriting. i\'m sick of writing. i\'m sick of writing this!   i think i need to totally fresh hobby. i mean, i can read like a madman but how many times can you reread the confederacy of dunces until you wanna puke and faint?   being poor is NOT conducive to being well-read, i\'ve found.   in kinokuniya th other day i spied a thrillingly lewd victorian-age novel called fanny hill: memoirs of a woman of pleasure by an obviously leery, perverty john cleland. i downloaded the first chapter yesterday and i was quite dumbstruck.   di: this novel mula dengan lesbian porn!  amir: ?  di: no yang, you see, this book was written in the 1700\'s!!!  amir: suka baca lesbian porn la tu.   i love porn. literary porn.   on to the painting: yesterday, amir was thrilled with my proposal and started plotting ways for me to earn some bucks through my paintings (and i was thinking \"macamana nak draw a nose? dah lupa\"). he bought me a huge tablet of sketching paper, a pencil, an eraser (\"raba\") and a pencil sharpener.   i drew a nudie pic and trashed it before he, ain, brother, atan or atan\'s gorgeous girlfriend could see it.   will badger amir to get me some color pots. i\'m going to paint tonight! no fooling around with sketches!   '),(2586,3038828,'22,April,2004','2004-04-22 00:00:00','    last night i followed amir to viper. he was gonna pitch some harleys to datuk and his friends.   it was the WORST pub experience ever!   i didn\'t touch a drop (remembering how it was the last time) and so i was sober and aware. i chain-smoked tho.   and later, datuk\'s friend brought over a chivas coke. had to gulp it down. threw it up back home. amir was so apologetic about everything:-PPP   tonight\'s dinner: fettuccine pasta with meatballs and tomato sauce  mesclun with thousand island dressing no weed   ;-)   '),(2587,3038828,'20,April,2004','2004-04-20 00:00:00','    anak kucing misses jebon.   '),(2588,3038828,'17,April,2004','2004-04-17 00:00:00','    i am having a bad hangover now.  when i reached the office i immediately headed to the toilet and puked some. the night before i actually puked in the trash bin and later, beside the bed:-PPPP   it was SO bad. i can\'t talk about it now. yes, we did go out with uncle men and yeh, he got as smashed as we did:-) ann was nice. it was all good. except for the hangover!!!!!!!!   sayang, jangan marah. lagu ni saja ja i upload because this song dah replace lagu welcome to the jungle la:-)   i had been listening to guns n roses\' live version of jungle daily for the past 3 months!!!  i think i will be puking again soon.   '),(2589,3038828,'16,April,2004','2004-04-16 00:00:00','    It could all be so simple But you\'d rather make it hard Loving you is like a battle And we both end up with scars Tell me, who I have to be To get some reciprocity see No one loves you more than me And no one ever will  Is this just a silly game That forces you to act this way Forces you to scream my name Then pretend that you can\'t stay Tell me, who I have to be To get some reciprocity see No one loves you more than me And no one ever will  Hook: No matter how I think we grow You always seem to let me know It ain\'t workin\' It ain\'t workin\' And when I try to walk away You\'d hurt yourself to make me stay This is crazy This is crazy  I keep letting you back in How can I explain myself As painful as this thing has been I just can\'t be with no one else See I know what we got to do You let go and I\'ll let go too \'Cause no one\'s hurt me more than you And no one ever will  Care for me, care for me I know you care for me  There for me, there for me Said you\'d be there for me  Cry for me, cry for me You said you\'d die for me  Give to me, give to me Why won\'t you live for me   '),(2590,3038828,'15,April,2004','2004-04-15 00:00:00','    oh gawd. smsed ann a moment ago and it is still on. asked her where to tonight, and predictably it went \"wherever you want to go.\" goddamnit. why is it that me and amir always end up the organizers? she asked us out, right?  i actually don\'t want to go. i REALLY am not in the mood. plus, it\'s ann... and my uncle... don\'t wanna go! tak nak!   i tried thinking of some excuse. they weren\'t all very feasible. fuck it to hell.   last night:  delia: yang, hungry. amir: ... delia: dah stone ke ni?  amir: ... delia: YANG! amir: yes di?   it has been like this for this whole week already. both of us zonked out of our minds. our weed is NOT very good.   '),(2591,3038828,'15,April,2004','2004-04-15 00:00:00','    another absolutely boring day at werk. i need a holiday. a lo-o-o-o-o-ong holiday.  surfed to airasia.com to check out the flights to kuala terengganu and goddamnit it\'s still expensive. will have to wait.  i need a beach! i need a beach!   hanging around in seawater until i get tanned all over and eating seafood all day.  i\'m so sick of counting pennies and looking at the clock all the time.   finished this months feng shui tips newsletter. the alias i work under, corey tsang, sounds sooooooooo gay.   :-P   '),(2592,3038828,'13,April,2004','2004-04-13 00:00:00','    yikes.   ann just called. she wants to meet up with uncle men this weekend.   i really don\'t want to. but i have to.   you see, i have to cuz i need to oblige my uncle. he\'s been reminding me about the jug i\'ve been promising him. sigh!   macamana yang? pegi ke?   i love my boyfriend more and more. more and more and more. it must be true love right? to go through extremes and yet love eachother with the same intensity?   '),(2593,3038828,'12,April,2004','2004-04-12 00:00:00','    the past weekend has to be the WORST of my life.   '),(2594,3038828,'09,April,2004','2004-04-09 00:00:00','    i am 24!!!!   i mean, i\'ve been 24 for 2 days already. still hasn\'t sunk in yet.  24 is pretty old.   24 to me sounds like a manager job, a pretty apartment in a gorgeous upper class neighborhood and a toyota vios. and a kid. and a beautiful husband.  reality: 24 is a struggling copywriter butting heads constantly with her peers, living with her devoted boyfriend and a another cute couple in a bedbug (read: PIJAT) infested hellhole, cooking up pasta with alfredo sauce for my devoted, laughing a lot, going out with friends and getting drunk a lot, sitting on my boyfriend\'s lap while he plays the chotaiti-gamble with his friends on our dining table, getting stoned and making lots of love, sleeping late on sunday and frying up bacon and eggs at 3 pm, sharing all sorts of secret ideas and feelings with my boyfriend and knowing that i am connecting with another human being who feels like a part of me.   24 is being there for my boyfriend and having him being there for me.  doesn\'t sound too bad actually:-)   '),(2595,3038828,'08,April,2004','2004-04-08 00:00:00','    can you believe this guy?????      Sayang,  In the real world, I so speak like a typical Singaporean (that may sound very cinak to you, but I\'m proud of that). That is who I am, although you might find it pathetic. I was a Singaporean (now a Malaysian... but I guess you know that) and no matter what, deep inside I am still a Singaporean.   Like, perhaps we can meet for a cuppa. Then you can tell to my face lah whatever you think is not right about me. From my experience, even meek looking akak-akak tudung, adik-adik jambu and abang-abang skema can turn into a monster in chatrooms, newsgroups and e-mails.   Okay now... enuf bitching. From you lah. Penat lah wei.. Ala2 if you\'re my drag queen friends takper tau... I know it\'s the side effects of taking too much hormon :)  Now I want to know more about you.   Delia sayang, you asal mana (lepas ni jangan marah kalau I cakap Melayu pulak tau!). Umur berapa? Kalau diberikan tiga buku... buku apa you akan pilih untuk ke sebuah pulau ala2 Survivor?  You tak suka dengan kerja you sekarang... beritahu kat I apa you suka buat sebenarnya. I mean now. Take this particular question seriously as if you are talking to someone that is looking for a job.   Delia wrote:  Dearest Nizam,   Naw, I am just pissed at your flippancy. If more people thought of what they have to say without thinking thoroughly first, THAT would make this place into a happier shinier world.   Anyways, I do not claim to \'expect great things\' from you. I am only expecting forethought. You have a measurable amount of traffic in your website daily, and if you do not take your words with gravity there is probably some idiot who read your blog that does. Think about that. Be responsible to the integrity of your work and don\'t just lambast anything you like just because you imagine your readers don\'t \'expect great things\' from you.    And if my Emails sound so terribly ominous to you, forget it. You\'ve been buttered up too much by other visitors you can\'t even see straight (no pun intended). This is honest advice, from one writer to another. If taking my words and calling them \'negativity\' will help you sleep easier with yourself at night, fine.    Yours truly, Delia   p/s You shirking away from my questions and starting off your email with pidgin English-Chinese is pathetic.    p/p/s I hate my job so just save your \'I\'m just kidding!\' insults to someone who does give a fuck about feng shui.    p/p/p/s i AM being constructive.    ----- Original Message -----  From: Nizam Zakaria  To: Delia  Sent: Thursday, April 08, 2004 3:28 PM Subject: Re: Fiction   Why you so marah2 one... I just don\'t get you. And please, don\'t expect great things from me. I don\'t deserve that. I do not take my writing seriously. If you doubt that, fine. But I guess you don\'t know me that well.   And really... why the anger towards me? Why all these negativity? Maybe... you should shift your table or something. You know, bad feng shui :) (and please... this is a joke. I really mean it!!!)  Okay like this lah... If I\'m your puching bag and you want to turn me into a free theraphy session for whatever issue that you are facing... Don\'t write back. I hate the idea of you wasting  your time like this writing to me. Like, whatever for, right?   If you want, we can start by you putting a stop to all the insults you\'ve thrown at me and we can talk about other things that are more constructive. You know what?  I\'ll be more than happy to accept you as a friend and listen to your part of the story and arguments.   Nizam \"shiny happy people\" Zakaria (No... I\'m not on valium. I guess, I must have excellent feng shui at home and at work)   Delia wrote:  Dear Nizam,   So I take it you don\'t want your work to be taken seriously? I doubt that.    If you do want to protect your opinion, I think a jab at your sexual orientation is too black and white. As a self-proclaimed critic of the arts, I was anticipating something better than that from you.     Are you telling me that your imagination as a child was not spurred by children\'s tales (fiction, by the way) in television, books, film? Are you so sure that who you are right has nothing to do with works written by your favorite authors (fiction again.)? Are you so sure that your grasp of language has not been altered or bettered as a byproduct of your fondness for books?    Chew on that and get back to me.   Truly, Delia   p/s Crediting bad grammar as a result to being well-read is pretty lame too.      ----- Original Message -----  From: Nizam Zakaria  To: delia@fengshuitimes.com  Sent: Thursday, April 08, 2004 1:37 PM Subject: Fiction   No...  Works of fiction can never change a person\'s life. Or else... After reading so many books that have so many heterosexual characters in it, I should be straight by, now innit?   Or at least, be able to write with proper grammar, even for something as mundane as a personal blog.   Haiyoh... why you take things so seriously lah dear? So salah one if you take my writings seriously.   Regards,  Nizam   '),(2596,3038828,'06,April,2004','2004-04-06 00:00:00','    kurt cobain died yesterday 10 years ago.  is it really ten years already? i\'m about to be 24 tomorrow. 24. 24.   it still hasn\'t hit me yet: that i\'m slowly getting older. i don\'t  feel  old. just a bit wiser. but i definitely still feel like a kid.  where am i going with my life? do i want to face 30 stuck in the same no-money situation? i can\'t even afford to give my mother any cash. it hurts.   kak enya is in edinburgh now. according to her text messages to my cell, she\'s having a fuckin great time. i\'m so jealous!   '),(2597,3038828,'21,May,2004','2004-05-21 00:00:00','    blur\'s good song. i\'m crazy about this! the video is just too sad for such a great song. :-PPP   Waiting, I got no town to hide in The country\'s got a hold of my soul TV\'s dead and there ain\'t no war in my head And you seem very beautiful to me  Sleeping but my works not done I could be lying on an atom bomb I\'ll take care Cause I know you\'ll be there You seem very beautiful to me  It is the rest of your life keeps a rolling and rolling Picture in my pocket looks like you It is the rest of your life keeps a rolling, rolling, rolling along   '),(2598,3038828,'10,June,2004','2004-06-10 00:00:00','    i\'ve been sick on and off for like 3 days. it\'s all about my black sweater and striped toe-socks now. minimal makeup, no jewelry, go to work and bleed some, eat lunch and bleed some more after... rush my baby to pick me up at 5.30 and i\'m home at last!   i\'m wondering whether this is just 1) me genuinely being sick or 2) my body telling me it\'s time to move on.   i feel like i\'m stuck in a goddam rut. there\'s no room for creativity. i write like a fuckin robot and hey, guess what, it\'s a damn thankless job.   i am just SO delighted to go home with someone who loves me and cares about anything i think. i can unwind, smoke some weed, talk a bit, read a little, have some sex then get sleepy.   i want to be a housewife!!! yang, dengar tak!!! i nak duduk rumah!!!   it\'s a bit shocking to actually discover than i have no ambition. save for the restaurant-pub thing. THAT i am passionate about.  i\'ve been thinking and thinking at the back of my mind to just be a writer full-time at home, churn out pretty little relationship stories, and hope to god someone reads them. but in this country, being a full-time writer is just like being a celebrity. you need a supplementary income and your passion is just that: a labor of love, nothing more.   i love my country but when i think of the inprisonment of artists and writers here, i just feel like knocking some people\'s heads and tell them there\'s more to life than just your radio station and tv channels.   there\'s a whole other world out there. and everyone\'s tuned in, EVERYONE\'s tuned in, except you, malaysia.   '),(2599,1960271,'20,june,2004','2004-06-20 00:00:00','   First Name :   Dom   URL or Email :   dom   Comment :   test      First Name :   alli   URL or Email :   happy14u2@hotmail.com   Comment :   Nice site...      First Name :   Jennifer   URL or Email :   jksynan@yahoo.com   Comment :   My grandfather Elario Tassone immigrated to McKeesport, Pa early 1900s and married Francis Vallelunga. I was wondering if Im related to anyone in Chicago, Toronto, or anywhere else. :)      First Name :   Lena   URL or Email :   ltossone@hotmail.com   Comment :   I am from the Tassone line that immigrated to PA 1910 and reside now in WV.      First Name :   Gregory   URL or Email :   gni552@msn.com   Comment :   Son of Nicholas J. Invergo Grandson of Frank Invergo. Typed in Invergo on Google and landed here.      First Name :   Kim   URL or Email :   kmk356@yahoo.com   Comment :   I was just browsing the net searching for my mothers mother - Josephine Mickey Nicosia. Josephine was a twin - do not know twins name. Josephine married James Patrick Young, had two children - Sandra Joann and James Patrick Jr. Husband went off to fight w      First Name :   Ronald   URL or Email :   ron.naccarato@sympatico.ca   Comment :   Enjoyed. Congratulations. Keep it up for your future generations.      First Name :   Don   URL or Email :   tassone.dp@pg.com   Comment :   My grandfather, Philip Tassone, was born in Calabria about 1900. He came to the U.S., via Brazil, as a young boy and settled in Bradford, Pennsylvania. My father, Donald, was born there and now lives in Cincinnati, near me. I would appreciate any informat      First Name :   GERRY   URL or Email :   G.BILEK@ATT.NET   Comment :   MY GRANDFATHER WAS VITO TASSONE, BORN IN SEPT. 1895 IN CALABRIA ITALY. ORPHANED, HE IMMIGRATED TO ISHPEMING MICHIGAN AND THEN ULTIMATELY TO CHICAGO WHERE HE RAISED 7 CHILDREN. DIED SEPT. 30, 1970. I\'M JUST STARTING MY RESEARCH, HAVE MANY STORIES PASSED DO      First Name :   Bruna   URL or Email :   Calabriabruna@msn.com   Comment :   Ciao Sono nata a Serra S. Bruno, Vibo Valenzia. My granfather was Giuseppe Raffaele Tassone (scianni) and my grandmother Carmela Maria Bertucci (Ninnola). Thank you so much for this wonderful website!!!! I will share my finding with my whole family \"Tass      First Name :   cousin concetta   URL or Email :   concettabell2000@yahoo.com   Comment :   dear cuz, grazi tutti para su salud y ayuda! I am now on home peritoneal dialysis 10 hours a night-seven days a week. As you can imagine, I am hoping for the precious gift of a kidney donor.Thank you for spreading the word to tutti paisane.To interested      First Name :   Anthony   URL or Email :   www.Malena.net   Comment :   Loved to see a site dedicated to Sicilian/Italian heritage. I\'m searching desperately for roots to the Malena family from Sicily... so far only a little luck. It\'s great to see some people interested in tracing their roots!      First Name :   Darlene   URL or Email :   deedleROCKS@yahoo.com   Comment :   Paul, I believe that you are closely related to my husbands family. They are from Toronto Area. Newmarket and surrounding area, we live in North Bay (surrounding area).... I\'ll check back in to see if you are related...      First Name :   Darlene   URL or Email :   deedleROCKS@yahoo.com   Comment :   The Tassone family that I am in, comes from Calabria, but I\'m not sure what town...      First Name :   Darlene   URL or Email :   deedleROCKS@yahoo.com   Comment :   Hi there, I\'m a Canadian female from Ontario... I\'m married to a Tassone....      First Name :   Paul   URL or Email :   p.tassone@rogers.com   Comment :   Hi i am from Toronto Canada and my parents and their parents are from calbria.We were from a small town called fabrizia.My family came to canada in 1940--1955      First Name :   Teresa Cline   URL or Email :   bufanofiles@webtv.net   Comment :   sarching for \"BUFANO\" surname....stopping in to say hello! Have A Nice Day! Teresa Cline      First Name :   David   URL or Email :   Figjamda@hotmail.com   Comment :   Great site Domenic! My mothers maiden name is TASSONE,Assunta and she was one of those who imigrated to Adelaide, Australia from Italy in the 1950\'s.All the best in locating the Tassone\'s from all over the world.      First Name :   Remo Faieta   URL or Email :   www.italianheritagetours.com   Comment :   Great Web Site! I conduct tours to Italy and to the town of people\'s family roots. Grazie Remo      First Name :   peter scoleri   URL or Email :   patrolo66@hotmail.com   Comment :   great site! i printed it off and showed my father and he enjoyed it my father migrated to australia from reggio calabria back in the 50\'s      First Name :   severino   URL or Email :   sev@totalelec.co.uk   Comment :   born in catanzaro calabria 1961 moved to england 1962      First Name :   Toni   URL or Email :      Comment :   Hello from Melbourne Australia, it was interesting reading your story. I am married to a Tassone, originally from Fabrizia. Best wishes and keep up the good work. ciao.      First Name :   Janet   URL or Email :   janjas@earthlink.net   Comment :   Hi Dominic, Love your site! Great Stuff! I just started doing genealogy research on my family and came across your website. How \'bout these for coincidences... my grandparents immigrated from Nicosia to Bridgeport/Chinatown in the early 1900\'s. My      First Name :   Kety Tassone   URL or Email :   K.tassone-sehn@soziales.saarland.de   Comment :   Hallo, ich bin Kety Tassone und mein Grandpa ist von Simbario, Calabrien. Mein Vater ist 1965 nach Germany ausgewandert. Ich bin 34 Jahre und mÃ¶chte gerne wissen ob wir miteinander verwand sind. Viele GrÃ¼ÃŸe aus Germany an alle Tassone!!! Mein Onkel Vin      First Name :   Luciana   URL or Email :   lcomishen@rogers.com   Comment :   There are a lot of Tassone\'s in Toronto. My mother\'s maiden name is Tassone and came to Canada from Calabria in 1960.      First Name :   Amalia   URL or Email :   aevola@dscr.dla.mil   Comment :         First Name :   giorgio   URL :      Comment :   say hi to cecily from us giorgio and vincent giurdanella      First Name :   Ralph   URL :   rclarosa@yahoo.com   Comment :   I see the pasted poem and part of my message didn\'t take. Send me your e-mail address and I\'ll post it that way. Ciao!      First Name :   Ralph   URL :   rclarosa@yahoo.com   Comment :   Hi, Dominic! I\'ve been meaning to tell you how much I enjoy the rich information you provide here. My sister Roseanne alerted me to it. I was recently in both Enna and Agrigento and wrote a little poem that includes some of what you cover here. It fudges      First Name :   John   URL :   Jfavale@ibjwhitehall.com   Comment :   I am a 2nd generation Italian with a grandfather also from Calabria. I am very interested in the geneaology of mia famiglia and was happy to see your page. Good luck with finding additional information. John      First Name :   Justin (boqueesha@hotmail.com)   URL :      Comment :   Crazy, I just decided to see if my name would come up on a search engine, and here\'s my family tree. I\'m the grandson of Anthony Invergo Sr. Great site!      First Name :   Geremy   URL :   gtassone@gates.com   Comment :   Wonderful sight      First Name :   Daniel labarbera   URL :   www.netpilar.com   Comment :   Tambien soy 4ta. generacion de los labarbera en Argentina, de la misma ciudad de origen que Uds. en Italia. Les envio un saludo y un fuerte abrazo.      First Name :   Dominic   URL :      Comment :   Hopefully this will not get deleted by GeoCities like the last one!  '),(2600,1960271,'20,june,2004','2004-06-20 00:00:00','   First Name :   Francesco Lepanto   URL or Email :   flepanto@yahoo.com   Comment :   ciao domenico, my mothers last name was Tassone, and her mothers was Bertucci. We live in eastern Canada. what do you know about the name; Lepanto? Is tassone roots greek? grazi..Francesco      First Name :   Angela   URL or Email :   angela0471@yahoo.com   Comment :   My mothers maiden name is Tassone. She was born in Serra San Bruno, Calabria. Parents were Luigi nickname Vito Tassone and Concetta would love to hear from Tassones around the world or think we may be related email me, would love to hear from any relative      First Name :   Debbie(finishing)   URL or Email :      Comment :   and I would like him to know about his origins...PLz email me if you have any knowledge...thank you      First Name :   Debra   URL or Email :   debbie_tassone@yahoo.com   Comment :   Hi...My name is Debbie...my great uncle is ..Robert Tassone...and his brother Gerry is my grandfather....Im only 20..and I dont know much about my family or its history...Im very anxious..to know if I may have family.. I dont know about...I just had a son      First Name :   Mickey   URL or Email :   none   Comment :   Hey, Im a 15 year old looking 4 my family history...its kind of hard when you dont know what youre doing...can someone help me out!? The surnames Im looking for are Calabrese, Destafano, and Pagarillo!      First Name :   Phyllis   URL or Email :   cheerful87@hotmail.com   Comment :   Anyone know four Tassone sisters from New Jersey. One born around 1948? Roman Catholic, five feet tall? Tell her shell be getting a happy call about me soon! 9-24-2003      First Name :   Claudio   URL or Email :   claudio.tassone@goldinger.ch   Comment :   Un saluto a tutti Tassone around the world      First Name :   Robert of Lorain, OH   URL or Email :   none, sorry   Comment :   I would like to hear from you other Tassones. My home address is 3419 McKinley Lorain, Ohio zip code 44052. Phone is unlisted. But maybe someday Ill let it out. But I do want you to write me. Good luck in your further search.      First Name :   Robert ( one more time)   URL or Email :   none. sorry   Comment :   Continuing... My dad and Phil Tassone are first cousins. My dad was Joseph John Tassone, born Dec. 1916, In Kane Pennsyvania. Died on Dec. 9th, 1986. Dont know when Phil Tassone died.      First Name :   Robert ( again )   URL or Email :   none. sorry   Comment :   As I was saying before, after my grandfather put his first five kids in the orphan home, he remarried and moved out to Calif., where Pete and Frank were born. So, there must be a connection with my Tassone Tribe and yours, because too many names fit each      First Name :   Robert   URL or Email :   none, sorry   Comment :   On behalf of Peter Tassone, who died in 1977 from cancer of the kidneys at the ripe old age of 89 years old. He first lived in New York, moving to Kane Pennsylvania, then back to New York, where his first wife, Mary died from TB. After tossing his first f      First Name :   Robert   URL or Email :   none. sorry.   Comment :   I made a mistake on my mothers maiden name spelling. It was Gollati. Ive always had trouble remembering that spelling. Anyway, there are two more Tassones livng here in Lorain,Oh. Christina Tucker, Tassone and Suzanne Tassone. Carmella Monia, who also re      First Name :   Robert   URL or Email :   none. sorry.   Comment :   I forgot tell you that I reside in Lorain, Ohio. I am fifty years old, married for fourteen years to Teresa, have step-children and five step-grandchildren. No real children of my own.      First Name :   Robert   URL or Email :   none. sorry.   Comment :   My grandfather was Pete Tassone, who moved from New York to Lorain, Ohio, and then to California. My dad was Joseph John Tassone. I had two brothers, John and Gerald. Both are deceased. My mother Albinas maiden name was Golatti, and she is also deceased.      First Name :   PATRICIA   URL or Email :   Bernard.BETTING@wanadoo.fr   Comment :   My father is Vincenzo TASSONE, born in Simbario(Calabria) in1920. He came in France in 1948.I have 3 sisters: Elisabeth, Concetta and Marie.      First Name :   Joe   URL or Email :   vegasdog22@yahoo.com   Comment :   Pretty slick site Dom, J smooth wandered onto it tonight      First Name :   Joseph Fragomeli   URL or Email :   jfragomeli@sbcglobal.net   Comment :   My father Vincent Fragomeli and his brother Fernando came to this country in 1920 through Ellis Island also. Ending up in Middletown NY moving to Stamford CT where seven children were born.      First Name :   alli   URL or Email :   happy14u2@hotmail.com   Comment :   Nice site...      First Name :   Jennifer   URL or Email :   jksynan@yahoo.com   Comment :   My grandfather Elario Tassone immigrated to McKeesport, Pa early 1900s and married Francis Vallelunga. I was wondering if Im related to anyone in Chicago, Toronto, or anywhere else. :)      First Name :   Lena   URL or Email :   ltossone@hotmail.com   Comment :   I am from the Tassone line that immigrated to PA 1910 and reside now in WV.      First Name :   Gregory   URL or Email :   gni552@msn.com   Comment :   Son of Nicholas J. Invergo Grandson of Frank Invergo. Typed in Invergo on Google and landed here.      First Name :   Kim   URL or Email :   kmk356@yahoo.com   Comment :   I was just browsing the net searching for my mothers mother - Josephine Mickey Nicosia. Josephine was a twin - do not know twins name. Josephine married James Patrick Young, had two children - Sandra Joann and James Patrick Jr. Husband went off to fight w      First Name :   Ronald   URL or Email :   ron.naccarato@sympatico.ca   Comment :   Enjoyed. Congratulations. Keep it up for your future generations.      First Name :   Don   URL or Email :   tassone.dp@pg.com   Comment :   My grandfather, Philip Tassone, was born in Calabria about 1900. He came to the U.S., via Brazil, as a young boy and settled in Bradford, Pennsylvania. My father, Donald, was born there and now lives in Cincinnati, near me. I would appreciate any informat      First Name :   GERRY   URL or Email :   G.BILEK@ATT.NET   Comment :   MY GRANDFATHER WAS VITO TASSONE, BORN IN SEPT. 1895 IN CALABRIA ITALY. ORPHANED, HE IMMIGRATED TO ISHPEMING MICHIGAN AND THEN ULTIMATELY TO CHICAGO WHERE HE RAISED 7 CHILDREN. DIED SEPT. 30, 1970. I\'M JUST STARTING MY RESEARCH, HAVE MANY STORIES PASSED DO      First Name :   Bruna   URL or Email :   Calabriabruna@msn.com   Comment :   Ciao Sono nata a Serra S. Bruno, Vibo Valenzia. My granfather was Giuseppe Raffaele Tassone (scianni) and my grandmother Carmela Maria Bertucci (Ninnola). Thank you so much for this wonderful website!!!! I will share my finding with my whole family \"Tass      First Name :   cousin concetta   URL or Email :   concettabell2000@yahoo.com   Comment :   dear cuz, grazi tutti para su salud y ayuda! I am now on home peritoneal dialysis 10 hours a night-seven days a week. As you can imagine, I am hoping for the precious gift of a kidney donor.Thank you for spreading the word to tutti paisane.To interested      First Name :   Anthony   URL or Email :   www.Malena.net   Comment :   Loved to see a site dedicated to Sicilian/Italian heritage. I\'m searching desperately for roots to the Malena family from Sicily... so far only a little luck. It\'s great to see some people interested in tracing their roots!      First Name :   Darlene   URL or Email :   deedleROCKS@yahoo.com   Comment :   Paul, I believe that you are closely related to my husbands family. They are from Toronto Area. Newmarket and surrounding area, we live in North Bay (surrounding area).... I\'ll check back in to see if you are related...      First Name :   Darlene   URL or Email :   deedleROCKS@yahoo.com   Comment :   The Tassone family that I am in, comes from Calabria, but I\'m not sure what town...      First Name :   Darlene   URL or Email :   deedleROCKS@yahoo.com   Comment :   Hi there, I\'m a Canadian female from Ontario... I\'m married to a Tassone....      First Name :   Paul   URL or Email :   p.tassone@rogers.com   Comment :   Hi i am from Toronto Canada and my parents and their parents are from calbria.We were from a small town called fabrizia.My family came to canada in 1940--1955      First Name :   Teresa Cline   URL or Email :   bufanofiles@webtv.net   Comment :   sarching for \"BUFANO\" surname....stopping in to say hello! Have A Nice Day! Teresa Cline      First Name :   David   URL or Email :   Figjamda@hotmail.com   Comment :   Great site Domenic! My mothers maiden name is TASSONE,Assunta and she was one of those who imigrated to Adelaide, Australia from Italy in the 1950\'s.All the best in locating the Tassone\'s from all over the world.      First Name :   Remo Faieta   URL or Email :   www.italianheritagetours.com   Comment :   Great Web Site! I conduct tours to Italy and to the town of people\'s family roots. Grazie Remo      First Name :   peter scoleri   URL or Email :   patrolo66@hotmail.com   Comment :   great site! i printed it off and showed my father and he enjoyed it my father migrated to australia from reggio calabria back in the 50\'s      First Name :   severino   URL or Email :   sev@totalelec.co.uk   Comment :   born in catanzaro calabria 1961 moved to england 1962      First Name :   Toni   URL or Email :      Comment :   Hello from Melbourne Australia, it was interesting reading your story. I am married to a Tassone, originally from Fabrizia. Best wishes and keep up the good work. ciao.      First Name :   Janet   URL or Email :   janjas@earthlink.net   Comment :   Hi Dominic, Love your site! Great Stuff! I just started doing genealogy research on my family and came across your website. How \'bout these for coincidences... my grandparents immigrated from Nicosia to Bridgeport/Chinatown in the early 1900\'s. My      First Name :   Kety Tassone   URL or Email :   K.tassone-sehn@soziales.saarland.de   Comment :   Hallo, ich bin Kety Tassone und mein Grandpa ist von Simbario, Calabrien. Mein Vater ist 1965 nach Germany ausgewandert. Ich bin 34 Jahre und mÃ¶chte gerne wissen ob wir miteinander verwand sind. Viele GrÃ¼ÃŸe aus Germany an alle Tassone!!! Mein Onkel Vin      First Name :   Luciana   URL or Email :   lcomishen@rogers.com   Comment :   There are a lot of Tassone\'s in Toronto. My mother\'s maiden name is Tassone and came to Canada from Calabria in 1960.      First Name :   Amalia   URL or Email :   aevola@dscr.dla.mil   Comment :         First Name :   giorgio   URL :      Comment :   say hi to cecily from us giorgio and vincent giurdanella      First Name :   Ralph   URL :   rclarosa@yahoo.com   Comment :   I see the pasted poem and part of my message didn\'t take. Send me your e-mail address and I\'ll post it that way. Ciao!      First Name :   Ralph   URL :   rclarosa@yahoo.com   Comment :   Hi, Dominic! I\'ve been meaning to tell you how much I enjoy the rich information you provide here. My sister Roseanne alerted me to it. I was recently in both Enna and Agrigento and wrote a little poem that includes some of what you cover here. It fudges      First Name :   John   URL :   Jfavale@ibjwhitehall.com   Comment :   I am a 2nd generation Italian with a grandfather also from Calabria. I am very interested in the geneaology of mia famiglia and was happy to see your page. Good luck with finding additional information. John      First Name :   Justin (boqueesha@hotmail.com)   URL :      Comment :   Crazy, I just decided to see if my name would come up on a search engine, and here\'s my family tree. I\'m the grandson of Anthony Invergo Sr. Great site!      First Name :   Geremy   URL :   gtassone@gates.com   Comment :   Wonderful sight      First Name :   Daniel labarbera   URL :   www.netpilar.com   Comment :   Tambien soy 4ta. generacion de los labarbera en Argentina, de la misma ciudad de origen que Uds. en Italia. Les envio un saludo y un fuerte abrazo.      First Name :   Dominic   URL :      Comment :   Hopefully this will not get deleted by GeoCities like the last one!              Guestbook Join       First Name :   Phyllis   URL or Email :   cheerful87@hotmail.com   Comment :   Anyone know four Tassone sisters from New Jersey. One born around 1948? Roman Catholic, five feet tall? Tell her shell be getting a happy call about me soon! 9-24-2003      First Name :   Claudio   URL or Email :   claudio.tassone@goldinger.ch   Comment :   Un saluto a tutti Tassone around the world      First Name :   Robert of Lorain, OH   URL or Email :   none, sorry   Comment :   I would like to hear from you other Tassones. My home address is 3419 McKinley Lorain, Ohio zip code 44052. Phone is unlisted. But maybe someday Ill let it out. But I do want you to write me. Good luck in your further search.      First Name :   Robert ( one more time)   URL or Email :   none. sorry   Comment :   Continuing... My dad and Phil Tassone are first cousins. My dad was Joseph John Tassone, born Dec. 1916, In Kane Pennsyvania. Died on Dec. 9th, 1986. Dont know when Phil Tassone died.      First Name :   Robert ( again )   URL or Email :   none. sorry   Comment :   As I was saying before, after my grandfather put his first five kids in the orphan home, he remarried and moved out to Calif., where Pete and Frank were born. So, there must be a connection with my Tassone Tribe and yours, because too many names fit each      First Name :   Robert   URL or Email :   none, sorry   Comment :   On behalf of Peter Tassone, who died in 1977 from cancer of the kidneys at the ripe old age of 89 years old. He first lived in New York, moving to Kane Pennsylvania, then back to New York, where his first wife, Mary died from TB. After tossing his first f      First Name :   Robert   URL or Email :   none. sorry.   Comment :   I made a mistake on my mothers maiden name spelling. It was Gollati. Ive always had trouble remembering that spelling. Anyway, there are two more Tassones livng here in Lorain,Oh. Christina Tucker, Tassone and Suzanne Tassone. Carmella Monia, who also re      First Name :   Robert   URL or Email :   none. sorry.   Comment :   I forgot tell you that I reside in Lorain, Ohio. I am fifty years old, married for fourteen years to Teresa, have step-children and five step-grandchildren. No real children of my own.      First Name :   Robert   URL or Email :   none. sorry.   Comment :   My grandfather was Pete Tassone, who moved from New York to Lorain, Ohio, and then to California. My dad was Joseph John Tassone. I had two brothers, John and Gerald. Both are deceased. My mother Albinas maiden name was Golatti, and she is also deceased.      First Name :   PATRICIA   URL or Email :   Bernard.BETTING@wanadoo.fr   Comment :   My father is Vincenzo TASSONE, born in Simbario(Calabria) in1920. He came in France in 1948.I have 3 sisters: Elisabeth, Concetta and Marie.      First Name :   Joe   URL or Email :   vegasdog22@yahoo.com   Comment :   Pretty slick site Dom, J smooth wandered onto it tonight      First Name :   Joseph Fragomeli   URL or Email :   jfragomeli@sbcglobal.net   Comment :   My father Vincent Fragomeli and his brother Fernando came to this country in 1920 through Ellis Island also. Ending up in Middletown NY moving to Stamford CT where seven children were born.      First Name :   alli   URL or Email :   happy14u2@hotmail.com   Comment :   Nice site...      First Name :   Jennifer   URL or Email :   jksynan@yahoo.com   Comment :   My grandfather Elario Tassone immigrated to McKeesport, Pa early 1900s and married Francis Vallelunga. I was wondering if Im related to anyone in Chicago, Toronto, or anywhere else. :)      First Name :   Lena   URL or Email :   ltossone@hotmail.com   Comment :   I am from the Tassone line that immigrated to PA 1910 and reside now in WV.      First Name :   Gregory   URL or Email :   gni552@msn.com   Comment :   Son of Nicholas J. Invergo Grandson of Frank Invergo. Typed in Invergo on Google and landed here.      First Name :   Kim   URL or Email :   kmk356@yahoo.com   Comment :   I was just browsing the net searching for my mothers mother - Josephine Mickey Nicosia. Josephine was a twin - do not know twins name. Josephine married James Patrick Young, had two children - Sandra Joann and James Patrick Jr. Husband went off to fight w      First Name :   Ronald   URL or Email :   ron.naccarato@sympatico.ca   Comment :   Enjoyed. Congratulations. Keep it up for your future generations.      First Name :   Don   URL or Email :   tassone.dp@pg.com   Comment :   My grandfather, Philip Tassone, was born in Calabria about 1900. He came to the U.S., via Brazil, as a young boy and settled in Bradford, Pennsylvania. My father, Donald, was born there and now lives in Cincinnati, near me. I would appreciate any informat      First Name :   GERRY   URL or Email :   G.BILEK@ATT.NET   Comment :   MY GRANDFATHER WAS VITO TASSONE, BORN IN SEPT. 1895 IN CALABRIA ITALY. ORPHANED, HE IMMIGRATED TO ISHPEMING MICHIGAN AND THEN ULTIMATELY TO CHICAGO WHERE HE RAISED 7 CHILDREN. DIED SEPT. 30, 1970. I\'M JUST STARTING MY RESEARCH, HAVE MANY STORIES PASSED DO      First Name :   Bruna   URL or Email :   Calabriabruna@msn.com   Comment :   Ciao Sono nata a Serra S. Bruno, Vibo Valenzia. My granfather was Giuseppe Raffaele Tassone (scianni) and my grandmother Carmela Maria Bertucci (Ninnola). Thank you so much for this wonderful website!!!! I will share my finding with my whole family \"Tass      First Name :   cousin concetta   URL or Email :   concettabell2000@yahoo.com   Comment :   dear cuz, grazi tutti para su salud y ayuda! I am now on home peritoneal dialysis 10 hours a night-seven days a week. As you can imagine, I am hoping for the precious gift of a kidney donor.Thank you for spreading the word to tutti paisane.To interested      First Name :   Anthony   URL or Email :   www.Malena.net   Comment :   Loved to see a site dedicated to Sicilian/Italian heritage. I\'m searching desperately for roots to the Malena family from Sicily... so far only a little luck. It\'s great to see some people interested in tracing their roots!      First Name :   Darlene   URL or Email :   deedleROCKS@yahoo.com   Comment :   Paul, I believe that you are closely related to my husbands family. They are from Toronto Area. Newmarket and surrounding area, we live in North Bay (surrounding area).... I\'ll check back in to see if you are related...      First Name :   Darlene   URL or Email :   deedleROCKS@yahoo.com   Comment :   The Tassone family that I am in, comes from Calabria, but I\'m not sure what town...      First Name :   Darlene   URL or Email :   deedleROCKS@yahoo.com   Comment :   Hi there, I\'m a Canadian female from Ontario... I\'m married to a Tassone....      First Name :   Paul   URL or Email :   p.tassone@rogers.com   Comment :   Hi i am from Toronto Canada and my parents and their parents are from calbria.We were from a small town called fabrizia.My family came to canada in 1940--1955      First Name :   Teresa Cline   URL or Email :   bufanofiles@webtv.net   Comment :   sarching for \"BUFANO\" surname....stopping in to say hello! Have A Nice Day! Teresa Cline      First Name :   David   URL or Email :   Figjamda@hotmail.com   Comment :   Great site Domenic! My mothers maiden name is TASSONE,Assunta and she was one of those who imigrated to Adelaide, Australia from Italy in the 1950\'s.All the best in locating the Tassone\'s from all over the world.      First Name :   Remo Faieta   URL or Email :   www.italianheritagetours.com   Comment :   Great Web Site! I conduct tours to Italy and to the town of people\'s family roots. Grazie Remo      First Name :   peter scoleri   URL or Email :   patrolo66@hotmail.com   Comment :   great site! i printed it off and showed my father and he enjoyed it my father migrated to australia from reggio calabria back in the 50\'s      First Name :   severino   URL or Email :   sev@totalelec.co.uk   Comment :   born in catanzaro calabria 1961 moved to england 1962      First Name :   Toni   URL or Email :      Comment :   Hello from Melbourne Australia, it was interesting reading your story. I am married to a Tassone, originally from Fabrizia. Best wishes and keep up the good work. ciao.      First Name :   Janet   URL or Email :   janjas@earthlink.net   Comment :   Hi Dominic, Love your site! Great Stuff! I just started doing genealogy research on my family and came across your website. How \'bout these for coincidences... my grandparents immigrated from Nicosia to Bridgeport/Chinatown in the early 1900\'s. My      First Name :   Kety Tassone   URL or Email :   K.tassone-sehn@soziales.saarland.de   Comment :   Hallo, ich bin Kety Tassone und mein Grandpa ist von Simbario, Calabrien. Mein Vater ist 1965 nach Germany ausgewandert. Ich bin 34 Jahre und mÃ¶chte gerne wissen ob wir miteinander verwand sind. Viele GrÃ¼ÃŸe aus Germany an alle Tassone!!! Mein Onkel Vin      First Name :   Luciana   URL or Email :   lcomishen@rogers.com   Comment :   There are a lot of Tassone\'s in Toronto. My mother\'s maiden name is Tassone and came to Canada from Calabria in 1960.      First Name :   Amalia   URL or Email :   aevola@dscr.dla.mil   Comment :         First Name :   giorgio   URL :      Comment :   say hi to cecily from us giorgio and vincent giurdanella      First Name :   Ralph   URL :   rclarosa@yahoo.com   Comment :   I see the pasted poem and part of my message didn\'t take. Send me your e-mail address and I\'ll post it that way. Ciao!      First Name :   Ralph   URL :   rclarosa@yahoo.com   Comment :   Hi, Dominic! I\'ve been meaning to tell you how much I enjoy the rich information you provide here. My sister Roseanne alerted me to it. I was recently in both Enna and Agrigento and wrote a little poem that includes some of what you cover here. It fudges      First Name :   John   URL :   Jfavale@ibjwhitehall.com   Comment :   I am a 2nd generation Italian with a grandfather also from Calabria. I am very interested in the geneaology of mia famiglia and was happy to see your page. Good luck with finding additional information. John      First Name :   Justin (boqueesha@hotmail.com)   URL :      Comment :   Crazy, I just decided to see if my name would come up on a search engine, and here\'s my family tree. I\'m the grandson of Anthony Invergo Sr. Great site!      First Name :   Geremy   URL :   gtassone@gates.com   Comment :   Wonderful sight      First Name :   Daniel labarbera   URL :   www.netpilar.com   Comment :   Tambien soy 4ta. generacion de los labarbera en Argentina, de la misma ciudad de origen que Uds. en Italia. Les envio un saludo y un fuerte abrazo.      First Name :   Dominic   URL :      Comment :   Hopefully this will not get deleted by GeoCities like the last one!             Build you own website | Yahoo! GeoCities  '),(2601,1960271,'20,june,2004','2004-06-20 00:00:00','   First Name :   Francesco Lepanto   URL or Email :   flepanto@yahoo.com   Comment :   ciao domenico, my mothers last name was Tassone, and her mothers was Bertucci. We live in eastern Canada. what do you know about the name; Lepanto? Is tassone roots greek? grazi..Francesco      First Name :   Angela   URL or Email :   angela0471@yahoo.com   Comment :   My mothers maiden name is Tassone. She was born in Serra San Bruno, Calabria. Parents were Luigi nickname Vito Tassone and Concetta would love to hear from Tassones around the world or think we may be related email me, would love to hear from any relative      First Name :   Debbie(finishing)   URL or Email :      Comment :   and I would like him to know about his origins...PLz email me if you have any knowledge...thank you      First Name :   Debra   URL or Email :   debbie_tassone@yahoo.com   Comment :   Hi...My name is Debbie...my great uncle is ..Robert Tassone...and his brother Gerry is my grandfather....Im only 20..and I dont know much about my family or its history...Im very anxious..to know if I may have family.. I dont know about...I just had a son      First Name :   Mickey   URL or Email :   none   Comment :   Hey, Im a 15 year old looking 4 my family history...its kind of hard when you dont know what youre doing...can someone help me out!? The surnames Im looking for are Calabrese, Destafano, and Pagarillo!    First Name :   Phyllis   URL or Email :   cheerful87@hotmail.com   Comment :   Anyone know four Tassone sisters from New Jersey. One born around 1948? Roman Catholic, five feet tall? Tell her shell be getting a happy call about me soon! 9-24-2003      First Name :   Claudio   URL or Email :   claudio.tassone@goldinger.ch   Comment :   Un saluto a tutti Tassone around the world      First Name :   Robert of Lorain, OH   URL or Email :   none, sorry   Comment :   I would like to hear from you other Tassones. My home address is 3419 McKinley Lorain, Ohio zip code 44052. Phone is unlisted. But maybe someday Ill let it out. But I do want you to write me. Good luck in your further search.      First Name :   Robert ( one more time)   URL or Email :   none. sorry   Comment :   Continuing... My dad and Phil Tassone are first cousins. My dad was Joseph John Tassone, born Dec. 1916, In Kane Pennsyvania. Died on Dec. 9th, 1986. Dont know when Phil Tassone died.      First Name :   Robert ( again )   URL or Email :   none. sorry   Comment :   As I was saying before, after my grandfather put his first five kids in the orphan home, he remarried and moved out to Calif., where Pete and Frank were born. So, there must be a connection with my Tassone Tribe and yours, because too many names fit each    First Name :   Robert   URL or Email :   none, sorry   Comment :   On behalf of Peter Tassone, who died in 1977 from cancer of the kidneys at the ripe old age of 89 years old. He first lived in New York, moving to Kane Pennsylvania, then back to New York, where his first wife, Mary died from TB. After tossing his first f      First Name :   Robert   URL or Email :   none. sorry.   Comment :   I made a mistake on my mothers maiden name spelling. It was Gollati. Ive always had trouble remembering that spelling. Anyway, there are two more Tassones livng here in Lorain,Oh. Christina Tucker, Tassone and Suzanne Tassone. Carmella Monia, who also re      First Name :   Robert   URL or Email :   none. sorry.   Comment :   I forgot tell you that I reside in Lorain, Ohio. I am fifty years old, married for fourteen years to Teresa, have step-children and five step-grandchildren. No real children of my own.      First Name :   Robert   URL or Email :   none. sorry.   Comment :   My grandfather was Pete Tassone, who moved from New York to Lorain, Ohio, and then to California. My dad was Joseph John Tassone. I had two brothers, John and Gerald. Both are deceased. My mother Albinas maiden name was Golatti, and she is also deceased.      First Name :   PATRICIA   URL or Email :   Bernard.BETTING@wanadoo.fr   Comment :   My father is Vincenzo TASSONE, born in Simbario(Calabria) in1920. He came in France in 1948.I have 3 sisters: Elisabeth, Concetta and Marie.     First Name :   Joe   URL or Email :   vegasdog22@yahoo.com   Comment :   Pretty slick site Dom, J smooth wandered onto it tonight      First Name :   Joseph Fragomeli   URL or Email :   jfragomeli@sbcglobal.net   Comment :   My father Vincent Fragomeli and his brother Fernando came to this country in 1920 through Ellis Island also. Ending up in Middletown NY moving to Stamford CT where seven children were born.      First Name :   alli   URL or Email :   happy14u2@hotmail.com   Comment :   Nice site...      First Name :   Jennifer   URL or Email :   jksynan@yahoo.com   Comment :   My grandfather Elario Tassone immigrated to McKeesport, Pa early 1900s and married Francis Vallelunga. I was wondering if Im related to anyone in Chicago, Toronto, or anywhere else. :)      First Name :   Lena   URL or Email :   ltossone@hotmail.com   Comment :   I am from the Tassone line that immigrated to PA 1910 and reside now in WV.    First Name :   Gregory   URL or Email :   gni552@msn.com   Comment :   Son of Nicholas J. Invergo Grandson of Frank Invergo. Typed in Invergo on Google and landed here.      First Name :   Kim   URL or Email :   kmk356@yahoo.com   Comment :   I was just browsing the net searching for my mothers mother - Josephine Mickey Nicosia. Josephine was a twin - do not know twins name. Josephine married James Patrick Young, had two children - Sandra Joann and James Patrick Jr. Husband went off to fight w      First Name :   Ronald   URL or Email :   ron.naccarato@sympatico.ca   Comment :   Enjoyed. Congratulations. Keep it up for your future generations.      First Name :   Don   URL or Email :   tassone.dp@pg.com   Comment :   My grandfather, Philip Tassone, was born in Calabria about 1900. He came to the U.S., via Brazil, as a young boy and settled in Bradford, Pennsylvania. My father, Donald, was born there and now lives in Cincinnati, near me. I would appreciate any informat      First Name :   GERRY   URL or Email :   G.BILEK@ATT.NET   Comment :   MY GRANDFATHER WAS VITO TASSONE, BORN IN SEPT. 1895 IN CALABRIA ITALY. ORPHANED, HE IMMIGRATED TO ISHPEMING MICHIGAN AND THEN ULTIMATELY TO CHICAGO WHERE HE RAISED 7 CHILDREN. DIED SEPT. 30, 1970. I\'M JUST STARTING MY RESEARCH, HAVE MANY STORIES PASSED DO    First Name :   Bruna   URL or Email :   Calabriabruna@msn.com   Comment :   Ciao Sono nata a Serra S. Bruno, Vibo Valenzia. My granfather was Giuseppe Raffaele Tassone (scianni) and my grandmother Carmela Maria Bertucci (Ninnola). Thank you so much for this wonderful website!!!! I will share my finding with my whole family \"Tass      First Name :   cousin concetta   URL or Email :   concettabell2000@yahoo.com   Comment :   dear cuz, grazi tutti para su salud y ayuda! I am now on home peritoneal dialysis 10 hours a night-seven days a week. As you can imagine, I am hoping for the precious gift of a kidney donor.Thank you for spreading the word to tutti paisane.To interested      First Name :   Anthony   URL or Email :   www.Malena.net   Comment :   Loved to see a site dedicated to Sicilian/Italian heritage. I\'m searching desperately for roots to the Malena family from Sicily... so far only a little luck. It\'s great to see some people interested in tracing their roots!      First Name :   Darlene   URL or Email :   deedleROCKS@yahoo.com   Comment :   Paul, I believe that you are closely related to my husbands family. They are from Toronto Area. Newmarket and surrounding area, we live in North Bay (surrounding area).... I\'ll check back in to see if you are related...      First Name :   Darlene   URL or Email :   deedleROCKS@yahoo.com   Comment :   The Tassone family that I am in, comes from Calabria, but I\'m not sure what town...    First Name :   Darlene   URL or Email :   deedleROCKS@yahoo.com   Comment :   Hi there, I\'m a Canadian female from Ontario... I\'m married to a Tassone....      First Name :   Paul   URL or Email :   p.tassone@rogers.com   Comment :   Hi i am from Toronto Canada and my parents and their parents are from calbria.We were from a small town called fabrizia.My family came to canada in 1940--1955      First Name :   Teresa Cline   URL or Email :   bufanofiles@webtv.net   Comment :   sarching for \"BUFANO\" surname....stopping in to say hello! Have A Nice Day! Teresa Cline      First Name :   David   URL or Email :   Figjamda@hotmail.com   Comment :   Great site Domenic! My mothers maiden name is TASSONE,Assunta and she was one of those who imigrated to Adelaide, Australia from Italy in the 1950\'s.All the best in locating the Tassone\'s from all over the world.      First Name :   Remo Faieta   URL or Email :   www.italianheritagetours.com   Comment :   Great Web Site! I conduct tours to Italy and to the town of people\'s family roots. Grazie Remo    First Name :   peter scoleri   URL or Email :   patrolo66@hotmail.com   Comment :   great site! i printed it off and showed my father and he enjoyed it my father migrated to australia from reggio calabria back in the 50\'s      First Name :   severino   URL or Email :   sev@totalelec.co.uk   Comment :   born in catanzaro calabria 1961 moved to england 1962      First Name :   Toni   URL or Email :      Comment :   Hello from Melbourne Australia, it was interesting reading your story. I am married to a Tassone, originally from Fabrizia. Best wishes and keep up the good work. ciao.      First Name :   Janet   URL or Email :   janjas@earthlink.net   Comment :   Hi Dominic, Love your site! Great Stuff! I just started doing genealogy research on my family and came across your website. How \'bout these for coincidences... my grandparents immigrated from Nicosia to Bridgeport/Chinatown in the early 1900\'s. My      First Name :   Kety Tassone   URL or Email :   K.tassone-sehn@soziales.saarland.de   Comment :   Hallo, ich bin Kety Tassone und mein Grandpa ist von Simbario, Calabrien. Mein Vater ist 1965 nach Germany ausgewandert. Ich bin 34 Jahre und mÃ¶chte gerne wissen ob wir miteinander verwand sind. Viele GrÃ¼ÃŸe aus Germany an alle Tassone!!! Mein Onkel Vin    First Name :   Luciana   URL or Email :   lcomishen@rogers.com   Comment :   There are a lot of Tassone\'s in Toronto. My mother\'s maiden name is Tassone and came to Canada from Calabria in 1960.      First Name :   Amalia   URL or Email :   aevola@dscr.dla.mil   Comment :         First Name :   giorgio   URL :      Comment :   say hi to cecily from us giorgio and vincent giurdanella      First Name :   Ralph   URL :   rclarosa@yahoo.com   Comment :   I see the pasted poem and part of my message didn\'t take. Send me your e-mail address and I\'ll post it that way. Ciao!      First Name :   Ralph   URL :   rclarosa@yahoo.com   Comment :   Hi, Dominic! I\'ve been meaning to tell you how much I enjoy the rich information you provide here. My sister Roseanne alerted me to it. I was recently in both Enna and Agrigento and wrote a little poem that includes some of what you cover here. It fudges    First Name :   John   URL :   Jfavale@ibjwhitehall.com   Comment :   I am a 2nd generation Italian with a grandfather also from Calabria. I am very interested in the geneaology of mia famiglia and was happy to see your page. Good luck with finding additional information. John      First Name :   Justin (boqueesha@hotmail.com)   URL :      Comment :   Crazy, I just decided to see if my name would come up on a search engine, and here\'s my family tree. I\'m the grandson of Anthony Invergo Sr. Great site!      First Name :   Geremy   URL :   gtassone@gates.com   Comment :   Wonderful sight      First Name :   Daniel labarbera   URL :   www.netpilar.com   Comment :   Tambien soy 4ta. generacion de los labarbera en Argentina, de la misma ciudad de origen que Uds. en Italia. Les envio un saludo y un fuerte abrazo.      First Name :   Dominic   URL :      Comment :   Hopefully this will not get deleted by GeoCities like the last one!  '),(2602,1960271,'20,june,2004','2004-06-20 00:00:00','   > This email was sent courtesy of The Mining Journal > Web Site: >  > Dear  Ishpeming Family >  > Dom Tassone wanted you to take a look at this story > on The Mining Journal Web Site! > Please follow this link > http://www.miningjournal.net/news/story/0716202004_new04-n0716.asp > to view this article. >  > Dom Tassone comments:  >  >  > Thank You > Dom Tassone > d_tassone@yahoo.com  '),(2603,1960271,'12,june,2004','2004-06-12 00:00:00','   Ciao a  tutti,         &nbsp;       &nbsp;     Dom     &nbsp;    ============================================  Mobilito.com &nbsp;&nbsp; &nbsp;&nbsp;  [m] 773.426.4662&nbsp;&nbsp;&nbsp; [f]  435.603.1620  ============================================     &nbsp;  '),(2604,1960271,'12,june,2004','2004-06-12 00:00:00','   Cool site for Nicosiane nel mondo  dt  -----Original Message----- From: plavigna@libero.it [mailto:plavigna@libero.it]  Sent: Friday, June 11, 2004 11:41 AM To: d_tassone Subject: notice from Nicosia   Hello, my name is Paolo from Nicosia (En) Look the site www.nicosiaweb.it See you soon  '),(2605,1960271,'12,june,2004','2004-06-12 00:00:00','   Dom  '),(2606,1960271,'13,june,2004','2004-06-13 00:00:00','   The ANNOTICO Report  \"Facing Windows\" is now being released in the US, after being a multiple award winner In Italy, stars Massimo Girotti, Giovanna Mezzogiorno, Filippo Nigro, and Raoul Bova.  Massimo Girotti died of a Heart Attack at the age of 84, January 2003, one month before \"Facing Windows\" opened in Italy.  Massimo Girotti, who appeared in over 100 films, was one of Italy\'s legendary leading men, was a championship athlete, before his 1939 screen debut, and gained popularity in virile action roles. Long before Kirk Douglas, Girotti played Spartacus in the same-named 1952 European historical epic.  Massimo posthumously won the David Di Donatello Award for best actor for his touching performance in \"Facing Windows\"â€” the only time in his career Girotti was nominated for the David.   Girotti\'s complete Filmography follows the La Times Article.   ================================================================================ MOVIES AN ITALIAN ACTOR\'S FINAL TRIUMPH Los Angeles Times By Susan King Jun 14 2004  The Italian import \"Facing Windows,\" which opens Friday, is not only a multiple award winner on its home turf; the drama marks the swan song of one of the country\'s legendary leading men, Massimo Girotti.  Written and directed by Ferzan Ozpetek, \"Facing Windows\" concerns an unhappily married young couple, Giovanna (Giovanna Mezzogiorno) and Filippo (Filippo Nigro), who discover a well-dressed elderly man (Girotti) wandering the streets in a confused state. Despite Giovanna\'s protests, Filippo brings the man home with the promise that he will take him to the police. But he doesn\'t, and soon the man\'s presence begins to have an effect on Giovanna\'s life.  Even with him in the apartment, though, Giovanna continues to gaze out a window to the flat opposite hers, where resides a handsome young bachelor (Raoul Bova). One night the bachelor helps Giovanna find the old man after he wanders off, and they begin a relationship. Meanwhile, the old man starts to put the pieces of his life back together.  In January 2003, Girotti died of a heart attack at age 84. A month later, \"Facing Windows\" opened in Italy and was a triumph for the actor. He posthumously won the David Di Donatello Award for best actor for his touching performance â€” the only time in his career Girotti was nominated for the David.  A former athlete and engineering student, Girotti began in films in 1939 and quickly became one of Italy\'s leading heartthrobs and a favorite of such directors as Roberto Rossellini (\"Desiderio\"), Luchino Visconti (\"Ossessione\") and Bernardo Bertolucci (\"Last Tango in Paris\"). Although Girotti worked on more than 100 films and made countless TV appearances, \"Facing Windows\" was his first project in five years.  ====================================================================== >From All Movie Guide: A champion athlete before his 1939 screen debut, Italian actor Massimo Girotti gained popularity in virile action roles. Long before Kirk Douglas, Girotti played Spartacus in the same-named 1952 European historical epic. He also showed up in such brooding melodramas as Ossessione (1942), an unauthorized Italian version of James M. Cain\'s The Postman Always Rings Twice. Massimo Girotti later essayed character roles, notably in 1971\'s The Red Tent and 1972\'s Last Tango in Paris. ~ Hal Erickson ======================================================= FILMOGRAPHY Massimo Girotti  Actor: Born May 18, 1918 - Mogliano, Italy   2003 Facing Windows Actor: Davide   1994 The Monster Actor: Distinguished resident   1992 Dall\'Altra Parte Del Mondo Actor: Aureliano   1989 Rebus Actor   1987 La Boheme Actor: Old Suitor   1985 Quo Vadis? Actor   1985 The Berlin Affair Actor: Werner Von Heiden   1981 Passione D\'Amore Actor: The Colonel   1978 The Mafu Cage Actor   1976 Monsieur Klein Actor: Mr. Charles   1976 The Innocent Actor: Count   1975 L\'Ultima Chance Actor   1975 Cagliostro Actor: Casanova   1973 Les Voraces Actor: Olmi   1972 Gli Orrori Del Castello Di Norimberga Actor: Uncle Karl  1972 Last Tango in Paris Actor: Marcel   1970 Medea Actor: Greonte   1969 Le Sorelle Actor   1968 Cuore Di Mamma Actor   1968 Teorema Actor: Father   1968 Scusi, Facciamo L\'Amore? Actor: Tassi   1966 Le Streghe Actor: Sportsman [\"The Witch Burned Alive\"]   1965 Le Meravigliose Avventure di Marco Polo :Marco\'s Father, Nicolo   1964 Oro per i Cesari Actor: Maximus   1963 Venere Imperiale Actor: Leclerc   1963 Romolo e Remo Actor: Tazio   1961 Lettere Di Una Novizia Actor: Don Paolo Conti   1960 Lupi Nell\'Abisso Actor   1960 Erode il Grande Actor: Octavius   1960 I Giganti della Tessaglia Actor   1960 The Cossacks Actor: Alexander II of Russia   1959 It Happened in Rome Actor: Ugo Parenti   1958 La Venere di Cheronea Actor: Praxiteles   1958 Asphalte Actor   1958 La Strada Lunga un Anno Actor   1958 Giuditta e Oloferne Actor: Holophernes   1957 Souvenir d\'Italie Actor   1954 L\'Amour d\'une Femme Actor: Andre   1954 Senso Actor: Marquis Roberto Ussoni   1952 Ai Margini Della Metropoli Actor   1952 Roma Ore 11 Actor: Nando   1952 Spartaco Actor: Spartacus   1951 Secret of Three Points Actor   1950 Cronaca di un Amore Actor: Guido   1950 Duello Senza Onore Actor: Carlo   1950 Persiane chiuse Actor: Roberto   1950 Anni Difficili Actor: Giovanni   1949 Altura Actor   1949 In Nome Della Legge Actor: Magistrate Guido Schiavi   1949 Molti Sogni Per Le Strade Actor: Paolo   1948 Gioventu Perduta Actor: Marcello   1948 Preludio D\'Amore Actor   1948 Fabiola Actor: Sebastian   1947 Caccia Tragica Actor: Michele   1946 La Porta Del Cielo Actor   1946 Desiderio Actor  1943 Ossessione Actor: Gino   1941 La Corona di Ferro Actor: Arminio   1940 La Tosca Actor: Angeloti  '),(2607,1960271,'13,june,2004','2004-06-13 00:00:00','   Message: 1   Date: Sun, 13 Jun 2004 00:13:41 -0000   From: \"tomdee59\" &lt;tomdee59@yahoo.com&gt;   Subject: The Alfano Digest June 14, 04   &nbsp;   &nbsp;         &nbsp;   1. LIGHTEN UP?   2. A NEW PERSPECTIVE ON PINOCCHIO   3. SHOULD TV BE CENSORED?   4. MEN OF THE CLOTH,A DOCUMENTARY FILM   5. ITALIAN AMERICANS IN BOSTON   6. TRUE WORDS FROM THE WRONG SOURCE   &nbsp;   &nbsp;   LIGHTEN UP?   The following is indication of the rampant Italian American bashing. Should  we lighten up and laugh at this continual degradation of our dignity? You  decide!   Once upon a time there were three little pigs, the straw pig, the stickpig,  and the brick pig. One day this nasty old wolf came up to the straw pig\'s house  and said \"I\'m gonna huff and puff and blow your house down.\" And he did!!!! So  the straw pig went running over to the stick pig\'s house and said \"Please let me  in, the wolf just blew down my house. So the stick pig let the straw pig in.  Just then the wolf showed up and said, \"I\'m gonna huff and puff and blow your  house down.\" And he did!!!! So the straw pig and the stick pig went running over  to the brick pig\'s house and said, \"Let us in, let us in, the big bad wolf just  blew our houses down.\" So the brick pig let them in just as the wolf showed up.  The wolf said \"I\'m gonna huff and puff and blow your house down.\" The straw pig  and the stick pig were so scared! But the brick pig picked up the phone and made  a call. A few minutes passed and a big, black stretch limo pulls up. Out step  two massive pigs in pin striped suits and fedora hats. These pigs come over to  the wolf, grab him by the neck and beat the living shit out of him, then one of  them pulled out a gun, stuck it in his mouth and fired killing the wolf, then  they got back into their limo and drove off. The straw pig and the stick pig  were amazed!!! \"Who the hell were those guys?\" They asked. \"Those were my  cousins from Jersey, the Guinea Pigs.\"   &nbsp;   &nbsp;   A NEW PERSPECTIVE ON PINOCCHIO   The following is an excerpt form the Annotico report.   Pinocchio is very much an Italian story. It is full of Northern Italian  landscapes and Italian dishes. It also embodies the traditional Italian belief  that the family is of central importance. You can have a good time with your  friends, but you can only truly trust your kin. Good parents will sacrifice  themselves for their children without a murmur, as Geppetto does when there is  nothing to eat in the house but three pears, and he allows Pinocchio to devour  them all. Good children will also sacrifice themselves for their parents, as  Pinocchio does when he and his father, like Jonah in the Bible, escape from the  belly of the Great Shark, and he carries Geppetto out of the sea on his  back.   For the full text please go to:     http://groups.yahoo.com/group/ItalianAmericanOneVoiceCoalitionMessage       Center/?yguid=49124464   &nbsp;   &nbsp;   SHOULD TV BE CENSORED?   By Adam Thierer   Like millions of other Americans, I will be glued to my television set on  Sunday night watching the season finale of HBO\'s critically acclaimed series The  Sopranos to see who \"gets it,\" or even just to hear Tony or one of his mobster  buddies say \"Forgetaboutit!\" one last time this season. But one person who  definitely won\'t be watching The Sopranos finale on Sunday night is my young  daughter. The violence, profanity, and sexual content are not something I want  her exposed to. I\'m not sure what the right age is for children to see such  programming, but at the point my wife and I think she\'s ready, we\'ll talk to her  about such shows before we sit down to watch them with her.   For the full article please go to:     http://groups.yahoo.com/group/ItalianAmericanOneVoiceCoalitionMessage       Center/?yguid=49124464   &nbsp;   MEN OF THE CLOTH,A DOCUMENTARY FILM   MEN OF THE CLOTH is now in the later stages of production. It is intended for  television broadcast, theatrical screenings, international film festivals, and  educational outreach programs. It has the support of The National Italian  American Foundation, The Italian Cultural Institute, and the American Italian  Historical Association. The New York Foundation for the Arts, a 501 (c) (3),  tax-exempt organization, serves as the fiscal sponsor for the film.   This film is being produced and directed by Vicki Vasilopoulos. She is a free  lance writer whose articles often appear The New York Times. She recently wrote  an article in that paper about Shark Tale expressing the concerns of many  Italian Americans.   A free private screening will be presented Sunday, June 27, 2004 4:00 P.M.  Montclair State University, Dickson Hall   For those who wish to attend please write to italicsenrty@yahoo.com for  further details.   &nbsp;   &nbsp;   ITALIAN AMERICANS IN BOSTON   Excerpts from a revealing article by John Keller   A few blocks away and four decades later, in a sight unthinkable to the West  Enders Gans knew, Bobby Travaglini of East Boston was sworn in as the first  Italian-American leader of either legislative branch in the 223 years since the  first General Court of Massachusetts convened. The significance of which was  hardly lost on Travaglini\'s Senate colleagues.   \"He\'s been called by many names &#8212; Bob, Bobby, Trav, Senator, Senatore &#8212; but  henceforth he will be known and answered to as Mr. President,\" said  Greek-American Senator Steve Panagiotakos. Gushed Irish-American Senator Therese  Murray of Plymouth: This is truly a great country \"when a kid from the  neighborhood can be elected the first [Senate] president of Italian descent.\"  Travaglini\'s own mother, Josephine, wept in the gallery when his nomination for  the Senate presidency was moved and seconded.   NOTE: This insightful and very revealing article is well worth reading in its  entirety. For the complete article please go:     http://groups.yahoo.com/group/ItalianAmericanOneVoiceCoalitionMessage       Center/?yguid=49124464   &nbsp;   &nbsp;   DEFAMATION IN THE POLITICAL ARENA   Maine Gubernatorial campaign   Jonathan Carter, the Green Independent candidate for governor, began airing a  television commercial Monday that questions - with what some consider  anti-Italian slurs - whether John Baldacci, the Democratic candidate, can be  trusted.   Governor John Rowland smeared with anti Italian American comment   In an Northeast Magazine article by Bill Heald a vowel was added at the end  Governor John Rowland\'s name to read Rowlando. This was a double smear since it  attempted to connect Rowland with criminal activities at the expense of Italian  Americans who are generally regarded to be criminally bent by most of the  American public.   Chicago councilman\'s comment   While discussing the selection of a particular contractor for The city\'s bus  stop ad benches, Chicago City Council Finance Committee meeting, veteran Fifth  Ward Alderman Lawrence Bloom Remarked, \"I don\'t know who\'s in that company, but  his name shouldn\'t end in a vowel\".   &nbsp;   &nbsp;   TRUE WORDS FROM THE WRONG SOURCE   For years now, a growing number of my constituents have been of Italian  descent and I\'ve come to know them well. They have honored me with their support  and with their friendship. Indeed I can proudly say that some of my very best  friends are Italian Americans...These hearings on the Mafia are in no way  whatsoever a slur upon the great Italian people. Because I can state from my own  knowledge and experience that Italian Americans are among the most loyal, most  law-abiding, patriotic, hard-working American citizens in this land. And it  would be a shame, Mr. Chairman, if we allowed a few rotten apples to give a bad  name to the whole barrel. Because from the time of the great Christopher  Columbus up through the time of Enrico Fermi, right up until the present day,  Italian Americans have been pioneers in building and defending our great nation.  They are the salt of the earth, and they\'re one of the backbones of this  country.   NOTE: These lines are from the senator in The Godfather I. Too bad real U.S.  senators including Italian Americans do not give such well earned praise.  '),(2608,1960271,'13,june,2004','2004-06-13 00:00:00','   urlLink'),(2609,1960271,'26,june,2004','2004-06-26 00:00:00','   ------------------------------------------------------------------------ --------  University student Michael Taormina was one of only 40 students in the nation to qualify for the National Italian American Foundation program, which sent him on a two-week trip to Italy with all expenses paid.  Erik R. Bishoff Photo Editor  Student explores heritage in Italy ------------------------------------------------------------------------ --------  An all-expenses-paid trip allowed an Italian American University student to learn more about his culture Reese Lee Freelance Reporter June 24, 2004   Many college students can\'t afford a trip overseas, so one option is to win an all-expenses-paid trip, which is what University student Michael Taormina did.  The sophomore physics major visited Italy from May 23 to June 5 on a trip sponsored by the National Italian American Foundation (NIAF) under its Gift of Discovery program.   \"It was an opportunity that presented itself,\" Taormina said of his lucky break.   NIAF gave Taormina and 39 other Italian Americans10 days in the Campania region of Italy, touring famous sites like the ruins of Pompeii and Paestum. Taormina then stayed in Rome for several days.   The itinerary included meetings with students at a high school and technical institute in Salerno, visits to Vietri sul Mare on the Amalfi Coast and its ceramic producers, as well as an intensive Italian lesson. Taormina was selected to travel with the group after NIAF reviewed his background, academics, letters of recommendation, and essays he submitted.   Brock Olivo, NIAF Assistant of Educational Programs and chaperone of Taormina\'s group, said Taormina \"obviously attracted someone\'s eye and got a high grade.\"   Taormina said he wanted to continue touring Italy, but had to return to Eugene for spring-term finals.   \"That was the worst part about the trip,\" he said.   Taormina said the best part of the trip was the group he traveled with.   \"There are 40 kids in a group, and you have to do everything together, so you have to get to know them pretty fast,\" he said. \"It was a cool experience.\"   He also liked going with people who know the country.   \"It\'s a happy medium, as opposed to if you had gone by yourself or with a commercial tour group,\" he said.   Prior to this trip, Taormina only knew about his Italian heritage through his family and relatives.   \"Growing up in a small town in Eastern Oregon, you don\'t get a big Italian community, like in New York,\" he said.   Taormina recalled an amusing story from his trip.   \"Two other students and I had strayed from the group, and we started talking to these three kids,\" he said. \"Suddenly their mother and grandmother came out and dragged us into their house to eat.\"   Olivo remembers the incident.   \"We were in this little mountain town, and we had just exited a church,\" Olivo said. \"There was a long walk through the town to the bus, and Michael and two other kids were literally dragged into a house!\"   He added, \"I opened the door and there was a table full of food and wine and cheese and kids, and there in the middle of it all was Michael and two other students, stuffing their faces.\"   Olivo also said Taormina got a lot out of the trip.   \"Michael is a shining example of an Italian American,\" Olivo said. \"He was [mature] beyond his years, receptive to the speakers and tour guides. He soaked it up and took it in.\" Rich Santoro, NIAF Manager of Youth Programs, said applicants have to show an interest in their heritage.   \"The goal of the program is to bond Italian American students with their heritage by bringing them to Italy, to show them the roots of where their heritage lies,\" he said.   More information about NIAF can be found at .    ------------------------------------------------------------------------ -------- Reese Lee is a freelance reporter for the Emerald.  '),(2610,1960271,'20,june,2004','2004-06-20 00:00:00','   It\'s fun reading through this piece and wondering how they Itlian-Americans got left out.   Dominic  '),(2611,1960271,'20,june,2004','2004-06-20 00:00:00','   Sounds like more unimaginative ad writers trying to ride the coattails of TV darling \"The Sopranos\".    Dominic  -----Original Message----- From: ladolcevita_italianinelmondo@yahoogroups.com [mailto:ladolcevita_italianinelmondo@yahoogroups.com]  Sent: Thursday, June 24, 2004 5:36 AM To: ladolcevita_italianinelmondo@yahoogroups.com Subject: [ladolcevita_italianinelmondo] Digest Number 1025  ________________________________________________________________________ ________________________________________________________________________  Message: 1    Date: Wed, 23 Jun 2004 07:09:01 -0500    From: \"Ben \"   Subject: RE: More on the Caravaggio Ad in the New York Times  Salve. Does anyone know if the ad in question can be seen online or can you post the exact text? It would make it easier to comment.  Grazie.  Ben Lawton  -----Original Message----- From: l.bonaffini@worldnet.att.net [mailto:l.bonaffini@worldnet.att.net]  Sent: Tuesday, June 22, 2004 7:40 PM To: ladolcevita_italianinelmondo@yahoogroups.com Subject: [ladolcevita_italianinelmondo] More on the Caravaggio Ad in the New York Times   I am posting my entire exchange with Nina Ozlu, Vice President, Government and Public Affairs, Americans for the Arts. As it is painfully clear,she frankly admits that it would be impossible to use negative stereotypes for other groups, but does not explain why it is all right to do so for Italian Americans.  If you would like to write to her, her address is nozlu@artsusa.org  Luigi Bonaffini    I was frankly appalled today when I came across your Caravaggio ad in the New York Times.  It is not enough for Italian Americans to be constantly bombarded by Hollywood and the media with the usual vulgar mobster stereotype, now we are getting it from an organization like Americans for the Arts. Mind you, I am not implying any malicious intent on your part, but this ad is misguided, ill-conceived, and downright offensive. I am assuming that you meant to suggest that not all Italian names are mob connected, and that Italian culture is much more than that. Considering the fact that only an infinitesimally small percentage of Italian Americans are involved with organized crime (0.002%, as close to 0% as you can get), it is really very hard to understand why an ad about Caravaggio by a cultural organization should have to mention the Sopranos. And if that were not enough, you mention Mario Puzo\'s novels and, finally, you administer the coup de grÃ¢ce, scraping the bottom of the barrel, with \"Fuhgedaboudit\".  No matter what you intentions might have been, the only effect your ad will have will be to reinforce the ugly stereotype of the Italian mobster and the fact that if you talk about Italians, then you must mention the Sopranos and the Mafia.  This ad was a terrible mistake and you need to pull it immediately.    Mr. Bonaffini,  We\'ve received wonderful praise for our ad series, including several Italian-American organizations. The point of the ad is that most Americans don\'t know who the greatest artists in the world are and by using pop references through word association, it really helps to make the point. Our ad campaign won the highest prize by the National Association of Newspapers for our arts education campaign.  This campaign has been in place for 3 years and you are the first person to have interpreted the ad as an offense towards Italian-Americans, which clearly was not our intention.  Thank you for taking the time to contact us. All opinions are welcomed.  Sincerely,  Nina Ozlu  Vice President, Government and Public Affairs  Americans for the Arts    Dear Ms. Ozlu,  I want to thank you for your prompt reply.  Again, I do not in any way question your intentions, only the choices made in the ad, and the need to use negative stereotypes to make the point. To be more blunt, if this had been an ad about Isaac Stern, would you have said : \"No wonder people think Isaac Stern is a Jewish Shylock\"? How about a Spanish singer? \"No wonder people think Amancio Prada is a kingpin\" If the answer is no, as I think it is, then you need to ask yourself why it is all right to use negative stereotypes when Italians are involved.  Since you say that several Italian American organizations praised the ad, I will post it on a couple of Italian American forums and we\'ll see what they say.  Luigi Bonaffini     Actually, you should take a look at our entire print ad series. We do a double play on artists\' names through similar sounding pop references. I think you\'ll then see how the examples you bring up would never happen for MANY reasons.  Martha Graham - snack cracker?  Louis Armstrong - 1st man to walk on the moon?  Man Ray - a poisonous jellyfish?  Tchaikovsky - gesundheit?  Thanks for getting back to me. Nina      I\'m afraid the examples you give confirm without a doubt what I said to you. They are not similar sounding pop references, as you suggest, because they are harmless wordplay and not one of them alludes to an ugly negative stereotype, except for the Caravaggio ad. You rightly say that the examples I gave you would never happen for MANY reasons, and I of course agree. It is no doubt unthinkable to use Jewish or Latino stereotypes, and rightly so because they are vulgar and offensive, yet you could not think of a single reason why using the mobster stereotype is inappropriate.  Thanks again for your response.  Luigi Bonaffini    At this point, we are going to have to kindly agree to disagree. My points are not changing your mind and your points are not changing my mind. The only thing we do agree on is that there was no intention to offend Italian-Americans.  Thanks, Nina  ________________________________________________________________________ ________________________________________________________________________  Message: 2    Date: Wed, 23 Jun 2004 14:40:39 -0000    From: l.bonaffini@worldnet.att.net Subject: Picture of Caravaggio Ad  I have put a  urlLink picture of the Cravaggio ad  on my website.  Luigi Bonaffini  ______________________________________________________________________ ________________________________________________________________________  Message: 3    Date: Wed, 23 Jun 2004 13:55:29 -0500    From: \"Ben \"   Subject: RE: Caravaggio Ad in the New York Times  The ad is both entirely typical and absolutely outrageous. Your letter  says it all, but nothing will change until the perpetrators are hurt financially. In the meantime I trust the various Italian American local and national organizations must bury Nina Ozlu and American for the Arts with letters of protest.  Saluti cordiali,  Ben Lawton    -----Original Message----- From: l.bonaffini@worldnet.att.net [mailto:l.bonaffini@worldnet.att.net]  Sent: Wednesday, June 23, 2004 9:41 AM To: ladolcevita_italianinelmondo@yahoogroups.com Subject: [ladolcevita_italianinelmondo] Picture of Caravaggio Ad  I have put a picture of the Cravaggio ad on my website. You can get it at http://userhome.brooklyn.cuny.edu/bonaffini/caravaggio.jpg  Luigi Bonaffini   ________________________________________________________________________ ________________________________________________________________________  Message: 5    Date: Wed, 23 Jun 2004 16:58:01 -0500    From: \"Ben \"   Subject: CARAVAGGIO AD IN NY TIMES  Dear Ms Ozlu,   Professor Bonaffini and Mr. Fiore have posted on a number of Italian American message boards their correspondence with you regarding the Caravaggio ad that ran in the New York Times of June 20 (see below). Your replies to Professor Bonaffini, culminating in Â“we are going to have to kindly agree to disagree,Â” are simply not acceptable. If you honestly donÂ’t understand what is wrong with your replies, go see Spike LeeÂ’s BAMBOOZLED. He has done a wonderful job of lampooning replies such as yours. While I would like to believe that your campaign and your replies were simply naÃ¯ve errors, the public affairs person in the film is clearly revealed to be deliberately and programmatically racist. But even you, however subconsciously, reveal that you are aware of the defamatory nature of the ad when you write Â“I think you\'ll then see how the examples you bring up would never happen for MANY reasonsÂ” in response to Professor BonaffiniÂ’s suggestion that an analogous campaign about Isaac Stern might say Â“\"No wonder people think Isaac Stern is a Jewish Shylock\"? How about a Spanish singer? \"No wonder people think Amancio Prada is a [drug] kingpin.\"Â” Among the Â“Many reasonsÂ” for which the defamation of Jewish Americans, Hispanic Americans, African Americans, Native Americans, etc. Â“would never happenÂ” is that there is no doubt that Americans for the Arts would have to grovel as it begged for forgiveness and that you would be instantly terminated.    I join Professor Bonaffini and Mr. Fiore in inviting you to pull the ad immediately. I also join them in asking you identify the Italian-American organizations that have heaped Â“wonderful praiseÂ” on your ad series and to share their comments with us.    Cordially,   Ben Lawton     New York Times, Sunday June 20th   NO WONDER PEOPLE THINK CARAVAGGIO IS A GUY ON THE SOPRANOS.     It\'s hard to believe. Here\'s a 16th  century Baroque master whose bold naturalistic painting style first created a sensation, then a movement, A guy whose life was filled with the turbulence and excess of more than a dozen Mario Pugo novels.  This guy who, while troubled, ultimately found redemption and immortality in his art. But does the average kid on the street even know who Caravaggio is?   Fuhgedaboudit.    Too bad. Especially when you consider how much our children ca earn from the conflicted life of a great artist like Michelangelo Caravaggio.   Hee grew up in less than ideal circumstances. Most of his family died in the plague. Much of his youth was misspent on the mean streets of Rome And as a young artist he struggled for years a living. He was angry. Yet the angry contrast\' between light and darkness in his work is the very reason why it now hangs in countless museums around the world. If nothing else, it\'s a case study of the importance of having art as an outlet. Unfortunately, one we\'re fast removing from our kids\' lives.   (Caption of a painting by Caravaggio)  Honestly, he wouldn\'t last ten minutes on the Sopranos         I am posting my entire exchange with Nina Ozlu, Vice President, Government and Public Affairs, Americans for the Arts. As it is painfully clear,she frankly admits that it would be impossible to use negative stereotypes for other groups, but does not explain why it is all right to do so for Italian Americans. If you would like to write to her, her address is nozlu@artsusa.org   Luigi Bonaffini       I was frankly appalled today when I came across your Caravaggio ad in the New York Times.   It is not enough for Italian Americans to be constantly bombarded by Hollywood and the media with the usual vulgar mobster stereotype, now we are getting it from an organization like Americans for the Arts. Mind you, I am not implying any malicious intent on your part, but this ad is misguided, ill-conceived, and downright offensive. I am assuming that you meant to suggest that not all Italian names are mob connected, and that Italian culture is much more than that. Considering the fact that only an infinitesimally small percentage of Italian Americans are involved with organized crime (0.002%, as close to 0% as you can get), it is really very hard to understand why an ad about Caravaggio by a cultural organization should have to mention the Sopranos. And if that were not enough, you mention Mario Puzo\'s novels and, finally, you administer the coup de grÃ¢ce, scraping the bottom of the barrel, with \"Fuhgedaboudit\".   No matter what you intentions might have been, the only effect your ad will have will be to reinforce the ugly stereotype of the Italian mobster and the fact that if you talk about Italians, then you must mention the Sopranos and the Mafia.   This ad was a terrible mistake and you need to pull it immediately.       Mr. Bonaffini,   We\'ve received wonderful praise for our ad series, including several Italian-American organizations. The point of the ad is that most Americans don\'t know who the greatest artists in the world are and by using pop references through word association, it really helps to make the point. Our ad campaign won the highest prize by the National Association of Newspapers for our arts education campaign.   This campaign has been in place for 3 years and you are the first person to have interpreted the ad as an offense towards Italian-Americans, which clearly was not our intention.   Thank you for taking the time to contact us. All opinions are welcomed.   Sincerely,   Nina Ozlu   Vice President, Government and Public Affairs   Americans for the Arts       Dear Ms. Ozlu,   I want to thank you for your prompt reply.   Again, I do not in any way question your intentions, only the choices made in the ad, and the need to use negative stereotypes to make the point. To be more blunt, if this had been an ad about Isaac Stern, would you have said : \"No wonder people think Isaac Stern is a Jewish Shylock\"? How about a Spanish singer? \"No wonder people think Amancio Prada is a kingpin\" If the answer is no, as I think it is, then you need to ask yourself why it is all right to use negative stereotypes when Italians are involved.   Since you say that several Italian American organizations praised the ad, I will post it on a couple of Italian American forums and we\'ll see what they say.   Luigi Bonaffini         Actually, you should take a look at our entire print ad series. We do a double play on artists\' names through similar sounding pop references. I think you\'ll then see how the examples you bring up would never happen for MANY reasons.   Martha Graham - snack cracker?   Louis Armstrong - 1st man to walk on the moon?   Man Ray - a poisonous jellyfish?   Tchaikovsky - gesundheit?   Thanks for getting back to me. Nina           I\'m afraid the examples you give confirm without a doubt what I said to you. They are not similar sounding pop references, as you suggest, because they are harmless wordplay and not one of them alludes to an ugly negative stereotype, except for the Caravaggio ad. You rightly say that the examples I gave you would never happen for MANY reasons, and I of course agree. It is no doubt unthinkable to use Jewish or Latino stereotypes, and rightly so because they are vulgar and offensive, yet you could not think of a single reason why using the mobster stereotype is inappropriate.   Thanks again for your response.   Luigi Bonaffini       At this point, we are going to have to kindly agree to disagree. My points are not changing your mind and your points are not changing my mind. The only thing we do agree on is that there was no intention to offend Italian-Americans.   Thanks, Nina     Ms. Ozlu:   Luigi Bonaffini has shared his recent exchanges with you regarding the above referenced ad by posting them on several Italian American message boards.     Mr. Bonaffini did a more than adequate job in addressing the double standard that\'s so pervasive in pop culture specifically regarding how one, particular ethnic group (Italian Americans) are abused with impunity in a manner that no one in their right mind would do with other groups.  So I will not elaborate further.    For the sake of clarity, however, would you kindly elaborate on your response \"...I think you\'ll then see how the examples you bring up would never happen for MANY reasons\"...one of those many reasons wouldn\'t be \"good taste and sensitivity\", would it?   Additionally, would you be so kind as to identify which Italian American organizations voiced approval of the ad?  Perhaps if we can contact them, they\'ll be able to help us understand and correct our hypersensitivity. Regards,   Don Fiore   ________________________________________________________________________ ________________________________________________________________________  '),(2612,1960271,'20,june,2004','2004-06-20 00:00:00','     '),(2613,1960271,'20,june,2004','2004-06-20 00:00:00','   Saw this last night on VHS. This was every funny, especially if you like Roberto Benigni.  http://www.foreignfilms.com/films/1427.asp  '),(2614,1960271,'20,june,2004','2004-06-20 00:00:00','   ________________________________________________________________________ ________________________________________________________________________  Message: 1    Date: Mon, 28 Jun 2004 01:30:54 -0000    From: \"dege59\"   Subject: The Alfano Digest, June 28, 04  1. CARVAGGIO LINKED TO THE  SOPRANOS   2. SHARK TALE, WHAT YOU CAN DO TO HELP  3. ITALY\'S PARMALAT SCANDAL  4. BOOOKS FOR ITALIAN AMERICAN CHILDREN  5. ITALIAN AMERICANS NOT A MINORITY  6. MEN OF THE CLOTH PRESENTATION    CARVAGGIO LINKED TO THE SOPRANOS  \"No Wonder People Think Caravaggio is a Guy on the Sopranos\", this  was the Americans for the Arts ad title in the June 20 issue of the  New York Times. The ad went on with more defamatory statements not  only connecting Caravaggio, the great Italian Baroque artist to the  Sopranos, but to Mario Puzo\'s novels as well. \"But does the average  kid on the street even know who Caravaggio is? Fuhgedaboudit.\" Was  another ill chosen phrase used in the ad.  However the most  offensive was the caption under a portrait of Caravaggio  reading, \"Honestly, he wouldn\'t last ten minutes on the Sopranos.\"  Thanks to Professor Luigi Bonifini who spotted the ad there was a  huge response from the ever uniting Italian American community.   Letters of protest were sent by NIAF, OSIA, IAOV, UNOCO and Italian  American Pride which sent a letter containing this excerpt; \"Let\'s  give you the substitution test and see if you\'d have printed this  ad: Replace Michelangelo Mersi da Caravaggio with Marc Chagall,  a.k.a. Mark Zakharovich Shagal and use the Yiddish phrase of \"oi  vai\" instead of fuhgedaboudit. Rewrite the caption to  read. \"Honestly, he wouldn\'t last ten minutes as Shylock in  Shakespeare\'s Merchant of Venice.\"  How do you think the Jewish  American community would have reacted?  We urge all who read about this ad to join our protest and ask for  the ad to be pulled.    Please contact: Nina Ozlu Vice President, Government and Public  Affairs at: nozlu@artsusa.org  You can also call to register your protest at:  212 223 2787 or 202  371 2830  Or write to:  Robert Lynch, Executive  Director                                          Americans for the Arts  1000 Vermont Ave NW - Sixth Floor  Washington, DC 20005    SHARK TALE, WHAT YOU CAN DO THE HELP  Message from OSIA  Make your voice heard with a short letter to the opinion leaders and  decision-makers below.  Stress that Shark Tale will pass on damaging  stereotypes of Italian Americans as violent gangsters to  impressionable children.  Keep your letter to one page.  Be careful  not to offend any other ethnic group to make your point!   .	Write to your local elementary school principal, urging him  or her not to allow DreamWorks\' curriculum to be used in the  classroom.    . Send a Letter to the Editor to your local daily and weekly  newspapers, expressing your concern about the negative ethnic  stereotyping that children will learn in Shark Tale  . Contact your U.S. Congressional representative and two U.S.  senators, asking them to criticize Shark Tale from the floor of the  House and Senate.  [To contact your representative or senators, call  202/224 3121.]  . Write to the CEOs of DreamWorks\' marketing partnerships, telling  them you will not buy any product of theirs that promotes Shark  Tale. [See below.]   NOTE:  This is the most serious threat to our dignity and to our  children.  If the Shark Tale film is allowed to be released as  presently scripted the consequences will go far beyond the damage  already done by the Sopranos.  We implore all to become actively  involved in this campaign and to speak out against this impending  film before it is released. Please take a few minutes to contact at  least one company on the following listing.  Shark Tale Marketing Partners  Brad Blum - Chief Executive Officer, Burger King Corporation 5505 Blue Lagoon Drive Miami, FL 33126  Stephen W. Sanger - Chief Executive Officer, General Mills, Inc P.O. Box 1113 Minneapolis, MN 55440-1113  Douglas N. Daft Chief Executive Officer The Coca-Cola Company P.O. Box 1734 Atlanta, GA  30918  Scott Livengood- Chief Executive Officer, Krispy Kreme PO Box 83 Winston-Salem, NC 27103  Rhoda Olsen-President and Chief Operating Officer, Great Clips Great Clips, Inc. 7700 France Ave, Suite 425 Minneapolis, MN 55435  Carleton S. (Carly) Fiorina-Chairman and Chief Executive Officer of  HP 3000 Hanover Street Palo Alto, CA 94304-1185           ITALY\'S PARMALAT SCANDAL  Although it will be more than a year before those involved in one of  the biggest cases of corporate fraud in European history are brought  to trial, the impact of the Parmalat scandal can be seen throughout  the economic landscape. Food giant Parmalat, Italy\'s eighth-largest  industrial empire, collapsed late last year amid fraud accusations  against top company executives and scandal involving several major  players from the world of international finance.  Several changes in governmental oversight of business practices in  Italy are taking shape, all allegedly intended to \"clean up\" the  system which led to shoddy practices. But while the changes will  likely make Italy more palatable to international finance, an  innocent party-the international working class, beginning with  15,000 Parmalat employees who face the loss of their jobs-will be  the ones punished.    Note. The only Parmalat employees who have been creating the any  real wealth for the company are the ones who will be let go.   BOOOKS FOR ITALIAN AMERICAN CHILDREN  Books for children on Italian/Italian American themes are available  but not widely distributed on the shelves of bookstores, libraries  or even in our homes. The influence of children\'s literature upon  the psyches of the young is known to be significant. The books have  to be present and be used in order to effect any positive influence  upon the children. The positive images of Italian heritage can be  learned during childhood via books read to them at bedtime, picture  and resource books being available on the shelves of their homes and  libraries, and giving them gift books for birthdays and holidays.  Positive self-esteem about their Italian Heritage can help children  withstand the barrage of negative ethnic stereotyping in the media  they will encounter later on.  For more information and a listing of recommended books please go  to: http://www.sonsofitalypa.org/booklist.htm    ITALIAN AMERICANS NOT A MINORITY  In the Philadelphia case of Joseph Ricco (real name Joseph Gavel)  six Italian Americans were excluded from serving because of their  ethnicity.  Although protested, the court ruled that Italian  Americans could be excluded since they are not considered to have  minority status.  Please read the following and come to your own  conclusion.  Ethnicity and race: General indexes  Typically, members of a minority group share a sense of solidarity  and a desire to preserve their culture, traditions, religion, or  language. A minority group can sometimes be a numerical majority in  a minority group position. Minority group status is not a matter of  numbers; it is determined by the presence of distinguishing features  such as discrimination. Central features characterizing a minority  group are:  .	The members of a minority group suffer various disadvantages  at the hand of another group;  .	A minority group is identified by group characteristics that  are socially visible;  .	A minority is a self-conscious group with a strong sense  of \"oneness\";  .	People usually do not become members of a minority group  voluntarily; they are born into it;  .	By choice or necessity, members of a minority group tend to  marry within the group.\"     MEN OF THE CLOTH PRESENTATION  The Genius of Vicki Vasilopoulos was clearly evident at the private  screening of her documentary, \"Men of the Cloth\".  The work in  progress was presented at the Coccia Institute for the Italian  Experience in America at the Montclair State University in New  Jersey.   It is indeed a credit to Ms, Vasilopolous\' talent for taking a  mundane topic as tailoring and turning it into a work of art. The documentary was filmed in the United States and in Italy and is  being funded by grants and donations.  NIAF is a principal grantor.   Additional funding is needed to complete this story of a fading art. The film is expected to be aired across the U.S and Europe.  PBS is  a likely outlet for this splendid documentary.  '),(2615,1960271,'20,june,2004','2004-06-20 00:00:00','   -----Original Message----- From: H-NET List on Italian-American History and Culture [mailto:H-ITAM@H-NET.MSU.EDU] On Behalf Of Automatic digest processor Sent: Sunday, June 27, 2004 11:11 PM To: Recipients of H-ITAM digests Subject: H-ITAM Digest - 25 Jun 2004 to 27 Jun 2004 (#2004-153)  ----------------------------------------------------------------------  -------- Original Message -------- Subject: 	Re: No IA theatre companies in USA? Date: 	Sat, 26 Jun 2004 08:13:02 -0700 (PDT) From: 	Lawrence Tamanini   To: 	D-Candeloro@govst.edu    The question WHY there is an absence of Italaian American theatre companies is a good question. When you  consider what is taking place on the boards right now, it adds to your question ten fold.  I have stated and have proved beyound a shadow of a doubt that Italian Americans do not support their artists. The only artists that receive any reconition are ones portraying Italian Americans in the most negative light. (Please, don\'t waste any time trying to prove this statement wrong.) It all comes down to that the powers that be make TONS of money ! Furthermore, the taste of Italian Americans has now sunk to a low that is equal to, or worst then , the general population.  As far as the Italian American organizations are concerned they are equally at fault and their taste for the arts is almost non exsistent. Yes, every year they\'ll honor some flat singing tenor but 99% of the time they, the leaders of these organizations, don\'t have a clue and it gives me the impression that frankly, they don\'t care. Yes, they\'ll wait and come out with statements against the portrayal of Italian Americans as gangsters et al . Stick their hands in their pockets....please, thats whats truly needed but then they have to form a committee composed of people whose knowledge of the arts can be written on the back of a match book cover.  If the past actions of Italian Americans and the organizations are any indicator then the future for Italian American artists as it pertains to FINANCIAL support from their own Italian Americans community will be as it is...lip service only, whinning about the explotation of Italian Americans in films, the theatre and litature.  Lawrence E. Tamanini host \"THE ITALIAN AMERICAN HOUR\"  WNWR 1540 AM on the web WNWR.com SATURDAY 5:00pm  WBCB 1490 AM on the web WBCB.com SUNDAY 12:07 pm     --- Dominic Candeloro   wrote: > -------- Original Message -------- > Subject: 	a message from LindaAnn Loschiavo, who > writes for FRA NOI, etc. > Date: 	Fri, 25 Jun 2004 20:06:27 EDT > From: 	NonstopNY@aol.com > To: 	undisclosed-recipients:; > > > > . >    There is no Italian American Theatre group in the > entire USA.  Not one. > Irish Americans have 8 drama non-profits that help > Brian Friel, et al to > win Tony Awards. > Jewish non-profits have more than 70 groups that > fund + support theatre. > [One group funded plays by 66 Jewish dramatists last > year.] > Blacks have over 80 non-profits dedicated to theatre opportunities for > African-Americans. > Hispanics have more than 90 groups to fund + support > Latino performers + > theatre. > >    My article, that analyzes why other ethnic groups > win all the writing > awards and lists 20  of the major writing awards > that NO Italian > American author has ever won [since 1917], will > appear soon in  FRA NOI > [published in Chicago].  Please tell colleagues to > order and read my > article. (Subscriptions to FRA NOI are encouraged > and appreciated.) > >    Since not one Italian Amer organization has seen > fit to dedicate > itself to encouraging playwrights, this is what gets > produced in New > York City instead: >                NOW IN ITS 8TH YEAR! >          \"MURDERED by the MOB\" Has an offer you > can\'t refuse! > > \"It\'s a HIT! No pun intended.\" > FOX TV > > \"...this interactive murder mystery is obviously a > hit by any > definition... funny and worth seeing...\" > ZAGAT Theater Guide > . > MURDERED BY THE MOB is an interactive mystery comedy > dinner theater. NOW > IN ITS 8TH YEAR! Mingle with mobsters and molls, > meet the new \"Boss of > Bosses\", break bread and heads with Wiseguys and > Mafia princesses. Sure > you\'ll be rubbing elbows with goodfellas and ducking > bullets over > Broadway, but it won\'t stop the fun! > > Eat, Drink, Dance, and be Merry, for tonight you > could die... laughing! > The two and three hour comedy mystery features a 3 > course sit down > dinner, dancing with a professional DJ, and > surprises at every turn. > Aaay! > > > > > Aaay!   You even have the chance to solve the case > and win some > > fabulous prizes. There are seven prizes in every > performance, > > including a collector\'s item any crime solver > would cherish and > > \"Academy Awards\" for the best actor &  actress > chosen from the audience. > > > Arno\'s Restaurant > > > > > $59.00 per ticket includes Dinner, Dancing, &Show. > > > > Special promotion: > > Through August receive a $10.00 discount on  . . . >  '),(2616,1960271,'04,july,2004','2004-07-04 00:00:00','   Each of the listed defamation incidents was handled within the  limitations of our financial and staffing resources. We have  therefore relied extensively on the power of the internet in  recruiting members of our community through the Italian American One  Voice Coalition website who have had willingness and passion to join  the battle against our detractors.    While we cannot claim a perfect success rate we have indeed been  effective in making our voice heard and have changed the thinking of  many who have initially slandered us through different mediums.  Our  campaign in prompting Coca cola to remove and offensive film short  drew a positive response quickly. Radio and TV personality, Laura  Nachman who was initially recalcitrant ultimately saw the  justification in our ire and issued a most sincere apology.    There has been great benefit from our efforts even where we have not  won the day since our protests were duly registered and that will  have a residual value for the future.  Following is the listing of issues we have responded to:   July, 03  It was stated on a radio talk show that the marital infidelity of  Kobe Bryant who is facing rape charges could be traced to his  exposure to the Italian culture.  Bryant spent his early years  growing up in Italy.  The talk show host did not waste the  opportunity to mock with an exaggerated accent.   August, 03  Yonkers mayor John Spencer refuses to apologize for his statement  that contractor Carl Petrillo was a backroom gommbah.    October, 03  John Discepolo of Fox 5 was asked by his boss Neal Goldstein to don  a comical hat and to play a character named Guido, Descepolo refused  on the grounds that is was offensive to his Italian American  identity. Goldstein did not see any offense in his request.   November, 03  Author Randy White used derogatory phrases of WOP Dago and Grease  ball in his book, Everglades.   Coca Cola movie short, Mafia Madness depicted Italian Americans as  running a theater and strong-arming the audience to recall the name  of the movie and the size of the Coke they were drinking.   December, 03  Comic Charley Chiarelli used extreme vulgarity in his performance  that blatantly stereotyped Italians and Italian Americans.   Cingular commercial in which a little boy sitting on Santa\'s lap  tries to bribe Santa into giving him a phone?  The little boy speaks  in a NY \"tough guy\" dialect and the commercial ends with him  saying \"capesch\" to Santa.    January, 04   \"Godfathers & Goodfellas\" billed itself as the organized crime  trivia game which is a take off from Trivial Pursuit bills itself as  the Organized Crime Trivia Game.   Homies toy figurines called Palermos Italian Mafia.   Human Resources Magazine article titled Resolving Conflict Soprano  Style.  Magic Hands Freddy, an Off Broadway play with offensive dialog whose  main characters are Italian American brothers and their Italian born  sweetheart who they share.   February, 04  PBS series, \"The Medicis: Godfathers of the Renaissance\" used the  Mob theme in denigrating an historical Italian family and  Catholicism in one script.   Laura Nachman who covers television and radio sports for the Courier  Times stated that, \"Since Dan Marino had decided not to join the Miami Dolphins front office, he\'ll remain with HBO\'s \"Inside the  NFL.\", maybe Tony Soprano gave him an offer he couldn\'t refuse.\"  Konica Minolta ad using an Italian sounding character with broken  English to dramatize the need for the unit\'s color graphics.   Staples commercial with the Don of the supplies and Godfather type  music in the background.  \"Good Morning America\" (Channel 7 WABC).  Weatherman Bill Evans gave a birthday gift to Bill Bartelstein.  Mr. Evans  reassured Mr. Bartelstein it was an authentically autographed  baseball. But, directly thereafter he said, \"no only kidding, I got  it from \'Vinnie the Goomba\' off the street\"   The Sunday edition of The Hartford Courant ran a piece that took aim  at Gov. Rowland\'s current political problems by presenting him as a  corrupt Italian American by adding a vowel at the end of his name.     April, 04  Justice Scalia shown in a newspaper cartoon hiding in the bushes  with Dick Cheney on a duck hunting trip.  Cheney is well dressed in  the standard business suit and holding a shotgun while Scalia in  unshaven dressed in a wide lapel pinstriped suit holding a  machinegun.   Rendering Unto Caesar\'s Subjects, an impending PBS series departs  from true history by depicting ancient Rome as a, \"dirty, wild  savage city akin to Mexico City and Bombay.\"    May, 04  A Port Jefferson, New York Town Trustee stated while walking on an  inspection of a new restaurant that it looked like it was renovated  by a \"Real Guinea\".  The owner of the restaurant, Guy Lombardi  requested our support.   June, 04  A full page ad in the New York Times by The Americans for Art linked  Caravaggio, the great 16 century baroque artist with the Sopranos  and used a sopranoesque caption under his portrait saying, \"Honestly  he would not last ten minutes on the Sopranos.\"       One of our most important undertakings has been in sensitizing the  media which has been has shown a complete contempt for our standing  in the American society and those in our own community who see no  defamation to Italian Americans.  This has been accomplished through  appearances on radio and TV talk shows, letter writing campaigns,  phone calling, Internet message posting and talks at group  gatherings. Our hope is to expand this effort in the coming year by  reaching out to college students of Italian descent.  The Shark Tale initiative has unquestionably consumed hundreds of  hours of our time, however, that is only one of our two major  concentrations. We have been principally involved in the upcoming  symposium on Italian American defamation to be held at Seton Hall  University in New Jersey.  After encouraging such a program for  several years it has now become a reality.  \"Real Stories:  Discrimination and Defamation of Italian Americans\" will be  presented December, 4, 04.  There will be film showings and panel  discussions.  Films will focus on the tragedies such as the 46  lynchings that have occurred throughout our early history in  America.  Panel discussions will conducted with prominent  politicians, scholars, media critics and representatives from the  major Italian American organizations.   Dr. Emanuel Alfano    Chairman, Unico National Anti-Bias Committee  '),(2617,1960271,'01,august,2004','2004-08-01 00:00:00','   Spielberg\'s Hipocrisy & Rudeness nicely Lampooned by Italic Studies Institute   Bravissmo!  '),(2618,1960271,'08,august,2004','2004-08-08 00:00:00','   Friday, August 06, 2004  Agriculturalists in Sicily have formed co-ops and taken over fields once held by gangsters. They see their niche produce as an anti-crime effort.  Los Angeles Times By Tracy Wilkinson Times Staff Writer August 5, 2004  SAN GIUSEPPE JATO, Italy â€” Antonio Castro, a young agronomist with mud on his boots and sun in his skin, says he can\'t let himself think too much about the previous owners of the land he farms: Sicily\'s most ruthless Mafia bosses.  If he spent time worrying that the owners might get out of jail and return to kill him, he wouldn\'t be able to get his work done.  But after dark, when he settles in for the night at the newly renovated stone villa (and future bed and breakfast) that was seized from convicted gangland murderer Giovanni \"The Butcher\" Brusca, Castro is not alone.  \"I say I\'m not afraid,\" he said, \"but I sleep with three dogs and a gun.\"  Castro and his agricultural co-op are part of a consortium that is farming thousands of acres of Sicilian land confiscated from Mafiosi and then marketing the fruits of its harvest â€” pasta, vegetables, wine, olive oil â€” under a label that proudly declares the products\' crime-fighting provenance.  The seizure of Mafia assets has long been an important weapon in Italian authorities\' arsenal, but only in the last couple of years has legislation allowed the transfer of property to groups such as Castro\'s.  It\'s not a big business yet, but the Libera consortium, the umbrella of 800 anti-Mafia organizations, including Castro\'s co-op, says the project is attracting notice from followers of the Slow Food movement and similar boutique foodies. Libera products can be found on the shelves of one of Italy\'s major supermarket chains.  More important is the symbolism: Turning the land against its onetime owners shows the possibility of breaking the Cosa Nostra\'s hold on at least one aspect of its considerable economic power, advocates say. The act is both redemptive and cathartic, bringing life and bounty from a land stalked by gruesome death and terror.  Castro\'s co-op â€” named after Placido Rizotto, a union activist killed in 1948, presumably by gangsters â€” is housed, appropriately, in the police station of this small town about 15 miles southwest of Palermo. The co-op chalked up about half a million dollars in sales last year, when it produced more than two tons of pasta; so far this year, it has corked 13,000 bottles of white wine.  When they started out, Castro and his fellow agronomists met resistance and chilly stares from residents of the Sicilian towns that were once the dominion of assorted dons. It was difficult to recruit workers for the land, Castro recalled. But now laborers seek him out, he said.  Castro, 31, studied in Florence but was always intent on returning to his native Sicily. He hated the fact that, everywhere he went, his home\'s name was associated with the Mafia.  These days, he traverses the island\'s sharp, golden valleys and lush green hilltops in a beat-up blue Fiat Uno, clocking more than 2,000 miles some months.  Down one rutted dirt road, in a sun-parched 15-acre field just outside the town of Corleone, Castro and his hired hands are grafting sprigs of the well-regarded Nero d\'Avola red grape to vines that have gone unworked for two years. (The combination makes a more resistant, higher-quality grape, Castro explained.) This patch belonged to a nephew of Salvatore \"The Beast\" Riina, notorious \"boss of bosses\" for years until his arrest in 1993.  One of the farmhands said he was not afraid to till such blood-steeped soil, though he would give only his first name, Antonio, and Castro later confided the man\'s nephew quit out of fear.  \"We are here to sweat,\" said Antonio, 54, pausing to chat and leaning on his hoe. \"If we don\'t work it, someone else will.\"  A few miles away, the co-op is raising grain, lentils and chickpeas on a 150-acre farm confiscated from Riina. Durum wheat is milled into semolina and hauled to a pasta maker in Corleone, the town made famous by author Mario Puzo in \"The Godfather.\"  At the shop\'s entryway, a poster painted by schoolchildren takes up the Libera spirit: \"Where there were bad plants from the Mafia, now there are grains of freedom. Free pasta for free people.\"  Corleone and seven other towns in the area have agreed to participate in the land-confiscation project and help the agricultural co-ops with equipment, housing and verbal support, no small thing in this still-suspicious region.  Up the road, the stone villa that once belonged to Brusca sits above a breathtaking valley covered in yellow and lavender wildflowers and freshly planted cherry and plum trees. Cows can be heard mooing in the distance, as hawks fly overhead. The initials of the Brusca family still hang in wrought iron above the heavy wooden door of the main house.  With about a quarter of a million dollars from the European Union, Castro and his co-op have renovated part of the building and installed a restaurant, though they are awaiting permits to open it to the public. With more refurbishing, they plan a bed and breakfast.  The Bruscas were the biggest crime clan in San Giuseppe Jato. Giovanni Brusca, who owned the villa, was arrested in 1996 and eventually confessed to dozens of killings, including one in which he had the 11-year-old son of a rival dissolved in acid.  The boy\'s picture decorates the label of a fruity white wine that is being produced by another anti-Mafia co-op, Tempio del Monte Jato (Temple of Mt. Jato), which has also opened a restaurant for weekend patrons.  Giuseppe Randazzo, who runs the place, said he\'s not convinced that the work of a few farmers, chefs and vintners is going to make much difference in challenging the \"very entrenched\" Mafia.  Castro and the president of his co-op, Gianluca Faraone, think they are safe, for now. The Mafia has turned its sights away from agriculture and to more lucrative businesses such as construction, they noted.  \"The Mafia is not interested in confrontations right now, and to attack a group of agronomists would be a major confrontation,\" the 30-year-old Faraone said.  Maybe. But a few weeks after The Times visited, arsonists burned 25 acres of wheat on confiscated land that Faraone\'s co-op was farming. The fires were set just as the co-op began celebrating the harvest.  '),(2619,1960271,'14,July,2004','2004-07-14 00:00:00','   Jul. 11--CHICAGO -- It is, as the King of Siam famously declared, a puzzlement.  After lots of early-summer discussions with Chicago-area real estate agents, it\'s hard to know what to make of the marketplace. Just about the only consistency we encountered: Nobody says business is bad.   Previous reports from the trenches have hummed a familiar theme since the Tribune began this series of occasional stories in March 2003. Real estate agents (that is, those who had the time to return calls) almost inevitably told us: It\'s busy, busy, busy. Or even: It\'s a madhouse out there.   Not even the threat of war fazed Chicago-area home buyers, who swam toward sales contracts like salmon fighting their way up the Columbia River, apparently drawn by the siren song of low interest rates.   But current readings of the marketplace are a little of everything. Some quarters continue to be in the throes of realty hysteria, with multiple offers a fact of life and some houses selling within hours of being listed. Others say business is calmer, and they\'re starting to see price reductions, a phenomenon that\'s been absent for a couple of years.   Agents say that \"move-up\" properties and homes in the upper brackets are sitting, sitting, sitting -- except for the agents who say they\'re selling normally, thank you. Agents say buyers are cowed by potential volatility of interest rates; others say they\'re not concerned at all.   In some areas, investors who hope to snag a bargain, fix it up and resell for a profit are regarded as a major factor in driving up prices; in others, the speculators have practically disappeared, agents say.   Some areas have almost too many houses on the market; there\'s a dearth in others. Buyers want fixer-uppers, they say; buyers don\'t want fixer-uppers, they also say.   Here are a few snapshots of local markets, as seen by real estate agents:   CHICAGO NORTHWEST SIDE   Sally M. Haynes, ERA ReaLife   \"If it\'s priced from $230,000 to $260,000 -- really, all the way up to $300,000 -- it\'s swept up literally overnight in Edison Park, Portage Park and Norwood Park. Our average sale price is quickly creeping up to about $325,000, and last year the average was $262,000.   \"But what\'s being bought and what\'s being asked are two different things. Some sellers are convinced that they can get a lot more than what their neighbors sold for a few years ago, and that\'s not always the case. For one thing, what people are looking for has changed. Nine years ago, when I started in the business, it was large living rooms, then it went to big kitchens and lots of closets. Now it\'s more than one bathroom, a first-floor family room, and an outdoor entertaining space, a deck or patio.   \"What\'s driving the prices is that we have a captive audience of city workers who can\'t leave Chicago. I\'ve been working with a police officer and his wife for four months, and they\'re looking in the $250,000 range. We\'ve written offers on five houses. The least competition we\'ve had is six offers on one house. Another house had more than 50 bids.   \"The higher end here is becoming the $450,000 to $500,000 range, and that\'s influenced by city workers, too, who are buying those houses and adding second stories. But I\'m starting to see prices level off there, probably because of interest rates, and we\'re starting to see price reductions there for the first time in a couple of years.\"   NORTH SIDE   Ashley Carter, Realty Executives City   \"It\'s been a generally strong market. Market times are reasonable now, but it\'s hard to put a number on it. I\'d say houses sell within reasonable time frames, 75 days or less, but you have to price it right.   \"I hear anecdotally about rates going up and their effect on the market, but I haven\'t seen them really affecting anybody\'s ability to buy lately. They did affect things a few months ago, when they were starting to rise. It was directly affecting how much people could borrow.   \"There\'s interest in two-flats, from people who have purchased before and are looking to make an investment over the longer term. There\'s a good number of those properties for sale out there because the rental market has been down, in general, and the owners are ready to sell.   \"First-time buyers aren\'t seriously looking at properties that need work. Their attitude seems to be, let\'s move in, we can have our cookout the same weekend because there\'s nothing else we need to do.   \"You do hear about multiple offers, but not on the homes that have been on the market for three days. I\'m seeing multiple offers on homes that have been on the market for 45 days. It\'s the luck of the draw.\"   Source: Chicago Tribune Publication date: 2004-07-11  '),(2620,1960271,'30,July,2004','2004-07-30 00:00:00','   Chicago, IL, Jul. 27 (UPI) -- The pampered pooches of Chicago may be barred from the new $475 million Millennium Park, but near downtown dogs are getting a park of their own.  '),(2621,1960271,'26,July,2004','2004-07-26 00:00:00','   U.S. Newswire via Yahoo! News Mon, 26 Jul 2004 5:19 AM PDTThe surge in Illinois home sales continues as sales of existing, single-family homes in June increased 15.3 percent to 14,608 homes sold, according to the Illinois Association of REALTORS(r). In June 2003, 12,670 homes were sold.  '),(2622,1960271,'25,July,2004','2004-07-25 00:00:00','   Northwest Herald Sat, 24 Jul 2004 7:52 PM PDTPatricia and George Riddiford sold their vacation home last month for $925,000, two years after they bought it. They reaped a 28 percent profit on the golf resort property near Las Vegas.  '),(2623,1960271,'01,August,2004','2004-08-01 00:00:00','   July 26 (Bloomberg) -- Is a U.S. residential housing bubble a fact or strange fiction? There are two schools of thought on the subject. Either speculation is rampant due to low mortgage rates and home prices are poised for a decline, or demand is at an unprecedented level and prices will rise unabated by ``measured\'\' rate increases by the Federal Reserve and inflationary pressures. In either case, it\'s wise to be more careful when buying residential real estate now as prices continue to climb in the largest markets. Ian Morris, chief U.S. economist for HSBC Securities USA, whose parent company is HSBC Holdings Plc -- Europe\'s biggest bank by market value and the holder of more than $1 trillion in assets -- is one of those who say a bubble is looming in the U.S. residential market. ``The bubble psychology has manifested itself in very rich valuations,\'\' Morris says. ``House prices relative to income, rent, replacement-cost and home equity have set new highs. Twenty states that account for half of the population look shaky.\'\' Bullish on Bubble Morris supported his argument in a recent HSBC study that examined valuations of home prices relative to rents, income and consumer debt. His general conclusion was that U.S. housing prices were ``bubbly and 10 percent to 20 percent too high.\'\' Possibly exacerbating Morris\'s bubble scenario is the U.S. Federal Reserve Board\'s stated policy of ``measured\'\' increases in short-term interest rates. The Fed\'s move will probably ``cause a reassessment of likely future house-price risks and its associated debt, thereby triggering housing\'s fall,\'\' Morris says. Particularly telling, in Morris\'s view, is the housing price- to-rent ratio (P/R), which he says is an indicator of home overpricing. ``The P/R ratios for New York, Los Angeles, San Francisco, Boston and Philadelphia have surpassed the peak of the late 1980s bubble, and suggest prices could be roughly 25 percent too high compared to current rents,\'\' Morris says, adding that Chicago and Detroit also look expensive to him, while Dallas and Houston look ``comfortably priced.\'\' Situation Normal While Morris\'s observations concur with leading economists like Robert Shiller, the Yale University academic and author of ``Irrational Exuberance,\'\' real estate analysts and the Fed say he\'s off track. Fed Chairman Alan Greenspan, for one, has repeatedly discounted the existence of a housing bubble. A report titled ``Are Home Prices the Next `Bubble\'?\'\' by Jonathan McCarthy and Richard Peach of the Federal Reserve Bank of New York, disputes the pro-bubble camp. The New York Fed\'s report finds ``little basis for such concerns. The marked upturn in home prices is largely attributable to strong market fundamentals: Home prices have essentially moved in line with increases in family income and declines in nominal mortgage interest rates.\'\' McCarthy says strong demand is driving prices on both U.S. coasts where there are building restrictions and a shortage of available land for building. ``Here in New York, there\'s not much land left and there are various regulations, so it\'s not like they\'re going to be building a lot of condos,\'\' McCarthy says. Being Cautious Even within the Fed system, though, there\'s some room for discussion on whether housing prices are peaking. A new study by Morris Davis, a Federal Reserve economist, and Jonathan Heathcote, a Georgetown University economics professor, focuses more on the increase in land prices over the past five years. Davis and Heathcote concluded that U.S. land prices may post a ``cumulative decline of 6.3 percent\'\' over the next three years with nominal growth in home prices at 2.6 percent over that period. ``This would be the smallest three-year nominal increase for home prices on record,\'\' the authors say. The debate on whether home prices will rise or fall is largely academic because the Fed\'s interest-rate increase may only spur potential homeowners into a frenzy to take advantage of the lowest mortgage rates in a generation. In a possible nod to that view, Freddie Mac, the second-biggest buyer of home loans, predicted U.S. sales of new and existing homes will total 7.3 million this year, beating last year\'s record of 7.19 million. Speculative Frenzy? Sales and prices of existing homes hit record highs in June, the National Association of Realtors said in Washington today. Sales of previously owned homes unexpectedly increased 2.1 percent from a 6.81 million-unit rate in May, the trade association said. Existing home sales were forecast to fall to a 6.65 million annual rate in June, according to the median estimate in a Bloomberg News survey. The median selling price rose to a record $191,800. If the U.S. housing boom is coming to an end, buyers certainly aren\'t acting like the party\'s over. ``People are buying without money down and using interest-only loans,\'\' says George Marotta, a research fellow and former financial planner at the Hoover Institution in Palo Alto, California. ``I think we\'re coming close to the end\'\' of the housing boom, he says. ``It\'s not a good time to buy. I hope I\'m wrong, but it looks like the stock market bubble.\'\' Like many in the bubble school, Marotta is concerned about high household debt levels, which he calls ``the biggest bubble of all.\'\' Marotta says it\'s troubling that while personal debt has risen, average U.S. home equity has ``dwindled to 57 percent, compared with 85 percent a half-century ago.\'\' Watch for Overpricing Randy Johnson, a mortgage broker for Independence Mortgage Company in Newport Beach, California, says it pays to watch for speculation and overpricing in your market if you are buying now. ``I would put an appraisal contingency in every offer and get an appraiser I trust,\'\' Johnson says, advising the use of the lender\'s appraiser instead of the agent\'s to ensure that you\'re not overpaying for a home. ``What happens in this stage of the cycle is the sub-standard properties come on the market under the `if I can get that much for it, I\'ll sell it\' philosophy,\'\' Johnson says. ``I have seen people putting these houses into escrow and finding out upon inspection that they need to do a lot more work than they thought.\'\' ``I think that there really are people who are buying property on speculation that they can quickly resell at a huge profit,\'\' Johnson says. ``You can\'t build a market on speculators selling to the next wave of speculators.\'\' Just as all real-estate pricing is local, so should your focus be on your long-term financial goals and debt level. Stop worrying about whether your market is frothy and check if you\'re saving enough for emergencies, retirement and college. That\'s the best way to avoid another, more personally damaging kind of financial bomb. To contact the writer of this column: John F. Wasik in Chicago at jwasik@bloomberg.net. Last Updated: July 26, 2004 12:00 EDT  '),(2624,1960271,'08,August,2004','2004-08-08 00:00:00','   New York Times Sat, 07 Aug 2004 9:33 PM PDT O.K., the house you bought for $420,000 is appraised at only $400,000. And the house is assessed at $300,000. What is going on here?  '),(2625,1960271,'30,July,2004','2004-07-30 00:00:00','   Interactive agencies discuss role in changing market The online ad marketplace is rapidly changing, or that was the upshot of the Jupitermedia Advertising Forum on Wednesday. The main challenge, according to Jupiter analysts, is that clients increasingly are demanding accountability from agencies, and agencies are seeking outside help to provide what clients need.  '),(2626,1960271,'29,July,2004','2004-07-29 00:00:00','   Cable companies, which have been weighed down for years by heavy investments and sagging subscriber growth, are starting to turn the corner. But as they do, they are meeting formidable foes in emboldened phone companies.   urlLink Comcast , the nation\'s biggest cable operator, said yesterday that it earned $262 million in the second quarter, after losing $22 million in the same period last year, thanks to an increase in the number of customers signing up for digital video services and an increase in advertising revenue. Comcast\'s revenues rose 10.2 percent, to $5.1 billion, compared with a year earlier.  Comcast, which is based in Philadelphia, also said it would double its share buyback program to $2 billion. The company is trying to jump-start its stock, which has sagged since a bid for the  urlLink Walt Disney Company  failed earlier this year.  Sales at Time Warner Cable, the second-largest cable provider, increased 10 percent, to $2.1 billion, partly driven by a 25 percent gain in revenues from high-speed data services and digital video services. However,  urlLink Time Warner , the cable company\'s parent, said total company profits slipped 27 percent in the quarter from the period last year.  Comcast\'s acquisition of AT&amp;T\'s cable service less than two years ago vaulted it to the top of the industry. With more than 21 million television, Internet and phone subscribers, Comcast is about twice as large as Time Warner Cable.  Shares of both companies fell yesterday because of concerns that the regional Bell carriers are chipping away at the cable industry\'s command of the high-speed Internet market and because cable operators do not have a significant business selling phone services.  Comcast shares fell $1.18, or 4.1 percent, to close at $27.56 yesterday. Time Warner\'s shares dipped 1.5 percent, or 26 cents, to $16.65. Shares in  urlLink Cablevision  and other cable providers also slipped. In recent months, the regional Bells have teamed up with satellite providers to offer television services along with their telephone and broadband lines. Though the cable industry still controls the majority of residential broadband connections, the Bells are now adding more new customers per month. And phone companies like  urlLink Verizon  and SBC, which own mobile phone carriers, can offer bundled plans that include wireless services, which presents an advantage over cable providers.  \"I think there are a lot of challenges to keep growing next year for Comcast,\" said Alan Bezoza, who covers the cable industry for Friedman, Billings, Ramsey. \"At the end of the day, they need to have a bundle that includes voice.\"   Comcast currently sells phone service over traditional lines, but that business is very limited. Its revenues for phone services fell 13.8 percent, to $177 million, and the number of subscribers declined 10.4 percent, to 1.2 million. Comcast is planning to introduce an Internet phone service to compete with the regional Bells, but not for at least another year.  Time Warner Cable says it is on track to introduce Internet phone services in all its 31 markets by the end of the year, but did not provide company-wide subscriber figures. Despite the challenges, Comcast, with more than six million broadband customers, is still the largest broadband provider in the country.  Comcast and Time Warner Cable are starting to reap the benefits from their spending and building spree started in the 1990\'s. The cable industry spent about $85 billion combined to upgrade its networks so they could offer not only broadband Internet connections and telephone service, but also high-definition television and video-on-demand services.  That construction is largely complete, so the companies have been introducing new products and increasing their marketing efforts to recoup their investments. Comcast added 1.1 million new digital television customers in the last quarter, but also lost 96,000 basic cable subscribers. The digital television service costs more per month than basic cable television, so the increase in digital subscribers helped raise the company\'s revenues and the average amount customers paid per month.  The company\'s operating margins have returned to levels before it bought AT&amp;T\'s cable unit, a sign it has absorbed the costs integrating the two companies.  '),(2627,1960271,'27,July,2004','2004-07-27 00:00:00','   7/26/2004 9:19:53 AM By Steve Gelsi 9:19 AM ET Jul 26, 2004  NEW YORK (CBS.MW) -- Google increased the size of its IPO by about $1.1 billion Monday, with the Internet search firm eyeing a price range of $108 to $135 per share for 24.64 million shares.  Google also disclosed that it plans to trade its shares on Nasdaq under the ticker symbol \"GOOG.\"  The Mountain View, Calif.-based company will raise up to $3.3 billion in the IPO. Based on the middle point of the estimated price range, Google will raise $2.99 billion.  Including shares held back to cover underwriters\' overallotments, Google could raise up to $3.8 billion in the IPO -- well above the $2.7 billion maximum estimate in earlier filings. The price range signals strong interest in the profitable Internet search firm\'s IPO. The filing of an estimated range often marks the start of an IPO road show, when prospective companies meet with investors.  Morgan Stanley and CS First Boston are lead underwriters on the Google deal.  '),(2628,1960271,'22,July,2004','2004-07-22 00:00:00','   The amount spent on advertising on the Internet is set to surpass that spent on on commercial radio within three years, according to new research  The Internet advertising market in the UK amassed an estimated Ã‚Â£350 million in 2003, which according to industry trade group, the Internet Advertising Bureau (IAB) UK, puts the industry on track to surpass the commercial radio market withinin 3 years.    The IAB\'s UK annual report revealed that advertising revenues surged 80 percent year-on-year, boosted by the demand for corporate-sponsored search listings and a surge in new users.  The growth figure is higher than the group\'s estimate published in earlier in the year when it said UK online advertising market would grow at around 50 per cent in 2003, the reuters news agency reports.  The IAB UK forecast online advertising revenues are expected to Ã‚Â£600 million by 2007, exceeding those expected by the commercial radio market.   22 July 2004     ......CONFIDENTIALITY NOTICE...... The information in this email may be confidential and/or privileged.  This email is intended to be reviewed by only the individual or organization named above.  If you are not the intended recipient or an authorized representative of the intended recipient, you are hereby notified that any review, dissemination or copying of this email and its attachments, if any, or the information contained herein is prohibited.  If you have received this email in error, please immediately notify the sender by return email and delete this email from your system.  Thank You.  '),(2629,1960271,'19,July,2004','2004-07-19 00:00:00','   July 19, 2004 Microsoft\'s Online Unit Shows Signs of Life By SAUL HANSELL   t an October 2002 gala to introduce Microsoft\'s MSN 8 Internet service, Bill Gates showed a video of himself wearing a butterfly costume and handing out software for the service.   His point was that after seven years of floundering in the Internet access market, Microsoft would do whatever it took to beat its longtime rival America Online. That included pouring $500 million into MSN to match AOL\'s most popular features and another $300 million into marketing, featuring MSN\'s butterfly mascot and a concert highlighted by rocker Lenny Kravitz singing \"Fly Away.\"  But all that flew away was the better part of a billion dollars. Microsoft\'s share of the Internet access market has declined steadily since.  In the years since the start of what was first called the Microsoft Network, Mr. Gates has tried dozens of different business models, from Internet access to Web sites to monthly software subscriptions. They all had one thing in common: they lost money.  Over the last year, however, MSN has finally started to see some profits. The unit began making money last fall and is expected to post an operating profit of about $200 million for the fiscal year that ended June 30, compared with a loss of about $531 million last year.  The reason has little to do with any of Microsoft\'s more ambitious Internet strategies. Instead, it was one of the businesses that it had put on the back burner - Internet advertising - that really started to take off.   Even without a lot of new investment, MSN\'s Web site has long been the third most popular destination on the Internet, bringing together its Hotmail e-mail service, the MSNBC news operation and a variety of other channels. Recently, it has worked to repair a frayed relationship with Madison Avenue, putting the company in a good position to gain as the ad market rebounded. Over the last year, MSN has ranked second in online advertising revenue, behind Yahoo and ahead of the longtime leader AOL, a unit of Time Warner.  Mary Meeker, the Morgan Stanley Internet analyst, said that MSN\'s revival was an indicator of the company\'s trademark doggedness.   \"They began their attack on AOL 10 years ago,\" she said. \"Since then they have lost billions, and now they are making money. It gives me a real sense of just how patient they are.\" Now basking in its newfound profitability, MSN is again on the offensive, competing with Yahoo, the Internet portal, even more than with AOL. It has a three-prong strategy: continue to cash in on the advertising market, sell software services for people who have broadband Internet access from cable or phone companies, and build a better Web search engine. MSN, which uses Yahoo\'s search technology, is trying to develop search technology to take on Google. It hopes to introduce that technology by the end of the year.   As investors consider Google\'s public stock offering, some worry that Microsoft will be able to squeeze out the plucky search company just as it pushed aside Netscape in the Internet browser market. After all, it plans to include links to its new search service on the Windows desktop and in many other products and services.   Yet that same considerable Windows advantage led many observers to predict that Microsoft would cause more trouble than it actually did for AOL and Yahoo. It turned out that most Internet users were willing to skip past the default services offered by Windows in favor of brands they preferred.   \"There was a concern that Microsoft would always win because they owned the desktop,\" said David Card, the research director of Jupiter Research. \"If the desktop is so important, then Microsoft should have beaten AOL, but they didn\'t.\"  One drag on MSN over the years, said several people who worked there, is that Microsoft has been conflicted about whether to treat the Internet as a media business or a form of software. It is not a simple either-or proposition because so much of the Internet is a hybrid - free software paid for by advertising.   \"One thing that amazed me is that the organization changed annually,\" said Merrill Brown, the former editor in chief of MSNBC. \"They seize upon a good idea, and throw lots of resources at it until it\'s time to move to the next idea or acquisition.\"   Microsoft certainly has had some successes on the Internet, including Hotmail, MSNBC (a joint venture with General Electric), the Slate opinion journal and Expedia, the online travel agency it sold to InterActiveCorp.   But there have been at least as many failures, ranging from Mungo Park, an adventure travel site, to the CarPoint automobile buying service, to the controversial Hailstorm system for connecting information from different Web sites.  But Microsoft\'s biggest online failure has been its Internet access business. It lost several billion dollars over the years by paying people $400 rebates to sign three-year contracts for MSN dial-up accounts, according to Internet industry executives. Those customers have quit the service in droves as their commitments expire.   By contrast, AOL\'s Internet access business has been hugely profitable, even amid turmoil from its merger with Time Warner and the decline in the dial-up, or narrowband, market.   \"We said that narrowband is like the buggy whip business,\" said David Cole, the Microsoft senior vice president in charge of MSN. \"We could try to be the best buggy whip maker, but that\'s not very aspirational. It\'s hard to get people rallied around that.\"  The broadband market, however, has proved exasperating for dial-up providers like MSN, AOL and Earthlink, all of which wanted to buy high-speed connections wholesale and market them to consumers.  After accounting for the customer acquisition costs, the costs paid to telephone companies and the support costs, Mr. Cole said, \"there is nothing left over.\"   Microsoft has invested $6 billion in Comcast, the nation\'s largest cable company, but that has not led to an arrangement to distribute MSN services to Comcast customers. Comcast, in fact, edged out Microsoft this year to become the nation\'s second-largest Internet provider, with 10 percent of the Internet access market, according to Solomon Research.  Microsoft\'s share of that market has fallen from a peak of 13 percent in 2002 to 8 percent today. (AOL\'s share fell to 24 percent from 29 percent in the last two years.) Microsoft claims a total of eight million subscribers to its various MSN services, of which about five million are Internet access customers, according to ComScore Media Metrix, an Internet ratings company.   Now, the company hopes to sell extra services to people who buy broadband service from their local telephone or cable companies. For $5.95 a month, it offers MSN Plus, which mainly expands e-mail services. MSN Premium, for $9.95 a month, adds firewall and virus software, and access to some proprietary content, most notably Webcasts of Major League Baseball games. Microsoft paid a reported $40 million over two years for the baseball rights.  So far, that market has not taken off. Microsoft will not say how many broadband add-on subscriptions it has sold, admitting only that the vast bulk of its users come from deals it has struck with Verizon and Qwest to include MSN software as part of those company\'s basic broadband services.   Microsoft is also developing more services, including some for small businesses, and a new music download offering meant to compete with Apple\'s iTunes.   But it is MSN\'s advertising business that is really starting to pay off.   The company had a reputation on Madison Avenue of being unresponsive and forcing advertisers to use its own proprietary formats. But in late 2001, it hired an experienced magazine executive, Joanne Bradford, to run its advertising sales staff and to reach out to agencies. It also moved to embrace the industry\'s standard ad formats. As marketers discovered ways to use more interactive ads, MSN was one of the few sites that could deliver a mass audience.  Still, MSN\'s audience appears to be more transient than those of its rivals. According to ComScore Media Metrix, 111 million people visited some part of MSN at least once in May. That is just slightly behind Yahoo, which had 113 million visitors and Time Warner (including America Online), which had 112 million. But MSN\'s visitors spent 147 minutes, on average, looking at MSN pages in the month, far less than the 260 minutes that users spent on Yahoo or the 340 minutes they devoted to Time Warner sites.  Yusuf Mehdi, a vice president for MSN, concedes that the company needs to find a way to engage its audience more.   \"We want to have the sort of evangelical feedback that people have for the iPod or the TiVo,\" Mr. Mehdi said. \"I want people to say, \'I love MSN.\' \"   MSN hopes one way to achieve that is to customize its service to the interests of each user. For example, it has introduced trial versions of an automated news service that displays headlines for articles based on the topics that users have shown interest in.  Another part of Microsoft\'s Internet strategy is to combine Web searches with an improved application that lets users find documents and e-mail messages on personal computers. The enhanced search function will be a feature of the next major version of Windows, scheduled for 2006, but Mr. Mehdi said that MSN would make a desktop computer search feature available much sooner, probably as part of the free downloadable toolbar that it introduced earlier this year. And last week Microsoft bought Lookout, a tiny software company, to help with its desktop search effort.  But even with this new focus on searches, Danny Sullivan, the editor of Search Engine Watch, said that Microsoft will have a tough time catching up with Google. \"To shift market share, Google has to get bad and they have to get good,\" he said. \"People will not go to Microsoft, even if Microsoft is better than Google - if Google is good enough.\"  Mr. Mehdi said that Microsoft was betting that none of today\'s search engines are quite good enough to hold users and that Microsoft\'s strengths will give it a significant advantage as the Internet evolves.   \"Search is not done until you can give users the answer they want instead of a list of 1,000 Web sites,\" Mr. Mehdi said. \"That\'s a big, tough software problem. And we said, thank goodness we\'re software guys.\"    Copyright 2004 The New York Times Company      ......CONFIDENTIALITY NOTICE...... The information in this email may be confidential and/or privileged.  This email is intended to be reviewed by only the individual or organization named above.  If you are not the intended recipient or an authorized representative of the intended recipient, you are hereby notified that any review, dissemination or copying of this email and its attachments, if any, or the information contained herein is prohibited.  If you have received this email in error, please immediately notify the sender by return email and delete this email from your system.  Thank You.  '),(2630,1960271,'16,July,2004','2004-07-16 00:00:00','   THE TOP 50 INTERACTIVE AGENCIES  Rank	Agency (Affiliation)	Headquarters	2003 projected InterACTIVE revenue (in thousands)	2002 actual revenue	percent increase	2003 employees	 1	Euro RSCG Interaction (Havas) 1	New York	$260,000* $239,000	9%	1,225*	 2	Grey Digital Marketing (Grey Global Group)	New York 185,000*	161,000	15	1,070*	 3	OgilvyInteractive (WPP) 	New York	176,000* 160,000	10	1,150*	 4	SBI.Razorfish 2	Salt Lake City	120,000	75,000	60	700	 5	Agency.com (Omnicom) 3	New York	84,000*	80,000*	5 348	 5	Digitas	Boston	84,000*	72,000*	17	440	 7	Zentropy Partners (IPG) 	Los Angeles	76,400*	69,500 10	490	 8	Tribal DDB (Omnicom)	New York	68,000*	52,000	31 475*	 9	R/GA (IPG)	New York	67,800*	56,500	20	264	 10	Digital@JWT (WPP)	New York	63,000*	55,600	13 240	 11	Modem Media 4	Norwalk, Conn.	60,500*	70,000	-14	218	 12	Wunderman Interactive (WPP) 5	New York	57,000*	49,198 16	305*	 13	Avenue A/I-Frontier (aQuantive)	Seattle	52,500	45,900	14 225	 14	Carat Interactive (Aegis) 6	Boston	47,000	36,000	31 271	 15	FCBi (IPG)	New York	45,500*	36,400	25	257	 16	DraftDigital (IPG)	New York	44,000*	39,192	12 237	 17	Organic (Omnicom) 7	San Francisco	43,000*	37,802	14 225	 18	iDeutsch (IPG )	New York	42,000*	38,056	10	131	 19	Tocquigny Advertising, Interactive + Marketing 8	Austin, Texas	41,000	33,000	24	80	 20	Critical Mass (50% owned by Omnicom) 	Chicago	37,000*	32,000 16	250*	 21	AKQA	San Francisco	36,000	30,000	20	280	 22	The Digital Edge (WPP) 5	New York	34,000*	30,000 13	195	 23	Nurun (Quebecor) 	Montreal	32,000	39,000	-18 398	 24	Medical Broadcasting Company (10% owned by WPP)	Philadelphia 28,100*	26,000*	8	141	 25	DVC Worldwide 	Morristown, N.J.	27,600	27,500	0 56*	 26	Universal McCann Interactive (IPG)	New York	25,100* 22,800	10	150*	 27	IconNicholson	New York	24,000	21,000	14	140	 28	Macquarium Intelligent Communications	Atlanta	24,000	20,000 20	116	 29	imc2 	Dallas	23,700	16,400	45	113	 30	VML (WPP)	Kansas City, Mo.	23,500*	18,800	25 85	 31	Planning Group	Miami	23,500	18,000	31	87	 31	Semaphore Partners (Publicis) 9	New York	23,000*	27,000 -15	140*	 33	Slingshot	Dallas	22,500	17,000	32	45	 34	Starcom IP (Publicis)	Chicago	20,900*	19,000*	10	125*	 35	Wirestone	Emeryville, Calif.	20,000*	20,000	0 109	 36	Blast Radius	New York	19,000	17,400	9	225	 37	Blue Dingo	New York	18,000	17,000	6	56	 38	Risdall Advertising Agency	New Brighton, Minn.	17,500 16,000	9	29	 39	Genex	Los Angeles	17,000	12,000	42	100	 39	Molecular	Watertown, Mass.	17,000	11,700	45 77	 41	Campbell-Ewald Digital (IPG)	Warren, Mich.	16,700*	16,641 0	85	 42	RappDigital (Omnicom) 	New York	15,500*	14,100	10 84*	 43	atmosphereBBDO (Omnicom)	New York	14,500*	7,900 84	80*	 44	Multimedia Solutions	Edgewater, N.J.	13,500	9,900	36 85	 45	Ten/Resource	Columbus, Ohio	12,500	10,500	19	60	 46	DNA Studio	Los Angeles	11,200	9,100	23	102	 47	Refinery	Philadelphia	10,900	10,400	5	78	 48	Cadient	Conshohocken, Pa.	10,500	8,700	21	50	 49	SimStar Internet Solutions	Princeton, N.J.	10,000	13,000 -23	54	 50	HSR Business to Business	Cincinnati	9,600	9,100 5	24	  *	Estimated numbers	 1	Absorbed Havas siblings Magnet, Brann, others from Arnold and Media Planning Group	 2	Acquired Razorfish, March 2003; Xcelerate, January 2003	 3	Acquired by Omnicom, July 2003	 4	Interpublic divests to 0.6% ownership, December 2003	 5	Wunderman Interactive media billings merged into The Digital Edge, June 2003	 6	Acquired Freestyle Interactive, May 2003	 7	Acquired by Omnicom, February 2003	 8	Fiscal year numbers reported, April \'02 through March \'03 /font>  9	Combined into Arc North America, January 2004	    IBCTV 165 N. Morgan Street  Chicago, IL 60607  P: 312-738-0740 x 235  F: 312-738-0748  www.ibctv.com  This message may contain information that is proprietary or confidential and is intended only for the addressee named above. If you are not the addressee or authorized to receive this message for the addressee, you may not use, copy or disclose to anyone this message or its contents. If you have received this message in error, please notify me immediately by telephone, permanently destroy any copies you may have and return this email to me at the email address noted above.       ......CONFIDENTIALITY NOTICE...... The information in this email may be confidential and/or privileged.  This email is intended to be reviewed by only the individual or organization named above.  If you are not the intended recipient or an authorized representative of the intended recipient, you are hereby notified that any review, dissemination or copying of this email and its attachments, if any, or the information contained herein is prohibited.  If you have received this email in error, please immediately notify the sender by return email and delete this email from your system.  Thank You.  '),(2631,1960271,'15,July,2004','2004-07-15 00:00:00','   -----Original Message----- From: adweekemail@adweek.com [mailto:adweekemail@adweek.com]  Sent: Thursday, July 15, 2004 8:41 AM To: adweekemail@adweek.com Subject: Digitas Acquires Modem Media  BREAKING NEWS FROM ADWEEK ONLINE * MEDIAWEEK ONLINE * BRANDWEEK ONLINE -- http://www.adweek.com -- ------------------------------- SPONSOR ------------------------------ THE NEXT BIG IDEA: THE FUTURE OF BRANDED ENTERTAINMENT event DVD If you missed the landmark event, don\'t miss this opportunity to purchase a DVD of top-rated presentations for only $195! Hosted by Adweek Magazines, The Hollywood Reporter, Billboard, Sporting Goods Business, Shoot, and Nielsen Media Research, in association with NYU - Stern School of Business.  Presented by the Hallmark Channel. http://www.thenextbigidea.com ----------------------------------------------------------------------  Digitas Acquires Modem Media  BOSTON Digitas today said it has agreed to buy Modem Media in a stock-for-stock transaction.  Based on yesterday\'s closing price for Digitas common stock, the acquisition is valued at approximately $200 million on a fully diluted basis, according to Digitas, a Boston-based interactive and direct marketing company with about 1,100 employees worldwide.  Norwalk, Conn.-based Modem has 270 employees and additional offices in London and San Francisco. Following the merger, Modem will continue to operate under its existing name and will be led by Martin Reidy, who currently is president of Digitas in San Francisco.  Both Digitas and Modem are independent agencies.  --Lisa van der Pool  ------------------------------- SPONSOR ------------------------------ Introducing Adweek\'s Marketing y Medios magazine, a new magazine for Hispanic marketers.  This monthly English-language magazine will launch in September 2004.  To find out more about this exciting new launch, click here. http://www.adweek.com/aw/images/Marketing_Y_Medios.pdf ---------------------------------------------------------------------- BREAKING NEWS CAN BE FOUND THROUGHOUT THE DAY AT   http://www.adweek.com  http://www.mediaweek.com  http://www.brandweek.com  Copyright 2004 - VNU Business Media, 770 Broadway, 6th Floor, New York, NY 10003 - All Rights Reserved.  ----------------------------- You are currently subscribed to the FREE adweekemail as: dtassone@i-dep.com To unsubscribe to this FREE service, send a blank email addressed to leave-adweekemail-2355945Q@listserv.vnuemedia.com      ......CONFIDENTIALITY NOTICE...... The information in this email may be confidential and/or privileged.  This email is intended to be reviewed by only the individual or organization named above.  If you are not the intended recipient or an authorized representative of the intended recipient, you are hereby notified that any review, dissemination or copying of this email and its attachments, if any, or the information contained herein is prohibited.  If you have received this email in error, please immediately notify the sender by return email and delete this email from your system.  Thank You.  '),(2632,1960271,'14,July,2004','2004-07-14 00:00:00','   New York Times Digital, the online arm of New York  Times Co., more than doubled its operating profit in the second  quarter thanks to increased demand for online advertising. The  Internet unit, which publishes NYTimes.com and Boston.com,  reported yesterday that sales grew 27 percent in the quarter  compared with the prior year, totaling $27.4 million. Operating  profit ballooned from $4.3 million to $8.9 million. http://www.dmnews.com/cgi-bin/artprevbot.cgi?article_id=29600       IBCTV 165 N. Morgan Street  Chicago, IL 60607  P: 312-738-0740 x 235  F: 312-738-0748  www.ibctv.com  This message may contain information that is proprietary or confidential and is intended only for the addressee named above. If you are not the addressee or authorized to receive this message for the addressee, you may not use, copy or disclose to anyone this message or its contents. If you have received this message in error, please notify me immediately by telephone, permanently destroy any copies you may have and return this email to me at the email address noted above.       ......CONFIDENTIALITY NOTICE...... The information in this email may be confidential and/or privileged.  This email is intended to be reviewed by only the individual or organization named above.  If you are not the intended recipient or an authorized representative of the intended recipient, you are hereby notified that any review, dissemination or copying of this email and its attachments, if any, or the information contained herein is prohibited.  If you have received this email in error, please immediately notify the sender by return email and delete this email from your system.  Thank You.  '),(2633,1960271,'14,July,2004','2004-07-14 00:00:00','   A rather think account by self-proclaimed marketing wonk Steve Hall.  '),(2634,1960271,'14,July,2004','2004-07-14 00:00:00','   According to an April 2004 survey published by the e-tailing group, the vast majority of Internet retailers rank customer acquisition and customer retention as their most important marketing strategies, followed closely by converting online browsers into buyers. Another 86% of online merchants rate the need to drive traffic to their Web site as an important priority as well. Not surprisingly, the e-tailing group has found that online retailers are increasingly shifting their marketing dollars to the Internet, due in part to the fact that online marketing is easy to test and refine. Furthermore, online marketing continues to be less expensive than most offline advertising channels.  www.eMarketer.com  '),(2635,1960271,'12,August,2004','2004-08-12 00:00:00','   Contact: Tom Genaro 908-580-1959  The Coalition has vented it\'s displeasure with A&E and Victoria Gotti for the \"Growing up Gotti\" TV reality show.  Executive Director, Dr. Emanuele Alfano, stated, â€œItâ€™s disparaging to note the A&E recent bent toward assaulting the Italian American Community with defamatory programming.  First was the Medici series, which was flavored with a Sopranosesque theme, and now a weekly program on the trials and tribulations of a Mafia daughter. Victoria Gotti does not represent Italian American women any more than Heidi Fleiss represents women of her community, but both are cashing in on their notoriety. Each has no concern of further tarnishing their personae as long as the linings of their pockets are thickened with dirty money.\"   Dr. Alfano went on to state that the media has fed the public\'s appetite for the mob genre by offering vowel ending surname characters who are brutal and immoral.  Unfortunately, this frenzied feeding has further deepened the prevailing myth of synonymy between Italianita and criminality. As the more astute critics have noted, Italian Americans are the last easy target upon which to hurl defamation without fear of retaliation.  However, let it be known by all detractors that that day has ended and a new and determined united coalition of Italian American Organizations will rise against any new attempts to diminish our identity as honorable American citizens.  Further statements accused A&E and of ill will towards Italian American women by continuing to portray them other than the Patricia Russos, the Carli Fiorinas and the Laura Savinis. The listing of notable Italian American women is extensive, but the lower end of that listing seems to be preferred over the well educated sophisticated women that the Italian American Community is indeed proud of.   Dr. Alfano urged both A&E and Victoria Gotti to recognize the damage which will be incurred if the \"Victoria Gotti Show\" is allowed to spew out weekly, a distorted image of the true Italian American woman. â€œWhat Ms. Gotti does not understand is that the audience will be laughing at her and certainly not with herâ€, said Dr. Alfano.  He ended by imploring A&E to withdraw the program as an act of decency.  '),(2636,1960271,'06,August,2004','2004-08-06 00:00:00','   NEW YORK 24/7 Real Media yesterday reported a 61 percent revenue rise on net income of $1.5 million in the second quarter versus the year-ago period.   The online advertising and technology company recorded Q2 revenue of $19.6 million, up from $12.2 million in the same period last year.   Second-quarter net income was $1.5 million, or 4 cents per share, compared to a net loss of $4.4 million, or 31 cents per share in Q2 2003. The net-income figure includes a gain of $2.9 million from a legal settlement with chinadotcom corp.   The New York-based company said it expects Q3 revenue of $19.5-20.5 million, noting that the company traditionally shows little growth during the period as a result of seasonality factors. Third-quarter diluted pro forma operating income, likely to be affected by the $600,000-750,000 onetime costs associated with the company\'s Sarbanes-Oxley compliance efforts, is forecasted to be 0-1 cent per share.   For the full year, 24/7 projected revenue of $80-82 million and pro forma earnings of 4-7 cents per share.   24/7 shares (TFSM) were trading on the Nasdaq SmallCap Market today at $2.78, down 24 cents or 8 percent.   --Ann M. Mack  '),(2637,1960271,'04,August,2004','2004-08-04 00:00:00','   After hearing the buzz about behavioral targeting, it seemed to make sense to put that term into perspective, writes Poindexter Systems\' Joe Zawadzki. Behavioral marketing, despite its five-year lifespan, had really just gotten off the  ground in terms of product evolution at a time when the space -- particularly online marketing and technology startups -- was contracting. Now, as companies (re)turn to data-driven marketing, some key techniques and principles should be incorporated at the start of this renaissance to ensure solid results and sustainable growth. To begin with, companies should look to their roots. Companies with the most success in  online behavioral marketing are those that leverage their offline expertise in this area. Marketers long have relied on audience segmentation and clustering to find the right segments for their varied messages.  '),(2638,1960271,'03,August,2004','2004-08-03 00:00:00','   Forbes.com is experimenting with an ad placement technique from Vibrant Media in San Francisco that puts links to advertisers\' sites directly into the text of stories. Some publishers worry the strategy might erode the traditional separation between journalistic content and advertising. But Jim Spanfeller, president and chief executive at Forbes.com, said \"We want to be trailblazers, whether from an advertising standpoint or an editorial standpoint\" and added that the practice will end if readers express confusion or disapproval.  '),(2639,2224636,'26,November,2003','2003-11-26 00:00:00','    congratulations. you are the kiss my ass happy bunny. You don\'t care about anyone or anything. You must be so proud    urlLink   which happy bunny are you?     brought to you by  urlLink Quizilla   '),(2640,2224636,'23,November,2003','2003-11-23 00:00:00','    this is just how i feel   have you felt like you know you exsist in the world but no one acknowledges it?  yeah because that\'s how I feel right now.  it\'s really funny when you look at your e-mail and all you\'ve got is spam.  not one thing from a friend.  or you try calling some one and then they never call you back.  you spend time with one person you love most in the world but feel like you\'re just taking up space this is just how i feel i feel like i\'m in a world that knows i\'m there but doesn\'t want me to be i would love some one to notice the things that i do or the things that i say  this is just how i feel   '),(2641,2224636,'21,November,2003','2003-11-21 00:00:00','    I HAVE GONE QUIZZLIA CRAZY.........  is this healthy???   '),(2642,2224636,'20,November,2003','2003-11-20 00:00:00','    urlLink     What Finding Nemo Character are You?     brought to you by  urlLink Quizilla   '),(2643,2224636,'20,November,2003','2003-11-20 00:00:00','    Bondage movie! You\'re into BSDM (Bondage & Discipline, Dominance & Submission) and chances are, you\'re fond of whips, chains, harnesses, and tight leather outfits. You like to mix a little pain with a LOT of pleasure, baby!    urlLink   What kind of porno would you star in?     brought to you by  urlLink Quizilla   '),(2644,2224636,'20,November,2003','2003-11-20 00:00:00','    Everyone remembers the \'faked-orgasm-in-a-deli\' sequence from your kind of movie When Harry Met Sally. It seems that you\'re falling for a buddy or have already fallen for them. Uh-oh. You\'re probably caught between the possibility of having a great relationship and wrecking the one you have now. You know what they say, it\'s better to regret something you did than something you didn\'t do.    urlLink   What Romance Movie Best Represents Your Love Life?     brought to you by  urlLink Quizilla   '),(2645,2224636,'19,November,2003','2003-11-19 00:00:00','    hee hee i just think the bear is cute....seriously   '),(2646,2224636,'08,November,2003','2003-11-08 00:00:00','    Tramp Bear    urlLink   Which Dysfunctional Care Bear Are You?     brought to you by  urlLink Quizilla   '),(2647,2224636,'08,November,2003','2003-11-08 00:00:00','    Quit Leaking DAMN YOU   You know what noses suck. Especially when you have a cold and no matter what you do they leak and leak and leak and leak. It\'s a very unattractive quality when the leak goes out of your body and nasty when it tries to retract itself into the back of your throat. And no matter what you do you can get it to stop leaking. You could stuff tissue up your nose and yet some how the liquid finds its way out or soaks the tissue to the point where it pushes it out. Then you take a bunch of cold medicine thinking that will help your nose b/c if I was a nose and leaking as much as mine is I would be tired of it. But no taking that medicine was a mistake because IT MADE IT LEAK MORE.  DAMN THE INVENTOR OF COLDS!!!!!!!   '),(2648,2224636,'07,November,2003','2003-11-07 00:00:00','    Just a Thought   Today at my school during the last period they had this festival celebration thingy for the renaissance card holders (some gay program that makes those of us who don\'t have it feel stupid) and me and a couple of my friends were \"left behind.\"  Michel and I were talking about random stupid things and some how we get on the subject of short buses. And here\'s that conversation: Emily- I\'ve always wanted to own a short bus I think that would be so great! Michel- you know what I wonder about short buses? Emily- What\'s that? Michel- Why the numbers are like 6 digits. Emily- (confused) Um ok? Michel- No I mean like mostly EC kids ride those buses and most of them can\'t remember anything so how in the hell are they going to remember 6 digits for a damn bus. Emily- Yeah it\'s like the number is longer than the bus!!!  (LOTS OF LAUGHTER even from the teacher)  Best quote ever from a teacher- Emily you spell masturbation with a \'u\' not an \'e\' I think I would know.   '),(2649,2224636,'05,November,2003','2003-11-05 00:00:00','    3.62 3.62 3.62 woo woo and buses   Yes that is my GPA. Which I am insanely proud of. Most people that are my friends and are in the really upper scale classes are like \"Well I got a 4 or above.\" And I\'m like \"Well I could care less this is the best I\'ve done in my entire school career.\" And run off giggling. Today was a good day. I was really happy about all my grades. Now maybe I could drive to school and not have to wait on this old fart woman who drives my bus. I wouldn\'t have to wait for a woman who waits until the entire bus parking lot is clear so she can turn the bus on. *eye twitches* It\'s really retarded like.  I said to my friend Hayley and also fellow sufferer of the damn bus driver while staring at the woman \"I can not wait to drive to school THEN I CAN LEAVE WHEN I EVER I DAMN WELL PLEASE.\" *angelic Simpson like music plays in the back ground* Yes I have issues. I know this leave me the hell alone. Oh I\'m in such a good mood. Well I\'m going to go eat pizza and go see my boyfriend. YAY!   '),(2650,2224636,'03,November,2003','2003-11-03 00:00:00','    OH lookit   OK thanks to Nick I have a blog and it\'s not all crappy and stuff thanks nickus. hmmmmm i really don\'t have much to say right now bye bye   '),(2651,2224636,'01,November,2003','2003-11-01 00:00:00','    Hello, peoples. This is Nick McIntyre testing out the Blog he made for Emily Green   '),(2652,2224636,'24,January,2004','2004-01-24 00:00:00','    A week ago   A week ago I had my son. A month early. It was so frightening I thought he might not live through the night. But he did. He\'s so strong to be so tiny. You never really understand was absolute true love is until your very first child is placed in your arms and you see that tiny life sleeping. It was so hard to let him go but  I know what I did was right and I know that the family that has him will give him all that love and the best life that he deserves.   '),(2653,2224636,'12,January,2004','2004-01-12 00:00:00','    A confession you\'ll never see   All I\'ve ever wanted in life was to be love unconditionally by someone.  (And no not by family if you have the wrong idea) One person in partcular. I would just love to have that person hold me in there arms again and kiss me. Making the world feel right again. But no, that person is gone. Doesn\'t want me. Why? It\'s great that we\'re friends and I\'ll accept that with time. But for all that I\'ve been through and all that I\'ve done, couldn\'t we give \"us\" one more shot? I always love the way you make me feel, the way you make me laugh, how you always say the right thing to make my tears go away.  You rushed right to the hospital with a moment to lose as soon as you found out it was time. You held my hand when I became to scared to deal with reality. You spent the entire time I was in the hospital right by my side. Wiping my tears away every time I thought I couldn\'t handle what had to be done. You were my sanity and my strength. Telling me what I chose to do was right.  You were there to hold me when I couldn\'t sleep. Calming me from my nightmares. You spent all your time that you could have had to work to make sure I was going to be ok. You even came home with me after everything was said and done and held me by your side so that I couldn\'t sleep. But no, that obviously doesn\'t mean as much to you as it does to me. I don\'t want to be your friend. I want to be yours.  Is it that hard to see you and I are supposed to be together. Or am I just crazy? I guess this is a confession of feeling you\'ll never see.   '),(2654,2224636,'12,January,2004','2004-01-12 00:00:00','    I won\'t cry I know I won\'t die But seeing you now is harder than it ever was before I didn\'t want to leave I didn\'t want you to go Was what we did right? Was it nothing to fight? I feel a mistake has been made Soon it will all fade  But not now I don\'t need this Not ready to take that bow I just need you a little longer Just to feel a little stronger I\'m not grounded I\'m lost In a desert full of frost  I have yet to be able to cry It stunned me that much LIfe is such a crutch  Come back Right now I mean it this time No more arguments No more turning on a dime  If there was ever a time I needed you most I think it would be now.   '),(2655,2224636,'03,January,2004','2004-01-03 00:00:00','    Wow   It\'s been a while since I\'ve gotten on and written anything. Hmmmm well what to say. A lot has gone on in the past month. Parents found out I\'m pregnant. OH wait no one knew that except me and Jerry. yeah so ok I\'m pregnant. Having a boy. Going through the adoption process. Found out that sex can really fuck up your life. Due in February. Yeah I\'m that far along. Had an ultrasound done and they say the baby looks great and healthy and that I\'m doing great. Got a puppy for christmas and it will not SHUT UP!! It has these like psychotic moments where it just barks and barks for no damn reason. It\'s actually quite funny. I really wonder if anybody is going to read this? My dad keeps cussing at the TV because of the football game that\'s on now that is really funny. God you think for not coming on and writing I\'d have more to say. Hmmmm. But I don\'t    \"If the world hated you so much, you\'d be dead by now\"   '),(2656,2224636,'29,February,2004','2004-02-29 00:00:00','    Gawd dammit YOU!   OK I just gettin a little confused as to why you act somethings there. I thought it was over. I thought it was done. I love you and you know that. But you want me to move on so damn bad but you\'re not lettin me. You keep invadin my thoughts and dreams. Thats makin it hard.   And you, you led me on and on for like what almost over a year now? So I tell you how I feel about you and all I get is a muthafuckin OH?? What the fuck? You\'re not that damn great. Oh wow every chick loves you. OOOOOOHHHHHHHH. But you make it that way. It\'s aalllll you dawg. I\'m tired of chasing you.  You want me? You come and get me!    Mood: Tired of it all Song: #2 on Mostastless by Twiztid   '),(2657,2224636,'28,February,2004','2004-02-28 00:00:00','    SKITTLES I LOVE YOU!!!!    Skittle said to me tonight: jerry loves you but doesnt want to be tied down right now..just give it time let him go play like I need to do and pinz can just be a jackass sometimes and thats part of him supposedly he didnt used to be like that but he knows he\'s a &quot;hunk&quot; and hes an asshole about it so forget him and act like he doesnt fucking exist anymore..   maybe then he\'ll see u as unatainable and be interested.. ;)    Damn as I read over that again and again it make more and more sense. I\'m gonna stop acting like I want them so damn bad and just act like I don\'t care how they feel anymore (SKittles make sure I do this K??)   I just need to go out and have fun and stop worryin about be in a damn relationship. Fuck it I\'ll get one of those soon enough. Maybe a great one.  But G-d almighty. I\'m 16. I had a son. I bounced back from that like it was nothing, even though it was. I need to go out have fun. BE A TEENAGER.  G-d, I gotta stop worryin over stupid shit. Thanks Skittles, I LOVE YOU.  MY BESTEST BUDDY IN THE WHOOOOOOOLEEEE WIDDDDEEE WORLD!!!!!   Mood: Totally enlightened and happy Song: IT\'S RAININ MEN!!!!!   '),(2658,2224636,'27,February,2004','2004-02-27 00:00:00','    NOOOO DAMN U SNOW   ::kick, thrash, bang, hit, punch, smack bang (if you know me you know what this looks like) smack, smack, hit desk, bang, kick, thrash!!::  STUPID DAMN STUPID STUPID SNOW!!!!!!! PHU-Q SNOW PHU-Q TO ALL HELL!! Me and Skittle were gonna go out of town to see a show (Smile Empty Soul) and we were already to go BUT NOOOOOOOOOOOO IN COMES SNOW AND BAD ROADS!!  We were ready to party and, and, and have fun, see some hott guys, I was gonna pick up some hott guys, and have a good time.  BUT NOOOOOOOO THE SNOW CAME AND RUINED IT ALL!! RUINED IT!! MESSED IT ALL UP!!  ::kick, thrash, bang, hit, punch, bang, smack, smack, hit desk, bang, kick, thrash, spazes out on the floor.::  (Oh look for the link to pictures and guest book. New things all the time)    Mood: SO SO MAD *GRRRRRRR* Song: Fuck the world by ICP   '),(2659,2224636,'22,February,2004','2004-02-22 00:00:00','    Both of you....OUT    OK so here I am smokin a cigarrett. Just finished talkin to Skittle about how shes confused about something and i try to make her feel better and not talk about my issues to her (for once time for me to listen). And I just burst in to tears. I\'m still semi-crying.  I sit here thinkin about how much I wish I could be loved once again. I mean Smurf sits there one night while were talking and tells me \"You\'ll always have my heart, and I\'ll always have yours.\" God how true it is. I mean him having mine. But will I always have his??   Then he busts out with \"God, you are such a trophy to have. Such a beautiful trophy.\" Really am I muthafucka?? Am I?? THEN WHY IN THE FUCKING HELL DID YOU GIVE UP ON US!!?? WHY?   And why is it that when I told Pinz how I felt, he just went \"Oh!\"  If I\'m such a beautiful person, trophy, whatever the fuck I am, why is it so hard for me to find someone to be with?? Why is it every time I fight for something, it goes further and further away from me?  God, I wish I could stop thinking about both of you and move on. Be truely happy? Why is this so difficult?  Both of you... OUT!!   Mood: So sad Song: More than a Woman by Calvin Richardson   '),(2660,2224636,'22,February,2004','2004-02-22 00:00:00','    Sick and Tired  Everclear    I break every day Stressed out in every kind of way I am sick and tired of bein\' sick and tired All I need and crave Loud life with the power to fade I am livin\' \'cause I keep it all in, keep it all inside Yeah, I keep it all in, keep it all inside Love don\'t work for me I want women who are out of my reach I am sick and tired of livin\' all alone All I need for sure Big love with the power to make more I am livin\' \'cause I keep it all in, keep it all inside Yeah, I keep it all in, keep it all inside I know I\'m close to the fall I know I\'m dangerous I\'m afraid I\'m going down  I blame my family Their damage is livin\' in me I am sick inside and tired of my life All I need, I swear Go out with the power of a nighmare I am livin\' \'cause I keep it all in, keep it all inside I keep it all in, keep it all inside Yeah, I keep it all in, keep it all inside I know I\'m going to fall I know I\'m dangerous I know I\'m not going down alone I keep it all in, keep it all inside I know I\'m close to the fall I know I\'m dangerous I know I\'m not going down alone I\'m not going down alone, I know   '),(2661,2224636,'20,February,2004','2004-02-20 00:00:00','    Somebody to Love   OK so I\'ve been thinkin..... i wanna give my love to somebody...... yeah i\'m fucking lonely.   *burn black smoke fills the sky*  I mean my friends are great and I love them but that\'s not the point. I mean there something inside me that\'s empty. Maybe it\'s all the rejection and heartaches I\'ve had. But why I still want to love soembody I don\'t know.  It\'s like a void I can never fill. Maybe I\'m just supposed to love someone for only a while and eventually move on to the next person. I guess I\'m just supposed to be in a \"forever\" type relationship. This blows.  What am I supposed to be? I don\'t wanna be lonely forever. I don\'t just want friendship with no strings attached (if you get what I mean)  I mean there\'s someone I want but can\'t have. There\'s someone else I have feelings for but the feelings aren\'t returned. They are but not like I want them to be. GOD why is this happening to me.  I just want somebody to love!   mood: crappy and depressed like Song: Sick and Tired by Everclear   '),(2662,2224636,'20,February,2004','2004-02-20 00:00:00','    AHHHHHHHHHHHAHAHAHAHA SLLLLLLLLUUUUUUUUTBUCKET   Yeah so me and skittle got to go to tremont. YAY! It was a kick ass concert. Local bands aren\'t all that bad. There was one band called Swift. They were pretty kick ass. Liked them.   Guys up there are soooooooooo easy to talk to and hit on. But most of them are like 20 or so. UGH!! There need to be hott guys my freakin age gosh darn it.   Skittle was like \"Yeah you were flirtin with like ur entire body. I had to laugh. No it\'s not a bad thing.\"  Yeah I\'m a SLUT BUCKET. But you know what I\'m not fucking them so it really doesn\'t matter. A little flirting never killed anybody right?? I mean it\'s healty. And you never know you might meet some one you\'d loooooooooove to talk to again. But not tonight. Tonight was all about hangin\' with my buds and flirting and havin fuuuuuuuuuuun.  BEEEEEEEOOOOOOOOOOOTCH!! *insano laugh i do*   Mood: So freakin happy nothing could ruin Song: Bring it by ICP (ICPAH!! LOL)   '),(2663,2224636,'19,February,2004','2004-02-19 00:00:00','    OK so....    Yeah man I was all having a good day today but then I think my mom is going to try and fuck it all up. I\'m hopefully going to tremont to night with skittle and a bunch of other people./ GOD IF YOU LOVE ME YOU\'LL CONVINCE MY MOM TO LET ME GO!!!!!!! ppppppppLLLLLEEEEEEAAAASSSSSSE   '),(2664,2224636,'19,February,2004','2004-02-19 00:00:00','    Yeah I love this song and really don\'t know why   Y\'all want a single say f*ck that F*ck that, F*ck that Y\'all want a single say f*ck that F*ck that, F*ck that Y\'all want a single say f*ck that F*ck that, F*ck that Y\'all want a single say f*ck that Boom Boom Boom Boom  What\'s going on today? We gotta break away We got a problem and I think it\'s going to make us go down  They think we\'re all the same And always we\'re to blame For sh*t I think is lame It\'s time to stop the game I think it\'s time to pay for everything you made me say  Y\'all want a single say f*ck that F*ck that, F*ck that [x4]  What\'s going on today? Why must it be this way? We\'re going nowhere and We\'re still knocking the need to bow down  They think we\'re all the same And always we\'re to blame For sh*t I think is lame It\'s time to stop the game I think it\'s time to pay for everything you made me say  Y\'all want a single say f*ck that F*ck that, F*ck that [x4]  We are the ones breaking you down We are the hope to drown out your sounds All Across the world you think we\'re to found  All Across the world you\'re breaking you down  Y\'all want a single say f*ck that F*ck that, F*ck that shit [x4] F**************ck  Y\'all want a single say f*ck that F*ck that, F*ck that [x6]   '),(2665,2224636,'15,February,2004','2004-02-15 00:00:00','    Play me Korn    (feat. Nas, JD)  [Nas:] Everybody\'s an Enemy telling me lies and it\'s killing me why they all want to get rid of me Everybody\'s my enemy Several try to disguise the devil in them Wanting to get into my cerebellum but I\'m Ready and willing to tell them that I can\'t f with them Exhale so hard it got my chest swelling like my dick does when watching nikkid women do sick stuff on my porn collection on television  [Nas:] Trust nobody I don\'t know who to call a friend they all just pretend to be Fuck Everybody this is to all of ya\'ll cuz everybody is an enemy [x2]  [Nas:] Watching my own back strapped in chrome, to my homies where the tombstones at, where the hoes at? Cuz to many didn\'t act like dikes, hermaphrodites with benzene, amen Look at the trash they biting The life in times is kinda weaker like the time and life of the sandman on Apollo theater imagine that another black with a hook who pulls the wack talent off the stage I\'m in raged  [Nas:] Trust nobody I don\'t know who to call a friend they all just pretend to be Fuck Everybody this is to all of ya\'ll cuz everybody is an enemy [x2]  [JD:] You fill your lies around me and you think you won You feel you can control me  with the things you\'ve done  [JD:] You think you can take me you think you can play me you\'re going to start to hate me I feel that you disgraced me [x5]  [Nas:] Trust nobody I don\'t know who to call a friend they all just pretend to be Fuck Everybody this is to all of ya\'ll cuz everybody is an enemy [x2]   '),(2666,2224636,'15,February,2004','2004-02-15 00:00:00','    Yeah I fucked up   So yeah a few days ago me and Hayley were in the car driving around getting lost as usual because we do that for fun. And we some how came up on the subject of our friend Pinz. I\'ve discovered that I like him  a lot . *sigh* Hayley was like \"Yeah you should call him.\" I was thinking to my self \"HAHAHAHA yeah fuck that.\"   But being the person I am I got this thing call \"spontaneous balls\" and decided to call him.  *bangs my head on my desk* So he picked up his phone and this was the conversation we had... Me- Hey buuuuuuuuuudy! Know who this is?? Him- Umm *laugh* no. Who? Me- It\'s Emily Him- OH!! Hey girl. Me- So yeah I got something to tell you. Him- What\'s that? -I keep laughing like a dumbass school girl through all this keep that in mind- Me- Well that conversation we had outside and I said yeah you need more fuck buddies. Well I didn\'t mean that I ment that I would really like to try something out with you. I\'m not saying like a relationship or anything just like go out or something. Him- Can we talk about this later cuz I\'m at a friends house OK? Me- Yeah sure.  So later that night I called him back and tried really hard to discuss what had come up and all I fucking got was \"You caught me at a bad time.\" I got off the phone and CRIED!!! I don\'t really understand why but I just bawled my eyes out. So I call up Hayley and told her what happened and she was like pissed like. Saying it wasn\'t right that he led me on like he did.  I should have seen it coming though.   But yeah I fucked that up. Now I\'m like all wanting to call him just to say hi but can\'t because I\'m afraid that it\'ll be all weird between us.  I really hate the feeling of having some type of feeling for some one and then they never return that feeling or they just keep it to themselves.  I really don\'t know what to do know. I do know that if I ever have feelings for someone again I\'m just keeping it to myself unless they something first. I\'m not going through this bullshit again.  Jerry being the sweetie that he is said he was gonna talk to him but I was like NO NO NO! I feel that if Jerry says something to him that he\'ll feel like he\'s trying to force him to like date me or something  UUUUUUUUUGGGGGGGGGGHHHHHHHHH!!!!!!!! AHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHH DAMN IT!   Song- Play me by KoRn Mood- Lost, hurt, confuzzled, wanting to run someone over with my car   '),(2667,2224636,'15,February,2004','2004-02-15 00:00:00','    Fuck the World ICP     lViolent J]  Fuck. Fuck this shit.  Fuck givin it to me.  [Chorus:]  If I only could I\'d set the world on fire  If I only could I\'d set the world on fire  If I only could I\'d set the world on fire  Sya fuck the world! (Fuck the world!)  If I only could I\'d set the world on fire  Fuck em all! (Fuck em all!)  [Violent J]  Fuck you fuck me fuck us  Fuck Tom fuck Mary fuck Gus  Fuck Darius  Fuck the west coast, and fuck everybody on the east  Eat shit and die, or fuck off at least  Fuck pre-schoolers, fuck rumors  Kings and Queens and gold jewelers  Fuck wine coolers  Fuck chickens, fuck ducks  Everybody in your crew sucks, punk mother fucks  Fuck critics, fuck your review  Even if you like me, fuck you  Fuck your mom, fuck your mom\'s momma  Fuck the Beastie Boys and the Dali Llama  Fuck the rain forest, fuck a Forrest Gump  You probably like it in the rump  Fuck a shoe pump, fuck the real deal and fuck all the fakes  Fuck all fifty two states! Oooo, and fuck you  [Chorus] X 2  [Violent J]  Fuck Oprah, fuck opera, fuck a soap opera  Fuck a pop locker and a cock blocker  Fuck your girlfriend, I probably did her already  Fuck Kyle and his brother Tom Petty, Jump Steady  My homie, fuck him, what are you gonna do?  (Fuck that bitch, fuck you!) Yeah well fuck you too  Don\'t bother tryin to analyze these rhymes  In this song I say fuck ninety three times  Fuck the president, fuck your welfare  Fuck your government and fuck Fred Bear  Fuck Nugent, like anybody gives a fuck  You like to hunt a lot, so fuckin what?!  Fuck disco, Count of Monte Crisco  Fuck Cisco, and Jack and Jerry Brisco  And fuck everyone who went down with the Titanic, in a panic  I\'m like fuck you, AHHHHH!!!!  [Chorus] X 2  [Violent J]  Fuck Celine Dion and fuck Dionne Warwick  You both make me sick, suck my dick  Fuck the Berlin Wall, both sides of it  And fuck Lyle Lovett, whoever the fuck that is  Fuck everybody in the hemisphere  Fuck them across the world, and fuck them right here  You know the guy that operates the Rouge River draw bridge in Delray on  Jefferson? FUCK HIM!  Fuck your idea, fuck your gonnoreha  Fuck your diarrhea, Rocky Maivia  Fuck your wife, your homie did, he\'s fuckin you  Fuck the police and the 5-0 too  Fuck Spin, Rolling Stone, and fuck Vibe  Fuck everybody inside  Whoever\'s on the cover, fuck his mother  Fuck your little brother\'s homie from around the way  And fuck Violent J!   '),(2668,2224636,'14,February,2004','2004-02-14 00:00:00','    My letter   Dear Mom,  What in the flying fuck have I done wrong to you?? Why do you always look at me like you want me to burst in to fucking flames?? I know that I have made a lot of mistakes lately but I really don\'t need you to be my damn enemy. I am SO FUCKING TIRED of you looking at me as if I\'ve been always doing something right. I\'m so damn tired of you acting like I\'m a fucking whore. Sorry to dissapoint you mother but I\'m not out there fucking every damn guy on the streets like you think. I was stupid in the past but I\'ve grown up an awful lot in the past 10 months. More so than you could ever think.  I know that I have had my ups and downs but when its the downs that I\'m going through and need to talk to you about it all I fucking get is \"Oh stop it Emily. Stop spiraling down you don\'t need to do that.\" I don\'t wanna fucking hear that. I need you to be my mother and help me the fuck out. Stop acting like you wish I was never born. Treat me like a fucking human being and not like a motherfucking piece of shit  And I really wish you\'d stop acting like Crissy and stop treating me and dad like she treat her husbend and daughter. That\'s not your life sorry. Excuse me for not wanting you to pick on me and put me down every chance you get. Will you ever be proud of me again? Or do you just hate me that much. I honestly don\'t believe you will ever love me again. If you ever have.  I\'m really tired of trying to please you in every way. So I\'m done. I am fucking done. I\'m not going to do it anymore.  Don\'t think that I have stopped loving you. You\'re my mother. You always will be. But I\'ve lost all respect for you. Even though you think I don\'t deserve it, you should give me some respect. So to get respect give respect. I haven\'t done anything that wrong. Let go of all your grudges against me please. I\'m tired of all the arguing. But unless you stop I\'m leaving. I\'m gone. I\'ve got places to go and I\'m ready to just pack my shit and go.   Song: Fuck the World by ICP Mood: SO fucking angry   '),(2669,2224636,'14,February,2004','2004-02-14 00:00:00','    EMILY   VERSE 1 It wasn\'t supposed to be like this Another dose of unhappiness I gave it all and managed to get shot down yet again So I got drunk Had sex with all your friends You told me to never call your house again  CHORUS Emily, you saved the day Emily, when you threw me away  She was always such a pretty girl Nobody like her in the world A little piece of heavenly That no one else could stand I see her in my dreams at night I see you when I close my eyes I just can\'t seem to shake you, Emily  VERSE 2 You got your money and I got cast Outside thrown out on my ass In the city with no one else, no where else to go So I hooked up with this model from Singapore Emily, I sure am glad you didn\'t want me anymore  CHORUS  She was always such a pretty girl Nobody like her in the world A little piece of heavenly That no one else could stand I see her in my dreams at night I see you when I close my eyes I just can\'t seem to shake you, Emily Yeah  Emily, you saved the day Emily, you saved the day  Emily, you saved my ass   '),(2670,2224636,'11,February,2004','2004-02-11 00:00:00','    Song   OH THERE IS A SONG WITH MY NAME IN IT!!!! WHO\'S UR DADDY BITCH?? LOL NOW EVERYONE CAN READ AND THEY WILL *GRRRRRRRRRRR* =)   '),(2671,2224636,'11,February,2004','2004-02-11 00:00:00','    It\'s Been a While Stained   And it\'s been awhile Since I could hold my head up high And it\'s been awhile Since I first saw you And it\'s been awhile Since I could stand on my own two feet again And it\'s been awhile Since I could call you  And everything I can\'t remember As fucked up as it all may seem The consequences that I\'ve rendered I\'ve stretched myself beyond my means  And it\'s been awhile Since I can say that I wasn\'t addicted And it\'s been awhile Since I can say I love myself as well And it\'s been awhile Since I\'ve gone and fucked things up just like I always do And it\'s been awhile But all that shit seems to disappear when I\'m with you  And everything I can\'t remember As fucked up as it all may seem The consequences that I\'ve rendered I\'ve gone and fucked things up again  Why must I feel this way? Just make this go away Just one more peaceful day!  And it\'s been awhile Since I could look at myself straight And it\'s been awhile Since I said I\'m sorry And it\'s been awhile Since I\'ve seen the way the candle lights your face And it\'s been awhile But I can still remember just the way you taste  And everything I can\'t remember As fucked up as it all may seem to be I know it\'s me I cannot blame this on my father He did the best he could for me  And it\'s been awhile Since I could hold my head up high And it\'s been awhile Since I said I\'m sorry   '),(2672,2224636,'11,February,2004','2004-02-11 00:00:00','    AHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHH   You know I am just so sick and tired of being confused and lost and hurt. I\'m so tired of being in love with someone who doesn\'t want to be with me anymore. I\'m so ready to let go of all the pain that I have stored up inside but I just can\'t. For some reason something is keeping me in this state of depression I could say for lack of a better term.   I\'m ready to just go out and be truly happy. I wanna smile and not have it be fake. I want to just go out and pick up guys with out feelin like I\'m betraying someone.   I damn sick of being tired because all these feelings are wearing me the fuck out. I would love to find another person that makes my heart skip a beat. But not right now.   I wanna be 16 again. I know I\'ve made right choices in the things I\'ve gone through. I\'m tired of feeling guilty about them. I\'m tired of the resentment that I have towards someone that I love so much.  I wish my mother and I could have a better relationship. I wish that we wouldn\'t end up in a fight every time we just sit down to shoot the shit. AND I also wish my brother would stop looking at me like I\'ve killed him inside. I\'ve made mistakes I know that. But I\'ve stopped dwelling on them. I wish they would too.  AAAHHHHHHHHHHHHHH! I JUST WANNA BE HAPPY DAMNIT! Why can\'t God just give me peace and happiness. What the hell have I done to Him that is so fucking wrong?? I do the best that I can. This world is all the sudden become  so  much to handle.   Stress shouldn\'t be l-i-f-e!! This blows. I wish things could be easier but unfortunantly they can\'t.  I guess I gotta take the punches as they come.   Song to tell the way i feel: It\'s Been a While by Stained Mood: So angry   '),(2673,2224636,'11,February,2004','2004-02-11 00:00:00','    Blue Eyes Crying in the Rain by Willie Nelson  In the twilight glow I see her Blue eyes crying in the rain  When we kissed goodbye and parted  I knew we\'d never meet again   Love is like a dying ember  And only memories remain  And through the ages I\'ll remember  Blue eyes crying in the rain.  Someday when we meet up yonder  We\'ll stroll hand in hand again  In the land that knows no parting For blue eyes crying in the rain.   '),(2674,2224636,'11,February,2004','2004-02-11 00:00:00','    (I Hate) Everything About You Three Days Grace  Every time we lie awake After every hit we take Every feeling that I get But I havenâ€™t missed you yet  Every roommate kept awake By every sigh and scream we make All the feelings that I get But I still donâ€™t miss you yet  Only when I stop to think about it  I hate everything about you Why do I love you I hate everything about you Why do I love you  Every time we lie awake After every hit we take Every feeling that I get But I havenâ€™t missed you yet  Only when I stop to think about it  I hate everything about you Why do I love you I hate everything about you Why do I love you  Only when I stop to think About you, I know Only when you stop to think About me, do you know  I hate everything about you Why do I love you You hate everything about me Why do you love me  I hate You hate I hate You love me  I hate everything about you Why do I love you   '),(2675,2224636,'11,February,2004','2004-02-11 00:00:00','    Why call me Beautiful?   *heavy sigh*  Yesterday I went over to Jerry\'s house to hang out with him. Well we\'re getting dinner ready and everything like and a friend of ours, Steve (aka Pinz), calls to tell him his sister is on the way home.  So Jerry and I sit down to eat and the door opens. And strollin in comes Meghan (sister) and Pinz and Blade (another friend). I saw Pinz and my heart little jumped. It was a weird feeling. I don\'t have that feeling with many people unless I have a very strong connection and there have only been a few and those people know who they are. If you have to ask you\'re not one of them.   But I think I felt what I did because of the past history that we have. May is a significant thing to my life (Skittle you know what I\'m talking about) Back then things between me and Pinz could have progressed but certain events stopped that.   Anyways again, Jerry and I finish eating and so all of us go in to his room and start watching this stupid shit called Girls Gone Wild. The most retarded and pointless crap I have ever seen in my life. So I get bored with stupid bitches running around showing their shit, so i get up to go smoke outside.   Pinz comes and joins me. So I\'m standing there looking around at everything and I happen to glance over at him. He\'s giving me the strangest look. So I asked him \"what\'s wrong?\" and he looks at me normal like and says \"When we were in Jerry\'s room I was just looking at you thinking to myself how beautiful you are.\" In my head I\'m like \"OH MY GOD\" but to him I said \"Nah I ain\'t that pretty now.\" I look over at him and he looks kinda hurt so I correct myself and say to him \"Thanks. that\'s sweet.\"   Then he busts out with \"Plus I was just thinking about how much I just want to kiss you.\" so I say to him \"Then why haven\'t you?\" He looks away from me and says \"Because I don\'t know how people will react if I do.\" So I say \"It doesn\'t matter what other people think. You should do what you feel like doing.\"  We\'re standing there and it\'s deadly silent and uncomfortable. So I said, \"You know I always wondered why you always look so sad?\" and he responds with \" I just don\'t know how to be happy.\" Then conversation progresses back to talking about may. I said \"You know ever since your \'drunken confession\' incident I\'ve thought a lot about you.\" I don\'t remember his response but I do remember that then i kissed him. And he looked all happy for a split second =).   But see now I\'m really fucking confused. I don\'t know what is going on with that boy. I mean if there\'s something that\'s there between us it really needs to be known. But nothing can really be done about that right now because unfortunantly I still love Jerry very much and still want to be with him.   I guess I just have to quit dwelling and take the punches as they come.   Song to tell the way I feel: (I Hate) Everything About You by Three Days Grace Mood: Needing questions to be answered   '),(2676,2224636,'11,February,2004','2004-02-11 00:00:00','    I want to bang my head on the wall and SHOOT MYSELF IN THE FOOT!!!   Yeah so here I am listening to Bush. A good song yet I don\'t know the name.  But anyways! GOD i want to bang my head on the wall!! So I got home from going on an adventure with my homie Hayley (SHOUT OUTS TO MY HOME GIRL SKITTLES) She\'s so damn smart on guys. Well she\'s really cool about explaining on how and why Jerry acts the way he does towards me. Sometime I honestly don\'t know what I\'d do without that girl. She\'s been there for me through some really tough times. I mean she\'s stuck by me as I lost my sanity when I had to give Joshua (my son) up. Plus she\'s been there to let me bitch to her about all the things in the world that piss me off so much. Plus she sits there and listens to me as I whine about how I hate Jerry but love him and how I get so confused. I swear she\'s been there so much for me lately that I would lay down and die for her just to show my thanks to her for being such a great friend. I hope that we\'ll be homies for life. I mean God, we\'ll be like 60 or older sittin\' on each others porch cacklin\' at the kids driving by saying \"Remember when we used to do all that shit??\" LOL   But anyways like I was saying: I really want to bang my head in to the wall because I was on the phone with Jerry today talking and stuff. And as we got off the phone I said plain as fucking day \"I love you.\" and he says \"What did you say?\" I go \"NOTHNG NOTHING. I\'ll see you when I get there for dinner.\" and he goes \"OK whatever.\" I felt like such a fucking fool. Imean my God, here I am trying to respect our \'friendship\' and there i go screaming out \"I love you\" because even though I still do I don\'t let him know that to the full extent because I don\'t want to run him away. I\'ve made to many mistakes in the relationship we once had and I don\'t want to do anything to fuck up our friendship. Cuz honestly I truely value the friendship we have but I don\'t want to say anything that\'s gonna mess it up.   Song to tell the way I feel: Blue Eyes Crying in the the Rain by Willie Nelson Mood: Shitty but Good   '),(2677,2224636,'10,February,2004','2004-02-10 00:00:00','    Numb by Linkin Park   I\'m tired of being what you want me to be Feeling so faithless, lost under the surface I don\'t know what you\'re expecting of me Put under the pressure, of walking in your shoes (Caught in the undertow, just caught in the undertow) Every step that I take is another mistake to you (Caught in the undertow, just caught in the undertow)  [Chorus:] I\'ve become so numb I can\'t feel you there I\'ve become so tired, so much more aware I\'m becoming this, all I want to do Is be more like me and be less like you   Can\'t you see that you\'re smothering me? Holding too tightly, afraid to lose control \'Cause everything that you thought I would be Has fallen apart, right in front of you (Caught in the undertow, just caught in the undertow) Every step that I take is another mistake to you (Caught in the undertow, just caught in the undertow) An\' every second I waste is more than I can take  [Chorus:] I\'ve become so numb I can\'t feel you there I\'ve become so tired, so much more aware I\'m becoming this, all I want to do Is be more like me and be less like you  And I know I may end up failing too But I know You were just like me With someone disappointed in you  [Chorus:] I\'ve become so numb I can\'t feel you there I\'ve become so tired, so much more aware I\'m becoming this, all I want to do Is be more like me and be less like you  I\'ve Become so numb I can\'t feel you there I\'m tired of being what you want me to be I\'ve become so numb I can\'t feel you there I\'m tired of being what you want me to be   '),(2678,2224636,'10,February,2004','2004-02-10 00:00:00','    I wonder   You know I wonder if I ever went and visited NWSA would anybody notice that I showed up? Or rather would anybody care. You know I left there with a sense that I would stay in touch with people, which in all honest I have tried to, but those people would stay in touch with me. I hate it that I lost contact with a really great group of friends. And they\'re reading this they know who they are. I\'d really like to get back in touch with them again but it\'s like every time I try I feel like the response it \"Oh her...\" But hey I can\'t blame them for not getting touch with me they have lives. It just be nice to know how everybody\'s doing.  Sometimes I wonder if I\'m ever actually alive? Lately I\'ve felt so far away from this world. Like I almost don\'t live. So much has happened to me I guess that I\'ve just shut down. I really don\'t feel any more. Well that\'s not true I just don\'t feel much any more. I rarely feel any true emotions. But I guess that can happen to a person when the two people you love the most are gone or don\'t want to be with you. How sad how things like that can effect me so much. But hey first love and first child what the hell can I say? I think if there was a song to tell exactly how I feel it would be Numb by Linkin Park. Funny. My life right now blows so much ass I don\'t know if I\'ll ever be truely happy again or if I\'ll be in love again. Sad how events can do that to a person. Oh well.   Song to tell the way I feel: Numb by Linkin Park Mood: Hurt and lost   '),(2679,2224636,'28,March,2004','2004-03-28 00:00:00','    You! HEY YOU! Yes You! Come \'ere!   Why is it that everytime we\'re together, I think everything is back to how I believe it should be? Why is it every time that you look at me its like you have something to say but the letters instead sit there and shake their asses at me and won\'t let you form the words? Why is it that I want to shake you and scream in your face until I\'m blue? Why in the fuck do I want you?   I\'m so sick and tired of wanting to be with you! It\'s like this uncontrolable urge to just need you. You, damn you, complete me in some strange way. You fill this empty spot in my soul. I wish I could kick my own ass for wanting you so bad.  Every time I wish to express to you how you make me feel I have to bite my lip until it fucking bleeds. I have to hold back on the things I know you don\'t want (or so I assume) to hear.  And yet, I\'m still lost on you. You still hold me and kiss me and make me feel all good about myself. Yet you can\'t be with me.   I\'m beginning to believe it\'s because you can\'t commit fully. You only said the relationship wasn\'t going any where because we couldn\'t move forward. WELL DUHHHHHHHHH DUMBASS WE CAN\'T GET MARRIED!!!! Is that what you want? Is that what you need? You just want to wait until I\'m ALL GROWN UP so we can commit like that? Well damn I\'m ready. My foots about to step down. IS YOURS??? NO NO ITS NOT!! I\'m so damn sick and tired of your excuses. BE REAL WITH ME! BE FUCKING STRAIGHT! Start remembering that you can tell me everything like you once did.   Being friends doesn\'t change a thing. You\'re the one who said that. YOU\'RE THE FUCKING ONE WHO SAID  You\'ll always have my heart and I\'ll always have yours.  Well damn it\'s true. And that sucks because I want you. Not physically (although that\'s a nice perk). I want you mentally, emotionally, spiritually! I\'ve had plenty of opportunities to connect with other people and I know you know that. Why? Because I tell you everything. You know more about me than anyone.   You want to know why I can\'t connect with other guys? Because it\'s just not there. That thing that I feel, it\'s not there with them. But with you, God, with you it\'s always there. Screaming and crying to get out. Wanting so bad to just run away but can\'t because you hold me in a way that no one ever has. You could let go at any time, any time you fucking damn well please but you just hang on. Hang on to those strings that are attached to my heart, my soul, my mind, my body!!! And it is fucking killing me.   I wonder is that why you haven\'t gone and found any body else like you did the last time? Is that the reason you don\'t date? Is that the reason you won\'t fuck anyone else?? Because you hold every string that surrounds my every being?  Why is it that every time something happens to me you forever say you\'re sorry? What\'s the reason behind that? Or is there? You kill me with your kindness. You kill me with your touch. It pains me inside.   Yet you make me laugh. You make me feel. You make me complete! There\'s not a damn doubt in my mind that we could make this work. But I know that we need time. Time to be friends. Time to just chill. Just know that it is so hard for me.      Mood: Just utterly lost Song: Any love song by Jewel   '),(2680,2224636,'27,March,2004','2004-03-27 00:00:00','    To see you once more   Sitting there Lettin things flow Lettin it all go You walk in the door Havin everything to score Smile my way You say Hi Sit down beside me My heart aches for your touch Is this love or is this lust What would it be like to fuck Is this wrong Why do I feel this You act mostly like I don\'t exist We\'ve expressed things Probably things that shouldn\'t have been said My heart feels like a ton of lead Look at me with those eyes Tell me all those lies They\'re like a bunch of vines Too many times We\'ve tried to start But we won\'t take the first step I\'m ready But are you Or is it that I\'m just bein a fool Feelin like a tool I\'m just a piece in your little game Your pawn To move as you please Such a damn tease Kiss so sweet Touch such a feat To see you once more My heart stops And yet I know theres nothin there You couldn\'t possibly care I know that you wouldn\'t dare     Mood: Just here Song: .45 Shinedown   '),(2681,2224636,'25,March,2004','2004-03-25 00:00:00','    YOU! LETTERS! STOP SHAKIN UR ASS AT ME!!   I\'m just a little frustrated right now. Not pissed off but not happy. Ever had one of those days. Here let me explain.  Ever since Jerry and I broke up there have been a lot of things on my mind. Just questions I need clarified. I want to move on from him. So bad. Because it hurts me some to still be in love with him. Yes, I openly admit to everyone that I\'ve told I\'m done with him I don\'t need him, I lied. I\'m pretty sure Skittle saw right through my fuckin lie. DAMN YOU J/K LOL  But its like... damn boy you\'ve got me confused outta my mind. I mean we\'re homies. Always and forever will be. But I mean you were or are the one who wants to be  JUST  friends. You\'re the one who asked for it. So to make you happy I gave you that. I respect that.    Yet when ever I go over there to chill or we go and hang out some where. No one could guess we ever broke up and that we\'re just friends. And for once it\'s not me initiating all the kissing and hugging and holding hands and everything that goes along with relationships.   I mean we go out in to public and he walks like there\'s only one inch of space between us. Our hands might accidentally brush and he grabs mine and holds it. He sees a guy checking me out, arm instantly goes around my waist. More intimate than the shoulders I believe. We sit there and joke around and he kisses me and hugs me. How the fuck am I not to be confused.  Now tell me this y\'all. When you go to a friends house do you end up holding their hand? Or if y\'all are like lying down on the couch does the person put his or her head on your chest and wrap their arms around you? No I don\'t think so.  Because you\'re just friends.   Jerry doesn\'t even act like we\'re friends. And get this when I try to act like  just  friends he gets offended. When he calls me \"baby\" or \"sexy\" (names he\'s called me for like ever and a day) and I jokingly tell him not to he gets kind of upset.  Then when he heard that I was going to the prom this year and NOT by myself he seemed to be hurt and offended. When he knows that I want him to take me to my Senior prom. The year that means the most. I want to share that with him because he\'s special to me. Not this year. This year is like the \"get ready for the senior prom\" prom. Ya know?  I just need to know why he\'s acting like this. My girl Skittle has told me her opinion on this situation time and time again because she\'s like a female Jerry (take no offense to that homie). But it\'s like its not clicking.  And when he and I are alone he just keeps looking at me like he desperatly needs to say something but can\'t. And no its not \"I love you\" he tells me that anyways. But its like he opens his mouth to tell me something and the letters ARE RIGHT THERE ON HIS LIPS JUST SHAKING THEIR ASSES AT ME!! Like they\'re saying \"YOU\'RE NEVER GONNA KNOW! YOU\'RE NEVER GONNA KNOW!!\" Oh it pisses me off.  Damn boy, if there\'s something you need to tell me JUST FUCKING SPIT IT OUT!!!!!!!   Mood: Don\'t really have one Song:Going under (i think) by Evanescence   '),(2682,2224636,'23,March,2004','2004-03-23 00:00:00','    WOW!! ME? REALLY?   Yesterday just had to be about the best day in the fuckin world. I was at lunch with my friend Brandon turns to me and says \"Would you like to go to the prom with me?\" I was like \"Are you being serious?\" cuz sometimes I can\'t tell with him. He nods and says \"YES!\" I was like \"Well damn ok!!!\" and then goes on to say we\'re wearing pink I WAS LIKE NO NO SIR THATS A NEGATIVE!! Then he suggests orange so I agreed to that. I already found the prom dress I want and everything. OK WHY THE FUCK AM I SO HAPPY. What makes it even better is that my homie skittle will be there too!!!! WOOP WOOP! WHO\'S YA DADDY BITCH!!!!!!???? THA CABLE GUY   Mood: flip floppin happy Song: Knock down by Chumbawamba   '),(2683,2224636,'21,March,2004','2004-03-21 00:00:00','    you are deeppink  #FF1493         Your dominant hues are red and magenta. You love doing your own thing and going on your own adventures, but there are close friends you know you just can\'t leave behind. You can influence others on days when you\'re patient, but most times you just want to go out, have fun, and do your own thing.  Your saturation level is high - you get into life and have a strong personality. Everyone you meet will either love you or hate you - either way, your goal is to get them to change the world with you. You are very hard working and don\'t have much patience for people without your initiative.  Your outlook on life is very bright. You are sunny and optimistic about life and others find it very encouraging, but remember to tone it down if you sense irritation.         urlLink the spacefem.com html color quiz   '),(2684,2224636,'19,March,2004','2004-03-19 00:00:00','    OH SHIT IT\'S YOUR BIRTHDAY.....WOOP WOOP   Damn so far today has been FAN-FUCKIN-TABULOUS!! I mean even tho I\'m only turning 17 today its all skippy.....  Go to pictures.... then go to guestbook AND SIGN DAMN!!!!!!!1 LMFAO   Mood: Bouncy Song: Birthday Bitches- ICP   '),(2685,2224636,'19,March,2004','2004-03-19 00:00:00','    STOP! JUST STOP!   OK listen up people. I am damn sick and tired of people bitchin about shit they want but don\'t have the balls to get. Look, this is what I have to say.  1.) To all those who love somebody or want to be with someone: walk up to the fucker and tell them how you feel tell them you\'d like to give you two a shot. Hell it could be an ever lasting relationship. But you might get rejected and feel like a fool. So what! At least you\'ve accomplished something  2.) To all those in a petty argument with someone you care about: STOP! JUST STOP! That petty shit you\'re arguing about is not worth it. Trust me I know. It could end something great.  3.) To all those with hate in your hearts: It\'s time to let go. Just let all the hate go. It\'s more than likely making you an angry person. Sad on the inside. You\'re hurting others as well. Stop. There\'s no reason to feel that way. Stop being negative start being positive  4.) To all those who are scared: Don\'t be. As some one once said \"the only thing we have to fear is fear itself\" words to live by. If there\'s something that scares you to death. Face it. Own it. Take charge and over come what you fear the most. Whether it be telling someone how you feel, standing up to a bully, becoming a new person. Its time to face our fears  5.) To all those \"pimps and playas\": Quit. It\'s time to straighten up. Be a one man or one woman person. This playin around on one another needs to stop. It spreading hatred, disease, and hurt. It\'s time to love each other.  6.) To all you racists: Even though you people disgust me, it\'s time for a change. We all need to get along. Love one another. You guys are scared. It\'s the only reason you act this way. Try meeting the people you hate you just might like them.  7.) To everyone in the world that reads this: It\'s time for a change in this world. TIme for the hatred to stop. Time for the pain and suffering in the world to cease to exist. We need to strive to saving this place or we\'re all going to end up in a hell that unimaginable to us all. It\'s time for love. It\'s time for peace. It\'s time to be at peace with one another. And it\'s going to start with me.   Mood: Ready for a change Song: What\'s goin on? by Marvin Gay   '),(2686,2224636,'19,March,2004','2004-03-19 00:00:00','    ::BAM:: BITCH!!!!   So yeah. After I got all the shit I had to done I tried to lay down to take a nap but instead got kidnapped by smurf because he wanted to take me out for my birthday TWO DAYS EARLY LOL!!  So anyways..... this bitch at the place we were eating was an old \"friend\" if you can call her that was there. She knew that I was pregnant and asked if we had any pictures with us. She saw it and said \"awwww he\'s so cute\" then she left.  I walked out to the car because I left my cigarretts there and overheard her on her cell phone saying to someone \"OMG the baby is like so ugly.\" I was so outraged that I walked up to her and just ::BAM:: my fist flew into her face. I just started screaming out all the profanities in the world at her. It pissed me off so damn bad. I mean yes i know that she\'s a two faced bitch but NO MATTER WHO YOU ARE YOU NEVER AND I MEAN    NEVER  TALK ABOUT MY SON IN ANY NEGATIVE WAY!   I walked back in the resturant and sat down. Smurf looks at my face and asked what did you do? I shook my head and told him not to worry about it.  But god I was so damn mad   Mood: SO PISSED but yet happy Song: Fuel ~ Metallica   '),(2687,2224636,'17,March,2004','2004-03-17 00:00:00','    SIMPLE MAN Lynard Skynard    Mama told me when I was young  \"Come sit beside me, my only son  And listen closely to what I say  And if you do this it will help you some sunny day\"  \"Oh, take your time, don\'t live too fast  Troubles will come and they will pass  Go find a woman and you\'ll find love  And don\'t forget, son, there is someone up above\"  chorus: \"And be a simple kind of man  Oh, be something you love and understand  Baby be a simple kind of man  Oh, won\'t you do this for me, son, if you can\"  \"Forget your lust for the rich man\'s gold  All that you need is in your soul  And you can do this, oh baby, if you try  All that I want for you, my son, is to be satisfied\"  chorus  Oh yes, I will \"Boy, don\'t you worry, you\'ll find yourself  Follow your heart and nothing else  And you can do this, oh baby, if you try  All that I want for you, my son, is to be satisfied\"  chorus  Baby, be a simple, really simple man  Oh, be something you love and understand   '),(2688,2224636,'15,March,2004','2004-03-15 00:00:00','    Simply Beautiful   I had just finished my Junior Exit. I absolutly hated writing that paper! I mean I had no input in it whatsoever. Our teacher wanted us to write the paper as if it was a Senior exit. So basically that meant lots and lots of note cards of what other people thought. It was gay. Just as I had finished this song came on. Simple Man by Lynard Skynard. My brother had followed me in to the room to read my paper but we just sat there instead. I was stunned. The song was simply beautiful. I was about in tears. Ya know everybody has a song that just makes them feel such compassion for what the song is talkin about. I was almost in tears. It was just that good. After the song was done me and my brother let out our breath at the exact same time. I think I didn\'t breath through almost the entire song. Damn.   Mood: Amazed Song: Simple Man by Lynard Skynard   '),(2689,2224636,'15,March,2004','2004-03-15 00:00:00','    UGH! BUT WHY NOT???   Today was OK but awful at the same time. My home boy Smurf decided to pick me up after school today. Nice surprise. We went to go and chill at his house. We started to wrestle and shit and he was playing and put his hands around my neck. I fuckin flipped out. I started screamin and crying for him not to hurt me and called him by and ex\'s name. The ex was extremely abusive. I took all the strength I had and pushed him away from me and locked myself in the bathroom still screamin the ex\'s name tellin him to stay the fuck away from me.   After about five minutes that I had locked myself in the bathroom room. Smurf unlocked the door and came in. I hid my face and tried my best to hide in the corner. I was so terrified. Smurf bent down and wrapped his arms around me telling me that he would never hurt me like my ex did. He told me that he could never hurt me in that way. Ever. He asked me why I couldn\'t look at him and I started to cry harder and told him that I was afraid to for the chance I might see my ex\'s face. He then made me look at him and said \"Khaos, you know that I love you so much and would never hurt you. You know that. I could never bring myself to hitting you. Please, please, I\'m begging you to stop crying. It tears me apart when you do.\"   He wiped away my tears and held me close to him. All the while I was thinking to myself \"God. But why not? Why couldn\'t we just be together again?\" I\'m trying so damn hard to let it go. To let him go but no matter what I do he invades my thoughts.  All I want is for him to be happy. I want to be happy as well. But I can\'t see my future with out him sometimes. Then other times I\'m just like UGH I HATE YOU SO DAMN MUCH I\'M GLAD WE\'RE OVER!!   N-E Ways....  Smurf and I went to his dads house and told him about Joshua (our son). He didn\'t yell. Didn\'t cry. He just looked at the pictures and all I could see was disappointment. In what I really don\'t know. But it felt awful. I hated it. I guess it was because Joshua wouldn\'t be around for everyone to see grow up. But I don\'t feel bad about it at all. I know that my son is in a loving home with beautiful happy people. I know that he\'s going to get everything that I couldn\'t or can\'t give him.  I did what I did out of extreme love for my son. I wanted him to be happy, healthy, and loved without the suffering that I would\'ve had to go through if I chose to keep him.  Smurf\'s mom is making me feel alittle guilty that she wasn\'t there. But I didn\'t want her there. Sometimes I wish Smurf had never told her. Because she told me that her sister couldn\'t have children and would have adopted him herself. But not to sound extrememly terrible I didn\'t want to have Joshua where I could get him. It would still be too damn painful. Her side of the family is saddened that they couldn\'t get to hold him and they\'re trying to get me to set up a date for them to. I told his mother no that I couldn\'t and wouldn\'t do that. It\'s not that type of adoption.   I don\'t think anyone understands that  I  want to hold my son again. That   I  want to see him again. And I don\'t know if that will happen again until he\'s 18. Or when he\'s ready. I will not force anything upon his parents to make them or myself happy. His mother was a little saddened by my desicion but I\'m sorry she\'ll just have to get over it.  I wish I could leave this town. I just want to get away from it all for awhile. I need to. I\'ve told my dad this time and time again. But he tells me not to because then he\'ll think he\'s done something to make me do that. But it\'s not him. It\'s not anyone. It\'s just me. I just need to get away from my thoughts and the pain that I have. The pain where I can\'t have my son with me and the pain that I can\'t be with the one I love the most with all my heart and the pain that mother can\'t seem to stand to have me around. I\'m ready for it all to be over. I\'m done. It\'s like I can\'t take much more of this. If I stay here another week I\'m going to break down mentally and just go insane. I mean the real insane not the funny kind.   Mood: Terrible Song: My Immortal   '),(2690,2224636,'12,March,2004','2004-03-12 00:00:00','    I CANNOT TAKE THIS   My mom has driven me to the brink off my sanity. I\'m about to go off the fucking edge. I have to get out of this town. I have to get out of this house. Away from her. I hate to leave my dad and brother. But I cannot take this. I can\'t take the bitching, the naggin, the whining, from her. I cannot stand her saying that she\'s \"so tired of my bullshit.\" When I haven\'t done a damn thing to her. God how I wish she\'d disapper out of my life for good. Not die just disapper. I can\'t take her anymore. I can\'t stand her face. I can\'t stand her presence. I hate it when she tries and makes me feel guilty about the mistakes I have made. I want so bad to hit her in the face, pack her shit, and tell her to fuckin leave. Me and my dad would be happier I know. I\'m so sick of her telling my dad he can\'t do anything right. He doesn\'t make her happy. Then when she wants something she acts all sweet like and shit. I CANNOT TAKE THIS.....I\'M GONE I\'M DONE I\'VE FUCKING LOST IT   '),(2691,2224636,'11,March,2004','2004-03-11 00:00:00','    AHHHHH MY FRIDAY SUCKS BALLSACK   Well I got to school and most of that went skippy. Had to do some athletic trainin with my coach for the track team. That was borin as fuck until somebody fell on their head.   I walked up to Dollar General and I was lookin around and heard \"Howdy Howdy.\" I laughed because I knew it was Jerry. While I was checkin out at the counter he kept lookin over the isle where you can see and the top and smilin at me and shit. It was great. Guess that stupid bitch he was chillin with yesterday wasn\'t good enough. LMFAO But I walked up to him and asked him how he was and that I had pictures of Josh (New ones) that he just had to see. Then I gave him a note. Starting off with \"WHAT! YOU CAN\'T CALL NOBODY???!!!\" LOL  Told him we needed to get together and chill and shit and he said yea sometime next week.  OH WELL!   Mood: TIRED AS FUCK Song: Wastin my Time ~ BoomKat   '),(2692,2224636,'11,March,2004','2004-03-11 00:00:00','    Damn... you\'re too good for me      To you       You\'re always there when I need you the most  Comforting me with sweet words To have you hold me in your arms would be a blessing To have you kiss me would be a dream You help me forget my toubles Ands and chase away my bad dreams You always call at the perfect time  You\'re voice calms me in the strangest ways I can\'t put my finger on it but there\'s just something about you that i\'m fallin in love with I promised I wouldn\'t but I just am Can\'t help it So this is to you A dream come true    Mood: *~*sigh*~* so happy Song: Fallin Alieca Keys   '),(2693,2224636,'10,March,2004','2004-03-10 00:00:00','    FOOOOOOOOD! CHINESE FOOD!!!! ::tackles mom and steals food::   Ahhh chinese food is good. My mom made the mistake of trying to keep it from me. MINE!!! *GRRRRRRRRRR* lmao...  So my day has been relativly good. Although as Skittle and I were heading up to UNCC to go to the library I was looking around for the bottom half of my seat belt and I was freakin out because I couldn\'t find it and then I look down at myself and start laughin because I had it on the entire time. Then I told Skittle I couldn\'t remember the drive to my house. And we had a little freak out moment because we both couldn\'t remember.  She thought I had like got her stoned or something and I forgot to tell her. LOL. I wouldn\'t do that to a person. Damn I\'m not that mean.  Even though everybody thinks I am.  OH WELL LOL    Mood: ::shrugs:: Eh! Happy Song: Highway by Audioslave   '),(2694,2224636,'09,March,2004','2004-03-09 00:00:00','    Smokin is bad   Yeah so I\'m in the car with Skittle and we\'re going to go scare off this guy that we\'re not to sure we want to hang with much anymore so yeah this is how it went down  Skittle: OK so yeah we\'re gonna go meet mike today Me: NOOOOOOOOOO :spazes in the bus seat:WHY Skittle: To scare him off of course (me evil look) Me: OOOOOOOOO fun Skittle: Yeah we\'re gonna meet him at Dollar General Me: OMG AND I THINK JERRY\'S WORKIN TODAY OMG LOL Skittle: That\'ll be great Me: We can get Jerry in on it too LOL Skittle: LOL Me: NO! NO! You know what I could do? Skittle: What? Me: I could act absolutly fuckin nuts and act like I have an imaginary friend that you see too. I\'ll hold her hand and everything and introduce Mike to her. (me and skittle laugh for like 5 mintues) Me: NO NO you know what else when we go to the park I could like drive off and you can tell Mike I\'m gonna go score somethng. Then I\'ll come back screeching tires into the park screamin at you THAT IT ALL WENT BAD. (we\'re walkin down the road and fall to the ground laughin our asses off)  Then we get our friend John involved in it and it turned out to be absolutly hysterical. Skittle couldn\'t help but laughin when we got in the car and I pulled my hoodie over my face.   Quote: Imitating me: I\'ve never been a quitter (puts a cig in her mouth) I ain\'t gonna start now  Mood: happy Song: I believe in a thing called Love the Darkness   '),(2695,2224636,'08,March,2004','2004-03-08 00:00:00','    OK I\'M SORRY YOU SAID YOU\'RE GONNA DO WHAT??    OK so I\'ve been told that I\'m going to be kidnapped by Chris and kept for like a week just to go to a big ol\' concert. I DON\'T WANNA BE KIDNAPPED!!! SSSSSSSCCCCCCCCCCCC----AAAAAAAAAA------RRRRRRRR---------YYYYY!!!!! I don\'t know it might be fun. No NO NO NO it won\'t. Kidnapping is bad.  And why would any one wanna kidnapp me of allllll people?? I mean c\'mon now! I will absolutly freak though if he does try and be mad and hyper all at the same time. Can one person possibly do that? I dunno Skittles says I have ADD so maybe I could lol ::runs and hides:: YOU\'LL NEVER FIND ME BAWHAHAHAHAHAHA   Mood: ::screams like a cheerleader about to get raped Song: Haunted Forest- ICP   '),(2696,2224636,'05,March,2004','2004-03-05 00:00:00','    Brain waves   OK so Skittle is like the most hilarious person I know. She has just sat there online and sent brain waves out to Chris to tell him to get online or call me. How funny is that shit??  He called me like three fuckin times yesterday I was like HA YEAAAAAAAyuuuuuuuuuuuuuuuuhhhh LOL. He was mean like the first and second time he called but then the third he was nice. I think either that or I was just really out of it. OOH SO PRETTY SOUNDING ON THE PHONE ::melts into a runny puddle:: LMFAO.....   Mood: FLYIN HHHHIIIIIIGGGGGGH BABY Song: there is none   '),(2697,2224636,'05,March,2004','2004-03-05 00:00:00','    OMG YOU SOUND SO PRETTY   So yeah this guy named Chris.... met him online like a while ago..... we were all mean to each other for like the longest..... he called me a hood rat for a while i really don\'t know why..... i wasn\'t that mean to him I don\'t think lol..... but then he goes and tells me that hes got feelings for me.... i was all like \"OH! SO SWEET!!\" I like him to...... I don\'t know there\'s something about him that I just can\'t put my finger on but theres something about him I like/love a loooooooootttttttt.   He asked me for my number so I was like w/e and gave it to him. He called me tonight and SCARED THE SHIT OUTTA ME!! BUT OMG HE SOUNDS SOOOOOOO PRETTY ON THE PHONE ::melts:: He\'s got this like voice that trips me out. Its kinda sessy like LOL!  Hope he doesn\'t read this LOL!! That\'d be kinda embarrissing like! Yeah I acted like a \'tard on the phone. BUT OOOOOOOOOHHHHHH MY GOD UUUUUGGGGGGGGGGGGGGHHHHHH!  OMG YOU SOUND SO PRETTY!!!   Mood:Elatedly floating Song: NONE THERES NONE TO DESCRIBE THE FEELING!!   '),(2698,2224636,'03,March,2004','2004-03-03 00:00:00','    FUNNY BUT SO SO WRONG LOL     Fuck Her Gently Lyrics    This is a song for the ladies  But fellas listen closely  You don\'t always have to fuck her hard  In fact sometimes that\'s not right to do  Sometimes you\'ve got to make some love  And fuckin give her some smoochies too  Sometimes ya got to squeeze  Sometimes you\'ve got to say please  Sometime you\'ve got to say hey  I\'m gonna Fuck you softly  I\'m gonna screw you gently  I\'m gonna hump you sweetly  I\'m gonna ball you discreetly  And then you say hey I bought you flowers  And then you say wait a minute sally  I think I got somethin in my teeth  Could you get it out for me  That\'s fuckin teamwork  Whats your favorite posish?  That\'s cool with me  Its not my favorite  But I\'ll do it for you  Whats your favorite dish?  Im not gonna cook it  But ill order it from Zanzibar  And then I\'m gonna love you completely  And then I\'ll fuckin fuck you discreetly  And then I\'ll fucking bone you completely  But then I\'m gonna fuck you hard  Hard   '),(2699,2224636,'03,March,2004','2004-03-03 00:00:00','    Could it ever happen     Could it ever happen to me again? That little thing everyone calls love? Or is that just a figment of my imgination? Could someone break down the walls around my heart? I believe they\'ve become too strong Even for a jackhammer If I ever find it again  Will it be happy and fun? Or will it be like all the others sad and hurt?     Mood: Lonely like Song Summer time by Calvin Richardson   '),(2700,2224636,'02,March,2004','2004-03-02 00:00:00','    OK so yeah a little depressing.... sorry    Today I got home and started to listen to \"My Immortal\" and  cried like a fucking baby!!!!!  I just started to think (for once not about Jerry or Pinz) but about my son. You know its not like I want the memory of him to go away but I miss him so much. And I would\'ve absolutly love to wipe away all of his tears and chase away all of his fears and hold his hand throughout the years to come. But I can\'t do that. I know what I did was right for him. I know that he will get the life and all the love he deserves. But no one will ever ever love him as much as I do! Not even Jerry. A mothers love cannot compare with anyone elses that they will feel for their child. I\'m dying to hold him again. I\'ve been told that he\'s getting bigger and stronger every day. I know that he\'s healthy. I know that he\'s alright. But to see him again will bring my spirits way way up. To all of you that are reading this, I know that you don\'t know how I feel right now you can\'t really understand, but to the future mothers when you go through all the 9 months of back aches stomach aches and everything else and then that tiny little life comes out of you and you hold him/her for the first time you will understand the feelings that I have. You will never want to let him/her go. You will promise so many things to him/her. You\'ll cry because you\'re so damn happy. To the future fathers out there, being with your wife as she goes through this will be hard but when it come to the end of it, and you see that person that you created i don\'t think you won\'t help but to shed a tear. I know Jerry did.  ::Mood- Happy but sad Song: My Immortal::   '),(2701,2224636,'29,April,2004','2004-04-29 00:00:00','    I\'d like to smile and for once completely mean it     I\'m tired. I\'m tired of so many things. You know I\'ve never really confessed that I hate my mom and dad for making me give up my son. In actuallity it wasn\'t my choice. It was there. And that hurts. I mean damn, they made me give up my first born. I don\'t think anyone can even imagine how much that hurts. The anger that I hold so far back in my mind is dragging me down.   I sick of acting like I\'m 13. I tired of acting like and idiot and not just ask that person out. I wish I could but I can\'t because I don\'t really think he\'d say yeah. I mean he might but would it because he doesn\'t want to hurt my feelings or would it be because he actually wants to?   I\'m really tired of every time I try to have a  decent  conversation with my mother we argue. It\'s like we can\'t act decent. I mean yeah there are those days that we get along and ladida but most of the time we\'re arguing over stupid ass things. OMG it\'s really ridiculous.  I\'m honestly ready to just be happy. I want to smile and not feel like I\'m being fake. I want to laugh and not think of bad things. I want my thoughts to be clear. I want to just be free of the pain and the hurt and the damn anger. I want so damn bad for every ounce of confusion I have in me to be gone. I\'d like to just feel rested. I don\'t want to be paranoid of everyone around me. I don\'t want to feel like I\'m being judged all the time.   Sometimes I just feel like I\'m being held back. Things are just keeping me down. God, now I realize why I did drugs. Why I drank. Because I didn\'t have to feel. I didn\'t have to think. But I guess it\'s time for me to face all those things and just deal. But I don\'t want to to be honest. I\'d just like to run away from it all. Give up. Not care. But all the thoughts are invading my head. I\'m just so tired of it all.      Mood: Lost Song: Fortunant Son~ CCR   '),(2702,2224636,'28,April,2004','2004-04-28 00:00:00','    OOOOO! AHHHH! I DON\'T FUCKING CARE!!     Man, my feelings right now are just hurt. I called over to Jerry\'s house because I\'d like to hang out with him. Well gee, DENIED!!! He\'s all the sudden become to fucking busy to hang out with me. BUT he can hang out with Korie and Stephanie?? WHAT IN THE FLYING FUCK???? He\'s all like \"Well I gotta work you know??\" UM YEAH I KNOW THAT I\'M NOT STUPID!! Gawd but he doesn\'t understand that I\'d like to hang out with. I have pictures of his son to see but damn you know what he says? \"When I have the time.\" WHEN HE HAS THE FUCKING TIME!!! That really hurt to hear. It really hurts that he\'d rather go smoke up and go to some god damned flea market to help sell shit that none of the damn profit is going to him than just chill with me. It makes me want to cry. I mean he\'s going to be gone for like three days and he can\'t chill with me for two hours. OH I\'M SORRY I\'M ATTEMPTING TO TAKE UP SOOOOOOOO MUCH OF YOUR TIME!! Shit now I\'m crying because I\'m so frustrated. Damn it! You know I\'m really glad that we aren\'t together because it\'d just be all heart ache and shit. I\'m so tired of it. I\'m so tired of his bullshit. You know I\'d just like to be happy. But everything around me is like saying NOPE SORRY!!! YOU\'RE GONNA BE MISERABLE FOREVER!! I\'m tired of being denied, rejected, pushed away, hurt!! I\'m so tired of all that bullshit. I just want it to end. End now! I can\'t take all that crap from him, I can\'t take the confusion I have over some matters, I can\'t get away from the thought of failing this really important ap exam, I\'m ready for it to be over.......... now. It makes me sick that I have all this stupid stupid bullshit in my life making me so down. God and you know what\'s even more retarded to pile more shit on?? Acutally I don\'t want to write about it. Fuck it I\'m gone.      Mood: Thouroughly pissed Song: What you are ~ Audioslave   '),(2703,2224636,'28,April,2004','2004-04-28 00:00:00','    The crowd is dragging me through the mud and spitting in my face!!      Well Monday is going to be hell for me. I have to go aalllllllllll the way to the West campus of CPCC to take the stupid English AP exams. At least I don\'t have to be in school that day like AT ALL!!!! WOO WOO!  I don\'t understand but all this week I have been feeling simply drained. Like I have no energy AT ALL! I don\'t really understand why. Maybe its because I\'m PMSing (yeah yeah TMI) Or maybe its just I have no energy.  Today I get to go and get this spray on tan for me. It\'s so I won\'t be so damn white when I go to the prom. *I\'M SO WHITE I GLOW BABY* Then I get to go get my shoes for my dress. Which you won\'t even see because my dress is long. OH MY GOD! My dress is simply gorgous it\'s black and pink. It kinda fitted in the top to show off my figure just enough and then it goes out just a little bit. It solid  black  with this bright strip of hot hooker  pink  across the top and then the flowers and vines cascade down the dress in  pink  as well. I LOVE IT!!  Ha! That reminds me. I was talking to my friend Stacey on line last night and I told her about the spray on I was getting and she said to me \"Have you no shame?? Us females work hard at laying down in those 100000000 degree tanning beds to get this dark and you\'re cheating!!! Cheating by getting a tan sprayed on!! You so suck you know that right??\" I said to her \"OK you know that shame does not run in my vocabulary. So no I\'m not ashamed of cheating. BECAUSE YOU KNOW WHAT I\'M EMILY! SO BOOYAH!!\" and we laughed and laughed and it was funny.  Ugh! Damn it! I was about to ask Chris out at lunch today but NO!!! The words wouldn\'t form. They sat there on my lips shakin their asses at me and laughing.  I just noticed I have heart burn and it hurts. LOL      Mood: In pain but happy but drained as well (oooo too many emotions) Song: Want you to want me   '),(2704,2224636,'27,April,2004','2004-04-27 00:00:00','    I\'ll just say whatever falls out      OK. So I told you I like you. I\'m kind of regretting that I\'ve told you now. I mean damn I know that Skittle is kinda pushing you to ask me out. What\'s the point when you like her better? I mean it\'s obvious.   But to be honest, if you actually do like me, I\'d ask you out. ::laughs  hysterically:: I don\'t think you can imagine how many times I\'ve actually tried to. I just can\'t get the words to come out right I guess. I don\'t know. But now that  you know how I feel about you, it\'s like you don\'t want to talk to me much anymore.  Yeah, that kind of hurts. I don\'t know. Maybe I\'m assuming some of the things  I\'ve said. But I\'m really not being proved wrong. I\'ve probably messed up our friendship by telling you how I feel. I mean regaurdless of whatever I still like  you. But hey man whatever.   I\'m just kind of tired of dwelling on the fact that you say one thing to me and  then a bunch of other shit to Skittle. I\'m tired of things being hid from me. I\'d  like to know what the fuck is going on. Somebody fucking clue me in. It would  be nice to know. But its funny that if you were going to talk to anyone about  liking me you\'d talk to oh gee possibly ME!!   OK. I\'m not trying to sound pissed off. I\'m not. I\'m just like at a loss sometimes.  It won\'t matter any way right?       Mood: Just confused Song: anything by Maroon 5   '),(2705,2224636,'25,April,2004','2004-04-25 00:00:00','    OMG MY EARS!!!! MY EARS!!!     OK! I can\'t take it anymore. American Idol. IT HAS TURNED INTO A TERRIBLE AWFUL HORRIFIC SHOW!!!! The people on there suck. They suck so bad they make me laugh with embarassment for them! Did the judges have to just fill spots with insanly crappy people. I mean damn, I know that I\'m not the best singer in the entire world but OMG I\'M NOT THAT BAD!!!! That one red headed guy about made me cry he was that terrible. Those people are going all off pitch, sounding all tone deaf and everything. I mean, the try outs there were some really really good people and yet they said they weren\'t good enough! What the hell?? Now I bet the people who created that show are kicking their own asses for picking the worst people in the damn bunch. What were they looking for?? Another Clay Aiken?? HELLO THERE\'S ONLY ONE!!! I mean did they honestly expect another geeky loser to come in and blow them away?? ::passes out:: This is so sad. I sit there watching that show hoping to God one of them does a fantastic show and I end up really really disappointed. I can\'t stand it. I bet that there won\'t be another season of that show. And if there is GOD SAVE US!!!!      Mood: Disgusted Song: THEY RUINED ALL OF THEM   '),(2706,2224636,'25,April,2004','2004-04-25 00:00:00','    I\'m ready to just curl up and die      As the end of the year comes closer and closer I find myself losing more and more of my sanity. I don\'t know exactly why but I just feel that way. There are just so many things going through my head. Screaming in my brain.    Worries begin to surface like the AP exam for starters. God and then the math exams have me worried off my ass because all I want to do is get the fuck out of Geometry and graduate on time. That\'s all I want. Plus you know I want to get those fabulous grades to make it that much easier for me to get into UNCC. My number one choice of course.  Then there are those questions that are unanswered. Most of which I really don\'t feel like expolting all over my blog. Because if you read all my entries you\'ll know what they are.  Then Kelly is worring me because she\'s in pain from her and Steve\'s break up. I sat here and tried to console her the best I can. I mean of course I know how she feels I went through all that crap with the break up with Jerry. You know you still believe you\'re supposed to be together yet you still can\'t convince the one you love to give you guys another chance. It sucks it really does. I know. But you know as they say with time pain will pass. It\'ll hurt for a while. I think Red from that 70s show said it best \"Things started to get serious and she broke it off. Man that hurt. But as time went by I started to think less and less of her. Then one day I stopped thinking about her all together. Then that hurt. But I knew it was going to be OK.\"    Sometimes I feel like what sanity I have left is just slipping from my fingers. It\'s odd how I just bounce back and forth from being \"OH WOE IS ME!\" to \"WOO WOO SOOOOOO HAPPY!!\"  Oh well.  On a better note. Prom is in 4 days and I\'m all excited. I still need to figure out if Brandon is picking me up here or I\'m going to his house. Cuz fortunantly enough my dad is cool about Brandon. Plus today I got some seriously cute pictures of Joshua. OMG they\'re so sweet I can\'t wait to meet his parents.      Mood:: Slipping further into my insanity Song: Running Blind~ Godsmack   '),(2707,2224636,'25,April,2004','2004-04-25 00:00:00','    I hate these dwelling thougts     OK so yeah I like this guy. Supposedly he likes me too. Yet its hard for him to decide if he\'d wanna date me or whatever. Why?? It\'s pissing me off to say the least. Its like dude if you don\'t wanna date me or whatever just say it damn. Or maybe it\'s not no your mind at all. Whatever.  Stupid me being the person that I am sits here and dwells on stupid shit like this. I feel 10 again and its really fucking sad. I should just get over it. But I don\'t know I really really like him. I don\'t know what it is but I do. I mean he\'s nice and funny and all that stuff.   God this is making me sick. I hate this. I hate how somebody says they like you and then they do  nothing  about it.      Mood: Kinda pissed Song: Play me ~ KoRn   '),(2708,2224636,'24,April,2004','2004-04-24 00:00:00','    The Numbness in my shoulder scares me     I woke this morning and I  hurt !!! OMG you don\'t even understand. It keeps going numb and that worries me just a little but I\'m sure it\'ll be OK.   So yeah. Not a whole lot happened today. I\'m not even sure why I\'m writting. So I guess I\'m gonna stop now.      Mood: Eh.....whatever Song: Under the Red Hot Moon   '),(2709,2224636,'24,April,2004','2004-04-24 00:00:00','    I just love this song it really fits my mood    Dare You To Move Lyrics ~Switchfoot    Welcome to the planet Welcome to existence Everyone\'s here Everyone\'s here Everybody\'s watching you now Everybody waits for you now What happens next What happens next [Chorus] I dare you to move I dare you to move I dare you to lift yourself up off the floor I dare you to move I dare you to move Like today never happened Today never happened before Welcome to the fallout Welcome to resistance The tension is here Tension is here Between who you are and who you could be Between how it is and how it should be [Chorus] Maybe redemption has stories to tell Maybe forgiveness is right where you fell Where can you run to escape from yourself? Where you gonna go? Where you gonna go? Salvation is here I dare you to move I dare you to move I dare you to lift yourself up off the floor I dare you to move I dare you to move Like today never happened Today never happened Today never happened Today never happened before   '),(2710,2224636,'24,April,2004','2004-04-24 00:00:00','    And the mood of the blog goes sad     OK LOOK HERE MAN!! I don\'t think you understand. I REALLY LIKE YOU!! I mean that honestly. I think that  you\'re sweet and funny and you\'re nice.  I mean its really really fun to be around you. You constantly have me laughing to the point I can\'t eat! I don\'t know really what it is about you but I\'m just attracted to you.  But this shit with the \"I gotta think about it\" that you keep telling Hayley is getting on my nerves. You say you like me to several other people but then you tell Hayley you\'re not sure. But then you say you do. DUDE YOU\'RE GETTING ME SO LOST I NEED A FUCKING MAP!!! Is it really hard to decide if you want to ask me out or what? Or is it just that you don\'t like me that way and you just said that to make me feel better.   I wish you\'d let me in on what\'s going inside your head. I mean dude, if you don\'t like me like that tell me so I don\'t have to dwell on what\'s going to happen next. Ya know? I wouldn\'t put you through something like this. It\'s sweet sick torture for me.  I mean damn Hayley\'s sitting there telling me to ask you out. NO! I can\'t! I can\'t risk asking you out and then you saying no. ME AND REJECTION DO NOT PLAY WELL TOGETHER!!! I can\'t stand the thought of it. It scares me. That is why I never tell any guy I like them unless I am absolutly positive that I know he likes me for like seriously for real!!  Just let me in on what\'s going on!!!!!!! Quit making me suffer lol!! Damn boy COME ON!!!!!!    Mood: Utterly lost Song: Dare you to Move~ Switchfoot   '),(2711,2224636,'24,April,2004','2004-04-24 00:00:00','    OK so the happy blog goes first      TONIGHT WAS SO MUCH FUCKING FUN!!! Me and Skittle and Stewie went to Tremont again for the second night of the punk show. AND IT SOOOOOOO KICKED ASS!!!! All night Skittle was like \"You have to mosh or dance with me or something.\" I was like FINNNNNNNNNNNEEEEEEEEEEEEEE! I couldn\'t get my self hyped up enough to but I was working on it.   Met some new people. Like Ross. He kinda cute in a goofy way. But he was drunk but hey he was talking to me so I was \"Okie Dokie this is fun.\" He only liked me because my nick name reminded him of a porno (i.e. Pretty Peaches)   Did a B- DOUBLE- E- DOUBLE- R- U- N BEER RUN!!!! For a couple of friends (JaR and Taco). We left them at the gas station though cuz piggy\'s rolled up. LOL.   The music was absoulty fabulous. I honestly like all this punk stuff. I used to listen to it and thought it was shitty. Straight shitty. But now I\'m getting into and its really cool. It gets you all hyped up and its so fun to just sit there and scream and dance and all that stuff while listening. I mean its not like I\'m going to run around screaming that I\'m a punk because I\'m not. But I really like the music. Are you happy Skittle you little punk you. I swear you are Heroin Bob and I\'m Steve-o and Chris is Mike LMFAO SLC PUNKS!!!!!!!  So it was almost time to go and I still hadn\'t moshed or danced or anything so when Minority Party (i think thats the name) came on I was like \"Okie Dokie Emily here we go.\" Thankfully Chris jumped in so he and I moshed by ourselves and then alllllllllllllllllllllll these people started getting. It was funny. It was even funnier when I fell on my ass. That hurt but I was laughing. Skittle put her arm around me and we flew in to people just spinning and shit. WE MADE THE BIGGEST MOSH PIT OF THE ENTIRE NIGHT!!!!!! OH OH GO US!!!! But we like got seperated because some asshole yanked my shoulder almost out of place. I kept going around and around and then I fell on my ass again and got stepped on and shit. I yelled out \"Fuck somebody help me!!\" And like 5 guys from the other side came rushing up and blocked moshers from me and got me up. I was like thank god man lol. My knee was like getting crushed. Man, omg, moshing give you like a total adrenaline rush. I AM NO LONG A MOSH PIT VIRGIN!!!!!! LMFAO!! It is so much fun. I don\'t know why I didn\'t start doing it sooner. It really hurts though when the drunk fuckers get in on it though. Because they like punch you and shit. One fucker while he was moshing and i was just standing there waiting for the right time to get back in came and like punched me in the chest. I was like bitch ass!!!!  Wow I just looked at that paragraph. That\'s sad that I can write that much on some moshing.     Mood: JUST SO HYPER AND HAPPY AND WOOOOOOOOOOOOOOOOOOO!!! Song: All the shit I listened to tonight!!!!!!!!!!   '),(2712,2224636,'24,April,2004','2004-04-24 00:00:00','    Fridays are fun!!      Yesterday was so much fucking fun. After a long ass day at school that wasn\'t all that great the events afterward made me happy.  First off me and Skittle and Stewie (lol funny), went to the skate park to go and see a friend. When we got there Price wasn\'t there after beggin us to come. Gayeness. Me and Skittle swore to kick his ass if he didn\'t show up. But after a few minutes of fucking around on the play ground he showed up. Amazing! And stupid me told them the story of Price \"making love to my peaches\" that i had given him for lunch and then stupidly I told them some girl was calling me peaches in my Sports med class and so now I have recieved the nickname Peaches! How gay. A nickname is one you can never give yourself but one you hate the most. But I can live. I got over it after Price sat there and taunted me with it for like 2 minutes. But we\'ve nicknamed him Bobby Hill cuz he looks just like him. LMFAO  After the skate park we went and got my money from my mom and headed off to Tremont Music Hall. I really felt out of place at first because it was a Punk show. I really don\'t listen to punk music unless I\'m with Skittle. But the people there were cool. I don\'t think I\'ve ever seen that many mohawks and leather vests or jackets in my life. Well I\'ve see plenty of leather vests and jackets minus the mohawks. Bikers ya know lol.  But damn, Skittle kept wanting me to get in the pit and I was like NO! These pits looked rough as hell! Plus I got a $5000 mouth maybe more now. I dont\' need to get that fucked up moms would have an absolute heart attack. I walk in the door with a busted mouth \"HI MOM!\" ::mother passes out on the floor:: \"What?\" Oh I can see it now lol.  The concert was hella fun and we\'re going back for the second one tonight. I\'m moshing tonight though. PEEZ DON\"T HURT ME LOL!!!    Mood: FAN-FUCKING-TABULOUS Song: Tilt-a-Whirl   '),(2713,2224636,'22,April,2004','2004-04-22 00:00:00','    OK NO NO NO YOU DON\'T LICK THE HEEL OF HER BOOT!      OK Jerry you and I have got an issue. I am sick and tired of being ditched by you. I am sick and tired of it. Thursdays is our day to hang to out. And you fucking know this.  But damn it all to hell if Sarah doesn\'t call asking you and Elf to go to Morey\'s with her. WHAT THE FUCK?? And you say \"Of course we\'ll come!\" Knowing full well that I can\'t go!! You know that I have school and I\'m the \"baby\" of the circle. Why in the fuck must you ditch me for her??  Are you trying to get some from her? Hun, I can tell you now it ain\'t gonna happen. She more than likely thinks she\'s too good for you. Everytime she comes around you and Elf are at her beck and call. I\'m sick of you acting like I don\'t exist.  What do you lick the heel of her boot now?? Are you her bitch?? No? Well fuck man. Stop being a dick and act like we\'re best friends.   Mood: Pissed Song: Fuck you!   '),(2714,2224636,'21,April,2004','2004-04-21 00:00:00','    OK. So you\'re over \"Us\" but you still love me. THAT MAKES NO SENSE FUCK BAG!!!!!      OK Jerry! You told me tonight after I so stupidly asked you that you\'re over us. As in the relationship. I knew it was coming sooner or later but damn it still hurts. When you told me that you were it was like 500 BILLION arrows flew straight into my heart laughing as they pierced it.  I mean its liike \"Well, damn!\" I understand we\'re both eventually gonna move on but it still hurts. I mean its gonna kill me to see you with another female. But sooner or later (more than likely later) I\'ll find somebody.   Don\'t get me wrong babe, I\'ll always love you. You\'ll forever have a place in my heart. And until we die you\'ll always be my best friend. I mean you and I are forever connected whether we like it or not because of our son. But you do need to understand shit like that still hurts. I don\'t know how else to feel.  Yet, it doesn\'t make sense that you sit there and tell me that you\'re over \"us\" but you\'ll forever love me. HOW DO YOU LOVE SOME ONE AND  NOT  WANT TO BE WITH THEM????  I\'m slowly accepting the fact that we\'re apparently not supposed to be together. As much as I want to fight for \"us\" I\'m gonna have to let it allllllll die. And that hurts too. I\'m a fucking fighter. I want something and I\'ll run myself into the ground trying to get it. But now I think for once, I\'m letting go and giving up on this certain fight. I\'m pretty sure you\'ll be happy about that.  At least we\'re best friends and the sex is still good!!! LMAO.    Mood: Sad but I know it\'ll be okie dokie Song: No More Love ~ Shinedown   '),(2715,2224636,'20,April,2004','2004-04-20 00:00:00','    I\'ve stepped outside myself and found a person I don\'t know     Well damn. Two of my friends who were going out broke up. Sadness. I don\'t know what else to say about that.  The school year is coming close to an end. I\'m so happy. This year has been rough for me. With being pregnant for like most of the year and all. That turned out good though. Joshua is getting soooooo big. He\'s a damn whopper. I love it. He\'s gonna be like a line backer or something for a damn football team.   The break up with Jerry has come and gone. I\'ve realized that I\'m so much happier being his friend. I love how he introduces me to someone new as his best friend. I swear in like 10 years we\'re gonna be \"Will and Grace\" with out Jerry being gay. LMAO. OMG if he turned out that way though I\'d die laughing. Just to imagine Jerry makes me laugh. Oh I love him. I always will. But that will never change. I mean yes, he and I will move. On maybe we\'ll get back together maybe we won\'t. Who know\'s who cares. Being best friends is all I could ever ask for. I LOVE YOU JERRY   Then Hayley (Skittle). My best homie in the entire world. She\'s been there for me through thick and thin. Letting me piss and moan bitch and groan. She\'s never told me what I  wanted  to hear but what I  needed  to hear. I love you for that homie. You and I have had our share of guy issues. ::clears throat:: Toad, Smurf, Pinz, ummmmmmmm all them lol. We have sooooo many damn memories isn\'t not even funny. Forever getting lost then some how making it home. Screaming Zug Island as we flew on down the road to some new county. LMAO. Playing the anarchy music in small closed areas for all to hear lol. All the damn inside jokes. So many. ROLLING PIN OF DEATH!!!! I love ya homie!!!!!  School. School has been just ugh. I mean even though I\'ve made honor roll all this year and held a 3.40 on average I\'ve hated Independence. It\'s just a fun school. It\'s stuck on it foot ball pride. How gay. I mean I like most of the people there but damn then you\'ve got those who are so prided in the schools 4X state champion. Wow that\'s not what schools all about. Am I the only one who believes that??   (more will be added later look for it)      Mood: Thoughtful like Song: Fly with me~ Zug Island   '),(2716,2224636,'18,April,2004','2004-04-18 00:00:00','    Remember when like 10 people in an entire school where the only ones who were depressed??   Holy shit man. I\'m sitting here last night just reading random peoples blogs and you know what fucking pattern I see? Everybody is fucking depressed? I mean holy hell you guys? What the fuck? Has being depressed become the new trend or something? I mean I know we all have our hardships and things that totally get us down but damn you\'re not AUTOMATICALLY depressed. I mean I don\'t think I\'ve actually sat here and actually written out that \"I\'m soooooooooo depressed.\" Until I\'m clinically diagnosed with the shit I don\'t say it.  And if you all are so damn depressed maybe it\'s time for a damn change. OMG! I just sat here reading y\'alls shit and was amazed at how sad you all were. I mean I know there are times where I get down but damn I don\'t say I\'m depressed. Well I do but I\'m only joking.   You all aren\'t depressed. You\'re just sad. SAD THAT\'S ALL IT IS! Sorry to pop your bubbles. You all can\'t be basket cases. Unlike me! LOL only kidding. Well wait..... THAT\'S A WHOLE OTHER SUBJECT! LOL.  You guys if you\'re so \"depressed\" change some shit around. Get away from the things that make you so \"depressed.\"  \"But I can\'t get away from my parents!\" AH FUCK YOUR PARENTS! You have like what 1 to 2 years until your out of there. Fuck give them hell. It\'ll be fun I swear! You won\'t be so depressed all the time then. You could actually have fun. And instead of sittin on your asses get UP AND DO SOMETHING. God have you fuckers out there are depressed because you have WAY too much going on at your schools. Be lazy like me. Fuck the clubs, fuck the plays, fuck all that shit. You honestly don\'t need it. \"But that\'s what colleges look at.\" If your grades aren\'t good enough then they aren\'t EVEN going to peek at your fucking extra ciricular activities. I mean damn.  You all are just stressing your selves to the point where you actually believe you\'re depressed. It\'s truely sad. Really. I mean I feel for you guys but you all just need to take time to chill out. Clear your minds. Cleans your soul and spirit.  NO WAIT WAIT PERFECT SOLUTION! Why don\'t you all just do drugs. Swear to God you won\'t be stressed or depressed any more. LOL. OMG don\'t even take that seriously. I\'m only kidding.  Y\'all just need to chill out. I say this cuz I love you guys. There\'s no point in thinking your depressed. Because in all honesty you\'re not.   Mood: Feeling Pity for once Song: Highway ~Audioslave   '),(2717,2224636,'17,April,2004','2004-04-17 00:00:00','    You talking scared the living shit out of me   OMG! So much was said. Most of it didn\'t make sense. I can\'t really figure out why. How come you said all that stuff to me then told me to forget it? You must have felt comfortable enough around me to tell me all that you did so how can I forget? Are you really that worried that\'ll I\'ll go back on my word and and tell the whole world what was said? Don\'t worry a promise is a promise and I keep those.  But damn, half the shit you said scared the shit out of me. Most of which didn\'t make any sense. You seemed so scatter-brained. Yet it was like you tried your damnedest to clarify it all. You sounded so hurt. There was so much pain in what you said. All I wanted to do was give you a big ass hug and hold you and tell you everything was going to be OK. But I couldn\'t. I couldn\'t even get a damn number right to call you because you wanted me to.   And another thing. Why did you want me to be there so bad? Why me of all people? I\'m not that great I can promise you that. And did it slip when you said you loved me. Was it just randomness or was it the truth.  What the fuck is going on with you sweetheart? I\'m really beginning to worry. I can\'t just forget the things you said. There\'s no possible way. It\'s kind of your fault. You burned them on to my brain. Reason being, you repeated yourself A LOT! But I\'ll try to forget because you asked me to.  You asked me why I was sad. Here\'s why. Because I care. I don\'t like to see the ones I care for suffer in any way. I want everybody to be happy and carefree. I mean I know that\'s not possible for everybody but at least I can give it a shot to help and make others happy.  You also asked me why I felt bad. For all the same reasons I was sad. I don\'t think you really understand how much I  do  care for you. Even if you told me not to care I couldn\'t help myself but to.  I\'m still wondering why you said those things you did. Was it to get reactions out of me? Were you trying to see what I\'d say? Were you trying to see how much I really do care? Because I can tell you all those things. I can tell you how I\'d react or what I\'d say or how much I do care.   You really worried me last night. I could have sworn you were going insane. But it felt like you were using me to keep you grounded in some sense of reality. Was that what it was? Are you that comfortable with me that in times of need I can some how keep you grounded?  What is going on with you? I\'m racking my brain and trying to figure out the answers but they\'re not there.    Mood: Lost! Really really lost! I NEED A DAMN MAP! HOLY HELL! Song: there is none right now   '),(2718,2224636,'17,April,2004','2004-04-17 00:00:00','    A random attempt of a poem (it probably sucks lol)   Eyes drawn out in black to hide the ocean of blue. Lips forever colored darker than they should be never letting the natural show through. Hair color changed for some vain attempt to be different. She\'s sold. She\'s cold. Using her body to flirt. Smile and smirk. Thinking the next guys gonna be just some asshole jerk. Wanting so bad to have another hold her. Keep her warm. She\'s sold. She\'s cold. Kiss her lip. See her smile. Never could you imagine her pain. Never could you know the scars that tear along her soul. Her heart. She\'s sold. She\'s cold. Many have tried to renew her blackened heart. Most have failed. She\'s cried a thousand times. She\'s lost too many things. She\'s starting just not to care. She\'s sold. She\'s so damn cold.   '),(2719,2224636,'16,April,2004','2004-04-16 00:00:00','    Pirates of the Caribbean!    urlLink   What movie Do you Belong in?(many different outcomes!)     brought to you by  urlLink Quizilla   '),(2720,2224636,'15,April,2004','2004-04-15 00:00:00','    You hold your life in your hands   Your Life Line reveals that you like to stay close to home. You are not very adventurous or aggressive and have a more delicate constitution.  Your Head Line reveals that you lack confidence, but have a sense of grievance. You carry a chip on your shoulder, but will vent this emotion through participating in political activities.  Your Heart Line reveals that you are a passionate person with a strong sex drive. However, you tend to be selfish and put your own needs before those of your partner.  Your Fate Line reveals that you will have an early and independent start to your career, and will be successful in your life\'s work.  You have a Fire hand. You are a person who is restless, outgoing, ambitious, confident and at times promiscuous. Your optimal career choices are to work as a lawyer, politician, entertainer, and an athlete in non-contact sports.   '),(2721,2224636,'15,April,2004','2004-04-15 00:00:00','    You will not understand it in its entirity unless you are the person I\'m writing about    (Actually written 4/14)   OMG! Boiling! Insanly pissed! God, does he think I am  that  stupid?? Does he seriously believe I am some   TOY   he can fuck around with?   The fires of hell will consume this Earth before that happens!   \"Oh, oh! But I don\'t want to hurt her! But I want to do thinkgs with you!\" Well damn, Monday, in the backseat you sure as hell acted like it!      You kissed me! You pulled me close! You laced your fingers with mine! IT WASN\'T ALL ME!!!     I don\'t want to hurt her either. But damn it all to hell if that doesn\'t make me want you more!   Yet you can\'t even  IMAGINE  how much it hurts to have you tell me what you did then throw your relationship with her  ALL UP IN MY FACE!!  Do you find that funny? Does it please you to see me suffer??   You of all people know my story with guys! You know the pain that silently rips me apart.  Fucker! I am  NOT  some one for you to toy with! My mind, my emotions! Those are not things to fuck with.  HA! You don\'t try to mess with an emotionally unstable bitch!    I WILL FUCK YOUR ENTIRE LIFE UP!    You don\'t even understand how  easily  I could! I could call her right now and tell her   EVERYTHING  that\'s happened between us!  But no, no! I just want you to be happy.  THAT\'S BULLSHIT!  I want you to step inside my head, my heart. Feel what I feel. Think what I think. Live my damn turmoil I silently suffer with every damn day.  You think you know? You honestly do? I mean, I know you have pain of your own. But shit, you don\'t know the half of mine. You try and deal with losing your first love over some   bullshit  You try and deal with having your first born child and have to give him up! Can you do that? Can you even fathom taht heart break?? No I don\'t think so! So how the    FUCK   do you think you have any right to fuck with my head??  Now, maybe I\'m accusing you of too much. Maybe I\'m completely and utterly wrong. You could actually care. You just might possibly have feelings for me. But    YOU    make it so damn unclear.  Get your shit straight. Please. You\'re putting me through hell. It\'s not right. It\'s unlawful.   You know I\'d love to be with you. Yet, you are the one holding \"us\" back. But I\'m not pushing. I\'ll let what could be come as it may. I fucking refuse to chase you. I\'ve done that one too many times. I f you want me you ask for me. But that\'s when you want  only  me.   Mood: Tired of the shit Song: She fucking hates me ~Puddle of Mud   '),(2722,2224636,'15,April,2004','2004-04-15 00:00:00','    I always believed the number 13 was unlucky. It\'s proved itself true    (actually written 4/13)   Damn, these past two days have been just ugh! Last night Skittle and I kidnapped Pinz, just to get him out of the house. And I knew \'something\' was going to happen. It always does. So we go and surprise Alex. Well we spent some time up there. That was all fun. Well Skittle and Alex were saying or doing their goodbyes ( =D ) me and Pinz walked outside. I asked him what was on his mind because, well, the entire car ride up to Alex\'s he just looked like he was thinking about something. You know he kept putting his head in his hands and sighing. He tells me then, he\'s got feelings for me. I was like  hold up, pause, rewind, play again   WHAT???  That took me totally by surprise. I mean twice now, I\'ve told him I have feelings for him and I\'ve gotten no response. Nothing. And now he tells me this? Damn. So the entire time on the ride home we sat there in the back talking. I mean I\'m sitting there thinking \"Wow, he might actually have  real  feelings for me. All the while we\'re holding hands and each other. When me and Hayley got back to my house I was like in the best damn mood. In awe and everything.  HA! That mood got killed tonight. We went and got Kelly. Don\'t get me wrong I love the girl to death. She\'s one cool ass mother fucker. She\'s also Pinz\'s girl.   There\'s your fucking twist  We took Kelly with us to go and get Pinz because he was going to go and stay at Alex\'s. Once Kelly and him got together it was like  ::SMACK::  ALL UP IN MY FACE all up on her. Throwing his relationship with her in my face. I\'m sitting there thinking to myself \"Damn guess what you said was BULLSHIT! Or was it? God, I feel like such a damn fool. I mean what the fuck??\" He goes and tells me one thing and then acts another?? Am I just something for him to play with? I refuse to be a toy. My emotions and my heart are NOT things to play with!  I\'m a damned human. I swear. It hurt so fucking much to see him and Kelly hands locked, kisses exchanged. And me sitting there trying not to cry. The entire night I held back my tears and anger. Not now. Not now since I\'m alone. Forever alone.  I FEEL LIKE EVERYONE\'S TOY!!     Somebody  please  PROVE ME WRONG!!  Mood:Kill me please Song: Fuck the World~ ICP   '),(2723,2224636,'12,April,2004','2004-04-12 00:00:00','    These three days have been a blissful blur part 3   Ok so this is the final part to all this aren\'t you so happy??  Well Saturday I get online to check my mail and Skittle pops up asking me to join her and 4 others on a Scavenger Hunt. I was like hell yeah that sounds like fun. So we have certain things we have to find. I got arrested and hugged by this hottie of a cop. YESH!  Highlight!   Nikki sat on the Easter Bunny at Carolina Place Mall. That bunny did naughty things to her back! ::shocked:: I thought they were good natured non-dirty things lol. But by then end of the hunt we won by like twenty points and it was all good.    Mood: Happy Song: No more love ~ Shine down   '),(2724,2224636,'11,April,2004','2004-04-11 00:00:00','    These three days have been a blissful blur and I LOVE IT part 2!!   (there is one more part)   Well gee friday. I love Friday! Friday was so much fun! I decided to skip school. What was the real point of going. No one was going to be there. I had no reason to be there so my mom actually was all cool and said I could. Sleep is nice. Sleep is my friend. Except when your homie calls you at 9:45 actually expecting you to be up (I love you Skittle but daaaaaaaaaamn come on now lol). Then sleep left me when Skittle called me again at 10:30 telling me to get up and that I was going to go somewhere with her. I was like \"Uggggggggggghhhhhhh I hate you but OK!\" LOL. So at about 10:30 she came and got me with Chipley. I can\'t stand that boy sometimes but stealing his weed is just too damn easy. LOL. Even though I can get that from like anybody for free. (I don\'t do it allllllllllllllllllll the time so its all good I could stop whenever I wanted swear to God try me lol). So we had to take him back to his house for a minute but he ended up having to stay. So Skittle asked me where I wanted to go. I said lets go to Weddington. I think Chris is skating down there today. So we went down there couldn\'t find the place and ended up in S.C. LOL. Lancaster is fun!!!! Well Smurf (    Part three coming soon.  Mood: Just so damn relaxed Song: Stranger inside ~ Shine down   '),(2725,2224636,'10,April,2004','2004-04-10 00:00:00','    These three days have been a blissful blur and I LOVE IT!!  (there is three parts to this so yeah)  OK so this entry is gonna be kind of long. So if you\'re like the one person that reads these things I hope you have time lol.  So we shall start with Thursday. After school my homie Skittle came and got me. We decided to go and find the \"campsite\" that\'s behind our houses where Blade, Pinz, and Tacks have been going. So we parked her car up at Turtle Rock at the pool and began on the trail.  So we have absolutly no idea where the place is except for the fact that it\'s really close to 485 and all that mess. So Skittle and I are walking around and we see this little circle of, well not necessarily litte, but circle of big stones with a fire pit and a cross in the front of it. Weird thing was the cross had the scriptures Genisus 17:7 and Matthews 6:24 on it. Looked them up. But anyways I told her I thought this was it and she said she didn\'t think it was. Then I remembered Mary (chick whos gone up there) telling me they slept in a huge 10 person tent and you could see 485 almost clearly from where the site was. So we kept going. We ended up way off the trail which was ok because ended up from the back of Turtle Rock to the back of our own neighborhood lol. Well we finally got back on the path and decided to turn around. WE GOT TURNED AROUND ALRIGHT LOL! We were going up the path when I thought I saw something from the corner of my eye and screamed \"HOLY SHIT! THERE IT IS! WE FOUND IT!!\" lol. Kinda bad I was that damn excited lol. So we snooped around there for a bit then decided we really needed to get back. Well we couldn\'t figure out how to get back. We ended up sliding down a hill that was right beside 485 and walking along it. We got honked at by like 30 cars.  Highlight!!  Well we ended up walking through briars. GOT ONE STUCK ALL UP IN MY ARM!! ouchies! We jumped a barbed wire fence. OMG THAT SUCKED LOL! Then we ran down an overpass on 485 with on coming traffic. That scared the living shit out of me! OMG I was about to cry lol. We got to a chain link fence and Skittle was like we have to jump it and then we\'ll be almost home. I was like I can\'t do it. There\'s something about chain link fences that scares me lol. I really don\'t know what it is. So I found a hole and crawled through it with Skittle flipping me off the entire time lol. We ended up on a road like 3 miles from Turtle Rock. So she  called a friend and had him pick us up and take us back. She then took me to Smurfs where I fell asleep. THAT WAS AN ADVENTURE!!!    look for part two!  Mood: Elated Song: Burning Bright ~Shine down   '),(2726,2224636,'10,April,2004','2004-04-10 00:00:00','    Tomboy    urlLink   What\'s your sexual appeal?     brought to you by  urlLink Quizilla   '),(2727,2224636,'09,April,2004','2004-04-09 00:00:00','    The scar along my heart is the silent turmoil I\'ve recieved from love    Oh, gentle feelings, soft sounds, the goodness and the gradual stilling of a soul that has been moved; the melting happiness of the first tender touching joys of love-- where are you? Where are you? ~ Ivan Turgeney, First Love   In recent situations, I have discovered, that I, myself, no longer wants to find love. I really don\'t feel like finding it anymore. There is this gigantic scar along my heart from love. There\'s too much drama involved in it. There\'s too many tears to be cried. Love isn\'t supposed to be like that. Love is supposed to be happy, make you smile, make you laugh. But when the love you want is so close your finger tips can almost brush it, it\'s so damn far away. That\'s silent turmoil that I have to deal with every day. It\'s the turmoil that has been haunting me for four months now. Love is what\'s killing me. Wanting it. Needing it. To be loved by someone, to be truely loved, can make one feel complete. I once had that type of love. But twice now love ran away from me and reared its ugly head. I can\'t stand it. I keep wondering why this keeps happening to me! Is this a test of some sort? Is God punishing me for some wrong I\'ve commited? Every guy that I believe there could be something with either ends up not having the same feelings, ignores what I tell them, or leaves me. I\'m tired of it. There\'s no use in finding it. If love finds me then great, fabulous. I\'m done looking. Yes, I have given up.  It pains me to give up. I\'m really not one to give up. I never have been. I\'ve always fought for what I thought was right, what was supposed to be. But I can\'t anymore. I can\'t fight a battle that I\'m not going to win. I can\'t battle with love. I can\'t force it to be mine. As much as I want to. Things just can\'t happen. Maybe I\'m not one to have love in my life. I mean I love my friends. I love my family. I love my son. But it\'s like I have all this love to give and no one to give it to. Or to anyone who wants it.  I fought for a 2 1/2 year relationship. I tried my damn best to bring it back. But I couldn\'t. I\'ve expressed to one person, I believe twice now, how I feel about them, and was rejected. Fine, whatever, I\'m done. I can\'t do it anymore. It\'s killing me inside. It\'s tearing my soul to pieces. My heart is brusied, cut, scard. I\'m done. There\'s no fight left for love in me.  What\'s the point? If there is a point in fighting for love can somebody give me the reason. Because I can\'t seem to find one.   Mood: Hurt Song: Love song ~ ICP   '),(2728,2224636,'09,April,2004','2004-04-09 00:00:00','    My friend that takes me to my magic place      Your addicted to.....  LSD!  Wow what you must go through every day. LSD gives intense hallucinations and offen makes you think you can do things that no human can. I wouldn\'t try to pretend your a bird and play on top of a tall building anytime soon.    urlLink   What are you addicted to? (pics!)     brought to you by  urlLink Quizilla   '),(2729,2224636,'08,April,2004','2004-04-08 00:00:00','    Laughter touches my soul      These are some quotes I\'ve collected from friends that have made me laugh so hard.....   1.) OK special. Go make cookies. ~Coach Suggs 2.) We should have bombed Cuba and then their bombs would blow up. Then they would be nothing and we\'d be like \"Well....damn.\" ~Lucas 3.) Never been a quitter (puts a cigarrett in her mouth) Ain\'t gonna start now. ~Skittle imitating me every time I say I\'m going to quit 4.) Come on now boys! I\'m only going 55! (says Skittle) Dude! Thats bad we have more testosteron than them! (says me) 5.) UGH! UGH! Why are you asking me about books when I\'m stacking them?? ~ Brandon telling me about his gay co-worker 6.) The Chimpmunk version of Roll another Joint by Tom Petty 7.) You\'re pregnant??!! But-but HOW?? (says Drea) Gee I can only wonder (says me) 8.) I started screaming like a cheerleader about to get raped! ~ Violent J 9.) Mack has decided to NaPalm and give herpes to France because they won\'t help us in \"wartimes!\" ASSHOLES!! ~Skittle 10.) I have anal glycoma. That means my eyes can\'t see my ass coming to school on Friday~ Chris B. 11.) I\'m eating porn and whacking off to macaroni. ~ Chris P. 12.) You are special. We all ride the short bus together. That\'s how increadiably special we are. ~ Sammy 13.) SQUIRRL BAT!! HOLY SHIT!!!!!~ me   '),(2730,2224636,'05,April,2004','2004-04-05 00:00:00','    Whenever I bang my head on the wall the screaming seems to stop   I\'m tired. I\'m sick and tired of it all. I give up. I\'m done. There\'s no use in trying anymore. I\'m sick of the blank stares. I\'m sick of the no responses. I\'m tired of pouring my heart out and no one really understands or listens. I would just like to be \"normal.\" I would like to be an emotionally balanced person. But no everything has to be so damn difficult. Things need to change. But they won\'t. No matter how hard I try. You called me the strongest person you ever knew tonight. If I was so strong I wouldn\'t be giving up like I am. YOu don\'t understand. You never will. YOu don\'t listen. You don\'t hear me. I could scream until I\'m blue in the face and I assume you still wouldn\'t care. So I\'m done. I give. I\'m never gonna be able to let you go. But I guess I\'ll try.   '),(2731,2224636,'04,April,2004','2004-04-04 00:00:00','    Behind these eyes my love burns   Love Song ~ 311    Whenever I\'m alone with you You make me feel like I am home again Whenever I\'m alone with you You make me feel like I am whole again  Whenever I\'m alone with you You make me feel like I am young again Whenever I\'m alone with you You make me feel like I am fun again  However far away, I will always love you However long I stay, I will always love you Whatever words I say, I will always love you I will always love you  Whenever I\'m alone with you You make me feel like I am free again Whenever I\'m alone with you You make me feel like I am clean again  However far away, I will always love you However long I stay, I will always love you Whatever words I say, I will always love you I will always love you   '),(2732,2224636,'03,April,2004','2004-04-03 00:00:00','    If I had enough courage I\'d give this letter to you....    To my everything,       I need to be honest with you and confess everything. I\'m ready to lay it all out on the table. I have  never  loved anyone. Until I met you. Never. That night I met you, I instantly fell in love with you. I acted like a little school girl, giggling and such, and yet you still asked to call me.        That summer morning you showed up at my house knocking on my door. I couldn\'t believe it but I was so damn happy. I couldn\'t believe the joy I felt. When we kissed for the first time, I thought my heart was going to run screaming out of my chest I was so damn happy. I wanted to cry tears of joy and excitment of finding a person like you.      When we began our relationship I felt like nothing could go wrong. And for a while it didn\'t. After a little over a year and half had past we began to argue more and more. I could never figure out why. Maybe it was because we were always with each other.      Then that ever fateful day. You left me. My heart was ripped in two. I swore to myself I would hate you for the rest of my life and never be with you again for how you ended it. You always said that if you were to leave me you\'d do it in person. But you were to stoned to deal with it personally so instead you called. That hurt. To this day it still does. I was a mess for those two months. Crying, sick, depressed. I loved you so much it hurt.      That one night in May when we began to talk again. When you started to act like I existed once again. My promise to my self waved itself good bye. That night we decided to get back together I cried. And you sounded so damn happy as well.      We made the mistake of starting our physical relationship again with out the protection needed. Two months later we found out I was pregnant. I wondered what we should do. You told me whatever would make me happy. I couldn\'t tell my parents. I didn\'t want to disappoint them. I was always labled the \"good child\". The one that never did anything dramatically wrong accept miss curfew once in a while. I wanted to get an abortion and pretend it never happened. But that didn\'t happen.       I went to the doctors the day you decided to go out of town. That\'s when hell broke loose and my parents found out. During the pregnancy we argued so much. Whether it was my hormons going nuts or it was the stress from my parents not knowing to the problems you had at home and needing a job. I don\'t know. Two weeks later you broke it off again. But you stayed around. Being the best damn friend you could have ever been.       You always called to check on me when you couldn\'t see me just to make sure I was feeling OK. You stopped by to see me. Just because.        That day I went to the hospital and called you because they had to enduce me, you came at a record pace. Staying with me the entire time. Holding my hand through out the night of waiting to deliever our son. Then when the time came to push you stayed right there telling me it was going to alright. That love you have for me shone brighter than it ever had before.         When I saw you hold our son for the first time I wanted to burst out in tears because I knew we would never be a \"real\" family so to speak. If only we had been a little older. Those four days I had to stay in the hospital you stayed with me. Never leaving my side except to get food or smoke a cigarrett. It means so much to me that you did all that.        The day I was released you came back with me to my house and just stayed there to make sure I would adjust. You  held me while I cried because I wanted our son back so much. I still do. But I\'m OK with it.     Lately you and I have talked about \"us\" a lot. Tonight I asked you \"Would you ever consider getting back with me?\" You told me yes. So what are you waiting for? For us to be older so we can be married? Are you scared things will turn our like they did before? Have you honestly not noticed that you and I have changed? Me, becoming more laid back, funny, relaxed person. You all the sudden have become a whole lot more sweet and sensitive. Making me laugh. Always willing to hang out. You make time for me when you know you shouldn\'t. When you have more important things to do.     So why is it so hard for you to accept that we should just be together. Everyone believes that we should. My parents even though they don\'t like it have told me that we\'d end up being together for the rest of our lives. So why can\'t you just put your hand back in my and live the rest of your days with me? That\'s all I want. All I need. I\'m ready but I have to wait on you.      Granted that may not happen for a while. I may be with someone else. But the day you want to talk about getting back together. As horrible as it may sound I\'d drop the other person in a damn heart beat. Because damn it that\'s how much I love you.       I just want to be with you and only you. You\'re the one for me. My soul mate. My love, my life. My everything.  I love you always and forever, ~Me    Song: Love song by 311   '),(2733,2224636,'03,April,2004','2004-04-03 00:00:00','    OOO I\'M A SEKZI VAMPIRE! YESH!   Well well well, the old fashioned 17th century vampire, one of my faves. You look for the good things in life, you posses a lot of classical class, and follow that of the original vampires, you have no shame in what you are, infact you embrace it, you love it and wouldn\'t have it any other way. Your wealth is unspeakable and your way of luring people with your mystical ways and looks is amazing, and most people would often call you The Seductress. Please rate this quiz!     urlLink What Kind Of Vampire Would You Be? (Cool answers, AND FOR GILRS UNLESS YOU ARE A CROSSDRESSER)!     brought to you by  urlLink Quizilla   '),(2734,2224636,'02,April,2004','2004-04-02 00:00:00','    THEY COULDN\'T GIVE ME SOMETHING BETTER???    Suicide! (and you know it, so... dont u have something to do?)    urlLink   Choose your Dramatic Death (Now w/pics!!)     brought to you by  urlLink Quizilla   '),(2735,2224636,'02,April,2004','2004-04-02 00:00:00','    I\'M NOT EVIL I SWEAR!!     You have a malicious soul! Malice, when defined means, The urge to see the sufferings of others But you dont really mean any harm to anyone, you just like to watch it. Mischievous, cruel, and seducing, you can hurt others with cruel words or just be plain mean. Prank calls, tripping people, and breaking the laws are your favorite past times. You can lie just as easily as laugh, even to a close one. People are intimidated by you because you never let people get close to you. Youre cold, ruthless, and pessimistic behavior drives people away, which show the pain you feel inside.    urlLink   What Kind of SOUL do you posses? (For Girls only) Incredible Anime Pictures!     brought to you by  urlLink Quizilla   '),(2736,2224636,'02,April,2004','2004-04-02 00:00:00','    To help you understand how I feel about you     Willie Nelson - Always On My Mind   Maybe I didn\'t love you Quite as often as I could have And maybe I didn\'t treat you Quite as good as I should have  If I made you feel second best Girl, I\'m sorry I was blind But you were always on my mind You were always on my mind  Maybe I didn\'t hold you All those lonely, lonely times  And I guess I never told you I\'m so happy that you\'re mine  Little things I should have said and done I just never took the time You were always on my mind You were always on my mind  Tell me Tell me that your sweet love hasn\'t died Give me Give me one more chance to keep you satisfied I\'ll keep you satisfied   '),(2737,2224636,'02,April,2004','2004-04-02 00:00:00','    This is the one of the most beautiful songs I have ever heard.      Finger 11 - One Thing   Verse 1: Restless tonight,  \'Cause I wasted the light. Between both these times, I drew a really thin line. It\'s nothing I planned, And nothing I can. But you should be in line, Across that line.  Chorus: If I traded it all, If I gave it all away for one thing. Just for one thing. If I sorted it out, If I knew all about this one thing, Wouldn\'t that be something?  Verse 2: I promise I might, Not walk on by. Maybe next time, But not this time. Even though I know, I don\'t wanna know. Yeah I guess I know, I just hate how it sounds.  Chorus: If I traded it all, If I gave it all away for one thing. Just for one thing. If I sorted it out, If I knew all about this one thing, Wouldn\'t that be something? (2x)  Even though I know, I don\'t wanna know. Yeah I guess I know, I just hate how it sounds. (2x)  Chorus: If I traded it all, If I gave it all away for one thing. Just for one thing. If I sorted it out, If I knew all about this one thing, Wouldn\'t that be something? (3x)   '),(2738,2224636,'02,April,2004','2004-04-02 00:00:00','    OMG! You just don\'t get it do you!?   I\'ve become so frustrated with you now. It\'s making my eye twitch. All I want more than anything in this world right now, is you! Can you not see that? How can you not? I know it\'s obvious! I mean come on!  You act like you want the same! I\'m putting my self through hell. The reason being is my stupidity of being a damn fighter. When I want something that means the absolute world to me, I go after it. I go after it until I can\'t anymore. And damn it all to hell that\'s what I\'m doing with you. I love you too damn much. I love you too damn much to ever,  ever  let you go. I just couldn\'t. That would kill me on the inside. It would break my heart completely. Not that it already isn\'t.   You just don\'t understand. The times that we spend together. I feel complete. That hole inside of me is filled. But no! No you just reject me and sometimes it\'s like I\'m just your damn toy. Even though that\'s not true.   You sat there and told me to my face and I quote \"I love you. I always have. I always will. That will never change.\" THEN WHY IN THE FUCK CAN\'T WE BE TOGETHER!!! HOW IN THE HELL DO YOU LOVE SOMEBODY AND  NOT WANT  TO BE WITH THEM?? That makes NO sense to me whatsoever! I could never put anyone through what you\'re doing to me! I mean I know you don\'t realize what you\'re doing and maybe someday you\'ll understand.  But damn it. There is so much we\'ve gone through together. Too much for people our age. I know all that you have to deal with is hard. I know that our relationship took a bad turn. We hit a really bad slump. It all happened out of stupidity.   But damn it, every damn time it begins to get to the point where our relationship begins to be tested YOU RUN AWAY AND GIVE UP! I can\'t. I can\'t run. I can\'t give up. That\'s not me! It never will be. When things get tough I fight to make them right. It\'s too bad that you don\'t.   So answer this for me if you only want to be friends then why do you: 1.) Kiss me like you fucking mean it? 2.) Hold me for no reason at all? 3.) Tell me you love me? 4.) Forever and a day want to hang out? 5.) Put your arm around me or hold my hand when we\'re out in public!? 6.) Make the first moves that lead in to something? and all the other things. That would be a long ass list if I wrote it all.   You are my heart. I love you and you know that. I always will. But this is becoming to painful to stand. I try to sit there and act like friends but you put out the energy and the vibes that you want  so much more.    What in the flying hell am I supposed to do? How in the fuck am I supposed to feel??  I just don\'t understand you anymore and your actions. And I believe you just don\'t get it either. I think you honestly don\'t know what you want anymore. Whether its being with me or not. Whether you want to stay in this town or not. Whether its being happy or not. I don\'t think you know what to do with yourself anymore.   I\'m tired of you running away. Stand up and fucking speak boy. Answer the questions I\'ve asked you so many times before. Stop confusing me. Show me what the fuck you want to be. Whether it be a relationship or friendship. I need to know. I need to know soon because it is DRIVING ME INSANE!!!   Mood: Just lost, hurt, pissed Song: .45 by Shinedown   '),(2739,2224636,'14,May,2004','2004-05-14 00:00:00','    Wanna read my thoughts and feelings? Want to add your beautiful comments? Then go  urlLink here  =D   '),(2740,2224636,'13,May,2004','2004-05-13 00:00:00','    Hey you guyz.....  Well since some people have asked me to add a comment thingy on here and I can\'t figure out how I\'ve gotten a blog type thing on xanga.com so I\'ll put on a link so you can get there. I\'d love to hear from you all =D.... oh for those of you who don\'t know I no longer have the klumsykhaosxxx s/n it is now GoTpEaChEzXxX@aol.com okie doke   '),(2741,2224636,'13,May,2004','2004-05-13 00:00:00','    KIDNAPP-NATION!!!!!!!!!!! say that 10 times fast!!!!     Man I swear I feel like \"Where\'s Waldo?\" except its \"Where IN THE FLYING FUCK is Peachez??\" Man my mom is sooooooooo slooooooooooow. I emphizes her slowness to the highest degree. She called me   AT THE HOUSE   and asked me where I was?? LMFAO. She really really needs to stop doing them drugs man. OMG!! I told her where I was and she was like \"Oh I thought I called the cell!\" \"Ummmmmmm no mom you have the car, therefore I wouldn\'t have the cell on me!\" DOOOOOOOOOOOOOOOOYYYYY!  Smurfykins called me and told me he was coming to kidnapp me. I was like YAYNESS!!! Then he was totally insensitive to something that I had felt bad about. Jerk off! Aw but I love \'em. I hate him and love him. I need something to drink really bad. ::runs and gets a drink::  Ah coke, popcorn, and the Real McKinzes! A really nice combo! But dammit Price you\'ve ruined popcorn for me completely with your nastyness lmfao. But that\'s ok its still good (;D)  ::sings~ and you\'re walking hand in hand and you\'re wondering what tha fuuuuuuuuuuuccccck...it\'s then you realize you\'re shit outta luck!::      Mood: happy ladifreakinfuckindaaaaaaaa Song:Shit outta luck~ R.M.   '),(2742,2224636,'12,May,2004','2004-05-12 00:00:00','    I had a dream last nigh     I had the utmost disturbing dream last night. There are 4 rooms. One of them is in the middle of 3 of them. One room sits Jerry playing his music, another is Skittle talking on the computer, and the other room is my dad. I\'m stuck in the middle room with a guy who is just beating the shit out of me. Like smashing my head into the wall, pounding me with his fists, kicking me, all that shit. As he\'s doing this I keep calling out everybody\'s names but no one can hear me. I guess then the guy gets his fill of beating on me and walks out of the room. Some how (b/c I look absolutly fucked up blood and everything everywhere) I get up and drag my self to the first room. The one where Skittle is. I ask her \"Why didn\'t you help me?\" and she sits there and just stares in horror. I drag myself to the next room where Jerry is and ask him the same question and he just sits there and laughs at me saying \"I told you so you stupid bitch.\" Finally I drag myself to the room where my dad is. I once again ask the question and just shakes his head and turns his back on me. I drag myself back into the room where I first started off and collapsed on the floor.   Damn, when I woke up from this dream I just cried. I couldn\'t help but to feel totally abandoned. But there\'s no point. Those are 3 people in my life who I\'m pretty fucking sure would never just up and leave me like they did in the dream. Right?      Mood: Scared Song: nothing   '),(2743,2224636,'12,May,2004','2004-05-12 00:00:00','    SWING BATTA BATTA SWWWWWWWWWIIIIIINNNNNNGGGGAAAAAAAA!!!!      Yeah so my day. Hmmmm where to begin? Well, I wasn\'t feeling well this morning as I wasn\'t yesterday. OOOOO CORRECT GRAMMER!!!!!!! LMAO!!!! So I suffered through E-Commerce this morning and got my work done. Thankfully we\'re doing partner work so I she told me to just make suggestions whenever the teacher walked by but I didn\'t have to do anything. I love starved-for-attention girls! They\'re great.  So, I got to Geometry this morning and I was just feeling completely ill. I really wanted to go home. But this chick can\'t miss anymore days. I got into class and immediatly put my head down. My teacher told me to sit up. I told her \"No I\'m not feeling good at all.\" So then she started yelling at me saying \"I don\'t care how you feel you need to put your head up and pay attention.\" So then I yelled back \"I already know this! Second year here! Unless you\'ve forgotten!\" So I put my head back. She then yelled \"I\'m calling security to remove you!\" I told her \"Go ahead. But I know as well as you do that they are going to come to the room, ask who needs to be removed, look at you and laugh, then leave! It happens every time you try to get me remove. So sure go ahead and make a fool of yourself!\"  It\'s insanly stupid as she hasn\'t caught on that the security aren\'t going to remove me! They never do! Awww the advantages of knowing the principal personally! LOL! And ya know what?? She never called security.  Then I went to science and that was funny. My science teacher Mr. Hess (the absolute most slammin teacher ever, I mean he sits there and lets me bitch and we talk about drugs and drinking and we bash on the kids in the class together LMFAO)he put me and two other girls in a seperate class (Oceanography) while the other kids are in the Active Physics part. We had a sub and he didn\'t leave any work for me or the other two girls so we basically didn\'t have to do anything. The sub really didn\'t understand that so she kept yelling at me and Mya and we just sat there looking at her and smiling.  Got to lunch and got molested! LOL! Awwww if you\'re names not Skittle you don\'t know and don\'t need to lmao.  Then went and was a hard core swinger and yeah my day progressivly got better.   ::Peachez steps up to the plate.....she looks hesitant....she swings a couple of times....the pitcher throws a curve (SWING BATTA BATTA SWIIIIIIIINNNNNNNGGGGGAAAAAAAA the crowd screams)......she connects.......going....going....GOING....GOOOOONNNNNNNEEEEE!!...take that victory lap Peachez::      Mood: Nice but it could  be better if someone could take a hint of the title (meaning STEP UP TO THE PLATE!!!) Song: Let it burn~ Usher   '),(2744,2224636,'11,May,2004','2004-05-11 00:00:00','    My throat is raw from screaming and never being heard     Last night was just hmmmm. Jerry called me and apoligized for not showing up. Even though he said he was sorry I still gave him a piece of my mind and told him that he could have called a helluva lot sooner and I could have gone and chilled with Skittle. But I swear we talked until like 10:30 last night about everything. Everything was just put out in the open. All the secrets and lies. It was refreshing even though there were still some things I didn\'t want to know. He asked me when I was going to start dating again. I was like \"Ugh what a question to ask.\" I told him there is someone I\'m interested in and trying to get things to move but eh I dunno. I asked him why he hadn\'t begun dating again and he wouldn\'t answer me. What am I supposed interpret from that?? Hell, I dunno.  OK so on to other things. Price, dude, c\'mooooooooooooonnnnnnnnnnnnnnnnnnnnnnn!  I\'ve asked you to hang out sometime and you\'re all like yeah sure. Then nothing happens because you know you and Skittle went out for like two days. Then you know you popped up one day online and asked if I still wanted to hang out sometime. I was like most def. Hmmmmmm still NOTHINGS HAPPENED!!!!! Is it that hard to just up and go \"Hey lets go this place or lets do this...\" Ha! Maybe I should just do that.     Mood:Eh sick Song:iono   '),(2745,2224636,'09,May,2004','2004-05-09 00:00:00','    From sad and depressed to happy and hyper!     OK...so today was semi-depressing. I mean you know it\'s mother\'s day and I can\'t really celebrate it like most mother\'s out there but I still can in a sense right? I can be happy that my son has a wonderful home, loving parents, and everything and anything that he needs. So that\'s how I\'ll celebrate Mother\'s Day. I\'ll be thankful for everything good my son has.   It\'s kind of amazing to me how my friends influence my emotions. I mean all day I wanted some one to just talk to besides my parents and my brother. I NEEDED MY HOMIEZ!! I was just so ugh and blah starting off today. It wasn\'t cool because you know \"I woke up on the wrong side of the bed.\"   And of course I cried a little bit because of the thoughts in my head and how I was feeling. Talked to my mom about and that made me feel better to just get it out.  I was wanting to talk to Jerry because you know he\'s a really good person to talk to. Really because he knows how I feel about the situation with Joshua better than anyone. I\'m so happy he\'s still here for me. And claims me as his best friend. I think I\'m really close to being settled with it all. I honestly do. But you know by this time in like two days I\'ll be saying \"But I still love him and want to be with him.\" How silly. Don\'t you agree?   Now I\'m online talking to all my friends. Me and Skittle drooling together over the R.M. band leader together. LMAO Getting \"raped\" by Chris =D I don\'t mind that though. LOL Just laughing and I absolutly love it.      Mood: Sad to Happy Song: Smoke on the Water~ Deep Purple   '),(2746,2224636,'08,May,2004','2004-05-08 00:00:00','    Easier to Run~ Linkin Park          It\'s easier to run Replacing this pain with something numb It\'s so much easier to go Than face all this pain here all alone Something has been taken from deep inside of me The secret I\'ve kept locked away no one can ever see Wounds so deep they never show they never go away Like moving pictures in my head for years and years they\'ve played (If I could change I would take back the pain I would) (Retrace every wrong move that I made I would) (If I could stand up and take the blame I would) (If I could take all the shame to the grave I would) (If I could change I would take back the pain I would) (Retrace every wrong move that I made I would) (If I could stand up and take the blame I would) (I would take all my shame to the grave) [Chorus] It\'s easier to run Replacing this pain with something numb It\'s so much easier to go Than face all this pain here all alone Sometimes I remember the darkness of my past Bringing back these memories I wish I didn\'t have Sometimes I think of letting go and never looking back And never moving forward so there\'d never be a past (If I could change I would take back the pain I would) (Retrace every wrong move that I made I would) (If I could stand up and take the blame I would) (If I could take all the shame to the grave I would) (If I could change I would take back the pain I would) (Retrace every wrong move that I made I would) (If I could stand up and take the blame I would) (I would take all my shame to the grave) Just washing it aside All of the helplessness inside Pretending I don\'t feel misplaced It\'s so much simpler than change [Chorus] It\'s easier to run Replacing this pain with something numb It\'s so much easier to go Than face all this pain here all alone It\'s easier to run (If I could change I would take back the pain I would) (Retrace every wrong move that I made) It\'s easier to go (If I could change I would take back the pain I would) (Retrace every wrong move that I made I would) (If I could stand up and take the blame I would) (I would take all my shame to the grave)   '),(2747,2224636,'08,May,2004','2004-05-08 00:00:00','    Sign me, Kiss me, I can now die!     Oh tonight! Tonight was just....fabulous. Absolutly fabulous. Hott guyz every which way you turn. Stupid little bitches you wish you could just punch. Fucking great music on stage. Thankfully nobody got fucked up tonight except one stupid bitch who apparently got assaulted by a guy and they had to call the cops. GAYNESS!!  If you can\'t guess, I went with Skittle and Alya (sp?) to Tremont to see another kick ass Punk show. Forgotten Youth, Minority Party, The Breifs, and The Real McKinzies (sp)!!   Man I was hopin to get myself up to it and mosh tonight but that didn\'t happen. Reason being, the guys in the pit were fuckin punchin and kicking. I\'m sorry I\'m not getting hit in the face for nobody. I would have jumped but another reason there were too many damn drunks. This one guy was so bombed he could stand straight without another persons help. Even then he couldn\'t stand straight. I was kind of sad.  OMG THE R.M KICKED SO MUCH ASS. The were the absolute best band of the entire night. I screamed my self horse through their show. They have the scottish accents, the KILTS MAN THE FUCKING KILTS!!, and bagpipes. THEY SO ROCKED!! I mean, people were clapping, jumping, singing along. That is a band that makes you wanna clap, move, jump, head bang, pump your fist in the air. After the show Alya bought one of their CD\'s and she wanted to get it signed so I went up to the band leader and asked him to. He started off with \"XOX\" and asked me if I knew what it meant and I said \"Hugs and kisses?\" so then he hugged me and kissed me I was like \"YESH!!\" Then I asked him to sign my pants and he did. He said to me \"It\'s cryptic.\" But I think I figured out what it says. He wrote \"XOX P X Mc X\" So maybe it says something along the lines of \"Kiss Hug Kiss Punk Kiss McKinzes Kiss?\" Maybe? I dunno and I don\'t care HE SIGNED MY PANTS AND THAT ROCKS SO MUCH ASS!!! Then he signed Skittles patch on her boob  and he said \"You like that don\'tcha??\" IT WAS GREAT.  Then on the ride home we listened to the CD. Fabulousness!!      Mood: Elated Song: Anything my the R.M.   '),(2748,2224636,'08,May,2004','2004-05-08 00:00:00','    Oh the 9th!      I\'m gonna be kind of upset on the 9th. It\'s going to suck. Try and figure out why! I\'m just gonna stay in bed alllllllllllll day unless somebody comes and kidnapps me.  Me and my brother:::::::  Brandon: ROCK!!!!!! AHHHHHHHHHHHHHHHHHHHHHHHHAHHHHHHHHHHH (singing) Me: OK go eat your fried poor ham now! (he goes and plays pointless notes on the piano) Brandon: NO I\'M GOING TO STAND HERE AND PLAY POINTLESS NOTES AND ROCK OUT!!!!! Me: Are you sure you\'re turning 21 this year? Brandon: NO! ROCK! AHHHHHHHHHHHHHHHHHHHHHHHHHHAAAAAAAHHHHHH! Me: OMG! Go away you tard bucket. (Brandon skips from the room)  Sorry just had to put that in cuz it was funny.  Yesterday was nice. To put it simply. I had absolutly nothing to do yesterday cuz Skittle went to City Fest (I hope you had bunches of fun homie!!). I started to clean and usually through my neighborhood people have loud bass systems and you can here them. So as I was about to start vaccumming I heard this loud ::booooooom booooooooom:: I was like \"Nu-uh\" cuz I thought that Jerry was going to be busy alllllll week. And then a blue neon speeds down my driveway and lo and behold who walks through the door??? JERRY-KINS!! I asked him what he was doing here and he told me he came to kidnapp me. I told him he had to wait until I cleaned up first and the he was sweet enough to help me. Well he takes me to his house where me him and another kid named Brandon (GOOMBA MAN I SWEAR TO GOD!! SKITTLE YOU KNOW!! LMFAO) played Mario Party 3. Those two kept laughing at me because the other princess (I was Peach) Daisy stole my star and my money. Then Jerry\'s mom comes in and tossed two packs of cigarretts in my lap and said \"You\'re mom came by!\" I was like WOO HOOO maybe these will last through Tremont man!! I gotta find out if Skittle and I are going.  I hope I hope! But last night was just nice. Just to chill with Jerry and have nothing happen was just....nice. And for once I didn\'t cry because we weren\'t together. And for once I didn\'t cry because he said he loved me and had it only mean \"as friends.\" I think I\'m doing well. Don\'t you?? LOL ::dances around::      Mood: Settled Song: Faint~ Linkin Park   '),(2749,2224636,'06,May,2004','2004-05-06 00:00:00','    You\'re making me miss you too damn much and I hate you for it!     Why? Why did you have to say that? Why did you tell me you love me? Don\'t you know it hurts to hear you say that? It pains me because we can\'t be together. I mean I know you\'re over \"us\" and I\'m getting to the place where I\'m over \"us\" as well. But I still can\'t help but to think there still might be something there! And that sucks. I miss you sweetheart. I just want to hang out and chill with you. But all the sudden you\'re so busy. I mean I can understand that you need to make money so you don\'t have to move in with your grandparents. I wouldn\'t want to either. Especially when mine are crazy. But God, it sucks that you can\'t even call me back when you say you will. I mean I know you have a life but I thought I was apart of it as well. Aren\'t I? I mean you and I are best of friends. So you say...... It\'s ugh! You\'re confusing me because you start to act all sweet again. You started doing that the last time you wanted to get back togehter. And NO NO you cannot do that to me. Not when I\'m really starting to accept the fact that I know you and I will never be together again. You are placing false hope with in me. Why why must this happen??  YOU CAN\'T DO THIS TO ME!!! I WON\'T LET YOU!!!! I deserve so much better. I know I do! I can\'t keep wanting to run back to you! I can\'t allow myself to have that pain of me chasing you and you rejecting me. I can\'t have that. I am sick to death of crying over what can\'t be!!! I\'m done wanting you! Haven\'t you realized that? We\'re not meant to be right now. Things are too damn tough for us to have a relationship?? So stop! Take a step back. Be my friend....be my fuck buddy......BUT DO NOT TRY AND PLAY YOUR MIND GAMES!!!! I learned from you. I\'m already a step ahead. Do NOT think you can sit there and actually think you can make me think that you love me and want to be with me again. I know it\'s not true. You\'ve said it too many times. I know you love me but you love me because I am the mother of your son and I am your friend. DON\'T TRY ME BOY!!!! I AM NOT PLAYING YOUR GAMES!!!!      Mood: Pissed and hurt Song: No more love~ Shinedown   '),(2750,2224636,'05,May,2004','2004-05-05 00:00:00','    I jumped and then I accidentally flew away     Well on Monday I heard that Jump Little Children was going to be at Manifest. So I talked to Hayley almost immedietly and she asked me if I wanted to go and my response you ask?   HELL YUHHHHHHH   So today we ran errends and then we went up there and I saw the one guy Matt. I think I almost drooled on myself. Then they played a couple of their new songs and I was like \"OH I LOVE YOU GUYS!!!!! I LOVE YOUR MUSIC!!!!!! I LOVE YOUR VOICES!!!!! I JUST OH LOVE YOU!!!!\" But I didn\'t say it. (::latin accent::)~~>But I said it with my.....eyes!! Im a crack monkey I swear. Well after the show they stared signing autographs and I wanted to get something signed so me and hayley got our progress reports signed. Well I got a piece of paper signed. ::plasters paper against computer screen:: SEEEEEEEEEEEEEEEE!!!!!??? It was hilarious what they said on Hayleys. But yeah I\'m in a good mood.     Mood: Happy happy jumpy happy Song: My Guitar~ JLC   '),(2751,2224636,'04,May,2004','2004-05-04 00:00:00','    Ah a song that describes me...wow     Better Version Shinedown     Excuse the mess, I didn\'t see you from behind I caught a glimpse, but the reflection\'s only mine It\'s almost like I\'m paralyzed and locked outside myself What I don\'t need is to concede because I won\'t be someone else I am not perfect and I don\'t claim to be And if that\'s what you wanted Well then I\'m so sorry  How about a better version of, the way that I am  How about a better version that, makes me understand How about a better version of, the way that I am The way I look, The way I speak, How about a better version of me  Excuse the wall, I put it up from time to time A silver shade, and the design is all mine It\'s just a maze that everyday I seem to be stuck in It never seems to fade away but I pray for the day it ends  I am not perfect and I don\'t claim to be And if that\'s what you wanted Well then I\'m so sorry  [CHORUS]  Show me my vital signs until I\'m realigned  [CHORUS]   '),(2752,2224636,'03,May,2004','2004-05-03 00:00:00','    AH! WHY! WHY DO I DO THIS TO MYSELF!!      Sometimes I seriously make myself sick with the way I feel. I went to Jerry\'s to show him the recent pictures of Joshua because you know he wants to see them as well as Brenda and all that.   So OK, I knew he was sick and wasn\'t gonna give me a big ol\' hug and kiss like he usually does but damn I walked in his room and didn\'t even get a fucking HI!! That hurt a little bit. I mean he could have been at least poliet and said HI! I say hi to people even when I don\'t feel good.   Although he gave me a small hug and smiled at me. So I gave him the pictures and tried that stupid small talk but that didn\'t work. It\'s just like he\'s only nice to me when he wants to fuck and UGH. I mean that\'s just how it feels. It makes me feel like a cheap whore and it makes me want to cry. Seriously. I hate that.  I hate all these damn feelings and I hate having to feel them. I mean I\'m not used to them all because I always got drugged up so that I couldn\'t feel. I don\'t want to feel but I can\'t do pills anymore because Hayley would kick my ass. But I want to. I just wanna be in a stoned stupor. I don\'t want to deal with them.  I don\'t want to have to deal with the fact that Mother\'s day is coming up and I have nothing to show for it. God that makes me feel like shit. I can\'t take it! I want to just give up. I don\'t feel like faking anymore. But I don\'t want to be real either. I don\'t want to be anything.  I\'m tired of Jerry acting the way he does. He needs to get his shit straight. I\'m tired of the feeling of hopelessness. Hayley keeps asking me what I want. What would make me feel better. Wanna know? Happiness. Pure true happiness. That\'s it. And to be satisfied with the person that I am. That\'s not much to ask for right??     Mood: Hurt Song: Left out ~ Shinedown   '),(2753,2224636,'03,May,2004','2004-05-03 00:00:00','    What a fucking blast!!!!     Prom was a fucking blast!! The music were fabulous. All my buddies were there. The guys were OH SO PRETTY!! The girls were simply beautiful. It was great. I danced for four hours straight. Me and my buddy Mary danced with each other and these two really hott guys. And it\'s always nice to get in between two guys and dance with them. Kinda whorey like but special like!! LOL. The guys at my school can actually dance I was really surprised. And it was really nice that guys were asking me to dance. Because YES this white girl has rhythem (omg i forgot how to spell cuz i know that\'s not right lol). The only thing that got on my nerves really was this guy who basically was interested in me for the fuck last year was there. GET THIS: hasn\'t noticed me alllllllllll year. I\'ve said hi to him just cuz, looked at me like a worm, saw me at the prom SMILED acted liked i existed and asked to dance. I did and then he was like lets go sit and talk I was like how bout no! I hated that how all the sudden guys act like you exist when they haven\'t acted like you\'re alive alllll year! How stupid. And it was really funny at one point me, mary, candy, and kari were all dancing in a line together and like 30 other girls joined us. Flashes were going off everywhere!! Oh wow I\'m scatter-brained when I walked in the doors with my date i heard \"OMG ITS EMILY\" ::flashes go off eveeeerrrrrrryyyywhhhere!!:: Me and my date Brandon were blind for a second.  After the prom me, Brandon, Torie, Scott, JC, Mary, and Lorenzo planned to meet at McD\'s and then go bowling. Torie and her man NEVER showed up so we left them. We all pretty much figured they were making babies on the side of the road. So we all went to the bowling alley. Me and Mary kept dancing the entire time we were at the bowling alley. It was great. Then me, mary, and jc lined danced in the bowling lane! It was so funny. And I got a lap dance that I did NOT WANT!!!! LOL. My night was fabulous! But today (this is monday now) I had that STUPID STUPID POINTLESS GAY ASS AP ENGLISH EXAM TODAY! FOUR HOURS OF STRAIGHT RETARDEDNESS!! Me and Candy sat there trying not to talk the entire time. OMG the essays were straight stupid. The first one made no sense whatsoever. Why would I give a flying fuck about Lord Chesterfield and his.....damn....values. OH RIGHT I\'M NEVER ALLOWED TO DISCUSS IT. TARDNESS!! LOL THE ANSWERS WERE EEEEEEAAAAAABBBBBBCCCCCDDDDDDD yeah how bout no!      Mood: Absolutly fabulous Song: Toxic~ B.S. and it has been stuck in my head since prom   '),(2754,2224636,'01,May,2004','2004-05-01 00:00:00','    I feel the need to say I\'m sorry     So yeah I\'m in a fabulous mood today. I\'m going to prom and really looking forward to it. I mean my dress is just awesome. My hair is gonna look so damn good when I finally get it done lol. I\'ll feel all pretty. I\'m sure I\'m going to have a great time. I\'m going with my good friend Brandon. I was like so surprised when he asked me to go. It\'s OK that I was his second choice. But that\'s only because his first date wasn\'t able to go. But hey that\'s always a good thing. I wasn\'t going to go at first of course unless somebody asked me. Then BOOM I got asked lol. It was funny because I thought that he wasn\'t being serious I guess. So me and Brandon are going to go eat at Showmars instead of McDonalds because that was just a funny suggestion to go there because you get free food if you show up in your prom etire (sp??). Then we\'re going to make a \"fashionably late entrance\" because Brandon is so funny like that. After that we\'re going to meet his friends up at a bowling lane to go Midnight Bowling. That should be really funny all of us bowling in our prom stuff. Makes you wanna bring extra clothes man. But nooooo!! I can\'t wait to see everybody there. A lot of my senior friends of course will be at the prom. Man for 50 bucks this better be a kick ass prom fo real!! I mean the last three days the school sold tickets they bumped the price up to 65 bucks. Dude all the other schools were like 35 to 45 bucks. That\'s cheaper than ours. In the end though you end up spending over 200 bucks anyway because you want to look good lol. Oh well. It\'ll be fun.     Mood: Calm and happy (for once its real) Song: Bottem of a Bottle~ SES   '),(2755,2224636,'01,May,2004','2004-05-01 00:00:00','    I can\'t even try to spiff up my blog right now I\'m so just ugh and arg and ::scream:: OK see I\'ll love my best homie in the world Skittle with all my heart. All I want is her to be happy and if you look at the convo below I\'m sure you can figure out why I\'m feeling the way I am right now.  I am soooooo damn sick of trying to \"let things happen.\" I\'m tired of sitting idioly by and waiting on shit to happen. I\'ve been there done that sick of it. But when I try to make things happen they BLOW UP in my face. Nothing works. I\'m at a fucking loss. You know I\'m pretty damn sure now Chris and Skittle are going to start dating now. YAY for them. I mean that honestly. As long as they\'re happy that\'s all that matters to me. But I still can\'t help the feeling of frustration and hurt I feel. I was pretty sure I\'d have no chance with Chris at all. I knew that he was in all kinds of love with Skittle. And yet stupidly I tried to start something with him. Actually thinking something might happen between us. Wow how embarrassed am I?? Entirely. You know I\'m really sick of trying. I\'m sick of being fake. But I\'m sick of trying to be real. I\'m sick of it all. I\'m ready to give up. I wish that things would work out for me. But usually they don\'t. That\'s my damn life. Every time I try for something it doesn\'t work out like I would want it to.  It\'s really sad that I cried over stupid ass bullshit like this. It\'s sad that I knew from the beginning I had no chance with him and yet I still fucking tried. Why? Why do I do this to myself. Here I am trying to jump back in to the dating scene and fuck I screw that up. Wow. How sad. I wish things would work out for me. I wish things were easier. But that can\'t happen. Too many things have happend to make the me the way I am. Sometimes I honestly wish I couldn\'t feel a damn bone in my body. I don\'t want to \"feel\" any more. I\'m sick of \"feeling.\" Having to \"feel\" puts me in a state of unbalance. It\'s times like these I would go back to popping pills or drinking into a state of stupor. It\'s destructive yes I know. But seriously I don\'t fucking care anymore. I\'ve lost my first and only love. I\'ve lost my son. I suck at trying to date. I suck at flirting. I suck I suck I suck. Yeah I know if Skittle reads this she\'s going to kidnapp me and yell at me because she hates it when I do this to myself. She hates that I drag myself down like this and beat up on myself. But you know what this is me. This is how I am. A sad miserable fuck. A fake fuck at that. I\'m just so tired of trying. But I\'m tired of fighting all my demonous battles. I\'m tired of fighting with myself. I\'m tired of fighting for things I can\'t have. I\'m tired of trying for things I can\'t get. I\'m tired of opening my mouth and letting things spill out when I know that they shouldn\'t. I\'m tired of it all. Sometimes I really wonder why I\'m living. Like seriously. What\'s the point? No you guys I\'m not going to kill myself what would be the damn point. I don\'t want to end up hurting everyone else because I feel like I\'m suffering. I\'m really sick of crying because I highly believe I\'m going to be alone for the rest of my life. I\'m tired of that. I really truely am. I can\'t go on being fake but I can\'t go on being real. So what am I supposed to do?? I need help. I need to get all this out. But I can\'t seem to do it. I know this is turning into insessenct rambling but I can\'t help it. I\'m breaking down and I can\'t figure out why! I think all the events that have happened to me in the past months are finally tearing me apart. Killing me inside. I\'m dead on the inside. I\'d like to feel without regret. I\'d like to say things and not feel like stupid later. I\'d like to be happy. I\'d like to smile and seriously mean it. I\'d like to feel that not everyone is judgiing me. I\'d like to not feel so damn paranoid. I wish people would stop talking behind my back and start saying things to my face. I need to get away. Just leave for a while. Get my thoughts and myself together. That\'s what I need. Just to run away. I don\'t know where I\'ll go or for how long but I can\'t stay here any more. Not for right now. I mean, I\'m not mad at anyone. No one has done anything to me to cause this. This is all  me . I\'m just breaking apart and I really can\'t take it. This is not supposed to happen to anyone because it SUCKS. Beyond all beliefe. I wish everything could be straigtened out with a snap of my fingers. I wish that I didn\'t feel the way I did. I\'m just tired. So very very tired. Sick of it all. I\'m done. I can\'t do it. I don\'t think I can feel any more. I honestly feel so dead on the inside. And I just don\'t care.   '),(2756,2224636,'01,May,2004','2004-05-01 00:00:00','    Being fake   CHAOticSkittle:  so in ALL HONESTY!!! how would you be hurt if me and price started goin out?   KlumsyKhaosXxX:  go out with him if you want to    CHAOticSkittle:  thats so not what I just asked you   KlumsyKhaosXxX:  as long as your happy    CHAOticSkittle:  (still not what I asked)   KlumsyKhaosXxX:   (not trying to sound mean) why would it matter to me if you guys went out?? ....i mean it would but it wouldn\'t.... it\'s not like i have a dying chance with the boy    CHAOticSkittle:  thats not the point.   KlumsyKhaosXxX:  what\'s the point hun?   CHAOticSkittle:  whatever   KlumsyKhaosXxX:  no please tell me    CHAOticSkittle:  just forget about it .   KlumsyKhaosXxX:  are you like mad at me now   CHAOticSkittle:  no   KlumsyKhaosXxX:  you sound like it but ok   CHAOticSkittle:  Im not its okay.   KlumsyKhaosXxX:  ok you want all honesty??   CHAOticSkittle:  yes   KlumsyKhaosXxX:  yeah i\'d be hurt but hey that\'s me.... but its not like i\'m going to go into some stupid ass emotional rampage over it... you guys like each other if you wanna go out with each other...as long as your happy i\'m happy for you.... like i said its not like i have a chance with him anyways.... i tried i fail ooo ahh go me lol....its whatever   CHAOticSkittle:  thank you.   KlumsyKhaosXxX:  and in all honesty its situations like this is the reason why i keep my mouth shut most of the time   KlumsyKhaosXxX:  but why do you ask??   CHAOticSkittle:  just wondering.   KlumsyKhaosXxX:  right lol   CHAOticSkittle:  shup   KlumsyKhaosXxX:  okie doke   CHAOticSkittle:  :-D   KlumsyKhaosXxX:  :-):-\\:-D   KlumsyKhaosXxX:  guess he isn\'t \"torn\" any more is he lol   CHAOticSkittle:  he is.   KlumsyKhaosXxX:  riiiiiiight   CHAOticSkittle:  no really..lol   CHAOticSkittle:  just I dunno anymroe   KlumsyKhaosXxX:  ::passes out on the floor::    CHAOticSkittle:  nooooo *throws COLD h2o on you*   KlumsyKhaosXxX:  arg   KlumsyKhaosXxX:  lol   CHAOticSkittle:  no passing out damnit.   CHAOticSkittle:  its BAAAD fer you   KlumsyKhaosXxX:  and why not lol   CHAOticSkittle:  ^^^^^^^^^^^^   KlumsyKhaosXxX:  no its good i can escape akward situations   KlumsyKhaosXxX:  *~*sigh*~*   CHAOticSkittle:  Im sorry.   KlumsyKhaosXxX:  for?   CHAOticSkittle:  i dunno   KlumsyKhaosXxX:  right   KlumsyKhaosXxX:  dobedobedooooo   KlumsyKhaosXxX:  haha this is awful we made a promise to each other to not get all \"omg\" over a guy we both like and look whats happening lol   CHAOticSkittle:  oopsie   CHAOticSkittle:  but what did we say we\'d do about it?..lol   KlumsyKhaosXxX:  oh well it always happens to me and the other chick usually comes out on top   CHAOticSkittle:  i didnt mean to   KlumsyKhaosXxX:  i don\'t really remember but i\'m not trying anything...... he\'s like all kinds of in love with you and me i\'m just some stupid little crush that really all i am...so yeah    CHAOticSkittle:  valley girl/psychiatrist voice &quot;a low self-image is sooo not what you need right now. Get back out there and just like be cuz thats like so great&quot;   CHAOticSkittle:  (I hear it in my head and Im just like)..LMFAO   CHAOticSkittle:  (sorry)   KlumsyKhaosXxX:  ::same voice:: that made like no sense but ok   KlumsyKhaosXxX:  lol   KlumsyKhaosXxX:  im just tired of trying and failing that\'s all   CHAOticSkittle:  yeah I know..it wasnt sposed to   KlumsyKhaosXxX:  oh well okie doke lol   KlumsyKhaosXxX:  jerry, pinz, and now chris WOW THIS IS SOME FUCKED UP PATTEREN FOR ME I SWEAR lmao   KlumsyKhaosXxX:  pattern*   CHAOticSkittle:  well who else do you like.   KlumsyKhaosXxX:  does it really matter?? in all honesty   KlumsyKhaosXxX:  its gonna be the same each time so whatever lol   CHAOticSkittle:  sure.. maybe we can work some &quot;hook-up magic&quot;   CHAOticSkittle:  Im not gonna let you think that way..   CHAOticSkittle:  your not allowed to at all   KlumsyKhaosXxX:  sorry but that\'s the way i feel   KlumsyKhaosXxX:  i just really regret ever saying anything lol oh this is so gay    KlumsyKhaosXxX:  i pretty much figured something along the lines of me \"not having a chance with him\" would happen in some form or another    CHAOticSkittle:  but you did and do have a chance with him.   CHAOticSkittle:  he has liked you, ligitimately liked you ..it wasnt just total bull.   KlumsyKhaosXxX:  in all honesty homes i don\'t think i do because if i did something would have happened by now   KlumsyKhaosXxX:  yeah \"liked\" not \"likes\" lol   CHAOticSkittle:  man you know the s is right next to the d on keybopards..lol   KlumsyKhaosXxX:  lol   KlumsyKhaosXxX:  SHUT UP STOP MAKING ME LAUGH   CHAOticSkittle:  hahhahaha   CHAOticSkittle:  cipotle!   KlumsyKhaosXxX:  ooo ahh   KlumsyKhaosXxX:  ::frustrated/sad like (and no not with you so don\'t start)::   KlumsyKhaosXxX:  but i\'ll just do my fake :-D and everything will be okie doke lol   CHAOticSkittle:  NO   CHAOticSkittle:  you cant hide behind that damn smile any more.   CHAOticSkittle:  I want to see a real smile on yer face man   KlumsyKhaosXxX:  ::Scared like::   KlumsyKhaosXxX:  i smile for real sometimes its not always fake   CHAOticSkittle:  oop..wasnt sposed to be skeery   KlumsyKhaosXxX:  just most of the time   CHAOticSkittle:  see exactly.   CHAOticSkittle:  what would make you happy?   KlumsyKhaosXxX:  ::shrug:: i dunno right now   KlumsyKhaosXxX:  jumping of a cliff in to the sea of embarrassment i\'ve put myself in   CHAOticSkittle:  why are you in a sea of embarrasment?   KlumsyKhaosXxX:  cuz i feel absolutly stupid hayley...like beyond beliefe   CHAOticSkittle:  because you like somebody and you told him?   KlumsyKhaosXxX:  yeah   CHAOticSkittle:  that shouldnt embarras you it has no grounds to. Its not like its been plastered all over school and everyone knows and is like hey look she likes chris kill her.. its no big deal.. he knows casey know and I know.. thats pretty much it was far as I know and thats not a bad thing.   KlumsyKhaosXxX:  thats not the reason i\'m embarrassed hun..... i just feel stupid for telling him and then ALL this happens that\'s why i feel stupid   CHAOticSkittle:  and a couple of us tried to help ya get him so what if that didnt work. somebody will work out and then itll all be alright   KlumsyKhaosXxX:  yeah   CHAOticSkittle:  can you like not say  &quot;hun&quot; it feels like uber-condescending   KlumsyKhaosXxX:  sorry my b   CHAOticSkittle:  its alright   KlumsyKhaosXxX:  i\'m just really tired of going after things i can\'t have thats all   CHAOticSkittle:  so am I   KlumsyKhaosXxX:  oh how so??   CHAOticSkittle:  lets see like soo many of the guys Ive gone after.   KlumsyKhaosXxX:  well most of them you\'ve gotten   CHAOticSkittle:  I end up going out with like 3rd string instead of what I set my goals to.   KlumsyKhaosXxX:  yeah well at least you \'date\' to say the least i try and ::smack:: nope doesn\'t happen lol   CHAOticSkittle:  I havent gotten what I actually wanted. I\'ve settled. and I\'ve lowered my standards like soo much its not even funny.   KlumsyKhaosXxX:  oh well stop   KlumsyKhaosXxX:  its been 5 months since jerry and i broke up and omg you can\'t understand how ready i am to move on so yeah i\'m like \"go date or at least try to\" and i can\'t really do that right either lol   KlumsyKhaosXxX:  its like \"damn god you hate me that much huh??\"   KlumsyKhaosXxX:  yes yes i know its only one shut down so far but that kinda effects me so hmm yeah   KlumsyKhaosXxX:  well actually two come to think of it but whatever   CHAOticSkittle:  okay well steve is the first right?   KlumsyKhaosXxX:  yup and i\'ve been shot down by him TWICE omg lol   CHAOticSkittle:  he was and prolly still is in love with you if what he says is true and that was only a shut-down b/c he was in a realtionship that he cared about.    CHAOticSkittle:  so its not a shut-down he shouldnt leave that to pursue something totally new to him.   KlumsyKhaosXxX:  yeah i know but still \"shot down\" lol and i don\'t know if what he says was actually true...things can be said and not really meant i guess   KlumsyKhaosXxX:  i\'m just really tired of trying lol   CHAOticSkittle:  so stop trying let something just happen   KlumsyKhaosXxX:  ok i\'ll be single for another 5 months then    KlumsyKhaosXxX:  lol   CHAOticSkittle:  rite   KlumsyKhaosXxX:  nvm you didn\'t get that lol   CHAOticSkittle:  i got it..   KlumsyKhaosXxX:  lol ok   KlumsyKhaosXxX:  wow that song makes a lot of sense now omg lmao dude ok sorry trying to put my mind some where else i guess i\'m trying to go back in to \"fake\" stage   CHAOticSkittle:  what song?   KlumsyKhaosXxX:  burning bright that one line that goes \"and as the darkness consumes me i pretend i\'m burning, burning bright\"   CHAOticSkittle:  o.   KlumsyKhaosXxX:  yeah that\'s me right there lol   '),(2757,2224636,'03,July,2004','2004-07-03 00:00:00','    ok damn it....... my links aren\'t working for some reason.... so if you want to read my life and everything like that just go to www.xanga.com and then find user and type in GoTpEaChEz.... if any one actually reads my stuff......if you can\'t get to it just IM me at GoTpEaChEzXxX (aol) and I\'ll send you the link okie doke.... *hugs and kisses*   '),(2758,3306195,'28,May,2004','2004-05-28 00:00:00','    There\'s been travel and there\'s been rain, so there hasn\'t been much going on in the garden. However, Marny put together some great container plantings for the deck. They really spruce things up and, once it stops raining, I\'ll take pictures of them and post them.   Within the next week, I have to set up the second vegetable bed. I already put the wood frame together, just have to clear out the grass from the site and fill it with dirt and stuff (i.e. cow poop).   '),(2759,3306195,'11,May,2004','2004-05-11 00:00:00','    This seemed like a good day to start this blog. As I write this we are having a shed built at the end of our driveway. Today is the last day I\'ll have to lug all of my garden tools out of the basement every time I want to do some yard work.  Over the last month, Marny and I cleared out the site. Then I had a large pile of dirt delivered so I could even out the site and tamp it down. I finished fine tuning it this morning and now the shed is going up!  Eventually, that area will also feature a new bench/garden arbor, gravel covering the ground and container plantings. But, all of that will have to wait a little while. Next up is the vegetable garden. I\'m doubling the size of last year\'s veggie garden and I need to start getting the new bed ready. My seedlings are starting to come up and will need to be replanted in a few weeks.   urlLink    the mess before...   urlLink    after clean up.   urlLink    the shed going up!   urlLink    All done! (Soon it will be painted white and blue, like the house)   '),(2760,3306195,'05,June,2004','2004-06-05 00:00:00','    Before I get into this new post, here\'s a picture of one of the beautiful containers Marny put together two weeks ago... We\'ll make a gardener out of her yet!   urlLink    nice, huh?  Since we moved in, there\'s been a big circle of sand in the backyard that used to house a pool. Early last summer I added a big pile (4 cubic feet) of loam to it, to be used for the vegetable garden. The intent was to then spread the rest of the loam out to fill in the sand circle, plant grass and have a nice lawn.   The pile of loam sat there until THIS summer, when we decided to add another vegetable bed and I had 4 MORE cubic feet of loam delivered.   Well, this past Sunday (May 30th), Marny and I put up the new veggie bed, spread out the loam and planted grass seed! The vegetable plants will go in tomorrow (weather permitting).  Here\'s a picture of the new and improved back yard...   urlLink    grass (and vegetables) coming soon   '),(2761,3306195,'09,July,2004','2004-07-09 00:00:00','    I\'m feeling a little better about my To-Do list. I\'ve got about 85% of it done (I obviously accomplished a lot since yesterday morning!), and still have all day Sunday to finish things up. Now, onto the things that have gotten done:  First of all, the shed is finished! I just put the finishing touches on it (marigolds in the window box and birdhouses hanging from each corner)   urlLink      Remember I mentioned that dilapidated storage bin/box that was falling apart and needed to be replaced? Here\'s a picture of what it used to look like:   urlLink      Yesterday I took a sledgehammer to it and three hours later it was broken up in bins for the trash. Unfortunately, they wouldn\'t take it! Regardless, I\'m thrilled that it\'s been taken down:   urlLink    note all the remnants that the trash men would not take away. I have to lug it all to the dump myself! :(    Do you also remember the circle of sand that became the circle with piles of dirt that became the spread out dirt with grass seed? Well... it\'s a lawn now!   urlLink    the new lawn  Though I\'ve still got a few things to do before my vacation is over, I\'m feeling very good about what I\'ve accomplished so far. Plus, I\'ve learned a valuable lesson.  Things rarely take as long as you think they will.  I kept putting off the demolition of the wooden box near the driveway, dreading it and not wanting to do it. Well, it turns out that 3 hours with a sledgehammer (great for any pent up aggression you might have!) was all it took. I now know that I\'m better off jumping right in and just doing it, rather than agonizing about it.   that\'s it for now. more to come early next week!   '),(2762,3306195,'08,July,2004','2004-07-08 00:00:00','    It\'s Thursday, my vacation is more than half over and I\'ve only gotten about a third of my To-Do list done.  uh-oh....  more to come!   '),(2763,3306195,'04,July,2004','2004-07-04 00:00:00','    Marny and I have been VERY busy in the yard and we\'ve accomplished a lot. I\'m on vacation this coming week, so even more will get done. We\'re putting the finishing touches on the shed tomorrow so we\'ll have pictures up soon.  I\'ve made up a to-do list for my vacation (don\'t get me wrong, I LOVE working in the yard, so spending a whole week doing that  is  a vacation for me). Here\'s the list of what I hope to accomplish:  -Finish painting shed/re-attach shutters and window box  -Demolish wooden \"box\" near driveway (a ramshackle storage bin that needs to go!)  -clean grill and replace briquets (which are coated with grease)  -install drip watering system in vegetable garden  -trim back trees along back of yard  -Put down landscape fabric and new mulch in front beds.  -install edging around front beds  -put up new mailbox (long story involving an \"unknown\" car that ran over our old mailbox - we know who did it, but he won\'t admit it)  -have Thai dinner with lovely wife (Marny\'s contribution to my To-Do list!)   It seems like a long list, but once I get into a groove I can get a lot done. I\'ll check in at the end of the week to let you know how I did and show off some pictures.  Happy Fourth everyone!   '),(2764,3306195,'01,August,2004','2004-08-01 00:00:00','    On Friday, Jeannette and I went to Tranquil Lake Nursery in nearby Rehobeth MA. Lots of back roads to get there, but it was well worth it! I picked up three new plants including a Yarrow, a Golden Curls Willow:   urlLink     and a Butterfly Bush. Within in ten minutes of planting the Butterfly Bush, we had our first butterfly on it!   urlLink      I finally got around to putting the lattice up around the bottom of the shed, giving it a truly finished look. Now I have to landscape around it. One thing at a time!   '),(2765,3972268,'30,July,2004','2004-07-30 00:00:00','   Time for my big vacation! It\'s really cheered me up, now that we\'re about to go. I just can\'t wait to go back to my home town! California here I come if you pardon the old phrase :)  '),(2766,3972268,'26,July,2004','2004-07-26 00:00:00','   Why is it me?  Why does it hurt?  my heart would sing,  I had the world   Do I have a chance?  Do I have a Hope,  of being with her,  the answer\'s nope   what hope I have,  I cling to still,  my heart is dead,  my heart is still,   I wish tonight,  With all my soul,  to have her still,  The one I wish to hold   &nbsp; ok....I\'m feeling horrible right now. Ever had the feeling you lost your insides suddenly? How it feels to see something you held dear seem to just float away? If I\'m not feeling that now I\'ve got something wrong with me. The story is as simple as it get but it\'s way to complicated for me. Boy meets girl, falls in love, finally tells her, after awhile she tells she likes another, Boy gets over it hoping there\'s a chance, Boy\'s put into his place and feels like jumping off a cliff.&nbsp; O_o what a day.....and it\'s not even noon time yet.  '),(2767,3972268,'24,July,2004','2004-07-24 00:00:00','   &nbsp;-_- *sigh* if being bored was fog, you could cut it with a knife right now. Boy has this been an uneventful day. I did finish some work on a dragon I\'m making but other then that it\'s just been one long day with nothing to do. Yeah, I\'m sure there\'s SOMETHING I could be doing but that\'d probably be either chore or something I don\'t feel like doing. Figures, I hate it when that happens......AND I STILL CAN\'T FIND ANY DECENT SOAP! ARG! WHERE IS ALL THE SOAP!  '),(2768,3972268,'23,July,2004','2004-07-23 00:00:00','   Well, fate has been kind to me lately. I\'ve gotten passing in all my APs, a 4 and a 5, so now I have 12hr. all together in Colledge credit, whipppy! I\'ve made some progress in a project I\'ve been working on for some time too. I need to start my exersising soon though, I do it now and then but I\'ll need the endurance for my riggerous trip to San Diego. I do have to remember one thing.....SOAP! must remember soap. I\'ve got to get my hands on it or I\'m doomed(exaggeratoin).*begins searching everywhere for soap bars*  '),(2769,3972268,'21,July,2004','2004-07-21 00:00:00','   Hm....well. Things havn\'t been doing the best lately. Yesterday I was EXTREMELY depressed, mostly about fate and all that stupid stuff like that. It\'s nice having a&nbsp;close friend, she helped me through it. Other then that not much else has come up, just drawing some various stuff and all.  '),(2770,3972268,'19,July,2004','2004-07-19 00:00:00','   &nbsp;Yello people! This is my introductory post so don\'t expect anything extravagant :P. Nothing much has been going on lately other then my impending graduation(I\'m a senior this year in high school so next summer I\'m free of it!). I\'ve certainly had a busy and...Ahem..slightly hard and painful summer...Cough..Cough.. but it\'s shaping up to be a good one, what with a trip to San Diego, CA coming up fast. I know it\'s a crappy poem..but I thought it up for someone...who obviously\'ll know who\'s it too when they see it :)    &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; My heart was torn asunder,  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; and we went our separate ways,  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; but forever more,  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; in my heart you\'ll stay.    &nbsp;  Adios muchachos! Must be off to my chores/internet surfing  '),(2771,3972268,'16,August,2004','2004-08-16 00:00:00','   Everthing isn\'t totally lost.. I\'ve had some sense knocked into me for hopefully the last time. Thanks kally, I don\'t know what I\'d do without you! Now if only I could hear your sweet voice once more....  '),(2772,3972268,'15,August,2004','2004-08-15 00:00:00','   ...has now meaning to me now :(  '),(2773,3972268,'12,August,2004','2004-08-12 00:00:00','   Alright, time to grit my teeth and live through the pain. >_  '),(2774,3972268,'10,August,2004','2004-08-10 00:00:00','   +_+ Blech, first day back and they sure don\'t give you any slack in seinor year...sure it\'s not all bad but.... +_+ Uye! *sigh* well I\'m off to school again, almost dreading what\'s going to happen next! All day yesterday I was bombarded with little acidents and of course the usual, You weren\'t here and missed something lecture. You\'d think they\'d have gotten sick of talking about that by now.... Kally wasn\'t on yesterday either... I really wish she was, I needed someone to talk to...although the way my days are going I\'ll need her even more today +_+ ! Well....wish me luck, I\'ll need it this year ; _ ;  '),(2775,3679805,'22,June,2004','2004-06-22 00:00:00','   Live everyday like its your last....... its called life and things happen ... good and bad... learn from them!!  Im out of that hell hole! THANK YOU GOD!!!    and yet im still not with the one i want to  be with... why?  whats wrong with me? is it something i did?  we talk all the time... i see you like everyother day  why?   ~kc  '),(2776,3679805,'21,June,2004','2004-06-21 00:00:00','   happiness is...  1. Falling in love.  2. Laughing so hard your face hurts.  3. A hot shower.  4. A special glance.  5. Taking a drive on a pretty road.  6. Hearing your favorite song on the radio.  7. Lying in bed listening to the rain outside.  8. Hot clothes fresh out of the dryer.  9. Chocolate milkshake. 10. A bubble bath.  11. Giggling.  12. A good conversation.  13. The beach  14. Finding money in your pocket 15. Laughing at yourself.  16. Midnight phone calls that last for hours.  17. Running through sprinklers.  18. Laughing for absolutely no reason at all.  19. Having someone tell you that you\'re beautiful.  20. Laughing at an inside joke.  21. Friends.  22. Accidentally overhearing someone say something nice about you.  23. Waking up and realizing you still have a few hours left to sleep.  24. Your first kiss (either the very first or with a new partner).  25. Making new friends or spending time with old ones.  26. Playing with a new puppy.  27. Having someone play with your hair.  28. Sweet dreams.  29. Hot chocolate.  30. Road trips with friends.  31. Swinging on swings.  32. Making eye contact with a cute stranger.  33. Making chocolate chip cookies.  34. Having your friends send you home-made cookies.  35. Holding hands with someone you care about.  36. Running into an old friend and realizing that some things (good or bad) never change.  37. Watching the expression on someone\'s face as they open a much desired present from you.  38. Watching the sunrise.  39. Getting out of bed every morning and being grateful for another beautiful day.  40. Knowing that somebody misses you.  41. Getting a hug from someone you care about deeply.  42. Knowing you\'ve done the right thing, no matter what other people think.  43. just shaven legs in a nice warm bed 44. getting to sleep in on a snow day  45. Smore\'s  46. realizing the person you like , likes you 2 47. winning a championship 48. getting cuddled  49. a nice warm bath  50. SUMMER  51. seeing old friends  52. making a SWEET save in a game 53. getting a test back that u thought u did bad on with a great grade on it 54. understanding something u didn\'t at first  55. finding something u have lost 56. ur friends remembering ur birthday  57. a call from some one u miss as u think of them  '),(2777,3679805,'20,June,2004','2004-06-20 00:00:00','   finals today.. SUCKED! who gives 158 multiple choice questions adn  2 essays?! BRAMANTE! THANK GOD IM OUTTA THAT CLASS!!! what a shit hole that was...    ...yah so all i could think about  today was how much fun it would be to  be skippin school with the kid i wicked like... just another dream  ..definatly didnt go to the beach today.. kinda upset but whatever          ~kc  '),(2778,3679805,'20,June,2004','2004-06-20 00:00:00','   i hate school! i cant wait till im outta this hell hole and on to better things!   finals absolutly suck!   but hey... tomorrow im going to the beach with one of my best friends ever!, NIKKI! gotta love the crazy girl!! ;-)   yah well.. study break i guess is over.. back to studyin for the wicked cool history final  tomorrow! I JUST LOVE BRAMANTE!...almost!   ~kc  '),(2779,3679805,'19,June,2004','2004-06-19 00:00:00','   i couldnt sleep at all last night i just kept thinking bout all the  shit thats going on right now in  my life... best friends, guys you know the works...  i dont know why im just sittin around not doing anything about ..us.. i really want this too work out but you dont seem to be making an effort either...  who knows whats true ne more... :-/   ~kc  '),(2780,3679805,'19,June,2004','2004-06-19 00:00:00','   To My Dear Friend,  A friend is someone who\'s    always there A friend is someone you know   who cares They\'re someone who will   share your dreams Someone who will hear   your screams They\'re someone who will   lend a hand Someone who will take   your stand They\'re more than a laugh and    a wonderful time More than someone who will    loan you a dime They\'re someone who accepts the    good and the bad Someone who will cheer    up the sad They\'re people who can   play around The people who won\'t   let you down They\'re someone who will   stay by your side Someone who you know   won\'t lie They\'re someone who knows    who   you   are Someone to help you reach    for the stars They help you to over come   your fears They\'re someone who will    wipe away your tears They\'re somone to teach you    the beauty in art Someone who always stays   close to your heart They\'re someone to keep you   on the right track Someone who won\'t just    turn their back They\'re someone to help you    keep your head clear Someone to help you drink    a beer They\'re someeone who is with    you throughout the day Someone who\'s trust just    doesn\'t fade away They\'re someone who will   always be Able to accept me just for   being me!             Thanks Friend!  '),(2781,3679805,'18,June,2004','2004-06-18 00:00:00','   yah so last night we went to a dance held by the RFL and it was gay... as usual  but while we were just sittin around doing absolutly  nothing i got to thinking... mostly about you and i realized that if im sittin here at a dance thinking about   you   than there must be something wrong with that picture...  i really do like you but i dont know why its taken me so long to realize...   i dont know ne more   ~kc  '),(2782,3679805,'18,June,2004','2004-06-18 00:00:00','   i was just straightening my hair when i saw a pic of me and  dennis hangin on my wall from grad dance last year and i  realized... i want that, but with someone else i dont know why i keep denying the fact that i do  but all i know is that i really wanna be with you having fun... but that may never happen and it gets me scared...  ..whatever..   .N.e.v.e.r.  .E.v.e.r.   .R.e.g.r.e.t.  .S.o.m.e.t.h.i.n.g. .T.h.a.t.  .O.n.c.e.   .M.a.d.e.  .Y.o.u. .S.m.i.l.e.    ~kc  '),(2783,3679805,'16,June,2004','2004-06-16 00:00:00','   today was just another day of finals... but 2 more days till im outta this hell hole!   thank god!    today tho.. i realized something... i dont need a guy to make me happy... sure i say that now and tomorrow ill be wishing im with someone..  but i have my friends there by myside and as long as i have that im golden  i also realized never to let a guy control your life... dont let him tell you what to do.. when to do it or who to take walks with.. its not healthy especially when you change so much that even your best friends are concerned dont  ever  change or do things just to make a guy happy even if you think your in love.... and more importantly... always, always save time for your friends.. dont consume yourself with someone just b/c you think you love em its just not right...    worried bout yah... yah know who you are :-/   ~kc  '),(2784,3679805,'16,June,2004','2004-06-16 00:00:00','   There\'s a boat, I could sail away There\'s the sky, I could catch a plane There\'s a train, there\'s the tracks I could leave and I could choose to not come back  How far do I have to go to make you understand I wanna make this work so much it hurts, but I just can\'t Keep on giving, go on living with the way things are So I\'m gonna walk away And it\'s up to you to say how far  There\'s a chance I could change my mind But I won\'t, not till you decide What you want, what you need Do you even care if I stay or leave Oh, what\'s it gonna be Out of this chair, or just across the room Halfway down the block or halfway to the moon   ...its kinda like that :- just likin the right guy for me at the wrong time!  '),(2785,3679805,'20,January,2004','2004-01-20 00:00:00','   I heard \"Get Low\" today adn all  i could think about was my field hockey ladies  I miss you girls so much I miss spending all of my time at the field with you all.  I miss the songs I miss the laughs. I miss the tears. I miss the fights. I miss the bonding. I miss the spice girls on the bus rides home. I miss how we comforted each other when we had a shitty came. I miss most of all is just being together, listening to angie tell us how she wants her    boobs to grow!    We dreaded....  the fact that kelly was leavin us  the fact that as soon as the season was over some of our bonds might break  jv season.. we all know we are in for an awakening   We loved...   the practices   the games   the parties   the bus rides   the inside jokes no one else got (ex. BLUE BALLS?!?, F.O.W?)   our coach!!   the spice girl songs angie use to sing on top of her lungs   the season   I miss you ladies already... we did such a great job going 8 straight games with  out loosing one! i love you ladies so much... i dont know what id ever do if i never met half of you!!! especially sully.. haha on fire are we? ;-) good times ladies good times!     ~kc   '),(2786,3679805,'08,January,2004','2004-01-08 00:00:00','   the banquet tonite was AWESOME!! i love all my ladies!  Freshman Season KICKED ASS!!!  12-1!! gotta love the ladies!   Coach Kelly... you will be missed!   SULLY...YOUR ON FIREEE!!!!!! hahah good time good times  Flannie- my little fish buddie! AWW!    ...can you say blue balls!? HAH gotta love our wicked werid convos!! ;-)...   '),(2787,3679805,'25,February,2004','2004-02-25 00:00:00','   yah.. so im over you why wont you just forget about us?   forget about me?   it is pointless!     ~kc   '),(2788,3679805,'18,February,2004','2004-02-18 00:00:00','   ehh.... im sorry bout this hole thing i dont know why i ever thought that we could survive another chance...   she screwed me over and she hurt you so dont blame it on me    ~kc   '),(2789,3679805,'06,March,2004','2004-03-06 00:00:00','   God saw you were getting tired And a cure wasnt to be So he put his arms around you And whispered \"Come to me\"  With tearful eyes we watched you And saw you pass away Although we loved you dearly We could not make you stay  A golden heart stopped beating Hard working hands at rest God broke our hearts to prove to us He could only take the best    Kevin Mitchell- known yah since i was 5 and i cant believe you left me so soon! missin yah more and  more everyday! its been a rough year without you hun! im sorry :-/!   '),(2790,3679805,'01,March,2004','2004-03-01 00:00:00','   sometimes you just gotta listen to your heart and not your head...  i apologize for not being the \'perfect\' girlfriend but i cant do it all...  But Never say I love you if you dont truely mean it because you\'ll end up hurting someone in the end.  But why cant friendship be enough?? Why does it need to be something more??  I have asked my self...       why back to him?      why not the one i truely like?      why not find someone new?  And i always come back to the same answer    \"b/c i thought i loved him and couldnt live with out him\"    WELL LOOK AT ME NOW! im doing fine on my own!!      ~kc   '),(2791,3679805,'27,April,2004','2004-04-27 00:00:00','   There\'s something about...  *butterflies before the game  *sweating out the stick check  *blocking the shot   *going on 1 on 1 with the enemy    *scoring the winning goal  *playing till you\'d drop dead    ..there\'s something about   Field Hockey   only a player could understand...   _______________________________________________________   - Coming home dirty and sweaty...  Coming home bruised and battered...  - feelin so tired that you dont even want to do a damned thing...  not wanting to get up in the morning b/c your so tired from the day b4 and you know that its just gunna happen again...  - Getting those weird tan lines...  Being up at 7 two weeks before school started...  - The yearn to just do it...  Having all your differences disapear when you get on that field with my favorite 18 ladies...  - The smell....                       ....what we do for the sport we love...   '),(2792,3679805,'25,May,2004','2004-05-25 00:00:00','   man i dont understand you   i try but i fail     i give up...    :-/       ~kc   '),(2793,3679805,'14,May,2004','2004-05-14 00:00:00','   ehh not much too say but  i like you wayyy too much!    ~kc   '),(2794,3679805,'10,May,2004','2004-05-10 00:00:00','   I AM: Kaitlin I WANT: him I HATE: the slut ;-) I WISH: she\'d just understand I HAVE: great friends I MISS: playing fielldhcokey every day I FEAR: not playing fieldhockey and not being with him I HEAR: country songs on my radio I SEARCH: for my dog.. hmm I WONDER: what you really think of me I REGRET: nothing I ALWAYS: get very protective of my close friends I AM NOT: the girly girl I DANCE: haha anywhere I SING: anytime.. especially to nikki and kate I CRY: when im frustrated..just ask nikki I CANT: not do anything I AM NOT ALWAYS: the nicest girl I WON: poker for once!!  I LOSE: my patience  I CONFUSE: myself I NEED: to learn that my friends are there no matter what  I SHOULD: tell *him* how i really feel I GET EMBARRASSED: sometimes     HAVE YOU EVER [*dated one of you best friends*] yah did work [*loved somebody so much it makes you cry?*] no.. i dont cry over that stuff..weird huh? [*drank alcohol?*] yes\'m [*done drugs?*] nope [*broken the law?*] maybee [*ran away from home?*] no [*broken a bone?*] no [*cheated on a test?*]haha only in science [*skinny dipped?*] yep [*played truth or dare?*] of course! [*flashed someone?*] hmm...:-/ [*mooned Someone?*] no [*kissed someone you didn\'t know?*] no [*been on a talk show/game show?*]oh can only wish [*been in a fight?*] yep [*ridden in a fire truck?*] yah  [*been on a plane?*] no [*come close to dying?*] no [*cheated on your boy/girlfriend?*] nope [*gave someone a piggy back/shoulder ride?*] hell yes [*eaten a worm/mud pie?*] no [*swam in the ocean?*] i love the ocean! ;-) [*had a nightmare/dream that made you wake up?*] wheen i was little yah but not ne more  LAST PERSON... [*You Touched*] umm... [*You Talked to*] my mom [*You Hugged*]  my mommy  [*You Instant messaged*] ALICIA [*You Yelled At*] my sister [*You Kissed*] hmm.. lets not go there  FAVORITE... [*Food*] pasta [*Drink*] water and ice tea (the rasberry kind) [*Color*] blue [*Album*] Rascal Flats [*Shoes*] my KSWISS! [*Candy*] ne thing chocolatey! [*Animal*] cows! [*TV Show*] one tree hill [*Movie*] Pirates of the Caribeen [*Dance*] NIKKIS DANCE! HAHA ALICIA [*Song*] hmm too many to count... [*Vegetable*] potatoe [*Cartoon*] spongebob!  ARE YOU... [*Understanding*] yah [*Open-minded*]  ehh [*Arrogant*] nope [*Insecure*] no [*Interesting*] haha i like to think so! [*Random*] hell yes! some of my friends swear i have add [*Hungry*] yah all the time! [*Friendly*] wicked;-) [*Smart*] yah..except in sciece [*Moody*] sometimes [*Childish*] not really [*Independent*] yah [*Hard working*] sorta [*Organized*] yah [*Healthy*] yah [*Emotionally Stable*] yep [*Shy*] never! [*Difficult*] sometimes [*Attractive*] i dont kno am i? [*Bored easy*] no im pretty easily entertained. [*Messy*] not really [*Thirsty*] ehh [*Responsible*] when i want too be [*Obsessed*] no [*Angry*] nope [*Sad*] not at the moment [*Happy*] always [*Hyper*] yes [*Trusting*] yah [*Talkative*] yes [*Legal* ] nope  WHO DO YOU WANT TO... [*Kill*] the slut!!! ;-) [*Slap*] ehh duno [*Get Really Wasted With*] my girls [*Get High With*] no one [*Look Like*] ALICIA MARSHIA I LOOVE THAT GIRL   Random [*in the morning I am*] tired and crabby [*all I need is*] food!! [*love is*] not for me appearntly [*i\'m afraid of*] getting outta here [*I dream about*] him  -H A V E .Y O U. E V E R . . [*pictured your crush naked?*] who hasnt? [*actually seen your crush naked*] nope [*been in love*] no [*cried when someone died*] yes [*lied*] hahah i cant lie are you kiddin me!  -W H I C H . I S . B E T T E R- [*coke or pepsi?*] coke [*flowers or candy*] flowers [tall or short] tall  -W I T H .T H E. O P P O S I T E. S E X- [*what do you notice first*] eyes, and smile [*last person u slow danced with*] haah o god... [*last person u enjoyed slow dancing with*] i duno [*worst question to ask?*] \"so whats going on between us??\" hahah damn that question  -W H O- [*makes u laugh the most?*] my girlies.. especially nikki! [*makes you smile*] everyone [*gives u a funny feeling when u see them*] someone ;-)  [*who do you have a crush on?*] well.. i dont wanna say [*has a crush on you?*] haha i duno [*is the easiest to talk to*] nikki  -D O. Y O U .E V E R- [*sit on the Internet all night waiting for someone special to I.M. u?*] naww [*save aol/aim conversations*] sometimes! [*wish u were a member of the opposite sex*] hahah yah [*cried because of someone saying something to u*] nope  -D O .Y O U / / A R E .Y O U- [*smoke cigarettes*] so gross [*obsessive*] nope [*could u live without the computer?*] NOPE! [*color ur hair*] no [*ever get off the computer*] yah [*habla espanol*] si senorita! me gustan mucho! [*how many people are on your buddy list?*] 200 [*drink alcohol?*] yep [*like watching sunrises or sunset*] yah man [*what hurts the most? physical pain or emotional pain?*] emotional pain [*trust others way too easily?*] nope.. i dont trust ppl that easily  -N U M B E R- [*of countries ive lived in?*] one [*of close friends?*] 5 reallllyyy close girls (kate, nikki, bailey, alicia n iggy) [*of cd\'s that I own?*] to many to count [*of scars on my body?*] wayyy to many [*of things in my past that I regret?*] i regret nothing   '),(2795,3679805,'30,June,2004','2004-06-30 00:00:00','   yah so i spent today and yesterday with the good old NIKKI!! haha gotta love the girl...   and for awhile i was sittin on my bed just thinkin to myself.. and i realized no matter how much i try i dont think i could move on not knowing wether or not  something could have happened... i guess im just gunna have to wait this one out... o well    ...yah sosummer league for field hockey starts next  wednesday july 7th! hells yes! the sooner summer league starts.. the soon Holy Cross comes and the sooner thats over its SEASON TIME BABY! i cant wait!   ~KC   '),(2796,3679805,'28,June,2004','2004-06-28 00:00:00','   \"are there any rocks ahead? if there are we\'ll all be dead! STOP RYTHMING AND I MEAN IT! anybody want a peanut?\"   ..yah i love the Princess Bride!! ;-)    As Another day passes by, I think of how much I like you more and more...  My friends see how likin you effects me but  they still seem to tell me \"dont worry  everything will work out\"..and i pray that they are right. And once in awhile ill get... \"aww you 2 would be so cute together\" and i just simply say \"yah i do like him\"   One Question \"Why?\" Why am i still chasing after someone that I  probably wont ever get... why am i fallin for you?  I\'m sorry...   '),(2797,3679805,'27,June,2004','2004-06-27 00:00:00','   oh man.. so much to say so little time.. where do i begin? well last thursday me and nikki got some disturbing news about one of our best friends being \"missing\" because  someone is a dickhead... anyways its all straigthened out now and she is safe...   then friday i hung out with my best bud katie.. i was going  through \"katie with-drawal\" haha  then sunday... we left for the cape at like 6 and took the nice 3 hr drive down to my aunt and unlces house down there... AND OF COURSE I GET STUCK NEXT TO ERIN THE HOLLEEE WAY THERE  AND BACK! 6 hours of her non stop talking! YEAY! hah it was fun tho... i got burnt, AGAIN! :-) oh man its not bad this time tho...   so yah im off to do something later     ~kc   '),(2798,3679805,'22,June,2004','2004-06-22 00:00:00','   im sorry i like someone who you think i shouldnt and dont like someone you think i should  its just not as easy for you because you have someone and i dont...  there is always that little thought in the back of my mind when i see you 2 together..\"why dont i have that?\"  im sorry but i still love you your my best friend for ever and always   much love to yah hun!!   ~kc   '),(2799,3679805,'21,June,2004','2004-06-21 00:00:00','   Live everyday like its your last....... its called life and things happen ... good and bad... learn from them!!  Im out of that hell hole! THANK YOU GOD!!!    and yet im still not with the one i want to  be with... why?  whats wrong with me? is it something i did?  we talk all the time... i see you like everyother day  why?   ~kc   '),(2800,3679805,'21,June,2004','2004-06-21 00:00:00','   happiness is...  1. Falling in love.  2. Laughing so hard your face hurts.  3. A hot shower.  4. A special glance.  5. Taking a drive on a pretty road.  6. Hearing your favorite song on the radio.  7. Lying in bed listening to the rain outside.  8. Hot clothes fresh out of the dryer.  9. Chocolate milkshake. 10. A bubble bath.  11. Giggling.  12. A good conversation.  13. The beach  14. Finding money in your pocket 15. Laughing at yourself.  16. Midnight phone calls that last for hours.  17. Running through sprinklers.  18. Laughing for absolutely no reason at all.  19. Having someone tell you that you\'re beautiful.  20. Laughing at an inside joke.  21. Friends.  22. Accidentally overhearing someone say something nice about you.  23. Waking up and realizing you still have a few hours left to sleep.  24. Your first kiss (either the very first or with a new partner).  25. Making new friends or spending time with old ones.  26. Playing with a new puppy.  27. Having someone play with your hair.  28. Sweet dreams.  29. Hot chocolate.  30. Road trips with friends.  31. Swinging on swings.  32. Making eye contact with a cute stranger.  33. Making chocolate chip cookies.  34. Having your friends send you home-made cookies.  35. Holding hands with someone you care about.  36. Running into an old friend and realizing that some things (good or bad) never change.  37. Watching the expression on someone\'s face as they open a much desired present from you.  38. Watching the sunrise.  39. Getting out of bed every morning and being grateful for another beautiful day.  40. Knowing that somebody misses you.  41. Getting a hug from someone you care about deeply.  42. Knowing you\'ve done the right thing, no matter what other people think.  43. just shaven legs in a nice warm bed 44. getting to sleep in on a snow day  45. Smore\'s  46. realizing the person you like , likes you 2 47. winning a championship 48. getting cuddled  49. a nice warm bath  50. SUMMER  51. seeing old friends  52. making a SWEET save in a game 53. getting a test back that u thought u did bad on with a great grade on it 54. understanding something u didn\'t at first  55. finding something u have lost 56. ur friends remembering ur birthday  57. a call from some one u miss as u think of them   '),(2801,3679805,'20,June,2004','2004-06-20 00:00:00','   hahah oh man  finals today.. SUCKED! who gives 158 multiple choice questions adn  2 essays?! BRAMANTE! THANK GOD IM OUTTA THAT CLASS!!! what a shit hole that was...    ...yah so all i could think about  today was how much fun it would be to  be skippin school with the kid i wicked like... just another dream  ..definatly didnt go to the beach today.. kinda upset but whatever          ~kc   '),(2802,3679805,'20,June,2004','2004-06-20 00:00:00','   FINALS!  o god!  i hate school! i cant wait till im outta this hell hole and on to better things!   finals absolutly suck!   but hey... tomorrow im going to the beach with one of my best friends ever!, NIKKI! gotta love the crazy girl!! ;-)   yah well.. study break i guess is over.. back to studyin for the wicked cool history final  tomorrow! I JUST LOVE BRAMANTE!...almost!   ~kc   '),(2803,3679805,'19,June,2004','2004-06-19 00:00:00','   couldnt sleep  i couldnt sleep at all last night i just kept thinking bout all the  shit thats going on right now in  my life... best friends, guys you know the works...  i dont know why im just sittin around not doing anything about ..us.. i really want this too work out but you dont seem to be making an effort either...  who knows whats true ne more... :-/   ~kc   '),(2804,3679805,'19,June,2004','2004-06-19 00:00:00','   Dear Friend...  To My Dear Friend,  A friend is someone who\'s  always there A friend is someone you know who cares They\'re someone who will share your dreams Someone who will hear your screams They\'re someone who will lend a hand Someone who will take your stand They\'re more than a laugh and  a wonderful time More than someone who will  loan you a dime They\'re someone who accepts the  good and the bad Someone who will cheer  up the sad They\'re people who can play around The people who won\'t let you down They\'re someone who will stay by your side Someone who you know won\'t lie They\'re someone who knows  who you are Someone to help you reach  for the stars They help you to over come your fears They\'re someone who will  wipe away your tears They\'re somone to teach you  the beauty in art Someone who always stays close to your heart They\'re someone to keep you on the right track Someone who won\'t just  turn their back They\'re someone to help you  keep your head clear Someone to help you drink  a beer They\'re someeone who is with  you throughout the day Someone who\'s trust just  doesn\'t fade away They\'re someone who will always be Able to accept me just for being me!  Thanks Friend!   '),(2805,3679805,'18,June,2004','2004-06-18 00:00:00','   ehh  yah so last night we went to a dance held by the RFL and it was gay... as usual  but while we were just sittin around doing absolutly  nothing i got to thinking... mostly about you and i realized that if im sittin here at a dance thinking about you than there must be something wrong with that picture...  i really do like you but i dont know why its taken me so long to realize...   i dont know ne more   ~kc   '),(2806,3679805,'18,June,2004','2004-06-18 00:00:00','   fallin for *him*...  i was just straightening my hair when i saw a pic of me and  dennis hangin on my wall from grad dance last year and i  realized... i want that, but with someone else i dont know why i keep denying the fact that i do  but all i know is that i really wanna be with you having fun... but that may never happen and it gets me scared...  ..whatever..   .N.e.v.e.r. .E.v.e.r. .R.e.g.r.e.t. .S.o.m.e.t.h.i.n.g. .T.h.a.t. .O.n.c.e. .M.a.d.e. .Y.o.u. .S.m.i.l.e.    ~kc   '),(2807,3679805,'16,June,2004','2004-06-16 00:00:00','   today was just another day of finals... but 2 more days till im outta this hell hole!  thank god!   today tho.. i realized something... i dont need a guy to make me happy... sure i say that now and tomorrow ill be wishing im with someone..  but i have my friends there by myside and as long as i have that im golden  i also realized never to let a guy control your life... dont let him tell you what to do.. when to do it or who to take walks with.. its not healthy especially when you change so much that even your best friends are concerned dont ever change or do things just to make a guy happy even if you think your in love.... and more importantly... always, always save time for your friends.. dont consume yourself with someone just b/c you think you love em its just not right...    worried bout yah... yah know who you are :-/   ~kc   '),(2808,3679805,'16,June,2004','2004-06-16 00:00:00','   There\'s a boat, I could sail away There\'s the sky, I could catch a plane There\'s a train, there\'s the tracks I could leave and I could choose to not come back  How far do I have to go to make you understand I wanna make this work so much it hurts, but I just can\'t Keep on giving, go on living with the way things are So I\'m gonna walk away And it\'s up to you to say how far  There\'s a chance I could change my mind But I won\'t, not till you decide What you want, what you need Do you even care if I stay or leave Oh, what\'s it gonna be Out of this chair, or just across the room Halfway down the block or halfway to the moon   ...its kinda like that :- just likin the right guy for me at the wrong time!   '),(2809,3679805,'08,June,2004','2004-06-08 00:00:00','   im movin on....   ive found someone knew...    im sorry about this...   sorry bout being a btich but its the way things are     ... why are you like this? im sorry!!!      ~kc   '),(2810,3679805,'26,July,2004','2004-07-26 00:00:00','   so i was sittin outside on my back porch with reily just thinkin... and you know sometimes when you just sit on your porch or like just outside, you can just feel like nothing else matters. All that matters is that one little moment in time when you are enjoying yourself. No one can bother you...... then reality hits you like a slap in the face.. yah well i hate that.. i was just begining to think about why im still likin erik and then all of a sudden \'kate, dinner\'... im like damn it i lost my train of thought... o well.. just thought id share....     &nbsp;    &nbsp;    &nbsp;   ~kc   '),(2811,3679805,'25,July,2004','2004-07-25 00:00:00','   I\'d sure hate to break down here.   Nothin\' up ahead or in the rear view mirror.   Out in the middle of nowhere knowin\'   I\'m in trouble if these wheels stop rollin\'.   So God help me keep me movin\' somehow.   Don\'t let me start wishin\' I was with him now.   I\'ve made it this far without cryin\' a single tear   And I\'d sure hate to break down here.   - Julie Roberts    &nbsp;   &nbsp;  i love that song...it makes me jsut wanna sit down and think about all the shit thats going on in my life and makes me wonder, why am i still trying to be with you...why do i even care ne more? am i this pathetic to want something this much?.. i mean honestly i can live without you as well.. but id really like to live with you... whatever    &nbsp;   &nbsp;  lunch time.. ill write back later   '),(2812,3679805,'25,July,2004','2004-07-25 00:00:00','   hey guys... oh man.. all i have to say is Field Hockey is apon us! hahah oh man girls so many great times @ camp! gotta love em all!     &nbsp;   &nbsp;  im jsut so afraid to ask you cuz i dont wanna sound desprite.. but i really wanna kow whats going on between us! ... oh man    &nbsp;   &nbsp;  ..&nbsp;i just dont know hwat to think ne more....   '),(2813,3679805,'15,July,2004','2004-07-15 00:00:00','   i just had this conversation with a friend of mine Danya from Gardner... it put alot of things into perspective for me.. like how much i really love you gals..just thought id share it with yah     &nbsp;   considine4 15 18: hey did how was your summer league game?    Miss Margarita07: alright-we lost though lol    considine4 15 18: :-( who did you play?    Miss Margarita07: lunenburg    Miss Margarita07: they have LOTs of players!    Miss Margarita07: it was crazy lol    considine4 15 18: hahah oh man... i dont know if we play them    Miss Margarita07: lol they were annoying!    considine4 15 18: hahah no MAHAR is annoyin!    Miss Margarita07: they ahd ppl like yelling 24/7 i was like AHHH shut up!!!!    Miss Margarita07: really?    considine4 15 18: actually.. my teams annoying.. we talk no stop    considine4 15 18: haha :-[    Miss Margarita07: lol    Miss Margarita07: we\'re quiet-i try to get ppl to talk    considine4 15 18: well we do..we have conversations that arent related to ne thing while we are playing    considine4 15 18: it distracts the other players and they get annoyed and then we jsut laugh    Miss Margarita07: LOL never really done that    considine4 15 18: hahah we have.. its kinda funny but then again..&nbsp; kinda distracting.. and on occasions youll get us to fight like sisters.. its sad    Miss Margarita07: lol like one big family huh?    considine4 15 18: oh yes!    considine4 15 18: haha    Miss Margarita07: yeah, i\'m not real close w/ ppl on my field hockey team. i do talk to them, but nothing like serious    considine4 15 18: whyy not?!? that is like the greatest part    Miss Margarita07: lol idk! we just never really hit it off. i don\'t like many of them anyway-i have more friends on the track team lol    Miss Margarita07: and a lot of my other friends play different sports    Miss Margarita07: one of my best friends plays soccer, basketball, and tennis lol    considine4 15 18: i see    considine4 15 18: its all good..    Miss Margarita07: yup    considine4 15 18: see we are all like sisters.. nothing can come between that bond.. its kinda weird tho cuz its like we are all from different \"cliques\" i guess you could say but we all love each other on the field    Miss Margarita07: awww thats cute :-P    considine4 15 18: yah man    Miss Margarita07: yeah i mean it\'s not like i don\'t talk to them at all-we seem to get along on the field tooconsidine4 15 18: thats always good    Miss Margarita07: like this girl leanne, i\'ve never really liked her    Miss Margarita07: but anyway, when we\'re on the field, we\'re completely nice    Miss Margarita07: she plays midi   &nbsp;and i usually play a wing    Miss Margarita07: and we always compliment each other lol    considine4 15 18: hahah i know its weird... like&nbsp;some of the girls&nbsp;usually never get along out side of field hockeybut when&nbsp; we are playin its a totally different story    Miss Margarita07: haha thats great! field hockey seems to be the only sport with that special bond :-P    c  onsidine4 15 18: hahah yah man     &nbsp;   ... i gotta say, we do share a bond that no one else can see.. its the tightest bond i think i have ever had with a team taht i was one.. i love you ladies, not matter what!   '),(2814,3679805,'14,July,2004','2004-07-14 00:00:00','   i really am so excited to see you.. only a week left until i get to acutally see you adn talk to you.. i miss our conversations so much.. and it sucks cuz in a week ill know how you feel.. and i dont know if thats a good thing or not... but all i know is that not seeing you for 2 weeks probably was the best thing to decided whats gunna happen.. it just puts everything into perspective for yah...     &nbsp;   but ne ways...2 more days man 2 more days! i cant   wait!!   its gunna be so much fun ladies! :-)     &nbsp;   jamo... haha PARTTTY.. i mean serious work mrs jamieson.. no worries! haha kim i cant WAIT!:-)     &nbsp;    &nbsp;   ~kc   '),(2815,3679805,'14,July,2004','2004-07-14 00:00:00','   aww man.. so little time between and sunday... i cant wait! we had a summer league game today!! we won... again! 4-2! and then there is sully.. oh man \"we got this bitch!\" hahah gotta love her.. and shei pullin the michael jackon moves on us! ;-) its all good! ... yah so 3 days till camp! whoop whoop.. what am i gunna do with out my friends tho!?! ahh!   but then i come back and i get the answer to the question i asked 2 weeks ago.. dun dun dun!   oh man.. im not quite sure what to expect.. i mean i wanna say yah you know everything is gunna go great! but i dont even know whats gunna happen.. and im kinda scared.. but whatever.. its all good     alright its late n im off to bed!    ~kc   '),(2816,3679805,'13,July,2004','2004-07-13 00:00:00','   yah sorry i havent been here in a while.. ive been busy! haha.. well lets see... friday night.. nikki came over after tom\'s party... we hung out saturday.. then sunday... i cleaned and JAMO came over adn we hung out and watched the RED SOX game.. monday.. hmmm cleaned, AGAIN! and then nikki cam over.. then today we went to the movies w/andy, iggy, mike, mike and cam showed up... weird.. but it was a gayy movie so (DONT SEE KING ARTHUR! GAYY!)   but yah so.. 5 days till camp with my FAVE ladies ever! (besides my gals ;-) you know who you are)  which means 9 more days till im home... hes home... but i guess it may be a bad thing too... what if you decide that its not for you right now adn nothingn happens between us... i dont know ne more.. all i know is that ireally do like you and this past week n 1/2 ive realized its not just something that i pretend to do.. its something for real...   well i g2g later    ~KC   '),(2817,3679805,'07,July,2004','2004-07-07 00:00:00','   hey guys.. sorry i havent been on... i went to the DAVE MATTHEWS CONCERT last night FUN FUN!  then today had my first summer league game! we won.. of course haha 1-0 now baby!! ;-)  gotta love my ladiess!      missin yah bud! even tho i know that  something may change im still prayin for the best!       ~KC   '),(2818,3679805,'04,July,2004','2004-07-04 00:00:00','   yah so this 4th of July was kinda shitty honestly... seeing danielle and dave all over each other couldnt help but make me think of you.. and i was  havin sucha good day too...   dude 18 days till i see you again.. man i miss you like crazzyy! i miss talkin to yah.. goofin on yah.. all that greeat stuff!   people say that i shouldnt be missin yah as much as  i do cuz we arent even together but i do.. and its unexplainable why.. i just do! UGH!   14 days till camp! woot woot 3days till SUMMER LEAGUE MAN!    ~kc   '),(2819,3679805,'01,July,2004','2004-07-01 00:00:00','   18 days till im outta this damned little town and  at my HOLY CROSS CAMP! THANK THE LORD! hehe  FIRST SUMMER LEAGUE GAME... Wednesday July 7th  @ MT Wachusett College in Gardner! hells yes! this is gunna be interesting... spendin almost my hole summer with my field hockey ladies!! hell yah!     ...some days i just wish that you\'d begin  to understand that not everything is my fault.... you have \'him\' and i  have no one so dont get pissed at me for things taht happens with out you...    just move on!  ~kc   '),(2820,3663542,'24,June,2004','2004-06-24 00:00:00','    Î”ÎµÎ½ Î¼Ï€Î¿ÏÎ¿ÏÏƒÎ± Î½Î± ÎºÎµÎ¹Î¼Î¹Î¸ÏŽ, Î­Ï„ÏƒÎ¹ Î¬ÏÏ‡Î·ÏƒÎ± Î±Ï…Ï„ÏŒ, \'Îº Ï‡ÏÎµÎ¹Î¬ÏƒÏ„Î¹ÎºÎµ ÎºÎ¬Ï„Î¹ ÎµÎ´Ï‰... Î½Î±Î¹ Ï„Î¿ Î¾Î­ÏÏ‰ Î¿Ï„Î¹ Î´ÎµÎ½ Î¼Ï€Î¿ÏÏŽ Î½Î± spell ÎºÎ±Î¸ÏŒÎ»Î¿Ï…... Ï…Î»Î¹ÎºÎ¬ ÏƒÏ„Î± ÎµÎ»Î»ÎµÎ½Î®ÎºÎ±... ÎŸÎš, Î±Ï…Ï„ÏŒ ÎµÎ¯Î½Î±Î¹ Î¿Î»Î± Î³Î¹Î± Ï„Î¿ÏÎ±... Î±Î½Ï„Îµ.   '),(2821,3663542,'24,June,2004','2004-06-24 00:00:00','    Î¿Îº, ÎµÎ¯Î¼Î±Î¹ Ï€Î¿Î»Î»Î® Ï„Î±Î¼Ï€Î­Î»ÎµÏ‚, ÎºÎ„Î´ÎµÎ½ Î­Î³ÏÎ±ÏˆÎ± ÎµÎ´ÏŽ Î³ÎµÎ¹Î± Ï€Î¿Î»Î»Î® ÎºÎ­ÏÎ¿.&nbsp; Ok the greek takes forever. I\'m just writing tonight because im kinda in a wierd mood.&nbsp; So yeah, this lacks content! Well see ya\' later, going to play soccer'),(2822,3663542,'24,June,2004','2004-06-24 00:00:00','    I\'m starting to think that maybe I\'m some kind of angel of death in training or something. K so maybe I\'m not that bad but in the past week and a half I have come up on 3 accidents immediately after they happened. Luckily I don\'t think anyone has died from them (well, not while I was there) but they have augmented in severity. The first was a guy fell off his motorcycle and went sliding down the road with his bike beside him; I actually saw that one. The next one involved two cars meeting on a really bad set of curve on this road her in Sylacauga; don\'t know as much about that one, but I was like the 5th or 6th car there.  Then today when I was coming home from a YSA conference in Huntsville, I was going down 208 and came up on another one. This one I didn\'t actually see, but that was because the cars in front of me were obscuring my view. But I was one of the first cars there, and so I stopped and I saw this man from behind where I parked running toward one of the two cars involved that was on its side. So I got out and went the same way he was running to see what I could do to help, by the time I was over there the man and several women were trying to push the truck that was on its side back over. My first thought was that it was a odd thing to do because the police would want to see things as the were left by the accident, however they were screaming something about a man being trapped, and then I saw why it was necessary. A man had been thrown from the truck, and now the truck was on top of him. I immediately started to help to right the truck and free the man, and luckily one or two more fellows showed up right behind me and we were able to push the truck back over. At this I could see the man, he look in pretty bad shape, I thought for sure he was dead or about to die, but I didn\'t have much time to think about that because the original man that was there was trying to get the truck door open, and was saying something about getting a woman out of the truck; I helped him with no avail with the doors until he jumped in the open window to try and pass the lady out... The lady in the truck was down in the floor, I didn\'t notice she was in there till then, I had no idea about her when we righted the truck, but her then I noticed her legs were poking out of the window so I went around and tried to help get her out, but the other two men were there and had her legs, and the man in the truck was lifting her head up and out. Once out of the truck we took the lady over to the side of the road and set her in the grass. From here the women who were there were awesome, they found rags and blankets to put under there heads and tried to comfort them and take care of the two as best they could. I stayed around in case there was something I could do, so the best thing I could think of at the time was to help direct traffic, as this was blocking both sides of 280.  Shortly before the paramedics arrived a surgeon showed up and started checking the two people, however the paramedics arrived really quite fast and started helping. I believe the girl in the other car was alright for the most part, apparently after it happened she got out of her car and staggered over to the median, but she was just really shaken. Well eventual the police came and more paramedics, like I said the guy was in a bad shape, and they decided to get a helicopter in and airlift him out. We kept traffic moving as long as we could but eventually we stopped both sides of 280 to allow the helicopter to land on the highway, and so the cops could take pictures and mark where the cars were. By the time the helicopter got there they had striped the man down and put a blanket on him to check him for internal bleeding and the like. We also found out sometime that the man had palsy and was in a wheelchair, witch we found in the road. Eventual the got the man air lifted out and the woman in an ambulance. While they were still taking pictures the sheriff of Shelby county came over to me and the original man that was helping, and thanked us for what we had done, and told us we did the right thing in getting the truck off the man and the woman out, then he gives us all these coin like things that are kinda neat. Luckily the also let us go before they let the rest of the cars that were waiting to get through go. However in my rush to see what I could do, I had left my lights on, SO I had to go back out and get a officer to come over and jump my car off, I don\'t think he minded too much because of the circumstances. Anyway, wow...  Well that\'s a chunk of text, but I figure its alright, I just wanted to get it all out. If I keep coming up on accidents like this I may need to change my e-mail address. But seriously I wonder if there is something else I could do in situations like that. I mean I don\'t think they were members, but should I have offered them a blessing, or the woman at least (she was more aware of what was going on). The thought passed though my mind briefly while I was there, and I dismissed it and said a silent prayer instead. But I always wonder what more I could do in situations like that. Maybe I should brush up on my first aid, and trama treatment, especially if this keeps up. Perhaps there is something I should learn from this. This just reaffirms what I already do, in that we should always wear our seatbelts!!! I already do, but after this past month, what I\'ve seen... wow...  other than that my weekend was good, the conference was good, and I enjoyed myself, meet some people, and such. It was odd b/c I spent most of the weekend with people I didn\'t know before, not typical thing I would do. Anyway I don\'t feel much like typing about that now sooo. Ill leave it at it was fun, meet some good people, had fun dancing and yeah... I enjoyed it. Perhaps more on that later, but don\'t hold your breath... But I should sleep, \'cause I didn\'t do much of that this weekend, if you know what I mean! (so I cant sleep on a flat mattress...)   '),(2823,3663542,'24,June,2004','2004-06-24 00:00:00','    I think the frantic romantics said it best when they said \"life ain\'t easy as a fresh RM, when all the girlies you be wantin\' say you ain\'t for them\" not that that really applies to my day or week or month in anyway, but its in my head.  I\'m about to go away, but I need to get hold of Holly first. I\'ve been debating what I should wear for the dance on Saturday, because its carnival theamed. I was thinking of wearing my capoeira pants, and such, but I\'m scared to stand out too much, and I would have to be barefoot, and that\'s not comfortable, besides I doubt anyone would get it. OR I could go the other way and walk around with juggling stuff all night, but I hate having things on me, I never know what to do with them, besides only clubs or devil sticks would be really cool, and I would do those for like 5 min, and get board, so I figure ill just go normal and such.  k... well...   '),(2824,3663542,'24,June,2004','2004-06-24 00:00:00','    Well, this is odd, im awake before noon, a novilty for me. My brain dosnte work as well in the morning, its more just sitting stairing, remembering im supposte to be thinking, and trying to think again... Humm... i was going to basicly rewrite what was lost, but i cant rememeber what it was exactly, that i wrote, well i kinda can, but not at the same time.  One thing i remember, that i wish i didn\'t, was the bad news i got yesterday. Yeah so i found out that Kevin, the kid i was going to be rooming with at Ydaho is not comming anymore. Yes that really sucks for me. Every stupid semeseter i have to get new roomates, every semester, and im sick of it. Its like some kind of cosmic rule that every four months i have to get a new pool of friends, just really annoying. Anyway so now i have now real guy friends up there i can hang out with, and i dont know a soul im living with. I dont want to have to get to know a new room roomate, thats probably the worst part. I just know how i am, it takes me awhile to get warmed up to people and to feel comfortable around people so that i can be myself, so im not looking foward to being uncomfortable for however long it takes.   Oh yeah i also decided last night (though i doubt ill actually do anything about it) that i need to learn how to read menues, er perhaps just will myself to read them. Anyway last night i went to a mexican restraunt with my family, now its normal for me to be the last one to order, simply because im undesisive, but of late i cant even seem to bring myself to read the menu. Last night was the worse i have ever been, i didnt read one word of the menu, till it was my turn to order. So i just skimed the menu and pointed to somthing, witch turned out to be somthing on the lunch menu. After that i ordered what my mom got, \'cause it sounded alright.*  I have a really bad tast in my mouth, im going to drink somthing... brb... not that you, the reader, could tell that im gone, but if i tell you my actions its more like im actualy here telling you, the reader, myself... k hold on...     Im back.  After dinner we went by wal-mart and i finaly got a belt that i could put my confedriate states belt buckel on, im really pelased. In adition to the belt i found some cowboy boots i used back when i showed cows, and i have a cowboy hat, so i was wearing the boots yesterday feeling all western. This is not my new style or anything, but i just think its awsome so now when i go to the weekly country dance** i can look the part, im excited.  I miss dancing, though i fear i\'ve forgoten so much. I really hope my sechdual works out so i can take the dance classes i want. I think it will as far as standard is conserned, i believe that the gols class is 8 in the morning, not happy about that but would be worth it. Unfortunatly i dont beleave latan will work out again this smester, and im sad for this thing. The ironic thing is i love to dance and take all these dance classes, but when i go to dances i hardly dance. Its partialy because how i dacne is too prpoer and needs way too much room for normal dances. If i was too dance like that at a normal dacne the girl would be wondering why my arms are so high, why im so close, why we keep running into people, and why i never look at her, i fear. (though the last thing she may appreciate..)  Wow if i just sit here long enough i can write a ton of stuff! Anyway sicne my power went out last night and i lost the post after that im really parinoid of loosing this post, i keep saving it to notepad. Did i mention that i want to build a computer?    * I would write what we ordered but i cant spell, and doubt i could spell what it was close enough to get on the spellcheckers radar... thats so sad.  ** I go to the country dance for the dancing not so much the music, but contry dancing is fun. In conjucntion with the country dance they have a swing dance in another room, witch i actualy perfer, and is more fun to dance. However less people go to the swing kids side because they go over there and get intimidated b/c if you are just learnting the Lindy hop looks really hard. To tell the turth im kinda intimdated myself to dance with a lot of the girls there, they all seem so much better than myself. Also the Lindy is a physical dance, and it really hard to talk while doing it, and it really tires you out, or gets you dizzy.   '),(2825,3663542,'24,June,2004','2004-06-24 00:00:00','    I wrote a good post after the power went out, but then when I tried to post it, the thing went all gay and I lost the post... so all ya\'ll get for now is a lame excuse... sorry kids, more later, I\'m hittin the hay.  Man, I really want my post back, I\'m tempted to rewrite it, but I really need to sleep more regularly, aahhhaahaahaa!!! Ï„Î¹ Î½Î± ÎºÎ¬Î½Î½Ï‰!!! (what should I do!!!) I don\'t really want to rewrite it, but I want what I said to be posted. This is lame, and all I can think about now is pick up lines... well ill leave with this: there are two fish in a tank, one says to the other \"how do you drive this thing?\"   '),(2826,3663542,'24,June,2004','2004-06-24 00:00:00','    Do be do be dooo... um... So after I showered today I was talking to my mom about my academic penalties that apparently I was under for 5 min today, and as I sometimes do when avoiding unpleasant conversations I started to mess with my hair, and I realized I didn\'t rinse out the conditioner. Man I hate when I do that; means I have to get wet again, at least partially (witch is worse than totally) and dry off again, and that\'s really annoying. Unfortunately for me since I started using conditioner I have done this quite often. See for most of my life I didn\'t use conditioner, thought it would make my hair too, something... But one day when I was having my hair cut the girl convinced me to start using conditioner, so I started. I kinda like it what it dose. Speaking of hair though I really need it cut, its beginning to really annoy me.  Humm... I thought I had 3 things I was going to talk about, but I can only remember two of them... I did the one, and now I don\'t really want to go into the other, but the short version is: G4techtv is gay, really gay. I was really into this one anime they were showing and the stopped it 1/2 way though the series, to show other gay one... This is not the only thing but its the last straw, I used to really like Tech TV, then they merged with G4 a gaming channel, and it all went to pot. All the G4 shows are sophomoric, you have to be a moron to watch them... Any way this really annoyed me, and I was going to rant and rave about it, but I don\'t feel like going into it deeper than this at this point.  I got a comment from someone I don\'t know on this thing. I thought it was awesome... his name was demitri, a greek name, so I went to his blog to find out more, and his is in spanish, and he\'s from mexico, so demitri if you read this thankx! I figure he came accrost this by searching for similar interests, b/c he had capoiera as a interest also...  I spend way too much time on  urlLink homestarrunner.com , they should make a homestarrunner trivia game somewhere, that would be awesome, I could hold my own at it. I come to this conclusion when my sister holly asks if homestar has a house and I tell her what Strongbad  urlLink e-mail  to go to (by name), where they are in H*Rs house. ah well, there are worse sites that I could spend time on, like g4techtv.com, man since the merger they cant even put together a decent site... its never updated, no content, (\'cept if you are looking for a cheat code). so gay sorry I regressed there.  Speaking of H*R I decided that it was time to change what my start button has on it. It did say \"Ben\'s Computer\" but I only did that to show my dad how I changed it, but it kinda stuck for a while. any way I have now changed my button to say \"eh Steve!\" it works for now.  oh yeah I really don\'t like AOL, I was on earlier and for the second time this summer I was kicked offline because apparently I signed on somewhere else. The last time this happened I changed my password immediately. Tonight however I let it slide, and figured it was a fluke, but about 8 min later it happened again. This time I changed my password and it hasn\'t happened since. So I\'m guessing that AOL\'s passwords aren\'t that hard to get past, so I started to figget with it and found that passwords on AOL can only be 8 charitors long, (the password I thought I was using was 9 charitors) any way just more for me to complain about about AOL. Cant waite till I get back to broadband...  I should go to bed, even though I don\'t have work tomorrow... its my last week of work and I only work 3 days! anyway I don\'t work tomorrow, I really should try to do something in that time, but it seems like so much effort. so ill probably sit at home and complain on this thing about something... OH!!!! I remembers the other thing, I think I found my problem with kicking a football (soccer ball) tonight. For some stupid reason when I kick with any power the ball will shank off to the right most of the time. Also I was noticing the other day my big toe on my right foot hurt. Well tonight I put 2 with 3 and got 8, saw that wasn\'t right took away 4 from the answer and got 4, figured that was close enough and turned in my paper. I found that I raise my big toe up when I kick, so its getting messed up, and is providing a uneven surface for the ball, forcing it right. So I taped my big toe to the one next to it and it made a huge improvement. the ball went straighter, harder, and I had more control. Now this may only be a \"feather\" (think dumbo) but for now it works for me, er it did, ill try it again tomorrow.  I love my dog and all, but right now he stinks. Yeah I\'m sure sometimes I stink too, but he likes it, I try to wash when I start to stink. He finds stink and will roll in it. At least he\'s not one of those annoying dogs that always wants attention.  la la la dum de dum... Oh I finale got my Berimbau strung tight enough so that it sounds right, now if I could only play it correctly for more than one iteration. Speaking of Capoeira, I\'m so gonna get beet down when I go back to the club, man I should practice more, but its so hard to want to do without a partner.  well its late, er, its early, so sleep yes   '),(2827,3663542,'24,June,2004','2004-06-24 00:00:00','    Man, I really meant to go to bed earlier tonight... But anime unleashed didn\'t show at 11, I didn\'t figure on it being on at 3 either, but I\'m glad I checked, cause it was... Any way the point is I watched it and while I did I messed with the HTML for this blog. One thing about this site, its not very user friedley, but you can mess with the HTML and change the way it looks, if you know the language... Anyway I changed some of the words on the site to things like \"hey kids!\" and so fourth. I plan on making more changes later, but who knows... I figured out how to make this box bigger, but I\'m not sure its a great idea, there may be a reason its this size. One reason I know its this size is to make the corners all round, if I enlarged it it would loose the roundness, at least on one side (because the corners are actually a definite sized graphic) that\'s why its always the same size, I could change it to accommodate the size of the window but yes... It was fun figuring it out though. I learned some HTML before, but that was a while ago... And I never did anything this complicated. Luckily I know enough about how programs are set up to figure it out and \"break the code\", so I guess I didn\'t need to get my sister after all, or a book. Anyway I\'m proud of myself.  Just thinking as I was writing that, I need some kind of symbol to let the reader know I\'m about to ramble mindlessly, and they may want to skip a bit. Like I could have used it up where I was explaining about the corners, actual I could use it on this whole post, actual 90% content you read on this blog you may want to skip. Perhaps a easer thing to have is a symbol that means this may be worth reading... humm.... Ok bed for real.   '),(2828,3663542,'24,June,2004','2004-06-24 00:00:00','    Yes! I\'m smarter than internet explorer! Stupid auto formatting was not allowing me to put the links in my favorites folder in the order that I wanted, but I beat it, and now its like I like... The secret was to remove everything from that folder and then return it in the order top down that I wanted it to appear. Sometimes I really hate when computers are smarter than they should be, like when you are on a login screen and you already have your name in and are 1/2 way though your password and then it decides you need to automatically be put on your name space, so 1/2 the password goes up before your name. Maybe its just me... Maybe its my computer, I probably need more memory, but then again don\'t we all. This brings me to what I really want to do... Build my own computer. Yeah yeah I really like my dell, but the challenge and coolness of making my own is very intreaging. I\'ve been thinking about it and if I use bits of my dell it wont cost as much as I was originally fearing, however I would have to sacrifice the dell. But I could use the chip (a P4), and maybe some memory from this box, monitor (though what I really really want is a flat screen) and the cheap stuff like keyboard and such.  My dad got some beads, and we put together some komboulois (or Begleri to be technical) now I have a total of 7. The glass one chipped today, but I have a new favorite, one made with beads made of horn, just the string needs to be improved. Also he had the great idea that dice would be heavy and probably wouldn\'t crack, and yeah they work great, a little different being cubes though.  k time for some milk...   '),(2829,3663542,'24,June,2004','2004-06-24 00:00:00','    ÎŸÎš Ï„Î¿ Ï€ÎµÏ‡Î½Î¯Î´Î· ÏƒÎ¿Ï… Ï‡ÏÎ¹Î¬Î¶ÎµÎ¹ Î²Î¿ÎµÎ¯Î´ÎµÎ±. Î ÏÎ­Ï€ÎµÎ¹ Î½Î± ÎºÎ¬Î½ÎµÎ¹Ï‚ Ï„Î¿ Ï€ÏŒÎ´Î¹ ÏƒÎ¿Ï… ÎºÎ±Î»Î»Î¯Ï„ÎµÏÎ±, Îº\' Ï„Î¿ Î¼ÎµÎ±Î»Î¿ÏÎ± ÏƒÎ¿Ï… ÎµÎ¯Î½Î±Î¹ Ï‡Î¬Î»Î¹Î±, ÎºÎ„ Î¾ÎµÏ‡Î¬ÏƒÎ¿Ï… Ï„ÏÎ¯Ï‚ Ï†Î¿ÏÎ­Ï‚! Î‘Î½ Î´ÎµÎ½ Î¼Ï€Î¿ÏÎµÎ¯Ï‚ Î½Î± Ï€Î­Î¾ÎµÎ¹Ï‚ Ï€Î¹Î¿ Î¼Î±ÎºÏÎ¹Î¬ Î¸Î± Ï€ÎµÎ¸Î¬Î½ÎµÎ¹Ï‚. ÎŒÎ¼Î¿Ï‚ ÎºÎ±Î½Ï„Î­ Î±Î¿ÏÏ‚ ÎµÎ¯Î½Î±Î¹ ÎºÎ±Î»Î»Î± Î³ÎµÎ¹Î± Ï„Î¿ ÏƒÏŒÎ¼Î± ÏƒÎ¿Ï…! Îš\' Î¼Îµ Î­Î½Î± Ï‡Î­ÏÎ¹ ÎµÎ¯Î½Î±Î¹ ÎºÎ±Î»Î»Î± Î±ÎºÏŒÎ¼Î±. ÎŒÎ»Î± Î±Ï…Ï„Î¬ Î¸Î± ÏƒÎµ Î²Î¿ÎµÎ¯Î¸ÎµÎ¹ÏƒÎ¹ Î¼Îµ Ï€Î¿Î´ÏŒÏƒÏ†ÎµÏÎ¿ ÎºÎ„ÎµÏ€Î¯ÏƒÎ¹Ï‚. Î•Ï„ÏƒÎ¹, Î½Î± ÎµÎ¯ÏƒÏ„Î±Î¹ ÎºÎ±Î»Î»Î±!   '),(2830,3663542,'24,June,2004','2004-06-24 00:00:00','    k I may have to start a new journal on one of the other fine sites, if not all of them. Mainly because I just posted a comment on another blog site, and I had to do that as anonymous, and that\'s kinda tacky, so if for no other reason to get my name on the stuff I do. Besides I\'ve been looking at them, er livejournal anyway, and it has some features that this doesn\'t have that I would desire unto myself, that this site lacks... Such as current mood, and music... Ah, I\'m so indecisive!!! But this site has some sweet features, like I can edit old entries, have more than one blog, or chage the date of the blog anywhere from 1999 to 2006! (it would be awesome if I could do a blog from some time B.C. or impersonate someone like Pythagoras, yeah that would be awesome) anyway that was my rant on journals for now.'),(2831,3663542,'24,June,2004','2004-06-24 00:00:00','    Since I\'m on here I might as well post. I think I have more fun messing with the settings of this thing... Anyway I\'m just debating weather to have the dates in greek or english, it doesn\'t really matter I guess, just makes the archive dates say Î™Î¿ÏÎ»Î¹Î¿Ï‚ instead of July, and Ï€Î¼ instead of am, and so fourth... ah I\'m so indecisive. However I really wish I knew/remembered a bit more HTML so I could really modify the why this page looks, mostly I just want to change the space these main posts are put in to a bit bigger, there is so much more room on the page that they could use! oh well ill figure it out, or ask my sister to help me (ok I admit it! she knows more HTML than I do!   but I know more about computers overall... eyah...  ).  I\'m really not excited about work this week, my schedule is kinda messed up, usually I go into work about 5, or 6, but tomorrow I go in at 10:30 in the morning, what is that? Like a normal time, cheah... Î¼Î±ÎºÎ¬ÏÎ¹ (better, or I wish) it could be today, then perhaps I could go catch part of institute... Also on friday I go in at 4, seriously who orders pizza @ 4 pm? k so I would but fish, (I use the conjoining word \"fish\" here because I just noticed I ues \"yeah\" way too much) not every day. I shouldt complain, and I\'m not really, I just think its odd, and I\'m confused, cause I\'ve never gone in at 10:30 before, I guess cause I\'m leaving they are giving me whatever slot is left to put me in. But only 8 more days of that fun so yes.  I like fanta orange.  I wonder if they ever tried to, or dose it exist, a dark chocolate icecream, or would that just be nasty. Dunno. Perhaps in mainland europe somewhere, they like a lot of coco in there chocolate there. In fact they tried to tell the english company Cadberries that they couldn\'t call there chocolate chocolate anymore because it didn\'t have enough coco... of course they told them to stick it in there ear, england \"discovered\" (or ripped off the idea from the indans, but made it what it is today) chocolate, and this is how they have always done it.  Well, I should e-mail some people now, but, really I cant be bothered. The sad thing is when its day I\'m all like hey, ill e-mail tonight, and when it night I\'m all like hey, ill e-mail tomorrow during the day, but neither ever gets done. ah well. ya\'ll be good now.   '),(2832,3663542,'24,June,2004','2004-06-24 00:00:00','    K that sucked, I forgot to spell check the last post... Man! Ah well, it could be worse. Could be in Spanish, then I wouldn\'t even understand it. But hey! My dad found some glass beads that make great symbolic*, but there glass, and I\'m scared I\'m going to break them. But they are the perfect weight and size... Maybe I shouldn\'t play so rough...  *greek worry beads, (the kind I use) string with beads on the end.   '),(2833,3663542,'24,June,2004','2004-06-24 00:00:00','    K\' so I\'m making a new post on my journal, you lucky people you! Actualy I\'m doing this for Hollys benifit, because she didn\'t think I had a journal, and I did think any one would ever read what I would write. oh and i just invented a new word for my vocabulary... eyah, cause I keep misspelling it... basicaly its a \"more of a mischeavious form of yeah\" pronounced like yeah, but with a Î­ sound on the front of it. Ok so now after talking to Holly I beleave it that posting here may be more fun than I originaly thought. SO... yes... I was thinking of puting down some guidelines for the reader of this journal, to help guid them thought the randomness of how I will most likley write, but after a second thought it lost the desire to make such a list, and it didnt seem like it would have too much of rules on it, and they wouldnt really be rules for se\' but more like \"Im gonna wright like this, so tough!\" and that could go without saying. So if you know how I am how I wright won\'t be a suprise.  Also if you start to read this blog latter that this post you wouldnt have read the things I would list, so maby I should put somthing like that in the \"about me\" page. ah... whateve... I really like elips marks, really I should probably just start a new paragraph every time I use them but Î­ Ï„Î¹ Î½Î± ÎºÎ¬Î½Î½Ï‰; (ok, first time translation... that means \"what am I to do?\") K so I never got that why is almost all the punctuation in greek the same as english, except the question mark? In greek they use the semi-colen to denote a intarogitive sentence...* Also I really dont get/like how I is capitalized when typing, takes too much effort, and I think it seemes kinda pompus. i guess it wouldnt look right otherwise... oh well. K\' so my left sholder keeps hurting, for a little back information it has been doing this off and on for over a year now and its getting really anoying, I think it must be a nerve er somthing. at least its not as bad as when it happened, that was painfull. Ahh now I\'m complaining, and sounding untough... um... yeah so yeah so im gonna go out and play soccer now till i bleed or break somthing, caus im so tough!!!! K so that covers the sounding untough... AAAAHHHH!!! i cant type any better than i can spell... though the spelling is not my fault, my dad cant spell, his dad cant spell, and i dont beleave his dad could eather, and we all spell the same... the funny thing is when my dad misspells somthing i cant tell.  **NOTE* after writing this paragraph i decided it lacked actualy content, thuss you may want to skip it and read other parts of this post/journal however it was left in here to make this post seem longer, thank you!** Well it seemes like its time to start a new paragarph... the old one was about me and the journal, this one is about ideas i had today. I had a excelent idea of a mod for a computer... though i dont want to say here so someone will steal my idea, though if you ask me in person ide probably tell you... Also i had a idea of how to make a ton of money, selling somthing that was already invented, though i dont want to give that away eather.... hummmm... i wonder if it has a pattent in the USofA? well this paragraph is excithing!!!! ya\'ll dont have a clue what im talking about do you... man! k sorry about that.  Well Im runnin out of stuff to rambel about... and the soccer idea is sounding grood right now... so... im gonna go get some icecream... er... somthing... to admit it i dont know how to end a post...wow... its like a letter but worse... well thats all for now... though i feel like im forgitting somthing, ah well, more fo\' laters!   *isn\'t it cleaver how im working in all the \"this is how to understand me\" things, and giving you a greek lesson to boot!   '),(2834,3471824,'31,May,2004','2004-05-31 00:00:00','   And it is now math class. I just finished off a graph about crime rates, so I have a few minutes before I need to go back to class.  I\'m considering quitting MUDwolf permenantly. It\'s not THAT good a game anyways- there\'s so many out there that are far better- and it\'s rigged. Rigged games suck. Especially when it\'s rigged against you. I\'ll NEVER become an Immortal, even if I do reach level 50 someday! And it\'s because of that little @#&*ing b#^&# Saskii! If I ever became an immortal, she\'d whine and complain about not being one (she\'s been bugging magden for longer to become one) and if she becomes an immortal, chances are she\'d kill me. Repeatedly. In the game, of course. But it\'s no fair! I\'ve been working so HARD helping newbies and such, and Az, who never does anything but CHAT, just became one! I could scream for all eternity. What\'s the point of working to become a level 50 (and an automatic immortal) if I wouldn\'t be permitted to become one anyways!?   '),(2835,3471824,'28,May,2004','2004-05-28 00:00:00','   Just another really boring morning. Once again I find myself sitting here at this same computer, trying to pass the time between getting off the bus and my first class of the day. And yet I have never been able to shake the realization that my friends are off talking to their real friends outside in the halls. I can\'t seem to forget that they\'d rather be with other people.  MUDwolf is seriously unfair. I\'ve been helping snivelling newbies for gods knows how long now, and still I\'m not even considered for immortality. Whereas Azuriel, who does NOTHING but chat all the time (seriously), has now become one. I deserve it way more than she does! I actually DID something. But nooooooooo. Supposedly it\'s because of that b**** saskii...   '),(2836,3471824,'28,May,2004','2004-05-28 00:00:00','   Like I said already, I can\'t think of a title. Ho hum.    My life is it\'s usual depressing, irritating self today. One good thing is that I found out what I needed to do in Pokemon gold today (hey, the story and show suck, but the game isn\'t all that bad. Plus I don\'t have anything better to do), which is defeat the gym leader in Goldenrod city. I could\'ve sworn I already did that...but everything in the game itself seems to disagree. Ah well, it\'ll be a snap anyways. My quilava, Fred (I name all my pokemon monotonous, boring names for no reason at all), is around level 32, and all those people who challenge me have wimpy pokemon that are around level 10! Hah! Pathetic!  But enough of my gameboy rambling. It\'s probably boring everyone to death. Not that anyone would be reading this anyways... I wish I could fully retreat into my own world, just to get away from this torture called life. If this is all life is, then I don\'t want to live at all.  Whoever said \'better to have loved and lost than to never have loved at all\' was just trying to make themselves feel special. It\'s exactly the opposite!  Just a few more hours of school left. And then I can go home! And go on the computer! Of course, I\'m on one right now, but we\'re not exactly permitted to play games at school (well, duh. It\'s SCHOOL). Most people play games anyways, but that\'s usually when they\'re not in the classroom when the teacher is practically looking right at you. That\'s just asking for trouble. And I\'m not too fond of getting caught playing games and being sent down to the office and getting a lecture from the vice-principle who probably doesn\'t even know my name or remember me. So I\'m writing in my blog- hey, it\'s not like I have anything better to do. Oh, yeah, and I\'m writing a list of roleplaying stereotypes, just to vent some anger. Here it is so far: Roleplaying stereotypes/clichÃ©s/over-done things 1)	Assassins 2)	Ridiculously attractive runaway elf princess warrior â€˜maidenâ€™ sluts 3)	Killing for stupid reasons 4)	Vampires 5)	Big, strong; masculine heroes who go off on whirlwind adventures and save the world too many times to count; screw half a million women who practically worship him; seduce some virgin empress, sleep with and then abandon her; Has rippling golden hair, perfect teeth and loads of charm; Becomes rich in the process; is a strong defender of â€˜goodâ€™; and has some magical sword with a long and majestic name. 6)	Mary-sues in any gender 7)	Young, heroic, brave wizards with half a million females madly in love with them   Yup, that\'s all I\'ve come up with so far. And they all REALLY tick me off. Vampires and Assassins have to be THE most over-done of all those though. I mean, come on! In all the rps I\'ve ever been in, only a handful of other characters have been something other than a vampire or assassin! Seriously, can\'t they be at least a BIT more original!? It\'s not like I actually go LOOKING for vampire/assassin rpgs and join them just so I can whine about how many of them there are! The only good rp I\'ve ever been in with vampires or assassins is this one on Elftown called Lethargy. In that, there\'s a total of one human (Mortis), around ten vampires and/or assassins(Kula, Eve, Wulfgar, Gelina, Jahn, someone or other....), and only ONE character of another race- my character (Syth). And there\'s no-one playing an \'evil\' character, aka antagonist. NO PC ANTAGONISTS! NONE!! NONE!!! ZERO!! NIL!!! Z- Okay, I think you get the point. Even Mortis, who is-or was, I can\'t figure it out- a vampire hunter, has joined the \'good\' side with the rabble of un-original vampire characters! Well, then again, so have I. At least for now. If I feel like it later on, I might decide to throw a wrench in the good guys\' plan and join the \'dark side\'. My character is pretty flexible in regard to loyalties- nobody really trusts her much, and she trusts them even less. Oh, the joy of messing with other peoples\' plots- OH, SHUT UP!!!!!!! Ahem. Sorry. One of the people in the classroom was being sick.  Like I was saying, my character\'s pretty distrusting. That\'s partially because she\'s anti-social. In most of the rps I join, I use Syth as my character. That\'s because I have a really lousy memory, and it\'s hard to remember all sorts of different character names for a million rps (I join alot of \'em. Yes, I need to get a life), so by using the same character for all of them, it saves alot of effort. Syth\'s a half-elf. Well, technically more of a half-demon...   '),(2837,3471824,'27,May,2004','2004-05-27 00:00:00','   Hello. To those of you who bothered to look at this, congratulations. Not like anyone would ever look here anyways.  This is the first post in my blog (well, duh) so if it sucks, don\'t complain. I\'m not too good at this stuff (duh again).    I\'ve created this thing for the sole purpose of being able to say that I own a blog. Yup. Not really any reason at all. No, wait, I\'m bored, AND felt like having a place to go on and on about random things. Oh gods, I\'m booooored. And angry. And depressed. I need a new hobby. Or a gameboy. Or both. But noooooo, my parents won\'t buy me a gameboy, or a game console of any kind! I\'m lucky to have PC games. So I\'m saving up to buy a gameboy. Which is harder than it sounds. You see, with an allowance of 10 dollars a month and the only other cash gaining opportunity being babysitting my little sister, I have to save up for a long time to get enough money to do anything with. Sure, I guess I should feel thatkful I live in a family that is well off and such, but oh well. I don\'t care. I don\'t FEEL very thankful at all. And my life SUCKS. Sure, it could be worse. I could be kicked out of the house and have to move in with the telletubbies. But it\'d be pretty hard for it to get much worse short of a nuclear explosion nearby or something like that.   People don\'t make sense. I try so hard to fit in sometimes, I really do, but no matter what I do I just can\'t. I know I\'m not like them. They know it too, and hate me for it. Well, that\'s how they act. I really don\'t have a clue what they really think about me. But their actions speak of disgust, of anger. I can\'t belong to this world. I don\'t know how to cope with so many things that life throws at me, and it\'s cost me so much. I\'m beginning to think it\'s cost me my sanity.  Am I sane? Is this reality of mine some twisted path that should never be used? Is it \'wrong\' to think this way? Even if it is, I cannot help it. I cannot change too many things about myself. There are so many things about me that I have no control over. And it seems I will never be able to control them. Yet all the other people around me can manipulate them flawlessly. Why does the world seem so alien through my eyes? It\'s the only one I\'ve ever known, but is it fake?  Death. My biggest desire. I crave it\'s cold grasp, to be released from this shattered vessel and disappear into non-existence, where there is no pain. I want my life to be ended, and I want to be the one who ends it. And yet, I cannot. Something is holding me back from it. Fear. How ironic, that the only one preventing me from reaching my greatest desire is myself. My fear of pain keeps me from my death. Fearing pain is a human adapting, made to prevent us from potential danger. But this danger is what I WANT. I need the consequence of that action- after all, it is the only reason I would even consider the pain.  And so I live to see yet another day. Time passes strangely here in my own universe. A minute can seem like an eternity, while forever can pass like a second. So I don\'t know how long I\'ve been isolated. Perhaps I am immortal and have lived for ages, or perhaps my existence hangs by a single thread and at any moment I could vanish. But I do not know. How can one know such things about themselves? Is it just me who is oblivious to fate, or do others know what will happen to them in the future? I suppose I shall never know. This illness keeps me locked inside myself, so that I can never escape to catch a glimpse of the world outside my cell. I seem to be spiraling downwards in a Maelstrom that I cannot stop. But what am I falling into? I cannot tell if this is insanity or sanity that I am approaching so quickly, but it seems that I will reach it soon, no matter what it is.  Is this real? Is my life a dream, a fantasy world created by who knows what? How I wish this was all superficial, that it never really happened. But how can I know? Even if I found out what this world is, it would probably be just another lie someone\'s using to manipulate me. Somehow I know that, even if this is a lie, it would change nothing. Even if I discovered this life wasn\'t real, I couldn\'t do anything about it. Why should I seek the truth when knowing it would not allow me to do anything about it?  Right now, I take things day by day, not thinking about tomorrow. For if I tried to think of the future, it would probably overwhelm me. I try not to consider the things that have the power to bring my world crashing down around me, because I do not know if there is any other world I could live in.   '),(2838,3471824,'20,June,2004','2004-06-20 00:00:00','   Happy father\'s day. Woop woop. Yaaaay.  Just had breakfast. No coffee. Not that I drink coffee anyways. Unless it\'s french vanilla. That stuff rocks! I don\'t have anythign better to do, so I\'m posting. School\'s finally over, woot woot, and tomorrow (which is monday) I\'ll get to gloat, since everyone else in my family still has school (those of them that go to school).  I\'m going to be dragged along to one of my sister\'s soccer games today, all the way on the other side of the city. The only thing more boring than sports is watching sports! I swear, I\'m going to VEGETATE. Ugh. But nooooo, today\'s father\'s day, so we have to stay together today! I hope I can sue my parents for this... We had a party last night. A bunch of hyperactive kids were running around the house, not to mention annoying adults who took over the house and played bad music all night. They probably got drunk, too. And now I\'ll probably haveto clean up after the obnoxious brats who messed up our basement. Fun.   '),(2839,3471824,'10,June,2004','2004-06-10 00:00:00','   Why can\'t I ever think of a title? Ah well. Not like it\'s a very important skill. I mean, putting \'able to think up good titles on the spot\' isn\'t exactly something useful you could put on your resume. But who cares? I\'ll just type in some random word or something I haven\'t updated in a while. I was too lazy, I think... Magden\'s going to abandon MUDwolf and start a circlemud game. Good idea, in my opinion. There\'s alot of really fascinating areas and such on the internet for circlemud. In fact, I downloaded the circlemud area creater (or AC) 9Don\'t ask why, I was looking for furcadia stuff at the time. Furcadia\'s boooooring. You can\'t carry more than one item at a time!). Maybe I could submit an area or something to him if I create it using the AC..... Too lazy to write more.   '),(2840,3471824,'12,July,2004','2004-07-12 00:00:00','   ...and so I live to see another day. Just got back from ottowa two days ago, on saturday. At writer\'s camp right now at the university.  &nbsp;Wraith\'s here too. It\'s nice to actually know someone in the camp ahead of time. Usually I would just hang around a bunch of people who didn\'t know me at all and probably liked me even less. But I\'ve changed since then, and if wraith wasn\'t here right now I probably wouldn\'t even try to make friends. But she is here, thank the gods, so I don\'t have to. She\'s been a better friend to me than anyone ever had before. I\'m really going to miss her... &nbsp; &nbsp;I\'m dying. I can feel myself slowly fade away. By changing schools again, all my parents are doing is alienating me (they seem to enjoy doing that to me...) and tearing away what I\'ve become. For once, I actually know who I am, and they\'re taking it away. I don\'t think I\'ll survive losing myself this time...   '),(2841,3921381,'31,July,2004','2004-07-31 00:00:00','   I am so fucked!!! Once again! i tried signing up for GRE test. itonly offers freaking Subjects tests this fall in US. Gosh, i don\'t want the subjects test. i want fucking general test only!!!!!!!!! shit! i may have to just take freaking computer science subjects test. other disappointment: there is no October test, just November and so on. barely get my scores in at end of December.  Did i do something i wasn\'t supposed to? ok besides the little stuff, i meant something really Mean to someone? Why is this happening to me? ------------------------------------------ since i\'m all angered, i mind as well degrade myself and show my weakness. I am so freaking upset and i guess jealous. i have not been so fucking jealous since the younger years. i\'m going back to the ex-friend who out of nowhere is getting paid more than me now and she only has a high school diploma (barely).  my boss offers me a job as management BUT i need a degree (which i do) AND the fucked up part is that the low end starts at $23.41/hr. the high end starts at $29.88/hr. And when i say offered, i mean nothing fucking official. haven\'t seen it yet and its been months.  now, my ex-friend(no life, no ambition, no education, no experience (under my terms), no professionalism ) gets a job that requires ONLY 5 years of experience in a general area AND HS diploma. I could\'ve easily applied for it but being i am still fucking waiting..... anyhoo, she gets paid $22/hr. ALSO, she was picked over someone with more experience and education, now theres MIS-Management for you. They picked her because she kisses ass and they know her more than the other person.  this makes a fucking joke of me! working my ass off to get this job that someone with the heart and brain of an ant can also get. When does paid for performance come in? for work? for life? God, how does that work in with life? Satisfaction/goodness for efforts/hardwork you put in.  HOW DOES THIS HAPPEN? WE WORK FOR THE SAME COMPANY. HOW DOES ONE JOB EXPECT MORE AND PAY JUST AS LESS AS ANOTHER WHICH EXPECTS LESS?????? HOW!!!!!?????  thats where my jealousy stems from. besides that, my ex-friend is going to love being closer to me so she can torment me. i hate it. i could easily file a sexual harrassment lawsuit on her. i\'ve got one proof and one theory. damn bitch!  okay, hows that for honesty? am i really jealous? i\'m confused. i think i am a little bit, just a little bit jealous but more upset because she gets to torment me and upset at the system.  I AM BETTER THAN SHE\'LL EVER COME CLOSE TO.  God, give me strength to not allow this person to annoy me in the early dawn of my sleep and make me have to get up and blog.  okay-i\'m feeling better now, i\'m going to breathe and never mention this again on here. it will all be over as soon as i discuss it with my boss. i really am better. the curse words tend to give off more heat than intended. but is that really true? hmm. okay, no more about this. '),(2842,3921381,'30,July,2004','2004-07-30 00:00:00','   God, am I the bad guy? Why am I getting the short end of the stick? I was so happy for a while until now. Is this a test? I will pass it but  I have so much frustration and anger right now.  You sent me your Son and I\'m so glad I\'ve found my way to both of you! I know you hear my prayers. I know you feel what I feel. I want strength and I want courage. I want to see Jesus again. I love the feel of no fear, no anger, the feel of warm embrace, warm light and pure unconditional love.......... I remember you sent him because I called onto you. Come back to me.  Show me the way.  Amen. '),(2843,3921381,'29,July,2004','2004-07-29 00:00:00','   I called it in to the secretary. taking a sick day. Sick of people who don\'t do jack and get paid, sick of working and not getting recognized, sick of work politics, sick of mis-managment. damn, here i go again. okay, just read my previous post. i\'m not going to dwell on them.  today is catchup for me. bills to pay, people to call, movies to catch up on, and internet. awesome day it will be.  definitely, i\'m signing up for GRE and CBEST today. thats about $100 for both, i think.  got paid yesterday. damn, paid all my bills due this week and i still got money left over.. i double checked it. can\'t believe it. it better not come back and slap me in the ass. '),(2844,3921381,'29,July,2004','2004-07-29 00:00:00','   Blogger.com, you guys need to make a search feature.   It will definitely kill my bored moments (like now) when I get bored of typing blogs and clicking&nbsp; to Dashboard to check out my name on the recently updated list. It gets boring after the 4th time.  &nbsp; '),(2845,3921381,'28,July,2004','2004-07-28 00:00:00','   God?? ... hello?? .................wheres my millions? oh.. you mean i have to work for it. darn.  sorry for my curse words today and all the other days and for tomorrow.  now will i get my millions? '),(2846,3921381,'28,July,2004','2004-07-28 00:00:00','   God, make me a millionaire! '),(2847,3921381,'28,July,2004','2004-07-28 00:00:00','   i\'m taking time off of work. i\'m 100% paid for performance and this job isn\'t cutting it.   it seems to me the lazy ones who give less than 5% from day one, get recognition when they show 10% of work. Here i am giving 100% from day one and now they expect it.   TIP TO ALL: act stupid. if they know you\'re smart they expect it. There is less expectations when you act&nbsp;stupid (hopefully you\'re&nbsp;not really stupid and its just all a show)&nbsp;which means less work and fucking recognition when you actually work.   &nbsp;   &nbsp; '),(2848,3921381,'28,July,2004','2004-07-28 00:00:00','   i just found out today my freaking ex-friend who is a genuine lowlife-low self esteem psycho is getting paid more than me and will be working at the same site as me. IT WAS ALL FUCKING MISMANAGEMENT!! who has the flipping education here? who is more professional and more experienced in the workfield? ME!   first of all , i\'m not upset she\'s getting more than me. i only say those things because&nbsp; its part of another issue at my workplace. i\'m upset she\'s at the same address as me.&nbsp; She is crazy! i mean empty inside but filled with evil thoughts crazy.   identity thief. she wants to be like me. she took all my previous jobs after i\'ve been promoted. she emails me with sexual harrasment. she spreads rumors about me. she called my husband and told him i was fucking someone. she wants to take me down because i am better and she\'ll never get to happiness like me.  best part, she\'s commented on how nice my ass is and she is a bisexual. she really WANTS me but since she can\'t have me, she\'ll rest on stalking me.   and i\'m not a lesbian.   she likes to chew revlon nail polish and thinks of the perfect crime on raining weekends. this is why i\'m so afraid she\'ll do something to my car or house.   damn mismanagment. they hired her because the other person (&nbsp;who is well educated ) was not on their \"In-Crowd\" list.   &nbsp; '),(2849,3921381,'28,July,2004','2004-07-28 00:00:00','   so i saw my chiropractor. he says i\'m developing a ... damn, whats the word.... a.......... damn! anyhoo, something like arthritis in one of my neck bones. i need to see him 4 times a week and put ice on my neck.   what else..hmmm...&nbsp; &nbsp;i have an arch to archy in my lower back. that explains my big ass! well.. actually nice size but.. anyhoo.. back to my back. on the last back bone, the joint inbetween the two is too small.. maybe no joint at all. not normal. thats explains why i\'m short.&nbsp; missing 1/2\" of joint goo.&nbsp; \"birth defect\" he says. thanks dr.  anyhoo, of course i want to be better. i\'ll go for better health, long life. besides insurance is paying for it due to the accident. its all good besides my lack of good sleep or sleep......period.  Thank you God and Jesus for the fact that i can still walk, breathe, eat normally and have full use of my five senses!  amen.  &nbsp; '),(2850,3921381,'22,July,2004','2004-07-22 00:00:00','   does blogger.com offer free photo storage? do i have to find my own online storage/web host? someone tell me..........please....................  &nbsp; all of a sudden i\'m feeling like i\'m on an island and no one is hearing me. '),(2851,3921381,'22,July,2004','2004-07-22 00:00:00','   does blogger.com offer free photo storage? do i have to find my own online storage/web host? someone tell me..........please.................... '),(2852,3921381,'22,July,2004','2004-07-22 00:00:00','   okay.. i divided up two posts up because i hate reading long ass posts.  anyhoo, i wanted to purchase a computer and setup linux on there. I wanted to startup on my php and mysql stuff. Man, all this free time... no school anymore until i pass the GRE, its summer... its awesome; and i am so lazy. i wanted to learn all these programming languages this summer and fall but just too damn lazy!  &nbsp; '),(2853,3921381,'22,July,2004','2004-07-22 00:00:00','   haven\'t posted in a while. okay... what has happened?  first of all ... lets look at my goals again and look at my progress or lack of.  1. sign up for GRE test - getting the books. 2. sign up for CBEST test - thinking about it. 3. buy incubus tickets for august 13 concert - asked my sister about it again 4. get resume completed for upcoming job!! - started it. 5. work on design for ADA database for sale - started proposal today; partial 6. work up an idea/design to simplify ASAM reporting for sale - nothing 7. Laundry and clean house - yeah! did my laundry; cleaned some rooms.  hmm..overall, i show good progress!  i didn\'t do much at work today. mental day off.   went to dr yesterday. i told her i was taking my medicine but really aren\'t. going to my chiro today. getting review of x-ray\'s from accident last week and discuss physical therapy.  got my boss to buy me programming books so i can optimize my workload. i\'m all for the easiest method. of course it also has to be the smart method.  i plan to look over my GRE books tomorrow and see what i\'m up against and then sign up the test in October.  &nbsp;  &nbsp; '),(2854,3921381,'22,July,2004','2004-07-22 00:00:00','   okay i\'m bored and i need to do something besides the goals i set for myself in my previous. here some junk about me....   3&nbsp; Loves:  1.&nbsp; My Husband, Son and Daughter 2. My Family  3. My Self   3 Hates:  1. EX-Evil-Sabatoge-Dark-Never-Happy-With-Herself-Friend 2. Auto Accidents 3. overworked and underpaid   3 Friends4Life:  1. Macy 2. SSV 3. Cyn   3 Can\'t Do withouts  1. Email 2. Wireless Internet 3. Curse Words   3 Deadly Sins:  1. Sloth 2. Sloth 3. Sloth  &nbsp;  &nbsp; '),(2855,3921381,'21,July,2004','2004-07-21 00:00:00','   Finished my big report at work today! It Feels GOOD!!!  on to the next item in my TO DO bin...................  I got Wonder Woman DVD original TV episodes for myBirthday! i\'m so excited. i started watching at it and wondered how it ever got my attention when i was younger. it looks so lame now. anyhoo, i do still like her and i\'ll watch it. its just its so lame looking compared to this era tv shows.&nbsp; it was the bomb then though. and i did love it then!  --------------------------------------------------- okay.. goals!   1. sign up for GRE test 2. sign up for CBEST test 3. buy incubus tickets for august 13 concert 4. get resume completed for upcoming job!! 5. work on design for ADA database for sale 6. work up an idea/design to simplify ASAM reporting for sale 7. Laundry and clean house  &nbsp; '),(2856,3921381,'21,July,2004','2004-07-21 00:00:00','   never get into an accident. of course because of safety/health issues but also if you survive it, the doctor visits and \"WHO is Paying??\" is such a hassle! NEVER PAY OUT OF POCKET EITHER.  it took me almost 1 hrs to do paperwork and crap of who is paying. finally, we got it all settled. my friends insurance will cover my expenses and then they insurance will go after the other guy.  damn.. i\'m feeling headaches. need.....................rest................................ '),(2857,3921381,'20,July,2004','2004-07-20 00:00:00','   i\'m twenty freaking seven years old today. wow. another age closer to the hill and almost accomplishing most of my goals. 3 more years to attain my goals before a number divisible by 10 jumps at me.  what to do todat? appointment. i want to get a tattoo on my lower back. something religious. a cross probably. maybe not today. soon.  anyhoo.. i\'m gonna shower. i took today off from work so i can play \"sloth\" today. what fun it will be! '),(2858,3921381,'20,July,2004','2004-07-20 00:00:00','   damn, i\'m all sore. my back. my shoulders. my neck. my chest.  i threw up dinner last night. was it from the accident? my first thought was actually \"pregnant?\". i tested. i wasn\'t. good. because the impact from the accident would\'ve made me very upset if i was pregnant.   anyhoo, this morning, as i type (you read), my right arm (mostly), is feeling numb. what the hell is it!??  ughh.. i hate to be reminded of an accident. this shit better go away!   last night i slept on my left side and there was this sharp pain at my lower left side. pulsating pain.   i\'m going to see the chiropractor this morning. damn shit! '),(2859,3921381,'20,July,2004','2004-07-20 00:00:00','   ughhh.. my back hurts. i got into an accident today.  co-workers and i were going to lunch. we were on the left lane and the car infront of us was turning left so she came to a stop. my friend, the driver, came to a stop also. 2 seconds later, BAM!!! BAM!!! and there was smoke in the air from the air bags. some idiot rear ended us at a very good speed. i was in the back seat passenger side. my other friend, who was&nbsp;in the passenger side, and i were both in shock. my friend, the driver, was very upset because this may be like the 5th accident in 2 years, all not her fault, and the second cadillac she\'s had. she got out of her car and went to bitch out the guy who hit us. his reply to her was, \"you stopped in the middle of the street\". well, yes, because there was a car infront of us turning. had his idiot ass paid attention, we\'d all be fine!  on-star came on. there was also a detective who was around when it happened. he directed traffic so we could move out of the street safely. police man came. took report. took names.  my friend, the driver, got her left thumb smashed. maybe be from the airbag. my other friend had her legs crossed at the time so her impact was harsh. her left leg was bruised. the seat belt squeezed on her during impact. her left hip was hurting along with her back. i felt shock. i had numbness and tingling around my face and all over my body, mostly my hands. i was scared.   the two girls went on the ambulance. i went back to work with a staff, who came over after we called him. i picked up my car and went to my doctor. i started feeling pain on my left side:leg, arm, neck. then there was more pain around my lower head and neck area.&nbsp;6 hrs later and i\'m feeling pain along my lower back and shoulders. i made an appointment to see my chiropractor tomorrow.  i had to get the claim # from my friend and i also need her insurance policy. anyhoo, i hope everyone is ok.  now for the twilight zone moment: last year 4 days from today, we all had gotten into an accident with a county car also going to lunch.   as for my friends: as stated previously, the driver--> this is her 5th accident within 2 years the passenger--> this is her 3rd accident of the year.   next time we go to lunch, i\'ll meet them there.  as for me, i\'m hoping this whiplash does not come back every so often to remind me of the today\'s accident.  thank you God for watching over us so that we can continue to breathe. '),(2860,3921381,'16,July,2004','2004-07-16 00:00:00','   Well.. i said yesterday that i was waiting on some free tickets today so i could take my sisters to the concert. Missed out on the time. Damn! Anyhoo, i\'ll take them another time. if i have to pay for it, it needs to be someone i like too!  &nbsp; Work totally dragged. Today\'s friday. It always drags on fridays. '),(2861,3921381,'15,July,2004','2004-07-15 00:00:00','   Same ole routine. wake up. work. come home. ponder what to make for dinner. watch a movie. ponder what to make for dinner. dinner. sleep.  &nbsp; theres a concert tomorrow and i think i might go. rock alternatives! yeah!!! anyhoo, tickets free. just gotta go pick up at radio station. i\'ll probably take my gothic/rocker sisters. they love that shit. they go at it at from a high school standpoint. kids. &nbsp; okay.. i\'m home now and i\'ve already pondered what to make for dinner. now i think i\'m going to watch a movie. &nbsp; &nbsp; &nbsp; &nbsp; '),(2862,3921381,'14,July,2004','2004-07-14 00:00:00','   Today was so boring at work. I\'m effective as always but just tired. overworked and underpaid. i got myself an office and new furniture but no promotion yet. maybe that\'s why i\'m so bothered and upset. the promise of a promotion that may not take place anytime soon.   I think i\'m going to work on my rubiks cube at work during my downtime. got to get a hold of some algorithm and apply it in hopes of getting all the damn colors back on each side. Its do-able. Just have to have patience and i think a brain is needed for the algorithm.  What a goal! i\'ll be figuring out the cube in my dreams, i\'m sure. '),(2863,3921381,'13,July,2004','2004-07-13 00:00:00','   Why do ppl blog?  boredom? loneliness? no friends? too much time?  why am i taking up blogging? hmmmm................. '),(2864,3921381,'13,July,2004','2004-07-13 00:00:00','   hmmm.. got a blog site today. '),(2865,3921381,'14,August,2004','2004-08-14 00:00:00','   I decided Thursday that i am going to forget GRE, forget going back for my Masters......FOR NOW, and take a dive into Real Estate. Its hot here! And i\'m grabbing my piece of the pie while its hot! We have a new U.C. coming in 2005, houses are appearing out of nowhere, the city is expanding in all directions. Supply and demand is screaming at each other. ppl from the Bay are buying homes here. Besides, i plan on selling my own home when my 2 year mark comes up and that\'s next year Christmas 2005.  I won\'t quit my current job though. i plan on starting real estate part time, then depending on how it goes, i\'ll contract for my services at my current job. i know they\'ll do contracted services because i have my competitive edge and they need that.  I am successful in most things that i do, so i have a clear picture of what i will get out of this career. I only plan on doing this for about 2 years.   my motivation is my drawing plans of a nice home on acres, landscaped with japanese garden and a fish pond, a nice huge library and a indoor rock climbing area in the house. It\'s going to be awesome.   i\'m so damn materialistic. next week, i\'m prepping for the test. ambition is such a wonderful thing! Its what drives me to the next level.  wish me success! '),(2866,3921381,'11,August,2004','2004-08-11 00:00:00','   Truth is that I have a promotion lined up. It\'ll happen before this year is over (Fingers crossed).  Anyhoo, my boss made a compliment today: \"It\'ll be a challenge but with you leading the charge I\'m sure things will go smooth.\" That made my day.   My reply was \"Should I go Hitler now or wait until everyone returns from vacation?\"  She told me to wait.   Moral of the story: There\'s a new Hitler in town and she ain\'t blonde blue-eyed. Goes by the name of BITCH. '),(2867,3921381,'08,August,2004','2004-08-08 00:00:00','   I want to say, i am mildly guilty of this. I like to think of it as Ab-Exercises.  You\'ll love this!   urlLink Oriental Chick '),(2868,3921381,'08,August,2004','2004-08-08 00:00:00','   urlLink    LOL. okay, first x-rated on my site. I don\'t think blogger.com has a problem because i see that they have other \"more visual\" blogspots.&nbsp; urlLink '),(2869,3921381,'07,August,2004','2004-08-07 00:00:00','   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ I got a new car stereo the other day in my car as the other one broke, and it has voice activation software installed into it!  If you say, \"Rock!\" it switches the settings to play rock music!   If you say \"Classical\", it switches settings and plays Classical music!   On my way home last night some kids ran out in front of my car.   I hit the brakes and yelled \"F*cking Kids\" and the radio started playing Michael Jackson. ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ '),(2870,3921381,'07,August,2004','2004-08-07 00:00:00','   Real Estate a Job for Me?  My good friend starting working in March and up-to-fucking-date she has made $80 grand. Now how stupid am I to be sitting my ass here and not taking a dive myself? Pure Logic comes to mind. I think it may be that I am scared and I think about my fucking $2000 mortgage and my other bills.......oh! and food. I\'m thinking about doing it anyways sometime down the line because i plan to sell my home in 2 years by myself.    GRE  I\'m taking the GRE test so now I have to shape up and get back into studying. I\'ve been so laxed for the summer, its miserable (in respect of the GRE). ughhhh... unfortunately procrastination won\'t work in this deal for me. btw, i\'m a procrastination at heart when it comes to school. The GRE looks terribly hard. What kind of words are contumacious, peregrination, loquacious, munificence. People don\'t use these words!!! not me anyways. ugghh.. and MATH!! i\'d forgotten how to do math. of course i had to take Calculus for my major but that was a long time ago. I find myself forgetting, at times ONLY (not all the time), what 8X7 may be. My son usually will confirm or make an idiot of me with the answer. God help me thru this one! '),(2871,3921381,'05,August,2004','2004-08-05 00:00:00','   urlLink    Here\'s a nice picture of Shi. &nbsp; urlLink '),(2872,3921381,'04,August,2004','2004-08-04 00:00:00','   urlLink    I like this art. It makes me think fearless, dangerous and sexy. This is David Mack\'s comic character Kabuki.&nbsp; urlLink '),(2873,3921381,'04,August,2004','2004-08-04 00:00:00','   I knew you listened, God!  My boss tells me they will place my ex-friend at another location! Halleluja! (or however you spell it). I feel good. I realize now that all that smack i talked about her getting paid more than me was mask. i was really hating on her being anywhere close to me. she\'s locust, leech.  anyhoo, back to the stuff that counts in life... my friends and I went to the bar after work. i was only able to take down half of my margarita. i had an empty stomach and that made things bad. Then i totally forgot there was a birthday party i had to go to. Just got back a few minutes ago. damn eye contacts still bugging me. must be the wind. its so dry.  Got Paypal? Check this out:  urlLink http://www.settlement4onlinepayments.com/  - $50 from paypal! Gotta Claim if you got Paypal. I haven\'t checked it out yet though. '),(2874,3921381,'04,August,2004','2004-08-04 00:00:00','   I have my eye contacts on and they are annoying the crap out of me.  anyhoo....  what happened today? watched  13 going on 30  at work today. played at work today. didn\'t work at work today. the beauty of it all is that i have my own office. ha! so nice sometimes. i hope my boss never reads these posts. i\'d be in alot of trouble.  i went to the tattoo store. so rude. after looking at the samples they had, i waited at the register to ask some questions. a kid walked right passed me. it\'ll be a while until i go back. i didn\'t find what i wanted anyways: too flashy, colorful and juvenile. its true, you tell more people about bad experiences than you do good experiences.   i just found out today that sleeping on your stomach is a no-no. especially for my neck after my accident. i never knew it was bad. its so comfortable at times though. '),(2875,3921381,'03,August,2004','2004-08-03 00:00:00','   As I browse thru the recently published blogs, I\'m thinking what annoys me the most is when ppl type using lower case and uppercase in one word alternating in no freaking order giving them an impression of a gangsta/punk rock profile. how annoying. actually.. more like how high school. sorry if i offend my friends who do this.  i used to do this myself. what an idiot i was in my younger years. off the topic a bit but i used to wear baggy ass jeans and wear dark lipstick and throw dumb ass signs thinking i was all bad. God, how pathetic i was.   o! i just remembered.. totally off the topic... i was smoking at work with the ladies. they laughed their asses off at how i held the cigarette. this marks my second attempt of smoking. once back in the younger gangsta days and now. believe it or not, my love is second hand smoking. and yes, i\'m aware i\'ll die faster.. thats why i don\'t intentionally go out to do it. however, it must be like when ppl have a fetish of smelling stinky toes or something. LOL.... read this correctly. i\'m not saying i like smelling toes. i\'m saying it must be like that. '),(2876,3921381,'02,August,2004','2004-08-02 00:00:00','   i figured out this Hello program! wheww hoooooo! okay, now just got to figure out how to get my picture to post on my profile.  if anyone knows, let me know. thanks! '),(2877,3921381,'02,August,2004','2004-08-02 00:00:00','   urlLink    sexy.&nbsp; urlLink '),(2878,3921381,'02,August,2004','2004-08-02 00:00:00','   urlLink    beautiful&nbsp; urlLink '),(2879,3921381,'02,August,2004','2004-08-02 00:00:00','   urlLink    She is so sexy. I\'m thinking about getting a cross tattoo like hers.&nbsp; urlLink '),(2880,3921381,'02,August,2004','2004-08-02 00:00:00','   I think there are ghosts in my house. (twilight zone music.....) Its a new home and we\'ve been here 7 months now. I feel different now.   perhaps the second month, i saw this white figure coming out of the garage as we were coming home. I thought it was my husband and I said something to it. It took maybe 2 seconds for me to realize it wasn\'t him. But I wasn\'t scared. Felt like it was watching over our house or us. From that point on, i believe if you saw something at the corner of your eye, you did see something.  Now its different. now i feel something. nothing evil.. i hope. yesterday i thought i saw something and i\'m afraid i did.   Why do they hide under the bed or in the closets or in dark places? My sister in law has the gift to see ghosts. She says they sometimes test her to see if she sees them. They would sneak up on her and see if she reacted. Spooky.   She says everything in The Sixth Sense is true. Scary shit, if you ask me. '),(2881,3921381,'02,August,2004','2004-08-02 00:00:00','   returned back to work. it was alright. talked to my boss about my exfriend. Boss said i should\'ve told her before. The situation would\'ve been much better than now. i agree. but who knew she was gonna get hired. They will try to find her another location. Worse case, i gotta suck it up and watch my ass. This is all in the past now in THREE..................TWO.....................ONE.  okay, new topic.  anyhoo, i didn\'t turn on the A.C. today. kind a hot. i\'m really feeling like taking my top off. getting sweaty here. i\'ll turn on the fan. '),(2882,3921381,'02,August,2004','2004-08-02 00:00:00','   took another sick day. i didn\'t want to go back to work. went fishing with the family instead. caught one striper.   i actually wanted to clean house or decorate kitchen but didn\'t. i\'ve been thinking about going out to buy a religious book so i can read about God and Jesus.   nothing much happened. '),(2883,3921381,'01,August,2004','2004-08-01 00:00:00','   Cancer\'s horoscope for today.....  \"Remember that you don\'t always have to echo the opinions of those whom you admire. Someone will greatly respect you for speaking your own mind. Be careful when dealing with others, as a miscommunication is very possible. So long as you keep your eyes and ears wide open, you should be able to calm a storm before it gathers momentum. Others may seem aloof, but realize that it may just be that Cancer sensitivity coming to the forefront. Since family is important at this time, rather than letting your feelings be hurt, retreat instead with the ones you love. No one else really needs to know about your internal conflict. Put it behind you, and move forward.\"  only rare occasions when horoscopes sometimes do match up with your life. '),(2884,3599631,'15,June,2004','2004-06-15 00:00:00','   So you can take a look at my desktop by clicking the purty little image below:    urlLink  '),(2885,3599631,'12,June,2004','2004-06-12 00:00:00','   So I got a  urlLink Gmail  account from  urlLink Fraser . So now you can send me an email at  urlLink jldera@gmail.com . However, my main email is still the one I\'ll read most often. Gmail seems like a nifty interface, we\'ll see how it works a bit.  '),(2886,3599631,'11,June,2004','2004-06-11 00:00:00','   So it used to be that  urlLink Google  was just a search engine. Now they keep coming out with all of these other cool technologies based on searching. For example,  urlLink Froogle  allows you to search the web for low prices on pretty much anything. Of course, this whole  urlLink Blogger  thing is Google-owned now.  The real fun, however, is over at  urlLink Google Labs . That\'s where you can find all the cool stuff before it comes out. It\'s like going to an auto show and seeing the concept cars that just never get to market. Though a lot of Google\'s stuff comes to bear. I\'ve been using Froogle for a while, it just recently got onto the Google homepage.  A pretty neat tool over in the Labs is the  urlLink Web Alert  tool. I don\'t know about you, but I\'ve been known to do some  urlLink egosurfing  every once in a while. With Google\'s Web Alerts, i don\'t have to do it by hand anymore. Instead, the  urlLink results  are emailed to me. A new item comes up, I get an email about it.  So now I\'ll know when someone is talking smack about me, without ever having to look for myself. Computers: making life easier since the 60\'s.  '),(2887,3599631,'11,June,2004','2004-06-11 00:00:00','   So, a little bit about myself. My name\'s Jason Deraleau. I\'m in my mid 20\'s, 24 to be exact. I was born in the year of the Monkey and am a Gemini (at least, according to my  urlLink Blogger profile ). I live in western Massachusetts, which is located in the northeastern United States, which is located in the western hemisphere, which is located on the third planet from the sun, which is somewhere in a vast amount of void and matter known as the universe. You may have heard of it, you may not. I make no claims of knowing my audience.  I work with computers. I\'m an IT administrator for a small business in Connecticut (somewhat south of that whole Massachusetts thing). My day to day involves managing several different servers of various operating systems. Going down the line, I have two running Windows 2000 Server, two running Mac OS X Server, one running Linux, and one running FreeBSD. I like to consider myself an equal opportunity OS employer. However, I honestly don\'t like Windows. I\'d say I don\'t like Microsoft at all, but I have an Xbox and I have to admit it\'s a pretty good gaming system.  My true passion is for the Macintosh. While I would never have said such as recently as four years ago, in September of 2001 I bought my first Mac and I haven\'t looked back. I sometimes joke that Macs never existed before OS X, but I know that\'s not true. For example, I have an original Macintosh that has been hollowed out for use as a fish tank. There aren\'t any fish in it, mind you, because I can\'t seem to keep anything besides myself alive. I fear for my children, wherever they might be. Just kidding, I don\'t have any offspring. I think many people are thankful for that.  No, switching to the Mac was probably the smartest thing I\'ve ever done. If I hadn\'t, I wouldn\'t have my current job, I wouldn\'t have written any  urlLink articles  for  urlLink O\'Reilly , and I wouldn\'t be typing this on a beautiful 17\" Apple Studio Display. I think I\'ve taken the whole Mac thing to an extreme though. I\'ve become quite a hypocrite, but also quite an evangelist. I love to be proved wrong and for every bad thing I\'ve said about Apple in the past, I\'m sure I\'ve said at least two good things about the Mac maker. And I have plenty more good things to say about Apple, the Mac, and other technology stuff, so stay tuned.  '),(2888,3599631,'11,June,2004','2004-06-11 00:00:00','   To be honest, I hate the word  blog . It just doesn\'t roll off the tongue very well. Blogggggggg. It\'s like saying blah, and we all know how dismissive  that  word is. In any case, I\'ve signed up for  urlLink Blogger . Sure, part of it is to see if I get any  urlLink Gmail  invites, but I\'ve been debating creating a more professional weblog that others can tune into.  Ya, I do have a personal  urlLink one  over at  urlLink LiveJournal . Plus I have  urlLink another  on the  urlLink O\'Reilly Network . That one\'s pretty professional, but it\'s also pretty Mac-centric. As much as I love the Macintosh, I do keep up with other technologies and such. I don\'t update the ON one very often, either.  So here we are, gentle reader. Enjoy my musings!  '),(2889,4096204,'31,July,2004','2004-07-31 00:00:00','   Okay.. This is my first blog ever. I know i\'m so not with the times. I\'m like in the stone age. I\'m so not a computer person so let\'s see how all this shit goes ay. Well i actually wrote some shit just then but yeah.. I did something which i have no idea what and yeah deleted all the crap i wrote..ahhh!! This always happens to me. I\'m so born not to touch computers.. haha! Okay. It\'s Sat nite and guesse what Shermaine\'s doing?? Nothing!! Yeah it\'s a hell fucking moodless day for me today. Practically slept the whole day. I came home bout 1-ish from work last nite. I stayed late cuz we were all just hanging out and all. They wanted me to go over to one of their place\'s today, and yeah i didnt\' go cuz i was sleeping and also i\'m kinda afraid of getting close to any of them cuz they are my boss anyways and yeah i really don\'t want myself shining through too much cuz i might get fired. haha! But don\'t get it wrong.. I\'m not a fake person infront of them at all. I\'m just drawn-back a tiny bit. But yeah they\'re awesome people and we all get along real well so that\'s all good. Guesse what?? I might meet Chris tommoz. Well, we were meant to go on a road trip.. Road Trip!! But yeah I don\'t know.. things always get cancelled with us. But that\'s cool i understand i guesse.. I don\'t really know Chris as well as i wish i did but nah he\'s a sweetie. He\'s outgoing and he\'s hot!! ;) haha.. I don\'t think tomorrow will happen but we\'ll see how it goes. If not, I\'d prob go to the city wit my friends from my old high school. Haven\'t seen them in agess!!! It\'s weird.. i thought i\'d never lose contact with them but it\'s actually really hard cuz everyone\'s doing their own thing and yeah it\'s not easy. But yeah i will definitely make an effort to catch up with them every now and then. We got to! Okay, John Lennon\'s on telly and i wanna watch it. '),(2890,4056798,'31,July,2004','2004-07-31 00:00:00','   Well, its me back again from San Francisco. A lot happened there, not a lot of big things but a lot. So the last day of school, i got A\'s on both of my finals and we got our grades, an A too, second highest in the class, my brother is first highest. Another thing: We were the only two to get A\'s in the class. And you wondered why i couldnt update! So..it was really sad leaving and i have pictures on my website now! I know! i konw it seems like forever! But the thing is...i love these pictures and there in perfect organization. Links and everything. Much love to Zack and to Sarah, and to Mr. Parrot and to Alex and Chi Chi and Yue-Cheng, and Paige...we all love you doll! So...anyways we went to San Francisco where they (the dentists) became happy with sticking very scary sharp objects in my mouth and things that were really loud. I really didnt like it and i kept thinking i was going to faint. I just get like that. So then after that we went to our house in Daly City, all peachy. And then the house in San Francisco...very nice little (NOT!) house there. The renters are keeping it nice. Then t buy groceries where last time i went with my dad we saw this guy that played on the Kings basket ball...could get his Autograph though...Grrr : (. Anyways...then we got our haircut! You would love my hair...its so tiny and with my glasses i look like a freaking genius!!! Anywhoo! So...then we went and ate and this is the funny part. I was totally bored...stuffed to the brim about to fall asleep and this guy comes in with this other guy... I dont know, friends or something....so then what happened is that i went to go use the restroom it was a two hour drive back home. So...i went and when i came back i stared at the guy. He had a freaking bible with him and he had glassed that were plastic like mine. So i stared at him and i guess he thought i thought he was hot, which he SO was not. So anyways...he stared at me and smiled and said something. I didnt understand so i just kinda smiled at him. I sat down and then i realized he had said \"Hey precious.\" Freaky deaky. I walked out of the resteraunt with my mom cause my dad was eating fish and that takes forever in a day. So...then what happened was my mom and i were about to cross the street when these gang bangers go by and they throw a glass bottle on the floor. It exlodes and i kinda take a step back. What idiots. Oh yeah, earlier i had seen this hobo women with a REAL Louis Vuitton bag, i personally thought it was hillarious. Cause she goes around with like dirty pants begging for money and she looks like she gonna die and in her hand a huge ass Louis Vuitton bag that costs more than she will ever have...she was holding it like it was her life. So..today i woke up with a fat headache, rememberd i was going Sushi-ing with Alexandra and Poof there goes my headache! So...we went to Sushi...i ate WAY to much. So much i thought i was gonna puke afterward. And the fact that we went out after for a cup of coffee that wasnt really a cup cause it was a venti white chocolate mocha from starbucks ( i cant believe i tossed it!!!) made me nearly gag. So...with the coffee in hand, cool haircut, neat glasses, loose faded jeans, carefree walk and black UCD sweater....we entered this hip little music store and sat down to listen to music, we were some fly little crackers! i felt totally sophisticated until i thought i was gonna puke. All the raw fish got to me. Oh yeah, huh! We went to Maren\'s mom store...a little funky. There was this TINY purse smaller than my palm and it had 20 on it. I went up to buy it...i mean tiny little purse made out fake leather...come one. twenty cents right? It was fity percent off. Of from 40 cents. I go up with my dollar...guess what? Its twenty bucks cause its \"From Italy\" Shoot man...i can kill me a cow in italy, make crappy purses, dye them, ship them over to the US and sell em for over a hundred bucks. These people are f-ing CRAZY! Not to mention they have a door smack in the middle of the damn store cause it shows off the too expensive hangers. Ridiculous. So...then after the music store, we went to the Salvation army...it was really cool. Even though it stunk so bad...there was some really good books. Then we went to go watch The Village. The movie...its not that bad, but if youre looking to get scared...dont go watch it. So anyways..after that...we went to Gap and we tried on some skirts and some pants and i found these pants...low on the waist cropped pants. THEYRE MINE! fifty bucks and this cute little hat and scarf...my birthdays coming up...*clears throat* Mommy, daddy? Wanna add that to the one hundred shopping spree at Pac Sun and the brand new dell labtop? Please...so anyways then we went to borders and we read some and then i came back home plopped down and watched Starsky and Hutch...hillarious and then im going downstairs to finish off the night with The Godfather one...i think i know my daddys new birthday gift... the godfather. Why are all guys obsessed with that damn movie anyways? So...heres to me and my two and half weeks of summer!!! You all know you love me!  '),(2891,4056798,'29,July,2004','2004-07-29 00:00:00','   Well, today is done meaning only one more day of school which is a total blast for me! I cant wait till its done except for the fact that then i wont be able to see those people again which makes me really sad! I really love these people! All of them...okay minus one or two or three or four or five or six...you get my point. Anyways Zack is getting nicer and sweeter and more into my personal bubble space. Alex is being a dumb ass he likes to take and touch people. He enjoys coming up to you and touching you. Touching your hair and your hands and your face and anything on your face and other things. he enjoys to take your things. Pencils, erasers, journals, books. Freaky huh? Well bryant didnt make it. He went to UCD Library, or he went to the MU. And he stood there and stayed there i suppose for too long. Like he left or something. I feel really bad i kinda didnt want to hop on the bus. He is all by himself over there....i feel awful! Anyways...tomorrow there is another final. Today there was a final, it was kinda sad though. Because you know what? There were only three questions on the test. The test was worth a hundred points, and talk about disaster if you missed one of the questions...!! Ouch! Anyways...tomorrow another final. I have a damn 88% in the stupid class. Im doing ec today...and a kick ass job in the test tomorrow. I mean hey, last time i came out with a 100% on the midterm. Who says i cant pull it off again? Zack was being awfully funny today. He stole my cell phone several times and then he sat on me and he yelled at two people who were making out to get a room...lol. What a doll! Anyways, i must go! Wish me Luck!  '),(2892,4056798,'28,July,2004','2004-07-28 00:00:00','   Dont we all love Barbie? Yeah...id say. Id write a lot except for the fact that i already wrote a lot but deleted most of it. Actually all of it. i think that being a webdesigner has gotten to my head. Anyways i am free in two days. So this freaky deaky mexican girl asked my brother to go with him. Well not go with him, like she wrote a note to stephanie her mexican friend and in it she said that she wanted to go with bryant...like date him. Freaky huh? It scares me to think that she wants to go with bryant. Which brought me to think...if my dad thought it was cute and funny that she wanted to go with bryant...what would he think about me? So i told him and then he stayed quiet and he said \"What?\" and then i started laughing and said just joking! Isnt that so unfair? my parents said that they would kill the guy who would want to date me! Lol, i find that kinda funny. SO i talked to paige today, shes a real doll! Lol, no pun intended! She is really sweet and not all that pretty but she has a face that you have to look twice at. like literally. She is just...so interesting. I love it. So anyways, the doll (lol) started to talk to me about highschool and stuff and how great it was and how i should take this class and that class and this and that. I loved it! No one except for her barbie guards can talk to her. I loved it! She was awfully nice until Zack started to bug her! Isnt Zack so cute! Anyways...it turns out that she is mexican and that she speaks spanish fluently. Yeah..thats right. She understands everything that we say! Uh-Oh! Anyways...people think that im half chinese. Im sorry but these people are just screwed up! I cant believe them...me chinese? They have the weirdest impression of asian people. They probably thought i was asian cause i was on top of everything...like knowing all the answers...thats how asians are. Always the best...its the truth, i know tons of people who want to break that style. But no...not me im not asian. Im full bred italian and spanish. You know you love me though!  '),(2893,4056798,'27,July,2004','2004-07-27 00:00:00','   Its so lovely to be one with the computer! its even better knowing that i will be able to go on the computer for the rest of my life and that in just three days i wil be able to go and work on my website. So, today Zach and i got to giggling like crazy. The father figure was upset with us. He was really clingy though and we were cracking up about capitol barbie. The truth of the matter is. That i see so much of her in me. Or me in her. Something  like that. I just see so much becuase you know what? the truth is that if i were in her position i would act like her. I talked to her with bryant today. Shes okay, tells the truth. She got three percent higher than i did on the test. Its okay, i wont kill her for it. I still love her like crazy. lol. Well anyways, i really should be going, working on homework. Schools almost out so that means new layout! Isnst that nifty? Anyways...new layout most probably. But without paintbrushes...im ultimetly nothing! (Spelling) Love you mucha!  '),(2894,4056798,'08,August,2004','2004-08-08 00:00:00','   Well nothing has happened lately...i have been awfully bored and doing nothing at all. Ive been sleeping a lot lately. Jo has gotten back today! Im so happy..but i cant pick her up, gabby\'s back from LA to and next friday im going major shopping with mommy! I almost have fifty buckaroos to buy those very cool Gap Capris. Woke up at five today to deliver news papers...must feed cat. Pellet has been getting Pissed off lately. We cleaned out the garage last night, i cleaned out the bird bath and Cindy asked me to baby sit! Money Money! My b-day is coming up soon to! Hurrah!!! HURRAH! im almost fifteen and school is almost starting. im blabbing like crazy...so please dont think im weird. Well...love you...but must run! BYE!  '),(2895,4056798,'01,August,2004','2004-08-01 00:00:00','   Well..its the first of the month...exciting huh? Not really that just means that its closer to school starting all over again. Really sad, i know...im actually really depressed...i wish that i could be back at summer school with all my friends...i really miss them all. I mean it would be nice to get in a get together again..im listening to Faith Hill, \"Breathe\" Exciting stuff huh? I love Davis and all, but it has NOTHING on San Francisco...i wish i were in San Francisco. I love San Francisco, its all i have. Its the only thing that i know that i can hold onto it. And then ofcourse some damn terrorists will come and blow it all up. I have to go...i dont really want to write right now...im probably gonna go finsish a layout i started and never finished. Okay then...and then code dolls again...i miss all my friends so much! I want jo and kristin back and zack and jennifer...i dunno.  '),(2896,3539593,'28,June,2004','2004-06-28 00:00:00','   Finally found the time to sit down and write again.  Nine-to-fiving it is not my cup of mint-tea, but I\'m slowly adapting (is that a good or a bad thing?).  Anyways, much has happened in these first three weeks that needs proper rambling to be put into place.  Let\'s see...  LARPed over the weekend.  If you are among the vast majority of sane Americans who don\'t know what that is then shame on you for not opening your mind, broadening your horizons, and all that other uncreative crap.  www.masi.us should give you an idea of what I\'m talking about.  So I spent the weekend in the woods, rain soaked, mosquito drained, bruised, bloodied, and shamelessly covered in mud - in otherwords: in absolute bliss.  I find it funny how I like spending a weekend busting my ass for the enjoyment of others better than getting blitzed and hanging out at a party.  I always knew college life wasn\'t my scene, but it\'s frighteningly easy to be sucked into believing that it is what I should like, even now that I\'ve gradumutated.  '),(2897,3539593,'11,June,2004','2004-06-11 00:00:00','   So I\'ve settled in to my first week of work and I just discovered that I\'m psychic.  All day I\'ve walked into the room just as someone said my name, or picked up the phone the second someone called me.  I was thinking of writing a letter to someone and they appeared in front of me.  Needless to say, things are good. The apartment\'s great - my roomates, likewise - and my neighbors are quite friendly.  I think this is going to be a great summer.  At least it\'s been a great week.   I was a little bothered that I wasn\'t going to be able to make it to my brother\'s bachelor party - I still have two more weeks until I get paid, and I\'m sufficiently broke to be unable to make it down to boston.  But true to uncanny form, the party I was fretting missing got cancelled.  Things have a funny way of working out. So I\'m making dinner with Kate and Laurel tonight - two classy chicks who redefine the meaning of cool.  This came as the end of a long held phobia of mine, asking a girl out to dinner.  Don\'t get me wrong, I\'ve been in numerous relationships of all levels of intimacy.  I\'ve just never felt comfortable asking someone out to dinner.  Requesting casual sex hasn\'t posed a problem, but, there\'s I\'ve always been reluctant to ask for their culinary company.  This may say something about my priorities. Anyways (stalling for thoughts)...   work has been good so far.  I like my job, picking up where I left off last summer after I got laid off during the school year.  Nothing\'s more embarassing than being laid off from a campus job.  Well, I can think of a few notable exceptions, but it still sucks. So I\'m here, back at college but without the homework.  Just how I like it.  I can tell the group of people I\'m living with (there\'s five more or less living in my apartment now) are going to be a load of fun.  We\'re already planning the 50 Drinking Games in One Night Party.  *Sigh*  Irresponsible young adult life is good.  '),(2898,3539593,'05,June,2004','2004-06-05 00:00:00','   Drove to Portland with my mother today to get measured for a tux for my brother\'s wedding.  The man working at the suit-store surprised me with his grace and professionalism.  I have seen few people in my life more respectable or confident in what he was doing.  Made me feel older, more professional myself.  Funny the places inspiration strikes.  It\'s utterly surreal that my brother is getting married this summer.  He\'s always been somewhat distant from me, but then I feel I\'ve always been somewhat distant from my life in general.  These brief islands of meaningful family events crop up every now and then, but it\'s far from the photo-album impressions of adulthood I had in my youth.  I guess this is part of my passage into manhood, letting go of the dream that I\'ll suddenly wake up one day and know what\'s going on.  I hate terms like passage into manhood and coming of age, but I can\'t help but chuckle at the truth of them when I come out the other side.  Whoever is writing this story of mine sure has a funny sense of irony.   The highlight of an otherwise disappointing (1:5 odds my ass) and depressing day (good-bye, Gipper) had to be the discovery that I know how to make pizza.  This noble task had always fallen to my brother when we were children, but I surpised myself and my parents for reinventing the not-quite-from-scratch art of pizza making in the Powell House.  Beats the Crispy Chicken from Burger King I had for breakfast.  For those of you who don\'t know me, I am one of the few people I\'ve ever met who has a bitter dislike with breakfast foods.  Muffins, bagels, eggs of all textures and viscosities, toast, fruit, pancakes, waffles, and worst of all breakfast ceareal - long have these morning delights plagued my strange and twisted stomache.  When I get up in the morning all I want is a burger.  Is that so unusual?  If you were starving in the desert which would you rather, a bowl of cold, mushy, sugar-satured grains, or a pound of sizzling beef draped in the warm embrace of cheese?  I must admit, I was a part-time vegetarian during my college years, but the thought of breakfast foods for breakfast still ran chills through my bones and waves of nausea through my stomache.  Anyways.  I\'m off for an icecream sandwich and some tums.  Maybe that\'ll cheer me up.  '),(2899,3539593,'04,June,2004','2004-06-04 00:00:00','   And how!  For the first day in over two weeks, there\'s sun.  Beautiful, life-giving sun.  And a nice cool breeze.  I feel inspired.  I even vacuumed this morning.  I must be coming down with something.  Breaking in the dancing feet again.  There\'s nothing quite like the guilty pleasure of blasting \'NSYNC and Backstreet and busting out some new moves across the family-room floor when you\'re home alone.   urlLink Tom Cruise  would be proud.  I seem to have gotten the spring back to my step again.  The knees aren\'t aching any more.  There\'s nothing worse than creaky joints before you make it half-way to middle-age.  (Side note: I seem to be going hyphen-happy lately - weird)  I just wrote something interesting, but lost it when I went to  urlLink dictionary.com  to look up the word \'sans\'.  I knew I knew what it meant, but I wanted to be sure.  Part of my newly blossoming, self-imposed OCD I suppose.  And now I\'ve gone and lost a paragraph that I was excited to write.  I mean, how many times do you get to use \'sans\' in casual conversation?   Paragraph sans excitement. Paragraph sans point. Oh well.  I find myself utterly paralyzed, staring at the trees blowing in the wind.  I thought it chintzy when John Travolta did it in  urlLink Phenomenon , but there\'s something to be said for simple pleasures.  Idle tree watching would be tops on my list.  Followed closely by supermanning into my super-soft futon, slipping on socks fresh from the dryer on a chilly day, mint lip balm, and of course killing kittens.  (Side note: I would never, in fact,  urlLink harm a kitten  - chickens on the other hand, they just creep me out and I\'m happy to  urlLink commit mass chickencide  - you see I knew this flock of free range chickens when I was in Australia - they were awesome - anyways, my favorite chook somehow hurt hurself and started bleeding when suddenly her carniverous band of merry hens decided to peck her to death and fight over her vital organs - if that doesn\'t convince you  urlLink this will )  Good God, I can\'t believe I\'m up before noon.  I just don\'t know what to do with myself.  I suppose I should go out and enjoy the sun while it\'s decided to make an appearance, but I\'m afraid it\'s just a hoax.  Mother Nature luring us outside so She can dump two more weeks of rain on our heads.  Just to punish us further for living in Maine.  I can\'t wait to get out of this place.  But I will miss it.  '),(2900,3539593,'03,June,2004','2004-06-03 00:00:00','   urlLink    Just think about it.&nbsp;  '),(2901,3539593,'03,June,2004','2004-06-03 00:00:00','   So, I finally came to that place. You know, the ultimate post-show let down.  Only the show lasted four years and was kinda called my College Education - or more hiply - my College Experience.  Like four years of putting up with cynical (but loveable) yuppies wasn\'t enough, now I either have to become a corporate potato or flee the country to avoid 40k plus in college loans that I *slightly* remember taking.   I think the biggest thing I\'m going to miss about old Bowdoin will be the food.  I loved that dining hall.  1st in the nation, damn straight.  That and the bathroom.  I will always be fond of a person that cleans my puke from the sink and smiles at me the next day.  Joyce, you are my personal angel for that.   So that place I was talking about.  You all know it.  Sitting at home.  Wondering how you survived four years of this through highschool.  Feeling like a stranger in the one place you\'ve called home all your life, but which now has the dull name of your Parent\'s House.  And everyone keeps asking you in hushed tones if you\'ve thought of your future yet, like it\'s some sort of insider joke - boy you must be asked that a lot.  Well no.  I haven\'t been asked that nearly enough.  And certainly not enough from the sort of people who would inspire me to actually want to find my calling in life.  You know the type of person I\'m talking about; the ones you ask for a cup of sugar and they somehow get you to join the Peace Corps.   I\'m ranting now, trying to sound like a real writer whose long-winded tirades somehow worm their way into your heart and make you feel all fuzzy inside.  I hate and love that shit, much like most of that shit out there I guess.  But where does that get me?  Four years of putting the liberal in my liberal arts career (we\'re talking Women\'s Studies, Government, Buddhism, Italian, Spanish, Native American spirituality, even a class on Japanimation and Manga!) and out of all of that, I think my greatest classes were in cynicism and lethargy.  Is this really what yuppies do for fun?  Am I doomed to analyze a life that I only half live in moments of lucid daydreaming?  Can I think of some witty or poignent end to this, my first entry, one that will let me go back to obsessively checking my email and wishing someone would come on instant messenger?   Or is this, as the movies asks, as good as it gets?   Well, I doubt it.   But if I\'m going to start living that life I\'ve dreamt of, might as well start here at home.  I guess I should dig that Computer Science diploma out of the glove box where I left it graduation day (I still don\'t understand a Bachelor of Arts in Computer Science) and put it to some sort of use.   Even if I couldn\'t afford the $100 frame for it.  '),(2902,3539593,'27,July,2004','2004-07-27 00:00:00','   Alright. So I\'ve been avoiding writing in my blog for too long.  I tried to be all clever and witty and expressive with my writing, but it just turned into a cheap immitation.  So I stopped writing.  I put my head down and tried to plow through this summer, as if something better were waiting on the other side.  Some other time I\'d have the inspiration to begin writing again.  Well, bullshit.  This is not a work of art, or at least that should not be its driving force.  This is my conscience, my voice, and my mirror.  I don\'t even think there\'s a single person out there who reads it, so it\'s about time to put these words to good use - that is, to start making some consistent sense out of my life.  I\'m always afraid to write about my philosophies - they sound so trite and self-riteous when I go back to read them later (I think trite and self-riteous describes most of philosophy for that matter).  Writing through self-consciousness is bit like shooting through a slump in basketball, only way to get your rhythm back is to keep at it.  So that\'s what I\'m doing.  The best thing I ever did for writing was keep a \'throw away\' notebook.  I made myself just write whatever, not worrying about its quality, all the time knowing this was just excercise - nothing I would want to keep for later.  It gets me out of my head.  Softens my focus, so I\'m not so worried about the product, and more invovled and interested in the process.  It\'s also really good for organizing my thoughts.  Not necessarily so I can sort out specific thoughts, but so that my mind becomes comfortable with following through on moments of inspiration.  It\'s a matter of tricking my mind into getting out of the way so my self can come through.  I like that description.  Anyways, that\'s what this was supposed to be, and that\'s what it shall be.  Maybe eventually I can stop writing about my writing and get on with my writing.  Here\'s to perfect simplicity.  '),(2903,3539593,'07,July,2004','2004-07-07 00:00:00','   Something beautiful just happened:   I was looking for apartments in San Francisco and came across a message written by a woman out there.  Something about it inspired me.  Here\'s the response I wrote:   ...I couldn\'t shake the feeling of laying on my back in a field of daisies, coaxing the clouds with my eyes into giant whales and majestic sky ships, sails billowing to carry them off to some distant horizon.  Lately I\'ve been reluctantly shutting myself off from lazy afternoons in my mind, trying to focus myself on a world that I can only half imagine somewhere out in the mysterious \'there\', where careers are made and my debts repaid.  I used to love weaving little stories, little patchworks of pictures and poems, little rhymes like daisy-chains to set on my head as king of the clouds, happy on my throne of words.  But the words left me.  My education sapped from me the nectar I squeezed from life, the simple pleasures I had idly sipped upon, oblivious of how good I had it.  Somehow the belief that I must reach for more - become something more - sucked the air from my sails and set me to dreaming of distant cities filled with busy people with busy jobs and busy lives.  But like the clouds I kept these visions far above me, distant, safe.  I stayed close to home, pushing my way through college in the town in which I was born.  Under the watchful wisdom of the white pines of my home I forced myself into a culture and a society that turned my stomach and suffocated my dreams.  Somehow I limped through, earned my piece of paper that qualifies me fit for fine society, and arrived here - half-way to middle age, searching for a new place to settle down and begin my second life.  This morning I was searching for an apartment, a career, a lifestyle in a city where I thought I could see myself living.  This afternoon, I found myself in a field of daisies.  I don\'t quite know why, but I feel like after reading your post I woke up - not from a dream - but into a dream, a familiar dream where the edges of the world are soft and forgiving and open to gentle breezes and light touches.  I don\'t know who you are, but I wanted to give to you my deepest thanks for reminding me how to see.  May peace, wisdom, and grace always be with you, A young man from a distant coast.  '),(2904,3539593,'07,July,2004','2004-07-07 00:00:00','   I\'ve done it; I\'ve decided where I want to live and stake out my fortune.  San Francisco seems to have it all: crazy stuff to do; crazy, openminded people; crazy weather; and crazy money to be made.  And crazy suits me just fine.  This comes after a near-miss with an impending depression that forced me to take a sober look at my life (well, I wasn\'t exactly sober at the time, but perhaps a sobering look at my life).  My brother got hitched over the weekend (beautiful ceremony, reception, and all - Virginia sure is purty) and his friends got me blitzed.  14 hours of riding cramped in the back seat of our Impala brought me back to my parent\'s house.  I\'m not sure which is more contricting, being sealed in the back of a car with a raging hangover or being secuestered at my former home without license or wheels to set me free.  '),(2905,3463885,'18,July,2004','2004-07-18 00:00:00','   Finally hitting a dry track Phantom Rose put in a much better performance. Carrying 119 lbs. she finiished 5f in 101.99.&nbsp; After todays performance there is hope that her mothers talents have been passed on. However there is still a long way to go. She will now go into a training program to bring out the best in her.  '),(2906,3463885,'14,July,2004','2004-07-14 00:00:00','   A 2 year old filly out of Quest for Glory and Cup Day.  She has hit the track twice at Lone Star but was unlucky hitting a muddy and sloppy track.  So far she has run sixth and seventh.   This has not put me off as I know somewhere within this filly are the abilities of her mother. I will find them a draw them out.  She is now at ELP and will cross our fingers for a dry track. She races again on the 18th of July.  Stay tuned we will hear more of this filly.  '),(2907,3463885,'20,July,2004','2004-07-20 00:00:00','   If you are after cheap claims keep an eye on CBY. Most will be in the $4 and $5 claimers. I won\'t tell you they are good because if they were they would not be there. But, perhaps there is something that you may want to work with.  '),(2908,3463885,'30,July,2004','2004-07-30 00:00:00','   Unfortunately Power and the Glory had to be a late scratching in the Scuylerville &nbsp;due to not enough training. Now she has transfered to Del Mar in hopes of lining up on the 7th in the next Grade II.  '),(2909,3463885,'23,July,2004','2004-07-23 00:00:00','   In Full Flight&nbsp; urlLink     urlLink  '),(2910,3463885,'17,July,2004','2004-07-17 00:00:00','   With high expectations for her this race, it was not to be. She started slow and when she had room to run I thought she would pick it up but instead DR Flores ran her straight into the back of Ramathorn. And then she was done for the day. I don\'t think i can take much from this race as I noted Easy Rhythm finished behind her. &nbsp; She is one tough horse to manage as she gains stamina at 9-10% per night. Her works say she doesn\'t run as fast with blinkers but perhaps it\'s time to try. Or perhaps I need to remove the shadow roll being as her best race to date was without it. &nbsp; So now it\'s off to SAR to contest the Grade II over 6f for fillies on the 28th. &nbsp; &nbsp;  '),(2911,3463885,'14,July,2004','2004-07-14 00:00:00','   Power and the Glory is a 2 year old filly which was born in January 2004.  She has now hit the track 5 times and been in the money every time. In her second start she put up a track record at Lone Star over 4 1\\2f on the dirt.  In the Hollywood park  Nursery  she came in third place.  She then traveled to Belmont to win the  Fashion  in fine style.  Her latest start was in the Tremont where she placed third not being able to match it with the boys.  Best workouts : 3f on dirt 34.1     Lone Star                 6f on dirt 1:12.2   Hollywood park  Her next race will be to contest the HOL Juvenile Championship - Grade III against the boys. Her light weight should play an important role here. Stay tuned for more.  '),(2912,3463885,'30,July,2004','2004-07-30 00:00:00','   In an expensive lesson Papa Bear scorches the turf in The Read Handicap - Grade I. After that race Brew will search for races just against 3 year olds. I guess he is just not ready to take on the older horses. It has been a valuablelesson though. Next up may be a turf race at SAR on the 9th.  '),(2913,3463885,'14,July,2004','2004-07-14 00:00:00','   urlLink    Brew....A legend in the making&nbsp; urlLink  '),(2914,3463885,'08,July,2004','2004-07-08 00:00:00','   A 3 year old named Brew started his racing career for the JHamilton stables in November 2003. Jhamilton immediately gelded him early in his career.  He started his career with a track record over 1 1\\8 on the Hollywood Park dirt in a maiden.  He then went on to win 2 of his next three races in allowance company.  Then it was on to  The Caltech Handicap  with a concincing win over 1 1\\8 on the turf carrying 122 lbs.  He then had concentration problems that led to failure in the SA Derby.  He then ran second to Knighthood in the  Coolmore Lexington   who has gone on to bigger and better things.  He then headed to Churchill Downs in the  Crown Royal American Turf   over 1 1\\16th on the turf to record a track record.  Then came  The Will Rogers   He ran fourth after not handling the pure speed in the race.  And then came the biggest race of his career,  The Belmont Stakes  He raced very well on the dirt against some of the very best 3 year olds going around at the time. He was able to run fourth.  And the most recent race was The Cinema BC Handicap back at Hollywood Park. The poor performance in this race can only be explained away due to a bad day at the track.  In another big test in his career and hopefully a return to form. Brew has been sent to DMR to contest The Read Handicap - Grade I. On arrival today he was sent out with the trainer to put in a 1 1\\8 workout on the turf. He was a little sluggish due to being over rested from travel but he did the trip in 1:51.4. Its not a great time but he will improve as get closer to race day.  '),(2915,3463885,'30,July,2004','2004-07-30 00:00:00','   In his 3rd race with 2 wins under his belt Arwon tackled the Sanford over 6f. He ran up to his ability finishing in 1:13. He managed a third place earning him $22.50 and 2 valuable BC&nbsp;points.  '),(2916,3463885,'27,July,2004','2004-07-27 00:00:00','   With the next auction approaching I gathering together my few decent horses and putting them aside. All the rest will be put in the auction on the 16th of August. probably up to 20 3 yr olds. and probably about 50 2 year olds.  '),(2917,3463885,'20,July,2004','2004-07-20 00:00:00','   For the first time in a year I have&nbsp;pulled most of my entries and have an almost empty entry list. It has come the time for a real stable assessment. &nbsp; For the next couple months stable horses will go into intensive training to see what they can really do. The best to DMR, average to ELP, and the rest to CBY. The only entries for awhile will be a maiden or two and stakes races for the best in the stable. Last night about 25 horses came back into work from the farm. Nothing spectacular from their first work.  &nbsp; I have&nbsp;sent about a dozen no hopers to the glue factory. And yet I still have 125 horses with at least a dozen more on the way at the end of the month.  &nbsp; Will be a quiet couple months but can only improve some of the decent stock in my stable.&nbsp;  '),(2918,3463885,'11,August,2004','2004-08-11 00:00:00','   Might and Power  - 4 year old gelding  Snowy                    - 4 year old gelding Kiss My Goodbye- 4 year old gelding Sun Rays               4year old gelding Squeeze Me Tight- 3 yr old filly Bananas in Pajamas 3 yr old filly Hopes Flame 3 year old filly Makybe Diva 3 year old filly Comanche 3 yr old gelding Van Der Hum 4 year old gelding Vo Rogue 3 yr old gelding What a Nuisance 3 year old gelding Rogan Josh 3 year old gelding I Did It 3 year old gelding  Because the new triple crown system was not in effect when I generated most of my 2 year olds. Any 2 year old with a January birthdate is BC nominated as well as Triple Crown nominated.   Power and the Glory  2 yr old filly  Carbine  2 year old colt  Arwon  2 yr old colt  Fields of Omagh  2 year old colt  Nimble Foot  2 year old colt  Phantom Rose  2 year old filly Evening Peal 2 year old filly Australia Day 2 yr old colt Dominate 2 year old colt Downtown 2 year old gelding Love to Dance 2 year old colt Oak Boss 2 year old colt Red Light 2 year old colt Roastin 2 year old gelding Ice Line 2 year old colt Frosty 2 year old gelding  Hopefully there may be something of interest.  '),(2919,3463885,'23,July,2004','2004-07-23 00:00:00','   Nosing out the competition&nbsp; urlLink     urlLink  '),(2920,3463885,'08,July,2004','2004-07-08 00:00:00','   Established July 2003  My stables comprises 130+ horses of all ranges of abilities and experience.   urlLink Brew    urlLink Phantom Rose    urlLink Power and the Glory    urlLink Stable News    urlLink Claims    urlLink Auctions  '),(2921,4149935,'17,August,2004','2004-08-17 00:00:00','   As for the list of CD\'s that I hope to get sometime soon: (2 hours later.)  Regulate...G-Funk Era-Warren G. Last Of A Dying Breed-Scarface Streets Disciple-Nas (When it is released) Illadelph Halflife-The Roots Heay Mental-Killah Priest Step In The Arena-Gang Starr Juvenile Hell-Mobb Deep The Chronic-Dr. Dre ##########-Snoop Dogg Power In Numbers-Jurassic 5 Only Built 4 Cuban Linx-Raekwon Uncut Dope-Geto Boys Ready 2 Die-Notorious B.I.G Liquid Swords-GZA/Genius Ironman-Ghostface Killah One For All-Brand Nubian Speakerboxxx/The Love Below-Outkast (May God Have Mercy On Me) Naughty By Nature-Naughty By Nature Love And Hate-Aceyalone Do Or Die-AZ Enter The Wu-Tang (36 Chambers)-Wu-Tang Clan The World Is Yours-Scarface Black On Both Sides-Mos Def The Infamous-Mobb Deep A Wolf In Sheep\'s Clothing-Black Sheep The Diary-Scarface Wu-Tang Iron Flag-Wu-Tang Clan Blackout-Methodman AND Redman Dare Iz A Darkside-Redman Tical-Methodman Things Fall Apart-The Roots 2pacalypse Now-2pac Strictly 4 My N.I.G.G.A.Z-2pac Don Killuminati: The 7 Day Theory-2pac Moment Of Truth-Gang Starr Hard To Earn-Gang Starr Daily Operation-Gang Starr No More Mr. Nice Guy-Gang Starr Resurrection-Common Quality Control-Jurassic 5 Later That Day-Lyrics Born Strictly Business-EPMD In Full Gear-Stetsasonic It\'s A Big Daddy Thing-Big Daddy Kane The Great Adventure of Slick Rick-Slick Rick Return Of The Boombap-KRS-One Keep Right-KRS-One Straight Outta\' Compton-NWA Geto Boys-Geto Boys Grip It On That Other Level-Geto Boys Sould Food-Goodie Mob Blazing Arrow-Blackalicious NIA-Blackalicious A2G-Blackalicious The Last Generation-Shyheim Criminal Minded-BDP Edutainment-BDP 3 Feet High And Rising-De La Soul Art Of War-Bone Thugs N Harmony Born Again- Notorious B.I.G Words From The Genius-GZA/Genius Operation Stackola-Luniz Dirty Weaponry-Killarmy The Last Shall Be First-Sunz Of Man Uncontrolled Substance-Inspectah Deck Life Of A Kid In Da Ghetto-Ed O.G. Original Stylin\'-Three Times Dope Wooden Leather-Nappy Roots The End-Three 6 Mafia Here Come The Lords-Lords Of  The Underground Mecca And The Soul Brother-Pete Rock And C.L. Smooth Aquemini-Outkast (God Have Mercy On Me Again) Ain\'t A Damn Thing Changed-Nice And Smooth 19 Naughty III-Naughty By Nature Return To The 36 Chambers-Ol\' Dirty Bastard Dead Serious-Das EFX Black Mafia Life-Above The Law Organized Konfusion-Organized Konfusion The Sun Rises In The East- Jeru The Damaja Critical Beatdown-Ultramagnetic MC\'s I Phantom-Mr. Lif Labor Days-Aesop Rock Deltron 3030-Deltron 3030 The Black Album-Jay-Z Both Sides Of The Brain-Del Tha Funkee Homosapien Paid In Full-Eric B. And Rakim Operation Doomsday-MF Doom O.S.T-People Under The Stairs Enta Da Stage-Black Moon Capital Punishment-Big Pun True Meaning-Cormega The Album-The Firm The Blueprint-Jay-Z  '),(2922,4149935,'05,August,2004','2004-08-05 00:00:00','   It\'s Been A While    Well, I haven\'t posted in a long time, so here ya go. A list of all my CD\'s, to let you know the style of music I like. Not much, but I have to prepare the Mockumentary Script a lot before I post it. Besides, this isn\'t just about my movies. It\'s about ME. (Eerie music in background coupled with strike of lightning.) They aren\'t in any order.  Title-Artist-Year Made  Illmatic-Nas-1994 In God We Trust-Brand Nubian-1992 Thug Life-Thug Life-1994 Whut? Thee Album-Redman-1992 The Ownerz-Gang Starr-2003 Brand Nubian-Everything Is Everything-1994 Tical 0: The Prequel-Method Man-2004 KRS-One-KRS-One-1995 Beneath The Surface-GZA/Genius-1999 The Low End Theory-A Tribe Called Quest-1991 Tical 2000: Judgment Day-Method Man-1998 I Am...-Nas-1999 It Was Written-Nas-1996 The Love Movement-A Tribe Called Quest-1998 The Collection Volume One-Bone Thugs N Harmony-1998 God\'s Son-Nas-2002 R U Still Down? (Remember Me) -2pac-1997 All Eyez On Me-2pac-1996 Greatest Misses-Public Enemy-1992 NIGGAZ4LIFE-NWA-1990 No Way Out-Puff Daddy-1997 AmeriKKKa\'z Most Wanted-Ice Cube-1991 Reachin\' (A New Refutation Of Time And Space) -Digable Planets-1993 Thug World Order-Bone Thugs N Harmony-2002 E. 1999 Eternal-Bone Thugs N Harmony-1995 Me Against The World-2pac-1995 Hits, Rarities, & Remixes-A Tribe Called Quest-2003 Stillmatic-Nas-2000 Queensbridge Finest-Various Artists\'-2000 The Lost Tapes c. 1989-2pac-RELEASED in 2000 Better Dayz-2pac-2002 Life After Death-Notorious B.I.G-? TRU-Master P-? Watermelon, Chicken, &amp; Gritz-Nappy Roots-2002 BTNHResurrection-Bone Thugs N Harmony-2000  Well, that\'s it. All the CD\'s I currently have. Soon, I\'ll post a list of the CD\'s I am going to get, and you can FEEL FREE to buy them for me.    No, seriously. Do it.  '),(2923,4149935,'05,August,2004','2004-08-05 00:00:00','   The Official Blog Of Levi S. Fairbanks    A \" Novel \" Idea  Written  By  Levi  S.  Fairbanks    This blog was created as a way for me to tell the world about a movie. This is know normal movie, it\'s not a fancy hollywood production, and it\'s not at all pretty. It is, in fact, a REVELATION. The official documentary .  Life And Times In McMinnville   Ah, yes, Levi Fairbanks\'s directorial debut, an epiphany of pure scientific genius. Not for sell anywhere, and a rare find, but if you have the luck to get your hands on a copy, you, too, will be able to revel in it\'s pure brilliance, such a gem that you might have to wear sunglasses while you are watching.  Anyway, this is LSF, the creator. This isn\'t just a blog for my movies, it\'s also my INNER-SANCTUM.  '),(2924,3856641,'29,July,2004','2004-07-29 00:00:00','   MUSIC:&nbsp; Radiohead -  The Bends  MOOD:&nbsp; Exhausted and Wired  Okay, so I\'ve seriously been addicted to   urlLink The Legend of Zelda: Wind Waker   lately.&nbsp; It\'s not funny how much I\'ve been into it, but it\'s a fact of life.&nbsp; I told someone the other day that I didn\'t want to start it because&nbsp;I didn\'t have the time to play a stupid game that was going to monopolize my time.&nbsp; I was told \"it\'s not a pet!\"&nbsp; That person doesn\'t play the v-games.  Looks like things are looking up for Camp Booze.&nbsp; The price went up, but it looks like it\'s just going to be a little less crowded than we expected.&nbsp; A cabin in Canaan Valley should be an interesting time.&nbsp; Video game systems have been outlawed.&nbsp; We will have cards and we will be playing Texas Hold-em, I guarantee that.&nbsp; Oh yes, we will be playing.&nbsp; Study-up now, assholes, cuz I\'m comin\' atcha full force.  Word association time.  Boobies. Milk. Cattle.  I\'m stuck.  '),(2925,3856641,'28,July,2004','2004-07-28 00:00:00','   urlLink My Cat Hates You   Look.  We all know that cats hate people but are total moochers.  They rub on your just for you to change their crapping boxes or feed their food-to-poop converters.  Finally, a website that proves what all of us have been thinking.  '),(2926,3856641,'28,July,2004','2004-07-28 00:00:00','   Sex in the City on TBS is just plain wrong.&nbsp; I mean wrong.&nbsp; And I think I know someone that agrees with me...   urlLink   Is it really that wrong?&nbsp; Yes.&nbsp; [At a church] \"We\'ll get there...&nbsp; we just have to have a little  faith .\"&nbsp; *Puke*  It seems like everyone in my life is sick in one way or another.&nbsp; That includes myself on Monday.&nbsp; Taking sick days at work are annoying.&nbsp; Some people have sore throats.&nbsp; Some people are going to doctors.&nbsp; I want to say something clever right here, but the sound of that damn TV show in the background is really playing with my brain...  ...in ways I\'ve never known could be possible.&nbsp; Is it epilepsy?&nbsp; Is it the sickness from earlier this week?&nbsp; I feel like I could get in a fight with Carrie, Miranda, Charlotte, and Ho-bag and LOSE.&nbsp; That\'s right, there clothing would distract me and they would burst my ears with conversation filled with irony and puns.&nbsp; Puns.&nbsp; I\'m feeling nauseous.  Let me think...&nbsp; Oh yes!&nbsp; We have Camp Booze coming up soon, don\'t we?&nbsp; Let\'s pray that we rally&nbsp;the troops and more people don\'t back out.  Speaking of backing out, if I see Carrie break up with Mr. Big one more time, well, then I guess that means that I\'m  urlLink watching  this next week, too, huh?  '),(2927,3856641,'25,July,2004','2004-07-25 00:00:00','   MUSIC: Soundgarden -  Down on the Upside   MOOD: Kinda hungover  BEVERAGES OF THE EVENING:&nbsp;Corona Extra, Bud Light, Miller Lite, Vodka and Tonic, Bombay Sapphire Gin and Tonic,     You feel 10 years too young to be in here? Why? Because Margaret Thatcher is over there having a good time?  -- Brian Leone on the crowd at Coyote Ugly in DC    Starting with the fact that my ATM card\'s magnetic strip no longer exists on most of the card in my wallet and getting physical cash was&nbsp;impossible.&nbsp; Evidently, it\'s impossible for stores to run an ATM card manually as a debit card transaction for cash back.&nbsp; That would have been helpful knowledge 24 hours ago.   Wow. What an interesting night. That\'s the easiest way to describe it, I guess. Interesting. Last night, my friends and me were really trying to give our buddy, Aaron, a great time for his bachelor party. Only problems were, in my opinion--and how can I put this delicately...    BACHELOR PARTY RULE #1:  Only invite your very bestest and closest friends.   Don\'t get me wrong, everyone there were pretty cool dudes for the most part. But when people, who ARE NOT the bachelor, are getting sick and throwing up in alleys and falling asleep at tables by the second bar of the night... well then we have a problem. It\'s hard to have a show Aaron a good time when you are babysitting and wrangling up disrespectful assholes.    BACHELOR PARTY RULE #2:  Don\'t invite more than 10 people who fit RULE #1.   It\'s gonna be a rough night. That\'s a given. The delusion of responsibility of showing the bachelor a good time dwindles to nothing-per-person. Who\'s job is it to keep everyone/everything on track? When you have 18 people and someone tells you that the group is leaving, then you leave or get left. HOWEVER, if someone who is just not having a good time and isn\'t man enough to suck it up and hang out for the bachelor\'s sake (who was having a great time)... Wow I\'m really pissed.    BACHELOR PARTY RULE #3:  The Best Man (or other designee) is in charge   Too many chiefs and not enough injuns will get any situation into a spiral of testoserone-filled cock fights. Mine\'s bigger! NO WAY! I can\'t fit mine into a 2 Quart can of baked beans! He says when you stay, he says when you go, he says where we are going next. Input is welcome, but if he declines it, then fucking tough. Wow. Still pissed about this one, too.    BACHELOR PARTY RULE #4:  Call ahead to every place that you plan on going that night.   We brought a large guarantee to any business that the group that is about to arrive at their club is going to spend an assload of money. Deals are available for such groups. However, they need notification of the event and the group size. We went to a normally-classy place in DC (yes, it was a strip club--great, now I can\'t run for office). But when I walk in the door (second to last in line) with 17 other guys to find out that I have to pay $20 for my admission... Well... I\'m not happy. Consider that the last time I was there a couple of months ago for another friend\'s bachelor party we got in for the price of $ZERO and we had 10 people or so.   Let\'s do the math: 18 people x $20/person = $360 for admission. That\'s&nbsp;fucked up. When we sat down at this long dance-table,&nbsp;all the girls swarm us.&nbsp;&nbsp;It became quite apparent that&nbsp;management had totally fucked up the operations of the club. We find&nbsp;out by several employees of the club--once we started complaining and refusing to pay for crap forced on us--that no one goes there anymore. $15 per shot? We\'re outta there. I don\'t want to point fingers or say don\'t go to this club, but don\'t. The club is called  urlLink Nexus  and it\'s located in southeast DC.&nbsp;My friends and I have scratched that place for all future bachelor events. I suggest you do the same. Period.    BACHELOR PARTY RULE #5:  The night is 100% for the benefit of the bachelor.   If you can\'t follow this rule. Don\'t go. It\'s just respect, asses.   BACHELOR PARTY RULE # 6 : Bear and Jeff should not be allowed to talk to one another.     He just compared Democrats to Socialists.&nbsp; That conversation has lost all coherency. --Mike Defede on Bear and Jeff\'s inane politics.  Thankfully, I think that Aaron had a really good time and we masked a lot of what was going on for him. Brian was doing a really good time of letting the specific assholes involved that they were not being cool. Sometimes I wish weren\'t as nice as we are and could have dumped some folks off at a taxi stand.   Whew! It may not sound like it, but all-and-all I had a really good time with some of my close friends.&nbsp; Aaron was in full force and seemed really happy that all his friends were there, so I suppose it was pretty successful.&nbsp; Hanging out on the back porch recapping the evening&nbsp;with Brian and Travis after everyone had fallen asleep in weird positions on couches, floors, and other peoples\' beds.  '),(2928,3856641,'24,July,2004','2004-07-24 00:00:00','   urlLink DecemberTeen Girl Squad   If you haven\'t seen Teen Girl Squad, then you don\'t understand summer fashions.  News flash, people, but summer is seriously on the brink of being over already.  That means that you are going to have to start wearing long pants and long skirts again...  Then how are you going to snare cute boys?  Good luck.  Seriously.  Good luck.  Don\'t you worry.  This ain\'t over yet.  Like rotten bananas, I\'ll be back in one way or another.  '),(2929,3856641,'22,July,2004','2004-07-22 00:00:00','   MUSIC: Metallica -  Fade to Black   Dear Diary,  I\'m really starting worry.  All the kids in class are going out with each other and I still haven\'t even kissed anyone. I\'m ashamed to not be cool.  They made fun of my Jordache shoes again.  I could die.  But not really, cuz that would suck.  On another note, Suzie McMasters asked me for my number again.  She wants help with her science project and to get advice about Tad Dudley (who sucks a big one! Just kidding Tad).  I think she likes me.  Like a friend, not a boyfriend!  Plus, I\'m not interested.  I should tell her that I\'ve had a crush on her since second grade, but she\'d probably call me a creep and stop talking to me.  Crap, I hope Tad doesn\'t read this.  Oh yeah, I forgot!  I got tickets to see C & C Music Factory at the  urlLink Ohio Valley College  auditorium!  I paid $15, so they better rock!  If you want to learn the words, go  urlLink here .  Gonna make you sweat!  '),(2930,3856641,'20,July,2004','2004-07-20 00:00:00','   MUSIC: Jude -  No One Is Really Beautiful  MOOD: Yearnful  You can always tell--your eyebrows look sad-style and you just can\'t seem to get the lips in an upward position.  I\'m not sad or depressed, just...  Well...  Bored.  If you can get up the nerve to just get out there and do it, then that makes you really ambitious, right?  So what does it mean when you can\'t seem to break out of your lightning-fast decision to just put it off until tomorrow.  I wouldn\'t call myself a procrastinator, really, well, kinda.  An interesting discussion I got into with Sara last night was about what drives a person to become an entertainer.  Is it ego?  Sure.  Is it a desire to be accepted?  Probably.  But here\'s the interesting question...  Why do some people perceive those people striving to be more than a bedroom balladeer as a braggart, a show off,  insert your own negative word here , etc.?  Sure, entertainers have to have some conceit to believe they are good enough for people to actually enjoy listening/watching their art.  But why are some shot down before they even get started?  I say let them get out there.  Support their desire.  The only thing that listeners/watchers can do is  be honest .  If they suck, don\'t tell them that your life has now changed from this moment on from their performance when in all actuality, you were in the bathroom for the whole second-half because their  art  induced rabid diarrhea.  Suck it up and tell them that your butt hurts because of their works and that you want your money back so you can pay for the expensive prescription drugs you are now going to be addicted to.  Which brings me to another point I\'m extremely angry about.  Why in the hell do I know William Hung\'s name?!  Assholes.  '),(2931,3856641,'20,July,2004','2004-07-20 00:00:00','   Music:  Whiskeytown -  Midway Park   Just to lead it off, my  urlLink friends  and I are heading off to the mountains of West Virginia for a little cabin get away a few weekends from now.  A few of my friends from my hometown (4.5 hours away) are excited for me to meet Jesus at the cabin.  I thought my friend may have become a born-again and the retreat was becoming more of a recruitment than a drunkfest.  Damn.  However, I was mistaken.  It is merely another guitar player with whom I\'m supposed to \"possibly combust.\"  I don\'t know what that means, but that\'s what I was told.  How was my  urlLink grammar  on that one?  You may have just witnessed the first time I\'ve attempted to use \"whom\" in a sentence properly.  Actually, the reason I brought this up is because I had a flashback of when I was really young (probably age 6-8) I was at a church with a family friend and some stranger took me aside and asked me if I wanted to let Jesus into my heart.  I was totally freaked out, but I went along with it cuz I thought it sounded like a good idea.  He made me close my eyes and pray something.  When I re-opened my eyes, the freaky dude seemed pleased and let me go back to playing with the other kids.  I remember trying to  feel  Jesus in me--like a physical change or something.  The only thing I think I felt was a touch of paranoia that He might be watching me a little bit closer now that I gave him the key to the house.  Were they trying to brainwash me at the time?  Is that a normal church occurence?  I\'ve asked around and no one seems to have gone through a similar experience.  Explanations?  Advice?  '),(2932,3856641,'18,July,2004','2004-07-18 00:00:00','   Music: Ben Kweller\'s -  Sha Sha    So I\'ve been running this website for a while now. Almost 3 years go by and I  finally  changed the freakin\' look of it. As you can imagine, I\'m freakin\' tired. But I have one topic to address...   My name is all over this website. I\'m trying to figure out a way to not make it so flagarant, but my ego just won\'t let me. I tried to get a rough estimate of how many times I say my own name, but I get confused.   Check out my nasty rough  urlLink music downloads  page and let me know what you think. I want to make an album, honest! Thanks for the patience, peeps!   Peace.  '),(2933,3856641,'14,July,2004','2004-07-14 00:00:00','   Music: The Killers -  Somebody Told Me   Dear Diary,  Today I ran into Sally in 3rd period.  She said she liked my Metallica T-shirt and I said \"thanks\" and she laughed at me.  I think she\'s got a crush on me.  Either that, or there was a booger hanging from my nose.  Ugh!     My life is so complicated!  Tim wanted to know if I could play basketball with the guys after school but I told him that I was going to the Mall with my girlfriend.  He called me  pussy whipped .  I don\'t think he knows what that means because it sounds just fine to me!  I\'m going to write poetry.    Black Hawk UP    Airplanes are stupid. Helicopters are hella coolz. You can\'t hover in an airplane. Utah Jazz rulez!!!  The End.  '),(2934,3856641,'13,July,2004','2004-07-13 00:00:00','   Music: Pearl Jam -  Vs.   Okay, so I\'m not one to get all nostalgic on my own, but when Pearl Jam popped up on the ole iPod, the pangs of a time in high school far far away really shot up my gut.  A few of my friends have completed college lately and I feel sad all over again.  You miss your friends.  Even the ones that you only went to their room because they had substances (alcohol, etc.) that you didn\'t have the access to (or you sucked at talking the 35-year-old gas station attendant in believing you were 21+).  Ah, memories.  Oh wait...  ... I had to study back then.  I had to spend my evenings dreading the fact that I hadn\'t read the assignment for tomorrow.  Yeah, I still have those dreams/nightmares every now and then, but isn\'t it totally worth that feeling of waking up knowing that you actually DON\'T have to worry about school anymore?  Oh wait...  ... It all gets replaced by something in your life.  Work.  Kids.  Money.  Money.  Eh.  Money.  *Steps off soap box*  How about I just talk about getting naked and writing music?  Sorry.  Just miss my friends an awful lot.  They know who they are.  If you want to get to know them, my friend Alana keeps a journal of the crap they do on  urlLink Blog Spot  and it\'s rather entertaining...   urlLink Check it out .  '),(2935,3856641,'13,July,2004','2004-07-13 00:00:00','   Music: The White Stripes -  There\'s No Hope for You Here   Wow.  I feel sweet.  There are a couple of moments when you feel the best and most clear.  #1:  If you get rid of a headache OR #2: After you\'ve had sex.  You\'re just a caring and interesting person just waiting to listen.  If I had some advice for you, I\'d put it right here.  Instead, I\'m going to tell you that music is coming back.  Oh yes, it\'s coming...  and when it does...  Well, it\'s going to feel like one of those two things from up above.  Probably #2.  Much love.  '),(2936,3856641,'11,July,2004','2004-07-11 00:00:00','   Music: Elliot Smith -  XO   Weird days spawn weird feelings. It\'s just a train wreck when you try to rationalize emotions. Blahedy blahepy. Different perspectives spawn different conclusions.   Okay, so that wasn\'t very deep and incredibly vague. I\'m not one to leave the away message that says \"I\'m out with someone who WON\'T stab me in the back and date my best friend\" so I\'ll just drop it.   Spring cleaning at its finest today. With all that has happened, maybe I\'ll write. I\'ve been such a slacker in that area lately that it\'s been kind of embarrassing. I still play every day, but I really can\'t seem to figure out where I\'m trying to go with my ideas. Maybe I could hire a shirpa to clean out my brain with mumblings and chants of higher mountain thought. Then I can write about the beauties of bumble bees and how unnecessary clean socks are when you go barefoot.   Music is important. Music is sleeping. It  has  to blow up soon. It just has to... Doesn\'t it? Damn.    Hibernation in Fm     Bears are Bears. Fleas are fleas. Large and small. Quite eager to sneeze.   Achoo.  '),(2937,3856641,'11,July,2004','2004-07-11 00:00:00','   Music: Faith No More -  Anne\'s Song   Whoa whoa whoa, crazy reader.  How\'s it shaken?  I\'m fine, thanks.  I started thinking about how stir-fry is much like human behavior.  Think about it.  You take all the residue from your daily lives and they end up tasting like each other in the future.  Things are going great with me.  Wine is yummy and keeps the tummy happy.  If you need some recipes for stir-fry, let me know.   urlLink E-mail me .  '),(2938,3856641,'06,July,2004','2004-07-06 00:00:00','   Music: Justin Timberlake -  Justified  (I felt like dancin\', get over it)  Okay, look.  I have been trying to get this damn blog working all night.  I\'ve learned what CHMODs are.  It\'s oh-so-very interesting, I assure you * yawn *.  Lord.  When I downloaded this thing, the freakin\' description said it was soooooo easy to use.  Anyway, things are going well with me.  I suppose that this will be the place to find out if you are curious in the future on my whereabouts and happenings.  Bed for now.  Ranting for later.  '),(2939,3856641,'06,July,2004','2004-07-06 00:00:00','   How conceded are web journals?  Pretty damn conceded if you ask me.  Maybe you didn\'t ask me, but I told you and you listened so I guess that means I win that one, huh?  You need to understand.  It\'s hard, but we\'ll get you there.  Breticious  '),(2940,3856641,'13,August,2004','2004-08-13 00:00:00','   MUSIC: Ben Kweller -  Sha Sha  (yes, again... shove it) MOOD: Ready for the Weekend  Thank God it\'s... I\'m not going to say it out of principle. Is that the right principal/principle? You can write me and tell me otherwise, but I\'m going to stick with my guns on this one.  Yes, it\'s Friday and I\'m ready for the weekend. I\'m heading off to Roman\'s wedding (you\'ll probably remember the bachelor party debacle ( urlLink click here for recap ). It\'s in  urlLink Morgantown, WV  or near there, anyway. It\'s going to be a long journey, but at least I\'ll get to see some friends that I haven\'t seen in a week. I sure do miss those dirty bastards from Camp Booze. Speaking of which, I still haven\'t posted those lovely Camp Booze pictures from my camera... I have a couple of pictures of not-happy people with their pants around their ankles and their bathing suit tops off. Nothing but towel, baby, nothing but towel. GRRRRROWL.  Well, if you made it this far into the e-mail, you\'re probably reading as to when I\'m going to play again. I\'ll be opening for the  urlLink Steve Hussey Band  on September 11, 2004, at a small sports bar called \"The Front Row\" in Parkersburg, WV. You\'ll remember Steve from the whole Stuck On Amber project from a few years ago. We\'ve both grown up in our directions... Awwww... how sad and cute at the same time. It\'ll rawk, don\'t you worry.  Wow. My neighbor is a freak. I was finger picking absolutely lightly in my room and she\'s pounding on the wall like I was jamming out and singing my ass off. I have a feeling there\'s going to be some falling outs going on once I get the studio up and running again. I will be producing whether she can take it or not so it should be pretty interesting. I\'ll keep you all posted. Good thing I\'m renting... She\'s an owner. Bad side is that I\'m stuck with  her .  I smell a showdown. *sniff sniff* Definitely.  Peace.  '),(2941,3856641,'11,August,2004','2004-08-11 00:00:00','   MUSIC:  Elliot Smith - Figure 8  MOOD: Distraughtly Working  I\'m starting to feel better from the debauchery that was Camp Booze. 3 Days of nothing but drinking can really make a person feel smart and refreshed once it all ends and you\'ve gotten a full night\'s rest without insane nightmares about your girlfriend being swept out from under you by  urlLink Fred Willard . Boy, that dream had one weird explanation attached to it that morning. I mean, he\'s just  so damn convincing .  I want to go on more about Camp Booze, but I really don\'t have the effort in me right now, so check back tonight, and maybe I\'ll even have a picture or two from the hot tub of tit soup. Bear\'s eyes will never work the same after the weekend. Never again. Also, have you checked  urlLink Alana\'s  website for an update? I smell a worker bee who\'s too lazy after work to update with all the quotes from the weekend.  Ah yes, I met Jesus... He plays a mean hand at Texas Hold \'Em. That damned Jesus. He also never gets tired of the Jesus jokes, which is absolutely astounding. They just keep coming. I even found myself apologizing around Him. It was funny, but damn did that get kinda old in just 3 days. And to think to oneself, \"Self, He has to deal with that on a daily basis.\"  I actually heard some of the weirdest words out of a pre-Graduate student looking for a way out of her current position. \"I just keep praying that she\'ll fire me.\" That\'s weird. And for some odd reason, it kinda gets me excited. Not like that, you pervert... Or  am  I? Dump dum duuuuuuuuuuuuuhhhhhhm.  '),(2942,3856641,'02,August,2004','2004-08-02 00:00:00','   urlLink Todd\'s Trip . You gotta feel bad for Todd.  Well, we got back from Cincinnatti last night and boy, that was a fun trip. Lots o\' driving and playing car games. I don\'t think I\'ll ever be able to think of another Celebrity who\'s name starts with a \"P\" or \"D\" ever again. I\'m even almost sick of the music on my iPod, which I thought would never happen.  All and all it was a pretty good trip, but getting up at 4:30am after a bedtime of 12am isn\'t that much fun and you tend to bite your trip-mate over stupid things like spilling water on the wall. But the best part is that the job has been completed and a nice little trip to Pottery Barn Outlet. Yuppies unite! Together we can rid the world of bad decorating. Or maybe I\'m an undercover agent for those who have no style... I just may be infiltrating the world of class without proper authorization from the Hiltons.    Future Song Titles  : My Girlfriend Sucks This Stinks (Points Below) Eat Cabbage, Just Don\'t Touch My Brotwurst I Blew Up Your Dog Fruit is for the Rich Ants in My Pants Ruffies Are a Bad Plan at Camp Booze  Stay classy, Internet.  Bret  '),(2943,4308812,'23,August,2004','2004-08-23 00:00:00','    Just when you thought an editor\'s blog couldn\'t get political... But I couldnâ€™t resist. These grammar bloopers deserved a category of their own. Please feel free to add.  From   urlLink  Still More Bushisms   : Along with a whole pack of other journalists, Jacob Weisberg covered Governor George W. Bush during the long months of the 2000 presidential campaign. Each of those journalists heard Bush make one stupid remark after another, day after day:  â€œI know how hard it is for you to put food on your table.â€    â€œFamilies is where our nation finds hope, where wings take dream.â€     â€œI donâ€™t want nations feeling like that they can bully ourselves and our alliesâ€¦â€ â€“ Des Moines, Iowa, October 23, 2000       â€œWhatever it takes to help Taiwan defend theirself.â€ â€“ On how far theyâ€™d be willing to go to defend Taiwan, Good Morning America, April 25, 2001 '),(2944,4308812,'22,August,2004','2004-08-22 00:00:00','    I hope you are among the hundreds or thousands of people who would have noticed the \"comma\" after \'eats. Yes, it is intentional.   urlLink  Eats, Shoots and Leaves    is an amazing book for people like you who get extremely annoyed when they see a misplaced comma or an unnecessary semicolon; whose blood boils with rage at the sight of a poor undeserving apostrophe hanging at the wrong place. The book talks about grammar, punctuation, and the evolution of language, with several funny anecdotes thrown in. Beware: the author writes (and is an advocate for) UK English. So expect to see a lot of  colourful organisation, full stops,  and lack of  serial commas . Author Lynne Truss also talks about the idiosyncracies of various authors, and the constant tussle that takes place in the offices of Editors.      Click   urlLink  here    for more info on the differences between UK and American English. '),(2945,4308812,'22,August,2004','2004-08-22 00:00:00','    urlLink Occupational Outlook for Writers and Editors  from the  urlLink Bureau of Labor Statistics , a fact-finding agency for the Federal Government in the broad field of labor economics and statistics  urlLink    Note from Miss Editrix: Don\'t take this site too seriously.    Hire an editor or become a freelance editor:   urlLink Associated Writers and Editors  (AWE) are independent professionals, based in Washington, D.C., with years of experience in national and international fields. Services include researching, organizing, and writing reports; editing articles, books, and technical volumes; designing and marketing products; preparing policy analyses, conference discussion papers, and journal articles; and starting up publications and publishing programs.   Note from Miss Editrix: I\'ve used AWE. They are great, especially for texts written by non-native English speakers.  The  urlLink Editorial Freelancers Association  is a national, nonprofit, professional organization of self-employed workers in the publishing and communications industries. '),(2946,4308812,'22,August,2004','2004-08-22 00:00:00','    urlLink    urlLink Write 101  contains hundreds of articles on the nuts and bolts of writing and editing. You are given the opportunity to examine grammar, punctuation and structure in a way that allows you to re-evaluate your writing style and improve your communicative skills.  urlLink    urlLink Information and advice for editors of magazines and other publications.   urlLink    urlLink Copy Editor  has been the leading newsletter in its field since 1991. Subscribers are a who\'s who of language professionals working in-house or freelance for newspapers, magazines, journals, newsletters, publishing houses, corporations, universities, government agencies, and nonprofits. Year after year, subscribers rely on Copy Editor to provide the facts, perspectives, and expert opinions they need to stay well informed and up to date.  Share some of your favorite editing resources! '),(2947,4308812,'22,August,2004','2004-08-22 00:00:00','    Yes, one reallly does exist.      urlLink  The Apostrophe Protection Society    was started in 2001 by John Richards, now its Chairman, with the specific aim of preserving the correct use of this currently much abused punctuation mark in all forms of text written in the English language.    The rules concerning the use of Apostrophes in written English are very simple: 1. They are used to denote a missing letter or letters.  2. They are used to denote possession.  3. Apostrophes are NEVER ever used to denote plurals!     Check out some examples of apostrophe misuse on the site\'s  urlLink Examples Page . '),(2948,4308812,'22,August,2004','2004-08-22 00:00:00','    In and out: Quotation Marks, Parentheses, and the Placement of Periods and Commas   The rule is pretty basic but hard to remember. Periods and commas generally go inside quotation marks (in U.S. English). Periods and commas always go outside parentheses because the parenthetical is still within the sentence. If an entire sentence is within parentheses, the period goes inside.  Citing sources within a text: when you cite the source directly after a quote, the parentheses go outside the quotation marks and the period goes after the parenthetical citation because it is not part of the quote but is part of the thought/sentence. '),(2949,4308812,'22,August,2004','2004-08-22 00:00:00','    Youâ€™ll get asked about these, too â€“ how and when to use them, what they stand for, and why they exist.   et al.:  This is the abbreviate form of  et alii  (â€œand othersâ€)â€”the others being people, not things. Since  al.,  is an abbreviation, the period is required.   etc.:  This is the abbreviated form of  et cetera  (â€œand other thingsâ€); it should never be used in reference to people.  Etc . implies that a list of things is too extensive to recite. But often writers seem to run out of thoughts and seem to tack on  etc . for no real purpose. Also, two redundancies often appear with this abbreviation: 1)  and etc ., which is poor style because  et  means â€œand,â€ and 2)  etc . at the end of a list that begins with  e.g ., which properly introduces a short list of examples.   i.e. and e.g.:  The former is the abbreviation for  id est  (â€œthat isâ€), and latter is the abbreviation for  exemplia gratia  (â€œfor exampleâ€). The English equivalents are preferable in formal prose, though sometimes the quickness of these two-letter abbreviations makes them desirable. Always use periods and put a comma after either of them.   [sic]:  You may be reading an essay or a text for a class when you run across this little guy [sic.]  Sic  provides information about the text youâ€™re reading. It only appears when the author is quoting another source. And it only appears after words that are either misspelled or used incorrectly.  Sic  always appears between brackets â€“ never in parentheses.  Sic  has been around a long time â€“ since the time of the Romans, actually.  The purpose of  sic  it to tell you that the author of what you are reading is not a complete idiot but is simply providing an accurate citation.  Sic  is translated as â€œthus in he originalâ€; it lets you know that the mistake is part of the text being quoted and not part of the text you are reading. '),(2950,4308812,'22,August,2004','2004-08-22 00:00:00','    Are you comma happy? Or do make your readers read on and on without letting them pause to take a breath?     Some people write well but allow themselves to be disabled by a fear of punctuation and grammar (not you, of course). These people know how to prewrite, organize, and revise, but proofreading for punctuation and grammar causes them difficulties. They turn to you for help. You know thereâ€™s no need to fear these conventions of standard English, but you also do not want to give them wrong advice. How can you show them that  urlLink punctuation is simple ? You can\'t give them good advice if you don\'t know the uses of and differences among the five most confused and misused  punctuation marks:  urlLink the colon ,  urlLink the semicolon ,  urlLink the comma ,  urlLink the dash , and  urlLink the apostrophe .       Use this guide on  urlLink Punctuation Made Simple  yourself and pass the info along to your friends who need them. You know who they are. '),(2951,4308812,'22,August,2004','2004-08-22 00:00:00','    Homonyms are some of the most confusing of the most confusable words. Homonyms are those pesky words that sound the same but have different meanings. You know that differences exist between  affect  and  effect ,  their  and  their ,  your  and  youâ€™re , and  principle and principal . Do you know what those differences are? For most of you, this will be a bit of a grammar school review. You might want to print these materials out and post them in your office. Just as the doctor in your family gets called for advice on every family memberâ€™s newest ailment, so do you get called for a quick proofread and advice on whether to use  elicit  and  illicit  in a certain sentence. Make them think youâ€™re a walking  AP Manual of Style .    urlLink Commonly Misused Words and Expressions     urlLink Words and Expressions Commonly Misused, William Strunk, Jr. (1869â€“1946). The Elements of Style. 1918. '),(2952,4308812,'22,August,2004','2004-08-22 00:00:00','    You can break the rules below because they never really existed. Yup, that\'s right. Sister Mary Charlotte, who lectured about all of the following non-rules, deserves a ruler slap to the knuckles. Now you can tell your bosses and proofreaders and non-editor friends who try to tell you you\'ve made a mistake that they are wrong. (And you can feel better about the B- you got in fifth grade grammar. Damn those conjugating sentence exercises.)   Splitting infinitives:  Weâ€™re told to stop the widespread practice of splitting the infinitive, as in the old Star Trek line, â€œto boldly go where no man has gone before.â€ But this prohibition is based on the fact that in Latin, itâ€™s impossible to split infinitives â€“ a ridiculous basis for a rule in marvelously flexible English.   Ending a sentence in a preposition:  Listen to the wisdom of Winston Churchill:     â€œThis is a rule up with which we should not put.â€      Starting a sentence with a conjunction :  Charles Allen Lloydâ€™s 1938 words fairly sum up the situation as it stands even today:   \"Next to the groundless notion that it is incorrect to end an English sentence with a preposition, perhaps the most wide-spread of the many false beliefs about the use of our language is the equally groundless notion that it is incorrect to begin one with â€˜butâ€™ or â€˜and.â€™ As in the case of the superstition about the prepositional ending, no textbook supports it, but apparently about half of our teachers of English go out of their way to handicap their pupils by inculcating it.\"        But  is a perfectly proper way to open a sentence, but only if the idea it introduces truly contrasts with what precedes.  Because  is also a perfectly proper way to open a sentence, but only if the dependent clause it introduces is followed by an independent clause (with a subject and verb). '),(2953,4308812,'22,August,2004','2004-08-22 00:00:00','    \"A Lady is one who never shows her underwear unintentionally.\" --Lillian Day (1931)    Itâ€™s okay to break some rules. When you know youâ€™re breaking them. When you break them on purpose.  Some good writing does break the rules. Sentence fragments exist in many of my postings -- and will exist in many postings to come. But you are an editor, and you know the rules. You know I broke them for a purpose. Certain sentence fragments encapsulate a thought, emphasize an idea, grab your attention. So, go ahead, break the rules. Intentionally. '),(2954,4308812,'22,August,2004','2004-08-22 00:00:00','    â€œSome editors are failed writers, but so are most writers.â€ â€“ T.S. Eliot   Only good writing communicates. To communicate, the words need a structure that enables the writer to convey his or her ideas to the reader in a way the reader can understand. Appropriate use of grammar and punctuation provide that structure. The mere fact that you havenâ€™t left this blog is a pretty good indication that you have an appreciation for good writing. You know that grammar is important. You probably have won a spelling bee or two in your day. You cringe when you see typos. Anywhere. On menus, billboards, Web sites, coupons. Even on the back of shampoo bottles. You read books, pen in hand, and you correct missing apostrophes and dangling participles. As much as you complain that typos should not exist in a book that you paid hard-earned money for, you also know that finding such errors is one of the most satisfying activities of your day. You are an editor. Your friends say you are nit-picky, anal, and obsessive compulsive. You like being right.  This blog is designed to be both a resource for editors (whether itâ€™s your job title or obsession) and a sounding board for you share the annoying and egregious errors you come across to others who will appreciate it (because your husband or sister or roommate is tired of listening to you). I hope you will share editing experiences and opinions on certain subjective edits like the serial commas and UK vs. U.S. English, as well as grammar bloopers and other questions and/or reference materials you have. If anything, I hope this blog spares you the embarrassment of asking a friend or colleague a question that they think (and you know) you should know the answer to. After all, you like being right.     My first few posts are some important â€œlessons learnedâ€ that I want to share. Hopefully, some of these will spark ideas and discussion. I am also posting links to some essays and Web sites that have either been particularly useful or particularly funny. I hope you do the same. '),(2955,4162840,'18,August,2004','2004-08-18 00:00:00','   Well, its been a little over a week since I last posted anything. Limited access to the Internet sucks. I just registered for my MPC classes today. I\'m only taking twelve units because of a SNAFU with the counseling department has greatly restricted my eligibility for classes. Of course my waiting to sign up for any classes until the day before the cutoff for Fall semester registration had absolutely nothing to do with it. At least my schedule for this semester has some fun and easy classes. Other than college stuff, I haven\'t been doing much of anything eventful. I am going to do some writing on my old Underwood typewriter (my graduation present from my mom, it is absolutely the coolest writing instrument ever devised). I\'ll probably start upon a new short story. Should be fun. My 18th continues to close in. September 5th seems to be just around the corner. Well... that\'s about it for today. Almost forgot, the Giants have won five in a row and hold the lead in the NL wild card race. '),(2956,4162840,'07,August,2004','2004-08-07 00:00:00','   .:General Info:.     Name::  Danny Brown     Age::  17     Gender::  male     Height::  6\'1     Birthday::  9/5/86     Hair Color::  blond/brown     Describe yourself in one word::  ummmm....     Describe your personality in one word::  unpredictable      .:School Life:.     What grade are you in::  going to college     What school do you go to::  Monterey Peninsula College     And where is that::  if you can\'t figure it out, slap yourself     Do you get good grades::  (laughs uncontrollably as he reads question)     Favorite Subject::  History      .:Love Life:.     Sexual Preference::  female, preferably alive     Do you have a bf/gf::  nope     If so,what is their name::  look, I\'m touchy on the subject, back off     How long have you been dating::  did you hear what I just said?     Do you consider yourself in love with them::  what we have here is a failure to communicate     If so,why::  one more question on this topic and we\'re going to have to step outside for a \"conversation\"     Have you had your first kiss::  nope     If so,when::  pretty dense, aren\'t you?     Do you have a crush::  yes or no     If so,what is their name::  they may have one, they may not (assuming said person exists at all)     Why do you like them::  maybe I don\'t (see parentheses above)     Do they know you::  Does Kim Jong-Il get enough fiber in his diet?     Do they know that you like them::  Do fish care?     How long have you known them::  maybe I don\'t know them (if they exist)      .:Friends:.     Best Friend(s)::  position is currently open to applications     How long have you been friends::  the answer to this one is rather obvious     Do you consider yourself a good friend::  I certainly hope I am     How many friends do you think you have::  I don\'t count     Most popular::  who?     Most conceited::  what?     Friendliest::  when?     Meanest::  where?     Prettiest::  me, of course     Craziest::  see above answer     Oldest::  why?     Youngest::  how?     Most recent::  Col. Mustard, in the kitchen, with the candlestick     Smartest::  smarter than me? You\'ve got to be kidding      .:Favorites:.     Food::  homemade pecan pie     Person::  John McCain     Number::  49     Show::  The Daily Show With Jon Stewart     Letter::  K     Song::  Voodoo Child (Jimi Hendrix)     Band/Singer::  tie: Hendrix and Stevie Ray Vaughan     Place in the world::  tie: Ireland and Germany     Season::  spring (start of baseball season)      .:Your-:.     Dream Vacation::  tour of World War Two battle sites     Dream House::  Pacific Bell Park, home of the Giants     Dream Room::  the pitcher\'s mound     Location::  in front of the computer     .:Last Time You:.:       Watched T.V.::  right now     Went to the bathroom::  how do you know I don\'t have a laptop?     Ate::  five minutes ago     Slept::  3-7 this morning     Listened to music::  right now     Used the phone::  earlier today     IMed someone/Got an IM::  IM?     Went to school::  last school year     Played a game::  earlier today     Took a shower::  this morning     Hugged someone::  can\'t remember     Went on a date::  (weeps softly)     Wrote a letter::  can\'t remember     Cried::  can\'t remember      .:Last Person You:.     Hugged::  a family member, can\'t remember who     Kissed::  (weeps loudly)     Laughed at::  my cousin     Cried over::  you can cry over people?     IMed/Got an IM from::  once again, what\'s an IM?     Hurt::  I don\'t know, thank God     Talked to::  my grandfather     Spoke to on the phone::  my older cousin     Ate with::  my grandparents     Spent time with::  me     Saw::  some boxer on TV     Missed::  my friends     Heard::  my little cousin (snoring)     Played with::  My grandparents & older cousin (a game of Scrabble)      .:Have You Ever:.     Been out of the country::  no     Been out of state/province::  yes     Done drugs::  caffiene     Done anything illegal::  yes, only mildly so     Slapped someone::  can\'t remember     Cut yourself::  accidently     Played an instrument::  poorly     Hurt someone for no reason::  yeah     Hurt someone::  yeah     Killed an insect/bug::  yes     Gotten stung by a bee::  no     Lied to your parents::  (laughs)     Stole Something::  yes     Kissed Someone::  no      .:This Or That:.     Rock or Rap::  Rock You Like A Hurricane     Singing or Songwriting::  I suck at both     Tennis Shoes or Sandals::  tennis shoes, I ain\'t a Communist     Phone or Computer::  computer     Biking or Skating::  biking     Analog or Digital::  huh?     Coke or Pepsi::  Coke, Pepsi tastes like flat coke     Sprite or Sierra Mist::  Sprite, Sierra mist isn\'t bad though     MTV or VH1::  MTV2 actually plays rock     R&amp;B or Country::  country     Cingular or T-Mobile::  don\'t know     Cats or Dogs::  dogs     AIM or Yahoo::  don\'t know     Bzoink or Quizilla::  don\'t know      .:Word Association:.     Birthday::  party     Peanut::  butter     Good Charlotte::  crap     Ravioli::  noodle     President Bush::  obtuse, lying sack of crap     T.V.::  remote     Rock::  and roll     Rap::  sucks     Chef::  Boy-Ar-Dee     Boys::  (draws a blank)     Girls::  confuse the hell out of me     Calendar::  (draws a blank)     Fan::  mail     Evil::  Knievel      .:Right Now:     Eating::  Twinkie     Drinking::  Pepsi     Watching::  boxing     What is on your mousepad::  don\'t have one     What are you doing::  typing on the computer, dumbass     What song are you listening to::  Paint It Black     What\'s in your CD player::  wires and some batteries     Wearing::  clothes     Time::  10:29PM     Month::  August     Day of the Month::  7     Year::  2004     Day of the week::  Saturday     What website are you on::  bzoink      .:Random Things:.     What color is your mousepad::  already told you     What color is your keyboard::  gray     What is the phrase you use the most online::  don\'t prefer any one phrase     Did you like this survey::  it was pretty good     Are you sad that it\'s over::  I might survive     What are you gonna do after this survey is over::  do something else     Do you like pop-up ads::  does anybody?     How long have you been online::  a while now      urlLink You Know It\'s All About You (Over 100 Questions)  brought to you by  urlLink BZOINK!   urlLink  EmotionDump   - 100% Anonymous Emotions and Confessions '),(2957,4162840,'07,August,2004','2004-08-07 00:00:00','   Damn Giants blew another game, this time reaching majestic new heights in futility by stranding over 10 baserunners. At least they lost to a guy of Greg Maddux\'s caliber. '),(2958,4162840,'07,August,2004','2004-08-07 00:00:00','   Giants actually won today. Schmidt picked up the win (14-4) and (I almost can\'t believe I\'m writing this) the bullpen didn\'t implode in a shitstorm of Biblical porportions. J.T.\'s red hot, smoked a three run dinger to crush any hope the Cubs might have had. Hermanson definitely going to make a good closer.  We should be moving into the new house soon (Glory Hallelluyah), almost definitely by my birthday. I\'m actually going to have a real room!!!  I badly need to develop some semblance of a social life. I will surely lose what teetering facade of sanity I have now if the status quo remains.  Just burned a new CD. Has \"Paint it Black\", \"Wanted Dead Or Alive\", and \"Smells Like Teen Spirit\" on it and \"Hotel California\" is the closer. Should be a good one to listen to at night.  Plan on maybe getting some knuckleballing in tomorrow, happy-happy-joy-joy.  50 some-odd days until \"The Dark Tower\" is released. Am happy, but also sad. Roland and the rest of the ka-tet are the closest friends I have right now. God that\'s pathetic.  Would commit genocide in a small Third World country right now if it meant I would get a girlfriend. Given the previous \"D.T.\" statement, prospects seem incredibly dim. Wish I had my Chris Isaak CD on me right now, \"Kings Of The Highway\" and \"Wicked Game\" make me feel less crushingly depressed. Don\'t know how listening to sad music while feeling sad helps the situation, but it does, somehow. '),(2959,4210644,'20,August,2004','2004-08-20 00:00:00','    music: only one by Yellowcard   mood: hella bored   TV: the olympics   food: ditto      heyhey so ya i just got home from my moms....and once again im hella bored...ugg i really just want this summer to be over...its so boring. I mean rachel is in the USA and Kelly is never home and Sharon can\'t do anthing so all i can do is go to the prk with brandon or the hundred stairs with kat, and after a while that gets kinda boring.  ya wen i was at ma moms we went shopping, i didn\' t get much, just a binder, 2 pairs of dress shoes *god damn uniformz*, and \'the notebook\' ...the BOOK, nit the movie...but as soon as that movie comes out i am buying it..same goes for the GC cd. im really excited bout the chronicals of life and eath, i think it gonna be really good..i heard there fisrt single on the radio this week, even tho it\'s not supposed to be played till the 23 *hey i no everthing bout GC, leave me alone* and it 3was really good...it was alot more...i dunno, almost, like.. harder..as sick as that sounds..put it this way, its more rock and less pop..thank god! ya so anways i gotta go cuz i have to go over to maleena\'s and try on the quilts shes giving me *thank gid shes such an angle* so tootles!   '),(2960,4210644,'15,August,2004','2004-08-15 00:00:00','    heyhey well as you can tell by the title of this entry im hella board. I slept over at rachel\'s last night **OMG really!!! you\'ve never done that before**  **ya i know...thats a first eh?!?!** and we didn\'t wake up until like 2:30pm! **ur such lazy asses!** **we were tired...and her bed was comfortable ok!**  **ya ya sure sure w/e u say**  **don\'t make fun of me!**  **lazy-ass**  **shut-up!...*cries*..**  **evil laughs**  but it was fun....we stayed up for like an hour talking about don....**hahahahaha ya i know, pathetic!**   **very!**   **would you please shut up voices!** and i told her i think the should go out...i mean they would make a cute couple and i dunno....i just think she should give him a chance. w/e if she does go out with him i promised her i wouldn\'t tell anyone.  omg i really want to read the notebook..and i want to see it!!!!! it looks s0o0o0o0o0o0o0o0o good! i will see it!!!! with god as my wittness i will see it! it will be mine!!!!!!!!!!!!!!!!!! **heh heh..sorry bout that...** well im gonna go look 4 more stuff bout the notebook...**heh heh** *obsess much?**  **it looks good..an i like romance ok?!**  **w/e**  **shut-up!**      tootles!   '),(2961,4210644,'13,August,2004','2004-08-13 00:00:00','    music: all downhill from here -new found glory   mood: alright   tv: Y&R   food: sandwhich      heyhey so im hella bored and im really pissed...i was trying to talk to cody and kyle but the assholes never talk and then cody was like \'wut do want? me and kyle don\'t like you dont talk to us cuz ur ugly..**smilez**\' i was like wtf! Its not like i really care but he was such an asshole! w/e...other than that i didnt do nething eventful today. I want to buy the rock against bush cd cuz its really good...i want volume 1 and 2. Dana has 1 and she brought it over and we listened to it and it ws fucking rad...ne ways im gona go ...tootles   '),(2962,4210644,'12,August,2004','2004-08-12 00:00:00','    music: hawiiene by les trois accords mood: pretty good tv: ditto food: ditto  ya so dana is coming over later and she\'s sleeping over....i haven\'t gotten to do that much with her this summer, cuz of Sevec and all. Well other than that nothing is really going on...it\'s 12:24 and the god danm phone woke me upat like 10 this morning ...again! I swear to god if that phone wakes me up one more time i\'m gonna rip it out of the wall then stomp on it 10000 times until it is only ahes left then throw the ashes outside and let them float away in the wind! mmmwwwaahhaahhhaa!!! sorry...i get carried away sometimes... anyways, yah i have to get ma uniform this weekend...or sumtime soon at least. I already know what im getting: 2 short french cut chirts, i long french cut shirt, 2 pairs of paints, the knee high socks, 1 sweater and 1 tie. Hopefully Maleena will give me one of her quilts since she has 3, but if not then i\'ll have to buy 1 quilt aswell.  yah well im gonna go listen to some......hawiiene bu les trois accord...ahh good one frenchies..well tootles!   '),(2963,4210644,'12,August,2004','2004-08-12 00:00:00','    W E  L  C  O  M  E'),(2964,1785539,'06,August,2004','2004-08-06 00:00:00','   It\'s getting near the end of the day on Friday and I\'m looking for a back door to sneak out of.  All of the office women are going crazy.  They are like apes in khakis and flower print shirts who know the promised treats of apple martinis and Julia Roberts movies are very close.  They are screeching (\"I want guacomole!\" \"Derek doesn\'t like crowded bars!\") and banging on their cages.  I feel the same way, but the tail end of a hangover is keeping me slightly more in check. I just called Kyle at Microsoft.  I love calling him there.  I always feel important asking the operator to please connect me with Kyle Kinkade (\"Last name spelt K-I-N-K-A-D-E\").  They always say \"Thank you, Sir. I\'m connecting you now\" and I feel like I\'ve had an agent of the machine do my bidding.   Sidenote:I like it when people call me sir, especially teenagers.  I\'m not one of those people who get caught up on the youth not finding me hip enough to call me dude anymore.  And I\'m deffinitely not going to be one of those lame-ass Jimmy Buffet listening people who tell the youngin\'s to call me \"Brent\".  Hell no, I\'m Mr. Kinkade, Snotface, now go mow a yard or play with fire.  Anyways, I called Kyle and he told me he has a top secret project that might keep him away from the apartment until Wednesday.  He\'s like Batman, he can\'t reveal what he does outside of the home.  He\'s so precious with his little red sports car and big brain.  Speaking of which, last night he brought up the excellent point that I know a lot more about computers than I let on.  Ever since I\'ve moved to Seattle I act like I\'m a hillbilly whenever he and his friends start talking about computers.  I even got a shirt made up that says \"Don\'t talk to me about computers\" (with a 66 on the back for esoteric appeal) to wear to one of his little socials.   The thing is, I was considered pretty proficient with computers back when I was in college.  Is this some self-defeating sibling rivalry of mine?  Why do I always pretend to fall asleep when he says \"CPU\".  Does it work the other way?  I\'ve haven\'t seen Kyle read anything since I\'ve been here, but I see books in his room. I think I got a promotion today.  Not any extra money, but extra responsibility.  I\'m replacing some guy named Roy who left recently.  He did other stuff with folders and paper that I bet many have never dreamed could be done.  I\'m looking forward to my new role and whole new cubicle!  Maybe I\'ll stay on and become a full-time (non temp.) Roy part 2.  This Roy fella always dressed kind of sharp and listened to trendy music.  Maybe if I do a good job I can make enough money to dress in Gap clothes and say things like \"Should we listen to new Radiohead or old Portishead\" and \"I say, Teeny, is this cabarnet or cabARET?\" I\'m ready to go.  I want to try to squeeze in some frolfing before I go see Ramones girl tonight.  She and her buddies caught some salmon in the ocean last weekend and I want to cook it up tonight (found a simple receipe for bbq salmon that sounds like it will make panties wet).  Then it\'ll be off to a party with people I don\'t know and more than likely alcohol (which I\'m all too familiar with).   Before I go I want to talk about something.  I keep thinking about my pops every time I use the copy machine at work.  Although, I\'m pretty sure my dad has someone who does his copying for him and I can\'t ever remember dad doing too much Xeroxing in the garage as a kid.  This got me thinking about father figures and how they affect us.  I\'ve been stuck with this image of Atticus Finch hunched over a copier blankly staring at the words  \"1 side/2 side\" with a look that says \"How did I get here?\" on his face.  Oh, to be the men we want to be, and yet we are the men we are.   Ok, the ape women are getting restless; I better start making a run for it.  Late '),(2965,1785539,'06,August,2004','2004-08-06 00:00:00','   I\'ve totally forgotten what I was going to say.  It was profound I think.  Hmmm. Last night was typical B-Man.  I got drunk on cheap beer while other people did sober things and told me how funny I am because of the drinking and smoking and punching and cursing and long sentences with too many conjunctions, or... I woke up at 8:05 today when I was supposed to be at work at 8:00.  I never used to be late for anything.  I used to get major anxiety attacks an hour before something was going to happen.  \"Holy Christ, the Wonder Years is coming on in 25 minutes!  Where\'s my soda!  Kevin!\"  Now I show up late and wearing a ball cap.  Fuck \'em, I\'ve been stabbed before.  Actually, I haven\'t: but, Goddamn that\'s a good line. So, it sounds like I\'ll be able to stay on at the space commander desk job after all.  I call it that because I spend most of the day pretending that my cubicle is a spaceship and I\'m fighting space asteroids and star people.  It all started because the ergonomically designed chair that I sit in has these arm rests with buttons and levers on them to control lumbar support and tilt.  But in my world they control the lasers and my tracking beam.  Sometimes I\'m wearing an eyepatch and have a big metallic space warrior suit on in my daydreams.  Usually I have my trusty crew with me who gives me advice on how to handle situations.  \"Captain, don\'t shoot that asteroid, it\'s kind of pointless and we only have a million photon rockets!\", they say.  \"Sounds like a challenge,\" I reply.  \"Huh?\"  And then I start my intergallactic genocide.  I\'m an anti space-monster-ite.   Then I come back to reality. Casual Fridays rock!  That\'s all that rocks about this place (except the world in my head).  I\'m dressed like such a frat boy today that I feel like pinching asses and tailgating out in the parking lot.  Maybe I\'ll jeer people as they go into work (\'You\'re office sucks, Washington Mutual has the best defense in the PAC 10! Whoooooo\").  Actually, I think I wore this exact outfit to a Clemson/Carolina game when I was a sophomore.  That night was crazy!  This guy started hitting on my girlfriend at the time (El Diablo) while we were partying on the frat hall.  Later on that night I found him on my bed and he did a \"come hither\" motion to me (this guy is a complete stranger by the way).  I said \"what the fuck are you doing?\" and he said \"Working on my paintball skills\".  He actually tried to convince me that he was laying on my bed drunk working on how to best squeeze the trigger of his paintball gun.  I\'m not a homophobe (I\'ve actually given thought to banging a dude just to prove that) and I\'m not sure if the guy was actually gay or confused or really did like paintball; but, he had to go.  The fucker had his shoes on my bed.  Try and do me like I haven\'t been done before, that\'s cool, I\'m honored.  Hit on my chick? You can have her, she\'s too skinny for me. But, shoes on my comfortor?  Get out. That was the same night I put the shaft of a six iron through my right arm, which I still have an amazing scar from.  Note to all new found frat boys out there: Cheap vodka is a pretty good anti-bacterial rub. Say something smart, Brent, you sound like an ass.  How do you get exactly 4 gallons by using a 5 gallon and a 3 gallon jug?  Fill the five gallon jug and pour it into the 3 gallon jug.  This will leave 2 gallons in the 5 gallong jug.  Empty the 3 gallon jug and pour in the 2 gallons.  Refill the 5 gallon jug and pour one gallon in with the 2 gallons in the 3 gallon jug.  This will leave exactly 4 gallons in the 5 gallon jug.  \"Form follows function\" is by far the most challenging architectural design method.  If lacking in aesthetic, it is made up for by it\'s almost alluring decadence in simplicity. This weekend I\'m going with the happy couple to an all-day cross country Bocce ball tournament.  I\'m looking forward to playing.  I keep having memories of playing the game with Kyle in our backyard in N.C.  \"Lawn bowling\" is the lamest term ever invented.  I can just see some spectacled fuck in shorts and suspenders in Kentucky who can\'t say \"Bocce\" without the \"K\" sound, just giving up and going \"Let\'s lawn bowl.\"  He\'d have a huge grin while all the other corn eating fucks all said \"Kurt, you\'re a riot.  Lawn bowling!  Funny guy!\".  Fuck Kurt, it\'s Bocce ball. Anything else happen to me?  Did laundry, got drunk, stayed up late.  Nope, pretty lame night and boring day so far.  I think I\'m going to go blow up Mars. '),(2966,1785539,'06,August,2004','2004-08-06 00:00:00','   Just had my third encounter with goateed-elevator-rule anarchist today.  Epidemic? '),(2967,1785539,'06,August,2004','2004-08-06 00:00:00','   I have a theroy.  Men with goatees don\'t know elevator manners.  Today I saw two men, both with goatees, walk onto an elevator before the people already on the elevator walked off.  It\'s like they forgot the golden rule of elevators, \"Don\'t walk before others as you wouldn\'t have others step in front of you\".   Maybe the goatee is evolution\'s anit-elevator sign post.  Maybe the men were just evil versions of themselves. On to real news.  I have succeeded in losing a job before I even started working at it. I had landed a well-paying gig at a certain place doing things.  It wasn\'t glamorous, but the benefits would allow for anti-depressants and cooler eyeglasses.  Well, I forgot to put on my application that I was arrested when I was a sophomore in college. Here\'s the story for the 4 people who don\'t know it: I was the social chair of my fraternity one year.  I was in charge of planning the annual orgy known as Carolina Cup (for those not from S.C., think of it as a Renior painting with Busch Light and less posture).  It was the end of the day and the pigs were on the prowl (power to the frat boys!).  I got into a fight with these two dudes from another frat about the property rights of a 3 foot sub sandwich.  In the end I offered that they take the sandwich ... UP THE ASS!  As I went to throw the deli monstrosity at them Dan Marino-style I was tackled from behind by 3 cops.  I thought I was getting jumped by the other frat and fought back.  So, more cops jumped on me and eventually the battle of Subway was over.  It took 6 cops to bring this hoagie fuck down, baby! Back to present day.  I forgot to mention this on my application.  So, when they did my background check this old charge came up.  They were forced to take back their offer, which made me upset for a while.  Now I\'m hoping the temp. company I\'m with will let me stay on with them.  Will Brent be umemployed as of Monday?  Stay tuned. Today I\'ve got that awkward confidence I always put on when I\'m unsure of the future.  I think I\'ll go play frisbee golf (game steadily improving.  driving is stronger, mid range sweet, putts oh-so-close) after work.  Then a sandwich and then Coors Light.  Perfect plan. '),(2968,1785539,'06,August,2004','2004-08-06 00:00:00','   I have to tell you who I am if you don\'t know me.  I could be ultra-alternative and say I don\'t believe in \"getting to know you\" stuff and ask you to judge me based on whatever else I do or don\'t do depending on what\'s considered hip by the alternative press.  Or, I could tell you I like the color blue and was born in Iowa.  Both of those things are for dorks.  And we all know now that dorks don\'t rock.  I do. So, I\'ll tell you the things I do.  Not things that I like to do, just a list of ten things I do throughout the day.   1.)I fantasize about sleeping with roughly 40 percent of the women I see.  This gets weird when I\'m at hospitals or drag shows. 2.)I chew Wrigley\'s Doublemint gum because one day I hope I run into my long lost twin and we\'re both chewing the gum at the same time.  I\'d call him B2 and we\'d laugh forever about the \"gum incident\". 3.)I drink Pabst Blue Ribbon.  Alot.   4.)When I know someone is coming over to visit I pretend to be watching the History Channel.  When they walk in they think I\'m smart when really I\'d rather just watch \"COPS\". 5.)If I like a girl I always forget her name and have to make up nicknames.  \"Russian chick\", \"Gas station girl\", \"The one that likes the Beatles\" are some examples from the past. 6.)I make faces at babies when no one is looking. 7.)I daydream about explosions and super spies. 8.)I play frisbee golf roughly 3 times a week.  9.)I don\'t eat microwave popcorn because I feel like I\'m cheating on movie theater popcorn. 10.) I get disturbed when people drop foreign words into english sentences.  Like this one time this guy kept saying \"Try some of this champagne (but said it the french way \"shom-pawn-ya\").  I wanted to say \"Why don\'t you just go-fuck-ya\".  We all know at least a little bit of a foreign language; so, stop trying to look like a Kennedy, you arrogant fuck.  That\'s why we have the History Channel. period goes here '),(2969,1785539,'06,August,2004','2004-08-06 00:00:00','   Holy Christ, dorks!  I just got a call from Ramones Girl that woke me up; I\'m two hours late for picking her up for camping.  The way she said \"Do you not want to go anymore\" made me want to buy her roses. Damned malt liqour and late hours. '),(2970,1785539,'06,August,2004','2004-08-06 00:00:00','   The name Brent gives you a strongly independent and highly creative nature, with drive and ambition to have experiences and accomplish things out of the ordinary. You can work intently at whatever is new and holds your interest at the moment, but your interest wanes quickly when drudgery and monotony set in. Obstacles to your progress or restrictions on your freedom to act create a sense of frustration which may cause you to feel resentful and even rebellious. You can then become intolerant of others, and caustic and belittling in your expression, thereby imposing stress on your personal relationships. Although you have a clever, quick, capable mind, your progress in life is restricted by instability in your affairs and misunderstandings with people. Your impulsive nature can lead to actions which you later regret taking, or to accidents. Relaxation is elusive, and depletion due to nervous tension can develop to the point where you become subject to moods of depression and morbid thoughts. Nervous tension centres in the solar plexus, with nervous indigestion and stomach ulcers a possible result.   Dorks, is this true?  I wish to think this is one of those all-encompasing generalities.  Dammit, my life can\'t be summed up by a random website, can it?  I\'d rather be shot by a Terminator than figured out by www.whoyouare.com '),(2971,1785539,'06,August,2004','2004-08-06 00:00:00','   I need to find out how to buy the rights to the Harlem Globetrotters theme song. I downloaded it earlier this afternoon and I\'m hooked!   I\'ve been playing it while I clean up, while I make a sandwich and while I take a little nap; it makes everything a little more whacky! Watch out Mr. Sandwich, that\'s not a jar of mustard; it\'s a jar of confetti mean ol\' Slippery Kinkade gonna throw on you!  Uh-oh, Mr. Nextdoor Neighbor, ol\' Slippery gonna pull your shorts down and then dunk while you look silly!  Hey there, Lamp, Slippery ain\'t pregnant; the ball is up his shirt! If my life was a movie and I had to have a theme song play whenever I did something cool, got into a fight or had rough sex it\'d be this song! '),(2972,1785539,'06,August,2004','2004-08-06 00:00:00','   I didn\'t write in this thing yesterday. I hope everybody didn\'t get shot. As you might be able to tell from my backwards hillbilly speaking, I\'m tired.  I had a lot of work to do today as Roy2D2.  One of the swing shift fellas totally forgot to fill out the HMDA forms and that set my closing process back a good 2 hours!  Can you believe it?   Here\'s what happened to me in the past 48 hrs. -A man with a goatee stepped on an elevator before I got off of it.  I was too shocked to say \"Hey\". -I rented \"Marathon Man\" and \"Rear Window\", excuse me \"Alfred Hitchcock\'s Rear Window\".  I wanted to redeem myself after watching \"Phone Booth\" the other night.  Forrest Whitaker? Hmph.  And Colin Ferrel?  This guy is supposed to be Hollywood\'s \"bad boy\"?  He\'s about as badass as a pair of Reeboks.   OOOO, take that you aussie or whatever fuck.  And Keifer Sutherland, listen up, I only want you making movies with someone related to or with the last name of \"Estivez\" from now on. -I got dragged downtown to go to a karaoke party.  By party these people meant a room with a computer that plays songs.  I was instructed to sneak beer in.  No.  I was told to have fun.  No.  I was already a little drunk when we got there, which allowed for one of my prouder moments in life.  I was in this \"room\" box thing for about a minute listening to the opening rifts of \"Superman\" by 3 Dicks Up My Face when a penguin came on the screen.  I yelled, \"And now there\'s a fucking penguin?!?!  That\'s it, I\'m out\".  And I was. -Kyle and I drove around downtown last night.  I had one of those surreal/ perfect life moments when we were stopped at a red light next to Seahawks Stadium around midnight with the top down on his cute lil\' car and The Sex Pistols\' \"Holiday in the Sun\" rocking everyone in ear shot.  It was a Visa commercial.  We were young and rocking; we were American Eagle. -I spent some quality time with Ramones Girl. She told me that I woke up during the middle of the night on Wednesday laughing my ass off.  She asked me what was so funny and I sat upright in bed staring at the wall and started saying \"Does anybody have any questions? Anybody? Anybody?\".  Over the past couple of weeks I have, in my sleep, asked her \"Where are the buildings?\" and \"When is the great race of 80 days on?  You don\'t know of the great race?\".  Most girls would bolt, but she says it\'s entertaining to have a psychotic in bed with her.  I\'m her R.E.M. cycle H.B.O.  That should say something about her. We have started to make plans for camping this weekend.  We are talking about going ot the Pacific Ocean (not the Caspian Sea) and camping on the beach.  That would be swell; I\'ve seen the sunrise over an ocean but not set behind one.  I\'m as squirrely as a prom night whore thinking about it. That\'s about it.  I have felt like I\'m coming down with a cold all week.  I\'ve been low on energy and kind of grumpy.  I\'m going to go take a nap. I hope all of you kids starting school are having a good time.  This is the first time in almost 20 years that I haven\'t been in some form of school during the fall.   HMDA.  Can you believe those stooges on swing shift? '),(2973,1785539,'06,August,2004','2004-08-06 00:00:00','   You know, for someone that spent 4 and half years studying journalism I sure make a lot of editorial errors. '),(2974,1785539,'06,August,2004','2004-08-06 00:00:00','   I like being the new Roy.  People even refer to me as \"The new Roy\".  Roy\'s job is so simple and keeps me busy enough to not notice the time; although it does allow for plenty of day dreaming.  Today I reenacted the scene from \"Desperado\" where Antonio Banderas is hiding behind a counter and has to silently check his pistol for bullets so the bad guys don\'t hear him.  I held my pen up to my ear and slowly clicked the button on top while I ducked behind a file cabinet.  It let out a little click (just like in the movie!) and I made that \"oooo\" face.  Such a hottie.  Antonio? No, me.  I haven\'t really checked to see if the other people in the office have a clear view of me.  \"The new Roy is making laser noises.\" There\'s a new annoying woman at work.  She doesn\'t scream about guacomole or look like a rat  like Screeching Woman of yore.  That chick looked like Minney Mouse came to life and then got shot in the face. This one has no inner-dialogue.  \"I have to tinkle\" \"I was just at the little girls room\" \"I bought new shoes because my feet stink\" \"My phone is black\" \"I like pizza\" \"Eye blinking, breath, eye blinking, breath\".  She sits right behind me.  Sometimes I go in the bathroom just to escape her.  And then I hear from the other side of the wall, \"The urine is coming out of my urethra.  Leno was good last night\"  What\'s worse,  she\'s the type of girl that makes a different voice for every occasion.  (British) \"Hello, sir, how are we today\" (Ghetto) \"I gave you that file, boy\" (LOUD) \"I told her I like pizza. PHONE BLACK\". I wonder if she gets hit at home; which, would raise the age old question, \"Which came first, the hitting or the self-focused ego-building lack of confidence babble fest?\".  I don\'t condone hitting.  But sometimes it\'s o.k. to shoot a dog. Enough about work.  That\'s not what defines me as a B.   I think the neighborhood kids are at the door.  I\'m not going to go find out.  They keep asking me to come out and help fight some unruly teenagers who apparently are throwing rocks at the poor kids.  I never see these kids and I don\'t want to start a fight.  Teenagers are mean man and I might end up punching of them.  \"Local man beats up kids with rocks\". Right now I\'m talking to some girl I used to bang; she is boring.  I don\'t even know what she is saying.  Something about boiling water and noodles.  God, I\'ve slept with some dumb girls.  My brain keeps playing the \"Harlem Globetrotters\" theme song. She keeps asking me if I\'ve met anyone new and I don\'t want to even have that conversation so I just keep saying \"No\".  She asks if I still smoke and I don\'t want a lecture so I say \"No\".    Now she\'s asking about my mom.  Has she even met my mom?  Ahhh, this is boring.  More boring the that horrible movie, \"Phone Booth\".  I need an out.  Quick brain, focus. \"Brent, you are hungry.\" YES!  I am hungry!  I need food or I will die.  You don\'t want me to die.  No, I haven\'t cut my hair.  No, I don\'t remember the dinner with the birds overhead.  I haven\'t been up to anything.  I live in a box.  I have no friends.  I\'m sorry, but I have to go.  I\'ll call you next week.  No, I\'m not still a liar.  Freedom.  The whole point of me moving to Seattle was to de-barnacle myself.  Just keep humming the Globetrotters and keep your head down when the \"tentacles\" (wink) of the past come up from the deeps.  \"Whore, she blows, c\'apn\"  \"\'Tis the great white skank\".   I just bitched through this entire blog.  I think I\'m going to go get some deli meat and make me a fatty sandwich and drink some coke.  Because that\'s what I like to do.  Let\'s end this on a pos note, B. When I was a little kid I went to the zoo with my mom, my brother, my godparents and their kids.  I was sitting on a wall surrounding the prairie dog pits.  I was having so much fun.  Until I fell in.  I hit my face and all the little prairie dogs ran to their holes.  As the people I was with scrambled to get the zoo workers the prairie dogs stuck their heads out, one by one.  I had gained their trust by not yelling and laying completely still. I got to go with my godfather to the world famous Mayo Clinic in Rochester, MN.  He was an anestheologist there.  I got checked out by a really pretty nurse and was giving a clean bill of health.  My godfather then showed me what doctors do during a day.  I got to see the operating rooms and he even gave me my very own pair of very baggy medical scrubs.  I wish I still had those.  And I wonder if the prairie dogs still talk about \"The Day That Kid Fell In\".  Have a good night, dorks. '),(2975,1785539,'06,August,2004','2004-08-06 00:00:00','   Dammit.  That story I asked to be read didn\'t really take place in the South.  That ruined my trifecta that I thought I had with the piece.  Who the hell cares about South Windsor, Connecticut skanks?  I don\'t.  They\'re just leftovers from the plethora of Kennedy skank bangings over the past few decades.  \"Would you say I have a plethora of skank bangings?\" \"Si, John, I would say you have a plethora\" \"Ted, do you know what a plethora is?\" I don\'t want blue blood hoochies, I require my trash pure Southern.  PBR blooded.  Read the story if you want, but it\'s ruined for me because I can\'t imagine her screaming \"Daddy paid for this dress; you ain\'t my Daddy\" no more.  Dammit.  She even jumped on the hood of a car and everything.   I\'m all alone.  Not figuratively, literally.  I\'m alone and sober.  This can\'t last long.  A sober B and an entire night alone don\'t mix well.  Like oil and vodka.  Mmmm, oil and vodka.  I think we have some olive oil in the pantry and the liquor store doesn\'t close until 9.  Boredom and sobriety problem solved, people.  Nothing more to say. '),(2976,1785539,'06,August,2004','2004-08-06 00:00:00','   Please read this story!  http://www.sun-sentinel.com/news/local/southflorida/sfl-819maritalbliss,1,7307132.story?coll=sfla-home-headlines  (oops, it doesn\'t create a link; I guess it\'s copy and paste time, dorkfaces)  Thank god for the South, alcohol and young tattooed skanks.  Without the first there\'d be no \"COPS\" for me to watch, without the next no reason for me to run this silly web dealy, and if there were none of the later there\'d be no porn.  God bless us, everyone. '),(2977,1785539,'06,August,2004','2004-08-06 00:00:00','   Not a good day.  I would say that it\'s because it\'s Monday, but that\'s for dorks.  Read the title, dorks.   I\'m sorry, Brent didn\'t mean to yell at you.  He\'ll buy you some ice cream.   First off, this little kid with way too much hair keeps knocking on my front door while I\'m writing this asking me if he can play video games.  Usually I\'d say yes.  They\'re good kids; but, it\'s adult time.  That means movies with cussing and Marlboro Reds. Second, I just got back from a horrible game of disc golf.  I started out strong by paring the first hole, which I had never done.  On the second hole I got caught up behind a 4some and decided to repeat the hole after a double bogey.  My chance for redemption, right?  No, my time to throw a jackass hook that landed me right in the middle of a blackberry patch the size of my truck.  Scratched up and soar I decided to go onto the third.  Redemption song time?  No, again.  I hook again and go over a fence and land in someone\'s back yard.  After retrieval I attempted teeing off again.  The damned thing goes shorter than the first time I ever played. That\'s it, I\'m done.  I packed up the discs and sped home.  My C.D. player was on the skids in the truck so I had to listen to the 80\'s station and a recount of the Butterbean/Larry Holmes pugalist spectacular on talk radio instead.  Man, that sounded like one lame fight.  A 50-year-old plus man beats the crap out of a super fat 30 something pop icon.  It\'s like Rocky 3 without the soundtrack. Before that I got a lovely letter from the Devil\'s cell phone company telling me that I had to pay a very large sum of money because Robocop now had my address and had been ordered to \"Fuck the prime directives\".  Curses, I knew this day would come.  So long cool new tennis shoes and premium whiskey and hello adulthood. And before that I had a very long and sleepy day at work.  Actually, work was all right, comparitively.  It was a slow day but I managed to stay very busy in spite of others mulling about the coffee machine chanting \"There\'s nothing to do\".   The salmon rocked last night.  Here\'s the recipe:  1/4 cup pineapple juice 2 TBS fresh lemon juice 4 (6 oz) salmon fillets   2 TBS brown sugar 4 tsp chili powder 2 tsp grated lemon rind 3/4 tsp cumin 1/2 tsp salt 1/4 tsp cinnamon   Cooking spray Lemon wedges (optional)   Combine first 3 ingredients in a ziploc plastic bag, seal and marinate in refrigerator 1 hour, turning occasionally.   Preheat oven to 400Â°F. Remove fish from bag; discard marinade. Combine sugar and next 5 ingredients in a bowl. Rub over fish. Place in an 11x7\" baking dish coated with cooking spray. Bake at 400Â°F for 12 minutes or until fish flakes easily when tested with a fork. Serve with lemon, if desired. Serves 4  After dinner I walked around with my redheaded sweetheart looking for the moon to try and spot Mars.  We couldn\'t find it.  But, isn\'t pointless wandering with a newfound someone always the best?  I think we reached a definite mile marker in the relationship.  She broke the Terms of Endearment barrier yesterday.   \"Wait, how do we get to Seattle?\" I asked. \"Honey, we\'re in Seattle,\" she replied.   This B is someone\'s honey.  Actually, I feel a little better now.   Fuck the promise of ice cream, reader person, I\'m going to go make a phone call. '),(2978,1785539,'06,August,2004','2004-08-06 00:00:00','   Just got back from my Seattle date with the Ramone\'s girl.  Here\'s a quick list of fun and exciting things that happened to me today. 1.) We went down to Lake Washington where little kids were swimming and diving.  There was this dorky little lifeguard riding around a boat in 5 FEET DEEP WATER!?!?  Best overheard line from a little kid today: \"You\'re going to get nunchucks on the head!\" 2.)We then went to a restaurant down by the docks on the Puget Sound where a.) the waiter said this classic line:  \"Hi, welcome to this restaurant, are you guys from around here? you live here so you know how we do it here, let\'s get down to business.  let\'s get this train a moving.  You probably want me to just be quiet and let you order, don\'t you.  She\'s thinking \"I wish he\'d be quiet\".  I can tell she wants me to be quiet.  Do you?\"  We did.  b.)Over excited waiter boy kept calling me \"chief\" or \"boss\" ($5 tip for a $10 tab right there), but he kept touching Ramone Girl\'s arm ($2 tip) then he called me buddy ($5 tip) and then he touched MY arm ($2 tip).   b.) we saw \"party boat\".  It\'s a boat (or type of boat) that goes around with 3 100 foot high streams of water shooting out of it it doesn\'t go near any of the boats.  When a boat comes near it it starts spinning.  I\'m serious.  It is like Looney Tunes in Seattle.  What purpose does this boat serve?  Who needs a boat that shoots water into the air and then spins?  Who said \"Dammit, we have it all a needle for the space and a bunch of other stuff; but, we don\'t have a boat that shoots water, avoids other boats and spins when excited.\"? c.) a bird flew under our table and started attacking Ramones Girl.  A woman passing by said \"Ahhh, I hate those things\". 3.)After the restaurant and a few minutes of standing on a deck trying to figure party boat out, we decided to head home.  I kept seeing a bunch of people wearing necklaces that had a bunch of pot leaves on them.  Then I remembered it was Hempfest, I grew curious.  This is an actual conversation I had with two stoner girls I saw at a crosswalk. B:So, where\'s this thing happening that\'s on your button? (\"Thanks for pot smoking. Seattle HempFest) Stoner girl #1: What? B: The HempFest thing, where is it? Stoner girl #1: You mean the pot thing?  What\'s it called? B: HempFest Stoner girl #2: Potfair?  Stoner girl #1: Weedfest. HempFest.  Yeah, it\'s way up there, man.  You can take a bus.  For a $1.25 they give you a transfer ticket, so you can transfer. B: That\'s nice of them.  What do they have at the HempFest? Stoner girl #1: Stuff.  Lot\'s of stuff.  If you want pot stuff, go there. I didn\'t go there.  I have enough stuff. 4.) We went to an arcade where I played that quarter game where you drop the quarters and the machine knocks over quarters, I don\'t know if it has a name.  But with one quarter and one quarter only I won ... 67 MOTHERFUCKING TICKETS! HELL YEAH! I RULE!  I didn\'t want any of the prizes (A Kit-Kat, a South Park sticker, a Spongebob Squarepants mini mug) so we picked out the cutest kid there and gave her the tickets.  I am Robin Hood, I take tickets from the machines and give them to the cute. 5.) Soon after that we passed by the greatest street musician ever.  I call him that because he made up a song about us.  He looked like a beefy Michael Stipe and kept staring at us as if there was no Seattle, only us.  The song went something like this.  He plays basketball for Winthrop.   He\'s the last of the basketball players and he doesn\'t care.   He likes to play basketball and there\'s a girl with red hair.   He\'s a basketball warrior.  He likes to play, but he\'s the last and he\'s alone.  He continued to sing as we kept walking; for all I know he\'s still singing (\"He\'s not here right now, but he\'s still in love with basketball\").  It\'s the greatest song I\'ve ever heard.  And it\'s all so true.   Today was such a great cure for a hangover.  No chance to be depressed, iced tea and vanilla ice cream.  I have to go cook Salmon now.  Apparently I\'m making an ass load for a lot of people. The pressure is on.  I know burritos pretty well, but I kind of lied and said I could cook salmon (\"Salmon? Fuck yeah, i\'ve got a million receipes for the fishy fuck.\") Why did I say I\'d cook salmon for all of these people?  Sometimes I feel like I\'m floating through life randomly grabbing beers and preparing fish.  Kyle told someone I\'m like a little drunk puppy that keeps wandering around and getting lost.  Bark, fuckfaces. '),(2979,1785539,'06,August,2004','2004-08-06 00:00:00','   Sunday afternoon.  Hangover very much a part of my life.  Like a parrot on a pirate\'s shoulder that punches the brain instead of the squawking.  \"B-Man want a migraine?\"  Yesterday ruled.  I woke up in time to pound a couple cups of coffee before the lovely couple picked me up to head out to the amazing Bocce compound party.  The drive out there was beautiful.  I don\'t use that word very often; but, that\'s the only way to describe it.  Fucking beautiful is another way, I guess.   We got to the event around noon.  This house was amazing.  Fucking amazing.  It looks like an architectural mastermind was blindfolded and sent about a forest to create.  This place was like a less-pointy St. Basils; there was lots of stuff and it didn\'t seem like it should flow but it did\". We played 3 games of Bocce ball (1-2 for the B) which I surprised myself by being able to pull out some amazing shots.  I\'m hooked on Bocce.  Hamburgers, oriental sandwiches, Jack Daniels and AmberBock.  More Jack Daniels.  Arrogant B-Man remarks (\"My boyfriend is a part-time cop\" \"Then I guess he\'s a piglet! Hahahahahaha\") More Amber Bock.  Some more Bocce.  More arrogant remarks (\"Well, they did give my idea a movie deal\" \"They gave \'Ernest Goes to Camp\' a movie deal too\").  Great conversation.  Even though a lot of the conversation was about musicals, it\'s amazing how much I care about musicals when I\'m half lit and in the company of theater buffs.  It\'s my ability to blend in and get drunk in any environment (\"Say, bishop, how \'bout a sermon and a Vodka tonic? I tell you, that Christ had one hell of a mother.\")  The other B and I played a rag-tag game of disc golf through this amazing piece of the Northwest (skins match score 2-5.  I can never seem to win anything) which I had to throw from the top of a compost heap at one point.  I argued that should give me at least an extra skin for props sake.  No luck with that. Anyways, I was obnoxious but I met lots of interesting people.  I just hope I didn\'t say anything too jackass to embarass the h.c.  I guess I\'ll find out. After all of that I had a good old fashioned sibling ass kicking argument with Kyle.  Why?  Why not?  He\'s so cute when he clenches his fists and goes \"Brent, I\'m serious\". There was more that happened but I don\'t want to think too hard right now.  Ramones girl is on her way over and I need to shower.  We\'re going to go down to this dock that we found on Mercer Island.  Maybe we\'ll watch some boats and drink some beer.  Maybe we\'ll just makeout in front of little kids.  I\'ll write more later. '),(2980,1785539,'06,August,2004','2004-08-06 00:00:00','   Do I dare it?  A drunk blog? Kyle took his socks off today and asked me to look at the tan lines on his feet.  Here were my responses: -Holy fuck, did Mom fuck a sasquatch? -What the hell?  Did God slap on some fallen trunks to the ends of your legs? - A family of 3 could feed for a year off of one of those goddamned things! - Are those fuckers registered? These feet are too huge.  Imagine a clown with huge red clown shoes.  Stop!  Now take away the shoes and there are my brothers feet.  It\'s like watching evolution forgotten.  \"We will create a race of men with huge feet and they will ... what? An Adam\'s Apple? Yeah, that sounds kind of interesting.  Let\'s go with that.\" So, I\'ve spent the night sipping Labatt Blue and calling people in Columbia I haven\'t talked to since I\'ve left.  Actually, it\'s mostly been people I forgot to tell I was leaving.  So, most of my night has gone like this ... B: Hello, someone I once knew Columbia People: B-Man!  Holy shit!  I\'m so drunk right now, where are you?  Are you still in Seattle? B: Yep.  I think I\'ll be here for a while.  Things are pretty good out here. Columbia People: I\'m so glad for you! Some random girl you used to know just passed out on my arm .... talk to her! Random girl: Brent? B: Yea, Hi, It\'s Brent. Random Girl: Where are you?  When are you coming back? B: Seattle.  You know what, this is getting boring put the other person on the phone.  Aww, I miss those guys.  Nowhere else have I found that you can walk around completely shitfaced and yelling \"Lube her up, I\'m not drunk yet!\".  Oh, Five Points, one of the last vistages where a P.B.R. chugging fratboy past his prime can find bliss.  For those of you that have never been there, imagine all that you know about the Old South with it\'s Hoop Skirts and Gone With the Wind\'s suddenly ripped down the seam, and replaced by Abercrombie and Bud Light.  (Upon later drunken editing I\'ve decided it\'s best to start playing Grand Funk Railroad\'s \"I\'m Your Captain\" at this point). Here you find the diehard and less political drunks out there sipping on $2 Beam and Cokes, but they stumble in between hoards of hair-spiked, \"Abercrombie\'s Ice Shed\" t-shirt wearing,  paying $4 for a Sam Adams trying-to-be-yuppies-in-the-middle-of-the-deadest-economy-in-the-country- boys who are somewhere between their sophmore and senior year.   Late at night, in the colder wind-whip of 5 Points, the pretty frat boys would run for cover and give us true alcohol soaked boys of the dark a chance to shine.  \"Oh, that\'s right, you said your mom was dead\".  \"You know what, this shot would look better if we were making out\".  \"Look, I\'m not going to remember your name; where the fuck did I park?\" And in the morning it was O.K.  It happened in 5 Points, somewhere between Columbia and a college sophmore\'s Heaven. I need to shower. '),(2981,1785539,'06,August,2004','2004-08-06 00:00:00','   I forgot, dorkbabies. If you want to contact me, whether I know you or not, my email address is capt_zeppelin@hotmail.com This is kind of an experiment.  I\'ve always thought about starting up a web business where I promise to respond to people\'s e-mails within 24 hours.  You know, help them the best I can with a problem, or be a buddy they can write to when they meet a girl or hit a double play during intramurals.  It\'d be kind of like being in a frat; you know, paying for friends.  Booyeah, take that you fratboy pieces of .... oh yeah. So, anyways, write me and I promise you that one of our staff will respond to your comment or concern within one B-business day, which is a 24 hour party day, baby! '),(2982,1785539,'06,August,2004','2004-08-06 00:00:00','   If I was 8 years old today would be one of those days where I\'d convince my mom that I was still a little too sick to go back to school.  Then after everyone went to work it\'d be M.C. Hammer and \"The Fall Guy\" \'til after 5, baby! I would tell you about the ape women and how work tested me and how I bested it like Conan the Fuckbarian; but, I don\'t think so ... Because it\'s 3-day weekend time, baby!  Party over here!  Party over there!  Throw your hands in the air like you\'s a true playa!  All the ladies say \"Heeey\"!   Today I\'ve been in the mood for music from my high school days.  So, it\'s been The Beastie Boys\' \"License to Ill\" and Ice Cube\'s \"Death Certificate\".   Went out to eat downtown last night with R. Girl.  I treated her to a lil\' spaghetti dinner (the worst I\'ve ever had, my meat sauce was PINK!) b/c for some reason we have to reward people for getting a new car.  It still makes more sense than tipping your hairdresser, I guess (\"Here\'s a tip, throw in a rubdown and we\'ll talk about cash.  Got it, sweets?\").  We went to a popular restaurant down by the place where they filmed \"The Real World: Seattle\".  While we waited for a table we looked for \"party boat\" and watched dolphins swimming in the exact spot where Stephen slapped Irene!  And for you Seattle-ites, you know where I ate.  Yes, I sold out, but the convenience factor was so convenient. The table next to us was this hyper-nuclear family with 5 little kids.  Three of them were singing this song that had only one line, but that was stuck in my head for the rest of the night. \"I love my Chi--nese food\". One little girl would only say \"I hate my Chi--nese food\".  I agreed, but mine was spaghetti. The other kid was this dark-haired ugly fatty in the midst of these blonde haired Heidi clones.  Damn, it has to suck not to look like the rest of your family AND be the ugly one ... Kyle? I had awesome dreams last night!  The one I clearly remember was being at a restaurant eating with some random people I know when all of a sudden a younger Ronald Reagan comes up and says \"B-Man why don\'t you look me up when you\'re in Chicago next week so we can catch up\".   \"Cool, Reagan,\" I said, \"Why don\'t you call up General Patton and we\'ll all go grab a beer.\"   \"Sounds good, I gotta run,\" said Ronny.   I felt like the diplomatic shit. An overheard story from today: \"Do you like my purse?  My house cleaner brought it back for me from Uraguay.  She\'s such a wonderful woman.  She\'s so sweet and tries so hard to get ahead.  I do what I can to help her out, because she has a hand in my home ... I mean my heart.\"  I thought the first one was much more honest.  Fuck yeah, take care of the people who have hands in your home.  Otherwise, you might wake up missing a Playstation2 one day. Well, I think it\'s time for me to get this party started B-styley.  I\'ll holla at a dork.  Damn, I feel hip hop.  Where\'s my Kangol hat? '),(2983,1785539,'06,August,2004','2004-08-06 00:00:00','   I\'m feeling better today, thanks for asking. I have that post-cold crummy feeling.  I\'m randomly breaking out into cold sweats and cigarettes taste like burnt glue.   I\'ve made plans for frisbee golf with R. Girl, but I think I\'ll have to cancel.  I don\'t think the body is up for 18, or even 3, holes of whipping discs and walking.  Instead I think we should go on a little jaunt about the city in her new car.  I love new car days.  They are very rare, and I feel we should take full advantage of this one.  You know, riding around going \"This is nice\" and \"How fast will it go?\".   I really don\'t have much news because all I\'ve done is slept and worked in the past 48 hours.  I did have some random dreams.  In one I returned to Columbia to say hello to everybody.  Everything had changed and nothing was in the right place.  Walking the streets made me very nervous and I kept slipping on wet floors and icy sidewalks.  Everyone I knew was in a very large park for what seemed to be a \"Welcome Home B\" party.  Instead, it was just a huge conincidence.  I was drunk and people kept saying to me \"You shouldn\'t have done that\" and I kept saying \"I don\'t care what I should have done; I did it\".  I quickly ran back to Seattle. When I got back here, in my dream, I was in a huge field looking up on a large 200 foot tall snowy cliff.  There were about 50-or-so snowboarders jumping off of it.  They were falling about 100 feet each and then landing in a large snow pile.  I kept thinking \"Are these kids cool?\"  I decided they weren\'t and kept wondering if they were hurting themselves.  They weren\'t getting up from the snow drift and more kept jumping on them.  Weird and vivid shit. I really smell like cigarettes.  I\'ve been trying to catch up on my blood to nicotine ratio and now I smell like my grandpa pre-cancer.  I need to wash up and drown in some Listerine.  I\'ll have more to say, dorkbabies, once I actually get out of here and enjoy this fucking beautiful Seattle day. '),(2984,1785539,'06,August,2004','2004-08-06 00:00:00','   Dorks. '),(2985,1785539,'06,August,2004','2004-08-06 00:00:00','   Two things learned from this cold. 1.) They use Sudafed-type products to make crystal meth, or crank if you like.  When you buy it at stores, old women look at you like you\'re a fiend.  If I had known earlier I would have twitched and asked where the best place to boost was. 2.) Blowing in someones ear and coughing in it produce two different affects. I don\'t recommend the later. I left work after only one hour today.  Right when I showed up with my head Sudafeded out (I was tweaking!), my nose running and N.I.D. Girl screaming \"Clyde, I think bagels are good\", I knew it was going to have to be a sick day.  What clenched it though was N.I.D. Girl turning to me and saying \"This day has been from Poopityville\".  I agree, freak; I\'m out. I slept and I had those fun halucanegenic (I don\'t care if I misspelt that, the dictionary is too heavy to pick up right now) dreams that come with a cold and crank.  I dreamt about dinosaurs and laser tag and for some reason it was fucking important for me to get to the grocery store.  Last night I dreamt that I had stolen a car and moved out to a Pacific Northrwest island to oversee a paper mill.  Anybody with one of them dream analyzing books want to take a stab at those? Today I\'m going to avoid sunlight, eat soup and try to ignore the nicotine addiction that is telling me to \"Fuck the chest; light up\".  I know he\'ll win eventually, just like my thirst for Sudafed will call me out into the Seattle streets looking for meth dealers and drag bars.  I like pretending that I\'m skirting the edges of hard drugs by taking these little white pills.  This might be why I eat poppy seed muffins and drink Coke. Last night I watched \"The Burbs\" with the comedic brilliance of Younger Tom Hanks.  The movie dragged more than I remember from watching it as a kid, but I blame that on Jerry Bruckheimer.  A great flick for stuffy nights.  Best parts: \"This is Walter!\" \"AHHHHHHHH\" and Tommy boy coming out of the ruins of the neighbors house. All right, I need to go watch T.V. or do laundry or eat.  I don\'t think I have the energy for all three.  Maybe an addiction to speed would come in handy. '),(2986,1785539,'06,August,2004','2004-08-06 00:00:00','   The cold I was dodging and drowning in Budweiser last week finally caught up with me.  I feel stuffed up and cold.  This damned thing came on fast too.  I was sitting around watching \"Dude where\'s my Car?\" after playing Bocce ball with the lovely couple when I became instantly sick.  No build up, I went from not sick to sick and now I\'m cold fever chest cough sick. I don\'t feel like writing and I don\'t have anything much more to say except it is hard to write.   I\'ll try and write later, but I don\'t think you want to know about runny nose updates.  So, it\'s time for soup and cough drops, thanks to the lovely advice of a very special girl named Liz.  Thanks, slut, and goodnight, dorks. '),(2987,1785539,'06,August,2004','2004-08-06 00:00:00','   So much to tell you about, sweet things. I\'ve had such a great and busy past few days.  But first I want to tell you what happened to my brother on Friday. Kyle is the blonder and more sober version of me.  He kind of floats through life with, as far as I can tell, the mindset of a 12-year-old suddenly thrown into a 21-year-old\'s body.  Anyways, this is what happens when my little brother goes boating. Kyle and his work buddies were out on a boat (I\'m guessing in the Puget Sound) and everyone is enjoying the scenery.  Well, except for Kyle; Kyle becomes obsessed with a duck that is floating near the boat.  His buddies challenge him to catch the duck and Kyle hops in the water to get it.  The duck gets away but this just pisses Kyle off.  After getting back on the boat Kyle goes \"We\'re catching that fucking duck\".  My words coming out of an aryan body.   So this boatload of dweebs goes tearing ass around for half an hour after this damned duck.  I can only imagine what the thing was thinking. \"Quack. Why? Quack.\"  They can\'t catch the duck and Kyle gets to his breaking point.   \"Pull up along side the duck at full speed.  I\'m going to jump out and grab him and punch him and break his fucking neck, then I\'m going to stuff the fucker and put him in my office.\"  Good idea, dork. So Kyle jumps off the side of the boat as they pass the duck going shitfuck fast.  Only he doesn\'t tell the guy driving the boat that he\'s about to jump.   Here\'s what happens when this scenario is set up.   While Kyle is in the air and before he hits the water, the boat turns sharply so that Kyle is now face to face with the front of the boat.  As Kyle hits the water, the boat hits Kyle.  Kyle\'s friends say they looked over the side and saw his legs \"going really fast\" in the opposite direction of the boats path.  Kyle is shot underwater by the force of the boat for about 10 yards and suddenly pops up way the fuck away from it.  He was O.K. but he didn\'t get the duck.   This is my new favorite Kyle story.  Mom just called and asked how everyone was.  \"Kyle got hit by a boat\"  \"What, what, what?\" \"Yeah, you should probably call him.\" Onto the B news. Work was work.  A goateed man ignored elevator rules again today.  This shit is getting weird. Camping was super fucking great.  After a late start (thanks to two late night Mickey\'s 40 oz.s) we got down to the state park in Ocean Shores.  When we get there there is a huge sign that says \"Camp Full\".  \"This doesn\'t apply to us,\" I say, and I proceed to go find me a park ranger.   \"Is the camp full?\" I said. \"You know it\'s full; you saw that big brown sign,\" said the park ranger.  And you can\'t sweet talk a park ranger, friends.  They\'ve heard it all.  \"C\'mon, sweet stuff, I\'ll let you see what I got for some kindling\". So, we went on to the next camping site.  This place was like that scene in National Lampoon\'s Vacation, you know the one after the one where the two girls get high.  Here\'s another conversation I had. \"Are you full?\" \"Only tent spots we got are next to that table out there.\" Next to the table also meant next to the road, a thorny bush and this guys office.  And this guy meant swatting flys and drool, he was all about that shit. \"You don\'t have any of those wooded areas?\" \"What? No! Those are all gone, I just said that!\" \"Yeah, I know.\" So, we went to the next camping site.  This place was like a yuppies wet dream.  For 10 bucks you got a picnic table, a fire pit and your own personal log to sit on and tell ghost stories.  We pitched our tent ... oh wait, about our tent.  This thing is fucking precious.  It became our common child.  It\'s a two person trail tent from Coleman.  It\'s one of those ones you\'d see Boy Scouts from the 1950s using.  It\'s this tiny red triangle thing with a yellow tarp bottom.  It was so cute that I wanted to makeout with it. After setting \"Tenty\" up, we headed out for the ocean.  But this wasn\'t as easy as going to Myrtle Beach and falling in the Ocean, you lucky East Coast fuckers.  We had to trudge through underbrush and then cross this treacherous log bridge across an inlet, which we let some teenagers cross before us.  \"What are you just going to watch us to see if it\'s safe?\" \"Fuck yeah I am.\"  After that we had to trudge through sand that would sink about 6 inches with every step.  After about 200 yards of this we finally reached the ocean, which was foamy brown and cold as a Minnesota handjob.  (Great fucking simile, B). It was gorgeous though.  The waves were choppy and harsh.  The wind made nipples hard in beach fashion but for different reasons.  I needed a sweatshirt and boots instead of my t-shirt and flip flops. Standing there on Danger Beach I realized that it would never be corrupted by dance clubs or cokehead bikini skanks.  Never would this beach hear \"Daddy, take me to the arcade\" or \"Hoooeee\".  It was perfect, and I felt like Leonardo DiCaprio having gotten there.   We walked along the beach looking for full sand dollars.  I\'ve never found an unfucked one.  Maybe that is my lot in life.  After a good while of looking we headed back.  R. Girl fell into the inlet when the log bridge gave out under her.  \"Sucks for you,\" I said and decided to strap on a pair and just wade the fucker.  I did, and I have the cock to prove it. I showed baby doll how to play frisbee golf back at the camp.  I ran barefoot through a patch of briars and got about a hundred stuck to the bottom of my feet.  \"Sucks for you,\" she said.  Actually she said I looked cute pulling them out with the blood on my fingertips.  Weird shit is cute to some people, man. We then went and bought firewood from a very crazy old man on the side of the road.  He told us he was the mayor of a Seattle subarb and that his face was on a statue of a very well hung dog.  I don\'t understand either, dorks.  He took me over to his truck and showed me how to judge the quality of firewood.  By his standards, I have to say, he did have some fucking great firewood.  He liked me and told me to \"stay here\" he wanted to give me something.  \"What could it be?\" I thought as he started walking deep into the woods.  Gold?  A treasure map?  A hatchet to my jugular?  Nope, just more wood.  But special wood he said to make our fire warmer.   Side note: This whole situation made me very nervous.  However, R. Girl was having a fucking ball with this guy.  After reading that \"The name Brent means you suck\" thing (posted below) the other night I have become worried that I will end up bitter and with a fucked up solar plexis.  I have started taking cues from other people on how to handle situations that make me nervous, like crazy old men telling me they want to show me something.  Normally I\'d say \"Not today, rapist\" and run away.  Instead I followed R. Girl\'s lead and laughed along.   We shook crazy fuck\'s hand and drove off.  He can be found somewhere on the \"what do they call it ... the intro-net?\", but he couldn\'t remember where.  Great guy. I made fire.  I am all that is beast and ape.  I took wood and I built the blaze of blazes which cooked the meat of dead animals for me and my woman.  I am a primal sweatshirted God. We had an early night, doing nothing with Ramones Girl is a very full day.  The less we do the more fun we have.  It\'s perfect for being a cheap anxiety-ridden ass. The tent started dripping condensation at some point in the night and when we woke up everything was wet.  And not in the good way, dorks.  I was up at 4 and had a fire going by the time the sun started to come up.  We decided to skip the camping breakfast since we were both damp and grumpy and find us a Denny\'s. All in all the camping trip was the shizzy and I think I could really like Ramone\'s Girl if things keep up the way they are.  But, that\'s personal, so back off you tabloid fucks. Other things to discuss: -The Twins are 1.5 games back, baby fucks!  They survived interleague play and can smell the spoils of the Fall Classic.   \"My baby does the Homer Hanky!\" Eh? Eh? -No Inner Dialogue girl was suprisingly quiet today.  \"My stapler smells like toothpaste,\" was the extent of her observations.  Although she did stab at me with her pen when I walked by her desk.  We\'ve never even talked and she\'s trying to stab me.  There goes my solar plexus. -Eva Moore is super cool.  Check our her fun adventures in linguistics and sexy alternativeness at www.clevertitle.150m.com  -I downloaded some stuff from www.suicidegirls.com from Kazaa Lite.  I was expecting buck naked goth girls going at it with Barbie Dolls and whips.  Instead I find out it\'s like some kissing, a couple tit shots and mostly just chicks telling stories like \"I like punk music\" or \"One time I had to walk down a hill\".  What a gip.  I\'m starting my own porn site.  It\'s going to be www.chickswhorocknextdoortoyourwhileyouarewatchingseinfeld.com Coming soon. I grow weary of writing now.  I\'m going to go play frisbee golf.  I love you. '),(2988,1785539,'06,August,2004','2004-08-06 00:00:00','   Just wanted to let you know I was thinking about you, dorks.  It\'s a perfect early fall day here.  I require plaid and football.  I\'ll tell you all of my dirty little camping stories later.  Outs. '),(2989,1785539,'06,August,2004','2004-08-06 00:00:00','   On those nights where you are all alone because sweethearts are tired and brothers are out watching gang members fight with his new found Justice Dorks of America, you sit alone and watch movies.  Then you make sandwiches.  Then you look up porn. And as a straight male all alone on those nights you ask yourself this question. \"Am I attracted to transsexuals?\" \"I am attracted to pretty girls.  I am not attracted to pretty boys.  But what if a pretty boy became a pretty girl?  Would that be passable prom date material for me?\" Then you take to the internet.  How these questions were answered pre-internet explosion I do not know. Come to find out, a lot of transsexuals are Brazillian.  Or so they say.  If they are telling the truth someone should really look up why so many pretty boys in Brazil have forgone soccer for transsexual porn.   \"Welcome to Brazil.  If you have a penis you can stuff it into soccer shorts or a thong.\" But all transsexuals have gone for that skanky \"I just got back from Myrtle Beach and got hair braids\" look.  I don\'t like that look and so I decided that I wasn\'t attracted to transsexuals.   I don\'t know if they have any Catholic school girl transsexuals (I\'m sure some preist would know) or Lauren Ambrose or Kirstie Alley chicks with dicks.  Those might do it.  Well, at least until they took their thongs off.  Not so sure I could swing that bat. Sweet dreams about that, dorks. '),(2990,1785539,'06,August,2004','2004-08-06 00:00:00','   I forgot to mention in the \"10 Things You Should Know About B\" thing at the begining of this silly thing that I\'m highly political.  If you want to talk a little jive, e-mail me at capt_zeppelin@hotmail.com '),(2991,1785539,'06,August,2004','2004-08-06 00:00:00','   I just finished watching \"Bowling for Columbine\".  The cover of the DVD has a review that says \"Funny\".  Yeah, I laughed; but, the they should have had a review that said \"Honest\" instead. I don\'t want to talk about the specifics of the movie or how I\'m kicking my ass for not going to see it in the ONE Columbia theater that showed it.  I knew all of the facts that were said in the movie.  I have studied poli sci, media literacy and race relations.  What I want to talk about is the one truth that was said in the movie. We are mad at nothing. I just gave Kyle my theroy that the U.S. is a high roller strung out from a 20 year poker game.  We won the cold war (purposely left in lower caps, fuckers) when Ronald Reagan played history\'s biggest bluff known as \"Star Wars\".  We were not going to put lasers in space.  We had no intention on building this silly project.  (In fact if you ever see some old coverage of that silly speech, note the reaction on half the people\'s faces ... I\'ve never seen more \"yeah, right\"s in my life).  It was a bluff.   http://www.geocities.com/CapitolHill/Lobby/1777/poly/rrcoldwr.html Russia had built up a nuclear arsenal larger than ours.  But, really, how many times can you blow up the world?   So, we bluff that we\'re getting ready to build this defense system that would make nuclear weapons obsolete.  Imagine you\'re a Russian. \"What?  Really?  We put all of our fucking money into this goddamned arms race and they are going to make it obsolete?  Can they do that?  Yes?  Oh, shit, we\'re fucked.\" That was it, the worlds greatest bluff ended the cold war. But, if you are America you can\'t just say, \"Just fucking with you.\" (P.S. I\'m not a Republican, but it\'s O.K. to admit that the other guys came up with some good ideas) Everything since that point has been a super inflated hyper active arms race to beat nobody.  Which brings me back to the poker game. We have noone to play but we are so excited that we won the \"big match\" that we have been on a 20 year(ish) bender.  \"Who\'s going to play me next?  You? You? You?\"  No one has ever wanted to play. \"Fuck that. Panama, you play him.\" But, we Shanghaied us a couple of apponents.  Now the rest of the spectators get nervous.  \"Maybe we should punch the fucking hyper Maverick before he gets us next.\"  I would. Which brings me to my big point that I wanted to say after watching this movie.  I have seen the pointless anger that is incorporated in us by a chase for goods and a lifestyle we can\'t ever reach.  Even those skinny blonde fucks in the ads go home to cocaine and bitchy girlfriends, people.   Wrap this up, Brent.  I wanted to share with you 2 typical Americans who are (or were) scared of the world outside. Case study #1:  When I was working at the great supermarket in high school this dude came in one day.  I was bagging his groceries and noticed he had F.T.W. tattoed on his arm.  Unlearned in the was of the angry world as I was, I asked what it meant. \"Fuck the world.\" I didn\'t say another word.  But I kept thinking, \"Why is he so mad?\" \"Fuck the world?  Fuck you!  Carry your own goddamned groceries that I just bagged for you so I can make enough money not to ask my single mother for gas money to go to school and take my fat girlfriend to a movie where I won\'t make out.  Fuck the world?  Buddy, I am the world.\" The second was this guy from just the other day.  It was when me and R. Girl were heading out to camp.  We were driving down the interstate when all of a sudden this fucker pulls into my lane.  He ran me off the road, which, naturally, upset me.  But the guy has the nerve to look in his rearview mirror and start telling me to \"Fuck off\" while giving ME the finger.  He cut ME off!  It was ME that almost died!  After a few seconds of me screaming \"I\'ll fuck your world\" and R. Girl giving him the doulbe bird (you know, both hands at once) did the guy calm down and realize his mistake.  He gave the \"I\"m sorry\" wave and as a citizen of this world I had no choice but to accept it and give the all encompasing \"It\'s ok, you\'re a good guy\" tilt of the hat. Good stories.  But within them are what\'s wrong with this country.  Daddy hated us or we got arrested for crank possesion and all of a sudden we say \"Fuck the World\" every time we reach for something?  Are we so caught up on fucking others over that we don\'t look at our OBVIOUS mistakes and realize we were wrong.   We can\'t even say we\'re wrong.  We have to tattoo ourselves to protect our silly lifestyles that are fake and based on GAP ads anyways.   You\'re mad at someone for something.  Don\'t fuck the world. You cut me off.  Calm down.  We lived.  I flip you off, you shake your head like an idiot and we drive on. There\'s no need for any of this anger.  Who are we mad at?  I think ourselves.  We watch 9/11, we watch Columbine, we watch all of these things and we KNOW that it\'s not the kids or the Middle Easterners.  It\'s us.   That\'s why we are mad.  We cut off the world and don\'t want to admit we aren\'t as cool as we thought.  So let\'s all run out and tattoo our arms, right? Think about that, dorks. '),(2992,1785539,'06,August,2004','2004-08-06 00:00:00','   I left work early today.  I was done playing Roy.  Before I left I did the \"Look At Brent\" round.  I don\'t know if this is a common office practice; it\'d be cool to say I invented the move.  On days where I want to leave early I go on a little jaunt around the office and talk to as many people as I can.  \"Say, fella with that box, how\'s the U.P.S. situation?\" \"Oh, hey, female superior, did you get that e-mail about that thing that involves something in Houston?\"  \"Say, person at the coffee machine, did you hear the one about the Jew, the penguin and the abortionist?\" \"Well, that\'s unappropriate, Brent.\" \"That\'s right.  And remember I said it at roughly 4-4:30 p.m. on Wednesday.\" The idea is that if anyone asks for me in the half hour to hour that I\'m supposed to be there, plenty of people will say \"I just saw him\" and lead whoever is looking for me to assume I\'m somewhere else.  And when 4:30 rolls around, \"I guess he just left\" will be in their head. The more I work this office job the truer my earlier theory about office work becomes.  People in offices don\'t really do anything.  Collectively, sure, they might spit something out.  But, as a single human, no one person is needed, ever. Over the past couple of days I have started to reach out to N.I.D. Girl.  After the Labor Day weekend I was sure she\'d be full of stories.  \"And then there were hamburgers and I ate one.  And then we saw clouds.  And then we drove to buy Listerine at the store.  And then I payed for the Listerine.  I like the blue kind.  The yellow kind hurts.  Hurts my mouth.  My mouth is on my head.  My head is connected to my body by my neck.  I once saw a man with a big neck.  Where are my fingers?  Oh, on my hand.\" Instead she only had one story, which, granted she did tell everyone that came by her desk.  \"I watched my sister\'s kid and she was late.  She said she was sorry.  I didn\'t have anywhere to go, but I was still upset.  The kid is a wild one.\" This made me sad.  The extent of her 3 day weekend was a little kid.  The extent of my 3 day weekend was balloons, making out, statues, getting ripped and Seinfeld.  Comparitively I win the weekend contest.   I realized that this woman doesn\'t have a boyfriend, or partner I guess I should say in the progressive and heavy sexual alternative lifestyle city of Seattle.  Is being gay still considered \"alternative\"?  Are the \"Spin Doctors\" still considered the same?  Alternative, not gay. Back to N.I.D. Girl.  I couldn\'t help but feel sorry that this woman only has this job to come to and talk to C. from Washington Mutual about what is happening to her at that exact moment.  Because, that exact moment is all this woman has.  She doesn\'t have a dorky brother to make fun of and try to get drunk, she doesn\'t have someone to kiss and tell they are sexy, she doesn\'t have a Lovely Couple to play alternative (not gay or Spin Doctors) sports with.   So, I\'ve started talking to her.  And I\'ve learned that if I speak to her I can stear the course of our conversations.  It might be harsh to hear her voice but at least I hear \"McDonalds is good\" compared to \"Boys are not the same as girls\". One thing she did say today that made me look up to see if it was actually true, because this is N.I.D. Girl we are talking about.  \"My ears are on fire\".  I was half expecting to see her ears fucking blazing when I looked up.  Oh well, it\'s still funny to think about. I am gross.  I spent the night with R. Girl last night, dorkbabies.  And that\'s all you get out of that.  She\'s cute.  There, there\'s a little tidbit of smut for your sick minds.  Go read porn! Anyways, I didn\'t want to leave this morning and calculated the Shower/Sleep ratio.  You know, it\'s a highly mathematical formula where you figure out the number of hours since your last shower, factor in number of physically exerting things done and relative ratio of closeness to stinky stuff and then come out with a number (x) which is then compared to the ammount of time it takes to shower (y).  If  x is less than y there is no need to shower and sleep can continue.   However, now I feel gross and nervous about skipping out on work.  I will shower and then watch Seinfeld or go play frisbee golf (which I\'ve fallen out of rotation of the past couple of weeks).   I hope all of the dorks out there are good and healthy little dorks.  I\'m sorry I haven\'t written much in the past couple of days.  I have been busy being caught up in the fun parts of a new relationship.  You know, \"Do you like Led Zeppelin\" and \"Is that a freckle?\" Later, dorks. '),(2993,1785539,'06,August,2004','2004-08-06 00:00:00','   I did nothing today. I could bore you and talk about how I cleaned my truck or downloaded 20 Seinfeld episodes.  But, you don\'t care and I don\'t want to bore you with that. Instead, in honor of Labor Day, which I told R. Girl should be renamed Non-Labor Day, I think I\'ll play James Brown\'s \"I Feel Good\" and tell you all a bad  joke. How do you spot an elephant on the moon? He\'s the one with a giant \"E\" on his space suit. Happy Labor Day, dorks. '),(2994,1785539,'06,August,2004','2004-08-06 00:00:00','   I am sad. Charles Bronson died on Saturday. This was a man who said his face looked \"like a rock quarry that someone has dynamited\".  He was the Tunnel King in the movie that bonded me and my father.  He was the man who put nails through a board that sprung up and killed trespassers in \"Death Wish\".   He was the man who I shared a birthday with. Every November 3rd I would say to myself \"Happy Birthday, Charles Bronson\". I need to watch \"The Great Escape\" and cry. Live life, dorks. '),(2995,1785539,'06,August,2004','2004-08-06 00:00:00','   There\'s a nice cool breeze blowing out the remains of a perfect Seattle day.  Friday night (in between drunk blogging):  Kyle had a dorkfest over here.  These are the kind of dorks I like though.  They drink beer and when I say things like \"I hope South Carolina kicks ass tomorrow!\"  they nod their heads and pretend to care.   We have discovered another thing that I do great with a few beers in me ... Soul Caliber 2.  It\'s this game that Kyle bought a few days ago and these dorks were going ape shit over on Friday.  After an hour or so of feeble attempts to get them to wrestle me (\"C\'mon I\'m the DorkKrusher, Bring it On!\") I decided to give their game a whirl.  I kicked their ass so hard and began to realize that I had played the first game in between drinking and class when I was a sophomore (\"C\'mon, professor, I\'m the ProfKrusher, Bring It On!\").  That was fun and then I passed out. Saturday morning I woke up late and in a non-showered sweaty hurry to find a present for Great Cousin because the Lovely Couple was throwing a b-day party for her that afternoon.  Kyle had come down with my cold and since I was hungover on top of that we bickered through every store that we were in. \"Dammit, dude, she doesn\'t read comic books!\" \"Dude, I\'m going to smack you. Shut up and get the hell away from the video games.  We\'re on a mission!\" \"No, Mom loves ME more.  She told me so ... wait, what about balloons?\" So, we ended up getting a lovely arrangement of balloons.  In the center of the arrangement was a Harry Potter balloon b/c Great Cousin loves Harry Potter. Here\'s how Kyle fucked that up. After getting together with Ramone\'s Girl (yes, this would be the day she would meet the family beyond the dorky blonde version of me) we decided to head over to the Lovely Couples for some par-taying. \"Dude, are you sure you want to take the balloons in a convertible?\" \"Brent, look, it\'s only a 3 block drive.  Nothings going to happen to them.\" \"Allright, but I\'m following behind you to make sure.\" \"Whatever, Brent, you know, sometimes ....\" \"Blow me.\" End of conversation.  So, me and R. Girl are following behind King Dork when not 30 seconds into the drive (but they were a fun 30 seconds of watching the balloons hit his face repeatedly) the fucking balloons fly out the back of his car and almost hit my truck.  They land behind me and in the center of the road.  The people behind us are verring and swerving to avoid the balloon baracade.  After a few minutes of pure Stoogian comedy we get the balloons back from a car full of girls who I\'m sure Kyle hit on in the process.  \"So, you hang out on the side of this busy road often?\" To cut the story short 30 more seconds into the ride the Harry Potter balloon sails off behind the car and was never heard from again. The party was great.  The Lovely Couple knows so many interesting people and it\'s always fun to sit around and have conversation time with them.  Ten minutes after being there I had forgotten that I was missing College Football kick-off Day.  Later on in the day we played a couple rounds of Bocce Ball.  It was boys v. girls (finally score 1-1).  R. Girl turned out to be a natural ass kicker at Bocce and became the ringer for the girls side.  I, on the other hand, was quickly labeled as the not-good player for the guys.  You know you\'re not doing good when your own team laughs when the other team says \"Oh, too bad you suck, Brent\". A great day nonetheless, and everyone liked Ramones Sweety.  I was nervous about the whole meeting, to be honest.  I don\'t know why, I guess I\'m just used to friends and family saying things about other girls I\'ve known like, \"Umm, does she abuse cocaine?\" \"Oh, yeah, she seems nice ... for a retard\" or \"Brent, that\'s obviously a man.\"  And, you know what, deep down I knew Suzzie-Carl was a man. It was an early night for all.  I decided to pass up on earlier promises to \"Get ripped and kick a little ass\".  R. Girl started having girl problems and the night was kept quiet and full of \"Your cute\" and \"Do you want me to rub your back?\". Today I slept in until noon and skipped showering.  I love being a man on Labor Day weekend.   After meeting R. Girl\'s parents and aunt in a rushed attempt to get medicine for girl problems, we went downtown.  We wanted to find a couple of statues. The first was the one of the dog with the face of the man that had sold us firewood last week.  We found it, but the mystery of whether it was the guy is still afoot. The next was a quick stop off at the Troll under the Fremont Bridge.  This thing is a huge concrete fuckbeast of an artpiece.  http://www.roadsideamerica.com/attract/WASEAtroll.html  Check it out here.  It can descibe better and with more sober brevity than I. That was about it, you perverts. I\'m tired of writing now.  It feels like work and this is my labor day weekend.  I\'m going to get ripped and watch \"Road Warrior\" and \"Animal House\".   You are all so fucking cute, ya dorks. '),(2996,1785539,'06,August,2004','2004-08-06 00:00:00','   And Jimmy Buffet\'s \"Margaritaville\" plays from No Inner Dialogue Girl\'s desk.  Sometimes life lacks couth. '),(2997,1785539,'06,August,2004','2004-08-06 00:00:00','   I love my fucking life, sweet dorks. A few minutes ago, at work, I mention that I wanted to go see \"Laser Led Zeppelin\" at the Science Center downtown.  The lady at the desk nearby pipes up, \"My friend is the manager there\" and after a phone call \"O.K. I have 4 tickets for you to go tomorrow night\".  Sweet. So, I decide to get a cup of coffee to celebrate.  I pass by Cool Guy\'s desk and stop to tell him about the game last night.  After a little sports swap I say \"Yeah, I\'d like to go see a Huskies game this year;  I sure miss college football.  Maybe I\'ll call up my frat and see if I can score some tickets.\"  Turns out that Cool Guy\'s good friend is the Alumni Chair for my fraternity.  I\'m alum, and now I have an in for student tickets at UW. And this is how my life has gone always.   \"I think this thing is cool,\" I\'ll say. Ring, goes the phone. \"Hey, B, it\'s someone you know.  You know that new cool thing you like?  Well, a semi truck flipped over in front of my house and now we have a bunch of them all over our yard.  Want one?\" \"Yes,\" I\'ll reply. \"Oh, and we have too much beer in our fridge; can you drink it?  Also, our just-turned 18 years old slutty daughter needs a prom date.\" Sometimes I wonder if it\'s a joke played by Fate that I suffer from low-level depression. '),(2998,1785539,'06,August,2004','2004-08-06 00:00:00','   So, it was Texas and Seattle, not Anaheim. And I didn\'t actually use the centerfield tickets last night like I wrote about. I used the box seat tickets that we got when some dude pulled up at the bus stop in an IROC or a Z and handed them over. He had adult braces, which always makes me a little uncomfortable. The seats were between 3rd base and home plate and about 30 rows back.  It was awesome.  We showed up around the 3rd inning because we wanted to go grab some bourbon from this little bar downtown that boasts \"The world\'s largest selection of bourbon\".  It was a lot of fucking bourbon to choose from.  And good too. The electricity went out at Safeco Field when someone drove a truck into a power line near the stadium.  I felt like I was back in the South again for a minute.  http://seattlepi.nwsource.com/baseball/139038_lightsout10ww.html I need to work ... or play space genocide.  I\'ll decide over some coffee. '),(2999,1785539,'06,August,2004','2004-08-06 00:00:00','   Take B out to the ball game. Take B out to the crowd. Buy B some beer and a hat watch him run on the field and steal a bat. So, let\'s loot loot loot the souvenier stands  God, I love this fucking game. Cause it\'s 4, 15, 34 beers and you\'re drunk  At the ol\' ball game!  Just got a call from the better part (ha) of the Lovely Couple and he has an extra ticket to go see tonight\'s Seattle/Anaheim game.   \"Want to go?\"  \"Fuck yeah, I do.\" Today has been such a good day.  I woke up so fucking hungover and an hour late for work.  But that was o.k. because of 2 things.  The first is I had an awesome night getting ripped with R. Girl, her friends and King Dork/Batman last night. \"Listen up, girls, I\'ll give one of you 5 dollars if you steal that guy\'s shirt.  I like it.\"  \"Muthafuckin\' STYX, baby!\" The second is that I got to take an amazing hour long nap today during my half hour lunch break.  So smooth. When I got back to work I was instructed to join in with a class about how to do things with computers and files and skulls.  So, I went ... like I was actually going to work after a nap.  Hmph. The class was full of Obvious Staters/News-Fueled Single Moms.  You college kids know these people well.  They are that chick in the class who is older than everyone and who obviously has been through some stuff so she decides to go back to school to show her father/ex-husband/boss that they are empowered and strong.  But instead of being a typical college drunk they sit and class and talk and point out things and talk about their kids. \"When I\'m watching television I don\'t let my child see the gore or the sex.  It makes him hit the dog and use foul language.  I\'ve heard about these movies where people actually lose their ...\" \"Shut the fuck up.\" My office is where these women come after they give up on college b/c none of the hot frat boys (booyea) won\'t bang them and the alimony tuition ran out.  So, when you put them all together it becomes this nonstop babble fest.  A lesson that would have lasted 30 minutes in college took 2 and a half hours. \"I see what your saying, but sometimes when I look at my computer I will see that I have a phone next to it.  I call someone and we talk about files. September 11th.\" \"I agree, but generally when I use a phone there is another person on the other end.  I will say \"Hello\" and they will say \"Hi\" or \"Hello\" also.  And roughly 2-156 minutes later I hang up.  It\'s a process that works for me.  SARS.\" \"I find that it\'s best to say \'Hello\'.  Yeah, baby, very shagadelic.  And when I hang up, I put the receiver back on the cradle. 9/11\" \"I already said that.\" But it was fun b/c I got to sit across from this younger woman in my office who has to be from South Africa or Uganda or something.  The way she talks is melodic.  \"I watch-ed the mo-vie Rush-MORE\".  I sat through the entire meeting imagining her reading me \"Huck Finn\" or The Playboy Advisor.   \"Gener-ally, the male cli-max lasts on-ly a few secONDS.  We should take the raft SOUTH to eventu-ALLY head NORTH.\" Wonderful. Well, I need to go grab some McFuckingDonalds before I head to the game, you nosy fuckers. Yippee, dorks, baseball. '),(3000,1785539,'06,August,2004','2004-08-06 00:00:00','   Folks, we have a winner.  Mz. Eva Moore, that\'s THE Mz. Eva Moore from Perfect Sleeper, has won the first Win Stuff From Brent\'s Whacky Life Contest.  Mz. Moore (why am I using the Mz. title?) will receive the original Roy nameplate as well as detailed instructions on how to use the space commander chair.  And for getting all of today\'s 5 songs at 5 correct (answers: The Monkees, P.M. Dawn, Gladys Night sans Les Pips, R.E.M. and R.E.M.) Mz. Moore has won the grand prize of No Inner Dialogue Girl\'s Sharpie marker!  Stay tuned for more chances to win with the Win Stuff From Brent\'s Whacky Life Contest! \"Where\'s my Sharpie?  It makes black ink go on white paper.  Sometimes the paper is blue!  Red cars pay more for insurance.  The Titanic was a boat!  A water boat!\" Moving on to Frame Stage in a relationship is harder than it sounds.  First you have to say goodbye to pre-frame stage.  You\'ve made a commitment now, buddy.  With frame stage comes the no more having Hustler issues laying around the room responsibility and the knowledge that flower buying is coming up very soon.  Next is toothbrush stage (which I\'ve already accomplished at R. Girl\'s apartment thanks to some desperate emergancy brushing after camping) then the dreaded tampon stage and then you die. But, the hardest part so far has been actually picking out the frame. I bought one yesterday at the Northwest version of Wal-Mart, Fred Meyer (who\'s slogan is \"We charge a dollar more for a 2 liter of Coke b/c we are hipper than Wal-Mart\").  I was half asleep and bought a frame much too big and I had to return it.  Actually, that wasn\'t too hard.  I guess I\'m just used to only having to remember to return my Soprano\'s D.V.D.\'s to the video store and eat McDonalds once a day.   Last night was sleepy night.   Laundry was forgone like a pretty young Brazilian boy\'s hopes of soccer greatness.  Instead I played \"Tiger Woods 2003\" with Kyle. I got him this for his 21st birthday, which is Saturday, by the way.  I played this game a lot in college and it always leads to beer drinking so I figured if he wanted to be a responsible young drunk male he should have this game.  It\'s fun.  Enough. We\'re switiching from UPS to FedEx at work!  Roy would be so jealous, dorks.   I require pysical activity today.  I don\'t think I\'ve told, y\'all, but this damned cold has stayed with me for near 2 weeks now.  It\'s been phlegm and sniffles for 13 days now.  I\'ve been sleepy and rundown and I need some sun and to watch little kids hit each other at football practice.  So, I\'m going to head over to the park across the street from me and try out my new frisbee that I got at the Adidas store on Sunday.   Now, it is time to sneak out the back door at work.  Say a prayer and hum the James Bond theme, dorks, the B\'s about to get crafty. Later, sweetdorks. '),(3001,1785539,'06,August,2004','2004-08-06 00:00:00','   Have I got something for you fans of the dorks out there. Today, while cleaning out my desk at work, I found an authentic \"Roy\" nameplate.  I was going to throw it away, but thought I\'d put it up on Dorkbay first.   So, if you would like an authentic piece of Dorks Don\'t Rock history, contact me.  There will be no charge for this one-of-a-kind piece. Shipping and handling fees will be waved if the buyer promises to put it on their fridge. Hurry, supplies are limited (to one piece) and time is running out for this amazing deal! '),(3002,1785539,'06,August,2004','2004-08-06 00:00:00','   Hey, hey, babydorks.  This has been a great weekend. Were any of you ever in junior high?  Did you ever have that Friday night were your mom dropped you and a girl off at the movies and tossed you a couple extra bucks to get some popcorn.  The movie was \"Tommy Boy\" or some Sandler/Farley/Segal epic.  You had your popcorn you laughed and you stole a few kisses.  After all of that you dropped the girl off (just a few precious but compromising feet away from your mom\'s Aerostar) and stole one last kiss. That was my weekend, you sweet and lovely dorks. Me and R. Girl had that one defining moment in a relationship.  That one moment isn\'t a second, that one moment isn\'t even a minute.  Nobody has ever defined how long a moment can last.  And, if Webster is reading, a moment can last a few days. We had pictures devolped today from our past few weeks together.  I really liked the one of us on top of the troll statue (read back for more info) and asked for doubles. This is a defining moment in a relationship, because.... \"I want to put this picture in a frame ... is that O.K.\" \"Yes it is, why wouldn\'t it be?\" \"Because it\'s the \"Frame Stage\" are we ready to go to that stage yet?\" \"Yes, let\'s go to that stage.\" On Friday I score some very potent and very west coast weed.  I haven\'t done drugs in months.  But when someone hands you a shitface fuckload of potent, grown in grandpa\'s back yard weed, you don\'t say \"No thanks\".   You say \"Who\'s got brownie mix?\" So, upon advice, I poured a quoted \"5 months worth\" of very potent grass into a batter of brownies. Let bake for 25 minutes. Become so shitfaced that .... \"Where do you hide it all\"  \"How about in your neck, shit look at that fucker\" \"Did I have legs\" \"Are there hills?  Help me with the hills.\" \"Did you want the chair? A chair?  Did you want to find the chair?\"  \"What the hell are you talking about\" You wake up at 6 in the fucking afternoon the next day!  6:00 p.m.?  What the hell do you do then?  Drink beer and rent movies, that\'s what. So, today I was more productive.  Me and my sweet, sweet baby went and spent our weekly day downtown.  We had a great day of window shopping and gyros.   An old woman with a camera set up this situation for me. Old woman:  Do you want your picture taken? Me: Yes.  What? Old woman:  Do you want your picture taken?  Do you want the camera, you can have it. Me: Yes, I want the camera. Old woman: Five dollars. Me: No way, I thought you were giving it away because you are high. (Walk on for a moment) British guy: Are you guys not from around here? (This guy is acutally sipping from a STARBUCKS cup ... only 400 yards from where Starbucks first opened.  Very Seattle.) Me and R. Girl: Yes, we are. British guy:  Well, you aught to be careful.  That\'s one of the oldest Seattle scams in the book. (This guy walked near us for 5 minutes without talking.  I know he wanted us to ask him more questions about how to avoid Seattle\'s downfalls, but I didn\'t care ... he made me feel like a tourist asshole enough anyways). What scam? Oh here\'s a camera for 5 dollars!  Fooled you, it\'s only worth 4.50!   A scam!  Like I\'m dumb enough to let some old crack sucking camera whore run off with my money? \"Oh, it\'s a Sony?  Well, here\'s my debit card, go shopping with this.  What? My social security number?  Sure, you DO have a camera.\" So, it was a good day except for British asshole ruining my good natured ribbing of the local crack heads.   Oh and there is the haircut I got on Friday at misogynist barbers. Short and sweet. \"You know what I do with a first date? I tie them up and beat them.  No, wait, how did that comic say it?  I tie them up and drag them by the hair and then I beat them unconcious.  No, it wasn\'t \"unconcious\" ... I think it was \"senseless\".\" So, needless to say, I came out of that place with a MOTHER FUCKING G.I. JOE, MARINE, HAIL TO THE CHIEF, HAND ME A GUN AND WATCH ME SHOOT UP A SCHOOL haircut that is so fucking short that I feel naked without a hat. Ahhhh, at least I have a sweety.  She jokes and tells me she likes it while others go \"Goddamn, did you fuck a weedwhacker?\" I need a new hat, dorks. '),(3003,1785539,'06,August,2004','2004-08-06 00:00:00','   Good morning, dork babies.  I had a great and not very restful weekend.  There were diabetes walks and birthdays.  There was R. Girl, broken toys and great fucking sandwiches.  Today I am sleepy and need just one more fix of a weekend before I can really get this week started.   Like I\'ve said before, I\'m not saying it\'s because it\'s Monday, so back off you fucks. No Inner Dialogue Girl had a good weekend.  She saw Tim\'s band play at the park.  (I have no idea who Tim is, people always do this to me.  They start talking about people I don\'t know and I don\'t care about) There was nobody at the park.  It was a nice park.  She went to Vegas for her 21st birthday.  She was supposed to work.  She didn\'t go to work.  She went to Vegas.  Her coffee is hot.  She needs a bagel. I\'ll write a more all encompasing blog later. '),(3004,1785539,'06,August,2004','2004-08-06 00:00:00','   I have a great fucking idea, my cute dorktits.  We will take down the internet. Or at least part of it. For this to succeed you each need two things. 1.) Consciousnes  and ... 2.) At least one friend And some of the dorks start crying .... When you see a pop-up when visiting the website, remember what the advertisement was for and don\'t buy any of that product that day. For example, I just saw a pop-up for Orbitz, which I think is a fun hip-time drink that kids put vodka in.  Well, I\'m not going to buy some on purpose today.  I wouldn\'t have bought any anyways, but imagine if this thing catches on and all of a sudden Orbitz has a zero profit day.  And then they ask why.  And then Chet tells them \"Dorksdontrock.com told people to do something to fuck with the man that wouldn\'t be obnoxious or require any effort and goddammit it worked!\".  And then I get to be on the Today Show.  Fuck yeah, let\'s do it! The B-Man motto: The easiest thing in the world to do is to not do something. It\'s reverse advertising and I really want this to work.  Please, please tell a friend about my idea.  We would be going against years of brainwashing from Mad Ave. and show those corporate advertising fucks (who I would like to work for one day if the pirate or forest ranger things don\'t pan out) who\'s the thinking man.  Rage against the Orbitz, dorks.  Don\'t not go gently into that convenience store.  Rage, rage against the lying and the hype. Dorkus Thomas, suck it. Did anyone else besides the better half of the Lovely Couple notice that my archives have disapeared?  Part of me has died and that\'s why I kind of have been taking a break from doing this silly typing with my fingers. B:Are you going to see James Taylor? N.I.D. Girl: Maybe, I don\'t have anyone to go with me.  Me is in reference to the person right in front of you.  You is in reference to the body you are living in and the soul that floats in your heart.  Angels put your soul there when you are a babity wabity scoobity doo! B: I\'ll go with you if you can\'t find anyone. So, I might be going on a little pseudo-date with the N.I.D. There will be no tongues and I declare it!  I get enough sugar from R. Girl.  And the good kind of sugar, not that brown cane shit that hippies and yuppies with no tongues put in their ass coffee. Speaking of which, my new favorite thing to do is waking up early when R. Girl spends the night and getting my showering and dressing done quickly and then jumping back into bed for about 5 minutes.  Sweet red-hair-in-my-face bliss, it is nice. Today I play frisbee golf and then pornography. Tickle me pink, dorks, and do my bidding. '),(3005,1785539,'06,August,2004','2004-08-06 00:00:00','   So, like I was telling you dorks yesterday. Last week was testing.  Testing in the way where you fight with swords, not in the way where you write your name in the top right-hand corner and make an A. First off, GreatTits/NoBrain and Uncomfortable Silence Boy were fired.  You see, as interest rates remain marginally high then the influx of new loans remains relatively low to last fiscal quarter, which allows for more space commander playtime but less purpose then before.  This made me fear that the Roy job would soon end so I volunteered to do extra work.  I decided to be come more valuable than gold to save my precious job.  So far it worked, but I was very tired all day last week.  While each and every one of my precious dorks were on my mind I was just a little preoccupied with visions of hobos and a disapointed Roy to do any more than cry and sleep. Not really. So, let\'s just skip over last week.  Last week is dead to me. I have snored my entire life.  I think I have at least.  I always hear people say things like \"Did you fight a walrus in your sleep last night?\". And for 23 years it has been funny.  I snore, other people don\'t sleep, I snore, and then Easter comes in the spring.  Well, I decided recently to buy some of that snore spray such and such and such.  It works.  I even tried the strip things and then ... anyways I sleep better now.   That was an all time low for the blog I think. Friday night: Me, R. Girl and the Lovely Couple all went to a high school football game.  It was a freaking blowout, dorks.  Some school kicked another schools ass by a large quantity of points.  I don\'t really know.  I was too busy checking out the hot 16-year-old in front of me and her sexy, sexy jailbait thongs.  Oh, and I\'m not hip anymore.  These kids these days don\'t tear up their hats or bend the brims.  They wear white jackets and Skechers.  I was fratted out, and I thought damned sexy.  But, no, I was the dork who a group of 10-year-olds threw popcorn at.  That really happened. Saturday:  Me and The R. went camping out by Mt. Rainier (I accidently typed \"Mr. Rainier\" there, that would be whacky). B: Gee, Mr. Rainier, we just wanted to do a little camping next to you. Mr. Rainier: You kids are all right. Nature fucking rocks.  There were trees and I lost a frisbee and we drank beer and R. Girl wouldn\'t let me keep a big stick I found and poked at a fire with and there was a river and these rocks and the toilets didn\'t flush and we had pancakes for DINNER! And I really like R. Girl.  She\'s like the female me. B: You know what I really like? R. Girl: Clouds that look like boobs. B: Yes, but Nirvana .... R. Girl: ... couldn\'t have lasted creatively if Cobain hadn\'t sucked on a lead slug.  Yes, I know.  Now, shut up and pass the vodka. So, there was that.   Today at work was this. N.I.D. Girl: My throat hurts and I can\'t really talk.  I said my throat hurts and I can\'t really talk.  Clyde, my throat hurts and I can\'t really talk.  My throat, which is where I breath and magic makes the letters into sounds, is hurty like hurtsville so I can\'t talky walky mcsquawky fawky.  I want ice cream because my throat is soar and those words I just spoke hurt more to speak but these are hurting and then I saw a cat. Thank god for the snore spray, dorksluts, because todays hangover would have been packed with stabbings if I hadn\'t gotten my deep R.E.M. cycle goodness.  Two trailer park dorks go \'round the outside, round the outside, \'round the outside. Guess who\'s back ... back again ... dorks suck face ... tell a friend. Guess who\'s back .... guess who\'s back ... guess who\'s back ... guess who\'s back ... guess who\'s back ... nanana I created a monster because everybody wants to read about dorks and more, nobody cares about B ... I\'m chopped liver.  Well, if you want dorks than this is what I\'ll give you a little Nintendo mixed with hard liqour.    I\'ll never leave you behind again for so long, dork baby slut fucks. '),(3006,1785539,'06,August,2004','2004-08-06 00:00:00','   For the time being ... DORKS DON\'T ROCK IS BACK, SLUTFACES! ... that is until I get sick of doing it again and give it up for a few months. But until then, let\'s get reaquainted, dorks. So, I know you are all asking \"Where the fuck have you been?\" or \"Why the fuck are you back?\" or \"Could a ninja beat a polar bear barehanded?\".  The answers are \"Seattle\", \"Creative recharge\" and \"I\'m going with yes.\" Quick list of things that have happened to me.  -The day before Halloween (4 days before my birhtday) I was laid off from the Roy job.  I wasn\'t too upset.  I saw it coming.  Something like 4,000 people were laid off that day.   I read somewhere that everyone should be laid off at least once in their life.   It\'s supposed to build character.  You know what else builds character?  Getting stabbed.  Do you want to be stabbed?  Sometimes people who come up with advice should be advised to eat glass.   Being laid off for a few weeks sucked.  It sucked harder than a pipe whore.  I felt like a bum, I had no money, I had no reason to shave or shower and I had to put up with questions from employers like ... \"Do you know how to program the 5RTX transtickledoober and create specs. in a 3P format?\" \"I know how to grow a beard.  Wanna hire me?\" or \"Hi, Brent, nice to meet you.  My name\'s Waste O\'Yourtime. Thanks for meeting with us today.  We aren\'t going to hire you.\" So, I no longer work at the space commander job as Roy 2.  I now am working at a place pretty much across the street.  I still do things with paper and boxes and stuff.  Mainly I listen to mix C.D. I made my Junior year in college with such hits as \"Rosanna\" by Toto, \"Werewolves of London\" by the late Warren Zevon, \"Baba O\' Reily\" (or \"Teenage Wasteland\" to Creed fans) by The Who, \"In the Meantime\" by Spacehog and ... I\'m tired of listing.  Lists are for fools and wives.  Moving on, dorks! -I\'m still with R. Girl.  In fact, my sweet dork faces, we have been looking at apartments together.  I am coming out of my bachelor shell ever so slowly.  Example, I went to IKEA TWICE in one weekend a few weeks ago.  I actually said these words, \"OH, we need a chopping block (a cutting board in non-me language)\".  Can you believe it? The same guy who just a year ago skipped graduation so he could make out with Ginger Lynn ( www.gingerlynn.com ) and the same guy who has had this conversation \"Dude, don\'t talk to her. She\'s a skank.\" \"Did you just call me a skank?\" \"No, I said .. I ... drank too much?\" is going to IKEA bi-weekly and picking out chopping blocks!  It\'s actually pretty nice.  I know you are still thirsty for the tear-jerking \"I love her like the dickens\" sap right here, nosefucks, but you still ain\'t getting it.  Just know that things are great and I\'ve discovered it\'s easier to have a meaningful relationship with a girl when you actually like her as a person.  Wonders! -I have cut back on my drinking and smoking.  I can go up to 4 days without a cigarette now.  I don\'t know if I\'m quitting or I just keep losing my cigarettes and my drive to buy more.  I actually went for two days without noticing that I had lost a pack of cigarettes.  And the best part?  I wasn\'t drunk when I lost them. So, I\'ve embraced a more sober and healthier lifestyle.  I actually drink Diet Pepsi and like it.  I also avoid McDonalds.  I\'m semi-healthy B now.   Why the lifestyle change? Two reasons. -I no longer live with Kyle.  The KingDork forgot that his lease was up at the end of December NOT the end of March.  So, we had a mad scramble last month to find a place to live.   I\'m not saying Kyle drove me to drink it\'s just that when one lives in a bachelor pad he must bachele (a new verb).  So where do I live now? I\'m staying at the lovely couple\'s place until R. Girl and I move in together, which should be at the begining of April.  It\'s fun. They don\'t have cable, but they do have a shitload of DVDs and a couple of dogs to play with.  Plus they have Risk.  The better half and I have talked about starting up a game. World domination here I come! -And the other reason for my newfound health awareness.  My mom has finally admitted she has a problem with alcohol and has started going to A.A. meetings.  It was a big ugly ordeal that led up to the proverbial straw and I don\'t really feel like talking about it, ever.  But she\'s apparently better now and such and such and such and such.  Not right now, dorks, maybe later. Other random things. -It snowed a few inches here last week.  The Better Half and I went and played frisbee golf in it.  It was awesome.  Oh yeah, after over a decade of shacking up together and scoffing marriage, the L.C. finally got hitched in a Britney Spears style weekend jaunt to the place where you sign marriage papers (not that far out of the bachelor shell to know what you do to get married proper) and they did it for ... INSURANCE PURPOSES!  Now that\'s straight up fucking the system.   -R. Girl got me tickets to go see the Lions and the Seahawks play for my birthday.  She also got me a B.B. gun for Jesus\' birthday.  We spent Christmas and Thanksgiving together.  She sprained her ankle by falling down some stairs on Thanksgiving so I got to play Super Boyfriend and take care of her.   There were other things, but you know what, let\'s just start fresh.  If I remember things from the few months I was amissin\' as I start this blog up again I\'ll relate them.  Word. I tried to start another blog.  It was a cool idea and I might still do it.  It\'s at http://mahatmajones.diaryland.com  and the premise is that I take an actual historical event that happened on a given day and stream of concious my own version of history. I interviewed with Microsoft the other day.  I\'m hungry and I writing this blog makes me feel like I\'m explaining to a needy girlfriend why I didn\'t call her after a night of boozing.  I\'m going to run downtown and pickup my paycheck now.  Maybe take in a skin flick or a bagel sandwich.   Later, dorks.  I\'ve always loved you. '),(3007,1785539,'06,August,2004','2004-08-06 00:00:00','   Moving so fast, dorks, taking a minute to say hello. What to say ... I\'ve been dipping in and out of work to run errands all day.  One of the ape-women from upstairs noticed my lack of being. \"Brent, I haven\'t seen you all day.  Just making sure you were here.\" \"SWITCHBLADE!\" No, I just did my little \"I\'m here, ape-woman\" jig for her, tossed her a banana and went about my own monkeyish work.   You know what?  If I had to holler something each time I was challenged or excited I think \"switchblade\" would be a good one. \"Sir, you can\'t supersize porn.\" \"SWITCHBLADE!\" Of course, I\'d have to get a switchblade for it to be convincing.  Hell yeah, one of those little springloaded plastic ones you played with as a kid.  You know, to have all of the humor of a stabbing and none of the blood and S.T.D. exposure. Well, I have a weekend to do after this.  Not sure what it will be.  More than likely it will be a quiet evening with R. Girl tonight and perhaps this neato thing at Safeco Field tomorrow.  They\'re having some kind of fan appreciation thing and for 10 bucks you get to run around the bases, throw some pitches in the bullpen, take some batting pracitce, etcetera, more ....  Mayhaps that would be fun after some Jack.  So, I\'ve been mainly eating high protein things the past couple of days.  I kind of feel like a dork testing out this new diet that all the fatty dorks are doing these days.  Today I ate half a jar of peanuts, some pork tenderloin (good), some bean salad (beany) and an apple (because I feel sexy eating fruit).So, fuck off, you grocery snoops.   \"Raised in Car-o-li-na, I\'m not like that.\" Forget it ... it\'s hard to explain. Well, I got an e-mail from my Mom yesterday.  She seems to be doing better and that is a good thing.  However, conviction to stay with something for life is not a common trait in my family.  I just hope some exceptions can be made for alcoholics, modern love and unsweet iced tea. That\'s right, I said I feel sexy eating fruit, dorks. '),(3008,1785539,'06,August,2004','2004-08-06 00:00:00','   Things to say, dorks, and things to say. Let\'s pick up where we left off ... After my little Tuesday rant about people and signs things pretty much started looking up.  People at this place have been nicer.  I guess you scream at one old lady and threaten a manager and all of a sudden you\'re Johnny Playground around here. So, after work on Tuesday I went to this here grocery store where the high school kid who was bagging my groceries was named Brent. \"Hey, my name is Brent too.\" \"Awesome.\" \"And, you know, I was a bag boy in high school too.\" \"Hey, maybe you\'re me from the future!\" \"If I am I don\'t have any advice for you.\" \"Aww, dammit.\" Which got me to thinking ... What the fuck would I say to the Brent of 8 or so years ago?  Probably to dump my fattie high school girlfriend and bang Emily Urbanek more.  I think I\'d have advised to stay closer to my family during college too.   But, I had no time for completing such thoughts because this happened next. An old man was standing in front of a mini-van yelling at two younger black girls. \"Are you stupid?  This is a one way drive!\" Then he goes up to their window and keeps yelling. \"Can\'t you read?  Do you have a problem with following rules?\" The girls drove off and the old smug anal plug walked off content.  But, I hadn\'t had my time with him yet. \"People make mistakes, sir,\" I said in my best Batman voice. \"When there are mistakes people get run over,\" he said, obviously flustered that someone had invaded his white-retired-cancer-bucket  post-berating orgasm. \"You didn\'t get run over sir,\" I said very Bruce Willisy.  Oh, and I was saying \"Sir\" all cool and calm to get under his skin Bond style. \"Mumble, walk, Mumble, Rumph, Hararumph,\" said Asshole. \"Asshole,\" said Batman. So, that kind of upset me.  To be honest, I had a southern flashback and wondered if the guy kept a Dixie flag tucked under his fashionable R.E.I. gear.  I then wondered if maybe the girls had been white and fat I would have laughed at the ordeal.  But then again, if the girls had been white would Asshole had yelled?  Dammit, I hate racism.   For those of you who weren\'t part of the Summer of B. listen to this story. In a post debauchle hangover stupor I decided that I wanted to play kickball so I went and bought one.  I was upset from the night before because this guy had called me a racist.  He really did look like Jackie Chan, I was drunk, he was trying to get me to hit him and then I said this ... \"Get out of here, Jackie Chan, we aren\'t going to fight.\" \"Oh, what are you racist now?  Going to use that against me?\" \"No, ass, you look like Jackie Chan and you want to fight.  Don\'t even pull that on me.\" So, I was upset that I was called a racist when I knew I wasn\'t.  The dude really was hopped up like Jackie Chan.  It would be like if I saw a white man killing a bunch of Russians I\'d go \"Hey, Stalin, calm down\". So, I had to redeem myself.  I wrote \"End Racism Now\" on the kick ball.  The idea was that if I could change one persons mind than I would be redeemed for my potential wrong racial doings. My idea was this, we\'d be playing kick ball and some dude with unjust racial convictions would be playing outfield.  I would kick the ball deep and it would land in his hands and he would read \"END RACISM NOW\" and he would think, \"Yes, I must change my ways\". We never played kickball, but the thing sat in my room for almost two years and I hope I helped one person. Burritos, ice cream and a movie with the female part of the Lovely Couple.  Apparently she found me inside Hollywood Video after we were split up and I was just staring at a wall.  When asked what I was doing I couldn\'t remember why I was in the store.  \"Huh, what the hell was I in this store for again?\" What\'s the big deal about sobriety? Kyle is now dating a friend of R. Girl\'s.  This should lead to some fun and wackiness.  They claim to not be dating, but he has already bailed on a trip planned for Portland this weekend so they can go to a museum and study the Amish or whatever it is that couples do. Shit, I\'m at work right now and listening to a C.D. that I made in college (a collection of bands playing live on Letterman and highlight from Badly Drawn Boys first album).  It has a bunch of orgasm sounds that I need for promos for the Phi Kappa Nasty show.  I forgot they were at the end and they just started playing in my office (glorified room).  Oops. I picked up tickets for me and the R.G. for The Shins show at the end of the month.  It will be swell.   Thanks be to Eva for introducing me to them.  And, while I\'m at it, thanks to God for my own pair of shins.  Much props. Last night was dinner with the beautiful to celebrate her new job offer.  Pizza and making out is a quality night.  She had to leave before me this morning and I really don\'t remember saying goodbye just being confused that people were moving.  She left me a love letter in my leftover pizza box.  Sweet, dorks. Well, I need to go work or shower. Later, dork dorks. '),(3009,1785539,'06,August,2004','2004-08-06 00:00:00','   Morning, dorksluts.  I have some busi to take to the streets.  To tide you over here is one of the only two entries I put into my other blog.  Kyle gave me an idea to just do my \"stream of history\" blog on this website every once in awhile, which is a good idea because, to be honest with you, sometimes I taint got much to say. Well, here it is.  I\'ll write more later.  ((2004-01-09 - 10:23 a.m.  Here\'s what happened today ...  IN 1984!   Long time beauty queen Clara Peller first appeared in Wendy\'s infamous \"Where\'s The Beef?\" commercials.   A transcript of that commercial is posted below.   An old woman: (pointing at a hamburger) Where\'s the Beef?   America: (laughs for 15 years)   The commercial prompted other companies to follow suit. However, not all slogans caught on as well.  For example Duracell\'s \"Where\'s the battery acid?\" campaign was quickly cut short when children, most notably child actor Punky Brewster (Sunbelt Farmface), went looking for the battery acid inside Double-A and Triple-A batteries.   Farmface was forced to leave her multi-billionaire Punky Brewster media empire in 1985 due to severe acid burns on her face and throat. She became a posterchild for the \"Hey Kids Battery Acid Ain\'t Kool-Aid\" organization, which spawned it\'s own catchy slogan, \"Acid Bad, Chocolate Good\".   Peller went on to enjoy critical success with her tell all novel \"The Beef Was Never There\". The book exposed the sordid history of the Wendy\'s corporation. Here\'s is an exerpt from the chapter title \"Frosty, French Fry and Freedom\"   It was the 4th day that I had been locked in the basement. I knew Wendy would be back soon and with her would come the sauldering iron and ball peen hammer. I had to escape, I had to get to Dave Thomas before that red-headed #$@! If the world was ever going to find out the truth about Frosties I had to make my move now.  From my time in the Marine\'s I was trained to kill with my bare hands, but those bare hands were handcuffed to a radiator. In a dark corner of the basement I saw a harpoon gun left over from the shark hunt Wendy and I had gone on last summer. Those few days in Florida were the good times, but they were gone. All that was left was pain, a red-headed @#$#, a harpoon gun, 4 stuffed sharks and a deadly secret ....   Peller did eventually escape that basement of horrors and her and Wendy \"Harpoon Face\" Thomas went on to rekindle their friendship and hit the road as a country-western duo with such hits as \"The Judds are Fatties\" and \"The Judds Suck\".  Peller died in a freak shark attack in West Texas in 2001. Witnesses say it was \"the most awesome, wicked\" thing they had ever seen.   At Peller\'s funeral, Ronald McDonald gave the eulogy.  \"When Clara gets to Heaven we all know she\'s going to ask God where the beef was all of this time. And I think we all agree that it was probably in her big heart. Super Size Amen.\"  Super Size Amen, indeed.)) '),(3010,1785539,'06,August,2004','2004-08-06 00:00:00','   \"Brent, what kind of people do you think are stupid?\" The kind in this story ... I\'m at a new job right?  So, it goes without thinking that only the people who had this job before me and God know what to do.  Well, maybe instead of getting upset at the new guy for not doing something for you when he had no idea he was supposed to do it  because your fat childless gray haired ass was too busy daydreaming about \"American Idol\" and how to look good by Easter on the Atkins diet to talk to me about it you should pull that grande latte out of your grill hole and talk to him.  But, nope, when you get asked by your manager why it wasn\'t done you blame it on the new guy so he gets talked to about it.  And then you have the nerve to call me and act as if I\'m holding up the progress of the free world because due to the inability of UPS to make things appear in people\'s hands you have to wait an extra day when really you\'d be O.K. and could keep eating your cake for Peggy\'s birthday and planning your weekend trip to Ann Taylor if you had been a fucking adult!   That person is the person I hate.   But, what can you do?  Some people have to be the type to step in front of buses.   Here is an actual handmade sign I saw at work today.  I guess everyone was invited to add their own two cents.  Someone laid down a nickel of stupidity.  What does the customer expect of us? 1.)To Listen 2.)Call Back 3.)Paitience 4.)Large adjustments 5.)To agree with them 6.) To be KNOWLEGDEBLE 7.) Even larger adjustments  KNOWLEGDEBLE?  Please, God, let me use my degree at some point in my life where the employees aren\'t so knowlegdeble.  Oh, and by the way, this sign has nothing to do with my job.  It\'s just something I saw.  I have no need for paitience or \"even larger adjustments\" to vague things in need of adjustment. Man, it\'s people like that that get under my skin. \"What will make us better?\" \"Large adjustments\" \"To what, Cindy?\" \"Even larger adjustments?\"  Back to my hip life.  Kyle has quit Microsoft.  Well, he was going to have to take his 100 day break soon enough and he was offered another job so what he hell.  He wants for me and he to gee on a road tree this wee to Portlee. I haven\'t spent any time in Portland and think I might take him up on it. I\'m now in search of a snowboard online.  Must find one soon so R. Girl and I can thrash together.  Also, started thinking about Valentine\'s Day ideas today.  I\'m leaning towards a pony or what I told her I got her last night. \"I hollowed out Mt. Rainier and it\'s now your house.\" How wicked would that be? Other things I have to do to maintain a certain level of chic. 1.) Rotate my C.D. selection for the truck. 2.) Get a new pair of crosstrainers and a sweater. I\'ve got a headache and no hangover.  I think I need to get out of this place for lunch.  I will, dorks. '),(3011,1785539,'06,August,2004','2004-08-06 00:00:00','   All right, back on track.  Monday morning, chugging a liter sized Diet Pepsi and digesting a PowerBar.  Give me a moment, dorks, just trying to remember what I did this weekend. Oh yes ... Friday R. Girl and I had a very romantic dinner at the International House of Pancakes.  You see, I feel it\'s important to treat the girl you care about with a fancy dinner at a foreign restaurant, preferably  at one boasting the all encompasing \"international\" label.  I didn\'t know they had never-ending pancakes there and now I have a new mission in life.  The man near us ate 6 pancakes and his son was in such awe.  Being the ass dork I am I quickly boasted I could eat 8 pancakes. \"You can eat EIGHT pancakes?!?!\" said an astonished R. Girl. \"Dad, that man can eat more pancakes than you.  I feel robbed of my innocence and my one true hero.  Goodbye, father, I care not for your hypocrisy of breakfast baked goods nay no longer,\" said the man\'s ex-son. But, I went with the patty melt instead.  Mainly because our waiter told us that someone ate 46 pancakes once and my hoped died and so did the boy. Saturday me and the R. went to check out a \"Mega snowboad and ski super sale blowout\" that boasted of sales that would make GAP girls cream.  Suck my dick if they were lying.  They were, so blowjobs around.  Oh well, I am excited about taking up snowboarding and the chance of messing up my evermore fucked up knees ... Speaking of which ... that is another thing that is new in my life.  Bad knees.  They fucking feel like someone poured sand in them sometimes.  I\'ve actually thought to myself \"Wait a minute, I need to rest my knees before I go any further\".  Shut the fuck up, body.  I finally listen to your hacks and hangovers and make amends to change and this is your fucking thanks?  Bad knees?  Whatever, I\'m back off the wagon again, ass. And back to the reality of non-interpersonal body arguments. \"Shut up,leg\" \"Bite me, molars\" Train of thought slipping ... was talking about something ... maybe it was ... Oh yeah, Saturday.  So, we skipped snow shopping and checked out some new sights (for me anyways) in Seattle.  We watched kids play in a fountain and I ate too much popcorn and got a stomach ache.  I won R. Girl a stuffed dinosaur playing skeeball and she won me an army man with the game where you put the quarters in and then more quarters fall but it was tokens not quarters.   Sunday was lazy day.   Talk about .... Emotions.  Maybe, sure.  There was some tension between R. Girl and me of late.  We talked last night and got to the root of the problem and dealt with it.  I definitely feel better about the whole situation (which I\'m not revealing what it is because it is embarassing), but not about the fact that there was something between us.  We still haven\'t had an argument or really been mad at each other and I like that lots.  Oh well, it\'s growth and growth is good and speaking of growth ... Remember, the girls in the middle are always the first to fall off. Shins segue right there, hipper dorks ... Oh yeah, e-mails I\'ve received today.  Three great ones. Eva finally wrote back.  Yippee.   Someone at the \"big paper\" (as my dad would call it) has written back to me about working as a sports stringer.  Must write back and follow possibility further.  Monster like write, write good sports stringer.  Sonics win game, make monster happy, freethrows many, rebounds help.   And I got one from a guy who runs an adult-industry \"the onion\"-based website where I could make a fatty G a month to write parody like articles about porn.  I feel weird about this, and don\'t see how I can really write too much parody about porn. \"In a stunning development today, people got naked and had orgasms on each other.\" I\'d feel like I was selling out my forward thinking sexual hipness for the mass puritan culture.  But then again, a G a month to write funny porn is good.  10 200-word articles a week though?  Man, I don\'t know.  Maybe. Oh, and K-Dogg the Killa called me on Friday.  My old high school and sometimes college drinking buddy might be moving out here in August.  Oh joy abounds and hearts leap like leopards, glory days of debauchary in Seattle! Later, dork princesses.  I might or might not be wearing women\'s underwear on a dare from R. Girl. '),(3012,1785539,'06,August,2004','2004-08-06 00:00:00','   Fucking Friday, bro. Dorks, I am full of spirits.  I\'m talking \'bout the angel tickles kind, not the Jack Daniels.  This new diet is kicking ass.  I feel my compounded-frat-boy-beer/LSD gut dropping off.  Although, last night the lovely couple and I went out for hamburgers and french fries.  I\'m allowed one naughty food day a week.  I cut it down to just one bad meal for the first week.  Now I\'m back on the protein kick. Shit, that last sentence reminded me it was time to eat and I just dropped a peanut into my keyboard it\'s stuck under the 11111 key.  Fuck it, it\'s not really MY keayboard. Yesterday I had coffee with one of the sports guys from the Seattle P.I. (the only newspaper with a mustache and a Ferrari 308).  It was a nice coffee as such ... We discussed pro sports versus college sports, LeBron James, journalism and the rigors of being a sports stringer.  It being the middle of high school basketball season and since I have a fulltime job that doesn\'t let me (I just got the peanut out!) get out until 4:30 I won\'t be able to start working with them for about another month.  Then it looks like I\'ll be covering baseball games or track.  Either would be fun. After that I headed back to work and then out for burgers, which was followed by a 2 mile walk with the L.C. and their dogs.  It was fun and wet, as are most good things in life. Tonight with the R.G. I have a well-earned steak dinner, which will probably be followed by \"Family Guy\" episodes and a box of wine.  I\'m so European.  Smirksome. Tomorrow is the big mini golf rematch between me and my lovetickles.  A few months ago I, trying to be a nice guy, let her win.  Only I wasn\'t anywhere smooth about it.  On the final hole instead of easily missing and making it look real I reared back and hit it as hard as I could. \"Oops.\" I forgot I wasn\'t dating that kid from \"Life Goes On\".  I\'m talking about the retarded one, not the girl who wore glasses. So, there\'s that tomorrow (lay money on me) and then the cruise and Shins concert on Sunday.  My, The Shins on Sunday. I wish the time would hurry up.   Speaking of which, does that new Ashton Kutcher movie look like it was made by 14-year-olds?  Not tempting at all.  It has that big fat kid in it who couldn\'t see the sailboat in \"Mallrats\" (Kevin Smith isn\'t original and has hepatitis).  I can\'t stand that guy.  Every movie he\'s in he plays this big tough guy when in real life he\'d be too fat to hit me.  I\'m more scared of the kid who played Goldberg in \"The Mighty Ducks\" than this clown.  Man, if I saw that guy I\'d totally switchblade him and his precious Goth hair and \"Der, I need a sandwich \'cause I\'m a meany\" attitude.  Dammit, I hate fatties and Goths.  Football.  Monster trucks.  Midwest culture.  Closed mind.   Whew, where\'d I go there for a second?  Back to ultra-hip land ...  Yeah, so the guy\'s kind of a joke if you ask me.  Hip recovery. On that note, have a great weekend, Dork Johnsons.  Let\'s all of us get laid. '),(3013,1785539,'06,August,2004','2004-08-06 00:00:00','   Well, shit, dorks, we\'ve been caught. The lovely and talented Eva Moore has announced on her own blog (http://themeasure.diaryland.com) the fucktastic return of Dorks Don\'t Rock. So, to all of you white t-shirt alternative vegetalbe people who might be reading this for the first time ... Hi and don\'t take my rants (and not boring Dennis \"Finger me, I\'m special now\" Miller rants) too seriously.  For I, born of Iowonian man and woman, am but a humble consumer of cheap liqour and lover of all things Zeppelin. And for you people who have been keeping up with the return of the blog for the past couple of weeks I think it\'s time you got an answer to the question of where the hell I was for 3 months. Well, you see, after the layoff I went into what psychiatrists might call a multi-month bender.  You see, since it was near the holidays and near the end of the year the only people who were in the market for hiring were cocaine dealers.  I didn\'t become a cocaine dealer, obviously, so I had to go with what work I could.  This was mainly sitting in front of a computer looking at screens to find problems. What problems? No one could really tell me what problems I was supposed to look for.  Yes, it\'s true.  So, for a couple of months I sat in front of a computer screen and scanned through several thousand pages of documents for 8 hours a day.  I was fried and poor at the end of the day.  But I had to eat and so I had to work.  What I didn\'t HAVE to do was write in my blog.  So, I didn\'t. Also, the days around here became supremely short.  I lived in darkness like a beast cast from job security into the pits of temping.   But now things are definitely turning around and I am on a life-improvement kick. This is why I\'m doing the diet, making amends with my alcoholic mother, focusing on maintaining the love between me and the redheaded R. sexpot, looking for a better job (although the non-brain-frying one is good for now and puts little rush on me to find a better one) and paying off debts occured by Christmas and such. Speaking of which ... So, I called this trendy clothing company today that I have an account with.  I was wanting to pay off my card and such.  So, perplexed by my balance, I asked when they received my last payment. \"November 8th\" \"What the hell?!? November?  That\'s not right!\" \"Sir, the information I have in front of me says you haven\'t paid since November.\" \"Well, dammit, the information is wrong ..\" \"Sir, what it says is ...\" \"I don\'t give a damn what it says ...\" \"that fines have been ... \"Stop talking.\" \"...added and that to remove ..\" \"Listen, I want you to stop talking right now.  I am done listening to you.  Put on your supervisor.\" ( I really said this, by the way, and he shut up). (\'only the lonely\' on muzak) \"hmmly la lmmmly ... mm, mm, mm ,ummm-m-mm-mm\" \"Hello, Mr. Kinkade, this is another annoying peson.  You wanted late fees removed?\" \"No, I was concerned because that guy said I hadn\'t paid on my card since November.\" \"It shows us here that you made a payment on January 8th.\" \"Well, you need to teach that guy the difference between November and January\" Well, I won\'t bore you with the rest of the story but they were jerks.  I like beef jerky not jerks, beefy. Yesterday I got in touch with someone from the \"big paper\" and set up a meeting for tomorrow.  We gonna talks about hows I\'d be a good little writer for their big ol\' paper.  I\'ll update you on that later. My The Shins tickets came. My, The Shins tickets came. My, The Shins tickets CAME? Moving on. Another paper has responded to a little 40 word start to a porno story that I sent in.  It\'s not the porno thing I mentioned before, it\'s another porno thing.  The world is full of porno things. The deal is this, I write a 40 word opening to a story and if they like it I finish up the story by recording it on a server.  Then they run the 40 word snipet and if people are titilatted they call up and listen to the end.  If I get the most people to listen I win something, probably more porn.  Well, I have a little problem.  First, read my intro.  After class I stopped at the cafeteria for a can of pop. I couldn\'t carry it because my arms were full of books. \"Got any lube?  I know a place to put that,\" I heard my sorority sister Heather say. Awesome, eh?  Lesbian soda can sex!  Oh my god, I\'m so wickedly pornirksome.  Except ... Who the fuck is going to listen to the story when they call up and hear a guy with a deep voice saying this ... \"So, we went to my room and got wickedly naked.  I mean all the way with tits and all of it.  We were such hot chicks.  Then I got that lube that was mentioned in the ad thing and Heather was like \'Damn, I want to do nasty shit with this soda can\'.\" I really want to win this contest and I think I have a good start.  I just need to convince some chick to pretend she\'s getting Sprited (dibs on this newfound euphanism).  Maybe it\'s time to use the old \"Get the girlfriend drunk and suggest the unsuggestable\" technique. Speaking of which (I am full of seques today) I spent the night with my precious yesterday.  We watched the \"Real World: San Diego\".  Two of them got arrested for really nothing at all.  We have a theory that people are just sick of MTv and the likes and are looking for any reason to lock up people who associate with them. That would be awesome.  So awesome that I do a chair-jig thinking about it. Well, I think that\'s enough for this smirksome Wednesday.  All of you S.C. Democrats, all 17 of you, vote for Dean.  He\'s the non-Botox-filled-non-Bush-loving candidate.  AND HE\'S A LITTLE NUTS! AAAAAARRRRRRRGGGGGGGGGHHHHHHHHHHHH, DORKS! '),(3014,1785539,'06,August,2004','2004-08-06 00:00:00','   Afternoon, dorks.  Let\'s get started. I\'m feeling very efficient today.  This stems from the high protein diet (Today: pork loin, diet pepsi, peanuts and little sausages ... the ape women brought in the sausages as an apeasment to their God of Beuracracy and I took part, bare chest marked with the stain of a sweedish meatball handprint and chanting \"Order toner, envelopes, overtime req-ui-si-tion\") and a very restful 9 hours of sleep that I got last night.   Things learned last night at the Lovely Couple\'s Den of Lack-of-T.V. 1.) An online Rorschach test told me that I was fueled by \"imagination\".  Imagine that.  Get it?  Get it?  AHHH-Hahahahaha! Fuck me, my mind and center of my being is smirksome!   The very sweet and kind member of the Lovely Couple\'s results said that she is driven by \"Peace\". Now isn\'t she a \"piece\" of work.  Holy shit! Where do I come up with this stuff? 2.) My I.Q. is 138.  I don\'t know if that is good or if like bank balances we\'re supposed to keep those a secret.  I made a 1280 on the S.A.T. also.  There, I\'ve said it all. I\'ve broken the secret bond of man by divulging my scores to the public.  My soul is secret no more; it is baren, 138 and 1280.  Gaze upon the naked essence of man.   Back to the I.Q. thing, the internet said that it means I can have my rule over 98% of society.  I don\'t know if I should get too excited though.  The internet also promised me free porn. I don\'t think I\'d want to rule 98% of society.  Maybe like 7 or 8 percent and have a little island to myself. 3.)The book \"Franny and Zooey\" by J.D. Salinger is not about lesbians.  I knew this, but I think I put off reading it (and thus completing the trifecta that is Salinger) because I thought it\'d be like reading Cosmo or something. \"Silly Franny, mascara should match your belt and not your shoes.\" \"Oh, Zooey, you are so 1941\" Welcome to dorksdontrock, folks, the home of \"Sexist recreations of America\'s great authors\".  Next week, Edgar Allan Poe. \"And quoth the stripper, \'Let\'s do blow\'.\" Actually, I put off reading it because one of my psychotic ex-girlfriends (one of the two that were certifiable, the others just smelt bad) read it in college and went on for two weeks about how good it was.  She also went on for two weeks about how she loved wearing clothes with cherries on them.  So, of course, the book was, by default, pre-judged as not up to par. I was wrong.  It is good.  Om.  Other tidbits. Kyle got into a car accident yesterday.  He cut off a big truck in his little Miata and got spun around.  Apparently he smacked his head on his roof and got a tail light knocked out. \"Yeah, dude, I saw that coming.  You are a pretty reckless driver.\" \"I think the cop gave me a ticket because of my blue hair.\" Naw, Kyle, it was probably because you were swinging a ton of steel around I-5 like it was a misappropriated yo-yo. Wait, I should probably call him and make sure that he didn\'t have a concussion. Well, he wasn\'t there.  They call when people go in comas, I\'ll find out eventually. R. Girl, her friend, Kyle and I all have a little cruise about the Puget Sound planned for this Sunday.  After that R. Baby and I go to see The Shins.  How ultimately hip am I? Well, I need to return to my work and my life.  It\'s been nice spending this time with you, dorks.  I hope all on the East Coast are warm and not buried.  By snow, not alive, that is, in reference to the buried thing. '),(3015,1785539,'06,August,2004','2004-08-06 00:00:00','   Sleepy, groggy, protein-packed and half awake.  And then this comes on ... \"Face down in the gut-ter, won\'t ad-mit de-feat though his clothes are soiled and bah-lack ...\" Damn me like the devil, dorks, them Dropkick Murphys (quoted song being \"Barroom Hero\") can pump up even the greyest Seattle blues.  Motivation to write about rain here. Before we go any further, dorkles, let me just get something off of my chest.  I don\'t give a fuck that Dennis Miller is a conservative now.  However, this is apparently a scoop to every news editor in the fucking country.  Everytime I open the paper or go to cnn.com or the other website I\'m addicted to even though the fucker that runs it is a puppet for the ragime there is fucking goateed Dennis with his irksome smirk.  He\'s fucking smirksome. \"9/11 changed me,\" is his usual quote.  \"I couldn\'t be a liberal any more, the terrorists made me think harder about this than I had to think about whether they should have pulled a Fredo fishing trip on Coppola\'s daughter after \'Godfather III\'.\" I just made that last quote up. This whole thing bothers me for two reasons. 1.) Who gives a fuck?  Mike Myers isn\'t funny any more, why should you stay the same. 2.) I am bothered by people who use 9/11 as a soap box for their moral staunchness.  (Typical asshole: \"When I saw the planes hit the buildings I asked God how things could be as such and he said to me \'You are so special, I wish I was there to tickle your belly\'.  So I ask you to tickle my belly as the Lord would do.\") Now if Dennis Miller had been hit by two fucking jet liners then maybe I\'d care more about his personal account of the situation. So, with that said, let\'s swap weekend stories like a couple little pleated honey bun-dos at the tea station, eh? Friday night was quiet as predicted.   Saturday was movies (porn and horror). Yesterday the sexy R. and I went on a little Seattle loop.  For those of you who aren\'t familiar with Seattle, look at a map.  We started near Kent, drove up the western shore of Lake Washington to Mercer Island and then took 90 into Bellevue and back down to Kent.  Why?  Because we hadn\'t done that before, nosefuck. We went to our dock and threw bread at the ducks.  However, the seagulls were more determined to get the bread and the ducks didn\'t seem to care whether they got any or not.  Maybe it was some twisted bird S&M, master/slave, gull/duck situation.  Maybe the seagulls were just faster. Do you know why one side of a \"V\" of geese is longer than the other? Because one side has more geese on it. R. Girl got me with that one.   She was virtually attacked by seagulls! Virtually (adv.) -about, just about, almost, most, all but, nearly, near, nigh, virtually, well-nigh -- ((of actions or states) slightly short of or not quite accomplished \"Virtually everything on sale!\" = \"Not everything on sale!\" \"Virtually lower prices store wide\" = \"Not lower prices store wide\" Just a little lesson for you dorks out there in the \"virtually\"-obsessed advertising-permeated world. Moving on. \"Cabin Fever\" is so wickedly awesome!  It\'s Dorks Don\'t Rock meets Evil Dead 2. Protein diet still going.  Little sign of progress so far.   I think that\'s about it, dorks. '),(3016,1785539,'06,August,2004','2004-08-06 00:00:00','   Anyways, dorkies, I had written a pretty negative blog earlier with copy/pasted e-mails and charts and graphs and testimonials describing just how rough yesterday was (too much bourbon on Tuesday night mixed with little sleep and ape women marching).  And then I came back to Hipville.   Hey, I\'m all right, just lost my esteemed level of cool there for a minute.  Now I\'m back and ready to rock you dorks who can\'t. Right now I\'m listening to a mixed C.D. I made called \"When You Need to Rock\". (sidenote: I made a little series with these, including \"When You Need to Rock ... Softer\"  and \"When You Need to Rock Southern\" both of which are amazingly wicked c.d.s that I\'m proud of.  The combination of the two genres mentioned above is my \"Go \'Way ... A Honky C.D.\" a fine milestone in my marvelous mixed C.D. legacy).  Right now I\'m hearing \"You Don\'t Always Hmm la la\" by the Stones (the C.D. is mostly Stones and Zep stuff).  This used to be my favorite Stones song when I was in Jr. High and early high school.  Then it was \"Paint it Black\" in late high school/very early college and became \"Honky Tonk Woman\" soon after that. Now?  I think it might be \"Jumping Jack Flash\". This morning was beautiful.  I woke up for a little morning smooching with the R. Baby (I spent the night with her last night so I could whine and get a back rub) and an unexpected shot of vodka.  Apparently lushly R. had some screw drivers the other night and apparently they were so good she forgot she had them.  I was feeling a little under the weather and asked if I could have some orange juice, which unknown to me was full of vodka.  So I got me a little 5 a.m. Stoli kick, some more kisses and it was back to sleep.  Great way to start a day ... well, and then quickly wind it down again. I get more done before 5 a.m. than  most whinos do all day. Tonight is cleaning night.  Yeeha, my life is exciting the past couple of days!   I need to get back to work, dorkles.  Just letting you know that I hadn\'t forgot about you and that you are all so precious.  Suck it. '),(3017,1785539,'06,August,2004','2004-08-06 00:00:00','   Afternoon, dorkies.  Or maybe nighty-night to you French dorks, you Forks. Today was a very boring day.  I did receive this e-mail string, I\'m sure you all realize that you\'ll read the last message first and then the first one last. So, to understand my frustration, read from the bottom of the list up (this is the actual e-mail corespondence, with names changed).  ((From: Me To: A girl I understand.  What I need is the address for the portal homepage.  The homepage on this comp. is set to something different.    ---------- From: 	A Girl Sent: 	Tuesday, February 3, 2004 1:57 PM To: 	Kinkade, Brent Subject: 	RE:   I gave you the path in the previous email. Go to the portal homepage, click on the Employee Services tab, then click on Procurement and Payment, and you will see a link to Business Cards and Stationery.   -----Original Message----- From: 	Kinkade, Brent   Sent:	Tuesday, February 03, 2004 2:56 PM To:	A Girl Subject:	RE:   What is the address for this website?  ---------- From: 	A Girl Sent: 	Tuesday, February 3, 2004 1:52 PM To: 	Kinkade, Brent Subject: 	RE:   The Business Cards and Stationery website.   -----Original Message----- From: 	Kinkade, Brent   Sent:	Tuesday, February 03, 2004 2:43 PM To:	A Girl Subject:	RE:   (A Girl), I\'m sorry, I\'m still learning ... what site are you talking about?  ---------- From: 	A Girl Sent: 	Tuesday, February 3, 2004 12:59 PM To: 	Kinkade, Brent Subject: 	RE:   Please go to the portal Employee Service tab, and click on Business Cards and Stationery. You can order letterhead through this site.  -(A Girl)   -----Original Message----- From: 	Kinkade, Brent   Sent:	Tuesday, February 03, 2004 1:23 PM To:	A Girl Subject:	  (A Girl), I need to order some letterhead for the Bellevue location.  I was told to go through you. If this is correct, could you please tell me how much we usually order?   Thanks, Brent Kinkade  (my work number) ))  It\'s like I\'m ordering letterhead from a Muppet. Well, that was the dorksdontrock highlight for my day so far.  I\'ve rented \"Lost in Translation\" and \"American Splendor\" and plan to start watching them within the next 10 minutes. \"But Brent, you are still at work.\" Not any more, my dork beauties ... I\'m outta here! '),(3018,1785539,'06,August,2004','2004-08-06 00:00:00','   Festive how-dos on this the holiest of days.  The day from whence the hog of ground bursts forth upon the land to beakon the call of but 3 fornights hence of winter\'s fowl demoness.  And, yea, that tradition, born of the noble clansmen and nazi-noos of Germania, has rendered upon us the icy grip once more.  Take heed ye of weak condition and cast on your blankets of wool! Fucking groundhogs. Very busy Monday for me here, dorkhogs.  People here must spend all weekend thinking of things for me to do on Mondays.  Naw, the probably don\'t.  They probably throw beurucratic orgies. \"I\'m sorry, to receive fellatio you must have been with the orgy at least 2 hours and have the consent of a sadist.\" But there was no swapping of the pinstrip-ed flesh for me!  There was ... a box of wine as predicted on Friday which led to ... my first (and 4th) cigarette in over a week and then ... a fucking massive hangover when I went with R. Girl to here place of work which ... is a place for kids who are sick and so ... I played with a paralyzed kid who is so much fun, what\'d we play? ... \"where\'d the paralyzed kid go\"? where? under the blanket.  and then ... there was a car-sick ride due to the wearing off of the wine and the settling in of the big part of the hangover, that ... made me so nauseated that I almost puked in R. Girl\'s carmobile which meant ... there was no mini-golf rematch (blamed on rain, Milli-style) so we ... went to bed early and tried to get healthy for ... the Puget Sound cruise we went on. This thing was awesome.  Kyle and his pseudo-squeeze went.  I was and still remain angry at Kyle for brotherly and otherly reasons.  This put a damper on the trip, but didn\'t ruin it completely.   We got to go through the Ballard locks and see this wicked awesome hill where multi-million dollar homes just fell off during a wicked winter a few years back. \"If the groundhog sees his shadow into the Sound with your house. (da-da-dum!)\" And then what did I do?  Hmmm, let\'s see, I returned my winter jacket b/c I was sick of it\'s lack of hipness and protection from the elements.  In exchange I got ... WAIT! I went and saw The Shins!  And they rocked so hard that I cried.  No, I didn\'t cry you tabloidorks.  I nodded my head and shook it when they said \"... at the morning bus stop, shaking their heads\".   We went to the all ages show so there were a lot of hip little 14-year-olds in CBGB t-shirts (\"It totally stands for \'Come bang girls, buddy\'\").  Besides the lack of hipness in the crowd there was a great fucking show on the stage.  I heard pretty much all of \"Oh, Inverted World\" played live!  Next stop, Badly Drawn Boy(\"I totally heard this guy doesn\'t have hands and so he can\'t draw, buddy\"), Radiohead and James Brown. Other things to talk about? Not really.  I\'m pretty tired from my run around weekend.  I haven\'t had to much time ... wait ... someone from Microsoft contacted me today about a position with them.  Should I do it?  Also, been trying to get in touch with the station manager of one of the big stations in town to work as a part-time d.j. (like I did in college all of you Phi Kappa Nasty fans). Will keep you updated on that. I\'m going to go eat protein and sleep.  You are all so precious to me, dorks. '),(3019,1785539,'06,August,2004','2004-08-06 00:00:00','   What to say, dorks?  It\'s muther-smirksome Friday, boyyyy! I\'ve got my truck fueled up, all of my Valentine\'s shopping done, some cash in my pocket, a horrifying addiction to crystal meth and two days with no ape women.  B\'s about to get freaky up in his Valentishizzy!   I woke up very late this morning after a night of weird dreams that included one that I met LeBron James on the street but I didn\'t know it was LeBron James.  I asked him for some reason if he knew how to get Sonics tickets and he hooked me up with season tickets.  I think I woke up at some point during this dream to some questioning by R. Sleepy.  I vaguely remember trying to explain that I had just got basketball tickets. I sometimes have dreams like this where I\'ll wake up and be expecting some cool new thing to be there, but it isn\'t. \"Cleydsdale with a rocket launcher attached to each leg, where are you?\" So, I woke up late and got to work about half an hour late.  No one noticed, nor have they noticed that I\'ve been running errands throughout the day. But beyond my growing psychosis (I did the stumble and mumble thing in my sleep again last night) there are great things a-brewin\' ... because it\'s Valentine\'s Day weekend.  And that means I get a present! And R. Girl and I are going out tomorrow for dinner at the place we went to on our first date and then a movie.  Sunday will be the zoo (if weather permits) with the Lovely Couple.  Will let you know how all of that turns out. They just announced on the intercom at work that \"The Mountain Man is in the lobby!\" The Mountain Man is this guy who goes to the offices around the area selling nuts and candy every couple of Fridays.  The people around here don\'t get too excited to see him, but when I had the Roy job the chicks over there would go crazy! \"Oh, I just love your chocolate cock.  I need it so bad.\" \"Give me some of that sugar covered crack-heroin.\" \"The Mountain Man is my God all other Gods are beseeched in the glory of Mountain Man and his delectable gummy worms.\" Naw, it\'s not like that, they just like candy and nuts.  Hehe, nuts. Anyways, whenever they announce that \"The Mountain Man is in the lobby!\"  I always get this image of some sort of deranged hillbilly who has escaped from somewhere and always finds his way back to this office.  I always picture him barefoot and in overalls with a straw hat just kind of frantically running around going \"Ah, hoo-eee, it\'s the lobby for me!\"  Then they have to announce that he\'s back so the gaurds will come and get him.   Of course, there are always banjos playing in the background as they chase him.   The next time you are bored in public play one of \"Brent\'s Imagination Games\".  Such as: -Next time you\'re at Wal-Mart or the D.M.V or wherever your happen to be picture in your head what would happen if a hillbilly ran in a hootin\' and a hollerin\' \'bout \"Heading for the hills\" or \"T\'ain\'t going back to no jail\".  What the hell would you do?  That one\'s always good for a chuckle. -Or.  Imagine the person you are talking to (a cashier or a cashier) suddenly had a spear go into their head.  NOt a killing spear, just a suprising spear.  How funny would it be to see someone with a goddamned spear in their head?  As long as they were still healthy and just happened to have a spear in their head. Dammit, I love the \"spear head to pass the time\" game. -OR, you could play the ever-Brent-popular \"The last few minutes of Return of the Jedi\" game.  In this game you are Han Solo and have just shot up the reactor core of the Death Star.  Uh-Oh, Han!  You\'d better get out of there fast!  This is best played at crowded malls when you are trying to get from Cinnabon to that smelly Yankee Candle store. Maybe I\'m still sleepy.   Well, have a good weekend, dorks.  Happy Dorkentines Day to your and yours (your nuts, hehe). '),(3020,1785539,'06,August,2004','2004-08-06 00:00:00','   \"Why didn\'t you \'put on socks\'?\" \"Because my job is to \'wear tennis shoes\'.\" Dammit, there are so many pricks around this office that I feel like I\'m at a gay bar.  That was a pun on the word \"prick\", not a homophobic slam claiming gay men are pricks. But I\'m still hip, dorks, because I\'m listening to U2\'s \"Stuck in a Moment\" while the sun shines outside.  This is GreatDayOutside Pt. 2.  It is such a nice reprieve from the usual Seattle \"What\'d you expect? It\'s Seattle\" rain and grey.   I saw Mt. Rainier yesterday for the first time in a couple of months.  I always forget how freaking huge this mountain is or exactly where it is on the horizon depending on where I am.  So, sometimes I\'ll be driving around a bend and Mt. Rainier will pop out and go, \"Ah-ha, it\'s mountain time!\" (A time zone joke!  Maybe the first of it\'s kind ... ever!).   Say, speaking of sequels, I just saw online that they are making Die Hard 4: Die Hardest.  Nothing really to say about it, just say they are making it.  Well, maybe they could have thought of a better sequel name.  Maybe \"Die Hard: Hell\'s Doorway\", \"Die Hard: Shotgun of Pain\" or \"Die Hard: 18-Inch Machete Wound\". My dad was in New York this week for the Westminster Dog Show (he was watching not participating).  Some dog won yesterday.   Oh, I looked up information today about an inn that R. Tickles and I want to stay at when we get our tax return money back.  I could tell you about it or you could look at the web site.  Guess which I\'m going with ... http://www.oceanshores.com/lodging/JudithAnn/ I made a budget for the next few weeks today and I got my watch back yesterday that has had a dead battery for two years.  I like being productive ... just like you like being a dork. So, it looks like Kerry might be the next great Clinton hope, eh?  Do they just have interns to bang them?  I mean can I just stop saying girlfriend and say \"intern\" instead?  They always told us in college to get an internship, they also say college is the time to explore your sexuality.  Maybe our universities are just whoring us out to the business people of the world.  Magna Cum Laude, eh?  Sounds like an escort agency name if I ever heard of one.  Hmmm, eh?, hmmm, then eh?.  It\'s time to go do laundry, dorks .... or maybe an intern.  Ah, boo-yeah! Oh yeah, Ramones (that\'s what the R. stands for, remember?) Girl said that I jumped out of bed last night and began walking around her apartment.  I vaguely remember stumbling around in search of a glass of water trying to explain that I didn\'t understand or didn\'t know.  What didn\'t I know?  I don\'t know.   R.G. had a more lucent memory, which is that I was stumbling around without any pants on talking about how confused I was.  She just rolled over and went back to bed.  Now that\'s what I call love, ignoring psychosis. '),(3021,1785539,'06,August,2004','2004-08-06 00:00:00','   Today is such a freaking beautiful day in Seattle.  The weather is just right for me to wear my new grey lambswool sweater over my hip light/dark blue plaid shirt.  And with my wind tossed hair, OOOOOOOO-EE, I am looking good and feeling a little like Al Joad in \'The Grapes of Wrath\' ... that\'d be after-church-pre-desperate-flee-to-California Al Joad.  But you knew that, my Okidorks. It\'s such a nice day that I was only momentarily considering an uzi to solve the following problem ... (For the following scenario I am going to replace the monotonous duty that I do every day with the phrase \"wear tennis shoes\", another activity that I do every day) Me: (Walking down the hall) Pretty Br-ent, walking down the street ... Pretty Br-ent with Sauconys on his feet. This woman: (Running up behind me) Brent! Brent!  I really need you to \"wear tennis shoes\" today.  It\'s very important.   Me: O.K., just so you know, I \"wear tennis shoes\" everyday. This woman: Oh, well I need you to (no joke) triple \"wear tennis shoes\". Me: (Confused, very confused) There was even a note attached to the thing that I was supposed to \"wear tennis shoes\".  It reads ...  Brent, This needs to \"wear tennis shoes\" - no acceptions! Thanks, This woman  She really wrote \"no acceptions\", it\'s not another of my AC/DC kilt mistakes.  This is very frustrating because the more I try to explain to her (and show her documentation) that I \"wear tennis shoes\" every day the more she becomes convinced that I don\'t. I wonder if this has been a pattern for her entire life. \"Columbus sailed the ocean blue in 14 hundred and ninety-two!\" \"1947, hmm?\" \"No, 1492.\" \"Yep, 1947 sounds right.\" \"1492.\" \"That\'s right, in 19 hundred and forty-seven all good Columbuses went to heaven.\" But, like I said, it was too nice of a day to consider uzi therapy for more than 12 seconds. I went and got R. Valentine her, well, her Valentine\'s Day present today.  What did I get her?  It\'s a really cool little ... can\'t fool me, dorks.  I know you\'re all agents for the man.   \"But we aren\'t agents for the man.\" \"Agents for the man.  I should have known.\" \"But this paper here says that I\'m not an agent for the man.\" \"That\'s right, that paper says you are an agent for the man.\" Actually, that would be a pretty fun outlook on life. Well, I need to run out of this office now, my blog buddies.  It\'s time to stash gifts, eat protein and drive to see my snickerdoodles for some pseudo-spring loving. '),(3022,1785539,'06,August,2004','2004-08-06 00:00:00','   Dorkensteins!  I just had my mind tripped like the clown/bourbon warp on Saturday.  I went to Google.com and typed in \"dorks don\'t rock\" (in quotation marks) to see what comes  up... why, instead of working?  Because my ego is grander than any one thing else EVER!   Anyways, the third find is this german website that lists \"german language blogs\" and has a link to mine.  Das es vat?  Weird?  Not so much yet.  I translated the page to English all Sherlock Holmes styley ... AND! .... Dorks Don\'t Rock in English to German to English translation is Dorks Don\'t SKIRT!  Maybe skirts DO Rock (skirt=rock).   Examples from history: 1.)Catholic school girls wear skirts.  Rumble my rocks. 2.)AC/DC rocked in a skirt (One day after writing this it has hit me that none of the members of AC/DC wore a skirt.  It was short shorts, which are very close to skirts on the efiminate scale, that I was thinking of.  But, I\'ve thought of another example ... Rowdy Roddy Piper.  I feel like such an idiot for messing up an AC/DC reference.  Hellls bells.). 3.)Those Scotish caber tossers wear skirts.  Sorry, kilts, me brethren.  And while we are at it, can\'t we just say skirt instead of kilt.  Hasn\'t a drunk man throwing a log really far and a drunk man wailing on the axe (turnbuckle) justified the skirt?  I mean, I know that they are traditionally called kilts, but my grandpa traditionally wore \"britches\" but I don\'t get all big-cocked when someone says, \"Nice pants, Brent.\" \"What? These are brithces, queerbo, only ladies wear pants.  Maybe I need to punch you a new zip code.\" \"What?\" So, you see my point.  What was I talking about?   Well, while my A.D.D. catches up with me, this reminds me of something else.  I really value R. Girl more every day for many reasons, but one is that she remembers all those things I forget (90% of the world).  Lots of times she\'ll say something funny and I\'ll laugh and I\'ll ask her what she\'s talking about and she\'ll go \"Remember when you said that when that thing happened\".  Or I\'ll go, \"Do you use doors or windows to leave and enter buildings?\" and she\'ll say \"Doors\" and I\'ll go \"The Doors are a band, not a portal\" ... Oh yeah, the German website.  Weird, huh?  I\'m pretty honored that somewhere a German kid might be reading this.  Maybe I should pass on some knowledge that I\'ve learned from watching Schindler\'s List and eating Bratwurst.  Why not, it\'s a free country, unlike Germany.  So.... Dorks Don\'t Rock presents Brent\'s Advice to Germans Reading This ... Hi, there, I heard you\'re German.  You\'re very handsome or pretty or ugly, did you know that?  I bet you did.  I bet you hear how handsome, pretty or ugly you are all the time down at the old Bierhaus.  You know you should really be careful how much you drink, your countrymen aren\'t known for their steadfast tempers.  Oh, I know you are different, I know.  It\'s just that you might want to know that the rest of the world thinks you are probably a Nazi.  What?  I don\'t make the rules, I just live by them.  Now, come on.  Let\'s see a smile.  You do make good cars and Ramstein was fun for a few songs.  There you go, that\'s my Hansel or Gretel.  Now, you go show the world how not a Nazi you are with you little sausages and warm beers. Well, dorks, on that diplomatic note I do think I am going to go do something else which doesn\'t involve talking to Germans.   Oh, before I go, is it just me or is Howard Dean becoming that dude that couldn\'t stop going back to his high school after he graduated. \"Come on, dudes, we\'ll get some beers and totally vote for me.\" \"Um, maybe later, Howard, we have lots of homework to do.\" Yeah, I know, he was my boy, but now he\'s just another dork like you. '),(3023,1785539,'06,August,2004','2004-08-06 00:00:00','   Jeez-um-craw, dorkwads, I\'ve had a busy day.  Or maybe it was a busy past couple of days, hmmm, hmmm?  And did I mention I am not feeling well? I\'ve been on a productive spree.  I have done my taxes, switched my cell phone to a hipper flip phone, paid bills, seen sites I\'ve meant to see, spent time with my R. babe, caught up on films I want to see, re-organized my office slowly but surely (my official name plate came today, I\'m officially non-official) maintained a healthy diet, caught up on sleep, and other things which I\'m sure a pointless to list because who is really reading this far. Update on my Hendrix-drinking experience the other night ... I have no idea what I\'m talking about.  I remember seeing a painting of Hendrix eating a bowl of soup and I was thinking that it was something \"at the Mecca\", but now I can\'t find it anywhere.  I think I just saw the place, thought \"Oh, a cafe, they have soup, Hendrix ate soup in a picture, Hendrix was there\" and went in.  Curses, my demi-tourist status/faultifourous logic fooled me again!  Anyways, I forgot to mention that there were a bunch of clowns there.  Not jerks, actual clowns.  Real live fucking clowns in a bar.  Can you wrap your head around this?  Imagine clowns being anywhere.  Weird, right?  Now imagine seeing a clown\'s face as you sip bourbon.  This is my life. Do you want to know how good a guy I am?  Really, do you?  Listen ... When R. Girl gets off from working the night shift (6 p.m. to 6 a.m.) at the place where they help sick kids I wake up early in the morning to go over to her apartment and crawl into bed with her.  This way it\'s like we spent the night together. Choke back the tears, dorks. Fuckin\' OutKast, bro, right on. And I just want to say that it\'s sad when we have to give almost as many awards to dead musicians than live ones.  No, wait, I want to say something more ... fuck you for dedicating every fucking thing to Johnny Cash when none of you have the balls to stand up to the corporate machine that puts out shit like Evafuckmyessence and that puts you on a 10 minute delay because they are afraid you might show some tit for a second.  Johnny mutha punching Cash would have stuck his mutha kicking finger in the gosh darn fuck air and said \"For Janet\".   And Warren Zevon?  Christ punch a medic, \"Werewolves of London\" was good, but dude that White Stripes bass lick is too tasty to not super size with the Grammy.  Mutha huggers, I hope someone interesting dies or gets cancer next year so we can have better tributes and awards. Like I mentioned before I am not feeling well.  I think I\'ve caught something that has been going around this office.  I feel like I need to rest but I feel to sick to get any real sleep so my brain is going \"Ah, why bother\".   All right, it\'s time for me to skee-mutha-lovin-daddle. '),(3024,1785539,'06,August,2004','2004-08-06 00:00:00','   Morning, dorkies.  Happy Monday to you ... I mean \"Fucking Monday\", you know, \'cause I\'m hip.   I have a bit of work to be tending to, me dorkles, so read you this about a lad and a wee bit of \"hooliganism\".  Didn\'a know it was a word.  http://apnews.myway.com/article/20040209/D80JQMM00.html '),(3025,1785539,'06,August,2004','2004-08-06 00:00:00','   Yeah, it\'s late, and maybe I\'ve been drinking, dorks, what do you care. Doing a little web-based research I realized I haven\'t typed in this thing for a few days.   Well, let me tell you something.  The R. had to work at the place where the save kids tonight so I decided to (I\'m not correcting any mistakes I make in pucntuation or grammar from here out, dorks, it\'s taking too long).   So, today I bought a couple of lamsbs wool sweaters at a GAP outlet.  That was spectacular.  They were only 10 dollars a pop and so a I coutgot two.  HMMM, and them I wasnt\' feeling too good adn R. baby had to go to wo rk and so I went asn saw Mystic River with the lovely couple.   Robert Downey jr. is good, but I don\'t mean robert downey jr. I mean the other guy, sean penn.  He was reall y good in that movie.  Every time I see sehan penn on the screen I am like \"dude, that\'s madoneas\'sas fuck buddy\" but no, the dude can act.  Like I\'m a judge of anything. \"dead man walking, more like madonnas tits a sucking ... hehehehehe, know what I mean, he dated her, dude, totally, he had to suck on those things ... hahahhahaha, dammit, dude, madonna\'s tits\" So, I went downtonwn with Kyeele tontight and had one drink at theis bar that I rmember reading about in a Hendrix bio or seeing in a Hendrix screen print.  Either way it coust me almost 16 dollars for TWO miller genuiine drafts and a jack daniels and a yager shot.  I don\'t think jimi would have paye dthose prices even with inflation.   (I had to to back and correct my misspelling of Jimmiy\" becasue I have to sound li ke I know my hendrix to remain hip, dorks) So, I guess I was taken becasue I liked \"all along the watchtower\".   I think that\'s about it.  I just wanted to let you know taht I\'m enjoying drikingin this beer and this bourbon and possibly this beer.  I\'m sleepy.  Whend is it time to say god night?  Now! Wait ... I bet you wonder how I got so drunk. Right? I drank the rest of the bourbon letft over from my touesday tirade.  This shith is good, dorks.  I\'d tell you I love you all but ehnn you\'d asay I was just drunk. '),(3026,1785539,'06,August,2004','2004-08-06 00:00:00','   I forgot to mention that my little brother (who is now bigger than me) has started a blog.  He also has blue hair.  Blue fucking hair.  And he\'s a proud geek.  AND he has blue hair, he used to be an aryan now he\'s a smurf.   www.omglolh4x.com '),(3027,1785539,'06,August,2004','2004-08-06 00:00:00','   I personally know nothing about how to make music.   I\'m going somewhere with this, dorks, so calm down a minute.  Dammit, you are impaitient.  Now, as I was saying, I don\'t know how to make the music come out of them music machines. I taught myself how to play the harmonica when I was younger.  The musical fruits of my labor (hehe, beans) was that I could play a pretty good \"When the Saints Come to March In Our Town ... Pilaging the Land!\" and \"Low Rider\" by WAR.  I can still play hobo-worthy harmonica but I can\'t wail like Eddie Van Halen.  But that doesn\'t mean I don\'t know who to rock your balls, dorks! What does this crazy guy mean?  Look!   I left the following message on my friend Eva\'s (http://themeasure.diaryland.com) answering machine the other night (and, yes, she correctly guessed that I was NOT drunk when I left it): EVA! I WAS JUST DRIVING IN MY TRUCK AND I HAD THE BEST IDEA! WOULDNâ€™T IT BE AWESOME IF IN THE MIDDLE OF A SHOW YOU JUST STOPPED WHAT YOU WERE DOING AND BROKE INTO THIS PART OF BOHEMIAN RHAPSODY BY QUEEN? HERE, THIS PART! [dull blare of car stereo] WOULDNâ€™T THAT BE AWESOME?!?! The part I\'m talking about is, of course, the classic Wayne\'s World head thrash part.  Eva\'s in a band called Perfect Sleeper so she can do this. But think about how wicked sweet it would be if a band did this.  They\'d be singing along all emo-alternative and then maybe at the end of a song they\'d do that cool thing where they hit the strings one last time and let it fade out slow while they look at the ground with their hair in their face when all of a sudden ... Wha-wha-wha-wha-wha-wha wha wha-wha-wha-wha SO, YOU THINK YOU CAN STOP ME AND SPIT AND MY EYYYE!  SO YOU THINK YOU CAN LOVE ME AND LEAVE ME TO DIIIIIIIE! OHHHHH BABEEE, CAN\'T DO THIS TO ME BABEEE, JUST GOTTA GET OUT, JUST GOTTA GET RIGHT OUT OF HEEERE. (I just wrote those lyrics without looking them up on the web!  I rule!) So, my point is, dorkensteins, that I know how to rock, but I just don\'t know how to make it roll.  I want to be in a rock band just so I can have awesome ideas like that and have them happen before me like a babe born anew of Zeppelin and G. \'n\' R. And I\'ve got ideas, oh yes, ideas in my head are brewin\', but I taint telling you none of them \'cause one day you will see them on MTV and your head with explode from THE EUPHORIA! I am hungover this morning, dorks.  I drank 2 40s and a few other beers and played some muthadominating RISK with the better half of the L.C., the other half being in Cali. or Haiti or something for the week. How\'d I do at RISK?  I lost.  Then we played Trivial Pursuit.  How\'d I do?  I lost.  Although with the Trivial Yah game I think the questions were a little easy for certain people who weren\'t me. Example question for the better half: \"You eat off of this and it rhymes with plate.\" Example question for me: \"14, Stevenson or that apple?\"    I guessed 14.  I lost.  So, I\'m hungover and very tired. I called R. There-for-me a couple times when I was wasted last night.  Always a good idea. \"Hey, I\'m wasted. Who wants a tickle?\" \"You\'re on a phone, you can\'t tickle.\" \"What?  You\'re cute.\" He,he I just called UPS and they say \"What brown can do for you.\" Brown do.  He, he.  Fuck I can\'t stop giggling about this. \"Sir, we lost your package.\" \"He, he, package. Do.\" Then they said \"Check us out at UPS.com\" \"He, he, check.  Like what I write at the grocery store.  Check.\" Allright, I need to fucking work.  Damn, I have been writing this thing for almost an hour!  Need to go, dork tickles. Wait!! One more thing. I went to return the digital camera that I got R. Camera for Christmas.  It broke when I was trying to take a picture of her riding a statue of a monkey at the zoo. Here\'s what happened at Best Buy. \"Sorry, we can\'t replace this.  The return date has passed and you didn\'t get the stupid, stupid extended warranty deal that we should have on all of our products anyways but we don\'t because we like to ass fuck you because we know you\'ll keep coming back even if we ass fuck you because we have lots of shiny things.   Lube up!\" \"But, it\'s a shoddy product.  It broke, it\'s not working, you should take it back.\" \"I\'m sorry (shoving the receipt at me) but the exchange date has passed.\" \"You don\'t need to ... so, you can\'t do anything for me.\" \"No.\" \"So then I guess I got (oh, it\'s soo good, wait for it just a second, rewind it and turn it up because this is going to be good!) ... \"So then I guess I got .... THE WORST BUY!\" I actually said this and many people heard me and I walked out of there like a king.  Goodbye, dorks, I rock. '),(3028,1785539,'06,August,2004','2004-08-06 00:00:00','   I tell you, dorkfaces, sometimes I think this whole world is full of people with rubber-filled heads.  That\'s right, rubber-filled heads.   I just showed someone in the executive office how to ... a.)Convert dollars to euros b.)Build a rocket to take them back to their rubberhead planet or  c.) Use a folder.  A regular freakin\' paper folder.  A folder that would not be unlike the folder that has been used for many a year by many a beurocrat.  A  regular (although it was red) paper folder.  I think it needs to be said again for effect.  A regular paper folder.  A red regular paper folder. Ok, now you\'re supposed to guess.  If you guessed \"c\" you are not a rubberhead.  If you didn\'t guess \"c\" I have a rocketship waiting for you. \"This folder not be da one me usededed to be using.\" \"Here, you open it up and then ...\" \"Oh, that\'s that there\'s the folder I be seeing when i use the folders.  Yep I likes the open folders.\" I\'m glad that rubberheaded people make 3 times as much as me; it really helps me stay focused and well-balanced.  Very zen. I hope your dorky weekend was nice and that you all ate chocolate and got hit with Cupid\'s arrow.  Not cupid the love guy, Cupid this meth addict I know with a fondness for crossbows.  That\'s right, I hired a cranked out crossbow nut to hunt you down, you smirksome rubberhead. But really, I had a great weekend.  Me and the R. chicky went out to eat and held hands.  I gave her a journal type thing and a nice pen for Valentine\'s.  What did she get me?  Well, I got porn. I know I told some of you that R. Virgin got me other movies.  This was a lie, a boldfaced lie, I lied to you.   Why? You see, I did this because I didn\'t know if I wanted to tell anyone that I owned porn and that I mutually owned porn and that I received it as a gift and yadda, porn, smut, other stuff, hmmm, television, I wonder what\'s on. But, then I realized this little fact. November 2003 - R. Touchdown gives me football tickets for my birthday. December 2003 - R. Christ gives me a b.b. gun for Christmas. February 2004 - My R. gives me porn for Valentine\'s Football, guns and porn.  I think that needs to be said again.  Football, guns and porn.  How perfect is this chick, I mean babe.  Plus, who the hell doesn\'t own porn?  I mean, the exception is that Catholic priests don\'t own porn and we all know what\'s going on there. That\'s right, if you don\'t own pornography you have to wear very tight collars.  And lord knows I like a loose collar. On Sunday (the day of the tight collar) the me-couple and the lovely couple went to the zoo.  It was very fun.  One of the elephants had half of his (her) ear gone like it had been ripped off.  That was awesome. There was also this sign that talked about how a certain type of monkey will get into fights with eagles.  It had a painting showing what a monkey/eagle fight looks like.  What does it look like?  Like this ... I have said in my life that a lot of things are the coolest thing I have ever seen.  Santa Claus, people falling while sking and gators jumping are a few of those things.  However, all of those things are lame ass shit-fuck compared to monkey/eagle fights.   A group of these certain monkies will be sitting in a tree just hanging out. \"It sure is a good day to be a monkey,\" they will be saying. Suddenly, an eagle will swoop in and sit next to the monkies in the tree.  He is the spotter.  \"No, it\'s a bad day to be a monkey,\" the eagle will say.  Then his buddy swoops in and starts picking up monkeys. (This next part was not mentioned in the zoo display, but here\'s how I think it goes) The monkies, flying high above the ground held tightly by the eagle, will fight and claw.  The eagle will do sky-cartwheels and zigzags to confuse the monkey.  The monkey is too smart for this and pokes at the eagle\'s eyes. His attempts are futile. The eagle pokes back with his razor sharp beak, slicing the monkey\'s left ear.  This angers the monkey who swings violently at the eagles face.  The blow to the head shockes the eagle who momentarily loses control of his flight.  Both the eagle and the monkey are falling to the ground at TERMINAL VELOCITY!  Just before impact the eagle regains control and swoops up.  The camera zooms in on the eagles claw (that\'s right it\'s a movie now) only to reveal it covered in blood.  The monkey could not escape TERMINAL VELOCITY!  Before I die I want to witness God\'s most perfect conflict ... the monkey/eagle fight. There were also bears and zebras and asian people.  But you\'ve seen all of that stuff. Well, I need to go home because I don\'t want to be at work any more.  You dorks are all right, ya crazy rubberheads. '),(3029,1785539,'06,August,2004','2004-08-06 00:00:00','   \"So, do you want me to have this in Portland by Saturday or Monday\"  (For this one, dorks, remember it\'s Friday) \"Um, which ever one will get it there faster.\" Pause. This is the lady that I had to explain a folder to.  Where the fuck did this rubberhead come from? \"That\'d be Saturday.\" \"Yeah, let\'s go for Saturday.\" That\'s a way to wrap up and summarize this past week.  Rubberiffic, and fucking smirksome.  Long.  Talking \'bout my week, not my ... That\'s it!  I\'m out of here.  Time for an oil change and red-headed loving.  Talking \'bout my girlfriend. Weekend, dorks. '),(3030,1785539,'06,August,2004','2004-08-06 00:00:00','   Yeah, I know, dorks.  I\'m sorry that my blogging this past week has been very none-done.  I did try to write one yesterday when I accidently closed the wrong window (the bloggin\' one) and lost it.   \"Aw, fuck it,\" I thought.  And fuck it I did. So, like I was telling you yesterday but you didn\'t know, this has been media week for me. I bought three new, used C.D.\'s last Friday. Green Day \"Dookie\" (Awesome reminder that those fuckers who walk around going \"This isn\'t punk.  Punk is punk I\'m punk, this ripped sweater is punk\" aren\'t punk.  And really no one is punk, dorks, it was a stupid, stupid way to sell some records in the mid-to-late seventies since disco and stadium rock had saturated the public ear.  The Sex Pistols WERE (don\'t ever be fooled into thinking they weren\'t ... it\'s not very punk to think otherwise) the SECOND boy band.  The first being the Monkees.   Punk music is still good, however, punks are dorks with spikes.) RZA presents \"Wu-Tang Killa Bees\" (Good truck drivin\' music) and  The Soundtrack for \"Dogtown and Z-Boys\" (or something like that.  I\'m not going to pretend I know the title correctly and falsely have you assume I\'m down with skating culture). All good C.D.\'s. Also bought \"Death Race 2000\" and \"Rollerball\" (the Jimmy Caan one) and rented \"Crouching Tiger, Hidden Dragon\" All excellent movies.  Best lines in each movie respectively ....  Cleopatra: It isn\'t my fault everyone scored before us. You should have gone after that boy scout camp like I told you!  Nero the Hero: I tried the goddamn boy scout camp. You know how fast those boy scouts move?  Cleopatra: Now here\'s something more your speed.  Nero the Hero: That\'ll be at least 200 points!  Cleopatra: If they scatter, go for the baby and the mother  Then ...  \"It was never meant to be a game!\"  Then...  \"Fooeee, where\'d I put my sword.  Oh-No!  A ninja!\"  I really want to use that Rollerball line at some point in life. \"Hmmm, this fish is a bit gamey.\" \"It was never meant to be a game!\"  And to top off Media Week I\'ve been reading (I think I mentioned this ... oh yeah, I did) \"The Dice Man\" by Luke Rhinehart. Very excellent book, please read it.   For Christ\'s sake, calm down.  I\'ll give you a summary b/c I know some of you dorks can\'t read. This guy invents a way of living where dice decide what he does with his life.  There are some rules, the major one being that he has to do what the die say.  Here\'s how you do it. Think of six things to do and let it represent one of the numbers on a die.  Or maybe think of 4 things to do and allow a couple numbers to represent one thing to do ... Example 1.) Find a hot homeless person and give them $5 2.) Get drunk  3.) Clean your home 4.) Call someone you haven\'t talked to in a while 5.) Hit on the next stranger you see 6.) Climb a large hill   ... you\'ll understand as you go.  They can\'t be things that you woulnd\'t do.  You have to do them without thought or ego once the die is cast.   So, I\'ve tried it a few times.  In the book the di command the guy to scream that he\'s Batman and bang lots of broads, the di commanded me today to go eat an unhealthy lunch in the park.  Of course, I don\'t want to bang a lot of chicks because .... R. Sense-of-humor and I are getting antsy about our big move.  I think we\'re both over the nervouse \"What the hell did I just agree to?\" stage and at the \"Fucks sake, let\'s do this already\" stage.  I remember when it was just \"frame stage\".    Boy, who doesn\'t want to see the new Jesus movie?  Me.  Is it just me or does watching Jesus get beat up for a couple hours sound morbidly interesting and therefore totally out of the realm of being able to go see because of the fires of Hell (that I don\'t believe in) that would await me if for just one second I had the thought \"Man, this Christ guy really just can\'t take a punch\". And what the hell is up with the F.C.C. since LaToya whipped out her tit?  I mean Mel-Gibson-sucker-you-all, you pull out a boob for a second (a black boob remember!) and all of a sudden no one can say \"pee-pee\".  But show two hours of Jesus getting whipped and everyone goes \"Hey, let\'s rent out a movie theater so my kid can watch this.\" And while I\'m going on about current events, I need to say something about Rosie O\'Donell.  This dork-with-tits is the only celebrity that the gay marriage movement could get to join their cause?  Well, maybe it\'s not a movement as much as a natural and A-O.K.-with-people-who-are-secure-and-don\'t-hate-their-fathers-and-jerk-it-to-Freddy-Prinze-Jr. progression in civil rights, but anyways ... getting Rosie O\'Donell to stand up for your cause is like having the one armed kid from my high school throw out the first pitch at the World Series ... THEY\'RE GOING TO FUCK UP SOMETHING IMPORTANT!   We all know Rosie is going to go to San Fran and be all snoddy to people like when she (DAMMIT!) would always give the person the correct answer on Hollywood Squares!  I mean for fucks sake Rosie, it\'s a game stop acting like you do anything besides eat and throw koosh balls all day.  She\'s going to piss off different leaders who will rush to shut down what would have been quiet civil disobedience. \"I\'ll take gay marriage for the win.\" Anyways, I think that\'s enough dorks rocking for the past few days I\'ve missed.  I\'ll talk to you all tomorrow and tuck you in for the weekend.   And the dork sayeth on to ye, go forth and do unto dorks as you\'d have dorks do onto you. '),(3031,1785539,'06,August,2004','2004-08-06 00:00:00','   O.K., Ameridorks, nows the time to saddle up and kick some old white guy ass.  We\'ve lost Dean.  That\'s all right, the Democratic Army can still win the fight .... IF, stupid Ralph Nader drops out of this race.   Ralph Nader is the biggest dork in the world.  Even worse, he will never be president because he doesn\'t have much more than economic ground to stand on.  And even worse he gave us that momma\'s-daddy\'s boy of a president we got now.  I\'m a seventies-era porn man myself, but I hate me some Bush. So I\'ve been sending e-mails to info@votenader.org (it\'s the only e-mail address I could find on www.votenader.org).  I\'ve probably sent out almost 100 in just 45 minutes of work.  Why not actual work, Brent?  Because I\'m working for America now, dammit!   You should all send similar emails to this address!  I\'ve just been writing things like \"No Nader = No Bush\" and \"Rember 2000?\" or \"Two parties bad?  Three parties worse!\" in the subject line and then pasting a simple message as such ... Please do not run for the office of President of the United States!  You know in your heart that this will only split the Democratic party and fuel a Republican victory.  Please support the Democratic candidate! Thank you, Brent Kinkade ... to the body of the e-mail. Tell your friends to do it too.  If this fucker runs than it\'s 4 more years of Bush.  That\'s four more years of your dad\'s friends running the country.  America won\'t be hip anymore kids.  The old white terrorists will have won.  So, how was your weekend?  Mine was superfanfuckular.  It was such a beautiful Seattle weekend that on Saturday R. Bedbuddy and I laid in bed all day and looked out the window.  We had grown sick of doing things on the weekends so we decided to do nothing.  And nothing we did. R. Girl had to go to work at 4:fucking:30 on Sunday morning ... and with no sleep thanks to a drunk me.   \"You asleep?  Christina Agulkaearea sucks on Saturday Night Live ... wait, it\'s not live here.  Did you realize that?  Hey, are you awake?\" But I didn\'t have to work, so I slept until 10 in the morn and then went and played a little disc golf with the L.C.  I have to say that I played pretty well for not playing in a few months.  Now that frisbee fever is back in my veins.  I need to get a couple more discs to round out my set and I\'ll be in good shape.  Also need to work on my drive.  Can\'t seem to get enough power behind the disc.  And, no, I\'m not a girl. After that the L.C. had another L.C. over for dinner, except the L stands for \"loud\" this time.  We had dinner and lovely conversation about lesbians, marijuana, sirens, dogs, dogs and sirens, and ladders.  I now know where to go for random gay sex in Seattle.  Heavens, what I needed to know is where to get a good slice of meringue instead I got the whole pie! Today I\'ve been reading \'The Dice Man\' by Luke Rhinehart. It\'s pretty good pre-Fight Club male angst stuff so far.   It\'s now time for ... The B-Man\'s \"After 47 pages\" book review! The book opened strong with much talk about disillusionment and disassociation with society.  Then there was some talk about this guys wife\'s ass which was cool, but was quickly replaced by talk about an egg breakfast with this guy\'s kids.  Then there was talk about murder and Christ and the Buddha.  I think someone took tranqulizers.  A must read! Thank you. Anways, dorks I need to get back to my real job.  But you need to do my bidding and send out those \"No Ralph\" e-mails.  If you don\'t, a New York City fireman will cry himself to sleep.  And you don\'t want that do you?  A good American would do anything for a brave New York City fireman.  Are you a good American, dork? '),(3032,1785539,'06,August,2004','2004-08-06 00:00:00','   Dorks Say Sorry. I was wrong when I said in a previous blog that Rosie O\'Donnell was going to piss people off with her marriage.  Kudos and kooshes to Mrsissz. O\'Donnell and her new husbawife. I wasn\'t wrong about the monkey/eagle fight. ');
INSERT INTO `msgs` VALUES (3033,1785539,'06,August,2004','2004-08-06 00:00:00','   Martha Stewart is guilty, Howard Stern is gagged, a 24 hour a day search for Bin Laden is underway (the CIA wasn\'t leading up this \"convenience hunt\" from the begining?  \"Ah, come on, we looked for him yesterday.  Plus, Dr. Phil is on.\"), Bush uses images of 9/11 to get your vote (\"It happened under me\" being the appropriate slogan), job market predictions have been cut back, my dad just told me he isn\'t going to be helping with my student loans, British troops plan to be in Iraq for at least two more years and many Iraqi \"leaders\" (yeah, and Bugs Bunny was a fucking rabbit, these guys are more puppet than Muppet), for some reason Ben Stiller makes more movies than the Beastie Boys put out albums, my goddamned shoes hurt, Bishops are gay, gays can\'t marry here, gays aren\'t supposed to be here, a tit has shook up the world, the F.C.C. is headed up by a Powell and so is our war, probes, leaks, Haiti, Nader = 4 more years, stamps for e-mail?, Atkins was obese when he died, Giuliani defends everyone, Kobe rapes ... he scores, little boys being touched, gays, bombs, is Hitler back?, high gas prices, lower wages, air, clouds, bird flu, mercury, carbon, hydrogen and oxygen ... and then you see this. http://www.local6.com/slideshow/news/2900374/detail.html?qs=;s=3;p=news;dm=ss;w=320 Have a good weekend, dorks.  If the world is ending, you\'ve seen it all. '),(3034,1785539,'06,August,2004','2004-08-06 00:00:00','   The female part of the Lovely Couple: I don\'t know the soup was good but the prawns were grey so, of course, they were a bit gamey. Me: (Snapping to attention) IT WAS NEVER MEANT TO BE A GAME!  I got to use my line!  Oh joy, oh meat, oh blessed dorks!  Skin, eyeballs, Nikes and all! '),(3035,1785539,'06,August,2004','2004-08-06 00:00:00','   I write no blog this week.  You know.  I know.  Hulk smash. I could not wake up today ... but boy-oh-boy I could wake up yesterday.  Because .... R. Girl: mmmmm (you know the morning groan of dream-bliss and awake-realization) Happy Birthday ... Me Boy: Thank you, but it\'s not my birthday. Even if it wasn\'t my birthday yesterday it sure was  great way to wake up.  I wish stuff like this would happen more often. Typical jerk sales clerk: Thanks for buying socks, Happy Anniversary. This reminds me that I used to have (when I had my own bathroom during college) a piece of paper torn out from a magazine that just said \"Christmas\" over my mirror.  I think the same principle is at work here. So, guess what I do this weekend, dorks.  Hulk guess. I\'m putting my security deposit down on an apartment that R. Cupcakes and I finally decided on.  It\'s damned near official, dorks.   Hulk nervous.  Hulk never live with girl before.  Well, Hulk live with girl roommates.  Hulk know isn\'t same thing.  When girl roommates have girl boo-boos Hulk say \"Drink, hush\" and girl roommates go to boyfriend.   Hulk sure Hulk be fine.  Hulk like redhead girl. You know I feel like I need to go off on all of the insanity going on under the Bush regime ... -Howard Stern being pulled off the air (\"But that was Clear Channel that pulled him with no pressure from the government.\"  Shut up and suck Dr. Phil\'s cock, Oblivious Boy, he\'s near climax ... And isn\'t that a good point.) - The march of the anti-gay bootstraps. - The anti-science \"But ma baby\'s imbilical cord should be placed in a holy trash receptacle \'stead of being used to help the cripple boy walk\" and \"T\'aint never been no water on Mars\" ... Oh, the Mars thing!  Why are people scared of water on Mars?  Why?  I know!  I know!  Dammit, dorks, I know!  If there was water on Mars then maybe there could have been LIFE on Mars!  And if there was life on Mars than what does that say for the creationists, the Baptists ... your God, What does that say for the entire South?  If God didn\'t just make life on Earth than maybe he didn\'t make just Adam and Eve, maybe he did make Adam and Steve (Arrrrrrrgggghhhh, the anger from this phrase makes me want to run my truck over a panda), maybe Moses didn\'t part the Red Sea, maybe Jesus could fly (awesome), maybe Cain and Abel were cyborgs sent back to decide the future of mankind (Cain: Step away from Eve.  Abel: Never Cain 5000!  You may have the shotgun but I\'ve got the dynamite!) ... Maybe an Army of gay men and women and people who don\'t want to see the new Mel Gibson movie and people who don\'t feel they need a gun to gain an erection or that we shouldn\'t hit Hindus and that a pious bastard is worse than one who commits adultery, maybe this army will rise up from under the grip of \"God\'s Army\" and say unto those who are left open mouthed and jaws in their limp pants licking and eyes bulged out BECAUSE they have witnessed that man can be strong and glorious without a God and they will see this new race of men powered by the knowledge of Mars life and ask onto they \"But what can I believe in now that my superstitous beliefs have gone\" and the new man will say \"Lo, but believe in thy self\"  and old man will say \"But I suck and I\'m scared\" and new man will smile and say \"You\'ve got some Dr. Phil juice on your lip.\" -Um, where to go from there? The Oscars were pretty good ... Oh yeah!  Why does every critic keep saying \"The Bored of the Rings\" \"I\'m so bored by it, Oh it was so boring, Bored\"  ... Harry Potter rim an atheist,  what\'s wrong with these people?  \"Oh, I have to pay attention for more than a minute and there\'s no Hillary Duff?  Boring.\"  and ... \"Lost in Translation was boring, so boring, nothing happened, it was boring.\"  No it was different not boring and don\'t worry too much some new \'Lion King\' sequel will be out soon enough for you.  \"He, he, that ol\' warthog pooted\". Now put your lips back on Dr. Phil he\'s randy again. - And Haiti!  I\'m not going to act like I\'m I know anything more than Haiti except that I do not want to go there ever, but I\'m not an idiot. Stupid president in the news: \"Maybe Aristide should step down.\" (Cut to the next day: Aristide \"leaves\" for Africa.) Aristide: \"I was forced to leave at gunpoint by the Bush machine.\" Bush machine: \"Naw, come on, America, the silly black man is lying to you.  We just happened to send 50 Marines to the country a few days before he left.  50 Marines seems like quite a small number to make any difference in a chaotic country doesn\'t it? But, it seems like more than enough to get one guy out doesn\'t it? Don\'t worry though because ....\" - Something like 180 people died in Iraq yesterday because of sucide bombers.  You don\'t know 180 people.  Can you wrap your head around that?  Everyone you know could have died in Iraq yesterday.  And now they\'re saying the guy that\'s heading up all of these suicide bombings and gureilla tactics in Iraq was within our reach when we first entered the country.  But, no, we had to focus on getting Sadam.  That\'s right, because that hole he was governing had to be liberated.  180 people died yesterday so Bush could make Daddy happy.  180 people.  Everyone you know.  Prior knowledge once again ignored.   JOKE TIME! How can you spot an elephant on the moon? He\'s the one with the big \"E\" on his space suit. And.... (here\'s a joke I made up while I was making a no-bake cheescake the other day) Me: Hey, what do you call a chicken who hits its young? R. Stumped: What? Me: An egg beater. And on another positive note ... naw, we\'ll save that one for later.  But until then, HAPPY BIRTHDAY, DORKS! '),(3036,1785539,'06,August,2004','2004-08-06 00:00:00','   Ok, dorkwads, I\'m a little hungover this morning and I had to pay bills, which might make my statements here a bit biased ... wait, maybe not biased, but skewed ... ah, Christ, dorks, just listen ... Two conversations I had today that made me start to think that something is starting to happen in the world that I should be aware of. (Calling up Best Buy to pay off my card and avoid interest rates, which I can\'t do online because ...) \"So, I can\'t pay it off online because the service makes me post-date the check and I want to pay it now and I can\'t pay it over the phone because it would cost me 15 extra dollars, which would make my efforts moot.  I don\'t know why you make me post date the check, every other company allows you to make the payment that day and have it applied instantly.\" \"Sir, you can not compare our website to another company\'s web site ...\" \"Wait a minute, yes I can.  As a consumer (and a Goddamned bomb dropping American!) I have the right to compare.  Your website is inferior to others.\" What the fuck?  I can\'t compare.  Oh, please be quiet Mr. Consumer and take your ass raping.  Shh, just be quiet.  Don\'t point out how we make it hard for you to pay things off so it makes us easier to get richer and dammit dammit dammit.  Don\'t fucking tell me I can\'t compare.  I dare to compare!  I\'m the compare king.  Your website sucks.  Burger King is better than McDonalds but not Arbys unless I really want a burger.  Apples are better than bannanas.  Leno is a douche, Letterman rules.  Argentina is nicer than Mongolia.  Fuck I\'ll compare all day, ass.  I\'ll compare shit you aren\'t even supposed to compare!  Why? Because I can. A skyscraper is tastier than the moon. Trees are taller than air.  My shirt is more alive than my Burt\'s Beeswax Lip Balm. Don\'t talk bad about the corporation, sir. Fuck the corporation! The corporation is more orange than my left nut.   Deep breath. Next, I\'m at McDonalds this morning ( I know I said Burger King, but McDonalds was closer.  Closer things are less stuff than stuff not closer.) and I see this sign. \"Lunch Policy! ... 16 cents extra for a dipping sauce for nuggets.\" First off, this isn\'t a policy.  Everyone is so excited for everything to be a policy.  Since 9/11 everyone wants to be Giuliani so fucking bad. \"I hereby decree that the department of nugget security henceforth charge 16/100 of one American dollar for BBQ sauce.\" It\'s a rate, it\'s not a policy.  Policies you can debate, there was no debating this RATE. \"Wait, you have to pay 16 cents now for your sauce? It\'s not free any more?\" \"Si. No.\" (He spoke Spanish) \"What? Are you guys going bankrupt?\" Dammit, I don\'t even eat McDonalds any more (except for my rarer by the minute weekday hangover) and I care about this.  I think I might go make signs and protest this weekend. 6 piece = 1 sauce! Say no to sauce fee! Sauce good!  \"Policy\" bad! I\'ve found my cause, dorks.  Have a good weekend.  Love you, mom. '),(3037,1785539,'06,August,2004','2004-08-06 00:00:00','   Dorks, oh dorks, sweet dorks ... you know that part in the 80s movie where the parents come home and the record player goes WHEEEREEEREEE because the precious Izod parents have come home to find that their precious sweet son has thrown an all night party full of strippers and Kelly Le Brock and booze and mutha punching sheep ... CHRIST, DORKS, I SAID SHEEP!  THAT MEANS FUCKED UP SHIT IS COMING UP!  SHEEP MEANS THIS!  Your Mother  Nice to read that you are hemorraghing and doing so well in your social life. Pick up the phone you little skamp!!   This is one of the comments to Kyle\'s little entry about putting his head through a fucking, sorry Mom, stupid window! This means that my mom (by linking to read the Kyle story) has read this site!  Fuck Christ Shit punch the Pope and inhale whippets while a lap dance girl does things and fuck and shit and dammit.  I mean shoot, ma.  Shoot it all to heck! No one else is going to use the word \"skamp\"!  Skamp means my mom!  My mom has read about all of this!  Ahhhhh, no way she could just read this little bit and then stop.  I\'m as adictive as nicotine, which I don\'t inhale on a daily basis, Mom. \"Brent, what\'s a skamp?\" \"Shit, I don\'t know, I only knows \'bout talking bout beers and boobs, I think my ma knows what skamps might sure be.\" And so we begin \"Brent Explains Everything to His Mother\" Hey, Ma ... remember how I call you Ma?  That\'s funny isn\'t it?  I know you read some things on this here website that aren\'t too pretty.  Um, so, I\'m still a virgin ... it\'s just that the other kids laugh at sex jokes.  Yep, and jokes about, um, your girlfriend buying you porn.  That\'s the funniest joke out there now for kids to tell ... \'Knock-knock\' \'Who\'s there?\' \'Your girlfriend.\' \'Your girlfriend who?\' \'I bought you porn and totally respect your mother.\' Hahahahahahahahhahaha. That IS the funniest joke out right now, Mom.  It feels like I\'m cleaning up my room, dorks.  Except it\'s not my teenage \"Hide the Marlboros\" room. Nope, I\'m chasing frantically around my \"Hide the, um, the yep let\'s stick with Marlboros and not dildos\" DAMMIT room. And now all of you dorks are saying, \"Well, hey Brent, why don\'t you ever mention your parents except in passing?\" Why don\'t I? I\'ve noticed that too, dork fellows.  So, now that I\'ve admitted to having a mother and not being the coolest hippest most idependent space commander on the sun I should probably open up about other things ... My parents are divorced.  My Dad lives in Michigan and is remarried.  My mom lives in Phoenix.  She recently admitted that she is an alcoholic and left her chump fiance that kept telling me that he loves me.  This was very uncomfortalbe for me. \"I love you, Brent.\" \"I don\'t know what you look like.\"  At least my stepmother is honest.  She once said to me, \"I don\'t care if you freeze to death.\"  This made me think, \"Hmm, maybe I\'m totally misinterpreting this freezing to death thing. Maybe that\'s how Iowa sluts say \'Hello\'.\" (Guy frozen in Iowa) Iowa girl: Hello to you too. I don\'t talk to either of my parents very often.  By not very often I mean think of the last time you talked to your parents and divide by 3.  That\'s how much I have talked to my parents this year. People:  So, what do your parents do? Me: Um, I don\'t know ... live? And that\'s all you get for tonight about, dorks.  I hope you all freeze to death. '),(3038,1785539,'06,August,2004','2004-08-06 00:00:00','   \"So, um, I have an entry for you ...\" \"What?\" \"You know, for your little blog thing ... I\'ve got something to write about.\" \"OK, Kyle, what?\" \"I\'m in the hospital right now.  I cut my head open and I lost a lot blood and had to ride in the ambulance.\" \"What!?!? Are you O.K.?\" \"Oh yeah, they\'re cleaning up my head wounds right now.  Look, I don\'t feel so good, I\'m going to go.\" This was the call I got last night on my way to R. Cookie\'s.   \"He seemed O.K.,\" I thought.  \"No need to go to the hospital.\" Then I thought, \"Wait, I wonder how fucked up his head is!\"  I couldn\'t miss this!  Kyle\'s head might be wide the fuck open! I might never have a chance to see his skull again! So, being the good brother and avid fan of head wounds that I am, I went to the hospital.  And why did Kyle have a head wound and massive blood loss last night?  This might top the duck story, dorks, it just might. \"Boy, I sure worked hard today, I think I\'ll go eat a pizza and play video games at home.  All I need to do is walk through this door and get in my car.  Yep, just walk through this door and I\'m home free.  I\'ve done this millions of times.  There won\'t be any problems here ...\" LADIES AND GENTLEMEN!  WELCOME TO TONIGHTS MAIN EVENT!  IN THIS CORNER WEIGHING IN AT MORE THAN A MIDGET AND STANDING TALLER THAN, AH, WELL, A MIDGET WE HAVE THE TYPING, THE SCREEN STARING, THE BLUE HAIRED ... THUNDER DORK!  AND IN THE DOOR CORNER STANDING 6 AND HALF FEET TALL WITH A WINDOW MADE OF SOLID GLASS WE HAVE A DOOR ... WITH A WINDOW!  LET\'S GET READY FOR .... HEAD V.S. WINDOW! FIGHT! Kinkade advances toward the door ... he\'s coming out of his corner fast ... the door just standing there, possibly trying to psyche out the Thunder Dork ... it\'s not working, Kinkade is still advancing ... he\'s now trying to push the door, who seems to just be giving in to his brutal attack .. WAIT! the door is swinging back ... OH! the window has laid a vicious left hook to the Thunder Dork\'s right side ... He\'s down!  Kinkade is down!  There\'s blood everywhere!  This fight is over, folks!  The window has won!  The window has won! So, yes, Kyle tried to open a door and instead put his head through a window.  Subtle. Personally I would have just gone through the door, but sometimes I guess it\'s healthy to try new things. So, he had a nice sized cut on his forhead, his right elbow was torn up pretty gory and he had a nice slice on his left arm.  And what did his charts say?  They literaly said ... \"Kyle Kinkade ... Window V.S. Head\".   What did I get to see last night being the only person who got to go back into the E.R. and be with him? -Kyle\'s elbows look like baked chicken on the inside. -When Kyle has to have his head cauterized because he has cut an artery in it and is losing a lot of blood fast ... well, Kyle doesn\'t smell too good when his flesh is burning.  At first I thought, \"Hmmm, someone is having bad Taco Bell.\"  and then I thought, \"Wait, that\'s Kyle cooking.  Someone needs to dump some salsa on him to cover up that smell.\" (Oh, dorks, it was so awesome!  The cool guy\'s in the E.R. let us keep the portable cauterizing iron!  I was going to take it home with it all covered in Kyle\'s blood and keep it forever, but I figured that he deserved it more.) -Kyle\'s heart rate jumps for cupcakes.   \"Hey, dude, you look pretty fucked up.  You O.K.\" Kyle: (Heart rate = 91) \"Yeah, I\'m not too bad.\" \"I brought you some cupcakes.\" Kyle: (Heart rate = 121) \"Really?  Awesome.\" beepbeepbeepbeepbeepbeepbeep So, anyways, Kyle was O.K. and pretty much laughing the whole time.  The worst was when he had to get stiches.  \"Oh, Heavens, please not stiches.\" Meanwhile: Some old lady in the room keeps saying \"Just let me live one more night.\"  Some pregnant lady next to us is talking to her doctor about getting brain surgery.  I\'m pretty sure a heroin addict died right next to us while some heroin addict down the hall is giving a lesson on the German/Phillipenes war, which I don\'t remember.  But Kyle can\'t see this (he stayed strapped to the strecher for 5 hours).  So, to him, his 20 or so stiches were the worst in the world.  Cupcake worthy, I\'d say. So, Kyle is fine.  He went home around 10 and was laughing about it.   \"Dude, look how covered in blood I am!\" You can read Kyle\'s take on it ( www.omglolh4x.com ... probably later today and possibly codeine induced) AND see the funny cartoon I drew about Kyle and his battery!  I\'m very proud of the cartoon ... and more personally I was scared last night, dorkos.  Very scary drive to the hospital not knowing if Kyle was O.K. for me.  I was glad he was O.K. and it was soulful to play the big brother role in another way than the jerk.  Other things since last blog: -Bought new shoes and a new jacket with my tax refund.  Adidas and North Face respectively.  Also, paid off some debts and, um ... shit, fuck, dorks.  What the hell did I do with all of my money?   -R. Di and I were approved to live at our apartment.  It\'s so nice to be approved.  We\'ve been planning our decorating patterns and color schemes.  We are very TLC.  I am very anxious to get this apartment thing rolling, dorks.   -I finally finished \"Franny and Zooey\", goddamn those lesbians whine a lot.  Seriously, it\'s 200 pages of ... \"Oh, but to be back yon in the day of childhood.  We could cast the off what we had wronged and were justly wronged by. The Jesus Prayer.\" \"No, Zooey, to cast off is wrong.  We, being child of superior intellect, know only how to cast on.  We must learn to cast sideways.  A sort of side toss of the soul.\" \"Mayhaps a knuckle ball, Franny.  A knuckle ball would cast back ...\" Jesus, I got a headache from reading that book.  No wonder people kill people after reading \"The Catcher in the Rye.\" \"Wow! Today is a beautiful day.  I think I\'ll head to the beach.  You know what?  I should really take a book with me. I really like that last book that I read.  That Holden guy was so honest and not a little bitch.  I wonder what else that Salinger guy has written.  Wow, I\'m already at the bookstore.  I sure walk fast and talk out loud.  Hmm, \"Franny and Zooey\", eh?  I think that I, Mark David Chapman, will read that one.\" Two days later. \"Fucking whiny little brats.  I mean seriously just eat the fucking soup and stop nagging.  Christ, I\'m so let down I could shoot a Beatle!\" And this has been \"Brent Distorts the Deaths of Beloved Cultural Figures\".  Join us next time when Brent takes on J.F.K. Jr. \"Hey, ladies, I can\'t remember if I got the plane that can fly under water or not.  My dad was a president, let\'s find out!\" I also read \"Fahrenheit 451\".  It ends with a nuclear war.  Biased, I like this one a lot more.  I have another book to read, \"How Soon is Never\" by Marc Spitz.  It\'s about a guy who tries to get The Smiths back together.  It sounded interesting in the library, but now I think I\'d rather just listen to a Smiths album. Well, dorks, it\'s been nice talking to you again.  I am tired and getting over a little sniffle thing today.  I hope all is well and that you have a limited number of bleeding family members.  Later, dorkpals. '),(3039,1785539,'06,August,2004','2004-08-06 00:00:00','   One more thing, dorks. A question I used to ask myself when I was a teenager and that I put before my friend Jared Stover one day.  Maybe it wasn\'t Stover, dammit, dorks, who cares who it was!  This question was asked ... \"Do you think that they\'ll ever run out of songs?  I mean, it\'s possible that there are only so many combinations of notes and chords and stuff. \" \"No.\" Maybe we have run into this block, maybe the combinations have all been found!  Maybe we just take old stuff and play it faster.  It\'s new!  It\'s new! -Dawn of the Dead -King Kong -Psycho -Puff Daddy -Starsky and Hutch -Walking Tall Remake, remake, remix and remake!  Now there\'s a movie idea, dorks!  A truly original one!  The world has rehashed everything so much that a man sells a copy of \"The Little Train That Could\" as an original work!  And everyone loves it! Monkeys typing . A movie about monkeys that take over our planet!  Original! '),(3040,1785539,'06,August,2004','2004-08-06 00:00:00','   So, close, dorks.  It\'s very close.  Can you feel it on the horizon?  I know you can\'t literally \"feel the horizon\", dorks, I know.  But, it\'s there, begging to be felt up.  Change is a slut, dorks, she wants you to take full advantage of her. When change blows through town she doesn\'t advertise $5 handjobs of herself, she begs you take full advantage of her services.   \"It\'s prom night and I\'m drunk on champagne.  Bomb city hall, tear down a historic tree, curse, fuck and show your tits.  Don\'t shoot Arabs, Islamics, Afghanis or non-rabid dogs. Love tits.  Tits made you healthy.  Black tits started your species, don\'t be ashamed of them.  Change, take me, knock it all down.  You want to, I want you.  You\'re drunk on champagne too.\"  Fuck. Boobs. Tits. Shit. Vote for Kerry. Fuck.   It\'s been a good week, dorks, it\'s time for a good weekend.  Sure, my girlfriend is sick and has cottage cheese in her throat and I can\'t kiss her for a week and that I have to drink beer alone.  Sure, my job is no closer to permanent and I still live out of a backpack between R. Girl\'s and the L.C. Sure, I have had major headaches all week from a snot fortification in my head. But, the slut is blowing into town, dorks.  Things are close to happening.  The precipice of the old, new shoes on, trail runners, perfect for whatever happens next.  Unknown next. New apartment, higher wages, new president, lower gas prices, electric cars and boobs galore! Good, bad, a meteor, a bomb, a Democrat or sunshine in Seattle. I\'ll take it all.  SCIENCE NEWS! The rubberhead planet has been found!  It\'s beyond Pluto!  Can you believe it?  I ask where these rubberheads come from and then from behind Pluto out jumps a fucking new planet?  Punch the world awake and the universe spits out a new planet!   \"It\'s a planetoid,\" says the scientist. \"Aw, fuck, scientist, don\'t rob me of this.  Just call it a planet.  It\'ll answer so many questions!\" \"But then we have to rewrite the books,\" say the beurocrats. \"Jobs, jobs, jobs!\" \"But then we have to name the planet,\" say the reporters. \"Rubberball!  Texas 2!\" (I just realized that the following doesn\'t make much sense unless I explain that I was thinking of the Sun also.  The solar system as a whole ((excluding moons)).  Sun, Mercury, Venus, Earth, Mars, Just, Set, Up, Nine, Planets.  10.) \"But ten was such an nice number,\" say the mathmen. \"So is 11!  It\'s two ones!  We\'re #1 #1!\"  Dorks, dorks, good weekend to you, dorks. '),(3041,1785539,'06,August,2004','2004-08-06 00:00:00','   Happy birthday to you, Happy first official and by official I mean first recognized by the media when in fact it\'s been an ongoing battle for more than a decade and not a first birthday to you, Happy birthday, War - in - I-R-AAAQ.   Happy birthday to you. Now, blow out the burning children and make a wish.  Weird times, dorks.  Things personally are getting better and better for me.  However, now that I don\'t have to focus on my own lot and now that I have your attention, I feel I should be punching the world awake.  Maybe it\'s just youth, maybe it\'s just pictures, maybe I\'ve caught my R. Matey\'s fever ...  She\'s doing better by the way, thanks for asking.  Well, she\'s either doing better, or she has mono. Either way, she\'s cute when she\'s sick.   A blind man with a seeing eye dog at his side walks into a grocery store. The man walks to the middle of the store, picks up the dog by the tail, and starts swinging the dog around in circles over his head. The store manager, who has seen all this, thinks this is quite strange. So, he decides to find out what\'s going on. The store manager approaches the blind man swinging the dog and says, \"Pardon me. May I help you with something.\" The blind man says, \"No thanks. I\'m just looking around.\"  Hahahahaha, fuck ... now, that\'s funny, dorks!  Happy St. Patty\'s Day, dorkles!  From all of your Scotch/Irish freinds at Dorks Don\'t Rock .. Get drunk, start a fight and have unfulfilling sloppy sex tonight! Blarney, dorks, blarney! '),(3042,1785539,'06,August,2004','2004-08-06 00:00:00','   Busy. Sick. Running around.   Drove sick girl to hospital last night, picked her up at 2 this morning.  Her kidneys went on strike. I hate hospitals, dorks, hate hospitals.  It\'s sad to see people stripped down to their blood and fear.   An e-mail I was going to write to Eva Moore today, but instead I\'ll write it here because I\'m killing birds with stones here, dorks. Eva, I went to the hospital last night.  (My girlfriend)\'s roommate was sick.  She got no kidneys.  Anyways, I saw something you might find interesting.  What, praytel, you ask? And I respond ... get the fuck out of here with that praytel shit. And then I say ... I saw you.  Well, I saw what you will be when you are 50.  When you are 50 you will cut your right pointer finger and wrap it up with a wad of either cheap gauze or good kitchen paper towels.  You won\'t be in pain, just a little worried.  Your mind won\'t be focused on the hospital.  Instead, you will be either thinking about a book you just finished or something your brother said to you. You will still wear glasses.  Your pajamas are hard to distinguish from your summer clothes.   You will get sick of waiting and realize you can patch yourself up.  You\'ll walk out of the hospital with a look that says \"My finger hurts and I\'m alone with this\" or \"I don\'t want to see this much of this part of the world for a long time\". If, when you\'re 50, you are thinking the later ... we\'ll have shared a thought 26 years apart. R. Fever is sick too, fever of 101.3.  Me help with ice water, lots of blankets and a cold press.  I\'m Dr. Frankenstein for fevers.  \"It\'s Aleve ... it\'s ALEVE!!\" All I want to do is go back to her apartment and crawl into bed, close my eyes and not hear anything.  Wrap my arm around my R., sleep, let my headache and snot drip away and dream about zombies, sweet blissful zombie hunting dreams.  Forehead kisses and zombies, dorks, that\'s all I need. Not much more to say except ... whew. Whew, dorks ... mutha punching, whew. '),(3043,1785539,'06,August,2004','2004-08-06 00:00:00','   \"Time and again, George Bush has failed to give those fighting the war on terror -- whether they\'re overseas or over here -- the weapons, equipment and support they need.\" Dorks, dorks, dorks.  Does this sound like a \"slam\"?  CNN claims \"Kerry slams Bush over war on terror.\"  I was thinking, \"Goddamn, a slam!  This is going to be good!\" \"That muthafucka ain\'t done shit for Kerry.  And if he ain\'t done shit for Kerry than he ain\'t done shit for none of you fuckers either.  That ass licker best start doing and stops lying before I go from straight chillin\' to straight killin\'.\" But instead we get the above quote.  See, dorks, it\'s all just a hyperbole to keep your mind fucked.  \"Kerry slams\" \"Nation mourns\" \"America attacks\" \"Terrorists threaten\" \"War mounts\" \"Security lowers\" \"Dorks Don\'t Rock\".  And now you\'re starting to see through the hype. PUBLIC ENEMY RULES! '),(3044,1785539,'06,August,2004','2004-08-06 00:00:00','   Morning, dorks.  Have a good weekend? Me too.  Me and R. Consumer went and did some shizopping.  Things we now mutually own together: -Plates -A Vacuum -Shower curtain -Bathmat -Welcome rug -A slip mutha covering cover and other stuff, etc.  It\'s weird to mutually own anything (besides a case of beer) with someone.  We are just a few weeks away from the big move in. Yip-mutha-readin-ee!  I am so ready to have a place to call home and not just a backpack full of clothes in different locations. What\'d I do this weekend.  Well, I was sick on Friday and so was R. Sniffles so that meant that we did jack shit.  I was so wound up from last week with it\'s sicknesses and exploding heads and other Kyle fueled anger that, in fact, I have no idea what we did on Friday.  I do remember that I went to Taco Bell and that I was dazed in my typical \"7 layer good, but 2 tacos for same price\" daze when I heard the Lovely couple saying my name behind me.  Apparently they had been standing in line behind me and I hadn\'t noticed. In turn, the female part of the L.C. didn\'t notice that I was right in front of her.  Guess what?  She\'s my cousin.  I guess there\'s some sort of \"Tacos have preponderance over people\" gene in my family. I like saying \"In turn\".  I\'m going to use it more often. \"In turn, I say to you, good sir of Wendy\'s that yes I shall indeed Biggie Size.\" \"So, I was wasted and, in turn, I took on SIX COPS!\" \"Dammit, I missed my turn ... in turn.\" Want another reason to laugh at Ashton Kutcher?  O.K., here you go. http://seattlepi.nwsource.com/lifestyle/164606_vondutch.html Isn\'t that the most ridiculous story about hats that you\'ve ever read?  It\'s like wearing an \"I Love Ghandi\" shirt to a fist fight, or a \"George Washington ruled\" shirt while you chopped down an apple tree, or making a very violent movie to spread the word (or your particularly warped view thereof) of Christ. Hey, who remembers Stussy?  Can you still buy a Stussy hat?  Or what about Yaga?  Or was it Yaya?  I was always afraid to ask because then I would ruin the carefully constructed image of hipness that I had created.  I had a kickass Stussy hat in high school, but then some dude said that if you read the signature all funny it said \"I am gay\".  So, of course, I couldn\'t wear it because my cock hadn\'t started it\'s growth spurt. That\'s right you started reading a simple paragraph about hats and high school memories and then all of a sudden (BA-DAM) you have to read about my junks.  I like to keep you on your toes and on my tip, dorks. So, on Saturday the R. Girl and me boy went and played disc golf with the L.C. That was fun and frustrating.  I couldn\'t seem to get my game together, so of course I spent more money on new discs.  They were having a tournament and all of these great players were out there, which ruined my self confidence faster than a Stussy hat. There was this group of deaf guys in front of us playing ( I now know how to say \"Knucklehead\" in sign language).  We were all slowed down because of the tourney ( I said \"tourney\", I miss the March fever that college life brings on) and frequently spent time together at tee pads waiting for other groups to throw.  At the 10th (and I know I might have that wrong and a part of the L.C. will catch it) hole, which has a huge hill just off the teepad that has a baseball field at the bottom, the deaf guy (sorry, guy who is deaf) with dreads and wearing the \"Weedies\" t-shirt throws a slice into the outfield.  So he walks away angry at himself, but fogets that his buddies haven\'t thrown yet.  But what can you do when a deaf stoned guy is walking in front of your disc\'s flight path?  (I think it was R. Girl that made the joke about stomping on the ground 4 times ... get it?)  So, that was funny for all the deaf and hearing people. I bought the original \"Dawn of the Dead\" on DVD this weekend.  It was awesome. Well, I really should work or something.  Later, dorks. '),(3045,1785539,'06,August,2004','2004-08-06 00:00:00','   Dorks, how\'s it going?  I\'m good.  Heading to the beach with R. Bubbles tomorrow.  We got us this here hotel room with a bay window that overlooks the ocean ... as you sit in a hot tub, mother fuckers!   But first more shopping at IKEA tonight. Want to hear something cool that happened to me? OK. How many of you dorks have watched \"The Sopranos\"?  OK, one, two ... um, 24.  OK, good, then you\'ll follow me. I was listening to the Sopranos soundtrack on the way to work this morning.  As I got closer to work the last track came on, which was a bunch of quotes from the show. \"Owee, I pinched my pinky.\" \"You really look good in green, Tony.\" \"Hey, who needs a hug and a cupcake?\" After these actual quotes the theme song for the show came on ... just as I was pulling into my office parking lot.  You woke up this morning Got yourself a gun, (Hell yeah, roll down my window) Mama always said you\'d be The Chosen One. (Pull my access card out) She said: You\'re one in a million You\'ve got to burn to shine, (That\'s right, can\'t no security gate stop me) But you were born under a bad sign, With a blue moon in your eyes. (Look at this fucker with the briefcase, what a mook) You woke up this morning All the love has gone, (That\'s right, mamagamook, keep walking.) Your Papa never told you About right and wrong. (Don\'t noone mess with T. I mean B.)  We\'ll fast forward a bit.  As I\'m parking my truck these lyrics come on.  When you woke up this morning, When you woke up this morning, When you woke up this morning, You got yourself a gun.  I parked my truck at the same point in the song that Tony parks his in the opening credits!  No fast forwarding, nothing, just perfectly in synch with my actions!  God loves a soundtrack!  Later, dorks, weekend! '),(3046,1785539,'06,August,2004','2004-08-06 00:00:00','   A conversation (circa. 2001). Kyle: (looking at meat grilling) Man, look at that.  That was alive once. Me: Don\'t be too sad for the cows, dude.  If it weren\'t for us, the cows would not be alive in today\'s world.  What purpose would a cow serve besides for food? Kyle: Then what purpose do we serve? Me: To eat the fucking cow!  Two signs I saw whilst on my lunch break. \"MOTORCYCLES USE EXTREME CAUTION\" (on a wide open road) \"Caution: CARS PARKING\" (In a PARKING garage. \"Is this where I park my squirrel?\") Manufacture paranoia.   Sell more Subarus, safety scissors and wars. '),(3047,1785539,'06,August,2004','2004-08-06 00:00:00','   I\'m eating peanuts (salted) out of the shell for breakfast.    Simple pleasures, like the ocean, help to keep us all in motion. Just like the waves below the gulls We all have crests, we all have lulls. But simple pleasures, like  peanuts, Christmas trees and girls\' butts, Home Alone, parts one and two and the deep Pacific blue will make you flip from a standstill, sure. \'Cause simple pleasures are the cure.  Yizzy, dorks, what\'s been going on in your respective lives then, eh?  Dammit, not all at once!  Respective, respective. I had a good weekend.   R. Throaty was feeling better but playing a little sick.   This meant I got to play the doctor and prescribe her some ... poetry time over, dorks. I played basketball with this kid on Sunday?  What kid?  I don\'t know, just some 11 year-old kid that wanted to play basketball at R. Yankee\'s softball practice (and by the way, she\'s good.  I saw her whipping them softballs around all better than the other girls and I got excited. I got me a star athlete for a girl, I\'m going to be the most popular kid at school!) Back to the kid, the Jordan to my Bird. At first, I let him win a game of HORSE. \"Aw, shucks, you got me.  I guess I didn\'t bring my A game today.\" Then he wanted to play one on one.  I just met this kid and from his shooting techniques during HORSE and the fact that he\'s 4 feet tall I was guessing that he couldn\'t be TOO much better of a basketball player than me. This,of course, meant I gladly welcomed a one-on-one game. And goddammit, the little duthafluffer started doing all of this Nike commercial around the back Harlem Globetrotter hug the line shit.  Dammit! And if I got too close to him he started screaming \"foul\". I was trying to teach him about the neccessity of the hook shot and he was throwing the basketball at my butt so he could drive the rebound home. And I couldn\'t contest his Harlem butttrotting antics.  Like I\'m going to get into a verbal argument with an 11 year old in a public park.   OK, so maybe I said a couple of times \"Bullshit, that was out!\" but other than that we played \"Dominic-ball\" \"You talked, your point doesn\'t count.\" \"I can throw the ball at your ass but you can\'t throw it at my feet, I hurt my leg.\" \"Mumble, jumble, Racky-too.  You can\'t understand me, because I\'m pre-pubescent.\" But then, the little punk hit me.   I had decided that he couldn\'t have all the rules the way he wanted them. \"No, you can\'t go walking around with the ball out-of-bounds going \'nay,nay,nay,nay\'.  I\'m drawing the line there.  It\'s my ball.\" \"It\'s my ball, check it to me.\" \"Hell, no, you have to have some limits.  My ball.\" So, the kid hits the ball, which I\'m holding, and it goes into my face and knocks my glasses sideways.  DAMMIT! I was Larry Fucking Hulk.   \"All right, your ball,\" I said. But the game was mine. I pulled no more of my punches. Literally.  The kid took a couple elbows to the side and got slammed one time when he drew a charge.  I ran the points back up in my favor (I had been letting him keep a 3 point lead).  I ran around him like he was a track and my name was Jackie. I won the game 21-20.  And, yes, as you can probably guess from the score, I wasn\'t a complete ass.  I let it come down to the wire so we could have a big finish.  I even gave him the ball a couple times and let him run to the hoop really slow while I did \"Sports Center\" highlights voice-overs. \"Dominic takes it up, this is for the win.\" But he couldn\'t complete the circuit and I got bored.  My win.  Well, technically my win, but we had to keep playing until he won.  My last goal didn\'t count because I had a hat on. And, yes, the kid had fun don\'t run off to call Social Services.  We have a playdate next Sunday for a little \"Around the World\" action.   Last week I read \"Slapstick\" by Vonnegut.  Good.  I\'m rereading \"Timequake\" by the same guy.  I wish Kurt Vonnegut put out a new novel every week. I also bought \"Children of the Corn\" It\'s now time for \"What Did Brent Say As the Credits Started to Roll on \'Children of the Corn\'\" \"What the fuck did I just watch?\" Join us next week when Brent watches \"The Cat in the Hat\". \"Ass lickers!  Ass lickers them all!\" Mike Myers is dead to me.  In my mind, Mike Myers got into a firey jet plane accident shortly after filming for the original Austin Powers ended.   Well, dorks, I really need to finish these peanuts. '),(3048,1785539,'06,August,2004','2004-08-06 00:00:00','   It\'s now time for \"Brent Explains the U.S. Foreign Theater by Pretending He\'s Iraq and Chicken Nuggets are Oil\"  Me: (sititng around in my apartment, eating chicken nuggets and watching T.V.) Man, I fucking love chicken nuggets.  No shit, these are awesome with honey or bbq sauce.  TV News Guy: Chicken nuggets make people fly. It\'s incredible but true.  Me: Shit, I better save some of these.  America: (busting down my door) Give us the chicken nuggets!  Me: No, I like them and they are mine.  America: (setting fire to my Adidas) There we burnt your shoes, now give us nuggets.  Me: What?!? No, now I want to give you nuggets even less.  America: You\'re insane! You kill babies! Give us the nuggets!  Me: Why are you here?  This is my apartment.  America: Are you threatening us?  Where do you hide your b.b. guns?!?!  Nuggets!  You are insane!  (Ten years later)  America: Can we have nuggets now?  Me: No.  Use your Texas nuggets.  TV News Guy: Dr. Phil has been murdered by not Brent.  America: Dammit!  You\'re insane and killed Dr. Phil.  Always remember Dr. Phil! give us nuggets! You will pay for Dr. Phil\'s death.  You own a flamethrower!  A nuclear weapon is your pillow!  (America breaks my plates and DVDs and sets fire to my couch ... and then shoots a puppy.)  Me: (Throwing an unbroken plate at America) Dammit!  Get out of my apartment without nuggets.  America: You attacked us!  You are insane!  You killed us!  Why did you not lay down and die?  Why do you fight back?  Why don\'t nuggets give us you?  We nuggets, you have, we kill, you no curl up and let us burn your puppies! NUGGETS! DR. PHIL! HIGHER NUGGET PRICES!  KILL YOU, DON\'T KILL US!  ... we will be back in Fallujah. It will be at the time and place of our choosing. We will hunt down the criminals. -Brig. Gen. Mark Kimmitt  \"Hey, how come you always step on my burnt flesh head after I kill your sister?\"  Stupid, dorks.  You\'re being lied to.  Punch the world and the elephants fall off their towers.  In happier news R. Pillowcase and I are about done with unpacking, electrical sockets and \"RARGRRRHH\".  After two long days of no sleep, Taco Time and up-down stairs we are ready to enjoy the fruits of our laboratory.   We (I\'m speaking for R. Silent here when really I\'m unsure of her feelings on this subject, but if I\'m wrong we\'ll just call it fiction) had the idea that moving in would be a magical process.  Much like Disney Land. Yes, dorks, exactly like Disney Land, no need to rethink that thought ... just run with it, Brent. \"Oh, look, Goofy and Donald have come by to help move the couch.\" \"Peter Pan did such a nice job setting up the shower curtain!\" \"Wow, Bugs Bunny, not only are you a Looney Tunes character and therefore an anachronism in this move-in fantasy, but you are also great at setting up a washer and dryer.\" And now I\'ve lost my train of thought and I\'m too wiped out to buy another ticket.   Dorks, dorks, dorks. '),(3049,1785539,'06,August,2004','2004-08-06 00:00:00','   Let\'s get de-lir-i-ous.  Nananana-na-nana.  Prince rhymed something with de-lir-i-ous.  Nananana-na-nana.   Future me was up very late last night lifting and opening. Color schemes and tidiness.  The fridge wasn\'t on, ice cubes to be made!  Too many DVDs, did I really need \"Firestarter\".  Oops, I forgot to clean out my coffee grinds from coffee pot ... 3 MONTHS ago.  Coffee is fur.  Girlfriend vomitous.  Lift a dresser, box spring.  No sleep yet.    I traveled to the future, my friends, on a machine fueled by Taco Time and redheaded smiles. In this future I have a toothbrush holder and a special cup for rinsing.  Ingenious.  In this future I hang my t-shirts on hangers.  I didn\'t know you could legally do that!   The future\'s so bright I gotta wear switchblades. Dorks, I want to crawl under my desk and sleep all day.  I could do it.  I could.   Dork-ir-i-ous. '),(3050,1785539,'06,August,2004','2004-08-06 00:00:00','   The prettier half of the Lovely Couple: No, she\'s not working with A.D.D. (ay-dee-dee) kids any more.  It\'s actually now \"English as a Second Language\" kids.  They still have some sort of learning disability. Me: Oh, so it\'s Aw-Day-Day kids, then.   That was a good one. Thank you, Lovely Couple, for a great and gas-heated warm three months. '),(3051,1785539,'06,August,2004','2004-08-06 00:00:00','   While coming back from the grocery store on my lunch break today I started thinking about a writing excercise we did in my sophomore-level English comp. class.  The purpose of the excercise was to write a few sentences that showed a progression of a thought. I think mine went something like: Kirby Puckett is a dedicated baseball player and a good man. He played until he went blind one day. He also was nice to kids in the community.  In retrospect (and because of Kirby\'s \"touchy feely\" charges) I think I should have instead written the following: Abraham Lincoln hated doing algebra. He was sure that Albert Einstein had invented it just to make him look dumb. So, Abraham focused on writing during his school days. Because of this we have The Gettysburg Address.  That seems a lot more interesting and creative.  Oh, dorks, I went to the beach.  Let me ask you, dorkwaves, have you ever laid on a full-size fold out couch, chugging wine while you stared out a window at the ocean a mere 200 yards away and your red hot girlfriend lounged sedated and bare in a hot tub only a few feet from you admiring the same view while R. Kelly\'s remix to Ignition played in surround sound throughout your villa including in your master bedroom where a King sized bed awaited you with comforters surely made from the hides of angels and baby rabbits?  No?  I have, fuckers, I HAVE!   This weekend ruled like no other. I flew a Ninja Turtle kite on the beach for hours on end. I bought a bumper sticker that says \"I love dinosaurs\". I ate buffalo jerky. I ate blueberry pancakes. I ate a pizza that had cashews and feta cheese as toppings. I slept in a bed for the first time since I was 3 where my feet didn\'t hangover the edge. And for two whole days I didn\'t talk to anyone except my lovely R. Tickles. You want to know how you know how you\'ve found someone special, dorkles?  You don\'t realize that you\'ve just spent 2 whole days without talking to someone besides them until you have to answer questions of \"How was it?\" bursting your bubble.  Bursting your bubble of waves and R. Kelly. And, lo, I say onto thee, the dorks of now and forever, remember this day for now and forever.  For this is the day that Bachelor Brent lays down in his cave of cigarette butts and Hustler back issues.  Yea, tomorrow and forever shall be the day when he known as Brent moves forth into his apartment with she known to all as R. Bunnycurls. The bachelor is dead. Long live the king!  He\'s checking his list, going to IKEA twice, tomorrow he moves in with his girl who is nice. Bachelor Brent is going away. He knows he must be tidy. He knows he can\'t stay drunk. He knows he has to wash his feet And not always be scratching his junk. Oooohhhhh, he now owns a rug He now owns plates. Who the fuck is this guy for Christ\'s sake? He\'s Future Brent and he\'s here to stay!  Dorks, I won\'t give you the sappy shit (\"she melted over me like putty\"? what the fuck?).  Just know I\'m smiling 147 % more often these days.    He\'s Fu-ture Brent and he loves you dooooorrrrrrks! '),(3052,1785539,'06,August,2004','2004-08-06 00:00:00','   Dorks, dorks, dizzydorkles. How\'ve ya been?  I hope that you all didn\'t blow up or get mono.  I did neither.  I be busy with my new apartment ... and my new XBox.   The apartment thing is going smoother than a fatty\'s trip to the store for some Twinkies.  And you know fatties make it easy because of the fat. The apartment is a perfect blend of Glade Plug-Ins and my Adidas, Pink and Knives, Utensils and B.B. Guns, Stuffed animals and Porn.  Many nights I stay up late to play my XBox and drink beer long after R. Slumber\'s has gone to rest and I think, \"I live with a girl and I haven\'t died yet.  Wicked.\" Plus, I\'ve convinced her to call me \"Best Rapper Alive\" (the new Jay-Z album is good), so that makes it awesomererererrr. I cook and I clean dorks.  I\'m getting matureereerrrer. Every morning I: -brush AND rinse. -put foot cream on feet -take calcium pills for weakening bones (especially my knees). -take antidepressant so I don\'t scream \"Lead pipes! Head blow!\" while at work. -take a Claritin b/c my body now hates grass, pollen, bees, peanut butter, milk, cheap beer, whores, Montana and cats. -make a lunch, which I actually do at night. -return my towels to where they belong, not in R. Bumpers car!   -put things in \"hampers\" I like it dorks, I\'m getting it down!   Hahahahahahahaha, fuck, Brent you are so funny.  It\'s easy to do what you do.  I bet one day you woke up and could write so goddamned good. NO! To play like the playa I be you gots to play lots like me. I found some of my old writings the other night while unpacking, dorktittyers.  The paper is fading and they might soon be lost, so I\'m going to post them here. And they will live for enternity in your hearts. Then you will realize the path of the genius. The following are journal entries I wrote during \"Career Math\" my senior year in high school (6 years ago now).  I had elected to take Career Math instead of Triangle Math because I wanted to slack and daydream about ninjas and Pringles (which will be evident here very, very soon, Dorklock Holmeseesssesss).   I fit in about as much as an eskimo fits in at a Subway. \"Mighty Quin like seal on sandwich.\" \"That\'s a very racist impersonation of an eskimo ordering lunch.  GET OUT! Lead pipes!\" \"Mighty Quin alone with hunger.\"  On the second week we learned to do addition.  What did we do the first two weeks?  Regression of horror keeps me from telling you.  I wrote most of these in 2 minutes (like my blog!). I won\'t change my typos, just for you.  Let\'s do it, dorks.   (Name three people you\'d like to meet and why.) Kirby Puckett - He\'s, in my opinion, one of the greatest baseball players ever.  But he\'s more than that he\'s a great human being, always nice to his fans and he held on to playing the game until he couldn\'t go any further. Teddy Roosevelt- He\'s the guy who said walk softly and carry a big stick, the way I think every man should live his life.  I\'d like to go on safari with him and shoot lions and Rhinoceroses and giraffes and bears. Beastie Boys (all of them) - It\'d be cool. We\'d break stuff and yell at people and other stuff.  I think they are great musicians I like that Sabotage son.  Bum-bum-bum-bum-It\'s a SABOTAGE!  (Philosophy on life) My philosophy on life is simple it comes in three easy statements #1 = never worry about anything that will not matter a week from now. #2=God is real #3= Do what you gotta do. It\'s that simple.  It\'s like one of them No Fear shirts.  They\'re pretty comfortable and they got some pretty witty statements like \"Beat Death Up when he comes for you.\"  I agree.  It\'s like that poem \"Don\'t go gently ...\"  It goes \"Blah, Blah, Blah, RAGE, RAGE against that thing Blah, Blah\"  It\'s a good poem, you gotta fight for your right. You know \'cause life\'s just like a big spinning wheel you gotta stay ahead of it or else it would crush you and you\'d go \"Ahhhh.\" Life is good.  (Something you wanted but couldn\'t have) I wanted a bear, a real live grizzly bear.  But my mom was like \"yah, yah, yah, it will eat you yah, yah, yah, blod and guts, yah, yah, yah, shut up\" so I felt bad and I went to the zoo and hopped in the bear pit.  Sure enough the bear was mad.  He tried to eat me.  So I ripped out his throat and showed it to him.  I was like \"Hi-Ya\" and his body went thunk.  The zoo people were like \"you killed him, we kill you\"  But my mom showed up and they wussed out.  Boy she was mad.  But I was like \"It was scary.\" So she gave me some Jello Pudding Pops.  I like them Jello Pudding pops.  I eat them up.  \"Slurp!\" That\'s what it\'s like.  I like bears a lot though they\'re pretty cool.  I wouldn\'t like to just fight with them.  (How to save electricity)  *Turn off lights *Be Amish *Live in the woods *Bonfires *Be homeless *Unplug dryer and shake dry *Take cold showers *Don\'t do anything *Blow up the electricity company *Rip out your neighbors electrical lines *Shoot out your lights *Get rid of your TV and hide in your neighbors bushes and watch thier TV  Just don\'t knock on their window and say \"Hey, turn it up.\"  They\'ll know you\'re there then.  (A good investment and a bad investment) I bought a tattoo that was cool.  The needle didn\'t feel nice and I think the guy who did it was into drugs.  But it was a good investment it was worth it.  I think I\'ll get lots of chicks b/c of it. A bad investment was when I got the salad bar and some cheez stix at a restaurant.  I was REALLY hungry and I had plenty of $ so I thought, \"Hey\" and I ordered lots but after like 3 bites I was Ll \"Uhh, I\'m stuffed.\"  That was stupid of me it cost a lot and I felt sick.  My tattoo is better than a salad.  Although, you can\'t eat my tattoo, unless you\'re a freak like on that movie \"Alive\" when they ate that ladies flesh.  They were just supposed to play soccer, not eat flesh!  (Something about the middle east and what you want to do with laws.  Not really sure, they start to get whackier.  And please remember that I wasn\'t this dumb, just this bored.) I think we should blow up all of the Middle East, even the countries we like.  Those sand people always get you in the end.  Sadam Hussein would have been a really bad leader if he was VOTED in.  I think our guns are pretty good now-a-days.  WE can shoot pretty far and do LOTS of damage.  Iraq is supposed to be really big on American TV shows like \"Mama\'s Family\" or so they say.  Those big ships we keep sending over there are pretty big.  I bet they\'d be pretty scary to see coming across the water.  If I were the president I would make it illegal to NOT turn right on red.  I always get stuck behind some doofus. I think it would be nice if it were legal to knock down people on bikes.  (People that bother you) My brother Mrs. (teacher in high school) Communists Chipmunks, especially when they make that noise. My boss Freakin Santa always wantin\' me to be nice. Drunk girls at parties Mexicans who stand in the road. Big mecanics who think you should pay $95 for new brakes like it\'s their car and then they say that they can\'t let you go. Drunk girls at McDonalds People who ring bells and ask for $ Old women who think you\'re cute Drunk girls who go trick-or-treating at 3 in the morning. Drunk girls that puke when you\'re trying to make a move.  (Opinion on prayer in school) I was thinking about prayer and I decided it would be best if kids could say a little prayer, it\'s like Madonna when she said \"poppa, don\'t preach, I\'m in love again.\"  It\'s more like \"teacher don\'t preach, these kids can\'t handle it.\" God is everywhere and everyone can talk to him so why can\'t you be in Geometry and be like \"God you\'re cool.\"  Because sometimes we all need to talk to God. especially at school.  Yep, God knows what we\'re going thru so I think the teachers need to get rectified (?).  We need a new Crusades where we go thru the schools and chop people\'s heads off if they don\'t let you pray (Tell me, dorks.  Was this written pre-Columbine?  Today\'s kids would be put in sand people jail for writing that!) We\'ll be like \"Pray\" and if they say \"No\" We\'ll go \"Arrr\" and hit them with an axe.  (Scariest movie?  What do you think of Halloween?) Scariest movie is the original \"Night of the Living Dead\".  Oh boy, I almost wet my pants every time I see that one.  They try to burn them zombies but they can\'t do it.  They all die but these two people live but they get shot.  This little girl eats HER OWN MOTHER.  Ahhhh!  My god, my mother would take my driving privelages away if I even thought about eating her (no snickers, you sick fuck dorks).  But this girl did it (hehe).  She went \"Ahh-Chomp-Chomp\".  And then she got shot.  I once saw this kid walking around on Haloween with a shovel sticking out of his head.  I thought it was a pretty scary costume.  I got scared.  The kid kept saying, \"For the love of God, it\'s real, call the paramedics.\"  Sure enought it was for real and he died, but he was very scary.  (Current event or space bugs.) Starship Troopers is a very current even.  It\'s the biggest movie about killer bugs since \"Schindler\'s List\".  I like it.  It\'s making a lot of $ I hear they are going to take all the $ it\'s making and give it to the poor people for welfare.  That will solve the problem.  What if those space bugs were real.  I bet I would hide if I saw one.  But if there was a girl around I\'d stop hiding and kill some space bugs.  BOOM-BOOM that would be my gun.  Those space bugs would be like \"AHHHHH\" and I\'d go \"Ha-ha-ha\" or \"Ho-ho-ho\" if I felt like all Christmasey.  I like Christmas.  But I like Starship Troopers better.  I\'ve seen it twice.  I want to see it again.  I only saw Schindler\'s List once.  It was good.  It was about Nazis.  (Something you want to change about yourself) I would like to change my height.  I wish I was 20 ft. tall and all the people would say \"Hey tall guy\" and if I didn\'t like them I\'d go \"Arrrr\" and step on their heads.  So there.  I want to keep my head though, it is good to have around.  I once saw a guy get his head taken off and he wasn\'t happy.  It was on a Popeye cartoon so he didn\'t die, but he was sad.  Popey was cool.  He like spinach.  Have you ever tried spinach on your pizza (teacher response: nope) it is good.  (New Year\'s resolution) My New Year\'s resolution can best be summed up in a poem by Shel Silverstein (a.k.a. Best Rapper Alive).  I came upon a box and opened it up wide and gasped to see a dead body dead inside. There was a note attached that he\'d decided to leave that only said... \"I forgot to breathe\". That\'s all .. I resolve to remember too breathe.  (Million dollar house or million dollar car?) I\'d want a car I\'d put all sorts of things on it like guns and wheels and paint.  It\'d be like r-r-r-r-r-Purring like a kitten.  All the girls would be like \"Hey, big Brent boy\" and I\'d be like \"Hey, baby\" than their boyfriends would be like \"Hey you, go away\"  than I\'d pull out my rocket launcher and be all KA-BOOM and they\'d go \"splat\" so than I\'d tell their woman to get in and she\'d love me because my car is really cool.  I\'d be like Burt Reynolds on Smokey and the Bandit.  \"This is the Bandit,\" I\'d say.  \"No, it\'s Brent,\" they\'d say.  Then I\'d run them over with my car.  I\'d be mean, real mean and I could live in my car.  I don\'t need a house my car is cool. Vroom-Vroom all aboard the fast car machine!  (Interesting?) Something that I find interesting is Madonna.  She\'s the material girl and she don\'t want her poppa to preach.  I think she\'s cool.  She\'s all like \"If I want to wear cones on my breasts and get freaky then that\'s my business.\"  She had that book named Sex.  Taht\'s a good name for a book.  If I were to write a book it not be about sex though it would be about grizzly bears.  I bet Madonna wouldn\'t think twice about shooting a grizzly bear because she\'s a superstar and probably a jerk.  (Write about mean, median and mode) (Nope.)  That\'s about it for my career math writing career.  You see, dorks, it takes a lot of time and talk about grizzlies to be a world class writer like I am.  dorksdontrock is what it is today thanks to my time in Coach Head Guy\'s class.  I\'ll never forget you, Coach Head Guy With Glasses. And I\'ll never forget you dorks.  You are all so cute. (Psssst, I found my creative writing stuff too..... shhhhhhh). Weekend, dorks. '),(3053,1785539,'06,August,2004','2004-08-06 00:00:00','   I\'m sorry, dorks.  I get into my rappin\' moods and I forget to tell you about what else I\'m up to. So, today I started carpooling with this chick who works at the Department of Ecology.  She is definitely not the kind of person who would like being called \"this chick who works at the Department of too much typing\". She was talking about carpooling and some van and her foster kids and trees and buildings and I wasn\'t listening and driving and not listening and trying to avoid bringing up lesbians, which I don\'t think I normally would have except I realized about halfway to work \"OH, she\'s a lesbian\" (maybe it was because she was going on about saving the \"bushes\" Ha, fuck). So, that was good and I like getting to work in 25 minutes and not an hour and 15 apples ... minutes, not apples, Brent.  You eat apples.  Anyways, my car smells like hippies. Man, I used to hate hippies in high school.  I think it was because they sat on the ground.  To this day I refuse to sit on the ground if I can avoid it. Man did not evolve millions of apples (dammit, years) to grow legs and hands and invent screws and wood to build chairs for hippies to sit on the ground!  \"It\'s a sit-in!\" \"Dammit, no it\'s a sit-on!\"  I should do a remake of that movie \"Hair\" and instead call it \"Chair\".  That\'s a really dumb idea.  In it I will teach hippies about chairs and show them how to use them. \"This is the dawning of the age of the Lay-Z-Boy ... the age of the Lay-Z-Boy ... THE LAY-Z-BOY!\" Yep. Just a few minutes ago I was getting on the elevator to go up (\'cause it ain\'t time to get down yet, a-huh!) and this lady who I haven\'t told you dorks about did something stupid.  I haven\'t told you about her mainly because I don\'t want her soul fouling up my precious dorksite.  It\'s like the Amish and their fear of cameras; I feel if I utter her existence here she will capture my soul!  Her name is Wig Wolf!  She is the darkness of Hell!  I hate Wig Wolf!  Sometimes I have nightmares that she is at my apartment very late at night to bother me.  I wake up in a cold sweat screaming, \"No Wig Wolf!  Rainbows on your shirt don\'t cover the Hell in your belly!  Your belly!\" And then R. Sleepy rubs my back. Witness the fire! (This time the e-mails are in order. I love you.)  [this is kyle, brent fell in a hole. he asked me to type this (\"kyle, i\'m in a hole. go update my site\"). These email correspondence are REAL. like Fo\' REAL. okay. i\'m done]  From: 	Wig Wolf To: 	Kinkade, Brent Subject: 	Sorter Importance: 	High  Brent, The sorter you dropped off, is made out of pressed cardboard.  Can you please return this one and get me a plastic one.  This one is not cost efficient because it is going to fall apart, especially the lettering.  The lettering is taped on to the ends.  I got approval for a sorter and I would prefer to get the sorter that I know will last me a long time, compared to one that is going to fall apart in no time.  Please order me a plastic sorter.  Please let me know if you need me to look at the Boise catalog or if you need to see (a woman)\'s sorter again. Thank you, Wig Wolf  ---------- From: 	Kinkade, Brent To: 	Wig Wolf Subject: 	RE: Sorter  Wig Wolf, I called around for you.  (Girl #2) has a plastic sorter that she says you can use when you need one.  She doesn\'t use it often. If you want to hold on to the one you have for future use you may or I will return it. In the future, please ask the different sales supports if they have supplies they can share or let you have. This is more cost and time efficient. Thank you, Brent ----------  From: 	Wig Wolf To: 	Kinkade, Brent Subject: 	RE: Sorter  Brent, (Woman #1) checked with (man) and he didn\'t have any. (Man) said that (Girl #2) had borrowed his sorter.  We need more than one sorter flooding around.  If (Man) should need it, I would not have one to use.  If (Girl #2) should need it, again I would not have one to use.    When I need to sort, I need to sort, it is not something that I can put off for very long and that is why I felt the need to ask you to order one for me.  You can check with (a woman), as to how many times we need the sorter at the same time.  There are times that (women) and I need the sorter at the same time.  It is not unreasonable to ask for a tool that we need to do our job and it is not unreasonable to get the quality that we ask for.    Brent, don\'t get me wrong, I believe that you checked, but I know that there are better qualities of sorter and I would like for you to order me one.  Please check for me, and please let me know if I need to escalate this request. Thank you, Wig Wolf  -------  Wig Wolf, I don\'t believe this issue warrants this much time or communication.  The problem is easily resolved.  Bring the old sorter down and I will return it, the new sorter will be here tomorrow.   I was looking for the most \"cost effective\" way to solve the problem as that seemed to be your original concern.   Thanks, Brent   This is Wig Wolf!  This is a taste of her viciousness.  She is vile, all hate Wig Wolf! Back to my original story.  Me: (getting on the elevator, which Wig Wolf is on by herself) Hey, hold the elevator.  Wig Wolf: I\'m going down.  Me: (noticing that the up light is lit, door closing) But, the light says your going ... (door closed)  I hear the elevator go UP! UP NOT DOWN!  WIG WOLF!  I wait for the elevator again and go up.  While I\'m upstairs I see Wig Wolf!   \"Have a nice weekend, Brent,\" she snarled. \"I will have nightmares about you for life,\" I responded. \"My soul is pure fire,\" said Wig Wolf \"Yes, and God shuns ye,\" I said. \"That\'s me!\"  I think I had something else to say, but ah well it\'s time to go pick up Janis Joplin and drink Wig Wolf away. Real weekend this time, dorks. '),(3054,1785539,'06,August,2004','2004-08-06 00:00:00','   I got ladies fight-ing over me,  Got the girls delight-ing under me.  The Kung-Fu way don\'t enlight-en me,  but, punk, you do not fright-en me.  \'Cause I pull out my shotty. \"Barrel, please meet Face.\" KRA-KOW too loud, I\'ll just stick with the Mase. Beause with Mase you only get a scream. And in space don\'t no-one hear you scream. Welcome to the void; call me Buzz Lightyear. Fuck kiddy shit, make it Doctor Fear. Sinister, like I\'m Spiderman\'s foe And Spiderman won\'t be \'round no mo\'.  I said Spiderman won\'t be \'round no mo\'.  \'Cause I got M.J. and Lois fight-ing over me. And Supergirl be delight-ing under me. That Kung-Fu shit don\'t enlight-en me, but, punk you still ain\'t shit to me.   Damn, dorks. I AM the best rapper alive!  I rock, I rap, I\'m not Kid Rock and Fred Durst is for super skinhead dorks!   \"I did it all for the nookie!\" \"You\'ve actually had sex, Fred Durst?\" Check out my album in stores June 6 \"DJ, get a Clue\". Throw the sign, dorks! Weekend! '),(3055,1785539,'06,August,2004','2004-08-06 00:00:00','   Hahahahahaha. I was just laughing to myself and I thought I\'d share the laugh with you, dorks. Here\'s a non-authentic transcript of a prank call I made my juniorish year in high school.  Operator: Thanks for calling Super Speed Readers, would you like me to set you up with our program?  Me: No, actually I\'ve used it already.  I bought it about a month ago.  I can read a lot faster now.  Operator: Well, I\'m glad you like it.  Me: Well, I like it, but I don\'t like what it\'s done to me.  Operator: What do you mean?  Me: I read really fast now, but maybe almost too fast.  I read so fast that some times I start puking.  Like motion sickness kind of.  Operator: You vomit?  Me:  Yeah, is there like a customer service person I can talk to?  I don\'t really want a refund, I just want to know if this is normal or if I need to go see someone.  Operator: You\'re saying you throw up because you read too fast?  Me: Yep, I just start reading and my eyes are moving so fast and I think it\'s faster than my brain can move so I think it pukes, like to maybe cover up the book so I can\'t read the words.  Operator: Please hold on while I get my supervisor.  Me: Nope. -Click-  Turned away from it all Like a blind man ... Bowie rules so hard!  Opens with \"Rebel, Rebel\" and 2 and half hours later closes with \"Ziggy\".   Me tired today, sore from sitting, didn\'t take dancing cues from limber balding gay men in front of me.  Full head hair straight boy stiff, bald gay men limber.   And the world turns on towards eternity. Outside they were selling t-shirts with a picture of 70s Bowie.  The shirt said \"I Fucked Mick Jagger\". I wanted one, but my stiff shoulders and full head of hair held me back. DORKWAD PLAYED GUIIIIIIIIIITAAAAAAAARRRRR! '),(3056,1785539,'06,August,2004','2004-08-06 00:00:00','   WASHINGTON (Reuters) -- Heavy social drinkers show the same pattern of brain damage as hospitalized alcoholics -- enough to impair day-to-day functioning, U.S. researchers said Wednesday ...  This is why I do most of my drinking under a blanket in a dark closet instead of socially.  Speaking of society ... is President Bush getting more darling looking every day?  The picture of him and Ariel Sharon on cnn.com makes him look so precious.  Kind of like a lost puppy/boy.  And his press conference?  Forget about it!  He is by far the most progressively-boyish-by-the-day president ever. \"Sir, I knew Jack Kennedy, he was a boyish-faced president, you sir are no Jack Kennedy... you\'re a pumpkin prince with chubby cheeks of gold and a tightlipped smile of pudding!\" Almost makes you forget he\'s killing kids in other countries and dumbing up kids at home. BOWIE!  SPREE!   (A fratboy infiltrates!) '),(3057,1785539,'06,August,2004','2004-08-06 00:00:00','   Busy life for me a busy life indeed when I move so fast no commas come and all day long the work is done and periods are nay nay nay all I feel I have to say is busy dorks with bad grammar and happy days can\'t be too far  What a country! -I found a disc golf park 1.5 miles from our new apartment!  It is so beautiful that I want to deflower it.  I could walk there if I wasn\'t a recovering BigMac junkie or if I liked to walk!   I decided the other day when I was on a ride out to R. Pickles parent\'s (the cucumbers) for a bbq that I really don\'t like walks.  I\'d rather stay in a relatively smaller area and jump around preferably with a ball.  I have to have a purpose to my leg motion. \"Let\'s go for a walk.\" \"Nope, let\'s do something useful.\" -I beat MaxPayne 2 last night.  I like running my proper nouns together.  BrentKinkade, MaxPayne, BigMac.  Very trendy looking. -For the last 2 and a half months I have taken part in a study to treat people with major depression. \"When a lack of headaches is making your sex life grand and you hate it , call MajorDepression.  When you\'ve been laid off and that pizza dough tosser boy job offer is looking good, call MajorDepression.  MajorDepression, he makes you say \'Maybe I won\'t have a burger\'.\" So, there\'s a 50% chance that I\'ve been taking a new drug that they are trying out, a 25% chance that I was taking, um, I can\'t remember it\'s name ... Luftextarasmiles and a 25% (50+25+25=100) chance that I was taking sugar or not sugar.  I won\'t know for a couple of years until the study is over. But anyways I feel much better and yes I will have that burger. Tomorrow I start Prozac.   Also, tomorrow I go see David Bowie and Polyphonic Spree.  Personally. \"So, David Bowie, let\'s see you do Ziggy.\" \"I\'m so hipper than you even though I\'m so older.  I will smile at you, but after 5 words I speak gibberish.  Flick flock puddy spoo.\" \"What?\" \"Tipple crackers.\" What was that about?  So ... the girl person of the lovely couple (who really needs a much shorter name so now she\'s Jill, which she really isn\'t) Jill\'s dad got sick and she had to go visit him in Atlanta.  So, I get to go to see Bowie in her place with Jack (who played frisbee golf with me on Sunday and who I actually outdrove and, i said it (or am about to (parantheses)), outplayed on a few holes). So, Bowie tomorrow night. Tonight? Kill Bill Vol. 1, dorks.  And Costco hot dogs, dorks. And redheaded kisses, dorks. And leftover girly Smirnoff drinks, dorks. But now? It\'s time for \"More Shit Brent Found\". Here is one more of my writing examples that I found when I moved.  This is from creative writing and a bit more structured and punctured. It\'s a \"Finish the story after Big Teacher Worked at Pizza Place finishes prompt\". The set up was this: It\'s Halloween (this was written in the same year, and more than likely on the same day, of the Career Math Halloween entry ... \"shovel in head\") and this girl is driving home from a party.  She is all alone and she stops on a bridge.  The bridge is supposed to be haunted.  She does not like this situation.  She starts to strip.  She is super super stacked.  She whispers, \"Brent, show me the triumph.\"  I whisper, \"Does that mean  ....  Back to reality.  The night is dark and the bridge is dark and she reaches in her back seat for her flashlight when all of a sudden she feels a man\'s head.  He he. (my story starts now) THE TRAGIC ODYSSEY OF A HEAD ON HALLOWEEN \"Hmmm there shouldn\'t be a head (he he) in my back seat (he he).\"  Carolyn thought. She was right.  Her car did come with many options; air conditioning, AM/FM radio, cruise control, but most deffinitely not leathery, bald heads. \"Gasp,\" she gasped.  (\"Brent, good writers don\'t do this.\" \"Yeah, well, good writers don\'t pee on your face.\") Suddenly the head spoke.  Its voice was a raspy voice one that sounded like sandpaper against concrete. \"Gack dat ackgah,\" said the head. \"Here use some of this,\" said Carolyn pulling some throat spray from her bag. \"Dack mak ugk,\" gagged the head. (I should have written \"dackmakugked the head\"). \"Oh yeah, no arms, just a head,\" observed Carolyn. \"Here let me bring you up front so I can spray it in your throat.\" So Carolyn picked the head up and plopped it on the passenger\'s side seat. \"Say \'ahhh\',\" said Carolyn. \"Aggkahk,\" said the head. She sprayed the head\'s throat.  He smiled. Carolyn liked his smile.  \"Mature,\" she thought. \"Hey thanks, babe,\" said the head. \"Did you call me babe?\" asked Carolyn, astonished. \"Sure enough, sweet thing,\" replied the head. Carolyn was shocked, she couldn\'t believe it.  She couldn\'t get a man even if he was just a balding decapitated head.  So she leaned over and opened the passenger side door and pushed the head out. \"All men are scum, even dead ones,\" she said. Now the head had been through a lot in his day but it doesn\'t matter if you were made part of the undead yesterday or last century, it still hurts to be shoved out of a Rambler going 60 miles per hour (which was stopped before, Brent). However, my future self is a picky ass, and as the head hit the pavement it bounced 15 ft. in the air!  This caused the head to fall over the bridge and land safely in the river below. \"Thank goodness all that gel I use solidified into a hard rubbery substance on my scalping,\" said the head, \"I may be bald, but I still know that gel is important.\" Well, it doesn\'t take the decapitated head of a rocket scientist to realize that heads sink in the water.  The head had to think fast.  Suddenly he remembered. (Here\'s a good question/lesson for aspiring young writers: when you realize that you have created an impossible situation for your character that doesn\'t even allow for much of a story do you  a.)rethink your entire piece b.) brain storm with fellow writers.  Your peers are your best resource. c.) come up with this ridiculous idea) \"I\'ll use my outboard motor,\" he exclaimed. He\'d almost forgotten about his signing bonus from the Dark Lord from when he joined the ranks of the undead.  So he yanked on the cord (how) with his teeth and the engine started instantly. \"Purring like a kitten,\" said the head. Well, the head had a grand time skimming across the surface of that lake.  He only wished he had some skis he could strap on (he he) to his chin. But once again there was trouble.  How does a decapitated head reach the choke button and thus cut off the engine? (?)  He doesn\'t (well played, past me). So the head had no choice but to try and jump the shore and land in some leaves.  Well, he cleared the shore all right, but he didn\'t land in leaves.  He landed in a bag of Halloween candy of a tot passing by. \"Hey, Ma, somethin\' done landed in my bag,\" said the little inbred. \"Well, what it be?\" asked the mother. Gazing in the bag the child replied \"A head.\" \"Well, I be ... is it one of them heads of the undead that have been going around?\" asked the mother. \"Yes,\" the child replied, \"and he has an outboard motor attached to him.\" \"Hooey,\" hooed the mother.  \"Mug him and take the motor.\" \"Sure, Ma,\" said the boy. So the little child proceded to beat the head with a stick. \"Hey, I\'m just a head,\" said the head. \"You could just take the motor.\" \"Shut up, devil head,\" said the boy, delivering a rather vicious blow. The last thing the head thought before he passed out was ... H A P P Y     H A L L O W E E N  Well, dorks, it\'s time to go do things and stuff.  Remember when things aren\'t working out as planned have someone get beaten with a stick.  It will draw attention from your problems and quite possibly land you a motor. '),(3058,1785539,'06,August,2004','2004-08-06 00:00:00','   The Ranger Creed  RECOGNIZING that I volunteered as a Ranger, disregarding video games, television, political propoganda, movies, the education system, the economy, my father, sports, chicks, beer and advertisments as influences on my naturally influencable human mind.   ACKNOWLEDGING that a Ranger is a more elite soldier who honorably gets to shoot twice as many dark skinned people in hot parts of the world.  NEVER shall I fail my comrades or ask my Ranger buddy, Pippytits, things like \"Hey, Pippytits, why did we just shoot that unarmed man?\" or \"Pippytits, do dark skinned people from this hot part of the world have a Heaven.\"  GALLANTLY I will show the world that I am a specially selected and well-trained soldier by wearing lots of tight t-shirts and Croakies on my Oakleys.  ENERGETICALLY I will meet the enemies of my country.  I will then invite them to my tent for soda and chips.  After a close game of Monopoly I will kill them.  I will always be the Scotty Dog piece, I will never be the Wheelbarrow.  READILY I will display the intestinal fortitude to say \"intestinal fortitude\" without laughing about poopy.  Laugh it out! Listen to me dorks!  Listen to the hype!   Pat Tillman will become a rallying point for the war hawks.  They will say things like \"Be like Pat\" and \"Pat was a good American\".  They will make Pat speak from the grave!  Pat will be the 2004 electoral zombie!  He is the wet dream for a war marketer. When you see Pat\'s face, dorks, and when you hear people talk about him remember this ... Pat Tillman died in Afghanistan so you could drive to the beach at lower gas prices, oil companies could punish their enemies and tycoons could have a hand in the world\'s biggest money maker called the U.S. Government.   Anybody who uses the dead as a puppet is an evil ventriloquist.  It\'s supposed to be nice this weekend in Seattle.  I hope it\'s nice where you are, dork reader.   Weekend. '),(3059,1785539,'06,August,2004','2004-08-06 00:00:00','   Productivity and desire to not work here fuel today. I may be up for a wicked new job.  It\'s another contract job, 6 months for sure, maybe 9 or a year.  Much more money, lots of experience and potential for something wicked afterwards.  No insurance though.  But, fuck it. When I was 10 years old I preformed surgery on my foot because some gross thing kept growing underneath it.  It was like yellow and bubbly. \"Don\'t pick at it, Brent\" \"Eat it, Mom, I\'m digging the fucker out!\" Yep, I sure as a fatty likes vegetables said that.  Then as sure as a fatty likes fresh pie I got a pair of scissors and a big piece of chewing gum.  If you bite down on Bazooka hard enough you don\'t need morphine. So, I dug out whatever was bubbly and sticky and I was fine. My point?  I don\'t need insurance like a fatty needs ranch dressing.  I have some scissors and bubble gum.  Cancer?  Fuck it.  You can only root deep in your body with scissors until you pass out or dig the fucker out.  Either way works for me. So, me and Wig Wolf and Man Face and Crumble Cheeks and Jersey Hair and Stupid Mustache and Smile Guy and Mole Teeth might be parting ways here soon.  Or maybe not, who knows, it\'s a crazy world out there, dorks. I got my truck registered in Washington on my lunch break.  I wish I had put some thought into a kick ass vanity plate like ... BLLKIKER STPDHIPY ILUVMACE GOFATTY URAFATS DRKSDTRK MMMCAKE or KITNKLR Those would all be good. So, thanks to a Pentagon mandate the media are not allowed to put out pictures of dead soldier\'s coffins coming home? What the fuck do you think goes on in war, dickfaces? \"I say, I\'ve been shot, Timothy.\" \"My, Pippytits, call a time out and I\'ll go get Mom.\"  In the war young Johnny did die then he came home so Mommy could cry. He died from a headshot near Kabul. They showed his headshot from high school, but not the coffin where he rests. \'Cause we don\'t want voters to get upset.   And poets die sooner than water skiiers?  Good thing I\'m a bad poet. Later, dorks. '),(3060,1785539,'06,August,2004','2004-08-06 00:00:00','   Morning, dorks.  Gotta a lot to say, so let\'s go! Went to the baseball game last night.  Mariners won it 2-1 in the ninth.  Made for a fun little date with R. Third Base. Dorks, are there times in YOUR lives when you go \"Hey, what the hell?\"  or \"What the hell, hey?\"  or \"Was that the Maytag Repairman that I just met?\", because in my \"cartoonish\" reviewed life I say that often.  Well, with the Maytag Repairman it was just last night that I said that. Yep, I met the Maytag Repairman.  Isn\'t that a random, random person to meet at a baseball game.  I know I felt up Mrs. Butterworth at a coke party once and played Tekken 3 against the Pillsbury Doughboy at a local arcade, but the Maytag Repairman?  He\'s like the living god of the unliving corporate mascots!  His ability to actually use paper towels makes him far more dietiesh than the Bounty man! So, after shaking his amazingly silky soft hand (\'cause he don\'t work, dorkitches!) I told him that I did as little work as him all day (and I do! I get paid to do this!). So he signed my glossy picture of him and his young assistant: \"To Brent, I know how you feel!  Depend on me!  From Ol\' Trusty, FARdlkjBarthly (Can\'t read his signature)\" His strikingly handsome counterpart (hey, I can admit it when a dude is Greekesque in his handsomeness.  I won\'t say guys like Harrison Ford are handsome, that\'s a little much.  But guys like the younger Maytag Repairman?  You have to respect their handsomeness ... or you\'re gay.) signed it: \"To Brent, Come join our team and do nothing. STAY CLEAN! Franck;jj OUUUllllw3 (again with the signature)\"  \"STAY CLEAN!\" it\'s like he knows about how I have to clean my feet and wear deodorant and wipe up spaghetti stains to maintain impress over R. Girly!  He feels my pain!  He really knows me for who I am.  I\'m messy, I like teams, he knows the real me!  Maybe I should call him.  No, too soon.  Maybe ... SNAP out of it! So the picture hangs in my office.  Read it, rubberheads and not understand.  To not understand is a rubberheads greatest natural resource. A treat from the creative writing vaults! For this exercise we had to take a horrible play called \"The Open Window\" and write a screenplay for it.  Fun!  A screenplay!  Awesome!  For \"The Open Window\"?  Lead pipes!  Lead pipes! Here\'s the premise of the play.  This nervousa guy goes to visit this family.  This little bitch tells him about her dead dad and brother and how her mom keeps the window open so their souls can come in and out and drink coacoa.  Then the \"ghosts\" walk through the door and it turns out the little brat was lying that they were dead and the guy becomes a Highlander. Get it?  Good.  Here\'s my screenplay for it.  THE OPEN WINDOW Screenplay by Brent Kinkade (1998) Adapted from A. Dumwriter (from Hell)  Cast Mr. Nutel-Crazy guy with a sleep disorder Vera(young lady)- Niece w/ an attitude Mrs. Sappleton - Not crazy, but thought to be Mr. Sappleton - Not dead Dog- Does his thing doggy style (I actually wrote that in high school and got a 20/20 for it!  Gold star!)  (Victorian townhouse living room with a large open window (!) that reaches the ground on the wall.  There are about 5 or 6 trophy heads decorating the walls and the seats are large and leather)  (Camera pan out of Vera\'s face while she\'s speaking) Vera- My aunt will be down presently, Mr. Nuttel, in the meantime you must try and put up w/ me.  (Camera continues to pan out until you see her entire body, she is thin and frail)  Vera- (Motioning for Mr. Nuttel to sit) Do you know many people round here?  Frampton- (sitting cautiously after wiping at the seat) Hardly a soul, my sister was staying ...  Vera- Yeah, yeah then you practically know nothing about my aunt.  Frampton- (shifting in his seat, perturbed by being interupted) only her ...  Vera- (interupting again) Her great tragedy happened three years ago.  Frampton- (Opens mouth to ask question, but...)  Vera- You may wonder why we keep that window wide open on an October afternoon  Frampton- It is ... (any of you dorks notice if the high school writer is getting bored with his material yet?  You will!)  Vera- (interupting) Three years ago boys leave.  Bog suck them down.  Many moons my aunt has kept window wide until the great fireball has set.   Oh, her ailment can be cured by no mediceman.  She believe boys come home soon.  Crazy white woman.  I\'ll scalp her.  Frampton-(looking confused) Why are you talking like an Indian?  (Enter Mrs. Sappleton) Mrs. Sappleton-I hope Vera has been amusing you.  Frampton-She has been...  Mrs. S. (Interupting- I hope you don\'t mind the open window but we are playing a cruel joke on you because we think you\'re ugly and smell bad.  Frampton-(looking up) Excuse me? I didn\'t hear you.  Mrs. S.- Nevermind let me rattle on cheerfully (Camera pans in on Mrs. S. until only her mouth is seen, switch over to Frampton. Again pan in but to his eyes.  Suddenly he looks shocked, becaus there are the dead men at the window.  He runs away)  Mr. S.-What was his problem?  Mrs. S.-Drugs.  Mr. S.- Oh.  Look little doggy-kins has learned a new trick!  (Dog dances around while balancing a ball on his nose.  Everyone laughs and the camera pans out .... the open window).   There you go, dorks.  I need to go not work now. '),(3061,1785539,'06,August,2004','2004-08-06 00:00:00','   Happy week to you, dorks.  Did you all have a pretty little weekend?  I did.  I saw me Kill Bill Vol. 2 on Friday.  I was supposed to go with the L.C. to see it later in the weekend, but I couldn\'t sit still knowing that Darryl Hannah was kicking ass.   After that I stayed up late and drank beer. The next day was Date Day with R. Tanktop.  We went shoe shopping(yes shoe shopping and BARGAIN shoe shop at that, dork princesses), bought a new ottoman, got a kickass margarita set and of course some tequila.  But Date Night was cut short when R. Sicky got a little upset stomach from eating too much mayonaise on her club sandwich (\"And I ain\'t even a member\"). There is nothing less sexy than mayonaise, dorks. If I wanted to punish terrorists I would replace their blood with mayonaise via slow transfusion.  Oh, it makes me sick! Miracle Whip is the money-shot-in-a-jar stuff.   I\'ve always had a problem with trusting other people\'s mayonaise (and still do!).   After my parents divorced I would go visit my Dad in Virginia and try to make sandwiches with his mayonaise but I just couldn\'t. I have to be present at the purchasing of the mayonaise for it to be O.K. by me. What?  Enough about mayonaise?  You know I\'m babbling to cover my Saturday emasculation?  What\'s emasculation?   So, yes R. Woods beat me at mini-golf on Saturday and as a consequence I have to make dinner and do the dishes every night this week. We were tied going into the 18th hole and then I choked harder than a clown on cock.   So, last night I made spaghetti.  (clowns---cock---spaghetti,  If this entry had one of the heavily-elipsed pull quotes it might say \"I .... choked ... a clown ... cock ... last night ... spaghetti\" hehe) Tonight!  I go baseballing.  Mariners v.s. Oakland!  Me v.s. free hat giveway mob!  First 21,000?  Fuck that I want the first one period! (actually \"I want the first one exclimation point?\" Dammit.) It seems like there is something else. Work sucks and apparently I am Dumbo.  No one will hire me away from here!  When they read my resume they must go \"Ah, yes, you have the solid background we are looking for, but you are also a cartoon elephant, we hate you, die, die, die.\" In the distance I hear a soccer mom laughing.  I know she watches \"Touched by an Angel\". I want to run away and grow a helicopter propeller out of my head and fly to a little island full of asian girls .... a little island full of my girlfriend and sip bourbon out of coconut shells and watch sunsets while I set of fireworks and chop with a machete, occasionally using my helicopter head to fly to the mainland for bologna, Miracle Whip, Anal Eaze, booze and funny pictures of monkeys doing man-like things.  I\'ll make up my own language and put out my own newspaper and then they\'d have to hire me as at least an editor, because I\'m the only one who knows the style guide of \"The Larpy Toontit\" like the back of my \"winkle\".   And I could do it, dorks, I just need to put the one foot in front of the other.  Run out the doors and through the parking lot, long jump over I-405 and the traffic, run Adidas run! But, I can\'t, I carpooled with the Hippie today and responsibility makes me sane. Later, dorks. '),(3062,1785539,'06,August,2004','2004-08-06 00:00:00','   Morning, dork nipples. First off, look at my links to your right.  Your right, dorknuts!  That\'s your foot!  Up, up, that\'s a bird!  Why are you outside with a computer!  Cave dork! There you go.  That political cartoon there is exactly like the joke I made at the begining of the week!  Dorks Don\'t Rock is so cutting edge I just cut my fucking finger off!  In fact, we\'re so razor sharp, suicide paitients aren\'t allowed to be near me without suppervision.  (It\'s now time for Brent Has a REALLY Good Idea and Then Runs It Into the Ground So Only He Finds It Amusing.) In fact, I\'m so razorrific sharp edge shittful that I\'m changing my name to Mach Brent 3 or even better, just call me Quatro. Yep, Quatro, that\'s my new name. \"Hey, hey, Quatro, new Adidas, eh?\" \"Quatro is so hot I want to buy him chocolate milk!\" \"If Quatro was a dinosaur he\'d have been a StegoQuatro Rex and he\'d get all the dinosaur chicks and all the dinosaur gold.\" Join us next time when Brent thinks of a good idea for a Halloween costume. \"A Lazer Ape!\" This weekend R. Butt and I are going to house sit for this lady with a lot of cats.   The house is very Old Lonely Lady and therefore not much fun, unless maybe she has painkillers or those nitro pills .... Focus, don\'t reveal your desire for old people pills, you just did, shit, think of cover up, go with something like So, R. Franklin and I had a great idea.  When we get home from work this afternoon (that\'s get home respectively, dorks, I don\'t do office romance ayight?)  we are directly changing into pajamas!  I\'m taking one t-shirt and one pair of boxers and that\'s it for the entire weekend.  I might even skip out on brushing, but that would not be good for my junks life so I will.   And don\'t forget to score old lady pills. But not the estrogen ones again, O.K.?  Last time I cried for a week. But your skin was softer and you did appreciate pink a lot more. And I finally understood \"Dirty Dancing\". Nobody Puts Brent In A Corner I\'m typing to myself.  Now they won\'t even say the dead troops names on some news stations?   Who do you think is dying in this war, dorkneck dick lickers?  Last time I checked it\'s almost 800 dead PEOPLE, not dolls.  Girls play with dolls, people die in battles. \"We Support Our What\'s-Their-Names.\"  Last night I had a horrible nightmare.  It started out with me watching the nightly news (Source:the position I\'m up for, watching \"The Daily Show\" before bed). \"That hummingbird in your backyard might not be a hummingbird after all...\" \"Shit.\" \"It could actually be a flying snake-like creature that is nature\'s most deadly weapon.\" \"Double Shit\" So, I look outside and I notice an old woman sunbathing (an old home video from my childhood). Back on the T.V. ... \"This unsuspecting woman thought the thousands of hummingbirds that had gathered in her backyard were awesome, until they attacked her.\" The flying snake birds (from yesterday\'s news story about the snakefish on the East Coast) attack the woman. \"One bite from the birds is lethal.  Victims first show signs of the flu and then die within days. (zombie movies)\" \"I can\'t let them kill that woman! (Boy Scouts)\" So I run outside and start attacking the snake-birds.  But once you attack one you attack them all (Playing Halo on XBox).  I was fighting off a thousand of them (Kill Bill), but they wouldn\'t die easily.  I had to crush then with objects lying around my yard AND avoid their jaws of venom (Wig Wolf)!  It was hard (porn). I had them down to the last two who I trapped in my dojo.  That\'s right, I had a dojo (Enter the Dragon). I was flipping and swinging and they were snapping and biting and I trapped one under a pillow.  I stepped on him and his venom sac exploded on my face. \"I\'ve been infected. (28 Days Later)\" The last snake-bird was laughing at me. \"A fool you were to risk your life for the old lady\'s.\" \"A fool you were to be born.\" (Splinter in \"Teenage Mutant Ninja Turtles\":  Death comes for us all, Oroku Saki, but something much worse comes for you. For when you die, it will be  [Shredder falls from the building and lands in the back of a garbage truck]  ... Without honor. ) Then I crushed him. I started to get sick and then .... woke up and peed (water, PBR).  That\'s it. Weekend, dorks. '),(3063,1785539,'06,August,2004','2004-08-06 00:00:00','   Wednesday to you, dorks.   It\'s quiet in the halls of the place that gives me money for porn.  Maybe the soccer moms are planning an attack.   Battlestations! A lady stopped me in the hall today, \"Just so you know, my 22-year-old daughter fell off my scooter.\" \"Oh yeah?  I once rode a bike off of a cliff and smacked my nuts on the handle bar when I landed.  So, who wins the stupid story contest, lady?\" Actually, I just filed her babble-dabble under \"Useless Stranger/Daughter Scooter Related Injuries\" in my brain and went on my way. Not much of words back from place me want work.  Send e-smokesignals today as requested by chief, short word back thanking for pow-wow.  Not respond again, young brave not want seem pushy.  Brent, did you yell at a foreign person or a retarded person last night? Both.   I went to buy some beers at a gas station. This chick in front of me accidently dropped a bottle of girly smirnoff drink.   \"I will pay for it,\" she said. \"What happens, man?\" said the unruly store clerk. Mumble, grumble, bumble (stopped paying attention for a few seconds because I noticed even NEWER flavors of Doritos!) Then... \"You don\'t have to call me stupid!\" said the girl. \"What happens, man?\"said clerk ass again. \"Your boyfriend got the gas, outside the pump, go him,\" said retard. Girl walks out followed by clerk and jerk.  They seem to all have a little talk in the parking lot.   When they come back in I decide that the place is too weird and put the beer back.   \"Sir, I help you!\" \"You guys are weird, I\'m leaving.\" So, I get to the car and R. Waiting asks me \"What the hell was going on in there?\" I explain, she tells me that the guys were yelling at the girl in the parking lot while her boyfriend just sat there. \"Hmmm,\" I thought, \"I must really have been into those Doritos, because I missed all of that.\" Then I said, \"Not while Batman is around!\" \"What?\" said R. Huh. So, I marched back into the store and looked the clerk right in the eyes. \"It\'s rude to yell at women,\" I let him know. Then I looked at old redneck retard. \"Rude!\" Then I left.  I am America\'s Boyfriend. Later, dorks, it\'s time for naps! '),(3064,1785539,'06,August,2004','2004-08-06 00:00:00','   ... we will be back in Fallujah. It will be at the time and place of our choosing. We will hunt down the criminals. -Brig. Gen. Mark Kimmitt  I don\'t think sealing off the roads and refusing to let families go home for weeks, advising \"bad guys\" to get out of town, have silly negotiations to get the enemies\' weapons and then carpet bombing the shit out of an entire city is really  hunting \"down the criminals. \"Hey, Mark, where are you going this weekend?\" \"Well, Neil, I\'m going hunting.\" \"Have you sealed off all the entrances to the forest yet?\" \"Yep, I figure anything left in there come Saturday will be deer and not turtles or Iraqi children.  Plus, I\'ve been in negotiations to get all of the deer to turn over their horns.\" \"Are they complying?\" \"Doesn\'t really matter, does it, Neil?  Come Saturday morning I\'m dropping bombs on Bambi! Hahahahahahaha!\" \"Hahahahahaha! Stupid deer!  Great white men!\" \"Fuel up my F-16! We got Iraqi venison to kill!\"  \"Hi, my name\'s Dick and for the 9/11 Commission Talent Show I\'m going to have my puppet George here sing \'Toxic\' by Britnney Spears while I drink this glass of cover-up ... I mean water. Right, George?\" \"I\'m a fire hazard.\"  LET\'S DORK!  Today I had an interview at Dream Job World.   \"Brent,\" Kyle asked me, \"what is your dream job?\" \"Cigarette-smoker Porn Watcher!\" This place is almost that cool.  I find out soon if I got it. I\'m whorish with anticipation.  Had a good-then-drunk-then-depressed-then-nappy weekend. Saturday night R. Throaty (who supposedly \"had a rushed tonsil removal surgery\" today, A.K.A. \"Brent rules at excuses to sneak out for interviews\"  \"Brent, can you please lift this box to save the world?\"  \"I would, but my arms are full of worms.\"  \"What?\"  \"And I fell asleep.\"  Back to ...) and I had people over to drink.  I met lots of cool new people.  I talked with tall and short people that I new (know), I drank lots of beer and vodka and touched R. booty. Then came Sunday. \"Brent, we\'re a thrash metal band that is going to move into your body.\" \"I say, what?\" (Thrashing metalically in my head, stomach and soul) \"YOU CAN\'T RUN FROM THE DEVIL, HE PUTS HIS EVIL IN YOUR EYES AND BLEEDS OUT YOUR NOSE!\" \"Prozac take me away!\" \"THE DEVIL ATE YOUR PROZAC NEXT HE\'LL EAT YOUR SOUL AND THEN YOUR FRIES FROM THE BOTTOM OF YOUR BURGER KING BAG!\" \"What?  He ate my bag fries?  What kind of Devil is this?\" \"HE\'S A CONTRACT DEVIL WE HIRED HIM ON A TEMPORARY BASIS REALLY HE\'S NOT THAT GOOD BUT WE\'RE STILL GIVING HIM A CHANCE.\" \"That\'s nice of you.\" \"NO, IT\'S EVIL!\" \"Oh, right, that\'s \"evil\" of you.\" \"THANK YOU!  THRASH METAL RULES!\" \"Yes, it does Body Snatcher Band.\"  Hi, this is the part of Brent\'s brain that is not affected by cafeine and cable T.V.  We have taken over because Brent has slipped into a creative coma, He has again lost track of where he was going with one of his silly conversations, we will return him shortly when ...  I Lead Pipe you, Mother fuckers!  Ok, dorks, so I had a goodish weekend and last night I bought Aqua Teen Hunger Force on DVDish.  I can\'t wait to watch more of it!   Body Snatcher Band, play me out!  Later, dorks! \"WELL, THE DEVIL DON\'T LIKE YOUR GRANDMA AND HE DON\'T LIKE SHOPPING NEITHER BECAUSE HE\'S THE KING DARK DEVIL AND THE LORD OF THE NETHER ... WORLD!\" '),(3065,1785539,'06,August,2004','2004-08-06 00:00:00','   More about \"The Matrix\" and my first ever thing about Shawn.  Shawn/Skills: So, what is the Matrix?  Isn\'t that the deal, noone knows what the Matrix is?  Is it a robot or something?  Me:Nope, it\'s like a dream world that machines made up so they can keep us locked up and use our bodies as batteries.  Shawn: What?  So, it\'s like they are doing this fighting in their sleep?  That\'s retarded, it\'s all imaginary! I can take on 200 guys in my mind!  (closes eyes) Look, there, BAM! I just beat up a thousand guys.   Words as true now as they were the day they were spoken, dorks.  Disney still sucks cock. '),(3066,1785539,'06,August,2004','2004-08-06 00:00:00','   There\'s Superman, Flash, Batman and Robin WonderWoman will always leave your heart throbbin\' But have I got a real hero for \'ya. He\'s Nail Gun man ... Isidro Mejia Nail Gun man fell from a roof Nail Gun man ain\'t bullet proof Or nail proof to be honest with you Because sure as hell those nails got through. Six in his head when all\'s said and done. Now falling from a roof ain\'t all that fun. But hitting your buddy on the way down  and having him shoot nails into your crown Well, that\'s injury to injury if I ain\'t wrong. Nail Gun man, you need a song!  Nail Gun Man, Nail Gun Man Fell off a roof, that wasn\'t his plan  He got shot up with six nails And then passed out before he could wail ... Oh wherever there\'s a nail gun We\'re sure to have some good fun \'Cause he\'s the Nail Gun Man!  (by the way, if you don\'t know what I\'m talking about, read up on news, fool dorks ... or just look at my new \"wicked\" link)   Morning, dorks.  I hope you are all reading well.   So, did you all have a nice little Cinco de Dorko?  I did, I made nachos for me and R. Amigo and then went to bed at 9:00.  Why?  Because I like nachos and I was tired. Other things happening?  I don\'t know, maybe.  This has been passive week for me.  I\'m totally phoning this one in.  \"Brent, we have a bit of a concern about ...\" \"Do I want a cream cheese bagel or an egg and bacon bagel?\" \"Um, what?  We are concerned about ...\" \"Yep, decisions are hard sometimes.\" \"Will you ...\" \"Hold on, start from the begining.\" \"We have a bit of a concern ...\" \"No, from the begining of time!  Act out Genesis for me!\"  We had pot roast and a walk with the Lovely Couple the other lovely night.  That was nice, thems peoples can cook.  Shoooo-eee I tells ya, and they sho got tables manners!  I was slopping up like a hog in Heaven, but theys all \"Use the fork and use the knife/That way dinner\'s extra nice.\"   Earlier that day R. Sandy brought me Subway for lunch and we had a picnic at a park.  Then we fed ducks.   That was a good day.  Well, except I rented \"The Matrix Revo-suckcocks\" (which lived up to it\'s name).  12:30 a.m. Me: (crawling into bed) Goodnight. R. Slumber: MMMM-was your movie good. Me: No, it was horrible.  4:30 a.m. Me: (waking up startled) What? What\'s going on?   R. Slumber: You O.K. Me: I\'m just having a Matrix dream. R. Yah: I thought you said it was a horrible movie. Me: Yep, and it\'s a horrible dream too.  Want another Wig Wolf e-mail?  Will that shut you up?  Dammit!  Here!  Brent, Per requests made by the (Stupid) Team.    	In the future when delivering paper, please move the boxes under the counter.  They are 	complaining that the aisle is narrow as it is and is difficult to walk through when the paper 	boxes are left partially out in the aisle.  Please push the boxes under the counter, which will 	allow the staff to walk down the aisle without the risk of an accident.    Thank you in advance for your cooperation and assistance in making our printer area, a safe work environment. Thank you, WigWolf  Wig Woooooooooooooooooooooooooolf!  This has nothing to do with my job!  She is so bored and fat and ranibowey that she must spread her venom before she dies, before Satan calls her back!  I know that doesn\'t make much sense, dorks, but neither does she. Here\'s a template for a WigWolf e-mail.  (Put persons name here), Please (do something that you don\'t ever need to).  This is because (put down some invalid excuse).  Please (repeat your request again, people are too dumb to remember things after reading two sentences).  You may also (With this paragraph, feel free to really let your insanity show.  Talk about other people, office supplies or your father.  Just go with the first thought you have and really run with it.)  I thank you in advance for (doing what you don\'t need to and probably won\'t   do).  I must now return to my master, The Dark Prince of Fear and Murder. Thank you, WigWolf   \"Hi, I\'m Charles Manson, do you want to go get Ice Cream?\" \"Hi, I\'m WigWolf.  No, let\'s kill people instead.\" \"Yes, lots of people.\" \"And carve swastikas into our skull.\" \"Yes, WigWolf, yes!\"  Later, dorks.  Disney sucks cock! '),(3067,1785539,'06,August,2004','2004-08-06 00:00:00','   Maria Shriver has made Pat Tillman speak from the grave.  The ventriloquist lives, the dead are silent heroes pushed into eternal rotting iconic roles!           Shriver said Tillman epitomized the message her uncle, John F. Kennedy, delivered in his presidential inauguration 43 years ago.  Â Â Â Â Â  ``My uncle once said, \'Ask not what your country can do for you, ask what you can do for your country.\' You, Pat, have lived those words,\'\' she said.   Dorks, dorks, dorks!  J.F.Mutha Kickin\'K. said those words so America\'s youth would HELP the people of the world through organizations like the Peace Corps!  He wasn\'t saying join the Army and get shot while shooting people! I don\'t remember the quote going \"Ask not who your country can shoot for you but who you can stack into a homoerotic Iraqi-prisoner pyramid for your country.\"  \"Ah, duh-er, we don\'t want no gays in the military and them homos shouldn\'t be getting ah-huh-der married in Oregon-io ... he-he, look, Pippytits, two Iraqi dudes all tied up and such.  Let\'s make them jerk off into each others mouths, them stupid homos.\"  And it happened and people die and homos hate and people lie and Haliburton and Daddy too get richer and you boo hoo hoo and quote the dead and make them praise any flag you\'d like to raise.  Did all of you pretty little dorks remember to buy your mother a present for this weekend?  I did, I did! I made her a card that goes as such in crayon: Outside:(A picture of my mom dressed in a superman costume with a giant \"M\" instead of an \"S\" with \" You are a Super Mom.  Inside: (SuperMom flying towards a T-Rex that has razor fingers and shooting fire out of its mouth!) Now kill that T-Rex and save THE WORLD!  Happy Mother\'s Day!  Love,Brent Other things are busy.  I have officially been offered my job as a permanent position. \"You\'re officially unofficial,\" said R. Emembers. She reads, she remembers!  Meat and New Balances sing the praise of a fan and girlfriend rolled into one! I need to go pee in a cup now, dorks. And then I might go to my drug screening.  Later, dorks. '),(3068,1785539,'06,August,2004','2004-08-06 00:00:00','   Hello, dorks. Are you glad it\'s Friday?  I know someone who is! Me?  Nope ... Well, yes ... But someone else! It\'s Weekend Delivery Man!  Monday: \"I can\'t wait for the weekend.\" Tuesday: \"One down, 4 to go!\" Wednesday: \"Almost there!\" Thursday: \"Feels like Friday to me.  Does that count?\" Friday: \"Didn\'t think it\'d ever get here.\"  Our entire relationship is based on how many days there are until the weekend.  I fear that one weekend I\'ll run into him at the porn store or the museum and it will be awkward. \"Can\'t wait for the ... just one ... three ... the wife said .... my doctor ... the weather.\" It\'s quite possible that me seeing this man on the weekend could rip the universe in half.  Or maybe it isn\'t, who knows?  We need billions more dollars to fund Iraq?  I thought we planned to be out by June 30th!  Here\'s a startling math fact for you (which I figured out by the way, thank you internet and BrentBrain4000)... If we funded the bare minimum of 2 billion dollars (the minimal plural ammount) and we leave Iraq on June 30th that will be an average daily (excluding today) spending of $42,553,191.  That comes out to being $0.15 per U.S. Citizen per day.  For less than the price of a cup of coffee you can kill a child in Fallujah.    Amazing!  So, this weekend I will do nothing.  Well, tonight I shop for breakfast sausage and dinosaur nuggets, but then I\'m retiring to my XBox.  I rented RedDeadRevolver.  It\'s O.K.   I watched the final showdown scene from \"The Good,The Bad and The Ugly\" to get myself pumped for the game.  That is without a doubt my most favorite movie sequence ever!   A close second is when those lesbians made out during \"Raiders of the Lost Ark\".  Oh, you missed that scene?  That\'s because you don\'t have Brent-Imagi-Vision!    Howdy!    I should probably go work ... nope. Weekend, dorks! '),(3069,1785539,'06,August,2004','2004-08-06 00:00:00','   Hello, dorks!  Welcome to the revampired Dorks Do Not Rock!  What they \"Don\'t\"?  That\'s right, they don\'t! As you can tell, this page is now blue, which is the color of my shirt and jeans.  Coincidence?  Maybe, I\'ll never tell.  It\'ll go down in history as one of the great mysteries, like why was Paul Bunyan so tall and who really built Big Ben? (pssst, it was the British) You can also leave me comments now.  I was wary of doing this, because I really fear that some dorks are going to leave stupid things.  No stupid things!  What\'s stupid?  Well, anything that isn\'t perfectly written or doesn\'t make me want to scream \"switchblade\" and/or laugh.   Feel free to leave your phone number so other dorkreaders can call you for phone sex. Not me though, me no call.  Me have girl who curls red. R. Esponsible had to work at the place where the save kids last night.  So, I had the whole place to myself.  I beat Halo, had a burrito, a few beers, tried to watch T.V. .... and then I was freakin\' bored out of my skull. I paced. I looked at DVDs (didn\'t watch, just looked at them.  I have a problem now that when I go to watch one my DVDs the movie plays through my head very quickly and I feel like watching it will be a waste of time ... plus all the girls can be naked in my imaga-movie!) I paced. I tried to decide between peanuts and chips. I ate chips. Contemplated whether my decision was the right one. Ate more chips. Paced in front of the peanuts. Settled on the chips. Went to bed.  This is what I do when I\'m alone.  I pace and eat chips and think about peanuts.  Ask anyone who has lived with me, although they weren\'t with me when I was alone (because I was alone) they know my pace habits.  I walk around, complain about boredom and not knowing what to do and then go to bed.  I really missed R. Gone too, dorks.  Calm down, shitdorks!  You aren\'t getting ooh-ahh-fodder out of me.  It\'s just that last night was the first night I had to sleep alone at our place.  It was kind of not as sad as you think but sadder than dropping a candy bar into gravel. All of the plastic trinkets and girly things that I complain about cramping my style reminded me of R. Nots.  I started to like them because they were part of her and I was there and she wasn\'t but her stuff was.  So ... I now love the plastic piggy bank!  The rainbow flowerpot is A-OK!   Hula man and pink pillow welcome to my world of love and understanding!    Ooooh Aaaah Dooooorks!  \"Why do you always chop off my head after I kill your kid?\" Are you angry yet, dorks?  Are you angry at the right people?   Vooooooooote!  Nervous to let you silly dorks tell me if my website works and talk about what I write about girls butts and why we fight.  Later, dorkarinas! '),(3070,1785539,'06,August,2004','2004-08-06 00:00:00','   \"My name is Nick Berg, my father\'s name is Michael, my mother\'s name is Susan.  I have a brother and sister, David and Sarah. I live in ... Philadelphia.\"   \"For the mothers and wives of American soldiers, we tell you that we offered the U.S. administration to exchange this hostage with some of the detainees in Abu Ghraib and they refused .... So we tell you that the dignity of the Muslim men and women in Abu Ghraib and others is not redeemed except by blood and souls. You will not receive anything from us but coffins after coffins ... slaughtered in this way.\"    And, yes!  Where the fuck are the mothers of America?  Your sons heads are rolling!  You will never receive handmade ashtrays like you did on Sunday!   There is no patriotism dying for Haliburton, Moms!  There are no heroics in having your head chopped off over a dick-measuring contest, Moms!  Your son could fight fires in California and still come home to you and pie.  He could teach P.E. or someone to read and make more of a difference!   Revolt, Mothers, revolt!  Revolt against the puppets and the killers!   And you, dorks, you see through the hype.  This execution was avoidable.  Anyone called to arms over it is being turned on like a robot.  Turn yourself off! Take out your patriotic Duracell! Fight fires!  Teach!   Dammit, dorks, dammit!!!!!!!! '),(3071,1785539,'06,August,2004','2004-08-06 00:00:00','   Morning and hopefully happy not-being dead to you, dorks.  If you have become a zombie and are reading this then shitty for you.  Shitty indeed. I have missed you all, dorkles.  I really did.  I\'m sorry I didn\'t call like I said I would, you are still the prettiest dork I\'ve ever seen.  No, I\'m not just saying that.   Me and the L.C. and the R.Gee went and saw the Super Size Me movie on Sunday.  I\'m done with McDonalds, dorks.  Done with McDonalds for life.    And this isn\'t like when I swore I wasn\'t going to use the Mach 3 because of rising prices.  Sunday: That\'s it!  Those arogant bastards can eat my ass!  I\'m going back to the pre-Mach 3 days!  Monday: Oooo, I do get such a smooth ball shave.  Kiss me Mach 3!  (I tongue a razor)  The blood!   I will never kiss a razor again!  Tuesday: Damn, Mach 3 have you been working out?  (Tongue the razor again)  Aaaaaaaaa!  My promise broken!  The world falling apart like tongue does too!  I will never break a promise to myself again!  Wednesday: Well, I guess it\'s O.K. to punch just one wheelchair-bound geriatric when no-one\'s looking.  (Punch an old guy)  Hooooeee, I missed that!  What fun!  Why\'d I promise I\'d never do that again?    I\'ll never make promises again!  Thursday:  I promise, it won\'t hurt.  (It hurt)  Dammit, I\'ll never suggest new things again!  Friday:  Hey, who wants to throw rotten fruit at the homeless and see if they eat it? (I do it.) They ate it!  I\'ll never doubt a homeless man again!  Saturday: Bullshit! You can\'t beat Slinky Boxcar in a shiv fight, Hobo Bob!  It\'s suicide!  And that has been a waffling week in the life of Brent! Back to life, back to reality, back to the hear and noooow-ow-ow. So, I\'m done with fast food.  I must regain my healthy 16-year-old physique so I can be a strong warrior when the rapture comes ... so I can look good at the beach.  Cows: \"Yippeeee!\" Potatoes: \"Spuderific!\" Whatever they make the vanilla ice cream out of: \"Glurp!\" McDonalds Exec: \"Our profits are down 21%!\"  Other news might include that I\'m close to beating Halo and therefore have put my love life in jeopardy. In Jeopardy. \"I\'ll take \"XBox = No Box\" for 500, Alex.\" \"Upon reaching level nine of Halo this unshaven ex-frat boy forgot that his hot girlfriend had a lovely rack AND a softball game he promised to attend.\" \"Who is Brent Kinkade?\" \"Correct ... however, in his defense, Brent did remember to tape the final episode of Friends for R. Luckiest Girl In The World.\" \"And unload the diswasher!\"  Did anyone else watch Friends finale done gone bye?  I did.  Why are they so zany?  I stopped watching about 4 seasons ago when they kept sleeping with each other and everyone had a secret or was pregnant. \"Don\'t tell Monica that Chandler has shrapnel in his brain because he\'s afraid it will cause her to sleep with Joey and then have Ross\' baby who is actually Phoebe in a fetus costume that Rachel stole from Ralph Lauren!\" \"Hmm, I\'m going to go eat lots of McDonalds instead of watching this.\"  R. Campy and I bought a new tent last night.  It is very big!  We set it up while watching \"The Swan\" my sort of new favorite not-liked show. Want to her something that sounds like it\'d be a good compliment but isn\'t? \"Baby, you\'re the swan.\"   Yeah, girls don\'t like to hear that.  You know what I don\'t like to hear?  That you are a lazy stupid dork, you dork!  Why hasn\'t President Bush\'s approval rating fallen to Negative 2000 %?  Why?  Because that isn\'t a real percentage?  Or is it because you have cotton in your dumb dork ears?  Or maybe it\'s because you are a stupid hippy-Democrat instead of a me-hit-rubberhead-Democrat?  Maybe you should stop crying in your polluted rivers and get a little more final-scene-of-Robocop! President Bush has maimed you and made you into that hideous man/machine who has been isolated from the love of your family and the compassion of society!  There he is in that pile of toxic waste!  He\'s melting, oozing his sludge guts!  Now\'s the time! Run him down, MurphyDork! Last week was prime go-for-the-jugular week and Kerry and all of us passed it up.  Your comander in chief (who is more in charge of the war than Bumsfeld!) had soldiers sexually and physically assaulting people in a foreign land ... WHO WEREN\'T EVEN OUR ENEMY!  We declared Sadam\'s loyalists our enemies, not Everyday Joe Iraqi!  The Red Cross says that as many as 90% of the prisoners were falsely imprisoned!  That\'s 90% of them dressed up as women and jerking off into each others mouths because our troops had run out of inocent people to kill!   DOES AMERICA NOT READ? Then President Bush AND Rumslick himself said it might be a good idea for him to resign!  Then Cheney calls him the best Defense Secretary ever?  All of a sudden everyone is backing Rum/Cheney again. \"Um, George no like Donald.\" \"Yes he does.\" \"Yes George like Donald.  George president.  George use phone!\" \"Yes, George is a big boy.\" Arrrrrrrrrrggggggggggggggghhhhhh!  When will the homoerotic slaying of children end so we can all get back to a good economy and nude girls surfing?   When, dorks, when?  Other than that, things are good.  Not much else to say, zombie dorks.  Behind you! '),(3072,1785539,'06,August,2004','2004-08-06 00:00:00','   My job is ...  5 minutes ago, singing in my head: Bite down through the wire, ya-ya la the fire.  Ya da da, da da ooooh yeah ... dum dum da, I use two lifelines ... OH SHIT! WigWolf at 9:00! Evade! Evade! She\'s on our 6! ((side note: You know that scene at the end of \"The Neverending Story\" when Atray... um, the injun meets the wolf of the nothing thing ... WigWolf is starting to look like that, which is to say Hot as Balls)) Left foot, right foot, left foot, right foot ... she\'s going to talk to us ... we can\'t have her talk to the precious Brent ... left foot, right foot ... dammit left foot keep up!  Too Late! Abandon foot, abandon foot!  Past the front desk and to the elevator, focus, focus ...  Jersey Hair: (At the front desk, covering for ManFace since she quit) There you are, you were supposed to be up here at 2:00.  (I smell foxy boxing!)  WigWolf: I was going to but then my father, you know The Dark Snake Prince, called and asked me to fetech the blood of bunnies and Straight-A students.  Jersey Hair: Well, I had to cover for you and now Blondie is here to cover for me. (Uh-oh)  WigWolf: I smite my lateness, but ..  Jersey Hair: Too late, you were supposed to be here.  Thanks again for being a real team player. (Oooooo)  (WigWolf turns into a mist of smoke and fades away into the air ducts ... and hopefully history.  I run for the elevator before she gains enough strength to return.)  Yippee Jersey Hair, the revolt of the human heads has begun!  Beware ye heads of rubber, your day is numbered.  Surely within a fortnight\'s hence ... Ah, fuck that, I ran out of fancy sounding words. Anyways, this might be a lame job, but it has it\'s moments.  Then again, there\'s never a lame moment when cat fightin\' is involved. Meow, weekend and down with the rubber heads, dorks! '),(3073,1785539,'06,August,2004','2004-08-06 00:00:00','   From CNN.com: BAGHDAD, Iraq (CNN) -- Iraqi Governing Council member Ahmed Chalabi said a raid on his compound Thursday was engineered by Baathists who control the Iraqi police and who are now protected by the Coalition Provisional Authority.  Arrrrggghh, dorks, Arrrggggh! Why the hell are the Baathists raiding anything in Iraq, dorks?  Who are the Baathists, you say? READ!  READ!  Oh, they\'re just this little group that used to be headed up by a guy named Sadaam Fucking Hussein.  What the hell is going on over there? We invade Iraq and declare Sadaams loyalists to be our enemies.  Why?  Because they weren\'t behind 9/11 and they don\'t have WMDs.  That good enough for you, you un-American prickhole?  Buy a flag and shut up! O.K., so we aren\'t attacking Iraq, we are attacking it\'s \"evil\" government. Then we kill some babies and bomb a wedding. (http://apnews.myway.com/article/20040519/D82LPPNO0.html) Don\'t worry, it was a gay Iraqi baby wedding. Ain\'t noone alive to bomb the SuperBowl. Now we are supporting raids by SADAAM HUSSEIN\'S FORMER MILITANT PARTY?  These are the people who would shoot you if you couldn\'t guess how many fingers were behind their backs!   I know a guy who knows a guy who\'s sister\'s husband was a Baathist and you know what?  Horns.  He had horns. Who the hell are we fighting over there? We have seriously run out of kids to kill and people to force into homoerotic poses.  We have actually sided WITH OUR ENEMY TO TAKE ON ... um, people who don\'t believe in naked Iraqi pyramids ... I mean Democracy.  But, don\'t worry, because I know what\'s going on in Iraq. A monkey with some dice is running the war. Crazy? Yes.  True? Yes. Think about it, dorks. Only a monkey with some dice could come up with a plan to invade the wrong country, kill children, bomb weddings and then a year later support our enemy to fight .... our enemy? \"What\'s it going to be today General Bongles?\" \"Ooh-Aaaaaaaaaah, Uh, Uh, Uh.\" \"That\'s it, men, General Bongles has rolled an 11.  That means all-out nuclear war against Spain and bananna splits for all of us.\"  When you put your hand into a bunch of goo that a moment before was your best bananna split, you\'ll know what to do.   I had a nice day off, dorks.  I tended to R. Sickly post-tonsilictamoyoney67yey.  That meant pudding and oxycontin.  For her we had ... hahahahahaha, just kidding, dorks. I laid around and watched movies and ate sandwiches.  I like getting paid for someone else being sick.  It\'s a sick day without all the sickness!   \"Can you please ...\" \"Sorry, R. Baby is sick and by dating default I\'m as useless as her.\" \"Would you like to ...\" \"...watch T.V. and not do anything? Yes.  Anything else? No!\"  And what did I return to?  WigWolf mayhem?  No, this time it was \"Crumble Cheeks\" (aka \"What\'s a folder\" Lady)  _____________________________________________  From: 	Crumble Cheeks Subject:	Training Boxes/// Importance:	High  (Brent) delivered boxes to the attention of (people) to me 1st floor Training Room.  Thank you, Crumble Cheeks  ---------- From: 	(Beaver) Subject: 	RE: Training Boxes///  Brent - Can you please bring them to the 1st floor training room? Thanks  -----------  From: 	Crumble Cheeks Subject:	RE: Training Boxes///  They\'re already there, Beaver!!!!  -----------  From: 	Beaver To: 	Crumble Cheeks Subject: 	RE: Training Boxes///  I\'ve been down here all morning and I don\'t see them  ______  From: 	Crumble Cheeks To: 	Kinkade, Brent; Beaver Subject: 	RE: Training Boxes///  did you place them in the lobby area, Brent?  ---------- From: 	Kinkade, Brent To: 	Crumble Cheeks Subject: 	RE: Training Boxes///  Beaver called me.   When you said \"training center\" I thought you meant the literal training center. Beaver has the boxes now.  ----------- From: Crumble Cheeks Subject: RE: Training Boxes////  what do you mean by that?   (end of e-mails)  What do I mean by \"I thought you literally meant to put the boxes in the place you said to put the boxes in instead of the place you were thinking of in your rubber head.  Slower? YOU SAID \"TRAINING CENTER\" AND MEANT \"LOBBY\".\" or \"Beaver has the boxes now.\"? Yippee, rubberheads!  I can always count on you to turn a simple situation (\"Hey where\'s that box?\") into a 17-alarm fire (\"The box is full of Anthrax!\" \"No, it\'s not\" \"It\'s not a box?\" \"No, full of anthrax.\"  \"A box full of anthrax!  What do you mean by that?\").  Bye, dorks.  Please, please, please someone find me a decent job in Seattle! '),(3074,1785539,'06,August,2004','2004-08-06 00:00:00','   \"Wait, I need the black toner.  This one says \'noir\'.\" \"\'Noir\' is french for \'black\'.\" \"I\'m glad one of us knows French or we could have had a problem.\"  Oh, silly rubberheads,I need a break from you. As we speak, dorks, R. Throaty is having her tonsils ripped out.  Tomorrow I stay home to play doctor and XBox.  Au revoir, les tetes des rubber! '),(3075,1785539,'06,August,2004','2004-08-06 00:00:00','   Howdy!  Are you smirksome, dork?  Then let\'s go! I had me a weekend full of nothing.  An empty weekend.  I bought groceries and then said to myself, \"Yep, that oughta \'bout do it for this weekend.\" And it did.  Wig Wolf\'s best friend, Man Face, quit today. I swear I heard a howl in the distance. Wig Wolf was sad today.  I found her in the lunch room eating only 2 newborn kittens for a snack instead of her regular 35. \"Oh, WigWolf, are you sad.\" \"Yes, and the strength fresh kitten meat usually gives me is fleeting today.  It hurts right here.\" \"Your heart?\" \"For you mortal hu-mans it would be a heart, but for me, a child of The Fearsome Lord FireDeath, it is a steaming lump of Hell coal.\" \"I know what it\'s like to have a broken heart, WigWolf.\" \"Steaming lump of Hell coal.\" \"Right.\"  I\'m just joshing, dorks.  I haven\'t seen WigWolf in almost a week.  I\'ll relate to you the way I did to Kyle me and WigWolf\'s last meeting. \"Maybe we should have the authority to (do what it is you do).\" \"Are you saying you think my job shouldn\'t exist?\" \"No, no, no .... I\'m just.\" (At this point I stop talking to her and walk into a supplies closet at the back of my office and hide in a corner until she is gone.)  Maybe my job shouldn\'t exist, but it\'s my job!  That works!  If I don\'t have my job then you dorks are un-American!  Hmmm?   New paragraph. I went over to the Lovely Couples for lunch today.  Jack made fancy pizzas.   Pizza #1: Yummy dough, cripsed potatoes, some wine sauce, chicken, angel smiles and puppy souls. Pizza #2: Wheat dough, dijon sauce, carmalized onions, chicken, spinach, Heaven rain drops.  They were good. Over lunch the L.C. reminded me of a show I want to produce.  It\'s called \"Beauties and the Beast\".  The premise is that 5 or so women go and get makeovers like on \"The Swan\".  EXCEPT one girl gets a fucked up makeover.  They make her look like a monster with scales and horns and bugeyes and a hook coming out of her head.  The girls have to stay bandaged up for months and are finally allowed to see themselves. That would be the most awesome show ever in the history of the world.  Could you just imagine some really fucked up looking girl with hunchback and blue skin and one of those robotic throat boxes.  That would be awesome!!!!!!!!  And she\'d have to have two mouths and a pig nose!  Hahahahahahahaha, the humor! I had lunch with the L.C. because they are going to Atlanta tomorrow.  So, me and R. Wonderful are going to house/dog sit while they are gone.  Mixed peanuts jar here I come!  And hey!  Way to go on getting Bushes approval rating down, dorks!  I knew you had it in you!  What else you got in there?  Any porn?   I paid $2.35 for a gallon of gas on Friday.   Keep up the good \"Bush-bashing\" work, Demodorks and Intelligent Republicans!  I need to go swimming or play frisbee golf after work today.  I have been sucked into my XBox and feel like a fatty due to lack of anything physical the past few days. I take a few steps and it\'s nap time! Look out supplies closet here I come! '),(3076,1785539,'06,August,2004','2004-08-06 00:00:00','   It\'s 8:36 in the morning and I\'ve run out of things to do. I\'ve been up since 5:30 this morning, dorks.   Just last year I would go to bed at that time.  What have I become?    My father, Brent Sr., wakes up with the sun. \"Cock-a-fucking-do,\" says Brent Jr. Brent Sr. now has to wear hearing aids at the age of, um, 52ish.  He blames it on rock \'n\' roll and woodworking.   \"I\'m sorry, R. Mumbles, could you repeat what you just said?  I wasn\'t facing you while you were speaking and someone is playing a radio in Texas.  Hard for me to hear you with all of those distractions,\" says Brent Jr. Brent Sr. has the paitience of a \"rabid wolf\" (to quote my ma). \"Once upon a time ... fuck this,\" says Brent Jr. Brent Sr. has the memory of a dead elephant. \"I\'m out of food and toothpaste.  Let\'s go to Kidco,\" says Brent Jr. \"You mean Winco?\" says R. Girl-I\'m-Pretty-Sure-I\'ve-Been-Dating-For-Almost-A-Year. \"Yep, Spifco.  Let me brush my teeth first,\" says who?   I have Monday off, that means if I can just play Statue game for 7 hrs. and 15 mins.  I have a three day reward and a paycheck waiting for me. Statue Game comencement in  10 9 8 7 6 5 4 3 2 '),(3077,1785539,'06,August,2004','2004-08-06 00:00:00','   Two voice notes left by yours truly on yours truly\'s cell phone.  5:16 p.m. 5/24/04 \"1 cross + 3 Nails = 4given .... huh-huh-ha\" (From a bumper sticker seen on the way home.)  1:05 a.m. 5/25/04 \"The mullet joke is not funny anymore.\"  (A truth discovered while P.B.R. influenced.  And really it isn\'t, dorks.  A dork who says, \"Look a mullet\" is really saying \"Look, my dorkhead is full of Jello\".  A mullet joke is on the same level as a Monica Lewinsky joke now ... move on!  The next big joke?  Chicken legs! (on humans))  If it\'s not the sun that gets you done or great white sharks that stop your heart or killer bees, or mad cow disease, or sarrin gas, or S-A-R-S, then don\'t you dread you\'ll still be dead before summer\'s through they\'ll still get you. The terrorists, says Uncle Sam have a scary summer plan to murder and kill and poison and shoot, maybe an explosion thrown in to boot! So run, hide and cover your head. Scream and cry \'til you see red. And don\'t ask \"Why?\" \"When?\" or \"Who?\" Cause Big Brother ... \'Cause Uncle Sam is there for you. '),(3078,1785539,'06,August,2004','2004-08-06 00:00:00','   Correspondence from a letter briefly glanced over at the office today. \"I faxed her a check but then never heard from her again.\"  Maybe next time you should e-mail the check, Rubberhead.  After that why don\'t you text message me a sandwich.  From cnn.com Bush outlined five broad steps to restore Iraqi sovereignty and get the country back on its feet:  *	Handing over authority to a sovereign Iraqi government. *	Establishing security. *	Continuing to rebuild Iraq\'s infrastructure. *	Encouraging more international support. *	Moving toward a national election in Iraq that \"will bring forward new leaders empowered by the Iraqi people.\"  From dorksdontrock Brent outlined five specific steps to stop killing Iraqi women and babies.  *Hand over authority to people who we won\'t turn around and kill/overthrow/raid/carbomb/naked pyramid stack/humiliate/intimidate in 5 years ... or days.  Then bring the troops home for a game of softball, some pie, whores and crank. *Establish security by not bombing any more mosques or forcing Iraqis to jerk off in each others mouths. *Avoid the labor and costs of rebuilding with a simple \"no carpet bombing\" policy. *Encourage more international support for more naked Janet Jackson!   *Move toward a national election in America that \"will bring forward a new leader elected by the American people\".  Bye, dorks! '),(3079,1785539,'06,August,2004','2004-08-06 00:00:00','   Good afternoon, dorks.  It\'s such a nice day in Seattle that I feel I should be running through a sprinkler or making out in a corn field. Instead I have to have inane office conversations, which I am no good at, by the way. There is something in-born in every beurodork and republican that allows them to chat mindlessly about things that I have never and will never care about.   Example? Example away! The following is word-for-word ver-fucking-batim a conversation I had today.  \"Hey, how\'s it going, Person?\" \"Hi, Brent.  Well, it\'s going.\" \"Well, that\'s a good way for it to ... move.\"  I sound like a gooddamned mentally exploded caveman when I try to small talk.  \"Sure can\'t wait for Friday to get here.\" \"Me either.\" \"Weekend never seems to get here fast enough.\" \"Hmmm, racecars go fast though.\" \"Yes they do, Brent\" \"Yep, Person.\"  R. Baby and I went to a wedding this past weekend.  Which wedding?  Doesn\'t matter, doesn\'t matter, dorks. (A rubberhead just walked into my office and threw trash into my garbage can without even talking to me.  Rude? Nope, Rubberheadish.  Word of my exploded caveman brain must be spreading.) Back to the wedding.  R. Extra\'s review of the said wedding?  \"Borderline white trash\". Indeed, dimple angels, indeed indeed. I wore a suit, no tie and was worried that I would be under dressed.  I wasn\'t thanks to \"These my Sunday denims\" Guy and \"T\'aint No One Telling Me My Kid Can\'t Wear His Barney T-Shirt\" Woman. More specifics? A conversation related to me that R. Girl had with the bride\'s mother.  R. Girl: Hi, Mother Person, remember me?  Mother: Wait, don\'t tell me ... are you with the groom\'s side or the bride\'s side.  R.: The bride\'s side.  Mother: Hmm, then who are you representing?  R.: Um ... your daughter.  Mother: Oh, you\'re R. Girl!  The woman forgot if her daughter was the groom or the bride!  I once forgot what Fruit Loops tasted like.  I was drunk. I was also drunk at the wedding. \"Wheeeeeeeeeee ... people I don\'t know AND free beer?  Recipe for disaster!  A disaster lasagna in the making!\" But I was ordered to be on my best behavior. \"Don\'t be dumb,\" R. Bossy ordered. I listened.  Stupid society and your rules. We left early because R. Throaty\'s throat got re-sore thanks to the tonsilectamony-mony.  R. Baby\'s throat saves me again!   Hip-hip-hooray!  And now\'s the time we say goodbye But frown no not and please don\'t cry. And tonight, dork, when you watch T.V. And see a message from the G.O.P. Via your prez (that\'s good ol\' George) Remember it\'s O.K. to yawn and be bored. And if you see through all the hype,  The Season Finale of \"The Swan\" is on Fox tonight.  Fight the powers that be, dorks! '),(3080,1785539,'06,August,2004','2004-08-06 00:00:00','   Wait, the bike doesn\'t have tires.  And it\'s missing a spring ... and now, so is my step. Witness, dorks, the saddest moment of my life. '),(3081,1785539,'06,August,2004','2004-08-06 00:00:00','   A story written by my friend Eva \"Goldfinger\" Moore has inspired me to get a bike.   I found a guy who\'s selling one for $25. My life just got 3,000 times better.  I wish all you dorks could see Seattle today! Weekend! '),(3082,1785539,'06,August,2004','2004-08-06 00:00:00','   Fifteen minutes ago.  The phone goes ring, ring.  I answer. \"This is Brent.\" \"Brent, do you have any of those big envelopes?  Not the skinny ones, the ones that are big,\" says Crumble Cheeks. \"Yep.\" \"Could you please bring me some?\" \"Yep.\" I Grab a handful of the envelopes and head to her office.  This is the same Rubberhead who was baffled by how a folder works until I opened it up for her. I am silently chanting \"I know something stupid is going to happen.  I know something stupid is going to happen\" as I enter the office. \"Here you go.\" \"No, these aren\'t them.   I need the big ones.\" Silence. I take the top envelope off the stack I have brought up.  I open it. \"Oh, OK, that\'s better.  My stuff can fit in there now.\"  Rubberheads, dorks, my life is full of Rubberheads. '),(3083,1785539,'06,August,2004','2004-08-06 00:00:00','   From the Tri-Valley Herald: Condoleezza Rice insists that (Bush) will one day rank alongside such towering pillars of 20th century statecraft as President Franklin D. Roosevelt and British Prime Minister Winston Churchill.  \"When you think of statesmen, you think of ... people like Roosevelt, people like Churchill, and people like Truman...And this president has been an agent of ... historic change for the better.\"  Repetition does not transform a lie into truth.         -Franklin D. Roosevelt  Never in the face of human conflict has so much been owed by so many to so few. - Winston Churchill  Humility is to make a right estimate of oneself. -Harry Truman  I know that human being and fish can coexist peacefully. -George W. Bush  Well, maybe he can rank with Aquaman and Dr. Doolittle.  A small glimpse into my life .... 3:30 a.m. this morning Riding a dinosaur, chasing robots, is that Michael Keaton? Yippee boobies and my childhood home ... Suddenly! I wake up on the floor of my bedroom. \"What the hell?  How\'d I get down here?\" R. Sleepy waking up, \"Mmmm, you woke up around midnight and started bumping into things and then grabbed a pillow, fell down there and went to bed.\" \"Why?\" \"I don\'t know.  I asked you and you just said, \'I don\'t know what I\'m doing\'.\" Crawling back into bed, \"The life I lead.\"  Have a good one, dorks. '),(3084,1785539,'06,August,2004','2004-08-06 00:00:00','   I woke up late today ... It\'s time for Morning Math! Current time (x) added to estimated time of shower (y) multiplied by the average number of backups on I-405 (z) over the ammount of words I\'ll have to use to explain to my hippie carpool partner why I\'m late (a) = haul ass!  Made it! \"Sorry I\'m late.\"  \"Yeah, you know, it\'s really hard to judge if you\'re going to make it, I mean you really need to know if you really know the other person, like if they are going to stay or if they are going to go ahead.  I mean I\'m on time because the kids had a game yesterday and they were tired so I could go to bed, but I\'m worried that their real mother is not taking care of business, and so ....\"  (Fading out.  Inner toughts taking over ...) \"Taking care of business? ... la da da if my train\'s on time I can be at work by nine. Start your yah-yah job and get your pay.  Hmmm- da second hand guitar  hmmm-da ... you\'ll go far if you get in with the right group of fellows.  ... \'Cause we been taking care of business, EVERYDAY. And we been taking care of business EVERYWAY.  We\'ve been taking care of business ... hmmm landmine.  We be taking care of business and working overtime.  Work out.  Dadadadadadaa DA DA\"   \"...so I told him that pudding shouldn\'t be gray.  Can you believe that?\"  \"Nope.\"  Nice save, brain. Have a good day, dorks. '),(3085,1785539,'06,August,2004','2004-08-06 00:00:00','   Welcome back, Me. I didn\'t Memorialize too much yesterday. I\'m feeling sick.  Between R. Streppy, the L.C.\'s dogs, a Rubberhead carnival last week and shopping at every freakin\' sporting goods store in Seattle to find a decent sleeping bag I am exhausted.  Exhaustion + Mac & Cheese/Natural Ice diet = ... too worn out for algebra right now, dorks. I\'m clammy, crabby and sluggish.  I\'m virtually made of molluscs!   Cheerful! This weekend included: -One of my birthday presents for R. Adult (June 18th, if you are present-inclined) was tickets to Mitch Hedberg on Saturday night.  He is our comedian-in-common.  Our movie-in-common is \"Malibu\'s Most Wanted\". Mitch jokes not heard before concert: \"I flew over an island and saw S-O-S written very big.  I yelled out \'Stop! Those people have dirty pans\'\" and \"I had an infestation of Koala bears ... I turned on my lights and Koala bears scattered everywhere ... Do not hide behind the toaster, you are too big to fit back there. Plus you are cute and I want to pet you.\"  -Shopping and disc golf with the Lovely Couple Saturday morning.  We bought chicken then took a nap.  If that don\'t say rock \'n\' roll you can eat my ass, dorks!  Why do I yell?  -Shrek 2 in the theater with R. Girl. It was pointed out to me that I started squirming like all of the toddlers halfway through the movie. \"I\'m tired.\" \"Shh, we\'ll get ice ...\" \"Look! Shrek!\"  -Bubba Ho-Tep on DVD Sunday night.  What would happen if Elvis had lived and then joined forces with J.F.K. (who is now Ossie Davis-like) to fight a mummy in a Texas retirement home? This movie. 14 tubs of popcorn out of 15!  A must-see ... if you are sick of history and mummy attacks not correlating!  Another voice-note left by me (Heavily Keystone Light influenced) 2:29am 5/31/04: \"Now you can eat Dorritos ... Fritos, whatever, no matter what you\'re doing ...\" For some reason I thought this was an important topic to write about.  It\'s magnitude escapes me now ... oh, wait, I can eat them ANYWHERE?!?!!?  Even while playing SOCCER?!?!!? Miracle of miracles!  I see the glory!  Hehe.  Memory! What I bragged to my buddy Kurtis I was going to tell this chick after I ditched her (circa. 1997): \"Truly thou art the fling of flings.\" Oh, Past Me, what ribald jest!  From cnn: \"Bush praises Iraqi interim government ... blasts kill 14, including 3 near \'Green Zone\' \"  \"And your president will be Bob. (Ka-Boom) Your president will be Carl.\"  I\'m reading \"Power and Terror: Post-9/11 Talks and Interviews\" sort-of by Noam Chomsky (it\'s pretty much the transcript of the related movie from what I gather). Very good, Ameridork. 15 out of 15 tubs of popcorn.  Hmmm, done. '),(3086,1785539,'06,August,2004','2004-08-06 00:00:00','   WATERBURY, Conn. (AP) - A food fight that started with fruit cup turned into a mealtime melee, serving up a full plate of arrests and injuries ... Three girls and fours boys ranging in age from 12 to 14 were  urlLink arrested   on charges including breach of peace, assault and inciting a riot.   Seriously? What\'s next? Speeding tickets for Big Wheels?  Background checks for Cabage Patch Kids?  I started a foodfight at  urlLink  Crooked Lake Elementary School   in Andover, MN when I was a wee hipster.  It remains one of the proudest moments of my life.  I can still see the potatoes flying, the chicken nuggets pelting innocent first graders and wounded students covered in ketchup screaming \"Where is God?\" Please, let us have our country back soccer moms and NASCAR dads!  Zealots and Rubberheads let our children go!  I signed up for an  urlLink  adult kickball league   yesterday. AND the dentist told me I had \"fairly good\" teeth today!   So, on that note ... Candy for all! I brought enough for everybody! Weekend, dorks. '),(3087,1785539,'06,August,2004','2004-08-06 00:00:00','   Proof!  An e-mail received today!  Just now!   From: president@rubberheads.org To: ME! Sent: Thurs.,  June 10, 2004  11:57 A.M. Subject: Please stop making fun of us ...  Brent,  Please stop making fun of us and letting people know that we exist. You are really screwing up our plans for world domination. Thanks.   -President McRubber of the Rubberhead Foundation>>  Incriminating evidence!  They exist!  It\'s like finding out Santa is real ... and really, really stupid. I wrote back!   From: ME! To: president@rubberhead.org Date: Thurs., June 10, 2004 12:11 P.M. Subject: RE: Please stop making fun of us ...  Oh, Rubberhead.  Your plan to take over the world by \"finding a big stick\" and \"hitting the Universe Emperor while he\'s asleep\" needs a little work. Keep up the horrible, horrible work! Love, Brent>>  But, it wouldn\'t be a Rubberhead story if all went right, right? So ... The message bounced back to me.  The rubberhead\'s e-mail address didn\'t exist.  Leave it to a rubberhead to send an e-mail from an account that doesn\'t exist!  Dorks beware! '),(3088,1785539,'06,August,2004','2004-08-06 00:00:00','   Just so you know that my life isn\'t all West Nile Virus and XBox ....  Last night R. Honey and I watched the last minutes of the sunset from a large piece of driftwood on Alki Beach.  The sun saying \"Goodnight, all\" on our left and the Space Needle blinking on the right. I had mint chocolate chip.  R. Girl had rainbow sherbert. We both had sugar cones. '),(3089,1785539,'06,August,2004','2004-08-06 00:00:00','   Head pounding. Head pounding. Head pounding.   \"So, are you FEELING BETTER? I mean, have you made HEALTH PROGRESS?\"  Oh, carpool lady can\'t you see my head beating like a drum?  That\'s not normal!  \"Yes, I feel better,\" I say softly .... slowly ...   \"What was it?  THE FLU? Did you HAVE THE FLU?\"  My brain just fell into my lap, lady.  Couldn\'t have anything to do with your Bose speaker lungs, could it?  \"Yes, I think so,\" I say smoothly ...  \"You know you really SHOULD BE CAREFUL.  I\'ve read reports that THE WEST NILE VIRUS has been detected in King County again.  It starts LIKE THE FLU.  You know THE SYMPTOMS.  My friend in college had it and you know she GAINED SOME SEVERE LEARNING disabilities because it caused her BRAIN TO DECAY.\"  It\'s 7:30 on a Wednesday morning and I might have some Egyptian Death Flu.   Well, at least Speaker Lungs has already blasted my brain out of my head before it can decay and turn me into some carpooling mummy.  \"The kids ATE TOAST this morning.\" '),(3090,1785539,'06,August,2004','2004-08-06 00:00:00','   I return to the rubberhead factory today after stomach flu yesterday. \"I had that a couple of weeks ago,\" says WigWolf, \"It gave me ... the runs.\" Hell Spawn!  I vomit more!   I avoid using words like \'vomit\' \'diarrhea\' \'puke\' \'the runs\' \'Reagan\' and \'vinyl\' when talking out loud. Be like me, rubberheads!  I fell into a bit of a coma this weekend and when I come out ... Reagan\'s dead.  Hm. Bush has got himself some lawyers lined up.  Hmm. This quote came out: \"This administration rejects torture.\"  Hmmm. But then they talk about The Torture Act. Hmmmm.  I am once again going to explain current Washington politics to you dorks by using fast food examples ... namely hamburgers equal torture.  \"Mr. President, some of our troops are eating hamburgers,\" said John Ashcroft. \"Hamburgers? Are we against that?\" ased Bush Jr. \"We should be.  Let\'s make a Hamburger Act outlining when it is right for us to eat hamburgers, when to eat them and how high the ketchup to mustard ratio should be (1:2) and then we\'ll say we are against Hamburgers,\" said John Ashcroft. \"Does that mean we are going to stop eating hamburgers?\" ased Lil\' Bush. \"Hell no.  But if anyone asks we aren\'t responsible for the Hamburger Act.  Just the Yippee Jesus Act and the Boo Homos Act .. all other acts are mysteries,\" said Ashy-Washy. \"I\'m confused,\" said Bushbo. \"It works!  Bush White House Mumble works again!\" said Assfingerer. \"I like pickles,\" said Bushy Boy gleefully. \"We know you do, kiddo,\" said Ashcroft.  Oh, it\'s so silly. Silly, dorks.  From cnn.com (\"Worldwide leader in sucking White House balls\") MTV plans to edit out a shot of (Eminem) exposing his rear end to the audience at the 2004 Movie Awards when the show is broadcast Thursday.  My name is what, my name is who, my name is what, chicka-chicka Slim Played Out. If I want sexism, violence, racial stereotypes and homophobia I\'ll just listen to a White House Press Conference instead of the guy who TRIED TO BEAT UP TRIUMPH THE INSULT COMIC DOG!  That\'s it! I\'m moving to Mars, dorks. '),(3091,1785539,'06,August,2004','2004-08-06 00:00:00','   Happy Birthday, R. Spaghetti Head Angel Wishes! Someone\'s getting a spanking ... '),(3092,1785539,'06,August,2004','2004-08-06 00:00:00','   Did anyone else notice how quickly the major news outlets dropped the \"9/11 Commission Finds No Iraq/Al Queda Link\" story?  Don\'t you think it deserves a little more attention than whether or not Reagan\'s face should be on the million dollar bill? I grow tired of lies and 9/11, dorks.  Three years of  urlLink fear  is enough.  \"They could have  urlLink 9/11ed the Space Needle !\"  Are we really going to let paralell universes scare us now? They didn\'t fly a plane into Safeco Field, get over it.  Did the Mariners win? No? Shit. Witness the  urlLink the greatest photo   ever taken by man or beast. Hahahaha, dorks, now that\'s something to get excited about.  P.S. I forgot to tell you, dorks!  I saw a hobo woman in Seattle\'s U District on Sunday with a sign that read \"Homeless mother yah, blah, rarrgh!\" only the \"r\" was turned around like in a  urlLink Toys \'R\' Us  sign. Why?  It\'s not even that cute when a kid does it, or actually when someone trying to show that someone is young so they reverse the R ... because everyone knows the \"Correct Positioning of the Letter R\" lobe of the brain doesn\'t develop until puberty. Back to my observation of the woman. No? A song? Yes!  I don\'t want a real job. I\'ll just beg for your change. And when I vomit on the street don\'t look at me strange. From crack to smack to Jumbo Jacks I really need to score some I don\'t want a real job \'Cause if I had one I couldn\'t be a  urlLink Hobo  Mom! '),(3093,1785539,'06,August,2004','2004-08-06 00:00:00','   Provided by the management for your safety ... From a pamphlet handed out today:  Would you know who to call if suddenly one OR ALL (!) of the following happen? -The fire alarms in your building start ringing, yet there is no fire ... -The radio is advising all listeners to evacuate the area due to an on-coming force 5 typhoon is heading your way, and your office is in Nebraska. (These are the actual words. \"Would you know what to do if you started writing a run-on sentence for a company handout and then forgot English grammar and the constraints of reality?\"). -A co-worker just slipped on the coffee you spilled earlier this morning and injured his (sexist bastards) knee. -A well-known TV investigative reporter is calling on line one wanting to interview you regarding a problem concerning (our product).  \"Who ya gonna call\" ... The answer is this resource telephone card (attached) ... The subject matter experts answering these lines will be able to help you with these issues and more...  It\'s nice to know that if I ever say, \"Ah, Christ SuperShit!  The building\'s on fake fire, I\'ve been wormholed to Nebraska, there\'s a Super Typhoon headed this way, Bob slipped in a completely carpeted office and has knee parts sticking out of his khakis and now Barbara Walters is on LINE ONE (the MOST DREADED OF LINES!) and wants to know the truth! What do I do? Who\'s I\'m gonna call?\" someone is there to tell me, \"RUN, fucker, RUN!\"  Before my disbelief can wear off ... A Rubberhead has asked me to send and inter-office memo to Denver, Alabama. Should I tell them that a.)Denver is in Colorado, b.)we don\'t have an office in Alabama or c.)nothing, because I\'m 75 percent sure the envelope just contains a crayon drawing of a kitten anyways.  I can only stare at the wall as my college education and soul fade away, dorks. '),(3094,1785539,'06,August,2004','2004-08-06 00:00:00','   The Rubberheads, Soccer Moms and NASCAR Dads are in  urlLink my backyard . They are so angry that the  urlLink Wizard of Oz  isn\'t real, afraid of their wars, and have so much hate for the Lakers that they are coming after 10-year-old children. \"There has to be an enemy!  That\'s what TV teaches us!\" Back Soccer Moms! Be gone NASCAR Dads! I have a zero-tolerance policy for Rubberheads!  P.S. Just snuck out of work to get the  urlLink Beastie Boys\'  \"To the Five Boroughs\" ... 4 tracks in and excellent! '),(3095,1785539,'06,August,2004','2004-08-06 00:00:00','   Today I shipped a phonebook to a psychic. '),(3096,1785539,'06,August,2004','2004-08-06 00:00:00','   Monday! Me: Hello, Mole.  Did you have a nice weekend? Mole: Nope, never really do much of anything ... Me: Oh. Mole: ... well, did some laundry .. Me: Yep, fun (Moving away) Mole: ... and grocery shopping ... Me: Important  Mole: ... and of course I rejoiced to see the Lakers lose. I HATE them ... so much.  When Mole says this his wrinkles become wrinklier and his brow furrows and his face forms into one that man has historically used to say  \"Ghengis Khan stabbed my dog\" and \"Those Nazis!\". Why has the Baby Boomer generation become an increasingly angry group? You see, dorks, first they were stupid hippies and they got hit a lot (actually, first they were neglected of paternal love due to the harsh traditions of non-emotional parent/child relationships their parents were used to), then they were stupid corporate raiders and they lost their souls and now their lives are filled with books like \"Men are from Mars and Need to Not Be Stupid\" and they cream their jeans when Hillary Duff makes believe that their world exists in it\'s pretty white universe that is .... lost myself there for a minute. So, they yell! \"I HATE red lights.\" \"I HATE David Letterman.\" \"I HATE Wendy\'s new salads.\" They used to be mad at their fathers, now they\'re mad at Shaq.  Moving on. So, I got R. Girl a bracelet for her birthday.  Turns out I got it a little small.  No problem, right? Wrong!  Every Jeweler in Town: Yes are you going to need new O Loops or new eye pins? Me: What? Every Jeweler in Town: Do you know the Quick ratio? Me: Um, I have a bracelet.  There\'s something about haircutters and jewelers that makes them think they can speak thier lingo and I\'ll just catch on.  Every Haircutter Ever: Do you want a fade or a crosssnatch? Me: I want to look like I got a haircut. Every Haircutter Ever: For next time remember your gaurd length is 5. Me: NO!  Want some fries to go with your irony? So, I went out with the L.C. yesterday to do some B-Day shopping for R.G.  We dec. to gr. some L. (wait, I can\'t abbreviate everything) so we went to this place for burgers. One of the burgers was named the \"Communication Breakdown Burger\" and I ordered out of Zeppelin Loyalty. Guess what? The cook read the ticket wrong and forgot to make the burger. Get it? There was a \"Communication Breakdown\" and that was the name of the burger. Get it now? Yes? Hahahahahahahaha!  The whole place was rolling when I thanked the waitress for the \"Irony Burger\", which was on the house. Free Burgers = Perfect Weekend, dorks. '),(3097,1785539,'06,August,2004','2004-08-06 00:00:00','   Mighty dorks and people with brains, I have just returned from Fahrenheit 9/11. Bore you with a review? Nope. Go read a paper for America. I will tell you that this ex-ameri-frat boy who once went after an innocent Clemson fan with a six iron because the dude \"was totally hitting on me\" actually got choked up at times.  That means it was a good movie, you looking-for-the-deeper-meaning dorks. I can sum up what I felt after the movie by what happened just minutes ago. As I drove home I realized I was angry at the world we live in and thus needed to smoke.  I had to destroy something and tobacco leaves are the easiest and least federally-crime-punished thing to burn. So, I went into the 76 station near our apartment and got me some Doral Lights in a box.   The old white guy next to me with a \"Something Important to Farmers and Republicans\" hat on starts yelling. \"Dammit, shit! I lost my goddamned lighter!\" Hmmm, hey I don\'t have a lighter either.   So, to the middle-eastern guy behind the counter I say, \"Hey, man, can I get a book of matches ... and give one to this guy so he stops cursing.\" The nice gentleman hands me the matches and a pack to the other guy. \"Nope, I don\'t use matches.  The sparks jump off the match and burn my clothes. (Really?) Matches are the most unsafe thing, I only use lighters.  I don\'t want holes in my clothes ...\" \"You know, cigarettes put holes in your LUNGS,\" I point out to this rambling aged Americana piece of work. The middle-eastern guy behind the counter gets the joke and we share a laugh.  The old white guy doesn\'t. \"Lungs? I don\'t care about my lungs.  I got clothes and they are ...\" \"Take it easy,\" I say and I leave. And somewhere in that story is the point of Fahrenheit 9/11.  The middle-easterners and the intelligent people of America get the joke.  There are no WMDs.  There was no threat from Iraq.  There has been no justice since 9/11.   The people who don\'t care about that and who blindly follow their leaders and their inanerubberist philosophies (\"Building a safer world through war\"/\"I don\'t use matches\") are the rubberheads who are more concerned about the guises they take on (Soccer Mom, NASCAR Dad, Compasionate Conservative, Neo-Patriot) than their lungs or the hearts of the Moms who cry because of soldiers who die. Put on your rubberkicking boots, dorks, it\'s time to save some Iraqi babies. '),(3098,1785539,'06,August,2004','2004-08-06 00:00:00','   I can\'t decide if the highlight of my day was Kyle buying me a hamburger at the Microsoft cafeteria or the fact that I came back to work after that and took a nap in the supplies closet. '),(3099,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink This , dorks, my doctors ball-grabbing prognosis that I am \"fit as a fucker\" after yesterday\'s check-up and my new urlLink  red Nike shorts  that I bought on my lunch break have combined like Voltron to make today a heroically great day.  \"This administration does not condone torture.\" \"Pssst, President Bush, condone means to like allow or overlook.\" \"Really? I thought I was just making up another fancy Yaleish word to confuse people.\" \"No, this time you made up a sensible lie.\" \"Shoot it all to Fuck Swamp ... um, what I meant to say is that this administration does not perpitify torture.\" \"That\'s our, Pinochio Prince!\"  Nike shorts, boogers, no testicular cancer, zen.  My two favorite things to yell at horrible drivers: -\"Are you alive?\" (I constantly suspect many zombies have learned to drive.) -\"Asslicker!\"  (\'Cause nobody expects to be called an asslicker.)  Two things R. Girl says I have yelled at horrible drivers that I don\'t remember: -\"Go fuck your mother.\" (In my opinion worse than saying \"Motherfucker!\" because it\'s an order not a noun word.) -\"Give us justice!\" (I don\'t remember ever saying this, but I think it\'s the coolest thing I\'ve ever said.  This will be written on my tombstone, dorks.) '),(3100,1785539,'06,August,2004','2004-08-06 00:00:00','   From  urlLink cnn.com   Implicitly, (Republicans are) also suggesting (Kerry\'s) out of touch with Americans. One Republican operative told the New York Times last year that Kerry  urlLink \"even looks French.\"   Looks French?  Is this seriously their campaign tactic?   \"Bush/Cheney 2004: A circus monkey + an angry dough pile = Not French looking\" Dorks Don\'t Rock presents ...  urlLink FIND!  urlLink THE!  urlLink FRENCH!  urlLink GUY!  Just click on these four links and guess which of these dudes is the war-hating, bread-liking, pant-wearing Frenchie for a chance to win valuable or non-valuable prizes or nothing!  Please, dorks, beat your television senseless when it tells you to vote for the \"least French\" candidate!  There\'s a surplus of televisions.  Put Chinese toddlers to work. '),(3101,1785539,'06,August,2004','2004-08-06 00:00:00','   As I escaped from the rubberhead factory today to get my truck\'s wheels aligned I drove up on some worker guys fixin\' the road. Orange barrels, vests and all. As I approached, the guy holding the \"Slow\" sign held his  urlLink hand paralell  to the road ( urlLink perpinicular to the road  being the well-known signal for \"Stop\" or \"Put your hand against mine to celebrate our football game victory or the completion of our  urlLink Pine Wood Derby Car .\").  As I was going 45 m.p.h., I decided to heed his advice. 35 m.p.h. ... his hand is still up, slow down more. 20 m.p.h. ... still up ... more brake. 10 m.p.h. ... hand up, foot down, time slowing 5 m.p.h. ... is this .... 2 m.p.h. ... guy serious ... Stopped. The guy nods his head to let me know I was now traveling at the proper speed, 0 m.p.h., the speed of rocks and pens.  I had reached Absolute Slow, dorks. My jaw dropped and my stare went blank to celebrate the occasion. '),(3102,1785539,'06,August,2004','2004-08-06 00:00:00','   Ah, dammit all, dorks. I\'ve got that  urlLink \"Ding Dong the Witch is Dead\"  song stuck in my head for no reason (maybe it\'s because I read  urlLink this comic  ... but that was months ago).  No witches have died that I know of! I don\'t even have a doorbell. For the last half hour it\'s been stuck on the \"Which old witch? The Wicked Witch!\" part.  Ding Fuck! Get out of my head. Which old head? MY fucking head!  Does it get worse? Yes! Sometimes it overlaps with the \"He\'s a Fucking Genius Wizard\" song. So at times my stereo brain is playing this fun hybrid.... \"Ding Dong the Witch is dead, because, because, be-cause, because, because, be-cause, because, because, be-cause .. The Wicked Witch is DEAD!\"  Welcome to the Jungle we got fun and games we got everything thing you want hmmla la the wicked witch is ..  Dammit, dorks! '),(3103,1785539,'06,August,2004','2004-08-06 00:00:00','   Wetbackorrists, Radical Nigics! So what if I\'ve been drinking?  Dammit, you deserve a drunk entry, dorks. I just finished watching an episode from my  urlLink Superman  DVDs called  urlLink Japoteurs . In it a radical Japanese man tries to fly a plane (!) into Metropolis.  He is dubbed the Japoteur, because Slant Eyed Bob and Nagosammy were already taken. In 62 years have we made progress? Hahahahahaha, dorks!  Just ask the next \"Radical Cleric\" you meet. '),(3104,1785539,'06,August,2004','2004-08-06 00:00:00','   Two highlights from my drunken Seattle-to-Raleigh conversation with Kurtis \"Baby Got\" Backs, the best drinking buddy I\'ve ever known.  Highlight #1!  Me: So, you read my blog, is it good?    Kurtis: Yeah, I have to say I laugh pretty often, except when you tend to over-politicize.  Me: What do you mean?   Kurtis: I mean I can\'t believe you like Michael Moore.  Me: I never said I like Michael Moore.  I think that it\'s important that he\'s putting out a message that the average American public has not heard before.  Kurtis: I guess.  Me: It\'s like Jesus.  Did you know Jesus never washed his hands after using the bathroom?  He\'d be like \"I need to pee.\" Then just say \"Hey, let\'s break bread.\"  People still liked him.   It\'s not the man, it\'s the message.   Highlight #2!  Me: Yeah, so I told R. Girl that when I die that I want my tombstone to say: Brent Kinkade Son. Brother. Husband. Father. Zombie?  Kurtis: Good one.  Me: Do you think it should have a question mark after \"Zombie\" or should it be a period?  Kurtis: How about some elipses?  Like, \"To be continued\".  Me: Perfect!  It\'s like I could come back as the Zombie King!  P.S., dorks, as much as I think Pete Yorn and The Smiths\' \"Panic\" kick ass seperately ... together they suck harder than a whore. Also, on Hwy-167 last weekend I saw a redneck driving a truck with a witty Calvin-peeing-on-something bumper sticker. What was it? \"City boys\" The redneck was peeing on city boys.  The irony is ... you have to usually find a city boy to find someone into that. '),(3105,1785539,'06,August,2004','2004-08-06 00:00:00','   I have tried in the past 18 months since leaving the drunken comforter of college to use Monster.com to my full advantage. Monster has sent me links to many exciting non-me career opportunities, including: Soldier, Welder, Pet Technician, President of Mobil Oil, Boat Welder and, of course, Carpet Welder.  Oh, Monster, truly today\'s the day for you to show me what I need!  Kyle, my brother and computer Khan at Microsoft, received an e-mail from Monster Networking (\"Today\'s the day for you to meet welders in your area!\") \"We\'ve got someone we\'d like you to meet!\" Who? Who? Why,  urlLink Robert G. , of course.  Who is Robert G.?  Just ask him!  \"60some writer, marketing/sales/PR person, fit and curious. I am in love with an Asian woman whose pop name is Emma.\"  This is the kind of guy with a hottub any Microsoft programer should know right? Read on, dorks!  Robert G.\'s Skills?  \"written/verbal communications, marketing/sales, le (Expert)	writer, philosopher (Beginner)\" He\'s a beginner philosopher, \"I think, therefore I ... um, need to go hottubbin\'!\" Also, he\'s a beginer AND an expert at writing!   Very zen, Robert.   He and Emma must discuss the inability to ever become an expert writer ,and as such, all one can really hope to achieve is a continued rebirth of the creative spirit.    Robert G.\'s Interests? Hold on to your universe dorks!  \"music, books, biking, hiking, mountains, beaches, good & straight sex, cuddling, sunsets\"  ....... what?  Oh, I see, Monster, the reason Microsoft has been having major security problems over the past year is because of lack of cuddling!  Of course, they aren\'t enjoying sunsets or \"good & straight\"-yet-curious sex!   Today\'s the day that Monster saves the world again, making it better for underemployed journalism majors, highly paid Microsoft engineers, Robert G. and of course the great American hero ... the welder! '),(3106,1785539,'06,August,2004','2004-08-06 00:00:00','   They have  urlLink TEEN NUDIST CAMPS ?!?!? All of those Friday nights in high school that I stayed home and watched  urlLink \"Boy Meets World\"  and thought \"I bet everyone is out having crazy sex and seeing  urlLink girls\' boobs \"  ... I was right!  Was I right about other things? Is  urlLink Starship Troopers  really the greatest movie ever? Does East Henderson High School\'s  urlLink basketball  team really suck harder than a whore? Should  urlLink 311  really run for president ... that\'s right I thought the entire band of 311 should be president! Was pops really straight trippin\', dorks? '),(3107,1785539,'06,August,2004','2004-08-06 00:00:00','   Brent\'s Top Ten Things to do to Keep Himself Entertained While R. Girl is Having Fun in Las Vegas  10.) Recreate the door kicking-in scene from Walking Tall (the  urlLink Joe Don Baker  one) every time I enter a room. \"You got a warant?\" \"Yeah, I keep it in my shoe!\" (Kicks in door, finds bourbon, rejoices! .... later gets shot and wears Darth Vader mask.)  9.) Have my yearly \"If I drink Zima alone noone will laugh at me\" Night.  8.) Find a slug on my porch when I\'m drunk.  Wonder at what point in the evelutionary process did slugs say \"You know what? I think I\'m good with the slime and the crawling.  You guys go on ahead.\"  7.) Decide that not wearing a shirt under my button up shirt is too sexy to pass up.  6.) Repeatedly get frustrated at Hitman: Contracts because I can\'t make it past the first level.  Just because I\'m trying to put a syringe into a SWAT team member\'s neck or just because I decide to use piano wire to choke a mental paitient in front of a horrorified orderly does that mean I have to get shot?  5.) Sing ... my own versions of songs! \"Well, I tip my hat to the big washing ma-chine And I say hello to the coffee ma-a-ker Smile because I\'ve got nuggets to eat. So I pick up my nuggets and say, \'Who wants nuggets today?\' Then I open up the oven and, HEY! I\'m eating chicken nuggets again!\"  4.) Catch up on my  urlLink reading .  3.) Call R. Vegas and ask what new presents she has bought me since the last time we talked!  2.) Get excited about presents!  1.) Stumble around Blockbuster for an hour trying to decide between \"They Live!\" or \"Naked Ski Hookers!\". '),(3108,1785539,'06,August,2004','2004-08-06 00:00:00','   R. Lucky left for  urlLink Las Vegas  yesterday.  According to her, our months of piggy bank savings have blown away in the desert.  I was hoping I\'d be bedding down with a millionaire when she got back, but I\'ll just settle for a derriere.   She told me about her first day of  urlLink liquour  and one-armed bandits while I made a quesadilla at 1:00 a.m. because neither of us could sleep. For the next few days I\'ll stumble around the apartment, forgetting pants and when to eat.   I found an interesting  urlLink on-line journal  from a soldier in Iraq. If the sucker\'s for real then our souls are fucked.   A story. When I was five-years-old or so my family went to Wisconsin Dells, which is for midwesterners who feel Disney isn\'t cheese-focused enough. One of the proudest moments of my life was when I got to drive one of the  urlLink Wisconsin Dells Ducks , a land/water hybrid tourist mover. I thought I was really driving the boat and loved every minute of it.  As I swerved the large boat around the captain told me I was in danger of wrecking it on the rocky cliffs.  But, I couldn\'t let that happen and always straightened the boat out in the nick of time to save my precious crew.   I thought I was a natural boat captain because it was all so easy.  I wanted to be a duck captain when I grew up. But then I did grow up.  I realized I wasn\'t in charge of anything.  I might have had my hands on the steering wheel, but one wrong move and the real captain would have yanked me out of control.   I can\'t say it was heart breaking to find out I\'m not a natural river boat captain, so don\'t cry for me Dorkentina. But it is sad that so many people will think that today\'s \"hand over\" of Iraq sovereignty is real, that the post-toddler steering the country is the real captain. But one wrong step and they\'ll lose their grip on the wheel. The  urlLink truth  is that nothing has changed, dorks.  The war is still going on.  Four or five guys signing a piece of paper doesn\'t stop bullets from going through soldiers.   But play with your MTvs and watch your Olsen twins dance, dorks.  Close your eyes and let it slip by, because it\'s not happening to you.  You aren\'t dumber or a zombie.  You\'ll get mad when they show up on your front door and take away your rights, right?  That\'s when you\'ll be upset. And then you\'ll be in a nice, quiet  urlLink Cuban ass-prodding camp. '),(3109,1785539,'06,August,2004','2004-08-06 00:00:00','   Cashier Woman: \"What\'s this?\"  Stupid Man: \"That\'s a tomatilo.\"  Cashier Woman:\"No way!\"  Stupid Man: \"Yes. I should know what a tomatilo is, I\'m from the South.\"  Cashier Woman: \"You better not be fooling me, I don\'t give you boys any special treatment at the fire house.\"  Top Ten Things I Should Have Said to The Chatty Rubberheads In Front of Me at The Grocery Store Delaying My Weekend Camping Trip:  10.) \"Hey! Buy one \'shut\' get one \'the fuck up!\' free.\"  9.) \"Hey! Can we get a price check on Bald Fuck and Chatty Skank?\"  8.) \"Hey! This week\'s special is life at an acceptable pace!\"  7.) \"Hey! Fuck up on ailse Bald Ass.\"  6.) \"Hey! Would you like some help out with ... these fists?\"  5.) \"Hey! It\'s paper or plastic; not paper, plastic or babble.\"  4.) \"Hey! Would you like to sample some mouth closing?\"  3.) \"Hey! I must slay you and have revenge for my father\'s blood.\" (Just to keep them on their toes)  2.) \"Hey! Your chat has angered me!  I shall summon your lord the Great Shift Manager to smite you!\"  1.) \"Hey! This line is 10 spoken words or less.\"  Weekend, dorks, I\'m going camping. '),(3110,1785539,'06,August,2004','2004-08-06 00:00:00','   Iraqi mother:  Why did your tank blow up my baby?  U.S. soldier:  Why am I missing a leg and suffering from nerve damage?  U.S. Taxpayer:  Why am I expected to bear the cost of a multi-billion dollar war?  Official Senate Intelligence Committee answer:  urlLink \"Group think.\"   Ooooh. Gotcha.  (Pssst, dorks ... \"This \'group think\' dynamic led Intelligence Community analysts ...  to both interpret ambiguous evidence as conclusively indicative ... as well as ignore or minimize evidence that Iraq did not have active and expanding weapons of mass destruction programs\" is a fancy way of saying, \"Liars!\".)  \"I think the American people - I hope the American - I don\'t think, let me - I hope the American people trust me.\"   -Your President & Savior Dec. 18, 2002 '),(3111,1785539,'06,August,2004','2004-08-06 00:00:00','   Hwy-167. Rush hour yesterday.  HOV (to the izay!) lane. \"So, the kids have these shoes and thier birth mother doesn\'t make them wear them and so their feet get dirty and I say \'Dirty feet?\'.  I mean come on, those are feet we are talking about, not potatoes ...\"  \"Hmm, sound like a couple of winners. Hey, look, our exit, time to merge ...\" I find a spot and move in.  HONK!!!! - HONK!!!! -HONK!!!! -HONK!!!!  I look in my side mirror and see some cow girl driving behind me making the universal  \"What\'s your problem\" face with accomodating hands in the air gesture.  \"What, you big (fat) girl (bitch)?\" I yell back. \"It\'s called driving!  People merge!  Guess what?!?!? The scenery changes too!  Go home and eat a pie you cow!\" I then proceed to mock her dumb face back at her in my driver\'s side mirror. \"Huh? A der!  What?  Who driving?  A hee dee huh-er ....\"  When I\'ve had my fill of fat-girl taunting I look over and see Hippie Chick slack-jawed and saucer-eyed looking at me.  Shit, forgot she was there.  This is the first emotion of mine she seen since \"Ah, sure, yep, uh-huh\".    \"So, tell me more about your lovely children,\" I say.   Yesterday.  Kidco.  Grocery shopping.   Thinking to myself:  Just put the cart over here.  Hey, look, orange juice drink.  That\'s not like real orange juice.  It\'s just water and powder.  Or just dye.  Mexican people like it a lot.  I always see them drinking it at the park or buying it at stores.  Look, those two Mexican guys just bought 4 gallons.  Wow.  I wonder if that stuff sucks as much as Chocolate Drink.  A knock-off Chocolate Milk?  Gross.  Is Mexican guy the correct term?  Are they Latins?  I know it\'s not Chi....  \"Brent!\" R. Girl says boldly behind me.  \"Hey, babe, what d\'ya got there?  Some half&half?  Mmmm.\"  \"I said your name three times and you just kept on walking.  You just walked over here and parked the cart and stood while I said your name louder and louder.  The entire grocery aisle turned around and looked except you!\" she says.  \"Really?\" I ask.  \"It\'s true,\" says a lady passing by.  \"See?  She looked, you walked,\" says R. I-would-have-thought-used-to-it-by-now.  \"Is your name Brent, too?\" I ask the lady as she walks on.   Later, dorks. '),(3112,1785539,'06,August,2004','2004-08-06 00:00:00','   Two short stories about food from yesterday (One mine and one R. Girl\'s)  My food story!  Kyle is notorious for keeping a messy car, room, home and pair of cargo pants. If there is space to be filled Kyle will fill it with something. That something is likely a trial AOL C.D. But not yesterday. After going out to lunch yesterday Kyle and I decided to spend the afternoon buying soap. To save time and the world gas, Kyle suggested we take his car. The little red Miata, described by some as \"a tube of lipstick on wheels\". Described by me as \"a car full of shit\". \"I\'m not going anywhere until you get some of this trash and stuff out of the passenger seat, dude,\" I said. So, Kyle started pulling trash out. Paper. Candy bar wrappers. Empty KFC Popcorn Chicken boxes. Broken CDs A rock. And of course ... a half-eaten whole chicken. I still rode in the car after we buried the fowl.  R. Girl\'s food story!  At work yesterday R. Girl and her co-worker ordered some teriyaki.  They waited and waited for the food to arrive, being unable to go pick it up due to the having to watch the kids on ventilators. Finally a small, old, rude Asian man showed up bearing gifts of rice and egg rolls. \"20.22\" R. Girl hands him a twenty-dollar bill and a quarter and starts to shut the door. Suddenly the old man pushes the door open and yells, \"This total does not include my tip! No tip for an old man? Where is tip?\" \"Well, this is the reason you aren\'t getting one,\" replies R. Witty. \"I guess it\'s going to be a bad day,\" says the rude Asian. \"For you,\" she says and closes the door. The food was good.  Two signs. One real, one in my head, one bad, one good.  The name of the coffee stand next to where I had lunch with Kyle yesterday: \"Drink & Drive\" Drive Thru Espresso Stand  The name of the price-friendly gyro restaurant I want to open in protest to the $10 gyro platter (beef gyro and fries) at Gyro World: \"Christ, They\'re Just Gyros.\" '),(3113,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    Wha...?  Violent Stick People '),(3114,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    Ha!  Violent Stick People '),(3115,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    Nice batting.  Violent Stick People '),(3116,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    What the hell ... here\'s my week\'s worth of work I have done.  I\'ll still post these on Wednesday\'s ... I just figure at my current rate I\'ll have close to 30 of them done by then.  Enjoy, dorks.  Violent Stick People '),(3117,1785539,'06,August,2004','2004-08-06 00:00:00','   Last night at the Mariners game against the Indians:  ( urlLink Section 143, Row 13, Seat 13)   Bottom of the 7th inning.&nbsp; Mariners at bat.&nbsp; 2 outs.&nbsp; Cleveland leads 1-0.    urlLink The Mariners\' Moose  is getting the crowd pumped with his \"Raise the Roof\" and \"I can\'t hear you\" antics.  The problem is, this is the most boring baseball game ever played.   The Moose orders us to raise the roof.  \"No, Moose.\"   The Moose orders us to make some noise.  \"Should we go get a hotdog.\"   The Moose says that he can\'t hear us.  \"Let\'s go to the Team Store.&nbsp; I need to buy a gift for the guy who gave me these tickets.\"   Suddenly, the Moose throws his \"NOISE\" sign on the ground and goes after a fan, arms raised making the universal mascot gesture for \"Do you want to fight a mascot?\"  \"Shit, look, the Moose is picking a fight.\"   The Moose regains his composure and his sign.&nbsp; Make some noise!  Suddenly the entire place gets on its feet and starts to yell and cheer.  \"Oh my god,\" I think for a split second. \"People have started to revolt against performers who say they can\'t hear them.&nbsp; I\'m witnessing history.\"   Then I realize that the&nbsp; urlLink game winning home run  has been hit.  That\'s exciting too.   Mariners snap their 9-game losing streak.&nbsp; I continue my 6-game Mariners-winning-at-every-game I attend streak.  Weekend! '),(3118,1785539,'06,August,2004','2004-08-06 00:00:00','   Please, dorks and rubberheads, urlLink have a brain in November! '),(3119,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    Here\'s another Violent Stick People cartoon for you dorks.  I plan to start publishing them every Wednesday (today is an extra special treat!) even though I could pretty much turn out about 8 or 174 of these things a day.  Violent Stick People '),(3120,1785539,'06,August,2004','2004-08-06 00:00:00','   \"Violent Stick People\" is a cartoon I used to do in high school and college history class. I think I\'ll bring it back, thanks to technology, boredom and you dorks\' insistence that I do more with my site. '),(3121,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    Violent Stick People  Violent Stick People '),(3122,1785539,'06,August,2004','2004-08-06 00:00:00','   Welcome back to typing, fingers, and to reading, dorks. Is it just me?  Does the world that I live in happen to other people? Why?   ! Last night I decided after a couple levels of Driv3r and some tacos, \"I need to just relax tonight.  No blowing up ninjas or hunting down prostitutes.  I will lay around and chat with R. Girl.\" Hours later ...  \"Really?  I didn\'t know you liked peanut butter.  The things you learn when the XBox is off ...\"  Ring, Ring goes my cell phone.   \"It\'s Kyle calling,\" I say. \"I\'ll ignore it ... tell me more about this \'vacuuming\' thing that you do.  Is it easy to pick up or do you need some training, like on Driv3r?\"  Ring, Ring goes R. Girl\'s cell phone.  \"It\'s Kyle,\" she says.  \"Hmm, it must be important.\"  I answer.  \"What\'s up?\"  \"So, I just got off the phone with Mom.\"  \"Yeah? Great, look I\'m kind of ...\"  \"Yep.  Apparently she\'s engaged ... again.\"  \"What?!? To who?  What\'s his name?\"  \"I\'m not really sure.  He has his own lotion business.  I didn\'t get his name because he was going on for half an hour about a foot lotion I should try.\"  \"What?\"  \"Yeah, apparently many adult men have major problems with their feet.  It\'s not uncommon.\"  \"The vital information, Kyle!\"  \"Oh yeah.  He has ten kids.  He said he\'s going to call you tomorrow and tell you all about it.  Don\'t let him talk about his lotions for too long, though.  He just keeps going and going.\"  \"This is too much.  I need to go.\"  What? Is? My? Life? Can\'t I just have one day to stop and smell the tacos without a family member flying through a window, a zebra cutting me off walking down the street or a crazy old man screaming at me \"You\'re a robot! A robot!\"? All of those things have happened to me and that is my life.  And now I have to welcome a new guy (who I have already nicknamed Newt Tragena ... get it?) into it.   Bound on, wackiness!  Bound on!  Oh, dorks, and ... During me and R. Woodsy\'s improper-English-yet-gloriforous camping trip we needed to go \"into town\" to get batteries for our flashlights. \"I don\'t need the Titanium batteries, Cashier Lady, just give me the plain ol\' Duracells.\" \"Are you sure?  These Titanium ones are good, especially if you are going to be using them  in anything electronic .\" \"Oh, really? Well, I was just going to use them for pancake toppings, but ... shoot, I\'ll give them ee-leck-tronick batteries a try for!\" (Banjos play and rubberheads sway.) '),(3123,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    Here you go, dorks.  I want you to have a great weekend so enjoy these and drink lemonade!  Violent Stick People '),(3124,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    Ponder this one.  Violent Stick People '),(3125,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    A cartoon about the 2000 election ... and leg ripping!  Violent Stick People '),(3126,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    Yes!  Violent Stick People '),(3127,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    Todd?  Violent Stick People '),(3128,1785539,'06,August,2004','2004-08-06 00:00:00','   Showing sign that my co-workers think I\'m a smelly 50-year-old woman:  Today I received an overnighted package from Denver whose&nbsp;contents were only a&nbsp;pink piece of paper with my name on it, a small yellow  urlLink tote bag  and four tins of breath mints.   Showing sign of N.P.R.\'s aloof eliteness:  \"The Mariners won last night 4-2 against the Oakland Athletics.\"   Showing sign that cnn doesn\'t live in the world you and I live in:   urlLink Wow!    A very Brent moment, from last night:  (Beer-buzzed at the grocery store purchasing a six pack of  urlLink beer , a pack of  urlLink cigarettes  and a cobb salad)  People in front of me in line.  Blurry Blonde: She what?   Blurry Man: She lost her lottery ticket.   Blurry Blonde: Did she win?   Blurry Man: She doesn\'t know.&nbsp; She lost the ticket.   Me: (Feeling I could contribute to the conversation) I lost a fish in college.   (Blurry Blonde and Blurry Man look at me.)   Me: (Shrug shoulders) I don\'t know how to take care of fish.   &nbsp;  A very non-Brent moment, from yesterday:  \"Hey, look, a pick-your-own lavender field!&nbsp; Let\'s stop!&nbsp; Hey, they have corn!\"   Weekend, dorks. '),(3129,1785539,'06,August,2004','2004-08-06 00:00:00','   Some days my job makes me want to roar:      Mon 7/19/2004 8:39 a.m.  To: Brent  From: Egg Belly   Thanks for delivering the large boxes to my area.   Next time could you please set them just outside my cubicle, I could hardly get in this morning. There should be room just to the left of the doorway (as you stand facing my cubicle) of my cubicle and the boxes could be stacked close to the cubicle wall and be out of the way until I return.   Thanks,  Egg Belly     My bad, Egg Belly!&nbsp; I forgot to do fatty math ... fatmatic!  One office cubicle:&nbsp; 5 feet long, 4 feet wide = 20 square ft. of floor space  5 stacked boxes: 1 foot by 3 feet= 3 square ft. of floor space  One beast: 4 feet long, 3.5 feet wide = 14 square feet.   I only left 3 square feet for your ham and Twinkie pie shake.&nbsp; I\'ll remember that next time.   And!    Mon 7/19/2004 7:14 a.m.    To: Brent    From: Crumble Cheeks     &nbsp;   Good god! From what I see here Brent, authorized to have you order. Double check for me making sure, as well.  Thanks again, Person.     &nbsp;   I have read this e-mail about 42 times.&nbsp; I give up. &nbsp; I am, unable to understand ... brain falling, double check, as well making, sure.  But sometimes I enjoy my job:   \"I wonder what\'s behind those file cabinets.&nbsp; Let\'s find out.\"   2 minutes later.   \"Is that a .. YES, IT IS!&nbsp; A yo-yo!!!!\" '),(3130,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    This is a more direct version of the below DDR-PSA.  Please post this across major metropolitan areas.  Violent Stick People '),(3131,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    I can\'t wait \'til tomorrow to post these!  I\'m so excited!   Violent Stick People '),(3132,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink       Violent Stick People '),(3133,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink       Violent Stick People '),(3134,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    A public service announcement from your friends at Dorks Don\'t Rock.  Violent Stick People '),(3135,1785539,'06,August,2004','2004-08-06 00:00:00','   Rubbersations for a Tuesday.   Me: Oh, hello, Disproportionate Torso (I really need to learn people\'s names).  Disproportionate Torso: Hello, Brent.  Me: Hey, did you get a haircut.  Disproportionate Torso: Actually, I got them all cut!  Me (Outside): Oh, ha ha, the glory of a joke! Me (Inside): I can feel my brain unlearning!  Another one.  Me: Hello, Told Me About Your Vomit Once, having an exciting day?  Told Me About Her Vomit Once: No.  I\'m soooooooooooooooo bored.  Me: That\'s a lot of ohs ... do you have a book to read?  T.M.A.H.V.O.: No!  I hate to read!  Me:(Outside) That\'s a shame. Me: (Inside) Ewww, she puked.  Last night.  The world was calm.  I was grilling a steak for dinner.  R. Girl was working late.  The world was quiet.  Children slept.  Old men remembered battles and victories.  Meat cooked.  The world was still.  Ring. Ring.  \"Hey, Mom, how\'s it going?\"  \"Well, OK, I just wanted to call ...\"  \"I\'m making a steak.\"  \"...becasue I wanted to let you know that ...\"  \"It\'s big.\"  \"... last night I came to the decision that ...\"  \"Sizzle.\"  \"... I was going to give the ring back to Newt.\"  \"Oh, yeah?  Why?\"  \"Well, it was all happening so fast and I was worried about his intentions.\"  \"Gotta look out for yourself.\"  \"Yeah, I\'m better off.\"  \"Yep, and my life is back to the way it was eight days ago. We all win!\"  \"I\'ve actually already got a couple of dates set up for this weekend.\"  \"I\'m making a steak.\"  \"Your mother rebounds really well, Brent.\"  \"That\'s great!\"  \"One guy was asking me if I make really good spaghetti.\"  \"Oh.\"  \"I told him yes.\"  \"Oh.\"  \"He was like \'I mean really, really good spaghetti.\"  \"Oh.\"  \"I was like, \'Yes\'.\"  \"O..\"  \"He was like, \'I mean really, really good spaghetti.\"  \"...h.\"  \"I was like, \'Yes\'.\"  I learned a lot about myself that night.  I learned I could listen to my mom talk about her spaghetti for hours and that I could not eat a full pound of rare steak without getting a severe stomach ache.  Most of all, I learned that losing a little piece of your innocence just means you\'re gaining a big piece of maturity.  I would never see that steak again.  He went off to Vietnam and would be dead a year later.  Mom?  She\'s now a full-time spaghetti cook.  To this date she has made over 2,500 tons of spaghetti for the single men of Phoenix.  They all claim it\'s \"really, really, really, really, really good\".  Newt\'s lotion business went bankrupt a few months later.  All along he had just been putting sour cream in old toothpaste containers and passing it off as lotion.  It turns out that one of his ten children, Lil\' Sue, had an addiction to Mexican food condiments.  She wiped out the entire family business during a Christmas Eve bender.  Police described the scene as \"The stupidest thing we\'ve ever seen.\"  Hobo Bob went on to beat  urlLink Slinky Boxcar  in that shiv fight.  Bums around the world would talk about the legend of Hobo Bob and his \"Steel of Anger\" for years.    And me, well, I went down to the tanning booth and later became a gorilla poacher.   Still, today I look back on that night ... in wonder. '),(3136,1785539,'06,August,2004','2004-08-06 00:00:00','   Did \"Red Heat\" really just call Democrats  urlLink \"girlie men\"?   Yes?  Oh, dorks, this guy is whacky.  What makes you a non-girlie man?  Killing kittens at a  urlLink faster pace?   or...  Running around with your  urlLink shirt off&nbsp; &nbsp;with&nbsp; urlLink Wilt Chamberlain  fighting urlLink  this guy ?  &nbsp;  And seriously, Arnold, a Hans and Frans quote?&nbsp; I know they were \"doing\" you, but come on get a little more current.&nbsp; And while you\'re at it, be a little kinder with your words ... that is, let\'s see a little more Spartan spirit.  &nbsp;  I say who\'s that sticking up for the G.O.P.?  It\'s me!&nbsp; It\'s me!  I say who\'s that sticking up for the G.O.P.?  It\'s me! It\'s me!  Uh-huh, uh-huh, uh-huh, uh-huh, uh-huh ...  FEMINIZE DEMOCRATS!  &nbsp;  &nbsp;  Also, dorks...  keep up with my kickball team\'s record  urlLink here .  We won last night ... yours truly being the starting pitcher and the guy who just happened to make the winning out, you silly girlie dorks. '),(3137,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    Ramones Girl ... the girl that makes me smile in the thick of rubberheads and ninjas.  Why \"Ramones\"?  The mystery continues, dorks. '),(3138,1785539,'06,August,2004','2004-08-06 00:00:00','   Me: What should I say about today being the first year anniversary thing of you and me meeting?  &nbsp;  R. Girl: I don\'t know.&nbsp; Say I\'m the coolest person in the world.  &nbsp;  R. Sticking-by-me-even-though-I-get-grumpy-and-smelly is the coolest person in the world.  &nbsp;  To celebrate the occasion we\'re  urlLink eating noodles  at the place of our first date.&nbsp; Yum, wonderful dorks, yum. '),(3139,1785539,'06,August,2004','2004-08-06 00:00:00','   \"They only poll people who have consistently voted in previous elections. ... If they wanted to be honest, they could say it\'s a 50/50/50 country because they never ask the other 50% how they feel. And I got to tell you, this is what they are in for a big surprise. Come November 2. ... \'it\'s a 50/50 country\'. Like if they just keep repeating it enough, it will be true. \'It\'s a 50/50 country. Put your heels together now. It\'s a 50/50 country.\' ... It ain\'t a 50/50 country. People are angry.\"  -  urlLink Michael Moore  on the  urlLink \"liberal\"  media  Be angry, dorks! Be American! Vote with your flag!   And don\'t forget to buy  urlLink Violent Stick People  shirts!   AND! According to the internet, there is a  slight  possibility that I am  urlLink pregnant . If I am, the internet tells me that I have been pregnant for 38 weeks.  That means I\'m not only slightly pregnant, but that I gave birth 2 weeks ago and didn\'t notice it. Anyone seen a baby? '),(3140,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    Here you go, dorks.    Violent Stick People '),(3141,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    A commercial!  Violent Stick People '),(3142,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    A plane!  Violent Stick People '),(3143,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    A bomb!  L.F.O.!  Violent Stick People '),(3144,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    More!  Violent Stick People '),(3145,1785539,'06,August,2004','2004-08-06 00:00:00','   Ah, yes ... I write this post as R. Girl gets her helping-the-kids sleep a few feet from me. I don\'t dare type faster than a wheelchair olympics.  Have you dorks ever been addicted ... to a song? I am! I am! The song is \"Invalid\" by Tubring. Why? Because it reminds me why rock rocks. Wanna hear? Just click on the  urlLink Invalid  link and rock out. Nevermind the pop-up ... you didn\'t win an I-Pod. '),(3146,1785539,'06,August,2004','2004-08-06 00:00:00','   Temporary fad or the  urlLink next Nike ?&nbsp; You decide! '),(3147,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    A comic blogger be me.  Violent Stick People '),(3148,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    It\'s tomorrow\'s comics ... TODAY!    Violent Stick People '),(3149,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    Punk\'d is for retards and retards.  Violent Stick People '),(3150,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    Nutty.  Violent Stick People '),(3151,1785539,'06,August,2004','2004-08-06 00:00:00','   The Democratic National Convention started today.&nbsp; Let\'s talk about the free speech zone, dorks.&nbsp; Or can we?&nbsp;&nbsp; I mean, what is this ....  SHUT UP, ME!  It\'s time to talk about porn!   A classic Brent porn story for the ages:  I was adventurous one night in college and decided to rent a dirty movie.&nbsp; I don\'t remember what it was called, something like \"Amazon Strap-On Golden Girls and Their Bitch Slave Boys Do Six Flags\".&nbsp; Whatever it was called it was dirty enough to make the \"climax\" (ha) of this story what it is.  I returned the movie the following day because porn rentals aren\'t the Blockbuster 5-day gracious ones (and we all know that no matter how exciting a porn is on the first day, by the second it\'s just 17 naked girls and an elephant).  But, wait, where is the pornography store?  Oh, it burned down.  I get out of my car, stunned to see a place that just a day ago was so alive with the aura of&nbsp;lotions, boots&nbsp;and 19-year-old strippers now reduced to a charred pile.  \"What happened?\" I ask a fireman/inspector.  \"It was an electrical fire,\" I\'m told.  \"When?\" I ask, still stunned.  \"Last night ... they asked us to hold all rental returns.&nbsp; I\'ll take that,\" the fireman says.  It\'s at that point I realize I am&nbsp;holding my copy of \"Amazon Strap-on Golden Girls and Their Bitch Slave Boys Do Six Flags\".&nbsp;&nbsp;&nbsp;And now a man who I\'m supposed to be worshiping in this post-9/11 world is taking it from my hands.  There\'s a moment soon after he glances at the title and before I run away, run away, run away that we look at each other.&nbsp;  I feel like I should say \"God Bless America\" or \"Never forget\", but instead I just run away, run away, run away.   A line heard this weekend that is truer than any I\'ve heard before in porn:  (Girl being violated by two guys) \"I should have stayed in school.\"   Let me tell you, dorks, this girl\'s words made me consider law school more than any ad campaign or family member\'s advice.  She should be the new poster girl for the anti-dropouts.&nbsp; '),(3152,1785539,'06,August,2004','2004-08-06 00:00:00','   Phi Kappa Nasty fans, your days of woe are no more!  You can now purchase the reasonably-priced  urlLink Phi Kappa Nasty: The Greatest !  Buy it! Copy it! Don\'t let your grandmother listen to it!  Hear  urlLink Ginger Lynn!  The B-Man yell about skanks! Skills yell about hobos! and much much much more!  If you have never heard the wonder of Phi Kappa Nasty than you have only lived 1 percent of your life! Hear samples now! Buy two minutes from now! '),(3153,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    It\'s the opposite of Violent Stick People ... It\'s Positive Jesus with Depth! '),(3154,1785539,'06,August,2004','2004-08-06 00:00:00','   Ha! At least the man knows he  urlLink blows ass balloons  when it comes to  urlLink Presidenting ! '),(3155,1785539,'06,August,2004','2004-08-06 00:00:00','   How I Know I Have an Addiction to Television by Brent Kinkade  Hey, what\'s that girl doing? Someone died? No, wait, it\'s a rock star. What are they doing now? Is that a high school? I couldn\'t do that at my high school. Who is that now? Are they friends? Hey, a jacket!  \"Brent!\"  \"Uh, yes, what?\"  \"This lady has been asking you the same question over and over again,\" says R. Girl.  \"Huh?\"  Oh, that\'s right I\'m at a grocery store. I\'m turning in a lottery ticket for a two dollar prize. Stupid in-store television display.  \"Oh, sorry,\" I say and focus on the cashier lady.  \"Everytime there\'s a T.V. on ...,\"says R. Girl.  \"Yeah, I know. What was that I was watching anyways?\" I ask.  \" Confessions of a Teenage Drama Queen,\"  R. Girl informs me.  \"Oh, crap, I have a problem ...\"   BUT! I\'m not the only person with a problem! Because .... The other night R. Girl and I stop by to visit her friend the girl-barber because she cuts hair like a downtown stylist at a reasonable rate. There\'s one catch. She has a very stoned and rubberheaded roommate. Here is my conversation with him.  Stoner: Hey, you want to smoke some?  Me: No, thanks. (I was just concerned that being high before a haircut could lead to a mohawk.)  Stoner: (Ver-fucking-batim) Hey, who\'s that guy that played the Riddler on that old Batman show? You know not the guy who was the Riddler, but the actor who played the Riddler.  Me: Frank Gorsham, I think. (Actually, Frank Gorshin, but I was really fucking close, dorks.)  Stoner: No, wait, I mean what\'s the name of that guy on Superman who if you say his name you beat him?  Me: What?  Stoner: You say his name and he ... hehehehe ... goes away.  Me: I am not following you.  Stoner: Wait! That\'s Rumplestiltskin.  Me: Wow, you are really combining a lot of fictional characters with this thought.  So, I get my hair cut and leave. Days later R. Girl tells me that the stoner was upset with her and I because we didn\'t smoke weed with him.  \"What?\"  \"Yeah, he said \'They come into my house and they gotta respect me.\'\"  \"What does that mean? I have to get high to be there?\"  \"I don\'t know. He\'s weird.\"  \"I needed to be high for that fucking Rumplestiltskin conversation; I tell you that!\"  I think I should consult Dear Abbey about this one.   Dear Abbey, When I go over to a stoner\'s house is it rude to pass up a bowl of kind bud? Thanks, Sober in Seattle    Dear Sober, If you want to show respect to your host you gotta toke that shit, dog!   Some people should come equipped with muzzles, dorks. '),(3156,1785539,'06,August,2004','2004-08-06 00:00:00','   Some of you dorks still have no idea how fetching a gentleman I am. Well, can you guess which one of  urlLink these people  is your fearless writer and future Congressman?  Hint: I am a man AND I own a green Nalgene bottle. '),(3157,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    Morning, afternoon or tomorrow to you, dorks. Here is this week\'s batch of V.S.P.s. It\'s been a while since I\'ve written an actual post ... opting for the Christ-teasing and stick-drawing instead. So, let me tell you dorks that I\'m doing good. I had a job interview last week that was promising. I had many daiquiris last night that were stong. Yes daiquiris. My  urlLink kickball team  is strong but we have lost 3 games in a row. I got caught trying to skip out on a $70 chinese food tab after concluding that the chow mein was actually \"noodles and ... is that fucking Vaseline?\". R. Girl and I stood our ground and made waitresses cry. The couple that berates the food service industry together is the couple that eats spit in the food together. More later this week. Enjoy my art! Fuckin\' A, Dorkbros! If you like these, you\'ll  urlLink love these !  Violent Stick People '),(3158,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    You\'ve thought about doing this, admit it.  Except after the flamethrower rampage you go over to Ashton Kutcher\'s house and totally punk him with a Mach 3, some sandpaper and a very hungry python ... oh, oh, and lots of lava!  Violent Stick People '),(3159,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    For R. Girl, who likes monkeys, me and karate video games.  Violent Stick People '),(3160,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    A debate settled!  A spear\'s use maximized!  Violent Stick People '),(3161,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    A different version of what you can find at  urlLink http://members.aol.com/JesusImages/  (Psssst ... the golfer one makes Jesus look like a midget.) '),(3162,1785539,'06,August,2004','2004-08-06 00:00:00','   Spreading my wings! Two  urlLink fan art  comics that I drew are up on  urlLink Dinosaur   urlLink Comics  today!  I might be prouder of myself now than when I won the \"Best at Canoe\" ribbon at summer camp. I can canoe. Canoe?  ___________________________   urlLink   Oh,         urlLink   my         urlLink   good        Christ     urlLink   in         urlLink   Heaven    ... '),(3163,1785539,'06,August,2004','2004-08-06 00:00:00','   Run, fuckers!  It\'s orange time!  The color of non-rhyming, financial terror and half a screwdriver!  Run, fuckers!  Run! Huh? The information that made the terror alert go orange was  urlLink 4 years old ? The plans were  urlLink pre-9/11 , eh?  Oh.  Well, I, um, think I read in a book somewhere about Germany wanting to invade Poland. Invasion! Pre-emptive orange! Run, fuckers!  Run! '),(3164,1785539,'06,August,2004','2004-08-06 00:00:00','   Our President is weird. Our President is a little nutty. Our President isn\'t really all there. Used to hearing weird things about the President. Nothing really shocks me anymore. After that fish comment nothing really makes me ... What? A man obsessed with Jesus and  urlLink potholes ?!?!?  The fun continues!  Watch  urlLink this ! '),(3165,1785539,'06,August,2004','2004-08-06 00:00:00','   From Reuters: Kerry said Americans wanted real leadership and vowed ... to be a president in the mold of Franklin Roosevelt, John Kennedy, Harry Truman and Dwight Eisenhower. \"They stood up and told the truth to the American people and they had a plan and they took this country to the future,\" he said.  ... well, except for Roosevelt. He didn\'t do much \"standing\" and was convinced that if the Nazis knew he  urlLink couldn\'t walk  THEN THE WORLD WOULD FALL! Hooray anyway!   An R. Girl story about booze OR nipples:   Girl at bar Saturday night: Oooo, my breasts are getting full of milk. That means my baby\'s hungry. I can feel my nipples leaking! (Chugs 20 oz. cup of Sex on the Beach).   A Me story about swimming OR nipples:   Girl at pool last night: (Puts toes in water) Ooooo, my nipples, my nipples, my nipples, my nipples, my nipples!   Check out the new  urlLink Violent Stick People t-shirt ! It was inspired by  urlLink this  comic. '),(3166,1785539,'06,August,2004','2004-08-06 00:00:00','   Holy brain just exploded, Batman! Dorks,  urlLink witness beauty beyond all ! '),(3167,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    Today\'s theme? Hamburgers!  Violent Stick People '),(3168,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    America has a hero again! (DAMMIT! I wrote \"hear\" instead of \"here\"!  Rubberhead osmosis, damn it all!)  Violent Stick People '),(3169,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    The prequel of Hamburgarr\'s amazing life!  Violent Stick People '),(3170,1785539,'06,August,2004','2004-08-06 00:00:00','   Billy Pilgrim has turned his back. Rabo Karabekian has closed the barn door. Kilgore Trout has  urlLink canceled his newspaper subscription .  Thanks for the war, urlLink  asshole .  You\'re the right man when the country needs to kill off its heroes. '),(3171,1785539,'06,August,2004','2004-08-06 00:00:00','   These are the Rubberheads in your aw-office.   In your aw-office.   In your aw-aw-0ffice.   These are the Rubberheads in your aw-office   The people who say things    that make you think they\'re ding-a-lings.   The people with heads made of ru-u-u-bber!   (Mere minutes ago by the coffee machine.)  Guy Who Kinda Looks Like Another Guy:  I\'m definitely going.  For that price to hear that many  urlLink motivational speakers .  Likes Her Coffee: I know.  Like  urlLink Goldie Hawn  and that governer from New York ... Giovanni ...  Guy: Giuliani.   They\'re the people that make you think   about going out to drink.   They\'re the people with heads made of ruuuu-u-u-ber! '),(3172,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    My weekend? Oh, dorks, this! I, for being America\'s Boyfriend, received a robotic dinosaur as a gift from R. Tops. And THEN we were endowed with the honor of watching a friend\'s kitten! What happens next? You should know! When R. Girl slips out of the bedroom for a second, it\'s Kitten Robo Dino Wars 2004! '),(3173,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    The battle rages on for hours upon hours! Homes are destroyed! Villages burned! '),(3174,1785539,'06,August,2004','2004-08-06 00:00:00','   urlLink    T. Rex is the winner! He stomps to celebrate his victory over the kitten! '),(3175,3877108,'13,July,2004','2004-07-13 00:00:00','    It\'s so unbearably hot and humid here outside.  I hate Mississippi summers.  Bear in mind that I mentioned our nursing class of 60 meets in an old dormitory building with air conditioning that works sporadically, if at all.  Makes for a miserably long day.    I think I have very poor size perception.  I don\'t know that I view myself as big as I am, or view others as small as they are, or something to that affect.  Did that make any sense?  All these skinny minnies in class are doing nothing but increasing my self-loathing.  Yes, these are my own issues.  Yes, it\'s not their fault.  I\'ll just blame society.  At least I\'m putting these nursing school lessons to good use...  using my defense mechanisms already.  Ha!  There\'s a girl in our class who was nearly Miss Mississippi one year.  Now all she can talk about is her engagement and upcoming marriage.  She\'s way too bubbly for 8am.    Another girl in class is recovering from major abdominal surgery and sits in those blasted chairs all day.  Bless her bones!  Our instructors were fully espousing the \"C=RN\" mantra today.  \"C is average.\"  \"C gets a bad rap.\"  \"You can be average and also be competent.\"  I\'m not sure I buy into this philosophy just yet, but my standards may have to withhold some lowering as we progress into more difficult classes/projects.  Here\'s a neat link we read at article from in class today: http://survivenursingschool-snurse-l.org Hope it\'s still active, check it out.  We didn\'t actually get to bed bathe each other today.  Instead, we viewed a demonstration of one of our instructors bathing another one.  They also changed the sheets while she was in bed (occupied), brushed her teeth, and gave her a back massage.  The recipient is well liked by most of us so far.  She\'s a clever, witty little Jewish lady who uses words like \"kibbitzing.\"  (did I spell that correctly?)  That particular prof\'s lecture today was on stress:  symptoms, illnesses, management, coping techniques, defense mechanisms, anxiety, psychotic tendencies.  When it came to the subject of worry, she included a very short but fascinating blurb on shiva.  If any folks out there care to enlighten me on this subject, I\'d be delighted.    Besides the lovely bed bath/etc demo, we viewed some of those PBS-type instructional videos that must have been made in the dark ages.  Very informative though - more of bed baths, changing sheets (more complicated than it sounds!), oral care (both teeth and dentures), back rubs, hair care, and that sort of thing.  One thing I was confused about and never got an answer to.  How do you do an occupied linen change when a client/patient cannot be rolled on their side?  A classmate broached the question to our prof and she said, \"Just extra help.  It has to be done.\"  Not the most effective of an answer...  This same classmate has already begun questioning authority.  He seems to have a tone of superiority in his voice and constantly wonders of the instructors, \"wouldn\'t it be better this way?\"  These ladies have more nursing experience combined than he is aged in years, but perhaps he hasn\'t heard the often rumored cardinal rule of nursing school is to shut up and take what you\'re given. Grin and bear it, anyone?  In addition, several classmates have mentioned being tired of this hygiene/universal precautions silliness and being ready to move on to loftier topics such as pharmacology (we don\'t have that until fall).  I actually wonder what expectations some of these people had of the field of nursing before they began class with us last week...  I\'m wondering if/when we\'re going to have a section on body mechanics and how we protect ourselves from those injuries nurses seem so prone to.  Maybe when we get around to the mobility/positioning lectures and labs we\'ll cover that subject.  I\'ve got to get a lab coat.  Today we received a little bit of information on our upcoming teaching/learning project.  We have to prepare a presentation for a group of 3-12 year olds on a subject like stranger awareness, oral hygiene, healthy snacks, etc.  So tonight our group of 4 decided we\'d brainstorm on topics and reconvene to make a decision tomorrow.  We\'ll be speaking to those kids at a church daycare, so no taboo subjects allowed.  What would you want to speak to a 3 year old about that might be offensive anyway?  We\'ll also end up presenting our information to our classmates as well.    So far the instructors seem to have assumed we automatically know how to do things like put up bed rails, etc.  I\'m looking forward to getting some hands on experience during skills lab the next few days.  I\'m ashamedly mechanically incompetent.  :)  Of course, nursing school seems focused on very independent learning, as well it probably should be.  We\'re expected to seek out a great deal of information on our own.  I think many of the community/junior colleges are doing their students an incredible disservice by some of the \"hand holding\" that I\'ve seen exhibited in their courses.    Shaving a man, not sure where we\'ll get to try that.  Wonder if my hubby would let me use him for practice?    We signed up and paid our SNA dues today.  Got our equipment orders - pen lights, clipboards, scissors, hemostats, steths, BP cuffs everywhere.  We won\'t know our \"big nurse-little nurse\" assignments until they host a picnic right after fall term begins.  My kids seem to be adapting fairly well to Mom being gone just a little more than usual and having a different pick-up/drop-off person.  My parents spoil them, but not to excess.  I wonder how long a two-year-old can survive a diet of only pancakes?    Tomorrow?  Values clarification and Vital Signs (lecture and lab).  First BIG exam is Friday, fortunate enough to make a 100 on my hygiene quiz today.  Woohoo!  Wish me luck!   '),(3176,3877108,'12,July,2004','2004-07-12 00:00:00','    Door prize to the first person who can identify what song my title quotes from today.  :)  All morning - group theory.  Types of groups, communication in groups, benefits of groups, group stages, group roles, group process.  I\'m already beginning to see what someone meant (on a discussion board I frequent) when they talked about their frustration at the amount of psychology, etc (less \"hard science,\" for lack of a better term) in nursing courses.  I\'m sure that subject matter I find more appealing will come along, we\'re just laying the foundation for now.  We also have a big section on teaching/learning this morning.  In our cohort/small groups, we\'ll be doing a presentation on something such as stranger danger, oral hygiene, handwashing, etc to a group of preschoolers, then a more formal presentation to our class.  Not sure what age group or specific topic we\'ll have yet, but should find out in the next few days.  Our cohort group hasn\'t really clicked yet and we all seem to be pretty quiet/shy/nervous about this whole thing.  The meetings of these small groups are out of classtime, so we have to organize those ourselves.  Again my commute/family obligations present play a part in the group dynamic among a class of 19-20 year-olds.    News broadcast today:  we don\'t do injections until fall, but thank goodness we don\'t get to practice on each other.  Big fun on the board for tomorrow in hygiene lab:  giving each other baths, brushing each others teeth.  Luckily we can just wear shorts and a tank top and we\'ll just be bathing extremities.  We have to bring a basin, soap, washcloths, toothbrush, toothpaste, etc to class tomorrow.  Also the big hygiene quiz.  Hygiene lecture was today, covering baths in all their varieties, perineal care, shaving, oral care (dentures, brushing teeth), eye care (contacts, prosthetics), hot and cold applications, back rubs, TED hose, and I\'m sure a number of other things I\'m forgetting.  We had a special guest speaker today - a massage therapist from a local day spa.  All I can say is, now I really want to go have a massage done.  Sounds like a really interesting job as well as training, something I could see myself doing if this nursing school thing hadn\'t panned out.    Felt even further isolated today.  I always try to reach out to those who are different, been doing that since high school.  I hate to see someone else snubbed or ridiculed, myself included.  Today at lunch was a stinger.  Another quiet, less \"trendy\" girl and I sat together at lunch.  In the cafeteria, a couple of classmates came over and asked the table next to us if they\'d like to come sit with the rest of the nursing students.  Not one word was mentioned to either myself or the girl sitting with me.  I\'m hoping she didn\'t intentionally leave us out, but it\'s obvious that we\'re all in the same class...   Bummer.  On tomorrow\'s docket, in addition to the hygiene lab/quiz and nursing paper due, we\'ll have a stress and professionalism lecture.  At lunch, the Student Nurse Association reps will come to deliver our supplies (steths, cuffs, penlights, scissors, etc) and take our applications.  We\'ll also receive some info about the \"Big Nurse-Little Nurse\" program where the senior nursing students mentor us lowly juniors.  Ought to be interesting to hear more about that...  I guess I\'m going to have to break down and see my plasma or something in order to buy MS Office Suite LOL.  I can\'t seem to get anything done without it - downloading PowerPoints, marking up docs in Office....  just what I need, more expenses in an already strapped budget.  Prayer request>That I\'ll be more disciplined with my own spending/bill-keeping practices and that somehow my money can keep up with my months...   Eek!  :)  I have so much more respect for nurses out there in practice now, after only 4 days of school.  The amount of material to learn is staggering.  Who\'d have known there was 42 steps in making an occupied bed?  Or that giving someone a bath could be so complicated?  Or that inumerable number of types a day you\'d have to wash your hands.  I don\'t see how floor nurses manage to give tihs kind of care and get anything done with the patient loads that I hear rumored...   So, at the end of this blue Monday, as I sit down to cram in my APA before my 8:30 am paper tomorrow, my hats off to all you nurses - whether LPN, ADN/BSN/diploma RN, MSN, or any of the other number of titles/degrees you hold.  You guys definately don\'t get the credit you deserve.  :)   '),(3177,3877108,'11,July,2004','2004-07-11 00:00:00','    Nothing exciting.  Spent the whole time at the in-laws.  Nieces were there, so my boys played with them both indoors and out.  Spent lots of time in the (above ground) pool.  Did manage to get to the university library and couldn\'t locate any of the books I wanted - found a few that might do the trick for my paper.    Finally got home Sunday evening around 5:00.  Stopped by Grandma\'s house and she fed us dinner and watched the boys while I worked on my paper a little while.  Have I mentioned yet how blessed I am to have such a supportive family?  I appreciate them so much.    Been sleeping late all weekend, so Monday morning at 5am is going to come way too early for my tastes...   '),(3178,3877108,'09,July,2004','2004-07-09 00:00:00','    The title says it all.  Last night was rough.  Boys up late, didn\'t want to sleep.  I have GOT to get them on a regular bedtime routine.  Didn\'t get much studying done.    Universal precautions quiz wasn\'t so bad.  I missed a question that I should have known the answer to (about the prodromal stage of infection) - ended up with a 95.  Lots of classmates seemed to feel that they over-studied.  That\'s something I like so far about our grading/school setting.  They post exam grades in our computer system (accessible from the web) on the same day they\'re taken.  Papers, on the other hand, are a different beast...    Our universal precautions lab turned out to be nothing more than a sterile technique demonstration on the part of our instructor, and the viewing of a CDC-produced video.  Actually got to leave a little early for a change, around 3:30!   We have a nursing history paper due no later than 8:30 on Tuesday morning.  If it comes in at even 8:31, our grade will be docked no less than 10%.  My topic?  Margaret Sanger, the birth control movement, her history/education/experiences and her contribution to nursing today.  I don\'t think this paper will be graded too harshly, as we\'ve been advised by the prof that this is to help us introduce some basic elements of APA style.  Unfortunately, none of my earlier classes have used APA and I wasn\'t wise enough to learn it on my own.  Looks like we\'ll get an in-depth education into it these next two years, though.  Picked up a student loan refund check before leaving campus.  Nice to carry around that pseudo-cash for a while, pity it\'s all going to pay daycare.    Today\'s lectures were all about communication, listening, that sort of thing.  From assertive communication to interpreters, dealing with patients in distress/crying to responding to sarcasm.  Interesting stuff, and I could definately use to learn to be more assertive myself.    Goals for the weekend:  Get my paper written, and my furiously scribbled notes transcribed into a more neat typed format.  Thank goodness there\'s no prep (reading) assigned for our Monday group theory lecture.  Now it\'s off to have dinner with my parents, brother/SIL/their kids.  I don\'t turn down free food, especially when it\'s steak!   After that, I think the boys and I will head out to the next town over (25 miles) to spend the night with my in-laws.  My husband should be coming over there from his house sometime tomorrow...    Hope y\'all have a great weekend.   '),(3179,3877108,'08,July,2004','2004-07-08 00:00:00','    I\'ve got to get up away from this computer, study, and lay out our clothes and things for in the morning.  There\'s something to be said for organization.  But first, today\'s journal entry:  Had to stop for a double shot espresso first thing this morning.  I was nearly snoozing the whole drive there.  Going to be a long year.    The espresso ended up giving me a tummyache (not to mention I have a nervous tummy anyway), but I guess it was either be in pain or fall asleep in class.  I\'m sure you know which of those two options seems more rational...  :)  Prayer request>That I don\'t sleep through my alarm (or turn it off in my sleep, which I\'ve been known to do)!     First lecture this morning (4 hours worth) - communication, roles, Maslow, basic human needs, all that good stuff.  Dumb me went and bought a mini-tape recorder last night, but then walked off and left it at home today.  Back to the organization being key, again.  Get organized beforehand, there\'s a useful tip.    Our lecturer around lunch hour must have been competing for the Guiness World Records fastest speaker award.  We were all relieved to see she gave us handouts with a variety of terms on them.  Nevermind that none of them were used in her lecture, and just forget about trying to write down anything off her Powerpoint slides.  I hope she publishes them on the web later because the whole class was frustrated with that one.  I\'ll admit that I don\'t have a lot of sympathy for my classmates so far.  Am I bitter?  Maybe.   Jealous?  Maybe.  I\'m not ruling out any possibilites.  Most of the conversations I heard this morning went as follows, \"did you do the assigned reading?\"  The answers were mostly, \"well my boyfriend called...\" or \"I went to Hunt club and didn\'t get in...\" (local bar), etc etc etc.    Social cliques are already forming.  This is the South after all, and Southern culture still reigns.  Brand names everywhere.  I didn\'t know that big hair was still big.    So far on day two I\'m still keeping to myself.  I guess I\'m a loner.  I really did want to join my classmates when they were all laughing in the lounge during break, but I feel awkward.  I don\'t have much in common with these people yet.  Our conversations end at, \"so where are you from?\" and other such superficial statements.  Very frustrating for me.  I\'m very shy, very self-conscious.    Distance and family obligations are going to prove a barrier for me in that area as well.  While other students can meet at the bar or in the library for a late night study group, I\'m driving home, changing diapers, washing clothes, up with sick kids...  This afternoon we had a huge professionalism lecture.  We also went over things like writing goal/missin statements, facility and school philosophies, objectives, goals, etc.    In addition to that, we were divided up into our Cohort groups.  There are 4 per group, and this is who we\'ll work with for the rest of the semester on group projects.  We also practice skills and do labs with this group, as well as peer check-offs.  I think the only faculty check-offs we\'ll do will be vitals and enema/ostomy.  I\'ve gone through at least 100 sheets of college rule paper already.  My handwriting is getting progressively worse.  Callouses are forming on my index finger.  My advice after 2 days:  buy stock in paper, black waterproof pens (not gel), highlighters....  It storms every afternoon here around 1:30.  Lightning, thunder, the works.  You\'d think we were in the tropics.  Maybe it\'s an omen.  Regardless, a nice rainstorm always makes me want a nap...  A few questions.    What\'s the deal with the two terms patient and client?  Is one preferred as a general rule?  Our instructors seem to use both interchangeably.  Is there a secret to keeping sandwiches in a cooler from getting both mysteriously soggy and stale at the same time?  I had a nice soggy/stale ham and cheese sandwich for lunch, with some baby carrots, water, and a peppermint.  Now if only I could keep my diet so sensible in the evenings...  Anyone know a good program (PC based) for making and/or printing calendars?  I\'ve got way too many things to keep track of, and each of our classes has individual calendars.  I\'d like to streamline them all into one.  First quiz is tomorrow, first paper due Monday.  Need to get studying.  Got to go get Scantrons at the bookstore first thing in the morning before class.  We also have our first lab tomorrow: Universal Precautions.    Good night all, \"see\" you again this weekend.   '),(3180,3877108,'07,July,2004','2004-07-07 00:00:00','    Today in one word:  overwhelming.    I really thought I was ready.  My overconfidence is staggering.  :)  Started out bright and early:  up at 5:00 a.m., didn\'t even dress the boys - just rolled them out of bed in their PJ\'s and dropped them off at my mother\'s (my Dad takes them to daycare on his wya to work).  Thank heavens for my parents, otherwise I don\'t know how I\'d manage.  Hats off to single parents without any kind of support system or network within their personal relationships.  Arrived to school early, around 7:45.  It\'s an hour commute each way, maybe 1 1/2 hours when the traffic\'s bad.  Lots of classmates were already there at that hour, though.  We all staked out good seats (in the front!) pretty early.  It\'s a very awkward shaped classroom we have.  They\'re remodeling a building on campus to serve as a brand new nursing department, but until its completion (next year) the building used in the meantime was once a dorm.  You can imagine, with all the walls knocked out, that the large support beams running through the middle of the now immense square rooms obstruct the view of the instructor and overhead quite a bit.    All morning, we had a sort of orientation.  Covering basic university and departmental policies, sexual harassment, grievance procedure, introduction of the 5 summer instructors, dress codes, etc.  We also had a library and computer resource lab tour.  Our schedule included a History of Nursing lecture, but we never got around to that.  After lunch (which I had in the cafeteria), we spent 4 hours non-stop listening to lecture and writing furiously, taking notes on universal/standard precautions, choosing, donning and removing PPE and infection control (basically a microbiology review).  Class let out at 4:30 and I think we all beat each other down getting out of there.  All of our lectures take place in the same room, so we stay put - only the instructors change.    There are 60 other students in my class.  Of those, maybe 3 including myself are married or have children.  There is one girl who is pregnant, due in September, right around the time we begin clinicals.  I told her she was a braver woman than I for undertaking that!  The only other \"adult\" student besides myself has grandchildren my children\'s age, an older Cajun lady from south Louisiana.  Maybe I should have tried the ASN program instead...  seems to be more non-trad students there?     Prayer request>That I find someone in my class to bond with.  Only one day in and I feel out of place, old, even alienated.  Lots of my classmates seem to know each other from pre-reqs or even high school (most early 20-somethings).  I\'m hoping that I\'ll feel like I fit in more with time, it could just be my stranger-shyness peeking through.    So it was a very exciting day at school, I was thrilled to be there..  on the way home and once picking the kids up, I sort of had a meltdown.  Out of the classroom, my exhaustion and responsibilities are much more visceral.    On a nicer note, when I called my husband on his cell phone to see how work was going, he told me he was at my house.  I got there and he was washing dishes and taking out my trash - a very sweet surprise to get to see him and just kind of let things go, knowing he was in control.    I\'m not sure if I can do this.  The 2-3 hours a day commuting, the 8 hours a day of class, the 2 little kids, the bills, the housework...   Even my grandmother asked, \"Are you sure you haven\'t bitten off more than you can chew?\"  I guess there\'s no turning back now.    That\'s all for now.  You\'ll have to forgive my atrocious grammar, I call myself a stream of consciousness type writer, and I just let the words flow as they come out - not giving much thought to my spelling and punctuation.  I\'ll be glad to accept any constructive criticism, though.   '),(3181,3877108,'05,July,2004','2004-07-05 00:00:00','    Glad to have you here reading my blog.  Hope you enjoy seeing a little bit about what nursing school is like, or just catching up with me if you\'re someone I know or perhaps knew personally.  I\'m going to make an effort to journal every day of my nursing school experience, but we\'ll see how big of an effort that becomes as time progresses.  Feel free to check back regularly for new postings!   A little background information on me, just in case...  27 years old.  Lived in MS all my life.  Small town, but I don\'t think I fit the associated stereotypes.  I have tattoos and live in a trailer though, so is that enough to be considered white trash?   LOL   :)  Shotgun marriage, going on 6 years.  My husband and I don\'t live together, but that\'s by necessity - not choice.  His job took him to a place I didn\'t care to move (the MS Delta - he\'s a service specialist for a commercial pest control company).  Two kids, both boys, ages 6 and 2 1/2.  Little one is still nursing and driving me crazy most of the time - typical toddler and a bully to boot.  Big one will be starting 1st grade in a few weeks.  Quit my full-time job last week.  Been working there as an engineering clerk for the past 6 years and doing school at night/online/part-time for the past 5.    Persistence pays off.  I was fortuante enough to get a nice scholarship package and will be starting the B.S.N. program at a local university this week.    Wish me luck..  and if you have any questions about me/school/my family/etc, just ask...  I\'m pretty much an open book, and I\'m always glad to help.   '),(3182,4121917,'03,August,2004','2004-08-03 00:00:00','   hey every1 just came back from school and how it was fun, thinking about subjects for yr 11/12 im doing whatever i want every1 who objects can go screw em selves yaha dats right any way---------------------------------------baiiiiiiiiii  '),(3183,4121917,'03,August,2004','2004-08-03 00:00:00','   heres a poem my hunny friend wayne sent me-hes a mad sort Tell me as a friend Do you see me in your future? Do I haunt you in dreams that make you glad that you forget them? Or do I visit you every sunday and we go to the park or get something to eat? Am I welcome to come to your aniversary? Will you still think of me the way that you do now? How i wonder Will you even remember me, even just walking down the street? I would like for you to at least think of me every once in a while And if it\'s just a passing thought I hope it\'s one that could bring a smile to your face I want you to think that I have always adored you Cause i always did I want you to think of me and remember how you trusted me with anything that you had to say I would like you to think that it was really important for you to have known me And having me in your life made you feel blessed Always knowing that I cared about you And as far as friends go I really do consider you the best Cause I know that I will think of you in the years, and maybe the next century to come And I will always remember a beautiful creature that came and did loving wonders to my soul Irreplaceable memories will always come up of us and all the little things we do Warming me up with some medicine...I don\'t knowI guess that I just want to say thank you Thanks for being alive Thanks for being my friend   DAMN WAYNE I LOVE U BRO!!!!!!!!!!!!!!!!!!!!!!!!  '),(3184,4121917,'03,August,2004','2004-08-03 00:00:00','   soooo how are all? well im nora but u can call me whatever-seriously, now i do poetry so i guess i let u guys see some u know what i hate just before i write my poetry? FUCKING SOCIETY-SCREW THE BASTARDS WHO WRECK IT what people dont understand is that schools are society and if people in schools change when they get into the big wide world everything has changed...ok im 15 and im aussie well leb/aussie but ok heres my poetry-dont think its sad cuz im leb shit happens to every1, im so fucking sick of people thinking shit doesnt happen to them cuz theyre ethnic-SCREW THE BASTARDS YOU CAN GO TO HELL. hmmm i cant find my good stuff so heres some shit i scribbled today \"situation in control but not in mine strive to be different yearning to rebel against the thing that influences the most society\"  i want to find my voice i want to scream i want a release (at this time when i was writing i got pissed off cuz i couldnt think of anything so then i wrote) THIS IS WORTHLESS SHIT I\'VE LOST MY SKILL  omg omg omg omg i just found a note hidden in the depths of my school bag ok this is a note from maths i carved a G in my knee and my frend saw it and this is a bit of the note me-i dont want to talk about it.period.i have to deal with it friend-not alone you don\'t. i want you to be able to talk about it cause i feel that we can relate of feelings. by the way i know you promised your friend with that pact but you should really tell some one.whats wrong? please dude you know you can trust me i won\'t tell anyone. me-i didnt promise her anything the G stands for guilt, the thing is i can\'t even recall the event without getting distressed. how am i supposed to talk about it? drop it please i dont want to talk about it friend-ok  well now i just read proporly what she wrote so okkkkk  man there is this girl in my group who strives to be different like she trys too hard- she lets people read her poetry and always sucks up to the teachers like what the fuck- she has problems and should be on prozac SERIOUSLY i\'d love to smack her one but i cant cuz she wont hit back  ok all for now bai bai  '),(3185,4121917,'04,August,2004','2004-08-04 00:00:00','   just some poems from the depths of my suicidal pen---- 1st i can feel it coming back again pray for the thunder pray for the rain give me a hand give me a blade help me stop the pain  2nd you have no problems in your life wouldnt you like to know everyone thinks i have a perfect life wouldnt they like to know fuck you and your subject fuck you ill see you in hell  3rd the pen the pipe from my heart the knife the blade of trust the alcohol the drink that loosens my heart my suicidal dream my acheivement  4th my ears the hearing of pain my mouth the keeper of pride my nose the inscence of blood my eyes hide the pain but let the pride go through  5th hall of fame FAITH G DIE HAIL  6th let some one go on with their fantasy at least they will be the happy one  7th i can\'t beleive it\'s happening second time,worst time it\'s an effort to breathe go home have a smoke carve a marking it\'ll be over soon  8th Safe Understandable In need Craved Irreversible Done Explainable  9th suicidal fantasy will only come out at night if you can do it dream my dream is to die leave without a trace i\'ll miss you i love you how come i dont get it in return? i try i try so fucking hard but you turn around and ignore ignore it mum fucking ignore it at least i\'ll die and then finally you will have soomeone to ignore your fucking attention seeking brat LOVE  ok thats my dumass poems anyone who reads it dont stress im not going to kill my self its just a fantasy and fantasies never come true-the reason i have this blog is so i have something to record my feelings and if i dont tell some one about it they can\'t get to it-SCREW YOU BASTARDS  '),(3186,4121917,'05,August,2004','2004-08-05 00:00:00','   I JUST PICKED UP THE MOST HOTTEST FRENCH GUY!!!!!!!!!!! OMG ARGHHHHHHHHHHHHHHHHHHHHHHHHHHHH  '),(3187,4121917,'06,August,2004','2004-08-06 00:00:00','   im sick of attention seekers, if u wna go and publicise that ur gna kill ur self-go and do it-no ones cares and no one is gna cry when u die, there are cases where people are so serious that they exclude themselves from society, thats what ppl should be caring about, not dumass attention seekers who say they r gna kill em selves u kno what they can do it, i promise u all i wont feel guilt or empathy for u when u die  peace, love, empathy                                         nora  p.s does anyone want to kno why people commit suicide? IT IS BECAUSE PEOPLE DO NOT WANT TO LIVE IN A WORLD WHERE INNOCENCE IS SO SHORT  '),(3188,4121917,'06,August,2004','2004-08-06 00:00:00','   u kno what who ever says to ppl they r gonna kill themselves can go and do it UR A FUCKING ATTENTION SEEKER AND DESERVE TO FUCKING DIE-NO ONE WANTS U IN THEIR LIFE SO GO AND KILL YOURSELF AND ILL PUT A DEAD ROSE THAT I PISSED ON AT YOUR FUCKING GRAVE  im so pissed off-thats right im pissed off i cant beleive that someone would do it, if u were serious u wouldnt tell anyone, u wouldnt want anyone involved, ***** UR A FUCKING ATTENTION SEEKER- i couldnt give a flying fuck about u or ur throat   peace, love, empathy                                      Nora  '),(3189,4121917,'06,August,2004','2004-08-06 00:00:00','   dumass  '),(3190,4121917,'12,August,2004','2004-08-12 00:00:00','   fuck i might be expelled-i didnt do anything though, the teacher thought i threatened him now there is a major thing happening---------------------------------------------------------------------------------------SCIENCE TEACHERS ARE DUMBASS  '),(3191,3749228,'30,June,2004','2004-06-30 00:00:00','   I just got home from the first tournament baseball game that Ethan and I had. The game was so close. IT was the top of the last inning we were up to bat and were down 6-5 we hit 1 home stole 1 home and Ethan walked 1 home. So now we are ahead 6-8 it was getting dark so we were afraid that they would call it for darkness. Our pitcher walked one who stole his way to third. We have two more people to get out. The next batter hits a grounder to second, and the guy on third gets home but we get the batter out at first. And finally we struck the last batter out! We had won! So now we continue in the tournament with a game tomorrow. '),(3192,3749228,'29,June,2004','2004-06-29 00:00:00','   After Josh woke up from his dose of LSD we were laughing about all the weired jobs people have at hospitals, like the first lady they talked to asked \"This was an accident right\"? Josh was thinking \"no I go around and break my arm on purpose. and another lady wasant allowed to say he had a broken arm even though it was obvious because she wasnt a doctor so she had to put \"posible break\". Another guy they talked to whose job was to make shure that women that are in labor are realy suposed to be at the hospital. Its just realy weired. '),(3193,3749228,'29,June,2004','2004-06-29 00:00:00','   This morning I woke up and was laying in bed and Josh came in with this big splint on his arm. I was like josh what did you do? He said that he had gone to somebodys house to work on their hay loft, and it fell, and he fell, and he screamed, and his arm snaped. So now Mom, Dad, and stacey are on their way to get his arm operated on. '),(3194,3749228,'26,June,2004','2004-06-26 00:00:00','   Today Ethan and I Had our last double header baseball game. We lost the first one and won the second. Yesterday Trenton and I went to the skatepark. I saw someone bust their board right in half. The Guy had oleyed up to grind and when he landed on the rail cracked right in two. '),(3195,3749228,'25,June,2004','2004-06-25 00:00:00','   this is my new blog '),(3196,3749228,'23,July,2004','2004-07-23 00:00:00','   I haven\'t posted for such a long time because Dad was running a phone line down to staceys room out of the computer jack.&nbsp; we are done with the dry wall, and have started with the mudding.&nbsp; this morning I got up and helped Mom get ready for the garage sale, then after it started I went down stairs and helped dad with the dry wall until lunch time then I ate lunch then I went to the tennis tournament, and won my matches, then I came home and helped Dad with the dry wall some more and now its almost time to eat, and after we eat Mom Nicole and I are going to go to a graduation thing at church. '),(3197,3749228,'15,July,2004','2004-07-15 00:00:00','   today was much like yesterday except instead of work at church I worked with Dad, and insted of a thing at the library I had tennis. so thats about all I have to say. By now '),(3198,3749228,'14,July,2004','2004-07-14 00:00:00','   I woke up, ate, and helped Dad in the basement. Then at lunchtime I ate again. At two I had this thing at the library where they showed you how to make stuff out of duck tape. Most of it I already knew. I brought my duct tape pants along, and the lady teaching thought they were so cool, and she was going to another library to put up a display and she asked if she could put my pants up so I was like sure. They,re at the College park library. Then after the thing mom took me to work day and then I came home and ate and now I\'m here.                                                                                                                                 Question: whose blog would you be more likely to coment on, Mine or Joshe\'s? '),(3199,3749228,'13,July,2004','2004-07-13 00:00:00','    '),(3200,3749228,'13,July,2004','2004-07-13 00:00:00','   Hi this is K_bob what\'s up?  urlLink Stacey\'s  going to show me how to link. It ought to work this time. On Sunday I had Trenton over after church and we skated and watched all the big battles on lord of the rings. On Monday I bought a roll of duct tape and made a pair of pants, and we started on Stacey\'s new room in the basement. If any body wants to buy a pair of pants let me know they\'re $5. Today I am just sitting around doing nothing. It\'s quite fun you ought to try it some time. '),(3201,3749228,'10,July,2004','2004-07-10 00:00:00','   Today Nicole had a swim meet so I got up early to go to that. Then we got home, ate lunch, did chores, and now we are about to eat dinner. Stacey and Josh are at a CYA so after we eat I will probably go take a shower, sit around till I feel tired then I will try to go to sleep, then I will wake up, and it will be morning, and I will be happy. THE END. Oh by the way the link to Joshe\'s blog didn\'t work so don\'t even try it. Good night. '),(3202,3749228,'07,July,2004','2004-07-07 00:00:00','   Sorry it took me so long to post so this will be a long one. Check this out I\'m going to try to make a link  to  urlLink josh\'s  blog! I hope it works. On Friday I had a ball game which we won, then on Saturday I went to the Owens. I was supposed to have another ball game on Saturday but it kinda got rained out. On Sunday I went to church and came home and went to church and after church we took the Magnuson downtown to watch the fire  works. We parked at the top of a parking garage, and Trenton and I had brought our skateboards so we skated around and watched fire works. It was really fun. Then on Monday Ethan and I had A baseball game which we lost. Then after I got home from the game I went around with some of my friends and blew stuff up with bottle rockets, that was a lot of fun! Then on Tuesday I wasn\'t feeling well so I slept most of the day. On wednesday I went to workday since I didn,t have a ball game. So today I am Just sitting around doing nothing. In the baseball tournament we ranked four out of thirty-two.        today\'s poll: Which is better STAR WARS, or Lord of The Rings? '),(3203,3749228,'02,July,2004','2004-07-02 00:00:00','   Last night we won our game against the River Dogs. Now we have semi finals tomorrow. We killed them in our game, the final score was 20 -10. In the last inning there was a pop fly right out over the shortstop, and he threw his glove at the ball. When you do that the runner gets to advance two bases, but if he had hit the ball the runner would have gotten three basses. Everybody except the coach thought it was funny. '),(3204,3130162,'29,July,2004','2004-07-29 00:00:00','   our heads will get fatter as we inflate it with the flatulence that has been so wanting to get out. we, as a collective, whole-heartedly have decided to write about our mission statement. no, it\'s not a goddamn memo; there\'s more purpose than that. so here we are discussing semantics and tautology when we should be getting down and dirty. we are all haunted by ghosts that have come and gone. we say \"keep it tight, keep it focused.\" but sometimes, like a silent-but-deadly fart, something just leaks out that was not experienced by the collective. however, i believe if we bring the topic up in the doghouse and we partake -- then it becomes part of the soup. my real question is this: we occasionally do things that are very climatic. it\'s usually just your run of the mill chit-chat, boozing, and hanging out, so it would seem like there\'s barely anything to really write about on a day-by-day basic that is blog-worthy in the HDM sense (as opposed to the 13 yr crush-prone girl sense). and if we are precluding those thoughts that were not shared by the mafia, then the topics are severely limited. i believe we have transcended the boundaries of activities buddies long ago and we are telepathically conjoined. sugar keeps it sweet. blue gives it color. and i ..... am an obsessive compulsive rambler.   we like to eat well. some of us like hotdogs; others prefer sausages. we\'ve all tasted cream corn soup. we are HDM.   .... in the news, candlelight bakery is dead and its place candlelight dumplings. shit, i\'ve been going there since 6th grade.   '),(3205,3130162,'27,July,2004','2004-07-27 00:00:00','   Keep It Together.  Let\'s remember to do that.  Just keep it together.     Happy Premise Number One: There are no aliens.    Happy Premise Number Two: There is no giant foot trying to squash me.      -----------------------------------    no, hot dog mafia is not a group of fat italian jews on a corner waiting for their pet pigs to finish their business.      make no bones about it, there is depth and substance to hot dog mafia. just how much? i a cave full of Fraggles, Doozers, Gorgs, Trash Heap, and others.      hot dog mafia is not to be defamed or slandered, its name is not to be tainted. there will be no misinterpretation or misrepresentation. there is no reason for any of this. participants (in every aspect..if you read this, you\'re an active participant), will add to, enjoy and bask in the warm glow of hot dog mafia. they will take comfort and be consoled by hot dog mafia. they will help right the ship if it is off course. they will help nurture and grow the living breathing presence that is hot dog mafia. challenge but do not deny. engage and contribute to figuring out what is good or bad, right or wrong, tasteful or distasteful, meaningful or meaningless, relevant or irrelevant, smart and clever or stupid and idiotic, funny or bland, interesting or boring...you get the idea. connection with US.   yes, i am unabashedly sentimental. WE are. deal with it b/c we do like cream corn soup.    -----------------------------------    the essence of hot dog mafia?    the element of hot dog mafia?   the essential hot dog mafia?      it is a feeling - a state of mind - it is a band - it is an entity of publishing, of thought leadership, of revolutionary action/inaction - it is THE idea - it is just people - it is a movement - it is a move - it is the culmination of many perspectives...do you see?   '),(3206,3130162,'26,July,2004','2004-07-26 00:00:00','   At 9:06am, I had cream corn soup for breakfast.&nbsp; Then at 12:02pm, I had cream corn soup for lunch.&nbsp; Then at 6:04pm, I had......&nbsp; just kidding you standards-upholding-bastards.&nbsp; Granted my much public, self-aggrandizing efforts on my eponymous website, I won\'t Harvey-Pekar (nor do I really have his talent)&nbsp;this shit to death.&nbsp; Nevertheless, I am more of a volume shooter (like iverson) and perhaps one or two of my ramblings might be interesting.&nbsp;   So while watching&nbsp;Trainspotting,&nbsp;Gerky and I were discussing that watching movies must definitely be a form of escapism and also that, in our memories, we were led to believe that Trainspotting was allegedly a decent movie.&nbsp; So two things come to mind.&nbsp; One, does a person\'s favorite movies denote what type of person, he/she is?&nbsp; And lastly, if watching movies is&nbsp;escapism and you watch at least two movies (4+ hours) a day, are you a goner?&nbsp; Whatever the case, Netflix is feeding a whole generation of reality runaways.&nbsp; With that said..... Hi my name is Jon and I have 110 movies in my Netflix queue........   While at Goldie\'s, Blue became Lance (aka Live Strong) because it was revealed that he had a triple layer of desensitizing latex fused to his manhood much like wolverine\'s adamantium graft (gives him invulnerability).&nbsp; I was thinking about how sometimes I don\'t perform to the best of my ability because I just stopped caring.&nbsp; I was thinking about pokerface-fucking and typewriter-fucking (if you don\'t know what that means, then too damn bad) and I was ready to add another one to the list to describe my current state of affairs: time-trial-fucking.&nbsp; Yes, much like Lance\'s determination to finish the time trial in the Alps ASAP, I found myself doing the same.&nbsp; It\'s a sad state to be in.&nbsp;&nbsp;The more often&nbsp;you get there, the more your designated hitter just doesn\'t want to get up to bat because if you have to leg out an inside-the-park homer, then that shit just ain\'t worth it.&nbsp; Yes yes, you might score and it might even show up on the ESPN highlights because&nbsp;it\'s a rarity.&nbsp; But you don\'t want to be known as a dinker, you want to knock the fucking cover off the ball, admire it, fist pump, dance around the bases a bit, do a little stutter step before hopping on home plate, and then double-handedly point to the sky and mutter some sweet nothings.&nbsp; Yeah, I was the man.&nbsp; And that\'s the way to go...... of course more often than not I find myself echoing the sentiments of the dude in Office Space, \"Now, if I work my ass off and Initech ships a few extra units, I don\'t see another dime. So where\'s the motivation?\"   .... can I get some WTF-you-talking-abouts?&nbsp; in any case, regarding sugar\'s #2,&nbsp;although I wasn\'t thrilled by the quality and&nbsp;was not truly able to immerse myself in that pretend-they\'re-not-only-interested-because-we\'re-paying-them-money&nbsp;set-up, i am very grateful for the efforts of blue and sugar and also it\'s something we\'ve never really done together.&nbsp; so even if the girls were annoying and aggressive (not to mention not too hot), we shared some laughs, some drinks, and some good times.&nbsp; amen to that.   '),(3207,3130162,'25,July,2004','2004-07-25 00:00:00','   casual chic. this is was the response to my inquiry of a dress code @ Tartare. so i purposefully ignored it. after arriving, i had whispers of regret for not looking pimp, but i decided to focus on the food instead.      definitely got to go back, i\'m happy to report that the meal was more than satisfactory and the mills boys track record of fine dining is still riding high. duck fat french fries? tuna foie gras? ostrich tartarte? GENIUS i say. george morrone - kudos to you for making this happen and for showing the community what good eats are about.    ------------------------------------------------------------------    i won\'t comment extensively on the activities at the gold club...    1) a couple of the strippers really pissed me off. aggresssive bitches who weren\'t sexy or hot. what\'s the point of dropping cash on tits and ass that i would kick out of bed. over all, VERY disappointing seeing the low quality of talent at this once esteemed establishment. i guess the gold club has slipped, so the question is - WHERE then, are the hot strippers?     2) found out that prior to last evening\'s activities, mr.jonmlee had never been the active recipient of a lap dance. wellllll, blue boy and i had to put a stop to that nonsense. so jon got one, i can\'t say that he was terribly thrilled about it, but at least its over with. besides, remember my point in point #1.    ------------------------------------------------------------------    ok, enough about that.  regarding bwliou\'s post.      blue boy with blue balls - truely a sad sad story. i wish him well and hope that he maintains his prior level of discipline. is he master of his domain?    ------------------------------------------------------------------      for the purposes of historical context and posterity - let me iterate that we will be thoughtful and insightful with our ramblings, musings, etc. this will be worthwhile content. this does not exist to show the world that we too can join the masses and become certified assholes by thinking that other people give a shit about details of our insignificant lives.      don\'t get me wrong. YES, we will have many posting that would be categorized as lame. and we will whine and bitch about stupid shit. if you don\'t like it, don\'t read it. we won\'t try too hard to be clever and witty - we\'ll sprinkle just enough throughout to keep you coming back.       what i\'m saying is, the focus of our bullshit won\'t be boring details of what i had for breakfast, how i slept last night, or how it burns when i pee. you\'ll get that in healthy doses, but we\'re smart enough round these parts to mix it up a bit b/c we\'re the best. that\'s right, we\'re the shit. don\'t forget it.       I\'ll leave you with this. when you get up in the morning, tell yourself that you\'re a good person, that you\'re absolutely good enough, that you like yourslef (even if others don\'t), and that today is going to be a great day (b/c you are going to make it a great day).   '),(3208,3130162,'21,July,2004','2004-07-21 00:00:00','   Preface: On some lonely nights, I throw this one stack of sticky notes (that I never use because it\'s too high to write anything on it legibly) on my pillow. It\'s sole purpose for its entire stickied life is to hold one specific message. It reads in big, capital letters: \"DO NOT RUB ONE OUT!\"    Sunday. July 25, 2004. 6:45a.m.  Mornings after a last night like last night shouldn\'t be painful. Sebaceous face caked with oil, I awoke with a piercing pang that I could have either taken as arousing or fucking unpleasant. Something was wrong with my right testicle. As if two small fingers held my ball like a vise, the right nut squeezed in on its own and collapsed. I didn\'t have to look down to know that my raging horniness carried over from last night and that my prick was rigidly beyond its usual morning form.    I\'ve never had blue balls, and again, I didn\'t look down to tell, but I\'m assuming that\'s how it would feel. But momentarily, I didn\'t do what my body gave me a jolting warning sign to do. For the past few weeks, I\'ve kept to a disciplined regimen. I do not ever, no matter what kind of torrid, subconscious sex I had with my coworker, try to fiercely beat it. Not if it\'s whimpering, not if I have a ten-minute window before I can get ready for work, not if it sheds a tear of semen. Never.    Why you ask? Dude, it keeps me up at night. Self-gratification is great. In the past, I\'ve armed my hand close to my boxers when I tuck myself into bed knowing the night will be a long series of kisses with longing eye contact, cameraman point-of-views of my cock ramming some bitch\'s wet slit and arching necks of flowing hair. Dreaming itself is tiresome. Imagine instinctively beating off while in la-la land. It\'s no wonder I\'ve been falling asleep on 280 during the morning commute.    But last night - man last night was fun. Granted, the ladies at Gold Club had more squarish jawlines than Tom, Jon and I combined, it was the tease that got to me. It was the hand holding, the lap sitting, the whispers and soft kisses on the ear. One even told me to yank her hair to demonstrate a typical move of mine and then proceeded to lecture me about the long memory of women who have been verbally abused.... TO BE CONTINUED.     http://www.wetcookie.com/pornnewcomer/090704/index.html   '),(3209,3130162,'21,July,2004','2004-07-21 00:00:00','   He\'s found and he\'s in Japan.&nbsp; It\'s old news now, but I was thinking back when I first started playing chess and Phil was bringing out the whooping stick on me.&nbsp; I was also thinking back when Charles first started playing.&nbsp; So much has happened since then and time sure moves quickly around here.&nbsp; It\'ll be nice to be known as the World Champion of something.   '),(3210,3130162,'21,July,2004','2004-07-21 00:00:00','   People DON\'T BE FOOLED,    Allright, I don\'t know who this guy is, but I certainly didn\'t sign on for this shit. So he invited me as a \"team member,\" so what. If he thinks he can just hog up all the space on this blog, he\'s in for a big ole\' lickin. There\'s too many blogs, not enough world wide web and I\'m taking mine, even if it just means taking up this ENTIRE BLOG FOR MYSELF! We\'re losing internet realty here, so I\'m posing as a bloggin groupie and backpacking on this \"Tommy\" fella. He\'s all about me, me, me. I\'m about me, me, me too.  Sure, we\'re buds - we\'re sharing the same blog aren\'t we? But I see the direction he\'s taking with his half of the blog. Like Tommy\'s \"sugarboy\" blog, each blog, if filtered into a ven diagram, all fit in the middle somewhere to self-promote. A self-promotion of their lives, either through discussion of what happened in a daily e-diary or through a collection of jpegs of moments capturing their daily lives.        What I feel is lacking in many of these blogs is ambition. Though any blog story is great in its own right, it doesn\'t do much. There are no answers given, and as a matter of fact, no questions raised. Enhancement of the mind is the ideal goal of the average blog, but really itâ€™s just become a scrapbook.        I want to change that. Though there is no departure from a chronicling of â€œme,â€ I want to record a new awakening of my mind. A diary of the mind. I somehow regret all that alcohol Iâ€™ve thrown down in the past 27 years of my life. Anyway, I want to condense the senses somewhat. Compress the ears, eyes, mouth and fingers for a second and give live mental static. Squish the daily experiences and the pulp to extract some juice. I hope to give snippets of what my mind is doing from day to day, either through snapshots of what Iâ€™ve read or perhaps something Iâ€™ve dug around to read. What these things do to me and how these experiences educate me, if they do at all, might be solely entertaining, but then again, they might ultimately show that blogs serve no one but oneself. Hopefully they wonâ€™t.     B.W. Liou, out.         P.S. This mission statement will be in constant flux as the focus becomes clearer.   '),(3211,3130162,'21,July,2004','2004-07-21 00:00:00','   ok, so i gotta give a shout out to my boys... \"the mills boys\". we always have a good time. we should, that\'s what being friends is about. we know bout having good times, relaxing\', good food, good drink. heck, we\'ve cooked an entire pig, had a crawfish boil, recreated amazing dishes from the commander\'s palace cookbook, grilled all kinds of meat up the yinyang.     very fun stuff.       we rock. that\'s what i say about my friends and i.'),(3212,3130162,'03,August,2004','2004-08-03 00:00:00','   in an effort to increase our vocabulary, here is something for you to chew on. i stole this one from my free subscription to stuff mag.  Main Entry: priÂ·aÂ·pism Pronunciation: \'prI-&-\"piz-&amp;mFunction: noun: an abnormal, more or less persistent, and often painful erection of the penis; especially : one caused by disease rather than sexual desire  PriÂ·aÂ·pus /prI-\'A-p&amp;s/, Greek mythological character. A god of gardens and fertility, Priapus was the son of Aphrodite, who disowned him because he had a grotesque little body with a huge penis. He was a member of the retinue of the god Dionysus and chased after nymphs.  other interesting finds:  urlLink http://www.air-2.com/  Since the dawn of time, humans have burned plants, inhaling unhealthy compounds from the burning that aren\'t found in the original plant. The Vapirâ„¢ line of vaporizers and vaporization products from Air-2 changes all that, releasing an herb\'s essential material without many of the toxic elements created by smoking. Imagine a smoker who uses a vaporizer with tobacco instead of smoking it. No visible tar, little or no carbon monoxide. Vaporization yields only the pure flavor and active substance of the tobacco without the carcinogenic byproducts.   '),(3213,3130162,'02,August,2004','2004-08-02 00:00:00','   in an age of virtual voyeurism, spyware and google-stalking, the dawn of the internet stalker has come to pass. we welcome it with open arms because we are curious and inquisitive creatures by nature. therefore, i would like to share with you the wealth of information at your disposal. for example, useless and pointless information can be found online including how much money your neighbors contributed ( urlLink http://www.fundrace.org/neighbors.php ) to a political fund, as well as your friends, coworkers and family. if you\'ve graduated beyond the rank and file of your typical web monkey into the upper echelon of a cyber psycho, then you can start looking up the corresponding name of a phone number ( urlLink http://www.anywho.com/rl.html ) or an IP address ( urlLink http://www.whois.sc/169.229.15.35 ). i guess this could work wonders if a certain girl gives you her home number / email and you doubt its veracity. the only safeguard against google-stalking is being altogether common ( urlLink http://technology.nzoom.com/cda/printable/1,1856,175193,00.html ). i guess having a non-unique first name and last name should ensure me of having cyber obscurity. then again, I am not really a person who strikes up too much curiosity.   '),(3214,3130162,'01,August,2004','2004-08-01 00:00:00','   Dear reader,  West Millbrae won\'t be an arbitrary Sunday bbq forever. That\'s been plain to see. After good talks with Chuck and Tom, it\'s clear - as soon as each of us are of purpose, crossing the chasm of laziness and self-pity, each will travel a road inevitably leading us out of this \"city.\" Not with a spit and scournful look over the shoulder, but with a tunnel-vision yet clear aim walking our legs. It\'s like the smooth rally of consecutive grounders between Jon, Matt, Jethro, Chuck, Tom, any of us - it will feel effortless and right.  And I will celebrate when that moment comes, though not without a furrowed brow and touch of sadness. Even now, the few of us who\'ve laid out a plan long ago have moved on. Ryan, Phil, BLeong, soon Jethro. But the great thing about us \"Mills Boys\" is that we won\'t forget. How can we forget? This is, and always will be, a great clan.  One simple measure of how unique we as a group are, have been the various comments I hear from fringe friends, coworkers and acquaintances. I remember recounting a full day of basketball, tri-tip, DVDs and beer to Theresa when we first started dating. The first thing out of her mouth was, \"Wow, you spent the entire day with them?\" Pausing, I reacted with 1. surprise: Days like that are natural, every-weekend occurences for us; AND 2. pride: Yes, what are typical Sundays for us are rare to others and yes, we had ten full hours of absolute fun. I made a sound between a harrumph and a laugh and retorted, \"You\'ve never hung out with your friends all day long?\" She answered with a long \"noooooo.\" Tom has told me his coworkers react similarly in shock.  Thus, we are both the envied and the oddities. We throw footballs in the street for hours, we cook a baby\'s weight in crawfish, we appreciate scotch and bourbon like old men, we puff cigars and tote on psychedelic pipes, we bob our heads and snap our fingers at places like Biscuits &amp; Blues. BUT, we do throw basketballs and tennis balls at each other\'s heads, we rip on each other, we punch our meaty fists at each other\'s clavicles, we give a good slap on the back and then promptly stick our middle fingers up with a loud \"FUCK YOU\" to each other. There is nothing we won\'t do.  Look, I\'m not sure whom I\'m writing this to, and I don\'t pretend to think our group is more exceptional than the next, but I\'m just encapsulating what this group means to me up to this point in my life. And at the age of 27, I\'m hitting \"two roads diverged in a yellow wood\" as are the rest of my brethren. I just hope change will be good to all of us, and whether we traverse the bridge with goals tucked in our arms or find stability not too far away, I hope we never forget all the fun we\'ve had in West Millbrae.  As Tom says,   \"we rock.\"  BLiou.   '),(3215,4023489,'27,July,2004','2004-07-27 00:00:00','   urlLink    Me, my grandma, and mom at lunch today.&nbsp; urlLink   '),(3216,4023489,'26,July,2004','2004-07-26 00:00:00','   my parents just went to sleep, in different beds, in different rooms.  i hope i don\'t turn out like that.  sometimes when i\'m falling asleep i think of people i used to cuddle up with, and it makes me sad.  but it just occurred to me; including childhood and being old and in between, modern women today might just spend more time sleeping alone than not.   '),(3217,4023489,'26,July,2004','2004-07-26 00:00:00','   mollie, moriah and i went to a restaurant titled the \"winward passage\" this evening, which was full of old people, wood paneled walls, and overcooked vegetables. it was exactly what i needed on a dreary, rainy columbus night like this. i love thunderstorms, and never get to see them in socal. why can\'t i see a rainstorm instead of all this drizzle?  spent the rest of the night at janine\'s, drinking red wine, and seeing her new house. now that we\'re 25, everyone (especially my friends in columbus from hight school) are getting to be so damn  domestic.  a house. we grew up together. have been best friends since 4th grade. she owns a house, and i\'m running around all summer sleeping with strippers named juan.   speaking of, in retaliation to last night\'s mean, mean email, said ex-boyfriend said he owes me nothing, he\'s glad we\'re not together, and called me a slut.   first i wasn\'t fucked up enough for him, and now i\'m a slut. i could waste my time typing all of the reasons i dumped him in the first place, but he\'s not worth it. i can\'t believe i\'m wasting this much time on him now.  maybe i need drama to survive?   '),(3218,4023489,'25,July,2004','2004-07-25 00:00:00','   what happened to me?  i used to be so nice. too nice. i stayed with an asshole who wasn\'t good enough for me for too long, and just when i\'m starting to get over him... i did a very un-Christian thing tonight.   i \"accidentally\" stumbled upon his weblog. not on purpose... not really. i was visiting a weblog of his friend and saw him connected, so of course i read it.   the internet is a dangerous thing.   in it he talked all about his new girlfriend,&nbsp;about how \"it\'s always nice\" to go over to her parents\' house for dinner (what, did they start dating the night we broke up?), how they did this, that, blah blah, puke puke puke. i don\'t want him but the idea of him being with someone else makes me want to vomit. am i crazy?   in said weblog, he also referred to his \"psycho ex-girlfriend\" (aka me), which pissed me off but then i thought, he\'s right! i\'m stalking him on the internet!   so i did what all sensible girls would do and in an email response to him from the other night, casually mentioned that i\'d slept with someone else while we were dating, which is true, but i was never going to tell him. i didn\'t want to hurt him. but now i do.   '),(3219,4023489,'24,July,2004','2004-07-24 00:00:00','   just got back from my family reunion. it seems they\'re getting older and shorter every year. it\'s always weird to me to have to introduce myself to relatives. i feel like we should just know who one another is, but we don\'t, so it\'s awkward. my aunt mabel brought a keyboard and was really getting down.  i drove with my mom and two of my cousins, and we talked the whole way about relationships and the like. i don\'t often hear my mom talk about such things, so it was fun but now i\'m all hopped up on coffee with nowhere to go. it\'s another crappy day in columbus, which is why i appreciate the sun in la so much.  still thinking about the ex, with his new girlfriend. i fear he treats her better, although everyone tries to remind me that people don\'t change that fast.  last night i went to a party down in the victorian village and my friend brett showed up DRUNK, and started wrestling me. at first i thought it was funny but then he started getting a little too close; was it just me or was he trying to lick my neck? his hug goodbye was lingered a bit too long, too. he\'s probably lonely, but it was weird. he dated one of my best friends for over a year-i just can\'t see him that way.   '),(3220,4023489,'24,July,2004','2004-07-24 00:00:00','   so i finally get an email from the ex that i dumped a few months ago. before and during our breakup he gave me all of these \"people should stay friends even if it\'s really hard\" speeches and then... nothing from him. and we were friends beforehand, so i\'ve been kind of hurt.   i broke up with him because he wasn\'t good for me-cheap, stingy, closed off, anger issues-but didn\'t expect silence. then, and this is the danger of the internet, i check his friendster page and see that his status is \"in a relationship.\" ahhh. bad memories of a girl from his photo class he was hanging out with before and RIGHT after our breakup came speeding back. again, i know i ended it, but he wasn\'t exactly begging me to stay, either.   so i emailed him, thanking him for not keeping in touch with me and for letting me find out via friendster that he\'s dating someone else. he emailed me back saying he needs his space, and i should respect that. i do.&nbsp;i will. the bastard just should\'ve told me this in the first place...  is she prettier? smarter? funnier? does he love her?&nbsp;fucked up?  he accused me of being too balanced, too \"normal.\" hope he\'s having fun in a dysfunctional relationship.   '),(3221,4023489,'24,July,2004','2004-07-24 00:00:00','   today i decided to start a blog. mostly because i\'m used to typing in my journal, as opposed to picking up a pen and doing it the old-fashioned way, and I\'m at my mom\'s house, sans my jump drive. so here we are.   '),(3222,3784161,'30,June,2004','2004-06-30 00:00:00','    At the moment, I\'m in an odd sort of mood. It\'s like a surreal incompletion. An empty serenity - a nearly eerie contentness.   It didn\'t rain today - it poured. The sky opened up and it seemed as if solid walls of water surounded us. Of course I started the day off right by running and dancing in it with my friends. I love the rain, espescially when it\'s appropriate for playing in. Really heavy rain can lighten everyone\'s mood, if you\'re out in it. Others see it as a limitation - an obstacle that leaves them trapped indoors. But what harm did a little water ever cause anyone? It\'s like a swimming pool in the air!   A freshman thought I was a sophomore. Some of my mother\'s friends thought I was 14. I don\'t wanna look old, I deffinately don\'t want to look like everyone else - I like being me usually, even if \"me\" appears to be 14. But I would like to feel atractive every now and then. I mean I\'m not a beauty queen or love goddess, but I like to think I\'m a nice girl, and that surely someone wants my company. It would be nice to be asked on a date or something - just something to make me feel not completely repulsive.   I mean, I\'m happy- I played in the rain and danced around with people I\'ve known forever and people I\'ve known a few weeks and they love me and I adore them. I just wanna be pampered I guess, once in a while. Ya know, feel like a girl. *sigh*  Anyway so I\'m in a kind of daze, a place of wonderment at my situation of happiness mixed with a false loneliness. I suck at putting this into words. Is it too cliche to say I just wanna be held?   ick. that just sounds so- girly mushy gushy. I disgust me sometimes. :/  My bed is calling however, and I cannot deny it my presence.   Dreams yo. '),(3223,3784161,'28,June,2004','2004-06-28 00:00:00','    Okay so I\'m still not completely done pasting stuff from my old blog, but I\'m close. This is stuff that I didn\'t write, but I think is absolutely beautiful.   ~~~~~~~~~~~~~~~~  Haiku Innuendo  Solace from such small words...   So rare a flower Delicate to touch and taste To own it blasphemy   ~Derek~  ~~~~~~~~~~~~~~~  I\'m \'back\' persay. Return? What the hell would that be?  I say it doesn\'t exist - any real \'return\'. \'You\' cannot return because \'you\' are no longer \'you\' when \'you\' leave a part of \'you\' behind.   ~ again Derek, written upon his return to TN, perhaps some of the most painful words I\'ve ever read ~ ~~~~~~~~~~~~~~~~  I find myself yet again a chrysalis of pain and hate. I don\'t think I\'ll ever escape it. Or want to. It\'s comfortable to me now. So easy to hate, to destroy; eventually we all succumb to it, I have only lept time and begun my ascention early. Or rather, reclimbed my old pedistol, rejoined my old self. Foolish of me. I cannot help but wonder what is in store for this aggression I hold so dear. To what end do I clutch it so dearly? At the same time, I clutch eagerly to the better qualities I possess. I\'m not all pain and hate, as no one can claim, and what many have; I am but human, and therefore have my own demons to subdue.    ~yup still him~   ~~~~~~~~~~~~~~~~~~~  La Belle Dame Sans Merci I saw their starved lips in the gloam With horrid warning gaped wide, And I awoke and found me here On the cold hill\'s side.   And this is why I sojourn here, Alone and palely loitering, Though the sedge is withered from the lake, And no birds sing.    ~an excerpt from a beautiful poem we studied in class~   ~~~~~~~~~~~~~~~~~  \"Nobody, not even the rain, has such small hands\" - EE Cummings.   ~~~~~~~~~~  A Touch. Like Cold Rain. Perhaps we should kill the pupa? Let life suck us dry? Maybe we should cut it down, maybe force our dreams to die. Perhaps we should make it suffer? Torture; make it squeel. Let us then stop the chrysalis! Let\'s finally crack the wheel, break the cycle, and end the game; Destroy the circle, and finally cover these fruitless shames.   ~Derek~  ~~~~~~~~~~~~~~~~~   The essence of your presence presents Me with a sunset, an eye set, something God-sent, To which I give thanks for, And why I speak frank more, And more is what you hand to me In each grasp of palms, Every contact of calm, And you come to me when it need not be, Though I love and need you like a sea inside me, A storm of care that dares to be bare and fair, All that which I find rare is borne from the scent of your hair, So that I forevermore am never sore In the softest place of my body, my heart, Because to me you embody the part That was left missing, even before we met, It was you I was missing, And now you are here and only your voice I hear, Even when you do not speak, you are the most clear, Because your language is love, not at all foreign, It is as universal as purity and sin, Yet you are the counter to all that is wrong in my soul, You are the counter I lean on, my ultimate goal. I love you.   ~Amin Issa~    Written for a girl he\'s apparantly been dreaming of for 3 years. That part is irrelevant. What is relevant is that in my entire day of anger at politicians and teachers and my best friends, in discovering with each new step that something else is wrong with the world, I come home and he says can I show you this poem I wrote? I have much to learn and think about.   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~   Never Let you down  We are all here for a reason on a particular path You don\'t need a curriculum to know that you a part of the math Kats think i\'m delerious but i\'m so damn serious That\'s why I expose my soul to the globe; the world I\'m tryna make it better for these little boys and girls I\'m not just another individual My spirit is a part of this that\'s why it\'s spiritual But i get my hymns from him So it\'s not me it\'s he that\'s lyrical I\'m not a miracle I\'m a heaven sent instrument  My rhythmatic regement navigates melodic notes for your soul and your mental That\'s why i\'m instrumental  Vibrations is what i\'m into Yeah i\'m makin my loot everyday But that ain\'t what gives me the heart of Kunta Kuinte I\'m tryna give us free like ---------------  I can\'t stop that\'s why i\'m hot  Determination dedication motivation I\'m talking to you of my many inspirations What I said I can\'t let you or self down If i\'m on the highest cliff On the highest riff And if you slipped off the side and clinched on to your life And my grip I would neva eva let you down And when these words am found Let it be known that gods penmanship has been signed with a language called love That\'s why my breath is felt by the death And while my words are heard it confined to the ears of the blind I to dream in color and in rhyme So I guess i\'m one of a kind in a full house Cause whenever I open my heart my soul or my mouth a touch of god reigns out  ~Kayne West~ '),(3224,3784161,'28,June,2004','2004-06-28 00:00:00','    Okay so before this blog I had another, which I allowed noone but my very old friend Derek to read, because for some reason which I\'ve really never understood, I\'ve never felt the need to keep anything from him - not that he cares about any of it, maybe that\'s the reason, but I don\'t have to be afraid to be human and be vulnerable with him. Anyway so I wrote and he knew it was there whether he ever read it or not i really don\'t know, but now I wanna progress to my own blog, partially because I forgot how to log in to my old one, but also because while I love him, and miss him, that friendship isn\'t what it used to be, which, is a good thing, for we both grew to a point of independence that I didn\'t know I could achieve. Anyway, so I\'m here now. These are excerpts from my old entries, things I didn\'t wanna forget. (be glad I didn\'t repost the entire journal).    ~~~~~~~~Jan 18~~~~~~~~~~~~  The other day I told Derek I\'d love to lean over a cliff, one of those that hangs over the rocky ocean waters of california. I want to lie down on it and hang my head over the edge and scream at the waters until they listen to me. Until they\'re quiet and I am the presence... with the power of the waves to crush anything against the jagged rocks or pull it away, with the beauty of the sunlight reflected in the waters, to capture any one\'s heart, and the serenity of the waves steady beating against the shore to comfort anyone who was hurting.    I want to know I\'m alive. I want to run screaming naked down the beach, or climb a snow-capped mountain and stand atop it and watch an avalanche somewhere beneath me.  I wanna go dancing, and I wanna flirt with the band until they let me sing lead-- sing the way I sing at home with a good bowl of ice cream and the right music.  I want to tell somebody I love them, in a way they can\'t help but love me back. I want to die in a storm out at sea, on a sail boat.    The other day, Mandy, only half joking, said she wished that she could have an adventure like in her video games, and books, and anime shows. \"Life is so boring\" she said. But it isn\'t. It\'s just us, we\'re too stubborn and sure that it has to be a certain way to be good. It\'s so sad. I hope that no matter what I do, I never forget what it is to feel like this.   ~~~~~~~~~~Jan 19~~~~~~~~~  ...So I tried to unleash the emotion on the piano, and I couldn\'t play anything right. Not even Eyes On Me. I wanted so badly to scream. It\'s always been kind of my dream, along with dying in a storm out at sea, I\'d like to let down my gaurd long enough just once to throw a real fit, to yell and kick and punch and throw things, breakable things. And then have someone there to catch me when I fall back crying. So many times I feel like doing that, but so far I\'ve been the well reserved, fair mannered child.   I made another list of wishes:   How I wish I could play the drums well. Just sit down and turn up the stereo and let it all out, and be happy again. Or run, throw on the sweats and run for miles until I\'m so tired I can\'t remember what was bothering me. Or drive, drive so fast and far that I get completely lost. I don\'t want to know where I am anymore.   ~~~~~~~Jan 21~~~~~~~~~  an excerpt from an entry inspired by an old note I found from my very dear friend John:   My Dearest Little Sister,  Sometimes in life we fail to see the happiness that one brings to our life to the fullest extent. Missing the subtle, selfless acts that some people do for everyone\'s benefit, we foolishly take these people for granted. You are one of those people, Sis. No matter how many times I look into your eyes, I can\'t help but see a beautiful and loving person inside.   With every smile and hug you give, you brighten another person\'s day. But we take for granted the fact that you are a very caring person. We take for granted the fact that without you there to brighten our day we would all be lost to fits of depression. We take your love for granted.  ~John~    ...The goodness of John. The innocent naieve lovable John. How could I have let him go ever? How I love him. He is the person described in his letter, Not me...I\'m so inspired. By John, by Charlotte, by Ben to better apreciate the people around me. I\'m so in awe of Monica to live so openly, so uncautiously. Mandy, the one most taken for granted, most often overlooked. She has a heart that most never see. She\'s more easily shattered than one might think, oh but she\'s so strong, for her mother. She\'d give her life away so that her mother would be happy -- Selfless. I want to become them.  ~~~~~~~Jan 30~~~~~~~ On nights like these he(Sam) makes me thankful that there are people in the world that don\'t put themselves first. I always wanna be one of those people, but I\'m not.   ~~~~~~~~Feb 05~~~~~~~~~~ My life is a song without Melody. I hate songs without melody. They tend to go on and on building tension or excitement, but never provide a release. My life is a song without melody because I can\'t prioritize. Familyschoolfriendsmusicmocktrialsyouthlegboyfriendsliteraturereputationsleepeternalquestionsfuturepastpresentverbalvommit. That\'s the name of my song. I wish I were a cat, with 9 lives. one to screw up with- to learn. a second to be rich with. a third to be dirt poor with. a fourth for law. 5th for psychology. 6th for music. 7th and 8th to love with. The last to teach with.   ~~~~~~~Feb 17~~~~~~~~ Ew. That\'s my reaction to life right now. It\'s a gooey green glob. Not even green. At least that\'s a vibrant color. No, it\'s grey. It\'s a murky grey glob. And someone has poked it with a stick.   ~~~~~~~~Feb 26~~~~~~~ How do you fix things like this? I envy John on some levels. Like I told Ben, I wish I believed enough in prayer to think it would help. Or like Tara in fate and Karma. I wish I could believe that everything I did was irrelevent in the divine scheme of things because it\'s \"meant\" to be a certain way. But I\'m stuck in a world where everything I say or do has consequences. . . And we will get past it but we will not mend . . . I hate that life is real now. It used to be so escapable...  ~~~~~~Mar 13~~~~~~~~~~ I will say this. This mood I am in right now, is only a half truth. It is not completely real. But then again, what is? I have lost respect for someone that I have always admired and I am hurt because of it. Last night, I had a lot to think about, and it consumed me, and reality crushed my naivety. Today I have changed but then when are we not changing? I used to run screaming from the thought, and now I accept it. Change is a part of life. The way in which we apply that change is where its worth is measured. I don\'t know a lot of things. I don\'t know what I want out of life, or an after life, or what of any of it is real. But I do know that if I ever stop learning from others, I have become someone that I do not want to be.   ~~~~~~~~~Mar 20~~~~~~~~~~~~ proof of beauty within... tonight was the first time in a very long time that I\'ve felt even remotely at home in the world. '),(3225,3784161,'28,June,2004','2004-06-28 00:00:00','    Written on Valentine\'s Day - There exists both a literal and symbolic meaning behind my riddle, but I would be more than a little surprised should someone understand my origional meaning. Anyway I thought it sounded pretty so I\'m keeping it from my old blog.    Velvet Red  So coveted. So Admired. Picked from a million only you, alone.  Your thorns gone away --  Only the soft red velvet to carress my lips. Only the sweet enchantment  to draw my nose to thee.  My life flows through thee; you open unto me.  How do I best enjoy thee? Your thorns gone away, I am not afraid.  Vulnerable Delicate Beauty,  Dare I reach? Velvet curls,  sooner worn and tired.  Dare I tear? The velvet spread around, to be loved, once. Never whole again.  Or do I preserve? In your entirety, your beauty and your sanctity. Your life denied, radiance stolen.  Always there are others. Always, I want you. '),(3226,3784161,'28,June,2004','2004-06-28 00:00:00','    So I\'ve screwed up a lot, yet I\'ve survived. I\'ve made new friends and by some miracle held on to the old ones. I\'m stronger. I still feel young and bewildered by my suroundings. Sometimes I feel lost, and always I am without direction. Disapointing, discouraging, dissatisfying - the three d\'s of my past year.   But I survived! I\'m me, more than ever before. For the first time ever I\'m not following people around with my tail between my legs trying to squeeze into conversation - I\'m just naturally welcome. It\'s so nice to not be afraid to be alone.   As with so many things, I wish I could explain it. There are a few things I would have done differently, but I didn\'t know then what I do now, as lame as that sounds.   Anyway I\'m happy. I look forward to tomorrow whatever it may hold. And I think I like my new blog. I wonder if I\'ll let people read this one. '),(3227,3784161,'22,July,2004','2004-07-22 00:00:00','    Last night was so great. I wasn\'t locked out of the great secret circle of people. There was no secret circle it was tons of fun and almost like old times. New kid- Nathan- very awesome guy. cept that he doesn\'t vote, that will have to be remedied, but for real,&nbsp;I think he\'ll fit well with the rest of us.     &nbsp;  I wish here was someone to be all excited for me about today. I can drive! woo hoo.    Dr. Dillaman, if something bad is happening to the animals, we\'ve got to tell the Wizard. That\'s why we have a wizard.  &nbsp; '),(3228,3784161,'19,July,2004','2004-07-19 00:00:00','    In the mood to be beautiful. Want a guy to take me out somewhere nice and treat me special just for one night.  &nbsp; I\'m not like other people. I guess I come off as immature, but my naievity is more than that. I respect myself, and I learn from the mistakes of others, as well as from my own. I\'m content. I just wanna be pretty. '),(3229,3784161,'15,July,2004','2004-07-15 00:00:00','    so today sucked and i have all sorts of ugly words running through my head right now. i want to be strong. i want to be happy. i want to be respectable. i want to be good at something. i wish i could control my tears better. i wish i could keep more inside.  &nbsp; but i\'m so glad for people. for people that hug me and make up stupid songs and dances for me and try and tell other people to go away for me and look at me with eyes that say what words just can\'t. &nbsp; people are so funny, how they can be so comical and hardened against any real thought or emotion, then suddenly become more sensitive than could be thought possible. Like that time I was at Jeffrey\'s and he was playing with his cat. His face got all soft, relaxed. Standing in line at Disney, little kids playing with older brothers, you know those teens don\'t act that way around their own friends. (btw there is nothing more sexy than a guy good with kids).  &nbsp; God I hope I\'m not on a gong. I don\'t wanna just be there, I want to have a purpose. I wish I could go running right now. Oh how good that would feel.  &nbsp; anyway, g\'night - here\'s hoping for a better tomorrow '),(3230,3784161,'14,July,2004','2004-07-14 00:00:00','    \"everyone is good at something. maybe you just haven\'t discovered it yet.\" &nbsp; Thankyou Andy '),(3231,3784161,'14,July,2004','2004-07-14 00:00:00','    Okay I danno how to spell Cochroach, but there was a HUGE one in band today. Jeffrey saw it first and scrunched his face up all funny, and then he showed Jessica and she kinda shrunk away from it with her \"Eew\" and then we all craned our necks to see what Jeffrey described the Mother of all Cochroaches. Dave made us ignore it. A little later it crawled past me. I feel sick thinking about it. I don\'t take well to bugs. Espescially big gross ones. And I mean he was soooooooo big and gross. And I\'ve never really studied a roach before ya know, so now I have this image in my head of this huge ugly thing and this line down the center of his back with little white lines going horizontally across it and lots of legs and it just crawled and crawled. Ew. I feel quezy now. Anyway it went over to Lindsey, and I caught her attention and pointed it out and she made faces, and Rickey saw it and came over and kicked it out from her playing spot and then squashed it. I didn\'t witness the actual squashing, but I saw him kick it and then I heard the stomp of his foot and the poor gross thing is now a pile of mush on the floor. With the stomp came chills. It took me a sec to shake them. It was gross and I don\'t like bugs, but it must\'ve known it was going to die, doncha think? With the kicking and all I mean, and it was so helpless. Just squash.   Dave missed rehearsal yesterday cuz he had to go to a funeral. Today we talked about death. Funerals we\'ve been to- we shared stories about touching dead people in their caskets as curious children. Suicides- jumping from buildings, the emotional scars that must leave on everyone that sees it. Fears- ways we absolutely would hate to die. Even hopes- if we could choose, how would we go.   Well I\'m tired and still feeling cruddy about the roach. I\'m off to bed, another long day awaits me. '),(3232,3784161,'03,July,2004','2004-07-03 00:00:00','    this has got to be like my 3rd or 4th entry today but I\'ll be gone next week so that\'s okay.   Derek lives!!!  nothing makes my day more than a random \"I love you\" in an IM. I bet it\'s been more than a month, maybe 2 since I\'ve heard from him. I had just convinced myself that we\'d finally grown apart and that somehow that was okay because we\'re both okay on our own now, and the important thing is we helped each other get to where we are now. But when his little messaging window popped up I knew I must\'ve been smokin somethin cuz I still love him like mad and no matter how much our lives have changed or how little we see each other, he\'s still Derek, and the magnitude behind that title, as most of my oldest friends know, is immeasurable. And, he said \"I love you\" - 6 years ago he would\'ve never done that. It\'s wild. But it meant that I\'m not clinging hopelessly to a lost friendship. \"Where\'ve you been?\" \"away\" - some things never change. He is the mystery of my life. Well, one of the many, one of the best. Maybe he\'ll come see me for my birthday. I wonder if he has his license yet......  okay so I\'m in love with the world and i get to go to disney world and party while mandy and charlotte and ben serve as my servants, checking my email and calling me to lemme know about the stupid school stuff. I\'m so happy. I want a trampoline. I wanna jump on it and yell and sing and then take a pillow and blanket out to it and sleep under the stars.  Happy 4th of July everyone. '),(3233,3784161,'01,July,2004','2004-07-01 00:00:00','    Incomplete and mediocre. That\'s me right now. Frustrated and very tired. Alone.   Melodramatic.  Okay so my aunt just called to \"see if I was excited about the trip.\" she is anyway. She went on and on about how much fun it will be. Maybe it will. I love her.   when i get back tho, i want a hug that lasts forever, from a boy. or from charlotte. lol. anyway ciao. '),(3234,3784161,'01,July,2004','2004-07-01 00:00:00','    When I was doing research on the stupid situation with the school board, I found this:   I shake with chills and wonder why   it\'s colder than outside.   My textbook pages need some glue;  in fact, it seems I\'m missing two.  The halls are packed, my locker\'s broke,  the lunchroom\'s fogged in hazy smoke.  The doors don\'t close; the window sticks.  My novel\'s dated \'46.   I gasp as my aged desk chair creaks.  It hasn\'t broken in two weeks.  We need relief; it\'s not too late.  Today the lights are working great.  With budget cuts and stacks of bills,  rain pours through cracks; the classroom fills.  Our classes were outside last week.   No need to worry; just a leak.  The ceiling\'s cracked, the walls are chipped.  The chalkboard\'s white and the floors are tipped.  We only need a little aid,   if just to buy a can of Raid.  In smoky, musty, freezing air,  I cough and shake. It isn\'t fair.  Jessica Boelter, age 17  New Ulm High School, New Ulm   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  They are going to cut all the first-year teachers. Why would the governor want to cut our education when he got his education? That is so stupid.  Mercedes Steward, age 9  Marcy Open School, Minneapolis ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  I think that cutting school budgets is a bad idea if we kick teachers out of their jobs. How do you think they would feel? Think before you do something. They should give schools more money because some schools can\'t even afford stuff like computers, laptops, desks, chairs, books and other important stuff. I thought Jesse cared more about education. Think about the students\' future. How can we get jobs in the future without an education?  Yeng Yang, age 11  Hancock Hamline University Magnet School, St. Paul  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  I don\'t really understand the budget cuts, but I may have come up with a solution that I think may work. We could sell all the schools and everything in them and use the money to buy a laptop computer for everyone. The schoolwork could be e-mailed every day to the students, and when the work is done, the day is over. If the work is not done, a private tutor paid by the parents would come to help. After a student learns so much, they can take a test that determines whether he can move up a grade or stay at the one they\'re on and learn more. Each district can keep one or two buildings for meeting once a week. In order to fit all the kids into one building, they would have to break the kids up into groups and have one group meeting a day. I think this could work someday. I know it would work for me. I wouldn\'t have to get out of bed as early.  Kyle Vasen, age 13  Coon Rapids ACE, Coon Rapids ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  At least the kids are trying. What\'s wrong with this country, that kids understand things so muhc better than adults do? '),(3235,3784161,'01,July,2004','2004-07-01 00:00:00','    When was the last time you looked at an infant and thought, \"well his life will be a complete waste and failure\"? When did you last know someone that died with absolutely no regrets?  If these occurances are so rare, then why is it okay to look at a school with 1800+ teenagers and not only accept but anticipate the likely failure of 70% of them?  I thought this was America, the land of the free. I thought that all men were created equal, and that everyone was supposed to be given a shot at life, liberty, and the pursuit of happiness. I guess I missed the memo that teens are exluded from these opportunities, espescially those from poor families. Since when is 15 old enough to label someone hopeless? When did it become okay to stop all efforts towards a decent furture for kids that made a few mistakes, or worse, because their parents did?  I think I know the answer- since certain adults found their way into power and money, they have the responsibility but not the time or concern for these \"hopeless\" teenagers.   I\'m disgusted. My local school board and county commission are too busy feuding with each other to do their jobs. The final vote on budget cuts will come on tuesday, and the vote on the supposedly neccessary tax comes on wednesday. I\'ll be gone for both. I doubt that even with these massive cuts I\'ll be dramatically effected in my last year of school, but others will, and in the years to come it can only get worse. Just the thought, of grown men and women - educated and trusted by the community, being so immature and at such an enormous cost, it\'s mind numbing. '),(3236,3784161,'01,July,2004','2004-07-01 00:00:00','    just wanted to say hello. today was interesting. bryan came. i miss him, adore him.   i was delerious at lunch, andy\'s sister\'s are probably scared of me now- sleep deprivation can do strange things to a person. i\'m just kind of hovering now, until time for bed. but still i\'ve enjoyed the day. tomorrow is a half day- so i don\'t have to do everything tonight.   It\'s not weird being a senior, tho i thought it would be. What\'s weird is looking at Jeffrey and Jessica and AJ and thinking about them being juniors. That makes me feel old. And none of them are taking AS. I\'m dissapointed, but I can\'t really blame them. band and AS and a job plus their very active families would be a lot to swallow. Still, they\'re old. Only they\'re babies. I wonder if they\'d be offended that I thought like that.  I love our teachers this year. Dave. Heather. Lee. I wish we never had to spend time with the rest of the band, or espescially other instructors. I wish it could just be us all day.  I\'m learning to march. The drum doesn\'t even hurt my back anymore. Of course I have like the smallest one, heh. My legs cramp a lot tho - not that that\'s new, it\'s just a pain when marching.   It\'s still raining. That makes me so happy. So relaxed, and still very delierious. Ignore the bad spelling. Ignore this whole entry - it\'s full of senseless rambling and poor grammar too.   my friends laugh at me cuz when they say something bad about the pit and i get defensive, or when i\'m excited over something like my car, my eyes get really big.  yup.  there\'s this freshman named Rickey, he\'s a doll. his brother graduated my first year, and then last year his sister came and we were friends, and now he\'s here, and so excited about everything. it\'s so easy to work with kids like him.   Billy is playing the timpani in the ballad. he\'s a brave kid. I\'ll miss him when I graduate.   I want something, but I danno what. Besides sleep- that\'s obvious. I danno why I\'m this tired, but it\'s kinda fun. Anyway, happy unbirthday. '),(3237,3784161,'17,August,2004','2004-08-17 00:00:00','    I\'m still here, believe it or not. Tara has kidnapped me and taken me to the land of Xanga, which is lame cuz everybody and his mother reads it. How can I write anything worth reading if I know someone will actually read it? Anyway I\'ve escaped long enough to write an update, cuz if I don\'t write now I just might explode later.  I like AJ. Complete honesty in this journal, if I put that in the other one he\'d end up reading it one day or some one else would be like.... ew she\'s a child molester or somethin. I can\'t spell molester. :/ Anyway I can\'t help it - I do like him, I mean, I think he\'s awesome and I love spending time with him. Really I\'d love to spend more time with him no matter what level it\'s on. But I\'m also catching a bit of spring fever here at the end of summer, so it\'s hard to be around him and not overflow with emotion.  It\'s also incredibly annoying for me to be so not in control of the moment.  Two years ago it would not have been a problem. But then, that was when my life was all about falling in love and raising happy little kids and spreading happiness everywhere.  Those things are still important to me, but in a much more realistic way.  After my break up with Coozer at the beginning of sophomore year, I wanted alone time, really for the first time.  But then I got so used to my own independence, that when Sam came along, I couldn\'t committ. There were a lot of issues that factored in to the failure of that relationship, but a big problem was that it was my first serious relationship. When I say serious, I mean that I grew a brain. For the first time I wasn\'t drueling over some picture perfect imaginary future or falling for lines old as time. He wasn\'t asking me to. But since I wasn\'t doing my usual naive cartwheels of joy, I knew something was missing. I couldn\'t adjust despite my own desires to. I wasn\'t happy cuz i was so busy thinking about what was wrong or would go wrong yatta yatta. I refused to let myself trust the relationship or any possible future. So that ended and I spent more time alone, and happy to be so, might I add.  Untill now anyways.  Now i\'m alone and I don\'t want to be. I can be, I\'m not afraid to be. I mean, couple events make me lonely kind of, but having a bf isn\'t gonna make me enjoy the concept of hanging with multiple friends and ignoring all but one.  I don\'t like couple things.  It\'s more the thought of having someone close to me.  All my best friends have their own partners to spend 24/7 with and tell all their secrets to.  I\'d be happy with a best friend, I mean, I have that in Mandy some, but even there boys have taken over a few of my duties. so yeah, that\'s life i guess. Anyway I could deal with it, it\'s not like i just went eanie meanie miny mo, and picked a crush. The aj thing just kinda happened. And it\'s far from head over heals, I don\'t know how to explain it.  It just bugs me cuz I didn\'t decide it was time for it to happen or that he would be a logical person for me to like. I really don\'t know him at all. It\'s lame. Completely. But I\'m around him and I feel like a middle school kid. I want to just tell him, take action of some sort and put myself in controll somehow. I feel like even screwing up and being disapointed would be better than complete helpless ignorance. Espescially watching Becca talk to him. She\'s awesome and really would be way better for him, if they\'re more than friends, which I wouldn\'t know, cuz I danno him! AH. But it\'s too early for me to say anything to him. One day, I\'ll explain all this, some how. If he\'s not taken by then heh. I\'m just frustrated. I hate talking to him and wondering if maybe tara told him, even tho she said charlotte was on crack about the matchmaking thing. Or maybe he can just tell.  I hate talking to him and being giddy instead of focussing on friendship. I feel like I\'m cheating both of us, in the same way that i used to handle my relationships. Bleh. Impatience.  In a kind of related note, I\'ve been full of energy lately. Even something as simple as breathing is awesome, it\'s like i\'m thirsty for air. I wanna be pretty, but more importantly I wanna be dancing, all the time. Music is bursting out of me. That\'s the only way I know how to explain it, I don\'t know if it makes any sense at all to anyone else. Anyway, I don\'t think that relates to aj, at least, aj doesn\'t cause that happiness, but it could be that happiness that causes my feelings for him. Who knows. I hate not knowing.  I do want a slow dance tho, with him. AHHHHHHHHHHHh. Ima have to tell him.  dang this drives me mad. toodles. '),(3238,2880960,'26,February,2004','2004-02-26 00:00:00','   Jewish World Review:   urlLink ZERO-SUM GAME  The British Columbia public school teacher was suspended for a month without pay and received a black spot on his professional record for writing letters critical of the practice of homosexuality to a local newspaper, the Quesnel Cariboo Observer.  '),(3239,2880960,'22,February,2004','2004-02-22 00:00:00','   Sunday Telegraph Leader: Gender games  (Filed: 22/02/2004)   Dora Ratjen, the German finalist in the women\'s high jump at the 1936 Berlin Olympic, was anatomically more equipped for the pole vault. Twenty years later, \"Dora\" revealed that he was actually Herman and had been persuaded by the Nazis to pass himself off as a woman.  In future, no such deception may be required. The International Olympic Committee (IOC) is to consider allowing transsexuals to compete in women\'s events without having undergone gender reassignment surgery. Males would simply need to convince a \"gender recognition panel\" that they were born into the wrong sex.  If one\'s sexual identity is to be reduced to a question of what one feels oneself to be, it is hard to see what there is to stop the principle being extended to other matters. If a Frenchman felt sufficiently British, for example, he could compete for Britain, and an able-bodied athlete who felt himself to be disabled could win a gold medal at the Paralympics.  The new rules could too easily reduce the Olympics to farce, although precedent suggests that, when men take on the image of women, they begin to perform like women. The IOC would almost certainly have rescinded Ratjen\'s medal had there been any medal to rescind. In fact, he came fourth.  '),(3240,2880960,'17,February,2004','2004-02-17 00:00:00','   urlLink   Sunday Telegraph Letters: On Birth Row   Cherie Blair should look a bit closer to home before criticising the morality of America\'s use of the death penalty from a Catholic perspective (News, February 8). Has she forgotten that her own husband has voted in the Commons for abortion on demand, abortion up to birth for disabled babies, and the use of the human embryo as a guinea pig in scientific experiments?  Her horror at the possibility of executing the innocent sits uneasily with her husband\'s enthusiasm for killing the innocent. Has any abortionist ever granted a stay of execution?  I think the scriptural warning to avoid hypocrisy is apt - one should not point out the speck in another\'s eye while ignoring the plank in one\'s own eye.  From: Stewart Finn, London SW17  '),(3241,2880960,'17,February,2004','2004-02-17 00:00:00','   WE DO MORE TO PROTECT WILD FLOWERS THAN CHILDREN    A week or two ago on a really bitterly cold day when I was in the town centre, I saw a man wearing an overcoat carrying a very young baby in a sling strapped to the front of him. What concerned me was that the baby had absolutely nothing on its legs, feet or head and appeared to be inadequately dressed for such cold weather.  Personally I thought this was negligent. I am not a mother but I do know that babies lose their heat very quickly and since they cannot walk about and generate their own warmth, I would be inclined to wrap a child up well and put a blanket over him/her when sitting in a pushchair.  Because of the nature of my employment, my employers expect me to take child welfare issues seriously and, although I would have thought something like this should be common sense, I thought I would take this opportunity to express my concern more publicly and to appeal to whomsoever it may apply, to make sure you wrap your children up well when you take them out in cold weather, as, sadly, I have noticed on more than one occasion children who seem to be inadequately dressed for the weather.  However, this does not totally surprise me because, even more sadly, we seem to be a generation that tolerates the legalisation of child murder.  I find it really strange that, although there is public outrage whenever innocent young children are murdered or children killed and wounded through war, when it comes to the horrors of abortion and its consequences, there is a deafening silence. I was not born under a cabbage patch myself and, as far as I am aware, nobody else was either. I was loaned two videos: one was about the development of the unborn child in the womb, and the other one was about what happens during an abortion; the first was delightful, the second horrific. How can anyone who has any knowledge about the development of an unborn child not have a conscience about this?  I really cannot see how we can ever hope to prosper as a nation when we are shedding all this innocent blood; I am told about six million abortions have been performed in Britain alone since the 1967 Abortion Act was passed. I have heard it referred to as \"Britain\'s weapon of mass destruction\". Are we really any better than Hitler? It seems to me that by ignoring the sanctity of human life from conception we have opened the floodgates to forces of violence and murder in this nation and the current social chaos and disorder as a result. Sadly, there seems to be more protection by law for wild flowers, bats and whales and concern over cattle and sheep than for future generations. I am told that Beethoven with his family history would have been a candidate for abortion. What talent and gifting and future leadership are we killing off? Personally, for conscience sake, as well as for the sake of those who cannot speak up to defend themselves, who have no choice, I find it impossible to remain silent about this issue.  Bath Chronicle News (Letters to the Editor)  '),(3242,2880960,'16,February,2004','2004-02-16 00:00:00','   urlLink Sunday Mail: I FELT I HAD BEEN GIVEN A SECOND CHANCE WHEN PAMELA SURVIVED ABORTION  '),(3243,2880960,'16,February,2004','2004-02-16 00:00:00','   urlLink Melanie Phillips\'s Articles :  \"Terminating the Parental Role, Daily Mail, 11 February 2004  The poignant, handwritten diary entry in its childish, rounded letters says it all. \'I had my termanation (sic). Then in brackets underneath: (killed my baby).\' Through these few words, one can only guess at the turmoil in the mind of this 15 year-old girl as she struggled to come to terms with what she had done. After noting the event in the impersonal jargon used by the professionals who handled her case, she revealed in childishly simple terms how it actually felt to her. Not a \'termination\', but killing her baby. Tragic as this was, the really appalling thing was her isolation from her parents. In need of adult guidance and support more than at any time in her life, she had been effectively abandoned to make this decision and cope with its aftermath on her own. For the doctors who dealt with her had not told her parents on the grounds of \'patient confidentiality\'. The girl, who was frightened of what they might say, had said she didn\'t want to tell her mother. So her GP, who took a mere 15 minutes to see her, simply referred her on to the hospital which carried out the abortion. The first her mother knew of what had happened was when she read the entry in her daughter\'s diary.  What have we come to when a girl barely out of childhood can have an abortion, a procedure with huge physical, emotional and moral ramifications, without the advice and care of her parents because they have been deliberately kept in ignorance of what is happening? Something has surely gone badly wrong not merely with medical ethics but with our society\'s whole attitude to sexuality, childhood and parenthood.  On every count, this incident illustrates a widespread breakdown of responsibility, care and common-sense. The first thing it shows is how far\"  '),(3244,2880960,'16,February,2004','2004-02-16 00:00:00','   urlLink Telegraph | Opinion | Christianity is outclassed  '),(3245,2880960,'16,February,2004','2004-02-16 00:00:00','   urlLink Office of the Deputy Prime Minister - Press Notice - GOVERNMENT ANNOUNCES Â£1.8bn SUPPORTING PEOPLE ALLOCATION  '),(3246,2880960,'31,March,2004','2004-03-31 00:00:00','   urlLink Telegraph | News | In Britain, one in five pregnancies ends in abortion : \"More than one in five pregnancies in Britain ends in abortion while the number of childless women over 40 \'increases substantially\', according to new figures.\"  '),(3247,2880960,'30,March,2004','2004-03-30 00:00:00','   urlLink Telegraph | News | Tory MPs give pro-gay summit a miss  The Tories\' first pro-gay summit ran into trouble yesterday when only five MPs turned up and Ann Widdecombe attacked it as \"misguided\".  '),(3248,2880960,'29,March,2004','2004-03-29 00:00:00','   Sir - Ann Furedi (News, Mar 26) has accused the pro-life lobby of being cruel to women by scaring them into continuing unwanted pregnancies by exaggerating the risks of abortion. Can the chief executive of the British Pregnancy Advisory Service really be so ignorant of the law?  Abortion on the grounds of unwanted pregnancy is not allowable under the 1967 Abortion Act. Ms Furedi must also be aware that there is no right to choose, but that the signed consent of two doctors is required to confirm serious grounds under the terms of the Act.  Is it not rather the case that the pro-life argument that the Act would lead to abortion on demand has been shown to be true? At the time of the Act, the pro-life lobby was vilified for suggesting that doctors might be unscrupulous in determining \"serious handicap\"; now this, too, has been proved to be the case.  One by one, the arguments in favour of the Act are being shown to be flawed. Ms Furedi should admit that we have social abortion in this country, contrary to the law.  From: Christopher Leather, Hull, E. Yorks.  '),(3249,2880960,'27,March,2004','2004-03-27 00:00:00','   Sir - Non-custodial alternatives to prison fail to prevent crime or address the widespread fear of crime. Your leading article (Mar 26) makes the point well. One of the main reasons for this problem, which is worsening monthly, is the fact that those in charge of the legal system do not understand how criminals really think.  Having spent nearly 40 years listening to them in cells and interview rooms, I know that the only punishment they fear is prison. It is precisely that fear that lies behind two of the most common pretexts used by criminals to avoid custody: \"You learn how to commit crime in prison\", and \"I did it to feed my drugs habit\".  Many criminals privately admit that they have no such habit; but it is a good way of getting a probation officer to recommend a non-custodial sentence. All too often, the strategy works. Too many courts fail to realise how smart some criminals are.  However, it is not only ignorance of the criminal mind that results in the failure to protect innocent citizens from crime. It is the lack of courage by the agencies responsible: the Government, the police, the probation service and the judiciary.  The Government refuses to fund a worthwhile prison-building project. The police do not have sufficient numbers to patrol our streets effectively - but two officers can be spared on a quiet road to photograph cars. Too many chief constables prefer to use their resources to trap speeding motorists rather than to combat real crime. Meanwhile, the probation service still views its main task as protecting criminals rather than victims.  The judiciary cannot bring itself to accept increasing public concern and criticism of weak sentencing. The Court of Appeal, in particular, includes many judges who know little about criminals or victims - and this is being noticed by the press and the general public.  Your editorial was quite right: nothing cuts crime more than imprisonment.   From: Keith Matthewman QC, Circuit Judge, 1983-2001, Nottingham  '),(3250,2880960,'27,March,2004','2004-03-27 00:00:00','   urlLink The Herald  The Catholic Church in Scotland has urged local authorities to abandon plans for \"gay marriage\" ceremonies.  '),(3251,2880960,'27,March,2004','2004-03-27 00:00:00','   urlLink Headline news from Sky News - Witness the event  The US Senate has easily passed legislation to make it a federal crime to harm or kill an \"unborn child\".  '),(3252,2880960,'27,March,2004','2004-03-27 00:00:00','   urlLink Population Myth  WASHINGTON , March 23 --- \"It\'s time for the population control movement to call off the dogs,\" said Steven Mosher, president of Population Research Institute (PRI), a human rights group dedicated to dispelling the myth that the world is overpopulated.  '),(3253,2880960,'27,March,2004','2004-03-27 00:00:00','   urlLink Stoddart attacks `too easy abortion\'  SWINDON peer Lord Stoddart has launched a blistering attack on abortion law, saying it is now far too easy for women to have the procedure.  '),(3254,2880960,'27,March,2004','2004-03-27 00:00:00','   urlLink Telegraph | News | Senate rules that foetuses have legal status  '),(3255,2880960,'26,March,2004','2004-03-26 00:00:00','   urlLink Telegraph | Opinion | More prisons mean fewer criminals at large  The appalling catalogue of failures that allowed a serial offender who should have been in jail to walk the streets, eventually killing a policeman who tried to arrest him, makes a mockery of the Government\'s efforts to expand the system of community penalties.  '),(3256,2880960,'25,March,2004','2004-03-25 00:00:00','   urlLink Telegraph | Opinion | Could cinema be the saviour of Christianity?  Is Mel Gibson\'s The Passion of the Christ anti-Semitic, as London\'s Curzon cinema group claimed yesterday? Too violent? Too literal in its adaptation of the biblical text? Did Pope John Paul II really say, \"It is as it was\", or was that just a piece of studio spin?  '),(3257,2880960,'14,March,2004','2004-03-14 00:00:00','   urlLink Telegraph | News | Teen pregnancies increase after sex education classes  '),(3258,2880960,'09,March,2004','2004-03-09 00:00:00','   urlLink BBC NEWS | England | London | My London: \'Cannabis change angers me\'  '),(3259,2880960,'01,March,2004','2004-03-01 00:00:00','   urlLink lifeissues.net | Aborted Fetal Cell Line Vaccines And The Catholic Family  An article which speaks about the use of aborted human tissue to create vaccines. Strange that I cannot find any evangelical articles on the same topic!  '),(3260,2880960,'01,March,2004','2004-03-01 00:00:00','   urlLink Telegraph | Opinion | Women may have it better, but not everyone wants to be one  Most people in the world know if they are male or female. A junior filing clerk called Vicky was my first encounter with a different response. Back then, in the early 1960s, James Morris, the travel writer, had not yet become Jan Morris, author of Conundrum, the compelling story of Morris\'s sex change. For Vicky, we were still in the Dark Ages.  '),(3261,2880960,'28,April,2004','2004-04-28 00:00:00','   urlLink ic Wales - Tory quits after gay storm  TORY AM David Davies is quitting the National Assembly\'s equality of opportunities committee, accusing other members of \"insufferable political correctness\".  '),(3262,2880960,'27,April,2004','2004-04-27 00:00:00','   urlLink The Observer | UK News | Nurses set to perform abortions  Nurses should be allowed to carry out abortions to speed up the process and defuse growing controversy over terminations late in pregnancy, according to family planning chiefs.  '),(3263,2880960,'27,April,2004','2004-04-27 00:00:00','   urlLink Telegraph | Opinion | Why the Palestinians are in such a state : \"There was an hilarious piece in the Washington Post on Sunday, under the plaintive headline, \'Why Did Bush Take My Job?\'\"  '),(3264,2880960,'22,April,2004','2004-04-22 00:00:00','   urlLink Telegraph | News | Humanism becomes one of the new gods of RE  Humanism, attitudes to sexuality and the ethics of wealth are to be taught alongside Christianity and other faiths in the first national curriculum for religious education.  '),(3265,2880960,'22,April,2004','2004-04-22 00:00:00','   urlLink Telegraph | News | Call to scrap daily worship in schools  The law requiring a daily act of collective worship in schools should be repealed, David Bell, the chief inspector of schools, said yesterday.  '),(3266,2880960,'21,April,2004','2004-04-21 00:00:00','   urlLink Telegraph | News | Bloody reality of abortion brought to our living rooms : \"Britain\'s first televised abortion lasted slightly less than three minutes. As part of Julia Black\'s confrontational documentary, My Foetus, Channel 4 last night showed a doctor\'s gloved hand manipulating a suction pump and forceps between the stiffly parted legs of an anonymous woman.\"  '),(3267,2880960,'18,April,2004','2004-04-18 00:00:00','   urlLink Telegraph | News | Should Channel 4 air My Foetus?  An abortion will be shown on British television for the first time tonight when Channel 4 screens its 30-minute documentary My Foetus.  '),(3268,2880960,'18,April,2004','2004-04-18 00:00:00','   urlLink The Herald  A PILOT health project which advocates free condoms and morning-after pill provision for school pupils has been criticised after the number of teenage pregnancies in the area where it was introduced soared above the national average.  '),(3269,2880960,'18,April,2004','2004-04-18 00:00:00','   urlLink UK Study Reveals Reproductive Health Clinics for Teens Increases STD and Out-of-Wedlock Births  LONDON, April 5, 2004 (LifeSiteNews.com) - A UK study has revealed that improved access to reproductive and sexual health clinics for teens actually increases sexually transmitted disease and out-of-wedlock birth rates.  '),(3270,2880960,'18,April,2004','2004-04-18 00:00:00','   urlLink Scotsman.com News - Opinion - Columnists - Pro-abortion lobby has got it wrong - guilt can be good for you  ON 20 APRIL, Channel 4 will show an abortion in a documentary called My Foetus, made by Julia Black who had a termination at 21 and who now has an eight-month-old daughter.  '),(3271,2880960,'18,April,2004','2004-04-18 00:00:00','   urlLink Telegraph | Opinion | The real victim  '),(3272,2880960,'15,April,2004','2004-04-15 00:00:00','   urlLink Telegraph | Opinion | So what\'s wrong with clubbing seals?  I donâ€™t know who handles the PR for these Canadian seal-clubbers, but it must be a hell of a job. Can there be any group, on the entire planet, that so excites the hatred of the British public?  '),(3273,2880960,'13,April,2004','2004-04-13 00:00:00','   urlLink Telegraph | Opinion | The Lords is the more democratic house  Why does Labour want to reduce the delaying powers of the Lords? Surely not even Tony Blair can believe that there are currently too many checks on the Governmentâ€™s power.  '),(3274,2880960,'06,April,2004','2004-04-06 00:00:00','   urlLink Untitled Document : \"Washington, DC-On April 27, thousands of young people across the country will participate in the second annual National Pro-life T-shirt Day by wearing T-shirts with pro-life messages to their high schools, colleges and everywhere they will travel that day. National Pro-life T-shirt Day gives America\'s pro-life youth the chance to publicly express unwavering support for the sanctity of all human life from fertilization until natural death.  \'This simple act of solidarity and courage on the part of America\'s pro-life youth will save more babies\' lives on this day than on any other day in the year,\' said Jason Scott Jones, director of American Life League\'s Rock for Life. \'This day is a day for healing; healing for a generation that has lost one-third of its members to the genocide of abortion. This event will save the lives of preborn children, and spare their mothers the terrible assault on body, mind and soul that every abortion inevitably brings.  The past decade has seen numerous situations in which pro-life youth have been challenged for wearing pro-life T-shirts. While pro-abortion forces use their free-speech rights to spread propaganda and lies unfettered, pro-lifers\' free speech rights are often suppressed. Examples of the bigotry students have faced include:  * Teachers, principals and school administrators have equated pro-life T-shirts to the Nazi swastika or profanity, and have attempted to force students to publicly remove shirts.  * School officials have verbally attacked students and have attempted to suspend students for simply wearing pro-life shirts.  * One student was physically attacked by a convenience store clerk for simply wearing a pro-life shirt while waiting in line.  * Golden Gate bridge police in San Francisco threatened to arrest pro-life students wearing their shirts while crossing the Gol\"  '),(3275,2880960,'05,April,2004','2004-04-05 00:00:00','   urlLink Telegraph | News | Teenage sexually transmitted infections up  Free contraceptives aimed at tackling teenage pregnancy have prompted a rise in cases of sexually transmitted infections (STI) among young people.  '),(3276,2880960,'04,April,2004','2004-04-04 00:00:00','   urlLink Telegraph | Opinion | Too many abortions  Television is fast running out of \"last taboos\", but later this month Channel 4 will break one with a vengeance: a film entitled My Foetus will show a woman undergoing a \"vacuum pump\" abortion, and will also broadcast images of dead foetuses.  '),(3277,2880960,'04,April,2004','2004-04-04 00:00:00','   urlLink Telegraph | News | Secular schools are switching to Anglicanism to fulfil parental demand for \'Christian values\'  Record numbers of secondary schools are \"converting\" and becoming Church of England schools to fulfil parental demand for Christian values and better discipline. Since 2002, eight previously secular schools have become Church schools and three more will convert this year.  '),(3278,2880960,'04,April,2004','2004-04-04 00:00:00','   urlLink Telegraph | News | Channel 4 defends decision to show film of foetus being aborted  Graphic footage of a live foetus being aborted is to be screened by Channel 4 in a documentary to be broadcast this month  '),(3279,2880960,'27,May,2004','2004-05-27 00:00:00','   urlLink Telegraph | Opinion | Face it: it\'s all your own fat fault : \"Just as I was cramming my mouth with another obesity-enhancing cheese \'n\' mesquita-flavour kettle chip, preparatory to washing it down with a draught of life-giving milk chocolate, the phone rang. \"  '),(3280,2880960,'25,May,2004','2004-05-25 00:00:00','   urlLink New Bill Informs Women About Pain Babies Feel During Abortions : \"New Bill Informs Women About Pain Babies Feel During Abortions Washington, DC (LifeNews.com) -- New legislation introduced in Congress will focus on an aspect of abortion that normally receives little attention -- the pain that unborn children feel during an abortion. The bill would make certain that those performing abortions inform women of the pain unborn children experience during an abortion, especially later in pregnancy.\"  '),(3281,2880960,'22,May,2004','2004-05-22 00:00:00','   urlLink New Bill Informs Women About Pain Babies Feel During Abortions : \"New Bill Informs Women About Pain Babies Feel During Abortions Washington, DC (LifeNews.com) -- New legislation introduced in Congress will focus on an aspect of abortion that normally receives little attention -- the pain that unborn children feel during an abortion. The bill would make certain that those performing abortions inform women of the pain unborn children experience during an abortion, especially later in pregnancy.\"  '),(3282,2880960,'22,May,2004','2004-05-22 00:00:00','   urlLink Telegraph | Opinion | We, the law-abiding majority, know the law doesn\'t rule : \"The train was late, but not late enough to shift us passengers from apathy to discontent. I sat on the platform reading. Suddenly there was a commotion and some shouting. A young woman with dyed blonde hair, carrying some sort of case, was trying to cross the railway line.  '),(3283,2880960,'21,May,2004','2004-05-21 00:00:00','   urlLink Telegraph | News | Police \'ignored sex abuse fearing racist accusation\' :  \"Complaints by a detective that she was sexually assaulted by a colleague were not properly dealt with because their police force feared being accused of racism, an employment tribunal was told yesterday.\"  '),(3284,2880960,'20,May,2004','2004-05-20 00:00:00','   urlLink Telegraph | News | Jail young thugs urge residents as \'softly, softly\' approach fails : \"Residents of a housing estate who set out to tackle youth crime by working with offenders have given up their \'softly softly\' approach and called for the troublemakers to be jailed.\"  '),(3285,2880960,'20,May,2004','2004-05-20 00:00:00','   urlLink Telegraph | News | Father convicted for shaking bully  A father who confronted a boy responsible for repeatedly bullying his 12-year-old son has been convicted of common assault.  '),(3286,2880960,'16,May,2004','2004-05-16 00:00:00','   urlLink Telegraph | News | Rebel parish to fund \'cure\' for homosexuals  A parish disgruntled by the appointment of Canon Jeffrey John as the Dean of St Alban\'s will divert thousands of pounds it normally sends to central Church funds to a Christian charity that seeks to \"cure\" homosexuals.  '),(3287,2880960,'14,May,2004','2004-05-14 00:00:00','   urlLink Telegraph | Opinion | Mother may not know best, but she should always be told  The case of Michelle Smith, the 14-year-old whose school \"outreach worker\" arranged an abortion for her without informing her parents, is one of those moral nightmares in which it is horribly easy to see everybody\'s point of view.  '),(3288,2880960,'13,May,2004','2004-05-13 00:00:00','   urlLink Telegraph | News | Abortion girl: who will take the blame?  No one was willing to take responsibility yesterday over the 14-year-old girl sent for an abortion without the knowledge of her mother.  '),(3289,2880960,'13,May,2004','2004-05-13 00:00:00','   urlLink Telegraph | News | School sent girl home to have secret abortion : \"A girl aged 14 was sent to hospital for an abortion without the knowledge of her mother, on the advice of a health visitor at her school.\"  '),(3290,2880960,'11,May,2004','2004-05-11 00:00:00','   urlLink \'Change law on death of unborn\' : \"A CAMPAIGN group today called for a change in the law after a drink-driver was jailed for just two months after a road crash in which a mum lost her unborn baby. \"  '),(3291,2880960,'30,June,2004','2004-06-30 00:00:00','   urlLink Telegraph | Opinion | Worrying look in the womb : \"New ultrasound pictures of a foetus show it toddling at 12 weeks, yawning at 15 and smiling at 18. What is the public reaction? Are we awestruck at this manifestation of the quickening within the womb that every mother feels?\"  '),(3292,2880960,'29,June,2004','2004-06-29 00:00:00','   urlLink Telegraph | News | Nearly half of pregnant teens opt for abortion : \"Nearly half of all teenage pregnancies end in abortion, according to new research published today.\"  '),(3293,2880960,'29,June,2004','2004-06-29 00:00:00','   urlLink the Mail online  Eyewitness to an Abortion  '),(3294,2880960,'27,June,2004','2004-06-27 00:00:00','   urlLink the Mail online  Shock Rise in Abortion Figures  '),(3295,2880960,'27,June,2004','2004-06-27 00:00:00','   urlLink the Mail online  Revolutionary photos in the womb  '),(3296,2880960,'27,June,2004','2004-06-27 00:00:00','   urlLink Campaigners Take To The Streets (from Epping Forest Guardian)  OVER 100 pro-life campaigners staged a prayer vigil outside a Buckhurst Hill abortion clinic after marching from Woodford Green. Organised by anti-abortion group the Helpers of God\'s Precious Infants, the police-escorted protest on Saturday was led by Bishop Thomas McMahon, the Roman Catholic Bishop of Brentwood.\"  '),(3297,2880960,'27,June,2004','2004-06-27 00:00:00','   urlLink IC Wales - Police probe children\'s \'street sex lesson\'  PARENTS are in outrage after a group of children were given a sex education lesson in the street.\"  '),(3298,2880960,'27,June,2004','2004-06-27 00:00:00','   urlLink Telegraph | News | \'Gay marriage\' Bill left in disarray  By George Jones, Political Editor (Filed: 25/06/2004)   Government plans to allow homosexual and lesbian couples to form legally binding civil partnerships were in turmoil last night.\"  '),(3299,2880960,'27,June,2004','2004-06-27 00:00:00','   urlLink Leeds Today : \"ABORTION ROW MUM WINS RIGHT TO PAYOUT By Louise Male   A woman who said she would rather have had an abortion than give birth to her severely disabled son is in line for substantial damages.\"  '),(3300,2880960,'27,June,2004','2004-06-27 00:00:00','   urlLink ThisisLondon  Leading doctors today called for a major overhaul to avoid babies being born alive after abortions.   Pregnancy expert Professor Stuart Campbell has demanded rules should be tightened after it was revealed that at least nine babies are known to have survived terminations in recent years.  '),(3301,2880960,'27,June,2004','2004-06-27 00:00:00','   urlLink Telegraph | News | \'Life without sex is hard but you can do it. I\'m living proof,\' says John, aged 14 : \"In a quiet corner of Surrey, 300 teenage souls are about to be saved. \'This,\' says a booming American voice, \'is a sexual exposure chart.\' \"  '),(3302,2880960,'25,June,2004','2004-06-25 00:00:00','   urlLink Telegraph | News | \'Gay marriage\' Bill left in disarray : \"Government plans to allow homosexual and lesbian couples to form legally binding civil partnerships were in turmoil last night.\"  '),(3303,2880960,'09,June,2004','2004-06-09 00:00:00','   urlLink    Who says the English are boring?...&nbsp; urlLink  '),(3304,2880960,'23,July,2004','2004-07-23 00:00:00','   urlLink Telegraph | Opinion | Designer babies are not wicked: abortions are, though : \"I swear I saw an elephant galumphing around the BBC studio last month, when the final item came up on the early evening news bulletin. But the newsreader appeared not to notice it.\"  '),(3305,2880960,'20,July,2004','2004-07-20 00:00:00','   urlLink the westgaz - news, entertainment, jobs, homes and cars : \"POLICE ARE TOO PC CLAIMS DAD   The father of an aspiring police recruit has condemned a constabulary for \'discriminating against white, able-bodied heterosexuals\' with its mentoring scheme for minority groups.   When Andrew Orchard, aged 19, of Rectory Lane, Norton-sub-Hamdon, phoned the Avon and Somerset Constabulary to inquire about applying for a job he was told that the force was not taking applications until January 2005. \"  '),(3306,2880960,'18,July,2004','2004-07-18 00:00:00','   urlLink   Re: Inconsistent views   There appears to be an inconsistency in Dr Rowan Williams\'s view of salvation. The Archbishop of Canterbury is reported as saying that salvation depends on \"a relationship with God\", but also that he \"did not believe any individual faced the prospect of eternal damnation\" (News, July 11).  If by \"prospect\" is meant absolute certainty, that is correct, since there is the possibility of salvation for anyone right to the end of life (as there was for the penitent thief).  If, however, Dr Williams holds the heresy of universalism (the belief that all without exception will be saved), he cannot reconcile this with the belief that redemption depends on a relationship with God, since many people have no such relationship.  The Archbishop also raises the suggestion of post-mortal penalties for sin that fall short of actual damnation. This concept is not taught by any mainstream denomination other than the Roman Catholic Church, and seems to derive from an inability to accept the full implications of the Atonement, as shown in the words of Christ to the penitent thief: \"Today shalt thou be with me in paradise\".  From: C M Rogers, Ipswich, Suffolk  '),(3307,2880960,'14,July,2004','2004-07-14 00:00:00','   urlLink Telegraph | Opinion | The Tories must confront Islam instead of kowtowing to it : \"Marcel Proust wrote that \'other people, as we get to know them, are like strips of metal dipped in acid: they gradually lose all their qualities - and their defects too, at times.\' Politicians, of course, only lose their qualities.\"  '),(3308,2880960,'14,July,2004','2004-07-14 00:00:00','   urlLink this is north scotland - news, entertainment, jobs, homes and cars : \"B AND B OWNER WHO BANNED GAY GUESTS GETS PRO-LIFE GROUP\'S BACKING  Pro-life activists have echoed the British National Party\'s backing for the Highland bed and breakfast owner who says business has boomed since he openly banned gay guests.  An organisation calling itself UK Lifeleague has branded homosexuals \'abominations\'.\"  '),(3309,2880960,'09,July,2004','2004-07-09 00:00:00','   urlLink Guardian Unlimited | The Guardian | CPS drops case against anti-abortion protester  The Crown Prosecution Service has abandoned its case against an anti-abortion protester whom it charged under anti-pornography laws because he demonstrated with a picture of an aborted foetus.  '),(3310,2880960,'09,July,2004','2004-07-09 00:00:00','   urlLink Telegraph | Opinion | Sheep may safely graze : \"An epidemic of finger-wagging is spreading through British public life. Rarely a day goes by without the BBC announcing, with lip-smacking relish: \'Campaigners are calling for a total ban on... \"  '),(3311,2880960,'07,July,2004','2004-07-07 00:00:00','   urlLink Telegraph | Opinion | Only Cromwell\'s Commonwealth was preachier than New Labour : \"People do not generally feel the need to confess their sins these days; but if they did, what would they say in the confession box? Something like this, perhaps. \'Bless me, Father, for I have sinned. It is four weeks since my last confession. Since then, I have smacked my four-year-old son, causing reddening of the skin. I have eaten Kentucky Fried Chicken ï¿½ not just one portion, but (God forgive me) a mega-bucket. Driving into town, I have strayed into a bus lane. And I do not recycle.\'\"  '),(3312,2880960,'06,July,2004','2004-07-06 00:00:00','   urlLink Telegraph | Opinion | Smacking Bill punishes the wrong people : \"On the face of it, yesterday\'s House of Lords vote on the Children Bill was a welcome defeat for anti-smacking zealots. Their Lordships rejected a total ban on smacking children and opted instead for a government-backed amendment that Dr John Reid, the Health Secretary, described as a \'middle road\'.\"  '),(3313,2880960,'04,July,2004','2004-07-04 00:00:00','   urlLink Telegraph | Opinion | Dr Williams, beware of false prophets  The Archbishop of Canterbury has again been pronouncing on British foreign policy. Last week, it emerged that Dr Rowan Williams had, with Dr David Hope, the Archbishop of York, written to the Prime Minister criticising \"recent developments in Iraq and the Middle East\".  '),(3314,2880960,'02,July,2004','2004-07-02 00:00:00','   urlLink Telegraph | News | Care call for babies who survive abortion : \"Babies born alive after very late abortions should be given the same treatment and care as babies born prematurely, doctors said yesterday.\"  '),(3315,3759734,'30,June,2004','2004-06-30 00:00:00','    What is the limit that a man will take to get into a woman cookie?  Today i have seen the lowest when a girl tells a guy that she only get down with bi-sexual or gay men, so he goes to tell her that he is a gay man just so that he can get that stuff between her legs.  Now i dont know no man that is would stoop that low just to get a piece of pus that is some ol nonsense. Now that leaves me to think he really have some issues going on within himself. '),(3316,3759734,'28,June,2004','2004-06-28 00:00:00','    Today was a pretty good day until I had to watch someone else\'s kids. I love kids to death but bad ass kids I can do without. Yea I know I should have and could have said no but I didn\'t. I dislike when I have to help someone especially when it is family and they have a problem and u try to bring them into your world and they find a way to fuk up things, that is y I like to be bymyself then I wouldn\'t have to worry about that nonsense at all, but when it comes to a point when it the person is starting to deserves u it is time for them to go give them all 5 toes in the butt :) '),(3317,3759734,'26,June,2004','2004-06-26 00:00:00','    CAN YOU FEEL ME? I AM RIGHT HERE WITH A WARM HEART. CAN YOU HOLD ME? I AM RIGHT HERE WITH A LONELY HEART. CAN YOU CHERISH ME? I AM RIGHT HERE WITH A COLORFUL HEART. CAN YOU RESPECT ME? I AM RIGHT HERE WITH A BROKEN HEART. CAN YOU HONOR ME? I AM RIGHT HERE WITH A LOYAL HEART. CAN YOU LOVE ME? I AM RIGHT HERE WITH AN OPEN HEART. CAN YOU SEE ME? I AM JUST A HEART '),(3318,3759734,'31,July,2004','2004-07-31 00:00:00','    GOOD BLACK MEN  \"Good Black Men\" you know and a few women that need to read it...   Good Black men are indeed all around. We pass them on the streets, in the malls and the halls at work. Most we can`t see because we don`t know what a good man really looks like.  He usually isn`t flashy enough or rich enough to turn our heads. He might not wear a suit or push a Ex. He might not have a \"body like Arnold with a Denzel face.\" But, as you mature, you realize it`s better to find someone who`s got your back than someone who turns your head.   A good man doesn`t agree wholeheartedly with everything you say. He doesn`t just tell you what you want to hear and do the opposite. He doesn`t declare how sensitive, sweet, caring, sincere, yada yada he is (he wont have to because it shows). He has his own opinions and you may clash, but he doesn`t have to degrade you to prove he`s right. He even admits at times to being wrong, especially if you are willing to do the same.  A good man is not going to meet every item on your checklist. He is human with frailties and faults mixed in with all of his wonderful, strong attributes. He needs your love and respect. He needs to feel that you don`t live to \"catch\" him doing something wrong so you can declare,  \"Aha! I knew you were a dog!!\"  A good man doesn`t necessarily give you a huge birthday or Valentine`s gift. He shows his love in the ways that are comfortable to him. Don`t judge him by TV standards. No one is living that fairy tale for real. You`ll miss out on your own fairy tale by buying into the myth that our men are no good. It`s just not true.   Black Men, we salute you, and thank you for who you are and all you`ve done. '),(3319,3759734,'30,July,2004','2004-07-30 00:00:00','    &nbsp; Let\'s see where do i want to start today.....................Why do people like to do and say things to bring other down, do they do that just to bring them up and make there day feel better.&nbsp; Also what give a person the right and the freaking nerve to try and correct and tell u what u should and need to do when they are the biggest fucking sinner. '),(3320,3759734,'28,July,2004','2004-07-28 00:00:00','    Loving you    As gentle as a springtime breeze   Sensual kisses to tease and please   As beautiful as a morning sunrise   Kissing and caressing your caramel thighs   A summers day sticky and hot   Longing to revisit your sensual spot   Sitting at the beach as the waves hit the shore   Crescendos of passion has us begging for more   As I walk barefoot in the morning dew   I can\'t keep my mind off   LOVING YOU '),(3321,3759734,'27,July,2004','2004-07-27 00:00:00','    Check this bullskit out while i was sleeping good doing the evening storm my cellphone ringing &nbsp;i wasn\'t going to answer it but they called me about 6 times back to back, so when i answer it.&nbsp; Once i heard this spongebob ass guy on the other end talking some crazy stuff, remind u i have heard or seen him in 2 weeks but he had the nerve to call me and say that someone he know saw me riding around with a guy (WTF) all my friends are males and he know it, but neither the less i didn\'t take the time or energy to to explain myself because for 1 i was sleep and 2 what nerve did he have to call me and that (BS) and i haven\'t talked to him in 2 week give me a freaking break so just to end the Fuk up conversation i hung up the phone. '),(3322,3924062,'15,July,2004','2004-07-15 00:00:00','   I\'ve just been thinking of revising the lyrics to my debut single (perhaps I\'ll post a draft - if someone reading this wants me to, leave a comment), when I sought inspiration in another song.  Now, as soon as I thought of a pop song to take ideas from, I thought only one thing.  Busted.  Inserting their debut album (their finest, and I can justify it being in my record collection as it was bought for me - it would surely be rude not to listen to it?) and skipping to \'Year 3000\', I sat back and prepared for aural pleasure (innuendo most probably intended...).   I won\'t lie to you, listening to that song has probably given me a clue as to how Busted are so successful.  Think about it.  They\'ve been to the Year 3000.  Do they not know the way the charts are going to pan out for the next 996 years?  Does this not grant them the surefire way to construct a song that will cater for the changing needs of the increasingly picky pop audience?  Or am I simply reading too far into this?  Only the future of Busted will tell...  '),(3323,3924062,'14,July,2004','2004-07-14 00:00:00','   I\'ve just realised the overwhelming burden of pressure, both from the expectations of myself and others, to succeed. I would say I can\'t sleep because of it, but its only 11.44am, so what the hell.  '),(3324,3924062,'14,July,2004','2004-07-14 00:00:00','   What is in a name?  They give us an identity, something people can call us, to distinguish us from the rest of the mob - unless of course you\'re in a mob of people with the same name as you, but thats statistically unlikely, isn\'t it? Of course,with a debut single in the pipeline, my thoughts must turn to my name.  Is Robert Pook (my full name, to those who may read this blog and know me only as Rob) a jazzy name, full of the sparkling pizazz associated with showbiz, parties and the luxurious lifestyle of someone famous.  In short, it isn\'t.  Now, I\'m not a group, so that rules out the pluralular (probably made that up...) style of name such as Westlife, 5ive, even the Spice Girls.  No, I need something short, punchy, something that hundreds of thousands of childrens can utter across their school playgrounds as soon as the new charts are announced, or following my appearance in Smash Hits magazine.  Darius.  Look at that, a successful popstar known only by his first name.  But Robert just doesn\'t have that ring, nor does Rob, Bob, Bert, just about anything that can be made using my humble little name.  I would go for Pook, it has been a nickname of mine for many years.  I say a nickname...its just my surname, so I guess it doesn\'t really work as a nickname of such.  But surely Pook corners me in as a performer aimed at the preschool category.  I want to have some credibility as a popstar, not be thought of as some bloke in a brightly coloured rotund costume who dances around singing old nursery rhymes but with the new politically correct lyrics.  All together now...\"Humpty Dumpty sat on the wall, Humpty Dumpty had a great fall.  All the Legal Aid Companies and Lawyers to boot, ensured he got compensated with plenty of loot!\". Perhaps I\'ll just change my name entirely?  Ramon, Diego, Carlos...something continental.  Ramon Valera...that could corner a market if I work on my dark, latin-based smouldering good looks?  You see, this is the trouble of the market.  I don\'t just need to be a good musician, in fact, I don\'t need to be a musician at all - which is rather good as far as I\'m concerned.  No, all I need to be is a package.  An image, a look, a name.  ...this should be fun.  '),(3325,3924062,'14,July,2004','2004-07-14 00:00:00','   Now, we come to the effective \'construction\' of my debut single, my first stab at the heart of pop stardom.  Continuing this analagy further, if I succeed, I will find my hands immersed in the warm scarlet blood of fame.  If I fail, I\'ll just hit the pancreas, joining many other failed acts such as One True Voice (where did it all go wrong?), Fast Food Rockers (I think we can see where that one went wrong) and Daphne and Celeste (see previous brackets).  Whilst this may not knock the Pop Industry dead, it will simply cause a minor injury, slowing it down, flagging it.  Like a wounded wildebeest, falling back in the moving herd - prime targets for the lions, the other genres such as Rock and Hip Hop.  Now, time for some cold hard facts.  My debut single will be self-financed (hence my profile listing my occupation as \'non-profit\' - I expect none whatsoever).  It will be a preliminary batch of 50 copies.  Hopefully, these will sell, and possibly fetch me up to Â£20 profit for charity.  With any luck, I will join the vast swathing ranks of the Charity Single Superstars.  Think Geldof, Bono, Band Aid, S Club 7, Gareth Gates.  My name will join them in lights.  In theory.  And there\'ll be a video.  I already have a slight demo for the single.  Its entitled \'The One Who Proposed\'.  As you may imagine, its about being the one who proposed, but being rejected.  The lyrics are, if I say so myself, jaunty, a little streetwise, a la Busted.  The demo was written, performed and recorded in one live take (after many hours practising alone in a darkened room) using only a basic laptop and a twelve year old keyboard.  Now, we come to the crucial stages.  Over the next few days, I need to jazz it up, give it that extra something.  We need a brass section, an unnecessarily long percussion intro, a synthesized middle-8 in which I can arse around in during the promotional video, we need a dance routine...  ...more will follow.  '),(3326,3924062,'14,July,2004','2004-07-14 00:00:00','   The blog has begun...  '),(3327,2311511,'02,July,2004','2004-07-02 00:00:00','   Hello everybody right now I\'m directing,producing and co writing a Kung Fu movie right now we are in the writing and planing fase but it\'s going to be called \"The Five Emeralds Of The Evil Cat\"  it sounds stupid but it will probably much stupider than the title.Don\'t worry I will definately alert you when my movie comes out and then you loyal readers can buy it on DVD from me over somehow on the net like ebay or e-mail or possibly by my website if I have mine up and running by then .Just to tell you I first got the insperation for my movie because of the T.V. show Kid Notorius and the fact that my best friend has an evil geinius cat who tried to kill me by biteing my right temple.Goodbye for now \"yall\"  '),(3328,2311511,'02,July,2004','2004-07-02 00:00:00','   I just read that horrible story about that poor man who\'s blog was discovered by his mother . I don\'t know if I could function if my ma saw my blog even with just 4 posts I find this story devistating there shoud be a ban of mothers reading there kid\'s blogs I could call it the anti-snooping mother act of 2003 . Oh yea I would like to tell you about a comeidian that I found out about and he is great his name is Bill Hicks if you like political humor but not like Ronald Regan political humor like anti-establishmetal comedy . Also here are some great web comics sinfest.com , joyoftech , mega tokyo and little gamers just to warn you these comics should only be used in an even of a crash over water .  '),(3329,2311511,'02,July,2004','2004-07-02 00:00:00','   I am back once again I just want to say that I hate the parent council at my school they are god damn retards that couldn\'t pull there head out of there asses for a single second to see what the students want and the reason I say this is because they canceled our class trip to Montreal the greatest city in the world It\'s Las Vagas , paris and the worlds largest smoking lounge all rolled up into one and those ass holes  took away all the funding for the trip but technicaly they didn\'t cancel it but if one person can\'t pay $600 per person we can\'t go and I\'ll tell you there are 34 kids in my class and it is bound that one of them will not be able to pay!!!!! Now for an update on my film we have started casting parts and we have found a few that will work for really cheap.Right now though I am looking for some good mac movie editing software because I fear I movie will not be powerful enough.If you want to give me some good sugestions just e-mail me at eprime666@hotmail.com and you might want to checkout my friend\'s blog at www.18thcenturypimp.blogspot.com he\'s the co director and co writer of the film and just to update you I have changed from director to co director and I\'m still producer and trying to live the Robert Evans lifestyle , well atleast as far as my mom will let me . Thank you and good bye  '),(3330,2311511,'02,July,2004','2004-07-02 00:00:00','   Hello readers you may think from my first post this blog would just be about my movie but it\'s about my opinions I just want some publicity for my movie. So this post will be about my opinions and in future if anything big happends in the production I will update you but until then it will be about my opinions and stuff like that. Now I would like to say something about the drug rate in the states you guys (I say you guys because I\'m Canadian) must have the highest heroin , crack and cocain rate in the world I mean look at the facts you elected G.W. Bush  , Richard Nixon , Bill Clinton and Jimmy Carter I mean the man said he was attacked by a KILLER RABBIT does this man just think the world is one big Montey Python skit or something. Those were actualy the stuff you guys did that wasn\'t that stupid now if  you want stupid two words Arnold Shwartenegger. I better stop goofing on your politicians now because the  FBI has probably already send a dark van out for me, good bye.  '),(3331,2311511,'02,July,2004','2004-07-02 00:00:00','   Hey sorry I wanted to post on Friday but my net connection was down for a bit but anyway blogman is here once again dum da dum . I\'m not sick anymore if I didn\'t say so before so ya I\'m back and ya . I just went out for a burger but it wasn\'t that great and I\'m kinda depressed because I\'m not old enough to see kill bill but I think my dad might be cool with me seeing it so I\'ll ask him when he gets back from his business trip and we can go see it when were Christmas shopping for my mom . Right now I\'m trying to see if natural soda is as good normal soda so I can think that I\'m doing something good for my body as my teeth are being ripped out by my dentist Dr.cloth or better know in some circuits as Mr. hied . I want to tell you that the cola is pretty good for natural but it tastes like someone put an unlit cigarette in it and took it out really fast but it still is pretty good I\'ll notify you when I try the lemon lime I\'ll be trying that in about a half an hour or a hour . Oh I have a message to bubble gum pop \"artists\" that are trying to be samantha Fox here is what you do make a girl on girl sex tape that \" accidentally\" is released to the public then go to India and find the joys of hashish and come back and collaborate with jimmy Page on a real album . Good bye  '),(3332,2311511,'02,July,2004','2004-07-02 00:00:00','   Hey  it\'s me it\'s official I\'m sick and I got to stay home today I have a fever of 38 degrees Celsius which is actually not that high but it\'s still a fever and according to a rule at my school a child cannot be sent to school and remain there if he or  she has a fever this is the only good rule at my school . I thought I wouldn\'t go to school yesterday but I pushed my self and I got rewarded , I didn\'t have to go to swimming at school ! Oh I\'m sorry I have yet to mention how much I hat swimming well I like swimming but I hate swimming at school I don\'t know why but I think it has something to do with the fact that one swimming class is  like 18 weeks of intense boot camp training for the marines not that it\'s even a class because they never teach us anything new they just work us and work us until the pool is red with blood like some scene from \"Jaws\" . Those swim teachers torture us so much I\'m afraid that the next time I go to pool I\'ll end up having electrodes stuck to my testicles and my eyes pried open to be brain washed by North Korean propaganda. Now I would like to say something about a new fad I found out about in blogs playing dead , what you do is you post a picture of your self on your blog pretending to be dead but the real trick is to find a creative and original death . I usual don\'t follow fads but I think this is really cool so around Christmas when I\'ll get my new digital camera I\'m going to make one I haven\'t thought of the exact death but I\'m thinking some where along the lines of a vomit induced death like choking on my own vomit or vomiting up my organs or actually if I wanted to make it simple I could just do a simple drug overdose . Oh well I\'ll figure it out good bye \"folks\"  '),(3333,2311511,'02,July,2004','2004-07-02 00:00:00','   hello I\'m back once again to tell you about my life . I think I\'m getting a cold I started to reallize this at about 10:30 last night when I was watching Unscrewed with Martin Sarget . I\'m not sure if it was a cold or just the side effects of the three cokes I had about a half hour before though I did feel a little screwed up in the morning more than usual anyway. Well goodbye I better lie down now so I can convince my mom that I can\'t go to school tomorrow  '),(3334,2311511,'02,July,2004','2004-07-02 00:00:00','   hey this isn\'t a really post I am preforming a test this is only a test had this been a real post I would have had something interesting to say  '),(3335,2311511,'02,July,2004','2004-07-02 00:00:00','   Hello everybody and good day to you loyal readers and first timers I am your leader and I would like to apologize for my weekend post it sucked ass but for those who\'s lives suck worse than  mine here\'s my review of natural lemon lime soda well I tastes like watered down mountain dew and ass though I have never tasted ass it was what I imagined ass would taste like so don\'t buy it . Oh yes I forgot because of my friends stupidness I must correct him he runs a blog and he said that family guy will be returning to the air and in one of the last lines he writes that it\'s the first series to be renewed but the original Star Trek was prematurely canceled and then NBC was flooded with letters and phone calls urging them to put it back on the air and they did so if I am not mistaken but because I haven\'t had a chance to talk to him about it so I don\'t know if he meant in FOX history or what ever but I would like to correct him . So he can thank me for taking his ass slack much like some kind of ass bra but I am much more than an ass bra , I do breasts to but seriously I will hold your breasts anyway I am also an invaluable fountain of tech news well here it is  Mozilla is launching a e-mail client a web browser merged together with a web browser , well don\'t worry it won\'t be like Netscape\'s communicator well I can really only guess that but you know it\'s Mozilla so it\'s probably good . Okay now let\'s play some hard ball and I\'m ah hell I got nothing we might as well just play T-ball . Now I\'m starting to Christmasfy my computer  you know screen saver wallpaper stuff like that . Well I guess that\'s all I have to say  oh accept have a happy and fun holiday and fight those establish monkey bastards but mostly the first on and the second one. Bye  '),(3336,2311511,'02,July,2004','2004-07-02 00:00:00','   Today in class they showed us a film that was an amazing classic and no I am not talking about Deep throat in sex ed but I\'m talking about inherit The Wind in English class it was an amazing film about the scopes monkey trial . It has the guy from Bewitched and the second colonel from M*A*S*H* and it also has in a small cameo of the guy who played Mr. Roper from three\'s company . The only draw back of having a movie is that my teacher will always assign something to do with the movie . Now I\'m going to publish my top 5 lists first my top 5 movies (not in any order) The graduate , M*A*S*H* , Office space , War games & inherit The Wind    top 5 bands : Led Zepplin , The Doors , The Clash , Pink Floyd & Devo  top 5 babes : Debra Harry(when she was young) , Live Taylor , Paris Hilton , Samantha Fox (when she was young) & that babe from weird science ( when she was young  thanks and bye  '),(3337,2311511,'02,July,2004','2004-07-02 00:00:00','   Hey I was just appointed the head of the social committee for the end of term party a kind of odd choice well because I\'m the least social but atleast now I can make it non conformist and stupid . I\'m going to get my friend\'s band to play but I\'ll have to go through a ton of sweet talking well atleast I don\'t have any home work tonight .  '),(3338,2311511,'02,July,2004','2004-07-02 00:00:00','   Hi I\'m here and it\'s only 10 school days until my winter vacation or only 8 or 9 if you consider one day we will have a full movie day because lots of people won\'t be there because of a volleyball game all day and another day we\'ll have a winter presentation for the little kids  . So really it\'s only 8 days . Oh ya I forgot to tell you my grades for school well all you need to know is that I got all \"A\"s exept for one crappy \"B+\" in french . And it also it occured to me that nobody is reading this GOD DAMNED blog so I don\'t care about me I\'ll just     rawl into this corner . Hmm.. wonder if they bought it  oh damn I\'m still typing . well good bye to any one who\'s reading  '),(3339,2311511,'02,July,2004','2004-07-02 00:00:00','   Hi I\'m here and it\'s only 10 school days until my winter vacation or only 8 or 9 if you consider one day we will have a full movie day because lots of people won\'t be there because of a volleyball game all day and another day we\'ll have a winter presentation for the little kids  . So really it\'s only 8 days . Oh yah I forgot to tell you my grades for school well all you need to know is that I got all \"A\"s except for one crappy \"B+\" in french . And it also it occurred to me that nobody is reading this GOD DAMNED blog so I don\'t care about me I\'ll just     crawl into this corner . Hmm.. Wonder if they bought it  oh damn I\'m still typing . Well good bye to any one who\'s reading  '),(3340,2311511,'02,July,2004','2004-07-02 00:00:00','   Oh hey sorry I didn\'t post this weekend I spent a better part of the it playing Pro-glove what in the holy name of god is Pro-Glove you retarded homicidal maniac well Pro-glove is basically Frisbee meets tennis meets looking like idiots . Well anyway what you do is you find a glove somewhere or use your own but I  recommend you find one well anyway what you do is you soak the glove in water until it\'s sobbing wet and then put it in your freezer for days and then take it out then go to a tennis court and then look for the first line after the net on both sides then one person goes to the one side and one person to the another and  you have to throw the glove over the net and on to the other side over the first line and when your throwing it you can\'t step over the line . WARNING you should really ware puffy or thick clothes because well all I have to say for you is aaaaaaaaaaaaaaaaaaaaaaaaaaaaawwwwwwwwwwwwwwwww DEAR LORD HOW COULD I BE IN MORE PAIN AAAAAAHHHHH!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  Well good bye  '),(3341,2311511,'02,July,2004','2004-07-02 00:00:00','   Well my last post was stupid but I just wanted to let you know that I was alive well I am and that\'s bad news for you assassins trying to kill me you may think I don\'t know but I know I also want to say that I want to change my rating I want to withdrawal \"inherit the wind\" and change it to \"Kingpin\" and I also want to say that because of my selections of music and movies you may think I\'m really old well I must tell you that I am between the ages of 10 and 17 and I also want to say that yesterday we typed up quite a bit of the script so we are getting of to a good start in 2004 . Oh yah here are some shows I found that are very good and you should go watch A.S.A.P. okay here they are : aqua teen hunger force , The Brak show , Harvey Birdman attorney at law ( cartoon network ) , Unscrewed with Martin Sargent (Tech TV ) , Tough crowd with Collin Quin ( comedy central )  well bye , yall come back now BITCHES  '),(3342,2311511,'02,July,2004','2004-07-02 00:00:00','   well my last post was stupid but I just wanted to let you know that I was alive well I am and that\'s bad news for you assassins trying to kill me you may think I don\'t know but I know I also want to say that I want to change my rating I want to withdral \"Inheirit the wind\" and change it to \"Kingpin\" and I also want to say that because of my selections of music and movies you may think I\'m really old well I must tell you that I am between the ages of 10 and 17 and I also want to say that yesterday we typed up quite a bit of the script so we are geting of to a good start in 2004 . Oh yah here are some shows I found that are very good and you should go watch A.S.A.P. okay here they are : aqua teen hunger force , The Brak show , Harvey Birdman attorney at law ( cartoon network ) , Unscrewed with Martin Sargent (Tech TV ) , Tough crowd with Collin Quin ( comedy central )  well bye , yall come back now BITCHES  '),(3343,2311511,'02,July,2004','2004-07-02 00:00:00','   yea I\'m back bitches right here with my friend crack head Jimmy other wise known as Eli say hi jimmy \"hi jimmy gimmie some crack \" well that is about it I guess I really don\'t  have anything to say but screw the government screw martha Stewart and to hell with G.W. Bush  '),(3344,2311511,'02,July,2004','2004-07-02 00:00:00','   Hello every body well I saw bruce all mighty and all I can say is one word relidgousandgovernmentpropigandafueledbygreed so don\'t see it and my site of the post is quizmo it alows you to make and take quizes to see what kind of person you are and if you go to it my friend made a quiz to see what one our classmates you are . Right now I\'m going to see what movies are going to be on this week on pay per veiw . Oh yeah we did the beep test in gym today if you don\'t know what the beep test is it\'s a test where the teacher plays a tape of a bunch of tones and every time the tone is played you have to run from one side of the gym to the other and it gets faster and faster every time it\'s pretty sadistic but we gotta do it well bye.  '),(3345,2311511,'02,July,2004','2004-07-02 00:00:00','   Hello readers I\'m baaack well nothing too new well there is the ipod mini but I don\'t care I got my full sized pod right here who\'s name is ipoddy mouth but any way ibook owners are complaining about problems so watch out apple released new music software for full news go to www.slashdot.org now to release my site of the week www.srangereports.com this lets you play pranks on your friends over the net the site of the week will now be a normal segment on my blog and I call it a segment because I have shatered dreams of being a tv show host . well my dad left this morning to go back to Jamaica he sais that he will be back and staying in Canada by April but I doubt it though it sounds more likely because he\'s getting all his imigration stuff together and just so you know he\'s not Jamaican he\'s Australian but he does construction work in Jamaica because he get\'s paid better and he kind of has a reputation as an amazing project co-ordinator in Jamaica . ah well bye \" monfrares \"  '),(3346,2311511,'02,July,2004','2004-07-02 00:00:00','   Well I\'m going to tell you my age I know that it was a little stupid to keep it a secret but I\'m 13 . Hopefully you guys won\'t discount my opinions because I\'m young but I think that most people on blogger are young . Oh well that\'s all I wanted to say wait I forgot today\'s site of the post here it is www.kfcc.blogspot.com  '),(3347,2311511,'02,July,2004','2004-07-02 00:00:00','   just  '),(3348,2311511,'02,July,2004','2004-07-02 00:00:00','   trying  '),(3349,2311511,'02,July,2004','2004-07-02 00:00:00','   to  '),(3350,2311511,'02,July,2004','2004-07-02 00:00:00','   get  '),(3351,2311511,'02,July,2004','2004-07-02 00:00:00','   publicity  '),(3352,2311511,'02,July,2004','2004-07-02 00:00:00','   sorry that didn\'t come through right but if you go there you\'ll find a geek godess !!!!!  '),(3353,2311511,'02,July,2004','2004-07-02 00:00:00','   \"http://www.thinkgeek.com/images/action/large/20fe8b8.jpg\"  '),(3354,2311511,'02,July,2004','2004-07-02 00:00:00','   I have confirmed that my teacher ( Mr. Fraser ) was an idiot when he was a kid . This coclusion is based on the fact that when he was a kid he licked a metal pole in winter . I think these findings speak for them selves . I just got to chapter 5 of 1984 and now I know what double think is but I\'m not going to tell you what it means so  you\'ll either read the book or die or just be really uncomfortable because I\'m going to use it and you\'ll be lost and kind of annoyed . Double think double think double think double think double think double think . Nothing much happened this weekend except I downloaded a bunch of songs and didn\'t go out like a hermit well a hermit with a satellite dish , computer and six pack of coke . Well my site of the post is ipodlounge.com it\'s all ipod all the time . Well that\'s all for me Bitch!!!!!  '),(3355,2311511,'02,July,2004','2004-07-02 00:00:00','   In response to a post made by the person who posts the blog Kentucky Fried Cow and chicken in which he or she said that in Canada we have only one road so we are bound to have atleast one collage well I\'m going to have to tell you off you yank bastard my fist choice school may be Yale but my second choice is Megil\'s in Montreal which I must say if you are going to go to a collage in Canada go there because of all the strip clubs and you can smoke where ever you want like in Toronto hell the smoking laws are so strict that I don\'t even smoke and they annoy me . So I must say that on behalf of the Canadian people either get drafted or finish of collage in Canada P.S. The legal drinking age here is 19 not 21 so come and suck down some beers and smoke a couple packs .  '),(3356,2311511,'02,July,2004','2004-07-02 00:00:00','   Hey sorry I haven\'t been posting but I had to stay at my grandma\'s house for the week and she doesn\'t have a computer but thank god I had my ipod or I would have gone insane . On Wednesday we had our graduation pictures taken even though we graduate in 5 months but what are yah gunna do . I have one thing to you people about Valentines Day I hate it and it\'s not just because I don\'t have a girl friend but I just don\'t like how there is a whole holiday just for love it just seems that because  most religions are based on love and caring that the government is just breeding a subcontinent theocratic society so I\'m taking a stand and  making February 12 international anti-love pro emotionless physical attraction and sex day the only bad thing is that the name can\'t really be slimed down . Well my site of the post is www.quizmo.com good bye and screw Valentines Day  '),(3357,2311511,'02,July,2004','2004-07-02 00:00:00','   hey sorry I havn\'t been posting but I had to stay at my grandma\'s house for the week and she doesn\'t have a computer but thank god I had my ipod or I would have gone insayne . On wednesday we had our graduation pictures taken even though we graduate in 5 months but what are ya gunna do . I have one thing to you people about Valentines Day I hate it and it\'s not just because I don\'t have a girl friend but I just don\'t like how there is a whole holiday just for love it just seems that because  most religions are based on love and caring that the govornment is just breeding a subcontios theocratic socioty so I\'m taking a stand and  making February 12 international anti-love pro emotionless phsyical atraction and sex day the only bad thing is that the name can\'t really be slimed down . Well my site of the post is www.quizmo.com good bye and screw Valentines Day  '),(3358,2311511,'02,July,2004','2004-07-02 00:00:00','   Hello everybody I now realize that I\'ve been treating my Blog like a TV show but I now realize how obnoctios that is so I\'m just going to treat it as a normal blog . On Sunday I had to go to lunch with my little cousins which actually wasn\'t to bad but it was funny because I woke up about noon and as soon as I get up my mom tells me that we\'re going to lunch with my cousins so I thought well I guess it\'s not to bad but I don\'t realize that it\'s noon so she says get dressed and I\'m wondering why we were going so early because I estimated that it was about 9 or 10 o\'clock and well I suppose you have to be there . I did find that the guy who posts Kentucky fried cow and chicken is actually pretty smart he is a front runner in a cause that I find very interesting which is using emoticons to end sentences as opposed to a period when your using an emoticon  e.g. I like cheese :-) . Would be changed to . I like cheese :-) well I have found that I kinda like 80\'s pop not like Paula abdul but like nena you know 99 red balloons . Tomorrow I\'m going to get a hamster I\'m not really sure how that connects to 80\'s pop but meh . Another thing last Wednesday \"officer Frank\" came to our school to talk to us about drugs he was how I would describe as exremly over dramatic by the end of the talk I thought he was having an epileptic seizure with all the hand waving and such . Anyway we got to see real drugs incased in plastic which was cool because I didn\'t know how small LSD was well I guess that\'s good news I always did have trouble swallowing pills . Bye you retards oh sorry bye you mentally challenged persons :-)  '),(3359,2311511,'02,July,2004','2004-07-02 00:00:00','   What the hell am I going to say no one wants to be my enemy am I not good enough to hate is that it well any way fish be fine . I think my teacher is going insane he thought there was an amusement when there wasn\'t one . We have a student teacher she\'s pretty cool but not as attractive as I hoped she actually has been helping on and of for like the past month but this week has been the first she has been there non stop . She\'s also our art teacher . Ah alas it\'s the weekend my small packet of freedom . I\'m going to ask out Alison in my class I think maybe . Meh well if you want to be my enemy e-mail me  '),(3360,2311511,'02,July,2004','2004-07-02 00:00:00','   The hand is an amazing thing able to grab and crush with amazing power but able to cradle the most gentile and smallest things it can even take a whole bodies force for hand stands . That\'s beside the point if you don\'t think this is worth reading don\'t tell your friends that tell them to read it please if there is even anybody reading it tell people apparently no one wants to be my enemy . C\'mon tell people to read it  I invented a new word soupgold it\'s the rewards countries that are big get by \"helping\" little countries like America\'s soupgold for helping Iraq would be oil . See it are fun to use . I realize that living vicariously through a 11 year old might not be fun but it could be . Oh if you don\'t want to I don\'t care screw you you don\'t have to read it just tell people about it let them decide . Anyway have you heard little Bill\'s little plan yes I am referring to my enemy and yours Billy Gates well he wants to force everybody to change to a new standard he\'s doing this by making hotmail inaccessible if you don\'t switch this will be terrible if a small bushiness cannot comply * cough juggernaut * now I heard this by one ear when I had tech TV on in the background so my facts may not be so very true so if there are any incorrect facts that you see feel free to e-mail me at eprime666@hotmail.com and I\'ll correct it . Bye for now  '),(3361,2311511,'02,July,2004','2004-07-02 00:00:00','   people I\'m not finishing my own anarchist manifesto because I found that the socialisim that they were talking about is means of production e.g. land , mines , offices or large tools like bull dosers and such not the actual product so you can have your own shit but you can\'t become a corporate jugernaut with big stores and mines and stuff . thank you and good bye  '),(3362,2311511,'02,July,2004','2004-07-02 00:00:00','   \"What is Anarchy?  \"Freedom without Socialism is privilege.  Socialism without Freedom is Tyranny\"  Anarchism is the most misunderstood of political philosophies, perhaps because it threatens both the established order and those who would become our new masters after we have overthrown it.  Anarchists are constantly having to deny the images of them presented by both Left and Right.  Taken in by these lies or suffering from their own preconceptions, few bother to investigate the reality.  With this in mind we have put together this summary of what we do believe.   Anarchists seek to maximize freedom for all. We believe all people should be as free as possible to determine their own destinies and activities within the limits required by a respect for the equal rights of others.  This freedom should actual and practical, unhindered by unnecessary legal prohibitions or material constraints.   Anarchists are opposed to authority and hierarchy. All persons must be considered as equals.  No-one has the right to coerce or expect obedience from others except where necessary to protect the equal rights of others.   Anarchists are opposed to patriarchy. As a coercive set of social relations based on gender hierarchy, patriarchy oppresses and silences women in ways that we are as yet not even fully conscious.  Patriarchal structures must be destroyed wherever they are recognised.   Anarchists are not opposed to organisation. Anarchy is about organisation. It is about co-operation amongst equals, free of oppressive power relations. Often a lack of organisation allows oppression to go unnoticed and merely offers the \"liberty\" of the strong to take advantage of the weaker.  We must organise to prevent this.  We are opposed, however, to types of organisation which are based on authority and hierarchy or which involve the unnecessary regimentation and subordination of individuals or which strangle individual creativity.  We are implacably opposed to the centralisation of power.   Anarchists believe in the necessity of *direct democracy*. Where disagreements exist amongst persons which cannot be resolved co-operatively, the will of the majority must be respected.  Consensus decisions are always the ideal but when a consensus cannot be reached then democratic procedures must be employed.  Considering peoples as equals requires this. Representative and Parliamentary democracy are frauds which separate the government from the people, deny us control over our own lives and encourage apathy amongst the citizenry.  Real democracy places power in the hands of the people by making all decisions at the lowest possible level by voting in workplace and community councils.   Anarchists seek the destruction of the State. The State, a government which holds itself to be separate from and above the people, is always an oppressor. It has its own interests: its natural form is a bureaucracy and the military, the police and the security forces are its weapons.  \"Control\" of the state is an illusion which corrupts all revolutionaries.  We will not be free while the State exists   Anarchists believe in the need for a total Revolution. There is no aspect of the existing evils, Capitalism, Patriarchy and the State, that we can afford to leave intact if we seek to build a world free of oppression.  Like cancers these structures will reappear and destroy freedom if they are not sought out where- ever they are manifest and destroyed utterly and simultaneously.  Because our goals are radical, involving the total overthrow of the existing order, we cannot hope to accomplish them  by reformist means within that order.   Anarchists deny the distinction between ends and means. Liberation and revolution are our activities not our ends.  For this For this reason we will never be able to achieve freedom through authoritarian methods or destroy the State by seizing control of it.   Anarchists follow no leaders. No-one can lead us to take responsibility for our own lives.  Only we can liberate ourselves. The only \"leadership\" we recognise is by example.   Anarchists believe in a brighter future. We envisage a future free of oppression, of people living in community and in control of their own lives. We see a society governing itself though workplace and community councils, making decisions at the lowest possible level and cooperating and organising together.  We believe in a society where \"economic\" decisions about production and distribution which affect all of us are made democratically rather than left in the hands of a privileged and unelected few.  Free of the destructive imperatives of capitalism, we will be able to live in harmony with the environment, use technology to our own advantage rather than our bosses and escape from the threat of war derived by the needs of Capital and State to extend their influence.  We will build our own future.   Anarchists know that it can work. We gain strength from the examples of anarchist revolutionaries at the forefront of progressive movements throughout history.  We learn from the Spanish Civil War where peasants and workers seized control of large areas of Spain and organised them according to anarchist principles.  We learn from other existing anarchist organisations and from our own experiences as we attempt to put anarchist theory into practice in our own lives.\"   this is all what I believe but there was a part that I left out about anarchy being socialism kind of so I want to say that what I believe is this and not the other stuff so I will say I\'m reformed anarchist and I will complete the reformed anarchist manifesto in my next posts or two  '),(3363,2311511,'02,July,2004','2004-07-02 00:00:00','   \" sex sex sex , sex and slapstick that\'s all it is \"                                           - Moliere plays paris  if only it could be  '),(3364,2311511,'02,July,2004','2004-07-02 00:00:00','   next week is march break ha ha ha ha ..............hmm well good bye  '),(3365,2311511,'02,July,2004','2004-07-02 00:00:00','   sorry I haven\'t been posting in a while I been spending alot of time at the kfcc shout box I\'m pretty sure  everybody there hates me I think mostly because of my habbit of dwindling on things .  '),(3366,2311511,'02,July,2004','2004-07-02 00:00:00','   Hey everybody sorry I haven\'t been posting in a while but I have been trying to get my website up , yes that\'s right a real website but don\'t worry I still will post because this is a different part this is my real life but my site is all about my work (cartoons and anarchy) the toons are rough but you\'ll get the point . I even have tutorials to make smoke bombs . The site is www.freewebs.com/evantoon  '),(3367,2311511,'02,July,2004','2004-07-02 00:00:00','   Hey everybody I just watched the episode of M*A*S*H where colonel Blake was killed and I must say no matter how many times I see that I am tempted to cry but then I realize that it\'s just a tv character but it still is sad because that show had some of the best tv writers ever . It\'s always presented creatively and well writen and flowing not matched by any show I\'ve seen .  '),(3368,2311511,'02,July,2004','2004-07-02 00:00:00','   ignore the cyberanarchist post because it was slopily writen though I am I tried to delete it but I didn\'t see it as deleted so if it is deleted for you then ignore this post so there is something for the whole family to ignore  '),(3369,2311511,'02,July,2004','2004-07-02 00:00:00','   hello everybody today I\'m taking a drum lesson from a hippie guy downstairs in my building . Guess what Alison in my class (she\'s the hotest girl in my class) just got braces hehe haha well you may be wondering why I\'m laughing well I kind of hate her and like her well it\'s complicated but anyway she\'s getting braces remember you heard it here first unless you were with Pema ,me ,Eli and Alison in the chat room then you really heard it first but who ever is reading this is one of the first on the net to see it .  '),(3370,2311511,'02,July,2004','2004-07-02 00:00:00','   don\'t get the idea that it\'s going to be some weird militarialistic blog with tutorial on how to bring down the Government . It\'s still the same blog just olive drab . Same weird shit and quotes from obscure plays and my life  '),(3371,2311511,'02,July,2004','2004-07-02 00:00:00','   don\'t worry it\'s still the same blog but different design and officialy now the film is scraped so meh......  '),(3372,2311511,'02,July,2004','2004-07-02 00:00:00','   Hey well......um ya the play is progressing well . I have to have my lines memorized by Monday or my ass is in a sling . If I have them memorized by Tuesday I only get my left butt cheek in a sling . In those commertials about ditec debt management why do they always throw away the credit cards ? why don\'t they smelt them down and make a statue resembling a pagan god to pray to ?  '),(3373,2311511,'02,July,2004','2004-07-02 00:00:00','   I know I haven\'t been posting for a while but I was making my site www.freewebs.com/evantoon it is all about my cartoons that I do and there is also my class site that I\'m helping with but I won\'t release the url until it\'s done . We got our parts for the play last week I\'m judge collins the cold judge who wants to kill criminals for minor infractions and I\'m also Lysander (which is a bigger role than Collins) who is love struck because of a fairly love potion (if you want to know more just look for the play mid summer night\'s dream) this play is a compilation of a bunch of plays so you wont be able to read it if you want to .  '),(3374,2311511,'02,July,2004','2004-07-02 00:00:00','   Hey everybody it\'s 12:55 on sunday and I still haven\'t memorized all my lines but don\'t worry I only have 9 more and they are small . Last night on MSN I went insane on a girl that we were doing the site with because she put something that could come across ant-semetic on her page , but it wasn\'t the comment it was the fact that when Eli and I found out about it I deffended her because Eli wanted to take it out because he was Admin even when I thought that she was a racist 100 percent (now I\'m only about 20 percent sure she\'s racist) because of free speech which I beleive in to the extent of getting killed but it sparked something inside me when I deffended her oppinion that freedom could be to free but then I just decided what I always decided that everybody even blaton racist have to have freedom to speak their minds . It can\'t be that freedom is for the only ones who agree with people .                                                                                                                  good bye and this little experiment of world government with Eli playing the dictator is running me ragged and on that note that\'s why anarchy works  '),(3375,2311511,'02,July,2004','2004-07-02 00:00:00','   yayayayayayay I finaly got the archive working again and I got links now to . I am now done with my renovations so you will now refer to me as MASTER OF THE HYPERTEXT!!!!!!!!!! just joking but this is a very liberating expeirience for me considering all the other times I tried coding a failed misurably .  '),(3376,2311511,'02,July,2004','2004-07-02 00:00:00','   ok ya its basicly not archiving but I don\'t give a shit just read my blog or you\'ll be lost  '),(3377,2311511,'02,July,2004','2004-07-02 00:00:00','   I just learned som HTML and well I recomend you don\'t fool around with the archive section because I\'m pretty sure I just deleted my whole archives which is bad for those new readers (ya right I\'ll be lucky if there are any readers) but I will say that I\'m just a 12 year old who is writing a blog about their life so when they find me dead they\'ll have something to pry out of my cold dead hands I\'m not sure if the blog is still archiving and collection the info . Since I know a little HTML now (stress little) I\'ll be putting a counter on and maybe pictures if I can but don\'t worry now I\'ll be able to improve my other sites because frankly there isn\'t much you can do with html on a free blog exept change colours make links and stuff but maybe my blog won\'t suck now (I doubt it)  '),(3378,2311511,'02,July,2004','2004-07-02 00:00:00','   well I have been reading blogs for a couple hours now and I must say in comparison to some this is driveling shit ofcourse compared to others it\'s amazing but mostly shit anyway I found this amazing blog about a guy who manages a strip bar and if you are planing to read alot of the posts download or get the song night moves by Bob Segar you\'ll know why . To get the link just go to a great blog on my links and you\'ll see it  '),(3379,2311511,'02,July,2004','2004-07-02 00:00:00','   today was pretty boring nothing much happend well Mr.Fraser gave us a boring speech about home work and the play but ya thats about it oh ya IT\'S FRIDAY IT\'S FRIDAY IT\'S FRIDAY YAYAYAYAYAYAYAYAYYAYAYAYAYAYYAYAYYAYYAYAYAYAYAYA!!!!!! oh I forgot to tell u guys (if anybody reads this) that I saw pulp fiction last weekend . We had our Dance-a-thon yesterday and it sucked monkey shit . I don\'t know why I bother to post no one reads this I know it and if you are reading it then thank you very much but appart from maybe 10 people who have seen it once there is no one reading this no regular readers and who wants to read this shit it\'s just my life infact you know I don\'t even read other blogs . I should talk about my oppinions more not my stupid little life . What is there to even say about my oppinions they mostly start and end in the phrase death to the government . I guess this is only what I think be it my oppinions or reveiwing what happend to me in my life but I don\'t know if my mind is interesting enough for you guys well lets try . I am going to put out a comic book and in it there is going to be a spoof of the peanuts comic with a nazi snoopy with a heroin problem . I\'m going to cruise some other blogs to see what they are like so maybe by the time I\'m 30 I can win the bloggie or atleast be in the running or atleast have more than 10 readers bye.  '),(3380,2311511,'02,July,2004','2004-07-02 00:00:00','   hello everybody for the last past couple of weeks or so there has been this girl from Mr.Fraser\'s(my teacher) past class to help us with the play and I just kind of realized today that I have a little crush on her . I\'m pretty sure she\'s a junior in high school and I know I have no chance but c\'mon she\'s amazing she has these green streaks in her hair and she always wares these great punk t-shirts and she has a pair of doc martins so she\'s not only a punk but a real punk not those poser punks who like good charlotte well I actualy don\'t know her musical preferance but I\'m pretty sure she\'s actual punk .  '),(3381,2311511,'02,July,2004','2004-07-02 00:00:00','   hey everybody I just got yelled at by my teacher for being an anarchist . ok fine it wasn\'t because I was an anarchist but it was because I said that if we got school uniforms it would turn our school into a facist police state . Well fuck him he tried to convice me that they were good well all I say is fuck him! oh well he\'s a good teacher but a capitalist dog . that\'s what happens when when your employed by the government . well I also want to say that I found this great show it deals with political issues and it pokes fun at everybody who\'s liberal or republican it\'s called this just in watch it or a great rath shall cease you in your steps ! just joking.....or am I? bye bye  '),(3382,2311511,'02,July,2004','2004-07-02 00:00:00','   well I haven\'t been posting for a while but now I\'m back baby! I was just kind of chillin mostly because I was making my own site which I should have the link there still and my new page on my class site (it isn\'t very good ) which is kinda not finished but you can go to I don\'t give a shit it\'s www.freewebs.com/room30 and my page is Evan\'s page of truth!!!!!!!! I have also found this great anarchist forum here it be www.pub44.ezboard.com/fanarchismfrm1 well that\'s about it but don\'t worry I\'ll be posting regularly (not that you fuckers care)  '),(3383,2311511,'02,July,2004','2004-07-02 00:00:00','   I saw kill bill fucking amazing thats all I have to say fucking amazing . Stefani is mad at me because when pema (one of my freinds) was over we made an off colour comment about where she could stick a spoon and she has blocked me on MSN messanger . oh well I guess I can\'t say sorry cuz I can\'t contact her (stef if you are reading this then I\'m sorry Domket groggy groggy).  '),(3384,2311511,'02,July,2004','2004-07-02 00:00:00','   I changed my links and now I didn\'t put my cartoon site up because I\'m saying fuck off to it and starting clean . So by the end of this weekend you should have a new link in all of your repituars . I took an IQ test , well actualy a couple of them and I\'m averaging a 122 to 123 IQ which if you didn\'t know is good is above average and boarderline gifted but unfortunately I\'m not a genius I am very intelligent but not a genius . The crush on the assistant teacher is about fading I still think she\'s hot but any real fealings are gone so thats a message to you Alison,stefani and the rest of the class I don\'t have a crush on her anymore :p up yer\'s . Oh ya today\'s my birthday I actualy am 12 now because as soon as its april I start saying I\'m 12 just to tell you so you know I wasn\'t 12 when you thought I was .  '),(3385,2311511,'02,July,2004','2004-07-02 00:00:00','   well for the last while I\'ve been following anarchisim like a religeon and wondering if I should buy things because it would be so capitalistic but now I have come to the conclusion that if I want something I\'ll buy it . Anarchisim is not about following ruels it\'s about doing whatever the fuck you want because the government has no right to tell you what to do . It\'s not about being a minimalist because you are wondering if you\'d be a hipicrate if you want something . We got new seating arangments for my class and now I\'m in the direct line of fire between Stefani and Alison who talk insane amounts but I guess it\'s ok I really don\'t care .  '),(3386,2311511,'02,July,2004','2004-07-02 00:00:00','   tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov tnod etov . Now that we have that out of the way I want to say that from now on whenever I see a picture of good charlotte I go into spasims well not so much as spasims as a disscusted sort of shaking like if you were to see a dead animal roting  . It\'s kind of disrupting me in class because whenever one of the girls in my class pulls out a stupid pop magizines near me I go into convolsions . good bye oh and if you didn\'t get it tnod etov means dont vote  '),(3387,2311511,'02,July,2004','2004-07-02 00:00:00','   well I made a promise today that because of stomach pains I would bitch about it to anyone I came in contact with so I thought this would the ultimate wyning machine so now my stomach hurts uhhhh its like a bunch of knives sticking into me uhhhh . there I\'m done ok now I want to say that because a bunch of people in my class have seen My blog they are joking at me for the crush on the high school girl but all I hope is that she doesn\'t find out because it was fine when some girls in my class knew they just made some stupid joke and that was the end of it but now that some of the guys (dickless retards) know it I think they actualy might tell her that I have a crush on her which would make it uberuncomfortable and the worse thing is that we have this other girl in our class with the same name and when they first hear it they think it\'s her and thats embarasing . (I really shouldn\'t have given those whores my blog url then this wouldn\'t have happend)  '),(3388,2311511,'02,July,2004','2004-07-02 00:00:00','   I think I might just abandon my cartoon site at free webs and start another one or even post my comics here (but I\'ll still probably have a site just for my toons) . I just wanted to start devoting most of my time with my blog for some reason .  '),(3389,2311511,'02,July,2004','2004-07-02 00:00:00','   well I was staying up late as I always do on saturday nights last night and I noticed that they were doing a little show on Much Music that was called much in your space where they give the person and the person\'s room a makeover so it fits the theme of their favourite band which in this case was good charlotte ,now if you have read my blog before you probably noticed I said some bad stuff about good charlotte well I want to quote someone one much that I actualy agree with to explain my point \"Is this what punk has come to ya know they ask why no one ever notices superman behind clark kent\'s glasses well I want to know how do people not notice the Backstreet boys behind a layer of mascera\" - Ed the sock now all I want to say is AAAAAAAHHHHHHHH!!!!!!!!!!!!!! I hate Avril Lavene because she\'s a poser of the skater culture well I hate the skater culture and when there are posers of the punk culture that I really like well I\'m tempted to lose it . On the show they dressed the guy up in a Ramones T-shirt WHAT!?!!?!? how retarded is that the closest good charlotte ever came to the ramones was when one was dating a girl with a leather jacket now c\'mon when are we going to pull the wool away from our eyes and say hey get off the stage you pussy posers . This is exactly like creed imitating Pearl Jam exept this is worse creed actualy had their own style and just butchered the music a bit but these guys butchered the culture to what was expected of punk and then used that image to make crappy music that sounds like your garden veriety rock/pop . From now on any one who likes good charlotte then don\'t count your self as punk count your self is conformist little terd . The punk movement is about being your self and being an individual and not being exepted ,and by the by just because you are a good charlotte punk doesn\'t mean you are an anarchist if I see a god damn little poser punk bitch putting an anarchy symblol on something and not knowing what it means well I\'ll cry then slit my wrists . These stupid retards are going to drive me to spill blood  '),(3390,2311511,'02,July,2004','2004-07-02 00:00:00','   well I am a anarchist and well I would like to say on behaf of my self and any other anarchist out there that are like me that just because liberals are more free doesn\'t mean that we are like liberals infact I would say I have a more conservitive veiw of the world and just because liberals and anarchist beleive that everyone should be equal then that doesn\'t mean that they are like us anymore than republicans that beleive we shouldn\'t pass bills protecting the environment because anarchy is freedom and equality and I hate facist conservitives but I hate facist liberals to that want to be politicly correct and censor everyone . You see conservitives get a bad rap just because everyone sais to ,I would take a true inteligent facist conservitive over an ignorant conformist liberal any day I might not agree but atleast I would respect them . I object to Bush because I actualy know that he\'s a gun toating idiot that can\'t take any critisizim of his administration and that to every problem his responce is \"Nuke\'m\" Iraq \"Nuke\'m\" Afganistan \"Nuke\'m\" Econemy \"Nuke\'m\" well you get the idea I hate him all because of this not because charlie sheen tells me that Bush is a poopoo head . I want to clear up that though I am not a pacifist I hate war its the most retarded thing anyone can sent people to do ESPECIALY when they have never been to war them selves (if you can\'t tell Bush over your yells of \"Nuke\'m\" that I am talking about YOU) I just hate conservitves because they are the closest to having church and state joined and hating abortion and I\'m an athiest I\'m not denying my personal beleifs influence but I think that I am the one with the clearest vision so I want to say : Let the Queens get married (sorry if your gay and you don\'t like the word queen but its my way of hating the censoring liberals and the facist conservitives) let the bitches kill the babies and Fuck the church! so to recap hating religous extremist Conservitive and politicaly correct Liberals who censor so just shut up I hate you all!!!!!!! Liberal or Republican you can all join in hands and say that I hate you ! so don\'t label me as a more extreme Liberal I\'m an ANARCHIST , Ta Ta  '),(3391,2311511,'02,July,2004','2004-07-02 00:00:00','   I\'m branding Donald Rumsfeld a war criminal personally on the grounds that he knew all this torcher was happening but refused to aknowlage it because he didn\'t want to lose support in Iraq so I will officially call for his resignation and hopefully he will peacefully step down because of my wishes but I will also request a indepth investigation into not only these atrocities but possible other war crimes committed by the U.S. I will make up a list of UN inspectors and investigators to look into this and I must say the only good thing that came from these terrible events perpetrated by the U.S. on the Iraqi people is the light that it sheds on America\'s \"good guy\" persona to find the real position . Hopefully the government won\'t \"pencil whip\" these events out of the world\'s eye . Good bye   (pencil whip means to use the position of power one holds to dismiss charges that they have committed )  '),(3392,2311511,'02,July,2004','2004-07-02 00:00:00','   Eli again. I was with a friend yesterday and while we were playing ball hockey his ex was walking donwn the road. First she was with her sriends and they were laughing lie crazy. And she yelled \" hey there hot stuff*  to him .we ran as fast as we good to my house. I went outside to check if they were gone.And they were. So we continued with the game. Later though they walked bythe road again still laughing like mad and we didnt have tyme to run. They yelled all these stupid coments and couldnt walk in a straight line. It was obvious that they are underage drinking. Stupid underage drunkies.  Well anyways in our class not that much has happened. Exept I asked Alison out. She said no. She only goes with people she likes.  '),(3393,2311511,'02,July,2004','2004-07-02 00:00:00','   hey everyone I wanted to go full on this weekend and replace my computers RAM and hard drive as j00 guys probably know but I wanted to get the full expeirience so I stayed up till 4 (in the morning)researching ram prices and writing the script for volume 2 of Mega Tokyo but I am only on like 4 or 5 hours of sl33p and now as I watch clocks I have no preseption of time what so ever what might s33m lik3 5 seconds to me is actualy like a couple minutes or even like 10 minutes . good bye  '),(3394,2311511,'02,July,2004','2004-07-02 00:00:00','   Hi Everybody. Eli here. I just got the invite to this blog so I decided to post. Well, Im doing the Megatokyo film with Evan. Im writing volume and he\'s doing 2. Its the first Quentin Tarrantino inspired techno film.  '),(3395,2311511,'02,July,2004','2004-07-02 00:00:00','   well I finally added Mega Tokyo on my links page and I finished all the Mega Tokyo strips . I\'m making my compy (I call it emacioveli) l33t with new RAM and a $l4v3 drive sorry I get all l33ty when I  talk about computers but don\'t j00 worry I\'m putting all of this to good use because I\'m getting unreal tournement 2004 for the m4c (well maybe) it\'s also because of the MT movie that I\'m making I have to edit it on here (here being emacioveli) . 8i b1  '),(3396,2311511,'02,July,2004','2004-07-02 00:00:00','   I haven\'t put mega tokyo up in the links section yet mostly because way to much mega tokyo I\'m addicted |)43 $3\\/3|\\| the l33t has sunken into |\\/|1 8r41|\\| I\'m n0t 3v3n aware when I t1p3 1n l33t I am on strip #426 only 128 strips more to go b|_|t 1 c0uld do 1t 1|\\| m1 sl33p inf4ct I\'|\\/| no7 sur3 I didn\'t on so|\\/|3 . I h4v3 b33n neglectin9 alm0$t every on3 of my d|_|ties (blog,reading other blogs keeping up on news) though for some r34son its as if 4s I\'m reading m3g4 t0ky0 everything is going faster cuz w3n I\'m waiting for Kevin to p0$T on tj\'s place he only does it like once a week and w3n I\'m re4ding mt he posts everyday and n07 ju$t little post$ big h0n|  '),(3397,2311511,'02,July,2004','2004-07-02 00:00:00','   well I\'ve been reading way to much Mega Tokyo (which you all should read atleast once) and don\'t worry I\'ll put it on the links page and I had a movie idea for mega tokyo so my freind and I will make a independant film and send it to sprokets (an all kid made film festival)  but all we have to do really is get permission from the author to use his characters . |31  '),(3398,2311511,'02,July,2004','2004-07-02 00:00:00','   7|-|15 p057 vv1|| |3 |3r0|_|9|-|7 70 joo 1|\\| |337 5p33k p|-|0r 7|-|053 0f j00 vv|-|0 |)0|\\|\'7 |  '),(3399,2311511,'02,July,2004','2004-07-02 00:00:00','   well stef isn\'t mad at me anymore well atleast not as mad as before . Today I was on Queen street west and used some birthday money to buy about $20 worth of comics (I want to be one of those guys that spends all their cash on comics,beer and art zines)and I also baught some art supplies and the kill bill sound track I\'m listening to it now and it kicks ass . I got scar face for my birthday and my freinds dad was all nerotic about giving it to me as a present because well thats how he generaly is he gets worked up about almost every little bit of \"inaproprietness\" . I was watching last nights episode of saturday night live and it was really funny y\'know Lisnsey Lohan is actualy not such an uptight little goody goody like in one sketch it was about harry potter when hermioni comes back to hogwarts and she has big boobs from over the summer and harry and uhh....what ever the fuck the other guy\'s name was kept drooling over her it was sooooo funny . Ta ta mother fuckers  '),(3400,2311511,'02,July,2004','2004-07-02 00:00:00','   well it looks like I have been kicked out of the class blog \"The Wall\" it\'s either that or sometime last night when I was half over caffienated and half exosted I got angry because I didn\'t understand quantum physics right away then deleted it but that one doesn\'t seem very likly mostly because I can understand the principals of quantum physics . I think I might start a knew blog a fake one just in case my parents start looking for one like my dad he\'s pretty tech savy and I know that if I have a blogger bookmark he\'s going to get suspicious so I could dangle in in the wide open kinda and then he\'ll  just say oh look at that my son knows some html thats good and think nothing more of it but if they find this thing it\'s going to be like balisticsville man because those squares are going to come down on my buzz . now goodbye all you cool cats and catettes  '),(3401,2311511,'02,July,2004','2004-07-02 00:00:00','   I was thinking about if I went on a vacation for while and couldn\'t blog which person would I get to guest post on it . I know that it wouldn\'t be any of my freinds because they would either not post regularly or just say stupid random shit \"like yum yum bumm cheese\" . So I started looking at other bloggers like the girl who has been posting comments here but her blog is waaaaaay better than mine and even having her post comments here is amazing and there are other blogs but most of them are either way better or are good but not like mine . Of course the very fact of having a guest blogger is a bit weird I mean it\'s supposed to be all about my life . Yknow what I\'ll think about the guest thing but I\'ll probably just end up using audioblogger . 81 81  '),(3402,2311511,'02,July,2004','2004-07-02 00:00:00','   I just decided to put the title line in my blog so you can find stuff easier it\'s definatly better than just having the first line of it  '),(3403,2311511,'02,July,2004','2004-07-02 00:00:00','   oh ya if you have any problem with my anarchist veiws in the last post , comment you facist son of a bitch .  '),(3404,2311511,'02,July,2004','2004-07-02 00:00:00','   yknow I don\'t think I have put anything on the class site since like emmanuel (a retard jerk ass homo in my class) broke into it and I had to change the pass and rag on him on my page . I can\'t beleive we actualy thought that site was good . It\'s kind of sad one minute it\'s the corner stone of your attention and then it just fades into everything else you may think you\'ll love it forever or give it the same amount of attention when you made it but your not it will eventualy just fade into the backround like everything else and you finaly notice this trend and you think that you might not be able to focus on anything in your life and you give exuses to your self but in the end everything fades into the backround nothing can really be valued higher than anything else . This is why people get divorced this is why there is no world peace because shit heads can\'t get it through their skulls that nothing truly matters well maybe a few things for your own self that you will never lose interest in but as collective nations you can\'t have real interest even if they are war crimes or anything . Look at afganistan when\'s the last time you heard anything about afganistan ? Sure you hear the occasional blurb like \"we apreciate our brave troops serving in Iraq and afganisan\" and that\'s if they remember to mention it but everything just melts into the fog thats why we have to do all the things that we want that we love and we can\'t let the world governments control us we must rebel we must get our freedom from their theocratic \"there is something greater\" tyranies . VIA VIA VIA!!!!!!!!!!!!!!!!!!!!!!!! (don\'t let my exesive use of exclamation points fool you I\'m very serious about this)  '),(3405,2311511,'02,July,2004','2004-07-02 00:00:00','   anti-social    urlLink   Which Personality Disorder Do You Have?     brought to you by  urlLink Quizilla   here\'s my results for \"which personality dissorder do you have?\" quiz that I found  '),(3406,2311511,'02,July,2004','2004-07-02 00:00:00','   Your True Nature by  urlLink  llScorpiusll        Username           The quality that most appeals to you:    Sense of Humour      In a survival situation, you:    Do what is necessary      Your hidden talent is:    Endurance      Your gift is:    Athletic ability      In groups, you:    Observe others\' behaviour      Your best quality is:    Your creativity      Your weakness is:    Your timidity             Created with the ORIGINAL  urlLink  MemeGen  !       well it isn\'t to bad though I do not have athletic abilty and the only thing I can edure is a Quinten Terrintino movie marathon :)  '),(3407,2311511,'02,July,2004','2004-07-02 00:00:00','   oh ya I want to say to those people who are just reading my blog for the first time ever that not all the stuff that is written is mine part of it is my friend\'s who was on here for a while but I kicked him out because of to much self indulgent shit about hating rush and his band playing at garden fete . That creative story stuff was his too . Comment on my posts god damn it there are people reading this god damn blog I know they are I have counter so just post a comment or 8 I already have one even if it\'s just to whore your own blog put the link in !!! now you sons of bitches better make comments or I\'ll nail your genitals to a tree with a rusty rail road spike then rip off your head and d3ffic4te in j00r neck stump , oh sorry its been really hot here in T.O. and I\'m not ajusting to the heat from the cold well it\'s taking it\'s toll on my brain so comment or I\'ll go nuts-o-futs-o- with 30 degreas centigrade under my belt .  '),(3408,2311511,'02,July,2004','2004-07-02 00:00:00','   well I found a free way to post my pics on blogger but it only works with windows so all us linux and mac people are totaly and utterly fucked  '),(3409,2311511,'02,July,2004','2004-07-02 00:00:00','   you may be saying to j00r s3lf that all you need to do is go to blogger help and lay a little code well its not that easy dick face for example the site that I am putting my toons on for some reason doesn\'t show the cartoons when I put it in . Oh well I guess I\'m just going to have to go on a rampage with black market small arms and kill everyone at the blogger office because we can\'t upload pics . Blogger you are aware that yafro.com lets you post tons of pics but I don\'t use it because you can\'t work with html so anyone who is not that great with html and wants to get their pics and message out there then go to yafro.com NOT blogger because they are shit heads and you can only upload pics if you have external hosting I\'m only 12 I don\'t have a god damn server!!!!!!!!!!!!!!!!!!!!!!! j00 d34d f00z aaaaaahhhhhh!!!!!!!!!!!  '),(3410,2311511,'02,July,2004','2004-07-02 00:00:00','     '),(3411,2311511,'02,July,2004','2004-07-02 00:00:00','     '),(3412,2311511,'02,July,2004','2004-07-02 00:00:00','   now I\'m devoting all my time to geting my comics on here so you can see them and love me .  '),(3413,2311511,'02,July,2004','2004-07-02 00:00:00','   well eli is attempting to put together a class blog instead of the sit which sux0r3d b3yond belief but I don\'t think the girls will come cuz for some reason they have somthing against blogs even stefani who I thought would like it and alison I know won\'t come because she didn\'t want to start one when I suggested it to her (which is weird because most people try one don\'t get past 2 p0sts and just disc4rd it like I did before this one) so I guess eli is screwed unless some other girl in th3 class will come .  '),(3414,2311511,'02,July,2004','2004-07-02 00:00:00','   I don\'t like sharing my blog I am an only child I\'m greedy and selfish I can\'t stand this . This used to be where I would say everything I wanted to say but now it just seems tanted I guess I don\'t like relying on anybody and if eli j00 c4n see this phuck j00 I\'m kiking you out elol.  '),(3415,2311511,'02,July,2004','2004-07-02 00:00:00','   Yes, I feel very guilty. Now Il go cry in a corner.  RUSH SUCKS 2004 ALL THE WAY! W00T! W00T! W00T!  My Band S.C.C ( Spontianusly Combusting Children) is playing at the school\'s Garden Fete. And I get to improvise a Solo on my YAMAHA! Three cheers! HIP! HIP! HORRA!  '),(3416,2311511,'02,July,2004','2004-07-02 00:00:00','   I want to say that though the knew templates are cool ass ! I must say that because I don\'t know that much HTML I can\'t do shit with them even though they might be cool I still liked having my own personalized template but since this is a multiple user blog I needed to change it cuz the other one didn\'t display our names when we posted so j00 couldn\'t tell who it was who posted what but I think you guys have known me so j00 c4n detect eli from me but still . I think I liked to have a single person blog a bit better and even know if we go back to the good template I\'ll have to do everything over because it deleted when I changed the template (does that make you feel guilty enough eli?)  '),(3417,2311511,'02,July,2004','2004-07-02 00:00:00','   Your True Nature by  urlLink  llScorpiusll        Username           The quality that most appeals to you:    Courage      In a survival situation, you:    Freak out      Your hidden talent is:    Spiritual wisdom      Your gift is:    Ability to acquire wealth      In groups, you:    Act as host/ess      Your best quality is:    Your creativity      Your weakness is:    Your lack of sensitivity             Created with the ORIGINAL  urlLink  MemeGen  !        Geez. That\'s scary. It\'s almost completely acurate   - Eli  '),(3418,2311511,'02,July,2004','2004-07-02 00:00:00','   Le Eli et Ici.Time for some creative writing.  True part  Alright yesterday At night I was at this party and there was a reaggae band playing there and they were really good. But I noticed something very strange theyre drummer wasnt a complete jackass! He wasnt drumming as loud as he could! He was actualy good!  False part ...why does it have to be false :(   Me and my band ( Spontaniusly Combusting Children) searched for why this was so. It was realy spooky. Me and the gang split up. We created a trap we but beer all over his drum set and farther awy in a line so hed drum slip fall and fall down the floor into a sticky web and bellow was Lava. But suddemly the other part of the gang were looking around and Drew lost his glasses. The drummer almost had him my beating the drums so lightly quickly Robert took his drum sticks and banged the drums causing Pampa the panda who was tackling the drummer to be the one who slipped. Unfortunately, it turned out that the web was not strong enough to hold a spider.So our Panda fell into lave and died. We mourned and then said \" What would quentin Tarrantino do?\". And we found out what to do................  TO BE CONTINUED..................oh screw it I\'m finnishing it now!  Adam took his base tore off the tuner part and there was a blade. I tore the tuner part of my guitar and it was a machine gun. Then Dimitri did the same with his guitar and it just became a plastic baseball bat. We gave him another guitar and it bacame and Adrienilen Needle. We just decided for him to use the plastic baseball bat. Phillip our keyboard player played rush for a weapon. We had to have earplugs or we\'d die.Then at the same time we all atacked the drummer. The rush music killed him. His head popped off like God was squeezing his stomzch as hard as he could.Bllod srited out like a firehose that couldnt stop and was out of control. Then he had a very spontiaus Combustion which was very bloody. And by Very I mean like putting 20 people in a small white room in one space then mosing them down with a Minigun. Then suddenly an Alien Spontaniusly Apperared! He said hello I am an Alien guitarist who just wanted to try the drums. We said \"oh\" and carried on with our lives.  '),(3419,2311511,'02,July,2004','2004-07-02 00:00:00','   I think this whole weekend I might get 9 hours of sleep (not an average but 9 hours for the whole weekend) I wanted to see what I\'m like when I\'m sleep deprived I\'m going to sleep under my comput3r desk with nothing but c0ke , Life cereal and pizza b4g3ls in my stomach . I w4nted to take computing to the next l3v3l so I think I\'ll be installing new RAM tomorrow if I can get my mom to go with me to a computer store on mother\'s day . PH34R MY CR4Z3 4SS SK1llZ bi bi  '),(3420,2311511,'02,July,2004','2004-07-02 00:00:00','   yknow I\'ve been wondering about some questions about life like what do they do with the limb after they amputate it like I could see you studying a heart or a liver or something but an amputated leg kinda seems weird like would they close up the severed part if they were going to study it or just let the blood freeze over or what and why would you really need to study just a leg wouldn\'t you want the whole cadiver experience so the students could steal it dress it up in clothes and bring it to the home coming game and then you\'d have to take it to a party then this really drunk girl would kiss it then maybe have sex with it and then she\'d be sober and the people who stole it would tell her and then she would freak out and drop out of school and shit , you just can\'t do that with a severed leg though there are alot of hallowean prank possiblities  '),(3421,2311511,'02,July,2004','2004-07-02 00:00:00','   I forgot to tell you about this blog that I came across about this girl in college and we\'ve been posting stuff on each of our comment sections ( like you wouldn\'t notice since she\'s the only one that posts here!) and so I decided to put her on my links section , I really should have before but I\'ve been so busy doing absolutly nothing and lazyness is a harsh and strict mistress . I\'m sure most of you guys will like it for the most part it\'s exactly like this blog exept cool interesting and looks good oh ya and there aren\'t constant posts bitching that noone is posting comments ;) so good bye and I leave you with this peice of information if you work half assed you have a half an ass you can use for your family  '),(3422,2311511,'02,July,2004','2004-07-02 00:00:00','   well here I am 4 in the morn with 3 cokes and 2 peices of pizza ( soon to be 3) are in me and the first stages of my caffienated paranoia is starting . It\'s comforting in a way I\'m not sure what way but I\'m sure there\'s some way . I hav found some new comics and will be posting linkage to them soon . I\'m on the computer but also looking at the tv which is on Much Music for hope that the \"hey ma\'ma video will come on because it\'s really hot it\'s really the only song I like from the blackeyed peas because everything else was crap and made me want to vomit on poo poo . I think some slight carpel tunnel is starting to show it\'s self well thatz 4b0ut i7 so 8i b1  '),(3423,2311511,'02,July,2004','2004-07-02 00:00:00','   yknow I was just cruising some blogs and I cam across one of a person with the same name as me and it seems odd that I took notice because another guy was named Evan but for some reason since I was ever called Evan I just feel some sort of connection to the name like it fits me perfectly and I guss whenever I see someone who is called Evan like me I kind of want to know more about them if they have some sort of connection to me in a way . It\'s completely stupid I mean it\'s just a name but it seems that when the name is chosen we have no real character and maybe the name it self forms you a bit like you see something in the name and that is what you want to live up to , the name itself will mean different things to differant people like there could be a psycho killer ted and a soup kitchen volontier ted but my name I don\'t know it just seems more apart of me more my identity and I want to know if it means the same to other people and if we see something together and I must say this guy was a bit like me not totaly but he had that alternitive essance , sure he was a bit of a hipster liberaltarian but still I got a vibe I\'m not sure what it was but it seemed real (if you can get what I mean) .  '),(3424,2311511,'02,July,2004','2004-07-02 00:00:00','   urlLink What Famous Leader Are You?    urlLink personality tests by similarminds.com'),(3425,2311511,'02,July,2004','2004-07-02 00:00:00','   well I have a whole bunch of news . First is that eli has turned the wall into his own blog it\'s actualy not that bad . Next is that I totaly eradicated all the crap from my links section and now it\'s just web comics and eli\'s blog , wait oh ya and the onion I couldn\'t bare to let it go but I think I just might soon but ya it\'s mostly web comics techie stuff so those not well versed in l33t shall find it hard to keep up $700p1|) |_||\\| l337z .  '),(3426,2311511,'02,July,2004','2004-07-02 00:00:00','   I don\'t fell like posting something really big so I\'ll just leave you with this one message ..............READ THE ARCHIVES FUCKNUT!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  '),(3427,2311511,'02,July,2004','2004-07-02 00:00:00','   thanks to all of you people for getting my blog counter over the 100 mark you are a truly insparational crowd exept for those of you who thought this was a blog dedicated to pink floyd , you S.O.B.s can go fuck your selves . I have noticed latley that whenever I go nuts on a rant I sound and look a little like Lewis Black , if you don\'t know who he is he\'s the modern angry comiedian a tribute to us raving lunitics out there infact he does this thing called back in black on the daily show with John stuart and is the equivilent to watch someone yell at the t.v. while watching news . Bye Bi and y3s t3y d3$3rv3|) t0 |)13 4n|) 1 h0p3 t3y  |3|_|rn 1n h3ll !  '),(3428,2311511,'02,July,2004','2004-07-02 00:00:00','   I just saw scarface today about 15 minutes ago actualy which is weird because I got it for my birthday and I haven\'t watched it until now but I loved it , in fact it\'s now tied with kill bill as my second favourite movie . It was really great and if you want to convice some one who is smart that anarchy works I definatly got it from that film but that might just be m3 . bi bi  '),(3429,2311511,'02,July,2004','2004-07-02 00:00:00','   well I stayed up till 1 o\'clock in the morning watching the end to fear and lothing in las vagas last night , not that I saw the beggining but I only saw the end but I must say that staying up that late on a school night is worth it for that movie . At school the rest of the school not my class because we have the play were planting stuff for our Brown goes \"green\" program (no pun intended i hope) and for a reward to them we had this big party at lunch but we got to come because well we\'re at the top of the food chain in the school so we do whatever the fuck we want ;) but anyway they had this terrible jazz band uhg you have no idea how much that sux0red , infact by the end of the lunch hour (more like 45 minutes but who cares) we were heckling them I was going to throw a button I found at them but I realized that if this is the kind of gig they are playing then I don\'t think they could get any lower .  '),(3430,2311511,'02,July,2004','2004-07-02 00:00:00','   well I just created my new blog that is a watered down version of this one so I can let my parents see it if I absolutely have to and I\'ll use it to try new stuff that I wouldn\'t dare do on here for fear that I would have to reconfigure all my colours put in all my new links (which I am still doing now thanks to hurricane eli . just joking eli) and get a new counter which I hate doing for some reason .  '),(3431,2311511,'02,July,2004','2004-07-02 00:00:00','   I\'m sorry I haven\'t been posting for a while but I\'ve been doing alot of stuff . Some of which is getting a full cartoon site up so I can have a web comic . I also have been paroozing the web comics to see what kind of stuff people are doing and I must say it will be some stiff competition but I hope it will work out . I\'ll be putting links to the new web comics tomorrow and my site whenever I finish it and I\'ll also be putting a free web hosting salution on here . I really hope that girl , Issabelle will put me on her links section soon .  '),(3432,2311511,'02,July,2004','2004-07-02 00:00:00','   Well it was announced that former president Rondald Regand had died . YEAHAW . I\'ve been so patient and finaly it was delivered to me ! I know it\'s mean but the dude was 93 not to mention a \"conservitive\". He was also the insparation (or I would imagine that he was) to Arnie or \"The Governator\" . Actors should stay out of pollitics and leave it to absolutly no one because no one should impose power over anyone . YAY anarchist dance time *anarchy in the uk plays*....get Emma Goldman and Lysander Spooner out on the dance floor *lookalikes of the famous anarchists come out and start dancing* *Evan holds up a sign that reads \"I use the little star things too much to tell action I should just learn how to tell the actions fluently in my writing well , bye bye\" then puts down sign and goes back to dance*  '),(3433,2311511,'02,July,2004','2004-07-02 00:00:00','   Well the play is done the certan is closed and all is done . Up on that stage there were preists of a sacred art . We are the one and only cast . no person will ever do that play again . And as the dust collects on our scripts . The basis of our affection and love that we formed to on another through this play . That peice of art that fuled the wrap party . That everybody wraped their scheduels around and later wraped their minds around . The only scrips . Full scripts will belong to us . Though this play was made up of many different parts . We have the full collection of scenes . The re-writen and re-writen scenes . Some so re-done that is was impossible to desipher them from the original . I don\'t think I\'ll ever feel closer to any other people that much again . Though we may say some are self obsorbed or weird or gay . We all became freinds tonight . We all knew we were part of something bigger . We let everything go just for that one second where we knew what the other was thinking . Though we might have been divided in the past now we were all together as one . Now is when the freindships are formed when we creap out of our little \"clicks\" and \"Posseys\" and when we finaly are truely together not only as freinds . Or even lovers . Just people . A true bonding of humanity . The only real time where I felt I was apart of something great . Good bye and as the certan closes remember me for what I did . for how I made you laugh cry and shudder . for how I made you question everything you knew . How I learned and loved and wished and dreamed . Let no man know what I knew for it is for me alone and the others too .  '),(3434,2311511,'02,July,2004','2004-07-02 00:00:00','   urlLink   What kind of superhero will you be?        NameÂ         AgeÂ         Favorite ColorÂ     white  black  red  orange  yellow  green  blue  purple  silver  gold       HowÂ     nuclear accident  genetic mutation  alien  super techno gadgets  medical mishap        Superhero Name Â     ?        Costume (not gender specific) Â          Power Â     cloning yourself        Number of people you save a day Â     53       popularity rating -  99%                                                                                   This   fun quiz   by  urlLink  aingeal311   - Taken 27338 Times.          New! Get  urlLink Free Horoscopes  from Kwiz.Biz'),(3435,2311511,'02,July,2004','2004-07-02 00:00:00','   I think I might just get into web design . It seems pretty because you get to see the source of it all and it\'s completely amazing  . just to see a page that you have designed out of the bits of code is awsome . Now I\'m on isi\'s blog . I hope alot more people will go to my blog because of it . If not then it still doesn\'t really matter because I\'m apart of a blogging community . Even though I was before this a sort of a fluke and is a spontanious link of 2 sites and even if I don\'t gain anything from it it\'s still merging together for a common goal of just keeping your self known . to leave a legacy to humanity even if it\'s \"the arrival of the mini-skirt\" . thats a line from the play my class is doing . Well sorta it\'s a version of the line . On the subject of the play . We did our second preformace today and it\'s so funny how after the first scene all the nervesness just drops away from you . buh bye bitches  '),(3436,2311511,'02,July,2004','2004-07-02 00:00:00','   well I haven\'t posted in while but this time I really don\'t care what you think I\'m not going to appologize for not doing something that I don\'t have to do anyway . Well we did our first preformance of the play today . it was really fun . There were all these little kids like grade 1 and 2 and after when I was leaving the school . I saw all these little kids pointing at me and saying \"oh cool thats the guy from the play\" . I guess I gave a note worthy preformance .  '),(3437,2311511,'02,July,2004','2004-07-02 00:00:00','   Hello everybody out there in blogland . As you know I am a bit of a music critic and while my colleague Eli (pimppanda) tends to focus on the greatness of the greats (Neil Young , Alice Cooper , The Beatles ect.) . I am a much much more petty man . I like to focus on the utter unlistenablility of modern popular music . Yes there are a few that are very good like Billy talent but there are also a rather disturbing amount of untalented manufactured so called \"artists\" . So that\'s why I did this post because I was watching Much Music at 4 in the morning and I need to rant about all the shit that I saw floating around . First lets start with the guys . My nemesis and yours good charlotte : see disguised boy bands .......oh god this is just too easy . Actualy I\'ll explain for those who haven\'t been reading my blog since forever . good charlotte or as I like to refer to them the unholy bringers of the end of existence . These people\'s sole purpos for existing is to drive me to the brink of insanity or suicide . Every time I hear on of their songs I block my ears and twitch and rock back and forth in the featle position . This band even sounds like the backstreet boys it\'s just instead of a synthisyser in the backround there\'s guitars , basses and drums  (which they don\'t play) . Ever wonder why they have a huge female following ? Ever wonder why the backstreet boys or n\'sync has a huge female following ? Well since I\'d probably spend the whole time yelling about how they suck lets move on shall we . Hubastank : (not sure if I spelt it correctly) They are just like G.C. but there style is  slightly  more punk because they don\'t over punk it . Yknow once you start refering to your clothes as a costume then it\'s not really style is it ? I feel basicly the same way to them as goodcharlotte but I think they kinda mean well . Either that or they are actualy more evil and plan on taking the minds of the world . All I have to say is  Rush is to Eli as Goodcharlotte and hubastank are to Evan . Now thets get to the chix0rz . First and Worst Hillary Duff : She seems to be a paradox in some ways . she\'s trying so hard to grow up and yet her fan base consists of 8 year old girls who look up to her . Hillary Duff , floresant lights and cow hormones . You figure the last scentence out for your self . Of course growing up fast is good it forces kids to be more expeirienced and more educated which for some reason other people don\'t like that . \"Ignorace is strength\" . You didn\'t here me say anything . Well atleast ol\'Hillary is whoring it up . In her videos she\'s either parading around in a white blouse in the rain , having sex with the microphone because in each scene she looks like she\'s at the moment of climax  or even doing ultrasoftcore lesbian porn with her sister who looks like Uma Thurman........well if ever wanted to see Uma (basicly) feeling Hillary Duff up then your in luck . In an interview she said she never rebelled as a teen . Daddy\'s little goodthinker . Her not rebelling I\'m kind of thankful for because if she rebelled she would be like Christina Agulara now and I think I would stop going to school just to watch those videos . And if she rebelled then she probablly would make good music I would have one less thing to rant about here on the blog . Now lets get to Avril Lavein : I want to say I hate her for being a poser bitch and she\'s anti-drugs . I also feel for her I think she actualy wanted to make good music but she is just a small town girl from canada and she got really exited about all the cash . warning : capitalisim may  cause disturbing after affects use with caution and if possible with a Liberatarian government  . Now the Scary think is that the closest thing we have to a popular rocker chick is Fefe Dobsone and you have no idea how much that scares me . Well atleast she writes songs about hating people and yells and looks like a coke head after a prefromance . Remember though Jesus died for your sins but I watched crappy music videos and interveiws for the simple fact to know who to hate and why so you don\'t have to and thats close enought to be a messia , right  :)  '),(3438,2311511,'02,July,2004','2004-07-02 00:00:00','   Since I now have a shout box this is the law of the shout box : ........uh........nothing.......there are absolutely no rules for the shout box exept no racisim or you\'ll get banned . Either banned or I\'ll hire very large convicts to rape you in the eye sockets .  '),(3439,2311511,'02,July,2004','2004-07-02 00:00:00','   I was watching a tv show at 7:00 last night and there was a news bulliten that said and I quote : \"Ray Charelles dies tonight at 11\" and I\'m thinking damn thats acurate . Just think about him watching that he\'d be like : \"fuck! ok well might as well make it creative ,hunny get the vibrator and a rope o\'ll charllie\'s going out in style\" ok I\'m pretty sure he didn\'t die haning from a rafter with a vibrator up his ass but still .  '),(3440,2311511,'02,July,2004','2004-07-02 00:00:00','   Don\'t worry this isn\'t the final draft of the template . I wanted to get the feel for the \"tic-tac\" and I didn\'t want it to be all normal so I figured I\'d put up my andy worhol skull pic for now . It should be completely changed by next friday .  '),(3441,2311511,'02,July,2004','2004-07-02 00:00:00','   Okay for the next while I\'ll be talking about the most asanine things . This is because I\'m going to montreal next week and I want to pack as much content on my blog as humanly possible with (what seems to be) carpel tunnel syndrome . I haven\'t had it proven by a doctor but I know it\'s some computer related muscle strain injury . It actualy happend while I was playing space invaders . Yes I\'m a sucker for the old video games . Anyway I was pownding on fire as hard as I could to blast those little freak alien bastards into oblivion and then I was killed . So I let off of the keyboard and as soon as I moved my arm it felt like I was pelted with very large rocks . Long story short I\'m probably going to lay off the site and the new template for a bit until this arm heals . Upside : I won\'t have to do as much work at school because it\'s on my right hand . For those of you who are gathering their  sick  opinions of how I got my bout of ( I don\'t want to say cts because I just realized that you can\'t get that in your uper arm) computer injury . That think that my injury was because of a certain type of site on computers you can just go to hell . Now back to Montreal . Since I Won\'t be posting all week next week I\'ll give you a bonus . I\'ll take all my daily accounts down and then put them up here so you can see everything you need to know that I knew then . Second I will also give you (almost) all the pictures that a take while on the trip so you can see what we all look like  . We may look frightening at first but you will learn to love us......just like monkeys! :)  '),(3442,2311511,'02,July,2004','2004-07-02 00:00:00','   Okay in my design for the \"new\" blog template I will incorperate sketches of mine . This will be feasible I\'m sure but possibly dificult because of how I plan to do so . You see what I want to do is take an existing template (most likely Tic-Tac) and take the original image out and replace it wil a collection of my sketches . Unfortunately these sketches will probably have nothing to do with the blog (little Japanse girls in school uniforms , crazy people , fan art that I made ect.) so bare with me if you don\'t like the sketches . Actualy don\'t bare with me if you don\'t like the sketches you can fuck off . Just joking . Well I better go I was playing a big ass game of space invaders while writing this and my hand is numb .  '),(3443,2311511,'02,July,2004','2004-07-02 00:00:00','   This is a follow up test of the emergency blogcast system  '),(3444,2311511,'02,July,2004','2004-07-02 00:00:00','   this post is a test of the emergency blogcast system if you are still reading this you are a moron that is all  '),(3445,2311511,'02,July,2004','2004-07-02 00:00:00','   For some reason everyone I know with a blog and a counter has very close amounts of hits as mine and alot of them have gotten their counters after me . So you can see my perdiciment my blog sucks and no one is coming . Thats why I\'m going to try another renovation over the next week or so . I\'ll either find that my template is fine or you\'ll have a whole new blog on your hands . Though it will be the same blog it will look different .  '),(3446,2311511,'02,July,2004','2004-07-02 00:00:00','   Now we\'re near to Ottawa where we\'ll be stoping for lunch and stuff . I\'ve basicly used up all my film on flash wars . What is flash wars you ask? it\'s when you take a surprise picture of a person with the flash on and scare them it also makes on hell of a funny picture :) I also tried to get some nice landscape shots for you guys . though a good amount of them will still be of surprised class mates  and irrecognisable blurs . Hopefully I\'ll atleast be able to show you the landmarks I will visit .   June 16 2004  Okay so I\'ve went through 2 disposable cameras but one was the one I won in a scavenger hunt to find out facts about Montreal . My group was surprisingly good . It was me , Eli , Alison , Stefani , Essery and Tasha . That just proves that intergration of sexes provokes the best in people.....well sometimes . We went to a big church on top of the hill here (TOUR GUIDE : It\'s not a hill it\'s a mountain!!!!!). Our Tour guide is obsessed with deffending \"Mount Royal\" as a mountain  NOT  a hill . Of course she\'s also obsessed with flirting with the Male teachers . It\'s funny (and a little disturbing) to watch adults (especialy teachers) flirt . The funny thing is that the all the male teachers are married . Looks like we have what the french call a pute ( think I might have spelt it wrong I think there might be an s in there ) .  '),(3447,2311511,'02,July,2004','2004-07-02 00:00:00','   you know what I just love ? Spam e-mails about how god damn big my cock could be . Do you have any fucking idea how much shit I get in my inbox about enlarging my prick ! It just makes me want to drive over to their flat and bash their heads in with a spanner . Those wankers should be shot and heaved into a burning heap of rubbish . Then put in a lock-up full of randy inmates that will shag them in their eye holes . Do you like my use of British slang ? If there are any mistakes then post on the shout box or on the comments . Oh ya if the shout box backround is black and the text is black then just refresh the shout box and if it\'s still giving you problems just post on the comments about it :)  '),(3448,2311511,'02,July,2004','2004-07-02 00:00:00','   sexual harassment : the momentary reminder by a co-worker or coleage that sex actualy exists .  un-conformist : some one who thinks they aren\'t conforming because they are into preforming arts vote demicrat and have peircings to satisfy some superfitial trend .   full life : living to be old enough to develope althiemerzer and die a degrating death in a pool of your own feacies . In a nursing home that is so costly that it often results in the loss of alot of funds .   (these are not direct definitions . The exact definitions can be found  urlLink  here   )  '),(3449,2311511,'02,July,2004','2004-07-02 00:00:00','   I should be putting this post to good use and putting in the notes that I made while in montreal but I wanted to talk about michael moore . I didn\'t see bowling for columbine but I always dismissed it as one of those exposeyes about how guns are bad and it was just a message to be trendy . Though now that I saw the trailor for Fahrenheit 9/11 I think that he\'s not sutch a whiney little bitch and that he actualy meant bowling for columbine . You see what I had against a movie about crappy gun control is that they would just be bitching about the second amendment for the whole time . The thing the U.S. won\'t tell you is the second amendment isn\'t EVEN ABOUT FUCKING GUNS GOD DAMN IT!!!!!!!!!!!!!! it\'s about the right for a people\'s army to over through an opressive force in the government . So if the government sais every one can have a gun to shoot squirells they don\'t have to deal with an uprising . Most liberals don\'t understand that and say get rid of the second amendment but in reallity the U.S. should educate the people on it\'s true meaning . Of course what government would sacrefice their own comfort for the well being of the pubilic ? But the great thing about  F9/11  is that it basicly conferms all of my opinions towards the U.S. government . So if this trend of me being right continues then watch out for Bill Gates to enslave man kind then rape goats and eat babies .  '),(3450,2311511,'02,July,2004','2004-07-02 00:00:00','   well I forgot to apologize for getting a bunch of the artists\' names wrong in my pop culture round up . I are sowwy . Does anybody want a hug?  '),(3451,2311511,'02,July,2004','2004-07-02 00:00:00','   I should have the rest of the posts digitised by tommorrow I\'m taking a day off to catch up and stuff . Though I noticed that no one posted on my shout box exept for eli . So post or I\'ll shit on your dog , cat and/or hamster .  '),(3452,2311511,'02,July,2004','2004-07-02 00:00:00','   Alright I\'m on the bus . I thought is would suck more than it does . Eli is beside me and is yelling to people . Stefani is showing me her plush duck that she brought for the trip . I\'m not sure why . I think is supposed to be a funny little quark but it\'s just fucked up and annoying .                Now we\'ve made out first stop (at tim hortons) . Because our teacher (Mr.Fraser) won\'t let us have coffee I had to get a coke and a double chocolate doughnut for the caffien . Keep in mind it\'s about 9 in the morning .  '),(3453,2311511,'02,July,2004','2004-07-02 00:00:00','   Well going to Montreal in about 9 hours and 15 minutes . I won\'t have access to internet but hopefully the thought that I\'ll be coming back with a post full of little notes I made while there and pictures !! yep pictures ! you get to see everyone in my class . Unfortunately for there protection I won\'t tell you who everybody is in the pics but it\'s still good . well I hope my crazy assness will inspire and illuminate while I\'m gone because it sure as hell didn\'t happen while I was here . Good bye , farewell , Amen  '),(3454,2311511,'02,July,2004','2004-07-02 00:00:00','   well I found out that I won\'t be leaving until tomorrow because of an air traffic controler strike in Costa Rica . You can figure out how that fits in with me staying at my grandma\'s on your own . But I did get the pics from montreal for you guys here they are .  this one was the first I took I think Eli was a little surprised to get a picture taken of him at 6 in the morning . It was taken just before we left Toronto .   this one was the first shot fired in the flash wars but I put it here because I wanted to show how crazy stef is .   now lets jump to our first night in Montreal . This was taken at the IMAX theater we went to . Well atleast I think it was . I can\'t think of any other place that looks like that .   there\'s eman or Emmanuel if you want to call him that . He looks like he\'s being confused by some thing maybe he was thinking about anything in the universe . I\'m just joking around . It\'s so fun to make fun of Emmanuel for being dumb and it\'s so very easy . Now here\'s one of me   I\'m not really sure what I was doing when that was taken . I don\'t know who it was that took it either .   doesn\'t this remind you of some crazy movie like Fear And Loathing in Las Vegasâ„¢ or some thing like that (I\'m sorry about the trade mark thing I just always wanted to do that) . well thats all the pictures I\'ll put in because it\'s going to be hell to publish this bitch .  '),(3455,2311511,'02,July,2004','2004-07-02 00:00:00','   well I just found out that I\'ll be going to my grandma\'s house for about 5 days . I won\'t be able to blog because my granny aint got a computer . She\'s one of those technophobes . I ask her \"why don\'t you get a computer it would make alot of stuff easy for you ?\" and she sais \"I\'m to old to learn anything new...and plus I always have my type writer\" . I kinda feel bad for her because her type writer only has dial-up . Anyway since I won\'t be able to blog or go on MSN or keep up on web comics I\'ll hang myself with my pants . Just joking don\'t worry this will give me a sort of break I\'ve been needing . I\'ll be doing comics and and little graphics for my new site and this one . I just won\'t be doing any html or importing images or touching up stuff on photoshop  (oh ya I\'ll be geting adobe illustrator soon too) . I won\'t keep a journal of staying at my grandma\'s house mostly because it will probablyl just be me watching tv and drawing all day so it doesn\'t matter . I\'ve been geting into showcase it\'s a movie channel here in canada (pretty much the only one on basic cable ) . It has alot of cool shows and movies . a couple nights ago I found this great movie on it about german lesbians . It was a really good film . They also have a great show that you probably never heard of called made in Canada . It\'s about a tv production company...long story short it\'s funny if you ever have the chance to watch it then do so or else you will be tormented with the regret of what would have been and it will tear you up inside until your last breath never forgeting you missed the chance to watch made in Canada :) I\'ll soon be putting all the stops out to promote my new template and my blog in general so put the word out NOW!!!! tell your freinds , family , enemies , teachers , camp counselers , people you just met like 5 minutes ago but find compelled to have a meaningful freindship with , pets or anyone else you can think of . Link me to your blog or your site . Better yet tell people about me on your sites and blogs . take out ads in your local news paper to tell people about me . I don\'t care just tell people . If your a kid who\'s going to camp tell your fellow campers , counselers or who ever . If you\'re an adult and you\'re going to work tell your fellow workers , employers or employees . NOW!!!! Run go tell them NOW!!!!! there\'s not a minute to spare . I don\'t care even if you think my blog is good  JUST DO IT!!!!!!  '),(3456,2311511,'02,July,2004','2004-07-02 00:00:00','   well since it\'s summer holiday I was watching the connan O\'brian show last night and I saw the U.S. air guitar champion . It was funny when she preformed because she put so much heart in it . She even did the pedals and at the end she lifted up her skirt so you could see her panties . That was the funniest part . Well I found a good free web hosting option . Well maybe not good but half decent . So the comic site should be up by the end of the summer .  '),(3457,2311511,'02,July,2004','2004-07-02 00:00:00','   well today was our last day at brown public school . Of course in about an hour it will be yesterday was our las day at Brown . I\'m glad that I\'m going to be going to Deer Park because \"Brown\" is the lamest name for a school . Today was just overall fucked up . I can\'t explain it any other way than totaly fucked up . I would go into the specifics but I have sworn to secrecy.....oh fuck it . First our last party was lame as hell thats why I ran through the school in only my underwear . Yerp I streaked (well I had my boxers on) through the school...ok fine it was only on the third floor but some people saw me . Mostly because I was yelling \"it\'s streaking time!!!\" . I didn\'t get in much trouble with the Fraseman . Infact he said he wouldn\'t tell the office but if one of the other teachers saw and reported me he wouldn\'t protect me . Fortunately I got off scott free (I\'m as amazed as you are) . So then (this actualy happend before I streaked but I wanted to mention that first because it\'s just cool ) Alison kisses Eli . Okay for those of you not in my class this is like the most fucked up thing that could ever happen...well not really but just damn unlikely . It was funny that he told me not to tell anyone and yet I\'m spreading it over the web . Actualy he told everybody during the day so I just suppose that it\'s okay to publish his personal affairs on the internet with no consent from him :) please don\'t sue . And then Alison anoucess that she likes Emmanuel . (sorry for being a gosip whore but I gotta put something on here with some action) The funny thing is that she already kissed him at a party like a week ago . Well actualy he kissed her but same damn thing . But you see Eli really likes Alison so when he heard they kissed at the party he was pissed . And later in a convo on msn Eman mentioned to Alison (I know because I was there) what Eli said about him kissing her that he wanted to be the first to kiss her . So one wonders was that kiss that she gave him but a simpathy kiss . Still though Eli got a little action from the hottest girl in the class so does it really matter ? This whole thing has made Eli a little paranoid . Not oddly that it was a sympathy kiss but because after school when Alison reveiled her love for Eman he was confused to if he could return the love . So in deciding not to have a relationship her because of his confusion and the fact that Eli started crying because he wanted Alison all for himself (or so it seems) . So because Eli assumes that Eman is so ruthless he is ploting some sort of evil plot conspiring against Eli to knock him out of the picture . This of course I fathom is just Eli being totaly insane and utterly and imencely paranoid . Mostly not because this is totaly self indulgent on Eli\'s part but because Eman isn\'t smart enough to do that trust me . He\'s as sharp as an egg in other words . Eli also claims that all of that crying and whyning wasn\'t because he wanted Alison all for himself but because he wanted to protect her from herself . Which I don\'t beleive for a god damn second but it\'s true in the past he was very selfless towards her (geting her dates with people she liked even though he liked her at that time ) . Despite all this evidence I feel in my gut it\'s a complete load of bull shit but whatever . So lets recap . I run half nekkid in my school , Eli gets a little action from Alison , Alison sais she likes Eman , Eli becomes more like me in being paranoid and then starting his own religon (it\'s a really cool religon I\'ll post a manifesto  of it when Eli works all the kinks out of it) . So remember kids Big Brother is watching so make sure to have alot of fun in the shower ;)  '),(3458,2311511,'02,July,2004','2004-07-02 00:00:00','   I forgot to tell you that I got my year book today and I might be able to put in a bunch of pic of my class . Though alot will be mixed with the french imersion class so you\'ll just have to guess who\'s who  '),(3459,2311511,'02,July,2004','2004-07-02 00:00:00','   I changed the template (once again) . I think this one works alot better but \'ll probably think something else will in a week . Of course in a week I will be sitting free from the torments of school . Yep it\'s summer holidays in exactly 2 days . Infact we had our graduation today . It was really fun because we had like a rave (well sorta) party after . Pema and I raided the supplies and got meter long pixie stix . Pema then went insane from over caffeination and lots of sugar . We haven\'t found him yet but I\'m assuming he\'ll be in his underwear passed out on the side of a highway . Thats my prediction anyway . That last day of school party will be one bitch . We went insane when it was just a little afterparty this is our last party at brown . Were going to blow the roof off that motherfucking school . Either that or it\'s going to be the worst party ever . I wonder if the Frenchies will french it up at the party ;-) (inside joke)  So in concusion I\'ll be here for the next while so I\'ll be posting almost daily making the template as beautiful as possible and......uh......ya .  '),(3460,2311511,'02,July,2004','2004-07-02 00:00:00','   I\'m not sure if I like the new template I made . Fear and Loathing in Las Vegas is one of my favourite movies but something just doesn\'t seem right . Comment and tell me what you like or hate about it . I think it might just be having tic-tac as my base template . just comment or post on the shout box . If you don\'t comment or post I\'ll bitch and moan until you do . And no one want\'s that do they ? because I haven\'t even ventured to bitching and moaning factor 5 on here yet . You people have not yet even seen the extent of my whining!!!!!!! INFIDELS!!!! YOU SHALL FIND MY REVENGE SWIFT AT THE HAND OF ME BITCHING ABOUT YOU NOT COMMENTING!!!!!!!! I shall smite all who do not post or comment . Your deaths will be swift and painful !  or yknow whatever......  '),(3461,2311511,'02,July,2004','2004-07-02 00:00:00','   well I would like to retract my list that I made yesterday because  I have been making numerous discoveries in the field of suckage . I would also like to retract it on the contition that there are alot of bands in the world and I couldn\'t possibly form an opinion on all of them or even most of them , hell I probably couldn\'t make an opinion on fraction of them . Now I would like to anounnce before you hear it from any one else I have that dee-lite song \"groove is in the heart\" stuck in my head and the only way I can get it out is to play it until I am completely averted to it . Now I would like to put new TV shows that have come to my attention watching canadian cable during the late morning and late night . The Chris Isaak show originaly aired on the U.S. channel showtime is now on Much More Music (the canadian VH1) . This show not only has funny situations consisting of orgies and public sex but also has Chris Isaak\'s music and his brand of bluesy rock is enough reason to watch . This hour has 22 minutes originaly aired on the CBC it is now on showcase . Though I started watching this show a long time ago I stoped for an even longer time and now with the new cast I would say that it\'s quite good especialy now that Gavin Croford is on it . Made in canada originaly aired on the CBC now on showcase . This show like this hour has 22 minutes started on the CBC but isn\'t like This hour has 22 minutes in any other way . It\'s witty and smart . Infact if it was a woman I would marry it then divorce it because marrage is man made prison that can only lead to unhappyness and regret :) The show makes an exelent point of what goes into making shows in a Dilbertâ„¢ (I love that little trademart thing) like environment .  '),(3462,2311511,'02,July,2004','2004-07-02 00:00:00','   I\'m going to list all the bands and artists I hate music . good charlotte , simple plan , hubastank , hillary duff , fefe dobson , evanesance , lilix , justin timberlake , britany spears ect. (ect. meaning the rest of the bubble gum pop artists) , sky sweetenam , Avril Lavein , celine dion , ABBA , bee gees , sammy hagar (DLR all the way), bocelli . This is all I could think of . I only put this many because most of the others that I hate are obvious in hating . I would also like to put together a list of bands and artists I like . The Clash , The Doors , The Beatles , Led Zepplin , Prodigy , David Bowie , Pink Floyd , Chris Isacc , The Sexpistols , Stiff Little Fingers , Beastie Boys , Jerry Lee Lewis , Billy Talent , Sweat Shop Union , Janis Joplin , Jonny Cash . Thats about it .  '),(3463,2311511,'02,July,2004','2004-07-02 00:00:00','   Okay to recap over the week . I was gone to my grandma\'s and she doesn\'t have a computer so here\'s all the news that you didn\'t know because you were to busy pleasuring your self to naked pictures of loretta swit . Canada had an election the liberals won......fuck the liberals won . alright this isn\'t really to bad I mean we could have had steven Harper of the conservitives who wanted to criminalize gay marrage and he wanted to ban abortions too . I would have much rather had Jack Layton of the NDP (new democratic party) or that guy from the green party . I\'ll explain what kind of elections happen in canada for you southern fried chickens . each part of a city or town has a representitive or an MP (member of parlement) these MPs are elected from the people in that area . The party with the most seats in the house of commons is the leader of the country and the leader of the party is the prime minister . So we can hold an election to elect a prime minister but this also gives other parties more power by how many seats they get so we can have more diverse government . Though only one party is ruling some other veiws can have more say in the governing without it being the ruling party . I\'ll also tell you about a minority government and a majority government . A minority government is when a party gets more seats than all the other parties but not more than half of the seats . So this means that all the parties have to work closer together and Rick Mercer said it best : \"These men hate each other it\'s not an act and a minority government would just bug the hell out of them and thats why I\'m in favour of it . Infact they should just divy up 24 susux drive and put in some bunk beds . That way Paul Martin , Steven Harper , Jack Layton and Gilles Ducceppe can stay up all night gigling and planing the fate of the world .\" Okay now more important things . The U.S. is issuing a Draft . Okay not exactly a Draft . They are taking people who were in the military and were discharged then puting them against their will into the army again . Kinda like a draft but they don\'t have to train them so they can get killed quicker . ATTENTION ALL HIPPIES WITH PREVIOUS MILITARY EXPERIENCE FLEE THE COUNTRY NOW!!!!!!!!!!!!!!!! Well thats all for world news on the EBS (Evan Blogcasting System )  '),(3464,2311511,'02,July,2004','2004-07-02 00:00:00','   This is the EBS world service broadcasting from Jamaica The GMT is 3:18 but here it\'s 10:18 . I just drove across the island and back all the way listening to the BBC world service I guess the soothing british voices affected me .I arrived in Jamaica no problems flying alone exept for the fact that as soon as I come to this damn country (every fucking time) it turns from nice and sunny to torrential down pour . Though now it\'s nice exept that I feel like I just to a bath in bacon fat . See ya whenever bye for now Evan signing off .  '),(3465,2311511,'02,July,2004','2004-07-02 00:00:00','   I\'m going to Jamaica tomorrow but don\'t worry I can blog from there . I just won\'t be doing it as often . Okay so I\'m a little behind on the new template . I have the general theme down and the title bar done but I don\'t have any graphics to replace those ugly little tic-tac tab things for the archive lists and the by line thing . I\'ll update you as the progress goes on (not like you care) . See ya and the next time I post I\'ll be in jamaica probablly .  '),(3466,2311511,'02,July,2004','2004-07-02 00:00:00','   I\'m going to Jamaica tomorrow but don\'t worry I can blog from there . I just won\'t be doing it as often . Okay so I\'m a little behind on the new template . I have the general theme down and the title bar done but I don\'t have any graphics to replace those ugly little tic-tac tab things for the archive lists and the by line thing . I\'ll update you as the progress goes on (not like you care) . See ya and the next time I post I\'ll be in jamaica probablly .  '),(3467,2311511,'02,July,2004','2004-07-02 00:00:00','   I downloaded that blues brother\'s song soul man and I when it downloaded I played it at the highest volume my speakers could go and I did the dan achroid dance . Y\'know the one where he\'s jumping around and kinda looks like he\'s playing hop-scotch . They actualy did some good songs . I guess some good came of watching that SNL music special . Did you know that once chevy chase had internal bleeding because of when he did his falls . I would like to finish with a salute to Dan Achroid : Jane you ignorant slut .  '),(3468,2311511,'02,July,2004','2004-07-02 00:00:00','   I would like to thank all the people who come to my site for helping me get 410 hits . I know that 400 would have probably been more traditional but by the time I checked the counter it was 409 so I thought I would make 410 because even though at the moment I\'m writing this it\'s 409 hits  who ever see\'s this post means that they went to my blog which would mean that they would push it to 410 . So......HAPPY 410 HITS EVERYBODY . This whole summer vacation thing is screwing me up . I\'m going to bed at 4:00 in the morning . Not getting alot of work done . It\'s just totaly fucking me up . I thought I would have the comic site done by the end of my first month of summer vacation . It\'s ended and I haven\'t even got my home page done yet . Oh fuck it I\'m just going to have a shower and try not to think about my wasted time .  '),(3469,2311511,'02,July,2004','2004-07-02 00:00:00','   well I saw dodge ball a while ago and I thought I would reveiw it . Though I don\'t really do reveiws I am just bored as hell so why not . Dodge ball : story line : 7.5 / 11 humor : 10.5 / 11 Use of pirates : 11/11 over all : 9/11 . The weird thing is that I\'m going to see Fahrenheit 9/11 soon . There are numerous unexplained phenomina on THE DARKER SIDE OF THE MOON !  '),(3470,2311511,'02,July,2004','2004-07-02 00:00:00','   I just started working with photoshop really hard and I must say I like it . I got a tutorial for it and now ihavefun (it\'s an inside joke) . I always thought that the catch to my layout would be that all my graphics are totaly hand drawn but now I\'m starting to have second thoughts . In my new layout for my cartoon site I think I might just use photoshop graphics but I can\'t help but feeling that I\'m giving in . well all I have to really do is just get good enough with photoshop that I can have exactly what I want I guess .  '),(3471,2311511,'02,July,2004','2004-07-02 00:00:00','   I was without internet for the past 5 days and at this moment I feel asthough I\'m having an Orwellian nightmare . You see because I wanted this to be a blog that my parents didn\'t know about (see what to do when your mom finds out about your blog) it\'s hard for me to write because my father is choping onions not 2 meters away from this little laptop and with the terrible little mouse on it and the wonky s key doesn\'t make it any easier . Though this little ibook . Through out the week it has provided me with the gift of DVD . I saw ghost world , El Meriachi , Resivoir Dogs , Wasabi , cold mountain and along came polly . I got a few souveniers for my freinds . I won\'t say what they are because I know they read this . I\'ll give you more details about the fact that a part of jamaica is going through a revolution . There are riots assasinations all typical revolution typed stuff . exept they are all just in one city . ofcourse it is the capital but whatever . I just glad I\'m in another part of the island . You won\'t hear any of this stuff on the BBC world service .  '),(3472,2311511,'02,July,2004','2004-07-02 00:00:00','   Today I played some tennis and ping pong then went for a swim . I hate this ibook I\'m blogging with . The keyboard is just generaly messed up and around the keyboard gets really hot . I saw Euro-trip and Big fish Both not to bad . No matter what anybody sais about the sort of animal house humour in Euro-trip it\'s really funny . Espescialy the Dildo thingy . Well see ya again real soon and remember I\'m funky Tut and I hope I\'m your favourite honkey .  '),(3473,2311511,'02,July,2004','2004-07-02 00:00:00','   I just decided to say fuck it to the new template . I\'m not going to not do it I\'m just going to change the orininal theme and the fact that I\'m not going to tell anyone . I started thinking gothy sorta black and white comic thing but now I\'m thinking andy warhol technicolor LSD trip thing . basicly the exact oposite but it will be easier . I figure it would be a bit more fun . So now you will see step by step what happends to my blog . I also have to say that any graphics that have been aproved for the links will be gone . I\'ll replace them with andy warhol or and tie-dye themed stuff . Oh ya here in Jamaica there\'s a religeon called rasta where you can\'t cut your hair and smoke alot of weed . I think just might not be an athiest any more . Well I\'m not a 100% on everything in rasta . I think that it\'s a branch of a christianity sorta . It\'s in the Jew-christian kind of belief system exept they warship another person like jesus but I\'m not sure why . It might be because the beleive he\'s the massia or maybe because he has a Buddah thing going down . Oh ya and since I\'m down here I might get a new cell phone because they have the weirdest ones down here . I saw one with a round screen , another that was as narrow as a pack of gum then some had these fake diamonds along the sides and when the phone rings the diamonds light up . They are all the absolute coolest . see ya EBS signing off .  '),(3474,2311511,'02,July,2004','2004-07-02 00:00:00','   Well I was without internet ever since I got back from Jamaica  so I couldn\'t post . A real Bitch aint it I go to Jamiaca where there isn\'t that great internet service and I look forward to coming back to a steady connection but it isn\'t here , Satan why do you mock me with a tangled web of irony ! Well anyway when I was on the plane here I met an amazingly hot girl but alas she was older than me so it wouldn\'t work ( girls don\'t like younger guys until they\'re atleast 30 ) but anyway I did get to meet her and bask in her babe-o-liciousness for a while . Ofcourse there was an even hotter girl that was on the shuttle bus . I didn\'t meet her but she was very hot and I think would have been more my type (by her look anyway) . Well anyway Bell sympatico internet sucks like a date rape victim . I hope you have a pleasant today and an overly complicated tomorrow .  '),(3475,2311511,'02,July,2004','2004-07-02 00:00:00','   My stupid ass mind has just desided to change preferences once again so now I don\'t like the idea of a psychadelic template . I should just get a lobeotemy so I won\'t have any other creative preferance other than grey . I guess over the years all my preferences will change and my blog will have massive face lifts . I think what I could do is make lots of templates for me and other people and then I can just click to what I want that day or week or month . From now on I am going to try and keep dedicated to the writing  not  just the template . I have lost tons of sleep and time over templates but damn near nothing to my blog it self . It was obvious all along why people weren\'t reading my blog . I am destined though to make sure I have a following on this blog . Maybe not today maybe not tomorrow but one day I will be on the Blogger\'s recomended list or something like that .Right now I just need a fucking vacation , Oh ya and by the way I might as well finish off what I said I was going to tell you about the assasinations and stuff . The government here has be showed to be tied with the organised crime in the ghetos . What the government did was supply these people with gun and money and they policed the gehtos for them so the police didn\'t have to go . the gangs the gangs would also make sure that all the people in that reagon would vote for a MP that gave the gang the supplies . A wonderful little system they have down here ain\'t it . I actualy do like Jamaica it\'s a place where all my whack job theories come true .  '),(3476,2311511,'02,July,2004','2004-07-02 00:00:00','   well I haven\'t posted in a while because of my film course . It really eats up the time . If I\'m not in it from 9 - 4 I\'m at Eli\'s discussing the day and when I\'m at home I\'m thinking of ideas for films . well today though I basicly got told by my teacher that the movie I made could never get into a festival which I was wicked pleased about because I always wanted to be the renigade director/writer that over does everything . The class seemed to really like it though even the one sort of Ã¼berserious girl that I almost never see laugh laughed and she\'s the most serious Director / Writer that is in the class .  I\'m really looking forward to doom 3 but I\'ll have to soop my little bitch ass computer up for it . If I get my movie idea voted on to be one of the 4 movies that we make next week I\'m going to try and incorperate a dancing backwards talking midget like on twin peaks . see ya  '),(3477,2311511,'02,July,2004','2004-07-02 00:00:00','   Ok I\'m at my film course . I haven\'t posted in a while but whatever . It\'s pretty fun here my movie got picked as one of the few that got made . oh shit! my teacher just walked in I hope  she doesn\'t see me . ahh ok I  gotta go .  '),(3478,1808929,'31,August,2003','2003-08-31 00:00:00','   Hey y\'all. I am so sick of studying. My test is on tuesday and I don\'t know if I can take much more cellular metabolism. You\'re probably thinking, \" man, if Doug is freaking out over the third test of med school, he\'s in for a long hard haul.\" However, most people say that this test is probably the hardest of year 1, so that makes me feel a little bit better. God dammit, I\'m tired.  [\"Slow Like Honey,\" Fiona Apple, 1996]  This last friday night I went to a Rangers game with my folks. I don\'t think I have ever been to a Rangers game where they\'ve actually won, and last friday night was no exception. They were doing well up until the 8th inning when everything just went to shit for the Rangers. I won\'t get into details, but it was so bad that at one point after a really screwed up play the home audience started booing the Rangers! Yeah, not too good. Anyway, it was still fun \'cause my mom got to walk out on the field and be honored with several other school teachers in the region who had been nominated for teacher-of-the-year. [\"The Note,\" Doug Supernaw, 1995]  In Texas we\'re getting ready to vote on a number of state constitutional amendments, including proposition 12 which motions to put caps on the amount of non-economic damages that courts can give to victims of medical malpractice. There\'s a lot behind this issue, but needless to say I\'m for it. Malpractice premiums have skyrocketed in recent years to the point that many doctors are closing their doors and moving elsewhere. And it\'s not like every doctor is getting sued. Only a few get sued, but courts have been awarding plaintiffs with millions of dollars in compensation that is for \"pain and suffering\" which does not include wage losses. Rural doctors and emergency medical physicians are particularly effected hard (which sucks, \'cause I plan to be both). Anyway, this is my plug for prop 12. Go vote! (Not that any of you who real this are in Texas, but whatever). [\"Green River,\" Creedence Clearwater Revival, 1976]  Have a great Labor Day, everyone.   Recommended Song: \"Like a Stone,\" Audioslave, 2002.   '),(3479,1808929,'28,August,2003','2003-08-28 00:00:00','   I have been studying my butt off for this cell science test coming up. It\'s September 2nd, so wish me luck. [\"I Wish It Would Rain Down,\" Phil Collins, 1989]  If I write with less frequency over the next week, that\'s why.  I did take a little time out to talk with some non-med student friends. Ashley Colburn is now all moved in at Temple in Philadelphia ready to start the physical therapy program. Laurie Glenn, err.....Laurie Biffer and I have now known each other for 15 years (since 2nd grade). We spent a while reminiscing about the good ol\' days at Krum. The crowd that we both ran around with is now sooo dispersed that I\'m not sure if we\'ll ever be toghether in one place again. [\"I\'ll Be Around,\" Joan Osborne, 2002]  \"The Big Chill\" is a really good movie about a circle of old friends, for years seperated, that find each other again after the death of one of the friends. I hope to find a lot of these people again, especially since my 5 year HS reunion is something like 9 months away. [ \"Vasoline,\" Stone Temple Pilots, 1994]    One of my newer friends here in Fort Worth, AnnMarie, was in a few of my public health classes this summer. We get along really well, for she is a former paramedic and we like to swap EMS stories. Well, over the summer her father died and that was a big burden for her. Long story short, I found out that her mother died just a few weeks ago, so in a period of 2 months she has lost both of her parents. I can\'t even imagine what she\'s going through; that\'s really scarey. [\"City of New Orleans,\" Arlo Guthrie]  Since I\'ve come back home I can tell that my relationship with my folks is much improved. Prior to Duke is was very strained, but now it\'s wonderful. I can\'t, and don\'t want to, imagine losing them at my age. If I were a religious person I would be praying for AnnMarie, but I\'ll be thinking about her while she takes the semester off.   Recommended song: \"Pets\" , Porno for Pyros, 1993.   '),(3480,1808929,'24,August,2003','2003-08-24 00:00:00','   It\'s been a pretty relaxing weekend. I went to see \"Open Range\" with some friends yesterday and it was a real good movie. The thing that was best about it was the character development. So many movies today just sluff off character development as a waste of time, but this movie (and many Westerns in general) showcased character development as its heart and soul. Of course the cinematography was awesome. I highly recommend this movie.  [\"Datura,\" Tori Amos, 1999]    I\'m going to call Jake Woods sometime today and talk with him. I was really hoping that I would have had time this summer to go see him in Wisconsin, but that just didn\'t work out. I\'m thinking about the possibility of a few days during spring break of next year, but I\'m not sure yet. He was asking me a while back if I\'ve really gotten involved with any organizations here at the medical school and while I\'ve looked into a whole lot of groups I can\'t say that I\'ve invested heavily into any. Med school is just very different from undergrad. Since I\'m a med student and also in the school of public health, my time is more strained than most other students, not in a bad way. I so far enjoy the program that I\'m in, but I\'m going to be way more cautious about what I dedicate myself to outside of the classroom. Besides, it took a year of bouncing around at Duke before I found the EMT option, so I\'m in no rush to committ.  [\"Ode to a Butterfly,\" Nickel Creek, 2000]  Have a good sunday, everyone.  Recommended song: \"Haunted,\" Poe, 2000.   '),(3481,1808929,'22,August,2003','2003-08-22 00:00:00','   I was talking with a good friend of mine today who told me how really lame it was to have a web log. \"Why? What are you talking about?\" I asked. [\"5 & 1/2 Minute Hallway,\" Poe, 2000]  He said that blogs are really pointless and that if anyone really wanted to know how someone was doing he or she would write/call or whatever. Perhaps he\'s right. Maybe this whole attempt is really just a waste of time. I\'ll have to think about this whole thing some more before making a definite decision. [\"Waste,\" Staind, 2001]  White Coat Ceremony was really cool. Afterwards my folks and I went out to eat at Joe T. Garcia\'s and the fajitas were aaaawesome! Jake Woods called me also, and though I didn\'t have time to really talk at length today we\'re hoping to catch up on sunday. Popo (paternal grandpa) will be in town tomorrow. This next week I\'ll have a lot more free time since they really worked us hard these first 3 weeks, but most of that free time will be spent studying. Studying is fine, but this new material is really tough. I haven\'t had biochemistry before and in some ways I wish I had taken it my last semester at Duke, but I took easier classes instead and mostly I\'m glad that I gave myself a break.  Recommended song: \"Old Man,\" Neil Young, 1972.   '),(3482,1808929,'21,August,2003','2003-08-21 00:00:00','   A few weeks ago I signed up for this running club. For those of you who don\'t know, I love running. I wouldn\'t call it fun per se, but it\'s similar to meditating in that it really makes you feel good after you\'re done. I\'ve been getting back into my running routine these past few months and I think I may have lost some weight. I\'ve never been a part of a running club, but I think I\'d like to try to take part in some local running events. [\"Mayberry,\" Rascal Flatts, 2002]  I\'ve never run more than 6 miles at a time, which is almost 10km I think, but I can currently do two miles in 15 minutes, not speedy but decent for me I think.  Ok, I think this 10 commandments in the courthouse thing is getting a bit out of hand. The same sort of thing happened 4 years ago in South Carolina with the Confederate flag controversy. [\"Brain Stew,\" Green Day, 1995] Just because one single judge wants his favorite statue in a federal building doesn\'t mean he can dismiss the seperation between church and state. However, people too quickly forget that one year ago we were all bitching about the \"under God\" part in the pledge of allegience. How can one justify keeping the ten commandments sculpture and yet keeping the pledge of allegiance untouched? [\"The Cinema Show,\" Genesis, 1973]  I have a lot of opinions on this, but basically I know that the founding of our country was intertwined tightly with religion and that the two, though now seperate, have a common heritage. Rejecting this fact is rejecting history, which is just goddamn foolish. We shouldn\'t have to get rid of all religious icons in politics, but we should be aware of  how  these things are being presented (without alternative religions displayed, without appropriate plaques describing the historical significance, etc). Whatever, Alabama\'s whack.  I\'m finding that a sense of community will probably be really strong here at TCOM. [\"The Power of Goodbye,\" Madonna, 1998]  Our class of about 120 people will surely become a really closely knit group. With White Coat ceremony approaching I think we all realize that we\'re all in this together. Some of it may be that we\'re a much more mature group (average age 25). I remember feeling the same kind of comaraderie when I was at TAMS (\'97-\'99), but I never felt that same bond at Duke, maybe between me and certain other individuals, but never as a part of a group. [\"Sweet Afton,\" Nickel Creek, 2002]  I guess I can\'t really say that. I remember living in the Arts Dorm and most of us in there really supported each other in a very sincere way. This was an exception. Duke was a wonderful place, but I will not be remembering it for its sense of community. [\"Putting the Damage On,\" Tori Amos, 1996]  I really look forward to the possibilities here in this new place.   My apartment is a wreck and I\'ve got to clean it and study more cell crap. Later.  Song of the day: \"Minerva,\"  Deftones, 2003   '),(3483,1808929,'20,August,2003','2003-08-20 00:00:00','   Today was a pretty good day I think. It\'s cool when people say happy birthday. It makes you feel like you\'re really part of the community. I didn\'t really do anything special today other than play ping pong for half-an-hour and allow myself to watch at episode of \"Enterprise\" that I hadn\'t seen. [\"Riot Poof,\" Tori Amos, 1999]  Oh, there was this former graduate of TCOM that came today to the emergency medicine club meeting to talk about his tactical EMS department in St. George, Missouri. He\'s a class of \'91 guy, but he\'s a medical director for the medical division of SWAT up there. He gave a very very interesting talk. For quite a while now I\'ve fancied the idea of being an EMS medical director, though I don\'t really think I\'d like being shot at. [\"Time Stands Still,\" The All-American Rejects, 2003]  Since I\'m more into public health and health policy issues I imagine that I\'d really like to delve into rural prehospital care research when I\'m not working in the ED setting. There\'s just so much goddamn untapped potential in this area. It won\'t make me any quick money, but it\'ll make me happy.  [\"Till Death Do Us Part,\" Madonna, 1989]    The UNTHSC White Coat Ceremony is this weekend and I\'m really looking forward to it. This is basically a med school ceremony where each entering student is cloaked with the white coat which symbolizes their new committment to medicine and health science. I\'m also going to go see \"Open Range\" this weekend; I\'m a very big fan of westerns. [\"Southern Man,\" Neil Young]  That\'s about all that I can think of. I\'m now going to continue studying my patient interviewing rubric and the hellish world of cellular glycolysis. When you thought you had it beat for the MCAT, it returns with a vengence. [\"The Rafters,\" Moby, 2002]  Song of the Day: \"Green River\"  Creedence Clearwater Revival   '),(3484,1808929,'19,August,2003','2003-08-19 00:00:00','   Tuesday, August 19, 2003       Ok, so to spice things up a bit I guess I\'ll talk a little bit about my non-existant love life. [\"Anytime You Need a Friend,\"  Mariah Carey] I guess in college my wild and crazy days were my freshman and sophomore years. I was dating someone all the time then, even at one point my own RA, but once I became an RA that came to an abrupt halt. I don\'t know what to think about med school, i mean I can\'t imagine that meeting people is going to be that much easier here. [\"Judy Blue Eyes,\" Crosby, Stills, & Nash]  I also have this problem of \"leading girls on\" when I\'m not interested. I\'m not sure why this happens, but it\'s a definite problem I have. I\'ve had several girls try to kill me because of this uncontrollable behavior. I\'ve thought about this a lot and I think that it\'s because I\'m a lot like my dad. My mom pretty much runs the household; my dad has worked hard all his life but pretty much supports my mom on whatever she thinks is best. [\"Precious Things,\" Tori Amos]  My dad, according to my mom, was and is a very quiet man. He was introduced to my mother by a mutual friend back in the \'70s. I\'m really lucky that they have such a strong marriage, but I have a feeling that my dad, twenty-three years ago, was generally fearful of women much like I think I am. That\'s right, I said it. I think I\'m afraid of women. My self confidence just isn\'t worth shit and I would much rather do my own thing rather than take a chance on talking to girl who I might want to get to know. In fact, I don\'t think that I\'ve ever \"made the move\" myself on asking someone out on a date (not including to lame dances and whatnot). Very fucking sad for a 23 year old. Ok, I\'m not helping matters, so shut up. Perhaps I unconsciously lead some girls on in some sort of ego boost thing. Yeah, I\'m really messed up. Did I tell you that I\'m an only child?       Ok, so about half of my medical school class is married. How much does that blow? I go from Duke, which is hook-up central, to a Ft. Worth, Texas medical program which is the matrimony capital of the state. Go figure. So yeah, anytime I meet anyone who I think might be cool to get to know, D\'OH! they\'re married. [\"You\'re So Vain,\" Carly Simon]  Arg, I gotta stop talking about this. It\'s gonna be my birthday, I don\'t want to be pissed.       I\'m probably done for the night. I don\'t know how often I\'ll be able to keep this blog up, but it might be cool to do a little bit every few days. I promise, all subsequent blogs will be shorter and more succinct. [\"Spit on a Stranger,\" Nickel Creek] Take care!   '),(3485,1808929,'19,August,2003','2003-08-19 00:00:00','   I really don\'t know what I would do without internet radio. I always have it going on in my apartment. It\'s amazing really; I can listen to any kind of music I want to at any time. [Currently I\'m listening to \"Trying\" by Lifehouse]. I had no idea that I would like this song so much, but the computer program anticipated what kind of music I might like based on the prior music ratings that I\'ve given it over the last 3 months. It\'s almost like some sort of symbiotic relationship between the internet and me: I get great music and the internet learns how I enjoy my music (along with millions of others) so that it can make more money. Sounds like a good deal to me!      I talked with my mother earlier in the day and she gave me the whole \"23 years ago I was pregnant and it was hotter than hell\" story. For those of you who don\'t know, the summer of 1980 was one of the hottest summers on record and it wasn\'t a great time to be pregnant. [Currently listening to \"Storms\" by Fleetwood Mac]   You know, it\'s funny what we chose to remember. I can\'t remember most of my birthdays, but I do remember my 5th, 10th, 17th, and 19th birthdays. There was something about them that just stuck with me. My fifth birthday was my FIREMAN-themed birthday party. It was awesome and I had lots of friends there. For my tenth birthday I went on a bike outing at a nearby state park. At 17 my friends at Krum, Texas threw me a surprise party, which was really special since I was leaving Krum and heading to TAMS (Texas Academy of Mathematics and Science).  [\"But I Do Love You\" by LeAnn Rimes]  At 19 I just got royally trashed with some friends the summer between high school graduation and college. Memory is quite a powerful force, but selective as all hell. If you don\'t believe me, just ask Marcel Proust.       I got a birthday card from one of my best friends from Duke, Ashley Colburn, yesterday and it had a lot of pictures included. [\"Live To Tell\", Madonna]  These pictures were like a blast from the past. Graduation, RA awards ceremony, Myrtle Beach, random crap, etc. In many ways the scenes from 3-and-a-half months ago seemed like days, yet years ago. It\'s just amazing to me how all of us at Duke lived furiously, one moment to the next, not looking back and rarely looking ahead, yet in an instant it all was over. It was sad, amazing, beautiful, and all too real. [\"Thank U\", Alanis Morissette]  Alanis really is an amazing singer. I really like her song \"Precious Illusions\" also. Some of the Duke EMS emails have started up again. I look back at the three years I spent with that group as some of the most awesome years of my life; I learned more from my experience there (truthfully) than most of my classes at Duke. [\"Cant Stop Loving You\", Phil Collins]  Ok, this song sucks. Phil Collins songs have just gone completely sucky since about 1995 or so. His songs are now all addressed to his kids and crap, which kinda sucks for us people who loved his \'80s, early 90\'s stuff.   '),(3486,1808929,'19,August,2003','2003-08-19 00:00:00','   Welcome! My name is Douglas Wayne Segars and I am yet another \"blogger\" that is just starting up. I really don\'t expect to keep up this habit, but who knows, maybe it\'ll stick. I kept a journal for about 3 years until I was 20, at which point I became a psycho pre-med f*ck-up and stopped putting my thoughts in writing. Now I\'m a f*cked-up first-year medical student in a whole new city, with a whole new set of peers and a whole new set of issues.       Tomorrow is my 23rd birthday. I think it\'s Blink-182 that said \"nobody loves you when you\'re 23,\" so I\'m not expecting much. However, maybe this blog can help capture in words some of the huge changes that I\'ve been through in the last 4 months, not to mention my new med school experiences. Ok, this\'ll do for an intro. Have fun reading, though I can\'t promise that it\'ll be spellbinding material. I lead a pretty low key life most of the time. Later.   '),(3487,1808929,'25,September,2003','2003-09-25 00:00:00','   I hate this f-in\' blog. I had a good one typed and saved, and now it\'s gone and I can\'t retrieve it. Fuck you Blogger!   '),(3488,1808929,'22,September,2003','2003-09-22 00:00:00','   This weekend was really awesome. It\'s probably been the best weekend since the July 4th weekend. On Friday night I had a small party at my apartment with about 6 people and we had a lot of fun. We were all really tame though. On Saturday I went to the Fort Worth Jazz Festival with 2 other people. On saturday evening I drove up to Denton where our family got together to celebrate Papa\'s 80th birthday. We had the party at my cousin David\'s restaurant at Valley View where I ended up having a very good grilled chicken salad for dinner! Pictures available at    http://f1.pg.briefcase.yahoo.com/dsegars@sbcglobal.net [\"Take Me Home,\" Phil Collins, 1985]  At first glance it seems like everything is just peachy with my family, and honestly I consider myself lucky to have such a good one. However, there\'s a lot of stuff going on with my mom\'s side. I don\'t really want to get into the nuances of our extended family relations, but it\'s enough to really stress Mom at times. Being away for four years was great, but I\'m glad that I\'m able to help her out now that I\'m closer. It\'s odd \'cause I always try to live with the idea that any relationship can be repaired, however times like these really make me see how some people cannot help but grow apart over the years, even if they\'re sisters.  [\"You and Me,\" Tom Petty, 2002]  This week I will be applying to become a member of the Tarrant County Medical Society. For some reason I have an interest in the \"public grievance committee.\" I have no idea what it does, but I\'m hoping to find out soon. Gosh, it would suck ass to be put on a committee that really blew. I need to talk to some people about it.  [\"Reasons Why,\" Nickel Creek, 2000]  In the last week I\'ve talked briefly with a number of my friends from Duke EMS, including Steve Barmach, Jake Woods, and Jen Farrell. I don\'t have a lot of time to talk usually, but occasionally I do, and when I do I really like keeping in touch with them. It seems to me that the squad\'s been going through a turbulent month. Those kinds of times really bring out the best and the worst in an organization. I\'ll be seeing Steve the second week in October, so we\'ll have plenty of time to chat about stuff.  [\"Suite: Judy Blue Eyes,\" Crosby, Stills, and Nash, 1969]  Gotta sleep now. Gross anatomy starts for me this week!  Recommended song: \"I Can\'t Stand the Rain (Supa Dupa Fly),\" Missy Elliot, 1997.   '),(3489,1808929,'19,September,2003','2003-09-19 00:00:00','   That excellent weekend that I was talking about earlier (see september 13th), well it\'s finally here! I\'m done with my cellular science section of medical school and my first round of practical exams. I think it went well for the most part. Gross anatomy and musculoskeletal science starts next week. [\"Go Your Own Way,\" Fleetwood Mac, 1977]  What am I going to do with my 3-day weekend? Tomorrow I am planning on watching the 10th anniversary show of Conan O\'Brian and probably some movie also with Melanie, Omar, and perhaps Juls. Saturday I\'m going to the Fort Worth jazz festival and that evening I\'ll be going home to celebrate my grandpa\'s 80th birthday. Papa is now among those people whom public health professionals call among the \"oldest-old\". He\'s still doing pretty well healthwise. [\"I Drove All Night,\" Celine Dion, 2003]  Kelli Modlin, one of my friends here at UNTHSC was in a car accident this afternoon. I think she\'ll be ok, but her car was pretty totaled. Melanie, Juls, Omar and I went to see her in the ER this afternoon. I got to talking to Melanie afterward about how odd this day was. First the weather was f*cked up royally today (alternating sunny and rainy patterns all through the day). Second....well I don\'t know how else the day was f*cked, but it just was, I\'m not here to figure it out. Maybe Isabel pulled some air pressure shit on us here in Texas. Anyway, I get so pissed and uptight about stupid crap all the time. Seeing Kelli in the hospital today really reinforced to me that I\'m really wrapped up in unimportant crap.  [\"Who\'ll Stop the Rain,\" Credence Clearwater Revival, 1976]  Ok, so now for more unimportant crap. It\'s now been 2 years and 2 months since I\'ve dated anyone. I\'ve gotta start looking in the public health school or the nursing school. This is rediculous, the best years of my prime are just going to waste. I\'m just sick of this crap...what\'s it going to take for me to get out there again and take a chance? Well, I\'m tired now, so I\'ll see y\'all later. [\"Whiskey Bottle,\" Uncle Tupelo, 1990]    Recommended song: Wilco, \"Kamera\" 2002.   '),(3490,1808929,'13,September,2003','2003-09-13 00:00:00','   Hey y\'all. I\'m still alive and doing all right. This weekend is no fun but next weekend should be EXCELLENT. By then I\'ll be done with cell science and ready to move on to musculoskeletal. I\'m not going to say too much tonight just \'cause I don\'t really have much to talk about.  [\"Don\'t Go Away,\" Oasis, 1997] Today was the proposition 12 day here in Texas. I talked about it a few weeks ago; it\'s a tort reform constitutional amendment that is supposed to help control skyrocketing malpractice insurance costs. We should know soon whether the amendment has passed and concurrently how many decades it\'ll take me to pay off my debts! Ah, no worries though, I\'m not in it for the money, but I would like to actually start a family before I\'m 40, so that\'s why I\'m concerned. [\"Juarez,\" Tori Amos, 1999] Aight y\'all, take care. Let me know how you\'re doing if you haven\'t in a while.  Recommended song: \"Traveling Alone,\" The Tony Rich Project, 2003.   '),(3491,1808929,'08,September,2003','2003-09-08 00:00:00','   What am I doing writing in this thing when I\'ve got class in 7 hours? I\'m going to be brief. I had a really nice weekend. The weather this past week has been unusually cool and pleasant. I got to be with my dad on his birthday and I also got to have lunch with 3 old friends from my first high school whom I had not seen in about 4 years.  [\"Tequila Sunrise,\" Eagles, 1976] It\'s just amazing to me that despite how much we\'ve all changed we all continued to act around each other the way we did at Krum, horsing around, making fun of each other, and having a great time. [\"Pink Moon,\" Nick Drake, 1994] Will McKnight is actually living in Fort Worth now. He recently bought a house and is now working at Lockheed Martin. I was really glad to hear that he is living so close, I had no idea before this weekend. Krista Wright.....er...Roseland is living in Killeen and is anxiously awaiting her husband\'s return from Iraq. She is working in the speech therapy program at an elementary school. Krista was the one who gave me my first bit of motivation to apply to TAMS, something that literally changed my life for the better. Nicolle Bowling is engaged I think and is working in accounting/marketing for the Farmers & Merchants State Bank in Denton. She is full of so much energy. Her and Krista are really the life of the party while Will and I just kinda look at each other and shake our heads in laughter as we watch the hilarity unfold before us.  [\"Harvest,\" Neil Young, 1972] Another test this wednesday, so I\'d better be going. September 11th is coming up again. Where were you when you heard the news? I was coming out of physics class at Duke returning a page I had gotten from my dad. My dad never usually pages me during the day, so I knew something wasn\'t right. He told me on the phone what had happened and I quickly joined the hundreds of shocked students in the Bryan Center as we looked at the tv screens in disbelief.   Recommended song: \"That I Would Be Good,\" Alanis Morissette, 1999.   '),(3492,1808929,'26,January,2004','2004-01-26 00:00:00','   Those two Mars rovers, the  Spirit  and  Opportunity  are awesome. I\'m hooked to those two robots just like I\'m hooked to ER and Enterprise. I mean think about it; those pictures that they\'ve sent back are from a completely different planet millions of miles away! How the hell did they figure out how to do that? Plus I\'m just so God-damned tired of the election year news: it\'s Dean....no, it\'s Kerry....no, it\'s Bush....no, it\'s whoever\'s got the jobs.  Spirit  and  Opportunity  are both out on the frontier of space exploration where one day man will hopefully live and have a chance to begin again.  Many people argue that we should be spending money to take care of our own first rather than trying to make it to Mars. My mother and I were talking about this the other day and that is exactly how she feels. While I really don\'t logical rebuttal to that, I just hope that my generation will accomplish something more than getting everybody health insurance, saving Social Security or whatever government program you want to think of. I don\'t want my generation to be known as the ones that had to clean up after and change the bed-pans of the baby-boom generation. If we have the ability to get out there into space then we should. There will always be shit to take care of here on earth. That will never change, but that shouldn\'t ever keep us from looking toward new frontiers.   This weekend was pretty good. I didn\'t really do anything special but watch some taped TV with a friend and study. I finally booked some tickets for Duke for spring break, so I\'ll get to go back and see my NC friends one last time. It\'ll be good and kinda weird all at the same time. Ok, I\'m off to bed. Have a good monday.   Songs of the day: The Flaming Lips,  Fight Test , 2003. Train,  Calling All Angels , 2003. Radiohead,  Creep , 1993.   '),(3493,1808929,'22,January,2004','2004-01-22 00:00:00','   Let me start tonight by reaffirming that I hate my damn computer sometimes. Oh yeah, now everything\'s fine, but just 5 minutes ago I was close to beating the screen in. There\'s no telling how many great blogger submissions this computer has taken to the graveyard. Somehow it senses when I\'m at my most vulnerable and it freezes up the file I\'m working on. Dammit I hate losing what I\'ve worked for.  I don\'t even know how to begin again anymore. Where do I start? I\'ll start again tomorrow....I\'m going to bed.   '),(3494,1808929,'18,January,2004','2004-01-18 00:00:00','   While I was at Duke I never watched TV. Three out of my four years there I didn\'t even have a TV in my room. One would think that in med school I wouldn\'t have time to watch TV; think again. I probably shouldn\'t be watching TV as much as I am, but I do. I mostly tape shows though and watch them on my own time with friends.       ER is one that I\'ve started watching in the past year. Yeah, it\'s been on for 10 seasons and I never really payed much attention to it until recently. I think I just like the characters more nowadays. I actually think I have a thing for Maura Tierney, the actress that plays Abby Lockheart. Her character is a nurse who\'s a 4th year medical student. She\'s had a whole lot of shit happen to her in the past few years so she\'s an underdog of sorts. I\'m really not sure why I think she\'s so attractive. Maybe it\'s because she\'s somewhat introverted like myself, I\'m not sure. Her character has been getting a lot of good episodes lately.      That show does sort of freak me out though. I know ER is fiction, but geez...am I supposed to be able to do all that stuff once I\'m in my rotations in a couple years? My first board exams will be June of 2005 which means I\'ll start studying for those in a year! All this stuff is going way too fast.    Song of the day Radiohead, \"Morning Bell\" 2000.   '),(3495,1808929,'16,January,2004','2004-01-16 00:00:00','   Wow, it\'s been almost 4 months since I\'ve posted to this thing. What the hell have I been up to? So much has happened, it\'s all gone by so fast. Right now though I am preparing for my second cardiopulmonary exam on monday. I am pretty mentally exhausted. I get about 5 hours of sleep each night, but tomorrow I am done by 4pm so I\'m gonna crash early. Today has not been a great day. It hasn\'t been bad, just not great either. I\'m pretty sure that I flunked my clinical medicine quiz today and I am probably going to have to see a doctor about a minor lateral scoliosis in my back. Oh yeah, my rent is going up too by almost 10% starting in March. Getting away from this recent shit, it looks like we\'re going to Mars according to Bush. If we get to Mars by 2020 I\'ll be 40 years old. That will be so freakin\' awesome. Hell, I wanna be an astrodoctor or something.   '),(3496,1808929,'21,February,2004','2004-02-21 00:00:00','   I\'ve recently posted a few photos on the following website:  urlLink http://photos.yahoo.com/dsegars@sbcglobal.net  My week off is just about over. I start back to classes on Monday, but I\'m really glad I had this chance to refresh and catch up with myself. I\'ve already mentioned how it snowed on Valentine\'s day. That was really cool. On Monday I had my public health class; nothing special there. Tuesday was pretty cool also. I got to work during the day for a home health service called Amedisys. Home health care is a service provided by Medicare part A, so being the health policy student that I am I had a great  time learning how home health works in Fort Worth. There\'s actually a lot of politics and economics that goes into home health care, which makes it that much more interesting.  Tuesday night I had about 5 people over to my apartment for dinner. I really spent way to much money on dinner for everyone; basically about a week\'s worth of food for myself was spent in one evening. I can\'t be doing that kind of thing often. It\'s time to tighten the belt a little.  About a month ago I learned that my Duke sophomore-year roommate, Alan Billharz, was going to be in Austin, Texas for a few days on business. On Wednesday I drove down to Austin to meet with him and three of his Microsoft coworkers. I had a really good time there. I got to stay with them at the Driskill Hotel (quite nice), I got to have a really good rizotto verde at this one Italian place, and we got to listen to a really good local band on Austin\'s 6th Street. I only really got to hang out with Alan for a handful of hours, but I\'m very glad I made the trip down. We got to catch up on lots of things; I really hope that I get the chance to get up to Seattle sometime.  Thursday I went back home, and Friday I got my eyes checked and talked to Jake Woods for a few hours. Today was a beautiful day to go running...and that just about brings me to right now.  I\'ve got about 3 weeks before I return to Duke for a few days. I think while I\'m there I want to try to do a lot of things that I never really got a chance to do: I want to revisit the Duke Gardens, lay out on the main quad and read the Chronicle, run along the Washington Duke trail and completely listen to my old friends without having a timetable running through my head. Those should be some good days.   '),(3497,1808929,'18,February,2004','2004-02-18 00:00:00','   I\'m in the middle of a much needed break. Having just finished my cardiopulmonary section I have about a week of very basic academic and out-of-class work to take care of. On friday the 13th I went home to Denton to try to beat the winter weather. Early saturday morning my mother woke me up after it started to snow and I leaped from bed acting like a kid in grade school. The snow was all over the place, so with my camera in hand I went to photograph this extremely rare event. The pictures are posted at the following site:  http://photos.yahoo.com/dsegars@sbcglobal.net   '),(3498,1808929,'04,February,2004','2004-02-04 00:00:00','   Tomorrow night is gonna be fun. For starters I\'m going to Hooters for a birthday party. While I\'m there I\'ll be watching the Duke-UNC game!!! Yep, and I\'ve even managed to get some of my med school friends interested in the game. It\'s going to be a good one.  It\'s been a rainy day today, and rainy days seem to make for reflective days. You know, the kind where you do more thinking than usual. Today I\'ve been working on a clinical medicine assignment that\'s required a little right brain thinking. It\'s called a \"spiritual history.\" It\'s just a quick statement that I thought I\'d add on to this blog. It\'s kinda deep though, so be careful reading it.   Faith & Practice I am not a religious person. I donâ€™t follow any faith though my family is Christian. However I do believe that there is a reason for everything that happens in this world and that there is an order at work that we donâ€™t completely understand. There are two somewhat spiritual practices that I follow through the best and worst of times, namely meditation and writing.  My meditation practice simply involves finding a calm place and clearing my mind with varying breathing exercises and postures. It is difficult to articulate exactly how meditation works for me, but I can say that it calms and centers me in a way that allows me to confront lifeâ€™s daily problems with a joy and clarity that I would not have otherwise. Writing, on the other hand, is my way of remembering important life experiences. My life is a collection of events and situations that have some meaning, and since my memory is so poor one of the best ways I learn and grow spiritually is to document meaningful experiences and reflect on them.   Importance Our society is extremely forward-looking, prospective, and it never looks back. Our days are mapped out in hourly schedules and even those individual hours are sectioned by successive powerpoint slides. Meditation and writing are both my ways of breaking out of this seemingly endless pattern. One of the primary goals of meditation is to take oneâ€™s conscious thought away from a timetable so that one may simply enjoy being in the present. Bringing myself back to the present moment free of all other external and internal influences is invaluable. I consider writing, however, to be my retrospective practice. It allows me to look backwards in time on my life experiences and give them some kind of meaning, great or small. Both of these practices help me appreciate my past and present as my foundation.    Song of the day: Neil Young,  Oh, Lonesome Me , 1970.   '),(3499,1808929,'25,March,2004','2004-03-25 00:00:00','   I\'m just about ready to kill me a cat. I\'ve got a particular one in mind. After I got home from class today I caught a cat chewing on one of my ferns. It bit a whole branch of one off and then acted all territorial when I chased it off. Yep, that mangy cat is goin\' down! Seriously, I don\'t take kindly to animals that just chew stuff for no reason. I mean it\'s not like cats usually eat plants or anything. If I were to just randomly put a dead carcass out on my patio, well then I\'d expect some cats to come snooping around, but plants? What the hell?! I wonder if cats are afraid of owls. I could find one of those decoy owls to stick on my patio; maybe that would keep the cat away. I\'m not too sure about wheather owls would scare \'em. If anybody knows please email me. However, I\'m pretty sure that cats are afraid of getting shot in the ass with a BB gun, but I don\'t think my neighbors would take to kindly to me shooting my gun around dinnertime. I\'m just going to have to poison the SOB.   Songs of the day  Toto,  Africa , 1982.  Badly Drawn Boy,  Silent Sigh , 2002.   '),(3500,1808929,'24,March,2004','2004-03-24 00:00:00','   Ok, I gotta make this quick, â€˜cause I want to sleep. Stuff is going well; renal physiology is slowly catching on in my mind. Did you know that the kidneys are 1/200th of our body weight, yet they receive 25% of our overall cardiac output?  Today I got to talk with two old friends of mine. Will McKnight, a friend of mine from Krum, left a message on my phone yesterday and I got back with him today. Iâ€™ve known for quite some time that he has a house now in Fort Worth ever since he started work at Lockheed Martin, but I have just completely neglected to keep in touch with him. I called him back today and I think we might be watching some basketball, TV, etc. this weekend. Iâ€™ve been saying now for months that I need to hang around with non-medical students sometimes for my own sanity, yet Iâ€™ve somehow forgotten about a handful of people I know in Fort Worth, including Will and my cousin, Jeff.    I also heard from one of my old HS roommates, Jacob Schofield. Heâ€™s now in flight school at Corpus Christi and today he sent me some pictures he recently took of him flying. I canâ€™t even fathom it, but Jacob is really going to be a pilot. I guess heâ€™s already somewhat of a pilot if heâ€™s flying certain planes around! Jacob is from Ponder, so heâ€™s one of those TAMS people that Iâ€™m pretty sure Iâ€™ll be able to keep tabs on over the years.   Thatâ€™s it. Have a good night.    Songs of the day  The Flaming Lips,  Are You a Hypnotist? , 2002.  Paula Abdul,  The Promise of a New Day , 1991.   '),(3501,1808929,'20,March,2004','2004-03-20 00:00:00','   Today has been one great day for college basketball. I think I\'ve watched about five games today and now I\'m watching UNC get snuffed by Texas. Tomorrow I have to return to med school mode so tonight is my last smell of freedom.   Today has been a big day of changes. My folks and I moved a 6-foot tall bookshelf into my apartment. I also cleaned out my dirty refrigerator and moved all my plants outside to my patio. I have quite a few large plants that have been in my room for most of winter, but not anymore. The biggest change was that I finally shaved. The great facial hair experiment of 2004 has concluded. It seems that most people don\'t like it, especially women. And even more so, I think that most of those who said it looked nice were just trying to be nice to me. After having shaved I barely recognized my own face. My chin seemed so small.  Anyway, it\'s gone now.  My parents, Papa, and I were on our way to the Golden Corral (Papa\'s favorite restaurant) yesterday when Papa asked me about the kinds of sports I like to play. I told him that I like running, volleyball, tennis, basketball and such. Papa, out of nowhere, starts laughing. He begins to tell a joke about something that Grandma Monie said probably about 70 years ago. See there were these two little girls arguing about what game to play. They went on and on until Grandma Monie told them to hush up and suggested that they both play \"dog\". Confused, the girls asked how you play \"dog.\" Grandma replied \"you come sniff my butt and I\'ll growl.\" Yeah, it\'s not exactly a joke; it\'s more rediculous than anything else, but it was funny as hell to Papa. After a few seconds his laughter became contagious and we all couldn\'t help but let out a quick giggle.   '),(3502,1808929,'18,March,2004','2004-03-18 00:00:00','   I woke up at around 10 a.m., took a shower, and packed up all my crap. At 11 a.m. I said my goodbyes to Shaomeng and headed to the Armadillo Grill where I meet with Dahl Clark one last time. She is looking forward to her future in physics research and starting her own lab equipment business. When she was talking with me I remember at one time she commented on how I seemed to have lots of friends. I was really baffled by this, for I donâ€™t think of myself this way at all. Itâ€™s all in the eye of the beholder. Her comment gets me thinking quite a bit. Will I ever just be happy with the way I am? 	Less than two hours before I was scheduled to leave campus I met with Jen Farrell at the Loop. She immediately expresses her disapproval of my facial hair. I tell her that sheâ€™s not alone, for just about every girl I meet has said the same thing. Jen is doing well and has recently returned from a WHO conference in South Africa. She and I share a common interest in public health and health policy, though she is more interested in international health issues while I am more drawn to regional Texas health issues. Childrenâ€™s health and development is important to her, while insurance coverage and obesity are my big issues. Finally we talked about her forced resignation from the squad back in the fall. I still canâ€™t believe that things happened to her the way they did. Itâ€™s incredibly unfortunate. She has a visceral dislike for Steve and the others involved with squad leadership. She said at one point that she felt like she was crying out for help so many times only to find that no one was listening. I canâ€™t imagine her anger, though now itâ€™s turned into a quiet frustration. I tried to let her know that she is tied in with the photographs and memories of so many people already graduated and we all think of Jen and DUEMS completely different from the way that she or Steve sees her involvement with DUEMS. She seemed to appreciate that sentiment. Seriously, before this year she was on the squad for 3 years. She and I joined the squad at the same time, yet our finale experiences couldnâ€™t have been different! Jen is going places though. Sheâ€™s going to be a serious asset to public health wherever she is and I hope to keep in touch with her over the years.  	I talked with Jen for nearly two hours and realized I had 10 minutes to buy my friend a t-shirt, get money from the ATM, and then get my luggage and meet Steve at the bus stop? Was it possible? Could Doug do it all in time? Well, not really, I was about five minutes late getting to the bus stop, but mainly because I ran into Lexie Riofrio and Matt McDaniel in the Bryan Center and Brandon Jones while in the EMS office. I wish I had longer to talk to all those people, but alas, no. As I was getting ready to leave, I heard a call go out! It was the first one Iâ€™d heard all week. Some person was disoriented after having been stuck in an elevator. It was a thrill just listening to the tones and dispatch once again.  	Steve volunteered to take me to the airport and we had a nice chat on the way over. Steveâ€™s a very quiet guy, so he didnâ€™t say much. We said our goodbyes at terminal C and I headed on my way. The best part of my vacation was over. It was time for me to leave and return to Texas. I was kinda sad at that moment. I mean, I really enjoyed being a Dukie once again. It was a strange combination of being welcome, yet out-of-place at the same time. On the plane I had to deal with this sadness and loneliness all over again. It was really the same feeling I felt on my graduation day a year ago. The plane started higher and higher until it finally broke through the rainy clouds. For the first time all week I saw blue, sunny skies stretching everyway. It was beautiful. This is corny, I know, but I believe that everything happens for a reason. At that time on the plane I thought about how lucky I was to have an opportunity to see so many college friends once again. How lucky am I also that I will probably keep touch with a few of them for quite a number of years?  	Iâ€™m now back home in Northern Texas and I wonder how I can bring that college passion to my life in Fort Worth. The city is great, there are nice people here and plenty of opportunities, but something keeps me closed in, detached from my world. What kind of bullshit is going on in my head? People are getting killed every day in Afghanistan and Iraq, 43 million people in this country are without health insurance, and millions of hypocrits in this country want to deny equal rights to gay individuals just because of religious fanaticism and/or bigotry. There is no excuse for my apathy. Unless I want to end up like Dostoevskyâ€™s Underground Man, I have got to push myself beyond my comfort zone. Itâ€™s time for me to move on.   '),(3503,1808929,'17,March,2004','2004-03-17 00:00:00','   I wake up in the freezing cold and peer out the window. Itâ€™s not raining, but it ainâ€™t sunny. I take the bus from east to west campus so that I can go visit Angie one more time and get a picture. Angie and I chat one last time and we take a quick picture. Iâ€™m very glad that I was able to have a courteous farewell picture with her. During the week of graduation last year so much was going on that I neglected to say goodbye to Angie. I noticed that one of my earlier pictures was on her wall, which made me glad. Angie kept all sorts of photos of students she knew over the years.  	My mom made some brownies before I left home. I brought them with me to NC and this morning I finished the big one that I brought. They were really good brownies; I brought them from home knowing that I really didnâ€™t want to be spending money on expensive Duke snacks while I was there.  	As Iâ€™m working on the ICS presentation for the squad meeting I get a call from Kim, which means that itâ€™s time for me to head over to east campus to meet her for lunch. We meet in front of the West Duke Building on East Campus, which is where we would always meet at the EMT-Basic class for 2 Â½ years. We head over to Georgeâ€™s Garage, which was really good. We spend most of the time talking about Kimâ€™s competitive cooking. Her boyfriend and her got started cooking as a couple and have had some recent success. I donâ€™t know that I exactly understand how competitive cooking really works, but she loves it. You know, itâ€™s all about passion. You have to find something youâ€™re passionate about and then go with it. Before we leave we briefly talk about how the squadâ€™s relationship with the EMT class has gone downhill. The squad leadership seems to have this â€œmy way or the highwayâ€ mentality that really frustrates Kim and Barbara. I tell her that Iâ€™m really full of regret about not leaving the EMT class more settled than it is. Months prior to my graduation I could tell that no one cared about the EMT class as much as I did. She assures me that it wasnâ€™t my fault and that these kinds of phases come and go. The current squad leadership will eventually graduate and others will take charge. I let Kim know that Omaira will be the new training officer and we both hope that she does a better job of working with the system instead of working against it. Change is a gradual process that takes persistent, patient effort. The current squad leadership attempted to push through immediate, instantaneous change and it only created a rift between themselves and the class.  	Sue-Ellen, Dyson, Sunny Wang and I all made arrangements to have dinner this evening. We end up going to Rickâ€™s Diner, which was where I must have eaten 25% of the time when I was a senior. Sunny is still Sunny, very funny and upbeat, but we are both able to commiserate about first-year troubles. Sunny is a first-year at Duke Med and she apparently is dating somebody she met while she worked at a firestation for two years prior to entering med school. I am real happy for her; I realize how much Iâ€™ve got to get out and experience new things, see new people. We also talk about trying to find Dyson a job. Sue-Ellen ends up buying mine and Sunnyâ€™s dinner because somehow I forgot to bring my money (is this a recurring theme?) and Sunnyâ€™s medical school flex doesnâ€™t work at Rickâ€™s for some reason. I need to remember that I owe Sunny about five dollars. Sunny comes to see the first part of the squad meeting before she takes off, probably to go study somewhere.  	Once I get to the EMS meeting the first person I see is Matt Womble. Matt Womble is another very inspiring individual. Why didnâ€™t I make an appointment to go see him this week? I know I thought about it, but I ended up thinking that heâ€™d be too busy to meet. We talk only very briefly and I can tell that heâ€™s starting to get some grey hairs. He is there to answer some initial questions about their MCI drill, which will happen next week. As he leaves he asks how long I will be in town; I tell him that Iâ€™m leaving tomorrow and I again I regret having not contacted him about my visit. What was I thinking? I will write him in the near future, for heâ€™s done so much for the squad and me.  	This is going to sound really corny, but I loved giving the ICS presentation and walkthrough this evening. I absolutely love presenting material to people in new ways and then watching them rise to the challenges I present them with. The squad members seemed very attentive and asked good questions. The ICS walkthrough went better than expected. We took it outside and actually put an ICS in place for the triage, treatment, and transport of 15 written mock-patients. It was pretty cool how all I had to do was set the walkthrough up and get it started before it took on a life of its own. Whatâ€™s amazing to me about ICS is how increasingly complex it can become in a matter of minutes. The amount of communication grows exponentially during that time. I fully believe that this group of EMTs will do fairly well on their drill. Roni Prucz asked quite a bit of questions; he must know that heâ€™s most likely to be the I.C. for the drill. Dave Strauss and others told me afterwards that they were very pleased and pleasantly surprised about how the training went. Ah, how great that felt. I still got the touch, eh?  	This evening Iâ€™m staying with Shaomeng Wang. Shaomeng and I were both RAâ€™s in Pegram my junior year. Heâ€™s very unlike myself, which is why I think we get along so well. Shaomeng is a â€œgangstaâ€ who doesnâ€™t like to take crap from nobody. However, heâ€™s also very thoughtful and a loyal friend. He judges people based on their character rather than superficiality, thatâ€™s a very admirable trait, especially at Duke. I feel kinda bad about staying with him that evening because heâ€™s got a project due the next day, but I get to talk with him here and there in passing. Heâ€™s overly concerned about how comfortable I am and so on and he goes overboard trying to be a good host to me. That was nice and very typical Shaomeng. He tells me about his girlfriend of about 7 months now. Sheâ€™s a very nice-looking young woman from Singapore, so now we both have dated Singaporeans. Heâ€™s going to Chicago next year to start a business job with ambitions to start an MBA in a few years. Very cool. 	The day is winding down and I start feeling a little sad â€˜cause tomorrow I will leave. These few days have flown by so fast and I really wasnâ€™t able to complete everything I wanted to complete. I did get to see just about every person I wanted, so thatâ€™s a major plus. Thereâ€™s really only one person left, the queen of busy, my old crew 1 member, and â€œmy baby momma,â€ Jen Farrell.   '),(3504,1808929,'16,March,2004','2004-03-16 00:00:00','   I wake up to semi-rainy weather, which keeps me from going on the run that I want to take. I get ready for the day and suddenly realize that North Carolina is in the Eastern time zone and that I have been going by Central time. This means that I have 20 minutes to meet Molly (my NC State entomology friend) at Mad Hatters. Yeah, I came really close to screwing that up. Molly and I both show up at Hatters at about the same time. We talk about Duke basketball, dating (or my lack thereof), and old Arts Dorm members whom are married, engaged, whatever. After Hatters, we roam 9th street, which has a handful of very interesting shops and eateries. Molly is looking for a wind chime to give as a gift for a bridal shower. The Regulator Bookstore is the next place we visit. While browsing I get stuck in the self-help section while Molly is looking at childrenâ€™s books. She finds an illustrated book on the life of Henry David Thoreau and the main character has been drawn as a bear. It was real funny. Franchescaâ€™s was our last stop. It has much more space after having taken over the building space next door. Molly recommends to me some mint latte or something; somehow in the years that Iâ€™ve known Molly I never knew that we both liked mint. We spend at least an hour in there talking about movies, our respective academic stuff, and whatever else. Itâ€™s funny because we talked about so much, I canâ€™t ever remember it later. I mainly just remember catching up and telling stories that made us both laugh. Ohâ€¦she and I have the same car. Theyâ€™re exactly the same. After 3 hours, we go our separate ways again. She drops me off at the WEL and then continues on the road back to Raleigh.  	I then call Stephanie, my old resident coordinator (RA boss). We meet up in her new office, a huge improvement from her office last year. Her hair is now short and curly. Stephanie seems to still be happy with her job so sheâ€™ll be staying around for a third year next year. Students who live in selective houses and fraternities have been pretty upset with the Residence Life Office decision to do away with the annual review, but Stephanieâ€™s view is that people are blowing things way out of proportion (some people claim that the university is trying to get rid of frats and selectives). Whatever. I tell Stephanie everything that Iâ€™ve heard from Michael Barham and Ashley Colburn, and we end our conversation laughing about the fact that there still is no furniture in the â€œdead-end alcoveâ€ on my hall. Go figure! 	In the EMS office Matt McDaniel and Michelle Hu are getting ready to do scenarios. I couldnâ€™t resist the opportunity to be a patient again so I ended up simulating a chest pain patient first and a hyperventilating mother second. On that last one I actually hyperventilated so much my finger and facial muscles kept consistently contracted for a few minutes afterward. I was real glad about how much Matt and Michelle have progressed over the year. Being a crew chief and doing things like this for my crew members was just about the most fun part of Duke EMS.  	What is it with the Duke food? I come back to campus and quickly realize that I am unable to finish the meals I used to get all the time. The food point system there on campus seems to promote convenient eating, which means I must have eaten more then than I do now. This must be true â€˜cause Iâ€™ve also had several people tell me that I look like Iâ€™ve lost weight.  	Matt catches me on the way to the bathroom and itâ€™s evident that heâ€™s anxious about something. I know through the grapevine that Matt has recently been demoted from crew chief. Heâ€™s actually had a pretty tough month and he seems to want some damage control advice; heâ€™ll be assuming the I.C. role in two weeks and heâ€™s beginning to feel a bit overwhelmed. We schedule a time to meet and I hope to help him in any way that I can.  	One of my favorite experiences was going with 7 Duke EMS members to the Durham County EMT-I class. They have the class on Tuesdays, Thursdays, and Saturdays in the evenings. Fortunately, Kathy Mellown is the instructor. Sheâ€™s a really good instructor who did one of our EMT-D classes about 2 years ago. The EMTs at the class (Steve Barmach, David Strauss, Jared Miranda, Steev Jones, Shelby Addison, Omaira Pacheco, and Mike Cheung) all had bruises from their previous class when they were practicing IV sticks. Tonight they learned about external jugular IV sticks and pharmacology. It wasnâ€™t a very interactive class, but I pretty much got the idea. I really wish I would have had that kind of opportunity sometime while I was at Duke, but then again Iâ€™m in med school and I wouldnâ€™t have had the chance to really use the EMT-I skills that I would have learned. What I thought was very good about the class was that Duke EMS was the most represented agency there and our members were able to network with several other people from different county agencies. I even got to meet with Josh Bordeaux who is an Orange County EMT. Josh (a UNC sophomore), Marcia, and I met probably about 2 years ago to talk about Joshâ€™s goal of establishing a campus EMS squad at UNC. While things havenâ€™t really panned out the way heâ€™d hoped, he is active with Orange County EMS and seems to be happy there.  	Cosmic Cantina is a staple of the Duke student diet. You can get large amounts of food for very very cheap. I actually forgot to bring money with me to the EMT-I class so I figured that I wouldnâ€™t be able to get any food with Dave Strauss and Steve Barmach. However, at Cosmic Cantina I was able to get a good- sized vegetarian taco for $2, which I payed for with spare change that I had in my wallet. Youâ€™ve got to love Cosmic. While I wouldnâ€™t call it the â€œbest Mexican food on the planetâ€ like they always claim, itâ€™s got to be one of the best deals on the planet. Oh, and itâ€™s also the place where youâ€™re most likely to get mugged in Durham. Fun stuff! 	David Strauss approached me earlier in the day about doing the incident command structure (ICS) training at Wednesday nightâ€™s meeting. How could I resist? This would be the third year in a row that Iâ€™ve done the training, and itâ€™s really no burden â€˜cause itâ€™s a training that I love to do. I didnâ€™t want the presentation to be â€œbusiness-as-usualâ€ though, so I needed to find a way to spice it up a bit. Before I went to bed I updated the powerpoints with a number of pictures from past MCI drills and I went about planning an outdoor activity for an MCI walkthrough. I loved having a training task ahead of me, for it gave me a focus and I began to get all wrapped up in it, just like I used to do.   '),(3505,1808929,'15,March,2004','2004-03-15 00:00:00','   I got up at 4 a.m. this morning to get ready for the day. I left my apartment an hour later in order to beat the rush-hour traffic. As I arrived the airport was pretty quiet, but as the sunlight started to gleam through the windows of gate C15 the amount of activity blossomed. I twisted and turned in my terminal gate chair unable to return to sleep. I peaked at my cell phone watch over and over again; it seemed as if boarding time would never come.      After about 7 cycles of CNN news I finally was allowed to board the plane. On the plane I could fall fast sleep, but only if I tried to read a recent JAMA article. The flight was uneventful, but as we descended to the RDU area I could see again the familiar tall treetops that always masked the horizon to the road traveler. Evergreens and bare deciduous arbors are everywhere.       While Iâ€™m in baggage claim I get a call from Melanie Michel, a friend from TCOM. We quickly catch up with either otherâ€™s respective spring break adventures. Mel offers her sympathies regarding the recent Duke loss to Maryland, and I thank her for the gesture. Yeah, Duke really sucked it up big time, yet somehow still maintains a #1 seed.       Mel and I conclude our conversation as I find my last piece of luggage. Outside as I wait for a ride I spot my first Duke contact. John Barrows was a Pegram resident of mine two years ago. Heâ€™s now a junior pre-med, and oh how I sympathize with the junior premed. Life is hell in that stage of life, isnâ€™t it? John and I are both waiting for our rides for about 15 minutes. John eventually sees his ride and bids adieu to me. Moments later, Scott Atkinson shows up in his new Passatt. He hasnâ€™t really changed much that I can tell. When I talk with him I realize what things have changed. Heâ€™s much more mellow. A year of carrying heavy Internal Coordinator burdens has brought him to a point where he must pick his battles carefully. The little things no longer concern him. With law school choices abound, he is looking forward to graduating and moving on to the next phase of his life.       Scott takes me all the way back to Dukeâ€™s campus, where there are all sorts of obvious physical changes that have happened over the year. Where else would I visit first upon arriving to Duke than the DUEMS office. In there I find Jared, Marshall, and Dyson. Itâ€™s really begun; Iâ€™m back at Duke and Iâ€™m seeing these people all over again. A number of things catch my eye first: the new office arrangement, the new squad uniforms, and yet another new computer. As we begin our hellos I suddenly feel welcomed. I wear no nametag; I am no second-class citizen; I am accepted, recognized, and appreciated all over again. More and more people arrive throughout the day at this seeming Grand Central Station. Most of the faces are familiar, but many are new, and some just downright confused me (the Jones twins have since exchanged each otherâ€™s hair style).       Without warning I walk into a newly refurbished Student Activities Office so that I might surprise Angie Thornburg. A quick â€œoh my goshâ€ sets the stage for another reunion hug. Angie is looking great and her day is fairly busy, but she skips a few phone calls for me so that we might have some time to catch up. We talk about her daughter turning 16, her search for a new house, and Jake Woodsâ€™s gossip. Angie was one of those women whom I was pretty conversant with day after day. She makes the Student Activities Office a fun place to be.       The sky eventually greyed so that it might rain at any time. I rushed to the biology building in a whimsical attempt to find Dahl Clark. Remarkably I find her in Bio 046. Her hair is shorter, neater, and her clothes look newer. In a complete academic role reversal she tells me that sheâ€™s switching back to physics after biology had rubbed her the wrong way over the years. She is on her way to teach her biology lab class so I leave her pressed for time.       David Strauss takes me over to the EMT-B class. On our way we talk about the successes and frustrations with the EMT class. Dave took on the heavy load of the EMT class last year, but since then he began to try and replace Kim and Barb. I told him that I regretting having left the EMT class situation on such unsure footing. I worked so hard to try to make the class a mainstay and an institution within our squadâ€¦their squad.       Barbara Jackson was one of my EMT instructors back in the summer of 2000, and sheâ€™s still teaching it. She had a pretty genuine smile on her face the whole time I was visiting with her, but I had to wonder what was really going through her mind      On the phone Sunny Wang was energetic and funny. One of my past crew chiefs, she is now a first year student (like myself) at Duke Med (unlike myself). We were making plans to meet on Wednesday for dinner.      Jake Woods is still Jake Woods. On the phone he was willing to drop everything he was working on just to talk to me. Heâ€™s got this new online startup project that he told me about, something like an online personal trainer. It sounds like a pretty extensive venture; he got the idea after talking business with his own personal trainer.      Steve and Jared have their own plans worked out for the future. Both will have their EMT-I by this fall and plan to be working for Durham County EMS. Steve, who will be taking this next year before entering medical school, will be living with Jared in a nearby apartment starting this summer. Steve has been quite nice and hospitable. Heâ€™s housing me for two nights of my stay. Like many other seniors heâ€™s biding his time â€˜till graduation. The MCI drill, the rugby tournament, and any possible NCAA basketball bonfire staffing will be all she wrote for Barmach before he moves on to bigger and better things. Jared, a junior, is tired of Duke. A crew chief now, he doesnâ€™t seem to have the same zeal for training that I remember having. Heâ€™s frustrated by an intense bureaucratic structure that is choking him. Duke EMS has taken him only so far. He needs to be challenged in new ways, and I hope that working for Durham County will be that new challenge he needs. Steve, Jared, and I stood for long time outside the West Union arch discussing these important life choices and other squad issues that have recently come up. A night-time rain begins slowly and at this point and I am once again amazed at how beautiful this campus really is. I took so much for granted.       As the rain starts to fall, I make my way over to Brownstone with Scott. As I arrive I immediately see Julie, the â€œEMS groupieâ€ and one of Suelâ€™s good friends whom Iâ€™ve come to know over the years. I was very happy to see Sue-Ellen again. She seems more vital and lively than I remember seeing her last. Ah, thinking of those days when we were very good friends. Somehow our dedication to the squad warped our relationship seemingly beyond repair, but once again time heals wounds and we can once again laugh about all those stupid things. Pam, Sarah, Trey and other Brownstone people I know come to say hello again; I begin to feel the symptoms of social anxiety disorder (that I self-diagnosed) return as everyoneâ€™s eyes are focused on my faceâ€¦and my new facial hair, which seems to be getting mixed reviews.       I eventually settle down on east campus after 1 a.m. in Steveâ€™s room. Before going to sleep Steve asks me what I think of the squad so far. Thatâ€™s certainly a big question and at first I donâ€™t know what to say. Everything Iâ€™ve seen thus far is mainly superficial; perhaps Iâ€™ll know more before I leave. But finally I tell Steve that Iâ€™m not here to judge or critique. I have my own life now away from all this. My hope is that all the graduating seniors can look back on their EMS experience in a positive light. If my visit can help some people put things in perspective then the tripâ€™s been worth a dern to me. All the while I think of Jen and her fall resignation. Iâ€™m confused by it all and wish it didnâ€™t happen the way it did.       Itâ€™s been one long day and I began to rapidly tire. My next few days would be booked, but I donâ€™t think about that as I drifted off. This stuff is so much fun. I realize once again why it meant so much to me. Actually, I donâ€™t. Itâ€™s an intangible sort of intense passion that I now associate with Duke EMS. Now in medical school, I hope to develop that same passion for â€¦something, just something that might give me the same sort of direction and personal satisfaction.   '),(3506,1808929,'14,March,2004','2004-03-14 00:00:00','   (From March 12, 2004 before falling asleep)  Itâ€™s spring break, thereâ€™s nothing to do, and Iâ€™m loving it. In less than 48 hours Iâ€™ll be back at Duke for who knows what kind of spring break craziness. I bought my tickets a number of months ago back when it felt like spring break would never arrive. Now that I have to think about packing Iâ€™m not real sure why Iâ€™m going back. Itâ€™ll be a fun trip, of course, but it really doesnâ€™t even feel like Iâ€™ve been away for a year. Why come back now? What is that all about?  Why is it that I always have to think of a reason why? Canâ€™t I ever be spontaneous at all? So what if I feel like Iâ€™ve burned all the bridges behind me. Damn, this is what I hate about blog writing. What I end up writing is always so damn self-depricating. I think ideally I would like to remind myself of all the people and places who influenced who I am and where I am in my life now. While at Duke I had mentors, role models and a sense of direction (mostly in my last two years). Now I feel like I\'m adrift once again, freshman-style.   '),(3507,1808929,'07,March,2004','2004-03-07 00:00:00','   Yep, I\'m still tired. One week left before spring break. I really haven\'t been doing anything lately besides studying. Hell, I didn\'t even leave my apartment saturday, that\'s how bad it is. I\'ve got 4 exams this next week, but that means I can spend spring break however I want to. This last thursday I did get to go see The Passion of Christ. Man, I still get chills when I think about some of the scenes there. I really liked the movie and I think it made it on to my top 100 list. Yes, I keep a listing. Whoa...now I have something to talk about, cool! Ok, a while back I made a listing of my top 100 movies. I rated them based on 3 independent criteria (quality, # of times seen, desire to see again). The top 20 are as follows:  1. The Last Picture Show 2. 2001: A Space Odyssey 3. The Wizard of Oz 4. Man of La Mancha 5. Terms of Endearment 6. Lost and Delirious 7. What\'s Eating Gilbert Grape 8. The Last Temptation of Christ 9. Contact 10. Lost in Translation 11. All Quiet on the Western Front 12. Lord of the Rings: Return of the King 13. Six Degrees of Seperation 14. Born on the Fourth of July 15. Playing by Heart 16. Ghandi 17. Good Will Hunting 18. Unforgiven 19. Dead Poets Society 20. The Big Chill   The Last Picture Show and Terms of Endearment were both written by Larry McMurtry, my favorite fiction writer. See them both and you won\'t be disappointed. OK, gotta study again, so see you later.    Song of the Day: \"St. Teresa\", Joan Osborne   '),(3508,1808929,'01,March,2004','2004-03-01 00:00:00','   I\'m so tired. I think I got 8 hours of total sleep this last weekend. I went down to Austin with our local American Medical Association (AMA) chapter for a conference. It was alright, but I\'m still trying to recooperate from it. So yeah, this morning I took my G.I. exam and I\'ve just been vegging since then.  ..... OK, I think I\'m too tired to really write paragraphs with a complete thought anymore, so I\'m just going to write random stuff. This thursday I\'m planning on going to see that movie the Passion of the Christ with a few people. Spring break is 2 weeks away. I\'m really stoked about these new tablet PC\'s that they have. I\'m hoping that my next computer will be one.  I only saw about 20 minutes of the Oscars, but I did get to see Jack Black and Will Farrell\'s duet. It\'s soooo hot right now. Later!   '),(3509,1808929,'17,April,2004','2004-04-17 00:00:00','   I\'ve got these huge reproduction exams on monday and tuesday. It\'ll be real good to get those done. But you don\'t want to hear about my exams, no sir. Well, how about this...I\'m taking a group of people to the White Settlement Health Fair tomorrow for our Texas Medical Association group. That\'ll be cool. No? Well how about this. Last week I went out to the Fox and Hound bar downtown for the first time with a couple friends. I actually had about 4 drinks, which is more than I\'ve had in quite some time. We spent a good amount of time trashing some of our more annoying classmates.   There\'s some way I\'m supposed to be able to put pictures on this biotch, but I don\'t know how just yet. Oh, I just remembered something I can write about. For the last week we\'ve been learning about the well-man and well-woman exams. These include breast, genital and rectal exams. Next week I will be doing all of those for the first time. We will actually be doing these exams on real patients, which is a huge deal to any first year med student like myself. I mean, huge as in a little nerve racking, yet intriguing at once. Seriously, how many times in your life does a normal person feel cervixes, prostates, and whatnot? That\'s really all I can think of. Peace out, fools.   Songs of the day   Heart of Gold , Neil Young, 1972.  Bigger Than My Body , John Mayer, 2003.   '),(3510,1808929,'06,April,2004','2004-04-06 00:00:00','   I hate UConn.   '),(3511,1808929,'04,April,2004','2004-04-04 00:00:00','   These last few days have been pretty good. The part about Duke losing in the NCAA semifinals really sucked, but everything else has been pretty peachy here.  I\'ve been meeting with a number of different people lately and having pretty good conversations with \'em, so that\'s what I\'m going to write about this evening.  Two nights ago I was over at Edmund Ho\'s apartment. He fixed dinner for Peter, Omar, Melanie and myself. It was real nice \'cause it got us to forget about our awful renal tests earlier that day.  Edmund is Houston Rockets fan and is actually pretty politically-minded. We got to talking about Bush vs. Kerry, the obesity epidemic, gay marriage, all kinds of things and I really enjoyed it. We share many social views, but differ a lot on the U.S. foreign policy/foreign aid ideas.  While I was at Edmund\'s place I got a call from Michael Barham (my old RA supervisor and good friend) and I was really surprised to learn that he was in Philedelphia with Ashley Colburn (my RA friend from Asheville, NC). He\'s there on a conference until sunday. It would have been fun to be there with them, but I was at least glad that they seemed to be having a good time together.   On friday evening I worked first aid at the Fort Worth Race for the Cure. I put in about 10 hours of work and was completely spent after it was done. While I was there I worked with Gabe Tarango and Gary Howard. Gabe Tarango is in my first-year class and it a pretty fun guy to talk with. Like me, he lives a somewhat tame and boring life right now, but last year during his senior year of college he was a party animal. He was also dating an older, very nice looking physical therapy student for about 7 months. We both talked at length about wanting to get out into the city and meet new people.   Gary Howard is an EMT and paramedic-student. He was pretty much running the first-aid show there at the Race for the Cure. Damn, I don\'t even know where to begin with this guy. I should say that one of the fascinating things about EMS is that the people are so damn interesting. Seriously, you have to some kind of seriously interesting character to want to get into EMS and Gary was certainly interesting. I\'d say that Gary is probably in his 50\'s, is a Vietnam War veteran, and came into EMS a few years ago after selling his previous two businesses. He\'s also a bit of a cowboy and has a shit-load of stories about saving stupid people\'s lives out in the middle of nowhere. We talked for most of the evening and it was pretty enjoyable. Talking with him helped remind me how much I miss EMS.   Will McKnight came over on saturday to watch the Duke game. We had fajitas. Will and I spent a lot of time looking at crazy crap on the internet and yelling at the TV. We did talk a little bit about the lifestyles of those on \"regular hours\" versus those on \"student hours.\" He said he misses being able to stay up past 10pm, whereas sometimes I feel like I would do anything to go to sleep and get up earlier. We had a good time I think. It\'s very reassuring to know that I have friends outside of my med school class.    Like I said, stuff is going pretty well right now and my renal grades turned out ok. The weather\'s nice and my allergies are subsiding, so I\'m pretty happy. I\'ve been in Fort Worth for almost a year now and I am starting to feel at home.    Songs of the day  Moby,  Porcelain , 1999. Collective Soul,  10 Yrs Later , 2000.   '),(3512,1808929,'31,May,2004','2004-05-31 00:00:00','   On Memorial Day I want to write real briefly about my closest connection with a war veteran. When I was 5 years old we were living in Cross Plains, Texas. Our next-door neighbors were Hubert and Celia Beleuw. They were an older couple in their 60s at the time. Both my mom and dad worked and I wasn\'t in school yet, so Hubert and Celia (everyone called her Sis) took care of me for a whole year. I have lots of stories about them taking care of me during the days, but I didn\'t find out until many years later that Hubert was in World War II. He was a decorated infantryman at the Battle of the Bulge. While he was in combat he started smoking and he kept on smoking for 40 years until the day that they saw that my mom needed help finding day care during the days. Right before they started watching me he kicked his cigarette habit cold-turkey. All that time I was just too young to really appreciate Hubert and Sis; they both passed away about 4 years ago. They were a part of \"the greatest generation.\"   '),(3513,1808929,'30,May,2004','2004-05-30 00:00:00','   Last night there was a party over at Joey Heitzman\'s place. Joey lives just two doors down from me. That guy can cook. He fed about 10 people and all of us were amazed at how much effort he put into it. John Abraham brought over some Knob Creek bourbon too, very good stuff. At midnight we all went out to the Flying Saucer down at Sundance Square. All of us were pressuring Molini to get a tatoo. She\'s the most straight-laced, level headed person among us, so it was funny...you just had to be there.   Today I\'m meeting with my folks at the Gaylord Texan Hotel. It\'s this new place that opened up that\'s supposed to be one of the nicest places in the Southwest. I guess I am getting older when I look forward to spending some time with my folks. Man, just 5 years ago I couldn\'t wait to be 1,200 miles away from them. Anyway, I need to start packing.   A local radio station is doing all 80\'s all weekend. Certain 80\'s rock can be very cheezy, granted, but it\'s unlike any other kind of music and it\'s great for when you just want to forget everything else going on. My songs of the day reflect my appreciation for the phenom that is 80\'s music.    Songs of the day   The Boys of Summer , Don Henley, 1984.  Every Breath You Take , The Police, 1983.  Your Love , The Outfield, 1985.   Pour Some Sugar On Me , Def Leppard, 1987.   If You Leave , OMD, 1988.   '),(3514,1808929,'29,May,2004','2004-05-29 00:00:00','   Memorial Day weekend is here finally, and summer is finally here. It\'s about damn time. I\'ve still got a month left of school, but this three day weekend is much needed. Pathology is pretty grueling work; I can\'t remember having ever studied harder than I studied this last week, but for now I\'m just going to try to forget about all that and hang out with friends and family.   Last night there were about 8 of us over at Omar\'s place \'till about 2am. It was awesome to finally hang loose a little bit. Finally there was someone there who would drink with me! I don\'t know what it is about the friends I know, but none of them ever seem to drink. I think I\'m caught in the middle sometimes: too introverted and reserved to have fun among the drinking crowd, yet too ... **I don\'t know the word** to be able to have fun with those who don\'t drink. Basically, I gotta have something to loosen the tension, be it EtOH, Zoloft, whatever. Anyway, last night was fun.   A couple days ago I heard from my good friend Fernando. He\'s one of my best friends from TAMS and is finishing up his business degree at UNT. I hadn\'t seen or heard from him since new years day. The last time I saw him he gave me a DVD of all the videos he took of our group of friends while we were at TAMS. It was a very cool gift. However, we got into this unfortunate discussion about how I\'ve changed. Fern thinks that since having come back from Texas I\'m much more of a hard-ass. I got really upset about that. I\'m not upset about what he said \'cause it\'s mostly true. I\'m not like I was at TAMS; People and situations change. Isn\'t that a normal part of the college experience? What I\'m upset about is that he doesn\'t appreciate how and why I am the way I am. What I\'m upset about is that Fern likes to live his life like a series of Friends episodes. The videos he took of all his friends in HS are picture perfect, but they\'re all taken out of context and detached from real life (which is always less sit-com and more prosaic). I don\'t know...this whole topic is a little depressing. Moving on now...  I don\'t know why people think that the Pirates of the Caribbean movie is so cool. I saw it today and it was pretty boring I thought.    Songs of the day   I\'m Always In Love , Wilco, 1999.   Desperately Wanting , Better Than Ezra, 1996.   Let Down , Radiohead, 1997.   '),(3515,1808929,'17,May,2004','2004-05-17 00:00:00','   So it seems that when I was writing my last blog I had some anger management issues. I\'m doing much better now and I\'m far less irritable. These last few days haven\'t been too eventful (studying and all), but one of the cool things is that I\'ve made some connections to and between old friends.        I recently got a random IM message from Anup Shetty and I found out that he\'s in his first year of medical school at UT Southwestern in Dallas. This was really cool news. Anup Shetty and I knew of each other all through our time at Duke. Our last names were close to each other, we took a lot of the same pre-med classes, we were in the same freshman dorm and we had a lot of mutual friends, however, we didn\'t (and still don\'t) really know each other that well. Now that I\'m in Fort Worth I\'ve realized that there are just some occasions (namely basketball season) where you wish there were some other Duke people around, and who else is better than Anup? He\'s a real nice guy and he was even more of a Cameron crazy than I was. I\'m going to have to get over to Southwestern sometime since there are now about 3 people I know over there.       I was chatting with Will McKnight last night. Will and I went through grade school together at Krum and we now both live here in Fort Worth. He said that he ran into and spent some time hanging out with Caleb Stone and Jeremy Villeneuve recently. Caleb and Jeremy were also friends of mine all those years at Krum, but I haven\'t seen or heard from them in 5 years. Will was also able to get a phone number for Tim Purnell. Tim was one of my best friends at Krum, but after he graduated from Krum I lost touch with him completely. See, this is what happens when I spend 4 years out of the state. Tim is married now and apparently lives in Sanger somewhere. I\'m planning on contacting him sometime soon. How do you make that call though? \"Hi, this is Doug. I haven\'t heard from you in 5 years. Sup?\"       Finally, I just found out that one of my current TCOM classmates, Liz Rankin, knows Will McKnight from Austin College. She dated Will McKnight\'s roommate at some point while she was there. I wouldn\'t have figured there\'d be a connection between the Krum and TCOM spheres of my life, but I guess there\'s only six degrees of separation between any one person and everyone else on the planet so I shouldn\'t be surprised.        Finally again, I found out that a Texas friend of mine is pregnant. I won\'t say whom, but she\'s already 5 months pregnant. It\'s been about 5 months since I\'ve seen her also.        That\'s it. I\'ve got to study now. This time around it\'s genetic disorders. Peace.    Songs of the day   Linger , Cranberries, 1993.   Ava Adore , Smashing Pumpkins, 1998.  Gold Dust Woman , Fleetwood Mac, 1977.   '),(3516,1808929,'15,May,2004','2004-05-15 00:00:00','   **This blog entry is rated PG-17 for language** I wish I could always say what\'s on my mind. Wouldn\'t that be great if we could all just say whatever we wanted without any consequences? See, there\'s this one person who I\'d really like to tell off. However, we work in an organization together so I can\'t exactly just tell her to go fuck herself like I would like to. Instead I just try to avoid talking to her at all, but man if I could only let the language flow from my mouth like water from a fire hydrant I\'d tell her to go lock herself in a dark room and fuck herself repeatedly. She\'s such a bitch, and not in that cool Meredith Brooks kind of way either. This is a cold, petty, stupid, and irritating kind of bitchiness. Do her bitch-powers have no bounds? As I recall the stories of young King Arthur pulling his sword, Excalibur, from the stone I wonder who but King Arthur himself can pull out the stick from this girl\'s necrotic ass?  In more political news, my mother is actually considering voting for some democrats this time around. My mom is a veteran teacher who\'s been screwed over along with countless other teachers. Teachers in her position have their salaries frozen at 20 years of teaching experience. When you combine that with the fact that their health benefits keep on getting cut and that their requirements for retirement will probably increase it means that she will be getting paid less each year for the next several years and that she\'ll be forced to work 5 more years just to qualify for retirement. Educators are getting fucked by the Texas State Legislature. It\'s been this way for years and it\'s been due to dems and reps alike, but generally democrats are more friendly to teachers. I\'m telling you something, whenever you get a conservative like my mother to consider voting democratic you know you\'ve really fucked things up for her.   Songs of the day:  I Am the Highway , Audioslave, 2002.  On Your Side , Pete Yorn, 2001.   '),(3517,1808929,'08,May,2004','2004-05-08 00:00:00','   Pathology is pretty tough. I just barely passed my quiz today, so I really need to spend some time studying and less time blogging and playing volleyball. Anyway, there\'s nothing really going on with me lately. Tomorrow I\'ll go home for Mother\'s Day and sunday will be one year to the day that I graduated from Duke. It\'s amazing because it seems like it was yesterday.   Speaking of it feeling like it was yesterday, today I got a call from Michael Barham of all people. He was on a 7-hour layover at DFW so I got to go pick him up and show him the Fort Worth scene. Michael Barham was my graduate assistant (RA-boss of sorts) my senior year at Duke. The last time I saw him was a year ago on graduation day.   I guess that\'s it. See ya later.   Songs of the day   Interstate Love Song , Stone Temple Pilots, 1994.   Lonely People , America, 1975.   '),(3518,1808929,'02,May,2004','2004-05-02 00:00:00','   These past two weeks have been pretty eventful. I bet I had something like 10 exams in the last 14 days. The good news is that my first year of medical school is almost over. I\'ve just got an 8-week pathology course left. Basically, that means I have no labs, no patients to see, no nothing, just a whole lot of book learning. I don\'t mind this at all because it means I can concentrate on one thing for once.  On Monday I worked for a few hours in the hyperbaric medicine department. High pressure oxygen is often used to expedite the healing process for your more difficult wounds. The most common hyperbaric patients seem to be diabetics (since T2DM is endemic in this country), but hypoxic patients also benefit from hyperbarics. Among other things I got to see an awfully narly food wound (caused by diabetic neuropathy and peripheral vascular disease) that made me a little bit queazy.  On Tuesday I had my final physical exam test. I got 237 out of 250 points, which is not bad I don\'t think.   On Wednesday I spent most of the day traveling to Corpus Christi for my 2-day preceptorship, which took place on Thursday and Friday. I was assigned to Dr. Vanderheiden who runs a family practice clinic at the Bay Area Medical Center at Corpus Christi. I quickly learned that Dr. V was a challenging physician to shadow. Dr. V frequently quizzes his students in front of patients and wastes no time letting you know what he thinks of you. At the end of the first day I was exhausted. I felt so stupid; I have got so incredibly much to learn. By the second day I had his personality figured out a little bit more and he had a better idea of where exactly I was in my training. By the end of the second day I felt much better and I now have a better idea of what my rotations will be like.   Today was all for me. I did nothing but relax and enjoy the beautiful day. Tomorrow I\'ll probably go and get a haircut.    Songs of the day   Love Profusion , Madonna, 2004.   Walk of Life , Dire Straits, 1985.  Good Stuff , Kelis, 1999.   '),(3519,1808929,'26,June,2004','2004-06-26 00:00:00','   Lessons Learned                                                           It\'s still hard to believe that my first year of med school is done. I now have all this time on my hands that I don\'t know what to do with...well, not so much time. There\'s really only 5 weeks.    alt=\"My friend Peter\">  It\'s about that time of year when everyone\'s getting back in touch with old friends and whatnot. In the last day I\'ve gotten emails from or talked with 5 Duke friends and Will, a guy I went to high school with. One thing I\'ve started doing is writing letters to people. Why? Well, \'cause email sucks, that\'s why. Shouldn\'t email just be used to send quick messages to people? More meaningful messages just don\'t work with email. For these kinds of messages you just can\'t beat a real hand-written letter. It\'s an art that\'s been lost among our generation.   In the last year I\'ve learned some pretty important stuff that isn\'t tested on board exams. I\'ve listed a few of these lessons below:  1. It\'s easy to blame the world or others around me, but usually I\'m the biggest cause of my own problems. See, a year ago I thought that coming back home would solve the problems that I had while I was at Duke. While parking is much easier here, the real problems I have don\'t resolve themselves just with a change in venue.   2. I spend way too much time feeling unsatisfied. I really should be more thankful for what I have, especially given that I\'ve never really had to fear for my own life.   3. I\'m not used to change and/or asking for help, but there\'s an absolute time for both. Pathology was probably the most difficult course for me this year; I would not have improved as much as I did without the help of my tutoring group.   I think those are the biggest 3. Oh, hey, I almost forgot to mention...I\'m going to Las Vegas! On July 11th I\'ll be going with 3 other friends to Vegas for a week of crazy, kick-ass fun. Awe-some!    Songs of the day   Breaking the Girl , Red Hot Chili Peppers, 1991.  Trouble Me , 10,000 Maniacs, 1989.  All Nite (Don\'t Stop) , Janet Jackson, 2004.   '),(3520,1808929,'25,June,2004','2004-06-25 00:00:00','   Stick a fork in me                                                       I\'m finally done with year one and it feels quite good. At last it\'s time to let loose.   '),(3521,1808929,'25,July,2004','2004-07-25 00:00:00','   Frustration with Election-Year Politics    &nbsp;&nbsp;&nbsp;&nbsp; I hate today\'s election-year politics. Month after month it\'s the same polarizing talk of status-quo and reform that leads to little or no action. And to think that the hours upon hours of televised political discourse only matter for about 10% of the voting population that hasn\'t already made up its mind yet.  &nbsp;&nbsp;&nbsp;&nbsp; It\'s ironic to me also that while there\'s so much political bickering there\'s very little difference between Bush and Kerry on the big issues. Both of them were in favor of using force in Iraq (though Kerry would have preferred to wait it out longer). Both of them support Bush tax cuts (though Kerry would repeal the tax cuts for a very small percentage of the population). Both of them are against gay marriage ultimately, though Kerry would pursue a more hands-off approach to banning gay marriage. Bush added a perscription drug coverage element to medicare while Kerry has not really opposed it or provided any details on health care reforms. &nbsp;&nbsp;&nbsp;&nbsp; Don\'t get me wrong, there are several areas where Bush and Kerry differ widely (education, affirmative action, the U.S. role in the U.N.) but these issues don\'t get near as much playtime as those mentioned previously. Why that is I\'m not sure. So while the political pundits keep on punting their political footballs back and forth I might as well go fix myself a drink, sit back, and watch some Star Trek.   '),(3522,1808929,'17,July,2004','2004-07-17 00:00:00','   Medical Preceptorship Anxiety                                          &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Iâ€™m a little nervous. In less than two days Iâ€™ll start a 4-day preceptorship with Dr. Perry in Sanger, Texas, but Iâ€™m about 3 weeks out of my first year of medical school and Iâ€™m almost 3 months out of my first clinical medicine class. I know that Iâ€™ll learn a lot, but when I think back to my first day working with Dr. Vanderheiden about three months ago I remember it being so rough. That experience and the final pathology course really gave me a kick in the ass and showed me how stupid I really am. I guess Iâ€™ve got to stop worrying about what I canâ€™t control and instead focus on doing a good job for the one D.O. that works in my neighborhood. Itâ€™s a cool opportunity.   '),(3523,1808929,'16,July,2004','2004-07-16 00:00:00','   Four of Us in Las Vegas    &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; I went with three friends to   Las Vegas   this last week. Each of us had different expectations, and different overall impressions of this place, but we all wanted desperately to have some fun before the realities of second-year medical school set in. John Abraham, Joey Heitzman, and Omar Villanueva are the three guys who traveled with me to Vegas; I came to know each of these guys during the last year, but we really had very little shared experiences. Iâ€™m very glad that this trip has given us something to talk and laugh about when times get tough in the next few months.&nbsp;  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; John, our blackjack expert, is probably the most vocal and extroverted of our group. His energy, humor, and ability to communicate well thought-out ideas make him a very successful conversationalist. Heâ€™s also very adept at studying personality types, for I learned more about the people in our medical school class just by talking with John than from my own knowledge. On many occasions I felt that he carried our group from one Las Vegas encounter to the next. He pretty much figured out most of our social agenda, which worked out well since he had been in Vegas only three months earlier. Iâ€™m not sure why, but John recommended that I listen to Stevie Wonder. I think his suggestion had something to do with helping me to quell my bad mood (at the time).&nbsp;  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Joey, the slots-master, is also very extroverted, but on top of that Iâ€™d say he has a knack for witty conversation and sense of style. Joey has been living right across the hall from me for about a year now, but until this trip I didnâ€™t realize how truly funny he really is. John and Omar bring out this side of Joey; itâ€™s amazing to me how John and Joey have such control over the mood of our group. Joeyâ€™s sense of style is impeccable; he likes very nice, expensive things. I have a very difficult time relating to some of these values, but I try to respect them.  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Omar, who doesnâ€™t gamble, yet still won $30 is a very considerate, perceptive friend. He always senses when something is wrong and wants to help resolve problems. Iâ€™ve spent the most time with Omar in the past year, but recently Iâ€™ve lost touch with him sense he started dating a mutual friend of ours. Omar didnâ€™t particularly like Las Vegas. I originally thought that Vegas had something for everybody, but I guess I was mistaken. Also, Omar does not drink. This personal choice has sometimes come between Omar and myself, but I have to admire his noble ability to find enjoyment without alcohol. &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; As we left Vegas today I think we could all sense that it was time for us to spend some time apart, for five straight days of the same people can wear on a person. However, as I drove home in the silence of my own car I actually started feeling a little lonely. All those fun days and nights had gone by so quickly that I even failed to really recognize the fun I was having. I miss it already. Vegas, though excessively materialistic, commercial, gaudy, and crowded is the place for making and building new friendships based on shared experiences that canâ€™t be replicated anywhere else.&nbsp;   '),(3524,1808929,'07,July,2004','2004-07-07 00:00:00','   The Catcher in the Rye                                                    Somehow I got through high school without ever having read The Catcher in The Rye, by J. D. Salinger. Iâ€™ve heard all about this book through the years; itâ€™s been alluded to in many of my favorite books and movies. Since Iâ€™ve had some spare time in the last few days I picked this book up and decided to give it a try. Now that Iâ€™ve finished it I consider it to be one of the best.   Salingerâ€™s main character is Holden Caulfield who is a very troubled teenager. The reader eventually sees that Holdenâ€™s account of his own problems is unreliable. The most important clues as to the causes of his problems is only briefly mentioned if at all; when one considers these clues and the entirety of his actions one sees that Holden is desperately searching for an intimate connection. Unfortunately, his self-destructive behavior often gets in the way.  There are several great ideas presented in this book the first of which is the human need for intimacy. Holden oftentimes is unaware of his own motivations saying heâ€™s â€œa madmanâ€ whenever he reaches out to others, every time ultimately ending in failure. I feel that this message is particularly important for young men. Our society and most of the world instills in boys a stoic, emotionless quality; this gender socialization, in my mind, really inhibits men and makes them more prone to mental and other health problems as is evident with Holden.   Another great idea from the character Mr. Antolini is the observation that some people, like Holden, are in a terrible kind of fall where they are unable to see or feel themselves fall. Mr. Antoliniâ€™s advice addresses the dramatic irony of the book, namely that the reader realizes Holdenâ€™s terrible fall while Holden himself does not. Salinger suggests that one needs the external environment to help us realize when one has fallen; Holden does not realize his fall because â€œfeelingâ€ and â€œseeingâ€ are both sensory perceptions stimulated by the external environment, of which Holden is detached.   The interesting title of the book represents Salingerâ€™s last great idea, namely that an essential part of life is being there for others. When pressured by his sister to name one thing heâ€™s interested in Holden can only say that he imagines a person catching other people running through the rye field who would otherwise fall over a nearby cliff. Holden finally finds a connection with his younger sister who keeps him from running away. As a medical student this message has been one of my central motivations. No matter how screwed up my own life is I can give my experiences more meaning by helping others.   I highly recommend reading The Catcher in The Rye. We can all relate in some form to Holdenâ€™s â€œlife-at-a-crossroadsâ€ experience. Personally, after having finished my freshman year at medical school I found myself sympathizing with Holden at times, for there were times when I felt that I didnâ€™t give a crap for anyone in my class. Everyone seemed â€œphonyâ€ to me, but unlike Holden I realize that I need to be more tolerant and not as quick-to-judge. While these are my own interpretations at this stage of my life, this novel no doubt has some important messages for everyone at various stages of life.   '),(3525,1808929,'08,August,2004','2004-08-08 00:00:00','   A Letter To Jake                                                Jake,  I just got back from a small party and I just got to remembering how college was. Anytime something was going on in my life I could go to the EMS office and 90% of the time you\'d be up there working your ass off on some project that was due in 12 hours or something. No matter what you had going on as soon as I walked in the door you always made time for me and we\'d have a good time just talking...about anything, it didn\'t matter. I would sit down in those shitty chairs and you\'d let me bitch, bitch, bitch about my life. You also let me in on your life and it was therapeutic like that. Why aren\'t there people like you in my medical school class? I really wish there was, but no, there can only be one Jake Woods.  So I\'m a week into classes now. Yep, we got practically no summer. I did get to go to Vegas for 5 days with some friends though. That was in mid July. Man did I drink. I actually got really smashed the night before school started. Can you imagine me coming into the first day of class with a heinous hangover? Who would have figured? Well I did it and it sucked. That night was fun as hell, but that monday sucked butt. You know one of the biggest reasons why it sucked butt? I discovered that morning that I broke one of my favorite shot glasses. I completely shattered it. I also apparently cut myself accidentally with it in several different places. Seriously, I woke up looking like I got the shit beat out of me but I don\'t remember a thing. That\'s very scary sounding, but don\'t worry, everything\'s fine now.  Jake, I\'ve got a question for you...more like a series of statements. I\'m starting to get concerned about my mental health. Med school is really fucking with my mind. I\'m probably just blowing things out of proportion, but I don\'t know. I\'ve no basis for comparison. I know about bipolar disorder and I don\'t think I\'m bipolar just because I don\'t think my behavior changes eratically, but I do go through these periods of hi-low-hi-low and when I\'m low it\'s like I have a really hard time getting things done. This is probably just normal, but it still sucks. I probably just really need to get laid. It\'s so frustrating sometimes.   Tonight, just before I was leaving for the party I went by friend\'s apartment to get him so that we could carpool. He tells me at the last minute that he\'s not going. It\'s a small thing, no big deal, right? I know that in one part of my brain, but the part I listen to is telling me that I\'m not being studious enough and that I should not be going out to a party. I become angry thinking that he\'s on this bandwagon of new habits and that going to a party with me is a bad habit. Of course these things probably aren\'t true. Later I find out that he\'s trying to avoid someone else...so I had been preoccupied with anger and confusion all night for no reason really.  I don\'t know jack about life. Sometimes I think I pick up on clues and signals here and there, but I\'m really just floating around from one interesting flicker to the next. Does being engaged alleviate some of this craziness? I\'ve got to go now in order to put a stop to my rambling. As I look back on this email it\'s so crazy and disordered. I really should delete it and start over again, but no, I\'m not going to. It is fuckin\' the way it is, right or not-so-right.   Here\'s to your Saturday night and write back anytime. I love email a little too much.  DS   '),(3526,4310425,'22,August,2004','2004-08-22 00:00:00','   I\'ve begun a scary and daunting quest; becoming a Roman Catholic.     Perhaps it doesn\'t sound so scary to you, the reader, but to a simple Southern Baptist girl with an old and much loved aversion to all things Papal, this is little frightening.      Let\'s start at the beginning - the vital stats      My name is Jaime. I\'m a 27 year old woman who lives in Washington state. I\'m originally from Southern California. My father is an excommunicated Catholic and my mother is Baptist. My father divorced my mother when I was just three, hence the excommunication. He would pick up my brother and I for weekends, and sometimes we would go to Mass. I don\'t remember much about it because I was so young. I remember sit, stand, kneel, pray. I remember saying, \"And also with you\", when the priest said, \"Peace be with you.\" I remember my father ushering us out of the church when it was time for communion.      But that\'s not all I remember...      I remember hating it with a passion. Sure, all kids hate going to church, but I mean I REALLY hated it. Why? Because of my father. He was Catholic so that made it bad. Unfortunately, a lot of families split up. Divorce is so common these days that to bemoan the fact that Daddy left is cliche. So forgive me if you\'ve heard this one before.      My father came from a religious Catholic family. His older sister was a nun. So, when he decided to leave my Mom after seven years of marriage, it didn\'t sit well with his family. He did it anyway, and he did it in style. He left my Mom for her brother\'s wife.       Did you catch that?       My mother\'s, brother\'s wife. My aunt. So, not only did he get a DIVORCE, he added a little INCEST to the steaming, stinking, bowl of sin that, by association, got placed in front of me. And I\'ve been choking it down for years.      My father\'s parents, his siblings, everyone on his side of the family sort of forgot that my brother and I exsisted. Sure, when Dad would pick us up for our weekends sometimes he would take us to see our grandparents (when he actually showed up). And they were always perfectly lovely. \"Good to see you, how are you doing, how\'s school, want a cookie?\" But once he dropped us off at home, we didn\'t exsist. No phone calls, no birthday cards, no letters, nothing.      This was incredibly painful, especially after my Mother\'s Mom died. My grandparents on her side were AWESOME! My grandmother was a saint. But, she died when I was only 11. Since I had such a great relationship with her, it really hurt me that my other grandmother chose not to have any contact with me. She didn\'t want to, she didn\'t want me.      Instead of thinking to myself, \'Well, she\'s just a backward, old fashioned meany who would rather believe you don\'t exsist than to admit the fact that her son got a DIVORCE\', I thought, \'If she wasn\'t Catholic she would be nice to me. If she wasn\'t Catholic, I would go over there all the time and she would love me and write to me and call me her grandbaby like my other grandma does.\' That\'s how I rationalized it. It wasn\'t about who she was, it was about what Catholicism was; mean. Catholicism is unforgiving, harsh, and cold. So damn cold.     So why would I want to convert after all these long years of hostility and pain? The answer is simple; because I\'ve been called. And I sure didn\'t see it coming!        The Call      I didn\'t have a vision or a mystical dream. I didn\'t see Our Lady on a rock with flowers at her feet or meet an angel in disguise. It was a gradual conversion. The Lord healed my heart by degrees. I realized one day that I didn\'t hate the Church so much anymore. I flipped through the channels on t.v. and landed on EWTN. I saw Mother Angelica praying the Rosary and thought it was \'kinda\' pretty. I liked how they chanted it.      Then came the desire to go to church. I never belonged to a church but I always believed. So, I got out the phone book and jotted down the addresses of Baptist churches in my area. I attended quite a few.      I dragged my brother (who is an atheist) along with me to a Southern Baptist church just down the street from me. After we sat down in the pew I noticed a copy of their bulletin on the seat next to me. I picked it up and scanned the pages. There was a little article in the bulletin about a member of the church. It mentioned him by name and said basically, since this man, even though he was councelled by the church, decided to go ahead and get a divorce, he was no longer welcome at the church and that all the members should no longer associate with him.     This was a huge thing in my life! I was completely repulsed by this. I wrote an angry note, put it in the comments box, and left before the service began. I strongly believe that what that church did was wrong. Not only was it wrong, it was mean! I thought only Catholics were mean and judgemental like that. It shows how little I knew about the religion that I belonged to.     So, after searching and searching for a place that \'felt right\' to me, I became a member of a LARGE non-denominational church nearby. I \'settled\' for this place because after almost a year of looking for a place that I could belong, none of the churches I visited felt like home. And man was I unhappy! I hated this place. I found I hated this place a lot more than I hated Mass as a child, and that\'s saying a lot. The pastor would stand up on this stage in his shiny shoes and his finely pressed, charcoal gray suit, flashing his gold watch and all the while, preach at me about the wickedness of greed. Baptist services seem to last longer than Mass. Perhaps that was because about half of the sermon would be about how we\'re all supposed to tithe ten percent and how much he needed our money. The whole service seemed to be about money. I would think to myself, \'If God is all powerful, why is he always broke?\'     I didn\'t last very long at that church. I stopped going. It left a bad taste in my mouth. One thing that has always bothered me about being a Baptist is that everybody at church is always so happy. Isn\'t it great that Jesus died on the cross for our sins? Isn\'t it great that he suffered through the scourging, was humiliated in front of hundreds of people and was eventually nailed to a cross so that I can get into heaven? Well, no it isn\'t great. Actually, it kind of sucks. What is everyone so happy about? Why is everyone clapping and shouting and saying Amen and hallelujah? I always wished that people would be...I don\'t know, a little more reverent. A little more somber, less happy and more apologetic that it took his death to save us.      (And quit wearing shorts and sandals to church! What\'s that about? You\'re not at the park. Your not out washing your car or grocery shopping. Your worshiping the One True God and Savior. Put some pants on and have some respect. )     Another thing about Baptists; they all KNOW they\'re going to heaven! They\'re saved and absolutely sure that when the Lord calls, they\'re going to the upper room.      Catholics don\'t know if they\'re going to heaven or not. They don\'t deign to know what God\'s plan for them is. They don\'t know, they HOPE. I like that. It shows respect.       Going to Mass        So, I went to Mass. I don\'t know why, I just did. Something inside me said the people there would understand my need for a little reverance. The people there would understand why I\'m not shouting hallelujah and amen. So, I went to Mass...and it was good.     It kinda comes back to you, you know. Like riding a bike. I remembered most of the responses and it felt natural. When I got home I went online and did a lot of reading. I wasn\'t convinced at this point though. I needed more information. Thank God for the internet!      There are a lot of great websites that had answers to the questions I kept coming back to over and over again. I\'m a liberal so it was really difficult to read some of the things the church had to say about the issues I felt so strongly about. Topics such as abortion, divorce, homosexuality, etc. kept putting \'a hitch in my get-a-long\' as my grandma (the saint) used to say.      It seems like almost overnight my thoughts and attitudes about those subjects changed. I went from being a far left leaning liberal to a far right leaning conservative! How does something like that happen? I certainly didn\'t have any priests whispering in my ear urging me to convert. I think the only one whispering in my ear was God.     Since my beliefs have changed, I\'ve realized something very important about Catholics; they\'re not narrow minded! This knowledge was a revelation to me! All this time I was one of those people who would shudder at the sight of protesters outside an abortion clinic. Now I\'m seriously considering the fact that not only are they right, but that they are on a mission from God.      A Double Standard When One Will Do        When did it go out of style to be a Christian? An even better question: Was it ever in style? Throughout the life of the faith, Christians have been persecuted. Many saints have been made out of martyrs. The patron saint of the Catholic church I attend is St. Stephen, the faith\'s very first martyr. As I progress in my research of the Catholic faith, it occurs to me that it\'s not so easy being a Christian, much less a Catholic. For one thing, it\'s so un-cool.     Growing up as I did in Southern California, I can remember driving past Buddhist temples, Mosques, Hindu temples, Synagogues, etc. You name it, there\'s someone who not only believes it but has the directions to 10 different places where you can go and worship it. Don\'t get me wrong, I\'m not knocking that. That is part of the reason that America is beautiful and we are all so lucky to live here. But it seems to me that if you\'re Buddhist or follow some obscure religion, you\'re laid back, easy going and forward thinking. On the contrary, if you\'re Catholic, you\'re rigid, uptight, and backward. I was one of the people who believed that. In this day and age, if you\'re getting your palm read while a Reiki Master balances your Chi then you\'re cool. If you pray the Rosary and can\'t wait for communion on Sunday, you\'re a crushing bore.     I was one of the people who thought that way. I am so sorry. So no more double standards, not from me anyway.        What Does it All Mean?        I\'m going to travel down this path, even if the way is dark, new, and scary. I\'ll be honest, I\'m afraid. I\'m afraid I\'m not going to be able to hack it. Being Catholic is serious business and I\'m not a serious type of girl. I\'m free and easy, liberal and laid back. I\'ve got tarot cards and when then mood strikes me, I consult my bag of rune stones.      Now I have to consider the possiblity that God doesn\'t like it when I do that. Maybe what I thought was true was a lie. Maybe I don\'t know everything. God help me, maybe I\'m not in control of my own universe and I can\'t go it alone. Being a Catholic is a big commitment and it isn\'t for the weak. I can think of nothing in my life that\'s been more difficult.      What does it mean to be a Catholic? It means, I can\'t have sex until I\'m married. It means if I do have sex, it\'s a big ol\' sin. It means I\'ll have to confess that sin to a total stranger. It means that if I get pregnant I better be prepared to have a kid because I sure can\'t have an abortion. It means loving the Mother as well as the Son. I\'ve never even considered talking to Mary let alone asking her to intercede or pray for me! It means having to police my every thought, word, and action when I have spent my whole life telling it like it is and not worrying about what the big, invisible man in the sky thinks about it. It means changing every single thing about myself.       About This Blog      As I go through the RCIA (Rite for Christian Initiation for Adults) classes, I\'m going to post my thoughts and concerns here. I\'m going to try to be as honest as I can. It is my hope that people who may be considering converting will be able to see the process, step by step, and therefore learn what becoming and being a Catholic entails. I haven\'t run across a site like this yet. I know I looked hard enough for one '),(3527,3986257,'27,July,2004','2004-07-27 00:00:00','   There is glitter where you are A spotlight in the dim hallway And there are people creeping in the shadows that you are not aware of.  You are lost, lost in your own perfection, While I silently wait for you to snap out of your&nbsp;shining world and into the hands of your&nbsp;over- the- edge admirers, Blind to their own jealousy.  And it seems that I can\'t stand here another minute.  So when I run off, Don\'t be alarmed, just be afraid. I\'m not protecting your useless causes anymore.  &nbsp;  &nbsp;  '),(3528,3986257,'25,July,2004','2004-07-25 00:00:00','   what a cool day. all i really did was lay in the grass on the picnic blanket and sleep, but it was still very cool.  i just about had the healthiest dinner ever...my mom is all macrobiotic, so tonight we had green beans in spices, quinoa, and stir fried vegetables...zuchini and carrots, to be exact. all organic.   &nbsp;  '),(3529,3986257,'24,July,2004','2004-07-24 00:00:00','   went to bed at midnight...so of course i didnt go jogging since i slept in. oh well. maybe tomorrow, haha.  anyway, i want to get another rachael sage cd, its only 12 dollars or something so im seriously considering it...maybe i\'ll wait until next weeks paycheck. in the meantime, i need to use up my giftcard at old navy. i cant get any pants there, coz they\'re all 30 dollars and i have 28 on my card...figures, right? and thats really all i need - jeans. i mean i haev jeans, but they\'re so annoying. jeans are like that - you can never find a perfect pair. one fits nice but it shows thighs, one gets loose after a while, one is too low - they fit perfect when you try them on at the store and then you take them home, and its like, \"crap.\"  mom asys i sholdnt get jeans, i shuold get khakis, but then you can only wear them at certain times. well not neccesarily. but i would.  another thing i need to decide - a picture day shirt. i dont have a lot of \"nice\" shirts, mainly ones that say something on it. i love the photographers in GCHS - they make you look really good in pictures. seriously. everyone looked pretty okay in their pictures.  i hope my aunt leaves today. not that i dont love her. but we just kicked the last houseguests out of the house, and now there are 2 more. its like, god, could we live in peace?  my grandma got me the prettiest bracelet from san francisco. she also got me a shirt - its super girly. why do ppl think i wear girly things? also one eyar she wrapped my preent in barbie wrapping paper. i didnt asy anything, but, its like, \"hi, im 15, i really dont play barbies anymore. they\'re all skinny blondes with big boobs.\" but i am grateful that she at least thinks of me.   not doing too much today - the usual practice and playing with my rats, and im going to start typing up my lesson.   hasta la pasta! gina  '),(3530,3986257,'24,July,2004','2004-07-24 00:00:00','   went to bed at midnight...so of course i didnt go jogging since i slept in. oh well. maybe tomorrow, haha.  anyway, i want to get another rachael sage cd, its only 12 dollars or something so im seriously considering it...maybe i\'ll wait until next weeks paycheck. in the meantime, i need to use up my giftcard at old navy. i cant get any pants there, coz they\'re all 30 dollars and i have 28 on my card...figures, right? and thats really all i need - jeans. i mean i haev jeans, but they\'re so annoying. jeans are like that - you can never find a perfect pair. one fits nice but it shows thighs, one gets loose after a while, one is too low - they fit perfect when you try them on at the store and then you take them home, and its like, \"crap.\"  mom asys i sholdnt get jeans, i shuold get khakis, but then you can only wear them at certain times. well not neccesarily. but i would.  another thing i need to decide - a picture day shirt. i dont have a lot of \"nice\" shirts, mainly ones that say something on it. i love the photographers in GCHS - they make you look really good in pictures. seriously. everyone looked pretty okay in their pictures.  i hope my aunt leaves today. not that i dont love her. but we just kicked the last houseguests out of the house, and now there are 2 more. its like, god, could we live in peace?  my grandma got me the prettiest bracelet from san francisco. she also got me a shirt - its super girly. why do ppl think i wear girly things? also one eyar she wrapped my preent in barbie wrapping paper. i didnt asy anything, but, its like, \"hi, im 15, i really dont play barbies anymore. they\'re all skinny blondes with big boobs.\" but i am grateful that she at least thinks of me.   not doing too much today - the usual practice and playing with my rats, and im going to start typing up my lesson.   hasta la pasta! gina  '),(3531,3986257,'24,July,2004','2004-07-24 00:00:00','   Its so pitiful. they say grandpa is deteriorating. deteriorating. dying, they mean. its just another word to take place of the word they dont want to say. god im scared. i know that sounds so stupid, but i hate grief. it just sneaks into your house and pounces upon you when you least expect it. and then it drips onto the walls and everyones bodies and takes them over. and the world seems to turn gray and cold, even when the sun is shining beautifully. it makes me sick to think of men cry, to think of all those sad crying faces. hes not dead. no, not yet. i have hope, but its so small - hes always falling down, a nurse visits him twice a week, they say that just this week hes been looking so much worse. i dont know what to do.  they say the goddess Kali teaches us to accept death and to rejoice life, its all part of the circle. but i hate grief. do i hate the fact that im going to deal with grief, or the fact that grandpa will not be around anymore? and what about grandma? will she die of a broken heart? or will she age and will her core be nothing but mechaniacal? i want answers, but there is no way to get them, and i am scared shitless.  he could die today, tomorrow, a week later, a day when i\'e gone out with my friends and have had a good time. its so sudden, so quick, like snapping fingers or blinking eyes. im so worried. we were never close. then again, uncle g and i werent close either. so then why do i are so much? why do i care, if i never cared before? im so spoiled rotten, to not care until the last minute - you dont know how precious things are until you know you\'re going to lose them.  i know i\'ll have to accept grief gratefully. but whats so grateful about tears and tissues, long, sobbed hugs and \"i\'m so sorry\"\'s? whats so grateful about going to mass, a church, a deity i dont worship and signing sad, mournful songs? i like to cry alone, and all this, its with other people. i cant seek comfort from other people, when their comfort might be fake. i cant bear to see my father crying all over again. i cant bear to see everyone in so much misery.   sometimes when you feel lonely and angry with yourself you say, \"god if i left this world no one would care.\" but they would, even the ones that thought they wouldnt care really would.   ugh. im going to go watch that comedy to cheer myself up. love, gina  '),(3532,3986257,'23,July,2004','2004-07-23 00:00:00','   &nbsp;  ...Among all of Gods creatures It is only men who seize Upon frienship and devotion As if they were a disease To be cured and violated Whitewashed and despised Never wondering why the heart is growing black and oversized...  &nbsp;  '),(3533,3986257,'23,July,2004','2004-07-23 00:00:00','   SIMPLE QUESTIONS: 1. Full name: Genevieve Ines Salazar   2. Nicknames: Gina, GG, Genie  3. what do you add to get three? 1 and 2...   4. Height: 5\' 6 1/2\"  &nbsp;5. Hair: dark brown  6. Siblings: 1 sis 1 bro  7  . a) Do u like to sing in the shower? Sometimes  b) Do u like to sing on the toilet? er, no.  8. Do u like to sing? yeeep  9. Birthday: may 25 1989  10. Sign: gemini  11. Address: its okay, really  12. Sex: female  13. Righty or lefty: lefty!!!  14. What do you want in a relationship most? good sex. sorry, im kidding. i just had to. trust, honesty, funniness, and the willingness to be open with each other  15. Have you ever cheated?&nbsp;in Sims, yeah...lol  Marital Status: Single  17. Do you have a car? I wish i did   18. What kinda car do you have? i wish  FAVOURITE QUESTIONS:  19. Movie? Chicago, Moulin Rouge, Bridget Jones\'s Diary, Love Actually, Johnny English...  20. song: either Diving to be Deeper by Sinead Lohan or Bravedancing by Rachael Sage  21.  word: fickle  22. place to go: Spain!  23. animal: rat or dove  22. Actor: hmm.....i dotn know his name. sorry.  25. Food: mexican or italian  26. Number: 6  27. Cartoon: bugs bunny!!!  28. Disney Character: dunno  29. Colour: blue, hazy green  LOVE LIFE ETC:  &nbsp;30. Do you plan on having kids? yep! not sure if i want to go through the pain though...adopting sounds good to me right now  31. do you know what love is? No  32. what is love: something beyond my comprehension   33. How old do you wanna be when you are married? 23   34. Would you have kids before marriage: no   35. Do u have a b/f or g/f: No, i dont need a guy at the moment - im pretty settled in   36. where do babys come from: the stork.  37. Do you have a crush? i dont think so. maybe, yes, no.   EITHER-OR (PICK WHICH ONE YOU PREFER)  38. Music/TV: music  39. Guys/Girls: guys  40. Green/Blue: both  41. Pink/Purple: dark purple  42. Summer/Winter: Summer  43. Night/Day: both  44. Hangin Out/Chillin: theres a difference?  45. Dopey/Funny: Funny  46. You know I\'m around when you hear: my laughter...lol   47. What school do u go to? GCHS. god bless it. coz it sure do need some blessing.   48. how many fingers is a fingerless person holding up: None   49. What\'s a major turn on for you? dark eyes, dark hair, and mature for their age   50. How far would you go on a first date?: not far at all   FRIENDS:   51. MosT blonde: dunno....  52. Nicest: Well all my friends are nice  53. FuNnIeSt: *shrug* thy all have their \"moments\"  54. TaLLeSt: you know, i think im taller than almost all my friends....   55. BeSt pErSoNaLiTy: All my friends have a good personality   56. Which people do you trust and are open with the most?: Megan and Shikha. and a couple others   57. What do you think of soul mates? they exist!!!  58. Is it right to flirt if you have a bf/gf? if its all fun n games, sure why not?   59. What was the last thing you cried over or got teary about? i think my grandpa...   60. What\'s something about guys/girls you don\'t get: Girls, how they can be so sickeningly flirtatious at times it makes you want to puke, and guys how attatched they are to their little gadgets and how egotistic they can be   61. What\'s the objects you can\'t live without?: my books and my \"love\" necklace   62. Did you ever get a pony for christmas? i never wanted a pony for christmas. they smell and they can stomp on you and kill you   63. Love or Lust: Love  64. Silver or gold: Silver  65. Diamond or pearl: pearl - it came from the bottom of the sea   63. Sunset or sunrise: both  64. Have you ever gone skinny-dipping?: Nope itd be cool to try though...  65. Do u sleep with stuffed animals: no, but i seep with a special pillow lol   66. Do you have any piercings: my ears   67. What colour underpants are you wearing right now: vanilla knickers lol. dont ya love the english?  68. What song are u listening to right now? none  69: Whats the last 4 digits of your home phone:&nbsp;i\'d rather not say   Where would you want to go on your honeymoon? france!   71. Who do you want to spend the rest of your life with? someone tall dark n handsome   72. What\'s the first thing you notice about the opposite sex: eyes  73. What sports do u play? hah!  74. What makes you happy? taking a walk, getting my chores done and over with, hanging out w/ friends  75. What\'s the next cd/s you\'re gonna get? rachael sage!  76. Do u wear contacts or glasses? neither.  What\'s the best advice given to you? love yourself before you can love others  . Have you ever won any special awards? Yep  79. What are your goals? To become wise, old, and happy with no regrets   81. Do you like Funny or Scary movies better? funny!   82. On the phone or in person? person   83. Hugs or kisses? Both all the way  84. What song(s) seems to reflect you the most? Child by Rachael Sage   85. If you died tomorrow who would you leave everything you own to?&nbsp; family and friends   86. Do u have any enemies? Not that i know of, hopefully I dont  87. if you could do anything you wanted with no consoquenced what would you do?&nbsp;steal from the rich and give to the poor   Would you rather be rich or famous? neither  89. What time is it in Albania now? WHO THE HELL CARES?   90. Have you ever been in love? nope   91. Have you met santa? no  92. Favorite Band: none  93. When did you last talk to the person that you like? i dont really like anyone.   95. when do you go to bed: 10 or 11   96. Last time you were depressed: summer of 6th and 7th grade  97. Are you an alcoholic? psh no  98. Who sent this to you? got it from navin  99. What do you think of this person: hes cool and very backwards  100. Do u want your friends to write back? whatever  '),(3534,3986257,'22,July,2004','2004-07-22 00:00:00','   Well today was all right - we went to On the Border to eat (yummmmm) and then took a walk by this one river. it was really peaceful, and we saw this huge white bird - i forget what they\'re called. and then we went home to eat cake, and we found out TS, who had not been w/ us and was supposed to go to Country Thunder, had been \"stood up\" by her friends.  she bought a ticket to the concert and was sure her friends were going to go with her, and they had bought tickets also, but at last minute they couldnt go with her, and they had used TS\'s mula to buy the tickets, so they owe her a chunk. her friend is a bit light headed and \"full of butterflies\" and never plans anything. so TS was in a tizzy and crying - she had just wasted about 300 bucks. maybe more. but she said it wasnt really about the money, but it was about how her friends had stood her up when they had been planning this for weeks.  mom said that she learned that lesson early - friends always stand you up. And you know, its true. friends do stand you up. they make these big gorgeous plans, and at last minute, \"nope sorry cant do that.\" its already happened to me. its a part of life, i guess, disapointment - and knowing that your friends arent perfect and that you will have to make sacrifices for them. its sad, but i guess thats the way it goes.   but, i dont like the philosophy that not having friends makes you a happy person. it doesnt, not really. you\'re actually more depressed when you dont have friends. family is all right, but when you need to talk about sex, about drugs, about the things you dont want to say to your family, who do you turn to? friends. and some friends are like vampires and they suck your blood to make themselves look cool and popular, and those friends turn out to be not friends at all. some friends are like dummies and ditzes&nbsp;and all they\'ll talk about is guys and make up.   im not saying that if you and your friends talk about guys and make up its stupid, but 24/7 yakking about the same subject can get pretty dull.  and then there are those great friends who will sacrifice a lot for you. like once, i was really, really pissed, and sad, and angry about nothing. it was just one of those days. and i told my friend, and we went on the bus, and i was like, \"god, i just feel like killing someone, like screaming.\" so, she had a crush on the bus back then. it was her fav time of day, to be near him. and she pulled me off that bus so we could walk home and she could hear me scream. and i screamed, and we talked, and i was so grateful to her. i know what its like, when you have a crush and something comes up so you cant be near them. but it made me realize that having a&nbsp;relationship is a tough thing - you sacrifice, you cry, you get mad at each other, you lie, etc.   but the funny, funny thing is, is thats its worth it. its worth the blood and the tears and the joy. and if you have had a friend that makes you work for those blood and tears and joy, you know that they\'re worth it.   ok, end of my long rambling.  love, gina  '),(3535,3986257,'22,July,2004','2004-07-22 00:00:00','   woohoo, i got my paycheck today! 60 dollars! except im being&nbsp;logical and putting half in the bank, so 30 dollars! this is so cool! today is a very good day.  Anyway, so today i\'m up to a lot, i forget what, but its listed on my calendar, and i was filling today in last night and i had about 50 things for me to do...  well, gtg, love, gina  '),(3536,3986257,'22,July,2004','2004-07-22 00:00:00','   She went looking for love and found her reflection,  She sought crystal clear&nbsp;and the utmost perfection,  Her life was in tears, her body in pain,  She wanted a sweetheart to make sure she was sane,  So she went on a walk to contemplate this,  And saw a white flower, which she gave a kiss,  And she glanced at the water, so light and so fresh,  And wondered how she ever got into this mess.  She stopped as she saw her eyes gazing back,  So ruined were they, they almost looked black,  She missed the warm days of sunshine and singing,  She missed those sweet&nbsp;days when her ears weren\'t ringing.   She splashed at the water and cried in the creek,  She wanted some strength,&nbsp;and she seemed&nbsp;so weak,  And the creek bubbled back, and it talked and it soothed,  And the girl was astonished,&nbsp; amazed, she was moved.   \"Listen,\" it said, \"listen, come here,  Why do you waste these cold&nbsp;bitter&nbsp;tears?  Can\'t you see that each day is your own?  Can\'t you see that you\'re free to roam?  Why do you weep, girl,&nbsp;why do you cry?  Don\'t you see your wings, that you can fly?  Don\'t let them strangle you, girl, you can fight,  And, most important, don\'t let your goals out of sight.\"   To this speech she listened with care and control,  And inside she felt a strange, little pull,  She wanted to love, she wanted to live,  And instead of taking, she wanted to give.  She laughed til she cried and cried til she smiled,  Her whole thoughtful walk had been well worth the while,  And she bent toward the creek and gave it good thanks,  She kissed its waters and its green, glowing banks.  She glanced at the water, so light and so fresh,  And wondered how she could get out of this mess,  She stopped as she saw her eyes gazing back,  So full of mirth, of joy, and so black.   She left the waters, but her soul was still there,  No longer was she in deep, doomed despair,  For she bubbled like the creek, and was in love with each day,  And she became an old wise woman, a&nbsp; healer, a sage.  So when her patients are in pain and in tears,  She tells them to&nbsp;touch their deepest,&nbsp;dark fears,  And go to the creeks waters and look for perfection,  And they usually find that it is in their reflection.  '),(3537,3986257,'21,July,2004','2004-07-21 00:00:00','   i actually jogged (bows before cheering audience) and then showered and fed my rats. the usual. went to work, i made the copier machine freeze twice, went home a half hour early, CDB came over, shes so cute, has a country accent now.   i almost finished the satchet, but i ground up all the herbs. i weeded the garden and read a book (or two!) and made dinner for my mom. egg salad. lol, during the preparation of it i spilled some mayonaise on my grandmas dog (oops...) and the floor, and my shirt and shorts.   and now im here. tomorrow im not going to work, its dads bday, so we\'re all gonna go out somewhere. Tia is coming over for a night on Saturday, and Yaya and Yayo and abuela are coming too. oh and Amir, the cutest one! luis hates Amir, i think its b/c he follows him around everywhere. but AS hates luis coz he follows him everywhere.  its a young role model type o\' thing.  i made a smoothie too! it was so good, i cant really remember the ingredients though, i just sorta tossed stuff in as i went along.   TS went to a concert w/ her friends. and i\'ve got to clean my room. i think im going to do a ritual on the full moon - it just seems like a special occasion.   note - ask MR if she\'ll mind if i turn in my packet late -make another date for lotus moon. -tell MR&nbsp; about help wanted job  love gina  '),(3538,3986257,'21,July,2004','2004-07-21 00:00:00','   I had FUN tonight!!! weeheeee!&nbsp; i swear, there&nbsp;is nothing better than a cute guy teaching you how to defend yourself. \'nuff said.   to do list:  -play with rats, trust train them -2 meditations  tomorrow -play with rats -do 1 ritual -GO JOGGING -clean room!!! -dry herbs completely, mortar and pestle them, put on altar  friday -do 2 rituals -play with rats -put cleansed herbs in satchet -place on altar -record dreams -mini workout  im hungryyy. well, nighty night. got stuffs to do.  '),(3539,3986257,'20,July,2004','2004-07-20 00:00:00','   NOT! hehe, j/k. kinda. work sucked today, cause i had to stare at the computer fo a half hour, and its got a blurry screen...ech. later today im going to make and create that satchet i promised MR, and theres self defense class, and i need to do some stuff for mi madre. oh and my lessons.   dads bday is this weekend - i didnt get him anything (i dont have any money!....yet) so i feel bad. but i got him something for fathers day.   yawnyawnyawn. i kinda want to go back to school. kinda dont. some people are really stupid there, ugh, they just get on my nerves. but i guess you get that with every school, huh?   mab is eating her medicine, she looves it! poor sweetie, hope she pulls through - i dont want her or breena to die!   speaking of death, wonder how grandpa is doing? i know, im morbid. but lets face it - hes not going to live for very long. hes yellow, his hair can hardly stand on its own, he sleeps all night and 5-8 hours of the day...   k, im gonna go do something else. hasta luego!  love always,  gina   ps - hi navin!!  '),(3540,3986257,'20,July,2004','2004-07-20 00:00:00','   I know, I already have an lj, but i figured i\'d just use this for a bit. things are going okay - as in nothing dramatic is happening. my god i went swimming today, my eyes were all red. i know i should wear goggles, but i feel so stupid wearing them....i know, i know, safety first, then looks, right? right.     &nbsp;  i guess im sort of waiting up for my sister, dont know why though. shes so bossy, gina do this, gina do that and gina you\'re not doing it right.     &nbsp;  work tomorrow. not very excited. and im going jogging. im so lazy, i have to will myself to get up to go jogging, and then L is like, \"gina, go, its good for you, its an adventure!\"    &nbsp;  i\'ve been seeing  him  a lot. ugh. gooo awaaayyyy. why\'d i like him again?    &nbsp;  oh, yeah. he was nice. funny, i dont see that characteristic in him nowadays.    &nbsp;  moving on, im excited. not sure about what. you know that feeling that says, \"something cool will happen...\" yeah, thats the one.    &nbsp;  *sigh* well, not much more to talk about. i watched The Sound of Music the other day. its so interesting, how a nun can just give it all up for a guy. but then again, does she erally give it up? its sad that they dont put more focus on the spiritual, just the romance.    &nbsp;  i think too much, dont i? lol    &nbsp;  love always,   gina    &nbsp;   &nbsp;   &nbsp;   &nbsp;  '),(3541,3986257,'14,August,2004','2004-08-14 00:00:00','   Stiff as a board You lay there, unaware Sleeping soundly but not snoring And I kind of miss that about you.  Flowers seem to poke out of nowhere In their glory presented to you But do you notice, do you care? Because all I wish is that you were here.  You were always a piece of string hanging in the air Not sure where to go or what the hell to do But you always bowed and smiled so humbly And my hearts mind is flooded with the grief Of having to finally let go of you.  I hope you\'re happy when no one else is You deserve it, you really do. And I hope you got all that you asked for up there And tell God I\'m waiting, too.  So you\'re stiff as ice up on that altar But you see, you\'re slowly melting... You\'re going into that place where we all dream We could go without pain.  '),(3542,3986257,'12,August,2004','2004-08-12 00:00:00','   -2 pairs of pants - one jean, one nice: \"dress pants\"  -lipstick. color: nude  -warmer shirts  -school supplies - no binders this year please. they are soo annoying  '),(3543,3986257,'12,August,2004','2004-08-12 00:00:00','   damn all onions to hellllll! arghhh i was making dinner and my sis was chopping the onions....theres a bit of a tear in my eye lol.  so zee parents are gone, im going to grandmas house tomorrow, she called me today to give me a 3 minute lecture on how i shouldnt open the door if someone knocks.  anyway. my brother is bugging me up the wall and my compass proves right. hes a total spoiled brat.  good buh bye  gg  '),(3544,3986257,'12,August,2004','2004-08-12 00:00:00','   you know, it just figures. im having the worst month of my life. every movie i see is making me cry, i dont want to do anything except be alone, and i want to stick my nose in a bunch of novels and forget my life here ever existed. im sick of my sister screaming and clawing at my back, im tired of looking at my moms tired face, and i\'ve lost the thrill for self defense. actually i\'ve lost the thrill for even writing. i cant write anymore, nothing works. i feel like throwing up in the face of this gray weather and if one more thing goes wrong i think i might crack.  i thought i had thought everything through the other night, that being miserable wasnt worth anything. i thought i had cried the last time. but god i am so pathetic, i was watching the end of Casper the Friendly Ghost and had to excuse myself to go sob quietly. everything jut really sucks. i dont want to go to self defense tonight or ever again, i dont want to go to my grandma\'s house to dwell in her past and grandpa\'s death, and i dont want to pick up my violin anymore. dad noticed and even suggested some welbutrin the other day. im not sure what everything im dealing with has to do with why im sad, depressed, angry. i really shouldnt be depressed. i dont have much to be depressed about, do i?  but i will go to self defense, b/c if i miss one more class they\'ll start thinking im a wuss. they\'re men there, i need to show them women are just as good as men. and i will go to grandma\'s house, because she needs me to comfort her. and i will pick up the violin and start writing again, and i will be happy.  just not now.  '),(3545,3986257,'10,August,2004','2004-08-10 00:00:00','   i am pissed. i want to get back to my life before all the shit of the blue moon happened. turns out this friday until next friday i am staying at my grandma\'s house with my brother, and returning home 3 days before school starts. where did summer go??  the thing is, i want to go school shopping, and make plans with a couple friends, and do things, but i dont know when. its like august 25 is a deadline or something. im just in a really sucky mood - i saw beauty and the beast today to cheer myself up, and i ended up crying. and you know, why cant i stay alone in the house just for a couple days? i dont want to be in grandmas house overnight - last year it was just for 3 days and i lost my head.  everythings just a fucking disaster  '),(3546,3986257,'09,August,2004','2004-08-09 00:00:00','   biggg day. i didnt go to self defense (again) so i feel guilty.  heres what happened:  4 kids. 5 1/2 hours.  and actually they\'re good kids. the only one that really needed watching was my own brother.  then around 11 or so my cousin came, so i had to watch over him too. but hes cute and sweet tempered.  i was happy - they all really liked me, i had made a good impression as a baby sitter.  but now im really tired.  i went to the pool instead of self defense, and it felt so good, just to swim. my brother said i swam like a mermaid! i was really happy about that compliment.  hey i just recieved news someone is interested in our religious group! yes!  i am happy.  love, gina  '),(3547,3986257,'08,August,2004','2004-08-08 00:00:00','   Another uneventful day in our family....hah! fooled you for a sec, didnt i? nope, today we went fishing, and of course caught nothing, but dad dropped his pole in the water and got his line caught on a sign that said \"fox lake\"...   then we ate out for lunch, went home, and went to jewel to get fish for dinner, so that at least we could pretend we had caught something.  tomorrow im going to be looking after 4 kids from 9:30-3. someone save me, i think i might lose my head.  actually, it wont be that bad - most kids like me. i mean, look at my cousins (the younger ones) every time i walk in the room they pounce on me and demand a spooky story or a story about prince charming and a damsel in distress.  heres my schedule:  9:30-10 = painting, crafts, drawing 10-12- outside games, tag, waterguns, set up the sprinkler, go to the park 12-12:30- eat lunch 12:30-1:30= quiet time, reading (the girls are archie comic book fans!), teaching spanish? watch movie until parents come? 2-3- chalk, maybe petting the rats, anything else i can think of from the top of my head  lets see how tomorrow goes. also, i\'ve got self defense alone tomorrow  since ts is hanging out with friends.  fuunnnn. did i tell you after my dad treated me with yet another medical miracle, my canker sore hurt even more???  oh, plus...you\'ll never guess that im spending 7-9 days with one of my grandma\'s, since mom and dad are going on their twentieth anniversary, ts will be moving back into her apartment, and the dreaded brother will be with  me.  i get back 3 days before school starts. some way to end my summer.  i cant say i\'ve had a bad summer, but it wasnt really the best, was it? too much death and pain and grief. that and all this stress is really getting to me.  but i\'ve been away from the world for a while, and its time to go back. i\'ve been hiding out in my room, crying, thinking, sleeping, being sick. i need to get into the routine of excercising and being friendly.  its only 7 and i feel dead tired. and i wake up tired, and i go through my day tired. so i just need to break through. i feel like such a disgusting slug lately. i dont care about how i look and i shower twice a day to feel refreshed. that and i\'ve only had a million canker sores.  i get at least $25 tomorrow. im going to treat myself to, if i can get a hold of her, a movie w/ my friend, lots of new tshirts, and maybe a CD.  whats 25 x 4?  my god that was a stupid question.  100 dollars. wooowwwwwww. technically, 130.  wooowwwww.  i loooove baysitting!!!  gina  '),(3548,3986257,'07,August,2004','2004-08-07 00:00:00','   I hate    glancing  into     your    kaliedoscope    eyes     Because  each time I do I wonder what form I take from your view.   Rainbow shards of glass seem to be reflecting off my eyelashes,   And you twist and blink until I become what you can take over.  '),(3549,3986257,'06,August,2004','2004-08-06 00:00:00','   wonderful day. went to the beach w/ mr, it was so pretty, i just love hearing the waves on sand.  i want to get 2 rachael sage CD\'s, Illusions Carnival and Smashing the Serene. god i love all the names of her songs and CD\'s. you can relate to them in the deepest way.  anyway, love all  gina  '),(3550,3986257,'05,August,2004','2004-08-05 00:00:00','   well....i was watching hidalgo last night and we didnt get very far into the movie before i went upstairs and threw up. twice. yuck yuck yuck. thankfully mom says its only a 24 hour bug and i should be fine y the time me and mr go mini golfing. but i slept in due to the throwing up thing so i didnt do my ritual. ugh, it seems like every time i want to perform it something gets in the way. first grandpa\'s death, the fairgrounds noise, my family, the canker sore, now its this stomach virus. im going to go insane. i wonder whats next??  i didnt eat at all yesterday except for some soup, tea, and a glass of milk, so i checked my weight this morning and i\'ve lost a couple pounds.  i dont feel like eating anything now, either. but i gotta brush my teeth. my breath probably stinks like mad.  gina  '),(3551,3986257,'05,August,2004','2004-08-05 00:00:00','   i woke up this morning feeling nacious. i showered, then got back in bed, fell asleep, and woke up at 1 in the afternoon.  i dont feel so good, but i dont want to get sick now - i want to go mini golfing with mr tomorrow and do my ritual at 4 am. yes i know its an ungodly hour - right after i do it im planning to go right back to sleep.  so, it seems as if i have a babysitting job all next week! i\'ll be sitting my brother and his friend from 9-3 mon-thurs. but i think i\'ll have fun with it, i love kids. most of the time anyway. but my brothers friend is a good kid, not rambunctious or noisy like him, so it\'ll be okay. i want every day with them to be special, so i want to make days like \"movie day\" and \"painting day\" and maybe \"water day.\" this way maybe zs will like me, and then his mom will want me to sit him more, and then i will earn a friend and some cash.  im still very much a kid at heart, so jumping over sprinklers and water gun fights and games of freeze tags are sooo much fun. especially if you have 5 or 10 ppl, then its great...  well, ta ta, off to watch a movie  gina  '),(3552,3986257,'04,August,2004','2004-08-04 00:00:00','   it was one of those days where...you wake up and think, \"god, it is going to be such a crap day....\" and then just to trick you, god decides to trip the opposite upon reality.  i went to grandma\'s house, hung out with my cousins. we watched some movies....some new ones, some old, funny ones. then we started playing with waterguns toward night. and me and cs were trying to gain up on ds, it was great....oh, did i mention my aunt is distantly related to bill nye the science guy? how cool is that?!  i\'ll miss them so much, but i know they\'ll come back. i want another family reunion. with everyone. its so much fun, just to sit there among conversations and know that you\'re safe and everyone in the room loves you. its truly a blessing that i sometimes take for granted.  no, i did not go to self defense. ts says B gave a talk about how we should devote more time to the art of self defense, but really....my grandpa died, im spending time with my cousins, and im figuring myself and my surroundings out for the first time in my life. excuse me for breathing.  then again, ts is stupid b/c she didnt tell them our grandpa died, she just sad i was sick. im going to tell them the truth. she says she doesnt want sympathy - i dont think theres anything wrong with it. she thinks shes so tough, so like a guy, so much armor. but i see her, shes as weak as the rest of us. i love her though.  i dont want to take life and the things i do and love for granted anymore. people, they take out their cell phones, their gadgets, their little shiny toys and flash them around. they dont want to be near mommy and daddy, dont wanna be close to the things they know they love most.  i had a dream last night...i saw a statue of a girl. she seemed to be lying on the grass, holding something, just sitting there, wishing and dreaming. yet she was living as much as anyone else, she wasnt wasting her time. and i picked up the statue and i wanted to buy it, but it didnt have a price tag. i promised myself no matter how much it cost i would buy it because it represented me, with my back on the grass staring up at the clouds. then i woke up.  i want to paint it now. the statue.  k, im tired. sleepy time.  love gina  '),(3553,3986257,'03,August,2004','2004-08-03 00:00:00','   such a good day today, like being born after so much death and pain. me and my cousins walked all over chicago, and the sun and earth reminded me that things are and can be reborn, even after the most defeating struggle. and i sweat and was strong and calm. feels like a tree sometimes, ya know?  but if im a tree, im one heck of a young tree. got so much to learn, to do, to finish and tie promises. and i can wait for my day of rest, because i feel so full of life.  and i will be honest, you wont hear me say this in winter, lol. in winter i\'ll be muttering curses about the snow.  love gina  '),(3554,3986257,'02,August,2004','2004-08-02 00:00:00','   i am still the puffy cheeked chipmunk - dads surgic thingy didnt really help. well, yes it did. but now i feel more canker sores coming on, and am having a ball trying to overcome the desire to chew.  mom came home and said, \"gina, i think your other cheek is looking a bit swollen...\" and of course, its not, shes just imagining things *hopes*  besides being a puffy cheeked chipmunk i am also rudolph the red nosed reindeer, with that zit on my nose. front and center. im going to have a simply lovely freshman year.  to top it off with a sour cherry, we\'ve got self defense tonight, and not only will i attract uwanted attention, i\'ll be missing a family barbecue.  im such a whiner. but at this point in time i\'ve really come close to saying, \"why me, god??!\" its not fun. its just one thing after the other, and maybe before i go to sleep tonight the house will burn down. or someone else will die. or i\'ll be eaten alive by grandma\'s next door neighbors dogs.  the possibilities are endless.  yaya is okay - shes got major bruises, but no broken ribs or anything, thirteen stitches, one cut really close to her jugular. she fell down the stairs and crashed headlong into a glass vase. she broke it with her head.  i am really, really, really stressed out. think of stressed out as stretching, and imagine putty stretched out all the way until it breaks. thats me.  i think sophomore year may be the worst year of my high school career. but, hey. bring it on, right?  right.  *rolls eyes* gina  '),(3555,3986257,'01,August,2004','2004-08-01 00:00:00','   this is just one of those times where optimism realllyyyy helps. heres whats going on.  you know that lovely canker sore?  it has puffed up my right cheek.  i officially resemble a chipmunk.  so im going into the office today to have it burned off. *starts praying profusely* dad promises that i\'ll still be able to eat afterwards, but seeing as i cant even eat now (i can barely open my mouth, and when the food does get inside, it hurts like hell on the open sores) i dont have much hope.  this isnt the first time i\'ve had inflammation due to a canker sore - last time it was on my lip. my parents were away and i had to go to orchestra, my sister was going to drive me, but i stayed locked up in my room and she decided to leave without me, which is what i had planned. i think i like a swollen cheek better than the lip.  i woke up at 4 am last night and was soo thirsty. i\'ve been thirsty ever since i\'ve had this sore. so i went downstairs, drank first regular coke, then det coke, then chugged down water. this morning after attempting and then failing to eat my cereal i drank about 2 cups of milk.  guess what? im still thirsty, lol.  its not so bad, if i dont talk or eat or smile it doesnt hurt. and hey, it\'ll be over in a couple days. i wanted to have fun at the fair (which i kinda didnt) and i wanted to have fun with my cousins, who are here until thursday, but my cousins can always visit and the fair will be here next year. and my cousins were really nice about it, i felt bad b/c i never felt like talking and i think they thought i was being rude and ignoring them.  i\'ll try making it up to them today.  so, my 2 activities left in the Isis lesson are postponed until all this is over. there will be more blue moons, right?  oh yeah, and guess what? yaya fell down the stairs and cracked her head open. its just thing after thing this summer, isnt it? but i always thought that her stairs were really steep, even as a child i didnt dare walk down. its a wonder she hasnt allen down until today. dad says she\'ll be fine. i sure hope so. damn stairs. im thinking of getting a one floor house - this isnt the first time one of my relatives have fallen down the stairs.  love, gina  '),(3556,3986257,'01,August,2004','2004-08-01 00:00:00','   i am in the pits.  i guess all the stress from grandpa dying and the funeral and everything has made a big impact on me, because now i have a cold sore. i havent gotten a cold sore since last halloween. it always happens with the big events. i was at the lunch after the funeral in the restraunt when i felt the \"symptoms\" of a cold sore, but of course i couldnt rush home and take some valtrex, which usually pushes it away before the cold sore arrives. by the time i got home, i officially had a cold sore. this was supposed to be a fun day. me and sg are going to the fair today. now i wont be able to properly enjoy it.  and im getting really sick of my sister. she always has to find time to correct me in everything i do. like, i dont mind if she says, \"your violin is out of tune, you should do this and this to make it better\" or \"that painting would be better with a touch of bule or green.\" but every single little thing i do makes her mad and shes ALWAYS correcting me.  \"you drink soda weird\" \"thats not how you\'re supposed to use your spoon/fork/knife\" \"nothing at all really goes on inside your head\" \"you use the toothpaste wrong\"  and on and on. and its with things i say, too. i try and make a joke, shes all \"that wasn\'t funny\" or \"that didnt make sense\" or something like that. she gets that part from mom - mom doesnt laugh a lot. i showed her the bday card i was going to give sg and she read it andgave me a tight-lipped half-smile. when i read the card myself i burst out giggling. people these days. sheesh.  and then half the time tasha acts like a 3 year old and the other half she tries to be cool and be in charge of everybody.  last night, she was all, \"mommy....daddy....i dont wanna go to work on monday...\" in this sappy childish voice. shes 19. give me a break.  \"im tired...\" \"i dont wanna go to grandmas house, i have a headache...\"  she always has a headache. she gets migraines (rarely) but in her life she has a headache every day. im startign to wonder if they\'re fake or not.  and then she is just plain bossy.  she doesnt act like this towards her friends, at least i dont think. just her family. and i dont think its grandpa dying that stressed me out, actually i was pretty okay with it. i think its her. i\'ve tried talking to dad about it, trying to convince him she needs to be on something like zoloft or welbutrin, but he just jokes about how she always \"lets off steam\" or has \"perpetual PMS.\"  i cant wait until the beginning of school just b/c thats when shes leaving.  i mean, part of e will miss her, but most of me is wanting to pull her hair, break her glasses, and give her a good punch to set her crooked nose straight.  '),(3557,3992688,'31,July,2004','2004-07-31 00:00:00','    A fellow blog has put up an excellent  urlLink analysis  of CNN\'s anti-Kerry bias during the Democratic National Convention. Newsweek is also reporting a \"baby bounce\" for John Kerry in a new  urlLink poll . It has Kerry at 49 and Bush at 42, a gain of four points from the last Newsweek poll. The only problem is, it\'s a bad poll! The poll was averaged from polling on two days, one of which was Thursday. Kerry\'s speech was Thursday night, meaning an accurate poll meant to measure a bounce couldn\'t be done accurately until Friday. Coincidentally, the polling done on  urlLink Friday  showed a much greater lead, with Kerry at 50 to Bush\'s 40 - a lead of 10-points! Though Newsweek does mention the fact that the results might be skewed because of this in the article, that apparently wasn\'t enough to stop them from publishing a poll, which will cited in articles and the media without the context the article provided. The truth is, no accurate polling of Kerry\'s convention bounce will be available until next week, with polls covering days from Friday, and not before.    UPDATE: Paul Krugman takes on CNN\'s bias too in his   urlLink  latest Op/Ed   .   '),(3558,3992688,'30,July,2004','2004-07-30 00:00:00','    I know, I know. The race is tight. The networks says Kerry and Bush are neck-and-neck. Most urlLink  national polls  show Kerry leading, but it\'s in the margin of error, therefore it\'s labeled a \"statistical tie.\" However, when one uses state polls to determine the standing in the race (as the electoral collage is what really determines who wins), you see Kerry has a  urlLink bigger lead . Of course, many of those state polls are within the margin of error too in the battleground states and some will change over the next three months, though the  urlLink trend  favors Kerry. Nontheless, the meme from most news organizations is that is a tight race. But is it really? We won\'t know the bounce from Kerry\'s convention speech until sometime next week, but John Zogby has very interesting early  urlLink analysis .   Basically, Kerry\'s leading all Democratic base constituencies, including the Catholic vote and even leads slightly in the South, while Bush\'s base is eroding. But what is even more interesting is if turnout is as high as many think it might be, this election might not be as close as people think:      \"Mr. Kerry is showing a 2-to-1 lead (50% to 25%) amongst voters who didnâ€™t vote in 2000, while winning three-quarters (75%) of Ralph Naderâ€™s voters and stealing twice as many (8% to 4%) of Mr. Bush voters in 2000 than Bush is stealing of Gore voters in 2000.\"      Now here inlies the problem with the current polling. Most polls take into account what they refer to as \"likely voters.\" Though any thinking person might assume \"likely voters\" means they asked who they were polling if they were likely to vote. Not so. \"Likely voters\" means those who voted in the last election. We all know many thought the 2000 election was a farce and stayed home. We all also know that\'s not gonna be the case this time. 9/11 happened since then. Bush happened! Turnout is gonna be high, and it will favor us. I will grant you the electorate is polarized, I see that everyday, but Kerry could be very possibly win over 50% of votes (which someone hasn\'t done in awhile). So let\'s start repeating our own meme (the right and the media do it all the time, so it\'s only fair):  Kerry, in a landslide .    UPDATE: William Saletan  urlLink analyzes the polls .   '),(3559,3992688,'30,July,2004','2004-07-30 00:00:00','    Last night, John Kerry gave his  urlLink acceptance speech  for the Democratic nomination for President.&nbsp; We here all believed it was a great speech, and that Kerry did what he needed to do to sell himself to undecideds. From the few I saw on the media channels last night, it looks to be so. We will see in the coming days what kind of bounce John Kerry gets from his speech and how long it lasts. For right now, I want to discuss the speech itself. I was certainly, like many others, impressed with the strength of the speech, that he didn\'t shy away from Democratic values like so many others,&nbsp;and that Kerry seemed like a new man. He seemed more comfortable, more personable, more presidential. He seemed  ready . Last night made me truly proud this guy is the Democratic nominee and I will proud to call him President.    &nbsp;  I have collected some highlights from the speech below:    &nbsp;  \"I\'m John Kerry, and I\'m reporting for duty!\"    &nbsp;  \"There is nothing more pessimistic than saying America can\'t do better.\"    &nbsp;  \"Now I know there are those who criticize me for seeing complexities and I do because some issues just aren\'t all that simple. Saying there are weapons of mass destruction in Iraq doesn\'t make it so. Saying we can fight a war on the cheap doesn\'t make it so. And proclaiming mission accomplished certainly doesn\'t make it so.\"    &nbsp;  \"As President, I will bring back this nation\'s time-honored tradition: the United States of America never goes to war because we want to, we only go to war because we have to. \"    &nbsp;  \"We are here to affirm that when Americans stand up and speak their minds and say America can do better, that is not a challenge to patriotism; it is the heart and soul of patriotism.\"    &nbsp;  \"That flag doesn\'t belong to any president. It doesn\'t belong to any ideology and it doesn\'t belong to any political party. It belongs to all the American people.\"    &nbsp;  \"I want an America that relies on its own ingenuity and innovation, not the Saudi royal family. \"    &nbsp;  \"Let\'s honor this nation\'s diversity; let\'s respect one another; and let\'s never misuse for political purposes the most precious document in American history, the Constitution of the United States.\"    &nbsp;  \"Maybe some just see us divided into red states and blue states, but I see us as one America red, white, and blue.\"    &nbsp;  \"I don\'t want to claim that God is on our side. As Abraham Lincoln told us, I want to pray humbly that we are on God\'s side. \"   '),(3560,3992688,'29,July,2004','2004-07-29 00:00:00','    Today in his  urlLink column  in the New York Times, Paul Krugman mentions this:   \"P.P.S.: Three weeks ago, The New Republic reported that the Bush administration was pressuring Pakistan to announce a major terrorist capture during the Democratic convention. Hours before Mr. Kerry\'s acceptance speech, Pakistan announced, several days after the fact, that it had apprehended an important Al Qaeda operative. \"    &nbsp; The Al Qaeda operative they\'re referring to is Ahmed Khalfan Ghailani, who has been wanted in connection with the embassy bombings in Tanzania and Kenya. He was captured last Sunday, but the Pakistani goverment waited until yesterday to  urlLink publicly announce &nbsp;his capture.    &nbsp; I didn\'t see this article in The New Republic when it came out, so I read it for myself. In it the authors discuss the pressure being put on Pakistan to produce \"HVTs\" in time for the November elections. I\'ve quoted the passage Krugman is referring to specifically below:    \"A third source, an official who works under ISI\'s director, Lieutenant General Ehsan ul-Haq, informed tnr that the Pakistanis \"have been told at every level that apprehension or killing of HVTs before [the] election is [an] absolute must.\" What\'s more, this source claims that Bush administration officials have told their Pakistani counterparts they have a date in mind for announcing this achievement: \"The last ten days of July deadline has been given repeatedly by visitors to Islamabad and during [ul-Haq\'s] meetings in Washington.\" Says McCormack: \"I\'m aware of no such comment.\" But according to this ISI official, a White House aide told ul-Haq last spring that \"it would be best if the arrest or killing of [any] HVT were announced on twenty-six, twenty-seven, or twenty-eight July\"--the first three days of the Democratic National Convention in Boston. \"    &nbsp; Of course the Bush administration denies putting any such pressure to produce high-ranking members of Al Queda with the elections in mind, but the timing of Pakistan\'s latest announcement seems a little fishy to me too. The point of delaying the announcement of his capture until yesterday would be to draw some attention away from Kerry, and help to focus some the attention on the Bush administration and their success in the war on terror. However, given the fact that Ghailani seemed to be mostly unknown by the American public before his capture, the advantage in announcing his capture to coincide with the DNC convention is slight at best. But it does give credence to the belief that there is no issue, not even ones of national security, that the Bush administration won\'t play politics with.   &nbsp;   &nbsp;   '),(3561,3992688,'29,July,2004','2004-07-29 00:00:00','    The Texas Democratic Party is trying to raise 15,000 over the next four days. If we win a net gain of just 2 seats in the next elections in the Texas House, we would be the majority again. This would be a great way to tell Tom Delay, Rick Perry, and Tom Craddick (who would likewise be out of a job)&nbsp;that\'s what we think of their efforts to redistrict out our US Reps, and that they can kiss our Texan asses.    urlLink http://www.txdemocrats.org/contribute/15k14days/index.php   '),(3562,3992688,'28,July,2004','2004-07-28 00:00:00','    As I mentioned in my earlier post on this topic, the run-up to any war in Iran begins first with the coordinated effort by various hawkish and conservative pundits to convince the American people of the \"rightness\" of any preemptive invasion. This first requires an airing out of the idea of the invasion, first in conservative media sources like  urlLink The Weekly Standard  or  urlLink The Washington Times . The next step is to take the op-ed columns to the mainstream media, like  urlLink The Washington Post , where they can try to reach a greater number of Americans and persuade them of the threat of a nuclear Iran. This is exactly what happened with Iraq, and the hawks are trying it again Iran. So I\'m going to keep track of the \"chatter\" out there in the conservative media, with a weekly post I\'ll call \"Iran Invasion Watch.\" Each week I\'ll post links to new articles, op-ed pieces and commentary that I see in the conservative media, advocating the invasion or Iran, or building up the potential threat of Iran to pave the way for talk of an invasion.  And of course we\'ll all keep an eye on what the administration has to say about Iran, to get a sense of whether what the pundits say is gaining traction.  Today we have this, from  urlLink American Daily . There\'s no specific recommendation for invasion, but rather an effort to set the mood by focusing on the threat of  urlLink Iranian-sponsored terrorism.     PS. Occasionally I\'ll post the left\'s response to Iran invasion talk. Today we have  urlLink this , from Znet.   '),(3563,3992688,'28,July,2004','2004-07-28 00:00:00','    The US-led coalition\'s failure to restore security has turned Iraq into a  urlLink battleground  for the likes of Osama bin Laden\'s Al-Qaeda network, a parliamentary committee said.&nbsp;&nbsp;In a major report on the war on terrorism released Thursday, the House of Commons foreign affairs committee said the lack of law and order had created a \"vacuum\" for criminals and militias, with \"appalling consequences\" for the Iraqi people.&nbsp; It added that Iraq\'s own police and armed forces are still \"a long way from being able to maintain security,\" and warned that ongoing violence could mar elections planned for early next year.     Well, this report agrees with what we already know - not only did the Iraq war take the focus off Al Qaeda and the war on terrorism, it probably made it worse! Now we are fighting an unwinnable war with dire consequences.     Thankfully, America has the option of correcting this course. Tonight, John Kerry will  urlLink accept  the Democratic Party\'s&nbsp;nomination for president. The choice is simple, and the action easy. All you have to do is cast your vote this November.   '),(3564,3992688,'28,July,2004','2004-07-28 00:00:00','    Fred Kaplan explores this issue of an arms race with North Korea in this  urlLink excellent article  at slate.com.    \"The parallel between then and now is not precise. North Korea does not have the resources that the Soviet Union had at the height of the Cold War. But with the deployment of the new missile-defense system, the United States has entered into an arms race with the North Koreansâ€”an arms race we are likely to loseâ€”and nobody in the White House or the Congress seems even to be aware of it.\"     &nbsp;  For nearly two years now critics of the Bush administration\'s policy towards North Korea have tried in vain to point out that of the three members of the \"Axis of Evil\", Iraq, Iran and North Korea, North Korea by far presents the most immediate threat to the United States in the form of a revamped nuclear weapons program. It\'s no secret that North Korea would like the ability to gain some leverage over it\'s Asian neighbors with it\'s nuclear weapons(and perhaps eventually the&nbsp;ability to  urlLink threaten the United States.)&nbsp; It\'s also speculated that North Korea may seek to sell some of it\'s nuclear weapons technology to other nations that we might not like to see armed with these weapons. Unfortunately our administration seems to think that installing an ABM system in Alaska while avoiding direct negotiations with N. Korea is the best policy for dealing with this threat.   The hawks of course are silent on the issue of nuclear weapons in North Korea. For one it doesn\'t fit into their grand schemes of remaking the Middle East, and second North Korea already  has  nuclear weapons, and any military effort to remove them would not nearly be as easy as the invasion of Iraq.   I for one believe only a stiff combination of diplomacy and the threat of action can push North Korea away from the further development of a nuclear weapons program. Our only choice is to make it worth more to them(in food and technical aid)to go without nuclear weapons.   &nbsp;   '),(3565,3992688,'27,July,2004','2004-07-27 00:00:00','    In the aftermath of the invasion of Iraq, many of of us who opposed the war began to wonder how long it would take the Bush administration to find another target for their doctrine of \"pre-emptive\" war. &nbsp;Sure enough some of the hawks, swelled with confidence from the largely successful invasion, began saber rattling in the direction of Syria, making accusations that Syria&nbsp;was hiding&nbsp;Saddam\'s WMDs, assisting leftover Fedeyeen in resisting us, funding terrorists groups throughout the Middle East, and before he was found, of even hiding Saddam himself. Then, amidst the nearly absolute failure of the Administration to prepare for the aftermath of war, the&nbsp;resistence to the occupation&nbsp;began in earnest and all talk of any more invasions withered on the vine as our soldiers found themselves bogged down battling a tenacious insurgency which continues to this day. I for one figured that the hawks had finally realized they\'d bitten off more than they could chew, and would realize that they\'re focus needed to remain on restoring security to Iraq. More war was almost certainly out of the question.   Until  urlLink now  that is.   It seems that I\'ve overestimated the hawks and their commitment to Iraq. I\'d always assumed that despite their incredible naivete in how to go about it, they were actually serious about turning Iraq into a shining example of democracy in the Middle East. Articles like Krauthammer\'s make it clear that, at least for some of the hawks, that\'s not the case. Krauthammer sincerely believes that the threat of nuclear weapons on Iraq is great enough such that we should currently take all of our soldiers battling to secure Iraq and send them flying across the border and into Tehran, to replace the Islamic regime currently in power. Who exactly will be left behind to guard Iraq is not addressed in his column. But it didn\'t take me long to realize&nbsp;that\'s not the point. What\'s left out is that the the reason Iran is pursuing nuclear weapons so feverishly is because of the failure of this administration to deal with the threat&nbsp;sooner, and the lesson they\'ve learned from North Korea that nuclear weapons will actually deter attack from us, not invite it. What we have here is another cynical attempt to distract the American people from a real problem, Iraq, to focus on another problem of seemingly greater urgency, Iran and nuclear weapons. It\'s an awful lot like what they did the first time around; distracting us from the real threat of Al Queda by pushing the case for war in Iraq. Some hawks seem to think you can play this prank on the American people twice. I for one don\'t agree, but be on the lookout for more articles like this to appear, as the hawks try desperately to divert the attention from Iraq as they make their case for war in Iran.   As for me, I\'ll talk more later about what we actually  should  be doing about Iran.   UPDATE: Krauthammer may be on the earliest advocates of war with Iran, but he\'s not alone. The first step in making this US policy is for the pundits and commentators to throw the idea out there on their blogs, op-eds and websites.&nbsp; Some examples:    urlLink New&nbsp;York Daily News(via FrontpageMagazine.com)    urlLink The Washington Times (predictably trumpeting the threat of Iran)   '),(3566,3992688,'27,July,2004','2004-07-27 00:00:00','    Turns out 10 Republicans who might be investigating Tom Delay  urlLink accepted some of his PAC cash &nbsp; and a computer crash  urlLink erased  detailed records from Miami-Dade County\'s first widespread use of touchscreen voting machines. Meanwhile, a Republican senator  urlLink speaks out  against the Iraq war, while George Bush\'s  urlLink best defender  against the accustations of \"Fahrenheit 9/11\" turns out to be one of Osama Bin Laden\'s brothers. Speaking of Iraq, three sons of an Iraqi provincial governor were  urlLink kidnapped &nbsp;and an army captain has been charged with  urlLink murdering  an Iraqi. But in better news, Democrats&nbsp;today will  urlLink formally nominate  John Kerry as their challenger to President Bush in November and hear from vice presidential running mate John Edwards in a nationally televised speech.     &nbsp;    &nbsp;   '),(3567,3992688,'27,July,2004','2004-07-27 00:00:00','    Tomorrow night a man named  urlLink Steve Brozak , who is running for congress in New Jersey, will be speaking at the Democratic National Convention about terrorism and defense. He plans to tell them that&nbsp;John Kerry \"offers a real vision ... for making sure we win the war on terror and making sure we have solutions for the Iraqi conflict.\" He is an  urlLink ex-marine, ex-Republican investment banker  who left his party because of their negative attacks&nbsp;on the patriotism of&nbsp;war veterans such as Max Cleland during the 2002 mid-terms (those ads and Paul Wellstone\'s death are the only reasons we lost the Senate). Brozak went on the say he saw too many Republicans with the \"arrogant attitude that if you ask questions, you\'re considered disloyal.\" He is a moderate Democrat&nbsp;who will be hard to attack because of his background and will be a formittable candidate.    &nbsp;  Though he is still a fiscal conservative and social moderate, I am glad to have this guy on our side.&nbsp;Being a military man himself, his split with the Republican Party over that issue is certainly understandable.&nbsp; It is great that he was invited to speak at the convention, and even better that he will speak so strongly about Kerry\'s ability to protect America in the face of more of the Republicans\' lies and distortions. Republicans might have that&nbsp;loon&nbsp;Zell Miller on their side this year, but I\'m much more pleased to have Steve Brozak.   '),(3568,3992688,'26,July,2004','2004-07-26 00:00:00','    An  urlLink op-ed piece  in today\'s NY Times discusses the Saudi connection to the 9/11 attacks, and the failure of the 9/11&nbsp; Commission to scrutinize that connection in great detail. I don\'t entertain many conspiracy theories, but in this case it certainly seems that there is more than meets the eye when talking about Saudi assistance of the 9/11 hijackers. I don\'t believe there was any official connection by the Saudi government to the terrorists, but I do not rule out the possibility that certain Saudi royals helped to fund Al Queda in general as the group was preparing for these attacks. I do believe that this connection is something that will never get the serious attention it deserves from anyone in our government. The truth is we have far too much at stake in our relationship with the government of Saudi Arabia, in that we have vast economic interests in the oil that is lying under their desert sand.&nbsp;We simply cannot afford to call the Saudi royals to the carpet for any participation they may have had individually or as groups with Islamic terrorists while we remain beholden to the oil that they sell us.&nbsp;Changing this relationship requires a fundamental change in energy policy on our part, away from our reliance upon vast quantities of oil for our cars and our industries. Unfortunately there is no incentive to do this either in the Republican or the Democratic party, as many politicians believe that we Americans are unwilling to make the sacrifices necessary to change our policy. For now we simply must accept the bland assurances of the 9/11 Commission, and wait for&nbsp;a day when our energy policy is not the prime determinant of our national security policy.   &nbsp;   '),(3569,3992688,'26,July,2004','2004-07-26 00:00:00','    Today, as I was filling out a job application at a local restaraunt out here in East Texas, I noticed an older&nbsp;man walk in to apply. The rugged man look to be in his 40\'s; your average&nbsp;conservative blue-collar man wearing a baseball cap with lyrics from a country song on it, blue jeans, and a T-shirt. He quickly started up a conversation with a another, younger man like myself about how his company layed him off two years ago and hasn\'t been able to find work since. His story was not atypical: He had put a lot of years&nbsp;into the job, but it was all for-not, and he was told to he had 2 days to \"get his shit out of there.\" As I was listening to this, two thoughts occured. The first was that while liberals and conservatives often talk a lot about the economy it terms of statistics, it is nice (or in this case, not) to see the reality of it all.&nbsp; The second, was that this guy is probably a Republican, or at least, votes for them and unfortunately, probably still will in these upcoming elections. I have no doubts&nbsp;about why, it\'s for the same reasons&nbsp;people have been voting against their own economic interests for awhile&nbsp;- Republican talk of values, less government, and as Toby Keith (who\'s actually a Democrat) would put it, putting a boot in our enemy\'s ass. Of course, what it really means is that this man doesn\'t have a job anymore, in large part due to Republican\'s disastrous&nbsp;economic policies (it didn\'t used to be this way, but the Pat Buchanan wing of the party is&nbsp;all but&nbsp;dead). This particular thought continues to sadden me, because this is the kind of guy who should be on our side. But I have no idea if Kerry\'s economic message is getting through to people like this man, at least down here. Though I don\'t know if his particular job was&nbsp;lost due to outsourcing, the same basic principle remains:     &nbsp;   Big companies have too much damn power in America and it is time we did something about it.     &nbsp;   Republicans surely won\'t, but Democrats, with progressives at the grass-roots level like us edging them on, will. So I again think of what I said earlier,&nbsp;we should support the Texas Democratic Party\'s and Glenn Smith\'s DriveDemocracy Organization, and get pro-worker Democrats in and pro-corporate Republicans out. It will hard to turn back the tide after a decade of Republican control in Texas and on the national level, but we have to start somewhere. And we will start with guys like the one I met today.   '),(3570,3992688,'25,July,2004','2004-07-25 00:00:00','    Naturalized citizens in Florida are saying that the Republican Party there has been registering them to vote&nbsp;as Republicans -  urlLink without giving them a choice .    &nbsp;  Is there nowhere too low for these GOP scumbags to sink to? Wait, don\'t answer that.    &nbsp;  Alas, this is really no surprise. It\'s not like they won it last time.   '),(3571,3992688,'24,July,2004','2004-07-24 00:00:00','    During his public testimony before the 9/11 commission, Attorney General John Ashcroft attempted to deflect criticism from his own lackluster counterterrorism efforts by pinning the blame on a 1995 memo written by former deputy Attorney General (and current 9/11 commissioner) Jamie Gorelick. Ashcroft said, \"The 1995 guidelines and the procedures developed around them imposed draconian barriers, barriers between the law enforcement and intelligence communities. The wall effectively excluded prosecutors from intelligence investigations. The wall left intelligence agents afraid to talk with criminal prosecutors or agents.\" Ashcroft called the memo \"the single greatest structural cause for the September 11 problem.\"   In their  urlLink final report , the bi-partisan 9/11 commission concluded that Ashcroft\'s public testimony was false and misleading. The commission bluntly stated that Ashcroft\'s public testimony did not \"fairly or accurately reflect the significance of the 1995 documents and their relevance to the 2001 discussions.\" Specifically, \"The Gorelick memorandum applied to two particular criminal cases, neither of which was involved in the summer 2001 information-sharing discussions.\" Any barriers between the law enforcement and intelligence communities were not created from written guidelines by internal Justice Department conflicts which \"neither Attorney General acted to resolve\" prior to 9/11. Even Ashcroft himself has recently backed away from his April testimony before the commission. In a recent document released by the Justice Department, Ashcroft conceded that Gorelick\'s memo permitted \"interaction and information sharing between prosecutors and intelligence officers\" and allowed the FBI to use the fruits of an intelligence investigation \"in a criminal prosecution.\" Ashcroft, however, failed to mention that guidelines issued by his own deputy Attorney General, Larry Thompson, were more restrictive because they affirmed the Gorelick memo and added additional requirements.   '),(3572,3992688,'23,July,2004','2004-07-23 00:00:00','    The U.S. military today admitted  urlLink ties  to Johnathan Idema, the Afghan vigilante, in that it held an Afghan prisoner for two months after receiving him from&nbsp;Idema\'s group&nbsp;which has been charged with torturing detainees at a private jail. The admission followed claims by&nbsp;Idema, who was in the Army reserve from 1975 through 1984 and received special forces training,&nbsp;that&nbsp;he was working for&nbsp;the Defense Department and had contact with Rumsfeld\'s office almost daily.   '),(3573,3992688,'22,July,2004','2004-07-22 00:00:00','    Iran\'s conservative-dominated parliament approved a draft  urlLink bill  that would allow abortion in the first four months of pregnancy if the woman\'s life is in danger or the fetus is malformed. I think it is sad that most Iranian clerics could support such a bill, whereas earlier this year the South Dakota congress tried to pass a bill that would effectively ban abortion even in these cases. Fortunately, the state senate killed the bill after it was sent back by the governor, but it is very disheartening that conservatives would try to so blatantly undermine the U.S.S.C. on a woman\'s right to choose. While the Iranian law would not support abortion in the case of unwanted pregnancies, I find it interesting that a place that we criticize as so religiously fundamentalist supports an obvious and pragmantic policy, when some of our own religious leaders and lawmakers at home do not.   '),(3574,3992688,'22,July,2004','2004-07-22 00:00:00','    As the House Ethics Committee decided to  urlLink extend their review  of House Majority Leader Tom Delay\'s illegal&nbsp;use of Enron contributions for Texas races, a recent study by the watchdog group, Campaigns for People, has found most of the 31 members of the Texas Senate draw on the kindness of big campaign donors to help support rich lifestyles they otherwise couldn\'t afford. The group\'s biggest discovery&nbsp;was that only 40% of the senators\' campaign funds go toward actual campaign activities. The remaining 60%, the report found, paid for lifestyle, office, or miscellaneous expenses. The study, aptly named \"Money in All the Wrong Places,\" is based on an examination of senators\' campaign-finance reports filed electronically with the Texas Ethics Commission. The data covered a three-year time period between Jan. 1, 2001, and Dec. 31, 2003.&nbsp;     &nbsp;   This is the kind of thing that would make even the most cynical political observers, including us, cringe.     &nbsp;   More disheartening is that this is happening in our home state, illustrating how much&nbsp;our system is hobbled by ineffective campaign-finance and ethics laws and the amount of influence corporate powers have, even moreso than at the national level. But until we stop electing corporate-backed Republicans and Democrats-in-name-only and change the laws, there\'s never going to be a difference. We must help the  urlLink Texas Democratic Party  and groups like  urlLink DriveDemocracy  to elect better people, from the grassroots to the national level. Let\'s start by showing Mr. Tom \'Teaching evolution caused the Columbine shootings\" Delay&nbsp;that even redistricting won\'t stop us from keeping a Democratic majority in our congressional delegation and winning as many state races as possible. From U.S. Senator to your local school board, every vote makes a difference.   '),(3575,3992688,'22,July,2004','2004-07-22 00:00:00','    Which passenger you ask? Oh yeah.  urlLink Annie Jacobson . She doesn\'t come off quite so well in this version of events. Again, I think we have a situation here of a hysterical woman over-reacting to a situation, and so making a big stink out of something that wasn\'t that big of a deal after all. Oh yeah, and then going on to write a  urlLink 3000 word column online  about it.   &nbsp;   '),(3576,3992688,'22,July,2004','2004-07-22 00:00:00','    After the Senate failed to pass the Federal Marriage Amendment, the Republican-controlled House passed today the so-called \" urlLink Marriage Protection Act \" which would prevent the Supreme Court and other courts from ordering states to recognize same-sex unions sanctioned elsewhere.&nbsp; In other words, the law is designed to prohibit these courts from ruling on the constitutionality of gay marriage bans.&nbsp; Did I miss something here? It is not up to legislators to decide what courts can and cannot rule on. This is the most blatant attack on the Constitution and the Seperation of Powers I have ever seen.&nbsp; What makes them think the courts won\'t just decide this law is unconstitutional (as it so obviously is)&nbsp;anyway? Fortunately, I don\'t think this will pass the Senate, either by vote or because the Democrats will fillibuster. What happened to conservatives wanting to protect the Constitution? Why would they even want this? What if, say, a Democratic House decided to pass a gun control law that no court could rule on? This is just idiotic.   '),(3577,3992688,'21,July,2004','2004-07-21 00:00:00','    So&nbsp;it turns out that all those menacing Arab guys on&nbsp;Mrs. Jacobson\'s&nbsp;plane were just  urlLink Syrian musicians . In fact as it turns out the security concerns that might have actually been worth heeding in Jacobson\'s mostly melo-dramatic piece, were  urlLink not that serious after all  (via Political Animal.) Look, I\'m not a liberal with my head stuck in the sand. I can understand how a group of dark-skinned Arabic gentleman milling around in my plane might cause some fleeting thoughts of terrorists to leap into the average passenger\'s brain. However, my response would not be to&nbsp;write a  urlLink 3,000 word mostly hysterical and moderately&nbsp;racist article  that contributes virtually nothing to the debate over airplane security. She leaves me with the impression that the only possible thing we could do to make her feel safer would be to ban large groups of Arabic men, or maybe all Arabic men, from flying in American planes.   '),(3578,3992688,'21,July,2004','2004-07-21 00:00:00','    Today the 9/11 Commission releases their  urlLink final report  on the attacks. With all the politicking surrounded the efforts of this commission, I\'d like to give the members of the commissions-Democrat and Republican-kudos for rising above their partisan affiliations to release a thorough and detailed report. I think they did as good a job as they could, considering the circumstances.   '),(3579,3992688,'21,July,2004','2004-07-21 00:00:00','    An American arrested in Afghanistan   with two countrymen for  urlLink illegally detaining people  he suspected of being Islamic militants said Wednesday he was working for the U.S. government, and he had evidence to prove it.     \"We were in contact directly by fax and email and phone with Donald Rumsfeld\'s office,\" he said, referring to the Secretary of Defense.     Rumsfel\'s office&nbsp;is denying a link, but columnist Ted Rall says  urlLink he met Idema  and has a different story to tell.    Hmm...     &nbsp;   '),(3580,3992688,'21,July,2004','2004-07-21 00:00:00','    Just as it was reported that the American  urlLink military death toll in Iraq  has reached 900 and the US Army is calling up recruits  urlLink earlier , House Republicans decided to back a  urlLink block  expansion of a program to improve military housing. This is hardly surprising considering the VA hospital budget and combat pay cuts they\'ve tried to pass through. Though they could care less about our military, Republicans are making sure  urlLink no lobbyist is left behind . Oh, and it looks like the war cost was a little  urlLink underestimated &nbsp;and war funds are dwindling according to the GAO.     &nbsp;   Oops.   '),(3581,3992688,'21,July,2004','2004-07-21 00:00:00','    Despite all of those bumper stickers that claim  urlLink Bin Laden\'s man is Kerry , the terrorists themselves seem to say  urlLink otherwise .   &nbsp;   '),(3582,3992688,'21,July,2004','2004-07-21 00:00:00','    The 2004 ACLU Membership Conference in San Francisco featured a keynote address from New Yorker magazine investigative journalist Seymour Hersh. The Pulitzer Prize winning Hersh has been at the top of his game recently, writing more exclusive stories on the Abu Ghraib torture scandal than any of his competitors.       &nbsp;   Seymour Hersh&nbsp;&nbsp;told the conference&nbsp;the US government has videotapes of boys being sodomized at Abu Ghraib prison. \"The&nbsp;worst is the soundtrack of the boys shrieking,\" the reporter told an ACLU convention last week. Hersh says there was \"a massive amount of criminal wrongdoing that was covered up at the highest command out there, and higher.\"     &nbsp;   You can see the video of the speech or read a transcript  urlLink here .   '),(3583,3992688,'21,July,2004','2004-07-21 00:00:00','    So says the  urlLink Sydney Morning Herald ... Whether this can be substantiated or not is still undetermined.&nbsp; They have promised an  urlLink investigation , but who knows what will come of it. I seriously doubt this occured, because I don\'t see what Allawi would gain from it. I think, with his past, he is certainly capable of doing something like this, but we will have to wait for further evidence.   '),(3584,3992688,'21,July,2004','2004-07-21 00:00:00','    A&nbsp;brilliant op/ed from Paul Krugman on Al Qaeda\'s choice for president...    &nbsp;  urlLink http://www.nytimes.com/2004/07/20/opinion/20krug.html   '),(3585,3992688,'21,July,2004','2004-07-21 00:00:00','    ...in the form of Arabic men  urlLink daring to fly commercial  that is. Would it be too much to refer to this woman\'s article as  urlLink racist hysteria ? I don\'t think so.&nbsp; Still, I envy anyone who finds so much drama in their everyday airplane flight. Last time I was on a plane I spent more time wondering if the  urlLink laws of physics would continue to work until I got back on the ground .   '),(3586,3992688,'21,July,2004','2004-07-21 00:00:00','    How liberal is Howard Dean? He supports  urlLink incest !   It is the National Review, so how can I NOT trust it?   &nbsp;   '),(3587,3992688,'21,July,2004','2004-07-21 00:00:00','    This is my new blog, which I\'ve created purely on a whim. I figure I\'ll post some commentary on politics, or things I find amusing, or somesuch. We\'ll see.   '),(3588,3992688,'18,August,2004','2004-08-18 00:00:00','    A top Republican congressman has   urlLink  broken from his party    in the final days of his House career, saying he believes the U.S. military assault on Iraq was unjustified and the situation there has deteriorated into \"a dangerous, costly mess.\"       \"I\'ve reached the conclusion, retrospectively, now that the inadequate intelligence and faulty conclusions are being revealed, that all things being considered, it was a mistake to launch that military action,\" Rep. Doug Bereuter, senior member of the House International Relations Committee and vice chairman of the House Intelligence Committee, wrote in a letter to his constituents.       \"Left unresolved for now is whether intelligence was intentionally misconstrued to justify military action,\" he said.   '),(3589,3992688,'17,August,2004','2004-08-17 00:00:00','    The trial of three Americans accused of jailing, kidnapping and torturing prisoners in Afghanistan was dramatically halted after the FBI returned a \"substantial\" amount of evidence to Afghan authorities.   Judge Abdul Baset Bahktiari adjourned the trial for seven days to allow the Americans and their four Afghan co-accused time to study the evidence, which prosecutors said had been held by the FBI for more than 20 days.     The defendants, arrested in July for allegedly running a private prison and counter-terrorism operations in Kabul, had earlier accused the Federal Bureau of Investigation of withholding evidence proving their links to US authorities.   '),(3590,3992688,'16,August,2004','2004-08-16 00:00:00','    \"Voodoo economics.\"   -how George H.W. Bush referred to Reagan\'s economic policies      One of the biggest Republican myths is that cutting tax rates for the richest Americans will improve the standard of living for the working class. Supposedly, top-bracket tax breaks will result in more jobs being created, higher wages for the average worker, and an overall upturn in our economy. Itâ€™s at the heart of the infamous trickle-down theory, also known as supply-side or Reaganomics after it\'s biggest supporter. Despite Bush Sr.\'s objections, his son and most Republicans ascribe to the same theory. It\'s hard to see why many people still buy into it, given that in fact, it has never lead to more jobs, higher wages, etc. I mean, anyone living today knows this has not been the case with the Bush tax cuts.  Obviously jobs have not been created and people\'s financial situation has not only not improved, but gotten worse. Yet, as I stated, many still believe this is a good way to stimulate the economy. This site explains   urlLink  how and why it has not worked    over the last few decades. So next time someone tries to tell you cutting taxes for the wealthy will help you, you can give them the facts about  such \"voodoo economics.\"   '),(3591,3992688,'16,August,2004','2004-08-16 00:00:00','    A   urlLink  study    by Michigan State University political science professors found that, contrary to the belief of political pundits, Americans\' fears of another attack actually makes people less likely to vote for Bush:      \"Of those who were \'very concerned,\' 38 percent said they planned to vote to re-elect President Bush, while 76 percent of those who were \'not at all concerned\' said they would vote for Bush.\"      So political pundits were wrong! Suprise, surprise.   '),(3592,3992688,'15,August,2004','2004-08-15 00:00:00','    Jeb would do anything for his brother.  The New York Times   urlLink  reports    that \"state police officers have gone into the homes of elderly black voters in Orlando and interrogated them as part of an odd \"investigation\" that has frightened many voters, intimidated elderly volunteers and thrown a chill over efforts to get out the black vote in November.\"  The more things change, the more they stay the same...  UPDATED: Also, the FBI   urlLink  tracks    GOP protesters.   '),(3593,3992688,'15,August,2004','2004-08-15 00:00:00','    I found this site for a group called the Republican Youth Majority, which   urlLink  promotes    a Republican agenda that is pro-choice, for common sense environmental policies, fiscal responsibility, and perhaps less surprisingly, investing in young people\'s future. Whether they\'ll have a voice at the upcoming Republican convention remains to be seen, but if it indeed represents the views of a majority of future Republicans there may be hope again for this party.   '),(3594,3992688,'14,August,2004','2004-08-14 00:00:00','    The Washington Post has an interesting  urlLink article  about how \"on the last day of 2003, in an action so obscure it was not mentioned in any major newspaper in the country\" the Bush administration eliminated the OSHA rules for tuberculosis testing. The article goes on to say how \"OSHA has altered its regulatory mission to embrace a more business-friendly posture. In the past 3 1/2 years, OSHA, the branch of the Labor Department in charge of workers\' well-being, has eliminated nearly five times as many pending standards as it has completed. It has not started any major new health or safety rules, setting Bush apart from the previous three presidents, including Ronald Reagan. The changes within OSHA since George W. Bush took office illustrate the way that this administration has used the regulatory process to redirect the course of government.\"      So instead of using government agencies to look out for the well-being of people, we are now using it to look out for the well-being of businesses.   '),(3595,3992688,'14,August,2004','2004-08-14 00:00:00','    Fred Kaplan asks this question is his latest \"War Stories\"   urlLink  column   . It\'s an interesting read than analyzes the current situation and our chances of avoiding failure. And well, it doesn\'t look good.   '),(3596,3992688,'14,August,2004','2004-08-14 00:00:00','    The Associated Press reports dozens of attackers raided a U.N. refugee camp in western Burundi, shooting and hacking to death at least 180 people.  A Burundian Hutu rebel faction claimed responsibility for the attack late Friday near the Congolese border, saying its fighters were in pursuit of Burundian soldiers who fled to the camp from a nearby army position. The assailants screamed war cries as they rushed into the camp and set it ablaze. The camp sheltered Congolese ethnic Tutsi refugees, known as the Banyamulenge, who fled fighting in Congo\'s troubled border province of South Kivu.      \"What we have seen so far are many, many, many bodies of children, women and men,\" Eliana Nabaa, spokeswoman of the U.N. mission in Congo said. \"People were sleeping when the attack happened. People were killed as they tried to escape.\"      Is it not time we took a real stand against human rights abuses in Africa?   '),(3597,3992688,'13,August,2004','2004-08-13 00:00:00','    Richard Reeves wrote an excellent   urlLink  article    on the current situation in Iraq:  \"So what are we doing in Najaf? Is killing the followers of a nasty Shiite preacher, killing them at the gates of the most holy shrine of Shiite Muslims all over the world, vital to the national interests of the United States and its allies? And why is it that we are killing Shiites, the wretched of the earth in the secular Sunni Muslim country of Saddam Hussein? That is the same Saddam who murdered the father of the preacher five years ago. Was that our clear intent and realistic objective in invading Iraq? Would the American people and Congress -- and our allies -- have supported a $200 billion war to get preacher Muqtada al-Sadr?\"  The answer is, of course, a resounding no.   '),(3598,3992688,'12,August,2004','2004-08-12 00:00:00','    The Bush adminstration admits   urlLink  no evidence of threat    to financial buildings. Also, the Washington Post finally   urlLink  admits    pre-war articles questioning Iraq threat often didn\'t make front page. Meanwhile, Halliburton contracts face   urlLink  new scrutiny   . Another reporter, this time for The New York Times, was   urlLink  subpoenaed    yesterday by a Washington grand jury investigating the disclosure of the identity of Valerie Plame. And a platoon leader\'s memorandum to an Army reservist charged with abusing Iraqi detainees   urlLink  praised his actions    as several GIs   urlLink  question the war   .    UPDATED: The CBO finds tax cuts did indeed   urlLink  favor the wealthy    and a federal judge   urlLink  orders    the U.S. to give documents on tortured prisoners to rights groups.   '),(3599,3992688,'11,August,2004','2004-08-11 00:00:00','    U.S. Congressman Porter Goss, President Bush\'s nominee for CIA director, could be his  urlLink own worst enemy  when it comes to making the case that he deserves to lead the U.S. intelligence agency.  \"I couldn\'t get a job with CIA today. I am not qualified,\" the Florida Republican told documentary-maker Michael Moore\'s production company during the filming of the anti-Bush movie \"Fahrenheit 9/11.\"    A day after Bush picked Goss for the top U.S. spy job, Moore Wednesday released an excerpt from a March 3 interview in which the 65-year-old former House of Representatives intelligence chief recounts his lack of qualifications for employment as a modern CIA staffer.    Of course, Republicans will probably contend that he was misquoted, Moore tricked him, etc.   '),(3600,3992688,'10,August,2004','2004-08-10 00:00:00','    \"I admit it -- the liberal media were never that powerful, and the whole thing was often used as an excuse by conservatives for conservative failures.\"   William Kristol, as reported by the New Yorker, 5/22/95      Starting today, I will be starting a new addition to the blog. From time to time I will be posting Republican/conservative myths and busting them. I will start with what is probably the biggest myth, that of the so-called \"liberal\" bias in the media, ironically, often perpetuated by conservatives in the media on mainstream media programs on mainstream TV and radio networks. Among them include Sean Hannity, Rush Limbaugh, Bill O\'Reilly, Bob Novak, Tucker Carlson, Tim Russert, Lou Dobbs, Judy Woodruff, Tony Snow, Brit Hume, Pat Buchanan, Ann Coulter, the \"Beltway Boys,\" Dennis Miller, Joe Scarborough and many more. Other major media figures such as Wolf Blitzer, Paula Zahn, Chris Mathews often give pro-Republican points-of-view. FOX News, of course, is a propoganda network for the Republican Party. CNN is little better. The vast majority of radio programs are right-wing. No one denies that. Clear Channel supports Republican causes. There are at least as many major newspapers that are conservative as liberal including the Washington Times, the New York Post, the Chicago Tribune, the Wall Street Journal and in large party, even the editorial pages of the New Tork Times and the Washington Post. Anyone who watched the coverage the Clinton\'s impeachment, the 2000 election, of the last four years knows whose side the bias supports. Don\'t believe me? Just visit   urlLink  Media Matters   , or read Al Franken\'s \"Lies and the Lying Liars Who Tell Them\" and other sources. But for most of you I\'m preaching the converted here. The important thing is to fight back. Quite watching CNN or any major news network, except for C-SPAN and PBS. Only listen to NPR and Air America Radio. Read your local alternative papers, like the Dallas Observer and Fort-Worth Weekly here in Texas, and forget anything bought and paid for by big corporations. Tell your representatives you won\'t tolerate further media consolidarity. And most importantly, point out this bias to your friends, family, and co-workers. We must expose these people for who they are and what they do. It\'s time for the truth to matter again.   '),(3601,3992688,'10,August,2004','2004-08-10 00:00:00','    The recent pressure on reporters to reveal sources in the Valerie Plame investigation could end in a U.S Supreme Court   urlLink  standoff   , according to two leading First Amendment attorneys, one of whom fears several leading Washington reporters could wind up in jail before it\'s over.      \"I think we are headed for a showdown and it would not surprise me in the least to see half a dozen reporters sitting in a jail in D.C.,\" said Lucy Dalglish, executive director of the Reporters Committee for Freedom of the Press, which monitors such cases. She believes Time magazine and other media outlets \"will be willing to take this all the way to the Supreme Court because the atmosphere here has become so difficult for journalists to promise confidentiality that the stakes are too high.\"      The Supreme Court last ruled on the issue in Branzburg vs. Hayes, the well-known 1972 case that offered a vague split decision. Five justices held that there was no privilege for journalists, but agreed that a qualified privilege should apply in some circumstances, which was the position of the four justices in dissent.   '),(3602,3992688,'09,August,2004','2004-08-09 00:00:00','    Jim Wallis   urlLink  argues    that if Democrats simply loosen their stance on abortion rights and made room for pro-life Democrats, many voters who would not normally vote for them now would. Many of these voters, he argues, are heavily religious, but share the beliefs of Democrats on social programs and are against the death penalty and hawkish foreign policy. However, they refuse to vote Democratic because they believe abortion is wrong. First, I would like to say this article makes a few good points. Democrats should take a \"pro-choice,\" but \"anti-abortion\" stance by promoting programs that would reduce unwanted pregnancies (especially for teenage and low-income women) and reform the adoption process to make abortions less necesarry. To paraphrase President Clinton, abortions should be safe and legal, but rare. No doubt Democrats would do better with pro-life voters to advocate this stance to show that no one really favors abortions. The article also attacks those that base their votes on this one issue, failing to look at other issues that may determine a party\'s \"value of life.\"       However, there are as many pro-life Democrats and pro-choice Republicans, and one party does not ostracize over this issue more than the other. Republicans may have more pro-choice Republicans speaking at their convention, but that is because these Republicans, namely Schwarzeneggar, Guiliani, and others are among the most popular in the country and the Republicans are hoping to use them to appeal to moderate, independent voters. The choice is purely political, and not because the Republicans believe in representing minority beliefs in their party.      Secondly, say Democrats did become more accepting of pro-life views, even bringing them into the mainstream. I agree that some voters that wouldn\'t normally vote for Democrats now would, but wouldn\'t the Democrats run the risk of losing female and liberal voters at the same time? I doubt there would be a net gain, and maybe there would be a net loss. It\'s the same thing with gay rights, or affirmative action. Many argue if the Democrats were less liberal in their stance they could gain more conservative voters. But where is the political gain in that? Everyone wants to hug the middle, but the farther you go the other side the more people you just drive away from your own base. Few are saying Democrats should start supporting gay marriage or unlimited welfare, but if they didn\'t take a stance somewhere on that side of the political spectrum, what would even be the point of voting for them as opposed to the Republicans? Really, what would the difference be?   '),(3603,3992688,'08,August,2004','2004-08-08 00:00:00','    As a federal judge on ordered \"Meet the Press\" host Tim Russert and Time magazine\'s Matthew Cooper to   urlLink  testify    before a grand jury investigating the leak of the identity of Valerie Plame, a new White House scandal has opened over the revealing of the name of a source.       Last week, the administration was desperate to justify their decision to raise the threat level to orange in three states based on activity that occurred over three years ago. National Security Advisor Condoleezza Rice admitted yesterday that the administration - during a background briefing to reporters - identified Mohammed Naeem Noor Khan as the source of the information that prompted the terror alert. According to Reuters, Khan \"had been actively cooperating with intelligence agents to help catch al-Qaida operatives when his name appeared in U.S. newspapers.\" His identification by the administration likely \"cost the United States avaluable source.\" It is not clear who originally disclosed Khan\'s name, which first appeared in The New York Times last Monday and was then confirmed by U.S. officials.       His unmasking triggered criticism across the political spectrum, and New York senator Charles Schumer   urlLink  has asked    for an explanation from the White House.       So why did the leak this name to the press? Carelessness, or is there a political motive? Discuss amongst yourselves!      UPDATED: The leak may have allowed al-Qaida suspects to  urlLink escape  and a Time magazine reporter was held in  urlLink contempt  for refusing to testify on the Valerie Plame affair.   '),(3604,3992688,'08,August,2004','2004-08-08 00:00:00','    The last article I posted was about a particular system that\'s coming down the pipe. This  urlLink article  is also about that but also touches on the planned overhaul of the entire military. And it\'s not all good.   '),(3605,3992688,'08,August,2004','2004-08-08 00:00:00','    Bet you guys didn\'t know about  urlLink this !   '),(3606,3992688,'08,August,2004','2004-08-08 00:00:00','    Iraq has issued an   urlLink  arrest warrant    for Bush administration buddy Ahmad Chalabi on counterfeiting charges and another for Salem Chalabi, the head of Iraq\'s special tribunal, on murder charges, Iraq\'s chief investigating judge said Sunday.      Ahmad Chalabi, a longtime exile opposition leader, had been a favorite of many in the Pentagon, and was used for much of the information about Saddam, but the administration has been distancing themselves for months and allegations have surfaced that he supplied Iranians with classified U.S. intelligence on American monitoring of Iranian communications.   '),(3607,3992688,'07,August,2004','2004-08-07 00:00:00','    This article is an interesting take on the transformation of the armed services that\'s going on now.  Read it  urlLink here.   '),(3608,3992688,'06,August,2004','2004-08-06 00:00:00','    Former presidential hopeful and talk-show host Alan Keyes   has  urlLink agreed  to become Illinois Republicans\' U.S. Senate candidate and will announce his intentions Sunday, a senior GOP official said Friday.   Keyes, a resident of Maryland, would face rising Democrat star Barack Obama in the race for the seat of retiring Sen. Peter Fitzgerald,   a Republican. The GOP was forced to look outside Illinois for a candidate after primary winner Jack Ryan dropped out in late June amid embarrassing sex club allegations in his divorce records and their second choice, former Chicago Bears head coach Mike Ditka, decided not to run.      Anyone who knows anything about him knows Alan Keyes is a far right wacko who will lose handily to Obama. Of course, the GOP really just want to put on a little show here by putting two black candidates against each other, going so far as to draft one from out of state. Speaking of, I find some of Keyes\' past remarks about Hillary Clinton\'s run for Senate in New York quite interesting in light of his recent decision:      \"I deeply resent the destruction of federalism represented by Hillary Clinton\'s willingness go into a state she doesn\'t even live in and pretend to represent people there, so I certainly wouldn\'t imitate it.\"      Is hypocricy a requirement for Republican candidates now?    UPDATE: Though running for election to the senate, Keyes   urlLink  argues against    election of senators.   '),(3609,3992688,'05,August,2004','2004-08-05 00:00:00','    As defense secretary under President George H.W. Bush, Dick Cheney   urged a presidential veto of legislation that would have given a national intelligence director budget authority over spy agencies run by the Pentagon. The letter Cheney wrote in 1992 addresses an issue now being debated: how much authority a national intelligence director should wield over budgets and personnel at America\'s 15 spy agencies. About 80 percent of the country\'s intelligence budget is controlled by the Defense Department.      The bills considered would assign \"inappropriate authority to the proposed director of national intelligence who would become the director and manager of internal DOD activities that, in the interest of efficiency and effectiveness, must remain under the authority, direction and control of the secretary of defense,\" Cheney wrote on March 17, 1992 to Les Aspin, then chairman of the House Armed Services Committee.    Could this be why the current George Bush is against the proposed intelligence advisor having full budgetary authority contrary to the suggestion of the 9/11 commission? Nah, it\'s unlikely a vice-president would have such influence over the president...   '),(3610,3992688,'04,August,2004','2004-08-04 00:00:00','    Four of the five Republicans on the House Ethics Committee, which will review the charges against DeLay next month, have received over $35,000 from an arm of the DeLay operation. They are in no position to conduct an independent investigation. The examination of Tom DeLay\'s political money machine should be free from the influence of that machine. That can only happen with an independent, outside counsel leading the probe.  urlLink Sign the petition now .   '),(3611,3992688,'03,August,2004','2004-08-03 00:00:00','    Serious fighting between insurgent and Iraqi police took place in the  urlLink city of Mosul today , 250 miles north of Baghdad. One part of the story which I found interesting:   \"The governor of Mosul, Dored Mohammed Kashmola, said in an interview on Al Jazeera television that eight civilians and four policemen had been were killed. He said Iraqi policemen were patrolling the roads Tuesday night without American troops, who, he said, had withdrawn to their bases.\"     I think it\'s become clear that, in large part, American forces are trying to turn over the brunt of the fighting to the Iraqi police and National Guard units. Of course this is good for our soldiers, who are getting killed at a lower rate(although they are still dying in roadside bombings and on patrols.) And it\'s good politically for the Bush administration, as the American people have more tolerance for the campaign in Iraq when fewer American soldiers are dying. This leaves the question however of whether or not the Iraqi police and military units are up to the task of defending themselves and the Iraqi citizenry. It appears they\'ve vastly approved since Fallujah and Najaf, when many Iraqi units flat-out refused to deploy against the insurgents. But so far the insurgency rages just as hot as it has at anytime in the past year, with the exception of the siege of Fallujah by US Marines. Will the insurgency enjoy greater success in battling home-grown Iraqi units, as opposed to American soldiers? I think this is an open question that\'s due to be answered in the next few months.   '),(3612,3992688,'03,August,2004','2004-08-03 00:00:00','    The Scotsman reports today on the threat of a North Korean missile  urlLink capable of reaching the United States .   \"A report carried out for the publication Janeâ€™s Defence Weekly said the two new systems appeared to be based on a decommissioned Soviet submarine-launched ballistic missile.\"      \"Ian Kemp of Janeâ€™s Defence Weekly, said North Korea would only spend the money and effort on developing such missiles if it intended to fit them with nuclear warheads.\"     The major threat comes from the sea-based launch system which, if mounted on submarine, would be able to launch from anywhere in the ocean and thus in theory be able to hit the United States. Of course the United States navy has been geared since the beginning of the cold war towards hunting down ballistic missile submarines and thus preventing this very thing, and the technological gap between N. Korea armed only with an old Soviet sub and the United States navy is considerable. However I do think it makes clear that North Korea\'s ultimate goal is to be able to threaten the United States with nuclear weapons, and makes the matter of addressing North Korea\'s nukes all that more urgent.   '),(3613,3992688,'03,August,2004','2004-08-03 00:00:00','    Much of the information that led the authorities to raise the terror alert at several large financial institutions in the New York City and Washington areas Sunday was  urlLink three or four years old , according to intelligence and law enforcement officials. They reported that they had not yet found concrete evidence that a terrorist plot or preparatory surveillance operations were still under way.      One piece of information on one building, which intelligence officials would not name, appears to have been updated in a computer file as late as January 2004, according to a senior intelligence official. But officials could not say yesterday whether that piece of data was the result of active surveillance by al Qaeda or came instead from information about the buildings that is publicly available.       \"There is nothing right now that we\'re hearing that is new,\" said  urlLink one senior law enforcement official  who was briefed on the alert. \"Why did we go to this level? ... I still don\'t know that.\"      It seems highly suspicious that years-old intelligence, with little to no indication of being relevant now, would be cause for the terror alerts we saw - unless, of course, the motivation wasn\'t public safety but rather political. I mean, it\'s not like it is the first time dated intelligence has been used for political gain...      UPDATE: The   urlLink  same terror alert    was reported on over two years ago.   '),(3614,3992688,'02,August,2004','2004-08-02 00:00:00','    Newsweek  urlLink reports  Rumsfeld maybe rebuked by his own commission on the prison abuse scandal, while Rolling Stone  urlLink details  earlier reports of rape and sodomy at Abu Ghraib. David Brock demands  urlLink equal  air time for liberal group spokesman in a letter to CNN\'s Ken Tucker. The Bush camp  urlLink solicits the race  of a Arizona Star staffer assigned to photograph Dick Cheney. Meanwhile, someone gets a peek at what the Bush campaign\'s internal polling  urlLink may look like . Also, the Niger \"yellow cake\" claim gets even more  urlLink complicated . And finally, the Center for American Progress analyzes the  urlLink many flip-flops  of George W. Bush.      UPDATE: Iraqi jail chief says prisoner abuse  urlLink covered up .   '),(3615,3992688,'01,August,2004','2004-08-01 00:00:00','    As reported by  urlLink Electoral Vote Predictor 2004 :      \"A group of Colorado citizens have proposed a change to the state\'s constitution specifying that Colorado\'s nine electors be apportioned strictly in proportion to the popular vote. Currently Bush is ahead 48% to 43% there, so under the proposed system, Bush would get five electoral votes and Kerry four electoral votes, instead of nine to zero. The group has turned in petitions containing 130,000 signatures. If about 68,000 of these prove to be valid, the question will be a ballot referendum in November. If it passes, the change takes effect for this year\'s election. If it makes the ballot, on the evening of Nov. 2, the TV news anchors will probably be saying: \'President Bush won Colorado with 55% of the vote, but we don\'t know how many votes he will get in the electoral college until they finish recounting the closely fought referendum on changing the Colorado state constitution.\' Whoever loses will appeal to the U.S. Supreme Court, which once again may have to rule on the sensitive issue of state\'s rights. To learn more about what may be the sleeper issue of the year, start   urlLink  here   .\"    The article points out that if this had been done this way in 2000, Gore would have won the election. Bush is leading Colorado right now, but is only a few points ahead like last time, so even if he won Kerry would also get a significant number of electoral votes, lessing Bush\'s chances of winning the election and increasing Kerry\'s.   '),(3616,3992688,'31,July,2004','2004-07-31 00:00:00','   Hey guys, this darwin award is pretty funny.   urlLink www.darwinawards.com/stupid   Click on Lawn Chair Larry.  '),(3617,3992688,'30,July,2004','2004-07-30 00:00:00','   So I\'m talking to Liz last night on the phone, and she tells me to hold on for a second. I say \"alright\" and when she comes back she says: \"sorry, I had to put some water in my bong.\"  Also, I had to remove the giant picture of the guy with all the piercings, it really freaked out the blog template so that it put all the links at the bottom instead of the side.  And no the picture below is not part of this post!  '),(3618,3992688,'28,July,2004','2004-07-28 00:00:00','   urlLink    A typical cell phone user at North Lake  So I come into work this afternoon, and see sitting out in front of the circulation counter a large fishbowl with green aquarium rocks in it and what appear to be one or two cell phones, on which is taped a large sign that says \"Please turn off cell phones...or deposit them here!!!\" Now, most people are not going to want to leave their cell phone in a bowl and out of their reach, let alone in a bowl mixed with other strange phones carrying God only knows what diseases. So one would be led to believe the implication is that we wish you to turn you cell phone off before entering. BUT...that\'s not what the sign says. Clearly, there are two options, including the option to leave the cell phone in the bowl...with no requirement that it be turned off before it is left in the bowl. So you can, in theory, leave your cell phone in the bowl  turned on . And being as this is a public library I can guarantee that at some point some patron will leave their cell phone in the bowl,  turned on . And naturally, that someone will try to call this clever patron while their phone is in this bowl of phones. And that, upon hearing it ringing, the circulation staff member currently occupying the counter will rush to the bowl and begin cursing frantically as they dig through phones and aquarium rocks to determine which of the phones in the bowl is ringing.  In truth I am not entirely opposed to this plan, as I welcome all arrangements that produce amusement in their absurdity, and the failure of the arranger to be aware of all natural outcomes of their action. It is my fondest wish to see this particular outcome before I leave.  '),(3619,3992688,'28,July,2004','2004-07-28 00:00:00','   urlLink     The new face of North Lake Library?&nbsp;&nbsp;   For those of you who do not already know, the dream that was the North Lake Community Library is coming to an end. Those brave and dedicated IPL employees that toiled, sweated and cried for the dream of a united Academic/Public library will be scattered like ashes to the wind among the various branches of the Irving Public Library system, broken and allotted like slaves at the auction block under the cruel&nbsp;Alabama sun. &nbsp;Those books which once stood so proudly on our shelves, virtually shouting their vast reservoir of knowledge to our patrons(student and public)will be divvied up like so much pirate treasure booty among the branches of IPL. The children\'s library, the gleaming jewel in the eye of knowledge that was NLCL, will close, it\'s lights dimmed forever, no longer able to battle against the ignorance that today runs rampant in the world of the child. Instead what will be left will merely be a shell of its former self, as North Lake college employees take over the many duties they once shared with their comrades-in-arms from IPL, and slave to provide at best 1/2 the service to 1/4 the patrons at 1/8 the level of competence(albeit at 2/3 the price.) Please take a moment to lament the passing of the dream that was NLCL. Alas, poor NLCL, we knew you well...  '),(3620,3992688,'27,July,2004','2004-07-27 00:00:00','   I made some adjustments to our blogsite.  1.)&nbsp; The time-stamp now shows the correct timezone.  2.)&nbsp; The posts will only cover the last 5 days, you can see the rest under the Archives.  '),(3621,3992688,'27,July,2004','2004-07-27 00:00:00','   Women Say Dentist Made Them Swallow Semen  1 hour, 13 minutes ago Add  urlLink Strange News - AP  to My Yahoo! CHARLOTTE, N.C. - Former employees of a dentist claim the man made female patients unwittingly swallow his semen during visits to his office. Dr. John Hall is accused by the state dental board of violating dentistry\'s standard of care, engaging in immoral conduct, and committing sexual assault or battery.   Six former patients say he tricked them into swallowing his semen. A seventh claims he jumped on top of her in the dental chair and \"began to gyrate against her lower body in a sexual manner.\"  The dental board suspended Hall\'s license Nov. 5, after police began investigating allegations by two former employees. In February, it conditionally restored his license, barring him from being alone with female patients.  Hall\'s lawyers say the dentist has been falsely accused by disgruntled former employees. He was collecting his semen because he was taking Propecia, a drug to promote hair growth, and was concerned about potential side effects that include low sperm count and diminished semen, his lawyer, Emerson Thompson, said.  The dental board heard testimony Friday, Saturday and Sunday. The hearing is to continue next weekend.  The five-member panel will decide whether Hall can keep his license. The case is still being investigated by authorities in Charlotte, Cornelius and Mecklenburg County.  Hall\'s former assistant Cheryl Lynn MacLeod testified Saturday that she found it odd that Hall began asking her to leave a patient\'s side to retrieve items that he never used.  She also became suspicious when she overheard him tell patients to \"swallow\" something, and when she saw him take a syringe out of his lab coat pocket while working beside a patient.  In May 2003, she and former office administrator Susie Hillman took the five syringes from office waste containers and Hall\'s lab coat pockets and gave them to police.  Hall\'s lawyer, Thompson, suggested the substance Hall asked patients to swallow could have been a dental product.  But in a sworn deposition read aloud Saturday, Hall\'s former dental assistant Rhonda Hamilton described several dental products used in Hall\'s office and none matched the patients\' descriptions of what was put in their mouths.  When Cornelius police searched Hall\'s office Nov. 5, they confiscated five more syringes with Hall\'s semen from his desk, according to the board and hearing testimony.  Five of the six patients testified Friday. The first, a 40-year-old woman, said she objected when Hall began to put a substance from a syringe into her mouth during a procedure on Sept. 4, telling him: \"That smells like sperm.\"  She said Hall drew back and said, \"You\'re crazy.\" She said she couldn\'t taste the substance because her mouth was numb.  Another witness, a 33-year-old woman, said Hall asked his assistant to retrieve something from another room during her appointment in May 2003. While the assistant was gone, she said Hall told her to lie back, open her mouth and swallow.  \"When I swallowed I tasted it, and it was semen. ... He told me it was cleaning solution.\" She said she drove directly to her husband\'s office nearby and told him what happened, but he and a colleague dismissed the idea. She said she tried to convince herself she was wrong, and contacted police only after hearing about the investigation on the news.  '),(3622,3992688,'27,July,2004','2004-07-27 00:00:00','   Alright everyone, I have a new library patron to tell you about.  An African woman comes into Northwest from time to time to get books, and her name is Princess Bempong.&nbsp; Yes that\'s right, the system has her listed as Princess Bempong (not sure if that\'s a title or a name, but either way...).&nbsp; She often comes in with her deaf servant who makes extremely loud, yet indiscernable, noises constantly.&nbsp; Sometimes grunts and sometimes groans, and sometimes just hollering.&nbsp; Here is the usual scenario:  Princess Bempong will come up to the counter and will make her servant ask for the books (which of course means just a lot of loud and embarrassing noises).&nbsp; Since no one can discern this fellow\'s gestures or sounds, she gets frustrated and signs with him for a minute or two (while he\'s still grunting and groaning) and then finally writes down what she wants, and makes him hand it to us.&nbsp; At this point we can usually help them, since what she writes down is an easy enough request.&nbsp; But what I don\'t understand is why Princess Bempong doesn\'t just tell us what she wants...you see I know that she can speak because I\'ve spoken to her on the phone and in person on more than one occassion!  I guess she just wants to make full use of her servant...who knows...  '),(3623,3992688,'27,July,2004','2004-07-27 00:00:00','   urlLink    I think he\'s dead!&nbsp; urlLink      Check out this site,  urlLink Simulaids ,&nbsp;that offers aids for emergency and medial personnel to learn on. There\'s some truly frightening stuff on there...  '),(3624,3992688,'27,July,2004','2004-07-27 00:00:00','   Here\'s a random trivia question for you all.&nbsp; Does anyone know what   \"the mystery\"   is, according to Paul of the New Testament?&nbsp; He mentions it quite a bit, and expounds on it some each time he does.&nbsp; Its nothing too tricky, I was just curious if any of you were familiar with it.&nbsp; Comment away....!  '),(3625,3992688,'26,July,2004','2004-07-26 00:00:00','   So here  \'  s the question at hand:      Do   I  , or do   I   not  ...  .paint my tuck somewhat similar to this?  '),(3626,3992688,'26,July,2004','2004-07-26 00:00:00','   Celebrating The Horadrim\'s 50th blog:    urlLink \"Inter-galactic Traveller\" Mystery Woman Safe in Hawaii     urlLink Man Found Dead on Hospital Lounge Couch     urlLink Accused shooter saw UFOs     urlLink Vietnam \"ghost\" killer gets 11 years in jail     urlLink Lightning Strike Energizes Man     urlLink Wisconsin man injured while flipping off trains     urlLink Researchers study \'Pokemon\' seizures  '),(3627,3992688,'25,July,2004','2004-07-25 00:00:00','   urlLink     Get a load of this  urlLink article .  Guess it kinda proves Italians are really different.  Or I guess it means you have to be extraordinarily fat to get your 15 min of fame.    But I finally got a picture posted!  '),(3628,3992688,'25,July,2004','2004-07-25 00:00:00','   This is a test post from  urlLink   , a fancy photo sharing thing.  '),(3629,3992688,'25,July,2004','2004-07-25 00:00:00','   urlLink http://www.joynk.com/cdg/   I stumbled across this site today in my daily rummaging through the internet.&nbsp; It\'s some web designer who gets persistent, weird calls from a drunk man while he\'s trying to work.&nbsp; My personal favorite is number 2.  '),(3630,3992688,'25,July,2004','2004-07-25 00:00:00','   For those of you who have been to my house, you can probably appreciate the comedy of this post.&nbsp; You have probably seen the absolutely enormous plant in our backyard and grows freakishly larger every week.&nbsp; It looks like this:   &nbsp;   urlLink        &nbsp;   &nbsp;   &nbsp;  And yes the swingset next to it is naturally dwarfed.&nbsp; Anyhow I was out in the backyard today and noticed that the top of our freakish plant was covered in seed-pods.&nbsp; I went and got my brother to show him how much bigger this plant had gotten and he had a brilliant idea.&nbsp; Here is a problem at our house:&nbsp; Mowing the \"garden\" (a plot of land overgrown by a furious variety of thick and persistent weeds).&nbsp; We have come upon a solution:&nbsp; taking these seed pots and covertly planting them all throughout the garden so that suddenly it will look like this:    urlLink       &nbsp;  &nbsp;  &nbsp; No more grass or weeds = no more mowing.&nbsp; Won\'t our mom be surprised when she looks out the back window one day and sees a veritable jungle of asian trees...  '),(3631,3992688,'25,July,2004','2004-07-25 00:00:00','   urlLink     I\'d like to tell you something....  There was this guy.&nbsp; He ate his  SHIT!   You see, I have another theory about how humanity can be subdivided, which I came up with in the 11th grade. It involves their reaction to the following statement said in the following way:  \"There was this guy (said just a little bit raunchily, but not too much)...(dramatic pause of about a second)...He&nbsp;ate his shit (extend the word \"shit\" and phrase the sentence in clear iambs.&nbsp; It helps if you raise your eyes a little as you proceed and allow your voice to become raspy.&nbsp; The whole process should flow like a gathering storm).&nbsp; People are guaranteed to react in one way or another.&nbsp;   Type A: Type A will laugh and always laugh.&nbsp; Sometimes they will find it funny, other times strange, but always amused.   Type B: Type B reactions will vary greatly, but the test subject will never laugh.&nbsp; They may look quizzical, vaguely disgusted, annoyed by your profanity, or presumably other, undocumented results.&nbsp;   I have my theories about what exactly being a type one or a type 2 means--perhaps one type is more able to find humor in a cruel world, better adjusted to hearing descriptions of disturbing behavior, or they themselves have been there.&nbsp; I encourage you all to test this hypothesis on your loved ones,&nbsp;and&nbsp;guarantee&nbsp;that you will be surprised by how they fall out by type .&nbsp;  '),(3632,3992688,'25,July,2004','2004-07-25 00:00:00','   As most of us know, our good friend Jason Asimikas has gone off to fight for his country.  Who knows what crazy adventures he\'s in store for.  Hopefully, he won\'t get an administrative discharge because of drug abuse like another good friend of ours.      Anyway, I was doing some \"web surfing,\" as it\'s called, and stumbled upon this  urlLink book  that pretty much sums up, I believe, what Jason is going through.  Hell, maybe the next volume will include one of Jason\'s stories.    OOH-RAH!  '),(3633,3992688,'24,July,2004','2004-07-24 00:00:00','   Kann jemand, dieses zu lesen? Gut dachte ich, daÃŸ ich versuchen wÃ¼rde, auf Deutsch bekanntzugeben. Ich schÃ¤tze, daÃŸ dieses ein Test fÃ¼r Daniel ist. Daniel kann Sie verstehen, was ich sage? Jeder sollte zum undersand in der LageSEIN: Ja und ScheiÃŸe. Ich verwendete Sherlock, um diesen deutschen Text festzulegen. Ich wundere mich, wie genau es wirklich ist. John und ich kamen gerade von der Penthouse-Verein Ã¼ber der Datenbahn von zurÃ¼ck, der ich gehe, fÃ¼nf oder sechs Tage ein Woche zu bearbeiten. Wir erhielten dort und sie schloÃŸ in 10 Minuten. John erhielt einen SchoÃŸtanz von einem heiÃŸen KÃ¼ken, aber ich erhielt nichts. Ausgenommen ich handhatte, Zwanzig Dollar zu sichern, von denen ich jetzt auf pornographischen Materialien verwende. Gut ist es spÃ¤t und ich bin mÃ¼de. SpÃ¤ter.  '),(3634,3992688,'24,July,2004','2004-07-24 00:00:00','   Alright, this isn\'t offensive in the same way tubgirl or goatse are, but it\'s pretty bad.    urlLink God loves you  '),(3635,3992688,'23,July,2004','2004-07-23 00:00:00','   urlLink     Hail loincloth!&nbsp; urlLink      This is the cover of my band\'s only album. I\'m obviously the dude with the big mustache. I was in a lot better shape back then.  '),(3636,3992688,'23,July,2004','2004-07-23 00:00:00','   urlLink    A little accident&nbsp; urlLink     This is what we do to library traitors.  '),(3637,3992688,'23,July,2004','2004-07-23 00:00:00','   urlLink     John and Daniel&nbsp; urlLink      So what the heck were you guys doing here?  '),(3638,3992688,'23,July,2004','2004-07-23 00:00:00','   For those of you who have encountered Meredith\'s strange form of reasoning in the past, this will come as no surprise.&nbsp; And I hope that Nathan can add to this post with his own tales.&nbsp; For those of you who have never dealt with Meredith, let me give you a little example.&nbsp; I pick this example only because it is the most recent:  The other night Patricia was discussing with Meredith the idea that we should count people who go through the wrong side of our entrance/exit gate using a stat-clicker.&nbsp; Meredith then says, \"oh yes, and we should add 25 to that count for the Junior Players who come through the back door and never go through our gate.\"&nbsp; So Patricia says, \"ok we\'lll start our daily count at 25\", and Meredith then says: \"but its easier if we just get someone to walk through the gate 25 times instead of adding it to our count.\"&nbsp; ...&nbsp; ...&nbsp; ...  So apparently its easier to walk through a gate in a circle 25 times than simply adding 25 to some other number.&nbsp; Right...  '),(3639,3992688,'23,July,2004','2004-07-23 00:00:00','   urlLink     Softcore lesbo sex intermingled with scifi parody.&nbsp; urlLink       So last night I watched the last half of \"Play-mate of the Apes\" on Cinemax, which I had heard about from a friend. Thankfully the porn aspect is all&nbsp;lesbian scenes (female astronauts and slaves getting it on), and no beastiality.&nbsp;However, to put it simply, this film is insane. First of all, they really just used whatever&nbsp;apes costumes they could find as opposed to one uniform costume for&nbsp;all the apes. This&nbsp;includes&nbsp; a pink ape, whom is gay. &nbsp;The \'alarm\" that goes off at one point is simply a police siren plugged into a wall. The sign pointing to the Forbidden Zone is a computer generated arrow saying \"Forbidden Zone.\" The acting was bad not in the typical porn way, but as if the actors were actually trying to say their lines in the craziest way possible. And, for some reason, at several points in the film the cast of characters suddenly bust a rhyme. But anyway, it certainly got the job done and I recommend it.  '),(3640,3992688,'23,July,2004','2004-07-23 00:00:00','   urlLink       Man down!&nbsp;  urlLink         I shall not say I told you so, only that I didn\'t vote for I, Robot. However, James wanted to avoid crowds so we didn\'t go see the Bourne Supremacy at a major theater, we went to see I, Robot at the Movie Tavern in Bedford. Now I didn\'t think this was a bad idea either, just not what I\'d prefer. The ill omens began almost as soon as the four of us (James, Amand, Kristina and myself) left Amanda\'s apartment. James took us down 121 to get to 183 where there happened to be an inordinate amount of traffic. However, we did manage to get there on time anyway. We purchased our tickets and headed into the theater. Our second warning was that the women\'s bathroom was closed due to some&nbsp;malfunction. The ladies had to temporarily take over the men\'s room. Not a severe problem. James and I went to find seating for the five of us (Lavonne had joined us). We spotted what seemed to be the ideal location. Now to give you an idea of the layout, this is one of those movie theaters where the attendants come serve food in the theater and the rows of chairs all have little tables in front of them. So we spot a good location over on the left side of the theater. Being on the right side, we began to walk through the row to get over to the left side. Tragedy! As James put his foot down just after passing the first seat, he slipped. I swear, it was almost like watching an act. First his foot went one way, then his foot went the other way, and after struggling valiantly for as long as a second to maintain his balance, James went down! He fell; his left knee took the brunt of the damage. However, he fell in a&nbsp;most spectacular&nbsp;manner. James tilted slightly to the left and came down hard on a cupholder armrest. I kid you not, the cupholder exploded! I\'m serious, I saw shards of plastic flying in the air. James quickly got up, but he was obviously wounded. His left foot was hurting and he had a scrape on his knee. James bravely walked over to our chosen seats, eschewing offers of help and protesting that he was ok. Still, when the manager came he did accept a bag of ice. I should point out that the staff of the theater neither covered the seats sharing the broken armrest nor did they mop the floor. Anyhow, with the apologies of the manager, we did temporarily move past this event. Inexplicably, the manager offered James  no compensation  for this mishap! Still, we decided to tough it out. We went ahead and ordered. Kristina and I ordered cheese sticks and a pepperoni pizza. Then the previews started rolling. It was the preview for the new Jet Li movie, but without the sound. We waited, as this sometimes happens, but no sound was forthcoming. But wait! We heard some kind of sound, shortly before we saw the Pepsi advertisement come up  simultaneously  with the preview! It was also playing the music that had been playing before the preview! It took them another preview to get this problem fixed. Finally, when the sound was fixed and the last trailer shown, the movie started rolling. But wait: where\'s the food? Yes, although the entire rest of the auditorium had been served, we waited. It only took our man another 10 minutes to get there with our order. Of cheese sticks. No pizza ever came our way, although initially we were charged for both. We convinced the server of the error of his ways though. Finally, after the movie was all over we went out of the theater where James and the manager wrote up an incident report (whatever they call it there) and I signed it as well, being the only witness of the event. We left without further event, although I suspect more would have happened if we\'d stayed. There was an ominous and ridiculous looking cop there in a huge white Stetson. And that was our night at the movies.  '),(3641,3992688,'23,July,2004','2004-07-23 00:00:00','   Before I crash for the night and get up early to catch the rat race, I thought I\'d share some things I\'ve learned over the years...  In no particular order:  1.  Starburst Twizzlers and Lime-flavored chips are a recipe for disaster.  2.  Dominoes buffalo wings after a heavy night of drinking is also a none-too-smart idea.  3.  Some milkshakes do not bring boys to the yard.  4.  Apples taste better than Windows.  5.  Never take a picture in a Tijuana strip club.  6.  Daniel is the only one in the world with an interest in forum browsing (click on it in profile if you don\'t believe me).  7.  It is possible to accidentally uninstall programs from a computer and have an uncle that can do anything.  8.  Expect the unexpected, unless your Nathan, in which case expect a life long struggle with the toilet.  9.  Transos are people too.  10.  FFUSFOSTAR  Thank you and good night!  '),(3642,3992688,'23,July,2004','2004-07-23 00:00:00','   I have now joined the blog.&nbsp;   As I write this, John is touching a statue\'s penis.&nbsp; In fact, David\'s penis (not to be confused with me).&nbsp;   Michelangelo would be proud.  '),(3643,3992688,'23,July,2004','2004-07-23 00:00:00','   I just saw the Bourne Supremacy.&nbsp; It\'s a very good movie, I highly recommend it.&nbsp; And on my way back Nathan must have called, I got his voice mail and called back the number he left, but to no avail...  Nathan if you\'re out there...no I will not make out with you...  '),(3644,3992688,'23,July,2004','2004-07-23 00:00:00','   urlLink    I\'m playing with Hello. This is my first picture.&nbsp; urlLink  '),(3645,3992688,'22,July,2004','2004-07-22 00:00:00','   urlLink    Too bad his sons bring on more shame than respect to his average American life.&nbsp; His only consolation is love from his golden retriever who lacks any vestige of discipline or self control.  '),(3646,3992688,'21,July,2004','2004-07-21 00:00:00','   urlLink    You know, I was editing my profile today, and I got to wondering why exactly my favorite stuff is underlined. Well, I clicked on some of them and, much to my dismay, I discovered that it shows you all of the people who share this interest. I think if we put our heads together , we can figure out who should be the most ashamed of his interest/favorite book etc. I think my interest in the book Foucault\'s Pendulum puts me in with a special class of high-powered mutants.  '),(3647,3992688,'21,July,2004','2004-07-21 00:00:00','   urlLink Find, Compare, and Buy 9 11 commission Books at Shopping.com   And while you\'re at it, buy me one too...  '),(3648,3992688,'21,July,2004','2004-07-21 00:00:00','   For some reason, weird stuff happens a lot when I\'m over at the Adams family house.&nbsp;This story happened on one such occasion.    &nbsp;  It was rather late at night and I was sitting on the futon playing some Mario Kart Double Dash (awesome game). John went into some Christian&nbsp;chat room to have some fun. He decided it would be cool to put on the webcam and offer images of me playing Mario Kart on for whoever in the chat room wanted to see.&nbsp; So some guy logged on and IM\'ed John.&nbsp; I&nbsp;forget what it was that the guy and John initially wrote, but shortly thereafter the guy wrote, \"Show me your ass.\" I kid you not. John asked me if I wanted to show the guy my ass, but after great consideration I declined. Anyway, I thought that if the guy wanted to see ass, he might like tubgirl. John IM\'ed him the link. After a few seconds, the guy wrote back, \"That\'s hot, show me some more.\" I may not be quoting exactly, but that\'s close enough. I forget what John did after that, but I was freaked out enough for one night.  '),(3649,3992688,'21,July,2004','2004-07-21 00:00:00','   Alright guys, we\'ve all seen this all to many times:  Like just now, for example.&nbsp; A large (or rather staggeringly enormous) woman (or man I suppose) waddles in to the library, goes and finds some books.&nbsp; (Perhaps resting a few times on the way, sweating and panting disgustingly as they go).&nbsp; They finally make it back to the checkout desk and slam their books down exclaiming: \"golly I\'m tired...\" or \"whew I\'m exhausted\" or some such other statement.&nbsp; And what causes all this exhaustion you might ask?&nbsp; Simply walking across a small branch library...what is the world of obesity coming to...  '),(3650,3992688,'21,July,2004','2004-07-21 00:00:00','   urlLink     Lovely Ladies \"A\" and \"K\"&nbsp; urlLink      Despite what you may have seen on TV, the library business is not entirely a male-dominated profession. Beside the many men slaving away here at IPL to answer your questions and find your books, many a hard-working woman works. Two of them, our personal favorites, are pictured here. \"A\"(on the left) is one of our resident children\'s librarians, whose job it is mostly to keep the little buggers out from under our big, manly stomping feet by entertaining them with stories, songs, crafts or by mostly allowing them to sit on the computers and download Dragonball Z and Yu-Gi-Oh pictures in black and white. \"K\" is entrusted mostly with returning many a valuable book to the shelves on which they belong, so that potential child-molestors have something to peep from behind when searching for underage girls who themselves are desperately trying to study and pretend not to notice the weird old guy checking them out from the gardening section. Yes, I daresay without their female-ness&nbsp;our libraries would be hardly as pleasant to work in.&nbsp; Thank you \"A\" and \"K\".  '),(3651,3992688,'21,July,2004','2004-07-21 00:00:00','   John has been imploring me to tell the story of the night I almost blacked out on his toilet. I don\'t know that there\'s too much to explore here, but I\'ll give it to you.   A couple of weeks ago I was hanging out one night at John\'s place. A natural urge came upon me so I went to the bathroom. Something was wrong though. As I sat there, allowing the natural course of events to occur, I suffered from what felt like an intense head rush. My vision narrowed down and I began to lose feeling in my body. I thought perhaps it was normal, but no! I could only see a small patch of wallpaper with my tunnel vision, but I knew this was no ordinary bm. I had to stop it or it might kill me. So I grabbed the rim and tried to slow the process down, but I felt like a truck was driving through me. I realized that strategy wouldn\'t work, so I accelerated the process. The effort it cost me was immense, and I was shaking and sweating. I only survived because I went empty before I totally blacked out. Still, at the end I was shaken by a violent convulsion and I fell back against the tank of the toilet, sending a candle holder to meet its doom on the floor. For a few moments, I had to sit there just resting. That was the most effort I\'ve ever had to put into an evacuation.   I still don\'t know why I had that one particularly bad episode, but all I can say is I hope it never happens again.  '),(3652,3992688,'21,July,2004','2004-07-21 00:00:00','   You guys check this store out; don\'t worry, you can view it at work.   urlLink Worship the monkey  '),(3653,3992688,'21,July,2004','2004-07-21 00:00:00','   urlLink     You sit down for a relaxing evening of computer use, but what\'s this? Somebody wants you to go to a website? Hmm, www.tubgirl.com, eh? Sounds enticing and sexy, doesn\'t it? OK, I\'ll bite, just type it in... Oh dear God, what is that!!!!, you exclaim, closing the browser in horror.&nbsp;&nbsp; Although you can purge your IE cache to expunge any physical traces of your visit after the shaking stops, you&nbsp;will never&nbsp;get Norton Systemworks to install inside your tortured mind.  If this scenario sounds familiar, then chances are you were, at one point in your life, tubgirled.&nbsp; Although some people are ultimately able to laugh at their trauma, others never quite seem to be the same after viewing that enigma wrapped up in a mystery wrapped up in a brown geyser.&nbsp; I have noticed two general reactions to the tubgirl phenomenon, which I will dub \"A\" and \"B\".   TYPE A: A type \"A\" tubgirl victim&nbsp;was initially horrified and confused by what they saw, and, in fact, may not be entirely certain what they just looked at.&nbsp; However, any morbid curiosity on their behalf will be quickly swallowed up by one single, all-consuming mania: \"Who can I show this too?\"&nbsp; My theory is that a type \"A\" tubgirl victim can only survive what thy have just seen by passing its unholy&nbsp;e-miasma along to another victim.&nbsp; A good analogy to help you understand this would be the movie  The Ring .&nbsp;   TYPE B:&nbsp;Statistical analysis of a sample of&nbsp;type \"B\" tubgirl victims reveals that most of them are exposed to the phenomenon through a type \"A\" acquiantance.&nbsp; The type \"B\" victim\'s symptoms display themselves similarly for the first few minutes, and they may be confused with type \"A\" symptoms by the casual observer.&nbsp; However, the type B victim never feels the need to pass the burden of seeing along.&nbsp; Instead, they feel a need to understand who, how&nbsp;and why tubgirl is pressing down on them like a crushing weight.&nbsp; Long-term morbidity data is forthcoming.   My research into this matter has caused me to speculate about the existence of type \"C\" victims, for want of a better term, who do not seem to be affected by the TG miasma.&nbsp; My best working hypothesis is that these \"victims\" are not fully human and therefore immune to the condition; however, I can not state this with confidence until I have performed several comparative autopsies.&nbsp; In this matter, I can only hope that Seamus\'s girlfriend Stacy will come through for me.&nbsp;   In conclusion, to all you type A\'s out there, rest easy--&nbsp;the danger has already passed.&nbsp; Type B\'s, you are in my prayers, and we are hard at work on a cure for your condition.&nbsp; Type C\'s, if you really exist, please come forward.&nbsp; And finally, for those out there who have seen no trauma gushing forth as a fountain&nbsp;&nbsp;as of yet-- if, one day, you see the tub loading on a computer screen, avert your eyes and save your soul!!  '),(3654,3992688,'21,July,2004','2004-07-21 00:00:00','   urlLink    Authentic Nat-Wu?&nbsp; urlLink  '),(3655,3992688,'21,July,2004','2004-07-21 00:00:00','   Here it is, of which we speak. What poor research assistant had to dig up all  urlLink this ?  '),(3656,3992688,'21,July,2004','2004-07-21 00:00:00','   hey all, as most of you already know i am part of the nintendo street team. anyway we are in a competion with the other street teams across america doing different contests each week. You can go to www.nintendostreetteam.com and vote for DALLAS, by voting you can win 5000 dollars too, so vote for us we need it. we also made our own website yall can check out if you want with pictures and stuff from our promotions www.streetteamdallas.com  '),(3657,3992688,'21,July,2004','2004-07-21 00:00:00','   Well whoop-de-doo...you can now look at me looking back at you...  '),(3658,3992688,'21,July,2004','2004-07-21 00:00:00','   urlLink    Disturbing?  '),(3659,3992688,'21,July,2004','2004-07-21 00:00:00','   urlLink    Here are some good photos for old time\'s sake  '),(3660,3992688,'21,July,2004','2004-07-21 00:00:00','   Many of you have e-mailed me to ask me why I choose the pseudonym Xanthippas. Well,  urlLink here\'s  a little more about him. In the annals of history\'s great face-offs, I\'ve always favored Carthage over Rome and Sparta over Athens, which makes this little guy my hero.   &nbsp; Oh, and historian Donald&nbsp;Kagan spells his name with an \"a\", but then he\'s British...  '),(3661,3992688,'21,July,2004','2004-07-21 00:00:00','   urlLink    Here goes nothing. You will note that I have blurred the face to preserve this person\'s identity  '),(3662,3992688,'21,July,2004','2004-07-21 00:00:00','   Somebody besides me figure out how to get a photo next to our profiles  '),(3663,3992688,'21,July,2004','2004-07-21 00:00:00','   &nbsp;          &nbsp;         http://www.msnbc.msn.com/id/5469808/?GT1=4244'),(3664,3992688,'21,July,2004','2004-07-21 00:00:00','   As I was playing with this last night I followed some links and found the Hello software that you may or may not have already seen.&nbsp; Its basically&nbsp; just a little program that lets you easily post images to your blogs.&nbsp; That\'s how I posted Mr. T.&nbsp; I think it also lets you post regular blogs, but it makes it much easier since it does it in an IM format, you just type up your blog and hit Send, you don\'t have to be logged in on any webpage.  You can get this software from  urlLink www.hello.com   You can also make blogs via email, I haven\'t done this yet but I know its possible.&nbsp; You just email the blog site and it posts what you send.  If things go as I predicted originally, we should have a flood of disturbing and funny images in no time...  Alex can post goatse for us, I think he knows the address for that image  '),(3665,3992688,'21,July,2004','2004-07-21 00:00:00','   Women. I tell ya. This girl who hasn\'t talked to me in months sends me a message out of the blue about what textbook she needs. I respond sarcastically about liking being the go-to guy for school info, and she acts like she doesn\'t know what I\'m talking about. What a b*tch!    &nbsp;  Anyway, this is a cool place.&nbsp; I don\'t work for any of the Irving libraries but I seem to know a good amount of people from there and what\'s been going on.    &nbsp;  It\'s kind of nice out here because it is cloudy.    &nbsp;  Hey, you ever notice the word \"believe\" contains the word \"lie?\"    Ah, my back itches...    &nbsp;    &nbsp;  (Yes, I am trying to make fun of the typical blog entry.)  '),(3666,3992688,'21,July,2004','2004-07-21 00:00:00','   Welcome to my first post! In a short 45 minutes I have gone from viewer and commentator on this blog, to fellow blogger. I also have a blog of my own that I created 10 minutes ago, which I will share with you when I figure out how to.&nbsp; I\'ve been invited by virtue of the fact that I too am a library employee, if only for another two weeks, after which I will become an unemployed law student. Still in that time I will cram in as much joyful library experiences as I am allowed by my fellow bloggers.  '),(3667,3992688,'21,July,2004','2004-07-21 00:00:00','   I just thought I\'d share a picture of myself with everyone. I am Nat-Wu, the  urlLink Chinese giant !  '),(3668,3992688,'21,July,2004','2004-07-21 00:00:00','   I was able to waste the first hour pretty well, especially since there were no ready carts, and some kid had played the push-all-the-books-in-that-you-can-reach game on the non-fiction shelves.&nbsp; So I fixed that kid\'s damage, wandered a bit, and now I\'m on desk until 4.&nbsp; It would be interesting to post images of the library mutants on this page...and hideous.&nbsp; I think I would start with Mironack.&nbsp; And then Guillauma the Frenchman.  '),(3669,3992688,'21,July,2004','2004-07-21 00:00:00','   urlLink http://www.colonialfsc.com/MASTERS/more%20photos.html  '),(3670,3992688,'21,July,2004','2004-07-21 00:00:00','   I\'m very glad to see that Nat-Wu has joined up.&nbsp; Didn\'t you like my picture of Mr. T?  I just got in to work and we have Vanessa helping us out.&nbsp; (Not sure if I spelled her name right.)&nbsp; Anyway today I feel like I should be lazy to compensate for all the work I did yesterday while Lorenza happily did absolutely nothing.&nbsp; Even Arthur (who was here yesterday) shelved a cart, and still Lorenza did nothing.&nbsp;   Anyhow...this is what Google images pulls up when you search for Nat-Wu:   urlLink http://www.colonialfsc.com/ MASTERS/Nat_WU_circle.jpg  '),(3671,3992688,'21,July,2004','2004-07-21 00:00:00','   Aren\'t blogs for losers and people who want everyone to read their crazy conspiracy theories? But anyway, I like that story about the texshare card. That\'s crazy. And what about the crazy guy trying to kiss your hand? Oh, and as far as interesting stories go, I only have a short one. Last week I was in my library one night after we\'d closed (having returned to use the phone). It was rather dark but I noticed what appeared to be a huge roach crawling up the wall over the Spanish books. I got the flashlight and indeed, it was a huge roach. I decided the only thing to do was use it for target practice and shoot at it with rubber bands. I had shot off several and was on the verge of giving up, but I decided to go for one more. I got the biggest rubber band out of the box that I could find (and I\'m talking an 8\" diameter here). Amazingly enough, I hit the thing! Unfortunately, I discovered that the rubber band wouldn\'t do a significant amount of damage. The roach got away. Perhaps I need larger rubber bands.  '),(3672,3992688,'21,July,2004','2004-07-21 00:00:00','   Mr. T and SeanBaby    urlLink  '),(3673,3992688,'21,July,2004','2004-07-21 00:00:00','   urlLink P. Diddy launches group for youth vote   Whatever happened to Sean Combs?  He used to be cool...  '),(3674,3992688,'21,July,2004','2004-07-21 00:00:00','   Ok here it is:  Not long ago I gave a friend of mine a TexShare card from the library.&nbsp; She then took this TexShare card out of town to a different library that is also on the TexShare system.&nbsp; That library took one look at the card and commented, \"oh Irving will give a card to  anyone , even someone from Ohio...\" and basically told her that they wouldn\'t issue her a card since they knew that Irving\'s policies were so open.&nbsp; Naturally I was unhappy since they offended my friend, my library, and my dignity...so here is what I did:  1.) I discussed this with Logan, who also agreed it was a travesty of everything right in this world, and I then emailed my story/complaint to him which I will forward to the Assistant Director tomorrow.  2.) I emailed the Texas Library Association and reported that a certain small-town library feels they have the right to be selective on TexShare cards they honor based on their interpretation of the other library\'s policies.&nbsp; We\'ll see what comes of that.  3.)&nbsp;I emailed the offending party a certain link that starts with  urlLink www.tub &nbsp;... oh wait, this has gone too far already...  '),(3675,3992688,'21,July,2004','2004-07-21 00:00:00','   Right now I\'m watching the X-Files and contemplating having the last Dr. Pepper in my refrigerator.&nbsp; I went over to Kevin\'s house earlier because he couldn\'t get his laptop to communicate with his desktop over the&nbsp;network and I\'m still puzzled about it, but I couldn\'t get it to work.&nbsp; Maybe Jason\'s luck has passed on to Kevin while he\'s away being ass-kicked at bootcamp.  I\'m also still kinda freaked out about the weird French guy trying to kiss my hand at work too.  Check my next post for some crazy library lore...  '),(3676,3992688,'17,August,2004','2004-08-17 00:00:00','   urlLink    The horror. The horror.  urlLink      when your shirt rips because the fat is trying to escape.  Look close enough and you can see the rip on the belly where fat is bulging.  '),(3677,3992688,'14,August,2004','2004-08-14 00:00:00','   urlLink     It\'s not really his car, but how the hell would you get out of this situation?  '),(3678,3992688,'13,August,2004','2004-08-13 00:00:00','   Imagine the laptop you could make with this stuff http://www.popsci.com/popsci/science/article/0,12543,676853,00.html  '),(3679,3992688,'11,August,2004','2004-08-11 00:00:00','   Rscheearch sowhs taht    it deosnâ€™t mettar what order the ltetres in a wrod are slpeeld, the iprmoetnt    thing is taht the fsrit and lsat ltteer are in the rghit palce. The rset of the ltteers can be mexid anruod and you can still raed it wouthit porbelm.    Taht is bacesue we raed the wrod as a wolhe, not eervy lteetr by it slef.  '),(3680,3992688,'11,August,2004','2004-08-11 00:00:00','   You guys ever check out the \"Blogs of Note\" on the Blogger homepage? Here\'s an entry from Aziz\'s blog:  \"Monsoon Miracles:  Today was an excellent day. The most runny of poos that have tormenting me for the past few days finally relented, i\'m sure because I got my hands on some sweet drugs, namely Ciprofloxacin, an antibiotic. This miracle antibiotic has, I think, saved my life at least once and maybe twice while i\'ve been here in India. Since Iâ€™ve already amply described my near-toilet experiences with Indiaâ€™s unique strains of explosive diarrhea in previous entries, and plan on writing a book to be named \'The Story of My Experiments with Diarrhea,\' Iâ€™ll just say that I have gone from near-deathly to definitely alive within the past few days.\"  Just in case anyone ever feels this blog isn\'t up to par...  '),(3681,3992688,'10,August,2004','2004-08-10 00:00:00','   You get this!    naked man in japan  ... the mud everywhere, the beer, the sweaty, naked bodies in the ... he could drink and run around naked without getting ... They came out \"glistening\" as Tim would later ... ducts.org/06_01/memoirs/naked.html - 26k -  Cached  -  Similar pages   This Fish Needs A Bicycle: Comment on talk to me, goose  ... ...glistening boys on the beach ... god robes on a pyramid, with a thousand naked women screaming ... And every single weekend they played hot sweaty volleyball in ... thisfish.com/cgi-bin/mt-comments.cgi?entry_id=715 - 19k -  Cached  -  Similar pages   Keith Harris - Fathers & Sons  ... His glistening, naked backside and my arms were reflected. I leaned forward to kiss him. ... He was still naked, his body sweaty and tense. ... www.ars-rhetorica.net/ Queen/Volume11/Articles/Harris.htm - 55k -  Cached  -  Similar pages   Rub-a-dub-dub, G-boys in a Tub! -- The Sequel  ... and Wufei entangled on the bed, sweaty and flushed ... fantasies, naked in an equally naked Heero\'s lap ... muscles hard and straining, sweat glistening, cobalt eyes ... jade1x2.com/Gundam/rubadubsequel.html - 17k -  Cached  -  Similar pages   Instinct Magazine  ... wages and tips Jonathan can skip home with a sweaty $200 to ... 3. Dancing nearly naked in front of strangers is: A. A ... B. Glistening, gyrating boys = World Peace. ... www.instinctmag.com/issues/0601/article3.html - 25k -  Cached  -  Similar pages   Boys\' School -- Part 1/2  ... I love savoring those rippling muscles under glistening sweaty skin straining ... to go upstairs with the other boys to my ... When he tucked naked me into bed I asked ... www.asstr.org/~YLeeCoyote/BoysSchool.htm - 25k -  Cached  -  Similar pages   The Sisterhood of SEX (3 of 5)  ... erotic oil that had covered her naked, glistening frame ... another orgasm wracked her shaking and sweaty frame ... how much more attention the college boys were paying ... www.mcstories.com/SisterhoodOfSex/SisterhoodOfSex3.html - 16k -  Cached  -  Similar pages   DRUNK OLD WOMEN YOUNG BOYS: : CANDID DRUNK GIRLS, DRUNK NAKED ...  ... Lick my sweaty feet...Yes my pet it turns me on to ... Soon your left foot is glistening with my saliva, shining in ... CANDID DRUNK GIRLS, DRUNK NAKED GIRLS PICTURES ... duni.bilji.org/drunk-girl-stories-movs/ drunk-old-women-young-boys.html - 7k -  Cached  -  Similar pages   Little gay boys porn nude  ... gay teen suicide, young gay guys, teenage boys naked gay, free ... As they wiped down the sweaty equipment, Chris stole ... as hot as any of the glistening buff men ... gay.mb4ce.com/dragonball-z-gay-pics-thumbs/ little-gay-boys-porn-nude.html - 5k -  Cached  -  Similar pages   FOR-LOVE-INDIAN-BOYS-4 \"Mother Moon Kisses Me\" {Part #4 of 9} ...  ... it easier for my shy eyes to watch his naked body straining ... At its finish, the man brings my glistening palm close to him. ... I bury my face in his sweaty chest. ... nifty.nisusnet.com/nifty/Vince_Water/ For-Love-of-Indian-Boys/For-Love-of-Indian-Boys-4 - 20k -  Cached  -  Similar pages  '),(3682,3992688,'09,August,2004','2004-08-09 00:00:00','   urlLink Perverted-Justice.com - Exposing wannabe perverts on the \'net   Guys you have to check this out.  I saw this website on TV last night.  People posing as young boys and girls have conversations with pedophiles and record them on this webpage.  They not only do this but expose the names, addresses, and screen-names of the perverts!  It\'s awesome.  I think Nathan is on there somewhere...  '),(3683,3992688,'09,August,2004','2004-08-09 00:00:00','     '),(3684,3992688,'09,August,2004','2004-08-09 00:00:00','   urlLink  '),(3685,3992688,'07,August,2004','2004-08-07 00:00:00','     '),(3686,3992688,'07,August,2004','2004-08-07 00:00:00','   AMSTERDAM, Netherlands -- Toe-licking is not a crime in the Netherlands -- but it may soon be.  A serial toe-licker is prompting lawmakers to consider making such unsolicited attention a criminal offense.  Police arrested a man in Rotterdam earlier this week after a female sunbather complained he unexpectedly licked her bare foot. Other women have also complained. The man has reportedly pursued his fetish for years -- but Rotterdam prosecutors say they are powerless to put their feet down.  A prosecution spokesman told the Rotterdam Daily newspaper, \"A lick over the foot doesn\'t qualify as a crime\" -- without an overt sex act.  Two Labor Party members have asked the justice minister to change the law.  In their formal complaint, the men ask how it\'s possible to \"prosecute someone for throwing a cigarette butt or soda can on the ground, but not for this kind of misbehavior.\"  '),(3687,3992688,'07,August,2004','2004-08-07 00:00:00','   urlLink    ...for Daniel.  urlLink     This is how John and Daniel play Super Smash Brothers.  '),(3688,3992688,'05,August,2004','2004-08-05 00:00:00','   You might think that the library would be a clean and quiet place to work, but you would be wrong. Today is a day like any other here. It is no better or worse. While cruising through life\'s journey this afternoon, I have encountered mold and a crazy lady shouting while carrying what might have been all her worldly possessions on wheels behind her. I didn\'t catch all of this one sided conversation but I did hear her shout angrily about a girl asking her for a dollar but she didn\'t have a dollar and the girl should have just asked her mother for it. But...there was no girl. There was however a very concerned citizen who was kind enough to suggest someone do something about her. Unfortunately, she exited the building just as it was getting interesting.  '),(3689,3992688,'05,August,2004','2004-08-05 00:00:00','   urlLink     urlLink http://www.blogger.com/profile/1866949  After careful scanning of the groups of people who share my various interests, I have selected The Laird of Madrona as the most shameful to be associated with. From his skirt-like kilt to his mead holster, his appearance is ridiculous to me. I deride his taste in music for nerds who think they\'re clever, funny and \"hip.\" Above all else, I despise his interest in \"cutting capers.\" Congradulations, Laird!  '),(3690,3992688,'04,August,2004','2004-08-04 00:00:00','   With skin:      Without skin:     Reading and stuff:  '),(3691,3992688,'04,August,2004','2004-08-04 00:00:00','   Alright, I have determined the easiest way to post images within the blogs. That is, for pictures that are already online, (for images on your computer you would have to use Hello or Flicker). Here is how you do it. To insert an image in your blog, (anywhere in the blog, you don\'t have to put it at the top) simply click the \"Edit HTML\" tab while you\'re typing your message and type this code tag:  I can\'t type it all in one line or it thinks I\'m making a real link to an image.  Type this first   Then   inside   those carats you type this:  IMG SRC = \"http.address.com\" get it?  So you have the first carat:   then you type:  IMG SRC = \" urlLink http://www.address.com \" and then you have the last carat: />  You can click over on the \"Compose\" tab again to see if it worked. Like I said you can put this tag wherever you want, and as many of them as you want. So you can fill your blogs with images of oh-so colorful things.  '),(3692,3992688,'04,August,2004','2004-08-04 00:00:00','   This email came to us yesterday at East Branch:   From Maria Redburn  Subject: Shirley Hinton banned from Phone use  Ms. Hinton is an older woman with white hair.  She has been using our phone to call social security.  Yesterday she had a 40 minute phone call and came back for another 20 minutes.    I\'ve seen this lady before. She has crazy hair; I mean, we\'re talking Don King here. The first time I ever saw her, she came to use our phone to report a crime. She called IPD to report that some lady was selling prescription pharmaceuticals, except it was over in Garland, according to her story. So they told her to call the Garland PD. She did that and spent about thirty minutes trying to explain herself, which involved a lot of speaking very loudly and repeating herself. I thought we should have banned her from the phone before she ever got on, based solely on her smell, but hey, it\'s not my decision. This last time, she didn\'t even ask, she pulled over Janis\' chair and sat down behind to use our phone, which happened to be behind our counter.  '),(3693,3992688,'04,August,2004','2004-08-04 00:00:00','   We\'ve all been waiting patiently for you to post on the blog...come on!  '),(3694,3992688,'03,August,2004','2004-08-03 00:00:00','   Since the two things that dominate my thoughts the most are sex and politics, I began to wonder if anyone had ever written any slash-fiction between political figures. Sure, enough I did manage to find one guy. Unfortunately (or not for you guys), it\'s all gay. Here\'s  urlLink Blair/Brown  and  urlLink Bush/Chirac . There\'s more listed  urlLink here , but no links, so you\'ll have to make do with those.  '),(3695,3992688,'03,August,2004','2004-08-03 00:00:00','   If you read my comment on the last post, and wondered, \"what in this sick world could Brainerd Diarrhea be?\" then I have your  urlLink answer.  '),(3696,3992688,'03,August,2004','2004-08-03 00:00:00','   urlLink    This comic is not a collector\'s item   Or rather, I have had some passing but nasty affairs with comics in my youth. Now, I happen to like comics a lot. No, they\'re not great literature. No, they don\'t make for the most challenging of reading. And no, calling yourself a \"collector\" doesn\'t allow you to escape the shame that you should feel at reading comics as an adult. However, I for one have never loved comics so much that I thought I should try my hand at writing one, or even  urlLink publishing it online . This gentleman\'s comic is less like the Sunday comic or the super-hero comic, and more like the comics they hide deep in the Dallas Observer...only not as good. But please browse his site and read a few. They may grow on you.  '),(3697,3992688,'03,August,2004','2004-08-03 00:00:00','   urlLink    What caused someone to look at this picture and think, \"It would be funny if we painted her green and called her Shrek\'s mom?\"  Would this classfy as \"dumps like a truck?\"  '),(3698,1807720,'31,August,2003','2003-08-31 00:00:00','    Well the party on Friday was fun and well yesterday we just all went and did whatever had to be done and then in the evening we went out for Chinese at that place Down Town called Ho Won.  Was that ever good, I totally just stuffed myself cause I had only eaten like a little breakfeast earlier that morning.  Today I was planning to go to church but my mom is like kind of against it I guess because like she kept persuing me not to go.  Because well there was this water main break and it flooded my sisters house so now I have to help them move their stuff from their house to our garage or else the storage fees will be very expensive!  And well right now they can\'t even live in their house because it is like ruined all down stairs and the walls are starting to cave in.  So well they bought a 5th Wheel to live in during everything and then in the Winter they will probably live with Jim, which is Ben\'s mother\'s boyfriend.  Kind of confusing but honestly I am not to crazy about Jim and all.  And well Ben\'s dad passed away from Cancer on Friday night and yeah even since I applied for some jobs I\'ve been having second thoughts - which is like NOT a good thing at all.  I guess I just still realize how important everything else really is to me, and I guess it kind of lets me know where my priorities are at the moment.  But well I am just concned about the time I will have available because most people don\'t take as long as I do to like finish work or study and things like that.  And well during the summer I sort of began to maybe question that like maybe there is something wrong with me.  I do get good marks and all but I don\'t think many people realize the countless hours I spend trying to figure something out or all the times I have cried because I tried and tried to figure out math and it just doesn\'t make sense.  Sometimes I think I might have some kind of  disability like associated with learning but when even I mention that people just tell me that I am crazy but I am not sure what to think I mean maybe I am crazy but I wish I knew why I cry like a million times per week and everything NEVER and mean never comes easily.  I have to work for all things and let me tell ya I work my ass off and yeah only I know truly what I go through.  But I don\'t think I am crazy, I really think that maybe I could possibly be right, but I suppose there is no way to know for sure and everyone has this ideal in their head that I am some kinda really smart girl.  And yeah so maybe I am smart in their eyes but if only they knew a little more and felt all the feelings and went through all the experiences I\'ve been through.  Especially when I consider how hard driving was for me to learn and I even did and extra class of drivers ED and like I still got a 70% and I was suppose to have 75% to pass.  So well I am putting of going for my license until about next year because I don\'t need the stress of driving in my life too.  But well yeah I don\'t know what to think when it comes down to about everything.  When I was a kid I never could count money or tell time very well, and I remember always being in some kind of room working with this lady who helped my learn how to count by tens.  And well it\'s not like math had even been easy, all my life it has been hard and like I read that sometimes disabilities are life long.  Sometimes I think that maybe someone knows something I don\'t, I mean maybe I do have some sort of problem but no one wants to tell me because maybe the way I would feel like I dunno maybe my parents or something, if anyone were to know anything it would probably be them.  But then again I might just be thinking way too much.  But I mean when I think about all the difficulties with everything it makes sense that it could be possible for me to have some kind of disability involving learning well at least in my mind it makes sense but people usually just think the opposite.  I never thought school or driving or anything was suppose to be soooo hard but I dunno maybe it\'s not just me and mayeb their are others who feel the same way.  I guess I will never really know for sure though and I just have to go along and do my very best because that is all I can really do.   '),(3699,1807720,'29,August,2003','2003-08-29 00:00:00','    I have been so busy!!!  Going to a party tonight from 7:00 - 12:00 PM and my dad is flying in from Vancouver tonight.  Today I took the C-Train to Chinook and applied at Old Navy then I took the bus down to Wendy\'s and applied there cause Chels, Derek, Eric, and Lawrence all work there so I think that would be kind of cool.  On Wednesday night I went to go see Uptown Girls and we got popcorn and later on some ice cream...YUM!  Last night we went to go eat out  *AGIAN*  and I bet you guessed it...Zellers.  Of course there isn\'t really many options because all my Grandma likes is Zellers and Dairy Queen.  Anyway Jenny Gear finally got voted  *OFF*  Canadian Idol!!  It\'s about time!!  And MY idol Billy Klippert is still in there!!!  I sure hope he is the idol, although I still love Clay Aiken from American, he will always be damn sexy and irresistable.  LOL.  And well the last Big Brother it was like really cool how Jun got to leave the house to go to the VMA\'s, like how cool.  I\'m kind of disapointed that Jack got voted off though cause I really did like him.  And well I  *CAN\'T WAIT*  for the finale of For Love of Money 2 on Monday.  I am like so into that show now and I used to think it was so dumb.  Canadian Idol was also really funny last night, my mother and I were killing ourselves laughing at these people who really could not sing.  It was the best of the worst show and man those are like sooo my total favourites.  Anyways school is coming soon and probably I will have a job soon and well now I have to think about how I am gonna handle everything!   '),(3700,1807720,'26,August,2003','2003-08-26 00:00:00','    This morning I baked some cookies to freeze for when my dad comes and also some for my sister and her future husband.  Then in the afternoon all I did was watch TV and just relax.  Yesterday Canadian Idol was also sooooooo awesome I mean Billy Klippert was awesome.  When my Dad was leaving from Calgary and going back to Vancouver we saw him at the airport.  He shook my Mother\'s hand and he hugged me.  Oh how lucky I really am!  Meanwhile I really think that it is Jenny Gear\'s time to go, in fact I never thought she should have even made it to Canadian Idol.  Toya was much better than Jenny, I mean Jenny can\'t sing - all she does is move her mouth and it is painful for me to listen to her.  My mom says she looks like one of those Hudarite people..LOL!  When I actually think about that it is so true!  I think that Canada judges way to much on appearance instead of like  REAL  talent!  And well my mom is at the Foothills hospital right now because Ben\'s dad is dying of cancer and they think he is going to go soon.  But well I think my Grandma is taking us out for dinner tonight - none other than the one the only...Zellers!  LOL.  That is like her favourite restaurant but well me and my mother would rather have a plate of ribs or Chinese...YUMMMM!  Oh well, I won\'t be greedy, I should just be lucky that I am actually going to go somewhere tonight  Lately everyone has been working!  Chelsey especially has been on the go non-stop and well she doesn\'t have a single class with me in grade 11 and well I don\'t know when we will even talk or see each other except for maybe Fridays cause we all usually hang out on Fridays.  School is only a week away now and well I am getting a bit more excited but a little nervous.  Just hopefully everything goes well and well I plan to ace Chemistry and hopefully that will happen!  Last year I don\'t really know what happened, I was good at anything that involved Science except for in Grade 10.  This year I hope to turn things around a bit, but of course still keep my good Math mark.  And knowing how Math last year was a big emotional roller coaster it will be even more difficult this year.  But well I took Physics and have it in the 2nd Semester and well last year I wasn\'t all that great at it and now I am having second thoughts about it.  I was so confident before that I wanted to do it though, and well I think after doing Math 20 Pure in the 1st Semester I should be able to handle it.  I will probably just try it for a while and see how it goes.  It would be pretty funny though if that ended up to be my highest mark.  I mean anything can happen can\'t it?!?!  I guess no one really ever knows.  I am just going to put my heart into everything I do and what ever happens was just meant to be!   '),(3701,1807720,'25,August,2003','2003-08-25 00:00:00','    Yesterday I didn\'t do alot except go to church, hang around the house, and bake some cookies.  Today, first I went to the mall and got some lunch at that new little place to eat in Sears.  Later I bought myself some socks - need those for sure!  And I also got myself a new bag for school.  Then I bought a pair of jeans for myself and that was basically my day.  Nothing too exciting because well I am trying to get in as much rest as possible before school starts.   '),(3702,1807720,'23,August,2003','2003-08-23 00:00:00','    It was just one of those days today - the ones where you kind of just take it easy because you don\'t really have anything planned.  Well basically all I did was *Majorly* clean my room today cause I all of a sudden had this like erge for everything to be clean.  Sure is funny isn\'t it.  And well tonight I think I am going to watch some sappy movie my Mom rented even though I don\'t really care for those but later at 9:00 PM they are having a special Big Brother Episode where there is going to be a surprise \'twist\' to the show.  I can\'t wait!  And well Reichen and Chip won the Amazing Race which I totally think they deserved and well the new Survivor is starting some time in September so I am really exited for that!   '),(3703,1807720,'22,August,2003','2003-08-22 00:00:00','    Well lets just say I am sooooo tired right now but I don\'t want to admit that or anything cause I am like still young!  And well today my Mom, her friend Linda from Winnipeg and me all went to Heritage Park today and we basically did it all.  We went on the boat, the train, and to tons of shops and exibits.  We even had a free breakfast and lunch there.  Then later on we went to Chinook Centre and went shopping!   '),(3704,1807720,'21,August,2003','2003-08-21 00:00:00','    Well this morning I went to get my hair done and I am kind of getting used to it now.  I wanted something a little bit more vibrant because I am not really the conservative type but it is very nice and I actually really like it.  I also met with Anne today to talk to her about becoming catholic and that went pretty good but I don\'t really know where I am going to find the time to do all of this while also trying to do my school work.  Also at Super Store today I saw Claire and her Mom and they couldn\'t stop saying how good I looked.  Maybe I lost more weight then I thought I did cause people are really starting to notice now and I actually feel really great I mean I really like it when I hear things like that.  In the past especially there are somethings I would like to forget and I haven\'t even told Chels about them which is how deep those secrets are but things are starting to turn around now.  I\'m getting much more confident, thinner, and much more happier and I just hope it all continues to rise.   '),(3705,1807720,'20,August,2003','2003-08-20 00:00:00','    Today I went to South Centre and got a new pair of leather shoes, a belt, and a new bag.  I had alot of fun but all my birthday money is going to be going before I can blink!  Tonight I am going to that teen night thing and we are having free pizza and cotton candy not like I really need that or anything.  I think I am really starting to get annoying counting calories and everything and I would like to think I am not obsessed but I am begginning to think I might be.  It has gotten as bad as skipping meals and things like that and everytime I eat I feel a huge sense of guilt afterwards.  Maybe this is just normal for teenagers though and everyone is probably just over-reacting.    Anyways I am still excited for chemisty cause well I think I want to either major in Chem or Bio but I am not sure which one I like better or which one I will advance at more quickly.  I am really thinking I want to become a high school teacher.  Like I am not the most talented at science or anything and even though sometimes it seems really hard but it just really interests me.  When I see people doing their thing and teaching kids about my age I think how cool it would be to do that same thing.  I really look up to alot of older people that have chosen the teaching profession - I mean if you really get to know them they are not as different from teenagers than everyone thinks - they are caring, compassionate, and really have an understanding of what others are going through.  I\'m sure it is not the easiest job with the best pay but I remember a quote I once read that went something like  \"Chose a job you love and you will never have to work a day in your life.\" The quote really speaks to me strongly because I mean I would rather be enjoying life than rich and unhappy.  Still I continue to wonder about the future - Like will I ever meet someone that loves me back and will I ever get married and have kids.  Just the normal things that people are expected of today.  Considering I haven\'t been with many guys really it concerns me but most people say high school relationships don\'t usually work out.  It is really quite confusing, in fact almost everything in life has never really made sense.  Like this summer I really had the worst drivers ed. experience ever, met a boy and fell in love but that turned out awful, and couldn\'t even find a job.  But really now that I think about it I am not ready to drive yet and I should take more time to improve, I know now what kind of guy I *DON\'T* want and exactly how I *DON\'T* want to be treated, and now that I have seen my school schedule I was thinking \"Thank Goodness I don\'t have a job because I know that would not be about to handle everything.\"  So I guess everything really happens for a reason and I feel that all the great and wonderful people I meet and have met in my life are there for a reason!   '),(3706,1807720,'19,August,2003','2003-08-19 00:00:00','    I finally got my schedule on August 18th the day after my birthday!  I\'ve never been so excited to open my mail before and what a relief it was!  I absolutely love my new schedule for next year and I am so excited for chemistry and I don\'t even know why.  Maybe secretly I really like it and I just won\'t confess...lol!  My summer however wasn\'t all that and well it was pretty bad except for the odd days here and there but I suppose everything happens for a reason.  And well yesterday Chels and I went shopping and we both bought the same pants for school - like how totally cool and she got me this awesome ring for my birthday that matches the one she has!  I\'m getting my hair done Thursay and nope I probably won\'t look like myself anymore but that\'s the whole point!  And well I had a little romance of my own over the summer but the book is still open so I won\'t say anymore until the chapter ends and the book is closed.   '),(3707,1807720,'30,September,2003','2003-09-30 00:00:00','    Well I made it to my comp finally to write this.  All that math and Chem..oh man!!  But well I REALLY wanted to prepare really well for my quiz tommorow and well I am feeling pretty confident - but not over confident of course but I KNOW that if I want to do good I really believe that I have to power to.  And well I\'ve got the Chem quiz on Friday but I am having some diffuculty understand the stuff and I did go to tutorial today which sort of helped but it\'s still confusing so I think I am gonna go for tutorial tommorow too and well what ever happens with my Chem quiz at least I know that I did my best and tried really hard.  And well this morning me and Chels got into a huge and I mean HUGE fight.  In fact I don\'t think we have ever had one quite as bad as the one this morning.  Because well I told her I liked Alex and all but we had a LONG and I mean long talk about how that stays between me and her so I found out yesterday that she told Lenny and well he talked to Alex and well apparently he likes me back which is a good thing and all.  But I was questioning her why she told him because well *I* wanted to talk to Alex and about the way I feel and for once *I* wanted to do something for me and instead of having it done the CHEAP way and having people set others up.  But I guess what\'s done is nothing and there is nothing that can really change that.  But I am actually not upset that we had that fight.  To be honest when she walked storming off I felt a huge sense of releif.  I feel like I am always trying to live up to her expectations and all and like she thinks she is almight and powerful and great.  Well I don\'t know how bad my attitude was but I was just questioning her and she took it the wrong way.  I think maybe though some part of myself is changing - espeically my attitude I just don\'t kow what is up with all that.  Maybe it is just a stage that everyone goes through but at different points in their lives.  But who knows.  I am feeling pretty happy right now actually despite being a bit well majorly confused in chem but math was SOOOOOOOOO awesome today.  I thought it was so fun, and like I was excited to do more practice problems.  Lol...if anything that so does not sound like me but it is so true.     The Untitled Poem  Floating on my feet once more I have looked on through this open door Steped inside to see what I face On this path to a better place  Knowing my attitude isn\'t always great But in life we must learn to appriciate All the things, the good the bad Even those that may be happy or sad  This is a stage I think all people go through Just some of us wish only we knew How to deal and get through each day Knowing that I can\'t always have my way  ~Vanessa   '),(3708,1807720,'29,September,2003','2003-09-29 00:00:00','    \"Sometimes I wonder why I am here\" those are the words that kept going through my head tonight.  I must have cried for over half and hour on an off.  I didn\'t really notice how bad things are going during school because I was busy with other things but tonight I was jus thinking and a huge rush of tears poured from my eyes.  I don\'t really think it is normal for a teenager to cry almost like every second night or sometimes there have been weeks where I just cry at night but it is getting bad.  I got a 35% on my social M.C. test but I did way better on the long answer and got a 75% so like I do know the material I just don\'t know why I have to fuck every damn thing up.  I was working on my math but I kept crying and so I decided I would write in my little journal thing cause I just need time to think about things in my life right now.  And well I don\'t understand the Chem stuff we did today and well I keep trying to inform my math teacher that I really don\'t understand but I just feel ignored and unimportant.  I am worried about the quiz Wednesday - after all who wouldn\'t be worried if they didn\'t understand.  And I tried to get some help today during tutorial but like there was only one lady in the room helping people and there were alot of people with questions so I could barely get any help.  To be honest the schedule and the tutorial thing worked way better for me last year and I was so much happier.  Although I try to put on a nice little smily happy face at school it is hard sometimes.  If someone where to look into my eyes and just \"look\" as they say it the eyes are the window to the soul.  And yes they truly are because when I stare into my eyes I see all the hurt and pain I feel.  And all the irrational thoughts I get once in a while were coming back into my head again tonight but all my dreams and hopes me so much, which is probably why I haven\'t given up.  But I don\'t know what to do and I don\'t know why I just can\'t understand like everyone else.  I mean I am going to the resource room, I\'m asking questions, I am going to tutorial and stuff and I mean I don\'t know what else to do.  And my marks worry me because that is the ONLY thing my parents ever talk to other people about, like last year how they were like \"Oh and Vanessa is on Hounors w. Distinction\" and blah, blah, blah.  Now what the hell are they suppose to brag about...that I can barely get good marks anymore.  Hmmm lets see...I can only wonder, they probably just won\'t talk about me at all like I am so unimportant and they only seem to worry mostly like about Ben and Melissa and everytime my Dad flys in he is always at his place and the only time I saw him last time was when we ate meals.  I just wish there were answers right now, I wish that someone could explain to me why everything is so difficult right now and is eating away at me self-esteem, self-worth, and at my heart and happiness.  I wish that I could wake up and everything would be ok - but in reality it just isn\'t ok.  And I wish that someone could look into my eyes and truly understand what I am going through and feeling and I wish for once that maybe someone could just be there for me to hold or hug.  To just stand there and feel someone else\'s arms around me reasuring me everything will be alright and just holding me for what seems forever until everything feels alright.   '),(3709,1807720,'28,September,2003','2003-09-28 00:00:00','    I went to church this morning which totally made me feel good afterwards.  Chelsey called me last night so we went out for dinner to Perkins cause we both used to work there and stuff and that was like so fun and well after we just hung out at her house and talked for like a million hours and when andreanna got home from work we were all listening to Christmas music..Lol.  Then I left around like sometime after eleven but man was it creepy walking in the dark.  Like I know self defense and all and even though I have my brown belt I kept thinking someone was like behind me or like hiddin in the bushes.  Then like I heard this noise and looked back and then when I looked forward again and I saw this massive shadow and I almost had a heart attack cause I thought it was someone else - luckily it was only my own shadow though.    And well I went to South Centre today and got a new winter jacket and some boots for when it snows and stuff and well I was doing my homework but social is no boring and I am behind and I just don\'t want to think about it anymore.  It is driving me up the wall cause I really hate how we always have to work through those guides and stuff cause I just don\'t have the time and patience to sit there and do that kind of thing - I wish they would just get up and teach already and well they should at least give an option if we want to be taught by someone or if we want to do it independantly but ya know I don\'t think that is going to happen.  And well AHHHHHHH math quiz Wednesday and I don\'t understand.  And well my goodness lately I\'ve been coming to have like a totally obsession with space and that.  I don\'t want to be no astronaght (sp?!?) or nothing like that but I actually find it really fasinating and really cool!               I mean sometimes I wonder...what am I good at anyway - I have a lack of answers when it comes to that and when it comes down to it I don\'t really know if I have any talents.  I mean lets see I am not very good at Chem, or Math, or Social, or English, or Physics and yes the list goes on.  I\'ve played sever years of soccer with fourteen medals, play the flute but I don\'t take band anymore because the theory is too hard for me, been sewing for a couple of years but really not very good at that, used to do a form of Korean Martials Arts and I have my brown belt but all that doesn\'t make me talented.  Why does it have to be like that...I mean just WHY???  I definetly KNOW it is not from lack of effort so I just really don\'t know and wish I did.  I wish I had some subject that I just would naturally excel at but Drama doesn\'t count because well anyone can act if they put the effort in and I wish that I just had something special that I could do that would define ME.  But right now I don\'t feel like there is.  And knowing that I am not doing well in Chem right now and I DON\'T want to major in Physics if I am not good at Bio next term I have no idea what my purpose anywhere is.  I mean I know that my HEART tells me that no matters what happens my dream is to be some sort of Secondary Science Teacher and I couldn\'t imagine doing anything but that but in my MIND I just don\'t know if I can do it.  And I KNOW that if I don\'t become some sorta of teacher that has something to do with Science I will regret it all my life and will always wonder well WHAT If.... and will probably not be very happy.  I was so confident at the end of last year and so happy but I don\'t know what happened to all that happiness and I don\'t know what happened to that girl I once knew will all that confidence.  But then summer happened, and it changed everything.  Everything right now is changing and I don\'t really like change very much.  I was happy the way everything was last year but over the summer like Drivers Ed totally ruined my confidence and self-esteem cause of the fucking bitch of a lady who called herself Kathy who thought that she could stomp all over teenagers and tell them they are no good at anything.  I mean I don\'t hate alot of people and that is not really my think but I do know that I HATE that Kathy girl.  And well even though I met Aaron over the Summer our intial relationship was good but then everything just fell apart.  Everything fell apart in the Summer just the way it is falling apart now.     Hopes and Dreams  Just feels like everything is falling Knowing what my heart is calling Wondering in this mind of mine Hope to know what is right in time  Dreams arenâ€™t what they used to be Confidence left I cannot find in me Searching in this life of mine Hanging on this one lifeline  Gasping for that breath to take Looking around for a purpose I can make Questioning this authority somehow Thinking of these dreams right now  Falling through this dirty floor Knowing I am not welcome anymore Left alone and on my own Searching for a brand new home  Where shall my heart sleep tonight? These dreams are drained from this flight Everything now keeps falling in Surrounded by hopes I can only loose or win  Looking through the hands of time Finally I found that someone so kind Special, rare, and precious too She is someone just like you  When I am lost, she guides me through She knows just what she is suppose to do Sometimes I wish I could be just like her All her hopes and dreams they are so sure  In this destination she stands still But here I am waiting for that day until I find my confidence and achieve these dreams Knowing I have the willpower and the means  ~Vanessa     Maybe in a couple years from now I will probably laugh at the way I feel now.  But right now it feels so real and it feels like my heart is being crushed by everything.  My hopes, my dreams, my goals - all that I have even wanted in life I don\'t even know if I wil ever have any of it.  And hopefully in a few years I\'ll be at the U of C.  I\'ve got it all planned out I just have to get the marks and everything which is a challenge now and I heard that now they only accept people with the highest marks and that is what it is based on.  But well I want to do three years in the science field and get a degree and then do the teaching program for two years and then get a job teaching somewhere is the S.W. or S.E. at a highschool and then I want to do evening classes and go for my masters in teaching.  That\'s my plan but somewhere along the way I want to meet my husband which I hope will be a big church go-er and someone who is also a teacher then I want two kids.  Not one, not three but two because one is not enough and three is too many.  I wish the sex of kids were in my control though cause I want two girls - Edele and Janine but ya know if so happens that I have a boy I like the names Eryk and Jaden.  So it\'s not like I don\'t know what I want in life because I KNOW EXACTLY what I want I just don\'t really know how to get there because it is not something that anyone can even have total control of.  and well I guess I just don\'t know and for now I will just have to keep going from day to day knowing where I want to go and trying to get there.   '),(3710,1807720,'27,September,2003','2003-09-27 00:00:00','    Did like nothing exciting today at all and well no one ever calls me so yeah.  My dad came in Thurs for the weekend but he spent the whole day at Ben and Melissas place and part of the day with my mom.  And well  I got part of my homework done but that darn ole\' procrastination ya know has to kick in sometime and boy did it even kick in.  And well I tried to finish my math homework but well I am really confused but she never finished the lesson and well luckily we get a work day on monday.  And oh man does this pic ever remind of worried I can be at times about Chem.  LOL.   '),(3711,1807720,'26,September,2003','2003-09-26 00:00:00','    Was today even better!  I actually felt happy today - really happy.  Today I actually smiled for once in like what seems a thousand years!  Everything just seemed to go right.  Lucky pennies really are lucky cause yesterday morning I found one on heads up.  I did awesome on my Chem test compares to other marks - I got a 74% but like that is really good for me right now and I studied like mad for that test.  And Britt just called!!!  I haven\'t talked to her in ages!!  She goes to Scarlet and stuff and we just had like a big long talk for an hour!  Today me and Alex went out for lunch and then we met up with Tyler, Ash, Sarah,  Ali, Lenny, and Trav.  I hope I don\'t gain like tons of weight cause well I couldn\'t resist the temptation and I HAD to get a chocolate brownie Frap.  God was that ever good and well I drank that thinking I would just skip supper and it would make up for all the calories but then I get home and my mom had ordered my fav. pizza - bacon and tomato and well I HAD to have some of that.  So yeah basically I feel like really guilty cause I had way to much food today and well that ain\'t good.  But well I had such a good time today and Chels wasn\'t there.  No one to fight with means no one to upset me and it actually was really nice.  Instead of speding all my time with chels I spent it with everyone else and it was so awesome.  I think my and Alex might have something between up - the romantic chemistry is totally there and he wants to walk with me to school one day even though it will be a long walk from my house.  Everything was just so awesome today and I even did the Chem lab like really well too cause for once I understood the directions and I actually felt like I did that lab without any help and like after I even cleaned out all the beakers for the teacher for her next class and well I stayed a bit later and the Chem teacher told me I didn\'t have to do it but it felt good to contribute to something and I felt like I could actually help with something for once.  It was nice.  I\'m happy at the moment and everything seems like it is going well.  To be honest Chels is alright but I had way more fun with out her.  It is sad to say but is true.  When we are not around each other we don\'t argue or disagree, I think for now it might be better for us to be apart.  Me and her are both going through alot of different and difficult times and well I think we just can\'t deal with all of it and it causes us to argue.  Besides I feel like I am getting to know my other friends in a greater depth.     Future and Past  These days are different Comes with a new foundation For times have been not so vast Searching for this sharp content  Then I saw the wrong,  Now I see the right Looking back in history, Shall not dwell on each date  Craft a future for what you crave In your spirit, one can recognize Where the heart slumbers in aspiration Only you know where you want to go  Take the stairs one step at a time Acquire the understanding of determination Know that amid willpower you can grasp a hold  Anything is probable, everything hugs potential  Lacking friends with choices firm Not one to steer, simply yourself to depend People transform and people depart Whether future or past, just follow your heart  ~Vanessa   '),(3712,1807720,'25,September,2003','2003-09-25 00:00:00','    Well I definetly thought my Chem quiz went good today but last time I thought something went really it good it didn\'t quite turn out how I wanted it too.  The only thing I really got confused with was dissociation vs. ionization but I thought everything was alright.  And well I got a 73% on my math test which is ok but I know I am so much smarter than that and could do better but I messed up two M.C. questions and they are worth 2 marks each.  And well last time I got a 72% on a test the math teacher told me that I could do better and she told me that she knows I can do the material and then today I got like one percent higher and she was like \"Oh Vanessa you did so awesome\" so yeah that doesn\'t really make any sense and I think she was just trying to make me feel good.    And well AGAIN me and Chels got into another fight this time it was during lunch.  I feel kind of bad through I mean I NEVER meant to be harsh or mean or anything like that but I guess it came out that way.  I just was informing her that I thought maybe she should eat a little more cause sometimes she skips meals and stuff and she is so small and thin and it worries me.  And then she was accusing me of calling her anorexic which NEVER once came out of my mouth at all, it came from her head.  I was never trying to upset her, or make her angry or mad at me, I was only worried.  But she got all defensive and told me she was sick or me and said that \"Every time we talk we always argue\"  which I totally can agree on cause that is EXACTLY how I feel too.  But well I don\'t think me and her stand on high grounds right now, we can never agree and I always feel I am competing about her.  I hate it how she never stops talking about Chem and \"oh I get this, and Oh this makes so much sense.\"  It is SO annoything and it makes me feel really bad about myself because it makes me feel like I can\'t do anything right.  Sometimes I wish Monet had never come to the school this year I mean she is a nice girl but ever since she came along me and Chels can do nothing but argue and fights and today Chels walked right by me in the hall and didn\'t even notice me - Go figure she was too busy talking to Monet to notice I was ever there.  I feel so lost, so empty it was like me and Chels and so much in common and she was my best friend.  But I guess people change, but I don\'t want it to be true I NEVER wanted it to be true.  I don\'t know who I really define as a good of friend as she was, I don\'t really have any other friends that I could share that best friend relationship thing with.  I am confused again right now but I suppose that is the whole point of being a teen.     Thoughts of my Heart  Always fighting, Nothing is right Monet comes along, Steals you away Why does it have to be like this?  Never wanted this friendship to end Something came between us We are different people, different now Different goals, different dreams  A same pure heart though Music taste we shared Fun times together Especially summer, remember?  What went wrong? What did I say? To make you feel mad and angry Everything is this way  I do not know what to do Have no clue what to say Do not know how to heal Cannot figure how to deal  Friends do come Friends do go Why though must it be you Not this time, one more change  Work things out Play with a better hand Get things straight Tell our story  Why can\'t we just work things out? Why give up? It is not that just easy It is the easy way out  And if I had the courage strong To talk to you to find the time Make things better Heal this flood tonight  Things are different now Must admit No one to talk to, to one to share those times together, precious and rare  To only remember tommorow just is One more day to work together Find a way to heal this scar Make this better, lighten this load  Sometimes only you can give Just some choices Seem so simple To die or to live  Wondering if I were gone or pleading for a life not worth Would anyone care? How would I know  Blood so vain Stab through my heart Consume to many pills My wrists are and art  If I were just struggling to make it through would you be there? Would you care too?  ~Vanessa     And lately Alex has totally been coming on to me, and to be honest I actually really sort of like it.  I never thought it would ever be him of all people and I don\'t think he ever though he would like me cause he always seemed to have a thing for Ash but as I said before - people change, times change, things change.  This whole world is like one big circle of change.  And now that I think of it Alex actually would be so my match.  I mean we have our differences but he is URAINIAN! and so am I which is totally cool cause everyone in my family ends up marrying a Ukrainian guy by fait I guess cause no one ever plans for it to happen.  Like my sis caught that cutie urainian Ben which she is getting married to next year.  Melissa is taking me shopping for my dress in December and I am the Maid of Honor.  It is gonna be so cool and I get a purple dress cause that is the theme color of the wedding!  My fav. color!  Man it feels like I have so much to say cause well who else am I going to talk to?!?!...I have no clue.  Chelsey used to be there to talk, but another reason why I just get soooo angry at her and others sometimes is I never get ANYTHING out but when I do get something out - I sure get alot more to come out than planned.  And well yeah usually end up exploding at people.  I wonder if anyone is going to want to eat lunch with me tommorow - everyone always goes to A&W but I don\'t really like it there cause there are no heathy kinds of foods.  Maybe I can drag Alex or Ash to Wendys to have a good ole\' healthy salad with me - HOPEFULLY.  Cause well I have been having lots of hot chocolate lately cause of my cold and sore throat and well during dinner for a couple nights this week I have been like the incredible eater or something because I just can\'t stop eating at night now which is probably why I haven\'t lost any more weight since the summer but everyone keeps coming up to me and saying how they didn\'t recongnize me like Alison that I met in Drama last year she told me today that she didn\'t even realize who I was until like the third week of school.  I didn\'t realize my weight loss was that noticable but HA..after like a million times to figure out how to loose weight the answer is so simple however I haven\'t have the will power lately to keep up with my good healthy eating habits quite often so once in a while I slip in one of those sugary pops that I am sure are like the worst things in the world cause of all the sugar.  But well I have lost about 40 pounds since April which is definetly something that I definetly could not predict a while\'s back but well I guess it was my sis in the first place that got me going.  She started to loose weight for her wedding and then I wanted to loose some because I didn\'t want her being thinner than me and well then me and her lost some weight and now we can fit in each others clothes so yeah when I get to see her I might raid her somewhat of a wardrobe cause she always raids my closet when she comes over.  But well they are living in their trailer in their backyard right now because of the water main break that happened a while ago.  It flooded there house and causes major damage so the house is basically like falling apart now and the walls are caving in a and buckling so they are probably gonna get a new house built after the old one is torn down and stuff and as long as no one disapproves of the plan and stuff.  It is complicated and they are young so it ain\'t like they have alot of money and the city sure didn\'t cover much for insurance.   '),(3713,1807720,'24,September,2003','2003-09-24 00:00:00','    I just finished my long studying and practicing of Chem and guess what I think I understand!!  I am feeling really confident about the quiz tommorow because I was thinking about what my math teacher said that if I had confidence I would be much better at things and right now I feel like I really believe in myself and that I can do anything.     Believing  Yes, Joyful it seems Understanding Makes so much sense I believe  And remember and Listened to the words Gave me advice Told me to be confident  With confidence ANYTHING is possible, achievable Reachable goals attainable  I believe With errors, comes success Disbelief, AND belief Sadness, AND Happiness  All makes so much sense Know I can make it through Have been searching for Found the light  knowing Thinking Wondering BELIEVING  In my heart I know What path I must follow CURVES along the way but we all get there  One way or another All works out, Gods plan Never know what comes next Just keep believing  and I know now Would not give up chem for anything these dreams, goals, aspirations mean something so special  So important, ALWAYS I know where my heart lies Realize what makes me heart beat and I believe that I can do anything  ~Vanessa   '),(3714,1807720,'23,September,2003','2003-09-23 00:00:00','    Ok well I am not sure how the math test went because I am not too sure if I did the last question right (word problem) ya of course ya know now really my thing.  And well Social was alright but I kinda totally blew up at Chelsey and I feel really bad.  It was kind of something like this \"I just don\'t understand, and if you try to help me how will that be different?  Everytime someone explains it I never understand.  I am so sick of never understanding and sick of trying and never getting anything. I never understand anything and teacher probably thinks I am like the stupiest person in her class.  It\'s just really frustrating.\"  Now that I think of what I said I sound so negative towards myself and I wish like yesterday the math teacher said, all I need if confidence but sometimes it\'s so hard to have that because I mean I really am not doing too well in Chem.  My feelings are something like this right now:     What I am Feeling Today  So tired  SICK  Trying to understand  Just do not get it   Frustrated  UPSET  Unhappy,  Underestimated  Wish that I would know   How do I make things right?  Correct mistakes  MARKS  So down, so down below   Is this just me?  The way I feel  Am I the only one?  How can I know?   Chem quiz  tommorow  Studying soon  Feel it is so important  Gotta make it through   Just when I can only think  things just might get better  Worse it seems to my eye  Light or Darkness now?   Crying in the silence of  my heart so deep and solid  breaking up the pieces so  to make them all but one part   Thinking in this gift of mine  What shall I do?  I\'m so confused right now  All I want is to make it through   ~Vanessa   '),(3715,1807720,'23,September,2003','2003-09-23 00:00:00','    Today well I dunno.  OK so I got a 60 something on my math quiz 10.5/16 but well I know it is not the end of the world and well it was just because I didn\'t understand the last question.  My math teacher was also being really nice today and she was so right.  Ya know she was telling me about how I am a really really great student and if I just had the confidence then I could do so much more.  She is right.  But well it is kind of the same for Chem but I am REALLY REALLY trying and well I think my average right now might be like a 56% and anyone who knows me knows that is NOT like me at all.  I am TRYING so HARD, but I really don\'t understand however I understand more than I did than when I began.  I am worried about the quiz for Chem thurs cause I am having diffuculty and the science wing is all closed tommorow cause it is Wednesday tommorow so I am not sure when I will get the help I need.  Anyways I don\'t know what to do about Chem cause well I could drop it and make life easier but I NEED chem for University like the courses I want to do, but well even if I do pass I don\'t think they are gonna want someone with like a 56 some average in there.  I am SOOO CONFUSED and I WISH I KNEW just what I do.  Well after I write a poem about how I feel right now I need to go practice my math for tommorows test.     Wondering Why?      Forever, it seems  One more breath, one more time  Wondering why, oh why, oh why  Is this happening?   Why Chem, why me, why now  Why  Why   Why    Why     Why?!  I sit here and think and wish  Things will get better   I must pass chem  Dig my way through this tunnel  Find the exit  Looking for it, forever   Wandering round,  Finding that escape somewhere  somehow  Out in darkness   Tears unending  Searching  Wishing for not much,  Someone to hug, someone to hold   A person so special  Not too young, never too old  With power so strong  To heal this heart   Passing on days  Like they mean so much  Why are things so hard  Why must this be happening?   And In the end  I still do not know  Will I be happy?  Where will I go?   How will I get there?  What job will I do?  Will I pass my classes?  If only I knew   Forever, it seems  One more breath, one more time  Wondering why, oh why, oh why  Is this happening?   ~Vanessa   '),(3716,1807720,'22,September,2003','2003-09-22 00:00:00','    I thought thing were starting to change but then TODAY happened.  I got a 60% on the Chem test and honestly in my heart I really though I did good, I studied more for that test that I usually study for tests just because I wanted to do good.  I spent half of the night crying over it and the fact that I don\'t understand Chem and my teacher is sorta trying to help but like she doesn\'t make any sense and I don\'t know who else to ask.  I mean I wasn\'t always amazing at balancing equations but now all the concepts are building on a skill I am weak at and I am confused about dissociation and non-ionic, total ionic, and net ionic equations.  Right when I got my mark back there again going through my head was \"why am I on this earth, and what is my purpse??\"  I know I want to teach science so bad by why does reaching my goal seem so difficult right now, I am having those bad thoughts again roll through my heads - the ones about suicide and death, the thoughts I wish I didn\'t have but are so real and deepest thoughts that I can\'t share with anyone.  And well I tried to get help with my math and all but well today I did the quiz and well I didn\'t understand the last question at all.  Go figure a word problem thing but well I couldn\'t get part A right and I couldn\'t even finish part B or C because part A\'s answer was crucial - so basically there goes 5 marks down that drain - five important marks.  I think my math teacher really understands what hard time I am having trying to make sense of everything and I am getting really frustrated trying to figure out things - everything is just becoming so difficult.  I don\'t want to admit it to myself but in honestly reality I don\'t know how I will even make honors, I mean it mean\'s soooooo much to me yet I am trying my BEST right now, my very best and I don\'t know what else to do.  And I was thinking too - do I have a talent, what is my talent.  When I thought about this there was not a single thing that I can think of that makes me talented at something, the only thing that ever made me proud was marks because it was something that was earned through hard work.  But what about now...where is all my hard work getting me to, I mean I guess no one\'s road to achieve all their goals and dreams is perfect but I never thought it was this bad ever for anyone.  I just pray that tommorow goes better, and that maybe I will understand Chem a little more, and maybe my math teacher will help me with my math and understand that I really am very confused.  Maybe someone special will see me and say \"hi\" or maybe they will just smile.  Maybe things won\'t be perfect, but I can try to make things better.  I never really know what the future holds so all I can hope and wish for is that tommorow will be better than today.   '),(3717,1807720,'21,September,2003','2003-09-21 00:00:00','    Well Friday was good.  I so needed that Friday.  Me and Chels had a BIG & LONG talk on the bus and C-Train and went to lunch with Eric.  Then after we went shopping and worked things out.  Later on we went to Matt\'s party which I actually had tons of fun and found out that Monet is actually a pretty cool person like Chelsey said.  I guess I just was a bit jealous before and didn\'t want to try form a friendship with her but now I realize how cool she is.  We might all be going to the mall next Friday or Sunday and maybe have a slummber party with everyone again.  It also felt real good to gorge on some pop and pizza Friday - I haven\'t been all strict about the whole healthy eating lately but well I\'m still doing good with that but well my Grandma made this apple cake and like all weekend long I have been addicted to it.  LOL.  I am like so confused for math though and have my quiz monday so like I might have to bother someone in the morning to get some help and well I am a bit confused about Chem but I think I did pretty good on the unit test so  *FINGERS CROSSED*  because I really need to do good since ya know the quiz didn\'t go over to well.  I have been studying social like all today like mad cause I have to write the M.C. and Long Answer test either monday or Tuesday but I think I will do it tommorow cause I keep thinkin bout it and it\'s starting to really stress me out cause I\'ve got so much more on my mind.  And well yesterday night I went to go see Cold Creek Manor but it was just and OK movie but nothing too exiting and well right now I am havin\' a big break from studying my social and listening to Hilary Duff - and man is she ever great and a good singer and yeah Chels bought her CD on Friday so we were listening to it then at my house before we went to the party.  And well this morning I went to Church and I always feel really good after Church and I feel I still have problems but I know in my heart I just have to keep going on.  And as Celine would say it \"One heart you are following, one dream keeps you wondering\" and it sure is.  Everything will be ok in time.   '),(3718,1807720,'18,September,2003','2003-09-18 00:00:00','    Finally a great day!!  Ashley was all supportive today and like everything was just going right.  I guess since I was trying to have a good attitude it really does work.  I still am a little confused about math but that can be dealt with tommorow and well I think the Chemisty test went pretty good.  My throat is kind of sore right now but I\'m sure that is not too big of a worry and well I felt kind of good today.  You know just one of those little simple nice things that happen in a second they make me feel so special.  Cause well I was walking down the hallway and Mrs. J was there was talking to some other teachers and well she just kind of smiled at me with a kind look.  Man did that even make me feel good, and it is those things that happen in like a second that mean so much.  Tommorow is the CALM carocell and well Matt\'s party is also tommorow and I can\'t wait to have some pizza.   '),(3719,1807720,'17,September,2003','2003-09-17 00:00:00','    Hmm lets see I got 72% on my math test but I could have had done better if I knew a few more things.  But I feel as if the math teacher actually understand now though that I really do need help with some things which is good.  And tonight I studied for Chemistry but well I have been reviewing since the weekend so I wouldn\'t have to cram and well I am kind of putting some pressure on myself to do good because my quiz mark wasn\'t so hot.  And well I had a better day today - didn\'t cry but well Chelsey and Monet do like everything together and I think how that once used to be me and she denies that we are going are separate ways sometimes but I think it is true.  But well I think I worry to much about that and tommorow I am going to try to just be happy!   '),(3720,1807720,'16,September,2003','2003-09-16 00:00:00','    Ok so the day wasn\'t like totally amazing but it wans\'t really bad either and I did cry tonight which is an improvement.  I am still having some problems with Chem but I am working to understand it and I do understand it a bit better and well the test is on Thursay.  But well there is lots of pressure on myself to do good because I got 45% on my quiz and I know it is not the end of the world and it is just one thing but it is very important that I do good on the unit test.  Anyway Ryan won Canadian Idol tonight so I am so happy cause well even since Billy got voted out I have been cheering for Ryan even though Billy was always my number one favourite.  And well I had a math test today and I am not really sure how I did because I had a hard time with word problems like usual but I mean they are just hard and it\'s not like I can just snap my fingers and everything will be fine cause it don\'t work that way.  But yeah things seemed to be going a little bit better today and well I just hope it continues.   '),(3721,1807720,'15,September,2003','2003-09-15 00:00:00','    Here I am again, a little upset again, actaully alot.  Today I REALLY REALLY put my heart into trying to understand chem and I am REALLY trying so hard.  But I just don\'t understand and I went to tutorial, I am getting help from my Neighbour, I asked the Chem teacher like a million questions in Chem, and even read over the chapters in my text that we are studying!  I have no idea what else do if and honestly if I could understand I would, but unforntunatly I can\'t just snap my fingers and understand everything right now.  I\'ve been having lots of irrational thoughts today like suicide and I want them to stop but I am thinking maybe dropping Chem would be my best option right now, but how could I even do that.  I would kill my dream of being a Science Teacher or a Bio Major (Biology Teacher) and well I need Chem to do the things I want to do.  And I  *CAN\'T*  kill my dream, if I were to ruin that it would be like some one stabbing a knife into my heart, and the thought that is demeaning.  I can\'t image doing anything else but teaching and I know in my heart that is what I life and breath for.  If I would drop Chem or fail I just don\'t know where my place in life is, it would feel as if there is no place and no reason for my being here.  I know suicide is wrong but sometimes it feels like nothing will even work out and I cry for what seems like forever wondering WHY??...Why can\'t God understand how much my dream means to me and how hard I am working to try understand.  Why can\'t God help me and WHY isn\'t there an answere to why I just don\'t understand things as good as everyone else??  I just wish there were someone to hug right now, it sounds silly but I wish someone could just understand what I am feeling and look into my eyes and see how much I am hurting.  I wish sometimes things would get better and I hope they do.   '),(3722,1807720,'13,September,2003','2003-09-13 00:00:00','    *Prayer to God*  Dear God,     I have had an interesting week with ups and downs but I feel that my attitude hasn\'t been all that great around my friends all the time.  Please forgive me, for I do not mean to say some of the things I do it\'s just that sometimes things just come out of my mouth and later I regret them.  I am feeling like the best friend relationship between Chelsey and I may be in jeopardy so I would ask if you please just help make everything alright.  I feel we are growing apart, even though she claims that is not happening it feels to me that we are.  But I also would like to thank you God for blessing me with the opportunity so meet so many amazing people and I am thankful for their help along the way.  I thank Mrs. J for trying to help me with Chemistry because I know she didn\'t have to do that, I thank you for her in my life because she is making Chem alot easier for me right now and she is alot more supportive than my T.A.  But I also do thank you for my T.A. because even though she can\'t help me with Chem it is nice to have some else that can be there more me.  I am also thankful for all my friends and I am thankful for having Chelsey in my life still because I really needed someone there for me this week more than ever, and she was there so thankyou.  But God you also know how hard I work at everything, and you know how much it means to me to make at least Honours, please help me to have confidence in myself that I can do this because lately negative thoughts have been going through my mind.  I suppose I am going through a hard stage in my life right now but in the end I know it will make me a stronger person but I ask that you help me through this hard stage right now.  I am having thoughts about my future career again and I am very serious about it but you know how I want to be a Science Teacher well I am not very good at Physics and not very good at Chem and well that is just how things are - but please, Oh God Please, bless me with the gift of Biology if you only had any idea how much I want this to work.  I could not image doing anything but teaching when I am older, it is what the voice in my head is calling.  My conscience is so strong right now and is has been for a while and I have known since grade 7 that teaching is my calling in life.  Please bless me with the gift of Biology because I know with my hard work and effort, and your blessing that I can accomplish anything.     Thank you God for listening.   '),(3723,1807720,'11,September,2003','2003-09-11 00:00:00','    Well today was sort of mixed, I felt happy at times and unhappy at others.  I was happy cause well I got an 85% on my math quiz which is pretty good from me but well I just can\'t seem to get the hang of Chemistry.  I just wish I knew why it is so hard to understand and learn for me.  I feel like no one else has as much difficulty as I do and it is very frustrating.  My math teacher said some nice things to me today which made me feel good because I thought that maybe she didn\'t like me or thought I was dumb.  I do think my self-esteem is a little bit low right now but I can\'t do anything about it.  I just wish, I wish, oh how I wish I just understood Chem though and well everyone thought the quiz was easy and well people after were like \"well that sure was easy wasn\'t it\" and well of course I don\'t want to say that I thought it was hard cause EVERYONE thought it was easy - so just WHY or WHY don\'t I seem to get anything?!?  And well even since Monet came along I feel me and Chelsey are drifting further and further apart.  I don\'t have anything against her but it\'s hard watching my best friend laugh and seem to have a better time with her than me.  Chels always seems to have room for Mike or Monet but never me.  And it hurts.  But well it kind of cheered me up when I found out that I didn\'t have to write the test for the Religion unit cause I finished my work.  Everything this just doesn\'t seem to want to work out though and I feel upset about things - there are just so many things, and things beyond my control.  I will probably get my chem test back tommorow and well I don\'t think I did to well so I don\'t know what I am going to do.  I don\'t understand the Chem teacher at all but I am really trying to understand.  I am having mixed thoughts about my career choice too, I mean I just KNOW that I want to teach something that has to do with Science in the High School level but I am not very good at Physics or Chem and well Bio (have it next term) is the only thing left.  I LOVE Science so much though and I don\'t know why I just can\'t seem to get the concepts of things and understand.  I just wish I was a little more happier, and that maybe I could alot more nicer to others, and I wish that maybe for once I could just UNDERSTAND.  I wish.   '),(3724,1807720,'10,September,2003','2003-09-10 00:00:00','    Well I do think there is something wrong and well I thought that might be like some kind of problem affecting learning or something but well I asked someone close if they thought it was possible but well they just laughed.  Maybe they are right but I feel there is something wrong.  Especially Chemisty right now, I really can\'t take it.  I mean I know I am stressed out when I cry every night.  Today was a bit better than the rest though, and yeah I still cried for like the 6th day in a row but I felt that someone actually kind of  cared  for once.  I was just sitting up in the Science wing and Mrs. J comes up and asks me how my Chemisty is going.  Little things do make a  big difference  because I feel like she actually wants to help me and help me to pass Chemistry.  She offered her help during tutorial which is nice because I am really stressed and I honestly don\'t understand a word the Chem teacher says.  It\'s so confusing, it\'s like she could speak another language and I still would know the same thing as if she were speaking English.  And her directions for the lab we did today were so confusing - I really want to be able to do stuff and not just look lazy but well I don\'t even understand her directions but well I always found it kind of hard to understand directions when people just tell me so well it would of been nice if she could have written them out but I guess no one really learns that way except me.  I still keep wondering why I just can\'t get the hang of things..Chem, Math etc.  I wish there was just an answer.  I feel like I have been so rude lately too or like when someone asks me something I don\'t even think before I speak I just... SPEAK .  And well I sit in like this little corner on the left side of the Chem class where there is like no one to ask when I don\'t get anything, and well it\'s hard to feel like part of the class because I feel so far away - I call it the little island and I have to be careful not to lean to far back or else I\'ll hit my head on the pencil sharpener..lol I actually did that cause well I have to admit I am such a major clutz!  I just wish things would start going a little better and well I wish I could stop crying at night cause I am at the point where I toss and turn for even and like hardly get any sleep.  It will be nice to talk to Mrs. J tommorow it\'s like she is the only person who understands what I feel and what I am going through with Chem, and I pray to God that she can help me so I can do good on my quiz.   '),(3725,1807720,'06,September,2003','2003-09-06 00:00:00','    It\'s like I am going through an emotional roller coaster.  Everything is so much harded this year and alot more stressful.  I don\'t really feel comfortable yet and it\'s really hard especially for Math because I really need help but I don\'t know really how to ask the right way because I don\'t feel a trust level with that teacher yet.  And well her teaching style is really not matched with mine at all - I never understand what she is doing because she will explain something but will not go through the steps of how to do the problem and she usually shows like one or two examples and expects the rest of the problems to be tried out on our own - well that  *DOESN\'T*  work because if I don\'t understand how to do something in the first place I won\'t know how to try it and get it right.  I\'ve cried just about everyday so far this week and well I had a talk with Chels yesterday and that helped me figure out things but there are still other things that I haven\'t talked to her about and that maybe I need to figure out on my own.  I just really hope on Monday my TA or McDougall will be there so they can explain the math to me because that is really my only hope right now.  I feel so energy drained and kind of dumb, I feel like everyone else is good at math and I\'m not, everyone is maybe better than me at chem, and so on.  That is just how I feel right now but I don\'t want to feel that way.  I am so confused and I am having feelings that I might fail math which would ruin everything.  It would ruin my whole life because of course my dream of becoming some sort of Science teacher would be ruined and well I need Pure Math 30 for that and well I have to pass my Sciences too.  I am having second thoughts about my career choice but not that I really want to.  I mean I  *KNOW*  that I want to be a teacher but I don\'t know what kind of teacher.  I really love science, in fact I always have really liked it but I am not the strongest at it, and well Math I would never want to teach anyways, English I am really not strong in reading comp at all and I NEVER understand that stuff so that would be out of the question, and even though in Social I usually pull off a good mark it just doesn\'t interest me the way in which Science does.  So I am just really confused, don\'t know how things are going to turn out, all I really want right now it to just be hugged - for someone just to hold me and tell me everything will be ok.  But I don\'t know if things will be ok or what is going to happen, there is no one here right now.   '),(3726,1807720,'03,September,2003','2003-09-03 00:00:00','    Today wasn\'t that great.  I feel like I\'ve turned evil, I mean I am all stressed out already and feel like I can\'t handle anything.  I feel like I have no control.  I have no idea what to do - I\'m just really confused right now I don\'t know what I should do.  I don\'t know if I should drop any courses or what...I just don\'t know.  I am starting to have some feelings that maybe I don\'t want to teach science, but it\'s confusing because that is my dream and I didn\'t even imagine giving up on it.  And I guess me and chels sorta had like a little fight today about stuff I mean we are going through so much right both and usually we support each other but lately there\'s been no one there and well I have to admit it\'s hard because I sit and cry for about half and hour each night.  She\'s always complaining about her problems and so am I but she gets mad everytime I complain yet she does the same thing.  Sometimes I have bad thoughts like suicide but I don\'t think I could ruin all my dreams like that, but I am just very confused and plan to talk things over with Chels tommorow.   '),(3727,1807720,'02,September,2003','2003-09-02 00:00:00','    First day of school today!!  Lets see it took me and hour and three minutes to get my yearbook photo, half an hour for my id, and about five minutes for my bus pass.  Then I was like starving to me, chels, ash, alicia, and sarah all went to Wendy\'s for lunch and I got myself a Chicken BLT salad with a big lemonade!  Yum!  Then we went back to the school to get something of Chelsey\'s and well we decided to see her sis and then we saw Mrs. J which we totally admire and think she is so cool and awesome.  It was nice to talk to her and see her again.   '),(3728,1807720,'01,September,2003','2003-09-01 00:00:00','    I had the biggest and greatest sleep yesterday night because it was technically that last time I really get to sleep in because school starts  *TOMMOROW* which is quite a scary thought when you think about it.  Today I helped my Dad and Ben unload the van and the trucks to store their stuff in our garage and well I am just kind of taking it easy today and hopefully I will get a hold of Chels to talk to her for a while.  And well I feel like I am the queen of eating or something like that because I had breakfast and then like an hour after that everyone was eating lunch and desert so I couldn\'t help but stare at the food so I decided to eat lunch to.  Little did I know though is that I would end up eating pie and ice cream too.  Hopefully I won\'t gain too much weight from that but I should be fine as long as I don\'t keep that up.  Now I think I am going to go and enjoy my last day of holidays!   '),(3729,1807720,'31,October,2003','2003-10-31 00:00:00','    On Wednesday night over to thursday night I had well the most interesting dream.  Usually I don\'t dream alot during the school week nights and stuff but this dream was long, sorta made sense is some ways and not in others, but I do feel it had a message.   My Dream that other night  My mother and I were going to this friends of my moms that lives in Ogden.  I remember first standing outside and then looking to the house on the left next to it.  I looked in and there was the living room with a couch and a TV but the on the screen was just that static grey kinda picture.  I remember going into my moms friends house and she took us to the backyard.  Then I remember seeing one my sisters rabbits and I said \"Oh Adora\" thened I look at that house on the left and Mrs. J was sitting there.  She looked at me as if she knew I was going to be there and said \"How was Friday\" and I just said \"It was good\" and that is all I remember from that seen.  Later on I was at a park that was close to the houses, but well I don\'t know how I got to the park.  Anyways there were these two swings and I sat on one of them and beside me on the other swing was this older woman.  She wasn\'t like wrinky and granny old but like she was \"older\" you know.  This lady seemed familiar and had like sorta yellowish, goldish, blondish, sorta whitish kinda hair - basically it was like a blonde but a mix of different colors but I am pretty sure I have never met her or seen her in my life.  She sat there at started telling me about Mrs. J\'s past.  She told me that the year before she taught at my school nothing in her life had been going right.  She said that she was at the lowest point in her life and she didn\'t think things could get better and the only thing that kept her going was her kids in her life but then the lady continued and said that she was going through a really rough time in her life and a really \"hard stage\" in life.  Then she described to me that we all go through these \"hard stages\" in life at one point or another.  This dream was like odd cause well I hardly ever dream and I swear that dream felt like hours and hours.  I think it was kinda like a message to me though.  Symbolism I guess.  I think since I am having a tough time right now my dream meant that it is normal.  That sometimes people just feel that way - we all feel that way no matter who we are, where we come from, what we look like.  We all feel pain, we all feel sadness.  That seriously is the most interesting dream I have EVER had.  And like it sorta makes sense but sorta doesn\'t and well I don\'t think it is true although it could be - I just think there is like stuff on my conscience and somethin\' is trying to get some kinda message to me. ___________________________________________________________________________________________   I went to Megan\'s party and slept over.  OMG it was SOOOOOOOO fun and I met lots of new people.  I feel a transition happening.  Even since I met megan my friend range has expanded and slowly I am being excepted into the more popular crowd.  I mean I like my friends popularity or not, that was never an issue.  But it is also real nice to like have other friends too.  We watched so many movies, we watched The Ring, the Rocky Horror Picture Show, Labarinth that dumb movie for kids cause David Bowie is in it and we like watched all the parts that he sings in [Oh he is a hottie] and we watched The Little Shop of Horrors and stuff.  Then we played like Cranium which was awesome and funny and great and we ordered pizza and stuff and ate like probably 40000 calories worth of junk too.  And well she lives in Mackenzie and holyman her house is HUGE.  Anyone who lives in Mackenzie does not know how lucky they really are.  And well yeah my mom had to go to faith day today cause she works for the Catholic School District thing or whatev. you wanna call it.  So yeah and then when she got back instead of having lunch the people she works with she left so she could have lunch with me and she took me out to B.P. and we ordered pizza...yum.   Then we went shopping to safeway and she bought me cupcakes so I can pack on another 5 pounds which I will have to loose like next week when all the halloween candy is gone and there is no junk left in the house.  LOL.  Well I am going to watch Survivor now cause I taped it from last night.   '),(3730,1807720,'29,October,2003','2003-10-29 00:00:00','    I don\'t know how to describe today.  There is no one word that comes to mind.  I didn\'t want to wait like the hour and some minutes for a stupid bus to come so while Chels, Andy, and Eric all went back to Eric\'s house I decided to walk to school.  They thought I was crazy but like it is just snow and a little windy.  And well I didn\'t want to really fess up that I love school and for the life of me I HATE being late for things.  We had a quiz in chem today which I hope I did well one but some of the answers just didn\'t seem right so I will see how I did tommrow or whenev she gets them marked cause she is pretty sick right now.  I am just glad that she was there today and not that sub. guy we had yesterday who had no idea what he was talking about.  At lunch we played asshole again for like a million times and then just before the bell rang me and chels kinda got mad at each other AGAIN but I am not explaining the story cause I am not in the mood for thinking about that stuff.  Today was good I guess, everything felt kinda odd though.  Felt like I was dreaming or something, but I suppose I am just not use to the snow and stuff.  And well I decided I would use my one re-write for the math test tommorow.  I have the \"feeling\" that that is what I am suppose to do.  It\'s my conscience I think, I have a very very VERY strong conscience.  Cause well if it were not loud and strong I would have given up my dream of teaching high school bio/science a long time ago, but I can\'t give it up.  It\'s my passion, the one think my heart wants, the one think I know that in the end will make me happy.  Well yeah that was today.     Conscience  Hear in this night my conscience is speaking â€œTake your time and live for todayâ€ Be yourself and donâ€™t be anyone else You are who you and there is no doubt  Dreams will be dreams until they make reality Turn them into the passion of your life Life for today but remember tomorrow Here in this life my conscience is speaking  ~Vanessa   '),(3731,1807720,'28,October,2003','2003-10-28 00:00:00','    Today was pretty good.  I did the bonus lab for chem at lunch so that kinda took some time away from my lunch but well I would rather have bonus marks.  And omg we had a sub. for Chem today, now that I think of it I had a sub. for all of my classes except for one.  Anyways the sub for chem didn\'t even know what he was doing and it took like half an hour to balance an equation and then he explained to us percent difference the wrong way, he was actually explaining percent yield but I don\'t think he knew that.  Then he was talking about SATP but on the board it was written as STAP...LOL!!  I got the biggest laugh out of that and well then he was like \"Oh and can someone explain to me how I would figure out so and so kinda problem\" and then someone answered that they would use Gravimetric Stoichiometry and he was like \"Well I don\'t understand that.\"  I hope my actual chem teacher is there tommorow cause I don\'t want another class with someone who doesn\'t have a clue about things.  My chem teacher is actually pretty good, smart, and nice.  Sometimes I underestimate her I guess.  And well I have gained 10 pounds in four days.  Scary as it sounds it is true.  I can\'t stop eating.  When I think about chem I eat.  When I think about not eating I just want to eat more.  Chocolate bars, timbit things from timmys, and I even had a doughnut tonight.  I can\'t stop but I want to.  I used to be sooooo good with the healthy eating but then everyone is like \"vanessa eat this, eat that\" and now I can\'t stop.  I am out of control.  I don\'t want to starve myself but tommorow I really can\'t eat ANY junk.  I am starting to worry.  Before I just wanted to be thin and on some level I still would like to be thin but I am more worried at being at a healthy weight.  I just want to be healthy and yeah thin would be nice too.   '),(3732,1807720,'27,October,2003','2003-10-27 00:00:00','    I went to the womens show yesterday and it was great.  I got to watch Billy Klippert perform live and then I got his autograph afterwards.  It was soooo awesome!!!!  I loved the free food and drinks.  And well I got to watch a fashion show and a hair style makeover show thing.  It was soooo cool and it was the first time I had ever been to a womens show and now I can\'t wait for the next one!!!  I also bought myself another charm from my bracelet, it says \"dream\" on it cause it is to remind me that everyone had to start somewhere and they started with a dream.  I think my mom is getting me a hello kitty charm for Chrismas but yeah that is still a while away.  And well today was actually good.  We saw Mrs. J in the morning and when she smiled and said \"hi\" that just made my day.  I was all cheerful and happy after that, it was kinda odd but that lady sure has something special about her.  She has this remarkable gift to just cheer people up.  And well I also got Billy to autograph a thing for Chels to and I ever got him to address it to her and she like loved me and gave me a big hug and I have never seen such a hugh smile come over her face.  Me and Chels are usually arguing but today we made peace.  Like I said we were all cheerful and happy today.  I was a little upset about extended TA cause I really needed help with Chem but then I was like about to start crying in chem cause we couldn\'t go to the lab until we figured out a calculation thing but then Paige came over and helped me with it, told me things would be alright, and told me that if I wanted she could even help me prepare for the quiz.  I think though that I could probably do the quiz really well it is just that I have a hard time with the basic skill of balancing.  That I find very very very hard and also like to write out the equation when it is in a word statement.  I got a 62% on my math test but I think I will use my re-write for the test cause I just have a \"feeling\" like this is what I should do.  There is like a voice in the back of my mind telling me to re-write it.  At lunch it was also really fun and I got to take my mind away from everything.  We played Crazy Eight\'s, then Poker, and then we had a few really fun and awesome games of asshole.  Social was like sooooooo boring but Chem well in the lab it was pretty entertaining.  Our chem teacher sat on like one of those stools in the lab and like she broke the chair leg and she was like \"whoa\" and also feel on her ass.  Oh man that was like the FUNNIEST thing ever, I like could not stop laughing for a good minute.  So yeah things are pretty good for now.  I mean classes are still hard, social is still boring, life right now is great even though I keep eating all the halloween candy and chocolate but I guess I have to live a little.  I have to admit it sure is good!  Well here is a song/poem I wrote about how I WAS feeling, but I think the feeling is slowly passing.  Who knows what is happening this time.  Maybe I am just \"adjusting\" to this crazy school life.  Hmm only God knows I guess or my heart - or course my heart can\'t talk, only my consciece can.      What I have to Say  A broken heart, wandering dreams Through the darkness in the latest hours Flames are fading, the colour is erasing Wish I could take this all away  What I am feeling, feeling today More than words could ever say Pain in this heart it canâ€™t be pushed away So you listen to what I have to say  This is today this is today, today, today, today Wandering away wandering away, away, away, away Thinking of death, just one choice, the choice of suicide Have no clue where these thoughts in my heart reside  This is today this is today, today, today, today Wandering away wandering away, away, away, away Looking at her looking at my future world that I want to discover Thinking how she could only be a true light and a hope to me  This is today this is today, today, today, today Wandering away wandering away, away, away, away No one just knows the thought it my head They truly are the ones I dread  This is today this is today, today, today, today Wandering away wandering away, away, away, away Wish I could talk and I wish I could share But sometimes I donâ€™t know who would really care  This is today this is today, today, today, today Wandering away wandering away, away, away, away Thinking what if my dreams donâ€™t come true I wish, I wish, I wish I only knew  A broken heart, wandering dreams Through the darkness in the latest hours Flames are fading, the colour is erasing Wish I could take this all away  What I am feeling, feeling today More than words could ever say Pain in this heart it canâ€™t be pushed away So you listen to what I have to say  ~Vanessa   '),(3733,1807720,'25,October,2003','2003-10-25 00:00:00','    I think I am gonna gain weight this weekend.  I wish I could stop obsessing about it but that is all I can think of.  I don\'t think I will eat tommorow and well the whole week I think I will just \"restrict\" myself from eating anything other than fruit, fruit, and more fruit.  And damn I knew I shouldn\'t of had the Chinese for lunch but I have to admit it was good but probably cost my about 5 pounds.  And then I went to Kirsty\'s party which was fun and all but we had chips, chocolate, pizza, and cake and then I had to come how and have another dinner!!!  I wasn\'t even hungry but me, my mom, my dad (flew in from Vancouver this weekend), Ben (Sisters Fiance), Melissa, Bens mother, and Bens mothers boyfriend all came for dinner and I *HAD* to eat again cause it would be rude to sit there and watch everyone else at the table.  So I barely ate but I had to eat something so I had like three meatballs and some veggies then I had to have dessert cause company made it and my Mom told me to eat what they brought.  I can\'t believe I didn\'t throw up from eating all that, I have never felt my stomach so loaded ever in my life.  On friday we all went so Starbucks and yeah actually I had more junk Friday which is most likely why I am gaining some weight.  I didn\'t want fast food but no one wanted to come eat a salad with me and I didnt want to eat alone for I just went to A&W to eat cause that is where everyone else wanted to go.  Then at Starbucks Travis bought me a chocolate brownie Frap. which was sooo yummy but I hardly get those cause they are like calorie loaded and later everyone left.  Apparently we \"wen\'t out\" really for the first time.  We played arcade games for a while which were actually really fun at the coliseum and then we went to go see Scary Movie which was absolutely hilarious.  He asked if I wanted to go for dinner and well I would have wanted to go if I had been hungry and hadn\'t eaten fast food earlier but I have to watch what I eat more than others.  Other people eat whatever they want, if I ate whatever I wanted all the time I would gain 20 in like a week.  So then after he just walked me home and well he is actually a pretty good person.  But now that I got to know him a little bit better I think of him as ONLY a good friend and nothing more than that cause I really can\'t see me dating him.  I don\'t like him in that way.  And well it does kinda of bother me that he is not really too serious about school cause I am, you know like I have goals, hopes, and dreams and I want to do things with my life and go places.  It bothers me a bit that I was talking him and he had told me that he cheated on his Science test and got caught.  I mean I am sure most people have cheated and yeah in my deepest darkest secrets yeah even I have cheated but not recently.  I mean I would never do that in High School and I don\'t really know why someone would - well actually maybe I might.  Now that I think back...Grade 6 the worst yeah of my life when every person hated me for no reason and I was the ugliest thing on earth and was always the \"bad\" kid that always got detention.  Anyone who knows me now would probably be suprised to learn that I was not a very well behaved kid.  Cheating for me though was actually I mean the act itself was bad but I cheated on three of my grade 6 achievments - all but the math test.  And well I only did it cause I was soooo scared of failure and maybe that is what gave Travis his motive.  But well cheated boosted my confidence and I really believe that I could do anything - and well my marks as a kid where never great no matter how hard I work so lots of things I found pointless.  So well that next year I worked my ass off, behaved myself, and had no detentions.  People started to notice me more, and were drawn to me.  Finally I was accepted for once with a good range of people.  So I mean I know Travis did a bad thing but there is ALWAYS something good that comes out of those things even though sometimes it may be SOOOOO hard to see the good.  And well I am proud of myself.  I got into the van today and drove for the first time since my drivers ed experience.  I kept thinkin bout\' drving so I drove.  But it was different this time.  This time I wasn\'t scared of the driving, or the road, or others, or anything.  I have to admit my turns are a little sharp and my judgement needs a little improvement but I am sure that will come eventually.  It takes longer for me to usually learn things anyway and I learned the hard way that sometimes things we have to work our asses off and yet that doesn\'t make us any wiser or better at things.  Although I was proud for driving again, my mother well she isn\'t helping anything right now.  I asked her what she thought of my driving and if she thought it improved [and yeah I would be the one to know if it has improved or not] and she was like ummmmm blah, blah, blah and started listing all this bad stuff about my driving.  Then I asked if I could drive home from the party cause well I drove there and she told me \"no.\" And well like if she won\'t let me drive then how can I practice and get better at driving?!?!?  Like duh you just do wake up and your a better driver, although sometimes I wish it worked that way.   '),(3734,1807720,'23,October,2003','2003-10-23 00:00:00','    I got 73% on the chem test!  Yay!  I studied really hard for that and well that mark isn\'t like amazing but it is not really bad either and right now I am just aiming for over 70% cause that is what is reasonable for my abilities for chem.  My average when from a 67 to a 68.89 or something like that which means it will be rounded to 69.  I mean that mark still isn\'t too great but at least it went up.  Afterall I would much rather see my average go up then down anyday.  However, the math test is a different story.  I worked my ass off to understand that unit and I really did believe I understood it but some of the questions just stumped me and I didn\'t really know how to do em cause I got confused.  I really need to boost my math average cause it is at a 70 but with the next unit test mark going in I can only pray, and hope, and wish that maybe I did do good on that math test and I am just underestimating myself.  Afterall I have done the very best I can, and truly believe in my heart that I derserve over and 70% in math and in chem.  I have worked my ass off to understand the stuff to the best of my ability and I have practiced math over and over, I don\'t know anyone who practices their math as much as I do.  I seriously don\'t even know who goes to as many chem tutorials as I do or who even studies as much as I do for that stuff.  I deserve way over a 70% in both things and sometimes it kinda bothers me cause well I KNOW that I have done my best but sometimes I wish my \"best\" were better.  I want to make honours more than anything and if I don\'t yeah I think I probably will cry.  I deserve to be on at least honours - I deserve it.  I have worked so hard to this point and it would be disappointing to see all my efforts and hard work go to waste.  I guess I should stop worrying so much about it right now though cause interm/mid reports and TAPS are not till ike mid November I think and from now til then I am gonna have to keep working hard and putting in the effort.  And well in the morning I told my TA *ALL* about what I thought of the schedule and stuff and we were getting into like a little bit more detailed conversation than she asked me if I was eating properly.  And well like EXCUSE ME but like that is my business and stuff.  Besides I don\'t skip many meals during school anyways and really should should stay out of stuff like that.  Besides I have been eating junk lately cause I have been worried about Chem and Math.  Today at lunch I hate 2 of those mini-chocolate bars and then my mom bought me two cookies from Tim Hortons.  I don\'t want to gain weight I see all the others girls so thin, I want to be like them and look like them, and yeah don\'t get me wrong I would like to also wear a size 2 but well I would be happy with size 5 cause I have giant bones so I don\'t think size 2 is realistic.  And chels man I am getting sick of her.  I hate to say it but it is SOOOOO true.  She is becoming self-centered and like she is always like \"Everyone always comes to ME, Me, Me\" and \"When they need to talk they come to me\" and blah blah blah.  I can\'t take it anymore!!! ARGGGG!  I am sick of hearing her complain and work, and chem, and people, and all the other shit.  I mean I complain too but not nearly as much as her and she is always in a bad mood, I hate it!  Eric is starting to get fed up too.  We are all annoyed.  She was my best friend and on a certain level still is a \"friend\" but we are not friends like we used to be.  It is her who has changed, certainly NOT me!!!  Everyone has noticed it too, and it is not only me...it is EVERYONE!!!  I feel like I have grown up while she is still stuck in her whatever-you-want-to-call-it-but-it-seems-like-she-is-so-imature-and-annoying-and-whatever Land.  So yeah there are alot of things going on right now and yeah some good and some bad.  I suppose I am still pretty happy though cause well I raise my chem mark a bit, but still a little worried about math.  I suppose it isn\'t the end of the world.  I am happy and worried...I don\'t know if that is a good thing or a bad thing.  I should probably go eat my dinner now but I am not really too hungry but I want to eat anyways.   '),(3735,1807720,'22,October,2003','2003-10-22 00:00:00','    Today yeah was good I guess besides my compulsive impulse to eat junk.  But I kinda resisted and got some pizza and an apple for lunch which I guess is not too bad when I think about it, that is not too junky or calorie loaded.  I am not too sure how the Chem test went cause somethings I really knew and other things I wasn\'t too sure and could not remember but I really need a good mark in that stuff.  And well since today was Wednesday and there are no Science tutorials I didn\'t go to any tutorial cause I wanted a *real* lunch so yeah it was nice to have *real* lunch and actually have a good *talk* with others.  Matt mentioned to me how he thought I was attractive and how alot of the other guys I know think I am attractive too.  That makes me feel pretty good and well I think I help Travis feel alot better today too.  He was really upset and disappointed in himself cause he cheated on one of his tests and stuff, and he feels really bad about it and does not even know why he did it.  I tried to make him realize it is not the end of the world, and people do make mistakes and things like that do happen and it is reality.  I think I helped him feel a bit better though cause I just cared and stuff so I feel good about that.  I gotta go eat my dinner now then I got a big math test tommorow and I need to be well prepared cause I want to boost my average.  I have been working hard enough that I really do feel like deserve it, I deserve for my Chem average and my Math average to both go up.  I always go to tutorial and get help when I need it and stuff, but well today I took a break during tutorial.  I need to re-energize myself somehow and sometimes we just need to talk to others and have a good lunch with others.   '),(3736,1807720,'21,October,2003','2003-10-21 00:00:00','    Again today was yeah...pretty good.  Well besides the morning.  I woke up crying because I had a dream about drivers ed and driving and well that is kinda what I am avoiding.  After drivers ed the lady I well didn\'t like very much and well I am not too good at driving.  Everyone is always talking about \"Oh I want to drive\" and \"oh driving is so easy.\"  Well it bothers me cause I find it really difficult to learn how to drive and well I feel the pressure to drive, and get my license and stuff.  I feel pressure but I know deep down that is not what I want right now.  Just this morning I did not want to wake up thinking about that stuff but I did, dumb dreams.  I guess they really are a reflection of the conscience.  Oh and Travis asked me today if I wanted to do something friday so of course I told him sure.  I mean why not try something out.  I have nothing to loose, afterall Alex did not turn out to good.  Well I have to go eat supper and then I have some major studying to do for chem.  I need to boost the average and I know that I can do this test.  Note to self: You can do it Vanessa!  You can do anything!   '),(3737,1807720,'20,October,2003','2003-10-20 00:00:00','    Today was a good day.  I talked to people and yeah they talked back.  I was happy.  I smiled.  CALM was even fun, me and Andrew talked alot [yeah so me and Andrew went on like one date and that went bad but yeah we are still good friends]  I even for the most part understood my math except for a few questions in the homework which my calc. was being a pain and would not calculte my decimals numbers to fractions.  I even found out Travis kind of likes me and then later and and Chester had one of our \"friend\" talks although Lenny seems to claim that I was like \"trying to pick him up\" LOL [that makes me laugh so hard] I have always sorta had this thing for chester but I think we make better friends ya know.  We are just those kind of people.  And well hmmm not sure how the Chem is going now, sometimes it is good, sometimes bad, sometimes just ok.  It\'s like being on a big roller coaster with all the chem students [I can actually picture that] and like I have people asking ME [yeah me...haha...this is when you laugh] and they were asking Me how to do things.  Ok...ya so I am not the only one who does not understand so I don\'t feel so bad.  I wish I would have not gone onto the U-courses sight tonight though to check my chem mark.  Down 1% from before [and yeah it may not be a big deal to anyone else but that 1% does happen to be a big deal to me] it was my worst mark and still is.  I HAD  67 but now I have a 66, but I just don\'t get it - like I am trying and all, going to tutorials, and putting in all my freaking effort, I just don\'t get WHY WHY WHY chem is not going well.  I feel like I work hard enough and study enough that I really do deserve an 80, of course in a perfect world I probably would get an 80.  Now I really have to think hard though.  I want my dream of being a science teacher more than anything [since I HATE chem now I would probably major in Bio] but well I don\'t think I could drop chem, besides I am way too far now and all that work for nothing...I don\'t think so.  I have to pass.  Just keep reassuring myself that I can DO ANYTHING, but it\'s hard.  I have to ace that unit test on Wednesday, at least get over 70 so my ave. should go up a bit, and over 70 is not an unrealistic goal.  I studied lots on the weekend and will do alot more tommorow night, tonight I just practiced the problems and stuff and worked out the math stuff from this unit.  Yeah well I am pretty happy besides chem, that is the only thing that is really dragging my down right now and upsetting a little part of me.  I will keep working at it though.  I suppose that is all I can do right now.   '),(3738,1807720,'19,October,2003','2003-10-19 00:00:00','    Ya so I fisished my book.  The end was the absolute best!!!  I want to read another book (scary thought - can\'t believe I just said that)  And yeah I feel good today again.  I went to church and well the feeling afterwards I love.  I think a good ole conversation with God was just what I needed.  And well when I got home from church I practiced my math for a while then had an apple (yummmm, I think I am addicted to fruit) [Scary thought when I realize fruit is actually healthy] and well yeah now I am back on my comp cause everyone works sundays and like there is never time to do anything.  I\'m like sunlight deprived or something and well yeah walking hope from church it was nice with the sun rays just bouncing off my face.  Kinda weird actually the sun put me in a better mood.   The chem teacher told us we might have a quiz monday (even though we just have one of Friday) so yeah I hope she decides not to have it.  Cause well lets see all the things I have crammed into my head: Chem, Math, Social, What I am making for dinner, how good my book was, friends, influences, Tv shows, lets see TOO MUCH!!!!  GRRRR and I just got one of those sales call things: this is sears blah blah blah survery.  Like K I don\'t want to hear that!! - It should be like a law that only people that know you or are your friends can call!  And yeah when I was practicing my math I realized that uhh I think I understand most of the stuff in this unit!!!  For once I understand alot of something that was really hard to learn, oh how proud of myself I am right now.  Now I just have to get that mark I need on the unit test thurs and boost my math mark.  I have a 70 right now but the way I work and effort I put in I feel it should at least be an 80 so yeah I\'m gonna try get that 80 I deserve.  As for chem, well all I can really do is study like usually and practice, practice, practice those calculation problem things still my head hurts and til I WILL get them right 100% of the time.  I think I CAN do good on this unit, but it\'s REALLY HARD to understand and yeah alot of times it is very frustrating and I honestly HATE crying over dumb stuff like that all the time.  But well I NEED chem for university so I am just gonna have to suck it up, get through the year, and I try to understand and do my best.  And well I am planning to get Chem over and done with doing summer school when it rolls around again.  If I do chem 30 during the summer that means no Chem during the regular school year [yeah I sure would like the idea of that wouldn\'t I]  I think the next Joe Millionaire starts tommorow.  Oh how excited I am - makes me want to jump up and down.  Of course if I was a kid that would be ok, but yeah I am think people would start to stare at like an older perosn.  And well I want it to be like almost Halloween.  I CAN\'T wait for Megans party!!!!  Like seriously we are gonna have a big sleepover with lotsa people and I CAN\'T wait!!!  Megz is like one of my close friends now(I would call her one of my best friends right now cause chels well hasn\'t been my \"ideal\" of a best friend lately) and Megz and I met in Drama last year.  In gr. 9 we both got medals for Drama even though we went to diff. schools.  That is like wicked.  And we all best pals in Drama Queenism forever!!! Lol [she wrote that in my yearbook]  And well mentioning yearbooks I have like closed eye syndrome in like almost all the photos I am in.  The only pics with my eyes open are my portrait one and the ones on the band page.  You know I really hated the thoery in band last year, and well the teacher for band last year doesn\'t score alot of points in my book, but I kinda miss playing my flute.  Megz wants me to join choir next year and chels always told me too take it.  I can sing and all and people that do hear me like drop cause they love my voice apparently but I dunno I think I will consider it.  I mean besides it would be someing FUN next year [A fun class besides Drama - now we\'re talkin\'] and yeah I think it would be kind of fun.  I am not sure if I wanna do the band thing again cause it\'s more theory than choir and with my load I think I will need less.  I can\'t even believe I am in gr. 11 now.  I am like getting old, and I will like serious cry for the longest amount of time when high school is over cause man even though life can be a bitch sometimes and all the times I go through.  School is like my second home.  I love school and ya know what I don\'t care bout anyone who thinks differently.  Lots of people call me crazy for thinking that but ya know what I got goals, hopes, dreams.  One day I will be at the front of a class and I will be teacher.  Yes, I will cause that is my passion.  The one true thing I have wanted to do since grade 7.  I could not imagine doing anything else, that would be like a sin.  Yeah and I think I am gonna get off my comp now cause it\'s cold downstairs where my comp is and NO light - I need light, thrive on light now.  I feel really good right now.  Chelsey went away this weekend - no friday with chels.  It felt kinda good though.  Instead of competing with someone else I could be me, people focused on me, talking to me, concerned for me, I am not self centered or anything but it gets annoying when CHELSEY always comes before I even will.  That honestly does both me.  But what the hell I don\'t need to be worrying about things like that.  If I always worried about things like that then I\'d think I would go crazy!!!  [yeah vanessa as if you are not crazy now?]  Well yeah this term in school is really stressful and I do want high school to last but I *CAN\'T* wait for next term...ya know what that means...DRAMA DRAMA DRAMA.  I am born to be on stage [now that I think about it teaching a class would be like the same thing] which is a neat realization.  I am a Leo (ROAR - lol)  Ok so that is like the reason I was born in August or something.  Now that I think of it getting into a Drama was not really even my choice, it was an accident, or well it makes more sense to call it \"fait\" or \"destiny.\"  I still remember at the end of gr. 8 we had to choose options and well I choose the one\'s I wanted and well never thought about it.  The next year in gr. 9 I got my schdule and was like all pissed off cause I didn\'t get like any of the options I wanted.  I made me soooo furious and Drama was on my schedule.  At first I didn\'t really like it, in fact day to day I claimed \"I hate this, I hate that\" but by the end of the yeah on of my fav. things we had to do was our monologues.  I blew everyone away - literally.  I went first and after that everyone was scared to go up cause no one wanted to follow after me.  After that I don\'t think I ever recieved so many compliments in my life.  Then I came to a realization and thought \"You know what I spent all this time saying I hated Drama, well ya know what I really LOVE drama, it\'s great.\"  So then the end of the year rolled around and they were givin out medals for \"Top marks\" and I got the \"Top Drama\" one.  It was cool, it still hands with all my medals from soccer in my room [soccer medals is a whole other story - got 14 medals for that] but yeah the Drama medal has my name carved on the back with the year and my grade at the time which was gr. 9.  I thought it was like the best thing even cause even though I have earned like a million soccer medals they don\'t engrave your name on the back.  And well so then I took Drama in gr 10 and LOVED it so now I am gonna do it till the 30 level and cause of \"fait\" I met Megz and like expanded my range of friends.  And well my hand hurts now from typing but I didn\'t realize I had alot of stuff to think about of vent about or get of my chest.  But it feels better.  Things feel better.  I feel better.  Chels better now ruin this mood tommorow - I won\'t let her.  She always ruins my happiness to make herself happy but not tommorrow.  Not this time.  I am gonna be me, happy me, old me, confident me, the girl I used to know.   '),(3739,1807720,'18,October,2003','2003-10-18 00:00:00','    Well I am not like the most happiest person alive right now but I feel better than I did on Thursday.  That like \"How\'s chem going?\" conversation thurs at lunch was some kinda miracle cure to my sadness.  On Friday I was worried about my Chem quiz but it certainly went way better than I thought it would.  Oh goodness I just got one of those calls again - you know the ones where they think they got the wrong number but they sound like someone you know.  Man that totally freaks me out and it usually happens when I am at my comp typing around like the same time!!!  I got a call like that once last year, and I could have sworn it sounded like one of my teachers..like way TOOOOOO freaky!  And well the one I just got kinda sounded like my science teacher from last year but it probably wasn\'t cause well chances of her getting the wrong # and calling me - ok yeah that just would not happen cause chances are too slim.  Man I think she has had to big of an influence on me.  LOL.  But well she is like the BIGGEST role model in the world and I would kill to be like her.  She has the ideal job, two kids, married, smart, pretty....everything I have always wanted.  But well yeah I am sure I will have it all some day too and someone will probably wish to be me.  Anyway yeah I am feeling really happy and good and stuff.  I feel really happy.  That\'s a change from the norm feeling of grade 11.  Also I am like not a big reader at all but Sarah and Ali both read this book called \"Speak\" and said it was real good.  And if anyone knows me they know that I would NEVER read any books outside school, I think I started one once or twice but I never finish them.  But I bought this book \"speak\" on Friday - that\'s right *I* bought a book (yeah the world might end) LOL and I am already half way throught.  It is SOOO good.  It totally relates to teenage life and like the whole time reading it I was like \"Ya this author sure knows what teens go through cause she hit it dead on with just about everything!\"  So yeah I want to finish the book soon cause it is sooo good...and I think I wanna read another one after this.  What is wrong with me?!?!?!?!...AHHHHH I am interested in books now..ok that is just way too strange!!!!   '),(3740,1807720,'16,October,2003','2003-10-16 00:00:00','    This morning I woke up feeling bad about things and just kinda upset about everything in my life right now.  The social test I thought did go actually pretty good and well I understood most of the math today.  I still haven\'t told Alex that I am not really interested in him in the way I thought I was.  I don\'t know how to tell him and I don\'t want to upset him and well thatt is actually one of the things I am least worried about right now.  After lunch today though I felt a bit better.  My science teacher from last year just stopped to ask how chem is going and I felt for once she understood how I feel about like stuff like chem and stuff.  I don\'t think anyone really realizes how stressful it can be to try really hard and not be able to understand.  And well I wish I knew why I don\'t always understand and things are really hard but I guess I will have to keep wondering.  I went to chem tutorial today and got some extra help the equations for boyles law and charles law and well I think the combined formula is pretty easy but I think the ratio thing we did was kinda confusing for me but the chem teacher didn\'t seem to understand where I was coming from cause she kept saying it was so easy.  She needs to realize that not everyone finds the same thing as easy as others might.  After Mrs. J little short talk with me after lunch though I actually kinda felt better about things.  Strangly enough and well tonight I went home and just took my time to go through my chem before the quiz.  I understand it alot better and I didn\'t cry tonight, and tonight I actually feel glad to be alive.  Survivor is on at 9 and well I found out today that the Clay Aiken Cd came out!!!!  I love Clay!!!  He is one of my role models so yeah.  Hopefully, after all my studying and hard work I will be able to do good on my Chem quiz and ya know what I believe in myself right now.  I really think that I can be successful and do good.  I feel like my mood has kinda been uplifted and for once I have a real good feeling.  I feel good inside which certainly is a rare feeling that only comes along every now and then.  I feel happy, and glad, and important.  I feel loved, and cherished, and like I can do the things I put my mind to.  Well it certainly is like the opposite of yesterdays feeling but I like it right now.   '),(3741,1807720,'15,October,2003','2003-10-15 00:00:00','    Well today at school I didn\'t really notice if I was happy or upset, I was like emotionless or something.  And well me, chels, eric, and andy all walked to school today cause the buses were all backed up but we got there in half an hour and beat the buses.  But well yeah tonight I just started finishing up my math and well then I got to my chem.  And well the chem teacher seems to think we have no life or something or she must think our lives should be chem cause I just say there starting at like a million questions and she have us like 2 examples and like no offence but I DON\'T learn from seeing two examples of something.  But I guess I realized how overwhelmed I am with everything and how stresses I really am.  It was apparent to me tonight because tonight was just different.  I have never felt that way I have felt tonight even.  I was crying on my bed for about half and hour and then well I went to the bathroom to blow my nose and stuff and well I just looked at myself in the mirror and said \"I just want to die, I want to kill myself\" and then I went through my drawers looking for a fresh razor.  I picked it up and just held in there my wrist.  I don\'t know why I did this, cause well I have never even come close to even thinking about that stuff and well as I held it there I felt a calmness.  It was strange cause all my fear it was all of a sudden gone, and everything just felt ok.  I can\'t really explain why I felt that way at the time but well I always cry at night forever and that is when all my sadness and pain really shows, but it never shows for anyone else and sometimes I kinda wish it did.  Things are just getting really bad and more than anything right now I would like to just to talk to someone about things but I don\'t know how to bring up the subject or how to talk about things in that kind of way, and I don\'t know who would take my thoughts and feelings serious.   '),(3742,1807720,'14,October,2003','2003-10-14 00:00:00','    I don\'t really know how to describe today.  But lately all I can do is eat and eat.  I am drowning my emotions in food and I gained 3 pounds and like oh God I have to stop like NOW...or I will blink at gain 10 pounds and that CAN\'T happen cause in high school every looks at people and judges them non - stop and yeah that includes the way others look and well I like the compliments of how good I look now.  You know it\'s weird though cause I thought loosing weight would solve everything and make me happier but ya know what is didn\'t change a thing, not one.  In fact I don\'t think that was ever the problems it\'s just I have many problems right now that some way or another I have gotta deal with em\' but it\'s hard.  And my TA didn\'t help that matter today, lots of the things she says makes me feel bad and makes me feel like I will never amount to anything.  And well she was like \"Well do you guys know how hard it is to get into university and she was like and did you know what the average acceptance average for a grade 12 overall is 82% and yeah that sure is high so you odds of getting into University are slim and you sure better get going now to get those marks you guys need\"  Like AHHHHHHHHHHHHHHHH I did NOT need to hear that cause frig it\'s not fair to people like me or TRY AND TRY AND TRY and just never can get over a certain mark no matter what and like I don\'t think acceptance into University should be based on only marks I mean cause then they aren\'t getting the really serious, dedicated, and focused people that know what they want in life that work so hard for it kind of people.  And really it is not fair to only choose the really smart people.  And well yeah she is kinda right though cause I am gonna have to work my ass off to get the marks and I HAVE to get into University.  I could just picture the outcome now if I didn\'t get in.  Lets just say suicide rates are really gonna go up and stuff and well I couldn\'t imagine not getting into University, that would like ruin my life forever and it would be like someone crushing my heart with a giant rock.  I could not let that happen.  Well I guess I still have to get through grade 11 and well 12 too but well grrrr why can\'t I just be finished with this all - I want a husband and kids and stuff and a REAL job...I don\'t want to go to school anymore.  I want kids more than anything right now - just kids that I love and will love me in return and a husband and stuff to love me too.  I guess one day that will probably come and later I will wish to be back in high school but I would soooooo any day rather have the stresses of adult life than teen life.  Like seriously ANY DAY I know I would choose that over what I feel now.  It will be sad when grad eventually comes and I say good-bye to some people but then I would be just one more step towards my future.  I can\'t wait right now actually but I have GOTTA STOP EATING!!!  I can like feel my stomach gettin bigger and stuff and well I just wanna be thin like all the other girls - it\'s not fair.  Of course it seems like nothing in my life is fair right now.  O well, I guess that is just the way things are.   '),(3743,1807720,'13,October,2003','2003-10-13 00:00:00','    Well on Sat I went to Calaway with Paige and Thy and we went on the bumber cars 12 times.  That was pretty fun and then when I got home I have a big turkey dinner with my family and with Ben and Melissa...well actually yeah they are family.  We had a lot of fun that night and it was good to see my sister again but when they left they left me feeling really upset.  They were talking about high school and university and stuff.  They were both like \"University is really difficult compared to High school and oh if I had to go back to high school I would have like 90\'s\" and well that comment scares me and makes me feel really bad about myself.  If I find high school hard now then how will I find university??...If I find high school stressful now then how stressful will University be??  I am trying my best right now but like they said it is suppose to be \"so easy.\"  So then I just cried when I woke up Sunday morning just lying in my bed wishing I were smart, better, wishing everything could be easier and that everything could just work out.  Wondering why I put so much effort into things and get hardly nothing out.  It feels like are my hard work and effort it for nothing right now.  I was even rethinking my future career choice but I am pretty certain I know what I want to do.  As I went over all the possibilities just NONE of them, not one interests me.  There is only ONE thing that I have mentioned over and over yet I just don\'t know why things are so hard.  I mean I don\'t seem very good at any of the Sciences or Math which is kind of related to Science and I can\'t snap my fingers and made me understand or be good at things.  I just cry over and over thinking about the future and what will happen, thinking if I will even get into University because my marks aren\'t that great now but it isn\'t like I don\'t try.  I DO TRY and that is what hurts me all the more is that I FUCKING TRY AND FUCKING try over and over.  And when I never understand or can never get the marks I feel I deserve I cry even more cause it hurts me.  I want SOOOO BAD for my dreams to come true, more than ANYTHING in the world.  Why are some people just nuturally smart and why am I not one of them?  I am sooooo sick of people always saying \"do the work and put in the effort than it will pay off\" but it NEVER does NEVER and it is so heart breaking just trying and trying but NEVER being able to do anything as good as others.  It hurts me that I am not good at the things I want to be, and I feel like I have never good at anything.  I wish that someone could just tell me right now where I will be in 10 years from now.  I am confused.  Always questioning if this life is worth living and right for me but certain people, hopes, and dreams keep me here for now.  But no one UNDERSTANDS what I feel or what I go through everyweek.  What goes through my head and what it is like being me.  I wish someone could feel what I am feeling and understand whatever I am going through right now.  And today I listened to the rain pour.  Like some kind of calming after affect of what I am feeling and I watched the drops roll down the window and realized that looking outside I feel as if though I am stading outside in the rain and everything is being washed away.  Everything.  I was so upset yesterday that I cancelled my plans with Alex in the morning and I don\'t really want to be around him or many people cause I just usually have like this attitude where I have no idea where it comes from but people end up hating me because of it.  And well I don\'t really like Alex at all, I thought I did at once but God what a mistake and besides I don\'t need for crap to think about or deal with.  I did go to Erics party through and I met Linkin which kinda cheered me up and I was trying to be happy and well thought I was doing a pretty damn good job but then later me and chels were on the couch and she was like \"Nessa what\'s wrong?  You are not yourself and I can tell you are thinking hard about something\" and well I felt that it was kinda odd she would notice something wrong cause we hardly even see each other anymore.  But well I just reassured her everything is fine and well she probably thought I was thinkin about Friday and what happed with Alex and Travis and stuff.  but well she is wrong and yeah I guess only I knwo truly about what I am thinking but I don\'t feel like talking things through with Chelsey at the moment cause she will just tell me it is silly to feel the way I feel and I will get over it.  I have tried to talk to her about things like this before and well this time I think I am just gonna keep things to myself unless someone other than her asks me about everything.  There is just soooo much crap in my life, and yeah a couple good things but the bad thoughts and feelings out weigh the good anyday.  I just hope tommorow I will wake up without crying and feel a little bit better about things.   '),(3744,1807720,'11,October,2003','2003-10-11 00:00:00','    Life at Sixteen  Crying on the inside I donâ€™t know what to do Have to clue of who to turn to I am lost, dazed, and confused  This crazy life is throwing me up And I am falling further on If High school is suppose to be so easy then why does it seem so hard?  To my sister and her fiancÃ© They claim high school is just a joke And reassure me University is where it really starts But I am just so confused right now  Right now everything seems so hard I have no clue of how I will make it through I donâ€™t understand just why things just are And I wish I new why I just donâ€™t seem to get things  Feeling as if though the teachers judge and stare It is like they feel that I am not smart enough Or just something enough, I am lacking something I just donâ€™t know what it is right now  But I was always told to follow my heart And that is what I feel I must do right now But sometimes itâ€™s hard to follow my true feelings Because I know what I want to do and where I want to go  However, I want to teach science, maybe Bio But I need chem. 30 and well 20 is hard And well Physics I canâ€™t really predict how I will do I have to be good at Bio next term I just have too!  If I am not good at Bio, or Physics, or Chem. then how the hell can dreams come true? How can I have a job that I love or pass University? In life what purpose would I serve?  There are so many questions, but not enough answers I wish someone could just tell me my future Cause sometimes I wonder why I am here?  What can I do? I feel as though my efforts just pour down a drain  It feels like I try and I try and I TRY But that is never enough it seems The one thing I want more than anything in this world Is to just understand for once, and have my efforts recognized  And sometimes I think no one really does know ALL the hours I spend studying or doing homework ALL the days and nights I try to figure out Chem. ALL the times I just practice my math over and over  Just hoping that maybe I will get a good mark Maybe all my studying will pay off, but does it? A lot of times I just feel and wonder why do I waste this time studying? I work so hard, but never get to where I want to be  It is always like that, but why does it have to be? I donâ€™t choose to be dumb and all I want to be smart But I feel dumb because I never understand what everyone else does And it is so FRUSTRATING almost all of the time  But no one can make me smart, or make me understand If I could make myself understand I certainly would And I KNOW that it is not from lack of effort Because there is so much effort I put into everything I do  So really I have to explanation to why things are the way they are Certainly an explanation would be nice And maybe I will reach my goals, and live my dreams But I just donâ€™t know right now cause I am so confused about everything  Chelsey and I always seem to argue almost all the time We still have fun together but things are changing now, people are changing And it is hard to talk to her about the things that bother me There is no one really there anymore because she was the only one  I wish sometimes I could just talk to people openly about anything But that is not the way things work even though I wish it were true And I feel as if the ONLY person to care enough about anything is Mrs. J I always felt she cared more than anyone I have ever met  But my mother always calls me annoying, and she might possibly think the same And I donâ€™t think she would want to hear my problems I mean when I think about it there is nothing I can do or anyone else No one can make me smart, or make me understand  Itâ€™s like I am stuck in the middle of some feeling so awful To feel like I am never successful at anything And to feel like I never achieve anything that I feel I deserve It is a very heart crushing feeling because I try really hard  And the feeling of love WAS something I thought was great Until I woke up and snapped out of this daze I WAS in I thought I liked Alex but I realize now I canâ€™t stand anything about him and he is not my type  He is the kind that keeps things to himself, but I myself am more extroverted And well I CANâ€™T stand those army shorts, and army pants They drive me INSANE cause he is obsessed with anything army And I donâ€™t want to upset him but I really donâ€™t like him how I thought I did  That feeling inside of me is a feeling of love But that strong feeling is developing for Travis now The one person I totally never liked, but I know him better now And he is definitely someone that I like  But he has been coming onto Chelsey over the last few days And I know she has a boyfriend but I think he DOESNâ€™T like me And well I guess that is not really one of the big concerns But I guess everything worries me cause I am a teenager  I am only sixteen but with everything there comes stress And sometimes it seems like things are just too much to handle I think sometimes how things would just be different If I just made a choice, one choice that would change things forever  Sometimes I am so certain, but other times I am not so sure This choice is the one of suicide but I donâ€™t want to hurt others around me And my dreams I want to come true, but if they donâ€™t Then I will be crushed forever, and I could not let that happen  And sometimes everything just seems to go wrong All I want sometimes is to be somewhere else, and be happier But even today I was looking through the yearbook and looked at the staff pictures There I saw Mrs. J and I just looked and stared at her picture forever  I thought to myself that I couldnâ€™t hurt this one person that seems to care so much And as I looked at her smiling in that picture I remembered a specific day The day of the social final, the day before the science final and what happened It was when I was walking down the hall upstairs and I wasnâ€™t exactly all too happy  But there she was, and there I was just walking down the hall and I looked up Then our eyes I guess connected and she just smiled and I never felt a feeling like that Cause it made me smile too and the feeling inside I cannot describe From then I knew she does care about me, and would care if I were gone one day  She would probably think I am crazy if I ever told her that I felt like she was an angel But I feel like as if she was sent by God to be in my life right now Honestly I donâ€™t know where I would be if it werenâ€™t for her, she has made a difference All the times I wanted to die, but all the times she stopped to ask how things were going  All the times I just wanted to give up, but she just smiled Sometimes a smile means more than words ever will, and sometimes that is all it takes I realize that words donâ€™t always fix things, or make them better But it is so true that if one person you know cares it can make all the difference  My life is my no means perfect, and I feel as if though I have a lot of things on my plate I mean I have some confusing feelings about love right now And I have many feelings about where I will go in life, and if dreams will come true I guess I canâ€™t forget that there are also great things and people that come into my life  Maybe these are common feelings for teenagers or maybe they are not My feelings are  not really things that can be discussed, which makes it much harder But I feel a lot better now that I have thought about what has been bothering me And maybe this is all normal, maybe this is just life at sixteen  ~Vanessa   '),(3745,1807720,'10,October,2003','2003-10-10 00:00:00','    OMG today/tonight was SOOOOO fun!!!  I have a 67% in chem right now and well I thought for a while that I just want everything to end, I just want it to all stop.  But then omg me and chels and travis went to Dairy Queen and he bought us those chocolate utopia things oh man were those things to die for!  Then later that night we went to Llody\'s and we went roller skating/blading and it was soooo fun.  I got to take my mind away from everything and it felt so good.  Alex bought me a rose and Travis got Chels a rose and Travis bought us some fries and like a million glow sticks.  He also gave us a ride there and back.  The funny thing is though I thought I liked Alex but I felt really uncomfortable with him tonight and well I felt this really kind of strong relationship/connection with Travis and not Alex.  I am still gonna go out with Alex on Sunday but I just don\'t thing going out with him is right at the moment but I don\'t really want to offend him.  I am really trying to give him a chance but everything he left when we were roller skating I felt better and more like me and stuff.  I thought tonight Alex was acting kinda like a total idot and well he was sort of embarasing in public.  And well I HATE it how he always were those army pants or shorts and stuff and GRRRRRRRR tonight he wore army shorts.  I HATE those.  I mean clothing and looks aren\'t a big thing but I mean when someone wears that same friggin thing for even and you ONLY see em wearing like one thing it is like a big turn off.  Like ewww man get some new clothes or something!!!  I just don\'t feel Alex is right for me like at all....I thing Travis would be way better and I still have had that crush on Chester for like what seems to be five billion years!!!  I said I would go out with Alex on Sunday so I guess I am still doing that but I don\'t really want to but just this one more date and if I still feel like I don\'t want to be with Alex then I think I am really gonna have to tell him cause I can\'t hide the truth and my true feelings!   '),(3746,1807720,'09,October,2003','2003-10-09 00:00:00','    I suppose today was pretty good but sometimes I just talk and I talk and whatever I feel comes out.  Then after I realize how negative I was being or some of the mean things I was saying about others and it makes me feel like I am a bad person.  And now I wonder am I a bad person?  I suppose I can never really know the answer.  And well yeah I had the greatest laugh after school today.  My back was turned and this guys comes behind me and is like \"hey\" so I turn around then I looked him up and down and thought to myself (do I know who you are) then he was like \"oh my gosh I am sooooo sorry, it\'s just you looked like one of my friends from the back\"  Man I had the biggest laugh after that but the guy was a cutie and really nice to me, and well I realize that at B.O.B. there probably are alot of awesome and really nice people I just don\'t know them all. And well I didn\'t understand the chem lab and well it WAS SO HARD to try figure it out so I went to tutorial to find out how to do some stuff and like the chem teacher I thought was like laughing at me and like for a while she made me want to cry cause she made me feel like I am stupid but I AM REALLY TRYING to understand.  But well after I ended up laughing at myself cause she was like \"yeah Vanessa thanks for playing around with this stuff, now I have a really good lesson to teach my students\" so then I thought that it was kind of funny and stuff and I realized how much I made her work to help me to understand.  boy I sure make teachers work for their money don\'t I?!?!  Lol!   '),(3747,1807720,'08,October,2003','2003-10-08 00:00:00','    Well this morning chels comes and like I was just talkin to her and she was saying how she really hated the yearbook cover and well I was saying that I really liked it.  And then she totally like spazed out and got like all mad and was like \"oh vanessa shut up and stop accusing me\" and well I just had my own opinion and she had hers but it seems that the more time I spend with that girl the more of a negative effect it has on me.  Infact I called her miss negative this morning cause she was being miss negative about every damn thing and after a while it starts to really rub off on other people and like I am trying to be happy and positive but she makes things like that so friggin difficult.  So yeah and well I am worried about that 70 in math and well I REALLY studied chem and REALLY knew the stuff but plain and simple - the chem test today I found really difficult but I KNOW that I know my stuff.  I just don\'t know why I found it hard like applying the knowledge is difficult and half of the questions I didn\'t know what it was asking.  It was really confusing.  And I am back again to wondering why I am not very good at anything.  Why are some people so good at school and subjects and everything and why NO MATTER HOW HARD I try or HOW MUCH I work on things they are still sooooooooo hard.  I want to know why.  I mean just WHY?!?  I mean I want to be a teacher when I am older but I have never heard of a teacher in my life that struggled through school and in their subjects....because there isn\'t such a thing.  I wish I had someone to relate to.  And all the teachers I ask if they were good at this or that always say that they ARE good at it or WERE good at it.  Either way they never had any of the difficulty that I am having.  I just wish there were an explanation, it would make me feel better if I just knew WHY..!  Like I didn\'t really understand how to do the lab write up for the last lab we did and the math word problem things I STILL don\'t know how to set those things up and it is REALLY FRUSTRATING because I REALLY WANT so bad to understand and just be able to get a right answer.  Honestly if I knew how to do it I would but the fact of the matter is that I don\'t and I feel like my continous efforts of trying and trying are getting me no where.  I am just falling into a big black hole where all my marks are falling in with me and everything else.  I want an answer to things.  Yes I want and I want and I want right now, but I only want know something.  I just want to know WHY...and forever I will probably always wonder why and wonder why there is no answer.  Why do I have to work at things so much more than others?  Why do I spend countless hours trying to figure out the material and studying something that I usually don\'t do good on anyways?  Why does everything usually take me longer than others?  Why does it seem like I feel as in I put in 100 times more effort than what I get out?  Just why??..!   '),(3748,1807720,'07,October,2003','2003-10-07 00:00:00','    Oh yes what a good day.  It was alright I suppose I mean my overall average for math right now is a 70% and if I get anything below that even I WILL die so yeah I\'m makin a promise to myself NO MATTER WHAT HAPPENS I *will NOT* get anything below 70 in math from here on in but I have been giving it my all it\'s just I can\'t always get the help I need which is really frustrating.  But well I feel the Chem teacher is really trying to help me and I feel like she finally understands that it is not easy for me.  The chem chapter 5 test is tommorrow and I studied for like ever tonight and did a little review on stuff the days before.  I really feel ready for this and all I can say is \"bring that chem test on\" cause yes I believe in my heart that if I truly believe that I can do good than I can do anything.  That\'s right...ANYTHING.  and well understanding the concentration of ions was really challenging but I think tonight when I just took a little bit more time to go through things I think I am getting it alot more and everything in this unit has been challenging but I have really been working hard that like right now in this current unit I have improved my marks drastically compared to the first unit.  Things are kind of coming together but well with that 70 in math it worries me a little cause well the whole \"I want to make honours thing going through my head.\"  Honestly I feel I really in my heart deserve it more than anything if only people knew how many hours I spend trying to work through my homework, studying, and even trying to figure out math.  No matter what I know in my heart that I have done my very best and worked my ass off to get where I am.  I\'m listening to a great song right now.  It\'s true and so touching by hilary duff.  It goes like this...     Inner Strength  Gotta find your inner strength If you can\'t then just throw life away gotta learn to rely on you Beauty, strength, and wisdon too  Your beautiful inside and out Lead a great life without a doubt Don\'t need a man to make things fair It\'s more than likely he won\'t be there  Listen girl gotta know it\'s true In the end all you\'ve got is you  ~A song by Hilary Duff   '),(3749,1807720,'06,October,2003','2003-10-06 00:00:00','    Well ok so I have to admit the math test I did today I don\'t think I did to good.  But I don\'t really blame it all on me.  I was TRYING so hard to get help and she kept ignoring me or saying she would help me but always ran out of time and well I went to the math resource room and asked that lady how to do some of the stuff but she didn\'t make very much sense.  If I don\'t do very well on the test though maybe it might be a good thing cause then the teacher will realize that I really really really....do NOT understand how to do many of the things and she always says \"when I talk to you though it sounds like you know it all.\"  But well honestly only I am the one who knows myself the best and I think if anyone were to know if I didn\'t understand something it would be ME!  Today had its pros and today had its cons but really I think there was much more great things that happened.  I got a 14 out of 15 on my chem quiz!!!!!!!!!!!  That is like a 93% and well I did good on something in chem.  I actully did good on almost everything in this new unit after the review one from last year that didn\'t go over so well.  I have gotten 100\'s on all my labs for the unit and well and 74% on the first quiz and a 93% on this one - all I need to do now if get a really good mark on the unit test.  And well Mrs. J came up to me and asked how Chem is going which even though I actually don\'t feel really bad about chem right now its nice to know people do care about others and that she does care about how I am doing in chem or even about how I feel.  She\'s different from alot of the other teachers, there is just something about her.  I mean my TA would never ask me how things are going like she does and well I think she really has a good understanding of what people feel and go through.  I\'m writing a poem right now which is not finished but I want to give it to her at the end of gr. 12, on the last day that I will probably see her for the last time and I just for once at the end of grade 12 to let her know how amazing she really is and how she had touched the lives of people around her and made them feel special and helped them to understand.  So here is what I have so far.     Thank you  People come and people go But you were so special I want you to know Iâ€™ve never met anyone so kind and so caring You were different, all that knowledge you were sharing  I still remember one day walking by You gave me a smile instead of a â€œhiâ€ That feeling inside was something so great A gift that anyone could appreciate  There is just that something special about you You know what to say and you know what to do You helped me with science and with chem too Helped me to deal and helped me to get through  Forever a memory of you will rest in my heart I didnâ€™t think you would be so cool right from the start But wonderful things they come and they go The thing in life is that we never really know  So here is a thank-you to all that you have done You have made such an impact on everyone You are an inspiration, and a role model too There is nothing about you that I would undo   Thank you  ~Vanessa   '),(3750,1807720,'05,October,2003','2003-10-05 00:00:00','    One Moment  Wishing just once, wishing for a moment One moment A moment in time we say good-bye to our problems Let them all fall away now Wishing just once, wishing for a moment One moment A moment in time these dreams are alive and flying There wonâ€™t be no more crying Here in this life wondering where Iâ€™ll go next Stepping onto a pathway, the freeway An I-donâ€™t-know-where-Iâ€™m-going day Wishing just once, wishing for a moment One moment A moment in time where friends can be in this line On this calling list to talk to, to help me to just get through To share our secrets and stories This moment, one moment Iâ€™m wishing for Here in this life, this crazy life Iâ€™m spinning, Not winning, just looking for a place to find Where in time I can find a line, a rope And hang on to that one true dream, this feeling inside of me Wishing just once or twice  Wondering why something canâ€™t just happen over night Wishing for dreams to come alive, friends by my side Wishing just once, wishing for a moment One moment  ~Vanessa      So I went to church today which I love going to cause the feeling I get from it afterwards is really great and it is kinda like a re-energizing thing before monday starts.  And well I spent HOURS and yes I mean HOURS studying for my social and preparing for the long answer and I spent a bit of time on math today but most of saturday I spend reviewing my math for the test tommorow.  I think I will studying a little bit more for the social tonight too cause I just still don\'t feel ready but well knowing all the time I spend I mean like yeah I probably am ready but last time I got a 75% on the long answer and like I know I am capable of like 90\'s and stuff and since the M.C. usually ain\'t my thing I\'ve gotta do good on this.  And well I have to do good in social cause if I want to make honours and gotta have good marks...lol and well I am a little uncertain about math and chem right now but I am sure that everything will come along just fine - it will all be right with time!   '),(3751,1807720,'04,October,2003','2003-10-04 00:00:00','    Open  Open your heart to the world, to your friends and let your light shine down  Open your eyes thereâ€™s the talent inside you In front of you  Open your hands to receive every gift from God   Open your mouth to speak words full of wisdom  Open you mind to see what lies beyond us  Be open to the world for you have so much to offer in the life  ~Vanessa     Well I did my homework and stuff today and really TRIED to figure out my math but like I realize that man I have NO CLUE about what I am doing.  I just hope there is not too much stuff on the test monday.  And well my mom is going to apply for a job opening at my school which is the textbook rental assistant which I actually think that would be like sooooo totally cool is she was.  And well I went shopping tonight man that was fun.  I bought a shirt and skirt, new jeans, and purple pants with my left over b-day money from August!  And I fit into a size 10 - I think I am gonna faint!  I used to be like size 14-15 man this is sooooo sweet I was like drowning in the 12\'s too.  So well I had a little treat tonight I got one of those twisty cones from mcdonalds besides on of those ain\'t gonna make weight gain happen over night and well I am just makin sure I am balancing the fruit intake w/ the junk intake.  Lol and well man I have sooo much social to do and I still have to finish my long answer question answers before I can memorize em\' for the test so that is what I really need to go do now.   '),(3752,1807720,'03,October,2003','2003-10-03 00:00:00','    One Thousand  One thousand words, one thousand miles One thousand dreams one thousand times One thousand takes, one thousand lines One thousand hopes one thousand times  ~Vanessa      Well yes today was good except for I felt I dunno really ignored in math and I have no clue when I am going to get help with anything and I wanted to get help but the teacher was busy and grrrr I just don\'t know who will help me cause I don\'t understand solving for min/max problems.  I keep thinking about the social test and it is driving me up that wall!!!  but well I am just one of those natural worry people so yeah.  And well Alex asked me out and stuff so we shared a salad and some fries for lunch and well then we disscussed how like we should go out a couple times and see what happens from there.  And well I suppose the Chem test today went pretty good.  I mean I think I did good at least and well I really prepared for that thing so yeah I don\'t really think there is anymore I could have done.  I feel pretty good right now again.  Chels came to my house for a while then my mom, chels, and I when out to dinner then me and chels went to South Centre to shop and I got some chopsticks for my hair and a new belt - oh it is soooo cool.  And well we invited Ash but she wanted to watch football so like whatever!  I just hope I understand my math a little bit more tommorow when I try to practice the stuff and well I should review my chem, but I have lotsa social studying and stuff cause I think I am gonna write the test on tuesday.   '),(3753,1807720,'02,October,2003','2003-10-02 00:00:00','    Wow!  Words can\'t describe how I feel right now but I feel *happy* yes really great.  I mean Chem is still kind of confusing but tonight I just sat there tonight and worked through it and it makes alot more sense then it did before.  I feel like I really have a good chance of doing good on the quiz tommorow.  And well I got a 70% on my math quiz but I think I could have done better but I guess I will probably get some help with the stuff I don\'t really understand to much and well that stuff we did today in class I found confusing cause I don\'t think she did enough examples so I am bit confused but tommorow we have a work class so it is all good.  Everything is great right now I feel so awesome it\'s like this odd feeling that have come over me.  I feel so calm it\'s so odd but a nice feeling and I feel so great.  Lol sounds silly doesn\'t it but I am actually not over worried bout the chem quiz tommorow cause I know I ready for this and I can face what comes in my way and what comes next on this path of mine.   '),(3754,1807720,'01,October,2003','2003-10-01 00:00:00','    All that social studying.  Man my head was starting to pound there for a while and well GREAT news....originally it said on my scantron that I got 35% on the M.C. social test well me and the social teacher went over it and like half of the things I put were right now yeah I actually DID pass - thank God he just said it was because he thinks he ran my test through the wrong key so lots of things were wrong when they should have been right.  And well I had the math quiz today and well I am just not too sure how I did.  My mind is kind of like ya know sort of went blank but I mean I know how to do the stuff so I guess I will just have to wait until tommorow to see how I did but I probably did fine because I really prepared for it.  And well I feel pretty good today and yes still confused with chem and all but I am sort of getting it a little bit more and my chem teacher DOES seem like she is trying to help me because I think she can see that I am really making an effort to do good.  And like tonight I was majorly studying for social because after last test I got a 60% on M.C. and a 75% on the written answer but I know in my heart that I could do much better than that.  And well yeah I feel good, I feel happy - there were no fights today and well I talked to other people that I don\'t usually talk to so much.  And paige invited me to Calaway on the 11th so maybe I\'ll see if Alex is a ride person cause knowing Paige she wouldn\'t care who comes she would just think \'the more the better\' so yeah that is an idea and I was thinking that maybe one weekend me and him could go during free skate and just hang out and well I can\'t wait til it snows so we can act like kids and go slidding down the hills and go skating at the rink like 5 minutes from my house and build a gigantic snow man and then when December rolls along we can go to the dance.  Man all the great and fun things I can think of.  I think a little part of my self esteem just took like a major boost.  It feels good, and I want to make it last.  Things have been going a bit better lately and I KNOW on Friday the Chem test will be challenging but in the end I think I have really got to realize that I CAN do things if I believe that sometimes we don\'t always do as well as we would like but it is not the end of the world and WE DO always get other chances, we have other quizzes, and other tests.  And that is right that one mark or one bad quiz doesn\'t mean I have to hate myself, or beat myself up over it.  It doesn\'t have to mean that at all, it can mean that I have made some mistakes along the way.  But mistakes are good, they are learning experiences and even though they may upset us at the time, when I look back on things I realize that the most important life lessons for me have come from the many mistakes I have made in my life time.  And thinking about this now reminds me of all the hard things I have gone through to get where I am now.  Like when I was a kid I don\'t think anyone even liked me because I wasn\'t the ideal \"thin\" like all the others kids and well in gr. 6 it got to the point as bad as when I was seriously considering suicide and I guess no one really does know what things in life have been like and I have been through alot that no one, not even my best friend will even know about.  And well I got through that but then two years ago I was older, wiser, but knew more about things and had a better idea of suicide but there was someone there for me and she always could tell something was wrong even though I thought I kept it hidden so well.  I have to admit man she had some sorta of gift.  Sometimes I don\'t think teachers realize the difference they can make for people.  Sometimes they can makes the difference of life vs. death and I would say that is a very powerful thing.  That is also one of the reasons why I want to be a teacher because of the difference that I could make too.  So I wrote this poem that has a lot of meaning to me:     All My Life  This is the story of all my life Can I tell you what is was like? It wasnâ€™t easy and it was so hard Take a look at me and see how far I got  Being pushed around day by day Called those names that I dare never say Told I did not matter, it hurt me so bad I had that feeling inside, told me I was sad  Down so deep and feeling alone, all on my own Why wasnâ€™t anyone out there to hold? I wanted happiness but all I had was one plan To take my life and that would be the end  Along came a hero, she walked in the door Held me tight until I hurt no more Told me everything would be just all right She gave me hope for a better life  Now in high school; Iâ€™m having the best time Who would of though I would of made it this far Thereâ€™s a hero out there waiting to help you through And she knows exactly what we all go through  The good and the bad The happy, the sad The joy and the misery That is what all my life has been filled with  When I was all alone, she helped me through I felt unloved, but now I feel the love inside of you Seemed uncared for, and now Iâ€™m cared for even more And this is how all my life has been like  Through the halls I pass all the people I know Hear the laugher as I walk the school floors Now I know so much more, things do get better And sometimes all you have to do is wait for that hero  This is the story of all my life Can you feel the pain that Iâ€™ve felt all this time It was never easy and it was all so hard But take a look at me; I got so far  And this is the story of a young girl She never really wanted to take her own life All she wanted was to be loved and cared for And this the story of all my life  ~Vanessa   '),(3755,1807720,'30,November,2003','2003-11-30 00:00:00','    I was listening to that new Britney Spears Cd.  It kicks some major ass!!!  I haven\'t bought it yet but I got it off the net in windows media player.  She is awesome.   And well I sorta studied my chem this morning but I did most studying last night but I am going to wait for later tonight to do some more stuff.  I think I am starting to understand the stuff and there are just a couple things that I really can\'t get the hang of that I should probably ask about on monday before the quiz.  Me and my mom went to our hair dressers house for this open house she was having for all her clients and we had a huge lunch.  And then later tonight my dad is flying back to Vancouver so my parents will be gone and it will actually be QUIET.  So I can actually study alot better.  I need to do good on this chem quiz tommorow.  I NEED to, and of course the unit test coming up is certainly important.  I haven\'t done the review questions from the end of the chapter yet but I might start on some tonight and well I will probably be done by tommorow night.  If I don\'t do good on these tests I don\'t know what I will do.  I don\'t know what else I can do really.  I\'m so scared for the future.  I wish I could understand chem.  If I want to be a Biology Teacher I will probably have to do chem stuff in university and I can\'t even understand the stuff that is suppose to be easy now.  But that is my dream, I have dreamed days and nights thinking about how much I want to teach something related to science and I am pretty certain it will be biology.  It is my passion, afterall I love science but not when it is frustrating.  Why can\'t I be good at the things I like?!?!?    Wouldn\'t it make more sense if God made me talented in the areas I am actually interested in?  I mean everything that has to do with science is LITERALLY my dream, I can picture myself infront of a class teaching too.  I can just \"see\" it, it is sooo clear.  But I am soooo afraid for the future, I NEED to understand Chem now, if I don\'t then when will I understand???...I mean I think I am getting a little bit more of the unit now that I have gone through the stuff over the weekend, but lots of times I feel I know the stuff and then I just get soooo confused on the tests and quizzes cause I need know what they want you to do or what they are asking.  I just hoep everything turns out alright.  I hope University still accepts me cause I don\'t know what my chem mark is going to be.  I am trying to get it higher but I don\'t think failing all the quizzes is going to raise it by any means.  I just worry soooo much, I wish I could stop.  CHEM, CHEM, CHEM, that is alllllll I can think about.  It\'s annoying and it ALWAYS on my mind.  I just want to score like at least an 80% on something.  I want a mark that actually reflects my effort for once, I hope that I do good on that quiz on monday and especially good on that test on Wednesday.   '),(3756,1807720,'29,November,2003','2003-11-29 00:00:00','    My dad flew in Thursday and yesterday we went out for Chinese.  Then I went to a party and OMG we had sooooooo much fun.  We went sliding down this giant hill that connects to a road so like when you get to the bottom of the hill you just keep sliding down the road, but well you have to be careful not to get hit by a car.  And like this one time when we were going down the hill we were going like REALLY fast and like almost hit this guys garage door.  We keep going from the hill and starting going up the driveway and then we started to scream.  It was soooo funny and fun.  And then we had this big snow fight and me and Ash loaded up this sled and poured it onto Lenny.  And then Lenny jumped on top of me and we like feel over and then we started like rolling down the hill and then when we stopped I landed on top of him.  LOL.  He was like \"Wow, that was some good fun but I got crushed by Vanessa\'s boobs\" so yeah I guess they were kind of digging into him but after he admited that he liked it.  What a fun night.  On Friday the CALM carosel was pretty good.  Some speakers were really boring and others were interesting.  Like the lady from Best of Bridge was really good and had an interesting story but when that football player Mark Mcglouchlin (sp...?!?) came is was like sooooo boring.  And the phys. ed teacher there would NOT stop obsessing over him and kept telling us over and over again how many grey cups he was in and how many he won.  The first lady was really boring and I don\'t even remember where she was from, I think she might have been some sort of career counsellor.  And well the last guy was a student from our school which was pretty good and interesting, and he was funny but he needed a little bit of work on the public speaking.  The carosel thing ended a little bit earlier so I had some time to talk and so I talked to Matt about like me and Travis.  See like Travis always buys me all this nice stuff and pays for like everything and well I like that.  I like when I don\'t have to spend my own money and everything but like there are soooo many qualities about him that I don\'t like.  He is sooo not honest, I mean I am NOT stupid and I CAN tell when someone is lying to me.  Especially when I asked him why do you always lie to your parents, and he was like \"I never lie, really I never lie, ever\" and so I was thinking, ok so if you really never lied to anyone then why last monday would he have told his parents he had school.  That is what he did last Monday.  He told his parents that we did not have school just so his parents would not make him work and he could spend the day with me.  Like that IS a lie.  And he is failing CALM, how in the world does someone fail CALM?!?!?  Like that just really tells me that he takes nothing serious.  He is a LOSER, the big L - O - S - E - R!!!!  I mean he won\'t graduate if he fails everything.  It makes me think you know, doesn\'t he have goals...or dreams.  He told me he wants to be a mechanic and well I guess that is alright I mean if it is your passion and stuff but like at least pass high school and get a high school diploma.  I am surprised he hasn\'t even dropped out.  Me and Travis did make plans to go to the mall on Friday though and well I felt that maybe I will try one more time to spend some time with him but like we have no kind of \"connection\" and I have felt NOTHING for him so it kind of bothers me that he still sort of comes on to me.  Anyways yeah we were suppose to go to his bank and then go to his house to get the truck and go to the mall but well I decided to go home first cause I wasn\'t feeling good near the end of the CALM thing and realized it was the stupid T.O.M.  So yeah I went home and met him at the mall and we spend about like half-an-hour in the dumb electronics store.  Like seriously he waists all his money on stupid games for his x-box.  Knowing how much he works and how much he spends, he could have saved up to pay for like University or something but I am guessing that he might not go on to post-secondary education.  It is not that I don\'t believe in him, it is just how he acts not and the way he is doing in school there is a very slim chance that he would get in.  But yeah we went to the mall anyway but like there is STILL NOTHING THERE.  Like he thought he kind of liked me and I was never really sure I liked him but now that I know him more I don\'t like who he is.  He doesn\'t even have a relationship with his parents and can\'t be open and honest with them.  That is like the most important thing and he does not have it.  He can\'t tell his parents anything and I KNOW for a fact he lies to them all the time.  I don\'t care is he claims that he NEVER lies, cause that is pure bullshit.  And well besides I am sure just about everyone has lied about something sometime in their life.  I just don\'t want to offend him, but I don\'t even like him as a friend anymore.  He turns me off.  He is not the definition of the kind of friend I want.  And even last week he started to follow me around everywhere.  I mean no offence but like it kind of starts to creep me out when someone starts following me around.  Like on that one friday were I was at my locked and he came up to me and was like \"hi\" and I was like \"how did you find my locker?  Who told you where it was\" and he was like \"I followed you.\"  No offence but like I don\'t want to be followed.  Besides it was afterschool and I just wanted to relax and stuff before I saw him and well nooooooo he shows up at my locker and I didn\'t even have time to think about things.  I hope he starts to clue into things, but gusy are slow when it comes to things like that.  I don\'t want to upset him, but I don\'t like him.  I want an intimate relationship but I KNOW that he IS NOT the one I want it with.  Not at all.  I want to find someone with goals and dreams.  Someone that wants to go to University and does take school seriously.  Someone that I can talk with and have a good time with and doesn\'t ONLY use money as a way of showing affection.  Money means nothing when it comes to feelings that that.  I want someone who is trustworty, open, and honest and has a good relationship with their parents.  I want someone who is responsible and that is about it.  If I had someone with all those qualities than everything would be perfect but like I have gone through a couple guys already this year which kind of stressed me out and this at the moment is still stressing me out but it makes me realize something.  It makes me realize what I DON\'T want in a guy and it makes me more aware of the qualities that I am really looking for.  So far I have not found that yet, but yeah it is not like I am 32 or something, I would start to worry if I was but I only 16 right now.  I have many years ahead of me and there is plenty of time.  Anyways that is like only one of my problems.  My other problem is understanding my Chem.  I would give anything right now, literally ANYTHING just to understand.  I feel bad that I don\'t really understand some stuff.  I feel it\'s my fault I don\'t understand and I wonder if it is my fault.  But I know I am trying to understand and I read over the stuff today and did the practice questions that I could do, but there were others that I am confused about.  I feel bad that I always have to ask \"How do I do this and how do I do that.\"  I feel stupid.  Why can\'t I ever understand anything on my own?!?  WHY?  And like alot of times I can\'t do something cause it realates back to Science 10 which I had a really hard time with.  Like I even remember the chem teacher saying \"well you should know this from science 10, right\" and yeah I agree cause yeah I \"should\" know it but I struggled with science last year I no matter how hard I tried to understand some of the concepts I just could not grasp some of them, even some that alot of people find very simple.  Since I couldn\'t understand some of the concepts from last year, obviously I cannot understand some of the Chem this year.  I don\'t know how to make myself understand all the concepts, if I could not \"get it\" last year and I can\'t seem to do it now then what do I do??  I don\'t know what I can do to understand some of the old concepts I should know and figure out the new ones.  No one understands how hard things are for me.  They just don\'t.  It\'s frustrating TRYING AND TRYING.  I failed my lasy chem quiz too.  If I wanted to fail I would have just done nothing but WHY did I have to fail.  I STUDIED, I STUDIED for all of the tests.  I feel like the chem teacher understands a little bit more of what I feel and realizes that I DON\'T understand some stuff and she is TRYING to help but I should not keep failing.  Someone who is actually trying to understand the stuff and studies and does all of her homework should not keep failing!!!!  I have another quiz on monday and well I will probably have to go to tutorial though and get help with the stuff that I don\'t understand because I HAVE to do good.  If my chem average ever drops below a 65% I am not going to be a very happy person.  I want to make honours again come the end of first semester but right now chem is really one of my big concerns but I KNOW in my heart that I am trying my best.  All the times at the beginning of the year where I thought chem was sooo hard that I could never do it or pass the course, I even thought of dropping it and giving up my dream to teach science.  But I didn\'t drop it, I kept going.  If only anyone knew all the nights I cried cause I didn\'t understand, and the nights I cried because I got a bad mark.  I have to admit that I have found NOTHING in chem easy and I guess yeah it is suppose to be challeging but I didn\'t it would be as bad that I would cried like a billion nights in a row and fail more tests than I have ever failed in a course.  Chem really is the first course where I have failed more things that even beofre, so far I have failed four quizzes and usually I don\'t fail anything at all.  On the odd occasion I might fail maybe once or trice in a course but NEVER have I failed four things in a course.  That is why it is so frustrating.  I know how hard I try and how much effort I put it and hardly even does it seem to pay off.  It\'s like I am constantly being drainned of my energy and often I wonder for what???  It\'s hard, I won\'t deny that.  I am trying, but for someone who wants to do good soooooo bad I sure ain\'t doing to well.  I am trying to be positive in my thinking for chem but when I reflect on everything it upsets me.  The way I feel about everything it upsetting and how much stuff I have failed in chem is also really upsetting.   '),(3757,1807720,'27,November,2003','2003-11-27 00:00:00','    Today I realized that the chem teacher is nice and is not as evil as I once thought she was.  I realized that she CAN actually know and realize what others are feeling, and for onceI felt like she tried to understand where I was coming from.  For once I feel like I told her my honest opinion about how I really feel about her classes and stuff and it feels good to get it all \"out there.\"  I mean I am still confused about alot of the stuff in Chem. but well I felt like she actually was NOT rude at all today and took her time going through the quiz and stuff so I would understand.  And when we were doing the lab she was like \"good work girls\" and like usually she gives us a lecture about how to \"propery\" do something but today she just told us we were doing good work.  She is sooooooo much more considerate and much more nice.  Tonight I was doing my homework for Chem. and I was doing those titration calculations/problems and like at first I totally did not think I could do them at all.  But then I thought about what the chem teacher had said, what she basically asked is that \"When you see a question that you might think you don\'t know how to do or confuses you do you panic or get scared?\"  When I thought about that, the truthful answer is \"yes\" and then she was like \"Well you can\'t let that get to you and you have just got to go with what you know and don\'t let that panic you.\"  So tonight I looked at this worksheet we got and like freaked and thought I could not do any of the questions, but then I thought about what she had said and well I started just to try doing the questions based on what I knew and stuff and and trying to put all the info together and since she gave us the answers for the questions I could check my work and I actually did do a couple right and there are only a couple questions that I need to ask her about.  I feel that she actually can be really nice, and can actually be really understanding.   I saw a totally different side of her today.   '),(3758,1807720,'26,November,2003','2003-11-26 00:00:00','    Ya, 80% on my math test...now that is more like it!!!      I got an 80%...Go me!!!!!!   I deserve it.   I sure hope I did well on that last Chem quiz Friday cause well I got kind of confused on it.  I think during tutorial tommorow I am going to go ask her what I got on that quiz...cause she still hasn\'t posted the marks and I cannot stop thinking about what I got. This guy today in chem like insulted her teaching.  Someone asked \"Why do we have to do this?\" and then she was like \"Well the students told me that they liked it\" and he was like \"Well not everyone said that\" and then she was like \"Well would you rather have me teach\" [cause yearh we were using labtops and he was like \"Uhhhhh NO!...Like no offence or anything\" and then she was like \"no offence taken.\"  She seemed to know that she needs a little improvement on the teaching.  I have to admit though she is getting better.  And well she is alot nicer to me now then she was before, but her teaching is getting there, it is improving.  One of my cats had to be put down today.  He had diabeties and well he went partially blind from the insulin and well then she would just lay on the floor and he wouldn\'t move or eat and the odd time he got up he couldn\'t even walk straight.  I had Socks since I was five but well I am glad I don\'t have to see him suffer anymore so I am not really upset.  I knew it would eventually happen one day so I am just accepting the fact.   '),(3759,1807720,'25,November,2003','2003-11-25 00:00:00','    Today was pretty good.  I was actually pretty quiet today.  I didn\'t feel like talking too much today.  When I got to school and went and got some math help from my math teacher and there were like a couple of us in her room watching some examples.  Socail was pretty good, we all spend the time practicing math and there were other people practicing for the test as well so whatever I didn\'t get we helped each other out.  Then I had the math test.  Surprisingly I didn\'t think it went all that bad.  I mean I know that there were some questions that I didn\'t understand but there were also lots of questions that I totally understood.  I really hope I did good on that.  And well after interviews thurs my TA knew I had some issues with some of the teachers and then in CALM today I don\'t know how we got to the topic but then she actually told me all the stuff they talked about.  I only asked her to talk to my Chem teacher but I think she talked to my math teacher too.  She said that they said that I doubt myself alot and that sometimes I can do things but I think that I can\'t do them.  And then she said that they said that I am really hard on myself.  I guess now that I think about it I can be a kind of hard on myself at times, and I guess sometimes I do doubt myself.  I think my TA also mentioned something about structured reviews to like the math and the chem teacher.  I am hoping maybe the Chem teacher will like have a little more structure to her teaching and stuff now that well I think my TA told them that.  Well yestarday well pretty good.  Me and Travis did some stuff together.  We were going to go into his hot tub but then I decided not too because I didn\'t want to lie to my parents because I can\'t lie - I hate lying, and well I just knew if we did his parents might come home or something.  So I went over to his house and we played X-Box, GameCube, and Playstation for a couple hours.  Then we went to my house and I made him some lunch, actually a really huge lunch.  We has that spiral KD stuff and then we shared some pizza too.  Then I made us some hot chocolate.  It was really good actually, he said I was a good cook.  Then we played my N-64 and I kicked his ass.  LOL.  Then we talked for like a million hours.  Later that night I watched Joe Millionaire finale which was soooooo sad and happy, and like a million other emotions.  When I was trying to go to sleep I just started crying for about five minutues, but I can\'t even remember why I was crying.  I think it might have been that I was really stressed out about the math test, that could have been it.   '),(3760,1807720,'23,November,2003','2003-11-23 00:00:00','    So I got up early this morning and then I started on my math.  I could do more of the review assignment because it was stuff was early on in the unit but the stuff near the end of the unit is really confusing me.  I am trying to understand it and I want to do good, but it is very hard to be optimistic about the math test since I don\'t seem to understand the stuff.  My dad can\'t help me cause he isn\'t coming to visit until next weekend and my mom knows nothing about math.  Basically I have to figure it out all of my own, and it certainly is not easy.  You know if I had one wish right now it would be to just understand my math.  That is the only thing that I really want at the moment right now.  It is stressing me out.  The more I try to understand the more I realize I don\'t understand.  The more I practice the more I worry about the math test being hard.  The more I don\'t understand the more I think that I won\'t do good on the math test.  I don\'t want to think negative right now, but it is hard not to.  I need a break right now though just to think about things and all this not understanding math is making me fell bad about myself.  Tonight after supper I will do more math cause during the night it is easier to focus and remember things.  And well yeah I can\'t wait til\' Friday cause my dad comes in from Vancouver on Thursday and that Friday we get to go out for Chinese.  Then Lenny is turning 16 so I will probably go over and visit with him and some other people later that night.  Travis is suppose to call me tonight to make arrangements for tommorow but I will see if he actually calls.  I don\'t know whether to believe him or not.  He told me he was working until three o\'clock today and like I logged on to my msn before 3 to check my email and he was online.  It kind of bothers me that I can\'t seem to trust him, although I want to trust him.  I feel that something is not right and he is not telling me something.  I don\'t know if I sould wait to find out what it is or come on to him with a strightforward question asking him what he is not telling me at the moment.   I got this in an email a while ago and saved it because it had meaning to me:     Sometimes people come into your life and you know right away that they were meant to be there, they serve some sort of purpose, teach you a lesson or help figure out who you are or who you want to become.   You never know who these people may be; your roommate, your neighbor, professor, long lost friend, lover or even a complete stranger who, when you lock eyes with them, you know that very moment that they will affect your life in some profound way.   And sometimes things happen to you and at the time they seem horrible, painful and unfair, but in reflection you realize that without overcoming those obstacles you would never have realized your potential, strength, will power of heart.   Everything happens for a reason. Nothing happens by chance or by means of good or bad luck. Illness, injury, love, lost moments or true greatness and sheer stupidity all occur to test the limits of the soul.   Without these small tests, if they be events, illnesses or relationships, life would be like a smooth paved, straight, flat road to nowhere. Safe and comfortable but dull and utterly pointless.   The people you meet who affect your life and successes and downfalls you experience, they are the ones who create who you are. Even the bad experience can be learned from... Those lessons are the hardest and probably the most important ones.   If someone hurts you, betrays you, or breaks your heart... forgive them, for they have helped you learn about trust and the importance of being cautious to whom you open your heart.   If someone loves you, love them back unconditionally, not only because they love you, but because they are teaching you to love and opening your heart and eyes to things you would have never seen or felt without them.   Make every day count. Appreciate every moment and take from it everything that you possibly can, for you may never be able to experience it again. Talk to people you have never talked to before, and actually listen, let yourself fall in love, break free and set your sights high. You can make of your life anything you wish. Create your own life and then go out and live it.   I wish you all the best in your endeavors as well as struggles in life. Have a fighting spirit and never hesitate to get back in the struggle!   '),(3761,1807720,'22,November,2003','2003-11-22 00:00:00','    Today was boring really boring.  I wish someone would call me.  No one ever calls me.  Apparently it takes way to much effort to call someone like me I guess.  I finished my CALM stuff today and did my Chem homework.  I finished some of my religion but I am going to have to go early Tuesday to finish my Religion stuff cause I brought home the wrong religion book.  I got distracted by Travis yesterday cause after school he just showed up at my locker.  And well it kinda creeped me out though cause I was like \"I thought you didn\'t know where my locker is\" and he was like \"Oh I was following you.\"  Like ok that kind of creeps me out when someone is following me and well since I have Chem when he had Science he was following me again on the same day.  And well I would have finished more Religion if I wasn\'t dumb and brought home the wrong book but I guess it happens and I will just have to live with that fact.  Tommorow is going to be ALL math.  I set aside ALL of my Sunday just to focus on my math.  I need to figure out what I am doing and PRACTICE, PRACTICE, PRACTICE.  That is all I can really do at the moment and try to understand it cause this unit is REALLY CONFUSING and yeah FRUSTRATING and just about everything else.  All I did today was eat mounds of food like seriously tommorow I am going to have to majorly cut back because well I don\'t want to gain any weight but everything I thought about how much I don\'t understand and how I dread the math test I just kept eating the cookies and chocolate.   '),(3762,1807720,'21,November,2003','2003-11-21 00:00:00','    Yesterday was pretty good.  Well I missed math basically cause no one had no idea what class to go to first cause they had to be dumb and switch the schedule around.  It was such a short day yesterday - it was great.  Chels and I went to go have some pizza for lunch then we walked back to my house cause the overpass is up no and there is no point waiting for a bus.  I showed her my dress for Winters Eve and well then we sat on my bed for three hours talking while we drank hot chocolate.  Then of course she had to work and well I started on my homework and studying and stuff.  Then later we went to TAPS and well I told my TA all about my true opinions.  My mom was kinda annoyed cause she thought my TA was ignoring her.  I am so glad Chels made honors too cause she really does deserve it.  Chels, Sarah, Monet, Lenny, and I all made honors.  Now we call ourselves the \"Charmed\" ones - Monet, Chels, Sarah, and Myself.  Lol.  We are all like obsessed with the show charmed which reminds me that I think it is on tonight.  I still have to watch the episode of survivor I taped and the finale of Bachellor that I taped.    Today was ok I guess.  Just ok like one of those days.  It is just that math that is really worrying my right now.  I know she is like trying to help me understand the truth is I DON\'T UNDERSTAND.  I have no idea how to do half of the stuff and half of my homework I can\'t even do.  I am really really REALLY worried.  Tommorom is gonna be dedicated to finishing my religion stuff and well maybe finish my CALM stuff in the evening.  Then Sunday I think I will use the entire day to prepare for math and then I just have like two Chem questions to do.  Monday of course is set aside for Travis cause his parents won\'t be home so we are going to spend some time in his hot tub.  *I Can\'t wait*  And yeah I do plan to make out in there and I am sure he is just as eager to do the same thing, but I don\'t want to take it all the way cause I am not gonna be stupied and get pregnant or anything stupid, but yeah just makeout that is all.  No parents - ALONE...yup that is right we will be all alone.  And well yeah I knew my Chem inside out and well on the quiz I started to confuse myself.  I sure hope that I still did good though because I really studied ALOT for that and I KNOW that I DO KNOW the material.  I need to get my Chem mark up and well my Math mark would be nice to raise but for now I think a realistic goal would be just to maintain it right now.  And well yeah Travis bought me lunch afterschool and stuff and well then he went off to work and I left.  All I have been doing today is eating...chocolate, pop, cookies.  Yeah so I guess it isn\'t a bad thing to reward myself for doing good is it??!?!...lol.  Yeah as long as I don\'t gain 5 pounds cause yeah of course my image worries me.  I want to look good of couse.  Guys seem to gravitate towards me now that I am 42 pounds lighter.  So yeah I want to go watch Charmed now and I think Boston Public is on tonight but chances are that it is a repeat.   '),(3763,1807720,'19,November,2003','2003-11-19 00:00:00','    Today was pretty good.  Well I still don\'t really understand my math though, and well even though we get monday off if I don\'t understand how to do the stuff then how will I have time to prepare for the test properly if I don\'t know what I am doing.  I really need to do good on a test in math, but how.....HOW HOW HOW...how am I suppose to do good if I don\'t understand.  You know actually today I felt like she actually wanted to help me.  For once I felt like she actually understood that I really have no idea about what is going on.  I really liked the fact that she actually sat down and took the time to help me instead of how she normally acts when I don\'t understand something.  Though I feel kind of bad, I don\'t want to have to keep asking her over and over again \"How do I do this, how do I do that\" because I think I will drive her up the wall and I feel really bad that I don\'t understand.  I wish I understood and I am trying to understand.  The Chem lab was really fun and interesting today.  We got to use a molecular model kit and build different bonds that form in a reaction.  We also had a sub today and I guess I was sort of happy.  But well I still have to admit that yesterday was seriously the best lesson she has ever taught.  I get this new unit and I think it is totally my thing.  For once something that I\'m good at!  Well the chem quiz is Friday but I usually do a 15-30 min. review everynight of the material so I don\'t have to cram it but I understand this new Chem unit.  It feels good to understand.   '),(3764,1807720,'18,November,2003','2003-11-18 00:00:00','    Ok well today was pretty good.  Well besides the point that I DON\'T understand my math that we leart today.  I was kind of freaked out to ask my chem teacher about a couple questions from the homework but I did and she was actually really nice to me.  And well not to mention that we had the most interesting Chem class ever and it wasn\'t boring at all and I kind of understand the stuff.  Chem was fun today!  Although I was trying to copy down notes and like the over head is my way and same with the teacher and like I think she DOES realize she is in my way but she contines to stand there.  I swear she does that just to tic my off!  LOL.  So yeah that was basically my dad and well today me and Travis actually connected.  We started talking and well it felt right.  I found out he has a hot tub too and well that is like MY BIGGEST romantic fantasy ever - to be in a hot tub with a guy!!!  Wait a minute that will be more like reality soon!!  So yeah that was my day and well I guess I should get going on my homework now.  I was just listening to this song...and I love these lines of it.      Looking back now, itâ€™s so clear to me That you were sent to shape my destiny Because there were times when I was ready to walk away But you gave me strength to face another day  ~Ryan Malcolm (Something more)   '),(3765,1807720,'17,November,2003','2003-11-17 00:00:00','    Today I came to school happy and said \"Hi\" to like a million people.  We had extended TA and I got to see my report card cause my TA had em\' and I DID IT!!!!!!!!!!!!!!!!!  I made honors!!!!  I think I am gonna jump up and down now.  LOL!  I worked so hard, I derserve it but I am still worried about if I can still make honors at the end of the entire semester.  I found out my math quiz mark and it was actually kind of bad and well I felt bad at the time and well felt like crying but I didn\'t thankful and well when I was going through the quiz with her she decided she was actually marking kind of high and gave me a higher mark so I felt a little bit better about that.  I was kinda scared about what I got on the chem test but I got a 61%.  I guess that isn\'t like amazing but I DID EVERYTHING I could to get that.  I was thinking of maybe to re-write it but I mean I did my very BEST on it to begin with and what I did on that test is my very best.  I hate to think of 61% being my best but only I know how much effort and time went into preparing for that test.  At least I didn\'t fail cause I know lots of people that did and well I was kind of scared about it cause I failed the two quizzes from that unit.  Hopefully it won\'t affect my average too much and well I have to keep in mind that there are other quizzes, other tests...more marks that go in for both Chem and Math.  Oh and well I think next term I am going to drop Physics.  The thought of taking physics keeps haunting me and if it is anything like Chem then I don\'t think I can deal with that.  I want to take Active Living 20 instead because that is what I origionally wanted to take but I did not have to room to fit it into the schedule.  I think I want to do that and I am almost positively certain at this point.  I felt pressure before cause my parents kept wanted me to do Physics and now my mom said \"Do what you want\" and well I want is to do Active Living.  Yesterday I wateched the game.  YEAH EDMONTON!!  Yeah they won!!  And well I went shopping earlier and got some new jeans and this new sweater and some bras [yeah girls need those] and well I bought my dress for winter Eve.  I looked everywhere but then I walked into this one store and looked at this dress on the manican and I said \"That\'s the one\"  It instand captured my attention and when I put it on I don\'t think I have EVER and I mean EVER looked so damn hot in my life.  All the guys are gonna think I am hot on Winters Eve lol....and the guy I go with man will he be in for a surprise.  I can\'t wait for that dance!!!  I think I am also gonna get my hair done professionally this year cause last year it didn\'t turn out the way I wanted it..so yeah.   '),(3766,1807720,'16,November,2003','2003-11-16 00:00:00','    Hmm lets see I get to go to see Love Actually after and well I am going to Chinook to go shopping.  Then we are getting pizza and watching the Grey Cup.  I think I am gonna cheer for Edmonton because afterall it is pretty close to Calgary.  Last night I studied some more for Social but I feel I am ready and then I played Tetris on my GameboyAdvanece.  And I decided I am going to do some self-discovery in my blog today....      01) First grade teacher\'s name: Mrs. Margaret - I am guessing that was her first name 02) Last words you said: Are the meatballs done already?!? 03) Last song you sang: S Club 7 - Don\'t Stop Movin\' 04) Last person you hugged: Probably my mother 05) Last thing you laughed at: On Friday me and Chels remember when we saw two chics walking through the Tim Hortons drive through instead of driving 06) Last time you said I love you: and meant it? I don\'t think I have ever said that, I just feel love 07) Last time you cried: Hmmm, actually I don\'t know the answer to that.  I have cried so many times I have lost track 08) What\'s in your CD player: A burned CD with old mucis like Kids in America, Toge+her, Celine, Michelle Branch etc. 09) What color socks are you wearing: Black 10) What\'s under your bed?: I have droors (sp?!) under my bed, and well the rest is dust 11) What time did you wake up today: I don\'t know, who keep track of that on weekends 12) Current taste:  In what?? 13) Current hair: Brown with some blonde, and red but mostly brown 14) Current clothes: Navy shirt with jeans (OMG I wear navy on weekends too..What is wrong with me?) 15) Current annoyance: nothing 16) Current longing: To be with my one true love...and to pass Chem 17) Current desktop: Plain old boring blue 18) Current worry: I really hope I passed that Chem test cause I studied alot 19) Current hate: I don\'t think I hate anything right now except maybe my hair 20) Current favorite article of clothing: Old Navy jip up jacket 21)Favorite physical feature of the opposite sex: eyes and hair 22)Last CD that you bought?  LiveOnRelease but I want the Celine - One Heart and Clay Aiken\'s CD 23)Favorite place to be: At School or at the mall 24) Least favorite place: None 25) Time you wake up in the morning: I thought I answered this question already 26) If you could play an instrument, what would it be: I already play the flute but I want to learn electric guitar 27) Favorite color: blue and black 28) Do you believe in an afterlife? Yeah of course! 29) Current favorite word/saying: Believe in yourself and you will succeed (It was on my fortune cookie once) 30) Favorite books: Speak but I don\'t know the author 31) Favorite season: Summer!!  duh!  Shorts, Sun, my Birthday! 32) Favorite day: Monday cause it starts a fresh new week 33) Where do you want to go: Hawwaii 34) What is your career going to be like: honestly? Well my full descrition of this could take me a while but bascailly I will be standing in front of many older kids and they will all look up to me.  I will be a role model and an inspiration to others.  I will teach Bio and be damn go at it.  It will be fun, enjoyable, and rewarding.  sometimes it might be hard but in the end the true rewards will out weigh the bad. 35) How many kids do you want: Two 36) What kind of car do you have: None at the moment but I might get my sisters silver Malibu next year 37) Random song lyric: \"One heart you are following, one dream keeps you wandering\" 38) Identify some things surrounding your computer: This is a dumb question     ___________________________________________________________________________________________       01) What color are your kitchen plates? Blue 02) What book are you reading now? My Social books cause I have been studying and Chem 03) What\'s on your mouse pad?  People on a spaceship pointing to the sky 04) Favorite board game? Twister, but if not a board game I like Uno and Truth & Dare Jenga 06) Favorite smell? Roses 07) Least favorite smell? I don\'t really think about that 08) What is the first thing you think of when you wake up in the morning? I thought about my dream 09) Favorite color? Blue 10) Least favorite color? A puke green 11) How many rings before you answer the phone? I answer when I feel like it 12) Future child\'s name? Girl:  Janine or Edele  Guy:  Eryk or Jaden 13) What is most important in life?  My grades and passing because it will determine my future 14) Chocolate or vanilla? Chocolate all the way! 15) Do you like to drive fast? No, not really 16) Do you sleep with a stuffed animal? Are you kidding?!?!...If I was five maybe 17) Storms - cool or scary? Very cool 18) What type was your first car? None, unless you count the small one we did in the Physics unit last year..LOL 19) If you could meet one person - dead or alive? Alive 20) What is your sign and your birthday? Leo: Auguest 17th 21) Do you eat the stems of broccoli? Yeah sometimes cause I hate wasting food 22) If you could have any job what would it be? Biology Teacher 23) If you could have any color hair, what would it be? Half blue and half pink hair 24) Half full or half empty? Half full except when I get down on myself about Chem 25) Favorite movie?  Bring it On, Ice Age, and Finding Nemo 26) Do you type with your fingers on the right keys? Yeah 27) What\'s under your bed? Drawers 28) What is your favorite number?  5 cause it\'s the number I hate in gr. 8 for the Volleyball and Badminton teams 29) Favourite sport to watch? Football - it\'s great! 30) What is your single biggest fear? Failing chem and not having my dream of being a Science Teacher come true 31) Favorite TV shows? Survivor, Bachelor, Joe Millionaire, Charmed, Boston Public, CSI 32) Ketchup or Mustard? Ketchup cause Mustard is nasty 33) Hamburgers or hot-dogs? Alberta beef!! 34) Favorite soft drink? Cream soda....yum 35) The best place you have ever been? I am still waiting to go there, I want to go to Hawwaii - it will be the best 36) What wallpaper is on your computer right now? Blue wallpaper 37) Burger King or McDonalds? Neither, fast food is nasty 38) Does Love conquer all? Nope 39) What\'s the best feeling in the world? Making Honors cause it feels good for effort to pay off 40) What\'s the worst feeling in the world? Failing 41) who do you wish you could be with right now? Andrew, even since I lost weight over the Summer he thinks I am hot and well I am kinda sad to find out that he does have a gf though but I think we are really into each other.  In gr, 9 we went out on like one date and it was horrible but now we have both changed.  If he is ever single I think you will know who I am going for   '),(3767,1807720,'15,November,2003','2003-11-15 00:00:00','    I got up real early today.  I have lots of social to do and study today.  I have been doing it for a while but once it\'s done it\'s done, and once I know the material it is locked into this head of mine.  Once I get that out of the way I get to go shopping and to the movies tommorow.  Then we are gonna watch the Grey Cup tommorow.  Anyways - lotsa stuyding ahead!  Oh and I found this funand interesting thing...It was this thing to determine which Finding Nemo character you are most like!!!  LOL.  I love that movie.    urlLink      Click here to take the test!   '),(3768,1807720,'14,November,2003','2003-11-14 00:00:00','    I had the math quiz today which I thought I actually may have done good on.  I am pretty sure I understood most of it except maybe like the odd question.  I still am kind of worried about how I did on the chem test cause even though I prepared for it I found it very hard.  Maybe I did good though, I can only pray and well for once the chem teacher didn\'t give me an evil look.  Usually I am like scared of her cause I always feel like she doesn\'t like me or something but then after school she was like \"Hi Vanessa\" and actually had like a non-evil smile on her face.  Chels, Mike, and I all went out for lunch and we had the funniest server guy.  He came up to our table and was like talking in the REALLY FUNNY accent and well we were trying soooo hard not to laugh.  At first I thought he was faking it but then he didn\'t stop the accent thing but well when we got the bill he stopped doing it.  So yeah I guess he was just playing around with us but it was funny.  Then I went to chapters with them and well was suppose to go to matts but I called ash I told her I was not coming.  I don\'t care if they are mad at me or whatever cause ultimately what I do is my choice not theirs!  And well I just know someone is going to say something on Monday to me about \"Vanessa why didn\'t you come to Matts\" so yeah I\'ll be expecting it.  I am just kind of relaxing tonight.  I have lots of social stuff to do tommorow but once that is done it will all be good.  I am a little worried about the new stuff in chem though cause I don\'t understand it really well yet.  I think I am starting to get a hang of the math though.   '),(3769,1807720,'13,November,2003','2003-11-13 00:00:00','    I feel like little miss nice Vanessa.  Over the past days I have let Kendra who never even talks to me borrow my lab book and then today I could not say no to Thy.  Then in CALM Stephan kept asking me \"how do I do this...how do I do that?\" and he kept saying \"tell me the answer.\"  It drove me up the wall and well I feel like all my effort is just being used by other people.  For now on, even if I may come across as rude I really have to learn to say no.  There is a point to cross the line, and there is also a point where the line should not be crossed at all.  The chem test was soooo hard today.  I had confidence though, and I practiced and studied for a million hours.  I even reviewed at lunch time.  I am not too sure how I did but I know that the test sure was hard.  I do deserve a good mark though.  I deserve a very good mark.  If I have to rate myself I would give myself an A+ for effort and working and trying so hard.  I deserve it.   Oh and me and chels almost got hit by a car today.  We were crossing at the cross walk and there were cars stopped for us and like another car went by in the other lane and was speeding I am pretty sure and I suppose they didn\'t see us or something or they just didn\'t want to stop.  A car that was stopped honked the horn and me and chels turned around and the car was like right behind us.  Literally like an inch.  What a day!  A close encounter with death kind of, and well it actually made me think that well you never what will happen so we should live life the fullest.   '),(3770,1807720,'12,November,2003','2003-11-12 00:00:00','    My head hurts like a !%&@ and yeah lets just say too much studying.  I studied and practiced for chem tonight...like ALL night.  I have to do good on that test.  I have got to!!!  I put so much effort in, I can do it!  It doesn\'t matter what marks I got on the quizzes in the past...all that matters is the future.  And I am going to make sure I do good on that unit test tommorow in chem!  Yes Vanessa you will do good cause I can do anything.     Anything  You can do anything I believe in myself I am capable of anything I put my mind too Hard work, effort, and dedication I have it all, and I can do anything  ~Vanessa   '),(3771,1807720,'11,November,2003','2003-11-11 00:00:00','    Here in this life, stuck in the place Time is racing us to face Challenges ahead, wondering why What or who can help us to get by?  ~Vanessa     Here is a little poem I wrote just now.  I keep thinking about chem.  cause well I want to good on the unit test but I can\'t just snap my fingers and all my balancing will be amazing and everything will go well.  I can\'t just make that happen.  I think Ashley knows I am not doing well in chem and on msn she was like well couldn\'t you be a Social teacher or and English teacher instead of teaching Science.  Yeah k like NO.  I felt kind of bad after she said that though even though I know it wasn\'t meant to be offending in any way.  I have to follow my heart.  I can\'t just give up my dream.  I thought about giving it up this morning, but God that would hurt me way to much to give it up.  I when onto the U of C website and looked at all the diff. courses I could do and the things that I could study.  Everything turned me off and just totally disinterested me.  And well I know I am pretty good at Drama and like being on stage and acting but I don\'t want to do that for the rest of my life.  I want to teach Bio.  Of course now that Chem and Physics are ruled out...yeah I will make sure it is BIO.  But I have to pass chem and I am worried right now.  Then quizzes are only worth about one mark but the unit test is worth about ten.  Either my mark is gonna go up or it is gonna go down.  And I mean I can\'t get it out of my head now...I mean I FAILED.....FAILED DAMN IT.  I didn\'t try to fail, that was never my intention.  See that is why I am sooooo confused right now.  How could I want to be a teacher and stuff and love science and stuff but not be good at it?????????....  I would give anything in the world the understand and just get an amazing mark on that next test.   '),(3772,1807720,'10,November,2003','2003-11-10 00:00:00','    Today was like really long it felt like and well math...math...math.  Lets just say I am going to be working on math for a very very long time.  It was so confusing for me...not that I wanted it to be.  I felt really bad cause I kept having to ask \"how do I do this....how do I do that??\"  I think I have a little bit of a better understand now but last time I thought that I just go confused so I don\'t want to get over confident.  I just hope I can really understand the stuff for the quiz.  I am very afraid of the chem unit test for stoich.  I am really trying but my work is non-reflective of my effort at the moment.  The first quiz of the unit I got a 48% and then today I asked to go over my second quiz and I got 25%.  If I can\'t pass the quizzes then well that of course makes me worry about the unit test.  I need to do good, but if I am trying for the quizzes and I am not doing well then what will make anything different on the unit test.  All I ever do is practice, practice, practice, but that isn\'t really paying off at the moment.  Today I actually thought the chem teacher was being nice.  I felt she was actually very polite to me and I think she felt empathy for me cause of how I did on the quiz.  It was actually a nice to change to how I feel she normal acts when I don\'t seem to understand something.  Everyone kept saying they liked me hair though.  I am not really sure if they were telling the truth or if they were sincere.  I wanted to be different today.  Going for the cute look ya know.  I think I wanna do something to my hair - but there ain\'t no way I am cutting it again..I HATE short hair...ECK.  Maybe a dark color...with a red through it...yeah that is what I like for winter.  I need a change.  I positive change for once.  Yeah treat myself of something.  Make myself feel loved.       Hmmm I am so gald there is no school tommorow.  I have so much to catch up on and do.  I have the religion test, the chem test, the social test, and the math quiz.  I have way to much on my plate and I thought cutting out extra curricular would be a good thing but I miss it.  I miss being always busy and doing things I like.  Now when I fail at school I feel like I fail at everything cause I don\'t do hardly anything anymore!  Sigh...there are so many things to worry about when your in high school.  Yesterday night Ben, Melissa, Me, my Mom, my Dad, Grandma, and Uncle had a big turkey dinner.  It literally felt like Chrismas holidays.  Oh and I bought...the Finding Nemo DVD cause I love that movie and it is awesome but my sis took it back to their trailer so I probably won\'t get it back til the weekend or next time I see her.  Newayz yeah it feels good to get all my thoughts down and clear em from my head right now.  I think I am going to go try to finish my math assignment.  I was looking at all the older pictures and I came across ones from my party and when me and Chels used to be like way better friends.  And well yeah over summer we basically spent our entire lives together, she even put up my patio set with me and we spend like a million days together doing a million things.  When I remember all the great things it\'s sad to think that we are not as close anymore.  Here is a pic of me and her in August on the day of my birthday at my house and yeah it reminds me of some great times.   '),(3773,1807720,'08,November,2003','2003-11-08 00:00:00','    Well Friday was pretty good actually.  I thought math was going to be bad but I asked her a few things and she reassured me I knew what I was doing and stuff.  It made me feel a little bit more confident and I can do a few more things from my assignment now but there is still lots I don\'t understand.  Luckily me dad flew in from Vancouver last night and so sometime this weekend he is going to have to help me.  He is good at math, very good cause well my mom can\'t do like any of my math at all.  I am trying not to let too many things get to me though.  I mean when I think about it I always WORRY WORRY WORRY.  But like what is the point about thinking about things of the past, that are over and done with and cah\'t be changed.  What does worrying about those do...nothing but make me miserable.  So now I am just going to look ahead to future.  What is done is done, but what has not been done yet can always been changed.  So even though I failed a chem quiz and well think I failed the other one....well I failed and that is reality that I have to life with.  On the next test I will just try my very best and that is all I can do.  I already know I put in tons of effort into doing work, trying, and understanding so I guess the rest is left up to fait.  What is meant to happen will happen, and what is not meant to happen will not happen at all.  Well yeah so friday afternoon I picked up a bagel from the Y cafe and ate it on the bus home and then dropped off my stuff and then left again to go to chapters.  Looked at some stuff their for a while and then Alex, Matt, Sarah, Ash, Thy, Travis, Raul, Lenny, and of couse myself all went to see Matrix: Revolutions.  I thought that action was awesome and stuff which is the only reason I go see those cause well I never really understood the whole story line too well.  I didn\'t want Trinity and Neo to die though but I guess they had to cause it was literally the end.  Actually I really liked the main quote and theme throughout the movie which was \"everything that has a beginning has an end.\"  I like that quote alot because when you think about it, it is soooo true.  And well after the movie I was going to go to South Centre but my mother wanted to wait for a phone call from my dad cause he though he would miss his flight.  The traffic was backed up in Vancouver and well he ran for like well far and JUST like JUST made it on in like the last minute.  And well YAY!!! I finished my religion projects....!!!!  What a relief...I got up at 8 so I could get a head start on em and ya it helped.  Not I actually have some time to spare [yeah like now lol] and well I get to RELAX....now that is word I sure don\'t hear everyday.  Well company comes tommorow.  My Uncle is flying down from Winniepeg and we are having a big dinner tommorow with my Uncle, Dad, Grandma, Mom, Melissa, and Ben.   '),(3774,1807720,'06,November,2003','2003-11-06 00:00:00','    Tonight was awful.  I just don\'t understand my math no matter how hard I try. I cried for about an hour off and on, and I kept trying over and over again to do the math.  I tried and tried to understand my math but I just don\'t \"get it.\"  I really wanted to finish at least one math assignment too but I couldn\'t finish it.  I feel so bad, like I am giving up but that is certainly not the case.  I don\'t think I am doing too well in chem either.  I don\'t know what to do.  I know lots of people have difficulties learning in life and stuff but things should not be really diffucult for me like this all my life.  Literally all my life things have been difficult and it makes sense that maybe I do have some sort of learning problem.  People don\'t seem to think so though, they think there is nothing wrong with me but man if they knew how many times I cry and get frustrated with things and how difficult everything is.  I put so much effort in and get hardly anything out and it is soooo annoying and not fair.  How come I can\'t get good marks if I put so much effort into trying and trying to understand and doing my homework.  I don\'t know what is wrong with me but something is...I know it.  I know that things in life should not be this hard or frustrating and they should not have to go through all I have been through.   '),(3775,1807720,'05,November,2003','2003-11-05 00:00:00','    Today was uh bad.  Yeah bad.  I did the math test in the morning and I got all confused with the word problem stuff and well I checked on the ucourses sight and I got a 69% which isn\'t too bad but it could be better.  I guess it wasn\'t too bad though.  The chem quiz was bad though.  I studied...I practiced...I did my work.  I did the first page fine and then I flipped the quiz over and the second side was like a nightmare in a quiz.  It was awful.  My mind just shut down and I could not figure out how to do some of the stuff.  I felt bad and well yeah I failed my other quiz and the one thing I really did not need right now is to fail another one.  I don\'t know what to do.  Everything I try to explain to her that I don\'t understand and I really need help all she tells me is that it is easy for her and says \"I don\'t understand why this is so hard for you.\"  She makes me feel bad and stupid when I ask her things.  It isn\'t my fault that I don\'t understand everything she teachs.  I just don\'t get it, and even when I did think I understood I failed.  I need to pass chem and well yeah get a good mark.  When universities see that I barely passed chem then I don\'t think they will even take a second look at me.  I was so down today cause I kept thinking about university and stuff and the future.  I know I have to follow my heart if I want to be happy, but this stage right now isn\'t the happiest of em\' all.  I think I might be one honours though for the half term.  Fingers crossed, I just have to pull off a chem average over 65% which yeah I think I am capable of.  My latest labs I got like 100%, 80%, and like a 93% so that will really help my mark cause the quiz thing certainly yeah did not go over well.  Now that I think of it today wasn\'t totally bad, I guess it was ok.  I am going to try be a little happier tommorow but the weather is really starting to affect me.  Also Travis kept talking to me about websites and games and stuff.  To be honest it was boring.  I got over that stuff in like grade 8 and like seriously if I could I would tell him \"move on man\" but that would be rude and I don\'t want to offend him.  I don\'t really share too many of the same interests as him but I notice that he is falling for me, and coming on to me, but I am not falling for him.  I am falling for Andrew B.  I remember in gr. 9 we did not really like each other and like went out and it went horrible but he is in my CALM class now.  I always end up talking to him like all the time and it is sooooo natural.  It feels right and he has goals in life and takes school very very seriously.  I like that about him and well I like her personaliy.  He is friendly and easy going and stuff.  I don\'t know if he feels the same way but I think it might a little bit cause I am getting the signs you know.  I don\'t want to be rude to Travis though and I don\'t want to let him know I am not really into him.  He keeps trying to like \"get me\" over to his house.  I am kinda freaked out about that.  He kepts condums in his pocket and keeps talking about how he wants me to come into the hot tub with him.  Like if he wants \"you know\" and stuff than I am not the girl for him.  I don\'t want hurt him though and I think I will still give it a go and do something on Friday with him.  I am afraid that if I tell him that I am not into him the way he likes me than he will go back to not doing his working and cheating on tests.  After I started listening and paying attention he made more of an effort to do work and didn\'t cheat again on science as I am aware.  See that is what bothers me though.  He isn\'t as serious, focused or dedicated as I am.  I have goals in life, dreams.  I want to go places, be someone.  For him I can\'t really say which in a sense is kind of sad.  I guess whatever happens will happens.  I believe in fait and destiny, so whatever will be will be.     You\'re Most Like The Season Summer ...  Whoa.... Passionate eh ?? Typically you\'re a fiery, zesty dominant person. As the hottest season, you certainly ooze Sex appeal. You have confidence which draws people to you, and you have the makings of a good leader. However sometimes your exterior is stronger then you are and so you scare people off before they can get close.  Well done... You\'re the most memorable of seasons :)    urlLink   Which Season Are You?  Take the quiz!   '),(3776,1807720,'04,November,2003','2003-11-04 00:00:00','    I hate everything right now.  I hate myself.  I hate my life.  I just hate all of this, just everything.  I can\'t take it, I can\'t.  I came home from school and there me and mom go.  Arguing off and on about stupid stuff.  And ARGGGGG she was like \"here have a muffin and shut up\" and I was like \"uh no, why the hell did you have to go and buy muffins anyway\" and then like I was gonna eat and she was like \"dinner is not ready\" and she was like \"have on of these\" and it was a stupid packaged dinner.  she DOESN\'T understand that those ARE NOT in anyways healthy and I told her that I would not eat something that nasty.  And then she was like \"well then what the fuck do you want\" and ARRGGGGG I can\'t fuckin take any damn fucking thing right now.  Stupid chem, stupid arguments, stupid everything.  I went onto the ucourses site during social and looked up my chem marks and there was my quiz mark - a 48% a stupid 48%.  A 48 for what...for working my ass off to understand, for trying SOOOOOOOOO damn hard, for trying to actually pass chem.  It is not fair, I hate it.  All my wasted effort and time and for what - a stupid fail mark.  It makes me want to just burst into tears but I haven\'t have the full out explosion yet, though I am thinking it might come sometime later in the week.  The chem teacher does not understand.  She doesn\'t understand that the stuff if REALLY HARD for me, I can\'t help it!!!!!!!   I feel like it is all my fault, like I am stupid or something but obviously that is not the case.  I want to know why everything is so damn hard.  I still wonder if this is normal for people, I mean of course not everyone will understand everything always but I am SO DAMN SICK OF IT.  I am NEVER good at anything, NEVER!  Not when I young, not now when I am older...plain and simple nothing has even been easy.  And you would think I would be used to it all by now but things get harder - way harded and defenetely not easier.  I wish I had answers for things but there are no answers right now, there is just feelings.  These feelings of anger, frustration, and yeah I do feel upset.  I didn\'t even think I did that bad on the chem quiz - I did all the questions and really put effort it.  How come a stupid 45%...HOW COME????????  What the fuck did I ever do to deserve to be doing to bad in chem.  Oh God it kills me to think that maybe I should give up my dream of ever being a science teacher, and tears come to my eyes.  But what do I do..I don\'t know what else to do about everything??  I just don\'t know.  I am so confused and have just about every damn problem in the world right now.  I don\'t want to say it and I don\'t want to believe it but I just want to be dead right now.  So much right now actually.  I can\'t take all this, this life - I still don\'t know what I am good at and there is nothing that I really excel at.  Other people have their looks, their grades, their talents but what I do have.  I don\'t know the answer to that, I don\'t know the damn answer to anything right now.  No one understands how it feels, they just don\'t understand, and would not be able to understand.   '),(3777,1807720,'03,November,2003','2003-11-03 00:00:00','    Today was alright I guess.  It was just the chem teacher that was really bothering me.  I needed help with stuff and so I asked her some questions about the lab and she was like \"I don\'t see why this is so hard for you.\"  I always get this vibe from her like she underestimates me and thinks I am not as smart as everyone else.  Maybe I am just imagining it, but I really feel strongly that that is how she thinks and feels.  And well then she checked everyones work before they left class and she was like to me \"Who did this?\" and like I mean yeah Paige helped me a bit but I was the one to do the stuff, and she was like \"this is good work\" but she didn\'t belive that I did it.  I felt really bad after that.  I mean that is just an example how she underestimates me and like she thinks I can\'t do anything apparently.  I am getting frustrated again, I am so sick of trying to hard.  Effort I put in is not paying off.  It feels like all my energy is being wasted, and for what.  My one true dream maybe, it is what I want most.  But I don\'t know.  I just wish someone had an answer to everything in life.   '),(3778,1807720,'02,November,2003','2003-11-02 00:00:00','    Today was good.  Matt came over and helped my with my math.  We worked through those circles booklets together and helped me to understand that stuff a little bit better.  I am still confused about chem but hopefully the teacher will be there tommorow and she can help me during tutorial.  When my mom got home from work (her other job doing demos for cat and dog food) she took matt and I out for dinner.  We went to Swiss Chalet and ate like tons and had dessert and everything.  It was so good but I made him promise me that he would stop me from eathing junk all week.  LOL.   '),(3779,1807720,'01,November,2003','2003-11-01 00:00:00','    ^_^ Six hours to finish my social project.  Six friggin hours - like who is soooo getting 100 - I better!!!  I did my chem lab, well what I understood of it but I can\'t answer the homework questions cause I don\'t understand what to do.  It is not that I don\'t want to do my homework, cause I always do homework.  But I feel kind of bad cause I can\'t really do it.  I think Matt is coming over tommorow and then he is gonna help me with math later cause I don\'t have a clue of what is going on and the math teacher didn\'t \"teach\" the unit so I am confused and she does not seem to understand that not everyone understands.  I think I have a religion project to do but I don\'t think I really want to work on that this weekend so I will probably do it next weekend as long as it is not due this week but I don\'t think it is or well if it is I am going to be one busy girl during the week.  Last night was Halloween like ya and it was funny cause my mom was like to some kid \"Oh boy you sure look scary\" and then he was like \"but I am a dog\" Lol...She thought he was an abominable snowman.  And then she was like \"oh faith day was good\" and then like in the night she is like \"oh faith day is so boring, I don\'t like it\"  like OMG I can\'t believe she would say that.  She does not realize how lucky she is cause she gets to go to that.  I can\'t wait til I am done University and I have a job teaching and then I will get to go.  And well yeah nothing too much this weekend.  I kept thinkin\' bout the homework I need to finish but weekends end up being me trying to catch up on everything from the week and I don\'t ever seem to get a \"real\" break from things.   '),(3780,1807720,'29,December,2003','2003-12-29 00:00:00','    Went shopping lots today and yesterday and got tons of new shirts *THEY ARE NOT NAVY OR WHITE* thank God too cause I am soooo sick of those colors and all my old shirts that aren\'t navy and white are like 5 sizes too big.  And well I got myself some clicky shoes, I have been wanting them forever.  And well this song/poem just came to me in my mind like about 10 minutes ago so I started writting down what I could hear in my mind.     Fighting the Cold  Out in the cold Iâ€™m lost In a reality of wandering Dreams trapped within themselves Stored up on the shelves  I think now is the time To unlock the cupboards Throw away the keys Let your dreams just be  You got to be you You got to be whom you are inside You have to believe That you can succeed Your willpower will override All your fears washed away Forget them for today these dreams are here to stay Follow your heart; donâ€™t do what your told In this life just keep on fighting the cold  You are the only one who knows you know where you want to go If someone comes along one day And tells you to throw your dreams away  Say â€œhey, do I look like a fool?â€ They tell you dreams are foolish and you should go away And you say, â€œdonâ€™t tell me where to go or what to doâ€ This is the way that it will be because I believe that  You got to be you You got to be whom you are inside You have to believe That you can succeed Your willpower will override All your fears washed away Forget them for today these dreams are here to stay Follow your heart; donâ€™t do what your told In this life just keep on fighting the cold  The reflection staring back at you Tells your something like dreams can come true If you believe and if you have faith You will get swept to a land far away  Your heart will guide you to the Promised Land Never know if dreaming comes to and end So if youâ€™re in doubt donâ€™t give up on yourself You have to believe that things will work out, and I said  You got to be you You got to be whom you are inside You have to believe That you can succeed Your willpower will override All your fears washed away Forget them for today these dreams are here to stay Follow your heart; donâ€™t do what your told In this life just keep on fighting the cold  ~Vanessa   '),(3781,1807720,'26,December,2003','2003-12-26 00:00:00','    Well I got back from Jim\'s place tonight with this big dinner thing.  Ben and Melissa were there, Doreen and Jim, and Ben\'s bother pete and his wife Judy, and their two kids were there.  It was pretty fun but when we opened our gifts one of the kids got a puzzle and I kept having to do it over and over again.  Like me and him would finish the puzzle and then he would dumb out the pieces on my lap and start putting it together again.  After he finished it for like the 10th time I said \"Why don\'t you go show mom\" but he just continued to dump the pieces on my lap.  After I finally got away me Ben and Melissa played Skip-bo which is like a really fun game.  It was fun and we AGAIN had lots to eat.  Then I just did some Social stuff when I got home and now all I have left to do is my Math but I don\'t feel like doing that at the moment cause I am on holidays.  Tommorrow I get to go shopping.     ___________________________________________________  I survived the crash, I survived the burn Survived the worst yeah baby but I learned Survied the lies, survived the blues Almost killed me but I survived the truth And when you wrote me off like I was doomed I survived you  ~Clay Aiken ___________________________________________________  Open to New Possibilities  Breathe and let all your worries fall away Just see what a difference it can make Let your mind just drift along today Itâ€™s OK and everythingâ€™s going to be just fine  Laugh a little and smile once more Walk along until your find that door Itâ€™s that one youâ€™ve been searching for Here I can see it opening up for me  Step inside cause youâ€™ve got nothing to hide Donâ€™t be afraid it wonâ€™t hurt at all Thereâ€™s a shinning light blinding me I have found my way and now I can see  Open your eyes and open you heart, I opened my soul right from the start Believe in yourself and belief in your friends It took some time but now the believing never ends  See in your heart and see in your mind That everything will turn out right with time Just open your hands and accept what comes Who knew the surprises of life could be fun  Follow your heart and follow your dreams Donâ€™t turn away because you donâ€™t believe Walk on the grass or walk on the road The choice is yours, go where you want to go  ~Vanessa ___________________________________________________  Somebody There  Tears descend from your eyes Let go of the hurting your concealing Be true to what your feeling There will forever be somebody there  They will seize you when you stumble And hold you to let you know they care When you think your lost and you think itâ€™s over There will forever be somebody there  ~Vanessa   '),(3782,1807720,'24,December,2003','2003-12-24 00:00:00','    So we ATE and we ATE and we ATE Today and Yeseday...and the day before.  Ok she so had junk all week.  Yeah I feel my stomach starting to expand actually - kinda scary thought and tommorow I have ANOTHER dinner thing to go to.  Ben and Melissa came over for Christmas Eve and we opened our presents that night and we had like a ginormous dinner.  Like EWWW though, how do people drink Eggnog, that stuff should be like illegal it\'s so nasty.  They should call it holiday Gagnog instead and whoever came up with the idea with putting dyed fruits in a cake was really messed up.  Fruit and cake don\'t mix!  So yeah then Ben and Melissa came again today and we had a HUGE breakfeast and ate junk and chocolate ALL day.  I thought my food was going to come up after dinner too.  I didn\'t realize I was full until after I finished eating.  Actually I felt like I was 5 million times way too full.  Too much food intake.  Too much food and junk intake ALL week.  And I have to eat a huge dinner again tommorrow out in Cochrane at Jim\'s place (Ben\'s mother\'s Boyfriend).  And well I liked all my gifts, but of course my sister was being her fussy little self and always complains about what she gets and stuff.  You would think since she is older she would appreciate things more but whatever.  I am not her and we are two entirely different people so I really shouldn\'t be on to judge.  Anyways here\'s the list of the stuff I got for Christmas:          -XBOX -Watch -Hello Kitty charm for my charm bracelet -New Diskman -Blue cord purse from Gap -Clay Aiken CD -Socks cause it\'s not Christmas without an article of clothing -Lipgloss -Pen with my name on it (lol) -Birthstone Ring -Money -Gift certificates to HMV, The Body Shop, and La Senza -Chocolate and mints -And five extra pounds from all the junk I\'ve been eating          I also updated my portfolio.  I didn\'t realize I had so many talents.  I didn\'t really that I AM REALLY SMART, and all this time I didn\'t pay attention to all the awards I have for like a million different things.  I am talented in the academic area (well I try REALLY hard) but I still call that talent, and I have tons of sports awards, and lots of music stuff too.  I am multi-talented.  I didn\'t even realize it.  And I totally forgot about the Winter\'s Eve dance, I don\'t think I mentioned anything about it in my blog ever.  It was really fun that night and I do think I look sexy in my dress.  I love it, but well since I gained some weight it doesn\'t look as good on my now but well I am still hot in it.  I don\'t care what anyone else says cause I like it!  That night was yes VERY fun, but I am way to lazy at the moment to go into the story of that night cause that would probably take about five hours to explain the night.  I don\'t have the time or patience to write for five hours.   '),(3783,1807720,'24,December,2003','2003-12-24 00:00:00','    This strange like song/poem just kind appeared in my head today.  I had the lines \"Just when I thought things couldn\'t get any better, I closed my eyes and you showed up at my door\" going through me head.  So then I thought I would try make a poem out of it but I didn\'t like even think of any of words in it.  They all sort of just \"came\" to me.  Strange isn\'t it.     A Feeling  Just when I thought Things couldnâ€™t get any better I closed my eyes and You showed up at my door I thought itâ€™s over But itâ€™s never to late begin On this journey I feel like I am starting to win  Closed my eyes, now I see Took my hand, now I feel Love is something that is real Thatâ€™s the way it should be Together, just you and me Feel like I am flying above and beyond all the earth I found my wings and now I can breath, I can breath This feeling inside is how I am  suppose to feel, this feeling is real  I never knew that life  could have so many turns, all those bumps in the road I never thought of how someone Can walk into your life And turn your life around I never dreamed before of a life like this A life full or respect for me That is how it is suppose to be I finally see, I finally see A feeling described, a feeling so close Something now, a feeling so real  ~Vanessa     So yesterday I had some people over and we watched Freaky Friday and played Mad Gab for like two hours.  Then we played air hockey and everyone seems to like that.  Then I went to the movie Something\'s Gotta Give which was absolutely hilarious.  I am almost done my social stuff too!  I feel like so non-stressed right now and that feeling is so amazing.  I feel like the enegizer bunny for once.  My head is clear and I realize that I can have so much more confidence than I think I have.  I can do so many more things that I had though.  I wish we had break earlier because realizing all these things would of really helped me a couple months and well when we get back to school it\'s pretty hard to raise a mark in two weeks, but anything is possible right!??!?  Tonight Melissa and Ben are coming and we are having a big dinner thing, like usual.  And you know what I am not going to worry about what I eat, life it too short to worry about that.  You know I really came to a realization today because I thought that when I lost weight all my problems would be gone and there would be nothing to worry about.  Like yeah right!  There is more to worry about it feels, and it didn\'t solve anything.  I am just going to enjoy my holidays and if I do happen to gain weight then it can always be lost again anyways and it\'s not the end of the world.  I just want to be happy and have fun.  And I am happy right now.   '),(3784,1807720,'23,December,2003','2003-12-23 00:00:00','    Today was sooo busy.  Wow, I feel happy.  I forget about my screwed up world for about two weeks.  My postive attitude is coming back, and hell I feel like I have the confidence to do anything right now.  Yesterday I went to see \"Elf\" with my neighbours and stuff and I think I did something that morning too but I can\'t remember what it was that I did.  Today I went to Melissa\'s house and Doreen was there and took us to Fabric Land and Fanny\'s Fabrics to get some patters cause Melissa is making some runners for the table and wanted some patterns for the window drapes for her and Ben\'s new house they are getting built.  Then we went to North Hill mall and then Market Mall and it took us like 50 stores and 5 times around the mall for me and Melissa to pick out Mom\'s Christmas gift.  There were so many people at the mall too.  And then we had to look for dresses for the Wedding because I am the Bride\'s Maid, but after lunch we were so tired that we didn\'t look long and didn\'t really see anything that was the color we were looking for.  So I went back to their trailer her and Ben are living in and Doreen went back to Canmore, then my Mom picked me up and went back to my house.  After dinner the neighbours came over to visit and we had a ton of food.  Tommorow is going to be a long day too.  Chels, Monet, Sarah, and Mike and coming over for a while and later I am going to see \"Something\'s Gotta Give\" cause it\'s cheap Tuesday at the Coliseum.  There is this song by Britney Spears I was listening to and it\'s soooo good.  It has so much truth to it and the like \"And everytime I try to fly I fall\" just so describes what I felt for like the entire Semester.  But after holidays I think I can totally get it together again.  I just feel so awesome right now.  I am rested and not tired, and just HAPPY.  That\'s right - HAPPY!!      Everytime  You notice me And take my hand So why are we In strangers\' land And love is strong Why carry on without me  And every time I try to fly I fall Without my wings I feel so small I guess I need you baby And every time I see you in my dreams I see your face, it\'s haunting me I guess I need you baby, oh  I make believe that you are here Is the only way that I see clear What have I done You seem to move on easy  And every time I try to fly I fall Without my wings I feel so small I guess I need you baby And every time I see you in my dreams I see your face, it\'s haunting me I guess I need you baby  I may have made it rain Please forgive me My weakness caused you pain And the sun smiles on me  Oh, at night I pray That soon you\'ll face will fade away  And every time I try to fly I fall Without my wings I feel so small I guess I need you baby And every time I see you in my dreams I see your face, it\'s haunting me I guess I need you baby  After all  ~A Song by Britney Spears   '),(3785,1807720,'20,December,2003','2003-12-20 00:00:00','    Yesterday was really fun, alot of good, and a little bit of bad.  The moring was really good and well in Math we just watched a movie and during social we got to watch the play done by the drama 30\'s.  I hope next Semester that the 20\'s get to do a play cause I would really love to do that, but I can always wait until 30.  Then we had Chem and like she told us to do a lab, Paige and I were like the only people that actually did it.  And well I only got a 30/43 on that Chapter 9 test.  The moring was going so good until Chem.  I love Chem though, I really really like it but I just keep trying and trying and I seem to be no good at it.  Our teacher always praises the people who score like really high marks on the test, especially Ashley and now Stephan.  It makes me feel kind of bad and it really bothers me because it feels like the only thing that matters to her is high marks and if your smart or not.  You know I feel like I have worked just as hard if not harder than any of those people, but no one comes up to me and says \"good job Vanessa.\"  I wish the Chem teacher could understand for once, I don\'t think she ever experienced \"not doing good when she put in a lot of effort\" cause she does seem really smart.  But I wish she could understand and just know what it feels like sometimes for things to be very hard and not come easy.  Anyways getting off the topic of Chem cause it\'s starting to upset me now.  After that we had CALM and like she didn\'t even care if we left of stayed so I stayed for half the class.   Then I just left afterward and well got a ride in Sarah\'s car and we went to A&W for lunch and then we went to chapters for a while.  Then we got in line for Lord of the Rings cause we had pre-bought our tickets.  That movie was so good!  I hated the first two movies for it though, but I liked this one because it had alot of personal discovery and reflection within each character, but they had like five different endings for it but it was really good.  Then we drove to Boston Pizza and got dinner and exchanged our secret santa gifts with everyone.  Then I went back to my house and got Mat\'s gift and then went to his house.  We exchanged gifts and he got me like these different scents for your body from the body shop and a whole bunch of Vanilla stuff cause that is my favourite smell.  Then we went outside and went sliding down this hill that was like pure ice.  Lets just say I have LOTS of bruises today and my ribs kind of hurt cause I like got banged into the ice.  I am sure I probably killed a couple brain cells to cause of all the times I banged my head.  Then well Mat\'s asked me if I wanted to go out, but it was kind of awkward.  To be honest I don\'t really want to go out with him, but I just told him I wasn\'t sure.  I don\'t like to upset people or hurt their feelings.  And Chelsey keeps telling me that I should just trying and experience a relationship but honestly I don\'t see Mat\' in that kind of way.  He is more of a friend and nothing more, that is how I really do want it to be.  Like at the dance we touched each other alot, but after a while he would not stop and then like he woulnd\'t let me do what I want to do.  He is a control freak.  He probably gets that from Cadets, and that is anothing thing that like he\'s obsessed with planes and flying and stuff.  It is one thing to be interested in something, but it\'s dufferent when your obsessed.  Also, I always thought to myself that looks don\'t matter, but now I realize that no matter how much I say that, they really do right now.  That is why at the dance I could not kiss him.  I didn\'t want to kiss him.  I don\'t want my first kiss to be with him, I KNOW that and I AM sure of it.  I just don\'t know what to tell him but I know sometime I will have to tell him that I don\'t want to go out.  Then everyone is going to be on my case and telling me to do something different, but whatever.  This time I am not listening to anyone else, not Chels, not anyone around.  I am going to do what I want to do.  And that is the way it is going to be.   '),(3786,1807720,'18,December,2003','2003-12-18 00:00:00','    Today was 1000 times better than yesterday, not to mention that our chem class smelt like a farm yesterday cause of some leak.  And well I did good on my re-write, I got 87 but when it\'s averaged with my old mark it becomes a 75% which still kind of sucks.  And that Titration test I REALLY did my best and I got a 59%.  I do have one more re-write left but if I did my best to understand and do that test then I don\'t think re-writing will do any good.  My chem average dropped again.  The more I seem to try, the more I seem to try and understand, the more my chem mark just keeps dropping.  I hope I did good on the Chapter 9 test we did today though.  I thought the M.C. was really easy but I got a little confused on some of the long answer questions, but I just pray that it didn\'t affect my mark too much.  I can\'t deal with my average dropping AGAIN because I put in alot of effort.    I totally don\'t get the new math unit.  Everyone is like \"it\'s so easy.\"  Why can\'t things be easy for me??  She even said that is was like the easiest unit they saved for last and like finance is NOT my strongest unit, if anything it is my weakest cause I am not good with money related concepts and stuff.  It is NOT easy, although I do wish it were sometimes.  I can\'t wait until math is over.  It\'s stressful.  And well I do like Chem, it\'s very interesting but I have a VERY hard time understanding.  Tommorow is the last day of school before holidays!!  Only one more sleep.  I need a break.  I am stressed out right now.  Although I do have alot of stuff to catch up on over the holidays and I have all that math to try and figure out and understand.  And well I should start reviewing for finals.  Especially my Chem final.  It is going to be like the most important final I have EVER written.  Depending on what I get on it there are two possibilities:  #1:  I make honors #2:  I don\'t make honors.  And if anyone knows how much I deserve to be on it that would be me.  Because I know how hard I try to understand and I KNOW all the effort I put in.  I want sooooo bad to be on honors because it makes me feel good.  Afterall who wouldn\'t feel good because of that?   '),(3787,1807720,'16,December,2003','2003-12-16 00:00:00','    I don\'t even feel like writing in my blog right now but I think I need to unleash some of this stress I am feeling.  Today was so stressful.  I have three tests, the math, the chem, and the chem re-write.  I didn\'t know that I could stuff that much knowledge in my brain at once.  And well I am really worried about how I did on that math test because I just lost my focus and concentration through it.  I wasn\'t scared or nervous or panicked for a math test for once and I think it kind of had a negative effect.  Kind of strange isn\'t it?  And well I am not sure how I did on that chem test.  I am sure the re-write went fine but for the actual test I just totally got confused.  I tried to understand stuff though, I really did try and I do feel really bad.  I always ask her over and over how to do things.  I feel like I I probably can get annoying sometimes, and it does may me feel bad because I never seem to understand alot of things.  She was nice today though, very nice.  She is always nice actually, I KNOW that she really did want to help me understand.  She even asked me how the test went, and like she has NEVER done that.  And then she was like \"some day you will understand.\"  I sure hope some day I understand, but well with math I didn\'t get it ever as a kid and it\'s still hard, so lets just pray that she is right.  I WON\'T give up my dream of teaching Bio no matter how hard I have to work to understand and no matter how things may be for me.   '),(3788,1807720,'14,December,2003','2003-12-14 00:00:00','    I can\'t concentrate right now.  The thoughts in my mind are still stuck on chem.  There is so much to worry about.  I went shopping for Christmas presents this morning and well when I got home I studied for chem for a while but like I said I couldn\'t concentrate so I am going to wait until later tonight when I can focus more.  I went to Britney\'s last night.  We haven\'t seen each other in almost a year and I got my like 4 month be-lated b-day present and we exchanged gifts.  She gave me a $25 gift certificate to South Centre and a $25 gift certificate to A & B Sound so yeah this moring I bought the Britney Spears Cd and I got these two sparkly shirts cause I love things that sparkle.  Ben and Melissa are coming for dinner tonight and I think the last episode of Survivor is on tonight.  I have to study for chem though.  I MUST do good, I NEED to do good, I HAVE to do good.  I hate always trying to hard and stuff and like average keeps dropping.  I want it to go up this time.  I can\'t even remember a time when my chem average actually went up.  I feel VERY stressed out right now just thinking about things.  Next week is going to be so difficult and stressful, but holidays are almost here.  I just have to hold on a little bit longer.   '),(3789,1807720,'12,December,2003','2003-12-12 00:00:00','    I don\'t really feel like writing anything right now but I do have alot of things to get out right now.  There are so many emotions I am feeling right now and so many things that I keep thinking about.  I just had to go onto that U-courses site and check my marks didn\'t I?  And there was a stupid 60% that made my waste half-an-hour crying.  I got a 60% on the chem. quiz we did Wedneday and there is NOTHING else I could have done.  I thought I did good, I can\'t believe I could do that bad.  I seriously thought I got like a 90% and I studied all night and even reviewed at lunch.  Why did I get  60%?????  I TRIED.  I want more than anything to understand and well I did think I understood that quiz.  And well if I thought I understood that quiz and I did bad, well now I am worried about the quiz I did today.  I tried to understand the stuff but I just DON\'T UNDERSTAND.  I can\'t help that fact that I \"don\'t get it.\"  When I started that quiz my mind just went blank.  I couldn\'t remember anything, of course it didn\'t help that I didn\'t get much sleep because of the dance.  But I even got up early to study and practice my chem.  My marks are NOT from lack of effort, and I do ask her ofter how to do things so my marks in chem aren\'t bad because I don\'t ask enough questions because that is not the case.  I don\'t know why I don\'t seem to understand.  It\'s frustrating.  Especially the unit now.  EVERYONE seems to understand the stuff, and I am TRYING to \"get it\" but I ask and ask and ask about how to do stuff.  And yes she explains it to me but it makes no sense.  I feel really bad that I don\'t understand her.  I want to understand.  I want to do good.  I just don\'t know what else to do.  My marks in chem don\'t reflect my effort and this unit seems to come very easily to everyone.  NOTHING ever comes easily for me.  I feel like no one understand me right now, and if I told them how I feel they just would not understand.  No one understand what it is like to wake up everyday knowing how much effort and time I spend trying to understand my stuff and my marks SUCK.  No one understands what it is like believe that you can do something and just when you think you understand and know it you end up failing another quiz.  No one understands how it feels to fail over and over when I try so hard.    No one understands how frustrated I can get with everything and how difficult things are for me.  I want to understand.  I feel so bad.  I feel like the chem teacher is puting in effort to try help me understand and I just never seem to \"get it.\"  I want to do good in chem.  I need to do good in chem.  I want to major in the Science field and I SUCK at chem.  How would that make anyone else feel.  Image if you tried SOOOOOOOO hard at EVERYTHING and your marks never showed your effort and you failed numerous times but you know you tried really hard.  Just imagine that you had a dream of being a science teacher someday yet some concepts from science you couldn\'t grasp from the previous year so in the present year it is really difficult.  So difficult just imagine trying to learn a new language but in a limted amout of time.  That\'s what it feels like - It\'s like chem is the forign language.  Think about how many times I\'ve tried over and over, how many times I have asked \"how do I do this?  How do I do that\" and I still don\'t understand.  Image how many times I have cried because I want to be able to finish my homework and do it correctly but I can\'t finish or do it correctly because I don\'t understand.  Just imagine.  For one day if someone could just picture that.  Image that for just one day then multiply it by 365 and then by how old you are.  Yes, EVERYDAY that is what is feels like.  EVERY SINGLE DAY.  I have lost count of all the nights I have cried because I don\'t understand, all the nights I have cried because I just want to understand, all the nights I have cried because I tried hard and my marks don\'t show it all all, all the nights I cried because I failed again, all the nights I cried because I never seem to understand, all the nights I have wished to be dead because it feels like too much to handle, all those nights that I wanted to drop chem because of the difficult time I am having with learning and understanding it, all those nights I thought about dropping Chem. but I didn\'t because I have a dream.  I will NOT give up my dream.  I want MORE THAN ANYTHING in the world just to be a science teacher and to be happy.  That is all.  So if I can make it through high school and do my best in my Sciences, if I can get into University and get through that, and if I can graduate from that, than that is really all I want.  I want also to be happy but it\'s difficult to feel happy if you are me.  There is so much in this world that I would change if I could.  The first that would be just to make myself UNDERSTAND.  All I want for Christmas is just to UNDERSTAND.  That is all.  The ONLY thing I want right now.  If I could just understand things and make sense of them then I know I could do alot better in a lot of things.  I am so worried about next week.  All the chem tests I have next week.  Just the thought of them stressess me out.  I have *3* tests on Tuesday - THREE.  I have to do the Math test and the Chem test, and the Chem re-write.  That is too much to handle in one day.  I don\'t know how I will remember everything and well I still don\'t understand those ICE tables and since Stoich is confusing for me titration calculations I always seem to mess up one way or another.  Then on Thursay and Friday there is the Chapter 9 test and I am very confused and lost when it comes to Chapter 9.  I thought the quiz was hard and that was just a quiz.  If I thought the quiz was hard then the unit test is probably like three times harder.  I HAVE to do good on those tests.  I don\'t want my average in Chem to drop again.  I am sick of it always dropping.  It seems like the more I try, the more my average just continues to decrease.  It doesn\'t make any sense does it.  That\'s why I am SOOOOO confused about what to do.  I feel like I have done EVERYTHING I could, but nothing is helping out my Chem. mark.  I want to make honours for the end of Semester 1 but my hope is slowing fading.  I deserve it and I know that, but if I don\'t make it I know I will just wonder what all that effort was for.  I guess for my dreams but I AM DOING MY BEST.  I feel though that my marks should be way better, and I think I will always wonder \"why can\'t my marks be better?\"  Like what am I doing wrong???  I just don\'t know and it\'s frustrating.   '),(3790,1807720,'10,December,2003','2003-12-10 00:00:00','    Today was a good day.  I understand my math for the most part and I am pretty sure I did good on that chem. quiz.  And I even got all caught up in religion today.  Today was good.  I am happy today for once.  I feel like I all of a sudden got some confidence back in me today and I really believe that I can do things right now.  Matt and I are going to the dance tommorow.  He is meeting me there though cause he has to go write some important test at Cadet\'s tommorow but then he is missing the rest to go to the dance.  I can\'t wait!  I still have to think of something to get him for Christmas but I have no clue what guys like.   '),(3791,1807720,'09,December,2003','2003-12-09 00:00:00','    Today was better.  I didn\'t cry unlike yestday and well it did seem like a very painful lesson to understand that chem stuff but I had to ask about it because I really didn\'t understand it.  I think I understand the stuff a little bit better now but it\'s still sort of confusing.  I feel bad that I probably like drove her up the wall today as she tried to make me understand.  I thought I would never get it but I think now just maybe I can do the stuff.  It is hard though, and frustrating when it seems like everyone else can do the stuff.  She was like \"oh this is so fun\" and \"you all understand, great!\" and like I didn\'t understand and like it is pretty hard to think something is fun when you get frustrated when everyone else seems to understand the stuff.  I think chem will get better tommorow though.  I know that I can do good on that quiz tommorow.  I just have to re-gain some of my confidence because after that chapter 8 test I was really hard on myself I have to admit because I really tried my very best and I got a 63%.  I am trying not to worry about it too much but it\'s hard because I want to still make honours and I still have two re-writes left so I think I want to use it for that chapter 8 test.  I know that I can do better then a 63%, I KNOW that I have the power to get like 80% and I really feel like I deserve and 80% and I earned an 80%, I just don\'t know why I never seem to get an 80%.  I just pray to God that all my studying will pay off.  Maybe chem can start going good for once, and well if I could understand some more stuff that would be nice.  I always study and stuff but the marks never show it.  I have a 68% average right now in chem and yeah that is what I see when I look online on that U-courses site but in my mind I see an 80%.  I feel like I should have an 80%, I should have an 80%.  I really should but I just want to at least get my chem average to a 70% before the final.       Prayer to God God I ask you to help me get through Chem and help me make honours.  You know how much work I do and how much I study.  You KNOW how much effort I put into everything.  All the effort that goes into trying, and trying to understand.  I pray that tommorow on the quiz I will do good because you truly know how much effort I put into studying for it.  I pray that I score an awesome mark on my re-write and that on those tests next week and the quiz friday in Chem. I understand so I can do good.  I pray for a mark that will reflect my effort for once.  I pray to you to understand my math and just score an amazing mark on the next math test because everyone knows I deserve a good mark, a really really good mark for once and you know it.  I pray for so many things and I do want so many things but if I just understood and had marks that reflected my effort then I would be much happier.   '),(3792,1807720,'08,December,2003','2003-12-08 00:00:00','    Today the lesson is chem was interesting but none the less confusing.  I like it better now though since people are doing it independantly our class is now SMALL, that\'s right a small Chem class [I never thought the day would come]I\'m so confused with chem and I don\'t understand AGAIN.  I am sick of never understanding.  It\'s frustrating.  Chem is really interesting and I really do like things that are related to science but I don\'t understand.  EVERYONE else understood today.  Literally EVERYONE.  I hate myself so much right now, if I had one wish right now it would just be to understand.  I listened and tried and everything today, I don\'t want to not understand it but I just don\'t.  I am sick of always having to ask her how to do things.  I am sick of always needing extra help.  For once in my life why can\'t I just understand.  Lenny even tried to help me with my chem in rel. after chem class and like he didn\'t make any sense.  Eric tried to explain it too.  No one makes any sense!  I NEED to do good in chem, I just want SOOOOOO BADLY to be able to do good.  I have put my heart into chem, being a science teacher is my dream.  I want that so bad.  More than anything else in the world.  But when I see how smart my chem teacher is and all the stuff she knows and understands it makes me wonder how I could ever be as smart as her and know as much, and just understand everything.  How could I ever be a teacher?!?  HOW???  And why don\'t I ever seem to understand things?!?  **FRUSTRATION**  I cannot emphasize that word.  I feel frustrated with just about everything, it\'s annoying.  The only things I am good at is CALM and Religion because those are courses that are based on mostly effort rather than being based on actual knowledge.  This brings me back to my original thoughts over the summer.  Maybe I have some sort of learning problem, but when I ask the people that know me really well they just tell me that I am crazy for thinking that.  No one understands though.  I KNOW I AM smart of course, I know that and I know others around me know I have a smart side to me too, but what they don\'t realize is how difficult it is for me to learn.  I don\'t work or anything and I have cut out EVERTHING extra curricular just to focus on school.  There is NOTHING I can change to make me do a lot better in chem.  I mean EVERYTHING I work my ass off for, and even then that doesn\'t mean I will get a good mark.  I put in COUNTLESS hours trying to understand and doing homework and stuff.  I NEVER seem to understand anything, but I want to, I want to understand.  That is all.  Why can\'t God just make me understand.  I would trade EVERYTHING, EVERYTHING I have in my life and this world just to understand.  ANYTHING.  I would do anything.   '),(3793,1807720,'06,December,2003','2003-12-06 00:00:00','    I can\'t concentrate right now.  I was TRYING to figure out my math.  I wish I could understand.  I was trying to do that trig assignment, the review trig stuff and I can\'t even remember how to do that.  How sad is that, I can\'t even do the review.  I can\'t finish my homework either cause I don\'t understand it.  It\'s frustrating.  I want to do so good on that test monday but it\'s hard to keep the hope that I can do good if I don\'t understand.  I tried out my chem homework today too, those ICE chart things don\'t make any sense.  I think she tried to explain them too fast or something.  I mean she sure knew what she was doing, but we don\'t.  I finished my religion project but now I have to study for dumb social.  I HATE social.  It\'s stupid and pointless and a waste of my life.  I have better things to do other than stuff about the past.  Why the hell do I need to know anything about that past, it\'s not like knowing WWI is really going to make some smart person or something.  I should get back to my social.  I keep avoiding it but it has to be done eventually and well I need to try figure out my math but if I don\'t hand in all my homework on monday I don\'t want marks taken off.  It\'s not that fact that I don\'t want to do my homework, I just don\'t know how to do it.   '),(3794,1807720,'05,December,2003','2003-12-05 00:00:00','    I hate myself right now.  I tried.  I worked hard.  I did EVERYTHING.  I got a stupid 63% on my chem test.  I mean I didn\'t fail or anything but I really really really tried.  It hurts when you know how much effort you put in and I just don\'t get anything out of it.  I can\'t take it.  It hurts and upsets me soooo much.  Thank God I don\'t have a gun right now, I would have pulled the trigger for sure.  Like I went to tutorials, asked her like a million questions, studied like a million nights, did all my homework....EVERYTHING.  Why can\'t I do good?!?  I don\'t get it...I mean what else am I suppose to do??????????  I want to die right now.  I just want to die cause everything I do is never good enough.  Everything I try and so things they just never work out.  How can I suck so much at Chem, WHY?  I dream nights and days of being a science teacher.  How though, just HOW can I do it if I can\'t even do good in high school chem.  All I could do is cry and wonder about the future and wonder about now.  Will things turn out, I can\'t give up my dream right now.  I just can\'t.  If I gave up my dream then I might as well give up my life too.  Some times things just aren\'t fair are they.  In fact, no damn thing is fair and that makes me sooooo mad.   '),(3795,1807720,'04,December,2003','2003-12-04 00:00:00','    Ok so the math teacher actually made an effort to help me today.  She actually like sat down and went through some stuff with me.  It actually helped me to understand some stuff.  She actually made time to help me today and I definetly appriciated that.  There is still lots I don\'t understand though and I am worried about the test because only *I* truly know what I understand and don\'t understand.  The lab in chem was actually kind of fun.  I sort of understood it a little bit better than usual but her directions are really confusing cause I hate when she like scribbles little arrows on the board to different thing and like it gets kind of confusing.  But yeah it was fun, and well I did more than usual this lab.  I actually like did stuff instead of just sitting and watching.  It is actually more interesting when you actually like do that lab.  I sort of understand it a little bit better but there is so stuff I sill don\'t get.  I hope I did good on that chem test on Wednesday.  I will be upset if I did bad.  I have soooo many things to do this weekend.  Tommorow is Friday and usually I am excited but there are soooooooo many things I have to do.  At least I feel better today and my stomach feels alot better.  I didn\'t have any meat or milk though cause those things don\'t agree with my stomach.  I didn\'t eat too much today either.  I lost 4 pounds since yesterday.  That is kind of scary when I think about it, I don\'t really think that is too healthy loosing weight quite that fast.  Oh and my horoscope today was actually really what I think describes \"me\" I thought it was really interesting:    There is really no such thing to you right now as \'try.\' When you set out to do something, you do it, no questions asked. The weight of the responsibilities you have on your shoulders might cause other people to stagger, but you have enough strength to carry on for as long as you\'re asked to do so. Your day isn\'t filled entirely with work-type activities, though. When you do get the chance to play, you play just as hard as you worked. The way you see it, anything worth doing is worth well.   '),(3796,1807720,'03,December,2003','2003-12-03 00:00:00','    I don\'t feel good right now and I didn\'t feel good last night.  I couldn\'t even sleep.  I woke up thinking it was about three and like I looked at the clock and it was 1:45 am.  I keep getting like these sharp pains through my stomach.  I didn\'t even eat breakfast today and for lunch I just had a slushie and well for supper I had some chicken nuggets.  I shouldn\'t have eaten though.  I started to feel better until I ate supper and put some food in my stomach.  I also threw up this morning but I still went to school.  I didn\'t want to miss school because I don\'t need to miss math because I already don\'t understand it as it is and well I didn\'t want to miss the chem test.  I am really annoyed with the math teacher right now.  I am not too pleased.  I admit at times she can really help and be nice but I don\'t understand this certain stuff and I keep telling her I don\'t get it, but like she keeps ignoring me.  I told her today I was like \"I don\'t really understand this\" and she was like \"well it takes to long to explain that so I will explain it later.\"  She never has enough time for anything or anyone.  She doesn\'t get it.  Like I DON\'T understand!!!!  And ignoring the fact that I don\'t \"get it\" isn\'t going to make anything better.   '),(3797,1807720,'02,December,2003','2003-12-02 00:00:00','    The math quiz was hard I have to admit.  Since I didn\'t understand the stuff yesterday it was difficult and well I tried to get some help with the stuff in the resourse room but like every teacher that tries to explain it to me was really confusing.  Tommorow I am going to see how I did on the quiz and hopefully my math teacher will go over more of it.  Chem today was great!  I did good on my quiz.  I got 16/20 which I think is an 80% which is much better than failing like previous times.  I have to study tonight alot but I know that I KNOW this stuff.  I studied soooo much on the weekend, I have to do good.  I believe in myself right now.  I really believe.  I can do it.  The chem teacher was really nice again, ever since I talked to her she has been extremely nice.  I think I ask her way too many qustions though and by the end of the term I will probably have driven her up the wall from \"How do I do this...How do I do that?\"  LOL.  I guess that is what she is there for.   '),(3798,1807720,'01,December,2003','2003-12-01 00:00:00','    Today was actually really good besides the fact that I am very confused in math and well we have the quiz tommorow but I am going to do my best tonight to prepare for tommorow.  The chem teacher was nice again today.  She even like reviewed this stuff that we are doing on the board with me and Thy and like when she was talking about it in class it all make sense because she had explained it to us already.  I think the quiz went really good.  There was one question that I got a little confused by but I did my best to figure it out and I think I will at least get part marks.  I am soooo sick and tired of failing her quizzes and I KNOW that there is no way that I failed this one.  I knew alot, and I know the stuff so yeah I think it went good.  I thought that math teacher was being kind of rude today though.  I mean sometimes she helped me but like I got this sort of vibe from her that like she didn\'t want to help me cause I think she thought I could figure it out of my own.  I tried tonight though and I am very confused!  I couldn\'t even finish the assignment cause I got lost on how to do that stuff and it is on the quiz tommorow.  In the morning I might be able to see if I can find an available math teacher and maybe they can help me.  Ususally my math teacher is really nice though, but today she was just kind of \"off.\"   '),(3799,1807720,'29,January,2004','2004-01-29 00:00:00','    I felt like the queen of zits this morning.  My face was like swormed and omg I have to get rid of em\' before school starts again cause I hate wearing make-up to non-speacial-occasions but if I have to hide my face then I will!  I am breaking out more now on this break then when we had school.  I assume it\'s from Mat that is really stressing me out.  I feel so good today though.  My sister came and we watched American Wedding which was hilarious and then like I got my hair done.  It\'s really dark with blonde and red going through it.  It\'s soooo awesome - I love it!!!   '),(3800,1807720,'28,January,2004','2004-01-28 00:00:00','    Things have been crazy.  Last night I had this big argument with Mat on msn that went to 1:00 in the morning, and thank goodness it wasn\'t in person cause I am sure someone would have been hurt.  I can\'t talk to him, or look at him, or anything.  I just can\'t.  After we finally finished our conversation/arguement - which was 28 pages, I just sat alone in my basement and cried as tears poured down my eyes.  I just wanted a HAPPY week off, no tears, no being upset, just a happy week, but NOOOO he had no ruin me just wanting to be happy.  Is it such a crime to want to be happy???  There\'s no chem, no math, but now there are different problems to deal with.  If it\'s not one thing, it\'s always another.  I went to BP for lunch today with Ash and Jesse cause I had to get out because I was feeling really depressed.  After crying during the night I thought everything would be ok in the morning, but that wasn\'t exactly the case.  I feel alot better now that I listened to my Britney Spears Cd lol, and I went to lunch, and Ash and Jesse came over and we played x-box and air hockey.  I am going to try forgot all the shit mat said about me, cause you know what I am better than all that sh*t that mat called me and said I was.  I am better than people who are so low, they have to insult people, make them feel bad, and put em\' down.  I am getting my hair done tommorow!!!  I can\'t wait!  Change, good change for once.  I haven\'t had my hair done since August, so YAY!!!  And when I get back to school, I will NOT talk to mat, I can\'t talk to an asshole like that.  I don\'t think I can even look at him, I don\'t want to try and find out.  He thinks he is so much better, well I will show him.  I am going to show all those people who constantly point out flaws and faults and everything that seems to be wrong with me to them.  I am going to show them that I don\'t need to change anything about me.  Cause I am who I am, that is the way it is.  I am me, that is all I can be.   '),(3801,1807720,'26,January,2004','2004-01-26 00:00:00','    When my sis came yesterday we went shopping for a dress for the wedding and I finally have a dress now!  It\'s the theme color of the wedding of course, that light purple, but a little shade off cause we both agreed that I don\'t want to match every flower and table piece.  And I even got shoes that I can walk down the eisle without tripping!  They are soooo awesome, and I love them so much I will cry if anything happens to them.  We destroyed like 5 million stores and tried on about 20 thousands pairs of shoes.  Finally, my sister found these ones that were made of ight blue leather and like, with a nice heel that isn\'t too high and I wont trip, and one the side of my small toes there is this little butterfly thing.  It\'s so classy, so sexy, and those shoes so say \"ME.\"  Kind of ironic too because the meaning of my name is butterfly.  When I saw them I just said \"those are the ones\" before I tried em\' on, and when I put them on they fit perfectly.  Today I am not really doing anything much.  Relaxing I guess.  I worked hard for exams, and now is my time to relax.  I am also working on my audtion for the play.  I have almost got my monologue down, it\'s going to be great.  And even f I don\'t get a part in the play (cross my fingers and work hard and hope I do) then I am sure I wil be busy enough with drama as it is.  Megz and I were talking to the drama teacher and he said that he thinks we might get to do \"travelling drama\" where we would go to other schools and show them the scene work that we did, or our script work maybe.  I sure hope we get to do that.  Semester 2 is definetly going to be time consuming and hard work.  I think after doing chem though, I feel like I can do anything.  I remember all the times I cried over chem, all the times I wanted to drop Chem.  I was actually reading only posts of mine in my blog and just about everyone second post mentions Chem. lol.  And well I have to go to dentist later, (dentists are evil) but well yeah I haven\'t been in a year, so at least my teeth will be white instead of yellow.  Something to look forward too.  lol.  I actually really badly want it to be semester 2 right now!  I think I love school a little too much, but yeah I am soooo excited for Bio too, and Drama, Drama, DRAMA!!   '),(3802,1807720,'25,January,2004','2004-01-25 00:00:00','    Yesterday night was fun.  Chels, Edi, Andre, and myself all went to glo-bowl.  Then we went back to Edi\'s house where we stayed there until about 1:00 AM and then he drove us all home.  Yup that was my night.  Chels sure flirted alot with Edi though for having a boyfriend, but whatever it\'s really none of my business.  So lets just say that when everything goes wrong and everything breaks - that\'s my house!  Ok so we got a new washer cause our old one sucked ass and well my dad flew in this weekend and he was setting it up and forgot to remove something so like for half an hour it sounded like the house was going to blow up.  It was actually really funny.  Then in the evening we had water pouring into the basement by the comp from like the ceiling.  That\'s when we discovered that the diswasher was leaking from the bottom so now they have to buy a new dishwasher cause it\'s already been fixed like 500 000 000 times, we could have had like ten new dishwashers by now.  Then my printer wouldn\'t work again and we just bought a new one, so there is a problem with the connection between the comp and the printer.  So now we have to buy cord thing that connects the two.  So yeah basically my house is falling apart, not like that is anything new though.  Melissa and Ben are coming later for dinner.  It was Ben\'s birthday on tuesday, so we invited them for dinner tonight.  We aren\'t having any cake though because we are all trying to loose weight for the wedding, so no cake.  We will probably just have fruit for dessert, it\'s way better anyway.   ');
INSERT INTO `msgs` VALUES (3803,1807720,'24,January,2004','2004-01-24 00:00:00','    So I spent all day Thursday at school preparing for the math exam.  I worked through like a million problems with the resource room lady and like went over my finance test with my math teacher.  Looks like practicing for that math test paid off.  I got a 71%!!!  YAY!  That was my goal, just score over 70%.  Last night all I could think about was  YES, I DID GOOD ON MATH , and when I woke up this morning all I could think of was: 71...71...71...I GOT a 71% in math!!!!!  Although I wish I could of done a little better on the Chem final I was looking at my marks and she changed one of my labs that I failed to 100% and it said:  Mark changed to reflect the participation in Bonus Lab.  I just hope my Chem average will be over 70%.  I sure hope.  I know I deserve it.  After the math final yesterday Megan, Erin, and Myself went to Moxie\'s for lunch and then after Erin left we shared one of those white chocolate brownie desserts: It was after final reward.  LOL.  And then we had to go back to the school so Megz could finish a social test she didn\'t write, and then we went back to my house.  We both choose our monologues for our audtions for the play and we are soooooo excited for Drama.  Can\'t wait!  Then my mom made us dinner but we were so full from lunch that we barely ate but by midnight we were pretty hungry but just waited til\' the morning to eat breakfast.  And we rented a Midsummer Nights Dream so we watched that, then we watched legally blonde, and on TV we watched Charmed, Jaon of Arcadia, and Third Watch, and then half of My Big Fat Greek Wedding until we we so tired that our eyes hurt from keeping them open for so long.  I had such a fun time.  I didn\'t sleep much last night for some reason.  It might take a couple days for the stress I have built up over the past week to wear off.  Either it\'s the stress or maybe it\'s cause I kept thinking about math last night.  I still can\'t figure out how I did better on the math, than the Chem.  I thought the long answer on the math final was soooo easy but like the M.C. part was challenging.  And like on the long answer there was this question that I didn\'t understand at first and I just remembered what she said:  If you get to a question you don\'t know then skip it and do it last.  When I skipped the question I didn\'t understand, I didn\'t loose my confidence through the test.  So this question was stuff during the year I really struggled with, but when I went back I just said to myself: Ok Vanessa you can do this.  I just sat there and I did the question right!!!!  Since it was like you have to set up two equations, and then solve for like what x or y represented I just plugged the values I got for x and y back into one of my two formulas and it gave me what it was suppose to equal.  I am so proud of myself, because all year I struggled with doing those kind of questions, and finally I got it right.      Life\'s Tug Of War  Life can seem ungrateful and not always kind. Life can pull at your heartstrings and play with your mind. Life can be blissful and happy and free. Life can put beauty in the things that you see. Life can place challenges right at your feet. Life can make good of the hardships that we meet. Life can overwhelm you and make your head spin. Life can reward those determined to win. Life can be hurtful and not always fair. Life can surround you with people who care. Life clearly does offer its ups and its downs. Life\'s days can bring you both smiles and frowns. Life teaches us to take the good with the bad. Life is a mixture of happy and sad.  So...  Take the life that you have and give it your best. Think positive, be happy let God do the rest. Take the challenges that life has laid at your feet. Take pride and be thankful for each one you meet. To yourself give forgiveness if you stumble and fall. Take each day that is dealt you and give it your all. Take the love that you\'re given and return it with care. Have faith that when needed it will always be there. Take time to find the beauty in the things that you see. Take life\'s simple pleasures let them set your heart free. The idea here is simply to even the score. As you are met and faced with Life\'s Tug Of War.  ~Unknown Author       ~Yourself~  What does your name mean?: Butterfly How old are you?: 16 How old are you mentally (as in are you mature?)?: I would say 18: Others might think differently Describe yourself in 5 words.: Frustrated, Stressed, Hard Worker, Dreamer What are your worst qualities: I bottle up emotions during the day and unleash them at home What are your best qualities: My work ethic How long does it take you get get ready in the morning: About 45 mintues    ~Sleep~   Do you dream at night?: Yes Do you remember your dreams?: Sometimes Describe one: One night I dreamt that I was in a pitch black room where I could see no one but I could hear voices calling out: Vanessa, Vanessa, over and over and it kept getting louder. What time do you go to bed usually?: 10 - 10:30 on week days and whenever I want on weekends What time do you wake up normally?: 6:45 AM What time do you wake on weekends?: Whenever I want Do you find waking late nice or annoying?: Annoying Do you sleep with one pillow or two?: 1   ~School~   Do you like school?: Yeah, of course Why/why not?: It\'s like a second home, and I want to be a teacher Whats ur fave subject?: Ok I confess, Chemistry Most hated subject?: Social Do you have a fave teacher?: Yes, two - My Science teacher from last year and my Chem teacher Ever had a crush on a teacher?: Yeah, who hasn\'t? Are you a maths/science person or an english/drama person?: I love Science and I LOVE drama   ~Friends~   Do you have heaps of friends?: I suppose, I talk to everyone Do you have a best friend?: Of course Do you have more guy friends or more girl friends?: More guy friends, but alot of girls too Do you ever get annoyed at any friend?: Yeah, all the time Have you ever lied to a friend?: Yeah, but only to protect someone else Have you ever stolen a friends boyfriend/girlfriend?: Nope, not desperate   ~Family~   Do you like your parents?: Sometimes  Ever run away from home?: No Ever thought about it?: Alot Do you have any siblings?: Yes, a sister  If so, do you like or get annoyed with them?: Nope How old are they?: 20 Do you feel your parents spoil you?: Always Do you not get along with any of your family?: Yeah, my cousins act like they are from Mars Do you have big family get togethers ever?: Sometimes   ~Relationships~   Do you have a boyfriend/girlfriend?: Apparently, but it doesn\'t seem like it If so, are you in love with them?: No comment Do they love you?: Yes How long have you been together?: Just over a month Most romantic thing theyve ever done for you?: Helped me with a social position paper (if that\'s romantic) Do you have a crush?: Of course If so, are you in love?: You can\'t call that love Do they know you like them?: yeah, I think so Is it serious or playful?: More serious How long have you liked them?: Forever Ever done something stupid to impress them?: Nope, don\'t need to impress Have you ever experienced unrequited love?: I have no idea what that means Do u find it romantic or hurtful?: Â¿?Â¿?Â¿... Even know what it is?: Nope   ~Sex~   Ever had sex?: No Believe that a person shouldnt have sex before marriage?: Yes Believe in casual sex?: No way When do you plan/when did you lose your virginity?:  When I get married Did you regret it?: I have nothing to regret as of now   ~Religion~   Do you have a religion?: Yes  Do you practice it i.e go to church?: No Do you believe in God?: Yes  Jesus?: Yeah  Satan?: Yeah Heaven?: Yes  Hell?: Yeah  If you died tomorrow what do you beleive will happen to you?: I\'d go to heaven Does death scare you?: A little bit   ~Morals~   Have you ever been drunk?: No yet, but going soon to a party with U of C boys Taken drugs?: NO WAY! Stolen?: No Shoplifted?: No  Tried to commit suicide?: No comment Lied to a boyfriend or girlfriend?: Always Gotten into a fight?: Of course Are you more innocent or guilty?: Innocent Would you date a drug addict?: NO WAY Have you ever had to look after someone who was a drug addict?: No Are you racist?: No Are you discriminatory to anyone?: NO Have you been a hypocrite in the past?: No Do you have an open or closed mind to other peoples beliefs and feelings?: Very open   ~Media~   Do you watch tons of tv?: When I find the time How many times have you been to the movies in the past 6 months?: Too many to count Do you listen to the radio often?: Never, except when someone else is listening to it Do you read the newspaper?: Yes  Do you read magazines?: Yes  Are you a couch potato?: Not really Do you use the internet too much?: Yeah, obviously if I am writing this right now, but I am bored   ~Music~   Whats your fave style of music?: Pop Do you play an instrument?: Yep, flute Do you sing?: Yes, all the time  Whats your fave band?: Live on Release and Britney Spears is my fav. singer Why?: They CAN sing and L.O.R. play all their own instruments Have you met them before?: No, but that would be cool Name 3 cds that youve bought in that last year: Britney Spears, Clay Aiken, Ryan Malcolm Why did you buy them?: Britney\'s voice is cool, Clay is cute, and Ryan is the Canadian Idol   ~Sports~   Whats your fave sport?: Lacrosse and Soccer Whats your fave sport to watch?: Football Do you have a fave team of any sort?: Nope Do you play a lot of sports?: All the time Ever won anything for sport?: I have 14 medals from soccer and my brown belt in martial arts   ~Personality~   Are you funny or serious?: Both, at times Creative or not?: I can be VERY creative  Logical thinker or lateral thinker?: I don\'t really know, and am not really sure Are you outgoing or shy?: Both, depending on who I am talking to, but mostly outgoing Are you lazy or active?: Active most of the time Have you ever been hyperactive?: Don\'t think so Are you a naturally hyperactive person?: Nope   ~Looks~   Are you happy with the way you look?: Yeah, I suppose What would you change?: I would have non-frizzy VERY straight hair and I would be thinner Do you wear makeup regularly?: Nope, only on special occasions Do you have a large wardrobe?: I would say it\'s average, besides jeans like 365 days are the way to go   ~Money~   Do you have a job?: Use to, but I quit Do you like it?: I hated it and it slowly eroded my happiness, which is why I quit Are you a saver or a spender?: Both  Do you work hard or slack off?: Work hard, always work hard Have you ever been fired?: No  In trouble at work?: Not from a manager, just a disagreement w/ another employee Made a major mistake?: Nothing major  Ever had money stolen from you?:  Yes Are you always broke?: Nope, I save alot   ~Embarassing Moments~   Your all time most embarassing moment?: That\'s for me to know, and no one else to never find out Ever snorted drink out your nose?: Actually, yeah believe it or not Ever giggled like an idiot?: Always Ever embarassed yourself and pretedned nothing happened?: Yep Ever tripped in front of someone you liked?: I dont think Ever said something really stupid?: You bet Ever snorted while laughing?: Yup Ever fallen off a bed?: Too many times, I always roll off  Ever sleepwalked?: Nope Ever sleeptalked?: I hope not   ~Memories~   Whats your best memory?: Making honor roll, it\'s the best Worst?: Grade 6 Whats the wierdest memory you have?: I have no idea Do you have a good memory?: I would say so Whats the coolest holiday you remember having?: Summer is always the best Have any regrets?: Nope, no regrets   ~Thoughts~   Ever had funny thoughts and laughed and no one understood you?: All the time Whats the first thought that comes to your head when you hear these names?  Melissa - My sister Kathleen- The girl from drama last year Vanessa - Me! Alyssa- This girl I met from my friends youth group Jess - Another girl from drama last year Brian - Annoying Charlie - Brown Olivia - Newton John  Drew - Andrew  Lily - Easter  Anita - Slut  Whats the firs thtoughts that come to your head while hearing these words?  Cheese - Yellow  Rubber - Tire Clothes - Style Big - Giant Dress - black Jacket - Cold  polyester - Chemistry Washing - Hands  Ok that was fun, and consumed my time off.  I might as well do what I want right cause it\'s not like there is school on Monday.....I get a wHOLE WEEK of break.  HOLIDAY!!  And yeah going out tonight with Chels and Eddie and some other guys from the universtiy.  I\'m so excited.  Those guys sure know how to party and we ALWAYS have the most fun and the best times with them.  The roads are kind of bad today though because of all the snow, but I am sure everything will be fine.   '),(3804,1807720,'21,January,2004','2004-01-21 00:00:00','    It sure is amazing how one day can seem like the greatest day ever, and others just totally suck.  Today is one of those bad days, one of those days that totally suck - except for this morning before the final when the Chem teacher said \"Hey Vanessa.\"  That was the only good thing about today.  I didn\'t think the chem final was that bad, but well they didn\'t give us enough time.  I would have liked more time to go back through my answers and go through the questions that confused me.  I felt rushed through half of the test.  I went online - big mistake - and wanted to see if the chem marks were posted and there was my final mark - 68%.  First of all, I KNOW that I DON\'T deserve a 68% and secondly why does my effort I put in never pay off?????  WHY?  I feel like I try so hard.  I try, and I try, and I try.  Over and over again I put the effort in, more and more effort, so much effort that I don\'t think I could do anything more or try harder.  I wanted so bad to do good on the final, I wanted to do good more than alot of things that I have ever wanted in my life.  I TRIED SOOOOOOOOO HARD to get a good mark on it.  Look what happened.  68%.  I suppose it could have been worse, but still that isn\'t the point.  I can\'t say I didn\'t try my best though, because I did.   I TRIED MY VERY BEST .  But I feel bad.  I have a pile of kleenex now after crying so much.  I don\'t deserve 68%, I just don\'t.  And to make things worse I somehow only got a dumb 68% on the social final too.  My m.c. wasn\'t that bad, and I thought that my position paper was 80%-90% work that I did.  Then online is says my mark is 68%.  I am so confused, I just kept staring at the mark wondering how I could have done so bad.  I thought that I would get at least an 80%.  Of course for both Chem and Social I do feel like I deserve an 80% but do I have an 80%...NO.  I certainly worked hard enough to deserve an 80%.  So then why don\'t I have an 80%?  I feel stupid at the moment.  I mean I don\'t mind not being \"naturally smart\" like Ashley P. but it would be nice if my marks actually reflected a little bit of effort for once.  I just pray that I make honors.  I can\'t deal with not making it.  I am so scared of the math final, and that is going to be a real challenge for me but I have to make honors.  And well going back the whole chem thing, it confused me alot because I want SOOOOO BADLY to be some sort of Bio or Chem teacher when I am older.  But just how could I be???  If I can\'t do the stuff now, how will I be able to do it for university.  That is what I want sooooo much though.  I WANT IT MORE THAN ANYTHING.  It\'s my dream, to be a high school teacher is my dream.  And the more effort I put in, and the more I try just leaves me wondering \"Maybe effort and trying isn\'t enough,\"  I need to actually be good at this stuff, but in reality I\'m just not.  I am not good at anything.  Hmmm Vanessa - name one thing your good at?!?...Funny isn\'t it, because NOTHING comes to my mind.  This means I have no talent I guess.  I feel like everyone has a place in this world, but where do I come in.  I want to be happy, I want everything to alright.  But I am not happy right now, and I don\'t feel alright.  I don\'t feel like everything will be alright either.  Finals are so easy for 99.9% of the population, and you know who falls into the remaining percent that don\'t find finals easy - ME.  Why does it comes so easily to lots of people, why is it that when other work hard it pays off for them, but not for me?  Just why??  What is wrong with me???  Why do I feel so different from so many people, but so similar to others??  I can\'t wait til\' all this final stuff is over.  I can\'t deal with the stress anymore.  I\'m glad I made plans with Megan on friday after the math final, because at least I have something to look forward to.  I don\'t even know what I feel anymore.  Of course upset cause of my chem mark, actually really upset, but I guess a little happy too for some reason.  There are so many things I am so confused about.  So many things I wonder about, so many questions, and no answers.   '),(3805,1807720,'20,January,2004','2004-01-20 00:00:00','    Matt came over this morning and then we ate breakfast.  Then we did the social test and well it was pretty fun and got the same score on the M.C. and then of course we each wrote our own position paper for the test.  Then I cooked some pizza - the stuff from the box.  Remind me never to eat that stuff from a box again.  The calorie and fat content in that thing was disgusting and I am sure not the most healthiest.  Then we watched some TV - comedy - which was absolutely hilarious.  It was nice to have a good laugh.  I wanted to write in my blog right now for some reason.  I keep avoiding studying right now because I studied so much yesterday.  My head was pounding when I got home yesterday but after I finish writing it\'s going to be back to studying chem.  I have all the stuff down, and the only thing that really needs some extra practice now is re-arranging the formula.  I sure hope everything goes well.  After the chem final tommorow I am going to do math review because I was going to do some today, but I need to know chem.  Chem comes first in my mind.  I want so bad to make honors though.  It literally means the world to me because of all my effort.  I don\'t want to screw this up, and I am not going to.  I am going to do good.  Anyways I have to study some more Chem!!   '),(3806,1807720,'19,January,2004','2004-01-19 00:00:00','    Today was great.  I am soooo tired right now though.  I got to school around 8:30 and well then around 9:30 did some Chem review for finals and then went to the quiet room in the public library to study more Chem!  Then Chels and I went to wendys and got taco salads then went to Chem in the afternoon.  That was my day!  Chem!  And well the Chem teacher was listening to music and like either I was imagining it, or she was like actually humming to the music.  I had to hold back my laughter cause it was hilarious.  I am like so into chem now and find it really interesting, and I don\'t want to admit it but I find studying chem fun!  And then chem teacher was like so incredibly nice.  She even told me that I had an awesome work ethic.  That actually made me feel soooooo good.  I always feel like no one realizes how hard I work at things and for things, and for someone to tell me that I have a really good work ethic, well that is an amazing feeling.  I feel alive for once, I feel like I am actually living.  I feel good, awesome, so many things.  I feel like I am on top of the world.  I feel like I can do anything I put my mind to.  I feel like things might just start to go my way a little, and I feel like everything will be ok.   '),(3807,1807720,'18,January,2004','2004-01-18 00:00:00','    Inspiration  I woke one morning Then I opened my eyes A flash before my face this feeling that I can\'t escape A memory that I can\'t erase I opened my eyes Then it was a new day  She was a inpsiration to me Was everything I wanted to be When I look in her eyes That is when I realize She is everything I have wanted to be Just like an inspiration to me  As I walk down the halls Passing through I can see This is how to live, a life so free Then I stumble upon the ground And I fall, I fall, I fall  Never thought that someone could see All the pain and the hurt inside of me But she always asked me about one class When she saw me it was always of chem she asked  Her career is my dream life as you see I could wish days ands night just to be So special, so rare, so full of care Someone just like her  She was a inpsiration to me Was everything I wanted to be When I look in her eyes That is when I realize She is everything I have wanted to be Just like an inspiration to me  I woke one morning Then I opened my eyes A flash before my face this feeling that I can\'t escape A memory that I can\'t erase I opened my eyes Then it was a new day  ~Vanessa   '),(3808,1807720,'17,January,2004','2004-01-17 00:00:00','    I am going to be a Chem expert by the end of the weekend.  I was reviewing my chem stuff and like I understand sooooo much stuff in the course now that I have like gone through everything we learned.  It\'s amazing!  I have a good feeling about the chem final because I really feel like I am REALLY REALLY REALLY getting this stuff.  of course ICE tables still continue to confuse me, but I partially understand them and well not understand the ICE table isn\'t exactly going to kill my mark.  I feel so good right now - but I am taking a little break right now because my brain will be fried if I don\'t take study breaks. feel so great and I am happy! :)  \"Believe in yourself and you will succeed!\"   '),(3809,1807720,'16,January,2004','2004-01-16 00:00:00','    Today was good.  Chem just made my day.  I thought I had failed that test and I found out I got 71% which is pretty good from me in Chem.  I am really hoping that my average will go up before the final because it would really help me alot.  Besides I worked hard enough that I really know that I deserve my average to go up.  I am so glad finals will be over soon.  I just want to get them over with, the thought stresses me out but I am going to put EVERYTHING into studying for Chem.  And well actually I already started studying for Chem and did some math review but I am not going to worry too much about math until after I am done with the Chem and Social finals.  After school ended, Sarah, Me, Chels, Mat, Alex, and Lenny all went to Swiss Chalet for lunch and we got soup with chicken and fries, and then pie afterwards.  It was soooo good.  I think starting tommorow though I am going to try eating healthy again, if I could loose 40 pounds once, then the last 20 I want to loose I think is definetly possible.  So yeah - Eat healthy tommorow Vanessa!  Oh and I confess.  I like the Chem teacher.  I know I always complain about chem, and at the beginning of the year I really didn\'t like the way she teaches.  But you know what, I really think she is really nice and actually really cool.  Although sometimes hard to understand, I really feel like after that day we had a \"talk\" she has put alot of effort into trying to help me understand and after \"talk\" I haven\'t heard her say anything rude to me or anything that would make me feel bad.  She is like different from ANY other teacher that I have EVER had, and I think she has the most coolest and most diverse wardrobe on the face of the planet cause her style is outta this world.  I would kill to like dress like her - I think her fashion style is really cool.  And she owns like 20 million types of high heel shoes.  Especially the ones she wore today when you could see skin, they were classy and I sooooo wanted them. But Mrs. J will always be my most favourite teacher in the entire world and will ALWAYS be number 1 on the list of my favourite teachers, but the chem teacher comes in second place, which in my books is extremely high.   '),(3810,1807720,'15,January,2004','2004-01-15 00:00:00','    Today was kinda bad.  Well chem was bad.  I sat there during most of the lab and well I thought that she thought I like was lazy or I just didn\'t want to do the lab.  The truth is that I just don\'t understand her directions.  Everytime she explains something it is so confusing.  It\'s not that I don\'t want to do the lab, I just never understand the lab.  I never found out what I got on that unit test either, because there was no time, but I really think I did bad and I keep thinking about it.  It\'s like ALWAYS on my mind.  Social was good today and I think Ali, me, Megz, Erin, Elise, Grant, Kenda, and another guy are going to do Model UN.  Anyways yeah I guess was ok, but Chem was bad.  I thought her demo was interesting, but I felt bad cause I never do anything during labs because I never understand.   '),(3811,1807720,'14,January,2004','2004-01-14 00:00:00','    Today was horrible.  It wasn\'t too bad until that chem test.  I got 72% on the last quiz, but there were just dumb mistakes that I know now how to do correctly.  That chem test today was so stupid.  Half of the stuff we didn\'t even learn or talk about.  I studied soooooooo hard for that and I am pretty sure I failed.  It upsets me so much.  Its so discouraging when I try and I try and I try and alot of times I fail and I fail and I fail.  It\'s like the more I try the more I fail.  It doesn\'t make anysense.  If I do fail that test like I predict then there is DEFINETLY something wrong because I ASKED A MILLION QUESTIONS and I STUDIED MORE HOURS THAN I CAN COUNT ON TWO HANDS and I KNOW FOR A FACT I PUT MY VERY  *BEST* effort forward.  I don\'t know what to do anymore.  I thought this was the chance to raise my mark, but I think that I was wrong.  I just feel so bad, and I think I really scared my friends or something.  They even said \"don\'t do anything rad...no knifes or razors.  I must have been more upset at school then I remember right now.  But I am really upset.  But I made plans with Megz after the Math final and she is going to sleep over that night too and we are going to prepare our monologues for our the auditions for the play.  I can\'t wait!  Once I get through finals I get to have fun, but right now I am very concerned about my chem mark.   '),(3812,1807720,'13,January,2004','2004-01-13 00:00:00','    Today was good, actually it was great!  I feel so good right now.  I think I did good on the chem quiz but I am not going to get too excited because you never know.  I hate thinking I aced something and then I find out that I didn\'t.  Today was really good though.  I finished social and religion and I STUDIED alot for chem.  I feel real ready for the test tommorow.  I just know I know this stuff.  I studied and practiced too much to do bad on this.  I KNOW in my mind that I can do good on this.   '),(3813,1807720,'12,January,2004','2004-01-12 00:00:00','    Today was alright.  The chem teacher was really nice.  The math teacher kept getting mad at us for talking - but hey we are humans and yeah we like to talk.  In social we watched a movie and in CALM we had a sub.  Interesting day wasn\'t it?!?  I have another chem quiz tommorow but I feel confident about this one.  Now that I have cried and whinned and complaned about not understanding all my frustration has led me to somehow understand this stuff.  It feels sooooo good to understand and I 84% on the quiz I thought I was going to fail.  I thought I didn\'t understand, but obviously I do and now I feel like I understandd too!  I just have to keep this up now.  This is my chance to bring up my chem mark.  This is my opportunity.  It\'s the last week before finals and well I know for sure that I can raise my mark for sure.  I have to make honors.  I worked too hard not to make it.  I have to do good and I really believe that I can do good.   '),(3814,1807720,'10,January,2004','2004-01-10 00:00:00','    I have never been so happy and so sad in one day.  Everytime something good happens, something bad happens that always outweights the good things in life.  Again I sit hear crying, thinking to myself that I want to be dead.  Wishing that there were someone here right now, but there isn\'t right now, and I just continue to cry in the silence of my basement.  I did my really stressful Rel. project today and then I told my mom that we were going out rollerskating and stuff, and that I would be going in the evening.  So Chels friends that she works with at Wendy\'s picked me up and we went to Eddie\'s house while he had a shower after work.  Then we played pool for a while until his brother got home.  Then we went off in Eddie\'s car to Perkins (Which is where Chels and I used to work) and we all got milkshakes with a Cheeseburger and fries.  It was soooo good.  Then we went to Lloyds for a couple of hours.  It was soooooo fun.  And OMG this Eddie guy is like my perfect match.  I never thought he would be sooo awesome.  And well he is a University student at the U of C and is studying to become a electrical engineer.  That\'s kind of ironic because my sisters fiance has EXACTLY the same profession.  And Eddie looks ALOT like my sisters fiance too.  When me and him held hands it was actually really really really natural unlike alot of other people.  He is soooooo fun, but since we had to drop some people off on the way back we got home around like 12:15 PM but like that isn\'t that late.  Then my mom was like questioning me when I got home and no offence but:  WHO THE FUCK WANTS TO BE QUESTIONED ABOUT STUFF WHEN THEY GET HOME AND ARE TIRED...hmmmmm....NO ONE!  And then she was like \"Oh I was worried, blah, blah, blah, and you never called\" and then we got into a huge arguement.  I told her where I was going, and was like \"well you never said what time you would be home\" and like she NEVER asked me what time I would be home so why the hell would I tell her if she didn\'t ask.  If she wanted to know so bad she should of asked.  Besides if she can\'t let go now, she never will be able to.  And then after our argument I told her I would talk to her in the morning and I went downstairs.  Then like half a minute later she followed me downstairs.  What the hell doesn\'t she understand about: I WILL TELL YOU IN THE MORNING, I WANT TO BE ALONE.  All she said is \"I want to talk\" and like she wanted to know about what stuff we did.  But I told her I would tell her in the morning.  I am sooo tired right now and have been talking and doing alot of stuff tonight, she doesn\'t understand that I DON\'T WANT TO TALK RIGHT NOW, ACTUALLY MEANS THAT I DON\'T WANT TO TALK.  She doesn\'t understand right now.  Besides I had such a good and fun time, and then I got home and she ruined everything.  Everything just started to turn around and I was feeling alot happier but now she has just put me is some depressed mood and I want to kill myself at the moment.  But that is nothing new.  There is so much to think about, so much to worry about, that honestly I don\'t need anymore fucking stress.  But obviously I am VERY stressed out right now.  Finals are coming up and those are SOOOOOO important and well I am TRYING to be happy, but that\'s an extremely hard task at the moment.   '),(3815,1807720,'08,January,2004','2004-01-08 00:00:00','    My life sucks right now.  I hate myself at the moment.  I don\'t think I have ever wished to be dead so many times in a night.  It\'s all because of the Chem stuff we are doing now.  I just can\'t get the hang of it.  I want sooooo badly in the world to do good, and well with the quiz tommorow and me not understanding how to do LOTS of the stuff scares me.  I WANT to understand, and I am TRYING to get it, but NO ONE understands how difficult it is for me.  I wonder sometimes if anyone has as much trouble learning everything as I do.  I still don\'t understand my Chem though and well I was studying and practicing and trying to \"get it\" but tears just kept pouring out of me for a good hour when I was trying to get my chem.  I DON\'T need to fail again, that is the last thing I need right now.  It\'s starting to set in though that it could be a reality.  If my mark drops again in Chem before the final and if I do bad on the final then it probably is possible to fail.  I CAN\'T fail chem.  If I fail my life is literally over.  The rest of my life will be ruined because my dreams for the future all depends on if I pass Chem or not.  I honestly don\'t want to harm myself or kill myself, but sometimes it feels like a better way out.  To just get away from everything.  No one understands how difficult this stuff is for me - I keep trying to do the stuff and like I just can\'t seem to understand it and I don\'t know how to practice something I don\'t understand.  The chem teacher doesn\'t understand, no one understands.  I feel horrible right now and I am so scared for that quiz tommorow.  I just feel like I put in sooooooooooo much effort into trying to learn the stuff and it feels like all my effort is just going down a drain.  Afterall it\'s not the effort someone see\'s, it a mark.  When it comes down to it, what defines me is my stupid marks.  University isn\'t going to give a fucking piece of shit if someone tried hard.  All that matters is whether you do good or not.  I am so worried.  I want to get into the Biological Sciences eventaully but if my marks in my sciences are weak they probably won\'t even take a second look at me.  I wonder if this is what all 16 year olds think about or feel like.  I wish I knew, and I wish I had answers.   '),(3816,1807720,'07,January,2004','2004-01-07 00:00:00','    Today was AWESOME!!!  I mean I think I did bad on the Math test, but other then that silght imperfection in my day it was really really REALLY good.  Chem was sooooo interesting today and she did a really good lesson today.  I am using my other re-write tommorow too because I KNOW that I can do soooo much better than the 69% I got on the test.  I am soooo happy right now and well worried about Math of course, but it\'s not the end of the world.   '),(3817,1807720,'06,January,2004','2004-01-06 00:00:00','    Things seem to be going pretty good.  I feel happy right now.  Although I think I lost my math text book and my CALCULATOR!!. AND I HAVE A MATH TEST TOMMOROW.  I NEED my graphing calculator for tommorow because a normal calculator can\'t calc. any mortgage questions. But hopefully Lawrence just accidently took it or something, or maybe I just forgot it in her room or something.  I HOPE!  And well tonight to review for my math test I am going to have to go over my notes and use those as practice cause I have no text at the moment.  I am VERY responsible though and it seems strange that I wouldn\'t have it but I am not going to worry too much about it.  I understand my Chem. a little more.  Mrs. J also talked to me a little bit today, and she is like my favourite person in the entire world to talk to.  She is so awesome and like one of the best teacher\'s I have ever met, and like really cares about people.  I feel like things are starting to go my way.  I still have alot of things on my plate though, I have the math test, the chem re-write, the Chem project, the religion tests, my project and M.C. for the rel. final, and the next chapter 9 pt. 2 test next week.  Yeah it\'s alot.  But I CAN do it.   '),(3818,1807720,'02,January,2004','2004-01-02 00:00:00','    I feel kind of bad right now.  Holidays are good, nice....relaxing.  They are wonderful, but what is not very wonderful is the fact that I feel forgotten.  I feel like I have no friends and like everyone has forgotten me.  No one calls me.  No one!  If I don\'t call them then I don\'t hear from them.  This is the first year that no one even called me to do anything New Years Eve.  I just stayed home on my couch watching the re-run of the Canadian Idol Finale.  Everyone else went out, while I just stayed home and was lonely and bored out of my mind.  Then I called up Chels today.  Usually me and her can have 2 hour conversations no problem and like just talk and talk.  When I called her today it seemed like the last thing she wanted to do was talk to me.  All she said within the time span of five minutes was: \"Yeah, uh-huh, maybe, and oh.\"  I was kind of like having a one-way conversation.  I KNOW for a fact that if it had been Monet that called she would have not been so boring and unenthusiastic on the phone.  I did offer to her that we could go to a movie today or something but I never foreced the decision on her, all I said was \"call me later then if you want to do something.\"  For some reason though I know that she isn\'t going to call.  I don\'t think she wants to do anything with me right now.  No one seems to ever want to do anything with me, but why?  What have I ever said or done that was so wrong?  What was it that I did?  After that conversation I just cried because we aren\'t the kind of friends we used to be.  Ever since she met that new girl Monet this year she has been way different.  I mean I have nothing against Monet and she is very nice, but all the stuff Chels and I used to do together is the stuff that Monet and her do together now.  I always hear about \"When Monet and I went to the mall\" or \"When we went to so-and-so concert.\"  No one invited me to the mall....and no one ever invited me to the concerts they have been to.  Chels used to be the ONLY person that I could talk to about anything, and it hurts me so much now.  There are so many things I want to talk about, so many problems that I want to discuss with someone but I feel alone right now and there is just no one to talk to about anything anymore.  It hurts alot.  It never occured to me that someday our friendship wouldn\'t be as strong as it once was, and I never thought that she would meet someone that has better qualities in a friend that she would like better then me.  It makes me wonder sometimes \'Where did I go wrong?  What did I say to make you feel this way\'?  I wonder why no one ever calls me.  I didn\'t do anything to hurt them or upset them, it\'s certainly not something I did or said.  I feel like I have no friends right now.  At lease no close friends at the moment.  As I get older the people I once were close with just somehow don\'t seem as close anymore.  I know this is a common feeling for a lot of people, but I don\'t want to loose all my so-called friends that I once had.  I can\'t do that you know, I can\'t do life alone right now.  I need people here for me more than ever.  I don\'t know what I would do without anyone else.  What I am trying to say is that basically I just feel very ALONE and FORGOTTEN and I need someone in my life right now that cares just a little.  That\'s all.   '),(3819,1807720,'29,February,2004','2004-02-29 00:00:00','    Here I am on my computer again writing in my journal, the one friend that always listens and doesn\'t criticize.  You know now that I think of it, I think I saw into the future.  I really think I had a vision of it before.  I was thinking about dreams I had in the past and I remembered the one were I was walking through the Science wing and all these strange people said \"Vanessa come it\" and I walked into Mrs. J\'s room.  Now that I think of it - it was basically the future.  I foresaw that I would have another class with her.  Either it\'s a coincidence or I really dreamt the future!  Scary thought!  I bought the followed up novel for speak - Catalyst.  It is absolutely AMAZING!  I can sooo relate to it, especially because the main theme through it is Chemistry.  I miss chem!  Anyways yeah I went to Vanessa W\'s place yesterday for her b-day.  I had a fun time, although I am all \"junked up\" from all the shit that I ate that I probably could have done without.  I also finished my Drama project too, and now I just have to finish my warm-up which I am pretty well almost done.  Today I also want to review the Bio stuff we learn cause she said we might have a quiz Tuesday.  I will do good on that, cause I can.  And well I think I might go to the doctors sometime this week because I think on my stomach I am having some sort of allergic reaction.  I don\'t what it is and last time I went to the doctors they told me it was acne - but it\'s not because this stuff like swells up, is itchy, and painful.  Ok no offence, doctor or not, I know for a fact that\'s not acne.  I really need to find out what it is too, because I was actually going to go get a piercing there without telling anyone.  However, I need to find out if anything is going to cause me to have a reaction, cause if my mom ever found out I would be one very dead gal!   '),(3820,1807720,'28,February,2004','2004-02-28 00:00:00','    Last night I had the strangest dream.  I dreamt that I kept eating all the chocolate I could find and I couldn\'t stop.  I would walk through like  narrow halls and all around were people that I knew just staring at me, saying nothing.  Then I went home, got on my scale, and yelled \"omg I gained 10 pounds, what am I going to do?!\"  LOL.  I guess the message here is to stop eating so much junk!  I am sooooo afraid of weight gain it\'s not even funny.  I suppose lunch yesterday didn\'t help.  Ash, Chels, and I went to East Side Mario\'s for lunch.  We actually had such a good time we were there for two hours.  We even had dessert afterwards - this big chocolate thing which me and Ash just went at.  Chels described it as: \"It was like two monkeys fighting over one banana.\"  Ok I have to admit I am a pig when it comes to chocolate.  I mean c\'mon who is like so junk deprived - ME!  Bio actually wasn\'t that bad yesterday.  I was a little worried about how I did on the unit test.  Turns out I got a 72% - not bad, but I am capable of an 80% or higher and that I know is possible.  Then unit we are getting into now does seem like it would be hard, but it\'s sooooo much more interesting than anything we studied on the biosphere.  What we are doing kind of reminds me of Chem in a way.  I feel like I am in my comfort zone.  I am also really motivated now because since she said it was a hard unit and I even more determined to prove myself capable of mastering hard tasks.  I have put that quiz that I did bad on in the last unit behind me.  It was one quiz, and a quiz is a quiz - nothing more.  Not ever worth much.  And I know that I am capable of great things.  Drama was also really great yesterday.  I am so proud of the work me and Megz did.  We put a scene together in about 5 minutes.  It couldn\'t have any words, just actions.  A beginning, middle, and end of course!  We actually weren\'t trying to make it funny, but everyone ended up laughing at it - so were like \"yes we actually made people laugh!\"  We even got Mr.B to laugh - mission accomplished!  LOL...  It was about the girl who goes shopping(me) and she sees these amazing shoes at a store, but the clerk(Megz) also wants the shoes, but she hasn\'t saved enough money for them.  While the shopper is convinced that she couldn\'t live without the shoes, the sales woman tries to sell her an ugly bag.  However the customer doesn\'t want it, so both the customer and the saleswoman run the the shoes and they each grab one.  They then circle each other trying to get the other shoe to make a pair.  However, the customer pulls out a $20 for the shoes and the saleswoman can\'t refuse.  The customer puts on the shoes, and sighs of relief then walks away!  That was our lovely speechless scene.  I am so proud of out work actually.  I felt it was some of my best work in Drama!   '),(3821,1807720,'26,February,2004','2004-02-26 00:00:00','    I don\'t know what to say about the Bio test except I thought the M.C. was really hard.  Everyone else that I talked to was like \"yeah that was easy.\"  I guess I shouldn\'t judge if I did good or bad on it right now, but I am not confident that I did well.  I got confused on a lot of questions, but I studied a lot for that.  It\'s like I know the information, but when I get to the test something happens.  I don\'t know what because I can\'t exactly explain or describe it.  There was lots of reading charts and number things to, and that\'s just not my thing.  Besides we hardly ever read charts or answer questions on them so it\'s odd that there would be a lot of that on the test.  I just pray I do good because I deserve it.  I can\'t deal with doing bad on something that I know for a fact I know very well.  I guess I will just have to wait for tomorrow and see.  We did go Off-Campus for Active Living today.  It was really fun!  We went curling and like at first it was really hard, but by the end of it we were getting a lot better and at the end of the game it was a tie.  Swain and I teamed up against the other girls in Off-Campus.  Drama was really cool today too.  We had a guy that specializes in make-up and costume design come it.  He did a demonstration on Megan which was amazing!  He made it look like she had wrinkles and bruises, and even scars he could do with the make-up.  It was amazing!   '),(3822,1807720,'25,February,2004','2004-02-25 00:00:00','    It was a good/bad day.  I can\'t really make up my mind how I feel.  I do know though that after the Bio quiz I literally wanted to kill myself.  In fact, I know that I KNOW everything on that quiz.  I do know it.  Just something happened, I lost focus, got discouraged by something I didn\'t understand.  In fact I thought the whole quiz was confusing.  Like some sort of unfamiliar foreign language.  I am really upset.  In fact I just wanted to cry during and quiz and after.  If I could I would of just let the tears roll, of course now being upset has built up inside of me.  I am still upset now actually.  I was doing so good in everything in Bio, and then this stupid quiz.  I know she said it was hard and that people do really bad on it, but I STUDIED SOOOOOO MUCH.  It hurts so bad because I know I deserve to do better, I know that I deserve it.    I know a lot of other people probably did bad too but I really really think I have a learning disability.  I don\'t care if it sounds stupid, or funny.  I don\'t care if people think I am really smart and that it\'s not possible, because of course I am smart but I DO BELIEVE it\'s possible.  I don\'t know what is wrong, but for a while now I have clued in that something about me is different.  Maybe not from the outside, but something is wrong!!  I used to even think I had a brain tumor or something crazy like that, until I realized that was crazy to think that.  I guess I just want an answer to my struggles.  When demands increase I can\'t keep up with it.  No one but myself knows really how often I cry trying to do homework and how many times I have wished to be dead because of my struggles.  All my life I have struggled with Math especially and aspects of English.  Especially the reading comprehension.  I used EVERY strategy that I have been taught throughout school and you know what I scored on the last reading comp. test in English - 35%.  The thing is that I have always struggled with it and people just ignore the fact, and it doesn\'t help.  Ever since I was introduced to reading comp. I have struggled with in, and I don\'t know how to pass those tests!  I used all the strategies and everything - so where do I go from there???  Honestly I don\'t know.  I don\'t know how I got through Chem, or math 20 last semester, I honestly don\'t - I suppose maybe it was because of my 200% effort I put into everything.  Of course it doesn\'t end there.  In fact jump in a car with me and you will learn a lot more.  I failed drivers ed.  You can\'t really understand what that does to your self-esteem until you have experienced it.  And guess who has experienced it - me.  Sometimes I wonder if I am just not ready to drive, but I feel ready.  Everyone always talks about how they want to drive and it\'s fun.  You know what I think when I hear the word driving - STRESS.  I\'ll admit it - Driving I find to be one of the hardest things.  Every time after the drivers ed class I would come hope crying.  I never did go for my license - why would I after failing drivers ed.  My parents just assumed that I am not ready.  No one ever stops to think that maybe that isn\'t the case.  Sometimes I wonder if maybe I am just a slow learner, but I don\'t think that\'s the right answer.  I mean everything points to some sort of learning disability.  Whether people see it or not, they are not the ones who live my life.  They should try walking in my footsteps  -Math being hard all my life -Failing/Barely passing reading comp. tests -Difficulty driving -Not understanding directions for labs (Which is why in Chem half of the time I just sat there during the lab - it\'s not because I didn\'t want to do it, it is because I didn\'t understand what to do) -Difficulty with relationships (Though I am sure lots of people have this problem)  To be honestly I find EVERYTHING difficult.  Whether I always show it or not, it\'s difficult.  No matter how good I may do in Social, or Religion or anything like that.  Plain and Simple - it\'s HARD.  All of it!  Not that I am saying life should be easy.  Other people excel at other subjects while they may not be strong in others.  I on the other hand am talented in the arts I suppose.  Not drawing wise.  Mainly Drama though I play the flute.  However it would be a lie to say that it isn\'t hard because learning and playing is a struggle and a challenge.  I wish I had answers.  Some sort of answers.  I wish someone could explain to me why everything is so difficult for me.  I wish someone could tell me why I struggle so much.  I wish I knew a way to change that, to be less stressed out, to not cry so often and have emotional breakdowns at night while \"trying\" to do my homework.  It\'s frustrating feeling like something is wrong, but the rest of the world doesn\'t seem to think so.  Maybe they\'re right, but I don\'t think so.   '),(3823,1807720,'24,February,2004','2004-02-24 00:00:00','    Today was really good.  Of course usually the days of semester II are good.  In fact I got 14/14 on my Bio assignment - that would mean 100%!!!!!  Tomorrow we have a Bio quiz which I know I am capable of doing good on.  Also I have an English test but if I prepare enough for it I think I will be ready.  I got like 30 something % on my reading comp test and well I can re-write it but the thing is that I TRIED MY BEST.  I used all the strategies that they have taught and everything.  It\'s frustrating not being able to successfully do reading comprehension all my life.  This is why I have always thought that I might have a learning problem of some sort.  It\'s confusing though because sometimes I can score amazing marks, and at other times I put my best effort forward and it is barely apparent.   '),(3824,1807720,'23,February,2004','2004-02-23 00:00:00','    Ok well I haven\'t posted in a long while.  I have been so busy, but none the less so extrodinarily overjoyed with the content and satisfaction of life.  I got 89% on my last Bio quiz, and so far have managed to score 70% or above on everything thus far.    Friday was amazing.  It was eventful and we were on the go since school ended.  First I had a Model UN meeting, then I ate a nice juicy chicken burger from the YMCA cafe.  Megz and everyone met up (about 23 people) and we divided into groups with one assigned driver and we given cameras and a scavenger hunt to do.  Some of the things we did were EXTREMELY CRAZY and are pictures to prove it!  Then we met up at Pizza Hut for pizza and Megz opened her b-day gifts.  I got her a g-certificate to body things which is a place she buys dance stuff from.  Then we went to East Side Mario\'s and ordered a million desserts til\' we felt like our stomachs were going to burst!  Saturday morning I woke up bright and early to get to MODEL UN.  We met sooooo many amazing and wonderful and smart and good-looking and intelligent people!  It was so fun!!  I learned soooo much and walked away from the experience just saying \"WOW.\"  I learned so much about myself, the UN, and gained plenty of knowledge that can be used in everyday life.  The University is soooooo cool and the banquet dinner was the most amazing this ever.  I felt like a queen for one night, like royalty.  No one even serves me dinner!!  I even made a really good friend.  Since in the Model UN you could write a note and then put the country name and the section such as GA: General Assembly (which is actually the room I was in) I wrote a note to Mr. Egypt I call him.  I think he is so perfect, just everything about him.  Lets say we sent a lot of notes to and fro and he even has my email now.  He goes to Grandin and he seems that type to keep his word.  In fact I know that he will contact me a while\'s after Model UN.  I walked away with more knowledge and a good friend.  I also met a lot of other brilliant, intelligent, and unique people!  Life seems to be going my way for once.  Everything is falling into place.   All I have to say is \"let it fall!\"  I mean Bio is going really well and I am doing good, MODEL UN was amazing and I met Jack (Mr. Egypt!) and I even got a part in the school play!!  What more could a gal such as me ask for?!?!  Note much more I wouldn\'t think!  I am more then just content, I am overjoyed.  Today I smiled for then I have in the last...hmm I can\'t recall how many days!  Of course I even said \"hey\" to LaCour and she didn\'t even have the dignity to say \"hey vanessa\" of course maybe she was just having a bad day.  But whatever, is people can\'t be polite and say \"hi\" and acknowledge that others exist then that\'s their problem!  Active Living was amazing today!  AMAZING I TELL YA!  Everyone wants ME on their team!  ME ME ME!!  That\'s right!!!!! - ME!  In jr. High I would always be like last, now it\'s like - hey we want the Work Horse!  LOL!  Swain kept telling the guys how they should play like me.  I bet out just about every guys out there, and out worked them all.  The guys on the other team were even complaining about not having ME on their team.  Swain asked if I could switch so they would stop complaining!  So I switched teams, and I still could out work them and beat em!  It was great!  I am really good at broomball!  I am turning out to be good at a lot of things.  Before I was just too blind from the discouragement in Chem and Math to see beyond the bad things, but I see further than even before.  I see the horizon where the sun comes up.  It rises beyond the hills and the sun shines on ME!   '),(3825,1807720,'19,February,2004','2004-02-19 00:00:00','    Today was good, but I have another Bio quiz to study for again.  Yesterday I stayed at the school and then got supper and then had to be in the Drama studio for 6:00 pm to do a rehearsal of our scences we were showing for the open house.  Then we performed from 7:00 and went past 8:30.  I didn\'t get home until after 9:00 pm.  I was so afraid I wouldn\'t remember the Bio I studied.  I think the quiz was ok today, but there was something that I just totally could not remember no matter how hard I tried.  It was even part of the topic that *I* researched.  Doesn\'t make very much sense does it!!  I also made a mistake with the last question on the front side.  I put that animals absorb water through cellular respiration.  Oh boy Vanessa, that\'s like not even close!! LOL.  I guess I can\'t change the fact that I made a stupid mistake like that, so I guess I will just laugh about it.  I have sooooo much to do this weekend though.  I have that Bio project, which is alright and not that bad but I also have an essay to write for English.  She isn\'t even giving us the question for it until tomorrow.  Also tomorrow after school I have a model UN meeting and then I \'m meeting up for Megz party and we are going for pizza after.  Then on Saturday and Sunday I will be at the university all day!  I\'m going to be a busy busy gal!  No rest for me!  Broomball today was really fun.  Active Living is fun!  I\'m glad I switched Physics for it.  It is such a nice break from hard work and everything.  It\'s nice to just have some fun.  Of course Drama is fun too!  And good news, no great news, no EXCELLENT NEWS!   ...I MADE THE AUDITION CUT FOR A MID SUMMER NIGHT\'S DREAM!   ...And I got Cobweb as my part  ...It\'s nothing major, but I feel good just to have a part and actually be chosen  ...This semester kicks @$$!  ...And did I mention I love life right now?!?   '),(3826,1807720,'17,February,2004','2004-02-17 00:00:00','    Words cannot describe how I feel today.  Feeling loved, cherished, important - that\'s what I feel like today.  In Bio my group got all their work done!!  I have trust issues you know and well it was a little bit (ok well not a little bit, but really) it was was really hard to trust some people I don\'t know.  And well they did finish their work and our project is done and it\'s all ready to present for tomorrow.  In Drama Mr. B wants me to like play my flute when the audience are walking in and while there are scene changes going on.  He wants me to do so much and it makes me feel so important.  I guess that is because I have a very important part in the Drama presentations tomorrow night at the open house.  After our little run through of everything today he even was like \"And lets thank Vanessa for all that she is doing\" and everyone clapped for ME!  Just for me - no one else.  It truly was an applause, a 100% dedicated applause just to me.  That makes me feel so incredibly special.  Even when I was sitting upstairs in the Science Wing and eating my lunch, Ms. LaCour said \"Hi Vanessa.\"  Of course if she said that to anyone else I don\'t think they would appreciate it as much as I do.  No one hardly ever stops to say \"hi\" to me anymore.  It\'s like sometimes I\'m invisible and the whole world can see, but they can only see beyond me and not what is in front of their eyes.  If that makes any sense.  I guess what I am trying to get across is that I felt noticed you know.  I do like it when people just say \"hi\" to me, because it doesn\'t happen a lot.  Active Living was the best today!  We played broomball and I like got my right hand all scratched up and even broke my nail off.  It actually really hurt and now there is blood on my new expensive jacket.  It was so fun though.  The guys even gave me a nickname - Work Horse!  LOL, ok Mr. Swain was like \"You know she has a name and it\'s Vanessa\" but you know what, I actually kind of liked being called Work Horse.  The guys said they called me that because I worked harder than all of them.  I would do all the work, then they would just score.  They even were like \"C\'mon Work Horse we want you to score for us\" and like they helped me score a goal!  That was awesome and I think I am going to make my MSN name Work Horse - cause I do work my ass off - I\'m a work horse!  How cute!  After class Mr. Swain even told me that I worked harder than anyone else and I beat out everyone out there.  That was an amazing compliment to get from someone like him.  It made me feel good.  Of course now I have to end my journal entry with boring ole\' English.  That class well was boring, but I mean C\'mon every good day has it\'s boring class.  Now that can\'t be debated.  You know I think the list of people who made in to the cast of A Mid Summer\'s Night Dream is being posted tomorrow.  I really can\'t wait because I know for a fact my audition was really good.  I just hope I am on the list.  If I am not then it\'s not the end of the world, but I do have confidence in myself that my name very well may be on there.  Fingers crossed!   '),(3827,1807720,'16,February,2004','2004-02-16 00:00:00','    I drove today and I drove good!  For once my parents didn\'t insult my driving, for once I wasn\'t told that it totally sucks - because it was good!  My driving is a lot better than it was before.  I think once my Dad\'s contract expires he is going to review with me uphill and downhill parking, and probably some parallel parking too so then I can take my drivers test.  I feel a lot more confident now, and a lot more ready.  Once I review that stuff I should be good to go.  I also went to the zoo today.  There were so many people there.  I think half of Calgary was there actually.  It was really neat cause I saw the new white tiger they have, it was so beautiful.  And I saw cute little (actually big) sloth bears playing - how cute.  And I said the cute lil\' adorable red panda.  There were other things I saw, but of course the tigers, sloth bears, and the red panda seems to stand out more in my mind.  Got school tomorrow.  I am excited of course but scared for Bio for some reason.  Silly me!  I am also kind of nervous because the model UN is going on during the weekend I am involved in it, so basically I am going to be very busy next weekend.  The closer it gets to it, the more nervous I get.  Though great people are also doing it with me, like Megan, Kenda, Elisa, and Maria to mention a few names.  I think I will be getting a ride with Elsie that day, but if not then probably Megz.  I think we will stick together cause we won\'t know a lot of people.  We had a sub for English on last day of classes before the holidays, but our English teacher was there and told us to read a story and that we were having a test when we get back on the story.  See the thing is - I don\'t remember the story that I read!  I mean I didn\'t understand it in the first place, but it was Wednesday that we read it.  What am I suppose to keep these things stored in my precious memory???  I don\'t think so!  So I am thinking that tonight I might read it again just in case.  I also can\'t wait for Active Living and Drama tomorrow.  Hopefully we get to play Kinball again in Active Living cause that was like so incredibly fun!  And well in Drama we are putting the final touches on our February 18th performances for the Open House at the school.   '),(3828,1807720,'15,February,2004','2004-02-15 00:00:00','    There is sun this morning!  SUN!!  The last few days have just been dull and gloomy it\'s hard to have a good attitude.  Now there is sun.  It\'s so refreshing when you wake up and your blinded by the light.  Guess you could say I am a true Summer gal, though Summer is still far away.  I had like the strangest dream last night.  I guess I have been thinking about school too much.  I dreamt that every time the bell would ring for a class change, I forgot something so I would go back to my locker and then be late for the next class.  When I was going to Bio I walked through the Math wing and in my old class our Bio teacher was there and I remember her saying \"Now see the reflection of the mirrors?\"  I wasn\'t sure if that was the right class so I went upstairs where Bio is suppose to be, but no one was there.  So I went back down stairs to the Math wing.  Strange dream!  I\'ve been thinking about Bio or School too much.  In fact, most of my dreams nowadays are based on school, Bio, or Chem.   '),(3829,1807720,'15,February,2004','2004-02-15 00:00:00','    Back online again.  Got lots more to say.  It seems that there is just so much I have to think and talk about right now.  I guess talking/writing to my computer will have to do.  I don\'t think anyone else will listen to my rambling, or would want to.  So yeah Ben and Melissa came for supper and we ate alot.  I am so afraid of gaining weight but I just keep eating.  Weight is a big struggle, and well yeah I have done good at maintaining my weight loss but I would like to loose more, however I am thinking if I keep this eating pattern up I might start to gain some.  Everyone gets sick of me talking about weight but they don\'t realize that while maybe a lot of people don\'t have weight issues and are thin, some people have a real struggle with it.  I would be one of those people that struggle.  Anyways I was talking to Ben and Melissa and they are getting a hot tub.  That\'s right a hot tub.  Since their new house is being built they are going to buy it soon and put in the basement because once the house is complete they won\'t be able to get it in.  And well I am allowed to go over there and use it if I want and stuff.  YAY!  Although it is a long way from the S.W. but I mean c\'mon it\'s a hot tub - that is like what I have always wanted.  Of course for those who know my romantic fantasy date, I dream of being in a steamy hot tub with roses all around with the dark moonlight sky above and a slight fog outside.  Ok that is enough of me fantasizing!  My Dad is in town this weekend - finally!  I feels like it\'s been forever since I last saw him.  At first I didn\'t mind him fly back and forth a couple weeks here and there to visit, but now it feels like he is not really my Dad.  I mean of course he is, but since I don\'t see him a lot it feels that way.  My mother says that sometimes it feels like she is single so I guess my feelings are actually quite common.  Anyways my parents said that next February break I might miss like a day of school or something but we are going to fly to Florida.  That is where one of the NASA buildings is located, and I made my dad promise me that we are going to go for a tour.  He never breaks promises, so yay we get to go!  I just have a long ways to wait.  That is basically like a year from now - so yeah it\'s a long ways away at the moment.  I think that tomorrow I am going to the zoo!  How exciting - I sound like a kid.  Anyways I want to see that new white tiger, at least I think that is what it is.  I love tigers, cats, you name it - anything feline.  Of course I think that I will sleep in on my last day off.  I miss school a lot because it is like a second home to me.  Half of my life is spent there, and it keeps me busy.  Today I also did that contest thing in the sun where you like go through the papers for the week and find answers to questions.  If you get picked you can get a $2000 scholarship to a post-secondary institution.  So yeah I am going to mail it in, and I think it goes for a couple weeks.   '),(3830,1807720,'13,February,2004','2004-02-13 00:00:00','    On Friday I went to Chelsey\'s house and Edi made us all dinner - Lets just say that was some of the best chicken I have had.  We watch the Lion King 1 1/2 which is hilarious and My boss Daughter or some movie like that.  Then we played Uno extreme and bunch of card games.  It was fun!  Yesterday I didn\'t really do much - just listen to a bunch of music and relax.  This break seems to be actually really short it seems like it\'s going so fast.  I went with my parents a couple places because my dad is in town.  Then we went out for dinner and Swiss Chalet because my mom wanted there chicken.  I felt really lonely on Valentines day though.  I just wished that I had someone in my life.  I just wished for one day that someone could love me, and they could just say \"I love you Vanessa\" and hold me.  I suppose one day there will be someone in my life, but I am afraid that I might not meet that person.    You know I have been thinking about Aaron a lot.  I mean some sort of twist of fate brought us together over the summer.  Then we just never talked after a few weeks.  Some stupid reasons kept us apart during a summer we could have spent time together.  I mean I know he lied to me about his age and everything when he met me, but I think he only did that because he didn\'t want me to turn away from him and I didn\'t.  I miss him a lot.  There have never been anyone like him.  When we met each other both him and myself felt like we were meant to be together for the rest of our lives.  That is not a common feeling when your young.  There was a spark with me and him, there was Chemistry between the two of us.  I don\'t feel that way with anyone or about anyone else.  I have never felt that way about or with anyone else.  Sometimes I wonder if I will ever see him again.  God can only do so much to create your destiny and let fate do the rest, but I guess sometimes that might not be good enough.  Ultimately we are the ones that decide what direction our lives go in.  Some of us take the right directions, some of us take the wrong directions.  Somethings we don\'t even know which direction is the right one and which is the wrong one.  I don\'t know for sure if Aaron and I are meant to be together, but I am starting to realize that what I felt around him, just with him was something so powerful.  When I think my future, I can see him there.  Ok so he lives in Halifax and I am in Calgary, but those are just slight differences.  You know when I met him, he though I was perfect.  And yet I was just being myself - image that myself, and that is perfect to him.  To him I didn\'t need to change anything.  No one sees me that way.  Chels was with me that day we met - July 18, 2003 - I even remember the specific day.  Usually when guys meet Chels and I, they are always attracted to her.  Maybe cause she opens up more right away, maybe cause she is thinner, who knows.  For once though, someone was attracted to me though.  Me for once.  And that was a special feeling in itself.  I think that I might have given up something that could have been great.  We only stopped talking cause he lied to me.  I have trust issues and don\'t like people who lie - but I realize now that he was doing that because he didn\'t want to loose me.  I see the picture clearly.  Though his plan back fired, and he didn\'t realize that I am the kind of girl that just takes people for who and what they are.  That is just me.  He also needed a little work on the hygiene, but that can be fixed.  Everything else is perfect.  To each other we were perfect and if I ever got to see him again, then I KNOW that we would still feel the same.  He was a good looking guy, cute.  Even chels had to agree with me.  I always look at the pics from the summer.  I look at the smiles on our faces.  We had so much fun, with Aaron it was all about fun.  And that is what I want.  Lets put it this way - I miss chem, I miss chem, I REALLY REALLY miss chem.  I am a confused girl.  I am not about to go make judgments about my future but I know that I want to be a high school teacher, but now the question is what kind of teacher.  I know that I want to teach something related to the science area and have already ruled out physics because that\'s too math based and just not my thing.  So what is left then - Bio and Chemistry.  We haven\'t got totally into Bio yet, but I find it kind of dry although since most if memorization I am pretty sure I am going to be way better at that.  However, Chem I found VERY difficult, I mean honestly I thought it was the hardest thing that I have even done.  I like it though.  You know I thought I didn\'t like it, or the teacher, but my views and my opinions have totally changed.  The chem teacher is one of my most favorite teachers of all time.  At the beginning of the year I thought she was evil, now I think she is the coolest.  See I am confused though, I always seemed confused.  Because well I kind of want to teach Chemisty, which would be ok I was as smart as the chem teacher.  I mean I can barely balance a chemical equation properly and I want to be a chem teacher...!  I can\'t explain it, it\'s weird.  I mean why do I always want to do what seems hardest.  It would be easier to teach bio probably, because I would most likely be better at it.  Eventually I do have to decide though.  Will I major in Bio or will I major in Chem???  Do I choose the area that I like more or the one that I am better it????  So many questions and I have to decide before I apply to university next fall.  You know I am accomplished more than I have even thought.  I am not the naturally smart, talented, or gifted.  I am your hard worker.  I work for all my dreams.  I turn those dreams into goals, and I accomplish them.  Of course University if whole other challenge in itself.  I am afraid.  Afraid that if I choose Chemistry to major in I might get stuck in a rut of not understanding because I do find it very difficult.  Then University and all the money spent would be a waste.  If I choose Bio, it would be easier but still challegning and I could probably do it, but I wonder since it doesn\'t interest me as much as chem, would I get tired of teaching Bio.  You see I just don\'t know, and I guess that\'s life because really none of us have all the right answers.  You know I do think that I would make an excellent teacher.  I know what people go through.  I understand that people do need help and that things can get very hard sometimes.  I mean I have sat down enough nights just staring at the math book and crying, looking at it like it\'s some kind of foreign language, and not being able to answer the questions.  By some amazement no one even knows that, I keep it so well hidden.  I even managed to finish all my homework last semester and everything.  No one will even know how many people I had to ask to understand, and all the nights I had to cry to finally get a little bit closer to understanding.  In fact lots of people think I am really good at social too.  You know I guess it could be partially true, but really I don\'t seem to learn anything from that class.  I have just come up with my own techniques for getting good grades, not actually mastering the material.  For the long answer tests since they give the questions before I wrote the answers out, and then just memorized the answers.  After the test was done, everything I had memorized seemed to be erased from my memory.  I admit my study techniques suck - What study techniques is the question I should be asking!  You know teachers spend so much time or material and teaching it, but what they don\'t teach is how to properly study.  I mean of course I study all the time, but I often wonder if I am doing something wrong.  After all if I can\'t get the marks I deserve and want, then don\'t you think teachers should teach us how to properly prepare for exams.  I learnt more in Chemistry then in Social, but my Social mark was like way higher then Chem.  So what does that show anyway.  Are marks really an accurate way of showing how much people learn.  You know really a mark, a percentage - it\'s just a number that\'s all.  It doesn\'t define how much you know, or how much you don\'t know.  It doesn\'t determine you future (well usually it doesn\'t) so then why is there so much emphasize on them?  You know I never thought about that until now.  So now that I think of it, why would I let my Chem mark stop me from doing what I want to do.  It doesn\'t define me, or what I know.  So when the time comes to apply for University, which will come sooner than I am imagined, I am going to major in Chem because that is what I want to do.  I don\'t care if it\'s my lowest mark, highest mark, whatever it ends up to be.  Why would I let a number determine my future.  I will decide what I want to do based on my feelings.  After all I wouldn\'t major in something related to Social just because I have a high mark, so why would I choose not to major in something I have a lower mark in???  You know just thinking about this makes so much sense, like I feel I have gained about 2000 points towards my confidence level.  I think that if you believe, then anything can happen.  You bet I am going to make what I want it life happen.  I have made my dreams before into goals that I have achieved, so why not do it again.  I think I know what I want to do in life know.  I know now.  After High School I just have to do it.  I will be a High School Chemistry Teacher.  That\'s it.  It doesn\'t matter how far fetched that seems right now.  The things that are harder are always more of a challenge, but somehow I always go for things that don\'t come easy to me.  It just means I have to work harder to get what I want, and to get where I want to be.  I see clearly now.  I understand.   '),(3831,1807720,'13,February,2004','2004-02-13 00:00:00','    Ok so I am back online!  I just got home a while ago and so much has happend since this morning.  You know today I really feel \"complete.\"  I feel like whatever was missing before has now just arrived right infront of my face.  Even I am not sure what this means, I just know I feel good.  I got my brows waxed today and went shopping again and bought myself - YES I TREATED MYSELF - anyways a pink spring jacket.  I LOVE it. , and I can\'t wait to show chels.  I think that her and I and some of the uiversity/college guys are going to Edi\'s house to play some pool and stuff.  And well yeah I guess that jacket was a reward to myself for all my hard work for making honors and I used part of my grandma\'s $40 that she gave to put towards the total value.  My mom also bought me a roxy shirt which I love.  And well I got my eyes checked today too, and my vision is still the same.  Did I mention that I also ruined a pair of my new underwear that had little glimmery jewels on em.  It wasn\'t suppose to be my T.O.M. til like next Saturday but you know how it is.  The monthly visit always comes when your least expecting it, lol - that sounds so bad!  Anways year grr. that was like my favourite pair, but well it could have been worse.  In fact, I have had and been though worse things than that.   '),(3832,1807720,'11,February,2004','2004-02-11 00:00:00','    Fun day yesterday!  A fun, fun day!  Went to South Centre and got some black pants.  Lets just say I have never seen my ass look so damn good in anything before.  Then I got a new pair of jeans form AE and a shirt for when it gets warmer.  In sears I went to the Clinique counter and this cute guy was working in cosmetics - like I have never even heard of a guy in cosmetics!  So yeah he was cute, and we were looking for some makeup for the wedding even though it\'s not til\' may, but the guy was like - \"do you want me to do your makeup\"and I was like \"Y-E-S.\"  So then he did one eye and then showed me how to do it like he did, it was really neat.  He showed me how to apply my eye liner and mascara because those two things I have never worn before.  It was really neat, and like I also got this lipgloss stuff - #7 Sugarblue.  Now that is some cool stuff and I also picked up this shimmer stuff with the happy perfume in it.  Oh I am beginning to like cosmetics - I get it now when people like make-up.  Of course I don\'t wear any to school though because I mean when would I have any time to put it on.  Saving it for special occasions makes it all the better.   Actually it was really warm yesterday.  There is actually sun again today too!  I love the sun.  It\'s like when I wake up I am blinded by light and I don\'t feel so depressed.  Semester II is so much better you know.  I mean I confess though that I do miss Chem.  I actually miss it ALOT!  I guess I spend like my life in that class - during class and tutorial and now it\'s like a part of me is missing.  I didn\'t realize how much I enjoyed it, until I didn\'t have it anymore.  Bio I think is good now though.  Of course I am not friends with many people in that class because they are mostly grade 10\'s, but for the group assignment I met this really nice girl, Lisa, and there is this other really nice girl, Aaron, that sits next to me.  I feel a little more comfortable in that class now then I did at the beginning.  It was really awkward and everything and well I am not sure what I think of these projects that we have to do and then like \"teach\" the class.  It\'s a little odd, because like usually we wouldn\'t do something like that for Bio.  I am afraid that I might not be able to learn the material as well as if it were taught.  But whatever happens I am going to know my stuff, cause studying will actually pay off in Bio, where as Chem not so much because it\'s a lot of application of your knowledge, rather than the just straight forward information of what you know.   '),(3833,1807720,'11,February,2004','2004-02-11 00:00:00','    Today was really awesome.  Drama was really fun today and we worked on our scences for the Feb. 18th open house performance.  I think that the Bio quiz went pretty good and there were only a couple questions on the back that I got confused on cause they were like those questions where you had to choose between option A or B.  I forgot to put my name on it though and realized afterwards.  I also forgot my name on the lab we did the other day.  I think what happened is that I wanted to get all that stored up info in my brain out and when I did the quiz and got it all out I worried about the info, but I forgot the simple part - my name.  When I told Mrs. J that I forgot my name she made like this comment that I actually kind of thought was rude - but really I don\'t think she was trying to be rude, that is just how I interpreted it I think.  She was like \"What\'s with this name thing?\"  At least of what I remember it was something like that.  I kind of got offended because I can\'t help that I keep forgetting to put my name on things.  How am I suppose to remember that?!?  If I don\'t remember, then I just don\'t remember.  In English we had a sub and we hardly did anything.  She just assigned a story to read and that was that.  So I read the story and then had free time so I played some games on the comp and finished off my bio.  Then I had active living.  SOOOOO FUNNNNN!!!  We played kinball and that was my first time playing it, but like the people on my team were so nice and awesome.  I had so much fun.  Our team even won, and we beat the four other teams.  Now I am just sitting here thinking about my day.  I am really having a big issue with my weight.  I keep eating alot, and I am gaining some weight.  I feel like I should stop, but sometimes I just have to keep going.  What happens if like for a week I will gorge on anything and everything.  Then the next week after I find that I have gained weight so I barely eat anything.  Of course it would be a lie to say that I don\'t like food - because I do.  I just don\'t know how to control that part of me right now.  Sometimes it\'s very easy, but at other times it is very difficult.  I am REALLY REALLY going to try eat alot more healthier and try not to gorge.   '),(3834,1807720,'10,February,2004','2004-02-10 00:00:00','    ARGGG!!!  I can\'t deal with old people.  I mean I love my grandma and all but she talks and talks, and TALKS!!!  Sometimes I just want to sit down and have a quiet peaceful meal but that is impossible.  When I got home she was like: \"look, look at the rabbit outside.\"  And like I have seen a million rabbits in my life time and like can anyone say...BORING!!!  She kept going on about that thing for like 10 minutes.  Ten minutes about a rabbit ok!!  Can you feel my pain??!?  In Bio I am kind of worried that some people in my group might not get their stuff done.  I have to be worried because my marks are VERY important.  There is this one really nice girl, Lisa, in my group and she is willing to do the work, but the other two people drew a dart board on the white board and used some sort of magnet for the arrow and they played that while they were suppose to be doing research.  Then they broke the first laptop we were given and they glued on the key they broke and then got another one.  When they got the other laptop they didn\'t even do their reasearch - like I said they just played that dart game thing.  I feel as if I am being portrayed as a bad student or something.  I mean with no one nominating me for habits of mind last term that especially makes my thoughts more realistic.  Especially I feel like Mrs. J feels I am a bad student, I don\'t know why I feel this way but something makes me think that right now.  I know I have a good work ethic and try hard, and do know lots, but sometimes I wonder if she ever forgot that and now has a different opinion of me.  I am probably just over reacting and thinking too much, but that is how I feel.  English was actually really good today.  We had to answer these really boring questions but then we got to write a poem.  And poetry is MY THING.  I LOVE writting poems.  So yeah that was fun, and something I am good at.  We played handball in active living and I thought that the guys might not let the few girls play but they did and they were totally ok with it, so yeah that was fun.   '),(3835,1807720,'09,February,2004','2004-02-09 00:00:00','    I had a good day today except for active living.  Katie was being rude and stuck up, and we had to shoot hoops for an hour and a half.  No offence but that is not my idea of fun.  Bio was really good today.  She explained the lab which was soooo nice, because it Chem they always left us in the dark and confused about what we were suppose to be doing.  I also auditioned for the play today and I think it went really well.  Drama was fun today and we got to work on our scenes for the Feb. 18th performance.  I am so excited for that, but I hope we don\'t get alot of homework and tests the day after.  English was pretty good too.  The other day she asked me to read, and well it was kind of creepy but like she was calling out names for attendance and she called me \"miss vanessa\" - that was a little creappy.  Overall I did have a good day, and mat and I kind of talked a little bit.   '),(3836,1807720,'08,February,2004','2004-02-08 00:00:00','    Yesterday night it was Andrej\'s 19th b-day so he took us all out to eat and we got pizza - lots of it.  And well I gained some weight too, so now I have to like cut back on the eating.  The we went bowling and then back to kari\'s house.  We had so much fun!  Earlier that day I just practiced my monologue for my audition for the play and then I was bored so I studied Bio cause she said we are having a quiz soon.  Nothing too exciting.  And today I am going to review the bio stuff we learned and maybe do some research for the Model UN thing I am doing.   '),(3837,1807720,'06,February,2004','2004-02-06 00:00:00','    Today was fun.  Drama was so awesome this morning - fun like always!  And in Bio the girl next to me talked to me - someone actually talked to me!  And well English was boring as usual, but she asked me to read so I felt special, lol.  In Active Living we did the 30 minute run and I was STARVING by the end of it, but I found it pretty easy.  So then after I was like \"LETS EAT, EAT, EAT, and I think I annoyed like everyone.  But I am good now, on a full stomach.  We had pitas today - YUM...   '),(3838,1807720,'05,February,2004','2004-02-05 00:00:00','    You know today was going pretty good, but I must be cautious when my days start to turn out good because they always seem to get ruined by one thing or another.  I guess today just proves that \"Life isn\'t fair.\"  Everything was fine until TAPS.  I made honors of course, cause I worked my ass off for it.  And well before I went to interviews I was talking to Chels and she also made it, but she was like \"Oh and I got nominated so many times for habits of mind.\"  You know what happened when I asked my TA: \"So did any teachers nominate me for habits of mind?\" and her answer was a \"NO.\"  When that no came out of her mouth I don\'t think I have even gone from so happy to so sad in that many seconds.  I mean I don\'t understand what I did wrong!  Was it something I did, or something I didn\'t do?!?! I KNOW for a fact that I have worked harder that half of the people in our school.  I have gone to more tutorials than 3/4 of the student population and that I know for sure.  I know that because out of all the days of the semester there has been 10 days or less that I haven\'t gone to tutorial.  Now try finding in the school how many people have gone to that many tutorials.  I bet that someone couldn\'t even find 10 people like that.  I worked my ass off.  And the sad thing is people in my TA who come in the morning are like \"Oh I didn\'t study, Oh I didn\'t finish my homework - I should umm finish it now\"  and sadly alot of the time those are the people who get nominated.  Well you know what, I finish all my homework before the school day and put ALOT of effort into it.  I put in hours of studying, and I know for a fact I put in more hours than alot of other people.  It hurts alot to know that NO ONE nominated me.  It makes me feel like no one even notices me, or even cares for that matter.  I feel like I even worked way harded this year to get the marks I have gotten.  I feel like I have put EVERYTHING into school, and if people didn\'t nominate me 1st semester it gives me little hope for 2nd semester.  Of course not all the teachers, but I presume that most teachers only nominate the advaced, the smart, the gifted, the talented.  I feel like none of the above.  I am a hard worker but it seems there is little room left for that category.  I just don\'t want things to be made worse by Chels bragging tommorow about: \"Oh I got nominated by so many people\" - and blah, blah, blah.  Just what I need to destroy my self -esteem.  But you know after I cried about enough tears to flood my bathroom I came to realize that you know what, I KNOW in my heart, I KNOW in my mind that I really worked VERY hard and put in my 110% effort.  If people can\'t see that, then it\'s their fault.  I know that I do deserve to be nominated, but I am not, so I have to deal with it cause nothing is going to change.  What\'s done is done.  Only God knows why no one nominated me, literally only God knows.  I mean I am trying to not take this so badly, but I didn\'t think not getting nominated would upset me this badly -but I knew I wanted it very badly.  Ever since I learned that you can be nominated for habits of mind I have purposly worked harder and tried by very best.  If that is not good enough then I don\'t know what is.  I feel really bad, but not as bad as before, and definetly not as upset as before.  It would of made me happy if at least one person nominated me, but like I said \"Life just isn\'t fair sometimes.\"  And right now it isn\'t fair because I know what I deserve.   '),(3839,1807720,'04,February,2004','2004-02-04 00:00:00','    I am not sure what to think of the new semester yet.  We haven\'t really got into anything interesting in Bio yet, but Drama cheers me up all the time.  Today Active Living was really fun and we did fitness testing which I did awesome on and tomorrow we are doing the 30 minute run.  I got really freaked out today though - I mean I didn\'t tell anyone but when we were doing a 5 lap warm-up around the gym I starting getting this stabbing pain right about where my heart is on the left side, then when I kept running the pain like stabbed through my stomach.  I had the same pain when I went to go work out at the YMCA one morning and well I am kind of scared.  I mean there\'s pain for a reason obviously and I have to run for half-an-hour tomorrow and I will do what I need to do to get a good mark - even if that means running though half-an-hour of stabbing pain but hopefully it won\'t happen again.  I talked to that Aaron girl in Bio again this morning too - she is the only person I talk to in Bio and well I am glad that at least someone talks to me.  I am scared for doing a lab though.  I am not scared of any labs or anything, I am just afraid that I won\'t know anyone and no one is going to want to do that labs with me.  Maybe I am just over reacting though and thinking too much.  Maybe the gr. 11\'s could just go in their own lab group.  Who knows really - I guess whatever happens, happens for a reason.  I sort of feel out of place in English too.  There are lots of people that I see as \"acquaintances\" but other than that I don\'t really have a lot of friends in my class.  Making friends was never really my thing - that\'s Chelsey thing, then she just introduces me to people - but I am going to have to learn something right.  But half of the people I would rather just not know, because I don\'t make friends with smokers, druggies, or people who can\'t focus, do work, or pay attention.  My English teacher does seem really nice though but she comes across to me as a \"favour picker\"  - those are the teachers that give students better marks because they are their favorite or they just suck up 1000 times, or they already know them.  It is only the first week of semester 2 though and I mean I can\'t really make too many judgments or assumptions yet.   '),(3840,1807720,'03,February,2004','2004-02-03 00:00:00','    Today was ok - but I have no homework because I finished it all.  It feels weird to not have homework  Without Math and Chem I sure have an easy semester - except for Bio.  I actually talked to someone in my Bio class today too which is good consider I was freaked out that NO ONE would talk to me because there are like no grade 11\'s.  It almost feels like I failed it really, but maybe I will make alot of friends or something, but it\'s hard cause most of these people already know each other.  I think I did bad on that fist Bio assignment too because she wasn\'t like really clear about what she wanted so I think I did some stuff wrong.  Even if I do bad on it though, I have to remember that it is only the first assignment and there are MANY others - because I mean I failed the first quiz in Chem but I didn\'t do totally horrible in that class - I did pretty good for my skills and abilities.  Drama was sooooo fun today.  That is like one of the best classes I have in my schedule.  We did so much and the class felt like it just kept going and going, and we were having so much fun.  Tomorrow we are doing fitness testing in Active Living tomorrow and I just know I am going to have like burn out by the end of it - especially if we have to do the 30 minute run.  But I can handle it - I\'m not in that bad of shape.   '),(3841,1807720,'02,February,2004','2004-02-02 00:00:00','    Today was filled with so many emotions.  More happy emotions than anything really.  This morning Mrs. J was like \"Hey Vanessa, your in my bio class\" and I was like \"WHHHATTTTT?!?!?!?????!!!!\"  At first I was shocked, then just happy.  I am soooo happy I got my favourite teacher in the entire world for Bio.  Of course it kind of blows that 90% of my class is in grade 10.  It\'s also ironic because I sit in the same corner and desk in Bio now, as I did in Science 10.  Notice how I always end up siting in corners?!??  -It\'s like everytime I take some sort of science, guess who ends up sitting in an isolated corner - ME!  But yeah today was good.  Drama was awesome!!  Active Living was cool cause our class is small and I know a couple people, and today Justin Pono - the guy that used to hate me and always call me ugly was like \"Hey, How\'s it goin\'?\"  So yeah people were nice, my schedule worked out soooo incredibly nice.  I am still mad at you-know-who but whatever.  And well I umm lost my wallet.  That\'s right, my wallet.  It had my new bus pas, my learners, debt card, YMCA membership, student ID, and well some money too.  Now I can\'t do anything fun cause I have no money!! GRRRR.  O well, hopefully it will turn up, but I feel happy for some reason even though losing my wallet was stupid and sucks, I still feel REALLY HAPPY! :)   '),(3842,1807720,'01,February,2004','2004-02-01 00:00:00','    I\'m not looking forward to see you-know-who tommorow cause I just know we are going to get into another big argument, but maybe something worse.  And tommorow and I just want to be nice, happy, good, all of the above.  Usually in the world of Vanessa that isn\'t possible, but I am going to make it possible.  I am going to be the new me tommorow - confident me, proud me, happy me.  I have wasted so much time being so unhappy just because I only looked on the negative side of things.  Tommorow is going to be different.  For the past couple days that I haven\'t wrote in my blog I can\'t remember what I did.  Stuff.  That describes it.  Today I went to go see The Perfect Score and yesterday night a watched a movie.  Of course that isn\'t all - it\'s just the rest of the stuff that I did - I can\'t exactly remember.   '),(3843,1807720,'30,March,2004','2004-03-30 00:00:00','    Come to think ok it, life at times is good, at times great, and at times just plain frustrating.  Especially Bio right now.  It\'s suppose to be the \"easiest\" unit.  I don\'t find it easy.  I find it the hardest, most difficult, most confusing.  I feel kind of stupid actually for not understanding half of the stuff.  It\'s stuff we did like way back when, but I don\'t remember what I learned like five or four years ago.  I need structure and so far, there has been like none for this unit.  It\'s hard to learn that way and I am extremely worried about the quiz on Thursday.  I don\'t feel like I really understand anything too well.  Honestly I don\'t know how I am suppose to learn from a bunch of fill-in-the-blank-handouts.  That is not my thing.  Why does every damn thing have to be hard for me.  WHY CAN\'T I JUST BE GOOD AT BIO?!?!  I don\'t get it, I mean there aren\'t very many areas I excel it.  Why is it that when I like something I always have to be confused by it.  Grrrrrr.  It\'s very, very, VERY, frustrating.  You would think I would be used to it by now, but let me tell ya that things never get easier.  They only get harder.  Ok so now I have gotten off my chest how hard I think Bio is right now, I might as well say how fun active living was today.  We did social/ballroom dance which was soooooooooo incredibly fun.  We learned the Cha Cha and Chataranga and some other things.  My partner and I were like so amazing at it!  The instructor always \"borrowed\" me from my partner to do demonstrations.  It was so fun and not too hard until like one step at the end.  I could do the guys steps and not the girls, lol.  So then Swain was like \"ok, I will show you\" and like omg I got to dance with him.  Ok so it was like under a minute, but he is seriously hot and good looking and a good dancer.  oh and I went to go see Secret Window tonight, which was good and Johney Depp is amazing.  Tomorrow is also going to be great cause I have rehearsal from 4-6 and well hopefully Bio won\'t be as stressful tomorrow as it seemed today.   '),(3844,1807720,'28,March,2004','2004-03-28 00:00:00','    I\'m finally sitting down, relaxing!  Listening to music, old music, good music - Ace of Base (Cruel Summer) and the new song My Immortal by Evanescence is really good and the song Breathe by Michelle Branch is good.  I have been out driving a lot.  I always drive out to Melissa\'s house in the N.W. all the time and I am getting way better at driving.  My mom even said I parallel parked better than her.  That is the first time I have ever heard my mother say something good about my driving.  I am worried about Bio now too.  I am doing \"ok\" in it, but this unit is suppose to be easy and like I find it confusing.  She keeps giving up worksheets and assignments and stuff but she hasn\'t taught us much about the unit.  So how am I suppose to know the answers?!?!  Also apparently marks have to be in at the end of next Friday, but I still don\'t know my English mark.  All I have to get is over 75%, that is it, then I am on distinction.  I worked hard all year, I deserve to be on there.  Hopefully I get nominated for Habits of Mind.  God and the world knows I deserve it.  No one nominated my last semester, and well someone is bound to nominate me.  Someone will and I know it, I know it in my heart that from the beginning of the year I deserved it, and I still do.   '),(3845,1807720,'25,March,2004','2004-03-25 00:00:00','    Well I just finished doing my Drama assignment write-up thing on my comp.  Last night Edi and I went went for dinner.  He took me to Swiss Chalet, lol, and we had chicken and fries.  Then we fed each other a strawberry sundae.  Ok so it\'s HARD to loose weight though I have done it before.  And I am thinking that maybe genetically I am just not meant to be super thin.  So maybe instead of trying to loose weight I think I will try to maintain it instead.  And you know, be healthy.  Ok I confess I had dessert last night and some chocolate tonight, but it\'s not like I eat that everyday.  I also eat tons of fruit so yeah that makes up for it.  I didn\'t get over 80% on the Bio test, but I got 72%.  Even though it\'s not what I wanted, I don\'t feel bad for getting a 72% because I know that I did my very best.  I\'m thinking that I may have ruined my chances of getting nominated by Swain for habits of mind cause last class before today I feel pressure to not participate a lot because none of the girls wanted to.  I know it was STUPID to not and I realize now, and I got an 85% in the unit instead of a possible 100%.  That is my fault though, my mistake and I have learned from it.  That is what counts.  So for now on I am going to try make it up, because I know that one day can\'t outweigh all of the other days.  I just pray SOMEONE nominates me.  PLEEEEEEAAAAAASSSSSSEEEEE!!!!!  Sometimes I feel like yelling that so freakin\' loud.  Arg, someone needs to see that I WORK MY ASS OFF, in ALL that I do.  Ok so call me crazy, maybe I am.  However, I feel different.  I know everyone feels different and yes being different is good, but just different in how I learn and understand things.  I know I don\'t fail my classes, or hardly ever assignments (Chem is an exception) but what I don\'t think people realize is how stressful and frustrating it is trying to finish them, or even trying to understand.  I don\'t think I am stupid, that\'s not it, and the reason why my marks are maybe average or higher is because I work so hard.  I am no expert in this area, but I honestly think that I might have like a learning disability or something along those lines.  I\'ve only told that to two people, one thought I was crazy, and then other one didn\'t but I don\'t know if I am paranoid of if maybe my instincts are right.  I was reading up on a lot of stuff about it and like a lot I can totally see in myself, except for the underachiement, but I need 120% effort just to do \"ok\" on things.  Like sometimes I will miss numbers or reverse letters, though it\'s not a very common occurrence.  However, when I copy from the board and then go back to my notes later I realize that I have either missing words, a sentence, or wrote the same thing twice.  I can read good, but when I finish reading I have no idea what I read, which makes novel study and Shakespeare really difficult.  I usually fail reading comprehension and not on purpose.  The harder I try to pass those tests, the worse I seem to fail, which explains why I am so scared for the English 30 diploma.  I have a hard time understanding directions, and that is especially evident when we do labs.  I hate labs because I never understand how to do them.  I struggle with understanding concepts in just about every area and it usually takes me longer to grasp a concept.  Sometimes I never understand the concept and we just keep going on.  I excel at Drama, and I\'m pretty good at sports, but Math, English, Social, and Sciences are difficult.  I do manage to do good in most of them, but firstly it\'s cause of my hard work.  More then half of the content in social I didn\'t understand and still don\'t.  I would do poorly on the M.C. tests, and for the long answers I would not be able to understand the content.  What I had to do was write out the answers previous to the test and then memorize them.  When I would do the long answer I would write what I memorized, but then shortly after I would just forget.  I have this bad habit of forgetting names on things, but I think a lot of people do that.  Often when someone tells me something, shortly after I forget what they just said and I feel bad having to ask them again.    In Elementary I did very poorly and even then Math was especially hard and everyone blamed me for not trying hard enough.  I strongly remember that I couldn\'t tell time, but I of course can do that now and I recall trying to copy off of people during a math test on money because I didn\'t know how to count it.  Junior High was also a struggle but somehow I managed to do \"ok.\"  I can do some simple math in my head now, but not a lot, and I still used my fingers to count certain numbers sometimes, but I don\'t make it very apparent cause if people saw me counting fingers they might look at me a little strange and think I am dumb.  Also it\'s not that big of a deal because we use calculators for practically everything nowadays.  I don\'t remember exactly why, but I remember working one-on-one with some lady when I was in grade 2 on math, and I think on math again in gr. 3.  In Elementary they have the resource room, which is were often I would go.  As far as I know I don\'t have a learning disability, and I think if I told someone else that I thought there was maybe something wrong they might think I am crazy or out of my mind.  I also don\'t like to admit it, but sometimes on quizzes and stuff I do need extra time.  I mean it takes me longer to do unit tests then most people, and well when someone asks \"do you need more time\" I usually say \"no\" because I don\'t want to admit it.  I like structured things, lessons, just structured ways of doing and learning things.  I think that is just more of my learning preferences though, and well when I color code my notes and underline stuff in my text I seem to to better.  I think forever this is like always going to puzzle me, just wondering about everything.  If I am right or wrong.  Last year I actually used to think I had some kind of brain disorder or something, but then I realized that IS crazy and like totally stupid, and I came to my senses.  Thinking that I have like learning differences or a learning disability I really don\'t think is stupid.   '),(3846,1807720,'22,March,2004','2004-03-22 00:00:00','    The sleepover with the gals was really fun!  We all had dinner and then we played Cat-opoloy and Truth & Dare Jenga.  Then we watched a bunch of movies and ate and stuff.  Though I was doing really good with the healthy eating, I kind of ate some chocolate this weekend and \"real\" food.  So yeah I gained a little bit back, but I still weigh less then I did before.  Besides tomorrow it\'s back to healthy eating, and I\'m going to try really hard again to loose some weight.  Just cause I blow my \"diet\" once, doesn\'t mean I have to let it drag me down.  Today was nice to have off, though after the gals left I studied Bio for the remainder of the day and then my mom came home, we watched a movie, ate supper, then I dropped off the movie and drove to Safeway to get more fruit.  Did I mention I crave fruit right now.  I am deprived!!  Strangely enough when I eat fruit the weight comes off, eat bread anything \"carbs\" then it just adds up.  Edi and I are going out for dinner on Wednesday night too, cause I am done my English and the Bio test is that day so I most likely won\'t have a lot of homework if any.  Though I am trying to loose weight, I will probably just get a salad.  Besides I will keep things cheap.  It just means that he spends less money.  And rule of thumb: guys are cheap.  My goals at the moment/for future include:   -Get Drivers License -Score 80% or higher on Bio test -Get over 80% in my English novel unit -Loose Weight -Get someone to nominate me for Habits of Mind because I DESERVE it and last semester no one nominated me -Make honors with Distinction -Make a new friend -Be a better friend -NOT get into any arguments -Look sexy -Have lots of fun with Edi -Maybe kiss Edi if it\'s right -Take up a new extra-curricular activity -Have a better attitude about things -I have TOO MANY goals to list!   '),(3847,1807720,'20,March,2004','2004-03-20 00:00:00','    I haven\'t wrote in a while, at least of what I remember.  I have like retired from being at my comp all the time.  I have better things to do now.  Edi and I went to Starbucks for an over two hour date on tues and yesterday night we went out to the movies.  They didn\'t even ID me cause we went to an 18A movie.  It was amazing.  He is so mature and I will admit he is good looking.  We have another date this Wednesday night.  I am excited.  Everytime he comes over he always wants to go eat, but I have always eaten already so this time we planed it!  lol.  Oh and after the movie last night we rented another movie and went back to my place and watched it.  The Bio quiz was slightly difficult I though, but I did my best in really trying to understand that stuff.  It was just \"ok\" though.  I have been driving everywhere and practicing all the time.  It\'s getting easier.  Although I do notice that is has taken me like 3 times as long as most people to get the feel for it, and just to be good at it.  I have to admit I used to find it difficult, but now I find it really fun!  I went shopping today and yesterday.  I got these amazing capris from old navy.  Pink!!  I own pink capris!!!!!!  NO NAVY, NO WHITE!!  AHHA!  And I bought some black sandals and I got a shirt from The Rocket that says: \"bubble cat.\"  It is so cute.  I also bought my bodysuit for underneath my costume for the play and I got the other part of chelseys b-day present.  It is really great and amazing, and I put so much effort into this present.  It\'s like a 3 part thing, but I ain\'t saying nothing here incase she happens to read this.  Did I mention I had a Wendy\'s salad for lunch yesterday and today.  YUM.  And tonight my sister, her fiance, her fiance\'s mother, and her fiance\'s mothers boyfriend are coming to celebrate my sisters b-day even though it\'s really on Monday.  Tomorrow morning I am going to check out the new value village in Midnapore Mall cause I have never been there and I have to practice parking.  Then Sara, Chels, Monet, and maybe Ash are coming over for dinner and then sleeping over.  I don\'t know when I am going to study for Bio, but I will find some time.  I suppose I could do some right now, but I need to relax a little.  I feel like I am just on the go all the time now.  But I kind of like being busy all the time.   '),(3848,1807720,'16,March,2004','2004-03-16 00:00:00','    AHHHHHH NO TIME!!  I am on my computer right now in what seems like a million days.  I have like no time and am like so booked.  I got a date tonight - Edi, and well lots of stuff to do.  I still have to put on my make-up too.  I am sooooo excited!!!  This upcoming weekend is so incredibly booked, I have like NO time.  I don\'t know when I will study and when I will do homework.  I am going to be up til 2:00 AM at this rate!  LOL!  And it\'s officialy my drivers test is booked for Saturday April 10, at 12:30.  I am keeping it on the down lo just in case I don\'t do well, but I have confidence in myself and really think that I can do good.  I know I can do good actually.  Anyways I have to go review some Bio and I have English to do, and make-up to apply.   '),(3849,1807720,'13,March,2004','2004-03-13 00:00:00','    Yesterday night I went to the car & truck show.  AMAZING!!  I love cars!  I also saw Brit, and we haven\'t seen each other since December.  By the way I want the new Malibu!  I also drove to the N.W. today and to south centre, wendys, petland, dollar store, and a whole bunch of places.  Later on I am going to drive back to the N.W. to pick up my Dad.  Tomorrow, we are practicing my parking and I am going to book my road test.  That\'s right, I feel it\'s time.  It\'s time to take my Drivers Test.  I am going to book it for a time over spring break near the end, so the days off I can keep practicing and then do it!  I am so excited!   '),(3850,1807720,'12,March,2004','2004-03-12 00:00:00','    So I guess you could say I have had quite the eventful life during the past couple days.  Yesterday was so busy for me.  In Drama we had a sub, and Bio I don\'t even remember what we did, I think it might have been the quiz that we did yesterday, and a lab.  Yeah I think that\'s it.  Then for active living we went off-campus and went bowling and I learned how to score keep for 10 pin.  It was actually not that hard and I am improving since last time we went, because I usually only do five pin, so 10 pin seems really odd.  Then Swain asked if I wanted to do volunteer hours for the leadership aspect of active living, and at first I didn\'t really want to, but then I said I would do it because I thought \"might as well get it over with.\"  So I helped with admissions at the door from 4:15 PM - about 8:00 PM.  I actually never knew bball games could be that interesting to watch.  Then I went home and watched the most boring episode of Survivor ever and finished my Bio homework.  Today is going pretty good so far.  I am addicted to the Y.M.C.A.\'s low fat muffins and their brown toast.  The food is sooooo good.  I also had the salad in which I have wanted to eat all week - my old classic Friday lunch - The Chicken BLT.  Yummmm.  Then when I got home I cleaned my room cause I was in the mood and well everyone else went to the movies, but my dad is in town for the weekend so I wanted to stay home.  Plus I didn\'t want to see that movie anyway.   '),(3851,1807720,'10,March,2004','2004-03-10 00:00:00','    This morning really sucked.  I don\'t mean to be so negative, but a girl can\'t control what she feels.  In Drama, Carli and I had to do our Helen Keller scene over and over.  Mr. B told us each time what we could change or do differently.  So we tried it again and did something differently, but then he like took me to a corner and gave me more suggestions.  Then he was like \"Vanessa sit in a chair over there and watch.\"  So I had to watch him do the scene.  It made me feel like he thought I was incapable of it or something like that.  It made me feel bad.  Maybe I just took constructive criticism the wrong way.  I mean I don\'t think that he thinks I am incapable, so I probably just took it the wrong way.  Bio was ok, except for the fact that I have difficulty reading graphs.  I mean I don\'t want to seem stupid or dumb, but like when I asked her about it, she explained it and well I still didn\'t understand.  Thank goodness I finished the assignment, finally!  And that is over.  Not to mention my head was pounding again.  I get headaches all the time now.  It was ok thought after I ate some lunch.  I think my body like needs certain foods or something, and if it doesn\'t get it, I get sick.  English was actually really good, and I got tons of work done!  Then in active living ALL the guys wanted ME on their team for floor hockey.  ME!  They were arguing over which team gets \"Work Horse.\"  Some of the guys were like \"whoa she\'s good\" when we were playing, and another one was like \"I commend you for what you can do, cause you\'re really good.\"  Another person was like \"I can\'t get passed her, she blocks my shots before I can even get a chance to shoot.\"  Someone else was like \"Hey she switched positions, she\'s playing offence.  She\'s got the whole package.\"  LOL!!!  OMG all those comments made me feel so good.  They made up for a lack of happiness throughout my day.  What started out as a bad day, turned out to be pretty good afterall!  I feel really good right now and I am ready to review some Bio for the quiz tomorrow.  And today I recall myself saying that I wasn\'t smart.  Well you know what I am a hard worker and I AM SMART.  If I wasn\'t smart, then I wouldn\'t be on the honor roll.  I AM SMART, and I should not put myself so low, as to say that I am not.  Cleary I AM.   '),(3852,1807720,'09,March,2004','2004-03-09 00:00:00','    Today was actually really good.  Drama was fun even though Sam wasn\'t there and she is CRUCIAL to my mark because we have not been able to rehearse our Helen Keller scene that we are doing in like forever.  Everyone seemed to like my skirt today, so it was nice cause everyone kept complimenting me.  Bio was really fun and we did a lab, and it was really interesting.  Mr. Proux I think marked me absent yesterday and he was like \"ok who is your Ta?\" and I was like \"Radomski\" and he was like \"Well tell her to jump in a river because I don\'t make mistakes.  The only mistake I ever made was getting married, but it\'s ok now because I\'ve trained her.\"  LOL!  By the way active living was boring - basketball - that explains it all.   '),(3853,1807720,'08,March,2004','2004-03-08 00:00:00','    My head hurts like a %&$!@ right now!  It\'s been pounding since English.  No one knows how much it sucks to try listen in English when your head feels like it is ready to explode.  I felt sick this morning after eating breakfast, but once I ate lunch I was better.  For A. Living we went off-campus and we got to go bowling which is really fun.  I am sort of upset with the English teacher right now because he was telling us that we should have finished lots of the questions that we had in these booklets that he gave us, and that he said is \"assumed\" most people had finished the book.  Well earth to Mr. Proux but:   I CAN\'T READ MINDS!!!   He never told us what questions to do, and never even mentioned what chapter we were to read up to.  That makes me mad, but whatever.  Things are the way they are.  Finally someone talked to me about just ANYTHING!  You know, other then just normal friends.  Mrs. J is amazing and one of the people I look up to.  We had quite the interesting convo during tutorial.  She is the only person who takes interest in aspects of my life.  A lot of other people do really care, but she cares and she\'s really cool.   '),(3854,1807720,'07,March,2004','2004-03-07 00:00:00','    Today I think I ate more pancakes then I have in the last six months.  I consumed more peanut butter today then I have in an entire year.  You know what else.  I reviewed my Bio - interesting stuff.  Now I want to be a Biochemist if teaching Bio fails, but most likely I will be a teacher.  I also finished my amazing Drama project, which I am certain it\'s my 110% best work.  I also come to a realization today.  I guess you could say a feeling came over me.  I just feel ready.  Feel like I know more now.  I feel smart, or more mature, or older, or I guess you could just say \"ready.\"  That\'s what I feel.  I feel I a ready to drive, to get my license.  I should have waited before doing drivers ed, but whatever, it doesn\'t matter.  My dad flys in on Friday.  I am going to talk to him about takin drivers with a diff company and doing the stuff again.  I think doing the drivers course again would benefit me a lot.  I am almost certain he won\'t have a problem with it.  I all of a sudden want to drive, and be free, and just drive.  Strange feeling.   '),(3855,1807720,'06,March,2004','2004-03-06 00:00:00','    Yesterday I just wanted to rip my phone out of the wall.  I was watching School of Rock and like the phone would ring - I would pick it up.  Then I would sit down and two seconds later - the phone would ring.  This went on for a couple of minutes!!  I wasn\'t exactly feeling up to par yesterday either.  I was a bit under the weather for some odd reason.  All I wanted to do was lie in my bed all day.  I was really tired too and I kept thinking about everything that I am worried about.  I did watch another movie last night too.  I watched The Missing but it was really boring.  I think today I might do some review for Bio and make sure my Drama project is ready to go, because he is such a hard marker.  This morning I kept thinking about Habits of Mind.  Kept wondering why I wasn\'t nominated.  As much as I say to everyone that I am over the fact, the truth is it still bothers me that no one nominated me.  I just don\'t understand why not.  Of course life is not fair, and that is just the way it is going to be.  However, I felt like I put in so much effort last semester.  More effort then I have ever put in.  That is why it leaves me so stumped, confused, and puzzled.  It leaves me in questioning.  Now since I wasn\'t nominated I use tutorials way less these days and I feel as if though no one notices that I work hard.  I\'m sure lots of people do know that I work hard, but I don\'t think they see it or know how hard.  Maybe that is why no one nominated me.  I\'m hoping that Swain will give me a nomination though because I feel like I am working so hard in A. Living and I know that he can see that.    I am not sure as for Bio though because I study a lot at home, and try to understand the material but have not gone to any tutorials.  That is only because I understand the info though, that is why.  I don\'t think she would nominate me again this year though.  I don\'t feel like I have anything special or stand out.  Of course that is slightly being negative towards myself and doubting myself, which I should not do because I am better than that.  The most important thing I learned from LaCour was to not to doubt myself because it\'s demeaning and degrading, if you think you can do something and believe in yourself, it\'s likely anything is possible.   '),(3856,1807720,'06,March,2004','2004-03-06 00:00:00','    After school ended, I went out for lunch with Travis.  I got a taco salad and he go fries and some chicken burger thing.  Then we played lots of XBox games which was nice cause I don\'t own many and plus I hardly ever have time to play.  We also watched Finding Nemo and got Hawaiian pizza for dinner.  Though I was still hungry after so I also had a bowl of cereal and some crackers with peanut butter.  I think I am growing or something, cause lately I have been one to just down food and the numbers on the scale are slightly going down though I keep eating like a hulk.  I miss having Chem, but I know lots of people dropping 30.  Scary, scary, thought.  I guess I am not them though and I shouldn\'t be too concerned if I am not taking chem til\' next year, but of course I am going to be a bit worried.  I am doing a lot better in Bio, then I did in Chem though.  The biochemistry stuff we are doing now is so interesting and it seems simple because I learned lots of the stuff in Chem so it\'s not the difficult.  I want to try do really good on the next quiz and I know I can get an 80% on the unit test.  I want to prove to myself and the world that I am cabable of such things.  I haven\'t failed anything in Bio yet.  I came close on the nitro/phos quiz, but I thought it was extremely hard.  Now that I have been stressed last semester, now that I have done Chem, it seems to be paying off.   Chemistry + Being stressed out -------> A good and hardworking Bio student     My goals right now are:  -Score 100% or close to in on the next Bio quiz, because I CAN -Get 80% on the Bio unit test -Finish memorizing my lines for the play -Try to pass my English reading comp because it\'s hard but I CAN do it -Be a better person, nicer, kinder, wiser -Eat healthy so I can loose some more weight   '),(3857,1807720,'04,March,2004','2004-03-04 00:00:00','    After school ended, I went out for lunch with Travis.  I got a taco salad and he go fries and some chicken burger thing.  Then we played lots of XBox games which was nice cause I don\'t own many and plus I hardly ever have time to play.  We also watched Finding Nemo and got Hawaiian pizza for dinner.  Though I was still hungry after so I also had a bowl of cereal and some crackers with peaunut butter.  I think I am growing or something, cause lately I have been one to just down food and the numbers on the scale are slightly going down though I keep eating like a mongrol.  I miss having Chem, but I know lots of people dropping 30.  Scary, scary, thought.  I guess I am not them though and I shoudln\'t be too concerned if I am not taking chem til\' next year, but of course I am going to be a bit worried.  I am doing alot better in Bio, then I did in Chem though.  The Biochemisty stuff we are doing now is so interesting and it seems simple because I learned lots of the stuff in Chem so it\'s not the difficult.  I want to try do really good on the next quiz and I know I can get an 80% on the unit test.  I want to prove to myself and the world that I am capale of such things.  I haven\'t failed anything in Bio yet.  I came close on the nitro/phos quiz, but I thought it was extremely hard.  Now that I have been stressed last semester, now that I have done Chem, it seems to be paying off.   Chemistry + Being stressed out -------> A good and hardworking Bio student     My goals right now are:  -Score 100% or close to in on the next Bio quiz, because I CAN -Get 80% on the Bio unit test -Finish memorizing my lines for the play -Try to pass my English reading comp because it\'s hard but I CAN do it -Be a better person, nicer, kinder, wiser -Eat healthy so I can loose some more weight   '),(3858,1807720,'04,March,2004','2004-03-04 00:00:00','    My lunch was like a balanced equation today.  Apple + Banana + Hot Dog + Cranberry Loaf  -------->  A semi-balanced/semi-healthy lunch  Ok I ate a lot, but whatever.  Live life a little.  I can\'t always be dieting and staving and trying to look thin all the time.  Maybe genetically I am just not meant to be thin.  In Bio we did a chromatography lab, but our groups really honestly sucked!  We only got like a yellow color, and barely any orange.  She told us that we had to staple our chromatography paper to our sheets when we hand them in too.  I don\'t even know if my group kept ours, cause I am not the one who put the stuff away.  I don\'t think the others in our group know the names of me and Amanda.  Whatever will be, will be I guess.  I still think I did like really amazing on that last quiz though so I will keep my hopes high.  Drama was pretty fun today, like usual, but I only got a 75% on my warm-up.  He just has to be such a hard marker doesn\'t he.  Drama was my thing though, it is my thing.  Something that I was talented at, something that I am talented at.  Thought I don\'t deserve a 75% but there will other things I suppose.  Other opportunities for marks.  I do have to remember that he is a hard marker.  But still, it was better then that.  Life seems just like going through the motions right now.  I mean I talk at school (except in Bio - I\'m like miss queen of silence) and well I laugh, socialize, do plenty of things.  But it just seems like I am going through the motions of everyday life.  That\'s what I fee like ever time I get home.  It\'s hard to explain.  I mean I struggle with a lot of things in my life, but now that Chen is over I don\'t struggle as much with everything.  I guess what I am saying that that before life was unpredictable.  *Will I fail the next test, or will I pass*  You know that kind of thing.  Now I can basically predict the future - what will happen - what will take place.  You name it.  There are a lot more things I can predict now.  I think I like life more when it\'s just random.  When it throws curves in your path when you least expect it and you have challenge to over come.  It may stress you out, make you cry for hours on and off, but in the end it is all worth it.   '),(3859,1807720,'03,March,2004','2004-03-03 00:00:00','    I swear by the fact that I did really good on that Bio quiz today.  I knew everything on it.  Of course it\'s cause I studied til\' my brain was wired to tap into my stored Biology information hiding somewhere in my mind.  Strangely enough we had a sub for Bio today even though our actual teacher was there.  Some meeting or something she had to go to, but like they should seriously make these meetings outside of class time.  Obviously I don\'t like substitutes, though the lady we had today wasn\'t bad.  She was actually pretty nice.  In Drama, since Megz did her warm-up yesterday she got to choose who got to go today.  Megz chose me, so I did my warm-up and everyone liked it and found it really fun.  And once again, everyone applauded to ME!     I am beginning to really think I like a lot of attention, of course being a Leo that is understandable.   In Active Living we just played basketball, which I find kind of boring.  I can\'t wait til\' we get to new units.  I think on Monday though we are going off campus again, at least that is what I have heard.  Apparently we are suppose to be going bowling, so that should be really fun!  In English we didn\'t do much, and right now we are doing novel study and reading Ethan Frome.   '),(3860,1807720,'02,March,2004','2004-03-02 00:00:00','    Bio I think I totally have things under my belt.  When she was reviewing the stuff from the previous days I knew every single answer.  I studied last night, so that could be why.  Of course I always so study and am going to study more tonight.  However, I know I can do good on that quiz tomorrow.  I know that I know the stuff, I just have to be confident and not bomb this quiz like that nitro/phos quiz.  That was hard.  Though it was cause of the wording that I did bad.  I knew all the info, just didn\'t understand what was being asked.  Drama was fun today and English is so much better now that we are doing novel study.  I\'m reading Ethan Frome.  It\'s actually pretty interesting so far.  We had open gym in Active Living today and about 12 of us played some games of volleyball.  That was my dad.  Nothing too exciting I suppose.  I can\'t wait for tomorrow though, I have all this info stored into my brain for tests.  It needs to be unleashed!!  - By the way our theatre is haunted!  Like no joke.  Megz and I were in there working on some scene work and like every couple minutes there would be whistling.  We thought it was just someone on the catwalk, but Mr. B checked and there was no one there.  Lets say I wanted to get out of there fast!  It didn\'t help that most of the theatre was dark because the house lights were off.   '),(3861,1807720,'01,March,2004','2004-03-01 00:00:00','    I guess today was good.  I didn\'t talk a lot though.  Drama was fun, but I thought Mr. B was being bit of a touch critic today, but whatev.  I know the Megz and I did a good job.  Bio was alright but at first I thought this handout thing that we got was like really hard.  I guess maybe I shouldn\'t assume so much, because once I tried the stuff and had a better attitude about it, it seemed to be a lot easier.  At lunch Ash and Chels just talked about stupid work ignoring the fact that I don\'t know the people they work with.  So basically I sat there for a very long time not saying a word.  Oh what a great lunch!  NOT!  Anyways English was nice cause we switched to novel study and now I have Proux (I think that\'s how to spell it) and like today was so nice for English cause he actually taught!  Unlike the classes I had before with Schamber.  In Active Living we played boring basketball.  I seriously should of signed up for rugby, should-of, could-of, but didn\'t.  So I will just have to deal with a sport I done like for a while, that I can deal with.    I keep hearing all about chem 30.  It\'s starting to freak me out.  I was talking to Thy and she told me she is going to drop Chem 30 cause she got 35% on the unit test.  Reminder:  Thy had a higher average than me in Chem 20.  Ok so thinking about that....I\'M SCARED.  It doesn\'t matter if I don\'t have Chem 30 now, cause next year I will.  So yeah I have heard LOTS of stories about chem.  I suppose I should worry about what I have to deal with now actually.  I\'m going to review my Bio some more tonight.  I am very determined to do good.   '),(3862,1807720,'26,April,2004','2004-04-26 00:00:00','    Play rehearsals are tiring me out.  We went from 4:00 - 6:30 today and on Wednesday we start the 4:00 - 7:00 rehearsals and other days we have some late night rehearsals from 7:00 - 900.  Not to mention that I have to miss three days next week to attend the High School Drama Festival at the U of C.  I am never sick and like I don\'t know how to catch up on missed work because I never really miss school.  I am going to have to find out the Bio work we have to do.  I don\'t want to miss Bio and at this point and I can\'t afford to.  I did bad on the last test - which yes I am still like killing myself over cause I didn\'t want to do that bad.  I can\'t even review the damn thing cause not everyone has written it and our Bio teacher is organizing the grad stuff or whatever it is.  Grrrrrrrrrr.  So frustrating.  Just goes back to my statement that no one ever has any time for anything.  I am still very sick of everyone condemning me for not understanding something or just plainly not knowing the answer.  Like leonard today I was like \"yeah we have a lab in Bio today\" and like that is ALL I said, nothing more and you know what he says to me.  He says \"omg this is the easiest shit ever, that isn\'t anything freakin\' hard Vanessa.\"  Like there we go again, I didn\'t even say it was hard, but then again making me feel stupid seems like the most popular game to play.  Even Brent asked me a question in Bio and I don\'t know what everyone seems to expect me to have ever damn answer.  When I told him \"I don\'t know\" all he says is \"I thought you were suppose to be smart.\"   Hmmm so here is the message I seem to be getting: No one ever has anytime, people feel good about themselves when I do bad on things, and I happen to be dumb.  Ok so we both know that isn\'t true but sometimes after hearing the repeptitive negativity towards me I begin to think it is.  But it is not.  I can\'t let that get to my head.  I need positive self-talk.  If I learned any lesson from Chem last term it\'s to believe in yourself.  To think: VANESSA YOU CAN DO IT.  And yes I can do it, I can do anything, I am capable of anything.  That is what I must say to myself.  Deep down I know it to be true.  I am the girl with so many dreams.  Dreams of going to University and majoring in a science like Bio and becoming the most inspiring wonderful teacher.  I have the ability to pass Bio and do good in Bio.  For some unexplainable reason like so many things seem so difficult that cry over all the time, but I have got to where I am right now somehow.  I am here right now, on Honors w. Distinction.  I must have done something right.  I suppose that something might be hard work, effort, work ethic, whatever you want to call it.  I shouldn\'t listen to what others say *unless it\'s positive, lol* because I am so much better than that.  I am better to call others dumb or stupid or tell them that everything should be easy for everyone.  If anyone knows how difficult things can get at times that would be me.   '),(3863,1807720,'25,April,2004','2004-04-25 00:00:00','    Call me the carb queen because that\'s what I am.  All I did was eat carbs and chocolate today.  Ok so I confess I ate junk all weekend.  Chocolate here and chocolate there but tomorrow I am going to eat really healthy and try keep it up for the entire week.  Also there is so really good news.  We went to the Saturn dealership today just to look around at some cars in the lot and next time my dad comes in we are going back to the dealership and I am pretty sure he is trading in our van and getting the L300 series for him and my Mom and buying me a used.  We saw a REALLY really nice 2001 LS-200 today which was like so nice and my dad checked up on it and said he really liked it and stuff.  NO ONE BETTER BUY THAT CAR OR ELSE.  It\'s my car...!!!  It is just like the car I drove for drivers ed (Is that like totally freaky or what?!?) and well it\'s the same color too.  It\'s like a really nice light blue color.  I love it!  Hardly anyone has cars like that and the interior is great.  And since I drove a car like it for Drivers Ed, I know where most of the things are in the car.  I want it so bad but my dad doesn\'t come in for another two weeks.  He is heading back to Vacouver tonight and it\'s his b-day tomorrow.  My mom also said that if he buys me a car then she is going to buy me a cell phone.  I was in a state of shock when she told me this.  I mean I have asked and asked and asked to the point on no return and now she is practically offering me one.  I guess she is right that I do need one.  She needs to be able to contact me at times and in case I need to call CAA/AMA then I am going to need it.  Like \"YES!\" if I get a a car then it is guaranteed that I will get a cell and she is paying the bill though I can\'t rack up charges but that is not something I would do on purpose anyways.  I still have to come up with a play for how to pay for gas but I do have money saved in the bank for a while that should last me a while.  My dad asked \"If I buy you a car and everything, Melissa payed for gas, so I would expect you to pay for gas too, so would you buy your own gas.\"  And of course knowing how badly I want a car I said \"yes\" because I have a lot of money now but depending on how much I use the car, gas could become quite expensive.  I don\'t have a job for incoming money either.  I know long ago I used to work at Perkins but I couldn\'t keep up with the demands of school and work.  So now when I apply for jobs I put I am only available from Friday to Sunday.  No one ever calls me back when I apply places that I actually do want to work.  There is nothing wrong with my resume.  Employers just think that your only life should be work.  No offence though I am not one to go pick and choose.  I just need a BALANCE of school and work, though to employers the word \'balance\' is almost unheard of.  All they seem to want these days is someone who is available 24/7.  They don\'t care how much experience you have, how qualified you are, or nice you are.  Plain and simple, they are going to hire the people that have a wider range of availability than those that don\'t.  And well if actually wanting some time to focus on school and continue to do well is interfering with getting hired or a job then whatever.  It\'s just sad that people put more emphasize on stuff like that, than actually caring about the things that really matter.  So to sum things up:  My dad bought me a laptop, I might get a car (though eventually forsure get one), and then a cell phone.  Hmmm I begged for all these things before and the answer was always no, no, no.  Slowing it changed to maybe, maybe, maybe.  Not it\'s a probably, probably, probably.  I have stopped bagging, stopped asking.  I am slowly but surely getting what I want.  I suddenly feel very lucky.  And I am very lucky.   '),(3864,1807720,'24,April,2004','2004-04-24 00:00:00','    More problems seem to arise today.  More so within my head.  My thoughts seem to be taking over or something cause it\'s really affecting the way I am feeling.  Flames lost today but whatev, the fight Iginla had was entertaining to watch and made up for their boring loss.  I keep eating like all the food in the damn house.  Maybe cause it seems to make up for all the damn feelings I have.  You know I do want to loose more weight, but then once I loose some I decide I am happy the way I am.  So then I eat what I want and then I get scared that I am going to gain weight.  I don\'t want to say I am obsessed, but really I think that is closer to the truth.  I weight myself AT LEAST 3 times on weekdays and AT LEAST 5 times a day during weekends.  I can\'t stop.  And my mom always buys a whole bunch of food that I really don\'t need and well I always go out to eat.  Actually now that I think of it, Thursday I went to Boston Pizza, Friday I went to moxie\'s before rehearsal (I have lots to say about how the play is going so far) and it\'s my Dad\'s b-day on Monday but he won\'t be in town because he is flying back to Vancouver so I am also going out to eat tonight.  Not to mention that I ate practically anything and everything that has the word \'chocolate\' in it in my freezer, we also had like ten thousand doughnuts during rehearsal on Friday.  So I would have the say that the play is like the highlight of my life right now.  I honestly wish I had signed for Soccer or something.  I miss playing it all the time and I am always stuck inside my lonely boring house.  I feel so anti-social like I have no friends.  I need to do something active - NEED...!  *Not a want* My mom is always working and every time that my dad flys in from Vancouver he always goes to Ben and Melissa\'s to work on their new house.  I feel so abandoned if you know what I mean.  I know my parents do love me, but they never have any time for anything.  No time to just talk and stuff.  I even have some doubts about who are really my friends.  You know during the Thursday of TAPS (by the way I made honors W. Distinction cause I worked my ass off) well see I have invited everyone but Travis to my house for lunch (Trav and I got in a fight therefore I didn\'t invite him) and so at first they were all like \"yeah we will come\" and then strangely the only people who came over were Chels and Sarah.  I mean I was glad that I didn\'t have to cook and clean as much but that really makes me wonder.  What was more important than me.  Apparently they were going to go to Extreme Pita for lunch.  That makes no sense.  I mean I have the vehicle and I could of driven them all and lunch would have been free.  That makes me feel like Chels and Sarah are my only true older friends.  I am becoming much closer to those that came from other J.H. schools then the one I came.  I feel as though older friends aren\'t so much like the people I would most likely be lifelong friends with.  Infact I still keep in touch with Brit that goes to Scarlett and well Megz and I declared that we will rule in Drama-Queenism 4ev :) and well I like Lindsey, Allyse, Margaret, Maria, Jess, Brit, Kenda, Bianca, Kirrie, Palavi, Erin, Kristen, and even Carli, Alison, and Jason and so many more people.  I mean I still like my old friends (most of them) because I have known them for so long but I think I am a very different person.  I don\'t really enjoy lunch hours with them either.  It\'s not fun or cool, it\'s just boring and I honestly find them immature.  I feel that I am like ten steps above em.  I mean Ash is nice, but honestly sometimes acts and dresses like more of guy.  By all means though that is just her so I can\'t really change that, but I think Ash and I will be really good friends for a long time because we both thrive off of cheering the Flames and thriving off Soccer.  Of course Chels and I forever will be friends, I mean that is a given.  I saw her at Safeway this morning and actually gave her a ride home.  And well possibly friends with Sarah.  I do like the rest of the people but I feel as though as high School progresses we are just growing more apart and more now than ever we seem very different.  Ah yes I also forgot (don\'t know how I could forget this but I did) what Leonard said to me about Bio.  You know maybe that is why I am like beating myself up for doing bad on it.  Well first of all I DO feel stupid because it was suppose to be like the easiest unit and like it was my worst.  That is also why it was so difficult to ask is she could go over it with me sometime, but I know that if I know now what I did wrong then I won\'t make the same mistake and if I want to make a career out of Bio I have to take it very seriously.  So anyways Leonard walks into T.A. one day and I was just looking over some Bio before the test and he was like \"What are you doing\" and I was like \"Just reviewing my Bio so I don\'t screw anything up and I really want to do good on the test.\"  Then he went ahead and said \"OMG I can\'t believe your actually going over that stuff, it\'s so easy, it\'s stuff that they talk about on The Magic School Bus.\"  I was so hurt by the comment, actually it still disturbs me because I remember it so damn well.  I\'ve heard it before \'Don\'t listen to those kind of people\' but how do you just \"not-listen.\"  It\'s like everytime I find something hard or difficult that most people find easy everyone has to condemn me and make me feel so low.  It\'s not a very good feeling quite honestly.  I actually remember what Ms. LaCour said to me way back when I was reviewing for the Chem final.  She started off by saying \"You have done so good\" but we both know that I failed ALOT of things, so I was like \"Well I failed a lot, I don\'t exactly consider that good\" and then she said \"Yeah but your work ethic is amazing.\"  That was the first time I felt that something she said truly and deeply came from the heart.  Sometimes being smart isn\'t everything and sometimes people who have to work for things work so hard that whatever their goals are, they always seem to accomplish them one way or another.  I am sure there is so much more that has happened over the past Spring Break and then some, but I guess if I don\'t remember the rest than it doesn\'t really affect me too much.  For Active Living Off Campus we went top rope climbing one day which I honestly hope I don\'t have to do again because heights scare the living day lights out of me.  One day we also went to archery (sp??) which was neat to try but no something I would probably even do again.  I think I will stick to being the amazing soccer player that I am, I don\'t have 14 medals for nothing.  Oh yeah I also forgot about the play.  I guess it\'s going really well.  We sold a lot of tickets already, but not everyone has their lines down pat so that is a little frightening since we start full runs at the end of next week and Wednesday we start the late night rehearsals for we go from 4:00 PM - 7:30 PM.  My entire life is going to be Drama - literally.  He expects so much too.  And well I honestly wish I had more lines in the play, but like I am a faerie so I dance a lot but dance ain\'t really my thing.  I think I am a better actress than dancer but you know what, if he cast me as a faerie then he must have belief in me for that part.  I have belief and confidence in myself too.  Friday\'s rehearsal never seemed to end though and I was so energy drained afterwards.  After Megz, Kirrie, and I went to Moxies for lunch we walked back to the school cause Friday rehearsals begin at 2:30 PM.  Then we rehearsed til like 6:10 PM or maybe even a bit later.  It was a LONG LONG night but I still once everything comes together it will be amazing.  All of our hard work and effort is bound to pay off.  I just really hope that everything with Drama and stuff doesn\'t interfere with my other classes *cough* Bio cause doing good in Bio is higher on my list and things that mean ALOT to me.  My goals, my dreams they are everything to me.  I cannot image doing anything else but teaching.  It is my passion.  ok anyways getting slightly off-topic yeah I just really hope the play goes well and well with the amout of work, time, and effort we are putting it it should be a good show.  Though in May I have so many things coming up.  There is:  May 1: Melissa\'s Mount Royal Grad Dinner May 5-7: High School Drama Festival & Rehearsal (Miss School those days) May 8: Weekend Rehearsal to make sure we are prepared May 10: Showing of a Mid Summer Night\'s Dream to J.H. Drama classes (Miss afternoon classes) May 11-14: Mid Summer Night\'s Dream Showings (No life for four nights but Drama) May 18: Drama performances at Father Whelihan (Miss afternoon classes) May 29: Melissa\'s Wedding  Ok wow, busy huh??  And between all that I some how have to fit time in for rehearsals for the play, and doing homework and studying.  When my dad comes in another weekend he is also taking me a whole bunch of places because we are looking for a car to buy me.  I think most of the things I need to get off my chest are free now.   '),(3865,1807720,'23,April,2004','2004-04-23 00:00:00','    Thank God my dad bought me a laptop cause my comp. just decided it would shut down because you know when I needed the most to write in my journal that just had to happen.  Things have been really rough lately.  No one has any time to listen to my problems.  I cried so much over Bio last night while trying to watch the hockey game.  I did really bad on my Bio test, at least for wanting to major in it.  And so then I wonder and think, and really do feel that Bio is the last thing that I totally don\'t completely suck at.  Worst of all I heard her say that is was the easiest until.  Well easiest for who???  I thought it was the most difficult and did the worst in it, so that especially makes me feel really stupid.  People that admitted they didn\'t study were like \"oh I\'m really surprised I got an 80.\"  IT\'S NOT FAIR!  I studied so hard, I mean I always study alot and I knew the concept questions really well.  They always say that if you can answer the concept questions then you will do good on the test.  Well yeah right...did I do the questions - yes, did I do good on the test - no.  See what I mean.  I feel like I did my best.  When I do tests I put my best effort forward and I just feel like I put 110% in the first time and I can\'t do better.  When I tell people that they just tell me that I am being too hard on myself and too negative.  They don\'t understand.  I SERIOUSLY DID MY VERY BEST, and the same goes for reading comprehension tests in English.  Everytime I really believe in my mind that this time I can pass, I always end up failing.  I mean of course I feel like I can\'t do reading comp very well because I just can\'t!!!  People always tell me don\'t have this attitude, but they have no idea how much I study for Bio and how important it is to do good in English - yet no matter the circumstances I still seem to screw things up.  I did however ask if there was a re-write in Bio and she said that possibly at the end.  And instead of just ignoring the test I did really I did really bad on, I am going to go over it with her hopefully sometime near the end of next week.  If I figure out where I am going wrong now then I think university might seem a little bit easier.  I also think I could of done better on the test in my comp had been working.  My computer is my outlet.  The one think that has to accept what I say and write no matter what.  The one thing that has no choice but to listen.  The one friend who is always there.  Everyone else works, or goes out, or is never home, or doesn\'t want to listen to me.  There is more I have to write.  I have been letting all my feelings and emotions build during the time that I didn\'t have internet.  Now I have so much to get off of my chest.  I am slowly starting to feel better already.  I have more to write but I still want to explore some of the new features of my new laptop so I -will continue with the rest of the very long \'What-has-happened-and-what-I-have-been-feeling-over-the-past-week-and-then-some\" journal entry.   '),(3866,1807720,'12,April,2004','2004-04-12 00:00:00','    I picked up Chels today and we went to the White Spot for brunch for her B-day and she opened her presents and she really liked them.  Then we were going to go see a movie but my Dad needed the van so we just went to the mall for a while and then I drove Chels home.  It was neat today and really cool.  I feel so independent.  It was just Chels and I in the vehicle - It was cool and I drove by myself for a while too like before I picked her up and just after I dropped her off.  I love driving.  It\'s so much fun!!!  Yesterday I also ate hard turkey!  LOL.  See my Mom was cooking it and she put in the thermometer that measures turkey and stuff and like it was way lower than it should be to be done but we had been waiting til seven at night so my Mom just cut off the top layers and put it in the microwave for like four minutes.  My turkey was hard!  It was good though and well at least it was food!  Also I love that song by Lindsey Lohan (Drama Queen/That Girl) It\'s sooooo good!!  She is like one of my newest role models.  So basically my main ones are:  Britney Spears, Hilary Duff, Lindsey Lohan, and some other people that I have met throughout my lifetime.   '),(3867,1807720,'10,April,2004','2004-04-10 00:00:00','    I PASSED!!!  Yeah ok so that wasn\'t that bad and now I can drive on my own.  This is great.  See now I can pick Chels up to go out for Breakfast on Monday and we can go alone!  I feel independent.  For once I did something right the first time.  Though it took my 3 tries to get my parallel perfect.  The first time I was too far, the 2nd I was too close and I hit the curb so I lost some points there, and the last time it was perfect.  I didn\'t even have time to merge because it took forever to parallel but I still passed and I only got 40 points deducted.  Tonight I get to go out for Chinese!!!!  OMG that is all I have been doing - EATING!!  But whatever, I look good and besides I am on top of the world right now.  Passed my test and pretty sure I made Distinction, hmmm not much more I want cept a good mark on the next Bio test!  LOL!   '),(3868,1807720,'09,April,2004','2004-04-09 00:00:00','    I am sick.  Woke up today and felt very sick, and now I sound sick.  Sick again during holidays, like that\'s nothing new.  It\'s official, I dumped Edi but I mean we are still going to go out and still going to be friends.  It comes down to the fact that we have a bit of an attraction for each other, but we are very different.  Our lifestyles and just what we like to do, what we would prefer, and just the way we were brought up.  Everyone is different, but there is a point to draw the line.  My Grandma is coming in from Winnipeg today and she is going be like \"oh poor you, your sick\" and I don\'t like it when people feel sorry for me when I\'m sick!  And then wait til\' she finds out someone stole my wallet.  Luckily my YMCA membership expires at the end of the month, and I had not renewed my library card, I don\'t care so much about the money but I had more than usual in there cause I was saving it up and was going to put it in the bank when I got the chance.  I got a new learners and my road test is tomorrow and *PLEASE GOD, I PRAY TO YOU HELP ME TO PASS* and I bought a new school ID, and went to the bank and got a new card issued and all that is left is to get a new S.I.N. card.  I didn\'t realize how important my wallet was to me until it was gone.  And Megz had her Cd\'s stolen - one of them I know she said was Lord of the Dance.  OK, so like right there that limits it down to a girl cause honestly what kind of guy would want to steal a Lord of the Dance Cd, like really?!?!  I think it might have been Jenna.  I don\'t want to blame or accuse, but everytime she shows up and then leaves something is always gone!!!  Hmmmm, coincindence - I think not!  And she does Celtic dance, so that would explain why she would want to steal Lord of the Dance.  I don\'t trust her.  And on Friday during rehearsal she was like \"Yeah I think Mr. B keeps candy in his office\" so she went into his office and got some candy.  I was thinking \'You have no right to go in there!\'  This holiday I am just going to enjoy my food and try to keep under 145.  I keep eating the paska my mom bought and she keeps buying more....ahhhh!  Satuday it\'s Chinese, then monday Chels and I are going to the White Spot for Breakfast/Brunch and maybe a movie if we have time, but if not then the movie will be another day.  Also I am going from 9-12 on Tues. and Wed. to help put together the set for the play and thurs we are getting together to shop for make-up.  I went to the mall today and finally found some zodiac t-shirts and tanks!  I have been wanting one forever!  Also I got a red sleeveles turtleneck tank think with a black skirt for my sisters Mount Royal grad.  And like EVERYTHING is happening in May.  Like there is the: High School Drama Festival, Sisters Grad, Sisters Wedding, J.H. Performance for Drama, Mid Summer Nights Dream running for four nights.  I am going to have like hardly anytime for studying or anything!  It\'s gonna be crazy.  tomorrow I just really hope I pass my drivers test.  I have practiced so much and I just want to do something right for once on the first try.   '),(3869,1807720,'08,April,2004','2004-04-08 00:00:00','    It\'s hard to believe I never seem to make it to the computer anymore.  I never have any time for anything!  I guess that might be a good thing.  My life is exciting, filled with excitement, and I am almost certain that I made it onto Honors w. Distinction.  Most of this week after school I have had rehearsal for the play.  Actually everyday except Wednesday.  However my wallet got stolen with money and all my cards so I have been doing lots of running around and then I had to file a police report to get a case number so I could get my learners cause I have my road test on Saturday.  Lots of running around!  Then next week I think I am going to the school a couple days because the Mid Summer Nights Dream cast is putting together the set and I think Megz, Justine, Jenna, Mandy, and I are going to finish all the dances.  So yeah I am a busy girl and filled with like tons of chocolate right now!!  That is like all I have had all week, but honestly it tastes good and it doesn\'t matter what I look like.   '),(3870,1807720,'03,April,2004','2004-04-03 00:00:00','    Yesterday was like the day of chocolate - seriously!  In the morning they were giving out those hershey kisses so I had some and then in Bio I had a chocolate cookie, then in English Megz was like \"hey, want to share smarties\" so I had some of those, and then when I was leaving school they gave out more chocolate.  OK so yeah you would think that is the end of my lovely I-ate-chocolate-all-day-story but nope it goes on.  So then I got soup and a sandwich at tims and I got the combo that comes with coffee, but I didn\'t want coffee so I got hot chocolate.  Then I don\'t have any idea why I ate dinner last night cause I wasn\'t hungry but I think I was just emotionally eating.  So yeah I ate like yogurt, a granola bar, crackers w/ peanut butter, and a plate of chile.  Yeah ok so that is it right - nope.  Then after I drove to my sisters and stuff and picked up their clothes I drove to timmy\'s cause my mom wanted to get a gift certificate and some tea and so I got a donut - with chocolate on the top.  It was my first time driving through the drive through too!  Then I want through Wendy\'s drive through and got some lemonade and I saw Edi there.  I really have to cut back on the chocolate though.  I was doing so good eating healthy and then my bad habits just creeped back.  I can just try again though.  It\'s been hard with the rehearsals for the play though because I have been eating bigger lunches cause we got during the week 4:00-6:00 and friday 2:30 - 4:30, next thurs we get off early and so I think he said we are going from 12:30 - 2:30.  Then Edi came over after work and we talked for like four hours.  Everything was going good until his kiss.  I mean c\'mon it was first kiss.  To be honest it was not enjoyable at all.  It was totally disgusting.  I expected a little kiss on the lips, but he like sucked the bloodly life out of me.  When he started kissing me her was like trying to like get his tongue in my mouth or something.  It was nasty.  I had a vision of a reptile kissing me.  I called it \"the lizard kiss\" cause it literally was like kissing some lizard trying to like get my tounge.  I finished my homework this morning, yay!  I actually had homework this weekend, but this morning I actually really had nothing to do.  So I finished my Bio, did Drama, and I am done my English though it is not due.  All I have to do before Wednesday is know my lines for our bonus scene we are doing for Drama.  Tonight I am going to Mat\'s house for a get together thing for his b-day with a bunch of my friends and we are having dinner and probably watching movie\'s and stuff tonight.   '),(3871,1807720,'01,April,2004','2004-04-01 00:00:00','    Well today was better than I expected.  Afterall the Bio quiz wasn\'t as scary as I thought, and I am pretty sure that I did good on it.  Although the little experiment thing we did I found like really confusing, so I had to ask Brent after class about it cause I had no idea what was going on.  After like 5 explanations from him I finally understood it, and afterall it wasn\'t as hard as it seemed.  That is why I didn\'t ask the teacher in the first place.  Sometimes I feel that I am just annoying, always asking questions, though maybe I am just super paranoid.  English was good.  We finished watched Good Morning Vietnam and now we are watched Billy Elliot for film study.  Though I am kind of worried cause on the ucourses site my average shows up as 0% with a comment mark entered.  Everyone knows I don\'t have 0 and when I asked my base teacher what I got she told me that I had an 84%, and well prouxl told me that something about the English is messed up because on another girls it said she had 15%.  Yeah, there is definitely something wrong.  Active living was amazing today.  We did social/ballroom dance today.  Yesterday we did break dance and I got to go in the break dancing circle!  It was soooo much fun!  Anyways today I swear it was like guys fighting over me.  I would start dancing with one, and Swain said that if a guy tapped the other on the shoulder, they had to kindly exchange partners.  Guys kept trying to dance with me, and this one creep kept trying to come back, but whatever.  Maybe he liked me or something, I suppose there is really nothing wrong with having a crush on a someone.  Also Swain pulled me over and was like \"I want to show you something\" and I was like \"oh goodness what are you going to do?\"  Then we had to like hook arms and \"sway\" into a certain position ready for the dance.  I did it once and then he was like \"ok now if you could just flow a bit more and loosen up\"  OMG, it was like so odd, but I like dancing with him.  And ok I can officially say that I danced with the gym teacher.  So I guess I am entitled to think a teacher is hot right...cause like damn Swain is an amazing dancer and he is hot.  Shhh....!  No telling, that is my little secret.  After school we did a read through of the play during rehearsal and then I went home, ate some Chile and did my Bio homework.  How exciting!  LOL.   '),(3872,1807720,'26,May,2004','2004-05-26 00:00:00','    Oh how I wish I did better on my Bio test.  I thought I got like an 80% and in my mind that is still what I KNOW I deserve.  I mean 73% isn\'t that bad.  It wouldn\'t be bad if I didn\'t study, but I studied like it was a matter of life and death for this test.  I want so badly to raise my Bio mark.  It\'s so frustrating just trying and trying.  People who don\'t even study score like what I score on tests.  It\'s not fair, not that life ever was but WHY WHY WHY?!?!?!?!  I want to major in Bio.  Can\'t God make me great, amazing, and talented in Bio.  Just something other than Drama, please!!!!  Sometimes I just feel like my effort goes to no where and it makes me question why I study in the first place.  It\'s cause of my goals and dreams, the places I want to go.  It\'s because I like Bio and I am going to be a Bio teacher no matter what.  The world knows I have the work ethic.  I certainly have that, it\'s just higher marks would be nice.  Sppp...Hear that God...I would greatly appreciate higher marks and I am doing everything possible and in my power right now to try and achieve.  I bought a shirt that says Calgary Flames on it today.  Mat and I went to the mall and got lunch then shopped.  It was really fun actually!  We tried to go out twice before and it ended bad, but today was really good.  It was calming and just really fun to have someone to do things with!  And now that I have my car it\'s even better and I am picking him up tomorrow!  I am also really busy and everything coming up.  Tomorrow I am getting a pedicure and it\'s my sisters wedding rehearsal tomorrow so I have to miss the game, but o well I will live.  Although I used to hate hockey, in fact I had NEVER watched hockey until the Vancouver vs. Calgary serious and then I got hooked.  I LOVE hockey now and the flames kick ass!!!  I love them and Jarome Iginla is like the coolest ever!  Anyways, on Friday I am getting a manicure and then we are having a cast dinner at Boston Pizza and then later we are going back to Megz house.  That will be soooooooo fun!!  Then Saturday is my sisters big day - getting married!  So in the morning we are doing our hair and make-up and we have to set up the hall and that is basically my Saturday gone.  Sunday is the opening of the gifts and when I get home I am going out with Chels.  Ok so that is my weekend!  Booked solid!  Yeah what is new - just hope there is no upcoming quizzes in Bio cause I don\'t really have a lot of time right now and my ultimate goal is an 80% in Bio, that is what I want.   '),(3873,1807720,'23,May,2004','2004-05-23 00:00:00','    Things are crazy, life is crazy!  Ok so like I think I am sick, my mom cracker her tooth down the middle, and now today me Dad flew in yesterday and was helping Ben with the house and he fell and broke his ankle.  Ok so yeah that wouldn\'t be that bad on \"NORMAL\" circumstances, but my sisters wedding is this upcoming Saturday.  So yeah my Dad will have to like hobble down the aisle instead of walk my sister down it.  He is more of a pain in the ass that sits and watches T.V. now - certainly not much of a help.  I am also really proud of myself right now.  I studied so much for my Bio, I did all the concept questions as I went through each chapter and I have read them more than once.  I plan to read them again maybe later tonight, but I really feel like I know the info so well.  I always study, but I have never felt this way before.  I feel as though I could be ready to write it tomorrow and you know what I probably am.  However, I will use the one day extra to my advantage.  I might as well, more review.  I mean no matter what I score on this test, I really know I put my 150% into it.  That is all I can do.  I know what I want, I know where I want to go, and I know what I need to do to get there.  So bring on the Bio, yeah cause I can handle it!  OH and here are some pictures from the play:   '),(3874,1807720,'21,May,2004','2004-05-21 00:00:00','    I think I am beginning to get sick.  Of course I haven\'t been feeling all that well and stuff and well am anticipating the results from the blood test cause I am almost convinced that something is wrong.  And well I did good on my Bio quiz!  YAY!  I just really need to do good on the unit test on Tuesday cause the last two I didn\'t do so well.  I don\'t know why though.  I studied for both, read over the chapter(s), and went over the concept objectives.  I know that I know that stuff.  It is so frustrating!!!!!  I just want to be good at something, I want to be good at Bio.  I don\'t just always want to be the hard working girl that is good at Drama and fit for a stage.  I don\'t just want to be good at Drama, I want to be at something that will get me somewhere in life.  I am trying though and I KNOW that no matter what happens I am passing Bio, but I want to do better.  So frustrating but I guess that is part of the process I have to do through to get to where I want to be later in life.  After school today we striked the set for the play, and it only took us an hour to get everything down.  Then I went home, ate a muffin and a half, two slices of raisin bread with jam, and about 7 crackers with jam and water.  That was really good!  Next Friday Mr. B is taking us all out for dinner at B.P. and I am also getting my nails done that afternoon.  It\'s hard to believe my sisters wedding is one week away from Saturday and the rehearsal is on Thursday.  Time flies by so quickly.  I wrote this poem a while ago, just thinking about things in the past.  It\'s called Beside me you Will Stand.       Beside Me You Will Stand   With time I\'ve come to grow and learn and in time I\'ve come to understand There are so many amazing teachers, friends, and people that can lend a helping hand When I thought that I just \"can\'t\" I was reminded that I \"can\" I realize now I am capable of anything and beside me you will stand  There were times when I thought I couldn\'t continue on Times that saw my life as short instead of extraordinarily long And though I must admit sometimes I felt as though I couldn\'t carry on There were friends and teachers there who saw it as duty to care I Would have felt to be all alone but I know there was always someone there  I cried so many nights thinking that \"I can\'t\" So many remarkable teachers reminded me that \"I can\" And there are those who I could never thank enough  For how many times you asked how things were going When school seemed just too much, when life seemed all too tough  Wonderful people that I did meet tell me \"Don\'t be hard on yourself\" At the end of the day I realize the many talents I do posses Though sometimes it takes more than just hard work and effort to gain success At times I have given up and failed, but forgot that there will always be another test Whether a test from God, test during school, or a test from friends to prove my cool  With time I\'ve come to grow and learn and in time I\'ve come to understand There are so many amazing teachers, friends, and people that can lend a helping hand When I thought that I just \"can\'t\" I was reminded that I \"can\" I realize now I am capable of anything and beside me you will stand   '),(3875,1807720,'20,May,2004','2004-05-20 00:00:00','    It seems like forever since I actually had time to go on this thing!  I have been busy with the High School Drama Festival which was soo fun, but left me confused in Bio.  And then like a million late night rehearsals for the play which was a HUGE success!!  It was so good!  People were comparing it to university plays and how we have met the standards and on the last night we got a standing ovation.  Tommow we have to take down the set though and well I think the cast party is next Friday, but I have to get my nails done for my sisters wedding.  Today for Bio we went to the Bridlewood creek, which was fun but really wet because of the rain.  And well this one day in Bio we were using one of those machines that measures blood pressure and I thought it was cool so I got my blood pressure taken and it was like really high.  Then I was kind of worried and have been paying attention to things more in detail.  I have also been having like a million muscle spasms in my right arm but I went to the doctor today and said that is normal, but my blood pressure was high and they took my pulse then decided it would be best to do blood work.  So yeah my arm is slightly sore, but not as bad as before.  I am really worried though.  I have lost a bit of weight over the last couple days and am worried because I keep having to eat to feel good.  I eat breakfast - a good breakfast but need something before lunch and usually after lunch I don\'t feel good by 2:00 pm.  That worries me, along with this strange thing that goes on with my hands and toes.  I feel warm everywhere on my body except the fingers and toes are absolutely freezing and like my mom described it like in-depth to the doctor today cause even she was kind of worried cause she is always like \"omg how can you be that cold?\" and like I really don\'t know.  Even on the really hot day that we went to Father Whelihan to do Drama scenes (It went really well and our scences were amazing) even then my fingers were freezing and the temp outside was like over 20 degrees.   '),(3876,1807720,'07,May,2004','2004-05-07 00:00:00','    I am so enegery drained, so worried about when I will review my Bio for the test I missed.  So stressed out right now but for good reasons.  In fact, I have had to take any available time (excpet now cause I am tired and just need to unleash) just to do school work.  I am trying really hard to keep up with the English but right now he expects way too much.  I can\'t keep up with it all, but I am trying.  On Sunday I will have to do an indepth review of my Bio, though we are doing a J.H. performance of the play during tutorial so I will probably have to write it Tuesday.  I think I am going to ask if I can write it 1st thing in the morning, like after 7.  There would be less distractions, I wouldn\'t feel time pressure, and if I am in a room by myself (hopefully) then I would feel less anxiety and more relaxed.  Drama festival was fun.  Too much food, and well I have had to leave early alot for play rehearsal (which I am sick of)  And well Wednesday I did Commedia, yesterday I did \"Unwrapping the Classics\" so basically breaking down Shakesphere, and today I did Musical Theatre - which was the most fun of all, and we even got to sing Jesus Christ Superstar with a dance we learned.  It was sooo much fun!!  Like I said though too much food - today alone I\'ve have p.b. & J sandwich, granola bar, muffin loaf thing, cookie, milk, chinese, pizza, ice cream cone, strawberries and cool whip.  I gained like 3 pounds from yesterday but I look the same, or atleast still feel like me.  Ok I can\'t write anymore, too stressed and I am talking to Brent on Msn and tommorow I have early morning rehearsal.  Rehearsal again, go figure.  I am really tired of it, but it will pay off.  It\'s just really hard to have a life.  My life seems to revolve around Drama more than anything else.   '),(3877,1807720,'01,May,2004','2004-05-01 00:00:00','    Wow I have been soooo busy.  Rehearsals all the time!  I think I spend more time at school then I do at home now.  Our rehearsals have been running from 4:00 and well they are suppose to end at 7:30 but we always end up going to at least 8:00 or 9:00.  Usually we order or bring food.  On thursday we got pizza and yesterday we had chips, cookies, and pop (well I didn\'t have any pop *by the way I have a story bout\' that*) and yeah Megz and I bought a 2 lb bag of animal crackers!  We bought the zoo!  Yeah so Friday\'s rehearsal was good and we got soup from timmys and a yogurt parfait from McDonalds (that is like the ONLY thing I will eat there) for lunch.  Oh and who got 100% on their Bio quiz - ME!!  It was the easiest all year but I still have to do good on the unit test coming up cause my last unit wasn\'t so hot.  Anyways like 100%.  Like I literally studied in the theatre during rehearsal and while watching the game.  Maybe it\'s because I wasn\'t so stressed out.  I usually study for hours on end and don\'t end up doing good.  I find that actually when I just relax and take it easy, the marks seem to come easier.  I guess it makes sense though.  Also the gym riot on Thursday was so much fun.  I was on team white and we came in 3rd, but we weren\'t last because there were four teams.  Team white was pretty pathetic though, because I mean c\'mon you know how hard it is to decorate with the color white when half of the gym is already that color.  We got to do a balloon stomp where everyone got to tie balloons around their ankles and stomp the other teams color.  Matt Clearly did the pie eating contest.  We got to make up a team cheer and like I said decorate.  Then there was a refridgerator box relay where three people on each side had to go under this giant box and to the other side to pick up other people from their team.  Then Chels got taped to the wall with duct tape in 3 minutes and we actually came in 1st for that one.  There were many more.  Also for the grab bag re-lay pulled out a can of Pepsi, but I like never drink pop.  At first I started to chug in really fast but I got like this rush of carbonated drink going down and I had to stop cause I couldn\'t breath and then my stomach LITERALLY felt like it was going to explode.  After I was done I thought I might actually throw it back up because I could feel some muscles or something in my stomach trying to force it back.  So I am sure that wasn\'t exactly too safe to chug pop, but whatever I did it for the team.    Yeah white! Alright! We look like snow We\'re ready to go We\'ve got the flow We\'re white, alright!  Go White!   '),(3878,1807720,'30,June,2004','2004-06-30 00:00:00','    I went to the YMCA again yesterday and worked out for about two hours.  I love working out.  If I had the energy I would stay at the gym all day all the time.  Afterwards I feel so good.  Exercise is like the medicine for happiness I swear.  When I got home my mom and I went to Tim Hortons and got a BLT bagel combo and we also tried those celebration cakes.  They were so good!!  Then we ran a whole bunch or erins and she also showed me how to get to Bishop Grandin.  I can\'t believe school starts again on Friday, but I am happy I signed up for Summer School.  It\'s better than being bored out of your mind.  I also applied at Safeway in Southland cause I haven\'t got any callbacks from all the places I applied.  My resume is awesome, I mean there is nothing wrong there.  I mean I am an honours student and do like a thousand things within the community and school, and still no call back.  Maybe it\'s the hours I request to work.  I mean I can\'t exactly work every hour of my life like other losers out there that don\'t have a life.  I mean I still have to focus on Summer School, that is my main priority and that comes first.  I woke up this morning and it felt so calming.  I had left my window open last night so when I woke up I could hear the rain pouring down and hitting the window and the side of the house.  It was peaceful.  I have to go shower soon and then drive my mom to the Rocky View.  Then basically I have to do everything - dishes, cleaning, feeding cat and bunnies, the list goes on.  I also have to drive her everywhere, but luckily I have a car and can drive now.  I also have to drive to the airport to pick up my Dad.  He is coming home from Vancouver tomorrow.  He is going to live with us again.  It is going to seem so strange.  It used to be strange enough when my dad came back for every second weekend, and now he like is going to be here permenantly.  I guess I will get used to it, but at first it\'s going to seem really strange.  Last night I also went to go see The Notebook.  That was a good movie.  Oh, I also found this really good scholaship I am going to apply for.  I am not going to say what the name is though cause then my friends would be competing against me.  There are like five categories you can chose to enter and I am choosing the Performing Arts Category.  I have to write an essay and stuff, fill out the application form.  Then you can also include four additional pages of evidence, proof, and like examples of what you did or can do.  I have the best idea.  It\'s too bad though that we couldn\'t video tape the play though cause video, cds, etc are allowed.  That is going to be the real competition.  Anyways I am going to write about the experience in the play.  Then for my examples and stuff I am including the program guide, two pic on two pieces of paper to show it, and then I am going to ask Mr. B for a letter of reference.  Except what I am going to do to make mine stand out is that I am going to use this tissue paper kind of art thing I was taught in grade 7.  Instead of just attaching the pics to blank pages I am going to make like a background kind of thing and since we did \"A Mid Summer Night\'s Dream\" I am going to make like moons and starts part of the background.  That is going to be like the theme.  Then on the envelope that I am going to send it in, I am going to do just a little simple art design like that.  It\'s creative, and original.  I don\'t think everyone will do something like it.  I hope that it will make in stand out, and just noticeable.  I believe I have a very good shot at this.  I am going to do my best to get this.   '),(3879,1807720,'28,June,2004','2004-06-28 00:00:00','    K, I mean how in the world can I be 140.  How was I at 134 one day, the next 137, and now 140.  I mean for christ sake I worked out for two friggin hours today!!!!!!!  I mean I ate healthy and everything.  I am hoping that maybe my weight is up because I am building muscle mass cause of exercise or because of the large water intake increase.  I am scared though.  I don\'t want to gain anymore weight.  My Mom and I am going to Timmys for lunch tomorrow and then on Thurs we are going out to Earls with the neighbor\'s and then to Peters Drive In.  That would be nice if I had like super metabolism and like didn\'t eat anything else those days.  Maybe when I go to Summer School the weight will come off.  I just don\'t want to go back in September and be like 10 pounds heavier.  Especially because Leonard said \"Well Vanessa before you came back you were pretty ugly, but now your hot.\"  I mean what if I were to gain back the 50 pounds I once lost.  What would people think of me?  Would their opinions of me change?  The truthful answer to those questions is yes.  You know I just want to be smart and pretty, really that is all I want.  I wish I weren\'t so self-contious, worried about weight and everything, but I mean that is what today\'s world is all about.  It\'s about being thin and hot.  That is all people care about these days, or at least it seems.  I wonder how people can be content with the way they look and the way they are.  I wish I could be, but everytime I step on the scale it\'s higher and makes me want to cry cause I feel like I have lost control over things in my life.  I am TRYING to eat healthy though and going to the gym all the time.  Other than that I don\'t know what to do, maybe I am just going through a temp gain of weight or something now.  I don\'t really know, I just know that I want to be back to the usual 137.  I feel ok with myself and confident at that.  I suppose it\'s not like being 140 is really going to make people say \"omg, she gained weight\" but I mean I have to keep the weight off some how.  It just doesn\'t make any sense though, you would think eating healthy and working out would get me somewhere.  I am more friggin active now then I was during the school days.  It doesn\'t make any sense.  I don\'t want to starve myself and I don\'t think that should be an option.  I am thinking maybe putting away my scale in my closet is a good idea so I don\'t weight myself like 4 times a day.  I need to keep away from it, and not seeing how much I weight might help with my self-concept here, but I can Physically notice changes too.  All the fat is going to my stomach, everytime I eat that is like the major fat storage area.  All my food seems to turn to fat.  My arms feel a lot muscular from weights and my legs are my more shaped.  If only there was something to make the stomach smaller.  Right now I wish I were a celebrity.  They can do anything.  I wish I were right and older.  Then I would get lipo, it\'s not like anyone would know you had it done.  I am scared though, if I can\'t keep the weight off at 16 then how I am going to manage when I have kids or when I am in my 20\'s and 30\'s.  Sometimes I think weight matter more than anything.  Your whole world practically revovles around it.  The thinner you are, the more prettier, which gives you high self-esteem and self-concept, which leads to you living a happy, successful, and productive life.  I wish I wasn\'t so obsessed with stuff like this, but I am and I don\'t know how to cope.  My Mom doesn\'t make it very easy since she is very overweight.  She bring home fast food (not that I eat any of it half of the time) and we always go out to here and there, she like NEEDS to buy junk when we go to movies, and compulsively she eats late.  Not to mention she buys way too much food for the house and there are always too many good options, which makes you try to resist the urge to always eat it.  She always seems to want to go get ice cream too.  I mean once and a while a treat is good, but not like every second week.  It\'s difficult trying to stay my weight, and I am doing my best.  I mean if I went more then 2 hours at the gym, then that would be over doing it, and I don\'t want to stave myself - that\'s just not right.  Well at least I got this off my chest.  In the morning I am going to wake up and just eat my meals, go work out, and NOT step on the scale.  This entire week we will be going out lots, so really there is no point in stepping on the scale anyways.  Once I start summer school their probably won\'t even be any time to eat.  Well that\'s that.  Wish I could be thin, but life isn\'t always fair so whatever comes my way I guess I just have to accept it and deal with it   '),(3880,1807720,'26,June,2004','2004-06-26 00:00:00','    I hit an all time high on the scale - 140.  I suppose going to all those stampede breakfasts and to dairy queen didn\'t really do any wonders.  I went to the gym this morning and am going to eat healthy - really healthy for the rest of the week except Friday cause my Dad is coming in, and I think this time he is stay and going to live with us again.  And well we are going to Peters Drive In on the Friday he comes in.  I did an hour and a half workout of cardio this morning at the gym and about half an hour of weights.  I am eating healthy, working out, I SHOULD be able to loose weight.  I mean I know I heard of what happens when people get older, and they gain weight and stuff but AHHHHHHHHH I am only 16!!!!!!!!!!!  Yesterday Chels and I went to Harry Potter and then got ice cream, though she barely ate anything.  No wonder she is so thin!  And well I went to the stampede breakfast at Walmart and got free food and stuff.  Today I have to vacuum later on, water the garden, and get all the dust out of the bathroom mats.  I have to drive my Mom to the hospital on Wednesday for her surgery and then I get to drive her back.  I think on Tuesday she is also going to show me how to get to Bishop Grandin so I know how to get there for Summer School.  I am excited for Summer School.  I will get to meet new people and well it was be a STRUCTURED Social class, instead of like the independant stuff they try to make us work with at BOB.   '),(3881,1807720,'26,June,2004','2004-06-26 00:00:00','    I went to the Stampede Breakfast at the Somerset/Shawnessy station today.  I went up twice for Breakfast because I got like the smallest pancakes I ever did see in my life and I was still starving and we had got there early so I ate two plates.  Then I saw Chels and her mom and Chels and I went up to the platform and we got to see the major driving the train into the station.  Chels and I tried to huddle by the photographers and the cameras, lol.  When he got off the train it was like endless pictures over and over.  It was really neat to see what it is like to be known and be in the publics eye.  It was kind of cool.  Then he went around shaking hands, taking pictures with kids, and when he came to Chels and I we though he was going to shake our hands, but then he put his hand up and we high-fived him.  It was funny.  My mom said that he probably just wanted to be hip and \"in\" with the young people.  You know what, I think it worked.  After that my mom and I did a bunch of just little stuff, just going here and there.  Then I went to the YMCA for a workout.  I did an hour of cardio, 45 minute walk around the track, and I think about half an hour of weights.  Then I got home, and was hungry. You know I had envisioned eating like an apple for lunch, but as I opened my cupboard I discovered chocolates that weren\'t there before.  It was like calling my name I swear.  Besides, I think I worked out enough that it won\'t really have an effect.  I have managed to stay at 137, but I have shaped up and everything, my arms and my legs I can totally feel and even notice physically now that there is a difference since I started to work out all the time.  And well it\'s almost Stampede time and though I start Summer School soon it\'s like tradition to go to as many Stampede Breakfasts as possible, so really I mean bring on the pancakes and sausages.  That is what Stampede is about!   '),(3882,1807720,'25,June,2004','2004-06-25 00:00:00','    I went to work out yesterday, I went to go work out today.  Do you think I lost any weight....NO.  I stepped on the scale and I am back up to 137.  It doesn\'t make any sense.  I am eating way less then I was before and working out all the time now.  How am I not losing weight.  Maybe I shouldn\'t of got a Wendy\'s salad yesterday with brent, but I mean I didn\'t even finish the damn thing and I took the rest home.  Oh well, I do feel in better shape then before.  I always thought the scale didn\'t lie, but my arms feel like super strong and my legs aren\'t as flabby.  If only my stomach would somehow shrink.  Oh, that would be nice wouldn\'t it.  I hope Chels calls today.  I want to do something, but if she doesn\'t then I will just find some exercise vids. in the basement and workout, or go for another hour long walk as long as it doesn\'t storm like half way through my walk.  It my bike wasn\'t in such bad shape I could probably go out for a ride, but I think the handle bars are still crocked from me dropping that thing a thousand times last summer.  I probably need a new bike cause I practically killed mine, but I don\'t really have the money for that now.  We are like ultimately broke now that I have my Car and my comp, but I mean those are like needs for me.  I went to go see the The Terminal last night.  It was a good movie.  If Chels calls we will probably end up going to see Harry Potter cause we both want to see that and we haven\'t seen it yet.  I am thinking I might go to that opening breakfast at the new station tomorrow.  It\'s not like I have anything else to do and Chels and I am probably going to the Walmart one on Sunday.  Although I am sure pancakes and stuff aren\'t the best choice for trying to loose weight, but whatev, I mean I probably will eat enough that won\'t need lunch anyways.   '),(3883,1807720,'24,June,2004','2004-06-24 00:00:00','    My stomach hurts so much.  I woke up and felt a pain that really I cannot describe.  I would have rather been run over by a truck then feel that.  I had to eat though.  I ate a fruit to hold me over til\' morning.  Maybe I should re-think my eating habits.  They aren\'t really good and they are starting to cause deterioration of my body.  I mean sure I want to be thin, but health is more important.  Anyways I am going back to bed.  I am tired and feel a bit better now.   '),(3884,1807720,'23,June,2004','2004-06-23 00:00:00','    I feel so sick, like I am dying or something.  Everytime I eat I don\'t feel good afterwards.  I can\'t go not eat anything all the time though.  People are going to begin to start to think I am trying to starve myself, but my stomach REALLY REALLY hurts right now.  I barely ate anything for dinner and I lied to my Mom about what I had for lunch because all I ate was a granola bar.  I don\'t want to eat anyways though cause it makes me feel like deadly ill.  Also, my Mom would have freaked out if she knew I just ate a granola bar for lunch, but she doesn\'t know I am not feeling well.  If I don\'t tell her, she won\'t figure it out.  It\'s not like anyone can do anything about it anyways and I just had a blood test done like a month or less ago and everything was fine.  My stomach really hurts right now, but it\'s more painful when I stand up and stuff, and like about an hour or half and hour after I eat is usually the worst times.  When I lie down it\'s not as painful.  I also have to start Summer School soon and I am going to have to go the entire morning there, trying to pay attention and stuff, and it will be so much more difficult if I don\'t feel well.  To be honest I am kind of afraid to eat.  I don\'t want to feel sick and I don\'t want to get fat.  I was 137 a while ago and that\'s usually the weight I try to maintain, and now I am at 134, which I guess isn\'t that much weight lost considering I barely eat these days and I work out like 24/7.  I did about an hour and fifteen minutes at the gym this morning.  It felt good to work out, I feel in my legs there is slowly more muscle building up, and at least at the gym you meet people and there are people to socialize with.  I have also noticed an increase in like my level of happiness, I think the self-esteem meter went up to, but that\'s because when I eat less I feel better about myself.  I feel thinner.  I don\'t actually really care if it\'s unhealthy.  I mean maybe this being \"sick\" or whatever you would call it is a good thing, because I want to be thin and beautiful.  I want to be like everyone else, have a nice body, and just look good.  I don\'t want to be feeling sick though, but whatever, with time I am sure it will go away.   '),(3885,1807720,'22,June,2004','2004-06-22 00:00:00','    Yesterday a whole bunch of us went to Laser Trek and played two games.  It was so much fun!!  I had more fun there then I have in a LONG time and I also got a good workout in that place.  The first game I came in third and the game I came in second.  The top three in both games were all girls - go figure.  We beat out all of the guys.  Then we all went to Fish Creek, but suck at starting fires so we went back to my place and cooked hot dogs and had chips and pop and stuff.  Then we went to the field out by the back of the school and some of us played Frisbee, Basketball, Soccer, or just went for a walk.  Then Chels left for work and the remainder of us went to Boston Pizza for dinner.  I barely ate though, cause again I didn\'t feel good.  When I did eat, like an hour lately I felt like deadly ill again, but that wasn\'t going to stop me from going out with friends and having fun.  After dinner we went to Chapters for a while and then to the movies.  Then I drove the people in my car home and Sara took the others.  It was a very fun day!  This morning I woke up and felt actually pretty good for once, compared to like how I usually feel in the mornings these days.  Then I went to go work out at the YMCA.  I didn\'t want to leave.  I was so relaxing working out and being in a busy, crowded environment.  I didn\'t want to like overdo it or anything, but I went from 8:45-10:00, which well isn\'t too long I guess.  Any longer then I will probably be like aching tomorrow.  When I got home though I felt sick again, though at least this morning I had an appetite for my breakfast.  My weight is slowly dropping though.  I am like not trying to avoid for or anything, it just makes me feel sick.  I suppose everything is fine though, as long as the weight comes down slowly then I don\'t think I have to worry too much.   '),(3886,1807720,'22,June,2004','2004-06-22 00:00:00','    Ok, so I swear there is like something wrong.  I mean I can tell when things aren\'t right.  I had a blood test a while ago and they said everything was fine, but I don\'t think so.  Everytime I eat, I wake up in the morning and feel really sick and extremely full.  Then I go to the washroom like a thousand times and then I am usually fine.  Except the rest of the day I don\'t really feel the need to eat.  It\'s like food is sickening.  I took four bites out of my cereal this morning and felt like my stomach was going to explode.  And I mean that is not like me.  I love food, I love to eat.  This morning it just made me sick, and yesterday I barely ate because I didn\'t feel hungry even though I went for a 45 minute workout and an hour long walk.  Last night I ate like some M & M\'s and I felt extremely full for like over two hours and I mean really eating M & M\'s shouldn\'t make you feel like you just ate like two steak dinners.  I also started to get the pain in my left side again while I was working out.  I was only into the 5th minute of my workout.  When I was talking to the doctor about that, they just said I over did it.  Ok, so how can I over do it is it\'s something that is reoccuring all the time.  I mean it happens when I work out, climb stairs, sometimes it just happens when I am sitting.  Usually the mornings are the worst though cause I can seem to keep the food in.  It\'s like I eat it one day and my body feel the need to just get rid of it ASAP.  I don\'t really know what this is, and if I go the Doctors again they are just going to give me the bullshit speech of how everything is fine.  I remember when my sister had kidney stones they kept telling her everything was ok, but she was like in extreme pain and like a month later they were like \"well maybe it\'s kidney stones.\"  Like...hmmmm...you could of thought of that a little earlier.   '),(3887,1807720,'21,June,2004','2004-06-21 00:00:00','    Ok, so I have to admit I spend a lot of time on my computer but I mean it\'s practically my other best friend besides Chels.  It doesn\'t tell me what\'s right or wrong, doesn\'t insult me, or give opinions, and it doesn\'t talk.  Sounds like something good to me.  Anyways, I was bored, yes bored of the comp so I went for a walk.  I actually went to the old wodden park near by and sat on a swing until I was dead bored so I went for a walk around the school and then around Evergreen and Shawnee and back home again.  Then I had some fruit - the other half of my lunch.  For lunch I only had a granola bar, but when I got back from my walk I thought that maybe not eating like anything isn\'t exactly the best thing.  So I had some fruit, and it was so good actually.  Better than junk food because you don\'t have to feel guilty eating it because it\'s healthy.  I had an oragne, bluberries, and a nectarine.  Yum!  I think my mom is also picking me up a side salad from Wendy\'s.  You know I am tired of hearing about the debate on TV.  That is all everyone always talks about now.  I mean C\'mon Paul Martin really doesn\'t have a chance.  The only people that will vote him are all the French people that don\'t know any better.  All Liberals are good for is wasting money if you ask me.  I wish I were old enough to vote because Conservatives are clearly best for what Canada needs.  Stephen Harper is the best, he is like a family man and everything.  He believes in what is important for everyone.   '),(3888,1807720,'20,June,2004','2004-06-20 00:00:00','    I had seriously the strangest dream last night.  I dreamt that I was pregnant.  Now that is just slightly scary considering I am an teenager.  I mean I don\'t really know what the message of that dream was trying to convey.  It was just weird.  This morning I went to the YMCA to work out, and I feel so good now that I have.  It\'s like seriously some sort of energy booster.  I feel good about myself, healthy, and just happy.  I miss school too, not like that is any surprise.  It was nice to be at the Y, kind of reminded me of school you know.  I also talked to Chelsey today.  I had an amazing talk with her this morning.  It was just like the old conversations we used to have when we were even better best friends then we are now.  No arguing or anything, just someone to talk to about my worries and everything.  God, now I realize why it\'s so important to have a best friend.  They are always there for you.   '),(3889,1807720,'20,June,2004','2004-06-20 00:00:00','    Yeah I was bored so I wrote a poem.  I know I don\'t have school and I am not obligated to do work, but I like poetry.     On the road ahead  Awaiting us is one grand course To where it directs I do not know As though we chose to go our ways What does this opportunity hold?  Cannot judge, nor calculate Where we will be in time Our position in life may be standing Or running against the hands of time   Where then I may find myself Slightly short of being blind To come across myself without a clue With only a insignificant sight in mind 	 These ambitions, these visions of reality They are not undetectable to the human eye Nor imaginable, just complicated to observe At times they are not easy to believe   One day we will stumble onto a path A route that lies further on The one we have bee waiting for With stones, rocks, and pebbles too  We may stagger, we may fall Trip over an obstacle or take off instead of fall Though we do not know what awaits us tomorrow Today we will follow the road ahead  ~By Vanessa     I also over-heard something today.  I don\'t know if I was supposed to know or anything, but my mom was talking about one of my cousins.  One of the sons of my Dad\'s brother and his EX-wife.  Well this cousin she was talking about is a years older than me and I found out that he has problems with learning and has a confirmed learning disability.  This is kind of scary, but kind of makes sense to me.  It makes sense because I think I have one to, it\'s just that no one thinks so.  They don\'t live my life though, so I don\'t know how anyone can speak on behalf of me.  Yes, I pass my classes.  Yes, I put in a lot of effort.  Yes, sometimes I can score good marks, but I also fail just as much as I pass.  I think one of the biggest eye openers was Drivers Ed.  That was one of the most difficult things learning.  I mean I REALLY tried to pass drivers ed, but I failed.  I don\'t think too many people fail, but I am sure that if I did, then others must have to.  I say that I think I might have one though by the unseen struggles I have with math and chem especially, and some aspects of Biology.  Like reading and interpreting graphs, reading comp for English, understanding what math word probs are asking and determining which methods to use, oh there is so much, and I listed more things (a lot more) in a previous post not to long ago.  I guess I try to hide most of my struggles.  Now that it\'s summer, things should be a little bit easier though.  No crying to hours at night and like almost having a heart attack from hyperventilating.  It\'s just now that I found out my cousin does have one, I know they run in families.  Therefore it could be possible that I might too.  Maybe I am just a slow learner though.  I don\'t really want to believe that, but it could be true.  That doesn\'t however explain why I fail sometimes, pass other times, and sometimes no matter what just don\'t understand.  My life is so confusing and the world is equally confusing.   '),(3890,1807720,'19,June,2004','2004-06-19 00:00:00','    Yesterday I babysat Marci\'s two kids - Danika and Megan.  They were like so well behaved it was amazing, and seriously I have never met kids that polite in my life.  We played monopoly, clue, soccer, basketball, then I had to pull them around in a wagon for like half-an-hour which actually was really fun, and we watched Monsters Inc.  Kids are so fun!  I can\'t wait til\' I have kids, but yeah I am just slightly young right now.  I made $40 for the night, so that will pay for Tuesday.  Some of my friends and me are going to laser quest, barbecuing in fish creek, going to Chinook to bowl, having dinner out, and then going to a movie.  We always have like a day of fun once school is over every year, so this is what we decided we are going to do this year.  This morning I went out for Brunch at the Carriage House Inn.  Ashley was working, so I saw her and apologized for missing her party, but I am seeing her Tuesday.  I will also give the gift I bought her on Tuesday to her.  Anyways, we went to brunch with Melissa, Ben, Collen, Martin, my Mother, my Dad (he\'s in town for once in a blue moon) and of course myself.  You know what the retard did at the gas station yesterday...He left EVERYTHING open - the fuel door thing and like the twisty thing that you have to unscrew to fill up the car with gas.  Luckily I went to go wash my car right after that.  While I was washing it, I got around the that side and noticed it open, right after I had filled up.  Errrr, that makes me so mad.  If it was raining my car would have been ruined if water had gotten in.  I mean it doesn\'t take a rocket scientist to know how to fill up a car with gas and it is only common sense that if you have to twist a knob to put the gas in, it\'s likely you also have to CLOSE it as well.  Next time I am just going to go to self serve, even though last time I swear the guys in the gas stating were laughing their ass off because I couldn\'t twist the cap thing off.  So then I had to ask some lady, because that was the first time I had ever filled up with gas.  I can\'t wait for Summer school to start, but on the other hand at least now I have some time to go to the gym and work out.  I haven\'t been for a long time, and I have been eating a lot more than usually because I am trying to get out of my habit of weighing myself all the time.  It\'s almost like an obsession I have.  I must weight myself at least four times a day, if not more.  Twice in the morning and twice at night.  It\'s hard trying not to gain weight though.  There is so much pressure to be thin and look good and everything.  It will be nice to get Social 30 out of the way too.  I knew that I wouldn\'t have time to use the strategies I used this year to get myself through Social.  I ended with an 82% and that is pretty good for not knowing a thing.  I mean all I did for all the long answer tests were write out the answers to the questions they gave us and memorize then word for word.  It doesn\'t matter in the world today whether you understand or not.  Everyone just looks at marks and uses that as an estimate, which in my opinion is stupid.  I mean you can\'t measure what one knows based on marks.  There is to much emphasis put on achieving, and everyone always praises those who are so smart.  Then again I guess I get praised for hard work.  It\'s like being complimented for TRYING to be smart - sometimes I wonder if that should be taken as a compliment of not.  That\'s a tough one.  It\'s hard to believe next year will be grade twelve.  I will be at the TOP of the food chain.  That feels good, but it doesn\'t mean everything.  There is so much pressure to succeed, and I already put numerous amounts of pressure on myself to do good.  I want to get into University so badly.  That is the one goal I am ultimately working towards.  I know that if I get in, I can do that stuff with hard work and well probably going to have to hire a tutor.  I might even need a tutor for next year because no one seems to take me seriously when I saw I don\'t understand my math.  I mean no one really knows just how frustrating it is, going home and just crying for like hours because you really want to understand and you just keep trying and trying, but can\'t seem to get the right answer.  That is why I doubt myself a lot.  I mean really, people expect me to believe in myself, but after trying and failing so many times you would think there would be a time and a place to draw the line of: I can and I can\'t.  If that makes any sense at all.  I probably will need a tutor for Chem 30 next year too, but I really want to try doing it without a whole bunch of tutors.  I want to be able to do the stuff myself you know.  I don\'t want to be dependent or reliable on other people all the time to \'teach me this, and show me that.\'    I hope I get all the hard courses in the first semester for grade twelve year.  I want at least just the Math and Chem.  That way second semester I won\'t be so stressed and will go through my depressive state at the beginning of the year, rather than the end.  I want to enjoy high school, it\'s just not always that easy though.  I\'m so scared of Diploma\'s next year too.  I mean C\'mon, why on Earth would they make them worth 50%.  Some people like me, try and they try, but doesn\'t necessarily dictate how good of a mark I will get.  My tests are my weakness for most subjects.  Like half of the subjects I do good in, is for the reason that I work my ass off for my future goals and dreams that I am NEVER going to give up on, and most of my marks are as high as they are because I do ALL of the homework.  I don\'t think anything is going to change though, it\'s just I don\'t chose to fail half of the tests I do, I mean I REALLY honestly try my best to pass.  I mean what will happen if I fail a Diploma.  Maybe I shouldn\'t be so negative, but I mean you always have to consider all the possibilities.  The University are especially going to pay close attention to my marks in the Science courses I have taken.  The thought of having to do Chemistry courses and Math courses in University scare me a lot.  I am the girl who still doesn\'t even understand how to balance an equation properly and needs to eventually pass University Chem.  I have proven to myself that I can do just about anything I put my mind to.  So when I make it in into University (Which I will because I didn\'t work my ass off for nothing)then I will probably have to face one of the biggest challenges of my life.  Mat and I argued this morning, about Tuesday that we are all getting together and stuff.  We argued about a stupid movie.  He was like \"Oh, we all agreed to see Chronicles of...??Can\'t actually remember anymore\" yeah I forgot what the last word was, but it\'s like some action shit I don\'t want to see.  Then I told him that like why would I pay for a movie I don\'t even want to see.  And he was like \"Well you can\'t always have your way.\"  Well you know what, if I don\'t want to see that fucking movie, then hell I am not going to go see it.  He pisses me off so much something.  I can\'t even believe I am dating him.  So sorry Mat if you somehow happen to find my journal and read this, but I really have to get off all the shit I am feeling so I don\'t end up almost killing myself like the thousand other times I came so close.  He is way too into my looks.  I mean no one was attracted to me until I came back after last summer 50 pounds lighter.  There is this one comment specifically I remember, he said: \"I can\'t wait until I get back from the flight camp because you will have lost more weight.\"  I mean buddy your being TOTALLY unrealistic.  Just because I lost 50 pounds one summer, doesn\'t mean I am going to come back to school again even thinner.  I look just fine (well I admit I would like to loose more, but right now I am not trying cause I have other issues I need to deal with) and like it\'s all about my stupid looks.  Looks are nothing, personality is everything.  The more I get to know him, the more turned off I am.  I miss Aaron, I loved Aaron.  I really hope he comes back for the Summer so I can see him.  He is like my perfect match.  I didn\'t believe in love at first sight until I met him.  Now that was love at first site, then our personalities just complimented each other like leaves in the wind.  I swear Aaron is the one.  Ok, not kidding, not joking, not being unrealistic.  I mean they say there is always someone out there that is meant for you.  With Aaron I feel so loved and accepted, so free to be who I really am.  To him, he loved me the way I was.  He didn\'t think I was too big, or too thin, just right.  We had so much fun together.  I had public affection with just about anyone, but with him it was so different.  I felt good showing public affection with him.  I just wish he didn\'t live so far away.  I mean I can\'t exactly fly to Halifax and I swear last summer it was fate that brought the two of us together.  In fact, the minute he stepped of the west side bus at C.O.P. for an off-trip, my attention drew directly toward him.  He made me feel confident and just wild.  When we all took turns going on the bungee trampoline he was like \"Vanessa, do a flip, you can do it.\"  I was so scared, but I did it for him and it was so much fun.  Then he dumped a bucket of water on me and we had a huge water fight, just the two of us without even knowing each other that much back them.    Nothing in life seems to ever make sense does it.  I mean I seem to have failed myself in so many different areas.  I mean I failed in relationships, I literally failed Drivers Ed, but with so much damn hard work and practice I finally got my license (Some miracle of God, seriously that was so stressful), and well sometimes I feel I failed myself in school.  All the times I really try.  Sometimes I think people underestimate how much effort and work I put in, trying to get a good mark.  That is why it\'s so upsetting to fail.  I guess that is the whole lesson in life though.  We try and we fail, and we try and we fail.  I suppose with failure we understand more about life and the rest of the world.  Maybe the world is trying to send me one big message or trying to teach me some important lesson.  Maybe I am just crazy.  That seems like a good option.  There is so much that is on my mind right now though, I am pretty stressed.  You would think that with the end of school, life wouldn\'t be as hectic, but thinking of the future and what awaits me is a pretty scary thing.   '),(3891,1807720,'19,June,2004','2004-06-19 00:00:00','    Yeah so I woke up this morning, but NOT early.  That\'s right, I slept in!!  Yesterday was so much fun.  Megz, myself, and a bunch of other people came to the Greek Festival.  We got lost like a million times, and yeah I made a couple u-turns.  Yup, I broke the law.  Then a couple assholes wouldn\'t let me in so basically I ALMOST hit them, but instead of clipping them, I just swerved back into my lane.  On out way back, so guy pulling out of a parking lot almost hit me to, but my car is fine and everyone else is too, so that\'s good!   It was so much fun though last night.  We ordered like tons of Greek food, and had lots of Greek desserts and everything.  Then we went to the face painting both and we all got mustaches with goatee\'s painted on.  OMG, it was so funny.  Then the kids that saw us with them painted on our faces, were all like \"lets get mustaches guys, then we saw little girls starting to do it too.\"  I suppose we started a trend, or we are pretty influence and great role models!  Then we went back to my house and we took a picture of us with our sporty mustaches on.  Then I drove everyone back to Megz.  The ride from the Greek Festival on the way home was the best though.  I still had to concentrate on driving, but when we stopped at red lights we all yelled OPA, and you have to remember we looked strange too, with all that stuff painted on our face.  They also had a whole bunch of Greek dancing to watch, and that was really neat.  Not that I am like into Greek a lot or anything, but it\'s like totally different from the Ukrainian culture - foods and dancing.  Anyways, my mom assumes that when I say I like something, it\'s some sort of indication for her to go out and buy like 10 of whatever it is that I said I liked.  Like these yogurt granola bars.  One day I was like \"yum, they are so good.  You know what she did...She bought a box of vanilla, a box of blueberry, and a box of strawberry.  Now she won\'t stop buying them.  I try to eat them quick so I don\'t have to see them anymore, then she just goes while I am sleeping to safeway and comes home with more food that I don\'t need!!  I can tell it\'s the beginning of holidays.  Already I feel kind of like a slave, but I mean I guess doing things for your parents is kind of expected.  My mom wanted to me take the newspapers from out garage to like a newspaper drive they are having at Bishop O\'Byrne.  Ok, so I was kind of upset by her telling me this, but then I just felt like I was doing something to help out the community.  It took forever to load the newspapers into boxes and bags, and then carry them to the car.  Like seriously I swear my mother has been saving since Christmas.  I mean if they world didn\'t have newspapers, they would save a heck of a lot of trees!   '),(3892,1807720,'18,June,2004','2004-06-18 00:00:00','    I got 71% on the Bio final.  I was aiming for 80%, but I thought I like failed.  I am proud of my 71% and my overall average only went down by 1%.  So I am very happy about that because it\'s didn\'t dramatically drop.  I worked my ass off trying to understand everything for Bio and I tried my very best.  71% is the best I can do.  I put my very best effort forth.  Mat and I went to the zoo today.  Our lunch was like the most expensive thing ever!  Tonight I am going to the Greek Festival with Megz and a bunch of my other friends.  It should be fun.  I also am eating what I want to eat now.  I mean I don\'t want to admit it, but I am obsessed with weight, and weighing myself.  I decided I need to get out of the habit, even if that means I gain a couple pounds (but hopefully that won\'t happen) Anyways....today has been great!!  My Bio marks just made my day.  I swear she like rigged the marks cause I thought I did bad.  Maybe my T.A. is right after all.  Maybe I am just like way too hard on myself.  I mean I do expect great things from myself though, I mean there is no doubt about that!  I also wrote this little thing a while ago.  Just about people who inspire me, I look up to, and mean a lot to me:      Making a Difference    Everywhere I go it seems that the message trying to be conveyed to students is that there T.A. is someone who they can rely on, find strength in, and just go to in times of need.  Truly that is what my T.A. is there for and she has been such an amazing support and guide through out my high school years.  However, at the beginning of grade ten my T.A. was away for the first few months.  To count the various substitutes I had gone through is nearly impossible, nor did I feel very comfortable is a new school environment.  There was this one day after tutorial had ended, and our substitute we had for T.A. had asked us to get a card initialed proving we had gone to tutorial.  It was my Science teacher, Mrs. Jagorinec who asked me who my T.A. was.  When I thought about this question, my answer came out as I donâ€™t know.  That was the truth.  I had met so many different people in the following weeks that the names of people became unimportant to me.  Once my Science teacher realized that really I didnâ€™t have a defined T.A. or just someone to talk and rely on, she did her best to make me feel welcome within the school community.  She talked to me, took interest in what I liked, and most importantly she cared.  To her I was more that just the student that sat in the back corner.  I was more than just another student in her Science class.  As the weeks proceeded, Mrs. Jagorinec began to notice my involvement in the school community and the different aspects of extra-curricular activities I took interest in.  These interests of mine became our foundation for weekly discussions.  The more she took interest and the more we communicated, I began to feel more welcome in the school community.  It wasnâ€™t long before I felt that she too was someone I could rely.  I discovered she was a teacher I could turn to for advice, someone that would support me in tough times, and just talk when no one else was around.  A couple months later my T.A. came back permanently and I learned that she was also an amazing person.  She was there to support me with learning and help me with choices, but I wouldnâ€™t take back the way Mrs. Jagorinec had made me feel in the past months.  She helped me to get through those months where I didnâ€™t have a supportive relationship with any other teacher, and didnâ€™t have any one else besides my friends to turn to.  Soon enough she became my favourite teacher and her Science class was also the place I loved to be most.  Not only was she an inspiration, but she was also a role model to me as well.   Since I got to know Mrs. Jagorinec very well in the last months, I began to feel more open when I needed to ask for help.  I didnâ€™t feel pressure to always be perfect in her class and she reminded me that making mistakes was an important step in the learning process.  At the end of grade ten she had mad such an impact on me that I realized I wanted to teach Science.  I thought I may possibly want to teach Biology because I excelled in that area more so than others.  When I reached grade eleven she only had the opportunity to teach one Biology class.  During the second semester I saw her in the hallway and she told me that I was in her Biology class.  I wondered how this could be; because on my original schedule I was suppose to have a different teacher.  It turned out that one Biology class had been switched from another teacher to her, so she had the opportunity to teach me for the second year in a row.    I cannot describe the difference she has made.  Times when she saw me in the hallways and just asked how things we going had value and importance to me.  Just a simple question of howâ€™s it going can make all the difference in whether my day was good or bad, just because she took time to care.  Even when I wasnâ€™t a student in her class, I was an important friend outside of class.  She has inspired me to become a Biology teacher in the future.  I plan to go to University and study Biological Sciences and then obtain my teaching degree to teach secondary education.  Before I had met Mrs. Jagorinec teaching was only a dream, but she has made me realize that becoming a teacher is only a goal away.    This is a poem I wrote about her, because of the difference she has made in my life, and I am certain that she will continue to inspire and touch others students such as myself in the near future.      Thank You  People come and people go But you were so special I want you to know Iâ€™ve never met anyone so kind and so caring So different, all that knowledge you were sharing  I still remember one day walking by You gave me a smile instead of a â€œhiâ€ That feeling inside was something so great A gift that anyone could appreciate  You never failed to ask how things were going Never failed at the compassion you were showing Always proved you wanted to help out Certainly the greatest teacher without a doubt  There is just something special about you You know what to say and you know what to do You helped me with Science and with Biology too Helped me to deal, helped me to get through  A memory of you will remain in my heart Didnâ€™t think you would be so cool right from the start But wonderful people they come and they go The thing in life is that we never really know  So here is a thank-you to all that you have done You have made such an impact on everyone You are an inspiration and a role model too There is nothing about you that I would undo  Thank you     	My high school years have been filled with wonderful memories of the remarkable experiences and the many magnificent people.  With anticipation this wonderful precedent has be set for the generations to come.  Just thought I would write a little essay kinda thing.  Just felt like it you know.  It gives me more time to reflect on everything.   '),(3893,1807720,'17,June,2004','2004-06-17 00:00:00','    Ok, I was reading some stuff online just looking into the possibilities as to why I struggle so much in school.  I came across a very reputable website that described what EXACTLY matches me.  I mean EXACTLY.  It totally describes me, it\'s really scary just how \'dead on\' it is.  Conceptual/Holistic Processing Disability:   When your weakness is in the conceptual/holistic area, you probably can memorize details and facts pretty well but really struggle to understand deeper meaning and general concepts. You may read rather quickly and easily but have difficulty understanding what you have read.     For this type of disability it is most important to take extra time to try to develop a deeper understanding of what you are learning. Although you may finish assignments quickly, you need to spend extra time to figure out why each assignment was important and what you really learned.    The following may be most suitable for you:   1.  Extra instruction  about underlying concepts and deeper meaning. Why is this information important?   2.  Overviews  and summaries to provide context before and after lessons.   3.   Concrete visual aids  (drawings, charts, video, etc.) to help understand the \"big picture\".   4.   Extra time for thinking  through tests and assignments.  5. Textbooks and assignments with  highlighting  of the most important information and concepts.  6.  Hands-on experiences to use your other senses to learn.   7.  Reduced assignments to avoid feeling overwhelmed.   8.   Concrete examples  and/or demonstrations.   9.   Linking information  to real-life situations.   10.  Encouragement  to go beyond superficial understanding.   11.  Clarification and repetition  of concepts.   12.  Outlines  listing key points and concepts.     And here are some things you can do for yourself:   1. Take more time searching for deeper meaning after completing assignments.   2. Pay attention, watching and listening for any important information you may need.   3. Ask for explanation when you don\'t understand concepts or meaning.   4. When reading, stop after each paragraph to see if you understood what you read.   5. Take time to break large assignments into smaller(manageable)pieces   This sounds just like me.  EXACTLY like me.  Especially the part about memorizing the details and not understanding them.  That\'s EXACTLY what I do for social.  When they give us the long answer questions, I write out the answer and then memorize what I am going to write.  This way it seems like I know the information, but really I just memorized it without actually knowing the meaning or understanding.  And the part about finishing assignments quickly, but not understanding is just like me too.  I do the assignments, but I don\'t really learn anything because I just plainly \'answered\' the questions without actually understanding them or going through what they actually mean.   I like concrete examples because I learn from totally structured environments, and often I do need time on tests and stuff like that.  I usually need clarification and ask questions beyond the original information presented - so that one basically hits me dead on.  It\'s so scary how accurate this is.  Also, this might explain why reading comprehension is so bloody difficult.  It\'s so true that I can read, I mean reading is not my problem.  My problem is understanding what I have read, which makes it very difficult to answer comprehension questions.  Maybe I have some sort of processing problem, which is why a lot of things are difficult and school can often be frustrating.   '),(3894,1807720,'16,June,2004','2004-06-16 00:00:00','    My Bio test was today.  I was really nervous before it.  I think I did bad on it.  I wanted to just burst into tears after it.  I feel really stupid too.  I didn\'t understand the ratio question and I think even after I asked I still did it wrong.  I think that question was supposed to be easy, but I found it confusing.  As well, I don\'t think I did so hot on the M.C. part and I got confused on the long answer.  I just feel bad about it.  I do feel a bit better though because I went to say good-bye to my T.A. cause I won\'t see her again.  We had a long talk and stuff about everything, about school, marks, the Bio final.  She made me feel like everything will be ok.  I just hope it turns out that way.  I studied a lot for that final and put in a lot of effort.  I just wish it would pay off you know.  I know I can\'t ace everything, but I really deserve to do good on this.   '),(3895,1807720,'16,June,2004','2004-06-16 00:00:00','    I thought earlier would be my last post, by I can\'t help expressing my feelings right now.  I need a break from studying Bio, or my head is going to explode.  You can only study so much til you get totally stressed and everything.  I was just curious.  As for me, curiosity gets me to feelings I don\'t really want to feel all the time.  I just had to go look at the English marks.  I just had to know so badly.  I got a 64%.  You know what though, I am not really that upset about it.  I put in my 150% effort.  I took my time.  I concentrated.  I believed in myself and in what I was doing.  I used EVERY strategy, yes EVERY strategy I have been taught for reading comprehension.  My average is going to drop, but what counts is that I did my very best.  I guess some people aren\'t English inclined, or Math inclined, or Chemistry inclined....the list goes on, but frankly that would be demeaning to list what I\'m bad at.  I guess I am just the girl who excels at sports, drama, and puts in a lot of effort and tries hard.  That is just me, and I suppose I just have to accept that.  This time I am not going to get upset about a bad mark.  I am going to learn from it.  What doesn\'t kill you makes you stronger.  And you know, my Bio mark has nothing to do with what I got in English.  I very well believe that with my effort, hard work, and amount of prep time I put in for this final I very well have potential to do good.  I am going to try my best and put my best effort forward, that\'s all I can do now.  I\'ve learned the material, I\'ve reviewed, I just have to prove what I know and I have to do my best.   '),(3896,1807720,'15,June,2004','2004-06-15 00:00:00','    This is truly the last time I write in my journal before the Bio final.  I will probably just do some last minute review of things I did a little bit more time to go over.  I feel very prepared though.  Today was good and I went to school and we performed our Drama Final Project.  Then I went to Tim Hortons and got my fav: BlT bagel, iced tea, and a crueler.  I felt really nervous this morning for some reason.  I didn\'t even have a final today, and I was nervous.  I was probably nervous cause I know a whole bunch of people were writing their Chem finals and Diploma\'s today.  Maybe I felt nervous for them.  I think the reality that it\'s actually the time of finals is starting to kick in.  I do know though that I am 100% capable of doing good on Bio.  I know the stuff.  I studied.  I worked hard.  I asked questions.  I know what I need to do and I am going to do it.  Believe in yourself and you will succeed.  Oh and I saw Ms. LaCour in the hall today.  It\'s odd how when I was in her Chem class she would say \"hi\" to me in the hallways.  Now she doesn\'t even say anything.  It\'s like she\'s pretending she\'s invisible or something.  She did smile at me, cause I smiled at her.  Maybe she thinks I don\'t like her or something, but that\'s not that case.  Actually she\'s one of the best teachers because she\'s different and she doesn\'t go changing her style of teaching or whatever to suit your needs.  She has the coolest style - love the clothing.  I actually look up to her as a role model, but shhhhh...that\'s my secret.  No one needs to know that.   '),(3897,1807720,'15,June,2004','2004-06-15 00:00:00','    One down, one to go.  I sit here in front of my computer, thinking to myself and realizing that grade eleven is really over.  After Thursday - that\'s it.  After the Bio final, I will be in grade twelve starting September.  I keep thinking about University more and more these days too.  I am pretty sure I know what I want to do, and I don\'t think it\'s going to change.  Getting into University scares me a lot.  I don\'t think I could handle rejection.  I mean I can barely handle Chem class as it is.  My biggest goals next year is to get into University and get a scholarship - Maybe one of those Habits of Mind Scholarships they give out to grade twelve\'s.  I have a good shot for that, especially since I am like miss-I-work-my-ass-off-for-ALL-of-my-marks-and-always-put-my-best-effort-forth.  I also thought that my English exam went really well today.  I took my time.  I concentrated.  I focused on what I needed to do.  I made a strategy.  I did the knowledge questions first, and then the reading comprehension.  This build up my confidence cause I thought the knowledge questions were easy, so when I got to the reading comp I was more relaxed.  However, I did get a major headache half way through cause the exam is like way too long.  Like sitting there for hours, just thinking and using your brain to ultimate capacity is hard on moi!  I so totally thank Proulx for helping my with reading comp. cause like after that it totally make me believe in myself and my ability to do well.  I also think that I might not necessarily have any type of learning problem like I thought before.  I think that I just get stressed out for tests and maybe like test anxiety.  I need for time for things and just need to be more alert and aware.  Maybe that\'s it.  Maybe that\'s my problem.  Yeah, I also forgot to post this pic.  It is the day of my sisters grad for graduating in the Youth and Child Care Worker program:   '),(3898,1807720,'14,June,2004','2004-06-14 00:00:00','    I have no idea where to begin.  My I-need-to-try-eat-healthy strategy well...I guess I have to try again tomorrow but I am going outs for lunch so that\'s a hard one.  Anyways I really need to stop worrying about weight, weight gain during exam week is allowed guess.  That is the exception to the I-must-eat-healthy-so-I-can-have-a-nice-figure.  I went to school at about 9:00 AM to review my English.  We went over my reading comprehension one-on-one which REALLY helped.  Then I stayed for the review cause he was going over terms that are on the exam tomorrow.  I wanted to do some Bio review too, but by the time the English review was over it was 11:00 AM and that is when the Bio tutorial ended.  I really have a lot of questions to ask and I want to go over a previous test, so I hope like no one shows up for tutorial tomorrow.  After all half of the people on my Bio class aren\'t like keeners like me, or have the work ethic of a grade eleven student.  Tomorrow is the big English final.  I studied lots today, and the day before that, and the day before that.  I studied my Bio extensively too.  I feel like I know it like the back of my head.  hopefully she will like totally help me tomorrow with some of the stuff that I need extra review and I am sure that all will go well during, before, and after the English final.   '),(3899,1807720,'13,June,2004','2004-06-13 00:00:00','    So here\'s the truth.  I gained weight...again.  Not like anyone is going to notice that much, but if I keep it up.  I can\'t help it.  I am addicted to junk, but I am REALLY trying to eat healthy.  I thought my scale was going to like break today.  140.  That is higher than I have been in a LONG while.  I got down to 132.  Then...I ate.  Soon enough back up to 137 where I usually.  Now 140.  I won\'t let it get any higher.  I can\'t deal with that.  After exams I need to go to the gym.  I suppose weight gain weekend before exams actually probably isn\'t that uncommon for most people.  I studied so much today.  I feel like I know my Bio so incredibly well.  I have lots of questions to ask Mrs. J, but tomorrow morning I am going in for English review of comprehension and so I might go do Bio review later, but I will probably just go on Tuesday because I have my English final that day.  Ok so here is my dilemma.  Well actually it\'s like my feeling  inside, feelings I have had for a long time.  When I talk to people about how I feel, they think I am crazy, insane.  I don\'t care about that though, because I know I am not.  I am not stupid.  I also don\'t fail my subjects, but I sometimes (often if it\'s chem) fail quizzes or tests within my subjects.  I think I have some sort of learning problem.  Actually I am almost certain.  I feel different a lot of times.  Different in the way I do things.  Different in the way I learn things.  There are MANY reasons for these thoughts and feelings.  They are valid and significant, sometimes I wish people would pay more attention to them.  I know I am doing my best to learn and doing pretty good, but there are some adjustments that would make my life much less stressful.    why I believe there is a problem:   -I am a good reader, but don\'t understand what I read  -Come close or fail reading comprehension  -When able to take read comp. test alone in a room w/ out time restrictions I can score much higher then usual  -Study MANY hours, sometimes all night and space out study sessions and feel as though my marks don\'t match up with the effort I put in  -Don\'t have clear understanding of concepts, need to use pure memorization to take tests (ex. Social long answer) I memorize the answer to the questions we are given.  I NEVER actually understood the material, it just seems like I do cause of memorization  -Failed drivers ed, couldn\'t remember parts of car or what they do.  Have trouble learning rules of the road and actually driving.  My right-of-way judgment is still bad and the parking (no comment), but the driving is getting better.  -When people tell me things, I don\'t remember unless I write it down  -Some concepts I just can\'t seem to grasp no matter how hard I try and how much effort I put it (ex. Chemistry ice tables, some Math concepts)  -Have trouble understanding explanations in texts books and understand what Math word problems are asking  -When I was younger I could not tell time or count money (Ex. I remember trying to copy off of another girls tests cause I didn\'t know what any of the money represented)  -Take and long time to complete tests, feel rushed cause there is usually not enough time to think things through  -Excel and most sports and Drama related activities, but cannot grasp (some) not all concepts for Math, Bio, Chem, English, Social  -Marks ARE good, but because of effort and because I do ALL of my homework, which makes up for the bad test marks  -For years I had hardly any friends and had to stay in during recess (grade 6) to do assignments because they were not \"good enough\" even though I felt I had done my best.  It felt like a punishment, although I didn\'t know what I did wrong.  -Thought of suicide many times, about twice I have considered it on a serious level cause of the frustration and stress I have been though?  -Cry on and off many nights because I want to understand and finish my math, but it\'s frustrating and like foreign and finish chem but no matter what or how many explanations of get, I just can\'t seem to grasp the concepts  -Marks vary drastically Ex:(English Unit Work: 80%E, English Comp test: 52%) (Chem Quiz: 35%, Chem Lab: 100%, Chem unit test: 73%) (Math Unit Test: 80%, Math Unit Test:52%)  Can anyone say: INCONSISTENCY!  And I just try my best, and those are some of the marks I end up with: 35%, 59%, 45%...That isn\'t the kind of marks you would expect for a girl like me who works her ass off  -Spend hours and hours on homework, studying, cut out most extra-curricular to focus only on school  -Get anxious and stressed during test, homework is frustrating  -Feel like I need more time one-on-one to understand things.  Ex: Time in math when she sat down (actually many times in this specific unit that I kept asking over and over for help) with me to go through a whole bunch of stuff and I got 80% (highest unit test mark in math for me)  -Feel like I try more that others, work more than others, need help more than other people (A lot more help)  -My mother was born at a low birth rate and also had difficulty in school (I read that those born at low birth rates are susceptible to have problems learning, and I read that learning problems run in families)  Yeah the list just keeps going and going.  So yeah that is why I truly believe there is something wrong.  I wish I knew.  I feel so different or something.  No one struggles as much as I do, and I wish I knew why.  I wish someone could just tell me.  Come straight out and tell me.  I am not stupid.  I study, I work hard, but so many things in my life are frustrating, stressful, and to accomplish things I must go to extreme measures of hard work, hours studying, etc...   '),(3900,1807720,'12,June,2004','2004-06-12 00:00:00','    I woke up this morning, reviewed Bio.  Most of my day consisted of...you guessed it...Bio review!!  I am like determined to do good on the final.  I did every question from the review booklet from every unit today.  So now I have like a million things to ask about stuff I am not sure about.  I am going for English help on Monday, and will go to the Bio review on Tuesday.  Wednesday I have to present our Drama final project.  Last night I went to Mat\'s house for dinner.  It was just the two of us, it was a candle lit dinner with rose pedals on the table.  He made the meal himself.  It was good and pretty nice of him to think of such a think.  I wasn\'t expecting it, and I was pretty surprised.  I am just worried about a few things.  I am pretty sure I am right too.  First of all, I am pretty sure he likes me for my looks, more than who I am.  He even said \"I like you cause you look good, and I assume you think that same for me.\"  When he said that in my mind I was thinking \'Actually I like you for who you are on the inside, I don\'t pay attention to looks at all.\'  The fact that like none of the guys liked me before last summer, but then I come back thinner then before and they all seem to like me.  Hmmm strange or coincidence, I will be the one to decided.  Then he also is going away for the summer and said \"I can\'t wait to get back, you\'ll have lost even more weight\"  I mean who does he think I am, some kind of super-I-can-loose-weight-really-easy-and-whenever-I-want.  I don\'t think so!!!  Just cause I lost 50 pounds before, doesn\'t mean I am going to loose anymore.  I am not trying to loose weight right now and I don\'t need to cause I look just fine the way I am.  So I think he only likes me for the way I look, that REALLY REALLY REALLY bothers me a lot.  I went out for an early dinner and then to see The Stepford Wives.  That was literally the only break I had from studying all day besides right now and taking a shower.  However, I am done for tonight.  Tomorrow I plan to read my Bio text books unit chapters - as many as I can get through in detail and know all the facts.  That is all I can really do for now, study, study, study, read, read, read, and ask questions.  And do my very best.  And I keep thinking how I got that Habits of Mind award thing for grade 11.  Oh how I worked my ass off for that.  It feels good in the end though.  To know my hard work and effort paid off.  It\'s a good feelings.  Anyways, so long blog until next time...   '),(3901,1807720,'11,June,2004','2004-06-11 00:00:00','    I feel good today.  I only got a 62% on the Bio test, but that is better than what I thought I got cause I thought that I failed.  I KNOW that I could have done better if I wasn\'t rushed, and had more time, didn\'t feel pressure to finish, and could just take the time and think through the questions.  I did feel really good today.  I got to use the scantron machine thing and got to scan through like a whole bunch of classes tests.  It was really neat to see how it\'s actually done, and it was really cool to be able to do it.  I felt good that my English teacher trusted me to do that cause it was sooooo cool.  Then in active living I got to take attendance for the class.  He is also helping me on Monday with my reading comp.  No matter how hard I try, it\'s difficult.  I think there is like something wrong or something.  It\'s hard to explain, but I am tired right now and I don\'t feel like trying cause my Mom wants me to talk to her and I am going over to Mat\'s for dinner.   '),(3902,1807720,'10,June,2004','2004-06-10 00:00:00','    Early this morning was really good.  I went to the recognition breakfast and got my awards.  I got the Habits of Mind award that I have wanted soooo incredibly bad.  Only one teacher nominated me last semester, and to get the award you need three nominations, and I got the award and this year they gave out pins too so now I have one of those.  I worked so hard for that.  All the years effort, hard work, and frustrations paid off for what I wanted really bad.  Chels wouldn\'t talk to me again this morning.  I didn\'t do anything wrong though and I never did anything that would make her mad. I am so confused.  I don\'t know what I did wrong, and well if she keeps this I\'m not-talking-to-you-thing up then I have a feeling we might go the entire summer without talking to each other.  I don\'t want that to happen.  We promised we would be best friends forever, and always lifelong friends.  We don\'t even talk anymore.  I am busy with things and other people.  She has to realize that I can\'t spend 24/7 talking to just her.  My world doesn\'t revolve around her and she needs to understand that.  Bio was really bad today.  I studied a lot for that test.  I study a lot for all the tests.  It was confusing, hard, difficult, and frustrating.  Not to mention it felt like I was in the middle of Hawaii cause the heat was like blasting from the vent near me.  There wasn\'t enough time for the test, and well I tried to finish it but she had to start class and well I didn\'t want to keep writing it.  I got confused on the long answer too.  I blanked out or something, it just didn\'t make any sense.  I know the stuff though, I just don\'t know how to explain it.  The chart thing we had to interpret made no sense to me.  That was the most frustrating part of all.  It\'s very very very frustrating knowing that you studied, put in effort, and tried really hard and then you feel like you bombed the test.  It makes you ask: What was my effort for?  Why is it I work so hard and doesn\'t pay off half of the time?  Why is it that people who study less then me, sometimes get way better marks then me?  I wonder, I wonder, and I wonder.  I will forever wonder I guess.  Why this and why that.  I think I actually got like 0 on that longer answer, seriously I think I got a zero.  I am so scared to see what my mark is.  The last thing I needed was to bomb and test to like totally bring down my mark before finals.  That is the last friggin thing I needed.  Sometimes I just wish things made a little more sense, and weren\'t so frustrating.   '),(3903,1807720,'09,June,2004','2004-06-09 00:00:00','    I can\'t really say anything exciting happened today.  If anything it was mostly bad.  I mean last night with Mat was fun at the Cadet Awards Banquet, but today well I don\'t know.  First I went to his house to print some stuff, but then his comp kept freezing.  That sucks cause my printer decided it would feed through 100 pieces of paper instead of 1.  So yeah now I have a jammed printer and can\'t print anything with color.  Go figure, last week of school before finals and I am like breaking everything in my house.  Mat then came over and we watched Shrek, then we ate too much damn food cause I seriously think my blood pressure must have gone like sky high when I saw that number on the scale today.  Haven\'t seen it that high in a while, but everyone around me claims I still look the same.  I suppose a bit of weight gain last week of schools and week before finals is to be expected.  It\'s like the ultimate stress weeks.  Not to mention tomorrow is a pretty big Bio test, cause it\'s the last chance I have to bring up my mark before the final.  I study so hard for those tests all the time, and everytime I always get a mark that I know I don\'t deserve.  For the amount of hard work and effort I put it, I truly, honestly, in the deepest part of my heart know for a fact I deserve an 80% on all of those tests.  Life ain\'t perfect though, and I guess in an un-perfect world, hard work and effort doesn\'t always pay off.  To make things kind of bad though is that well I haven\'t talked to Chels in a while.  That in itself isn\'t that bad, except she was talking to Mat about how I didn\'t talk to her anymore, and she thinks I am mad cause \'there\'s never enough time to spend with her.\'  So I don\'t know exactly what her problem is at the moment.  I asked her to talk today and you know what she said \"no.\"  So now a girl who was my best friend won\'t even talk to me, in fact I have said maybe like three things to her in like the last two weeks.  I don\'t want to be like non-communicative or anything, but it\'s kind of hard when someone doesn\'t want to talk to you.  I mean there isn\'t a lot I can do about anything, and I can\'t force her to talk.  I didn\'t even realize there was a problem cause I don\'t have any problems with her right now.  It just seems that she has some sort of problem with me and it\'s bothering me that I don\'t know what it is.  tomorrow I am picking up Mat and going to that Recognition Breakfast.  Finally the moment of truth when I get to open up my awards envelope and see if any other teacher other than my T.A. nominated me for Habits of Mind.  I want that award again this year so badly, that is what I am aiming for.  I worked so hard all year.  You have to have three nominations though, which means that two of my teachers would have to nominate me from the current semester.  I can definitely say I tried very hard, and hopefully two of them did.  Although see the thing is that two of my teachers from last year that nominated me I have this semester.  I don\'t know if they would nominate me again, but I mean if I have habits of mind one year then I would assume no much would have changed.  So I know I deserve it again.  I think I should stop thinking about it though cause I am going to worry myself too much.  What happens, will just happen tomorrow I guess.  And whatever way things turn out, then that is just the way it\'s going to be.   '),(3904,1807720,'07,June,2004','2004-06-07 00:00:00','    It\'s the second half time of the flames game.  We\'re losing anyways.  My mom told me she was going to Ben and Melissa\'s and then pick up a milkshake for me.  She told me this like three hours ago.  I waited forever for her, I was worried something happened.  Then she calls and says that she is still at their place and staying like a hour longer.  Like what the FUCKING HELL I AM SUPPOSE TO DRINK A MILKSHAKE AT 11 AT FUCKING NIGHT JUST CAUSE MY SISTER WAS MORE IMPORTANT THAN ME.  FLAMES SUCK AS RIGHT NOW.  You know I was going to go to a party, but I stayed home cause she always complains how I always leave during the game.  I hate myself right now.  I want to die.  I want to kill myself.  You have no idea how fucking tempted I am to find some pills in the cupboard and swallow the whole damn thing.  If only someone was here right now.  But no, I\'m all alone, no one is here.  I feel like no one cares.  I feel like I don\'t even have a mother.  That\'s nice and dandy that I own a car, and a laptop, and million other friggin things.  That doesn\'t matter ,though.  I want to be happy.  I want someone to care about me.  I want someone to be here right now.  My mom doesn\'t barely ask any questions or anything.  She\'s too busy.  Too busy with everything but me.  I wish Mrs. J was here right now.  I just wish I could talk to her.  She is the one person in the world who truly cares about people.  She cares.  She asks questions.  She takes interest in things.  I just want a mother like that.  Some one who cares and puts me first, makes me feel important, and makes me feel like I possess talents and can accomplish a a lot of things in life.   '),(3905,1807720,'06,June,2004','2004-06-06 00:00:00','    It was a half day today.  We only had to go for the afternoon, so I picked up Mat and we watched finding Nemo til\' school started.  Also the people at the gas station must have thought I was retarded.  I couldn\'t get my fuel door thing open, like the nob thing you twist.  So after like a minute I decided that the smart thing to do is ask someone, so I asked a lady and she could barely get it off too!!  But well she was like \"you just have to twist, and give it a pull\" so yeah I think the guys in the building were like laughing their ass off, but whatever.  All that matters is that I have a full tank now, lol.  Drama was good today, watching Finding Nemo was the highlight of my day.  Mrs. J talking to me was the 2nd highlight of the day.  She is so damn cool, I want to be like her someday.  And well yeah that was my day, and this morning I studied Bio.  Exciting stuff!!  Lol!   '),(3906,1807720,'06,June,2004','2004-06-06 00:00:00','    I barely slept last night.  All I could think of was the Flames loss, my Bio test on Tuesday, the fact that the upcoming week is the last week of school, and the fact that finals aren\'t this week, but one more week from that.  There is so much to worry about.  Friday I spent working on our Drama Final Project, then I went to Mat\'s and helped Ash and him on their Chem project.  Then I had dinner at Mat\'s and spent the night with him.  Saturday morning I reviewed a bit of Bio, went shopping (I bought some tanks, skirts, shorts, and two dresses - one of them for the Cadet Awards Banquet Mat and I are going to, which he is going to fall over when he see\'s me in that dress cause it\'s hot n\' sexy), then I got home, relaxed, and went to a Flames Game Party thing at Ashleys and we had a barbecue and party.  It was fun, though would have been better in the Flames actually won.    This morning I woke up and read over my Bio text, ate a lot - Way too much but I think it\'s the weather.  The rain is putting me is some kind of I-want-to-keep-eating-mode-even-though-I\'m-not-hungry and some sort of depressed state.  Well actually a depressed state until I studied my Bio tonight.  Studying at night is easier than studying during the day.  I feel like I am kind of getting a grasp on this stuff now, and before it felt very foreign.  It\'s kind of a bad thing though that the game is on tomorrow and my Bio test is the next day - but I will just study in between periods and well partially during the game.  Once I studied during play rehearsals while I was off stage and I did good, so I am not going to miss a game for studying.  I am however going to miss another Flames Game Party though, cause I turned down the offer.  I know that if I go to that I won\'t have time to study - and quite honestly Bio is more important than hockey.  Bio is going to be my life and career so I better do good.  Anyways, yeah that is what\'s new.  Tomorrow I going to the bank, the filling up my tank, then picking up Mat for 11:00 AM and going to school a bit earlier to eat and just talk and stuff.  Yeah that\'s all for now.  GO FLAMES GO!    And I pray to God, For all my effort, all my hard-work help me get through Bio and raise my average.  This is the last week I have to raise my average before the final.  God I also thank you for the amazing and wonderful people in my life.  You know who I am talking about - There\'s so many!  ;)   '),(3907,1807720,'04,June,2004','2004-06-04 00:00:00','    Yeah, ok so I am suppose to be in English right now but we have a sub and I happen to be like totally tired today.  Ok, and a little (no actually majorly) yeah majorly stressed out.  Bio makes no sense.  It\'s great for all of those I-understand-every-damn-thing-always-kind-of-people, but it just sucks that I am not one of them.  Last time I studied my life away and my mark sucked, and this unit makes no sense.  I don\'t know what to do to understand.  I actually felt really stupid asking her \"how do I do this?\" and \"How do I do that.\"  Bio is suppose to be my thing.  It\'s something I like, but sometimes hard work doesn\'t get me to where I want to be.  I suppose what ever will happen is going to happen.  No matter what I have basically passed it course it\'s just University standards are becoming way higher and it\'s harder to get it and that scares me sometimes.  I think all will be ok, and things will work out, but sometimes I just really have to question things like that.  I need to question choices, options, etc... but I am absolutely certain that I want to be a Bio teacher and I am going to do anything that will get me there.  I feel like such a bitch today.  Maybe it\'s cause I feel lazy or ate too much junk for breakfast.  Maybe that\'s it, but I really don\'t know.  I\'m glad flames won yesterday, but really I would rather have like a thousand other things then just a win.  Maybe I am so stressed out cause finals are coming, school is almost over, and it seems like I have no time.  I am so busy this weekend!!!!  After school today Ali, Carli, Meg and I are working on our drama project for a couple of hours and then I might go over to Mat\'s house and help him and Ash on their Chem project.  Knowing how things usually work I will probably be the one doing the project instead of them.  Ya that is how it usually works - guess that is what you get for being totally nice.  But I like it, nice is the way to go.  I finished my English project of poems and that\'s why I am taking a break.  I haven\'t written anything down in my journal forever.  I just need to get things of.  Not to mention I have been eating like massive loads of junk food all the time.  It\'s strange though cause I can only eat large amounts of food when I am alone and no on is looking.  Whenever there are others around I eat really healthy and not very much.  They sometimes think I stave myself, but that\'s not that case.  If anything I eat too much.  I got drunk at the wedding, but it was all in good fun.  Me and this uncle Mike we like doing a disco solo to \'staying alive\' which I kind of remember - it was crazy, wicked fun!  And like Ben\'s side of the family kept buying me alcohol - a lot of coke with rum\'s, I had white wine, red wine, punch, and water.  All day it was like RUSH, RUSH, RUSH!!  We had to put up signs, basically swallow our lunch whole cause there was no time to chew, and get our hair done, do our make-up, and yeah just way too much.  On the Friday before the Saturday I got my nails done, went out to Boston Pizza with the Mid Summer Night\'s Dream Cast.  We ordered like 3 appetizers, 2large piazza\'s, and like 5 desserts.  Then we went back to Alex\'s house, jumped on the tramp, ate and drank more, and played a million improve games like party quirks etc...  So fun, so fun!  The day after the wedding we went to Jim\'s place and ate more.  We had French toast, eggs, sausage, left over cage, brownies, chile, butter tarts, cinnamon buns, muffins- Yeah, ok, get the idea - We had WAY TO MUCH to eat!  tomorrow I am going shopping for a dress for the Cadet Awards Banquet, I need to also get some tanks and shorts that actually fit too!  All my old clothes are too big.  Early that morning I am helping my Mom with the garden and then in the evening there is a party at Ashley\'s house, dinner, food, and we are watched the game on her big screen T.V.  Yeah GO FLAMES GO!  And then Sunday I have to memorize my lines for our Drama final project and believe me there are a lot of lines.  Then I need to study my Bio - well actually figure it out.  Analyze it.  It\'s like a friggin foreign language.  So difficult.  So hard.  Everything in school and life in hard though so it\'s not like there is any difference fro before.  Monday night I am also going to bust my ass studying.  It\'s my last chance to raise my mark before the final.  That\'s it.  The end.  Last chance.  Scary thinking about it.  But I pray to God that I do good, because I deserve it, work hard for the marks I have and the marks I want.  I want to go places.  I want to be something in life, and I will.  I know I will.  I can do whatever I want to do because I know that I can.  I am capable of anything and the world knows that.   '),(3908,1807720,'30,July,2004','2004-07-30 00:00:00','    I can\'t even remember the last time I was online. I am trying to think where I left off, I think it was last Friday. I went to the Cher concert that evening. Saw her wearing outfits with more colors that could ever be imagined by the human eye. That was pretty cool and only my second concert ever. (My first concert was Back Street Boys when I was in gr. 5) lol. Oh, the memories. I think I went to the mall on the Weekend, but I can\'t remember. All that comes to mind is my Summer School teachers voice saying \"Don\'t make plans for the weekend, you have too much to do.\" That\'s pretty sad when all I can remember is that. I swear I am like memory challenged in some ways. This is going to sound strange, but like I forget sometimes if I washed my hair in the morning. Sometimes I will be in the shower, look down at the shampoo bottle, and think to myself: Did I already wash my hair!! (As far as I know, no one else does that) I have thrown cutlery in the garbage by accident, took my empty bowl to the laundry room, and when carrying down my dirty clothes from upstairs I took them to the garbage can. It\'s kind of scary all the things I forget and mix up now. These are things I do at 16 (almost 17....August is coming) and well I hate to imagine what I am going to be like at 40!  Tracking my life back to Monday I remember that at teen night, we all played Kinball. (Fun game) and Chels and I got to teach everyone how to \"properly\" play the game. I came up with our team name and we called ourselves Neon. Yeah! Of course we won...cause I mean I am like amazingly talented at sports. That\'s if you exclude baseball because I could only hit myself in the head with that thing like five times. Later that night we went out for dinner at like 8:30 but didn\'t end up eating until like 9:00. I had an amazing rice bowl at moxies, Andy had steak, and Chels had pasta. Then Chels and I split our favorite - the White Chocolate Brownie. On Wednesday, we went bowling (I won both the 5 and 10 pins games I played) and well we had pizza and pop. That was like the first time I have had pop in months cause I don\'t drink it much anymore. I am anti- pop. Too much sugar, and bad for the body baby!  Today it was my last official day of Summer School. I suppose you could say I am finished Social 30. I just have to write the diploma next week and that\'s it. I wouldn\'t do Summer School again though. I didn\'t exactly get the mark I know I could have gotten if I had done the program at B.O.B. I got a 71% for now, until the diploma. I can\'t blow that test. My life depends on it, or ay least it seems. University will be looking at marks, and that is the only 30 mark I will have in at the time they look at my transcrit. I don\'t want them to be thinking like \"Well, if this girl got a 90% in Social 10, 82% in 20, and 70% or somethin\' in Social 30...then what is she going to get in her other courses. I suppose it\'s because I scored BELOW the class average on all the unit tests we did except for one (The one that was based on FACTS, not reading and interpreting) My essays were good for the most part (The last one could of used some work cause we had to write on a topic I didn\'t feel I could defend my position strongly) Also, if I were to hand in those same essays I wrote at B.O.B. I would have for sure got 100%. I realized how easy and slack our School\'s Social program is. I did good on all of my assignments too. Plain and simple, I am bad at tests. I understand why now too.  COMPREHENSION . That sums it up in one word. I don\'t understand what I read. It\'s frustrating. How do you make yourself understand what you just read???? I still had yet to figure that one out. This is becoming a big problem for me though. I can\'t achieve what accurately understand or answer questions because comprehension affects my performance on social exams, math word problems, English (that is a big one) and when I study. It kind of reminds me of the time I asked Mrs. J how she like studies and what helps her to go good. She told me: You need to read your texts book. However, when she said that, for EVERY unit test I had previously done I had read over ALL those chapters MORE THAN ONCE. Then one morning that confirmed by issues with reading comp. is when I asked her some Bio questions about how the book said like one thing and the notes totally describe dsomething different. Then she was like \"Well it\'s just the comprehension, they are really the same things,\" but in NO way, shape, or form did I believe they were the same. I am still confused about that actually. I was confused on the unit test, and confused on the final. I guess you could say that it\'s frustrating knowing that you know a lot of things, and just can\'t score the marks you feel should reflect your effort. It\'s so frustrating you know. Well I mean I know, and highly doubt anyone would care to understand. And GRRRRR my Dad doesn\'t understand, and was like well you should be studying more for that Diploma. I would like to see him take the course. My Mom at least can except that fact that work my ass off for the marks I get and have, and they they solely reflect MY hard work and effort. It\'s never good enough for my Dad. I am just never friggin good enough. And I need to score at least 70% on the diploma to keep me near that 71%. 71% is so far the lowest mark going on my transcript...it just can\'t be any lower. It can\'t be!!!!!!!!!!!!! Just can\'t, and I won\'t let it.  I actually had second thoughts about my career choice today, but yes I am still going to be a Bio teacher no matter what it takes. No matter what I have to do. I just thought about the alternatives. Everyone always says \"Chose a job with your strengths and talents.\" The really of the statement - Well hmmm, I am going against the grain on this one. Math isn\'t my talent, Chem isn\'t my talents, and well I would exactly consider Bio one of my talents either. These are my interests. Things I am passionate about (INCLUDING Math, even though it stresses me out and makes me cry like 235 days out the the year trying to understand) This is what I love. I love Bio, I love Chem, and I love math. So, I guess I am just following my heart. I am going to do what I love. Though today I thought about the easy way out. I thought about maybe applying to like some drama related program at University. I thought about how much easier it would be. I wouldn\'t have to struggle with the math, or be stressed about now understanding Math problems or Chemistry labs. I thought about it. I thought about it until I realized that it\'s the easy way out. I am not about the easy way. When have I ever taken the easy way out???LOL...Yeah good question. Answer = Never. Some how I always surprise myself with what I am capable of with my 250% best effort towards everything I do. Some people see this effort, some people are blind. Other people just don\'t really give a shit, but that is just the way it is. Anyways, I realized that I would regret it forever if I chose Drama over Science. Drama is fun, but not a passion. I want create a life that revolves around something that will keep me interested forever.   '),(3909,1807720,'22,July,2004','2004-07-22 00:00:00','    Life\'s been fun!&nbsp; On Wednesday Chels and I went to one of those off-trips the City has and we had a barbecue and were taught how to play Frisbee golf.&nbsp; It was a lot of fun but really tiring because there were too many hills!!&nbsp; Each group had their own \"specialist\" or expert that were part of the Frisbee golf association.&nbsp; Chels and I got a hot guy named Kevin.&nbsp; Anyways, I didn\'t see Aaron.&nbsp; I guess he didn\'t come to visit this year.&nbsp; I suppose Halifax is far away.&nbsp; I wish I didn\'t loose him number and just ignore him at the end of last summer.&nbsp; I constantly think about him and wonder where he is.&nbsp; I wish I could see him because I miss him.&nbsp; Am am strongly guided by my heart and intuition and I KNOW that I loved Aaron.&nbsp; If I saw him again, I wonder if that feeling would come back.&nbsp; I just miss him a lot.&nbsp; I had TRUE love with him.&nbsp; Never felt anything else like it since!!&nbsp; All the Calgary guys are strange, at least the ones I know.&nbsp; Thank goodness Mat and I broke off.&nbsp; Clearly their if NO love at all, and I question our friendship because of some of the things he has said.&nbsp; I am still kind of pissed off with him and I don\'t think he will be getting invited to my party at the end of the summer.&nbsp; I don\'t care what he thinks of me not inviting him either.&nbsp; Plain and simple it\'s my choice and that\'s final!!!   Anyways I was pretty nervous for my next position paper in Social today because we didn\'t get the topic before hand, but&nbsp;I mean it was so amazing!!&nbsp; If I don\'t get like 90% on that I am going to have to kill someone.&nbsp; It was purely amazing writing with supporting examples - far more than I am sure anyone else provided.&nbsp; I feel so good because I just knew the information.&nbsp; Unlike Social at BOB I always had to purely memorize the info.&nbsp; It felt SOOOOOO good just to not memorize pages and pages of information.&nbsp; It was SOOOOOOOO less stressful.&nbsp; I mean I was sill nervous, but I mean that just took stress to another lever.&nbsp; A very low lever.&nbsp; Not to mention everyone was like: \"Nice dress\" and well it made me feel good.&nbsp; It sparked convo with a lot of people in my class that I haven\'t even talked to.&nbsp; I talked to this really nice girl Andrea that is from Grandin and this guy from Marys.&nbsp; Our convo started with school of course, but then once I started it he just continued and we talked for like the next ten minutes.&nbsp; It was good.&nbsp; The unit we are doing right now is a bit confusing because it\'s like an overload of history in a few days.&nbsp; I also have to do an individual unit on the weekend.&nbsp; Then a test Monday, test Friday, and third position paper on Wednesday.&nbsp; Once Friday comes it\'s all over...and then the Diploma.   I just realized something.&nbsp; I can do just about anything I put my mind to.&nbsp; If I say I want it, then I get it.&nbsp; If I say I am going to do something, then I certainly do it.&nbsp; I also discovered why I have problems with so many things sometimes.&nbsp; It all comes down to comprehension and memory.&nbsp; Tests are confusing cause half of the time I don\'t understand what the question is asking.&nbsp; I don\'t always learn info from texts cause since I can comprehend 100% I miss some important details.&nbsp; When I read I forget what I just read until I read a thousand times again and again.&nbsp; Then there is memory.&nbsp; I can study a lot, then get to a test and only \"partially\" remember something.&nbsp; Since I only remember part of something it makes me mix things up in my mind.&nbsp; I find myself debating with my own conscience half of the time.   Summer School is great.&nbsp; I have become more accepting of the fact that YOU CAN\'T ALWAYS DO GOOD ON EVERYTING!!!&nbsp; I have turned negative thoughts into positive ones.&nbsp; It goes a long way.&nbsp; When I score bad on a test or&nbsp; quiz I tell myself: It\'s only a test or a quiz, there will be others.&nbsp; You can\'t always be perfect.&nbsp; Now that I think of it, chosing to do Summer School has helped me gain confidence I didn\'t even know I had stored in me.&nbsp; It has helped me to become independent cause I mean for the longest time I didn\'t talkto anyone and has to \"go it alone.\"&nbsp; It has also helped me to be less critical of myself and long the the brighter and more positive aspects of life!!   '),(3910,1807720,'20,July,2004','2004-07-20 00:00:00','    Ok, so I will give my story of my life since I last wrote in my journal.&nbsp; Saturday I ATE too much again and went to the Chuck Wagons and Outdoor Show - that was exciting and my some gift of God I have managed to eat like a cow (wait a min...cows eat grass...lol) and gain NO weight - scary, but it\'s a good thing.&nbsp; Then Monday we had a substitute&nbsp;for Summer School (not like I haven\'t seen enough of those in my lifetime) , but I liked him a lot.&nbsp; He was pretty good, maybe even a bit better&nbsp;at explaining things.&nbsp; Then during that night I picked up Chels and Andy and went to the Teen Night thing out here cause it\' fun and something to do - Even though I have too many hours of homework.&nbsp; However, like two minutes after getting there we got rained out so we went back to my house.&nbsp; I ate dinner&nbsp;and then we played Catopoly.&nbsp; Later on it stopped raining and walked back up the the teen night.&nbsp; We made shirts again this year that say: South Side Gals, and then our nicknames on the back.&nbsp; They also have a karokee system this year, and we are having an 80\'s night where we all have to come dresses as someone from the 80s and everything, with 80\'s karokee.&nbsp; Oh, I am so excited.&nbsp; Chels and I am OBSESSED with the 80\'s!!&nbsp; That was the coolest decade ever!!!   Today I got my mid-term report for Social 30.&nbsp; 78%!!!!!!!&nbsp; How did I manage that one.&nbsp; I like practically failed half of my tests and quizzes.&nbsp; I even beat the guy from St. Marys on the average.&nbsp; By the sound of it I also beat everyone on their essays.&nbsp; Everyone was like oh...50, 60...I got 81!!!!!!!!!!!!!!&nbsp; Anyways, that is all I can write.&nbsp; I am a busy gal.&nbsp; Gotta study tonight and go out right now!!&nbsp; So long...   '),(3911,1807720,'16,July,2004','2004-07-16 00:00:00','    I think taking summer school may have been a mistake, but I know how badly I just need to get a course out of the way.&nbsp; I like Summer School, the class structure is good - but my marks are anything but good.&nbsp; I found out I got a 60% on the last unit test and a 55% on the quiz.&nbsp; I did my best and I studied.&nbsp; I have done every single homework assignment assigned and have studied for each quiz and test.&nbsp; I am so afraid that Social 30 could affect my chances of getting into University.&nbsp; I hadn\'t been obsessing over the possibility of getting rejected for a while now, but our teacher keeps mentioning that \"not everyone has the right to education\" either because of financial situations or admission requirements.&nbsp; He told us to take a look around and realize that not everyone in the class will get accepted into University.&nbsp; That just made me realize the reality of it all.&nbsp; There is no such thing as a \"hard work\" mark or \"effort\" mark.&nbsp; There are only numbers to define students.&nbsp; I will be just another number, and just another statistic - % of people rejected/% of people accepted.&nbsp; I am so afraid.&nbsp; I know how badly I want to get in.&nbsp; I know how much I want to go to University.&nbsp; I am still going to stick through to the end of Social 30 though.&nbsp; I am too far now, and whatever my mark I am pretty sure I AM NOT redoing it because I need spares this year.&nbsp; Last year I didn\'t have any and that just about killed me.&nbsp; I just about killed myself some nights too, trying to deal with the Chem and Math.&nbsp; I don\'t want to give up on myself, it\'s just that marks like 69%, 60%, and 55% have me worried.&nbsp; I am doing all I can though.&nbsp; A part of me just wants to cry for some reason, I guess because I am really trying to succeed, but the I think to myself what good would that do??&nbsp; I mean I like school, I am dedicated, and constantly keep up with homework and studying.&nbsp; There is not much more that I can do, so I guess I just have to stick it through to the end.&nbsp; What ever happens happens, because I seem to have very little control over my mark at the moment.&nbsp; Everything always works out in the end one way or another, so I suppose this is no different.   Well, ok so that is my like depressing little side to me right now, but I did just come back from Stampede.&nbsp; I went to watch the hypnotist twice, and ate A LOT of food.&nbsp; I had some ice cream, cheese burger and fries,&nbsp; Stompin\' Strawberry drink from Jugo Juice, half of a lemonade drink and half of a strawberry fruit shake - so good by the way.&nbsp; I think I am going to get another one tomorrow when we go the the Chuck Wagons and the outdoor show.&nbsp; I also got an amber heart pendant thing which I am going to attach my white gold chain too.&nbsp; anyways, I am going to some Stampede lunches and breakfasts tomorrow.&nbsp; I am going to Stampede again, and well Sunday I have lots of Social work to do.&nbsp; I just pray to God that no matter what happens, no matter what mark I end up with in Social 3o, just PLEASE make sure I get into University.&nbsp; I want to get in...and I deserve it.&nbsp; Just thinking about rejection upsets me, and I can\'t handle being upset!!!&nbsp; God, I pray to you.&nbsp; Please.   '),(3912,1807720,'14,July,2004','2004-07-14 00:00:00','    Today alone I was assigned 23 questions (by the way each of them had like four parts to them), 4 chapters to read, and an assignment due Tueday that I haven\'t started yet because I am too energy drained from everything else I have been doing.  I don\'t believe those rumors of \"summer school is easy.\"  If anything it\'s A LOT more work then ever!!  I had another unit test today - guessed a lot of questions actually.  It\'s difficult to do old diploma questions and infer from article, sources, and pictures.  Knowing knowledge is one thing, but tests are another.  Truthfully, I actually do care what I get on it.  I did my best - that is what counts.  I am not going to worry or get fed up with anything.  I am simply going to just keep on working hard.  We have already written one position paper and have another like next week or the week after.  I have had like two unit tests and about five quizzes if not more.  I lost track along the way.  I haven\'t really made any friends in Summer School other than the ones I know, because all the people are strange and most of them know each other.  I am truly finding out what it is like to \"go it alone.\"  Between Stampede, Monday Teen Night\'s, and Wednesday off trips that begin next week I am going to be a busy girl.  Next Wednesday if what I must wait for.  It is when I get to find out of Aaron is back for the Summer.  If God meant for us to be together he will be back because I LOVE him.  I miss him so much and breaking off at the end of last summer was a big mistake.  I gave up something that I have never had with anyone else.  I gave up a feeling that since I have not seen him, I have not felt it.  Love.  Aaron is a true example of love at first sight.  I didn\'t believe in it until I met him.  I need to see him again.  HEADLINE: Calgary Girls Searches for Long Lost Halifax Boy - That is going to be me if I don\'t see him for the next few years.  I can\'t go on that long with out him.  I need him.  I need to see him.  I need to talk to him.  I love him.  On Monday we are making shirts again this year and I think Chels and I are suppose to be brainstorming a good name to put on em\' this year.  Last year we were the \"South Side Girls,\" this year we have the cleaver idea of being the \"South Side Womanizers\" yeah now that\'s cool!! LOL ;)  On Wednesday we are doing some BBQ thing and doing disk golf.  I can\'t wait til\' Friday cause I get to go to Stampede again and this time I am going to see the hypnotist cause he is like so amazing.  Then Saturday Meg and I are going to a Stampede breakfast in my area, then we are going to some Stampede lunch in McKenzie.  Then later that night I am going to the Chuck Wagons and Grand Stand Show cause that is like the best.  OMG I love Stampede, and it will such when it\'s over.  I am just about half way through Summer School too!!  Thank goodness.  I have such a passionate hate for social and history.  It makes no sense really, not like it ever did.  I have more common sense of current issues it todays world rather than the \"old time\" issues and conflicts that were occurring.  Whatever though, as long as I pass Social 30 all is well.   '),(3913,1807720,'10,July,2004','2004-07-10 00:00:00','    Today was fun despite the morning.  My parents were arguing again.  I think I liked things better when my Dad lived in Vancouver.  Now I understand why I lack confidence half and the time and doubt myself.  Over and over my dad kept saying \"You guys are so stupid\" to me and my mother.  Now I remember why I lack confidence in things like Math.  I remember I used to ask him how to do things and then if I didn\'t understand he would get upset with me.  I kept thinking in my head this morning: I wish my parents would divorce.  I felt so strongly like that this morning.  It made me want to cry just listening to them go at it, and there I was right in the middle of it.  The argument didn\'t even have anything to do with me, yet my Dad put it as \"you guys are so stupid.\"  I mean sometimes my parents are fine, but old age has turned my Dad into Mr. Grumpy.  I don\'t like it.  That is not the person I know or thought I knew.  Anyways, despite that we picked up Megz and went to Stampede.  We walked around inside and checked out the art exibits and all the both things.  We also pretended to be really interested in products so we got tons of free samples.  Then we got pizza and mini-donuts.  I don\'t even want to think about how much fat were in those, but they were good and I don\'t eat those everyday.  Then we went to the Rodeo which was pretty excited.  Except I am not really into bull-bustin or any of that stuff.  The barrel racing was pretty exciting.  I have to admit that my favourite parts were before the Rodeo when the Stampede Showband performed, during the Indian Hoop Dance, and part of the show where the rodeo clown dressed up as if he came from different eras: 70\'s, 80\'s, 90\'s and it was really funny.  To say the least all the non-related Rodeo things were my favourite, lol.  It wasn\'t that bad though, it was pretty entertaining.  Then later on we went for dinner and then Megz and I went to those boths where u get old fashioned picks and we dressed up like the older days and got a pic taken like that.  It is sooooooooooo hilarious.  My Mom was killing herself laughing.  Then we went to see a whole bunch of animals and stuff and that was about it.  When I got home I had to work on my position paper for social 30.  We get Monday\'s class to write it, but I wanted to get a head start.  Here\'s that picture we got taken at the Stampede today.  And by the way that outfit makes me look fat, my stomach is NOT that big.  It is a misleading misconception.  Anyways, still looking good!  lol ;)   '),(3914,1807720,'09,July,2004','2004-07-09 00:00:00','    Now I know why my marks seemed to shoot up when my Dad lived in Vancouver all the time.  Here my parents go again.  Blah, blah, blah, fight, fight, fight.  It makes me feel like I did something wrong or something.  It\'s actually not about money this time though.  Believe it or not.  My Dad can\'t find some important envelope so he is like raving and ranting all over the house.  I chose to sit here at my laptop and just be silent - invisible.  Things are better that way.  Other wise, I will end up being yelled at and blamed.  I really don\'t need that.  On Monday I have to write a position paper in class for Social 30 and tomorrow I get to go to Stampede.  I am excited about that.  Megz and I are going to the rodeo too!  I also have tons of homework again!  So much!!!!  I mean I am crazy to sign up for Summer School!!  The teacher is really really good though, but the info we learn is fast.  I will admit I find it difficult to learn, but I am doing pretty good.  Now I know what it feels like to \"go it alone\" cause I don\'t really know or talk much to anyone.  Most of them are kids that normally go to Grandin and they are all strange.  There is one guy from Mary\'s - my competition, and the rest from O\'Byrne that all the smoker-drinker people.  I chose NOT to socialize with those that smoke and drink.  Besides it\'s again getting a good mark, not about socializing and making friends.  I just have to keep reminding myself: Three more weeks, diploma, then I am DONE SOCIAL for the REST OF MY LIFE.  I can almost taste it.  That is my motivation.  Considering I hate Social with a passion, except for Model UN.  That is really fun, and the way the UN works and is organized is actually pretty interesting.  I don\'t even know why the hell Social is mandatory anyways.  I mean thirty years from now I am not going to be like \"Hey, remember the tactics that Stalin used?\"  Like, I DON\'T think so!!!!!  And for English, why do I need that to the 30 level??  I want to be a Bio teacher not an English teacher!!  Who cares how well you can interpret Shakespeare and use guessing ability on reading comprehension tests.  What gets me the most is how people DON\'T need Math and Sciences to the 30 level.  I think ANY DAY logic and reason are MUCH MORE important than stupid history and English.  I know I am right too.  Those English and Social nerds just hit it lucky that\'s all.  Then there are people like me - who like Science and Math. (math is hard, but yes I admit it I find it interesting) They are just making the Science and Math nerds do more shit that they really don\'t need in their schedule.  I have enough courses as it is.  If I didn\'t have to do English and Social I could spend more of my youthful years doing more productive things and things that would enjoy.  I could take more option classes and have more time to relax.  But...NOOOOOOO...Government...school boards...who ever thought it was a good idea to make make Social and English to the 30 level mandatory and only Science and Math to the 20 level mandatory had too many screws loose.  History and English are merely a waste of time.  Society should be educated to use logic and reason.  To critically examine situations, learn to interpret, and experiment.  Those are the essentials.  Anyways, talking about that is not going to change anything.  Lately I have been having nightmares and thoughts about university.  I even dreamt I was in University one night, but I was taking some performing Arts program, lol.  I like acting, drama, dance, and performing but would NOT do that as a career.  I am scared though.  Scared for the future.  I wish my marks were higher from the previous years, but I tried my best.  Sometimes your best isn\'t always enough though.  I want to go to Univeristy so badly.  I want to get in so badly.  When the applications come out online in live Oct/Nov I am going to apply as soon as possible.  It\'s so stressful though.  Just all the articles in the newspaper talking about how difficult it is to get accepted now a days, and I mean even all the teachers during school time telling us all how important it is that we focus and concentrate on acedemics.  I read about a month ago in the Herald that what Universities do now is that they start at the top - those with the highest marks, and then work their way to the bottom - those with the lowest marks.  I have to compete with those people wanting to be doctors, dentists, all that other stuff that requires you to be like super smart.  I just want to get so badly.  I don\'t want rejection.  I don\'t know if I could even handle rejection.  I need to consider the possibility that I might not get in.  My marks are nothing amazing and don\'t stand out as much as other peoples do.  My marks don\'t suck, and are not bad, but compared to others with 90 averages I just can\'t compete.  That is not in my range.  I already try my best and get in the 70\'s on most cores and in the 80\'s and 90\'s on other things.  My mom knows this lady whose son got rejected and had a 90 overall average.  I mean EVERYONE in Canada should have a right to higher level education.  Marks should not be everything, but really they are.  Marks are everything when it comes to your future.  Marks determine whether you will be accepted or not accepted.  That then would determine the career path.  One more year of high school and then I am off to bigger and better things.  I work so hard for what I want soooooo incredibly bad.  It\'s my dream to teach.  When I am standing infront of a high school bio class teaching bio, that is when I can sit back and know that my hard work, effort, dedication, and persistence though the challenging Math, Chem, and a variety of other subjects.  That is the one day I dream of.  I can picture it.  I can \"see\" it in my mind.  I don\'t want to anything else but be a teacher.  That is what I was put on Earth to do.  I can hear a voice in my head telling me that is what I am to do.  It\'s just so right.  I love school, love kids, and would love to be the one to influence and help and support others one day.  All that depends on one thing though.  If I get accepted when I apply for University.  That can make or break everything.  I have to consider it though.  I don\'t even know what to do if I don\'t get in.  I guess try again next year, but for the mean time if that happens I guess I will go along with things.  See where my life takes me.  I NEED to get in though, I don\'t really want to think of the other possibility of being rejected too much.  I believe that I can get in, and I want to get in.  I deserve to get it.  Ok, and now I am thinking about this way too much!  I guess I will just have to wait until the time applications come out in the fall.  Then I will go from there.   '),(3915,1807720,'08,July,2004','2004-07-08 00:00:00','    We had done so much.  Social 30 is challenging.  It\'s difficult in some aspects and like a lot of people in my class seem super smart.  It kind of bothers me or intimidates me.  Mainly for that fact that we wrote a unit test today and it just took me a long time.  Some people finished really quick and scored really high.  I want to be one of those people.  No matter what though, it takes me a long time.  I kind of lost my focus near the end of the test too because everyone was being a distraction cause half of the people were done.  I felt like I was under pressure.  That probably lost me a few percent.  The test was difficult, but it was also composed of previous questions from previous diploma\'s.  I felt really \"slow\" though, like everyone was good at test taking and like I just couldn\'t answer anything as fast as anyone else.  I don\'t know if that is a bad thing, because maybe I just think things through.  I only got a 69% on the test though.  I know the information.  I know I know it.  However, I stopped my self-defeating thoughts this time and told myself that it\'s \"ok.\"  A mark is a mark, and marks don\'t determine your future.  Twenty years from now, no one is going to care what I got in Social 30.  Anyways the questions were difficult.  The answers were all so close.  Social exams remind me of reading comp, and anyone who knows me knows that reading comp is the one thing I LEAST excel at.  Not to be degrading or anything, that\'s just plainly the truth.  I have more things to talk about.  My nightmares about not getting into university, my fears of next year, so much stuff.  I am really tired right now though from all the reading and homework and studying.  Summer School is a lot of work and requires a lot of effort.  This is yet just another challenge for me.  Just another challenge in which I will prove myself.  My dedication, hard-work, effort.  It will be all worth it in the end.   '),(3916,1807720,'06,July,2004','2004-07-06 00:00:00','    I AM SO SICK.  I thought I was going to die yesterday like two hours after lunch.  I went to the third floor of sears and got a salad w/ chicken for lunch w/ a muffin.  It was good at the time.  Then like two hours later my stomach really hurt.  Then this morning it was sooooo difficult to concentrate in Summer School and I can\'t miss classes cause then you get kicked out.  I even had to throw up in the middle of class.  I could feel the acid coming up and I know it wasn\'t going to stay don\'t so I quickly asked to go the the washroom and pucked up green and yellow stuff.  I am thinking my salad from yesterday.  All I ate this morning was two bites from my cereal and when I felt a little bit better (not 100% better) I wanted some fast food.  I have lost a a lot of weight too going to the washroom and throwing up, so I really didn\'t care if I had a happy meal from McDonalds.  That was so good!!  I haven\'t had one of those for years cause I usually don\'t eat anything from Mcdonalds.  The burger was small and so was the drink - a nice portion size for someone who is sick.  My mom had to finish my fries cause I was full and I was proud that I actually finished my small burger cause that was hard to get down.  That is the only real meal I had all day.  Usually I don\'t drink pop either, but something in it makes me feel better.  Luckily I felt ok when I wrote my quiz today and well my unit test is Thursday so I still have tomorrow to study.  I just really want to feel better tomorrow.  I don\'t want to be sick now.  I CAN\'T miss Summer School.  I just want to get Social over with.  I do like Summer School though.  This is the first time I have actually ever found Social interesting.  Anyways I am going to watch TV cause I have been doing all the readings and assignments for the course earlier.  There is A LOT and it takes a while!  Anyways, hope I feel better tomorrow.   '),(3917,1807720,'05,July,2004','2004-07-05 00:00:00','    I think it\'s a good thing I am keeping my brain active.  I think since I am doing Social 30, when I got back to school for the regular year I won\'t have forgotten so much.  In fact, I recall most of the recent stuff I have learnt.  There is a lot of homework though.  I have e quiz tomorrow and a unit test either Wenedsay or Thursday.  It is so fast!  I mean it\'s a good thing though, cause I just have to do this and NO MORE SOCIAL FOR THE REST OF MY LIFE!!!!!!!!!!!!!  That feels so good, but our teacher makes the info pretty simple to understand and it is fairly intersting.  I also feel really smart cause like this morning no one was talking so I was answering all the questions.  I didn\'t have time to workout today though, and probably won\'t for a while.  That is just a scarific I must make.  I will have time after Summer School is done and on the weekends, except Megz and I are going to Stampede and got Rodeo tickets for Stampede.  The weekend after that I have Grandstand tickets and ChuckWagons cause I love watching the races and the show.  Anyways, I got to go read over the next chapter in the Social text cause I got a quiz on the stuff tomorrow! ;)   '),(3918,1807720,'04,July,2004','2004-07-04 00:00:00','    I went to the gym again today, not like that is anything new.  I am excited to go to summer school tomorrow.  I don\'t actually care that gained weight cause I look good and feel happy except for when my Mom and I argue.  Last night I just kept thing about one thing.  University.  My mom told me that she new a lady whose son had a 90% average and didn\'t get it.  I am so scared.  People always tell me I should have a back up plan just in case - but I don\'t.  I only want to do ONE thing with my life, and that is teach.  I can\'t imagine doing anything else, which is why I WILL become a teacher.  I mean it\'s like their in a inner voice telling me this is my purpose in life and the role I am to fulfil.  I am so scared about getting into University.  I don\'t think I could handle rejection.  I don\'t know what I would do, and I don\'t have a back-up plan.   '),(3919,1807720,'03,July,2004','2004-07-03 00:00:00','    I still can\'t wait for Monday.  I am so excited to go back to Summer School.  I had so much homework already though and it took me a while this morning to finsih it.  I really like the stuff though.  It\'s interesting and FOR ONCE I KNOW something in social - I LEARNED something.  I went to go work out again today for like two hours.  When I stand on my toes you can literally see how muscular my legs have become.  I was 138 this morning and then in the evening I was 140.  However, right now I am not too concerended with the weight.  I mean I drank three bottles of water today and did the full circuit of weights.  I think maybe I am reducing some of the fat in my body and just building more muscle cause I really don\'t see a big change in the way I look.  I feel so great.  I feel so happy.  On the treadmill I just ran for half and hour stright and just thought about all the stupid bad things that occurred of have happened.  My Goodness that is such a good way to release anger.  I love going to the YMCA.  I am addicted!  Last night Ali and went for a swim after our workout too.  That felt so good.  I also made these energy granola bar things tonight.  It was in the paper like last month and was created by a dietician who won some contest thing.  It is soooo good, and really healthy.  Man I wish all food could be like that.  LoL!!   '),(3920,1807720,'02,July,2004','2004-07-02 00:00:00','    Today was so awesome.  It was so fun much fun.  It was so different, and cool, and just really neat.  I started Summer School today.  Already I have like tons of homework and the teacher told us to expect HOURS of homework.  I am used to doing that anyway though, so it shouldn\'t be that big of a deal.  Bishop Grandin is a really old school though and like there aren\'t many windows.  That school is like ancient but I have a really good Social 30 teacher.  I actually REMEMBER EVERYTHING he taught today and understand it all.  He is very good, and the class has so much structure compared to BOB\'s Social program.  There are only about 16 people in the class, and like hardly anyone talks so there are like no distractions or anything.  I love Summer School, or at least the first day.  I have my first unit test next week already and we got like five million assignment today.  I drove Alicia back to my house from Summer School and then we had the left over fries and onion rings from Peters last night.  We ran a whole bunch or erins and then picked up a vanilla cone each from McDonalds - by the way their ice cream, apples, and yougurt parfait are the ONLY things I will eat there.  The rest of the stuff is nasty.  For like an hour Ali and I sat outside on my swing.  Then we went to the gym, did half an hour cardio and the entire circuit of weights, then we went for a swim in the pool and hot tub.  It was a fun night and a fun day.  Once I got home from the YMCA I was planning to do some more social homework...lol...I am interested in Social for once (that is like a first!!!) and well then I ate my dinner at like 9.  What a day!!   '),(3921,1807720,'01,July,2004','2004-07-01 00:00:00','    I went to the gym last night and for about an hour and fourty-five minutes today.  That is like my second favourite place in the entire world!!  I love working out and I met this really nice girl at the gym today.  I weight a few pounds more then when I left school, but when I look at my body in the mirror it looks nicer then before.  I like my stomach now cause I barely have one, my arms feel better from all the weight lifting, and my legs are just nice and not really thin like half of the other teenagers on this planet that starve themselves.  My Mom\'s surgery was good, but now I have to do like everything.  My Dad is flying in from Vancouver for the final time and then we are going to go to Peters Drive In for dinner.  Tomorrow I start summer school.  I am excited, but nervous.   '),(3922,1807720,'11,August,2004','2004-08-11 00:00:00','    I love Lindsey Lohan.  This is like my favourite song in the entire world right now.  It makes me happy.  It\'s fun.   I love it.  What can I say, the girl makes me feel good.    It Freaks Me Out  You hate the way I\'m always watching videos Eleven-thirty and you just can\'t sleep You think whatever I\'ve got playin on my stereo Is so pathetic that it makes you weep  And the beat goes on when I get back home And I run stright past you to the telephone And my feet don\'t fit in your sensible shoes But you just won\'t quit until you\'ve killed my groove  Wooo it freaks me out I\'ve got to scream I\'ve got to shout Wooo it freaks me out like Halloween I\'ve got no doubt  I think your weird, it\'s writen on your resume You and I\'m stupid and your best freind too And if I smile you think I\'m doin it to get my way And if I don\'t I\'ve got an attitude  And the beat goes on when I get back home And I run stright past you to the telephone And my feet don\'t fit in your sensible shoes But you just won\'t quit until you\'ve killed my groove  Woo it freaks me out I\'ve got to scream I\'ve got to shout Woo it freaks me out like Halloween I\'ve got no doubt   '),(3923,1807720,'10,August,2004','2004-08-10 00:00:00','    My legs hurt right now. I got to the gym way too much, but I can\'t help it cause I need to like look good. Besides working out is good for the body. My muscles just hurt a little from too much weight lifting. I went to see Spiderman 2 on Sunday - believe it or not I really liked it. And I usually hate comics and thought the first one sucked ass. I liked Kirsten Dunst though, she is one of my favorite actresses besides Lindsey Lohan and Elisa Dushku. I wish there was more time for Summer too. Summer School caused an entire month to go to waste. I could have been having fun. I guess on the positive note I am DONE SOCIAL 30. DONE DONE DONE DONE. Oh yeah........!!! And I found the best song by Lindsey Lohan - Ultimate, listen to it sometime cause it\'s good. Yesterday besides working out and watching Ellen, Chels and I went for dinner. We went to Jack Astors cause we are sick of all the other restaurants around here. We have been to them like a thousand times (we eat out too much) but yeah it was pretty good. Chels thought the meal wasn\'t big enough, but like it was pretty big. I usually don\'t consume that much food for dinner, but whatev. So we also got dessert cause Chels wanted some (I honestly would not have eaten chocolate if she didn\'t want it) I mean really, honestly I was trying to be healthy :) It\'s alright though, cause I toned my body and stuff and well lost some weight, so I look good. Chels and I then made new truth and dares for my Jenga and played. I had the best time cause we are way too a like. We are wild, fun, and crazy. I can\'t believe some people think I am shy and reserved!! Now that is seriously funny! Yeah, it was a fun night. We were so loud my mom probably thought we had been drinking, like she thought last time. I was trying to ask her something but I mixed all the words up, and like everyone I was with was laughing their ass off, even me once I realized what I said. It was funny. No wonder she thought we were drinking, but she knows I would never drink and drive. This morning I went to the gym again - yeah like that is a surprise. Then my mom and I get went to South Centre and got some Chinese fast food. It was good. I love that new place - Bourbon St. Grill. Love it!!! After trying it at Market Mall with Melissa I wanted it again, and then go figure it\'s in South Centre. I also went to guess and my mom bought my this belt I wanted for my b-day (Well I think my b-day is like a week from today) but she can never surprise me anyways. Now she just buys half of my gifts when I am with her. And later on I think I have to take the papers and cardboard to recycling then we\'re going out for Coffee with the neighbor\'s. That\'s my day!   '),(3924,1807720,'07,August,2004','2004-08-07 00:00:00','    Ok, so I did something totally embarrassing today. My mom was like \"Vanessa, take me shopping to Safeway.\" So we went to Safeway and I knocked some guys basket things over while he was setting up a display. I guess I misjudged how far away it was from the cart. It also made a really loud bang. And then she was like \"Geez Vanessa, can\'t you be more careful.\" Like, GRRRRRRRRR I was friggin being careful. It\'s like it\'s always my fault and well I guess it was, but I really didn\'t mean to. She doesn\'t get it. I didn\'t try to knock all that shit off the shelf.  I had another break down today. When the stress builds up it\'s not a good feeling. I always have this recurring notion that I am different. I mean yes we are all individuals, but there is something just not right. I know it, and I have known it for a long time. By now I am practically certain. People are always like aww sweetie your a hard worker and everything, and seriously your just being too hard on yourself. And like seriously people don\'t know shit about my life or about myself. They don\'t see what goes on behind the scenes, they don\'t see out of my eyes, they don\'t know!!!!!!!! I wish I could scream: Can\'t someone PLEASE understand!!!!!!! Your know just to understand what I am going through would greatly be appreciated. I mean right now isn\'t as stressful as school, but nonetheless it\'s a stressful time. I admit I am usually pretty happy during summer, but not today. Maybe it\'s the the PMS kicking in for once, but I highly doubt that cause usually turn into the nicest gal you\'ve ever met at that time. I have always considered the possibly of having a learning disability, but everyone else seems to shrug that idea off. They don\'t think it could be possible. They laugh at the idea. I don\'t get it though, I mean why??? Why do they just automatically shrug of they idea when they no nothing about what I struggle with and what I have trouble with???? Why...Why...WHY???? I KNOW for a FACT though, something is wrong, wrong, WRONG.... I do think it could be possible and am open the idea that is it may very be possible. I guess people just really don\'t see or understand what I deal with and go through. I suppose I hide the things I find difficult, but I mean if I don\'t have some kind of learning problem then what is wrong????  I wish I knew how to talk to someone about this. I mean first lets start with the reading cause that like affects my life in too many damn ways. When I read I don\'t understand what the hell I am reading. It makes no sense unless I read it like ten times. Even then sometimes I just don\'t get the connection. When we did Reading Comp in English this is like a seriously problem. There just is NOT ENOUGH TIME to read something like ten times over. When I read something once I don\'t even remember what I just read. I mean no wonder I never understood how to do the math word problems. I didn\'t understand what the hell they were asking. During the English Final I got the biggest friggin headache of my life. It\'s hard to concentrate when your heads hurts. When I was doing the Social 30 diploma recently I started to get this pounding in my head. The diploma was also when I realized there was something is seriously wrong. I could not focus on the words I was reading or well trying to read. I am going to sound like a major nut case but like the words were like moving. I usually don\'t have problems like focusing, but this was different. It seemed like the text was too dark and the words were all too close together. I know it sounds like really weird, but it indefinitely was happening and there was nothing I could do. I just tried to focus as much as possible and infer the meaning to the best of my ability. Of course by the end of the diploma I started to get a headache, but not as bad as the English final. The more I read, the more my head hurts. I am at the point where I would rather do just about anything but read.  Then there is my job situation. I feel like such a loser. Practically half of my friends that want jobs have jobs. I used to have a job too, but only because I got it through a friend of a friend. You know how the story goes. That lasted about seven months believe it or not. I just worked hard and most managers liked me because I was kind and friendly. It didn\'t really matter though because no manger ever had the courage to fire anyone. Anyways, getting back to my point I have applied at every place that classifies you at the bottom of the work chain. Wendys, Burger King, IGA, Co - op, Safeway twice (diff. locations) Half of the time I even dropped off my resume to the managers, which is a must these days cause half of the people just throw your application in the garbage. Do you think I got one call for any of those places. All the fuck I wanted was a job to earn money so I could pay for the stupid gas my car needs to get me places. By all means it\'s not my resume, it\'s practically perfect and my list of extra-curriculars are astounding. It\'s not like I have never worked before because I do have experience. How pathetic do you think it feels to not even be able to get a job at burger king or wendys. You always need people for fast food. And everytime, no call, no interview, no job. I mean I suppose it might be the fact that I am NOT going to work every damn day of the week. Sorry, I had a life and would like to maintain at least half of the good part. These days it\'s always like everyone wants you to dedicate 110% of your time to work and work only. Sorry, but I won\'t and I refuse. I wonder if I filled out the stupid application\'s wrong, they were all confusing. Too many boxes and confusing words, except for the Wendy\'s application. Maybe they just assume people applying there aren\'t that smart so they need a simpler method of obtaining employment.  Then there was this day when some strange guy stopped at my house and asked me how to get to MillView Bay. Ok, I mean Millrise is composed of streets that all friggin start with Mill so you tell me how the hell would I know how to get there. Then comes my mom and pulls out a map. \"Here Vanessa you figure it out.\" Guess what, I couldn\'t figure it out. I didn\'t feel that bad though considering the guy was trying to read the map too. To both of us it made no sense. So I suppose it isn\'t so abnormal to not be able to read a map. It was kind of stupid though cause it took us like 15 minutes to find the street on the map. Really, that honestly tells your nothing. Ohhhh a street name with squiggly lines all around it, which too many colors...like how the hell is that suppose to tell you where to go. Ok, so there is my next secret - I wouldn\'t say I cannot a map (saying you can\'t do something is degrading) but I would simply have to classify myself and an inefficient map interpreter. That sounds better than can\'t read maps. I still remember when me and my friends when to greek fest, I got lost three times.  Wow, this is a long journal entry but I feel a little bit better now. I can finally unleash some of what I am feeling. I am not sure if you would call it stress or not. I\'ve always wondered if my childhood ever had a slight affect on me and the way I am now. I ran into a wall when I was five if that says anything. I honestly didn\'t see it, maybe that is why I can be such a bad driver. I just don\'t \"see\" things sometimes. I have almost run over people like four times now, and once I almost did it with four other people in the car. I swear my driving is going to scar them for life, but they said they still trust me. Go figure, they must be real true friends if they have that much trust in me. Failing drivers ed was yet another bullet shot at me in my life time. That was hit me pretty hard though. I couldn\'t remember any of the parts of the car and the fluids colors. At the time it seemed so stupid to have to learn that stuff, but now that I drive more I realize that it is pretty important. I waited six months before driving again after the drivers ed. My self-esteem had never taken at hit as hard as that one. I had failed things before, but this was worse because half of the world talks about how fun driving is. I think it is fun now, but it took me about a year to realize that. After eight months following driving ed I got my licensee, but NO ONE knows how much effort I had to put it. I mean more effort that ANYTHING I have ever done in my life. Anyone who knows me knows my level of effort is like 500%. I put more into learning to drive. Once Feb 2004 hit, I practiced with my mom EVERY night and MORNINGS and NIGHT practiced my parking. This went on for months until the Easter weekend in April. I swear I will always remember that day. Just the feeling of: I got my license and actually did something right for once. That is a feeling you can never replace. Only I know how hard it was to get that license that occupies its glorious space in my wallet. I also had sleep apnea when I was a kid. According to my mother I would snore so loud she could hear through walls, then for a minute it would be silent. I would be like practically lifeless for a minute. I had an operation to remove my adenoids and my tonsils were taken out. After that it never happened again, but I think that could have like had a major affect on me - like killed maybe like a couple thousand brain cells or say....affected my learning.  Call me paranoid because maybe I am. Call me crazy because maybe that\'s the answer. I just wish I had an answer though. To a lot of things. A solution. Something that could make me understand just why sometimes things to challenge me in ways sometimes far beyond where I ever thought my limits could extend. I don\'t know what\'s wrong, I only know what I THINK is wrong. I wish I knew what was wrong me with. Maybe I am just a teenager, maybe that is the problem. The difference between me and everyone else though is that none of my other friends and people I know and keep in contact with experience anything close to what I experience. I know for a fact they don\'t go through as many times of struggles and frustration as I have been through. From the outside I seem like the happy, hardworking girl who at times doubts herself. On the inside I am a confused and struggling teenager.  If we are to look back in time I also thought I was pretty different from other people. I was always the kid who didn\'t understand, the slow one, the misbehaved girl. I was pretty badly behaved during grade two, but only because no one cared that I didn\'t understand anything. I would always be ignored and I suppose you could say I wanted attention. That year I drew so much on my desk in pencil, it was completely covered. I don\'t actually blame myself for the horrible marks I got that year or any year after. I tried to copy off of others math tests all the time. Being a cheater in school was how I passed half of my school stuff anyways. I didn\'t know how to count money or tell time. Thank goodness I know now, but I didn\'t know how until about grade 4/5. I also used to work with some strange lady one-on-one with counting and everything. That helped, but as I got older people had no respect for those whose marks shot up when they got individualized instruction. A couple years ago I also found an old report card - it was from grade 3. I started to read the comments on it and mentioned was remediation of certain math skills. To this day I still don\'t know half of my math shit. It\'s sad to say that I still don\'t understand how to round even though I have been repeatedly show, most graphs I can\'t seem to understand or interpret so most tests that involve interpreting graphs I simply just guess, and I don\'t know how to reduce fractions or ratios. I just guess on all of those things and hope I am right. Somehow I manage to get through math every year by hard work, practice, and help of others. And well I can tell time by now of course, but I can\'t just take a glance at my watch, I have to like really stare before I know the time. Memorizing my multiplication table - that didn\'t happen until grade five. My parents made me work on that every night and I still don\'t even remember half of it. Thank God for calculators though, I don\'t know what I would do without one.  Then came the year I lost most of the friends that I did have. I didn\'t have many in second grade because no one wanted to be friends with the \"misbehaved dumb girl.\" In everyone\'s eyes that is what I was. Growing up was hard for me. At times I feel glad to be a teen, because being a teenager is better than being young. Anyways, I was little miss violent in grade four but some how friends with the \'popular\' kids. No one like me though especially after this time I beat up a girl named Brittany. I became quite and reserved after that spending every recess alone. It\'s funny though cause me and Brit still keep in touch today even though she goes to a diff. high school and are amazing friends. They year after I met Chels, my best friend and partner in shopping forever. My life changed for a while. I still almost failed half of my classes, and even Chels admits to this day: \"I remember you being so bad at school.\" That year Chels and I were probably the two most unpopular people. Grade six eventually came along - what a nightmare. The few friends that I had previously made before I was separated from that year. Before long I became the most hated person - literally. I almost killed myself that year because at times I have to admit the verbal abuse stared to get to me. That was also the year of ultimate weight gain. Not to mention the teachers I had were horrible, they had no respect for me or anyone who didn\'t seem to understand anything for that matter. I tried my best, but it was never good enough.  If people wonder why I doubt myself so much...well there is the reason why. Blame my grade six teachers, they stomped all over the self-esteem I did have. I didn\'t get very many recesses that year either, because they always kept me inside. My English work was never good enough, they would make me re-answers to novel questions during recess. I would have to do homework again that I had \'tried\' the night before. It seriously was not my fault thought that I didn\'t understand what I was reading. I realize that now. Everytime I would spent that fifteen minutes crying wonder why everyone else got to go out. It wasn\'t fair. When I stayed in at recess to do the work again, my answers were usually better than the first time only because I had read the same chapter like five times by then. This led the teachers to assume I just didn\'t care about work and wasn\'t trying. That was a frustrating feeling and made me almost not want to try anymore because no one realized how hard I worked. Because of this frustration I wouldn\'t complete spelling assignments and when I asked about math questions the teachers explanation never made sense. I didn\'t finish half of my math assignments. And again the explanation was \"she\'s not trying.\" How did they expect me to finish them when I didn\'t understand. If I understood half of the work that year, I honestly would have finished it. I honestly would have done it. It seemed like a better option to look as if I didn\'t care than to look stupid. Instead of being frustrated with work though, it was easier just to not do it. Shortly after my mom met with those two teachers and to this day I still have no idea what they talked about. I asked what they talked about to my mom, but there was never a response. My mom wasn\'t even that upset after that meeting either. I don\'t know why, but to this day I think she is still hiding something from me. There is something she has not told me.  After that I was sent to the resource room for math, but it was like already passed the half way point of the year. I do admit I desperately needed help with math, but those people did not need nearly as much help as me. I am not exaggerating. First of all the people already knew the ladies, I didn\'t. They understood the math, I didn\'t. They knew how to ask for help, I didn\'t. There were too many kids and I was actually ignored more than anything. As far as I am concerned, the other kids there honestly did not need any help at all. They all understood perfectly well. That is when I realized that no one else struggled with all this math as much as me. No one else really had any of the same difficulties I had experienced. I mean out in the world there are definitely people like me. They are out their somewhere, but they sure weren\'t in that school. I would NEVER cheat today, don\'t get me wrong but back then cheating was the only option. I had a D in math and was afraid I was going to fail all of my courses. I was scared so I cheated. I cheated on EVERY sing grade six achievement test except math. I know it would look suspicious if like all year I practically failed math and then got like an 80%. That would just be too obvious. I copied off this smart kid sitting in front of me. He doesn\'t know it to this day. I got away with cheating. Most people find it a bad thing, and I do admit it is very wrong but it motivated me. I never believed I could do anything from all my experiences in the previous years. And then I realized here was this girl who just got away with cheating, you know wait a minute but if I could cheat then I could do anything.  Grade seven rolled along. My life changed. This guy Joseph that was the only one who talked to me all the time in grade six stopped me from taking my life. I mean that was a horrible year. I found out he was moving away to a country overseas. I was scared at first wondering what I would do without him, but once he left I realized that he helped me get through a rough time and now all I can do is just make friends, be happy, and try my best. So that is exactly what I did. People came from different schools and most of them didn\'t know me. It was my chance to start over. I made honor roll every term that year. I went from the girl with a D in math to the girl on honors. I did it without cheating. I never cheated again after grade six. I realized to work hard and to fail is always better than to cheat. And when we work hard and we succeed it is an even greater reward because only you know truly how hard your worked for something you wanted to badly. I also used to have a really bad acne prob and like all over my back too. That also why ne of the reasons why no one liked me in grade six. I was \"disgusting\" according to most guys. I couldn\'t really change anything though, you can\'t change genetics. Both of my parents had acne, which means 100% chance of me getting it. By grade seven though, no one really cared. In fact, today I have it under better control after trying like 5 million things I finally found something that works pretty good. A lot of guys say I am hot and beautiful now and I get hit on at the gym all the time. It\'s pretty funny.  I worked hard in grade eight and managed to make honor roll only once. I worked hard though and everytime missed honors because of my math mark. Always the math. I guess you could say I tried new things though. I used to be reserved, quite, shy, barely talked. I had played Soccer for many years, even made div 1 before (and yes I am still pro at soccer) Anyways, I tried out for the Senior girls volleyball cause I am a sports person. I had never tried Volleyball before. Out of the girls there were over fourty girls to begin with. I made it into the top twenty but didn\'t make the team. So I tried for juniors and definitely got on. I wouldn\'t say I was nearly half as good at it though as Soccer, but it was fun, different, and at least I made friends and talked with people. I also tried out for the badminton team, I made the cut for that. Most games I lost, but really I had fun. I am way better at badminton today than I was back then. I sucked at doubles though, I realized that. I am the dominant one...likes to take control. I rock at singles badminton.  So grade nine eventually came along. Nothing really changed much except I went into another state of depression for a while. I admit I was unhappy for various reasons. Well, first of all English comprehension was a struggle like usual, but I was a good writer. There was no doubt about that so the teacher always assumed I didn\'t need any help because I did good on English. Social was just one big confusion but for like everyone. Our teacher always gave up tests that were like two grade levels higher and would do stupid things on M.C. tests like made eight answers in a row all A so like half of the class would fail. It was the class where my ability to write didn\'t pay off cause I had no idea what was going on half of the time. Math was the worst. What a nightmare. I still remember being kicked out when I said something \'apparently\' rude. I wasn\'t trying to be rude. Half of the teachers just thought I was doing it on purpose. Anyone who knows me knows I would NOT purposely try to be inconsiderate. I quit my job that year, the only job I have ever had in the last well about four years. Eventually balancing school and work became to much. It was just too difficult. It was either fail or quit and pass. So I quit because I wanted to get into all the highest possible classes in high school. I wanted to be with all my friends. Getting the mark I needed for math pure was by far the most difficult, but I did it. I sure proved that I could do it to. I have passed math 10 and 20 with fairly good marks now - better than the marks I got in j.h. math. I always had this knack for science though. That was one my things. I studied day and night to get the marks I got in that course. I think was only like a low 80% but that is good and I work damn hard for it. Science interested me and still does today.  It was strange though because in grade nine I also ended up with Drama as one of my options and I didn\'t even put it as one of my choices. I hated it the entire year, I complained about it every day but I was good at it. When the head of the board for drama or whatev her title was came, the teacher told me she wanted me to do my monologue first cause she needed to make a good impression on this lady. I never felt so good after that. Everyone - even people who never talked to me, they came up and told me that they were scared to go up now because mine was so good. I know mine was good. I realized I had a talent. Drama, acting, performing, whatever you want to call it. At the end of the year the school gave out medals to those who got the highest mark in all the classes except like comp and things like that. Then they called out: Top Drama........Vanessa Boila. I knew it too, in my mind I just knew it would be my name. I had a 98. To beat me you practically had to be perfect. I had hated Drama so much that year, and initially what most people don\'t know if that when I chose my courses for high school Drama is not what I initially chose. When I got my schedule mailed to me for grade ten I looked at it and was like: This isn\'t right. So I went to the counselors and switched fashions to five credit Drama. That was fate right there working in the hands of time. I could have not been more happy. I met Megz, we are awesome together, partners on stage. We both played fairies in A Mid Summer Night\'s Dream and we do tons of stuff together now. I know for a fact after high school we are going to stay in touch. There is no doubt in my mind.  Grade ten I would probably have to rate as the best year of my life. Sometimes I wish I could just go back in time and relive that year. I mean for the most part it was good and less stressful than many other times in my life. It was kind of embarrassing though when I was asked \"who\'s your T.a.\" and well I didn\'t remember. At least I had a good excuse though. I mean we had gone through three other people before my permanent T.A. came so no wonder I didn\'t remember who anyone was anymore. High School was a big change for J.H. I realized that there were teachers that cared, ones that did hate me cause I wasn\'t \'perfect\' or just didn\'t understand. They didn\'t, blame or critisize, they just accepted whatever came their way. That kind of difference goes a long way. No teacher had ever asked me about anything I was ever interested in before like Mrs. J. No one ever cared the way she cares. In High School there are teachers like Ms. L who even though they appear scary they encourage you, push you, and actually BELIEVE you can succeed. That is the biggest difference. Teachers in j.h. would just assume that I wasn\'t capable, but then high school came and here were a whole bunch of people who just supported me, accepted me for who I was, and believed that I could succeed.  That doesn\'t change the fact that I struggle with math and some other things. It doesn\'t change the fact that I failed half of the things in Chem. Really it\'s not that big of a deal. I learned a lot in that class. Directions for labs confused me a lot. A lot of times I just sat there watched everyone else cause I was confused. I wanted to understand though and I tried to participate as much as possible. Math was difficult, but a lot more difficult in grade eleven. Especially because the teacher more like ignored me pleas of \"I don\'t understand.\" Everyone is like: Oh just practice and stuff. I did so bad on most of the tests except the one where she actually sat down and went through the stuff with me. Once I even told her specifically that I didn\'t understand and need help, and the teachers reply was: \"Oh do you have ____ day free, ___ time free\" but like I just simply was not available those times. Then she just said: \"Well I can\'t help you then\" and I was completely left in the dark for that unit. The schedule worked way better the first year. At least there was time to eat lunch.  In grade twelve I need to focus on getting into university and keeping up that marks. Ever since grade seven I have known I want to be a teacher and since grade ten I have known I want to do something in the science area. I don\'t care how difficult it is cause I am going to do. I have gotten this far now there is no turning back. I just got to keep trying my best. Trying and working my ass off is all I can do. I want to major in Bio and do a minor in Chem. Yup, that\'s what I am going to do. Diploma\'s are a stupid part of grade twelve that I don\'t want to have to deal with, but that is just a fact of life. Before I know it they will be over and I will old...ahhhhhh. I want to meet someone who is not into himself and an asshole. Can\'t someone half decent come along...???????? I miss Aaron from Halifax but chances are I won\'t see him again. I don\'t know if I should feel this way or not, but I think that I am attracted to Brent. I don\'t know why. A lot of people would be like: \"Why him?\" you know and well he is easy to talk to. Once I got to know him better I found he was pretty nice and is fun. He always says I am smart too, while most guys claim that I should have been born a blonde. Most guys like me for my looks only and that is annoying. I have always wanted to marry like a crazy hot science guy though or a teacher, or someone totally into sports. My requirements are:   Must like to go the gym often  Likes sports and puts priority of having fun first  I get to shop and cook, the guy cleans  Must want two children  Go to Catholic church on a regular basis or does not go but is eagerly willing to go  Must be a part in the kids lives, I don\'t want to do it alone  Must have a successful career or head in the direction towards one  Likes to eat healthy, no junk in my house - No pop! No chocolate!  Does NOT speak another language, I hate that  Does NOT have an accent cause half of the time I don\'t know what those poeple are saying  Likes to go out a lot and get away from the house  Likes cats  Funny and good personality  No really long hair...yuck! (Must be well groomed!)   Ok yeah that like has to be the longest post in the history or my journal writing. I think I am done now. My head hurts from thinking so much.   '),(3925,1807720,'06,August,2004','2004-08-06 00:00:00','    My mom and I picked up my sister today and went to Doreen\'s condo in Canmore for breakfast, then went to Banff later on.  We shoped for a while, then went to lunch, and got lots of ice creak and chocolate from my favorite candy store where you can look the the window and watch them make the chocolate.  We went back to canmore for about an hour after that, and then dropped off Melissa, and went home.  That was basically my day today.  And well I need to go watch the news now - love the news.  And well I have no idea what I am doing tomorrow.  A couple more weeks to my birthday...maybe two or something like that.  I am looking forward to generous gift of money which I most desperatly need.  Since you know I like killed my debit card when I went shopping with chels I am like going to have no money for gas.   '),(3926,1807720,'05,August,2004','2004-08-05 00:00:00','    Finished! Done!! No more Social for the rest of my life!!!!! No more Summer School. It feels so good. The diploma I wrote today for Social 30 wasn\'t even scary. For once in my life I didn\'t panic or get nervous. I was clam. I don\'t know how I did it. It was a lot easier when your not freaked out and panicked. There was also LESS people in the comp room writing the essay, and less people in the gym. It made me feel not so cramped and more confortable. I didn\'t feel pressured or anything. I guess because it\'s summer I just relaxed and took it easy. My essay was the best thing ever. If anyone marks me bad on that thing I am going to have to kill someone. I know for a fact that it was good writing. Damn good writing. The multiple choice was a little difficult for me, but that is always to be expected with those tests. I find them a lot more difficult than the writing part. Anyways I am sure I did fine. Everything always works out. Just one way or another I always find a way. It was kind of strange during the multiple choice though. I wasn\'t sure if it was the lighting in the gym or my eyes or something. You know those questions that involve reading an except or quote, well when I was reading those it was hard to focus. The words would like \"mix\" togehter and it was hard to read it and understand it cause I couldn\'t make out what I was reading. It was like a blur of dark spots and I like blink excessively, which can be extremely annoying for like half a page of reading. If I just tried really hard to hold my eyes open without blinking, it was better. Except, it was hard to dom or something, cause it totally sucked. But whatever, it\'s over now and done with. It was probably just the lighting in the gym. Other than that I don\'t really know how to explain it.  I haven\'t had a lot of time to do one of my favorite things - write poetry. I don\'t really care if a lot of people don\'t like it, cause I do. I like it, and that is just me. Here\'s my most recent poem I wrote, I don\'t know if this one is half as good as any other ones I have written, but I guess I will still put it in my diary:     Tower Twelve    Alone at night I am left to fear Darkness of this night so queer For when I reach the watch of doom two arrows then unite as one  These hands they strike in the name of vein Forgotten solidarity is but yet contained For on this night I am to discover what remains This darkness has left me on the verge of insane  Cannot uncover a way to depart from this tower I am caught up in this daunting midnight hour Chasing stairs far beyond my reach Stumbling upon an appalling fatality in which my eyes do reveal  Before me lies a rusty board covered in such scarlet stains Flooded across the broken plank my eyes do not deceive And through the air such an awkward silence lingers Misplaced is this sense once so common to me, what shall I do next?  My breath becomes grave and my figure turns stiff As I look the rusted window I can hear such a riff The hands belt twelve and the sound hammers hard I remain in the tower of doom as two arrows become one  ~Vanessa   '),(3927,1807720,'04,August,2004','2004-08-04 00:00:00','    It\'s funny you know. I am the girl who is always so worried and concerened. This time is different. This time I really mean it. I keep avoiding studying for the diploma. That is not like me. As you can tell, I am avoiding it right now by going on the comp, today I went out for breakfast, then shopping with Chelsey, then out to eat, and for a walk. The list goes on. I mean I know how much doing good on this means to me, but I guess studying has not paid off in Summer School. Everytime I studied for a test I either failed or almost failed. If that is what I got on those then I mean, really why bother studying. I am not saying I didn\'t study at all. I just didn\'t study as much as usual. I remember most of the material and I KNOW that I am an amazing writer and from MANY people have been told I am very talented in writing. I KNOW that. I need to prove it tomorrow. The Summer School Social teacher loved my essays except one. The essay I know is my thing. MC is a little scary, but I mean you practically can\'t study for that shit anyways. All you do is read things and interpret - comprehending material...AHHHH. It will be ok though, everything always works out. Besides if I want to be a Bio major they aren\'t going to care if I am not a hard core history nerd (Well except Model UN cause that is just fun) and well I know it means a lot to me to get into university, but I have learned something greater. I CAN accept things the way they are now. I CAN deal with failure. After all I failed a lot of things in Social 30. No, I am not saying I am going to fail the diploma. I am simply just going to try my best like always. That is all I can do. The rest is up to the other half of the world ----> My fait lies in the good ole\' scantron machine and markers in Edmonton (Which are damn well going to give me at least 80% on my essay, I am going to make it so good that we will feel I deserve nothing less. After tomorrow I can stop worrying about stupid Social 30 and Summer School. My Summer can OFFICIAL begin. I can have more fun and not worry for a while. I like that idea!   '),(3928,1807720,'03,August,2004','2004-08-03 00:00:00','    Yesterday was so amazingly fun, but so amazing weird. I went to Bragg Creek for the first part of the day - ate a really awesome lunch and have some chocolate chip cookie dough ice cream. That was so good, and like that is the like 2nd week of eating ice cream every single day. I haven\'t gained a pound. I have been like hungry all the time. I practically eat like six meals a day. That is pretty sick when you think about that. Six meals a day!! When I got back I felt really stuffed from that ice cream I ate so I went to the gym. I had not been in a while so I thought it was best to go. I am in pretty good shape still, but usually I can run 30 mins straight on the treadmill no prob, but I was kind of hurting yesterday and couldn\'t do as much. That was only cause I haven\'t been keeping up with it more often than usual.  So then I like ate again when I got home from working out. I went to Wendy\'s to pick up Chels from work, then stopped at her house, and went to....THE PARK!! Yeah ok, it was only to talk. We just talked about what we wanted to do the rest of the night. I was like \"lets go buy a game we don\'t have or something.\" So believe it or not everyone took to my idea. So I drove to Walmart and we checked out the prices, but it was all too expensive. We wanted hungry, hungry hippos but it was over $20. So then we went to walmart and checked out the prices there. Hungry, hungry hippos was more expensive, but then we found some like no name brand game. \'The Frog Game\' or well Hungry Hungry Frogs!! Before we played though we were all starving by 8:00 PM so we went to Moxies for like the second week in a row. We all ordered chicken fingers and fries cause Mike was working and he is responsible for cooking deep fried crap like that. Anyway you know how they like never give you enough of the sauce stuff I asked the lady \"Do you think I can have extra plum sauce with my meal\" and of course she says \"No Problem.\" A couple minutes later our meal shows up and guess what she brings. She brought a plate of FOUR extra plum sauce, but Andy and Chels didn\'t want theirs so they gave me theirs to eat too. If anyone knows me, I HATE wasting food and would rather eat it then leave it knowing that there our people in this world who don\'t have as much as we do. Anyways:   RULE 1: Don\'t ask for extra plum sauce RULE 2: If you are a retard and do ask for extra, be SPECIFIC RULE 3: If they bring you too much plum sauce, don\'t eat it ALL    So anyways I slowly worked my way through eat little container thing. One by one. So much friggin plum sauce. Too much. Where was my Dad when I needed him. I mean I sure could have used some help eating that. I am guessing I probably ate like half of the bottle. By the end of my meal I had ate four of those container things of plum sauce and one container thing of ketchup. I thought Chels was going to get sick just watching. I had a stomach ache after that. When I was driving everyone back to my place, we put on that song by Ace of Base - Cruel Summer. We sang along with it but in the words: It\'s a cruel, cruel, cruel summer, we sang to: It\'s a cruel, cruel, cruel Dinner! That was so hilarious. When we got to my place my mom thought we were all drunk. She was like \"What did you drink tonight?\" We went to my room and played hungry, hungry frogs for like hours. That was SOOOOOOOO much fun. I mean I have NEVER had that much fun playing any game before. It\'s like my new favorite - Hungry, Hungry Frogs. Best game ever I swear!! I encourage all of you to go out to Zellers and buy it! It\'s half the price of hungry hippos and twice the fun!   '),(3929,3840157,'04,july,2004','2004-07-04 00:00:00','   Last two nights dream about Kevin,luckily last night don\'t have.hehe...I even forget the content of the dream liao...Well,yesterday is quite a nice day aS we got 5 periods have no teacher.Hooray!I kept on&nbsp;talking and discussed about the trip with others AJK.Then,Khai Wan and I ended up go to phoned the agency again.Then when KH period,we practised for the Litereture Drama again.Sometimes.....I just feel bit boring.Emma came to my class because it\'s Morale period.She brought a purple flower and said it\'s someone gave one.Then I found a funny picture in her magazine and I almost showed to everyone.In bus,Pei Shi and I also have a nice chat. &nbsp; At night,I chat with Chen Yeong and Shan Lei.They seemed can get along well.yeah!hehe....it\'s quite a nice day though.. &nbsp; &nbsp;   '),(3930,3840157,'04,july,2004','2004-07-04 00:00:00','   A OK day,rushing homework and play gunbound(just downloaded yesterday).Then weetat phoned me to help him do homework.He too kind to help Ca Wen do homework(I also)hehe...Then go to school,chat with shan lei,but a bit cold because nothing to talk.Then when in class,I nearly fall asleep when geography period.When it\'s KH period,some other friends and I acted the Prisoner Of Zenda.I\'m Sapt,Khai Wan is Rudolf,Hazel is Johann,Ting is Princess Flavia and etc.It\'s fun.But Yan Yi insisted Romantic but in the end...it seem to turn our to be a funny drama as Hazel is just so funny.She is so interesting.WHen I go back home,my attitude to Emma and Pei Shi not really good.I just don\'t feel like talking.sorry...... Well,not a bad day after all..hehe   '),(3931,3840157,'04,july,2004','2004-07-04 00:00:00','   Well,tommorow have test and i haven\'t finished studying and doing homework.I just feel bit annoyed and stress.Actually today after tuition,I going to buy taufiq present but then I couldn\'t make any decision.So,I mostly will buy next week.Just now watch the TV 8 show,Cried until so badly.Keep saying sorry and asked God to granted their wishes.Sigh....life is just so unpredicted.                In darkness I start to find                   Just want a dim of light               No growth                  No voice                     No strenght                         No movement               What am I?                Cries linger in the room under the moonlight               Pain tortures me Orders stayed with me               When can the little wish of mine               Ever been granted?                Looking through my hopeless sight               Seeing them caring me with endless love               Until the end of their life.               When can I have voice to call \'Mum\' and \'Dad\'?                              Have you heard a sound from me?               From my heart?               The sound of my broken heart... ...  Save every life that you can,Good deeds shall never end. May God bless everyone....   '),(3932,3840157,'04,july,2004','2004-07-04 00:00:00','   Hehe...My day really become much better only a bit spoilt by the KRS.I today with LAy Yen ,Angelin and Ting present for the literature\'Prisoner of Zenda\'.I quite enjoy practicing but very tired.I breathed hard.But the performance was quite a success.But got an accident part is Lay Yen\'s ruler was broken.   Then when BM period,Kid Yie and Farha presented the \'Bahasa\'.At first,I and Taufiq busy discussing.Then Nazalea,Lek Wen and Arvind also gave their opinions.it sounds liked debate and it\'s very fun.   When Moral period,I had played truth and dare with some friends.It\'s kinda of fun.Encik Lim also started to recite his life experience.hehe...Then I today only know that Emma didn\'t believe in spirits.   Today got little rain.....but still a good day ^-^   '),(3933,3840157,'04,july,2004','2004-07-04 00:00:00','   This morning,Yean Ruey phoned me and told me that the photo had being deleted.Yeah!   Well,eventhough Ting didn\'t come to school because of piano test,my mood is still quite ok.I just felt a bit lonely.Luckily,Bryan got came and accompany me for a while.But very tired cause kept chasing him to get my diary back.QUite fun too..haha,seem to back to primary school. Homework seem stacked up.But still lazy to do,hehe.I know I shall be doom(hope not lar).   When recess time,Siew Huey came and talked to me.Both of us had a laugh quarrel.Sometimes I think,shall I give up this friendship?My heart is just to soft to do it.hehe...   Oh,when Morale period Emma gave comment on my article.She said it\'s not bad and she helped me to correct my mistake.Thx ya.I maybe try to post this article to school magazine.   I also had a litle chat with Tang.I begin to know him a bit more better.Oh,I also got chat with Sheau Tian.Today is her birthday and she receive lots presents.I got chat with her about she and Kelvin.Hmm....hope everything is alright.   Well,quite a nice day though,thx god  ;)   '),(3934,3840157,'04,july,2004','2004-07-04 00:00:00','   Don\'t know why,these days a bit unlucky,hope everything will be fine soon.Siew Huey is angry with me again.The reasons is:I didn\'t listen to her when she talk.If I really did that,I\'m very sorry.But I don\'t know what can I do now?Apologize?I not really sure for that.hehe...I\'m a very bad girl..haha   Yesterday got replaced class of Science.Before going to the laboratory,Vinod came and had a chat with Ting.I was so bored that I opened my History revision book and read few sentences of it.But,this is just not my type.So,I go find Tang for a chat.I asked him the way he study,his computer game and his sister.Quite a nice chat ^-^.Then I went to the lab with Ting and Hazel.Then Hazel suddenly said something that make me puzzled.ok,i knew it,I did something wrong again?   Oh that day is Khai Wan b\'thday too.She seem very happy and her eyes seem have shone lights.That\'s just so special.Then I handed her a persent.I so glad that she seem very happy that day.I guess she had a great b\'day.   Then at night,I met Mark online.Chatting with him made my mood become much better.Thanks for him so much.   But still...I just maybe need to change some of my irritating attitude.sigh....Hope everything will get better soon.   '),(3935,3840157,'04,july,2004','2004-07-04 00:00:00','   Today still quite worried about the pic.haih..ANd today Yean Ruey called,we have a chat.His voice really a bit loud(hurt my ears). Then went to school,I handed the chocolate to Ting and Lay yen.I also chat with Shan Lei about the IU day(but mostly got class tambahan).I don\'t think I\'ll be going to the IU day.   Then when back to class,I threathen Bryan to delete the pic with his cow.We two start a fight but i managed to get the cow.But then,he used the teacher to threathen me.....sigh....a bit angry with him,but then he got said sorry,so everything alright.I felt very sleepy at the Science period and become energitic when History period.   When recess,I listen to Fhei Fhen mumbling about Weetat.Weetat just sometimes not that gentle to girls lar.Too bad for that.....When Chinese period,I and Emma kept on talking and talking.Quite a nice day to chat.   At last,it\'s Math period.Puan Katinah was sick but she still insist to teach and gave lots homework.Today was my turn to call people clean up the class but they just didn\'t listen to me and making fun on me.I\'m very irritated to that.Luckily Taufiq got helped me do some work.Thanks him a lot.   Hope my luck will turn better soon ^-^   '),(3936,3840157,'04,july,2004','2004-07-04 00:00:00','   Yesterday evening,i suddenly have a feeling to have a great chat.So,I picked my my phone and pressed the buttons. \"Hi,ShanLei.Haiching here.You busy?\" \"Hi,not really.\" \"So,how were you with chui wei they all?\" \"Oh..you want have a chat is it?\" \"Yup\" \"Wait ar..I go off the fan first\" Then we start to chat.I can feel that she quite enjoy herself with chui wei,jia hui they all.They seem very friendly to her.Many interesting happens at city square.I can\'t just stop laughing.Then,we chat gossips,my old fren and us.It\'s so enjoyable chatting with her.hehe....I\'m such a chatter box.We chat very long until my mum called me for dinner.Have no choice ,I had to hang up the phone.Hope can have such a nice chat next time ^-^.   '),(3937,3840157,'04,july,2004','2004-07-04 00:00:00','   urlLink    SIn Cheng Yeong and his hotdog stall&nbsp; urlLink   '),(3938,3840157,'04,july,2004','2004-07-04 00:00:00','   urlLink    the 1st performance on stage&nbsp; urlLink   '),(3939,3840157,'04,july,2004','2004-07-04 00:00:00','   urlLink    the spastic hall.i took the pic in front of the toilet.haha&nbsp; urlLink   '),(3940,3840157,'04,july,2004','2004-07-04 00:00:00','   urlLink    The pink color shirt is Nazalea and the one beside her is her partner.They are FOMAD mc.&nbsp; urlLink   '),(3941,3840157,'04,july,2004','2004-07-04 00:00:00','   urlLink    The red color shirt,cute girl is my friend Shan Lei.I\'m the one besides her.We in the FOMAD.&nbsp; urlLink   '),(3942,3840157,'04,july,2004','2004-07-04 00:00:00','   just created my blog in the very early of the morning(5.58 am),hehe...juz too boring.Well,I\'m so lazy to study as the exam is coming so near.I should start my studying soon afterward.So boring about that.... Last 2 days,I had just gone to a hall for FOMAD(Festival Of Drama & Music).On the way going to the Spastic hall,i get a really good nagging from my dad.Soon,i saw Shan Lei coming towards me.SHe seem moody about the way she going to chruch.Then Chui Wei came and we had gone approached her.Shan Lei is just so shy.Then we went to the counter and I saw Yean Ruey sitting there and hand me the CD that I had wished to borrow.Having a simple chat,we get a hand band from Abel and go into the hall. In the hall ,I just being so attracted to a girl that wear red band.So cool...She\'s the 3rd person I had being attracted too.Shan Lei and I enjoying ourselves playing the digital camera.We had took few photos.Soo,Dharine and Farha come and join us.Silence again....Shan Lei and I decides to walk bout.Then Yean Ruey introduced my net fren to me Sin Cheng Yeong,hotdog stall\'s boss.haha... Shan Lei and I back to our seats and Weetat the gang came.Weetat and Yean Ruey jus came beside me and start to sabo each other like kids.So funny..I just couldn\'t stop laughing for that.Nazalea also meet her partner,a boy from EC(cool hair). Then I borrowed Kenneth my camera,but in the end,he only took 2 useless pic and let the battery ends.Shit lar..I should take more picture be4 hand the camera to him.Sigh... The beginning part of the show is quite interesting.A special intoduction by Li Jie and his singing band is just great.Then drama,singing and....CONVENT DANCE!The girl that attracted me was on the stage and their dance is great.Soon,the show become lame.Shan Lei and I decides to buy something to eat.There got sell sandwiches,hot dog,ice cream,burger and drinks.SO little than I thought.But still I didn\'t buy the hot dog as ONE of my fren said is not very nice.hehe...But Shan Lei still got eat a hotdog with no bread and I went to buy a chicken burger.Then Shu Pei came to me  and we have a little chat. I not really remember what happened next.But nvm,hehe.Shan Lei and I back to the hall and watch a really stupid(funny?) drama of \'The Pearl\'.Then Mark joined us after the drama.He said Shan Lei looked like Chui wei(Weetat said b4 too).hehe...Then cause got a irritating person sit beside me,I quickly went to find Yean Ruey for a chat.But....we just good friends since primary school,and he has girlfriend liao,the other people thought we what what pula.Futhermore still took picture...At first I teasing Yean ruey cause none of my business actually,but then I start to think will his girlfriend come and find me and giveme a slap?Hope not...I and Yean Ruey just a very good friend. Then Shan Lei,go to church with Chui wei and her gang,hehe..She seem enjoy her day. I was so lonely and I phoned my sis to pick me up earlier.After had a short chat with Mark,Yean Ruey,Bryan and Weetat,I went to the main gate to wait for my sister.Weetat and bryan came and acompany me(make fun of me i guess?).Almost after half an hour,my sister came in her boyfriend car.I bid Weetat and Bryan goodbye and went straight back home. I quite enjoy my day as I met many of my friend.Opps forget to say about wei-xun.He had become taller than me......sigh.....   '),(3943,3125462,'02,June,2004','2004-06-02 00:00:00','   In School...    Right now, I\'m in the computer lab, supposed to be looking for information on garbage user fees. But, I\'m checking out Will\'s   Howitzer Puppies . These are genetically engineered puppies, to take over the world. More on this later.  '),(3944,3125462,'03,June,2004','2004-06-03 00:00:00','   I am in charge of the secret farewell present for someone I will name Lousie. Mid-June, she\'ll be leaving for South Africa. It turns out that she has already planned her own farewell party and  everyone  must bring something for her that will remind her of us. So the entire hackey group (except for Thomas) took 15 minutes of Science to do a digital photo.  We took two photos, actually. On the first one, Lauren gave Samie bunny ears.  Real  mature. Seeing as it was a  digital  image, we saw it instantly, and Samie found out instantly. She then, of course, demanded another shot (it  is  Lousie\'s farewell gift after all).       Front of the T-Shirt for Lousie  Low Quality'),(3945,3125462,'03,June,2004','2004-06-03 00:00:00','   Man, Ã‰tudes Canadiennes is really boring. We\'re doing  more  research for our stupid debates. Mine is really simple. The proposition, you ask?  User Garbage Fees Should Be Implemented . The answer is really simple. If Garbage Fees are put into place, then a  whole  lot of illegal dumping will take place.  Yup, you guessed it, I\'m posting instead of working again. That\'s because  I\'m done all the research I need!  Crazy.  Today marks the day when I enter the 21st century. I just bought a new digital camera. It\'s a little piece of crap (600 kilopixels) that is just used for point-shoot-download-then-upload to the web. Expect my photo here anytime soon.  '),(3946,3125462,'05,June,2004','2004-06-05 00:00:00','   Man, oh man, was that a crazy party last night. But I\'m getting ahead of myself. It all started in math class, where every day starts. I mean, seriously here people, math is always the first class of the day. Our regular math teacher couldn\'t be there, because she was coaching the girls\' soccer team. Therefore, we had a sub. He normally taught Social Studies, but was brought in anyways. Seeing as we had a sub, everyone came over to sit at my table, mainly because I\'m so cool. Will and I were working together on the math homework, and the sub came over and left us an interesting anecdote.   You know what happens when you wear a toque like that in the middle of a hot, summer day, don\'t you?   Um.. no.   It\'s the main reason why I don\'t allow hats or toques in my regular Social class.    [obviously disinterested]  Really.   Your toque doesn\'t let any moisture escape, and you know what that does to your scalp, right?    [fairly monotonous]  No, why don\'t you share it with me?    [not picking up on Will\'s hints]  Because of all the excess moisture, your scalp becomes all saturated and pruny&#8212;   uh-huh   &#8212;and then you know what happens?    [evidently annoyed at this point]  No.   Your hair falls out. It\'s an early sign of male pattern baldness.   Wow. Thank you for that disturbing thought. I\'m doing my math now. My toque is still on my head.    After that was &#201;tudes Canadiennes, with  another  sub. As always, even though we already have a load of crap due very shortly, M. Sokalski left the sub an assignment due at the end of class. We had to find 10 difficult words and define them, and then summarize the promises of  all the federal political parties  of Canada. Most people in the class were complaining about the fact that it was too hard, so the sub decided that a communal effort was in order. I decided to forge my own trail and do my own work. To aide my brain, I decided to listen to  urlLink Strong Bad Sings .The sub, however, didn\'t like that. &ldquo;The human brain is incapable of doing two things at once.&rdquo; What he didn\'t know is that Will (who was also listening) and I are  excellent  debators. His main argument was that talking on a cell phone while driving leads to a lot of deaths, I mean, a  lot  of deaths. We countered with the fact that every car is equipped with a radio, and if you can  drive  and listen to music (i.e. multitask) then you  should  be able to  write  and listen to music without major problems. It was also been proven scientifically that the human brain can concentrate on up to 4 tasks  at the same time!  However, since the classroom is an autocracy, he won.  We still forged on our own. I wasn\'t  really  listening to him and the rest of the class, however, I do know that he talked about the Mandarin language for at least 15 minutes. During this time, I became really bored  (wink wink)  and rested my head in my hand. My hand happened to be holding an earpiece. I don\'t know how it got there. After a couple of songs, though, he noticed and threatened to confiscate my discman if I did it again. So, I turned it up real loud (which with headphones in isn\'t loud at all) and listened with it on my desk very quietly. In case anyone\'s keeping score, the class finished about 5 of 10 difficult words, and I finished about 5 words too. However, I decided that a summary of the political partied would be more beneficial to me and finished that as well. It\'s something I call \" time-management \". All while doing the impossible multi-tasking! Hah!  Next came lunch, which was extended by about half-an-hour to incorporate a live band, BBQ, and the  Big Ollie Contest ! It was enjoyed, however, with a good toss of the disc.  Nothing else exciting happened except for the party. The walk to Louise\'s house from Kelvin is about 15-20 minutes. A ride would take about 3. And it\'s less energy expended by my body. So when Samie offered Lauren and me a ride, I naturally accepted. I didn\'t know however, that it would include a trip to Costco and then Samie\'s house. It added about an hour and a half to the 3 minute trip. But it was still a cool party.  Phhew! I have to get out of the house. This will be finished later.  '),(3947,2607577,'21,june,2004','2004-06-21 00:00:00','    urlLink    urlLink audblog audio post   '),(3948,2607577,'21,june,2004','2004-06-21 00:00:00','    urlLink    urlLink audblog audio post   '),(3949,2607577,'21,june,2004','2004-06-21 00:00:00','    Thank you all for you prayers and love.  We had some MAJOR breakthrough just within 14 hours after my last blog.  God is so faithful to us (His children).  We are such a blessed people.  Those who call him Daddy know who I\'m talking about.  Well I will be leaving for Atlanta, GA tomorrow morning until Saturday.  I will however be leaving Audio Blogs on my site so make sure you check it out and listen to what God is doing in Atlanta.  I have been asked to assist the 2st Baptist Youth Pastor (Katie Cox) along with her adult and student youth leaders on a trip to \"Church on the Street\".  We will be staying who knows where in Atlanta, and going and ministering to the homeless on the streets in Atlanta.  This is a very important trip to me primarily because I was asked to go with them.  It is a little unusual for leaders from different churches to link up and do a trip together.  WOW,  God is doing something supernatural in our city.  There are a few pastors that are starting to pray and seek Gods face for what He wants to do in our city.  It is very exciting when different denominations get together and go before the throne.  Transform our cities oh Lord!!!  Most importantly, I want to thank God for just being God.  In control of all of my circumstance and my future.  I am so greatful for His Grace and Mercy, He is the Most Holy God!  Thank you for my congregation, and for my dear friends that would die for me, because of Your love in them.  God you are my God, and I will ever Praise YOU!!!!!!!!!!!   '),(3950,2607577,'21,june,2004','2004-06-21 00:00:00','    Ya\'ll we are in need of your prayers.  We are going through an intense testing period, and we are trying to discern God\'s voice.  I am very excited about all of the awesome things God is doing since we went full-time as missionaries 3 months ago, and I see the immediate fruits of our labor.  Doors have opened in some major ways that have been a confirmation of going full-time, but on the downside, it takes a while to build up support (which I knew going into it).  I am a strong believer that where God\'s vision is, He will provide the provision.  I am also a strong believer in stewardship.  Now, I know that God is Soverign (in control of ALL things), even in the areas of finances.  My heart wrestles with the fact that God is providing provision (miraculously) in every area except finances.  Now, I have no problems believing for finances.  Matter of fact, there was a time when I gave everything I owned away to serve Jesus (cars, computers, T.V.s, stereo systems, $4000 diamond ring, etc).  I have no problem giving God ALL the Glory, cause EVERYTHING I own has been given to me by Him for me to steward.  When I gave up my Job, I got rid of my 2003 Cadillac Esclade, the day I got rid of it, someone gave me a 2001 Nissan Pathfinder (Loaded).  You see, I have no problem giving abundantly, NONE!!!  Where I need prayer is discerning TIMING!!!  I know that I have the favor of God on my life and ministry.  What I need prayer for is when I set my life before Jesus, and He provides all of my needs (which He has), but there is that stewardship thing.  My house payment was due on the 18th of July, and I don\'t have the money for it.  Let\'s get something straight!  I am not writing this to get some emotional christian to save the day, I need PRAYER!!!  I have to be a good steward of what God has given me, in all things, according to His will He provides ALL my needs.  When the bill doesn\'t get paid.  I have to look inward.  Am I being obedient, is there something in my life that I am not dealing with?  Have I missed the timing of the Lord?  Please know that I am not shakin\' or lacking in faith, just sharing what is in my heart.  I just want to know God\'s will for my life.  I wanted to share this with those who care about my life, and obviously you are interested in whats going on in my life (or you wouldn\'t be checking my blog).  Please pray with us and share anything that you get from your Spirit (not emotion).  We love you and thank you for your prayers and love!!!   '),(3951,2607577,'21,june,2004','2004-06-21 00:00:00','    As many of you know, I work out on the streets.  One of the areas that we go is to a place called Hospital Heights.  HH is a neighborhood with about 130 homes, and about 300-400 people.  HH is known as being the worst area in our city (drugs, theft, fighting, abuse, some minor prostitution).  In HH there are 4 known drug houses that are constantly moving and shaking all day.  Anyone that comes from HH is automatically stereotyped as trash.  Since we have been out there, I have found that to be FALSE.  These people in HH are some of the most loving and caring people I\'ve ever met, the only thing is, they have been hurt serverly by the community.  These people have up great walls due to the fact that they have been mistreated by our community.   In Janurary 2003, we started a house church in a garage in HH.  Since then we have consistently had 15-35 people there every thursday evening.  Well the house church wanted to have a cookout for the neighborhood as an outreach to their friends of the neighborhood.  WOW, did GOD show up.  We had over 70 people show that would have never stepped foot into a bible study/church.  We fellowshipped, and just hung out and loved on these people (all ages), and it was just a great time of community.  God gave us an idea that we should take some food to the house next door, which has been know for drug activity.  So we did.  We took over plates of Hot dogs and chips.  About 20 minutes later, one of the main guys (the one who hasn\'t liked me since he met me, and the guy who just cussed me 2 days before) came over to the church and walk very aggressively through the crowd, came up to me and said, \"Hey, I wanted to come over here and thank you for the food you sent over.  That was very nice of you guys.\"  Of course, when I first saw him coming, I have to tell you I was very unsure what was going to happen.  I saw a side of that man that many people probably never see.  I swear I saw his heart fully visible, genuine in all his ways, sincere to the depths of his soul.  I saw a man that I really liked.  This was a tremendous confirmation of Gods works here in HH.  It made my heart so greatful for all of the seeds that we have planted, and the increase that God was bringing through it.  It\'s funny what God can do with a HOT DOG!   '),(3952,2607577,'21,june,2004','2004-06-21 00:00:00','    I must have gotten up on the wrong side of the bed this morning.  Everything is like slow motion.  I am sitting hear with absolutely no thoughts of anything.  Actually, it is very difficult writing this blog.  Anyway I\'ll press through.  I usually have one day a month (my time of the month) where all my fruit seems to be gone.  When I have to have a day just to myself, with little or no human involvement.  For any of you who know me, you know I love to be around people.  But during my time of the month, I don\'t want to talk or see anyone, I am depleted of all relational juices.  It\'s a good thing though, I get to spend a day resting in the Lord and soaking up the Word without the worry of people coming and taking the fruit off of my tree.  Don\'t get me wrong, I am a guy who loves to freely give my fruit to whoever needs it.  That\'s why Jesus gives us fruit, to freely give it away.  But guess what, today is all about ME!!!  Me and Jesus, I am gonna crawl up in His lap, and rest my head on chest and let Him hold me ALL DAY LONG!  I am so blessed that I am a child and not a slave.  I have the responsibility of doing my Fathers work, but at the end of the day (unlike a slave) I can walk right into my Daddy\'s throne room and get up on His lap and He listens and loves me, cause I\'m His son!  WOW, what an awesome revelation.  Sonship!!!!  I love being His son, see I\'m already in a better mood.  Just thinking about how much He loves me as His son...  It truly changes your perspective.  That\'s enough from me, I am going to the LAP of JESUS!!!!   '),(3953,2607577,'21,june,2004','2004-06-21 00:00:00','    As I go down this journey of faith, I learn new and exciting (and sometimes painful) things about myself and God.  The big three (Purpose, Plan, & Timing) that must be in place for spiritual success and growth.  Of course, along the way their is our friend \"the Process\", that brings us to and through the big three.  As I go through the process I learn more about God\'s love and grace, and a whole lot more about my wicked heart.  I get busy in my mind and works and many times I forget to include God.  Then when I come to a wall, I say God what are you trying to tell me.  I sometimes hate the fact that I am intelligent, resourceful, self-motivated, a doer, a forerunner, a breaker, and of course so dog-gone good lookin\' (just kidding).  It is a dynamic tension that I struggle with daily.  Using and submitting my DNA (how GOD designed me) to His process.  I am made a specific way, to not be that way.  For my own gain that is!  God designs each of us different and special.  Special in the way that no 2 people are the same.  I bless God for His process, and I submit to the process.  I don\'t want to be 10 years down the road before I realize that God has been waiting to use and bless me this whole time.  I need Gods chastening in my life, He knows whats best for me.  I am in the heat of the process right now, where the heat is being put directly to all of my person.  Drawing out the many impurities in my life, all the hinderences that seperate me from God.  Everyday one step closer to you Lord.  It sucks, but it\'s lovely.  Two extremes present and pulling, and somewhere in the middle is Truth!  Allow God to take you through the process, be refined by the fire.  It is worth it, all of it!  We must see the beauty of the process and embrace it wholeheartedly in order to get where God desires and had destined us to be!   '),(3954,2607577,'21,june,2004','2004-06-21 00:00:00','    I was so blessed last weekend when I came into church and there stood Jack Hannum.  You see it didn\'t surprise me that Jack came to church.  Jack was a Pastor.  It wasn\'t because he was a fallen pastor, or back sliden in his walk with Christ.  It was because Jack is the Associate Pastor of 1st Baptist Church in New Castle.  Jack had never been to our church, and in my home town you don\'t see this sort of thing happen.  I excitingly said, \"Jack thanks for coming, what are you doing here\", and he said that he was on his last day of vacation, and wanted to visit our church.  WOW, a Baptist Pastor coming and visiting a Full-Gospel Non-denominational church.  You see, Jack is a forerunner in our community.  He is what we charismaniacs call a \"BREAKER\".  Jack has a deep passion and anointing to be a bridge builder to bring churches together.  I was so blessed to see Jack at our church, blazing the way for unity within our community.  If our cities are going to be transformed, we are going to have to get a revelation like Jack, that it is going to take a joint effort, not just one church to change our communities.  I thank God for Jack, his heart has taught me a valuable lesson.  So if you know Jack, thank him for his heart, and for his obedience to be Christ to the churches and people of our city!  We love you Jack!!!!!  So get out of the box, and play THE WILD CARD !!!!!!!!!!!!   '),(3955,2607577,'21,june,2004','2004-06-21 00:00:00','    This is a word that we are just unfimiliar with these days.  With the lack of community and commitment in our society, we truly are a people who lack availability.  Sometimes I think that being an American is a curse, due to all of the activities that the world has to pull at us daily.  I sometimes wish I could move to that little town in \"Northern Exposure.\"  I remember when I was working in my clinics full-time and a need arose, and I couldn\'t tend to it the way I felt I need too.  Just the fact that I was tied down to something that controlled my ability to serve someone else, really discouraged me.  I have to tell you all, now that I am in full-time ministy, I LOVE MY JOB!!!!  There is nothing more awesome to me then to be available to be used by God.  People miss out on sooooo much just because they won\'t make themselves available.  I\'m not saying that everyone should quit their jobs and go into full-time ministry(unless that is what the Lord is telling you to do!), but there is a fulfillment that is beyond comprehension when God uses you to be a vessel for Him JUST IN TIME!!   For example, Saturday evening at 11pm I received a call from a 16 y/o boy from the neighborhood that we minister in frequently.  He was very upset, cause a girl he knew was getting ready to commit suicide.  He was very frantic and concerned for his friends well being.  I told him that we would be over immediately.  So I called a dear and anointed brother of mine, Chris Hanson and we went over to their home.  When we got there, she was crying and shaking and desperately in need of Jesus.  Chris and I talked with her for 2 hours, and at the end of our conversation, she agreed to get some help.  Sunday evening myself and my wife (Jessica), and Chris and Cheryl went to see her.  She was still a little shakin\' and withdrawing from her drug addiction.  Currently, this 18 y/o girl is staying with us in our home so Jessica can continue to minister Love to her and to keep her engaged so that she is less likely to make a bad choice and go back to the drugs.  She is doing great and she is getting ready to have an interview with Teen Challenge, and hopefully within a week she will be off to T.C. It is funny how God will use someone if they will make themselves available.  For those who won\'t make themselves available, I find them looking at us in a very unique way, wondering why we do what we do.  All I can say to them is scripture, in Luke 7 it says \"Those who have been forgiven much, Love much\".   And believe me, I know how much I\'ve been forgiven!!!  I think understanding forgiveness is something that is lost these days, but I\'ll right more on that later.  I love being available for the Lord, in whatever way or capacity that I can.  It\'s not out of duty, but out of dedication to His Love for me and for His children that are so desperately in need of His love.  We have to get that revelation, we may be the only Jesus that some will ever know!   '),(3956,2607577,'21,june,2004','2004-06-21 00:00:00','    urlLink    urlLink audblog audio post   '),(3957,2607577,'21,june,2004','2004-06-21 00:00:00','    When I was growing up I never really new what life was all about, and quite frankly I didn\'t really care.  life was all about me and mine.  What I could get out of it or someone.  Boy was I wrong!  As I grow in my walk with Christ, I find that my heart and joy comes serving others with disregard for myself.  I look around today and I see so much depression, loneliness, and lost identity.  People desperately want to be significant in their lives, but they just don\'t know how to find it.  You see, the world determines ones identity by what someone does for a living.  We always ask, \"So, what do you do for a living?\"  This seems to determine so much in our society.  We try to prove ourselves based on what we do or have accomplished, instead of who we are in Jesus Christ.  The sad thing is the Church as a whole has been infected by this virus.  People have a hard time engaging in church without being a part of a position, or having a title.  I believe that God desires for us to have our identity through understanding who we are in Him.  We are made in His image, our example is Jesus, and we are joint heirs to the Kingdom of God.  Many of us are slaves to God, we think as long as we work and do things in the Kingdom, we will please God.  I tell you, God is not looking for slaves, God is looking for Sons and Daughters.  Him our maker and we the created being.  God delights over us (His creation).  He doesn\'t want us to be like a slave that does all the work and then retreats back to our servants quarters.  God says,\" My son/daughter, thank you for your work for me, now come on up here in My lap and tell me about your day.  After that I will hold you until you sleep, and I will tell you of all the great things I have for you.\"  God desires to have INTIMACY with you.  He loves you with a searching and conquering love, that brings joy into and out of the depths of our being.  That joy is what I seek.  The joy of dwelling daily in His presence.  The joy that changed my life!!!!!!!!!!!!   '),(3958,2607577,'21,june,2004','2004-06-21 00:00:00','    On my blog entitiled \"Wonder what they\'ll say?\", I spoke of myself and Tom Preble (Chris Kelley went too) going out and asking some very specific questions to random people on the street.  First of all, we went to the mall and after interviewing 5 people we were asked to stop taping in the mall (liability in case someone went postal on us).  So we went to the village at BSU.  In our findings, we noticed that there were no two answers really the same.  We asked questions on Sin, Absolute Truth, and Trust.  I think the closest answers were based on trust, primarily because each person had been burnt by a friend at one time or another.  I have been doing a lot of study on the PostModern culture lately, and first and most importantly, there are no real definitions to anything.  Everything is relative to each individual person.  There are no absolute truths that apply to all people, nor absolute authorities.  On our journey, we found that this is very true with the people we interviewed.  We received an incredible amount of different answers in all of the questions.  Their answers were truly based upon their individual mindsets.  This was very disheartening to me, but in the same sense didn\'t surprise me at all.  It is what \"Bob Preston\" calls Dynamic Tension.  The battle between two extreme differences tugging back and forth at each other, and somewhere in the middle is the TRUTH.  I believe that there is a real identity problem in the world today, this is why there are so many definitions in society.  I mean, everyone desires to be significant, and needed.  Everyone wants and desires to be understood (as my friend Tom puts it).  Everyone is trying to find out what their individual purpose is for this life.  It is like walking around in the daytime with blinders on.  People that are trying and doing and being a part of anything they can find meaning in, but my question is, \"is there really true fulfillment?\"  I would think when someone comes into their purpose, and starts to find their own identity that there would be joy all over them.  Come on!  Look at Patch Adams, that man had been purposed in life to be a doctor.  The joy the fulfillment.  His job wasn\'t work, it was pure joy.  Something that he was passionate about!  Something that he was made for.  Do you see an abundant amount of joy in the world today?  NO, absolutely NOT!!!  I see a world that is frustrated and fed up with all of the crap that goes on in the world.  We live in a world without purpose.  Anything that has been created has a purpose.  A salt shaker\'s purpose is to hold and distribute salt.  Paper towels are created to wipe up messes.  Cars are created to transport things.  We have to start to investigate why and what we were created for, in that we will find purpose.  I will pose this however, we will never understand what we were created for, until we have a revelation of our Creator!   '),(3959,2607577,'21,june,2004','2004-06-21 00:00:00','    Our church is really starting to get into cellularizing its people.  We believe that as you get larger, you must get smaller.  I mean, every organ in our body is made up of tissue, which is made of cells.  So everything in our entire body is cellular.  The smallest living organism, the human cell.  Singularly you have life, Multiplied you have a body, made of a variety of different tissues and organs (personailties and cultures).  There is something to say about being big, and staying small.  You see, when you grow you lose intimacy within your body.  The cells in the toe never get to hang out with the cells in the hand.  But the thing that keeps them engaged, is that they are a part of a cellular group that is very tied together in the toe or the hand.  Because of their involvement in the specific area of influence they are intimately engaged in a very important area that has function and purpose.   My beautiful wife (Jessica) and I are engaged in a cellular group every sunday evening called \"Marriage Enrichment\".  It is led and hosted by Mel & Melodi Gideon.  Last night I was sitting and pondering this organism, looking at when it started well over a year ago.  It started with 8-10 families every sunday evening, and since then has almost trippled.  When I look at this organism, I see people who are committed to intimately engage in a function and a purpose.  It\'s not just because there is an incredible spread of food there (but it doesn\'t hurt), It\'s not the way we all squeeze into a 12X18 room, It\'s not even all the joking and cutting up that we do (Although we to get pretty crazy at times), but it\'s the since of belonging, the sense of community that brings us together.  Just being a part of something that has meaning to our lives, the sense of being a part of a family that genuinely cares and desires to experience life together.  It is the lost key of society, the smallest function of a city \"Community\".  I set back in my chair last evening and I said, you know what?  This sorta thing doesn\'t happen very much anymore!  It\'s a very sad thing, losing a key that unlocks so many doors............................   '),(3960,2607577,'21,june,2004','2004-06-21 00:00:00','    Today Tom Preble and I are going on an adventure to see what various people will say about certain issues in our world today.  We are going to ask some very pointed questions to these people to hear there thoughts on trust, sin, absolute truth, and community.  We are making a documentary for our church to show those of us who don\'t see what is going on in the world (with the breakdown of community) what\'s really the mindset of people these days.  I hate that there has been such a breakdown in the morals and values the were so present in the premodern era.  I guess in a way it will be better this time,  I know in the premodern era religious morality was based heavily on the Law and works, so I beleive the movement in the postmodern movement of Christianity will be based on Grace and Lifestyle, where our actual value is stated through our lives, and not our words.   '),(3961,2607577,'21,june,2004','2004-06-21 00:00:00','    Last evening I went to a local church to share the vision of what we thought God was doing in our community.  I shared a video of all the research and resourcing we did before we even launched into the community.  When I shared the video with the people of this church, they were blowin\' away by the amount of drugs that were in our community.  You see, Henry Co. (New Castle) has one of the most severe drug problems in the nation per capita.  So when I shared the statistics I could see that it was the first they had heard of this.    This is a sure sign that community really isn\'t community!  The community that we build these days are so disconnected in community that it makes me sick.  There are truly walls within the City, that seperate and disconnect and desensitize the people of our communities from the horrible things that are happening in our community.  We, as the body of Christ, have to make a change in our lives.  We have to bring back the value and importance of true community.  Unfortunately, we see this same disconnect in the majority of the church bodies of our community.  There is a definite disconnect between the churches of our community, instead of looking at our differences as strengths, we look at them as weaknesses.  And even within our churches are walls of \"Clicks\" that segregate us within our own family.  We have to set the example of community within the Church, if not, communities of Mormons, Jehovahs Witness\', Wicca, and other cults will draw our brothers and sisters into an inferior and non Christian community.  It\'s time Church for us to have our actual value, being our stated value \"The local Church is the Hope of the world\".  Why would anyone want to serve a Jesus when they look at us in our defunk moods, and constant depression that offers no hope?  We have to get a greater revelation of \"WHO\'S\" we are, so we can know \"WHO\" we are in Him.  Then and only then, will we be able to minister the \"Hope of Glory\" Jesus Christ.  No one cares how much you know, until they know how much you care.   '),(3962,2607577,'21,june,2004','2004-06-21 00:00:00','    I wanted to state this only one time and go from there.  If you need to know this is \"What we Do!\"  The Dream Center is a non-denominational ministry that is a gift to the city of New Castle.  We are a front-lines ministry that gets out into the street and assesses the needs of the people.  We serve the same house/neighborhoods every week in order to develop relationships with the people of our city.   Currently, we are in an area called Hospital Heights.  The Heights is one of the \"worst\" areas in our city.  The mixture of drugs, poverty, and just an overall spirit of depression/oppression, the Heights was the prime candidate for us to launch in.  We will be celebrating our 1 year anniversary there the 1st week of Aug 03.  We have came to this neighborhood every Saturday for 1 year, except 1 weekend, we went to see the some of the 14 people that we have put in Teen Challenge (Christian Rehab Facility).  God is radically transforming HH.  We have seen over 40 people come to know the Lord in less than 1 year.  We have a house church that was started in a garage in Jan 03, and we have 15-30 people that come weekly.  We are so blessed to be used by God in this capacity.  We are getting ready to launch our METRO KIDZ (Children\'s Evangelism) at the 1st of Aug 03.  We were just recently given a 20\' box trailer, and we are cutting the side out and putting it on hinges to make it into a stage, in order to be able to roll up in any neighborhood and do children\'s/youth services out of it.  We also just linked up with Westminster Center (Jeff Kundert), and we will be bringing in a food truck with 6000lbs of food every week to give out for FREE to the people of our city.  All of the Dream Center Ministries are non confrontational in our approach.  What we do is, go to the doors and say, \"Hey, I am Robert from the local church, is there any way that we can serve you today?\", and out of that we don\'t plug our church, or preach Jesus verbally.  What we do is BE Jesus through our service to them.  It is the easiest thing in the world to do....  This is the Evangelism for our times,  relationship evangelism, making our stated value our actual value, not preaching, but being through a lifestyle and a passionate love for Jesus, we are able to be a fragrance of love to a hurting city.  People don\'t care about your Jesus, or what you know about Jesus, until they know how much you care about them.  The one value of the Post Modern culture that I love, is the fact that they want to experience something that is real, not a crappy religion, but experience the love of a Savior.  This won\'t happen unless we experience this first!!!   '),(3963,2607577,'21,june,2004','2004-06-21 00:00:00','    OK, this thing already sucks.  I just went to look at the templates and it erased everything I already wrote in it!!!  Anyway, this is my first Blog, and I was turned onto it by Bill Roberts (pastorbill.blogspot.com).  As I read his blogs I noticed the importance of blogging.  It seems to be a great way to journal your thoughts, as well as, communicate your thoughts and feelings to others that care.  My name is Robert White.  I am a missionary to New Castle, Indiana.  I have a lovely wife name Jessica, a 19 months old son (Judah Zion), and a baby girl in the oven due in late Oct (Zoe Victoria).  I am blessed beyond messure.  I have the great honor of serving the people of New Castle and our region.  It is so exciting to see what God is doing in our city and region, so much so that I look forward to every day and even my nights when God speaks to me in dreams and visions for what He is going to do in this region.  New Castle is a city in east central Indiana.  It is a small city, but has one of the biggest drug problems per capita in the nation.  The drug of destruction is primarily Oxy\'s, with a little bit of Crystal Meth, and cocaine.  Our desire is to see the banner of Jesus Christ raised in our city, not by preaching, but by living a lifestyle of the Love of Jesus to a hurting community.  The ministry that I work through is called the N.C. Dream Center, which is an affiliate of the L.A. Dream Center (Pastor Matthew Barnett), and our covering comes from New Covenant Worship Center in New Castle, Indiana.  I can\'t wait for you to hear all of the cool things that God is going to share with you through me on this Blog.  Get ready, I believe that you will be blessed in the upcoming blogs!!!!!  If anyone ever needs to get in touch with me, here is how you can:  Email:  NCDreamCenter@ureach.com Phone: (765) 599-HELP Add:  815 W. Western Road, New Castle, Indiana  47362   '),(3964,2607577,'21,june,2004','2004-06-21 00:00:00','    urlLink    urlLink audblog audio post   '),(3965,2607577,'21,june,2004','2004-06-21 00:00:00','    One of the hardest things that I have had to get over is my perceived value of who I am.  I have always desired to be someone who made a difference,  was respected, and most of all, someone who has authority.  Through my 32 years, I have had many positions and titles that I thought defined who I was.  As I submit my life to Jesus, and start to look at His ways, I am gripped by a reality that my position or title doesn\'t determine who I really am.  Think about it, when you ask someone to tell you about themselves they immediately start to describe what they do, because that is how society defines who we are.  It is entirely based upon what we do.  So we go around with our important little titles and our twinkie badge and shout it from the roof tops (especially if it sounds important).  Believe me, I have been guilty!  The funny thing is, that in the Kingdom, EVERYTHING is the exact opposite of the world.  The world says, eye for an eye.  The Bible says, bless those that curse you, pray for those who despitefully use you!  The world says to do what feels good to the flesh.  The Bible says to not give into the temptations of the flesh.  The exact opposite.  And for identity, the world would say that what we do defines who we are.  Guess what?  Just the opposite in the Kingdom.  Our identity in the Kingdom is defined by who we are in Christ, not by the status of the world.  It\'s like this, Jesus\' title isn\'t Lord, It\'s who He is.  It\'s not a position but a function.  Paul is identified as an Apostle of Jesus Christ, not the \"Apostle Paul\".  That is a title that we give him.  Pauls title isn\'t Apostle, that is who he is, or how he functioned.    Let\'s go a little futher.  You say, ok Rob, I am getting this a little bit, but isn\'t it your title or position that gives you authority?  I say only in the world does a title or position give you any authority.  In the Kingdom, it is a function of who you are.  Authority in the Kingdom is not something that is titled, but something that is given to you by another person.  When you get pulled over by a cop for speeding, you didn\'t call him earlier that week and say, I am going to be driving this route later, would you please keep me accountable for my speeding.  No, that is a position that he has been given, and it is his discretion who he pulls over and who he lets go.  He controls his authority over you.  But in the Kingdom, authority is something that is given to a person.  My authority in anyones life, is only as much as someone give me.  It has absolutely nothing to do with my position, but who I am and how I am perceived by any individual I have relationship with.  If I have no relationship with someone, the will be reluctant to give me authority to help steward their life.  Authority is not something the is taken, but it is something that is given.  Think about it!  The only authority that I allow in my life is that which I give to someone.  That is why my identity is only based on my revelation of Jesus Christ, and what He says I am.  Many people never get their because they look at authority through the eyes of the law and not through the eyes of grace!  I am so thankful that I understand who I am, not based on my performance or my works, but entirely on who He (Jesus) says I am.  WOW, what a load of my back, no more earning my way to the top, do ya get it now?  I am sure you do.  Here something:  There is nothing you can do to make Jesus Love you any more, and the good thing is, there is nothing you can do to make Him Love you any less!  Stop perfoming and start being who He made you to be!   '),(3966,2607577,'21,june,2004','2004-06-21 00:00:00','    I wanted to personally thank the 1st Church of the Nazarene for bringing Mr. Pickle into town and sharing his testimonies.  I am personally touched and encouraged by the Grace of God that is on his life.  I was gripped by the stories he shared, and the truth that was revealed through them.  He is definitely a blessed man with a powerful message of how God can change a manâ€™s heart.  It is evident that the Lord has kept him in order to use his testimony to change lives, and I know that there are lives being changed.    While listening to him speak, one of his statements really challenged me.  Mr. Pickle stated, â€œif we wanted to do something about the drug problem in our area, we needed to stop being scared of the drug dealers/pushers and snitch them out.â€  When he stated this, I immediately had a check in my Spirit.  I have never felt the answer is for us to snitch on them.  So I prayed, searched my heart, and dove into the Word of God.  As I was praying the Lord revealed to me 2 Chronicles 7:14,   â€œif my people, who are called by my name, will humble themselves and pray and seek my face and turn from their wicked ways, then will I hear from heaven and will forgive their sin and will heal their land.â€  Our land is very sick right now.  It is sick with bondage and oppression like drug/alcohol abuse, poverty, hunger, broken families, and division within the Body of Christ.  What we need is to come together as â€œ1â€ Body, seek the Lords face, repent for our wicked hearts, and then He will heal our land.  Only Jesus can bring change to a manâ€™s heart, just as he did to mine.  It is only through a revelation of Jesus Christ that can change a person heart, and that is what it will take.  A heart change in you and me!  Please donâ€™t twist my words by thinking that I am disrespecting Mr. Pickle, because I have the utmost respect for Jesus in him.  However, we must be very careful to take authority outside what has been given to us by God.  We must be wise as a serpent, and gentle as a dove, in how we approach this situation, because lives are at stake.  Itâ€™s not only praying for the down and outers, but I believe we need to pray for our City Officials and for the Law Enforcement Agencies just as much as we pray for the oppression and bondage in our city.  We must pray for all areas good and bad, and touch the heart of God by uniting in Love just as He desires His Body to be.  Then we will be used to radically change this community as Jesus heals our land!   '),(3967,2607577,'21,june,2004','2004-06-21 00:00:00','    Wow, where do I begin this blog.  I have been ravished by my Gods Love for me.  It is truly changing my life.  I am realizing that as I give God more, the more  He takes, and this changes my heart to His.  It is a painful yet beautiful process.  I am so refreshed with His faithfulness in my life.  He is such a faithful God, even in my weakness, He loves me.  He cherishes everything about me.  Man that is good to know!  We don\'t serve a God who beats us down, but a God who is for us.  He is always looking out for our best interest.  Unfortunately, our twisted selffish minds don\'t allow us to see past our own desires.  It is very funny when I look back on the ways that I used to be and how immature I was even in my perception of who He is.  God thank you for being so patient with me.  Thank you so much for Your Mercy that is new every morning.  You truly are the God of Grace!  Ya\'ll listen to me, we have a God who adores us in every way!  That excites me.  There is nothing in the world that can change you more than a revelation of who you are in Jesus, and how much He truly Loves and Desires intimacy with you.  I am so Blessed, knowing that I am above all things, I am Loved by my King!   '),(3968,2607577,'21,june,2004','2004-06-21 00:00:00','    I am so stinkin\' tired.  We have been working overtime on the Block Party doing this weekend.  We are teaming up with Isaiah 61 Ministries (Donna Stephens), Westminster (Jeff Kundert), and a whole bunch of others which include my buddies from 1st Baptist Youth Group (YEAH).  I am so blessed to have brothers and sisters in the Lord be a part of this outreach.  God is truly transforming Hospital Heights (where we have been working for a year now).  When you sit back and watch, you can see the hand of God orchestrating something incredible.  He is pulling together His resources to change a community.  I am so greatful.    Anyway, I have PERFECT POWER today!  \"His power is made perfect in my weakness\".  I thank God for His power through the Holy Spirit.  So many Christians walk around with such defeat and weakness.  Many of them are ignorant to the Baptism of the Holy Spirit.  I am thankful that I have the Baptism of the Holy Spirit.  It is the POWER behind the fullness of God.  A free gift to all who believe and receive it.  It is like have a television plugged in into the source of power (wall socket), but never turning the power on inside the T.V.  The T.V. is hooked up to the power source (through our receiving Jesus as our Savior-being saved), but never receiving the power from the source through turning the T.V. on.  So many people never experience what life is like with the Power of the Holy Spirit.  If you don\'t have the Baptism of the Holy Spirit, seek it out through the Scriptures and allow God to change your life forever through the Power of the Holy Spirit Baptism.  You\'ll never be the same.  Trust me, I NEED IT!  We all do, in order to have victorious Christians lives that are constantly conquering sin, not constantly getting our butts kicked and staying bound by it!  Seek it, Know it, Own it!   '),(3969,2607577,'21,june,2004','2004-06-21 00:00:00','    I was awakened to a new revelation this weekend.  It was the revelation of understanding the price Christ paid for all, and within the revelation was an understanding that it is our responsibility as Christians to bring Him His inheritence.  I know, you\'re saying, \"what the crap are you talking about?\"  I am glad you asked!  Christ was the ultimate sacrifice for all sin.  Saving us from eternal damnation in hell.  He paid the price of sin for all, and that is death.  Because of His death, His heart is turned towards us as His vessels to deliver to Him His inheritence.  What is His inheritence you ask?  His inheritence is everyone you come in contact with, and those you don\'t.  His inheritence are all who have been, are, and will be created.  He created all for a purpose, so He could have intimate fellowship and love for them.  So many of His creation don\'t understand they are His.  And many times, because we call ourselves His (Christians), many people never experience that they are His, because many times we really aren\'t fully His!  So what those who don\'t know Him see, is a lukewarm bowl of a person who only thinks about their needs.  This is not His heart towards those who don\'t know Him.  He has a desperate pursuing Love that persistently cries out for their souls.  How would you feel as a father or mother, if your son or daughter had not ever known you, or if they knew you, and they rejected you.  Some of you know what I am talking about.  You experience that everyday!  Just imagine if you were Jesus, He created all.  And the ones you reject as you call yourself a Christian (being Christ Like) defiles the nature and the person that He is!  Forgive us Lord.  For looking through tained lens.  Lord our desire is to see the way you see.  We must have your eyes Father, give us your eyes to see Your heart.  We are such shallow and selffish people Lord.  Please have mercy on us.  Show us your heart towards your inheritence.  Give us eyes of desire for what Your desires are.  Allow us to experience Your heart so we can be vessels willing to bring You Your inheritence!   '),(3970,2607577,'21,june,2004','2004-06-21 00:00:00','    I really don\'t know what I am going to write in this blog.  I am just going to start typing and see what God does.  I went to a conference this weekend in Indy.  It changed my life forever!  The simplicity of the Gospel came alive to me when Shawn Bolz (whitedoveministries.org) shared.  You know we try to complicate things so much.  We try to get this deep understanding, like we figured something out in God.  We wear it on our shirts like a medal, and then we base our whole ministry on this revelation.  I have to tell you Shawn has done this as well, except this isn\'t some deep theological revelation of the Word.  It is as deep as it is shallow, as long as it is short, and as wide as it is narrow.  In a whole it summarizes everything up in just 4 letters (LOVE).  Yeah we hear a lot that God Love\'s us, but this was shared very differently.  It was a man who has been ravished by God\'s beauty, a man who was lovesick for his Bridegroom to come sweep him off of his feet.  Shawn truly had this fragrance (atmosphere) of Heaven on him.  His message was so simple, yet deeper than any revelation that God has brought my way.  I am drawn by the realness of this.  I am captured by the trueness of this.  I am left only to desire the same atmosphere in my life.  WOW, Jesus a beautiful Bridegroom, and I am His Bride.  Even in my weakness, He Loves me.  I don\'t think the church on a whole gets this, I know that my life has been changed forever.  My heart has been wrecked for less.  I have to have and know more about my Saviors Love.  Just talking about it makes it burn within me.  Living with the Atmosphere of Heaven in my life, what do you think God could do with that?  Miracles, signs, and wonders through anyone who will put themselves in that position, the position of being ravished by His Love.  It\'s coming guys, we have to catch this revelation!  I goto go!   '),(3971,2607577,'21,june,2004','2004-06-21 00:00:00','    I was at a conference last night (Friday) called \"Fire in the Heartland\".  It was a conference on Prophetic Worship & Healing.  During worship I had my little boy Judah (19 months) with me.  As the music began to intensify, I began to dance with Judah in my arms.  It was a very intimate time with my son.  While we danced, we would spin around and around, and I began to feel the joy that Judah was receiving as we danced.  At that moment, the Lord touched my heart and showed me a glimpse of His heart towards me.  I began to weep as we danced.  The tears intensified the more I saw joy and laughter come from my son Judah.  I was moved to be changed forever with this encounter with my God.  I realized that the father was dancing His son, that was dancing with his son.  You see, the first son was me!  As I danced with my son, Jesus was dancing with me.  And as I experienced the Joy from my son, Jesus was experiencing the Joy from me.  There\'s nothing like touching the Fathers heart, and having Him smile down on you from Heaven.  No one will ever be able to take that away from me!  Thank you Father for the dance!   '),(3972,2607577,'21,june,2004','2004-06-21 00:00:00','    As I search my heart, I see many wicked and lonely places.  Wicked in the sense that there are still so many wrong motives and selffish ambitions.  Lonely only because I refuse to allow God to go to the depths of my heart.  My heart is very shallow right now.  I have to get in line with Gods ways for my life.  I have to seek and understand this Love that pursues me so perfectly.  I don\'t want to be busy for busy sake.  God change my heart to Yours.  Teach me Your character and give me Your integrity.  Show me the intensity of how You Love Your children, so my heart will be purified of all the filth I feel in my flesh.  Only You know what the future holds for me.  Please Lord, do not increase me in any way until I am ready and wholly submitted to You.  I want to represent You, not Rob White.  I give You all I can, help me give You more!   (Please check out \"My Photos\", I want to introduce you to the Love that pursues me)   '),(3973,2607577,'21,june,2004','2004-06-21 00:00:00','    I am one of the busiest people I know.  The sad thing is, is that it is my fault.  I read this article on Jordon Cooper Blog the other day that convicted the crap out of me.  The article went a little like this.    Christians pastors/workers are some of the vainest people in the world.  They get there identity from being needed.  They pack there schedules in order to make themselves feel important.  Look at Doctors, no one wants to goto a doctor that has an empty waiting room.  You see, our importanct in the world is based upon how needed we are.  A doctor who is busy all of the time is perceived to be an expert or the best in his field.  Many times you see this in the Christian world, where pastors are keeping such busy schedules with their flock, but neglecting their home life.  Do we really need to be so busy in our lives.  Many times we are so busy that even God gets put on the schedule.  This is a major problem.  I am sure that this wasn\'t word for word, so don\'t get upset.  But there is truth to us finding our identity in our work, even as ministers.  There is a rush that you get when someone depends and needs you.  Unfortunately, we as humans, feed off of the feeling of being wanted.  I know that I do.  I have to change.  My identity must come from a revelation of who I am in Christ, not by what others perceive me to be.  I had to repent to my wife the other day, and explain to her that I was wrong for ever putting the needs of the people above the needs of my home.  I have to tell you, she cried, because she has seen this and has been praying that God would reveal this to me, and He did.  Imagine her relief.  We all need a check in this area.  We all desperately want to be needed and understood.  But let us focus on the one who created us and understand how much He needs us, and how deeply He wants us to understand Him and His heart.  If we would only get this revelation, He would take care of all of the needs that WE try to be God in!   '),(3974,2607577,'21,june,2004','2004-06-21 00:00:00','    Joe, the number that was on my caller I.D. was a calling card, call me back.  Thanks   '),(3975,2607577,'21,june,2004','2004-06-21 00:00:00','    I have had so many calls since the last article I wrote in the Newspaper.  Everyone is so excited, but no one knows what to do.  So I am going to share a story with you  Mother Teresa was on her way to a speaking engagement at a major convention center, where men and women from all over the world came to hear her wisdom of how they could help the dying and hungry people of their areas.  Unfortunately, she was running late to the meeting.  Thousands of people sat waiting for her to get there.  When she finally arrived, she came upon a homeless man who was hungry and weakly postured on the steps of the convention center.  Mother Teresa approached the man and turned to the other sisters and said, I am sorry we are going to have to take this man back to our hotel room and take care of him.  One of the other sisters stated, \"But mother, we are already late for your speaking engagement\".  Nevertheless, Mother Teresa helped to load him in the car, she took him back to her hotel, she bathed him, and the she personally fed him.  After she had taken care of the man, she then went to the convention center to speak to the people.  When she got there, everyone was very excited.  Then she spoke these words, \"All of you came to this city to learn how to take care of the dying and hungry, and at the same time, each of you walked by the man that was hungry on the steps.  You want to know how to cure hunger and need of your city, you must met the need that is before you!\" (Story told by Pastor Andy, Church on the Street Atlanta)  It isn\'t a complicated thing, all we have to do is open our eyes and look around.  There are needs everywhere.  So many of us see needs, and we do nothing about it.  We are so selffish with our time and resources it makes me sick.  You want to be like Jesus then do something, stop talking about it.  Give someone the hope that you have been given.  Maybe it\'s just through a smile, or giving someone grace who has offended you.  Maybe its seeing someone in need, and instead of giving them money for food, take them to applebees and then to Goody\'s and getting them some clothes.  Or maybe its just calling me 765-599-HELP, and we will get them the resources they need.  Do something, act like you are a Christian, stop talking about it!   '),(3976,2607577,'21,june,2004','2004-06-21 00:00:00','    urlLink    urlLink audblog audio post   '),(3977,2607577,'21,june,2004','2004-06-21 00:00:00','    I have decided that this is the new moto for the Dream Center (Shut up, and do it!).  Very direct, and harshly pointed.  I think that the people of the world need a wake up call that is very direct.  I am seeing more and more that people who go to church call themselves Christians, but in reality, they really don\'t strive to live that way.  They talk about all of the things they should be doing, but they don\'t do them.  I don\'t know, I am sure it is a mixture of self-fishness, pride, hurt, past experiences, and many more.  People are just to busy for God.  I promise you that there are only 2 currencies that account for anything in heaven.  1.  Your Money 2. Your Time.  There is only one thing that the two of these (Money/Time) have in common.  Do you know what it is?  I will tell you.  They both represent your LIFE.  Your time is the only thing in life that you can never get back.  Your time is so precious, even your employers give you money for your time based on how much you are worth to them.  So when God tells you that your life is not your own, he is speaking of your Money and your Time, cause this is the only thing that truly represents the importance of your life.  God desires for us to give our life to Him in every way.  But what we say is, \"OK God I will give you this area, but don\'t go there!\"  We are such winey little kids, we live our micromanaged lives and when something doesn\'t go our way, then we cry out to God.  Come on!  When are you all going to get it.  God isn\'t interested in your failed attempts to run your life, so He can make it better and you\'ll just go back to your own ways.  He wants ALL of you.  Your time, money, thoughts, love, effort, gifting, heart, children, YOUR LIFE!!!!  He knows what is best for you.  He knows the plans He has for you, plans to prosper you, to give you hope and a future.  He desperately wants to give you the desires of your heart, but He will not do it if He can\'t trust you in your stewardship of your life.  He wants us to be able to give all things to Him, so He will get the glory for our lives.  Trust me, I know what it is like to do EVERYTHING on your own, I lived that life for 30 years.  Only in the last 2 years have I allowed God to really have areas of my life that He could have, and this is still a process I am dealing with.  But guess what, I am trying, and God honors our efforts.  Anytime that Jesus had an encounter with someone, whether it be healing a lepor, or radically changing their lives in some way, you know what He would tell them?  He would say, \"Go and tell no one, but go and show yourself to them.\"  People don\'t care what we say, all they care about is what we do, and believe me they are watching.  So let me challenge and encourage you when I say, \"shut your mouth, and do it!\"   '),(3978,2607577,'21,june,2004','2004-06-21 00:00:00','    u-ni-ty (yn te) the state of being one, or united; oneness; the quality of being one in spirit, purpose; harmony; agreement; unification; a unified group or body.  The bible tells us, \"how good and pleasant it is when brothers dwell together in unity\".  Unfortunately, this is something that we do not walk in, YET!  I believe that God desires for us to make every attempt to come together.  Where I see the problem, is that everyone is trying to build THEIR kingdom instead of Gods Kingdom.  We have our priorities way out of wack.  We must allow God to change our hearts and motives, so we can experience the beauty of Unity.  I am so blessed to be a part of a group of spiritual leaders who are coming together every other week.  Putting down their agendas and schedules and meeting for prayer for Unity in the Kingdom and our city.  These brothers are dear to me, there lives encourage mine.  They drop their titles, positions, and doctrines to come together in Unity in prayer for our city.  You see, God doesn\'t really care about Bethany Tabernacle, 1st Baptist, South Memorial Church of God, Christian Missionary Alliance, New Covenant or any other church more than another.  It is not our doctrine that makes us any better than any other church or group.  The only thing that God cares about is that we love Him with all of heart, and that we love our neighbor like ourself.  I want to thank John, Dan, Jack, Mark, Ted, and Eric for being forerunners in the area of Unity.  You guys really move me, and let me know through your lives that God is ALIVE and doing something in this city!   '),(3979,2607577,'21,june,2004','2004-06-21 00:00:00','    urlLink    urlLink audblog audio post   '),(3980,2607577,'21,june,2004','2004-06-21 00:00:00','    My heart is stirred with a multitude of emotions.  I am tired of seeing the people of Henry County in bondage and oppression.  I am distressed by the fact that my brothers and sisters in this community stay bound to drugs and alcohol.  I am broken by the 2 recent teenage deaths due to an overdose of street drugs.  I am sad that there is so much division in the Body of Christ.  Most of all, I am sick of people saying that they are Christians, sporting the T-shirts, wearing the WWJD bracelets, and making Christianity their stated value, but not living it out through there lifestyle.  Itâ€™s like someone going to the Texas Roadhouse and ordering a steak, and the waitress bringing them a can of spam, wouldnâ€™t you be a little disheartened too?    You see, God has put the Church on this earth to be Jesus to our community.  The problem is, there is very little community within the Church.  We have over 150 churches in our community, and very few have any contact with each other.  Are we really building Gods Kingdom?  God desires for us to be Kingdom people who are members fitly joined together (The Body of Christ).  Coming together with our foundation firmly planted in Jesus Christ, and radically changing the face of this community TOGETHER.  We have to come together in this war of injustice that is keeping our brothers and sisters in bondage.  We have to mobilize our people to bring justice to the oppression in our communities, not the justice of the â€œlawâ€, but the justice of â€œLOVEâ€.   We must unite in order to bring the full expression of Jesus Christ to our people.  Jesus didnâ€™t come just for the Christians, but for those that live in the world. (John 3:16, For God so loved the WORLD that He gave His only begotten son) The local church is the hope of the world, letâ€™s start \"BEING\" that!   '),(3981,2607577,'21,june,2004','2004-06-21 00:00:00','    urlLink    urlLink audblog audio post   '),(3982,2607577,'21,june,2004','2004-06-21 00:00:00','    urlLink    urlLink audblog audio post   '),(3983,2607577,'21,june,2004','2004-06-21 00:00:00','    Hey everyone, sorry for such a long time away.  I have been in a process of soul searching.  Thank you all for your patience and prayers.  It\'s is so refreshing to allow the Holy Spirit to renew you from time to time.  Sometimes we get so busy that we forget about what\'s really important, ie:  Relationship with God, Wife, Children, then ministry.  God has used this time to reveal to me my heart, and how disconnected I had become, even from my family.  So know we are making plenty of time for family (keep me accountable).  Thank God I have such and awesome and loving wife.  She is incredible.  I am so blessed.   Anyway, God is doing so many cool things in our life, and we will be telling you all about it in the upcoming blogs.  Right now, God is having us to look foundationally at the structure of the ministry so we can withstand the storms of tomorrow.  God is so Good.  Kelly Boles just became our 1st graduate from Teen Challenge, she is doing great and she is very excited to get out and help others in need.  There will be an article in the Courier (Front Page) in the next couple of days.  We are very proud of her, and even more excited about what God is doing in and through her.  Also, for those of you who have loved ones or know someone who struggles with drugs/alcohol/or any other life controlling issue, a rally team from Teen Challenge will be at New Covenant (815 W. Western Road, New Castle) at 10am on the 5th of Oct, and also at Crossroads Church on S.R.3 at 6 pm that evening.  Please if you know anyone that struggles in these areas GET THEM THERE!  It will be life changing.  We love you all and please be encouraged in the Lord today, Jesus is Lord of all.  I want to give a shout out to my two little angels who brought us a truck load of food the other day.  Your love is so beautiful to us, thank you for your servant heart.  We appreciated and needed that more than you will ever know.    I will talk with you all very soon.  Tell everyone you know that I am back online, check it out, God is moving and He wants to use you to be His vessel!   '),(3984,2607577,'21,june,2004','2004-06-21 00:00:00','    I am so tired of seeing things the way I do.  I hate it!  All we do is look at things the way the world sees things.  Isn\'t that pithetic, all we do is look at the negative in people.  We very rarely look at the postive or even the potential of someone.  It is funny for me, I can look at a drug addict, alcoholic, or someone battling with depression and I can see greatness.  But when it comes to my wife, or those who are close to me, all I look at is the weakness.  How ironic is that.  Someone who has a visual, fully apparent weakness has greatness, but when someone close to me does something small, it defines who the are as a person.  YUCK!    I need Your eyes Lord, the eyes that look past the critical judgements of my heart.  I need Your eyes Jesus so I can Love like You Love.  I must have Your eyes God, so I can have unlimited patience in my life.  Most of all God, I need Your eyes so I can see what I have to change in my life!  That\'s the real reason I need Your eyes!   Psalms 139:23-24   Search me, O God, and know my heart;  test me and know my anxious thoughts.  24See if there is any offensive way in me,  and lead me in the way everlasting.   '),(3985,2607577,'21,june,2004','2004-06-21 00:00:00','    WOW, I just got some great revelation yesterday.  David Mallonee was speaking at our Church, and then at Aletheia later in the evening.  He teaches on Biblical Finances, not the prosperity message that we\'ve heard before on T.V., but true biblical principles for Gods prosperity in our lives.  It was very freeing to me.  One of the things that freed me up the most was when he shared the story of Paul the Apostle.  He shared the power of his ministry, and how he was probably the most powerful minister of all times, and at the same time he worked part-time.  That\'s right, Paul the Apostle was a part-time minister.  He was a tent maker by trade, and he did it all the days of his life.  He was a self funded minister.  This excited me for many reasons and here is why.  1.  I hate to ask anyone for money!  2.  If I am working outside my ministry, then I am not bound to someone elses feelings of how I have to run the ministry because they are helping to fund it! (Like, federal grants or even people).  3.  I can still make my own hours and make time for ministry.  4.  I would have the financial freedom to get my debt paid off, and not depend on the big 10,000 dollar check.  5.  I can be an abundant giver  6.  All the money that comes into me for my support can be put into the Dream Center account for the Dream Center.  7.  I will be able to take my wife out to dinner when ever I want to!!!!!  These are just a few of the reasons.  You see, the point is not to be a millionaire, just to increase the gospel in whatever way we can, and while doing that we increase our lives too.  I know that my thoughts are very skattered but it is 7 am, give me a little grace.  I am blessed to know that my giftings outside of the kingdom, can still be used to increase the kingdom.  I\'ll write more on it later!   '),(3986,2607577,'21,june,2004','2004-06-21 00:00:00','    urlLink    urlLink audblog audio post   '),(3987,2607577,'21,june,2004','2004-06-21 00:00:00','    Sorry everyone, I have taken a little time away to re-focus my life.  I have been so busy lately that sometimes I don\'t know whether I am coming or going.  But still, my God is faithful.  It\'s funny to me, even when I get wrapped up in me, and the things I do, He just looks down with a deep longing just to communicate with me.  WOW, what a loving and sweet God I serve.  He Loves me soooo much.  I am very greatful.    My summer has been a rollercoaster, mostly of what I have created, but there have been some ordained activities of God (thank God I don\'t miss it all the time).  As I look back, I look in the mirror and see a person that has so much passion and confidence in His God, that sometimes he overlooks the more important things in life.  Of course that person is me!  When I go deeper into the Heart of God, I see that we are a people that get stuck on the 2nd Commandment, and neglect the First one.  2nd Commandment, \"loving our neighbor as ourself\" trying to meet the needs those around us, and the 1st being, \"Love the Lord your God with all your heart and with all your soul and with all of your strength.  Even the best Christians get this turned around.  We are such a society that needs affirmation in the things that we do, we actually feel good about ourselves when we do good for each other.  For those of us who are servant junky\'s, we can easily get these two commandments flipflopped.  Trust me, I have to fight not to stay in that place.  I love to serve, I feel closer to God when I am serving and helping someone more than any other thing I have experienced.  Is that right?  I don\'t think that it is a matter of being right or wrong, but what is important is the motive of my heart.  The real question is, \"what is the motive of my heart when I serve?\"  Am I trying to fill my empty heart with doing good works, or are my good works and overflow of my 1st commandment devotion to my King?  Com\'mon you know exactly what I am talking about.  We all have wicked and selffish hearts, even those of us who have the purest motives.  Thank God that His mercy is new everyday morning!  Lately, I have been going through a heart surgery (not physcial, but spiritually), and as God has been exposing me to my heart, I have been able to acknowledge that there are areas in my life that need special attention.  I am so greatful for the Processes of God.  Lord thank you so much for never giving up on me, thank you so much for your unlimited patience in my life.  You are such a gracious Lover of my soul.  I want to be a 1st commandment Lover!  I want to know God\'s Love in such a greater measure, so I can, in some small way represent the character of Jesus Christ.  God ravish me with your Love.  Please Lord, bring a greater revelation of your Love to me.  I need Your Love in my life, it\'s only through Your Love that I can live my life to Glorify You!   '),(3988,2607577,'21,june,2004','2004-06-21 00:00:00','    Lord I thank you for the incredible feeling of wellness I have today.  I feel your grace saturated in my entire being.  Lord, I want you to be famous in my heart, and I want to make you famous on the earth.  Only You Oh Lord, can transform a mans heart, Lord I ask you to transform my heart.  I need your Holy Spirit to empower me to conquer the areas in my life that are unholy.  I love you Jesus, You are my King.  Thank you for your never ending love in my life.  You are beautiful to me.  Lord strengthen my brothers and myself as we journey to new depths in you.  We need your anointing in our lives.  I feel such a beautiful peace in my Spirit.  Such a feeling of victory in my Spirit.  I thank you Lord for all that you are doing in my life and the lives of so many in our community.  Lord let your Kingdom be magnified, and for you to be glorified.  You are Holy Lord.  I love the way You Love me.  It is so sweet to me.  Deepen Your Love in my life.  I love You Lord.  I need a greater revelation of who You are and whoâ€™s I am.  Make me a 1st Commandment lover of You!  I donâ€™t want to just get involved in activities that are just activities.  Lord, send Your Heavens down, Open the windows of Heaven, and allow Your people to be Your vessels of honor.  Let us be the ones who Glorify Your Name on the earth.  Let Your Kingdom come Lord.  Come Oh Lord, Come sweet Jesus.  We need you.  My heart is so excited to see what You are doing through Your people.  I am so anxious in my Spirit to see Your Glory fall.  I feel my spirit stirring in such an incredible way.  To You be the Glory and Honor forever, and ever!  You are everything, and all things rest in Your hands.  Come Lord, Come Jesus, Come and inhabit Your Bride.  Do whatever it takes to awaken Your Love in our hearts.  I love you Jesus.  You are my King.  Thank you for never giving up on me.  I am Your faithful servant for eternity.  Change my heart oh Lord.  I need to love like You Love.  I need to think the way You think.  Put in me a clean, new, heavenly heart oh Lord.  I love you Jesus with such a limited love, increase Your Love in me so I can love you more.   '),(3989,2607577,'21,june,2004','2004-06-21 00:00:00','    God has really been stirring me in the areas of intimacy and fellowship.  In many ways they are individual desires, but also they lived out by sharing these gifts with others.  I have been so enlightened with the depths that God is desiring to take His Bride.  I just got finished listening to the \"The 7 Perils of the Bridegroom Fast\", by Mike Bickle.  What an inspiring, eye opening revelation about our hearts towards living a fasted lifestyle.  I am so encouraged by what the Holy Spirit is doing throughout the earth.  I am starting to see a people who are desiring to be a genuine representation of what Jesus has always desired.  Of course, that goes along side of my very limited and immature understanding of Who He is.  Thank you God for Your patient heart, and long suffering with me.  Think about it, God created us to be in intimate fellowship with Him, and then to share that intimate fellowship with others (1st/2nd Commandment).  I met this guy on PBR\'s blog, and his name is Thom.  I don\'t know this guy at all, but I can tell you that I love him.  Thom if you\'re reading, please know that my heart is genuine towards you.  I really enjoyed the conversations that we had.  What I see is that our current revelation of what is going on is defined by many factors.  Some of those factors are:  how we grew up, are social environment, school/work experiences, and even our past church experience (both positive and negative).  Many times our differences really aren\'t a differing belief, but most of the time our differences are merely communcitaionally.  The way we communicate our hearts is so important to God, and unfortunately he uses peoples personalities to bring those revelations through.  We must desire to be more intimate with each other, we must strive to live in fellowship with each other.  I need to learn to communicate better with other, not so much for my sake, but for theirs as well.  I shouldn\'t expect everyone to think the way I think, or understand the way I put things all the time.  There are some big differences even in locality, and ethnicity.  Wow, if we could only commune together, so then we would know each others hearts.  Thanks Thom, for keeping me accountable to bringing understanding to my words, because those who don\'t know my heart, might misrepresent it!   '),(3990,2607577,'21,june,2004','2004-06-21 00:00:00','    As you can see, I have taken some time away from my blog.  I have been going through some deep introspection, and what I have come to understand is that I am incredibly unsatisfied with my personal walk, and with the face of Church as we know it.  As we work out on the streets, I realize that it is only through the deepening my commitent and understanding of who Jesus Christ is in my life.  I find myself looking at the immaturity of my heart and even in my love for the Lord.  We are on the 38th day of our 40 day liquid fast, and God has revealed so many things to me, mostly God has shown me the wickedness and the immaturity of my own heart.  It has been a very good revelation, don\'t get me wrong.  I know if I am going to boast, I need to boast in my weakness.  I have also gotten a new revelation of God\'s Grace in my life.  I love His Grace, it is only His Grace that allows me to do anything.  I am very excited about some new revelation God has given me from the standpoint of the fluidness of His Church.  I will reveal more later as God allows me to share it more.  I believe that He is preparing a wineskin that will greatly increase the effectiveness of His NT Church.  It is really exciting.  So stay tuned.  Anyway, I don\'t ever want to be satisfied with my walk, I desire to long after Him, and to pursue and prepare the way for my Bridegroom King!  I\'ll write more later!   '),(3991,2607577,'21,june,2004','2004-06-21 00:00:00','    In today\'s culture all you hear are buzzwords.  It seems that in our \"postmodern culture\", definitions are no longer relevant, but in the same sense there are new definitions for definitions.  New words, and thoughts that don\'t \"define\", but they do.  It is funny to me to hear postmodern philosophists dialog and debate, about not debating.  It is truly the circle dance of the mind.    In the language game, philosophists use a process called deconstruction to break down a thought (Metaphor) or a word to it\'s basic possible understanding.  After that process, they feel that they can fully understand the thought or word.   What a crazy bunch of crap.  The word game only brings a greater level of confusion, and very little clarity.   On the otherhand, in the Scriptures God uses a series of layering that helps to build a beautifully elaborate picture of who Jesus is.  That is how we learn, through the Scriptures.   We use illustrious metaphors like:  the Cross, Vine/Branches, Living Water, Salt/Light, the Blood of the Lamb, and many more.  All of these are used to build layer upon layer the character and beauty of who Jesus Christ is.    It is a process of building poetic imagination and the true power of the Gospels.  The indentity of our Saviors Love, and His desire for us.  Layer upon layer of Faith, Hope, and Love.  The Love of our Father projected through the Life of the Son, and imparted to us through the Holy Spirit.    The miraculous creativity of our Father.  Always building, never tearing down.  Never bringing confusion, but bringing clarity, upon clarity.  The funny thing is, is that the Word of God is very simple, we are the ones who deconstruct it to try to figure it out!   '),(3992,2607577,'21,june,2004','2004-06-21 00:00:00','    I was listening to a teaching by Mark Driscoll (Soularize 2003), and if any of you have heard Mark, you know that he is not one who candy coats things.  In his teaching, he made a profound analogy of how many of today\'s Christians view our God.    He said, \"We treat God like He\'s a some sort of sky fairy.  He is like a pinata, and we all send our prayers up and hope we hit it, and then we wait and hope a lot of goodies fall out.\"  WOW, that was a profound statement to me.  Think about it, how many times do we look at God as this big sugar daddy in the sky.  Sure, many of us don\'t intentionally look or even act that way.  But lets think about our true devotion to Christ.  We sit back with our mediocre lives, and our busy schedules and believe that we are really spiritual.  Where is the generation that will give it all?  Where are those that God calls, and immediately they respond by giving all.  I mean, when I go out and I listen to people talk (even my Christian friends), God is rarely the center of conversations.  What I hear most of have to do with worldly issues, activities, and upcoming programs.  Lord, increase Your revelation in my heart.  Help me to boast on the fact that I know You!  Give me wisdom Lord.  Lord, you are magnificiant.  I have made You too small in my life, forgive me!  Lord, be magnified, and let me shout it from the rooftops!  Help me to view You for who You are, High and Lifted Up!  You are the Glorious Lord, the Famous One.  Be made Famous in my life, so I can make You Famous on the earth!   '),(3993,2607577,'21,june,2004','2004-06-21 00:00:00','    Man there is nothing like the family of God.  I mean when you really encounter true geniune relationships with those of like precious faith.  We had our Church christmas party tonight, and we had such an incredible time.  I really love my family of God.  I am so grateful that they love me so unconditionally.  It is so nice to have intimacy with so many people.  I wish everyone understood this feeling.  I believe the world would be a much better place.  I want to say thank you to all of those who a put me and my family in a special place in their hearts, and given us the opportunity to grow in God Love together.  Thank you all for accepting me into the core of your hearts.  I love you all, and you have given me a great christmas gift, the gift of life and love.   '),(3994,2607577,'21,june,2004','2004-06-21 00:00:00','    Value is a big buzz word that is constantly floating around in my head lately.  As I start to see the end goal of where God would desire me to be (in part of course), I realize that my values must change.  We have 9 interns at the Dream Center, and it is so incredible to me, how we can think we are on the same track, but in reality we aren\'t.  I am sure there are many reasons for this, but mostly I believe it is because of \"value\".  Have you ever had someone ask you to do something and you said, \"you didn\'t have time for it?\"  What we are really saying is, is that we don\'t really value what you asked me to do, therefore I am not going to make time for it.  Or even our spouses, we tend to not give them the grace that we would other people who aren\'t so close to us \"VALUE\".  We see them in a different way that makes us (or me) not value our wifes because we continue to look at them through tainted lenses, based on knowing their weaknesses.    We talk about this revolution, and how everyone is going to be \"sold out\", but the real result is when someone actually takes the stated value in their mouth, and applies it to their lives, and it becomes the actual value of their heart.    What can we do to make our stated value, become our actual value?   '),(3995,2607577,'21,june,2004','2004-06-21 00:00:00','    urlLink    urlLink audio post  powered by  urlLink audblog   '),(3996,2607577,'21,june,2004','2004-06-21 00:00:00','    I was so blessed to hear a life changing message from a brother named Paul Vieira.  In his message, he was speaking about the emerging church and what it could look like.  One of the analogies he used was that the emerging church will be amphibious.  (forgive me Paul, I am going to do the best I can) For example,  When we were in the world (without Christ) we were like tadpoles, but after we had a metamorphosizing experience through excepting Christ as our Lord, we developed lungs (the ability to breathe new life), and then have the ability to live in a completely new and exciting environment.  He said, \"Much like a frog that can breathe and flourish both in water and on land, the early disciples lived naturally in two environments\".  They lived with God/Jesus on the earth, but also lived perfectly naturally in the world.    These days the church treats the world (where they came from) as if they had leprosy.  Unclean, Unclean!  You see, it is perfectly natural for frogs to live both on the land and in the water where they came from.  Matter of fact, when a frog gets ready to reproduce, the frog goes to the bottom of the pond and burries itself into the mucky mudd and plants its seed.  These frogs don\'t reproduce above the water, but in the water, where they came from.  WOW!  So, it is perfectly normal for us who have been born again to desire and plant our seeds into the lives that are in the world/pond.  IT\'S NATURAL, and God has made us that way.  (Thanks Paul for this beautiful analogy, it has transformed my life)  Two days after hearing this message, God started to stir me inside.  I couldn\'t get this analogy out of my heart, it drove me.  After two weeks, this is what God had given me. \"THE POND\"!  The Pond is the world, and it is perfectly natural for us to live in that environment, as well as, above water. (Keep tracking)  God spoke to me that He desires to have a pond that is filled with Lilly Pads (home fellowhsip churches/gatherings).  Within the Pond, we would start to develop \"Pads\" in it that we and others could rest (fellowship, Commune, Grow).  As more and more of the Frogs get this vision, more and more \"Pads\" start to emerge.  The cool thing about the \"Pads\", these pads are connected and get their source of life and nurishment from the bottom of the Pond, in the mucky mud (where we plant our seeds).   I believe that God is desiring to use this model as a way to help those of the water with their transformation of being frogs.  It is also a place where other frogs can hang out and experience life together, and then be released themselves and go back into the mucky mudd and reproduce after themselves.  Can you imagine,  thousands of \"Pads\" where those who have the life of Christ, bring those tadpoles into a non-threatening, non-churchy environment.  Where the goal is to commune, and experience life together.  \"Pads\" would pop-up all throughout \"The Pond\", and encouragement, edification, and life would be produced and reproduced as we and others went out to start their own \"Pads\".   This is a lifestyle that I believe God would desire for us to investigate, and implement!  We are starting our \"Pads\" for \"The Pond\" as we speak!   '),(3997,2607577,'21,june,2004','2004-06-21 00:00:00','    This is a repeat post from Bills Comments I wrote a few days ago with a little added:  I feel that the church in general has become a great industrial/Assembly Plant.  Our plans and strategies our carefully planned out (which isn\'t so bad), the executives get together and have meetings about increasing production, we have our set schedules, and functions where production happens, when our members/employees don\'t show up then there is something physically/spiritually wrong with them (and we judge their health), we desire to produce a product that the consumer can and will desire to buy into, and then after it is all over, we go out the door and seperate ourselves from work/church and become what the world calls couch potatoes for Christ.  If were blessed, we get to plant a new factory of coach potatoes in another city, so we can teach them how to be successful at being lukewarm robots, never producing life, but reproducing a quality function.      What is Church really?  Let\'s look at this from a simple mindset (cause that is all I have). God tells us that if we, confess with our mouth that Jesus is Lord and we believe in our heart that God raise Him from the dead, then we will be saved\". (of course that is a punctiliar salvation, see Bill I was listening). But if that is so, then the Sperma of God has been sown into our spirits, and we too become a new creation (Born Again). Knowing this, we must also see that Church is not a building, but an inhabitation of His Spirit. So we are the Church. If we are the Church, and God commands us to \"go into all the world and make disciples\", and we do that then through the Holy Spirit in us, we will help to make disciples, or inhabitations of the Holy Spirit, or even a more simple terms, a Church. So when we are imparting our lives into others through the love of Christ, we are also, being a vessel that God is using to plant a Church! We make Church good or bad. We are the representation of Christ on  the earth. This is a great responsiblilty. Thank God for his Grace and Mercy. When we strive to become 1st commandment people, then we will be able to make quality disciples. We must be passionate lovers of God, then we will be the Church that God desires to visit!  We must increase our devotion, and be a more accurate representation all of the time.   '),(3998,2607577,'21,june,2004','2004-06-21 00:00:00','    What is a biotic christian? A biotic christian is an organism (vessel) that is a living, mulitiplying, and eventually has the ability to reproduce the life of Christ.  WOW, I love this term.  I believe that God desires us to be biotic christians, ones that give and inhabit the life of Christ, not to hold onto for ourselves, but to be fruitful and multiply.  Ones that desire to mulitiple ourselves (make disciples) naturally.  Naturally developing disciples is a product of life, experiencing life together.    I think that the Church in general, of course I am not talking about you or yours, but for those who I am talking about it is you or yours!  We look at church as a set time, place, and a function.  It is like taking a look at a robot and a human.  In one hand we have a robot that is made up of may diverse pieces, each one is put into place for a specific function, and when the robot is completely assembled, it can look, and function in many ways like a human.  But it isn\'t.  Although a robot can function and look like a human, it cannot reproduce after itself like a human.    I think that this is one of the shortfalls of the church these days, we are so caught up in function and structure.  We are so busy creating new and user friendly robots (activities), that we miss the essence of being biotic.  We get so caught up in the organization of the church, that we become irrelevant to culture, because we are not a life giving organism.  Since we were made in Gods image, it would benefit us to learn more about the human body and how it functions, there we will find the diversity, complexity, and simplicity of the Godl.    We are yet one body, members fitly joined together!  We need to learn how to play well together, functioning together in life.  Experiencing life together spontaneously, that is what a biotic christian looks like to me!   '),(3999,2607577,'21,june,2004','2004-06-21 00:00:00','    My favorite season is the fall.  I love the changing colors and the way things have to die in order to become new again in the spring.  This happens every year.  Well, when I fall, it\'s the same thing but there\'s a slight difference.  I fall and it\'s a death that I die daily.    When I first heard the saying that we each have to die daily, to be perfectly honest, I hated it.  I rebelled against it.  I did not want to have to die daily.  I wanted to live, to love, and to develop in the life God gave me without having the pangs of death constantly on me.  What I couldn\'t understand is that the only way to develop is to revel in my \"fall season.\"  I mess up.  SO MANY TIMES!  I\'m only human, and not anywhere close to being fully cooked or finished in the kiln and firing process.    Monty said today that he felt like he had been taken off the Potter\'s wheel and is ready for what God has next for him.  That\'s where I feel I\'m at.  I\'m ready for the fire, I\'m ready for the kiln.  What I didn\'t expect was how the fire would burn me.    Working with clay is one of my favorite things in the world to do, so I know what happens when a pot is fired and there is some kind of imperfection or an air bubble in the clay.  The pot becomes something different.  It can take on a different shape or, in the worst case scenario, it can break into a thousand tiny fragments.  Well, I\'m in the kiln.  The fire has been turned on.  And there\'s been an imperfection revealed.  My bubble has burst.  Since I can\'t see myself through God\'s eyes, I don\'t know the extent of the damage.  Have I simply taken on a different form and am still usable?  Or are there pieces that need glued together?  Only time can tell.    This pot has become broken, I think.  And has fallen in between the spaces of the shelves.  What kind of glue is God going to use to put me back together?  Will it be a long process?  Will it be one of those 5,000-piece puzzles that takes even the puzzle-pro some time to figure out?  What I want to know, most, is how will He display His craftmanship?  Will He want to?  Yes, He loves to show off how creative He is.  But, will I want to be shown off?  How completely humbling to have to show my cracks where the glue has not quite made invisible.  And who will it help?  God, please use me.  Use as much glue as You can.  Just piece me back together so I can be filled with the Water.  Fill in the gaps so You aren\'t leaking out, but are POURED out onto Your people.  And keep using me to pour and to serve.  Just, please, don\'t leave me broken.  And don\'t let me hide my imperfections anymore.  Let me flaunt them like the battle scars they are!   '),(4000,2607577,'21,june,2004','2004-06-21 00:00:00','    I was just on Bill Roberts page and blogging with his people.  He was talking about pendulum swings, how we find something out to be wrong and, instead of trying to find out what is right, we automatically think we have to \"cross to the other side of the road.\"  For instance, when I was divorced, I assumed all men were bad.  This was only natural for a hurt human being to think, but it was completely wrong for a Christ-loving woman to think.  Afterall, Jesus is a man.  I was so busy hating men and thinking the worst of them, I almost missed an awesome opportunity to be ministered to.    When I found out I was lonely and no better off hating men than before since I wasn\'t healing, I went the opposite way.  I tried to find out what was wrong with me by having meaningless relationships.  I remember looking into each guy\'s eyes and trying to see what he saw in me.  Did he see ugliness?  Poverty?  Death?  Or did he see in me life?  Love?  Beauty?  It took me a while, but I realized I didn\'t even see anything in my own eyes when I looked in the mirror.  That was a revelation and I only received it because God smacked me in the head with it!!!!!  Girls, so often, are hurt by men and do one of these two things:  hate all men and are guarded for the rest of their lives, or think that it\'s their own fault they were hurt and search for answers in countless fleshly relationships.  Notice I used the word \"girls\" instead of women.  I was a girl when I was married and tried both paths after the marriage died.  More than anything, I KNOW that girls have to see men following after God with their whole hearts if they are going to have any faith in the male species.  So, now when I see a man totally in love with my God, then I\'m encouraged.  There is hope for girls and there is also hope for the Church in America.  Revival will happen with the men.  They are the missing component.   '),(4001,2607577,'21,june,2004','2004-06-21 00:00:00','    Lately, it\'s scary how much God has been directly talking to me.  And I do mean directly!  For instance, there was something very specific He told me to say to someone I didn\'t even know.  Thank goodness I did it.  What would have happened if I kept my mouth shut?        Coming from a church so divided in its thinking as well as its love, before I may have been hindered by my mind over-riding what God told me.  That\'s a scary thought.  But, and praise God for this, I am in a place now where worship abounds and isn\'t restricted.  The Word of God is taken as the \"real McCoy\" and not judged by the translator.        Freedom.  That\'s what Ms. Alliece said I would find when I came home.  I did.  And I am home.  *********Update********* I have a job now, praise God.  And it\'s amazing how He worked it!  I will be baby-sitting my nephews.  This allows me plenty of time to do whatever I can for the Dream Center and even the church, if they want me, and to counsel any and everyone who wants to talk!  Not only that, but I will be able to write.  REALLY WRITE!  Thankyou, Jesus, for taking my petition to our Father.  He heard me loud and clear!:)   '),(4002,2607577,'21,june,2004','2004-06-21 00:00:00','    When I think of all the things God has brought me through, and all the things that are yet to come, I become confounded!  As if this past year hadn\'t been enough of a change and an adventure,  now God is bringing to life new things in me.  Scary things!  He is putting dreams and urgings in me that were, heretofore, unheard of!  Example:  God has given me a heart for teen girls, specifically ones that are going through major crises.  What I don\'t understand is why He has broken my heart for them while I\'m still walking out my own healing.  I know, I know.  The only way He can use anyone is when they are broken.  I simply cannot put my mind around it.  I feel so insufficient for them and, as a result, back off from them.  My greatest fear is that I will lead them astray.  How many times have I heard that men should not ask to be teachers due to the responsibily and the greater judgement they will receive?  Christine Caine came to the Dream Center in June and I asked her this same question.  How involved can I be while trying to overcome the giant chasms in my own life?  She kicked my butt!!!!!!!  She told me I had no right not to be completely and totally involved with every person God led me to.  Now, how to overcome the fear?  And do not say that I should repeat that scripture...  \"God did not give a spirit of fear...\"  This is more than a recital for me!  This is hard and REAL!  Yes, scriptures help and that very one has helped me in past times.  But this is more!    God, give me the answers...  And the peace that will definitely pass all my understanding.  *************UPDATE**************  He did give me the answers.  He made me see it myself.  Funny thing is, He used someone else\'s blog!:)  He used the blog from Bill Roberts on Baggage.  God made me write out what I thought about people bringing baggage around with them wherever they go.  He doesn\'t want me to hurt.  I have to give that to Him, I have to hand over the baggage.  The scars are still there and, yes, I will still have to deal with the open wounds that are still there.  But, and this is an awesome thing, God has already been dealing with my pain.  Only, it\'s not mine any longer.    FREEDOM!!!!!!!!!!!!!!!!!!!!!!!!!!!!   '),(4003,2607577,'21,june,2004','2004-06-21 00:00:00','    Rob told me to get on here and start writing, that it would help us keep each other accountable.  (The interns and other Dream Center people)  I\'m guessing that I am supposed to write my inner-most thoughts and put them out there for any and all to see, read, and comment on.  Terrifying!  What should I write about?  Drama, suspense, romance, spiritual adventures...  There is a never-ending list of topics.  Hmmm...  I\'ll ask a question, then.    Where do you think I should go next?  To help in answering this, I\'ll give a little background.  I\'m 21, have about a year\'s worth of college credit, divorced, and jobless.  I just recently got back from the St. Louis Dream Center where I finished a one year\'s intership doing inner-city ministries.  I JUST broke off a relationship with the man I was sure I was supposed to marry and still have the same assurity.  I\'ve left my home church where I was not growing and have started going to a, certainly, charismatic church.  I don\'t know where I fit in.  Now, does anyone feel up to answering that question?  God knows that I have been pleading with Him to open a door somewhere.  All I get is,   \'Be still and know that I am God...  \'  Fun, huh?   '),(4004,2607577,'21,june,2004','2004-06-21 00:00:00','    I haven\'t written about this weekend on here so far.  I wrote on my own blog site, but I didn\'t want to reveal all just yet.  Guess I wanted to hang onto it for a bit, for myself to enjoy and revel in.  But now, I\'m ready to rejoice with you all, (whoever \"all\" is) and tell you what God did this weekend.  On Sunday, I almost didn\'t make it into church that morning.  We\'d been out late from the radio program, WJCF 88.1, theG.I.G., and then I\'d gotten a call from a friend in need of some help which kept me out about an hour and a half after that.  So, Sunday morning didn\'t produce the same bright-eyed and bushy-tailed Anna it usually did.  Nevertheless, I couldn\'t stay home.  45 minutes before I was supposed to be at the church, I hopped in the shower and hurriedly got ready.  Thank God I did!!!!!!!!!  The service was wonderful though for some reason, I wasn\'t as engaged as I normally am.  I generally take tons of notes and get a lot out of the service, knowledge-wise.  I couldnâ€™t figure out why I was so completely relaxed and at rest during the service.  If anything, I should have been anxious and uptight because of the kids from the Block sitting with me making an unholy amount of noise!!!  In fact, several people had moved during the service due to the noise they were making.  They didnâ€™t seem to penetrate the â€œtranceâ€ I was under.  I was at rest, thatâ€™s all I can explain it as.  When the speaker gave an altar call, before I knew it, I was up at the altar! (Apparently, I needed it more than my mind had thought I did!)  4 different people prayed with me!  Thatâ€™s odd.  I guess I looked pretty prayer-needy!:)  Anyway, when someone was praying with me, they said I needed to let God unlock my heart!  As he was praying that, my eyes were not seeing the back of my eyelids anymore.  I was seeing myself in a vision.  (Before, if I had visions, they werenâ€™t about me.  This one was!!!!!)    I saw my hand brush aside a curtain or a veil hanging over a doorway.  The curtain had been torn down the middle, from top to bottom.  Note: I pushed the curtain with my left hand.  When I walked in the room, it was this amazing room, lavishly furnished and beautifully done.  In the middle of the room stood a Man of such beautyâ€¦  He was dark skinned with dark hair and eyes.  And He was holding out His hand for mine.  I took it with my left hand: the hand that wears the wedding ring.  What I was seeing was the bridal chamber.  God had been waiting in my heart all along.  I just didnâ€™t let myself look that deeply into my heart anymore and hadnâ€™t wanted to find what was waiting for me there.  Stupid me, He has been waiting there all along.    This vision wasnâ€™t scary or anxious or even overwhelming.  Like I told Rob, it was perfect.  Thatâ€™s all I can say.  Perfect.  Afterwards, though, I became overwhelmed with what it was I really, truly saw!!!!!  The reality hit me like a brick!  I felt so unworthy, so uncleanâ€¦  But in the vision, He had taken me just as I was, as I am.  He didnâ€™t say anything.  He didnâ€™t have to.  My mind simply couldnâ€™t handle it.  Trebby came over and said something I knew God was trying to tell me.  Actually, Iâ€™d seen this girl over in a different row and thought God was telling me to give her a message.  I wrote it down and fully intended on giving it to her when the opportunity presented itself.  Unfortunately, sheâ€™d left by the time I came back from the altar call.  What Trebby said was what Iâ€™d written for the girl.  God was trying to tell me how much I am worth to Him, how much He adores me.  Hmmmâ€¦  Such wonderful stuff!  And then!!!!!  As if this wasnâ€™t enough, Monday night at Cleansing Stream, I had some major break through!  We are dealing with the consecration phase of the class.  We were asked to write down all the things we felt God wanted us to consecrate to Him and then burn that piece of paper in a symbolic fire on a make-shift altar.  Iâ€™d done something like this before in St. Louis on a retreat with the other interns.  I consecrated Paul, my ex-husband.  I figured he was permanently consecrated.  I didnâ€™t know you could pick back up something youâ€™d let go of before.  (Found that out the hard way!)  Anyway, Monday night, I consecrated him again along with Jason, the man I thought for sure I was going to marry.  I meant this consecration with all my heart and was ready to be released from all the pain and agony of stressing and worrying about both of these problems.  Along with several other issues I wonâ€™t go into, I burned that paper and was released.  I am released.    Funny thing?  Pastor Eric and the video teacher from class warned we would immediately be attacked in the coming week with what weâ€™d just consecrated.  The battle was to come next.  I just didnâ€™t expect it to come the minute I walked in the door of my home!!!!  My father, not even two minutes after Iâ€™d taken my coat off and sat down beside him, said that Jason had been at the YMCA and came over to exercise with him chatting up a storm!  No way, Satan.  Not again!  Iâ€™ve given that over and heâ€™s not mine.  Heâ€™s not mine to worry over, to think on, or even love.  Heâ€™s just not mine!    Well, what do you think?  Pretty amazing stuff, huh?  There is so much God has been doing lately.  And so much Heâ€™s changed.  Itâ€™s hard to even think what I was like just a few months ago, back in St. Louis.  Iâ€™m a different person.   '),(4005,2607577,'21,june,2004','2004-06-21 00:00:00','    I know what you mean, Trebby.  Before I went to St. Louis or even had thought about going out there, I\'d been asking God how to be a leader, a GOOD leader.  I was looking for all the training I could get and pestered my youth pastor I don\'t know how much!!!:)  God kept telling me, \"Be still and know!\"  I thought this meant He wanted me to wait for the teaching to come to me.  It took me going all the way to St. Louis where I thought these great teachings would change me into the right kind of leader to realize what God really meant.  Be still meant to shut my mouth from saying what it wanted and to shut my mind from doubting.  He had already made me a leader and had put in my heart and spirit the words me wanted to say.  I was already a leader, and a good one.  I just didn\'t walk it out in the Spirit.    I figured that the best leaders were learned men with round glasses at the tip of their noses, grey hair and all.  The St. Louis Dream Center showed me street-wise men and weirdo women who were my age and leading a church of over 2,000.  Leading is following.  We follow God and lead the sheep.  We have to just know that we won\'t make a mistake if we are walking in the Spirit and following God, His Word, and His path.    Be still and know, Trebby.  You are a leader.  You wouldn\'t have anyone following you if you weren\'t.  And that laundry list?  More areas for you to glory in.  God still has more work to do in you.  Be glad there is a list you can see and He has revealed it.  You could be walking around with a plank in your eye and hypocritically trying to remove the speck from eveyone else\'s.  Remember?  The only way I have been able to say any of this in TRUTH is because I\'m still working and walking that out.  We all do, Trebby, if we are perfectly honest.  And we are all trying to get the \"be still and know\" part down.  It\'s a life\'s pursuit and mission to get that!!!!!  ********UPDATE********  Be praying for me.  I had an interview this morning with Anchor Families.  I so want to be working in a place like that.  It will show the world later on that I\'m serious about starting a girl\'s home.  I also know a couple of girls there from my old youth group.  They were some of \"My Girlz\" and I would love to be in aposition to help them again!!!!!!   '),(4006,2607577,'21,june,2004','2004-06-21 00:00:00','    Glory to GOD!!  I want to talk about this self-centered approach that I battle. God has things he wants to do through me but I am so self-centered, I respond, \"I don\'t think I am trained. I don\'t think I am capable of doing  it. I have been of those self-centered people for my entire life. Then the moment I sense God is moving in my life, I break out my laundry list of why he has the wrong person or why the time is not right. Here lies the problem as usual, it is my perspective not God\'s. God knows that I can\'t do it! But he wants to do it through me. I wish to seek God\'s perspective every single time.   I can\'t take the initiative on my own, God will take the initiative every time when he comes to me, He will reveal himself and his activity. This is always an invitation for me to adjust my life to God. You will never be the same after encountering God. God -centered is the only way we will ever have the true revelation of God. I must learn to see things from his perspective. I must allow him to develop my character in me. I must allow him to reveal his thoughts to me. Only then will I have a proper perspective on life.  Can you relate with me here?  Treb   '),(4007,2607577,'21,june,2004','2004-06-21 00:00:00','    It\'s about time, Trebby!  And you are most welcome here!  Your \"adjustments\" have been seen by us all.  (It\'s that transparency thing again!)  We see an amazing man changing into an even more amazing man, amazing MAN OF GOD!  Praise God, you are going through all of this now.  And with us.  There is a perfect timing and placement in your situation, you know that.  Be assured your love for God shines through your skin!  You glow, man!  *******  I wanted to go back and re-read some of the blogs I\'ve put on here.  Well, honestly, I didn\'t WANT to, but I felt like I needed to.  (Not to mention that when people comment on something I\'ve written, nine times out of ten I can\'t even remember what I\'ve written that day!)  You know, it\'s amazing how much junk I\'ve put myself through in this short amount of time.  And it\'s even more amazing to see how far God has pulled me out of the \"muck and mire.\"  When I first wrote on here, I wasn\'t sure of anything.  All of my pre-conceived notions about what would happen when I came back to New Castle...  Let\'s just say they were trampled, crumbled, and pulverized!  (I say this laughingly, now!)  I was ready to save the world in a big way!  The training I\'d been through showed me how to be strong and steady for God when I\'d gone past all my physical abilities.  But it was in an environment already set up for success.  I was put into a working ministry with most of the kinks worked out.  And home?  A whole new ball game!  (Thank God!)  In St. Louis, I just had to worry about how things were going to get done in a certain amount of time.  There wasn\'t much time for personal evaluation.  True, there was a lot revealed while I was at my wit\'s end in the middle of a new fiasco which needed my immediate attention.  But, I didn\'t have time to think!  I was just reacting to situation after situation and learning from my responses what God needed to work on me next.  Here, back home, I\'ve had more than ample time to catch up on self-examination time!  UGH!!!  I already knew what I\'d been through.  I\'d faced it.  But, I didn\'t think anything new had to be dealt with.  Like any other human being, I\'d assumed the past was in the past and I only needed to deal with the future.  Well, if you haven\'t dealt with the past yet, then it will re-appear in the future.  I\'m living proof!  Ms. Alliece described it as a giant wheel I was running in.  I\'d made a mistake, hadn\'t dealt with it, and couldn\'t figure out why I kept having to deal with that mistake re-appearing.  She said I was on the same wheel and seeing the same rock stuck in the treads.  I had to either get the rock out, or jump ship and land on a different wheel.  My blog, \"I don\'t know, yet,\" was a primary example.  I know where I am supposed to go, I just let my brain become confused and jumbled with doubt.    So, what do I do now?  Jump wheels?  Or get the rock out?  Which would you choose?   '),(4008,2607577,'21,june,2004','2004-06-21 00:00:00','    Greetings,  I have been experiencing this thing that business people like to call \"Deep Change\".  I was introduced to God in 1994. That was about the extent of my experience until lately. I became very sick on October 11, 2003 diagnosed with bacterial meningitis. During this same time some very personal problems were revealed. (I am not very proud of) I was released from the hospital on Friday, October 17, 2003...Wow!Little, did I know the change that was forthcoming to this wicked heart.  I spent the next week in brokenness and for that, I will forever be indebted, unfortunately my brokenness came at the expense of my Wife and her heart. I spent one entire week crying 24/7 out of my shame and embarrassment. The awesome thing is my Jesus was there beside me, the whole time, his Love is Authentic, it does not change. The mercy and grace that God poured down on me was like a cup of hot chocolate on a cold night. It sometimes takes circumstances that are man made to reveal your total dependence on God and His will. I like to say \"it is the best of times and the worst of times\" I have discovered a new life and new purpose for this \"Wicked Heart\". It is to bring forth the Gospel of Jesus Christ!!. After trying to be all I could be. I have surrendered to Jesus and that means, doing all I can do with Jesus in my heart to convey that message to the other Trebby\'s of this world. I am not sure, what will happen in my personal life but one thing is sure, my Jesus. I know that some of you have wondered about my wife and wondered who is she? Why is she not with me? Well, my wife is an awesome woman that teaches 5th grade. We are currently separated because of my poor decisions. I hope you all understand me a little more now that everything is in the open. I have been Moving from a self centered life to a God centered life and this requires adjustments...Friends,job,family,etc. so please, keep us in your prayers, and a big shout out to my bro...The candy man! (haha)  I want to thank all of my new friends for your friendship and love.  Treb   '),(4009,2607577,'21,june,2004','2004-06-21 00:00:00','    I\'m finding that I like this \"Cleansing Stream\" class more than I\'d anticipated.  Nearly everyone I met out in St. Louis went on and on about how much Cleansing Stream changed their life.  In fact, while at one of the pastor\'s houses, I met a lady named Carole who actually used to lead Cleansing Stream in New Castle!  I was in St. Louis, in a private home, and met a lady from New Castle!  How strange is that!  I\'m smart AND stubborn.  This hasn\'t helped me like you\'d think it would.  I have all kinds of head knowledge from reading the Bible, but the heart knowledge?  I guess I\'m too stubborn and prideful to let it penetrate past the upper extremities and into the place it needs to be hidden the most.    I help people.    I can\'t help myself.    The prophet, Dennis, who came to New Covenant, hit this right on.  He said God was telling him all about how I needed to take care of myself and not everyone else in the world.  How do I do that when I don\'t want to?  And don\'t think I deserve my own attention?  Mindsets.  It\'s all about mindsets.  My former mindset, (and, yes, it is FORMER) is that I am not worth it.  I\'m not worth my time or anyone else\'s.  So, I need to be helping everyone else to have SOME kind of worth.  Tuesday night and then again last night, my worth was re-examined.  What kind of God would be so lavish with someone He didn\'t think was worth His time?  What kind of God would think me worth it?  And Who would think me worth wasting on?  Lavishing on?  Any blessing I received from God, I would see it as a necessary thing He needed to give me.  There was something He was going to use it for and that was as far as it went.  I wouldn\'t see a blessing as something He simply wanted to give me, to bless me with.  A revelation on this came when I hit a crisis point this month and called my best friend.  God told me He didn\'t give me Craig as someone just to be there when I needed, but He gave me a friend I could talk to about anything and everything.  It was and is to be a relationship of friendship, and easy friendship.  I don\'t have to call Craig only when I need to or if I know I need to minister to him.  I can call him and just talk for hours about anything and everything.  This month has been one of the hardest of my life.  There has been so much transition and change.  AND ON TOP OF THAT!!!!!!!!!!!  God has been changing old things in me.  So, I\'ve been talking to Craig a lot more than I did before.  Two o\'clock in the morning, during his classes...  Anytime!  And he is there to talk.  I haven\'t really had that kind of friendship before.  I\'ve been that person before, but in a counseling type of situation.  I haven\'t RECEIVED that kind of friendship.  God wants to bless us with so much more than we desire.  And He wants different things for us than we would ask for ourselves.  How long is it going to take me to get this entire revelation?  God loves me!!!!!!!!!!!!!!!  Plain and simple.    What else do I not get, Lord?  Start cluing me in!   '),(4010,2607577,'21,june,2004','2004-06-21 00:00:00','    Christy, welcome to life as a transparent piece of glass!  Getting all that out in the open is the hardest and best thing you can do.  Teens, people, are going to see it, anyway.  Better to put it into a way you want it to be understood.  And how awesome is it when you think that what you went through may actually help or prevent someone going through the same junk you went through?!?!?!  I think it\'s safe to say it takes a breaking point in us to get to transparency.  Tuesday night was certainly a breaking point for me.    I\'d never thought of God as the Prodigal before.  Yes, I knew He lavished love and blessing on us way past our understanding and worth.  But, when put together with the parable of the Prodigal Son...  Man!  Powerful!  And I hit my breaking point.    For a while my emotions have been raw.  You know how you can cry so much that the tears stop, but you\'re still crying?  It\'s kind of like that.  I was so worn out, tears wouldn\'t even come anymore.  And I hated feeling that way.  I wanted the comfort of tears to wash out the impurities.  I wanted to feel the cleansing sensation only a good cry can bring.    You know what?  I haven\'t really cried like that since...  I can\'t remember when.  It was back in St. Louis, I remember that.  Yes, I\'ve cried, but bitterly so over some really painful things.  I haven\'t cried those clean tears like Tuesday night.   I\'d been crying out of sheer emotional reaction to situations, not because, I don\'t know.  It\'s hard to describe.  Ah!  I\'ve got it!  I haven\'t cried or been overwhelmed by God\'s extreme love for me in way too long!  That\'s how I want to put it.  I would see people in church, especially at Harp & Bowl, and be so jealous of their open brokenness before God.  My own brokenness was different.  It wasn\'t because I chose to be broken before Him, but situations had broken me like a toy boat crashing against the rocks in a hurricane!  And I hated how selfish that made me!  I COULDN\'T MAKE myself see how much God loved me and wanted me to crawl up in His lap and cry myself to sleep.  I could sing the songs of how great His love for me is, but my heart refused to understand it.  I was so hurt by circumstances that it took my tenderness away.  Not now, thank You, God.  Guess where my favorite place to worship is?  House Church in Hospital Heights.  I\'m home, there.  That is where I can truly worship God.  Not in some nicely fabricated building with people who know God and how much they are worth to Him, but in a converted garage/church with God\'s \"outcast flock\" can I feel free to worship Him like He deserves to be worshipped.  If I could, I\'d take every single person from my old church, one at a time, and bring them out to the House Church to see how God\'s children really worship.  It may not have the most technologically advanced screen and PowerPoint system, or even restrooms separated for men and women but it has something my old church is missing: it\'s purely humble.   '),(4011,2607577,'21,june,2004','2004-06-21 00:00:00','    I have never posted anything before.  Mainly because I am not a writer and I am not all the excited to share my thoughts and feelings with others.  I have always been a person who stayed to herself and a lot like Anna I was depressed growing up!  I honestly cannot remember much of anything up to maybe 4 years ago.  I am 26 married with two kids.  Now you would think I would have a lot to remember.   I cannot even tell you what happened on my wedding day.  I was just so consumed with sadness I never thought I would be happy.   Then I found God. Yeah he has always been there but I never once turned to him for any kind of help until a little over a year ago.  I did not know he was there! I knew something was wrong, but really did not have anyone in my life who directed me where to go.  This past year a lot has changed inside me.  I have a huge passion for teens.  I don\'t want anyone to go through what I went through and have no body to turn to.  I am determine to let God use me in any capacity to do whatever he needs done so kids will always have someone in their lives they can turn to. I want them to feel the JOY that I now feel and the peace that I have.  The other day I was at work and I started to write in my journal.  Just now as I was reading different things on Rob\'s blog I felt God telling me to share my journal pages with you.  Now if my vote counted well lets just say I would definitely keep this to myself..  Its all about being obedient and stepping out even when you are uncomfortable.  Well here you go!!!!  How I feel I can\'t express with my mouth nor can I imagine with my eyes. It is a faith that nothing of this world can break.  A white light so pleasing and beautiful you can\'t directly look at it.  Its to wonderful to even begin to express.  You think its a love like you love your children, but its not.  Its a million times greater!!  When you love something with all your heart and you don\'t think its possible to love one more drop.  Then you realize your love has grown deeper.  You can feel a burning in your chest a pounding of your heart and you want to share the awesome experience but no one or maybe, just maybe one person might understand.  I just so desperately want to share and for someone to understand my joy, my love and my peace.  The best part to understand you are beloved!!  You are loved like this by your creator.  That God has this feeling about each one of his children.  Then I realized that I don\'t need anyone else to understand because I know my father understands my heart perfectly.  And in the end he is the only one that matters...  HE is the ONE TRUE LOVE!!!   '),(4012,2607577,'21,june,2004','2004-06-21 00:00:00','    MY shield is not going to cut it.  Yesterday, in youth group, Christy was talking about the armor of God.  We\'re supposed to put on the armor every morning for a week and see what kind of difference it makes in our day.  (Who\'d have thought God would use a 15 minute lesson to break through some of the thick skull I have?)  When she came to the shield of faith, something clicked there.  We were having fun pretending to put on actual armor when it hit me: I\'ve been using MY shield.  What kind of shield was I carrying?  One of self-reserve, boyantly crowd-pleasing, and critically judging...  AND DEPRESSION!  In St. Louis, Ms. Alliece, my spitirual mother, asked me to write down 5 things I didn\'t like about myself and 5 things I wanted to change.  One of the things I wanted to change was to be able to go from depression to steady joy.  The most defining thing in my life has been dealing with depression.  Even as a child, I battled with keeping a core of peace and joy without it being shaken and jolted by outside circumstances.  What I didn\'t realize, the shield of faith is supposed to keep those poisonous soul darts from penetrating that core.  I\'d made a shield out of circumstances I\'d been through.  Duh!  This automatically means I will see and live my life as one already predjudiced against whatever came my way.  I was already afraid.  Depression has hit its mark because I knew it would! I knew depression would hit me, not from lack of trying to fight against it, but because I\'d seen it come before.   God has given me the gift of faith for a reason.  I\'d been using it to believe many different things for other people, but very rarely for myself. The very gift God gave me to believe wasn\'t good enough for me since I didn\'t believe I was good enough.  I\'ve accepted lies from Satan that I\'m not good enough.  My many failings are thrust in my mind to make me think I\'m not what God wants.  I know, in my over-analyzing brain, that by myself, I\'m nothing.  My heart doesn\'t get it.  I\'ve been hurt, yes, and I\'m tired of holding on to it.  My shield was made by using the pain to defend myself.  Never trust a man because I was hurt by a man.  Never write again because I will only be rejected like before.  Never trust myself when I think I\'m happy because I will do something to screw it up.  I\'m only causing myself more pain.    My shield has fallen completely apart and I find there isn\'t much else defending me.  My attempts at self-preservation have failed.  Your shield of faith?  Yep, it\'s pretty heavy!  I don\'t know if I can believe by myself.  Too many times have proven I can\'t do it alone.  Faith is the things hoped for, the things not YET seen.  I hope I will see me grown in faith.  In the meantime, the faith You gave me?  Will you make it grow?  Make me know that You are God, I\'m not.  You are love and won\'t ever stop.  You never fail even if I always do.  You are my God.   '),(4013,2607577,'21,june,2004','2004-06-21 00:00:00','    I don\'t have anything specific to write today.  I just wanted to get on here and praise God.  He\'s just so amazing to me.  And I don\'t do this nearly enough.    Besides the obvious, HE SAVED ME, He is God.    I drove around last night with windows down and hair flying out through the sun roof of my mom\'s car just looking at things.    When I first got my license, I did this a lot.  I would just drive anywhere and everywhere I could feeling the freedom of BEING ABLE to drive but also something more.  I was feeling the freedom of being able to live and serve God.  No matter where I go or what I do, I have that freedom still.    I know I won\'t always FEEL this way, but that same freedom will be burning inside of me.  I\'ll never be held captive by this world again.  The moment I was set free, I was SET FREE INDEED!!!!!!  My God is my Captivator, but also my liberator.  His captivation isn\'t demanded, it is asked for.  He asks if He can captivate my imagination, my thoughts, my dreams.  He is a Gentleman.  He never demands anything of me, but He asks.  That is my freedom.    I don\'t revel in the idea that my freedom is mine for the taking, but that it has been so graciously given to me without any price to myself.  Sure, I had to give up despair, sorrow, pain, agony, self-loathing, self-pity, and just general self-hatred.  God, that\'s just not a very fair trade, is it?    But outside of myself and my freedom, there is this entire other universe He dwells in of which I have no idea.  He\'s made things I will never see in this lifetime, people I will never meet, and places even the greatest scientist or explorer could never find.  And He made it just because He was pleased with it.  How many other \"gods\" of this world can claim that?  I think it is the Hindu faith that claims reincarnation, but I could be wrong.  (Spaced it for a second.)  Anyway, reincarnation means that they COME BACK to the same place they left.  They may come back as a \"higher being\" but they still come back to this defeated world and will still face the same end I will: death.  That\'s not exactly something I would look forward to.  What if they saw what I\'ve seen?  Or something even greater?  A life free of having to come back!!!!!!  A life free of the past\'s sins.  India, when will you be free?  Ok, I\'m gettin wordy.  Sorry.  I\'ll end with this.  My God is great.  And even if I didn\'t know it, He\'s still GREAT!!!!   '),(4014,2607577,'21,june,2004','2004-06-21 00:00:00','    I was really impacted by the teaching tape this morning at intern training.  When everyone left the room and thought I had fallen asleep, I was actually thinking and praying.  I remember when I was in high school, I used to love to just be in God\'s presence.  At one point in time, I remember I would get up early to worship, study teaching tapes after school, and read, pray and worship right before I went to sleep.  I still love his presence, and at times can sit for hours just being with him.  It is hard to find the time now to do it daily, though.  Sure I read and pray daily, but I don\'t think that is enough.  On the tape he said, \"Your ministry to God\'s heart is the only ministry that only you can do.\"  Everything else we do could be done by someone else.  The entire dream center could be done by other people, but no one can stand in the face of God for me or you.   '),(4015,2607577,'21,june,2004','2004-06-21 00:00:00','    Hello,  My heart has been burdened lately with some stuff and I feel that now is the time to share. Who goes to heaven the good people? How do you quantify good? When it is good enough? Let\'s explore this together.  I was under the impression that only the good would make it to heaven and well, I was disqualified from that at about age Fifteen. I can remember believing that twisted propaganda for the most of my adult life. I truly believed, that my chance of being good enough to make it to that glorious palace in the sky had long been ruined, because of the incredibly bad choices made as an adolescent and adult.  Then it hit me, who is good enough to go to heaven, no one. You cannot be good enough to go to heaven, good is a word that has many definitions to many people. Good game Billy, Jimmy has a good Heart. Those two phrases have one thing in common and that is only the word \"good\".  The context of  how you apply the word good is broad and encompassing. I believe the word is used to freely and has become mundane in it\'s very nature. My point is that the word good should be applied in the appropriate situation and not thrown around like a medieval ball and chain to render people helpless. We must as passionate lovers of Christ spread the word to all people, that it is the FORGIVEN that make to heaven.  Trebby   '),(4016,2607577,'21,june,2004','2004-06-21 00:00:00','    Working with the girls at Anchor Families is hard stuff!  There\'s a lot I can and can\'t do with them.  Thank God, though!  I can take them to church!  They are so hungry for companionship, love, and attention.    I totally can relate.  They are living in a house with people they would not have picked and are not required to love since there aren\'t family ties.  Actually, the ties they DID have with family were severed due to circumstances they didn\'t create.  It\'s hard.  One of the girls was just weeping yesterday.  She had misbehaved and had to face the consequences set up by the rules of a system not set up to discipline in love.  She wants so bad to be able to go back to her mom, of course.  And she fears she has messed up so bad she won\'t be able to go back at all.    How many times have I said that?  And to God?  And how many times have I been reassured there was nothing I could ever do to keep myself away from God if I didn\'t want to?  These girls, (and the boys, I assume) don\'t have that reassurance.  They don\'t know how to get that all-consuming love!  Ugh!  My arms just ache to grab onto them, hold them tight, and tell them all about this awesome Love that wants to wrap them up forever!    God, make the girls\' home happen, quick!!!  They need to be in a home where God is shown off, leaned upon, and taught.  They need to see what it is to be a woman of God...  BELOVED!    Yesterday, while I was working, I sat and looked at one of the girls.  I saw that her face was missing something.  A few minutes later, I looked at one of the staff members whom I knew to be a Christian.  I saw what the young girl\'s face was missing.  While looking at the staff member, I could see what her face was like at peace with herself.  She was completely confident of being loved: by God.  The young girl, she didn\'t have ANYTHING in her life that was permanent, let alone love.  I couldn\'t see peace in her at all.    This description isn\'t exactly right.  What I couldn\'t see?  BEAUTIFUL, feminine peace...  A face that had a Lover.  Jesus.  That\'s better.  That\'s what I saw.  God, grant the speed to progress the house.  Let other people catch the vision.  (Especially those you have blessed with the $$$ for this specific purpose!:))  These girls need You!   '),(4017,2607577,'21,june,2004','2004-06-21 00:00:00','    I was looking at this list Ms. Alliece, my spiritual momma from St. Louis, had me make up.  Here it is:  5 Things I Don\'t Like ABout Myself: 1.  My Mouth 2.  Self-Esteem 3.  Unorganized Personal Life 4.  Time Management 5.  Thought Process  5 Things I Want To Change: 1.  Depression to Steady Joy 2.  Dependency on Men for Self-Image 3.  Thought Process 4.  Organization 5.  How I Deal with Confrontation  I\'m happy to say that God has been working miracles in these areas already.  And you know how it all started?  With Him MAKING me go to Ms. Alliece and TELL her all my junk.  I let every single thing out there, recognized it, and renounced it.  Cleansing Stream has a similar process, but Ms. Alliece started it all for me.    God, thank You for keeping me sane!  And thanks for working on me.  I know there is a lot to be done in order for you to get me into the right timing, frame of mind, and condition to be able to do Your work.  I\'m just amazed that You\'d even want to use a poor, pitiful wretch like me!!!!!!  Thank You!!!   ****Mini Update**** Whoa... This week has been hard!!!!! That\'s the only word I can think of to describe it. Hard!  Jason re-entered my life and I went to bits for a little while. Dad attacked me and I told him where I stand on everything. And I\'m still here!  God, there is so much. So much.   I\'ve been sick for the past 2 1/2 months and it\'s coming to a head. God, please heal me or show me how to GRACIOUSLY live with this. It\'s confusing me.  I don\'t ache for Jason anymore.  I don\'t fear being alone anymore.  And I know that since I\'ve said that now, satan will try to attack me with those emotions, but I know how to deal with them now. Thanks to Cleansing Stream.  I\'m not worried about not having a husband right now. I\'m not worried about not being in Africa yet. I\'m not even worried about not having my own place yet or a job that pays well. I\'m just not worried. And satan can\'t take that away from me.   '),(4018,2607577,'21,june,2004','2004-06-21 00:00:00','    Hey everybody, I just wanted to write a little sumpin\'sumpin\'!  I know, I don\'t usually scribe on this bad boy, but I have seen the light. haha  Anyway, I want to encourage all of you in the work that you are doing.  It is very obvious to many that God has His hand on our ministry.  The way He is exploding our ministry is sometimes a little overwhelming, but exciting at the same time.  I would ask that we would all continue to allow God to go deeper in our lives, that we would allow Him to ravish our hearts, so that it will not be us living but Him.  He has so many great things that He wants to do through us, and it is our responsiblility to make ourselves ready to fully inhabit His Spirit.  I am so proud of all of you, what a gift mix God has given us, I am blessed above all men to work along side such quality lovers of God.  Thank you so much for making my life so enjoyable as we labor in love together.  It is so important that we continue to seek humility as we see His greatness revealed.  We must stay meek, and contrite!  He deserves all of the Glory, and I know we will give it to Him.  Keep being radically violent, passionately destructive lovers of God.  Always remember, LOVERS MAKE BETTER SERVANTS!!!!!!!!!!!!!   '),(4019,2607577,'21,june,2004','2004-06-21 00:00:00','    Greetings,  I have been like a sponge the past six months soaking up every morsel the word of God reveals. Anna, you should continue to listen closely to what the holy spirit is telling you, regarding your spiritual gifting the one thing you do not want to get caught up in, is questioning that unmistakable transformation the holy spirit is wanting to make within you. I have always thought, I was a relationship person, but the context was strictly about business relationships. I have attributed my success in the business world strictly to \"my ability\" in gaining the trust and bonding with the people, whom I wanted to forge a new partnership with. WOW! Then it hit me like a ton of bricks this was God preparing me for the work he had planned for me all along. I am not in hospital heights for some unknown reason, God has dropped me there and through this experience, I have found that my heart is for those people. This is truly what relationship evangelism is all about, you see relationships are truly about people and nothing else not the next big business deal.  I love being with everyone on Saturday and going out an administering the grace of God Through Jesus. I sometimes get in my truck and take a drive on Sunday afternoon and inevitably, I am drawn to hospital heights and the surrounding area, I find myself thinking about certain people who have spoken to me with a concern or a problem, or I think about the person who seem to be desperately wanting something but just did not know how to articulate that to me.  I am so blown away by all that has happened to me in the last six months, it has been the best of times and the worst of times. I had such a hard heart towards people, that did not look a certain way or act a certain way and now I find my self loving all people regardless. God has done a quick and miraculous work in me, and I  praise him for that. Glory be to God!!!   Trebby   '),(4020,2607577,'21,june,2004','2004-06-21 00:00:00','    Ok, I need to confess this.  Confess it and let it out there!  I haven\'t told anyone for fear it was me wanting this and not what God was telling me.    On several different occasions, people have prophesied that my hands would be used by God to heal.  (This is terrifying me to TELL people.)  Sunday morning, God told me to go and pray for this woman and I almost didn\'t, thinking it was just my mind wandering and not God actually speaking to me.  (I almost don\'t want to say anymore for fear that satan will try to dement my mind anymore than he already has!)  So, seeing in my mind\'s eye what would happen if I left and just walked out of church as usual, I went and prayed with her.  Please don\'t misunderstand me, I don\'t want to bring attention to the fact of this, this... prophecy.  I want to be held accountable for what I do and what I say, so I\'m putting this on here.  And I want to make sure that my actions follow the Word.  What do I do with the nagging notions coming into my head?  They are doubts and definitly from the devil himself.  He\'s trying to tell me that I\'m taking pride in this.  THAT IS EXACTLY WHAT I FEAR!!!!  I DO NOT want this to be anything of me and that\'s why I haven\'t said anything about this until today.  The prophecy was first said to me long ago and I\'m just now putting my faith into it.    Of course, I know that as soon as I put my faith into a new arena, satan is going to try his hardest to destroy that faith.  I just don\'t know how to cope with it right now.  Having so much internal tormoil over unrelated issues and coming into this new......  \"calling,\" I guess you\'d say, is making me question everything and every action I\'m taking.    God, HELP!!!!!  That\'s all I know to say!   '),(4021,2607577,'21,june,2004','2004-06-21 00:00:00','    When I said there was more, I didn\'t mean about Jason.  I meant about my dad.  I\'m finally resolving some issues!!!!!!!!!  Sunday morning, I was completely ready for church and was actually in the car when my dad came out and said he wanted to talk!  Of course, I was frustrated since I was going to be late for church AND I\'d had a horrible night and was not in the mood for an argument.  My dad took me inside with my mom and let everything out he\'d been holding in.  As he sat there and basically degraded my entire existence, I was calm.  I was amazed beyond belief, but I was calm.  I was furiously angry, naturally, and incredibly hurt, but I was able to sit there and listen to it all until he finished.  On top of THAT, I was able to rationally reply to my father.  Dad was angry (and is still working on) with me about being divorced, going through financial crisis, moving out to St. Louis for a year without getting paid, and coming back to live at home again.  Sound like a lot?  Yep, it was.  My dad and I\'d talked about most of this before and I thought we\'d resolved it and were moving on.  Apparently not!  Dad harbors such resentment towards me, especially when I decided not to go back to college after the divore was final.  So, when he re-vented all of these things, I was righteously indignant!!!!!  I told my father that he was not following the Bible word-for-word since he STILL had not forgiven me!  I also told him he did not come about this whole situation the biblical way since he\'d been telling my mother to deal with me and would never bring up the topic to me, personally.  He let it fester in his own soul, and now he was angry...  AT ME!!!!!  I don\'t mean this to sound disrespectful, but I let him have it!  I told him WHY I\'d made so many mistakes and WHY I always searched for the answers in relationships with men EXACTLY LIKE MY FATHER!!!!  Understand, I told him I knew that I MADE those decisions and no other person could force me.  But, Dad also had to realize how much he influenced those decisions.  What I could not find in my father, I searched for in too many relationships that ended horribly.  I was looking for the love and acceptance I wasn\'t shown by him.  Encouragement, loving correction, and just plain THERE!!!!!!!  By the time I told him all of this, I was in tears and couldn\'t hold them anymore.  But they were good tears.  I was able to tell my father just where my boundary was with him now.  I am no longer dependent on his opinion of me and he can no longer affect my decisions as he did before.  I am a woman of God, and DAUGHTER of God.  True, I would love to be able to have a strong relationship with my father.  But, I don\'t crave or long for it like I did growing up.  I know who I am now, through running the gauntlet of trial and error.  It may have been the hard way to figure things out, but it taught me just the same.  I do not need Dad\'s approval anymore.  I know on my own what God wants for me.    Now, God.  Teach me how to honor my father in the purest form I can.  I cannot agree with my dad on very much, so teach me how to honor him without having to become dependent on him for my emotional behavior and decisions.  Teach me how to live dependently on You!  *********  There has been so much more that\'s happened this week.  I\'d love to tell you guys more, but the words are too many right now.  I don\'t want to strain anyone\'s eyes right now!  If you want, you can ask me. But for now, I want to keep it just between me and God.  Just for a while.  It\'s a secret!!!!!!!!   '),(4022,2607577,'21,june,2004','2004-06-21 00:00:00','    We are just as blessed to have you on board Sarah.  It is so nice to see the nieve young kid go off to school, and a powerfully focused and passionate woman of God come back.  We are excited about what you are going to do for the Lord.  As you can see Interns/Staff, God is bringing people from the N/S/E/W to help with our work.  New Castle is a seed bed for Gods anointing and power to be demonstrated.  WOW!!!  Just think, in 5 years we won\'t even be able to recognize ourselves, let alone our city!  City transformation is coming, and we are in the breaker stages of that.  Isn\'t that exciting.  I am so blessed to work along side such passionate and dedicated servants of the Lord.  Thank you all for making Jesus famous on the earth!   '),(4023,2607577,'21,june,2004','2004-06-21 00:00:00','    Hey everybody!!  This is my first blog; how exciting is that?  I don\'t really have a lot to say (y\'all thinking that\'s a joke, quit laughing).  I just felt like I should blog since I\'ve been here a few weeks now.  I am loving it here.  I am so excited about working with the Dream Center and, what over the past few weeks I have discovered, an amazingly unique group of people.                        Rob, you have become an incredible man of God since I left for college three years ago, not that you were all that bad before, but now I can truly say I look up to you and admire your heart possibly more than any one else I know right now.      And, you still give the best hugs on earth.  Trebby, I love you so much.  I think you are just the neatest person.  You are so quiet, but there is so much power inside you.  I love talking to you and just walking the block.  Chris, I am just beginning to get to know you, but already I love hanging out with you.  When you smile, it fills the room.  Anna, Cynthia, Kristie:  You have all become good friends quickly and have helped make my transition to Indiana much easier.  I know that all three of you have incredible hearts for people and for God and I am loving getting to know you.  Everybody else rocks, too.  Working with the Dream Center is the most incredible thing I have ever been involoved in.  God has given me such a heart for the people of New Castle, I have been driving around some nights after work, praying.  I can\'t wait to start Metro Kidz this Spring.  We have such an incredible opportunity to affect a generation before they become involved in the same things as the generation before them.   '),(4024,2607577,'21,june,2004','2004-06-21 00:00:00','    Ok, I only have a few minutes, but I\'ve been dying to get this all out.  This weekend was...  Extreme.  I don\'t want to say that it was horrible, but it was pretty darn close.  It was definitely filled with twists and turns!!!!!  Last week, I knew something was about to happen, something was coming that would change up some things.  So, as a result, I was very unsettled and restless.  (It also didn\'t help that I\'m adjusting to working 3rd\'s!)  And Saturday morning dawned after a sleepless night from work and left me with a definistley uneasy spirit.  I assumed it was because I\'m a woman and over-emotional.  Nope, for once I wasn\'t just being hormonal!!!!:)  Something WAS up!  TheG.I.G. was an awesome show, and towards the end, I got a call from a guy who said he wanted to talk to Rob.  Well, I asked his name, \"Will.\"  And last name, \"Tester.\"  And asked him what the call was regarding, \"It\'s an emergency.\"  So, I gave the call over to Rob not getting what the name was at first.  When Rob got off the phone, it hit me.  I asked him who was testing my will, in a joking way.  Rob told me to wait 2 minutes and I\'d find out.  I was thinking it was Joe or Abram, someone who usually comes to the show with us.  I wasn\'t expecting Jason.  AT ALL!!!!!!!!!  For once, my mind was completely clear of him.  I wasn\'t wondering, obsessing, worrying...  I was totally involved with working the phones and computer.  So, when Jason walked in...  My whole being just dropped.  Keep in mind, I am the best actress I know.  (I\'m not bragging.  I hd this performance to prove it!)  I kept it cool and ordinary.  (As best I could with my mind screaming and my heart wailing.)  He came in at the tail end of the show and got on the air with Rob and Marita for a few minutes and talked about First Assembly\'s youth group.  In between times, he asked about my family, how I was, what I was doing...  Nonchalant.  He goes to leave and tells me to say hello to my family since he doesn\'t get to see them very often.    He leaves, I break.  A flood of uncontrollable tears I\'d been holding that whole time just broke me.  Thank God Marita grabbed me, or I\'d have been running through the streets...  There\'s more, but I have to go.  The Passion tonight.  God, I need something.  More of You.  Less of me.  Show me.   '),(4025,2607577,'21,june,2004','2004-06-21 00:00:00','    Hello,  I want to start by giving God all the glory!!!!  A couple of weeks ago a group of Jesus lovers prayed for my wife and I, out at Everitt\'s house.  My wife was preparing at the time for a surgery, that was complicated by the fact she has a rare enzyme deficiency that will not allow her to be put under by normal medication. She had the surgery and is recovering at light speed, the doctor who performed the procedure told her that \"he has never had anyone recover as quickly as she has\". Let me back up just a little, my wife let me pray for her also, now this might not sound like a big deal to all you happily married folk but for me, it was a huge step forward, considering where my relationship has been with my wife. (Translation: not to good) I have thought about what this Doc said and the more I think about it, the more questions I have. See, I am not surprised at her recovery because the power of God supersedes any medical abnormality, and with a bunch of Jesus freaks praying together in the house of a Man and Woman of God, how could the outcome have been any different. Getting back to the questions I have, and something for us all to think about, was this Doctor just being compassionate, nice or could it be that none of his other patients ever received such a powerful message from God, through miraculous healing. You see, I hope that is not true because if it is, what message do you clearly HEAR?. Something to think about  Thank you all for loving me, just as I am. God is working in my marriage and some great things will happen. I love my wife and have found an appreciation for her that I never new before. I am a work in progress but I know one thing, God will never give up on me and he will not give up on you either. (You know who you are)  God is good!!!!!!!!!!  Treb   '),(4026,2607577,'21,june,2004','2004-06-21 00:00:00','    urlLink    urlLink audio post  powered by  urlLink audblog   '),(4027,2607577,'21,june,2004','2004-06-21 00:00:00','    Hello!!  Today opened my eyes to some things.  A couple interns read there testimony.  Let me just say first of all.  GOOD for you!! You guys earned a deeper respect than even what I had before. I LOVE YOU!! See personally that is not something I want to do.  Especially after I realized the testimony I gave Rob was not really my testimony.  In a part yes, but it was missing SOOOOOOO many things.  Things that I just have started to remember again.  I had pushed them so far down that really they never bothered me.  To be honest I thought I had forgiven whoever because emotionally I am ok!!  It\'s weird even as GOD has revealed these things to me, I still don\'t believe they happened.  You know what I mean.   I KNOW they did but I am just not feeling it yet.  Its like I am in a certain place and someone is telling me all this stuff however its happening to them.  Don\'t get me wrong I have spoken them out loud and it make me sick to my stomach.  See that\'s where I am at right now!  I have NO clue how to deal with this stuff.  I mean REALLY deal with it.  Ok GOD you showed this stuff to me now fix me!!  I realized I am a very emotional person.  REALLY I am..hehe  I just don\'t show it to people as much!!  Yeah the fun laughing side of me I show ALL the time.  Its fun to laugh!  But the stuff that I deal with is not something I personally want to deal with in front of anyone.  I guess I am kind of scared that when God finally does say hey ITS TIME!! I am going to be around everyone and their Momma!!  I know I am like anyone else who needs prayer and help.  However I feel especially with working with the teens that when I am around them its NOT about me.   AND I know since this is how I feel about things GOD is going to show me in some way something about myself.    I also realized this HAS to happen so I can heal and be able to share.  I want to be able to have that testimony!!  The one that says I went through all this crap, but guess what!  GOD was there for me!  He healed me of everything and put people in my life to teach me what I need to know.  I have NO DOUBT whatsoever that is exactly what is going to happen.  GOD is going to stretch me in a way that is very uncomfortable to me.    I bought this book yesterday at a Christian bookstore in Indy!  Beauty for Ashes by Joyce Meyer!!  The funny thing is about six months ago I bought it for someone else.   I honestly did not think I needed it.  I am not completely sure I do now.  However I need something, someone to teach me what the heck I am to do.  The funny thing through all this I am not a wreck!!  I am still happy and full of Joy... Because in my heart I know my latter will be greater.   I am ready to jump in I just need help learning how to jump!   '),(4028,2607577,'21,june,2004','2004-06-21 00:00:00','    I know what you mean about the roller coasters, Treb.  Staying off the ones I know will make me throw up is my problem!!!!!  As you guys now know, I have a job and I love it.  Due to privacy reasons for the girls, I can\'t write out a whole lot, but I can tell you that two of the girls have asked if they can get permission to start coming to church at New Covenant.  They had been going to Spiceland Pike but they just don\'t feel involved or included in a lot of things, so they haven\'t been going in quite some time.  We\'ll see what God does there.  I recently have had the urge...  No, that\'s not the right word...  I don\'t know.  Let\'s put it this way.  My heart has been aching for Africa unusually strongly lately.  I know this isn\'t the right timing, but I can\'t help but wonder if that time is drawing nearer.    For those of you who don\'t know my entire life\'s history as of yet, (give it some time :)!!!) then you don\'t know that I\'ve been called to Africa.  More specifically, I\'ve been called to Zaire, or what\'s now known as the Republic of the Congo.  Right now, there is a national upheaval with so many civil wars and uprisings that no missionary has yet dared enter the country for fear of their lives and those lives of the people they would be tending.   About 10 months ago, Dave Ohlerking, the president and founder of Children\'s Cup, an international mission program as well as relief program, visited the Dream Center.  He spoke to the church about the call and how to define the specifics of the call.  Along with several other men in the Children\'s Cup foundation, he has set up many different orphanages for children left alone due to parents dying of aids and other calamities and Dave told us about the severe need to minister to the natives without any sort of hope.  Don\'t misunderstand me.  While his message was eloquent and very moving, his fine speech wasn\'t what convinced me to go to Africa.  The call has been there since I can remember.  I just haven\'t had any idea how to fulfill it.  Dave may be helping with that.  A few months late, Dave came with two of his colleagues from Children\'s Cup who were workind in different countires.  They took all of the interns aside for a time of questioning since they are starting an intern program for those interested in African ministries.  They wanted our input on how to run an internship program and gave us the opportunity to interview them about what they do in Africa.  Dave already knew about the call for Africa I have, so he didn\'t really say much to try to provoke me into questioning, he said later on.  But I couldn\'t help but ask for as much information as I could about what was going on in Zaire and the status of the people.  It was so hard to hear...  God, help them!  But, anyway, after the interview, Dave and Dave, (a different Dave) and Ben took me aside and told me they would be honored if I would join their internship program.  I would study under them for a year, whenever I was ready, and then be launched into Zaire with their support and blessing.    God, why is this so heavy on my heart right now?  I\'ve commited to THIS internship program...  After just going through a PREVIOUS internship...  And another one?  What\'s going on?  AM I supposed to go?  If so, when?  And how in the WORLD am I going to raise that kind of support?  You know the problems, God.  And only You have the RIGHT answers.  Any time You\'re ready to give those up, let me know!!!!!   *****Insert***** The reason I put this on here?  I need to hold myself accountable.  Especially when it comes to God\'s plan for my life.  I need to make sure this stuff isn\'t festering in me, like it has been, and then pops out in an unmanagable way.  I also know that getting this out helps me to think it through and let my mind process what\'s going on inside it.  Make sense?   '),(4029,2607577,'21,june,2004','2004-06-21 00:00:00','    Greetings,  I have always identified my self worth based off my Job. I guess hiding my identity within my work made me feel warm and fuzzy. WOW!!! What a bunch of crap!. I have been driven by my job the last Fifteen year\'s with my emotions hinging upon my success or failure. Needless to say, what a roller coaster ride, when it was good, man was it good, yes, there is a but, when it was bad, it was bad. I had gotten so caught up in my job that all the really important things in my life became irrelevant...Wife,kids,house,GOD. I always knew in the back of my mind that I really had my priorities all screwed up, but was afraid to admit it, you know that thing they call pride. The moral of the story is a job should be put in proper perspective for what it is. It is not who I am, it is my means of supporting my family, I have made $10,000 a year and 80,000 a year it makes no difference, I was still miserable. I have recently been offered another position within our company, which will allow me to work closer to home and give me an additional 10 to 15 extra hours a week. I listened to God before making a decision and the answer I kept receiving was think about others before self, that has made my decision simple, I will accept this job and do the best job possible, but I refuse to become the job. I want to spend extra time with my wife and the ministry...Hopefully, those two will be bound together someday.  I love this journey God has me on, you know, it\'s like your favorite ride at the amusement park, you want to get back on as soon as the ride ends.   Treb   '),(4030,2607577,'21,june,2004','2004-06-21 00:00:00','    Cleansing Stream...  I don\'t know about this class.    We are talking about unforgiveness and we listened to a teaching by Joyce Meyer last night in class.  (Suprising, I haven\'t heard that one.)  And she was telling us about how unforgiveness will block us from receiving our breakthrough.  Duh, I know this in my head, but another revelation came through.  (I need a new word for revelation.  There have been so many, it\'s lost its flavor.)    My own unforgiveness was against my ex-husband and all the other men in my life I\'d allowed to hurt me.  There was unforgiveness against them as well as against myself for allowing them to do that to me.  And what have I been praying for?  A husband, a man of God to come along side me and lead me.    I need to write a letter to them, ALL of them, and ask for their forgiveness and to forgive them.  I don\'t dare do it.  What will happen?  Will the unthinkable happen and God reconciles my ex-husband back to me?  Or Jason?  Or anyone else I think of as dead in my eyes?  That is the unthinkable, the unimaginable, the horrible thing I dread happening.  Does this sound unrealistic?  My heart DOES NOT WANT reconciliation with any of these men...  And maybe not with myself, either.  I just don\'t know if I am actually willing to let all of that go and let those possibilities haunt me.  I\'ve consecrated these things, even myself.  They are set apart and do not get int the way when I am trying to walk out my life.  The thing is, there is a bitterness there I can\'t overcome alone.    I\'ve consciously forgiven the men.  I haven\'t myself.  What do I do?  I can\'t reopen those wounds for inspection again.  I especially cannot open them for the very people who inflicted them...  God, help.  Those wounds still haven\'t healed properly.  And I am desperately afraid of having to put a new bandage on an old wound.  What do I do?  How can You heal something I can\'t re-reveal?   '),(4031,2607577,'21,june,2004','2004-06-21 00:00:00','    Hello all!! God has been doing a lot of changes in me the past couple of weeks.  I have had a lot of decisions to make and a lot of things that have been coming up.  Mainly attacks against me and what I should be doing.  The attacks of course are coming from the people that are closest to me.  Some of them are just in my own head and I am trying to figure it out.  Its weird because last night is when I realized I was trying to do it on my own.  At least I thought I was!  The same questions and situations have been coming up for a while and I have been so consumed with doing it the way God wants me to do it.  That it makes me question things!!  Does that make sense!  I so badly want to be in his will and do what I am called to do, however I am impatient.  I want to do more for people.  I don\'t want more things to do, I want more quality time with people.  Especially with the teens in the youth.  The funny thing is God is telling me I am doing what I am suppose to be doing.  I am exactly where I am suppose to be and doing what I can.  I think personally I just want him to equip me right now at this very moment with everything I need.  So I can start working for him.  I have not been on the block the past couple of weeks and I am really missing the people.  I want to help more and I think God is trying to teach me and get me to learn that the little stuff counts just as much as the big stuff.  I might be doing something that in my mind isn\'t a big deal, however I am  not realizing how much it means to someone else.  So I am feeling like I have not done enough.  Thank you GOD for working on me and through me.  I know I need a lot of work.    Sunday Eric was saying how he was watching The Passion and got that revelation of God\'s LOVE!!  When these men were beating Jesus Christ he had this compassion for the men.  When he said that I was like WHAT!  I wanted to jump into the movie and start beating people.  I was wanting GOD to punish them... Then by the time it was over I was like WOW!!  I totally get it!!  I understand Jesus Christ loved the people that tortured him like I love Jesus.   I thought I understood Gods love!!  Do you ever think you already had a revelation of something and then realized oppppss maybe not!  I want that and I NEED God to help me love like he does.   I NEED him to search my heart and pluck the negative things that I don\'t need to make room for all the wonderful things that are coming.  I am so ready for that I get so upset especially with my husband when he does something wrong.  I am trying so hard to show him that God lives inside me.  There are times though I am showing him that but deep down I am upset.  That is something I KNOW God is working on in me.  Then I am thinking WOW my heart is not in the right place in that area.  It does not matter how much I act like Jesus is in my life.. If he isn\'t in my heart its almost pointless to try!!  God help me not to be so hard on my husband.  Help me to get a BETTER revelation of your love!!! Hey I am yours and I am willing to do whatever I have to do to be your servant.  I am so excited for what you are about to do in me.  I know its coming and I know its going to be great.  That\'s the best part for me.  I don\'t need to see things to know that they are coming.  Praise God!!  Well those were some of my thoughts for the day..   Sometimes they don\'t come out the way I would like them to.  So I am not a big fan in sharing them..  Luv you guys have an Awesome week!   '),(4032,2607577,'21,june,2004','2004-06-21 00:00:00','    Visions and dreams have been coming more and more each day.  Well, I asked for them, so what else should I expect?:)  No, really, God has been freaking me out lately.  Lots of things He\'s been showing me...  Frightening!  On Sat. night, I had a weird confirmation from God.  Back in St. Louis when I was debating on what to do next after graduation, I asked God to give me confirmation after the fact of making a decision.  I asked specifically if He would give me a dream and that dream would come true.  Make sense?  I wanted to know that if I was doing the right thing at the right time, then my dream would have been a confirmation.    I was sitting at the computer at the radio station with about 15 min. left of the program when I looked down at my hands and knew what else I was about to see and what the conversation behind me would be.  I\'d dreamt about it.  Thank You, God!    I was so afraid that I\'d made the wrong choice coming back home.  See, I had two amazing offers I\'d turned down.  They would have been offers leading to great provision and prospering.  I could have elected to stay at the Dream Center and enter their resident missionary program and immediately become what I knew God was calling me to be: a missionary.  I would have been able to travel the world on different missions trips with a team of amazing people from Joyce Meyer\'s ministries and learned from them!!!  How cool would that have been!?!?!?!?!  The other offer I was made was to go to Alabama to an amazing church and group of people I already felt were like family after only having known them for a few days at the Dream Center.  They\'d asked if I would come to their church and help start a campus ministry in association with their already growing college and career group.  They were going to pay room and board, transportation, and a small salary on top of that!!!!  And I was going to a place where I was desperately needed and wanted.  How much better could that have been?  I would also have had time to go back to school....   But I came home, much to the consternation of both interested parties.  I knew I had a purpose for being here.    I\'ve had confirmation out the wazoo telling me I did not screw up.  This was right.  This was good.  I didn\'t fail this time.  The main confirmation I\'ve had came on Friday.  I\'d been praying for a job but hadn\'t been able to get ANYTHING, let alone the job of my dreams!  In fact, I was told I was over-qualified to work at Taco Bell!!!!:)  Funny.    My heart has always been to help teen girls in trouble.  My prayer was answered in this job.  I\'m working at Anchor Families now, praise God!!!!!!  And I know two of the girls who live there.  They actually used to be \"my girls\" at my other church but had lost contact with during this last year.  And I know there are other kids I\'m going to meet there who need to hear what God would say!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  Praise God!!!!  THis is my heart, this is my passion.  A Missionary is only a title and can live and work anywhere God sends her.    Guess I didn\'t screw up after all!   '),(4033,2607577,'21,june,2004','2004-06-21 00:00:00','    Yes!!!!  It\'s about time some new revelations were shared on here!!!  I was beginning to think y\'all didn\'t have time to share the awesome stuff (and even not-so-awesome stuff) God has been showing you.  I wanna be able to see your revelations, your growth as well as my own.  It gets a little boring reading mine!!!:)  These past few days have been great.  God has really been answering some tough questions I put before Him some time ago.  And, yes, even some answers on the possible \"Mr. Right.\"    Someone recently asked me if I truly thought there was a certain someone specifically for me.  It made it sound like this person thought that there is a certain TYPE of person each different person is meant to be with, of their own kind, as you might say.  I totally don\'t believe that.  At all!  I honestly believe there is just one person out there meant to be with me and that\'s that.  He is made to compliment me in every way and I, him.  Our weaknesses and strengths will mesh together so that we fit...  Like a puzzle piece.    How awful to have the menality that there is only a certain TYPE of person meant to be with you.  That implies God only makes types of people and not individual people.  Who wants to be a type?  I enjoy the perfection of imperfection which makes up Anna, which makes up every person I meet. True, I may not LIKE the person, but I like the fact I\'m not what they are, even in the same \"class.\"  Thank God there aren\'t classes He puts us into.  But instead, we are \"fearfully and wonderfully made.\"    So, my husband is \"fearfully and wonderfully made.....\"  I like that.  Yep, he\'ll probably frighten me sometimes, hence the \"fearfully.\"  And, yep, I will wonder about him sometimes.  That\'s the good part.  I won\'t be able to predict him.  He is his own person, made by God just specifically to fit me!!!!!  Thank God!!!!!!!  And WOOHOO!!!!!!!!!!!!!!!!!!!!!!!!   '),(4034,2607577,'21,june,2004','2004-06-21 00:00:00','    I have been bad about blogging much...sorry Anna...I don\'t always know what to write.  I guess sometimes I\'m not really sure what the whole blogging thing is about, like what you are supposed to put on it and stuff. Lately God has really been stirring in me how we have the ability to bring change.  I was in someone\'s living room with Rob yesterday...I don\'t even know the man\'s name...he was depressed...he said he was worth more dead than alive...that he wasn\'t worth anything...Rob held him as this grown man cried in his arms...Rob told him that Jesus loved him, and wanted to change his life, give him hope...Rob offered to pray for him...the man said it wouldn\'t do any good...I just wanted to cry, hold the man, give him hope, convince him Jesus loved him. I was praying last night after I went home and was just talking to God, asking him what I could do, how could we begin to see the change in people like that man, and communities like Heights.  I was reading Isaiah 58, and I got it.  I had a new level of revelation about fasting.    Isaiah 58:6-12 reads:  \"Is not this the kind of fasting I have chosen:  to loose the chains of injustice and untie the cords of the yoke, to set the oppressed free and break every yoke?  Is it not to share your food with the hungry and provide the poor wanderer with shelter --when you see the naked to clothe him, and not to turn aweay from your own flesh and blood?  Then your light will break forth like the dawn, and your healing will quickly appear; then your righteousness will go before you and the glory of the Lord will be your rear guard.  Then you will call, and the Lord will answer; you will call for help and he will say:  Here am I.  If you do away with the pointing finger and malicious talk, and if you spend yourselves in behalf of the hungry and satisfy the needs of the oppressed, then your light will rise in the darkness and your night will become like the noonday.  The Lord will guide you always; he will satisfy your needs in a sun-scorched land and will strengthen your frame.  You will be like a well-watered garden, like a spring whose waters never fail.  Your people will rebuild the ancient ruins and will raise up the age-old foundations; you will be called Repairer of Broken Walls, Restorer of Streets with Dwellings.\"  Wow.  I want to see those kinds of changes.  I want to see a community changed.  I am just crazy enough to believe that is what God wants to see, too.  Well, enough for tonight.   '),(4035,2607577,'21,june,2004','2004-06-21 00:00:00','    Ok!!! Anna I will help you out and give a little update on what is going on with me.  In a earlier blog I wrote that God was showing me lost memories!  That it was like I was watching a movie and all this bad stuff happen to someone else.  I was feeling ashamed and guilty!  Like I said I was not a wreck, however I did not understand why God was showing these memories now!  Well as the weeks have went by I have remembered in detail a lot of things.  It really explains where the source of my depression came from starting at a very young age.  I can understand why I did not want to remember this stuff because it was horrible.  Anyway God was protecting me in a huge way.  By not letting these memories take over me or consume me!  Well I also went to the cleansing stream retreat!  As I was getting prayer God was revealing to me a lot of different things.  I was just like WOW... and like...oh yeah I forgot about that!  God was showing me new stuff as well.  I remember the first retreat I went to last year I had no clue what was going on.  I did not understand it really.  I realize now God showed me this stuff so I could be broke free.  He knew that right now in my life I have complete trust and faith in him.  So I never doubted for a moment that what he was showing me was for some purpose.  I just realized last night at work what happened.  I was set free!!  I am not feeling the shame and guilt anymore!  I had abandonment issues as well!  I actually had a lot of what we were praying for, some was worse then others!! Guys you have to understand something.... I have been thinking about how I was going to talk about what happened to me.  I knew I needed prayer, but felt like I was going to get sick everytime I went to open my mouth to talk about it.    I was even to the point where I was going to try to set up some kind of counseling appointments.  I knew I needed to deal with it, but I just did not know how!  Well last night I started talking to a friend and I felt like God was telling me to talk to this friend about what had happened to me.  I was like GOD... how can I when it makes me sick to even open my mouth to speak the words.  He just said trust me Kristie!!  So I did!  you guys... it was easy!  I mean easy!!!!!  I just talked to her and told her about what happened to me and she started to cry.  I told her that it was alright that God set me free. RIGHT then I realized I had been set free since the retreat and just did not know it.  You guys GOD is so AWESOME!!  I can\'t imagine ever not having a desire to know him.  He is the only person that has made me feel completely safe.  I have never had that before!  THANK YOU JESUS!!!!   '),(4036,2607577,'21,june,2004','2004-06-21 00:00:00','    Not many people have been blogging lately.  What\'s the deal?  Anyway.  The continuation of yesterday.  There have been some major attacks going on since the end of Cleansing Stream and my newfound freedom has been gained.  MAJOR!!  The Devil has been workin over-time on my case lately.  Not a chance!!!!!  Ok, back to what happened this weekend.  Friday night, Kat came with me to Cleansing Stream and we both were looking forward to the time of worship and seeing what was going to happen at the prophesied \"life changing event.\"   Neither one of us had ever been to one, so we didn\'t know quite what to expect.  The worship, though we came in a bit late due to a site problem, was awesome!  They had a guy come over from the Indy House of Prayer.  Amazing!!!!!!  But, we didn\'t have much time to worship since we came in right at the tail end.  The session began and they immediately started with what we were expected to see that night.  They would talk about certain demons and strongholds that could be binding us, tell us what the sypmtoms were of that demon or stronghold, and then they\'d have us repeat a prayer of repentance, renouncement, and breaking.  After we prayed the prayer jointly, we all came up to the front to be prayed for by the leaders of the session.  (There were about 25-30 workers in all, intercessors and prayer partners.)  EVERYONE came up to be prayed for, no matter if we felt like that was something we dealt with specifically.  The prayer partners prayed with us in agreement that the demon or stronghold was broken and was demanded to flee.  They also prayed that God would fill that void left with His love, comfort, strength...  It was awesome.  Ok, now that you know the general scheme of things, here goes what happened to me.  One of the areas they first talked about was the spirit of unforgiveness.  I knew and had known for quite some time that God was dealing with me here.  I was STILL finding it so hard to forgive my ex-husband.  But more than that,  I still couldn\'t seem to forgive Jason for what he did.  And harder still, I couldn\'t forgive my dad for not being what I needed him to be.  So, we prayed the joint prayer and went up to be prayed for.  Joe had been sitting by me and had come in even later than Kat and I, so we chatted for a sec in line while we waited to be prayed for.  I KNEW I needed to concentrate back in on what God was doing right then, but I kind of ignored it.  (Alright, I\'m chicken.  I didn\'t WANT to deal with the tough stuff just yet!)    Well, when I got up there, some serious stuff was dealt with and that spirit was demanded to let go of me.  GOOD DEAL!  I went back to my seat and Joe was there again, so I chatted with him and started trying to disengage myself from what was going on...  Yeah, I\'m a chicken........  God decided to call me on it!!!!!  WHO\'D  A  THUNK!!!!  The service was interrupted because one of the prayer partners had a word from God saying that someone needed to move and get away by themselves so they could give their whole mind and attention to Him!!!!  She asked if we would move if we felt like that was us...  Nobody moved...  And then she decided to pray and get on to the next part... I got up and moved, alright?!?!  I moved to the back and got alone with God!!!!  I was the only one who moved!!!!  HA, HA!!!  Thank God, He made me move!!!!!  Because the next area was the spirit of rejection.....  MAN!!!!!!!!!  That was hard.  Really hard.  Pastor Eric was the one who spoke on it and I just couldn\'t stop those big \'ole fat tears from soaking the front of my shirt.  I was weeping, guys.  And hard.  Rejection has held my heart for so long, it was second nature to feel rejected.  In fact, I didn\'t know until this weekend what it was to be accepted, truly, by another human being.  Honest!  I\'m not exaggerating.  I did not feel accepted by anyone and felt I\'d been rejected by even God...  Why had God rejected me and left me alone when I was with Paul?  Why did He let me get hurt by so many men in my past?  Why did He reject His daughter?  Oh, gosh!  Did I ever have to repent of that, y\'all!!!!  Big time!  When we said the prayer, even a repeated prayer was hard to get through.  Then time came to come forward for prayer and the same lady who prayed with me about the unforgiveness got me again and prayed with me....  Ugh!!!  HARD!!!!!  I don\'t want to tell everything just yet about what happened during those prayer times.  It was intimate and I want to leave them as that for a while...  Just me and God for a bit.  But, it was something every person who feels like they\'ve been left waiting for someone to come and take them home realize that they\'ve always been home.  He\'s always been there, we just pushed him as far away from us as we could.  WE had rejected ourselves, and God.  The night went on and then it was time for Saturday!!!!!!  WOOHOO!!!!!!!!  Talk about some freedom!!!!!   My big breakthrough was when they talked about the spirit of abandonment...  UGH!!!!!!!!!!!!!!!!!!!  Tough stuff!!!!!!!  And not only did I have abandonment issues with my dad and the other men in my life, I had abandonment issues with the child I lost.    Yes, it\'s time I talked about it.  I lost a child not long after I was married.  I\'d felt such guilt about the miscarriage, guys.  You don\'t understand.  I WANTED, more than anything, for that baby NOT to live.  I didn\'t want it to have to endure the abuse I was seeing, hearing, and feeling...  So, I wished it would leave me, it would die.  I\'d thought that meant I had murdered it in my heart.  What I didn\'t see until Saturday was this: I felt like the baby had abandoned me.  It left me just like my husband did.  This may not make sense, but this abandonment spirit had such a firm grip on my heart and soul, it had me convinced that even the baby had abandoned me.  Oh, guys!  There is so much deception out there just waiting to grab our souls!!!  Don\'t let it start!!!!  Not even one drop of it!!!!!!!!!!!  I went up to be prayed for about that and it broke off.  Finally.  Not long after that, the leader of the retreat asked us all to close our eyes for 60 seconds.  She wanted us to think of one thing we would do if we didn\'t have a single ounce of fear around or in us.  I thought about it and I remembered what I wanted to do when I was first saved.  I wanted to sing, to dance, and to play the piano for God!  In front of His children, no less!!!  And not be afraid to praise Him like I absolutely wanted!!!!!!!  And, you know the drill, we prayed about it, and the guilt, the shame, and the condemnation LEFT!!!!!!!!  I ran to the back and let my feet do the leading!!!!!    I danced, guys!!!!!  I DANCED!!!!!!!  And this isn\'t the end of it!  Now that I\'ve danced and now that I\'m free, I can\'t be caged again.  I don\'t care if this bird is subjected to blights and bouts of whipping wind and storm, this bird it going to FLY!!!!!!!!!  I\'m free.  I can\'t help but keep saying it.   I AM FREE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!   '),(4037,2607577,'21,june,2004','2004-06-21 00:00:00','    Set free...  Wow.  Set free.    There\'s so much to tell and so little time to do it.  I wanted to get on today, really quickly, and ask you all to pray for me while the rest of my deliverance is coming to fruition.  But, guys, I am set free!!!!!!!!!!   '),(4038,2607577,'21,june,2004','2004-06-21 00:00:00','    Yep, as Joyce Meyer says, \"New Level, New Devil!\"   I\'m livin, walkin proof!    Remeber how I said that I was being tempted with three different opportunities to settle?  Yep, they are still trying to come back!!!!  Satan is just not giving up!  He knows I was tempted before, but he also saw that I resisted the temptation!!!!  Ugh!!!  Yes, I can resist this once again.  And yes, I will!  But, God, he has no right to try to come back in again.  I\'ve told all three of the men that I am just not made for the kind of life they would give me.  I\'d never be satisfied and would probably make their lives living HELL in the process.  That wouldn\'t be fair to me or them.  And yet, they keep on comin!    I turned off my cell phone for this reason.  (This and the fact that it came with a bill of almost $300 for one month!!!!!)  The voice mail is still working, for some odd reason, so I\'ve checked it to see if some of my girls have called it instead of my home.  And guess who has been on there?  The three temptations!  (This sounds like some kind of 50\'s rock band groupie story!)  Lord, get rid of them!  I don\'t want anything in my path except what YOU bring to me.  Nothing, Lord!  I don\'t need distractions or temptations.  I NEED more of Your love and reassurance surrounding me.    Don\'t get me wrong, folks, I know I screwed up by letting those temptations affect me too much.  But...  GRR!!!  STILL!!! What\'s the deal?    Satan must know that my real hubby is on the way...  Ah!  That\'s it!  And my hubby is such an awesome man of God that it\'s got Satan panicking and trying to set up every single pitfall he can!  WOOHOO!!!  Praise God, then!  And Satan can\'t do anything about it!  My man of God apparently is so strong in Him that he\'s coming quick!:)  I\'m in a weird mood, y\'all.  In case you hadn\'t guessed already.  I\'ve been fighting demons in dreams and in some of the people I\'ve been working with.  My methods and train of thought have become a little unorthodox because of the warfare!:)  SATAN IS GETTING HIS BUTT KICKED!  Praise JESUS!!!   '),(4039,2607577,'21,june,2004','2004-06-21 00:00:00','    I only have a couple of minutes before I have to be at work, but I needed to get on here and write for a bit.  My calming moments before the storm!  Trebby...  You are in His hands.  And He is your Father as well as your God.  Be still and know He is God and nothing in this world is for certain.  NOTHING!  My God, your God is bigger than any situation this world or any of the people can throw at us.  He is our Watchman and Guardian.  Prayin for ya.  I have seen new fruits and I have also found the new responsibilities that lie behind them.  I\'m in charge of more as far as what God has sent me...  That\'s frightening!  Especially for someone who all her life has prayed for a simple life.    When I was born, my mother went through 81 hours of labor.  My dad went through one.  Mom and Dad didn\'t tell me this until a couple of years ago, but they have always said they knew before I was born that I was called by God to do something for His kingdom.  And since I can remember, I knew I WAS called...  And I prayed for a simple life instead.  My dream was just to be a wife, a mother, and live on a farm with no other responsibilities or demands on my life.  That\'s all.  Just me and my family living out our lives.  What God showed me later on scared me!!!!  Yet, there was always this desire, this hidden and supressed urge to see where God was taking me or what adventure He had planned...  Hmmm...  What if?  What if I followed God?  Where would I go?  Who would I meet?  What would I see?  Miracles?  Angels?  Demons, even?  Who knows?  But I was finally ready to find out.    I still have to fight off the want to pray for only the simple life.  I guess, deep down, I know I\'m too complicated of a person to be satisfied with anything less than complication!!!  Ha, ha!  I want the unexpected, I want the unseen...  I want the adventure of not knowing my next step.  It\'s scary.  But the reality of being scared of the unknown is nothing.  God already knows!!!!!  And He\'s revealing His secrets to me gradually...  (Sometimes TOO gradually!)  And I know my hubby-to-be is going to have to be just as adventurous as I am!!!!!!  (Wink, wink!)  God is good, y\'all!!!!   '),(4040,2607577,'21,june,2004','2004-06-21 00:00:00','    Hello,  It has been some time since I last blogged and a lot has been going on in my life. My Wife and I have been separated since October 27, and I  had hoped that restoration would take place, those dreams took a hit this last weekend when I was served Divorce papers. I have been very sad since receiving that news, but I have no one to blame but myself. I was a very self centered person seeking first the kingdom of \"Trebby\" I have been unfaithful, dishonest and a terrible example of a husband. I have had so much junk in my life as a child, that unfortunately I just never dealt with that and it has effected every other aspect of my life. The mistakes I have made are issues of character and discretion.  God,  I am so immature please grow me in your Love and Wisdom. I seek your will and your vision not mine. I for the longest time have viewed myself as failing, utterly unlovely and worthless. These things have conditioned me to put forth a front of fasle pride growing accustomed to the constant barrage from the enemy, and my own accusing thoughts. That I hardly no what is like to live without a sense of sin, failure and rejection clouding my heart.  Here is the good news! My beloved Lord does not see me this same way.  He has allowed me to lift the Vail that has covered my heart and get a glimpse of the beauty that lies within.  I have for so long could only imagine God being angry with me, who was constantly holding me in judgment. The judgment was coming from me, I have been seeking the counsel of Men of God and through the word of God and believing it. I have begun to see myself as God would see me-passionately loved and forgiven. This is the only way to feel the enjoyment that God has in me.  Trebby   '),(4041,2607577,'21,june,2004','2004-06-21 00:00:00','    One of my girls was baptized this weekend. Man, I cried like a baby!!!!  And I  was glad I did.  She needed someone who really, genuinely cared that she decided to live for Jesus and show the world her decision meant something to her.  Now?  The hard part!!!  Living it!  Ha, ha!  She has had some major stress demons coming after her since that day.  In fact, the entire house has been a ball of tightly strung stress-lights on a tree that\'s lit by pure rage!!!!!  And I love working there!!!!!  Some of you may think I\'m lying, but I totally know that I\'m meant to be there, I\'m meant to work with those girls.  And I\'m meant to show them what they haven\'t seen anywhere else.  JESUS!!!!!  Praise reports galore!   1.  Got a car! 2.  Have an interview tomorrow with Flying J! 3.  Losing weight! 4.  The women are coming in from the rain and are ready to start talkin!  Aw, guys.  There is so much I could be complaining about.  Way too much!  But there is even more to be thankful about.  God is confirming, steadying, and strengthening me.  And this is all with me living in a glass house.  Nothing I have done or am doing is hidden.  SCARY!!!!  My mind is still a bit fuzzy from lack of sleep this weekend.  Work has come to rely on me and I love that fact.  I\'m just not used to NOT sleeping!  :)  Pray for me, y\'all.  If I get on at Flying J, then I will be able to relax more with the whole financial side of my life.    I\'d been praying and asking God when I was going to see the physical fruits of my laborings.  He\'s faithful, guys.  He\'s faithful!  The fruit is rollin in!   '),(4042,2607577,'21,june,2004','2004-06-21 00:00:00','    Today was amazing!!!!!  These past couple of days have been amazing!!!!  And God is AMAZING!!!!  This morning, at work, we had a horrible time with the girls and Satan was at work trying to corrupt the girls, cause trouble, and literally kill them if he could.  There was so much stress and oppression in that house...  When I took the girls to school later on, you could feel the oppression still in the air like a thick smoke!  God, there is so much there.  And the girls are searching.  They have seen in me something they desperately want for themselves.  In fact, they\'ve asked me how I can be so calm when all of this is going on!!!  I then get to tell them how I know Someone Who loves me more than anything in this world and gives me peace beyond even my understanding!  One of the girls, an adorably stubborn creature, has honestly been trying to let God start taking over some of the horrible areas in her life.  AND SHE IS DOING IT!!!!  Praise God!!!!!  She has requested, along with a couple of the others, that I see if I can work on Wed. and Sun. nights so I can take them to church!!!  That just blesses my socks off!!!!  Another amazing thing which happened today: I got to sit down and talk to a girl I\'ve been praying for for quite some time now.  I hadn\'t met her, but her parents are good friends of my family\'s and they had asked me to pray for her.  Well, today she was ready to talk!!!!  And boy did she!!!  She just let everything out and spilled her whole life into my lap!  Thank you, Jesus!  She let every single hurt, every injury be seen by the light and THEN, THEN she asked me for help!!!!!!!!    Before I went to meet her today, I was praying so hard.  \"God, please!  Don\'t let this be a single word of \'Anna.\'  Let this be Your words!  It\'s You she needs anyway, not me.  Talk through me!\"  He did, guys!!!  He did!!!  I don\'t even remember most of what I said, but I do remember sitting outside of myself, out of my head, and watching the transformation on her face as she listened to what God had to say!!!!!!!!  Oh, guys!  That transformation was miraculous!  That\'s the only way to dexcribe it!  It was a miracle of God!  See, she has been diagnosed with a social anxiety disorder and DOES NOT meet new people.  She met with me today and saw GOD!!!!!!  HALLELUJAH!!!  Glory to my God!!!!!:)  I\'m absolutely giddy with what happened!!!    I\'ve promised to keep myself accountable, in good things and in bad, so I now need to speak with Pastor Eric about all this since Rob isn\'t here.  But, I just had to share this with you all!!!!!  What a weird God we serve!!!  Good, but dang weird!!!!  Ha, ha!   '),(4043,2607577,'21,june,2004','2004-06-21 00:00:00','    A lot has happened since I blogged last.  A WHOLE LOT!    When I said in that last blog that I was being tested and tried with this giant fork in the road, I was right!  I had 3 different opportunities to choose the settled life.  3!!!!!  That\'s quite a few!  And all in one week!  I was being tempted with the life of the settled, not the life of the chosen.  Thanks to Sandy listening to God, I know that the desires of my heart will be fulfilled someday with the husband, family, and home.  But I also know, now, that that\'s not all!  What I chose at this testing time was the road covered in jungle roots!!!!!  And tigers laying in waiting!!!!:)  I chose the harder, unseen path.  I\'m ready to be radical.  I\'m ready to be a spitfire for God!!  And I\'m ready to WAIT...  For my husband, family, and home.  Ok, to be totally and completely honest with everyone on here, I was tempted with three different men.  (Yes, being married is a huge temptation.)  And all three of them represented a different kind of settling.  Each man was pursuing me!!!!!!  Man, I haven\'t been pursued by a man...  Not ever, really.  And then I was bombarded with three men in a week!  If that isn\'t Satan, himself, I don\'t know what is!  And all three men wanted to take care of me.  They could see the huge call on my life and could also see how vulnerable I am to... to attack and pain.  I guess I have that \"deer in the headlights\" look!:)  But, and this is amazing, I was able to tell them, all of them, that I need more.  I need more than just protection and love.  I need a challenge!  I need the love of my life to be an inspiration to find more God!!!!!  He needs to be my spiritual leader, not just a safety net.  That\'s what the three men were: a safety net.    God let me be tempted.  He did!  He wanted me to see for myself what I wanted.  He also wanted me to be strong enough to choose it for myself.    Did I pass the test, God?  I think so, though not without a few more bumps and bruises.  That\'s ok.  The more I suffer for God, the more I realize what He suffered for me.  And I ain\'t seen nothing yet!!!!!!!!:)   '),(4044,2607577,'21,june,2004','2004-06-21 00:00:00','    Ok I just wanted to share something that has been building in my heart today!  Thanks Rob for helping me see that it is ok to be where I am at! (to be content)  I have this huge ball of joy right now in my heart!  On my way to work (yes I am writing this while I am working..lol) I just prayed to God and began to thank him for everything he is doing in my life.  I told God today if I stay exactly where I am for the rest of my life I could still find Joy in everyday living.  I know in my heart that I can!!  I don\'t care if I am doing what it is I am doing because either way I am serving the lord.  Who cares if its by leading millions to Christ or just loving the unlovely.  I guess what I am trying to say it does not matter to me what my purpose is.  Just knowing I have one is good enough.  Right now MY purpose is to serve my Lord in whatever way he wants me to.  It is not up to me to decide if that\'s good enough or not.  I just want to encourage you if you are not sure what it is you are going to do.  If you are unsure what God has in store for you!  That is ok!!  Actually that\'s even better!  It gives you an opportunity to praise him and thank him everyday for what is about to happen.  Its an opportunity to show him my faith.  PRAISE GOD for the opportunities and what he has in store for me.  Even if I am exactly doing what I am suppose to be doing praise GOD for that too....  It just not important to me what I am doing, what is important to me is serving GOD!!  I am content with that!!   '),(4045,2607577,'21,june,2004','2004-06-21 00:00:00','    Well, Kristie.  You know, in your head, what is truth.  But, it is in your heart that your mind is re-affirmed.  Keep letting your heart breath in that love God keeps sending down to you.  Especially now.  Satan wants you to lose all want for order in your life!  He wants that feeling of, \"Why not just let everything fall since it\'s just going to anyway?\"  Only thing is, he\'s a liar.  HE\'S A LIAR, KRISTIE!  You are the RIGHTEOUSNESS of God through Christ Jesus.  That means you are the King\'s Kid.  Duh, you\'re going to fall.  Duh!  But, guess who\'s waiting to catch you, patch you up, and send you off with a big \'ole hug and kiss?  Your Father.  He loves doing that stuff.  Sure, He doesn\'t want you to get hurt, but He loves babying you when you do get hurt, and then nurse you back to health.    Anyway, sorry I got a little preachy.  Seems to be a running thing with me, lately.:)  I haven\'t written in a while.   I\'ve been quiet.  And I think I\'m waiting on something.  Something is about to happen, a decision is yet to be made, and I will know which way to turn at this fork in the road.  Ok, that sounds a little prosaic.  I\'ll explain.  I\'ve been given a choice, here.  I\'m at a point where there are two options: 1. A life completely safe with every comfort of home, family, love, and God, or 2. To become radical, unsure of my next step, and a spitfire for God.  My problem?  I\'ve always wanted option 1.  In fact, when I was little, that\'s what I prayed for.  I wanted the husband making enough money for me to be able to explore my artistic endeavors and still be able to send my kids to private school or home school them.  I wanted the house of my dreams my husband and I would build together.  I wanted the family surrounding me and needing me.  I wanted the doting, cherishing husband whose purpose in life was to be my compliment and provider/leader.  And I wanted to have the life with God that was...  Settled.  That\'s all I\'ve ever wanted, to be settled.    My family has moved several times growing up and I have never felt at \"home\" in any of the houses we lived in.  When I did find \"home,\" it was in church with God wrapping around me.  And now?  He\'s asked me to choose for myself what I want now.  AGAIN!!!!!  I\'ve been at this fork in the road before.  Only, I chose a husband that wasn\'t adoring, no home of our own, no financial stability, no family...  And that road ended abruptly.  Am I now at that SAME road?  Lord, help.  You know the desires of my heart.  And I know which choice to make.  I WANT to be settled, but I NEED to be on fire.  I need to be consumed with this, this passion for more.  I hate the song, but Stacie Orrico\'s \"There\'s Gotta Be More To Life\" pretty much describes it.  Someone recently asked me what makes me tick, what drives me.  They didn\'t believe me when I told them it\'s when I\'m out on the streets and one of the baby girls needs to talk.  Being there for them just makes my heart jump into gear!  That\'s how I function, that\'s how my heart breaths.  And when I\'m asked to choose between that and what I want out of life?  That\'s unfair.  God, I know what You\'re trying to tell me, here.  There\'s a reason that \"Seek Ye First\" has been my theme song since I can remember.  I am seeking You first!  I AM NOT choosing the settled, though I long for it.  I know You will ground me, root me.  And I know that a husband, family, home, and financial security will never do it for me.  Only You can do that.  But...  When do those other things come along?  How long, God, until I begin to see the PHYSICAL fruit from the SPIRITUAL fruit?    I sound awful whiny today.  Sorry to anyone who reads this.  I have to write this, though.  That whole transparency thing, again.  Right, Rob? (WINK, WINK!)   '),(4046,2607577,'21,june,2004','2004-06-21 00:00:00','    Hello everyone!! It has been a while so I thought maybe it was time to let some more of what is in me out.  I don\'t think about writing on the blog because I personally feel I am not all that important.  I don\'t mean that to put myself down.  I am just saying who has the time to read my stuff and really why do YOU want to.  Then I thought hey this might be a good way to START getting to know me.  I can express more of who I am and what I am feeling by writing it instead of being put on the spot.   I went to Acquire The Fire this weekend.  I really believe the enemy was trying to make this time not very enjoyable at all.  I left Friday basically with little sleep because I worked the night before.  I am already cranky because lack of sleep put people in that equation it probably was not going to be good from the get go.  I realized something a few weeks ago.  I have been struggling with rejection issues.  I mean I KNOW in my head that nobody is making me feel unwanted or unloved.  It is completely ALL me!  So I started to go to ATF with all this and when you think you have disappointed someone you are close to you tend to feel even worse if you know what I mean.  I am not sure if this person felt disappointed or even aggravated.  However I was feeling bad because I was being sarcastic and I was getting aggravated myself.  This was all just in the first 5 hours...lol.. The Good part was there was one time on Saturday I was just sick of it.  I was sick of feeling bad, sick of feeling the way I was feeling(guilty).  So I just prayed to God and told him to take away all this junk that is on me.  It was so AWESOME how he God just instantly took away the junk and I was filled with so much joy! From then on out I was good..   It was an awesome experience and I am very blessed I got to share it with the youth. Well going back to the real me!! I said in an earlier blog God has been showing me things in my past that I have not remembered.  I am still trying to figure out how to deal with that stuff. (have not yet) The newest thing God showed me was this rejection I was feeling.  I mean I know in my head that I am loved and even appreciated.  However these feelings come in from nowhere and are like telling me that I am a fake that I am not loved.  Its crazy and I now they are all lies.  For some reason though its been happening A LOT more than usual.  I don\'t know why exactly. It has been a constant battle lately and I feel overwhelmed with thoughts.  Its come to the point that I have lost the desire to do things.  I have lost the desire to do homework and even to pay bills.  I use to be so on top of that stuff! Its just thoughts that why try I am going to fail anyway.  SO I was thinking maybe I am not where I am suppose to be!  Still not completely sure I am doing what I need to be doing.  Am I alone in this??  The ONLY cool thing is I have not lost my desire to know God to be in his presence.  I want to know him more I want to be around people who know him.  I want to talk about him!!  I know in my heart that this is just one of the many things I am going to have to go through in my life time.  I guess in some ways I am lost and just trying to find my way! To be honest I am sick of it!! I hate the feeling of being lost!!  I hate feelings all together.  IN my head I am where I am suppose to be....But the enemy strikes me right when he sees an opening.  I think I will stop for now I could go on forever.  Keep me in your prayers please.  God is Awesome and I know that Gods way is the better way.  That alone in itself gives me JOY!!    Kristie   '),(4047,2607577,'21,june,2004','2004-06-21 00:00:00','    Geesh!  It\'s been forever!  Finally have a few minutes to blog, though not many.    So much has happened...  And I can\'t seem to think straight right now, so sorry, guys.  You won\'t be gettin the whole \"Anna\'s Story Time!\"  haha  I do want to ask you all to be prayin for me right now.  Hard stuff.  And I\'ve had to give two weeks\' notice to the Flying J.  Too much stress and not enough sleep or down time from there.  Be prayin that I get another job soon.  BILLS!!!!!  OR, you guys could pray what Rob has been sayin lately...  \"Miracles happen in finances!\"  Dear Lord!  Let the miracle come quick, then!  There\'s so much I have to do, to pay for, bills...    God, I\'m under so much pressure.  Help me to STOP and rest in You.  Lord, I pray for those quiet times with You I CRAVE and need.  Since the whole \"2-job\" thing, I haven\'t been able to spend time with You like I want and I can DEFINITELY tell.  I know You are still there and still moving like You always do.  I know that miracles and breakthroughs are happening for me.  I just need time to rest, Father.    God, please.  Heal me.  You know where and how.  Just please, heal me.  And work in my specific situation so that I can be freed to work for You.  Not a stumbling block or hindrance, God, but a blessing and encouragement, a new platform.  You know what I\'m talking about.  Help, God.   '),(4048,2607577,'21,june,2004','2004-06-21 00:00:00','    Hey guys, I wanted to clear something up that might not be so clear.  The only people that can blog on this site are those who have been invited (ie, the interns and myself).  This DOES NOT mean that we are the only ones who visit this site.  We have over 100 hits on our site every day.  These hits are from people from all over the world, and from many churches within our region.  Please be very careful what you put on this blog, knowing that what you put on here will be viewed by 100\'s of people every week.  Please know that there hasn\'t been anything that I would deem \"inappropriate\".  There have just been some very personal issues that some people have shared, and they just thought they were sharing them with the interns who have the ability to write in the blog, and that is not true.  Continue blogging, and being real.  If your going to boast, boast in these 2 things:  1.  Knowing Jesus, and 2. Your weakness!  Just realize that others will read this!  I love you guys, you are doing great. Keep doin\' whatcha do!   Poppa Rob   '),(4049,2607577,'21,june,2004','2004-06-21 00:00:00','    Hello again!! I just got back from the chapel at work.  On my break I go down there and pray!  Its really nice and quiet especially at night. I was thinking back to when I was a kid and even a teenager. I remember that I was one of those kids who were always wondering about the other kids and why they would act they way they do.  I could always go past the service and knew that there was a reason. Even if I did not understand but I knew it was there. I have always had a big heart.  I have always had a lot of love I have wanted to share but really usually loved the wrong people for the wrong reasons.  Now ever since I have been saved my love has grown so much as well.  I have been getting to the point that my chest just wants to burst.  Sometimes I have a hard time expressing that love to people and I can go to one extreme to the next.  God lead me to some scripture tonight in Romans 12 and 13.  I don\'t want to type out the whole thing because it would be way to long.  There are some that I would like to share.  Like Romans 12:9-13 Love must be sincere. Hate what is evil: cling to what is good.  Be devoted to one another in brotherly love.  Honor one another ABOVE yourselves. Never be lacking in zeal, but keep your spiritual fervor, serving the lord. Be joyful in hope, patient in affliction, faithful in prayer.  Share with God\'s people who are in need. Practice hospitality.  God is so awesome and I want everyone to know him.  I want to know him so more deeply than what I do. I so want to do the things he mentions because it will honor his name.  I want to be the friend, wife, mother, leader that really HONORS other people above myself.  I want all that is in the above scriptures. So for those who want to know me and know my heart.  There it is.... it just wants to love all people saved and unsaved.  I want to find the lost and show them that there is someone out there that wants to protect them and love them.  I want so much to walk in LOVE at all times. I desire to be used by God in whatever way he wants.  I have so much to learn and I know I cannot do anything without him. So know that anything I say anything I do..when I breath, when I sleep, when I talk, and when I LOVE its all him. ITS ALL GOD!! I wouldn\'t want it any other way.   '),(4050,2607577,'21,june,2004','2004-06-21 00:00:00','    Hi everybody!!  The past couple of days have been awesome.  I have found my joy and really its nice.  Then again I would say I always had it but decided to start living it. Today was a testing day and actually I have a lot of testing days.  Its funny cause after church today I realized something pretty awesome!! Well let me tell you first my hubby as most of you know isn\'t saved.  He is an awesome man and I love him deeply.  The enemy uses him A LOT to try to steel my joy.  A lot of times it works, but today was different.  We had a disagreement about some stuff.  After I hung up the phone I started to loose my joy.  Well as I was driving back to church I was talking to my best friend..(Jesus) He was sitting in my passenger seat as he usually does.  Well I just started to tell him \"Man I know my hubby does not know what he does...He does not understand the enemy is using him to try to take away my joy.\" In that moment I felt so sorry for him.  He is so missing out on God right now. I CHOSE not to let the enemy take my joy! (it was a good feeling to..hehe) Guys I usually have always got mad at him and his attitude.  To be honest though I have a little relief to know that its not all him.  He don\'t understand what is going on not completely.  I can\'t wait for that day that he invites Jesus into his heart.   Thank you Jesus for that day!!  Every trial you face is an opportunity to let him work in your life for abundance. If you invite him into your circumstances he will keep his promise to work everything together for your good Romans 8:28  Now on a completely different note..haha I am so thankful for the youth ministry and the people God has placed in my life.  I really love all the adults that help out.  They are really awesome! Each one of them have something that I can respect and enjoy very much.  I am so excited to see what God has in store for us.  I am even more excited when I think about doing ministry with these people.  We are all human and make mistakes.  There are times that we are going to have a bad day. I just love knowing I can consider them family.  Its a process that I am willing to be patient about because I know the end result. God thank you so much for the many blessings I receive each day! I just want to pray special blessings over Chris, Jessica, Jon, Exie, Nathan, Sandi, Tony, Cliff, and anyone else that I might have forgot to mention. YOU GUYS ROCK!!   '),(4051,2607577,'21,june,2004','2004-06-21 00:00:00','    Good Morning,  I just felt the need to vent. This is one of those times, when, I\'m just down in the dumps and feeling sorry for myself. \"I know get over it\" I agree, but with that being said, I\'m still human and feel the ups and downs of life. I just dread what faces me with my impending divorce and all that junk that goes with it. You tell yourself that this can be relatively painless, but what a bunch of crap that is people are making life changing decisions, and rethinking their future plans. I had no intention of ever being divorced for a second time in my life, but that is exactly where I\'m at. I shoulder a huge part of that responsibility, and will have to live with that. I always viewed people that have had multiple divorces as real losers, and vowed that \"I\" would never be divorced for a second time, spoke to soon. This is what the Holy Spirit has been revealing to me over the last Six months.   I could not have loved anyone intimately, and had any relationship that would last beyond the physical attraction that has been my problem since I was 13 years old. I have had a dozens of relationships that were nothing more than physical or at best just superficial, on the surface things. Now my wife Cindy, did her best to make me go deeper than surface stuff, but I couldn\'t on my own. Like Rob said the other night at house church, I never had an example of love. Anyway, Jesus has revealed that I must first Love him,and seek after him passionately in this experience love will be revealed through scripture, truth, and a deeper relationship with Jesus. Jesus is my Lover, and maybe someday, I can love a woman the way she deserves to be loved. \"you know like Jesus loves\"  I first must discover my Jesus, and who knows he may be my lover in every sense of the word. Jesus is love right?  PS:  Sarah,  You are in my prayers.   Trebby   '),(4052,2607577,'21,june,2004','2004-06-21 00:00:00','    Do you just have those days when you are feeling sorry for yourself.  When you see Awesome things happening to other people and you are happy for them but you wonder why its not happening for you.  Maybe you just haven\'t been patient enough or maybe its not Gods will..   I am so sick of it!  I am so sick of the enemy trying to beat me down. I have been really sick the past couple of days(physically).  I am not one that gets sick too often.  I might get a headache or something but I am not usually so sick that it hurts to move.  Being sick yesterday was one of the worse days of my life.  My son who has no fear found a way to get out of our yard.  We have a lock on it thinking that would stop him.  But together him and the dog manage to find a way out.  There was only a 3 to 5 minute time span from when he went outside to when I decided to check on him.  HE was GONE!! Ok... You know I was looking sick anyway not wanting to move.  I got my sweatshirt on and went outside looking for him.  Lets just say a half an hour later I found him.  I was good until I got him into the house.  I LOST IT!  I was crying uncontrollable because I thought he could not get out of the yard and believed someone took him.  I was crying and I couldn\'t control it all the sadness just was being pushed out of me.  I really don\'t know how people who have missing kids go through it without wanting to die.  For those few minutes I was hurting so bad I wanted to die.  Even though I had found him and he was right beside me.  Well later that day I had to take him to the walk in clinic because he had poison oak all over his face and it was swollen by his eyes.  So he needed a shot and some medicine.  So I decided to go to work anyway which I am not sure why cause I came home an hour after I got there cause I was feeling bad again.  Well if that was not bad enough I was feeling down anyway I got calls from friends etc.. On how great they are doing.  Lets be honest I love them and I am truly happy they are doing great, but today was not the day I wanted to hear about it.  I did hear about it though and told them I was happy and tried really hard to not tell them about poor old me.  I guess the whole point of me getting on here and telling you guys this stuff is that there are going to be days that suck!!  There is no getting around them! However I realized something when I had that terrible feeling that I lost my son or that someone took him.  I realized that is how GOD feels about every single one of us.  That is how he feels when the enemy is trying to steel our soul and joy.  Heck I am the first one to tell you that with yesterday, today I was loosing joy because of my circumstances that my life is in.  Trying to figure out what I am going to do about that and this.  In the End I just let the enemy take yesterday away from me!  The day is over and gone!!  He is NOT going to get today!!  If the enemy somehow wiggled his way in and Got a day from you.  Let it go and move on.  Don\'t let him get another one!  God Please forgive me for feeling sorry for myself and being selfish. I know what you have in store for me is better than what I can even imagine. I know you want to give me the desires of my heart!  It says Delight yourself in the Lord and he will give you the desires of your heart. Psalm 37:4...  I am ready to continue to delight myself in you! No Worries! In Jesus Name...Amen   '),(4053,2607577,'21,june,2004','2004-06-21 00:00:00','    Hello,  Anna, it was good to hear about your mom, and how important she is to you. I have such a weird life can any of you relate? I am Forty years old, and feel as if a my life is just beginning in a lot of ways. I am preparing to move into a home with a bunch of people, that I have never even camped out with. I have lived on my own, since I was Nineteen so this is a huge step for me. Then at the same time, I am preparing for a divorce and moving out of a home that is so large you could get lost. I have a giddy excitement about this whole adventure, and I know it is a God thing. He has placed me in this position with these people for a purpose, and I want to be obedient to my Father. I look at all this stuff as a season in my life, but not the people, who I have come to know over the last year. We will be friends for life. In the midst of these incredible personal trials, God, is opening doors and bringing forth the truth about who I am in him. I am weak and struggle daily and that is a good thing for God\'s power is made perfect in my weakness.   I look forward to the experiences that await me in the future. I look forward to entering into a deeper more intimate relationship with God over the next year. I am sure all of us will be forever changed by the new wonderful experience that is awaiting all of us in the \"Big Blue House\".  Love ya,  Trebby   '),(4054,2607577,'21,june,2004','2004-06-21 00:00:00','    I HAD to get on here and write about my awesome Momma and how much she\'s meant and done for me.  You know, there is a country song where one of the lyrics says that her momma is her biggest fan...  Well, that totally applies.    When I made the decision to go to St. Louis and see what God would do there, Momma was scared to death her little girl would never come back.  (Thank God I did!!!  She\'d have killed me!!!lol)  Not only did she take me out to St. Louis herself, but she also was my biggest financial supporter.  Out of her own check, she gave me $100 a month!!!!  Man!!!  She believes in me so much, she gave up what would have been saving and spending money and invested it into her not-so-reliable daughter.  Know why?  Momma knows what God told her.    Before I was born, He told her that I was meant for Him, to do a work for Him that Momma hadn\'t been able to do....     Mom told me during one of my home breaks this: she was proud to be able to help me because I was doing what she couldn\'t.  I still cry at that thought.  Momma was called into ministry but wasn\'t able to do that in the way she THOUGHT after she was married and had children.  Her ministry?  Momma may not realize this, was to prepare me for the next step in His extremely confusing and awesome plan.  And she did a heck of a job, if I do say so myself.  Momma is still my biggest supporter today, and my best friend.  Mom keeps reminding me of Africa, of my perfect husband, and the girls I am and will be helping.  And she does this without any words.  She does this with her hugs and smooches, playing with my hair when I put my head in her lap, and in the shrieks of rebuke!!!!!lol  I love my mom.  And not just for what she\'s been for me, but what she hasn\'t been.  She hasn\'t been a doubter.  EVER!!!!!  Thanks, Momma.   '),(4055,2607577,'21,june,2004','2004-06-21 00:00:00','    I know what you mean, Trebby, about the \"Mask.\"  I don\'t know how many sucky poems I\'ve written about masks.  Especially as a teenager, I discovered that the mask was masking me.  I was covered, hidden, and couldn\'t find out how to take it off and live as ME, as Anna.    When we are all young, or young in Christ, we put on this shell, this protective covering we invent for ourselves.  We think this is who we want to be and this is who we SHOULD be.  This is what the world wants.  And this is how I\'ll make it, how I\'ll keep safe.  DECEPTION!!!!!!!  I see now what that mask really is.  It is a mud bath!!!!  Explanation of MUD BATH:  A mud bath is just what it sounds like.  It\'s mud.  It\'s used to protect us from the harsh environment of the world.  For instance, elephants jump right into the mud and revel in the ewy-gooyness of the mud to protect them from the brutal sun and the annoying bugs.  Well, God gave them the protection they needed already!  Their thick covering of skin.  Have you ever seen an elephant\'s skin?  That thing is TOUGH!!!  And thick.  Yet, they are luxuriating in the mud as an added protection.  Problem?  The mud is filled with the very dung of the elephants as well as other animals.  They are wallering, as Momma says, in their own crap.  And they do that to protect themselves!!!!  Anyone see my point, here?  We waller around in our own dung to protect ourselves.  We cover what is us with the dung of us and the dung of the world as an additional shield.  What we don\'t get?  The dung, the mud keeps us from developing our skin like it should be.  The sun is supposed to burn it, the bugs are supposed to annoy it.  We are not supposed to LIKE what this world is doing to us!  And we don\'t need another shield to protect us.  We have One already: GOD!!!  He is our shield, our strength, our protection.  If we\'re rolling around in our own crap, He gets deprived of His job, His asked-for duty: to protect His children.    If you all don\'t get burned or annoyed by the world, your skin will never get used to what it feels like.  With the mud on there, you will be indifferent to the reality of this EVIL world and what it wants to do to us.  The mud gets into even our eyes until we don\'t know where we are going, which could be we\'re going straight to our death.  Oooo!!!  Parables!  Like em lots!  Don\'t think I\'m preachin, y\'all, unless you remember I\'m preachin to myself, here!  I still grab my dung and slop it on sometimes.  I\'ve been known to make an entire self-pity manure bath for myself and stay there for days.  Good thing is, I know what I\'m bathing in and CAN\'T STAND the conviction that comes when I waller again.  Be encouraged.  Know that God is the best manure shoveler and remover I know!!!  HAhaha!!!!  God is cool!!!   '),(4056,2607577,'21,june,2004','2004-06-21 00:00:00','    Greetings,  It is really cool to read all about what God is doing in all your lives. It is good to know such awesome people. I have been working on this thing called fear...I have let this thing guide my destiny for way to long. Christ tells us not to fear, we are free to be vulnerable. Being vulnerable means being REAL. This has long been my problem, you see, I did not want to risk rejection by laying  my life out before others, and saying this is what I am and what I am not. I have so desperately wanted to be accepted by others, so I created this facade that mask the real me and then found myself hidden behind a barrier of secrecy and hypocrisy, leading to constant compromise. Here is where God has been working, he has told me that being accepted by him is more important than being accepted by others. If my faith is in God, I don\'t need to fear being REAL. The stronger my faith is in God, the more I am driven to be REAL. I have struggled with the fact that God has accepted me, but I know now he has and I can face the rejection by others. I can now admit my struggles, my failures, my sins, and my emotional weaknesses...I don\'t have to pretend, that I am always on top of the world. It is OK to swallow that ugly thing called Pride and just be REAL.  Dear Heavenly Father,  I am a broken piece of pottery put the me back together using your own special Love. I am a sinner, who struggles with many things. I know that your Love reaches all the universe and covers all, but please let me have a personal relationship with you this very day.  Treb   '),(4057,2607577,'21,june,2004','2004-06-21 00:00:00','    Rob and Treb, thankfully, sat down with me the other night to kinda counsel me through some of my crap.  While they were advising/patienly listening to me, some new ideas popped into my head.  Well, they aren\'t \"new\" as far as revelations the world has had, but new as far as \"new to me\".  And even though I knew these things in my head, my heart\'s been ignoring what it could not deal with.  (For a while now.)  I\'ve been in a working mentality for too long.  And I\'ve purposely MADE myself be busy.  You don\'t have to think when your mind is busy with details.  True, the work is worthwhile and for God, but it hasn\'t left me any time for what Rob calls \'gazing.\'  Ugh!  I hate gushy words like that.  I\'m so not a froo-froo girl.:)  BUT!!!!!  I haven\'t simply STOPPED and listened for a long time.    I\'m afraid of what I\'ll hear when I do stop, finally, to listen.  I\'m sure I\'ll hear the voice of a disappointed Father.  He has to be disappointed.  I AM!!!  So, how much more is my Father, Who has SO MUCH MORE for me than I can see or imagine?    WHY do I get back into this cycle?  WHY?!?!?!?!?!?!?!?!    I settle.  For very little.  And not enough.    There\'s more than I\'ve seen.  THERE\'S GOT TO BE!!!!!  Otherwise, what in the heck am I living for?  WHAT??!?!?!?!?!?!?!  The things God\'s shown me He wants for me...  Way too lofty for my puny, miserable self to accomplish.  And I have to convince my not-so-open mind to accept that fact.    I haven\'t been living a life as a child of God.  I\'ve been living it as only the SERVANT of God.  If I\'m serving, somehow I\'ll redeem myself for all the screw-ups I\'ve made recently.  That\'s how I work.  I WORK things off.  I don\'t accept them as forgiven and forgotten.  Y\'all are thinkin, \"She still hasn\'t gotten this!!!\"  I know.  My head knows.  AND I SURE AS HECK know how to preach it to others.    What\'s going to happen in the girls\' home when I have to LIVE this principle out?  For them all to see?  They have to learn it from someone who LIVES this, not just preaches it.  God, please.  Don\'t allow me to be a person whose witness is of words only and not as a lifestyle.  I hated those people.  In fact, I despised them!  DWYSYWD...  Do what you said you would do...    I am forgiven.  And the sin has been forgotten.    There is, therefore, now no condemnation for those who are in Christ Jesus.    Know something?  ALL of the interns in St. Louis would randomly quote that to me throughout the day.  Funny, huh?  I\'ve lived a life in the shadow of condemnation and guilt.  That has defined me!  God, what\'s it going to take to allow me to live a life of PEACE?  And victory?  When am I going to be able to wake up in the morning and feel GOOD about who You made me to be?  When will I realize I am Your daughter?  And Your CHOSEN bride......  Lord, help.  Heal me.  In all the ways You know I need.  You know me.  And You still love me...  Why?  I\'m not looking for answers.  I already have them.  I just wanted to write all these things out so that they cannot torment me anymore.  GET IT OUT!  In the name of Jesus Christ, thoughts of condemnation, abandonment, and rejection are NO LONGER allowed to plague me!!!!!  And I WILL see You, Jesus.  I will.  I have to!!!  I need You so desperately!!!!!!!!!!!   '),(4058,2607577,'21,june,2004','2004-06-21 00:00:00','    Hey all,  Man I am reading something that to me seems like hits the spot!! This is what I want to teach people!! I believe there is a principle here that you and I can apply to any difficult situation.  Holding onto joy starts with a decision to believe the best.  Then we need to line up our mind and mouth with Gods word.  This keeps an open channel for the oil of Gods joy to flow.  No matter how bad your cirumstances look, God LOVES you.  Never let go of that truth.  keep it in your mind and keep it in your mouth.  If you\'re in the midst of a situation that\'s stealing your joy, begin praying.  GOD I don\'t like this, but for some reason I need this experience to fulfill your destiny in my life.  I believe it will work out for the good, even though it does not feel good right now.  I cast my cares upon you, and I ask you to carry me through this situation.  In Jesus name amen!! As believers, we are more than conquerors because we have Jesus walking with us through our problems.  Our worst day with Jesus will always be better than our best day without him!!   '),(4059,2607577,'21,june,2004','2004-06-21 00:00:00','    Hello,  The Lord put the word perseverance in my heart and then compelled me to blog about it. I think that James, described it better than anyone ever. We live in a world today that is so easy to just throw in the towel, when things aren\'t going the way we think they should be going. It can be home, work, school, Marriage, Ministry, etc. If the apostle Paul would have had this attitude where would Christianity be today? If the Disciples would have had this attitude how would they have spread the gospel and been such worthy ambassadors to boldly proclaim Jesus as Lord an Savior? The one common denominator that set these men apart from all others was their, I can do attitude. Unlike most of us today, who immediately succumb to the pressure. Let\'s face it, we have struggles everyday and temptations that can impact our lives in a positive or negative manner. Our character is formed from these times of great peril. God has promised that when we persevere the reward is great. I want to encourage and challenge all who read this to stand tall and press on through, when faced with the decision to flee or face it. I am uniquely qualified to write on this subject, because for most of my life I choose to flee, but NO more. I stand tall and choose to go head first into the abyss. The darkness does not scare me anymore for I posses the light of Christ.  Trebby   '),(4060,2607577,'21,june,2004','2004-06-21 00:00:00','    I want you to know that if pride is a sin, then I am one of the biggest sinners around, becuase I am so proud of the team that we have!!!!  Our team is the most dedicated, sacrificial, and committed group around.  I have been so blessed to see the radical growth in all of you.  The places that you will let God go in your hearts encourage me to keep pressin\' hard after Jesus.  Each and everyone of you are so special in many different ways.  I am honored to serve along side of you, and I want you all to know that I would die for you!  Thank you so much for submitting to imperfect authority, I am empowered by your love and devotion to the Dream Center.  Thank you all for making it such a joy to be in ministry.  I appreciate each one of you, and I love you all very desperately!  Continue to \"BE\" the example, the true representation of Christ on the earth!  Much Love, PR   '),(4061,2607577,'21,june,2004','2004-06-21 00:00:00','    I\'ve asked Rob to hold me accountable.  (PRAY FOR ME!)  And I sincerely hope he does.    Am I crazy, you ask?  Yep.  But I also want to be able to dream again and see visions from God.  I\'d lost my heart, as some of you may have heard on Sunday.  And I didn\'t know why I wasn\'t hearing anything new from God.  DUH!!!!  How in the world could He encourage my heart and move my heart if it wasn\'t even there?!?!?!  Doesn\'t make a whole lot of sense, does it?  How am I supposed to change that?  Well, going to L.A. DEFINITELY helped.  I cried for the lost, the hopeless.  I haven\'t cried for them in I don\'t know how long.    When I came back from St. Louis, I was ready to save the world and start a girls\' home.  In that order!  When none of the above happened, I lost sight of what God wanted me to see.  He needed me to slow down and listen for a while.  At the SLDC, I was moving and working CONSTANTLY!  Sure, I grew leaps and bounds, but I couldn\'t hardly stand to sit still and wait on God.  So, this was my waiting time.    Stupid me, I misconstrued waiting as God just abandoning me.  God was only wanting me to chase after Him, to pursue Him.  He\'d done the initial work by waking up my heart to the dream of speaking His Words to the hopeless.  Now, He wanted to spend some intimate time with Him.  Boy, did I ever handle that wrong!  I\'ve gotten into so much trouble lately.  Not like you think.  I haven\'t robbed any convenience stores or shot up any coke.  Nothing like that.  But I have let my mind wander back to things it CERTAINLY did not need to revisit.    I let my mind think about the mystery behind the dark side rather than the mystery behind the light.  Man, I can be dumb!  I\'ve already figured out the dark side!  It\'s gonna end in fire and damnation!  (Can I say that on here?  Guess I did.)  The light?  That\'s something altogether separate.    The last time I had an actual vision of God or of His plan...  Too long ago.  And when I had that vision, it mystified me!  I wanted to know more and more about what I saw!!!!!    God, more!!!!  I\'m greedy, now, for more knowledge about You!  You\'ve heard my prayers asking for dreams and visions to return.  I know I was named \"Anna\" for a reason.  Anna was not only the GRANDMOTHER of Jesus, but she was also a prophetess.  You\'ve given me words before, bring it on!  If that\'s what You want, God, then please do it!  Just please start talking to me again.  I need to hear more from You.  Without Your Word, I\'ve fallen and nearly killed myself in the process!   '),(4062,2607577,'21,june,2004','2004-06-21 00:00:00','    Today was so awesome.  When we first started blitzing to get the kids to come over, Brooke, who I met two weeks ago and have seen twice now, ran up and threw herself at me.  She almost knocked me over.   I have no idea how many kids were there after we got started.  At first there were like 15 kids, but they just kept coming.  I\'ll have to get an exact count from Cassie before I overexaggerate in print.  It was so awesome, though.  The girls side of the tarp was full and the guys side was over half full.  And I couldn\'t believe how many adults were there from the neighborhood.  I can\'t wait until we have enough people doing adopt-a-block to expand into other neighborhoods like stonegate and Belmont.  There are so many areas in our communtiy that need to experience Christ\'s love instead of just hearing about it.  The church has been preaching Jesus instead of being Jesus to these people for too many years.  Quit asking what would Jesus do and just do it.   '),(4063,2607577,'21,june,2004','2004-06-21 00:00:00','    Hello,  I am still feeling the after shocks from our California trip.(No Pun intended)  I gave my testimony on Sunday and that was my frivolous attempt at describing what the trip meant to me and how it impacted my life. I have yet to find the words that could accurately describe how I feel or the impact it made on me. I hope, that I will never be able to accurately describe my trip. If words could reflect what is in my heart than by all means let them flow like a river. I do not desire an emotional experience alone, those come and go with the changing of seasons. How do I know it has been permanently etched into my very soul? Because, all I can say about the trip is that it was awesome.    Trebby   '),(4064,2607577,'21,june,2004','2004-06-21 00:00:00','    Man, this week at the Pastor\'s School in Los Angeles was incredible.  Today we gave testimonies and it was hard to even try to find words to describe adequately what is going on in my heart.  I think everyone who went feels that way too.  There is such a seriousness in my heart about what we are doing.  This is for real.  I have never been so passionate about anything.  I know this is easier for me than some of the other interns becuase Metro Kids is already up and running, and many of their dream ministries are still in the very earlies stages of being developed.  Tommy Barnett said while we were there that you can tell a vision is from God when you can let it go but it won\'t let go of you.  Anna, you will see it come to pass!!    At the very beginning of the conference, they did what they call the Parade of Ministries, which is where all of the DC ministries walk across the stage holding signs for their ministry.  This procession took an hour, just to walk across the stage; it was incredible.  Then was the parade of miracles where several of the ministries came back up with people in their programs giving their testimonies.  A phrase several of them said in giving their testimony was \"cancelled out.\"  Their parents, freinds, siblings, etc. had cancelled out on them.  Some of them cancelled out by comitting suicide, others by just leaving, and others by being physically there, but not really there.  Praise God, we will not cancel out on these people.  We won\'t give up on them, and we won\'t give up on ourselves.  The success of the dream center in LA comes through consitancy and relationships.  It has nothing to do with the programs or pastors, but the individuals who go to the neighborhood and knock on the doors, hold the kids, encourage those who are down, etc.  We must dream big and never cancel out.  There are so many divine appointments out there, people we need to encounter.  We must walk with open eyes.  So often we just don\'t see the need.  Trebby has said that he hasn\'t felt a specific calling, he just wants to meet needs, but that is the call.  To see a need and fill it.    I can\'t wait to see all the doors that God is going to open up and all the people\'s lives that are going to be impacted by the staff and interns of the Dream Center and the volunteers from different churches in the community.  There were so many great realtionships and connections started while we were in LA, and I know God is going to continue networking his people to accomplish his plans.   '),(4065,2607577,'21,june,2004','2004-06-21 00:00:00','    Man!  This week was incredible.  We went to the L.A. Dream Center and got our butts seriously kicked on some topics.  And, yes, I did too.  What\'s the deal now?  The passion that I gained from L.A. is ready to bust out of this pathetic shell commonly known as \"Anna\" and wants to rock this city.  You\'re all thinkin\', \'Well, isn\'t that why she went?\'  Yeah, I did go for that.  What\'s killin\' me is this: I can\'t do everything yet!!!!   It\'s the waiting game now, waiting for God\'s perfect timing and all that...  Ugh!  And there\'s more to this story than I\'ve told.  Before we even got on the plane, the attack started.  True, I was tired already and the enemy loves to jump right in when I have the least resistance.  Guys, this is some tough junk!  I\'m insufficient.  I\'m inadequate.  I\'m weak.  I\'m unstable.  And I just don\'t have what it takes.  Anything else I\'m missing?  I know, I know.  Being so wrong for this job is exactly what makes me RIGHT!  Because it\'s going to be ONLY God and not a darn thing of Anna.  Man, though, I\'d like to have some reassurance on this one.    My personal life is in such shambles right now, I don\'t feel like I DESERVE to be doing anything for my Father.  I\'m not anywhere close to where He\'s gonna need me to be for this task.    I\'m not ready, God.  And I don\'t know how to get ready.    Please, Father, take my ineffectiveness and over-ride it somehow.  Anyway You can, please!  Just do it.  I\'m so stinkin\' tired and not much good right now.  So, it\'s literally going to have to be COMPLETELY You and not a thing of me.  And that\'s scary.   '),(4066,2607577,'21,june,2004','2004-06-21 00:00:00','    Wow I just wrote a blog but somehow it messed up so I lost everything. I hope this time it works..lol 1 Corth 2:16 For who has known the mind of the Lord that he may instruct him?  But WE have the mind of Christ. Hmmmm I have the mind of Christ :) I would like to think so, and I am actually starting to believe it!  Confidence in who I am in Christ.  WOW that\'s a hard one for me.  I believe I can hear from God!  I believe all wisdom comes from my spirit which is from God. But I have been believing this lie from the enemy for so long that the person I want to be is not me.  Even though I know God loves me and I know Jesus died for me.  It still is not me!!  I knew people liked me and thought I was cool, however I thought if they really knew me then they would not like me!  Another lie!!  But God really knows me and loves me for me and that is good enough for him.  So why can\'t I take that and stand with confidence in who he made me to be.  Well guess what I am starting to!!  WOOOHOOO!!  I decided not to believe that stupid lie that the enemy wants me to believe.  I believe that I am an awesome person who loves God and seeks his face every chance I get!  I am a person who walks in love!  I am a person who helps others, encourages them and builds them up.  I am living my life to Glorify him and to share the word with anyone willing to receive it.  I am one of his favorites and with him I can do anything, without him I am nothing.  So everything that I am God I give you thanks.  Thanks so much because without your love and grace I would not have a reason to live.  Thank you JESUS!! I don\'t know if most of you know but I work at Methodist hospital.  I work 2-12hour shifts a week. (7p to 7am) Well with the drive it turns into 16 hour days.  Then I am trying to catch up on sleep the next day or so.  Needless to say sleep is something I lack.  ANYhoo I decided to ask my boss if I could maybe work three days a week 8 hour days.. So I only would have to work 7p to 3:30am.  Which for me would be awesome!  I would get home in enough time for my hubby to go to work and we would not need a babysitter.  PRAISE GOD she agreed!  It saves me money on babysitting which I need all the money I can get..lol.. I believe it will help me out not to be SOOO tired!  So I just wanted to share some of my good news with my family.  Without your love and kindness I would not have ever been able to experience God!  Its people like you that helps make it worth wild!  Let me rephrase that its people like us!!   '),(4067,2607577,'21,june,2004','2004-06-21 00:00:00','    Here is a new look for our blog.  It has the comments available at the end of each blog.  Please take some time to comment, encourage, and exhort each other in love.   '),(4068,2607577,'21,june,2004','2004-06-21 00:00:00','    Man, so many things are running n my head.  I though I would just put a few of them on here, free up some space in my head.  Man, I am falling hard and fast for these kids.  I don\'t think I have ever been so committed, so passionate about anything in my life.  something inside me desperately wants to take the older kids, the 10, 11, 12 year olds and erase all their hurt, make them forget everything bad that has happened, fill up thier voids with God\'s love, teach them to dream again; and take the little ones, keep them innocent, sheild their eyes, cover their ears, impart Jesus love in such a way they will never leave it, and teach them to walk in his plan.  I want to tell them they are worth more than any one has ever given them credit for, that however big their dreams are, their potential is even greater than those dreams.  The potential contained within one child is beyond what we could imagine.  Those we think of as great people were all once children with dreams:  Billy Graham, David Wilkerson, Mother Teresa, etc.  They were all kids, and someone took the time to believe in them.  The other day Pastor Eric was talking about the difference between volunteers and called ones.  He said something to the effect of volunteers will give their time to do what needs to be done, but called ones will give their lives.  Called ones are committed to give whatever it takes.  That is why I have been saying I feel like I am doing what I was created to do.  Before when I have done other ministries, I have been a volunteer, serving, giving my time and energy, because it was needed, and I wanted to serve.  But I have never given my life to something like this before.  Never before have I understood just how empty money and possessions are.  Not that there is anything wrong with having money, but it doesn\'t do anything for me.  If I just had a bed to sleep in and food to eat, and time and transportation to do Metro Kids and spend time mentoring these kids and loving them, that would be enough for me.  I\'m not just a volunteer anymore!!!  For those who are volunteers, there is still nothing wrong with that, becuase it is necessary, and God calls us to be servants, but it is amazing to begin to walk out my calling, and i know there is so  much more to come.  I praise God for leaders like Robert who will lay down their life to see others begin to walk in their destiny.  I know it has to be hard when you are a man of excellence to release others to do the ministry.  Thank you PapaRob.  It\'s cool cuz Rob didn\'t just leave me hanging, but started out doing a lot with Metro Kids, and is slowly backing out as I am more comfortable with doing it.  It\'s like going to kindergarden.  The first day, he carried me all the way to the classroom and held me for a while, then the next day he let go at the door to the room, and soon he will just pull the car up to the curb and wave as I get out.  That is enough for tonight, I guess.  I need some sleep.  (This has been another episode of Ramblings of a Tired Sarah)   '),(4069,2607577,'21,june,2004','2004-06-21 00:00:00','    I told you guys that new dreams were being birthed in me.  And the old dreams are being redefined.  Guess it\'s time to dream with y\'all!  Lately, I\'ve been in this sleeping mode.  God\'s been allowing me to rest and recuperate from the MAJOR spiritual battles I\'ve gone through in these past 4 years.  (MAN!  I just realized it\'s been 4 years of constant battle!  I hadn\'t really thought about how long I\'ve been going through all this!)  And now, He\'s calling me to wake up.    When I got home from St. Louis, I was still on this mega high of wanting to use what I\'d learned and get out on the streets to save the world, one girl at a time!  And then Mom got sick.  And Jason came back.  And I started ANOTHER internship program.  And I had no money OR car...  I became overwhelmed and lost the zeal I\'d come home with.  I got busy again, losing myself in the duties of being a servant  only  and not a daughter of God.  I thought throwing myself into the work of being a servant would peove to everyone where I was at spiritually.  This was a BIG problem for me with Rob.  Coming into the New Castle Dream Center, I was admittedly cocky and assumed I knew what I was doing.  WRONG!!!!!  I\'d learned integrity in my work ethic, but I didn\'t learn how to rest and just shine.  I didn\'t and DON\'T need to prove myself through works anymore.  People see that I will work to the best of my ability in whatever is put before me.  Now, people need to see me glow for Jesus.   One of my spiritual mothers from my old church saw me a couple of months ago and noticed I\'d gotten quiet.  Well, she described it as not lighting up like the 600 watt light bulb she remembered.  Ouch!!!!!!!!!  I thought, \'I\'ve only just gotten back from St. Louis!!!  I should be glowin in the stinkin dark!  I\'ve earned it, haven\'t I?\'  (Paradigms keep shifting, thank God.)  Wow.  Talk about needing to re-evaluate where I was at!  (Will the evaluation process ever end?)  Where exactly did I lose that glow?  And why?  Well, I got busy, and that\'s when I noticed the definite change.  Not only was I busy, but I was worried.  Do daughters get worried about what their daddy is making and if there will be food on the table?  NO!!!!  I don\'t think so.  So, why was I worrying about all the unnecessaries in my life?  Sure, I thought the situations needed to be resolved and IMMEDIATELY!  But I forgot...  IT\'S NOT ABOUT ME!!!!!!  OR MY TIMING!!!!  You are asking, \'what does this have to do with dreams?\'  I\'ll tell you.  After I realized I wasn\'t glowing, I asked God what it would take to get that glow back again.  What does He then do?  He takes out everything that\'s in me, (rather painful process) and slowly begins to put back in me only those things He wants.  Yes, I will be a firebrand for God again, but the fire I had had too much of \"Anna-Spirit\" and not the Holy Spirit.  Yes, I will be a prayer warrior, but He wanted to make sure I could even concentrate on ONLY those things He wanted me to pray about and not make my prayer time into a worry session.    My friend Christy (you don\'t know her) had a vision God gave her about me.  She said that she saw me and then I turned into a glass half-full with water.  She then saw the hand of God pour me completely out and throw me, the whole cup, into a huge bucket overflowing with water.  That\'s exactly what He\'s doing now!!!!!  (It\'s cool to see the things He\'s shown me and other saints coming to fruition.)    And just how is this redefining my old dreams and giving me new dreams?  What I\'d thought of as FOR SURE things which God gave me and ended up twisting to my own understanding were taken completely away from me so God could give them back to me in the pure form He intended.  Woohooo!!!!  Painful process, but AWESOME!  And now!!!!  He\'s giving me new dreams and old dreams with a new twist.    This is cool, guys!  The girls\' home?  I don\'t know just yet.  God is still taking the \"Anna\" out of it and hasn\'t given it back just yet.  But I\'ll let you know as soon as I find out more details!   '),(4070,2607577,'21,june,2004','2004-06-21 00:00:00','    Finally, some victory I can see and feel!!!!!  Praise God!  Sunday night, my friend from 1st Assembly asked me to come to church with her and visit afterwards.  Well...  I was admittedly nervous about going back since I\'d not BEEN back since I broke off things with Jason.  Yes, I did go to my youth pastor and explain what all was going on, but I decided it was best not to go back to any services.  On top of that, I had to almost completely cut off any contact with the youth just in case any of them wanted to come with me wherever I went and didn\'t stay planted.  I went, though, and Sarah went along with me.  She\'d gone to 1st a long time ago and wanted to see how everyone was doing.  THANK GOD SHE DID!!!!  My friend couldn\'t make it because of a family emergency and that left me in a bit of an awkward position.  We were LATE on top of that!  (Funny.  God has seriously been teaching me some humility lately.  Side note: locked my keys in my car once and Rob did it a second time...  IN THE SAME NIGHT!  More humility.)  EVERYONE stared, but smiled.      Ah....  Better than I could even have hoped for!!!!  And the best part?  I wasn\'t even bothered by the fact that Jason sat right in front of me, flirted with me later on, and was just plain THERE at the service!!!  I WAS JUST FINE!!!!    God, You\'ve done some amazing things in my life.  But this freedom?  Whoa.  This is cool.  I\'m not BOUND anymore.  I don\'t even feel nervous around him.  And that\'s thanks to You.    Thanks, God.  I love You!   '),(4071,2607577,'21,june,2004','2004-06-21 00:00:00','    I am very proud, and NOT surprised at the way you are handling this Anna.  You are a quality woman of God who has a heart for Him.  Please know that I am always in your corner, and that I desperately love you!  Thank you for embracing the process!  Much Love, PapaRob   '),(4072,2607577,'21,june,2004','2004-06-21 00:00:00','    What a night.  I am really tired so I am going to try to make this a short entry (everyone who knows me very well is lauging out loud at that right now).  Tonight I had a flat tire on Sulpher Springs road, out in the country near the church.  So of course I called PapaRob to be my knight in shining armor (if girls were meant to change tires, God wouldn\'t have invented cell phones).  So I got the spare tire out and waited for Rob to come rescue me.    Then the guy whose yard I was parked in saw me sitting out there and came to help me.  We didn\'t make very much conversation, just names, thanks, etc.  Turns out he was one of Rob\'s friends (like, who isn\'t).  Anyway, the guy was a Christian, too.  So Rob was like, hey did you tell him anything about Jesus.  And I was like, no.  So he finished changing the tire, and Rob said, \"tell him your testimony.\"  So I got embarassed, but started talking.  After a few mins, Rob said, \"Tell him what God did in your life while you were in Texas.\"  So I started talking about it.  Then, when I got back in the car, I started thinking about it.  What exactly had happened in Texas.  As I began to look at everything that has happened in the last three years to get me to this point, I am just amazed at how God gets his plans to work, one step at a time.  When I left for college, I planned on getting 3 things:  a husband, a degree, and the training I needed to be a missionary.  What I got was an ex-fiance, a lot of lifelessons, and the training I needed to be me.  I never wanted to do childrens ministry, or street ministry.  I qualified for work study, and my chices were work at a Boys and Girls Club or do paperwork for a professor.  So i jumped on the Boys and Girls Club job.  I was timid at first, and the workers there didn\'t think I would be any good at it.  But after a week or two, I started loosening up, and I found out I was actually good at it.  I could take a group of over 100 kids between the ages of 6 and 13 and get them to listen to me.  Most of our kids came from high risk nieghborhoods and broken homes, and I fell in love with them.  I loved my job, putting in twice as many hours as they could pay me for the first year, then taking a position with them as Director of Educational Programming.  I eventually just knew it was time to move on, and I was heart-broken.  I loved those kids.  They loved me.  We had a bond deeper than any of the other staff.  I cried harder over having to quit that job than almost anything else.    But I still didn\'t see any future in children\'s ministry.  I had decided to do urban ministry instead of cross-cultural, but never knew that it would have anything to do with kids.  Then, I was supposed to go to Israel over spring break, just over a year ago.  They cancelled the trip do to the instability and violence.  So I decided to go to the Los Angeles Dream Center instead.  I began to read about it, and it was like passion just rose up in me.  I was so excited about the trip.  When I was there, what impressed me the most, was that the interns and staff there give their entire lives to ministry.  It isn\'t on of the many roles they play; it\'s who they are, and it\'s who I wanted to be.  I still didn\'t think about doing kids ministry, but wanted to do a Dream Center.    When I came home at Christmas, I was talking to Rob about this, and I had been praying about moving back to Indiana.  When Rob was talking to me about the internship, and developing your dream ministry, I blurted out that Metro Kidz was what I would want to do.  I haven\'t told Rob this, but I didn\'t realize that was my dream ministry until the moment I said it, and it just hit me.  Then Rob was like, I have the trailor, I just need someone to run the ministry, and on the outside I was like wow and on the inside I was like holy cow.  But I knew then that it was God.  Now, I\'ve told several people this before, but I feel like I am doing what I was created to do.  It is the most incredible feeling in the world.  It is amazing how God used all the little things in texas, the things that I just thought were details, to paint the bigger picture.  I can\'t wait to see what is in store next.  I have been doing a lot of dreaming and praying lately.  More about that later--gotta tell PapaRob the details before I tell the rest of the world.   '),(4073,2607577,'21,june,2004','2004-06-21 00:00:00','    Yep....  Thought I\'d hear back from you, Rob.  Actually, I wondered why you didn\'t yesterday when I dropped of the moolah for the LA trip.  Thanks.  Needed it.  Be encouraged, y\'all.  This seems like I keep on the same cycle of obsessing, getting over it, obsessing, getting over it...  And in a way, it was the same cycle.  Only this time, I\'ve asked the right questions and got some real answers...  And I\'m living out the results.    Before this, I would never have submitted my thoughts about my personal relationships with ANYONE, let alone hundreds of people on the N.C. Dream Center Blog Site!!!!!:)  And definitely not ROB WHITE!!!  (I\'m growin up, Rob.)  The cycle is broken and will not be something I struggle with so much anymore.  Yes, my own evil desires drug me away and enticed me.  BUT, and I lke this \"BUT\", I AM THE RIGHTEOUSNESS OF CHRIST!  And I am victorious.  Kristie has been talking with me and counseling me about this.  To tell the truth, I ASKED her to counsel me on this topic.  She\'s a woman who hears from God and I\'ve submitted my total and complete thoughts to her.  I\'M EXPOSING!  God is walking me through this, praise Him!  He\'s taking His time because He\'s seen what I\'ve done in the past with all the stupid relational mistakes I\'ve made.  Now, He\'s given me a gift.  AND He\'s given me the grace to be able to wait........  Patiently.  Last night, Kristie and I went walking and I told her how I wanted that peace back which I\'d lost by obsessing over the situation.  THEN, I talked with my friend Craig from AL and he prayed with me about it over the phone...  (I\'ve been doing a lot of talking)  I can truthfully tell you I have peace.    I love Ecclesiastes 3.  Well, I love the whole book of Ecc.  (Been studying it lately.)  There is a time when all these things will come to pass.  God\'s Word does not return void.  He\'s told me secrets and whispered instructions.  I\'ve only to wait for His timing.   I know, Rob.  Wait.  And I will, joyfully not thinking about it.  Honest.  In the meantime, there are OTHER things He\'s been saying and quietly dreaming into me.  I don\'t know when to tell, but soon He\'ll release me.  There is so much more to what I\'ve seen in this life.  And God is SOO STINKIN EXCITED about showing it to me.    Guys, He\'s my Lover.  And yours.  \'Lover of my soul...\'  I prayed for that countless times as a child.  I wanted a lover of my SOUL who knew me inside and out.  AND didn\'t care I didn\'t even know myself yet.  He reaches His hands down, covers my heart, and soothes, energizes, and strengthens me.  That\'s the best way I can describe it right now.    There\'s so much He\'s got planned.  He\'s allowed this time of relative peace since I returned from St. Louis.  I\'m not the \"cheerleader\" like I was.  Not that I won\'t ever be, just He\'s given me a rest.  I believe that once I\'m in the intern house, there\'s going to be an explosion!!!!!!!  IN A GOOD WAY!  He\'s allowed new dreams to appear and new directions to take me feet where He wants.  This next phase?  Who knows?  I just know He\'s gonna keep me BUSY!!!!:)   '),(4074,2607577,'21,june,2004','2004-06-21 00:00:00','    urlLink    These are the ones who make it happen!  Thanks guys for all that you do.&nbsp; urlLink   '),(4075,2607577,'21,june,2004','2004-06-21 00:00:00','    I am blessed to see how God is working in your life Anna.  You see, since I have known you, this has been a great area of weakness.  It\'s not that your future husband needs you for him, but it is your future husband that needs you to be Gods.  If you were to pick this up too quickly it would ruin both of you.  It is so important that you see past your current feelings and get into the vision of what God has for you.  Good leaders are able to see the consequences of the decisions further than their peers.  And this is what you must do.  You know that you are a born leader, and you must die to yourself, your dreams, your will and start to walk in maturity.  God desires to have you in all areas of your life, and this is the one that He has not ever got a commitment from you in.  God needs you to be His, and His alone.  Your choices no longer affect only you.  As a matter of fact, your choices will never again only affect you, but the countless 1000\'s of people you will touch by your leadership and servanthood.  You must look further beyond yourself, this has absolutely nothing to do with you, but it has everything to do with you (dynamic tension).  If you are going to be his sister in the Lord, then you must act like it.  Dying to your own desires, and protecting his purity, even when it comes to flirting, and spending time together.  You must stop being so selfish and acting like you are actually being a ministry tool for him.  In reality if this continues, all you will become is a stumbling block for him.  God doesn\'t need you to help to change him.  God would never put you into a position that would cause you to be a stumbling block to someone.  But in our own evil desires we are drawn away and enticed, and we become the bait that satan uses to snare someone else, and we call it righteous acts!  That\'s a bunch of hyper spiritual BS!  All it is, is a way that we justify our weakness so we can get what we want.  The process is a painful one, that requires us to continuously look inward at our heart and our motives, for God to reveal any wicked thing within us.  Anna, we are all guilty of this, no one is exempt.  You are an incredible woman of God with unlimited potential.  But potential never got anyone anywhere!  It is something that must be lived out, through the pains of growth and perseverance.  Keep pressin\' hard baby girl, and stop laying down!  Fight for righteousness and holiness, these are the prizes we strive to attain.  Live the standard, be the model, set the example (1 Tim 4:12).  If you don\'t want others to look down on you, then this is what you must do!  I love you and I am very proud of you for allowing God to go to those places that are painful.  This is the price of righteousness, this is what you were called to, this is the example that Christ has set.  LIVE IT LOUD!  Much Love, Rob   '),(4076,2607577,'21,june,2004','2004-06-21 00:00:00','    My last night at the Flying J was last night.  Peace.  Funny story, I almost didn\'t even make it in last night!:)  I\'d set my cell phone alarm to wake me up at 9pm to have time for a shower and all that.  I don\'t know what happened, but the phone didn\'t wake me up!!!!!  It did decide to go off at 11:38 when it let me know I had a voice mail mess!  Of all things!  Work had been trying to call me on the cell and at home.  My family, God bless them, thought I wasn\'t there!!!!!!  Gotta love them.    When I did finally wake up, I was SOOOOOOO tempted to just go back to sleep and forget about it.  Ugh...  Conviction.  It was my last night, so who could blame me for going ahead and sleeping in?  I was already going to be EXTREMELY late...  They probably already called someone else in to cover for me........  Just couldn\'t do it!!!!!!!!  God kept telling me, \"Integrity, Anna.  You asked to grow in it, well this is how it\'s done.  You want others to be able to show the same kind of integrity towards you, so you show them what you asked for.\"    I got up, scrubbed my WinnDixie feet (volleyball in the mud) and washed my face, put on my uniform and ran out the door!  My manager, who is a big fan of Joyce Meyer and knows that I was an intern at the St. Louis Dream Center supported by her, was so stinkin proud of me!!!haha  She honestly didn\'t think I\'d come in and wasn\'t going to blame me if I didn\'t.  Needless to say, I had a great night.  I may also be able to help her get on at the park where my dad works.  She\'s been trying to get on there and needs a reference.  Ta da!  Me!    Anyway, that doesn\'t have a lot of significance compared to what else has been going on in my life since my last \"official\" blog.    I\'d asked for prayer for a certain situation the last time I wrote and still need that prayer.  To give you a little insight as to what\'s going on, here\'s the deal.  I know God has this amazing man of God coming my way and He\'s preparing him in the most awesome and amazing way.  My problem with this process?  I obsess.    Ok, I\'m gonna be dead-seious honest here.  I know who that man is.  Yes, Rob, I know.  And I\'ve asked God to give me confirmation, confirmation, confirmation, confirmation, confirmation....  I\'ve had it.  And I\'ve also had some warnings from people about this whole thing.    Now, confirmation from God and then warning from man doesn\'t bode well for the well-being of my already worried brain.  Yes, I worry too much, I know.  And God has grown me in the area of true peace...  Especially in these areas.  But, (I hate but\'s) I\'ve slipped back into the old way of worrying, fretting, and over-contemplating.  I get to dreaming about what it will be like to have someone constantly with me, encouraging me, needing me and I get side-tracked with all the dilemas and road blocks to the upcoming wedding!lol  (Yes, I said wedding.  I even think about that.)    Grrr!!!  Sometimes I really hate being a woman.  We women are so multi-tasking that we get ourselves into trouble in the area of ruling and tempering our thoughts.  One minute, I\'m praying for my husband and his walk with God and the next, I\'m thinkin\' about what colors will be in the wedding, will he want me to work, what\'s going to happen with the girls\' home....  Men, you just don\'t understand.  It\'s bad!  So, all these thoughts are plaguing my mind and not allowing me to really break out the prayer for my husband.  He really needs it right now and is under attack.  I guess I\'m being attacked in the mind so I won\'t be able to pray and intercede for him like I need and want to.    This is so hard.  Not only do I need to pray for my husband, but right now he needs to be my friend.  AND JUST MY FRIEND!!!!!  If he\'s anything more than that to me at this time, I\'ll want to up and run away with him.  MUST THINK OF HIM AS BROTHER!!!!  Right, Rob?  Any other female alive will agree with me that this is one tough task.  And I am treating him as my brother.  I truly want him to hear what God has to say.  And I need God to continue kicking my own butt on MANY different issues.  So, he\'s my brother I obsess about.    See my problem?  In all honesty, this is affecting me more than it needs to.  From too many messed up relationships in the past, my nerves are shot.  This means now when I get worried or upset, my body goes into melt down.  It hasn\'t helped that I\'ve been working too much, sleeping too little, and not spending quality \"God-time.\"  Well, I\'ve been physically ill again.    Guys, I prayed and asked God to heal me of this.  And He is.  Right now, I need you all to be praying I don\'t pick this back up again.  There is something coming up and God needs my full attention.  Pray my mind is eased from worry about him.  Pray I rest even my mind.  And please pray that peace returns.  I could use it.     *********************** (This has been another \"Whoa-is-me-Blog.\"  Sorry.)   '),(4077,2607577,'21,june,2004','2004-06-21 00:00:00','    I have so many things I want to tell you!&nbsp; So many great things God has been doing in me!&nbsp; I am so blessed because of the people God has put in my life.&nbsp; So Trebby I definitely agree with you enjoy it because it stretches you but knowing God is doing stuff to help you grow is just an awesome feeling.&nbsp;&nbsp;  I wanted to share this story with you guys!! especially you guys!! While I was at camp Brandon Bealls told us a story.&nbsp; After that story I was like man a simple story that helps me explain what I am feeling.&nbsp; Anyway he talked about how one day he brought in a tank full of gold fish.&nbsp; During his youth meeting he took and dumped all these fish out on the floor.&nbsp;&nbsp; You know what happens when fish are not in water they flip around until they die!&nbsp; Well the kids were looking at these fish and got to the point where they could not take it anymore.&nbsp; They jumped up running to these fish and started grabbing them by the handfuls putting them back into the tank.&nbsp; There hearts were going out to the fish.&nbsp; They could not handle it that they might die so they had to save them.&nbsp; Come on now if your heart goes out to these fish and you are one who would of ran and scouped them up!&nbsp; Tell me why a little fish would mean so much however people who don\'t know God are dying!&nbsp; There are so many lost out there who need the fish tank and all it would take is for us to scoop them up and put them in there.&nbsp; I guess I am waiting for people to take their stand and stop being scared.&nbsp; Who wants to live of life of being scared all the time.&nbsp; I know I don\'t!!&nbsp; When are people going to realize its not about them!&nbsp; So your scared so what!&nbsp;YES its that easy to say so what! I will be real I get scared talking in front of people.&nbsp; I get to the point where I think I am going to puke! ITS&nbsp;a real feeling and I even start shaking!&nbsp; SO WHAT!&nbsp;I just deal with it and keep on going forward cause I know&nbsp;eventually God is going to take that crap away. &nbsp;I don\'t know about you but for me its what can I do for him.&nbsp; Yes I make mistakes on the way and maybe sometimes don\'t make the best decision!&nbsp; But I am enjoying the learning process.&nbsp; If you look at it like I need to be this good at this or wow if I fail I should give up!&nbsp; Then tell me how can you be happy worring all the time those things might happen.&nbsp;&nbsp; Brandon also said one thing that I was just like YES YES! I FEEL SO STRONGLY ABOUT!&nbsp; He said you know what the enemy hates more then a praying man,&nbsp; a pastor a evangelist........ He hates someone who won\'t GIVE UP!&nbsp; He hates more than anything someone who says sorry devil but I know what God has in store for me and its so much better.&nbsp; That alone guys makes me not even think about giving up.&nbsp;&nbsp; I have a little story I want to share about not giving up....  Sunday God gave me two visions!&nbsp; The first one was me sitting in church and chris coming up beside me and talking to me.&nbsp; Well not five minutes later it actually happened.&nbsp; Exactly as I saw it!&nbsp; I was just thinking weird!&nbsp; Well later on I realized why God did that.&nbsp; During worship God showed me something else.&nbsp; It was during worship!&nbsp; In my vision or whatever you want to call it I was also worshiping and my husband walks past me and kneels down at the alter!&nbsp; HOW AWESOME IS THAT! Then I realized he showed me the first one with Chris and it happened as I saw it.&nbsp; NOT TO GIVE UP! that its going to happen...my hubby is going to be saved.&nbsp;&nbsp; The funny thing is that is the one thing I know in my heart is going to happen.&nbsp; I know God is going to capture his heart.&nbsp; So I just thought God even though I trust you and have complete faith in you thank you I needed that.&nbsp; I needed to see that!&nbsp; So even when you don\'t know what you need God does.&nbsp; That is just so awesome to me!!&nbsp; I guess I could just express in words how awesome I feel.&nbsp; How HAPPY I am!&nbsp; I don\'t care what happenes to me!&nbsp; I don\'t care what God wants to do with me.&nbsp; I don\'t care how he does it.&nbsp; I am just glad he cares enough to help me and change me to be the best that I can be.&nbsp; That is sooooo exciting to me.&nbsp; THANK&nbsp; YOU JESUS!!&nbsp; I just want&nbsp; you to know that I love you guys and you are in my prayers daily!   '),(4078,2607577,'21,june,2004','2004-06-21 00:00:00','    Hello,   I have not blogged in quiet some time, but I have enjoyed reading Kristie and Anna\'s blogging.&nbsp; I have been living at the intern home for almost a month and have really enjoyed it. I had great reservations before moving into this home, because I have lived on my own since I was 18 yrs old. I am 40 and most of the other people living there are half my age, but the ironic thing is the chemistry between myself and the other interns has been fantastic. I am blessed to be living under the same roof with such passionate lovers of Christ. We continue to use the word stretch, when describing what living communal does to each of us, and I couldn\'t&nbsp;agree more&nbsp;with that&nbsp;definition.&nbsp;This current situation does stretch all of us any many ways, but let\'s face it we could all use a good stretching occasionally...Living this way you give up&nbsp;certain privacies that you never really thought about before.&nbsp; This stretching for me has been long overdue...I plan on making the most of this opportunity and using it to draw closer to God, and in the process finding out what I am made of. I want everyone to know that I have chosen to live in this home, and quite frankly it has mystified most people that know me, because I have been on my own for so long...See I am the kind person that would be perfectly content with never having a phone, or for that much a computer...These things just don\'t matter to me. I am not sure how long I will be at the intern home, but for now I just want to enjoy this life experience   Trebby   '),(4079,2607577,'21,june,2004','2004-06-21 00:00:00','    Where is everyone?&nbsp; I thought this was supposed to be a DC thing?&nbsp; And no one else but me n Kristie are ever utilizing this thing!!!!!:)&nbsp; Kristie, we must be too good of writers and are intimidate everyone else.&nbsp;   Anyway, here\'s the scoop on what God has been doing lately.&nbsp; AND THERE HAS BEEN QUITE A LOT!   God has asked me to go up on the mountain.&nbsp; And I\'m scared to death of heights!&nbsp; Funny, but entirely too true.&nbsp;   Seriously, I\'ve been used to being in the field and working for God.&nbsp; I\'m a self-admitted \"Martha.\"&nbsp; Now, though, God has asked me to stop working and start taking the trip up the mountain to come see Him for a while, to become a \"Mary.\"&nbsp; What will I find up there?&nbsp; Who knows?&nbsp; And will I like being a \"Mary?\"&nbsp; Doubtful.&nbsp;   When I\'m working, I feel needed.&nbsp; There\'s the absolute truth of the matter.&nbsp; And when I\'m not working, I\'m obviously not needed.&nbsp; So...&nbsp; There\'s a loss of importance.&nbsp; SELF-importance.&nbsp; And self is what has to die if I\'m even to see God.&nbsp; I know, I know.&nbsp;   My family was and is a household where work ethics are STRONGLY pushed and adhered to.&nbsp; If you weren\'t working on SOMETHING constructive, then you were the weak link in the family.&nbsp; Growing up, I don\'t know how many sports or extracurricular activities I was in.&nbsp; But if I didn\'t have practice of some sort, then I was told I had to get a job and work.&nbsp; Once I was older, working became an even higher priority than all the extracurricular activites combined.&nbsp; Mom and Dad, especially Dad, said that work was more important because it was making me a productive member of society as well as the family.&nbsp;   See where being a \"Martha\" was exalted over the \"Mary?\"&nbsp;   If I wasn\'t working, then I wasn\'t productive and productivity was the only thing praised in my family.&nbsp; Obviously, if I wanted any praise, then I must work and excell to receive that praise.&nbsp; And MarthaAnna was born.&nbsp;   As a child, I was a dreamer.&nbsp; And one of the biggest liars ever!!!!&nbsp; I\'d make up stories you wouldn\'t believe and had my parents usually convinced!&nbsp; Needless to say, once they learned that they were just made-up stories, that\'s when the enforcement of productivity kicked in.&nbsp; If I was working, I hadn\'t any time to dream, to make up stories...&nbsp; Or enjoy being a silly kid.&nbsp;   Not only did the dreaming stop and the work begin, but by this time, both my parents had to go into the work force.&nbsp; And I had to take care of my brothers, as well as myself.&nbsp; I became Mom and annoying older sister&nbsp;overnight.&nbsp; My poor mom didn\'t have any choice but to go back to work.&nbsp; And I didn\'t have any choice but to become what my mother couldn\'t be.&nbsp; My dad, well he was&nbsp;so tired after work, he didn\'t have time to be both Dad and Mom let alone the energy.&nbsp; So, the MarthaAnna kicked into gear and I forgot what it was like to dream, to be silly, and just be.&nbsp;   Now...&nbsp; Oh, God.&nbsp; Now, I have to go back to that point where I stopped growing dreams, stopped imagining and start where I left off.&nbsp;   I\'ve known God all my life.&nbsp; Thank God for Godly parents!&nbsp; But, I stopped dreaming about Him, seeking His calming presence and I started working for Him.&nbsp; Work was praised in the home, so God must praise the work, more than just being.&nbsp;   And I wasn\'t touched as a child.&nbsp; Not once my brothers were born.&nbsp; Sure, I was picked up when I fell.&nbsp; And I was hugged when I cried because one of the boys hit me.&nbsp; But outside that, I wasn\'t touched just because.&nbsp; There wasn\'t that affirmation of love for me, so I assumed that WASN\'T an affirmation.&nbsp; Know what I mean?&nbsp; And I carried this over to my relationship with God and thought that since I was working, and THROWING myself into His work, then I was affirming and having affirmed God\'s love for me.&nbsp; Therefore, I didn\'t expect a touch from God outside of the instances when I fell or was hurt by someone else.&nbsp; God doesn\'t just HOLD His children.&nbsp; He has them WORK for Him and for the glory of His kingdom.&nbsp; Holding is for newborn babies and very young children, not for the \"big kids.\"   See where I\'m going with all this?&nbsp; God has called me up on the mountain so He can touch me, hold me, and just let me be what He made me be.&nbsp; There aren\'t any demands or expectations.&nbsp; There isn\'t a code of work ethic to follow during this time.&nbsp; There\'s just me and God.&nbsp; And a new kind of love to experience.&nbsp;   I\'m terrified of this kind of love.&nbsp; I\'ve never had it before.&nbsp; And it\'s exactly what I looked for in the countless wrong relationships I\'ve had with men.&nbsp; I looked for the touch, the holding.&nbsp; And found a dead end.&nbsp; Will I find another dead end with God?&nbsp; Will He leave me when it\'s over?   '),(4080,2607577,'21,june,2004','2004-06-21 00:00:00','    Hi everyone I just got back from Impact Youth Camp this week!!&nbsp; I don\'t have the words except WOW!!&nbsp; Hmmmmmm I personally have never attended a church camp.&nbsp; SO this was my first one!! I was blessed enough to go with seven of our Youth, Chris, Nathan and Exie.&nbsp; God showed me A LOT of things this week.&nbsp; Things about myself things about others.&nbsp; It was a learning experience for me.&nbsp;&nbsp; Have you ever been in a place where there were 300 kids just going crazy for GOD!!&nbsp; I am not just saying clapping and screaming.. I mean jumping, dancing, running around....I MEAN CRAZY for God.&nbsp; I was just blessed beyond anything I could explain.&nbsp; YOU know what\'s better than seeing 300 kids go crazy for God.&nbsp; When some of those kids YOU know and have a relationships with are going CRAZY too.&nbsp; They did not care who was beside them...They did not care who was looking...&nbsp; They just went crazy...PRAISE GOD!!&nbsp; Now I want to have a building with thousands of kids going crazy for God.&nbsp; I think we can have that here in NEW CASTLE!!!&nbsp; I am not going to give up until I see that!!&nbsp;&nbsp; Please pray for Chris and I and all the helpers to!!&nbsp; Just keep us in your prayers that God gives us Wisdom to do what needs to be done.&nbsp; That the kids(youth) keep the fire that they have right now!!&nbsp; Things are&nbsp;going to happen soon!!!&nbsp; God is going to rock this town and surrounding areas!!!&nbsp; I have more but I don\'t want to&nbsp;give to much away just yet!!!&nbsp; God bless you all and&nbsp;I pray extra special blessings for you&nbsp;this week.&nbsp;&nbsp; Luv you!!&nbsp;   '),(4081,2607577,'21,june,2004','2004-06-21 00:00:00','    My title is tow-fold.&nbsp; There is everybody, and where is everybody.&nbsp; Where is everyone on here?&nbsp; Why haven\'t they been blogging?  &nbsp;  And where is everyone when I need them?&nbsp;  &nbsp;  Yes, I\'m alone primarily because I CHOOSE to be alone.&nbsp; But this is different.&nbsp; I\'m alone right now.&nbsp; And I feel (yes, more EMOTIONS) completely without real, true friends.&nbsp; Know what I mean?&nbsp; Kristie does, I know.&nbsp;  &nbsp;  Not that any of my current friends aren\'t as real as could be.&nbsp; It\'s just that I\'m made differently.&nbsp;  &nbsp;  Here\'s a good example.&nbsp; Anyone evern read \'Anne of Green Gables?\'&nbsp; Kindred spirits.&nbsp; That\'s what I don\'t have.&nbsp; In fact, the only time I\'ve really ever had a kindred spirit...&nbsp; They just disappeared.&nbsp; Was is beacuse of me?&nbsp; Who knows.&nbsp; Was it that they were in all reality just flakes?&nbsp; Could be.&nbsp;  &nbsp;  Right now, God, I could use someone who already knows what\'s in my head.&nbsp; They can see it, call it out, and confront me with it.&nbsp; My friend Craig is like that.&nbsp; Problem?&nbsp; He\'s in Alabama.&nbsp; And he can be a bit of a flake at times.&nbsp;  &nbsp;  Yes, God.&nbsp; I know You are all of the above.&nbsp; But I could really use You using someone here on earth to do all of these things.&nbsp; (Sounds like I\'m choosing my own blessings!:)  &nbsp;  I need someone in my face, in front of me, who knows what is going on and can help me think through all these things without me feeling like they are unequal to me.&nbsp; I need someone here on my level to hash these things through with me that won\'t come at me in an authorative manner, \"spiritual\" manner, or ignorant manner.  &nbsp;  Man, that sounds harsh!&nbsp; But I don\'t mean it that way.&nbsp; I just mean I need someone completely unbiased by other opinions but has enough knowledge and wisdom to wisely help guide me through this time.&nbsp; Make any sense?  &nbsp;  God, I honestly don\'t care whether or not it is a man or a woman, same age or not.&nbsp; I just need someone THERE with me.&nbsp; Maybe they don\'t even have to talk, or maybe they have wisdom enough not to talk at all.&nbsp;  &nbsp;  I\'m pushy, I know.&nbsp; Lord, help.&nbsp; That\'s all I know what to say.&nbsp; Help.   '),(4082,2607577,'21,june,2004','2004-06-21 00:00:00','    Oh, guys..  This crap is hard.  And it is CRAP!  Had a talk with Rob last night.  Started out as a conversation about the job situation and it ended up being about the girls\' home...  And timing.  Guys, this just guts me.  WHY, GOD?  Why?    Rob wants and believes this is my winter season.  I know. I knew this already.  And I didn\'t want it to come.  Especially not now.  I wanted ACTION!  Let me have the winter AFTER the girls\' home starts.  Not now, God.  I\'m ready to go NOW!  Let\'s get this thing started so the many hurting girls I see have someplace SAFE to go to.  And not have to worry about every burden placed or brought on by them.  Especially after what I learned about one of my girls today.  Just found out that one of my girls, just a teenager, was forced to have an abortion by her mother.  She\'s had it and there\'s nothing anybody can do about that now.  BUT SHE IS HURTING!!!    She had no where else to turn.  So, now she feels condemned for life because of her actions.    It is so hard trying to show her God\'s love and all-covering grace.  She can\'t see through her pain right now to even remotely try to comprehend the fact that Someone loves her.  AND NOW!!!!!!!!!!!  I have to wait on God to finish a work in me.  To wait and just serve....  And watch the hurting continue with there being nothing in my power to do.  Yes, it never has been in my power.  And it never will.  But, God!!!!  Why is it burning so in my heart?  Only for You to say, \"Wait?\"  Dear God...  Help!   '),(4083,2607577,'21,june,2004','2004-06-21 00:00:00','    What\'s the deal, y\'all?  Kristie, the very person who usually is so quiet about personal things, is kickin out butts on here!!!  She\'s blogged more than any of us now!:)  Praise God for it, too!!!  You rock, Kristie.  Man, when I think of when I first met you...  How reserved you were with things of yourself and your life...  And now!!!  ROCK ON, SISTER!   Ok.  Now that I have a little bit of time, I wanted to tell you all about what happened this Sunday.  (Yeah, I know.  It\'s Friday now.  Who cares??!?!?!)    For a while, my joy had disappeared and I\'d literally been PUTTING ON joy...  I wasn\'t living joyously.  In fact, when I looked into the mirror at myself, my eyes were so droopy and void of life.  It\'s hard to explain unless you knew me.  See, I\'m one of those annoying people that is joyful at all times.  It takes more than a monsoon of emotions to usually overwhelm me.  And not being able to be joyous just because I WAS, that was hard.  I\'m very good at acting, I\'ve found, and can fool even the most discerning person.  (Used to be proud of this fact.)  But these past few months, the mask of the actress was slipping.    Anyway, back to Sunday.  I was fighting hard to get back the joy I\'d lost along with the HOPE I couldn\'t seem to find.  In the back of the sanctuary, I was worshipping and struggling, like usual, to keep my mind off of the things which were PLAGUING me.  At one point, I just got really quiet with God and was waiting.    My eyes were closed, but they FELT open.  I was standing in the middle of a field full of flowers and SCREAMING at the top of my lungs, with my arms thrown wide, \"FREEDOM!!!!!!!!!!!!!!\"  There\'s no way in the state I was in that I could even FATHOM this kind of freedom.  It used to be I would go down to the lake, any time day or night, and sit on my favorite rock and sing songs, whisper secrets, or cry out to God.  I hadn\'t been down there in MONTHS!!!!  Actually, it\'s been over a year since I went down by myself and waited for God.    So, when God showed this vision of me screaming out to Him...  There was no way that was me!!!!!  Not even two SECONDS after the vision left, a lady from church came forward with a word of knowledge from God.  She said someone had been struggling with this thing, this dark crevice.  It was such a struggle that it was in the stomach and in the bowels.  She said it was so painful, she couldn\'t stand and couldn\'t breath and that\'s how this person was living...  That was me.    This thing that stole my joy, it had affected my very health and had made me sick for several months now.  My weight has fluctuated from the stress put on my body through this.  And it was in my stomach and in my bowels.  I went forward, sank to my knees from the weight of this thing, and the woman held my head in her lap and prayed.  (I saw later many other women were there praying as well.)  TALK ABOUT SOME MAJOR BATTLE!!!!!  These amazing women of God went to the mattresses for me!!!  They fought for my freedom in the spiritual realm!!!  And they won.  After we were done praying, I didn\'t know what else to do.  So, I stayed at the front a little while longer and worshipped.  Well, the vision came back.  This time, the field of flowers changed to the faces of my church.  And when I screamed freedom in the vision, I screamed freedom in the church, in reality.    I DIDN\'T CARE!!!!!  Freedom was MY gift.  It was HIS sacrifice for me to have that freedom.  And I was claiming it!  And I was joyous in it!!!!!  Freedom...  I was free!!!!!!!!!!!  A while later, I went to the back of the sanctuary again.  I wanted a little distance to find God alone for a bit.  When I went back, the dance began!!!!!  Guys, I\'ve not been one to just start dancing.  But Sunday, I was.  And I didn\'t care if I looked like a fool.  I just didn\'t care.   A little girl came to the back to dance with me.  She was adorable.  God showed me I was a girl again.  I WAS DANCING like I was a child again.  No longer the servant, but the CHILD God made and then later chose.  I was young, new, and pure.    Purity.  With hope.  New hope.  And new life, like Rob said.   New life.  No bars or bondages.  Just freedom.  With joy back.  :)   '),(4084,2607577,'21,june,2004','2004-06-21 00:00:00','    I am so sorry God for being so selfish! It is not about me! Being here is not about me! It is about making a choice to let you use me to help the lost, forgotten and the broken.  I am so tired of the selfishness that people have.  I am tired of being one of the those people.  Help me have a heart that is all about you.  It is so hard Lord to really break down the walls to forget the words that are spoken to me.  The words are like a million paper cuts in my heart! Help my attitude even to be always full of joy!  I am so sick of the sadness and the hurt feelings.  I don\'t want it anymore I only want more of you Lord..ONLY YOU!!  I want to see my husband as YOU see him!! I beg you to change my heart...Please change my heart.   '),(4085,2607577,'21,june,2004','2004-06-21 00:00:00','    I only have a few minutes before I leave work!  However I wanted to share something with you guys!! God has been doing amazing things in my life!  He is teaching me how to die!! He is teaching me and taking away all the hurt, bitterness and unforgiveness.  In August I have been married for 10 years and there is a lot of emotional pain especially when God is not in your life.   There have been times that I hurt so bad that I could not feel anything.  I just did not care from one day to the next.  I put so many walls up and Praise God he is showing me how to break them down.  The enemy is not going to win the war!! Yeah I have not been strong in the past (and lost many battles) because I did not know the glory of God in my life.  I am taking my marriage and my husband back.  The enemy is NOT going to win..   With Gods help I continue to die and learn to pray for my hubby.  Its funny cause I finally realized in order to change my hubby I NEED GOD to change me.  AND even then its not me that is going to change him.. I can\'t put that fire in him that I have for the Lord.  He has to get it on his own.  I am just so excited because I love him Soooo much!!  I can\'t wait until we are praying every night together.  I can\'t wait to have the blessings that we will get for not giving up!!  I know deep down what God wants for me and my husband!!  WhoooHoooo!! Thanks for all the prayers and keep them coming I have a long road ahead.  BUT every step will be worth it!!   '),(4086,2607577,'21,june,2004','2004-06-21 00:00:00','    This one is for me just as much as for anyone reading this. And, *****DISCLAIMER***** this won\'t be easy to read.  Yes, Faith Without Works... Is dead!!!!!  Remember Jesus saying this? That\'s right, He said it. And He meant it.  Now, why aren\'t we listening and FOLLOWING this teaching? Perhaps it\'s because we each define works as something different. Ok, then. Let\'s look at a perfect example of the works of Jesus. THERE\'S where we can find our standard of works.  When Jesus was allowed to start telling the world the truth of Who He is, He immediately began walking and bringing up under Him. DISCIPLES! What does \"The Great Commission\" say?  Go out into ALL the world and make disciples. THIS DOESN\'T MEAN GO OUT AND FIND SOME PEW WARMERS! He said to make disciples out of all the world!!!!! No exceptions.  Ok, Anna. What have you made? What kind of people have you brought in? Are they pew warmers? Or are they true disciples? Do they have a BURNING within them to go out and make disciples just like they were made? Do they want to continue the legacy begun by Christ Himself then perpetuated by you?  Well... To be completely honest, there have been more pew warmers made than actual disciples. Why is that? I haven\'t continued with them. I showed them the truth and lived the life of a Godly woman in front of them.... But I didn\'t turn them into disciples. Why?!?!?! I thought,  \'That\'s all I needed to do. I just needed to show them the way. God will do the rest.\'   Yes, my God is an awesome, AMAZING God. But He needs  US  to be His hands, His feet, His mouth to his people. And not just for a one time experience, but for a lifetime of discipleship!!!!! This isn\'t a \"program\" where we are enlisted to sign up 5 different people and then have them sign up 5 more people... etc. This is a LIFESTYLE! This is a life totally different from selling magazine subscriptions or Mary Kay. Get this! This is a way of life! Not to be taken lightly, I\'ll add.  Y\'all are thinkin,  \'Ok, what\'s gotten into her britches!?!?\'  I went to Harp and Bowl last night, our prayer time at church, and had to get up and leave the room because I was so frustrated. And the frustration was with myself as well as with those in the room. I know I wasn\'t the only one who felt this way either.  I was  \'righteously indignant\'  about how our city, county, and region has been praying for hope to come for   GENERATIONS   ! I\'m not talking about praying for a revival of this area with praying for a few weeks. What I\'m saying is this; since this area was populated, we\'ve been praying for hope to come here. We\'ve wanted hope for the indians, hope for the \"wayward\" laggard or drunk, hope for our government, and now hope for EVERYONE not saved and living a life of sin.    Why hasn\'t this \"GREAT HOPE\" come?  DUH!!!!!!!!  Because we\'ve been praying for it inside of our churches and in our homes. WE HAVEN\'T BEEN OUT GIVING THE VERY HOPE WE ARE PRAYING FOR FOR OTHERS!!!!!!!!!!!!!!!!!!!!! How can we expect God to work a miracle of hope if we are terrified of getting out of our home churches or plain ole homes?   WE ARE THE LIGHT OF THE WORLD!    And where there is light, there is hope.  \'Light at the end of then tunnel. A light in dark times.\'    We are light, we are hope, AND we are the only thing keeping us from finding that great revival we\'ve all so desperately wanted to come to our area. Heck, our country!   Sure, I\'ve been prideful and full of myself and said, \"Nobody else can say that they are actually living for God like I can. I\'m living in the intern house, afterall.\" God, forgive me!!!!! I haven\'t been living. I\'ve been HIDING! I\'ve been hiding in the intern house, behind programs for God, and forgot all about the WORK for God.  I have faith and have been called a woman of faith. But if my works aren\'t there, and the fruit certainly isn\'t, then my faith is   DEAD   ! Do you guys get this? DEAD!!!!! As in not LIVING!    What can I do, God? Yes, ask your forgiveness. Ok, then. God, PLEASE forgive me for being so stuck in what I thought was living for You. Change my ideas, change my heart so that I might have YOUR heart. And now? MAKE ME BOLD!!!!!! Show me the ways You want me to work for You. And don\'t just show me. SHOW THEM!!!!! Show whoever is reading this. BURN their hearts, burn with correction and with passion. Burn me, God. BURN ME!!!!!!!   '),(4087,2607577,'21,june,2004','2004-06-21 00:00:00','    Wow, it\'s been a long time since I wrote.  It has been an exciting, crazy month.  Moving into the intern house, the Project, has been awesome.  I love everyone who lives there so much and we are really able to encourage and minister to each other.  It has also been a very stretching month for me.  Last night I was reading my journals from over a year ago, when I was still in school.  I guess I had forgotten what it felt like, but I wrote over and over again how I felt so frustrated going to school and not being able to do ministry the way I knew I needed to.  I wrote that I had to bury the passion that consumed me so deep inside just to make it through the school days.  Man, am I glad God is allowing me to begin to walk in my calling.  Rob always says, \"I\'m living a dream.\" That is truly how I feel.  At the Kidz Crusade, one of the girls God really placed on my heart from day one got saved for real.  I am so excited to see what God is going to do  in her life.  53 kids got saved total, out of 75 in attendance, 10 of whom were my helpers.  What was incredible was that most of the kids at the cursade came from Stonegate and Hospital Heights, the 2 neighborhoods in which we do Metro Kidz. When I first came here, I figured I would be a dream center intern, and run Metro Kidz in Hospital Heights and that was it.  Boy was I wrong.  I had no clue how God would enlarge my heart towards our community.  I can\'t imagine not doing this now.  I used to question whether or not I even wanted to do ministry full time because I wanted to live more for me.  I thought that would make me happier, yet I am happier now than ever.  Focusing on others makes it so my issues seem so small.  I know that my strength only comes in my weakness, so that Christ\'s strength can be made perfect in me.  That is the only way I can do what I do.     We\'re living a dream!!!!!!   '),(4088,2607577,'21,june,2004','2004-06-21 00:00:00','    This weekend was another \"God\" weekend.  Well, it\'s all God\'s anyway.  It\'s just that sometimes I actually REALIZE that and give Him what\'s His already.   Anyway!  I was supposed to have gone to the Gig like usual, the radio show we do on Sat. nights, but God had other plans.  I couldn\'t even tell Rob I wasn\'t going without breaking down into tears.  (SUCH a girl!)  But, I did.  And God was there.  I did nothing that night other than play my keyboard and do a ton of worship.  At one point, I had to even stop that.  There was some worship music in the CD player, so I let God do what He wanted and laid down on the floor, spread-eagle, and waited.   Man, did He ever tell me some things.  Keep in mind, I didn\'t let them go into full effect until mid-Sunday.  Point is, they were spoken into me on Sat. night, then God started RAINING DOWN!!!!!  (Those worship songs were right on.)  I\'ve told you all that God has me up on the mountain for a while.  And it sucks.  But ROCKS!  I have an excuse for saying \"no\" now.  I can say no.  BECAUSE God said to.  No more working for the time being.  No more delaying and putting off what God has been trying to teach me.  It\'s NOW.  My time is now.  And I am SUPPOSED to be completely selfish with what God is currently giving me.  Yes, later on God will want and demand that I give out what He\'s given me.  Though, for now, He\'s given it to me FOR me.  I\'m not used to that.  I\'m so different from when I came home.  Just six months ago, and I\'ve completely changed.  Hard to believe I needed to change that much.  Guess I did!  I must have been a real JERK!:)  Anyway.  I\'m healing.  And I have less to say.  I\'m not supposed to.  So, WHEN ARE THE REST OF YOU GOING TO START SAYING ANYTHING?!?!?!?!?!!?!?!?!?   '),(4089,2967519,'28,March,2004','2004-03-28 00:00:00','   yummmy!!! i love this new diet coke! its sooo good! you all should go out and get some...anyways im getting my hair did next week but i havent decided what color yet.Lots to choose from lol went and got a bunch of movies last night...But ive got to say that Le Divorce has been my faveortie...Lots of French being spoken but its a really great movie. I think i had better go finish my dishes and ill write more later! Kayla  '),(4090,2967519,'27,March,2004','2004-03-27 00:00:00','   Im just now wakeing up its 9am here in AK and im supposed to be going to Acupuncture with my grammpa but we had a very exciting night last night! my moms friend kelly came to visit from Pennysylvania so we had to go get her at the air port at like 2am (when i say we i mean my mom lol) and her plane was late so they didnt get home till 4:30am and i was waiting to make sure everthing was okay so im still a little tired...Not much else to report! have a great weekend! Kayla  '),(4091,2967519,'17,March,2004','2004-03-17 00:00:00','   So this weekend will be pretty un-eventful so i wont be writting much...my mom and step dad are doing there aniversary thing,going to dinner,hotel the whole nine yards(by the way did you know they were making a sequal to that movie?) so my sister and i are going to town tomorrrow and getting movies and pizza and junk food and relaxing without parents for one peaceful evening. I also just found out that i have enough sky miles to fly for free! YAY!HMMMM.....where shall i go? any ideas? *ME* ps i dont know if they put this on here but if you wanna contact me my email is blonde_chick_6353@yahoo.com im always happy to hear what im doing wrong with this whole web site thing! :)  '),(4092,2967519,'14,March,2004','2004-03-14 00:00:00','   Okay,so,most of you know that im obsessed with Soap Operas...Passions,The O.C...but mainly Days Of Our Lives and everyone just keeps dieing on this show! And not like newbie people but people that have been on for like 30 years!! Crazyness and know they are saying that they want to kill off one of the characters that has been on the show since the begining in 1964!! Yeah...i was mad.Okay now to the non-trivial parts of my life lol j/k i had to get a shot yesterday...so im pretty sore b/c they gave it to me in my thigh...ouch...well i hear wood being split outside and its likely that ill have to bring it in so ta-ta! Have a GREAT Spring Break for all you in AK! *Kayla*  '),(4093,2967519,'12,March,2004','2004-03-12 00:00:00','   Sorry for not writting...ive been up to my elbows in school work trying to get everything finished to go see my dad in May.Hopefully this will be a less eventful trip than the last ones that i have taken to see him and my step-mother.It looks like its going to be a HOT summer in Colorado this year...the last time i was there it was 112* the ENTIRE time i was there.Way to hot for me. It has also come to my attention(and the attention of my friends) that i havent really said anything about myself on here so here it goes...My name is Kayla Hosea im 16 and live with my mom,step dad and little sister in North Pole Alaska, im fashion obsessed and hope to make that my future buisness venture in life. ummm...hmmm...yeah ithink that covers the basics  *MissKayla* Ps wonder where the Miss Kayla comes from? Its a nickname given to me by my grandma and grandpa McLean who come up with creative little nicknames for all of us grandchildren mine includ Miss America and Muffin....Just though y\'all would like to know  '),(4094,2967519,'06,March,2004','2004-03-06 00:00:00','   I know that its only been a few hours since my last post but so much has happened since them...A certin guy who shall remain name-less that is continuessly leading me on and them dropping me like a freaking hot potatoe has agian done it! Men...what makes them think that they can act like complete dick heads and then turn around and expect us to forgive them? Because like idots we always seem to..well i propose a strike for every woman who has had to deal with a selfesh,manipulative jerk of a man....Okay i feel much better *Miss Kayla* Ps Cold hard bitch by Jet is a great song---Check it out!  '),(4095,2967519,'05,March,2004','2004-03-05 00:00:00','   Im suposed to be cleaning right now...but what mom dont know dont hurt her right? right...My cuz Jody emailed me this joke i think y\'all will like  The following scene took place on a BA flight between Johannesburg and London.   A white woman, about 50 years old, was seated next to a black man. Obviously disturbed by this, she called the air Hostess. \"Madam, what is the matter,\" the hostess asked.   \"You obviously do not see it then?\" she responded. \"You placed me next to a black man. I do not agree to sit next to someone from such a repugnant group. Give me an alternative seat.\"   \"Be calm please,\" the hostess replied. \"Almost all the places on this flight are taken. I will go to see if another place is available.\" The hostess went away and then came back a few minutes later.   \"Madam, just as I thought, there are no other available seats in the economy class. I spoke to the captain and he informed me that there is also no seat in the business class. All the same, we still have one place in the first class.\"   Before the woman could say anything, the hostess continued. \"It is not usual for our company to permit someone from the economy class to sit in the first class. However, given the circumstances, the captain feels that it would be scandalous to make someone sit next to someone so disgusting.\"   She turned to the black man and said, \"therefore, Sir, if you would like to, please take your hand luggage because a seat awaits you in the first class.\"   At the moment, the other passengers who were shocked by what they had just witnessed stood up and applauded.   This is a true story which is not usually told. If you are against racism, please send this message to all your friends, please do not delete it without sending it to at least one person.   *Mss Kayla*  '),(4096,2967519,'05,March,2004','2004-03-05 00:00:00','   Well this is my first post not too sure about this whole web-site thingie but ill give it a shot...pretty un-eventful day,just getting over a cold..ive been hacking and gagging my lungs out for the past 2 days..YUCK...i know...It seems that NYU is more in the distant than I thought (for those of you that dont know im going to NYU to major in Fashion Marketing) My cuz Amanda loves it there and says that since i like to eat just at much as she does New York will welcome me with open arms! lol okay its late ill try and keep this posted as much as possible *Miss Kayla*  '),(4097,2967519,'30,April,2004','2004-04-30 00:00:00','   I just watched 8 seconds,have you all seen it? The most sad movie i have EVER seen. Its a true story about this bull rider Lane Frost, And he dies when a bull broke his rib and it stabbed him in the heart.Okay now if that isnt sad enough at the end of the movie they have his partner Tuff Hediamen and Cody ? his family and everybody talking about how he effected thier lives.Yeah i cried lol. Im getting sick i think. Yuck i hate being sick. Coughing is the worst possible thing. It makes you sound like youve been smoking crack since you were two! LOL sorry, but its true. Okay gotta go email Amanda!  '),(4098,2967519,'26,April,2004','2004-04-26 00:00:00','   Went to town today. Jeanette skipped school and went with us so she and I could get an application for wal-mart, Had a semi-fun time riding around and running errands with my mom. We also went out to Dot and Brains house and she is going to start me a pink christmas cactus(my fave) they are soooo pretty and i love them. I think that Jeanettte is comming over this weekend to veg- out and watch movies. I really need to Call you Amanda and i will i promise!Right now i have a major headache and are in desperate need of some pain killers so ill write more tomorrow.  '),(4099,2967519,'23,April,2004','2004-04-23 00:00:00','   ARRG! I have a cold sore on my lip! It looks like someone punched me there! So gross! Anyways its my moms birthday tomorrow so im making her my famous lasanga and french bread dinner and carrot cake for desert b/c lets face it im flat broke! But hopefully ill have a job soon working as an assitant to a travel agent while im in Colorado visiting my dad. Monday I have to go and get a state ID, then hopefully whithin the next week ill be doing some major shopping for my trip.Got to get lots of shorts cuz its going to be like 112 degrees the entire time! Okay,i need to finish the dishes and call and see about dinner!  '),(4100,2967519,'21,April,2004','2004-04-21 00:00:00','   Im having kinda a blah day....school and stuff was easy, and i have to eat fish for dinner. Yuck i hate fish  '),(4101,2967519,'19,April,2004','2004-04-19 00:00:00','   SOOOO excited! Imagine my surprise when on wednsday evening i come down the stairs to see who is at the door and its...JEANETTE! I screamed and like fell down the stairs to hug her! I was so surprised,she wasnt supposed to be back untill June when I was leaveing and we were all bumed out b/c we wouldnt get to see eachother! YAY! So she came and picked me up and took me back to her house where she talked non-stop about North Carolina which is cool. Everyone was shocked to hear that she was back. YAY! And dont you ever run away agian!!! -Other news- Im starting this new Global studies thing which is actually pretty interesting if you can belive that! And also a cosmotology course which is an independent study,looks pretty spiffy if you ask me but i havent started it yet. Okay more laters Kayla  '),(4102,2967519,'14,April,2004','2004-04-14 00:00:00','   Boring day so far...Woke up this morning at about 9am, made myself a pretty spiffy breakfast,took a LONG shower and im sure that I used up all of the hot water...Then I watched Days Of Our Lives which was AWESOME today lots of good flash-backs of Alice and Doug at their funeral. Now im just waiting for my mom to get home so I can yell at her for hiding all the cookies on me. Very relaxing day indeed. Other than that I think that ill just keep on relaxing,right now im watching one of the greatest movies of all time....Three Men and A Little Lady...Very very cute! Ta-Ta  '),(4103,2967519,'12,April,2004','2004-04-12 00:00:00','   Oh man, if you could see my drive way right now you would think that we built our house in the middle of a lake! We cant even get our car in and out and had to park it at a neighbors house! LOL I talked to my cousion Justin yesterday,he got his drivers licence! YAY! Great for You! And I got a letter from Jeanette...so happy she wrote me back b/c no one ever does. Easter was good,went to Jd and Marys for dinner at about 1:00, then had to go to town and didnt get home till like 6:00...but it was all good. Lots of great food and everybody seemed to be in a pretty good mood. Someone was killed at Artic Man yet agian this year. What it basically is a big snowmaching party,lots of booze and drugs and aparently one guy was caught in an avalance and sufocated....something like this happends every year but it is still really sad to hear about. Nothing else much going on. Hope you all had a good Easter weekend!  Kayla  '),(4104,2967519,'07,April,2004','2004-04-07 00:00:00','   Okay im so scared!I have to go and see a new chiropractor on friday and im so scared that she is going to do something and like break my neck or something! Isnt that like entirely possible? On another note Easter we have to go to my step dads aunt and uncles place...that should be interesting,and loud, and full of insults! No lie! And I just got my new bed/futon,very pretty and soft. Now im all moved into my new bedroom with my sister instead of my aunt. Jody just sent me pictures of her prom! She looked so gorgous! Well g\'night everyone!  '),(4105,2967519,'02,April,2004','2004-04-02 00:00:00','   People are so stupid!Yes it is true the idiots of the world have united and are taking over! Everyday while talking to a friend or something they will tell me what there verison of a crisis is,A failed class,a boy who dosnt like them or being grounded...\"Hello!!! are you people not listening to the news? Did you not know that the world dosnt revolve around you?\" The petty little problems that people have are inevitable to the crap thats going on in the world right now and untill you are on the outside of it you will never see it. Sometimes i just want to shake my friends and say \"Shut-up!\" \"You are being so petty and stupid!\" And usually its not unitll you look back from high school and see it,but lemme tell you those of us that see it while we are in high school,are going to be the ones who dont stress out over a failed 2 week relationship or a c- on a test when you thought you studided hard enough to get a B! Its like why is that a problem agian? I read some of my other friends blogs on here and i find myself laughing b/c of the things that happen in there days that they think is bad.  lol sorry guys but its true...take a look at what your writting...Making a decision not to smoke or not isnt a hard one,who cares what people think! Screw them if they dont like you b/c you dont fit into there little clique! Which justifys my point that stupidity should be painful! Anytime a person says something or does something stupid they should be smacked over the head with a stick! And yes ive had my stupid moments,who hasnt,but when your are just plain dumb in your decisions or you thinking all the time you need to get a grip on reality and mabey making it painful will scare people away from being stupid! Thats my thoughts on that...whoa that was alot of writting...I still love my friends but guys,seriously,you would be a whole lot more fun if you lots the pettiness! Kayla  '),(4106,2967519,'01,April,2004','2004-04-01 00:00:00','   I just read about this guy who went to see The Passion of The Christ and after the movie confessed to killing his pregnant g/f! WOW! if that is the kinda reaction Mel Gibson was goingf or than good for him! I was in complete shock b/c this guy wasnt even a suspect! He made it look like it was a suicide...Sick Sick ppl out there! But at least he is behind bars! Kudos to the movie and Mel! LOL Making cookies and watching Oprah, Kayla  '),(4107,2967519,'28,May,2004','2004-05-28 00:00:00','   This is going to be my last post for a few weeks as im flying off to Colorado to visit my dad. We are having a going-away BBQ tomorrow night and then sunday morning i fly out at like 6:15 am. Yuck, You have to get there 2 hours early so its going to be a little hectic. I never sleep before I get on a plane so by the time i reack my dads house im completly wiped out. And of course, everyone wants to stay up late and talk and eat and all i want to do is sleep for 2 weeks! I went to the doc and had that EKG done and everything is good with my heart so i dont have that to worry about. I got some really cute shorts and stuff too. Hopefully Jeanette will be here tomorrow, If she has to work i wont get to see her before i leave, Bri will be here and Christine. NO babies this time! YAY! Im not in the mood to play babysitter. Well i have to mop and put away dishes, Ill try and post as much as i can when something important happends. GoodBYe for now!!! Have an AWESOME summer everyone!!! Kayla  '),(4108,2967519,'25,May,2004','2004-05-25 00:00:00','   Went to my doctor appointment yesterday and everything went fine, she gave me my depo shot and drew my blood,but then she listened to my heart and she wants me to go have and EKG thing at the hospital b/c i have and irregular heart beat and it worries her. Im already in a bad mood b/c i have to get stuck with needles and then she tells me that i need to have someone moniter my heart! So im freaking out wondering whats going to happend tomorrow. When we were done with all of that we had to go to cell one b/c Liz\'s cell bill was WAY out there and ended up there for almost 2 hours! It was so borring. I have alot of stuff to do tomorrow so i prolly wont get around to updating till thursday or friday. I leave on Sunday, I dont like planes or being crammed in like a pack of sardines. anyways still gotta do it no matter how much i hate it. byebye  '),(4109,2967519,'24,May,2004','2004-05-24 00:00:00','   I just woke up, Somebody called here at like 4am...I was sooo pissed off,no one answered at the other end! grrrrr i hate ppl that do that.I have to go to the doc today which im not looking forward too. Im not good with needles and machines and all that. my sisters birthday was saterday and my mom invited Tawnie and Christy who have hunter whos 2 and Siera whos 3. Well Bri and i were dubbed babysitters so the adults could smoke and talk without distractions and the other kids didnt have to be bothered by the little ones. Im so sore still from throwing those two on the trampoline over and over! lol it was fun tho,kids say the darndest things! Okay i need to go take a shower before Tawnie gets here to take me to my appointment, have a great day!  '),(4110,2967519,'20,May,2004','2004-05-20 00:00:00','   It was sooo cold on Tuesday! A friend of mine and I went 4 wheeling and ended up going all the way to a store about 3 or 4 miles from my house and froze the hell out of ourselves! It was fun though.Next week is my last week ill be able to see everyone before I leave. Im excited to be going but really sad that I wont be able to spend the better part of my summer with my friends. I havent seen my best friend, Heather in a year and usually summers are our time to be inseperable but since my dad moved i dont get to see her! I guess ill just make the most of it and ill be writting everyone and sending pictuers whenever possible. Ive got to go get my sisters birthday present here soon and a couple of books for me to read on the plane. Okay im depressed now, lol j/k.  Kay*la  '),(4111,2967519,'17,May,2004','2004-05-17 00:00:00','   Im in the party mood or something lol. This weekend we are having a party for my sisters birthday but that will mostly be just family and Bri i think. Then the day before we leave for Colorado im throwing this huge BBQ, kinda like a end of school,beging of summmer,going away party thing. Lots of food and drinks,lots of people. Hopefully it wont be too eventful and nobody will end up in a fight or getting pissed off and leaving. Today was good. gotta go to town tomorrow, Oh im appling at WalMart when i get back to work in the infants, Jeanette got an interview there too so Good Luck to her on that. Season Finales have been horrible! I hated The O.C.\'s and last thursday i watched E.R\'s...That was horrible,no way to leave things. I didnt like Fraisers either, It was borring. Friends was the only good one so far. Okay enough of that!I should be able to get in a few more posts before i leave, and mabey a couple this summer but its very doubtful i will be able to update that much. Last day of school for all of us in North Pole and Fairbanks! YAY! lol  '),(4112,2967519,'14,May,2004','2004-05-14 00:00:00','   We just got about 50 baby chickens today. they are SOOO cute! Tomorrow we will get the 2 baby pigs. I usualy dont have anything to do with the animals but since my mom is working and nobody is home all of the time but me and my sister I guess i have to pitch in. Its also like 100 degrees right now in my house and im wondering now why we dont have air conditioning. I have to go get a shot on Tuesday and im not looking forward to that, I also have to get a physical just in case anything is wrong i know about it before i leave. I think that the hardest part for me, on my trip, is that i wont be able to check my email. That is the way i communicate with all of my friends. Its going to suck but ill survive somehow!  '),(4113,2967519,'12,May,2004','2004-05-12 00:00:00','   Im having a really good day so far! Nothing really terrible has gone wrong and no one had pissed me off to badley today,YAY! I wont be able to update for a while starting on the 30th. My dad dosnt have the internet at his house and so they only way i can use it is if i go to my Aunt Char\'s. Lots of plans to do stuff though it should be a really cool summer for the most part. Im sad that i have to say goodbye to everyone though, even if it only is for a couple of months. I usually send out all my friends an email lol! I hate the whole goodbye thing, it just isnt fun. Im going to the bookstore here soon to get some reading material for my trip, and i still need to do a little more shopping but im almost done!  '),(4114,2967519,'07,May,2004','2004-05-07 00:00:00','   My computer got a virus so I havent been able to use it for like 3 days! Drive me crazy! plus on top of all of that ive gotten one of my horrible kidney infections. I had then so bad when i was a baby i almost didnt make it. They SUCK! Ive already gone through like half a gallon of cranberry juice in like an hour. The whole week was pretty good tho. i got to go shopping agian and get some stuff for my trip. Watched the finale of Friends last night and it turned out just as i had hoped,cant say the same thing for The OC but whatever! lol Jeanette and i tried tanning out on the trampoline on tuesday and it was so nice for like 20 minutes unitll the wind came out and it got REALLY cold. We also died my hair,we got bored and put bright red streaks in it, but of course i have blonde hair and after i washed it they turned orange! GREAT! so now im walking around looking like a mupet.They canceled Days of Our Lives today b/c of some stupid presidential thing,that really ticked me off,it was the one good thing i had going in my day and it was canceled. I also watched Tuck Everlasting for the first time after owning it for almost a yr, and  I really liked it up untill the end when she dies and dosent drink from the spring.Now okay, if you had the choice to drink from a spring that can give your everlasting life would you turn it down? NO!!! Okay ive rambled on enough. Have a great weekend! Kayla  '),(4115,2967519,'01,May,2004','2004-05-01 00:00:00','   Jeanette came over on friday after i forget to pick her up at the bus stop and she had to walk like 4 miles down our road, but then we went 4 wheeling and had a BLAST getting all muddy and wet going into unknown areas out behing our house. Great time. So when we got back we had to wash everything like 3 times to get all of the mud out and just finished not too long ago. Im really tired and so im going to veg out the rest of the weekend and watch movies and eat junk food. YAY! lol Wednesday were all going to town agian and jeanette is going to skip school and come with us. She has such cool parents to let her do that. Im getting my hair cut then too and hopefully going shopping. So great weekend so far! Hope you all are having the same! Peace(ha ha ive always wanted to see if i could say that and pull it off lol did i?)  '),(4116,1298865,'07,July,2004','2004-07-07 00:00:00','    Well, all in all things are going well. Day two went well yesterday and I tried to report it to my journal on www.BFL101.com but it is down. Since Monday a.m. I\'ve dropped 4 1/2 lbs of water weight (who\'s kidding who here - I KNOW it is water) but weight is weight and it feels good. I\'ve eaten clean for two whole days (except for a nip of Brandy last night) and it feels terrific.  Yesterday I rode my bike indoors on the windtrainer for a whole 20 minutes! I can\'t wear my regular brace when I do that because it only gives my 90deg of movement, and I need about 115 to ride. So I wore the tensor, which seemed to be enough. And I iced it down after.  Today is a weights day. I will get going after I drop the kids off.  Kyle had a friend over yesterday and they spent hours on the computer going to funny links. A couple are great star trek TNG clips. I\'ll have to try to find them.  LCS   '),(4117,1298865,'05,July,2004','2004-07-05 00:00:00','    Which I already knew, of course, but I am feeling pretty sorry for myself. Three and a half weeks ago I tore ligaments in my left knee, effectively curtailing most fun for the next several months. PMO!!! Anyway - the good news is that the bruising has now faded, the swelling is down, I can walk further than I could even a couple days ago. The bad news is that I am still wearing this silly brace and using a cane. The physiotherapist said I \"did a real number\" on my LCL and illotibial band (sp?) and can expect the healing process to take a while. It could have been worse right? I could have broken my leg or completely torn the LCL. It could have been my right leg and I wouldn\'t be able to drive.   So much for paddling this summer. At this point I still can\'t bend my knee far enough to get or out of a boat. I CAN still do weights though and I had a short somewhat successful trial on my indoor bike a couple days ago. I walked around the bloack today too - not exactly the 10km run I wished I was doing, but a few days ago I could only make it halfway around and today the whole loop was easy. Progress, progress.  My plan for tomorrow is to get up and ride my bike (indoors again) after I do my physio exercises.   '),(4118,4132315,'14,August,2004','2004-08-14 00:00:00','   Message: Peter and Tina are sitting in the park >>> >doing >>> >nothing, but just gazing into the sky, while all >>> >their friends are having fun with their beloved >>> >half. >>> > >>> >Tina: I\'m so bored. Just wish I have a boyfriend >>> >now to spend time with. >>> > >>> >Peter: I guess we\'re the only leftovers. We\'re >>> >the only person who isn\'t with a  date  now. (both >>> >sigh n silence for a while) >>> > >>> > >>> >Tina: I think I have a good idea. Lets play a >>> >game >>> >Peter: Eh? What game? >>> > >>> >Tina: Eem..It\'s quite simple. You be my >>> >boyfriend for 100 days and I\'ll be your >>> >girlfriend for 100 days. what do you think? >>> > >>> >Peter: Oookay..Anyway I don\'t have any plan for >>> >the next few months. >>> > >>> >Tina: You sound like you aren\'t looking forward >>> >to it at all. Cheer up. Today will be our first >>> >day and our first date. Where should we go? >>> > >>> >Peter: What about a movie? I heard that there is >>> >a really great movie in theater now. >>> > >>> > >>> >Tina: Seems like I don\'t have any better idea >>> >than this. Lets move. (went to watch their >>> >movies and sent each other home) >>> > >>> >Day 2: >>> >Peter and Tina went to a concert together, and >>> >Peter bought Tina a keychain with a star. >>> > >>> >Day 3: >>> >They went  shopping  together for a friend\'s >>> >birthday present. Share an ice-cream together >>> >and hugged each other for the first time. >>> > >>> >Day 7: >>> >Peter drove Tina up onto a mountain and they >>> >watch the sunset together. When the night came >>> >and the moon glowed, they said sat on the grass >>> >gazing at the stars together. A meteor passed >>> >by. Tina mumbled something. >>> > >>> >Day 25: >>> >Spend time at a themepark and got onto >>> >rollercoasters, and ate hotdogs and cotton >>> >candy. Peter and Tina got in the haunted house >>> >and Tina grabbed someone\'s hand instead of >>> >Peter\'s hand by accident. They laughed together >>> >for a while. >>> > >>> >Day 67: >>> >They drove pass a circus and decided to get in >>> >to watch the show. The midget asked Tina to play >>> >a part as his assistant in the magic show. Went >>> >around to see other entertainments around after >>> >the show. Came to a fortune teller and she just >>> >said \"Treasure every moment from now on\" and a >>> >tear rolled down the fortune teller\'s cheek. >>> > >>> >Day 84: >>> >Tina suggested that they go to the beach. The >>> >beach wasn\'t so crowded that day. They have >>> >their first kiss with each other just as the sun >>> >is setting. >>> > >>> >Day 99: >>> >They decided to have a simple day and is >>> >deciding to have a walk around the city. They >>> >sits down onto a bench. >>> > >>> > >>> >1:23 pm >>> >Tina: I\'m thirsty. Lets rest for a while first. >>> > >>> >Peter: Wait here while I go buy some drinks. >>> >What would you like? >>> > >>> >Tina: Eem...Apple juice will be just fine. >>> > >>> > >>> >1:43 pm >>> >Tina waiting for about 20 minutes and Peter >>> >havent return. Then someone walked up to her. >>> > >>> >Stranger: Is your name Tina? >>> > >>> >Tina: Yes, and may I help you? >>> > >>> >Stranger: Justnow down there on the street a >>> >drunk driver has crashed into a guy. I think its >>> >your friend. >>> > >>> >Tina ran over to the spot with the stranger and >>> >sees Peter lying on the floor with blood over >>> >his face and her apple juice still in his hands. >>> >The ambulance came and she went to the hospital >>> >with Peter. >>> >Tina sat outside the emergency room for five and >>> >a half hours. The doctor came out, and he sigh. >>> > >>> > >>> >11:51 pm >>> >Doctor: I\'m sorry, but we did the best we could. >>> >He is still breathing now but God would take him >>> >away from us very soon. We found this letter >>> >inside his pocket. >>> > >>> >The doctor hands over the letter to Tina and she >>> >goes into the room to see Peter. He look weak >>> >but peaceful. Tina read the letter and then she >>> >burst into tears. Here is what the letter said. >>> > >>> >Tina, our 100 days is almost over. I had fun >>> >with you during all these days. Although you may >>> >be greedy sometimes and less thoughtful, but >>> >these all brought happiness into my life. I have >>> >realize that you are a really cute girl and >>> >blamed myself for never taken the time to >>> >knowing that. I have nothing much to ask for, >>> >but I just wish that we can extend the day. I >>> >want to be your boyfriend >>> >forever and wish that you can be beside me all >>> >the time. Tina, I love you. >>> > >>> >11:58 >>> >Tina: (sobbing) Peter. Did you know what was the >>> >wish I made on the night there was a meteor. I >>> >asked God to let us last forever. We were >>> >suppose to last 100 days so Peter! You can\'t >>> >leave me! I LOVE YOU, but can you come back to >>> >me now? I love you Peter. I LOVE YOU. >>> > >>> >As the clock struck twelve, Peter\'s heart >>> >stopped beating. It was 100 days. >>> > >>> >NOTE* >>> >Tell the guy or girl that you love them before >>> >its too late. You never know whats going to >>> >happen tomorrow. You never know who will be >>> >leaving you and never return '),(4119,4132315,'09,August,2004','2004-08-09 00:00:00','   Date of Birth Group 01,06,11,16,21,26,31                 A1 02,07,12,17,22,27                      B2 03,08,13,18,23,28                      C3 04,09,14,19,24,29                      D4 05,10,15,20,25,30                      E5 Group A1 You consider love as the most beautiful thing and you love to fall in love, but a number of your group members think that the person they love is not sincere with them. You love to be with your friends and you are always found to be a dutiful friend. You can hardly control your patients, which is a very big drawback of your nature. The Person who is ruling your mind & heart these days is perfect for you.    Group B2 Your dreams and ambitions are much important for you and you can do everything to fulfill your dreams. Love is much valuable in your life but you always search for someone perfect. You hardly trust someone. Your friends are really important for you but normally you hide a lot from them. You are a deep thinker you always study the negative view as well as positive. You can lead a happy life with a person for whom you care a lot these days.    Group C3 You always prefer mantel decisions more then emotional ones due to this you don\'t have a limited number of friends. You consider life a very beautiful gift and you love to enjoy its colors. There are a number of peoples who are your ideals and you loved to spend a big part of your time with them. You are found to be a very sincere lover. You have a perfect control on your emotions but sometimes your decisions really effect your beloveds. The person who has just appeared in your mind and you has decided to forward this mail especially to him/her is your real and special friend.    Group D4 Your always have goals to achieve and you can do everything to fulfill the dreams of those who loves you. You have a sketch of an ideal in your mind and you always search for that personality. Your friends means a lot to you and you can do everything for the sake of your friend, you a normally found to be an emotional personality. You have a very less control on your patients and due to this sometimes you over react. The one who sent you this mail &amp; the one to whom you will send this mail first are your real friends.    Group E5 You are found to be a person who love to love. You prefer emotional decision more then mantel decisions. You consider life just to enjoy; you are the one who is perfect to call FLURTIES. You love to increase the list of your friends and beloved???s. You have a number of dreams but you never work hard to make your dreams come true which is the biggest drawback in your nature you take everything much lightly. '),(4120,4132315,'07,August,2004','2004-08-07 00:00:00','   As I sat there in English class, I stared atthe girl next to me. She was my so called \"bestfriend\". I stared at her long, silky hair, andwished she was mine. But she didn\'t notice melike that, and I knew it. After class, shewalked up to me and asked me for the notes shehad missed the day before and handed them toher. She said \"thanks\" and gave me a kiss on thecheek. I wanted to tell her, I want her to knowthat I don\'t want to be just friends, I  love  herbut I\'m just too shy, and I don\'t know why.11th gradeThe phone rang. On the other end, it was her.She was in tears, mumbling on and on about howher  love  had broke her heart. She asked me tocome over because she didn\'t want to be alone,so I did. As I sat next to her on the sofa, Istared at her soft eyes, wishing she was mine.After 2 hours, one Drew Barrymore movie, andthree bags of chips, she decided to go to sleep.She looked at me, said \"thanks\" and gave me akiss on the cheek. I want to tell her, I wanther to know that I don\'t want to be justfriends, I  love  her but I\'m just too shy, and Idon\'t know why.Senior yearThe day before prom she walked to my locker. \"Mydate is sick\" she said; he\'s not going to gowell, I didn\'t have a date, and in 7th grade, wemade a promise that if neither of us had dates,we would go together just as \"best friends\". Sowe did. Prom night, after everything was over, Iwas standing at her front door step! I stared ather as she smiled at me and stared at me withher crystal eyes. I want her to be mine, but sheisn\'t think of me like that, and I know it. Thenshe said \"I had the best time, thanks!\" and gaveme a kiss on the cheek. I want to tell her, Iwant her to know that I don\'t want to be justfriends, I  love  her but I\'m just too shy, and Idon\'t know why.Graduation DayA day passed, then a week, then a month. BeforeI could blink, it was graduation day. I watchedas her perfect body floated like an angel up onstage to get her diploma. I wanted her to bemine, but she didn\'t noticeme like that, and I knew it. Before everyonewent home, she came to me in her smock and hat,and cried as I hugged her.Then she lifted her head from my shoulder andsaid, \"you\'re my best friend, thanks\" and gaveme a kiss on the cheek. I want to tell her, Iwant her to know that I don\'t want to be justfriends, I  love  her but I\'m just too shy, and Idon\'t know why.A Few Years LaterNow I sit in the pews of the church. That girlis getting married now. I watched her say \"I do\"and drive off to her new life, married toanother man. I wanted her to be mine, but shedidn`t see me like that, and I knewit. But before she drove away, she came to meand said \"you came!\". Shesaid \"thanks\" and kissed me on the cheek. I wantto tell her, I wanther to know that I don\'t want to be justfriends, I  love  her but I\'m just tooshy, and I don\'t know why.FuneralYears passed, I looked down at the coffin of agirl who used to be my \"best friend\".At the service, they read a diary entry she hadwrote in her high school years. This is what itread:\"I stare at him wishing he was mine, but hedoesn\'t notice me like that,and I know it.I want to tell him,I want him to know that I don\'t want to be justfriends,I  love  himbut I\'m just too shy,and I don\'t know why.I wish he would tell me he loved me!I wish I did too...I thought to my self, and I cried.I  Love  UI  Love  UI  Love  U...\" '),(4121,4132315,'04,August,2004','2004-08-04 00:00:00','   FrOm tHe vErY dAy ThaT wE wErE bOrn, gOd hAs AlrEadY dEsiGneD tHe PaTh tHaT wE\'rE InTenDeD tO tAke. NoBoDy ReaLLy KnOwS AnYtHiNg AbOuT ThE FuTUrE. It\'S LiKe We\'Re PlaCeD In a MaZe wHeRe In, We\'Ve tO LooK fOr tHe WaY Out. BuT thEn, aGaIn iT\'s nOt GoInG tO rUn OuT oF tImE aNd wE\'LL JusT ReAliZe ThAt wE HavEn\'T eVeN ReAcHeD anYthInG YeT. It\'S jUsT sO DePreSSiNg tO hAvE ThOugHt wHaT wE HaVe AlL tHe tIme In tHE WoRlD aNd eNd Up wItH nOtHinG bUt ReGreT ....... '),(4122,4132315,'04,August,2004','2004-08-04 00:00:00','   When you are together with that special someone, you pretend to ignore that person. But when that special someone is not around, you might look around to find them. At that moment, you are inlove. Although there is someone else who always makes you laugh, your eyes and attention might go only to that special someone. Then, you are in love. Although that special someone was supposed to have called you long back, to let you know of their safe arrival, your phone is quiet. You are desperately waiting for the call! At that moment, you are in love. If you are much more excited for one short e-mail from that special someone than other many long e-mails, you are in love. When you find yourself as one who cannot erase all the messages in your answering machine because of one message from that special someone,you are in love. When you get a couple of  free  movie tickets, you would not hesitate to think of that special someone. Then, you are in love.You keep telling yourself, \"that special someoneis just a friend\", but you realize that you cannot avoid that person\'s special attraction. At that moment, you are in love. While you are reading this page, if someone appears inyour mind, then u are in  love  with that person... -nD- '),(4123,4132315,'04,August,2004','2004-08-04 00:00:00','   HOW TO TELL IF A GUY LIKES YOU: 1. The guy will be extremely nice to you. 2. He will tell you that you did something good, even when you did it horribly. 3. He might make fun of you. 4. He will want to be your best friend. 5. He might complement you on something normal like, your hair, even if you wear it that way everyday. 6. He will stick up for you. 7. He will start hanging out with your friends. 8. He will flirt with you. 9. He will call you for no good reason. 10. He will make eye contact with a serious look on his face. HOW TO TELL IF A GIRL LIKES YOU: 1. She always talk about the different kind of guys she has. 2. She stare at you with a smile on her face and won\'t look away until you do first. 3. She always seem to be talking about how nice or cute you are. 4. She laugh at all your jokes, no matter how stupid they are. 5. She will ask you who you like, continuously. 6. She talk to your friends about you a lot. 7. She always flirt with every other guy, except you. 8. She always try to make you jealous. 9. She beg you to do everything for them. 10. She always ask you what to do in a bad situation. '),(4124,3889796,'31,July,2004','2004-07-31 00:00:00','    well today was quite interesting to say the least... had a french test...quite hard wif this song that\'s now stuck in my head...then had ace, where we sorted out the teams for the qualifying problem in fps that\'s future problem solving...then had electronics...where we did literally nothing...except talk...and then maths...10 quick questions as usual...e.g. what\'s the capital of new jersey? well its not the big apple...its trenton...  then we had soccer training...played a game...and we won...with praw, johannes, and some others..like this new boarder guy justin...he\'s pretty gd...  anyway catcha later   '),(4125,3889796,'24,July,2004','2004-07-24 00:00:00','    here\'s another installment of things of the week  song: heaven - los lonely boys  quote: you have 30 seconds or in other words 30 billion nanoseconds..plenty of time - einstein  site: www.acronymfinder.com - find all those annoying TLA\'s.  word: supacallyfragalisticexpiallydocious  joke: why did fred fall off his bike? coz his mum threw a fridge at him.  here\'s another 1 i heard from garry  why did the plane crash? coz the pilot was a loaf of bread...   '),(4126,3889796,'24,July,2004','2004-07-24 00:00:00','    this week has been pretty interesting...first week bak always is... had the westpac...did really bad in that...anyway we had a badminton comp in p.e. and i vs prawi and he beat me 11-10, so damn u praw...anyway the a\'s (that\'s my team) got thrashed about 7-3... anyway hope 2 do betta next time...yesterday we played hale and we lost...again! 3-1 this time....and mr hess didnt\' coach us so mr de bes (some athletics dude) did...and we played pretty well...but then in the 2nd half we did really bad and we lost...  here\'s my favourite songs rite now:  1. pieces of me by ashlee simpson 2. leave (get out) by jojo 3. spiderman by michael buble 4. my happy ending by avril lavigne 5. super duper love by joss stone 6. bail me out by pete murray 7. she will be loved by maroon 5 8. that girl by miranda murphy 9. learn to fly by shannon noll 10. the morning after by zinc   '),(4127,3889796,'24,July,2004','2004-07-24 00:00:00','    well me and praw hav started a riding thingy jus like the badminton thing we started as well...well we just ride around the river wen we get time...this time we went around the long-way...that\'s from the south perth foreshore to the graham farmer bridge in burswood, winding through east-perth, outside trinity, on the perth foreshore on riverside dve, then on the narrows bridge and back to the south perth foreshore...we normally hav 2 routes the long way and short...roughly the record for the short was is just less than half an hour and for the long way, which we\'ve only done once is 41 minutes...  we\'ve been trying to go early in the morning to try keep cool and to ride while there are less ppl on the roads etc...   '),(4128,3889796,'22,July,2004','2004-07-22 00:00:00','    song of the week: is she will be loved by maroon 5  word of the week: winchestonfieldville/bitcheroni...off mr deeds...if u\'ve seen it  quote of the week: \"well i tried my hardest and threw the kitchen sink at him but roger went to his bathroom and got his bath tub\" - andy roddick after losing the wimbledon final to roger federer...  site of the week: www.bonzi.com, this is a really weird and kinda lame monkey kinda-desktop buddy...it can say anythin u type but its not always accurate so don\'t rely on it and it has some amazing facts and really lame jokes e.g. where does the dog go to after its lost its tail? to the retailer...lame...huh? anyway some ppl mite like it so check it out if u hav time...  joke of the week: is comin\' soon...i\'ll find 1   '),(4129,3889796,'22,July,2004','2004-07-22 00:00:00','    today\'s been interesting...pretty exciting...right now...yeh doin stuff  anyway i gtg,  cya later guys   '),(4130,3889796,'20,July,2004','2004-07-20 00:00:00','    hey again,   here\'s the things of the week:  damn u praw for copyin\' me!  song of the week: pieces of me by ashlee simpson  quote of the week: Be sincere, be brief, be seated.  word of the week: goolgle  there u go...gimme some more suggestions 4 thing\'s of the week...   '),(4131,3889796,'11,July,2004','2004-07-11 00:00:00','    hey, today\'s been pretty boring. i was gonna go and play badminton with prawi, but the courts was packed with ppl...so yeh we just played soccer on ward oval instead... anyway i\'m at prawn\'s house now and i\'m pretty bored... so yeh cya later   '),(4132,3889796,'10,July,2004','2004-07-10 00:00:00','    hey again, i jus finished practising clarinet and now i\'m jus downloading some songs...its pretty boring..so i think i\'ll be going soon...anywa on my previous post....the date was satuday and its meant to be today...so disregard that...its meant to be monday the 12th of july 2004...  so yeh...  here\'s the songs i\'m listening to now:  bail me out - pete murray mercury4,  maroon 5\'s newie j-kwon   '),(4133,3889796,'10,July,2004','2004-07-10 00:00:00','    hey guys,  today\'s been a comme ci, comme ca day, really bitta french for ya... anyway the day started in the early morning when most of u guys would\'ve been asleep...as normally i would be as well...but i was watching the british f1 gp from bout 11pm till 12:45am today...so yeh... anyway wen i got up i watched it again on tape and checked out some of the highlights and lowlights...in case u haven\'t heard michael schumacher won for the 10th time this yr out of 11 races from johannes\' fav driver kimi raikkonen and 3rd was schumacher\'s ferrari team-mate rubens barrichello.... the highlight was definately jarno trulli crashing and ending up with a canoe after exiting bridge corner at 250kph...  anyway to other stuff...in the afternoon i went 2 praw\'s house and me, him and michael sherry played soccer...and leonard played as well instead of his usual swimming training....  and that\'s bout it  no1 this week is learn to fly by shannon noll...  if u want the full charts go to www.aria.com.au and if u want lyrics go to www.musicmademe.com or www.lyrics.com...both r cool,though i suggest musicmademe for a more extensive search...  anyway that\'s my day...   '),(4134,3889796,'10,July,2004','2004-07-10 00:00:00','    hey, today\'s been a boring day...did jack... anyway in the last few days...i did nothin... rented out some dvd\'s... like master and commander, lost in translation, ocean\'s 11, and some others...rode round the river couple of times... doin some homework from teachers... anyway here\'s a good song... its called pieces of me by ashlee simpson (lil sister of jessica) and if any of u guys watch 7th heaven...ashlee plays cecelia...the girl that simon likes...anyway its a pretty good song...i heard it ages ago on rick dees and the weekly top 40...but that\'s been replaced now by AT40 with ryan seacrest...anyway another good song is 8th world wonder by kimberley locke...she came 4th i think in american idol 2...so check both of them out if u hav time or if ur lucky enough 2 come across it..  anyway cya guys later...  peter   '),(4135,3889796,'10,July,2004','2004-07-10 00:00:00','    hey guys, this is my new blog so check it out if u hav time... and i\'ll try update it as often as i humanely can thx, peter   '),(4136,3889796,'13,August,2004','2004-08-13 00:00:00','    well 2day...as it is friday...we had another game of soccer, this time against guildford...well it was another disappointing game coz we lost...again... anyway here are the highlights: sukhum was bout 25 metres from the goal and he the ball came 2 him and somehow he booted it backwards into sheahan\'s goal...then liam (the ex goalie) went to clear it and it came flying off the side of his boot and he almost bounced into the goal...so that\'s a own goal and another very close shave...then michael ierace shoulda been sent off...coz he just tackled this guy really roughly wen he tried to get the ball...and the guy chucked a huge psycho at him...and they almost started fighting...luckly ierace got taken off just after that...then there was sam missing quite a few gd oppurtunities in front of goal...especially this one just be4 half time...where it was a absolute sitter....anyway in the end it was 2-1 and jaron scored this hell gd goal...  well betta luk next week...   '),(4137,3889796,'12,August,2004','2004-08-12 00:00:00','    well 2day....we did nothin as usual...1st was ace...academic enrichment, where we did a play 4 tournament of minds....and then we had health....where mr reidy rambled on as usual...right now i\'m in maths where i\'m studying about sin, cos and tan...but i\'m actually not doin it as u...next i\'ve got sose...then sport against guildford in soccer...  anyway i\'m goin 2 do some work..  cya   '),(4138,3889796,'12,August,2004','2004-08-12 00:00:00','    2day....well today 1st i had english and we did nothin except read...and mr deegan (my p.e) teacher was our relief teacher 2day...then we had p.e. and i played prawi but it wasn\'t really competitive...it was just a session where we showed the teacher our skills 4 an evaluation... then afta recess...i had electronics....where we did nothin as usual...then was sose, where we watched a video and heard some really bad singing....then afta tutor, i had house badminton...where i teamed up wif danny and we beat a hell gd team of leonard and this yr 10 guy sam...16-14....whoa...what a relief....and right now i\'ve just wagged biology and i\'m in the wyvern centre doin this thing for tournament of minds...  anyway cya later guys...   '),(4139,3889796,'11,August,2004','2004-08-11 00:00:00','    well starting on sunday....i played badminton at wesley wif prawn, danny and mo...then afterwards me and danny went to mo\'s house and played soccer....we played some randoms... then i got home at about 5, on monday we didn\'t do anythin special....except a maths investigation....which was pretty hard and also a french test...then we had soccer training....which was the a\'s vs the b\'s then yesterday we didn\'t do nothin...except we started human bio for science....  anyway 2day we\'ve got a debate....  p.s. will and shu say hi   '),(4140,3889796,'06,August,2004','2004-08-06 00:00:00','    well this aftanoon, we played soccer against trinity...at trinity....it was a pretty bad game in general...and in the end, our team (the b\'s) lost 2-1, and if any1 out there eva vs trinity again, there\' a bunch of idiots....coz they cheat so much and i\'m not just sayin that coz we lost...they really are...all the 3 yr 9 teams lost...which was pretty crap....now i\'m in the library....wif praw and josh allen sending me unlimited net send messages....  anyway i\'m goin now cya guys later   '),(4141,3889796,'05,August,2004','2004-08-05 00:00:00','    well this mornin\' wen i got to skewl i was really pissed coz i found out that i didn\'t bring my soccer shirt...damn! well 1st was badminton, and i played A-1 as usual and i won both my singles games against Graham and Bradie 15-10 and 14-5, then me and aaron chew played the doubles and we won like 16-4, against graham and bradie....so i won all 3 games i played in, but overall my team   lost 4-3, which means that we were winless through the whole comp....that\'s pretty crap, but at least i won all my games 2day which was gd,  anyway jus had electronics where we did the usual of nothin\' next i\'ve got maths in the library...where we muck around for a hour or so... then sose where we\'ve goin 2 do some map skills...  then we play the usual game, where 2day we\'ve vs trinity in soccer   '),(4142,3889796,'05,August,2004','2004-08-05 00:00:00','    hey...last nite we had debating and we lost...against pc...neil had the longest speech in history...7:02, and it was hell funny....and pc were pissing themselves laughing....and so were we...that\'s me and sheahan....and brodie totally lost it   anyway we lost and that was bad....but apart from that it was a gd debate coz neil came up about all the macca\'s characters who were luring kids to eat at macca\'s and how hamburgler was promoting kids to be crooks ....praw came 2 watch and support us....that\'s the 1st yr 9 that\'s eva come 2 watch us...aftawards we wanted 2 celebrate our loss wif a trip to macca\'s  anyway we had badminton this morning in p.e. and i won my 1st match against arran parker 16-15...anyway then i played the d\'s number 2   and i lost 16-8...damn...oh well my team the a\'s lost again....  next i\'ve got geography and english....that\'s bout it....   '),(4143,3889796,'04,August,2004','2004-08-04 00:00:00','    Okay, here\'s what you\'re supposed to do. Copy this e-mail and paste it onto a new e-mail that you will send. Change all the answers so that they apply to you, then send this to a whole bunch of people you know INCLUDING the person who sent it to you. The theory is that you will learn a lot of little facts about your friends. Starting time : 13:40pm Full Name: Peter Xiao Nickname(s): pedro, eziao, pete, shower power, pedor, petor Birthday: 21/02/1990 School: Wesley College Colour of eyes: dark brown Hair colour: black Height: 170 cm >------HAVE YOU EVER--------- > > Ever been so drunk you black out?: no way Put a body part on fire for amusement: errr...burnt with a soldering iron... Been hurt emotionally: yep Kept a secret from everyone: nope Imaginary Friend?: nah Cried during a movie: no way Had a crush on a teacher: hell no Have you ever been on stage: yeh every wednesday Cut your hair yourself: errr...y would i? Been sarcastic: ummmm...err u guyz should know >---------FAVORITES------------- > > Shampoo: anythin Color: green/red Day/Night: night Summer/Winter: summer Lace or satin: what\'s this? Cartoon Character: fry and lila off futurama Food: watevea 1 get Advertisement: anythin Movie (s): spiderman 2 >---------RIGHT NOW------------ > > Wearing: skewl uniform Eating: salami sandwich Drinking: orange juice Watching: the comp screen in the wyvern centre Listening: ppl yell around ideas for fps >---------IN THE LAST 24 HRS-------- > > Cried: no Worn a skirt: no Met someone new: no Kissed: my mum Done Laundry: was gonna but stuff that  >---------DO YOU BELIEVE IN------- > > Yourself : yes Friends: absolutely Santa Clause: nup Tooth Fairy: nope Destiny/Fate: yes GOD: yep Angels : yes r they the guyz with wings?  >--------FRIENDS AND LIFE-------- > >  Like anyone?: yeh Who\'s the weirdest: matthew carter Who do you go to for advice: family and friends Who do you cry with?: me, myself and i When you last cried?: hey gd ? What\'s the best feeling in the world: love What\'s the worst: despair A wish: i am successful life and career in wateva i do Finish Time: 2:55 (i kinda stopped and started)   '),(4144,3889796,'03,August,2004','2004-08-03 00:00:00','    hey...today\'s been a normal day really....except that i\'ve got a french test next and i don\'t know anythin and we got our results for out biology medium test and i passed....c\'mon! in english we did nothin...except read anne frank\'s diary....then garry acted like a fully nerd coz the library teacher didn\'t like him and he had 2 go borrow a book and it was hell funny...well apart from that, that\'s about it up 2 now...      cya later   '),(4145,3889796,'03,August,2004','2004-08-03 00:00:00','    ok this is really blown outta proportions....johannes has this crazy obsession with me and my favourite f1 driver kimi raikkonen....ok let\'s stop this mess now...  p.s. johannes and prawi are single (so they tell me)  anyway cya later guyz   '),(4146,3889796,'02,August,2004','2004-08-02 00:00:00','    today we had a biology medium test and i think i passed...hopefully...then i had a french reading test which was pretty hard if u learn french...english we did nothin....except borrow a book about the 2nd World War...next i\'ve got health....then finally electronics...then afta skewl i\'ve got 2 do a fps thing where i go into lockdown for 2 hours...  and now i\'m in the com labs doin nothin...  anyway i gtg now,  cya all later   '),(4147,3889796,'02,August,2004','2004-08-02 00:00:00','    here\'s another installment of things of the week  song: heaven - los lonely boys  quote: you have 30 seconds or in other words 30 billion nanoseconds..plenty of time - einstein  site: www.acronymfinder.com - find all those annoying TLA\'s.  word: supacallyfragalisticexpiallydocious  joke: why did fred fall off his bike? coz his mum threw a fridge at him.  here\'s another 1 i heard from garry  why did the plane crash? coz the pilot was a loaf of bread...   '),(4148,3532367,'30,June,2004','2004-06-30 00:00:00','   We went to see the girls get their awards tonight from their Club. It was good to see their hard work get rewarded by the presentation of badges and certificates. They had good fun and it\'s given them a late night. We are all pretty much shattered, me especially and I could do with some decent sleep.  '),(4149,3532367,'29,June,2004','2004-06-29 00:00:00','   Well as I type, there are big arguments going on in the office about annual leave. It doesn\'t effect me directly as I booked mine earlier in the year, but I really feel for the guy who it does effect. The atmosphere at the moment is painfully tense and after having no sleep, its quite difficult to cope with. Unfortunately, \"new\" rules were brought in at the beginning of the year. The only problem is, no-one consulted us and they are only local (effecting only our department).  '),(4150,3532367,'29,June,2004','2004-06-29 00:00:00','   I had half a day annual leave yesterday to attend a meeting in Stafford about computer strategies. Following that I went straight to football practise, then once home, it was time for tea and housegroup. Busy, busy, busy.  By 22:00, I was totally shattered and went to bed looking forward to getting loads of sleep. How wrong I was - another night of disturbed sleep. It\'s not because I\'m lying awake thinking about things though. I just can\'t sleep. I\'m sure it\'s to do with the medication, but the doctor won\'t give me anything at this stage, preferring to get me to exercise. The theory is all well and good, but today I am totally shattered and can\'t concentrate on my work at all....Oh well!  '),(4151,3532367,'29,June,2004','2004-06-29 00:00:00','   Morning world! And what a beautiful day it is today!  Yes you are right, it is 05:00. I woke at midnight, 02:00, 02:30 and 04:00. Don\'t know why. By 04:30 I decided enough was enough and I ventured out for my first \'morning\' run. While I was getting changed, Sandra woke and asked what I was doing. I told her I was going for a run....unless she wanted sex. She was silent and I took that to be a \"no\"!  I jogged up to the local football pitch and did a couple of circuits, with short bursts of \"running\", before jogging back. Half an hour in total, but more than enough for me to experience bursting lungs, aching legs and a sweaty back....nice.  And just to prove the point, I decided to blog it now, rather than later ;-) I dont feel tired though....  '),(4152,3532367,'28,June,2004','2004-06-28 00:00:00','   I got paid today, so I have given Sandra details of how to access my accounts and what money to move where so that the bills are paid. The money hasn\'t come through from the bank yet, but we seem to be coping quite well financially at the moment (either that or the pressure has been taken off me by Sandra looking after the finances).  '),(4153,3532367,'27,June,2004','2004-06-27 00:00:00','   I managed to get the car serviced tonight. Changed the oil, plugs, filters and rocker cover gasket. It took the best part of an hour and it went quite smoothly. I feel a lot happier now that it\'s out of the way and I have more of an idea of the state of the engine. The parts cost 33 UKP and Sandra is due to pay for those tomorrow.  '),(4154,3532367,'27,June,2004','2004-06-27 00:00:00','   I\'ve made a conscious decision to avoid Sharon. I\'ve put her number on ignore and her messenger and email on block. I\'ve got nothing against her on a personal level, I just need to avoid creating further damage to myself and my family.  I don\'t really know what is worse - having sex with someone else behind Sandra\'s back, just for the sake of \"needing\" a sexual high or continuing in the same manner as we are with each other. I\'ve chosen at the moment to remove the third party. Let\'s see if I can stick to it.  '),(4155,3532367,'27,June,2004','2004-06-27 00:00:00','   Today seems to be going so slowly. I got in a bit late due to heavy traffic, I\'ve already eaten most of my lunch, even though I had breakfast before I came out (which is unusual for me). I would just rather be somewhere else, but I suppose most people feel like that when they are at work.  '),(4156,3532367,'27,June,2004','2004-06-27 00:00:00','   What a great afternoon - not!   We took the kids on a bike ride, after loads of nagging from them. We literally got half way down the hill and my youngest daughter fell off. Fortunately, she had a helmet on but she was badly damaged on her hands, chest and face. A passer-by gave us a lift back, by putting the bikes in his van and taking the casualty home - Sandra went with her. As soon as I had arranged for someone to look after the other kids, we took her to hospital with a suspected broken finger and to get cleaned up.  Fortunately, it transpired that nothing was broken, although she was very sore. The hospital was quite quick, only 2 hours start to finish but by the time we got home it was raining, so the idea of a b-b-q had to go out of the window.... we settled for a chinese instead.  '),(4157,3532367,'26,June,2004','2004-06-26 00:00:00','   The buyer came to collect my ebay item at lunchtime. Lovely lady, paid in cash and I gave the 25 UKP straight to Sandra for her to put towards the parts for my car service (due to pick up parts on Monday). The kids want to go on a bike ride this afternoon (I feel knackered already!)  '),(4158,3532367,'26,June,2004','2004-06-26 00:00:00','   Well I started off this morning with a 13 mile bike ride. I really enjoyed it as the weather was great. Sandra had sent me out as I had started to pester after she had woken up. We had had sex the night before (usual routine on a towel) but she seemed more accepting of it, although again I had felt pressured to go ahead. We finally went to sleep at about 01:00, but I was awake again at 04:00 and couldn\'t really get back to sleep. It was very difficult trying to keep to myself, but I couldn\'t get myself up and out of bed to start any execrcise either.  When I got back from cycling, it was time for the chores - a little vaccuming and time out to mow the lawns. Just before lunch, we had a cuddle and I ended up undressing Sandra, but stopped short of full sex. I decided to distract myself with a little more gardening and Sandra came out and helped me.  '),(4159,3532367,'26,June,2004','2004-06-26 00:00:00','   Sandra suggested we went out for a meal tonight. After some discussion we settled on a place, where we thought the prices would be reasonable. It was the first time we had been to a \"Harvester\" pub, but we really enjoyed the layout and the food. We had a really good time and for less than 25 quid all in. Sandra said that we should put it on my card for now, which I did.  On the way back, Sandra said she wanted to do the food shopping. I had my reservations, but she was understanding to a point and offered to drop me off at home. We discussed it further and I said that I would have to face up to it at some point. I got a couple of things on 3 for 2 offers, but Sandra seemed to be OK with them.  '),(4160,3532367,'25,June,2004','2004-06-25 00:00:00','   After having dropped the girls off so that they could go on their organised day trip, I went down to the bank to pay some of my Saturday job money into the savings account - along with the refund, I am now back over the 1000 UKP mark.  Next up was the cycle shop - it was open, they fixed my gears while I waited and charged me 4 quid! Bargain! Even if he did close early on me yesterday.   From there I got home and found the batteries had gone flat in a piece of equipment I use. I asked Sandra if she would fetch some for me as I had to be out in 30 mins and still needed to get changed - she agreed. 40 minutes later, still no sign of Sandra. I told my son to jump in the car and phoned work to ask them to cover me as I would be 15 minutes late. Half way to work, the phone goes, it\'s Sandra \"where are you?\" she enquires. I tell her how I needed to get off and that she would need to come and get Rob off me. She slammed the phone down.  I arrived at work 15 minutes late, there was cover, but everything was OK. 15 minutes later Sandra turned up, handed me the batteries and took Rob home, without muttering a word.   Now I felt like the one who had been put out, I was OK about everything (even though I was flustered, I wasn\'t angry or \'down\') yet it was Sandra that was in the mood - why? OK she went out for me, but knew what time she had to be back - she had gone in plenty of time and should have done it easily.  Nevertheless, I have come back to an empty house, which means I can\'t go out and get the parts for my car service, as I have asked Sandra to only give me cash for things. As she is not here I have no money, and the shop closes in 30 mins....  '),(4161,3532367,'25,June,2004','2004-06-25 00:00:00','   Good news and bad news from the postman today. First I got a refund from a finance company I have paid off. Just under 90 UKP (I had paid them to close the account, then they took the monthly direct debit as well).   The bad news was the insurance company had charged me twice (once by card on the phone and once by direct debit) for the changes following the change of cars. Oh well, I suppose another phone call will resolve that one - its just a hassle.  '),(4162,3532367,'24,June,2004','2004-06-24 00:00:00','   Well I started off well. Got home and jumped on the bike to take it down the shop to get the gears repaired (in preperation for my exercise routine set by the doctor). I got to the cycle shop at 16:50 and it had shut - everything had been put away and there were still 10 minutes to closing time! Bastard!  Decided then it would be best not to waste the opportunity and took off down the canals - with only 6 of 21 gears working, pedalling like a mad man. Half way round, would you just know it, I got a thorn in my tyre. It had gone all the way through. I decided the best option was to leave it in so that it acted like a plug. It seemed to work, I completed the 9.5 miles without much problem and even decided to come home a different route so that I got more hills! The puncture was fixed in a couple of minutes once I had got home and the whole adventure from start to finish took less than an hour. All in all I am quite pleased with myself.  '),(4163,3532367,'24,June,2004','2004-06-24 00:00:00','   I\'ve told Sandra I would like to service my car this weekend. It\'s been standing with the previous owner and the oil has congealed at the top of the rocker cover. I asked her if she would ask her Dad if I could do it there, as it\'s flat and he has a bigger tool kit than me. I\'ve also indicated that I want to get my push bike in to get the gear cable fixed. I\'ve estimated 25 UKP for the car and 25 UKP for the bike. I am hoping she will come with me to pay so that I don\'t have an opportunity to deceive her.  I finish early on Friday\'s so I am hoping to get the bike up this afternoon as part of my exercise therapy. It\'s about 2 miles away, so if the weather stays nice, I might go in shorts and try to jog back - LOL  '),(4164,3532367,'24,June,2004','2004-06-24 00:00:00','   Well it didn\'t really seem to work. When I got into bed, Sandra offered me sex on the same basis as last night. I refused, she pushed and I refused again. It would have been so easy, but I knew there would have been a low on the other side of it.  I woke at 02:30 and again at 04:00 and again at 06:00, but I was not awake enough, so there was no way I was going to get up and go for a run. At 06:00 I started to give her cuddles and hugs which inevitably led to me trying it on, which she rebuffed. I recognised what I was doing, apologised and tried to remind her that I was trying, but it was hard for me. When we got up, I asked if she had slept OK, she said not that well and returned the question. When I told her I had woken several times, her response was \"well I gave you the opportunity\".....Aarrgh!  Sandra thinks I am using the depression label as an excuse for all of my problems, sometimes she almost seems jealous that I am getting help and she isn\'t, but she won\'t do anything about it.  '),(4165,3532367,'24,June,2004','2004-06-24 00:00:00','   Having watched England lose on penalties and with a load of pent up frustration I asked Sandra if we could have a cuddle when we went to bed. I got the \"do what you feel you need to do\" response. I said no and I didn\'t want it to be like that, to which she replied \"the damage is done, its not repairable any more\".  I decided to go for a bike ride. Texted a friend who said he would come, then decided he had no lights. I did nearly 4 miles - flat out, on hilly terrain. I am totally unfit and hate exercise of any kind so it was a feat in itself.   On the way back all I could think of was cuddling up to Sandra in bed, so on the last hill home (a big climb) I really worked it hard. I am at the point now that I can\'t type straight, I am sweating profusly and feel unco-ordinated. I hope it bloody works!  '),(4166,3532367,'23,June,2004','2004-06-23 00:00:00','   Another of those questions off the doctor and one which perhaps goes someway to answering comments. When I feel bad about something I\'ve done, does it have a stronger feeling than the high?  Well at the moment no, the high lasts for longer and is more powerful, whilst the low is very intense but may only last a couple of minutes - usually while admitting to it!!!!   Take the speakers, every time I get in the car I get a high from the music they produce as a good sharp, clear sound. That is more frequent than the low I got from hiding the boxes in the dustbin or admitting my actions to the doctor. Also the high was experienced both before and after the low and thus cancelled out the feeling of being low - does that make sense?   Sex with Sharon was the same, I was OK at the time, OK after, but felt bad when admitting it to the blog or doctor, but OK again next time I spoke to her on chat.  '),(4167,3532367,'23,June,2004','2004-06-23 00:00:00','   Today I told a friend at work about the existence of the blog. I needed some feedback on how others interpreted my problems. Lets face it, the anonymity of typing it out to a website is a damned site easier than sitting face to face telling someone. They were really good about it and it was nice to get the thoughts from an \"outside\" source.   I also got my first comment today - horray! It\'s really important to me to know that people are out there who feel the same, or are able to share experiences in this way. It makes me realise I\'m not alone and I am very grateful to everyone who feels able to comment.  '),(4168,3532367,'23,June,2004','2004-06-23 00:00:00','   Well first the good news - I can have a drink (upto 2 pints) anything more will react with my medication.  The rest of the appointment was really stressful, I was the last to go in and spent a good 40 minutes with the doctor. I confessed to everything, from the nose bleeds, not sleeping, feeling down and angry, to eating, spending 120 quid when i only went out for 20 quid speakers, having sex with strangers off websites and sexually abusing my wife.  She phoned through a counseling appointment while I was there and told me that I am now identifying things as being wrong, so that is a sign of getting better. I have 2 months supply of tablets and will be on them for \"a long time to come\" and have to see her again in a month.  Her suggestion to when I am feeling out of sorts (or not able to sleep) is to exercise - run, walk, cycle - anything to increase my endorphines, even if its the middle of the night!!! She said no to sleeping aids at this stage. She also said that this is what I should have done instead of registering for the dating website.   I should also confess to Sandra about the true cost of my spending for speakers in order to humiliate myself and to go out only with cash, not cards.  Her concern with the marriage is Sandra\'s refusal to see Relate. She said I should try and mention it again and if Sandra was not in agreement then it must be that deep down, she may really want a divorce. In her words \"you are trying really hard, your wife needs to put something into it as well\", \"the lines of communication need to be re-established\". She also suggested that Sandra may well know about my sexual activities and has chosen to ignore them for the sake of the kids - something I have not contemplated before. Sandra\'s last response to me mentioning Relate was \"It\'s you with the problem don\'t bring me into it again\" - basically she is scared of opening out to a stranger about her intimate feelings, and I am scared of asking her again.  '),(4169,3532367,'23,June,2004','2004-06-23 00:00:00','   Help!   I\'ve got Sharon on Messenger offering me a blowjob tonight while watching the football - every mans dream she says. Too damned right! But I can\'t, I shouldn\'t - how do I put her off? Do I want to put her off? Sandra never gives blowjobs, she hates it! What do I do?   It\'s not as if I even have feelings for Sharon, I don\'t care about her, I don\'t even fancy her, its purely a sexual release, like a prostitute but without having to pay. That seems to make it even worse - socially unacceptable, but a requirement, a craving that I need to \"feel\" better.  '),(4170,3532367,'23,June,2004','2004-06-23 00:00:00','   I\'m still not sleeping, I\'ve woken every day at 01:00 and 04:00. I\'ve got the doctors today so I must remember to ask for some help.  I\'ve got mixed emotions from last night, Sandra got into bed and said \"right lets get it over with then\" (referring to sex). I told her I didn\'t want it to be like this but she insisted saying \"come on, then you can leave me alone.\" She pulled down her night clothes and lay a towel on the freshly made bed and lay there. I wasn\'t allowed to touch or kiss and when we had finished she got up and had a shower. I know I have used her and she probably feels like she has been raped, but despite the situation I just couldn\'t say no!  I know there is something seriously wrong with me to feel like this and treat the preson I love the way I do, but I can\'t stop it and I am getting increasingly worried about it - I need help!  '),(4171,3532367,'23,June,2004','2004-06-23 00:00:00','   Sandra is trying to be very supportive tonight. She knows how bad my day has been and has asked me if there is anything she can do to help. We discussed several things about the day, looking at it from different angles. I have asked if we can make love tonight, but so far no answer.  '),(4172,3532367,'22,June,2004','2004-06-22 00:00:00','   My boss, has hinted that he needs to do my return to work interview. I have requested someone else does it as I feel he does not beleive I was legitimately off. He has referred it to his boss but the request was turned down on the basis of it just being office banter! I started to write a complaint, but he aplologised, which meant I had to hold onto it, instead of submitting it.  '),(4173,3532367,'22,June,2004','2004-06-22 00:00:00','   Oh my God this is SHIT! I\'ve had nothing but sarcastic comments from my manager about being off. He\'s been trying to make out that I\'ve not been ill but have been doing my car (I\'m not going to tell him am I!).  I\'ve had to walk out of the office to escape the atmosphere. I called Sandra almost in tears and it\'s really getting too much for me. She suggested I talk to Dave, who knows my predicament at the moment. Dave took me for a walk for an hour while we discussed my boss and God and it gave me the chance to calm down. Dave is a Christian and has also worked with my boss and knows what he can be like.  The end result is a headache that is splitting me in two - I can\'t concentrate.  '),(4174,3532367,'22,June,2004','2004-06-22 00:00:00','   I arranged with a friend to drop off my old car this afternoon. I took it down to the auction rooms and they had it off me within 20 mins. I was quite happy as they didn\'t do an inspection that I could witness, so if I get a bill later on to say this and that \"this was wrong and that was wrong\" I can say I don\'t agree! Daft or what?  '),(4175,3532367,'21,June,2004','2004-06-21 00:00:00','   I\'ve fitted my amp in the boot and we have sound! I\'ve hidden the boxes in the bin so that Sandra doesn\'t find out - silly or what, I\'ll have to tell her eventually. I\'ve spent the morning, cutting out 3 huge holes in the parcel shelf for two 6\"x9\" speakers and a 10\" sub woofer.   I had to take some medicine into school during lunch for my son, who claimed he had a headache too. It gave me an opportunity to see how the new lunch system was going and everyone seemed to be enjoying themselves. I couldn\'t get over how quiet it seemed to be.  '),(4176,3532367,'21,June,2004','2004-06-21 00:00:00','   Decided to follow up my day off with another one. This will allow me to sort out my new car and get the old one ready for handing back. I went to the local car spares place to find speakers, and air freshener and a rear eye level brake light. I ended up with speaker, air freshener and a 300W amplifier. I just coudn\'t help myself - not a good start, the first sign of any money and it\'s gone.  '),(4177,3532367,'21,June,2004','2004-06-21 00:00:00','   Yes! good result for the boys today, with a 4-2 win over Croatia. Defence is appauling, but Rooney is the wonder boy up front. I would love to be able to have a drink though. It\'s the first time we\'ve got past the group stages while playing on foreign soil.  '),(4178,3532367,'20,June,2004','2004-06-20 00:00:00','   I arranged to pick up my car tonight. I took the kids with me as Sandra dropped us all off before heading off to her diet club. The kids enjoyed their ride in \"Daddy\'s new car\". I\'m quite pleased with it, nice and tidy but it needs a radio (which shouldn\'t be a problem).  '),(4179,3532367,'20,June,2004','2004-06-20 00:00:00','   Went into work this morning, for a couple of hours. Sharon was chatting to me on messenger and I decided I would be better off meeting her, so I claimed a headache/migrane and left on sick leave. Got the money out of the bank for my car on the way home and went over to pick Sharon up. The date started off OK, we talked while walking in a local park, then went back to her house as a storm brewed up. We finished up going upstairs and having sex in her bedroom for 2 hours.  '),(4180,3532367,'20,June,2004','2004-06-20 00:00:00','   Today we skipped church in favour of celebrating fathers day. Got some nice presents off the kids - England tshirts and shorts. Popped over to see my Dad, who was well considering and the kids enjoyed seeing him. We also visited Sandra\'s Dad on the way back. During the afternoon we all went for a cycle ride down the canal\'s - my bike lost its gear cable, but we managed. Tea was great - roast beef and ice-cream. The evening was frustrating, as I thought Sandra and I might have a night cuddled up, but it didn\'t happen.  '),(4181,3532367,'19,June,2004','2004-06-19 00:00:00','   Strange day today. This morning we finalised the re-mortgage. After work I was due to meet a girl (Sharon) from the internet dating site, but I cancelled in favour of looking at cars. I found one eventually, after being held up by roads being closed due to a major fire and getting involved with a road traffic accident - a boy had been run over on his bike, while trying to cross the road. He was fine, but abusive to everyone in sight, including the Asian lady involved and another witness. Anyway, I put 100 UKP down as a deposit after settling on an 1994 Fiesta from a very nice couple as it had only dome 37000 miles and seemes a genuine car.  '),(4182,3532367,'18,June,2004','2004-06-18 00:00:00','   Last night was like a rollercoaster ride. Got home annoyed at having effectively missed the first half of the football, due to helpful colleagues making me work the late shift. We won, but not having played the best. Sandra was in a good mood, we cuddled up and had a hug and the kids went to bed early, things felt good.   Then it all went wrong - a door slammed in the wind, knocking a cross-stich picture off the bookcase, the glass broke. The kids got it in the neck for messing around (they weren\'t) and having a messy room (it was) and I got it in the neck for not having hung the picture properly. The kids were made to get out of bed and tidy, which they did for an hour, going to sleep themselves at exactly 21:00 (the time I had told them to stop) - they were golden.  Sandra didn\'t speak for the rest of the night and went to bed early, still in a mood, having said she lives in a hell hole, where no-one cares. She still won\'t go to a doctor.  My reaction? I stole the details off a credit card that was lying around and registered on a dating website! A \'double whammy\' straight for the 2 things that I use to make me happy - spending and sex! What made it worse, was that this time I knew I was doing it - and why, but it didn\'t stop me.  '),(4183,3532367,'17,June,2004','2004-06-17 00:00:00','   Sandra\'s had the day off work today (headache). I said she would pick a day I can\'t have off (because of the England match, it would look too obvious) but I had to ring in for her. I was up at 04:00 again last night so I could have done with the time off to catch up with sleep.   I filled the car with petrol this morning, went to pay and realised I didn\'t have any cards with me - oops!. Left my driving license with the lady behind the till, then got Sandra\'s cashpoint card from home to pay for it. Made me late for work, but the woman was OK with it.  I work with 3 other people, non are English, non are into football (soccer), but they are all going home early at 16:00 to watch the game while I have to stop late to cover the office - where is the justice in that?  '),(4184,3532367,'16,June,2004','2004-06-16 00:00:00','   The mortgage offer arrived today and everything seems to have gone through OK. It says they will NOT allow any further second interests (secured loans) on the property - Yipee!  We are going to get a friend to go over the paperwork with us as its a bit confusing with the terminology. The total amouont of the loan (including our existing mortgage is 135,000 UKP! Frightening really. I\'ve still not heard off the counselor which is worrying me slightly. Still I will chase it up with the doctor when I see her next.  '),(4185,3532367,'15,June,2004','2004-06-15 00:00:00','   Paul took to trying to ring Sandra on her mobile this afternoon. She ignored the calls but got all the messages! In the end she sent a text to Jill, who didn\'t realise Paul had been in contact again. The girls seem more settled at the moment, so it may have broken the ice, although Sandra says they will never be friends again.  '),(4186,3532367,'15,June,2004','2004-06-15 00:00:00','   Sandra keeps getting text messaages from Paul, the husband of a former friend - (Jill and Sandra had an argument after she used Sandra as an excuse for her and Paul splitting up). They are both back together and have just returned from 2 weeks holiday, but now Sandra has had a series of texts from Paul saying he has missed her and can\'t get her out of his head!!!!  There is absolutely nothing in it from Sandra\'s point of view, although she is gullable, she was just a shoulder to cry on while he was going through a rough time. There is no doubt in my mind though that Paul fancies Sandra and would like to sleep with her.  '),(4187,3532367,'15,June,2004','2004-06-15 00:00:00','   Still feeling hungry although sleep levels were high last night - thank God! Still not heard off the car dealer though and having rung after him, it seems he is off on leave today. I need to get the cancellation confirmed so that I can buy a cheap car, preferrably for the weekend. I\'ve got to work on Saturday too, so it should be a busy time.  I feel better within myself over the last couple of days, although I have a boss who moans when I get into work 15 minutes early (thats got to be a first!). I try to explain that I travel 18 miles down motorways, through 3 sets of roadworks at rush hour - some days I may be early, some days I may be held up - its just the way it goes, but he just doesn\'t seem to get it! He never used to moan when his mate stopped over an hour later until 6pm everyday because he arrived and hour late after taking his kids to school.  '),(4188,3532367,'15,June,2004','2004-06-15 00:00:00','   At least I managed to get some sleep last night. I woke at around 01:00 and 04;00 again, but it was better than the night before. I feel al little more relaxed although I\'ve not heard anything from the dealer about cancelling the car order (despite chasing him up yesterday). I\'ve started the second hand search again and found a Proton Compact, Vauxhall Corsa, Toyota Celica and a Rover 200 - what a mix!  I should have a fairly busy day today, what with work and a school meeting to attend this afternoon - and I\'ve got to make an appointment for the doctor for next Friday too. I\'ve been feeling hungry today again - eaten most of my lunch already.  '),(4189,3532367,'14,June,2004','2004-06-14 00:00:00','   I phoned the dealer up this morning about cancelling the car. I told him the position I am in and he said to leave it with him as his manager isn\'t in until later today, but he couldn\'t see it being a problem. I am still really worried though. I\'ve been having nose bleeds all day. One minute I am sitting at my desk, the next it just starts pouring! I have a headache and blurred vision too (although I think that is down to looking at the screen too long on my PC).  '),(4190,3532367,'11,June,2004','2004-06-11 00:00:00','   Another night of no sleep, I spent the whole night worrying. I really think I\'ve done the wrong thing with the car, despite Sandra\'s blessing. I am going to ring up today and try to get it cancelled. I\'ve even phoned the man up with the old Clio for sale (the one I saw on Friday that was tidy but was too basic), it\'s still available. I just hope I can cancel OK, the terms and conditions don\'t read well, but when does small print ever make sense? I\'ve worked myself into a panic and now I am scared.  '),(4191,3532367,'11,June,2004','2004-06-11 00:00:00','   Sunday morning started well too, lovely sunny day and we all went out together as a family. After lunch, I asked Sandra if we should test drive one of the Clio\'s I had seen from the day before. She agreed and off we all trooped. The first one was a diesel - nice car but the salesman was well dodgy and rude to Sandra so we bunked off - after he had given me figures that were 10 UKP a month higher than what he had quoted the day before. We went along to another dealership. They didn\'t have a diesel, but did have 5 years interest free. The only problem is the deposit was 2345 and 99 pounds a month. Sandra and I discussed it and surprisingly she said we could afford it, so we signed up.  '),(4192,3532367,'11,June,2004','2004-06-11 00:00:00','   Better late then never. Spent friday lunch and after work looking at cars - one was very nice but basic spec, (not even PAS). Later on I helped a mate move his freezer, washing machine and bits back into his kitchen, after havinig a new floor laid.  Saturday was fine, spent the morning looking at cars - none to my liking. Came up with the idea of going for a slightly better model by going for a newer used car via a dealer. I thought that with a deposit of 1500 I would be ok aiming for 50 quid a month. Found a nice couple of Clio\'s - over budget though. Took the family out for a cycle ride in the afternoon, along the canal, which the kids, Sandra and I enjoyed. Later in the evening we went over to some friends for a meal. Enjoyed the evening, even though I had to go tea-total. I joked with Sandra about having skipped the main course (there wasn\'t that much and we both thought it was the starter).  '),(4193,3532367,'11,June,2004','2004-06-11 00:00:00','   Last night was hell. Hardly slept at all. Woke up at 01:00 and couldn\'t get settled back to sleep all night. It was really difficult trying to keep my hands off Sandra, but I managed it. We had discussed it earlier in the day and she said that it was due to time of day - morning too busy, evening too tired. I tried to ask her if we could have an early night, she said \"don\'t know\" which is her way of agreeing, without saying so. But by the time we got settled down it was 22:00 and she said she was too tired again and her peroid had started.   That just led to frustration and anger on my behalf, although I also felt vunerable - I think that this was down to me recongising for the first time that this was one of my problem areas. I was very aware of my reaction to being denied the \"chemical highs\" the doctor had been on about and was ashamed of it in a way.  '),(4194,3532367,'10,June,2004','2004-06-10 00:00:00','   I viewed another couple of cars on the way into work this morning. One was supposed to be a private sale, but when I got there it turned out to be a scrap dealer with 6 cars for sale - all insurance write-offs. I\'m going to look at another at lunch with Sandra. She\'s off today and has just had her haircut, it sounds really nice - the first time she has been happy with it for a long, long time. She asked to meet me at lunch, so I presume she want to show it off.  '),(4195,3532367,'10,June,2004','2004-06-10 00:00:00','   I had a look at some brand new cars on finance yesterday. It seems I could get quite a few for under the \'100 quid a month\' mark - Citroen, Renault, Hyundai&#8230;.But is that what I want? I don&#8217;t know!       Rang up a chap about a Clio - he wanted 1350, but I said I would offer him 1000 (no tax and solid colour, slightly over priced) - he said no! Another had his car on auction for 1500, with no reserve and only 1 day left. I said I would travel the 80 miles with the cash, and pay the full amount without haggling, if the test drive was OK. He called me a &quot;tssr!&quot;&#8230;.and his auction closed without selling.      I\'ll take a look at another second hand one at lunch. I\'ve got to go into town this lunchtime, to renew my season ticket for the football team I support (the one luxury I was allowed, if I could save for it in time). I\'ve decided that if I put the season ticket on a card, I can gain interest on my money in the bank, but pay off the amount outstanding interest free when the statement arrives - and it gives me more time to save, thanks to my wages not being there! Again, my Saturday job employer hasn\'t paid me, so I\'m 100 quid down this week (that\'s twice in the last month that they haven\'t paid on time).  '),(4196,3532367,'09,June,2004','2004-06-09 00:00:00','   Sandra moaned at me again this morning for not helping get the kids to school. I\'ve tried to explain that I owe nearly 3 hours to work, so have to get in early to make the time up. In the next breath she is telling me about school assemblies and sports days - more time off needed! I also have a school meeting next week and I need to get time off for that too. I didn\'t pester again last night - that&#8217;s twice in a row now - small steps forward from my point of view (expected from Sandra\'s). I bet reading this back makes me sound like a sex pest!  '),(4197,3532367,'08,June,2004','2004-06-08 00:00:00','   Last night was better, woke up around 05:45. I\'ve got to try and motivate myself today to sort out a car, otherwise I\'m going to be without transport. Trouble is I dont know whether to use all the money for a half decent car for 1500 or just go for a \"runner\" at 500 quid.  I need to look at pricing up a couple of projects that I have to do at home. A local organisation want a projector set-up and mounted and a friend of a friend has a PC on XP home that he wants to network! With trying to fit it in everything else I have to think about, its getting a bit awkward, especially with transport doubts.  The worst thing about it is that I wont be paid for either job!  '),(4198,3532367,'08,June,2004','2004-06-08 00:00:00','   So much for getting sleep last night. The dog woke me at 01:00 scratching around and howling, while Sandra accidentally woke me at 05:00 when she moved sharply in her sleep. I am the person who can sleep anywhere at anytime - even standing up! It\'s 08:45, I\'ve been in work for 45 minutes and I am yawning already (and I feel hungry, as the rest of the team eat saussage and egg McMuffin\'s).  '),(4199,3532367,'07,June,2004','2004-06-07 00:00:00','   I hope I manage to get a good nights sleep. What with the heat and the tablets keeping me awake.... I also hope I can manage not to pester Sandra, but in the situation I am in, I need to feel loved and supported and I tend to show that physically, with hugs and affection.  '),(4200,3532367,'06,June,2004','2004-06-06 00:00:00','   I phoned up the vehicle auctions today to arrange a suitable date to deliver the car to them for its VT. They said anytime I like - nice response, but has thrown me a bit as I wasn\'t expecting it. So I\'ve been checking out the car magazines to try and find a replacement for between 1000 and 1500 UKP - why are all the cheap, low mileage cars in Scotland?  Also contacted my credit card company. I have been trying to transfer 300 UKP for the last 2 weeks, to close off 2 store cards. Nothing has gone through though. She tried over the phone, with no joy so I\'ve had to move the money into a bank account to then move it to the store cards.  I will be glad when the mortgage is through and all this lot can be closed off.  '),(4201,3532367,'06,June,2004','2004-06-06 00:00:00','   Work is going well at the moment, but all I seem to be doing is eating. I\'ve felt hungry since last night, (I had a couple of bags of crisps after a roast dinner with all the trimmings and pudding!) and I\'ve already eaten half of my lunch - not normally unusual, but since having my medication, I\'ve not felt like the munchies.  '),(4202,3532367,'06,June,2004','2004-06-06 00:00:00','   Another 04:00 wake up call. I just couldn\'t get to sleep again for over an hour, yet when the alarm when off at 06:30 I couldn\'t get up! Sandra says if I pester her for sex again at 4 in the morning, she will kill me. Typing this I can\'t say I blame her, but at the time I just can\'t think like that....why?  Back to work today too - should be fun!  '),(4203,3532367,'06,June,2004','2004-06-06 00:00:00','   We needed to go to the local supermarket yesterday. The thought of it actually worried me as my tactic had been one of avoidance. Sandra and I got the things we needed - beer, pop, wine and crisps - to entertain the friends that had given me a lift with my bike. When we came out, I told Sandra that I found it difficult to go shopping in the current situation, knowing that I couldn\'t spend. I had wanted a CD and if it had been up to me I would have just picked it off the shelf and put it on my card, but as it was, I didn\'t even mention it. It was something small, but it was still just as hard to \"avoid\" that section of the shop.  The evening went well, but Icouldn\'t drink any beer or wine because of my tablets - again a frustration. I don\'t drink to excess, but like to have a social drink now and then. With Euro 2004 coming up, I am already thinking about what I am going to do when I get asked to watch the match down the pub....  '),(4204,3532367,'05,June,2004','2004-06-05 00:00:00','   The bike sale went OK and I now have 1550 UKP towards a small car. The paperwork for the voluntary termination of my existing car has also arrived, so things are beginning to move forwards. When Sandra got home she gave me a newspaper clipping of a dealer advert for a brand new car for 99 UKP per month with 5 years 0% finance. I thought it was a bit naughty of her to \"tempt\" me with another finance deal, so I asked her if we could afford it. She said we could afford 99 UKP/month once the money had come through from the mortgage. I said I would prefer to put the money in a savings account for a year instead, thus having another 1200 towards a better car next year - she seemed in favour, and I was quite surpised at my excitement at wanting to save.  '),(4205,3532367,'05,June,2004','2004-06-05 00:00:00','   Well it\'s Saturday and I\'m off to work - I\'ve taken a part time job to help on the income front. The good news is, it pays 100 UKP, the bad news is, it\'s only summer work for a few months. Still its helping us get out of a \"tight spot\". Should be finished by lunch, then I\'ve got to get my bike over to Worcester.  I got an apology this morning from Sandra for her comments to me last night, it helped a little, but I can\'t help feeling that she meant what she said. Still I am usually good at forgiving people, so it should all be in the past by the end of today.  '),(4206,3532367,'04,June,2004','2004-06-04 00:00:00','   We came back from the bike ride fine, everyone on a high having enjoyed themselves, although the kids were tired and hungry.   Vicky, a friend of my wife had asked me if she could take Sandra out tonight, but didn\'t want me to tell her - a surprise to cheer her up. The problem is when Vicky rang to confirm with Sandra, she said she had already arranged it all with me....needless to say I got both barrels!  \"Why didn\'t you tell me....I don\'t want to go out....have friends treating me like a charity case....its alright for you you\'ve got the label \'depression\"....its not you that wakes up wanting to be dead.\"  I had wondered how long it would take to be thrown back in my face - 2 days was quicker than I had thought. \"You were the one that made me get counseling....if you feel like that you should go to the doctors too\" I retorted.  \"You weren\'t supposed to go to a doctor!\" came the response, she ignored the part about her going to the doctors.  The whole experience has me feeling as though I have taken another 2 steps backwards. We had been getting on OK for the last few days, making progress, getting a dent in the finances, helping each other through. Now it all seems to have gone again. I feel lost. I feel drained.  '),(4207,3532367,'03,June,2004','2004-06-03 00:00:00','   Today I\'ve got a cycle ride planned with the kids (I\'m off work this week on leave). Nothing too special but its a case of finding things to do with them, that are cheap but can involve everyone. I\'ve also set things going with the bike. Handover has been arranged for Saturday afternoon, I\'ve been offered a lift back by a good friend and I\'ve spoken to the finance company with regards to paying off the amount outstanding. I\'ve got to put this on a credit card until the mortgage money comes through and put the cash into a savings account to buy a cheap run around car, after mine has been handed back. I still have my car at the moment, but this is in the process of going back to the finance company under their voluntary termination agreement - this is where I have paid more than half of the finance and I can hand the car back without penalty. Unfortunately, I wasnt able to do this with the bike as it was on a loan instead of HP! Something to note for the future.  '),(4208,3532367,'03,June,2004','2004-06-03 00:00:00','   I managed to sleep until 4am last night, then spent the rest of the night being restless, which is progress on last night - hooray!  '),(4209,3532367,'03,June,2004','2004-06-03 00:00:00','   Well, having travelled just over 100 miles it looks as though our man is interested in the bike. After a bit of haggling we agreed on a price and he has just phoned to say his bank will have the money ready for him on Saturday - yipee! Fingers crossed it all goes through OK.  Have been getting headaches today, something I dont usually get. I think I had one last night when I went to bed too. I feel tired now, but that is probably due to riding all afternoon - lets hope I manage to get some sleep.  '),(4210,3532367,'03,June,2004','2004-06-03 00:00:00','   Just had a call about my bike (its currently up for sale to clear the finance). A chap from Worcester who is interested but has no way of getting over to me, as the person who is giving him a lift won\'t come this far out (50 miles). For some stupid reason, I said I would take the bike over to him to see it, so I now have a 2-3 hour ride ahead of me and its due to rain. Need to organise kids \"babysitting\" too.  '),(4211,3532367,'03,June,2004','2004-06-03 00:00:00','   Last night was awful. The medication (Fluoxetine 20mg) kept me up all night. I must have literally woken up every 15 minutes, but I dont feel tired now. We had gone to be early - 22:00hrs, together.   To be honest I had hoped we might have intercourse, but we didn\'t and I lay there trying to go to sleep while making an effort to keep my hands off. I \"felt like zebedee\" (from the magic roundabout) I could have gone for a 20 mile run. When I did drop off, I was very aware that every time I woke up I turned to my wife for a hug and more often than not tried to fondle her, kiss her or try to get some form of sex from her (I couldn\'t stop myself) - she must have felt very drained waking up to tell me to \"get off\" every couple of minutes.  '),(4212,3532367,'03,June,2004','2004-06-03 00:00:00','   Having got home a bit flustered, I managed to lock my keys in the garage while putting my bike away (no spare key!) but at least I was in the house. Panic set in because of the way I was thinking, following the events at the doctors. The realisation a doctor thought I had depression was just kicking in......and it was depressing! I found myself thinking of needing a hug from my wife, then questioned whether I was only thinking that because I was low and needed a high.  After an hour or so I rang my mother-in-law (my parents are not supportive), explained what I had done with my keys and told her that I had been to the doctors, got a referral and was on anti-depressants. I wasn\'t sure what her reaction would be (because I felt I now had this huge label \"depression\" hanging round my neck) but she was very good about it.  My father-in-law came round and managed to break into the garage - either because of a mis-spent youth or because he was able to put his engineering skills to practice (I don\'t know which!)  Now the only hurdle was how to tell my wife. This actually proved easier in practise than it did when thinking it over. She is a nurse and has strong beliefs about people with depression. She told me to try not to get hung up on it, or use it as an excuse. The discussions that followed, made me realise that my fear was myself and that my problem was treatable.  '),(4213,3532367,'03,June,2004','2004-06-03 00:00:00','   Yesterday morning is where I am going to begin in any detail. I had driven my bike over to the local dealer to get its service done. While waiting, I thought I would ring the GP for an appointment in the hope of getting some information on counseling. I wanted some help to prevent me getting spending impulses and finding anything suitable was proving very difficult. The appointment was set for 10:45 and I eventually got there at 10:55, having waited nearly an hour longer than I was told for my bike.  The GP was great, when I explained to her what I wanted, she said she would need some more background information and proceeded to ask lots of leading questions. The whole thing took over 30 minutes - the longest I\'ve ever been in a doctors surgery (waiting excluded). The outcome was that she thought that I used spending, eating and sexual activity to get my \"happy kicks\" when feeling low, and if I couldn\'t get these I would turn to internal frustration and moods instead of external displays of anger and violence. The good news was that I hadn\'t developed into using alcohol and drugs!  She would refer me to the in-house counselor, who may in turn refer me to someone more specialist - RESULT!  BUT, in the meantime she wanted me to go onto Anti-depressants....Prozac!  \"But I\'m not depressed\" was my immediate reaction and \"Anti-depressants have a stigma!\"  She calmed me down and said it was either a stigma or getting help. The drugs would help my body with the need for a chemical high, while the counseling would help to re-train my thoughts and approach. The two would work together to reduce my need to spend, eat and have sex.  '),(4214,3532367,'03,June,2004','2004-06-03 00:00:00','   Welcome to my blog. I hope you find the entries useful and informative. Please feel free to leave any comments or ask me questions. I will try to answer them as fully as I can.  '),(4215,3532367,'21,July,2004','2004-07-21 00:00:00','   Sorry, my language is always the first thing to go when I am stressed. I\'ve had a couple of texts from Angela, but we\'ve not been able to chat because she has gone out for the day with her friend. I feel like I have lost a limb. She says we need to chat ASAP and I agree with her. I just hope so much&nbsp;that it all works out.  I went for a walk at lunch to try and stay awake as I was falling asleep at my desk. When I got back I had a call from my director to go and see the department that were messing me about last week. They had started again, complaining about a problem which it transpires was fixed by an external engineer last week. I asked why they had not said anything about it and was told by the staff they their boss had told them not to tell us. The frustration has compounded and its now taking itself out on my head again.   It hurts. I am so low, so down and things are getting to me so easily. I\'m not sure how much longer I can cope.  '),(4216,3532367,'20,July,2004','2004-07-20 00:00:00','   That\'s what I feel inside today. I am waiting, hoping, that Angela will come online. I miss her so much, even though she said she would not log in today. I have not slept at all, I feel totally drained and have been running on reserves for too long now. I need to know she still wants to give us a try. Sorry seems such an inadequate word a times like this, but I am - for lots of reasons.  Can\'t keep my eyes open....  '),(4217,3532367,'20,July,2004','2004-07-20 00:00:00','   Gutted, betrayed, violated, raped, unclean. A mixture of emotions, expected when&nbsp;people&nbsp;find out the truth.  &nbsp;  I actually feel as though I have cheated on Angela, not Sandra! I am devastated at the thought that I might lose her. The mixture of emotion is strange, new and scary. How can these feelings be so strong for someone? If only we knew.  &nbsp;  One thing is for sure, I\'ve let her&nbsp;and myself down and Angela is going to be the one left to pick up the pieces.  &nbsp;  I hope we are&nbsp;strong enough to get through this.  '),(4218,3532367,'19,July,2004','2004-07-19 00:00:00','   Well by now Angela will be part way through June. I\'ve done it, exposed my soul, my life, my thoughts, my actions. I feel totally exposed....very vulnerable! &nbsp; I hope I\'ve done the right thing. I\'ve done it out of trust. I have total trust, total faith....total lots of things, for a person I \'ve known little more than a week.&nbsp;It&nbsp;not about&nbsp;lust, nor&nbsp;sex, nor a&nbsp;craving of affection. Its more than that....its special. &nbsp; ....and now I\'ve got to try and get to sleep - lmfao!  '),(4219,3532367,'19,July,2004','2004-07-19 00:00:00','   I am really agonising over whether to let Angela know about my Blog. I\'ve been thinking it for a while, but I am scared she won\'t like what she reads and it might just be a case of it being too much for her to cope with. &nbsp; Its tearing me up inside that I have a secret from her - probably my only one - but it contains lots of other little secrets. What to do? - I have come so close to telling her on more than one occasion, but I don\'t want to lose her.  &nbsp; We had our 1 week anniversary today. Only a week, but it feels like she knows more about me than 13 years with Sandra. Angela - Angel....mine anyway. God sent? Who knows? .....probably!  '),(4220,3532367,'19,July,2004','2004-07-19 00:00:00','   The lady from Personnel got in on the act today. Insistence from my boss that my screen is the cause of headaches at work, meant he referred it to Health &amp; Saftey for an assessment. When she arrived I told her I was an assessor myself, so she asked what the real problem was.  &nbsp; She cited a poor working environment and possible conflicts amongst staff, so I gave her the whole history of my move over here 2 years ago, how it was really poor for 12 months, but is generally OK now, although the idiot who thinks he is boss cannot man manage a paper bag. She said I was showing signs of depression (she used to be a nurse) so I told her about the doctor and the counsellor, although I didn\'t mention the home issues. She said I should keep a diary - LOL and that I looked shattered and should have the rest of the day off - PMSL  '),(4221,3532367,'19,July,2004','2004-07-19 00:00:00','   I had no&nbsp;sleep at all last night (Rob came into our bed at midnight and stayed there) he\'s such a damned fidget. I could hardly get out of bed this morning and slept right through the alarm for an hour. When I did get up, I had to iron my trousers and shirt and find my footie kit for tonight, so I thought I would be very late. Sandra actually gave me a hug this morning. She put&nbsp;one arm round me and sort of snuggled me. Our kiss \"bye\" was still just a peck though.  '),(4222,3532367,'19,July,2004','2004-07-19 00:00:00','   So far tonight has been quiet. The meeting at school finished at 20:30 and I\'ve been moping round like a lost sheep for the last hour. Sandra has gone to bed and if it weren\'t quite so early, I would go too. I don\'t want to be waking in the night and I don\'t want to be tempted to have sex. I don\'t enjoy it with Sandra, it is just a comfort - does that make sense? &nbsp; Oh, its my Mom\'s birthday today - another reason to go to bed early - LOL, but that is a different story (we dont get on)!  '),(4223,3532367,'18,July,2004','2004-07-18 00:00:00','   Thank you God for Mia, she is going through shit and is still there for me. We have been chatting on MSN all day, with friendly chat, flirtation, vulgarity, blantant sexual comments and care. I have a thumping head (half from all the typing and half from the annoying call with the counsellor), yet still she is there, offering advice and suggestions.  '),(4224,3532367,'18,July,2004','2004-07-18 00:00:00','   Just phoned up for another appointment to be told there weren\'t any until the 2nd August (this woman is seriously pissing me off, I have no faith in her). She is going to put me on the list for cancellation cover. I have given my mobile, just in case. The passing comment.....\"Shame you didn\'t ring earlier, you could have come in this afternoon\" ...........aaaarrrrgggghhhhh!  '),(4225,3532367,'18,July,2004','2004-07-18 00:00:00','   I managed to get back in time for tea. The scratches on my arms were covered with the story of lending a mate some steps and getting roped into cutting down&nbsp;HIS conifers.&nbsp;We had an&nbsp;arrangement to pick up the kids from their Nans, where we would all meet up and have dinner, before coming home. A nice roast beef and yorkshire puds put pay to my fitness and healthy eating day.  &nbsp;  Sandra had an early start in the morning with work, so the kids stopped over again. When we got home, Sandra went for a bath, while I sorted out my work stuff. When she came down, she asked if I wanted to go and see Shrek 2. I asked if it was a date and she shrugged her shoulders.  &nbsp;  While she got ready, I decided on Harry Potter 3, as none of the kids would want to see it. It was a good film (predicatbly too scary for the kids) and the popcorn and coke went down well. We didn\'t talk a lot and held hands just once, when she felt scared of some yobs as we were leaving.  '),(4226,3532367,'17,July,2004','2004-07-17 00:00:00','   With Sandra being at work and having no kids because they stopped at their Nan\'s last night (because we went out) I had an afternoon to myself. Following on from the cleaning, washing-up, polishing and a spot of gardening,&nbsp;I decided to give Angela a ring.  &nbsp;  Her kids were back home after their holiday, but she didn\'t mind, although I would have to bring my steps over to help chop down the hedge. Well this bloody hedge comprised of at least 16, 20 ft high conifers. And my steps were 6 ft. So that settled it, the new hedge was going to be 10 ft by the time I finished. Nearly 3 hours later, I was only half way though. As the electric trimmer was only capable of shaping the edges, I had to manually saw through the rest.  &nbsp;  It still needs finishing off, but Angela decided I needed to be finished off before I left too, so she dragged me upstairs and locked me in the bedroom&nbsp; - thank God for Stuart Little - LOL  '),(4227,3532367,'17,July,2004','2004-07-17 00:00:00','   The gym really surprised me. I thought I would find it boring, but on the electronic machines, I found it a therapeutic way to relax and work up a sweat while burning calories. I wasn\'t that keen on the resistance machines (weights) but I did them nonetheless.  &nbsp;  &nbsp;     Spent over an hour and a half there having really enjoyed myself and even arranged to go again during the week. Its not at the gym I would go to regularly, but it was OK.  '),(4228,3532367,'17,July,2004','2004-07-17 00:00:00','   Well after getting the girls out to their Saturday club and having dodged the rain showers at work, I got home to a grateful rest. Well thats what I thought.  &nbsp;  Sandra had arranged for us to go out and play happy couples at our friends house round the corner.   While Sandra had a shower, I lay on the bed trying to grab some&nbsp;rest. The temptation while she was getting ready was too much and I asked if we could have sex. Her reply of \"if you have to\" was met by her taking her towel&nbsp;off,&nbsp;lying on the bed, closing her eyes, parting her legs and saying \"hurry up\"....I did!  &nbsp;  Armed with a bottle of red, we made our way over for tea (at 20:30) I was starving, having already rung to see if I could have a pack of chips at 17:30 to tide me over - our host suggested a banana.  &nbsp;  We arrived and sat at opposite ends of a 4 seat sofa, which I found amusing&nbsp;&nbsp;   &nbsp;  The meal was nice, a mousaka, although it wasn\'t that filling, with just a salad to go with it. By the time we had finished coffee, I was shattered and we got home before midnight, with Paul and I having arranged to go to the gym in the morning.  '),(4229,3532367,'16,July,2004','2004-07-16 00:00:00','   Tonight it was the turn of Rob to have an awards ceremony. He has only just started at his group so he just got a welcome present. The band was really good, they have won loads of competitions this year, so it was nice to see them perform for us. It seems Louise may be able to join, but she needs to go on Monday night to see what instrument she could play - almost thought of joining her - lol.  '),(4230,3532367,'15,July,2004','2004-07-15 00:00:00','   Having slept quite well, it was up early for work. I got an odd comment from Sandra about being late back, but with the people I was supposed to be with, it was covered well. I am hoping to be able to make Thursdays a regular event, with cover from me going out with friends.  &nbsp;  I had time off this afternoon to enable me to go to the counsellor appointment. Got there, having had to book 2 hours off work, only to be told she had gone off sick. I am still really pissed off about it.   Apparently they treid to get hold of me, but I have to ring back on Monday to re-schedule.  '),(4231,3532367,'15,July,2004','2004-07-15 00:00:00','   Last night went well. Managed to get over for 20:00. Spent the night drinking coffee, while Angela downed a bottle of red wine as we both felt too tired to go out. We chatted, things got very deep (conversationally) and we made love into the late hours and cuddled. I got home late, but it was worth the time spent being happy, just to enable me to re-group.  '),(4232,3532367,'15,July,2004','2004-07-15 00:00:00','   The plan is to get over to Leicester again tonight, perhaps a bit of bowling or a drink. I\'ve already sowed the seed for being out with Sandra and she seems OK with the excuse I gave her. Oh well, fingers crossed the kids go to bed good and don\'t play up!  '),(4233,3532367,'14,July,2004','2004-07-14 00:00:00','   Bloody up at 02:30 again, eventually got out of bed and made a hot milky drink at 03:00. I had hoped it would settle me down, but it seemed to wake me up all the more. I watched TV for half an hour, then hit the PC and mailed Angela. I got a reply almost instantly as she was on night shifts at work, so that was it; stage set for 2 hours of chatting online via email - I dont think she got much work done.  Eventually went to bed at 05:30, dropped off at about 6 then was woken 30 mins later by the alarm.....lovely!  I got into work really late. Simon collared me and said he thought he had dropped me in it. Apparently when his Mom rang him later that day he asked how I was as I had gone off sick. She of course had said she had been there all afternoon and hadn\'t seen me.   Talk about quick thinking, I used the phone excuse and not wanting to go home with MIL and kids there. Bluffed my way through saying I went to the park. It seemed to work, but I was so involved that an hour later I even told Mia the same story. While I was telling her I kept thinking \"she will only read the truth later, why are you feeding her this crap\" but I was too far gone to back track and it would give me practise in sounding convincing if I was quizzed by Sandra later.  Sorry Mia  '),(4234,3532367,'14,July,2004','2004-07-14 00:00:00','   Got back from Leicester at 18:45, an hour later than when I usually get home. There was no-one home. Managed to get in a quick cup of coffee before going out to a school meeting at 19:00. Inevitably, that over-ran, home again for 21:30 then off to bed an hour later, after some tea.  Sandra asked where I had been and I said I had expected to have to work over due to being late back from the bank (she didn\'t know I\'d booked the afternoon off sick). We went to bed, in silence and both tired.  '),(4235,3532367,'13,July,2004','2004-07-13 00:00:00','   Got a text off Angela this afternoon, wishing she could have a hug. Well I don\'t need asking twice so I headed over to Leicester, seeing as MIL had put a spanner in the works (with regards to my original plan). We had a couple of hours together, went to get some car parts for one of her friends and shared a coffee while cuddling up on the sofa. We talked and laughed and kissed loads. I told her I was really happy just being there with her and wasn\'t at all bothered about not having sex with her, as with other people I had met, because she meant more to me than that. The next thing you know she took my hand and lead me upstairs.  The strange thing was I actually tried to stop her. I asked if she was sure (more than once) and said I didn\'t want her to feel used and I didn\'t want to spoil anything. Nothing I said would prevent her though, she \"had an itch that needed scratching\" as she put it.  Well I won\'t go into details other than to say we \"made love\" and I use that term rather than \"had sex\" because it is exactly what it was, - two people forming a bond, coming together, (excuse the pun) being as one.  It made me late home though  '),(4236,3532367,'13,July,2004','2004-07-13 00:00:00','   OK, so I started off well. Met Sandra at the bank and went for a coffee before heading off home. Halfway home my mobile rang and by the time I pull over its gone to the voicemail. I get the message \"Relate have rung home and left a message with Mom, bit embarrassing but can you call them back.\"  Well I don\'t mind her taking the message but I\'m not going home if she is there cos the kids will be there too and I won\'t get any sleep - what should I do?  '),(4237,3532367,'13,July,2004','2004-07-13 00:00:00','   I\'ve had enough now - time to go and get some kip! They have been drilling and banging in the room next door and its killing me on top of being tired. I\'ve got to out to the bank to meet Sandra, but I don\'t fancy coming back, so will take the rest of the day off sick and try and get some sleep.  '),(4238,3532367,'13,July,2004','2004-07-13 00:00:00','   Relate phoned again this morning. I spoke with the manager and explained my position, she was very understanding and said I could pay as much as I could afford now and discuss the rest with the counselor. I opted for 15 UKP, which she said was fine. Sandra had said she didn\'t want to come, but the lady said if she could come to the first one to be assessed that would be better, even if she didn\'t want to come to any more.   She is due to call back to arrange the appointment.  '),(4239,3532367,'13,July,2004','2004-07-13 00:00:00','   OK, it was 2:30 when I woke up this morning. Insomnia stikes again! For the rest of the night I was just tossing and turning and now I am back at work it\'s all I can do to keep my eyes open.  '),(4240,3532367,'13,July,2004','2004-07-13 00:00:00','   Took an extra hour for lunch and went into the City Centre to meet Angela of all people. She had come over with a friend who had some work to do. Totally spontaneous, not pre-arranged or anything, she just emailed and said she was coming over to do shome shopping. I\'ve not even shaved, brushed my teeth and I\'m wearing trampy clothes today....bloody typical!  We had agreed on just meeting as friends for coffee - nothing more, but during the last couple of days we had had a couple of heart to hearts and she had come to realise that I was indeed telling the truth and not just after her as another woman on the side, or sexual conquest.  After walking to the shopping centre on the other side of town, I went to put my arm round her as a form of support when she was telling me something - just to say I was there if she needed it. She immediately came forward and kissed me - for ages - in the middle of the street, with loads of people around. The most unromantic place, the most unromantic time, but it just felt so right, so natural, so .....nice! We never made the coffee place, we stood overlooking the church, talking, hugging, kissing, listening and caring about each other. It felt soooo nice to be wanted and to feel appreciated. Not sex, not perverted, just simple feelings of happiness.  WOW!  '),(4241,3532367,'12,July,2004','2004-07-12 00:00:00','   I have a department here that has been giving me grief today.     I used to work with them and there are very complicated public sector political links that are too deep to go into here, but basically I have been playing piggy in the middle all day.  They put a call in, I am told by my director to close it and refer them to the supplier, they contact supplier, supplier says they need to speak to someone in IT. Director says close it - supplier can send someone in to cover it under their agreements and the circle continues, with me just being the political pawn.  Interestingly my director said \"well you can handle it, your robust.\" To which I replied with a one liner about being delicate. After we had finished the work discussion he went digging for more info on my one liner, so I admitted to him, very bashfully, that I had been diagnosed with depression and that I was on meds, not sleeping and up for counseling and further referral. I also mentioned my lack of concentration and having to take things one day at a time. He asked what the cause was, and I said I wasn\'t sure at the moment, but put the trigger down to the difficulties I was experienced at work when I moved department.  '),(4242,3532367,'12,July,2004','2004-07-12 00:00:00','   \"I am not strong enough to do anything about it and I dont think you are either, so it looks like we are going to be unhappy for the rest of our lives.\" Said Sandra. She continued \"Yes I have been like it since we got married. I begged to have you back (we split on the run up to our wedding), I got pregnant (during the making up process) and I agreed to things I didn\'t want just to keep you. I have done ever since....we have never been a happy house\".  She is not strong enough for Relate or to see the doctor as she doesn\'t want to humiliate herself again. \"It\'s too soon, why can\'t we just sort out the money side first?\" She said. \"I don\'t want the pain agiain, it\'s not something I am brave enough to do like you.\"  '),(4243,3532367,'11,July,2004','2004-07-11 00:00:00','   Just managed to get through to Relate. She took my details, then said there was a 45 UKP charge for each session! I told her I couldn\'t afford that and her manager is going to ring me back to discuss options for reduced payments. Lets hope they can get it down to 10-15 quid or something. I have said she can leave a message if she needs to.  '),(4244,3532367,'11,July,2004','2004-07-11 00:00:00','   Just dropped a couple of links in to other blogs of note. \"Not a Lot\" comes from my good friend Mia, while \"Demons in Me\" comes from Ecks, in Australia. I like his blogs cos while he is in a totally different situation, in another part of the world, there seems to be a lot of similarities.  '),(4245,3532367,'11,July,2004','2004-07-11 00:00:00','   Simon just popped in and asked how I was. Then asked how things were at home. I told him I thought Sandra was jealous of the fact that I had been out enjoying myself while she was at home with the kids.  His comment was \"so she\'s had them for a day for a change, rather than dumping them at my mums, you mean\". From my point of view it seemed a strange thing to say and I took it as though he was sort of telling me he supported me. Almost as though it was his way of telling me she should buck her ideas up.  I can\'t quite work it out though, so I don\'t know if I am on the right wavelength.  '),(4246,3532367,'11,July,2004','2004-07-11 00:00:00','   Tried them 3 times over the weekend and all I get is an answer machine. Checked the opening hours, they seem OK, but the email address bounces back with an NDR. I\'ve notified Relate HQ but so far no answer, despite the fact they have read my email to them.  '),(4247,3532367,'11,July,2004','2004-07-11 00:00:00','   Well it was carry on as we left off after I got back from Silverstone. Sandra was in  a strop. I tried to give her a hug but she didn\'t want to know. She moaned about having been left with the kids all day while I was off enjoying myself. I said that perhaps I should move out - her comment was we can\'t afford it. So I suggested seperate beds but she said there wasn\'t enough room. She still says (although not convincingly) that she wants to be with me.  She continues that any affection is hypocritical as the other day I was contemplating killing her, but now I wanted sex. I had told her how I was feeling to try to shock her into realising how bad I had felt. It doesn\'t seem to have made any change.  '),(4248,3532367,'10,July,2004','2004-07-10 00:00:00','   Sandra spent the evening writing cheques and closure letters for several of the main creditors, so hopefully they will be piad off soon. I need to hadd 500 UKP from the savings to allow us to pay off thelast of my cards. Its a bit of a blow, but I\'d rather have it out of the way.  '),(4249,3532367,'10,July,2004','2004-07-10 00:00:00','   Fantastic day today. Got up (before the alarm again) and was waiting outside for my lift - good start! During one of the warm up races there was a major accident right in front of us, and the GP itself was good too, although the winner was a bit predictable! The traffic wan\'t too bad on the way home - only took 90 mis to get to the M40, which by previous years was a total bonus apparently.  Highlight of my day - a 50 minute phone conversation with Angela. She sounds older than she looks (although at 37 she is 4 years older than me) and while she knows how to have fun she is sensible too. We shared some serious moments about married life but had a great laugh too. I even managed to get her to agree to coffee one day ;)  '),(4250,3532367,'10,July,2004','2004-07-10 00:00:00','   The alarm went off at 05:30 - I had already been awake for over 30 mins. I got ready to go after Sandra had threatened me \"don\'t you dare waste my brothers money, by not going\". Sandra had said that \"they ware going between 06:00 and 06:15\" and that I \"should take my rucksack to help carry the food and drinks.\"  I drove over to my FIL\'s house - no car, but it was only 06:10 - where were they? 5 mins later I had a call on my mobile.   \"Where are you?\"  \"I\'m at your house.\" \"What you doing there I said I would pick you up at 06:15 from yours.\" \"Sorry, thats not the message I got, there was no suggestion you would be picking me up.\" \"Stay there, I\'ll come and get you.\"  Great, here I was in the wrong place...fantastic! The traffic would be shit, everyone would be late and miss the practise session, possibly qualifying as well and I would get the blame for not following instructions. Why couldn\'t Sandra have told me or been clearer? I looked up at the window of the house, my MIL was there and she mouthed \"you pratt....you pratt\" at me. My heart sunk even further, I couldn\'t take much more of this. It was all I could do not to cry.  The journey there was in virtual silence, Simon tried to dispell any atmosphere by saying it was history and to forget it, but I still felt burdened. Fortunately, though the traffic was light, we arrived in plenty of time having gone a different way - I was washed by a wave of relief and the day picked up as the racing began.  '),(4251,3532367,'09,July,2004','2004-07-09 00:00:00','   Nasty comments from Sandra tonight. Basically she feels as though I am making a fuss so that everyone treats me nicely. She cited Simon and the GP tickets and the fact that we have been looking at Health Clubs just \"so that Jon can feel better, bugger everyone else\".   There were a few other comments - basically silly ones - but with the way I am feeling they just get to me, when at any other time they would be water off a ducks back. I went to get up to go out for a cycle to get rid of my feelings but Sandra prevented me from going, with more cutting comments.  We sat on the sofa in silence, for over half an hour my thoughts ranged from which knife do I use and how do I kill her, to how do I hurt myself, what would be the best way? It was very scary. I can remember thinking at the time, this is dangerous and being really worried about it, but it didn\'t stop me.   The room physically shook from side to side as I looked at it. If I closed my eyes, my body felt as though it was physically spinning - over and over. The only way I could stop it was to get up and go, when Sandra left the room. I drove round for a bit, pulled up at the park down the road and went for a run, as hard as I could, in the pitch black of the park.  By the time I got home and hour later, Sandra was in bed, apparently asleep. I did feel better, but I was scared, tired and low. That is the first time I can recall wanting to kill Sandra or even self harm myself.  '),(4252,3532367,'08,July,2004','2004-07-08 00:00:00','   I completed the wireless network after finishing work today. It took a couple of hours, slightly longer than planned, as I had to remove a virus off the laptop the customer was using. The money paid for the parts and the extra (profit) went to the next door neighbour for the work he had done to his car, after I had put a dent in his wing last year. Easy come, easy go, but al least it was another debt busted!  Sandra says that money will be tighter than she had originally thought as we still have to pay off the overdrafts and her credit card. Lets hope the mortgage rate doesn\'t go up any more.  '),(4253,3532367,'08,July,2004','2004-07-08 00:00:00','   My Brother in Law (Simon) offered me a ticket to the British Grand Prix this weekend. As I am not doing any work I thought it would be good to go with him, my FIL and my SIL. I told Sandra and she said it would be OK. Simon has a contact at Jordan so as he had not mentioned any money (he usually does), I asssumed it was free. It turns out later that he has paid 75 quid for the ticket. I challenged him about this as I don\'t want to be treated like a charity case, but he just dismissed it as being his treat.  '),(4254,3532367,'08,July,2004','2004-07-08 00:00:00','   Last night was great. I decided to stop up late to cool down from my run and to wait until I was tired enough to go to bed. I ended up chatting to Angela on MSN. Angela is a person who I met on the dating site. We have chatted, she knows all about me, but won\'t meet cos of my position. However as we have chatted, we have discovered that we get on really well. We have loads in common, like similar things and chat for ages, when it only seems like 5 mins. She knows that as we have chatted my fondness for her has grown. I told her last night that I wouldn\'t want to jump into bed with her just for sex because she was becoming more special (I meant it as well). We also discussed her holiday and I was able to give her some pointers as she was looking at South Devon. She basically booked it there and then over the net, having looked as some websites I sent her. It was fantastic, we have never met, I\'ve never spoken to her or heard her voice, but we seem to have some sort of bond.  I had teased her previously by giving her my mobile number, but she scrubbed it. However things came full circle as we said good night at 01:00 and she offered to give me her details. I texted this morning to see how she was - the reply was \"knackered\"!  '),(4255,3532367,'08,July,2004','2004-07-08 00:00:00','   I wanted to go for a cycle tonight, but after having got the kids to bed and got myself ready, Sandra made stupid comments. They were along the lines of \"I\'ll have to get into this depression lark so that I can pop out for an hour to relax\". That made me feel worse. I didn\'t go in the end because of it, and although she apologised, the damage had already been done.  I tried to make up for my lack of exercise by taking the dogs at 22:00 and having a run round the park. I got all sweaty for 5 minutes, but my legs and shins suffered on the uneven ground.  '),(4256,3532367,'07,July,2004','2004-07-07 00:00:00','   The kids had their school report last night, in general we were very pleased with them. The only problem is that Rob is moving into a class, with a teacher that we are not that keen on. At this stage, they keep the same teacher for 2 years, so Sandra had some concerns. We both went in this morning to discuss it with the Head. She was surprised, but tried to allay our fears about the teacher concerned. We can\'t pinpoint the issue, just a gut feeling, that she is watching her time go by, until she retires. The appointment made me a bit late for work, but who cares, sometimes the kids come first.  '),(4257,3532367,'07,July,2004','2004-07-07 00:00:00','   It seems as though we might be short on the amount to clear our debts. The 2nd charge on the house has a higher settlement figure than we thought, by 4,000 UKP. The result is we don\'t have enough to clear our overdrafts and Sandra\'s Visa. I\'ve asked her if it can all go to clearing my cards as a priority, so that  I am not tempted to spend on them. She seems happy with this, although the dicovery caused some tension.  Just before bed, I decided to take the dogs a walk and while I was out, I did a few laps of the field. I surprised myself by the distance I did, before getting out of breath. By the time I got back, Sandra had gone to bed, so I spent 30 mins on the computer, chatting, while I was cooling down.  '),(4258,3532367,'07,July,2004','2004-07-07 00:00:00','   Insomnia - the first song on the radio on the way in to work. \"I can\'t get no sleep\"....How apt!  02:30 was the best I managed, before Rob came into our bed feeling unwell. He has been for a couple of days, but last night was a bad night for him....and me! I just couldn\'t get out of bed today - 06:30 the alarm goes off, it turns off at 07:30 and I was still in bed at 07:45. Just about got into work for 09:00!  I feel a bit stiff after the football last night - it was a good work out, I broke into a sweat (for the first time ever) but felt afterwards, that I could have done more. I also got a nice friction burn on my knee, from the astro turf pitch.   I went straight from there to a parents evening with the violin teacher. Some good comments about the work being put in by Louise to her ability to read music, rather than just remembering the tune. By the time we got home, there was just enough time for tea, before going to bed at 22:00.  '),(4259,3532367,'06,July,2004','2004-07-06 00:00:00','   I am struggling today to stay focussed. I don\'t know whether it\'s because of lack of sleep or if it\'s due to boredom through not being motivated to do any work. I\'ve done nothing but eat again and spent most of the day chatting on the net. I\'ve had 2 blatant offers of sex, one for Saturday and one for Monday - all arranged with date, time, place.   I\'ve just gone off at a friend for no reason. I was being crude, thinking at the time it was funny, then as I read it, I realised I was abusing the trust and friendship that had been built up. To her credit, she is being an absolute rock, if it were me I would have thrown it back 10 fold.  I am hoping I can get motivated sufficiently for football later. I want to work so damned hard that I collapse, just to try and get this shit out of my system. I just hope my body will allow me to do it. Perhaps I should look at the gym sooner, rather than later, or perhaps I need to find something else, or just get some sleep.  '),(4260,3532367,'05,July,2004','2004-07-05 00:00:00','   I spent pretty much the whole of yesterday eating. Crisps, biscuits, snacks, fruit, dinners, puddings, chips....the list goes on and today I feel fat, but I\'m still hungry.  Last night Sandra and I spent some time trying to find a cottage to holiday in during October. Sandra feels guilty that she is spending money, but I pointed out that we will need to spend some money on ourselves, we need a break and it will be the first time it will be paid for before we go.  Last night was pants. No sleep at all from 04:00 - again! Didn\'t get up, although I was very tempted this time. The trouble is, I could fall asleep standing up now, I\'m shattered!  '),(4261,3532367,'05,July,2004','2004-07-05 00:00:00','   Set off early today and got in for 07:45. The traffic wasn\'t too bad at that time, but a lot of people struggled in late due to the major roadworks on the M6. The office is quite subdued at the moment, with 2 out of the 4 being in a foul mood. Sandra has had the day off today with suspected chicken pox!!!  Oh the joys! We had sex last night too. Nothing to shout home about, just trying to keep a record. She cryed afterwards and I felt empty, devoid of feeling - it wasn\'t nice. I will try and get hold of Relate today or tomorrow.  '),(4262,3532367,'04,July,2004','2004-07-04 00:00:00','   Well a call from the \"Mother In Law\" suggests we have 3 kids with chicken pox - they all stopped over last night as we went out to the BBQ. A mad rush over there confirmed our \'fears\'. They all wanted to stop put for the day, which suited me.  Sandra and I decided to have lunch out - we went to the local supermarket cafe and lunched for 4 quid. Next stop was the bowling alley to drop off a present to a birthday party that Rob was supposed to be going too, but now couldn\'t.  Following that we decided to look round the local health clubs, very nice but a tad expensive at 45 UKP per month (6 months min) or 30 UKP month at the local public gym (but with 12 month min). Sandra decided a  review of finances was required first - fair enough - neither of the clubs offered a one month trial without committment.  We got back home late in the afternoon - via the garden centre for 12 quids worth of plants - and planted out the front garden. It was quite nice considering the budget! A quick tea and while Sandra went to pick up the kids, I decided to cycle to church for the evening service (not the usual service we do). It was a stand-in to the usual minister, and he was very good, splitting the sermon into 4 parts of 5 mins each.  '),(4263,3532367,'03,July,2004','2004-07-03 00:00:00','   We went over to a friend of Sandra\'s tonight for a BBQ. They used to work together. The evening was very strained, I didn\'t know anyone, and Sandra only knew one other person. I had 1 bottle of beer, with my burger then moved to soft drinks as I was driving. We came away at 23:00, it was pleasant enough but not exactly relaxing if you know what I mean.  '),(4264,3532367,'02,July,2004','2004-07-02 00:00:00','   This morning is mad. Sandra is working, so am I from lunch time. We have kids who were up all night chatting and laughing and now can\'t get up out of bed for dancing. Sandra is panicing about getting them and herself ready and everyone is copping out.  I have to get ready to take the girls and friends dancing at 10.00. I have to take Rob to his Nan\'s to be looked after, then get back in time to get ready for work. It looks like it could rain as well, which is going to make this afternoon fun!  '),(4265,3532367,'02,July,2004','2004-07-02 00:00:00','   Tonight was interesting. On my way home I had popped in to assess a network install and do some email faults for a customer. With the money from that and the job on Thursday, I bought a drinks bottle and pump for the bike and a rear \"eye level\" brake light for the car.  I was in the process of fitting the parts to my bike, when Sandra phoned. Our friend over the way had called her as their daughter had come off her bike - could I go round? Despite being a nurse, Sandra hates first aid, so I obliged and the girl went off for X-Rays to her elbow.  We kept hold of the younger daughter, while the rest of the family went to hospital. By the time they got back, we had next doors lad in with a fractured collarbone. He had done it previously, but wanted his sling re-done. There were kids and parents everywhere.  We ended up having the youngest daughter over on a stop over, when they asked at the most awkward time. We felt as though we were in a position and couldn\'t say \"no\"......KIDS!  '),(4266,3532367,'01,July,2004','2004-07-01 00:00:00','   Well I don\'t know what to think about my first \"session\". The counselor was young and didn\'t really seem to be asking the right questions. \"Do I wash my hands a lot, or do I have to have things in order, or do things have to be in a straight line?\"  I knew exactly what she was trying to get at, but her questions came out of a leaflet and did not relate to me personally. I found it very difficult to talk to her (not like with the doctor at all) and came away feeling disappointed.   She originally decided to give me a leaflet to take home on obsessive, compulsive behaviour, but then changed her mind at the last minute to one about Depression. She also suggested I contact Relate.   The one thing she did do, was to suggest a referral to the Mental Health Unit for a full assessment. I said yes to this, but mainly because I got the impression that she didn\'t know what she was doing.  '),(4267,3532367,'01,July,2004','2004-07-01 00:00:00','   Just had news from Sandra about Rob. He had his health check today and failed his eye sight test. He couldn\'t read out of his right eye. In his words \"he couldn\'t see it\". So it looks like we\'ve got a trip to the opticians with him this weekend!  When will we ever get a break ?  '),(4268,3532367,'01,July,2004','2004-07-01 00:00:00','   I got home just before 21:00 and was very wound up. I couldn\'t settle or relax at all, but didn\'t want to go out running or cycling. In the end I had a bacon sandwich and watched some TV. By about 22:30 I was feeling shattered and eventually we went to bed at about 23:00.  I read for a bit and tried to get some sleep. Unfortunately, our youngest daughter came in after having had a nightmare at about 01:30. She stayed with us all night, which meant I felt uncomfortable and couldn\'t sleep at all. By about 4 or 5 I was considering going into her bed, but last time I did that I couldn\'t settle either, so I gave it a miss.  Today I am exhausted, but I am trying to keep myself busy so that I don\'t sit and think about this appointment later. Just typing about it, I can feel myself becoming uneasy.  '),(4269,3532367,'01,July,2004','2004-07-01 00:00:00','   We have a new person joining our team today. He is young and overly keen and of course has a lot to learn. A lot of the attention is focussed on him today, which should be a good thing. The problem is, his desk is next to mine, so it\'s like \"New Street Station\" here at the moment. I am finding it very unsettling.  I dont know why, but I seem to view everything as a problem, or look at the negative in daily situations or things I do. Today I feel down or as if the slightest thing is getting on top of me. I feel as though I am creating problems out of nothing.  I have just completed a training session (which I enjoyed) but I spent the latter part of it thinking about fishing. Buying a new rod and tackle and spending hours by a river bank. Is this a serious healthy interest or is it a fad to enable me to spend money so that I feel better in myself? Does anyone know? Oh I wish I could curl up and go to sleep.  I have to work on a \"foreigner\" when I finish today to network someone at home. It could take a while, but I can think of a hundred things I would rather be doing.  '),(4270,3532367,'01,July,2004','2004-07-01 00:00:00','   Well one out of 2 anyway!   Sandra\'s reply was \"if you need too\" when I asked if I could be inside her. I think she agreed on the basis of getting more sleep if I left her alone. As far as my sleep went I woke twice during the night, then just before 06:00 I was restless because I couldn\'t get comfortable as my joints seemed to ache.  I finished my first box of tablets today, already I have 2 more to replace them (another 60 days), but they are by a different manufacturer. I have my first appointment with my counselor tomorrow afternoon. The thought of it is starting to make me a bit tetchy.  '),(4271,3532367,'11,August,2004','2004-08-11 00:00:00','   Another argument with the phone company today and yesterday. They were supposed to have ordered a new SIM for me last week. It\'s still not arrived and I have now been told it wasn\'t ordered in the first place. They have now supposedly ordered another one, but I have to wait another 7-10 days.  '),(4272,3532367,'10,August,2004','2004-08-10 00:00:00','   Another couple of emails from Angela, the first saying she thinks she has lost me, the second saying she knows she is being demanding, but lets meet, talk and put the fun back into our lifes. This has made me happier this morning I must admit, but it is all very up and down at the moment. I am still sure it will work out in the end.  '),(4273,3532367,'10,August,2004','2004-08-10 00:00:00','   So now the new kid has been in 2 minutes, takes his coat off and says \"Has anyone boiled the kettle?\" My reply...\"Yes about an hour ago.\"  Well that starts it doesn\'t it. \"Do you book 08:00 on your time sheet, I didn\'t think you were allowed too?\"  ....The debate continues (and if you haven\'t guessed its another sore point) what with all the roadworks, poor weather conditions and time I owe due to having to go to the doctors. If I set out early in case of a bad journey, is it my fault that I get in early due to lighter than expected traffic? Or would they prefer me to come in late, just in case, and owe them another 30 mins?  FOR FUCKS SAKE!........It is me isn\'t it... I\'m just being stupid aren\'t I?  '),(4274,3532367,'09,August,2004','2004-08-09 00:00:00','   Just read an email from Angela. She wrote it last night, but it is very painful to read. She still wants us to be just friends, but nothing more. I suppose it\'s down to all the baggage I have. To be honest who can really blame her? - a single free happy life with the only problems being your own or taking on me.... and a \"married at the moment\" me as well.  As if I haven\'t lumbered her with enough, without her being dragged through someone elses divorce too.  I can\'t blame her, I don\'t blame her and if I am truthful I can\'t expect her to be there \"on my terms\" - how fair is that in a relationship?  '),(4275,3532367,'09,August,2004','2004-08-09 00:00:00','   I got very little sleep last night, after talking to Angela most of the night. It seems she wants to revert back to being \"just friends\". Its my own fault, she wanted me to spend the weekend with her so that we weren\'t clock watching. A reasonable request, but one that I can\'t make at the moment.  I want to move out, but with everything that is going on I know I don\'t have the mental strength to cope with the lonely times or the low points on my own. I don\'t trust myself. These last few months have seen me out of character....agressive, angry and tearful. It hit home with the letters last week and I don\'t think I\'ve got over that.  Today is going to be a bad day, I can tell already. I\'m not in the mood and my team leader has come in and the way he is, I can tell he\'s going to get to me. Our day sarts at 08:30, he walked in at 08:35 and the first thing he says......\"you\'re in early!\".........arrrgh! Whats the point?  IS IT JUST ME? AM I BEING UNREASONABLE?  '),(4276,3532367,'09,August,2004','2004-08-09 00:00:00','   Why does life always seem to be a case of 2 steps forward and 1 step back. I was really looking forward to chatting to Angela after so long, but it seemed as though all the frustration came to the surface. Is it because we haven\'t seen each other for over a week or is it the situation or both?  It seems we both have really strong feelings for each other, yet those same feelings seem to be creating an obstacle between us. Are we both scared? I certainly didn\'t intend to feel this way about someone so quickly, but I dont\'t want it to spoil anything either.  '),(4277,3532367,'05,August,2004','2004-08-05 00:00:00','   Think of a number between 2 and 4, then steer well clear of them. Their customer service is shit. I\'ve spent 2 weeks on the phone to India, spending anything up to 90 minutes at a time on hold. They wont put me through to the relevant department (collections) without me explaining everything to 3 other people. I\'ve been cut off 4 times while on hold - make that 5....   THEY DONT KNOW WHAT THEY ARE DOING - ARRGH!  Call 6 today - \"Will I be getting a new number?\" \"Yes!\" \"Last time I rang they said I could keep my old number. \" \"No actually you will get a new number.\" \"I\'m not happy having a new number.\" \"My mistake you can keep your old number.\" \"Sorry, but I dont believe you, can you put me through to collections in the UK please?\" \"No you can\'t speak to collections.\" \"I am the customer, I want to pay a bill, put me through to collections please.\" Silence......\"OK I will put you through, please hold..........\"  '),(4278,3532367,'05,August,2004','2004-08-05 00:00:00','   For those of you who may be wondering, yes thats right there has been a distinct lack of sex recently. No mention of it, means that I haven\'t had any thoughts or actions in that area.  Why?  Well it may seem like a strange concept to those looking in on the life of a married man, but Angela has been on holiday and any previous thoughts or needs I have had to just get my end away with the woman lying next to me have gone.  Don\'t ask me why, it all seems to be subconcious. I could say the reason I think it\'s happening like this, but the mention of the word or words would be unfair and sometimes is just plain scary. I also want to be really sure of myself.  '),(4279,3532367,'05,August,2004','2004-08-05 00:00:00','   Things seem a little quieter at work this week. After running round like a tit, all last week it comes as a refreshing relief and has allowed me to catch up on important things like blogging!  I\'ve still got my sick note and it seems to be helping me to cope (knowing I\'ve got a saftey net I suppose). I\'ve also just had an email from Angela and it looks as though she will be home early due to the bad weather, so that in itself is a help. I would love to get over and see her before the weekend, so fingers crossed.  '),(4280,3532367,'05,August,2004','2004-08-05 00:00:00','   This weekend has flown by. I was working in Luton on Saturday, in scorching temperatures, it turned out to be a long slog, but it kept me busy and active. Sunday was spent down the gym and doing housework - lawns, washing, cleaning floors, washing up, ironing, dropping films off etc etc. Before I knew where I was it was 20:00 Hrs. The highlight of the day was a half hour phone conversation with Angela. It got a bit heavy, but I didn\'t mind - such are our feelings for each other. How can one person make you feel so different about yourself?  '),(4281,3532367,'05,August,2004','2004-08-05 00:00:00','   That\'s me! It\'s official!  I was having a good day today - work was OK, I was feeling good. Even had some messages off Angela.....but then I got home to two letters.  The first one (off the counsellor) said she was really concerned about me telling her of the knife incident, she had discussed it with the doctor and was strongly advised to contact the mental health team for a more urgent referral. Well that made me feel really good didn\'t it!  The second one (off the mortgage company) said that they refuse to cover me on any life policy until the outcome of any medical investiagtions, and will only cover Sandra. Basically they have labelled me as too high a risk to insure - very very scary! So even if I wanted to kill myself now I can\'t and if I die by accident my kids suffer - great!  '),(4282,3532367,'04,August,2004','2004-08-04 00:00:00','   I managed to speak to Anglea yesterday. We were both disappointed at not being able to meet up this week, but it was great to be able to hear her voice again. She sounds like she is having a great time with the friendly locals - hehe! It helped me get through the day at work too. Every time I felt as though I was slipping, I would just think of this sweet accent.  It was the MIL\'s birthday as well, so I popped in to say hello and to see the kids. They were really great - had loads of hugs and I even stayed for tea. Sandra left early to go to her fitness class, so I was able to eat my chinese in peace. By the time I had finished it was time to go down the gym and work up a sweat, although it didn\'t help me sleep any better!  '),(4283,3532367,'04,August,2004','2004-08-04 00:00:00','   I had my latest appointment last Friday. It was another waste of time. We went through the basics and although she didn\'t ask the right questions again, I made a point of just telling her all about the last few weeks in one big swoop. Right at the end of my \"little bit\" I happened to mention that I had had a letter from the mental health team telling me I was on the waiting list. They expect to see me in 3 months time, although haven\'t issued a date yet.  It seemed to act as a cue for her to say \"right well in that case its pointless starting anything with you, do you want to make another appointment with me or just wait until the referral comes through?\"  I chose to wait - stupid fucking incompetent cow! That might be a bit harsh, but its the way I feel.  '),(4284,3532367,'03,August,2004','2004-08-03 00:00:00','   She is away at the moment, camping in Devon....is that why we have rain and thunder storms? - LOL.  I really miss her, we have met once since our \"split\" and we got on really well. We went down to her local pub on a glorious evening and spent the night catching up over a pint! It was good to be able to see her and talk to her properly. We had been limiting ourselves to the odd chat on MSN and I had decided not to be pushy and give her the space she needed to make up her mind where she wanted to be, in terms of our relationship. I am glad I did, because things seem to have turned out OK. I can\'t wait to ring her and hear her voice, but at the moment my moble phone has been disconnected, but that\'s another story!  I had really hoped to be joining her on the camping trip for a couple of days, but with the combination of poor weather, the reluctance of the doctor to give me time off and that fact that the company I work for on Saturdays have left me an urgent message to contact them in the morning about my job at the weekend, it doesn\'t look as though I am going anywhere.  '),(4285,3532367,'03,August,2004','2004-08-03 00:00:00','   So while I was off, I was beginning to get some sleep at night (well for 2 of them anyway) but I was definately getting more relaxed. Two days back at work and here I am at 02:30 again - wide awake. A pattern has been emerging since the weekend as the prospect of going back to work got closer.  Interestingly I didn\'t identify work as a problem area in my original lists. I thought I had got over a fairly significant \"rough patch\" from this avenue of my life. So why does it seem to be such a big issue again now? Is there a real problem there or am I creating one again due to the way I feel?  '),(4286,3532367,'03,August,2004','2004-08-03 00:00:00','   Went to the Doc\'s again this morning, looking to get time off work. She was very reluctant to do anything though as in her words \"it would be a slippery slope\" and I wouldn\'t be getting back for some months (until after my mental health referral at least - 3 months down the line). She did offer me 2 weeks off on the basis that it would be non extendable, but she explained that she wants me to carry on at work unless I start to be a risk to myself or others.  My options now are 2 weeks off, then back to work in the short term (probably not feeling any better) or sit it out and wait until I chin someone at work or want to jump in front of a bus, in the knowledge that I would get long term time off.  Somehow the second option seems more appealing.....  '),(4287,3532367,'02,August,2004','2004-08-02 00:00:00','   This is just a total nightmare. I\'ve come back to a desk full of clutter, as it\'s been used as a dumping ground for the last week. My line manager is just being his normal rude arrogant self and he is unhappy with my attitude towards him. Things got worse this afternoon as my manager decided that rather than discussing things with me, he would go and tell the directors..... good management skills there eh?  '),(4288,3532367,'02,August,2004','2004-08-02 00:00:00','   Helloooooo!  Well in case you haven\'t guessed by the lack of my blogs I have been on holiday. Nowhere nice, just sitting at home, but I thought I should take a break from computers as well. I have lots of catching up to do - good news, bad news and lots of events to tell you all about, so stay tuned for the updates.  '),(4289,1733718,'23,July,2004','2004-07-23 00:00:00','   I\'m currently reading The Purpose Driven Life by Rick Warren, and I\'d like to share an excerpt that I thought to be quite interesting and enriching.... made me think...  ===== He (God) created the church to meet your five deepest needs: a purpose to live for, people to live with, principles to live by, a profession to live out, and power to live on....Worship helps you to focus on God; fellowship helps you face life\'s problems; discipleship helps fortify your faith; ministry helps find your talents; evangelism helps fulfill your mission. ===== I hope this little piece has given you something to think about, not necessarily changed your whole perspective, but if anything has made your life a little more enriched.  Take Care &amp; God Bless  '),(4290,1733718,'22,July,2004','2004-07-22 00:00:00','   Just came back from the Madonna Re-Invention Concert at the Air Canada Centre and what can&nbsp; say, she is a PHENOMENAL performer.&nbsp; With a mix of her \"old-school\" and \"new age\" songs, she kept the audience on their feet.  I was blown away by her talent in playing the accoustic and electric guitars to her old songs, such as Material Girl, and was awestruck by her rendition of John Lennon\'s \"Imagine.\"&nbsp;  I don\'t think you truly come to appreciate an artist till you\'ve been to one of their live concerts.&nbsp; I saw a side of her that enabled her to connect with her audience, her fans here in Toronto.&nbsp;   It was a great concert all together, the energy, the excitement, the anticipation of what the next song was gonna be and how she\'d perform it.&nbsp;  '),(4291,1733718,'19,July,2004','2004-07-19 00:00:00','   After an exhausting weekend moving things from upstairs to downstairs, I\'ve come to the realization that there is still soooooo much to be done before my parents leave for California.&nbsp; They only have about 9-10 days (including today)!!!&nbsp; The frustrating things are not so much what has to be done in the basement, we can get that done when they\'re gone, but the things that must be moved to the basement (practically all the contents of the kitchen, bathroom, etc.).&nbsp; It is tremendously exhausting. &nbsp; And even when all this is set and done, there\'s still a matter of sitting down with my parents to decided who\'ll take on what responsibilities like maintaining my parents banking accounts, my grandma\'s, backyard and frontyard maintainance, monthly bills, etc.&nbsp; So although freedom seems to be coming around the corner, responsibilities are looking to take over.&nbsp; It\'s kind of bittersweet, gaining the freedom of living in our own \"place\", yet taking on the responsibilities of the entire household. &nbsp; Oh well, we\'ll see how the transition&nbsp;goes through.&nbsp; Only time will tell.  '),(4292,1733718,'03,July,2004','2004-07-03 00:00:00','   I would have liked to start blogging \"officially\" on July first, but due to circumstances that WERE in my control, I figure, oh well.....  So here it is, the first BLOGG!!! (woohoo!) Canada Day 2004 will without a doubt go down in my books as a memorable one.  Actually, the festivities started a little early this year thinking I\'d take July 1st as a day off to recouperate before coming in (reluctantly) for work the next day....Uhmmmm didn\'t happen :P  Here\'s what did...  June 30: PreCanada Day BBQ Great food, Great Drinks, Great Music, Great friends!!!  And all set to a Hawaiian/Beach-type theme. Thanks to Karlo, Gerard, Laura, Don, Jen, & Rodney for coming out.  July 1st: Canada Day  Don, Jen & Rodney stayed over.  I Woke up with a mild headache (didn\'t drink as much as I would have liked to) but nothing that a good cup of java and a hearty filipino breakfast couldn\'t fix.  Jen was the first up, Don passed out (way too much to drink) and Rodney just was tired out.    Fast forwarding to some night action.... Darlene & Marcella (long time highschool gal pals) came over for some pre-drinking before we\'d head down to Seven (on Adelaide and University) to meet up with Diana (another highschool bud).  On the way to, we picked up Nadine, yes another Ex-Poper like the other 3.  Needless to say, the night was truly memorable.  Pre-drinking consisted of a Sambucca shot (notice the absence of \"shots\", eww), Cran and Raspberry Vodka chasers, Tequila and Lemon Drop shots.  We had a nice buzz going.  Just when we thought the night couldn\'t get any more exciting.... my dog Tamiko managed to slip passed us through the door and bolted with the sound of the fireworks.  It took us a long and scary while for find her (shaking).  Who knew that Marcella (the sprinter), Darlene (the pick-up) and Me (the negotiator) could manage to pull this off in a semi-drunken state.  Well after this fiasco, we kinda lost our buzz (darnit all).  My sister was gracious enough to drive us downtown after what happened and we arrived at Seven just as the club was starting to get busy.  The rest of the night seems like a blur, and I\'m still trying to piece most of it together....  Like I said earlier, this one is definitely \"ONE FOR THE BOOKS\"!!!  '),(4293,4022192,'24,July,2004','2004-07-24 00:00:00','   i figure if you wanna get to know me, you gotta know my friends first. thoughit\'s a little more complicated than that. you see, i call myself a floater.not because i can swim or anything, which, as terrible as it is, i cannot,but because i travel between different groups of friends. sure i miss somethings. but hey, i love all of them in their own ways, and i don\'t not mindnot being a part of only one group, i love it that way.    Group One  description: seven mostly cute, straight-edge, soon-to-be college freshmen (except one), one of whom is my  urlLink boyfriend .  reasons why i love them: eight boys + one girl = attention galore. greatmovie-partners, advice-givers, and cuddle-ers. easy to tickle. always a sourcefor good \"clean\" fun (though all have dirty minds). can feel like i\'m corruptingthe innocent.   reasons why i don\'t fit in: not a guy. i like to, as they call it, party.i have created some jealous tensions, and i have more sexual experience thanten times them all combined.    Group Two  description: five fabulously fun, dare-to-be-different, new york city girls, all except one my age.  reasons why i love them: ROCKY HORROR GAL PALS (equally in love and obsessedwith both the cast and the movie itself). actresses all. intelligent conversationalists.internet trend-setters. seem to feel blessed to have me. best underwear partiesever.   reasons why i don\'t fit in: they\'re too damn intelligent. while i maybe up to their acting, i cannot sing for the life of me. i cannot keep upwith their trivia and have not been around from the beginning, and in fact,was hated in the beginning.    Group Three  description: three almost as short, same-aged, fly chicas.  reasons why i love them: goofy as all hell. love categorically girliethings (shopping, boys, etc.). always excited for a night out. have wonderfulsame-size-as-me clothes.   reasons why i don\'t fit in: they\'re too damn pretty, and i\'m afraidof competition over boys, grades, etc.. i have friends of whom they don\'tapprove.    Group Four  description: too many to count riotous alchies and the like. aka, any thing\'s funny when you\'re that fucked up.  reasons why i love them: source for any drug, and center for drunkendebauchery. understand a need for \"relaxation.\" parents are always away.serve up cute guys and good times.   reasons why i don\'t fit in: have drifted away from them recently findingother passions. i care too much about school and am never up on the gossip.   Group Five  description: two i-hate-everything-so-choose-to-be-outcast sorta-punk boys.  reasons why i love them: really chill people with interesting and differenttheories on life. fun to scare people with. good music and movies.   reasons why i don\'t fit in: rather soap-opera-like relationship withone of the two. i have other friends besides them. i do not hate the worldor myself (most of the time).    Group Six jess  *group order does not at all mean order in my heart  **royal apologies to those who object to being placed in stereotypicalcategories such as these. you have my permission to hit me.   ***there are also some glaring exceptions of which i am fully aware,people are part of a group but who do not reflect the picture of it i havemade. you know who you are. please don\'t kill me.  '),(4294,4022192,'24,July,2004','2004-07-24 00:00:00','   at the request of some very wonderful people (one in particular), and as you, my readers (that\'s right, all two of you), can see, i have started a blog. now, as a self-proclaimed commitment-phobe i must first warn you all that you should not expect consistent updating. you see, it is my belief that the people who would have the most to write about are too busy doing it to actually be writing about it. this is not to say that i recently have nothing to do, just that i\'ve decided to devote an occasional hour of my life (from sometime between two and four in the morning) to the maintenance of the blog in question. but i digress. i also probably spelled that wrong. that is why for this said entry i shall propose a disclaimer:  warning. the material you are about to read (maybe next week) is not suitable for people who don\'t already know and love me. for only those eight million will be able to forgive the constant spelling errors and probable obsessiveness that shall comprise my ramblings to come.   my advice? sound stuff out and skim (says my insecure self)....please love meâ€”iâ€”meanâ€”it (says my vulnerable self)....suck it up and die (says my irritable self)....but most of all, read and love because anything i write turns to pure gold (says the happily me self)  '),(4295,4022192,'24,July,2004','2004-07-24 00:00:00','   this is just a test. please remain calm. i repeat. this is just a test.  '),(4296,4022192,'24,July,2004','2004-07-24 00:00:00','   she wakes up. gently batting her eyes, the color has been drawn from her face (and the bed and the carpet and the...). \"i just had the most wonderful dream\" she says, bewildered, (pointing to the other people in the room) \"and  urlLink you  were there, and  urlLink you  were there and  urlLink you  were there....\"   but unlike my southern friend, mine was not a dream. eeeeeee.  '),(4297,3069909,'07,April,2004','2004-04-07 00:00:00','   Okay I made it... so does anyone else actually use this yet?  hello? (hellohellohello?) echo! (echoechoecho!) '),(4298,3069909,'03,April,2004','2004-04-03 00:00:00','   i just want to say that i think this is the beginning of something good.  there, i said it. '),(4299,3069909,'02,April,2004','2004-04-02 00:00:00','   so who wants to talk some spiritual talk? '),(4300,3069909,'02,April,2004','2004-04-02 00:00:00','   Im in this is Craig '),(4301,3069909,'02,April,2004','2004-04-02 00:00:00','   This is for the Old Bald Guy with an inspiring vision. Blog Away Chuck! '),(4302,3069909,'25,May,2004','2004-05-25 00:00:00','   Hey UH-GNAW-KNEE-MOOS  .... can I shorten your name to UGKM?  If you post by signing A? How will we know who the real UGKM is?  Any Anonymous bloak could come pretend to be you even though you aren\'t really any identity... you could quickly get a really bad identity....  so if its agreeable to you, perhaps those posting and reading right now can come to know you as UGKM. (Of course I guess that means now you have an identity?) What a condundrum. Glad your here.  In fact, I think I\'ll change my identity while I\'m at it.  From now on I\'d like to be just \"B\".  Anyone want to be \"C\"?  TTFN - B. '),(4303,3069909,'24,May,2004','2004-05-24 00:00:00','   ANYONE under the sun can add comments to a post, they do not have to be a registered user.  All you do is click on the \"comments\" text at the bottom right corner of a post.   AND MOST IMPORTANT When you are \"Creating a POST\", you need to click on the \"More Post Options\" just above the \"Publish Post\" button.  Then you can select to \"Allow Comments\" to the Post. '),(4304,3069909,'21,May,2004','2004-05-21 00:00:00','   Growth demands a temporary surrender of security. -- Gail Sheehy '),(4305,3069909,'21,May,2004','2004-05-21 00:00:00','   Has anyone else seen   Man on Fire  ? I saw it on a Sunday afternoon--a time when my emotions are typically raw. Everything about the movie was cliche except . . . why was this burned out, alcoholic, ex-CIA able to quote Scripture and why did he read the Bible? Who put that in the script? Why?  Something got stirred in me and I felt sad, like I was watching my own life through a window. A moment came when God said to me, \"Pay attention to this!\" I knew that the next couple of lines were going to be really important, so I listened carefully. One of the odd characters in the movie spoke a sentence that was God\'s message to me. I was bowled over, and so thankful that God would let me be there, alert me to the fact that something was coming, then hit my heart with a message He had been working in me for several weeks.  Want to know what it was? '),(4306,3069909,'27,June,2004','2004-06-27 00:00:00','   For those who live in the world and desire to embrace true virtue, it is necessary to unite together in holy, sacred friendship.  -St. Francis de Sales '),(4307,3069909,'25,June,2004','2004-06-25 00:00:00','   Originally posted by: overflow At 2:36 PM,   overflow  said...  question. what is an evangelical christian? is immersion part of an evangelical community and if so, does immersion have a cohesive core belief and/or purpose with this? '),(4308,3069909,'21,June,2004','2004-06-21 00:00:00','   \"When I saw others straining toward God, I did not understand it, for though I may have had him less than they did, there was no one blocking the way between him and me, and I could reach his heart easily. It is up to him, after all, to have us, our part consists of almost solely in letting him grasp us.\" -Rilke and Benvenuta: an Intimate Correspondence '),(4309,3069909,'15,June,2004','2004-06-15 00:00:00','   Below is a teaser to the article you may want to check out (click the \"Link\" above).  I\'ve never been really interested in trying to wrap my brain around how science proves or disproves the existence of God.  I am completely mystified still by the whole concept of radio in my car.... so this article mostly just furthered my belief in the mystery of God.  Yet somehow it helped my logical thinking understand the co-existence of our Free Will and God\'s Predestination.  I am interested in hearing if anyone else knows this stuff or cares or already gets it..... whatever.  I feel more enlightened but I also feel like it\'s worth checking out....  The teaser to the article on the link.......  We\'re talking about the ancient smackdown between God\'s resolute, unchallenged predetermination of all things and, on the other hand, humankind\'s unfettered right to choose our own destiny. In one corner stands God\'s ultimate prerogative to rule; in the other, man\'s primal birthright as a free-agent.  But here, in little more than the time it\'d take you to delete the spam in your inbox, you will encounter a mesmerizing synthesis of science and philosophy that will forever change the way you look at the universe. '),(4310,3069909,'14,June,2004','2004-06-14 00:00:00','   After digesting the discussion on saturday I thought that I would give my perspective on the topic clip.  When I first saw the Animatrix, Matriculated jumped out as something that was just ripe with Immersion-style content.  In fact I felt that the clip was a handy metaphor for the Immersion dynamic.  The first important factor which may not have been made absolutely clear in the selection itself was the setting.  Being that it takes place in the world of The Matrix, the machines and the humans are in a state of war.  Those humans we saw in the clip were only in their position after having escaped imprisonment by the machines.  They all had existed in a world that was FORCED upon them by the machines in order to keep them enslaved.  Even with this in mind they still choose to try to create a dialouge with their mechanical counterparts as a means of creating resolve.  The basic form of interaction between the human and the machine is to seek to kill the other for their own survival.  The idea that this small renegade band is struggling to find peace in the midst of such chaos was a worthwhile discussion in and of itself.  We the viewer witness one encounter between these humans and a pair of scouting machines.  It is implied that the machines are scouting for the purpose of eradicating the humans (the transmitter being placed to summon reinforcements).  When the humans gain the upper hand, they manage to take one of the invaders hostage (with the help of a machine fighting and giving its life on the side of the human cause).  This begins our focal point for the story.  What follows is an attempt on the part of the humans to communicate with (not reprogram) the machine mind.    The director specifically states in an interview that the method used was that of the human dream state.  He also clarifies that the machine represents a \'rigid mind\', an unbending mentality rather than simply a \'robot\'.   The protagonists of the story had to create a common ground in order to inact a constructive dialouge.  They had to envision a \'language\' which they could use to reach those who would seek to destroy them.  The war metaphors being used here are to exagerate the problem at hand.  How do we communicate with someone/thing different from ourselves?  They make the point that coersion is ultimately useless because to force is just another way to destroy.  If they can find a common ground with the machines, only then will there be hope for peace.   The peice is concluded with a very important image.  It is the machine waiting on the shoreline in same place where it first encountered the woman in the beginning.  To me this was the most telling part.  After all was finished, and everyone on both sides lay dead, this lone machine carried on the same vigil of the humans he had destroyed.  There can be no more certain way to show the decision made by that machine.  No one was left to pressure, coerce, threaten or controll him and he still chose to wait on that beach to show the next one the things he had learned.  I thought this was so applicable because as christians we fall into the same patterns.  We tend to try and fight or reprogram those who believe differently than us rather than find a common ground and share experiences in order to communicate and find peace. '),(4311,3069909,'12,June,2004','2004-06-12 00:00:00','   Email From Will Smith to Immersion Team:  I have been searching  the entire internet for culturally relavent bits and pieces of post modern  expression. I am proud to anounce I feel I have finally succeded and I am proud  to share my glorious findings with all of you.  Please give me feed back and let  me know how we can effectively use this golden nugget of media genious. I think  this could be a good topic for a saturday night.  Copied to Blog by: '),(4312,3069909,'09,June,2004','2004-06-09 00:00:00','   Today I was having a conversation with a co-worker when I had what I call a \"spiritual experience\".  The content of the conversation is not important. But the experience was.  While I was speaking to her I began to visualize a spider landing on her and spinning a web. The spider deposits its first blob of sticky threadlike filaments and begins going from post to post to deposit additional blobs of sticky webbing until ultimately the spider has spun a complete web and is satisfied that its web is perfect for ensnaring the food required to sustain its life.   What I saw in that picture was how my conversation was like that.  I engage in conversation with someone and it deposits a sticky residue of me on their soul. That person goes from post to post depositing parts of that conversation, each time creating a stronger more viable web structure.... but for what?  What is the web for... ensnaring what?   I feel the web can represent two things: it can ensnare and kill its prey. Our conversations with people can indeed ensnare and kill people - their passions, dreams, hopes, trust, freedom... etcc.   But I also feel the web ensnares to fortify the life of the spider.  And so too our conversations with one another can fortify, nourish, edify the life of others.   What I also got from this experience is that my relationship with God is alot like this.  He deposits these sticky filaments in my life that represent a time when He connected with me for a \"purpose\" and then as I go to and fro, from post to post trying to figure out this purpose, the web is spun..... when the web is completed: I am either at the end of myself and surrendered to the outcome like the spider\'s prey -  OR I am like the spider, sustained and fortified until it\'s time to begin another web.     Anybody else feel like a spider? '),(4313,3069909,'08,June,2004','2004-06-08 00:00:00','   Check out Brian\'s link. He is going to be with Immersion on Saturday June 19th and the 26th as host for the evening. He has a wealth of knowledge we need to glean and be challenged with.  Also, below is the link to his book recommendations.... some of them look really interesting for me, I thought you might like to check out what the emerging church leaders are reading to inspire their passion.  http://www.anewkindofchristian.com/recommendations.html '),(4314,3069909,'08,June,2004','2004-06-08 00:00:00','   I was just wondering, if Vegan\'s are pro-life... meaning \"for all living creatures\", which is why they don\'t eat things that were alive at some point.  What do they think about \"meat-o-sauras\" animals.  Ya know, animals that prey and feed on other animals?   And.... do they have pets?  Cuz if they had like a dog or a cat, what do they feed them? Simulated protein... cuz most of the cat/dog food I see in the markets has meat byproduct in it. And if it is simulated protein, isn\'t that kind of unfair to the dog/cat to put their \"vegan\" lifestyle on their pet. I mean, it\'s not like God gave the dog/cat a brain to be able to decide whether or not it was cruel for them to eat other animals... it\'s just their created nature to do so.  I\'m just curious about this...  anybody know any Vegan\'s?  P.S.  Are Vegan\'s against abortion then?  Hmmmmm. '),(4315,3069909,'08,June,2004','2004-06-08 00:00:00','   Dear Neohippie  I wasn\'t there last Saturday night so maybe commenting is out of line, but here goes anyway.  O.K. I get that your pissed off. I\'m not clear at who, what or why?  1. Are you angry that in a theoretical exercise on how to best spend 50 billion dollars to solve the worlds problems someone came up with the     lame idea of translating bibles into different languages? 2. Are you angry that some christians are still in favor of evangelism? 3. Are you angry that in an experimental evening the format required small group activity? 4. Are you angry that well fed christian missionaries risk their lives by going into third world countries to torture starving children with meatball subs? 5. Are you angry that you argued with your mom and you think she is disappointed in you?  Seems to me getting angry about children dying, human suffering and injustice in the world is an admirable quality.  How would Jesus spend 50 billion dollars? Hmm... My guess is he would give to Caesar what was Caesar\'s.  My question to you neohippie is how can you make a difference in a world of need? '),(4316,3069909,'03,June,2004','2004-06-03 00:00:00','   Okay, so after digesting our most recent (and very weird) Immersion session, I wanted to blog something.  Isn\'t it interesting how, depending on the group of people we\'re with, we seem to shift back and forth between being \"conservative Christians\" to being no-good-dirty-hippie-liberal-heretical beatniks. Hmm. It seems as though we all had someone in our small groups (did I mention I hate small groups?) that really believed the best way to spend 50 billion dollars was to translate bibles into every language and then set forth into the world converting people, in hopes that before the poverty-stricken, homeless, starving, 3rd world peoples of our globe  DIE , we can convert them to Christianity. HMM. So.... Okay, I get it. We should also make sure that the money is spent nourishing the good Christian missionaries who do this, so that they are healthy and fed and physically ready to take on this task. THEN, when they get there, they can dangle their foot-long meatball sub in front of bloated ethiopian children and tell them that if they promise to believe in Jesus, they\'ll get a sandwich!!   Wha? Alright, so I  don\'t  get it. I argued with my mother about this when I explained the service (and the aftermath) to her. I often argue with my mother about Immersion. She\'s disappointed in me because I\'m into all this \"new-age crap.\"  Isn\'t a better example of God to care for the needy without putting requirements on it? To love unconditionally? Wouldn\'t it represent our Most High God BEST if we let our beliefs shine through our actions? I\'m not saying we should never mention Jesus or pretend that we\'re something we\'re not. But wouldn\'t someone who hasn\'t eaten in weeks and sleeps in the mud rather see random foreign people giving away food, clothing, medicines instead of listening to strangers preach at them about a God they\'ve never heard of? Look how pretty our God is... he has blond hair and blue eyes and speaks with a lovely british accent.  But then, I guess that\'s just me JUDGING these other Christians. Now its someone\'s turn to judge me for judging them. Then I can judge you back for judging me for judging them.  What would Jesus do with 50 Billion dollars?  neohippie '),(4317,3069909,'03,June,2004','2004-06-03 00:00:00','   I was sitting with my thoughts one day, thinking about some of the different paths I visited and these thoughts started rolling something like this...  What would you say if today was the day that you didn\'t see tomorrow? Where would you go?   How do you know?     Who would you see?       Did he hang from a tree? Some people say there\'s a shiny white light up at the end of a tunnel. Now let me see the opposite would be getting sucked down a black funnel. Others have said your tied to a thread, your going to get recycled. A princess   A queen     A jellybean? You can\'t really mean things that aren\'t seen, go and get you frightened. Orange and red, a Lama once said \"time to get enlightened\". Follow your bliss was really a miss, addictions get real boring. Try follow your fear, for it draws you nearer to the lessons that are adoring. Nirvana was the scene, if you know what I mean. His music led to disaster. Don\'t waste your time standing on a life line, shortcuts are always faster. If you are cool, yogas the rule, stretching isn\'t traggic. Be a goddess instead but don\'t lose your head, witchcraft is really magic. The prince of thieves is disquised as these, laughing all the way to the bank. All you must do is get a clue, assurance is what you seek. I know there are times, TBN signs, people that look like freaks. Don\'t be misled, the book should be read, the contents may surprise you. It speaks of a man, who walked the land, parables were his bag baby. The headlines read, 5000 fed, he rose the dead, he also said,\" I am the way, the truth, and the life\". With his healing touch, lepers lost there spots, the blind were made to see. The message was clear for all to hear, he\'d come to set us free. The court of the day, in a media way, just like today, had their way.   They said that he lied.     Let him be crucified!       Hammers fell,         pungent smells,           crimson tears fell on deaf ears.              They pierced his side.                 He let out a cry. Forgive them Father, for it was no bother, I\'d do it again to save all men. My blood on the cross, was no loss, it has set things straight. In three short days, I\'ll stand at the gate and wait. Now that you know, will you go with the flow? Take the time to bust a ryhme. Or won\'t you please,    fall to your knees,      and take hold of the hand,       the hand of the man,         the Holy Lamb of Galilee. '),(4318,3069909,'31,July,2004','2004-07-31 00:00:00','   I have an idea that when longblogs (I mean it as one word) show up ... like the last one of mine,  Emergence-y!  ... some people may take it as though the writer\'s intent was, \"Well! That about sums it up. No more to be said. (And don\'t even try, please.)\"  Even if that were so (it shouldn\'t be), I hope everyone would see  any  posting as never really \"done,\" more like an opportunity for thought and reflection, tasting and sampling; if it\'s a long post, then it can be a buffet table, a sort of smorgasblog where you can select any little tidbit that strikes your taste. Maybe none of it; maybe only one word or turn of phrase, or a single illustration.  But you can add to the thoughts, as many do, and surely you have insights that no one else has. A painting is always more than the artist intends, music more than the composer or musicians express, words more than writer may think is being said. One person may offer some thoughts;  you  have opportunity and insight to open any of those and take others still deeper, a door whose inside is bigger than the outside.  Is it about who has the last word? Maybe it\'s that  no one  has the last word, there\'s always more to explore and express. If I go on for pages, you could take the smallest phrase and open it for others to your own insights, wider and deeper than all of it together!  the well\'s mouth is small, the ocean\'s face so endless. but how deep? immerse.  ---seussac '),(4319,3069909,'30,July,2004','2004-07-30 00:00:00','   Here is a story review that might spark some interest:   \"Hell is the Absence of God,\" on the other hand, is very much an attempt to examine the idea of faith, specifically by imagining a situation in which faith is no longer a part of religion. In our world, religion relies on faith because definitive proof is lacking. This lack of proof allows some people to reject one religion and choose another based on which makes them feel better, e.g. \"I don\'t like the judgemental god of Religion A, so I\'m going to worship the kind and gentle god of Religion B.\" We have that option because neither deity is unambiguously present, but if a particular god were here right now, we\'d have to deal with him whether we liked him or not; faith would have nothing to do with it. I thought that would be an interesting scenario to explore. '),(4320,3069909,'26,July,2004','2004-07-26 00:00:00','   \" It\'s not our sin that hurts God, it\'s our unresolved pain. Sin is the result of not knowing how to find the transformation of the human condition.\"  Richard Rohr- '),(4321,3069909,'23,July,2004','2004-07-23 00:00:00','   We talk about the \"emergent\" church: it might be worth thinking about that  emerge  itself derives from \"to come forth from  immersion \"---which resonates delightfully, and as I&nbsp;see&nbsp;it,&nbsp;even in a kind of&nbsp;providential artistry, with this group\'s name. Was that thought of when the name was chosen?&nbsp;We go in, I hope to see people emerge as something newer than when&nbsp;we came.  Before about 1425,  emergent  also referred to the year of the Israelite exodus---and I hope that many people here&nbsp;are also finding their own exodus into clearer perspectives of God, themselves,&nbsp;and one another. (Hopefully without plagues of flies and such. The incense might help keep them away.)&nbsp;On the other hand, an  emergence  can refer to a plant outgrowth such as a thorn. The question there is: Whose side are we in (not \"on\")?&nbsp;(And are we maybe  supposed  to be there!)  And even the alarming word  emergency  began simply as referring to an  unforeseen circumstance--- but like so much else in life, maybe really not so unforeseen, if only we had recognized how a genuinely normal part of life is that much often comes from little. A seed somehow ends up as Jack\'s giant beanstalk (or a giant redwood); water, air, heat, and a spin from the Earth are all that make up our endlessly complex weather.  John Holland\'s fascinating book,  Emergence: From chaos to order,  is a marvelous adventure in that kind of complexity---the&nbsp;notion of much coming from little, of systems \"where the behavior of the whole is much more complex than the behavior of the parts,\" and where \"the deeper explanation, once understood, inevitably gives the imagination an exhilarating boost.\" Which would be good to have in understanding our own lives!&nbsp;Things we might puzzle over as confusing---the sprawl of a rain forest or a city, the churn of traffic or just of our own lives---can take on a whole new perspective,&nbsp;with even an exhilarating jolt of recognition,&nbsp;once we step back to see order emerging&nbsp;from confusion.  Which, it might&nbsp;seem, is&nbsp;easier to see in a rain forest than in our own&nbsp;(or others\') lives---until we also step back to see the potential in each person\'s life. There is a fresh view of that&nbsp;by Roger Lewin and Birute Regine, in their book with the really long title,  The Soul at Work: Listen ... Respond ... Let Go: Embracing complexity science for business success.  Once you get past that, and the maybe drab sound of \"business,\" they&nbsp;have a&nbsp;refreshing view of creative potential in  any  groups, drawing on observations of animal groups, such as a flock of birds. \"The complex behavior of the system as a whole---the coordinated motion of the flock---emerges from a few simple rules of interaction among individuals, not from a single leader.\"  So what&nbsp;about all this? So they conclude by thinking about  emergent teams;  see for yourself how this sounds, considering what we\'ve hoped to&nbsp;do in Immersion: \"In emergent teams, everyone jumps into action; everyone intuitively knows what needs to be done; different people guide the collective action at different times ... and it\'s all done in the spirit of goodwill and care. And the results are often beyond the imagined .... This capacity in people, and in teams, to [break] through the normal day-to-day structures is the soul at work.\" Actively including diversity, especially diversity in ideas; openness for people to contribute and have a say; the central need for caring relationships; the importance of story in building understanding and trust---these are some of&nbsp;their pictures of&nbsp;how&nbsp;this idea looks in action, how complexity emerges from individuals\' gifts.  And this sounds a lot like what I hear people hoping for (and often practicing) in Immersion. If any of this sounds like what you\'ve wanted to do, or see, or be a part of, I hope you dive in and add your part, because no one can tell how important your part is until it\'s in the mix; and it takes  all  these unique lives for us to see how the whole design will emerge. The whole is very much greater, richer, deeper, more complex,&nbsp;than the sum of the parts.&nbsp;Lewin and Regine were told,&nbsp;with regard to teams, \"If someone asks who the leader is, then that person has never been on a good team.\" What\'s important in this \"team\"? Look around: you\'re immersed in it. '),(4322,3069909,'19,July,2004','2004-07-19 00:00:00','   During the modern era, \"story\" as a form of communication was minimized as compared to \"report,\" \"essay,\" \"argument,\" and \"documentation.\" Intelligent people seemed to assume that if you wanted to entertain, then tell a story, but if you wanted to get at the facts, you should present an argument. Therefore a scientific report that was well-researched was a better medium for communicating truth than a story. &nbsp; The modern assumption has been challenged in the last two or three decades. We realize now that a good story (that is, one that obeys the rules established by the story world and genre--fairytale, sci-fi, murder mystery, etc.) can contain and communicate truth as well as any other form of communication. In fact, many scholars (including cognitive scientists) believe that \"story\" is necessary for human thinking, for ordering our worlds, and for forming a rational and coherent mental map. We need stories to&nbsp;make sense out of the events and circumstances of our lives. Stories help us figure out how everything \"fits.\" &nbsp; I\'ve been reading N. T. Wright\'s,  The New Testament and the People of God  (and loving it!). Here\'s a quote: \"Stories are a basic constituent of human life; they are, in fact, one key element within the total construction of a worldview.\"  &nbsp; I don\'t know how to fully articulate this yet or explain why it is so, but I can feel its truth. My worldview is my total story, my explanation of reality. &nbsp; obg '),(4323,3069909,'16,July,2004','2004-07-16 00:00:00','   How did you first hear of Immersion (Saturday night meetings/community)? When did you decide you wanted to be part of it? What does it mean to you? What do you look forward to when you go? What do you dread? &nbsp; This Saturday night we want to think about the Immersion narrative. In a way, it is many different narratives, because we all have different stories. But what is the one common story that holds us together? Granted, we are now writing the first chapter, but still something is beginning to emerge. We love each other, enjoy being together, and sometimes even find as much fun and benefit from the things that fail as the things that succeed. &nbsp; If you have a story, email it to me or else respond to this blog. It would be really nice if I heard from you by this evening (Friday, or tomorrow morning at the latest). &nbsp; Thanks! &nbsp; obg '),(4324,3069909,'10,July,2004','2004-07-10 00:00:00','   \"There is no use trying,\" said Alice. \"One can\'t believe impossible things.\" \"I daresay you haven\'t had much practice,\" said the Queen. \"When I was your age, I always did it for half an hour a day. Why, sometimes I\'ve believed as many as six impossible things before breakfast.\" --Lewis Carroll,  Through the Looking Glass   Sounds perfectly reasonable to me. '),(4325,3069909,'06,July,2004','2004-07-06 00:00:00','   Kick   ASS,   Jyro! The song ROCKS! It\'s exactly perfect for Immersion. I\'m very impressed. Can\'t wait till everyone hears it!  Also, I\'m really digging the new look of this site. Very cool.  I\'m so excited. Things are coming together, guys!   neo '),(4326,3069909,'02,August,2004','2004-08-02 00:00:00','   dominic just emailed me your emails about Saturday night\'s topic of sin.  i was gonna email back but got a little scared i was gonna be a perpetrator of an email flurry.  so i\'m putting my response here, my virtual saturday night replacement until my show closes (blatant plug: if you didn\'t see Dominic direct my play \"Evolution of Sunday\" in OC, come up to LA and see the LA version!  urlLink www.splitid.com . it\'s about god and guilt and stuff. buy tickets at  urlLink www.lastagealliance.com  or tell the house manager that night that you\'re from Immersion and I\'ll have a comp ticket for you.)  Okay, back to topic....  Here was my e-mail and two cents....  I so wish I was there! Sin is my favorite topic!  How funny that I was trying to find the term for sin that other religions use and all I could think of was Buddhism and I guess it\'d be \"suffering\" and wasn\'t even thinking about the Rohr quote.  I remember a philosophy class where we debated sin vs evil and this class had every kind of religion. It was impressive how the non-Christians felt so much more comfortable about the topic than the Christians. Maybe because there\'s no spiritual investment in the term. No baggage? Either way, it certainly added clarity for those who called ourselves Christians.  I heard someone say that they thought of a person\'s spiritual journey as a straight path with a light up ahead. And if you keep your eye on the light and keep walking, it\'s all you can do..it\'s all you need to do. It\'s really all God needs to help you get to heaven/nirvana/bliss/happiness/etc. Keep your eye on God, move forward, and the rest is taken care of.  So on occassion, you\'re on the road and you stumble. But you keep walking. That\'s a sin. You tripped. You hit a little pothole on the road where your eyes weren\'t on God, your eyes were on the ground watching the dirt meet your face. And the result can be a skinned knee or a broken foot, but either way, you heal and the only one you hurt is yourself.  On that note, evil was walking on that path and making a conscious decision to take your eyes off God, turn around, and walk the other direction.  I think we\'re sometimes afraid God won\'t know if we\'re tripping or walking the opposite direction. We think sin is gonna put us on the judgement list. And judgement is for evil. Judgement is for the person who doesn\'t want to be on the journey. Most (everything?) Jesus talks about when he talks about wrongdoers is about intent. So much of his teachings were \"They do this, but their hearts...\" It\'s about where your facing, not what you\'re doing.  (personal belief is I think hell is the life you lead with your back to the light).  So anyway, as a Christian, I often get confused and think, \"If I trip, God will see it as me intending to leave Him. But really, I just didn\'t know I was gonna trip. It took me by surprise. But God won\'t understand! I was on my butt! I meant to be on my feet! How can I tell God I didn\'t mean to!\"  And that\'s just silly. Since God is rumored to be fairly smart. '),(4327,3069909,'02,August,2004','2004-08-02 00:00:00','   \"We must be the change we wish to see in the world\"       -Mahatma Gandhi '),(4328,3595127,'16,June,2004','2004-06-16 00:00:00','   Feeling a wired feeling today. Kind of sad or low, one of my mood swings I guess it is happening a bit more often than it should now days. I donâ€™t like myself in such moods because then I become reclusive and donâ€™t like talking to anyone. I just want to sit alone and depress myself more than I already am. Those reading (if any) may be wondering what has happened to but me in this black mood, well the answer I assure is that nothing drastic. Itâ€™s just that my stupid hormones are playing against me (not for the first time I assure you). Maybe its got to do with the fact that people close to me or people I have known for long have found them self hooked while I am waiting to bite (not literally)   More often that not I have been told by my friends that they feel the need to have someone special in there life and I used to wonder how could it be possible to miss a person you donâ€™t even know!! But then one fine day they would walk upto me with a smile on the face and twinkle in the eye to announce, â€œI found him/herâ€ and used to wonder how can they be sure that they are the one. Now I see them happy with there mates laughing, smiling and sharing moments at all spheres of life. My friends tell me that I am lucky for them because knowing me for a short while leads them to the love of their life (donâ€™t ask me why they say it)  You know I have a confession to make as happy as I am for my friends I feel a twinge of jealousy for not what they have but because I know that the moment someone special enters life friendship takes a back seat, not intentionally but unintentionally. You cant call up 12:30 at night to just talk about your day to them because they might be busy with their mate, you cant ask them to come out for a coffee because they might be out with their love, u cant watch a much awaited movie with them because they might want to watch the movie with someone else. All of this is perfectly understandable but does hurts a bit because admittedly I am very possessive about my friends. Not a good thing I know yet something I am. I donâ€™t know how to conclude but I just hope that when I find someone (if I ever do) I would be very conscious of the fact not to ignore people close to me because I would not want to hurt the ones I Love.  '),(4329,3595127,'15,June,2004','2004-06-15 00:00:00','   I liked being a kid because I could be carefree and the only tension in my head was that how could I convince my brother that i was better than him, or convince my parents to get me a new doll, or how i would finish my homework in time to go out and play till late. But things are different now, at the age of 22 I am getting to know the ways of the world. Not that it is always bad to be a part of the grown up world. I admit it is fun at times but more often than not it is very complicated and tough. I did not understand it to begin with, i am not sure if I understand it now. I do not understand why certain things that seem so trivial shape up to be the discussion point for days. I do not see what harm can be done by getting wet in the rain. I do not see the logic of telling a girl not to laugh too much, I do not see the resons that tell me that I talk too much(maybe true :-)) But what I do know that these things are something I did a lot when I was a kid and nobody stopped me then, so why now?  '),(4330,3861317,'30,July,2004','2004-07-30 00:00:00','    I\'m back to myself as early as yesterday. Amazing coz normally I took longer period to recover my anger + bored . Yeah, I guess when you have someone so dearly in your heart, things will be very much easier...Thanks so much&nbsp;for being there when I most need him...  And last Tuesday, I wasn\'t plan to go anywhere&nbsp;but I just can\'t wait for meeting someone. So, KLCC &nbsp;was the destination. Guess who..??&nbsp;Norarnida @ Doda :-) We had great quality time spent together...Thanks Doda, it was nice having you around. All the memory that we\'ve been through in campus keeps flying in my memory...  Just can\'t wait to&nbsp;see my fiance tomorrow. We\'re supposed to&nbsp;have crazy shopping but not until we drop..hahaha. But the most important is to recap of what happened between us last 2 weeks. So guys, I will be not around until Monday as I\'m planning to take a breakaway this weekend. See you all again. '),(4331,3861317,'27,July,2004','2004-07-27 00:00:00','    Ha ha...seems like my fotopages can link with my blogger. Two thumbs up for the clever mind of mine...hehehe.   Boring Boring Boring Boring Boring. For the past few days, and actually already&nbsp;2 consecutive&nbsp;weeks to be exact, I have no job aka assignment to be done. Still in the office and doing nothing. Arghhh...suddenly I\'m so sleepy, tho\' I went to sleep early last nite...early?? but not so early as I watch \'Pontianak Harum Sundal Malam\'...hehe. Give that 3.5 stars out of 5.   My mom has just call me and said that Bariah (my best friend since secondary school) has delivered a baby girl but&nbsp;something miserable happened as her&nbsp;baby died earlier. Pity her *sob* *sob*. It was already six month and...only God knows. Nothing much to say but trying to comfort her&nbsp;as I can.&nbsp;My tears drop and suddenly I feel so gloomy. Sometimes, my heart can be so fragile at times...maybe its a sign of woman instict. To my best friend, a condolence for you from both of us (read this as me and my fiance). '),(4332,3861317,'26,July,2004','2004-07-26 00:00:00','    &nbsp; Isteri tau suami sayangkan isteri...  walau tak kerap bunga ros dihadiahkan  walau jarang sangat ada candlelight dinner atau makan special yang romantik  walau kurang hadiah intan emas permata  walau tak selalu suami kata \"I&nbsp;love you\" dengan penuh perasaan  walau begitu jarang suami nyanyikan lagu cinta  walau hidup tak semewah mana Isteri tau suami sayangkan isteri....   Dan isteri tau bukan bunga ros jaminan kejujuran  bukan kata-kata romantik jaminan taat setia  bukan intan berlian jaminan kasih sayang dan  bukan juga kemewahan jaminan kebahagiaan...  Tapi isteri hanya perempuan...  Nafsu yang sembilan kengkadang membuat isteri tertanya-tanya  Mana bunga ros merah... mana cincin berlian habib jewel  mana itu dan mana ini dari insan tercinta  Namun walau tiada itu semua... Isteri tetap tau suami sayangkan isteri...  Dan suami tetap no.1 di hati isteri....   Dan suami pula  Kalau tak mampu cincin berlian habib jewel....  Bunga ros merah setengah dozen pun dah mampu menceriakan suasana  Chocolate&nbsp;yang dalam kotak bentuk LOVE dah boleh memujuk isteri...  Kerana sibuk seharian kita terlupa untuk bercinta....  Dengan itu, hangatkan kembali api cinta.... dan kembalilah bercinta....  seperti waktu mula-mula bertemu dulu....    Conclusion :  Things are great even if it\'s only a very little stuff when you enjoy doing it together with your love one...   &nbsp;  p/s : Syukur alhamdulillah...After all the hard times and tough obstacles being tested on me and the other half for a past few days, now I\'m glad it\'s over and everything is moving smoothly. InsyaAllah, we\'ll both go thru all the ups and downs together and do pray for our happiness till the rest of our lives. Abang... I love you... ( ^_^ ) '),(4333,3861317,'22,July,2004','2004-07-22 00:00:00','    Hey out there, have a glance with my new uploaded fotopages in my blog. Hopefully there were no such an error...at last I made it :-)  The picture mostly, ooppsss; to be exact were all captured during Family Day at Reviera Beach Resort -->    urlLink craps by Azni Haida '),(4334,3861317,'22,July,2004','2004-07-22 00:00:00','    Yesterday and until today, my beloved fiance and me had a silent treatment. After so long since the past 3 years we didn\'t get into problem in our relationship, and suddenly yesterday it happened. I\'m still terribly sad at the moment and I bet it\'s going to be longggg to recover. This sadness turns me into a very very mellow little girl. Anyway, I believe the ever-showering blessings from the Almighty will help to reduce the pain.  Yet, I\'m now really missing him sooo much. He starts calling me 10 times yesterday and I\'m purposely didn\'t hang it up. This morning and just now, he calls me again. Sorry my dear, I didn\'t meant to but I need to empty up the head so that I can start smiling again...  Honey, thanks for being my shoulder to cry on especially&nbsp;last night although we are in a silent fight. Thanks for your ears and patient and of course your sweet romantic words that makes my tears drop heavily. I will never forget what you\'ve saying last night ever. But just give me some times to recover it back.  Just to let you know;&nbsp; there will never be other man as sweet as you.. I love you with all my heart forever and ever.    &nbsp;   &nbsp; p/s: Ida sayang Abang '),(4335,3861317,'20,July,2004','2004-07-20 00:00:00','    *Sigh* Hectic weekend during the Family day at Riveara Beach Resort. Now, I\'m in the office doing my routine works again and again...bila la nak siap jurnal aku nie. Sorry, don\'t have any time to update my bloger yesterday and today, curi-curi masa masuk blog ni utk release tension...berpinar dah bijik mata tengok excel jurnal yg beribu-ribu line tu.   So, as I told earlier that I\'m participating the treasure hunt on Family Day, YES!!! my group won the top ten prizes on it. We managed to answer all the questions (although get 6 wrong answers), we get a number 8th prizes among all the 33 participation. Proud haaa (read this as its my first time treasure). And malam tu jugak my group buat persembahan Boria Prodata. Wow!!! ramai gak mata terpegun tengok kitorang menari mlm tu...mesti tak percaya. Apapun, mlm tu group kitorang yg paling best antara 3 group yg lain. tak sia-sia practise setiap petang selama 2 miggu. A big thanks to our lead, Irma. Then, the next day, we have fun playing all the childhood games in Sukaneka Prodata. Lawak tu...tapi memang happenning lah. Lupa semua pasal kerja, SKB, SAP, GL journal etc. Banyak tu ambik gambar but sorry, I\'m still cannot managed to upload the fotopages inside the bloger. InsyaAllah, by the time I\'m free, I will upload it to share the happinest with you all.   And for yesterday, my heart is no more in working mood. Low motivation at all and one thing for sure, I was going back early yesterday. Going to Carefour Wangsa Maju and bought my moisterizer cream and a novel  \"Bisikan Rindu\".  That night had a cold argument with dearie but everything get to normal again this morning as he made a morning call for his appologize. Hemmm...don\'t do that again tau.  &nbsp;  &nbsp; '),(4336,3861317,'14,July,2004','2004-07-14 00:00:00','    Pagi tadi ada ad-hock meeting with En. Ishak bahagian Hasil Sawit dan Getah. Mentioned about problems due to crops, users and management. So, I take this opportunity to review (read as complaining to be exact) all the problems that I had to face in Segamat especially about the users attitude and their work commitment.  Just now I attend a Treasure Hunt briefing at Bilik Gerakan Wisma Felda. It is conducted by \'I Hunt Malaysian Club\' which I give two tumbs up for the great and excitement brief. Macam tak sabar aje nak tunggu hari Sabtu utk pegi treasure hunt. I learn more on the new terms such as tulips, odometer trips, tiebreaker ext.  That\'s it. Time to continue my work. There\'s a lot of G/L (about 3,000+++) lines to be journal. Arrggghhh!!! boring. Tak de short cut ke nak buat jurnal ni...   p/s:  ...more practising for the Family Day presentation this evening...and suddenly, I miss my Abang *sob* '),(4337,3861317,'13,July,2004','2004-07-13 00:00:00','    Di mana pun ku berdiri Kau kan tetap ku ingati Kau selalu dekat dalam hati ini  Dan tak pernah sehari pun Hari-hari yang berlalu Tanpa aku merindu mu kekasihku  Saat kita berjauhan Jangan ada rasa bimbang Kerana cinta menyatukan perasaan  Tak mengira detik waktu Ingin ku lafaz selalu Aku cinta kepada mu  Walau seribu liku di depanku Akan tetap aku tempuh Dalam hatiku kau tak pernah jauh Walaupun terpaksa ku menunggu Tanpa jemu dan selalu Sentiasa dekat di hatiku  Kau membuat hidup ini Semakin lebih bererti Kau miliki kelebihan tersendiri  Tak mengira detik waktu  Ingin ku lafaz selalu Aku cinta kepada mu    p/s: For the one I love...whatever else may complicate our lives, one thing is pure and simple, I Love You heart and soul. '),(4338,3861317,'12,July,2004','2004-07-12 00:00:00','    Today maybe a bad day for me as I am so  frustrated  with one of my user and always with my  job commitment  (to be exact). God please help....  I need more motivations somehow to go thru all the hassles in the office. I think, my work will be never ending, habis satu, ada lagi yang datang, pressure jugak woo... but then, pikir2 balik, they are just work... but i still have to do it with commitment.  Rasa nak nak marah je dengan user yg sorang ni!!! SABAR...   p/s: AEV 319 is now mine... '),(4339,3861317,'11,July,2004','2004-07-11 00:00:00','    Gosh...pagi tadi kereta kena sita. Ni semua keje pak guard la ni. Alah, parking makan line orang sikit pun kena sita ke..?? Teruk betul management!!  Terpaksa pulak jumpa bahagian security utk settlekan problem ni. Mujur akak kerani tu baik. Ni dok tunggu akak tu call, mintak mintak la pak guard bukak gari kereta tu. Kesian AEV 319, kena gari macam orang bersalah aje. Luv u my sweet car... '),(4340,3861317,'11,July,2004','2004-07-11 00:00:00','    Buat jahat sikit hari Ahad tu, balik awal dan tinggalkan kerja sbb nak pi dating ngan abang. So, at 12.20 tu lari naik LRT sampai Sungai Besi dan tunggu abang kat sana. Malas nak bawak kete, hujan satu hal..jem lagi teruk. Tapi, sampai aje kat stesen Sungai Besi, terpaksa tunggu abang dalam setengah jam, hujan pulak tu. Macam nak marah aje, tapi sabar.  So, dalam pukul 1 dia sampai, terus pergi Mines. MasyaAllah, punyalah banyak kereta sampai jem tempat nak parking tu. Menguji kesabaran betul. Nasib abang yg bawak kete. At last dapat jugak parking.   Masuk je Mines, punyalah ramai manusia. Ni semua gara-gara Spiderman2 punya panangan lah ni. Nak beratur beli tiket pun satu hal..queue panjang!!! Abang dah try booking tiket tapi kat Mines tak boleh book online. Nasib baik dapat tiket, itupun dapat seat 4 dari depan...sakit leher pun takpe la, janji dapat tengok..hehehe.  So, sementara tunggu nak tengok wayang perut pun dah berkeroncong. Maklumlah hujan, cepat lapar. Nak makan pun kena beratur panjang. So, at last makan KFC aje. Chick a cheese tak la sedap sangat, biasa aje..tipah tertipu!!  Lepas makan, gi jalan-jalan. Ingat nak cari tudung dengan seluar slack hitam utk persembahan Family Day nanti tapi tudung punyala mahal kat Mines. So, slack je la yang dapat.  Wayang habis dalam pukul 9.30 mlm. Terus ajak abang balik sebab dah penat sangat mata ni pakai lens. On the way balik Selayang pun jam jugak!!! This is what we called KL, JAMMED! Sempat singgah restoren utk supper sebab perut dah mula berbunyi minta diisi. Kesian abang, balik Damansara lewat pagi, pas tu kena bangun awal utk short course pulak. Sori, sayang abang :)  Anyway, terima kasih abang untuk segalanya. I love u so much. Segala kesyukuran untuk-Mu kerana menemukan ku dengan seseorang yang baik budi pekerti, romantik dan yg paling penting memahami dan penyabar.  p/s: Public apologize for my fiance for doing something that makes you unhappy and disappointed. Sorry my love, I did\'nt mean to. Till the death do us part '),(4341,3861317,'10,July,2004','2004-07-10 00:00:00','    My blogger just screwed up my head. I only managed to write my jurnal today since the blogger updates something in the page.   Here I am...in the office doing my work on SUNDAY. Betul-betul giler kuasa. Tak lah, ni semua sbb nak habiskan keje coz the due date is next week. So, forget about all the thingy work plus people surrounding me (in this office of course)... Gosh!   Last nite, my fiance, Syed and Azi pick me up at Selayang and we were going for a great dinner at Selayang Baru Seafood. Nice place and the food quite nice. Nothing much to say for both of us since I had faced a hectic work on Saturday and my fiance tired driving from Kuantan to KL.  For Syed and Azi..heemmm...bila lagi, apa lagi nak tunggu ?  p/s : Nice to see Syed and Azi again since the last time we met was on my engagement day. '),(4342,3861317,'10,July,2004','2004-07-10 00:00:00','    Yeah...yeah!! My fiance is coming to see me in KL. Actually he had a short course training in KL at IBM and Mesiniaga for a week. Yeeahaaa!! Boleh jumpa :-D  Tonite I will meet him and going for a dinner. My plan to mengular ngan member (read as Liza and Lin) was cancelled, not because of my date tonite but Lin have to stay back at her office in TNB Bangsar and Liza\'s mom is not feeling well.   So, better going back home now and take a shower. Tak sabar nak kuar ngan abang malam ni...hehehe '),(4343,3861317,'08,July,2004','2004-07-08 00:00:00','    Working on Saturday makes me feel bored and not too much if I said that it\'s not productive. Nothing much work to do but the crop problem makes me headache. Benci, muak :(  Meeting with En. Ishak was cancelled...haaa..lega, but will be posponed next Wednesday.  This evening will have \'boria\' practising since my company Family Day will be held next week at Riveara Beach Resort, Melaka. Much more steps to discover and learn. Waaaa..penat, dah la keje sampai petang. '),(4344,3861317,'08,July,2004','2004-07-08 00:00:00','    First of all, I have to listen all the Kak Ma\'s annoying about the user\'s attitude. Habis semua hasil sawit dan getah problem, tu yg Kak Ma bengang tu. User pun satu, bergantung sangat pada OSS (On-side System Support). Ni, baru je selesai settlekan problem Air Hitam. Muak dah...thanx to Syahman for helping me.  Then on Saturday, all the OSS including me, have to see En. Ishak regarding the crop problems. Gosh! Segamat banyak problem tuhhh...better get ready all the SLP summary report before the day come (and of course to pretend myself, hehehe).  Bila-bila masa je dari sekarang dah kena buat job baru. That  \'girl\'  tengah dok compile file nak assign job tu. Oh ooo..she\'s coming, now I have a job to do. Better pen off now and do my work. '),(4345,3861317,'07,July,2004','2004-07-07 00:00:00','    Planning to mengular pi jusco tak jadi...hujan. Malas nak drive dlm hujan lebat tu. So,  urlLink time  to get back  urlLink home  now. Errrkhhh, I\'m starving, since don\'t take any meal for lunch. Can\'t stop thinking of keoteow tomyam pedas..yummiiee. Better get  urlLink home  now. '),(4346,3861317,'07,July,2004','2004-07-07 00:00:00','    Tahukan anda biodata IBLIS :  Nama: Iblis laknatullah  Tempat lahir: bermula dari keenganannya sujud kepada Nabi Adam a.s  Status: Fasik kelas pertama  Agama: Kufur  Tempat tinggal di dunia: Di dalam hati mereka yang lalai  Tempat tinggal di akhirat: Neraka jahanam   Kawasan legarnya: setiap tempat yang di dalamnya tiada usaha untuk mengingati Allah.   Arah perjalanannya: tiada haluan dan bengkang bengkok  Isteri di dunia: setiap wanita yang tidak menutup aurat dan tidak mentaati Allah, Rasul dan suaminya.  Kaum kerabat: semua golongan yang menentang dan memusuhi Allah.  Modal pusingan: angan-angan kosong  Tempat pertemuan: Pasar dan pusat membeli belah.   Musuh ketat: Orang islam yang beriman dan bertaqwa  Kegemaran: melalaikan orang dari mengingati Allah, menggoda dan menyesatkan.   Matlamat: supaya semua makhluk di muka bumi ini menjadi kafir dan sama-sama masuk Ke neraka jahanam.  Kelemahan: Bila orang membaca istighfar  Lambang: sebarang tatu  Pejabat operasi: tandas, bilik air, tempat bernajis serta tempat maksiat.  Rakan kongsi: golongan munafik serta syaitan dari kalangan jin dan manusia  Sumber: segala harta yang diperolehi dengan jalan haram dan riba  Makanan kegemaran: daging orang mati, segala najis dan kotoran.  Yang ditakuti : orang mukmin yang beriman dan bertaqwa kepada Allah.  Tempoh perkhidmatan : sehingga hari kiamat   p/s : dgn itu sesame la kite menjadi org yg bertakwa dan berdoa kpd Allah SWT moga dijauhkan dari iblis laknatullah....insyaAllah... '),(4347,3861317,'07,July,2004','2004-07-07 00:00:00','    a) Jika kita mencintai sesorang, kita akan sentiasa  mendoakannya walaupun dia tidak berada disisi kita  Jangan sesekali mengucapkan selamat tinggal jika kamu  masih mahu mencuba, jangan sesekali menyerah jika kamu  masih merasa sanggup, jangan sesekali mengatakan kamu  tidak mencintainya lagi jika kamu masih tidak dapat  melupakannya.   b) Jangan simpan kata-kata cinta pada orang yang  tersayang sehingga dia meninggal dunia , lantaran  akhirnya kamu terpaksa catatkan kata-kata cinta itu  pada pusaranya. Sebaliknya ucapkan kata-kata cinta  yang tersimpan dibenakmu itu sekarang selagi ada  hayatnya.   c)Cinta ia tetap hadir, ia datang bukan dengan nafsu  atau keinginan tetapi kepasrahan tentang yang mutlak,  itulah cinta terulung.  d) Mungkin Tuhan menginginkan kita bertemu dan bercinta  dengan orang yang salah sebelum bertemu dengan orang  yang tepat, kita harus mengerti bagaimana berterima  kasih atas kurniaan itu.   e) Sungguh menyakitkan mencintai seseorang yang tidak  mencintaimu, tetapi lebih menyakitkan adalah mencintai  seseorang dan tidak pernah memiliki keberanian untuk  menyatakan cintamu kepadanya.   f) Cinta adalah keabadian...dan kenangan adalah hal  terindah yg pernah dialami.   g) Kita tidak pernah tahu bila kita akan jatuh cinta.  Namun apabila sampai saatnya itu, raihlah dengan kedua  tanganmu dan jangan biarkan dia pergi dengan sejuta  rasa tanda tanya dihatinya   h) Cinta bukanlah dari kata-kata tetapi dari segumpal  keinginan diberi pada hati yang memerlukan. Tangisan  juga bukanlah pengubat cinta kerana ia tidak mengerti  perjalananan hati naluri.   i) Cinta seringkali akan lari bila kita mencari, tetapi  cintajua seringkali dibiarkan pergi bila ia  menghampiri.   j) Cinta pertama adalah kenangan. Cinta kedua menjadi  pengajaran. Dan cinta yang seterusnya adalah satu  keperluan kerana hidup tanpa cinta bagaikan masakan  tanpa garam. Lantaran itu bajailah cinta yang  dianugerahkan itu sebaik-baiknya. Agar ia terus mekar  dan wangi sepanjang musim.   k) Cinta tidak selalu bersama jodoh tapi jodoh selalu  bersama cinta   p/s: Nice  urlLink quotes ...suddenly remind me to my beloved fiance :) '),(4348,3861317,'07,July,2004','2004-07-07 00:00:00','    Not much productive works for today. All the  urlLink bank  and cash reconciliation works have been sucessful done. Tomorrow my life will ruine up again since a new  urlLink job  will be assign. '),(4349,3861317,'16,August,2004','2004-08-16 00:00:00','    Hello, I\'m in Prodata HQ again doing my work, but this time the productive job (hehehe...at last). I was assign to do a training module as we had a brainstorming for the last 2 days preparing the end-user SLP training module. And for today, I had finish my work....la la la la....  I feel like jumping when my sister has just call me and said that she get a job as Trainee Engineer at ENC company in PJ. I guess that this would be the grand present for Abah as he will celebrate his birthday tomorrow. Happy Birthday Abah, semoga dipanjangkan umur, dimurahkan rezeki dan diredhai Tuhan dunia Akhirat. Ida sayang Abah :-) '),(4350,3861317,'05,August,2004','2004-08-05 00:00:00','    Battery Low Demotivated as usual. Pending job of uploading template.  Haaacuuum..!! I have flu + not feeling well inside my body + my skin starts dry. I\'m freezing like death in this office as the air-conditioning starts operating well after a monthly services.  Mental Block Huuaarrgg...sleepy. Do not know what to do. I\'m surfing the net and chats on MSN Messenger like nobody\'s business.. Heh.. Who cares by the way..? I have done all my job excluded the pending one. And the one that I loved to do is my blogger...hehehe  Mood Swing ?? My body is in the office but my soul flies through my hometown. I  miss mak and abah. Can\'t wait to act like a sleeping beauty girl on the bed with my teddy and along (read this as my cute cat). Pray for a safety journey tomorrow. I think, nop...exactly I\'m putting my weight already and this makes me uncomfortable. I feel like soo lazy, heavy, sleepy and to be worst my mouth keeps chewing any food on the table especially the junk one. Huh..what am I supposed to do.  Non productive works = eat + eat + eat  (which is bad habit at all)  Non Next week would be my last day in Prodata HQ as I have to go back to Segamat and do my routine work as usual :-( Probably I will never have a time to update my blogger since my office there don\'t have any connection/online excluded the SAP line. But I\'ll try my best to write at least once a month as I will miss my weblog soon...hihihi   Hey, my  \'daddy\'  syahman has just join my blogger...hehehe, statue your name into my journal. Proud haa ^_^  Another craps from me quoted from Bluehikari :  \"Buang kusut benang putus asa, kaitkan ia menjadi sabar dan jadikanlah ia alas ketabahan. Sempurna semuanya, tawakkal kepada Allah biar ia kekal dalam diri menjadi hiasan...\"  Nice...very nice. Isn\'t it right ?    p/s : All the best for my colleague @ housemate - L**D* as she will tie up into another new phase of her life which called a  \'merisik\'  during this weekend and getting engaged probably end of this month. A big congratz also to my friends H**A and A**A for their married. Happy Ever After. '),(4351,3861317,'04,August,2004','2004-08-04 00:00:00','    Jika kamu memancing ikan Setelah ikan itu terlekat di mata kail hendaklah kamu mengambilnya terus ikan itu Jangan sesekali kamu lepaskan ia semula ke dalam air begitu saja Kerana ia akan sakit oleh kerana bisanya ketajaman mata kail mu Dan mungkin akan menderita selagi ia masih hidup...  Begitula juga setelah kamu memberi banyak pengharapan kepada seseorang Setelah ia mula menyayangi mu, hendaklah kamu menjaga hatinya Jangan sesekali kamu meninggalkannya begitu sahaja Kerana dia akan terluka oleh kenangan bersamamu dan Mungkin tidak dapat melupakan segalanya selagi dia mengingatimu.  Jika kamu menadah air biarlah berpada Jangan terlalu mengharap pada takungannya dan Jangan lah ia begitu teguh...cukup la sekadar keperluanmu Apabila sekali ia retak, tentu sukar untuk kamu menampalnya semula Akhirnya ia dibuang...sedangkan jika kamu membaikinya Mungkin ia masih boleh digunakan lagi  Begitu juga jika kamu memiliki seseorang...terimalah seadanya Jangan lah kamu terlalu menganguminya dan Janganlah kamu menganggapnya begitu istimewa Anggaplah dia manusia biasa Apabila sekali dia melakukan kesilapan, bukan mudah bagi kamu untuk menerimanya Akhirnya kamu kecewa dan meninggalkannya Sedangkan jika kamu memaafkannya boleh jadi hubungan kamu akan berterusan hingga akhirnya...  Jika kamu telah memiliki sepinggan nasi Yang kamu pasti baik untuk dirimu...mengenyangkan...berkhasiat Mengapa kamu berlengah, cuba mencari makanan yang lain Terlalu ingin mengejar kelazatan Kelak nasi itu akan basi dan kamu tidak boleh memakannya Kamu akan menyesal.  Begitu juga jika kamu telah bertemu dengan seseorang insan Yang kamu pasti membawa kebaikan kepada kamu Menyayangimu...Mengasihimu Mengapa kamu berlengah Cuba membandingkannya dengan yang lain Terlalu mengejar lesempurnaan kelak kamu kehilangannya dan kamu akan menyesal Apabila dia beralih arah...    p/s : Satu nasihat buat teman ku L**A :  If you loves someone, set him free. If he comes back, he is yours and if he doesn\'t, it was never meant to be... '),(4352,3861317,'04,August,2004','2004-08-04 00:00:00','    Another craps also from \'hanyasayang\' weblog as I have no idea what to do in this office hour....boring. I have a template to upload into the system but have to pending it since the commitment type of GL does not have enough budget. Then, I have to wait until the budget approved by Financial  team.  Back to the test, so...I\'m group of  AUTUMN  season. Hemmm....like a environment in Autum In My Heart :-)  The Season You Are Most Like Is : + Friendly + Confused + Shy + Warm + Approachable + Caring  It described :  You\'re warm, and the most approachable. You have that gentle prescence about you. People can relate to you, and find you easy company. However, it\'s likely you\'ve been hurt in the past and it has left you scarred so things can become rather chilly with you at times. Being the third season in, you\'re mature, thrustworthy and loyal to your friends but prone to depression and negative thinking. Well done...You\'re the shy and sensitive season :-) '),(4353,3861317,'04,August,2004','2004-08-04 00:00:00','    I had done a personality or enneagram test which I craps from \'hanyasayang\' weblog. It is call  urlLink Riso-Hudson Enneagram Type Indicator (RHETI)  Sample.  The RHETI can help you determine your Enneagram type which you have to answer 38 of questions in this saample test. While there is no guarantee, your highest score can indicate your basic type.  So, here it is...a conclusion of my enneagram.   Type 1 : The Reformer. The rational, idealistic type.                               (0)                           Tyoe 2 : The Helper. The caring, naturing type.                                       (6)   Type 3 : The Motivator. The adaptable, success oriented type.             (3)  Type 4 : The Artist. The intuitive, reserved type.                                    (5)   Type 5 : The thinker. The perceptive, cerebral type.                              (2)   Type 6 : The Skeptic. The committed, security-oriented type.              (4)   Type 7 : The Generalist. The enthusiastic, productive type.                   (6)   Type 8 : The Leader. The powerful, aggresive type.                                (4)   Type 9 : The peacemaker. The easygoing, accomodating type.              (6)    Hemm...among my top score, I conclude that I am type of  Helper ,  Generalist ,  Peacemaker  and  Artist  person. Is that me..?? Yaaa, I think I am but the most important thing is what people see in you right :-)  So, have a try with this test. '),(4354,3861317,'04,August,2004','2004-08-04 00:00:00','    Now I can imagine what it\'s like when they said about pimple and it\'s unpleasant ache. Why so...because at this moment i got pimple in one of my chicks. Gosh, I hate this dull ache..this is one of those very rare situation for me. Please just go away...   Sorry, another  urlLink craps  from me...'),(4355,3861317,'03,August,2004','2004-08-03 00:00:00','    Hai, here I am again after being hibernating for 3 days. Still in the  urlLink office  and doing nothing much productive works, as usual...boring and demotivating. I feel like a dummy sitting on the chair all day long and off course it\'s not something good at all.  urlLink Time  flies, huh, it\'s been almost 3 months staying here in HQ KL.     Being with beloved fiance the whole weekend makes me cheer up again. Thanks dear for being such a wonderful man. And again I had a good  urlLink time  for last couple of days. And I want you to know that I really appreciate what you are and have been doing so far. Only I know.... :-)    Recap for the past few days, we were  urlLink shopping  at Mid Valley and OU. Plan to go to IKEA but don\'t have enough  urlLink time , maybe some other  urlLink time  ok dear :-) I bought a simple but nice color of blouse, two-way clear smooth maybelline as I got 20% voucher of it (worth it huh) and my other half a pair of renome slack and of course his PS2 vcd  urlLink games . Also catch a  urlLink movie  of \"A Day After Tomorrow\" and \"I-Robot\". Due to the tiredness of both of us during the day  urlLink time , we managed to take a nap in the middle of the peak scene. Hehe.. ;-) Honestly, I had a marvelous nap next to my darling....but as far as the  urlLink movie  is concern, it was pretty good I consider.    Thanks again dear, thank you so much for the wonderful moments together... Well, every moments with you is the moment I treasure most in my life... starts mising u badly.'),(4356,3861317,'03,August,2004','2004-08-03 00:00:00','    Mampukah aku menjadi seperti Siti Khatijah? Agung cintanya pada Allah dan Rasulullah. Hartanya diperjuangkan ke jalan fisabilillah. Penawar hati kekasih Allah. Susah dan senang rela bersama...    Dapatkah kudidik jiwa seperti Siti Aishah? Isteri Rasulullah yang bijak. Pendorong kesusahan dan penderitaan. Tiada sukar untuk dilaksanakan...    Mengalir air mataku, melihat pegorbanan puteri solehah Siti Fatimah. Akur dalam setiap perintah Taat dengan abuyanya, yang sentiasa berjuang. Tiada memiliki harta dunia. Layaklah dia sebagai wanita penghulu syurga...   Ketika aku marah, inginku intip serpihan sabar. Dari catatan hidup Siti Sarah....Tabah jiwaku Setabah umi Nabi Ismail. Mengendong bayinya yang masih merah. Mencari air penghilang dahaga.    Di terik padang pasir merak. Ditinggalkan suami akur tanpa bantah. Pengharapannya hanya pada Allah Itulah wanita Siti Hajar....    Mampukah aku menjadi wanita solehah? Mati dalam keunggulan iman. Bersinar indah, harum tersebar. Bagai wanginya pusara Masyitah....            p/s : buat pedoman kaum Hawa dan peringatan diri.'),(4357,3487773,'29,May,2004','2004-05-29 00:00:00','    HMm...just testing this out...gonna have to figure how to link this to my page...erm....   '),(4358,3487773,'19,June,2004','2004-06-19 00:00:00','    Hey BVC ppl...  Have u guys thought about ordering ur gowns for Call? I cant seem to find my list of things i need to order. Just want to see how much it will cost. I tried to go to www.gownhire.co.uk, but after clicking Gray\'s Inn, the page is invalid! So how in the world are we gonna hire our stuff?  Oh...and how do i change my fonts? I dont like this one... hmm...will go find out... BRB..   '),(4359,3487773,'17,June,2004','2004-06-17 00:00:00','    We all can breathe a sigh of relief by tomorrow. Final papers are due and its over!  Its been a very stressful, strenuous, taxing and challenging year for everyone on the BVC. This has been the most challenging course Iâ€™ve undertaken and Iâ€™m quite happy its finally over.  However, having said that, it was also one of the most enjoyable course Iâ€™ve taken. (aside from French)  Iâ€™ve been pushed to do things and not slack throughout the year. Its true what the lecturers told us in the beginning, that this is going to take over our lives for the next 9 months.  Iâ€™ve had little time for anything else; friends, social life, etc. I found myself coming home from classes and studying or starting on the next subjectâ€™s homework. Iâ€™ve literally turned into a Nerd! And you know what? Its not that difficult to put a little more effort into what we have to do. After all, what is 9 months compared to your whole life?  I must admit, its never been smooth sailing the whole way. I may look ok on the outside, but inside, I am a wreck! Iâ€™ve had a few nervous breakdowns this year alone!  Maybe it was tougher this year coz Iâ€™ve always been â€˜spoon fedâ€™ all my life and now I have to do it all on my own. Donâ€™t ask me how I got my law degree, coz I have no idea what I was doingâ€¦ Ok, this is not gonna be too good on my CV!  Some people say that I must be smart to come all this way in something I donâ€™t like. But I think its â€˜guiltâ€™ that really drives me. I feel I owe my parents something for spending so much money on my education and I should be grateful, for not everyone is as lucky as I am. They want me to be successful and live comfortably in the future and I donâ€™t see anything wrong with that.  Studies aside, what I really enjoyed this year is meeting friends that Iâ€™ve made throughout the year. I feel very lucky to have these people as my friends, just as I thought Iâ€™d never find good friends that easily. I know this is very clichÃ©, but I really thank God Iâ€™ve found them. And I am so grateful to those who have been by my side when I needed them most.  Its not been an easy year for me, emotionally. But I always could talk to someone about my problems and whatever happens, I know theyâ€™ll always be there for me and never let me feel alone.  I am really going to miss them all when I leave.  Some of my friends have left already and although I donâ€™t feel it yet, but its gonna hit me in a few weeks. But Iâ€™ve always been optimistic about life, and I know somehow, someday Iâ€™ll see them again.  We might not grow old together, but Iâ€™ll always know thereâ€™s someone out there I can count on, and they on me.  I am really looking forward to going home to Penang. I miss the lifestyle and food, But it will be very difficult to fit in again. Iâ€™ve lost contact with a lot of my friends back home, but I know that I will be fine.  I never really believed people who tell me that University life is the best. But now, I CONCUR!  Iâ€™m just feeling all melancholie now. I want to go back, but I donâ€™t want to leave my friends. I realise that I donâ€™t have to have 1000 friends, just a handful that really matter to me.  I just want to give a shoutout to my dear friends that Iâ€™ve met here and will miss a lot when we all part.  Madelene Tan(Iâ€™ll c u back home, right?) Jenny Liau Irwin Lo Lee Shih (seeing u guys in Aug-Sept!) Serene Tang Thomas Teh Tai Yong (gonna hire me as ur PA when uâ€™re rich and famous, rite?) Charmaine Goh Alexandra Houssaye Alexandre Martellozzo Helene Guibert Tia Po Yew Helene (Fishponds) Miriam  Raphael Cadoux (my hsemate) Derek Fox (my hsemate) Louise Howard Karina James  I love you guys, and hopefully weâ€™ll keep in touch! Ciao!   '),(4360,3487773,'16,June,2004','2004-06-16 00:00:00','    Have you LOVED before?  If to love means to be selfless, then I have not loved before, For I have always been selfish.  And why shouldnâ€™t I?  Maybe I have never loved anyone then. Maybe I want to be loved and when this is not given to me, I feel the need to somehow extract love from the people I think I love. But I realise that I have not loved, but need to be loved.  Affection. I have given so much and still have so much to give. One never exhausts affection. I find it so hard to withhold affection to the ones I care for.  Maybe thatâ€™s why I love to shower my loved ones with gifts. Not just the various partners I have had, but even my close friends. Whenever I see something that I think they might like, I feel like getting it for them. No matter the cost, for I am satisfied if they are happy.  But is it really necessary to give gifts?  Maybe Iâ€™m too materialistic. Tell me that to my face. I have no qualms. For I know it is true. Material things do make me happy. But Love?  My heart has been broken too many times to know what LOVE is anymore.  WHAT IS LOVE? Is it an emotion of giving but not expecting anything in return? Is it a warm feeling for another person? Is it a feeling of security? Or is it just plain physical?  Relationships puzzle me. Why would a guy want to spend time with a girl? And for that matter, what does a girl want from a guy? Other than to procreate, I donâ€™t really understand what couples really need from each other. Attention, care, affection, emotional support? Any human being is capable of providing that. Maybe more females than males. Maybe thatâ€™s why more females turn to their own kind.  Iâ€™m not saying that I am homosexual. I just donâ€™t really understand a heterosexual relationship anymore.  Attraction And why is it that Iâ€™m always attracted by guys who break my heart? Maybe its because I pick the â€˜wrongâ€™ kind of guys. Iâ€™ve always been attracted to the rebellious ones, the bad boy look, etc. I never seem to be attracted to the nice â€˜rightâ€™ ones.  The ones that would listen to me, talk to me and hold me like I want to be held.  Maybe they do. Maybe I just expect too much in a relationship. Does that mean Iâ€™ll never be happy? Should I lower my expectations? Will I be happy then? I doubt it. I need too much. Would anyone ever understand?  Am I overly sensitive? I think I feel emotions more than a lot of people. I used to be aloft and blissfully unaware of other peopleâ€™s feelings when I was younger. But now, everything I do seems to reflect how I think people will feel. Iâ€™m not impulsive anymore and take an awful lot of care before I open my mouth.  Have you ever felt, when watching a movie, that you wanna shake a character to just say what they have inside? For example, a girl/guy who wants to confess their love/emotions for another and is holding back for various reasons? Iâ€™ve always wanted to just shout â€œJust say it!â€ But now, I find myself holding back. Biting my tongue, suffering in silence.  A friend once said â€œHe who sighs never gets what he wantsâ€. It could be true. It so frustrating!  You want to let them know how you feel, but you think about the consequences. Is it worth it? Whatâ€™s gonna change? And then thereâ€™s regret.  I used to say â€œI donâ€™t have any regrets yet!â€ I used to think that everything Iâ€™ve done made me who I am today. But who am I today, really? Do I have regrets? Definitely!  Of course weâ€™ve all made mistakes in life. What would life be without mistakes? What would we learn without them? But it doesnâ€™t prevent us from regretting, being sad, being hurt.  Câ€™est La Vie!  I just want someone to love me as much as I love him. Is that too much to ask?   '),(4361,3487773,'11,June,2004','2004-06-11 00:00:00','    Last night we (BVC STUDENTS) had a BVC dinner at the Royal West of England Academy and it was pretty boring. Maybe coz it was too bright in the dining hall, or maybe it was coz we had to sit and talk to people we dont normally have conversations with. I was also having a headache, so that didnt help...  Was separated from my friends and so was pretty lonely...  Took LOTS and LOTS of pics with all my dear friends, and cant wait to develop them...i have another roll, with the Birthday BBQ pics to develop as well! oh man, looks like i\'m gonna be eating grass soon!  After the dinner, we all headed to Fuschia for a party and was admitted into the VIP room, which we didnt even spend much time in anyway! But thanks to T(hiru) who organised it, we had fun! at least i did! I was given a Tequila shot to down and i DID! but 1/2 hr later i was totally out of it...dunno if i was drunk or what, but i had a massive headache and it felt so heavy...and my vision was turning black! maybe i was gonna pass out and had to sit down every few steps! but once i got outside, i felt loads better...got home safely, thanks to Jenny, Maddy and Ian and plonked into bed, after taking everything off!  Woke up at about 12.30pm, so no going to class for me! hehe it started at 11am! but had to come in to pick up my International Trade brief anyway... should have brought my IT notes, so i could do my work here..coz i dont have internet so i always have to come into Uni to do my online research. sheesh! what a hassle.  well, better stop talking and start cracking on.  I cant wait till the 18th! then i\'ll be FREE!!!   '),(4362,3487773,'08,June,2004','2004-06-08 00:00:00','    I\'m bored...at Uni..online...think i should start cracking on my Employment brief...but its so hot! i keep sweating! sheesh...should go sit in a tub of cold water...  My earrings!!! thanks to Alex and Leesh, i now have a new earring design...what a coincidence...\'LUCKY\' Me... ... and This ALEX is so LAZY! sheesh...   '),(4363,3487773,'07,June,2004','2004-06-07 00:00:00','    Hey, i have 2 pic in the last crunchie party of the year...its in the uwe website...there\'s a link...i think its on page 15...  been so tired...and had negotiation today...was against Chris Jones...that must send some shivers down a few spines! anyway, it was his second and i was so dazed...but we managed to reach some kind of compromise...and the lecturer said i did give some good points that others didnt pick up on...YaHoOOoO..yea...thats what u get when u work in a half comatose state haveing slept only 1 hr!  ok, gtg now....pretty blur and going cross-eyed! ciao!   '),(4364,3487773,'19,July,2004','2004-07-19 00:00:00','    As u probably know, since not many ppl read my blog anyway... i\'ll be leaving Bristol, UK soon. Heading for Paris, then my lovely Penang in August.  I\'ll also be called to the BAR at Gray\'s Inn on the 22nd July.  Shipped like 4 boxes home to Penang and had to wait like forever for that fella to come pick it up..from 7pm-1am!! and he finally came at 10am the next day!! stupid shit!!  Ok, so now my room looks like it was hit by a tornardo..and being me, i.e. lazy as hell, i am not bothered to clean it up.. and no one can make me...HAH!  anyway...i booked like my barrister\'s gown and bands and still dont know what shirt i am suppose to wear with the bands...  booked my coach to london, hotel and hopefully everything goes smoothly.  If i knew i had to ship my things by last friday, then i would have gotten the eurostar back to Paris with Julz from London on 25th!! sheesh...talk about bad planning...but how was i to know that the shippers were so unflexible? can only pick up on friday and no other day for Bristol!! sheesh...what kind of business is this? who\'s the client here?  so now i have to come back to bristol on 25th and leave for Paris via London on the 28th!  but luckily the flight is not that expensive... only Â£27.99 with easyjet. eurostar was Â£50!  But its good that i can come back to have a farewell dinner with my friends before i leave UK, for good (maybe).  will certainly miss my friends, but never the weather!! HAH! i cannot understand how ppl can still live here! and dont get me started on the lifestyle! *rolling eyes!*  ok lah, better go now, coz i am very very hungry and shih\'s blog about malaysian food does not really help hungry ppl in UK!! :P  ciao!   '),(4365,3170057,'30,April,2004','2004-04-30 00:00:00','   urlLink The New York Times > Technology > Circuits > In Class, the Audience Weighs In   cool...wish I had this one in back in college... Argh! Miss school, miss sociology   '),(4366,3170057,'29,April,2004','2004-04-29 00:00:00','   shame. I cant stop smiling.  gawd. He made my heart leap, and I can\'t stop looking at him. geez. He kissed me, and it felt so sweet. I can still feel his lips on the lower portion of my left eye.  yeah. that is where his first kiss landed. uhuh. but he just skyrockted in my heart.   '),(4367,3170057,'29,April,2004','2004-04-29 00:00:00','   Ah, John. This would be my vow to you...  â€œThere is no room in my body for anything but you. My arms love you; my knees shake with blind affection. My mind begs you to ask it something so it can obey.â€   Thanks Buttercup, or should I say William Goldman =)   '),(4368,3170057,'28,April,2004','2004-04-28 00:00:00','   urlLink Yahoo! News - Fresh Hope for Fat Bunnies :  \"Rabbits are being killed with kindness\" --- wish I can say this about humans too. justa thought...   '),(4369,3170057,'28,April,2004','2004-04-28 00:00:00','   urlLink Yahoo! News - Most Americans Select Money Over Sex  Money over sex? You got to be kidding me!!!! Hahahahaha   '),(4370,3170057,'28,April,2004','2004-04-28 00:00:00','   I am scared, and at the same time excited about the things going on in my life. I really donâ€™t know which emotion to hang on to. Especially with regards to this great guy.  He is such a wonderful person. A very pleasant surprise.  But I have to play this one coolly. Think things through. But gawd! He is so much wonderful. And our conversations keep playing in my head like a sweet lullaby.  Oh dear johnâ€¦   '),(4371,3170057,'27,April,2004','2004-04-27 00:00:00','   \"When you feel it in your body you found somebody who makes you change your ways like hanging with your crew said you act like you ready but you dont really know and everthing in your past- you wanna let go\"  damn! I have it so bad for this guy.  I\'m dead!   '),(4372,3170057,'27,April,2004','2004-04-27 00:00:00','   hmmm...  the silence is deafening... i can almost hear my mind processing... scary... total silence can be sickening... i want some music...i crave for some noise... yet, all that is playing is silence... and this is making me sad...making me very sad... is this a preview of my next 20 years? a life of total silence... deprived of sound?   '),(4373,3170057,'26,April,2004','2004-04-26 00:00:00','   Damn! Damn! Damn! Why am I in limbo again?  Damn! Damn! Damn! All I can think of is usâ€¦together! Damn! Damn! Damn! But I shouldnâ€™t be doing this, I shouldnâ€™t be thinking about us! Damn! Damn! Damn! Stop beating you foolish heart!  Damn! Damn! Damn! I am sick, and the only cure for this tattered soul of mine is him! Damn! Damn! Damn!   '),(4374,3170057,'26,April,2004','2004-04-26 00:00:00','   urlLink Yahoo! News - Violence Among Girls Increasing in U.S.   This makes an interesting sociological study...hmmm. I miss schooling!!!   '),(4375,3170057,'26,April,2004','2004-04-26 00:00:00','   Damn! I miss him. Damn! \"Reality bites!..and I have the teeth marks to prove it!\"  Officially Missing You   All I hear is raindrops falling on the rooftop Oh baby, tell me whyâ€™d you have to go â€˜Cause this pain I feel it wonâ€™t go away And today Iâ€™m officially missinâ€™ you  I though that from this heartache, I could escape But Iâ€™ve heard it long enough to know There ainâ€™t no way And today Iâ€™m officially missing you  Chorus: Ooh...canâ€™t nobody do it like you Said every little thing you do, hey, baby Said it stays on my mind And Iâ€™m officially  All I do is lay around, two years full of tears From looking at your face all over Just a week ago you were my baby Now I donâ€™t even know you at all,  I donâ€™t know you at all  Well, I wish that you would call me right now So that I could get through to you somehow But I guess itâ€™s safe to say, baby Safe to say that Iâ€™m officially missinâ€™ you  Chorus  Well, I thought I could just get over you, baby But I see thereâ€™s something I just canâ€™t do From the way you would hold me To the sweet things you told me I just canâ€™t find a way to let no one through  Chorus  Itâ€™s official Hoo, you know that Iâ€™m missinâ€™ you, yeah, yes All I hear is raindrops, oh, yeah And Iâ€™m officially missinâ€™ you   '),(4376,3170057,'26,April,2004','2004-04-26 00:00:00','   argh! im sooooo sleepy..my eyes are ready to jump out of their sockets to my tabletop, and then free fall to the floor in order to roll far away from my scurrying hands and hide in the darkest corner of my workstation so they can slumber soundly. *sigh* if only my brain can do the same... or my heart even so i wont be thinking about this man day in and day out...   '),(4377,3170057,'23,April,2004','2004-04-23 00:00:00','   urlLink The New York Times > International > Asia Pacific > The Hostages: Freed From Captivity in Iraq, Japanese Return to More Pain   Hmmm...I\'m a bit speechless actually...quite shocked....hmmmm   '),(4378,3170057,'22,April,2004','2004-04-22 00:00:00','   urlLink Yahoo! News - Rampant Rhino Gets Amorous with Car  Woohoo! Animal instinct unleashed! Hehehe   '),(4379,3170057,'22,April,2004','2004-04-22 00:00:00','   urlLink Yahoo! News - Gay Lovers Climb Tree, Then Have Sex   wow! kudos to their bravery! Oral sex while sitting on top of a tree! Woohoo! Coolness to the max!   '),(4380,3170057,'22,April,2004','2004-04-22 00:00:00','   urlLink The New York Times > Education > As Wealthy Fill Top Colleges, New Efforts to Level the Field :  \"They say they are worried that their universities are reproducing social advantage instead of serving as an engine of mobility. \" ---Education should never be enhancing one\'s social stratification. But with how things are looking now, it seems that it is only the rich who can go to the best schools. And the less privileged can\'t do anything about it.   '),(4381,3170057,'21,April,2004','2004-04-21 00:00:00','   urlLink Yahoo! News - You really may be addicted to that chocolate cake   Ahhhh....this is true! hehehehe   '),(4382,3170057,'20,April,2004','2004-04-20 00:00:00','   urlLink BBC NEWS | England | \'I sold myself on eBay\'  Hmmm..London is really interesting....   '),(4383,3170057,'20,April,2004','2004-04-20 00:00:00','   urlLink BBC NEWS | Technology | Curiosity fuels anger at mobile chat  Maybe these people are enraged because no one is calling them on their mobiles...sheez, sometimes i just can\'t understand humans!   '),(4384,3170057,'20,April,2004','2004-04-20 00:00:00','   urlLink CNN.com - \'Toothing\'ï¿½craze goes underground - Apr 18, 2004   Aaahhh! Makes me want to change my mobile and move to London...dang! wish I could do toothing here...oh well, i guess we just have to rely to the age-old ways of telling strangers you want to have them in bed...   '),(4385,3170057,'20,April,2004','2004-04-20 00:00:00','   gawddamit! i am going nuts over this guy...grrrr. reality check please? PLEASE! grrr...   '),(4386,3170057,'20,April,2004','2004-04-20 00:00:00','   urlLink Yahoo! News - Survey Unveils What Happens in Bedroom  Woohoo! Why didn\'t they ask me huh? I have some nice answers as well! hehehe   '),(4387,3170057,'20,April,2004','2004-04-20 00:00:00','   What my desktop calendar reads today... \"I don\'t SUFFER from insanity, I ENJOY every minute of it!\" Harharhar! Bullseye!   '),(4388,3170057,'20,April,2004','2004-04-20 00:00:00','   urlLink CNN.com - Is it OK to switch off from work? - Apr 14, 2004   hmmm....i have to think about this really hard....   '),(4389,3170057,'20,April,2004','2004-04-20 00:00:00','   Argh! I can\'t think straight. I have this manuscript to process, but damn! All I can think of is this great guy I am having the most wonderful conversation with (yeah, just convo no sex....yet!) gawddammit! my head is going bonkers and my heart is now utterly useless...ugh! hate this situation. hate this. grrr.   '),(4390,3170057,'20,April,2004','2004-04-20 00:00:00','   Hey, i am in!   '),(4391,3170057,'25,May,2004','2004-05-25 00:00:00','   urlLink BBC NEWS | Health | Low male sex drive down to mum?   Crap. Men have another reason to cough up when they don\'t feel like making love.   '),(4392,3170057,'18,May,2004','2004-05-18 00:00:00','   Shit. It feels like a first kiss. Warm, tingling sensation. Darn. I haven\'t felt like this before. I AM DEAD.   '),(4393,3170057,'13,May,2004','2004-05-13 00:00:00','   urlLink Yahoo! News - Aging Octopus Finds Love at Last   Awww! Good for J-1 and Aurora. I hope the rest of humanity catches up.   '),(4394,3170057,'12,May,2004','2004-05-12 00:00:00','   urlLink Yahoo! News - 3D Church Opened to Woo Internet Faithful   on-line worship? I wonder how the Almighty will react...   '),(4395,3170057,'11,May,2004','2004-05-11 00:00:00','   urlLink BBC NEWS | Health | Morning coffee \'not best boost\'   oh crap! now what? =D   '),(4396,3170057,'11,May,2004','2004-05-11 00:00:00','   urlLink The New York Times > Health > Personal Health: The Risks and Demands of Pregnancy After 20   \"...a woman\'s fertility declines steadily after age 20 and precipitously after age 35.\"   Oh no! mY unfertlized eggs!!!!   '),(4397,3170057,'06,May,2004','2004-05-06 00:00:00','   hmmm...should i stop texting him? e-mailing him? gawd, i don\'t want to...not after being honest...girl, hold your horses! let him miss you..and keep your finger crossed that he does...   '),(4398,3170057,'06,May,2004','2004-05-06 00:00:00','   urlLink The New York Times > Technology > Technology Special > In Europe, Cellphone Profits Go Up as Clothes Come Off   Ahhh! pornograhy on the go! interesting   '),(4399,3170057,'05,May,2004','2004-05-05 00:00:00','   urlLink The New York Times > Business > Economic Scene: How Much Does Information Technology Matter? :   \"According to Mr. Carr, knowing how to use information technology is like knowing how to run an assembly line. It is a utility now, like telephone service or electricity.\"  Nice insights. hmmmm   '),(4400,3170057,'05,May,2004','2004-05-05 00:00:00','   urlLink The New York Times > Technology > Circuits > For Technology, No Small World After All   This is cool! I like her study! Nice. Darn! I miss the academe!   '),(4401,3170057,'05,May,2004','2004-05-05 00:00:00','   urlLink The New York Times > Technology > Circuits > The Internet\'s Wilder Side   Nice. But this is just one of the stories about I.R.C.   '),(4402,3170057,'05,May,2004','2004-05-05 00:00:00','   urlLink The New York Times > Technology > Circuits > The Bionic Running Shoe  Ahh! This is why I am in this field...Good lord, I want that sneaks!   '),(4403,3170057,'05,May,2004','2004-05-05 00:00:00','   urlLink Yahoo! News - CDs and DVDs Not So Immortal After All  crap. memo to myself...chech CD and DVD collection ASAP for rotting.   '),(4404,3170057,'05,May,2004','2004-05-05 00:00:00','   Feeling sexy with my fishnet stockings  Start 10:00 am End   10:25 am  Morning! Morning! Morning! Mwah! Mwah! Mwah! Come on! Come on! Come on! Dance with me! Laugh with me! This day  And if you want Until tomorrow.  Will you be driving through the metro? A little break here And then again move on Alright then I suppose I will just be waiting For you For your stop For your go.  Think of me will you? Even for a split second That will do. Just knowing I occupied your mind That would be enough To make my heart skip To make me smile a lot.   Musing of a coffee-laden mind. For my dear john.   '),(4405,3170057,'05,May,2004','2004-05-05 00:00:00','   urlLink Mirror.co.uk - BRAD: MY MID-WIFE CRISIS :   \"\'We don\'t cage each other with this pressure of happily-ever-after,\' he says. \'You figure it out rather than setting policies and rules. Jen and I always made a pact that we\'ll see where this thing is going. \'I\'m not sure it really is in our nature to be with someone the rest of our lives, just because you made this pact. \'You keep going as long as you keep growing. When that dies, we do.\' But he adds: \'Just when you think you\'ve gotten all you can out of it, you get knocked upside the head. It\'s good fun.\'\"  This makes sense! Trully!   '),(4406,3170057,'04,May,2004','2004-05-04 00:00:00','   I can feel the stick in my heart. At past 4 in the morning, I awoke sensing the something dreadful will come my way. And I was correct. I had my heart pierced, left vulnerable to a deluge of never-ending pain. He just spoke two words and these were enough to extinguish the glow from my face, the optimism in my life. â€œMy honeyâ€¦â€ Why do I have to always play the second best? Why canâ€™t I be the only one?   '),(4407,3170057,'03,May,2004','2004-05-03 00:00:00','   Fuck! I canâ€™t get him out of my headâ€¦Was trully hurt by him telling me he is with his honeyâ€¦Why canâ€™t he be insensitive regarding that matter!!!! Fuck! Fuck! Fuck!   '),(4408,3170057,'03,May,2004','2004-05-03 00:00:00','   â€œI was in the airport WITH MY HONEY to pick up her younger brotherâ€¦â€ What the FUCK!!! Why do you have to rub it my face!!! Damn! It hurts, it freaking HURTS!!!   '),(4409,3170057,'03,May,2004','2004-05-03 00:00:00','   urlLink Yahoo! News - How Google Became a Cultural Phenomenon  Will google continue to run our lives? Hmmm....   '),(4410,3170057,'01,May,2004','2004-05-01 00:00:00','   urlLink Yahoo! News - BMW drivers get most sex  Oh my! I must find an owner of a BMW ASAP! Woohoo! Superb car! More sex! Can life get any better?  Hahahaha!   '),(4411,3170057,'01,May,2004','2004-05-01 00:00:00','   urlLink Yahoo! News - Pet Owners Warned of When Cicadas Emerge : \"once-every-17-years mating dance\" --- what a life! =) hehehehe   '),(4412,3170057,'01,May,2004','2004-05-01 00:00:00','   urlLink Yahoo! News - Airline Pilot Caught Dozing in Flight  what the @&!%?!!!!! Hahaha. Good thing planes have auto-pilots now...geez!   '),(4413,3170057,'25,June,2004','2004-06-25 00:00:00','   urlLink Yahoo! News - Farrell\'s Full Monty Cover-Up   whoa!!!!!!!!!!!!!!! Massive member? what a way to describe colin farrell\'s anatomy!!!!!   '),(4414,3170057,'22,June,2004','2004-06-22 00:00:00','   urlLink IHT: The boss who\'s a bully   cool study   '),(4415,3170057,'18,June,2004','2004-06-18 00:00:00','   urlLink Yahoo! News - Octopus Lays Eggs After Successful Mating   Wow!!!!!!!! Darn, why can\'t that be me? crap.   '),(4416,3170057,'17,June,2004','2004-06-17 00:00:00','   urlLink The Globe and Mail   commitment pill... neurology of romance... nice! nice!   '),(4417,3170057,'11,June,2004','2004-06-11 00:00:00','   urlLink Yahoo! News - Economists find no link between sex, money   COOOOOOOOOOOOOOOLLLLLLLLLLLLLLLLLLLLLLLLLLL! =p   '),(4418,3170057,'01,July,2004','2004-07-01 00:00:00','   urlLink Poverty - Jul. 01, 2004   NICE @--->--   '),(4419,3590306,'30,June,2004','2004-06-30 00:00:00','   I seriously need to learn how to play the game of love better. Or I need to learn how to play hard to get. It seems that if I like someone, I\'m honest about my feeling. But I guess guys don\'t like that. So from now on, I will not be calling any guys after we make out or any such thing. Chris hasn\'t called me in 2 days. I\'m kind of upset about it. I just don\'t get it. He told me I was his dream girl. I swear that line was just to kiss me. But I told him that I wanted to date him because I feel like a slut just making out with him all the time. So I kind of have a feeling he isn\'t calling me is because he doesn\'t have any money to take me on a date so he\'s just avoiding the whole situation. He is making me so angry. Well, I\'m moving on. I\'m keeping my options open. I don\'t want to loose out on any good opportunites. Plus my cousin Hannah is coming tomorrow to stay with me for the whole long weekend. I\'m so excited to see her. So I\'ll make myself busy so I won\'t hang out with Chris when he comes crawling back. I can play these games too!! I went to institute tonight and it was so good. I learn so much about the church and the scriptures when I go. I feel so good when I\'m there. It\'s awesome. I need to make more mormon friends. I feel like I\'m getting a lot, but there always room for more. I like living in Long Beach. It has been the best decision I have made in a long time. I\'m happy here. Ok, talk to you later.   '),(4420,3590306,'27,June,2004','2004-06-27 00:00:00','   Well, today is Sunday and I really should be getting ready for church but I feel like procrastinating. I\'m actually hungry right now and should get some breakfast. Ok, well Chris called me on Friday but we couldn\'t hang out because we both had different plans, but he ended up coming over late at night and we ended up making out. It was really, really fun. I\'m not sure how much I like him, but I like kissing him. So that should mean something, right? He\'s actually really cool and down to earth. He\'s kind of had a crazy life, but he turned out surprisingly normal. So, on Saturday, I went to Disneyland with Mandy because work got us free tickets, and I was so tired. Then I came home, took a nap, and then got ready for my date with Layne. He\'s a super nice guy, I don\'t think my type. But we had a crazy night together because all these crazy things started to happen. Anyways, it actually turned out to be really fun and we had a good time. I seriously doubt he\'ll ask me out again. I think I am too wild for him. Which is ok, because he needs someone better than me. He wants to live in Utah for goodness sake! Ok, I better get ready for church. Byee!   '),(4421,3590306,'24,June,2004','2004-06-24 00:00:00','   I have been perfecting my new song I wrote and I played it for Thorey tonight and she said she really liked it. She told me I was really talented and stuff. It made me feel good about myself. Sometimes I just get so down on myself as a musician. I dunno, I just need to get better. After playing music, me, Thor and Isabel went out to dinner at some mexican restaurant on 2nd street. Isabel leaves tomorrow for NY for the rest of the summer. I\'m going to miss her, but she\'ll have a lot of fun. I also bought a new shirt tonight for my date on Saturday. It\'s a super cool layered shirt. I look hot! Um, I thought Chris would call me tonight because he got home from mexico today, but maybe he is tired or something. Maybe he\'ll call me tomorrow. Maybe he doesn\'t like me? I dunno, but our last conversation was so good. I mean it was good for me, but I don\'t know about him. He\'s so mysterious, I just want to know more about him. I hope he calls me. But if not, I\'ll just keep going out on my dates with dorky boring guys. Ok, I need sleep. Later.   '),(4422,3590306,'22,June,2004','2004-06-22 00:00:00','   I ate some bad food tonight and now I feel sick. I went with Matt and Jeff to some Irish Pub and had really horrible food. Irish food sucks. Anyway, it was fun hanging out with those guys. It\'s kind of refreshing, but then I don\'t think I could hang out with them all the time. I serisouly don\'t know what\'s up with my social life. Dating totally sucks. I have a date on Saturday with some guy I don\'t even know. He seems nice and all, but it\'s wierd to just go out with some person you have no interest in or have no idea what\'s he\'s about. But whatever, I\'m just going to go and try to have fun. He seems like a total dork, but I just have to go once and I\'ll never have to go again. I\'m trying to keep my options open because you never know who you\'ll like. But then again, I know I won\'t like him. I don\'t even think I\'m attracted to him. So, my bad luck with guys will continue. But I feel like I learn a little about myself with each one. And since I did absolutly no dating while I was in college, it\'s now my time that I did. I\'m so sick of talking about boys, but that\'s all I have to talk about. Ohhh, today I picked up my diploma. I got all sad and cried a little bit. It was offical that I\'m done with college. It made me a little sad. I was just thinking about all the times I had and all of the memories, and it was all over. Geeze, it went by so fast. I can\'t believe it\'s over. In the moment, it seemed so slow, but now, wow. I\'ll be sad about it for awhile. Ok, talk to you later. Byee!   '),(4423,3590306,'16,June,2004','2004-06-16 00:00:00','   urlLink    Horn&nbsp; urlLink   '),(4424,3590306,'15,June,2004','2004-06-15 00:00:00','   So I wrote a new song tonight. I actually wrote the lyrics a while ago (like 6 months or so) and I have just been waiting to use them. I made up song catchy little tune and I think it\'s an okay song. I\'m so proud of myself because write a new song was one of my goals and I have already accomplished it! I can scratch that off my list. So, why does money always make things so aweful?!! At work, they are trying to figure out some sort of bonus structure, and I never cared about a bonus before, but now it\'s getting me to think about money and I\'m getting money hungry! It\'s totally sucks. I don\'t want to be like that. I just need to stop thinking about all the money I could potentially get and be happy with what I got. Because it\'s not like I\'m going to stay with this job for the rest of my life. One of my goals is to be able to quit my job. Ahh, well, tomorrow I\'m going to go out and socialize and try to meet some new people. Okay goodnite!   '),(4425,3590306,'14,June,2004','2004-06-14 00:00:00','   So, I made my list of goals and it feels good to write them all down. I already feel better about myself and my life. If I accomplish everything on my goal list, I will be a happier person. Plus I will have to satisfaction of crossing things off a list, that\'s always fun!! My hip hurts today because I went for a run, when I haven\'t ran in a long time. I\'m totally out of shape but one of goals is to loose weight, not that I\'m fat or anything, I\'m just not where I used to be and it\'s bugging me. I also played volleyball today at church today and it kind of scares me. I feel like I\'m going to get hit in the head. That would suck to break my nose or something. Work is going okay. I guess I\'m going to start getting some sort of bonus. Whatever, I don\'t even care. I have never gotten a bonus before, so I guess anything would be nice. It really doesn\'t matter to me. Well, I\'m going to go to bed now, See ya.   '),(4426,3590306,'13,June,2004','2004-06-13 00:00:00','   I went to church today and it was okay. It was testimony meeting and there are some crazy people at church. I didn\'t really get to meet anyone new, but I have an okay time anyway. I like my friends at church, they are all pretty cool. There is this one guy who is a friend, but he\'s way old, and I think he likes me or something because he keeps on trying to go out with me. I try to avoid it as much as possible. I like him as a church friend, but nothing else. Plus he has a reputation for being a womanizer. I\'m not even attracted to him. Today we learned about setting goals, so I\'m going to start making and tracking my goals. I have a few things in mind ... like paying off my credit cards, going to that music school in hollywood and eventually quitting my job. All good things. Even though it\'s only been 5 months since I graduated college, I feel like I\'m doing nothing with my life. I feel like my life is so pointless right now. And since I\'m doing nothing extra curricular, I focus all my attention on boys, even though I have never been boy crazy or anything. I think boys should be a priority, but not my only priority. I\'m going to go make my list right now. See ya.   '),(4427,3590306,'12,June,2004','2004-06-12 00:00:00','   Last night was the first time I saw him after we stopped hanging out with each other. It was wierd not to talk to him and spend time with him. I didn\'t realize that I was so used to hanging out with him, that when I didn\'t, it made me sad. I just need to get over it and move on. I know that he isn\'t the one for me, but it was nice to have someone for a little bit. I went to the movies today and saw Napolean Dynamite. It was awesome. Really funny. It\'s going to be the next cult movie. It reminded me of Wes Anderson movies. I felt like today was totally just flew by and I just kind of wasted it. I woke up late, went to the movies, hung out with my sisters, watched tv, rented a movie and didn\'t take a shower until 5:30. And I have felt tired all day. I feel more tired today waking up at 10 am, then I do when I wake up at 7am. I\'m excited to go to church tomorrow, but then again I\'m not. I want to go and feel the spirit and take the sacrament, but I just don\'t want to deal with everything that comes along with it. I dunno, I always feel like this, but then when I get there I feel better. Maybe I\'ll make a new friend. We\'ll I guess I\'m going now. I feel tired and want to sleep.   '),(4428,3590306,'09,June,2004','2004-06-09 00:00:00','   I\'m 23 and living in Long Beach.  I graduated from college and now working a pretty much a dead end job. I only stay here because it pays my rent and my car payment. I\'m just waiting around to see what I want to do with my life. But all in all, I\'m happy. I do not have a boyfriend right now, which is okay, but I would like to get one. How does one get one? Well, I\'m trying. I have been on so many dates, and nothing ever comes of them. Well, I just haven\'t met Mr. Right yet. I need someone creative and musicical. Someone that makes me laugh and challenges me. I think sometimes that I\'m too picky, or I will not like someone for 1 bad trait. Maybe I need to evaluate my standards. I dunno - but, I have been in love (or maybe lust) and it\'s fun, but it sucks at the same time. This is all for now. I will write more later. Peace out.   '),(4429,3590306,'27,July,2004','2004-07-27 00:00:00','   Tonight was fun. I went to the movies with Matt and we saw the Bourne Supremacy. It was good. Now I\'m just relaxing and dreading another day of work. Working full time was good for awhile, but now it\'s starting to get to me. Everyday is so beautiful here and I\'m stuck inside an office. I wish I had a little more freedom. Other than that, my life is okay. Saturday is Mandy\'s baby shower and I\'m kind of in charge of it. I seriously don\'t have any money. I\'m so broke. I\'m going to have to so skimpy on her baby shower present. Oh well, I\'m throwing the whole thing for her, so that should be her present. Ahh, I wish I made more money. I need to get a raise, or find a new job. I just want to get out of debt. If I didn\'t have to pay credit cards every month, I would be a rich lady. Credit cards are evil! But yet I still get them. I had my fun in the past, but now I have to pay them off. Well, I\'m off to have some chocolate milk. Later.   '),(4430,3590306,'23,July,2004','2004-07-23 00:00:00','   I have been thinking a lot about life lately. Mostly because my nephew goes into surgery next week and I\'m a little scared. This will be his last surgery, thank goodness. We are just praying that everything will go well and he will be healthy. My sister in law and brother have been through so much, and especially little Owen, he doesn\'t even know what\'s going on. My heart is just so heavy. I can\'t wait until it\'s all over with and everything is fine. I\'m going to their house tomorrow to pick up Sam and spend the day with him. I haven\'t seen him since Thorey\'s wedding. I miss my family. I guess I\'m just lonely right now. I\'m just thankful for all of my family and how much love they give me. I know that if I follow and keep the commandments I will be fine. Sometimes I just go through some panic attacks, like I don\'t know what I\'m going to do with my life and it scares me. It\'s like, college is the only thing I thought about for a long time, and now that it\'s over, I don\'t know what to do. I actually have a lot of options, but I don\'t know which road to take. The 1st counselor in the bishopric told me that I should consider a mission. I told him that I was too old for one, and he said I wasn\'t. I still could go if I wanted. I don\'t know. Now I\'m thinking I want to go on one. Or do I just want to escape reality? I\'m really confused. The only way I could go on a mission is if I paid off all of my debts. I have a lot, so it might take me a while. I seriously embrace change, I hate for things to stay the same. I hate my life to stay the same. I want to be doing something different all of the time. Or I just need some more life experiences. College gave me some, but not enough. I think praying and reading my scriptures for revelation is the only way I\'m going to figure things out. I have faith that Heavenly Father will provide my way. Ok, bedtime. Goodnite.   '),(4431,3590306,'21,July,2004','2004-07-21 00:00:00','   Tonight I went to see my friend Matt\'s musical he play percussion in.&nbsp; It was a lot of fun.&nbsp; The musical was really entertaining.&nbsp; Matt is super talented.&nbsp; He will be famous one day.&nbsp; I seriously need to start excersizing more!&nbsp; I\'m getting really fat and it\'s depressing.&nbsp; I thought that working full time would make me loose weight, but it\'s the exact opposite.&nbsp; I hate it.&nbsp; That\'s it! I\'m loosing weight.&nbsp; I went off my budget a little bit.&nbsp; I love tracking all of my expenses.&nbsp; It\'s really interesting to see where all of my money goes.&nbsp; It\'s just like a business, you have to track where your money goes to save money.&nbsp; It makes sense.&nbsp; So, I have no dates for this weekend, nothing fun to do.&nbsp; I\'ll probably just hang around and do nothing.&nbsp; I wish I liked someone.&nbsp; Blah, blah blah. Ok, bedtime. goodnite.   '),(4432,3590306,'19,July,2004','2004-07-19 00:00:00','   Today was good.&nbsp; I just went to work and after I tried to exercise but I pulled a mussle in my back.&nbsp; It hurts really bad.&nbsp; Then I went to FHE and it was okay.&nbsp; It so depressing when you don\'t like anyone.&nbsp; I feel kind of lonely right now.&nbsp; I hate it when I feel loney because I do stupid things like call guys that don\'t like me and look pathetic.&nbsp; Well, I\'m not going to do it.&nbsp; I\'m going to be tough and not call anyone.&nbsp; I just feel like I\'m at a point in my life where I want to get a boyfriend.&nbsp; I\'m so sick of first dates and all that stuff, I\'m ready for a relationship.&nbsp; I just don\'t know how to get one.&nbsp; Plus there isn\'t anyone I like good enough to start one with.&nbsp; I just need to wait it out and it will happen for me.&nbsp; I\'m trying to get my life in order.&nbsp; Today I made a budget to track all of my expenses.&nbsp; I\'m going to be way more careful with money.&nbsp; I\'m going to start by always paying my tithing.&nbsp; I haven\'t in a long time and I need to start.&nbsp; It\'s a commandment I need to keep.&nbsp; This week is going to suck because we are traveling doing billing.&nbsp; I just hate it so much.&nbsp; But I need to keep a positive attitute because if I don\'t, I\'ll end up hating my job.&nbsp; Ok, well, I guess that\'s enough.&nbsp; Bye!   '),(4433,3590306,'17,July,2004','2004-07-17 00:00:00','   I am feeling better, not 100% back to normal, but better. I slept a lot this weekend. I didn\'t end of up going to traffic school today because I was too tired.&nbsp; I stayed out with some friends last night way too late, and being sick and all, it took it\'s toll on my body.&nbsp; So, I\'m going to have to reschedule.&nbsp; Tomorrow I\'m not going to church because I have threate tickets.&nbsp; I really don\'t like going to the theatre on Sunday, but my grandma loves for me to go, so I really don\'t have a choice.&nbsp; I really don\'t get to see her that much, so it\'s good quality time together.&nbsp; It\'s sooo hot right now I think I\'m going to pass out of heat exgaustion. Seriously, I need to buy a fan, but I\'m too poor. On Tuesday I\'m buying a fan!!&nbsp; I\'m sweating to death. So, what else, I kind of just speant the whole day by myself. Tonight Suz made dinner and we just stayed in and watched movies. It was nice.&nbsp; We really didn\'t have a choice because we are so poor. Ok, I need to go to bed. I\'m going to pass out! Goodnite.   '),(4434,3590306,'15,July,2004','2004-07-15 00:00:00','   I\'m getting over my cold quite well. Now it\'s all in my nose and sinus, so I just can\'t breath. Which sucks because I can really only breath from one side of my nose anyway. I was so tired today because last night Suz and I stayed up until 2 in the morning talking to our friend Brian. It was really good talking to him, but we stayed up way too late. Today at work I was hurtin\' so bad. So I came home and instead of going to enrichment night, I fell asleep. I should have gone to enrichment, it would have been fun. Bummer. I just couldn\'t get myself out of bed.&nbsp; I officially have no plans for this weekend. Suz is going to Palm Springs tomorrow and I want to go but I have traffic school all day on Saturday.&nbsp; I hate traffic school! It\'s my second time going. I don\'t ever want to get a ticket ever again! I am so poor right now. I only have like 50 bucks in my checking account. How did I get so poor? I feel like I don\'t make enough money, so much to the point where I want to get a second job.&nbsp; I don\'t know what else to do. I need a raise. I\'m never going to get one. I need to quit my job and do something else, but that\'s not going to happen either. Well, I\'m getting really fat. I haven\'t excercized is a long time, and it shows. As soon as I get over my cold, I\'m going to start working out again. It\'s totally getting way out of control! Ok, well I better go to bed now. Goodnite.   '),(4435,3590306,'13,July,2004','2004-07-13 00:00:00','   Well I\'m officially sick. I went to the doctor today and he put me on antibiotics. Which is good because I need to get better. I hate being sick. I had to cancel my date with Chris tonight to go to the Pagant of the Masters, which I really wanted to go to because it would be cool, but I\'m too sick and contagious! Hopefully by this weekend I will be okay. I have traffic school on Saturday and really not looking forward to it. Today I bought a new computer workstation. It fits my apartment so much better and it\'s really cute. Now I need to buy a entertainment center and a new tv, then I will be complete. I can\'t believe how much I have been sleeping since I have been sick. It seems all I do is sleep. Which is nice because I love sleeping. I don\'t have any plans for this weekend really. I have theatre tickets on Sunday, so I won\'t be able to go to church. But I\'m still going to hang out with my friends afterwards. Ohhh, geeze, it\'s been so hot lately. I need to go buy a little fan because it\'s totally miserable. I hate not having air conditioning. Ohh, wait, I won\'t be complete just with an entertainment center! I still need to buy a little kitchen table. Just a little two seater table. Then my apartment will be complete. Ahh, I can\'t wait. I just need more money! Ok, I\'m going to go get ready for bed. I\'m already feeling better :)   '),(4436,3590306,'11,July,2004','2004-07-11 00:00:00','   Having Hannah here was so refreshing. It was good just to relax and hang out and have fun. We did a lot of fun stuff together, but mostly we just loved being in each others company. I wish we lived closer and could hang out all of the time. I often think about living close to my cousins and having that sort of relationship with them and how it would go. I bet I would just love it. I\'m trying to convince Hannah to go to grad school down here next year, so that way we could live together. That would be so much fun. Well, Chris eventually called me, but I\'m so over it. I talked to him a couple of times, but he\'s so shady and I have no clue on what he thinks about. I had a date on Friday with some other guy named Chris and he is totally nice. But I don\'t know if I totally like him. He\'s a little bit shorter than me, and I don\'t know if he\'s my type. It really doesn\'t matter though because he\'s moving to Utah at the end of the month anyway. Which puts less pressure on me. Then Mike from Huntington Beach called me Friday to go out with me. When it rain, it pours. He is a totally cool guy too, and I could actually see myself liking him, maybe. I dunno. I should give him another try. He has his life together, which is better than the other guys I\'ve been going out with. Enough with my love life, I\'m so poor right now. I wish I made more money. I thought living on my own I would have more money, but I was wrong. I need to get my finances in order. There are so many things I need to do. Geeze. Okay, I better go to bed. Later.   '),(4437,3339644,'31,May,2004','2004-05-31 00:00:00','   Hura was born on a cloudless night. It was so very calm that one could hear the leaves outside. Inside that small hut a midwife was trying to deliver the child. Inishia was in a lots of pain and was cursing everyone she knew. Outside the hut Hakulla was pacing up and down with a couple of vilagers. He knew that bad was going to happen. Finally the midwife came out running and said something in Hakulla\'s ear, he had lost Inishia. Hura came to the world and was branded as a devil -- thee devil who killed his mother and started growing up like one.  There was not a single day when Hakulla wonk beat him up. When Hura was 6 years old he learnt that in order to survive the world he has to cut throats. His world at that time was his family of a ever drunk father and 10 other brothers and sister. Hura learned his surviving skill very early and proved his skills at the age of 10 when he sold his eldest sister for 20 dollers. It was a big sum of money those days.  Growing up wasnt easy but he did very well. By the time he was 19 he had 19 brutal murders and a couple of espionage under his name and had gathered enough money to escape to Russia.  I met Hura on one of my missions, i had to hire and train one *out-of-the-agency* and *expendable* agents and he seemed to be best suited for the job. He had a passion inside him -- a passion to find things out and kill!! He had those clear eyes and there was one thing very clear in those eyes -- he meant business -- he was the best natural killer that i ever met.   '),(4438,3339644,'25,May,2004','2004-05-25 00:00:00','   Well i know you guys must be wondering where i am.. well dont think that i am still inside that hotel room at Karachi,Pakistan with Tequila (Ha!!)  Guys i am just busy doing what i love doing -- my disappearing act... I will be back!!! sooner than you realise, with another lesson to the \"Diary of a Spy!\" believe me the wait will be worth it..  Neo aka Vikas   '),(4439,3339644,'21,May,2004','2004-05-21 00:00:00','   Friday, 21 May, 11:45 PM Karachi,Pakistan  I tried her number in vain a couple of times. I got so sick of her answering machine that i poured myself a very large vodka, sitting and dragging on to my cigeratte i was thinking about the day that went so well. The Mole field is very big. It is a very planned and organized operation and i have my doubts that it is not one organization that is running it. It has to be a combined effort of KGB, CIA may be ISI too. The roots go very deep and are very well hidden. It is not going to be as easy as i thought. I was shaken from my deep thought by the constant ring of the phone. It must have been ringing for some time now.  \"Who is it\" i barked. There was a silence on the other side, i could hear the person lighting a cigerrate. \"So you are the one.. huh?\" He said in a heavy hungarian accent. \"Listen to me very carefully coz i wont repeat. That girl is going to die a very slow death. Tell her to stop nosying around my business\" The click on the other side was heavy. It was a Beretta 105, after 12 years of field job i knew the weapon by its sound. \"Who is this? What girl?\" I said.  The phone went dead. The first thought that came to my mind was that it was some drunk who was playing midnight pranks, but there was a certainity about his voice. Then it hit me i have to reach to Tequila. I dialed her number again but was greeted by the same machine \"I am out leave me your number and i shall call\". I had to do something.. i cannot let her go waste like this.  Tring!!  The doorbell. God they still have those old fashioned bell in Pakistan. Ha!! I collected my Beretta and opened the door it was her looking exactly the way she looks. Time never seem to have any effect on her. She was wearing a pink evening gown. Gosh she looks so good in Pink, she always has. She pushed me inside with the same primitive laugh and said gear up it is going to be one long night.   '),(4440,3339644,'20,May,2004','2004-05-20 00:00:00','   Friday, 21 May, 4:45 PM Pakistan  Here i am standing, in the worst downpour of my life, waiting for my contact to deliver the package. This mission is a Mole Hunt. I was tipped by my contact in KGB about a potential and widely spread mole activity happenning right inside the CIA. We have lost men - good men. The director briefed me about the gravity of this mission. I have to be very careful.  Deja Vu!! There goes the same Red Ambassador  again (Why red!! Ha!!)  This cannot be a coincidence!! \"When you are in the field remember one thing there are no coincidences. A coincidence means that your position is compromised\" said the director during one of the training sessions.  Suddenly the Red Ambassador passed me again and stopped a little ahead of me. I was already ready with my Beretta 66. Ready to confront whatever steps out.  It was Tequila suddenly i started feeling the same \"Weak-knees\" syndrome. So she was a KGB afterall. I havent been wrong. She approached me, she was wearing the most alluring smile (ofcourse apart from that shear Red Gown!! Ha!!) She came very close to me, i could smell her cigerrate. So she hasnt changed. \"Can you smell the rats?\" she said in a whisper. \"The rats are all over\" i said. The packets changed hands and she turned away and walked towards the Car.  I wanted to stop her, ask her where she had been all these years. She turned and threw something towards me. I ducked. She let out a primitive laugh. \"If i wanted to kill you then i would have done that long time back\" she said. She was gone.  Bingo!! i have her number now.   '),(4441,3339644,'20,May,2004','2004-05-20 00:00:00','   Friday, 21 May, 3:45 PM Pakistan  Here i am on yet another mission. Outwardly this mission is no different than any of the earlier ones. This life has always facinated me but it has  its drawbacks too. I never understood what Tequila wanted, i always thought she was a KGB operative. Ha!!  Looking back i realise that she was actually as innocent as Hura (thats another story - shall write about it in detail someday) now i wonder that all the time while i was thinking that she is looking for my gun, trying to disarm me, what was on her mind?  She was actually looking for thee Gun. She was actually trying hard to disarm me.. Ha!! But that was a different time - a different era.   '),(4442,3339644,'20,May,2004','2004-05-20 00:00:00','   Sorry guys i hadnt posted for some time. I have been running places to get *something* done. It is a \"classified P4453$R\" project so cant tell you a  lot about it. All i can tell you is that there are some agents that need to be straightened up. I know you will be worried about a lots of things (the extermination of the entire human race for example) but dont worry \"Mai hu na\" * \"Mai Hu na\" is a hindi phrase that means \"i will be there\"   '),(4443,3339644,'17,May,2004','2004-05-17 00:00:00','   Friday, 13 January, Munich  Waiting for Tequila. Don\'t know whether my contact in  Korea has passed on the correct information to me. These damned CIA guys messed up Langley too. Waited for long...no sign. I have already called Tequila through the scrambler phone in Zurich to get past the taps by KGB. See a tall girl furtively making her way towards me. The British intelligence was right. She had big boobs...I hope she is not hiding a transmitter in them. She whispered, \" The evening feels great\". I whispered back, \"My underwear feels brown\". You have to whisper when your underwear is brown. Those years as a mole at ISI were not in vain. Contact established. We made our way past the thick crowds in the even thicker darkness towards a coffee rendezvous.  Saturday, 14 January, Vienna  Could not write more last night as I think I was being followed. Tequila quickly briefed me on my assignment. I was to make sure I charmed her. This looked difficult. The last time I charmed a person, she turned out to be a â€œheâ€. Tequila gave me the inside information on the BAR (must find out what it stands for). The BAR was fabulously organised. Wasn\'t sure how I was going to get past. Her sources were impeccable said she as she quickly tried to pass me some files under the table. I think she was just testing me. It was just her foot making its way up on my thigh. Think she was looking for my Beretta Magnum 33. Luckily, I have kept it in a place where she will not find it. Ha! Even I cannot find it. I have been looking for it for almost 3 months now. Tequila said that she was in the loop for 6 months but her cover as a waitress won\'t fool me.  I had a coffee ...black...just like my future seemed to me given the task at hand at breaking in BAR. Damn, I should have got my Mauser along. The BAR had a front. It turned out to be some place where they serve drinks. (Still donâ€™t know what BAR stands for). There were 2 very muscular men at the front. Things did not look good. Tequila just smiled at them and unbuttoned the top of her low cut blouse and walked past the bouncers swaying her hips. I too did exactly the same. Must do that to blend in. Don\'t want to attract too much attention.  Sunday, 15 January, London  Sorry â€¦.I had to leave the adventure halfway yesterday too. I suspected some bugs. So where was I, yeah Tequila went to the bar table and ordered a Screwdriver. She said she liked long hard things and winked. I think she knows about my rifle, the Telescopic Gewehr 45. Damn, she was good. Things don\'t look good. The dance floor was alight with people. She took my hand and led me to the dance floor. I checked the ceiling strobes, it was difficult as Tequila was trying to French kiss me so I had to dance with my back towards her. Ha, I am good too. The strobes looked goodâ€¦no cameras.  Suddenly Tequila slid close to me and asked me whether I wanted to go to some place alone. Ha!!! She was obviously thinking of disarming me with unarmed combat. I slid back. I was pretty sure that the KGB had tipped her off. It was time to make my move. I pretended to take a call on my hand phone and told her I had to leave. She said she would call me. Nice girl, I look forward to our next date.  Thursday, 19 June, Paris  Tequila never called. I wonder why.   '),(4444,3339644,'17,May,2004','2004-05-17 00:00:00','   There\'s a hole in the world tonight There\'s a cloud of fear and sorrow There\'s a hole in the world tonight Don\'t let there be a hole in the world tomorrow  They say that anger is just love disappointed They say that love is just a state of mind. But all this fighting over who is anointed, Oh, how can people be so blind?  Oh, they tell me there\'s a place over yonder Cool water running through the burning sand Until we learn to love one another,  We will never reach the Promised Land  There\'s a hole in the world tonight,  There\'s a cloud of fear and sorrow There\'s a hole in the world tonight Don\'t let there be a hole in the world tomorrow   '),(4445,3339644,'15,May,2004','2004-05-15 00:00:00','   Dont know what to write today!!!  Ever tried to hold on to something only to realize that it is slipping away from your hands??? They say the more you hold on to something the more difficult it will be to actually keep it in your hands. It is like holding sand in your hands.. if you keep your grip tight then the sand will seep out of your hand. Keep it loose and it will stay in your hand. But human mind looks for control!! We try to hold on to our loved ones close and tight. Should we let them go?  Something inside me told me that i should let her and her thoughts go away.. and this thought made me so sick. I thought that i can never let her go. But then all of a sudden it was so simple. I just have to kill something inside me.. this slaugher needs to be slow. It will resist but it will eventually die. So let it be (dead).   '),(4446,3339644,'15,May,2004','2004-05-15 00:00:00','   So folks B-day Still continues and thus todays Blog shall appear tommorrow...   '),(4447,3339644,'14,May,2004','2004-05-14 00:00:00','   The first time i saw her i thought she is preety, but so are almost all the girls around the globe. We didnt talk to each other for a long time, i was afraid that i will start liking her and a romantic relationship is the last thing i would want to do at office. But i always admired her, she was beautiful, simple, elegant. She is the most soft-spoken girl that i have ever seen and has those big eyes that poets talk about.  And then i thought what is the harm in having a simple friendship, i was decieving myself. Things were great those days, we used to chat so much, we used to meet up for coffee and would just chat and chat. It was like a dream - a beautiful one. Somewhere in the middle of those innocent chats she make her way into my mind. And then soemthing happened, something that has never happened to me before, something that was totally alien to me. I would think about her all the time just remembering what she said, how she looked at me, how she laughed. I would run to my cell phone as soon as it beeps thinking that probably its her and for the first time in my life i felt what it was to be in love.  The more i thought about her the more i realized that she is *thee one* for me.  I started giving her subtle hints about how i feel and was on top of the world when she didnt tried to stop me or push me away.   Ok so you read it till here.. $ 100 to the person who suggest an ending to this love story :)  Please leave your names in the comments so that i can add a note of acknowledgement.  Happy thinking!! :)   '),(4448,3339644,'14,May,2004','2004-05-14 00:00:00','   One of the good things about this cubical job is that u work(work!!!!) 5 days a week. So a saturday is an official off, except when u have one of those deadlines :) looming over ur head. It is no deadline around and still i am at office.. WHY???? Coz they have bloody declared it working and have snatched one of the virtues of this cube job off us... Not fair!! But then whoever said life was meant to be fair..  She is looking beautiful today... just like everyday.. too bad now i cant tell her how beautiful she looks.. Have a party tonight. its Hirdu\'s bday so it is going to be a blast.   '),(4449,3339644,'14,May,2004','2004-05-14 00:00:00','   Someone asked me define pain? Pain!!! Imagine this... it is Friday 4:35 PM and you are sitting in ur cubicle and trying hard to pretend that you are a busy man... trying hard to keep those eyelids open.  Imagine that she is not talking to you.. actually she is clearly avoiding you coz you again said something (she doesnt want to hear) and she is lamely pretending that she didnt understood. Imagine that you have checked you mails zillion times with the hope to see a new mail.. but looks like it is a dead world.   Do i need to say more?   '),(4450,3339644,'14,May,2004','2004-05-14 00:00:00','   So folks, its TGIF time again...Leaving for Priya at 1600 hrs.   '),(4451,3339644,'14,May,2004','2004-05-14 00:00:00','   What now!!! This is one BIG question that haunts me since the day i joined HCLT. There is just not enough work to do. Well there is a lots actually but who wants to write word docs on how to do quality work and implement CMM practices when actually no REAL (!!) work is being done. Just did what i love doing at office. Lunch time with her is something i look forward to everyday. But now she is back to her seat and talking on Realpop is not that attractive.  Anyways captain is on the deck so lets start pretending that i am very very busy :D Cya later folks i have docs to fight.   '),(4452,3339644,'14,May,2004','2004-05-14 00:00:00','   Call it inspiration or boredom from my regular cubical job.. Today when i woke up the first thought that came to my mind was that may be i should start jotting down my random thoughts...  And here i am :) well frankly i started some time back on Rediff Blog somehow it didnt work. Lets see how long do i stay here.   '),(4453,3533397,'03,June,2004','2004-06-03 00:00:00','    A tiring day indeed.   We had our class chalet 4 days ago. Which means on the 31/05/04. Err,rite? It was fun though.   1st got up in the morning ard 10am. (which is so not me! usually i wake up at 3-4pm!) I make so much noise till ayn got up and start to swear and curse me. Hahaha! Funny!   Went to eat at Long John Silver and bumped into Syidah,Syikin,Hanisah & Reza. They are as pretty as ever.   Smiling and waving. Haha!   Then off we go to Yishun. Took 969 to meet Aidil coz i wanna borrow the Honey vcd.   Aidil called to say he couldnt send the vcd. Instead he asked his youngest bro to send it to me.   Reach Yishun den me and ayn were like talking and talking. Laughing2 den suddenly this cute kid just came up to us and say, \"Nah!\"   I\'m shocked! Hes so cute! He\'s actually Aidil\'s lil bro. Hes name is Alfian. Damn cute! Ayn and me was like DAMN! HES CUTE! Haha!   Then off we went to Tamp back. We walk to tamp library bus stop to take 15 to Pasir Ris. We sat at the bus stop and were like arguing weder 15 pass that bus stop den came bus no.28 and there is Shasha and her lil bro. Her straight hair irritates me. But i\'m soooo happy i got to meet her! At last!   We took 15 then met Syed and his new girlfren. She\'s very pretty. Her name? Nurul. Schling in Ngee Ann Poly thiz July. Takin engineerin or sometin. Forgot. After that met Chal and Fattah. Fattah so cute. Hahaha! Alwais bump into him.   The most funny incident is tat we all lost our way while findin for Aloha chalet. It really sux! But rily funny. Nurul was like so-so tired. Can see her face. It was like,God help me. Haha!   At last,after struggling finding for the chalet,Chal finally called Serina to pick us up. We reached the chalet, evryone was like hello2! A bit crappy and awkward but we just sit one corner and enjoy ourselves. I ate lots of crabsticks while Shasha and Ayn were like aiming on the satays. Esp the beef satays. Haha! We never help with the BBQ but we help them to eat. It was nice.   The bad news of the day! Syed told me dat Acit was already married! That means i dated a married guy! Yucks! Tat never have happened b4 among all my frenz! Oh my gawd. I almost broke down and cried but nevermind. I\'ll take it as a lesson. Never to get to noe a guy anyhow. It really sux. But nvrmind.   Then i took pic with Fattah. Haha! We were like all over each other. He hugs me,i hug him. Haha! Not to worry. he\'s like a bro to me. A lil bro. No speacial feelings. Took many pics. Click here. Click there. Posing2 for the pics. Gerek! Went home then later watch Mummy Returns.   Later got to go aunt\'s place coz my dad called to say tat his coming home late.   Well, its sooo my dad. Typical.   Den got to know from someone that liza and faj said that they dun wanna talk to the ITE students. Wat the hell. Its kinda a discrimination against the ITE students. Ahh.. Who cares anyway coz im no ITE student.  Haiz, tired already. Got to go to sleep. Catch up with ya\'ll later. Bubyez!   '),(4454,3879020,'28,May,2004','2004-05-28 00:00:00','   Peep  urlLink this .  I mean really in maybe 50 years, nobody will be on the outside.  How will anyone get a congigal (sp?) visit?  And no sex is as good as that congigal visit sex!  '),(4455,3879020,'17,May,2004','2004-05-17 00:00:00','   I just wanted everyone to take note of a few fun websearches on google.  Try these two,   urlLink douchebag,  and  urlLink failure.   Both funny, both appropriate.  '),(4456,3879020,'16,May,2004','2004-05-16 00:00:00','   Peep  urlLink this one , a Mexican jet spots some weird unexplainable shi... stuff.  My theory is that  urlLink this little incedent  was probably caused by the moon men in some way shape or form.  Go Brewers!  '),(4457,3879020,'16,May,2004','2004-05-16 00:00:00','   Beyond the obvious reasons why one want a girlfriend, the one I miss the most is the movies.  I have a stack of ticket stubs on the headboard of my bed from each movie that I have seen since, well, sometime in high school I guess.  And the one thing I notice when I glance through them on occasion, other than the huge bloat in ticket prices since I was in high school, is the fact that many, in fact probably most of them come from the year or so I was dating Tammy.  Now the coverage was OK in HS.  Most of the big movies that were released during those days I saw in the theater.  Some, just seeing the ticket stub makes me laugh.  Like the stub from the debut of  urlLink Scream II,  when for some reason I had difficulty finding the correct theater after visiting the restroom.  Or the infamous  urlLink Summer of Sam,  where all of the boys from our group went to see a movie that still gets more than a few mentions as the \"crappiest movie ever\" when that topic comes up in conversation.  Seeing movies in the theater is fun, and it breeds memories.  HS was not the time of my life where I had the most bling to throw at my entertainment budget, but neither was the period when I was dating and saw more movies in the theater than any other time in my life.  But that was the time it was easiest to come up a reason to go see a movie.  It was still before we turned 21, we couldn\'t go to bars yet, but I\'m not going to get off topic talking about how that was the real killer of creative fun in my life.  Plus the dinner and a movie date is just so  easy.   Since then though I can only think of maybe two or three movies at most I\'ve seen in the theater.  One being the premiere of  urlLink Star Wars: Ep2,  but that is a movie where it didn\'t matter that I wasn\'t on a date, because the geek atmosphere was in full swing, and there were maybe three girls in the whole packed theater, and I don\'t think they were on dates.  In fact, I\'d go so far as to guess that many of the people in that theater hadn\'t been on a date in a real long time.  But this is just such an isolated case.  Movies are better with a companion, and with the summer movie season fast approaching again, I can look forward to missing the big blockbusters  urlLink Troy,   urlLink The Day After Tomorrow,  and hell, maybe even  urlLink Harry Potter.   I\'m sure as hell not going to see that one alone.  So if you know, or you ARE a reasonably good looking (said: not butt ugly) female, who can carry on a conversation well enough to not make me want to kill myself every few minutes, and is sick of missing all the good movies in the theater.  Get in touch with me.  PS. Making out is optional, but encouraged.  '),(4458,3879020,'14,May,2004','2004-05-14 00:00:00','   I am SICK and tired of spring.  Who could find this season beautiful?  A duck?  I got the wet butt thing happening again this morning, because my car leaks.  Fricking roof!  I just want it to be summer so I can go to a Brewer game, and get my tailgaite on.  Just me, a brat, some booze, my buds, our helmets, our outfielder defamation pages, and the sunshine in a parking lot.  That\'s all I ask for.  Oh a large group of random hotties sitting next to us would be ok too.  '),(4459,3879020,'13,May,2004','2004-05-13 00:00:00','   In some  urlLink good news  today Michael Moore\'s newest documentary, Fahrenheit 9-11, was picked up by Miramax.  In case you didn\'t know, Michael Moore is the \"Bowling for Columbine\" guy.  This strikes me as good, even though because it\'s Michael it will probably be a bit over the top, because after reading more on Dubya and 9/11 it just seems like there was a lot of fishy shit going on that I think would change a lot of people\'s mind about the current administration if they knew about it.  That is if your opinion can get any lower... :)    I guess since this is my first quasi political post on my blog, I\'d be remiss to not mention  urlLink this  beauty of a site.  I just love that title.  Pretty much sums up the way I feel about this election to a T.  '),(4460,3879020,'13,May,2004','2004-05-13 00:00:00','   In an effort to raise my  urlLink geek score  a few more points, I have started a blog.  Really that has to be the reason, because I doubt anyone I know is interested enough in reading what I think to visit here on a regular basis, and therefore make this entire project worthwile.  You know, I just realized that there is a high likelihood(sp?) that many of you that I\'ll initally tell about this blog, probably have no clue what the hell a blog even is.  Well for those of you who still do not have \"blog\" in your vocabulary I may as well give you the low down.  Blog is the shortened version of web log.  Weblogs became the most popular type of page in the days (late 90\'s) where people who cared enough to have a web presence started putting up personal pages.  I did this as some may remember, but my site 1. sucked 2. eventually got taken down by the Nimda virus and 3. wasn\'t updated nearly often enough to warrant coming back even onec a month let alone every few days.  Well, anyway some smart people got the idea to make the process to post a web log and keep it updated and looking good extremely easy.  Like sending an e-mail easy instead of publishing a website easy.  These sites are basicly the creation of a generation that grew up watching  urlLink Doogie Howser M.D.  typing away in his \"personal journal\" each night.  Damn that was a good show looking back on it, but I digress.  You also may notice that most blogs are riddled with links on the relevant topic.  Such as the Doogie Howser deal of a few sentences ago.  So neway recently the blog has become uber popular and it just seemed like a really good fit for me, since I type slow and often things that I\'d like to e-mail and tell several people about I only tell once, because it is a bitch to type out several times.  I do plan to use this blog as my outlet for just about anything that pops into my mind, some of which may be mired in controversy.  So that\'s about all for now.  I\'ll probably get a few more posts up than just this first one before I start sending out e-mails or telling people about this little doozie.  Peace out!  '),(4461,3879020,'27,June,2004','2004-06-27 00:00:00','   Hello, well the period of anti-politik is over, and I think not a moment to soon, really b4 I launch into all of that Iâ€™d just like to give some general announcements.  First off, It looks like Fatty has taken me up on my offer to become a co-poster on this page, so that is great.  It will be good to have some fresh blood and some fresh perspectives on here, and with his recent NPR kick I think he is joining up at the best possible time.  Iâ€™m excited to see his first post.  My weekend ended up not being as packed as what I had expected.  There was no trip to Gurnee, which in hindsight is probably a good thing as I think I would have been tapped by then.  It still sounds however like weâ€™ll be going to Summerfest Monday night so Itâ€™s still quite a bit of running for me, but I can run for fun.  No doubts there.  So letâ€™s see, any good movies come out over the weekendâ€¦  Hmmm.  You know I saw one called Fahrenheit 9/11 that was pretty good.  :)  But seriously, I had to drive to Madison as it was on limited release and was only shown in about 1/3 of the theatres that a movie would normally release in.  Still it grossed more than any other documentary release ever, and made more in the first day and a half than Bowling for Columbine made in itâ€™s entire run, so needless to say, there was quite a demand to see this film.  I went to the first showing  in Madison, and there was a line like Iâ€™ve only seen for the opening of â€œStar Wars: Ep2â€ in LaX formed for tickets.  It made me very feel very happy that I had bought VIP tickets online and could walk right past the craziness.  It actually felt a lot like Star Wars: Ep2 in many ways, which was my favorite movie theater experience thus far in my life.  First off the people.  Star Wars fans are loyal and had been waiting for Ep2 for a long time.  Many of them were dressed up as characters from the first film, it was a geek paradise.  Lots of camaraderie between the fans, and people who didnâ€™t know each other from Adam were engaged in deep conversation in line.  It was cool.  Same thing here at Fahrenheit.  Lots of chatter while we were all waiting in the second, longer line to get into the theater.  Broad range of people represented, many retirees and college students, these are of course the people most likely to be able to see a movie at noon on a Friday afternoon, but you also saw a few people who either snuck away from work, or got off early like me to be there for the opening.  And actually more conservatives there than I would have thought.  As you can imagine the audience was mostly liberal in make up, and strangely about 75% female for some reason, but the conversations sometimes were loud and you could sometimes hear where some very unpopular pro GW rhetoric had been spouted.  The other thing that was similar to Star wars was the â€œcostumesâ€ if you could call them that.  Many T-shirts with various slogans on them.  Some anti-war, some anti-Bush, some Pro-Kerry, some just silly.  I was wearing my â€œA shout out to the guy who lives at 1600 Pennsylvania Ave.â€ T-Shirt.  The girl next to me in line was wearing a shirt that said â€œI shaved mine to make this loud and clear â€œNO MORE BUSH!â€  She was pretty cool and funny.  I had seen the slogan on other shirts before, but it was funny to see one in Madison.  But Iâ€™m glad she didnâ€™t sit next to me, because I know I would have been distracted.  So we get inside and I got primo seating.  Like 15 rows from the screen, right in the center, an acoustic sweet spot if you know anything about the THX rated speakers that they had at this theater.  After about 20 minutes of previews some were getting antsy and booing the ads.  Here is another similarity, a very rambunctious in theater crowd that cheered parts they liked.  For instance when about halfway through the film when Wisconsin Congresswoman Tammy Baldwin popped up on screen for a maybe 15 second snippet there was a roaring cheer and clapping.  It was cool.  So the movie then I guess.  I came in knowing what this was about, obviously knowing that the film was a scathing view on Bushâ€™s handling of foreign policy as well as his actions on 9/11.  I also knew from Mooreâ€™s past film that two things always appeared in his films.  â€œStretches,â€ and on camera confrontations of those in power.  I have to say in this case they were not stretches of truth, but rather stretches of what I think logic would bring from certain pieces of news or connections between various people, but the connections, statements and events are true, so if you care to take more from them than I did than that is certainly your right.  The Saudi/Bush connections are all outlined in Mooreâ€™s book â€œDude, Whereâ€™s My Country.â€  When I read the book (which is quite well written btw) I went and checked the sources that were quoted on all of his footnotes, and found all to be there and correct.  So that was my only complaint if I had one at all.  This film blew my fucking mind.  It easily bests past documentary favorites of mine including Mooreâ€™s â€œBowling for Columbineâ€ and Steve Jamesâ€™ â€œHoop Dreamsâ€ from 1994.  I was blown away by the frankness, writing and pace of the film.  Parts are just gut wrenching, but everything I thought was covered very tastefully, which is difficult when working with topics like 9/11 and Iraq.  The thing that is also a hallmark of Moore films is a very sarcastic extremely dry wit.  I cried at this film, which is a pretty big rarity for me.  The only films I can think of that have moved me that much in recent memory are Forrest Gump, and I am Sam.  I state these two facts because at the same time that I was moved to tears, there were those around me chuckling away at Mooreâ€™s wit.  And I think the wit works well, but me personally, I am past a point where a lot of this seems funny to me anymore.  It either saddens me to a point where I just donâ€™t know what to do, or pisses me off to a point where I canâ€™t even talk about it in a normal tone of voice.  I donâ€™t want to sit and discuss the actual content of the film to any great deal, because Iâ€™ve already gone quite long here, and I know some who are planning to see it and I donâ€™t want to give anything away.  I would however strongly, strongly recommend that everyone who can get to this movie does so.  Itâ€™s something special.  So yeah, I may gloss on this again, but we have the looming June 30th deadline upcoming, and people being beheaded by the day so Iâ€™m sure there will be plenty of other stuff to talk about until then.  Hope you all enjoyed this weekend, and thanks for reading.  '),(4462,3879020,'25,June,2004','2004-06-25 00:00:00','   I wished this weekend would come, but now that it\'s here I\'m overwhelmed.  This week wasn\'t so bad after all.  I spent some time with Tammy this week, which was nice.  I did however sort of rudely fall asleep while watching one of the Seinfeld DVDs I gave her.  It\'s those chairs, and the blankets over there.  They are like Mrs. Smith\'s voice in Physics, they just make me want to go to sleep.  I was on the end of a 24 hour awake too.  So that may have played into it.  But today is the day though.   urlLink Fahrenheit 9/11  is here.  Well not here like Dubuque exactly, but  urlLink here.   Actually at only 15 theatres here in Wisconsin.  Luckily most of you live in urban areas that are showing it like Milwaukee, LaX, and The Mad City.  I plan to see it today at 2:30.  I\'ll be letting you know how it is for sure.  Then it\'s anniversary, Pasture party on Saturday, and then Fatty has invited me to Great America on Sunday, so I\'ll prolly do that.  So I\'m Chock full of fun this weekend.  This seems like it will be an expensive weekend to me.  I\'m just guessing tho.  In any event I hope you all have good weekends.  Take it easy and be well.  '),(4463,3879020,'21,June,2004','2004-06-21 00:00:00','   This was typed a few days ago, but didn\'t get posted, sorry  I read some sad news  urlLink here  this morning.  Mattie Stepanek the poet/kid with muscular dystrophy died today at age 13.  I remember seeing this kid either on TV or hearing him speak somewhere a few years ago, and even at age 10 he sounded like a person twice his age.  He had such tough luck in life, but he always had such an optimistic look on life from what I read about him in the past, that it just amazed me.  I remember him saying in his interview or whatever, that he did not fear death, which is amazing since his 3 older brothers all died of MD and his mom has it too.  I\'m nowhere near death (hopefully) and I\'m scared as shit of it.  Quite a kid.  It reminds me of the saying \"I complained that I had no shoes, until I saw a man with no feet.\"  You can read one of his poems on his homepage  urlLink here .  Where it says \"For Our World.\"  I could talk at great length about this kid, but I think I\'ll jyst leave it at that.  \"Big ups\" to you Mattie.  Tough to follow that, but I did read some good news in the State Journal this morning.  You can read about it  urlLink here.   It looks like the Badger Men\'s Basketball team will be playing a perennial powerhouse in just their second game of the pre-season.  This team is not a Division I powerhouse though.  Not even a Div 1-AA or a Div II school.  It\'s our little UWP.  Yep Badgers vs. Pioneers.  Beege vs. well shit any of the Badgers big name forwards.  Maybe even in-state phenom Brian Butch.  Mark me down for a ticket right now.  Other news of note, it looks like I\'ll not be getting around to the top 100 before Thursday, mainly because this week most of my musical focus has gone into collaberating with Fatty on the Summer CD 2004.  Details are limited at the moment, but I think my summer CDs have been fan favorites, and I\'m sure Fatty and I\'s collaberation will bring the boom.  Be well.  '),(4464,3879020,'21,June,2004','2004-06-21 00:00:00','   Ahhhh the weekend.  If only life were one long weekend.  Actually I\'d only have enough money for a few weeks if all days were weekend days.  But this was a pretty good weekend, well it was probly a super ok weekend.  I\'m not so sure I\'d go so far as good.  Neway, I went and saw the Brewers again.  Good game.  Anytime we beat the stupid \"Twin\"kies it\'s a good thing.  If there is one thing I can\'t stand It\'s being at Miller Park and seing over half of the fans there with shirts for the other team on.  Especially Minnesota and Chicago fans.  The just seem to annoy me more.  But Fatty and I eventually saw a close Brewer victory and that\'s all that matters.  We were like three feet away from a foul tip, that I did not manage to come up with, but the dude who got it gave it to a little kid sitting next to me so that was pretty cool.  I also got to see Fatty\'s new place which is a spacious 3 bdrm with storage garage attached.  It\'s nice.  I did not ask how much a month, but I\'m guessing It\'s more than my single ass can afford.  I did have to jet in the middle of the night to get home, so sorry Fatty, but I hope Connie enjoyed my illustration of the car battery and genital clamps.  :)  Saturday night was pretty uneventful.  Go out.  Have Drink.  Repeat.  Repeat.  Eat Cheeseburger.  Eat Slice of Elvis\'s Pizza.  Play Jukebox.  Have Drink.  Repeat.  Repeat.  Repeat.  Repeat.  Play Golden Tee.  Have Drink.  Repeat.  Go to Bathroom.  Piss and moan to Andy about how gay bars are, and swear never to go out again.  Have drink.  Go home.  This is pretty much the template of my weekends for the past few years.  I could live without it to be honest.  Hope everyone\'s Father\'s day wen\'t well.  I got my dad the air compressor of his dreams, but it\'s going to have to be a birthday present too, cuz damn, industrial air compressors are not cheap.  But the thing came with an impact wrench a brad and nail gun, and a paint sprayer too, so it was prolly worth it.  I was trying to find a picture of it, but Lowe\'s website uses gay Active Server Pages.  Uber Lame.  Well that wraps up the weekend.  Next weekend is The 11th Annual Pasture Party which I\'ll be attending for the 4th year, yipee.  In swoops the pasture party to break up the drudgery of drinking in bars.  Also my G&G Francis\' 50th Wedding anniversary is next weekend, so that will be cool.  I\'m happy to see anyone married that long, and when it\'s my family, I\'m all the happier.  But I\'m drawing long here, so have a good Monday if that\'s possible.  '),(4465,3879020,'17,June,2004','2004-06-17 00:00:00','   Ok I know I said no politics, but this is almost completely detached from the actual election so I hope you guys will let it slide.  In past elections, for everything from president to town board I have decided to give points to a canidate based upon how hot the canidates daughters were.  Well It\'s been tough lately.  Bush/Gore was a tight race, because  urlLink Karenna Gore  (pictured with Tipper) was not bad looking at all, but a little old (in her 30\'s when I was 20) to get the nod over the  urlLink Bush girls  (4 pictures down), who are not bad either, and also have as Roach would say, a  urlLink party slut  (also  urlLink here )reputation.  But now throw new contender Alexandra Kerry into the mix, Alexandra which is a hot sounding girl name and this... (Wait if you are at work I\'d wait on clicking this link until ya got home.  It\'s not porn or anything, but there is some incedental (whispering like Mrs. Hensley) ...nipple showing)  Ok  urlLink here .) incedent recently scores her points in my book.  She  is  30, but quite a looker.  And it really sems like older girls are very cache at the moment.  That was funny for all the right reasons.  Honestly I know this is stupid, but we would not want a repeat of  urlLink Chelsea Clinton .  God Damn!  Maybe I\'m wrong, but I just get a strong Jessie U. vibe from Chelsea, and that isn\'t good.  Oh well, be well and have a good day.  '),(4466,3879020,'14,June,2004','2004-06-14 00:00:00','   Well I\'m still settling down from all the walleye fun I could handle this weekend.  I\'ve decided that until June 24th I will not blog about anything political.  This is due to the fact that Fahrenheit 9/11 comes out on the 25th, and the handover of power in Iraq (which I\'m sure will go just as smooth as silk... eyeroll) is on the 30th.  So I don\'t want to be inundating you poor souls with the politik.  On the T-Shirts, I\'ve been looking into the screen printing facilities and should have pricing info very soon, so that is good.  I was also thinking of opening up the blog.  It\'s nice to see the recent comments from Temp and Snydes (even though it took Ryan like 7 tries) and I found a feature in my options settings that allows multiple posters to the blog.  It basicly means other people that I assign to be posters may make direct posts to my blog.  I think it could increase the free flow of ideas that might otherwise not come up in conversation in the limited time we get to see each other anymore, as well as help in the coordination of summer events.  And isn\'t that what it\'s all about, having an organized summer.  LOL Kim would be proud.  Yeah, so let me know If you\'d be interested.  Lastly as my big week before politics project, I\'ve decided to post one of my \"Top 10\" lists.  Actually this one is a top 100 list, but I was going to do my top 100 favorite songs.  I think people enjoy lists, or maybe they don\'t, it really doesn\'t matter, because I\'m posting the list on Thursday and there isn\'t much anyone can do about it.  Then you can call me a dumbass for raking some songs too highly, or omitting a favoite of yours later.  This is really a list of relatively little contriversey (sp?) and I figured I\'d start you out on the soft stuff.  Hopefully building up to better lists in the future.  So that ends this post and my Monday workday.  Have a swell night.  Oh BTW if you are the person using Broadwing Communications as your ISP let me know.  I\'m baffled as to who it could be.  '),(4467,3879020,'13,June,2004','2004-06-13 00:00:00','   Someone, possibly me, posted a whole bunch of gibberish last night when they arrived home from the bars.  I hope that this does not discredit me as a blogger or as a sane and rational human being when I\'m sober.  Took in some mini golf today with Snyder and his roommates.  I was beating Snydes by 2 strokes, but then I had some \"issues\" with a hole with a water hazard, and he ended up beating me by 2 strokes.  Lame.  But I beat him in the off road go carts which was fun as hell, so we end up even for the weekend.  Well I best get goin\' home.  Peace out blog readers.  '),(4468,3879020,'10,June,2004','2004-06-10 00:00:00','   Peace up my peeps.  I have just got done taking in the spendidness that was walleye weekend in Fondy.  The Snyder\'s lo9fty prediutions wretre far from true, but hey, trying to lure m,e anywhere with women is just silly.  I did however play an ugly chick with a nice rack at darts right before bar time, because I was running my mouth and I had to put up or shut up. I won by 6 marks, but she may have taken a dive because I have such a sweet ass.  Snyder is making us a Pizza.  He has the hookups large.  I cleaned out the passenger seat of my car , that is big news.  Hey snyder just made a hell of a wingman gesture to me  (even though we are at home now) so big up s to you.  I\'m hungry and tip[sy, so I\'m gonna eat andhit the hay.  Walleyes 4 life!  '),(4469,3879020,'10,June,2004','2004-06-10 00:00:00','   Oh BTW Monday and Tuesday of this week both tied the record for number of pageviews in a day... at 9!  Wohoo!  Peep the stats  urlLink here  if you like.  '),(4470,3879020,'09,June,2004','2004-06-09 00:00:00','   I was stunned to read  urlLink this news.   I wasn\'t stunned about what the article even says.  I was floored that someone actually thinks that this is news.  I mean anyone who has ever even  known  a female could have told you this.  5 days they hate you, 21 they don\'t really care either way, and the other two they are trying to rape you.  Every guy looks forward to those two days.  But I digress.  This post was supposed to be about something else..... ummmm maybe it is Fatty\'s recent run-in with all things electronic.  When he got his new mobile phone and was cut off 6 times during the inagural call that was actually pretty funny.  Fucking cell phones.  Now it is just down to me and Lou in the last man standing without a cell phone contest.  But I should prolly just go for it, because Lou is a welcher.  I have a better chance of seeing a hot chick in (This isn\'t) Oshkosh when college isn\'t in this weekend.  But on that that topic, Big ups to Snydes for being the first person to post a comment to the blog.  I\'d also like to give props to Tam and Josh as I see their ISPs pop up on the recent visitors list a lot.  Thanks guys.  AOL and Qwest, hard to miss, of course Fatty may have a different provider now that he is jocking the DSL.  Don\'t touch that modem Fatty!  You know I just came back to the original thing I was talking about, almost by accident.  Damn I\'m good.  Anyway, then Fatty\'s Xbox goes on the fritz, which was much less funny since mine is sitting on the server table at school in pieces.  Damn.  If it wasn\'t SOOOOOOOOO worth the effort to mod the damn XBox I\'d just let the poor thing alone.  But when it\'s working it\'s the coolest and most capable piece of electronics I own, and that includes my new stuff, my PC, and my PDA.  We figure it out tonight Fatty.  Talk nerdy to me.  But when I get an e-mail that looks like it\'s from fatty today with a virus attached, I gotta start thinking that it might just be you Fatty.  And now to wrap up with the mandatory politics,  urlLink I see  Kerry now has a 7 point lead on Bush.  A douchebag like Kerry normally would not be able to be leading a race for president of the Cuba City High School Student Council let alone president of the USA, but he is running against Dubya, and Ross Perot might even be able to beat Dubya.  Oh yeah, Regan died.... and I really don\'t care.  '),(4471,3879020,'09,June,2004','2004-06-09 00:00:00','   I, like most Americans, really only care about politics when I\'m really really pissed off.  Well, maybe I care a little more than average, but if things are good it doesn\'t really matter who is at the helm of this country.  I can\'t say things are good now.  I saw  urlLink this one  today and was stunned.  I mean honestly.  Checking people at random, making them present IDs?  That sounds exactly like what the Nazi\'s did to the Jews in the ghettos.  Can this practice really continue that long?  I mean the fourth ammendment really says this is exactly what you can not do.  And it\'s not foggy.  Here:    The right of the people to be secure in their persons, houses, papers, and effects, against unreasonable searches and seizures, shall not be violated, and no Warrants shall issue, but upon probable cause, supported by Oath or affirmation, and particularly describing the place to be searched, and the persons or things to be seized.     I mean unless they plan to issue like 50,000 warrants a day for no reason in Boston.  If searches like this make you feel safer, you are a dumbass.  If you are afraid to fly because of terrorists, you are a dumbass.  And if you are willing to give up your civil liberties, what makes this country great, for a false sense of security, you are the biggest dumbass of all.  You know I feel myself slipping into a tirade about the Patriot Act, Patriot 2, the DMCA, and all the other liberties I have seen stripped or proposed to be stripped away from us in the last few years, but I don\'t want to do that.  I just want to say keep your head up and be aware, because there is a lot of crazy shit going on there.  On a happier note, the water here at school is turned on again, and the toilets are flushing beautifly.  I was thinkin\' about going up to Snyder\'s this weekend or the next.  Whichever is not father\'s day.  I\'ve not been up there in so long I\'m willing to go see the country music event of the summer just to visit.  Country music, I think I\'d almost rather hear opera music.  You know what I think?  I think that this may have something to do with my alarm clock being accidentally tuned to a country station.  I\'m sleeping like a baby and all the sudden I get 106 decibels of \"Redneck Woman\" in my ear.  Grrrrrrr.  I mean who does not hate the sound of their alarm clock?  Oh well, It will be fun to \"Make the rounds\" with Snyder again.  Peace out.  '),(4472,3879020,'08,June,2004','2004-06-08 00:00:00','   Man do I have to go!  I drank a \"liter of cola\" this morning to wake me up, and man was that the wrong idea.  Now since they are working on the streets out in front of the CCHS they have shut off the water so they can replace the pipes or something.  It reminds me of the first time I met Josh\'s girlfiend\'s parents.  \"Open Up!  I gotta go and no foolin\'!\"  I\'m a dumbass.  Yeah probly only one person who reads this will get that inside joke, and that is a very good thing.  So nothing flushes anywhere around here.  It\'s horrid!  I may just sneak out behind the building or something.  It\'s times like this I\'m glad I\'m not a girl.    I finally got my DVDs all cataloged last night.  The  urlLink DVD player I use now  has a little database program that tells what disk is in what slot, and shows a little screenshot of the movie, and lists movie genre and stuff.  The player holds 400 disks, of which I have about 150 filled, so it was a bitch to find a particular title before.  \"Lets see Star Wars: Episode 1, Maybe disc number 58.... Damn!  Maybe 64?.....Damn!   Ok 76?... oh wait here we go.... nope, Damn!  Aw the hell with this!\"  But this player should solve my problem of having discs strewn all over the house.  What me not organized?  Yeah I know.  I bought a TV too.   urlLink This one  actually.  It is the same size and type as the one Tammy\'s parents have, if you can remember that one from New Years\' past.  So I figure I\'ve caught up with Tammy\'s parents in one aspect of my life, now if I could just catch up with the Temperlys.  Then I\'d really be a high roller.  You know that joke isn\'t as fun to tell if you can\'t hear the people around you instantaneously burst into laughter, but I must admit it was still quite gratifying.  I tried changing the way the blog looked today, all I really wanted to do is make this white column here a little wider.  But I quickly realized that I do not know XML at all.  It\'s painful when your nerdy inadequacies stare you directly in the face, not quite as bad as being depantsed in the winter, I mean that is real inadequacy, but not fun.  So it stays the same, which is prolly good because some of you are looking at this page on itty bity monitors, and wouldn\'t want this column any wider.  Oh well I\'d better get back to reading the excitement that is the owner\'s maual for Dell Optiplex series desktops.  Youch this job is lame.  '),(4473,3879020,'06,June,2004','2004-06-06 00:00:00','   Just because I\'ve been pondering styles of dress and whatnot lately I decided to throw  urlLink this one  in.  Thought It was good for a laugh.  I wish this freakin blogger had a spell check, cuz I seem to be having a lapse in the spelling area of my brain lately.  In case anyone cares other than the band members, it looks like the 2003 version of the Lawn Chair Debacle T-Shirts will be printed soon.  We made the shirt for last summer, but it was not printed because of our lack of screen printing facilities (because of Mom\'s changing jobs) and also because the design this year is so intricate we decided it would prolly be best to have them professionally done.  I did not go over the tour dates recently, but I remember there will be big shows at the Mississippi River, and also at the Hazel Green Sewage Treatment Facility, which I really think will be a big show.  The tour dates were trimmed back from the 2002 tour, because we had way to many girls chasing after us after our extensive touring scedule, and also because the shirts were kinda hard to read.  But this tour is off the hook.  Fatty or Andy can verify, but we cut this final list from about 60 legit funny ideas, so we trimmed the fat and left you, our fans, with nothing but the solid good shit.  The reason I mention this is two fold.  One because I thought that some of you might want one of these super swell high fashion t-shirts, and two, to get you people using the comment area.  Temp told me this weekend he was on the verge of commenting, so jump on it guys!  If you\'d like a t-shirt, comment with size and number of shirts and I\'ll hook you up.  They will be either yellow or gray (not ash, I hate ash tshirts, but gray) and we will get them to you at our cost and not a penny more.  Show your support for the LCD today.  In other news, I had a pretty swell weekend, most of which was spent \"sleeping it off.\"  But I always have a good time going out with FT (or TT or MST or whatever.)  And it was cool to see the Madison \"city slickers\" that were in our group this weekend.  My only gripe is if you are going to bring people home, bring home people without boyfriends.  Jeesh, what is it with couples hanging out with other people in relationships these days.  Also because of some off the cuff sing-a-long-ing on the way home with Drewsy I\'m on a Jimmy Buffet kick.  Not bad at all.  Well I\'m running long, so thx for reading, and have a kick ass week.  '),(4474,3879020,'04,June,2004','2004-06-04 00:00:00','   I just had to say something about the 15th anniversary of the Tiananmen Square protests/massacre.  I\'ve looked at the picture of the guy standing in front of the tanks hundreds of times over the years, in fact when I say \"Tiananmen Square\" it is the image that will pop into most peoples heads.  If an image pops in at all.  I was kind of suprised in talking to people around school today that some had no clue what Tiananmen Square was all about, and this is something that is huge news that happened during our lifetime.  I\'d devote at least 5 minutes and a quick Google search if you don\'t know anything, or even much about it.  But I look at the photo which has to be one of the greatest and most expressive of the last century, and it just inspires the hell out of me.  The lone guy, nothing but two bags of groceries in his arms, standing there balls to the wall staring down three tanks, because of a belief.  Someone who is willing to die for their cause cannot be defeated.  It\'s the struggle that makes the eventual success great.  I don\'t really know if I have an ideology that I\'d be willing to die for, but in this country we already have it so good, there isn\'t the urgency like in China.  In any event though I gotta give props to the dude, who is to this day unidentified, read more about that  urlLink here.   If I don\'t check in on this busy weekend, have a good one!  '),(4475,3879020,'03,June,2004','2004-06-03 00:00:00','   urlLink This one  made me laugh, well not the actual article so much as when I started imagining a streetball style spelling bee.  Rambunctious?  (African American) please!  Yeah it goes like \'dis R-A-M-B-U-N-C-T-I-O-U-S  Yeah, bitch!  What\'s up?  I before E except after C mother (frogger!)  It of course would be voiced over like that because it would be on regular TV.  Well, I thought it was funny at least.  '),(4476,3879020,'03,June,2004','2004-06-03 00:00:00','   Sometimes when I\'m walking around the halls at work here, especially now close to summer, I think back to the things that I thought were super racy on girls when I was in high school.  And I realize just how far skanky clothing has come in the past few years.  I\'d be totally fucked in this situation.  I had a kid who I was talking to the other day, and I asked him why he had all kinds of homework left the hour after his study hall.  (Normally this kid has stuff done by the time he sees me.)  Anyway, the kid is like, \"Well to be honest, I spent all hour staring at (Name omitted)\'s thong sticking out of her pants.\"  My time in HS was full of time wasted/lost to sexual fantasy, I can only fathom what kind of sensory overload I\'d go into if you shipped \"1994 Jeff\" ahead ten years in time to today.  My head would explode!  And some would say, \"Those girls today, they are trashier than when we were in school.\", etc.  (And girls would say that, cuz they are catty)  But you know what?  The same amount of people are getting laid as any other time in the past I would guess, nobody is trashier now than before, just the clothes are.  I believe it was Dave Chappelle who said (or something to this effect) \"I have had women come up to me and say, \"Just because I dress like this, it does not make me a whore!\"  And that\'s right, just because a person dresses a certain way, does not mean that they are a certain way, and we have to respect that.  But ladies, you got to understand, that is fuckin\' confusing.  You may not be a whore, but you\'re wearing a whore\'s uniform.\"  Oh Chappelle you funny bastard.  Oh wait, check  urlLink this article .  It\'s why I went on this tirade in the first place.  Big ups little girl!  On the topic of Chappelle, anyone interested should check out the new standup DVD of his I came across.  It is great.  Hey, come check it out on my new TV.  It\'s times like these I wish I lived in Madison, cuz we could have movie night.  Not much demand for an event like that here.  Frickin job in the sticks!  Oh well, it has it\'s perks, I get to stare at high school girls thongs all day long! (  '),(4477,3879020,'29,July,2004','2004-07-29 00:00:00','   urlLink It may seem wierd , but come on, you can\'t turn down a triple dog dare.  These dudes would have lost all their playground cred.  '),(4478,3879020,'28,July,2004','2004-07-28 00:00:00','   Hola, all.  I just wanted to throw out some quick info today as I had some stuff I had planned on writing about this afternoon, but it didn\'t really fit with this stuff I just found this morning.  First off a funny story, or so  not  funny that it  was  funny I guess.  I was sitting at home watching the convention (Democratic National Convention in case you live under a rock) and marveling at the only 1 hour of coverage they have decided to give to the conventions this year.  One hour for four days.  Four hours total.  12 on PBS where I watch, but that means we devote 30 times the TV coverage to choosing an American Idol as we do an American President.  Yes, even in an election where there is actually a difference between the candidates, nobody still gives a fuck.  Sad shit, but the funny was when my dad is like dissing some comments made about Kerry.  Just picking a fight it seemed, begging me to say something.  But I remained silent.  Because the last time we discussed politics, by the end I was really letting the expletives fly, and we didn\'t talk to each other for like 4 days.  I really don\'t swear around my parents that much, but that day it was muthafukin\' zero hour, and I just flew off the handle.  But back to last night, this went on for a bit until dad is like \"He won\'t even get a third of the vote, Bush is going to win by a landslide.\"  And that was about all I could take, so I\'m like, \"Have you lost your fucking mind?  Victory by a landslide?  Kerry is up by two points now, and so called \"undecideds\" end up voting overwhelmingly democrat.  For Bush to win by a landslide, everyone in the middle east would have to make up, hold hands, and sing \"Kumbaya,\" Gas would have to drop to about 98 cents a gallon, and Michael Moore would have to make another documentary that shows Bush playing with baby bunnies or something.\"  I think Bush\'s campaign should hire my dad, because nobody apparently has a more rosy view of Bush\'s chances in this election than my dad.  I\'m not saying it\'s good or right, but people are too high strung and divided these days.  If Bush gets re-elected somebody is going to go Lee Harvey Oswald on him.  But I don\'t think he\'ll be re-elected so I don\'t think we\'ll have to worry about this.  On a much more humorous news report, check out  urlLink this sweet little mama  I found.  If news on the web thinks it can hide from me, I laugh heartily.  I just enjoyed the picture so much, because it reminded me of the camo chair.  Black sweater, black background = Kim\'s floating head.  She\'s incognito.  :)  Talk to you later hopefully.  '),(4479,3879020,'26,July,2004','2004-07-26 00:00:00','   So the sixth annual camporee was this weekend.  The big story of the weekend, but stuff happened while we were gone I figured I should touch on first.  First off, Lance Armstrong won number six, which is incredible.   urlLink This one  kind of compares the feat with other great sport streaks.  I\'d certainly pick him in my list of the top 5 athletes of all time.  Cheers to you Lance!  But the biggest news of the week or the month (possibly the year) is Steven Hawking\'s comments on black holes.  Hawking, although I\'m sure many of you have heard of him already, is the number one guy in quantum/astro physics.   urlLink What Hawking is suggesting  (and I tried to find an article that explains it as simply as possible) really changes a big part of what we had held as truth in the field of relativity.  I\'ve been looking to find some physics junkie type people ever since to chat about it.  Fatty called me at like 10:30 the night he drove home, he\'s like \"Turn on NPR, there\'s a show on about black holes!\"  I love these calls, half because Fatty is listening to NPR and half because my friends know my holy trinity of nerdiness (black holes, giant squids, tidal waves) and will call and tell me when a show is on about them.  I still hear from Snyder about the squid spermatozoa.  I may write more on this when I have a better handle on the new info.  Ok, so camping now.  Good site, Good weather, OK food, OK time.  I give it two and a half stars.  With no disrespect to Joe and Kreul I did miss having  my  guys there on Saturday, but it was probably about as interesting as it could have possibly been.  For about 45 mins. to an hour on Saturday, we actually carried on what I would consider a rational adult discussion, which was amazing but nice.  You see I say I love to hear the big fights we get in, that is really not true.  I say I enjoy them because sometimes it happens when we come to an impasse on something we were discussing.  Other times it\'s just someone \"cheating\" at Outburst or Sequence, which I can live without.  I really find that most people do not want to have to defend an opinion in discussion, some don\'t even form opinions on important topics, that is why I was happy to see Saturday.  The conversation, on religion no less, never broke down to just yelling and shouting the \"F-Word\" at each other.  Or did it?  I did hear a stream of obscenities flowing from our campsite all the way over at the bathroom.  :)  Very relaxing all in all.  Tammy and I left at like 4:30 Sunday afternoon, right before my head exploded in a giant allergy ball.  I believe a fun time was had by all.  We should start on those lucky 7th anniversary t-shirts right now you know.  Then, just maybe they\'ll be ready on time for next year.  '),(4480,3879020,'20,July,2004','2004-07-20 00:00:00','   Oh Fuck, here we go again!  Can we get this nutjob out before the world is nothing but a blob of nuclear sludge.   urlLink Check it!   I guess the nuclear bit doesn\'t make sense without  urlLink this one.   In-fucking-sane!  and Yippee & hurray!  urlLink Check dat.  '),(4481,3879020,'19,July,2004','2004-07-19 00:00:00','   Not that Michael Moore needs to score more points with me, but I read  urlLink this one  today and about choked on my Twizzler. Of all of the many rants that Moore has gone on in years past this is by far my favorite one, because it deals with peer to peer and digital rights. Honestly in the last election, handling of digital copyrights was about 25% of my decision for who I voted for for both president and governor. I\'d say that\'s borderline fanatic about the issue. This election it will be the last thing on my mind, because I\'ve come to see that it really doesn\'t matter who is in power between the two major parties, no improvements will ever be made in this arena by the government. Unless a Libertarian or something were to be elected president. But that will never happen. That is about as likely to happen as a Tupac Shakur/Notorious B.I.G. ticket winning the election in 2008. But in any event (I got  way  off track there) I\'m glad to see Moore giving his blessing to the people like me, who paid for a ticket (and probly will see it again), plan to buy the DVD when it\'s released, and are only looking for one of these \"illegal\" downloads to tide us over till the DVD release. Even Moore\'s movie house Lion\'s Gate seems not to have a problem with it. So if anyone is interested, I may or may not have a copy...  urlLink  This is funny    urlLink This is funny    urlLink This might come in handy this weekend    urlLink and This is just getting stupid  '),(4482,3879020,'16,July,2004','2004-07-16 00:00:00','   After talking for a full paragraph on the new spell checker, I failed to actually run it on that last post, and it\'s riddled with errors.  I\'m a stupid dumbass.  '),(4483,3879020,'13,July,2004','2004-07-13 00:00:00','   What be the lowdown my G\'s? I gots ta say that here on the backend, the blogspot just got a lot dope-er. I mean if I really cared enough to edit the HTML/XML before, I could have done all the cool shit I can do now, but now all kinds of options only a button push away. Thankfully they added a spell checker so I can stop composing in Word and then pasting into the blog post application. But you can Change  fonts  now, change  c  o  l  o  r  s , change alignments and stuff and even an add pictures button it looks like.  I\'ll have to test that out in some later post.  I\'m not really one for bling bling posts, or even bling bling e-mail for that matter.  I\'m the one who turns off the rich text so I can just type an e-mail that isn\'t in stupid Times New Roman or some other dumb thing.  So I would not expect any visual changes on the site other than better spelling maybe.  So it\'s Friday huh?  \"It\'s Friday, you ain\'t got no job, and you aint got shit ta do...\"  If you haven\'t seen Friday with Ice Cube that was just a really stupid sentance.  I do in fact have a job, but I don\'t have shit to do.  Temp said something about comin\' home to do some Buncome fishin\' but I have not heard anything more since.  I\'m not sure if i wrote anout this but I went fishing in Buncome the other evening and got raped by the gnats, or some other little bitey ass bug.  If I had a parrot or something else that would eat those mothers, I woulda been having a great time.  After an hour and a half or so (no bites but the water is pretty high and fast right now) and the trip was cut short.  Why?  Either a broken or slipped fucking knot.  Can you believe it?  Me, the one who held the knot tieing record for Troop 86 and 122, and the second fastest time ever at Canyon Camp over the ropes course (Boyscout ropes course is eight knots:  urlLink bowline,   urlLink sheet bend,   urlLink square knot,   urlLink two half hitches,   urlLink sheepshank,   urlLink timber hitch,   urlLink clove hitch,  and  urlLink taut-line hitch ) forgot the second loopthrough in  urlLink my fisherman\'s knot.   Mother of pearl!  Now I need a new chartruse single spinner.  Oh well it\'s better than fucking up on a  urlLink Rappala knot  and losing a five or six dollar lure.  But homo nonetheless.  So yeah if anything else is going on let me know, cuz I\'m tapped.  Maybe I\'ll call Andy.  Or go with Kim to F911 since no random women from Madison have contacted me lately wanting to see that film.  Hope all of you have something fantabulous planned for the weekend, and all goes well.  '),(4484,3879020,'13,July,2004','2004-07-13 00:00:00','   Well it\'s my wrist actually not my elbow, but perhaps my mom was right for once.  I\'m not sure I ever posted about getting a skateboard on here, but I did.  I have always kind of wanted one.  Not because of Tony Hawk, or Bucky Lasek, but because of  urlLink Julie  from Real World: New Orleans.  She was the Mormon from Wisconsin, who incidentally rode a skateboard.  I\'m not sure if they showed her on her board more than once, but I do remember that that is what made me want a skateboard.  Wierd.  She seemed ok on the show, she wasn\'t a huge skank, but I really gained respect for her after reading a bit into her after (like a year after) the show was over.  A girl who understands the beauty of one of the top 5 films of all time, Fight Club, and can identify with it\'s message as evidenced by  urlLink this,  is cool with me.  Anyway she was skateboarding through the airport and I thought to myself \"Man, I really want a skateboard.\"  So anyway my dream of skateboard ownership recently became reality, but I did not get out riding till yesterday.  Mom was like \"You are going to wear pads and a helmet right?\"  What can you say to that...?  I\'m sure the typical kid response would have worked, \"But mom I don\'t want to look like a dork!\"  but I chose the much more truthful and less traditional \"Fuck that, I paid to much for the board, I\'m not buying all kinds of accessories too!\"  So I am looking like a total fucking dork even without the pads up in Southwestern\'s school parking lot practicing standing up while pushing the skateboard, and then kind of slowly rolling down the almost nonexistent incline and trying to pactice turning.  I can not yet push the skateboard with my one foot and get going and then transition into getting on with both my feet.  In short: I really suck so far.  But I can remember a time in my life when I couldn\'t ride a bike, and as it ended up I could do that pretty well once I could stay up without crashing.  So I had the first of what will probably be many skateboard crashes.  I was kinda just standing on the board trying to practice balancing, and I leaned forward to try to get rolling.  I went forward, my skateboard went backwards at an incredible rate, and my wrist stopped my fall.  For a second I thought my Mom may have been right about the pads.  Those thoughts were quickly replaced by thoughts like OWWWWWWWWWWWWWWWWW!  I stand by my no pads position though.  I only hurt my LEFT wrist, so I hardly notice.  :)  In any event I just wanted to give congrats to Tammy on her new temp to own (hopefully) job.  With that I\'m done for today.  Pray for my stupid ass\'s saftey.  '),(4485,3879020,'12,July,2004','2004-07-12 00:00:00','   I decided to throw in that slightly homo reference as an allusion to something later in this rundown of my weekend, so stay on the edge of your seats for that one.  I ended up going up to Madison to visit Candy and Ryan and take in some of the ESPN Great Outdoor Games.  If you donâ€™t know the website for the GOG itâ€™s  urlLink here.   (that was kind of an inside joke, but informative I hope)  Anyway I got to see the team timber relays which was pretty exciting as it included pole climbing, chain sawing, axe chopping, and other various talents into one big relay event.  I canâ€™t recall who actually won gold, but the gray teamâ€™s saw girl was the best looking, and so I give them the gold in my mind.  Sunday was exactly what a Sunday should be, some laying around while waiting out the rain, watching nature programs on PBS while switching between a NASCAR race, and Tour de France coverage.  Then after the rains came, Ryan and I joined Roach on his spiffy new boat and attempted to trick fish into biting our rubber lures.  No fish fell for my delightful presentation of several plastic and steel baits.  No fish fell for Temperlyâ€™s presentation of various plastic and steel baits.  Roach caught three white bass, because I think he cheats somehow.  I did however buy a new fishing rod (those reading only to see where the â€œrod thingâ€ would come up can stop right here) and reel.  Itâ€™s a  urlLink Fenwick Eagle series 6.5 ft rod,  and a  urlLink Shimano Sedona 2500FB Front Drag reel.   Temperly didnâ€™t call me an idiot after purchasing either, so they must be at least ok.  We revisited the infamous â€œJeffâ€¦..Jeff Langeâ€ Arbyâ€™s on our little shopping trip and Iâ€™m pleased to report that there were no incidents this time.  I did not spot anyone with a Tec-9 charm on their chain either so maybe theyâ€™ve got some better help at that franchise.  Being out on â€œThe 10,000 acre lakeâ€ just before dark was beautiful and when all was said and done I wasnâ€™t to upset about getting skunked fish-wise.  So now back to another crappy work week.  Gotta make the ends I guess.  Oh, btw it sounds like, in a new development in my continuing quest for an apartment, I may be living with Cindy for the next 11 months.  Iâ€™ll keep you posted on that one.  Big ups and peace out.  '),(4486,3879020,'09,July,2004','2004-07-09 00:00:00','   Anybody remember that root beer commercial with the guy who is like \"I can\'t wait to work for you Mr. Dum(b)ass...  I\'m Dumass material... So what can I do for you Mr. Dumass\"  and the old dude is like \"The name is Du-Maas...\"   Remember that?  That commercial was really... really.... funny.  '),(4487,3879020,'08,July,2004','2004-07-08 00:00:00','   Fresh off another ass raping at the hands of violent inner-city youths Fatty (aka. Soars with Eagle, Kicking Horse, and Smells Like Turd) has figured out how to work his PC and accept my blog team invite.  As the only person who responded to my open call for co-bloggers, Fatty was subjected to a rigirous screening process that involved several full cavity searches.  Even though many strange items  were  found in his various body cavities, Fatty has been accepted to the Blog sidekick training program.  He has the same authority in the blog as I do which means he can add posts and/or delete posts if a blog viewer were to get unruly and start oh I don\'t know, knocking my ravishing good looks, lol.  Yes, I\'m looking in your direction Cripes!  I\'ve yet to hear any weekend plans either, Snydes are you comin\' home?  I\'m not going to sit home for a second weekend in a row.  I\'ll go hang with FT if I have to.  Speaking of FT, I was driving home via a slightly different path than I normally drive last night trying to search for unsecured wireless networks in Hazel Green. (there are 8, which is all of them, in case neone was wondering.)  And I see all kinds of digging going on over at the \"Old Temperly Ranch.\"  And since HG is the nosiest fucking place on the face of the earth, I decided to swing in and ask Brooke what was going down.  I figured it was something to do with the city and FT\'s flowage that he\'s always bitching about.  No, apparently some old people are building a house there.  Such a typical FT move.  Sell some land, pay off your new truck, get some oldies to build a house so you don\'t have to look at the city flowage running through your property anymore.  If only we could all live the FT lifesyle.  Ahhhh... well that about wraps it up.  Once again my big welcomes to a new blogger, hope your time spent here is very key.  And once again an open call if you have an opinion and you\'d like to express it on this blog, join the team.  I\'m really very gentle when cavity searching...  ;^p  '),(4488,3879020,'08,July,2004','2004-07-08 00:00:00','   Alright meow, this is my first official Blog post.  I feel honored that Jeff decided that the things I say may be relevant enough to grace the screens of his blog\'s readers...I am truly touched.  I do have to say off the bat that I plan to post very unapologetically.  It\'s gonna be like listening to a movie\'s commentary here (that is if anyone besides Jeff or I do that) my views aren\'t necessarily those of the owner of this blog.  I plan on posting my sincere feelings and opinions on many different subjects in a multitude of areas.  Chances are, everyone that reads the posts won\'t agree with what I have to say (including Jeff).  I just want to post in the manner that I think Jeff wants this Blog to stand for...a place to view the world through the eyes of someone else, even if just in small doses.  I am just hopeful that what you see through my eyes is interesting enough that you return for more.  In saying this, I would like to leave you with  urlLink this , to see if you are just that little bit \"off\" enough to truly walk away from this experience undaunted.  '),(4489,3879020,'08,July,2004','2004-07-08 00:00:00','   Articles like  urlLink this one  piss me off.  You see the same thing about the music industry, how they are losing billions to music pirates.  Wake up people!  For the most part, you aren\'t losing dick.  Just because somebody pirates some software or music does not mean that they would have actually bought it were they not able to get it for free.  I have some songs by the Backstreet Boys and the Dixie Chicks, does this mean they lost sales revenue because I didn\'t buy those albums.  Hell no, because there isn\'t a chance in a million that I would have ever actually purchased those albums to get the songs I wanted to hear.  I would have learned to live without them.  I heard that William Hung (the uber crappy American Idol\'s) rendition of Ricky Martin\'s \"She Bangs\" is one of the most popular downloads on the internet lately.  You know why that is?  Curiosity.  Everyone wants to hear that train wreck of a song, but nobody is dumb enough to buy the CD.  If you think those people who downloaded that crappy song would have actually gone out and purchased the CD had they  not  been able to download it, you\'re in a fucking dream world.  Maybe one out of every ten or twenty thousand people if they are lucky, and if you divide the number that these industries say they are losing by 20,000 they really don\'t look that bad.  '),(4490,3879020,'07,July,2004','2004-07-07 00:00:00','   It\'s sports day here at the blog spot.  I choose to highlight two sports that are at the top of my mind at the moment.  The first is of course  urlLink The Tour de France.   It\'s July, and for me that means the tour, just as March is college basketball\'s month.  I had been meaning to mention it for a bit, but just had to today as Lance Armstrong is now again in the yellow leaders jersey.  Usually I\'m not a big fan of the type of cyclist who wins the tour.  They are usually the distance runner types who weigh as close to nothing as possible.  Even in my cycling days I was not this kind of rider, and now... well yeah... I\'m not as close to nothing as I could be, we\'ll just say.  I was always partial to the big sprinters like  urlLink Mario Cipollini  and  urlLink Marty Nothstein,  because this is how I rode.  These are the guys that are all about speed and fast finishes, really the only thing exciting about cycling for the layperson.  Lance is really kind of an enigma, because he was one of these big power guys back in the day before he got cancer, but now he\'s kind of turned into one of the distance guys.  Oh well.  It will of course be another hotly contested tour, and I plan to cheer for Lance again this year, but this year is it.  This year if he should go on to win will be his 6th in a row, the most tour wins in a row ever, and the most overall ever.  He\'ll be beating out tour legend Miguel Indurain (who won 5 in a row in the 90\'s), and cycling\'s Michael Jordan, Eddy Merckx (who won 5 spread out in the 70\'s.)  Next year I\'ll be cheering for Tyler Hamilton.  He was one of the youth counselors when I was at cycling camp in \'93 and he was the counselor that stayed in our room.  So I know him pretty well.  We still shoot an e-mail each other\'s way a couple of times a year.  But this year he\'s in 8th right now, and in future tours I\'ll be cheering for him to win it all, especially since Lance has now kind of joined the \"glitterati\" by dating Cheryl Crow.  I know you guys will prolly not be watching TDF coverage on the Outdoor Life Network or anything, but I just thought I\'d let you know it was going on.  Moving on to a much more \"bling bling\" type of sport, my second one is street hoops.  I got to look at the And 1 Mixtapes and they are amazing.  The things these inner-city kids can do with a basketball are fucking amazing.  The best footage in the tapes comes from the Rucker Park Tournament in New York.  \"The Ruck\" as it\'s called is basically street hoops\' Mecca. They have tournaments there, and it\'s common for NBA stars to go there to play against the street guys to try and gain some street cred.  And it\'s funny as shit to see these millionaires get worked by the playground legends.  It\'s really my favorite kind of basketball.  If you can get your hands on the And 1 Mixtapes do so, because some of the stuff is just amazing.  A lot more amazing than Fatty bouncing a basketball off my face while trying to show me his new moves he picked up at Ethan Allen.    Hey before I go, is anything exciting going on this weekend?  '),(4491,3879020,'06,July,2004','2004-07-06 00:00:00','   I heard about  urlLink this  on NPR this morning, and people were calling in very pissed off from both sides of the issue.  Sooth or slander?  Your call.  '),(4492,3879020,'06,July,2004','2004-07-06 00:00:00','   How\'s it going blog people?  I have been remiss in posting for a bit, but hopefully everyone got a well needed break from my rants and the monotonies of my life.  In one of his comedy sketches Chris Rock is talking about having conversations with his wife/girlfriend, and he says â€œThere comes a point in a relationship where youâ€™ve heard every fucking thing this person has to say, and your thinking, go out and get kidnapped or something, get some new stories.â€  While I didnâ€™t go out and get kidnapped, I did have an odd Thursday, and I was thinking of this line the whole time.  I use a program called Skype, which is kinda like what AOL IM was only with voice chatting.  Only itâ€™s not like voice chatting which has been around forever, itâ€™s actually Voice over IP which is what big companies use to route phone calls over the internet instead of the telephone lines to save money.  Itâ€™s a cool little application.  I use it to talk to people I play games online with in Alaska and South Dakota, but I digress.  I just wanted you to know what the hell I was talking about.  So I usually get about 2-3 â€œcallsâ€ per day, usually from like Uzbekistan or Denmark or someplace weird like that, people who are setting up the application and trying to test it out by calling overseas.  I never really answer those calls.  So Thursday I get a call from Madison, WI from somebody named Bonnie, and I figure maybe this is the internet moniker of somebody I know.  So I send a text message that I donâ€™t have my headset hooked up and through text messages I realize Iâ€™ve never met this person before.  This is noon.  To RAPIDLY fast forward this story for brevity, by 3:30 I have decided that it would be a good idea to go to Madison to re-see Fahrenheit 9/11 (her idea not mine,) by 5 we are having dinner at a State Street cafÃ©, by 8 Iâ€™ve realized the movie is sold out at all theaters for the rest of the night, by 10 by some odd twist I am assisting in the purchase of (ahem) marital aids with her and her friend, and by 11:30 I was on my way to Fattyâ€™s to try to sort out a very odd night.  Itâ€™s the stuff that I left out there that I think make the story so strange.  But it is not something I really do normally.  Iâ€™m not really the person that hauls off and meets internet wierdos (myself included.)  There are people Iâ€™ve known for eight or more years on the web who Iâ€™d never dream of actually meeting.  But it was kind of a fun thing to do â€œjust for the hell of itâ€ for some life experience, and Bonnie was cool.  Btw besides the vibrators, no weird sexual shenanigans took place here.  Iâ€™m not  that  dude.  Iâ€™m just glad the â€œJeffymobileâ€ wasnâ€™t stolen, and I wasnâ€™t gang raped by burly men.  Itâ€™s not something I plan to do again on a regular basis, but a very interesting Thursday indeed.  I did get to work more with Fatty on the Summer CD when I was up there, and itâ€™s sounding really good so far.  The weekend was kind of gay though as nobody was really doing anything that I heard of.  I missed the Rhythm and Booms fireworks in Madison for the first time in 10 years, which is really quite a shame as my traditions are dieing around me.  They did get rained out Sunday though so it wouldnâ€™t have been the best year to be there.  So now camping trip my only remaining yearly tradition will have to be extra good.  Iâ€™m crossing my fingers as Iâ€™ve just not had that much fun the past few years.  But hopefully this is the year that restores the good name of the camping trip.  So as not to make each post a mile long Iâ€™ll wrap this mother up now.  Have a good day.  '),(4493,3879020,'13,August,2004','2004-08-13 00:00:00','   I was reading  urlLink this  news story today and thinking about the butt raping we will all be taking in a few months when the over 25% increase in crude oil trickles down to gasoline. Now of course this looks bad for Bush, and the more pissed people are about gas prices in November, the better for his removal. So bring on 2.50 a gallon. But seriously my thoughts quickly wandered to the extreme changes in gas prices over the past 10 years or so. On our class trip one of the gas stations we drove past in western Nebraska had gas for .69, when I was 5 and my parents and I went to South Dakota we bought gas for a quarter at some ghost town gas station. Dad in his usual cynical way predicted that it was probably half water. :) But cheap gas played a big part of my life. In high school I was for the most part flat broke. But three bucks would buy a couple of hours of driving time around Buncome, listening to CDs, shooting the breeze, trying to find Scales Mound. My hand on the wheel, Cherry Skoal in my lip, Temperly knocking back a few penicillan flavored beers from his old man\'s fridge in the barn, windows down, the wind in my... stubble. That was independence, that was good times. Would we have done this if we had to pay twice as much or more for gas? Probably, but not much. The joyride has been a hallmark of high school males since there were cars around to drive. Whatever kids are doing these days I\'m sure it is not as cool as driving around Buncome, and I lament that loss. It\'s six years later and I already have one of those \"When I was a kid such and such only cost a dollar.\" I think that\'s kind of funny.  Oh BTW sorry for the sparse posts this week. I usually blog on my lunch break, and I\'ve been home feeding the kittens at noon all this week. As soon as they can eat solid foods I\'ll be back to normal pace.  '),(4494,3879020,'09,August,2004','2004-08-09 00:00:00','   Well I just wanted to give some blogspace and a proper eulogy to my cat Hemi who died Friday of car tire related complications.  Hemi was 16 months old and is survived by five two week old kittens.    I officially claimed Hemi as my pet early last summer.  She had been hanging around our house for about a week, and following dad around as he was beginning to build the retaining wall behind our new garage.  He preliminarily named her â€œPeskyâ€ because he considered her quite a nuisance.  Well, one night after Andy and I got home from the bars a bit loopy, and Hemi was there to greet us, Andy suggested that a pet might be good for me.  I agreed and decided to name my cat Hemi after the line of engines used in Mopar vehicles.  I felt it was a fitting name as this cat would get to purring almost before you even pet her.  Well that and those Dodge commercials are hella funny.  I always contested that Hemi was less of a pet and more of a utilitarian animal.  From the first few days that I had her I would play chase the ball of string with her.  We worked up from golf ball to tennis ball sized â€œmice,â€ and eventually worked up to larger targets.  Iâ€™ve owned cats before and some will not chase something even half their size, but Hemi eventually would go after a football sized target, which was really rare but good, as she on occasion would kill a rabbit that seem to love to eat baby trees.  She was good at what she did.  But as much as I said that she was there strictly for rodent control, I realize now that I really grew attached to her as a pet.  Some say that all cats want from their humans is food, and to an extent that is true, but it always made me feel good when Hemi would greet me outside my car door after work, even if her dish of food was full.  It was nice to have someone that was happy to see me home every day, and I miss it already.  Nobody comes running across the yard now when my car rumbles down the driveway, and although Iâ€™m eye dropper feeding her babies, I doubt I will keep any, as I canâ€™t take the emotion of losing more pets.  Hemi was buried Friday afternoon at the foot of the retaining wall where I first met her, I thought it was fitting that she would be as close as possible to the garage that was basically built during her lifetime.  Big ups, and peace out Hemi.  Dad took over the kitten feeding duties this weekend for a day and I got to visit Fatty.  We went to Great America in the late afternoon and got to ride 3 rides.  Superman, which was key but short, coulda used like 1500 more feet of track, Vertical Velocity (or V2 he he) which was pretty dope except for the complete stop on the second time up the back portion of the coaster.  Itâ€™s a jolt that really just seems like something that is designed to knock off glasses and Birkenstock sandals, and didnâ€™t really add to the ride I felt.  And then Batman, which is just a classic, and is about the perfect length and speed.  Fun times.  The night ended nicely enjoying some cigars and the nighttime air, and watching some of the Bowling for Columbine: Disc 2 content.  Good times.  Oh yeah and e-mail me if you are having trouble posting messages, or replies or whatever, because I am and I was wondering if it was just me.  '),(4495,3879020,'06,August,2004','2004-08-06 00:00:00','   I went last night to Fahrenheit 9/11 for the third time with Kim, Candy, and Jenny B. That would set the new record for most times I\'ve seen a single movie in the theater during it\'s original run. I\'ve seen one other movie 3 times but it was twice on the original release, and then once in like a re-release theatre like the Rivoli, or something. That was Home Alone. But I\'m glad to see this record fall, because although Home Alone was a hit in it\'s day, and still is one of my faves, a comedy other than Dumb and Dumber didn\'t deserve the top spot. But it was an odd showing in a really crappy theater. Obviously much smaller crowds than the first two times I saw it (Opening day in Madison, and a week later when it opened in Platteville) which really changes the feel of the show, and in a very antiquated theater that I could not get over the whole showing. I\'ll never see it with better crowd involvement and interaction then in the art house type theatre I saw it in on opening day in Madison. This was a sellout show, at noon on a weekday, and obviously comprised of people who cared enough to take a half day or something like I did to be there. Cheers and boos came from the crowd at certain parts, and there was genuine emotion. With maybe 30 people in the theatre in Dubuque the other night that just wasn\'t there, but these are the people who care less. If you haven\'t seen the movie by now, it pretty much means you fit into one of a few groups:             People who don\'t want to see the film at all, which I don\'t understand because if Fox News or something put out a documentary tomorrow showing all Bush\'s good deeds I\'d be there just to have both sides of the story. I mean how long could a movie like that be? Six maybe seven minutes? I could take that.        People who are interested mostly because of the buzz and less because of the politics, which I also don\'t understand, because I\'m amazed that anyone can\'t have a decision made by now. But hey, that\'s what we call a swing voter I guess.        And lastly people who have been living in a hut in Africa for the last 12 months. Which I think is pretty self explanatory.         But anyway the theatre sucked. The screen was too small and was cutting off the edges of the picture so you couldn\'t see the names and titles of some people, which is really pretty important. And the sound was stereo at best. There were no speakers behind us. You need to be enveloped to truly appreciate the film I think, which is why I\'m inviting anyone who wants to come over to my apartment (wherever that turns out to be) the day after the DVD release and see the film as it was meant to be seen. In crystal clear high definition on a Trinitron tubed screen with Dolby DTS-ES/THX Select audio. I\'ll let you sit as close to the screen as you want, but as you know from your mom \"If you sit to close to the screen you\'ll go blind.\" Nobody really was just balling their eyes out though last night, and I\'m not sure if that was because of the kind of people who made up the crowd, or because the cruddy theater made me feel like I was watching a movie instead of just standing right there. I still can\'t make it through dry, but it\'s less of a sad tears thing, and more of a seeing red from anger eyes watering thing. I can remember very few things I\'m more passionate about than the removal of Bush and his ilk from the white house, which is why it was odd to have Jenny along last night. I didn\'t want to just flip out, which I\'m sometimes known to do when I\'m talking about something I\'m really passionate about. Kim and Candy have both seen me get all worked up about something and go on a rant, but I didn\'t want to look like an idiot in front of Jenny who I doubt ever has seen that. Plus I just get a genuine feeling of intelligence and knowledge from her. I mean for someone who has been in central Africa she\'s better informed on both topics and issues than many people I know here in the states. And I get the feeling like if I were to actually get into a debate about something we differed on that I could lose, which I think is really cool and I don\'t know many people like that. But the discussion is what I love. I mean people like Ryan or Josh are happy to throw down on just about any given topic, and have an opinion, which I dig. It\'s probly why they are my friends. But wholehearted discussion can\'t really be forced, it has to just happen, and a movie like this, like any other movie with a message (that means something  not  like Spiderman II, White Chicks, or the Terminator. :) can be just the stimulus some people need to begin the dialog. I mean you can\'t walk out and not at least have an opinion of the film. And I like that. I hope everyone has a good weekend doing weekend stuff, or as much weekend stuff as you can get done this weekend. Wow I accidentally put the word weekend 4 times in that last sentence. The Jeff Lange weather station predicts rain this weekend. Hopefully it holds out long enough for Fatty and I to go to Great America, and long enough for Temperly and I to get some fishing in. Be well everyone and peace out!  '),(4496,3879020,'05,August,2004','2004-08-05 00:00:00','   urlLink Ha Ha Ha Ha....... Ha!  '),(4497,3879020,'04,August,2004','2004-08-04 00:00:00','   urlLink 417.5 billion dollars for defense spending?   Yeah that sounds like a good amount for that.  Let me see I\'ll just pay upfront for that right now.  Oh wait, that\'s right we are  already  running a massive deficit.  Do you take Master Card, can I write you a check?  Don\'t worry Jeff Lange\'s kids will eventually pay for this, we don\'t have to pay for this all right now...   What the fuck are we doing?  Who are we fighting against?  China?  This  defense  budget is  urlLink twice what China spends in it\'s  entire  budget.   That\'s right twice their entire national budget, not just defense.  Oh and North Korea?  Yeah, we spend three times more than  urlLink North Korea\'s  entire gross domestic product,   to research the creation of just the Joint Strike Fighter (JSF) aircraft.  Just one of hundreds of defense spending programs.  If you include the war in Iraq as a counter terrorism measure we will spend close to a trillion dollars on counterterrorist activities over the next few years.  In the US, you have a better chance of killing  yourself  (28,332 suicides in 2000), or accidentally being poisoned (9,839 in 2000) than being killed by terrorists.  In the worst terrorism year ever, 2001, 3023 people were killed in the trade center bombings.  Last year the number was zero on American soil, as was the number zero for the years of 2000, 99, 98, 97... well you get the point.  Not that the events of 9/11 should be brushed lightly to the side, but I have come to realize two things in my life.  #1 we are going to die (surprise!) and #2 No amount of spending can stop this first fact.  The massive amounts of money thrown at this unseen threat are a product of the unwarranted fear in this nation.  I have not lost a single night of sleep worrying about being killed by a lightning strike, which kills more people per year than terrorism.  The only time I really  do  think about terrorism is when I have to do something queer because of what has happened in the past.  Yes, check my beat up Birkenstock\'s  twice  for bombs on my way to Amsterdam.  Pull me aside to answer \"a few questions\" when I get back to O\'Hare because I refuse to give \"voluntary\" fingerprint identification for US citizens while passing through customs.  Stick five or six weird yellow sticks in my luggage on the way to Vegas that disappear basically into thin air (can I get a witness Fatty), and my fave because it happened so recently, make sure you have a black federally tagged Suburban parked across the entrance of my driveway as John Kerry\'s motorcade passes my house towards Dubuque.  Why did none of my neighbors get a friendly secret service visit?  It all seems a bit absurd to me I guess.  Today is the night Kim and I are going to see Fahrenheit 9/11.  I plan to be as relaxed as I can possibly be (without drugs) before we go, because if the first two times I saw the film are any indication of tonight I\'ll be close to a coronary from pent up rage by the end of the film.  Also as I\'ve said in the past, although I feel some of the facts are irrelevant, you cannot deny their truth.  So in case you doubt the authenticity of certain statements check it out  urlLink here  and you can find the exact quoted source.   urlLink This  was a local story I saw on the Journal Sentinelâ€™s page that I thought was pretty amazing and it kinda fits in here.  Lastly I got the hookup with Badger tickets this fall.  I couldn\'t be more excited.  The only game I\'ll probly miss is the Big 10 home opener against Penn State... at night.  :(   Why not elope the week before when we are at Arizona, Kim?  Oh well, priorities I guess.  And I guess I\'ll just leave \'er there, so as not to babble all day.  Peace out!  '),(4498,3879020,'04,August,2004','2004-08-04 00:00:00','   Well I do, but before I get to that I just wanted anyone who thought that I cut off half of my weekend while posting yesterday, to take a closer look.  It\'s all there, 2 separate posts.  I\'m sure nobody else missed the second post, but you can thank Temperly for making me doubt you guys a bit, as he missed it.  Oh BTW have a good time in Vegas, buddy.  What I found today was a list that showed up on the MSN start page.  The first one I read was  urlLink the top 10 schools for academics,  and this is the one that got me all worked up.  Every military school with the exception of the Marines\' Citadel makes the top 10.  The fucking Coast Guard Academy?  Come on, and Harvard doesnâ€™t even make the list.  If two job applicants come in to interview with the same qualifications except for one went to Harvard and one went to the Coast Guard Academy, then sailor boy had better realize that he is not walking away with that job.  I can see maybe one, but four of five?  Please.  Iâ€™m guessing whoever is writing these top 10 lists is either high, or is getting a little dough thrown his way by the Pentagon.  The one I did see as pretty correct is Madison as the #2 party school, night Badger football games are insane.  Iâ€™m probably lucky just to be alive after some of those nights.  Other things I noticed was Berkeley not making the most liberal colleges list.  Berkeley is the definition of liberal.  Cripes.  Oh well the lists are fun so check â€˜em out.    The big list of all top 10 lists can be found  urlLink here.   Click on parties, or politics, or social to bring up other categories in their respective fields.   UW- Madison\'s page can be found  urlLink here.   Including prestigious #1\'s in both beer and hard liquor consumption.  Go Badgers!  ');
INSERT INTO `msgs` VALUES (4499,3879020,'03,August,2004','2004-08-03 00:00:00','   Ok so after the weekend blogs you may be sick of reading, but hey, don\'t read this now if you don\'t want.  You can save it till later.  Kinda like a doggie bag, or saving your piece of pie till later at Thanksgiving.  But in any event here goes.   urlLink Meet the Parents  was released in 2000 which is what I think I said, but Tammy seems to differ.  But if anyone can remember (other than you Tam) what was said please leave a message at the end of this post.  If I was wrong I accept my but kick with my dignity.  News Flash:  Kreul is still a Protestant, so is Fatty in case you were wondering.  (maybe a bit of an inside joke there)  Another thing I said I planned on blogging on recently was divorce rates.  Now before I get started I\'m not trying to put a bad omen on anyone here, we have a lot of marriages coming up here, Kathy, Kim, and my Dad, and mostly because of him have I been thinking about this.  And any statistic can be looked at in either a \"glass is half empty\" way, or a \"glass is half full\" way.  If 40% of all high schoolers have tried weed at some point in time, that means 60% have not.  I look at it through the eyes of a cynic however because my life has brought me to a cynical outlook on marriage and divorce.  like the  urlLink alien poster  says though, I want to believe.  Nothing makes me happier than being proven wrong on this one.  I can\'t tell you how happy I was for my Grandma and Grandpa Francis on their 50th Wedding Anniversary this summer.  But with that all said the  urlLink Us Census Bureau reports  that an estimated 50% of all marriages end in divorce.  The actual number is in the 43% range however 4 states including a biggie in California do not report divorces for some reason.  Of course like with gun deaths, obesity, and all the other good things to lead the world in the US is one of the top countries as far as the divorce rate goes, but thank god we are not in the lead.  We still trail Russia, Sweden, Finland, and Britain.  65% Russia?  My fucking god, what the hell is the matter over there?  And 64% in Sweden?  Why even bother.  More disturbing to me perhaps as someone who  isn\'t  getting married this summer is that 1 in 20 people never get married.  Those odds scare me... a bit.  But confuse me a bit too, because 10% of people are supposedly gay and therefore can\'t get married, so why only 5% don\'t get married.  That isn\'t even figuring in priests and the extremely ugly.  The couple of stats I really did want to find, which was the median length of all marriages that end in divorce, and are children of divorced parents more likely to divorce than those from normal marriages, eluded me though.  I kind of wanted to know what the \"hump\" year is on the bell curve, so if I get married I can say, \"It\'s all downhill from here.\"  And kind of know If I was statistically fucked from the jump, coming from a family with divorced parents.  I\'m not really concerned about any of my friends, because there are plenty of 72 hour Brittany Spears style marriages out there to even out the stats, but as Adam Sandler said in Big Daddy \"Think about it, you know.\"  Don\'t break up or get divorced over something stupid.  All the couples out there are like \"Thank you Captian Obvious\" but it happens more than you think.  That took longer than I expected, So I\'ll be quick on these last couple.    They\'re both DVD stories.  First off October 5th is the day.   urlLink Fahrenheit 9/11 on DVD.   Just as everyone expected just a bit before the elections.  I\'m excited to see the second disc as I\'m a big DVD geek, and I just recently got to see the second disc from Bowling For Columbine, and it was one of the best second discs I\'ve ever seen.  Also if you haven\'t seen or bought the Chappelle\'s Show: Season 1 DVDs, what are you waiting for.  I almost wet my pants on several occasions the first time I saw them.  That\'s right literally almost actually pissing, my stomach muscles were sore from laughing.  And the just signed my man Dave (been on my top 5 comics list for almost 10 years straight now) up for  urlLink two more years of the show for a ridiculous sum of money.   I\'m rich Beyaaach!  Oh P.S. and BTW  urlLink \"Code Oranges\"  can bite my bag.  \"Big N\" please!  '),(4500,3879020,'01,August,2004','2004-08-01 00:00:00','   Yeah Don\'t start here, go down to the first post like this and start from the beginning.    So anyway Saturday brought together some of the greatest golfers of all time right here to the midwest.  I\'m not talking about the Buick Open in Michigan, I\'m talking about a star studded best shot team comprised of Myself, Candy Jo, Eric Heim, And Casey Droeszler.  God damn!  How could we lose?  Well I\'m not sure how, but we did, posting a meet worst score of eight over par.... in a best shot.... a four person best shot.  Some highlights of the day included Candy outdriving Beege on the par 3 160-yard \"beat the pro\" hole, Me nearly getting pegged on the first hole by an errant golf ball, Heim and Casey polishing off about a 30 pack of beers during the round, and probably the best drive of my life on hole number 9.  It was our last hole to play and since Casey and Heim had basically just been swinging for the fences all day I decided to \"let it eat\" even if I would probably shank it into the pond.  But somehow the ball took just a nice straight flight path and faded over the bunkers.  I defy anyone that day using my hoopty driver to hit a better shot.  Which brings me to my next point, a lot of the shots we were playing were mine.  Which is a bad thing.  Normally when I play a best shot like one putt and 3 of my short iron shots count... if I\'m playing well.  So when we were using my shots, I knew it was not going to be a good day scorewise.  But we had fun I think, and after paying a $21 bar tab at The Annex (if you knew how crazy that figure was compared to reality, you\'d fall off your chair) and FT ranting about \"owning Platteville\" I\'d say it was a pretty good day.  I was in pretty bad shape by the end of the night, and really didn\'t notice that according to everyone else there was a \"Shiite load\" of extra cops out that night.  But I did get to see a lovely chase/beatdown on Mineral St.  Attention criminals of the world:  Do some fucking windsprints, or some cardio or something.  Homeboy looked like Forrest Gump running down the street.  Only he couldn\'t \"run like the wind blows.\"  Anyway Sunday, Ryan and I went to Buncome and did some fishing, with great success, by my opinion.  \"The Fenwick\" finally struck paydirt and reeled in an 8\" smallmouth as it\'s first catch, so the new rod and reel definately works.  we caught about 25 total and I had 11 so it was almost equal which is amazing for both me, and us.  Because usually I don\'t catch fish on fake bait, and  usually when Ryan and I do any kind of sporting event (fishing, hunting, etc.)  We strike out bigtime.  So it was nice that we bouth caught a fish that would have been about 14 inches had we measured them.  I held my fish next to my junk and it was about the same size so I figured it was 14 inches.  Ha!  So that was the weekend, good times, now I can cover the news and I\'ll be done.  '),(4501,3879020,'01,August,2004','2004-08-01 00:00:00','   Apparently Fatty is alive and well and posting at a pace of one blogpost a month. :) But in all seriousness, nice to see you back. I wish I could explain the Hawking/black hole thingy a bit more clearly for everyone, but I don\'t intend to do that until I have a better grasp on it myself. I need to find one of those \"Quantum Physics for Dummies\" books. Oh BTW way to read ahead there Joshie, like I wouldn\'t post big Hawking news. Sheeeeeit!  So I guess I\'ll start with the weekend, the things that happen all seem to tie together to make an interesting blog, but I guess I\'ll start on Thursday night. Temp calls me at home to ask me if I wanted to play in the best shot golf outing to support the Pioneer men\'s basketball team, that I had committed to earlier. I said yes and he told me to go talk to FT and see what the skinny was. So I walks out the basement door of our house and hear some kind of unruly commotion going on across the street. I walk over to the Temperly\'s and see that a hot tub party is already in progress. Oh this is like 1:30 in the morning BTW. I planned on going over, getting the info, maybe having a drink, and then heading home, because I know the Temperly hot tub parties usually are. 15 basketball players, 2 or 3 girls that are dating the players and enough snacks to stuff a buffalo. More sausage both literally and figuratively than a Saturday Night Live \"Da Bears\" skit. But when I get to the deck it looks like a fucking P. Diddy video. More scattered ass than I could count, and for once it wasn\'t Brooke\'s high school homies. So I\'m like you know what \"I\'ll hang for a little.\" (  You know I think I\'ll leave it there for a sec and come back to this post.  The tip of my one finger hurts.  Ta Ta til then.  '),(4502,3879020,'01,August,2004','2004-08-01 00:00:00','   Ok, all apologies to Jeff on this one cus I can\'t read and seemed to skip the blog where he talks about Hawking and his announcement. Fuck it, it was worth repeating neway. STOP LAUGHING AT ME!!!  '),(4503,3879020,'01,August,2004','2004-08-01 00:00:00','   Well, here it is, my second post...gives you warm fuzzy feelings similar to the ones you got during those \"special\" moments with your uncle, doesn\'t it? Neway, I find it hard to believe that Jeff hasn\'t wrote about  urlLink this  yet, but since he hasn\'t I will. Now, I know a majority of those of you reading this aren\'t even remotely interested in Stephen Hawking or what he has to say...er...type but this is some pretty amazing shit. MC Hawking has basically came out and said that black holes aren\'t the portals to the end of existence that they have been thought to be for so long. Now, this may take some of the intriguing factors away from black holes, but they\'re still cool as hell and this is some major geek news (which means it fits in well considering who\'s eyes this world is being viewed from on this blog). Enough geekness for now though.  As some of you prolly already know, the first collaborative effort ever on a Summer CD has been completed and is ready for the grabbing. Both Jeff(J. Diddy) and myself (FlipFlop LumberJack) are excited about our creation and are eager to hand it out. There was a problem with indecision about CD art but we figured individual art for each cd would be a nice touch. I recommend that you listen to it front to back, but that\'s my opinion and it will be seen how many of you do so.  Well, that will about do it for my second/first August post. Feel free to drop me a line if ya want a cd and keep an eye open for the new LCD tour shirts...hopefully coming soon. You\'re going down sucka, peace out!  '),(4504,4129852,'17,August,2004','2004-08-17 00:00:00','   I just saw spider-man 2... I HAT EHAPPY ENDINGS right now.. mainly cause i get a bad feeling that im not gonna have one...Man i am tired..im tired of feeling and hurting.. and all this shit.. but I can see light... its just too far... i dont know if i can make it.. I will never lay down... i will not keel over... I WONT GIVE UP... i cant ive done too much to stop now      Its funny last night I got 2 offers for dates im such a looser.. i turned them both down... and these are girls that I have crushed on for like 4 months they both out of the blue are now showing interest in me... and i still cant back away.. my dreams for so long were answered and I stepped away... for her      I wish to dissapear completely....  how may I if you read this tell me  please   '),(4505,4129852,'16,August,2004','2004-08-16 00:00:00','   I Feel Small... I feel Like im in a dream.. beacsue Real Life couldnt be like this... I Understand The Song mad world after last night... i had a dream....Sam was walking away from me I was sprintin Ive never ran so hard or fast in my life.. I felt like an olympian.. I couldnt catch her... i couldnt get her attention she was slipping away..... then the ground seperated there was a hude cliff between me and her I kept running I knew the one thing i wanted was her so I lept... And i Failed on the way down i looked up and sam blew a kiss at me. then a bullet wound appeared on my chest..... I was thinking on the way down I realized this is wonderful if i cant have her I will just die.. i then saw myself it was like i was watching myself fall i closed my eyes put a smile on and enjoyed what seemed to be like 4 minutes of falling it was beautiful.... a few seconds Troy appologized for everything he has ever done to me... and i died  I find It kinda funny, I find it kinda sad The Dreams in witch I\'m dieing are the best Ive ever had  Good Bye -Rem   '),(4506,4129852,'16,August,2004','2004-08-16 00:00:00','   My fucking God  Im gonna SWEAR SO much in this so be prepared..... Im fucking in PAIN... I ended prolly the best fucking relationship of my life so far today... cause im in love.... in love with a girl... THAT FUCKED A STONER LIKE 3 NIGHTS AGO.... I dont know what to think....I cant even be fucking angry im just hurt... i ran i ran for like 15 min It HURT it HURT SOOO FUCKING BAD I Can move my god damn LEGS.....IM honestly SICK TO MY FUCKING STOMACH.. i dont know what to feel or say or do... i dont know what the fuck is going on....I cryed today more then i think i ever have in my life....i guess my alter ego poped by today... ive never met him but i can hear him sometimes  he is a fucking asshole he\'s aslways fucking laughing at me...I HATE HIM!!!!!!!!!!!!!!!!!!!!!!!! RRRAAAHHHH..... so yah she fucking ripped my heart out... but its still hers i cant not love her.... i think this is worse than when one of my ex\'s told me she fucked a drug dealer for drugs then got pregnant... hey what an interesting FUCKING PATTERN maybe i should become a drugy again i would get more PUSSY  FUCK           -Rem   '),(4507,4129852,'15,August,2004','2004-08-15 00:00:00','   You Know.. it\'s truely amazing how much of an asshole rem can be soemtimes.. he is inconsiderate, and a jackass.... he is a moron.. he is hurt because a love of his is interested in another guy... its killing him... im not to sure if he can fucking sleep and you know she is amazing if she can do this to Rem... so something must be done...  in the next week a change will happen im sure what it will be but it hopefull wont kill the guy... well see yah later all          --Kalvin   '),(4508,4129852,'05,August,2004','2004-08-05 00:00:00','   well  why do ppl feel the need to improve apon themselves?  I am NO prize but i am happy with me. and some ppl are amazing and it hurts my heart to see them not see it. I\'de kill myslef if I could show the world its own beauty. i would bleed myself to show the heart of ppl. I would cry all the rivers full to know ppl can be happy.  i need ppl to see there beauty cause i am an empty vesle I hold nothing in me i am what surrounds me I am a canvas that fades i fade fast so i need to be around happiness or blackness falls on me and is hard to get out.  I wish i could be a white canvas  to show the world what purity is so they may feel the happiness that I would imbody i would gladdly give my life to show this to everyone or even ONE PERSON.  I am nothing waiting to be something  so incurage you all be something so I may be aswell  THANK YOU FOR READING THIS I LOVE ALL OF YOU   '),(4509,4129852,'04,August,2004','2004-08-04 00:00:00','   I need to spill I am feeling my heavy heart everytime i get htis feeling it feels like my chest is getting thick and heavy with like concrete or something. I feel this way when I am feeling emotion that i dont know how to interpet i dont really know if its sadness or joy i feel now all i know is im feeling a lot of it. I can often feel the way the ppl around me feel I interpret what they are feeling and i in turn feel that way. that is the reason im the way i am. i focus on understanding what other feel and how others act sometimes i forget about me. But hey there are a lot of ppl who are more interesting than me. thats why I am a rounded person I take the cool parts of my friends and i become that, so i feel better about myself. I guess that makes me a huge phony but hey if you are gonna do something might as well be convicted.... or not whatever   i guess im done for now   '),(4510,4129852,'03,August,2004','2004-08-03 00:00:00','   Wow  one thing you may realize while hanging out with rem is that he is very immoral he has almost no conscience its rediculous. I guess he makes up for it kinda he is nice to like evry one. But he has a serious problem unless he hurst some one he really cares about he wont give a shit. He has conned a lot of ppl out of money he is good at it he doesnt mind. But he hates it when ppl dislike for no reason he tries to make everyione happy see he is the con-man that wants to be everyone\'s best friend.  One thing he needs to work on is evaluating risk and having the balls to fallow through with the capable some situations are easily avoidable if you can think through the possibilities. Normally Rem is good at it but lately he has been lacking in this skill  \"hey, Love does funny things to yah\" says rem.  I dont even know why im writing this one Rem just wanted to sill his guts on some things so here I am writing for him.   '),(4511,4129852,'03,August,2004','2004-08-03 00:00:00','   Hey, this is my first blog so you may wanna get to know the character well his name is Remington North he is 17 and he has had an interesting life so far.  When he was 3 he cracked his dad over the head with a frying pan When he was 5 he gave his 14 year old brother Ryan a black eye At the ripe age of 10 he was hospitalized for 2 months and almost died once and actually dies once. it was the most inriching feeling he had ever had waking the morning afetr even though he didn\'t know what had happened he just felt wonderful. At the age of 12 he left the country be himself for the first time on a plane to Cali to see his sister Stephanie, it was a wonderful experience. Age 14 he went to high scholl and fell in love for the first time. Unfortunatly she was a drug addict and got pregnant with her dealers baby then lost it cause of drugs... so we wont talk about her much. Age 15 he fell in love again and lost his virginity to a lovely girl who he still loves Trista she is one of his best girlfriends ever. he finds he loves anything sexual wether it be pleasuring a girl or being pleasured he quikly learns all he can about pleasuring a woman. Age 16 he meats one of the coolest guys ever Jay Vanderheart or \"jesus\" he is one of his best friends ever he teaches him indirectly how to live life and not be afraid when you dont need to be. At age 17 he falls in love twice but that is another chapter.   '),(4512,2831727,'26,February,2004','2004-02-26 00:00:00','   It snowed!!! It snowed!! It snowed!!!  It was so nice to wake up to a white apartment complex.  It would have been a whole lot nicer if my electricity hadn\'t gone out at 10:30 last night and if the skies were blue this morning. I guess all of my wishes couldn\'t come true.  Especially since I did have to go to work today. Yuck.  Plus I made the mistake of lowering my window on the passanger side to get rid of the snow.  Huge Mistake, bucause it wasn\'t snow it was ice and it came flying into my car.  I had to spend the next twenty minutes clearing the ice out of my car!!! Then I got stuck in traffic, which was ok, because I really deep down did not want to go to work.   I hope it snows again tonight.  I don\'t think it will, but it might!!   '),(4513,2831727,'19,February,2004','2004-02-19 00:00:00','   Washington D.C. here I come!!!  I just made my flight arrangements.  I will be in the DC area March 5-7!!!!   '),(4514,2831727,'18,February,2004','2004-02-18 00:00:00','   Funny how in your head there is so much to say, and when you go to write it down, somehow its just not there?    Ok - I\'ve got a topic.  Every year my family gives a scholarship to a student at UNO as part of the Foundation for my sister.  We would like to find a way to have more applicants to the scholarship and are brainstorming ideas of how to reach the UNO students.  The ideas that I have come up with are as follows  1. Use the UNO internet site.  If there isn\'t a scholarship site for UNO, find      some way to create one. 2. Each class or department has to have an internet program (like Learnlink at     Emory) where the teachers can post the application. 3. Attend departmental meetings for each major and minor and tell them     about the scholarship. 4. Post a flyer at the student center/library/every building on campus 5. Write an article for the school newspaper 6. Send a flyer to all of the student advisors.  Am I missing anything or are there any better options or ways to get a bigger applicant pool?  Please post or email me with ideas?    As you can probably tell, I\'m a little bored at work.  I was extremely busy the last couple of days.  I even worked til 10 pm on Monday night, so today I haven\'t done a thing all day.  I\'ve been internet searching, because I would like to find a way to get to DC in the next few weeks to visit Angele and Mandi!!!  I am so excited about visiting DC, if I can only speak to my boss who is in and out of her office all day today.  (Warning: Below is some good old fashioned whining in Lissy fashion.  You do not have to read)  Also on the note of what the hell am I going to do with my life - You know how I have thoughts of being a therapist.  I have serious doubts.  Today a coworker came into my office to bitch about her family and all I wanted to do was say get a life.  You can\'t control your family, you just have to role with the punches, otherwise you will keep getting frustrated with them for the rest of your life.  On the plus side, I said nothing and just nodded my head, so maybe I wouldn\'t be so bad at it.   '),(4515,2831727,'16,February,2004','2004-02-16 00:00:00','   I hope that all of my friends have had a good Valentine\'s Day!!!  Mine was pretty interesting.  I went to a party with Angie and Lauren.  We were definately the cutest girls there.  We may also have been the youngest. Because I didn\'t have a Valentine I went shopping on Sunday and bought myself a new outfit!!!  I LOVE clothes.  On Sunday evening I went to dinner to celebrate Lauren\'s 23rd B-Day. I\'m just sorry that I have work today, otherwise I could\'ve stayed out to celebrate longer.      I\'ll try to post another entry this week, because today I have a headache and can\'t seem to think.   '),(4516,2831727,'10,February,2004','2004-02-10 00:00:00','   I\'m at work again.  Boo hoo.  I\'m getting back into the swing of things after a very humourous weekend.  Saturday was my friend Gail\'s Birthday, so we went out on Friday to celebrate.  And where does Gail pick?  Could it be simply dinner and a drink? Nooooo!!! She picked a male stripclub!!!! I couldn\'t believe how funny it is to watch men strip nekid.  And of the six who went I was one of the three that had a good time!! (no I did not have a lap dance you naughty people).   Saturday I went out again, but this time only to sushi with Lauren and Angie.  I just love sushi, hmm but now I\'m in the mood for Indian or Thai.  Yum.   '),(4517,2831727,'05,February,2004','2004-02-05 00:00:00','   Hey Everyone -  I don\'t know if anyone is going to read this, but I decided that to start my own blog so that I can talk even more about myself :P  I am currently at work and as I always am doing, trying to figure my life out.  I\'m trying to figure out what to do in my immediate future.  I know that I don\'t want to stay in my job (as a paralegal assistant), but I\'m not sure what to do next.  I have so many ideas running around my head, I think I\'m going to use this blog as my journal.  First is that I want to eventually move back to New Orleans, whether this means within the next few months or within the next few years.  I want to see what living in New Orleans is like as an adult.  It\'s scary to think of going back because I was unhappy living there in my last years of high school and none of my friends really live there anymore, so I\'d be starting from scratch again.  Plus it\'s not like the economy is jumping in NO, so my opportunities will be limited.  On the other hand my family is there and I want to be near them.  So should I move back there, get some rinky dink job while I figure out what grad school I really want to go to? or should I stay at a job I don\'t love, but don\'t absolutely hate either? or should I apply to this program I just saw on the internet which would be an internship in London for a year?  Which brings me to my next thought process.  What career path do I want to follow next.  Options are, Psychology because there is this big part of me that wants to help people through their problems.  or Business school so that I learn about business because I have such a limited understanding and then try to find a job in the business world, like event planning or hotel management or advertising/pr.  Ok, I think that these are enough thoughts for now, plus I\'m at work so I should get back to it now.   '),(4518,2831727,'23,March,2004','2004-03-23 00:00:00','   I\'m sorry that its been awhile.  Although who knows if anyone is actually reading this.  I returned from D.C. a lifetime age.     Washinton D.C.  I landed in D.C. around 10 pm where my friends Angele and Latorial picked me up.  Angele has been my friend since Kindergarten and I\'ve known Latoria since my Junior year of High School.  We dropped off my suitcase at Angele\'s and then we went out in Georgetown.  I met the neatest people from New Zealand.  The were so much fun!!! I danced well into the morning with my new and old friends.  I didn\'t get to sleep til 4 or 5 am.  I had five hours of sleep. I woke up on Saturday aroun 9 am.  Angele\'s housemate Claudio (I think that is his name) cooked us breakfast.  We had crepes with nutella and jam, and homemade bread (not from a machine, but actually baked in an oven).  I spent the day visiting Angele\'s hangouts and friends and then visiting Georgetown with Latoria and her friend Alex.  I then had to take a quick nap and then was off to a party with Angele and her boyfriend.  The party was not so much fun, but then we left to go out where we met up with Latoria and had so much fun bouncing from one place to another.  I had so much fun, I did not even mind when my jean jacket got stolan :(  I got home around 5 am again. Sunday Angele and I woke up and had brunch around her house at this little hole in the wall that seriously had the best breakfast potatos.  She and I then drove around D.C., constantly changing our minds as to where we wanted to end up.  Finally we went to this cute little coffee shop where I found my favorite coffee drink.   Big Trains!!! For any of you that don\'t know what Big Trains are, they are my frozen coffee drinks I get in Colorado and at Chocolatte!!!  Angele and I then went to get hot dogs for an early dinner, and then I was driven to the airport.  I had such a fun time visiting with my friends and seeing D.C. from their perspectives.   I\'m sorry that its taken me so long to post this, but work has gotten seriously busy.   '),(4519,2831727,'06,April,2004','2004-04-06 00:00:00','   I just returned from a quick trip to New Orleans to visit the family and to host a bridal shower.  The shower went really well.  We had it the country club near my parent\'s house.  It was very simple.  The food and service was provided, so all we had to do was decorate and show up.  My cousin seemed to enjoy the shower and all the gifts she got.  By far, my gift was the most stylish!!!  The worst gift was of course from my crazy great Aunt Nanan (the one who has been hit by a car, not once, not twice, but three times). I spent my entire saturday looking for silver shoes to go with my bridesmaid\'s dress, since the ones picked out by my cousin were ugly, only to buy the ugly shoes.  I attended a friends wedding and found out another friend is moving to Atlanta, so I am excited!!! Monday, I went shopping yet again, but this time for clothes.  I currently have five new skirts!!!!! And have replaced my stolen jean jacket.  Today, I\'m worn out and a little bummed.  The only bright spot is that I\'m going to the Braves/Mets game tonight with Angie, Lauren and Courtney.  I\'m excited about that.   '),(4520,2831727,'13,May,2004','2004-05-13 00:00:00','   urlLink    Me &nbsp; urlLink   '),(4521,2831727,'13,May,2004','2004-05-13 00:00:00','   I know that it\'s been awhile since I\'ve posted, but I\'ve been extremely busy.  I have suddenly decided to make some changes in my life.  I decided I no longer want to work where I\'m working and that I want to act young while I am young.   I have applied to the program that would let me live in London for one year!!!  I am going to be in New York for an interview over Memorial Day weekend!! I hope that everyone I know in New York will be there that weekend.  I have also decided to go to Steamboat for 6 weeks over the summer!! So if anyone wants to visit, there I\'ll be!!!    If I don\'t go to London in August, I don\'t know what I\'ll be doing, but I\'ll be doing something.  Maybe I\'ll do something zany.  Also next week is my cousin\'s wedding, so I\'ll be in New Orleans from tuesday May 18 to sunday May 23.  I\'m going to be traveling a lot in the next three months.  Not that I\'m not already going bonkers with all of my traveling already.  So far I\'ve been to New Orleans twice and Washington DC once in the past 2 months.    I also want to congratulate Gail for graduating Law School.  She will be kicking other lawyers\' asses in the future.   '),(4522,2831727,'21,July,2004','2004-07-21 00:00:00','   Colorado is Beautiful!!!!&nbsp; I love being here!!! The only thing that I hate is that my internet out here is so slow that it is painful to check.&nbsp; Right now I\'m in an internet cafe updating this site.&nbsp; I\'m not sure if anyone checks this anymore, since I am so bad about posting.&nbsp; Here goes anyway.&nbsp; I have a tan!!&nbsp; I spend my days hiking, studying for the GMAT and running errands.&nbsp; I am itching to go to work though.&nbsp; I have way too much time on my hands!!!!  Nola (my cat) has turned into the greatest huntress that ever lived.&nbsp; She caught a mouse two days ago!!! It was an extremely exciting moment.&nbsp; She is now called Queen Nola!!!  As for my plans for London.&nbsp; I have an interview with JP Morgan tomorrow!!!&nbsp; So I should hear soon about whether I am going or not!!!  I am getting really excited about London and everything.&nbsp; I will keep you guys posted!!!   '),(4523,4163223,'10,August,2004','2004-08-10 00:00:00','   For quite sometime now, I have felt that I have been wasting away precious time from my life. The time in which I could have built something useful not only to me but to others as well, the time in which I could have gained more knowledge, become wiser and moved onto the next level of enlightenment (I assume realizing this is the first one). During one of my frantic and desperate searches to do something worthwhile, I came across this really great statement by  Scott Stark, CTO of  JBOSS  . The link is -  urlLink http://jboss.org/community/overview    For those who have no clue of what Open Source is, read this you might end up being one of the super stars that Scott talks about.  JBOSS is an open source Java application server, I was looking for ways as to how contribute to it. This article is a must read for any programmer who has ever thought of contributing to open source.  '),(4524,4163223,'09,August,2004','2004-08-09 00:00:00','   For all those action movie fans, Vin Diesel symbolises the ideal action hero. He\'s got lot of attitude, muscles and style. ;-) This was the second time I was watching this movie. Riddick (Vin Diesel) is a most-wanted fugitive, being searched by a number of planetary systems. He is being hunted all the time by mercs (mercenaries) who are looking for a prize. While saving his own ass, and kicking asses of such mercs he ends up at the planetary system of Helion Prime.  Where he meets one of the Elementals (those who calculate future using probability, we havent got such people right now I think, but the concept is not bad) , who make him realize that part of him which he has forgotten. That he is a survivor of planet Furia, that he is a Furian.  Helion Prime in the meanwhile is under the threat of attack from the Necromancers (a breed of people led by the Lord Marshal, how is half human and half something-that-nobody-knows-what-he-is). Their motto is \"You keep what you kill !\", a very bold faith which leads to some interesting dialogue scenes in the movie.  In the events that follow we see a lot of great action sequences especially the ones on the triple security prison on planet Crematoria. The movie proceeds in a good speed, great special effects and amazing scenes. It is a must watch for all sci-fi action movie fans.  PS : The theatre we visited was Plaza. It is a no go ! Pathetic screen, pathetic sound, pathetic movie print. But if u have no other choice, I suggest take the cheapest tickets at least you will be nearer to the screen to listen to the sound.  '),(4525,4163223,'08,August,2004','2004-08-08 00:00:00','   Had the opportunity of watching this amazingly disconnected oddball of a movie yesterday. Mujh se shaadi karogi another attempt of mash-mixing hit hollywood flicks by David Dhawan.  Anger Management, Meet The Parents, Matrix  the movie draws from all of these  foren  films. The movie is basically a comedy masala movie. ( lots of oomph ! like the  kanta laga  girl playing the role of  Bijli  and doing an item number) .  David Dhawan is banking on making money by attracting  desi  audiences by showing some  maal . Having  Baywatch  attire through out the movie. (Though I give credit to the movie for having a great collection of songs).  I spent half the time waiting the movie to finish.  Sallu baba and Akshay kumar spend most the time showing off their muscles. And for a change Akshay kumar is the saving grace ! The guy did a good role in Khakhi too, I guess he gives a good performance whenever he is cast in a multi starrer, so for the want of competition he ends up raising the level of his show.  I should say this as a final verdict, like my friend Makkhi says : \"Dimag mat lagao !\". A thing easily said but tough to do, when u have nothing else for 3 hours to do.  PS : Those new to Bglore from Delhi, the theatre was NAGA in Ulsoor. I suggest dont go for the front row, its out right discrimination they have actually put chains between the front stall and others so you have to come out of the hall to goto refreshments ! Besides the chairs in front stall are hard wood ! Talk about difference between a 30rs. ticket and a 50rs. one.  '),(4526,4163223,'08,August,2004','2004-08-08 00:00:00','   After graduation from college you look for a job, in my case I landed up in Bangalore (25th July 2004 - for the record), ending up to work for off shored IT division of a US company. Work hard, earn lots of bucks and party hard. The office is on MG Road, its as famous here as the Silk Route is in the Eastern World. Banks, companies, malls, theaters, Dominos, Pizza Corner, stolen goods market, smuggled goods market, Baristas, Cafe Coffee Days, book shops, video game parlors and of course a lot of indian population - a good big package of things compressed on to a long road. (no doubt roads are lifelines of a country)    The weather in Bglore is much like that in Europe, only the traffic is much more and lot more indisciplined. The first Saturday night I ended up buying a Tag Heuer watch.... a cheap imitation of course ;-). Bglore is a young city with fast life and a hubbabub of multicultural torrents mingling with each other. Its like a young teenager trying to look cool too fast, living life in the fast lane but only his style is cramped the lack of space and bad roads. In every other corner u would find a tech company, each one trying to become a behemoth - a collective effort though disconnected that might turn Bglore into a spank new version of Silicon Valley, outside the US of A.  '),(4527,3463044,'26,May,2004','2004-05-26 00:00:00','   I\'ve decided that I need to write.  Like drinking water or getting sleep, it\'s something I have to do.  I am better when I put words down, and while I prefer the act of longhand on paper, I\'m going to try this medium in order to hopefully do it more frequently, more consistently, and more creatively than I have been able to do until now.  My spiral notebooks have become coloring books for Mackenzie and scratch paper for David.  The computer is safe, for now, and the idea that the blog is out there, even if no one knows about it (at least not yet), will compel me to write and add notes.  What I plan to write about is completely up in the air.  My husband, my kids, my life, what I hear and see and read, what I do or don\'t do, it\'s all food for the writer in me that I hope to channel into something productive via this medium. '),(4528,3463044,'17,June,2004','2004-06-17 00:00:00','   Apologies to all, but only stay-at-home Moms (or Dads) are going to get this column.  I mean no disrespect to the many working parents out there, I used to be one, too.  And I know you\'re with your kids on the weekends, but you\'re just not going to get this column.  I have two kids, nine and three.  David is from my first marriage, and consequently has some stuff to deal with, which has gotten easier as he\'s gotten older.  Suddenly, in the last three weeks or so, the maturity level has gone up a couple of notches.  Don\'t get me wrong, his room\'s still a mess, I get the eye roll at bed time, I still can\'t figure out how the water gets out of the shower when there are two shower curtains to keep it in the tub, etc., but he\'s become attentive, doesn\'t argue as much, accepts direction and finishes a project before moving on, I\'m almost sad (I said   almost  ), I think he\'s growing up right before my eyes.  Mackenzie is my three year old and from my second marriage.  She is amazing to watch because she is so much like my husband and I, throwing back my own words at me can be sometimes hard to hear.  Her preschool teachers constantly tell me that sometimes they forget she\'s three because she articulates her thoughts so well.  She idolizes her big brother and he adores her, but they are brother and sister, and when that fact comes into play, it can be ugly.  David is home for the summer, school has been out for a couple of weeks now.  Mackenzie is in preschool three days a week, mornings only.  So today I had both kids home this morning.  David had day camp at school at 9:30 a.m, the same time Mackenzie had her swimming lesson across town (Mom didn\'t have her organizer with her when she scheduled this mess).  It therefore was imperative that we leave our house clean, dressed, and fed by 8:40 a.m.  No one wanted to get up, no one wanted to get dressed, no one liked my breakfast choices.  St. Vincent de Paul was coming by to pick up the remnants of the garage sale leftover for donation, and I needed to get that stuff out of the garage and leave it up front for them to pick up.  The dog and parrot both needed attention. And for some reason, today was the day my phones started ringing at 8:00 a.m.  David has pulled himself together and I ask him to get himself and his sister in the car.  Mackenzie is all over the driveway, running from flies (our newest fear) and not getting in the car.  David has made the valiant attempt and I tell him not to worry about it, I\'ll handle her.  At the precise moment he decides to swing the door closed, Mackenzie decides she better cooperate.  Her hand didn\'t get caught, there were no broken bones, but she did get some skin scraped on the edge of her finger.  She proceeded to cry about her finger for the next four and a half hours -- all the way to David\'s school (20 minute ride).  All the way to the YMCA (25 minute ride).  Through the entire swimming lesson (30 minute class).  Over to dad\'s office (20 minute ride).  At dad\'s office (a very quick ten minutes).  Back to pick up David (35 minute ride) and drop him off at a friends house (10 minutes).  (Now here I must interject the next piece of good fortune.  At dad\'s office, we utilized part of our visit in the bathroom, she is fully potty trained and I haven\'t carried a diaper bag or a change of clothes in at least six months.  As she went to the bathroom, I noticed she had what I\'ll call RB (short for Runny Butt, I\'m sure you get the idea).  I\'m surprised, she hasn\'t had RB in a while, and I\'m chalking it up to the fact that she\'s made herself so upset this morning.  I\'ve tried a variety of things at this point to distract her from her pain and get her to smile and nothing is working.)  After the friends house, I decide to take Mackenzie to a local park close to the friends house, thinking I won\'t have to make the big loop around when I pick David up in a couple of hours.  It was hot, so I decided to take Mackenzie to an ice cream store, counting on the sugar to give her energy at the park and make her happy and give Mom a break because the last nerve was about ready to blow.  She smiled.  Things were good.  Just as the ice cream got to the table, we had major RB.  She lost it again.  I haul her into the bathroom and survey the damage.  It couldn\'t have been much worse.  I had baby wipes but that was it, no change of clothes, no diapers or underwear, and she\'s screaming the whole time.  Chucked the underwear in the trash, not salvageable given my current situation. Washed out the pants from the inside, had her put them back on, and I slid baby wipes and paper towels around her bottom.  Cleaned up the bathroom as best I could.  Got my drink to go, left her ice cream (more screaming) and back home we went (25 minute ride).  Changed her at home, caught my breath for five minutes and back for David we went (25 minute ride).  Picked David up and now it was time to take him to his dad\'s house, he\'s there every Thursday (35 minute ride to dad\'s).  Poor David, he pretty much sucked up to his sister every time he saw her, and she still wanted nothing to do with him -- he tried singing to her, making funny faces at her, telling her jokes, I felt bad when he reached for her hand and tried to kiss it and make it better, she completely blew him off.  As I walked him to his dad\'s door, I told him not to worry about it, she\'d be better tomorrow when he saw her.  Let\'s hope that wasn\'t an extreme stretch of the truth.  When kids have diarrhea, the things you worry about most is dehydration.  You\'ve got to keep the fluids going down their throat, the idea is to get the fluids in faster than they come out.  Of course, the job is twice as hard when they don\'t want to take them in at all.  By 3:00 p.m., I was on my way to the doctor\'s office, she just wasn\'t right, the constant crying was starting to worry me even though it didn\'t sound like the pain cry, it was definitely the wimpy cry, but nonetheless, it was time.  And going in to it I knew there was no medicine I could give her, nothing to stop the RB, but I had to try something.  She blew twice while we were there.  (Thank goodness I picked up a change of clothes.)  While the doctor told me lots of stuff I already knew, he did mention something I didn\'t.  Since he figured out her particular case is viral, he said it would be contagious and adults could get the same thing, but in a much milder form.  Hmmmm, double extra hand washing for me.  Called the husband to tell him what was going on, and was told he had left a little early, had said he was feeling a little tired and run down.  Hmmmm, he hadn\'t mentioned that to me when I talked to him earlier in the day.  We got home, and there he was lying down on the bed, said he didn\'t know why he didn\'t feel so good.  I did.  Why do I direct this column at those stay at home parents?  Because this was one of those days that all these things were going to get done and not a darn thing got accomplished.  Now the health and well being of my child is paramount and my life can go to hell as long as my kids are OK.  But what was shaping up to be a good day fell apart so quickly.  My laundry is still on the floor.  The dishwasher is still loaded with clean dishes and the sink has dirty ones in it.  My kitchen floor is, well, let\'s not talk about my kitchen floor, OK?  My coupons are still on the table waiting to be cut and sorted and my bed didn\'t even get made this morning (but I don\'t think the husband cared when he came home and rolled into it) and I didn\'t even cook dinner today (again, not high on Keith\'s priorites right now).  I know I didn\'t do a damn thing today and I feel guilty and useless.  So it was a surprise when David called me from his dad\'s, to see how Mackenzie was feeling and to thank me for taking him over to his friend\'s house today.  And Mackenzie announced to her dad that she didn\'t want him right now, she wanted Mom because I had taken care of her all day.  I went to bed thinking I had done a good job for today after all.  And maybe if I get up extra early before she wakes up I can unload the dishwasher and fold the laundry, etc. And if she wakes up early, well then we\'ll just cuddle in Mom and Dad\'s bed for awhile. '),(4529,3463044,'16,June,2004','2004-06-16 00:00:00','   We had a garage sale this past Saturday and I\'m happy to say we netted about $450 for a day\'s work.  We have one every couple of years when there\'s too much stuff and not enough room.  I\'m sure as my kids get older the frequency will diminish, right now my kids are still young enough where we\'re trading up (crib for twin bed, big dresser giving way to smaller dresser plus desk, etc.), and my house is overrun with the small things of infancy and we\'re almost through toddlerhood at this point, so from time to time the expulsion of all these once needed necessitiies is paramount to ramaining sane.  David had 343 remote control objects of various sizes and shapes, how many can a single person control with just one pair of hands?  (David\'s original answer -- 343.  His revised answer after my anger started fogging up his glasses from three feet away -- 342).  Mackenzie had so many young child toys, stuff she got on her own merits plus some of the good ones I saved from David little kid days, they were everywhere.  I had kitchen paraphernalia I don\'t have a chance to use anymore, Keith had a bunch of really nice polo style shirts with lots of big names on them, but since we\'ve been working out his shoulders and arms don\'t fit in them anymore, plus other stuff that was once vital and alive, but now gathers dust waiting to be used again.    The decision to discard one\'s own stuff can be either heartbreaking or made without a second thought.  Some of the kid\'s stuff was easy to put in the discard pile.  Some brought back memories of a child\'s face, lit up like Las Vegas, and for just a minute, you think if you hang on to the item, you\'ll hang onto the look on that child\'s face, maybe you can bring back a better time, a happier time, an easier time, you choose the adjective.  Coming to the realization that it\'s just stuff, and the past is just memories can be hard to deal with.  Then you have to price the stuff, and it\'s agony again as you try to put a value on your memories, or worse, a value on your taste.  Five dollars for that!  I paid thirty dollars seven years ago when it was new and before everybody and their brother had one!  I told David that whatever he contributed to the garage sale would be his money if it sold, but if I pulled items from his room and I sold them, the money would be mine.  Even my nine year old suffered from the stress of his decision making, I could hear him rationalizing about the train he got for Christmas that year, or how that car was his favorite gift from his dinosaur themed birthday party.  In the end, he did contribute about a dozen items, in a range of sizes and shapes and ultimately made $16.  I\'m not much of a garage sale shopper myself.  There\'s a time constraint involved, plus the fact that I can\'t do much shopping with a nine and three year old around, it puts the damper on shopping of any kind.  When I lived in Southern California, a garage sale might net a great Disney find, or a piece of transferware to add to my collection.  Here in Idaho, the chances of that Disney find are more rare, and while I might be able to find something of value worth my time, it\'s not high on my list right now.  I wanted to sell my stuff Saturday morning, but it was a great way to watch people for awhile. Watch and see what they were looking for, watch them and see them criticize my stuff, and then of all the injustices, watch them as they turned their nose up at my valuable stuff.  What do they seek among my discarded memories?  I watch them and wonder -- what are they looking for?  Are they trying to replace something they once had?  Are they looking for a more inexpensive version of something?  Having been on the extremely tight and fixed budget at one point in my life, I can certainly understand shopping with quarters and dimes, but was still taken aback when I actually saw my former self buying a small toy for a child.  A two dollar toy with one dollar being scraped out of the ashtray holder in her car.  I used to do the same thing for french fries for my child, and often couldn\'t scrape the lousy 99 cents together.  And it wasn\'t that my child needed french fries everyday, but I wanted to be able to provide them.  I took her money and put the toy in a bag for her, and as she was leaving I offered a couple of cookies to her child, who eagerly shot out a hand.  They were cookies that I was selling, a quarter a piece.  As she went to stop her child (she had seen the sign), I told her if it was OK with her, it was OK with me, it\'s hard to say no to a child and I had been giving them away to kids all morning long.  Maybe it was just wishful thinking on my part, but I think she understood I had been there once, too, and let her child take the cookies.  My husband said he thought I was going to just give her the toy.  I told him that I wouldn\'t haven\'t robbed her pride from her for anything.  To me it was a two dollar toy, but that child needed to see his Mom buy that toy and hand it to him, and she needed to buy it and give it to him.  He was so happy in the back seat of his car, clutching his new toy and wiping cookie crumbs from his face.    A simple reminder that one person\'s trash is another\'s treasure.  That toy brought my daughter joy and I\'m glad to see it continues to do it\'s job elsewhere. '),(4530,3463044,'05,June,2004','2004-06-05 00:00:00','   I suffer from migraines.  I have since I was about sixteen.  So seeing as I\'m well into over twenty years with them, I consider myself an expert on the subject.  I\'ve read everything I can get my hands on about them, I know all about the latest pharmaceutical inventions in order to seek relief, I have a personal list of how and sometimes why I get them, and how to deal with them when they attack.  And yet with all my knowledge about them, and even what produces migraines in my own head, I still suffer their agony.    Sometimes, the pain is on one side of my head, sometimes it\'s the other.  It can be in the front or in the back of my head.  And that is where the differences end.  My migraines are always an unrelenting throbbing pain, I literally can hear the pounding inside my head echoing through my ears.  And while I wait to see if my medication will do me the favor of easing my pain, I assume gymnastic positions in order to put pressure on the offending part of my head.  I\'m in the dark, because the most miniscule drop of light on my eyes is nothing short of blade-like steel being pushed through my brain.  My eyes are closed, I pull a blanket over my head, bury my face in the pillow, and yet I still see light -- damn that knife!  The medications are at once wonderful and frightening.  I remember when I could take one Vicodin and blissfully sleep for twelve hours on just the smell of the drug alone; now, I take two Vicodin and stay awake long enough to feel the headache burner get turned down from scorch to simmer before I sleep for a couple of hours.  I\'ve been through a handful of drugs to help in easing my pain, some didn\'t work for me at all, some worked great until my body refused to allow them to do their job.  My body, rejecting what could save it from the headache inferno, deciding instead to suffer the pain until something better comes along.  At times, I am powerless, and I must seek the attention of someone with more power than I have, someone who can give me what I need and quickly.  The doctors string me out often enough, having me answer their stupid questions  (can\'t they see I am an expert in the suffering of migraine headaches?)  before giving me the relief I seek.  Their medicines are much better now, I feel the immediate effects of tranquility and know that when I wake later there will be no drug hangover for three days.  Why do I write about my migraines?  Because I had one yesterday, a small headache that ibuprofen had no effect on, which through the course of the day started developing into something that could have undermined my plans for the weekend.  I was at the point of the second to last step, when thankfully my medicine decided I had had enough for today and slowly eased its way back to wherever it is it comes from.  And I think about my migraines and wonder why.  Why did yesterday\'s migraine retreat after just a minor skirmish?  Why didn\'t it stay and fight, like it has so many other times, and win?  Did my body have something in it yesterday it didn\'t have before, something that turned it into a killing machine against the headache ravage?  If so, can I figure out what it was and bottle it, to save for the next attack?  What possible use could there be for having a migraine headache? '),(4531,3476799,'27,May,2004','2004-05-27 00:00:00','   Sitting here in this room.  Waiting for her to make a move.  I\'d like to hear the phone ring, I\'d like to have the inbox filled.  An instant message.  A call from someone else letting me know she asked about me.  Maybe an old fashioned letter.  The problem is, she doesn\'t exist.  Somewhere out there she does, but she hasn\'t met me.  She doesn\'t know my name.  She can\'t even tell what color my eyes are.  Never heard my voice, never asked to hear mine. Somewhere she\'s happy with the guy she has now, or is she alone asking the same question I\'ve wondered for sometime. Where is the one for me?  I may find her tomorrow, I may find her next year. Perhaps she does know me, but it\'s my fear of failure that keeps me from getting to know her.  Not in idle conversation, or the casual question, but asking her to let me know her, and peering into her soul through her eyes and getting the answers I seek.  I will meet my fate one day, and I hope my fate is her.  '),(4532,3476799,'27,May,2004','2004-05-27 00:00:00','   I\'ve seen the commercial over and over again.  The one where the guy and the girl get in the elevator with the older black lady and she exchanges a glance and smiles at the girl.  The guy tells the lady to have a nice day and they walk out with the girl doing a meekish lookback at the lady.  The back of the guys shirt reads \"Child Abuser\" or something to that effect.. That\'s all I need to see again right now.  I want to call. I just need to see that commercial or for someone to tell me to call.  I saw something the other day more obvious then the guy with the shirt clearly spelling out that he beats his kid.  Standing at the door at a retail location the other day a small child, thin, probably thinner then he should be entered the store. He went for the gumball machine and when he did his father grabbed his arm, nothing there bothered me. Lord know I\'ve grabbed my nephew and had to pry him from something but what the child did was what disturbed me.  He didn\'t flinch or cover his head like I\'d expect, hell I still flinch when my older brother is around and I see something flash out of the corner of my eye.  This kid however turtled, covering his head with both arms and curling up in a fetal position before his small body even had time to hit the ground.  This wasn\'t the reaction of a child that had been swatted on the but a dozen or more times in his life, this was the animal instinct of a child that had the fear of being pummeled once again.  His father didn\'t hit him that time.  He yelled at him sternly and as I said, had it just been the stern yell, and the arm grab, and even a smack on the ass it wouldn\'t have bothered me one bit, but the reaction again is what I go back to.  Through work while the guy was there I was able to get his name and adress.  His wife I\'m certain is also a victim.  As the exited he questioned her about something they bought and when she told him not to worry about it he yanked the recepit off of her saying \"Fuck you\" and off he went in the other direction.  I don\'t want to get involved, but if I don\'t then what?  Whne I was growing up I was a small child, I\'m still a small man.  My older brother and myself looked like we where malnourished while in fact, that\'s just the way we where, still are.  When the school nurse reported my mother for child abuse on that alone it was horrible.  Half a dozen blood test on my brother later it was just determined it was genetics, nothing that my mother was doing.  That\'s what I fear, alienating someone else that may be a decent person like my mother.  But decent people don\'t tell their wife \"Fuck you.\" in private, let alone infront of a complete stranger in a public place over something so mundane.  Decent people don\'t have children that fear for their safety at a sudden movement by their children.  I\'m a decent person.. if I don\'t do something, will I still be a decent person? Or just like him.  '),(4533,3191071,'23,April,2004','2004-04-23 00:00:00','   Things  There are things inside me I donâ€™t understand, Like why my life seems so happy when I feel so sad, Why do I think itâ€™s time to play when thereâ€™s work to be done, Maybe cause there is no time for fun,  I wish I could just fly away in to the morning light, At least then I wouldnâ€™t have to fight, Fight for all the things that donâ€™t matter  Like how I comb my hair Still the things wind tight around me  So tight I can not bear,  So tight I canâ€™t breathe or even think to do  What I want to do with you  Still I grit my teeth and try to understand  How my life is still around, How itâ€™s not gone like all the otherâ€™s here lying on the ground  Itâ€™s not all wasted on drugs on booze   Or why I didnâ€™t try  I work hard to live not knowing why I work hard to live, I work hard to stay alive  Still this my reward wondering why  They say donâ€™t let your emotionâ€™s take control  Just put them in your back pocket and donâ€™t let them grow But when you suppress your feeling  This is all that comes out   Anger, hatred of your self and all that people have done And years and years of fighting that never will be won  Still we sit and wonder why things are  Things that need to go away far.  '),(4534,3191071,'30,May,2004','2004-05-30 00:00:00','   So any way how have you all been no new poetry today just I feel like typing and so I am. OMG so any way this really ishy girl was being bitchy to me one day and so I said \"someone\'s being bitchy today\" (inside joke friends movie theater)anyway she\'s all did you just call me a bitch and I\'m like all innocently no.then as if she can\'t even talk for her self the \"queen bee\" of their group decided to come and have \"a word\" with me. So she\'s talking and saying how I should wach my mouth or I\'ll end up with my ass getting kicked. And I told her I\'m like \"you don\'t want to mess with me I know people people high up\" (like cops) I told her about my sisters probation officer and they must have miss heard me or something because next thing I know they all think my sister is a cop and shes only 16. so in conclusion I think that people should stand up for them selves!! Love ya :) xoxoxoxooxoxoxoxoxoxoxoxoxoxoxoxoxxoxxoxox, this is my hugging and kissing you! jk!  '),(4535,3191071,'06,June,2004','2004-06-06 00:00:00','   Omg so much news so little time, my Birthday is in 2 days and my party is in 6! This is going to be the best party ever! But I still have to clean (the bathroom, the living room, the kitchen, and my room), finish tileing the basement, go get  and  put up decorations, and get food so Iâ€™ve got my work cut out for me! So anyway I just got this creepy message from this really biznichy girl monkey something and she wouldn\'t tell me anything about how she got my name or why she had it just that her friend gave it to her so that was kinda creepy! But yea if you read this and your her friend or her tell her that she is/youâ€™re a biznich and you/she is not to ever IM me again and she/you can KISS my BUMBUM!  '),(4536,1359631,'27,March,2004','2004-03-27 00:00:00','   Change is supposed to be the only thing constant, and with change very human being aspires to grow, grow in terms of career, in terms of social influence, in terms of wealth, respect, power and what not. Growth, like change, has no limits. From time immemorial man has always being growing, be it with the discovery of wheel, metals, industrial revolution, cyber revolution, etc. Horizontal growth and vertical growth have been the major contributors of growth. When man grows with the available means in the space already available, the growth is horizontal. When space is no longer available, man grows vertically to the next stage where he creates space for himself as well as for others. This is typically influenced by discoveries, inventions, etc.   In an organization, the primary motivators for an individual could be either career or money. In both he tends to keep growing. Let me just try to look a bit deeper into growth in career. Growth, like all other adjectives, is relative. We always want to grow towards a goal, a target, something that could be long term or short term. When I speak about goal, it could be an idealistic goal based on position, designation, achievement, etc. However mostly we tend to look at seniors and say to ourselves, â€œYes, I need to be like himâ€ â€“ basically you need to have role models. When there is no role model to look forward to, the growth is often achieved by the hard means. Why is this so?    Growth by Reference  I believe that it is easier to grow if you have a reference target, a target that you can keep referring to and check if the actions that you are taking are in line with your career growth plans. For example a technical architect could look forward towards his enterprise/vision architect who could mentor him. When it comes to mentoring, you know what you want to achieve and there is someone who has already reached that target and knows what exactly has to be done in order to reach that level. In this case it is easier to plan your growth and also avoid the pitfalls or diversions that would take place if you were to do it all by yourself. For e.g. doing a MBA may not be of any help if your aim is to continue in the technical line.    Growth by Trial  This usually more time. Even though the person gains more experience, ultimately you would be way behind the person who has grown by reference. It is highly possible that the plans you have laid is not in line with what you need to achieve. This may prove disastrous, as you would have to spend more time in turning around and taking the path you like. Moreover it is highly likely that such growth may not always bear fruit. If the organization does not have space for the goals that you want to achieve, then your growth could be hindered. At this stage, when horizontal growth (here I am considering the org. as the horizontal boundary) is saturated you need to look for vertical growth (probably move over to a new org.).   Organizational Growth  Leaving the personal growth aside, let us look at how and when an organization grows. An organization grows when its employees grow and flourish in their career. Employees taking on more responsibility within the same job profile characterize the horizontal growth. For e.g. software engineer may start off with just maintenance. He could then start coding new modules, reviewing, testing, low-level design, etc. Thus he has space for horizontal growth. Once the horizontal growth is completed he should have an opportunity for vertical growth. Often people would leave when there is less space for vertical growth. Since growth is more like a pyramid, only few people can grow vertically within an organization. Hence attrition may not really be a sign of weakness but it could also be a sign of growth, provided the number of people entering a horizontal space is more than the number of people leaving it.   Growth is a fundamental aspect of life. We need to keep introspecting whether there is growth in our career and life. We also need to keep checking the space around us available for growth and influence our decisions accordingly.  '),(4537,1359631,'12,March,2004','2004-03-12 00:00:00','   This poem is a reflection of the last 1317 days with which I have been associated. This is my first attempt at poetry.     derived from dreams  driven by heart  for the morning tomorrow We sweated it so hard   when morning came and time for bread  came the call  it was time for bed  and back to dreams!   little did We see the morning little did We see the tomorrow in today  We did see our dreams We did see our heart  down the lanes of memory  not so far from today  there was a dream, realized  morning had not come but  We did convert dreams to reality  now We dare to dream again in the hope that morning will be none too far !  not just bread but with milk We need to prepare for a journey too far !   ~ Satba  '),(4538,1359631,'06,March,2004','2004-03-06 00:00:00','   March 6, 2004 - DIscussion @ Southern Spice. For some it is supersition, for others it is a branch of science, but all agree it is good business. What is this about astrology ? What is it that give astrology so much following, hatred, and envy ? Let us look more into this ..  '),(4539,1359631,'29,June,2004','2004-06-29 00:00:00','    This site contains some good articles and content about cryptography told in a unique way filled with all sorts of games and articles that provides information and at the same time adding the fun aspect to learning. A very good site for people who are interested in knowing more about encryption, etc.   urlLink SimonSingh.net   Dont forget to visit the black chamber where you can learn how encryption was used in the ancient times.   '),(4540,1359631,'28,June,2004','2004-06-28 00:00:00','    Stepped over this site when preparing for a quiz that incidentally contains some questions on Brazil. The article provides information on Rio de Janeiro in a humours way. Looks a bit outdated though esp. on information relating to currency, etc.  Good one if you want to have a laugh !    urlLink Rio De Janeiro, Brazil, Earth   Some favorites ...   Beer is served cold, the beach is served 40 degrees centigrade hot.     During four days the city is turned upside-down and rightside-left. Sometime inside-out also, due to excess of an alcoholic beverage called Batida    Players usually play soccer, while spectators develop boxing, judo, karate, and survival skills.     Weather is hot in spring, very hot in summer, hot in autumn, and quite hot in winter.     You should not take a honeymoon in Rio unless you robbed a bank in the last 3 weeks before your marriage.    Actually, it goes down faster than any other currency known, including Monopoly money.   '),(4541,1359631,'23,July,2004','2004-07-23 00:00:00','    If mythology interests you, then this one needs to be in your favorites ! The site is well organised and well structured. I havent read much, but stumbled on it when I was searching for something related to Greek mythology. Explore it out !     \"Timeless Myths is a website containing a collection of myths and legends, from ancient Greece and Rome, and from frozen Norse lands and the magical Celtic realms. Or follow the medieval adventures of Arthur and his unconquerable knights.\"      urlLink Timeless Myths   '),(4542,1359631,'05,March,2004','2004-03-05 00:00:00','    Spain â€“ Time to Say â€œAdiosâ€  It is time to say â€œAdiosâ€ to the travelogues that I have been penning down, to the wonderful place on earth called Spain. The last week started out as another busy week. But this was the week when I had to do my shopping, attend some of the farewell treats and also to catch my flight back home.   One thing that I have been doing constantly is basically trying different kinds of food. This week it was time to try out some different kind of breakfast. One of such breakfasts that I had was a French specialty, a couple of break slices with a slice of ham topped with a specially prepared milk cream. The cream provided the needed â€œkickâ€ to the breakfast that made it one of the best during my stay in Spain.  During my 3 week stay, the â€œEstacion de Servicion Arjeâ€, a small cafÃ© opposite to the place I work, had become the most visited place by me for food. This was mainly for lunch and breakfast. The Salmon sandwich had become one of my favorites. Apart from this, â€œCorralâ€, which is basically chicken and salads sandwiched between long crispy bread (that resembled more like hot dogs). Another favorite of mine here was the Paehllia Bahilla (I am not sure if this is the correct spelling, but it sounded like this). This resembled more like our tamarind rice mixed with some pieces of squid.  The last dinner that I had during this trip was in â€œSamar Kandaâ€, a romantic restaurant beside the artificial tropical forests in the Atocha Railway Station. I had been informed that this was a Kazakhstan or Uzbekistan restaurant. They had an English menu and one of the dishes that attracted my attention was â€œOstrich Meatâ€. My hosts also informed me that this was really good and is infact much softer that chicken and does not cause any stomach related problem. Even though I had to catch the flight back home in another 9 hrs, I decided to take the risk as this was once in a lifetime opportunity to try Ostrich Meat. The Ostrich meat was really good, but I am not sure if it was softer than Chicken as I had to chew a lot. Anyhow it is worth tasting and if enough amount of spice is added it would taste really great.   By the end of the trip I had managed to add Ostrich, Squid, and Ostrich to the growing list of animals that I have eaten in my life. This trip is the one where I managed to taste foods from different regions round the world, from Peru to former Soviet Republics. I would like to keep writing about the different kinds of food that I had during the trip but it is time to say â€œAdiosâ€ to the different kinds of food there and say â€œhelloâ€ to the wonderful Indian food that never fails to satisfy your taste buds.   There was no other major experience during the last week. Towards the end of the week I was more worried about catching the flight early in the morning. However a couple of experiences with the Spanish culture help me shed my worries for some time. For one last time before I left Madrid, my hosts drove me through the center of Madrid at around 1:30 in the morning of Saturday (4th October, 2003) after a very good dinner. I had just 3 hours left to check out and return back.   Journey Back   The journey back was without much of an incident, the only accident being my decision to watch a complete hindi movie â€œSupariâ€ starring Nandita Das. In this movie the hero keeps repeatedly telling his friends, after half of the movie is complete, that it is wrong to kill someone. That was a great message right !!?? But inspite of this his friend keep bumping people not for money but because one of them wants to sleep and the other wants to forget his dad and his typewriter. [Good Excuses for bumping people off]. Anyhow at the end of the movie, all the friends of the hero are killed, and the hero had already killed around 5 people, 1 person before he delivers the message of the movie to his friends and 4 after he had delivered the message. But even after murdering 5 persons, he is pretty confident that he would definitely come out of jail and hence requests her to wait for him. A â€œMustâ€ watch movie for those who have just a frustrating day at office. For the rest of the journey I preferred to sleep.    Looking Back â€¦  Looking back this is a trip that I would remember for â€¦  -	The trip that I enjoyed the most -	The trip that helped me in rediscovering myself -	The trip which helped me taste food from different regions of the world -	The trip that helped me in knowing more about the Spanish Identify, the Spanish people, Spanish art and architecture and also about Spanish culture.   The last 3 weeks have definitely gone into as one of the most memorable weeks of my life never to be etched out of my memory forever. As I bring this travelogue to an end, I would consider the following events as the best moments in Madrid.  [3] Lunch at Estacion De Servicion  [2] The Visit to the Salons at Palacio Real [1] The Bull fight  Yes, I cannot also forget the beautiful female in the Metro. But then with the passage of time, her face would keep getting blurred until all that I would remember would be the event rather than the face itself. Anyhow, someone, somewhere will be waiting for me, so why bother about the ripe grapes J  I would finally like to sign off with the following words.  â€œ All journeys have to end, end in the by lanes of memory that can be visited in the journey of thoughts without the constraints of time.â€   Signing off before going to sleep after a memorable journey, â€œThe Journey to EspaÃ±aâ€. Sathish on October 6, 2003 at 22:05 hrs '),(4543,1359631,'05,March,2004','2004-03-05 00:00:00','    The Matadors and the Bulls     One of the main items of interest to me when I embarked on this trip to Spain was the bullfight, something that symbolizes Spain. From day 1, this was one of the events that I was eagerly looking forward to. On the first day at office, my colleagues at the Spanish office showed total disinterest for the bullfight and were surprised that I wanted to see it. Well, I was in fact surprised that they werenâ€™t interested in something that represented Spain.   Today, I decided to go and witness the fight between the Matadors and the bulls. The arena (Plaza de Toros) was located at Ventas, in Madrid and was a magnificent stadium that looked similar to the Colosseum in the Gladiator movie, though not matching it in size. For me bullfighting was a contest between the bulls and the matadors, which the matadors always won. I had expected that the matadors would â€œcheatâ€ the bulls for few minutes and once the bulls were tired and given up the fight would be complete. But I was in for a surprise.  At 6:00 in the evening, the matadors entered the arena and saluted the box were the king would have sat in the olden days. Once this formality was completed, 4 matadors, one in each corner of the arena, waited with pink clothe (donâ€™t know the exact term) for the bull to make the entry. As the gates opened, the first bull came charging in and the matadors kept inviting it and then â€œcheatingâ€ it. This continued for few minutes, by which time the bull had made around 3 â€“ 4 rounds around the arena. The trumpets blew once again and a couple of men riding on armoured horses with long sticks (I initially assumed it was a stick, but it was more like a spear with a sharp point). The bulls were then made to attack the horses and at that point, the men riding the horses, inserted the spear on to the bulls back. I was initially shocked when I saw blood oozing out of the bulls back. This happened a couple of time and then the matadors, came out with pointed sticks. The bulls were made to charge towards these matadors and in a skilled manner, the matadors inserted these sticks also to the bulls back. At the end of around 10 minutes, the bull had blood oozing from the holes on its back. I began to realize that the bullfighting was not just about winning it was about killing!   After 10 minutes, the chief matador made the entry. After completing the formality of saluting to the imaginary king, this matador was handed a sword and the red clothe. At this stage all the minor matadors moved to the back stage. It was now a one to one contest between the bull and the chief matador. The matador demonstrated his skills to the applause of the crowd gathered there. After about 10 minutes, the matador replaced his sword with another sharper one. It was time for the kill. After another couple of rounds, the matador positioned himself straight in front of the tired bull and pointed the sword at its head. In one quick motion, he had the sword inserted completely into the region between the bulls neck and back. One could only see the handle of the sword. But the bull had not given up. Skillfully, the minor matadors removed the sword from the bulls back, and the chief matador again went for the kill. After the third attempt, the bull moved a few feet and fell on the ground â€“ DEAD! The Spaniards were clapping to the victory of the matador over the bull. Probably around 1000 years back, people used to applaud similarly to the sight of the victorious gladiators who would be standing beside the people they had killed over a sport that meant nothing to mankind!   Another matador came and similarly but less skillfully, killed another bull. The third matador was something of a special. There was a general excitement in the air when this matador made his entry. People were cheering and applauding his every move. His name was Ismael Lopez and he was a star of the crowd and he did demonstrate why he was one. The bull was made to run in circles around him many number of times and this was something that others had not done skillfully. Finally, when it was time for the kill, Lopez had the sword inserted in the right spot at the first attempt. The bull was vomiting bottles of blood and running around. There was blood all around the arena. People were standing and cheering and Lopez was running behind the bull pointing at the bull in a symbol that meant, â€œHe had done itâ€. The bull tried to force its entry through the gate through which it had entered the arena. But it was in vain. In seconds, the bull dropped dead. The Spaniards gave a standing ovation to Lopez and at the same time they took a white kerchief from their pockets and waved at him. There was applause all around. The display of white kerchief by the crowd meant that they appreciated his fight and he was awarded a medal by the chief organizer. Probably in those days, there would have been a princess who would have liked his act and probably would have considered other options for the night!!    Death in victory becomes a sight of joy!    I couldnâ€™t appreciate it anymore and hence left the stadium and found that many others were also leaving. Now I realize why my Spanish colleagues didnâ€™t want to join me for the bullfight.    Experiencing Life â€“ The Royal Way   Last week, I had mentioned about the Royal Palace that had about 2800 rooms. I decided to visit the palace and experience the Royal Life of Carlos, Alfonso, Fernando, Isabella, etc  The Jardines of Sabatini This has nothing to do with the cricketer â€œDouglas Jardineâ€ or the tennis player â€œGabriella Sabatiniâ€. â€œJardineâ€ is the Spanish word for â€œGardenâ€, and these were the gardens which were â€œarchitectedâ€ by Gabriel Sabatini who was one of the chief architect of the Royal Palace. He was responsible for designing most of the rooms in the royal palace and hence the Garden was named after him. There is nothing much to mention about the Gardens, except that it is located in the Palace courtyard, has some fountains (the Spaniards are crazy about fountains!) and has some flora so that it can be called a garden!     The Armory   The palace is a massive structure and on entering inside there is a large square where the soldiers would assemble in those days for a parade. The armory and the pharmacy was located on the sides of this massive square. The armory has a very good collection of (you guessed it right!) armors J and this includes the massive protection gear for both people and horses that you would normally see in Hollywood movies.     The Pharmacy   More than the armory, I liked the pharmacy â€“ one because I felt that the Salar Jung museum in Hyderabad had better armors and secondly because it was for the first time I was seeing a well structured pharmacy, where you could easily retrieve all kinds of herbs. There was also a distillery, weighing scales, leather boxes used to transport medicines across continents and also during the travel of the king. There were so many herbs that I was left wondering that either this pharmacy was also open to the general public or the royal family was a â€œsickâ€ family J. Probably, just like we have a section in our projects, this might been a necessary part of the risk management for the Royal family.     The Royal Salons   The entry into the main palace was through a staircase that is very similar to the ones we would have seen in old hindi or tamil movies â€“ the only difference is that this is real and much more grand. I was first led into a room that was dedicated to a chief of the Royal Guards, which protected the royal family and the palace in those days. This person was responsible for the formation of the Royal Guards. This room led me into another hall which was a banquet hall used by the king to throw parties. This hall was also used to sign Spainâ€™s accession to the European Community in 1985 and was the place where the Middle East peace conference took place in 1991.   Some of the highlights of all the Salons were â€“ Costly (it has to be costly!!) carpets, paintings hanging everywhere, painting on the ceilings with a theme for each room, etc.   The third room was the best of all, a room that every palace would have. This was where the thrones were present. This was the best ever throne that I had seen in my entire life. I just felt like running towards the throne, sit on it and then proclaim myself the king of Spain. Even if it were for a moment, I would still be one of the few who would have sat on the throne. If luck was on my side, I could also become one of the very few to have farted on the throne J.  The steps leading to the throne were flanked on both sides by roaring lions that had the world at their feet. As far as I have noticed, the lions are the royal symbol of Spain. Any monument linked with the royal family would definitely have a lion somewhere.  The painting on the ceiling of this room was a special one â€“ it represented the Spanish kingdom stretching from America to Philippines (via Spain of course J). There was the kingdom of Valencia, the Indians (Red Indians), the America discovered by Columbus, the Latin America, etc.  I was also led to the dining room, the reading room, the dressing room and the bedroom of the Kings. The dinning room had a massive table that could probably seat about 100 people. This table was made to celebrate the marriage of one of the Kings. Other rooms that I visited include the billiards room, smoking room, etc. There was also a section where Isabella De Catolica and her husband Fernando De Catolica lived. Isabella is one of the most famous queens of the Spanish kingdom. I am not sure if there were many Isabellas, as the Spanish kings generally have either Carlos or Alfonso as their name in alternate hierarchical sequence. However you can find many references to the queen Isabella in Madrid. Infact one of the hurricanes that struck USA recently was also named after her! ;) Anyhow, this section was temporarily closed and hence we could not see the life Isabella had lived.    Temple De Debod   What is the link between Egypt and Spain? This is the question that came to my mind when I reached this temple. The Temple De Debod (Template of Debod) is an Egyptian temple located near the palace, overlooking the ravine bordering Madrid. This temple was build some 2000 years ago, not in Madrid, but in Egypt. In recent history, there was danger to some of the Egyptian Shrines because of certain invading forces and Egypt approached Spain for help. Spain obliged and the shrines of Egypt were protected from the invading forces. As a token of thanks, the Egyptian Government decided to gift this shrine to Spain. In this manner the Temple of Debod was reconstructed in Madrid in the year 1968. There is nothing much in this shrine, except for some rock carvings depicting some King and his life. I was not interested and decided to help out some girls who wanted their picture to be taken in front of the temple. Since it was a temple, probably they wanted the photo to be taken by a person who was coming from the land of Ahimsa, Land that gave birth to religions like Hinduism, Buddhism, Jainism, Sikhism, blah blah â€¦ The bullfight is still in my head!!   Search For the Missing Indian  There is a famous story in Kerala. When Neil Armstrong set his foot on the moon and was about to proclaim himself as the first man on moon, he saw a smoke coming from a distant place. He decided to investigate the same and hence moved in the direction of the smoke. To his surprise, he came across a tea shop run by a Malayalee. Forget Malayalees, Indians are supposed to be everywhere. At least that is what I was imagining. But in Madrid I havenâ€™t come across a true Indian. Why true? Read ahead! This weekend after receiving the directions to some of the Indian restaurants in Madrid from one of my colleagues, I decided to search for the missing Indian. On Friday night, I decided to check out an Indian restaurant called Delhi. Unfortunately, after about 1.5 hrs of searching around, I couldnâ€™t locate this restaurant and hence opted for another Indian restaurant called Adrish. This restaurant had proper Indian menu but it was run by Spaniards with a Bangladeshi Cook. The food was also a disappointment. I ordered for Chicken Madras expecting that it would be similar to Chettinadu Chicken. However it was chicken is tomato sauce!   Today, I finally managed to find the Delhi restaurant. Infact this happened accidentally. On Friday I had called up the Delhi restaurant a couple of times and asked for directions but still couldnâ€™t locate it. Probably having missed a customer, they had decided to make corrections. After visiting the palace, when I was walking towards another restaurant, I came across an Indian (looked like an Indian) who was fixing a new board that provided the directions to the Delhi Restaurant. This board was much more clearer and probably helped them improve their business as they had an increased crowd for lunch today (This information is from what they told me). The Delhi restaurant had a better ambience. There was also Songs of Nusrat Fateh Ali Khan, which included the â€œAfreen, Afreen, â€¦â€ song. One of the specialties of this restaurant was the option for the customer to customize the food according to his needs. This restaurant also offered fish curry and when I was given the option of customization, I, expecting the host to be Indian, asked for fish curry in Kerala style. But to my disappointment, these guys were Bangladeshis and were not aware of something called â€œKerala Styleâ€ L I decided to give up on fish and opted for the prawn curry. But these guys were really good. They came and enquired about the Kerala style and I, who always like to live thinking that I am a good cook, described about the Kerala style of preparation. When the prawn curry arrived, I was totally surprised; they had tried to put into action whatever I had told them except for the turmeric part. There was a good amount of coconut and the prawn curry was fabulous. And at the end of the lunch, they gave me free â€œIndian Masala Chaiâ€ as a token of affection. It has been long time (2 weeks has become a very long time for me! ) since I have had such a good tea. Well, even though they were not Indians, they had put their best to give me a proper Indian Lunch. After thanking them for making my day, I decided to go for the bullfight.    The Story of an Icelandic and an Indian  What is common between an Icelandic and an Indian in Spain? Both are foreigners, both have similar problems communicating to the Spanish people, and both find a companion among each other during work, lunch, dinner, and sight seeing. During this week, Inkimar joined me from Iceland team. He was here for a week for the same project on which I was working. He was 37 years old and had similar problems that I have been with the Spaniards. The main topics of our discussion were the culture and habits of Iceland and India. During the last 2 weeks, apart from my normal work, I have also been promoting India a lot as a tourist destination and he is the second person who has taken a strong inclination in visiting India in the near future. These guys (Europeans) generally travel a lot and that is where I feel they gather a lot of knowledge. Since travel is very cheap here, they can afford it. We Indians should also travel a lot. Only then can we learn a lot about the positive aspects of other work culture, technologies, etc and how we can incorporate the same into our lifestyle. Only then can we become a truly developed nation. But as I have been reading, things are also changing in India. There was an article in one of the online sites, on Indian farmers who have started visiting countries like Israel, Australia, Netherlands, etc to learn about their technologies, process, etc and incorporate the same into our agriculture.    Bagets   I donâ€™t want to drag much, but this is a Spanish food item that I liked. It is more like a sandwich. An elliptical bun is cut into 2 and a slice of stuffing of your choice is placed in the middle. This is served along with potato chips. We had cheese ham, spicy ham, Salmon, tuna and chicken bagets. It seems there are 200+ combinations of baggets.  Before closing down â€¦ as I was coming in the Metro train after the bullfight, there was this beautiful female who was sitting right in front of me. She had a very good smile and was well dressed and a very special glow on her face. She has been one of the very few females that I have admired in my life. The unfortunate part is that She was not an Indian. Anyhow, my station came and I had to get down, probably never to meet her again.   The coming week is my last week in Madrid before I return to India. I was once mentioning to one of my colleagues here during a particular incidentâ€¦ â€œIn India there is order in chaos and here there is Chaos in Orderâ€. I am not sure if they understood my statement as they gave a confused look or probably they like many of my colleagues in Hyderabad, chose to ignore it.  Looking forward to my trip back, Sathish on September 28, 2003 at 11:39 PM '),(4544,1359631,'05,March,2004','2004-03-05 00:00:00','    A busy and tiring week!!  This is how I would summarize the first week in Spain. Busy and tiring because of the amount of work that I had to do. However I still managed to take time of the busy schedule and managed to visit places and enjoy and discover the different aspects of Spanish (Madrid) life.     The Spanish Identity  A very good start to this process of â€œDiscovering Spainâ€ was given by one of my colleagues â€œCarlosâ€. He was a very jovial guy and is from Barcelona. He takes pride in calling himself a Catalonian rather than a Spanish, but then is very proud of Spanish Identity, language, etc. In fact, for me, he represented the face of Spanish Identity.   The Spanish people are very proud about their language. One of the first questions I received from one of the colleagues here was â€œWhat do you know about Spain?â€ and when I mentioned some of the things I know about Spain he also told me that Ricky Martin and Enrique Iglesius also sing in Spanish. His point was that the whole world is listening to Spanish and he hoped that Spanish would become a universal language. At the same time, there is also a realization that English is â€œslowlyâ€ becoming the universal language for communication. Here if you need to survive, you have to know Spanish. For me, since I had company for most of the time, it was not a problem. Moreover one of the first things I learned in Spanish was â€œEspaÃ±ol, No Entiendoâ€ (I donâ€™t understand Spanish), apart from â€œHolaâ€ and â€œGraciasâ€.   One of the best experiences that I had with EspaÃ±ol was in the McDonalds. I was in the queue and the nearby queue became empty. A lady who had just come offered me the empty queue and I obliged immediately. The person at the counter knew only Spanish and She kept on telling things in Spanish. Finally I had to release the only mantra that I knew â€¦ â€œEspaÃ±ol, No Entiendoâ€. Well after bit of sign language and help from another person who knew decent English, an order was placed. In the mean time, the person who offered the queue to me, called the manager of the place and was mentioning something and repeatedly pointing at me. At the end of the shouting session, she gave up and left the place without having lunch. The only thing I could do was to keep quiet and when she had left, I asked the person at the counter â€œAny Problem?â€ and She said â€œNo Problemâ€. Anyhow I still donâ€™t know exactly what the problem was, but what I had interpreted from her body language was that She had offered me the queue as she had just entered and I was already in the second queue (Spanish Culture) . However when the second queue became empty, an American who had just entered did not offer her the second queue (American Culture). She was really pissed off!! Hope she had a nice lunch J  The Spanish also are not that very fond of their neighbours, i.e., French. They believe that EspaÃ±ol has wider reach than French (which may be right if we include Latin America) and they also believe that Spanish wine is better than French (may be some wine expert can comment on this).   Madrid â€“ City of Fountains, Restaurants and Castles  Madrid is a city famous for Fountains, Restaurants, and Castles. If 4 or more roads meet at a point then you can be assured that you would find a fountain there. And this fountain would definitely have significance. For example, an important road in the city is named after a fountain that is located along that road.    Madrid is also famous for its night life and the enormous number and variety of restaurants that you can find. If you are in Plaza De Cibelles, all you have to do is to go to one of the by-lanes and knock on any door. It would definitely be a restaurant. These are not typical restaurants, but are more famous for â€œtapasâ€ i.e., along with a glass of beer or wine, etc have some snacks. The Spanish dinner is an extended one. Generally they have tapas around 9:00 â€“ 10:00 pm and then have a late dinner that would extend to 12:00 PM. Most of the Spanish dishes contain either Meat, Sea Foods or Pork. Vegetarians would definitely have a tough time surviving here.  I have been eating mostly sea food and have tried squid, Salmon, tuna, octopus, etc.   Madrid is a very old city. Except for a brief period, it has been the capital of the Iberian Peninsula / Spain for many centuries. As a result any tourist can easily trace the gradual development of the city. This is more visible in Castellana (where most of castles are located) that houses the Spanish Congress and most of the ministries and offices like the Central Post office, Central Bank, etc. All these buildings were constructed a couple of centuries back and are good examples of Spanish (European) Architecture.  There are also many modern day â€œcastlesâ€ and these include a Europe Tower designed by the famous Japanese Architect who had also designed the WTC towers in New York.    Typical Spanish Dinner  On Tuesday (September 16, 2003), Carlos and myself decided to go for a typical Spanish dinner. At around 9:30 we entered one of the places where they serve you tapas and he had a glass of beer and I gave him company with a glass of Coca Cola. We also had a slice of salted ham that was well cooked and tasted good. At 10:00 we again took a long walk and then walked into a restaurant at 10:30. We decided to go for 2-course dinner. As part of the first course I tried a platter of seafood, which included octopus, Salmon, tuna, prawns, etc. The second course included a stuffed squid.  In Madrid, I havenâ€™t seen a single Indian yet! However in this restaurant, there was a Spanish girl, Soledar (something like that) meaning lonely.  She worked in the restaurant and served us the food. The first thing She asked to my colleague on seeing me (She knew only Spanish and hence couldnâ€™t communicate with me directly) was whether I am a Hindu. She then showed me a chain that she was wearing and asked if I recognized it. It was a â€œRudrakshâ€ chain and she said that she was a devotee of Sai Baba and had got that chain in a temple. Well that was the one of the very few Indian Connections that I have encountered so far here.    Spice Contest â€“ Peru Vs India  One of the major complaints that I have been having about the food here is the amount of spice. It is totally zero! My colleagues here decided to take me to a place where the food would be really spicy. It was a Peruvian restaurant, and Peruvian food is supposed to be very spicy (or at least that is the impression they have created). I decided to take on the challenge and opted for the spiciest food in that restaurant. It was a rice-based food, with lots of sea organisms mixed into it. Well it was not a contest of evens, as the Spiciest Peruvian food could only be compared to our sambhar in the amount of spice. Anyhow it was a different experience and I really liked the Peruvian food.    Annapurna â€“ Indian but not really Indian   On Friday, I had given up. I wanted something spicy and decided to look out for an Indian restaurant. There are very few Indian restaurants and one of the popular one is the Annapurna. The name itself brought joy to my stomach and I took the metro to Alonso Martinez and from there I walked for about 10 minutes before I reached the restaurant. I expected to meet some Indians over there, but then even though the name was purely Indian it was run by Bangladeshis ! There was a Bangladeshi cook, Ali, who reminded me of the Bangladeshi cook in the serial â€œTandoori Nightsâ€. The food was anyhow Indian. I decided to go for a Paratha, Murgh Jalfrezi and a plate of Chawal with a glass of lassi. Finally my stomach had got what it wanted. The food was spicy enough and I left the place satisfied enough. It is an expensive place, but I think I will visit this place one more time before I leave Madrid.    Museum De Prado   One of the many highlights of Madrid is the Prado Museum. It is supposed to be one of the 3 largest Museums in the world (in Paintings Category). It has a large collection of Italian, Spanish, German, French and British paintings and sculptures. Though I hate seeing paintings, I decided to visit it just for the sake of not missing it. It is more like tourists wanting to see Charminar. After around 30 â€“ 60 mins in the museum I decided to take the bus trip of Madrid. Madrid Vision â€“ Discovering Madrid  The Madrid Vision is a bus trip, wherein I can take 3 routes to see the complete City. I decided to take 2 routes as you the 3rd route contain the places covered by the first and second route.   The first trip that I took was that of the Modern Madrid. The main aim of this trip was to visit the Real Madrid stadium. Luckily there was also the English commentary and hence I could get to know more about Madrid like the place where the rich lives, the embassy area, the parliament, the post office, stock exchange, etc  The second trip was more historic, and the places they covered included the Discovery Park (this is in dedication to the discovery of America by Christopher Columbus. There is no mention about India at all, although it was a case of mistaken discovery).  Other places included the Royal Palace (containing around 2800 rooms), the Teatro Real (Opera house), the town center (where the 0 km mark of Spain starts), the Plaza Mayor (Mayor in Spanish means big), some historic fountains, etc.   Even though it was a tiring one, the trips were worth taking and this gave me a better insight into the history of Madrid. The city itself was founded by the Arabs in the 700 AD when they conquered the Iberian peninsula and built a fortress here and called it â€œMadhireedhâ€ (something like that) meaning the land of water / fountains J   Satiago Bernabeau â€“ Inside Real Madrid C.F.  The Real Madrid Football club is situated very close to the place where I stay. The stadium is on the other side and is called â€œSantiago Bernabeauâ€. On Tuesday, there was a match going on between Real Madrid and Marseille. Outside we could here the crowds roar when Ronaldo scored a goal for Real Madrid. It was a great experience. If possible I need to go for a match.  The stadium Area had a large collection of trophies right from 1904 to last years Champions Trophy. I also decided to get a small memorabilia from the stadium shopping center.   Retiro   Today, I again took the Metro and reached Retiro. This is a famous park in Madrid and along with another park forms the lungs of the city. The park has some famous fountains and a â€œlargeâ€ lake [donâ€™t even compare with Durgam Cheruvu. It is much smaller than that]. This lake is more like a beach for the Spanish, and you can find many people taking a sunbath. There is also a Crystal Palace in the park and is famous. But all I could see was a glass palace with some glass tables and some speakers around it. There was a booklet in Spanish that described its significance, but then I canâ€™t read Spanish!! If you have visited either the Lal Bagh or the Ooty Botanical Gardens , etc then just forget this park. It is not worth visiting. The only highlight was some good Latin music being played by some local people.   City of Plazas  Madrid is also a city of Plazas, Plazas in Spanish means Square. So you have the Plaza de Castilla (Square of Castilla), Plaza de Cibelles, Plaza Mayor,  etc. Most often you will find a fountain in the center of the Plaza.   The Latin Connection   Latin America was once a colony of Spain and that connection is very evident in Madrid. You can find lots of Latin Americans and also places that are named after countries / places in South America. There is a district in Madrid called â€œLatinaâ€ and this has places like Paraguay, Colombia, Republica Argentina, etc.   Other Highlights of Madrid    Puerta De Sol [Gate of Sun]  This is where the town hall of Madrid is located. This place is significant because this is considered to be the center point of Spain. The 0 km mark starts from this place. This was constructed around 300 â€“ 400 years ago. Recent studies by satellite have shown that the people in those days were nearly correct as the exact center of Spain is located just 10 km from this point. This point has around 10 (probably more as I saw only one half) roads leading in a circular fashion. It is believed that all roads in Spain start from this point.     EspaÃ±ol â€“ Fast Track    	Sala â€“ Not the hindi â€œSalaâ€, but it means â€œRoomâ€.  	Salida â€“ Exit. Follow the â€œSalidaâ€ signs blindly when you are going out of Metro or other places.  	Chicala â€“ Meaning girl. Probably the work â€œChicâ€ had its origin from this word!  	Libre â€“ Means â€œFreeâ€. If you need to catch a taxi, check if the board near the driver says â€œLIBREâ€.  	Occupado â€“ Occupied. If taxis are not free then it would be â€œOCCUPADOâ€  	Perfecto â€“ Meaning Perfect. Have been hearing this and Correcto lot of times, especially in television.  	Correcto â€“ means Correct.   	Naranja â€“ This is pronounced more like Malayalam for lemon and it means the same J Strange! It also is used for Oranges.  Another busy week starts in 7 hrs!   Chao, September 22, 2003 1:05 AM '),(4545,1359631,'05,March,2004','2004-03-05 00:00:00','    Spain, the land that is famous for many things, things that I am yet to find out. But as many of my friends have been telling and also based on the very less knowledge about Spain that I have, I was sure to discover new and interesting things. 	 My journey started from Hyderabad on September 13, 2003. I am undertaking this trip for business reasons. Hence the time to â€œdiscoverâ€ Spain is very limited. Around 500 years back Christopher Columbus set off from Spain to discover India and now I am moving to â€œdiscoverâ€ Spain.    Hyderabad â€“ Mumbai    I had to travel by the 9:15 flight of Jet Airways. It was a Boeing 737 aircraft and unlike my previous experiences and that of most of my colleagues, the journey was very smooth and not like traveling in an Auto through the lanes of Chennai. To top it all the Airhostesses were too good and caring. There was this hostess, Swati, who caught my attention for most of the journey. She had a different hairstyle that made her forehead look like the â€œohmâ€ symbol (if you drew the line along the boundary of her forehead skin and hair). The one factor that I would appreciate in this whole journey to Madrid is the waiting time. The waiting time was just appropriate to refresh yourself and make the move. Earlier there used to be so much waiting time that you would get totally bored waiting for a flight that would take off only after around 6 hrs.    Mumbai â€“ Paris Charles De Gaulle [CDG]   On landing in Mumbai my first job was to get a reliable transit service from Mumbai â€“ Hyderabad. The experience of Jan Westrate at the Mumbai airport during transit was fresh in my mind. But to my joy I found an Air France representative waiting for me and other passengers. Even though I managed to find the Air France representative, She (Yes it is a She and her name was Rupali and She couldnâ€™t be better) couldnâ€™t find my name in the list of passengers. After searching for a while in the list of names, we finally managed to dig out my record using the point of destination (Madrid), as I was the only passenger in that list traveling to Madrid. Air France had arranged for a separate bus (similar to KLM).  As I boarded the Air France bus, a well-known face was staring at me. It was Manikandan, ex-colleague who had worked in my team earlier. He was traveling to London via Paris. He said that he was going for â€œshort-termâ€ which was 3 months long. I basically felt the need for re-defining the terms â€œshort-termâ€ and â€œlong-termâ€!!   I was probably one of the very few who didnâ€™t have a hard copy of the ticket but still had a ticket that was re-confirmed. I met Clara (J) at the Air France counter and she too was a bit surprised that I didnâ€™t have a hard copy! However because of the help of Rupali I had already managed to get an early check in done and hence all I had to do was collect my boarding pass and return tickets.   The Air France flight (Airbus 340) was decent but not as good as the Austrian Airways flight for several reasons.  -	The push back was very limited. Austrian Airways had a much larger scope for â€œpush backâ€. -	There was lot of (relatively) dust in the aircraft. -	The blankets had lot of cotton dust. -	There was a movie of Esha Deol and Sanjay Khanâ€™s son!! Most of the other movies were not that great.  Still the flight was a very pleasant one. The landing at the CDG was one of the smoothest landings. During the last 3 hrs of the flight I once again got the sneezing problem because of the cotton dust. I had only 4 hours of proper sleep. My company through out the flight was Gopi, a Malayalee from Palakkad. He was going to a university in France for doing some post Ph.D research.   The CDG is a very big airport. But I felt that Amsterdam was better. Probably I havenâ€™t seen much of CDG. However the airport Architecture looked good. Everything that you need was within 15 min of your reach if you take the proper means and for that you need to properly interpret the signs, which I did not do initially.     Paris â€“ Madrid Bajaras   It was again an Air France flight and this was much better, with very good food, beautiful Airhostess and a sleeping babe for company, who slept throughout the 1.5 hrs of flight! There was one food item that I liked a lot. A Croissant with Apricot Jam stuffed inside. It was â€œdeliciousâ€ and I loved it. Need to check out if I can get one here somewhere. Else need to wait for the return flight.  The Aueropuerto De Madrid Bajaras, is not that great compared to Schipol or CDG or even Mumbai. It was a simple airport with the basic needs. Marcelo Cury, my host in Spain, was waiting for me and we took a cab to the hotel.     HUSA Chamartin   This is where I am staying. It is a four star hotel and has a mini bar, a safety box, blah blah. The hotel is pretty decent. The Metro runs right below the hotel. Infact this hotel is a property of the Railways of Spain.     First look at Madrid   The city is a very calm city. It was also a Sunday and nearly every shop was closed. In Spain, most of the shops are closed on Sundays and the Saturday of the first weekend of the month. Some of the other features of Spain are,  1.	Lunch is taken for 2 hours unlike the 0.5 â€“ 1 hr period back home.  2.	Restaurants are open only from 1 â€“ 4 for lunch and 9 â€“ 12 for dinner.  3.	Shops generally close early at around 8:30.  There was one aspect that caught my immediate attention. There were posters stuck on the walls very similar to the film posters that we would find in Chennai or Coimbatore or any other town / city in India (except Hyderabad J ). Madrid looks more like a costly City to live in. The minimum charge for a decent dinner is â‚¬ 10.00 and everything is costly. Probably it is more to do with the place I have visited today. May be things are cheap in other parts of the city.  The time now here is 9:00 PM and the Sun is just setting down. Have to prepare for a busy week ahead.  Closing down, September 14, 2003 20:55 '),(4546,1359631,'16,April,2004','2004-04-16 00:00:00','    Why do we sleep?   Looking back, it was a usual night. During the day I had been part of the team evaluating a couple of design tools for object oriented design. The evaluation was proceeding smoothly and it was one of those days when life seemed to be moving slowly and suddenly there was plenty of time to look into the where my life stood in the road to the future. It was not a road to perdition but it was hazy and the path forward was not clear.   It was a long time since I had seen some unusual dreams, dreams filled with creativity that presented exciting scenarios that rarely happen in our day-to-day life. Sometimes, dreams help you achieve the impossible, though it only happens in dreams. I have often felt that I am at my best whenever I have such dreams. It had been a long time and life was at its lowest ebb and then this dream happened.  Life, as a whole, presents a lot of perplexing questions. There must be some reason, I thought, as to why we humans need to sleep everyday. The very purpose of sleeping daily! It could have been weekly or there could have been some other recharge mechanism. But why sleep? I thought to myself, I sleep for about 8 hours a day. If I could use those 8 hours for some other activity, then would my life have been different? Would the world be a different place? Probably there could be some ways of achieving the impossible. 1000 years ago, no one would have even imagined about the various technological gadgets that we take it for granted as part of our daily life.   As I kept wondering about the need for sleep, I began to realize that my body was not a single piece of human body. It was abnormal or probably it was the way all humans were. My body was made up of various objects and I could see that classes defined these objects and the classes were associated using UML style relationships. It was definitely not a comfortable feeling. I had known myself as being represented by a single framework made up of skeleton and various muscles and organs and covered by a single lining of skin that made the entire body look like a single component. The realization of the new representation of my body gave me a worrying feeling. I wanted to do something about it. I wanted to assemble all the components and deploy it as a single unit. But I just didnâ€™t know how to do it. Was there a tool? Was there a physician who could assemble it? If yes, how do I find the tool or the physician? Do others also experience the same? I was deeply into finding ways to see myself as a single unit of representation. Then this thought came to my mind. Probably this is the correct representation of the human body. And that could be the reason why we sleep. When we sleep, the body was getting un-deployed and necessary corrections and fine-tuning was being done to the components and by the time we are through with the ritual of sleeping, the components get deployed again. That is why we need to sleep daily as the human components are not scalable beyond a day without loosing performance. The human system probably needs to get restarted once a day so that it does not loose its performance.   A strange feeling came to my mind. I wanted to view the objects and their classes and their relationships so that I could study the architecture of the body. Even though I had realized the same, I couldnâ€™t see it. In an attempt to see the classes of my body, I opened my eyes to view what could be the best architecture ever known for mankind. As I opened my eyes, I was getting out of my sleep and the components were getting deployed and the system was restarting. My eyes were wide open and the world was at it was earlier. My body was still a single unit of human body. The components were not visible. I was confused whether what I had seen was just a dream or the realization of the unknown! '),(4547,1359631,'22,June,2004','2004-06-22 00:00:00','    Very few things have given me company for long period of time. My second jeans, the first one that really impressed me is one such item and recently after 4 years of public service, I had to retire active service. A painful decision, but it had to be executed.    The Origin of Jeans    From the days of total ignorance about denim, Jeans has become a part of my everyday life. I have been always thinking that Jeans originated in some Texas ranch or as a necessity for the mineworkers during the American Gold Rush. But I was surprised when I got to know that it originated in Italy and the word \"Denim\" is actually a French word. Surprised??!! Read the below paragraph taken from the wikipedia.   Jeans were first invented in Genoa, Italy when that city was an independent Republic, and needed all-purpose pants for sailors in the Genoese Navy. The first denim came from (french:de) NÃ®mes, France ... hence the name denim. Those sailors required a pant that could be worn wet or dry, and whose legs could easily be rolled-up to wear swabbing the deck. These jeans would be laundered by dragging them in large mesh nets behind the ship, and the seawater would bleach them white. The French word for these pants was very similary to their word for Genoa -- this is where we get the term \'jeans\' today.     Stepping into the New World   I remember the day when I along with my friends had gone on a \"selective cloth hunting\" spree to Shopper\'s stop. The aim was to grab as much as possible within the limited budget of each individual. I had selected the jeans sanctuary in the 3rd floor. It was filled with denims of all the popular international brands - Lee, Levis, Wrangler, Pepe, etc. Jeans definitely was not my area of specialization. The only jean that I had purchased earlier was during my school days and it definitely was not a experience worth remembering. The brand was called \"Currency\" and I always felt that my weight increased by around 5 kgs whenever I wore that. The shelves were packed with all varieties of jeans, in different colors and fits. I just hoped that I wouldn\'t get lost in the denim jungle! Suddenly a voice emerged from one of the shelves .   \"Pssssst !! Here ... look at me! I am the original one. The rest are fakes. Just check me out\"   It was none other than a Levis jean trying to grab my attention. Unfortunately, the other jeans had heard the Levis! From the neighbouring shelf there was a loud laughter. It was Lee trying to divert my attention.  \"Levis!!\" shouted Lee accompanied by the laughter that sounded more like the ramming of doors during a tropical thunderstorm.    \"Levis Strauss what not!! Such a pathetic name that would make the medicines sick!\"  The tone was particularly harsh that made Levis look totally lost and depressed.   \"Look at me!\" exclaimed Lee with a tone that had changed from harsh to one that was more aimed at giving an impression of being a statesman. \"I am Lee and you would always be glee with me!! Lee ... Bruce Lee ...Brett Lee !! Don\'t I sound familiar?\" I knew that it was a poor joke and being an expert in that field, I dint want another buddy with similar traits! I knew Lee was not the one. I looked around and found that Pepe was busier trying to break away from the wrangles of wrangler! The decision was simple. I wanted to give a lift to the exponentially decreasing morale of Levis and give it the life it deserved, away from the company of Lee and Pepe. Without any hesitation, I grabbed the Levis one and in minutes I was going on my way back home with Levis breathing the fresh air of Hyderabad, outside that of shopper\'s stop. Levis had just stepped into active public service.   When life Changed !    I donâ€™t remember when I actually started liking it, but all I remember is that I was not sure as to how and when to use it. Yes, I always wanted some dress that I could wear anytime and anywhere without having the need to keep changing my attire for each and every occasion or every quarter of the day ! Even though I had purchased the Levis jean by chance, it definitely had a great impact in my life. Slowly it was the preferred option to work (casuals are allowed only on Fridays :) ), dinner, home, outing, drive, travel, etc. Apart from the 4 working days of the week, I never felt the need for going for another option. It didnâ€™t matter if it was raining or the temperature was low enough to freeze your bones or whether it was a scorching day with a feeling of being grilled live, it always made me feel comfortable. With all the patience it ever had, it tolerated all the dirt, dust and roughness that it was subjected to! It also gave me company during the chill Christmas nights in Copenhagen, or during the nervous days spent in Netherlands after losing my passport and also when the matadors were taking the bulls by their horns in Madrid.  Life was as comfortable as it could be!    Time and toughness did have its toll! The charm of the youth was being lost gradually as it was losing its shine in a slow process. The frills of old age began to appear and dirt was slowly eating into its life. A stage had reached when I had to take a tough decision!   It wasnâ€™t an easy decision! The hard work had its toll and it was necessary to amputate the lower folding so that it could be saved from further damage. This meant that continuing the active service would be difficult, as it would loose its immunity to the constant wear and tear that it was being subjected to! But the options were limited and amputation had to be performed. With amputation done, it was time to retire it from active life. It is often said that the only constant thing is change. It was true in this case also. It was time to recruit new ones and I decided to get in 2 new ones into service so that the workload could be effectively distributed.   The levis jean would still be in limited service and would definitely be preferred over the young guns in occasions like games, limited outing, etc. It always was an effective companion and would remain to be so but with limited usage.  It definitely is the first jean that I liked and would remain to be so ! '),(4548,1359631,'15,June,2004','2004-06-15 00:00:00','    You never know when boredom strikes you nor can you predict when someone would score a goal in a soccer match!  Fortunately or unfortunately, both of these struck me this week and it has the potential of driving me crazy!     The Goal Part â€¦  It was fun and frustration to watch France snatch victory from the jaws of defeat in the final minutes of the Euro 2004. It was fun because the match was a treat to match with both sides exhibiting quality football. It also made me feel that it wasnâ€™t worthless (also read as senseless) to waste my precious sleeping hours and watch the live feed of the match at around 2:00 AM. I could have been dreaming about some princess in a far off land waiting for me and battling the pressures of life or probably the war of the solar system where my space ship would be engaged in a needless fight to destroy the bloodless aliens from the planet of morons!    It was also frustrating because it took away the opportunity that I would have had had England one. The opportunity to take a good-natured gibe at one of my friend who revels in supporting France! I also had to spend extra minutes in arriving at a counter strategy to mitigate the effects of hearing all the praises that would be showered on France and its players by that friend of mine. All due to Beckham and Victoria!! Probably the effects of personal life affected the performance of Beckham.    BTWâ€¦ Who cares about the English!! The â€œSpanish Armadaâ€ or the â€œAzzuriâ€ are going to rule!     The Bore Part â€¦  In software industry boredom happens to be our lifelong friend. Just when we think that life is at its best, boredom knocks on our door and asks us â€¦   â€œBuddy, isnt it long time since we were together! Would you mind if I joined you in your moments of happiness!!â€   Even before we can say a loud â€œNOâ€, it jumps in and occupies a corner of our beautiful house and tries to make itself comfortable in whatever space available. Thereafter, it is a battle, a battle to get boredom out of our house. If we donâ€™t get boredom out, we will soon find that boredom has entered our kitchen and is cooking something really fishy for us! Beware â€¦ tasting it would be dangerous and could lead to a romance with boredom. Once we begin to like boredom, we will soon find that its friend \"Laziness\" wouldn\'t hesitate to join us! (Does that lead to threesome !!)   Anyhow â€“ boredom also provides us with opportunities. Just cleaned my mailbox today. Something that I wanted to do for months. It can also get the best out of us â€¦ who knows, boredom might just be cooking a new idea that could get me ignited (Ahem!! I meant it in the positive sense). Its Smoking !!    To conclude, just like Zidaneâ€™s free kick that turned the match in favor of France, I just kicked boredom out of my personal life (at least for some time). All my senses are now glued to Euro 2004. But there is much more to follow - Wimbledon, Asia Cup Cricket, Olympics and of course my trip to Coimbatore. It is monsoon time and time for me to take a Monsoon trail and recollect the good old memories that I have about Monsoon. However, boredom still exists in my office life! Probably what I need to do is try for another Goal and get it out!    â€œHey, where is the damn football !!â€ '),(4549,1359631,'28,July,2004','2004-07-28 00:00:00','    July 26, 1999 â€“ This was the day when I joined my first company after graduation and on Monday, July 26, 2004, I would be completing 5 years in the same organization! 5 years have passed by in a flash!   Till 5 years back, my entire life was confined to a limited area around my house at Coimbatore. My schooling and graduation was done at institutions very close to my house. On July 23rd, 1999 when I was leaving the Coimbatore railway juntion in Sabari Express to join this Organization, my father had remarked to my sister, â€œ His independent journey in life is finally starting! â€. I am not sure how far I have traveled in this journey and am not going to do an analysis of the same right now!   On the first day at office, one of the first things that impressed me was the reception! There were various clocks that displayed the times of India, Netherlands, etc. On the left side was a small room with a large coffee vending machine. On the right was a conference room where you could see â€œexecutivesâ€ discussing various strategies! It must have been one of the many boring meetings that I am used to nowadays, but the impression was great. To top it all was the receptionist who seemed to be in total control of the proceedings in that room. This multi national company impressed me!   Trainings were more painful. It started sharp at 0830 hrs and the trainers were busier trying to scare the shit out of us than to make us feel more comfortable. There were only 2 systems outside the training room that had Internet connection and these systems were in great demand during the tea breaks! Finally it was the day when we were allotted our departments or areas of work, and I was among the â€œeliteâ€ gang to be put into the E-Enterprise division. This was a new division and it was in great demand then! From the next day onwards we were in separate training and this was the best part of training. We had total Internet connectivity, were dumped into some â€œjunkyardâ€ kind of room where we made the most of the opportunity and surfed through whatever the Internet could offer us and kept tracking the happenings around the world. It was a hectic period then and the company was not doing well. Our stocks were going down in NASDAQ while those of all other companies were shooting up the sky! From around 20 dollars per share we saw it go down the $1 mark! Many people were also leaving the organization, but I decided to stay put. â€œ I was in the most sought after department. There must be something in it! â€ I donâ€™t want to validate this thinking now, but that was the basic thinking that made me stay put in this organization during the crisis. The organization has since been moving slowly and steadily up the curve after a series of acquisitions and restructuring and this was a great learning experience.   10 years back if I had remarked to someone that I have completed 5 years in the same organization, people would have just shrugged it out and probably would have commented that I have a long way to go! 20 years in the same organization was the criteria when you can start even talking about how long you have been in the same organization. During the dot com boom, the same discussion would have made people look at me with disdain. They would have commented â€œ 5 years!!!! That is a long time to be in the same organization! â€. People would have also wondered if there were something wrong in me that made me â€œstickâ€ to the same organization! However in todayâ€™s time 5 years is something that would give you some decent respect in the industry! At least in this respect, I have caught up with the times!   Let me go back to the journey 5 years back to Hyderabad. At that time, I donâ€™t think I even imagined that I would stay in this organization for 5 years. My aim was to get 2 years experience and then move to a place closer to my family. Bangalore was the place that I was aiming for. But as I completed 2 years in the organization, I had an excusing of staying here for some more time due to the dot com burst! This was more like a moving target, from 2 years it shifted to 3, then 4 years and finally 5 years! I think now I have given up! No more targets, take it as it comes and make decision based on the current circumstances. That makes it more realistic! :)  When we joined 5 years back there were 40 members in our batch. This figure has come down drastically and now only 11 members are left! Every year we celebrate the completion of 1 more year in this organization. Last year we celebrated the occasion at Mount Opera water theme park. However I have wondered if this was really an occasion for celebration. Anyhow you just need to have an excuse for having fun and I donâ€™t mind that! This year though there are no celebrations. People have become busier to find time for an outing. Probably they have better things in life. However I feel this year was the right time for a bash as it is a significant milestone and also because we are getting our gratuity and a gift voucher worth Rs.10000. That was more than a reason for celebration.   On Monday when we would be moving to our sixth year in office, there is nothing much to expect out of the day except for the customary email from the management thanking and congratulating us for blah blah blah reasons! Else it would be business as usual  ! ;-)  To end this blog, let me pen down the top 10 memorable events that have dotted my life as far as association with this organization is concerned. These are events that I feel would be worth remembering or at least worth enough not to be forgotten! As any countdown would do it, I will list it down starting from the bottom.   10. (a)   One month of Internet â€œexposureâ€ at the main building junk room.  We were left to ourselves for around 1 month in a â€œjunkâ€ room and during this period we knew how to use the Internet, what are its pitfalls and pluses and what are the main resources, etc. This was a fun filled period and something that we still keep talking about.     10 (b)   Kingfisher Corporate 5-a-side Football Tournament 2002 for us was the year of football! We used to play every weekend at the Gymkhana grounds here in Hyderabad. Finally when the tournament was announced, 2 teams from our organization took part. It was the first time 2 teams were taking part in that tournament from the same organization. In that regard we were trendsetters. I was part of the B team which met the A team in quarterfinals. Finally a team went on to win the finals against Wipro in a thriller! It was pleasing to see all the hard work paying off finally. I enjoyed it!    9.  Playing  around with Yahoo IM to find my first and only 100% Internet buddy!  I was one of the first in my batch to use yahoo messenger and as part of my experiments I happened to find my first and only buddy who has remained a good Internet friend of mine all these years. However even after 5 years I havenâ€™t met her yet! J   8.   Preparation for Bali Trail  This was a quiz event that my friend and I had organized in the organization. I was involved in preparing the presentation. The preparation and effort reminded me of the college days and was similar to the effort that I had put for Nexus â€™99.   7.   Trip to Mahabaleshwar  In 2001, we organized a trip to Mahabaleshwar from the organization. The place and the trip was great! I especially loved the stay at Panchgani and the table top, as well as the strawberry ice creams at Mahabaleswar.    6.   Weekend Fever â€“ Movie and Dinner @ CafÃ© Bahar For around 1 year or more (2002 â€“ 03) I had a great time with our gang of friends, who were also colleagues. Every weekend you could spot us at either Sangeet or Skyline and this was usually followed or preceded by dinner at CafÃ© Bahar or Paradise. Chicken Biriyani and Tandoori Chicken were our favorites! Now however this has stopped as some of them have left the organization and city and others have become busier! Weekends were never boring!    5.   Trip to Denmark This was my first trip outside India. I had to make the trip alone and was totally ignorant of the practices and procedures that one has to follow during such trips. I still remember the way I used to observe other passengers as to how they were using the controls (for lights, push back, food tray, etc). Somehow I managed to go and come back without any issues!    4.   Approval engine   This was my first major activity for a product. I had to redesign and modify most of the portions of this engine and I really enjoyed doing it.    3.   Trip to Spain  Boyâ€¦what a trip it was! Read the initial blogs to get a better idea. This was the best trip that I had and enjoyed it to the maximum extent without getting bogged down by constraints.    2.   Losing my passport This one is totally contrary to trip to Spain. During the trip to Netherlands in 2002, I lost my passport and had really tough and anxious moments there and thereafter.    1.   Product 2.1 Architecture  The best moment in the 5-year span is obviously the effort spent in designing and developing the product 2.1. Product 1.0 did not come out all that well and I wanted to address the shortcoming that caused Product 1.0 to fail. When I look back now, I feel great about the way I had approached to 2.1. It may not be the best, but definitely it was lot better and I think I did a good job there!   On this note, it is time for me to sign off and have a good sleep. Tomorrow marks the dawn of another year in this organization. With each year comes new hopes, new dreams and new events that can be blogged ! ;)) '),(4550,1359631,'15,July,2004','2004-07-15 00:00:00','    Belpahar is an industrial township located in Northern Orissa near the industrial belt of Rourkela. I had left from Hyderabad on 12th July and returned back on 15th July 2004. Time to take you on the journey!    The Calcutta Alliance !    This was the first time I was flying Indian Airlines, or so did I think ! Only while boarding did I realize that the flight from Hyderabad to Calcutta stopped over at Bhubaneshwar. Hence the total flying time was approximately 30 mins. The aircraft was an old one and this could be easily made out from the very old overhead control panels which most of the time dint seem to work ! The air hostesses were totally emotionless and were more like zombies out to complete the task assigned ! It is then that I realized that we were flying Alliance Air, a subsidiary of Indian Airlines which owned the old aircrafts.   After 1.5 hours of flying we landed at the Biju Patnaik airport in Bhubaneshwar. Ours was the only aircraft in that airport and the airport was totally void of any activity. From the sky Bhubaneshwar looked more like a town rather than a city. Unlike other citiies, there wasnt any major activity visible from the sky. The only major structure that I saw was of a typical Oriya temple with the architecture that was similar to the Puri temple.  We were initially alloted the last row in the aircraft. However we entered into agreement with some generous Infosys guys because of which we had a relatively better seat. But we were not aware that those guys were getting off at Bhubaneshwar. At Bhubaneshwar, there was one passenger who was looking confused, as we were occupying his seat. The emotionless airhostess was called, and with a totally hopeless expression (which could be read as \"What the shit is happening here ! Get off quickly ! Why do u guys give me much work ??!!\") she asked us our seat numbers and asked us to move to our seats. Well, we wondered if these people were alloted these flight as part of their punishment !!   Finally we reached Cal, the City of Joy ! The airport pillars were drapped in East Bengal Football Club colors. Clearly the atmosphere was changing. Our immediate task was to catch the Gitanjali Express that would be leaving the Howrah station in another 2 hours. Lot of time ... or so did we think !   One of the highlights of Cal is the bright yellow Amby Taxis. The streets of calcutta are flooded with the yellow taxis that is colorful for first timers like me ! Just like Hyderabad, here also it was the fight for survival ! The taxi drivers did not have the modern mechanisms like power steering, etc. However the ease and tact with which they manouvered the heavy taxis would definitely belittle all the schumachers and alonsos ! The climate was totally humid and for the first time in several months I was sweating, sweating so hard as though the Ganges was flowing from my head. The river Ganga (Hooghly river) did not flow from my head, but it definitely was flowing beneath me when I was over the Howrah bridge. It was a feeling that was totally \"meaningless\" ! ;)) To be frank, Howrah bridge wasnt as grand as I was expecting. After 90 minutes of battle for survival in the streets of Cal, we were finally at Howrah junction. Wait a second ... the train was at platform 20 and we had just around 20 mins more ! During the search for platform 20, there was an overbridge that never seemed to end ! For a while I thought we would miss the train. With just 10 mins to go, we were finally standing besides the Gitanjali express. With rains pouring down, we had made it just in time on our onward journey to Jharsuguda.   Tanaya â€“ â€œ Woh Kaun Thi?? â€   It was a day journey from Cal to Jharsuguda. The train passed through the famous places of Kharagpur (has the longest platform in the world !), Tatanagar (Jamshedpur), Rourkela, etc. We had to kill time ! Along with we had a Bengali couple in our bay who were travelling to Bombay. They were young and had the typical bengali looks. Along with them was Tanaya, a cute little girl, who barely managing to stand independently and who was still very much attached to her mother. She must have been around 1 year old and was very active during the entire journey. She was not bothered or worried by the strangers around and easily migled with us as well as others in the train. Tanaya\'s mother was particularly very expressive whenver she spoke about Tanaya. Very often, Tanaya would let off a cry which was an indication to her mother to get prepared to take care of her basic hunger needs. And each time her mother used to try to pacify her with the words \"Oooi Baaaba \" in a typical bengali style which would have made even the best of the sweets envious ! Bengali is truly a sweet language and I wouldnt mind trying my hands on Bengali if given a chance !    After around 8 hours of journey we reached Jharsuguda station at 2100 hrs approximately. It was a relatively small station, but being a junction handled a relatively heavy traffic. We were expecting a pick up till Belpahar. However since that did not happen, we had to hire a taxi to Belpahar. The entire distance was about 30 kms and it was a totally dark road. With only hope and the determination to fight it out if cornered, we took the risk and soon we hit the road to Belpahar. En route we crossed the L & T cement plant and also the Ib river, a branch of Mahanadi. After about 45 mins we saw the first signs that we were being taken to the proper destination ! The logo of Tat group were visible and we could see the welcome sign of \"Tata Refractory Limited\".  We were directly taken to \"Tata house\" where guests usually stayed. This was our base till we departed !    Tata House â€“ In Competition with Parents!   The Tata house was the first stop and the place where we had the first interaction with the locals inb Belpahar. It had around 20 rooms of various sizes, a canteen that could accomodate around 30 people and around 5 people at any time who knew the art of taking care of the guests. Their smiling faces, the concern shown towards the guests and the care showered on us were unmatched and has no parallels ! During the breakfasts, lunch or dinner they used to keep serving rotis, and rice in a manner where you just cant refuse their offer ! I was reminded of the story of snow white and the dwarfs where the dwarfs took total care of snow white. Of course, we werent any relatives of snow white, nor were they dwarfs, but the attitude was well appreciated by one and all and was one of the highlights of the visit.   Let me talk about the food, a topic that I really like talking about. One of the major difference in the food there was the lack of spice and less amount of salt. Coming from South India where spice is a necessity rather than luxury, this was something that could be easily noticed. In fact one of my colleagues, who had stayed there for 2 weeks earlier, had to have a diet of rice and pickles only so that he could get back his taste buds in proper shape ! One of the food items for breakfast was \" urlLink Chakuli \". To be frank, I expected something exotic, but what was served was plain \"kal dosa\" which was referred to as \"Chakuli\" in orissa. Another item worht mentioning was \"Chanar Pravesh\". This was served as dessert and was made from milk and very similar to \"Rasmalai\".  Apart from the Tata house, there was also the Tata club and the Tata Park. Belpahar is a small locality and has one major street where you get everything from vegetables to electronic items.This is the only place (a 100 metre stretch) where we could also get the mobile signals for Airtel and Idea services. Hence this was a spot that was frequently visited. A raliway gate served as the boundary beyond which the signals were not available ! Though my colleagues were able to get connected occassionally, I was the unfortunate one who always ended up getting the network busy message ! The street also has a Lord Krishna temple, whose architecture is typical Oriya style with cylindrical domes that reach out into the sky !    The Drive Down to Jharsuguda  After 2 days of hectic work, it was time to return back. It was the 14th of July, and it was also my star birthday ! Having totally forgotten about it, I ended up eating mutton and fish which was the first time that I violated the basic birthday code that I had followed right from my childhood days. This was bound to happen one day and it happened this year ! Without any regrets (I like to think it that way;) ) I packed my back and after visiting the Lord Krishna temple there I was on my way back to Jharsuguda raliway station.   Shivanchal Panda was the taxi driver on m journey back. He was a brahmin and was proud of the fact that he had married the lady whom he loved and that too without any hassles from both the families. His father had worked for the Hirakud dam , which was about 80 kms from Belpahar and had later moved to Belpahar. As for himself, he was a taxi driver right from his young days, and had travelled to places like Mumbai, Bangalore, Kerala, Tamil Nadu, etc.He often kept saying that the southern states were far more developed than his state and was aware that education was one of the major reasons. He took much pride in talking about a particular stretch of road on the highway that we were travelling in. This stretch was built by L&T and he claimed that once upon a time this was the best road in whole of India. The project was initiated by Biju Patnaik, whom he respected a lot, and was contracted to L & T. As per his story, Biju Patnaik had tol the constructors that once the road was completed, the amount would be paid only if the milkwalas were able to carry the pots of milk without spilling a drop of milk. Once the construction was completed, it seems this was tested out and there was no drop spilled !! The road from Belpahar to Jharsuguda passed through a jungle which was a haven for dacotis until 4 -5 years back. Now there is a police post for every kilometer which makes travelling during the nights more safe. Around 1 hour we talked about various topics from politics to Kerala to education to his life, etc. Finally with just 10 minutes for the train to arrive, we were at the Jharsuguda station. As soon as I entered the platform, in one of the chairs I found a pretty woman who had this beautiful round eyes, a glowing face and the typical Bengali looks. It was a perfect ending to the stay at Jharsuguda and a perfect beginning to the day ahead in Calcutta the next day !!    Kolkata â€“ The City of Joy   It was the morning of 15th and the time was somewhere around 0520 hrs. Daylight was already shining through the landscape and I wondered if my watch had stopped ! Being in the eastern part of India, this is mainly due to a mismatch formed because of the single time standard being followed in India. Anyhow, I dint feel like sleeping just because the landscape was so green and beautiful and there was a light drizzle that added the shine to the atmosphere ! The paddy fields, the bushes, the coconut trees , etc were all so vibrant as though they were ready for action and were enthusiatic about being part of the new day ! After 2 hours of journey broken by the crossing of the electric trains, I could finally see the Howrah bridge. Soon I found myself searching for the prepaid taxi counter in the railway station. It was an ocean of yellow cars, all packed into one shed. The cars were almost struggling to get out of the shed. often the gaps between each cars would be in cms and it looked like the taxi drivers spent most of the time there shouting at others to get off their way ! Finally I was into one of those taxis. It was a yellow Premier Padmini taxi. Believe me, I never saw a yellow premier padmini taxi ever in Cal, so I think I was among the lucky few to have taken a ride through Cal in a bright yellow premier padmini ! I wanted to sit in the front, but the driver politely told me that the door in the front is not that strong and could fall out any moment !! Well, I dint want to take any chances ! :))   I was dropped at my destination - Hotel Rutt Deen. It is definitely not a place that I would visit again. Though they claimed to be of 5 star category, it was a very shady place and the room stinked badly of old linen ! Toilets werent maintained properly and there was just no AC !! But still the cost was high ! I was a victim of incorrect internet advertising ! \"This was definitely not the place to stay for long\", I thought to myself. It was better off taking a walk through the lanes of Cal !   Finally I was walking through the Park Street at 0900 hrs. There was another 5 hours for my flight to take off and hence decided to walk for around 2 hours. On enquiring I found that Victoria memorial was just 20 mins walk from the place where I stayed. All I had to do was walk straight down Park Street till it joins Chowringee Stree, take the left there and walk straight till it reaches a junction where you would find the sign board to Victoria Memorial. On the way, the crowd kept increasing and what an amazing crowd it was ! Bong females are really beautiful ! One of the main reasons of their beauty is their beautiful eyes, which are large enough to convey the expressions in total and to keep you glued to the eyes till eternity ! I dont think I ever saw a bong female that was not worth looking at :)) John Keats had mentioned that \"A thing of beauty is a joy forever ! \". Calcutta was a city of beautiful people and no wonder it was rightly called \"The City of Joy\" !   Finally I was walking towards the Victoria Memorial, its reflection on the nearby lake being disturbed by the gentle ripples caused by the monsoon breeze. Against the backdrop of the dark monsoon clouds, the building was a majestic piece whose grandeur would face no competition in the years to come. The lawns were well maintained and it was a place where you could relax and forget all the tensions of the material world. From Victoria Memorial I decided to walk around the golf course / park, and in the process also saw the Hooghly bridge (Vidyasagar Sethu) and the Eden Gardens stadium. The park was located at the center of Calcutta and you could find various football teams discussing and practising the final tactics before taking a break for the oncoming monsoon to sing over Cal.   With 2 hours of good walking, what else could be in your mind ? It was time to search for FOOOOOOD ! I wanted real good bengali stuff and the best hotel in Cal that served Bengali food was Aheli restaurant in the Peerless Inn Hotel. Unfortunately it was closed and would open only at 1230. The time was just 1100 hrs. I did not want to give up. When I reached my hotel, I found that there was another restaurant called \"Suruchi\" that was being run by All Bengali Womens Union. This was just a km from the place where I stayed and I tool the McLeod road to reach the Elliot road where this hotel was located.   The hotel was open from 1030 for lunch and I was lucky enough to find a restaurant open for lunch so early ! I wanted good Bengali fish based lunch and this was conveyed to the incharge there. Soon, I was served with a bowl of rice and \" urlLink Rui Fish Curry \". This went on well with the rice. Though the spices werent as extravagant as in Kerala dishes, it had a different taste mainly promoted by the mustard that is regularly used in bengali dishes. I also had a fish fry (some Jhal) which was cooked really well so as to really excite the taste buds. Having had a sumptuous meal, it was time to catch my flight on my way back to Hyderabad. The flight was without any incidents except for very minor turbulences immediately after take off. The flight was even more exciting because of Hita, the air hostess who accompanied us on our way back to Hyderabad and also because of the good lunch served on the flight. With a truck load of memories, I landed in Hyderabad on July 15th at 1630 hrs to get back into the crowded life of mine !   Signing off ... K.V.Sathish Babu '),(4551,1359631,'11,July,2004','2004-07-11 00:00:00','    There have been many memorable moments, but three events stand out in my memories as far as organizing the events are concerned. In chronological order, the first event was â€œLiscar 96â€ which was an inter school contest conducted in our school. We had just passed out of school and being the immediate we were part of the organizing committee. It was a big hit, had its own bit of controversy, its share of fun, learning, etc. The way we resolved the winners for the tightly contested dance competition after hectic discussions, the manner in which the girls from Alvernia school protested to us that we were partial, the sight of a contestant and her brother protesting that we had no sense of music after we had announced the winners for the solo music competition, etc are all fresh and are memories that would remain as long as the air flows through my lungs. The second event was the sequence of events, preparation and effort leading to Nexus â€™99, the National Level Technical Symposium conducted by our department in my college. This one obviously is the best event that I have been part of. The memories related to this event are so vast that I may have to write a book rather than just blog !!   The third event is of course the Bali Trail (held of June 29th and 30th, 2004) in my organization. This was a quiz event were the contestants would be taken to various places of interest around the globe depending on their answers and choices. The teams are supposed to collect certain clues en route to Bali. The first team to reach Bali and submit the key (based on the clues) would be the winner. In the first week of June, we were approached to do a quiz event with a difference for Baanquet. Though I wanted to do a Rapid fire quiz, during the course of the discussions, our thoughts and ideas were leading to organizing a trail. The idea originated from a discussion about the Carmen game that was popular during the DOS days!   There were two main parts of the trail â€“ (1) Tracking each teams trail and displaying the same on the screen, (2) questions for each stop on the trail. The first part had to be done using power point and being a self-proclaimed expert in power point presentations, I agreed to take that responsibility. But somehow I was not comfortable with this. Certain events during Nexus â€™99 had made me decide then not to get involved in such activities in future. This time something kept telling me that I was doing exactly what I wanted to avoid. Many times I just dint want to take this forward and wanted to just give it up and lead the silent life focusing more on work than on other activities. But the efforts, enthusiasm, and nervousness of my quiz partner didnâ€™t allow me to go back. As far as this specific quiz is concerned, lots of credit goes to her as she did a great job that was totally contrary to my expectations. In the days leading to the quiz, she was so nervous that at times it seemed that she dint know what she was doing. Most of the time her expressions spoke more than she could ! But in the finals, she did a marvelous job.   When preparing the PPT, the first thing that struck me was the mainframe screen and the typical computer screens of Hollywood movies. Also of special significance were the screens from Matrix movie in which you would find a black screen with green alphabets falling all around. With the real estate on the screen divided and allocated for specific purposes, it was time to put in the content. But suddenly there was a bombshell. My partner wanted one more branching and all of a sudden the number of presentations that I had to make increased from 16 to 81!! â€œMan! That is some more work!â€ I thought to myself. I could feel the fumes coming out of my ear. There were two periods when I had to put in lot of extra effort â€“ first was for preparing the 80 presentations and second was for pumping the trivia in for each place.    urlLink      A Screen shot from \"The Bali Trail\"    During the entire course of the preparation I was sure that the presentation could be prepared in advance. But what worried me later was that during the trials, the presentation wasnâ€™t as smooth as expected. I was doing lots of mistakes in swapping, selection, etc. After the trials on the day before the finals, I was a bit worried. Any goof-up on my part would make the entire effort look worthless. Apart from my effort, it would also take up the shine from the questions that had been prepared by my partner by burning the midnight oil. I also had a history baggage wherein all the events that I have been part of werenâ€™t flawless! I was a bit worried and could feel the hairs turn grey with worry!   The finals had surprises in store for me. When the first slide was projected, the screen layout was looking much more beautiful than I had expected. From the experience of previous trials, we had taken some precautions and had modified and fine-tuned the way we wanted to approach and take the contestants through the finals. All through the way, the progress was smooth. Except for a minor blip the entire show went on flawlessly. I also realized that I could have avoided the trivia part. This was because the audience was more involved in the heat and speed of the finals than on the trivia section. Probably a picture of mine would have been a better replacement. I am not sure which of the two people would have liked to ignore the most!!   It definitely was the first time something (event organizing) went on flawlessly and it definitely found its way to being one of the most memorable events that I was part of (as far as organizing was concerned). However Nexus â€™99 would still remain the best event that I was part of especially because of the scale of the event and because of the efforts that we had put and because of the amount of learning that I was subjecting myself to ! The next question that obviously comes to my mind is â€¦ â€œWill I do it again, if opportunity knocks once more?â€ Probably not! Probably never!   Exceptions should not get generalized! '),(4552,1359631,'01,July,2004','2004-07-01 00:00:00','    urlLink    Its my Birthday tomorrow ! The collage of Garfield strips brings out my feelings on the eve of my Birthday ! '),(4553,1359631,'03,August,2004','2004-08-03 00:00:00','   â€œShruthiâ€ in carnatic music means â€œmelodyâ€, the musical pitch that imparts the melody. Basically the very substance or life of music lies in its shruthi. No folks, this is not a story about Carnatic music but it is basically a travelogue of Ajit\'s life as he travels through the self imposed obstacles of life is his search for Shruthi - his life ! Finally does he get the Shruthi of his life? This travelogue will take you right there !    Presenting Ajit !    Ajit, the main character of this story, popularly known among his friends as â€œAtom Shotâ€. Just like his character the origins of his name is also a mystery. But the most often talked about origin of his name dates back to the good old college days when during the course of gathering materials for an assignment he stumbled upon a story â€œAtom Shot - The Hero of Endless Spaceâ€ by Rob Dudwood, in a particular school website. He had hyped the story among his friend and in return they gifted him the name â€œAtom Shotâ€. Times flew by and lile most of his fellow Indians,Ajit was also absorbed into the booming field of softwares and projects ! Though belonging to a hitech world he is always boged down by the feeling of being like an adolescent most of the time, being neither orthodox nor modern, neither village bred nor city bred, neither here nor there, always in search of something! Probably that is why he has left behind a trail in search of Shruthi !    But who is Shruthi ?  Shruthi, Shruthi, Shruthi. I have referred to \"Shruthi\" so many times that you must be wondering \" Who the hell is Shruthi ??? \".    \"Is she a damsel, a tropican beauty, a babe, the love of Ajit\'s life, his girl friend, crush, wife, sister, ...\"   A beautiful word that provides the much needed positive energy in a stressful and surcharged environment. Shruthi is a word that provides melody to Ajit\'s soul at times of distress and varying emotions. Shruthi, the word that he would love to die for! Somehow or the other he has had the fascination for this word right from his childhood. Shruthi does not refer to a specific person. During the course of the trail you would find Shruthi taking various forms, shapes, etc. (Did I say shapes ?? :) ) Shruthi is the trail and the trail is for Shruthi!    Oops ! I have been so busy trying to give you guys an sneak preview into the life of Ajit that I almost forgot that Ajit has to catch a train to his native town ! I need to get to him as soon as possible so that I can travel along with him and take you through the journey of his life.   Catch you soon ! '),(4554,4259479,'20,August,2004','2004-08-20 00:00:00','   ...on my French reading, REALLY WEIRD BOOK, RhinocÃ©ros, although i\'m only like 50, 60 pages into it, i wouldn\'t suggest it to maintain mental health... but anyway... i woke up at 7 to go swimming.... that didn\'t happen, then woke up at 9h50 and i had to be at my Japanese teacher\'s house in Pleasant Hill at 10h00... zoomed over there feeling all gross and she stuffed me with food like i was a turkey for thanks giving or something, she doens\'t understand when i tell her i don\'t want any... and that last for like three hours.. finished up my third-year Kanji (the characters) and then came home around one or after... need to go to registration, but would rather go on monday... my mom is busy with the counselor anyway... i need to do like five loads of laundry, read at least 40 pages of french ...and study a bit of my spanish before school starts, and then i took a shower, need to eat breakfast... and mucho ist aller on (that was awesome: sp, ger, fr)   and .. later i\'m goin to Robert\'s to watch the one and only Princess Mononoke! then chill or whatever... anyway, i\'m psyched... but i still need to manage my time better... and i wont have too much to do anything before school starts cuz i work on Sat and Sun from 9-5... so not a whole lot of time to do my homework... although i will be getting much money for the amount of time i\'m working... i love money... and this is Cullen REDDY, signing off with Channel CTR news...  '),(4555,4259479,'19,August,2004','2004-08-19 00:00:00','   i had the most boring day today... i work up at 6 something... and strated to read my ap french book a bit till i passed out.. then woke up at 10h30, and read some more, returned the rental... that Durango was so much fun... came back... and kept on reading until i passed out again, until the phone rang, and woke me up.. and now i\'m here at my computer with no plans for tonight.... et on n\'a que cinq jours jusqu\'Ã  la rentrÃ©e...   i thought i was psyched for school to start... but i need another week to just sleep... unfortunately, that will never happen...  '),(4556,4259479,'18,August,2004','2004-08-18 00:00:00','   i had the most boring day today... i work up at 6 something... and strated to read my ap french book a bit till i passed out.. then woke up at 10h30, and read some more, returned the rental... that Durango was so much fun... came back... and kept on reading until i passed out again, until the phone rang, and woke me up.. and now i\'m here at my computer with no plans for tonight.... et on n\'a que cinq jours jusqu\'Ã  la rentrÃ©e...   i thought i was psyched for school to start... but i need another week to just sleep... unfortunately, that will never happen...  '),(4557,4259479,'18,August,2004','2004-08-18 00:00:00','   so last night i got to go to iMusic Cast and chill with Robert and meet his friends and see his friends\' band play; so really enjoyable night... and also finally saw Signs again... with Chris... and going back to watch Elephant;  then i need to do swim and smell like Chlorine- but with my goggles so my eyes dont burn up and shrivel like they always do when i forget::::  And then Mike is coming home today!!!!!!!!!!! havent seen him in like two weeks:::: but we\'ll be all good once he\'s home::: and nothing else is planned except that i need to start AP french homework:::: we\'ll see  '),(4558,4259479,'17,August,2004','2004-08-17 00:00:00','   I\'ve decided to metion a couple songs or artists that i\'ve discovered while i was in Europe every so often because STUPID america doesn\'t like good music.... well ... ...anyway, download them and see if you like...   Jamelia - Stuper Star  (it\'s in english, but i don\'t know if she\'s british or not, so didnt know if we\'d have it here...) K Maro - Femme Like You  (all you frenchies out there, good song) Tiziano Ferro - Perverso  (there\'s an italian, a spanish, and a french version.... check out his other music too, good stuff)  '),(4559,4259479,'17,August,2004','2004-08-17 00:00:00','   So a couple days after i got back from Europe, i noticed that my hermit crabs weren\'t moving too much altho they didn\'t do too much of that in the first place... and then noticed one of the limbs fell off one of them, and i found out my first crab was dead, and then a coupel days later, the other died.... how sad.... AND THEN a couple days later i step out of the shower and notice somethign squirming on the counter, and it\'s one of my crabs! A miracle!  so my favorite one was still alive.... (yes i often leave dead crabs on my bethroom counter for days on end...) and then this morning i realized he was dead again.... that\'s sounds weird... i hate those crabs messing with my mind...  '),(4560,4259479,'17,August,2004','2004-08-17 00:00:00','   So summer is ending, i\'ve been back from Europe, I still have to read that damn AP French book and do the packets, and we only have a week till school starts... but on the bright side, we\'re gonna be seniors! and then there\'s also Robert!  My really cool boyfriend in the making....Quel type!... : ) and then what\'s with the whole senior picnic not being a picnic and actually us going to a baseball game!?  I hate baseball! anyone else? And if you want french toast, don\'t go to the Egg Shoppe in Monclair because they SUCK! and anyway, i\'m just trying to pull shit out of my ass cuz this is my first entry and so... pretty much...  '),(4561,3905963,'13,July,2004','2004-07-13 00:00:00','    I got my sleepy self up this morning to take my Boo to the airport at 3:30 am. I\'m really glad he decided to see his best friend, but I have the feeling that Craig will convince my Boo to move down to Georgia in the near future. After all they are best friends and have lived on the same block for the past 20 years. When Craig decided to move to ATL he was crushed, and he even cried. Yes, this is the man I would like to spend the rest of my life with and father my children, so of course I would go with him to Georgia, but not for another 4 years.   I must finish my second Bachelor\'s degree in NYS first and I would like to at least get a year of work experience under my belt before I relocate. Let me stop jumping the gun and just wait to see what happens when he returns.  On a lighter note, my homegirl got into the Master\'s Program at Brooklyn College and I\'m thrilled. My friends and I all got our stuff together and the older we get the more productive we get. So hopefully on Friday we can make it a girls night in and get the Pinot Grigio flowing and watch some DVD\'s.  Oh wow...I just realized that the BBQ at the house is in a week and a half and I have yet to tell Jasmin about Abby and Junior. So let me start from the beginning. When I was in college Jasmin was my roommate. One weekend this guy named Junior came to visit her (they know each other from grade school), she had a crush on him since then and decided to act on it. So he and I meant and that was that. Well they didn\'t hit it off after all and I didn\'t she Junior until March of this year. We all went to a party at the Roxy and I was toasted from the alcohol, but I thought I seen Abby speaking to Junior (Abby is my roommate). After a week or two Abby is telling me how she meant this guy at the Roxy and he drives a Lexus, has a good job and is a bit freaky. But every time he came to the house I was at my Boo\'s house. Finally he came over one night when I was home and Melly happened to be there chillin as well. So when he walked in and Abby introduced us, I said \"Junior, oh I know him already, he is a friend of Jasmins\'.\" Well Melly bust out laughing and so did I. Maybe Abby took offense but hey Brooklyn is big but small at the same time.   Okay so here we are 4 months later and I have yet to tell Jas that those two are seeing each other none the less having sex on a regular basis.  The BBQ is July 24, what should I do? '),(4562,3905963,'12,July,2004','2004-07-12 00:00:00','    These pass couple of months I have been on the edge about getting into nursing school, but finally I got the acceptance letter and Friday I took a 2 hour trip to the Bronx (a different world) to handle my business. I went to the College paid my enrollment fee, handed in my health forms,went to Financial Aide, spoke to an Academic Advisor and registered for classes. Already having a BA from a state school I knew what to expect when I got there and had my scissors to cut any red tape. College Administration is usually full of \"red tape\" and I felt so bad for the freshmen that were there for orientation. No only are they forced to participate in these pre-college seminars but they are told \"hey we need this paper, that paper, and proof of this.\" Me being a vet to this game I came prepared. After handling my bees-wax I was able to enjoy my weekend.       This weekend was great, I got my hair done and seen my man on Friday. I spent my Saturday running around Manhattan looking for some type of Superhero shirt for my b/f and settled on a Punisher shirt. Then I ventured to Kings Plaza to get another shirt for him. God I\'m such a good girlfriend, but at times I can be a real b*tch. Sunday was even good. I visited family and then capped my day off with a good ole game of Double Dutch. Yes, I said Double Dutch. A couple of us (all over 21) decided it would be fun and even got the guys involved. It was a sight to see. A bunch of grown-ass people playing DD late at night like we didn\'t have to go to work in the morning. '),(4563,3911122,'16,July,2004','2004-07-16 00:00:00','    Today i\'m writing in blue HAHAHA ok i\'m seriously really bored. well today .The most exciting thing i did today was that&nbsp;i went to the movies. I saw The Notebook the 2nd best movie ever lol the first it A Walk to Remember with the huge hottie Shane West lol okies well that movie made me sad and made me want a special somebeody lol *tearing up* actually no im not lol ok well thas all cause nothing happened I\'M A BORING PERSON!!!!!!!! lol '),(4564,3911122,'15,July,2004','2004-07-15 00:00:00','    okies peoples i skipped a day but oh well. YES! i am writing in purple lol well yesterday i went to gymnastics and worked my butt off. 1 thing of advice don\'t eat dinner before going to gymnastics lol. okies well then burgi and amy came over&nbsp; to spend the night. Amy did my makeup which was very good by the way amy and burgi did my hair which was really coolies she did all of our hair and and and I&nbsp; picked out the outfits lol. it was fun. WELL when i was sleeping (on the floor) my arm somehow got on my bed and went numb i couldn\'t feel it at all it was pretty funny so i oculdn\'t move it i had to use my other hand to pick it up heehee i\'m so silly then we went to SD (splashdown) and i was sorta dissapointed but i won\'t say why cause i\'m just nice like that even though the person prob. hates me but i had to.......... ok anyway we won\'t get into that lol my hair looks really pretty right now and i\'m really relaxed well me madi and Ally made up names for each other my name is xSPLASHx madi\'s name is xPOPx and Ally\'s name is xBUBBLEx yes we are weird lol oh well if u don\'t like us then SCREW YOU i love everybody heehee okies i\'m done       &nbsp; '),(4565,3911122,'13,July,2004','2004-07-13 00:00:00','    ok weeeeeeeeeeell heehee today was unusual. heehee me Burgi and Karen went to the movies where we saw sleepover which was really funny u should see it and and and i bothered everybody in the movie theater with my extremely obnoxious laugh lol well i think it is anyway. Then we decided to go to the mall while were in the car and i picked up this scented oil ball thingy and sqeeuzed it and there was a hole in and it got all over my face which was nice it was sooooooo funny i was laughing so hard i was crying tears mixed with oil very funny. ok well thas it heehee so interesting and just to let everyone know i love my PJ\'s!!! '),(4566,3911122,'12,July,2004','2004-07-12 00:00:00','    Okies well today was EXTREMELY boring....... i sat around and sat around some more OMG AND sat around some more. My cousin mooned me which was really nasty and i slapped him for it. Me and my cousins rode up to Subway for lunch which was yummy yummy yummy and when i was riding down a hill some lay opened a door and i almost went flying into the door luckily i think fast and went around the door heehee.Then i came back home from many days at my dads house. Oh and the most exciting news ever i found out madi got a haircut!!! WOOT WOOT so amazing heehee '),(4567,671748,'30,May,2002','2002-05-30 00:00:00','    Okay. Patrick Roy is innept.  His Colorado Avalanche had a chance to win the Western Conference against the Detroit Red Wings, but his \"cocky\" attitude got him in trouble tonight.  When three Redwings rushed him, Roy bent down to cover the puck with his glove hand. In an attempt to hold it up to the referee in his glove hand, he misses and leaves the puck there on the ice. One of the alert Red Wings players knocks the puck in the net for a first period goal! After Patrick Roy had found out what happended, he certainly had egg on his face.  Patrick, you blew it for your team. Instead, the Red Wings tied up the series 3-3!   '),(4568,671748,'30,May,2002','2002-05-30 00:00:00','    Okay.. as much as it dissapoints me to do this, but I guess it\'s a good reason to start using this online journal. As the first official entry to this Hockey journal, I will use this to congradulate Ronnie Francis and the Hartford Whal... \'er I mean Carolina Hurricanes for becoming the Eastern Conference champions last night. They defeated the Toronto Maple Leafs in yet again an overtime game.  This is the first trip to the Stanley Cup in their franchise\'s history. (Even back when they played in Hartford). Somehow, I don\'t think they can match up pound for pound against the Western Conference heavyweights. Either Detroit or Colorado will pick them apart!  Sorry, Carolina fans.   '),(4569,671748,'22,June,2002','2002-06-22 00:00:00','    Drama and excitement is the story from Toronto at the NHL 2002 entry draft.   First off, the opening welcome was responded to the \"rude\" and \"disrespectful\" fans in the Air Canada Center by booing the NHL Commissioner Gary Bettman. He also was booed for congratulating the Carolina Hurricanes and the Stanley Cup champions, Detroit Red Wings.   There was some drama as to how the #1 pick was made. The Columbus Blue Jackets wanted to wanted to insure that they got the person they wanted. So, to screw over the Atlanta Thrashers, the Florida Panthers gave up on their #1 pick and selected third. Rick Nash of the London Ontario Knights in the OHL was the first overall selection. Jay Bowmeester went #3.  My hometown Pittsburgh Penguins have taken the defenceman Ryan Whittney of Boston University with the #5 overall pick.   On a personal note, my uncle works for the Washington \"Crapitals\" and got on TV when he was talking with their first draft pick. He got to hob-knob with the rich owner \"Trillionaire Ted\" Leonsis. He got lots of on air time as he was seen down on the floor at the represenative table. He does some of the media guides and most likely will do the profiles for the Washington draftees. My family and I are all proud of him!   '),(4570,671748,'18,June,2002','2002-06-18 00:00:00','    2001 MLB Baseball season - 55 bench clearing brawls   2001-2002 NHL Hockey season - 0 bench clearing brawls   '),(4571,671748,'14,June,2002','2002-06-14 00:00:00','    urlLink       Well, now that the Stanley Cup finals are over, we can begin to talk about the upcoming NHL draft weekend on June 22-23. The Pittsburgh Penguins fans, as the other 28 teams in the league that were not in the finals have been anxiously awaiting the Entry Draft.  There has been much speculation as to just who would be the number one draft pick (that belongs to the Florida Panthers). The top player ranked #1 in the scouting lists is Jay Bouwmeester. He did well this season with the Medicine Hat Tigers in the WHL. Florida is most likely to pick him.  Click on the link above to find out just who is expected to be the future stars in the NHL!   '),(4572,671748,'14,June,2002','2002-06-14 00:00:00','    Congratulations to the Detroit Red Wings for buy... I mean.. winning the Stanley Cup last night. I\'ll talk about this in another post later. Winning the Stanely Cup is the best thing - the most coveted prize in all of sports and hockey history. The Red Wings have certainly won it before. The last time they were there was the 1997 and 98 seasons.  I don\'t know what kind of spin on this to give. I hope that the Detroit Red Wings aren\'t percieved as buying the Stanley Cup this year because of all the high priced free agents that they added to their roster. Brett Hull, Luc Robitaille, and Dominic Hasek to name a few.   '),(4573,671748,'12,June,2002','2002-06-12 00:00:00','    Okay, I officially denounce my Hockey Fanaticism for 24 hours. Therefore, I hereby bestow the \"hockey fanatic\" title to Mr. and Mrs. Nick Arena. The mother, Sara, gave birth to a 9 lbs. baby boy. Having thought of a name for him, they both settled on  Joe Louis Arena . They named their child after the place where the Detroit Red Wings play!   urlLink  Read about it here from ESPN.com!   '),(4574,671748,'07,June,2002','2002-06-07 00:00:00','    Okay, one more weekend to go before I can hit the ice again!  The ice surface at Penn State\'s Ice arena is being yearly maintenance. It has been closed since the middle of May, I have been trying to contain my need to go into \"hockey withdraw.\" It is set to go back into working order on June 15th. Then, I could finally get my exercise!   '),(4575,671748,'05,June,2002','2002-06-05 00:00:00','    Okay, I was surfing around on the newly improved and updated Pittsburgh Penguins website and I come upon a message board post that had a link to this website  urlLink NHL 2 Portland, OR . It\'s about bringing the NHL to Portland, Oregon.   Seems like a good idea, just as long as it isn\'t at the expense of the Penguins team and it\'s fans!   Back when the Penguins NHL team was in limbo and facing a lot of bankrupsy debt, they were talking about moving the team to Oregon.   '),(4576,671748,'02,June,2002','2002-06-02 00:00:00','    As seen on a t-shirt featured at the  urlLink  McLanahan\'s | Penn State Room :   Eat, Sleep, & Play Hockey (Eating and Sleeping Optional)   '),(4577,671748,'02,June,2002','2002-06-02 00:00:00','    Okay, It\'s been decided and Patrick Roy is still innept.  After blasting Roy with 4 goals in the first period, the Colorado Avalanche lost 7-0 to the Detroit Red Wings. Detroit won the best of 7 series and the Western Conference title.  It looks like an all red Stanley Cup finals between the Carolina Hurricanes and the Detroit Red Wings. How are we supposed to tell these teams apart? Both teams colors are red and white.  Congratulations to the Detroit Red Wings!   '),(4578,671748,'29,July,2002','2002-07-29 00:00:00','    I took my last skate on the ice Saturday. *sniff, sniff* Since I will be moving away from State College, PA soon, I will not have access to a rink as much as I enjoyed going to Penn State\'s hockey rink. There, it serves like a sanctum to my troubled life sometimes allowing time to escape. Skating is like a gracefull act that starts fluid series of movements and perhaps be the only real way to travel! For those that actually have laced it up and enjoy the sport know what I\'m talking about.  The last 10 minutes or so, I had all to myself. I skated as hard and as fast as my burned out legs could take me. So, when I couldn\'t go anymore.. I gave it one last lap for good measure. That\'s how much I enjoy skating - kinda says it all really. I wish I played hockey when I was younger or that I chose it over football in High School. I would have had so much fun. I\'ve been skating for over a year now. Hopefully, I can find a rink close by my parents and maybe drive out to Greensberg, PA and skate at Kirk Nevin Arena. I heard of a few other rinks in Pittsburgh, but right now I don\'t know how to get to them.   '),(4579,671748,'24,July,2002','2002-07-24 00:00:00','    Create A Caption      Mario: Okay... that was a great job on the skating drills. Now, to teach us how to get your dad or mom to manage your career, please clap for our special guest, Eric L.   Kids: Yea!!!!!   OR  Mario: Okay kids, I want you to take a good look at these gloves and hat. They\'ll be what you\'ll be making in the warehouse next door.   '),(4580,671748,'11,July,2002','2002-07-11 00:00:00','    It was the oddest thing! I was playing NHL 2002 for the Playstation and had a match up of Minnesota Wild vs. the Pittsburgh Penguins. With less than 1:30 left in the game and the Wild on a power play down 5-3, Jim Dowd had the puck in the Penguins zone and from the top of the key, just somehow started skating in circles. This stunt effectively run out the last few remaining seconds to their Power Play and with it their hopes of mounting a comeback. I shouted, \"Shoot it!\" to imitate a fan sitting there in real life. Straka eventually sunk the \"empty netter\" for the Penguins to win 6-3.   This isn\'t the first odd glitch that I\'ve seen the computer simulate. Another odd moment a while ago was when between periods the Rink PA announcer had given a warning to the fans (and presumably to the gamer) that \"the puck could enter the stands at anytime, so be alert.\" This wouldn\'t be so bad if it had not zoomed in on #21 Espen Knutsen and Team Norway leaving the bench going back to the locker room for the 2nd Intermission. I\'ll remind you that it was Knutsen\'s blocked shot that went into the stands that claimed the life of a young Columbus Blue Jackets fan. Don\'t you find that eerie? A computer game and the programmers that made the game with no premonition of the tragedy, somehow later have everything line up for an insensitive twist of irony? What are the chances of a close-up of Knutsen and that random voice over at that time?   '),(4581,671748,'10,July,2002','2002-07-10 00:00:00','    okay.. after one day of being a part of the Hockey Pundits.. I almost got myself fired! After giving it some thought, the column that I wrote about Jarome Iginla, intending to be satirical, turned out sounding borderline terrible and it was published with the less offensive edited version. Whoops. I guess I\'d better fly right from here on out.  Come on!! A guy with the exotic name like Iginla should have an equally exotic pronouncing first name... try it once... \"Yar-o-may\" Iginla! See??! It makes more sense that way! It\'s along the same lines like a sports broadcaster we have here in Pittsburgh, Guy Junker. Someone once gave him a french name \"Gee Junk-air\".   '),(4582,671748,'07,July,2002','2002-07-07 00:00:00','    Hey hey! I got a job!  I\'ve just been accepted as a writer on   urlLink  The Hockey Pundits  . It\'s a great web page that shares their writing talents as personal columns are written in to response of hockey news of the world. I think they are all Canadian, but that\'s a good thing. ;)  Check it out! I\'ll be writing some on there to add some of my own brand of humor to the hockey world!   '),(4583,671748,'07,July,2002','2002-07-07 00:00:00','    According to a  urlLink Faceoff.com  article that was given to them off the European waver wire (kind of like the associated press briefs that American journalists get), the IIHF [International Ice Hockey Federation] based in the notoriously neutral country of Switzerland has determined through research that the Victoria rink in Montreal is the birthplace of Hockey.  People in Toronto are already pissed as they take their Hockey seriously. A response to the article by a reader of faceoff.com asked  \"Oh, my God! What will Bob Cole and Harry Neale do now? Toronto isn\'t the birthplace of hockey? HNIC  [Hockey Night in Canada]  will be crushed to hear this and somehow rewrite history to prove otherwise.\"   Regardless of where it was first played, thanks, Canada, you\'ve given the world a great game!   '),(4584,671748,'06,July,2002','2002-07-06 00:00:00','    Well, the frenzy has left everybody wondering what ramifications this free agency period has had the NHL. Salaries have gone up, the lower tier teams can\'t afford the better players to improve their team, and owners are reportedly losing money hand over foot.   \"unstable eddy\" Ed Belfour went to the Toronto Maple Leafs, CuJo [Curtis Joseph] signed with the Detroit Red Wings, Adam Oates became a Mighty Duck, Teemu Selanne knows the way to San Jose (and decided to stay there), the Columbus Blue Jackets have signed Scott Lachance (someone whom Penguins fans wanted in Pittsburgh) and Luke Richardson, and Robert Lang signed a big money deal with the Washington Capitals. But the most obsurd signing of the Free Agent period is Bill Guerin who signed with the Dallas Stars for 5 years and 45 million dollars.   '),(4585,671748,'05,July,2002','2002-07-05 00:00:00','    NHL Commissioner Gary Bettman dies and goes to heaven (yes, it\'s still possible). Once there, he sees the nice little house just inside the gate. An angel says to him, \"This is where you will be spending eternity, the house that has been prepared for you.\" Soon, Mr. Bettman looks around and inside the moddest little house is decorated with little pennants, hockey sticks, banners, and many other rare memoribilia. He looks out the little window of the house and sees off in the distance a large mansion. The mansion is adorned with black and gold pillars. The windows are large and expensively designed with black and gold stained glass. Mr. Bettman squints and looks closer and notices that the windows have the Penguins logo on it while the big oak door is adorned with the Penguins secondary logo. Mr. Bettman asks, \"Angel, why does Mario Lemieux get such a big house?\"  To which the Angel replies, \"That\'s not Mario\'s. That\'s God\'s house!\"   '),(4586,671748,'01,July,2002','2002-07-01 00:00:00','    Happy Canada Day!    Canada Day 2002 Hockey Tournament (as simulated by EA Sports NHL 2002)   Round One:  Toronto vs. Ottawa Montreal vs. Quebec Edmonton vs. Calgary Vancouver vs. Winnepeg  Round Two:  Toronto vs. Quebec Edmonton vs. Vancouver  Finals:  Quebec vs. Edmonton   Winner of the 2002 Canada Day tournament - The Edmonton Oilers!   '),(4587,671748,'01,July,2002','2002-07-01 00:00:00','    Want to own your own Hockey Franchise?  urlLink eBay item 1745220456 (Ends Jul-08-02ï¿½09:43:35 PDT ) - AA Hockey Franchise For Sale - Anchorage Aces   Check out the ebay auction offering the Ankorage Aces a WCHL franchise currently asking for 1.8 million dollars.  How soon will they do this for the Buffalo Sabres?   '),(4588,671748,'01,July,2002','2002-07-01 00:00:00','    Okay, NHL free agency period has started!! Let the shopping begin!  The trend in the NHL now is to trade away your high priced free agents to gain an extra compensatory draft pick when the players sign with a big market franchise.  Already, Theo Fleury went to the Sharks, Curtis Joseph went to the Calgary Flames, Ed Belfour and Tie Domi became property of the Nashville Predators. These smaller market franchises are hoping that they can either sign them or get the compensatory pick for \'em.   '),(4589,671748,'27,August,2002','2002-08-27 00:00:00','    My brother Dave came to visit me here in State College on Sunday. So, I talked him into going ice skating with me at the PSU Ice rink (and yes.. it\'s not the last time, Iike I previously reported). I wound up meeting a Coach who was there to help teach little kids how to skate and wound up giving pointers to help my skating technique... like bending at the knees more and skating backwards. It felt great to be able to share my favorite sport with someone I care about like my brother. Normally, I go skating on my own. I wonder if I can talk a date into going someday?   '),(4590,671748,'24,August,2002','2002-08-24 00:00:00','    I can\'t wait!! NHL 2003 comes out on September 26th for Playstation! Look for a review here and on The Hockey Pundits webpage.   '),(4591,671748,'22,August,2002','2002-08-22 00:00:00','    Well, the Penguins have finally unveiled the new home jersey to match their now away jersey. They wanted to throw away their old jerseys to which I am very upset that they are doing. Now, I have their \"secondary logo\" jerseys with my name and number on them that are now old and out of style. I wish they could keep using their jerseys they used last year, but the change was necessary. I don\'t think anything was wrong with the last ones, certainly, just because the team changes uniforms doesn\'t exactly make them Stanley Cup contenders. I will probably cave and go out to buy one once I have the money. I read up on the teams web page that some of the other fans think that changing the uniforms will instantly make us better. Those idiots!! Clothes don\'t make the man.. or in this case a hockey player! Now I have a closet full of penguins jerseys that I cannot wear as much with my name and #21 on them!   '),(4592,671748,'18,August,2002','2002-08-18 00:00:00','    This morning I want to talk about a young woman who I know is a big hockey fan. She loves hockey about as much as her Canadian herritage. Because of her, I have come to respect the Maple Leafs- her hometown team. You might have already seen  urlLink Trish Stratus  on many Fitness Magazine covers, on TSN Cable TV, or here in America as a WWE Superstar! She also was voted as  urlLink  2001 Babe of the Year . To say I have a crush on Trish would be an understatement. Her energetic personality shines through every Monday Night on WWE Raw as her stunts, interviews, and gorgeous body entertain me along with the rest of her fans. All the photos that I do collect of her, Trish appears in my favorite in an oversized hockey black and white game sweater sponsored by the NHLPA. I will post it up on here sometime soon. On a recent  WWE Confidential  program, she gave a tour of her home while showing off her action figure collection of Mats Sundin and Curtis Joseph.  I dedicate this blog journal entry to my \"hockey hunny\"... TRISH STRATUS!  Then, Trish is not the only hockey fan in the WWE! Former champion, Chris Jericho a native of Winnepeg, had his dad play professionally for the NY Rangers. Chris also is active playing in celebrity hockey games.   '),(4593,671748,'17,August,2002','2002-08-17 00:00:00','    I played an exciting hockey game this afternoon on NHL 2002 for Playstation. I used the updated rosters from the most recent 2001-02 season. I was Toronto vs. Vancouver Cannucks. It was an exciting high scoring affair! 12-10 Toronto! The lead changed hands at least 3 or 4 times. Domi got the empty netter at the very end.   '),(4594,671748,'16,August,2002','2002-08-16 00:00:00','    I found some really cool additions to improving a web page! Here\'s a Guest Book that puts you on the map!  urlLink guestmap  Click on this link and let me know who sees and reads my hockey blog!   '),(4595,671748,'11,August,2002','2002-08-11 00:00:00','    Have you been to NHL.com lately? Something\'s going on there!  For the past several days to check out the NHL\'s web site, it returns a \"Unable to\" message. My best guess is that they are redesigning the site for the new season only 2 months away! Oh boy I can\'t wait!   '),(4596,671748,'29,September,2002','2002-09-29 00:00:00','    I had my day all planned out today. I was going to go to work in the morning, then go ice skating, then meet my dad, then take my sister and her boyfriend to church. September 29th is a special day for me because in the Catholic tradition it sets asside a day to honor St. Michael. I try to keep him in my thoughts and say little prayers thouout the day.  While I was at the ice rink today, I met a nice girl that works there. Chrissy. She had basically just stood on the bench side of the rink and watched everybody go by during the second hour of Public Session skate. She told me that she was a student studying psychology. I told her that I worked for WRSC and was pretty impressed. I might go back and talk to her some more some day.  Then, I noticed all my old girl-friends hanging around in the rink\'s lobby. They had just gotten done with a game. They lost, but I was glad to get a chance to talk to the coach and invite them on the air to be a guest of Sportsline on WRSC/WBLF sometime.  I also watched this supposed father/son outing at the rink today. The poor kid must have either been on the ice the first time or he wasn\'t going to use the metal \"walkers\". His dad held his hand as the poor kid was afraid to actually skate. The kid had these huge gloves, helmet and these really nice pair of CCM skates like he was playing hockey. I looked at the dad and his facial expressions were that of disappointment at his kid\'s progression. They moved an inch at a time and hugged the boards closely. I think the poor kid only went around the rink once. When I passed him just getting off the ice, I clapped for him and applauded his efforts. If I were that kid\'s dad, I would have been more patient and understanding than the grownup that I saw. Off the ice, the kid played on the benches right outside the rink and looked happy. I hope that the guy didn\'t waste a whole lot of money on the kid. Maybe he\'ll want to come back again.   '),(4597,671748,'28,September,2002','2002-09-28 00:00:00','    Well, score one for old school hockey! The Avalanche went back to their old stomping grounds of Quebec and Le Colisee. They played a pre-season exhibiton hockey game against their old rivals the Montreal Canadians today. Adam Foote, Joe Sakic, and Peter Forsberg were two of the few remaining players who remember playing in Nordique blue. They told the press that they were nostalgic and glad that the 15, 339 fans gave them enough cheers. Montreal still one the game 4-1.   '),(4598,671748,'18,September,2002','2002-09-18 00:00:00','    The Alexei Kovalev saga that is going on right now is getting interesting to follow. From what I understand, the GM Craig Patrick and Team Owner Mario Lemieux oferred Kovalev a ton of money to extend his contract twice. Kovalev turned both of them down. The second and most recent time, Kovalev turns it down and decides to talk it over with his agent to make a counter offer. Since then, GM Patrick has said that if it doesn\'t get done, he will not continue contract extension talks after October 10th when the season begins.   '),(4599,671748,'18,September,2002','2002-09-18 00:00:00','    \" You could be seeing the Leafs in late May again next spring. But missing the playoffs is also a distinct possibility. \"  - Tony Gallagher, Faceoff.com Columnist   Way to go, Tony! That\'s the real sure-fire Hockey journalism that people need! Take a stand! (For some of you Canadians without a sense of humor, I was being sarcastic there.) Also, it gives me hope knowing that this is my competition in the world of Hockey Journalism. I know I\'m better than this guy! Geez!   What really needs to be said... The Leafs do have a pretty good shot at going back to the playoffs this year. They\'ve got one proven goaltender in \"Unstable Eddie\" Belfour, picked up some key acquisitions this summer with Fitzgerald and Shvela. The biggest sneak in the whole summer was re-signing franchise player and notorious goon, Tie Domi after they gave him up to Nashville in a trade over the summer.   '),(4600,671748,'18,September,2002','2002-09-18 00:00:00','    \" If I can get serious for a minute,... \" The Hockey Pundits  and I have had a falling out. I can see clearly that they do not have freedom of speech at that web site. I can\'t really say what I have been wanting to say against Canada on this post. I really don\'t see how a country as ass-backwards as that can produce good Canadians like Trish Stratus!   Okay.... I\'ve washed my hands... I\'m through with those \"un-american\" imbisiles!!!   Things just aren\'t as happy here in  Happy Valley  right now.   '),(4601,671748,'17,September,2002','2002-09-17 00:00:00','    It\'s that time of year again!  The time for contracts, schedules, and/or strategies are over and now it\'s time to lace \'em up!  The NHL has begun to open training camp. This means that my beloved Penguins are gearing up for a new season with a lot of new players on the roster. Daigle, Vuijtek, and Salivanov are some of the veterans in camp on a try-out level.  I\'m thinking it\'s about time that I hit the ice again too. I\'m planning on skating from 1-3 PM at the ice complex this afternoon.   '),(4602,671748,'11,September,2002','2002-09-11 00:00:00','    I wanted to use this time to put up a small little remembrance post about all of us who were affected by the Attack on America. All of our lives were changed that day. In a special way, the entire world was changed. The Hockey sports world also was affected and did what it could to heal America.   Directly, The Los Angeles Kings were affected by losing two of their scouts. This past season they wore special patches in their honor. The New York hockey teams were also at the top by helping to raise money for charity. Buffalo Sabres created their \'one time only\' third jersey replacing their buffalo head logo with a diagonal \'New York\'. The Philadelphia Flyers wore special American flag patches on their orange away jerseys on the back of their uniforms. The most visual and most lasting tribute to the tragedy is the placement of a large red, white, and blue ribbon behind every goal net directly on the ice. The Washington Capitals were also sorrowful with the loss at the Pentagon.   Hockey did the best it could to heal the wounds. It focused many North Americans on who were the real heroes.    9/11/01... We will never forget.   '),(4603,671748,'04,September,2002','2002-09-04 00:00:00','    urlLink  Hockey Fanatic\'s Dreams Come True as Voice of Alma Matter Hockey Team   Michael Fedor has announced plans to become the new play-by-play announcer for the 6-time ACHA National Champions Penn State Nittany Lion Icers. After a successful trial run with area high school football games, Michael was announced on air that WRSC 1390 AM / WBLF 970 AM will be there for the Penn State University Hockey team this comming season. The team has just come off a three-peat year in the ACHA, hoping to do it once again for a 4th straight championship year. Like a big fish in the little pond, plans for the program are to join Division I NCAA competition in the future.  Michael has had experience in University varsity sports as the Public Relations/PA announcer for the Penn State Lady Icers last season. No word as to any plans if Women\'s hockey will be covered on air.   '),(4604,671748,'31,October,2002','2002-10-31 00:00:00','    Things that are scarry: 1.) (If) Eric Lindros wins the Lady Bing Trophy this season. 2.) (if) The Flyers win the Stanley Cup 3.) A back and blue fight victim by Crapitals inforcer, Stephen Peat. 4.) A Gary Bettman costume!!   '),(4605,671748,'29,October,2002','2002-10-29 00:00:00','    As I was surfing along Reuters on the Internet looking for something to discuss on the [much hated] PSU student radio station as a special guest of the DJ there, I found a story that I should mention here as well.  A Professor of English Literature at the University of Saskatchewan, Michael Kennedy, has a unique class on the curriculum. At the college, Prof. Kennedy teaches a course called  Reading Culture Canadian Hockey . The class there is so popular that for the Spring 2003 semester, the allotment of 39 students has already been filled.     \"Themes such as isolation, survival or the immigrant experience are more common threads in Canadian literature.\"  But Kennedy argues that hockey -- a sport that Canadians love with a near-religious passion -- is often used as a metaphor in the country\'s writing.  \"It\'s a device for protagonists to \"find themselves,\" and a means for defining Canadian culture\", he said.      To achieve this, Kennedy has a reading list of novels, poems, and a play script  \"Les Canadiens\"  about a history of Quebec within the confines of a hockey rink.    The class really seems legitimate as the students are expected to learn from many different sources of fiction and nonfiction sports articles and applying critical thinking to distinguish fact from opinion. One assignment saw students learn about jargon, rampant in hockey commentary.      \"Students learn to research and write essays, defending their positions on the game,\"  Kennedy said.   '),(4606,671748,'28,October,2002','2002-10-28 00:00:00','    Tonight,   MY   Pittsburgh Penguins face off against the rival Washington Crapit... \'er I mean.. Capitals. The feature matchup is Mario Lemieux playing against his former employees - Jaromir Jagr and Robert Lang.  I would have loved to see this game, but I have to work. I gotta stick around and let \"Ironhead\" go on the air and promo the Monday Night Football game from  urlLink  Damons Restaraunt.  Luckilly, \"Ironhead\" is keeping me updated in the  urlLink WRSC 1390  studio on the hockey game.   '),(4607,671748,'26,October,2002','2002-10-26 00:00:00','    Congratulations Joe Batista for winning 400 games at Penn State to be their all time winning Icers coach.   '),(4608,671748,'24,October,2002','2002-10-24 00:00:00','    Jason Spezza made his NHL debut tonight in Boston.  Anointed as the \"next best thing\" among hockey experts, Spezza wore #39 for the Ottawa Senators. Spezza was called up from the AHL because of an injury to Radek Bonk. He collected his first point by assisting on a Marian Hossa goal. Fans expect Spezza to be the next superstar having done something that only three other NHL players have done - play for Team Canada in the world Juniors at age 18. Only Lemieux, Gretsky, and 2002 draftee Bouwmeister have that honor.   '),(4609,671748,'22,October,2002','2002-10-22 00:00:00','    See Minnesota in High Definition TV this season!  The NHL along with HDNet announced that they have made plans to air some games in high definition picture quality to appeal to those who own HDTV sets. However, the schedule of games to be broadcasted isn\'t exactly going to permit people to see all of their favorite teams in that quality. According to the NHL website, of the 7 games listed to be broadcast in HDTV, 4 of them involve the Minnesota Wild. You would think that they would at least try to get all the teams on there once! Personally, I don\'t mind. I like Minnesota with their star player, Marian Gaborik. These first 2 games against Colorado sound pretty good!  Sun., Oct. 27 9:00 p.m. Minnesota @ Colorado  Tues., Oct. 29 8:00 p.m. Colorado @ Minnesota  Sun., Nov. 3 7:00 p.m. Nashville @ Phoenix  Mon., Nov. 4 10:30 p.m. Minnesota @ Los Angeles  Wed., Nov. 6 10:30 p.m. Nashville @ Anaheim  Thurs., Nov. 7 9:00 p.m. Minnesota @ Phoenix  Fri., Nov. 8 8:30 p.m. Toronto @ Dallas   This almost makes me wish I had the money to buy an HDTV now.   '),(4610,671748,'22,October,2002','2002-10-22 00:00:00','    Evgeny Nabokov has received an enormous pay raise today as he ended his holdout, signing with the San Jose Sharks for 2 years 7 million dollars [according to SportingNews Radio]. Last season, the Kazakhstan native earned $575,000 US last season.  Some quick notes about Nabokov is that he is the 7th goaltender and first European-born goaltender in NHL history to score an actual goal during a regular season NHL game. He ranked 3rd last year in the NHL with 37 wins. He won the Vezina trophy as the leagues top goaltender and in 2000-2001 was selected as the Rookie of the Year.   '),(4611,671748,'19,October,2002','2002-10-19 00:00:00','    Thanks for all the extra prayers and good vibes.  Last night I made the on air broadcast of the Penn State Ice Hockey team as the on air color analysist. Things were looking pretty bleek there for a while, I didn\'t know weather or not my bosses would give me the chance to voice the Icers, but when the opportunity finally came down to the wire, I was in the studio when I was told to go ahead over and help out with the broadcast.  I met Jason Mals, a crossover from the PSU student station, at the Greenberg Ice Sports Complex. He was worried about doing the broadcast by himself so he originally got some of the PSU hockey players not dressed for that game. When I met Jason, he was also profefssionaly dressed in a shirt and tie. We did our best with what was available to us. Originally, we had some problems dialing in, but once we got everything together, we were doing pretty well! We talked at length about the Icers and the upcoming Football game with Northwestern. Later, we brought the stuff back from the studio and got a tape to keep with our performance.  Penn State won 9-1 vs. Kent State.   '),(4612,671748,'17,October,2002','2002-10-17 00:00:00','    urlLink  Faceoff.com  has a poll on their page that I noticed last night and it\'s still up this morning, but I was facinated with the choices. The question was...  \"What has been the biggest surprise so far this season?\"  The choices were/are:      Tampa Bay is undefeated   Minnesota is the best in the Western Conference   Mario is healthy and on fire   Colorado is the west\'s lowest scoring team   Toronto is the east\'s highest scoring team   '),(4613,671748,'17,October,2002','2002-10-17 00:00:00','    I gave my boss and the sports programer the tape of me auditioning for the voice of the Nittany Lion Icers. I hope everything goes well with that...   The PSU Icers open their season against Kent State University on Friday night. I may get sent elsewhere to cover a stupid high school football game, and it\'s homecomming for PSU this weekend. T-R-A-F-F-I-C!!   '),(4614,671748,'15,October,2002','2002-10-15 00:00:00','    Already starting behind a huge hole after the first game of the year with negative point values due to a Commissioner\'s Snafu, the   Ruthless Aggression   have taken over first place after 5 games into the NHL season. I\'ve here is a dead heat with second place as one point separates the   BeeGees   of   2099   and the   OYLERS   and the fourth place team is the   Imbeciles   who seem to be struggling a bit. It\'s still anybody\'s lead for the taking! I hope the league stays active and remembers to set the lines everyday. (but you can \"cheat\" a little and set the rosters days in advance)  It\'s an 82 game schedule, we\'ll see how this all works out in the spring!   '),(4615,671748,'15,October,2002','2002-10-15 00:00:00','    Revenge was on the minds of the Pittsburgh Penguins and their fans as they all traveled to Toronto to wash their hands of the embarrasing 6-0 opening day loss.  There were some really stunning goals scored by Aleskey Morozov, Mario Lemieux, and Ville Neminen. Morozov hit a top shelf glove hand goal late in the third period and Neminen scored an amazing goal by skating out to the top of the key, then doing a 360 pirrouette and then with the final follow-through used a wrist shot to score a goal earlier in the third period Mario also scored an amazing goal from behind the net by banking it off of the goaltenders back. The goalie was later credited this quote:  \"He\'s done that 100 times.\"   The final score was 5-4 in favor of my Penguins, but the revenge wasn\'t as fulfilling enough. The goaltender that was in the net the first game for the \"Leaves\" has been injured and did not play this game. Also, I should congratulate Mats Sundin for scoring his 400th career goal last night.  I have been really gaining respect for Toronto because I know they are really good and they are Trish Stratus\' favorite team. I hope that they can build this into a new rivalry.   '),(4616,671748,'12,October,2002','2002-10-12 00:00:00','    The EA Sports Trax is the next best big improvement that EA Sports has given to their long running gaming franchise on  NHL 2003 . EA Sports thought instead of having some really boring, innane, and repetitive instrumental music that covers the menu section in NHL 2002. (I usually just hit mute on my TV when I\'m working with the rosters or create a player mode) A little black box pops up in the bottom right hand corner of the screen to let you know what song is playing, who performs it, and what album it\'s from. I was so surprised and glad that they chose to contain two songs that are on my station  urlLink  WQWK 97.1fm - Qwk Rock . The adrenalin pumping \"Sweetness\" by Jimmy Eat World and Papa Roach\'s \"She Loves Me Not\" are two of the best known songs included in the lineup. \"Modern Rock and Hockey go together like peanut butter and jelly.  As far as the gameplay goes,  NHL 2003  emphisized more realism with the hockey players giving you more control over stick handling, deke-ing, and shot selection. I\'d bet that their game developers, designers, and focus groups told them they wanted more realistic movements. You can\'t have your cake and eat it too though, with all the detail in the skaters movements, it slows down the game play. I guess people who complain about it being too slow is used to 16-bit \"hat trick hockey\" games where all players looked the same and molded the same way.  NHL 2003  is unique in that way. Diving over a sprawling goaltender, cutting back around a defenseman makes for great hockey action immitating a real world senario.  Some negatives with the PS2 version compared to the PC version is that the some rosters and uniforms are not in the game. You won\'t see the Penguins new \"vegas gold - skating Penguin\" home white uniforms in the PS2 version (which is okay since I like last seasons home jersey more) and the Los Angeles Kings (who have the nicest uniforms in the NHL) won\'t have updated uniforms since they dumped their coat of arms crest for a silver crown logo made popular from their 3rd jersey. Their royal purple, silver, and black are really nice. Team USA has no updated jersey from their silver medal Olympic team. Also, my two favorite teams to use Pittsburgh and Team Slovakia have their rosters slightly off. Guys like Beranek, Brown, and Dollas have not played in years, but still remain on the Pens roster. Machala somehow found a way to appear on the Slovak roster after being MIA for 2 years since they can fill a whole team with Slovak players.   '),(4617,671748,'10,October,2002','2002-10-10 00:00:00','    I was waaaaaaaaaaaaaay off on that one. *pouts*  I went to my new favorite hangout tonight to watch the Penguins drop their home opener 6-0. It was aweful to watch. Another thing that I\'m going to have to get used to is the new home jerseys. They look way too white. There\'s no more yellow only vegas gold. Mats scored two goals and got goal #399 on his career tonight.  On a lighter note, I got to see a familiar face in Kristen the waitress and meet a really sweet young lady named Meghan at \"my favorite place to be\".  I also got to go to Blockbuster and rent EA Sports NHL 2003. I will post a review of the game real soon.   '),(4618,671748,'09,October,2002','2002-10-09 00:00:00','    So, the Pittsburgh Penguins will be at home against the team that I am gaining a lot of respect for rival-wise in the Toronto Maple Leafs. I expect the teams to really test out the goaltending with a barrage of shots at the net. We get our first look at \"Unstable Eddie\" Belfour in the net for Toronto tonight. Look for him to do something highlight worthy to make a good first impression.  Prediction: Pens 4 \"Leaves\" 2   '),(4619,671748,'09,October,2002','2002-10-09 00:00:00','    WOOOOOOOO FREAKIN\' HOOOOOOOOOOOOOOO!! It\'s time to DROP THE PUCK ALREADY!!!!!  The Los Angeles Kings are at home to play the Phoenix Coyotes. They plan to raise the #99 in the Staples Center and honor the visiting team\'s owner - some guy named Gretsky.  Then, the Colorado \"\'Lanche\" are in action tonight against the team formerly known as the Minnesota North Stars.  My beloved Penguins are in action tomorrow night,... I\'m going to make sure to watch that game!! It\'s what I\'ve been waiting for all summer long! Go Mario!! Go Kovy! Expect a pregame and postgame analysis from me about the Pens vs. Leafs.  I\'ve signed up and created my own Fantasy Hockey League on Yahoo... the teams are stacked! I am thinking about making it mandatory for all the teams in the league to check this page out. It\'s going to be a real fun year!   '),(4620,671748,'06,October,2002','2002-10-06 00:00:00','    In an effort to diversify my reporting on things other than NHL and Pittsburgh Hockey, the Erie Otters have made the front pages on faceoff.com.  Last year, I watched a game between the Kitchner Rangers at the Erie Otters. Saturday night, those same two teams opened up the Ontario Hockey League schedule. Erie raised their 2001-02 OHL championship banner in front of a sold out 5,280 Tullio Arena crowd. Alas, Erie wasn\'t able to pull off the win. Those Rangers from Kitchner, Canada won the game 8-4.   '),(4621,671748,'01,October,2002','2002-10-01 00:00:00','    Former Owner of Montreal Canadiens Dies  MONTREAL (AP) -- Hartland de Montarville Molson, a former Canadian senator and longtime owner of the Montreal Canadiens hockey team, has died at age 95.  Molson also headed the family brewing dynasty, but is best known for his association with the National Hockey League franchise. Molson was inducted in the Hockey Hall of Fame in 1973.   Molson served as a director of the brewery for more than 50 years and held the positions of president, chairman and honorary chairman. He was appointed to the Senate in 1955 and he served 38 years until the age of 86. Born in Montreal in 1907, Molson joined the Royal Canadian Air Force in 1939 and was wounded during in the Battle of Britain.    ~Awww.. Rest in peace, Mr. Molson. I\'ll have a bottle of Golden in your honor tonight.   '),(4622,671748,'01,October,2002','2002-10-01 00:00:00','    \"Fedor Fedorov wasn\'t a first-round draft pick, was never considered a star in junior and wasn\'t even signed by the first team that drafted him. Last season, his first as a pro, he suffered a career-threatening eye injury and played just 10 games. If his name was Fedor Eastwood, in fact, he would have been a complete afterthought coming into the Vancouver Canucks training camp. But in three short weeks, the 21-year-old centre has morphed from a project into the Canucks\' best prospect. We don\'t know what that says about the Canucks\' recent drafts. We do know what that says about Fedorov.\"   Ed Willes,  The Vancouver Province    No relation.   '),(4623,671748,'30,November,2002','2002-11-30 00:00:00','    Congratulations to Peter Stastny, Stan Mikita, and Vaclav Nedomansky for their inagural induction to the Slovak Hockey Hall of Fame today. The museum itself will be stored in the beautiful Bratislava Castle.  They were the forerunners to the NHL, which helped the league grow into a world game. They paved the way for the European stars of today, Peter Nedved, Jaromir Jagr, Martin Straka, and Marian Gaborik.   '),(4624,671748,'27,November,2002','2002-11-27 00:00:00','    Today, my brother and I went to the Mellon Arena on our Thanksgiving vacation to see the hometown Pittsburgh Penguins vs. their hated rivals the Philadelphia Flyers.  We did get great seats, but enjoyed the birds eye view from the \"F\" section. It was so high up that the new netting behind each goal was not a factor to enjoying the game, in my opinion.  The Pens went on a scoring rampage, netting 4 goals in the first period and scoring 40-some seconds into the next. I was hoping that the game would turn ugly so that there would be a fight or two!  Pens win!! 7-2!!   '),(4625,671748,'26,November,2002','2002-11-26 00:00:00','    An article that I found written by Terry Bell on Faceoff.com is titled  urlLink  Lemaire says Gaborik Not Quite a Superstar . I suggest you read it.  The article starts out by mentioning that my favorite player not on the Penguins team was the first person to score a goal for the new Minesota Wild. The \"experts\" and teammates in the article also says that he is a great player, but has a lot to learn still. His coach told reporters that he wasn\'t ready to put \"Gabby\" Gaborik in to the superstar category just yet.   '),(4626,671748,'19,November,2002','2002-11-19 00:00:00','    The LEGO company and the National Hockey League announced today that they\'ve agreed on a internationally cross promotions merger with LEGO. There will most likely be LEGO toys featuring the NHL brand and the All-Star Entertainment will have some sort of LEGO tie-in.  I\'m skeptical that this will work. Apparently, Bettman wants to get like the cigarrette companies and market to children.  The only thing that they did right was announce it at the \"21 Club\" in New York. Twenty-One is a lucky number.   '),(4627,671748,'15,November,2002','2002-11-15 00:00:00','    The long standing NHL veteran was supposed to make a big return tonight after coming back after tearing his ACL in last season\'s playoffs.  Bad Luck would hit him again during his comeback, Pecca did not finish, injuring his knee in the game!   '),(4628,671748,'12,November,2002','2002-11-12 00:00:00','    The NHL has been reported to issue heavy penalties to a few New York Rangers because of what happened in a recent game against the Columbus Blue Jackets. What makes matters worse than having a high-priced talented team like the Rangers getting their collective ass beat by a young expansion team, but with 2.5 seconds left and losing 6-3, Coach Bryan Trottier put noted \"enforcers\" Krystof Oliwa and Sandy McCarthy on the ice to take the faceoff. Once the puck was dropped, the \"enforcers\" dropped the gloves, clobbering Grant Marshal.  Oliwa (a former Penguin and Blue Jacket) was suspended 5 games and forced to give up his salary for those games. Coach Trottier (a former Penguin player who I thought was a real class individual) was suspended for two games and McCarthy is now $1,000 less. If that wasn\'t enough, the Rangers organization was also fined $25,000 dollars.   '),(4629,671748,'07,November,2002','2002-11-07 00:00:00','    When you combine my two favorite things in this world, you can sure bet that I\'m interested!  On November 6th, the Women\'s Team Canada defeated the Women\'s team USA in the Four Nations Tournament. The ladies from USA were outmatched because they only had only 9 players return from the Salt Lake Olympic team to Team Canada\'s 16. Goaltender Kim St. Pierre made another outstanding performance worthy of her predecesor Manon Rheaume by stopping 26 shots and 6 powerplays. The games are being held in Kitchner, Ontario and surprisingly the USA vs. Canada tournament game was played in front of 5,371 fans Kitchner Memorial Auditorium.  The other two teams involved in the tournament are sports international rivals Sweeden and Finland.   '),(4630,671748,'05,November,2002','2002-11-05 00:00:00','    Most Americans would like to take one of our rights seriously. Especially those rights dealing with censorship. Sometimes, we take those rights for granted. I would like to remind you all that it is important in a democracy that people vote. It\'s how decisions are made that affect your lives!  That being said, please go vote at the NHL.com page to vote for your favorite NHL all-stars to compete later this season in the 53rd annual All Star game in Miami, Florida on February 3rd, 2003!  I am glad that they have finally dumped their North American vs. The World format for the more traditional East vs. West Conference. For a Pittsburgh fan like me, it\'s hard to pick which all star team to root for because most of the Penguins all-stars are European. That would take away the attention of my home country and players for North America. At least this way, I know to personally cheer for the Eastern Conference! In my opinion, I think they should save the America vs the World format every four years when the Olympics are played.   '),(4631,671748,'02,November,2002','2002-11-02 00:00:00','    I got another amazing opportunity in the hockey world.  Working on another assignment for WRSC/WBLF, I went over to the PSU Ice Rink for the PSU vs. Towson game. I recognized the head coach of the Wilkes-Barre Scranton Penguins, Glen Patrick in attendance. Knowing the Patrick family as friendly and approachable, I politely made conversation with him and introduced myself as a radio broadcaster. I asked Glen if he would go on the air for an interview after the game. I had already run into his brother Craig at the icers games - Craig being the GM of the Penguins.  My friend Dan was really impressed as I talked at length with Coach Patrick about the Penguins franchise, it\'s players, and talking about recruiting from the college ranks. Yes, I did get a recording of it and hopefuly use the experience in my career as a broadcaster again someday soon.   '),(4632,671748,'28,December,2002','2002-12-28 00:00:00','    Keeping up to date in the Hockey world, the World Junior Championships have been going on and according to the Erie Otters website, some former and current Otter Alumns have been doing great thus far. Carlo Colaiacovo helped out this native Canada shut out the powerhouse Czech Republic team today. In other news, former Erie player Michal Kokavec scored two goals to help Slovakia win against a imerging Belarus national team on Friday December 27th.They exploded with 11 goals! Way to go Slovakia! The U.S. team is 1-1 so far and is lead by many players from the NCAA level.  As of tonight, the standings look like this:   Group A:    Russia , 2-0; 4 pts.  Slovakia , 1-1; 2 pts.   Switzerland , 1-1; 2 pts.  United States , 1-1; 2 pts.  Belarus , 0-2; 0 pts.   Group B:    Canada , 2-0; 4 pts.  Finland , 2-0; 4 pts.  Czech Republic , 0-1; 0 pts.  Sweden , 0-2; 0 pts.  Germany , 0-2; 0 pts.   '),(4633,671748,'25,December,2002','2002-12-25 00:00:00','    Merry Christmas  everybody!!!!   '),(4634,671748,'19,December,2002','2002-12-19 00:00:00','    urlLink  ESPN vs. NHL Petition online   Please sign the petition to keep hockey on ESPN and don\'t let it be cut back by the current Commissioner Gary Bettman. ESPN has decided not to air their recap show NHL 2 Night on their flagship station ESPN and let it be only shown on \"The Duce\" ESPN 2.   '),(4635,671748,'17,December,2002','2002-12-17 00:00:00','    In an interesting move in Denver today, the Colorado Avalanche fired their head coach Bob Hartley and replaced him with the assistant coach Tony Granato.   Note: Cammi was the color commentator for Sharks hockey when her brother Tony played for them in the mid to late 90\'s   I guess that the management \"higher ups\" think that Colorado lost their premier team status in the NHL and is now considered not an invincible team. Roy has had his problems in goal. Sakic has been on the shelf twice this season.  Stop the bleeding.  No word as to weather or not Cammi Granato will put her career on hold to be an assistant or the Avalanche\'s braodcaster.   '),(4636,671748,'17,December,2002','2002-12-17 00:00:00','    ...from the same person as heard on the Pittsburgh Penguins Message Board -  snagswolf     \"Unless you can back it up I\'ll just write you off as just another Flyer fan talking out of his ass.\"       \"F--k the Lady Byng!!! Get out there and kick some ass!\"   '),(4637,671748,'15,December,2002','2002-12-15 00:00:00','    The Pittsburgh Penguins set a team record, scoring eight goals in one period (the second) vs. the San Jose Sharks.   Why couldn\'t they do that the other day??!!   '),(4638,671748,'15,December,2002','2002-12-15 00:00:00','    Some of the Stanley Cup Champion RedWings fell ill to a case of food poisoning on a recent flight from Dallas to Detroit. Now, I\'ve flown to Detroit once and it does get you sick, but on the menu for the team were Cheesesteaks (hopefully not Philly style) and Coleslaw. Holstrom and Larionov were the most affected and the Wings to turn Dandenault into a makeshift foreward for today\'s game against St. Louis.   Sometimes these things happen, but I\'d bet soon teams will add dietitians to their coaching staffs.   '),(4639,671748,'14,December,2002','2002-12-14 00:00:00','    Nothing says NHL hockey like soaking up the sun, beaches, palm trees, and Sheryl Crow.  On Decmber 10th, the NHL annonuced that the weekend concert called \"NHL Block Party\" (sponsored by Nextel) will have Grammy Award winning singer Sheryl Crow for the NHL All Star Game weekend. The concert itself is set to be February 2nd. It is $50 for \"premium\" seats, $35 for \"reserved\" seats, and $20 for General Admission.  I guess the NHL is going after the \"easy-listening\" crowd to expand their reported slumping fan base. I would rather see bands like Jimy Eat World, Disturbed, or a new band that I\'ve heard played a lot lately - Trapped performing their hit single \"Headstrong\" to play a more Modern Rock concert.   '),(4640,671748,'03,December,2002','2002-12-03 00:00:00','    Who says star athletes don\'t get special treatment?  Peter Worell was arrested on June 16th of this year and sentenced to 10 days in jail for a DUI. According to the Associated Press, he does not have to report to jail until the team finishes it\'s season. Worrell, the Florida Panthers winger is supposed to turn in to the county jail on April 7th 2003! How many of you would like to choose your own punishment?  Along with jail time,. Worell was also fined $500, had his licence suspended for 5 years, and placed on probation for 1 year.   '),(4641,671748,'28,January,2003','2003-01-28 00:00:00','    Okay.. I\'m torn.  On a mission to the College [Chinese] Buffet to feed a friend and I during late night working at WRSC studios... I was listening to our competitors in my car... just to spy and see what they were up to. While a commercial spot was running, they aired a local promo for ESPN Radio saying that they are the home for next Sunday\'s NHL all star games and games for the rest of the season. So, you know how much I love the game of hockey, but can it overcome my guilty feeling to listen to my competitors? If Nielson calls, I won\'t be eligible so I guess me listening doesn\'t count!  Tonight also marks the NHL debut of Konstantin Koltzov, the Belarussian player that they drafted in the first round in 1999 and had trouble signing. He spent most of the year in Scranton minors.  Also, I learned today that my hero Mario Lemieux would not play in the All-Star game because of a lingering groin injury.   '),(4642,671748,'24,January,2003','2003-01-24 00:00:00','    A good question was posed to the pundits at ESPN recently...     Hey! Of all the currently active players, who do you think would be the most deserving/likely to have a trophy named for them in the future? I\'d say Ron Francis, Patrick Roy, or Lemieux. And do you think there will ever be a Gretzky Trophy?      My response..... I think there should be a  Mario Lemieux Ambassador of the Game of Hockey Award  someday.   '),(4643,671748,'23,January,2003','2003-01-23 00:00:00','    From a correspondant Kayla (Thanks, KayKayHoliday)... here is a few quotes from Gary Thorne and Bill Clement... the two biggest boneheads in the ESPN Hockey Broadcasting crew!    \"I wonder why Martin Straka went over to his teammates Josef Stumpel before the faceoff\"   Geez!! Since when are they teammates? Stumpel plays for Boston!    \"Ian Moran and his little friends almost didn\'t play tonight, him, Dan LaCouture and Keith Primeau almost didn\'t play tonight cause of whiplash because of a fender bender on the way to the arena.\"    Keith Primeau plays in Philadelphia. I think he got confused with his brother Wayne who is in Pittsburgh. And I would like to know who the hell Ian Moran\'s little friends are!!!   '),(4644,671748,'18,January,2003','2003-01-18 00:00:00','    Well, it couldn\'t have lasted for Theo Fleury.  As reported from a Columbus strip club, Theo Fleury was involved in a fight suffering a black eye and bloody lip. Apparently, the NHL thinks he should get another chance since being on Stage 3 of the NHL\'s substance abuse policy. It is basically on probation status. The Blackhawks signed Theo in hopes to turn his act around after the NY Rangers (of all teams) dumped his sorry ass this past summer.  I am not surprised at this decision that NHL will allow Fleury to play after he blatantly broke his probation status. After what happened to Peter Worrell the other day (see blog entry dated 12/3/02) that athletes get special treatment when it comes to breaking the law.   '),(4645,671748,'18,January,2003','2003-01-18 00:00:00','    As Penn State misses their Head coach and several of their big star players... Team USA is down in the dumps over in Taraviso, Italy for the World University games.  Team USA lost to their rivals to the north in Team Canada 5-1. The lone goal from Team USA came from Neal Price... a PSU player.   '),(4646,671748,'17,January,2003','2003-01-17 00:00:00','    Okay, just to give you guys an update on the World University Games in Italy... the Team USA is not doing very well.  Coached by my friend and Penn State Icers head man, Joe Batista, the USA squad comprised of Junior Hockey status players from Universities all over the country. They unfortunately lost today against team Finland 3-2 and are now winless in the first 2 qalifying rounds. They are set to play rivals Team Canada on Saturday.   '),(4647,671748,'16,January,2003','2003-01-16 00:00:00','    The Ottawa Senators team scored 23 goals (the most goals scored in one NHL game) was played many years ago. Final Score 23-2.   '),(4648,671748,'12,January,2003','2003-01-12 00:00:00','    Rick Tocchett was named as a assistant coach to the NHL\'s Colorado Avalanche yesterday. Congratulations to him!   '),(4649,671748,'12,January,2003','2003-01-12 00:00:00','    The NHL\'s longest active consecutive games played streak has just come to an end.  Tony Amonte sat out during the Coyotes game vs. the Nashville Predators due to a rib injury suffered at the previous game. The 8.5 year NHL veteran\'s streak lasted 453 games. Before that span, Amonte has only missed 2 other games in his career.  Talk about loving what you do for a living!  On a personal note, I have never really needed to take a day off during my jobs. I commend him for his work ethic and dedication.  Now, the record for consecutive games stands at 964 games by Doug Jarvis running from 1975 to 1987. The next closest guy to that is Maple Leafs defenseman, Robert Svehla at 343 games.   '),(4650,671748,'06,January,2003','2003-01-06 00:00:00','    I can\'t remember where I saw it last, but I once read over this article online about the state of hockey with regards to paying European players. Most European imports to the NHL perfer to be paid in American dollars (along with \"smart\" Canadians who know that our dollar is worth more). With the upcomming labor agreement... it is a posibility that a more feasable way to pay European players is to pay them in Euros with garunteed money.  I don\'t know what effect it will have on the hockey market, but it sounds like an interesting plan.   '),(4651,671748,'04,January,2003','2003-01-04 00:00:00','    To keep up with a few friends online through instant messanger... I sent out this line to them...    While you were sleeping, I married Canadian fitness model Trish Stratus, traveled the continent attending hockey games and own/general manage a little junior league team in Regina, Saskatchewan. Oh... wait... that was while  I  was sleeping     I thought it was funny! Hey, it  might  happen!!   '),(4652,671748,'03,January,2003','2003-01-03 00:00:00','    Tonight, the Mellon Arena in Pittsburgh is host to the University Div. I Ice Hockey Series. My Penn State Nittany Lion Icers are playing Washington and Jefferson College. They are going to be aired on my station 1390 AM WRSC at 8 PM. Earlier, Ohio State\'s Division I team also played Princeton University. (I was rooting for Princeton.)   '),(4653,671748,'02,January,2003','2003-01-02 00:00:00','    (from his column on NHL.com)     \"Big-name stars like Pittsburgh\'s Mario Lemieux... aren\'t likely to dealt before the deadline.\"     Yea... the owner of his own team is going to trade himself so he could play against the team he owns! Good one, Wigge!! Way to go out on a limb! Take a risk here! This is an example of cutting edge journalism that the world needs!   '),(4654,671748,'01,January,2003','2003-01-01 00:00:00','    Happy New Year................  2003!!!   '),(4655,671748,'23,February,2003','2003-02-23 00:00:00','    urlLink Pittsburgh Tribune Review  Article Link    Goalie Robbie Tallas left a 5-2 loss to the Utah Grizzlies early on Wednesday when a spray of ice scratched his cornea    Oh-no. If a little bit of frozen water can take this guy out... umm.. maybe he should not be playing hockey.   I don\'t understand... how could that scratch his cornea? It would melt upon touching his eye, no?   '),(4656,671748,'22,February,2003','2003-02-22 00:00:00','    Today in 1980, the  Miracle on Ice  team scored a victory in the 1980 Lake Placid Olympic Winter Games against the rival Soviet Union Team. Two days later they won the Gold Medal vs. Sweeden.   '),(4657,671748,'21,February,2003','2003-02-21 00:00:00','    While at dinner with my new friend Ben from Brisbane (who looks a lot like a young Jaromir Jagr), I was watching an Isles and \'Lanche hockey game and they had this amazing stat on there to share with all of you.  Arron Asham, defenceman for the New York Islanders played 40:14 in a game last October which was an NHL record.   He was certainly earning his paycheck that night!   '),(4658,671748,'16,February,2003','2003-02-16 00:00:00','    \"She\'s an Original,\" reads the front cover of Parade magazine, the Sunday newspaper insert. Avril Lavigne posed provocatively with microphone in her 18-year-old hands advertises the breif interview by columnist James Brady.  The cover insists that one of her favorite sports is hockey, she played on a boy\'s team from the age of 10!  Holy cow, she\'s awesome! Here\'s an excerpt from the online version of the article    In boys hockey there\'s a lot of body contact,\" she says. \"I would\'ve gotten killed on the ice. So I dropped out and played on a girls\' team.\" Like most hockey fans, Lavigne is a big fan of the great Wayne Gretzky. \"Gretzky is huge,\" she says. \"He\'s still a hero in Canada. I\'d like to meet him.\"   '),(4659,671748,'14,February,2003','2003-02-14 00:00:00','    Happy Valentines Day!!!   '),(4660,671748,'13,February,2003','2003-02-13 00:00:00','    Yesterday, my friend Dan and I took a trip to Pittsburgh to see the Penguins. It was a long trip from the middle of PA, but it was worth it going home again.  We get there and I was supposed to meet a friend of mine, but things got crossed up and I got in before with tickets. After we sat down in Mellon Arena, a row of obknoxious people kept getting up for beer. I guess they missed the announcement to wait in your seats unil the next period and be considerate to other fans.  The game between the Penguins and Senators was a shutout. Unfortunately, it was a shutout for the Senators. They won 3-0.  Hedberg the goaltender tried - really, he did.  The second goal was scored by Marrian Hossa and the other two was scored by an AHL call up that I never heard of. For the Penguins, my first look at the new acquirees proved just what the rest of the Penguins fans on the Penguins message boards were expecting. God, WE WERE TERRIBLE!  On the way out, it was just as frustrating as watching the game. To redeem the cupons that all fans received entering the Mellon Arena, Dan and I had to go through this labarynth of hallways down to gate three to pick up the Joe Mullen bobblehead collectable.   '),(4661,671748,'10,February,2003','2003-02-10 00:00:00','    As soon as I got up, I knew it was going to be a bad day. I woke up this morning with a bad headache that left me incapacitated for at least most of the morning and afternoon. Later on, I got up (a hot shower really helped) and visited my buddy Dan. The day just got worse!   \"Hey, guess what.... they traded Kovalev!!\" says my radio partner, Dan, as I walk into the door.  Prior to this, all he did was make jokes and tell me they traded Kovy. This time, I didn\'t want to believe him. Sure enough, I looked at all the Internet sites and it was plastered there in big black lettered headlines.  I still didn\'t want to believe it.   '),(4662,671748,'09,February,2003','2003-02-09 00:00:00','    Ah-Ha!! Now my fantasy is at least half way realized!   Maple Leaf Sports and Entertainment   urlLink  Toronto Maple Leafs for Sale , with Raptors and ACC thrown in.  If I only had the money to buy the Leafs, that would be awesome! I could co-own it with Toronto native Trish Stratus and really go on a marketing promotions frenzy to juice up the NHL and rescue it from the Public Relations disaster that it is in.  I am currently searching for all loose change in my pocket and couch cusions!   '),(4663,671748,'08,February,2003','2003-02-08 00:00:00','    Mario scored 4 assists today at Boston\'s Fleet Center to get his 1,000 assist. The clinching goal was scored by \"enforcer\" Steve McKenna on a fluke-ish soft goal that newly acquired Boston Goalie Jeff \"can\'t\" Hackett let in late in the game.   '),(4664,671748,'07,February,2003','2003-02-07 00:00:00','    Two hockey players in the NHL have been in the front page headlines recently looking for Milestone goals in their career.  Jaromir Jagr scored his 600th goal this past week and Detroit\'s Brett Hull is trying to reach his 700th. Brett was shut out yesterday against the Colorado Avalanche and now has waited 5 games to score that elusive milestone.  Mario Lemieux needs 4 assists to get 1,000 for his career. The next game is at Boston with Regional coverage on ABC.   '),(4665,671748,'02,February,2003','2003-02-02 00:00:00','    The first period was missed due to my covering the Penn State Wrestling on WRSC, but I did get to see the better part of the show. Danny Heatly of the Atlanta Thrashers (ick!) scored 4 goals and became the youngest player by one day to have done that in an All-Star game. The action was really fast and ended with a fantastic display of goalie saves that took a 4-4 tie into overtime. Even the casual fan could tell that the players were actually getting into it and tried their best to get the game winning goal. Both Marty Turco of the Dallas Stars and Patrick Lalime of the Ottawa Senators were phenominal in goal in the third period. My boy Marian Gaborik even had a chance for the game winner at the very end, but clanked it off the post.  It took a shootout to determine the winner! Three of the five Western Stars had scored for their Conference to secure the game victory, but Danny Heatly had the last laugh. With a toothless grin, he won the MVP and a new Chevy Blazer.  I\'m going to look forward to going next year to Minnesota!!   '),(4666,671748,'02,February,2003','2003-02-02 00:00:00','    Yesterday was the skills competition. I had stayed up late to see the replay on ESPN last night. I was really impressed with the Young Stars or Rookie All-Star game. Although there was only two names I actually recognized, the speed of these players coasting up and down the ice was amazing. Of course, the idea of a running clock probably had them in a \"hurry-up\" offense too!  The Skills competition by the other veteran all stars was not too bad. The Western Division won 15-8 I believe. The most impressive thing was seeing Martin St. Louis stretch out and beat one of the faster stars in the NHL, Teemu Selane in a puck-control relay race. The objective was a full sprint down the length of the ice and then weaving back through maintaining control of a puck. I have to be glad for Marian Gaborik who turned out to be the fastest skater with the quickest lap time. The oddest thing, and probably overlooked, was that the Toronto Maple Leafs were not represented. Ed Belfour and Mats Sundin were both selected to go, but had to bow out at the end. They didn\'t bother to replace them with another teammate. I guess it was an oversight.  If the Penn State Wrestling doesn\'t go overtime, I may just make the beggining of the All-Star Game that starts at 2:30 PM Eastern time on ABC. Be sure to read here what I thought about the whole show either later tonight or tomorrow evening.   '),(4667,671748,'31,March,2003','2003-03-31 00:00:00','    Tonight marks the night that the Philadelphia Flyers play the Pittsburgh Penguins during the season. I want to go downtown to Sports Cafe & Grille because it may be my last time to relive my \"glory days\" of when I used to go there all the time dressed in my Penguins jerseys surrounded by Flyers fans.  There was one time where this Flyers fan shoved me in the back celebrating a Flyers goal. I didn\'t retaliate because the bartenders threw him out! hehehe... it\'s good to be one of \'da regulars.   '),(4668,671748,'30,March,2003','2003-03-30 00:00:00','    There may be hope for the National Hockey Leage as it announced a brand new mareting agreement with the appropriately named, \"The Hockey Company\".  I can\'t write it any better than the paragraph or two mentioned in the article taken from NHL.com.   To capitalize on the opportunities and new demands of the agreement, The Hockey Company and the NHL will create a Research and Design Center in Montreal dedicated to developing innovative, high performance apparel products. The new NHL/THC Research and Design Center will serve as a key resource for the development of the NHL Member Teams\' authentic uniforms. The agreement also includes an annual multi-million dollar marketing commitment from THC to the NHL and each of the 30 Member Teams.   '),(4669,671748,'30,March,2003','2003-03-30 00:00:00','    I got this article from the  urlLink  Baby Pens Website  talking about the growing concern that some hockey fans felt that the new safety netting was obstructing the view of people in the arenas. The First Union Arena at Casey Plaza took some focus group polling and found that the transparent fishing line was less bothersome than black nylon.   '),(4670,671748,'27,March,2003','2003-03-27 00:00:00','    I\'m sure that you\'ve seen some of the  Best Buy commercials where the customers get lost in their daydreams while playing with the displays. While watching ESPN waiting for the Thursday night game to come back on, I was really genuinely surprised at the commercial I saw next.  A man dressed in street clothes was on the rink in the middle of a game. It seems as if he is the only one playing. Ducking and weaving on the ice, but from out of nowhere and before he is pancaked along the boards by 4 NHL hockey players in full gear (and recognizable jerseys) he passes to his female companion (wife/girlfriend/mistress??) who is also in street clothes, does this great little spin move at center ice and then throws a backhand through the goaltender for a goal! From there the fantasy scenario turns to some reality as the man and lady were shopping at Best Buy  The recognizable team was the Minnesota Wild featured in the commercial. This would make sense since the Best Buy Company has their home office in Minneapolis, MN.   '),(4671,671748,'26,March,2003','2003-03-26 00:00:00','    Ever since I talked about the controversial and shocking story of fans booing the National Anthem, \"some guy\" named Don Cherry waxes poetic on the issue, which in turn creates a bigger issue. Should athletes and celebrities way in on the war; swaying public opinion? This article from  urlLink  Mark Spector  of the National Post delivers an opinion column of the way Cherry addressed the war. In the segment \"Coaches Corner\" in CBC\'s broadcast of Hockey Night in Canada, legendary NHL coach Don Cherry gave support to the U.S. for the war on terrorism and condemned his Canadian contrymen for being disrespectful to America for their lack of support.  Don Cherry, always a stand up guy, chose to take some flack, but has not strayed from his \"old school\" views.   '),(4672,671748,'24,March,2003','2003-03-24 00:00:00','    Harvard University lost the NCAA Women\'s Championship to University of Minnesota-Duluth. It took 4:19 left in the second OT for Nora Tallus to put one past goalie Jessica Ruddock.  I\'m sure my gals here at Penn State could kick both team\'s buts!   '),(4673,671748,'22,March,2003','2003-03-22 00:00:00','    Congratulations to the Minnesota Wild for earning a playoff berth in their 3rd year of existance. They defeated the division rival Detroit Red Wings in a 2-0 shutout.   '),(4674,671748,'22,March,2003','2003-03-22 00:00:00','    FOX Sports Radio News had reported today that since word had gotten out that our anthem was booed by some really rude cannucks in Montreal, the Canadian anthem was booed by many just as rude fans in Florida in retaliation.  Okay guys, this is really gettting out of hand. Stop it!!   '),(4675,671748,'21,March,2003','2003-03-21 00:00:00','    Since I had been early to work, I chose to go to the QWK Rock studio next door and talk to the young lady working Saturday mornings.  She had on this long sleve red t-shirt on. I noticed that there was a little logo on it that looked like the state of Idaho. The blond female D.J. told me that she was originally from there and it was an  urlLink  Idaho Steelheads  t-shirt. She told me that is where she was originally from.  I researched the franchise after I told her that I wasn\'t that familiar with the team. The Steelheads are an WCHL minor league team that plays other west coast teams. The WCHL is comprised of several teams in cities like Las Vegas, Fresno, San Diego, Bakersfield, Long Beach and Ankorage. Idaho is in first place in the WCHL with 100 pts after 65 games played.  Go Idaho!   '),(4676,671748,'21,March,2003','2003-03-21 00:00:00','    Thank God for someone in Canada with some class! Montreal Canadians President Pierre Boivin issued this appology:    On behalf of the entire Montreal Canadiens organization, I would like to express my deep regret over the conduct displayed by certain fans prior to Thursday night\'s game against the New York Islanders.During the performance of The Star Spangled Banner, some of the fans in attendance at the Bell Centre chose to express their political beliefs by jeering the American national anthem. It is our firm belief that this kind of behaviour has no place in the context of professional sports. The Montreal Canadiens organization has always held a high respect for its neighbours and friends in the United States, and we look forward to maintaining this strong and positive relationship. We apologize to anyone who may have been offended by this incident, and would encourage all fans at the Bell Centre to conduct themselves in a manner worthy of our game and our two great nations.\"   I think I might write this guy and say thanks for trying to set things straight.   '),(4677,671748,'20,March,2003','2003-03-20 00:00:00','    From the  urlLink  Ottawa Citizen     The Ontario Hockey League has fined the Sault Ste. Marie Greyhounds $50,000 and suspended former coach and general manager John Vanbiesbrouck indefinitely from the league for the racial slur (the \"n\" word\") he directed at team captain Trevor Daley.  \"It is imperative that the Ontario Hockey League make a strong and unambiguous statement to the effect that any statements or comments of this nature will not be tolerated by the league,\" Branch said in a release. Vanbiesbrouck may apply after five years for reinstatement, he said.    Cute, guys. Real cute.  I am not in the kind of mood to hear this news. I\'m not defending at all for what he did, it was wrong and that word should be erased from American-English lingo, but my gosh!! The thought police are on a roll! Way to take a man\'s livelyhood for one stupid remark.  Like I said.. I\'m not having a good week and just reports like this are rubbing me the wrong way.   '),(4678,671748,'18,March,2003','2003-03-18 00:00:00','    According to an article in the  urlLink  NY Post   Canadians that boo us during our National Anthem at the time we are trying to do the good thing and take down a dictator that will surely kill us all in a bink of an eye.  Okay.. this is going too far.  Classless people, truely classless.   '),(4679,671748,'18,March,2003','2003-03-18 00:00:00','    Tonight Martin Straka scored his 500th career point against the Detroit Red Wings at home by scoring a terrific backhanded goal on Manny Legace. Congrats to him! Unfortunately, The Pens lost another one 5-1 tonight.   '),(4680,671748,'11,March,2003','2003-03-11 00:00:00','    Well, the deadline for the NHL to trade was at 1 PM MST.  Many teams went nuts trading for draft picks, vying for the best player to get them either into the playoff hunt or carried through the Stanley Cup playoffs.  Lots of key veterans, overpriced and taking up salary on teams that can\'t afford to keep them, guys like Tone Amonte wound up with his best frien Jeremy Roenick and the Philadelphia Flyers. Phil Housley went to the Toronto Maple Leafs along with old man Doug Gilmour.  If you are a Penguins fan, you are pissed like me to see the entire team get thrown away and young guys only making $40,000 a year.   '),(4681,671748,'06,March,2003','2003-03-06 00:00:00','    Congratulations to Mats Sundin for scoring his 1,000 career point in a game at Edmonton yesterday. Upon returning to the bench, he picked up the puck, pitched it to a equipment manager and plopped down in his seat. Job well done, Mats!   '),(4682,671748,'06,March,2003','2003-03-06 00:00:00','    After returning to State College from a breif vacation to see my brother perform as  The Near Professionals , I checked my email & the news around the NHL world.  I was surprised to see that the San Jose Sharks had traded Owen Nolan to the Toronto Maple Leafs. In return, they got Alyn McCauley and Erie Otters standout Brad Boyes.  The trade looks great to the Maple Leafs who are looking to a strong push towards the playoffs. Their offence was lagging a little and looked for a way to boost a quiet Mats Sundin. I was really impressed with Brad Boyes having seen him play in Erie for major Junior OHL hockey. He looked to be a great prospect. Both new S.J. Sharks Boyes and McCauley have both been rumored to be offered to the Penguins for Alexei Kovalev.  The trading deadline for the National Hockey League is March 11th.   '),(4683,671748,'03,March,2003','2003-03-03 00:00:00','    Okay... so I got my brother, my best friend, and my radio co-worker into a Yahoo! Fantasy Hockey league over the internet. For the most of the begining of the season, I was ahead in point totals. Then, people started to lose interest because it wasn\'t fair that the one guy running the leage was ahead. It wasn\'t any fun.  After teaching some tricks and advising the other players, I found myself and my team in a serious scoring slump. I am now in last place. The  Ruthless Agression  has now been reduced to a mild advancement.  I have now decreed that for the rest of the year my players will wear a patch with \" WTF \"   '),(4684,671748,'03,March,2003','2003-03-03 00:00:00','    This week\'s   The Hockey News   has a cover story that has the hockey world gone ape shit.  Sergei Fedorov, center for the Detroit Red Wings has admitted in an article that he did have a serious fling with tennis/model (and fellow Russian) Anna Kournickova. In the article, he admits to a brief marriage to her. Something like this was kept a very closely guarded secret, but comments by Barry Melrose would have you believe that it is not that big of a deal to a hockey team such as a Detroit. They all new about Sergei and Anna\'s relationship.  I think Sergei is one of the luckiest guys on the Earth! But much like most Hollywood marriages, it never works out due to conflicts in interest. Stars and famous celebrities often marry or \"hook up\" because very few people can relate to fame the way other stars do. Sergei and Anna have simmilar cultures and used to the limelight, working in America they more than likely formed a common bond. However, the questionable nature of their relationship is that Anna is very very young. Sergei and Anna were also never in the same country together long enough to make a relationship work. The constant traveling and touring that celebrity pro athletes have to do is very stressful and makes for a life of living out of a suitcase. Their eventual split was no doubt an effect of the afformentioned problems.   '),(4685,671748,'02,March,2003','2003-03-02 00:00:00','    While flipping through the channels & getting my weekly Monday Night \"Stratusfaction\". I watched a little bit of the NY Rangers and NY Islanders game. During the intermission, the ESPN Commentators John Buccigross and Barry Melrose gave their picks for the Stanley Cup Finals.   Barry: New Jersey and Dallas John: Toronto and Detroit   I would like to agree with fellow Pittsburgh fan Buccigross be right, but Barry knows the teams better. Since I hope Pittsburgh overachieves and sneaks into the playoffs I\'d natually pick them for the finals. Realistically, I\'d want the two other favorite teams to root for to get to the playoffs. I have a growing respect for the history and tradition for the Toronto Maple Leafs. They have a realistic chance of making it in the Stanley Cup finals representing the Eastern Conference. The team that surprised the leage in the Western Conference is the team currently sitting in 6th place. The Minnesota Wild should surprise somebody and play spoilers in the playoffs like the San Jose Sharks a few years ago.   Mike: Ottawa and Vancouver   '),(4686,671748,'02,March,2003','2003-03-02 00:00:00','    Montreal Canadiens set an NHL record for most goals in a single game. Montreal defeated the Bulldogs 16-3 in Quebec City.   '),(4687,671748,'02,March,2003','2003-03-02 00:00:00','    ACHA D-I Championships @ Athens, OH (Ohio University-Bird Arena)   #2 Penn State is looking for a 4-peat and 7th national title overall.   Today they face off today at 2:00pm   ACHA Division II Championships @ Anaheim, CA (USC-Disney Ice)   #2-Northeast, Penn State (D-II) and #3-Southeast, Indiana University of Pennsylvania both look for their first ever national championship. Both teams have been to the 16-team dance several times in the past few years.  IUP begins with #2-West, Colorado State today at 12:45pm on the NHL rink and PSU takes on #3-West, Weber State at 4:00pm on the Olympic Rink.   '),(4688,671748,'30,April,2003','2003-04-30 00:00:00','    Today, I just went ice skating at Penn State Ice Pavilion. I had a good time.   '),(4689,671748,'25,April,2003','2003-04-25 00:00:00','    Slovakia\'s defense of the World\'s Top Hockey Country begins today with 16 international teams participating in the IIHF World Championships beginning Saturday April 26th. This year, the tournament will be settled in Finland.   A lot of the NHL stars will represent their respective countries as they have done so last year and in the Salt Lake Olympics. Jaromir Jagr is not on the Czech Republic roster since the Capitals reached the first round of the playoffs. Many of Canada\'s stars turned out on thier roster. In the peliminary rounds already, USA lost to Denmark and Canada defeated team Belarus. Since when did Denmark get a hockey powerhouse????   GROUP A   Slovakia   Germany   Ukraine   Japan    GROUP B  Russia USA Switzerland Denmark   GROUP C  Sweden Canada Latvia Belarus   GROUP D  Finland Czech Rep. Austria Slovenia   '),(4690,671748,'24,April,2003','2003-04-24 00:00:00','    Okay! The first round is over... time now for the elite 8 in the NHL to survive to the Conference finals.   Eastern Conference   Philadelphia vs. Ottawa  In a rematch of last season\'s first round, Ottawa and Philadelphia square off in what is sure to be another classic playoff series. Tony Amonte has to score a hole lot more to prove acquiring him for the sole reason of the playoffs wasn\'t a bust. It\'s time to start serving the Wheaties in the locker rooms. The Senators proved that they didn\'t need Yashin to not get \"voted off the island\" and sent the former holdout with the resurgent Islanders packing. The Senators are the top seed in the East and it will be just as difficult eliminating them as chopping down the 6 ft. 9 in. \"tree\" named Zedno Chara.  Prediction: Senators in 6     New Jersey v. Tampa Bay  The second seated Devils had a short time taking care of Boston in the first round. Though New Jersey had them on the ropes early in the series, Boston was eliminated in 5 games. Martin Brodeur continues to be the clutch goaltender he always was stopping a barrage of shots day in and day out. The matchup of the series is Brodeur versus Martin St. Louis who took more capital offense in the last series. This is the furthest that the Lightning have traveled in the playoffs, so the underdog Tampa Bay team has low expectations on their side.  Prediction: Devils in 5    Western Conference   Minnesota vs. Vancouver  NHL Playoff hockey has not been played in Minnesota in 10 years since. The tremendous support that the Wild franchise got in the first round proved moving the Stars to Dallas was a huge mistake. It took 7 games to shockingly upset the strong Avalanche franchise. Overtime helped the Wild get a second chance at winning taking their last two games into OT. The Wild are also undefeated when they score the first goal this year. Vancouver is also a team that had to fight back from being on the brink of elimination. Marc Crawford\'s teams always come ready to play and are poised on representing the Western Conference in the Stanley Cup Finals.  Prediction: Wild in 7    Dallas v. Anaheim  If you checked my predictions in the first round, somebody was going to be swept, but who\'d thought it was going to be the defending Stanley Cup champions. I\'ll admit that the Mighty Ducks shocked even me when they sent Detroit to the geriatric ward. The Ducks got \"jiggy\" with it and never looked back. Dallas once again proved to be the Oilers nemesis in the playoffs and fought off a tremendous crowd support in the Skyreach Center last series. The Ducks must remove the rust and shoot for the stars to continue to the next round.  Prediction: Stars in 7   '),(4691,671748,'22,April,2003','2003-04-22 00:00:00','    I enjoyed watching NHL 2night crew get ready for the upcomming games as they planned a segment featuring comedian Denis Leary.  Denis is a long time hockey fan and helps organize a celebrity show for the Neely house. Being a Boston native, Leary sported the Bruins away jersey and struck a pose or two. His message was short and only to bash the heck out of Barry Melrose. His verbal parsings targeted Barry\'s mullett. Of course, Barry being the good sport he is, took things in stride and even threw in a few insults of his own by laughing at Denis\' choice of favorite teams.  I guess you had to be there. It was really funny.   '),(4692,671748,'19,April,2003','2003-04-19 00:00:00','    Happy Birthday to me!!!  Happy Birthday to me!!  Happy Birthday to me!!   '),(4693,671748,'15,April,2003','2003-04-15 00:00:00','    The Pittsburgh Penguins today have relieved their head coach Rick Kehoe of his duties.  Coach Kehoe has spent years behind the bench as an assistant coach to the Penguins team, a team that he recently spent more than 25 years in association with. No word as to who the successor will be or if Rick will be maintained as a scout or assistant coach with the Penguins for the 2003-04 NHL season.   '),(4694,671748,'14,April,2003','2003-04-14 00:00:00','    The NHL playoffs have begun and already there have been some great games played out with the bracketed matchups that the NHL provided. Fighting it out in 4th and 5th seeds respectively are the Toronto Maple Leafs and Philadelphia Flyers. These teams have had a long history in the past and the series has proved to be an exiting one. However, the Maple Leafs were reminded of their past when  urlLink  a Flyers fan brought a law suit  from an incident two years ago. The Flyers fan is suing for damages after alegedly had been provoked into leaping the plexi-glass while being taunted by known goon Tie Domi.  \"It might be Philly trying to stir up something two years later because we\'re playing against them. It\'s two years later but, who knows, I wouldn\'t put it past Philly,\"  Domi said.   Since my beloved Penguins are not in the playoffs this year, many Pittsburgh fans have asked me who should we root for now? I gave them this advice:  1.) True Penguins fans NEVER NEVER NEVER NEVER root for the Flyers. If it was a game between the Flyers and Iraq with Saddam Hussein and his sons as the top line, You still don\'t root for the Flyers!   '),(4695,671748,'14,April,2003','2003-04-14 00:00:00','    Yesterday my dad came for a visit. He brought a surprise with him... my mail    \"To Mike,... a fellow hockey fan, love Trish Stratus xox\"   reads the 8x10 photo of the perfect figured blond Canadian fitness model. This WWE Diva is standing as the Women\'s Champion and hailing from Toronto, Canada. Nobody comes out of Toronto or lives there and not be a Leafs fan. On Spike TV\'s WWE Confidential program shows a recent tour of her home, Trish showed off her action figure collection, complete with CuJo and Mats Sundin. I have another photo of her in a NHLPA jersey. It\'s one of my favorites of her.   '),(4696,671748,'13,April,2003','2003-04-13 00:00:00','    University of Minnesota Golden Gophers have won the NCAA National Championship. This is the second year in a row for the team, making the first team to go back to back national champions since 1921. The Frozen Four as it\'s called was played this year in HSBC Arena in Buffalo, NY. They defeated the University of New Hampshire 5-1 on Saturday.   '),(4697,671748,'11,April,2003','2003-04-11 00:00:00','    Congratulations to the first Slovak and European born player to win the Hobey Baker Award as NCAA\'s 2002-03 Player of the year. Peter Senja the forward for Colorado College.   '),(4698,671748,'08,April,2003','2003-04-08 00:00:00','    Tonight, I was watching the resurgent edition of NHL 2night on ESPN2 and the usual suspects were behind the anchor desk. John Buccigross, Barry Melrose, and Ray \"Parm\" Ferraro.  During one break, Barry teased his friend John by calling him \"Ashley Banfield\" because of the glasses that John wears are simmilar to the MSNBC info-babe.When they came back from the break, John re-introduced himself by naming his co-anchors and reffering to himself as \"Ashley Banfield\".  I got a good laugh out of that. It only supports the belief that Buccigross and all the self-deprocating humor (especially in his ESPN.com colums) that make him out to be a nice guy.   '),(4699,671748,'08,April,2003','2003-04-08 00:00:00','    Okay.. for starters.. lemme just say that John Buccigross is a great guy... he has a great sense of humor and I think he sees things on the  right  side - if ya know what I mean.  The Stanley Cup Playoffs begins tomorrow night and people and pundits are ringing in with their predictions. I think I\'ll do the same.    Eastern Conference    Ottawa v. NY Islanders Ottawa enters in with a few negatives. 1. The President\'s Trophy winners don\'t have an above average chance to win it all. The Detroit Red Wings were the first team to break that curse last year. 2. Ottawa is still strapped for cash, they need to win to earn more money. 3. Alexei Yashin is an Islander now. Their once top pick has come back wearing a new uniform. The Islanders have proven that they can win, but dealing Osgood to St. Louis will prove to be their downfall.  Prediction: Ottawa in 6   New Jersey vs. Boston Boston\'s managerial and upper management got a shake up when they fired Robie Ftorek their coach and the GM is now sitting in temporary. Glen Murray and the B\'s have been on a mission all season long to return the once proud franchise back to it\'s legendary status, New Jersey is riding on the hot goalie in Martin Brodeur. They also contain a lot of veterans from Dallas in a blockbuster trade a while back. The Devils are the experienced playoff franchise. Sorry Beantown, you\'ll get \'em next year.  Prediction: Devils in 5   Tampa Bay vs. Washington If you asked who the higher seed is in this series, you\'d be surprised that Tampa Bay won the Southeastern division over their rivals Washington. Tampa has been quietly making waves thanks to their good goaltending in Graham and Khabibulin. Jaromir Jagr is \"dying alive\" with his skills in Washington and fighting the discomforting process of healing a broken wrist. Peter Bondra scored 2 goals in their last regular season game. Look for the Caps and my Uncle (in the PR office) to find a lightning rod and control the young and inexperienced \'Bolts.  Prediction: Capitals in 4    Philadelphia vs. Toronto The two teams taht made the major deals to re-tool and improve for a serious playoff run find themselves against one another in the first round. Roenick has his buddy Amonte back and are playing well. Hitchcock has coached the Stanley Cup Stars; he knows what it takes to make it for the long haul. Toronto awaits their return on investment with goalie Ed Belfour and RW Owen Nolan. The Leafs look stronger and battle tested than Flyers goaltender Checkmanek. The Flyers are looking at another first round exit if Amonte and Roenick can\'t score.  Prediction: Maple Leafs in 7  .    Western Conference   Dalas vs. Edmonton The Stars were the best team in the Western Conference earning the home ice advantage throughout the playoffs. The new world Oilers retooled themselves for a better showing in the playoffs this year. Dvorak for Carter will prove that it was a lopsided trade in favor of New York. Tommy Salo treats every shot like a sweedish meatball. Dalas has the kind of firepower to knock a few of those back in on 2nd and 3rd tries.  Prediction: Stars in 6   Detroit vs Anaheim The Mighty Ducks are the surprise team in the playoffs. Led by Peter Sykora and Paul Karya, Anaheim is a scoring threat that could hold their own. However, Detroit is a team that has the right mix of age and youth. They can score like water off a mallard\'s back. Brett Hull scored over 700 goals for his career this season and won\'t hesitate to score some more. Detroit is the defending champions and will play like that through the series. Curtis Joseph is trying to achieve what he couldn\'t for Edmonton or Toronto. Maybe the agless veterans on Detroit\'s team all retire at once for going out on top.  Prediction: Wings in 4   Colorado vs. Minnesota In a surprising turn of events, the Minnesota Wild fought hard to get a playoff berth in only their third season. The young team is lead by Slovak phenom Marian Gaborik and the defensive trap system. Colorado had a bum year losing Joe Sakic for most of the first part of the season due to injury. Patrick Roy hasn\'t looked like his old self this year and even let the team down when Minnesota won a few games at their expense. This is a tough call because Minnesota is good enough to shock some people, but Colorado has the dominant team.  Prediction: Avalanche in 6   Vancouver vs. St. Louis It\'s put up or shut up for Vancouver. The hottest team in the league scoring wise is Vancouver with Bertuzzi, Naslund and Brendan Morrison as the saviors of most fantasy leauge teams. They gave the defending champs Detroit a run for their money last year in the playoffs and the fans support their efforts to give it another go. St. Louis stands in their way this year bringing back captain Chris Pronger from injury. The Blues were treading water with the goaltending they had only to pick up Chris Osgood at the trading deadline. Also, look for newly acquired Mr. Candice Cameron (Valerie Bure) to make an impact in the playoffs.  Prediction: Cannucks in 5   '),(4700,671748,'06,April,2003','2003-04-06 00:00:00','    WORLD CUP OF HOCKEY 2004 SCHEDULE               Preliminary round- Aug. 30 - Sept. 5             Quarterfinals - Sept. 6-8             Semifinals - Sept. 10-11             Championship- Sept. 14  European Pool (E): Czech Republic (CZE), Finland (FIN), Germany (GER), Sweden (SWE)   North American Pool (NA): Canada (CAN), Russia (RUS), Slovakia (SVK), United States (USA)   The complete World Cup of Hockey schedule follows:      Aug. 30  Hartwall Arena, Helsinki  CZE vs. FIN   Aug. 31  Globe Arena, Stockholm  GER vs. SWE *  Bell Centre, Montreal  CAN vs. USA   Sept. 1  Globe Arena, Stockholm  CZE vs. SWE * Bell Centre, Montreal  CAN vs. SVK   Sept. 2   Cologne Arena, Cologne  FIN vs. GER  * Xcel Energy Center, St. Paul   USA vs. RUS    Sept. 3   Sazka Arena, Prague  GER vs. CZE  *  Xcel Energy Center, St. Paul   USA vs. SVK     Sept. 4  Hartwall Arena, Helsinki  SWE vs. FIN * Air Canada Centre, Toronto  CAN vs. RUS   Sept. 5  Air Canada Centre, Toronto  RUS vs. SVK   Sept. 6  QUARTERFINAL - Home Arena of E1  E1 vs. E4   Sept. 7 QUARTERFINAL - Home Arena of E2 E2 vs. E3   QUARTERFINAL - Xcel Energy Center, St. Paul NA1 vs. NA4 or NA2 vs. NA3  Sept. 8 QUARTERFINAL - Air Canada Centre, Toronto NA1 vs. NA4 or NA2 vs. NA3  Sept. 10 SEMIFINAL - Xcel Energy Center, St. Paul   Sept. 11 SEMIFINAL - Air Canada Centre, Toronto   Sept. 14 CHAMPIONSHIP - Air Canada Centre, Toronto    urlLink  *From- Wild.com   '),(4701,671748,'05,April,2003','2003-04-05 00:00:00','    So would you believe me if I told you that I was friends with a former stripper in California??  This afternoon, I was talking to her via one of my instant messanger names. I asked her how she was doing and she responded by telling me that she went to a hockey game last night! Bakersfield Condores v. Idaho Steelheads. The Steelheads won that night. She said that she went and had a good time, got Suite seating, and loves hockey!   '),(4702,671748,'04,April,2003','2003-04-04 00:00:00','    Continuing the debate about talking heads and TV personalities commenting about the War in Iraq, I found an article at  urlLink  Faceoff.com  that says CBC completely hung out their most popular and controversial personality Don Cherry out to dry.  The article says that the Canadian Broadcasting Company (CBC) gave a statement that the comments that Cherry made on Hockey Night in Canada about the war in Iraq was his opinion and did not necessarily reflect the views of the CBC or it\'s management. They felt compelled to release a statement after being deluged by letters of complaints.  Cherry came out and supported the U.S. in the war against terrorism.   '),(4703,671748,'03,April,2003','2003-04-03 00:00:00','    World Cup 2004  - Toronto, Canada  urlLink  NHL.com article     The tournament, a joint effort staged by the National Hockey League and the National Hockey League Players\' Association, in cooperation with the International Ice Hockey Federation, will feature the national teams from the top eight hockey-playing nations in the world. Tournament organizers discussed many of the preliminary details of the tournament during Wednesday afternoon\'s press conference at the Air Canada Centre, which will host the final game Sept. 14, 2004. \"It\'s a real honor for us to be the host of the championship game for the World Cup of 2004,\" Ken Dryden, Toronto Maple Leafs President, said. \"We are very pleased to have preliminary round games here, a quarterfinal game, a semifinal game and the final game of the 2004 World Cup.\" According to organizers, the tournament\'s final will be a one-game affair, as opposed to the best-of-three format that was used in the 1996 World Cup of Hockey, and also in several Canada Cups. Canada, the United States, Russia, Sweden, the Czech Republic, Finland, Slovakia and Germany are scheduled to take part in the tournament.   '),(4704,671748,'03,April,2003','2003-04-03 00:00:00','    I should be in Toronto right now.  I had this plan that we would go on Spring Break with a friend of mine from work, but when he got sick, the plans were changed. We then decided to make a road trip to Toronto to see the Hall of Fame on Wednesday night and see the Wild vs. Leafs the following night. However, that didn\'t happen as a lack of money made us change the plans.  Oh well, I\'ll get to go...............  someday .  Maple Leafs 2 Wild 1   '),(4705,671748,'02,April,2003','2003-04-02 00:00:00','    Today, Mario Lemieux played in his last game..... again.  In a game vs. the Carolina Hurricanes, he set up Miloche for the game winning goal. Before the game, he decided to sit out the season ending game pondering retirement once more.  Mario... please come back... one more time!!! The team needs you! If we wanted to see a minor league team, I\'d go to Wilkes-Barre. Give it one more time to try and turn things around. It sucks that all the good players had to go.   '),(4706,671748,'28,May,2003','2003-05-28 00:00:00','    Earlier today, \"St.\" Patrick Roy announced his retirement from the NHL after an 18 year career.  Regarded as thee best goaltender in the history of pro hockey, he owns many NHL records including most wins. He also holds 3 Stanley Cup rings he won with Montreal and Colorado. The French Canadian was highly criticized for his departure in Montreal after a contract dispute with owners. He then went on to rival organization Quebec/Denver and won two Stanley Cups with them. Oh, yea, he also assaulted his wife after a temper tantrum that included ripping a closet door off it\'s hinges.  After one full year of  Confessions of a Hockey Fanatic  I began talking about Patrick Roy. Fitting that I now post about him one year later.  In other goalie news, the scapegoat of Flyers fans, Roman Cechmanek was traded to the L.A. Kings for a second round draft pick. Poor Felix Potvin, looks like he\'s on the move again from L.A. Potvin was in Toronto and was outplayed by Curtis Joseph and in Vancouver wore out his welcome. No word yet if Felix \"the cat\" will accept a backup or a pack up role.   '),(4707,671748,'25,May,2003','2003-05-25 00:00:00','    In my last installment to the Stanley Cup predictions, here is how I think the teams match up.   Stanley Cup Finals   New Jersey Devils vs. Mighty Ducks of Anahem  This series might not be the match-up that will bring in all kinds of interest continent wide, (especially in Canada when they get pissed off when an American team wins the Cup) when neither team are among the elite franchises in the league. The Mighty Ducks, a Disney company/team, are in their first Stanley Cup finals ever. They sport Paul Karyia as their team captain and best player. They also have a few former Devils on the team having worked a few trades at the deadline and in the previous off-season. Peter Sykora looks to get his third championship ring as a member of the Ducks. The Ducks have to battle the iddleness waiting 11 days for the Eastern Conference to finish up. J.S. Giguere, the phenom goaltender, hopes that he could carry them though one more series.The Devils are a determined crew of hard hitting defensive minded franchise. Scott Stevens is their Captain and team leader. This defenceman leads with his shoulder. They\'ve been able to win the Stanley Cup twice in their history.However, Pat Burns, their coach, has never been able to win it coaching other teams. The Devils have a star goaltender that can take control of the game and shut an opposing team down. Martin Brodeur will be huge in this series. It will be a series of which goaltender can be better. Prediction: Devils in 6   '),(4708,671748,'24,May,2003','2003-05-24 00:00:00','    Who would name their arena  urlLink Gaylord Entertainment Center  anyway?  The Nashville Predators the 5 year old NHL expansion team has filed a lawsuit against Gaylord Entertainment Company for not paying in January for the naming rights to the arena.  This is getting silly, folks. With the issues of Enron and the internet website that had the naming rights to Baltimore Raven\'s field, I think stories like this should change the minds of pro teams who get corporate sponsors. At least, they should be more thorough in their investigations and not give the rights over to the highest bidder.   '),(4709,671748,'23,May,2003','2003-05-23 00:00:00','    Congratulations to the Conference winners in the NHL 2002-03 playoffs. Tonight, the New Jersey Devils defeated the Ottawa Senators 3-2. They go on to face the surprising Mighty Ducks of Anaheim.   '),(4710,671748,'14,May,2003','2003-05-14 00:00:00','    The summer of 2003 issue of Hockey Digest has an article covering the tragic story of a youth hockey player near Pittsfield, Mass.  Quinn Connelly, according to the article, was diagnosed with the learning disability called dyslexia. But Quinn would find an outlet for his frustration by finding something he was good at and something he loved - hockey. His sister was his biggest supporter and fan. However, tragedy struck in early December of 2000 where Quinn was at practice. Quinn dove his body out with the purpose of blocking a shot. He turned his head back only to have the shot strike the back of the neck. It\'s the one small part on a hockey player that is not protected. The 1 oz. solid rubber puck crushed two arteries that connected to the brain. Quinn later died at the area hospital. The Connally family want to give his son a lasting legacy and build a two rink sports facility for the Massachusetts community. You can read and donate to the Memorial fund by clicking on the  urlLink  Quinn\'s Legacy Foundation  link.   '),(4711,671748,'10,May,2003','2003-05-10 00:00:00','    Today, the World Championships of International Hockey is set for the Gold Medal Game. Sweden will take on Canada tomorrow. Though the teams did not match the same \"dream team\" rosters from the Salt Lake Olympics, both teams are stocked with NHL talent from their home country. The loser of this game will earn the Silver medal and second place. In the Bronze medal game, the defending World champion Slovakia will face their traditional rivals of Team Czech Republic.It\'s a dissapointing finish for Slovakia who had high expectations of repeating as champions. Czech Republic a hockey powerhouse in recent years hopes they return to their former glory since winning the Gold Medal in the Nagano Olympics. This is the biggest match between the two countries since they parted ways more than 9 years ago. In a sureal turn of events, the two teams will have to face goaltenders who work together in the NHL. Tomas Vokoun (Czech) and Jan Lasak (Slovak) both play goal for the Nashville Predators.   '),(4712,671748,'08,May,2003','2003-05-08 00:00:00','    Center.  urlLink Denver Post  Columnist Woody Paige wrote in a column dated May 21, 2001    \"However, St. Louis somehow is supposed to win three consecutive games, two at  The Can , against the Avalanche\"    It\'s nice to know that the citizens and media in Denver have nicknamed their indoor sports facility after a slang term for the toilet. In reality, the Official real name is called the Pepsi   '),(4713,671748,'08,May,2003','2003-05-08 00:00:00','    Well, I blew it big time! I had to leave Hooters last night because they were closing at midnight and the third period had just started. I can\'t believe that I missed a great comeback from the Minnesota Wild to win 4-2!  So far I am 66% correct on the Playoff Round Predictions. (Hey, a blind squirrell finds a nut every now and then)     Eastern Conference   Ottawa v. New Jersey  The team that just won\'t go away in the Eastern Division is the Ottawa Senators. They feature a European scoring threat of Alfredson (captain), Hossa and Havlat. Despite filing for bankrupsy, the Senators have continued to find ways to win - including winning the President\'s Trophy as the team that won the most points in the standings.The undersized and defensive minded New Jersey Devils have been riding the coat-tails of goaltender Martin Brodeur. Headmanned by Pat Burns, this team looks primed for their third Stanley Cup.  Prediction: Ottawa in 6      Western Conference   Anaheim vs. Minnesota  Hot Goaltenders have been the major focus in this years playoffs. J.S. Giguere has taken the Mighty Ducks of Anaheim the furthest the NHL franchise has ever gotten. They shut down Detroit in 4 and hope to do the same to Minnesota. The current goal and point leader in the playoffs is phenom Marian Gaborik. He hopes to score Minnesota into the finals who haven\'t been there in 10+ years.  Prediction: Mighty Ducks in 5   '),(4714,671748,'08,May,2003','2003-05-08 00:00:00','    Tonight, the Minnesota Wild will make history in the NHL as they play Game 7 in Conference semi-finals tonight against the Vancouver Cannucks.  The Minnesota Wild, comming off a 7 game series defeating the vaunted Colorado Avalance when they were down 3 games to 1 now travel to Vancouver tonight to accomplish the unthinkable. The 3 year old expansion franchise is in their first year of post season play. If they win tonight, they will be the first team in NHL history to come from behind 3-1 twice in the same year.  ... Told ya it was going to 7 games! *wink*  I\'ll be watching the game from a comfy seat at Hooters!   '),(4715,671748,'05,May,2003','2003-05-05 00:00:00','    Come on everybody... sing along with me... \"na na na na..na na na na.. hey hey hey... goodbye!\"   The Flyers lost tonight to the Ottawa Senators 5-1. They should really consider getting a new goaltender over the summer. Chechmanek is the softest goaltender in the league. I watched the game tonight and 7 minutes into the game the Flyers are down 2-0. The first goal catches Cechmanek out of position. Two more goals were also scored just like that. Ottawa just put the puck into the empty net.  Philly\'s big scorers Primeau, Amonte, Roenick, et al only managed 2 or 3 goals the whole series. The offensive production is what eventually got them in the end.  Ottawa moves on to the Eastern Conference Finals against another Penguins rival - New Jersey.   '),(4716,671748,'04,May,2003','2003-05-04 00:00:00','    Hmmm... In a online interview given by some newspaper in England, m\'girl Trish Stratus lists her \"perfect man\"     \"The main thing I look for in a man is personality. I always say, and I know it sounds funny, is that I\'d like a male Trish Stratus someone who\'s quirky, with a good sense of humour and into jokes. I like sports, to watch movies and just to chill out, so if they like that then they could be my man.\"   Sounds like me to a \"t\"!      Oh.. here comes another mention that I need to get out of the way..   Thanks to Hootress, Jay Currie, and the Elephant Rants by linking my website on theirs. Thanks guys, I appretiate that!  wait.. what does this have to do with hockey??? *imitating cackle like Barry Melrose*   '),(4717,671748,'02,May,2003','2003-05-02 00:00:00','    Team USA almost missed not qualifying for the medal round in next Winter Olympics by defeating Belarus 4-2 in Relegation rounds after losing the first 3 games in Preliminary action. The loss would have had made team USA play a few more games to qualify to the Olympic tournament in Tourin, Italy.   '),(4718,671748,'02,May,2003','2003-05-02 00:00:00','    In a somewhat shocking move, the Penguins organization has fired the Wilkes Barre - Scranton coach Glen Patrick.  I remember talking to him early last fall during one of the PSU icers games and interviewing him on my station WRSC. He spoke about developing talent to the NHL level - something that he and the organization was always blamed for. Glen convinced me that he did hve a lot of hockey knowledge and could be a very compitent coach elsewhere.    \"After reviewing all of the circumstances, we decided it was time to make a change,\"  said Penguins GM and brother Craig Patrick.  Glenn Patrick compiled a record of 115-152-38-15 in four seasons with Wilkes-Barre/Scranton. His son Curtis played for PSU and Team USA in the World University Games in Italy earlier this year.   '),(4719,671748,'30,June,2003','2003-06-30 00:00:00','    My friend Jim would be interested in this story...  NHL.com says that the NY Rangers and the Edmonton Oilers have come to an agreement to send the rights of aging veteran defenceman Brian Leetch up north for backup goalie Jussi Markkannen.  On the surface, it looks like a good deal, but if you think about it, it doesn\'t make sense for NY. The Rangers let go one of their leaders for another goaltender when they already have a platoon of goaltenders to start with. Blackburn, Richter, and Dunham have all been established as the #1 guy. Plus, this doesn\'t mean that Leetch will even put on a Oilers jersey, he has to sign a free agent contract first.   '),(4720,671748,'27,June,2003','2003-06-27 00:00:00','    Besides covering my favorite teams here, I think I should mention a few news-worthy items for the rest of the league happening since Blogger was converting my account to the upgraded version.  1. Dominick Hasek wanted to go out a champion and retire to spend more time with his family in the Czech Republic. After a year, Hasek has considered a return to the Red Wings. I guess he sat at home and watched the team leave the playoffs after 4 game sweep made him wonder.  2. Bondra and Halpern have decided to go down with the ship re-signing with the one team that seems to always be going nowhere. The Washington Capitals exercised their contract options on their players.  3. Toronto GM/Coach is now just a coach. Pat Burns demoted himself as a part of the upper management to the club. President and one heck of a former NHL goalie Ken Dryden also stepped down as team president. No word was said to when the positions would be replaced. (Gentlemen, get your resumes ready!)  4. \"Mr. Candice Cameron\" was picked up by his former team he was traded from at last years deadline. Valeri Bure was released by the St. Louis Blues and picked up by the Florida Panthers solidifying the idea that the NHL trading deadline is the \"rent a player for the playoffs\" sweepstakes.   '),(4721,671748,'21,June,2003','2003-06-21 00:00:00','    In a shocking move, my hometown Pittsburgh Penguins moved up to get the #1 pick in the draft and used it on highly sought after goalie Marc-Andre Fleury. They traded up with the Florida Panthers who selected in the number three spot. Here is the top 5 selections in the first round.   1. Marc-Andre Fleury (G) - Pittsburgh 2. Eric Staal (C) - Carolina 3. Nathan Horton (C) - Florida 4. Nikolai Zherdev (W) - Columbus 5. Thomas Vanek (LW) - Buffalo   In Nashville, TN the host sity of the draft, the Nashville Predators selected defenceman Ryan Suter with the 7th pick overall.  With a very classy move, the NHL paid respect to the memory of Roger Neilson who died this morning. He was currently serving as assistant coach of the Ottawa Senators. He also worked for Florida, Philly, NY Rangers, and other teams. When represenatives of teams Neilson worked for, they expressed their condolances. They also paused for a moment of silence. Roger Neilson died of cancer at his home in Peterborough, Ontario at the age of 69.   '),(4722,671748,'15,June,2003','2003-06-15 00:00:00','    Today, I laced up my CCM\'s and hit the rink as Penn State\'s favorite rink rat and skated the public session at the Penn State Ice Pavilion from 1 to 3 PM. It felt great to get back on the ice after a long month\'s layoff! I had a wide open rink to myself through most of the first hour, then at 1:55 the PA system played \" Bring Me To Life \" by Evanescence and \" Fighter \" by Christina Aguilera. That helped propell me even faster around the rink as I pushed myself to skate harder and faster. The great news is that I can skate backwards better!      Then I went to the local mall and found a Franklin SX Pro Extreme In-line Hockey Goalie\'s Mask reduced from a lot of money to very very affordable money! It was probably last years model. It\'s a black mask with a skull of a fanged creature of some sort. It looks pretty fearsome. I have an idea that if I am not satisfied with the look, I just may take the helmet to a air brush place and get something really interesting on it! Like a Trish Stratus tribute!   '),(4723,671748,'13,June,2003','2003-06-13 00:00:00','    Well, the season is over... time to hand out the awards for the year!       HART MEMORIAL TROPHY (MVP) Peter Forsberg, Colorado JAMES NORRIS MEMORIAL TROPHY (Best Defenceman) Nicklas Lindstrom, Detrot FRANK J. SELKE TROPHY (Best Defensive Forward) Jere Lehtinen, Dallas VEZINA TROPHY (Best Goaltender) Martin Brodeur, New Jersey CALDER MEMORIAL TROPHY (Rookie of the Year) Barret Jackman, St. Louis LADY BYNG MEMORIAL TROPHY (Sportsmanship) Alexander Mogilny, Toronto JACK ADAMS AWARD (Coach of the Year)  Jacques Lemaire, Minnesota [TOTAL RIP OFF! Mike Babcock of Anaheim deserved it more!]   '),(4724,671748,'12,June,2003','2003-06-12 00:00:00','    IF you thought your job was hard, consider....  Larry Robinson, NHL head coach, is reported to have turned down the head coaching job in Calgary because after consulting the wife and kids. He is expected to meet with Glen Sather and the Los Angeles Kings next week. I wonder if his family will approve this move. The former coach of the NJ Devils is a higly sought after no nonsense coach. According to this quote from the New York Daily News; he knows how to get things done.    \"I coach a certain way and I\'m not going to change my philosophy because somebody else feels the game should be played a certain way. It would be my rear end on the line. And if it is, I\'m going to do it my way.\"   Kick Ass!   '),(4725,671748,'12,June,2003','2003-06-12 00:00:00','    Hey!!! That\'s Coach Edzo to you!!!\"   or   \"Oh, by the way, I\'ve had to announce the resignation of my membership to the Pittsburgh Penguins Message board under the name of foreczech\"   or this could only happen in my wildest dreams.. but who knows.. maybe someone will pick up on the hint!    \"I have called this press conference to name my successor to the color analyst as little known broadcaster from WRSC AM - in State College - Mike the Biggest Pens Fan\"   '),(4726,671748,'12,June,2003','2003-06-12 00:00:00','    It\'s Hockey Hall of Fame selection time again. You know that you\'re getting old when players you remember watching as a kid are retired and getting into the Hall.  The first black hockey player to get into the Hall is Grant Fuhr. He rose to fame by goaltending for the really really good Edmonton Oilers during their dynasty run. He then moved on to other teams, most notably with St. Louis. He won 5 Stanley Cups with Edmonton and won several championships representing his native Canada in international tournament play. No mention that he was once in trouble for substance abuse.  Pat LaFontaine is one of the few members of the NHL that have played for all three NY teams. During his 15 year carreer he jumped from the USA Olympic team to the NY Islanders. He then became captain for the Buffalo Sabres, and only played one season with the NY Rangers before sustaining a career ending head injury in 1998. Today, Pat LaFontaine raises money for cancer and promotes the Pat LaFontaine chalenge cup to the winner of the Islanders/Rangers rivalry game.   '),(4727,671748,'11,June,2003','2003-06-11 00:00:00','    Well... the search is over... and the Pittsburgh Penguins have decided to go in an unexpected direction.  The Penguins have chosen former color analyst and former player Eddie Olczyk to be the next head coach for the franchise. New fans would recognize his voice as the NHL Radio announcer for this years Stanley Cup playoffs. He was also partnered with Hall of Famer Mike Lang for 3 seasons for the local Penguins broadcasts.  I guess this was as much a monetary decision as much as it has been a directional descision. Names like Herb Brooks, Larry Robinson, and the like would have commanded a high salary to coach the money strapped team. They probably got \"Edzo\" for cheap.  The  mother of all rumors ... with a big emphasis on the rumor part is that I will be named to replace \"Edzo\" soon.   '),(4728,671748,'10,June,2003','2003-06-10 00:00:00','    Congratulations to the  New Jersey Devils  as they won the 2002-03 Stanley Cup Championship last night in front of their home crowd. Pundits and critics talked about it being a home rink advantage in the series as both teams won their games at home. It was fortunate for the Devils to have Game 7 land in their favor.  A teary eyed J.S. \"Jiggy\" Giguere did in fact win the Playoff MVP as one of the rare times that the losing team\'s player had won the award.   '),(4729,671748,'08,June,2003','2003-06-08 00:00:00','    hmmm.. Somehow I overlooked the Civil War brewing among this Stanley Cup series. If you are not sure still of who to root for in the Finals, just ask Mr. and Mrs. Niedermayer.  Traded to the Mighty Ducks at the trading deadline this past season, Rob Niedermayer has enjoyed his first trip to the Finals. His older brother Scott has been a mainstay for the New Jersey Devils defence. He\'s already got two rings.  I guess they must have some additional motivation for the series. I have a brother and if these guys were as half as competitive as my brother and I, they want to win the Stanley Cup badly.  Rob and Scott aren\'t the only brothers to play in the NHL. Eric and Brett Lindros, Wayne and Keith Primeau, and Peter and Antonin Stasny are some of the better known Family legacy names.   '),(4730,671748,'03,June,2003','2003-06-03 00:00:00','    On the 1st, \"Jiggy\" maintained a 166.4 minutes of shutout hockey. Mighty Ducks goaltender has set a new record for longest shutout streak in OT in the playoffs!  Today, Mighty Ducks of Anaheim goaltender J.S. Giguere earned another overtime win in the playoffs over the Devils to extend his OT shutout streak record. Congratulations. I think he\'s won the Conn Smythe Trophy with his efforts this year. [Note, the trophy is given every year to the MVP of the playoffs]   '),(4731,671748,'30,July,2003','2003-07-30 00:00:00','    Mario decides to come back for at least one more year!  According to a Yahoo.com Sports article, (and you could also check out the story on NHL.com) that Mario Lemieux will again lace up his skates to play for the Pittsburgh Penguins for the 2003-2004 season.  It\'s been received around the league with mixed emotions. Some people are still jealous of the greatest hockey player and think he\'s just doing it for money and/or self interest. Other fans like me think this is great for hockey to have him around one more year. The Penguins really need him to survive.   '),(4732,671748,'29,July,2003','2003-07-29 00:00:00','    The Philadelphia Flyers have announced the re-naming of their arena from the First Union Center to Wachovia Center. I think this is a good choice to go from a vulgar nickname to a gothic eastern Europe castle sounding name.  The Wachovia Corporation is a financial institution that not only will hold your money, but they\'ll tell you how to manage it for you.   '),(4733,671748,'27,July,2003','2003-07-27 00:00:00','    Bowing to fan and player complaints, the NHL last week amended the rule for goalie leg pads to meet. Forever known as the \"Roberto Luongo Rule\", goalies leg pads must be 12 inches in length.. no more.  Anyone violating the standard equipment rule will be fined $25,000.   '),(4734,671748,'16,July,2003','2003-07-16 00:00:00','    It\'s a little bit closer to the new 2003-2004 hockey season!  Today, the NHL released their regular season schedule. A few surprises are in store to boost fan interest and TV coverage.   1. Edmonton and Montreal will play an NHL game outdoors at Commonwealth Stadium making NHL history Nov 22 2. Toronto will play pre-season games in Sweden on Sept 18-19 and Finland on Sept. 16. 3. New Jersey will raise their newest Stanley Cup Banner on Oct. 16th vs. Buffalo 4. My Penguins play their rival Flyers for the first time on October 11th and open the year against the LA Kings the day before.   '),(4735,671748,'15,July,2003','2003-07-15 00:00:00','    Speaking of the \"baby Pens\", today they named a successor to my buddy Glen Patrick.  Michel Therrien was announced as the new head coach of the AHL minor league affiliate to the Pittsburgh Penguins. He was the head coach of the Montreal Canadiens in the NHL. But, his employers had little patience for re-building the franchise. When your a hockey fan in Montreal and French Canadian losing is not an option. So, what Canada\'s loss is Pennsylvania\'s gain.   '),(4736,671748,'14,July,2003','2003-07-14 00:00:00','    So, riddle me this... what was the  urlLink  Wilkes-Barre/Scranton Penguins  doing in the Western Conference anyway?  In the summer meetings for the American Hockey League (AHL), the league commissioners voted in favor of realignment. The new divisions and conferences were marketed around local rivalries.  The \"baby Pens\" will compete against their rivals in the new Eastern Conference - East Division with: Bing-hampton (Sens), Albany (Devils), Philadelphia (Flyers), Hershey (Avalanche), Bridgeport (Islanders), Norfolk (Blackhawks).   '),(4737,671748,'09,July,2003','2003-07-09 00:00:00','    Well, since there isn\'t a whole lot of newsworthy events going, I\'ll tell you about my hockey dream I had the other night.  There wasn\'t much too it. I remember suiting up for the Penn State Icers. I was in the locker room and preparing for a game. I remember hopping up and down on my skates trying to get the adrenalin pumping. The anthem was playing, I was about to take the ice,  but.... I was approached by a real player who told me that I shouldn\'t be here, that I wasn\'t really a part of the team and that my services wasn\'t going to be needed.  Oh well, I wish it was a more pleasant dream. I was woken up by the alarm that next morning.   '),(4738,671748,'04,July,2003','2003-07-04 00:00:00','    Happy 4th of July!!!   '),(4739,671748,'03,July,2003','2003-07-03 00:00:00','    The free agent signings started off with a bang by the announcement that the deep pocketed Colorado Avalanche have signed both friends Paul Kariya and Teemu Selanne to million dollar one year contracts. The two wanted to play on the same team again since their first stint with the Mighty Ducks. The most shocking news is that they both chose to sign only one year deals, making them free agents again and most certainly affected by the new Collective Bargaining Agreement (CBA) that will be voted on and go into effect in 2004. Subsequent to the retirement of Av\'s goalie Patrick Roy, Colorado\'s plan is \"the best defense is a good offense!\"  The Red Wings signed away one of their rivals best players in Dallas Star\'s Derian Hatcher. Hatcher was the Stars captain and big hitting defenceman, but Detroit gave him over a suspected million reasons to go home to Detroit. Dallas then announced that franchise Stars player Mike Modano will wear the \"C\" on his jersey next season.   '),(4740,671748,'02,July,2003','2003-07-02 00:00:00','    Veteran and Controversial NHL referee Paul Stewart has announced his retirement. I can\'t say enough about him and the effect that he had on the league.   '),(4741,671748,'01,July,2003','2003-07-01 00:00:00','    Let the free agent signing begin! Today, the NHL begins it\'s Free Agent season allowing players and teams to sign contracts. Here\'s the link to the NHL.com\'s list of  urlLink Free Agents .  In other news... The New York Rangers have decided to just make Glen Sather the coach too. Sather, the GM of the Rangers, took over when the Trottier experiment last season did not prove for the better. Pat Quinn tried this in Toronto. Look how they turn out!  Jeff \"cant\" Hackett signed a free agent deal with the Flyers. He couldn\'t do it in San Jose. He couldn\'t do it Montreal. He couldn\'t do it Boston. I swear this guy is on his 18th team in 2 years.   '),(4742,671748,'01,July,2003','2003-07-01 00:00:00','    Happy Canada Day!!   The Second Annual Canada Day Cup 2003. [Results]  Last years winners were the Edmonton Oilers who got past three rounds of very tough tournament action. They defeated a tough Nordiques squad to set the standard the first year. This is the rundown All original 8 canadian franchises (Quebec, Montreal, Ottawa, Toronto, Winnepeg, Calgary, Edmonton, and Vancouver) were simulated out in tournament fashion using NHL 2002 for EA Sports. This year, I decided to change it up a bit. Instead of a one-and-done format, I brought the up to date July 1 rosters and simulated an international tournament to first decide the tournament seatings. Then, I watched the final \"gold medal\" game between the Maple Leafs and the Avalanche (sporting their original Nordiques uniforms). The game was a classic. In the first period, Toronto took a couple of stupid penalties, but still squeezed out a 2-1 lead. Despite franchise players Sakic and Forsberg, Toronto scored 4 and added an empty net 5th goal. Aebisher stood on his head at times for the retired Roy, but his counterpart Ed Belfour was solid.  Toronto wins the second Canada Day Cup!  \" Blame Canada, Blame Canada... they\'re not a real country anyway!! \" JUST KIDDING!  Thanks, Canada for a great game, great beer, and a great woman [ urlLink Trish Stratus ]!   '),(4743,671748,'31,August,2003','2003-08-31 00:00:00','    Well, today\'s the last day in August... so it\'s the last chance to post something. Maybe I should update it more often, even when things aren\'t going as fast in the off-season.  My Penguins gave away one of their goalies in the \"plethora\" of goalies that they had. Johan \"Moose\" Hedberg was traded to the Vancouver Canucks for a second round draft pick. When, Caron got the bulk of the playing time, it looked as if he was \"the chosen one\" for goalie of the future.  That\'s all for now.. have a safe Labor Day weekend!   '),(4744,671748,'19,August,2003','2003-08-19 00:00:00','    In a follow up story to the tragic death of USA hockey\'s legends, Herb Brooks, several sources including Yahoo! Sports have said this...   Sen. Cal Larson (R-Fergus Falls) recommended Monday that the National Hockey Center in St. Cloud, MN should be called the Herb Brooks Arena.   The Gov. Tim Pawlenty said he\'d think about it and talk to the people who own the Arena rights. It is home to the St. Cloud University men\'s and women\'s team by seating 6,000 and contains two Olympic size rinks.   '),(4745,671748,'12,August,2003','2003-08-12 00:00:00','    I just caught the last few minutes of Colin Quinn\'s comments on his Comedy Central show. In closing, he said this:   \"How do you get more black men to watch hockey? Tell them that white guys won\'t allow them to,\" said Quinn.   '),(4746,671748,'11,August,2003','2003-08-11 00:00:00','    The Associated Press and NHL.com reports that legendary coach Herb Brooks was killed in a car accident in his home state of Minnesota today.  The entire hockey world is both shocked and saddened to learn the news. Mr. Brooks was the head coach for several NHL teams including the North Stars, Rangers, Devils, Penguins (funny how NHL.com article omitted that fact from the list) and was also the head coach for the 1980 U.S. Olympic team when they won the Gold Medal. He was currently the scouting coach for the Pittsburgh Penguins.  No doubt, the NHL will honor this man with memorial patches. Expect the Penguins to have one night of tribute for opening night.  Herb Brooks was 66 years young. He will be missed.   '),(4747,671748,'05,August,2003','2003-08-05 00:00:00','    On an internet surfing trip to Tim Harvey\'s Hockey Blog, I was disappointed to read that the most \"original\" hockey team name the Macon Whoopee have decided to cut out and rubber stamp themselves \"defunct\".   '),(4748,671748,'30,September,2003','2003-09-30 00:00:00','    Can you imagine Bobby Orr, Ray Bourque, and Cam Neely on the same Boston Bruins team?  Well, yesterday, my favorite comedian and fellow hockey fan, Denis Leary, sponsored a celebrity hockey game benefiting the Boston Firefighters. They successfully raised $500,000 ($673,000 Canadian) for the worthy cause. You can read the results of that game at this link on  urlLink NHL.com   '),(4749,671748,'29,September,2003','2003-09-29 00:00:00','    Since today\'s date is really important to me, I would like to dedicate this entry to my patron saint. On that note, I would also like to make others aware of an Ontario Hockey League team by promoting them on this website. The  urlLink St. Michael\'s Majors  play up in Toronto, Canada. They have had a lot of distinguished alumni play for them. Rick Tocchet and Jason Spezza. But I use that term loosely as Eric Lindros also played his Junior career for the Majors.  The funniest thing about the team is that their mascot is named \"Mikey.\" Their official web site is seen above.   '),(4750,671748,'26,September,2003','2003-09-26 00:00:00','    Knowing that the NHL 2004 from EA Sports came out yesterday, I ran to my nearest \"Mart\" and got a copy. I\'ve been buying the EA Sports series since I\'ve had my PS one.  I love the way that they have really updated the NHL series game from last year. Some of the new features of the game are similar to the Madden game where the game player can now become the team\'s GM in franchise mode. As GM, you can manage the salaries of the players, set ticket prices, and make trades. If your team is successful during the years, you can upgrade your simulated GM office. Other new features for the game that I really like is the on-line play where you can hook the Playstation up to the Internet and challenge others from around the globe. Secondly, since it\'s hooked up to the Internet you can also download player updates and new rosters. Speaking of rosters, EA Sports also cleaned up the mess of out of date rosters from the 2003 version. The team rosters are pretty up to date. Any free agents that were unassigned to a team can now be found in the Free Agents column. On 2002 version, a few complaints were raised that there was a small limit as to how many players could fit onto the rosters. Teams now begin with 4 or 5 goalies when 3 were the limit on past versions. Also, I was genuinely surprised that the Phoenix Coyotes were up to date with their new logo and uniforms. (See Sept. 4th blog entry) Last year, the Penguins had changed their jerseys, but it was too late to fix it in 2003\'s version. (It also made my decision not to buy last years for that reason.) They have also improved their fighting sequences by a new fight engine that provides a no-holds barred realism to the game. Instead of a button-mashing attack, instead you must strategize bunching the body and head. The helmets come off as you grapple for dominance. You may also chose when or when not to drop the gloves. Additionally, you are also rewarded by playing other EA Sports titles and in the beginning, you can chose your favorite team to appear as the wallpaper background to the menu screen. Getting to the few things that I don\'t like about the game is the incredible waste of space that putting the European elite leagues on the game. You can also chose to play as one of Sweden, Finland or Germany\'s elite league teams. I only know a couple of those teams thanks to the NHL expanding their interests to playing exhibition games there. Modo and Jokerit. In my humble opinion, they should have listened to their fans that enjoy the unique All-Star feature of the skills competition. EA\'s rival ESPN has included the skills competition in their hockey game this year. Look for the skills competition in NHL 2005. I am also having an adjustment period learning how to maneuver the players around the rink. EA Sports found it necessary to incorporate the analog joysticks on the PS 2 to get more realism to the game. EA Sports... If it\'s in the game...It\'s definitely in this game! (The makers of this game have not paid me to endorse their product... but if they wanted to...*wink*)   '),(4751,671748,'25,September,2003','2003-09-25 00:00:00','    Stop reading this blog.... just for today. Come back tomorrow for up to date hockey wall to wall coverage!  Go out to your local Mart or Electronics store and pick up  EA Sports NHL 2004!  It comes out today! They are the top of the line when it comes to Sports video games. If you are like me, try it on the PS2. I\'ll let you know what I think of it in my review in an upcoming entry.  (they didn\'t pay me to endorse their product... but if they wanted to... *wink*)  Also, I\'d just like to wish my sister Lisa a happy 22nd birthday.  (she didn\'t pay me to say that... but if she wanted to... *wink*)   '),(4752,671748,'21,September,2003','2003-09-21 00:00:00','    Canada is #1!!    After surfing around the latest link that I added to the Essential Hockey Websites listed at the left,  urlLink NHL Network  (actually it was TSN) had an article reporting that for the next World Championships and carrying over to the new format in the Winter Olympics. The top 8 teams automatically qualify for the Turin Olympic games.  Based on recent international competition wins and overall reputation, Canada was listed #1 like a national champions poll in NCAA Football. In no particular order... Slovakia, USA, Finland, Germany, Czech Republic, Sweden and Russia round out the top 8. The last place team was Turkey listed as 43rd. Turkey has a hockey team?   '),(4753,671748,'17,September,2003','2003-09-17 00:00:00','    Well.. the Toronto Maple Leafs are the ambassadors for NHL brand of hockey this month. They are currently playing exhibition games in Sweden. Yesterday, they defeated the Jokerit team 5-3. Newly acquired defenceman Richard Jackman (no, not the guy that played Wolverine... that\'s Hugh) scored twice for the Leafs. The lowlight for the Leafs was that team star Owen Nolan was the recipient of a check in the previously injured back. Later, he assured the fans and team on numerous websites including the  urlLink Leafs official page  that his back is fine. One thing for sure, team captain Mats Sundin played host as he hails from Sweden.   '),(4754,671748,'08,September,2003','2003-09-08 00:00:00','    When the announcement came that Richter called it a career, it was one down one to go.  Today, 20 year veteran Doug Gilmour announced his retirement from the NHL as a Maple Leaf. He played nobly for the Canadiens, Sabres, \"Leaves\", Flames, Blackhawks, and Devils. He won the Stanley Cup with the Flames in 1989. He also stands high on the all time scoring list putting in 450 goals and totaling 1,414 points.  I was amazed when Gilmour signed a contract to play one more year in Montreal last year. (check out my archives on what I said then). Gilmour was overdue for retirement, but gave his best shot for one last run.   '),(4755,671748,'07,September,2003','2003-09-07 00:00:00','    Don\'t be caught with the wrong jersey for the 2003-2004 season...   ...warns the scrolling internet banner advertisement.  The NHL will break with tradition and switch the jersey sweaters worn for all NHL teams during the season. The dark jerseys will now become the home jerseys and the away jerseys will be the white jerseys.  I think this is a terrible idea! This isn\'t the NFL! If it \'aint broke, don\'t fix it!   '),(4756,671748,'04,September,2003','2003-09-04 00:00:00','    The Phoenix Coyotes, owned by NHL Hall of Famer Wayne Gretzky, has updated their look to the franchise by creating the new jerseys. You can check them out at this link  urlLink  Official Coyotes website .  I think they look terrible. They tried to incorporate the red and white of the Detroit Red Wings and ended up in the same vein as fashion victims the Carolina Hurricanes. The Coyote logo is a slight improvement.    The original press conference from the organization said that they were \"going to incorporate the rich heritage of the franchise by honoring it\'s past.\" Everyone thought that this would be a nod to the Winnipeg Jets that the franchise started as. The Jets were blue. I don\'t see any blue.   '),(4757,671748,'03,September,2003','2003-09-03 00:00:00','    Too little too late for some people, but it came down to the personal honor as veteran goaltender Mike Richter has called a press conference. He is expected to announce his retirement from the NHL.  Richter spent the bulk of his time on the IR the last 2 years of his career. However, he was earning his paycheck by manning the pipes as the New York Rangers won their first Stanley Cup in more than 40 years in 1994. He also distinguished himself as a USA born goaltender selected to represent his country in the Winter Olympics in Nagano, Japan.  The Rangers are trying to turn things around by investing in goalie prospects Jussi Markkanen and Dan Blackburn. They also picked up veteran goaltender Mike Dunham at last year\'s trading deadline. The writing was on the wall for Richter.   '),(4758,671748,'03,September,2003','2003-09-03 00:00:00','    10 days after re-signing with the \"Leaves\", defenceman Brian McCabe will be out indefinitely after Yahoo! Sports and Sporting News reports McCabe injuring his knee in practice yesterday.   In other news, I\'m starting up my Yahoo Fantasy Hockey League again this year. Stay tuned with the details. My team, \"Ruthless Aggression\", came in last place last year.   '),(4759,671748,'31,October,2003','2003-10-31 00:00:00','    Here are the top five scariest costumes this year!  5. Philadelphia Phantom (AHL) 4. Eric Lindros 3. Bob Goodenow 2. A Collective Bargaining Agreement  1. Garry Bettman   '),(4760,671748,'30,October,2003','2003-10-30 00:00:00','    The new ECHL this year has begun this season by including the west coast hockey teams in the Western Hockey League. This infusion of two leagues is reminiscent of the old WHA teams becoming NHL franchises. It also legitimizes a league that now has the capability to provide minor league development all over the United States. Check out the ECHL hockey league link to the left of the page for the teams. Included in the new super ECHL league is an expansion franchise Gwinnett Gladiators from near Atlanta, GA. I recommend that you visit their  urlLink official site .   '),(4761,671748,'29,October,2003','2003-10-29 00:00:00','    Congratulations to Mario Lemieux who collected his 1,700th career NHL point today in a game against the New York Islanders.   '),(4762,671748,'28,October,2003','2003-10-28 00:00:00','    Two NHL franchises chose to honor their all time great goaltenders The Colorado Avalanche have decided to honor their franchise goaltender by raising #33 to the rafters of Denver\'s Pepsi Center. Patrick Roy called it a career earlier in the summer, but gave a passionate speech about what it was like to play hockey professionally.  Earlier in the season, new Hall of Famer Grant Fuhr had his #31 retired by the Edmonton Oilers. What was unique about his ceremony was that he was allowed to skate on the ice in full goalie gear donning the classic 1980\'s Oilers jersey.   '),(4763,671748,'22,October,2003','2003-10-22 00:00:00','    I am deeply troubled at the entry that I found on my buddy Tim Harvey\'s hockey blog. Here is an excerpt from it...    \"[bleep]  you, Gary Bettman. I can\'t think of four words to better describe how I feel about the National Hockey League as I make my final entry into Rink Blog. On second thought, those four words should be: Thank you, Gary Bettman. Thank you for saving me a thousand dollars on tickets for poor to mediocre hockey.... Thank you for saving me a $150 on NHL Center Ice, which shows poor to mediocre hockey....Thank you for saving me $49.95 on The Hockey News, which covers poor to mediocre hockey.... After weeks of documenting your bullshit spat with the NHL Players\' Association, I don\'t give a [bleep] anymore. Those are the four words I leave you with, Gary Bettman. NOBODY GIVES A  [bleep]!\"    I think this recent tirade is with merit and speaks for a lot of hockey fans. It should focus on the frustrations of a sport that should have risen above the petty disputes that have plagued the NHL for quite some time. For a fan that should sit and watch the product and find it watered down so much that it gets this kind of reaction, then something must be done.  I think Mr. Harvey\'s rant on the conditions of professional hockey go a little bit to far. For him to sit there and throw his hands up in the air and give up on the greatest game in North America. My advice to Tim and others who feel like him is this: Donâ€™t give up just yet! The actions of a few small-minded people controlling must not tarnish this great game. An analogy to this is where I disassociated the people from the institution. People make mistakes. We are lucky to have some institutions like true American freedom that stand alone as something closer to perfection. Bettman, Goodenow, and the like should be kicked out for what they are doing, but DON\'T EVER STOP LOVING HOCKEY! Go see a peewee game, go to a minor league game, or if you can, just go out and skate around the rink. The game will survive this. Believe.   '),(4764,671748,'21,October,2003','2003-10-21 00:00:00','    Okay, how could I ever miss this story?  urlLink Slam! Sports  and  urlLink TSN  report that the legendary Maple Leafs Gardens in Toronto, ON, Canada is to be turned into a giant  urlLink Loblaws  grocery store.   There  goes my idea of turning it into Canada\'s largest  urlLink Tim Horton\'s!   '),(4765,671748,'17,October,2003','2003-10-17 00:00:00','    Here\'s a bit of trivia for you...  The last time that two NHL games had finished a tie 0-0 was 69 years ago. Recently, NY Rangers and Atlanta, Philadelphia and San Jose games ended in a tie.  The Elias Sports Bureau reports that Boston at the New York Rangers and the New York Americans at Detroit in 1934 was the last time this occured.   '),(4766,671748,'16,October,2003','2003-10-16 00:00:00','    According to a FOX Sports Net TV report... The \"It takes some serious grapefruits to pull that stunt\", in a pre-game ceremony against the New Jersey Devils at home against the Toronto Maple Leafs, the Devils reportedly offered former teammate Joe Nieuwendyk a place on the Ceremony to collect the 2002-03 Stanley Cup commemorative ring.  Nieuwendyk responded by saying \"Uh-uh... new team; different attitude!\" as a new member of the opposing Maple Leafs.  The Devils and Maple Leafs skated to a 2-2 tie as Toronto gave up a goal in the last seconds of the game.   '),(4767,671748,'16,October,2003','2003-10-16 00:00:00','    This just goes to show you that not always the squeaky wheel get the grease.  Last season, Latvian-born goaltender Arturs Irbe complained about his diminished role on the then defending Eastern Division champ Carolina Hurricanes. The team responded by demoting him to the AHL to give him playing time with the Lowell Lock Monsters.  Since the team picked up free agent Jamie Storr, Irbe became a man without a country. NHL Network and Fox Sports Net report that Irbe was again sent to the ECHL for the \"legendary\" Johnstown Chiefs.   '),(4768,671748,'11,October,2003','2003-10-11 00:00:00','    Just as the new hockey season is upon us, many teams start out fresh with a new outlook to the season. This season is going to be different for me and for my family as it will be the first hockey season without my maternal grandfather, Stephen M.  As I am consoling my family with the loss, my thoughts and memories of the good hockey fan my grandfather was over the years. We talked about the Penguins a lot. I remember him in his final years positioned on his favorite recliner rooting for the hometown franchise.  The classic hockey rivalry starts again in the NHL as the Pittsburgh Penguins play the Philadelphia Flyers. I could think of no other way to salute my grandfather by sitting back, cracking open an \"Iiirrn\" (Iron City Beer to you non-Pittsburghers), and watching the hockey game. [Flyers and Pens tied 2-2]     Stephen M. 1918-2003   \"Hang in there, pup-pup. We still need you!\"   '),(4769,671748,'09,October,2003','2003-10-09 00:00:00','    Here\'s a bit of news from  urlLink  Glen Colbourn of the Toronto Star     Coming tonight to a TV screen near you: Shania Twain telling hockey fans how she gets it. Get It? is the tagline of the NHL\'s new promotional ads, which feature Twain, Jim Belushi, Denis Leary, Kiefer Sutherland and Cuba Gooding Jr.  The 30-second spots try to show new fans how to watch hockey games and explain terms like \"roughing\" and \"high-sticking.\" And they are intentionally cheeky. Twain, for example, describes \"changing on the fly\" while wearing a tight, midriff-baring, orange T-shirt. \"Shifts are so exhausting they last less than a minute,\" she says. \"Which leads to something you only want to hear in hockey ...\" A clip of Marc Crawford then pops up with the Vancouver Canucks coach barking to his players: \"Short and hard here, short and hard.\"    What a great idea. Shania is really beautiful!   '),(4770,671748,'08,October,2003','2003-10-08 00:00:00','    Tonight the NHL season gets underway with three games on the bill. Tonight, the Mighty Ducks of Anaheim with Sergei Fedorov will face off with the Dallas Stars on ESPN 2.   '),(4771,671748,'06,October,2003','2003-10-06 00:00:00','    Things can\'t get any worse for the Atlanta Thrashers franchise. In late Sunday night and on to early Monday morning, the word was out that Dan Snyder lost his life in the hospital after a tragic car accident driven by teammate Dany Heatly.  According to several NHL sources, the Atlanta P.D. announced that they will indeed follow through with the investigation of the accident and charge Dan Heatly with first degree vehicular manslaughter.  I will be following this story as it continues to unravel.   '),(4772,671748,'05,October,2003','2003-10-05 00:00:00','    Mario Lemieux, heralded as the (3x) savior of the Pittsburgh Penguins franchise and the greatest player ever in the history of the NHL, has a birthday today.   '),(4773,671748,'04,October,2003','2003-10-04 00:00:00','    NHL Radio and upstart satellite Radio Network Sirius have reached a deal to broadcast their games on the network so that from where ever you are, you can always hear the NHL Radio Network. Of course, you can hear every game from NHL.com, but if you\'re not around the internet, I\'d suggest this is the way to get your NHL ice hockey fix.   urlLink From NHL.com   '),(4774,671748,'01,October,2003','2003-10-01 00:00:00','    First Michael Vick goes down, now Dany Heatly 2002\'s NHL Rookie of the Year for the Atlanta Thrashers and NHL 2004\'s poster child, Heatly and another teammate was involved in a serious car crash. According to the  urlLink  Atlanta Thrashers  Official site, Heatly sustained injuries such as a broken mandible and multiple bruises. His teammate wasn\'t so lucky. Dan Snyder was listed in critical condition with a fractured skull and underwent emergency surgery to relieve the pressure to the brain. Investigations are underway as to the cause of the tragedy.   '),(4775,671748,'30,November,2003','2003-11-30 00:00:00','    The last remaining fan-beloved Pittsburgh Penguin was traded today to the LA Kings. Marin Straka and his $4.2 million salary had to go in return for 2 minor leagers.  Straka just wasn\'t producing much in a Penguins uniform and missing Mario Lemieux\'s help. He will try and turn it around personally on a struggling Kings team that has a lot of players on the Injured Reserve. If he could play with quick players Deadmarsh, Palffy, and/or Robitaille then the Kings could contend for a playoff spot in the West this year.   '),(4776,671748,'27,November,2003','2003-11-27 00:00:00','    I\'m thankful for a lot of things... mainly I\'m thankful for my kind and understanding family, friends, and co-workers... then I\'m thankful to blogger for allowing me to make a news journal about my favorite sport, then I\'m also thankful that I can still enjoy the sport of Hockey in Pittsburgh. I\'m thankful for the Penguins and Mario Lemieux\'s efforts to keep them in Pittsburgh.   '),(4777,671748,'24,November,2003','2003-11-24 00:00:00','    When you think of a hockey rivalry, some people think of the Esposito brothers Tony and Phill. 23 years later, Mathieu and Martin Biron continue the legacy of sibling rivals in the NHL. In a game with the Florida Panthers v. Buffalo Sabres, brother Mathieu scored a goal to help his team win.    \"I was happy,\"  Mathieu Biron said.  \"I knew it was going to come sooner or later. I\'m glad he didn\'t put me on the ice right after the goal. I didn\'t know what was going on. It\'s hard to explain.\"   '),(4778,671748,'22,November,2003','2003-11-22 00:00:00','    History was made today as the Montreal Canadiens and Edmonton Oilers by playing in the first outdoor NHL game at Memorial Field.  more news and comments soon!   '),(4779,671748,'21,November,2003','2003-11-21 00:00:00','    In AHL news, the 2005 All Star city and logo was recently announced. Congratulations to Manchester, NH as next years host city.  This season, the AHL will display the best of Canadian v. \"planet American\" players in Grand Rapids, Michigan.   '),(4780,671748,'20,November,2003','2003-11-20 00:00:00','    This  urlLink NHL Network  article reveals that the naming rights to the Oilers home rink will now be changed.  The Skyreach Center for Rexall Place.  I really don\'t like the new trend of naming rights to buildings. Now my NHL video games are all outdated. What if this Rexall company falls into disfavor with people like Enron did?   '),(4781,671748,'15,November,2003','2003-11-15 00:00:00','    Overheard during a hockey game on UPN. Pittsburgh Penguins v. Florida Panthers   \"That\'s why he get so much time on the P.P.!\" - Pens color analyst Troy Loney   Okay, read that quote aloud and tell me what\'s wrong with that!   '),(4782,671748,'12,November,2003','2003-11-12 00:00:00','    The Buffalo Sabres will play the defending champion New Jersey Devils on the first neutral site game in Rochester, NY.  Fittingly, that is the home of the Sabre\'s minor league AHL franchise Rochester Americans.   Update:  The Sabres and Devils tied 2-2.   '),(4783,671748,'11,November,2003','2003-11-11 00:00:00','    Come to papa!       Gee.. do all women look that good in the new  urlLink  Vintage Hockey   line from the NHL?   '),(4784,671748,'09,November,2003','2003-11-09 00:00:00','    The first one to get the ax this year is Mike Kennan the head coach in Florida. Rick Dudley (who\'s prior credits include a good young Senators team) will take over duties as GM/Head Coach.   '),(4785,671748,'08,November,2003','2003-11-08 00:00:00','    I got in the mail a catalog for the new line of merchandise from the NHL. Just in time for the holidays. In it was the jersies, pucks, and other usual assorted stuff. Also, they included the section devoted to this years NHL all star game in St. Paul, Minnesota. They unveiled what the All-Star jersey\'s are going to look like. They are modeled after the \"old style\" knitted sweaters of yesteryear. They look nice, but expensive. Lastly, they want to appeal to the adressee, the hockey fan, as on the back page there is a picture of Karyia and Selane flanking a male model in a Flyers jersey. They personalized it with my name on the Flyers jersey!  Ack! They put my name on a Flyers jersey!   '),(4786,671748,'06,November,2003','2003-11-06 00:00:00','    During the Boston and San Jose game on ESPN 2, they flashed an interesting stat that since 1984 the Bruins have had 18 goaltenders protect the pipes. They threw out names like Andy Moog, Brian Rolston, Jim Carey. Hopefully, they can count on their current goaltender Felix Potvin to stop the revolving door trend.   '),(4787,671748,'04,November,2003','2003-11-04 00:00:00','    Congratulations to Mark Messier for surpassing Gordie Howe as second in the all time NHL point scoring race. His late empty net goal in the Dallas game was his 1,581th career point. He has at least 1,000 more points to go to surpass Wayne Gretsky.  Speaking of Gretsky...  Team Canada announced that he will be the GM of Canada\'s World Cup team in 2004. Canada is not only the defending Olympic chamions of 2002 but also the World Cup champs of 2003. He will be selecting the players for next years team.   '),(4788,671748,'25,December,2003','2003-12-25 00:00:00','    Merry Christmas!!   '),(4789,671748,'20,December,2003','2003-12-20 00:00:00','    The one night that I had marked on my calendar was today. The Atlanta Thrashers were in Pittsburgh tonight and as a special promotion, the gave away Hall of Famer and Penguins Broadcasting voice Mike Lang talking bobbleheads. I was sure to go to see the game at the Mellon Arena. However, I found out later that my plans would change when my sister was already scheduled to have a Graduation ceremony at Penn State the same day. Oh well, there\'s always a next time.   '),(4790,671748,'19,December,2003','2003-12-19 00:00:00','    I am very proud of my Alma Mater and former employer PSU Women\'s Ice hockey team as I was looking at the College Hockey news digest via Yahoo News. I was genuinely surprised to see that the Lady Icers got a headline as they coppied a Daily Collegian atricle detailing the rescheduling of games planned against Massachusets. Keep in mind, these girls, though they work and do the very thing they love, they are just a club team to the University with little or no deep history or NCAA status. As a matter of fact both Penn State Men\'s and Women\'s teams are subjugated to the ACHA conference.   '),(4791,671748,'15,December,2003','2003-12-15 00:00:00','    While I was flipping around the TV dial, trying to learn the new cable channel switch, I came on The Learning Channel (TLC) and airing the show \"While You Were Out\". It is about a team of redecorating experts that collude with people challenged with redesigning a room in the house while another person is away.  During this particular eppisode, a loving wife decided that the downstairs bar/hockey room should be converted to the ultimate Hockey bar. I loved it. They had hockey sticks coming down the bannister, white walls with a Center Red stripe accross the middle, and a large digital TV.  It was even cooler that the family were Canadian and Maple Leafs fans. The TLC team all worked in replica Toronto jerseys.   '),(4792,671748,'10,December,2003','2003-12-10 00:00:00','    My insider sources to the Washington Capitals tell me that Coach Cassidy has been fired and replaced with Assistant Coach Glen Hanlon.  Despite having one of the leagues lowest point totals (currently tied with the Pens), \"sweaty teddy\" has dropped the hammer on one of the leagues youngest coaches.   '),(4793,671748,'07,December,2003','2003-12-07 00:00:00','    It was announced today that \"phenom\" 18-year-old goaltender for the Pittsburgh Penguins, Marc-Andre Fleury, will be loaned to play for Team Canada in the World Junior Championships. Fleury was the goaltender for Canada last year which earned a Silver medal. Team USA is favored to win this years gold.   '),(4794,671748,'05,December,2003','2003-12-05 00:00:00','    From  urlLink NHL Network   A man known only as \"Dale\" (my theory is that it\'s Dale Hunter) was reported to have procured the winning bids to Wayne Gretsky\'s and Mark Messier\'s auctioned off jerseys of the Classic outdoor game played earlier in the month. The Gretsky jersey goes for $26.6 thousand ($34,827 CDN) and Messier $9.6 thousand ($12,569 CDN). when asked why he bought both jerseys, he was reported to have said  \"it seemed only fair to keep them together\".   '),(4795,671748,'03,December,2003','2003-12-03 00:00:00','    Well, today I just felt like I needed the chance to clear my head and find the Kirk Nevin Arena in Greensburg, PA. The afternoon sessions are the ones least ocupied. Today, I was the only one on the ice. I didn\'t even have any skate guards. For 2.75 hours, I had the pleasure of skating alone.  The Pens play in Philthydelphia (gee.. did I lose my journalistic integrity with that one?) tonight. The Penguins have managed 2 ties in 2 games thus far this season.   '),(4796,671748,'01,December,2003','2003-12-01 00:00:00','    The Toronto Maple Leafs have  urlLink on their Official website  the results of the Breast Cancer Awareness game. They raised $26,000 for the CBCF (Canadian Breast Cancer Foundation) Here are some of the excerpts from the article.    On-line auction items included the Leafs pink ribbon helmets with Tie Domi\'s helmet fetching the highest bid at $2,130. Helmets belonging to Gary Roberts, Mats Sundin and Alexander Mogilny\'s garnered $2,010, $1,400 and $1,400 respectively. HNIC (Hockey Night in Canada) personalities Don Cherry and Ron MacLean also participated in the auction by donating Cherry\'s pink jacket, which raised $5,010, and MacLean\'s pink tie that garnered $610.    It\'s amusing to see who\'s personal donation took in the most money. Domi, despite being a popular Toronto franchise player, is also widely known for goonish play. It\'s surprising that he would bring in more than superstars Nolan, Mogilny, or even their captain Sundin. Plus, Cherry v. MacLean\'s earning is amusing thinking back to their stance on the Iraq War recently. The disparity is of course due to a heavier bigger jacket to a tie, but 5 thousand to 6 hundred is far from being close. It shows you where the star power is on the HNIC program.    A \"pink ribbon\" theme was successfully incorporated into the Leafs  [November 8th]  game against the Edmonton Oilers with the distribution of 20,000 pink ribbons to game attendees by CBCF volunteers and Leafs wives and partners, while players from both teams wore exclusive pink ribbon game night helmets.    If you look at the website they show some of the players wives... and gosh some of them are HOT!! There is a lot to be said about Tellqvist\'s Swedish Blond wife! *wink*   '),(4797,671748,'28,January,2004','2004-01-28 00:00:00','    In the latest news of comings and goings in the NHL today,...  Dany Heatly returned to the ice tonight after 4 long months of rehab after suffering injuries sustained in a car accident. The accident also killed teammate Dan Snyder. Since that has happened, it has colored my opinion on Mr. Heatly. His reported careless actions behind the wheel does not signify a strength of character. The Thrashers tied the Blues 1-1.  Speaking of \"strength of character\"....  Eric Lindros leaves tonight\'s game against the Capitals for sustaining a yet unheralded 8th concussion. Skating with his head down along the bench boards, Lindros runs into a wall #54 [Jason Doig]. You\'ll remember that Lindros was traded to the Rangers after suffering a 7th concussion while playing for the Flyers.   '),(4798,671748,'27,January,2004','2004-01-27 00:00:00','    In what seems like a nightmarish two weeks for the Washington Capitals PR department, it\'s now gotten worse as the owner known to some fans as \"Sweaty Teddy\" or \"Trillionaire Ted\" Leonsis has reportedly attacked a fan tonight. Details of the attack are rumored to have happened like this... a fan named \"Chris Hammer\" held up a sign that read \"First AOL and Now the Capitals - See the coincidence?\" Leonsis wrestled the fan to the ground.  Pundits and people are shocked at Leonsis\' actions as he was also known as being very fan frendly.  Expect a heavy fine coming the Capitals way.   '),(4799,671748,'25,January,2004','2004-01-25 00:00:00','    From The United Hockey League  urlLink Official Website   The United Hockey League (UHL) announced Friday that the UHL\'s Board of Governors has approved the application of Glens Falls Professional Hockey, L.L.C. to operate the Adirondack IceHawks  One of those owners just so happens to be NHL analyst for ESPN, Glens Falls resident, and all around good guy Barry Melrose.  That\'s right! Barry Melrose owns a hockey franchise.   '),(4800,671748,'24,January,2004','2004-01-24 00:00:00','    I don\'t vote democrat, but if you\'ve been following the Democratic Presidential primaries, you might have seen the campaing trail pick up game that featured Presidential candidate John \"F\'n\" Kerry.  I was impressed with Kerry\'s liking for the game, at least, as hundreds of his supporters came out to watch him at the JFK Colliseum in Manchester, NH.   '),(4801,671748,'23,January,2004','2004-01-23 00:00:00','    \"The reality is I\'m used to seeing more shots, more quality shots in Albany. This was an easy game for me. What\'d I have, four shots in the third period? It\'s nice playing for these guys.\"   -- New Jersey goalie Scott Clemmensen who breezed to a shutout victory in his first NHL game Tuesday night in Pittsburgh.   Oh, quit gloating already, Scott! You are lucky to even be playing with Brodeur hurt. Plus, the Pens played down to your AHL level with the minor leaguers that they put on the ice every night.   '),(4802,671748,'23,January,2004','2004-01-23 00:00:00','    Jaromir Jagr finally got his wish as he was traded to a major market franchise (just about the only franchise that could afford his 10 million a year salary). The New York Rangers, fearing missing the playoffs again this year hope that they can spark a path to the post season. Jagr could be the first in a long line of players that leave the Capitals as reported by TSN\'s Sportscenter. Anson Carter now finds himself on the Capitals.   The Rangers have also added journyman defenceman Jamie Pushor from the Columbus Blue Jackets. This is Pushor\'s 3rd team this season after spending time with the Maple Leafs and Blue Jackets this year. He also was a stopgap in the defence as a member of the Penguins last season.  In other moves around the leage, Charlie Stephens of the Avalanche goes to Ottawa for veteran forward and career AHLer Dennis Bonvie.   '),(4803,671748,'22,January,2004','2004-01-22 00:00:00','    Here is the list of players that will represent their respective teams at the All-Star Game in February. If you are a fan of one of the NHL\'s smaller market franchises. You have been denied a represenative. Pittsburgh, Carolina, Buffalo, Chicago, Edmonton, and Anaheim have not played well all year. The first draft of the list was announced on ESPN\'s NHL 2night during the 1st and 2nd intermissions of a NY Rangers v. Philadelphia game.    Eastern Conference  Robert Lang, Washington Mark Messier, NY Rangers Keith Primeau, Philadelphia Jeremy Roenick, Philadelphia Mats Sundin, Toronto Daniel Alfredsson, Ottawa Marian Hossa, Ottawa Jaromir Jagr, Washington Gary Roberts, Toronto Adrian Aucoin, NY Islanders Nick Boynton, Boston Wade Redden, Ottawa Sheldon Souray, Montreal Roberto Luongo, Florida Jose Theodore, Montreal   Western Conference  Pavel Datsyuk, Detroit Patrick Marleau, San Jose Joe Sakic, Colorado Shane Doan, Phoenix Jarome Iginla, Calgary Rick Nash, Columbus Markus Naslund, Vancouver Alex Tanguay, Colorado Keith Tkachuk, St. Louis Filip Kuba, Minnesota Mattias Norstrom, Los Angeles Chris Pronger, St. Louis Kimmo Timonen, Nashville Dwayne Roloson, Minnesota Tomas Vokoun, Nashville   '),(4804,671748,'17,January,2004','2004-01-17 00:00:00','    From  urlLink  the AHL official site    With native son Roger Clemens among them, the largest crowd in the AHL this season  13,342 fans  were on hand Friday night to watch the Houston Aeros battle the rival San Antonio Rampage at the Toyota Center in Houston.  It was the largest crowd at an Aeros home game since the team joined the AHL in 2001.     I thought I\'d just squeeze a little AHL news and notes in to the fray to try and uphold a more well diverse reporting on this Hockey Blog. Roger Clemmens did a good thing by getting out and meeting the community. He\'s associating his career himself to the city he calls home!   '),(4805,671748,'16,January,2004','2004-01-16 00:00:00','    urlLink The Moose Returns    About a while back, the Penguins playoff phenom goalie Johan \"The Moose\" Hedberg, was famous for having the fans rally around him after comming from the San Jose franchise minor leagues of the Manitoba Moose. The fans would all shout \"Mooooose\" when Hedberg made a great save. Since those playoffs, things didn\'t pan out in Pittsburgh and Hedberg was sent to the Vancouver Cannucks. According to the above links, this season, the Moose is the AHL affiliate of the Cannucks. Johan was sent to the team that made his nickname famous.   '),(4806,671748,'15,January,2004','2004-01-15 00:00:00','    The NHL is doing a better job with promoting it\'s sport. The NHL is now holding a contest to find the  urlLink  7th Man .  Fans must compose a 250 word essay about why they should be considered. The winner will get the \"honor\" of being the NHL GM for a day and ceremoniously drop the first puck in New York City. I am entering, athough I don\'t think I can make it 250 words. I am thee Hockey Fanatic. I should win - hands down!   '),(4807,671748,'13,January,2004','2004-01-13 00:00:00','    While checking the news around the world of hockey, I happened upon the news link headline that read  Leafs douse Flames  on NHL.com. As a trained writer and journalist, I have become more aware of the cliches used. It\'s dissapointing for Flames fans having to put up with seeing titles like that. Come on guys, be more creative!  1. Flames flicker out 2. Flames burn bright 3. Smoldering the Flames   '),(4808,671748,'12,January,2004','2004-01-12 00:00:00','    Boston Bruins forward Cam Neely had his number retired tonight. Cam\'s decade long career was cut short after injury because of the physical style that he used on the ice. He led the team with 55 career goals.  Perhaps his best known achievement is the establishment of the Neely House Foundation. It\'s mission is to provide housing for families that suffer with cancer disease.      urlLink  Bob Snow  summed it up best in his NHL.com article.   Last night, Bruin Nation recognized a unique player. From here on, anyone can look up at the No. 8 banner, smile, and realize how great Cam Neely truly was in the history of the Boston Bruins.   '),(4809,671748,'11,January,2004','2004-01-11 00:00:00','    What an amazing turn around it has been for Phoenix Coyotes goaltender, Brian Boucher. Tonight, he added his name to the legacy of dominant Western Conference goaltenders as his streak of shut-out hockey ended against the Thrashers.  Boucher has already made history by setting a NHL record of 5 concecutive shut out games. It totals his time between the pipes 322 minutes and 1 second without a goal being scored.  The Coyotes fans hope that this could resurect the franchise who were struggling to win at home all season long. Boucher, a castoff from Philadelphia, was playing himself into oblivion while watching goaltender Sean Burke get most of the workload this season.   '),(4810,671748,'08,January,2004','2004-01-08 00:00:00','    Well, we\'re at the half way point in the 2003-2004 NHL hockey season and the starting players for the All Star game in Minnesota in early February.   Eastern Conference  Coach Pat Quinn (Leafs) Center Joe Thorton (Bruins) Wings Ilya Kovalchuk (Thrashers) Martin St. Louis (Lightning)  Defensemen   Scott Stevens (Devils) Scott Niedermayer (Devils) Goaltender Martin Brodeur (Devils)   Western Conference  Coach Dave Lewis (Red Wings) Center Mike Modano (Stars) Wings Bill Guerin (Stars) Todd Bertuzzi (Cannucks) Defensemen  Nicklas Lidstrom (Red Wings) Rob Blake (Avalanche) Goaltender  Marty Turco (Stars)   '),(4811,671748,'07,January,2004','2004-01-07 00:00:00','    In an all too familiar senario, Pittsburgh fans are dissapointed at the decision that Mario Lemieux has decided to call it quits for the year. Many fans have speculated that Mario was only interested in himself and money. However, the truth is that Mario Lemieux, now 38, has given his selfless time, money and energy to save the Penguins franchise not only once but 3 times.  Blaming a bad hip, Mario plans to have surgery to repair the tendonitis in his left hip. Hope for a better year came with the admission from Mario,  \" It\'s disappointing that I won\'t have a chance to return to the ice this season, but I look forward to coming back next year. \"   '),(4812,671748,'05,January,2004','2004-01-05 00:00:00','    A failed clearing pass by Canada\'s goaltender Marc-Andre Fleury ended up in his own net with 5:12 left and gave the United States a 4-3 victory in the world junior hockey championship game Monday. Zach Parise, left, of North Dakota was chosen MVP of the tournament   Another legendary moment goes down in history fueling the rivalry between the United States and Team Canada. The talk of the town is that Marc-Andre Fleury will be undergoing crutiny as to his confidence level and how he would perform in the NHL post-clearing pass mistake. Some die-hard Canadian fans are taking the loss pretty hard as a big dissapointment. The World Juniors are supposed to be a exhibition of young talent for the future, but what does that say for hockey in North America to compete at a dominant level? Team Canada may have fallen a few pegs on the food chain.   '),(4813,671748,'01,January,2004','2004-01-01 00:00:00','    Hooray! It\'s the first blog entry of 2004!!   '),(4814,671748,'28,February,2004','2004-02-28 00:00:00','    Imagine my surprise when I go to check the rounds of my favorite web pages, that Kelly The Patriette has again linked my Hockey Blog to her latest entry. She sounded off on her opinions of the battle of the sexes which included talking about women and hockey. (for sure, two of my favorite things) Here\'s and excerpt on what she had to say on her page. It\'s good stuff!    Whew. I\'m not a big fan of seeing women go into most men\'s-only sports or men going into women\'s-only sports (yay for the differences between men and women), but I used to play ice hockey... And I played because I love the sport - I loved the speed and cold air and the strategy involved in moving the puck across the ice... I played because it was fun.    Read all of the post  urlLink  Patriette TrackBack Link  here!  If you want some insight to some Current Events news from a great gal, please support her cause!   '),(4815,671748,'25,February,2004','2004-02-25 00:00:00','    The Penguins have ended their losing streak of 18 games in a row by defeating the Phoenix Coyotes in overtime 4-3. The newly acquired Jackman scored the game winning goal. The Pens now have a little time off to get the losing out of their system as they go to San Jose and play the Pacific leading Sharks. (yes, that\'s right, the Pacific leading Sharks)   '),(4816,671748,'18,February,2004','2004-02-18 00:00:00','    I can\'t believe that my insider source to the Washington Capitals didn\'t give me a heads up on the Capitals trading that winger, Slovak, and all around great guy Peter Bondra was traded today to the Ottawa Senators. He\'s the latest hired gun in a league that obviously cycles the stars arond during playoff times. \"Loaners\" as we call them. To keep up with the Maple Leafs, the Senators must deal for some players. The Sens also acquired Rob Ray from retirement.   '),(4817,671748,'16,February,2004','2004-02-16 00:00:00','    Well, what do you expect from the game two nights ago the Pittsburgh announcer instead of selling the home team as a reason, but extols the virtues of the away team! John Steigerwald on the radio suggested that the fans come out to see the Mon. night game against the Maple Leafs because they would be the more exciting team.  Of course, he was right. The Pens lost the game 8-4.   '),(4818,671748,'12,February,2004','2004-02-12 00:00:00','    Depending on how you look at it, it could either be the first nail in the coffin for Bob Goodenow or the NHL President Gary Bettman\'s.  Former SEC Chairman Arthur Levitt says in an independant audit that the NHL  \"$273 million on revenues of $1.996 billion\"   Yikes! That\'s a lot of money! I think it\'s a serious bargaining chip in the hands of the NHL owners who want change in the form of a salary cap and revenue sharing.  The story continues on  urlLink  NHL.com  as...   The 24 page study (plus 11 exhibits) reveals that 19 of the League\'s 30 teams had operating losses in 2002-2003 averaging $18 million and that only 11 teams were profitable averaging $6.4 million in profits.   Holy Shnikey!   '),(4819,671748,'08,February,2004','2004-02-08 00:00:00','    Well, the NHL half way point has come and gone. The 2004 NHL All Star Game is but a memory. Joe Sakic walked away from the game as MVP with his hat trick, but the East Conference came out on top. I think last years uniforms were a lot better. The old fashioned knitted sweaters just didn\'t translate well to TV. I don\'t see a lot of people buying them. What was wrong with last years Blue and White combo.  (or they could have gone to a blue/red combination if they want to promote dark colored jerseys more). The next years all star game is going to be in Atlanta...  if  there is a next year!   '),(4820,671748,'07,February,2004','2004-02-07 00:00:00','    On this date in 1976...  Darryl Sittler sets the NHL standard for most points in one game (10) with six goals, four assists, versus Boston. The Leafs defeat the Bruins 11-4.   Well, add this to your trivia collection! I found it on the Maple Leafs page.   '),(4821,671748,'06,February,2004','2004-02-06 00:00:00','    Well,.. the long wait is over.. I can now tell you what I thought of the movie, \"Miracle\"!  First of all, let me get some of the negatives out of the way. I suspected that since Disney produced it, they must have some sort of slant. Nevertheless, they included a voiceover of President Carter claiming in a old speech \"our country\'s goals aren\'t always the best.\" Yea, like the recent Iraq war. Subtle slam on America. But I wasn\'t fooled.  Again, a banner that read, \"Hey Russia, get the puck out of Afganistan\" was shown in the movie. I believe that history was correct in this instance Russia did in fact invade Afganistan, but who\'s in Afganistan right now fighting the Taliban and Al Queida? Subtle Disney-Liberal dis to the Red, White, and Blue.  With that asside, Kurt Russell playing the late coach Herb Brooks deserves an Oscar nomination.  He really showed the driven nature of \"Herbie\" and fought to keep a bunch of college amateurs playing as a team. My favorite scene is when he made the entire team run wind sprints after a game until they realized that they represented a cause larger than themselves.  Go see the film or go read my review at  urlLink  The College Press  in greater detail.   '),(4822,671748,'04,February,2004','2004-02-04 00:00:00','    I guess they outlawed freedom of speech in Canada. Even after my experience in Toronto, I\'m convinced that some sort or aparteid is going on there.  I read on TSN\'s NHL Network page that co-host of CBC\'s Hockey Night in Canada, Don Cherry, is undergoing investigations after comments he made on his editorial segment called \"Coaches Corner\". Coach Don Cherry said that the only hockey players who wear visors are europeans and \"the french\". Well, that pissed off a lot of liberals in the country. (Despite the statistics that back up Cherry) A statement like this in Canada qualifies as a \"hate crime\". The liberals have been trying for years to get Cherry off the show, and really dislike him for backing the U.S. during the Iraq war.   Well, besides Portland, OR, and Winnepeg, Manitoba, the latest news out of Ontario is that a law firm in Hamilton is looking into acquiring an NHL team and move them to their city. Already an unidentified NHL team was looking to relocate to Winnepeg. It\'s another sign that all things are cyclical, Canadian teams were once being lost to U.S. and now teams are trecking north of the border.   '),(4823,671748,'01,February,2004','2004-02-01 00:00:00','    urlLink create your own visited states map   or  urlLink write about it on the open travel guide   This map is to show you how many states I\'ve visited in my lifetime. Surprisingly enough, they all have hockey teams in those states. Actually, I think I can say that every state has at least one hockey team in it. It\'s a universal game.   '),(4824,671748,'01,February,2004','2004-02-01 00:00:00','    Today, I posted my first column on  urlLink The College Press  and of course, I talk more about the game I love so much... hockey. Please check up on it weekly as I plan to post a new column every month. If you want, I encourage and reccomend that you read the other columns and articles on the website   '),(4825,671748,'31,March,2004','2004-03-31 00:00:00','    Mark Messier took a final lap around MSG tonight. It is the final game of the year for the rebuilding Rangers. He took a long drawn out, but well-deserved bow to all the fans. I could tell that he was trying not to get emotional during the moment.  This by far is not the official signing off. He hasn\'t called it \"quits\" just yet. Messier said good-bye to the fans today in a really classy way. He even got to score one final goal.   '),(4826,671748,'29,March,2004','2004-03-29 00:00:00','    ...order an Aeros ticket and get to see  WWE Womens Champion Victoria  at the game, plus receive a ticket to next Monday\'s WWE event, for only $30!    I\'m sure you all were in dire anticipation over just who the Houston Aeros were going to have promote the WWE Raw and the Aeros game.   '),(4827,671748,'25,March,2004','2004-03-25 00:00:00','    The NHL continues to punish the \"evildoers\" by cracking down on the smacking down. Two more player suspensions have been handed out today.    1. Marty Turco (4 game suspension) for losing his cool and giving a scythe-like slash to the face of an Oiler player.  and..  2. Jody Shelly (3 game suspension) for sucker punching and getting wild to a Minnesota player.   '),(4828,671748,'23,March,2004','2004-03-23 00:00:00','    The mainstay to the New Jersey Devils franchise is just how well Martin Brodeur has done in between the pipes. In a time when athletes life span with any one team averages 3 seasons, Brodeur has spent his entire career with the Devils. And he wins a lot of games for them.  Tonight he won career game #400 against the Florida Panthers. His career marks are 400-215-105 in 12 seasons.  Congrats to him.   '),(4829,671748,'21,March,2004','2004-03-21 00:00:00','    The new crackdown \"Zero Tolerance\" for fighting and other misbehavior since the Bertuzzi-Moore incident has brought on a bevy of fines and suspension.  1. Wade Belak (Suspended 6 games and 2 playoff games) for chopping an Avalanche player with his stick. 2. Mark Messier (Suspended 2 games) for pitchforking a Penguin player in the gut.  and now...  3. Chris Simon (Suspended 2 games) for kneeing a Stars player.   '),(4830,671748,'21,March,2004','2004-03-21 00:00:00','    Houston, we have a problem!   I think the  urlLink Houston Aeros  of the  urlLink AHL  didn\'t get the memo! Aparently, they missed on the idea that hockey is trying to get away from the goonism and fighting. But if you go to their website, you\'ll see the press release featured on it.  Here\'s a quick summary of where Houston who has a great idea of a cross promotion, but the timing of it just stinks! Judge for yourselves.   The Houston Aeros and WWE are tagteaming to bring a WWE Superstar to the Thursday, April 1 game against the Cincinnati Mighty Ducks. The Superstar will be announced next week.  In addition, by calling the Aeros office, fans can purchase an Aeros corner seat for the April 1 game, plus a WWE ticket for the Monday, April 5 appearance at Toyota Center, for just $30.   '),(4831,671748,'18,March,2004','2004-03-18 00:00:00','    When in times of need, some of us forget to accentuate the positives in life. We could fret over all the negative press coverage that the \"Bertuzzi incident\" caused or I could tell you about what the mainstream media did not.  The  urlLink Jennifer Ferraro Foundation  was set up to give help and hope to people suffering with cancer disease. Ms. Ferraro was the sister to Chris and Peter current players of the Washington Capitals franchise. Last Saturday night (4/13/03), the Springfield Falcons and Portland Pirates played a game where the Falcons wore specially designed \"Team Jennifer\" jerseys and auctioned them off to the public for charity.  Do yourselves a favor and go visit the website and donate what you can to worthy causes to fight cancer and other diseases. I\'ve added a special section to collect website links for the charity efforts.   '),(4832,671748,'17,March,2004','2004-03-17 00:00:00','    Happy St. Patrick\'s Day!!!!!   '),(4833,671748,'16,March,2004','2004-03-16 00:00:00','    I have to way in on more of the media coverage to the Bertuzzi-Moore incident. I laughed my ass off at the post that  urlLink Joe Taska  has up recently at his Hockey commentary blog. I suppose I should credit the original source as  urlLink Sally Jenkins from the Washington Post     Real hockey people wear \"Free Bert\" T-shirts, like those fans in Vancouver who rallied in support of Bertuzzi over the weekend, led by rally organizer Sean Carl. According to Carl, who is a real hockey guy, Bertuzzi is being persecuted and Moore\'s injuries have been exaggerated by the Colorado Avalanche.    Histerical!  I do think that Bertuzzi does deserve a little support in all of this. He\'s still a human being and shouldn\'t be villified to the point that we\'ll be running commericals on the TV comparing Bertuzzi to Hussein.   '),(4834,671748,'14,March,2004','2004-03-14 00:00:00','    Okay, so I\'m definitely enjoying watching the Pittsburgh Penguins! They have a stretch run of rivalry games in the schedule. So far, they\'ve played up to their competition. Bob Errey (the Pens commentator) has mentioned that they played that way against \"non-playoff teams\"  This afternoon they had the dreaded Philladelphia Flyers. They gave a tough attempt, but could only keep the tie. 3-3. Tuesday, the Pens play the \"new look\" (translation = salary-dumped) Capitals at the Mellon Arena. Here\'s the rest of the Pens \"rivalry stretch\" schedule.    March 7   @ Rangers March 9    Stars March 11 @ Maple Leafs March 14  Flyers March 16  Capitals March 17 @ Devils March 19  Hurricanes March 21  Rangers March 23 @ Rangers   '),(4835,671748,'11,March,2004','2004-03-11 00:00:00','    With the help from  urlLink Vertical Alliance \'s PR digest newsletter, The Florida Panthers have released this public relations ploy to get more people out to the game. Here is an excerpt from that clever email.    Knowing that many Canadians migrate to Florida during the winter months, the Florida Panthers created a \"Canadian Night\" promotion targeted specifically to Canadian citizens. Canadian fans were able to purchase tickets in Canadian dollars, which gave them a significant discount as the Canadian dollar is currently worth about 24 cents less than the American bill... The Panthers held the first \"Canadian Night\" promotion at a game against the Toronto Maple Leafs. The event was only promoted through the Canadian press. The promotion was a success for the Panthers, as a sellout crowd of 19,000+ attended the game... (National Sports Forum\'s \"Selling It\", 3/3)   '),(4836,671748,'10,March,2004','2004-03-10 00:00:00','    As reported to the media, the Indianapolis Ice CHL team has chosen not to offer maligned Figure Skater turned \"white trash ho\" Tonya Harding a roster spot. She will then only be religated to an autograph signing at Fridays Ice game.  I would have liked Hardings credentials given her skating background, she\'s killer with a stick!   '),(4837,671748,'10,March,2004','2004-03-10 00:00:00','    PM to hockey: \'clean up your act\'    In what I thought would be a true return of some interest to covering hockey, \"we fans\" now see the public relations disaster spreading in the media. Their bias against the NHL is only fueled by the latest controversial incident involving Todd Bertuzzi #44 of the Vancouver Cannucks and Steve More #36 of the Colorado Avalanche. Apparently, there was some bad blood between \'Couver and Moore who was alleged to have taken a hit on Naslund some time ago. Bertuzzi saw fit to stalk him on the ice and sucker punch him from behind. Moore hit the ice with such force it broke his neck and now puts his career in jepardy.   \" I don\'t think it\'s given Canada a black eye, but I certainly don\'t think it\'s helped hockey, \'\' said Liberal PM Paul Martin.   Indeed, Mr. Martin. This is not good for the casual fan who won\'t put up with said antics. ESPN, CNN, CNN-HN, Fox Sports are all talking about it, but holding a negative referendum on the violence. The anti-hockey crowd is just licking their chops with the negative spin.  If it can\'t get any worse. Look what I found on the AP wire...  AMERICA\'S BAD GIRL IN ACTION FRIDAY AS PART OF \"GUARANTEED FIGHT NIGHT\" ...  reads the headline in the Indianapolis Ice on their   urlLink official web site  as disgraced and maligned Figure Skater Tonya Harding has signed on to play in a CHL game this coming Friday.  If this isn\'t a publicity stunt, I don\'t know what is. Hockey is officially out of control!  I now pronounce myself as Hockey Czar and ruler of all things on \"the frozen pond\"! *smile*   '),(4838,671748,'09,March,2004','2004-03-09 00:00:00','    With the trading deadline over now today at 3 PM EST, there were indeed some shocking trades and developments in the National Hockey League.  Veteran Center and all around classy guy, Ron Francis, was traded to the Toronto Maple Leafs in exchange for a 4th round draft pick. In what is sure to be seen as a playoff rental, Francis was expected to finish his career with the organization that he is mostly associated with in Hartford/Carolina. He also spent a long time earning respect as a role player on the Pittsburgh Penguins teams during the 90\'s winning a cup with them in \'92.  The Vancouver Cannucks warned everybody that their moves to improve their roster would be calm and logical. They went out to several sources like the Rangers and got Rucinsky (for R.J. Umburger who grew up in Pittsburgh and many Pens fans wanted R.J. to come home) and Geoff Sanderson from Columbus.   '),(4839,671748,'08,March,2004','2004-03-08 00:00:00','    Has this ever happened to you???  I was playing NHL 2004 for Playstation 2 simulating a game using the up to the minute rosters of the Toronto Maple Leafs and Dallas Stars. In the first period, I noticed that former Leaf, Shayne Corson, did not have the correct jersey number on him when I started the game. He did haunt his former teammates with a goal in the 1st period. Dissapointed that my new favorite team in the Leafs were losing 2-1 at the first intermission, I turned off the game in favor of updating a few things on the Dallas roster.  When I went back and re-simulated (meaning I played the game, as I often do, by having the computer control both teams) the Toronto v. Dallas exhibition, I noticed that it picked up right where I had left it. The first period goals were exactly the same. Morrow and Corson both scored in the same manner exactly.  Is this how things work out in real life? I don\'t know!   '),(4840,671748,'07,March,2004','2004-03-07 00:00:00','    This is another gem from my new friend over at  urlLink Tasca\'s Take  talking about the Hockey fighting trends and the ogres that root them on! (a.k.a. Flyers fans)    The Rogers Sportsnet broadcasters seemed completely befuddled by the euphoric reaction of the Flyer fans during the course of the fisticuffs. The reason why is anybody\'s guess, as Philadelphia is widely known for its appreciation of bloodshed.   '),(4841,671748,'06,March,2004','2004-03-06 00:00:00','    Seeing the need for a reputation revamp, Dallas Stars Shayne Corson lashes out at his former teammates, organization, fans and Toronto media by offering his reasons for quitting on his team during last year\'s playoffs. You can read it here, or you can go to TSN\'s NHL Network web page for the story.   \"Hey, if I screw up on the ice, criticize me. I\'m man enough. I\'ll take the heat. It\'s part of the job. But some of that stuff . . . I just wasn\'t healthy enough to make any sort of contribution to the team. People can believe what they want, but that\'s the truth.\'\'  He blames his behaviour last year on panic-anxiety attacks, for which he\'s sought out additional professional help. He\'s also changed medicine and the results have been remarkable.    Spin it any way you like, I hope you got help for your issues, but you still  QUIT !!!   '),(4842,671748,'05,March,2004','2004-03-05 00:00:00','    With special thanks to my new friend  urlLink  Joe Tasca , I read a very amusing story on his site as reported to him through pressconnects.com. It seems as if Bing-hampton Senators goaltender Ray Emery has had a very shaky life in the AHL. It seems as if the Wilkes-Bare Scranton Penguins are his favorite team to play against. In Wednesdays game, Emery was given a match penalty a 5 minute major for striking WBS player with his blocking pad. It was the final factor to the game as the Pens won 7-3 off two power play goals. Tasca has featured on his site has the rap sheet on the volitaile netminder.  In December of 2003, Emery skated the length of the ice and pummeled Wilkes-Barre goaltender Andy Chiodo (now the starting goaltender in Pittsburgh) during an all out brawl.   '),(4843,671748,'04,March,2004','2004-03-04 00:00:00','    I\'m sure you all have lost sleep worried on weather Russian sensation Nikolai Zherdev would have to be returned to his junior league equivalent team because he left to play for the NHL\'s Columbus Blue Jackets too early! According to the two major sources, Canada\'s TSN and NHL.com both have article headlines saying that Zherdev can stay in the NHL. The Russian millitary\'s team, CSKA Moscow, claimed he did not get IIHF aproval and still had time on his millitary commitments.   '),(4844,671748,'03,March,2004','2004-03-03 00:00:00','    WINNIPEG (CBC) [2/11/04] - Don Cherry was right when he claimed mostly French-Canadian and European NHL players wear hockey visors, says a Winnipeg lawyer who crunched the numbers. Curtis Unfried told CBC Newsworld on Wednesday that his investigation of data showed 59 per cent of European players wear visors, followed by 55 per cent of French-Canadian players. Among North American players born outside of Quebec, only 20 per cent wear face shields to guard against injury.    Well! It looks like some Gov\'t officials and fans owe Coach Cherry an appology!   '),(4845,671748,'03,March,2004','2004-03-03 00:00:00','    It\'s a dis-arms race between Washington and New York!  In the past two days, the Rangers have given up on Petr Nedved, Alexi Kovalev, and Brian Leech shockingly trading them to Edmonton, Montreal and Toronto respectively!  I\'m disapointed that all these stars are getting moved before the trading deadline! All of the playoff teams are bolstering their rosters, but when it comes time for March 9th, things will be pretty boring to report on!   '),(4846,671748,'02,March,2004','2004-03-02 00:00:00','    BRATISLAVA, Slovakia (AP) - Prime Minister Mikulas Dzurinda said Tuesday that former NHL star Peter Stastny will lead his party\'s list of candidates in elections for the European Parliament.   How about that! If you don\'t pay your taxes! He\'ll send you to the penalty box!   '),(4847,671748,'01,March,2004','2004-03-01 00:00:00','    I\'ve been reading up on hockey going on in the rest of the continent and I\'m interested in what is going to go on with the American Hockey League if or when the NHL goes on strike. It looks to be that most of the NHL owners are going to keep a lot of their young talent to the minors. That way, they can be exempt and not sitting on their hands during the strike. Good idea!   While the NHL heads toward what could be a lengthy work stoppage, the AHL is flourishing. Should there be a lockout or strike in the fall, its top development circuit might benefit in popularity, exposure, even in significance. But the last thing the AHL wants to see is a hockey Armageddon.   They also want to tweak with the game to make it match more of the NHL rules. Some of the NHL GMs made a few suggestions on   urlLink this article     A system that will give three points for a regulation victory, two for an overtime victory, and one for an overtime loss.  A game-deciding shootout. A shootout winner will get two points, while the losing team gets one. Tie games would be eliminated.  Expanding the width of the blue and red lines.   '),(4848,671748,'01,March,2004','2004-03-01 00:00:00','    Well... the NHL Trade Deadline is coming up on March 9th and people are talking about the teams that may get geared up for this upcomming post season dance.  According to my friends at the  urlLink NHL Network , trusted source - Bob McKenzie, says that Atlanta Thrashers goalie Byron Dafoe is rumored to be traded to the Maple Leafs.  This would be an incredibly stupid move for the \"Leaves\" since they have proven already stocked in the goaltender dept. despite \"Unstable Eddie\" future with a back sprain. Of course, the well-traveled veteran Dafoe is under some questionable scrutiny. He recently got hurt and embarrassed for slipping on the ice outside his hotel not too long ago. As a hockey player, you\'d think he\'d be able to traverse the ice! Dafoe even got attention from Paul Harvey\'s commentary by focusing on his ironic mishap.   '),(4849,671748,'30,April,2004','2004-04-30 00:00:00','    While surfing around and listening to music online related to hockey, I found this one song sung by Canadian rockers \"The Tragically Hip\". Listening closely, I heard the singer mention something about \"the last goal he ever scored won the leafs the cup.\"  I\'d say that song was about hockey. However, I really just enjoyed the music without knowing what exactly the song meant. I had to watch tonight\'s game of Philly v. Toronto on ESPN as explained by Steve Levy to get the full story.  The \"fifty mission cap\" song is about the mysterious cycle of how the Leafs won the Cup in the early 50\'s. The guy who scored the winning goal unfortunately dissapeared that summer after taking a plane to go fishing. The Leafs didn\'t win the Stanley Cup championship for over 10 years. They finally found his remains and that same year, the Leafs won the cup in 1962.  Believe it or not, read for yourselves at a web page honoring one of the Leafs players to have his jersey number actually retired.  urlLink Barilko WebPage  Today, Darcy Tucker of the Toronto Maple Leafs hope that Brilko\'s Ghost or Legend will work one more time. He hopes the media will pick up the story and fans re-discover Brilko in hopes that Toronto wins the Stanley Cup again this season.   '),(4850,671748,'28,April,2004','2004-04-28 00:00:00','    Folowing up on the controversial issues surrrounding the CBC and their segment \"Coaches Corner\" I\'ve compared two Toronto newspapers  urlLink Toronto Star  and the  urlLink Toronto Sun  talking about what to do about the controversial host, Don Cherry.  It seems as if Cherry will no longer be a part of the traditional Hockey Night in Canada on the government owned CBC.  However, if their government were at all interested in listening to their citizens and believed in free speech, they\'d cite this online poll from Slam! Sports:   After 12,000 votes on  Slam! Sports Webpage  60% said keep him.   Globe and Mail of Toronto (the guys that do the TSN stuff) offers a poll of their own:     If Don Cherry leaves the CBC, will you miss him? No    12752 votes   (51 %) Yes   12348 votes   (49 %)   '),(4851,671748,'28,April,2004','2004-04-28 00:00:00','    Yes.. you saw correctly!  Union President Bob Goodenow and NHL Commissioner Garry Bettman sitting together in Torono\'s Air Canada Centre to watch the playoff game Maple Leafs v. Flyers. They were initially there to re-establish talks to extend the Collective Bargaining Agreement.  ESPN\'s cameras picked it up and I\'m guessing that CBC and TSN picked it up as well.  The Flyers got their asses beat tonight against the Leafs. I think the few days off from the series, plus getting out of that shit hole called Wachovia Center helped big time. The \"Leafs Nation\" were all there in Full force today waving their banners, towels, and stuff. Melrose on NHL 2night said that the Leafs always play with heart and desire. I believe him now for sure!   '),(4852,671748,'26,April,2004','2004-04-26 00:00:00','    Continuing with my run of diverse topics to write about, here\'s a little story off the wire concerning Team USA going down the tubes.   OSTRAVA, Czech Republic -  urlLink (Yahoo! News)  Defenseman Richard Lintner scored with 6:52 left in a wild second period Monday as Slovakia kept the United States winless at the World Hockey Championships by rallying for a 3-3 tie. Lintner\'s equalizer capped a five-goal outburst and came just over five minutes after Pittsburgh Penguins rookie Ryan Malone put the United States (0-1-1) in front on the power play.... The Americans must earn a win or tie against Ukraine on Wednesday to reach the qualifying round.   '),(4853,671748,'25,April,2004','2004-04-25 00:00:00','    Okay.. I\'ll do some coverage of the American Hockey League playoffs for you now. The baby Pens from Wilkes-Barre Scranton have pulled it off in Game 7 and pass the Bridgeport Sound Tigers (Islanders) in Overtime. M\'boy Konstantin Koltsov scored the game winning goal. Chicago Wolves (Thrashers) and the Worcester IceCats (Blues) also have advanced to the next playoff round taking care of Grand Rapids Griffons (Red Wings) and Manchester Monarchs (Kings) respectively.   '),(4854,671748,'24,April,2004','2004-04-24 00:00:00','    Since NHL has currently dropped the puck on growng the sport, the Board of Commish\'s have proposed drastic measures to scale back the NHL season.   urlLink NHL Network and TSN  report that the NHL will shorten their season to 72 games. The worst of it is that they want to emulate Major League Baseball so much to restrict the Eastern and Western teams from playing each other. The theory is that the regional rival games bring in more fans. But NHL Network reports that the Western Canadian teams like it when the Leafs come to town.   '),(4855,671748,'24,April,2004','2004-04-24 00:00:00','    New Type of Arena?  The  urlLink Pittsburgh Tribune Review  may break the story that a new type of arena may be built in Pittsburgh. A group of Wall Street investors would like to build a $150 million arena for the Penguins and the city to use. Only this would be the first underground arena in the world!  I\'ll be watching the developments of this story as it continues to unfold.   '),(4856,671748,'22,April,2004','2004-04-22 00:00:00','    Well.. the first round was really exciting.. time now for my predictions for round two! I got most of them right! Montreal really surprised me.   Western Conference   Detroit v. Calgary    The age is starting to catch up with Detroit. But their greatest weakness can be their greatest strength. Detroit\'s veteran leadership got them through the first round. Stevy Y, Hull, and Lang have been solid. Calgary is the best story of the year. Iginla is the star that he should be and it\'s been a while since Calgary has had anything to root for. Matter of fact, The last time the Flames and Wings met in the playoffs, the Flames were in the old Omni in Atlanta.  Prediction: Wings in 6    San Jose v. Colorado  What a difference a coach and a year make! San Jose has been in the playoffs before, but never this high. The \"shark tank\" will be rocking tonight! Selane and Granato will treat it like a homecomming as they both have played for the Sharks. The Avs think that 4th place is a fluke and hopes to make it to the Conference finals. The have the firepower, but will the Sharks no-name defense eat them alive?  Prediction: Sharks in 5    Eastern Conference   Tampa Bay v. Montreal  Montreal faked their way into the second round. The controversial Ribiero move cost a lot of luster from the \"storied\" Montreal legacy. Tampa Bay is just happy to be there! However, in a reversal of fortunes, Tampa is the king of the eastern hill. Martin St. Louis for Tampa is getting a lot of pundits to dub him Hart Trophy and Lady Bing at the end of the year.  Prediction: Lightning in 6   Philidelphia v. Toronto  It\'s deja vu all over again! These two teams met last year with the Flyers on top. \"The Flyers are a team that you don\'t have to work hard for to hate,\" said Toronto coach Pat Quinn. You can\'t ask for any better bulletin board material. Flyers are banking on their goalies and the \"blackhawk line\" to pull them through. Roenick, Amonte, and Zhamnov were all teammates in the windy city.  Prediction: Leafs in 6  - hopefully.   '),(4857,671748,'19,April,2004','2004-04-19 00:00:00','    Happy Birthday to me.... !!   Happy Birthday to me.... !!   Happy Birthday....      to me.... !!   Happy Birthday to me.... !!   '),(4858,671748,'14,April,2004','2004-04-14 00:00:00','    Stanley Cup Races turning ugly...  With more drama than a Presidential race.. the Stanley Cup Tournament Playoffs have had more mud slinging than than a girl \'wrastlin\' at the local bar. *wink*  Ottawa\'s Team 1200 caught some serious heat when some of their broadcasters lampooned and inferred that Toronto\'s enforcer Tie Domi beats his wife. Domi\'s response is that he\'s thinking of taking legal action.  Toronto\'s Coach countered by sullying the image of the Ottawa Senators saying \"The Senators misled the fans by giving off a vibe that they\'re all good guys when they\'re just as dirty\".  In the Bruins v. Canadiens series... things have gotten so ugly that it threatens the very fabric of the NHL\'s credibility. Mike Ribiero layed down on the ice and got the refs to stop action. Later, Ribiero was seen on the bench laughing. The Bruins questioned Ribiero\'s health and said that he may have just been faking. Then during the Bruins v. Canadians on Tuesdays game, Alexei Kovalev got slashed on the hand, ran into his own player and Boston scored the game winning OT goal. Experts theorised that Kovalev could not be that hurt through the glove.  The best part was that big Barry Melrose definitely looked pissed that the Canadians have stooped this low on ESPN\'s NHL2night.   '),(4859,671748,'07,April,2004','2004-04-07 00:00:00','    The playoffs begin today, so.. just like last year, here\'s my interpretation and predictions of the first round.   Western Conference    Detroit v. Nashville  This is the most watched playoff series as Detroit hopes to not repeat last years sweep by the Ducks. Nashville tried to garuntee a playoff run last year to their season ticket holders, but they made it this year. Nashville earned their spot down to the last day of the season. Manny Legace is expected to start in goal.  Prediction: Wings in 5    San Jose v. St. Louis  The quietest team going into the playoffs and the team that gets the least amount of respect in Division winning San Jose is looking to advance. St. Louis dumped their coach and was riddled with injuries for the past two years. I don\'t know how they got into the playoffs again. Osgood is huge in the playoffs. That should help the Blues.  Prediction: Sharks in 5    Vancouver v. Calgary  Vancouver minus Bertuzzi is interesting to watch. Fresh faced Calgary is a blue collar team looking to build upon their rebuilding. Watch for the Canucks to pay dividends in their deals at the deadline.  Prediction: Canucks in 7    Colorado. v. Dallas  In the most evenly matched series of the Western playoffs, rivals collide! Aebisher is the key to Colorado because their big gun Paul Karyia may be done for the year due to an injury. He took a pay cut believing the \'Lanche to be his best shot at winning the Cup. Dallas isn\'t the team they used to be.  Prediction: Avalanche in 5    Eastern Conference    Tampa Bay v. NY Islanders  In the East, the least respected is Tampa Bay (\'cuz who thought of Hockey in Florida?). They were in the running for the Presidents Trophy. The Isles have gone far with Yashin and hopes their investment in #1 pick Rick DiPietro will get them to the promised land.  Prediction: Tampa in 4    Boston v. Montreal  Original 6 rivals collide! No two teams have met in the post season than these two teams! Boston rides the back of Andy Raycroft, a Calder Trophy nominee. Montreal is just happy to be there where losing is no option. Prediction: Bruins in 5    Philidelphia v. New Jersey  Atlantic Division and local rivals go at it where the defending champs have slipped from grace. The Devils are missing Stevens, but have the equalizer in Brodeur. Philly lost veteran Dejardins after hurting his hand playing catch with his son. Philly is stocked for a deep playoff run. Sorry, Jersey no repeat this year Prediction: Flyers in 7    Toronto v. Ottawa  In the best evenly matched teams in the East, the \"Battle of Ontario\" look to take repeated shots at each other in what is turning out to be a great rivalry. Toronto was active at the trading deadline to match up with thier rivals. Ottawa also hired ex-Cap Peter Bondra. Sorry Sens loyalists, Go Leafs Go!  Prediction: Maple Leafs in 7   '),(4860,671748,'06,April,2004','2004-04-06 00:00:00','    The Penquins got screwed!! Big Time!!  The Washington Capitals walked away from the Draft lottery getting the first overall pick, despite the Pens reluctantly getting last place in the NHL. Here\'s how the top five are set to pick in June 2004  1. Capitals 2. Penguins 3. Blackhawks 4. Blue Jackets 5. Coyotes   '),(4861,671748,'04,April,2004','2004-04-04 00:00:00','    I got my one chance to see the Pens at Mellon Arena today as it\'s nice to have \"insider\" connections to get such great seats! :) 16,000+ people were there to cheer on the Penguins versus the Capitals. I got a chance to see my favorite player \"Millenium Man\" Milan Kraft score the game winning goal.   '),(4862,671748,'02,April,2004','2004-04-02 00:00:00','    In Other News...  Adam Oates held a press conference today to announce that he finally discovered that he was too old to play and retired. (Now that his Oilers missed this years playoffs).   '),(4863,671748,'02,April,2004','2004-04-02 00:00:00','    Tonight, the Thrashers organization paid tribute to #37. Dan Snyder was tragically killed in a car accident. \" He\'ll have to live with that the rest of his life, \" said Pens Hall of Fame announcer Mike Lange during the game about the perp responsible, Danny Heatly.   Also, Kovalchuk was going for a 40 goal season. You could tell how hard he was trying when all of his teammates were just serving it up to him every chance they got. Ilya had 8 or 9 shots all by himself in the first period. He was so visibly frustrated at how Pens goalie J.S. Aubin covered him well by stopping every chance.  However, Kovalchuk would not be denied! He got a goal with less than 30 seconds remaining in the game to make it an even 40 goals during the regular season. Hollywood can\'t come up with such natural drama!  He gets one more chance to tack on more goals during the Thrashers last game of the year against their division rival Tampa Bay.   '),(4864,671748,'01,April,2004','2004-04-01 00:00:00','    I\'m wishing that this is a April Fool\'s Joke, but sadly it seems legit! Aparently, they didn\'t get the memo that hockey is trying to move away from all the negative stereotype that fighting and goonism has brought to the game.  Read on...   ATHENS, Ohio -- If you ever have uttered the words, \"I went to a fight and a hockey game broke out,\" you are exactly the type of person who is encouraging people like Darryl Wolski. Wolski is the man behind an upcoming Pay-Per-View event called \"Hockey Gladiators,\" an affair that takes former pro and semi-pro hockey players known more for fistcuffs than puck control and lets them beat the heck out of each other. Fans get enjoyment, and the players get a few bruises and a chance at a $100,000 prize. \"The idea about Hockey Gladiators came about in a lounge after a senior hockey game five years ago,\" Wolski said in an e-mail interview. \"Hockey highlights were on the screen and a fight was shown. The entire place went crazy.\"  urlLink   '),(4865,671748,'29,May,2004','2004-05-29 00:00:00','    In American Hockey League News, the Wilkes-Bare Scranton Penguins have advanced to the Calder Cup Finals by beating the Hartford Wolf Pack (NY Rangers organization) in OT. They will move on to play the Western Conference Champs in the Milwaukee Admirals (Predators organization) in a best of 7 series. This is the second trip to the Calder Cup Finals for the short franchise life of the WBS Penguins. The Penguins made the Finals in 2001, but lost to the now defunct Saint John Flames. Milwaukee this year is one of the teams the AHL admitted when the International Hockey League folded.   '),(4866,671748,'27,May,2004','2004-05-27 00:00:00','    Happy Birthday,  Confessions of a Hockey Fanatic  Weblog    I am in the process of upgrading and remodeling the Weblog in anticipation for it\'s second Anniversary and I hope you like the new layout template.  Some of the new features to the new site layout would be an expanded profile section - so you can learn about where I\'m from and what my interests are, a Comments section so you can give me feedback to my coverage. Let me know if I\'m fair, even coverage, or way off the mark!  Please come back later as I will get all of the blog, news, league, and other links up soon. Some of the links will be discontinued, some of the links will be moved around, and some of the links I\'ve endorsed over these past 2 years will no longer appear on  Confessions of a Hockey Fanatic  Weblog due to inactivity or contain content that just does not fit in with the Hockey format.  Thank you all for reading up on my  Confessions of a Hockey Fanatic  Weblog and I\'ll see you when we have a Stanley Cup winner!   '),(4867,671748,'26,May,2004','2004-05-26 00:00:00','    It takes a well-trained journalist to spot a fake. Pittsburgh Sports Writer Bob Smizik got himself in hot water today with a column article that he wrote for my hometown Pittsburgh Post Gazette. You can read it if you want at this link  urlLink Pittsburgh Post Gazette & Smizik . I must admit that I to was sucked in initially to Smizik\'s opinion, but giving it a little bit more thought, it\'s clear that all he wants to do is attract attention to himself driving up his readership.  (aparently it worked)  In the article, Bob Smizik attempts to criticize the way the playoff rosters are set up for the American Hockey League. He claims that had the Flyers not gone deep into the NHL playoffs, then they too could have loaned their NHL players to the Phantoms and the Baby Pens of Scranton could not have advanced as easily as they did. The truth of the matter is, Bob, if you don\'t like the way the final rosters are set up for the playoffs and how they are played simultaneously to the NHL playoffs, DON\'T tear down what this Penguins organization has worked so hard for. Don\'t point the finger at them. They play by the AHL rules.   '),(4868,671748,'25,May,2004','2004-05-25 00:00:00','    Time for the Stanley Cup Finals - winning now means you are riding high as champions of the NHL and able to party with Lord Stanley\'s Cup all summer long!   Stanley Cup Finals   Tampa Bay v. Calgary  Calgary won the Stanley Cup back in 1989 with quite a number of future Hall of Famers. (Lanny McDonald, Mike Vernon, Al McInnis to name a few.) The Flames team captain Tampa Bay is shaped up to look just like the Montreal of the late 80\'s (that\'s back when their team was really good). All of the Lightning stars are from Quebec providence (Richards, St. Louis, and Lecavalier) and \"The \'Boulin Wall\" Nikoali Khabibulin is playing just as strong if not stronger than what Patrick Roy did in his prime. I am very upset at how the mainstream media has been covering this Stanley Cup Finals matchup. ESPN (the Disney-owned sports broadcasting company) should be the ones to take the lead and build the news hype and cover the Finals the way they should. Instead, I hear negativity of only how the ratings are down and the Pundits would most likey want to see the league just die off! To the American media credit, they have featured Calgary Flames captain Jarome Iginla. Not so fast!! They have, however, taken a considerable amount of coverage time to Iginla\'s ethnic background. Since when is a players\' ethnicity a story? IT DOESN\'T MATTER WHAT SKIN COLOR HE IS! Tampa Bay is just happy to be here. In their decade long NHL run, they\'ve reached the pinacle and are Bolt-ing to the finish line. Tampa has 4 lines that can score and as I mentioned before a goalie who has a record worth of shut outs in these playoffs. They were the #1 team in the Eastern Conference and represenative of the \"southleast division\".  Prediction: Flames in 7   '),(4869,671748,'24,May,2004','2004-05-24 00:00:00','    From  urlLink NHL.com .. like... woa! Dude! He\'s like totally a fan & stuff!   Movie star Keanu Reeves is the latest in a series of celebrities to participate in the NHL\'s \"HOCKEY ... GET IT? GET IT!\" brand advertising campaign. Reeves, raised in Toronto and a Canadian citizen, is an avid hockey player and fan. One of his childhood jobs was sharpening skates at a local ice rink skate shop.\"   '),(4870,671748,'23,May,2004','2004-05-23 00:00:00','    Do you feel that sometimes decisions in sports are the results of meetings between greedy owners and incompetent union leaders?  Please, I recommend that you all go to  urlLink National Hockey League Fans Association  (NHLFA) website and sign up to become a member.  The NHLFA\'s mission statement is to become the first Professional Sport that recognizes the fans for their tireless dedication and offers input on mediation to some NHL issues.  If this group was around earlier, could they have possibly been able to prevent Hartford from moving to Carolina? Could the NHLFA be the voice of reason when negotiating the Collective Bargaining Agreement?   '),(4871,671748,'23,May,2004','2004-05-23 00:00:00','    Congratulations to Tampa Bay Lightning and all their fans for making this year the best in their decade in the NHL.  \"Safe is Death\" is Head Coach John Tortorella\'s motto.  Tonight they defeated the Philadelphia Flyers in Game 7 by the score of two to one.  Fredrik Modin (#33), Lightning player for 5 seasons, scored the game winning goal in the 2nd period. The happiest player on the Lightning roster is team captain and noted \"old\" veteran Dave Andreychuk. This is his 22nd NHL season and he has never won the Cup.  Tampa meets up with the waiting Calgary Flames in the Stanley Cup Finals on Tuesday.   '),(4872,671748,'19,May,2004','2004-05-19 00:00:00','    Everybody in Alberta can breathe easier tonight.  The Calgary Flames defeated the San Jose Sharks 3-1 tonight in Calgary to become the Western Conference champions. Congratulations to Calgary and a whole country who longs for a championship after creating the sport.  The Flames last won the cup in 1989 against Montreal.  They wait the winner of Tampa Bay and Philadelphia.   '),(4873,671748,'18,May,2004','2004-05-18 00:00:00','    Thank you Peacock! I could barely believe my eyes! This morning I read the news headlines and on  urlLink Yahoo! Sports  reads and article about how NBC wants to air NHL games next year and include a revenue sharing plan that could very well work out for all the 30 NHL franchises and bring in some new money to help everyone out.  Great news! Things are looking up!   '),(4874,671748,'18,May,2004','2004-05-18 00:00:00','    Okay.. stop what you\'re doing!  Go out and pick up a copy of  Miracle  on DVD. It comes out today.  It\'s the Disney version of the 1980 U.S. Hockey team led by Coach Herb Brooks and defeated a powerful Soviet team on their way to win the Gold Medal at Lake Placid, NY.  I reviewed it when it first came out, you can TrackBack that entry  urlLink [Dated: 2/06/04]  on my blog archives if you wish. For a longer version, check out my column  urlLink A Hockey Knight in Pittsburgh  for  urlLink The College Press   '),(4875,671748,'15,May,2004','2004-05-15 00:00:00','    The defending champion team Canada have announced their roster for the World Cup team to be played later this summer. The team was selected by \"the great one\" Wayne Gretzky and Kevin Lowe as a comitee of Executive Directors.  Two names are obviously on it... Gretsky\'s current employee Shane Doan (who is currently a Phoenix Coyote) and Mario Lemieux was penciled in as a posibility.  From  urlLink WHC2004.com     Canada\'s roster includes 14 players that won a gold medal at the 2002 Olympic Winter Games in Salt Lake City, six (6) players that played for Team Canada at the 1996 World Cup of Hockey, six (6) players that won a gold medal at the 2004 Men\'s World Hockey Championship in Prague last week, and seven (7) players that won World Championship gold in 2003 in Finland. In total, Canada\'s roster accounts for 686 international games played in either the World Juniors, the Olympics, the World Championship or the Canada Cup/ World Cup of Hockey.  The WCH 2004 rules state that teams will dress 18 skaters and two goalies for each game. The World Cup of Hockey 2004 runs from August 30 - September 14, 2004.   '),(4876,671748,'14,May,2004','2004-05-14 00:00:00','    With thanks to Reuters news service, it must be bittersweet news to all hockey fans living in Toronto, Canada today.    TORONTO-  Billionaire Eugene Melnyk, founder and chief executive of Biovail Corp., Canada\'s largest publicly traded drug company, said on Friday he wants to stop a rival bidder for Toronto\'s Maple Leaf Gardens .   The aforementioned rival bidder was  urlLink Loblaws , Canada\'s biggest retail grocer. They wanted to create a large grocery store with an indoor parking garage. Mr. Melnyk wants to rescue Maple Leaf Gardens and turn the legendary building into a hockey shrine and feature junior teams.  Ottawa gets the last laugh though... Mr. Melnyk was the billionaire who rescued the Ottawa Senators from bankrupsy - Toronto Maple Leafs NHL and hated Providencial rivals.   '),(4877,671748,'13,May,2004','2004-05-13 00:00:00','    With thanks to  urlLink TMLFans.ca , they have passed on the information that  urlLink TSN  reports that legendary NHL goalie and current Maple Leafs president Ken Dryden will represent Toronto and the Liberal Party in the Canadian Parliment.  I guess Canadians have caught on to the trend of putting famous people into politics. Jesse Ventura, Steve Largent, and Arnold Schwarzenegger are just a few of the American names who have or had held public office.  I guess it is all about popularity and name recognition in politics.  On a related Maple Leafs note, the team\'s official site said yesterday  in bold print  that MOJO radio have renewed their contract to air Leafs games for another 7 seasons. However, they had a picture of Brian Leetch next to the write up with the caption \"Brian Leetch will be back next year for the Leafs\". Based on what? Leetch has nothing to do with the MOJO story. At first glance I was even misled to believe that Leetch signed for 7 more seasons.  Just an example of poor journalism on their part.   '),(4878,671748,'12,May,2004','2004-05-12 00:00:00','    Want to increase your teams attendance and season ticket sales?  Then why not follow in Tampa Bay Lightning\'s footsteps and offer free unlimited beer to the season ticket holders.  An AP story from  urlLink Yahoo! Sports  says that the area police and the local chapter of Mothers Against Drunk Driving (M.A.D.D.) are dead set against such an idea and according to the article called it \"irresponsible\".   '),(4879,671748,'11,May,2004','2004-05-11 00:00:00','    For you hockey novices, I hope you don\'t get confused with some of the names and division lineups.  Wachovia is a fianical bank and securities firm. They own Eastern Pennsylvania. They bought up all the First Union naming rights not too long ago  If you see \"Baby Pens defeat the Phantoms at Wachovia Arena\" would you know which team was home and away?  The Flyers play in Wachovia Center The Phantoms play in the Wachovia Spectrum The AHL Penguins play in the Wachovia Arena at Casey Plaza.  Confused yet?    So, I think I\'ve made my distain for Philadelphia teams and their fans quite clear... but is it wrong if I\'m seen in a Philadelphia Phantoms AHL Third Jersey?  I was elated that my Baby Pens have taken it too the Flyers minor league team in what I understand is a very important rivalry. I interviewed Glen Patrick a couple years back on the radio and he addressed the rivalry situation.  Check out the  urlLink Calder Cup Official Site   urlLink Wilkes-Barre Scranton Penguins   urlLink Philadelphia Phantoms  Websites to keep up on what\'s going on in the minors.  In other news, the Rochester Americans have swept their series and await Chicago vs. Milwalkee.   '),(4880,671748,'10,May,2004','2004-05-10 00:00:00','    I got this email from \"the Hockey Mistress\" not too long ago. She wanted me to check out her website dedicated to Hockey Fiction... but with a twist. It\'s Hockey Erotic fiction.  If you\'re bored and need a hockey fix of a diferent kind, I\'ll link over to the website  urlLink Hockey Erotica .  I haven\'t gotten a chance to read some of the featured fiction yet, but I\'ll enjoy reading about women who love hockey and it\'s players.  The author(ess) claims she\'s not a \"puck bunny\" but I\'ll let you judge for yourselves. The bio claims she\'s a So-Cal Gal and got turned on to it via \"at the Pond\" (She\'s a Mighty Ducks fan apparently) Be warned though, the content is intended for an adult audience.  Perhaps I\'ll try and submit my Hockey Fiction short story to her soon. I wonder if she\'ll publish it.   '),(4881,671748,'09,May,2004','2004-05-09 00:00:00','    If once wasn\'t enough, USA and Slovakia lock horns once again in the Czech Republic for the IIHF World Championships of Hockey, but this time it\'s for the Bronze Medal and Third Place in the tournament. Andy Roach scored the games only goal as Team USA came out on top to shut out Team Slovakia 1-0.  Canada won the Tournament\'s top honor by defeating Sweden.   '),(4882,671748,'08,May,2004','2004-05-08 00:00:00','    Who\'s up for some minor league hockey????  In the shadows of all that is... the American Hockey League (AHL) playoffs are still underway.  Since Pittsburgh doesn\'t have all that much to look forward to now, the Pens fans have turned their attention to the AHL and their Wilkes-Barre Scranton Penguins. I got a chance to hear them live as  FOX Sports Radio 970 AM  bent to the will of the people and decided to cover the Conference Finals round between WB Scranton Penguins and their hated rivals the Philadelphia Phantoms.  The \"Baby Pens\" took the first game of the series in a dramatic come from behind win. Philly took game two.  More updates on the AHL to come!   '),(4883,671748,'07,May,2004','2004-05-07 00:00:00','    I just want to thank all my readers at this Confessions of  a Hockey Fanatic website. I have had a great time reserching all the news stories and giving them my own commentarial twist.  Unlike some of the Blogs that have made it linked to this page, I plan to continue blogging all about hockey 24/7/365.  There is still a whole lot to talk about even with the World Championships over in the Czech Republic and the AHL Calder Cup playoffs. Soon, the NHL Entry Draft is coming up in Mid-June in Carolina. (Pittsburghers are still pissed about not getting the #1 pick.)  Keep up on the coolest game on earth right here - Hockey Rules!   '),(4884,671748,'03,May,2004','2004-05-03 00:00:00','    Time for the Conference Finals - winning now is only one step closer to playing for Lord Stanley\'s Cup.   Western Conference   San Jose v. Calgary  Calgary hoists all Canada\'s pride by being the remaining Canadian franchise in the playoffs. San Jose and their \"no-namers\" have made it this far in a gritty system well-coached by Ron Wilson. Calgary is hoping to beat the system with upcoming stars such as team Captain Jarome Iginla and  Robyn Regehr with former Sharks goalie Kiprusoff.  Prediction: Sharks in 6    Eastern Conference   Tampa Bay v. Philadelphia  Much to my disappointment, Philadelphia is in the Conference finals. I can\'t stand this team. Though, they seem to be building a line chemistry with the \"ex-Chicago\" line Zhamnov, Roenick and Amonte. Tampa Bay are riding high on the \"safe is death\" philosophy. Martin St. Louis can clinch the Hart Trophy (NHL MVP) by stepping up and sending the Flyers home packing Prediction: Lightning in 6   '),(4885,671748,'03,May,2004','2004-05-03 00:00:00','    Na na na na... na na na na.. hey hey hey.. goodbye!  The Detroit Red Wings were eliminated tonight by the Calgary Flames to advance to the Conference finals.  Canadian fans who have been getting upset that our \"Yankee\" streak of winning the Stanley Cup and keeping it to American soil can breathe a sigh of relief. Calgary still has yet to face the winner of the Avalanche/Sharks series. They also have the ace in the hole, riding their success on the back of Sharks castaway Miikka Kiprusoff. \"Hey guys! Remember me?\"  Farewell to Steve Yzerman,  the \"Rock of Gibralter\", on the aging Detroit squad. He was hurt in a previous game in the Flames series ending his season. With the CBA/Strike making things hazy for a future season, it\'s speculative that Yzerman will retire.   '),(4886,671748,'30,June,2004','2004-06-30 00:00:00','    Either in a serious change of strategy with the NY Rangers or simply a choice to go separate ways, Eric Lindros has been given his walking papers so to speak according to this  urlLink Yahoo! Sports  article.  After suffering yet another concussion from Washington Capitals Jason Doig earlier in the year, Lindros\' career was again put on hold. The debate now is trying to determine if it\'s even worth signing with another team to continue his career and risk death.  In the past, Lindros had expressed interest in playing for his hometown Maple Leafs team. I hope that doesn\'t happen or it will be hard for me to root for Toronto.   '),(4887,671748,'26,June,2004','2004-06-26 00:00:00','    Today in the RBC Center in Raleigh, NC. The NHL held their annual entry draft. I have just one word to say regarding the coverage and the events that took place.  Booorrrrring!  I was attenative for the first 3 draft selections as the Capitals (who got lucky to get the #1 position) selected the much-hyped Alexander Ovechkin and Pittsburgh Penguins had to be satisfied with Evgeni Malkin.  So, despite rumors of teams trading up, things went along as they were \"supposed to\". No shocking deals and no unknown draft picks were taken in the first round.      Some  COHF  Draft notables :    My \"Insider\" to the Capitals was on the tv screen a total of 30 seconds. The joint broadcast of Canadas TSN and USA\'s ESPN 2 folowed around the tables on the arena floor, but failed to capture any real drama.  History was made when two Russian-born players were selected #1 and #2 overall.  I like what I hear from the Penguins late round Picks. Dave Brown a Canadian Goalie from the University of Notre Dame and Chris Peluso a Minnesota native defenceman could be major contributers.  Christopher Bourque, son of Hall of Famer Ray Bourque, was selected in the second round by Washington.  Standing as giants among men, two blue chip Slovak players Boris Valabik (6ft. 7in. 212lbs.) was selected #10 by the Thrashers and Andrej Meszaros (6ft. 198lbs.) was selected #23 by the Senators.  Robbie Schremp, a London Knights OHL player, was selected by the Oilers, but scouts and pundits say that Schremp believes his own hype and may cause problems on the bench.  Ted Leonsis is  still  fat and sweaty.   '),(4888,671748,'24,June,2004','2004-06-24 00:00:00','    When a professional athlete does well, the agent who negotiated his/her contract should be able to reap some of the benefits, right?  Well, according to this  urlLink Yahoo! Sports  article, Slovak hockey star and current Minnesota Wild player, Marian Gaborik is being sued by his former agents who encouraged the young and impressionable star to hold out this past season.  Gaborik made bonuses after making the All-star team in 2003 and now the agents who arranged that contract want him to pay them an additional $275,000.   '),(4889,671748,'24,June,2004','2004-06-24 00:00:00','    The speculative news report on  urlLink Yahoo! Sports , tells the tale that former Penguins and Flyers winger Rick Tocchet may look to move on from his assistant coaching position.  The article stated that he had \"differences\" with head coach Tony Granato who was his teammate in Los Angeles at one time.  The bulk of the speculation was that Tocchet may want to get the Phoenix Coyotes job since having a good friendship with Coyotes owner Wayne Gretzky.   '),(4890,671748,'22,June,2004','2004-06-22 00:00:00','    I got up this morning with a feeling of \"woa, what just happened\", as the dreams that I had last night were still fresh in my mind.  I could remember the first part, I was high up in a generic arena seats, perhaps even looking down from the broadcast section. I was at a hockey game. Interestingly enough, neither team was my hometown favorite. I understood later that I was at a Buffalo vs. Washington game. I remember being asked by some unidentified man about which team I was rooting for. \"Washington, I guess, because I know somebody who works with them,\" I replied. The game\'s action was fairly decent, but later I would get a first hand look.  The second part to the dream included me on the ice, in full uniform, ready to line up for the face off. It wasn\'t the same game, but I remember feeling like I was wearing the red logo of the Carolina Hurricanes. In typical fashion, I could remember skating well, but when it came time for me to take the pass for a shot at the net, my stick was out of position. The other team went the other way with the puck and scored. Again, I pretended to know what I was doing and lined up wherever for the next faceoff. Again, I kept up with the play, but the pass to me went wide. The other team scored again. I remember skating off the ice with my head low. Unfortunately, the first line took the ice dissapointed as I left already at a -2.  It\'s not the first time that things haven\'t quite worked out the way I wanted them to in my dreams. I wrote about  urlLink another dream  here where I didn\'t quite fit in.   Oh well, I\'m sure that Sigmund Freud would have a field day trying to analyze those dreams.   '),(4891,671748,'18,June,2004','2004-06-18 00:00:00','    While surfing the web this afternoon, I came upon a article on the  urlLink World Wrestling Entertainment  website that breaks from their own storyline and interview several of their well-known Canadian stars and ask them about how important it was for Calgary to be in the Stanley Cup finals this year. Lance Storm, a Calgary native, shared some of the feelings that Canada had for their represenative team.  \"You just canâ€™t go anywhere. Every business has â€˜Go Flames Goâ€™ up on their billboards and their signs. â€¦ At my daughterâ€™s school, there are several classes that are completely done up in red.\"  Of course, \"m\'girl\" Trish Stratus is a hockey fan and showed her support of the Flames, despite being from the Toronto, Ontario providence and seen in ads modeling the Leafs jersey.  \"Itâ€™s just a national pride thing,\" Trish Stratus said. \"Hockey is Canadaâ€™s game. To bring the Cup home to Canada is where we feel it rightfully belongs.\" Stratus, born and bred in Toronto and a fan of the cityâ€™s Maple Leafs, was rooting for the Leafs. Like many (or most) Canadians, she jumped on the Flamesâ€™ bandwagon once her team was eliminated from the playoffs. \"But as a Canadian Iâ€™m allowed to,\" she said.  Also, Christian, another Toronto native brought up this interesting point to the article linked above.  Christian adds, however, that the Flames vs. Lightning series wasnâ€™t as simple as \'U.S. vs. Canada\'. \"Itâ€™s funny because there are more Canadians on Tampa Bay than there are on Calgary,\" he said.   '),(4892,671748,'16,June,2004','2004-06-16 00:00:00','    One of the hardest working goaltenders in the NHL, Nikolai Khabibulin, has decided not to attend the World Cup games this year as reported in the  urlLink The Hockey News  and seen in other sites such as  urlLink National Hockey League Official Site . This would lead you to believe that he needs time off from winning the Stanley Cup in 7 games and needs more time to prepare. However, TSN\'s  urlLink NHL Network  has a more in-depth write up as to why the \"Boulin Wall\" decided to no-show.   \"We have no chance of winning based on the current situation surrounding the preparation of our national team,\" the Tampa Bay Lightning goaltender told Sport-Express. \"I have already had a similar experience before the 1996 World Cup when there was total chaos in our organization.\"   He\'s right. The Russians have really fallen from hockey graces in international play. They were used to dominating everybody in the 70\'s and 80\'s, but since communism fell, the program hasn\'t gotten over the slump with the surgence of Czechoslovakian, Swedes, Fins, and even Americans. The furthest I remember them finishing was placing 3rd at the 2002 World Cup Tournament.  Russia was also hoping that San Jose Sharks goalie Evgeni Nabokov would cary the team, but is recovering from recent knee surgery this spring. That leaves Russia with Sokolov, a Russian league player, in charge for the World Cup games.   '),(4893,671748,'14,June,2004','2004-06-14 00:00:00','    When a local guy from South Western Pennsylvania makes it big, the fans will always support and owe loyalty to one of their own. However, should someone turn their back, then they are most certainly unfogiving and the \"star\" is reduced to \"sellout\" status.  Enter: R.J. Umberger.  As a Plum Area native (suburb of Pittsburgh), he decided to pursue hockey at the collegiate level at  Thee  Ohio State University. That should have been the first warning sign to his loyalty when everybody rivals the Buckeyes. At least the Patrick family had the decency to play for national champion Penn State for less national exposure. However, he was a hero to the area having been the first local kid to make the NHL from the \"Mario era\" which created interest in Pittsburgh area children to play the game. He was drafted by Vancouver, but when he was unable to come to a contract, they traded him to NY Rangers.  Follow the money.  Umberger signs with Philly according to this  urlLink NHL Network  news release. Umberger chose his wallet over loyalty and proved to a lot of NHL and Pittsburgh fans that he is only in it for the money.  Also in the news, the Flyers also signed Keith Primeau to a $17 million 4 year deal.   '),(4894,671748,'09,June,2004','2004-06-09 00:00:00','    As reported in  urlLink The Hockey News  the American Hockey League as passed down their edict on who should and will be eligible if an NHL lockout occurs.  The remaining players who haven\'t signed contracts to Europe leagues who wish to play through an NHL lockout will have to reconsider. The AHL will grant status to those players with 50 NHL games or more.  That screws a lot of players like Marc Andre Fleury who has only 1 or 2 games to his AHL credit.  I say that all those who want to stay in the states and wait out the inevitable lockout, they should sign temporary contracts to ECHL and UHL teams. That way, they could expand interest in those leagues, drive up ticket sales in those markets, and improve the overall exposure that the minor leagues should get. It will only benefit the future of hockey when the league has a better minor league system to work with, that way, they can develop more star-calibur players.   '),(4895,671748,'09,June,2004','2004-06-09 00:00:00','    Okay, I don\'t realy feel like I need to say a lot about this issue, but it\'s been known for quite some time that people and investors wanted to revive the WHA of the old days. Much like the NBA had it\'s rival ABA and the NFL had it\'s USFL, hockey fans will soon be given an alternative to the NHL. TSN\'s  urlLink NHL Network  has an article about it. I\'ll outline some more of the finer or more interesting points here.    WHA Commisioner is NHL legend Bobby Hull   Entry Draft will be open to players 17 years of age   Halifax, Quebec City, Detroit, Dallas, Orlando and Jacksonville are the original 6 teams. Toronto and Hamilton to soon join   $15 million dollar salary cap    Rule Changes :  no red-line, no-touch icing, touch-up offside and three-on-three overtime followed by a shootout.    Halifax (Metro Centre), Quebec City (Le Colisee Pepsi), Detroit (Pontiac Silverdome), Dallas (Reunion Arena), Orlando (TD Waterhouse Center) Jacksonville, Fla. (Memorial Arena), Toronto (SkyDome) and Hamilton (Copps Coliseum) are all old hockey arenas expected to hold the WHA games.    The NHL had better get to signing that CBA soon or the WHA will have a leg on them and steal away all their good players. They have their eyes on phenom Sidney Crosby and hope to \"steal\" him away before he turns 18 from the NHL. Is that Salary Cap in Canadian, American or Australian Dollars?   '),(4896,671748,'08,June,2004','2004-06-08 00:00:00','    With thanks to  HockeyNation  for the story, most Tampa residents and hockey fans are wondering why their NHL championship team does not get any respect. What makes it worse is that their own hometown  urlLink Tampa Tribune  newspaper didn\'t give them a chance to win.  Apparently, the early morning edition to todays paper will consist of a  Mea Culpa  (see: \"my bad\") to the Florida citizens for preparing their Sports colums as if the Flames had won.  Don\'t believe me? Canada was also almost duped in the hopeful tease.  urlLink Toronto Star  has an AP article.  Listen fans, if you just want to get the correct hockey news first, keep it right here at  Confessions of a Hockey Fanatic !   '),(4897,671748,'08,June,2004','2004-06-08 00:00:00','    According to a TSN\'s  urlLink NHL Network article,  Bourque and Coffey might as well mark Nov. 8 in their datebooks now because that\'s when the formal ceremonies will take place.  Yup, sometimes news like this makes me feel old. When players that I grew up watching have all retired and qualify for the Hall of Fame, then it\'s time to say \"where has the time gone\". Definitely Bourque and Coffey belong in the Hall of Fame. They redefined their blueliner (see: Defenceman) position. Not only were they solid on the even strength battles, but they showed their skill by shooting and passing the puck on the powerplay. Both players racked up points, especially Coffey who collected them like Cups. Coffey was an integral part to the Edmonton Oilers and Pittsburgh Penguin dynasties.   '),(4898,671748,'07,June,2004','2004-06-07 00:00:00','    Congratulations to Tampa Bay and NHL veteran Dave Andreychuk for winning the Stanley Cup, the NHL\'s historic championship trophy. It was a gritty first period and both teams were trying their best to beat one another up. Calgary got caught with a couple of boarding calls leading to a Ruslan Fedotenko Power Play Goal (PPG). In the end, \"the Boulin Wall\" stood taller than the Finnish phenom Kiprusoff. Both teams goalies stood on their heads and made great saves. Tampa finished the game with a 2-1 win. The best moment of the night was Andreychuk who got to finally hold the Cup over his head after a long NHL career. They featured him several times during the game as well as the first period intermission. The St. Pete Times Forum in Tampa was filled along with a large gathering outside the gates watching the coverage on big screen.   '),(4899,671748,'06,June,2004','2004-06-06 00:00:00','    My Penguins couldn\'t quite do it tonight and got blown out despite a sold out white-wearing Whachovia Arena at Casey Plaza. Milwaulkee Admirals are the Calder Cup champions of the American Hockey League. Their veteran-enhanced roster blew out the Rochester \"Amerks\" and took care of Scranton in timely fashion, winning that series 4-0. The Admirals are the third IHL team to win the Calder Cup since the IHL disbanded and the AHL welcomed the additional teams with open arms. (Chicago and Houston are the other two teams)  I forgot to keep you all updated on the ECHL - the \"AA\" of minor league hockey.  Idaho Steelheads are this year\'s Kelly Cup champs defeating the Florida Everblades on May 22nd. Another hockey team from Florida? Wow. Hockey has really grown into a North American sport, reaching for fans from all over.   '),(4900,671748,'05,June,2004','2004-06-05 00:00:00','    With thanks to \"Punkette\" a member on the Pittsburgh Penguins message board, she translated an article on  urlLink RDS  (an ESPN-type French Network)   The Quebec Junior League (LHJMQ) will become the only junior league in Canada to use a drug program which will be implemented next season.  The league still has to produce a list of illegal substances but said it will follow the Olympic International Comitee as soon as they go public with theirs list.  Players will have to submit to random tests which will help to prevent this problem.  1st infraction: 5-10 games  2nd : 25 games  3rd : 2 years  The suspensions will have their causes published in public.   '),(4901,671748,'03,June,2004','2004-06-03 00:00:00','    The Collective Bargaining Agreement is starting to scare a lot of the European and Veteran NHL players that  urlLink The Hockey News  reports that the latest player to sign a contract with a European league is Tommy Salo. The Swede, formerly of the Edmonton Oilers, He signed a 10 year deal with Modo of the Swedish Elite League.  10 years, eh? I guess he\'s not comming back.   '),(4902,671748,'30,July,2004','2004-07-30 00:00:00','    Dave Andreychuk captain of the Tampa Bay Lightning returned to his native Hamilton, Ontario, Canada with the one prize that had aluded him for 25 years - The Stanley Cup.  He was greeted as a hero in Mountain Arena according to this  urlLink TSN article  and showered with a lot of gifts.  A jersey that he first wore when he was a little leaguer, a framed photo of him first holding the Stanley Cup, a parade, fireworks display, an honorary captains jersey to the AHL Hamilton Bulldogs, and the really impressive item... the arena became The Dave Andreychuk Mountain Arena.  I suppose Hamiltinians will do whatever it took to get a NHL team in the city. I\'m sure they also may have alterior motives for honoring their city\'s favorite son. They put on a grand show to make it look that Hamilton loves hockey and will support a pro team there. This may be true, but the CBA agreement has to come first before any of those questions are answered.   '),(4903,671748,'26,July,2004','2004-07-26 00:00:00','    Okay.. this is getting rediculous.    urlLink Lindros says he wants to sign with a Stanley Cup contender    Eric, did your daddy ever tell you that we don\'t always get what we want?   '),(4904,671748,'24,July,2004','2004-07-24 00:00:00','    Today, I played with some of my fellow Pittsburgh Penguins fans in a roller hockey game at  urlLink Bladerunners  in Harmarville, PA. I had a great time with memories that will last a lifetime.  One of the things in my life that really I wish I had more of, it\'s friends who enjoy the kind of things I do. I used to skate with my brother and some friends in the neighborhood, but that only lasted one summer back in High School. I played some in college, but that was only a couple of times too.  As it turned out,  urlLink Bladerunners  was behind this grocery store that I drove by all the time on the way to my brother\'s current apartment.  The game itself was awesome. We had a good turn out at Bladerunners. But when I got to the door, I was surprised at the idea that it cost money to play. Okay, so I was out $10.  I gave it my best shot. However, there were people there with better speed and skill than I, so after about an hour. Nobody would pass me the puck. So, I did the \"sensible\" thing and volunteered myself for goaltender. (For some reason, the other goaltender couldn\'t make it. So, without any pads, I did my best stopping shot after shot.) The Message Board said I was crazy others said I was courageous, but I didn\'t care what they thought.  Afterwards, we all went to the local  urlLink Primanti\'s  for a bite.   '),(4905,671748,'20,July,2004','2004-07-20 00:00:00','    The cable channel  urlLink E! Network  is no stranger to T&A television, so this new TV show  urlLink Scream Play  should be no surprise as&nbsp;what to expect. This obvious \"Fear Factor\" rip-off, suddenly got my attention as the opening stunt featured a tie-in to a hockey movie.   One of the competitors named the 1977 movie \"Slapshot\" as his favorite&nbsp;movie of all time. So, the makers of the Scream Play program&nbsp;thought it wise to recreate the scene where a player in the movie strips off his uniform to&nbsp;help his teammates score a goal during the distraction. Scream Play thought they could&nbsp;close off a rink and shoot a stunt where&nbsp;3 teams of 2 players played broom-ball in the buff! The winner would spend a weeks vacation in St. Thomas, Virgin Islands.   That\'s right. Naked Hockey.   The female to male ratio was 4-2, so you know it had my attention. The game started at center ice, the two players pulled off their jerseys and proceeded to knock their balls into the goal as quickly as possible. Standing on street shoes and not on skates, the danger was hoping that no one fell hurting a boob or something.  The host did the play-by-play and obviously did not see the movie. He tried citing \"Oggie\'s\" name a few times just to get the tie in along with any other random name/movie reference he could. He was a tool.   I don\'t know if I would\'ve done it, I don\'t have the kind of body to be an exhibitionist like those players on TV. From what I could tell through the pixilation blur, I liked the one contestant\'s back tattoo.   '),(4906,671748,'18,July,2004','2004-07-18 00:00:00','    The rogue WHA hockey league held their draft in a casino gambling on which NHL Free Agents that the league can possibly persuade to sign.   Simon Gagne was the first free agent selected by the Quebec Nordiks.&nbsp; Followed by Trent Green of the Boston Bruins.  Then on Sunday, the WHA decided that it could probably lure some young stars to play in the league at age 17. Of Course, Sidney Crosby was selected first.  I don\'t know how sucessful their draft is going to be. They didn\'t ask anyone if they actually wanted to join, they just took turns randomly selecting names.  This reminds me when the RHI (Roller Hockey International) started. It was a real joke. They even drafted people who weren\'t even hockey players. I think Tonya Harding was drafted.  I don\'t really have too much to say about this, so go check up on some of my friends websites and see what they have to say about the WHA.   '),(4907,671748,'14,July,2004','2004-07-14 00:00:00','    It looks as if another NHL veteran has decided to call it a career. According to  urlLink Yahoo! Sports , Ron Francis was offered a front office job with the franchise that he has known the most in his 20+ years.   Ronnie \"Franchise\" was originally drafted by the Hartford Whalers, but after 9 seasons, he was traded to the Penguins to win the Stanley Cup. He also brought the Whalers to the Finals in their current Carolina Hurricanes form. Last season, he made a run for the Stanley Cup as a Toronto Maple Leaf. Originally reported in the  urlLink Raleigh News & Observer  newspaper, Ronnie \"Franchise\" will be a welcome relief to the franchise he\'s helped build.   '),(4908,671748,'11,July,2004','2004-07-11 00:00:00','    In light of a NHL lockout season, the National Hockey League has made public their 2004-05 schedule.   Begining with October 13th, the NHL drops the puck on it\'s 88th season. The Tampa Bay Lightning raises the Stanley Cup banner (plus a Southeast Division and Eastern Conference banner) on that day with a game against the Philadelphia Flyers. The Lightning and Flames won\'t get a rematch until Nov. 11th.  Some other dates most notable on the schedule will be the Hall of Fame game featuring the Leafs and Bruins on Nov. 6th, the All-Star Game in Atlanta on February 12th, and Tampa Bay will also take part in the leagues only neutral-site game against Buffalo on Oct 18th in Rochester, NY.  Hockey Day in Canada, which is considered a national holiday, will be covered by CBC on Feb 19th with Ottawa at Montreal, Edmonton at Toronto and Calgary at Vancouver.  My Pittsburgh Penguins start the year on the road against the NY Rangers on October 15th and come home to play the Rangers again on the 16th.  From a Yahoo! Sports article, I find this most newsworthy  For the first time in NHL history, all 30 teams will be in action on March 26. A lineup of 14 games has taken place nine times, most recently on November 1, 2003.   '),(4909,671748,'11,July,2004','2004-07-11 00:00:00','    The American Hockey League, the minors to the NHL, have elected to make a few changes to the way the game is played for next season. In this article at the  urlLink Pittsburgh Post Gazette  it underlines a few things being changed for the upcomming 2004-05 AHL season. If all goes well, they will be suggested to the NHL and adopted there as rule changes to their games.     Adopting a no-touch icing rule as is used in international play. Play will be blown dead as soon as the puck crosses the goal line.  Using a shootout to decide regular-season games tied after a five-minute overtime.  Returning to the \"tag-up\" offside rules the league used from 1986-96. Players who are caught in the offensive zone in a delayed-offside situation can negate the offside by \"tagging up\" with the blue line.  Increasing the size of the neutral zone by moving nets from 13 feet in front of the end boards to 11 and expanding the width of the blue lines and center-ice red line from 12 to 24 inches. Passes from the defensive edge of one blue line to the offensive edge of the other will be legal.  Also, for the first seven weeks of the 2004-05 season, the AHL will experiment with rules that restrict where goalies are allowed to play the puck.    I don\'t see why any of these drastic rule changes are going to solve much of anything in the NHL. The game is fine the way it is. Experts argue that the rule changes, especially the one about \"no-touch\" icing would make the game flow faster. The icing rule won\'t stop anything. The players will only ice the puck as much as they did. Only this time, there won\'t be as much time lost for a player to get to the puck. It won\'t do a darn thing in strategy. It\'s also bad on the goalies who can move the puck well. If you restrict their play, there will be even more stoppages. Errors do occur with goalies who mis-handle the puck and when that happens, usually a good team can take advantage of that and score. The overtime rules is just REALY dumb. I don\'t know of anyone who will be willing enough to sit around even longer to see a game end, most \"fair weather\" fans are already making their way to the parking lots at the end of regulation.   '),(4910,671748,'09,July,2004','2004-07-09 00:00:00','    In the summer off season months, the NHL owners and gemneral managers assess their standings while going over their financials and player needs. But this year is unique as both players and owners don\'t know what is in store for the future of the league.  This summer has brought on some surprising signings already since the free agency season started on July 1st.  Mark Recchi has returned to the team that drafted him today as the Penguins have signed him to a deal with the terms of the deal not disclosed. The theory around town here is that the Penguins and Recchi don\'t want to affect any other NHL signings in the market by \"comparison shopping\". Thus keeping it on \"the down low\".  Gretzky\'s Coyotes have been throwing around some money signing 2 NHL veterans so far. Red Wings\' Devereaux and the Sharks Ricci have signed deals to play in Glendale, AZ. Two good pick ups for them... a team that still doesn\'t have a coach.  Boston Bruins captain Joe Thorton decided to go to Europe during what he expects to be a long lock out. There is a clause in the Swiss team contract that states he can go back to the NHL once the CBA is resolved.   '),(4911,671748,'04,July,2004','2004-07-04 00:00:00','    God Bless America.   '),(4912,671748,'03,July,2004','2004-07-03 00:00:00','    With a humongus three patriotic cheers for Debbye living and working in Toronto, ON go check out her blog. I highly endorse it.  urlLink Being American in T.O. Trackback  The link shown takes you to a post she made about the ongoing saga of controversial coach turned broadcaster Don Cherry of the CBC. The Editorial in the  Toronto Sun  suggests that Coach Cherry would have to undergo some sort of sensitivity training to understand why the Bloc Quebeccois is so super-sensitive and hyper-offendedâ„¢. It goes on to say that an independand audit supports that hockey players who wear face shields are \"europeans and french guys\". I just can\'t understand what the heck is so controversial about this statement. Coach Cherry is now proven to be stating a fact and people have to get all huffy-and-puffy about it. Get over yourselves, Quebec.   '),(4913,671748,'02,July,2004','2004-07-02 00:00:00','    Okay. I\'m glad that this is actualy following through. I\'d hate for pro athletes, especially one as careless as Dany Heatly was, to go free when he was responsible for the unfortunate death of a teammate. In this  urlLink Yahoo! Sports  article link, the Atlanta athorities are planning to move ahead with the investigation and file charges against the one time rookie of the year.  The most serious charge Heatley could face is vehicular homicide, which carries a maximum sentence of 15 years in prison.  I think that\'s a real sobering quote in the article.   '),(4914,671748,'01,July,2004','2004-07-01 00:00:00','    Back in 2002 and inspired by a surge of interest in Canadian culture, I had decided to spend July 1st simulating hockey games on my Playstation. This year, the tradition continues! Like last year, I chose to simulate a national style tournament instead of a \"one and done\" format like 2002. I set up two pools, eastern and western, with the 6 remaining NHL Canadian teams and the 2 transplanted teams in USA that used to represent Quebec and Winnepeg. There they each played two games a piece to decide the tournament seatings. I also used up to the minute rosters barring free agent or injury status, however, I did remove the players who retired or chose to sign with European teams. This tournament this year was an all Quebec final as the Nordiques met up with their rival Montreal. I suppose it was the additions of Bonk and Kovalev that improved the Canadians this year. It was a strong match up even to the overtime. Steve Begin scored in the extra tie breaking session this year to put Montreal 3-2. Quebec put up a valient effort, though. The leading scorers in the tournament was Joe Sakic (6-6-12) and Ryan Smyth (7-4-12).    Previous Canada Day Champions 2002: Toronto Maple Leafs 2003: Edmonton Oilers   '),(4915,671748,'06,August,2004','2004-08-06 00:00:00','    One of the most highly sought after Hockey Free Agents have landed in the desert. Brett Hull signed a 2 year deal with the Phoenix Coyotes according to this  urlLink Yahoo! Sports  article. The money part was not released. I suppose that having Gretzky for an owner is starting to pay off. The Coyotes have really been active this season acquiring quite oa few. Brett is no stranger to the \"Great One\" I\'m sure that influence was the deal breaker. The Penguins and the Stars were also interested in the Free Agent Hull. Brett is also friends with Mario Lemieux... and has said he would\'ve liked to play with/for Mario sometime in his career.   '),(4916,671748,'04,August,2004','2004-08-04 00:00:00','    What could they possibly be waiting for?  Today, the NHL and the NHLPA met in Toronto, ON, Canada and according to all the major sports sources (and several other fellow bloggers). The two sides failed to reach any progress on the Collective Bargaining Agreement.  This is really starting to make me worry. My best guess is that they will stretch out the talks until the 11th hour and sign anything to get the season in on time.  NHL and NHLPA did however agree to meet again on August 17th.   '),(4917,671748,'04,August,2004','2004-08-04 00:00:00','    Well, I was shocked and dissapointed to read on my hometown team\'s website that the Pittsburgh Penguins are going to create a new squad called  urlLink \"Penguins Patrol.\"   They are calling for young women (at least 19 years of age) to walk around the Mellon Arena and do the PR work with the fans. It\'s a lot like the \"Spirit Squad\" that they had this year, but not with the costumes or dancing.  Of course, I could take this post and start talking about the cheerleading in hockey, but I\'d skewer off to a tangent and talk about my ax to grind with an evil Canadian website.. I\'m not in the mood to right now though.  Feel free to check out the link above and give me your feedback about hockey maybe needing cheerleaders.   '),(4918,3869854,'10,July,2004','2004-07-10 00:00:00','   The  urlLink CGMC  show opened last night, and despite a very modest crowd, I think it went very well.   This was the first show that I had people from work come to see-- and this wasn\'t just a random co-worker or someone from another department, either.  One of my direct managers came and so did THE boss, the Chief Marketing Officer (as her title reads) and her daughter and husband.  So yeah, it better NOT have sucked.  They loved it, though-- raves all around.  I was beaming.   I\'m disappointed by the ticket sales, though.  We sold out nearly all of our Holiday shows for this season, which was because of a combination of a huge push in marketing, a surge in season subscription tickets and the Chorus talking up the show.  We do three shows a season... the Holiday one, one in the spring (usually in April) and one during Pride.    We switched the dates for our Pride show this year, however, and I think that hurt sales.  Our reasons were valid-- there are a MILLION things to do during Pride weekend, and our show was always one of them.  But I think that\'s what drives sales in the first place.  One of the things people do during Pride weekend is see the Chorus.  And when we weren\'t there to fill that need, they found something else.  Now it\'s TWO weekends later, and we\'re still singing about Pride.  Call me crazy but I\'m Prided out.  I\'m sure others felt the same, hence the dropoff in sales.  Rumor has it we will go back to \"normal\" next year.  I hope so!!  It really is a fun show, with some gorgeous singing.     urlLink    Here are those German boys.  Aye Aye!&nbsp; urlLink     The German set is adorable.  They aren\'t always together and you can\'t always understand their English so well, but you have to admire their gusto and their love for staging.  And the pirate costumes are pretty cute too... and flexible.  One song struck me as hilarious-- they are wearing these little bandanas on their heads throughout most of the set, tied in the back.  At one point they sing a song about swabbing the deck, and they stop, turn the bandanas around to mimic \"Aunt Jemima\" or something along those lines, and it is simply hilarious.  There\'s much more but it\'s just worth saying GO SEE IT and find out.   urlLink Windy City  sounds fabulous of course, and their love for musical perfection is evident in their set.  I also sing with the  urlLink Windy City  organization, however not this particular chorus, so I performed with them a couple weeks ago for their concert.     urlLink    The boys from Windy City&nbsp; urlLink     Their musical director is taking them in a good direction, I think.  Highlights are a sampling from Rutter\'s \"Requiem\" and a nautical set that is highlighted with \"What Do You Do With A Drunken Sailor\".  I can\'t help but think that if CGMC did that song, we would change it to \"What Do You Do With A Drunken Drag Queen\"... the lyrical possibilities are endless.  Since we have the German boys to be campy and funny and silly,  urlLink CGMC   keeps it a little serious this time around, and frankly I\'m glad.     urlLink    Patrick Sinozich, CGMC\'s Artistic Director&nbsp; urlLink     Patrick Sinozich, our director, has assembled one of the most ambitious and challenging sets for our chorus in a long time, and we are really doing a wonderful job with it.  Highlights are:   Daemon Irrepitus Callidus , a truly demonic piece (whee!),  Arise My Beloved , a lovely verse from Song of Songs that Patrick wrote a lush and gorgeous piece of music to accompany,  Jimmy Jimmy , which is an upbeat ditty based on an old folk tune and tells a story that ends in a gay wedding (how topical!), and my personal favorite,  This Moment As I Sit Alone , the marriage of the third movement of Rachmaninoff\'s Concerto for Piano and Cello (Andante) Op. 19 and a gorgeous poem by Walt Whitman that expresses his love for men and the question that if one can love a man in his own land, he could also love a man in other lands... it is moving, gorgeously arranged (again a Sinozich creation) and lovingly sung.    So again, if you haven\'t got your tickets yet, you have two more chances.  Performances today are at 5:00PM and 8:30PM at the Athenaeum Theatre and are on sale through Ticketmaster and at the Box Office.  Click  urlLink here  for a visit to CGMC\'s website and click on the concert\'s logo for ticket ordering info.  So I get to rest until our call, which is at 4:45.  I think some guys are at the beach today, but the temperature is only 67 degrees at 10AM.  Hm.  Not prime beach weather, but I could use some fresh air.  We had an opening night party at one of our members\' homes last night and I was up far too late and drank one too many.  Ouch.  Why do I not remember to pop two Tylenol before bed?  Silly me.  So I am recovering and hoping this headache goes away soon.  Fresh air could help, no?  Yes.  So I may have to join them.  Whee!  Please leave a comment if you\'re out there reading this!  I have no idea if you are otherwise, and I\'d hate to think I\'m the only one that gets to enjoy my ramblings. :)   '),(4919,3869854,'09,July,2004','2004-07-09 00:00:00','   Last night I posted my old Journal to my Geocities web account for your viewing pleasure.  You can find the links at the sidebar to the left.    I was nice enough to make them look pretty too... I had some god-awful background image on them before and they were UUUUGLY!  But hey, that\'s how web pages looked back in 1997. ;)  Reading those old posts, which were written starting around the time I got the job that moved me to Chicago, I\'m reminded how \"Mary Richards-ish\" (from \"The Mary Tyler Moore Show\") I was back then.  So idealistic, so hopeful, so giddy... I was moving to the big city, just getting my start in \"adult\" life.  Who knew what would lie ahead?  I could only wait and see.   Seven years later, a lot has changed.  And a lot more will change in the future.   '),(4920,3869854,'08,July,2004','2004-07-08 00:00:00','   Wow...  My first few postings are really... ranty!  I guess I really needed to vent a few emotions here and there.  Now that I\'ve got that out of my system, maybe I can do more fun stuff. :)    Oh no!  I smileyed!  I told myself I wouldn\'t do that here, but dammit, I couldn\'t help it!  Please, someone, make me stop before I go too far with the smileys!!!  Well it\'s dress rehearsal tonight with the Chorus.  We sounded damn good last night, but we can sound even better.  I really can\'t wait until we do this show.  It\'s going to be tremendous.  Gotta run.  Call is at 6:30 and it\'s almost 5:30.  Driving from downtown to the Athenaeum is always a joy.  Whee!  Oh and wish me luck on meeting a cute German boy.  Whoo hoo! ;-)    Aaaaugh! :)  Oh well. :-P   '),(4921,3869854,'08,July,2004','2004-07-08 00:00:00','   BREAKING NEWS!   urlLink Ridge Warns of \'Credible\' al-Qaida Plot   Oh great, more terror news.  It seems to me that every time the Department of Homeland Security issues a bulletin about possible al Qaida plots or strikes, the only ones getting all worked up is the Department of Homeland Security.  Aren\'t these the same people that told us, way back when, that we should just go about or daily lives and not let ourselves get too worried about terror plots and the like?  This increased paranoia is neither healthy nor productive.    I have never seen an administration manufacture more hysteria than this one.  They raise the terror level to ORANGE if someone\'s shoes don\'t fit right.  They lower it back to YELLOW when they drop a few pounds.  Everything\'s a crisis.  Everything is a potential terrorist plot.   Sure, I am worried about another strike.  I think it\'s wholly possible.  And it\'s our own fault.  Instead of dealing with what happened on September 11 the right way, we have set ourselves up for strikes from more sides than we could have possibly imagined.  We\'ve ruffled feathers and angered nations.  We have gone on the offensive where we had no reason to.  We\'ve become even more of a bully than ever before, and that, along with our excessive lifestyles and abuse of power and money, is what got us into trouble in the first place.  Whatever happened to humility?  Are we destined to be, now and forever, the nation that slaps the hand instead of the nation that holds the hand?    Yeah, I\'m scared.  I\'m scared of the life that the children of my friends and family will have to lead.  I\'m scared of the decisions this administration is making.  If I can be at all hopeful, it is that the next administration will be able to quell some of my fears.    But that doesn\'t stop the fact that damage has already been done.  Things will never be the same.  I\'ve accepted that as truth, but I don\'t like it.   '),(4922,3869854,'08,July,2004','2004-07-08 00:00:00','   59 Degrees.  Is this really July?  I can\'t get over how chilly it\'s been. Where\'s the heat?  Where\'s the humidity?  Where\'s the shirtless men running around everywhere?  I need a vacation.  Oh yeah, that\'s right-- I\'m taking one.  OK granted it\'s in  urlLink Montreal, Canada , where who knows how cold it will be (I\'m no genius at Geography or atmospheric changes in temperature) but it sure will be nice to get away from HERE for a while.  Plus, a nice long break from work would really be nice as well.  I haven\'t taken too many vacations in the time I\'ve been at  urlLink Jenner & Block  (3 years this past April.)  This is what we call \"Hell Week\" in the  urlLink Chorus .  They ain\'t lying.  By the way, have you got your tickets yet?  Off to shower.   '),(4923,3869854,'08,July,2004','2004-07-08 00:00:00','   Well here we go.  Thanks to my friend  urlLink Jake , I am off and blogging away!  (Thanks, Jake!)  I trust that if you\'re reading this, you read my profile and all that, so I won\'t go into the specifics of who I am, what I do, etc.  But if you haven\'t, take a gander at my  urlLink profile  and you\'ll get all that info.  Actually, this isn\'t the first time I\'ve done something like this.  I had a \"blog\" of sorts when I first moved to Chicago in 1997.    It\'s actually funny, because I recently just found all the old HTML files from that old website I used to have online.  To think I actually sat down and created all the layouts and designed logos and everything.  Could I do that again now?  Probably.  Do I want to?  Hell no.  Too much work.  This is much easier.  All the designs and everything are already done.  All I have to do is type, and frankly, I think I do that pretty well.    Now that I\'ve started this, I might just post those old \"Diary\" pages here, just for fun.  Actually, they paint  quite  an interesting picture of how far I\'ve come since I moved to Chicago.  It was 1997.  I was 26 years old, I had moved out of my parent\'s house for the very first time, and I was officially \"on my own.\"  How scary was that?  I look at the things I wrote there and I realize how incredibly naieve and \"green\" I was about a LOT of things.  Thank goodness I\'m jaded and bitter like everyone else in this fair city.  OK, maybe I\'m not quite that bad, but hey, I  am  still single after 7 years... Hmm.  What is the point of a Blog anyway?  And who decided it should be called Blog?  As if chopping the letters \"W\" and \"E\" and smashing the rest of the letters together would make it somehow \"hip\" or \"cool.\"  (I love \"Quotes\" can\'t you \"tell?\")  I suppose it\'s a great way of getting to know how someone thinks and feels-- and for that matter, how they write about it!  It\'s also a great way for me to vent my frustrations and feelings about situations that happen in my everyday life.  It essentially is an online journal.    My original Journal was inspired by someone who I consider to be one of the pioneers of the \"Blog\".  His name is Justin Clouse, and he currently lives in San Diego, CA.  Although it seems he isn\'t posting to his diary anymore, he kept it going strong from 1995 until late in 2003.  You can check his site out  urlLink here .  It\'s getting late and I have a long day tomorrow.  I\'m rehearsing with the Chicago Gay Men\'s Chorus for their Pride show this weekend.  I hope you\'ll be there... if not, click  urlLink here  and buy some tickets, will ya?    I\'ll do more posting soon.  See ya then!   '),(4924,3869854,'16,July,2004','2004-07-16 00:00:00','   Well I\'m just waiting for the cab to come and get me.&nbsp;  &nbsp;  I\'m on my way to lovely Montreal!&nbsp; &nbsp;&nbsp; I hardly slept a wink.&nbsp; I spent the whole night packing and cleaning my apartment.&nbsp; Think I\'m nuts?&nbsp; Well, there is nothing nicer than coming home to a spic-n-span apartment.&nbsp; And thanks to my busting my ass, I will have just that!  &nbsp;  My friend Rafael is watching Pippin & Screech while I\'m away.&nbsp; He\'s more of a cat freak than I am, so they\'ll be in good hands.  &nbsp;  So I will see y\'all in 10 days.&nbsp; Whoever you may be!  &nbsp;  Drop me a line (a comment would be nifty) and let me know you\'re out there!  &nbsp;  -Rick   '),(4925,3869854,'15,July,2004','2004-07-15 00:00:00','   Fantastic news... thanks to everyone who called their Congresspeople, Senators and other politicos about the preposterous Constitutional amendment to deny marriage rights to gays and lesbians, the amendment was DEFEATED 48 to 50.    George Bush lost this battle-- big time-- and that should have interesting consequences when election time rolls around.    Of course the Repugnantcans won\'t stop at that.  They\'ll do everything in their power to make sure that people like me and millions of others are denied as many equal rights as possible.    Quoth The Bushman:  \"Activist judges and local officials in some parts of the country are not letting up in their efforts to redefine marriage for the rest of America and neither should defenders of traditional marriage flag in their efforts,\"   \"It is important for our country to continue the debate on this important issue, and I urge the House of Representatives to pass this amendment,\"    Seems the only one who is making this such an \"important issue\" is the Bushman himself.  And why not?  He has an election to win.  Or so he thinks.  For every step forward there is a step backward.  Expect one soon. But at least the steps forward are GIANT steps, while the steps backward are the tiniest of baby steps.  We will win, but we MUST keep fighting.   '),(4926,3869854,'14,July,2004','2004-07-14 00:00:00','   Ditka for Senator.  Can you believe it?  Did we learn nothing from the Schwarzenneger nightmare?  Oh wait, that\'s right, these are Republicans we are talking about.  I forgot.  So let\'s see, we have a complete moron as President, who stole the damn office in the first place (thanks to the Supreme Court and the legendary \"hanging chad\"), who can barely form a complete sentence without rehearsing it first, who was the Governor of one of the largest states in the union (geographical-wise) probably because of his name alone (of course that name wouldn\'t have influenced  my  decision, but I digress), and who has run our country into the ground financially, diplomatically, and politically.    We also have a completely unqualified actor born in another country who knows nothing more about running a government than the fact that he made millions upon millions of dollars uttering such simple lines as \"I\'ll Be Back\" and \"It\'s NOT A TUMOR\" in movies as the governor of the largest state in the union (population-wise), elected to office in the greatest circus of an \"election\" known to man (after the Bush heist).  So, then, in theory, why shouldn\'t Illinois elect a completely unqualified candidate for Senator, whose claim to fame was one (count \'em) one Super Bowl trophy for Chicago\'s Bears (Sorry, I\'m a Packer fan, so I just don\'t like the guy anyway), whose speech patterns make Richard Daley look like a Fulbright Scholar, and whose knowledge of politics and making sound decisions lie in his decisions whether or not to play certain linebackers and whether to go for the 1st down or try for a field goal. (Scare ya that I know something about Football?)  Yes, folks, the Republican Party is proving once again that they are truly trying to ruin our country.  Why put qualified people into office when we can push buffoons and rejects to the front of the line to represent us all?    These are scary times, folks.   Please  register to vote, please pay attention to what is going on, and please, I urge you, do what you can to make sure that Bush gets hedged this election year.   '),(4927,3869854,'13,July,2004','2004-07-13 00:00:00','   (This story was written in 1995 for my first website.)   - - - - - - - - - - - - - -  I drove over to the Citgo station.  I could see him in the window.  I was quite pleased. =)  I gassed up, drove the car up to the building to park, and got out.  When I walked in, he immediately recognized me.  \"HEY!\"  he said happily.  \"I was hoping you\'d stop in!\"  My heart leapt with glee.  \"I love the box.. I\'ll keep it... but I wanted to know what you knew about car stereos..\"   He proceeded to ask me a bunch of questions  about car stereo equipment... we talked about that for a while... then talked about where he was from... then talked about families.. then talked about music.. .then about TV.. then about God knows what else... I got in there at maybe 10pm.. next thing I knew it was 1am and it was time to close the store!   He closed up, didn\'t kick me out.... we talked some more about whatever... finally he asked me \"So where  do you go out?\"   \"Well, here and there.. nowhere in particular...\" I answered.   \"Ever been to a place called \'Club 94\'?\" He asked. (Club 94 is a gay bar in Kenosha)   \"Well.. yeah... once or twice.. it\'s fun.. .I guess I like it.\" I answered, getting VERY excited.   He asked me if he could sit in my car while his warmed up (this was about the dead of winter of \'92, a VERY cold winter.  He had a Renault.  Need I say more?).  I of course obliged.  He locked up, we got in my car while his ran to warm up and he says,  \"OK I\'m not gonna bullshit you anymore.  I\'m gay.\"   \"Well that\'s good,\" I said with a smile,  \"because I\'m gay, too.\"   We talked a bit more.. exchanged phone numbers.. and that was it for that night.  We both had to get home.   His name was Donnie.  I was on CLOUD NINE for at least 48 hours after that.   I visited him nearly every day after that.  We\'d talk all night long, sit in my car and wait for his to warm up, then go home.  We never would kiss, we never got mushy.. it was just this strange new friendship thing we had.  I felt silly standing around a gas station convenience store for all hours.. and my dad wasn\'t pleased with it either.. especially because I was there to see a guy!  (My parents knew about me for a while before this)   Anyway about three weeks into my visits with Donnie a friend of his came in and hung out.  I could see that they knew each other pretty well... so at one point I asked him about Donnie.   \"How well do you know Donnie?\" I asked.   \"Why, do you like him?\" (he was rather blunt)   \"Yes.. why?\"   \"Well let me just tell you.  You\'re not his type.  I wouldn\'t try too hard\"   I was crushed.  As far as I knew, he knew Donnie better than anyone else, so I had to trust his word.  Still, I thought that if Donnie didn\'t like me that much he wouldn\'t have been talking to me so much and letting me stay so late... I mean, there had to be SOMETHING there.  However, I was very new to the whole gay thing and as far as I was concerned, his friend was right, so I made my visits to the gas station less frequent.   But eventually that changed and I would stop in again just as frequent as before.   One night we went to a bar behind the gas station for a quick drink after he got out of work.  We had known each other for about three months at that point.  I still didn\'t know what he thought about me, but I knew how I felt about him.  I was falling head over heels in love with Donnie, and it was driving me nuts.   After the bar closed, we sat in my car and talked.  \"This is it,\" I told myself.  \"It\'s now or never\".   \"Donnie, we\'ve been seeing each other now for about three or four months... and I\'ve really grown to like you .... a lot.  I just want to know... what do you feel?\"   \"I feel the same way Rick.\"   \"Really?   \"Yes.\"   I then asked him if I could kiss him.  He said yes.  It was the first time I had ever kissed anyone.  And it was wonderful.    Part 3 to come... stay tuned!   '),(4928,3869854,'13,July,2004','2004-07-13 00:00:00','   So I\'m getting ready for my trip to Montreal!  I leave this Friday for 10 days in sunny (?) Montreal as part of the GALA Choruses VII International Festival in Montreal, Canada.  This is my second GALA, but I\'ve never been to Montreal before, so I\'m really excited.    I\'m going with CGMC and the Slickers... but I\'ll have plenty of time to see other choruses (and plenty of the sights as well!)  Yesterday I got some shopping done, printed out all my itineraries (Gotta love Orbitz for the online tix), a map of the downtown area, and got all my paperwork together.  All I have to do now is laundry and pack.  Last night I went online to Gay.com and visited the Montreal room, where I chatted with this super cute boy from Ireland and another from Columbus, OH who saw our chorus when we went to a regional festival in Cincinnati.  I think the super cute one from Ireland wants to meet... I hope so!  He\'s a hottie!   Anyone Out There?   Hey if you\'re reading this, how\'s about ya click on \"Comments\" and leave one?  I\'d love to hear from you.  Let me know what you think.  Am I keeping you interested?  Do you have suggestions?  Do I totally suck?  Do you want to bear my children?  Let me know!   '),(4929,3869854,'12,July,2004','2004-07-12 00:00:00','   Please visit  urlLink Moveon.org  to write and send your own letter to the President and your Senators and Representatives.   Dear President Bush, Senators, and Representatives:   As not only a gay man, but also as a citizen of the United States of America, I cannot help but be disturbed by the removal of freedoms from our citizens that is occuring right in front of our faces.  This is shameful and frightening.  Our country was founded on freedom, and by removing the possibility of freedom from one class, group or sect of people, we are deconstructing the very foundation of our nation.  This is wrong, and I urge you to realize it.   I urge you to reject the politics of hate and division.   Everyone has the same rights. There is no place in America for a Constitutional amendment denying marriage equality to anyone.   Sincerely,   Richard L. Aiello, Jr. Chicago, IL   '),(4930,3869854,'12,July,2004','2004-07-12 00:00:00','   Check out  urlLink Im\'mature  - the blog of Flip, the first outside visitor to my blog!   '),(4931,3869854,'12,July,2004','2004-07-12 00:00:00','   Well I got my wish.  Damn it\'s hot out.  And humid.  Let the parade begin!  And just in time for me to go to Montreal.  I hope to meet a hot French Canadian boy.  Or two.   '),(4932,3869854,'12,July,2004','2004-07-12 00:00:00','   Donnie was my first love- but he got away.  This is our story.   (This story was written in 1995 for my first website.)   - - - - - - - - - - - - - -  It all started in the winter of 1992....   I was working at a Best Buy in Racine, Wisconsin at the time.  I had a 20 minute drive to get from there to Kenosha, where I was living, and the majority of the drive was along a highway that connected with my street.  As I entered Racine, I would always pass by a Citgo gas station.  On the way home from work, I would frequently stop at this gas station to fill up my gas tank.   One night I stopped and went inside to pay my bill.  There at the counter stood the most adorable, albeit gorgeous guy I\'d ever seen.  He had deep blue eyes, brown hair and was about 6\' tall.  I immediately knew he was gay... that instinct \'gaydar\' we all have... but being rather shy I didn\'t say too much.  I simply paid my bill and left.   Needless to say my visits to that gas station became all the more frequent.  It seemed that he worked there EVERY night.. so I was almost never disappointed.  Once in a while I\'d make small talk.. maybe just smile... he had the cutest Southern accent... I couldn\'t begin to place where it was from, but that combined with his adorable looks just made him all the more irresistible.   At Best Buy I was a Product Specialist in the Audio Department.  I had been there a few months so I pretty much knew what I was doing.  One day I was at our \'answer center\' and I looked up... whom did I see but Gas Station Guy, walking into the boombox aisle. I quickly scooted over there and said \"HI\".   He recognized me from the visits to the gas station... and smiled.  He wanted to buy a box... so I showed him all the features, gave it to him, and said, \"See you around\".  That was about it.  I was in bliss.   The next day I was shocked to see that he had returned, with the boombox in hand.  I waited for him to finish at Customer Service, then when he headed toward the boombox aisle again I asked him \"Didn\'t like that one?\"   \"No, I decided I wanted one with a CD player.\"   \"Well let me show you some that have one.\"   We talked a little more small talk and I sold him the new boombox.  I couldn\'t tell if he liked me or not.. and for that matter I couldn\'t tell if he was gay or not.. but deep down I knew he was.. he just had to be.   That night, I looked at my gas gauge and... what a surprise!  I needed gas! =)    Stay tuned for Part 2...   '),(4933,3869854,'12,July,2004','2004-07-12 00:00:00','   The wedding was just lovely.  Like I said, I\'d never participated in a gay wedding before.  It was quite a sight.  It was also very multicultural.  The grooms, David and David (yes, how cute.) wore kimono-type outfits, the attendants (Is it a Bridal Party if there is no Bride?) wore tuxes, the ceremony was Irish, Indian and Jewish in influence (replete with the glass breaking and \"Mazeltov\" at the end), the food was Italian, Vegan and Seafood, and the fashions ranged from the sublime to the casual.  Joan Rivers would have had a field day.  The Slickers sang two of their songs, \"Come What May,\" which I had a solo in, and \"Alone in the Universe/Nature Boy.\"  I also had a solo... the processional song, which was \"Someone Like You\" from  Jekyll & Hyde .  Unison also sang a set of songs.  So the music was Broadway and Movie influenced.  It was a fun time for all, really, and emotions ran high.  David E, who I am in the Slickers with, read his vows, which borrowed heavily from the lyrics of \"Someone Like You.\"  I was touched by the meaning of the song, which I really hadn\'t latched onto when I actually sang it.    I\'m still living the first verse of the song, up until \"Someone Like You\" comes along.  Had this moment occurred years ago (around the time of my  urlLink Old Journal  posts) I probably would have become all melancholy and full of regret, wondering if I\'ll ever find \"the one\"...    Ok maybe I had a moment or two of that...  But then it passed, and it was time to party!   '),(4934,3869854,'11,July,2004','2004-07-11 00:00:00','   No sooner than it starts, it\'s all over!  The shows on Saturday went fabulously, and although the crowds were modest in the afternoon, they were both very receptive, and we absolutely performed our best.  I was so proud of us (and the others.)    Afterwards was a huge (and I mean HUGE) party at Charlie\'s Ale House in Andersonville (Yay, just stumbles from home!) and a chance to celebrate our wonderful show.  Indeed, we should all be proud of what we are bringing to Montreal this year.  It will be a very good showing.  Today I am attending my first-ever gay wedding.  In fact, I am singing a song and performing with the Slickers.  The wedding is for the board chair of Windy City Performing Arts.  I\'m excited because I get to sing, of course, (where I will get the voice to do it I have no idea) but also because I\'ve always wanted to witness an affair such as this.  Of course it is not a government-sanctioned WEDDING, as those are not truly legal in Illinois (yet), but just the fact that these two people have decided to commit to each other for the rest of their lives and do so in front of friends, family and the like is so wonderful and so moving to me that I could not say no.  Of course this begs the question, what on earth do I wear!?!?  Well it\'s not happening for a while, and I\'m only awake because the cats wanted food.  I will sleep on it further and see what comes to me.   '),(4935,3869854,'31,July,2004','2004-07-31 00:00:00','   urlLink    Screech in the Sunshine, my contribution to \"A Minute in the Life\" for  urlLink Gay Bloggers on Tribe.net . Isn\'t he adorable? :)   Picture hosted by Hello by Picasa    urlLink   '),(4936,3869854,'30,July,2004','2004-07-30 00:00:00','   Sometimes I wonder if I\'m posting stuff for my own entertainment.  Maybe I am.   In any case it doesn\'t look like anyone\'s reading this thing.  Am I that boring?  Would you rather hear of my sexual escapades?  Oh wait, there aren\'t any to tell about.  How about my dating life?  Oh yeah, that\'s right, I don\'t have one.   I\'m sorry, I am just foul and bitter about things lately.   I went to my friend Eric\'s wake tonight.  It was, to say the least, a very sobering experience.  Three weeks ago I was talking to him.  Now he\'s gone.   He was only 37 years old.  Ridiculously young.  His body failed him.  Bad kidneys, bad liver.   Why is it that someone so wonderful had to be struck down so quickly?   His family was all there.  Eric was Filipino.  There were family members everywhere.  They loved Eric.  They knew he was gay.  They loved him because he was their son, their brother, their nephew or cousin.   It made me feel even more love and appreciation for him.  I could see where he got his wonderful personality from.   Tomorrow is the funeral, and the Chorus is singing.  I don\'t know how I\'ll get through it, but that\'s what we do-- we sing, even when times are rough and our hearts are broken.  It\'s the greatest gift we can give to Eric.  He would be so thankful.  As we were for him.  Sorry, I\'m overly melancholy tonight.  I actually tried to sit online and coordinate a hookup.  Bad idea.  I\'m just not in the mood.  In fact, I haven\'t been for quite some time.  I just don\'t want to bother with that anymore.  Too much work, and not enough reward.  What\'s the point?  If you\'re reading this (if anyone is) you may have also noticed that I jumped on the \"100 Things About Me\" bandwagon.  Of course, I added a few \"sub-points\".   There might be some surprising things in there, there might not be.  But it\'s all about me.  I\'ll move them to my other server after a while so they don\'t clog this page up so much.   Boy is this ever a true RcktRambling.  I\'m just spouting out what\'s in my mind.  Random thoughts.  I guess that makes a true weblog.  Or a journal.  Or whatever you want to call it.   A boy in distress   Cries out for the hand   That comforted him   Throughout his life   From childhood   To adulthood.   He reaches for it,   Finding the air,   And grasps at nothing.   Thing about me #93:  I like to write poetry to clear my mind.   '),(4937,3869854,'30,July,2004','2004-07-30 00:00:00','   1. I am single. 1a. This rarely ever changes. 2. I love to sing. 3. My favorite food is Italian. 4a. I make a kickass meatball. 5. I was born in Kenosha, Wisconsin 6. I am a Packer fan, but not a Brewer Fan. 7. I am a Cubs fan, but not a Bears Fan. 8. I used to have a Cabaret show at Gentry in Chicago. 9. I worked for Six Flags Great America for 12 years before the bastards fired me. 9a. It was the first time I\'d ever been fired. 9b. It was the best thing that ever happened to me. 10. I grew up with dogs, but I own two cats. 11. I have one sister, and she is married. 12. I have no nieces or nephews (yet). 13. Most of my friends are in the Chicago Gay Men\'s Chorus or the Windy City Performing Arts. 14. I am a member of both groups. 15. I am not related to Danny Aiello. 16. My favorite artist is Elton John. 17. The RcktMan in my name is an homage to Elton John\'s song \"Rocket Man\" 18. My first online experience was on Prodigy back in 1992. 19. I used to consider myself an Independent voter, but I now consider myself a Democrat. 20. I was raised Catholic, but do not practice it anymore. 21. My uncle is a priest. 22. I have only one first cousin. 23. My mom had my sister because she did not want me to grow up an only child. (She was an only child.) 24. We named our second dog \"Cubby\" after the Chicago Cubs. 24a. His name was my idea. 25. My cat\'s names are Pippin and Screech. 25a. Pippin is named after the musical, not the basketball player. 25b. Screech is so named because he\'s so chatty, not after the \"Saved By The Bell\" character. 25c. My ex named Screech. 26. I have friends all over the country and in many areas of the world. 27. I have never been to Europe. 28 I have never been to paradise-- or to me, for that matter. 28a. I don\'t know what the hell that means. 29. I never knew either of my grandfathers. 30. I have relatives in Italy who are enormously wealthy because of their coffee empire, \"Aiello Coffee\" 30a. I have bought and tasted Aiello Coffee. 30b. I don\'t know any of these relatives, therefore will probably not inherit a penny when they die. 31. I did the layout for Carol Moseley Braun\'s resume in 2001. 31a. She paid me $200 for the service. She paid with a personal check. I have a copy of the check. 32. I have never kissed a girl. 33. I have never had sex with a woman. 34. My first kiss and my first time having sex with a man was when I was 23 years old. 35. I am now 33 years old. 36. I have a solo on the Chicago Gay Men\'s Chorus\'s first CD, \"Cool Yule.\" 37. My favorite drink is Rum &amp; Coke 38. I have never done drugs of any kind. 39. I have never smoked a cigarette. 40. I have smoked cigars but only puffed on them. 41. I am a top. 42. I love to kiss. 43. I love to cuddle. 44. I am extremely affectionate. 45. I can be a bit of a loner at times. 46. I have been called \"introspective\" 47. In college I was heavily involved in DECA, a student marketing organization. 47a. I was elected Wisconsin State SE Region Vice President in 1992 and State President in 1993. 47b. I was elected State Alumni President in 1998 (after having moved to Chicago). 48. I served on the board of CGMC for 2 years. 49. I will not serve on any boards again for a long time. 50. My last ex and I were together for a year. 50a. It was my longest relationship. 51. One of my fondest memories is going to the Wisconsin State Fair with my family and going down the huge slide. 52. Another fond memory is going to see my Grandma in Antioch, Illinois. It always seemed so far away. 53. I had my tonsils removed in July of 2003. 54. I was diagnosed HIV Positive in June of 2003. 55. I am mentally, physically, and emotionally doing just fine. 56. I am surrounded by a network of wonderful, loving and supportive friends. 57. My family does not know about my HIV status. 58. My father is Diabetic and my mother suffers from Emphysema. 58a. They have enough to worry about without having to worry about me too. 59. While growing up, I wanted to be a teacher. 60. I graduated with an Associate Degree in Marketing in 1994. 61. I still want to complete my Bachelor\'s and Masters degree someday. 62. I live way beyond my means. 63. I have horrible credit. 64. I own a car, but still rent my apartment. 65. I would like to buy my own condo within the next 3 years. 66. I cry at sappy movies and songs. 67. I love comedies. 68. I love gardening 69. I grow my own herbs. 70. I hate housecleaning and laundry. 71. But I love when my home is clean. 72. I have a gym membership that I haven\'t used in 4 years. 73. My bike has sat with a flat tire for 3 years. 74. I thoroughly believe that if I worked out, I\'d be \"Hot\". 75. I am too lazy to make that happen. 76. I listen to music when I fall asleep. 77. \"Queer Eye\" is about the only TV I watch anymore. 78. I hate seeing movies by myself, so I rarely see movies. 79. I love to cook for myself. 80. I rarely ever go out anymore. 81. I cut my own hair. 82. My favorite breakfast food is Eggs Benedict. 83. I love Tombstone Pizzas. 84. I would fail miserably at the Atkins Diet. 85. White Castle hamburgers are my nemesis. 86. I have very few enemies. 87. I believe there is a direct correlation between one\'s relationship with his or her family and the type of person he or she is socially. 88. I have a very good relationship with my family. 89. I am not a morning person. 90. I am a definite night owl. 91. I spend a lot of time online but rarely ever \"hook up\" anymore. 92. I\'m ready to \"settle down\". 93. I like to write poetry to clear my mind. 94. My favorite author is John Irving. 95. I think Dolly Parton is one of the most underrated artists alive today. 96. I think The Beatles were geniuses. 97. I think George W. Bush was the biggest failure of all of the US Presidents, ever. 98. I think John Kerry will do a fine job as our President. 99. I have voted in every Presidential election since I became eligible. 100. I think the future of America is pretty bright.   '),(4938,3869854,'29,July,2004','2004-07-29 00:00:00','   The same day I got back from Montreal, I found out that one of my friends in the Chorus had died.    &nbsp;  Eric was one of those guys who was always around, and you always expected him to be there.&nbsp; He was quiet, unassuming, and warm.&nbsp; Yet he never wanted anything special.&nbsp; He just was a good person.    &nbsp;  He had a penchant for drag.&nbsp; In fact, he made a knockout drag queen.&nbsp; And he was fiercely talented when it came to makeup.&nbsp; Many of the boys in the Chorus came to him for makeup tips whenever they needed to be done up for&nbsp;a show.&nbsp;     &nbsp;  He had a radiant smile.&nbsp; It lit up the room.&nbsp; And a hearty laugh to back up the smile.    &nbsp;  He wasn\'t one of my best friends, but he was, indeed, a friend.&nbsp; Four years ago when I was working a dead-end job in retail, he stopped in at the store and we had a long talk about how I wasn\'t happy with my job and needed to find something new.&nbsp; Eric helped me.&nbsp; He recommended me for a position where he was working.&nbsp; I went in for an interview, but it wasn\'t the right fit for me at the time.&nbsp; However, hedidn\'t stop there.&nbsp; He got me in contact with a former co-worker of his who was leaving her position at the law firm she was working at.&nbsp; I interviewed with them, and have been there now for the last 3 years.&nbsp; Without his encouragement and that extra push that he gave me, I might not have made the transition at all.&nbsp; I have Eric to thank for that.    &nbsp;  Having said that, I certainly hoped I appropriately thanked him.&nbsp; I\'m pretty sure I did, but if not...    &nbsp;  Eric, thank you.&nbsp; You were a true friend.&nbsp;     &nbsp;  I will miss you.&nbsp;   '),(4939,3869854,'28,July,2004','2004-07-28 00:00:00','   Well I\'ve been at this now for about a month.&nbsp; So far it\'s pretty fun.     &nbsp;   On one hand, it\'s given me something fun to do instead of sit on gay chatrooms all night long.&nbsp; On the other hand, I\'m still up til all hours reading blogs and posting comments.&nbsp;The rest of the time I\'m tweaking this blog, adjusting the template, trying to make it just right.&nbsp; I could use some tips if something&nbsp;looks awry.&nbsp;&nbsp;So far, though, I think I\'m doing pretty good.    &nbsp;   I put my blog on a couple search engines in hopes that I get some more readers.&nbsp; Question is, will they want to read it?&nbsp; I think I have some interesting things to say and some good stories to tell.&nbsp; There\'s definitely more to come.&nbsp; What\'s here is only the beginning... I\'ve got other ideas conjured up, but I can only do so much.     &nbsp;   One thing I have to stop doing is blogging at work.&nbsp; I\'m gonna get caught.&nbsp; Can\'t let that happen.&nbsp; I got a really great review just before I went on vacation (plus a 5% raise.&nbsp; Can\'t beat that in these times.) So I just have to cool it.&nbsp; It\'s fun though.&nbsp; I&nbsp;could very easily be a powerblogger.&nbsp; I just need to set the right paramaters and stick to them.     &nbsp;   In the meantime, I\'m lonely, I\'m horny, I\'m bored, and I need some excitement in my life.&nbsp; Oh wait, I\'m also poor, at least until the 7th of August.&nbsp; Damn this job\'s stupid paydates.&nbsp; I need money, and NOW!&nbsp; (Montreal was an EXPENSIVE trip!)     &nbsp;   At least Chorus starts up again in a month.&nbsp;     &nbsp;   God, I\'m pathetic! &nbsp;   '),(4940,3869854,'27,July,2004','2004-07-27 00:00:00','   As promised, here are some of my favorite pictures from the GALA festival in Montreal:      Friends     urlLink   &nbsp;&nbsp;   Me and my friend Sonelius, formerly of CGMC, at the CGMC bash.&nbsp;       urlLink      A group of us at Arnie\'s birthday dinner &nbsp;    urlLink      The Slickers pose outside the performance venue before our Tech rehearsal. &nbsp;       Quebec City     urlLink      The Hiiills are Aliiive... with the sound of Muuuuusic....&nbsp;       urlLink      ChÃ¢teau Frontenac, the most photographed (because it\'s the most beautiful) building in QuÃ©bec City.   &nbsp;    urlLink      How gorgeous is this city? &nbsp;       Concerts      urlLink      CGMC members pose after their triumphant performance that opened the Festival. &nbsp;     urlLink      Unison performs \"My Soul\'s Been Anchored In The Lord\" &nbsp;     urlLink      New York performs \"You Can\'t Stop The Beat\" from \"Hairspray\" (The best of 8 different renditions at the Festival.)     Pictures hosted by   Hello   by Picasa&nbsp;  urlLink   '),(4941,3869854,'25,July,2004','2004-07-25 00:00:00','   If you haven\'t become a fan of this comic yet, please do so.&nbsp; I think you\'ll enjoy it immensely.&nbsp; It makes me smile.&nbsp;(Click on the comic to view at full size.)       One Big Happy &nbsp; by Rick Detorie     urlLink     Visit the One Big Happy website by clicking  urlLink here .&nbsp;   '),(4942,3869854,'25,July,2004','2004-07-25 00:00:00','   The last part of the story of Donnie, my first boyfriend.   (This story was written in 1995 for my first website.)    - - - - - - - - - - - - - -   About a month later I learned that Donnie would be moving to Houston. His mother and sister would be moving back to Atlanta, where they originally came from. I was crushed. I was really falling for Donnie and now he had to leave. We decided that even though he was leaving, we would continue things up until he left, thinking that maybe someday I could move down with him or vice versa.   It killed me to think he would be leaving... but eventually he did.   He was gone for about three months when one day I got a call from him.   \"What would you do if came back to Racine?\"   I answered, \"Well of course I\'d be incredibly happy.. but what would you do? You have no job, nowhere to live, and I can\'t move out because I\'m not ready!\"   He said, \"I could probably figure something out.\" I figured that was that and left it there.   The next evening I got home from work and checked the answering machine. There was a message.   \"Rick, this is Donnie. I\'m back in Racine. Give me a call tomorrow at this number..\"   I cried.. tears of joy and tears of sorrow. I didn\'t know where he was or where he was staying. But he was home. And as far as I knew he came back for me.   We saw each other the next day and holding him in my arms was the most wonderful feeling I\'d ever felt. He got his job back at the gas station and was living with the owners. I knew that wouldn\'t last long because he never got along that well with the owners... but for now he was here and that\'s all that mattered.   We dated for a while but for some reason the magic just wasn\'t there like it used to be. I became busy with school and then started working in Illinois. He worked every day at the station and we hardly ever saw each other.   Eventually I met another guy and we dated for a while. He was cute, younger, and fun to be with.. but he was 18 (and for some reason that bothered me)...   ...and I wasn\'t over Donnie.   I broke up with him about two months into our relationship. I told him I loved Donnie. He understood. He was a good guy.   I went to see Donnie the next night. To my shock he informed me that he would be moving to Lexington, Kentucky to live with his father. He wasn\'t happy in Racine and needed to get out. I was crushed more than ever before. I lost my chance with Donnie.. the guy I truly loved. We made out a bit in the back room and hugged a lot. Then I left.   About halfway down the road from the gas station, I called him on my Cell phone. He was still there. I was crying.   \"Donnie?\"  \"Yes?\"  \"It\'s Rick.. .I\'m in my car. I was just thinking after I left...\"  \"What\'s the matter?\"  \"I love you, Donnie. I will always love you.\"  \"I love you too, Rick\"  \"Don\'t go. Please.\"  \"I have to, Rick.. there\'s nothing else I can do\"   We talked for a little while then I hung up.   I saw Donnie one more time after that. He had a going away party but I couldn\'t go because I was sick. I cried so much I didn\'t think I\'d have any tears left. He called me from his party. We were both crying. I told him I loved him again. He told me he loved me.   The next day he was gone.. .and I haven\'t seen him since.   We have kept in touch via phone and mail since then. He now lives in San Jose, CA and has a room mate that he says likes him a lot.. but he\'s never stopped loving me or thinking about me. Judging from my tears as I type this, I realize that I still love him. Very much. And I always will. Perhaps we were meant to be together, and if so, I hope that someday we can and will be. Until then I hold very precious memories of the times we spent together, and great regret for all the times we could have been together but weren\'t. We parted against our will. We still have contact.. but it\'s not enough. I wish he were here now to dry my tears, to hold me and comfort me and tell me everything will be OK. But he\'s not and I must deal with that. I miss  him. I will always love him.   I guess what I learned from this story is, \"You don\'t know what you\'ve got until it\'s gone..\" as the song says. I never realized just how wonderful Donnie was until I realized he would be gone. Life is unexpected with the way it works. You could be walking down the street, minding your own business, when all of a sudden fate smacks you in the face with a new adventure, a new love, or a new tragedy. All you have to do is keep your eyes open and your senses aware. You never know what will happen next.   --------------------------------------------------------------------------------   A few months after this was written, I realized I had lost Donnie\'s phone number. I was pretty devastated, looking all over the place, and even phoning directories trying to find him, but to no avail. One day Donnie phoned me and informed me that he was, indeed dating the man he was living with in San Jose. They had been together for some time and he was happy with him. I was more than happy to hear from him again, and I finally got his phone number and could call him whenever I wanted, but I didn\'t. I felt it was time to close the wounds from this experience, and I was doing a good job of that already. So I called him very sporadically.   A while later, I heard from Donnie again. He was moving out of the apartment he was sharing with that guy, and they had broken up. He was moving a few blocks away into an apartment with a friend of his. I was slightly happy to hear that, but again I felt like it was useless getting excited about it. He still wasn\'t going to be moving back anytime soon. He said he\'d call me with his new phone number once he got settled into his new place.   I never heard from him again.   Closure has happened. I am over Donnie. Though I will still have a place in my heart for him, as most everyone does when it comes to their first love. I am still single, still looking for someone to fill that space, but I guess I compare everyone to Donnie. That\'s not fair, but I guess that is an unwritten rule somewhere when it comes to relationships. I had hoped he would call before I moved to Chicago, but he never did. It just wasn\'t meant to be. And I guess that\'s just fine. I\'ll live.     --------------------------------------------------------------------------------   Epilogue, 2004:     Many relationships--one much more serious than Donnie--later, and looking back at what I wrote nearly 10 years ago about him, I see now how young and naieve I was back then about love.&nbsp; Having been through the pain and the sorrow of the ending of a relationship one too many times, I know it never gets easier.&nbsp; But when you\'re 23 or 24 and still new to everything, it seems like the end of the world.&nbsp;   Donnie was special, though I wonder if I really  loved  him.&nbsp; I\'m not so sure now.&nbsp; I still think of him fondly, occasionally,&nbsp;but I don\'t really know if I understood what love was back then.&nbsp; I think I do now, but I also think that love takes on many different characteristics depending on who he is, what the relationship is based on, and the when, where, why and how you fell for him.&nbsp; The fact that I haven\'t had a relationship that has lasted over a year probably speaks to my still-evident inexperience in this area.&nbsp; Maybe, hopefully,&nbsp;that will change someday, but it is good to know that I am capable of loving.&nbsp; I know that will never change.   -R   '),(4943,3869854,'25,July,2004','2004-07-25 00:00:00','   MontrÃ©al,  mon ami , our time was shortened but I had a wonderful time.     &nbsp;   I came home a couple days early due to a lack of funds and a huge dose of homesickness.&nbsp; But that didn\'t stop me from having a wonderful time and enjoying nearly every second of the trip.&nbsp;     &nbsp;   Nearly... well, there were a few moments where I wondered whether my trip was ever going to be a  vacation  and not a whole lot of work. But then it turned around and became everything I had hoped it would be, and more.     &nbsp;   First of all, yes, I did have sex. Once.&nbsp; With one guy.&nbsp; And then it was over.&nbsp; It was fun but it could have been better.&nbsp; Oh well... at least I wasn\'t completely dry all weekend.     &nbsp;   Second of all, yes, I did have a few lap dances by a few stripper boys.&nbsp;  Oooh la la  are the boys there BEAUTIFUL.&nbsp; Everywhere you turn, another gorgeous man would walk by.&nbsp; Talk about distractions.&nbsp; So why not more sex?&nbsp; Well considering that I am painfully shy in public and have the confidence of a groundhog who sees his shadow around all those hot boys... well, I would probably say that sums it up. &nbsp;   Third, Montreal is not as pretty of a city as I had imagined.&nbsp; It makes Chicago look like the crown&nbsp;jewel of cities.&nbsp; I have never seen so much graffiti in all my life.&nbsp; And if you ever thought the&nbsp;panhandlers and&nbsp;street people of Chicago were bad, you haven\'t seen&nbsp;(or smelled) the ones in Montreal.&nbsp; Not pretty.    Don\'t get me wrong, there are parts that ARE pretty, I just&nbsp;didn\'t get a chance to see&nbsp;many of&nbsp;them.&nbsp; Instead, I took a&nbsp;day trip to Quebec City, and I would have much preferred to stay there.&nbsp; It was gorgeous... romantic, charming, everything I wanted from a destination city.&nbsp; Everywhere you turned was another perfect postcard waiting to&nbsp;be snapped.&nbsp; I will definitely be making a return trip there. &nbsp;    &nbsp;   So anyway, back to GALA..     The premise for the weekend was the GALA Choruses quadrennial festival where all of the LGBT choruses from all over the world who wish to participate come together in one city and present concerts for all of their peers.&nbsp; The larger choruses do the 30-minute sets and small ensembles of the larger groups present the 15 minute sets.&nbsp; I am a member of the Chicago Gay Men\'s Chorus&nbsp;who did a 30-minute set&nbsp;and the Windy City Slickers who did a 15-minute set.&nbsp; And let me tell you, when you perform in front of literally thousands of people who know what it\'s like to do what you are doing and appreciate you for coming out and singing your heart out as much as they enjoy doing it, the response is the most gratifying and stimulating thing you will ever hear.&nbsp; It\'s like being at a rock concert and YOU are the reason they all came and paid $300 a seat to see you.&nbsp;     &nbsp;   (Ok so maybe the $300 a seat is a bit excessive, but it\'s not too far from the truth considering the travel expenses, registration, etc.&nbsp; But I digress...)     &nbsp;   CGMC opened the entire festival, which is a pretty big responsibility to take on.&nbsp; Of course, since there are always two concert blocks going on at any given time, we were up against the Boston Gay Men\'s Chorus, which also is a very good chorus.&nbsp; But CGMC has opened every one of the last 7 GALA Festivals, so people come to see us.&nbsp; We had a full house.&nbsp; The set was extremely well-received and I heard nothing but raves the entire weekend.&nbsp; In fact, a documentary crew was filming our set and wants to include one of the numbers in the final film.&nbsp; I think that\'s pretty exciting.     &nbsp;    The Slickers also had a fun set.. although I was  much  more nervous about their set than I was for CGMC, and for obvious reasons... there\'s only 10 Slickers, whereas there were over 70 guys in CGMC.&nbsp; The same screw-up in the Slickers is seen or heard a lot more clearly than in CGMC.&nbsp; Luckily there were none.      &nbsp;   The rest of the time was spent watching other choruses perform and catching some of the sights, sounds, tastes, and other sensual pleasures of the MontrÃ©al/QuÃ©bec area.&nbsp;     So, here is a list of the most&nbsp;memorable moments (good and bad) during my trip to MontrÃ©al:     Performing on stage with  urlLink CGMC  and hearing the audience roar-- numerous times.     Performing on stage with the Slickers and getting such an amazing response.      urlLink QuÃ©bec City .&nbsp; In a word:  Magnifique .     Seeing my friend (and former duet partner) Sonelius, who was formerly in CGMC but now lives in New York and is now in the  urlLink New York City Gay&nbsp;Men\'s&nbsp;Chorus .     Getting to know the men and women of  urlLink Windy City Performing Arts  better.     Cringing and pretending I could disappear during The Rainbow Chorus (\"Flowers Have Agendas!\"&nbsp; Honest.)     Stunning performances by:                            The  urlLink New York City Gay&nbsp;Men\'s&nbsp;Chorus                                 The Los Angeles Choral Artists (only in extistence for 10 months before coming to the Festival)                           urlLink MUSE- the Cincinnati Women\'s Choir                                  urlLink Lavender Light Gospel Choir                                  urlLink The Gay Men\'s Chorus of Los Angeles                                  urlLink The Columbus Gay Men\'s Chorus                                  urlLink Connecticut Gay Men\'s Chorus &nbsp;(Them boys is just out of control!)                           urlLink Heartland Men\'s Chorus                                 and of course all of the Chicago choruses.                            The comedy stylings of  urlLink Lily Tomlin  and  urlLink Kate Clinton , forever goddesses.     Singing \"The Hiiills are alive with the sound of MUUUUUSIIIIC\" on the hills of the  Citadel  in QuÃ©bec City     Lunch at  urlLink Le Saint Amour  in QuÃ©bec City, where I had some of the most delicious Mushroom Creme soup I have ever tasted (along with everything else I had there that was amazing)... and I hate mushrooms!     The Travelodge MontrÃ©al Centre, the tiniest hotel room I have ever been in.&nbsp; God I am glad to have my bed back.     Birthday dinner for Arnie at Food.&nbsp; Yes, that\'s a restaurant.     Parties on the Sky Bar terrace.     Arnie, Jeremy and John in the hot tub on the Sky Bar terrace.       The cute boy that joined them in the hot tub after they had been in there a while.     The pictures&nbsp;I have to&nbsp;bribe them with.      Naked strippers at Campus.    Lap dances at Campus.    Being on vacation with all of my closest friends.    Coming home to&nbsp;a clean apartment and my two adorable kitties!     I\'ll be posting pictures shortly to accompany this post, so check back.&nbsp; Til then, cya.     &nbsp;   '),(4944,3869854,'06,August,2004','2004-08-06 00:00:00','   So I\'m sitting here, putting off my trip north to Kenosha to see my parents.  This is oh-so-typical.  It\'s not that I don\'t get along with them-- I do. It\'s just that I loathe sacrificing my weekend for a weekend of parental bliss. Is that selfish?  Again, please, don\'t get me wrong, I LOVE my mom and dad. I really do. But sometimes, a little dose is just enough.  This weekend is looking to be a big dose. I\'m going up tonight because tomorrow morning we get to do chores all day. That\'s me, my sister and her husband.  The chores are being planned because we are getting ready to move Mom and Dad out of their house. It\'s a monumental task. There is nearly 40 years worth of shit in that house. The basement, for years, was a dumping ground. If a toaster broke, we bought a new one. But instead of throwing out the old one, it went in the basement.  Why?  \"Someone might need to use it.\"  HUH!? It\'s BROKEN! THROW IT AWAY! Nope. There it sat, along with three other toasters, 5 electric frying pans, countless broken pots and pans, a few blenders and even an ancient mixer that my grandma once used.  My parents are pack rats. And they are the worst kind of pack rats-- major junk savers.  So this is round two of junk disposal. Round one was my sister and her husband spending a week in the basement, dumping crap. They filled up three huge dumpsters. THREE. And it\'s still not done.  Of course, the fact that I wasn\'t there to help was a major bone of contention. In fact, it caused my sister and I to have a HUGE fight, something we rarely ever do. But it was Pride week... and I told them I couldn\'t do it that week, but that was the only week that worked for them. So they did it, and I wasn\'t there.  This weekend we are hitting the garage, which is another post for when I get back. Imagine nearly 40 years of paints, chemicals, and fertilizers. Oh I bet you can. I can\'t believe the damn thing never blew up.  I need to get going before it gets too late. Wish me luck.   '),(4945,3869854,'05,August,2004','2004-08-05 00:00:00','   The recent Superior Court decision in Seattle is passionate and articulate about several reasons why  marriage equality just makes sense.  For those who might be interested, here are some excerpts from King County Superior Court Judge William L. Downing\'s opinion in Andersen v. Sims:    urlLink (Click here for the full text of the Judge\'s opinion.)   There was no deeply rooted tradition of interracial marriage at the time of the U.S. Supreme Court\'s consideration of anti-miscegenation statutes in  Loving v. Virginia , supra; yet, the Court analyzed the issue of their constitutionality in terms of the broad right to marry and found that right to have been infringed.  There was no deeply rooted tradition of marriage while delinquent in child support payments at the time of the U.S. Supreme Court\'s consideration of statutes prohibiting this in  Zablocki v. Redhail ...(1978); yet, the Court analyzed the issue of their constitutionality in terms of the broad right to marry and found that right to have been infringed.  There was no deeply rooted tradition of inmate marriage at the time of the U.S. Supreme Court\'s consideration of statutes restricting this in  Turner v. Safley ... (1987); yet, the Court analyzed the issue of their constitutionality in terms of the broad right to marry and found that right to have been infringed....  It is true that marriage has long been defined as the union of one man and one woman. It is equally true that the shape of marriage has drastically changed over the years. It took a very long time for the courts (with legislative bodies sometimes understandably following just a little behind) to break down the traditional stereotypes that relegated women to second class status in society and in the marital relationship....Serving tradition, for the sake of tradition alone, is not a compelling state interest....  Some declaim that the institutions of marriage and family are weak these days and, in fact, stand threatened. Any trial court judge who regularly hears divorce, child abuse and domestic violence cases deeply shares this concern. It is not difficult, however, to identify both the causes of the present situation and the primary future threat.  They come from inside the institution, not outside of it.   Not to be too harsh, but they are  a shortage of commitment and an excess of selfishness.  Before the Court stand eight couples who credibly represent that they are ready and willing to make the right kind of commitment to partner and family for the right kinds of reasons.  All they ask is for the state to make them able....  Of course the laws never have placed a requirement on marriage that the parties procreate nor do they prohibit from marriage those who are unable or disinclined to procreate....  The legal question is not whether heterosexual marriage is good for the replenishment of the species through procreation. It is. The precise question is whether barring committed same-sex couples from the benefits of the civil marriage laws somehow serves the interest of encouraging procreation. There is no logical way in which it does so....It rationally serves no state interest to harm certain...children by devaluing the immediate families that they have joined....  It is good for children to be raised in stable families with a father and a mother. There is not the slightest question about this. It is a situation to be encouraged by the state. But, can it be said that fewer children will have this stability because couples consisting of two men or two women are allowed to have a relationship that is state-sanctioned?  There is no reasonable explanation for why this would be so.  There is no reasonable expectation that, should such a legal result come to pass, married fathers and mothers will abdicate their parental responsibilities or young would-be parents will defect from the ranks of heterosexuals....  On the other hand, when one peers into the future, one circumstance is far more certain to occur. Many, many children are going to be raised in the homes of gay and lesbian partners.... One, then, must try to envision two categories of future children. The first category consists of those whose heterosexual parents will either neglect them or never conceive them because same-sex marriage has been legalized. The second category is those children who will be raised in a home with same-sex adult partners and who would enjoy enhanced fmaily stability and social adjustment if these adults were granted the benefits of civil marriage. The only reasonable conclusion is that the very real second category greatly outnumbers the first theoretical one.  Therefore, the goal of nurturing and providing for the emotional wellbeing of children would be rationally served by allowing same-sex couples to marry; that same goal is impaired by prohibiting such marriages.... Unlike the documented impact of children\'s exposure to domestic violence and substance abuse in the homes of lawfully married heterosexual couples, as to children raised by intact same-sex couples there is no science, only questionable assumptions based on stereotypes.  The Court concludes that the exclusion of same-sex partners from civil marriage and the privileges attendant thereto is not rationally related to any legitimate or compelling state interest and is certainly not narrowly tailored toward such an interest.   If there is indeed any outside threat to the institution of marriage, it could well lie in legislative tinkering with the creation of alternative species of quasi-marriage....Better, perhaps (in terms of simplicity, fairness and social policy) to allow all who are up to taking on the heavy responsibilities of marriage, with its exclusivity and its \"till death do us part\" commitment, to do so - not lightly, but advisedly.... The privilege of civil marriage and the various privileges legally conferred by that status are not being made equally available to all citizens.... [In addition, the] denial to the plaintiffs of the right to marry constitutes a denial of substantive due process.   [T]hese plaintiffs...include exemplary parents, adoptive parents, foster parents and grandparents. They well know what it means to make a commitment and to honor it. There is not one among them that any of us should not be proud to call a friend or neighbor or to sit with at small desks on back-to-school night.  There is no worthwhile institution that they would dishonor, much less destroy....The characteristics embodied by these plaintiffs are ones that our society and the institution of marriage need more of, not less.    Beautifully spoken words. Your thoughts?   '),(4946,3869854,'05,August,2004','2004-08-05 00:00:00','   OK, this creeped me out. A bit too close to home, folks.     Dirksen Federal Building   Chicago, IL   Photo Â©  urlLink Chicago Tribune      Ill. Man Accused of Courthouse Bomb Plot   By NATHANIEL HERNANDEZ Associated Press Writer    Published August 5, 2004, 5:03 PM CDT    CHICAGO -- A convicted counterfeiter with an apparent grudge against the courts was arrested Thursday on charges of plotting to blow up a federal courthouse, but he never actually had materials to make a truck bomb, authorities said.  Prosecutors said Gale William Nettles, 66, was arrested with a pickup truck containing 1,500 pounds of fertilizer that he allegedly thought was volatile ammonium nitrate, the farm chemical used to blow up the Oklahoma City federal building.  Nettles had planned to sell the chemical to terrorists who would blow up the Dirksen federal building, U.S. Attorney Patrick J. Fitzgerald said in announcing the charges Thursday. But all the other people involved, including the \"terrorists\" and the people who sold him the fertilizer, were cooperating witnesses or federal agents, Fitzgerald said.  Nettles told the undercover agent he could make a 3,000-pound fertilizer bomb.  \"He had a rational plan to build a bomb. We weren\'t going to wait to see if it would work,\" Fitzgerald said.  Timothy McVeigh used a bomb made of 4,800 pounds of ammonium nitrate to blow up the Oklahoma City federal building in 1995, killing 168 people.  Nettles was arrested at a park early Thursday with the pickup truck when he met the undercover agents who he thought were terrorists, according to the criminal complaint. The fertilizer he obtained in the sting does not have the explosive potential of ammonium nitrate.  According to the complaint, Nettles met July 26 with an undercover agent he thought was a member of a terrorist group. In a recorded meeting, Nettles said he had a half ton of ammonium nitrate in New Orleans that he could have in Chicago in two days and that he had a target in mind -- the U.S. courthouse downtown, the complaint said.  A federal magistrate Thursday ordered Nettles held until a detention hearing Tuesday. No plea was entered, and defense attorney John Theis declined to comment afterward.  Nettles was released from prison in 2003 after serving time for counterfeiting and apparently retained a grudge against the court system, Fitzgerald said. The Dirksen federal building in downtown Chicago houses federal criminal and civil courts and the U.S. attorney\'s office.  Fitzgerald said Nettles was not working with any other groups but had asked cooperating witnesses about contacting Al-Qaida and Hamas.  Nettles also had a previous conviction for armed robbery in Cook County and was on parole at the time of his counterfeiting arrest. In 2001 he claimed he had mental disabilities, according to court documents.  \"He\'s 66 and he\'s spent -- probably conservatively -- 45 of those years in prison,\" said attorney Ronald J. Clark, who represented Nettles in the counterfeiting case. \"It surprises me that Gale would seriously be involved in something of this magnitude,\" Clark said. He described Nettles\' counterfeiting as a \"fairly pathetic attempt.\"  Nettles had been living at a transient hotel in Chicago since his release following the counterfeiting sentence, authorities said. Sam Gelles, 35, another hotel resident, said Nettles was an introverted computer whiz who went by the nickname \"Pops.\" \"You couldn\'t really read him,\" Gelles said.   Copyright Â© 2004, The Associated Press   '),(4947,3869854,'04,August,2004','2004-08-04 00:00:00','   My department manages the website for our Firm. It\'s usually a pretty easy task, and doesn\'t require much participation by me or my co-workers. There are those who have the duty of updating the content, and that\'s their job. Aside from formatting an image or two a day, I don\'t have much involvement.  So yesterday, we posted a story on the website. When this happens, an Email is sent out to the department and all involved attorneys about the posting.  Well damn it if the floodgates didn\'t open up.  Something was apparently VERY wrong.  First one partner REPLIES ALL and says he found an error. Then ANOTHER partner does the same thing. Then ANOTHER. Soon they\'re having a virtual E-chat about the error. And my mailbox is piling up with these pointless, annoying E-mails that have absolutely NOTHING to do with me whatsoever.  And it didn\'t stop there. After they got the problem resolved, they proceeded to banter. And let me tell you, attorneys who banter are NOT FUNNY in the slightest. Oh they try... but oh, they fail.  Meanwhile, my mailbox was bursting at the seams. My \"New Message\" indicator popped up so many times I was about to scream. I wanted to REPLY ALL to everyone in 72 point lettering \"FOR THE LOVE OF GOD, MAKE IT STOP!\"  When all was said and done, I had nearly 20 Emails on the said subject. Don\'t these people have billable hours to worry about? And don\'t they know how to reply to one, and NOT to all? It\'s not just that we aren\'t involved... we just don\'t CARE!  Besides, I have much more important things to do with my time... like bitch about things like this on my blog.  Silly people.   '),(4948,3869854,'04,August,2004','2004-08-04 00:00:00','   This post was originally a comment to a post by  urlLink Hot Toddy . (click the link to read his post.) But it was too long for a comment, so I\'m posting it here in its entirety...    Wow what\'s with all of us getting reflective lately?  I know when I first moved to Chicago, I wanted to get away from the quiet, rural life of Kenosha (a/k/a Kenowhere) Wisconsin and plop myself right in the middle of a big city, with bright lights, big buildings and all kinds of crazy people everywhere I turned. I got what I wished for. I moved to an apartment in a not-so-great area that had an incredible view of the skyline. Every morning I woke up and looked at it and thought \"Damn, I finally live here!\"  After a while, the novelty wore off. I missed grass and trees. I missed sidewalks that weren\'t caked in beer bottles and litter. I missed kids riding bikes and laughing and the feel of a neighborhood. The area I lived in didn\'t have that. It was a concrete jungle, and I was in the heart of it.  I looked around and found a great neighborhood that had what I was looking for. Then I found a friend who needed a roommate. And I made my move.  Now I live alone in the same neighborhood, and for all intents and purposes, I still have what I was hoping for. It\'s quiet, I feel safe, I look out my window and see birds and trees and people doing things... that remind me of home.  You never give up who you are. You may supress, and even reject a little of it. But it\'s always a part of you. And wherever you go, and in whatever you do, and in the people you meet and make friends with, you find more parts of yourself. That\'s one of the joys of life.  I like it when you get sentimental. It reminds me of someone I know quite well. :-)   '),(4949,3869854,'03,August,2004','2004-08-03 00:00:00','   Keep watching this site. It will happen.  (Thanks to  urlLink Jake  for the link.)   '),(4950,3869854,'03,August,2004','2004-08-03 00:00:00','   Like the smallest pebbles   The drops of rain fall   Gently and soothing   Upon the windowsill.   As the wind blows,   They change direction,   Spattering against the glass.   Cleaning away the grime   Of a city caked in dust.   Bright white flashes of light   Brilliantly flare in the sky   As the sound of thunder   Crashes against the walls   And shakes them to their core.   I sit quietly in my room   And close my eyes   And imagine myself   Drenched in rainfall.   The wild weather\'s wrath   Encircling me in fury.   Washing me clean.   For the water cleanses   My soul and senses.   And as the rain subsides   So do I return   To everyday life.   -RA   '),(4951,3869854,'02,August,2004','2004-08-02 00:00:00','   Lola sure knows how to put it right. I\'ll let the comic do the talking.   urlLink   '),(4952,3869854,'02,August,2004','2004-08-02 00:00:00','   Bzzzzzzzz  Bzzzzzzzz    Bzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz    I went to bed VERY late the night after I got back from Montreal. I must have been wired still from the trip, even though I was completely broke (still am) and painfully exhausted. Nonetheless, I had blogging to do and guys to cruise. I had to catch up!  Very early the next morning, I was awakened by my doorbell. I remember wondering if I had forgotten about plans I had made with a hot boy and quickly started to grab the essentials-- lube, rubbers, etc...  But first I had to answer the door. I was more than half-asleep still, but I made it. It was the Fed Ex guy delivering a package to me. (Insert crass remark here.)  I buzzed him in and he left it on my front doorstep. I must not have been fast enough. (Or I fell asleep again somewhere in between the buzzing in and opening the door.) Before I went back to bed I opened the package... it was from my friend in NYC who\'s moving to Australia-- an American Pride Flag. How cool!  She left me a nice note.  I read it, still groggy and nodding off.  So I went back to sleep.   I don\'t know where I\'m going to hang/fly this thing.  I live in a half-courtyard apartment building with no flagpole.  I already have an American Flag that I bought after 9/11 that is still folded properly and put away for the same reason.  But it\'s cool.  I\'ll definitely keep it and treasure it... and someday I\'ll use it.   '),(4953,3869854,'02,August,2004','2004-08-02 00:00:00','   This is creepy... it\'s like it KNOWS me...  Thanks to  urlLink TunaGirl  (whose blog I just love) for this one.  I know it\'s an old thing but what the hell. I\'m new to this. Get over it. :)       R     Radiant       C     Complicated       K     Keen       T     Tame       M     Misunderstood       A     Amazing       N     Noisy            R     Rounded       I     Insane       C     Cheesy       K     Kind        (Cue \"Twilight Zone\" music... I swear. This is me to a \"T\".)     Name / Username:      urlLink  Name Acronym Generator    From   urlLink  Go-Quiz.com   '),(4954,3869854,'02,August,2004','2004-08-02 00:00:00','   It\'s been a while since I\'ve done a shout-out, so I need to catch up.  I\'m finally seeing some comments now and then, and I just want to say THANK YOU! :)  So here\'s a shout out to:   urlLink Toddy ,  urlLink Jake , David H.,  urlLink Archerr , and  urlLink palochi .  Thanks guys!   '),(4955,3869854,'02,August,2004','2004-08-02 00:00:00','   In  urlLink Gay Bloggers  on  urlLink Tribe.net  today, we are discussing a fellow blogger  who recently quit his blog because someone was hurt by something he said.  It rang home pretty true to me.  I\'ve been down that road before.  If you\'ve perused my blog anytime before, you know that from  urlLink 1997-1998 , I kept an online journal.  I started it when I was getting ready to move to Chicago and I let it chronicle pretty much everything that happened to me from the time I found out about the job and the move to the actual move into the city, to my adjusting to life there.  I probably didn\'t even put that much thought into it then, but reading it now, it\'s interesting to see what was going on in my head at the time.  Even though it was only a 6 years ago, I was in such a different place than I am now... a very different mindset.  I stopped posting to the journal because someone I cared about was upset about something I had posted.  I realized that I was venting TOO much of my emotions.  I was being mean.  I was expressing anger through my words.  When it got back to me that those words really hurt someone, I stopped.   I didn\'t do any type of journal until I started my blog just this past month.  The pages of my old journal are linked on my blog now.  I find them fascinating and read them from time to time-- maybe to recapture some sort of \"Newness\" that life had back then, or maybe to remember some of the really good times I had.  (Oddly enough, sometimes I have a hard time remembering people and details.  Time does march on.)  I want this blog to be more about me than the old journal was.  That was essentially just a diary without \"Dear Diary\" before every posting.  I like the blog concept because you can really share yourself more-- your likes and dislikes, your fears and your joys, and of course, the funny, crazy moments in life.  I hope that nobody ever gets hurt by anything I say here... as that\'s never the intention.  But a wise woman once said, \"Fuck \'em if they can\'t take a joke\"...   Or just fuck \'em... that would be nice too...  If you\'ll excuse me, I\'m going to go wash out my potty mouth and go to bed now.  Gnite.   '),(4956,3869854,'02,August,2004','2004-08-02 00:00:00','   Sex! Lies! Passion!  Oh I could be such a smutmonger if I wanted to be. Alas, aside from a few failed attempts online, there was no sex to be had. Sorry.  However, there were plenty of flirtations at  urlLink Sidetrack  last night.  I met up with Rafael, my friend who watched my cats while I was in Montreal, for some drinks and showtunes. I also had plans to meet a guy I had been chatting with on AOL for the last few days. So I had plenty of bases covered.  I ran into AOL Guy first, and he offered to buy me a drink. First impressions being what they are, he was attractive, but he didn\'t stop me in my tracks. I greeted him with the customary hug and said I was also looking for a friend and would be back in a minute. He said he\'d wait for my return.  I found Rafael in the adjoining room and told him about the AOL guy. Of course, he accused me of being the \"Second string\" guy to the AOL Guy, but I assured him that he clearly was main reason I was there.  So back to AOL Guy, who by this time had downed a few too many bottles of Miller Lite. He was attractive in a baked-in-the-sun-too-long kind of way, and nice enough. So I accepted his drink (also a Miller Lite, not my favorite, but I wasn\'t going to complain at the prospect of free alcohol) and chatted a while. He introduced me to some of his friends (Thing about me #101 - I am horrible at remembering names.) and watched videos and boys and that was about it.  Eventually Rafael found us and I introduced Rafael to AOL Guy (more to the point, I said \"This is my friend Rafael\" and let Rafael introduce himself. I admit, I forget the guy\'s name. I am so lame.) Also in the group was Richard, another Chorus guy. I sort of wedged myself between Raf & Richard, keeping AOL Guy at bay.  But then he made his move.  First his hand was on my ass... then wrapped around my back, then he was getting all touchy feely in ways that I wasn\'t too comfortable with. Now don\'t get me wrong, normally, I wouldn\'t mind being touchy feeled in public if it was with someone I liked or was really interested in.  However, this quickly was not becoming the case with (Increasingly Drunken) AOL Guy.  Finally he had me from behind, wrapping his arms around my waist and tickling me. I\'d had enough. I was facing Rafael and said \"Help!\"  He didn\'t hear that one, and (I.D.) AOL Guy had me in a Half-Nelson chokehold sort of thing on the floor at this point. I pleaded a little louder, \"HELP!\"  Rafael suggested we go to the roof deck. I gladly obliged.  (I.D.) AOL Guy seemed disappointed, but at that point I didn\'t care. I told him I\'d be right back. (Oops!)  He caught me later on, and looked dejected, slurring his words. By now he was simply Drunken AOL Guy.  Drunken AOL Guy: \"Sscho, what\'sss wrong?\"   Me: \"Well I was just a little uncomfortable. You were a bit too touchy-feely for me.\"   D.A.G.: \"Oh... I\'m sschorry\"   M: \"Don\'t worry about it.\"   *lengthy, uncomfortable pause*  D.A.G.: \"Well, I sschould prob\'ly go ffind my ffriendssss\"   M: \"OK, bye!\"     And that was the last I saw of Drunken AOL Guy.  After I got rid of him, I ran into Terence, who is someone I have always had a crush on (and slept with a few times as well). He was being overly flirtatious. I was being overly receptive. But Raf was around and I didn\'t want to be a complete whore and forget about him, so I controlled myself. And then he went home. Dammit.  Then, I ran into Jonathan.    Backstory:   Jonathan is this studly dude that I met at a party a few months earlier. He\'s Chicago born n\' raised, and I initially thought he was straight, only to find out (to my pleasant surprise) that he was gay after talking to him for a while. We shared a kinship over our mutual friend\'s record collection. (That\'s LP\'s, not CD\'s.) But, just when I was developing the hots for him, he mentions his boyfriend.  DAMMIT.  Anyway, we exchanged Emails and kept in touch for a while, but lost touch after a bit.  A few days ago, he Emailed me again, and said he was moving up near me. I thought that was curious, since I knew he lived with his partner. But I didn\'t pry. He also mentioned a roommate that he hardly knows. I grew even more curious. We agreed to go to a Cubs game together in a couple weeks (Gay Days at Wrigley) so I felt a glimmer of hope.  Then I ran into him last night.  Turns out that he has, indeed, broken up with his (now-ex) boyfriend, who turned out to be a psycho (don\'t they all?) and a loser (speak the truth, brother) and a user (AMEN!) In fact, he just got done moving into his new closet of an apartment and stopped by the bar for a few drinks. And there I was to comfort him.  I know what you\'re thinking.   You: Comfort him means bring him home to have your way with him, you floozy.   Me: Who, ME?   You: Yeah, you... you\'d be online looking for sex anyway, and now it\'s fallen RIGHT IN YOUR LAP (so to speak)!   Me: I think you have me confused with someone else.  I had no such intentions. Honestly. But we did spend the rest of the night together at the bar, talking about bad Ex experiences and sharing life stories. And making catty comments about people and videos playing at the bar. It was fun. And we  are  getting together to see the Cubs game in a couple weeks. So the night was not a bust, at all.  Even though I went home alone and sat online for another 3 hours.  I am a good boy... honest!   '),(4957,3869854,'01,August,2004','2004-08-01 00:00:00','   Early to bed Early to rise Makes a man  A really lame loser   A big fat dork   A brown-noser   A goody-two-shoes   A guy with no life  Healthy, Wealthy and Wise!   '),(4958,3869854,'01,August,2004','2004-08-01 00:00:00','   Oy.  I just woke up.  It\'s nearly 11:45.  I had another one of those \"can\'t-sleep-because-I-need-sex-but-I\'m-not-getting-it-so-why-am-I-wasting-my-time-online\" nights last night.   That coming after a day where my emotional rollercoaster was going higher and doing more hairpin turns and loop-de-loops than ever before.   OK, I was a mess yesterday.   First it was Eric\'s funeral.  That would do it to anyone, right?  It was a nice service, held in the big Catholic church just steps away from Boystown.  I\'d always passed by the place but never gone in... being the good ex-Catholic that I am.  The service was nice, but again, it was oh-so-typically Catholic.  Long.  Lots of standing and sitting.  Communion.  The usual drill.  Communion struck me as odd.  All these gayboys were going up and getting Communion.  I had to wonder how many of them actually do the \"right Catholic thing\"-- go to Confession (now called \"Reconciliation,\") and tell all of their sins to a Priest.  If I had to do that, I\'d keep the poor old guy there for a week.  It\'s not worth it.  Might as well let him eat and sleep and do the things he needs to do, rather than sit there and listen to me blather on.  (And starve to death in the process.)  A few of us joked (predictably) that we were suprised that we weren\'t struck down by lightning when we walked in the church.  Yes, it\'s an old one, but it\'s still effective.  So anyway, the service was nice, the  urlLink Chorus  sang (although we didn\'t sing our first choices-- the songs had to be religious or Biblical in theme or we couldn\'t sing them.  Luckily in our last concert there were two such songs ready to go, or would couldn\'t have sung anything.  Catholics.)  I was touched by the large number of people that came to pay their respects to Eric.  So was his family.  While I was extremely sad, this warmed my heart greatly.  I was honored to have known someone who made so many wonderful friends.   Then I started thinking, \"How well-attended will my funeral be?\"  Boy what a downer.  After the funeral we had lunch at  urlLink Ann Sather\'s , a popular restaurant not far from the church.  The Chorus was supposed to sing again but we decided it would be inappropriate so we didn\'t.  So we paid our final respects and moved on to  urlLink Sidetrack  for the video viewing party for the Chorus\'s spring show, \"Low Hanging Fruit.\"  Talk about a shift in emotion.  Now we were all together and laughing and having a fun time.  But I was still in one of my moods.  I wasn\'t feeling overly social.  I enjoyed myself, but I had no money with which to buy a drink, so I didn\'t have one.  I felt out of place.  I actually thought about leaving a few times, but I stayed until the end.   When I got home, I decided I was \"in for the night.\"  I was supposed to go to a couple more parties (one hosted by old friends of mine that I haven\'t seen in a long time) but I just wasn\'t up for it.  Instead, I made dinner, played with the cats, and sat online.   I\'m still in a funk today.  I need to snap out of this.   I know my friends are probably at the beach.  I should maybe go and see if they are there.  Or maybe not.  See how conflicted I am?  What should I do?   And I still have no money.  I won\'t for another week.  Damn that trip to  urlLink Montreal  for being so damned expensive.  Oh look...  urlLink Manhunt ....   '),(4959,3869854,'13,August,2004','2004-08-13 00:00:00','   Last night, I met another AOL Boy out for drinks at Sidetrack.  As you may recall, I had a disastrous outing with the  urlLink last such AOL-type guy .  This time was not much different. Aside from the drunken part.  This time, the meetee (I being the meet-er) was much younger... 9 years younger than me in fact. Except he should have been 19 years younger. He was young in just about every way possible... as in naive. As in he still lives with his parents. As in he just wasn\'t right for me.  I\'ve been chatting with this boy for probably over 4 years on AOL.  We haven\'t necessarily been the greatest, closest of chat buddies.  I would sign on, he\'d say \"Hi, how are you?\" and I\'d say \"Fine thanks\" and trade a few updates and that\'d be it.  I should have known meeting him would be a bad idea.  When we would chat, he would get frustrated if I didn\'t respond right away, and would get upset if I didn\'t have much to say.  He always would try to pry information out of me (or so I thought.)  To be honest, I found him irritating.  But yet, I was cordial and always nice to him.   I knew he had a crush on me.  I don\'t mean that to sound arrogant or anything... it\'s just a feeling I had.   Apparently I was right.  A few nights ago, he caught me online again.  The same banal conversation ensued, and finally he said, \"When are we going to meet?\"  I must have been tired or weak or whatever.  I said, \"How about Thursday?\"   So we agreed to meet at Sidetrack at 10 PM.  I know what you\'re thinking right now.   \"Rick, how evil of you.  Leading this poor kid on.  You knew you weren\'t interested in him.  Why put him through this?\"   Uh-uh.  I am not evil.  I just wanted to get this over with.  Is that evil?  I don\'t think so.  Maybe he would end up being a sweet guy, and I\'d fall madly in love with him and we\'d live happily ever after.  Or not...  Thursday arrived, and I got home from work on time.  I was tired, still feeling lousy, coughing, sneezing, sniffling, overall feeling miserable.  I didn\'t want to go.   He called at 9:15 PM.  \"What are you doing?\"   \"I\'m making dinner, what are you doing?\"   \"I just got done with work...  I was wondering if you wanted to do something 45 minutes early.\"   \"Uh... no... I am going to make dinner first.\"   \"Oh... ok....\" he sounded dejected already. (Oh great.)   \"I\'ll meet you there between 10 and 10:30.. ok?\"   \"OK, see you there.\"  (OK... I REALLY didn\'t want to go now.)  I finished dinner, got changed and checked a few messages online (Damn that Tribe for being so addictive!)  He called at 10:30 right on the nose.  \"Where are you?\"   \"I was just about to leave...\"   \"OK, I\'m here.\"  I had no choice.  I was committed.  I had to go now.  So I got there and we met, finally.  OK.  He is a nice guy.  But he\'s just not my type.  Like I said, he\'s very young.  He\'s very naive.  Comparatively, I have been around the block... many,  many  times.   It was Comedy Night at Sidetrack.  He had never been before.  (He lives out in Crystal Lake, a very far-away suburb, with his PARENTS.)  I had to explain most of the routines being shown to him.  He either didn\'t get the humor or had never heard of half of them.  (Thankfully he knew Mommie Dearest, or I would have had to have left.  Immediately.)  Then he got touchy-feely.  You may recall  urlLink the last time this happened .  I didn\'t like it then and I didn\'t like it this time either.  In no way was I giving him the signals that I was at all interested in being touchy-felt.  Yet he persisted.  An arm around the shoulder, a touch on the back, a hug at a moment that didn\'t really need a hug.  Standing MUCH too close.   He was like a puppy that craved attention.  And I wasn\'t in the mood for another pet.  Now don\'t get me wrong.  I wasn\'t mean to him or anything.  In fact, I tried to keep the conversation going and keep things as light as possible.  But he must have sensed that I wasn\'t in the mood to be... snuggly.  So he got self-conscious and finally said,  \"Am I making you uncomfortable?\"   \"Well... in a way, yes.  I came here to meet you... but I\'m not comfortable with the touchy-feely stuff.\"   \"OH ok.. I\'m so sorry.\"  He immediately got distant and self-conscious.  I felt bad... for a minute.  And then I decided it was getting late and I had to go home.  So I said good night, gave him a hug goodbye, and left.  In hindsight, maybe I should have been up-front and said  \"Look, I\'m not really looking to date you.  I just am here to meet you as a friend.  I don\'t want to give you the wrong idea.  I hope that\'s ok.\"       But I didn\'t do that... and now I feel a little guilty.  And why is it that these guys that I\'m not so interested in are so touchy-feely but guys that I\'m ga-ga over want nothing to do with me?  Men are so damn confusing.   Of course, I don\'t help that situation any by being so ... me.   ARRRGH.   '),(4960,3869854,'12,August,2004','2004-08-12 00:00:00','   I feel like a steamroller flattened me today.  And yet I still went to work. Bleah.  The weather hasn\'t helped much. It got up to a blistering 63 degrees today. Brrrr! Talk about a chilly August.  I\'m coughing and hacking up lungs left and right. Now ain\'t that a pretty sight? :-)    OOOH I forgot something cool to tell you!    This weekend I\'ll be with the  urlLink Chicago Gay Men\'s Chorus  as we sing the National Anthem at the  urlLink Chicago Cubs  game!  Yup that\'s right, we sing the National Anthem before the Cubs vs. Dodgers game at 12:20 PM CDT on Saturday, August 14 at beautiful  urlLink Wrigley Field .  This will be the second time we will have done this-- the first time was September 27, 2003, at a certain double-header game where the Cubs clinched the division title for the first time in 14 years. And after we sang, we got to see BOTH games! It was an awesome experience. Here\'s a few pics of that glorious event!   urlLink    Lining up to go in and SING! (That\'s me front and center. I was the first one to walk on the field. It was INCREDIBLE.)   urlLink    Oh say can you see....   urlLink    There were 90 of us there-- the most that\'s ever sung the National Anthem at Wrigley. And the FIRST Gay-identified group, ever.   urlLink    Our name up in lights on the Scoreboard!  Pictures posted by   urlLink  Hello       So anyway, if you have cable and get Fox Sports, watch for us... Saturday, August 14 at 12:00 PM CDT. We might just be on TV :) (We were televised last time.)  OK that\'s all the big wonderful news I had to post... back to my regularly scheduled boring life. :-P   '),(4961,3869854,'11,August,2004','2004-08-11 00:00:00','   We interrupt the smell of baking bread with this... the latest in the \"Poop Juice\" saga....   (click headline for the latest)    In case you missed it,  urlLink here  is the original story.  I swear, I\'m not making this up.   '),(4962,3869854,'10,August,2004','2004-08-10 00:00:00','   So I was home sick from work today (really.)  I was hungry, so I went to make some toast.  OH NO!  No bread in the house.  So what\'s a housebound sick boy to do?   urlLink     urlLink    Posted by Hello   Mmmmmm :)  Nothing like freshly baked bread to cure the pangs of hunger. :)  And it was damn tasty too.  Eat your heart out Martha \"Jailbird\" Stewart.   '),(4963,3869854,'09,August,2004','2004-08-09 00:00:00','   The plot thickens... Finally, some REAL news in this city! ;-)   (Click on the title for the story...)   '),(4964,3869854,'09,August,2004','2004-08-09 00:00:00','   I was someone\'s Blog of the Day! Really! Check it out...   urlLink Banshee\'s page   '),(4965,3869854,'09,August,2004','2004-08-09 00:00:00','   Could you imagine? Ew...       Bus waste scuttles boat tour    By Angela Rozas and Brett McNeil   Tribune staff reporters    Published August 9, 2004    At the outset of a Chicago River boat tour Sunday, Capt. Sonja Lund deadpanned that passengers should not look up with their mouths open when passing under the city\'s grated bridges, lest they get a true \"taste of Chicago.\"  About 15 minutes later, passengers got more than that. They got a deluge when the boat was doused with a foul \"liquid substance\" that appeared to drop from one or two charter buses as the boat passed under the Kinzie Street bridge, passengers and tour officials said.  Witnesses said they had just begun the Chicago Architecture Foundation\'s 1 p.m. tour when they passed under the bridge and a cascade of a \"brownish-yellow\" substance rained on them. About two-thirds of the passengers seated on the upper deck of Chicago\'s Little Lady were soaked. Passengers said they looked up to see at least one, possibly two charter buses rumbling above.  A volunteer tour guide at the boat\'s helm reassured passengers that they had been splashed with water. But the foul smell prompted suspicion.There was \"stunned silence initially. Then sort of this horrible realization as they began to smell themselves as to what happened,\" said Steedman Bass, 35, of Boston, who was a passenger on the boat but was not sprayed. \"It was horrific.\"  Passengers shouted that police should be called and demanded the boat be turned around. Within minutes, the tour headed back to the dock at Michigan Avenue and Lower Wacker Drive, speeding as \"fast as the boat would go,\" Lund said.  Some passengers became sick from the smell and made use of the boat\'s lower-deck bathrooms, Bass said.  At the dock, the 120 passengers disembarked and were given refunds for the $25 tickets, exchanges or open-ended tickets for later tours. Some were given cab fare to get back to hotels and some passengers\' dry-cleaning bills will be paid, foundation officials said. The boat\'s crew swabbed the deck clean and set off only slightly delayed for the 3 p.m. tour.  \"We feel very badly for these people who have had this day\'s event ruined,\" said Bastiaan Bouma, vice president of marketing and tours for the foundation. In 50 years of operation, the tour has never experienced something like this, he said.  \"Obviously, this wasn\'t our fault. But we were just trying to make it as comfortable for the passengers as possible.\"  Bouma said that he was told that someone recorded license plate numbers from the bus or buses and reported the plates to police. He said he hoped to know the name of the company by Monday morning and would demand the company\'s buses be inspected for defects.  \"Tomorrow morning we\'ll find out what buses were involved, and hopefully we\'ll get some answers,\" he said.  Police confirmed a report had been made, but said it was not being investigated as a crime.  A local tour bus company official said he had never heard of a bus unloading liquid waste into the river from a bridge. Charter bus companies typically contract with disposal companies to get rid of their waste or dump it at their own disposal sites, he said.  Passengers said they were shocked at the awful timing of the improbable hit.\"If you tried to orchestrate this, it would be an almost impossible scenario,\" Bass said.   Copyright Â© 2004,   urlLink  Chicago Tribune   '),(4966,3869854,'09,August,2004','2004-08-09 00:00:00','   I\'ve found a new talent for myself.  The topic  du jour  on  urlLink Tribe  lately is... \"If They Had A Blog.\"  People are posting names of people they think would have interesting blogs.. if they kept one. Common ones are Salvador Dali, Oscar Wilde, Nelson Mandela (the attention whore, as  urlLink Toddy  says,) Janeane Garafolo, and countless others.  Yet one name caught my eye... and my imagination. (Thanks to  urlLink EViLL ) - Jesus Christ.  What would Jesus\'s blog read like? Fascinating stuff.  EViLL started it off:      \"Woke up. Washed feet. Learned I could walk on water - how cool is that? Wept. Said 5 Hail Marys and she was all like, \"I heard you the first time, Jesus Christ!\"       I found that funny... so I added my $.02      \"Went in the fridge and there was nothing there but some cold fish and a stale loaf of bread. I was freaking out because I had people coming over in an hour. So I just talked to Dad and next thing you know, the whole fridge was full-- all kinds of fish and enough bread to feed a multitude. Had to fill bushels with the extra. It was nuts. Sometimes Dad doesn\'t think about leftovers. Oh well, at least my friends were full.\"      I instantly saw my  urlLink Hell-Score  go up 50 points. And a new inspiration was born. I couldn\'t be stopped. So I decided to ghost-write for a few more \"famous folks\"...   Joan Crawford      \"Got up at 12:30 AM. Started weeding roses. Decided they were in my way, so I chopped them down. Yelled at Christina. (She\'s so insolent.) Drank a fifth of gin to ease the pain. Strangled Christina because she was obstinate. Sent her to boarding school. I love my children, really I do...\"       Bette Davis      \"Woke up, smoked a cigarette, went for a drive in Beverly Hills, went shopping. Smoked another cigarette. Fought with my daughter. Pissed off a few media types, but hell, I don\'t care. Went on Letterman and put him in his place. Still waiting for my next movie deal...\"       Martha Stewart      \"Guards woke me up. I told them a nice pleasant \"Good morning\" is sufficient, but they insist on banging on the rails of my cell. Rude. All we had for breakfast was oatmeal. I wish I could add a sprig of fresh mint at least to make it presentable. These people know nothing about presentation. *sigh*...\"'),(4967,3869854,'08,August,2004','2004-08-08 00:00:00','   I got home around 2:00 from Mom and Dad\'s and quickly got ready.  This weekend was Halsted Street Market Days.  In Chicago, there are three weekends that are considered \"Gay High Holy Days.\" One is IML Weekend (International Mister Leather) which is always Memorial Day weekend. The second is Pride weekend (The last weekend in June) and the third is Halsted Street Market Days.  For those who have never been to this conglomeration of shirtless men and tacky merchandise, Market Days is the largest street festival in the City of Chicago. It occurs every August, usually the first or second weekend. It happens on Halsted Street, between Addison and Belmont. Yes, that\'s right in the heart of Boystown.  For two days, the street is closed off and the show begins. Shirtless men, Hetero and Homo, adorn every nook and cranny of the street. Peddlers from far and wide display their wares-- some cool, most tacky, never cheap. There\'s music stages, featuring artists who are up-and-coming and a few that are definitely has-beens. (This year\'s highlight was  A Flock of Seagulls . Really.)  And then there\'s the food. Chicken on a Stick is popular, as is hand-shaken lemonade, and of course, BEER. Plenty of booze to go around.  Add in a dose of (usually) potent sunshine and steamy temperatures, and you have a weekend of sights, sounds and smells that is quite remarkable.  So I dashed onto the El to make it in time for some prime viewing. This is people-watching central folks. The men are something to see. It\'s as if they come out of the woodwork for this event, proudly displaying a summer\'s worth of tanning at the beach and action at the gym. You can\'t sneeze without seeing another bronzed specimen waddling past, his arms too big to hold down to his sides. Oh yes, this is definitely another parade.  I ran into a few friends of mine and then made it to the booth for the  urlLink Chorus . This is the Chorus\'s prime recruiting time. We hawk new members, season subscriptions, potential donors, and people to add to our mailing list at Market Days. Without this, it\'d be hard to continue adding  fresh meat  new members each year.  So my initial reason for going was to help out at the booth. Yet upon arriving, I could see that things were well under control. So I stuck with my friends and we proceeded to drink, eat, and be merry with everyone else.  What\'s sometimes disturbing (and amusing) is the lack of control people have over their alcohol consumption. It\'s my theory that alcohol and hot sunny sweltery days do not mix. People are stumbling over themselves, throwing up, and getting into fights left and right. It\'s amusing in a repulsive sort of way. Some people just don\'t know their limits.  One particular gentlemen accosted us verbally as we were leaving the festival. We were standing on a corner waiting for the light to change, when he slurred \"Get outta my way... move it, please\"  We didn\'t say anything. To him.  The light changed. We started to cross the street.  \"Yeah, Go Kerry. Get that fucking Bush outta office\" he blurted, after seeing one of us wearing a \"Kerry for President\" sticker.  \"Yes, that\'s right, ok, thank you sir,\" we said, as we continued to walk past him.  \"Yeah you fuckers, better vote. Yeah...\" he kept on going.  I turned around and shouted, \"Don\'t drink and walk!\"  We exploded in hysterical laughter.   '),(4968,3869854,'16,August,2004','2004-08-16 00:00:00','   Cubs Game    The Cubs game went off just fine.  We had over 70 guys this time, and once again made an amazing impression on the fans and on the Cubs organization.  Pics are coming soon, I\'m just too lazy to get them all formatted and ready for posting.     The only thing that sucked about it was... well.. two things:  1. They announced us... and showed our name on the scoreboard... as the Chicago Gay Men\'s CHOIR.  (As opposed to CHORUS.)  2.  We weren\'t on TV or the radio.  BUMMER.  Now mind you, the Choir/Chorus thing doesn\'t COMPLETELY suck... because: a.  They at least announced as and showed our name as Chicago  GAY MEN\'S  Choir.  That is what\'s important. b.  Nobody booed, nobody sneered.  But EVERYBODY cheered!   It was a gorgeous day, and the Cubs won AGAIN.  So I think we have to consider ourselves lucky... and good luck to the Cubs.  Can\'t wait til we do it again next year!   - - - - - - - - - - - - - - - - - - - - - -    More Strange Happenings in the Online World   I had THE most bizarre chat with someone today.  I have no idea who this person even was.  But it was just... WEIRD.  Read on:    Strange Guy:  hey       Unsuspecting Me:  Hey        SG:  whats up       UM:  Not much       SG:  koolio       SG:  do you have a kitty?       UM:  Yeah, two       SG:  aww i have 1       UM:  Cool       SG:  do you have any other pets       UM:  Nope       SG:  kool  i have 21 rats       UM:  (Inside Voice) [What the  FUCK!?!? ]  (Outside Voice) wow       SG:  yeah i had a suprise litter....a huge litter lol       SG:  they look like little puppy dogs       SG:  the grandma rat has a crush on my cat and chases him around   And that is where the chat -- with this bizarre, freaky, wacko guy that should be in a movie with the theme song sung by a young Michael Jackson-- ended.  Hello Twilight Zone.  Yikes.     - - - - - - - - - - - - - - - - - - - - - -   The Summer of Sad News Continues      We got the news today that another Chorus member has passed away.        Bill wasn\'t a close friend of mine or anything... he\'d been a member for many years.  But he never looked very well.  He hadn\'t been around much lately either.  So the news of his passing didn\'t come as a huge surprise.  Still, after  urlLink Eric\'s  passing so recently, it is going to be a somber reconvening of the Chorus come September.       - - - - - - - - - - - - - - - - - - - - - -    Random Ramblings         â€¢  I just love the   urlLink  Olympics   .  All those hot studly little short gymnastic men.  All those swimmers in their short shorts and bikini bottoms.  Why it\'s enough to make a boy get ... a little hot & bothered... ;)      â€¢  I\'m meeting with fellow bloggers   urlLink  Palochi   ,   urlLink  NoFo    and   urlLink  Andymatic    on Monday at Sidetrack for a few drinks \'n Showtunes.  I already know Palo and NoFo but I haven\'t met Andy yet.  It\'s the first \"official\"   urlLink  gaybloggers.tribe.net    Chicago meeting since I\'ve been a member there.  And the first since Andy moved to Chicago... should be a fun time!      â€¢  I\'m getting ready to go camping next weekend.  I still haven\'t recovered financially from the   urlLink  Montreal    trip... but I need another getaway.        â€¢  I had a wonderful phone conversation with my friend Ellizzette yesterday.  She is always such a joy to talk to.  If you\'ve read my old journals,   urlLink  you will find her in there   ... she used to live in Chicago but now lives in NYC.  Since she\'s been gone we probably have become even closer friends... although I miss seeing her in person.  Oh the memories I have of sitting up and watching TV (usually Frasier reruns) and eating pizza or hot cocoa.  Good good memories. :)  I need to go see her.  Very soon.      â€¢  It\'s still only in the 50\'s here at night.  I fear that autumn has come early to Chicago.  Where have our summers gone?      â€¢  I so do not want to go to work tomorrow... but oh well... bills must be paid.  Sigh.   '),(4969,4077627,'17,August,2004','2004-08-17 00:00:00','   Don\'t know where it originated, but recieved it from my lil sister this morning.   Hello,  my name is Lewis and I suffer from the guilt of not forwarding 50 billion  fucking chain letters sent to me by people who actually believe that if you  send them on, a poor 6-year-old girl in Arkansas with a breast on her  forehead will be able to raise enough money to have it removed before her  redneck parents sell her to a traveling freak show. Do you honestly  believe that Bill Gates is going to give you, and everyone to whom you send  \"his\" email, $1000?   How stupid are we?   \"Ooooh, looky here! If I  scroll down this page and make a wish, I\'ll get laid by a model I just happen  to run into the next day!\"   What a bunch of bullshit.   Maybe the  evil chain letter leprechauns will come into my house and sodomize me in my  sleep for not continuing a chain letter that was started by Peter in 5 AD and  brought to this country by midget pilgrims on the Mayflower.  Fuck  \'em.   If you\'re going to forward something, at least send me something  mildly amusing. I\'ve seen all the \"send this to 10 of your closest friends,  and this poor, wretched excuse for a human being will somehow receive a  nickel from some omniscient being\" forwards about 90 times.   I don\'t  fucking care.  Show a little intelligence and think about what you\'re  actually contributing to by sending out these forwards. Chances are, it\'s our  own unpopularity.  The point being? If you get some chain letter that\'s  threatening to leave you shagless or luckless for the rest of your life,  delete it. If it\'s funny, send it on.   Don\'t piss people off by  making them feel guilty about a leper in Botswana with no teeth who has been  tied to the ass of a dead elephant for 27 years and whose only salvation is  the 5 cents per letter he\'ll receive if you forward this email.   Now  forward this to everyone you know.   Have a nice  day.  '),(4970,4077627,'14,August,2004','2004-08-14 00:00:00','   From This Modern World   As you read this, keep in mind that Bush can\'t win without Florida, and that elderly blacks in Florida are overwhelmingly Democratic:  Bob Herbert, in today\'s NYT, informs us that armed police officers have been showing up at the homes of elderly black voters in Orlando, many of them active in get-out-the-vote activities, and scaring the crap out of them.  The \"investigation\" (which officials refuse to discuss, or course) is explained by a lawyer representing the 73-year-old president of the Orlando League of Voters, one of the recipients of this fair-and-balanced treatment:   The fear generated by state police officers going into people\'s homes as part of an ongoing criminal investigation related to voting is threatening to undo much of the good work of the league. He said, \"One woman asked me, \'Am I going to go to jail now because I voted by absentee ballot?\' \"  According to Mr. Egan, \"People who have voted by absentee ballot for years are refusing to allow campaign workers to come to their homes. And volunteers who have participated for years in assisting people, particularly the elderly or handicapped, are scared and don\'t want to risk a criminal investigation.\"   Is there anything the Bush machine won\'t stoop to?   These people aren\'t just unfit for office; they\'re unfit for basic citizenship.  '),(4971,4077627,'14,August,2004','2004-08-14 00:00:00','   Want to know why your prescription drugs cost so much? I\'ll tell you why, Ronald Reagan. -------------------------------------------------------------   Beginning in 1980, Congress enacted a series of laws designed to speed the translation of tax-supported basic research into useful new products &#8211; a process sometimes referred to as \"technology transfer.\" The goal was also to improve the position of American-owned high-tech businesses in world markets. The most important of these laws is known as the Bayh-Dole Act, after its chief sponsors, Senator Birch Bayh (D-Ind.) and Senator Robert Dole (R-Kans.). Bayh-Dole enabled universities and small businesses to patent discoveries emanating from research sponsored by the National Institutes of Health (NIH), the major distributor of tax dollars for medical research, and then to grant exclusive licenses to drug companies.    Yes, that\'s right drug companies are getting patents and exclusive licenses on discoveries paid for with our tax dollars. Where is my royalty check?    Until then, taxpayer-financed discoveries were in the public domain, available to any company that wanted to use them. But now universities, where most NIH-sponsored work is carried out, can patent and license their discoveries, and charge royalties. Similar legislation permitted the NIH itself to enter into deals with drug companies that would directly transfer NIH discoveries to industry.   Does this sound a little like \"conflict of interest\" to anyone else?    The most startling fact about 2002 is that the combined profits for the ten drug companies in the Fortune 500 ($35.9 billion) were more than the profits for all the other 490 businesses put together ($33.7 billion).   There is something wrong here, what the fuck were they thinking?  '),(4972,4077627,'13,August,2004','2004-08-13 00:00:00','   Thanks to Les from  urlLink SEB          urlLink What Famous Leader Are You?    urlLink personality tests by similarminds.com'),(4973,4077627,'13,August,2004','2004-08-13 00:00:00','   After claiming he has no political affiliation with the GOP it turns out that O\'Neill has made nearly $15,000 in contributions to federal races -- all Republican   Appearing on CNN\'s Crossfire on August 12, John O\'Neill, co-founder of the anti-Kerry group Swift Boat Veterans for Truth and co-author of the Regnery book Unfit for Command: Swift Boat Veterans Speak Out Against John Kerry, denied he had links to the Republican Party: \"I\'ve had no serious involvement in politics of any kind in over 32 years.\"  But, as Media Matters for America noted on May 4, O\'Neill has long-standing ties to the Republicans, having worked under the direction of the Nixon White House and clerked for Chief Justice of the United States William Rehnquist in 1990.   Moreover, since 1990, according to the Center for Responsive Politics, O\'Neill has contributed $14,650 to federal candidates or national political organizations -- all Republicans:  2004: $2,000 to Duane Sand (ND)   1999: $1,000 to Peter Staub Wareing (TX)   1998: $250 to Rudy Izzard (TX)   1996: $1,000 to Brent Perry (TX)   1994: $2,500 to Texas Republican Congressional Committee  1993: $2,500 to Texas Republican Congressional Committee   1992: $1,000 to Texas Republican Congressional Committee  1992: $1,000 George H.W. Bush  1992: $1,000 to Clark Kent Ervin (TX)   1991: $1,000 to Clark Kent Ervin (TX)   1990: $400 to Hugh Dunham Shine (TX)   1990: $1,000 to A Tribute To Ronald Reagan  '),(4974,4077627,'13,August,2004','2004-08-13 00:00:00','   WASHINGTON (Reuters) - One-third of President Bush\'s tax cuts have gone to the wealthiest 1 percent of Americans, shifting more burden to middle-income taxpayers, congressional analysts said on Friday.   The report by the nonpartisan Congressional Budget Office and calculations by congressional Democrats based on the CBO findings fueled the debate over the cuts between Bush and his Democratic challenger in November, Sen. John Kerry.   Using the CBO\'s figures, Democrats in Congress said the top 1 percent, with incomes averaging $1.2 million per year, will receive an average tax cut of $78,460 this year, and have seen their share of the total tax burden fall roughly 2 percentage points to 20.1 percent.   In contrast, the report showed that households in the middle 20 percent, with incomes averaging $57,000 per year, will receive an average cut of $1,090 while their share of the tax burden would move to 10.5 percent from 10.4 percent.   The CBO report said about two-thirds of the benefits from the cuts went to households in the top 20 percent, with an average income of $203,740.   People with earnings in the lowest 20 percent, which averaged $16,620, saw their effective tax rate fall to 5.2 percent from 6.7 percent, the CBO said. But Democrats said that meant their average tax cut was only $250.   Democrats said the CBO calculations, which they requested, confirm the view of independent tax analysts that the tax cuts enacted in 2001 and 2003 have heavily favored the wealthiest taxpayers.   What will he do for us next?  '),(4975,4077627,'13,August,2004','2004-08-13 00:00:00','   I\'ve had enough trying to keep my own blog going on my web server so we will give blogspot a try and see what happens. You can reach my blog here on blogspot at impissed.blogspot.com or at blog.timlewellyn.com they are one in the same.  I am pissed about a lot of things at the moment, mostly about political issues and the pending election, (if we have one). I am also pissed about religious issues that seem to be gaining popularity on the right and their impact on the rest of us.  So  feel free express your opinions and by all means rant if you feel the need.  Thanks  Tim  '),(4976,3839683,'26,July,2004','2004-07-26 00:00:00','    This post may offend some people so I will try my hardest to be politically correct on this one. Yesterday I was working.. On Sundays the mall closes at 6pm so I was all hyped when 5:30pm rolled around and I could almost get my ass out of there. Then walks in a mom, daughter and a son. The son is what you would call \"mentally challenged\". He\'d stand around \"dancing\" to the music on the radio while wearing his fanny pack too high and staring at me. Perhaps like we were the only ones there... I really don\'t know. I\'d look over and he would just give me a blank stare back. You know when you feel like someone is looking at you? Well usually when you look at them they look away.. not the case here. It was kind of getting uncomfortable so I would change the way I was standing like lean against the counter or squat down like I was picking something up on the carpet. But when I\'d pop back up there he was still. The mom was pestering him across our tiny store by saying \"oh I\'m going to buy you these Oscar the grouch socks cause your grouchy like him.\" This annoyed him a lot and he would yell back \"mom that\'s kidâ€™s stuff! I\'m over the age of 21!!!\" I kind of had to giggle to myself at that one.. I can keep my cool though so I just acted like I wasn\'t interested in their convo. That was until his next comment when his mom said \"you like Oscar don\'t you?\" to where he replied \"Fu*ck yes!!!!\" I about died when that came out of his mouth... Jesus H Christ that had to be one of the funniest things I have heard a \"mentally challenged\" person say... EVER. I think it even replaced the  \"have you seen my baseball?\" or \"frank and beans!!\" comment on Something About Mary. After saying that he bashfully looked over at me.. just like every other time he blew him mom off. I couldn\'t get them out of my store fast enough. So I came home last night and told my buddy Davis about it and he came up with a pretty cool idea. You see I have trouble a lot with \"mentally challenged\" people liking me.... It doesn\'t matter if they are male or female it\'s like I am a big shiny piece of metal that attracts them. So Davis and I are going to market shirts that say \"tart magnet\" instead of the ones you see floating around that say \"chick magnet\" with a baby chickadee on them. Instead our shirts will have a sweet tart candy on it.. Just to be politically correct of course.   '),(4977,3839683,'25,July,2004','2004-07-25 00:00:00','    In 1981 (the year you were born)  Ronald Reagan becomes president of the US  Minutes after Ronald Reagan becomes president, Iran releases 52 American hostages that had been held captive for 444 days  President Reagan is shot in the chest by would be assassin John Hinkley  Space shuttle Columbia, the world\'s first reusable spacecraft, is sent into space  A suspended walkway in the Kansas City Hyatt Regency Hotel collapses, killing 113  A female former lover files a \"palimony\" suit against tennis star Billie Jean King  President Reagan appoints Sandra Day O\'Connor to be the first woman on the Supreme Court  The Pac-Man video game is introduced in the US  Alicia Keys, Elijah Wood, Kelly Rowland, Anna Kournikova, and Britney Spears are born  Los Angeles Dodgers win the World Series  Oakland Raiders win Superbowl XV  New York Islanders win the Stanley Cup  Raiders of the Lost Ark is the top grossing film  \"Bette Davis Eyes\" by Kim Carnes spends the most time at the top of the US charts  MTV debuts on cable television, playing music videos 24 hours a day  Gimme a Break premieres    urlLink   '),(4978,3839683,'22,July,2004','2004-07-22 00:00:00','    Four days bat free and counting yet I\'m still scared to go to sleep before dawn. Yes I know I\'m a whiney little bitch. Yahoo pool is keeping me pretty busy so it\'s all good. I really hate playing it sometimes though. It seems like people would rather talk then play. For Christ sakes I don\'t go there to make friends. A/S/L? Has to be the most annoying thing I\'ve ever been asked or seen. Last night this 12-year-old kid asked me if I was a virgin... I decided to be nice for once and answer his question with \"I\'m married what do you think?\" Where he then proceeded to ask me if I would have sex with him. Now I don\'t know if this was a sting operation by cops or something but I had to chuckle to myself. Bless his little 12-year-old horny heart. Things sure have changed from when I was 12 which was about 11 years ago. I don\'t even recall the thoughts that I had back then but at least I knew that my biggest concern was playing with my Barbie\'s. Kids scare me now a days. I\'ve told Rob that if we were ever expecting a baby and it was a girl.. I would make her go back in until she came out a he.  My cousin is 15 years old and she\'s part of the \"I\'m a popular kid that will do anything to make people think I\'m cool\" club. This club includes hanging out with 18 year olds and drinking beer until you throw up all over yourself. She\'s always been hell on wheels as long as I can remember. Whether it was cussing at the age of 6 or locking you out of the house while baby sitting her. I honestly believe she was the reason birth control was invented. Don\'t get me wrong kids can be fun too. Like all the times that we fed my youngest cousin, \"lil dick\" AKA Richard, a dog biscuit. I still can\'t get over the fact that when he threw up.. more then one bone was visible. Or even the time when we shot him with a paint ball gun and he ended up pissing his pants over it. Even the dragging behind a 4 wheeler on a sled and kicking dirt in his face was good times. If that isn\'t ground enough to have a boy then I don\'t know what is.   '),(4979,3839683,'18,July,2004','2004-07-18 00:00:00','    Hey hey hey so what\'s up wit jew? Nothing here really.. maybe that\'s the problem, eh? So I was sitting back lately thinking that i should take stock in AT&amp;T.  I spend about $400.00 a month on recharging phone cards through them. This is so Rob can call me from Iraq.  It\'s crazy cause I will buy 2600 sum minutes at a time and it will convert to only 500 minutes.  Big bummer indeed. The sad part is I have been spending most of the time we talk lately on nonsense such as the problem I have been having back here at home.  Bat\'s have been finding their way into my home after dark. My house is only 10 years old but it has managed to give me carbon monoxide poisoning twice and infested my room with bats. I don\'t know what kind of sick joke that is being played on me or who is trying to disturb me but I feel like i\'m trapped in my own personal final destination.  My dad is \"trying\" to get rid of these bats but i\'m really not seeing any progress in it.  Let\'s just say my lack of sleep isn\'t going to win me a beauty contest anytime soon. I\'m trying to switch my sleep pattern into day time.. but i\'m missing too much and my body doesn\'t want to cooperate with it.  I\'m sick of hearing \"it won\'t hurt you\" from people. LIES LIES LIES!!!! I\'ll tell you one thing though if i wake up looking like bat girl my life is seriously going to take a social dive... even though it\'s nonexistant already...   '),(4980,3839683,'13,July,2004','2004-07-13 00:00:00','    This summer has been pretty boring so far. I don\'t know what it is but time is dragging by really slow. Possibly it\'s the fact that at the end of summer rob gets to come home for 15 days. It\'s something to look forward to but seems really far away still. I haven\'t really done half the things that I had wanted to this summer and its already the middle of July. Oh well there is always next year right? I went swimming yesterday.. Well actually it was just suppose to be a boat ride down on the lake. It turned into me having to swim behind the boat while my brother laughed at me and kept me from getting in. I really didn\'t mind though since it was really hot out. I can\'t imagine being over in Iraq right now with 120+ temps. The thought makes my face want to melt. We are experiences a severe thunderstorm right now so it\'s pretty nice to get a cool down.  The most exciting thing about my summer so far is that I downloaded the new yahoo messenger and have been jamming to the yahoo launch music on it. It\'s pretty cool... You can listen to almost anything on it. Wow I really need a life. I wish I didn\'t live so far into the country..the ice cream man could visit me.   '),(4981,3839683,'09,July,2004','2004-07-09 00:00:00','    I worked all day today and left at around 5:30pm. Today was pretty stressful. Kids sure have gotten to be little punks lately. No please or thanks you... Shame shame shame. Anyways I drove home as usually and once I got home I realized there was no food in the house so I ventured out to the grocery store. As I started to back out of my driveway I looked in my rear view mirror and noticed something was placed behind my back windshield wiper. I got out of my car to see what it was and the first thing I saw was \"god loves you!!!!\" I sighed.... Now I believe in the dude but... This is..Well not something I wanted to see attached to my car after driving 20-25 mins home in rush hour traffic. I could only imagine what the people behind me thought as I was blaring the \"naughty words\" music and having \"god loves you!!!\" plastered on the back of my car. I\'m mad at the man above anyways.. This sign could mean one of two things. First would be that this was a sign to stop my pouting. Or that religious freaks get a discount on paper products in my town. Either way I\'m gonna have a few beers and relax. Maybe he will join me... I think he deserves a break.   '),(4982,3839683,'07,July,2004','2004-07-07 00:00:00','    So someone stole my yellow ribbon off of the back of my Jeep at the mall. It really meant a lot to me too. Makes me feel pretty sad right now. Rob bought it for me from his commander one day at work and brought it home to me. It\'s a magnetic yellow ribbon that has the first cav patch on one side of the ribbon. People can be such asses sometimes you know? He was mad when I told him about it and I feel really bad about it too. It\'s not like anyone in this town would have known what it meant anyways since there is no army bases anywhere near me. There is a national guard though but nothing with first cav, so I don\'t know. I just know that if and when I find it on someone\'s car I am taking it back and giving them a piece of my mind. The lack of respect that a person has for another is unbelievable. Especially when the person has a family member that protects you everyday, it is so sad in this country. How could someone be so low? People have no clue what so ever as to what soldiers and their families go through... So one could only imagine how much that ribbon meant to me...   '),(4983,3839683,'06,July,2004','2004-07-06 00:00:00','    Not much happened today so far. Just the usual day off routine of waking up and watching Maury Povich. Yes I know the whole show is a pile of crap but it\'s such entertaining crap at  that! Unfortunately it was one of the episodes of people over feeding their babies. I think the funniest part of the whole thing is that when they bring them out onto the stage, all 60 lbs of their 9 month old frame, they are wearing diapers. So not only are they really big but they look like little sumo wrestlers. I\'d like to send these parents a newsflah while watching it. It\'s something I like to call \"obvious\". How bout you don\'t feed your kid steaks and 4 grilled cheese sandwiches in one sitting. Also lets not fill a baby bottle full of root beer. One of the main problems the parents have isn\'t the size but the fact that their child stops breathing in the middle of the night. I\'ve said it once and I\'ll say it a million times more.. Even I could make a better parent then parents. I hang my head in shame knowing that this is what America is.   This reminds me of my favorite thing by the way. It\'s called Walmart. I know that they have been in the news lately for their sexes acts with employees but that isn\'t what I am talking about. My main thing with Walmart is the people who walk the aisles in it. You walk into the place and all you see is barefoot children walking over broken glass or running around screaming with a permenate kool aid mustache planted on their dirty faces. Would it kill you to rub a wet one over their faces before they get out of the car? It\'s like Walmart is the breading grounds of the local inbred in my town. Actually it\'s not MY town it\'s the small towns that surround my metropolis. Here\'s a word of advice though. Next time your kid sneezes on me I\'m gonna go out to your pick up and steal your dale Earnhart bumper sticker! I\'m not saying that they shouldn\'t be allowed to shop it\'s just that maybe you could clean yourself and your children up before Ma and Pa leave the house to buy food for the family hoe down. But they sure do know how to \"roll back\" prices eh? *thumbs up*.... *sigh* will American ever get better?   '),(4984,3839683,'05,July,2004','2004-07-05 00:00:00','    So I\'m not impressed with the postal system. For some reason they think that they can take the day AFTER a holiday off. I hate that... The postal system has been so wacky this year after taking Reagans burial day off too. He was a great man but I\'m sure good ole Ronnie would have wanted me to get my new US weekly magazine.. Cause life does go on don\'tcha know? I don\'t know why so many postal employees go crazy working there anyways. What\'s so horrible about selling cat stamps to old ladies, driving on the wrong side of a car and having every holiday off?  I actually enjoy going to my mail box and finding letters that I never usually get! It\'s a sad story yet it\'s very true. My pathetic 23 year old life lives for the daily mail. But that part of joy in my life was taken away on July 5th 2004. It all started like any other day. Me walking out the door around noon and having to drop a few letters off in the mailbox expecting them to not be there when I got home. I was wrong like usual. This is when I realized that the postal system workers have a better life then I do. I had to work today and they didn\'t, it\'s as simple as that. Yet I\'m sure they got paid today for it...And I get paid less then them... I am almost positive. That makes baby jesus and I cry....  The bank people make me mad too come to think about it. They also had this Non holiday day off too. Also they get to touch thousands of dollars a day and get paid for it. I think if I worked at a bank I would sit there all day and smell the money. That\'s just me though. I would get fired for holding on to it too long probably too. Maybe get caught in the bathroom  just \"touching\" the money all day long. I would be on your 6 and 11 o\'clock news.. and you\'d say.. \"nothing wrong with that, I would touch money too.\" Yet the sad truth is that money doesn\'t stay long enough in my custody without saying byebye!  I believe the bank in my town already thinks that I am some sort of drug dealer. You see when Rob gets paid the only way I can get access to the money to pay bills is threw an ATM because the bank the army deposits our money into is a Texas bank. So the ATM likes to spit out 20 dollar bills only. Let me tell you sending $400 dollars in 20 dollar bills ($400 a day bank card limit) threw the drivethru bank window at a time is not my idea of fun. I could only imagine what those people are thinking. Perhaps one day the local Narcotics team will show up at my door and give me a much needed vacation away from my life. I\'ve always wanted to learn how to pee standing up!   '),(4985,3839683,'04,July,2004','2004-07-04 00:00:00','    Wow I guess I\'m starting my blog off on the 4th of July. I had one of these on Xanga but have more friends on here anyways so I decided to do the switch over. Well I guess I should start off by saying Happy 4th! Thank you to all who protect us everyday...being in the military family now makes me open my eyes more and realize what it takes to make sacrifices. Hope your day is going well... I have ups and downs about today but push through like any other day.   Perhaps I should introduce myself to those who do not know me. My name is Rachel I\'m  23 years old, into punk music, alt music, some 80\'s music, Detroit tigers baseball, and UofM football. My sweetie is in Iraq right now in the Army\'s 1st cav out of Fort Hood in Texas. He is a mechanic and I couldn\'t be any prouder of his accomplishments and what he has done for himself. I\'m currently living with my family in Michigan while he is deployed.  You will see me talk about him a lot since that is all that is going on in my life right now. Also this will serve as a journal to him to see how I am doing while he is gone. To cap it all off I try to stay busy by finishing up my dental assisting degree and I am 6 months away from getting it. That\'s about all that\'s interesting or worth talking about right now.    I\'m pretty excited about today because Rob getting his combat patch sown on. I know he has been waiting even before he went over there to get one. It\'s what most military guys dream of. I just talked to him a few minutes ago on yahoo messenger and he is a bit tired from pulling a 14 hour guard duty but that\'s all I seemed to really get out of him. I\'m used to it though cause it\'s pretty strenuous over there.     I don\'t know if I will be going to the fireworks tonight. In my town we have this huge festival called the cherry festival.. And it\'s just that.. All about cherries. There are tons of people here.. More then you would ever want to come in contact with in a town this size. There\'s a few good movies on tonight so maybe I will just stay in and catch those. There\'s nothing like a good 80\'s movie like The Karate Kid or Top Gun. The only thing that would make this day better is Peewees Big Adventure after those 2 movies... I don\'t care what anyone says... peewee herman is my favorite still... All guys do what he did.. He just got caught and that\'s sad! heh. ahhh the good ole\' days... How I miss them... Well welcome to my blog if after my peewee herman confession you are still interested to read!   '),(4986,3839683,'06,August,2004','2004-08-06 00:00:00','    Yo.. Life has been going slow lately. I don\'t know what it is... maybe because I actually have things to do in September and I am counting down till then. It always seems that when you have something exciting going on in your life it takes forever to get there. My little cousin is participating in the JR Miss Michigan competition next weekend so that is a bit exciting. She must have got all the good looks in the family! heh heh.. oh wait that\'s kinda sad. So anyways I didn\'t say anything before but Rob and I are having trouble with a yahoo group. It\'s some group on yahoo that writes letters to troops. Yes the idea is nice and all but somehow they got a hold of his address over in Iraq. Looking in from the outside you might think that this is ok but Rob was a bit pissed about this. He received a letter from some girl and immediately came back to me and asked who the hell it was thinking I gave his address out. But of course I didn\'t give it to anyone. She gave us a little lead by saying she was from this group on yahoo. Well he wanted me to get them to pull his name off of the site. Although it was nice that they were sending letters to soldiers he did not want any from people he didn\'t know. That\'s just the way he is... he\'s a pretty private person and only wants letters from friends and family. Which is understandable to me. So I kindly wrote the person who ran the site and asked them to pull his name off of it and if they could please give me the name of the person who gave the address so we could get them to stop spreading it. Seems nice enough don\'t you think? Well they went into this big thing about soldiers having privacy and they don\'t know if I in fact knew him and they wouldn\'t remove the name for me. After this happened Rob started to email them and they still hit us with the privacy issue.. although we both commented that it\'s against his privacy for them to have his address on there without his permission. Anywho this yahoo group of about 23 memberâ€™s think that they are the cats meow. Like I said before it\'s cool what they do for other soldiers but when it comes to common sense they have none. Well this girl writes a final letter that he received today saying that I am some enraged jealous girl and I\'m a bad person cause I won\'t let them write to him and she\'s going to stop writing. We both had a good laugh at that because I wasn\'t the one who had the problem, it was him. Anyone who knows me knows that I wouldnâ€™t attack someone for doing something as simple as writing a letter. He actually thought that I was too nice to them. Now this little group thinks that they are going to reach his commander and ok if he is a real soldier...Like a commander in the Army has nothing better to do in a time of war then to make these little girls happy. *sigh*   '),(4987,3938699,'29,July,2004','2004-07-29 00:00:00','   Once again, this is another post which Jake will hate.&nbsp; It is simply the retelling of a regular&nbsp;day I had.&nbsp;   6:45 A.M., Wednesday, July the 28th, 2004.&nbsp; My alarm sounds, followed by me hitting the snooze three times.&nbsp; 7:15 I wake up and get showered.&nbsp; After jumping in my car and pulling out of my driveway, I begin my morning prayer.&nbsp; Just as I get ready to turn onto eldridge, I remember I forgot my social security card.&nbsp; I U-Turn and pick that up with very little time to spare.&nbsp; God Answers all prayers.  I make it to Nesbits in ample time and begin work for that day.&nbsp; Nothing particulalry interesting happened that day at work.&nbsp; Outside of having a short, mexican lady come and tell me why I\'m stupid.&nbsp; I hate people who never graduated high school tell me I\'m stupid.&nbsp; There\'s just something wrong with that.&nbsp; Nesbits is not to bad.&nbsp; I can actually go to the restroom or sit down every so often, much unlike Rib-Tickler once was.&nbsp; I rue the days I worked there!&nbsp; After six hours of work, I make my way home and begin preperation for the bible study I would be leading in about 5 hours.&nbsp;   The \'preperation\' that lasted from 2:15-4:45 was mostly me staring at my computer screen and then a bible.&nbsp; Followed by about an hour and a half of watching television.&nbsp; I print the bible study sheet out, jump in my car and drive towards Prince of Peace.&nbsp; Already there is Matt R. and his wife Stephanie with their baby girl of 3 months, Mary Clare Regitz.&nbsp; As I tend to do around kittens and cute babies, I stammered around and expelled some giberish.&nbsp; Smith showed up and we sat down to practice.&nbsp; Thankfully enough, I was able to neutralize a majority of the monopoly he originally had over our bible study (which was originally MY idea to begin with).&nbsp; I sat him down and told him how things were going to be.&nbsp; \'I do this, you do this.&nbsp; Got it?&nbsp; Good.\'&nbsp; I made a few copies, and he wrote some \'burdens\' down on popsicle sticks for the bible study.&nbsp; We go to church and say a quick prayer, then he starts practicing for Praise n\' worship.&nbsp; I go and bug Matt for a little bit, followed by him kicking me out of his office and into the youth room to start PnW.&nbsp;   For praise n worship, I sat in the front corner and nodded my head to the music.&nbsp; I felt overly charismatic that night!&nbsp; As people begin entering the room, my head looks for her.&nbsp; The one person I actually wanted to see that night, the one who I wanted to see me.&nbsp; A while ago, I promised myself that I would finally confess that I have feelings for her the next time I saw her.&nbsp; After thirty minutes and her not being there, I came to the obvious truth that she wouldn\'t be there.&nbsp; Praise n Worship ends, and bible study begins.&nbsp; I tell a crappy joke, and then introduce our topic for tonight.&nbsp; \'What burdens us from God\'?&nbsp; The bible study started off rather smoothly, but eventually picked up after Smith and I got the hang of it.&nbsp; Just as we finished our discussion on the first verse, she entered the room.&nbsp; Smelly, unkempt, wearing a dirty&nbsp;PoP t-shirt and nasty shorts.&nbsp; Beautiful.&nbsp; It took me about five seconds to draw my attention back to the bible study.&nbsp; I was enthralled.&nbsp;   Smith begins his popsicle stick plan... Disaster.&nbsp; The thing was a debachery and poorly executed.&nbsp; I looked to Matt and told him I wish we had thought about splitting into groups beforehand.&nbsp; Oh well, it wasn\'t a complete failure.&nbsp;&nbsp;I felt like going around to see how the small groups were interacting.&nbsp; Yeah, I basically went to see her and that was it.&nbsp; Was it very obvious?&nbsp; Probably.&nbsp;   Finally, we had the people there write down their burdens, and put them on nails in a big cross we have.&nbsp; It was hilarious though, having people come up there and have trouble putting their paper on the nails.&nbsp; Half of them ripped.&nbsp; We ended quietly with Dave playin some music.&nbsp; I\'d say the whole event was a success.&nbsp; Matt wanted to know if we\'d ever do it again.&nbsp; I said if he ever wanted us to.&nbsp;   I went to go talk to her as soon as I could.&nbsp;&nbsp;I needed to talk to her, to hug her, to be with her.&nbsp; We stayed and talked for a good 5-10 minutes.&nbsp; Many a time did she mention her boyfriend though.&nbsp; Not necissarily what I wanted to hear, but at least it was her who was speaking.&nbsp; I enjoyed the evening well enough.&nbsp;   I gave her a hug goodbye and she was gone.&nbsp; Didn\'t tell her I had feelings for her.&nbsp; Just said i \'might\' have sent her&nbsp;an email.&nbsp; Didn\'t tell her it was an email I made just for her so I could send her one email.&nbsp;   Stephanie had left with Mary Clare earlier so Matt was without a car.&nbsp; He asked me if I could take him home, and seeing as how he did the day before we started bombing Iraq, I was more than happy too.&nbsp; Plus he lived about 2 minutes from my house.&nbsp; So we drove, and we chatted the whole way there.&nbsp; It\'s great talking to him.&nbsp; We talked about CtR\'s inferior bible study, our bible study, retreats and what not.&nbsp; He said he wanted to continue that conversation with me over coffee some time.&nbsp; I\'ll be meeting him tomorow 2:30 at common grounds coffee house.&nbsp; I hate coffee.&nbsp;   Well, I\'m in lakewood park, and lost.&nbsp; Who do I know who lives there?&nbsp; None other than THE shannon lam.&nbsp; So i called her, she rescues me from my being lost, and we conversate for a while.&nbsp; I was pissed we didnt hang out last week, but oh well.&nbsp; She said she\'d get her sister\'s-boyfriend to pirate me anime, so I\'m happy.&nbsp; She went back to her house to pack, I went home to get online.&nbsp;   She was online.&nbsp; First time I\'ve ever seen her online.&nbsp; She was checking email or something.&nbsp; We spoke for a while and it was awesome.&nbsp; She signed off, and I went to bed.&nbsp; My day was over at 11:30...    See you space cowboy...'),(4988,3938699,'27,July,2004','2004-07-27 00:00:00','   I constantly question myself.&nbsp; Am I doing the right thing, and if so, then who is it right for?&nbsp; Few people can scratch past my surface.&nbsp; Very few in fact.&nbsp; That is because I quell most of the hateful feelings that I have.&nbsp; That doesn\'t mean I never complain, it just means I could complain a lot more.&nbsp; From what i have gathered, I do this because of my parents.&nbsp; Every time I\'d come home from school and tell them that friend x pushed me or called me a dumb name, (honeybutt was fairly common back in the day), my parents would immediatly issue a fatwa against them.&nbsp; Nothing that person could do would redeem themselves in my parents eyes.&nbsp; They would openly express every other time I mention that person, in good light or bad, how little they think of them.&nbsp; That really bugs me.&nbsp; I\'ve been a jerk to all of my friends and vise versa.&nbsp; My parents hear about next to none of this though, because I don\'t want them to bash my friends whom I forgive on my own accord.&nbsp; I HATE when my parents do that.&nbsp; What are they trying to prove?  For example, I invited a few guys to my bible study this week.&nbsp; Not one of them attended even after having said they would enjoy going.&nbsp; Yes, I was hurt, and I felt slightly betrayed.&nbsp; Yet, I shall not mention this to my parents because I have slipped on&nbsp;other occasions as well.&nbsp; I am not perfect, nor are my friends.&nbsp; My parents believe themselves to be though, and consequently anyone who hints imperfection is immediatly dead to them.&nbsp;   I hesitated to tell them about Jackie.&nbsp; Still to this day, they are unaware of about 75% of our relationship.&nbsp; Inheretly I knew it would flounder in one way or the other.&nbsp; I knew something would blow up.&nbsp; I hate being right all the time.&nbsp; But I dare not tell them the what the real conclusion of that book was, for in the unlikely event I ever speak to her again, they would frown upon it.&nbsp; I carry these burdens, these lies on my back.&nbsp; Though I know not the reason for doing so.&nbsp; What good have I accomplished?&nbsp; Who am I serving?&nbsp; Not myself, that much is true.&nbsp; I do not have a friend whom I can converse with about everything.&nbsp; Ben for politics, Jake for girls/girl problems, Nathan for gaming/school crap, and Gus for obscure and unimportant things.&nbsp; I dare not talk to Ben about girls, nor about politics with Jake.&nbsp;   So now we\'re at the conclusion.&nbsp; I met this one girl.&nbsp; Jackie is physically more attractive than her.&nbsp; But she is so much more  beautiful  than Jackie could ever be.&nbsp; I could stare into her eyes for two eternities.&nbsp; It\'s not just the beauty in her eyes though.&nbsp; I can see past them into an keen and unbridled intelect.&nbsp; Aye she is smart.&nbsp; Words fail me when attempting to describe her.&nbsp; I am a writer: pictures to words, events to words, words to even better words.&nbsp; Still, I couldn\'t conjure up anything to capture her.&nbsp; She just is.&nbsp;   Enter inreconcileable flaw: Boyfriend.&nbsp; How that accursed word plauges me when not attatched  to  me.&nbsp; She cares for him deeply, every time she\'s off doing something, she is with him, or so it seems.&nbsp; She once told me she loves me as a friend, and him as a boyfriend.&nbsp; So it\'s all about choice.&nbsp; Altruism or selfishness.&nbsp; Putting myself in the boyfriend\'s shoes, I would hate me for doing what I want to do.&nbsp; Break them up and seat myself in his position.&nbsp; Yet back inside my own shoes, which I am more fond of wearing, that seems perfect.&nbsp; Why concern myself with someone I\'ve only met once?&nbsp; Does my being kind prevent others from hurting me?&nbsp; What debt do I owe, what cross must I carry, but more importantly why?&nbsp; Perhaps my selfishness is not right, but it would ease my pain.&nbsp; What must one do to attain true contentment?&nbsp; What journey, what storm must be braved?&nbsp; Am Ieven ready?    Fight on noble paladin...'),(4989,3938699,'27,July,2004','2004-07-27 00:00:00','   Ordinarily I think this goes without saying, but my Catholic beliefs are VERY liberal.&nbsp; In fact, I\'m just a liberal Christian, Catholic nothin\'.&nbsp; I believe the Church has the most legitimacy amongst the other Christian faiths, which is why I choose be Catholic.&nbsp; When I say \'legitimacy\', do not get defensive, for I do not believe that our \'faith\' is more realistic than any other.&nbsp; Faith is simply that, faith.&nbsp; No faith is more or less valid than the next, it\'s simply a matter of your beliefs.&nbsp; I weigh Catholic legitimacy in the sense that we abide by some tradition, in fact, almost total tradition.&nbsp; There are over 3,500 Christian denominations in the US alone.&nbsp; The reason is not a matter of \'faith\' per se.&nbsp; We all have faith in God, it is a matter of practice.&nbsp; \'I don\'t like that preacher\', is all you need to split off and form a new church nowadays.&nbsp; Now, to many churches and christian denominations are fighting over the most moronic of things, the most insignificant of practices.&nbsp; Perhaps I am blind to truth, but it seems that modern day Christianity has inverted Christs\' original message.&nbsp; To love.  Christ was a liberal, there is no doubt about it.&nbsp; The moral of Jesus\' life was not about proper rites on the sabbath, or how much you put in the collection basket.&nbsp; It wasn\'t about praise and worship, or praying through saints.&nbsp; It wasn\'t about \'accepting him into your heart as your personal savior\'.&nbsp; Those are all creations of men, and because of that, they are flawed and give rise to sin.&nbsp; The moral Jesus came to teach, was to love without ceasing.&nbsp; Be a good person.&nbsp; Clothe the naked, feed the hungry.&nbsp; Don\'t push a bible into someone\'s hands and tell them why they are cursed to hell if they don\'t read and believe.&nbsp; Amen I say to you, a bible forced down a man\'s throat shall not change them, but showing them the kindness spoken from within shall.&nbsp;   Most would call me atheist, or hethen because I do not subscribe to the orthodox belief that saying Christ\'s name grants entry into heaven.&nbsp; Indeed, it was said, \'no man may come unto the father unless it is through me\'.&nbsp; I believe that to mean that Jesus came as a means of salvation for all, not those who learned about him in Sunday school when they were young.&nbsp;   Let me allude to the end of Matrix Revolutions for a moment.&nbsp; Neo struggled to stand on both feet in the pouring rain, Smith clones all around him.&nbsp; By this time, every side of the war was aimed on him.&nbsp; He was hated by the machines, by Smith (satan), and even by some of humanity (Jews), but in the end, he realized the choice he had to make.&nbsp; Sacrifice.&nbsp; Neo died not to save just humanity.&nbsp; Neo made himself the lamb of the passover to save all life, machine and flesh.&nbsp; Likewise, Jesus did not come to save the righteous, for they already were saved.&nbsp; Jesus came to save the lowly prostitue, or the foul tax collector.&nbsp;   It has been said that Jesus exists in all of us.&nbsp; Many a person attends \'cougars for Christ\' at Tomball and expects to see Him most present there.&nbsp; Nay, I believe if you look at the hated, the despised of Tomball, you shall find Jesus.&nbsp; Outcast, abandoned, alone.&nbsp; I freely admit, I am in great fault.&nbsp; I have taken pleasure in hurting others.&nbsp; I am not inclusive.&nbsp; I hate many others.&nbsp; There are few redeeming qualities that I possess.&nbsp; I am just a voice crying out in the desert.    Make straight the way of the Lord...'),(4990,3938699,'23,July,2004','2004-07-23 00:00:00','   \'can\'t explain all the feelings that you\'re makin\' me feel...\'   Oh heckly, what a week.&nbsp; I suppose I\'m in that transition phase of my post-relationship-angst.&nbsp; It starts off with that strong pain in your stomach at the very sight of the person.&nbsp; Then the adamant hatred and malice.&nbsp; Shortly thereafter, was the \'I desperately need a warm body to fill the gap left behind\'.&nbsp; Now it\'s the, \'life is life, now its up to me to live it\'.&nbsp; At the current rate, do I wish I never met her, yeah.&nbsp; But who knows, she may have taught me a valuable lesson I have yet to learn, or sown seeds that will inspire me to find my true companion.&nbsp; But because I do see time as a four demensional spectre, I\'m pissed, and wish we never met.&nbsp;   Time for me to live life again.&nbsp; Jake got back today, well... I guess yesterday, I\'ve been lanning for almost 10 hours, so the whole time concept is blurry right now.&nbsp; Jake is really the only person I can sit down and talk to about this whole catastrophe.&nbsp; Who else?&nbsp; My parents, haha.&nbsp; Samantha, why would she care?&nbsp; We\'re mavericks as he once said, he shall help me brave this storm, and I him.&nbsp;   I hate uncertainty.&nbsp; I hate not knowing what comes next, or where to go with this.&nbsp; That\'s why I can\'t stand those who are smiling one minute, and stabbing you in the back the next.&nbsp; I\'m okay with assholes, but shifty assholes, I cannot stand.&nbsp; I wish I could peer into the future and see that everything will be okay.&nbsp; To see that I have conducted my journey in the best possible way, to see that I will not be alone forever.&nbsp;   I was talking to Dave (youth minister), and he asked me, \'what one certainty is there in this world, what always is always?\'.&nbsp; After thinking for a couple of seconds, my answer was, \'death\'.&nbsp; Life is always followed by death.&nbsp; That is the only certainty I have.&nbsp; No guarentees were made to me at my conception but that one.&nbsp; No one told me I would get married, have kids and live a happy life.&nbsp; Only that my life would be followed by death.&nbsp; That I suppose, is the nihilist point of view.&nbsp; Where I differ from the seemingly pessimistic and apocolyptic fate that nihilism describes, is that I have no fate.&nbsp; No one guarenteed me a happy life, or a beautfiul wife.&nbsp; Therefore, I must be the reaper of my own fate.&nbsp; I must control my actions, not submit to temptation.&nbsp; May I have strength to carry on.&nbsp;   I once described my journey as a walk down a road.&nbsp; Something set and simple to follow.&nbsp; I was in error.&nbsp; I am lone on a small raft, floating in the torrents of life.&nbsp; Temptation, hatred, doubt and fear all crashing against the few logs I\'ve put together as my life.&nbsp; I will be strong, I have enough strength to defeat these waves, to brave this tempest.    He sheaths Xylothin and looks out into the endless abyss, with&nbsp;a smile...'),(4991,3938699,'21,July,2004','2004-07-21 00:00:00','   So it\'s about 4 am right now.&nbsp; I just played Tales of Symphonia for a good while and decided to look back on past entries.&nbsp; I scrolled down to my first few that mentioned Jackie.&nbsp; Gosh, it was probably 4/22.&nbsp; Three months ago.&nbsp; Three months ago I could just see the tip of her destructive force creeping out.&nbsp; \'I\'m not a good girlfriend\'.&nbsp; Why didn\'t I listen?&nbsp; I was so blinded, so suckered in by those brown eyes.&nbsp; My God, I am a fool.&nbsp; I was a hold off, a&nbsp;temporary.&nbsp; She would say things like, \'Bryan, if I stopped talking to you I would just die\'.&nbsp; Well, as far as I know she\'s still alive.&nbsp; WHY DIDN\'T I LISTEN TO MY FRIENDS?&nbsp; MYSELF?&nbsp; MY PARENTS?&nbsp; Everyone was telling me to get out.&nbsp; Everyone.&nbsp; Samantha kept instilling faith in me, saying we were a great couple.&nbsp; THERE WAS NO COUPLE!&nbsp; It was her dragging me around to satisfy her own interests.&nbsp; Jackie kept telling me she\'d mature, she\'d  want  a relationship,  eventually .&nbsp; Even Jackie\'s mom said that she would do this.&nbsp; Why could I not see a definite problem here?&nbsp; Am I so ignorant, so blind to my own senses?&nbsp; If the mother of the girl you\'re trying to date tells you not out of malice, but sympathy that her daughter will posion my heart, there must be some great wrong.&nbsp; Nay, I was arrogant.&nbsp; I thought I could walk through, unscathed.&nbsp; I wanted to brave the storm and prove to myself what a good guy I was.&nbsp; How her inherent nature would be overcome by my charm and good nature.&nbsp; How decieved I was.&nbsp;   I could see the end from the very begining, but I was paralyzed.&nbsp; I have once been told, that \'dating is for dumping\', and that \'you should draw a realization of what you want in a wife, from every person you date\'.&nbsp; I would agree.&nbsp; I learned two things from Jackie.&nbsp;  First, if your friends, your girlfriend, your girlfriend\'s mom, and your own parents tell you something is wrong, they\'re probably right and you should reevaluate the relationship you are pursuing.&nbsp;  Second, I  need  a relationship that is a constant battle.&nbsp; All she ever talked to me about was, \'yeah, sarah beth blah blah\', \'james showed me his wang\', \'I\'m wearing that skirt you hate\', \'American Idol is cool\', and other useless information, not worth a neuron in my skull.&nbsp; I want someone whose very look sends me into argument.&nbsp; If she would argue, it would end with, \'I hate arguing w/ you\', or some rhetoric which she is saying simply to change subjects.&nbsp; I need my relationship to be a battlefield, and my girlfriend and I to be opposing armies.&nbsp; I want debate without concession.&nbsp; I want the battlefield to be stained red with our blood.&nbsp; My father\'s wife does not argue w/ him.&nbsp; He comes from an earlier era and that is to his liking.&nbsp; My brother however, cannot go ten minutes without arguing with Elizabeth, and they love each other intensely.&nbsp; That is what I want.&nbsp; I want to find a girl who is on par with me.&nbsp; Someone I can cross swords with and be unwary.&nbsp; If she was a gamer, all the hotter.&nbsp; But she  MUST , be someone I respect and more importantly  fear  intellectually.&nbsp;     I believe in a thing called love...'),(4992,3938699,'20,July,2004','2004-07-20 00:00:00','   Well, within not five entries of my new journal, I have recieved quite a hateful comment.&nbsp; I do have that effect on people, no?&nbsp; As always, the person left it unsigned, thus preventing me from knowing where my justification lies.&nbsp; Though I have my assumptions, I dare not make accusations, for that is never a productive attempt.&nbsp; Let me move on to why the commentary is&nbsp;false, and where it is not false, biased and ludacris.&nbsp;  &nbsp; The first attack made upon me is saying how \'I\'m too hard on people\', and that \'no one crosses the great Sir Bryan\'.&nbsp; Let me first ask who lives with humility.&nbsp; Is it the person who has countless shoes and matching outfits, someone who addresses herself as \'Princess Jackie\'.&nbsp; Does a person who live in humility say, \'Everyone at my church says I could totally get any guy I want\'?&nbsp; I fail to see then how I am pompous and arrogant when I wear the same pair of shorts for a week, and am in a constant state of abusing myself verbally.&nbsp; Nay, I say to you that I am not arrogant. &nbsp; Second, is the statement that, \'I took advantage of her because she was so young\'.&nbsp; What advantage was taken might I ask?&nbsp; You then bring up the stunning and unavoidable argument, \'If you really liked me than, blah blah blah\'.&nbsp; Tell me what it is that I forced her into doing?&nbsp; If I can recall correctly, it was her who said, \'Bryan, don\'t you ever just want to go out and have fun, forget your morals and ethics for a day?\'.&nbsp; I would argue that I put forth an attempt to instill some morals into her.&nbsp; Did I now have her promise me she wouldn\'t drink so long as we went out?&nbsp; Did I not tell her how badly I cringed when she explained to me of her \'fun\' summers in Mexico?&nbsp; Did I not convince her (partially) that homosexuality is not a choice?&nbsp; Amen I say to you, I would go so far as to say that I did more good for her than bad, and wise men would be at my side.&nbsp;  &nbsp; Third is the lies this unsigned individual spewed forth not three sentences later, claiming that Jackie told me, \'she just wanted to be friends\'.&nbsp; She did not, nor does not have any idea what it is that she wants.&nbsp; I say this without insult for she has freely admited this to myself, her mother and others on multiple occasions.&nbsp; If she truely wanted to be friends, then why did it matter if I did not recipricate her love, or that I had feelings for Lindsay.&nbsp; What then, if she truely wanted to be friends, would seperate me from the countless others at her church, or at school (James for instance) who she flirts with?&nbsp; It makes little sense to me that I am considered to have been pressuring her when I responded after her \'I love you\', phone call with, \'I\'m just not feeling the same about you\'.&nbsp;  &nbsp; Finally, let us consider the attacks made upon me at DYC.&nbsp; I, \'pretended not to notice her when she did her skit\'.&nbsp; In truthfulness, had I pretended not to notice her during the skit, I would have looked directly into her eyes and seen nothing.&nbsp; But nay, I looked away, for I knew that seeing her would only swell the pain that was growing inside me at that time.&nbsp; Yes, I too felt pain through the debauchery that was our hollowed relationship.&nbsp; I too was distraught throughout the weekend.&nbsp; What person would follow someone around who treats them as though they are nothing and plays them as a yo-yo? Next on the list was my phone conversation w/ Nathan that Jackie was present for.&nbsp; I will freely admit that I said on the phone to Nathan, \'Bye nathan, now give me kisses\'.&nbsp; Let me pose the question though, does that surprise any of my friends?&nbsp; I say no, for all are aware that though I will act gay and blatantly support gay rights, I am the straightest individual whom has ever walked on American soil.&nbsp; It is because I am aware of this and not afraid of homosexuality that I can do these things.&nbsp; Pardon me for not being ignorant and lifeless.&nbsp; Jackie more than likely simply used that as an excuse to end the relationship when she was already in cohorts w/ John. And finally I was attacked for arguing with the lady.&nbsp; It was \'embarassing how condescending I was\'.&nbsp; I admit, I was more abrasive than I intended to be during that \'great debate\' I had with the speaker.&nbsp; That however, was not my intention, after I left the room I told my PoP friends that my message was not to offend, but to inspire and enlighten.&nbsp; I told them that I felt the Holy Spirit flowing through me as the lady and I clashed swords.&nbsp; Not only did I humble myself afterwords to my friends, but after everyone had exited the room I returned to thank the speaker.&nbsp; I told her, \'Thank you very much for your talk ma\'am.&nbsp; It was very powerful and I was touched by your story, I can only hope and pray that there was someone who is going through what you did and understood what they need do\'.&nbsp; Is that the rebuttal of someone who is condescending?&nbsp; And if it was so embarassing to have listened to me, why did so many clap, why was I congratulated? &nbsp;   Please sign your name next time, this debate is not over...'),(4993,3938699,'19,July,2004','2004-07-19 00:00:00','   ...has an end, Neo.'),(4994,3938699,'15,July,2004','2004-07-15 00:00:00','   To summarize my experience at DYC, I would have to use the word \'bittersweet\'.&nbsp; On one hand, it was a great way for me to unleash the furies of my debating prowess and strengthen my own beliefs.&nbsp; And simultaneously, it was a weekend filled with malice and angst.&nbsp; So while I was becoming a stonger Catholic, I was being weakened physically and mentally in the process thereof. &nbsp; First let us go into why my weekend was rewarding.&nbsp; I am one of the few in my church who believe that gay marraige, or rights for that matter is acceptable.&nbsp; This became a big topic of discussion throughout the weekend between me and others.&nbsp; At one point in time, I was sitting in on a workshop regarding the Church\'s views on certain political issues.&nbsp; The speaker came out and said how she came within a day of aborting her child, but a vision from God allowed her the wisdom not too.&nbsp; Very touching story.&nbsp; Afterwords, she went into checking off the do\'s and don\'ts when it comes to politics and our faith.&nbsp; Capital punishment=bad, Abortion=Very Bad, Masturbation and Contraception = bad, etc.&nbsp; With five minutes before we had to depart, I had Samantha raise her hand to ask the question, \"what is the Church\'s view on gay marraige?\".&nbsp; Promptly the lady responded, \"Why obviously we\'re opposed to it\" &nbsp; Just before she began to close, I raised my hand and asked, \"Excuse me ma\'am.&nbsp; Did you  choose  to love your husband?&nbsp; Was that a concious decision you sat down and made?\" &nbsp; Lady, \"Well, actually, the Church does not believe that being gay in and of its self is gay, merely the sexual act of being gay since it is out of marraige.&nbsp; The Church calls homosexuals to live a life of chastity.\" &nbsp; Me, \"Well then wouldn\'t it make sense to allow gay marraige, thus making their actions inside marraige and therefore unsinful?&nbsp; Should I be prevented from living a full life with my partner because I was born differently than you?\" &nbsp; Lady, \"Well, actually, the Church is pro-life.&nbsp; They believe that too people are wed and have sex to reproduce.&nbsp; Obviously two men or two women can not have sex and reproduce.&nbsp; Therefore, if two people have sex without the intention of reproducing, its immoral.\" &nbsp; Me *grand finale*, \"So ma\'am, are you trying to tell me that a sterile couple shouldn\'t be allowed to have sex?&nbsp; They are just as incapable of reproducing as the average gay couple.&nbsp; Where is the difference?\" &nbsp; Lady, \"that\'s all the time we have for today\" &nbsp; I then get a standing ovation.&nbsp; Ah, there are few things in this world that equate to, going into a room to learn a thing or two, and then being the one who gets applause.&nbsp; So outside of learning more about what I believe in and getting into constant battles over faith and our religious practices, I got little else out of the weekend.&nbsp; But that was a hella lot. &nbsp; On the opposite side of this coin, I went in attendence with many others from my church who went to Steubenville, a charasmatic event.&nbsp; Lots of singing, adoration, etc.&nbsp; From what I hear it\'s eight fold what challenge retreat is.&nbsp; So half of my PoP people were running around me the whole time saying that this weekend sucks in comparison to Steubenville.&nbsp; That\'s cool to believe that and what not, but to come off as arrogant and opressive is quite another thing.&nbsp; It seemed like segregation.&nbsp; And personally, I would give up the chance to EVER go to Steubenville just as long as I got to see my brother put that golden ring on Elizabeth.&nbsp; I don\'t care what a religious experience that is, I love my brother more than anyone, living, dead or celestial.&nbsp; I fight and bitch at him, but he is my brother and I love him.&nbsp; So to hear I\'m a bad person for not going, makes me a little angry.&nbsp;  &nbsp; To top that off, Jackie was floating around a multitude of guys, explaining how I mean nothing to her.&nbsp; It\'s that feeling you get in your stomach you know?&nbsp; Jake knows it, and I know it.&nbsp; And few things can erase that feeling.&nbsp;  &nbsp;   The sun is always set on my side of the world...'),(4995,3938699,'15,July,2004','2004-07-15 00:00:00','   When we first started conversing, Jackie told me she was afraid of commitment.&nbsp; Apparently some time ago, she had a major crush on some guy, who later dumped her for someone she found to be less attractive than herself.&nbsp; After this \'heartbreak\', she wrote her crush a fat email explaining what a horrible person he is and how he\'s doomed to a life of angst and suffering.&nbsp;  &nbsp; How fitting that she left me for not someone, but something.&nbsp; She left me to be \'free\', to be capable of doing whatever she wants without fear of that bad feeling you get, knowing you\'ve hurt someone.&nbsp; Likewise, I do not believe what it is she left me for is more attractive than I, so I wrote her an email, thus allowing her to experience what she did to another.&nbsp; To show her that she is no better than he. &nbsp; \" Jackie,  &nbsp;&nbsp;&nbsp; I hope you can accept my apology.&nbsp; Since Battle of the Bands I have been holding you back and doing you wrong.&nbsp; You\'re nothing but a flirt.&nbsp; That\'s all you do, you flirt then walk away.&nbsp; So by attempting to be your boyfriend, I prevented you from spreading your wings.&nbsp; Likewise, you were/are a burden to me since we first met.&nbsp; How foolish I was to suspect \'you\' could be a girlfriend.&nbsp;  &nbsp;&nbsp;&nbsp; Jake has once told me, \"you don\'t go out w/ someone to be cool, you go out with someone because they make you happy.\"&nbsp; In all honesty, you haven\'t made me happy in proportion to your ill treatment of me.&nbsp; You lead me on like a siren and then do an about face.&nbsp; You call me when you\'re on vacation, no doubt checking out every guy there, crying and telling me you\'re in love.&nbsp; I don\'t care.&nbsp; Don\'t call me, don\'t IM me, don\'t talk to me if you ever come to my bible study.&nbsp; I don\'t want you to pretend you give a shit about me.&nbsp;  &nbsp;&nbsp;&nbsp; What\'s more, you treat me like a platonic friend, but cry whenever I mention another girl who I\'m interested in.&nbsp; Then, at DYC you hang on some guy and tell Sam about it.&nbsp; I felt a little crappy for a while after seeing and hearing about that.&nbsp; Could I blame you though?&nbsp; All I could do was weep for the poor fool you were seducing next.&nbsp; Maybe he just wants a \'fun\' couple of minutes.&nbsp; I\'ll bet you ten bucks he won\'t put a smile on your face when you go to sleep.&nbsp;  &nbsp;&nbsp;&nbsp; You don\'t want that though.&nbsp; I mean come on, you flirt w/ everyone, that is what you want.&nbsp; No need at all to stay true to one person when you can have them all right?&nbsp; Well, all except one now.&nbsp; And don\'t shed any more tears over me, I don\'t really care, or have any desire to bear the responsibility.&nbsp; I figure if you\'ll be selfish enough to do what you have to me, I have some right to take interest in myself.  &nbsp;&nbsp;&nbsp; I just hope that you can find someone who will take interest in you.&nbsp; Someone who can put up with your capricious nature and passion for clothes.&nbsp; Hopefully you marry into wealth, or else you probably won\'t be really happy.&nbsp; Keep searching though, there has got to be someone out there who loves you and who makes you feel good.&nbsp; Someone who can make you smile, and will smile just the same when you ignore them.&nbsp; When you treat them like they\'re nothing.&nbsp; When you value the chase more than the catch.&nbsp; But as for catches, I\'m the one that got away.&nbsp; Damn.  &nbsp;  Sir Bryan  &nbsp;  PS: enjoy the rest of your natural life \" &nbsp;   It\'s all about the choices we make...'),(4996,3938699,'15,July,2004','2004-07-15 00:00:00','   I remember the simpsons long ago where Lisa tried to determine if a mouse was smarter than Bart.&nbsp; During one of her tests, she attatched electrodes to a cupcake, so by attempting to touch it, Bart would get a sudden shock.&nbsp; He grabbed it, got shocked, then did it several other times.&nbsp; Bart did not realize that even though the cupcake would be tastey, it only hurt him. &nbsp; That is my relationship with Jackie in a nutshell.&nbsp; I think to myself, \'hey, she\'s a pretty gal with a great deal of wisdom and knowledge on her head.&nbsp; A relationship is possible and would yeild good things for me\'.&nbsp; Yet every time I tried to get close, every time I showed interest in her, she stung me.&nbsp; Either by claimed to want a platonic friendship, or just ignoring me.&nbsp;  &nbsp;   \"you know that feeling you get deep in your stomach when someone you care about betrays you?\"'),(4997,3938699,'15,July,2004','2004-07-15 00:00:00','   Prolouge to a Mack part I - 12/8/2001  &nbsp;  &nbsp;  Up until about 8th grade I had not had a girlfriend before. Finally I asked out this one girl who I barely knew. It wasn\'t really a physical relationship, or a speaking one, it was more just flirtatious gestures from time to time. To my dismay it ended because of a lack of trust. I now consider it more of a test run for my relationships.  Then came the time when I had sharpened my skills around females and began to actually hit on someone. This ended worse than I could have possibly imagined. The person I began to mack on was a band person, which made things difficult to begin with. Band took up two class periods and we had a completly differnt lunch so I could spend a maximum of one period with her. Every chance I got I would call her and try to be with her, seeing as how that would only be about a half hour a day. Our school was having an astroworld field trip that year and my girlfriend and I were both going. So I was thinking, \"un un uh, I\'m gonna get me some this trip.\" We picked partners before we started going out so I had to walk all over the park looking for her. When I found her I was like, \"hey hows it going, wanna step out for a moment...\"  \"uh no I\'d rather sit here and flirt with your best friend.\"(didnt actually say this)  So I was following her around trying to chat with her and every time I\'d get a word in she\'d start talking to Mr. X, my best friend. Needless to say it was a bad field trip.  Somehow I forgave her, and our relationship was fine for about a week or two more. Then I\'m like, \"alright, astroworld sucked, so now being my birth-day, I\'m really gonna get some... Oh yeah.\" So I walk up to her before school started and she goes in a strange tone \"oh Bryan here... happy-birthday...\" and hands me a note. I get to class and read the first two lines:  Happy birthday Bryan!!!  I\'m breaking up with you for Mr. X  &nbsp;  &nbsp;  Prolouge to a Mack Part 2 - 12/9/2001  &nbsp;  &nbsp;  Continued from \"Prolouge to a Mack Part 1\"  I was crushed, nothing can match the heart-break I felt. That single note sent my self-esteem to a dark ages. And I was like that until the summer...  I finally got involved with my churches youth-group that summer and I was overwhelmed by what a pleasant enviorment it was. I had never been in a happier place. Soon a some of the hot girls there began to hit on me, so I\'m all like, \"well ya know mabye the long-term kinda relationship aint for me. Ya know mabye I was put on this earth just to knock someone.\" Soon enough came this confrence thing where the whole youth group goes for about two days or so. During that time I hit on a total of three girls.  Girl #1: I began to get my mack on, and it was working pretty well until this stalker guy started to rival me so I backed off.  Girl #2: This time I started and right off the bat she goes, \"uh get off me; you\'re gay\", well that sucked.  Girl #3: Finally I started to make some progress; This girl and I gave off some flirtatious gestures from time to time, small talk, ya know the good stuff. Finally when the dance came time I walked up to her and she said, \"Oh uh... I can\'t be around someone all day... bye!\"  I began to realize an actual relationship is what I really needed in my life.  &nbsp;  &nbsp;  Prolouge to a Mack Part 3 - 12/12/2001  &nbsp;  &nbsp;  Okay, so I\'m not that great of a mack, lets face it. It was that time I said, \"God, I\'m not a Mack, and my relationships don\'t work out so well. Give me some guidence.\" Now for people who think God doesn\'t listen... he does. Sure enough, I found the perfect person in my youth-group for me. Smart, funny, we had almost everything in common. The only disadvantage was she had a boyfriend at the time.  Well, for some reasons, many including the boyfriend was kinda a depressed person (I don\'t really know the details for reasons which will be explained). To me this was a total victory, now if only I could just sway her a bit... Ah-ha I thought! If I took her to my homecoming she\'ll surely fall in the deep undeniable love with me that I had with her. Well, as it turned out she went with me and... nothing happened. Well a few weeks later I start to talk to her and ask her some menial issues, and she just kinda shut me out. I have no idea what I had done. She wouldn\'t really talk to me at all. She seemed so depressed, and thinking she was a great friend I tried to comfort her. Unfortunatly, she just wouldn\'t have it.  So as you all can see my life has had many relationship flaws from time to time. Alas, nothing can hold me down, as mentioned earlier I have a never dying happiness. Unfortunately I would be slightly happier if I had someone in my life who I could share my happieness with...  &nbsp;  &nbsp;  Can\'t Nobody Hold me Down!!! - 12/12/2001  &nbsp;  Well now that I\'ve finished the Mack Saga I can start talking about my current life.  I\'m 14, I live in Houston, Texas, I\'m in to rap, girls, IM (Btsunami2001), and all that other mess. I guess I\'m pretty funny and like to chat alot.  g2g so if anyone wants to chat:  Holla back!  &nbsp;  &nbsp;  &nbsp;  &nbsp;  Finals... Arrrrrrrrgh!!! - 12/14/2001  &nbsp;  &nbsp;  Ah I cannot stand finals. Though this is my first year to take them, I am utterly overwhelmed by them. It doesn\'t help having some really mean teachers who give unquestionably hard exams. I had to do like six differnt projects today and am very stressed out.  I know this one girl who is all like, \"Hey you are looking at a girls body... don\'t you care at all about personality... you stupid pig!\" Now if thats all I did I could understand that argument, but that garden tool has had a boyfriend for about a year now and is giveing herself off to half the guys she knows (excluding me thankfully) and she and her boyfriend knock boots everytime they get a chance. All I said was that I could not go out wit\' a girl that took no care of her physical apperence. I don\'t care how smart, friendly and funny you are, you\'ve gotta take care of your body. Brilliance does not mean you have to throw your apperance away. I mean come on, I wouldn\'t expect a girl to go out wit me if I didn\'t bath comb my hair or trim my nails, I should think that girls should have to at least be cleanly and shave. Only exception: A mortal sin against your religion.(Which I know of no religions that have that, but why bother relying on me I\'m an ignorant pig)j/k. Well aside from the village bicycle on my bus, today was fine. Not much going down, I\'m still having relationship problems with my youth-group girl, but ah well, not my whole life.  Peace, love, and victory  HOLLA  &nbsp;  &nbsp;  Still Dre - 12/19/2001  &nbsp;  Today I found out that my closest friend, indeed hates me with a passion. I was crushed, I thought that I could tell her everything. Alas, my life is still great, I\'m having fun and that\'s all that matters.  Finals will soon be over and I will enter two weeks of bliss. Ah the Christmas holiday... goofing off, sleeping in, ah how miss thy blissful times. Well to all of ya\'ll who be hatin\' on me I don\'t really care cause when Christmas roll\'s down, I\'m still Bryan... and no one can take that from me.  Peace, Holla,  Bryan Mack  &nbsp;  &nbsp;  You can hate me now... - 12/19/2001  &nbsp;  &nbsp;  \"...But I won\'t stop now, cause I can\'t stop now.\"  This is just a shout-out to all ya\'ll playa\'s who be hating on me... you know who you are! Like I said, you can hate on me now, but I dont give a $#@* cause nuthin\' you can say will bring me down or stop me from what I do.  Hey Megan, yo in the hall you said thanks for something? Sorry, I was a bit messed up at the time. Well thats all I gotta say for right now.  Peace, (except for all you hatin\' on me now)  &nbsp;  &nbsp;  Victory! - 12/21/2001  &nbsp;  &nbsp;  Finals are over, I have triumphed over that which has all but killed me... now for the next two weeks I\'ll be sitting on my a$$ and watchin\' TV... the good life.  Well my finals were really hard but wit\' a small bit of studying and goofing off I pulled them all off pretty good. I know that I got an 89 in english and bio. so I\'m pretty happy about that.  Well now that I\'ve told a bit about myself I guess I\'ll describe myself.  Height: 5\'7\'\'  Weight: Hopefully 150 lbs.  Dark thick brown hair.  Medium build, and thats about it.  HOLLA  &nbsp;  &nbsp;  Girls, Girls, Girls, Girls - 1/9/2002  &nbsp;  &nbsp;  Man, I\'ve been so confused. I gave up on that youth group chick... heck I\'m gonna stop goin\' to that youth group period, but not cause of her. There be this famliy there, and every-single one of them make me wanna go on a killing spree.  Bad thing going down now is that all my playa\'s (You know who you are) think I\'m just some guy who\'s in it for the physical... that just aint true at all. I\'m very concerned with personality, unless the girl is like REALLY, REALLY ugly I would like her.  I kinda like three girls right now:  the first, I don\'t know her last name and I haven\'t actually spoken to her.  the second, I don\'t know her last name and don\'t go to her school, but on the plus side she really likes me  and the final one, well lets just say I\'m not in her favor.  Anywayz,  Peace, Holla, Gettin\' Your game on  &nbsp;  &nbsp;  Lets get it... - 1/10/2002  &nbsp;  &nbsp;  \"Make this money, take this money, aint no way you can take this from me, aint s*** funni, shake it honey, take this money; now lets get it\" G-Dep, lets get it  Shout outz...  This is for my playaz who\'ve been wit\' me long as I can remember...  Glenn ya crazy foo\' keep pimp\'n, Johnny you just need to go clubbin\' some time, Ryan we need to work on you a bit.  For those I met few years back...  Nathan you got some issues boy, Jake... wonder ho powers... activate!!!, Meagan\'s wow there are like too many Meagans I know, Macy and Lauren go pimpin\' with me sometime alright, Shannon and all the smarties\' you\'re giving me some competition... stopit, Guardbabe you\'re a sweety holla! hehe, Christian you\'ve got the son of God for a bro. thats cool.  &nbsp;  &nbsp;  &nbsp;  Changes - 2/26/2002  &nbsp;  &nbsp;  First lemme make a few shout outz, Sheena, don\' feel hated on; like tupac shakur said shortly before his death,\"...buried as a G while teh whole world remembers me, until the end of time...\" i will be ur hommie until  the end of time.  Nathan, you have now made it to king pimp, one, two, three, four, five... just kidding, you know that will always be my job... hahahahahaha!  BTW, don\' none of ya\'ll be hating...\"Really? get smacked silly, you get smacked silly, F***ing with these N*****...\" G. Dep-Lets get it  A revalation which has just recently occered to me  :  Everyone in the THS band or guard are obsessive and exclusive people who are probably the worst part of our school. Now Glenn, johnny, vann and ne i missed, this don\' mean u so stop ur biznitching.  As mentioned earlier... rap keeps me alive... it is probably my life force, it inspires me to become great and has lead me to a state of near mental balance (don\'t leave a note bout that nate, k? lol)  Alright, well G2G ya\'ll, Peace, Holla, and straight pimpin on wax... nuffin else matta\'s  &nbsp;  &nbsp;  &nbsp;  I\'m an school anthropologist... - 2/26/2002  &nbsp;  &nbsp;  ...that of course is provided they deal with studying the formation of social structures...  Being the overanalytical person that i naturally am i have mapped out the entire nerd clique at THS by tribe   Ekyd Group:  Two members inhabit this group, Kassie and Erica  They stay together because of their common hate for men and resentment to physical apperance. Also, they happen to be oversensative which makes them cling together in a host like way.  Central Nerds:  3-4 Memebers; Shannon, Nathan, Ben on occasion, and I. All of us tend to sling sexual jokes off at each other and have a lot of fun, cept when we make fun of ben, lol. Also are rather thick skinned and except for me are extreamly smart...  Kob and his followers:  7+ members, Kob is Jake and his lil\' sub clique thingy goes down in that group, I really have no idea what holds this group together other than that they friends with jake and taht mess... very mystirous   Uncommons:  2 members, Emmit and Austin, they really arent nerds but they sort of hang aroudn with us... very strange.  1/2band nerds:  2 members, they are sort of band nerds but do not really stay in the band hall thus not making them zealous band nerds.  &nbsp;  &nbsp;  &nbsp;  Diary to a Mack Part 4... - 2/27/2002  &nbsp;  &nbsp;  alright i said part three was the last but as u can see... it isnt...lol  PART 4... Playa Hata:  Alright as you know bout my g/f who broke up me on my birth-day was a band nerd...(worst peeps alive)And also like i\'ve pointed out frequently I am a very forgiving and overly friendly person... so as you can imagine i forgave HO-BAT  I always talked to her bout shmit over teh summer, things were great and we were kickin it. we get back to school and so we be chattin and her and her all of a sudden like week into school she figuretively bish-slaps me... I\'m like WTF HO-BAT, I forgave your a$$ and your all ridin me, so i basically stopped callin her right. Ahh but the pure treachory and that mess for band-nerds  did not end there, I sort of met this girl whom i\'m gonna call guardbabe05  by accident, i\'d go into the details but i don wanna blow my storage quota on one entry... but newayz, we met and i was tellin her my girl trubles on day so she goes, hey i\'ll hook u up wit this friend of mine. Being reluctant at first  i finally said, k wuteva. So i called this girl  and was kinda givin her the scoop on my classes and shmit. Durign this time i told her how i STRONGLY dislike my art teach and wished for her to... lets just say stop blinking for an excess of time. So GUARDBABE05\'s b/f, who is a hrny bstrd, was all like yo bryan, don say tat mess to a girl. Ok i know he has a g/f and i don but lets see he\'s a junior and GUARDBABE05, is a fish, comon he cant have that much game. it wud be like me goin out wit a 7th grader... ewwww... back on track... so i call up GUARDBABE05\'s friend again and am gonna pologize for slingin that crap and when she picks up the phone she\'s all like... um yeah, GUARDBABE05\'s friend is um, in the shower...  Quick synopsis for those who look at notes first: Forgave, got bishslapped, accidently met h0, chatted wit h0\'s friend, got jail R*****   &nbsp;  &nbsp;  Prolouge to a Mack Part 5... - 3/1/2002  &nbsp;  &nbsp;  Part 5, coming to terms:  Ya know people be saying how i need to hop up off the band nerds... I say F*** that S***, if i wanna throw flames i muthaf****** will and who ever is a ho a$$ enough to try and stop me can suck my MuthaF****** D***, but thats not the important thing...  The important thing is that last year i was all up in the school trying to fit into a certian clique: I wanted to be wit tha nerds but if we\'d be talking about gaming or sumthin like taht and a prep walked by i\'d get embarrased and stop talking... i wanted to be accepted by the band nerds last year too (blough, worst mistake of life... worst people concocted by God  ) but they cud only talk wit me for about ten mins a day so that went poorly... then i went to my youth group and i sort of lost my sense of embarrasment right... that was the best thing that could have happened to me. I now accept i\'m one of the biggest nerds in the school... i goof on AIM for most of the day, play RPGs a lot and chat about computer S*** all day wit my peeps. The only thing i\'m prouder to be other than a nerd is a Catholic  but still, ever since i realized that I WAS a nerd i became happier and happier,   NTM, I\'m also gettin along a lot better with my playas from last year, and since most of us have open diaries its easy for us to chat alot. I\'m really good friends with nathan the pimp-in-training, Jake who hates his Kob clan  and Shannon/sheena who are the smartest girls in the school  , it really amazed/humbled how i kick it wit all the smartasses of our class.  Quick synopsis: Once wanted to be a dumbshit band nerd, now i\'m a nerd... NERDS ARE THE COOLEST PEEPS ALIVE, PREPS ARE ARROGANT, GOTHS SCARE THE F*** OUT OF ME, BANGERS CUT THEMSELVES  , PUNKS TRY TO BE NON-CONFORMISTS BUT BY DOING THAT THEY ARE CONFORMING TO THEIR OWN NON-CONFORMITY, AND BAND NERDS COMBINE ALL OF THOSE S**** TRAITS IN ONE CLIQUE... I\'M TOO ANALYTICAL FOR MY OWN GOOD, WHY AM I WRITING IN CAP LOCK???, HOLLA BACK  &nbsp;  &nbsp;  Changes... um some more i guess - 3/10/2002  &nbsp;  &nbsp;  \"that\'s just the way it is... things will never be the same... that\'s just the way it is...\"  People be asking why I won\'t delete those lovely messages I get from whomever... BTW, the crazy ho-rat Amanda who left me the last note can suck my fat C***. Neway, the moral of this story is... this is my opendiary... i\'m glad if people leave me notes, even if they are threatning of my life, so on and so forth. I\'m not gonna delete them just cause someone put something mean on my S***. So all the peeps that be deleting my S***, grow up, its an open diary... notes are for everyone to put... you put ur s*** on my diary and i\'ll do the same to you... Reason i have an all note function is fo\' my hommies Nate/Jake, not for cowards who wanna be unsigned so i can\'t find their ods. I\'ll keep leaving notes on band nerds\' s***. BTW... to all ya\'ll hating on me, cris, band nerds, that mess...  \"I aint trippin\' i\'m a playa, i aint sweatin him, I sex his sister and had her screaming like a mexican, with no remorse it was meant to happen; \'sides rappin only thing good i did was scrappin\'; until the end of time.\" 2pac shakur, until the end of time  &nbsp;  &nbsp;  Gin and Juice... - 3/11/2002  &nbsp;  &nbsp;  \"...sippin\' on gin and juice; laid back, with my mind on money and money on my mind...\"  Reason for this title was... well my ex-friend\'s name is Cris which reminded me of crystal, a type of champaign frequently mentioned in rap, which then reminded me of the song Gin and Juice by Snoop Dog. I dunno, my mind wanders. neway this is bout her.  Timeline of my friendship wit her...  Begining of school: thought i was hot  we talked alot. was told she wanted to go out wit me, because she was a junior i thought that was just a joke...  month or two into skool: Started going out with this one guy. Didn\'t think much of him because of his rep. never actually saw him yet though. became friends instead of love interests... note: this girl is very hot.  Two or three months into skool: talked to her openly bout michelle, luv interest at time and currently... we chatted frequently and were good friends  Just before christmas break: recieved an awful, awful letter from her tellin me to back the f*** off and saying i shudnt go after michelle. now i was thinking that its okay for her to F*** her b/f every week  but its not ok for me to luv a girl.  Newayz the moral of that story is... one ho is better than two??? well actually that had nuffin to do with that but oh well, that\'s not the important thing.  Um, neway  if u read this Cris... suck my D*** u fat ho.  and if u read this Shell: I still sorta luv u... despite ur lack of emotional involment with me...  (michelle is really really hot... ask nathan.)  &nbsp;  &nbsp;  j\'adore michelle... - 3/12/2002  &nbsp;  &nbsp;  That isnt a rap song either but... eh F*** it.  Loosly translated in french that means \"I love michelle\"  Well if you don\'t know then don\'t bother thinking about it and if so here\'s tha deal.  Ya know I\'m always trying to ask her stuff but she like never says nuffin to me, still i am drawn to her. I\'ve got the damndest feeling that even if we were to go out that it would fail horribly leaving our friendship a smoking crater  see, she views high-school relationships as \"practice\" for future ones. Ah, if only she knew how i felt... for the love of God no one tell her... wait, no one who reads this crap knows her newayz... haha, the beauty of my plan...   Funny story; Spring break, monday, five or so days before end of break... 8 page minimum paper, business presentation, French scrapbook, biology report; all of that is due in about five days... ARRGH!!! who told me high skool was gonna be a bish...? well everyone but ah well, its my own damn fault for being a procrastinating dumb $***  Newayz,  Peace, luv, HATING ON S***Y A$$ BAND FAGS; nuffin else matters.  HOLLA  &nbsp;  &nbsp;  Until the End of Time Part 1 - 3/13/2002  &nbsp;  &nbsp;  Guess what band nerds??? I can hate on you all I want... I\'ll still be me... Until the end of time...  &nbsp;  Perhaps I was addicted to the dark side  Somewhere inside my childhood witnessed my heart die  And even though we both came from the same places  The money and the fame made us all change places  How could it be - through the misery that came to pass  The hard times make a true friend afraid to ask - for currency  But you could run to me when you need and I\'ll never leave  Honestly, someone to believe in, as you can see  It\'s a small thang to a true, what could I do?  Real homies help you get through  And come to know he\'d do the same thang if he could  \'Cos in the hood true homies make you feel good  And half the times we be actin\' up call the cops  Bringin\' a cease to the peace that was on my block  It never stops, when my mama ask me will I change  I tell her yeah, but it\'s clear I\'ll always be the same  Until the end of time  So take, these broken wings  I need your hands to come and heal me once again  (Until the end of time)  So I can fly away, until the end of time  Until the end of time  Until the end of time  Please Lord forgive me for my life of sin  My hard stare seem to scare all my sister\'s kids  So you know, I don\'t hang around the house much  This all night, money making got me outta touch, shit  Ain\'t flashed a smile in a long while  An unexpected birth worst of the ghetto childs  My attitude got me walkin\' solo, ride out alone in my lo-lo  Watchin\' the whole world move in slow-mo  For quiet times disappear, listen to the ocean  Smokin \'Ports, think my thoughts, then it\'s back to coastin\'  Who can I trust in this cold world?  My phony homey had a baby by my old girl  But I ain\'t trippin\' I\'m a player I ain\'t sweatin\' him  I sex his sister, had her mumble like a Mexican  His next of kin, no remorse it was meant to happen  Besides rappin\' the only thing I did good was scrappin\'  Until the end of time  So take, these broken wings  I need your hands to come and heal me once again  (Until the end of time)  So I can fly away, until the end of time  Until the end of time  Until the end of time  So take, these broken wings  I need your hands to come and heal me once again  (Until the end of time)  So I can fly away, until the end of time  Until the end of time  Until the end of time  Now who\'s to say if I was right or wrong?  To live my life as an Outlaw all along  Remain strong in this planet full of player haters  They conversate but Death Row full of demonstrators  And in the end drinkin\' Heneessy made all my enemies envy me  So cold when I flow eliminatin\' easily  Falls to they knees, they plead for they right to breath  While beggin\' me to keep the peace (haha)  What I can see closer to achieve  In times of danger don\'t freeze, time to be a G  Follow my lead I\'ll supply everything you need  An ounce of game and the trainin\' to make a G  Remember me, as an outcast Outlaw  Another album out, that\'s what I\'m about, more  Gettin\' raw \'til the day I see my casket  Buried as a G while the whole world remembers me  Until the end of time  So take, these broken wings  I need your hands to come and heal me once again  (Until the end of time)  So I can fly away, until the end of time  Until the end of time  Until the end of time  &nbsp;  &nbsp;  &nbsp;  &nbsp;  Coming to Terms... - 3/17/2002  &nbsp;  &nbsp;  Yes I\'m also well aware that this title was not rap related at all... but  who cares, no one who reads my $hit likes rap anyway.  I have learned over the years that you can have all the friends in teh world, but if none of them like you, it makes no difference... instead, if you have at least 1 good friend, than it is unquestionablly better than one  thousand fake ones. Reason for me saying that is for two reasons. For the past 6 years of my life I was trying to follow the Preppy crowd... not really understanding where my ambitions lay or who I wanted to be with. I thank God that I have finally realized what I truly am: a nerd. I talk on AIM frequently and talk to my friends about games... and now that I have finally accepted and been open about this I have never been happier.  The second reason is because I do have a few truly good friends. What we all have in common is that we all wear masks. Though Nate appears a cold-hearted conservative to many people, I have had the chance to get to know him better and he is truly a fun loving person who... well he\'s still a conservative but, eh... Jake also seems distant many times. He will give you the impression that he is a jock and despises the ground everyone else walks on. Jake on the other hand is a very passionate person with a sense of humbleness that it is astonding. Then we have Shannon, upon hearing that she is the smartest person in the school you would think that she is arrogant and has no life. She is none of these, Shannon is witty, slightly insane, and humble for her  abilities.  I consider these three to be my true friends   We all have something in common and for the first time in my life I feel that I\'m on the same frequency as someone else.  But the moral of this story is... if you\'re a playa with thousands of ho\'s and their all ugly... you\'re not really a playa, but if you only gots a few hot a$$ b****s then you\'re the best playa round... This is for my hot a$$ b****s,  HOLLA my pimps and pimpettes!!!  &nbsp;  &nbsp;  I don\'t believe in psychics but... - 3/18/2002  &nbsp;  &nbsp;  At this time you are feeling \'uptight\' and you are urgently in need of rest and relaxation; but perhaps even more than that you need to overcome that feeling that you have been \'hard done by\' and treated with a complete lack of consideration. Maybe you have, but whatever may have been the cause of your inherent anxiety, you regard the situation as intolerable. Your are, however, sufficiently competent to turn that situation around - you have overcome similar problems in the past, and really this one isn\'t too different.  You are looking for something different. Your imagination has been working overtime and you are seeking adventure - and you\'d like to share that adventure, the new experience, with someone like yourself: Imaginative, Enthusiastic and Sensitive.  You feel truly deprived - not getting your fair share, but you have accepted the fact that that is the way things are at this time and that it is prudent to let matters slide and not hit your head against the wall -so conform and agree for a while. Accept the situation - nothing can last forever.  There is considerable amount of stress present in your life at this time and this is perhaps due to some considerable mental and physical frustration. There are various physical needs that are necessary for your well-being but whatever the reasons - mostly of your own making - your needs are not being fulfilled. We wonder why? You are under the impression that nobody seems to care for you. This predicament is most uncomfortable and it is because of this that you are experiencing far more stress than you feel you can cope with. You need to find a soul mate - someone who truly understands you and whose standards are as high as your own. As matters stand you would like to break away from the vicious cycle that you find yourself entrapped but this is easier said than done. You refuse to compromise with your opinions and essentially you are unable to resolve the situation because you are continually postponing the making of the necessary decision. You are stubborn but this should be no deterrent experiencing a happy life.  You are completely worn out and you are not in the mood for any further demands on your resources. The situation - such as it is - has rendered you quite helpless, unable to continue the mental battle that you have been pursuing for some considerable time. Enough is enough. All you would like to do now would be to have some time for yourself, to find a peaceful situation where you can recuperate in your own time.  ...I was just stunned when I read this... I always try to analyze myself... understand who I am... for the most part this summed it up... www.colorgenetics.com  &nbsp;  &nbsp;  &nbsp;  Bad Intentions... - 3/19/2002  &nbsp;  \"I don\'t care for F***ers, I\'m just drinking, smoking straight west coasting, B****s put your A$$ in motion, P**** poppin sex on lotion, raise it up... blaze it up... all these n**s, I dont give a F***. All i really know is that yo ho wants you to be with me, she aint playin\', what i\'m sayin\'...\" Doc Dre ft. Nock\'Turn\'Al, bad intentions...  Here\'s my average day... wake up at 6 and get my clean on, ride the lorenzo sportin\' bus 2 skool, kick it with my pimps an\' playa\'s, roll on down to trick a$$ Evan\'s or Moore-head, hop up an\' on the bus \'gain, get my sleep on, wake up at six and dine and gin and juice, holla at my hommies again, listen to some tunes and think my thoughts til three, then hit tha muthaf***ing sac..  Translated: wake up at six and shower  , ride the bus to skool, hang out with my friends before skool starts, go to one of my least fav. classes, get back home and sleep till 6, eat dinner and goof off on AIM, then i goof off and listen to rap till 3 in the morning and wake up again.  the moral of this story is: if you smoke marijuana next to a russian nuclear silo, you\'re bound to end up destroying the entire universe...   Holla back my pimps and playas...  &nbsp;  &nbsp;  Gin and Juice... Da Remix... - 3/21/2002  &nbsp;  &nbsp;  Alright i\'m well aware that gin and juice was not remixed... as far as im concerened...  Well, reason fo\' writing this is cuz i need to update \"Gin and Juice\" see the moral of thsi story is... Cris really didnt hate me, just her b/f did... haha... so me and her are cool now cuz we were like such great friends it wasnt even funni, we only incoded about 2% of our convo. nathan... haha.  But here\'s wtf\'s been going down recently... I am the happiest mutherf***er in the whole world. I got good friends, kazaa, a kick A$$ bro, life be good...  Hit me back my W****z  &nbsp;  &nbsp;  Severing Ties... - 3/24/2002  &nbsp;  &nbsp;  \"you aint neva seen a N**** like me, never in yo\' life; and that\'s what you can\'t understand\" You can hate me now-P.D.  all of my playas been asking y i don\' go after that shell girl... here\'s my reason... i have decided that it is simply a waste of my efforts and resources to persue this girl. She simply does not talk to me about anything of interest/seriousness, we go2 dif. skools which makes it hard, and she\'s a crazy ho... end of story...  also, to those of you who have heard my legacy of POP (my church) this will make more sense, if not than ask. Anyway, the moral behind this story is that I shall no longer support POP\'s youth group and simply continue as a regular catholic.  ah hell w/ it, i\'ve got plenty of time...  THE LEGACY OF POP... dun, dun, dun...  I came to POP\'s youth ministry by chance; I should have been up at Washington state w/ my northern homies pickin\' strawberries and makin\' money... instead i stayed home and went to youth group... I was changed from a cold hearted indivig\' to a loving and open person... the people there were so nice... especially the seniors... they all made me feel at home... welcome... I finally was able to be myself... or who I thought I was at the time... then skool rolled on in and as I walked into skool for the first time for three months, i sensed a forign, almost evil presence... From POP I learned to make fun of myself instead of others (its sooo much easier, think, if you\'re trying to come up w/ an insult you have to sit there and think and think... instead i just added on to their dis... and that got whoever to shut up)... As skool started i felt like everyone there was evil, blasphemers... i would constantly look down on people and mentally acuse them of being un-christian, hypocrits... then as I would go back to POP youth to sort of \"power up\" my happiness... when i looked at my youth group a second time I noticed that something happened... the friendly enviorment had been tainted by two main sources... one of which being the \"Oil\" family... and the second spawned from the Oils... let me just lable these as the \"Shawn Kara Stones\"... well Shawn Kara whom had a high influence on my skool\'s peeps and $hit decided to flock his friends to the youth group... by that time I had realized that I was in fact a nerd and decided to lable out the entire clique system in my skool (i\'ve got an actual gif. map if neone wants to see it... u\'ve already seen it Squalhan)... and from that I had come to the realization that Preps are arrogant, stuck up fools... and Shawn Kara brought such people to POP... within a matter of weeks the entire system fell to the ground and died... no longer was my church a happy place to be... it had been infested by swarms of dark preps... I could no longer be happy there... Luckily I had found my tru playas at skool by then... I am very lucky to have been in POP when I was... it showed me who I truly am...  &nbsp;  &nbsp;  Selective Hearing... - 3/30/2002  &nbsp;  &nbsp;  I\'ve got one thing to say \'bout all ya\'ll pissin on me... HAHAHAHAAHAHAHAHA!!! You think that your words can hurt me??? You\'re talking to me... ya know what? so what if i write like i\'m black... You can all kiss my W*gga a$$. Just cause I don\'t act like some stuck up band F***er who could easily burrow a holein their fat a$$ b/f and live inside of him don\'t mean nuffin\' i don\' act white. My life is better than ya\'lls cud ever be... I KNOW MY FRIENDS, MY FRIENDS ARE KICK A$$, THEY RNT IN BAND... exculding van, AND MOST IMPORTANTLY I\'M NOT TRYING TO EMULATE SOMEONE... SO THE REST OF YA\'LL CAN GET ON YO\' KNEES AND SUCK MY D***  hit me back playas!!!  FYI... LEARN TO F***ING LEAVE A SIGNED NOTE COWARDS!!!  &nbsp;  &nbsp;  Is this the end??? - 4/1/2002  &nbsp;  &nbsp;  \"momma told me one day it would happen, but she never told me when; she told me it would happen when I was much older, wish it would have happened then (is this the end?)\"  Here\'s wuts been on my mind recently...  Is life really worth living? I mean if this life is hell, logic would suggest that dying on earth would send you to either heaven or cause your conciousness to disipate... if your family betrays you, you anticipate failiure in life and death within at most a decade, you have a knack for harming all around you, and your lovelife is biting u in the gonads is life really worth living??? Not talking about suicide of course, its just... well i... i dont know, i just dont feel too good, lol. But none of ya\'ll shud feel bad for me... I have reached a state of mental clarity... hmmm, very strange. Haha, i know ya\'ll think i\'m f***ed up but when have i ever cared... that\'s why i\'m so proud to be me... cause i am me... Holla back foo\'s  &nbsp;  &nbsp;  Death, Remorse, Rebirth - 4/10/2002  &nbsp;  &nbsp;  Death...  Shannon once said, \"Bryan if you keep saying this S*** it will come back at you.\" Natuarally I laughed it off and kept on making rude and uncalled for jokes not thinking of the consequences. One day I said something which, unknown to me, horribly effected a good friend of mine... when I heard what it did to her I couldn\'t believe it...  Remorse...  When I found out why it upset her so much I was sick with myself. I stepped out of my body and looked what i had become. i was a mean and hateful indivig\'. The weekend after what I did, my parents, brother and I got in a large fight which I guess was my pennance. What i did to her was inexcusable and can only pray will be forgiven. Still, days after I found out I was still bitter and cold, even my best friends couldnt reach me. I realized a change was needed.  Rebirth...  I\'m now trying to reconcil my sins with her adn others. haha, i comb my hair now too.  but the moral of the story is: if u do bad S***, it\'ll come back adn bite ur ballz off  other than that, peace love and pimpin\'... nuffin else mata\'s... holla  &nbsp;  &nbsp;  &nbsp;  One Mic... - 4/23/2002  &nbsp;  &nbsp;  \"Yo, all I need is one mic, one beat, one stage One ni**a front, my face on the front page Only if I had one gun, one girl and one crib One God to show me how to do things his son did Pure, like a cup of virgin blood; mixed with 151, one sip\'ll make a nigga flip Writin names on my hollow tips, plottin s*** Mad violence who I\'m gon\' body, this hood politics Ackowledge it, leave bodies chopped in garbages  Seeds watch us, grow up and try to follow us Police watch us {*siren*} roll up and try knockin us One knee I ducked, could it be my time is up But my luck, I got up, the cop shot again Bus stop glass bursts, a fiend drops his Heineken  Richochetin between the spots that I\'m hidin in Blackin out as I shoot back, f**k gettin hit!\" Nas, One Mic  Here\'s a funni story those close to me should enjoy:  -Alright, there\'s this guy rite?  -He was thinkin\', \"ya know what? there\'s something missin\' in my life. I need to hook up w/ someone.  -this guy did a lot of thinkin bout who he\'d agsk out. he took the time to elmininate the fat, ugly, manish, annoying and emotionless people.  -so finally he found this one girl... this guy did some more thinkin\'  -\"Hmm... the last reltionships i\'d been in i rushed in adn got screwed over rite? Lets try a dif. approach this time\", thought the guy.  -\"maybe if i got to know her and didnt force nething, stuff wud turn out better\"  -so the guy waited and waited... boring... til that girl got curious as to who the guy was mackin\' on.  -the guy was wise though and didnt tell her cause he didnt think the time was rite  -during the time the guy waited he noticed she really wasnt the person he wanted  -finally the girl hounded the guy enough he had to tell her  -\"i really don\' know if i like you nemore...\" the guy said to her.  -the girl using all of her wisdom decided to ignor/avoid the guy in order to piss him off  -keep in mind that until the guy told that girl how he felt they were good friends... but the girl thought that they cudnt be friends nemore after he said taht.  -so had the guy kept his d*** mouf shut they\'d still be buds  &nbsp;  Tell me what ya\'ll think of that!!! HAHAHA  &nbsp;  &nbsp;  &nbsp;  One Winged Angel - 5/16/2002  &nbsp;  &nbsp;  [CHORUS]  Estuans interius  ira vehementi  Estuans interius  ira vehementi  Sephiroth  Sephiroth  [CHORUS]  Burning inside  with violent anger,  Burning inside  with violent anger,  Sephiroth  Sephiroth  [REPEAT CHORUS]  [REPEAT CHORUS]  Sors immanis  Et inanis  Sors immanis  Et inanis  Fate - monstrous  and empty,  Fate - monstrous  and empty,  [REPEAT CHORUS]  [REPEAT CHORUS]  veni, veni, venias,  Ne me mori facias  Veni, veni, venias,  Ne me mori facias  Come, come, O come,  do not let me die  Come, come, O come,  do not let me die  [REPEAT VERSE]  [REPEAT VERSE]  Veni, veni, venias, (Gloriosa)  Ne me mori facias (Generosa)  Veni, veni, venias, (Gloriosa)  Ne me mori facias (Generosa)  Come, come, O come, (Glorious)  do not let me die (Noble)  Come, come, O come, (Glorious)  do not let me die (Noble)  [REPEAT VERSE]  [REPEAT VERSE]  Sephiroth  Sephiroth  Sephiroth  Sephiroth  ahh my week has been a bitch, first i find out i cant exempt Evans\' final, then become the first in all 12 years of her teaching to be sent out into the hall, then my birthday half sucked: (when i got to dinner my parents were biznitches and evil, but i got matrix poster n mp3 playa... woo-hoo) and to top that off my application form for kroger got rejected. \"but i aint trippin im a playa i aint sweatin him...\" 2pac,  I\'m still content... damn mask... about that. if any of yall didnt know, i wear a mask...  2 emotions-happy, furiously angry  perverse and rude  stupid and ignorant  ... well that\'s about it...  under this mask,  all emotions imaginable  sensitive and compassionate  witty and ... well non-iggnoarnt(poorly spelled btw, haha)  but most importantly a good person... haha, i should start taking off my mask thou... problem is that its not nearly as fun!  well newayz, shoutoutz to my playaz!!!  to the big five, see ya at astroworld  shammhammon, hey, you\'re my pimp  glenn, AHHH you\'re crazy  and the rest can kiss my... dog...  peace, love, becoming fused with the world\'s life force... nothing else matters...  &nbsp;  &nbsp;  &nbsp;  (sigh) Life blows... - 7/19/2002  &nbsp;  &nbsp;  Darn, its been a while since I\'ve written... hmm, well it\'s just because I\'ve been so damn bored... haha. Well here\'s wuts been going down for the last... well since i last wrote:   As i\'ve said before; \"Life\'s a b*tch, drink it up\"... sadly, to this day I have no idea what that means. but i\'m guessing its that no matter how much S*it goes down you\'ve got to suck it up and get on w/ it. Well, its official, I\'ve learned that when i tell myself my life is going well or that i\'m good friends w/ sum1 i get screwed over somehow.  Now, there are many perfect examples of this but i\'ll just put in the few that recently occured:  1.Before: Hey, nate\'s a good friend, i really understand him and vise versa  After: Uh... well I\"m still working that one out, some bad S*it is still going down, i dont know wut... hey, i\'m still in the dark... haha.  2.Before: YOUTH GROUP IS AWSOME AND EVERYONE THERE IS COOL.  After: youth group sucks and i despise all who go there.  3:before: my parents and brother understand and really love me.  After: Well they think i\'m some freak because i dont try to put myself on some pedistal above everyone else like they do and therefore treat me like dog S*it.  Well that concludes the three main things that are on my, I\'m pissed off for the following reasons list. Now on to more pressing matters, (i talk like a complete idiot for one)  Hmm, I know ya\'ll(regarding all who actually read this crap) think that i\'m only 15 and taht i should have absolutly no idea what i should do w/ my life... but i am iggnorant as always (just ask nate, he\'ll agree) to that and want to plan my entire life out.  I am really pissed off because three of my 3 life ambitions I\'m going to have to decide one to pursue... sadly none of them are respected by my family members much less understood.  -Business dude: I duno, i just always loved my intro to buz class in 7th grade and would love to be in a suit w/ a bunch of other guys debating whether or not to buy or sell, (for some damned off reason that just appeals to me)  -Anthropologizer: (yes, i\'m aware that is not a real word) as most of you know i love analyzing things, especially things w/ no relavence to real life. teh main con is that i\'d have to actually do \"work\" and as Mr. Ego mentioned, no airconditioning.  -Gaming Dude: I LOVE GAMING!!! and i love making games all the more, but in order to really be a famous one i\'d have to start my own business and all that BS, but still it appeals to me. Unfortunatly no one in my family thinks thats a good idea cuz they think comp. sci. is some witch craft BS.  In conclusion... life blows... end of story, but the moral of this story is... don\'t drink and drive.  &nbsp;  &nbsp;  &nbsp;  To Kob... - 7/22/2002  &nbsp;  &nbsp;  DUUUUUDE ITS SOO LATE!!! Yeah, but newayz.  I\'m writing this because Kob has blocked EVERY SINGLE ONE of my screen names. (you gotta give him props for that, it took him like an hour.. haha) Here\'s what went down that fateful day (or evening, morning or any other part of the day it happened.);  Squalhan had this conversion rite,(i hope few peeps know who that is, o well) and despite my analasis of his personality he still believes that he will forever remain \"saint paul the second\". At the time it was really buggin\' the shi* outa me so I asked a few of his friends what they thought about this sudden change in beliefs (dont get me wrong, he was Christian before but... well he\'s obsessive now). Kob happened to be one of the people i was inquioring about this issue. I said to him, \"don\'t you think nathan will stop being so preachy?\"(well that was roughly it) and he replyed (also rougly quoted) \"yeah, he will become less active about it) So I was talking to Squalhan about him and his... evangilizing and the issue that people aggreeded w/ me that he would eventually change and I mentioned Kob was one of them. (now understand that in Squalhan\'s eyes, to change at all from that personality would be to \"fail\" somehow... i never got that) K, so Squalhan and Kob are good friends and some crazy person said, well if Kob thinks you\'re gonna fail u shouldn\'t be as good friends w/ him or at all. So Squalhan told Kob this and... well apparetnly there was some miscomunication on all three of our parts.  Kob: Thought he would change his personality somehow  Me: Told Squalhan that he would stop evangilzing  Squalhan: Though Kob was out to make him fail thus ending their friendship.  Ending note... if any of you who actually reads this crap and knows who either of them are please tell them to read this... as I\'ve said, Kob blocked ALL 32 OF MY SN\'s ARRGH!  &nbsp;  &nbsp;  Emotions... - 8/21/2002  &nbsp;  &nbsp;  Yeah, i\'ve been trying to kill them off for the last year. They\'ve brought me nothing but dispare. So I now only have 2 emotions. Bliss and rage. As the Count of Monte Cristo once said, \"one cannot feel ultimate bliss if he has not felt ultimate dispare\". So unfortunatly I can no longer feel ultimate bliss or ultimate dispare... well i\'m sure i could feel ultimate dispare but I choose not to. Eh, rambling is good for the mind, I reccomend it. Friends stray, cliques disapate, but my mind shall always remain unchanged. Eh... and for this very reason is the cause of my turmoil. I bi+ch to da** much. I wish I could just talk to myself. I\'m sure i\'d have a great time.. heck i would get all of my jokes. Heh :-).  What is the meaning of life? to better yourself or those around you? Perhaps niether. Perhaps life is simply a trail before your last breath... what does it all matter thou? If you only worry about death then you can\'t worry about life, which defeats the purpose of worrying when you die, because you never really lived. I am now wondering how I became such a hating person. My parents, life, friends, or was it simply a path I chose to take?  I\'ve also become to dam* analytical. This causes me to find flaws in everyone, also a reason why i can\'t be content. The irony of this all is that thou i can usually tell why someone duz something or acts a certain way I never learned to figure out myself. I guess I am the ultimate riddle, what am i... and why.  ~Roac Wind~  &nbsp;  &nbsp;  Empty... - 9/8/2002  &nbsp;  &nbsp;  That\'s all that I have left. My friends have left me abandonded, my parents just don\'t get my personality, laurel doesnt get the concept of homecoming (explained further down) and I don\'t understand love.  Over the summer my friends and I grew an unbreakable bond. When school started it broke, shannon decided i was a lazy F*** and nate fails to realize that I am bryan and always have been. plus i cant figure out whether or not cailyn hates me or not. I\'m so alone...  I\'m a god D*** nerd. my parents need to get over it, I am more comfotable in front of my computer screens than in front of them. They dont get anything about me, or wut i\'m like. I\'m so alone...  I met this great girl, asked her to homecoming (a ninety dollar investmetn) and she basically says she\'s gonna hang out w/ a group of freshman the whole time (which includes dancing w/ all of them. That is total bullshi*, if i\'m gonna fork out a near hundred on a dam* nite ur practically my whore, so u damn well best act like it. (shell was the acception, i was working up to asking her out) I\'m so alone...  The girl i was gonna go 2 homecoming w/ is awsome, we agree completely, but becuz i\'m a sophmore i get the feeling she sees me as a superior and not equal... I dont want her to agree w/ me cuz i\'m older, i want debate, argument, I dont want a slave (lest i spend 90 dollars on her lol) I\'m so alone...  I need help, a friend.. something, please tell me you hear my cries...  &nbsp;  &nbsp;  &nbsp;  &nbsp;  on a less biatchy note... - 9/8/2002  &nbsp;  &nbsp;  Visit my site... it has my story on it and i\'ll be posting all my diary entries on it soon too. enjoy  http://www22.brinkster.com/nezro/rwproductions/base.htm  email me...  &nbsp;  &nbsp;  Love... - 9/15/2002  &nbsp;  &nbsp;  For many years now I have been trying to find it. Though I thought I was in love several times they proved to be flawed. Nothing more than a self-created illusion to hide me from truth... taht I am indeed alone. I suppose taht I choose to be alone. Or is it that don\'t know what love is... perhaps I shall never know.  Michelle was my \"first love\". She was beautiful, I thought we were alike in every way. Aside from the simple fact that she had absolutly no feeling for me.  I guess I loved Cris as far as my idea of love goes. She bore similar feelings for me, but her boyfriend drove us apart (in more ways than one)  A year later... I thought perhaps I was on the brink of falling for Laurel, but as I tried to fawn Michelle by taking her to homecoming, Laurel proved that our mindsets on relationships are different. Perhaps my mentality is greedy, I want Laurel to be mine and mine alone, however she wishes to share a tight \"friendship\" with other guys... and w/ my past experiences w/ men in my g/f\'s relationships (diary of a mack part 1) that scares me... perhaps my last few emotions clinging on. Not to mention she constantly speaks of how other guys are cooing over her, and being the passive Xy (major inside joke) person that I am, I can do nothing about it. So is this God telling me that I am doomed to live my life alone... empty... w/o love? Or is this my primary challenge in life, to break my curse, and be loved?  &nbsp;  &nbsp;  Dying on the inside... - 10/11/2002  &nbsp;  &nbsp;  Well, looking back on this diary I\'ve noticed a sort of rhythm to my entries. I go from being in a state of total bliss, for instance when I was in youth group. To a state of anger and depression. But for the time being I haven\'t slipped outa that phase of depression. I can\'t honostly say who my true friends are anymore. I once thought the nerd clique was inclusive and friendly... but recently found out they are arrogant, hateful and narrow minded. No longer do I belong to anyone or anything... I can\'t conform for I have found nothing of any level to conform to. I can feel myself slipping further and further away into darkness. I don\'t know if i\'m going insane or just horribly depressed. I guess it\'s always been true that I could never tell what I was/am thinking.  Im no longer as articulate or strong as I once was... I am... I am, I don\'t know... perhaps I\'m dying on the inside... but I can never show it. No one knows...  &nbsp;  &nbsp;  10/17/02 - 10/17/2002  &nbsp;  &nbsp;  Externally, Bryan is a very happy person full of energy and life. Willing go above and beyond to get a laugh outa someone.  Internally, my life sucks, friends have been treachorous (not all), my lovelife shall never get better... much less exist and lastly as i mentioned before, I\'m slowly dying on the inside... life is fading away until i shall become mentally unstable.  I suppose before I leave this place.. into a deep pain where there is no return I might as well get out a few last thoughts... I\'ve never been good at writing down what I\'m feeling until recently. I am happy at school, but I think its just a shell, I\'m empty... alone... and I don\'t think anyone is responsible for it; yes, it is my own hands that have caused me to spiral into this dark pit of depression. When I come home and try to sum up why life is good... I come up with nothing.  Friends from last year, nathan and I have just been drifting apart since he \"found Jesus\", he just doesn\'t get me and plus I\'m like so far down on his \"friends to talk to list\" that when i want ot talk to him, I have to wait in a line, and when I get up there i\'m pushed away. And he wonders why I blocked him.  Shannon, well, as many of you know, she had a lengthy convo w/ nathan(me on his sn) that was nothing but bashing me (several pages long) and you can imagine how that affected our friendship... its gotten better. But I guess she\'s just tolerating me.  Ben, i guess now that we\'re not in PE we don\'t have enough time to get philosophical... or perhaps too much time. We argue on most everything now, and without fail every time he will end with \"you\'re dumb, don\'t talk to me anymore\"... hmm, reminds me of kevin.  Bret, the only person I can say nothing negative about, even during m darkest hour and my first falling from the clique he stayed with me (first falling was that convo w/ shannon) He has never been bitter or an asshole... and I should thank him indefinetly for that...  It doesn\'t help that I have several classes primarily with freshman... and quite frankly, every single one of the \"nerd\" freshman are arrogant bastards. Never in my experience of analyzing the sophmore nerds have i seen more arrogance, Calvin holds no ground to any of them (btw, that guy is so fickle on whetehr or not he\'s going to be nice to me I didn\'t even bother noting him in the friend list thing)  I shall leave my precious readers with this, One cannot find a brilliant mind if all they can do is call them stupid.  &nbsp;  &nbsp;  The end of it all... - 10/17/2002  &nbsp;  &nbsp;  Externally, Bryan is a very happy person full of energy and life. Willing go above and beyond to get a laugh outa someone.  Internally, my life sucks, friends have been treachorous (not all), my lovelife shall never get better... much less exist and lastly as i mentioned before, I\'m slowly dying on the inside... life is fading away until i shall become mentally unstable.  I suppose before I leave this place.. into a deep pain where there is no return I might as well get out a few last thoughts... I\'ve never been good at writing down what I\'m feeling until recently. I am happy at school, but I think its just a shell, I\'m empty... alone... and I don\'t think anyone is responsible for it; yes, it is my own hands that have caused me to spiral into this dark pit of depression. When I come home and try to sum up why life is good... I come up with nothing.  Friends from last year, nathan and I have just been drifting apart since he \"found Jesus\", he just doesn\'t get me and plus I\'m like so far down on his \"friends to talk to list\" that when i want ot talk to him, I have to wait in a line, and when I get up there i\'m pushed away. And he wonders why I blocked him.  Shannon, well, as many of you know, she had a lengthy convo w/ nathan(me on his sn) that was nothing but bashing me (several pages long) and you can imagine how that affected our friendship... its gotten better. But I guess she\'s just tolerating me.  Ben, i guess now that we\'re not in PE we don\'t have enough time to get philosophical... or perhaps too much time. We argue on most everything now, and without fail every time he will end with \"you\'re dumb, don\'t talk to me anymore\"... hmm, reminds me of kevin.  Bret, the only person I can say nothing negative about, even during m darkest hour and my first falling from the clique he stayed with me (first falling was that convo w/ shannon) He has never been bitter or an asshole... and I should thank him indefinetly for that...  It doesn\'t help that I have several classes primarily with freshman... and quite frankly, every single one of the \"nerd\" freshman are arrogant bastards. Never in my experience of analyzing the sophmore nerds have i seen more arrogance, Calvin holds no ground to any of them (btw, that guy is so fickle on whetehr or not he\'s going to be nice to me I didn\'t even bother noting him in the friend list thing)  I shall leave my precious readers with this, One cannot find a brilliant mind if all they can do is call them stupid.  &nbsp;  &nbsp;  Homecoming... - 10/18/2002  &nbsp;  &nbsp;  I\'m so damn lonely it isn\'t funny... this entire event is nothing more than somehting to make you feel lonely. well guess what, it\'s succeeded  &nbsp;  &nbsp;  My wonderful parents... - 10/21/2002  &nbsp;  &nbsp;  Hah, that\'s a load of bull$hit. As I\'ve mentioned I can\'t stand stupid and ignorant people. But it\'s only worse if they handle money. I spent aproximately 4 hours writing a debate case the nite before and becuz my printer blows a$$ I have to keep transfering the file to print it out. As I finished both cases I put it on a disk and took it to the shit computer to print it out on and it was completely corrupted... coluldtn be opened. So I spent the next 2 hours in the nite trying to get it to work... to no prevail. Hence, I went to sleep 5:35 and work up 5:42. My case was hoss boss too, but becuz of my shit ass old computers I LOST EVERYTHING. Now this was a mistake that could have been prevented but would be EASIER BY ALL MEANS IF I JUST HAD A RELIABLE PRINTER/ COMPUTER. So i bring them teh paper that shows i have like 12 hundred in their account and tell them I\'m buy a new computer. THEY WONT LET ME AND HAVE ABSOLUTLY NO REASON FOR IT. IF YOU CANT SEE WHY IM PISSED JUST TRY STAYING UP AN ENTIRE DAY, ALMOST PASSING OUT IN 3 CLASSES AND GETTING *UCKED OVER ON A GRADE CUZ UR PARENTS ARE STUPID. GAWW I HATE THEM  &nbsp;  &nbsp;  hug and a wish... - 10/23/2002  &nbsp;  &nbsp;  A hug... That\'s what I could use right about now. As for my wish... I guess someone to share my life with... to be with, someone who understands me. I guess in all these years trying to rid myself of anger, hatred and cruelty I have grown weak. Am I being weak by needing someone? By wanting 2 be w/ someone? Or is that just my true self emerging. *sigh* how I desire I could understand myself better  &nbsp;  &nbsp;  Counting my blessings... - 10/25/2002  &nbsp;  &nbsp;  What I found odd was that ben actually reads my diary... and that he left a note. Initially I was like, \"wtf? can\'t you b**** at me some other time? But i now realize that he is correct. I have been taking the \" i need special treatment and am now a demi-god. I want to take some time and openly thank those I may have been an A$$hole 2.  Ben, first and formost I am sorry for not realizing my fault. I tend to do that alot if you havn\'t noticed. I understand that we both get personal in our debates... well every time and I suppose I should either argue w/ you less or less passionately to maintain a friendship  Shannon, she has really set me straight this year. I now realize that i had been a slacker (and still am) for the first part of the year, i\'m still working on it but yeah, thanks shan.  Nate, He taught me about how greedy I am... unintentionally but I need to realize taht I\'m not the center of all life, (though I wish it to be tru :) And that friends shud stick w/ each other  Bret, though I havnt biatched at him I\'d just like 2 thank him for never... uh, disenting. He\'s been nothing but kind to me even when I was quite the oposite to him. Btw, Catholics pown major a$$ (pown is like own and... i didnt find out what hte p was for but heh, its still cool)  Moral of todays story is... uh, \"Xy\" haha  &nbsp;  Adiue, I bid thee good-bye - 10/30/2002  &nbsp;  &nbsp;  As some of you know I have/had a friend who recently got into drugs. Though her personality is similar to mine as it was when I was a freshman she has chosen to handle her problems... her depression differently. I once thought to have a relationship further than friends w/ this girl. But when I heard of her drug problem my heart felt compassion and so I tried to help her. However, as I extended my hand to help this girl she did not just turn it away, but push it away. This girl has tought me that you cannot help someone who doesn\'t want to be helped. Wrong decisions give birth to wrong decisions and she is making one of them. Though it is sad to see a friend slip into a darkness taht even I have not crossed I cannot allow myself to be dragged any deeper into that pit.  To those of you afflicted by depression, I beg you don\'t turn to drugs or sex. I hate sounding preachy but it will honostly do nothing but send you into an even further depression. I\'ll give you the one reason I never did or will do drugs. Friends. When I was faced with the opportunity to smoke weed (which i researched heavily leaving me moderate on the subject) I wanted to try it. But when I looked back to where it matters; I suppose my friends taught me how to live. And for that I thank you dearly.  Thank-you Bret, Ben, Nathan, Shannon and Jake. You may not know it, but you have saved my life... time and time again  &nbsp;  &nbsp;  Being Driven Slowly Insane... - 11/4/2002  &nbsp;  &nbsp;  Why do those who don\'t love you cause you the most grief? That is what I\'ve been asking myself recently. I have decided that love is over rated and is the reason that I have tried to kill my emotions. Unless you are one of the select few capable of finding someone whom you can share your life with, you find yourself tearing your hair out wondering why God chose you to be alone. So often to I try to laugh something off but deep inside my soul it burns, and it shall not stop burning until that which I have saught so long is found... Love. I have been searching for such a period of time I think my mission has blurred my vision. I havn\'t been able to figure out why those who look forsomething can\'t seem to find it, but when you don\'t give a damn it just pops up. That is the way I am. So often do I see people mindlessly using the term \"love\" and \"relationship\" but it means nothing to them because they needn\'t a journey to find it. But if I ever do find my love it will be held above all else because it is precious to me. Sadly I don\'t see any in the near future.  I ask again why those who don\'t want you cause you the most pain, and steal away your soul by constantly running through your mind. So often do I want to just run away from the idea but it is ever persistent. The hope for love hunts me down but always strikes me in the heart. This strike leaves not happiness in it\'s wake... but trials and anguish. What sickens me all the more is the people who do this to you and find pleasaure in it. They woo you over like a siren and most times you are too desperate to avoid the painful call, so time and time again you are wounded in the heart... wounds leave scars, scars leave pain and thus you have my life. My heart is now scarred and gushing with a terrible pain which I can only express to one person. (they may or may not know who tehy are)  Perhaps it is not just the person I vie over that scars me but myself. I too often make the illusion that there is hope, there is chance in my pursuit of happiness. But when I do try and find hope, it is snatched away, murderd by the bringer of pain, the bringer of love. How ironic those two qualitiese are held by one person. I am nieve... I am hurt.  &nbsp;  &nbsp;  Links in My Chain - 11/5/2002  &nbsp;  &nbsp;  Well, I do this every few months or so because my whole \"people I\'m friends w/\" status changes like the weather.  Jake: Prolly the only person who understands what I\'m going through right now because we have the same problem. (if you don\'t know what it is just look back the last five or so entries) He\'s hilarious and also an avid gamer (NERDS ROCK)  Nate: Meh, we\'ve had better times but we\'ve also had worse. Nate and I debate over anything and everything irrelatvent in thsi world. From the credibility of fan gamers to who\'d win in a fight between the all mighty Sephiroth... and some weak hobbits. (guess who i\'m rooting for)  Shan: I still think she hates me but \"meh, wutever\"  Benn: This guy has really taught me that I was/am stuck up and should try and not be... which I have for the most part. I\'d also like to thank him for all the help he\'s been in regards to debate.  Bret: He has been nothing but nice to me (he even took me home which was completely out of his way) And I also thank him for putting up with my constnat questions about debate and such. (I\'d have told myself to screw off by now if I was in his shoes)  To those of you I did not mention don\'t think you aren\'t part of my life (well, some of you aren\'t so stop bitchin) but these are the people whom I think have affected my personality the most and changed my mentality on several things.  +Peace  (haha, if you guess who i stole that from u get a cookie)  &nbsp;  &nbsp;  For no one else but me... - 11/5/2002  &nbsp;  &nbsp;  I fail to see how greed in a relationship is bad. My view on relationships are that your g/f or b/f is someone a step up from a good friend. I want to be able to share my feelings with my girlfriend. Why is that greed not justified? A friend of mine quoted this Weezer song and though I dislike that genre of music I felt the excerpt conveyed my current view on the topic of a \"relationship\" (the text in quote is the song and below it i have reflections)  \"I want a girl who will laugh for no one else:\"  --I want my girlfriend to know I will be there for them and th'),(4998,3938699,'06,August,2004','2004-08-06 00:00:00','   Today at work (Nesbits today), I realized just how much I hate my last name.  Honeycutt.  It\'s hard not to make fun of that name.  My friends all have cool names.  Cooper, Thompson, Hartman, etc.  The only insult for cooper is poopers cooper, which was uttered by Alice McDonald, only because everything she says needs to be accompanied by one.  Because all she says is s**t basically.   So this big black lady goes, \'hey bryan... uh, honeycomb?  Uh, honeysomething?\'.  I hate myself.  And i\'m putting up all these clothes that belong to people with freaking awesome last names.  I\'d give an arm and a leg to have names like some people have: Valore, Machines, Terranova.  Freaking awesome names.     I am forced to live in perpetual angst and self hate...'),(4999,3388444,'22,May,2004','2004-05-22 00:00:00','   and Dave... and Brian  urlLink     Pic  '),(5000,3388444,'21,May,2004','2004-05-21 00:00:00','   and Holzer...  urlLink     Pic  '),(5001,3388444,'21,May,2004','2004-05-21 00:00:00','   and Brian...  urlLink     Pic  '),(5002,3388444,'21,May,2004','2004-05-21 00:00:00','   After the drinking commenced... the pictures would follow.  urlLink     Pic  '),(5003,3388444,'21,May,2004','2004-05-21 00:00:00','   Oh yea, Here is a picture of my lovely lady... You can find her faggot blog  urlLink Here .  Rawr  urlLink     Pic  '),(5004,3388444,'20,May,2004','2004-05-20 00:00:00','   So yea, I didn\'t go to work today... It\'s Friday, and who wants to be at work? I watched Gandalf fly on his uber horse lastnight until 6:30 in the fucking morning. So getting up at 11:00 wasn\'t happening for me.   I think since it\'s Friday, I get to drink... although I\'ve been drunk EVERY night of the fucking week... I\'m such a lush, uber.  Dance Dance you fucking homo\'s..  Rawr  '),(5005,3388444,'20,May,2004','2004-05-20 00:00:00','   \"How about my hot pink dildo\" is the conversation in my left ear between Holzer and his 37 year old girlfriend... according to Holzer, thats for \"when im not there\". So yea. Enough about those...   So I played some  urlLink EverQuest  lastnight, it was fun... although we had to do some some CR.  urlLink Holzer  always tends to n00b it up for us... gg pathing bug chardok.   I bought the Probot album this week and don\'t really understand why I didnt buy this fucking album like the day before it came out. Dave Grohl is the man... I cant stop listening to this album. If you really loathe 80\'s metal then you are a faggot and shouldnt buy this album.   How is it, that I\'ve had digital cable for years and just now discovered On Demand... I mean, wtf. I realize its only about 6 months old... but how did something that fucking badass slip by me????? Free movies and I was retarded and thought I had to pay for them... so lastnight I rented like 16 things just to do it... all free, fucking something good finally came of Comcast.   I think that coming in to work today at 2pm was amazing... and that I\'m done posting  Later  '),(5006,3388444,'19,May,2004','2004-05-19 00:00:00','   Wait a FUCKING second... I didn\'t mean all my friends were gay. Im sorry Keelan, youre good game buddy. Holzer on the other hand, is a seasoned douche kouche...   So yea  '),(5007,3388444,'19,May,2004','2004-05-19 00:00:00','   Oh, good fucking game... I made a blog. These thing\'s are fucking retarded, yet I feel compelled to create one so I can fit in with the rest of the pretentious cock suckers ( deadwood pwns ) on the internet. So yea, I will probably post a lot of shit about  urlLink EverQuest , because it is indeed the greatest fucking game on the planet. I will probably post a lot about my hyper homosexual boss Mr. Betz. I will probably not talk about my gay friends... as most of them are indeed... gay.  Ok, enough rawr... I LIVE IN ATLANTA WHERE IT IS 90% GHETTO AND 10% RICH WHITE PEOPLE WHO TRY TO ACT GHETTO SOMETIMES. ITS SO AWESOME HERE>>> ALTHOUGH I FOUND A SMALL COMMUNITY HERE WHERE ITS WHITE PEOPLE WHO ACT WHITE SO IM GTFG RIGHT NOW>>>  Good Game...  '),(5008,4248403,'23,August,2004','2004-08-23 00:00:00','    No one speaks English at the corner of 42nd street and 5th Ave.   '),(5009,4248403,'21,August,2004','2004-08-21 00:00:00','    He started a small restaurant in China and failed recently. Around the same time, another Chinese person in New York opened up a restaurant just outside of Manhattan. They where both working 14-hour days but one business survived and the other failed.  The person in China has a huge debt while the New Yorker now has two businesses a Japanese and a Chinese restaurant â€“ he still works 14-hour days just to make $10,000 in profit every month.  They say he comes from a part of China called the Jiangsu province, apparently most Singaporeanâ€™s come from this province. They say if you are from Jiangsu and you make 1 you actually want 100 and if you have a 100 you want 10,000.  I imagine he lost a lot of money in China. But how is this possible with all the growth in Asia? I remember Mark Mobius mentioning rather casually during an interview I watched recently that money is made by growth and investments attempt to capture that growth or something like that. I sent him my resume after hearing this - I never did hear back from him.    urlLink  http://www.pbs.org/wgbh/commandingheights/shared/minitextlo/int_markmobius.html     50% of all businesses fail â€“ here is the statistics.   urlLink  http://www.census.gov/prod/2004pubs/03statab/business.pdf     Maybe profit is the bottom line in business. Maybe he failed because he didnâ€™t make a profit he just broke even for years until he broke.   '),(5010,4248403,'19,August,2004','2004-08-19 00:00:00','    urlLink      urlLink    urlLink'),(5011,4248403,'18,August,2004','2004-08-18 00:00:00','    urlLink http://washingtontimes.com/upi-breaking/20040818-105443-7992r.htm   I was in Times Square Last night debating with my friend about the future of China. But within 50 years, India will overtake China as the world\'s most populous nation. It\'s hard to project things this far into the future, and in 50 years I might not be around anyway. Is China the center of the universe or is India?   '),(5012,4248403,'17,August,2004','2004-08-17 00:00:00','    Well, I spent most of the night last night pretending to be homeless and wandering around the city talking with various security guards.  I got cold so I decided to make a bunch of clothing made up of old Wall Street Journal scraps (they worked better for some reason then The New York Times and that helped.  I found a decent sized bench looking out over the Hudson River crumpled up a few wads of paper to make a pillow and presto I was set for a good nap.   '),(5013,4248403,'16,August,2004','2004-08-16 00:00:00','    urlLink CNN.com - South America\'s \'tri-border\' back on terrorism radar - Nov. 8, 2002   This is one of the articles that convinced me that I should be very careful traveling through this region.  It is hard to say weather there is true terrorism happening here or that what is going on effects North America that much here.  For sure, the problems in Paraguay migrated later to Argentina with the bombing of the Israeli Embassy.  The Argentinean side is a lot harder to get into without documents then the Brazilian side.  It\'s still possible to just walk across the \"freedom bridge\" into Brazil without any documents.  However, I did have the sense that the Brazilians are at least making some effort to control crime and drugs going across this border.  Clearly, on the Paraguay side they couldn\'t care less what goes into their country or out of it as long as they get paid off...    I have many other things that I would like to talk about in more detail regarding this tri-border region but it is very risky to talk about details.   '),(5014,4248403,'16,August,2004','2004-08-16 00:00:00','    urlLink Israel may try to tempt hunger strikers with aromatic cookouts   Funny idea...   '),(5015,4248403,'16,August,2004','2004-08-16 00:00:00','    urlLink Excerpts from Google co-founders\' Playboy interview   An interesting interview... I do not see why so many people are going crazy about this interview with Playboy.  After all Yasser Arafat once interviewed with Playboy too.   '),(5016,4248403,'16,August,2004','2004-08-16 00:00:00','    urlLink Bush unveils plan for homeward military shift   Seems like we are entering a new era for military.  I don\'t know what this really means shifting around all these troops.   '),(5017,4248403,'16,August,2004','2004-08-16 00:00:00','    urlLink Chart - Proportion of Sites Blocked by Google Search Term - Documentation of Internet Filtering in China   I\'m in this grand debate with some of my Chinese friends right now since I\'m planning a trip to China... The other day I was walking through Battery Park in Lower Manhatten and came across a bunch of people protesting torture in China. I mentoned this Harvard study... anyway, I wanted to go review this study again just to see if my data that I was telling the guy was still true... anyway, I found the study. It\'s quite interesting... looks like the Saudi\'s block a lot of things too.   '),(5018,4248403,'16,August,2004','2004-08-16 00:00:00','    urlLink CRAIGSLIST / On the record: Craig Newmark   This is an interesting interview with Craig of Craigslist... However, I can\'t believe that Craig actually told the SF Cron. where he goes out for Coffee in the morning. SF is such a small place, what is he thinking???   '),(5019,4248403,'16,August,2004','2004-08-16 00:00:00','    urlLink Brazil to loan Paraguay US$300min for infrastructure    Brazil should be carful about loaning money to Paraguay. There has to be some other reasons why the Brazilian government is loaning so much money to Paraguay. I smell corruption...   '),(5020,4248403,'16,August,2004','2004-08-16 00:00:00','    New York, I keep telling myself that I have to leave before I start to like the place.   '),(5021,3893150,'27,July,2004','2004-07-27 00:00:00','   1. What is your favorite word? - Psycho or Aquarius  2. What is your least favorite word? - Brown  3. What turns you on creatively, spiritually or emotionally? - pain  4. What turns you off? - a controlled environment  5. What is your favorite curse word? - Motherfucker  6.&nbsp;What sound or noise do you love?  -horses running (weird and a little pretentious I know)  7. What sound or noise do you hate? - TV Static  8. What profession other than your own would you like to attempt?  -New York cabbie  9. What profession would you not like to do?  -Funeral Parlour manager  10. If Heaven exists, what would you like to hear God say when you arrive at the Pearly Gates?  -\"You did good, kid.\" '),(5022,3893150,'22,July,2004','2004-07-22 00:00:00','   so yesterday was pretty cool. my friends, Bartlett (Devon) and Reynolds (Jeremy) came over. we sat around, watched Boondock Saints and discussed the many ways to make and eat a PB&amp;J. also had a nice discussion on how kids today are not as innocent as kids were when we were kids. I personally think it started with Super Mario Bros. 2. what a fucking trip, you have to be a savante to finish that game, without going insane, that is. but it\'s true. the worst thing we had as kids was Mortal Kombat. nowadays, kids have The Suffering, Manhunt, Max Payne, GTA. certainly not bad games, but still..kids are getting desensitized...which isn\'t a good thing. but whatever, I\'m sounding a little too grizzled for a 16 year old. later that night, we met up with my other friend Farrell (Andrew) and played some water frisbee, sans Devon. good times. for the next three minutes, I have to help my dad with our cabin..it\'s finally getting taken off the stilts it\'s been sitting on for 12 years and put in a foundation. so I\'ll be working on that until saturday evening. it\'s my punishment for not having a job. better than flipping greasy cowheads all day I guess. I guess thats it for now. '),(5023,3893150,'19,July,2004','2004-07-19 00:00:00','   ok...so my week consisted mostly of watching movies and playing vice city. I spent the wonderful day with Claire last wednesday so we know that. but yeh...I had some movies rented, so I watched them. one a day for five days. good-ness. watched Dazed And Confused, Up In Smoke, Half-Baked, Game Of Death and Boondock Saints. and some others on TV and what not. now, prepare for a little spiel here. it\'s taken me quite some time to see Boondock Saints. came out in 1999. I\'ve heard about the movie for a while, from random friends from forums and reading threads about it on IMDb. I\'m kicking myself now for&nbsp;not seeing it sooner...one of the most stylish movies I\'ve seen since I first seen Kill Bill. so well shot..everything about it I loved. Willem DaFoe owned, as usual. he\'s one of the most underrated actors in Hollywood. so versatile. also seen Spiderman 2 tonight, with claire and some of her friends. I\'d say it lived up to the hype and everything. not extremely polar-opposite from other superhero movies, but it was developed well..with the Peter/MJ love struggle thing and what not. good times. the third movie is gonna be cool, although I think every Spidey fan in the world is waiting for the movie with Venom and Carnage and all those cool cats. yeh...maybe that\'s it for now? claire is leaving tomorrow. thats sad...but she should be back soon. hopefully. tata, internet stalkers. '),(5024,3893150,'14,July,2004','2004-07-14 00:00:00','   just spent the day with one of the best people in the world, Claire Barbour. honestly, she is one of my favorite people in the entire world and I\'m a better person having met her. we sat around all day, talking about everything, catching up. we went for a walk out around, watched The Neverending Story with subtitles in Spanish. one of the best days I\'ve ever had. '),(5025,3893150,'11,July,2004','2004-07-11 00:00:00','   driving today, I swirved into the other lane and almost took out some lady in a silver Windstar. I was staring at a fat lady having a yard sale on the side of the road, dressed in her pyjamas. hicks fascinate me way too much. '),(5026,3893150,'10,July,2004','2004-07-10 00:00:00','   OK so this is my first post in my new blog. well, I suppose I should get to the personal bs.  I\'m a 16 year old kid from Newfoundland, in Canada. I\'m pretty normal for my age, I think. I like watching movies, playing frisbee, hanging out with pals and the occasional pinner. as far as movies go, I\'m a big Quentin Tarantino fan. love all his stuff. I like the LOTR series, Trainspotting, Kevin Smith\'s stuff, The Godfather: Part 2 and The Goonies. music..favorite bands are the Misfits, Queen, Talking Heads, The Clash, Alexisonfire, Dropkick Murphys, Gwar and Radiohead. that it? I think so.  that should be ok for now. I\'ll update again when somethign interesting happens. '),(5027,3893150,'12,August,2004','2004-08-12 00:00:00','   friend o\' mine, Katie, had a party. close to 20 people in a very cozy house. smoked some cannons and hotboxed a bathroom. watched Enemy At The Gates with my dill pickle-lovin\' compadre Reynolds. what did YOU do last night?  PS: I have a new AIM with no one on it. jacktheworld13. add me if you\'re reading this. '),(5028,3893150,'03,August,2004','2004-08-03 00:00:00','   so last weekend, it was family-o-rama. went to my grandparents house and spent some time with relatives who were home for a visit. it seems as though every person I\'ve ever known that moved away was home this summer at some point. didn\'t get to see all of them, which was a bummer. anywho, family time all last weekend. picked up a couple books when I was out there, too. Bankok 8 by John Burdett. about a cop in Bangkok who witnesses the murder of an American marine and his partner. big conspiracy, who-can-you-trust thriller. damn good. well, I got home that sunday, just chilled out monday. Tuesday, me and Reynolds went swimming and we smoked up. good times. chilled out on my own the rest of the week. whole lotta PS2 and movies went down. Friday, went out with Bartlett, Reynolds and Shauna. met up with Farrell later on. smoked a doober around 12, hung out at a playground, being high fools for a couple hours. didn\'t do anything over the weekend. maybe swimming later tonight...maybe not.  on a sidenote, I have this sick fascination with trash TV that comes on really late at night, roundabouts 4 am here. shit like Jerry Springer, Elimidate...it just makes me laugh. I\'m glad there\'s fools like this to keep me occupied during the wea morning hours.  this summer has flown by. I have a supplementary math exam one month from today.  gotta start studying for that sometime soon...within the next couple of weeks, for sure. fuck man....my last year of high school. never thought I\'d see the day. King Of The School, baby! hah...not likely. I guess that\'s it for now. it being 4:35 in the PM, I should probably get dressed and showered soon. chow. '),(5029,3206777,'30,April,2004','2004-04-30 00:00:00','   This may be esoteric or pointless but I got stuck in a trail of thought one night and ended up with an obvious but personally inspiring conclusion. This is long but writing it help settle a bit of silt in my mind. It is writing for me, not necessarily for anyone passing though. A new purpose for my swampy bogspot!  There is a line of thinking that says that there is such a thing as fate. That is, that there are quite strong elements of pre-determination in life, if not total pre-determination. For example, any decision you make is due to the experiences you have had and the circumstances beyond you. Are you not just doing what all the pre-conditions say you will do? You were always going to have the experiences you had because of the situations that preceded it. Even when you feel like you are making a random choice, the number might just be the number that your brain would choose if it had to. Last night, I did try to think of a random number and suddenly thought \"132\"... which after a moments thought realised was just a number that was raised earlier that night.   Why am I yapping on about this?   I just read in a book of apparent fiction:  \"These people (Australian Aborigines) believe everything exists on this planet for a reason. Everything has a purpose. There are no freaks, misfits, or accidents. There are only misunderstandings and mysteries not yet revealed to mortal man.\"  I thought this was rather interesting and wanted to think about it but it was late and I was in need of sleep. When I tried to sleep, my mind returned to this discussion.   The section of the book contrasted with something James said to me in an e-mail:  \"If everything is predetermined, then it gives me an empty feeling about life and existence.  How meaningful is anything you do, if you were predestined to do it anyway?\"  Pre-determination hits upon all sorts of raw nerves. â€œJudge, I was always going to steal that car, circumstances meant that it was always ever going to be my only reasonable choice with respect to my state of mind, my past experiences, the air temperature and an a butterfly flapping its wings in Brazilâ€  So, it damages faith in real freedom, meaning and other issues, as illuminated above, like justice.   The statement about aboriginal views of everything having a purpose and reason stirred me a little though. And is still stirring me a little. For me, I don\'t have a belief that everything has a purpose, in a fixed way i.e. this book was NOT always meant to be read by me personally, so that I will have a revelation that will lead me down the track to change the world. Sure, there was a purpose to the book, but it was not specific anyone except perhaps the author but not in a â€œfateâ€ sense for me. It however can be used for â€œaâ€ purpose. Recreation, enlightenment, paperweight etc. But â€œtheâ€ purpose is undefined.  â€œEverything happens for a reason.â€ This sentence could be taken in two ways. The aboriginal way is in the â€œfateâ€ sense. This is to say, a meaning akin to â€œThe Lord works in mysterious waysâ€, that all events are part of a master plan. There are some times when life does appear to be doing that. On my birthday 2002, I had a birthday party where I selected the girl to feed me my birthday cake out of a bag at random. There were the names of about 20 girls in that bag and I drew the name Xin, who was to become my girlfriend.  But then you could say that there could be an element of causation there. Did feeding me my birthday cake become a precondition for us becoming close? But soon as we suggest a master plan, there would be implication on something greater considering our getting together to be a â€œsuperiorâ€ outcome (Or else why?) Having no affiliation with a deistic religion, I can say that I can lay this argument to one side till I have one.  Everything has a reason. The less fate-based interpretation of this sentence is much easier on the brain. It is easy to see that any occurrence has a reason for happening because we already know what the occurrence is.   Everything has a reason. Any accident, intentional or does have a reason for occurring. When I crashed my car into the side of a Morris Minor, while attempting a U-turn on Khyber Pass, I was influenced by tiredness, distraction and the desire to take a suggestion of a friend to go back up the road the other way. The Morris Minor was logically there because the driver needed to go there, maybe circumstances. Two sets of circumstances added up to the collision of motor vehicles. It was no accident (in the sense of being something that shouldn\'t have happened), although quite unintentional. Accidents are often seen as aberrations. They do not seem to be if we have perfect knowledge.  So is this pre-determination? Was it always going to happen?  A few years ago I had the argument against fate as follows:   â€œThe perspective of fate relies on a perspective that is all-knowing, and that is beyond time. You can only consider the future fixed if you have these two capacities. That is, to be a god. Only those who believe in God/ a god need to worry themselves about the implications of fate (I would suggest to Christian believers that the omniscience of God, if it is not just popular belief that he is, threatens the idea of salvation and damnation). The rest of us have no need to be concerned as it occupies a perspective that is impossible and irrelevant to our existence. Like worrying how we look to a fly. Our equal lack of knowledge is makes us equally ignorant of what all the implications of current actions are so fate is irrelevant to our human lives.â€  That was a nice â€œfend-offâ€ in my agnostic times. A lot of my arguments usually lead to the implication that I didnâ€™t have to think about anything.  But as I said, something jumped into my mind. Our personal agency as one of my lecturers would name it. It is true that every event has reasons that led to it, although the mystery might be what the reasons are exactly. So all one can do is see how their own actions and processes lead to any particular action. We do actively participate in most of the things that affect us.  In the car crash, I know tiredness and distraction led to it. I can see exactly where I went wrong. But there were some things that I did that played an active role in causing it. My choices, attitudes and my subsequent actions, with the background of uncontrolled circumstances, led to the collision.   Between circumstances and fate, we should look at it from the perspective that we are active agents in fate. For any occurrence, we can see our responsibility for them, and if we consider the occurrence adversive, we can do something about it. To take ownership of all things that happen because of us.  It is quite an empowering conclusion, but obvious conclusion.  This loops into my existing values, of responsibility. And late at night where the relevance of any particular issue is hard to see, it was a nice revelation.  I also love the line: â€œThere are no freaks, misfits, or accidents. There are only misunderstandings and mysteries not yet revealed to mortal man.\" Thatâ€™s all. You can imagine that sleeping with that in mind, waiting idly to be forgotten is something of a sleep-inhibitor. Thanks for reading, if you are.  P.S. This almost item equals the word count in my recent essay which I struggled to write. '),(5030,3206777,'29,April,2004','2004-04-29 00:00:00','   I am a person who easily slips into comfortable routine.  But in that way, I also find joy when I find the switch that puts me into activity (different from the normal).  I wonder what would happen if I had always been switched on. '),(5031,3206777,'28,April,2004','2004-04-28 00:00:00','   I\'m back only moments after my dream post to say:  I will be having Rooibus tea when I next go to the best tea cafe in Auckland. I have drunk it before, and unusually I also drank it in Singapore (served to me by a ethnically Chinese friend whom I had many interesting discussions). It is a South African tea with a \'different\' taste. I didn\'t think of much of it when I first drank it but in Singapore it was a divine taste. Now, I hear it is apparently one of the healthiest teas around. I want to see what it tastes like loose leaf and of high quality. '),(5032,3206777,'26,April,2004','2004-04-26 00:00:00','   With my blog spot still in diapers, I suddenly find myself with ideas for posts seeping from my eyes and ears. Two nights ago, an idea or revelation kept me up. I wanted to write about it but found myself too tired. \"Tomorrow!\" And off to sleep I went. Yesterday came and while walking to pub quiz, I was again struck with a compelling daydream idea. \"That would be a great post!\", two to write about now. Regrettably time did not allow me to write either so I just scraped together notes on the Uni computers. Bussing to Xin\'s house, I suddenly remembered parts of my dreams. For now, that is what I will blog, \"Fate\" and \"Leadership\" will come later.  I dreamt last night about being in a boat. A shark, Jaws perhaps, burst headfirst from beneath the deck. The boat was damaged badly, some went into the water and pushed the boat, kicking in the water. There was no fear in that part of the dream. Later on, we got to an island, suddenly I found myself in a cave passage, and both the entrance and exit were guarded by snakes. I had no exit. I awoke.  I dream of vicious animals a lot. I can remember numerous dreams with sharks in them. Dream interpretation is not a hobby of mine. I just checked now, apparently sharks can mean \"unpleasant emotions or difficult and painful materials coming up from the unconscious\" or \"dishonest friends\" or \"financial difficulties\". None of those things seem a part of my life right now. Water is meant to indicate \"subconsciousness\". Maybe I should consider my own thought as to what \"sharks\" mean in my dreams. I am interested because I can\'t remember there being fear with such vicious animals in my dreams. '),(5033,3206777,'26,April,2004','2004-04-26 00:00:00','   My negative thought for the day:  The birth of the mass media signalled the baby boom of the mass passive audience. '),(5034,3206777,'26,April,2004','2004-04-26 00:00:00','   My first post on BlogSpot.  Surrealism is an interesting concept.   Surrealism n. A 20th-century literary and artistic movement that attempts to express the workings of the subconscious and is characterized by fantastic imagery and incongruous juxtaposition of subject matter.   Here was my surrealistic day~  Life has no gravity, horizontal.. vertical.. horizontal... vertical. Day after day. Spinning through space time. Rotating.  So it was on this morning, my spine became perpendicular again (by most mathematical measures). The erection of my body from the depths of my sack of sheets was a laborious effort. Then it was down the chute. My head followed the same old watercourse, no control, none needed. I washed up at school. No splash. Just a stagger to my seat.  We learned of the blood and bones of language today. The dissecting of tongues and the discovery of the little bones in the ear sent some to bathroom in an awful hurry. Schwa. Schwa. We could see the sinews and tendons of the lecturer\'s words, not really listening to the meaning of his lesson.   He was just a pile of blood and bones anyway.   Higher I rose, climbing the walls. My company was stagnant, humming quitely. I talked endlessly.  He was filled with my silly ideas, taking note of my every point. A soulless friend. Higher again I rose. Till I came to a crossroads. A fairy danced up right next to me and took me by the hand. We talked of tales, till food came to our heads. Night came and I was whisked away by time.  I went down, and down, till there was only darkness, guided by a bright light. I went far away, and saw in the cosmos, a mad world had caught a small girl (5). Of milk, a man would drink yet still take. Of water, there seemed so much in a land so baked. Yet it was all the same. The curtain of night fell even further, and the girl was cloaked in the gloom. That was the last I saw of her.  Flying home in my seat. I could reflect on a day near complete. My body yearned for another turn, to meet with my mattrice in another horizontal slumber. '),(5035,3206777,'29,May,2004','2004-05-29 00:00:00','   Relfection after 5 days in my practicum:  It is Saturday morning and my feeling on Thursday afternoon has become reality, I have got a cold or flu. This week has been interesting but I really do need to think about the implications of my observations now and set a plan for the following week.  Observations:  # A week has gone by and I haven\'t much except thoughts to show for it. That means that I better get good mileage out of these thoughts for the time being, and planned to be more productive in the following weeks.  # Despite her experience and my limited experience, I think that my supervising teacher\'s teaching has a lot to be desired. Examples of poor classroom management abound. Two examples I will cite:   In order to do one-on-one maths and reading testing, she has repeatedly given the class loose instructions to play with some things while she works. Within 5 minutes naturally the class is doing all the things they shouldn\'t be doing. She yells at the top of her voice \"HOW DARE YOU! MY HEART IS SORE!\". In lectures, we learn that if you give poor instructions it is your mistake if the kids get diverted. I think this makes sense. Having useful activities available and prepared would be the best way to go. During planning for day of lessons, it is obvious that if you are doing 1-on-1 work then you need something to engage the other students. Getting them to play is not productive for them, and gets the kids all excited.  Another case during the one-on-one testing, was when she asked the child being tested \"What number comes after 7?\", quickly a student who \"should\" be playing with something else, comes running up and says \"8!\" before the child being tested can speak. Again, \"HOW DARE YOU! GO AND SIT BY THE RUBBISH BIN!\" So the child goes and sits by the rubbish bin and the teacher, who hasn\'t specified when the child will come back, forgets about him during the testing. After a while, I walked over to him and asked him whether he knew why he was being punished. He didn\'t. I explained it to him and asked him if he was ready to behave, to which he nodded his head. I let him join his classmates again. If someone doesn\'t know the reason for which they were being punished, it doesn\'t stand to reason that they will learn anything at all from it and will probably either resent the teacher or have a reduced self-image.  # Being that I am not impressed with the habits that have been drilled into them, I do have to bear in mind that I will need to be doing some assessments with them. So, I need to have a sense of control, otherwise when I am observed by people at the University it will not be a pretty demonstration. I read a book to them \"The Story of the Gingerbread Man\" but at the same time experimented with a few instructional techniques and I have some reason to think in that small, low-risk environment, I held the classes attention and controlled the less controllable students. But that is only a start, and I will need to build up for more challenging instructional tasks.  # Compared to the last school that I went to for practicum, this school has put little thought into how they will accommodate student learners. At the last school, students teachers had a guided tour by the school, a powhiri welcome and a proactive mentor who arranged times to meet with us (a requirement for which they get paid for by the University). The mentor at Avondale didn\'t seem to know that he had to meet with me, even though presumeably he gets paid to do so. And when we did set a time, Friday afternoon, he found that he was busy with the staff socialising event and cancelled it.  # The talented students are really talented. The behind students are very behind. We have several ESOL children, one from Russia also has some developmental problems. I had to find out the other ESOL students by observation. Two of the young polynesians were recent immigrants. It challenges a preconception of mine that Polynesians at the school would all be born here and hence have good English.  # I am not sure how well I will be by Monday, and I may have a sicky that day just so I have recovered sufficiently.  That\'s all for now. There are a few other thoughts that lie beneath the surface of my memory, so maybe I will append this later. '),(5036,3206777,'26,May,2004','2004-05-26 00:00:00','   Malcolm in the Middle  has been a favourite programme of mine for a few years. I once thought it was some stupid series copycat series, when I first heard of it and could never be bothered watching it. But one of my language students told me about how funny it was in one class, which eventually lead me to try watching it. I\'ve been hooked since. It is a superior version of Wonder Years. Far more creative, diverse in direction techniques, yet centred enough to be endearing.  One of the hidden themes, I would say, is almost a Yin-Yang kind of reality, where under normal peaceful external conditions the family is chaotic, but when they resolve their issues, the world becomes chaotic. In the episode I just saw, the family came home early from their fifth regularly timed vacation to find the whole neighbour was holding the fifth annual Neighbourhood fair. It turns out everyone else covertly planned it at this time to exclude the most disliked family in the neighbourhood.   The parents didn\'t mind this, \"Hey, there is always someone that everyone hates. If it wasn\'t us, it would be some minority group! Don\'t be so worried about what other people think, you\'re born, they hate you, you get worried about it, you die. Why do that?\" \"Mum, is this meant to make me feel happy?\" \"Why does everything need to make you happy?\"  They took part in the fair with gusto, and suddenly all their neighbours found they weren\'t all that bad... but on finding out that their previous united object of hate was no longer realistic, a chain reaction of disagreements and accusations erupted amongst everyone else dividing them and causing everyone to leave the fair in mutual acrimony immediately (the fair was desolate in moments). It was only the neighbourbood\'s disliking of them that made them so united. And then the family could do what they want.  \"I\'m gonna put dogs on the ferris wheel!\"  Gotta love it... Funnier than Simpsons (which is kinda dragging) and Whose Line is it? (which seems rather repetitive with the current host and not rotating guests). Scrubs is OK for light crazy humour too. '),(5037,3206777,'24,May,2004','2004-05-24 00:00:00','   Day Three review of my practicum:   On Monday afternoon, I felt energetic after my first day. I thought that was odd as I usually find myself drained upon leaving a teaching setting. Tuesday afternoon, 3:05pm and I felt as flat as the proverbial pancake. I had taken two guided reading groups, and saw my maths lesson turn the usually raucous class into a tremendously raucous class. I did try to maintain order, and I believe some learning was done, but once I had done all that I could, I quickly gave the reins back to the reliever teacher.  Wednesday, I had a more effective reliever teacher who was quite keen for me to learn from how she did things, and she was quite a proficient teacher in my estimation. I did more guided reading but otherwise just interacted with the kids. I felt less run down than Tuesday. Needless to say, I have been hitting the chocolate again to get me through.  Tomorrow, the regular teacher should be back. Hopefully I get to see her current methods of giving directions and getting attention and order. That is one thing that gives me some explanation for a lack of classroom control. Teachers must install some cues to get the classes attention or to bring them to the mat etc. In the schools I attended, the clapping \"tah, tah, tah-teh-tah\" was a signal for you to drop everything you are doing, copy the clap and then pay attention to the teacher. But each teacher has their own way. The Tuesday reliever did that ever-so-familiar clap yet the students didn\'t stop at all. They must have thought she was amusing herself. Once I get a hold of her cues, then I will have the commands to orchestrate the lessons I need to do for my assessments and then some.  This school doesn\'t have the spirit of my last practicum school, or a feeling of professionalism that I like. My last school felt so on to it yet collegial. Avondale however, is just a friendly staff, that is all you can say. There are advantages to it, in my last practicum school, it was hard to really mix with the people, they had their cliques but were experienced and could discuss issues with you. At Avondale, I don\'t feel left out, I can just start talking to someone about anything.   But their professionalism seems lacking, exemplified by my current supervising teacher (a huge contrast to my last supervising teacher). This supervising teacher, although sick, should have had at least prepared something. Instead we have muddled through two days of essential education for students at an important stage of their learning. That kind of neglect is abhorrant to me. I remember when I was sick at my language school, I still faxed in lesson plans and objectives, directions and all that I could to not leave the reliever in the lurch. In retrospect, that might be because I am possessive of my class, and desire control, and would hardly ever want some person to wander in and subvert my intended lessons.  In some ways, this leads me to thoughts of culture and values in schools, which is an awfully convenient pivot considering a presentation I will be doing soon for AIESEC.  So 3/24th of the way through my practicum, I march on. '),(5038,3206777,'24,May,2004','2004-05-24 00:00:00','   It has been a while since I last posted, with good reason too, I have been preoccupied with every sort of thing. I am rather preoccupied now too but rather centred on one-thing for now, which is rather pleasant.  I have been placed for my practicum at Avondale Primary School, a year one class. The teacher who is supervising me fell sick with the flu but dragged herself to work so not to leave her student (me) welcomeless. She gave up the fight and went home at midday but in tha time I was there, I could already see that she wasn\'t the best teacher to learn from. But her absence tomorrow will herald the beginning of my attempt to practice as much teaching as I can. I am in the zone.  I have had my first taste of teaching year one (those turning 6 in this calendar year), which was intriguing. They are undoubtedly adorable but missing some rather basic conceptions. They struggle to write numbers (3\'s backwards). They cannot guess words. Some don\'t seem to know their alphabet at all (the one\'s that don\'t just guess,  . One of the most fluent orally has little number or letter knowledge (he burst into life telling the whole class about what happens when a woman gives birth, rather vivid!). I also saw some of the potential trouble students did have a thirst to learn.  There is also one Ethiopian boy whose reading ability (and apparently his maths ability) are extraordinary, he read a reasonably advanced book for that age group and read it cover to cover aloud with very little prompting.  So anyway, with my supervising teacher ill, and a reliever who is fine with me taking control of some aspects, I have put a fair bit of effort in tonight so I have preparation to be experimenting tomorrow.   Maybe I am in a honeymoon period... Anyway, I will update with how my efforts go. '),(5039,3206777,'13,May,2004','2004-05-13 00:00:00','   It is cool to Koorero   Teenaa koutou katoa.  Ko Daniel ahau. Ko Ngati Pakeeha te iwi. Rua tekau maa whaa taku tau. Ko Susan raaua ko Warren aku maatua. Ko Brenda raaua ko Joanna aku tuaahine.    Noo reira, teenaa koutou, teenaa koutou, teenaa koutou katoa.  I am learning te reo Maaori, so I thought I had better see how much I can remember and actually produce a few sentences, the above being a self-introduction. I made a discovery that was obvious to many people. That is, I have access to the newly started Maaori channel. I had thought I needed a SKY aerial. So, on the advice of a classmate, I watched Koorero Mai, a programme designed to teach te reo Maaori. Boy, did it go fast! I think I learned a lot, and also got some listening practice as well as added to my stock of words. '),(5040,3206777,'09,May,2004','2004-05-09 00:00:00','   I want to change the world. I do. I want to be a lot active in the way I live out my life so I crystalised my current mental state as the following...    Books and changing oneself or the world   (this might be a little messy - I may have suffered a paradigm shift between the beginning and end of writing)   I have probably read more recently than any other time in the last 5 years. That being said, I have often had a rather negative view of books (for various reasons).   Right now, I just have negative thoughts about books that try to change your life or change the world. Books could outline how to do something, but in the end I think that it usually just results in a lot of head-nodding and no action. Reading is usually too passive for me. I think that knowledge comes from books, but wisdom (I think there is a distinction) comes from learning as you act in every action of your life (this has become a tenet of my recent life).   It is fine to read \"Dance like no-one is watching\" but to do this is actually a process that is done in action and developed, not just read and done. To dance like no-one is watching, you would need to address your self-confidence, your self-image, your self-consciousness, your dancing self-efficacy (a term from a social cognitive theory, meaning confidence in a particular skill, \"Can I pull it off?\" feeling), and find the real purpose for why you dance (perhaps one of the most important). These are not resolved by reading. By actively widening each of these aspects in every action you do is what is most important. Only then can you \"Dance like no-one is watching\".  There are many self-help/new-age save the world books out there. Some have a vital heart but malformed limbs. The intention is there but they do nothing, change nothing. I am starting to have the perhaps not completely unfounded opinion that self-help books are literary drugs, making those who read feel good. Chicken Soup for the Soul is typical. Empty platitudes that inspire good feelings.   Probably my perception of all art affects my view of books and writing. I see the process as more beneficial to the writer than the reader. I love writing, and sometimes entertain ideas of publishing, but it would probably be more as a personal attempt to develop my ideas. So reading someone else\'s enlightenment should be secondary to breaking down your barriers to greater wisdom.  There have been some books in history that have changed the world. Right now I can only think of one person who by writing a book changed anything (for good or bad):   Karl Marx\'s \"Das Kapital\". His ideas changed the world (better or worse). But in his life he was an activist too and he did have a good advocate in Engels. He was attempting some sort of leadership too. Hitler wrote a book too \"Mein Kampf\", but he backed that with very firm leadership (probably the greatest mover in his success). Most strong leaders don\'t necessarily leave books. I think most religions were only sustained by the actions of their figureheads and not the actual texts themselves.  A Chinese philosopher, Zhuang-zi, wrote a story to say how reading other people\'s thoughts might not get the \"real message\".  (http://acc6.its.brooklyn.cuny.edu/~phalsall/texts/chuangtz.html) Translated by Patricia Ebrey: ------------- Duke Huan was reading a book in the hall. Wheelwright Pian, who had been chiseling a wheel in the courtyard below, set down his tools and climbed the stairs to ask Duke Huan, \"may I ask what words are in the book Your Grace is reading?\"  \"The words of sages.\" the Duke responded. \"Are these sages alive?\" \"They are already dead.\"  \"That means you are reading the dregs of long gone men, doesn\'t it?\"  Duke Huan said \"How does a wheelwright get to have opinions on the books I read? If you can explain yourself I\'ll let it pass otherwise, it\'s death!\"   Wheelwright Pian said \'\'In my case I see things in terms of my own work. When I chisel at a wheel, if I go slow the chisel slides and does not stay put; if I hurry, it jams and doesn\'t move properly. When it is neither too slow nor too fast I can feel it in my hand and respond to it from my heart. My mouth cannot describe it in words but there is something there 1 cannot teach it to my son and my son cannot learn it from me So I have gone on for seventy years, growing old chiseling wheels The men of old died in possession of what could not transmit. So it follows that what you are reading is their dregs.\" ------------  I agree with this. Except a writer whom you have no contact with is as good as dead. Guidance from someone is essential. Zhuang-zi is dead so I guess these are the dregs of dead men. Zhuang-zi didn\'t write it, his followers did.  To make people move, you don\'t need a book passively giving them the ideas.   So if books don\'t change the world, what does the world need? You need a person to lead.  So if books don\'t change people, what do people need? You need a person to guide and aid reflection.  These of course are more difficult but rather more profound for me.  Leadership is a way to change the world. History proves it. Napoleon and Gandhi didn\'t write books. The former just needed to be in power and that seemed to make armies strong. The latter was too busy doing things like fasting and leading. I have been inspired by recent observations and an event (NZWA) for what I personally think leadership requires but that is another message. That still needs a bit more time for me to ruminate on the verbal cellulose in the paddocks of my brain.  Mentorship seems logical for a person to develop. It seems that in the rather individualistic society in which I live that deaing with our personal life planning our by ourselves to be foolish. Planning could always benefit from a different perspective. Would your physical training do better with a personal trainer / running buddy or by your own solitary efforts? Are you more likely to adhere to your plan and goals if you have had a witness who will remind you of those promises, or those made to yourself? It seems obvious. Why am I only discovering this now...?  Wouldn\'t the real benefit of a book come from reflection with someone, some planning and goal setting based on your conclusions? If you did read \"Dance like no-one was watching\" and were inspired and thought that that was what you wanted to do, wouldn\'t it be better to think about it concretely and objectively?  The concept of being interactive with a book\'s text is mindblowing. Maybe the problem lies in my own mind that takes reading books was the purpose in itself.  Writing also suffers in that it is distant from those you are trying to move. Compare an e-mail, to a letter, a letter to a person encounter. What would move you heart, mind and body to act in a new way that someone advocated, would save the world? E-mail has words but sometimes seems to lack soul. Letters have soul but sometimes lack the face. In AIESEC, we were often told that when cold-calling we should not discuss things on the phone or elaborate too much. If we give them information on the phone, they could easily just say no and hang up. If there is a meeting, then it is harder to hang-up, and also you have a bond created in meeting the person.  To have a leader is to have someone with the message burst from their every word and action. You can see their vitality and passion and it brought people along. See old clips of Hitler, Martin Luther King Jr. and Churchill. Their demonstration and words changed history and changed people. Come to think of it, have all the compelling, strong orators gone, never to be seen again?  If a person has written a book that inspired you, and they are not leading a movement to actively go out and change anything, nothing may happen of it.   So, I think I need to re-evaluate my mode of constructing the process that define the course of my life and actions. I will do that this week.  Okay, that\'s all. Anyone who wants to offer any thoughts of what I am trying to express, tell me and we can have tea (on me). Later. '),(5041,3206777,'07,May,2004','2004-05-07 00:00:00','   Playing chess at a reasonable level has many facets that non-players might struggle to undestand (like any skill perhaps).   Before you get to that level, there are the fundamentals, which themselves are quite difficult; knowing the pieces and their moves and value, having awareness of the whole board so that if one of your pieces can be taken, you prevent it. Later more complicated \"future\" awareness develops so that tactical tricks can be avoided. After this, most players have a little plateau because you have the tools to play but how do you go about getting that end result: checkmate. Perhaps a few basic attacks were learnt in developing the basic skills but mostly these are futile against players who are beyond the basics.  So then, ideas of strategy come into it, how to start the game (opening, a chess players\' serve) , techniques nearer the end of the game when pieces are few (endgame) are slowly absorbed so that the game starts to have structure and almost a sense of predictability.  Most of the above is knowledge-based. Then comes where the brain-power is really required. Calculation into the future is a skill that takes a long time to master. You need to organise your thoughts into a logical order, then use your imagination to move the pieces on the board  in your head .   Watching a good chess could be more engaging than looking at art (Xin will argue of course). Art tends to have one surface to admire and maybe some context or meaning embedded there. Just like a classical music symphony, there are themes on the surface and maybe a setting or feeling that it wishes to invoke. Chess, however, has invisible symphonies, and implicit ideas in every frame. Every move the chess player makes is just one of the many choices that were analysed, with multiple responses and possible counter-responses and so on and so forth. Each position has so many potentialities, only one of which is lived out, but all of the arrays of variations and possible lines contain drama, complexity, unbridled aggression and staunch, creative resolute defence.  When I was at the peak of my chess interest, I could stare at a position for days and find new ideas and complicated lines. I have done that again recently too. One such position I found so many interesting aspects that I was compelled to write them all down (in chess notation). I still missed the best line! (which was pointed out by my friend Paul). Again recently I have been playing another chess friend Aaron. One game, although a little dry at first livened up into a double-edged struggle, with the advantage see-sawing between white and black. The end position could have been forefully drawn by either white or black (boths sides could \"perpetual check\" or check the other king for enternity which is decided as a draw) but also both of us had a chance to win or lose. I didn\'t like my chances so sacrificed a rook to force the perpetual check.  Despite it being error-ridden (he should have won and perhaps so should have I), it was a delightful game with a lot of hidden variations.  That is the joy of chess. '),(5042,3206777,'05,May,2004','2004-05-05 00:00:00','   Ever since my return from Malaysia, I have only once had a decent night sleep (the first night back\'s record-breaking 10 hour sleep). Most of this sleep-deprivation hasn\'t been forced on me (except for one night\'s essay-writing). Today, I bought the Listener because of an article about sleep-deprivation and this has motivated me to shift my sleep-time closer to the 11pm mark. This means a cut-down on chess-related procrastination, maybe frequency on blogspot and more early afternoon essay writing.  In passing, I might buy the listener more often. It seems to be my kind of magazine. It has good topics and doesn\'t have dumbed down language. It also seems to have a more level-headed perspective (I dislike Time magazine for what felt like tabloid-journalism on it).  Anyway, with less than 50 minutes before bedtime (and I am quite tired), Oyasumi~ '),(5043,3206777,'04,May,2004','2004-05-04 00:00:00','   Two themes have collided in my brain today.  I returned from Malaysia just a few weeks ago. One of the observations I made while there is how little involved the Malay population (indigenous people to Malaysia) and the immigrant Chinese and Indian population (their ancestors mostly immigrated there many hundred years ago) are. I stayed with mostly Chinese company and I noticed a slight prejudice against Malays in terms of their culture, intelligence and other aspects. This being said, the Malays are in power and noticing the disparity in quality of life, academic achievement and wealth, the Malaysian government has positive discrimination in favour of themselves. This is an interesting contrast to New Zealand, which also has some policies that favour the Maori, but whose government is mostly controlled by the \"immigrants\". I might do some research on the extent to this positive discrimination in Malaysia just for comparison sake.  Anyway, the one aspect I am focussing on is intelligence. We just had two lectures on Gifted Children and discussed intelligence and now we are talking about behavioural problems and then Maori academic achievement. A Malaysian Chinese girl in my class commented that Malays\' intelligence is lower for genetic reasons, and that it was observable that Maoris also had dark skinned and were in a similar situation. I mentioned to her, that in NZ most people would consider this racist. I didn\'t think it being genetic was the case, or at least didn\'t think that it could be proven.  Instinctively, most liberal minded people in the western world freeze up at the implication that intelligence or achievement is decided by genetic factors. Of course, it is conceivable that it is this way. This mixes with the fate issue and self-determination issue of course. (\"If your achievement doesn\'t come from your genes, where does it come from? If it comes from your environment, then you are just fortunate that external factors beyond you have made you be successful and intelligent etc.\") But let\'s not get sidetracked.  I mentioned to her about the situation with the Maori people, that with their lower achievement and discipline problems (in comparison to other ethnic groups), it could be said by some that they \"had lower intelligence\". But this could be due to the effects of colonisation, being subjected to a few cycles of poverty, and a foreign education system that they had to adapt. It is hard to put a reason to the phenomenon.  However, Malays (in English that does have a horrible homophonic ring) did not suffer as much the brunt of colonial life like the Maori did. They would have experienced the same oppression from the Dutch and British that the Indians (often taken from their homeland as slaves) and the Chinese (who were used as manual workers) yet still they are quite behind their ethnically different compatriots.  The only thing I could really think for this is that there are cultural differences in terms of their attitude to education (which could be said for Maori too). Chinese have long put emphasis on education and I think Indians do too (although I am not as familiar with them as I am Chinese people). It is easy to consider that some cultures are more inclined toward education and working in professions. Some cultures could be more hardworking too (although the corollary of saying this is that perhaps some cultures are lazy, which is something I find hard to say). But then, soon as I emerge at this end of the discussion I thought that that assertion (\"of other cultures being more education-minded or industrious\") could be viewed as \"racist\" too.  I am a liberal-apologist at the best and worst of times. This is another issue that I think will crystalise better in my mind when I have more exposure to the lives of Maori and Pacific Islanders and maybe Malays too. It is too easy to think for a long time without actually having direct experience.  Passing view: What do you call an Indian living in the US? (you can answer that). Also, we saw a dated tape on aquatic education (swimming) where one game was \"Cowboys and Indians\". Our teacher-teacher commented that you have to call this something else now, of course. '),(5044,3206777,'03,May,2004','2004-05-03 00:00:00','   Another dream with menacing animals. There were huge, friendly, spindly spiders (though not as big as the Lord of the Ring spider, was Sheila its name?) and also a menacing lion. The lion offed a rather stupid lamb or goat that surrendered so meekly (to the slaughter!) which gave me and some other person a chance to run for it, behind a hedge. Not sure if we fooled the lion or not. '),(5045,3206777,'02,May,2004','2004-05-02 00:00:00','   A few months ago, GeoCities decided to dismantle my 5-year-old website. I hadn\'t been using it much or altered more than a few dozen words on it in the last few years so maybe I was asking for it. When friends who visit my site noticed and told me, and I found it was not a technical problem, I wasn\'t that sad. There were some things that were certainly gone forever but it wasn\'t a big loss to me personally. Now, I am starting a new ambitious project, a more complicated labyrinth of ideas.  Just last night, my girlfriend Xin told me she had saved some philosophy pages from my old website (back then when she was obviously \"checking me out\", she had even hacked into my directory and found files I didn\'t even have presented on my website, that were not complete) and she sent them to me today.  It is curious to look through because I had forgotten the things that I have written but suddenly think: \"I agree with this guy a lot!\"  Anyway, here is an excerpt of a story I started writing in 2000. I had written all my ideas in stories then. It is a discussion between an enthusiastic Christian (Devon) and an agnostic guy who masquerades as a homeless person (Davisun):  Devon enthusiasm picked up suddenly with a conception of a new idea, \"OK well let me ask you another question then. You admit that you are, in terms of reality, a blindman. Why would you seek to tell people anything about the world. Like the phrase goes, it would be like the blind leading the blind. Why convince people of their blindness or maybe, to put it another way, pull them into a strange sort of voluntary mypoia?\"  \"Tis\' true that I would probably admit that I am blind, and cannot necessary acknowledge much to be true. So what is there for one to do? Well I chose to philosophically sit down and not bother to think about the hard stuff unless the whim should come to mind. Running around willynilly, however, trying to get to a place that one has heard about, but has to charge into the darkness, following a scent, denying one\'s blindness, now what is that? They might step on a few people, trip over and hurt themselves, knock other people off their respective blind rushes to really GET THERE. Or maybe they don\'t, they might walk around in confidence. I dunno that is other people. I am thinking, why not sit, maybe flag a few people down to sit by the road side. Being low to the ground so there is no fear of falling. Being still and calm, so no risk of bumping into things. Do you not think that you are also blind to some degree?\"  Devon turned his eyes toward the group that sat across the park from him to ponder for a moment. \"Yes, I would say that I am. Which is why I would say we need guidance. I would regard the Bible as a map to find the only thing that can lead me, namely God and Jesus. I guess I should add the manditory joke saying, \'And then I saw the light\'.\"  \"But can you also say that despite the guidance you have fallen down or stood on people?\"  \"Of course, to be human is to err. We all will sin in our lives at some stage. Even if you metaphorically say that you are sitting down I don\'t see how you can claim to avoid such either though.\", Devon questioned.  \"If you think of it from my point of view, I don\'t need to. Give me an example of how I could fall? Give me an example of how I could stand on someone?\"  \"For you to fall... To fail in your studies.\"  \"How is that to fall?\"  \"How is it not to fall, why is it that you study?\"  \"Because I enjoy to learn. Regardless of failing, I still learn. How is that any more a failure than not being able to eat all of what is on your plate. There is no requirement to it.. unless you consider what someone has chosen as a standard to be taken as your goal. As it is, because I enjoy to learn it is no problem to pass exams. I like to learn things not for the exams too. If I didn\'t it do well that would be just the same as my tennis game or my squash game. I love to run around whacking the ball although I might never beat my friends who in turn might not be the greatest players themselves. What is there to compare? There is nothing to fall from.\"   Just an excerpt. I liked it anyway. '),(5046,3206777,'01,May,2004','2004-05-01 00:00:00','   Hmmmmmmmmm... I am terribly left-brained, aren\'t I! '),(5047,3206777,'01,May,2004','2004-05-01 00:00:00','   Another lateral rant. Indirectly linked to my idea of responsibility.  The following is something I noticed yesterday in class. We were learning about the legal implications of teaching. Things that seemed small but specific, some seemingly odd but with a potential reason. Anyway, here are three of the legal implications that we were told. Which do you think received the most surprise/annoyance by student teachers?  (a)  Being told we couldn\'t write student\'s full names on their art work (b)  Being told we couldn\'t show let any other student know the results of other students, or putting stars on the wall to show achievement of some. (c)  Being told we\'re not allowed to show videos in class or video more than 3% or 3 pages of a book. (d)  Being told, if we wanted to keep anyone in after 3pm we needed to give parents 1-2 days notice.  Chosen yet?   Well, even though all of the above seem to be rather odd, it was (c) that caused the most murmur and discussion. I don\'t know what made people particularly annoyed in particular but this is one rule that I am in favour of, or at least believed it was an obvious rule.  Even though we have all been raised in a capitalist society, I don\'t think people really understand the implications of being in it. People produce things for us to buy because of a desire to make money. The society as a whole will produce most of the multifarious things we want and need because they want to earn money. Movies, videos and books are all produced (and priced) with consideration for how many will be sold and how much profit will be produced.  The outrage of not being able to use videos seems to be the same desire that makes people buy illegally copied music, internet distributed music and CDs. People see this as a right, and actual prices are a \"rip-off\". The way companies fund the wonderful people who create music, write books and make movies is by selling it and licensing it out.   Maybe somewhere there lies a \"honesty quotient\" for a market. That is a number that reflects the degree of honest purchases and the the leakage through illegal actions and selling. The leakage quotient would be used in pricing legal copies. The ironic nature of the system is that the \"rip-off\" price is high because of the illegal leakage of profit that would otherwise go to the creators of the goods. People who take illegal copies are harming those that don\'t.  In the end it is more or less a \"prisoner\'s dillema\". If everyone bought and acted legally, prices would be lower. If some people cheat, the legal price needs to be higher harming those who are honest. If everyone cheated, then there would be a collapse and there would be no goods for anyone to use as the creator would not be making a profit.  Maybe we have an implicit understanding in society that there are understanding, obedient, moral people in society who will always buy legally so that means that we can allow illegal copies or illegal use of rented goods. That since there is someone who is doing the buying, then it is OK to buy illegal copies. Not many people would know any friends or acquaintances who have been fined hugely or put in prison for buying an illegal movie. The consequences as fierce as they might be, are not \"real\" in anyone\'s eyes.  Back to showing videos in school, movie companies specifically license out videos for personal use and outline the places that such can be shown. All of these things are built into the price structure of videos, books and DVDs. Even in our particular context, we are receiving money for educating and we are exploiting the creative works of someone else to do it without recompensating them. When we copy books for whatever reason, we are using what   The next time you complain about the cost of textbooks, software, music, maybe think about this. It is only your own personal interface with your values that you can come to the conclusion of whether you will or won\'t buy or not buy. I know I have bought and was thinking of buying things that were not legal copies (NZ$3 on a Coldplay album, $5 on a DVD) on my last overseas trip to Malaysia.   But I am not at all surprised by legal implications and I will have to do more thinking to see how this will change my actions in the future. '),(5048,3206777,'28,June,2004','2004-06-28 00:00:00','   Three   Sick with a cold again. Twice in a month. Thrice in three months. And three  months since I had a blood test which indicated a lowering in the number of  one kind of white blood cells (neutrophil)...  I am due for another blood test but I can\'t help but wonder whether that my  recent run of illness is a good predictor of my result in that blood test. I  might be off to the haemotologist, and earning a puncture in a  marrow-bearing bone. This is what would have last time if the doctor I saw  didn\'t mention that the decrease could have been explained by the cyclical  nature of the body.  Anyway, tomorrow the test shall be. And in three days, a result should come. '),(5049,3206777,'23,June,2004','2004-06-23 00:00:00','   Unadulterated   When do you start to feel like an adult? What does it mean to be an adult? Do you actually know? Or is it just something you realise? Or are we always the same inside, the same as the child that skinned its knee on concrete at primary school all those years ago?  I don\'t know. I don\'t feel like an adult at times, especially when I am in a new context. And when I become more familiar to a context, everyone else just seems like a \"non-adult\" like me. A knowledgeable experienced child. But not as dynamic as a child.  This is more apparent to me being around children. Although I can patently see that they are cognitively very different to present-day me, I still feel the same as them at times. '),(5050,3206777,'22,June,2004','2004-06-22 00:00:00','   Two steps forward, One step back   Oh, the vicissitudes of personal progress. Oh, how I remember consoling ESOL students with the nature of learning. I use to draw it on the whiteboard, graphically, the curve of ability over time, going rocketing up and then plateauing, maybe even inexplicably receding a little, before climbing once more to all new heights.  Students didn\'t get much relief from their frustrations of course. I think they recognised it as my scientific, complicatedly logical ploy to make them feel better when they failed to reach their own expectations.  Well, the tables have for have a year been turning and turning as I have become teacher and student, learning as I taught and teaching as I learnt. And with time, my frustration has grown. Today was when this was clearest to myself. I made several breakthroughs but at the same time tripped on unexpected hurdles, which have worn me thin.   With now just three days left in my practicum and almost all of my required teaching done, hitting hurdles is mentally taxing not because they are hurdles but because my brain recognises that the necessity to act is no longer there, and to do something requires a bigger push. But today, on several occasions, I reverted back to the most elementary, thoughtless kind of classroom management. And I was too tired to really care.  I could skeptically look at the graph, hoping to see it as a fallacy, in order to sustain my disappointment. But I know it is right, and that is my respite.  I DID make some breakthroughs. Three days to go, I could make this into something that I can really build in on my next placement. I can say that, but will I do it... '),(5051,3206777,'16,June,2004','2004-06-16 00:00:00','   Mental Stamina   I have survived the first three, rather busy, days of this week without much tiredness at all. Today, I took so many different roles in the class because of several incidental occurrences. I think I was in charge for 80% of the day. And still I feel fine...  Touch wood.  Two more days to go in this week, shouldn\'t count \'em chickens to early, no? '),(5052,3206777,'10,June,2004','2004-06-10 00:00:00','   5:00pm Oh, dear. . . . . . . . .  Mentions of teaching stamina seem so distant in the past. I am feeling flat as cat, stuck to a railway track. At about 12pm, my will crumbled and I was left just lurching around the classroom clueless. Fortunately, I wasn\'t alone and the class was not chaotic. After the lunch break I did a lesson which idea-wise was superb. The management-side of things was slightly lacking in forethought but another good class. Exhausting though. Most of my tiredness came from the late night that formulated that and other classes, so what goes around comes around.  The reliever had his last day today, back with my original supervising teacher from tomorrow. The reliever seems indefatigable. He controls the class, writes down copious notes for me and does so at a higher speed than most others. After class yesterday I commented that he should have a good rest now that the class is over. But no, he was off to tutor two Korean girls English for another 2 hours.  That is the sort of endurance I would love to have.   Now, I need to balance some time to recuperate, with some time to plan and some time to not do anything but go straight to bed. '),(5053,3206777,'09,June,2004','2004-06-09 00:00:00','   My current reliever is a great teacher, despite the fact that he has broken the law, asked me to break common teaching practice, and touches children in ways that would not be approved by lecturers at Auckland Uni (not \"BAD\" touching).  To say why he is great, he has not growled at anyone, he has contained the class, the class listen to him and enjoy him as a teacher (it makes me look bad, but it is great to watch, a child said to me yesterday \"We want Mr Newman to read us a book\". Sigh).   He is much better than my supervising teacher at giving feedback. He gave me copious observations about my teaching, all written up and explained after the lesson In two weeks, my supervising teacher only gave it to me once. Although both of their notes are similiarly coated in niceties not to hurt my confidence (I have to try hard to translate what I need to do!).  But this teacher tickles the children under their chins, which is technically not proper, he went through a boy\'s bag, which is a legal no-no, and he asked me to remain and teach a boy one on one in the class, which is a position that is not permitted.   Today had its share of highs and lows. One of my lessons was so difficult. Another went better beyond my wildest expectations! I returned energy sapped. Then the power went out, and I spent 2 hours without the necessities like hot tea. And then when I decided to take a nap, the power came on.  Now, after desperately thinking up more material for tomorrow, I neeeeed sleep. Oh dear. Night. '),(5054,3206777,'08,June,2004','2004-06-08 00:00:00','   Another important consideration for teaching is that of mental stamina. I think we all know that when you have a double lecture, or even a whole lecture straight, you can become pretty jaded.  Well, mental stamina is something that can be built up. Young children have next to no mental stamina and expecting them to sit attentively for more than 10-15 minutes is ridiculuous.  Now I know some of my lesson plans are ridiculuous.   Another reliever in today. My supervising teacher is going to be away for 3 days at a tangi (funeral). Soon, I will have seen more reliever lessons than her lessons. Typically she has set nothing much up. Fortunately we had a discussion about what things I can do this week on Friday last week. So I am set and planned.  Making lessons sometimes takes inspiration. It took me so long to think of anything decent for The Three Little Pigs. When I did think of a good order, then things ploughed through to a whole plan for the week. But it is not a brilliant idea, just a normal idea that you would expect. This sort of planning shouldn\'t be rocket science so hopefully in the future when I have a bit more experience I can produce a set of lessons (a unit), quickly without agonising over it. I had lesson one, and it went mostly according to plan. I am happy.  Talking of mental stamina, teaching stamina is another thing. After my first two weeks being exhausting, the comparatively intensive day today has not dented my energy levels much. Then again it is only the first day of the school week. I might be a mess by Friday 3pm.  Time to get my Wednesday together so I can enjoy a night out at the pub quiz. '),(5055,3206777,'03,June,2004','2004-06-03 00:00:00','   Trouble?   I think the class which I am doing my practicum has been described recently as \"off the wall\", many of the members responsible for what one teacher described as \"the two worst duties in my life\" (duty being when a teacher walks around children playing during morning tea and lunch breaks). Today, the reason was five of the boys in my class being involved in an all tit-for-tat punching war (out of the six involved).   Earlier in the morning, one of the \"trouble\" kids was particularly agitating but when told off, he started to cry. This my supervising teacher noticed as unusual for him, and asked him what really was up with him. After much persistence she found that his mother had hit him in the head that morning.   There seems to have been an escalation of these sorts of disruption since I arrived (although I can\'t see how I have had anything to do with it). These are only 5 year olds. Born in 1998. And the teacher, I don\'t see as having many productive techniques to restore order. But is it possible to do so anyway? Some of it seems so affected by external factors.  How do you regain control? Or don\'t you, just doing the best you can under the circumstances.  Next week, I do take a more administrative role over the class, and will be formally required to do some lessons. I will be thinking about this hard.  This two quotes, passed onto us by our lecturer John Hope, give me courage:  \"The Myth of the Good Teacher\" \"This is the teacher who doesn\'t have behaviour problems, who is always in control, who never calls anyone for help. \"The myth is nonsense. No teacher, no matter how skilled he is or how much experience she has, is capable of working successfully with each and every student without support. This myth is especially damaging today because of the increase in the students whose behaviour is so disruptive that a teacher must have help from the principal, the parent(s), and if possible, a counsellor, in order to deal effectively with unacceptable behaviour.\" Canter 1992  In other words, I really should let these current experiences pour some cold water on my idealistic expectations for myself. I will be hit with the worst case of disappointment if I let my expecations define an impossible goal. Plan for the best, prepare for the worst. And this one:  \"The Frightening Conclusion\" \"I have come to a frightening conclusion. I am the decisive element in the classroom. It is my personal approach that creates the climate. It is my daily mood that makes the weather. As as teacher I possess trememdous power to make a child\'s life miserable or joyous. I can be a tool of torture or an instrument of inspiration. I can humiliate or humour, hurt or heal. In all situations, it is my response that decides whether a crisis will be escalated or de-escalated, a child humanised or de-humanised\" Ginott 1972  A balance to the previous quote, because we must always remember we are the strongest agent in the classroom. We define so many of the factors. To always point to external factors (mums hitting children etc.) is to look for scapegoats and evade your own place of responsibility.  Anyway, to bed I go, to save my energy for the last day of the week. '),(5056,3206777,'01,June,2004','2004-06-01 00:00:00','   I had a sicky on Monday due to sickness, but soldiered through a busy Tuesday. More about that later.   While killing time prior to the pub quiz, I went to Borders where I found an interesting set of philosophical books. One of them, \"Life, Sex and Ideas\" by A. C. Grayling, had the following quote, explaining why people think mores of previous times were morally superior to today (and the hysteria in relation to new moral dilemmas):  \"But a more important reason is that moral panics occur because the increased availability of information about what happens in society is not matched by a public capacity to reflect upon and make sense of it.\"  I think this is so pertinent. Things like the attempted change on laws regarding legal consensual sex caused an uproar here. Maybe this is why.  The author goes onto say, for whatever advances we have made in media technology, it has not been matched to any degree with an ability to understand the significance of the moral dilemmas that occur these days. We are still rather prehistoric in that regard.  Sleep time for me. '),(5057,3206777,'30,July,2004','2004-07-30 00:00:00','   New words in my block   The primacy affect is quite pronounced when it comes to learning words by Word Of the Day. The first few words are still quite clear. Flummery et al. are easily recalled, yet recent word acquisitions are not so salient in my mind. Anyhow, I am sure my story making is etching&nbsp;at least some of the&nbsp;words into my mind,&nbsp;like a&nbsp; lapidary  quote on a ancient stone wall. Here is my latest mnemonic story:    The princess sat in the courtyard of the castle twiddling her thumbs, when suddenly her  beau ideal  for a&nbsp;suitor strode in. Her&nbsp;knowledge of charm and seduction was restricted to a  sciolism  of such skills observable in&nbsp;nursery rhymes&nbsp;and the folk-tales her mother told her as a child.&nbsp;So with a  treacly   song, she tried to&nbsp; ensorcell  him.     &nbsp;  On hearing her voice, his feet appeared to stick to the floor, as if glued, preventing his progress. He looked around to see her. She was jaw-droppingly stunning, and while his guard and jaw were lowered, his gaping mouth dried rather quickly. So he pulled a small bottle from his pocket and swigged from it twice.  The drink acted as a  roborant ,  strengthening his machismo, and making his cheeks a bright  incarnadine .  Momentarily, his&nbsp;advance toward her was halted in order to release an  eructation   of some note (a note much higher than you would expect). Then he came to her side and in a well-practice manouevre, reaching for her hand and kissing her from her wrist to elbow.    &nbsp;  \"What\'s a pretty princess like you...\" \"Where can we go?&nbsp;We need a&nbsp;place where no-one will find us!\" she interrupted&nbsp;assertively. \"To the  aerie !\" He cried. Lifting her over his shoulder, caveman-style, and ascended the stairs.    (to not be continued) '),(5058,3206777,'29,July,2004','2004-07-29 00:00:00','   When&nbsp;good feet go bad   Or should that be when bad feet go worse? The day I was going to resume running, my foot&nbsp;almost failed my fitness test. The weather was terrible so I pushed my run day further back. I have worn my foot brace for most of today, and will do so tomorrow. If it passes the test Saturday morning, a run will be done.   Cryptic when we meet   It seems wherever I go, there are  crypticians about . Or at least people to be enticed into the fold. At my last practicum, I found a couple of teachers chomping at the bit to do cryptic crosswords together. It has become more apparent that I am far from alone, nor expert, compared to some of my classmates in my course too. One such classmate is Tiri is a case in point. She tells me she hardly had a chance to start the crossword, then reveals half of the gaps being filled with letters. Her knowledge is what is even more stunning, cracking the following two:    Where those for the ferry were always late   (4) ->  STYX  (which I can understand but wouldn\'t have a chance in Hades of getting)   Well, it\'s near a disturbance   (8) ->  ARTESIAN  (which is a well, apparently)  This is only a sample of what was an amazing demonstration of clue-cracking. I am training another classmate at solving clues but she may have surpassed me in two days penetrating the concise:    Outstanding Manager   (4) ->  BOSS  (as in outstanding is something that has been embossed).  My contribution was restricted to the more arcane ones:    Attractive perch on top of the world   (8,4) ->  MAGNETIC POLE   (to which I leaped in joy for getting)    Not an area of growing importance in Asia   (4) ->  GOBI  (which is hardly direct, but proved ARTESIAN, which Tiri intuitively remembered being a word)  On the bus, with only&nbsp;one go, I cracked the final really cryptic \"cryptic clue\".    Not a figure of speech!   (6) ->  STATUE  (like who could get that? I was lucky to have two \"t\"s so I could aimlessly fit words into it).  Mind you, the above might not be right (we\'ll see tomorrow morning), but it is complete. My newly learnt&nbsp;words from this single crossword are: Artesian, outre, unbred, mal-de-mer&nbsp;and&nbsp;sea fever. Quite extraordinary.   I am feeling I am bringing a nerdish element to the lunchroom at the School of Education. But everyone is convinced I am&nbsp;a genius. \"If in doubt, ask Daniel.\" '),(5059,3206777,'27,July,2004','2004-07-27 00:00:00','   Coz\' this is the aftermath   The first week back from practicum was a struggle but it is only in the second week that the smoke has really started to clear. Post-practicum trauma was a hidden epidemic that is only now coming to light. I was stricken but with the eventual meeting with the course co-ordinator allowed me to release my burden and by doing so I have&nbsp;prevented another person being stuck with such a pedagogically deficit practicum school. Yesterday (Monday) I bumped into another agrieved classmate who was quite determined to go to the student who was due to go to her school in the next practicum and tell him how terrible it will be. I calmed her down and said that we should go to the administrator first. But she didn\'t come today. Then, in the middle of a lecture another disgruntled student suddenly unleashed pent up frustration upon a lecturer mid-lesson. It was embarrassing and awkward for everyone. The lecturer, being put on the spot, politely asked to meet with this person but was spurned twice.   After class, I met with another few students and we discussed things and then other concerns and worries arose.  There is a right way and a wrong way to air such problems. I think it is interesting that people still struggle to air their worries in a proper forum. '),(5060,3206777,'25,July,2004','2004-07-25 00:00:00','   Das Experiment    The lab  rat  that I am, I am using myself again in an experimental way. People say that eating  fish  is good for the brain. Well that is fine, because my  brain    is really getting drunk on its own stupidity. It is&nbsp;scrapping&nbsp;the bottom of the cerebral barrel, at the moment. I have lost my umbrella again. I almost lost my second orange cap. I lost my bus card and now owe $9 in overdue fees for a library book.   If I were to start eating fish again for a period of time, would I noticeably remember more or&nbsp;think more clearly? Well, I am going to find out.     Fish   are back on the menu. '),(5061,3206777,'24,July,2004','2004-07-24 00:00:00','   Endless  Sunshine  on a  Spotless   Mind   ( ESoaSM  for short)    (this is long, first is a bit of theorising,&nbsp;which may&nbsp;be fiction but is my analysis based on my minimal knowledge of&nbsp;film-making&nbsp;history, but I like the writing which provokes thinking, and then after all that, there is a review of the movie itself)    The end of the 20th century might be noted for a change in the key players in&nbsp;movies. In \"arty\" movies of the past, the director was key. Looking at such movies, you might not note any star actor. Hollywood movies were the opposite, the director is just a hidden man behind the stars who are the ones you remember about the film. There were a few blips in the past of course.  Stanley Kubrick &nbsp;movies were clearly his.  George Lucas  also became a famous name&nbsp;Maybe they were the leaders.   But later, in&nbsp;the 1990s, \"Hollywood\" directors seemed to become more prominent. They seemed to want the movies to be identified as their own, and in that way reduce the&nbsp;importance of the stars.  Quentin Tarantino  and  Steven Spielberg  are directors that easily come to mind. Tarantino\'s style was clear but&nbsp;Spielberg\'s movies are varied in their style though, and the content still not particular to his (this is not to detract from the quality of his movies, but to say that they are not uniquely his in style).&nbsp; M. Night Shyamalan  began created movies with a particular distinct style. He was also brave enough to make his name bigger on the advertisements for Signs, than  Mel Gibson  (!). This&nbsp;seemed a big statement to me.&nbsp;Perhaps, there is now an trend towards bringing \"Hollywood\" and \"arty\" together. Or at least fill in the middle-ground. More arty movies are showing at Village Cinemas, which could be showing that they are becoming more financially rewarding.   I like this trend.&nbsp;   &nbsp;&nbsp;  Charlie Kaufman , like  M. Night , has created a genre of films that are virtually identifiable as his own. They usually are bound to an unlikely, unreal pretext. Take  Being John Malkovich  where there is a porthole to a particular person\'s mind.&nbsp;They often&nbsp;contain&nbsp;eccentric characters seemingly unrealistic characters, take  Adaptation \'s lead brothers.&nbsp;His movies&nbsp;are all swathed in complex formats, with&nbsp;flashbacks, circular plots and voice-overs (which as one of his movies intones, is a movie-making sin!). But soon as these factors are set, he sets about making all the characters and the film as serious as possible.&nbsp;Any parody or irony is done with deadpan seriousness. All&nbsp;plot-tricks are done with pinpoint precision. All oddities are applied consistently.    (the following is the review section, there are spoilers in it)     ESoaSM  seems to be a distilled pure form of his previous movies. It is much more simplistic than  Adaptation  in format and concept, but still is delightfully complex (if you can understand that).&nbsp; The concept is flawlessly pulled off. The concept being:&nbsp;Joel\'s girlfriend impulsively gets her memory wiped of all her memories of you after a tiff, so he does likewise but in the middle of the procedure finds he wants to keep the memories and tries to fight the treatment, trying desperately hide his memories of her wherever he can.   The lead actors  Jim Carrey  (Joel) and  Kate Winslet  (Clementine) are wonderfully in with the spirit of the film.  Carrey  shows again that he can act very competently outside of the realm of facial and scatiological humour. Of course, he has his chance to show that experience in that form of humour can pay off in other movie types. When his childhood bully picks on him, it is just hilarious.  Kate &nbsp;is just awesome to watch. She&nbsp;has distinguished herself as a capable actor in all the films I have seen her in. But this might be her best. I might see  Iris  another day to see her act there.    Frodo  (read:  Elijah Wood ) also appears early, requiring a double-take, in the beginning and through the movie. And so does the technique that made Frodo look so small in the&nbsp; LOTR  trilogy. I just love the size of that  c  o  o  k  i  e  toddler Joel was holding!   The format is a jumping timeframe with flashbacks and the present time spliced, which at first is disorientating, but is easy to follow once you observe the cues. The dent in his car and their \"second/first\" courtship serve as signposts to align the timelines correctly. Of course, many of his memories are jumbled and the timeline&nbsp;can be&nbsp;filled in. And the movie wouldn\'t work without the jumping timeline, and it serves as part of the magic of the film.   Also interesting is the fact that there is a nice moral to the story. Memories, painful or not, are important. If you wipe them from your mind, you are almost bound to do the same mistakes. The ending is smooth too. It doesn\'t try to make things sweet and stick them back together. The reality of their past is too much to keep them together. Forgetting their differences doesn\'t make them any more compatible.   I got into the film for $7.50 because I am a poor beneficiary and got in before 5pm on a Friday. I would say this ranks equally  exceptional  with  Adaptation. Adaptation  had more variety and was more witty, but  ESoaSM  was clinically done and done with  simplicity  that I enjoy. (For the record, I only ranked  Being John Malkovich   good  but with aspirations for exceptional, while ranking&nbsp; Confessions of a Dangerous Mind &nbsp;as  mediocre  , with aspirations for craphood)     And that is the end of my film watching spree. I won\'t watch one till I use my free ticket sometime between now and mid-September. I must say that this has been an enjoyable set of movies.     MISTAKE: Sorry,  Charlie  Kaufman is the  scriptwriter  not the director, but is the scriptwriter for all the other movies too. I think my point in the theorising part is even more compelling if you think that even scriptwriters are being identified through their work! '),(5062,3206777,'23,July,2004','2004-07-23 00:00:00','   Sunrise, Sunset     &nbsp;  A now to be expected \"unexpected\" cancellation of a lecture allowed me the time to take advantage of $10 B-grade movie tickets on a Friday morning. I saw  Before  Sunset  , a sequel to the movie  Before  Sunrise  .   If you haven\'t seen the first movie (a recommended movie), it goes a little something like this:&nbsp;A young American and a young French woman meet by chance. They wander around Vienna incessantly chatting about philosophy, life and love. They irritate each other at times, but naturally and believeably fall in love. But the hook is that he has to leave on a flight early in the morning so their whole relationship is the night. They promise to meet&nbsp;9 years later. It is thought-provoking, humorous and touching.    Before&nbsp;Sunset &nbsp;is an incidental&nbsp;meeting between the two after&nbsp;circumstances prevented the promised meeting to come to fruition. It is in the same format, walking, sitting, sailing&nbsp;but all that time talking. He will be leaving back to America at  sunset  so,&nbsp;of course,&nbsp;they must do their catch-up within that time. The movie is quite different from its predecessor in terms of emotion. The first movie was primarily optimistic. The second film we can feel the pessimism, the bitterness that their experience earlier had prevented them truly bonding with their partners. It is a logical contrast of the enthusiasm of  youth  versus middle-age  dissatisfaction .   Both movies in the series are&nbsp;surely the complete opposite of  Goodbye, Dragon Inn  with any action in the movies&nbsp;being completely incidental, and&nbsp;the lengthy dialogue the heart and soul. The first movie was very likeable&nbsp;and also touching, but I always had a feeling of it being too fantastic. It reads like a story I would write (but not as cryptic) where philosophical points are illuminated in each turn of sentence. The endless fascinating conversations flow far too easily and it seems they both had saved them up just for this encounter.&nbsp;They are all the discussions you dream of having with someone.&nbsp;But if you can surpress that feeling of disbelief, these movies are wondrous.   In terms of the two movies, the second tends to trump the first in terms of feeling, but that is natural. It is more affecting and you can feel the frustration that exists between the two. But I don\'t think these two movies should be judged separately, as the second could not happen without the first. Both of them are  very good movies , but not exceptional in my estimation  (addendum)   But I would say together they make an   exceptional   movie experience. '),(5063,3206777,'22,July,2004','2004-07-22 00:00:00','   The  urlLink Corporation     Well, a 140 minute documentary, enthralling?  The Corporation  does what  Farenheit 9/11  could probably never do (although I have not seen it). It  captivates  you with brutal facts and the unambushed words of real people.    The Corporation  takes the words of CEOs, real life people and cases to both shock and give hope. There are plenty of voices including  Noam Chomsky, Naomi Klein, Milton Friedman  and of course  Michael Moore.  In fact, it is mostly composed of voices of others rather than the manipulating voice of a narrator (a la Moore),   One of the interesting sticking points of the whole movie was that in America, through the unlikeliest of processes, an incorporated company has the same rights as an human individual. This is true in New Zealand too apparently. Therefore it is legal to hold property, sue, be sued etc. But, as the film would like to point out, if it is an individual, does it have morality? What is its personality? If we consider the actions of some corporations,&nbsp;they would fulfill&nbsp;all the requirements to be considered  psychopaths  as per the DSM-IV psychiatric assessment. Do we as a society have control of these undesireables? Yes we do, if we want to (the whole punchline of the film).   At some points, it made you cringe (Monsanto\'s sins take a lot to swallow). It is often ironically humorous ( Michael Moore  was comic relief but the words from actual people\'s mouths is perhaps the most startlingly humorous). Its production is excellent. At the end, you are certainly in a state of information overload, but you can\'t help but be affected.   The director  Mark Achbar  was there at the start and end willing to do Q&A and explain deeply some of the details of the \"making of\" and his thoughts on the current system. Interesting questions did come up although he had to give it to other people sometimes, \"How much would our tickets cost without the sponsorship of Telecom?\"   As a documentary, it MUST rate as an  exceptional  one. It was&nbsp;affective, and served for me as an antithesis to my musing on political liberalism.&nbsp;It didn\'t answer everything though. For example, it raised the example of  urlLink Cochabamba  (a city in Bolivia) of why water privatisation was flawed proposition. BUT, it didn\'t say how the necessity for water privatisation was controlled. Of course, I can research that.  The&nbsp;last&nbsp;phase of the movie was one thing that set it apart in that it made many suggestions for the forms change could take.&nbsp;I am glad I watched it. '),(5064,3206777,'20,July,2004','2004-07-20 00:00:00','   Total Eclipse of the Herald~   And so it has come to pass, that in a single  day , within 7 hours which included 5 hours of  lectures , I completed both the  cryptic  and the  normal  crossword for the first time ever. '),(5065,3206777,'20,July,2004','2004-07-20 00:00:00','   Another interesting   dream ...&nbsp;  This was quite interesting just because of the  negativity  of emotion (not that common for me).   For some reason, I was on a  trail , sometimes following a few people, sometimes on my own.&nbsp; On the trail there were numerous  obstacles , and sometimes I had to walk on along a&nbsp; precipice . Eventually I crashed or fell or maybe just got stuck in a room. And suddenly their were lots of people around&nbsp;me. It  annoyed  me a lot. One put his hand on me so I grabbed his hand and tried to  pierce  it with my thumb. Everyone joking said, \"He\'s  annoyed , leave him alone.\"   Well there was more to it than that but my grip of the images was loose although I think this is the same as what remembered at 9am (it is now 4pm). Sometimes I wonder if my memory constructs some of the details to my remembered dreams. '),(5066,3206777,'20,July,2004','2004-07-20 00:00:00','   And now for another episode of WordCity:   \"The   leadership   of the city was ineffectual in every extent of the word, leading the citizens into  debauchery  that&nbsp;is unheard of in the last few millenia.&nbsp;In a desperate &nbsp;attempt to stop the exigent threat of&nbsp;complete  chaos , the councillors implemented  draconian  laws to restrain the rioting. They obviously lacked the social acuity to notice one important&nbsp;salient detail: The rioting was for the purpose of overthrowing their&nbsp;foolish governance, of which their current measures were a symptom!  Pandemonium  ensued. Suddenly a  supernal force  flew down from the skies entreating them all to be silent. The   drunken   quit their  libations . The&nbsp;ringleaders halted their agitations.&nbsp;The anxiety-pandemic, that had since infected all other regions, ceased momentarily and   peace   prevailed ever so slightly. Everyone was so shocked that they wandered in a  hypnagogic  state not knowing if they were  dreaming  or otherwise. But the  aliens  really did want a quieter Universe so were quite unsure whether extermination or persuasion was necessary to cut down the noice pollution of this&nbsp;excessively raucous   planet  .\"   Don\'t ask me what happens next~! '),(5067,3206777,'20,July,2004','2004-07-20 00:00:00','   Total   Eclipse   of the Timetable   &nbsp;    University   administration must be a  diabolical  job. At least, what I have seen today strongly evinces such.&nbsp;Most students (many of whom were rather stressed about impending   assignments   and the  thorn -in-the-side Thursday  test ) got up early to come to the timetabled 9am lecture...&nbsp;in a classroom, which in  radical  style had  OCCUPIED  by a group of DipEd( Secondary )... for their scheduled lecture. And the room timetable attested to the fact that we  were not  the rightful people to fill that lecture-room. OK. Most&nbsp;of us&nbsp;then  bull rushed their way up to the  computer  room to seize the scarce supply of  computers.  10am came. And again we clogged up the hallway to the lecture room (about 80 of us all in a small hallway), to find, no our class would not be here as another class was booked in. We held our printed timetables aloft and said  \"No!\"  in  solidarity .&nbsp;It was then that&nbsp;a  frazzled  looking replacement course co-ordinator jumped up the stairs like a  spring   rabbit , with apologies blazing, and explained a compromise, that we start our paper at 12pm and go until 3pm. \"All-righty!\" we called and swarmed our ways back to the  computer  room, which in our distraction, had been filled with the DipEd(Secondary)s! Agh.  &nbsp;  Anyway,  order  has come back to this  chaotic  campus  offshoot . In the meantime, I have probably given my assignments the biggest  growth  spurt since day one (when it burst from  nothingness , like the scientifically not 100% substantiated  Big Bang ). '),(5068,3206777,'19,July,2004','2004-07-19 00:00:00','   Blackout~  &nbsp; You know your holidays are over when you get home and  collapse  on the couch. Oh dear, not even the   invigorating   effects of  Lady Grey  could stir me.  &nbsp; Once I arose, however, in an unusual burst of  energy  I tidied the  decay  and  tartar  that had accumulated all over my desk and floor, increasing available space two-fold. I am still to  floss  the cupboard, but that can be done in time.  Strange  how this never really happened during the holiday. &nbsp; Two assignments handed in today, but they are not the ones to be  feared . '),(5069,3206777,'18,July,2004','2004-07-18 00:00:00','   Oops,&nbsp;it did it again~     &nbsp;  My morning fitness test (getting out of bed) revealed a still very sensitive, and not so kind inside ankle. Being a patient person, I am prepared to give it a bit more time to mature and refrain from being so temperamental. Till then, no run.    &nbsp;  Being at school again is quite a treat. I got up at the wrong time in that I set my alarm clock for the time I thought I usually set it, but&nbsp; that  time was actually&nbsp;the time I&nbsp;had to&nbsp;get up&nbsp;for my practicum school on time, rather than University. But I feel rather relaxed with so much time to spare. I will even hand in my assignments in the morning to get them off my mind immediately. '),(5070,3206777,'18,July,2004','2004-07-18 00:00:00','   The  Family   Thicket ~   &nbsp;  My latest review will be on&nbsp;the locally made movie headlining the film festival,  In my father\'s den . It is a story that was the portrayal of a story by a   NZ   author ( Maurice Gee ).  &nbsp;  Small town life is something that I was born into but left before things really got tricky. Some movies go into the depth of the hidden side to this quaint life, where the secrets are thicker than  water , and so is the  blood  of the families that form small communities. This is one such movie.  &nbsp;  To understand the characters, we all need to have some sense of   depth   perception. The protagonist,  Paul , is&nbsp;quite pained at the best of times. In fact, we can see pain in many of the characters. But the search for the root of the pain is the quest of the audience, as much of it is known to the characters.  &nbsp;  Simply put, it is a story of  homecoming,   alienation ,  friendship  and a  disappearance &nbsp;possibly due to&nbsp;  foul   play, or is it a run away? There are  hsalf \'s and  flash forwards that are crucial to the story but often to the detriment of the orientation of the viewer.  &nbsp;  As a production, it is most impressive. The acting shows that   New Zealand   does have much depth.&nbsp;The cinematography and sets&nbsp;were stunningly  realistic . As it was a premiere we got to see those at the   heart   of bringing it together including most of the cast.  &nbsp;  One thing that came across quite coincidentally was the links between the very different film I reviewed before,  Old Boy . Both have strange hidden plots of  real ,  accidental &nbsp;and  imagined &nbsp;incestuous relationships. The hearts to both  puzzles  lie with the question that is not focussed on. And similarly, when the final pieces of   the  puzzle  is put in the story becomes credible. For me, I would rate it as a   good   movie, also possibly stretching into the  exceptional  category. '),(5071,3206777,'17,July,2004','2004-07-17 00:00:00','   Things that go.....    &nbsp; In the night, I had an interesting  dream . It was diverse with lots of different strands, with a dark  gothic  feel. Somehow the first dream was converted into a book I had written in the second dream. Apparently Ying read it in that dream and said it was good and I should publish it. And the third dream had a nasty supercreature in it. Anyway, in the morning I found one of my curtains had been pulled open and my pyjama top removed. Aside from the possibility that there is some  c  r   e   a  t i  v    e  burglar who doesn\'t take things,&nbsp;I assume that it was  I  who did these things. I have no memory of it whatsoever. Maybe it is the sleepwalking streak coming out again. It has been a while since I did something  completely  without awareness and memory in my sleep (of course sometimes I can\'t remember things well when awake...). When I was young I once let my eldest sister into my house without remembering. '),(5072,3206777,'17,July,2004','2004-07-17 00:00:00','   Strains~       &nbsp;  So, it had to happen some time. But this morning I gained a cranky right ankle/shin. It was fine last night, I know because I checked. And then this morning it was rather tender and not nice to walk on. So back goes on the ankle support and I will rest from my alternating-day running schedule for the weekend. '),(5073,3206777,'16,July,2004','2004-07-16 00:00:00','   Oh man~    Crypticity sure is abound!      It is now quite apparent that        urlLink www.blogger.com &nbsp;are'),(5074,3206777,'16,July,2004','2004-07-16 00:00:00','   How do you  feel?    I have seemingly dropped in a pit of melancholy without the choly, from which my limbs can get no traction. I am not one to feel sad about much. In a recent \"Bar News\" described depression as anger without enthusiasm. Well I might be stricken with depression without sadness. Ennui, but without the exotic sounding charm.  I hate to say this but: The sooner this holiday is over the better. I can get start doing stuff with consistency. I will hand in substandard assignments and have to wait weeks before the \"punishment\" in receiving a grade.  Anyway, distracting myself for a moment, I was thinking about the depression of others.  Nuch of the modern music I listen to is often done by seemingly-depressed people. I cannot imagine Radiohead, Tricky, the Eels, Coldplay, Massive Attack, L\'arc~en~ciel, Tool/Perfect Circle or TaoZhe producing a cheery song (David Bowie can though). There is real feeling there though. Not the unfortunately popular strains of pop you can hear to readily.  Looking through the blogs of others, it is easy to see that there is a self-tormenting, dare I say it \"artistic\" bunch that like to wallow in sadness. Perhaps it does make for affecting words, music and art. I have a theory for this of course, although it is hardly original. To have a desire for greatness, people need to suffer. I think that is something I have lacked in my life actually. In the past though, there was plenty to be vexed by. Poverty or war and the like.  But those who have some desire towards experiencing and producing exemplary art (regardless of its form) need to delve into those things that can cause suffering in life, bring them close to mind, taking themselves into the coldness of life, and its natural vicissitudes, a strain of existentialism perhaps. Even from the comfort of modern life, there is suffering without pain, or with pain if they want to self-administer.  And hopefully from this hurt emerges a truly feeling voice or a simple expression of real expression and not the saccharine pipe-tune of some life \"charlatan\". Like me.  Probably real genius doesn\'t depend on the tempering one\'s heart with fiery coals of despair. It can emerge with force and passion without. But such tapped talent is rare. I cannot think of such. Or maybe all geniuses are destined to suffer a little which triggers them. Or maybe those talented who are not subjected to despair or self-brutalised are those that don\'t find an emotion on which to hinge their ability onto.  Either way, I am not seeking to be an artist of the kind that is recognised as an artist. I simply want to get some direction to overcome the inertia. '),(5075,3206777,'15,July,2004','2004-07-15 00:00:00','   Interlooper~   I jogged around the greater block in 16:35. I should be able to shave 1-2 minutes off that should I actually press myself to run (provided I can make it to the end).   Movie news   Anyway, my movie review for today is \"Goodbye, Dragon Inn\", which I saw last night.   Before going to such a film, it is advisable to be familiar with the kind of movies the director makes.  Vive L\'amour  was the first movie I saw of his. It was devoid of almost any dialogue or script. It made an art of understanding feelings and intention purely from the action observed.   Goodbye, Dragon Inn  is done by the director  Tsai Ming-Liang , a Taiwanese director who seems to enjoy this sort of movie. It takes some adaptation of the viewer to enjoy watching them. If someone were to inadvertently see this film without understanding that they might be somewhat confused.  The movie is set in a cinema, whose popularity has somewhat dwindled. That being said, there seems to be many people wandering through its cavernous architecture. There are few seeming real people there, but one is a young man who sneaked into the cinema, who we only find out later, is Japanese. He seems out of place in amongst the other denizens (much like someone who doesn\'t understand such films might feel). All he wants is someone to light his cigarette. So he tries languageless cues to get people\'s attention. He fails most of the time. All through the sounds and artificial dialogue of the movie being echoes throughout the place, emphasising the lack of speech from the \"characters\". There are some bizarre scenes. With a door shutting by itself. Another opening. Possible homoerotic themes. And then the endless people passing through the passageways the Japanese guy finds himself. In the first dialogue of the film, eliciting applause from one of my fellow theatre-goers, it is revealed: \"This cinema is haunted,\" and then followed up by \"It is haunted!\" We never know if the Japanese man understands but he finally gets his light before saying in Japanese-accented Chinese \"Wo shi ribenren\" (I am Japanese). \"Sayonara\" comes back the reply. Regardless of whether he understands what was told to him, later, in probably the most comic moment he is scared witless and runs from the cinema.   All the time, the lady at the front desk (who has a whithered or prosthetic leg) struggles around the cinema on different quests.   On the whole, I enjoyed the movie (more than  Vive L\'amour , but perhaps only because I had checked my expectations for the movie). I wouldn\'t rank it any higher than \"Good\", just because it doesn\'t have in a movie everything I look for, but it was a nice excursion into a completely different world. It is these kinds of movies that you can lose yourself in easily though. Another theatre-goer remarked the similarity of the screen flow and that of the movie. The movie on the screen in the movie was a martial arts movie. But often in such movies, it is just lurching from one fight scene to another with justification in between, before ending and fading without a trace.  Goodbye, Dragon Inn  had this without the fighting.  I may see one of his other films  What time is it there?  in the future. '),(5076,3206777,'15,July,2004','2004-07-15 00:00:00','   Heh. This will probably be of no interest to anyone but Radiohead. But \"this is what you get\" when you take songs from a band and put them to a banjo.  http://www.hardnphirm.com/rodeohead.html '),(5077,3206777,'14,July,2004','2004-07-14 00:00:00','   What was that you said?   In amongst my chronic short-term amnesia stories, comes the truly classic. This one precedes my morning run just reported. I already knew I had misplaced my watch the previous day so I knew that had to be found first to time my run. I then realised my ankle support (which I use for my running) had also wandered and needed a search party. The ankle support was in a conventional hiding-place (the laundry basket) but the watch came out in a truly bizarre fashion. I was hunting through the piles of Uni-related stuff for it when suddenly I saw a book I had on 3-day short loan that had to be back two days ago. Short-loan books sometimes are charged overdue by the hour, so I raced to my computer to check if I had received a notice etc. But just before I arrived at my computer I found that I had, in my hand, my watch...  I have no idea at what stage I picked it up. But there it was.  After the panir cheese experiment of 2 days ago with lemon juice (very successful, very yum!), I am now trying the alternative recipe using the yoghurt that I made a few days previously, it looks promising so far! '),(5078,3206777,'12,July,2004','2004-07-12 00:00:00','   Where is the asymptote?   I ran this morning and registered 9:10, a mere 7 seconds faster than last time. My ever-receding target time is now just to inch my way past the 9 minute mark. I think where any extra speed will come from is the beginning where I am usually sluggish and usually have the same doubts whether I will actually be able to do the whole circuit without stopping. I have two time markers to check my progress. The 2:48 intersection and the 6 minute roundabout.  So up steps plan B! I will shift to the greater block for a week before coming back to the lesser block. Practising the greater block will give me a bit of variety and stretch my endurance rather than speed. '),(5079,3206777,'12,July,2004','2004-07-12 00:00:00','   Use it or lose it   \"In every rowdy party, where most of the revellers drink with gusto, congeries of teetotalers are to be found. Sometimes there is one in each corner of the room. Dry but controversial discussions are started and sometimes things get heated. Overweening self-righteousness amongst a group tends to lead to acrimony after all. The more rational would speak with impassioned polemics, which the more radical take as pure unctous flummery. Of course, there would also be the happy-go-lucky member whose ludic comments and mimicry would break the tension momentarily. Usually the discussion was something as irrelevant as how we should remunerate the politicians. This kind of subject would drive most to drink in short time.\"  Well, that was a week of \"Word of the Day\" words all incorporated into a vaguely sensible story. Don\'t ask me what happens next. '),(5080,3206777,'11,July,2004','2004-07-11 00:00:00','   Decreasing  MRI    Another run this morning, and my marginal rate of improvement has plummetted! Ah well: 9 minutes 17 seconds.  It was rather frosty this morning which almost made me unable to turn off my stopwatch due to numb, uncoordinated fingers. Instead of nausea after the run, I gained a cough. Maybe my respiratory system has noticed that in order to increase efficiency it has to clean up my airways.   I think so too, at the start of my run I was struggling and thought I wouldn\'t even get close to my last time. But as is usual when you are improving, speed gains just come from nowhere. '),(5081,3206777,'11,July,2004','2004-07-11 00:00:00','   Predetermination   To have disbelief in predetermination is a rational thing to do. That is, if you subscribe to the same logic as Pascal\'s wager.  If you believe in predetermination, and predetermination exists, you enjoy that things are secure and will work out. But if predetermination weren\'t true, you\'d be left disappointed or confused.  However if you don\'t believe in predetermination, and predetermination is true, you will end up wherever you are meant to end up, and if predetermination were a false belief, you\'d be right and thinking correctly.  So to disbelieve predetermination is a win/win situation, whereas there is a risk of disappointment in belief.  That being said, I have had lots of moments to think that something guides the coincidences in my life. But it is far more interesting to think they are pure coincidences. '),(5082,3206777,'10,July,2004','2004-07-10 00:00:00','   urlLink    \"Who are you calling old? Eh?\" (not an actual quote from the film.&nbsp; urlLink '),(5083,3206777,'08,July,2004','2004-07-08 00:00:00','   I will review movies I see from now on on my Blog.  I started my Auckland Film Festival campaign last night with the Korean film \" Old Boy \".  The premise of the film is simply that one night, a quite drunk man, is abducted and kept in a makeshift prison hotel room for 15 years. Every now and then valium gas washes over him. Knocking him out for haircuts, new clothes and things like that. Food is given. He doesn\'t know that it will be 15 years that he is in for and doesn\'t know why. He verges on madness many times. And then... He is released, waking up on a rooftop, with nice clothes and eventually he is given a cellphone and a wad of money. At this point I could honestly say that I couldn\'t imagine a way that this would ever acquit itself logically.  If a film hinges itself on plot, then the plot must come through for it. I like movies that can come up with a unexpectedly decent end. I can say that was part of the reason I liked  Matchstick Men , it was well-crafted (although, apparently with a couple flaws) and it has a kicker of an ending.  I enjoyed  Old Boy  in a similar way, perhaps more so. It delivered a visceral, explicit tale of revenge (in more ways than one!). There was nothing concealed except the plot, which although I had my doubts at different points, clarified all of those doubts and more in the long end sequence. It is a story about revenge bur perhaps more importantly shame.  The acting is great. Some of the scenes are improbable, but entertaining all the same. If you cannot take scenes showing the excessive infliction of pain, or sex scenes (that become much more \'illicit\" by then conclusion of the movie), then don\'t see it. It was a funny, enjoyable movie. Recommendable by me. In my ratings of films it easily makes it into the category of \"Good film\", while potentially challenging \"Exceptional Film\".  It won the Grand Prix award at Cannes.  One small side note is that the epilogue (the bit at the end, concluding the drama) was filmed in New Zealand.   Park Chan-wook  was the director, whose work includes  Joint Security Area  (an exceptional movie in its own right). So I think I will be looking out for his future efforts or his other work, the interestingly named  Sympathy For Mr Vengeance . '),(5084,3206777,'08,July,2004','2004-07-08 00:00:00','   \"Don\'t you wonder sometimes \'Bout sound and vision?\"    Sound:  In my classical music tastes, I don\'t particularly like Mozart (except for Symphony no. 40) to anything like the appreciation I have for Beethoven. And there are many composers I like better than Beethoven. Dvorak is one that I am interested in exploring. Serenade in E major is to die for.  Well, with the sound out of the way, now for the vision.   Vision:  For most of the year, actually since the disappointment of the way my previous birthday party logisitically went, I have been contemplating another kind of an event. This crystallised in to \"Strange Night\", a possibly regularly organisable event, which is constructed by people, rather than by a person.  Well, that was the idea. But ideas tend to just sit and observe as time goes by. Without the vision for the event, there is nothing to provide one with the drive to get it done~ Suddenly last night, I was inspired. It was not that what I thought and then proceeded to scribble down was anything revolutionarily great, but that suddenly I caught how I would like such an event to go. An idea of perfection, a goal, something to strive for. And so now all I need is to clarify a date, send out an e-mail and it will be done. Cool, eh.   \"I will sit right down Waiting for the gift of sound and vision And I will sing Waiting for the gift of sound and vision. Drifting into my solitude Over my head  Don\'t you wonder sometimes \'Bout sound and vision?\" '),(5085,3206777,'08,July,2004','2004-07-08 00:00:00','   Time and Change   I whipped around the lesser block in 9:33 removing 1:04 off my time, which was pleasing at first, but easily rationalised. What runner, when starting running again, does not get big gains? Also, I can remember that the first time I ran I was being gentle on my cobweb bound legs, and started at a conservative speed. This time at the start I automaticallk launched into what would have been a natural speed, which with my fitness was not going to be easy to maintain. I have to learn what my sustainable speed for a distance is, and then constantly push this. I felt a little nauseous after the run, meaning that my body was stretched. Either way, I hope to get within +/-5 of 9 minutes on my next run. '),(5086,3206777,'08,July,2004','2004-07-08 00:00:00','   urlLink    I can see you! Yes, this is a test~&nbsp; urlLink '),(5087,3206777,'05,July,2004','2004-07-05 00:00:00','   Slow start   Holidays bring out the worst in me, I am sure. I am so lethargic and distractable and currently running at about 20% efficiency at everything.  Anyway, on a mildly different strand.  I, for the nth time, started running again. Getting back into running fitness is very much a new challenge for me, because ever since I was 8 I went running with my sister (she was a budding athlete) I had a good degree of latent fitness. When I went running at high school, it took only a few runs to get into \"gear\", find the \"zone\" and dash around.  Now, after the inactivity of university and teaching and the intervening foot problems, that foundation has been eroded. I struggled to keep running for 10 minutes 37 seconds for a run around the \"lesser\" block (there are two different \"blocks\", the lesser is naturally the smaller of the two). In my running days, I think I may have been able to whip around it in 6-7 minutes.  Remembering back to when I was at high school, I would almost have the fantastic motivation for running that I had to be fit and ready in the case of an imminent terrorist take-over of the high school (because you would need to be able to outrun etc.). Fortunately it was just the fitness that I saw as necessary, I didn\'t go to the firing range to practice using a firearm.  I need to get something to make me continue running. Maybe I should go back to competing against myself, trying to ground that 10:37 down to 8 minutes, while run walking, then running the \"greater\" block. Ah well, it is a plan. '),(5088,3206777,'05,July,2004','2004-07-05 00:00:00','   Fruitfall  (a.k.a Fruit F\'all)  They covered the ground, Which was more orange than green Dewy grass wet my feet The wind blew again And no matter how many I pick up There always seems like so many on the ground  Is the neighbour watching? Donâ€™t catch her eye, or else she may wave and come out Pick up the oranges And check for holes and slugs, Fungi and bugs. But still so many remain fallen down  The Foodtown plastic bag is near full, It might very well break Throw the little ones back And feed the tree and birds I have too many to eat Does anyone want any?  Honestly! I have lots, come and get them '),(5089,3206777,'05,July,2004','2004-07-05 00:00:00','   What problem?    And so, despite my prognostications and mildly uneasy open thoughts on the subject, my blood test has returned without any cause to be alarmed. No need to even go in for a chat with the doctor apparently, or so said the message on my cellphone. I want the raw numbers to compare with those from my preceding tests but otherwise I am pleasantly relieved.  But then what of my run of illness?  Maybe I wear myself down too much.  Maybe there is something wrong with what I eat. Or perhaps, I should just remind myself that getting sick is all part of the rite of passage for a beginning teacher, and that is what I am becoming~  Maybe I have been thinking too much. A chronic illness perhaps.  I wonder whether I should go through with getting the health insurance I was getting. But I can wipe medical fees from my budget for now, and that has to be good for my liquidity until October the 1st.   Any~ways... '),(5090,3206777,'07,August,2004','2004-08-07 00:00:00','   Language exchange   After a hiatus of 2 years, I have restarted doing language exchange. This time it is with my newly acquired language te reo Maaori. My exchange partner worked in China for a brief time and had tried to pick up Chinese at that time but had forgotten a lot, so I used my Chinese knowledge to trade for Maaori practice. She was quite shocked that I had quite substantial Maaori accumulated in my memory banks and was making sentences quite readily and even asked questions about how to say sentences that she struggled to know how to say.  It was productive though, it was so nice to hear sentence structures and be able to get an explanation. Sometimes the meaning just flowed into my mind, other times my brain froze under the intensity of a string of words which it declared as familiar but didn\'t want the further commitment of identifying. '),(5091,3206777,'06,August,2004','2004-08-06 00:00:00','   Evidence of the advantage of developing a cryptic mind!   urlLink http://crpit.com/confpapers/CRPITV30Simon.pdf '),(5092,3206777,'06,August,2004','2004-08-06 00:00:00','   The Finest Cryptic    For a cryptic crossword fan like myself, crosswords bring pleasure and frustration. The biggest frustration known to crossword solvers would probably be a clue that you have many of the letters (provided by intersecting answers) yet still cannot crack the clue or fit a likely word.  Two such clues have a risen in the last week:  1.  Time-worn tea ceremony  (5)  letters supplied: T _ I _ E  This clue was impregnable to both me and the cryptic magician previously mentioned in this blog. We concluded that it must be a specific technical or archaic word that neither of us knew. It was only when we got the following day\'s paper that we discovered we were wrong. The answer is a logical and reasonable answer indeed yet we could not crack it.  Just yesterday, there was another:  2.  With which the Orientals show agreement  (4), letters supplied: E _ E _  This clue I was contemplating many times during an eight hour period where I also got the intersecting letters. The answer, which is actually reasonably obvious in retrospect, came to me intuitively rather than logically when I was tired and frustrated. But getting it brought ecstacy.  These two reflect a fine distillation of crypticity. A real mystery waiting to be cracked. Often clues are far too illustrative of the method by which they are to be solved. These two with their utilitarianism and simplicity show that although frustrating they are the finest cryptic clues.  (the answers will be in the comments page later, clue, they are essentially the same mechanism)   This message is a reconstructed version of a previous D.O.A blog I wrote last night. '),(5093,3206777,'05,August,2004','2004-08-05 00:00:00','   The previous blogs have been eaten by www.blogger.com. Oh well, I guess some things were never going to happen. Goodnight.  (this is posted in via e-mail, a good function)  Find local movie times and trailers on Yahoo! Movies. http://au.movies.yahoo.com '),(5094,3206777,'03,August,2004','2004-08-03 00:00:00','   The Return of the Fierce Creatures   Another assorted dream brought back the big cats to my subconscious.  I have had several dreams lately about being in my teaching course. The classes are nothing like mine but we are learning teaching. None of the people in the class are from my teaching class either (or perhaps one or two that I can only vaguely remember). Anyway, last night I had to collaborate on a lesson to show our class with an Indian classmate. We did it on   tea  , a good topic. We were preparing the lesson when the guy suddenly put two  eggs  in the cup we were using and I just said \"Oh, that is not the kind of   tea   I make.\" Anyway, while he got the tea ready, I found some wonderful ornamental vessels from which to drink the   tea   from, really nice ones although the shape was a little odd. My memory gets a bit fuzzy, and I think my classmate had rather illicit relations with a teacher in the interlude before we presented...  Then suddenly I was in a wall-less zoo, where the   lions   and the antelope are in one big field with me. Apart from being mildly apprehensive about the proximity of such carnivores nothing happened.   Fanciful things dreams are. '),(5095,3206777,'02,August,2004','2004-08-02 00:00:00','   Another new column   Haha, well I have found a category for myself. The Listener has a new column too called \"Words Worth\" which introduces readers to \"new\" words. Some of them seem a joke but maybe people use them somewhere..  Anyway, one of the words today was \"Generation C\", which means:    \"Obsessive content generators. Often to be found snapping photos - of themselves - on their mobile phones for instant distribution to friends and family. Or constantly updating personal blogs with boring personal trivia.\"    Yeah that\'s me. But he/she makes it seem so narcissistic... Hmmmm maybe I am..... '),(5096,3206777,'02,August,2004','2004-08-02 00:00:00','   Today\'s Thought   Today\'s Thought is a little column in the New Zealand Herald daily. It is usually a little Biblical excerpt but with the facelift of the Herald there was an interesting change. The Weekend Herald now carries one Bible excerpt and one from a different religion. This is the one from last weekend:   Blessed is the man who fears the Lord, who   delights greately in his commandments.    - Psalm 112:1    I have been insulted! I have been hurt! I    have been beaten! I have been robbed!   Anger ceases in those who do not harbour   this sort of thought.   (Buddhism, Dhammapada)   Interesting! '),(5097,3821963,'29,July,2004','2004-07-29 00:00:00','   HmM..feelinG sOOOooo...dUno hOw to explaiN..i nOe it\'s nOne of my busIness..but dUno y feeL sOoooo...haha...aiyA..mUz be wonderiNg waT i tOkkinG rite..haha..slowLy gUess...'),(5098,3821963,'26,July,2004','2004-07-26 00:00:00','   heya..i getting lazy..seldom write blog nowadays..Hmm..i admit i m a lazy piggy..haiz..getting bit paranoid nowadays..alwis think someone is tokking bad abt me..or is it true..sobx..anywae..i can feel michelle tt grp not vry happy wif me..thou for watever reasons i duno..did i step on their toes??haix...ParaNoid...  Oh..example..tt dae i brought a new bag to sch..deN dey all were lyk staring at mi lyk a freako..n so bu shuang..scary lehx..deN i thot i hearD deM saying smth abT copyiNg shih ying..haiz..but i didnt..i buy tt bag becoz i lyk it..watever i do some ppl will sure not happy..i noe i shud juz ignore deM..but some remarks can be real biting lorx..i reali hope everyone can be happy lorx..m i not sensitive enuff......................................  Another thing..y m i trying so hard to try to please everyone n let dem be happy when dey dun even care abt my presence n care abt my feelings..i feel so sadddd...saded..  buddeN..the bible says we must be forgiving and nv expect smthing in return when u help someone or etc..but smtimes it juz so difficult..afterall..we are humans..fine..lets juz sae i m childish or immature..  i also feeling depressed nowadays actually..so slack..juz cant bring moi to study for Os..i juz cant get myself into study mood...my momentum kena broken larx..but force myself...but vry difficult leh..haix..see bks liao den feel so lazy to pick dem up n concentrate revising for few hrs..haiz..but in the past i can..hu can help mi bring back the momentum??argh!!! reali wan do well for Os de leh..if i cannot take sci i think i will reali break down..but dun worry..will not commit suicide..bleahx..but life\'s juz lyk tt..my usual quote..instead of blaming y life is lyk tt..y not make an effort to improve it to the way u want it to be..??whoa..sound vry chim horx..hee..k larx..i wan go do maths le..zOOkx..   '),(5099,3821963,'24,July,2004','2004-07-24 00:00:00','   tOdae weather sOOOoo cOld..earLy in the MorniNg raiN..deN i gOt mOCK exAm..No sweaTer tO keep mOi waRm..cOld..buT thiS tYpe weatHer beTa deN Hot aftErnOOn..unbearaBle aNd sticKy..toDae sEE pikaChu sO siaNx..is it trUe all woMen hu reaCh a certaIn agE will naG naG anD naG(menopause??)caNt stand hEr lorx..her VoicE wiLL bE 4eva ringinG iN my ears..yiKs..actUally caN toleratE de Lorx..buT smTimEs hOrx..sHe caN reaLi juZ gO on naggIng lOrx..eveN if we make tHose nastY remarKs..heeX..hOrrible ritEzx..oh ya..hOw cOme sHe nv evEn get foOd poisOninG when she tastE thE pastrY aR..??pEngz..we tasTe a bit dUno lyk waT le..XinyI eVen vomit \"PINK\" substanCe..haha..yeT she seeMs oKie..sHe tasTe so many ppl\'s pastRy..lOlx..   '),(5100,3821963,'19,July,2004','2004-07-19 00:00:00','   Hi..i\'m baCk..it\'s awfullY lOng sInce i lasT enTered a nEw pOst hOrx..bleHx..shuCks..todae iN tHe aFternOOn feeL sO sicK unTil almOst vomit..bUt nOw feeliNg betA..haHa..btW herE iS a sOng dedicateD tO evEryOne..  &nbsp;   The cAllinG -&nbsp;oUr liVes     &nbsp;   Is it love tonight    When everyone\'s dreaming    Of a better life    In this world    Divided by fear    We\'ve gotta believe that    There\'s a reason we\'re here    Yeah, there\'s a reason we\'re here     &nbsp;   Cause these are the days worth living    These are the years we\'re given    And these are the moments    These are the times    Let\'s make the best out of our lives     &nbsp;   See the truth all around   Our faith can be broken    And our hands can be bound    But open our hearts and fill up the emptiness    With nothing to stop us   Is it not worth the risk?    Yeah, is it not worth the risk?     &nbsp;   Cause these are the days worth living    These are the years we\'re given    And these are the moments    These are the times    Let\'s make the best out of our lives    Even if hope was shattered    I know it wouldn\'t matter    Cause these are the moments    These are the times    Let\'s make the best out of our lives     &nbsp;   We can\'t go on    Thinking it\'s wrong    To speak our minds    I\'ve gotta let out what\'s inside     &nbsp;   Is it love tonight    When everyone\'s dreaming    Can we get it right?    Yeah, well can we get it right?     &nbsp;   Cause these are the days worth living    These are the years we\'re given    And these are the moments    These are the times    Let\'s make the best out of our lives    Even if hope was shattered    I know it wouldn\'t matter    Because these are the moments    These are the times    Let\'s make the best out of our lives   '),(5101,3821963,'16,July,2004','2004-07-16 00:00:00','   waH..i m deaD beat..toDae help tHe geRs wiF f&amp;N practicaL..hMm..weird hUh..duNo y nowaDays feel tirEd sO easily..diE..izit becoZ nv ExercisE foR lOng tiMe le..i m lyK..a batteRy..neeDed tO be recharGe tt tyPe..hAha..tmL goT gEog moCk eXaM..muZ studY woRx..cAn anyoNe teLL me whetHer dO i nEEd tO studY develOpmenT..cOz shuD be dUn neeD ritE..aiYa..tMl go scH asK..if need deN dO sOme last minutE browsinG loRx..eH eH..tMl goinG out wiF yOke liNg aNd micheLLe i thiNk to buy LipiNg\'s bdaE preseNt..shuD be buYing a soFt toY..coZ hearD frOm yOkes tt shE wanTs a biGgy sofT tOy..hEe..my bdaE alsO cOminG sOOn lEhx..sO fRens hu saW tiS bLog..rMb me bdaE k!!!iT\'s On 18th Aug..niCe nUmbeR hUh..i borN in 1988 sOmemoRe leHx..haHA..craPPy..'),(5102,3821963,'13,July,2004','2004-07-13 00:00:00','   pEngz..goodness knOws wat happen to me..aft scH feel sO exhausted and weak..had a bit of difficulties breathing..den reach hOme suddeNly feel so feverisH..siaNz..nOw got feveR..feel so giddY..muscle ache all oveR the bodY..duNo tml wan gO scH nT..maybe not..but tml got the project B planning thingy..lata i duno how to do how..reali ma fan and terribleto get sick loRx..haiz..dun wan type anymore..wan go rest again..getting reali giddy..seeing the surroundings go round n round..haha..tt\'s juz bin to lie ya..not tt kua zhang larx..byez..   '),(5103,3821963,'12,July,2004','2004-07-12 00:00:00','   haha..the blog title anyhow write de..nth to do wif the content i m writing..juz type it in when i was listening to \'i believe\' by tata young..hee..one of the verse..hmm.let mi recall wat happEn toDae..eh..o ya..pengz..my class actually spend 1 whole hr deciding the design for class t-shirt..i listen until sianz..so difficult to fit into everyone\'s liking..but wat to do..poor xinyi..muz be very stress..xin ku le..actually i m fine wif the design as long it is appealing to look at larx..N not tOO expEnsiVe..otherwise gonnA be brOke agaiN..hahA..another thinG..i juz found out todae jiehui quite gd at writing short stories worx..althou some of the story quite AHEM..groSS!!especially tt rat thingy..haha..but the weird thing is..jiehui wrote tt herself n yet she wanted to vomit when xinwei told her a more gross way of trting the rat..loLx..anywae..gtg do hw..erm..i mean revise..todae no hw..bleahX..   '),(5104,3821963,'11,July,2004','2004-07-11 00:00:00','   Hmm..actually toDae waNna gO out de..bUt nO onE pei wO..stay at HoMe will die de..especiaLLy haV nTh tO do..oH ya..tOOpid larX..my parents weNt gambLinG agaIn..eveN my grandma alsO gO..pengz..wat\'s the worLd cOminG to..my grandma instead oF teLLinG deM not To gamBle actuaLLy WENT wif deM!!!argH!!but suaN le..i alSo caNNot saY anyThinG..i think dey went on leisure worLD ba..haiZ..i reali canT imagIne iF the sIngapoRe governMent decide to set up CasiNos in Singapore..pengz..den it\'s rEali a goner liaO..dun wanNa thinK abt it..since i noT goiNg ouT..beta finisH my Hw deN..bubBYez..zOOkz..   '),(5105,3821963,'07,July,2004','2004-07-07 00:00:00','   shiT shiT shiT!!!!i haTe iT..wat The heLL..sHe is sittiNg beside Me aNd i reaLi feeL lyk retortiNg heR..y caNt i geT a raiSe iN allowaNce..it\"s nOt beCoz i waN To paY oFF my deBts..okIez..maYbe partiallY i dO..buT overaLL..aiya..fuCk oFf laRx..i in a daMn blooDy baD mOOd nOw anD i reali dUno whaT to saY..i juZ caNt stand iT wheN deY theMselveS bOrroW mOney frOm otHers anD yet teLl mi Not to..dUn deY nOe whaT is Yi sheN zHuO zhE..i reaLi hatE tOkkiNg tO deM nOw..juZ wanNa shUt myseLf iN my owN rOom..arGh!!i wanT tO crY..i caNt stand iT!!!i reali caNt staNd iT!!!i waNt to scReam!!!nOw for no reaSon..deY inVite a few monKs to oUr hse..oh goOdneSS saKe..iF deY aRe siNcerE deY wUn eveN gaMble kNowing it\'s is one of thE principLe of budhism..or waTeveR it is speLLed..shuT up..i caNt eveN staNd heariNg their Voice nOw..reaLi feel Lyk teLLinG deM to SHUT UP!!!arGH!!!!!!!i goinG bersek..i dUno y..i reali so sad anD depressEd noW..tearS arE gOnNa flOw lyK a rivEr anY MomeNt..i m tryIng to control deM..buT i canT..i beta gO and hide mYselF iN a corNer of a Hse..   '),(5106,3821963,'06,July,2004','2004-07-06 00:00:00','   HaiYo..todae nTh mucH except tt i so bRoke..reTurn meL tHe mOney den livIng iN pOvertY..haHa..bUt afT scH verY faRni..coZ todaE afTerNooN raIn ma..den jIe hui deY all suGGesTed walkIng to whitesaNds..actually feeL lyk taKing bus..buT siNce it\'s lyk onli drizzliNg i agrEed loRx..but daMn suaY..walk half waY raIn get biGGer..i sharinG umbrella wiF xin yi den her umbrella sO lite until the wiNd hOrx..blow the umbrella uP..eH..noT exactly explosiVe tt type laRx..haha..hOw to descriBe leHx..deN xIn Yi sO amuseD by it tt shE doubled ovEr loRx..dUn understaNd alSo wat\'s sO farNi..kena drencheD lyk duNo wat..in the end..we stoP oveR at tHe pavilion in The park but still drencheD lyk siaO..afteR reaching whitesands we weNt to the library tO return sOme boOks beFore going oLd chanG kEE to buy currY puffs..i gOt addicteD tO deM sia..bUt it\'s fattEning..heeZ..   '),(5107,3821963,'05,July,2004','2004-07-05 00:00:00','   Cant stand my parEnts smtimEs..willinG to take Out mOney gamBle but wheN it cOmes tO gettiNg alloWanceS frOm thEM..alwis saE nO moNey..oR so stinGy..will it diE to GiB me extrA moNey..deY arE sO calculaTinG..gibIng mE onLi 4 bucks a daY..whEn i gO out alsO nV asK deM foR moneY anD yet saE i speND moNey lyK wateR..plEasE!!if i ever waN to Go ouT wiF my freNs foR a moVie i wouLd haVe to savE foR at leasT 2 wks..provideD i go on \"DIET\" foR thE twO weeKs..buT whEn i dUn eat..deY will sCold mE..saE lata caNNoT groW tallEr or watevA..nOw i bought a nEw dreSs anD i borroWed tHe moNey frOm my frEn..i thInk i juZ havE tO starVe fOr 3 weeKs..beFore i Can retuRn heR all tHe moNey..haiZ..y liFe is lyk tt For me..all mY frEns almoSt get twiCe my allowAnce pEr daY & caN eveN get xtrA moNey whEn deY gO out wiF theiR frEns..whEn i gEt juZ 10 bucKs frOm my mOm to Go out..she giB me the moneY reluctantlY onLi aft i beG heR anD eveN call me aN unGrateful braT..saE wO mei liaNg xIn..wth..if my parEnts arE williNg to foRk ouT 500 or 1000 bucKs to gO gamBle and PreparE to losE thE moNey..i juZ sEE no reaSon whY shE is so stingY wiF the moNey wheN it cOmeS to buyinG smTh of valuE and be ablE to use it..my parEnts\' retoRt iS..gambliNg haS wInninG aNd losIng..shucks larX..whaT caN i dO..i m stiLL a studEnt..stiLL havE to DepenD oN deM to pay my scH fees etc ma..franKly speakIng..Our family wuN be iN dirE straiTs if deY did nOt becUm addictEd to gambling..we usEd tO be quitE wealthy oKie!!nOw my fatHer havE to resoRt to sEllinG his MercedeS beNz juZ to pay ofF  his gambling debts..eveRy nOw aND theN..my parEnts also quarrel ovEr thE issUe of moneY..it\'s reali Full oF shiT!!!!!dUno waT theSe adUlts arE thInkinG..theiR frEns arE alsO nOne the BetteR Hu encouragEd deM to go gamblIng wiF deM..idiOtic FellOws..iF i cAn i reaLi feeL lyK basHing dem uP..i dUn understaNd..parEnts teLL yOu not to mix wiF baD company anD yeT do it demselveS..hoW to expect their ChildreN to respect deM..tell me not to learn gambliNg but Go gambling demselvEs..fatHer tell mE not to smOke but smOkes himSelf..wat\'s the woRld cuMing tO maN..smTimeS i reaLi wiSh to retort deM..my fatHer saE i m stubborN..but i m at least in a gD way..unlIke deM..teLL deM not to gamBle oredi and stop smokiNg buT deY alSo nV lisTen whaT..if i\'m stuBborn i guess it\'s probably inherited From dEm..smTimeS see deM alsO keP sIm..haiZ..   '),(5108,3821963,'04,July,2004','2004-07-04 00:00:00','   strEssed strEssed streSSed!!!i hAte iT!!!toDae my papA asK mE bOut goinG jC agaiN..keeP tellinG mE tO go jC sO i caN get iNto a Uni moRe easilY..bUt deY nv tHot bout waT i reaLi waN to studY..deY waN me to tAke businEss sO i caN take Over his businesS one daY oR woRk in some bIg coMpanieS to Earn mOre mOney..sO overaLL it\"s still aLL aboUt the SalarY..The \"MONEY\" issUe..soBx..aND whEn i toLd dem whaT are my Plans deY sae tt\"s meaNs iT\'s all My decisIon anD dUn need tO evEn tell deM..saE whaT i nv sparE a thot for deM..pleasE..iF i didNt spaRe a thOt for deM i woulDnt eveN tell dEm anytHing..i reaLi duNo whaT to dO..to listen tO my heaRt or to Dem..if i listeN tO deM..i wiLL jUz bE anotheR geR whO will regreT heR lifE to turN ouT lyk tiS iN futuRe..howevEr..if i listEn to mY heaRt..i will alsO be responsibLe foR the Path tt i haVe chOsen..is it reali whaT i waN??i will haVE no quaLms abOut goiNg to a jC if i can geT intO the scieNce couRse but thE problem iS..i CANT!!!it\"s difficuLt fOr a studeNt without studyinG A maths and pure science tO get inTo the Science courSe in jC!!!let\"s juz agREe liFe\'s lyk tt aNd it sucKs..thE decisIon i made noW will surely affect thE resT of my liFe anD i seriouslY need tO consideR it carefuLLy..coZ i duN waN to regreT it in The futurE wheN i diE..althOu my MummY saE shE will let mE choosE what i wAn to Study anD tell mE nOt to worrY abt the \"mONEY\" issuE..i cAn still seNse tt going jC is betA coz it\'s Cheaper..sHe sae tt if i gO poLy it\'s up tO me..but make surE i geT intO the locaL university..buT if i caNt..i juZ haVe tO go Out aNd worK..if deY caNt afford to send Me to ovErseas uNi..well..at Most aft my diploma i go worK fOr feW yeaRs b4 pursuinG my degree overSeas..sHe scoldEd mE anD sAe oversEas is not neceSsarily beta..sO juZ what the heLL dO dey wAn..althOu ppl miGht thiNk i muZ be maD to thinK about university whEn i haven even completed my \"O\" levelS..buT tiMe is rUnning ouT..sOon i will graduate aNd it wiLL be tiMe fOr mE to makE a decisiOn..i wanT tO havE a goaL iN liFe..i duN waN to liVe an aimLess liFe..   '),(5109,3821963,'02,July,2004','2004-07-02 00:00:00','   PeNgz..firSt tiMe writE blOg a bit Weird aNd fUnny..hAhaz..toDae dUno whAt haPPen to me..cAnnOt thInk properLy..miNd blaNk blAnk..faN yiN cHi dUn agAin..hAiz..daMn siAnz todaE..nOthiNg muCh haPPen iN sCH except dArn loTsa weeKenD hw..tML chInese LC sia..scaRy sIa..scarE forgeT brinG IC anD entrY prOof deN i gOner liaO..tML waNna gO tO tOwn walK walK leHx..bUt sO brOke..nVm..aiYa..dUno whAt elSe tO wriTe alsO..betA gO do my ProjeCt..zOOkx..   '),(5110,3821963,'12,August,2004','2004-08-12 00:00:00','   Hmm..cancel itEm NumbeR 8 in bdae lisT cauSe i oredi got a new pencilcaSe..hEex...pinkie waN..  YeaH!!!todae got back \'O\' lvl results..sHOoo happi k!!!sO relieved...tt i dUn need taKe chi fOr the resT of thE yeaR!!i secuRe my firsT A1 out of the 7 subjecTs..& i think it\'s came as a gd bDae prezzie..!!! but sOmehOw or ratHer..it juz seems lyk few daes agO we werE sec 1s lorx..time pass reali faSt..within a few moNths i will be graduatiNg..can u believe it??sOOn i will be either going poly or jC den followed by university..   '),(5111,3821963,'08,August,2004','2004-08-08 00:00:00','   BDae wisH LiSt     gEt reaLi GoOd results For mY \'O\' levels..at leaSt 5 distinctiOns aNd let thOse hU thiNk i aM a bimBo OpeN eyEs biG biG..waHaha...  gEt iNto s\'pOre pOlytechniC biOmedical oR bioteCh coUrse..  gEt a samSunG E700A pHone..  sLim doWn my biG biG thigHs..  haB a \"NO PIMPLES\" fAce..which meaNs flawless...heex...  wAnts tt pinkie T-sHirt(quiTe biG sizE oNe) frOm CenturY squaRe tt \"yellow\" shOp..  wAnts a deniM skiRt..  Hope tO haB a new pencilcaSe..  HaB a nEw sweatEr thaT i lYk..  waN to buY loTsa wittY eNglisH storybOoks tO reaD...   weLL..tt\'s foR Now...i guEss...    To Be conTinued...   '),(5112,3821963,'03,August,2004','2004-08-03 00:00:00','   hMm..wah..tis few daes goona go broKe..haha..not reali larx actually..tis week muz pay for a lot of things..muz pay for class t-shirt, graduation night fee, maths paper, bbq..pengz..a lot sia..hmm..but tis week got long holidae...but i guess i wun be going out..instead i will be mugging at home i think..it\'s like \'O\' level oral is juz 2 weeks away?oMg..i cant believe i m still slacking around..i mean get alive!!!it\'s oredi Aug n i haven reali start my revision..wat m i doing!!!tis week muz reali start mugging..by hook or crook..locked myself in my own room with my bks...hmm..sounds lyk self torture horx..haha...i think muz start preparing for the eng oral topics..otherwise when tt dae i not prepared surely stutter or cant speak a word..oh ya..\'O\' level chi results most probably out aft National Day..  *tensed uP  beta dUn be a B and below..i will go KILL myself!!haix..but dunO y hab tis weird feeling didnt do very well..sianz..y m i telling myself to mug but not taking any actions???tis wun do!!!..muz stick to my plaN...haiya..if i m suppose to be sticking to my plan..y m i still here writing my blog??!!??lolx..k larx..i m off to stardee planet..zoOkx..   '),(5113,3821963,'02,August,2004','2004-08-02 00:00:00','   Lydia è¿·ç¦»çš„çœ¼çœ¶   ä¸ºä½•æµæµª å¿ƒç¢Žçš„æµ·æ´‹    å—äº†ä¼¤ è¿žå¾®ç¬‘éƒ½å½·å¾¨    Gypsyå¥³éƒŽ ä¸ºè°è€Œå”±    ä½ ä¼šçœ‹è§é›¾ çœ‹è§äº‘ çœ‹è§å¤ªé˜³   çºªå¿µå’Œå¤§åœ°é‡å¤ç€æ‚²ä¼¤     ä»–èµ°äº†å¸¦ä¸èµ°ä½ çš„å¤©å ‚   é£Žå¹²åŽä¼šç•™ä¸‹å½©è™¹æ³ªå…‰   ä»–èµ°äº†ä½ å¯ä»¥æŠŠæ¢¦ç•™ä¸‹   æ€»ä¼šæœ‰ä¸ªåœ°æ–¹ ç­‰å¾…çˆ±é£žç¿”    Lydia å¹¸ç¦ä¸åœ¨è¿œæ–¹   å¼€ä¸€æ‰‡çª— è®¸ä¸‹æ„¿æœ›    ä½ ä¼šæ„Ÿå—çˆ± æ„Ÿå—æ¨ æ„Ÿå—åŽŸè°…   ç”Ÿå‘½æ€»ä¸ä¼šåªå……æ»¡æ‚²ä¼¤    Feel sO irritaTed by sOmeonE toDae..dUno y..tt perSon diD nTh tO offeNd mE but juZ feeL sO uNeasY anD irritaTed wif tt perSon..jUz iN caSe You aRe wonDerinG..tt persOn nOt frOm my clAss jiU dui leRx..haix..weird mE sia..nowaDays feeL sO unliKe myselF..i m sufferinG frOm temporary depressiOn..   '),(5114,3735544,'24,June,2004','2004-06-24 00:00:00','    T0dae woke up so late.. at 11am.. yesterdae watch TV n read my story book until so late.. to woke up n hav hav my lunch .. the f00d was so damn hot sia..(BEEhon)... then took alot of water.. likE water taNk..ritE..!! 3 m0re dae$ to school liao.. my mum haveN seeN my report b00k.. haHax.. later surE kaNa scolding.. so fast school reopen..aiyo.. sch00l reopen dat dae hor.. sure n01 going t0 bring anY bookS... onli all thE go0d  0neS.. theN all the timEtable also change liao.. siaN.. g0na play gamE liao.. byEE... ~{tIRED}~ !![tiRED]!! '),(5115,3735544,'23,June,2004','2004-06-23 00:00:00','    T0dae is so tiring.. muZ go take care my cousiN.. Wa they are noise maker lucky neva stay there.. came back frm my cousin houser not long.. been there 4 hours.. n back home reading Harry Potter bookS... i already long time neva received any calls frm my est frenz.. mayb they are busy rite.. but it so long.. i hav not been toking to anyone 4 so long.. i onli hav 1 true frenz dat can hear mi out .. it hav been to tough 4 mi.. Do bE$t frenZ b tru$ted..?? i think NO.. i hav been betrayed alot of time .. but i still  reali cared 4 my frenZ.. wa should i do..?? Alot of ?? in my headS.. What i tell them they went around telling ppl.. n i tell them not.. they still go n tell ppl.. wad should i do..!!! i reali treasure this ~frenZshiP~..!! '),(5116,3735544,'09,July,2004','2004-07-09 00:00:00','    Wad happEn to me.. Sick 4 week$ n i lost my voicE 4 days.. My frenz they call me muTe.. that doesn\'t concern me the most.. the thing that make me tik veri hard.. should i patch back wit my EX-.. so any opion.. i dunno wad to do.. wen i was sick.. he find my a jacket.. at that time i also neva realise that he still like me.. so wad i muZ do...!! '),(5117,3735544,'02,July,2004','2004-07-02 00:00:00','    Today we celebrate youth day... then i came to school late.. then i hav fallen ill during chinese period.. then i was veri sian...i had fever.. my whole body can cook egg... then we went to the hall 4 concert.. it was fun..n the music they played is veri nice... then the claz that is beside us.. is veri the wad .. the bully their own clazmates.. n they kept pushing their friend .. he nearly sat on my friends lap.. then my friend wacked the boy who pushed his friend... then i went home staight after school as i m sick.. my friends they all went to Kopitam n eat their lunch.. as i was walking down to my house.. my heads hurt alot.. hav to rest liao.. byee '),(5118,3735544,'01,July,2004','2004-07-01 00:00:00','    Todae after sch00l i was walking back to my sch00l canteen.. bcoZ i had to wash the use can 4 my claz project.. then as i was walking back i saw him.. then i saw him.. lucky.. i did..!! i miZ him s0 much... he gone to austrila to study.. n0 wondor i can\'t find him any where in the school..but im so lucky to find him.. n c him.. i was so happy...!!! '),(5119,1007188,'12,July,2003','2003-07-12 00:00:00','   1/03 DrKioni.com Awarded ByRegion.net Healers Best of Websites Award DrKioni Live Internet Radio can now heard on AuctionWitch.com.  2/03 Announcing strategic web partnership with AuctionWitch.com.  2/03 Site Popularity: Metaphysical Link Directory 1383 distinct links, MSN 143 Links,  Site Popularity: AllTheWeb 148 Links, Alexa 76 Links. Google 166 Links, AltaVista 117 Links  Site Popularity: Yahoo 166 Links, Lycos 97 Links, Teoma, 71 Links, Ask Jeeves 80 Links,  Site Popularity: Google/AOL 166 Links, HotBot/Inktomi 142 Links, Excite 30 Links, i won Search 166 Links.  September 2002 Dr. Kioni celebrates 10th Anniversary online!  September Geo Listener Stats: 483 people tune to Internet Radio for a total of 408 hours!  August Geo Listener Stats: 303 people tune to Internet Radio for a total of 114 hours!  September 2002 Life Success Empowerment Internet Radio goes live!  DrKioni.com averages 3528 hits per day! July 2002 Statistics.  August 2002 Website redesigned - added separate products and services page.  August 2002 DrKioni.com offers site visitors discount telecom services through Cognigen; A leading global provider of discount telecom services.  June 2002 DrKioni.com Internet Radio premiers. Powered by Live365.  May 2002 World\'s Largest Metaphysical Link Directory added!  2002 Adds Personal Development giants Nightingale.com and mp3Motivators.com.  Jul 1999 Publishes the online eZine Esoteric Prosperity ISSN: 1528-7874. Awarded Topica\'s Pick for Valuable Content.  Oct 1998 DrKioni.com evolves online as a theme oriented virtual ecommerce website first began as the Exalted Living Institute.  1998 Becomes a member of Commission Junction. Begins marketing affiliate sites online.  1996 Began marketing telecommunications products and services online.  1992 The beginning - Owner and Sysop of Computer Bulletin Board. Service (BBS).  '),(5120,1007188,'07,July,2003','2003-07-07 00:00:00','   WDRKLOVE Positive Vibes - DrKioni.com Web Station  904 Total Listening Hours in the last 30 days.   Countries tuning in to WDRKLOVE Positive Vibes: United States, Portugal, Japan, El Salvador,  Ukraine, United Kingdom, Switzerland, Hong Kong,    Israel, Belgium, Canada, Mexico, Brazil, France, Germany, Singapore, Chile, Australia, Korea Republic of, Thailand, Hungary, New Zealand, Norway, Indonesia, Yugoslavia, Netherlands, Spain,  Denmark,   Austria, Sweden, People\'s Republic China, Bahamas.  '),(5121,1007188,'07,July,2003','2003-07-07 00:00:00','   Some html and java script code modifications, mostly for asthetics and faster loading pages.  Added digital prayer wheel from http://www.dharma-haven.org/.  Added link to http://www.auctionwitch.com/shoppes/Raven/ for DrKioni.com Logo design. Thanks Raven! Added Psychic Rx banner to index page for greater visibility.  Also tweaking keywords for higher search engine placement.  DrKioni.com now places #2 in the Google Directory  under the keywords Spiritual Advice.  '),(5122,1007188,'07,July,2003','2003-07-07 00:00:00','   In Feb. 2003 a query of search engines reported 1568 Links to DrKioni.com from websites on the WWW. The report is as follows:  Site Popularity: AllTheWeb 148 Links, Alexa 76 Links. Google 166 Links, AltaVista 117 Links  Site Popularity: Yahoo 166 Links, Lycos 97 Links, Teoma, 71 Links, Ask Jeeves 80 Links, MSN 143 Links Site Popularity: Google/AOL 166 Links, HotBot/Inktomi 142 Links, Excite 30 Links, i won Search 166 Links  '),(5123,1007188,'07,July,2003','2003-07-07 00:00:00','   DrKioni.com is proud to choose Blogger Pro for site news, updates and press releases. Check back often.  '),(5124,1007188,'18,July,2003','2003-07-18 00:00:00','   From: Dr. Kioni  Re: WDRKLOVE POSITIVE VIBES WEBCAST OUTAGE  Date: 07/18/03  You may have noticed I was having technical difficulties with the Positive Vibes Webcast yesterday. The cable company said the problem was on their side and it was repaired in a short period of time.  Nevertheless, the live Webcast was spotty at best and the outage occurred when a lot of you were listening. I sure hope you didnâ€™t mind listening to the Best of Positive Vibes earlier than scheduled.  On the upside, I am now keeping the live stream online until 11pm EST weekdays.  The evening Webcast of Positive Vibes will feature music only â€“ no success talk, nor public service announcements.   mp3motivator Nightingale-Conant Personal Development http://drkioni.com/products.html  '),(5125,1007188,'16,July,2003','2003-07-16 00:00:00','   The following Affiliate Merchants offer great products and services on DrKioni.com. See individual Merchant\'s sites for content and economical bargains! Your satisfaction guaranteed - read individual website Secure Shopper Guarantees and Return Policies.   Nightingale-Conant Personal Development Henry and June Sexy Lingerie Gift Tree eHarmony mp3motivators Keen Tarot by Saskia AzureGreen Dr. Kioni\'s Online Health Rx - Vstore.com  '),(5126,1007188,'15,July,2003','2003-07-15 00:00:00','   DrKioni.com is now an official eHarmony Affiliate and will offer Free Personality Profiles and memberships.   eHarmony utilizes intelligent matching for compatible singles.  Free Personality Profile ($40 value) offered with registration! Take the eHarmony tour and fall in love for all the right reasons.  2 thumbs up review by Dr. Kioni!  '),(5127,1007188,'14,July,2003','2003-07-14 00:00:00','   Â DrKioni.com Logo by Raven\'s Graphics Web Shoppe - See Ravens Banner under Web Master Resources  urlLink   at DrKioni.com.  Digital Prayer Wheel provided by Dharma- Haven.org urlLink    Thank you for your contribution to DrKioni.com!  SLH, Dr. Kioni  '),(5128,1007188,'17,August,2003','2003-08-17 00:00:00','   Press Release August 17, 2003   Dr. Kioni announces a new plugin Community Forum powered by Every1.net has been added to DrKioni.com. Exciting , fresh content awaits each day as hundreds of people share online interest in Community Forums.   urlLink DrKioni.com News Updates  '),(5129,1007188,'29,August,2003','2003-08-29 00:00:00','   Dr. Kioni joins the Ryze Business Network and also announces affiliate associations with ELOAN, Cheapairlines.com, AstroStore, Catalog City and Chopra Center Essentials as Commission Junction Affiliate Merchants. DrKioni.com continues to add quality internet products and services for your secure online shopping experience.    urlLink DrKioni.com Galleria Mall, catalogs and search engine shopping. Toscano, Bombay. shop malls online  '),(5130,1007188,'29,August,2003','2003-08-29 00:00:00','   ELOAN, Inc. Nasdaq Symbol EELN, a leading online lending company with over $13 billion in consumer loans has approved DrKioni.com as an Affiliate Merchant website. ELOAN offers a variety of financial packages for Home, Auto, and Refinance Loans.   Also joining DrKioni.com is the Chopra Center Essentials featuring the products and services of the reknown Dr. Deepak Chopra.     urlLink DrKioni.com News Updates  '),(5131,1007188,'21,September,2003','2003-09-21 00:00:00','   Java Script errors prohibited the site from loading over the past week and a half for some visitors. Please accept my apologies for the inconvenience. Thank you.  Dr. Kioni    urlLink DrKioni.com Live Advice, Spiritual Advice, Internet Radio, Metaphysical Links & Online Shopping  '),(5132,1007188,'01,October,2003','2003-10-01 00:00:00','   urlLink WDRKLOVE Positive Vibes Internet Radio, New Age, Smooth Jazz, Rhythm & Success Talk http://DrKioni.com    The Positive Vibes webcast appeared to be offline a lot last week overnight due to my broadcast server malfunctioning. Live365 engineers have cleared the  problem and normal programming is resumed.  Thank you for understanding and being loyal listeners.  Dr. Kioni  '),(5133,1007188,'13,November,2003','2003-11-13 00:00:00','   Greeting Card interface back up.   urlLink DrKioni.com News Updates  '),(5134,1007188,'13,November,2003','2003-11-13 00:00:00','   Website has undergone an entire renovation.  Users and Visitors will notice new template interface. Navigation buttons are consistent throughout the site and always located on the left side of the page. Positive Vibes listening button is now  available on all pages. The Google search engine p   urlLink Dr. Kioni Conscious Creation Coach. Advisor, Live Advice  '),(5135,1007188,'26,November,2003','2003-11-26 00:00:00','   urlLink DrKioni.com News Updates  '),(5136,1007188,'22,December,2003','2003-12-22 00:00:00','   DrKioni.com began hosting the TVRadio Network today. The TVRadio Network is a rapidly growing FREE TV/Radio Guide, that offers access to over 1000\'s of ONLINE television and radio stations worldwide.   Services include: all stations organized by both LOCATION and CATEGORY, plus a full station database search option. REAL PLAYER and WINDOWS MP options. (when available).  Other user features include a Plug-ins Center, \"Favorites\" options, \"Send Station to a Friend\", and much more...  http://drkioni.com/media.html  '),(5137,1007188,'10,February,2004','2004-02-10 00:00:00','   urlLink DrKioni.com WDRK-LOVE Internet Radio - TVRadio Network - Smooth Jazz, Cool Jazz, Streaming Audio - Live Audio   The Internet radio station  WDRK-Love  and the popular program  Positive Vibes  is back on the air! Anyone can listen 24/7 by visiting DrKioni.com and clicking on the gold speaker icon located on every page or the Internet Radio tab on the main menus. Keep in mind that  the FREE stream is limited to only to 20 listeners .  However if having a choice of the best stations on AudioRealm appeals to you, then  listen to WDRK-Love\'s high quality digital stream on the AudioRealm Network where the stream is never full!  Only $4.99 per month or $49.90 per year (get two months FREE!)  urlLink FREE 7-day trial (No credit card required)Streaming Audio - Live Audio  '),(5138,1007188,'24,June,2004','2004-06-24 00:00:00','   Luckymojo.com and DrKioni.com are proud to present the Luckymojo Hoodoo Rootwork Hour. Every Wed evening at 10PM Eastern Time, participants in the Hoodoo Rootwork Course come together on a conference line to discuss hoodoo (African American Herbal Magic - Conjure with Cat Yronwode from Luckymojo.com and Dr. Kioni from DrKioni.com. The Luckymojo Hoodoo Rootwork hour also features old time hoodoo blues and luminaries from the hoodoo and metaphysical community. The show can be heard live on WDRK-Love via DrKioni.com.  '),(5139,1007188,'01,August,2004','2004-08-01 00:00:00','   Added: Hoodoo Rootwork Forum Added: Hoodoo Rootwork Photo Album Added: Personal Weblog Added: Hoodoo Rootwork Annoucements and Events Calendar Added: I Am Success Email Links to above pages  '),(5140,3936941,'25,July,2004','2004-07-25 00:00:00','   Finally I got a look at  urlLink Whidbey Beta 1  and SQL Server 2005 Express Edition Beta 1 over the weekend.   So far I am really impressed with the new products. Even the Express Editions which are available for free while in beta have the most improved features built-in.  Cool things:    new keybindings (they definitely make more sense than the VS.NET 2002/2003 ones. If you don\'t like them, just change to the old ones)  Tracepoints (right-click -> Breakpoints -> insert Tracepoint): No more Trace.WriteLine when debugging. Extremely cool.  much improved intellisense. No more need for some addIns that ongoingly crash the IDE ...  Visual Studio settings can be exported to and imported from an xml file. No more redundant or registry based hacks when setting up more than one machine.  Many more settings available via the options dialog in the IDE  the MultiLine Property of the WinForms TextBox-Control is automatically set to true when you drag the textbox larger. Cool!    What sucks:    still no intellisense for sql.  Inconsistent Project defaults. SqlServerProjects put AssemblyInfo.cs in the root, not within Properties Directory (like winapps). I think this will be fixed.    GewÃ¶hnungsbedÃ¼rftig    Partial classes. Cool, but my first reaction was: Where is the damn code? (after adding controls to an asp.net page). Designer generated code goes to a hidden file. MVC!    I like it!  '),(5141,3936941,'21,July,2004','2004-07-21 00:00:00','   I relocated to Hamburg (largest city in the north of germany) at the beginning of july and on monday the world\'s largest cruise ship, the Queen Mary 2 from Cunard visited Hamburg for 24hrs. When the ship arrived, several thousand people were at the river Elbe at six o\'clock in the morning to watch the ship enter the harbour. During the 24 hour stint almost 400.000 people came to Hamburg to watch the ship. The coolest thing was that it was extremely good wheather (sunny and hot). But after the ship had left, it started to rain almost instantly :-(  '),(5142,3936941,'15,July,2004','2004-07-15 00:00:00','   Great BizTalk Blogger Martijn Hoogendoorn has published a  urlLink pipeline component wizard  and there is also the (really helpful)  urlLink Adapter wizard  from Scott Woodgates blog. Makes development for BizTalk 2k4 even more delightful.  '),(5143,3936941,'15,July,2004','2004-07-15 00:00:00','   Microsoft has published a PDF version of  urlLink Integration Patterns  from their Patterns and Practices Group.  Just gallopped through it but seems worth spending a weekend. Grab it!  '),(5144,3936941,'15,July,2004','2004-07-15 00:00:00','   I participated in the Microsoft EMEA Security Contest 2004 and made it all the way to the finals in Amsterdam. Eight people (out of several thousand who participated in the Microsoft Security initiative 2004 either by ordering the Microsoft Security DVD or participating in special user group tour events or Microsoft Tech Talks) were invited by Microsoft EMEA to an all expenses paid trip to Amsterdam, including tickets for Tech Ed EMEA 2004, for the Finals. One candidate from South Africa missed the event (visa problems I think) and the seven other candidates came from Ireland (1), Great Britain (2), Germany (2), Austria (1) and Portugal (1) participated in the final round.  The Final, a knock-out quizshow, was filmed in a small TV Studio in Amsterdam on June 28th (the day before the Tech ed began). I made it to the last round but fell against Eamon Oâ€™Tuathail (with a score of 4 points vs. 5 points). Unlucky, but it was really a great experience. You can watch the winner\'s bio at  urlLink here  and be sure not to miss the webcast (live stream with  urlLink 56k ,  urlLink 100k  or  urlLink 300k  WMV format). It is definitely fun watching.  The whole final was done in english and we did not see the questions as in the video.  '),(5145,3936941,'15,July,2004','2004-07-15 00:00:00','   A few days ago I watched the Blue Man Group live at Berlin, Germany, where they perform \"The Complex\". I knew only little about the group (almost everything from the intel pentium commercials) and hence I did not know what to expect from the show. But it was extremely cool. The show is a ingenious combination of rock music, light show, very funny and sudden actions which also includes a wobble pudding. The show is definitely worth every penny.  '),(5146,3936941,'15,July,2004','2004-07-15 00:00:00','   Just started working at a new company. Main focus lies on developing EAI solutions with   urlLink BizTalk server 2004  . Nerver used the software before, but it really seems to be pretty easy developing software things like pipeline components or adapters. Microsoft really did a nice job integrating all the tools into VS.net.  The only things that really seems to suck is  Enterprise single sign-on , which seems to offer only command-line tools for configuration. Nice work, M$. The only thing which cannot be administered using graphical tools or wizards is a security feature.  '),(5147,4067087,'31,July,2004','2004-07-31 00:00:00','   In a state of too many thoughts broiling through my brain.. chatted nonstop to my aunt and a good friend while waiting for the flight.. Reflecting on the kind of person I am.... What direction I want to go... am I nuts?... Chatting nonstop about the week... Missing my Hubby Missing my Kitties.. Flew home with a High School basketball crew... Interesting..  Got home, shoes off, bags dropped.. Wine poured.. (rubiat-Cakebread cellers yum) Couch, cats, laptop... I\'m home.. Ahh  '),(5148,4067087,'28,July,2004','2004-07-28 00:00:00','   What a day!       &nbsp;   Our day started off with  importing our videos.. an exercise in mechanical frustration.. Trying to get the video downloaded and distributed took up a chunk of our time.&nbsp; Not only finding the \'right\' firewire, but, getting the \'correct\' spot on the video camera.. so working with unfamiliar hardware.. arghhhhhhh... We sort of kept going back discussing it during the day to figure out how best to manage splicing together the video, so it\'s something that is still bothering the team. We\'re not frustrated with each othibut rather.. but how to best approach the project & overcoming technical difficulties and a sense of losing time.     &nbsp;  The Lego project-   This one definately fit the serious fun, but also \'challanging\'.&nbsp; We all talked about how to \'create\' a good rover. Charleen had a great idea to go out to the hallway for a quieter space.. So we spread out on the comfy chairs and tried out different ideas & pieces on how to best make a the solar panel cleaner. Everyone had suggestions and we were all picking up pieces and trying out different ideas.&nbsp; Everyone \'saw\' where the other was going.. and modified or contributed to \'well.. how can we make this work.. will it work?&nbsp; Some questions definately need to be answered like.. Just what is the range.. etc.. on the sensor.. so not having all the information in terms of what the \'material could do\' was sort of frustrating in the \'planning\' dept.&nbsp; Tommorrow will be a lot of trial and error and we\'re expecting that.     &nbsp;  Dreamweaver-   Blogger-    &nbsp;  '),(5149,4067087,'18,August,2004','2004-08-18 00:00:00','   ARP\'s.. changes... etc.......  Well.. informal chatting with coworkers.-they seem to like my digit photography/Website building ideas..  I do have a pretty good tech savvy staff I work with.  I swear everyone is ending up with the statement, hey, take a survey tho! Baldridge in Education seems to be establishing roots. (lol) With that thought in mind, for one of my \'adjunct\' duties, I signed up to be one of the data/survey coordinators.. which is taking parent, student & teachers surveys 2-3 times a year and collaborting the data.. It seemed to be a good match for this year.   COOL thing today was asking my 21 first graders-\"Who has a computer\" instead of the 4 to 6 I usually see, about 3/4ths of the students raised their hands.. (YES!) and since I only have 2 out of 4 computers set up so far, two of my youngsters wandered over to the one computer I had turned on and was creating some paperwork for the class... and started clicking away to see what was there...  (good thing I saved) There was no hesitation whatsoever in their movements either.. WAY too comfy..  I knew one of them was tech savvy when I went.. \"ACK, I was working on a paper.\" He got that look of \'GULP\' his eyes went wide and he told me, \"I\'m so sorry teacher.\"  yep.. he\'s worked on a computer before.. I told him I saved, so should be okay.. but.. LET me log you onto the student account..   Hmm.. I definately think my 6 & 7 year old students will be changing the world... afterall.. born.. around 97... The Tech BOOM babies...  One of the parents was asking another teacher about computer use.. (that other teacher has FOUR computers in her room.. you only have one.... and who is the computer for, you or the students?)  Signals of good things...  That kinder/first grade combo class looks like it may be a real possibility, very.. very soon.. spent about an hour talking about logistics and possibibilties today.. look, another gray hair..  '),(5150,4067087,'16,August,2004','2004-08-16 00:00:00','   Now that I have the hobbit\'s attention...  I think I\'ve spent more time away from home than not this summer.  Flew down to Laguna beach to see the Pagent of the Masters for the first time this weekend.. and I\'ve been working 10 hour days setting up the classroom..  Today, was the first day of school.  A good friend was my sub today. I had a very pleasent shock, walking into a quiet classroom.. (hey, I still showed up at lunch after all, no pay, but ya gotta be there.. )  I do have one autistic/ADD student from Mexico who speaks no English... I have a smattering of Spanish.. After having a full blown panic attack last Thursday, My very kind, patient and listener of a principal set up aide time to help out in the classroom. My major concern is simply if something upsetting happening.. I don\'t have the language to help the student out! He does like visuals tho, tv, computer, photo\'s.. (hmm... something in common I see... that could be usful)  When I walked in, they seemed to be doing a pretty good job.   wheee.. I\'m inviting my friend every year I think to go over rules & regs I think-she did such a fantastic job.   After the training Wednesday and realizing that I am such a visual learner.. I was thinking of having new fellow up front take the occasional digital picture with the inexpensive camera the teachers got at the Science Institute early in July. (little point and shoot)  Hopefully, if he has a picture with some \'ownership\' to it.. hopefully, that will encourage him to write about it.  And so that things are fair, perhaps \'sharing\' the camera with classmates.. to spur their writing. (I do have a goodly number of eld students-so if I can push the visual learning/language develop/ownership/fun factor..... may be interesting....  and after all this thought.. I may even end up with a kinder/first grade combo in a week or two....  That curse of \'may your life be interesting...\' till holds true today..  and since it\'s is soo bedtime right now... gnight all and happy blogging..  '),(5151,3320022,'21,June,2004','2004-06-21 00:00:00','   The program subclasses the main form and watches for the  WM_WINDOWPOSCHANGING message. When it sees that message, the program  moves the satellite forms to stay with the main form.  \' Process messages. Public Function NewWindowProc(ByVal hwnd As Long, ByVal msg As Long,  ByVal wParam As Long, lParam As WINDOWPOS) As Long Const WM_NCDESTROY = &H82 Const WM_WINDOWPOSCHANGING = &H46      \' If we\'re being destroyed,     \' restore the original WindowProc.     If msg = WM_NCDESTROY Then         SetWindowLong _             hwnd, GWL_WNDPROC, _             OldWindowProc     Else         \' See if the window is moving.         If msg = WM_WINDOWPOSCHANGING Then             \' The window is moving. Keep the others with it.             Form2.Move Form1.Left + Form1.Width, Form1.Top - 360             Form3.Move Form1.Left - Form3.Width, Form1.Top - 360         End If     End If      \' Continue normal processing. VERY IMPORTANT!     NewWindowProc = CallWindowProc( _         OldWindowProc, hwnd, msg, wParam, _         lParam) End Function  This program removes the title bar from the satellite forms so you  cannot move them directly. You could try to subclass them as well but  moving the group as a whole would be more complicated. For example, you  might need to be able to ignore WM_WINDOWPOSCHANGING messages while you  are moving one of the forms. This example should get you started,  though. ==========  '),(5152,3320022,'18,June,2004','2004-06-18 00:00:00','   The main form\'s KeyPreview is True so its KeyDown and Keyup event  handlers get keyboard events. Those event handlers set a value in an  array to indicate whether a key is up or down.  One odd case that this handles is when the user presses multiple keys.  The array will know if the user is holding down the Left and Up arrows  at the same time.  \'This vector contains the state of all keys \'I think KeyCode goes from 0 to 255, but you \'can change this if you want Dim Keys(255) As Boolean  \'Its used to stop the Do-Loop of the Form_Load Dim StopLoop As Boolean  Private Sub Form_KeyDown(KeyCode As Integer, Shift As Integer)    \'The key KeyCode is pressed now...    Keys(KeyCode) = True End Sub  Private Sub Form_KeyUp(KeyCode As Integer, Shift As Integer)    \'The key KeyCode is NOT pressed now...    Keys(KeyCode) = False End Sub  Private Sub Form_Load() Dim str As String     Me.Show     \'The form must handle the key events    Me.KeyPreview = True     StopLoop = False     Do       \'Just check for the arrow keys to see       \'if they are pressed. To check form more       \'keys, just ask for the correct subindex       \'in the \"Keys\" array       str = \"\"        If Keys(vbKeyUp) Then str = str + \"UP\" & vbCrLf       If Keys(vbKeyDown) Then str = str + \"DOWN\" & vbCrLf       If Keys(vbKeyLeft) Then str = str + \"LEFT\" & vbCrLf       If Keys(vbKeyRight) Then str = str + \"RIGHT\" & vbCrLf        \'Refresh the label       lblKeys = str        \'Important!! (to not freeze the thing)       DoEvents    Loop Until StopLoop End Sub  Private Sub Form_Unload(Cancel As Integer)    \'Important!! To stop the loop, so the    \'form can be unloaded    StopLoop = True End Sub ==========  '),(5153,3320022,'18,June,2004','2004-06-18 00:00:00','   This program uses the ClientToScreen API function to convert the target  coordinates from the screen\'s coordinate system to the screen\'s  coordinates system. It then uses the SetCursorPos API function to move  the mouse there.  Sub MoveMouse(x As Single, y As Single) Dim pt As POINTAPI      pt.x = x     pt.y = y     ClientToScreen hwnd, pt     SetCursorPos pt.x, pt.y End Sub  '),(5154,3320022,'18,June,2004','2004-06-18 00:00:00','   This program simply loops through the pixels in the images. For each  image in top-to-bottom order, the program looks for a color other than  the one defined as transparent. When it finds such a color, it stops  looking at the images and sets the output pixel\'s color using PSet.  Private Sub cmdCombine_Click() Dim max_input As Integer Dim transparent As Long Dim wid As Single Dim hgt As Single Dim X As Single Dim Y As Single Dim i As Integer Dim clr As Long      cmdCombine.Enabled = False     MousePointer = vbHourglass     DoEvents      max_input = picInput.ubound     For i = 0 To max_input         picInput(i).AutoRedraw = True         picInput(i).ScaleMode = vbPixels     Next i      picResult.Width = picInput(0).Width     picResult.Height = picInput(0).Height     picResult.ScaleMode = vbPixels      transparent = picTransparent.BackColor     wid = picInput(0).ScaleWidth     hgt = picInput(0).ScaleHeight     For Y = 0 To hgt         For X = 0 To wid             For i = max_input To 0 Step -1                 clr = picInput(i).Point(X, Y)                 If clr <> transparent Then Exit For             Next i             picResult.PSet (X, Y), clr         Next X         DoEvents     Next Y     picResult.Picture = picResult.Image      cmdCombine.Enabled = True     MousePointer = vbDefault      Beep End Sub  Note that there are faster ways to access color values in V 6 and VB  .NET, and that there are faster methods for merging images if you have  an overlay mask. Note also that VB .NET provides tools for setting a  transparent color for an image so this problem is trivial in VB .NET. ==========  '),(5155,3320022,'22,June,2004','2004-06-22 00:00:00','    Iran seized eight Royal Navy and Royal Marines personnel in three patrol boats on the Shatt al Arab waterway yesterday, claiming that they had strayed across the border with Iraq.  The boats, which were being delivered to the Iraqi riverine patrol service, were flying the White Ensign. They were travelling up the waterway towards Basra.      British troops on patrol on the Shatt al Arab in southern Iraq  Teheran said: \"British boats entered territorial waters of the Islamic Republic of Iran and officials of the naval force, in accordance to its laws, seized the boats and arrested the eight crew members aboard.  \"Interrogation of those detained will continue until the matter is clarified.\"  An Iranian Arabic-language television station which broadcasts to the mainly Shia population of southern Iraq said the Royal Navy personnel had \"confessed that they have made a mistake\".  British sources in Baghdad appeared to confirm that the boats had crossed into Iranian territorial waters in the waterway, which has long been a matter of contention between Iran and Iraq.      There was immediate speculation that the arrests were linked to the row over Iran\'s nuclear programme.  The Iranians were infuriated after Britain helped to draft a highly critical resolution at the International Atomic Energy Agency, the UN\'s nuclear watchdog, condemning Teheran\'s failure to co-operate fully with international inspections.  Although the Iranians insist that the programme is purely for civil purposes, there are fears in the West that it is a cover for developing nuclear weapons.  The Foreign Office said it was in contact with the Iranian government to find out more details and \"try to reach a satisfactory solution\".  It played down suggestions of a tit-for-tat action by the Iranian government in response to last Friday\'s resolution by the IAEA. But privately officials admitted that there might have been an \"opportunistic\" attempt at retribution by hard-line elements associated with the Revolutionary Guards, the troops in charge of protecting Iran\'s Islamic revolution.  Islamic students regularly demonstrate outside the British mission in Teheran over Britain\'s \"occupation\" of holy Shia sites in Iraq, something they could not do without the sanction of the Revolutionary Guards.  Iranian state television played down the seizure of the vessels in its evening news bulletin, giving only a brief mention of the arrests.  The seized craft are believed to be Italian-made patrol boats which were bought by Saddam Hussein\'s regime but never delivered because of UN sanctions.  The Royal Navy is in charge of training Iraq\'s new naval forces and a joint Royal Navy-Royal Marines unit was set up for that purpose.  Behind the scenes, Britain has been playing the good guy to America\'s tough guy to try to raise relations with Teheran from the low point of the 1980s and the late Ayatollah Khomeini\'s regime.  Contacts at a number of levels, which are believed to have included an MI6 back-channel set up in the mid-1990s, have led to five visits to Teheran by Jack Straw, the Foreign Secretary.  British officials said that, although the standard Teheran response to any action it disagreed with was a tit-for-tat retaliation, such moves tended to be much more measured than the Shatt al Arab incident.  \"Of course they do not like the IAEA resolution and they have made that clear,\" an offical said. \"It would be no surprise if there was a frostiness to our diplomatic relations as a result.  \"But it would be unusual for them to link high policy matters to something like this. This smacks more of opportunism.\"   '),(5156,3320022,'21,June,2004','2004-06-21 00:00:00','    Tony Blair predicted yesterday that he would win a battle \"between myth and reality\" over Europe as he gave himself two years to persuade the British people to accept the new EU constitution.  He defied Conservative demands to call the promised referendum as soon as possible, making clear he would not so until after the next general election.      Tony Blair in determined mood during his interview  As a hard-core of Labour MPs prepared to launch their own campaigns against the constitutional treaty today and fresh opinion polls showed a public majority opposed to it, Mr Blair accepted that he faced a \"tough battle\" to win over its critics.  But, if given time, he would succeed. \"This is going to be a fascinating political battle because it will be a battle between reality and myth.  \"If you actually have a debate about the reality then people will say \'well what\'s wrong with that? Why aren\'t we being part of that\'?\"  The constitution, agreed on Friday after an acrimonious summit in Brussels, does not have to be ratified until the end of 2006, meaning that Mr Blair can hold off the referendum until well after the election, expected in early summer next year.  He said on BBC1\'s Breakfast with Frost that there was \"absolutely no need\" to call a national vote in a hurry. The only reason his opponents were trying to force the pace was to limit the time pro-Europeans would have to expose the \"myths\" being peddled about the treaty.  The Tories and some Labour MPs replied saying the British people deserved the chance to have their say as soon as possible.  Last night Frank Field, the Euro-sceptic former Labour minister, said it was \"nonsense\" to say that a referendum could not be held until after the election on the grounds that there was insufficient time to complete the formalities in Parliament. \"It could all be done in a few months,\" he said. Mr Field feared that the whole issue could do serious damage to Labour at the election and said that Mr Blair had become a liability to his own party.  \"The Prime Minister has this naive view that the more he explains the constitution the more people will come round. But the reverse is true. The awful thing is that he has become a handicap.\"  The battle with the Tories will switch to the Commons today when Mr Blair reports on the summit.  Michael Ancram, the shadow foreign secretary, accused Mr Blair yesterday of ignoring the British people, who were against the constitution by a clear majority. Some Labour and Tory MPs believe Mr Blair was delaying the referendum in the hope that one of eight other EU countries expected to hold referendums will vote the constitution down.  Robert Kilroy-Silk, the former television chat show host who won a seat in the European elections for the UK Independence Party, likened Mr Blair to Neville Chamberlain, Britain\'s wartime Prime Minister, and his appeasement of Hitler.  \"There\'s Tony waving a piece of paper saying \'It\'s OK, I\'ve only given a little bit away of our sovereignty\' when in fact this is the beginning of the end of Britain as a nation state governing itself.\"   '),(5157,3320022,'18,June,2004','2004-06-18 00:00:00','    Horror and anger swept the United States yesterday after Islamic militants beheaded Paul Johnson Jr, a 49-year-old grandfather from New Jersey, who had been kidnapped a week ago from his home in Riyadh.  His corpse was later found dumped in the al-Munisiyah district, a rural area of farms and villas 25 miles north-west of the capital, Saudi security sources said.      An image from a video showing Paul Johnson in captivity  The militants - calling themselves \"al-Qa\'eda in the Arabian peninsula\" and the \"al-Fallujah squadron\" - released gruesome video images and photographs of the killing.  Still photographs showed a severed head, placed on the back of a body wearing an orange jumpsuit, the face turned towards the camera. A dagger rested on the corpse, its point apparently buried in Mr Johnson\'s forehead.  They released a statement on the internet, claiming that the killing of Mr Johnson - an aeronautical engineer employed by Lockheed Martin, working with a Saudi manufacturer of high-tech equipment for Apache helicopter gunships - was just vengeance for the suffering of detainees abused in Abu Ghraib prison, outside Baghdad.  It added: \"The infidel got his fair treatment. Let him taste something of what Muslims have long tasted from Apache helicopter fire and missiles.  \"As we promised the mujahedeen, we have beheaded the American hostage Paul Marshall [sic] after the deadline that the mujahedeen gave to the tyrannical Saudi government.\"  The kidnapping of Mr Johnson sparked unprecedented public appeals from moderate Saudi clerics and lawyers, saying that he was an invited guest to the country, protected by Islamic law.  One Saudi colleague of Mr Johnson, using the pseudonym \"Saad the believer\", said he had often invited the American to his home for dinner, and given him books about Islam as presents, and invoked the traditional tribal protection over him, known as ijara.  If Johnson is harmed, the Saudi wrote: \"I will never forgive you, I will curse you in all my prayers.\"  The kidnapping triggered a fierce debate on prominent Islamic websites. One senior Saudi cleric, Sheikh Abu Bassir, re-issued a religious ruling stating that foreigners arriving in Saudi Arabia on valid visas should be protected.  The extremists behind the killing had demanded the release of hundreds of al-Qa\'eda suspects from Saudi jails, setting a 72-hour deadline that passed yesterday afternoon. They also demanded that all Westerners leave the Arabian peninsula.  Mr Johnson was the fifth Westerner to be killed in Saudi Arabia in just over a week, including two other American defence contractors and a BBC cameraman. Late last month, 22 people were killed in a two-day shooting spree at a residential compound for foreigners.  Though Mr Johnson\'s wife, sons and relatives in the US begged President George W Bush, the Saudi authorities and even the kidnappers to save his life, family members told reporters that the impossible nature of the terrorists\' demands left little doubt he would meet a horrible death.   Colin Powell, the United States secretary of state, called the killing \"barbarism\", and predicted that it would reinforce the Saudi resolve to crack down on militants. Last night Saudi police shot dead three suspected militants in Riyadh.   '),(5158,3320022,'22,June,2004','2004-06-22 00:00:00','   LF commander Samir Geagea\'s prison conditions in Yarze have been improved and a new medical checkup proved him in good physical shape, the Beirut media reported on Tuesday. The latest checkup a few days ago was conducted in Sacre CÅ“ur Hospital. Judicial sources pronounced the outcome of the tests satisfactory and said doctors have recommended additional vitamins to his meals, An Nahar said.  The doctors also had no objection to the daily exercises Geagea undertakes in his prison cell or at the prison courtyard. This led to extension of the exercises from one hour to 90 minutes a day, An Nahar said.   The time frame of the weekly visits of his wife Strida was also extended from half an hour to a full hour, and increased form twice to three times a week.  Geagea has been held in solitary confinement at the prison of the Defense Ministry compound in Yarze for 10 years thus far. This made him the longest-held political prisoner since the end of Lebanon\'s 15-year civil war in 1990.  An ongoing campaign by LF activists to win back his freedom is believed to be hinging on whether President Lahoud\'s term would be extended or a new president would be elected in Lebanon by next November.  '),(5159,3320022,'21,June,2004','2004-06-21 00:00:00','   Israeli warplanes have attacked south Lebanon after shrapnel from Hizbullah anti-aircraft fire rained down on military positions straddling Jewish settlements on Israel\'s northernmost flank, re-igniting tensions on the last active Arab-Israeli warfront. Lebanese police and Israeli army spokespersons said no casualties were reported on either side of the frontier from the sudden flare-up of hostilities that came hard on the heels of U.S. orders to Lebanon-based Americans to stay off Hizbullah strongholds in south Beirut and south Lebanon.  The warning notice that was posted by the State Department on Saturday had aroused fears in Beirut that the Bush administration maybe anticipating a major Israeli offensive against Hizbullah military bases in south Lebanon and command headquarters in south Beirut.  The air raid on suspected Hizbullah anti-aircraft batteries between the Lebanese border towns of Haddatha and Tiri was staged late Sunday night, about three hours after Israeli fighter jets patrolling Lebanese skies came under anti-aircraft barrages form south Lebanon.   The Israeli army said shrapnel hit an army position close to the Shlomi border settlement, warranting the swift retaliation by the air force although there were no casualties from the falling shrapnel. Israel had rarely struck in the past unless there were injuries among Jewish settlers.  Hizbullah said enemy jets fired four air-to-surface missiles in two runs on the strip between Haddatha and Tiri, causing bush fires and shattering window glass in the two townships.  '),(5160,3320022,'22,June,2004','2004-06-22 00:00:00','   BroadSoft announced that Ericsson and BroadSoft have signed a global OEM agreement to integrate components of BroadSoft\'s BroadWorks(R) technology into Ericsson\'s IMS based Engine Multimedia 2.0 (EMM 2.0) solution. This integration will enable service providers for the first time to use the same platform to offer a uniform package of advanced managed communications services across wireline and wireless networks based on the Third Generation Partnership Project (3GPP) architecture.  Ericsson will market the EMM 2.0 platform with integrated BroadWorks technology as part of its Softswitch offering, which is based on the 3GPP architecture. With this solution, Ericsson\'s customers will be able to deliver multimedia services, including hosted PBX, IP Centrex, residential broadband and collaborative conferencing services, transparently across different networks. The integration of the BroadWorks media and application servers into EMM 2.0 is based on the Session Initiation Protocol (SIP.) It follows the architecture developed for 3GPP that will be used in next generation fixed and mobile networks to provide multimedia services regardless of location.  \"Today\'s announcement marks a major step forward in the evolution of uniform multimedia service delivery to fixed and mobile 3GPP networks,\" said Deborah Workman, vice president of market research at Dittberner Associates, Inc. Ericsson is uniquely qualified to provide this converged offering given their leadership position in wireline and wireless networks.\"  \"By working with Ericsson, we\'re helping operators to keep their core technologies and investments in place, while enhancing their ability to offer hosted multi-media services across multiple networks based on the 3GPP architecture,\" said Michael Tessler, CEO and president at BroadSoft. \"This announcement further confirms BroadWorks commitment to working with industry leaders to successfully penetrate new market segments.\" '),(5161,3320022,'21,June,2004','2004-06-21 00:00:00','   US : With more than forty operators already delivering advanced mobile services to customers in 22 countries and territories across Asia, the Middle East and Europe, 3GSM is asserting itself as the global 3G standard.   Delivering a wealth of innovative new services to consumers such as video on demand, high-speed multimedia and mobile Internet access, 3GSM exploits the potential of content rich information and communication around the world.   3GSM also brings substantially enhanced capacity, quality and speed of data transmission. Incorporating a radio interface based on Wideband-CDMA, 3GSM is the 3G-upgrade for GSM, developed jointly by standards bodies from China, Korea, Japan, the United States and the European Union.   3GSM is the technology choice of 98 percent of operators who have been granted spectrum in the 2GHz band identified by the International Telecommunications Union for 3G. As 3GSM rolls out across the world, customers in Japan, Korea, Hong Kong, Australia are already enjoying its many benefits, with launches to follow in Singapore, Malaysia and Taiwan this year. In addition, Indonesia and New Zealand are moving toward launch.   â€œAll 70 networks that will be delivering 3GSM services by the year-end are utilizing the 2GHz band in which China intends to deploy 3G,â€ said Rob Conway, CEO of the GSM Association and a member of its Board. â€œOnly one operator anywhere in the world has so far deployed an alternative 3G technology in this band. This global commitment is fundamental and demonstrates where economies of scale in the 3G market will be found.â€   Growth is expected to accelerate from the current 3GSM global customer base of more than five million, as the number of commercial networks increases to at least 70 by the year-end and handset launches by all the worldâ€™s leading manufacturers eliminate supply issues.   â€œThe greater availability of 3GSM handsets in Chinaâ€™s 3G trials programme is a further reflection of 3GSMâ€™s superior maturity in the 2GHz 3G spectrum available in China,â€ said Conway.   Craig Ehrlich, GSMA Chairman and board member of Hutchison Mobile, said: â€œThe opportunities for China with 3GSM are immense. Operators benefit from international roaming and economies of scale; manufacturers gain equal access to global markets; IT, software, application developers and content players are not locked out as in other systems, and ultimately consumers enjoy value from a greater variety of globally compatible products and services. These benefits can only be delivered by an open system and a non-proprietary philosophy.â€   Made up of board-level representatives of 21 operator groups, including the worldâ€™s 12 largest operators by subscriber count, the GSMA Board met this week in Beijing to acknowledge Chinaâ€™s contribution to the success of GSM, and to demonstrate the Associationâ€™s commitment to supporting Chinaâ€™s Government, operators, vendors and customers during the transition to 3G. China Mobile and China Unicom both sit on the Board.   The worldâ€™s billionth GSM user was connected in February 2004. Chinaâ€™s unrivalled contribution to this milestone was reiterated when the Chinese GSM user base passed the 250 million mark in April.   â€œThe impact of GSM in China has been profound,â€ added Ehrlich. â€œIt has been the primary driver of mobile penetration and has propelled China Mobile to the status of the worldâ€™s largest mobile operator and China Unicom to the number three position. China Unicomâ€™s GSM customer growth in 2003 exceeded that of its CDMA network. Also during 2003, Chinaâ€™s GSM users made a staggering 585.4 billion minutes of voice calls and sent 131.2 billion text messages. GSM accounted for 93% of Chinese mobile revenues.â€*   China alone has more GSM users than the global customer base of the next most widely used mobile technology - CDMA. China Unicomâ€™s GSM customers base is bigger than the combined customer bases of the worldâ€™s three largest CDMA operators â€“ Verizon Wireless, China Unicom & SKT.   â€œGSM is unique among mobile technologies in having a clearly-defined user benefit â€“ international roaming â€“ at the heart of its design,â€ said Rob Conway. â€œRoaming creates a discipline amongst operators and their suppliers to comply with open standards, promoting inter-operability and creating economies of scale that benefit end-users. 3GSM builds on this heritage.   â€œ3GSM is changing the mobile world. Users in Japan and Korea will benefit from improved roaming services as their deployment of 3GSM brings them into the GSM community. There will also be wider economic benefits as their manufacturers will gain greater exposure to the global GSM/3GSM market, concluded Rob Conway.â€   â€œAs we look toward the Beijing Olympics of 2008, the opportunity for China to serve a worldwide audience, and many millions of inbound visitors with compelling next generation mobile services has never been so apparent,â€ concluded Craig Ehrlich. '),(5162,3311843,'24,May,2004','2004-05-24 00:00:00','   urlLink NewsForge | Don\'t Know What PearPC Is? Now You Do!  MacOS on the x86 platform? This is a dream we\'ve all had for years, but although it\'s a nice try, we\'re not there yet though... But, we can always hope for the future!  '),(5163,3311843,'12,May,2004','2004-05-12 00:00:00','   urlLink Chris Anckaert  - sorry, it\'s in Dutch, as I\'m a Dutch-speaking Belgian...  '),(5164,3311843,'12,May,2004','2004-05-12 00:00:00','   I went over to my brother\'s yesterday evening to give him a demo of Knoppix. It was a CD which I got with my Dutch magazine PC-Active. The version is 3.3 His impression before putting the CD in was that Linux was overall much to difficult to learn, after all, he\'d been using Windows since he has a PC - which is a fair few years now. When the KDE desktop appeared with working sound his face simply lit up. He didn\'t have to say anything, I could hear him think \"wow!\". We had a look through the main menu-structure and just ran a few programs - Mozilla, KMail, KStars, the Control Centre. He was simply impressed. I have to say, even from the CD, Knoppix is fast and reliable. He\'s now going to play around with the CD for a while and will probably get more enthousiastic as time goes by. In a couple of weeks (days?), I expect a request to make him a permament install on his hard-drive. He had just had a lot of virus-related problems and his windows-box had grind to a halt a couple of days ago. He was very interested in learning that with this operating system, no more of that would happen. He was also worried he wouldn\'t be able to do the things in Linux he did under Windows, but after showing him OpenOffice.org, this was no longer an issue.  Oh, by the way, there is a new version of Knoppix, 3.4 - read all about it on the  urlLink Knoppix website  '),(5165,3311843,'12,May,2004','2004-05-12 00:00:00','   I would like to welcome everybody who happens to visit this page. Discussions I\'m interested in are about Linux, Open Source, programming, etc...  Feel free to reply or make a posting yourself.  Chris.  '),(5166,3311843,'24,June,2004','2004-06-24 00:00:00','   urlLink PC Pro - Computing in the Real World  Yahoo is now offering more mail-storage space. Funny how a vaporware product (GMail) can actually make others change their real and actually existing products... strange world sometimes.  '),(5167,3466740,'30,May,2004','2004-05-30 00:00:00','   Okay so it\'s not exactly Disneyland..but..The Buccaneer rules. I was there last night bein merch maid for the shakes and lemme tell you..it\'s full of pirates there..Captian Kidd, Captain black beard,Captain Morgan..  The Glossines were there too. They have a pretty rockin van.  Best part about the Buccaneer..besides the pirates...Phuket island is just up the street.  thats all for now.   '),(5168,3466740,'29,May,2004','2004-05-29 00:00:00','   Mc who? Mc what?   After viewing Super Size me with The Shakes and friends, I\'m never visiting the golden arches ever again!!!   HOWEVER..after learning about the demise of Mister Baskin..ice cream is still my morphine!!!  My favorite ice cream:  1.Mashti Malone\'s rose water with saffron 2.Pancakes Alaska at Du Pars 3.101er Brownie sundae 4.Hollywood Hills Schooner Sundae 5.All other ice cream!!  mmm eat some ice cream today and dream about me.   '),(5169,3466740,'27,May,2004','2004-05-27 00:00:00','   Those s\'cool girls songs stick in your head like ribs in your belly sometimes. I started singing L-E-T-S G-O quietly and pretty soon several 8-year olds caught on and started to sing too. There you have it. Another s\'cool day.   '),(5170,3466740,'26,May,2004','2004-05-26 00:00:00','   Okay, So I know I\'ve been kind of invisible lately, doing behind the scenes Teenacide work..but expect to see me around town real soon! You may have already caught me last nite at the Teenacide viewing party for Godzilla at the Nuart. That\'s right, last night I made a special appearance at the Nuart. NO, I wasn\'t searching for one of the Laurens...Although you should buy their record as soon as it hits the shelf..anyways..decked out in my \"I will Destroy You\" tee by Tom Neeley, I watched Godzirra destroy Tokyo alongside Teenacide\'s elite.   Well that\'s it for now!  xxteentern   '),(5171,3466740,'26,May,2004','2004-05-26 00:00:00','   2 more weeks!  Till then.  xxteenternxx   '),(5172,3466740,'25,June,2004','2004-06-25 00:00:00','   In celebration of the Checkers nomination for \"Best New Artist\", I put on my spikiest 4 inch heels and headed down to the Henry Fonda with Jim. Upon our arrival, we spotted George from the Checkers and headed inside where we were showered with free Stoli..(wink!).  There were lots of shiny booths there like tattoos and free hairstyling and amoeba..    Suddenly, I got some hair goop stuck on my hands..Heads turned as I squealed..until this guy took it off and put it on his mohawk. I realized that hair styling should be left to professionals only. So I let this cool girl named Cazzie style my hair while the suicide girls were on. Those suicide girls have a van you know. It is silver with SG on the side..I thought it stood for s\'cool girls..but I guess not. Anyway, we headed for the balcony where we began our evening of being kicked out of places. I got kicked out of at least 6 different places.Oh and the Like\'s lead singer gave me the skunk eye.  So after they kicked me off the roof for the second time..we decided to head out and look for better trouble.   P.S. If you are wondering what else I was wearing besides heels..you\'ll have to look for me in the weekly.   '),(5173,3466740,'08,June,2004','2004-06-08 00:00:00','   Okay camping at the beach and surfing all weekend..life seriously doesn\'t get much better! I spent my weekend in a gyspy camp on a bluff overlooking the sea..joined by Teenacide\'s own Janet Housdon, my rock n roll mama, betsy cuttlefish the mermaid, stitch and capn morgan, we took on the surf and had a blast.  You should try it. Better yet..you should join us next time.   '),(5174,3466740,'03,June,2004','2004-06-03 00:00:00','   That\'s right. Surfing was ranked number 3 (next to Softball and boogie boarding) as Teenacide enthusiasts\' favorite sport! I\'ll be catching waves all weekend and listening to my copy of \"Too Hot to be Bothered.\" Did I mention I\'m camping out right on the beach too?! Times like these require a summer van..  xxteenternxx   '),(5175,4044445,'19,August,2004','2004-08-19 00:00:00','   Helllllloooooooo... welcome. Thank you for reading this. My vision for M-power is to share with you things that raise my thoughts up to a level where I am thinking of good, looking for the good in my life and appreciating what I have and what is possible. I experience freedom every day. There are qualities, that most New Thought/Metaphysical teachings call the Qualities of God. My favorite list is Abundance, Balance, Beauty, Freedom, Harmony, Joy, Love, Order, Peace, Unity, Wholeness, Wisdom.  I admit that I view this list a bit different than other metaphysical teachers, because in my head \"God\" is that space between the molecules, which of course is everything and every where. What I call God is all possibility, all potential, existing in a timeless everywhere. I believe that this list is a list of high spiritual qualities and that the divine exists AS them, and I exist because of how I see myself in the relation to them. For instance.. the Universe is Abundant... you can see that with your own eyes.. how much sand is on the beach.. quite enough as it is in all of life. Now, how I feel and believe in relationship to Abundance is how I am experiencing prosperity. Either I believe and pray from a place of abundance and are prosperous, or I believe and pray from a place of lack of abundance and I am not prosperous. Most of us are some where in between, aren\'t we?   The Universe is Joy and I, my believe in Joy experience and express happiness! Some people believe in a lack of joy and they experience a lack of joy!     Either we see Balance or lack of Balance.. and our experience is chaos or order, depending on our belief.  Beauty or lack of Beauty... see pleasing things or the lack of pleasing things....   You get the picture...     The Universe is inclusive... and we create the way the Universe creates and we include things... we include in our thoughts and our experiences the things that we observe in the world. Literally we attract to our senses.. our sight, feeling, touching, tasting that very thing we think we deserve to have. We either attract our good OR our lack of good. Abundance or lack of abundance. Wisdom or lack of Wisdom.  I have a tendency to ramble...  To Empowerment.. here\'s to finding our way in this world by just including what we want in our thoughts! Here\'s to group spiritual practices...  For more about me and what I do, visit  urlLink www.empoweringwords.com   Namaste\' Mitzi '),(5176,4044445,'29,July,2004','2004-07-29 00:00:00','    Happy Birthday to my angel baby, Tonya Sioux...&nbsp; she was born at 1:40 PM on this day, in 1984.&nbsp; I guess this is the way a mother is suppose to feel 20 years later, but today I feel so much good, so much love, so much awesomeness at what has happened, to both of us, over the last 20 years!     Today I want to talk about  Beauty  as a quality of God,&nbsp;opps,&nbsp;rather&nbsp;&nbsp;a quality of our highests Selves; what&nbsp; beauty is, and what it means to be a beautiful human living in a spiritual world.   We are all beautiful, each and every living thing on this planet and that includes spiders and creepy bugs.. Because, why? Because Beauty is in the eye of the beholder and it is subjective to&nbsp;an individuals&nbsp;&nbsp;beliefs!&nbsp; It is an individual thing, yet  there is more beauty in this world that we see or experience with any of our physical senses!    So, what is beauty from a spiritual standpoint?&nbsp; I believe it is a state of mind, or consciousness in which we detect with our physical senses things that are&nbsp;&nbsp;pleasing.&nbsp; Beauty is what it is to each conscious being.&nbsp; It is just for each individual.&nbsp; Beauty is the vibration that is pleasing to our senses... Again subject to each perspective.&nbsp; We all filter every thing through our physical senses and our imaginations and what we know to be true. Learning comes from allowing different view points and opposite thoughts in so that we can find better and better ways of feeling, seeing, thinking, being.&nbsp;   We each, individual and together, see beauty where we want to.&nbsp; Beauty is the highest vibration we can pick up with our physical senses, it is in itself a sense... The sense of Beauty.&nbsp; Choose beauty, by looking around you and noticing things that are pleasing to your senses and the more you do this, the more you\'ll notice pleasing things.&nbsp; It would seem to me that the outcome of seeing more pleasing things comes from seeing pleasing things that it would feel good to see more and more things that brought pleasure.. And not just seeing, but feeling, tasting...&nbsp; An abundance of beauty surely leads to Joy!     Breath and as you do so, deeply and fully, think about how what you see with your eyes, or taste with your palate is beautiful, looking at it, eating it, smelling it allows you to feel pleasure...   Why wouldn\'t you want that?     This is Beauty, this is being tuned in and tapped into that Universal Beauty... The higher Truth is that EVERYTHING in the Universe, is beautiful to some one, to some creature that shares our oxygen.   Imagine that black widows find their mates so&nbsp;pleasurable, they eat him !!! &nbsp; There are many things in your life that are pleasing to your senses... When you focus on those things,&nbsp; you will see more of them, and the more pleasing experiences you feel, the more you\'ll want to feel...&nbsp; feeling good brings us joy and when we feel Joy, we feel Balanced.   Again, qualities of high spiritual living,   not just qualities of God, any more!     I want to say to each and every person that reads this blog(hahaha.. that words still makes me giggle)... YOU ARE BEAUTIFUL!&nbsp; No matter what, there is an inner beauty, that only is covered by your belief that you are ugly...&nbsp; If you can just be willing for one small moment, to see beauty in your own eyes, that will open up the door to your beautiful soul... Then if you can just see one small thing of beauty in your life and imagine things of beauty and see your self doing, saying, feeling those things of beauty, it will comes to pass in your experience... and more will be revealed.   Any human being can see beauty because beauty is in the eye of the beholder, it is in your own eye, if you want to see beauty in your life, look at the beauty that there is now and choose to see more, to feel more, to taste more, look for more and more beauty and more will be found!&nbsp; You decided what is beautiful for you.&nbsp; Don\'t give any negative thought to other\'s beauty that you don\'t find beautiful.&nbsp; Remember it\'s all subjective and all the beauty you see, or the lack of beauty (for this works for the lack of, as well as the abundance of, remember) is what is beautiful to each person.&nbsp;     &nbsp;   When we all share a common vision, then we will see a even greater beauty. Find those that see and want your same vision of beauty.&nbsp; Make it an intention to see beauty and call forth those friends that see the same beauty as you.&nbsp;   Next quality of spirit is Joy...   Tonya, I hope that you\'re having a great time tonight...&nbsp; I love you!   &nbsp;  I see a world of beauty in an ever changing landscape..   Mitzi '),(5177,4044445,'27,July,2004','2004-07-27 00:00:00','    Today the Birthday\'s start... and the anniversaries, too..     &nbsp;   20 years ago today I went into labor and it\'s been painful ever since... two days later I had a beautiful baby girl that came into the world scarred, had a mother that was numb and a father who was drunk...3 weeks later some friends of mine came and took me away and home to my family..&nbsp; thank you, Tom and Michelle...     &nbsp;   93 years ago my Grandmother, Mabel Yuill VonLindern was born...     &nbsp;   some where around 60 years my Aunt and Uncle where married and over 18 years my cousin and his wife were married...     &nbsp;   &nbsp;20 years ago...July, 29th Tonya Sioux Lambert came into the world...     &nbsp;   5 years ago on the 30th my brother married his second wife     &nbsp;   .... and it goes on.. but we\'ll talk about August in August!     &nbsp;    &nbsp;   So, it\'s a busy week, or it has been for my family...&nbsp; but for me, the most important thing is that Tonya have a great birthday, even if it won\'t look like she wanted it to... she\'ll be safe and have fun and that\'s all that matters...&nbsp; Twenty years ago...&nbsp; it\'s a mile stone, that\'s for truth.         Hey, weren\'t we going to talk about&nbsp;Abundance?       &nbsp;   So, the first Value that we assign to God/Spirit is Abundance... and that\'s just what I am talking about, Abundance!&nbsp;     &nbsp;   I have a rich family, abundant amount of love from family, from friends, from strangers on the  street! &nbsp; I make more friends easily as well...     &nbsp;   What is abundance? Its\' a matter of perspective, isn\'t it?&nbsp; Would an old man from the deep jungles of Africa have a different idea about abundance than man from Wall Street?&nbsp; And women, men and vise versa, old people and young. &nbsp;Every human being has a different perspective on what abundance is to them, so how could it be a \'quality\' of God?&nbsp; God is all things, so yes, abundance is a quality of God, but Abundance is also what it is for us and a thing that each one of us names for ourselves and each one of us, when we put away the belief in lack, can have and have it defined perfectly.      How?&nbsp; Well, there\'s the question...&nbsp; and here\'s the answer:    We each see, hear, taste, feel... in other words, perceive what we call either abundance or a lack of abundance in our physical body.&nbsp; Depending on where we are looking (either at abundance, or the lack of abundance) we are seeing and the more we experience it with our physical senses, the more we experience in our lives.&nbsp;     GOOD VIBRATIONS     No, that\'s not just some&nbsp; hippie from the 60\'s talk... we are vibrations.&nbsp; Look at it this way, Quantum physics teaches us that everything that exists, comes forth from what seems to be nothing and everything resides in possibility until we observe it with our physical senses and then it becomes real to us. Truly, that\'s how it creates.&nbsp; Here is a great web site to check out about it&nbsp;  urlLink What The BLEEP Do We Know &nbsp; Explore the possibilities.    When we vibrate with abundance we are thinking it, feeling it and experience it in some form in our lives.&nbsp; Here\'s a trick to get more abundance in your life.&nbsp; Sit down twice a day, morning and night and jot down 10 things to be grateful for.&nbsp; You will start noticing more and more things to be grateful for and you\'ll be grateful for more and more things...&nbsp; And guess what, it\'s still back to perspective, isn\'t it?&nbsp; Because what we are grateful for is only for us to say!    Abundance is having&nbsp;and being that which truly brings us Joy (another quality of God...or....).    Abundance is having loving family, if that\'s what brings&nbsp; you Joy... or a fat bank account, or a hunky partner, or a new deck build with friends, or friends that give with out asking...&nbsp; We are all abundant in some way... if you think I am not talking to you, then just try finding 10 things to be grateful about 2 x\'s a day... any one can do that.    Tomorrow we will be exploring Balance...    Live, Love, Laugh and be happy...    Mitzi     &nbsp;    &nbsp;    &nbsp;    &nbsp;'),(5178,4044445,'26,July,2004','2004-07-26 00:00:00','    There are 12 so called qualities of God that many New Thought teachers talk about.&nbsp; They are; Abundance, Balance, Beauty, Freedom, Harmony, Joy, Love, Order, Peace, Unity, Wholeness and Wisdom.   In New Thought God is defined at Spirit, all knowing, all being, every where present.&nbsp; Omniscient, Omnipresent, Omnipotent&nbsp; I don\'t think that we can even dare limit what God, aka Spirit is&nbsp;by 12 qualities.&nbsp; I do think that these 12 qualities are some of the highest&nbsp;states of consciousness...&nbsp;states of being human, that we&nbsp;are achieving these days.&nbsp;&nbsp;   God can also be defined as that highest&nbsp;vision that&nbsp;the&nbsp;\"I am\" within us&nbsp;has and thus, they are qualities of God.&nbsp; Yet living spiritual in a physical world and&nbsp;being human in a spiritual world gives us the opportunity to define our spiritual&nbsp;truths the way we want them to be.&nbsp; Each day, I am going to study the 12 qualities of being human in a spiritual world and find ways to apply them to&nbsp;my life and hopefully you\'ll join me by&nbsp;reading and finding ways to apply them to your &nbsp;individual lives. The 12 qualities tie into the lessons that &nbsp;   urlLink  Emma Curtis Hopkins    teaches in many of her books.   I will be writing about them&nbsp; from a perspective; a point of view of the human being living in a spiritual world and how I can use these 12 qualities in my daily spiritual practice and my daily life&nbsp;(which, ideally to me, is the same thing)!   I will write about them in the order that they are in here, alphebetically, with no other signficance related to that.&nbsp; I think that in order to evolve spiritually and live in this spiritual world, we all must start from where we are at, right here, right now! Every one is in their right place, working either from a place of each of these qualities, or a lack of these qualities...i.e. each human being is either living in a place of abundance, or lack of abundance and every point inbetween.     &nbsp;  I will start tomorrow then, with Abundance.&nbsp; '),(5179,4044445,'26,July,2004','2004-07-26 00:00:00','    Good Morning!     &nbsp;   This blog will be all about empowerment.&nbsp; All about learning how to become empowered from with it.&nbsp; Ernest Holmes, the founder of  urlLink Religious Science  and the author of  urlLink Science Of Mind &nbsp;says; \"train your mind to think what you want to thing, be what you want to be, feel what you want to feel and place no limit on Principle\".&nbsp; That to me, is empowerment!&nbsp;     &nbsp;   I am new to this whole thing, truly following the voice of Spirit and this web log is born out my communion with that Highest Self that I am.     &nbsp;   Many Blessings and things to come!    Mitzi '),(5180,4044445,'07,August,2004','2004-08-07 00:00:00','    Our friend Bonny has reached that right age of retirement, the age where she can start collecting her share of all that money they\'ve been taking away from her for all her working years! We are happy she\'s getting it, because we probably won\'t it. The way the government is going, we are just going to have more problems and no one is going to be able to take care of any one else!  Oh, ye of little faith.. I affirm that the consciousness level is raised enough, when the time is right, that we actually \"get it\". I don\'t think that we can survive four more years of George. I don\'t think we\'ll have to.  At any rate... at Bonny\'s party that Robert and I put together, Jon and Kevin helped, also... we had a \'proclamation\" from \"we the people to Bonny Johnson. Here is the speech we gave her and then presented her with her first \'gubment\' check!   On behalf of the United States gubment and the honorable Mr. President, Mr. Bush, We are moved to be issuing your first gubment check. Bonny Johnson has slaved her way through the last 50 years or so and by the sweat of her brow and the strength of her back, she has helped build America to the place that it is today! And for THAT Bonny Johnson, we salute you!!  We, the American people, the rest of us, are going to take care of you! We are going to send you a gubment check every month. Now noticeably it will not be as big as it once was, we have a war going on, youâ€™ve noticed, I am sure. There is also the matter of your sexual-orientationâ€¦. It does not quite come inline with the morals of America. For that, we have to issue a â€˜sinâ€™ tax on your gubment check. Add to the fact that you are single and weâ€™ll have to really take care of you in a few years, we are going to have to reduce the check by some, you know, to save away for the future! We want to always be able to take care of you, Bonny Johnson! And I know that youâ€™ll want to continue to help educate Americaâ€™s youngest citizens and incarcerate Americaâ€™s evil people, so weâ€™ll have to take education tax and prison tax out of your gubment check. But donâ€™t worry, Bonny Johnson, we are going to take care of all your medical bills! Medicare CARES, Bonny Johnson, and we will provide some things for your well-being. Live, love, laugh and be happy, live that life of carefree existence you deserve! You are in good hands with the American Gubment at your side! Bonny Johnson, here is a note from our esteemed President, that he sends with his thanks:  â€œDear Mrs. Johnston: It is my gratitude that goes out to you at this time, your daily devotion of service to this country at a time like this was a Christianly thing youâ€™ve done. I mean, because of your morals and values, this country is a better place! Congratulations on the start of a great journey! Please remember our motto: â€œweâ€™ll leave no senior citizen behind!â€ Thank you, President Bush.â€  We are all there for you, Bonny Johnson. Go forth and prosper!  The American People '),(5181,4044445,'04,August,2004','2004-08-04 00:00:00','    Robert had seen on a church sign one time, \"let the Lord have His way\"..and after laughing, laughing and laughing about the visual, we decided that letting Spirit have It\'s way was a pretty nifty trick for the human being!  The thing is, when we can live from our spiritual mind and not the animal-human instinctual mind, we some how seem to have a more rewarding, enriched, loving life. Things seem to be more easy and life seems to arrive at our awareness in a way that allows us to make choices that are more related to the \"Qualities of God\".  I think that letting Spirit have Its way with my life means that I live more in harmony with that greater Life that is around me. I have no deep understanding of the why grass comes up in the spring, but I know that it does and I know that there is intelligent people on this earth that understand that process more than I do. So, I refer to them. But there are things that no one seems to understand, and to that I defer to something that, while with in me is more than I am. That is Spirit! Living in harmony with that greater awareness of Spirit we find that life is more peaceful.  The answer, I think in living more in Harmony with Spirit is in doing spiritual practices.. We practice them until we embody them, then we are living those spiritual practices. The ones that work for ME are meditation, journaling, group spiritual practice, Treatment and just conversing with people who are like-minded!  It\'s all a journey that we are on.. and I just want to remind people that it\'s easy to live in Harmony with that Spirit within when you do *some* spiritual practice!   Let the Spirit have It\'s way with your life!  Mitzi '),(5182,4044445,'04,August,2004','2004-08-04 00:00:00','    I was reading where Elizabeth Edwards, wife of our future VP, John Edwards is a large woman... I am sooo excited about this fact!!! Being a well rounded, juicy peach of a woman myself, I am delighted as goat cheese that she is going to be in the public eye! She is beautiful! I also love the spirit if our new first lady, Theresa!  BTW dear reader, do you vote??? If so, please vote for Kerry and if you don\'t vote, please do so this year, it\'s sorta important. But vote for Kerry. It\'s been a rough four years.  Anywhooo... I love that more women of size are getting in the media. I love that we are valued more for our worth as valid humans then what form our form is. Size, color, sexuality, religion, gender.. etc.. should not be an issue in relationships with others. We all come from the same planet, we all are on our paths, take away the skin and social structure, we are all the same.  This is what conscious evolution is... we realize our UNITY with each other and with Life it\'s self.  There IS something more intelligent than we can be at this time. There are forces at work for the good of this planet. There is only One Life and I can\'t be convinced otherwise.  There is one mind and we all use it... This one mind manifests as human and it has a physical body and brain, but is the breath of Spirit. That\'s that Unity of mind and body and spirit. One, three modes of manifestation.  Unity... We are all One..  I don\'t know what else to say on that!  If you live from the ego consciousness, one would be a lonely number...  Unity.  Mitzi '),(5183,4044445,'03,August,2004','2004-08-03 00:00:00','    Interesting show today... why do people have to pick on people that are happy.... all these women that are going after the younger man....hmmm.. Hey, the best thing about being 40 is that at forty, I can take a man twenty years younger! So, OK, he was 5 months shy of being and 18 and he was her student... I can see the student thing.... there does have to be that sacred trust.. but she lost her career, but now a baby and a cuuute young man.. He\'ll be horny for years to come... hehe.  Hello! Good day. When the koan \"what is the sound of one hand clapping\" first came out, didn\'t they know how to make noise with just one hand? Didn\'t they watch Simpsons??  It\'s amazing to watch both my daughter and husband going through their own growth process.. they are both angry at everything and mostly me.. which has been a challenge because I can\'t *stand* it when people are mad at me! I used to be the HUGEST people pleaser... now people can serve themselves if they are going to be mean! But truthfully, I *love* to serve and if Tonya and Jon would figure that out and treat me like the queen that I am, I\'d be much nicer to them.. does that make as little sense to you as it does to me? Seriously, though she is my biggest challenge, my funniest puzzle and my greatest love... And he is, too.. .although I have no idea where we are going, but I thought as long as I am here, I might as well be the best I can be.  So... check in with yourself...breath, relax, throw your shoulders back, smile a good smile..  Do you know that it\'s difficult to think of negative, worrisome things when you are smiling really big and cheesy??? Try it.. paste your biggest, brightest smile on your face, take some deep breaths, hold them, smile for the count of 10, but try to think of something \'negative\' in your life... hard to do.. So, think of having an abundance of what you want!  Smile, breath, smile, breath.... just don\'t chew bubble gum!  waiting for autumn.. Mitzi '),(5184,4044445,'03,August,2004','2004-08-03 00:00:00','    Does it really? I mean here is one day with the intention to write *something* every day, and here it is three days later?  And nothing much has been happening; except that my friend Robert and I are building a web space together...  our computers and idea factories are shacking up, hooking up and becoming one.  Here is our Space  urlLink http://www.winwinsolution.biz   You can\'t buy anything just yetâ€¦  but check it out, get on our email list, please...     It does feel good to be moving together in solution with Robert and our future customers, partners, friends, and so on!  I don\'t remember what quality I said I would talk about next, but I am feeling most is Freedom.  Freedom is a quality of high spiritual living that expresses as a feeling or experience of freedom.  The feeling of freedom is different for every one, individual in the way it expresses It\'s self in life.  The feeling of Freedom is yet another way that we feel our Divine Connection!  And Freedom is felt in our lives as we believe Freedom feels like! Imagine your highest feeling of Freedom! When I was a kid, that feeling was roller coasters or motorcycles... man what a free feeling, to be flying in the wind, knowing that your fate rested on KNOWING fully that you would be All RIGHT... that was a feeling of pure connection and great feeling of energy rushing through my body! I have that same feeling when I jet ski as an adult.. a rush of fear followed by an instant knowing that I am going to be OK...  So, what feeling can you remember about Freedom and the way it feels to you?  It does not matter when it happened, or how short it was, it\'s all about the feeling.  Sit back and relax and try to remember feeling Freedom.  Feel it in your mind by remembering, feel it in your body, in the middle of your stomach and lower... feel it as deeply as you remember feeling it by remembering the experience in your mind...  Our human mind does not know the difference and will release the same chemicals that were first produced if we give it time...  So, relax, remember and save a moment for the feeling.  The next question is what can you do, now... to bring that feeling.. Remembering is great and remembering with feeling can bring thoughts, desire, feelings, people, advents, etc that revive that first experience of Freedom.   But taking action on the feeling brings those advents, people, thoughts, etc. a bit faster and freer than you could quite imagine... you see, you are adding your agreement with the highest Truth you know.  And any time that you agree with Spirit, your experience of those High Spiritual Truths like Freedom flow more easily, more naturally into your experience!!  Imagine what brings you Freedom!  Feel that through out your whole body as you re-imagine what Freedom feels like...  take action towards Freedom that comes to you...  as you agree with Spirit and see Freedom every where in your life, you raise your vibration and more and will be revealed.  This is evolution! This is Freedom!  Freedom... for me, that feeling that the Universe is in Divine Order and as I connect to that Highest Truth with in me, I recognize that that feeling as Energy that flows through my body with Joy and Trust and Understanding.  Freedom of choice, freedom of Life is the Highest Truth I can imagine. I know that as Freedom manifests in all things, Freedom manifests and expresses as me, in me and in my experience! I experience freedom in my expression and in my affairs, in my spiritual practice and in my physical life.  I am free to be my highest expression always... and I am.  I speak my word for every one who reads this Blog and I know that they have with in them the Power to feel Freedom as well.  I know that as they read, they experience what every Freedom means to them and Freedom expresses in their lives.  I affirm that at the level of their acceptance of their good and of Freedom, it expresses in their lives.  I see each and every individual as a perfect and clear channel for his or her highest Good! I give thanks for Understanding and Clarity and Freedom!!  WooooHooo..... :) :)  I release my word, my prayer.. into that unified, harmonized consciousness that is One!  It is done as we believe! and so it is!!  Yes, this is good!  Moving together in Freedom!   Mitzi '),(5185,4044445,'12,August,2004','2004-08-12 00:00:00','    The Nature of Spirit One Mindâ€¦ present everywhere, always, in all things, as all things. Spirit/Intelligence is the Earth and the rock. It is the Ocean and the raindrop. It is the Fire and the spark. It is the Hurricane and the summer breeze. It is the Mountain AND the molehill.  Energy/Intelligence does not care if it is the storm or the calm, it is both, and water does not mind if it flows through the rapid cursing river or the wide, meandering stream, it is both. This Energy and Intelligence that becomes all things comes from nowhere, according to our physical senses, yet it is everywhere and is all things.   The Nature of Me The difference between me and the rock, raindrop, spark and breeze is that I have some level of conscious awareness. I can choose what to think, how to feel, where to go, what to experience. The rock and I are made of the same things, yet I breathe and I move and I sing and laugh and dance. I choose and that makes all the difference.  Being all things, this Energy and Intelligence is what I am. As I move and as I breathe and as I think, I am. Because I duplicate the nature of Energy and Intelligence, I can be any thing. As I think, as I feel, I create.  I am vibration, which is energy. I am consciousness, which is Intelligence. I have the capacity to KNOW who I AM. I have some intelligence to choose WHO I am and WHAT I am. I duplicate the nature of Spirit as ME. I can do ANYTHING that I can conceive in my mind. I canâ€™t do everything; I have physical limits, at this time. Evolution of my Spirit-Self renders even physical limitations a choice. The All does not care what the I is or does, IT is all things. It does not delight, It is. I call it Spirit because for me, that is the highest Ideal I can imagine with my limited physical senses, but this is good because as I evolve in my own spiritual expression, which means Spirit-Self, more and more becomes revealed to me.  Duplicating the Nature of Spirit in ME I affirm for myself the power of choice in every single thought and feeling and action! I affirm that I choose thoughts, feelings and actions that bring forth joyful experiences. Iâ€™ve decided that Joy is the highest consciousness for me, and it is my Intention to think joyful thoughts and feel joy and take action that creates experiences of joy as my life! I simply allow Spirit, which is everything, to BE Joy as me. My life that is as Spirit, then becomes as I desire it, as I intend for it!  My thoughts bring my feelings, which brings my experienceâ€¦ my own unique experience. I USE the Energy and call it Mind or thinking or feelingâ€¦  I use it uniquely according to my unique thoughts! Not one person has all the same thoughts as another, yet we all use the same mechanism for thought and feeling. Everything I experience in the outside world is created from the same energy and intelligence that is everything. Everything is a vibration. Vibrations like to match. I think, I feel and I take action, which creates a Vibrationâ€¦ I include in my vibration the things i think about, which creates feeling, which tells the vibration which frequency to vibrate in. Depending on the frequency of my own special vibration, i attract TO me, the same Energy frequency. The question then becomes; what do I want to have/be in my experience? How long do I want it to last? How good do I want it? We canâ€™t seem to answer what our Spirit-Self has as lessons, but what I believe is that the more evolved I am the better it gets, so what ever I have now, regardless of how Good I think it is, has the potential of getting even better! So, I want good. I want experiences that are fun, joyful, loving, abundant, opulent, balanced, beautiful, free, harmonious, peaceful, feelings of unity, wise. I take time to remember this Intention and change the very thoughts in my own unique mind. I do this to reflect the Intention of LOVE that Iâ€™ve set, the highest vibration I can feel, to experience the best adventures I can. I can only do this for ME. We all have the potential for anything we can imagine. This is good stuff!  Thank you.. Gratitude feels good. Gratitude gets me to thinking of good things in my life. I remember that I am in charge of my thinking and I relax and remember to appreciate the good in my life. I want more good, I appreciate whatever form that good is appearing in the moment, in the Now. When I groove on the good in the NOW, I automatically get Good in the Future. What feelings I think from, I experience. Thinking thoughts of appreciating brings more things to appreciate! It is my intention to stop often and appreciate how Life is showing up for me. '),(5186,4044445,'19,August,2004','2004-08-19 00:00:00','    out of no where, an odor offensive enough to make a growing boy proud. I looked around for the dog, naturally, because Jon was not home, but Buford was nowhere in site, nor where the cats. Spooky was eating and looked at me with sheer indignation when I suggested that the odor might be her expulsion. Then I realized what it was... One of the \'girls\' had left a morning sooper-dooper pooper.   See what I have to put up with????   Tonya came  home in an altered state last night... (family members would be humored to remember, \"here, kitty, kitty, kitty\" )  If you don\'t know ask about it in the \'comments\' section and my mom can answer them, she would LOVE to tell you all about it!   Annnnyyywayyyyyy....  she figured I\'d be mad at her, but I know that the consequence of her choice will hit her when she wakes up... I couldn\'t put that kind of punishment on her... It\'s Karma...  although Tonya insists on call it bad luck.  Ha!  Ernest Holmes said that there was no sin but a mistake and no punishment but a consequence... and you can\'t get it wrong, the Universe is folding out excatly like it\'s suppose to.  Don\'t you agree? '),(5187,4044445,'17,August,2004','2004-08-17 00:00:00','    I did nothing but watch tv and the remote\'s been lost since breakfast!  Well, that\'s NOT true, and I did have a great day, but a very strange one as well... I think lots of healing on a deep level took place and I also realized today how far I\'ve grown.  My daughter proclaimed her hate and disgust for me. Apparently, even though I have been working hard the last month building our e-biz\'s, she thinks that all I do is sit on the computer and play... well, OK, I admit, I am having fun, fun seems like playing and I have been doing it on the computer! What appears to be one thing, is really another!  Now when she was doing her ranting and raving, I was remembering that this is about her confusion in her life, her frustration and her pain of her best friend being on the streets, strung out on meth.. Tonya cares so much, perhaps too much. I think I was her age when I read, \"Women How Love Too Much\" for the first time.. perhaps I should find her a copy.      So, this is what I learned today.    I have a vagina!!   No, that\'s not true. I mean, it is, but what I learned was far more reaching.. .. what I remembered was that I was not so much one to find my own way, but to define a way that is already present. I have always liked to redefine words. I think that we redefine words all the time, but no one wants to admit that we do. Remember the 1st gay nineties.... the last one was just not the same gayness... gay meant happy, now not many  remember  happiness, but no one confuses it with gay! Have I made my point?  For years I\'ve been trying to find and refine my definition of God. I have tried NOT to call that \'Thing\' that is everything, God because of what it meant for so many people.. Ugh, an angry, controlling, evil watching, hard ass, do as I say, not as I do type God... again, ugh, my mother was like that, too...  (just kidding Mom!)        So, my higher self asked my lower self, I mean; in my meditation I had this flash of intuition.. \"this isn\'t working, redefine and teach them your definition of God\". I just said Yes!   Here is my story and my re-definition of God:  God is that space between all the molecules. Since we can literally see, going small enough, there are really not any molecules, God is all that space. God is all  possibility  and all  potentiality . God is the wave and the particle. The particle is life, the wave is energy and it is every where, all the time.  The particle becomes form, life... and the wave energizes the form . God exists as time and space, all possibility and all potentiality. This space between the molecules, this space that is every where, at all times has the possibility to potentially BE anything.   What is it that makes nothing become something  ???  My answer is that it is the observer. What is the observer? Good question. I think that it is the sum total of all potential, everything that is possible. What is possible? What is NOT possible? Really, think about it.  So, the sum total of all things is the Observer .  ( I felt an urge to capitalize this really huge thing, the Observer)  But here\'s the best part... I think WE are the Observer! Or rather, the Observer(being the sum total of all things) becomes ALL THAT IT OBSERVES. Hence, we are the Observer. We have form, and we observe. We are made of particles and energy. We have potential to become anything that is possible. Anything is possible.  Dare to believe it !  Aren\'t convinced? Just imagine the possibility if that were true! Would believing that make you experience more freedom?  I bet the Wright brothers  imagined  flying; soaring through the skies, feeling a great and holy sense of Freedom. I bet they studied all the previous wisdom on flying, I bet they studied how God accomplishes it. If you can imagine it, you will see it happening some where in that space between the molecules. Remember that space between the molecules contains all possibility and the potential to become what it is that you see! So, they imagined, they saw themselves doing it, studied the wisdom of the ages and experienced it in nature and the figured out how to do it. Do you imagine that when the Wright brothers flew for the first time, that they felt FREEDOM? I imagine they did!!   Back to the Observer . When the Observer observes the space between the molecules where everything is possible and they see a potential, that potential is immediately formed. Depending on the amount of observation on this potential, the stronger, larger, denser the form. The form becomes a mold in which becomes filled with that space of all potential and all possibility. What we observe becomes form. And even though form is always filled with all possibility and any potential, it becomes...inside the form...that which observed IN the form. IF in that space of all possibility I see fearful, sad, bad, evil things, that is what form possibility is going to take. It becomes potential fear, sadness, badness and evil. The more fear, sadness, badness and evil that the Observer sees in the form, the more the space of all possibility becomes filled with that potential. The potential becomes form, and the more the Observer observes the form, the stronger, bigger it gets.    So, the lesson there, Dear Reader is to see only good things in the space between the molecules.        Mitzi '),(5188,4044445,'16,August,2004','2004-08-16 00:00:00','    yet my computer has slowed way down... What is *that* all about??  I just finished a three day project at Beloved Deborahs, so today I am taking a rest from physical work to love work, working on the websites. I am having a great time; mostly because it\'s almost 5 pm, and I am still in my pajamas and NO BRA!!  I have a question I was pondering and I thought I\'d share: Why does the physical world seem to be such a contrast to spiritual? Answer is that it\'s not. Yet I started this thread of thought, prompted by the Biblical quote, \" . . . In the world ye shall have tribulation: but be of good cheer; I have overcome the world.\" John 16:33. Of course that\'s Jesus being quoted, and I am sure it\'s right;  after all , it was only a few hundred years after he left this planet that it was wrote down! Also, remember Jesus was quoted as saying something like, \"anything I can do, you will do and more..\" So, it must be true that  WE  must over come the \"world\".  The world in biblical means the physical form. Imagine this; everything in the world, our physical world, the earth and everything that is created; either by  US  or  Something greater than us,  is made from the same thing and it comes from the same source! The world exists because it is NOT contrast to the spiritual world,  but a slower, massive vibration of the spiritual world .   Having awareness  of the spiritual world, there is a whole new set of rules to play by! We are living  in  a spiritual world and  having  a spiritual experience!! Yet, the Biblical world is that world that has slowed down... a place where humanbeings are   infinite Spirit having an individual expression as ME!         I\'ve evolved to a space where I know that the only game in town is that this is a spiritual universe. The world, the earth, the matter of this plane of existence is still  spiritual , just slowed down in vibration to the point that mass gathers! We are always fully supported by the laws of the Universe! Physical beings in this dimension attract energy to themselves the same frequency as we release energy. The energy that is attracted produces the experiences that we have.  So, since Jesus was just another   \"infinite Spirit having a individual expression as ME\",   then I am too. And since Jesus (and other masters, that have come to teach and lead us) said that what they could do, we could do and more,   then I too will over come the world!      Sure I will... it\'s as simple as continuing to spiritually evolve! This is what  urlLink Empowering Words  and  urlLink WinWinSolution.BIZ  are all about... products, news, education and information to empower every individual expression of the Infinite!          Won\'t you come check us out? '),(5189,3677153,'18,June,2004','2004-06-18 00:00:00','   I\'ve got waaaaaaaaaaaay too much time on my hands. It\'s the longest summer break I\'ll ever have, and on only day two I\'m bored. Thought that\'s probably the fault of these things-  1) Rain, insecent rain!!!!! 2) two huge bruises on my thighs, a nailed wrist and two twatted elbows. Cool bandages though. 3) The fact that everyone appears to have vanished. Hello?  So, I\'ve decided to make this. It\'s for when things actually kick off. And they will- this summer I\'ve got loads of stuff planned. On the summer soltice I\'m off on an uber bike ride from sun up to sun down. Thing is, my other bike mates are either in school, at work or they\'ve got the food tech exam. I\'m just praying for good weather- according to the BBC sun up is at 4:36, so I\'ve got to wake up at 4 o\'clock. Sun down is twenty to ten. And it\'s lighting storms. WOOOOOOOOOOOHOOOOOOO! I don\'t give a toss- I\'m doing it anyhow.  The weekend before that it\'s the southern area championships for our class of boat. Mmmmmmmm needless 5-6 hour drive burning 122kg of carbon dioxide. To put things into perspective, walking for 6 hours burns only 300 grams. But what dad wants, we must do.   Then there\'s the olympics to watch (sailing and mountain biking only for me), two weeks in Greece ourselves, the national sailing championships (where I\'ll get 1st junior even if it kills me), my birthday the day after that, then 4 days in the lakes with Oli and Will, home on the results day. Then it\'s back to the grinder at college.  It\'s going to be mint, and even if it hoons it down all holiday, I\'m going to make it mint. '),(5190,3677153,'21,June,2004','2004-06-21 00:00:00','   This Weekend Thanks To- Big fat waves, surfing, sunshine, rainbows dancing in the waves, talent  No thanks to- Big fat waves, rain, falling in, crap rudder fittings, huge southern prices  Quote of the weekend- Bloomin\' \'Eck  OK, I know you\'re not all sailors, so I won\'t bore you to fuck with technical stuff.  Getting dragged out of bed at 5am (FIVE FUCKING AM!!!) by an angry father is never a good start to the day. Off and out by 7, my dad angrier than ever, and onto the motorway.  Why do so many people walk their dogs at such unsociable hours? Why are they not in bed?  Anyhow, after enduring 6 traffic jams (howl!) and getting stuck behind just about every slow driver in Kent and Sussex, we arrived at Bexhill-on-Sea (the name says it all).   It\'s howling a gail, and the sea\'s rougher than a Barnsley lass. Oh dear.  Throw the boat together (we\'re late), get out and sail. We survive (just) the huge 6 foot waves, I get soaked through and we finish 5th. Out of only 13.  The evening livened up when we went to the club and joined int eh getting drunk. I wandered round the town (my god, there is a lot of talent on the streets of Bexhill- and some sail!) for a while, and then we headed back to the tent.   The 2nd day was windier, wetter nd it rained for a bit. We were doing ace- and then the rudder fell off! After that we capsized, and so we gave in on that race. The 2nd race of the day resulted in me nearly getting knocked out, and two capsizes which really knocked the energy out of me. But it was fun. We came 9th overall.  The drive home, well, the less about that the better- 2 hours in a traffic jam...  '),(5191,3677153,'21,June,2004','2004-06-21 00:00:00','   Today Thanks To- Bandages, cotton wool, solitude, The Streets\' Dr Your Eyes  No thanks to- lonliness, Lucien, The Streets\' Dry Your Eyes  I\'ve got waaaaaaaaaaaay too much time on my hands. It\'s the longest summer break I\'ll ever have, and on only day two I\'m bored. Thought that\'s probably the fault of these things-  1) Rain, insecent rain!!!!! 2) two huge bruises on my thighs, a nailed wrist and two twatted elbows. Cool bandages though. 3) The fact that everyone appears to have vanished. Hello?  So, I\'ve decided to make this. It\'s for when things actually kick off. And they will- this summer I\'ve got loads of stuff planned.   This weekend it\'s the southern area championships for our class of boat. Mmmmmmmm needless 5-6 hour drive. But what dad wants, we must do. It\'s looking wet with no wind, and I\'m wrecked as it is.  Then there\'s the olympics to watch (sailing and mountain biking only for me), two weeks in Greece ourselves, the national sailing championships (where I\'ll get 1st junior even if it kills me), my birthday the day after that, then 4 days in the lakes with Oli and Will, home on the results day. Then it\'s back to the grinder at college.  It\'s going to be mint, and even if it hoons it down all holiday, I\'m going to make it mint.  '),(5192,3485630,'29,May,2004','2004-05-29 00:00:00','    We just finished our gig at the warehouse, i think it went pretty well.  I couldn\'t hear myself so i just was hoping i was in-tune.  I can\'t wait till Ian graduates.  He forgot that i had to be at the warehouse at 3:30 today, so he went to jennys instead.  Then when he got home he was blaming it on me.  He was like, \"You shoulda walked bitch!!!\".  I hate him soooooo much.  Then my mom got home today from picking up my dad at the airport, she came to see us practicing at the warehouse.  She lectured me on how i need to \"ask\" before i do anything.  Our practice was re-schedualed so i just showed up at three when it was re-schedualed for, and my mom got mad because i didn\'t call her cell and ask if i could go a little later.  What does she care??? She wasn\'t going to be at home anyways, she just wants to know EVERYTHING that\'s going on in my life.  Then she watched our ENTIRE concert, and insisted on driving me home afterwords.  I was gonna help load the truck back up, and she was just sitting alone at a table.  AAARGGGHHHH!!!  This summer she\'s gonna make me study for next year, she\'s buying me all of the textbooks for sophmore year and making me read them all.  Let me have a LIVE PLEASE!!!!!!!  Anyways, today was pretty nice, i watched harry potter all afternoon (DRACO IS SOOOOOOOO F***ING HOT!!!)  I guess i\'ll be studying tomorrow all day.  I NEED at least a B on my math final to pass the class, i hate that class.  My teacher, johnson, thinks i have a learning disability so they are going to test me.  I have been tested by SOO many people for an LD but the can\'t find anything big wrong with me.  I guess they\'ll just keep trying till they can stamp me on the forehead with the word TARD.  All i need is a lil extra help in math, i\'m not that stupid...really.  This one time when i was taking a tard test (that\'s what i call my LD tests) they had one of those big mirror things that people watch you through.  I forgot that people are accually behind those things and i started making funny faces at the mirror!!  That day they diagnosed me w/a minor comprehention problem(don\'t laugh at me, i can\'t spell anything!!).  I can\'t wait till the next HP comes out, all the guys will be sooooooooooooooooooooooooooooooo hot!!  I could eat them all *chomp* *chomp*.  I like these blog dealies, they are fun.  Ok, there\'s this one dude that i like, but he never notices me...he notices everyone else, but me.  I try to make myself stand out, but to him i\'ll just be Jordy.  Maybe i\'ll just wait around, and someday he\'ll finally look at me and notice that i\'m a lil more than just Jordy.  I\'M JORDY!!!!!Maybe i should scream that in his face, tha\'d be funny.  Hey, what\'s up? I\'M JORDYYYYY!!!!!!!!!!!  Damn it, now i can\'t do that, because for all i know he\'s reading this or maybe is friend will read this and tell him about it and he\'ll be like, \"if anyone says I\'M JORDY!!!! Then i\'ll just ignore them more\".  Eh, i doubt that, i don\'t even think he knows i have a blog, good.  I\'ll just say his name over and over again....nah, maybe not.  What do u think i should do to get his attention?  He likes this other girl though, and i don\'t know if i should do anything to interfere just incase this makes him super duper happy.  The least thing i wanna do is make him worried or stressed or sad.  I\'ll stop rambling now, and write a list of all that things that make me happy so i can be happy again.  Everyone should read this list and i might make them happy too.  This is not in order of importance. 1. warm sheets out of the dryer. 2. waking up on saturday morinings, and knowing that i could either get up or stay in bed. 3. Having my brothers be nice to me. 4. Orange Sherbert 5. Seeing my guy smile. 6. knowing that i have numerous packages of sugary candy hidden under my bed. 7. Dancing w/a good dancer. 8. Shaking my tambourine 9. People laughing at my stupid jokes. 10. People knowing my name, when i don\'t know them. 11. Being noticed by people, anyone really. 12. Eye contact w/hot dudes in the hallway. 13. Wearing my favorite jeans and me chucks. 14. I bet you\'re bored now so i\'ll leave 15. If you\'re still readin this the you have no life 16. Bye This song is for that one special boy that i was talking about earlier. \"White Flag\"  I know you think that I shouldn\'t still love you,  Or tell you that. But if I didn\'t say it, well I\'d still have felt it  where\'s the sense in that?   I promise I\'m not trying to make your life harder  Or return to where we were   I will go down with this ship  And I won\'t put my hands up and surrender  There will be no white flag above my door  I\'m in love and always will be   I know I left too much mess and  destruction to come back again  And I caused nothing but trouble  I understand if you can\'t talk to me again  And if you live by the rules of \"it\'s over\"  then I\'m sure that that makes sense   I will go down with this ship  And I won\'t put my hands up and surrender  There will be no white flag above my door  I\'m in love and always will be   And when we meet  Which I\'m sure we will  All that was there Will be there still  I\'ll let it pass  And hold my tongue  And you will think  That I\'ve moved on....   I will go down with this ship  And I won\'t put my hands up and surrender  There will be no white flag above my door  I\'m in love and always will be   I will go down with this ship  And I won\'t put my hands up and surrender  There will be no white flag above my door  I\'m in love and always will be   I will go down with this ship  And I won\'t put my hands up and surrender  There will be no white flag above my door  I\'m in love and always will be   '),(5193,3485630,'28,May,2004','2004-05-28 00:00:00','    Well, this is my first blog EVER.  I\'ve wanted one of these, but i never had the time to write one.  I dunno who\'ll read this, i guess anyone can i guess.  Well, today was fun.  Nothing interesting.  I had lunch w/Maja today and met ben Brizell.  Today i accually talked to this one dude that i\'d had my eye on for a LONG TIME!!!  He\'s a senior so i figured what the hell?  I\'m watching friends, i like friends.  I wish it wasn\'t over.  I\'m glad freshmen year is pretty much over, but i\'ll miss all of my senior pals.  I don\'t wanna be a sophmore, they are like 2nd year freshmen.  Does anyone but me have dreams of people and they are really really nice in your dream, and when you go to school you like them, but they don\'t really like you.  Damn you dreams!!! Eh, ian\'s having a keg of root beer at his grad party, that\'s sooo awsome!  At Ian\'s party i\'m gonna try and see how many times i can hit on Jon Stack, and maybe get a reply.  Hee hee, i\'m brutal.  I hope something interesting happens soon, my life is boring.  I can\'t wait till the new freshmen come into band, it\'ll be soooooo much fun! I was at rocks yesterday and i saw a table of these completely hot guys, and i couldn\'t help but stare.  After that i felt so wierd, like someone that\'ll never get a guy like that.  I felt inferior, and ugly.  I don\'t often feel like that, i\'m a confident person.  I hope that that feeling goes away, it\'s lingering.  I think like I\'ll leave now, and continue eating my M&M\'s. Bye, -Jordy   '),(5194,3624333,'18,June,2004','2004-06-18 00:00:00','   The purpose of this meeting was to arrange for Dave\'s 40th b-day party. We are longgggg time listeners of WFMU the local free form radio.  And my gift to Dave was to get(hire) a DJ...so he doesn\'t have to \"do\" the music at his own party! I thought long and hard and asked for many recommendations from friends and colleges on who I should see/hire.  Then....one Friday afternoon...it hit me like a ton of bricks....My favorite DJ on FMU was playing some really wicked shit!  Could I ask him? I did.... and he does...do parties that is....and he likes good food....Beth and I cooked for him while He & Dave looked over our \"stuff/vinyl\". We had a mixed greens salad with gold and red beets, avocado, heirloom tomatoes, cucumbers, red and orange peppers, jicima, and chilpolte/ maple glazed Shrimp.  Then for dinner we made \"Patagonian Snaggle(i added the snaggle part)tooth fish\" seared with light dusting of curry pwdr. over mixed sprouts, in a infused coconut broth with cilantro oil.  Dinner was GREAT ! Company was GREAT!  Couldn\'t have asked for a more pleasant, intriguing, fun night!!!!!!  After having to deal with shit stains who roam the earth aimlessly...My faith in mankind is restored, after having met DJ Small Change AKA Jim Dier!  Thanks for a blast of an evening!  '),(5195,3624333,'13,June,2004','2004-06-13 00:00:00','   See previous post....Trying to get him on the WWW!  '),(5196,3624333,'13,June,2004','2004-06-13 00:00:00','   Well my first rant will be about how my weekend started off with a face to face with my nightmare contractor Angelo Rella from All pro construction aka All Pro demolition in West orange, New Jersey.  They are listed in \"Orange, N.J.\".  I find out later that people who know him refer to him as \"that crumb\". I am the unfortunate neighbor, who suffered the consequences of all pros\' incompetent work. What started out over 10 months ago quite amicably, ended with threats against my life... Anyone who reads this pls. never get involved with these ALL-PRO people!!!!!!!!!!!!!!   But soon after that things got better...Saturday was the most beautiful day on the beach at Gunnison that I have ever seen in my 15 years going!  The ocean was a sheet of dark blue glass, and the sky was so clear that could see the parachute ride in Ocean Beach Coney Island.  The sea water is another story....ewww.  But then again, we don\'t really go for the water...LOL ;) It\'s Sunday now...and the BBQ is ready...ciao! Peace...  '),(5197,1788355,'18,August,2003','2003-08-18 00:00:00','   At some point in the near future, it will be possible to know for  urlLink certain  whether or not someone is lying, fix depression, and essentially read their mind. The technology is edging closer and closer, to discovering which areas of the brain are active during certain thought processes, it is now only a matter of time.   Caplan gives a convincing argument as to why it is ethical for us to develop the tools to detect lying, and modify our brains and the way the work. Such modifications may enable us to learn at incredible paces. French in an hour, read faster, remember things perfectly. Who wouldn\'t want to learn a new language in an hour? Or read a book in minutes. Have perfect memory, or erase past traumas.   If we had a machine that could tell if someone was lying, one would have to define a lie. You couldn\'t really find the truth in such a machine, just what the person believes to be true; what actually happened versus, their perceptions. (That point came up in talking to JohnnyMac) Scientists who looked at brainwave outputs saw differences in activity in regions of the brain from lies and truths. Apparently it takes less work for the brain for one to tell the truth. Of course one could tell a lie with impunity if one would believe the lie they told, but then at that point, it ceases to become a lie.   But at what point do our enhancements stop? Just because we make ourselves better, faster, and smarter, should we? At what point do we become less human?  Earlier this year a Nasa project  urlLink surfaced  to setup \'mind readers\' at Airports to \"detect passengers who potentially might pose a threat.\" Do we trust the government to use such a power? One does not have to be paranoid to see such a power being abused by political powers. If Stalin had the ability to detect someone\'s \'intentions,\' do you not think he would have done it? What would a president do to protect a nation? Scan everyone to find the terrorists in our midst? We live in a time where the government sees everyone as a potential threat, a potential terrorist. Are we as a people responsible enough to handle such a power? I think not. Sounds like  urlLink 1984 ? So does many things that our  urlLink government  is  urlLink currently   urlLink doing .  At the point that it becomes possible to give people perfect memory, it will become possible to fabricate memories, or erase them. At the point one could fix brain defects and cure people of schizophrenia, you could give people these diseases. With the power to create, and fix, comes the power to destroy.   Is it worth opening this box? Do the benefits outweigh the dangers? That\'s something that will have to be debated. But certainly, just because we can do something, doesn\'t mean that we should.   I for one, do not wish to have a perfect memory. There is something about how memories fade... Often one remembers thing better than they actually were. Perhaps, there is bliss in my ignorance. I don\'t mind it, it is part of the human experience. Perhaps that is something that will change with time.  ~Ben   '),(5198,1788355,'15,August,2003','2003-08-15 00:00:00','   Now, as far as Video and Computer games go, I do not shy away from the violent ones. But  urlLink this  is just ridiculous. It does confirm my suspicians that the war was just entertainment for most Americans. I mean how many of us actually know someone in Iraq? Actually, I do. But for most of us there is a large degree of seperation between what happens to people in the news and what affects you. Our nation was at war for months, yet in America, you wouldn\'t be able to tell if it weren\'t for the newspapers or TV news.   I remember when Doom was called a  urlLink murder   urlLink simulator  - so I wonder what \"good-natured\" senators are going to have a hissy fit about this game? In Doom you fought monsters from hell, most far from human looking. All of those who talk about media bias (be it conservative, or liberal), you should take note. The news is a form of entertainment, the reality show of life. It is of course no more real than what producers choose to show you. My film teacher once told me: \"Television is a sequence of lies that you believe to be true.\"     Murder Simulator? psh...  What you see and hear on television and the radio, or read in the newspapers is carefully chosen with one thing in mind: someone will watch, read, listen or buy it. The news is a consumer product. This of course will not be news to anyone who\'s paid any attention to the world. People often blame the media for the ills of their society. They see the violence in a game their children are playing, or a move they saw. They then connect that to the violence they hear about on the news. Our movies are violent, cause our culture is violent. We are violent. No one needed to see a movie to learn how to kill someone, we\'ve been doing that thousands of years before we ever had televisions.  ~Ben   '),(5199,1788355,'15,August,2003','2003-08-15 00:00:00','   Hey, I suppose this is my first post. So to whomever will actually read this, hello. I wasn\'t sure if I was going to create this today, or ever, really. But since I couldn\'t sleep I figured I\'d do something semi-constructive. First was applying for jobs, second was creating this blog.   So for a start, a little about me: I am a 23 year old male who lives in the  urlLink Chicago  area. I will be attending  urlLink graduate school  this fall, but would prefer gainful employment. That, however, doesn\'t seem to be working for me. So for the moment, school it is. I have a B.S. in Computer Science, and that is exactly what it is: BS. I\'ve been working an extremely boring job devoid of anything intelligent to do at the  urlLink university  I graduated from. When you graduate from school, you\'re supposed to leave, and I still haven\'t. I\'m slightly ashamed of that.   I play guitar, write crappy songs, watch movies, and spend way too much time online in my freetime.   So that\'s a little about me. So what should you expect if you stop by? Mostly me bitching, probably. I cannot gurantee it will be entertaining, but at least it\'ll be something.  ~Ben   '),(5200,1788355,'26,September,2003','2003-09-26 00:00:00','   Looks like someone changed the image on me!     Ben   '),(5201,1788355,'18,September,2003','2003-09-18 00:00:00','   Wow...   '),(5202,1788355,'09,September,2003','2003-09-09 00:00:00','   Jesus: The Virgin Mary a \"Frigid Bitch\"    In the urlLink  diaries  Jesus calls the Virgin Mary a \"frigid bitch\", and describes his experiences of taking marijuana and LSD and having sex with Mary Magdelene.   Now this sounds like my kind of book. I do agree that god must have a sense of humor I mean, look at the platypus for chrisake.  There are way too many reactionary christians that get offended way too easily. First reaction of course is to ban the book! Ban ideas that are different than our own. Insanity. If you don\'t like the book don\'t read it, no one is forcing you.  In the mean time, at least admit it\'s funny...   '),(5203,1788355,'09,September,2003','2003-09-09 00:00:00','   Is David Blaine Insane? Probably....     I do like that picture though. So apparently this nutty \'magician\' (as if there were such a thing) decided to lock himself up in a glass box and suspend it over London. Why? No one knows?  Apparently urlLink  Londoner\'s  are doing everything from flashing their breasts at him, to throwing sausages and bacon.  Suppose he could fast himself for something more useful..like a cause?  But hey if Sky news were paying me to sit in a glass box for 44 days, I might do it too.   '),(5204,1788355,'09,September,2003','2003-09-09 00:00:00','   So what happens when mutli-billion dollar corporate trade entities don\'t have enough money? They  urlLink sue  12 year old girls for more. The industry continues to blame \'internet piracy\' for their declining sales, while not looking at the fact that just maybe their music just really sucks, and that\'s why no one buys it.   As usual,  urlLink The Register  sums it up best:  \"Brianna thought the $29.99 fee her mother paid for the Kazaa music trading service entitled her to download songs at will. Nothing like a lawsuit seeking $150,000 per song to correct that misconception. This is what some refer to as a growing pain.   This settlement has taught us a few valuable lessons about the RIAA\'s methodology. Apparently, young teens hit the copyright infringement scale at the $2,000 mark. College students, by contrast, must cough up between $12,000 and $17,000 for their violations, as we saw earlier this year. So any parents out there with children under 12 can expect their precious tots\' crimes to cost around $1,000. That\'s comforting.\"   '),(5205,1788355,'03,September,2003','2003-09-03 00:00:00','   Hey, you want something else to piss you off? Read  urlLink this.  I\'m really starting to wonder what Northernstar won\'t print. I\'ve got a good mind to write a response, but that requires time and effort.   '),(5206,1788355,'03,September,2003','2003-09-03 00:00:00','   urlLink Liberals need to lose the hypocrisy and realize old traditions are still important     I would like to complement Nick Arhos in his recent column written in regards to liberals and their ideas.   His article really captures the overpopulation of the liberal presence on campus. I believe he is correct in his statement, â€œLiberals also refuse to acknowledge the overwhelming evidence that the United States is a Christian nation founded upon laws derived from the Holy Bible and the Christian faith.â€   The Alabama Supreme Court situation makes me sick. Our laws, that the Supreme Court enforces, are based on the Ten Commandments (thou shalt not kill, thou shalt not steal, etc.). Those tablets have a place in the courts.   I also would like to voice my approval in his statement exploring the hypocrisy in many liberalsâ€™ beliefs. I am tired of complaints about SUVs and gas costs. Unless these individuals are riding a bike or walking, they have no place in harassing vehicle drivers.   Do I classify myself as a conservative or Republican? No, I consider myself a citizen of the United States of America: In God we trust.   Joshua Huseman  Junior, communication   Normally, I do not like to participate in the name-calling and misinformation that is modern political discourse. However, the complete lack of historical knowledge of this nation by Joshua and Nick is amazing. Firstly, if the United States is a christian nation, what denomination is it? Christians among themselves cannot decide on a set of beliefs to all agree on, so it seems unlikely one denomation would be favored over another. Then there is this document that the nation was founded upon, the constitution. The constution doesn\'t even make a mention of god, or christianity, even in general. Although it does mention religion, right in the first amendment: \"Congress shall make no law respecting an establishment of religion, or prohibiting the free exercise thereof;.\"  To argue that our nations laws are based on the 10 commandments is nonsense. Firstly, I don\'t remember there being a federal law forbidding from using God\'s name in vain. Further, it is not illegal for me to worship or create an icon.   more later...  Ben   '),(5207,1788355,'31,March,2004','2004-03-31 00:00:00','   empty-armed and half a soul to go  and all i wanted was you here next to me a little sunshine and sympathy  now everybody knows that i\'ve been hanging down so low \'cause now i\'m feeling up soon i\'ll be feeling out so cold wondering, will you call and now i\'m feeling high soon i\'ll be feeling left for dead sometimes someone saying yes changes what you\'ll bet  and all i wanted was just to hold you close a little sunshine just to butter my toast and your love next to mine  i have to let you know that we were meant to be just right heaven sent, not sympathized by everybody\'s lie and now i\'m feeling high now i\'m feeling left so dead kicking up the dust in bed wondering, i guess   Zwan( urlLink Billy Corgan ) - El Sol   '),(5208,1788355,'30,March,2004','2004-03-30 00:00:00','   Now this article is just...sad:   urlLink http://www.reason.com/hod/nr032604.shtml   '),(5209,1788355,'30,March,2004','2004-03-30 00:00:00','   Heard this new song just the other day and it summed up the moment perfectly, i like it a lot.  The Postal Service - Such Great Heights:   I am thinking it\'s a sign that the freckles In our eyes are mirror images and when We kiss they\'re perfectly aligned And I have to speculate that God himself Did make us into corresponding shapes like Puzzle pieces from the clay True, it may seem like a stretch, but Its thoughts like this that catch my troubled Head when you\'re away when I am missing you to death When you are out there on the road for Several weeks of shows and when you scan The radio, I hope this song will guide you home  They will see us waving from such great Heights, \'come down now,\' they\'ll say But everything looks perfect from far away, \'come down now,\' but we\'ll stay...  I tried my best to leave this all on your Machine but the persistent beat it sounded Thin upon listening That frankly will not fly. you will hear The shrillest highs and lowest lows with The windows down when this is guiding you home   '),(5210,1788355,'29,March,2004','2004-03-29 00:00:00','   My life has taken a turn towards the surreal   '),(5211,1788355,'24,March,2004','2004-03-24 00:00:00','   Mar\'s had a  urlLink sea  (of water!)!  I\'m not a person that deals with change well. And for a person that doesn\'t deal with change now, there have been a lot of changes in my life in the past year. I guess I\'m not sure what I fear, if anything, other than change itself. I get comfortable in my routines - to a fault. But I actually get used to new routines rather easily, just makes me uncomfortable.   Anyhow.  Places I want to go to:     urlLink Japan  (Tokyo! I think I\'ve always wanted to go to japan..) My studying in Japan kind of got put on hold since I got a job,  I still plan on doing it however, so Perhaps I should not get used to an extravagant lifestyle!    New Zealand (Looks beautiful and interesting, and more fun that Australia!)    India (Taj Mahal! Calcutta, Bombay, Delhi)    Spain(a friend keeps telling me how great it was there, over and over, and over... ;))    Egypt(Pyramids of course... and the Sphinx)    Ireland(who doesn\'t wnat to go to ireland?)    England (Big ben, need i say more?)    China (Great wall, Hong Kong, Bejing..)   South Africa Morroco Argentina Mexico Florida! Alaska Norway Lithuania Poland Germany France Australia   '),(5212,1788355,'22,March,2004','2004-03-22 00:00:00','   Bush pushed to Bomb Iraq after 9/11    After the president returned to the White House on Sept. 11, he and his top advisers, including Clarke, began holding meetings about how to respond and retaliate. As Clarke writes in his book, he expected the administration to focus its military response on Osama bin Laden and al Qaeda. He says he was surprised that the talk quickly turned to Iraq.   \"Rumsfeld was saying that we needed to bomb Iraq,\" Clarke said to Stahl. \"And we all said ... no, no. Al-Qaeda is in Afghanistan. We need to bomb Afghanistan. And Rumsfeld said there aren\'t any good targets in Afghanistan. And there are lots of good targets in Iraq. I said, \'Well, there are lots of good targets in lots of places, but Iraq had nothing to do with it.   \"Initially, I thought when he said, \'There aren\'t enough targets in-- in Afghanistan,\' I thought he was joking.   \"I think they wanted to believe that there was a connection, but the CIA was sitting there, the FBI was sitting there, I was sitting there saying we\'ve looked at this issue for years. For years we\'ve looked and there\'s just no connection.\"   Clarke says he and CIA Director George Tenet told that to Rumsfeld, Secretary of State Colin Powell, and Attorney General John Ashcroft.   Clarke then tells Stahl of being pressured by Mr. Bush.   \"The president dragged me into a room with a couple of other people, shut the door, and said, \'I want you to find whether Iraq did this.\' Now he never said, \'Make it up.\' But the entire conversation left me in absolutely no doubt that George Bush wanted me to come back with a report that said Iraq did this.   \"I said, \'Mr. President. We\'ve done this before. We have been looking at this. We looked at it with an open mind. There\'s no connection.\'     urlLink http://www.cbsnews.com/stories/2004/03/19/60minutes/main607356.shtml   Unbelievable.   '),(5213,1788355,'20,March,2004','2004-03-20 00:00:00','   Violence begets Violence.   Something I learned while I was still young. Something that Israel, or the Palestinians have learned. Terrorism is generally regarded as a dispicable act. Strap a bomb to yourself and take out as many of the \"enemy\" as possible - often times civilians. But all warfare is horrible. To condemn assymetrical terrorism as a crime against humanity, and yet not \"conventional\" warfare, seems downright hypocritical. Does not the seemingly random bombing and missleing of Palestinians terrify them? Is it not itself a form of terrorism? To make the people of Palestine - those who are fighting - submit to the wishes of the Israelis? Isn\'t that the definition of warfare - make one side submit by force. It would seem that the end\'s of the would-be terrorists aren\'t irrrational - they just want their homes. Some of the more radical groups want the destruction of Israel - but that seems to be the fringe groups. Most of the Palestinians want what every person on this planet wants, peace, and a place to live.    What\'s left of the Hamas leader. What was widely unreported was that he was wheelchair bound...  This situation can most easily be put upon the British, amoung other nations (however it was the British who occupied the area which is called Palestine and Israel now), that decided to make a homeland for the Jewish people displaced by Hitler and his indescrible act of inhumanity. The British gave up, and handed it over to the newly formed United Nations to handle. The land was divided and Israel was formed, and it was immediately at war with several Arab countries. Israel ended up conquering land from Palestine, and Egypt too.  Why the United States decides to take sides in the conflict (the Israelis) is unknown to me other than political reason - to have a strong ally in the middle east which doesn\'t like the US much (and the irony is lost on me, one of the reasons they don\'t the US is because of our support for Israel).   The Palestinians are fighting back in the only way they can. They have no armies, no tanks, no aircraft. The Israelis have American tanks, american weapons, american aircraft. This military support is fueling as what is seen as a largely one sided conflict. The Israelis moving deeper into Palestinian territory, and bulldozing homes; the palestinians fighting back with suicide bombings, and the Israelis retaliating.  The palestinians won\'t stop their bombings until the Israelis pull back, the Israelis won\'t stop retalliating (by shooting missles into crowds of people to get \"terrorist leaders\"!?), or pull out until the palestinians stop bombing. So its a vicious circle of violence, with no end in sight.    urlLink US running scared...   '),(5214,1788355,'20,March,2004','2004-03-20 00:00:00','   One sign that you may be doing something that is wrong - world wide protests!    Hong Kong.  I remember being educated all my life about how great the United States of America is. \"We\'re the greatest nation in the world!\" Why? Is it because of all the nations we\'ve \"freed\"?   One of the factors leading to radical terrorists is our propensity to overthrow governments we don\'t like. This is why I was against going to war with Iraq. We were playing right into the hands of Al Queda. The suffering of the Iraqi people was great, but no greater than what the United States has incured against other peoples by supporting even more terrible dictatorships.     Kashmir.  I attended a talk about the morality of the United States invasion of Iraq at Benedictine University. It was the typical catholic \"just-war\" theory stuff. But there was one very interesting moment. The only Iraqi to attend - someone from the audience stood up and said (during the q & a session): \"I find it absolutely hypocritical that you would all talk about the evil regime of Saddam Hussein, all the while it was the you who had put him into power, supported him, gave him the weapons of mass destruction. Then you sit here and debate the human rights issues of government.\"   I am paraphrasing of course. But this put the whole room into shock. And I am glad he said it. It is hypocritical. Our nation is warring contradiction. Freedom for all, as long as you\'re american, or have oil, and only if we can\'t get along with you\'re autocratical government or replace it easily.    The Philippines  Now, you may think I\'m a looney here. What other nations has the United States government overthrown? I haven\'t read about this in history class? Well, no you haven\'t. Or rather it\'s there - between the lines.    Turkey.  Here is a short list (made short as I\'ll only list nations after WWII ended):  Iran (overthrew their democratically elected government no less, because their president was going to nationalize their Oil company! OH NO Socialism!!!) South Korea Guatemala Vietnam Cambodia Zaire Brazil Dominican Republic Haiti Cuba Indonesia Chile Greece Nicaragua Greneda Libya Panama Iraq Afganistan Yugoslavia  this isn\'t even a full list.   One suggestion for the future. STOP. MESSING. WITH. OTHER. PEOPLE\'S. GOVERNMENTS. That pisses them off!   I think we can eventually restore respect to this nation, may take a long time though.    London.   We are, along with the world, are the victims of our own foriegn policy.   '),(5215,1788355,'18,March,2004','2004-03-18 00:00:00','   Bleu Bleu, something new for you.  I gots me a new a job. So long to my old place of not so gainful  urlLink employment . About all I\'ll miss is the people, really. Lots of memories there. But time to create new memories in new places.   I was thinking today, that I think I will be a teacher - at least for a little bit. Watching the professor that (as of this second)  I still currently TA for, completely lose the class over and over again - made me really want to take over and teach the class.  My current plan, finish my masters. Go teach somewhere and get my PHD. Marry some great girl (perhaps Jewel will be availible?) Work for Nasa, go to Saturn, then die a happy man with 10 kids.   '),(5216,1788355,'17,March,2004','2004-03-17 00:00:00','   urlLink Donald (foot-in-mouth-syndrome) Rumsfield   That\'s quite funny actually.   About  urlLink curious George W ... It seems the latest campaign slogan is: \" urlLink A vote for John Kerry is a vote for Osama bin Laden .\"   My  urlLink nation  is progressing well - although I\'m not sure how meat eating is discouraged...oh well.  Happy Saint Patricks day! For those of you reading that aren\'t from Chicago, I\'ll let you in on a little odd tradition. Every St. Patty\'s day,  urlLink The City of Chicago dyes the Chicago River green  - its quite fun to see actually.  This will be the first time I celebrate it realizing I\'m a whole quarter Irish. Yaayyyyy.   Another note: Doctors suck major ass.  I have kidney stones. Oh wait, no I don\'t. Yes I do, No i don\'t. So back and forth they went.  I give up, just drown my sorrows in pain meds, i suppose!   '),(5217,1788355,'25,April,2004','2004-04-25 00:00:00','   guess where i am!!  knock one of those places i wanted to go off my list...   '),(5218,1788355,'09,April,2004','2004-04-09 00:00:00','   random pictures    U.S. Marines pray over a fallen comrade at a first aid point after he died from wounds suffered in fighting in Fallujah, Iraq (news - web sites), Thursday, April 8, 2004. Hundreds of U.S. Marines have been fighting insurgents in several neighborhoods in the western Iraqi city of Fallujah in order to regain control of the city. (AP Photo/Murad Sezer)     Granada    the kiss by Gustav Klimt      ahhh sailing...so much fun.   '),(5219,1788355,'08,April,2004','2004-04-08 00:00:00','   i can\'t put into words the beating and relentless thoughts of my head that constantly bombard all that i am and if i could write all i thought or knew im pretty sure i\'d either be damn smart or completely insane. most likely the latter... i know that for about a day 4 months ago i actually thought i would have everything i ever wanted, but it slipped away, like most things.  as Eddie says \"hearts and thoughts, they fade, fade away...\" I guess nothing is permanent, but I\'d think that at least I\'d be happy for more than a day. But reality then sets in. and then the possibility of that i may never have what i really want out of life. Who am I to expect that anyhow...   '),(5220,1788355,'06,April,2004','2004-04-06 00:00:00','   it\'s hard to say what i would\'ve done if i hadn\'t met you but somehow i know it wouldn\'t be the same, now that i know you i walk the same streets  i always have, barren and without light but i find my way despite my blindness and i end up here with you and i wonder what you\'d do if we lived in a different world i have to learn to stop thinking in permanence but it feels like the sadness of life won\'t ever end and maybe you\'ll find in your heart what i always felt was there but nothing to count on except the empty roads to my soul and my life but i\'ll find the star and you won\'t be far behind.   '),(5221,1788355,'06,April,2004','2004-04-06 00:00:00','   always too early or too late. fate.  never enough, or too much for you. just blue.  and i find everything, and nothing here. more tears.  i am blind, and see it all fall.  i drown and swim around never found.   '),(5222,1788355,'06,April,2004','2004-04-06 00:00:00','   i think i should really see a doctor, the last two days the pain in back has gotten a lot worse, i couldn\'t even get up out of bed the first time.  funny how my life is - so strange. Things I thought that would never happen - that happen. Things that I never thought of that come to life. It\'s an odd, odd life.  Suppose anything is possible as long as you don\'t rule out that possibility.   '),(5223,1788355,'04,April,2004','2004-04-04 00:00:00','   i don\'t know kids, whats wrong with me.  so i finally move out of the house, but for some reason miss it. it was too late to drive back to my apartment last night, so i stayed home, one last sleep in my bed eh?  I have a new one though....much better than my old one. so perhaps i\'ll delay judgement until then.  that old beaten up room still holds memories for me though, and its hard to leave it on one level.  strange life, strange days, strange air  now, my recurring back pain is getting much worse. some days i feel completely debilitated. probably moving all my stuff yesterday didnt help anything either.   '),(5224,1788355,'27,May,2004','2004-05-27 00:00:00','   insanity. absolute insanity.  Want to stop viruses?  urlLink Execute  the writers. That\'s write exectue someone who writes a virus. That\'s the most ridiculous thing I\'ve ever heard. Now you will go to  urlLink jail  for up to 3 years in italy for \"downloading and uploading copyrighted content.\"   This is insane. People don\'t understand technology, the record and movie lobbyists are scared to death of it. Instead of changing their business to fit the new economy they want to  urlLink sue  people back into the old one.  A single mother sued for $540,000 - because her daughter downloaded some songs off of the internet is ABSOLUTE INSANITY. That is the worst abuse of corporate power I\'ve seen in a long time.  Crime comparison:  Speeding ticket: 95$ (for speeding 95 in a 65 zone, I know I did this). Downloading one song in the United States and sharing it: 750$.  Let me ask you, what is the more dangerous activity?  What should our concern be? Why are we wasting time with this? We have terrorists that want to kill us, half the world angry with us, and we\'re wasting our time tracking down \"music pirates\"?? GIVE ME A FREAKING BREAK.   some days i wonder about this country.   '),(5225,1788355,'20,May,2004','2004-05-20 00:00:00','   I now have a gmail account. Apparently, that\'s a  urlLink big deal . Now, if I knew I could sell it....  Well I gave my brother and Samreen the accounts. So lucky them. I have bestowed them great riches, or something.  It is pretty nice though. I like the way the page is setup - nice and simple. Its faster than yahoo or hotmail, and uncluttered. It organizes the email into \"conversations,\" sort of like a messageboard.  Privacy \"advocates\" are having a field day - but they just prove they don\'t understand the technology. A computer looks for terms in your email and gives you relavent text ads. Though, I have yet to see a \"relavent\" text ad (pictures later..). That\'s it. Nothing sinister.   And you know what? You don\'t have to use it.  adios for now   '),(5226,1788355,'18,May,2004','2004-05-18 00:00:00','   i guess i\'m officially important now. I now get woken up in the morning whenever anything breaks, or looks broken. And thanks to the idiots who came before me, things break plenty.   So what to do? Fix it, I suppose. But I thought I was hired to program not babysit servers...   '),(5227,1788355,'16,May,2004','2004-05-16 00:00:00','   I don\'t even have to wait 10 mins for a social engineering example!   Here is an email I recieved:    Dear U.S. Bank valued member,   Due to concerns, for the safety and integrity of the online    banking community we have issued this warning message.     It has come to our attention that your account information needs    to be updated due to inactive members, frauds and spoof reports.    If you could please take 5-10 minutes out of your online experience and renew    your records you will not run into any future problems with the online service.  However, failure to update your records will result in account suspension.  This notification expires on May 20, 2004.     Once you have updated your account records your internet banking    service will not be interrupted and will continue as normal.     Please follow the link below    and renew your account information.     urlLink U.S. Bank Internet Banking        Hmm? I don\'t ever remember having a US Bank account? Is there such a thing? Sounds a little generic!  Now if we look at the source you see where the link really goes:  a href=\"http://203.131.92.146/www/\" U.S. Bank Internet Banking   It seems, however, someone caught this and has disabled the website. But how many people went to that website and gave him their bank information?  I wonder..   '),(5228,1788355,'14,May,2004','2004-05-14 00:00:00','   So what am i doing up at 6 am on a sunday morning?  to go to work. Some days I like my job, others - not so much.  So I thought perhaps I would bore you with some random thoughts, whoever it is that reads this.  A  urlLink flying car  is less of a gas guzzler than an SUV. Holy carp. (holy fish?)  Some of you may know, I am deeply interested in security, especially in the work place. I think I\'ve read 5 or 6 books on the topic (starting my 7th and 8th). I don\'t think many people actually realize that the biggest security weakness are the people. If you call up and ask someone for information 9 times out of 10 they will give it to you. All you have to do is claim authority.  Any some good and interesting reads are:   urlLink Social Engineering Part 1   urlLink Social Engineering Part 2   '),(5229,1788355,'14,May,2004','2004-05-14 00:00:00','   Have $200,000 and some time? Create a new virus. Or a cure for aids..  urlLink  http://www.wired.com/news/medtech/0,1286,63441,00.html?tw=wn_tophead_1     BERKELEY, California -- It took Adam Arkin and David Schaffer just $200,000 and a grad student to develop a potential treatment for AIDS. And that scares them.  That\'s because the therapy itself is a virus. The Lawrence Berkeley National Laboratory assistant professors created a virus altered to latch onto HIV and mute its ability to become AIDS. They\'ve tested the theory in a computer model and in cells in a dish. The results have been promising, and if they continue in that vein, the researchers could begin animal testing by the end of this year.  Arkin said this week at the International Biotech Summit at the University of California at Berkeley that it was almost too easy for him and his colleagues (Schaffer and then-grad student Leor Weinberger) to build the anti-HIV virus.  \"If I can do it, anyone can do it,\" Arkin said.  \"That\'s going to be a problem.\"     Understatement of the year?   '),(5230,1788355,'13,May,2004','2004-05-13 00:00:00','   Odd. I don\'t post for a whole 10 days and WHAM the entire interface is different. Maybe I can change how my rather generic website looks.  Anyhow.  What Won\'t Happen but should: John McCain joining John Kerry and running for president. Vice-president is a rather useless role, but one that has been redefined by Dick Cheney. (Who really runs, or rather ruins this country). Bush admitting this war was an expensive mistake, in every sense of the word.   A good read:  urlLink http://www.indystar.com/articles/0/146292-3880-021.html   A note about Torture.  It seems some people aren\'t shocked by the torture of Iraqi prisoners. I can\'t believe that. We\'re supposed to be above that. Once we compromise our principles to fight the terrorists, we\'ve lost. Not to mention WHY THE HELL ARE WE IN IRAQ? How many reasons have we be given? How many more will be fed? Why do we still accept them? It was a nice political spin that the talk of the missing weapons of mass destruction has completely left public debate. Now we\'re their to liberate the iraqi people - or are we there to torture them? So how are we bringing democracy to the middle east? By raping and torturing their people? There seems to be little mention that the prisoners that are at that prison are largely held there for no other reason that \"looking suspicous.\"    urlLink http://news.scotsman.com/latest.cfm?id=2915028   War brings out the inhumanity in people. You can\'t establish a democracy for another people by force. It is something they must do themselves, it is something that we did ourselves (with a little help of the oft-forgotten french).  It will probably be a long and violent process - most revolutions are. But it\'ll be better than americans being over killing them, and \"freeing\" them (of their lives..).   '),(5231,1788355,'28,June,2004','2004-06-28 00:00:00','   so there was an earthquake this morning. How strange.  http://earthquake.usgs.gov/recenteqsUS/Quakes/uskgad.htm  I felt the building swaying, it was quite strange. I thought some lovers were getting a little into it upstairs (happened before), but then the swaying got stronger. I thought it strange, but since i\'ve lived in a wooden house that swayed a lot before I lived in the apartment building, I didn\'t think much of it.  But then, John Im\'d me.  [01:11:27] John: um [01:11:31] John: the ground just shook [01:11:37] me: you serious [01:11:41] me: it did here [01:11:43] me: wtf  [01:11:49] John: what the hell [01:11:50] me: i swear to god it just did here [01:11:59] John: it was like an earthquake [01:12:04] John: hehe [01:12:04] me: weird [01:12:10] me: or a nuke [01:12:16] John: ever been in an earthquake? [01:12:22] John: our computers wouldnt be working [01:12:40] John: nor would our eyes, skin, vital organs, peripheral organs...  Then Matt called me and told me the ground shook by him in Dekalb. It was quite freaky.  Anyhow, 4.5 earthquake is pretty good, but mild.   '),(5232,1788355,'22,June,2004','2004-06-22 00:00:00','   Some of you asked what does \"nani betsuni\" mean? Well, it\'s japanese for \"what? nothing!,\" well roughly, anyhow. It\'s for that annoying interaction with certain people, when you don\'t hear something they said and they say \"nothing!,\" then you get really curious what it was they just said, since they just won\'t say it again.  I went to Great America yesterday. It decided to rain most of the morning though, and I got absolutely soaked. But we recovered, and went again later, with no rain, and waiting in line. I\'ve never been on that many roller coasters in a row that quickly before. I was dizzy, and felt a little drunk afterwards. It\'s a bit strange, because I dont remember most of what happened, it\'s all a haze.  I have lots of memories there...strange ones, nice ones, and bad ones. They were pretty much haunting me all of yesterday. It\'s funny how sometimes you can\'t get certain things or people out of your mind, i just keep replaying the moments, and i really dont want to remember or forget at the same time. That\'s been the last 4 years. In one sense i want to forget them in another I want to remember them.  every couple of months my dreams change, well some of them. Others stay the same, and I can\'t shake them. I keep having this recurring dream of the past, then waking up, realizing the moment is gone, and lost forever. So I have to live through all the sadness again, and I\'m not sure why my brain is relentless in its self-torture.  more and more i just come to the conclusion i live in a fantasy world most of the time, and when i come back to the real one, i get really sad, then fall back to my fantasy that will in reality never be fufilled.  i never much liked the real world anyhow   '),(5233,1788355,'08,June,2004','2004-06-08 00:00:00','   The creator of the algorithm, and my profession:   Al-Khwarizmi From Wikipedia, the free encyclopedia.  Abu Abdullah Muhammad bin Musa al-Khwarizmi (Ø£Ø¨Ùˆ Ø¹Ø¨Ø¯ Ø§Ù„Ù„Ù‡ Ù…Ø­Ù…Ø¯ Ø¨Ù† Ù…ÙˆØ³Ù‰ Ø§Ù„Ø®ÙˆØ§Ø±Ø²Ù…ÙŠ in Arabic), also called Muhammad ibn-Musa al-Khwarizmi, Muhammad ibn-Musa al-Khowarizmi, and Mohammad Bin Musa Al-Khawarizmi, (flourished early 9th century), was a Persian scientist, mathematician, and author. He may have been born in 780, or around 800; he may have died in 845, or around 840.  He was born in the town of Khwarizm (now Khiva), in Khorasan province of Persia (now in Uzbekistan). The name al-Khwarizmi means the person from Khwarizm. His family moved soon afterward, to a place near Baghdad, where he accomplished most of his work in the period between 813 and 833. There are various guesses at his native languages, including Persian or more probably Khwarizmian (now dead). All of Al-Khwarizimi\'s works were written in Arabic.  He developed the concept of the algorithm in mathematics (which is a reason for his being called the grandfather of computer science by some people), and the words \"algorithm\" and \"algorism\" come from Latin and English corruptions of his name. He also made major contributions to the fields of algebra, trigonometry, astronomy, geography and cartography. His systematic and logical approach to solving linear and quadratic equations gave shape to the discipline of algebra, a word that is derived from the name of his 830 book on the subject, Hisab al-jabr wa al-muqabala (Ø­Ø³Ø§Ø¨ Ø§Ù„Ø¬Ø¨Ø± Ùˆ Ø§Ù„Ù…Ù‚Ø§Ø¨Ù„Ø© in Arabic).  While his major contributions were the result of original research, he also did much to synthesize the existing knowledge in these fields from Greek, Indian, and other sources. He appropriated the place-marker symbol of zero, which originated in India, and he is also responsible for the use of Arabic numerals in mathematics.  Al-Khwarizmi systematized and corrected Ptolemy\'s research in geography, using his own original findings. He supervised the work of 70 geographers to create a map of the then \"known world\". When his work became known in Europe through Latin translations, his influence made an indelible mark on the development of science in the West: His algebra book introduced that discipline to Europe and became the standard mathematical text at European universities until the 16th century. He also wrote on mechanical devices like the clock, astrolabe, and sundial. His other contributions include tables that included trigonometric functions, refinements in the geometric representation of conic sections, and aspects of the calculus of two errors     urlLink http://en.wikipedia.org/wiki/Al-Khwarizmi    P.S. wikipedia rules.   '),(5234,1788355,'27,July,2004','2004-07-27 00:00:00','   We are as the flute, and the music in us is from thee;         we are as the mountain and the echo in us is from thee.             We are as pieces of chess engaged in victory and defeat:         our victory and defeat is from thee, O thou whose qualities are comely!             Who are we, O Thou soul of our souls,         that we should remain in being beside thee?             We and our existences are really non-existence;         thou art the absolute Being which manifests the perishable.             We all are lions, but lions on a banner:         because of the wind they are rushing onward from moment to moment.             Their onward rush is visible, and the wind is unseen:         may that which is unseen not fail from us!             Our wind whereby we are moved and our being are of thy gift;         our whole existence is from thy bringing into being.    -  urlLink Rumi ,  urlLink Sufi  Poet About Sufism, which somewhat mirrors my beliefs...  Sufis believe that their teachings are the essence of every religion, and indeed of the evolution of humanity as a whole. The central concept in Sufism is \" urlLink love \".  urlLink Dervishes  -- the name given to initiates of sufi orders -- believe that love is a projection of the essence of  urlLink God  to the  urlLink universe . God desires to recognize  urlLink beauty , and as if one looks at a mirror to see oneself, God \"looks\" at itself within the dynamics of nature. Since everything is a reflection of God, the school of Sufism practices to see the beauty inside the apparent ugly, and to open arms even to the most evil one. This infinite tolerance is expressed in the most beautiful way, perhaps, by the famous Sufi philosopher and poet  urlLink Mevlana  (also known as    urlLink Rumi ) :  \"Come, come, whoever you are. Worshiper, Wanderer, Lover of Leaving; ours is not a caravan of despair. Though you have broken your vows a thousand times...Come, come again, Come.\"   '),(5235,1788355,'26,July,2004','2004-07-26 00:00:00','   i won\'t find another way to the top  and i\'m falling down, down, down  all that i know will leave me,  and all that i\'ve seen keeps me  and as fade away i breathe to see you  in my hands i hold another fleeting thought  of all that i was  and my vision blurs and fades away, darkening  and i miss it all, especially your moments  memories fade, but the feelings remain  i feel your warm breath on me  i reach out my hand   '),(5236,1788355,'25,July,2004','2004-07-25 00:00:00','   I think a society totally motivated on greed, consumerism,  and materialism is a sick society.   I don\'t think that there can be a sufficient advancement in society until people generally care for each other. Despite all the great things about this country one its greatest assets - individualism - is one of greatests downfalls as well.   Because we are an individualistic society, there is an apathetic attitude among the general populace for anything but their individual lives. Hence the lack of participation in political systems of this nation. I think that most of us would be surprised if you asked the general population if they beleive that health care, and other social programs should be provided by the government - i think you\'d find that answer would be yes. Yes, That is complete conjecture - but not unfounded, I am sure I could find a survey to back that up.        Ashcroft, who has performed an unprecedented power-grab as attorney general of the United States.    The part that disturbs me about the Bush administration is their psychotic, and odd politics. They are no longer for less government (as government has grown dramatically under the Bush administration), they continually attempt to claim more and more executive power (although this has been a general presidential trend), and are for further government interference in the general lives of the people (the amendment to outlaw gay marriage, institutionalized discrimination). The only thing they have done remotely conservative is - ramp up the nationalism of this nation and cut taxes. The rest have been very un-conservative moves.   I\'m for less government involvement generally, and especially in areas that will hurt the general populace. I believe that a national or state based health care system is badly needed. I believe that we should leave other nations to self-determination, with the exception of actual self-defense. The interventionalism of the past 50 years has created more problems than it has solved.   And most importantly - there is no one way to solve a problem. Not all liberal ideas are correct, not all conservative ideas are correct. I think this is a point most of you are missing, there are multiple ways to solve a problem - because that is generally what we\'re trying to do - solve problems.   These are more or less my observations.   '),(5237,1788355,'22,July,2004','2004-07-22 00:00:00','   urlLink A Good Read or rather depressing.                     So I might be a terrorist, but Iâ€™m not completely sure; Iâ€™m still waiting for the government to let me know. Like many Orange Countiansâ€”150,000, give or takeâ€”I fit the description of a possible terrorist. Whatâ€™s that exactly? Does she walk around with a .38 Smith & Wesson, slaying heathenish infidels? Nope. Does her normal wardrobe include an \"I [heart] Osama\" T-shirt? No way; I tossed that out with my \"I [heart] the New Kids on the Block\" shirt ages ago. Does she conveniently create fictitious wars around election years? Not me.        I donâ€™t feel like a terrorist, but who am I to question the sophisticated methods used in our nationâ€™s counterterrorism agencies, as infallible as our nationâ€™s leaders. With new, cutting-edge advancements in technology, todayâ€™s terrorists can be easily tracked down based on two factors: their last name and their country of origin. Forgive me, Big Brother, for I, having the last name of \"Afghani\" and my father being from, well, Afghanistan have thusly sinned twice. My being Muslim is simply an added bonus.        My family was contacted last August via post cardâ€”I suppose our Morse Code machine must have been down at the time. The piss-colored post card was addressed from Detective Boyd of the Los Angeles Police Department and asked us to \"PLEASE CONTACT.\"         My father, the original Afghan in the family, had the honor of placing the call to the Anti-Terrorist Division at the LAPD. The conversation between my father and the detective basically resulted in the understanding that we, the Afghani family, were on some sort of a list that their highly developed computers had created. The detective wanted us to set up a meeting with him sometime in the near future. Where? Well, it seems the preferred meeting place for police and enemies of the state can now be found at your local coffee shopâ€”doughnuts were not discussed.             urlLink Read the rest here   '),(5238,1788355,'12,July,2004','2004-07-12 00:00:00','   one step further to a dictatorship?  Tom Ridge wants to  urlLink suspend the election  if there is a terror attack.   The better solution would be to make the election several days, instead of just one. That way an attack one day wouldn\'t make sense - it wouldn\'t be as disruptive. Not to mention that the election shouldn\'t just be held on one WEEKDAY. (Making it difficult for most normal working people to make it to the polls).  Ben   '),(5239,1788355,'06,July,2004','2004-07-06 00:00:00','   the media is like a single mind. When one newspaper writes something, the next one has to write something about it too. So when you check the various news papers/TV/websites you find they all have the same stories - details may be different but the news is always the same. They all seem to miss the same things, then they all seem to make the same mistakes.   urlLink For Example :  What most (nearly all) people fail to realize is that sending email is more analogous to sending a post card than it is to sending a letter. Since all emails are sent through the network and internet in clear text and can be easily read. If you don\'t want people reading your email - you should encrpyt it.   Not to say that I am in favor of government wiretaps of internet connections. One just has to be pro-active about security.   '),(5240,1788355,'05,July,2004','2004-07-05 00:00:00','   A really good  urlLink article .  I\'ve been trying to make this point to people for a while:  We\'re better than torture. We know better. We have laws - a supreme law - the constitution.  If we give up our values, to fight a war - then we\'ve lost. If we torture someone to get information, we\'ve sunk to the level of the terrorists - we\'ve become the terrorists.   How do you win a war against an enemy that knows no boundaries, that recruits a new member everyday, who doesn\'t have any problem beheading innocent people?    You be better than them. You cannot every fully defeat idealogy, but you can make it obsolete.   '),(5241,1788355,'01,July,2004','2004-07-01 00:00:00','   whatamigoingtodoblue? if only she knew... dancin\' fingers and sprinkly kisses whatamigoingtodoblue? if only i knew... flash a smile and rub your nose whatamigoingtodoblue? the minutes were too few...   '),(5242,1788355,'11,August,2004','2004-08-11 00:00:00','   I\'m a fan of  urlLink Lawrence Lessig , who is a law professor at  urlLink Stanford Law School  (which is where I would go, when I go to law school. I think I\'d like to do that one day...). He used to teach at the  urlLink University of Chicago Law School , with  urlLink Barak Obama  (someone else I\'m a fan of).  Lawrence has an interesting  urlLink article  in  urlLink Wired  (my favorite magazine to read):   In August, Robert Greenwald will release an updated version of his award-winning film,  Uncovered: The Whole Truth About the Iraq War . Greenwald has added a clip of President George W. Bush\'s February interview with Tim Russert on  Meet the Press , NBC\'s Sunday morning talk show. In the clip, the president defends his decision to go to war - astonishingly unconvincingly.  Greenwald asked NBC for permission to run the one-minute clip - offering to pay for the right, as he had done for every other clip that appears in the film. NBC said no. The network explained to his agent that the clip is \"not very flattering to the president.\" Greenwald included it anyway.      and       This question is crucial, and thus Greenwald has decided to defend his fair use right, even if it means staring down a bunch of lawyers in court. The argument: It\'s hard to tell \"the whole truth\" about the Iraq war when you censor bits of that truth because a network tells you to. But what this incident demonstrates most is what many increasingly fear. Concentrated media and expansive copyright are the perfect storm not just for stifling debate but, increasingly, for weakening democracy as well.     This does mark a disturbing trend...    And one of the reasons why i like Obama:   \"There is not a Black America and a White America and Latino America and Asian America - there\'s the United States of America.\"  - Barak Obama at the Democratic National Convention.   '),(5243,1788355,'10,August,2004','2004-08-10 00:00:00','   NASA Gives OK to Fix Hubble Telescope'),(5244,1788355,'09,August,2004','2004-08-09 00:00:00','   here i am heading to san francisco, with another person named erik( urlLink eric ). This time I have a feeling it will be less fun, however. At least I won\'t accidently be on a flight to Alaska this time. Which, by the way, isn\'t on the way to SF.  Sometimes, I think it\'s worse not to know, than to think about the possibilities. But I think, that is my flaw. Because my life is all possiblities. Where do I go? What do I do? They\'re more or less - all my choices.  My biggest goal in life is to have a family, and to be a better father than my own. And I\'m sure I\'ve found the person I want that with. The problem with primary goals is when they include people who have their own lives, and their own goals. I was never good at that. Sometimes, I like to think to myself how I\'ve never gotten what I\'ve wanted. Perhaps, I should think rather, that I\'ve never wanted what I\'ve gotten. And I wonder about the things I\'ve done to be in this position. I know she is right. Only thing is, it feels like everything is closer than it is, and I have to be reminded that Christmas doesn\'t come this year. Especially for someone who\'s done what I\'ve done. But I\'ve learned how quickly one can descend into weakness. I don\'t want to be there again either. So comes possibly the hardest thing I\'ve ever done. Letting go. Either it\'ll come or it won\'t. One of my major faults is forcing something when I\'m frustrated - to make it work - instead of patience and a steady heart. Whether or not things work out the way I want them to - is ok. There will be tomorrow, other people, other things.  It helps, not to be so self absorbed. But I think that I\'m often a poor observer. I\'d like to think I wasn\'t always caught up in my head...but I am. Perhaps its the caffiene, or my nuttiness.  There is this baby boy in the seat ahead of me, his kind, new eyes staring back at me. How can\'t you help but stare back? The little girl across from the aisle is playing with her cat, entertaining the boy. Sometimes the people are more entertaining than the movie...  Another flight back to the west coast. Not for long enough. Even thought it feels like my heart is being stretched like a rubberband from San Fransisco to Chicago.  So i pray now and forever that I can be the person that I want to be, and need to be. That those around me can forgive my sins, that God can too. That despite my wrongs, that he will have enough mercy on me to let me achieve my happiness, and let it not be at the expense of someone elses.  So, I guess you are right, Sammie, tears are my muse. They aren\'t tears for me though, I won\'t cry for me anymore. I\'ve been blessed, and i think it\'s tough to see, because I can readily see everything that\'s wrong - instead all that\'s right.   '),(5245,1788355,'07,August,2004','2004-08-07 00:00:00','   Guess where I am at?        I think tomorrow, I will buy a camera so I can take my own pictures.   '),(5246,1788355,'07,August,2004','2004-08-07 00:00:00','   Sometimes - to be completely accurate - often I try to make music. Here is a few of my recent experimentations.   urlLink Blue  and  urlLink New   '),(5247,1788355,'07,August,2004','2004-08-07 00:00:00','   your eyes calm the storm of my soul and to my heart you stole and everythin, everythin you\'ll ever desire with your eyes afire stands before another day for all that we pray another way... another day  every dream another hopeful night only to find another height and ill leave it all to wander the world with you   '),(5248,3811749,'02,July,2004','2004-07-02 00:00:00','    my second true love  Aldo Andres Mores.I met him on 21 March 2004. He is the cutest guy I have ever met. He is so wonderful...:)   INITIAL IMPRESSION   He\'s too good to be true.He has all the physical appeals, that took my breath away.And the softness that melts every pores on my body.I m falling for him..so so deep...  There\'s many things that I like is also what he likes. It is so surprising ...a person mirroring you....:) unspoken understanding, no words to say, no way to express it, just feeling and feeling..it is soo peaceful being..with him..:)  I cant describe him because it is something I feel within..  An amazing honesty, a true understanding, a pure feeling. I really love him. he is gone from my life. If you can\'t touch it, you can\'t feel it, it is GONE. And probably it is better living with our memories..   If I could ask, I would ask for him to be mine. He has never hurt me, not a word..:) He has never say goodbye to me..and I don\'t want to say it too because he will took my heart away with him.  I will always remember his eyes, his voice, he\'s singing, he\'s dancing..everything of him..the sun and the tiger..his angels...:)  mi amor, te quiero..muchos..muchos.. '),(5249,3811749,'01,July,2004','2004-07-01 00:00:00','    i miss you..bellisimo david..:)  i dont know what to say..i m thinking about david so so much.. '),(5250,3811749,'01,July,2004','2004-07-01 00:00:00','    my first true love  Florian Geraud Demoisson.I met him on 5 April 2000. And he changed many things in my life. I will always remember him, of course.To love someone that long a time, and that deep a feeling. It is amazing..:)     One thing always right is, everything changed through time. Better or worse still nothing\'s stay.   Love is the best gift from god. How can you ever know the person you laid your eyes that moment changed the way you feel about life, the way you see the world and how strange it will turn your life upside down. It is still a miracle..hmmmm..:)  The first time you knew you are in love, with those pain in your stomach. Churning and warming. And the need to just hearing his voice, to just feel him near you, to just knew he is the person that you love. To get up in the morning to know that you long for his smile and to just think of him as the last thing before you close your eyes every night.You would not asked for more..:)  But then, to love someone is to not hurting the person but at the same time you are just a tiny human, who\'s so fragile and too vulnerable. To just possibly wanted to ask from god to watch over him all the time and not letting any harm ever touch him. It is wearing every emotions you had down. But in return, the smallest gesture of him pays all the miseries.  But c\'est la vie..monsieur..:)  je t\'aime..beacoup beacoup..toujours! ta fille.. '),(5251,3811749,'01,July,2004','2004-07-01 00:00:00','    Part Une   i loved him and now..i want to love again.. '),(5252,3851968,'30,July,2004','2004-07-30 00:00:00','   this movie is so cute... the guy here is so guapo!! wala lng... funny and entertaining... Ãœ   Rating: 10/10 - a must-see.. Ãœ  '),(5253,3851968,'27,July,2004','2004-07-27 00:00:00','   our adviser, Mr. Guinto is thinking of leaving us and thus, making Mr. Roldan our new adviser... hayy... kainis...   sorry... my mind is kind of blank right now.  '),(5254,3851968,'27,July,2004','2004-07-27 00:00:00','   here we go again... the week or i mean weeks of unending projects...     &nbsp;  1) CL project- Portfolio about the outreach and recollection.   2) AP Project- Comics about the outreach....   3) Eng. Project- individual Survey Reports....    &nbsp;  haay.... so many naman!! so tiring!! grabe! hehe ÃœÃœ oh wel... i guess that comes with school life. (obviously)     &nbsp;  basta, kaya natin to! dubah FC?? ÃœÃœ    &nbsp;  gosh.... still have to do the freakin P.E. assign... grrr...    &nbsp;  '),(5255,3851968,'26,July,2004','2004-07-26 00:00:00','   well... 2day, i heard a quite shocking news... our science teacher or i mean, bio teacher is going to make lipat departments... oh no! sayang... i was looking forward to submitting experiment papers and getting a low grade on the investigative project pa naman... haay... de... jke lng yon!! R u KM?? (go norms!) nway... despite the bad news, i had lots of fun today... especially with my seatmate forever... JAMINA! hehe ÃœÃœ basta.. we had lots of kalokohan today... Ãœ tanongin nyo na lang sya.. tinatamad ako ikwento eh.. hehe Ãœ also, Jap was fun... ichi-ni-roku-go-san-nana-watever.... muntik na tlga ung kanina noh grp. yon??? (4) hehe Ãœ well... un muna kwento ko.. my sis is still gonna use the comp. .... ÃœÃœ TTFN!! (tata for now! ehhhhhhhhh... BADUY!)  '),(5256,3851968,'24,July,2004','2004-07-24 00:00:00','   another nice movie... funny as always... basta andyan c jackie chan, nakakatawa sigurado ung movie...   did you knw? u can travel around the world less than 80 days in the modern times... Ãœ ngee..... Ãœ  '),(5257,3851968,'21,July,2004','2004-07-21 00:00:00','   nice movie... dats all i can say... watch it yourself... ÃœÃœ  '),(5258,3851968,'21,July,2004','2004-07-21 00:00:00','   our outreach was fun... my partner was non other than of course, my bestfriend, NotNot. our foster family was composed of Nanay Linda, Tatay Leby, Ineng and Jon. We expected to be doing lots of stuff like cleaning the house, doing the laundry... but instead our nanay told us not to do anything but take care of Ineng so that\'s what we did... Ãœ we took her around... (actually she was the one who gave us the tour.... hehe Ãœ) and we bought her food from the tindahan just across our house... where 3sha and robbie were staying... Ãœ (promise 3sh u\'d make a good tindera! hehe Ãœ) we took pictures and when it was about... 11:00 am, i started cooking the porkchop... NotNot helped a little bit... Ãœ we ate lunch and they told me it was good.... NotNot loved it actually... diba Nots?? hehe Ãœ nway.... when it was about 1:00 pm, we gathered together... and the children played games with Norms and Pielle as the \"gamemasters\" !! Ãœ Go Pielle and Norms! hehe Ãœ after that we performed a not-so-prepared dance number... diba jam?? hehe Ãœ (sorry, i forgot the cd)... but still, jam danced gracefully as always! Ãœ (nax) hehe Ãœ after that, we had a prayer service or something then, we went back to skul.... hehe ÃœÃœ the end.... Ãœ (do i make this sound like a bedtime story or something...) i\'m kind of like an ingleshera 2day... la lng.. Ãœ  '),(5259,3851968,'21,July,2004','2004-07-21 00:00:00','   ang saya saya!! so far.. this is the best recollection i have ever experienced. Ãœ there were lots of activities that were fun and i think they really thought about what activities we can do to make the reco more fun... Ãœ from the start of the day, i had a feeling that this reco was going to be the best and i wasnt wrong! hehe Ãœ i was part of the honest group!! (daw) hehe Ãœ i liked the afternoon part better than the morning part because i had lots of time to relflect...Ãœ i cried, you know.... i think i always cry during recollections... it\'s like a routine... hehe Ãœ i also read the reco letters given to me by my classmates and the one that touched me the most is from my bestfriend, NotNot.... Ãœ grabe... haay... i wish this day didnt have to end.... Ãœ today, we also tested our trust to one another and our friendship!! Ãœ KAPATIRAN dude! hehe Ãœ i liked the dragon activity! (nanalo kmi!) and the chair activity... Ãœ go soj! hehe Ãœ basta... i always remember this special day because i bonded more with my favorite class ever and daddy rob! hehe Ãœ i\'ll never forget our friendship dudes and dudettes!! Ãœ labshu all!! Ãœ belated hapee reco!! Ãœ DONT 4GET MY OUTREACH LETTER! is there such a thing?? Ãœ i love this day! Ãœ  '),(5260,3851968,'20,July,2004','2004-07-20 00:00:00','   dismissal time. \"people, announcement!! ung mga gustong mag- \"audition\" sa dance pls stay after dismissal!\"- sinigaw ko yan prahmis! nway, we had another practice 2day at the classrum and we taught those who wanna join d \"FC Dancetroupe\" hehehe Ãœ i learned new steps 2day... i mean HARDER steps!! as in!! but, cguro i can do it naman... ehe Ãœ jam is soooo good... she\'s d best dancer in our class! basta... watch out for our dance na lng! Ãœ (medyo matagal pa&nbsp;nga lng... hehe Ãœ  '),(5261,3851968,'17,July,2004','2004-07-17 00:00:00','   it\'s our reco 2morrow!! ang saya saya! i hope its gonna be fun.... nways...&nbsp;ang dami nanananamang reco letters ang gagawin ko... oh wel.. i think it\'s better writing them you know.. i mean, i not that straightforward person.. i express myself better through writing especially when i have to say something...... important. Ãœ wel... magpopost na lng uli ako... dami ko pang gagawin eh... ya knw.... d reco letters and stuff... Ãœ  '),(5262,3851968,'17,July,2004','2004-07-17 00:00:00','   haay.... grabe na ito!! Ãœ &nbsp; Anyone can be passionate, but it takes real lovers to be silly.&nbsp;&nbsp;  &nbsp; The best thing to hold onto in this world is each other. &nbsp;  &nbsp; Don\'t go for looks; they can deceive. Don\'t go for wealth; even that fades away. Go for someone who makes you smile, It takes only a smile to make a dark day bright. &nbsp;  &nbsp; The door to the human heart can only be opened from the inside. &nbsp; &nbsp; For it was not into my ear you whispered, but into my heart. It was not my lips you kissed, but my soul. &nbsp; If you love someone, put their name in a circle, instead of a heart, because hearts can break, but circles go on forever. &nbsp; &nbsp; Immature love says: \"I love you because I need you.\" Mature love says: \"I need you because I love you.\" &nbsp; &nbsp; Just because someone doesn\'t love you the way you want them to, doesn\'t mean they don\'t love you with all they have. &nbsp; &nbsp; Love enables you to put your deepest feelings and fears in the palm of you partner\'s hand, knowing they will be handled with care. &nbsp; &nbsp; You learn to like someone when you find out what makes them laugh, but you can never truly love someone until you find out what makes them cry. &nbsp; &nbsp; You never lose by loving, you always lose by holding back. &nbsp; &nbsp; Your soul-mate is the one who brings out the best in you - not perfect but always perfect for you.&nbsp;&nbsp;  '),(5263,3851968,'09,July,2004','2004-07-09 00:00:00','   &nbsp;  ang kukyut ng mga quotes nya prahmis!! ÃœÃœ  &nbsp;    \"It takes a minute to have a crush on someone,an hour to like someone, and an day to love someone...but it takes a lifetime to forget someone.\"    \"Lucky is the man who is the first love of a woman,but luckier is the woman who is the last love of a man.\"    \"Love me now, love me never,but if you love me, love me forever.\"    Time is too slow for those who wait, too swift for those who fear, too long for those who grieve,t oo short for those who rejoice, but for those who love, time is eternity.    I like you because of who you are to me I hate you because of what you did to me I miss you because you want her, not me I love you and that\'s how it will always be.    Spoiler:  My friend, I hate to do this, but our friendship just can\'t last. The time we shared together has all faded in the past. I told you this would happen... Our friendship was bound to end. Although I know you care, I can\'t consider you a friend. Please don\'t try to argue, but try to understand... That time can change two people like the tide can change the sand. Our friendship has been lovely, but you see it just wouldn\'t do. For now I see you in a different way. I\'ve fallen in love with you.&nbsp;           &nbsp;     i especially love the poem at d end.... la lng.. ÃœÃœ   &nbsp;  '),(5264,3851968,'09,July,2004','2004-07-09 00:00:00','   luk   @ marga.... marga...\'s post... i luv it... awww.... ÃœÃœ visit her blogger @  urlLink http://authentic_genuine.blogspot.com     &nbsp;   Argument   Q:&nbsp; What is Love? &nbsp; A:&nbsp; Love.&nbsp; There are three kinds of love... first is real love, crush, illusion. Pinakamasakit yung illusion kasi, for some of us... we say we love someone but we do not really mean it when we say it.&nbsp; Though sometimes involuntarily blurted... hindi mo pa rin matatawag na true love ang isang pakiramdam kung ito lamang ay isang paku-chaba ng pag-iisip. &nbsp; I don\'t know where I got that idea at all... bigla ko na lang nai-type.&nbsp; It\'s strange when I blurt out stuff like this.&nbsp; I say what I think involuntarily, regardless of how it might affect other people.&nbsp; Ganun talaga ako...&nbsp; There are times when I say what I want to, just because.&nbsp; It is never an issue though sometimes people take it the wrong way.&nbsp; Having to think back... there were countless times when I said stuff I meant but I never bothered to insist modesty.&nbsp; I was oblivious to what other people might\'ve or rather... have felt.&nbsp; Thinking about these things, made me remember the times when I was very free-willed.&nbsp; I had never really paid that much attention... it\'s strange when I just... say things.&nbsp; I keep talking and talking and talking, until I don\'t have anything left to say... or until I\'ve run out of breath... &nbsp; Now here comes the realy personal part...&nbsp;my insensitivity towards the world around me shifted when I entered the Honor\'s Class.&nbsp; Though it&nbsp;may seem as if it was alright... I was alright, it really never was from the start.&nbsp; I got pressured every single time, I always had to keep up... it was intense.&nbsp; Then and there have I realized that my&nbsp;insensitive self had not been active, at least no till later.&nbsp; I remembered it vividly.&nbsp; Grade 6 ako noon.&nbsp; Gosh... it seems like&nbsp;yesterday!&nbsp; Honestly, thinking back seems it all happened just yesterday...&nbsp; Anyway, there was this time when I blurted what I had in mind...&nbsp; It was surprisingly unexpected on the 6G honor students to hear me say such things.&nbsp; I guess they didn\'t&nbsp;really expect to hear me say such things... at least not yet.&nbsp; Afterwards, they began to accept me as someone who isn\'t always giving and all.&nbsp; I am opinionated and sometimes insensitive too... &nbsp; But you know... that is exactly what I like about this class.&nbsp; They\'re very open.&nbsp; And now that I am a Freshmen, it\'s all very clear.&nbsp; What I admire about the Honor\'s class is not the fact that they\'re all very intelligent and that they can withstand every \'jolt\', but it\'s they\'re attitude towards people... Everyday it\'s the same thing since Grade 6.&nbsp; I come to school, they\'re all giddy about something, again.&nbsp; It never stops till bell time.&nbsp; For a short period of time in line... they remain quiet.&nbsp; As if to show a distilled disposition.&nbsp; And when that\'s over... the noise retaliates once again.&nbsp; It never stops... I silently watch what they do... the smiles on their faces, the laughs they make, the homeworks they share!!!&nbsp; It\'s all so... them.&nbsp; It\'s a unique bond between friends.&nbsp; Despite the fights, they remain bonded and close.&nbsp; Though not all of them are so close to one another... they still \'stick together\'; siyempre, may pingsamahan kasi eh... &nbsp; It\'s nice to think that I am a part of this bond...&nbsp; I may not be tightly tied to them, I still think that were all good friends.&nbsp; Kahit na may mga napag-aawayan, it doesn\'t mean were not friends.&nbsp; And that\'s what I like most about this class... alam nila na dapat ang bawat tao na bago sa kanila... tinatanggap.&nbsp; It\'s very admirable...&nbsp; It\'s nice to know that I am welcomed amongst them.&nbsp; They understand me perfectly...&nbsp; The achievements, the reputation, the image... these are not those of which that make the Honor\'s Class... the Honor\'s class.&nbsp; It\'s their openness.&nbsp; I am happy that they accept me&nbsp;for who&nbsp;I am.&nbsp; I am happy that they like me for who I am.&nbsp; And most of all... I am happy that they welcome me into their bond... just as I am... &nbsp; Minsan ganun talaga ang buhay.&nbsp; Bigla ka na lang magugulat sa mga pagbabago.&nbsp; Kahit man naiilang ka, natatakot ka, nag-iisa ka... laging may ibinigay na gabay ang buhay upang ika\'y maihikayat upang tumindig.&nbsp; Pasalamat ko sa Diyos na binigyn ako ng pagkakataon na mapasama sa kanila...Ãœ &nbsp; Sorry... is it too much?&nbsp;&nbsp;        &nbsp;  ang kyot noh????? ÃœÃœ ;p  '),(5265,3851968,'09,July,2004','2004-07-09 00:00:00','   sorry haven\'t blogged for a long time.. nway... today we practiced @ jam\'s haus and it was okay... we had a great time watching micmac and miggy\'s video when they were dancing for trumpets... nway.... this year\'s dance will be da bomb!! im syur its da best weve ever made... although we only practiced 2day, we had lots of steps on our puny minds... (jke lng guys!) and we\'re just trying to put them together with the perfect song/s.... ÃœÃœ i hope we\'ll win this year... we\'ve been second place for the past 3 years so ngyn, desidido na kming manalo.... prahmis!! k.... still have 2 do the reco letters... haay..... ANG SAKIT NG KATAWAN KO PRAHMIS!!! GO COPS AND ROBERTS!! Ãœ galingan ntn ha???? ;p hehe ÃœÃœ kaya ng powers ntn 2!!  '),(5266,3851968,'09,July,2004','2004-07-09 00:00:00','   well... i already watched the movie yesterday and it was better than the first movie... ÃœÃœ  '),(5267,3851968,'08,July,2004','2004-07-08 00:00:00','   2day was an okay day... the usual happened... go to school... study... go home... BORING eh?? kelangan eh... hehe ÃœÃœ doi! ÃœÃœ we had a quiz in filipino w/c was kinda hard... the questions were related to AP too... ÃœÃœ oh and 2day was the club promo... it was okay but partly boring.. there was a fashion show... how nice! Ãœ I was thinking of joining the dance troupe but when i saw the sophomores dance, i realized i couldnt do what they just did and it was like- whoa. so, for now, im choosing KKK... ÃœÃœ i hope there wont be any history lessons there... k lng kc c sir guinto naman ung moderator... Ãœ i hope apple joins that club too! ÃœÃœ hehe... asa pa ako... ÃœÃœ o btw, i saw him 2day!! sooper dooper guapo!! ya knw, i dont like talking much about my luv lyf but i feel so bored and there\'s nothing to do... oh yea!! almost forgot!! Quiz in CL and AP tomorrow!! gotta study!! ÃœÃœ -signing out- (waddaheck) â˜º  '),(5268,3851968,'06,July,2004','2004-07-06 00:00:00','   ello!! its friday tomorrow and im gonna buy a guitar!! yay!! cant wait... ÃœÃœ and tomorrow, were gonna watch Spidey man 2... la lng!! ÃœÃœ  i wanna say belated happy bday to my classmate robbie!!! and Advance Happy Bday to Steph!! her bday is 2morow!! ÃœÃœ  dats all... till nxt tym.. ÃœÃœ  '),(5269,3851968,'06,July,2004','2004-07-06 00:00:00','   well... its my first tym here... its kind of like, helpful to express yourself more so i joined...ÃœÃœ well... i still have to read and do assignments so im gonna logout now...ÃœÃœ i\'ll post more sum oder tym.. ÃœÃœ  P.S. yehey!! pep rally tomorrow!! la lng! Ãœ  '),(5270,3851968,'07,August,2004','2004-08-07 00:00:00','   hey guys.. sorry for not blogging for a long tym...   This week.   this week was okay... the usual stuff... projects here, projects there, tests here, tests there... oh well.. guess datz highskul lyf... honestly, i haven\'t totally adjusted so im kind of freaking right now. Ãœ (O.A.) nway, im totally nervous about my english Long test and my science quiz... i feel like i\'m gonna have failing scores... nway, we had club last thurs. and i already changed my club. Now, i\'m part of the dancetroupe and it\'s okay... at least it\'s about something i love.... DANCING... i guess many people know that particular detail about me.... that dancing serves as one of the things that makes me happee...    of course. isnt it obvious? (medyo lang naman) hehe Ãœ   Last friday, we also had our mass...    yun lang. haha Ãœ      TODAY.   today is a saturday. duh. today we practiced again at jam\'s house. doi. it was really fun though.... i think there were twelve of us there... actually, it\'s more of like hanging out than practicing for our dance which we are going to perform in like a freakin month! Ãœ the merienda part was the best! prahmis!! we were talking about things... (ehem) hehe ÃœÃœ you guys know what i\'m talking about! Ãœ and, we made new steps which are not actually polished yet but it\'ll do... basta, saturday\'s at jam\'s house are unforgettable especially if you share it with the people you care about... aww.. Ãœ thanks guys for making my weekend complete! Ãœ (grabe.... so mushy naman me!)      HIM. â˜º   him. i don\'t know if he\'s the one but he does make my heart skip a beat whenever i see him. Lots of things happened between him and me this week though not very personal (ehem... FC, ya know what im talking about) it\'s very kilig to the bones!! haay.... i think im infatuated. Help! Ãœ      English project.   need help! can u guys suggest a decent topic to make a survey on?? ÃœÃœ      well... guess this is too long... im gonna stop now.      HAPEE WEEKEND! Ãœ  '),(5271,3851968,'02,August,2004','2004-08-02 00:00:00','   laurs is hereby authorized to take the sea laurs is one of australia\'s leading cello soloists laurs is now principal cellist with the adelaide symphony orchestra laurs is referring to things like laurs is also the principal cellist with the adelaide symphony orchestra; elinor levy is associate concertmaster of the tasmanian symphony laurs is laurs is because of the corrs laurs is fresh i adore her laurs is a funny kid laurs is my bestfriend laurs is age 29 (daw) laurs is a german national living in germany laurs is currently an associate with the consulting firm of booz allen hamilton laurs is my best bud laurs is always happy laurs is fixing to update mine this weekend  what??? if you want a good laugh, search your name in  urlLink www.googlism.com  '),(5272,3851968,'02,August,2004','2004-08-02 00:00:00','   maybe i was really overreacting. sorry to whoever has a violent reaction. i just read miggy\'s blog entitled acceptance and i guess he\'s right.  maybe i just had to spill my feelings... everybody does.  now, i feel much better. Ãœ  thanks miggy for enlightening me! hehe ÃœÃœ  '),(5273,3417999,'27,May,2004','2004-05-27 00:00:00','    Ah, yes, now you can make sure that your kids don\'t hear someone who unwittingly triggered Armaggedon in Terminator 3 say \"Dear God\" while still allowing them the wholesome pleasure of seeing Arnold Schwarzenneger shove his female opponent\'s head into a toilet.  As David Pogue puts it in his  New York Times  article,  urlLink \"Add \'Cut\' and \'Bleep\' to a DVD\'s Options\" : \"...as it is, the evidence suggests that ClearPlay\'s technology is not intended for families at all. It\'s for like-minded adults, specifically those who are offended by bad language and sexual situations but don\'t mind brutality, destruction and suffering.   \"Maybe every ClearPlay-sanitized movie ought to begin with a message: \'This film has been modified as follows: It has been formatted to fit the taste, sensibilities and religious beliefs of a couple of guys in Utah. That\'ll be $1.50.\'\"  Once again, it\'s the people crying the loudest about the downfall of U.S. culture that are degrading our society the most. And sticking our country\'s fists in places we should not be, like Iraq.   '),(5274,3417999,'26,May,2004','2004-05-26 00:00:00','    I\'m surprised not to have heard any guffaws about George Bush\'s mangling of the pronunciation of \"Abu Ghraib\" a couple of days back. When I first heard him say it, I thought there was a glitch in the tape. But then when he got more and more confident in his mispronunciation, I realized that his handlers had once again forgotten to do those flash card exercises with him before he tried to talk.  Of course, I don\'t watch the Daily Show â€“ or much TV at all â€“ and I can\'t imagine they\'d pass up that opportunity. Maybe it\'s appropriate he mangled the  prison\'s name, much as the soldiers there have mangled any sense of human decency. Of course, the entire administration has shredded human rights, civil rights, constitutional rights, and any idea of lawfulness, so it\'s not as if the soldiers who committed those despicable acts didn\'t have role models. I don\'t believe for a minute that these were simply a few rogue soldiers â€“Â this reaches up to higher levels, and the U.S. administration, starting with Donald Rumsfeld and George Bush, should be mature enough to take responsibility. Where\'s Truman when you need him?   '),(5275,3417999,'25,May,2004','2004-05-25 00:00:00','    I wrote a couple of days ago about  urlLink the positive aspects of a car-free Manhattan , and possible ways to do it. Now let me tell you what pisses me off about those metal boxes of death:  When I\'m walking at a good, steady clip, trying to get some exercise as I get to where I\'m going and I have to stop at every damn traffic light just so people in cars can go faster, I become a frustrated young man.  When I have to literally stick my neck out into the street to see past a parked van or SUV to make sure I won\'t get creamed by oncoming traffic, I contemplate reducing its height a few inches by letting the air out of its tires.  When I\'m squeezed on the sidewalk by too many pedestrians and see how much space all those cars are taking, it makes my blood boil.  When drivers act like the second line of the crosswalk is where they should stop for the light, I feel like walking over their hood. Very, very hard.  Don\'t even get me started on car alarms.  When some car honks at me because I\'m riding my bike ahead of them in a narrow lane, I want to put my fist through their horn. As it is, I just pull into the very center of the lane to make absolutely sure they don\'t try to get by me and scrape me between themselves and the parked cars. Hey buddy, you don\'t own the damn road!  When I\'m forced to breathe your SUV\'s second-hand smoke, I want to scream. But I don\'t, because that would mean I\'d have to breathe in even more of the air you\'ve fouled.  And when I hear that over half of the people getting hit by cars in New York City are  in the crosswalk , I feel like taking a sledgehammer to the grill of any car that gets near me.  See, especially in a place like New York City, there is so much bad that is associated with cars, that it really outweighs any of the good. So get out of your car and freakin\' walk! And while you\'re at it, check out  urlLink Transportation Alternatives, NYC . They\'ve got all kinds of information on making our cities more human-friendly.   '),(5276,3417999,'24,May,2004','2004-05-24 00:00:00','    As we all know, the Bushies love a good lie. But come on,  urlLink trying to pull a fast one about the weather ?!?   '),(5277,3417999,'23,May,2004','2004-05-23 00:00:00','    This blog seems to be turning into a discussion of urban issues, but since that\'s what\'s interesting me at the moment, I\'m going to keep going with it.  I guess today I\'ll do a little bragging. Not about me, but about my neighborhood. You see, I grew up in a small ag town in California and, despite its relative diversity, I am still sometimes bowled over by the day-to-day richness of where I live.  My block on the western edge of Park Slope in Brooklyn, New York has people from an amazing variety of ethnicities, races, and nationalities. My own incomplete recollections enumerate:  	- Korean 	- Egyptian 	- Puerto Rican 	- White Boy Mutt (that\'s me)  That\'s just in my three-apartment building. On the rest of the block there are also: 	- Japanese 	- African American 	- Italian 	- South Asian 	- Yemeni  Okay, the Yemeni guys actually run the corner store. Which brings up the shops and restaurants in a one-block radius (i.e., less than a minute\'s walk away): 	- the aforementioned corner store 	- a fish and chips shop that also serves curry and is the inventor of the deep-fried Twinky. 	- a Peruvian restaurant 	- a Mexican restaurant 	- a Thai restaurant 	- an Italian restaurant 	- an excellent deli 	- a barbershop 	- a grocery store 	- a hardware store 	- a big chain video store 	- an odds-and-ends store 	- a travel agency 	- a furniture store 	- a real estate agency 	- a carpeting store 	- some other stores I\'ve forgotten 	 Walk another block in either direction and you\'ll find a greengrocer, a tattoo parlor, a comic shop, a fancy florist, my bank branch, a fried chicken place, etc., etc., etc.  My block is quite residential and the area around it has only in the last couple of years begun to come into its own. With all that\'s going on and available so close to me, the area is still human-scale, and walking around is a pleasure. I know a goodly number of my neighbors. The guys at the corner store would pull down prepaid calling cards as soon as I walked in when my girlfriend was away in India. Much as I still feel like a Californian, Brooklyn is New York City\'s saving grace, and my neighborhood is a very integral part of that.   '),(5278,3417999,'23,May,2004','2004-05-23 00:00:00','    Yesterday I biked over to the newly renovated  urlLink Brooklyn Museum of Art  to check out its new look. I wanted to see how well the new glass atrium, the radiating lawns, the ampitheater, and the dancing fountains worked for the space. I was not disappointed.  The renovation gives the plaza and entryway a much-needed multi-level aspect and encourages the surrounding community to utilize it for a myriad of purposes. Before, the museum stood as a blank, vertical, neoclassical monolith across a flat expanse of lawn. This has been broken up by the curving glass atrium that pushes out from the building. The series of narrow lawns, along with the ampitheater to the eastof the entryway, radiate out from the entrance, drawing passersby into the museum\'s sphere of influence.  The steps of the ampitheater are so broad that I could sit on one step, lie my bike flat in front of me, and still afford plenty of room for someone to sit in front of my bike. These platforms, rather than being solid concrete, are instead framed by it, with the vast majority of the surface area made of wood planking. This planking also features on the walkway that leads from the top of the ampitheater around the curve of the glass entryway and affords interesting views into the entryway itself as well as across the plaza. The stairs down from the other end of the curve are also made of the same wood. While not visible if you are simply passing by, the wood softens the features of ampitheater for those who take the time to sit and conveys a sense of being on a boardwalk, along with the leisure and relaxation that connotes.  Children are drawn to the ampitheater space not just because of the opportunity to run up and down the stairs and make up multilevel activities, but also because of the dancing fountains that provide a visual focus at the bottom of the ampitheater. It has been stated by a number of people that the fountains are a legacy of mall gimmickry. But these fountains are placed right out front rather than inside some courtyard or behind an impenetrable barrier of hedges or tulips. Maybe this is why metal police barricades were set up around their perimeter yesterday. I hope they take those barricades away - looking at how the fountains are situated, I get the sense that the architects wanted to maintain a feeling of openness. Is the museum afraid of liability? That the fountains aren\'t robust enough for children to play in and amongst them? If they are afraid of these possibilities, and insist on making the fountains themselves off-limits, I would hope they would consider doing it in a way that retains an overall feeling of openness.   I would also be remiss if I didn\'t note that the fountains use a fair amount of water and I\'m sure a not-small amount of energy to run. Much as I feel curmudgeonly for criticizing the fountains on sustainability grounds - since they seem to be an effective draw to the plaza - it is important that we always look to principles of sustainability especially when building new projects. It\'s possible that the fountains run with treated greywater from the museum and solar cells from the building\'s roof, and if that\'s the case, so much the better. In that case, they could become a public educational piece about the possibilities for sustainable design. But somehow I don\'t think they\'re doing what I just suggested.  What defines the success of an open space is whether people use it. Yesterday, despite the spurious threat of rain, many people were using the space. People sat on benches under the trees at the west end of the plaza, and a few folks lounged on the lawns in the middle. At the ampitheater, parents lounged on the steps watching their children playing down by the fountains, and not a few kids got themselves purposely drenched when the water blew their way. It was a diverse crowd on the plaza, and it was apparent that many of the people there that day came from the surrounding community. This last point is especially important and positive, because it means that the museum has succeeded in providing a service to the people who live and work around it, and so has integrated itself into the community a little bit more.   '),(5279,3417999,'22,May,2004','2004-05-22 00:00:00','    What if Manhattan were a place with no passenger automobiles? What if, for the most part, trucks were also banned? It seems to me that because of its density, the borough is one of the most amenable places for getting around without cars or motor taxis or the like. We could have an extensive networks of electric trams, a fleet of pedicabs, people on bicycles and on foot. The subway system\'s capacity and extent could be increased. Cargo could be brought in via train and carried to centralized points of the borough using a tram-like system, which could possibly be all underground. Small electric vehicles or pedicabs could then be used to take goods from distribution points to stores and institutions. Many streets could be either wholly or partly converted away from vehicular use into parks, gardens, outdoor shopping, play areas, and the like. Emergency vehicles such as fire engines and ambulances, as well as certain service vehicles, might need to have the same flexibility of movement they do now, along with their current sizes. But police could patrol more on foot, on bikes, or in smaller electric vehicles similar to what the meter cops use now. With a reduction in asphalt, the city would become less of a heat sink it is now. With less fossil-fuel burning vehicles, air would be cleaner. As a pedestrian-oriented city, even more tourists would be attracted. And vehicle-related stresses would be reduced. These are a few initial thoughts, and I would welcome comments, additions to this narrative, and the like. I would love to see someone come up with visual sketches of what this all could look like. One group that is working at moving New York City away from the automobile is  urlLink Transportation Alternatives .   '),(5280,3417999,'22,May,2004','2004-05-22 00:00:00','    It\'s very late, and my brain is rather mushy after setting this up. So at this point, all I am posting is a little experiment to see whether I can get this whole thing to work. It looks like it might be, and I\'m fairly amazed at how easy this whole process is. Though I need to change the layout at some time.   '),(5281,3417999,'24,June,2004','2004-06-24 00:00:00','    The Center for American Progress has posted  urlLink A Declaration Calling for the Resignation or Removal of John Ashcroft . I would encourage everyone to read over this declaration and, if they see fit (I did), to sign the petition that goes along with it. John Ashcroft is, as Paul Krugman of  The New York Times  put it in at least two of his columns, the worst Attorney General in U.S. history. His affronts to our democracy are legion, and his performance as our country\'s top law enforcement officer are abysmal.   There are currently questions swirling around the fact that although he will blast the trumpets whenever a person of color or Muslim is arrested on the flimsiest of charges, he has made not one public peep about the  urlLink arrest in December 2003 and subsequent guilty plea of William Krar , a white supremacist who was caught with automatic weapons, pipe bombs, a stash of cyanide gas large enough to kill everyone in a 30,000 square foot building, and false UN credentials and Defense Intelligence Agency IDs. This man would have been able to carry out the most devastating attack on U.S. soil since 9/11/2001, and our attorney general (along with the vsat majority of our country\'s media) has completely swept it under the rug. Might this be because Mr. Ashcroft has a history of being supportive of white supremacist groups?  Keep in mind that John Ashcroft was appointed by a president who was illegitimately placed into power by a partisan Supreme Court. Keep in mind also that the people of Missouri so reviled John Ashcroft that he lost his race for U.S. Senate to a dead man.  So let\'s put John Ashcroft where he\'ll do us the most good â€“ as far away from government as possible.   '),(5282,3417999,'11,June,2004','2004-06-11 00:00:00','    Let me not be charitable towards Ronald Reagan. So many are doing this for him now, and so undeservedly. Yes, it is fine to mourn the passing of another consciousness, though in this case much of that was gone or at least disastrously altered. But to engage in the wholly uncritical historical revisionism that has been going on is wholly repugnant. Even so-called liberal commentators seem to forget that this was a man that stood against everything they themselves stand for and, in the process, against human decency.  For those who would point out that Reagan started out his political career as a labor leader, let us examine his most important actions within the Screen Actors Guild, of which he was indeed president. His most important activity within SAG was to denounce fellow union members and others in the film industry as Communists during the McCarthy era, destroying careers and lives in the name of a hysterical political witch hunt that even conservatives now invoke as a black spot on our history.  This is the man who, as governor of California (beware Californians becoming Presidents), called out a violent National Guard against free speech protestors. This was the man who vetoed a state bill to allow farmworkers to have the simple right of collective bargaining.  This was the man whose people, in order to put him over the top in a tight race for the presidency, arranged for Iran to prolong the holding of US hostages until after the elections so as to further damage his opponent, Jimmy Carter. It is no coincidence that the hostages finally were released on the day of Reagan\'s inauguration.  Let us not forget, in this time of war on Iraq, that Reagan\'s administration actively supported the regime of Saddam Hussein, selling him weapons, including the chemical variety. Many of the weapons used against US troops in 1991 and even today had and have US serial numbers on them. The chemical weapons that we helped provide were not only used against Iranian troops during the Iran-Iraq War, in clear contravention of international treaties. They were also used against Kurds during the infamous episode that has been hauled out repeatedly in order to vilify Saddam and justify our waging war on the Iraqi people.  Reagan is most celebrated by his proponents for his supposed victory against Communism and the fall of the Soviet empire. Oh, how we forget about Mikhail Gorbachev and the people of the Soviet bloc! It was the people themselves, without our help, that made change in the Soviet bloc. Let us not forget how surprised our own government, including the CIA (whose track record suggests the \"I\" should simply be dropped from its name), was when the Berlin Wall fell and the Soviet Union shortly thereafter.  Yes, Reagan did actively fight communism, or what he perceived to be communism. But where did this take place?  In the case of Nicaragua, Reagan\'s administration violated the very laws of Congress in order to support the Contras, a band of murderous thugs who worked to bring down a democratically elected government. Reagan\'s \"freedom fighters\" rarely engaged Nicaragua\'s military, preferring instead to target civilians in their villages, destroying fields and hospitals and killing hundreds in the process. This strategy was undertaken not only with the blessings of Reagan\'s administration cronies, but at their urging. And how did we fund the Contras? Ah, yes â€” with drug money, and arms sales to Iran.  In the name of quashing the communist threat in the rest of Central America, Reagan propped up regimes in El Salvador, Guatemala, and Honduras. Hundreds of thousands of people died in these tiny countries at the hands of US-trained death squads.  Reagan even conducted an out-and-out invasion of a country in order to supplant a democratically-elected government. As if Grenada would have ever posed a \"threat\" to the US. And as if it even wished to. Basically, in this case people were killed for a publicity moment, to show that we were big and tough and nobody should mess with us. Especially if they were from small, poor, defenseless countries.   While he didn\'t seem to like democratically elected regimes, Reagan\'s record elsewhere suggests he never met a righ-wing dictator he didn\'t like. From Marcos to Suharto to Pinochet, he and his administration made sure that these blights upon humanity had the backing of the United States of America.  His love of dictatorships extended to the race dictatorship, otherwise known as apartheid, of South Africa. Reagan was not only vocal in his support for the apartheid regime, he actually  vetoed  a bill to place sanctions upon South Africa. Much to The Great Vetoer\'s chagrin, and our relief, his veto in this case was overridden. And let us remember he labeled Nelson Mandela â€” yes,  the  Nelson Mandela â€” a notorious terrorist, along with the ANC, which now rules South Africa. Reagan\'s policies served to lengthen the time apartheid held sway, and he himself is therefore responsible for the prolongation of death and suffering in that country.   Not that Reagan wasn\'t active on the home front. Of course he was. Why, under Reagan the homeless population in this country rose from relative negligibility to two million. How\'s that for making an impression? Reagan also has the distinction of being the president under who the US for the first time became a debtor nation. The massive amounts he spent on weapons, including the nuclear kind, surely aided in that. How ironic it is that vocally \"small-government\" Republican administrations these past decades have been the ones who have been the most profligate in their spending! Reagan is at the root of this pattern.  This is also the president who took the unprecedented action of firing all of this country\'s air traffic controllers who had gone on strike. This was not only a blow to the right of working people in the US to organize, but also left our skies much more dangerous by bringing in, en masse, a whole cadre of barely-trained controllers to manage the air lanes.  This is the man who consistently attacked civil rights, and then affirmative action, from the time he was president of SAG to the time he was president of the USA. He even began his 1980 presidential campaign in the infamous Philadelphia, Mississippi, where three civil rights activists ahd been murdered in the 1960s, without so much as acknowledging the incident.  So as people remember Ronald Reagan, we should not only remember the cheerful face, but remember that this face masked a combination of bullheaded ignorance and outright malevolence. Let\'s remember that this is the man who nearly brought us to the brink of nuclear annihilation with his funny little aside that he had outlawed the Soviet Union and bombing would commence in five minutes. His personal charm masked a hateful boorishness that is reenacted now, more baldfacedly, by George W. Bush.  Ronald Reagan, go away and leave us in peace. And take George with you.   '),(5283,3417999,'10,June,2004','2004-06-10 00:00:00','    When I was a child, my parents took me to the Exploratorium in San Francisco. Making even more of an impression on me than an original stand-up arcade version of Pong and simple laser light demonstrations was a curious little device that demonstrated the power of optical illusion. The entirety of this highly sophisticated machine consisted of a foot-diameter black-and-white spiral rotating slowly.  urlLink 60 Second Trip  brings a slightly more complex version of that spiral to your Web browser.  urlLink Little Fluffy Industries  says \"This neat optical trick will give you the same trippy feeling you\'d get from spinning around real fast,\" but it\'s actually much cooler than that. Bascom-Bob says check it out.   '),(5284,3417999,'08,June,2004','2004-06-08 00:00:00','    An open letter to the Metropolitan Transportation Authority (New York City and surrounds) regarding some of their  urlLink proposed new rules :  To the MTA:  I strongly object to the proposed ban on photography in the New York City subway system. This will NOT enhance passenger security or safety. It will do nothing to keep a determined person from using such devices as cell phones, Palm Pilots, and the like from taking pictures. Instead, it will insidiously extend a culture of fear into our transit system. It will prevent passengers and visitors from taking with them a little of the romance of New York City. I have been taking spur-of-the-moment pictures in the subway system for the ten years that I have lived here, and they have been important in showing my friends elsewhere in the US and the world what makes NYC such a great city. Diverting police resources to enforcing a no-photography rule will only keep them from work that truly keeps us safe, and will make the atmosphere more hostile for everyone down below. Please do not play into the culture of fear that the Bush administration wants to impose on us.  I must also object to your criminalization of frustrated passengers who have to deal with your extremely imperfect Metrocard technology. While I have never jumped a turnstile, I would sympathize greatly with those who do so due to a malfunctioning card. I sometimes wonder what real-world testing you actually did with Metrocards before setting them loose on us. I am not opposed to a card-based technology in principle â€” indeed, I got along famously with the SF Bay Area BART system\'s cards. But your cards have caused me to miss trains more than once, and because of the often unreliable nature of your train service, one missed train can eventually mean delays of up to 20 minutes, when missed transfers are factored in, during the day. Late at night, the waits can be even more disastrously long. So my suggestion is to get your own technological house in order before criminalizing your paying passengers.  Finally, do not prohibit people from transferring between train cars. I know of many cases when people transfer because a car that they entered is hostile in some way: there is no air conditioning in the summer, there is an especially foul smell, there are people acting in a threatening manner. In addition, people in New York work to squeeze whatever efficiency they can out of the subway system, getting themselves to a car that will leave them closer to a transfer point (see above point) or to their exit. Why not instead come up with ways to make those inter-car transfers safer?  Thank you for your time and consideration, Bascom Guffin Brooklyn, NY   ');
INSERT INTO `msgs` VALUES (5285,3417999,'06,June,2004','2004-06-06 00:00:00','    I just sent the following letter to Roman Catholic Bishop Michael J. Sheridan regarding his  urlLink call to deny communion  to politicians who support issues that violate specific Roman Catholic teachings.   Bishop Michael J. Sheridan The Diocese of Colorado Springs  228 North Cascade Avenue  Colorado Springs, COÂ  80903   Dear Bishop Sheridan,  I am trying to understand why, in your call to deny communion to Catholic politicians and voters who support abortion, stem cell research, or same-sex marriages, you do not call for the same measures to be taken against those who support the death penalty. As you yourself state, \"As Catholics we have the further obligation to give assent to the doctrinal and moral teachings of the Church because \'to the Church belongs the right always and everywhere to announce moral principles, including those pertaining to the social order, and to make judgments on any human affairs to the extent that they are required by the fundamental rights of the human person or the salvation of souls.\' [Catechism of the Catholic Church, 2032 and Code of Canon Law 747.2].\"  One of these moral teachings reads thus:  \"Assuming that the guilty party\'s identity and responsibility have been fully determined, the traditional teaching of the Church does not exclude recourse to the death penalty, if this is the only possible way of effectively defending human lives against the unjust aggressor.   \"If, however, non-lethal means are sufficient to defend and protect people\'s safety from the aggressor, authority will limit itself to such means, as these are more in keeping with the concrete conditions of the common good and more in conformity with the dignity of the human person.   \"Today, in fact, as a consequence of the possibilities which the state has for effectively preventing crime, by rendering one who has committed an offense incapable of doing harm--without definitively taking away from him the possibility of redeeming himself--the cases in which the execution of the offender is an absolute necessity \'are rare, if not practically non-existent.\' (NT: John Paul II, Evangelium vitae 56) [Catechism of the Catholic Church, 2267].  What this teaching tells us is that the death penalty, especially as practiced today in the United States, is morally wrong. Why, then, do you not call to account the many politicians who so strongly support the death penalty, especially in the face of a growing mountain of evidence that many innocents have been put to death by the state in this country?  If you think this question smacks of partisanship or particularism (i.e., why single out the death penalty to be against), then mightn\'t the same be true of your own pastoral letter? Why not simply, in your letter, call for a denial of communion to those who support violations of any of the Church\'s teachings? This would at least give some sense of consistency and fairness to what you propose.  I look forward to your reply.                  Sincerely,                 Bascom Guffin   '),(5286,3417999,'02,June,2004','2004-06-02 00:00:00','    What follows is a wish list of some ways the subways in New York City can become more pleasant. I am putting these down without regard to cost-effectiveness, though if the subway system is more pleasant, there is a good chance ridership, and therefore revenue, would go up:   Make them quieter. What is especially loud and disturbing is the express trains hurtling through local stations. Maybe there are new wheel technologies that could be applied on the current tracks. The Paris Metro runs on rubber wheels. Another option would be to fit ceilings and pillars (including those separating tracks) with sound-absorbing materials. It is possible that there are hard rubbers or polymers that could do this while at the same time being extremely durable and easily washable.   Fix the sound systems. The loudspeakers in the stations sound like tin cans. Coupled with the echo-chamber nature of the stations (see above), this makes announcements at most locations incomprehensible. Make sure that the system is set to the appropriate loudness.   Include digital visual signage for announcements. New York City is far behind almost every other metro system when it comes to this. The signage can normally give the current time, and wait times for trains coming to that platform. It can give announcements about delays or route changes. Up-to-the-minute visual signage is also critical for people who are deaf. By the same token, it should not replace voice announcements, which are critical to blind people.   Daylight stations more. Those with darkening grills that let air in might be replaced by something that allows more light through. Or the ceilingway between the grills and the stations can be modified to allow more light. Utilize plexiglass, metal grilling, and maybe even mirrors to get more natural sunlight into the train stations, and maybe even along the tracks. Along with more sunlight will come lower energy consumption, as well.   Electric lighting should use full-spectrum bulbs, meaning light bulbs that emit the same spectrum of visible light as the sun. Alternately, or as a complement, various colors of lighting can be used in various arrangements to make the lighting of stations and trains more interesting and inviting, while still conveying a sense of comfort and security.   Make them free. This contradicts my argument about revenue going up, although there might be a way to obtain more public moneys as ridership inevitably goes up. Making subways free will eliminate turnstile jams. It will eliminate frustrations with faulty Metrocards, Metrocard readers, and vending machines. Stress levels associated with using the subway system will decrease.   Have personable employees at every station, and maybe every platform, to give information to riders. They could be stationed at small kiosks on each platform, and would also serve as a visible sign of security for riders at all times of the day and night.   Make all stations and trains easily wheelchair accessible. This will not only aid people who use wheelchairs, it will also be a boon to the many people who push carts and baby carriages and the like. And will allow ambulatory people with difficulties walking to get around the subway system more easily. This will also mean less jams on stairways, that now act as bottlenecks trying to get to and from trains.   Make the system much more extensive. Add the Second Avenue line, yes. But also add lines in the Bronx, Brooklyn, and Queens. And not simply all radiating out out of Manhattan. It should be easier to go places between and within boroughs outside of Manhattan.  These measures, and more, would make people\'s subway experience more agreeable, and possibly even pleasurable, and would go far toward upgrading the quality of life in New York City.   '),(5287,3417999,'01,June,2004','2004-06-01 00:00:00','    Nelson Mandela, icon for those who value freedom, compassion, and humanity has announced that he will effectively be  urlLink retiring from public life . This is a sad day for all of us. I wish him well as he finally takes time to be with his family and himself. We salute you, Madiba, for all that you have done for us and inspired in us, the world over. May you live to be 115, always in good health. Amandla!   '),(5288,3417999,'06,July,2004','2004-07-06 00:00:00','    Paul Krugman  urlLink writes today in  The New York Times  : \"Bush campaign ads boast that 1.5 million jobs were added in the last 10 months, as if that were a remarkable achievement. It isn\'t. During the Clinton years, the economy added 236,000 jobs in an average month. Those 1.5 million jobs were barely enough to keep up with a growing working-age population.\" This says a great deal about how our economy is doing - times are not getting any easier for working folks (and especially for folks who are trying to  find  work. So much for the miracle-generating powers of Bush\'s tax breaks (you know, the ones that have made rich folks richer and put us as a nation trillions of dollars in debt). This is just one more example of why we should look skeptically at every single word that comes out of the Bush administration\'s collective mouth.   '),(5289,3853532,'09,July,2004','2004-07-09 00:00:00','    I want to be one of those old hippies who wears mumus because they\'re comfortable and wears her hair long even though it\'s gone grey.  Is it weird that I\'m looking forward to aging? I mean, I\'m not looking forward to aches and pains, and I\'m not hoping to rush the process or anything. But it seems like I will feel less pressure to be \"pretty\" when I\'m old.   I know that a lot of women feel it\'s the opposite: that it\'s easy to be young and hard to age (and if I worked in Hollywood--where a career can be ruined by the natural aging process--I might agree). But I think it\'s terribly difficult to be young and not fit cultural standards, physically speaking. People look at you more when you\'re young. It\'s when you\'re supposed to be finding a mate and orchestrating your fairy tale ending. A task that poor self esteem makes quite dicey, I must add.  When you\'re old, all that stuff is--supposedly--over and done. Decided. There\'s no pressure.  Maybe I\'m just delusional. It\'s easy to pontificate this way from 24 years old. I might feel differently when I\'m 60. '),(5290,3853532,'07,July,2004','2004-07-07 00:00:00','    Just wanted to say a quick hi, and I\'m here, and I\'ll be back. Like the terminator..  ;) '),(5291,3853532,'06,July,2004','2004-07-06 00:00:00','    I hate to diet. I absolutely  abhor  it.  It seems like I am always setting myself up for failure. I set goals--sometimes tiny, sometimes enormous and unrealistic--and when I fall short, I feel guilty.  Guilt and dieting, for me, go hand in hand.  I think I have built up a resentment to the entire concept of dieting  because  of the guilt I feel when I break the regimen.   It\'s a bit silly really, because I have had self-imposed dietary restrictions for as long as I can remember. I limit my dairy intake (because of lactose intolerance, and also, dairy isn\'t that good for you!), I don\'t eat red meat or pork, and I limit my carbohydrate intake (despite constantly craving carbs). This is just something I do, but if I called it a diet, I\'d eat half a cow and a tub of ice cream tomorrow, just in spite of myself.  I wonder why? '),(5292,3853532,'05,July,2004','2004-07-05 00:00:00','    Just a short note by me testing this out and letting you know I\'m here. '),(5293,3853532,'05,July,2004','2004-07-05 00:00:00','    Here\'s the deal: I have never been supermodel material.  I look at women who spend their days on the catwalks and my mind boggles. No amount of proper eating, exercise or metabolism-enhancing pills will ever give me their physique. And being so bombared with adverts and images telling me I\'m TOO FAT and I need to LOSE WEIGHT FAST puts me into sensory overload.  I can\'t be all bad, because I\'ve had boyfriends, and people tell me I\'m pretty quite frequently. But I don\'t feel attractive, and I haven\'t for as long as I can remember. I have never been, in my estimation, a pretty girl.  But why the fuck am I so hard on myself? Why are we all so hard on ourselves? When did life become a constant obsession with dieting, instead of something to be lived and loved and  enjoyed ? When did Atkins become more important than God?   I recently met a man on the Internet, and he\'s really wonderful. I\'m so enjoying getting to know him that we\'re slated to meet in person in early Autumn. Already I\'m worrying--will I compare to other women he\'s been with? Will he be disappointed? Mind you I went through all of these worries with my ex boyfriends, and it was all for naught. They still took me to bed (where I work magic, because nothing makes me feel better about myself than pleasing my mate), and in some cases, took me into their hearts.  I decided to start this blog because I\'m tired of being holed up in this prison of poor self-image, which in turn just fuels the fire of depression (which seems to come and go cyclically). I\'ve asked some friends to come on board and join the \"Strangely Shaped Girl\" team to post their thoughts too, so that together we can all start to reform our bodies, minds, and spirits, one post at a time. '),(5294,3853532,'15,July,2004','2004-07-15 00:00:00','    I\'ve been thinking a lot about relationships lately. Specifically, I\'ve been thinking about what I bring to the table, and why I have a hard time selling myself to potential partners.  I\'ve been blessed to have some really meaningful relationships in my short life. And I think they\'ve bene formative, too. I\'ve learned a lot about how I behave in a partnership with another human being. All in all, I think I make a pretty good S/O.  I am kind and caring. I am compassionate. I am generous. I am passionate. I am faithful. I am honest. I am supportive. I\'m a very good lover (and honestly, I probably wouldn\'t be if I were catwalk material and all I had to bring to the bed were my looks). I\'m attentive. I listen well. I give good feedback. I love to experiment. I am adventurous when able so to be. I am open-minded. I am willing to compromise. I appreciate little things. I like to rise to a challenge.  So if I bring all of those qualities (and more) with me to a relationship, a sort of personality dowry, why has it always been so hard for me to be confident about seeking and securing a partner? I read through personals adverts and a lot of men are looking for the sort of things I have to offer.   I am getting better, though. A little bit every day. I can\'t take all the credit for my development, though. I have friends who do a great job at boosting my ego! '),(5295,3853532,'13,July,2004','2004-07-13 00:00:00','    A Palm Beach Post writer, making the point that America\'s obesity problem is not limited to humans, reported from the Boca Greens Animal Hospital (Boca Raton, Fla.) in June that \"Pumpkin,\" a 12-pound Chihuahua, was up and moving well after her recent liposuction surgery. However, the 12 ounces of fat she lost still left her among South Florida\'s overweight pets, said to be two-thirds of their population. As Pumpkin\'s owner was reminded, surgery is not to be a substitute for sensible exercise and a modest number of treats. [Palm Beach Post, 6-22-04]   urlLink Click Here for More   What the heck? Honestly, it\'s not bad enough they\'re doing liposuction in the stupidest places on people now, but dogs? DOGS? '),(5296,3853532,'12,July,2004','2004-07-12 00:00:00','    Self-esteem, for me, seems to come and go. Almost as cyclically as depression (which, mercifully, I haven\'t suffered from in quite some time).  Sometimes my self-confidence shifts multiple times in a day. It depends what I\'m doing, who I\'m with...if I am working on something and accomplishing a lot, I feel very self-confident and proud. Somehow that all spills over to my body image and I feel better about myself. Probably more realistically about myself--because I am my own worst critic.  If I am with someone romantically I can easily feel good about myself, if not great about myself. Love is quite a tonic. There was a bit on  Xena  earlier on in the show\'s life, wherein Xena tells Gabrielle (paraphrase) that she can only believe in herself when someone else believes in her, first. Written out like this it sounds rather sad, but I can relate to that. Having someone fancy me makes me feel like maybe I\'m decent human being material after all. '),(5297,3853532,'23,July,2004','2004-07-23 00:00:00','    Many people tell me that one symptom of depression is a lack of libido. For it, it\'s always been the exact opposite. When I\'m very depressed, I am hypersexual, because giving another person pleasure makes me feel good about myself. I get a sort of high from it, not unlike being buzzed on drugs or alcohol.  As a general matter of course I\'m a sexual/sensual being. I love sex and I love making a man feel good. Even when I\'m not depressed, nothing makes me feel sexier than to watch a man\'s face when I\'m doing something naughty to him.   But sex, to me, is only good when it\'s with someone I care about. I can\'t derive the same boost of self-confidence from attempting to please someone I know nothing about. It just feels hollow and lacking in any substance.  When I am not depressed, sex becomes almost a holy act, for me. I can escape into a \"sexual trance\" where the only thing that matters is what I\'m doing and what\'s being done to me, focusing entirely on the union of my lover\'s body with mine. Being in that moment, sharing something so personal...it really is the stuff of dreams.  I haven\'t been in a deep depression in a long time, and I am making so many positive changes in my life, I don\'t think I\'ll be sliding down that slope again anytime soon. Hopefully, if I ever do, I\'ll be in a relationship with someone who can recognise the depressive tendencies in me and help me to nip it in the bud (either by himself or by encouraging me to seek therapy) before it gets out of hand.  Meanwhile, as I continue to make positive changes and sound steps forward, I am finding my self-confidence creeping towards a better place. The support of loved ones helps immeasurably, but at the end of the day, I\'ve had to be responsible for my own happiness--or misery--and now that I am starting to attain some joy, I\'m working even harder at doing what must be done to hold on to it. '),(5298,3853532,'03,August,2004','2004-08-03 00:00:00','    Here is a picture of me at my thinnest. I was 15, 4\'11\", and 115#. You really can\'t see it in that picture, but that sweater was literally just hanging off of me in bags, as were my jeans. This is what I\'m  supposed  to look like according to all of those wonderful-happy charts. Gag me with a spoon. (can you see the circles under my eyes?) '),(5299,3853532,'02,August,2004','2004-08-02 00:00:00','    So I was tooling around on MSNBC.com\'s Health news section the other day and came across this little bit of news:  \"\"Anyone with a body mass index (a ratio between your height and weight) of 25 or above - thatâ€™s someone, for example, who is 5-foot-4 and 145 pounds - is considered overweight, according to the National Institutes of Health. Anyone with a body mass index of 30 or above - such as someone who is 5-foot-6 and 186 pounds - is considered obese.\"\"  Isn\'t that happy news??  Honestly, I sat here and cried. It made me think of me waaaaay back when I was a teenager (20 years ago). An anorexic teenager. I was anorexic because I never felt like I was good enough. I was always too fat, too short, too stupid, too something, too anything, too nothing.  I feel bad enough about weighing 220# as it is, although I\'m not ashamed to admit that I\'m that heavy - people look at me and can\'t believe I\'m that heavy; I must \"carry it well\" or something. But then I read articles like the one I quoted,and I end up feeling completely devestated. According to all the so-called \"health charts\" someone my height is supposed to way somewhere in the vacinity of 110-120#. At the height of my anorexia years, I weighed 115#. I looked awful. I looked like I had one foot in the grave. And I was *not* healthy. My freshman year of college, my swim teacher/coach did that body fat thing they do with callipers, and I came up with *negative* body fat; there wasn\'t anything on me to pinch!  Reading this made me think of Dina\'s post about self-image. The American woman is bombarded almost constantly with images and sound-bites about diet and appearance, or at least it seems that way. Ultra-thin women on magazine covers, in advertisements. Advertisements for weightloss supplements and diets, 99% of which never work.  I honestly think that the biggest reason I\'ve had trouble losing weight, aside from not getting more exercise, is that somewhere in the back of my head, I\'m terrified of becoming anorexic again. I admitted that to myself a couple of years ago and erroneously thought, \"HA! I admitted it, now maybe I can lose weight!\" Nope. I\'ve gained 15# since then. And lost it. And gained it back. And lost it. And ... well, you see ... :sigh: Too, I\'ve heard quite often since Tayler was born 13 years ago that women who have c-sections have a hell of a time losing weight, especially when they\'re given a vertical cut like I was. Maybe I\'ve stuck that into the back of my head and my subconscious hangs on to that for some reason.  I have no idea.  I do know that at one point I lost 75#. I\'m not sure how. I do know, though, that 40# of it came right back.  I don\'t know what it is. I don\'t overeat; in fact, if anything, I don\'t eat as much as I should. I eat an apple with every meal. I drink almost two gallons of water every day. I don\'t drink soda anymore. I pay attention to what snackfoods I eat - I don\'t eat very many of those anymore, and I make sure I don\'t eat more than the given \"serving size\" listed on the nutrition panel (sometimes I don\'t even eat that much). I don\'t eat a lot of candy, but what candy I do eat is the new kinds which are sweetened with Splenda. I\'ve cut my caffeine consumption down by 95% - something I thought I\'d never do nor ever be able to do. I stopped smoking this time last year; I don\'t even crave cigarettes. I do Yoga every day, and I bellydance like a maniac once a week. I\'m supposed to take a walk every morning, but I refuse to do so during the summer while it\'s 90* out; with the climate where I live, that\'s a good way to kill yourself, and I *did* give up being suicidal a long time ago. :sigh:  At this point, I honestly don\'t know what I\'m doing right and what I\'m doing wrong. I had a long talk with both of my doctors - once in April and once in June - and they both agree that I *look* good, that losing weight shouldn\'t be my focus, that being healthy should be. Honestly, right now, my only huge concerns are my blood pressure and my on-going, continuing problems with panic disorder; and truly, everybody seems to agree that my blood pressure problems are largely due to stress issues. (which sets off my anxiety problems ....)  Anyway. I\'m glad this blog is here just for this, and that we\'re here to help and support each other. That means a lot. '),(5300,3853532,'09,August,2004','2004-08-09 00:00:00','    In 3 weeks I\'ll be leaving for Paris. Tanning, primping & preening, getting myself into some sort of shape (whatever shape it may be, remains to be seen). Three weeks! Cannot possibly be coming that quickly...can it?  I can\'t help but think back to the AbFab season 4 episode where Edina decides to go on a crash diet to lose weight quickly. (Though that\'s hardly what I\'ve been doing for the past few months, there\'s very little \"crash\" about it...) Quote: \"In three weeks I want to be on the cusp of organ failure.\"  Good old Eddie. '),(5301,3853532,'08,August,2004','2004-08-08 00:00:00','    The \"BodyFlex\" exercise routine marketers just plain got greedy. And it\'s a shame, too, because they have a product that I think genuinely works.  When I bought the BodyFlex videos & isometrics bar a long while ago, they weren\'t making any outrageous claims. They were simply telling the facts: increased oxygen into your body is good! Add resistance for muscle building and you\'ll have a sure-fire way to fight flab. I had been practicing pranayama for some time, and it seemed like a natural thing to combine both structured breathing and physical exercise. Plus, isometrics just feels  soooooooooo good . (As does deep breathing!)  I was immediately thrilled with the BodyFlex program and loved the isometric \"gym bar\". It\'s a little wonder. And I liked the sensbility of it all--no hype, no pressure, just a lot of positive reinforcement and good vibes. Greer Childers, the spokeswoman, has a very gentle way about her. Her little words of encouragement are quite nice, despite the fact that she looks  exactly  like Barbie, right down to the unnatural proportions. She\'s still huffing and puffing away whilst doing the exercises with you and that\'s kinda neat.  Plus, you do the entire routine sitting in a chair, but at the end your heart is pumping, you\'re feeling the burn, and you\'re definitely aerobic--no doubts about it. It\'s cardio in its own oxygenated way.   Then the marketers decided to try the tactic of \"lose 7-14 inches in 7 days without dieting\", which is ludicrous. Of course, the FTC sued. I don\'t think the BodyFlex program is still being produced, which is a pity. If they hadn\'t tried the gimmick, they\'d still be selling the product--and at not an inconsiderable profit (but people would be feeling good about it instead of feeling cheated).  FWIW, I did not lose 7-14 inches in 7 days, but I did lose a combined total of 25 inches (from all measured points) in the first month, just doing BodyFlex by itself. I still do the routine and still like it very much. It, alone, is not enough to cause substantial weight loss, despite my initial dramatic results (they were not sustainable). It takes proper nutrition & diet, lots of water, and I find that if I do yoga after the BodyFlex routine the results are much more permanent. It\'s always a combination of things.  Nowadays it\'s hard to weed out the genuine treasures from the hyped-up junk. Every company seems to be trying to one-up the competition with increasingly incredulous claims. The FTC must have several full-time employees doing nothing but sueing fitness companies. But I consider \"BodyFlex\" to be a gem. YMMV, of course, because exercise is nothing if not utterly subjective!  I do not think I would have purchased \"BodyFlex\" had the claim of 7-14 inches in the first week been front and centre when I bought it. I don\'t recall that phrase being bandied about on the web site or infomercial. I wonder when the parent company decided to jump on the hype bandwagon? At any rate, there is no magic bullet for weight loss, and every time we try a gimmick & fail, we just run the risk of feeling worse about ourselves (\"why does this work for them, but not for me?\").   I try to steer clear of any product that raises the \"too good to be true\" red flags. But if you like breathing exercises and you like stretching & resistance training, look for a BodyFlex kit on eBay. You\'ll want the instructional videos and the \"gym bar\" (sometimes referred to as \"BodyFlex Plus\"). Just don\'t expect to lose 7-14 inches in 7 days--I\'m not sure that\'s possible unless you\'re dead! '),(5302,3877932,'29,June,2004','2004-06-29 00:00:00','    Some times it’s amusing to see what web sites look good and which do not. It was really interesting when looking for a college all the design schools had really bad web sites. Later it became obvious that no school uses its design department to do actually design work for the school. Possibly this is just because the accounting department will want to have a look at the books as well but even still students have always seemed like a great work force on unleash on a school practical problems.  Just now a trip over to  urlLink W3C  was very shocking. No, they do not have a new site design, but after looking at websites about site design, and ones that look very good at that, it was amazing to see how badly the W3C is designed. It’s literally painful to look at as it doesn’t look good and seems to be haphazardly organized.  This might not seem like a big deal to some people but for those who don’t know the W3C have taken it in to their hands to create standards for HTML, CSS, XML and the whole forest of acronyms the Internet has grown. The W3C are, literally, the experts in these languages and maybe they could get a couple of volunteers to spruce the place up.    '),(5303,3877932,'27,June,2004','2004-06-27 00:00:00','    urlLink Adaptive Path  has a look into why some design firms give their web sites to out side statistics organizations to discover if what they made is useable. They argue that usability should be kept to the design house with people who knows what is trying to be achieved and will be able to come up with more innovative solutions to problems as they arise.  It is always interesting reading about industry standards and practices. At the esteemed  urlLink TCNJ  websites are not a large part of the curriculum. Eeking out two classes of web was a bit more maneuvering then one would like. Though out of those two classes there is no talk of usability testing by Q&A and anonymous tests there is testing as you go. Though this mostly comes out of their practices in other classes where critiques are on going in the creation process and simply flows in to the other classes. Unless this boring, sterile practice better suited to non-designers was discussed through the first few sleepy days of Web Design 1, ALT 255 does not remember such ways to discover usability.   This method cannot be used too much as why would the client care as long as it does what they want. It will be a sad day when they start asking to see the numbers to back a web site up.    '),(5304,3877932,'27,June,2004','2004-06-27 00:00:00','    Like Mark Twain ALT 255 trys to never let school get in the way of education. Graphic Arts is something that definitely falls in that category. As a growing graphic artist ALT 255 is an experiment to constantly learn about graphic arts. Poke it, prod it, or pull its pigtails ALT 255 will do what it takes to try and figure out what graphic arts is and where it might be going next.  And if ALT 255 is really lucky, maybe some one will even read it and learn, form an opinion, or just get angry. So let the experiment begin.    '),(5305,3877932,'30,July,2004','2004-07-30 00:00:00','    Well I think it’s official. When it comes  urlLink to   urlLink the   urlLink leaders   urlLink of   urlLink industry  they web has been standardized and so have a large portion of their followers. I include my self in that category even though I have yet to finish standardizing this site to begin with. I have the want to standardize to have my peers in web design accept me.   This  urlLink web site review by François Briatte  shows that after standardization we have resolved to nit picking. 80% of the links are underlined. How does that effect usability?  urlLink Jakob Nielsen  has the answer.  Since when were we all in the  urlLink Bauhaus  with  urlLink Paul Rand ?   Not that it’s bad at all as simplicity in design is a wonderful ideal to try and achieve. Yet it seems like everyone is going for the same goal.  It was the great and legendary  urlLink Fanky Chak  of  urlLink The College of New Jersey  that really introduced me to the diametric opposites of  urlLink Paul Rand  and  urlLink David Carson . Both have such opposite styles yet archive such visual beauty. Where are the Carson’s of the Internet? Where is design communicating while still obscuring type?  Perhaps it is just that I feel a strong Carson influenced site would not score high on the usability list. I’m not sure that is true, but as I try and think of ways to mess with type and integrate it with images while still keeping standards intact I am having a hard time.  urlLink David Carson’s own site  doesn’t seem to help as I still have not figured out how his site works.  With the Bauhaus seeming to come in to full-blown swing will we see a revolt? Will people throw some standards out the window and challenge users to think? We can only wait and see.    '),(5306,3877932,'29,July,2004','2004-07-29 00:00:00','    The Pathfinder pallet is one of those Illustrator pallets that unless explained to you it can be a little difficult to figure out how it works. The Illustrator help file gives a basic over view of how it works but just by experimenting it is not always obvious what the tools in there actually do.     One of the most useful tools is the Subtract from Shape Area tool. This tool takes the front most object and subtracts it from the backmost object when 2 objects are selected. This allows anything in the background to show through. This is important if any white areas in the illustration are just supposed to be whatever color the background is instead of the ink color white.'),(5307,3877932,'28,July,2004','2004-07-28 00:00:00','    As a preliminary post, I’d like to discuss an intriguing aspect of design, one which is often overlooked and easily forgotten or even unnoticed and unappreciated.  As a graphic design student, I often look to various sources for inspiration for designs, keeping an eye out for aesthetically pleasing devices.  Sources include nature, machinery, and music, but also in the area of film.     Most often, people go to the cinema, watch the film, and toward the end, as soon as the “Directed By…” text flashes across the screen.  It’s up, leave the spilled popcorn, and out to the parking lot to find your car.  There’s little care for those talented individuals who actually put their time and effort in to make the visual and audio craftsmanship that goes into making a motion picture.  Yesterday, with my sister, I had the good pleasure of seeing Alex Proyas’s unique Sci-Fi thriller I, Robot.  While the movie was exceptional, in its style and pace, one of my favorite parts was the end-credits!  Sure enough, the first credit displayed, the music came to a crescendo, and the theater was cleared before the cast list even finished crawling up the screen.  My sister and I sat, giving credit where credit was due.  It was in those few moments in reading the many Visual Effects companies and their workers that I noticed the typeface used in the credit text.  Such a remarkable font, very different from what I’d seen elsewhere, I debated whether the font was already in existence or was it created specifically for the film.  Colored in an electric blue, the font was sans serif, but in all caps.  Thin simple lines, but not condensed.  Graceful and pure, it appeared to have been carved through laser onto the screen.    The typeface reflected the futuristic minimal nature of much of the film’s motifs. The set designers said it right when they suggested, “We’re making a documentary of the future,” not the typical silver and blocky nature often seen in other Science Fiction films.  The film itself had remarkable visuals, very different from other Sci-Fi genre films, such Star Wars or Trek, but more like that of Minority Report with a mature appreciation of how aesthetics would develop with the increased use of technology.  Given the fact that the movie is only set 30 years into the future, we see a connection between design of our Modern Day, and let’s say where design would go 50 years down the road.  I thought it was interesting how, as our modern culture shifts to a more minimalist idea of design, this movie takes the idea that we continue on that route, but also take a step back into detail for the sake of interest.  This reflects the movie’s theme that humans can try to be more efficient and robot-like, but still revive our humanity through personality and taste.  Now I want to know where I can get a model of Will Smith’s “new-hotness” Audi!    '),(5308,3877932,'22,July,2004','2004-07-22 00:00:00','    I enjoy finding new sites and promoting them. The hardest part of any new site is getting the word out and getting people to become regular visitors.         urlLink Funfurde  is a good collection of new and interesting furniture design. The design of the site is oddly reminiscent of the clean  urlLink gizmodo  design as seems to be popular these days. I’m not really a fan of overly wacky furniture designs though they are fun to look at.      Good luck to Funfurde!    '),(5309,3877932,'22,July,2004','2004-07-22 00:00:00','    urlLink          While I am more liberal then conservative I’ve become very interested in these conventions. They really are a great place to see some truly heartfelt speeches from strong speakers. I look forward to the Republican convention for more strong speakers.                One thing is clear though. That is one large ass screen. And for such a large screen they really don’t seem to be doing much with it. With the speaker featured prominently in the center there is a lot of space on either side. Of course there is the waving flag in the background and the tried and true color swirls. These are all good background fillers but for such a large screen they should be getting more for their money. Some sort of graphs or montage to highlight the speeches given could have been good to fill the space.               I’m sure they did not want to take the focus away from the speaker but it seems there could have been some sort of festive graphics to get the crowd going. What would you have put on the giant democratic screen?    '),(5310,3877932,'22,July,2004','2004-07-22 00:00:00','    urlLink Internet Explorer has started a Team blog . Larger corporations are definitely taking advantage of the blogger medium and what they post shows their corporate culture. This blog seems to be pretty energetic&nbsp;and should offer some good discourse. Does any one else think that  urlLink google\'s blog  is a little reserved? Little too much corporate polish for my taste.   IE\'s new team blog allows comments (something google does not) and I imagine they will get riped apart by some people but in the end it will probably very beneficial. That is if they can abstain from flaming back.   **Update**   urlLink jessey.net &nbsp;points out that the site does not even  validate as HTML 4.0 Transitional .    '),(5311,3877932,'21,July,2004','2004-07-21 00:00:00','    We here at ALT don\'t get much traffic. It\'s a new site and I\'m still trying to convince some of my friends and colleagues to throw in their 2 cents. We\'ll see if that ever happens. I seem to be adding content fine now but we\'ll see when my senior year of college starts with my other commitments.  urlLink Concerts  don\'t plan them selves you know.    Anywho, when some one does me the pleasure of stopping by I notice on  urlLink SiteMeter . A site called  urlLink Design Detector  stopped by and poked it\'s head in. Seems fairly new and any site that amasses design sites so I don\'t have to look for them gets an A in my book. So from one new site to another I welcome thee. It seems they don\'t offer a category for blogs. If you stop by put in a request for a category. Us bloggers can never be ignored for long.    '),(5312,3877932,'20,July,2004','2004-07-20 00:00:00','    Who knows what is going on with the next iteration of Internet Explorer. Not to long ago, due to a post on  urlLink Design by Fire ,  urlLink Robert Scoble &nbsp;has said he would pass along some feature requests to the staff of IE. It seems Microsoft is being much more open about IE and are trying to talk to people to find out what they want. This is wonderfully as the word on every web designers mouth is “standards are god” except that IE seems to be an atheist&nbsp;since their standards support sucks. Yet even with all this openness the next I read is that  urlLink new versions of IE will only be released with new a OS (operating system).   &nbsp;  For those of you that’s are not rolling on the ground like they just got a really good kidney punch, let me explain. Internet explorer is up to version 6. Yet people still run internet explorer 5, or even 4. This is because it works fine for them and they see a need to change. Web designs know these people are out there and jump through hoop after hoop so these lazy people can view the web page in a pleasing way. This tends to hinder advancement of the web because every one is dumbing down their web site to work with older versions of browsers.  &nbsp;  You though people didn’t update their browser before, how about IE 7 and above which, to update your browser, will need a OS update that will cost over a hundred dollars. Corporations that run hundreds on computer will have to pay hundreds of thousands of dollars. Well browser update will grind to a halt. I know people that are still running Windows 95, people upgrade even less when it costs money.  &nbsp;  The hope is for any one reading is that every one will migrate over to  urlLink FireFox , or the  urlLink Mozilla  browser of their choice. I don’t see this happening until  urlLink Dell  starts shipping PCs with it installed. It’s a combination of laziness and simple computer illiteracy of the masses that will keep people stalled with whatever browser is on their OS.    '),(5313,3877932,'20,July,2004','2004-07-20 00:00:00','    urlLink Andy Budd  has an organization going on called  urlLink Skill Swap . It’s an ingenious idea where design type people get together and teach each other what they know. Some cool People over at  urlLink Eris Free  pointed me to the wonderful idea.   I don’t know what I could teach others. It seems I am still at the point where I know little about many things but a lot about nothing. Even still I wanted to check it out and see what they are doing.   Being new to the design bogging sphere I am not innately aware of people’s locations.  urlLink Eris  boldly pronounces she is from Dallas Texas, but then who from Texas doesn’t boldly pronounce everything?  urlLink Andy , it turns out, is from  urlLink Brighton which is about 50 or so miles south of London . This way to far for this New Jersey native to go.   This got me thinking about designers and location. Surly there are designers in the New York/Philly area that blog. While having a world wide group of people to read is amazing it would also be amazing to be in contact with the blogging savvy in my immediate area.   Just poking around I have found little. I can find most anything on the web and it is only a matter of time before I put the puzzle together and find them. Who else has a group of local designers they look to? I can’t be the only one looking for local camaraderie.    '),(5314,3877932,'18,July,2004','2004-07-18 00:00:00','    urlLink Graphic Push  has an article about  urlLink taking PowerPoint design seriously . While many do not want to think about having to put together PowerPoints as a living it\'s still good to know the standards and practices of any graphical tool.    With PowerPoint seemly everywhere how does this effect how information is seen? After just putting a PowerPoint presentation together for a school it’s very obvious how painful it can be to animate anything in that environment.    With Flash MX starting the way to be the next PowerPoint (at least in their eyes) will Microsoft set up and start adding some features that will take PowerPoint to the next level? It sure would make any designer using it happy if the damn thing was easier to make exciting. Those animations are painful.    '),(5315,3877932,'14,July,2004','2004-07-14 00:00:00','    As a designer, we learn visually. No matter how much CSS, XHTML, or PHP a designer programs in the end we are visual people. With this in mind one of the many tutorials you can find online are video tutorials. These really are great inventions that quickly explain how too work the program in question. Most of the time they just cover the basics but for anyone trying to learn a program for the first time the basics are often the hardest thing to learn. Some of the best ALT knows about are  urlLink videos for Milk Shape 3D by Brian Hayes .  The thing about these video tutorials is that they were put together by people who not only knew the program they were talking about but how to record and capture these films. So, not to put them down, their voice was a bit dull and boring. Though no one was looking for anything but a clear voice to explain what to do. Enter  urlLink Designs By Mark . These video tutorials are not free but there are a few examples on his site. Give one of them a whirl and a wild and crazy man talking about Photoshop will quickly jar you awake.  Who is this guy marketing to? Designer do not want frills when they are looking for an easily understood answer are they? I can’t imagine who is buying these but I’m going back to the soothing voice of my techie friends.    '),(5316,3877932,'09,July,2004','2004-07-09 00:00:00','    Designing any web site should be a learning experience, period. If you\'re just churning out the same web page every time it\'s safe to say you\'re not really challenging your self or the rest of the web. While ALT is never going to assume we are challenging the web in any way with our masterful layout but we do grow stronger with each attempt.   This time more advanced stylesheets was the attempt. CSS is wonderful unless you want it work in more then one browser. And more then one browser it will be viewed, as graphic artists are stubborn people that still use MACs, ug.  Anywho… to get it right we went to  urlLink Position Everything  and grabbed this very nice three column layout to get out selves started. It is a beautiful piece of work with all the hacks built right in. Never the less it still took a couple of hours to get it looking right, or at least to have it not look bad. It was much easier to test with a nice version of  urlLink FireFox  on the computer, since with all the hype about it how can you pass?  As is tradition the problem that came up was matching the image color with the background. There are two choices with this, go transparent or fix the color. Transparence is no picnic either and also with bugs from browser to browser. One-day  urlLink .png  will be a fully supported file format across the board, but not today.  Already having tried to get transparency to work, it was now time to, once again, try and get the colors to match. It was always baffling why Illustrator would say it\'s the right color and Photoshop to say it too but for the browser to disagree so adamantly. Having never really come across the answer, the decision was to just dive in and fix it.  It seems that when the images were brought in to Photoshop from Illustrator they loose their contrast. Image > Adjustments > Levels…   is the way to correct this off set. This handy tool lets you set the black, white, and gray of the image. This seemed to fix the problem and when I refreshed my page the colors matched exactly.    '),(5317,3877932,'07,July,2004','2004-07-07 00:00:00','    There is  urlLink an article  over at The Atlantic about campaign ads and why they are so bad. People are not amused after they see these dry, uninteresting, and repetitive ad that have no right to be on TV. The real question is of course, why are multi-million presidential campaign ads so incredibly bad?  The article goes in to how fearful candidates are of messing up and how consultants, campaign managers, and even the candidates insist on having creative input. With reaction to ads measured and plotted it is boiled down to a science. Earlier today on  urlLink MSNBC  one of the men involved touted that it was a science and talked about the amount of user testing they do with slightly different versions of the ad. ALT does not know who is being subjected to political ads with a slightly different accent on that third paragraph or that different angle of the candidate in his hard hat, but we would like to extend our sympathies.  While John Kerry and George W. Bush are off metering the lights on their next commercial to get statistically good lighting, where are the designers that are actually making these atrocities? This just plays up to  urlLink an article  that was in Communication Arts but is featured on AIGA\'s own web site, which say that graphic artists need to design with the confidence that we know what we are doing. This is no reason any ad need to be committeed to death.  The is a light on the horizon though as MoveOn.org as well as other online sites have be pioneering privately made campaign ads that have design and idea in the forefront and statistics in a far second. Might candidates squelch at ads that are made better then their own? Perhaps, but in an age where many political decisions are based on polls and stats it is not looking likely.   Designers need to take responsibility for the drab boring ads they make and need to take charge in the creation process. The candidates have no reason to change their ways, but if the industry changes as a whole and enforces design principals in all they do candidates will have little lass to do then follow, kicking and screaming, to a future of interesting ads that attract attention and create a buzz.    '),(5318,3877932,'02,July,2004','2004-07-02 00:00:00','    Here at ALT, we are not afraid to stretch a bit to make out content fit our mission. Though it\'s really not that much of a stretch it\'s better to warn you in advance so you don\'t get all riled up later, *wink*.  After just finishing, after a long day, reading a 35-part retelling of the historic  urlLink Clone Saga from the pages of Spider-Man . Something that many people know about but, it seems, few actually know the whole story. This article seems to have been written in 2001 so it\'s not a new endeavor but it would be rather painful to try and read this long article if you had to wait for their weekly/monthly publishing.   It was an amazing read as rarely are comic book arcs completely represented out side of comics. This is more then a retelling though as Glenn Greenberg give long, detailed, and painfully interesting tales from behind the sense. As much as this is a story about compete chaos and corporate politics, the behind the scenes that show how comic books really come together really show through.   What is really staggering is, honestly, the staggering size and detail of this history of Ben Reilly. Enough detail is given that Ben becomes an endearing character that you care about. This will make you long for these books to be TPBed so you can experience this great character. ALT can do little more then personally thank the two wonderful writers of these articles.    '),(5319,3877932,'12,August,2004','2004-08-12 00:00:00','    In the UK two companies have started a  urlLink fight over the color orange . The cell phone company Orange and EasyJet’s new company EasyGroup both use orange as their primary and only brand identity color. The article goes over how several companies have defended specific colors when it is identified with a product or identity and might cause confusion to the consumer. While legally they do not “own” the color they effectively block out the color use from their competition in a specific market.  While it makes sense that a company would want to protect it’s identity, it would be interesting to see a situation where all the color choices were taken. Eventually issues like this will have to come up. If  urlLink GE decides it really likes all of it’s colors  and that costumers respond to them, are they going to try and defend all the colors? I’m sure it’s an unlikely situation to ever happen but it would be an interesting legal president if ever all the colors in the rainbow were defended as an identity.    '),(5320,3877932,'10,August,2004','2004-08-10 00:00:00','    Well, this is a bit more of a question than some design revelation, so bare with me please. I\'m currently doing some design work for the company where I have my summer job (big note: it is not a design company in any way). The work I\'m doing is purely for the reason that I\'d love to get any work I can to use for a portfolio and they\'re a bit too cheap to go to a design firm, either way, good for me. The problem is while I\'ve been given the information and the go-ahead to design a product catalog, the many people that are reviewing it have some very odd preferences to what it should look like. Some suggestions are fine and don\'t bother me one way or another, but certain suggestions are very off the wall and I think to be very bad ideas for design (no specific examples except they seem to want me to use Word Art effects for headings). I have one coworker who at least has a good design sense that I discuss my work with, but otherwise, no one has any design experience. Does anyone have any suggestions? Do I try to convince these people that their ideas are not good design (I don\'t really feel like confronting them since they seem to be very set on what they think \"looks good\") or should I just pander to them and do whatever they want, risking the fact that I may not want to use this as portfolio work? Well, just needed to get this out there and hopefully this little venting will prevent me from killing someone with a mouse and keyboard.    '),(5321,3877932,'03,August,2004','2004-08-03 00:00:00','    As this site tries to find it’s legs and create a voice for it self within the online community the format of the posts will change. I really didn’t want it to. I wanted to set up a professional style and get some Google Ad Sense going. Well I’ve given up on trying to keep a consistent style for now and I’ll just let it flow till we all find a style that works and excites us.   urlLink Speak Up  has a post by  urlLink Peter Scherrer  about  urlLink making deadlines . I have always been happy that I have finished my projects for class in whatever allotted time I have given my self. I know this bodes well for when I have clients and have to set and meet deadlines. Of course the article made me think of Jimmy Casa who never seems to make a deadline and rather have a perfect project then an on time one. That’s right Jim, I’m calling you out. So hopefully you’ll post some sort of response so when your future employs you he will read your wonderful explanation along side this scathing comment. I doubt you’ll have to much trouble explaining it away.  Though it might be unfair to call Jim out as he is the only one of the Contributors on the left that has actually fulfill that title.  p.s. it’s early and I’m a little cranky.    '),(5322,1729882,'30,May,2004','2004-05-30 00:00:00','    I am going to start exercising more.  My doctor has been trying to get me to do this regularly for awhile.  My wife has bugged me about it too.  Now that school is out, I\'m going to attempt it.  I want to start walking 1 or 2 miles a day at least 4 times each week.  I also want to start lifting some light weights at least a couple of times each week.  I hope this blog will keep me at it by updating my progress.  Cross you fingers.   '),(5323,1729882,'28,May,2004','2004-05-28 00:00:00','    urlLink    Two of my favorite things:  baseball & food&nbsp; urlLink   '),(5324,1729882,'28,May,2004','2004-05-28 00:00:00','    I was reading the paper this morning when I came across a letter to the editor.  The basis of the letter was that anyone who is opposed to the war in Iraq or even the atrocities at the prison should get out of America.  \"America, love it or leave it,\" sound familiar?  It should if you know anything about the war in Vietnam.  It went on to resurrect the ghosts of 9/11.  \"They\" caused the horror of the WTC and at the Pentagon and \"they\" can only get what they deserve.  I heard similar arguments when the story of Mi Lai (spelling?) came out.  Nobody can tell me that history doesn\'t repeat itself.  Iraq is this generations Vietnam.  We shouldn\'t be there, and we have no way of winning.  I don\'t want to see another wall with a list of the names of the dead who died for no reason.  Don\'t get me wrong, one thing is going to be different about this war.  There won\'t be any \"blame\" forced upon the soldiers who are doing their duty.  However, wrong I see this war, these soldiers were given orders to carry out and have done so admirably, despite a few rotten apples.  You can no more win a war than you can win an earthquake. --Jeannette Rankin   '),(5325,1729882,'27,May,2004','2004-05-27 00:00:00','    I\'m going to scream if this rain doesn\'t stop soon.  This is killing our baseball schedule.  I know farmers like rain, but enough is enough.  Give me a week of sunshine and warm weather.  If you have an extra million dollars you want to get rid of, I\'ll take that too.  I\'m not picky.   '),(5326,1729882,'25,May,2004','2004-05-25 00:00:00','    Both sons had their last day of school today.  #1 graduated from 8th grade, made the honor roll, and got a Presidential Academic Award.  #2 graduated from 5th grade and also made the honor roll.  Despite what I\'ve said before, they are turning out to be pretty good kids.   '),(5327,1729882,'23,May,2004','2004-05-23 00:00:00','    #1 is fourteen years old and as they say, I\'m not sure he\'s going to make it to fifteen.  One minute he is depressed for whatever reason, the next minute he thinks he is God\'s gift to the \"ladies.\"  He argues with his mother, his brother, and me.  He starts \"going out\" with a girl and two weeks later he breaks up with her.  Working hard for something was always his way of doing things until lately.  Now, if his mother and I would let him, he spends all of his time on the computer.  Yes, we do have time limits set.  I\'m hoping this is his early teen phase and will outgrow it.  SOON, I hope!  Love him, just don\'t always like him.   '),(5328,1729882,'20,May,2004','2004-05-20 00:00:00','    #2 pitched five innings of scoreless ball, drove in the go ahead run, and scored the winning run.  Not bad for a day at the \'ol ballpark.   '),(5329,1729882,'18,May,2004','2004-05-18 00:00:00','    A positive attitude may not solve all your problems, but it will annoy enough people to make it worth the effort. --Herm Albright   '),(5330,1729882,'17,May,2004','2004-05-17 00:00:00','    First of all let me say this:  I love my wife with my whole heart!  But I also want to say that sometimes she makes me want to  SCREAM!!!!!!    YYYYYAAAAAAAHHHHHHHGGGGGGG!!!!!!!!!!   '),(5331,1729882,'16,May,2004','2004-05-16 00:00:00','    One of the weaknesses of our age is our apparent inability to distinguish our need from our greed. --Author Unknown   '),(5332,1729882,'15,May,2004','2004-05-15 00:00:00','    Nothing much today.  My #2 son played a baseball game.  We won 3 - 1.  He still didn\'t have a hit and he played second base the whole game.  No pitching today.  \"A\" pitched all seven innings.  I hope that doesn\'t come back to haunt us later in the year (worn out arm).  #1 son plays tomorrow night.  I have to tell his coach that he will miss a game the 26th.  He won\'t be happy.  Such is life.   '),(5333,1729882,'15,May,2004','2004-05-15 00:00:00','    Every parent of every child thinks that child is the star of the team.  Someone once told me that and I have no doubt that it is true.  I know that I feel that way about my 2 sons.  The difference between parents is that I refuse to yell and scream at a coach if they do not play my sons the way I think they should.  When a player is just learning to play baseball let\'s say, they need to learn the game and play as much as possible in a lot of different positions.  But after playing for 8 years, there are some players that don\'t belong in certain positions.  They can\'t handle the ball and they might get hurt.  No one wants to sit on the bench the whole game.  I think that they should play somewhere every game, but better players deserve to play more.  We have become so concerned with not hurting someone\'s self-esteem, that we forget that is exactly what we do when a   team   loses because a poorer player makes mistakes that a better player might not have made.  Parents need to lighten up and let a coach do his/her job.  We won\'t always agree with everything they do, but that\'s what makes it fun.  Arm chair quarterbacks and second guessing is the best part about sports.   '),(5334,1729882,'14,May,2004','2004-05-14 00:00:00','    I don\'t know half of you half as well as I should like; and I like less than half of you half as well as you deserve.   --J. R. R. Tolkien   '),(5335,1729882,'13,May,2004','2004-05-13 00:00:00','    #2 son had a baseball game tonight.  He started the game as pitcher.  Pitched 3 innings of no-hit ball.  Awesome huh?  He is depressed because he can\'t get a hit.  He wants to hit the ball so bad he can taste it.  He is pushing too hard.  As far as playing the field goes, he is one of the best players and is  the  best pitcher, but he can\'t get that hit and it bugs him.  What can I do?   '),(5336,1729882,'11,May,2004','2004-05-11 00:00:00','    Success usually comes to those who are too busy to be looking for it.     --Henry David Thoreau   '),(5337,1729882,'10,May,2004','2004-05-10 00:00:00','    Day after my birthday.  Thank God that\'s over!  Had batting practice for my son\'s team today.  Why is it that kids that haven\'t been hitting the ball refuse to make any changes which might make them better hitters?  Uncoachable is the word I would use.  Today\'s quote applies here I think.   '),(5338,1729882,'10,May,2004','2004-05-10 00:00:00','    urlLink    Family Picture&nbsp; urlLink   '),(5339,1729882,'10,May,2004','2004-05-10 00:00:00','    Don\'t worry about the world coming to an end today. It\'s already tomorrow in Australia.    --Charles Schultz  I know not with what weapons World War III will be fought, but World War IV will be fought with sticks and stones.    --Albert Einstein   '),(5340,1729882,'10,May,2004','2004-05-10 00:00:00','    Been there, done that for 48 years, wish it would stop rearing its ugly head.  I don\'t want to be dead, I just want to stop observing birthdays.  I get pissed off if people don\'t remember and I get pissed off when people make a big deal out of it.  Sounds like I get pissed off a lot, doesn\'t it?   '),(5341,1729882,'30,June,2004','2004-06-30 00:00:00','    Surprised my brother and his wife on their 40th wedding aniversary.  The whole family was there.  My sons got to meet cousins they have never seen before.  All my sister-in-law could do was scream for about 5 full minutes!  We had a blast.   '),(5342,1729882,'22,June,2004','2004-06-22 00:00:00','    I\'ve always wanted to go to the Boston area.  My wife has to go next week for a few days to work.  I don\'t get to go.  The boys and I get to stay here.  Isn\'t that a scary thought?   '),(5343,1729882,'17,June,2004','2004-06-17 00:00:00','    OK, laugh your ass off:  \"I\'d rather be a could-be if I cannot be an are; because a could-be is a maybe who is reaching for a star. I\'d rather be a has-been than a might-have-been, by far; for a might have-been has never been, but a has was once an are.\"  -Milton Berle   '),(5344,1729882,'14,June,2004','2004-06-14 00:00:00','    #1 son\'s high school career started today.  He had his first Cross country practice today.  Upperclassmen were there and I think he was a little nervous about that.  Everything went great.  He enjoyed himself and actually met a few new people.  YEA!!!!  We\'ll see how things go once the newness wears off.   '),(5345,1729882,'10,June,2004','2004-06-10 00:00:00','    --#1 son doing better.  I think he got the piss scared out of him.  Handcuffs will do that to you.  --#2 son made the All-Star team, then turned around and his regular team played the worst game of the year.  --Can\'t find my drill.  --Bushes need to be trimmed - House needs a new roof - Trim needs to be painted - Garage needs to be gutted.  --Planning for next school year needs to be done.  --At the time, I didn\'t always agree with Mr. Reagan, but maybe he was smarter than I gave him credit for.  --Can\'t stand Bush, don\'t particularly like Kerry.  What now?  Not a shred of evidence exists in favor of the idea that life is serious.   Brendan Gill   '),(5346,1729882,'06,June,2004','2004-06-06 00:00:00','    Yesterday was bad enough.  This morning we got a call that our #1 son had been  arrested !  He spent the night with some friends and they decided to go out late and throw rocks at cars.  Must have dropped him on his head when he was little.  The sheriff was called, they were handcuffed, taken to jail,  and we were called.  He will have to appear in court soon and will probably have to pay damages to the car owners.  He says that he wasn\'t throwing at cars like the others were.  Unfortunately, even if true, that means squat.  The sheriff did say that he didn\'t show any attitude to them and was very polite.  Hopefully this scared the  SHIT  out of him.  His life is about to become quite different for the next several weeks (months?, years?).   '),(5347,1729882,'05,June,2004','2004-06-05 00:00:00','    #2 son had 11 year old All-Star tryouts today.  He fielded the ball with no problems.  He caught the ball in the outfield with ease.  He even hit the ball better than he has all year.  But, the one thing that he has excelled at the whole year, pitching, he didn\'t even attempt to do.  When I asked him why not, all he would say is, \"I don\'t want to!\"  Needless to say, I was livid.  He has pitched awesome all year.  He is getting the same attitude that his older brother has - If I have to work for it, I don\'t want it.  I think he is afraid that he will be facing tougher hitters in tournaments and he won\'t do as well.  His brother isn\'t even going to try out for the team because he hasn\'t hit well this year.  I have offered numerous times to take him to the cage to hit, but he never wants to.  He wants to be good without any effort.  Being the baseball nut that I am, this is driving me crazy.  I want to yell and scream, but I don\'t want to push them into something I want and they don\'t.  Don\'t know what to do.   '),(5348,1729882,'02,June,2004','2004-06-02 00:00:00','    Just got back from a workshop on teaching writing.  It was not what I wanted or hoped for.  Writing is the \"thing\" in education right now.  I am looking for a plan on how to teach writing.  As usual in education, we are given a lot of theory but no practical application.  Guess I\'m getting to old and set in my ways to figure this out for myself.   '),(5349,1729882,'01,June,2004','2004-06-01 00:00:00','    The only alternative to coexistence is codestruction. --Jawaharlal Nehru   '),(5350,1729882,'31,July,2004','2004-07-31 00:00:00','    I DON\'T WANNA GO!!!!  Taking the boys and a couple of their friends to a nearby amusement park and I\'m not amused.  It\'s going to be crowded as hell.  A wasted day, yippee!   '),(5351,1729882,'22,July,2004','2004-07-22 00:00:00','    Reality TV sucks.&nbsp; It has no basis in reality.&nbsp; I\'ll take a good baseball game anyday.  Why can\'t coaches be fair.&nbsp; I know some kids are better than others, but to play a kid 1/2 of an inning when others are playing the whole game is not right.  In keeping with the baseball theme -   GO CARDINALS!!     Bush has got to go!&nbsp; (I\'m just not thrilled with Kerry either.)   Motivation and procrastination - a winning combination.&nbsp; Especially when it comes to diet and exercise.  School starts in 3 weeks - yippee?  Everyone talked about how Steven King\'s   Dark Tower   series was so good.&nbsp; I am reading the first book and am not that excited about it.  Should I buy a Palm PDA?  We are taking the boys to the   Brickyard 400   in Indy.&nbsp; What am I getting myself into?   '),(5352,1729882,'15,July,2004','2004-07-15 00:00:00','    Jess: (over a game of Pictionary) \"Baby talk\"? That\'s not a saying!  Harry: Oh, but \"baby fish mouth\" is sweeping the nation?  --When Harry Met Sally  Andy Dufresne: Get busy living, or get busy dying. --The Shawshank Redemption   '),(5353,1729882,'13,July,2004','2004-07-13 00:00:00','    My son keeps telling me that times now are different than they were when I was a kid.  He says that I don\'t understand how things are now.  I used to argue with him.  I am a teacher and I see kids everyday in their stressed out world.  I would trade their day for my day anytime.   But,  I got to thinking(Scary huh?), I\'m comparing apples with oranges.  Sure an adults life is tougher than a kids, most people would agree with that.  But do kids today have it tougher than I did when I was fourteen?  Yeah, I have to agree that they do.  Many things are still the same especially when it comes to peer pressure.  The number of distractions that kids have to put up with today are tremendous.  My boys go from one sport right into another sport all year long.  I didn\'t even know what soccer was when I was fourteen.  His mother and I are constantly running them hither, thither, and yon.  Studying is more intense now than it used to be.  We are pushing kids to learn more every year.  All this said, there is nothing new here.  Kids today have it tougher, I had it tougher than my parents did when they were fourteen, and my grandkids will have it tougher than my sons do.  This is the natural course of things.  The way it has been and will always be.  So I agree that times now are different than they were when I was a kid.  I will continue to argue with him that I don\'t understand what it is like to be a kid.  I may be 48, but I remember what high school was like and I wouldn\'t go back there for winning lottery ticket!   '),(5354,1729882,'04,July,2004','2004-07-04 00:00:00','    Watching the movie \"Independence Day\" and Judd Hirsch looks at Jeff Goldblum and says, \"I\'m proud of you son.\"  It made me think how many times I\'ve said that to my sons.  Not nearly enough.  I think I need to change that.    If you want to be free, there is but one way; it is to guarantee an equally full measure of liberty to all your neighbors. There is no other. Carl Schurz   '),(5355,1729882,'08,August,2004','2004-08-08 00:00:00','    Just got back from Indy and the Brickyard 400.  The boys went for the first time and I think they really enjoyed it.  We got to see our favorite driver, Jeff Gordon, win.  It was a fun day.   '),(5356,1729882,'05,August,2004','2004-08-05 00:00:00','    Freshman orientation day.  My #1 son starts high school today.  Man, am I getting old or what?  I can still remember my first day of high school - almost 34 years ago!  :(      I\'m not ready for this.   '),(5357,4015298,'29,July,2004','2004-07-29 00:00:00','   Ok, well I hope to be better about posting to this than I seem to have been following through a great many other things in my life.     &nbsp;'),(5358,4015298,'23,July,2004','2004-07-23 00:00:00','   urlLink    This is a picture of my trio - it was taken in December, 2003.  &nbsp; urlLink  '),(5359,4015298,'23,July,2004','2004-07-23 00:00:00','   After months of wondering exactly what the heck a BLOG was, I came across&nbsp;one by a young man from Dallas, TX that explained it - and I was HOOKED!!     &nbsp;'),(5360,4015298,'17,August,2004','2004-08-17 00:00:00','   Wow - time really flies whether you\'re having fun or not. It\'s been quite a while (over a week, in fact) since I actually made it here to put some words down on, er - what exactly is this?? It\'s not paper - and it\'s not even necessarily in black and white - so I don\'t know quite what to call it. Cyber-perpetuity? (is perpetuity spelled correctly? I suppose I\'ll have to check now to see.)  Any-hoo, back to my dilemma. For a week or so now, every time I MADE myself find the time to come out here put down something for someone somewhere to read, I got sidetracked. I supposed I could blame it (initally) on Jen and Paula - or maybe on blogger itself - for listing soooo many other more interesting blogs to look at!  I have read so far about everything from the x-rated antics of an oversexed exec and his barely-legal office assistant to logs about a professional massage \"therapist\" who offers to see her clients through to \"completion\" to a typical Chinese college student\'s life and experiences with her boyfriend and an assorted cast (this was quite interesting as it was written in English - perhaps because the UK still holds Hong Kong, or used to, whichever). I also have read some very interesting commentaries by several conservative political pundits - which I found particularly interesting and which kept me busy for a couple of hours as I clicked on link after link after link, getting myself deeper and deeper into the chain of blogs. I read one blog by a young girl whose birthday is the same day as my oldest daughter - maybe a couple of years earlier - so I read it simply for that reason. She talked about college life and piercings and tattoos and which of her friends were sleeping with the others also. Let\'s see... oh YES - there is the continuing saga of the california bar bouncer and company - quite amusing.  Most of these well put-together cyber journals makes me wonder why I even bother attempting to entertain myself - let alone anyone else - with my own rants and raves. But then again, I only started this thing for the sole purpose of bashing a certain scum sucking sewer rat and exorcising a few of my own demons.  Maybe I should just have another piece of cake and a glass of milk and call it a night - right now I\'m much more in the mood to read about Kashmir\'s travels in the Peruvian rain forest than to pour out my guts or worry about correct use of grammar and punctuation.  I do have a few ideas for future posts, so I\'m making a list here - in case I forget:  1) reconnecting with old high school friends/classmates  2) why I\'m ready for my 20 y/o daughter to get out from under my house and out on her own  3) why so many men are homophobic and yet - they all seem to have a fascination with trying to convince a woman to try anal sex (ok - just threw this one in to see if you were still awake)  4) my ex-husband, the moron and his slutty sidekick I call my stepwife  5) funny resume blunders I\'ve actually read first-hand (this one should be put into book form, published, and make me a rich woman)  6) people who drive down the e-way in the left hand lane - but not going faster than the right hand lane - as if they consider themselves the official pacecar of the canyon e-way  7) why i\'m a selfish horrible person and I know it, but can\'t seem to help myself  8) are there ANY good men left out there who can appreciate me - oh hell - are there any good men out there, period????  later.... osm'),(5361,4015298,'09,August,2004','2004-08-09 00:00:00','   Ok here - here I sit bawling - I\'m supposed to be working, and I have been - with the radio playing country music in the background. There was a PLAIN and simple reason why I gave up Country Music after I separated and then divorced - it just freakin\' depresses the HELL out of me!! But my oldest daughter likes it, and some of the stuff today is really good, so lately I\'ve found myself listening to it more and more - BIG MISTAKE.  And now, as I sit here sorting through resumes and making my daily calls to hiring managers and candidates, in my subconscious I hear the words to a song by Gary Allan come pouring out over the air waves. It\'s a song about a daddy\'s love for his little girl - I believe it\'s entitled \"Tough Little Boys\". It\'s all about how this grown man is turned all to mush watching his little girl grow and change - and all the inevitible emotional tugs there will be - watching her take her first steps, go off to school, get married, etc....  Which brings me to the cause of my tears. My own little angel lies sleeping in her bed - my GOD she is beautiful - with her golden blonde hair, but dark, dark THICK eyelashes laying on porcelain skin. Just as beautiful when she\'s awake, because she has eyes of the BLUEST, truest BLUE you\'ve ever seen - not sky blue, and not dark - just brilliant BLUE. Anyway - this song makes me so painfully aware that my baby\'s father goes about his daily routine without even sparing a thought for her existence, let along her well-being - and certainly she will never hear him say how he watched in with the most thrilling trepidation as she took her first steps, or how he cried like a baby and followed her school bus all the way into town on her first day of school, or how he knows some day he\'ll have to give her away - he\'s already done that. He did it the night we made her. He did it the minute he told me he loved me and wanted to live with me forever - with the first lie that came from his lips. It makes me want to pick up the phone and call Mr. Randy Brison (see, I made it almost a whole month before I actually used his name - hence the tootsie pop reference) and tell him what a pathetic person he is - and how I ALMOST feel sorry for him.  She is awake now; I can hear her over the baby monitor - lying in her bed, singing a song that only a two-year-old can compose. She is precious. She holds the world in her tiny little hands - it is hers to hold fast or to throw away. Anything is possible for her - except her father\'s love. I hate him for that. Fuck him. What a loser - who could do this to their own child?? I don\'t mind that he changed his mind about wanting me - I\'m a big girl, I\'ll get over it. How will she?  She already has used the word \"dad\" to mean ANY person of the male persuasion, not knowing that it is supposed to be reserved for the male person who, if he didn\'t donate the dna to create you at least cares more about you than any other in the world and would die so that you might live. Don\'t get me wrong - I know there are LOTS of dads who suck - my ex-husband has let my older kids down so many times it is unreal, but at least they KNEW who he was. At least he managed to have a few bright and shining moments - and has wanted to know about them and hug them and tell them he loved them. It\'s a different kind of pain when they at least KNOW what \"dad\" is supposed to mean than it is when trying to explain to a child that \"dad\" isn\'t just a random word. It is supposed to have purpose and meaning - and a face to see yourself in, and arms to pick you up and hold you, and lips to kiss away your fears/tears/boo-boos.  AAARRRRGGGGHHHH - these are the days when I wonder how stupid I really am. How could I have thought it would be fair to her to bring her into this world and to keep her myself??? And yet - anything else would just not have been an option for me. Which is the bottom line - she is for me. I wanted her. I needed her. Far more than she will ever need me - I hope that doesn\'t become a weight around her neck. I will have to work very hard to let her be her and to live her life without worrying who\'s going to be there for Mom.  Like I said - I\'m a big girl, I can take it. I just wish my little girl had a daddy - HER daddy - and in a moment of true weakness - I wish he had her, too because he\'s the one who\'s really missing out.... osm'),(5362,4015298,'07,August,2004','2004-08-07 00:00:00','   I just read a post on another blog -   urlLink  http://ridor.blogspot.com/    . The post was entitled \"I Have No Right To Tell Them How To Live Their Lives!\" I actually wrote this post as a comment, but it turned out so long, I couldn\'t post it there (they have a 1000 word limit). So, I just said I enjoyed the comments and left my URL. So, without further comment - here is what I wrote:'),(5363,4015298,'02,August,2004','2004-08-02 00:00:00','   I have no idea what I\'m going to write about today. I just know that I haven\'t been here in a whole week and I probably should write something.'),(5364,4015298,'02,August,2004','2004-08-02 00:00:00','   Hey! I received my first comment tonight on my blog. That was a funny feeling, I must say (thank you, Ed Grimly).'),(5365,4178914,'09,August,2004','2004-08-09 00:00:00','   Alrighty, i think i have this bitch worked out, i think that if i use it in a fairly dumbed down state, i can get away with dealing with the signing in bullshit only like once or twice.  Anyway, i am gonna go for a trial run on this site, i have another journal thing going on over at a site called AnimeOnline.net under the screen name Jabroni, for those interested go ahead and go there and read some of the shit i write on that site, most of it is quite long... i get bored like that.  I am thinking i might go into long ones like that over here.  Well right now i am currently listening to some Zakk Wylde, this guy is an awesome musician, he used to be in Ozzy Osbourne\'s band, not Black Sabbath, but his other band.   But now he went off on his own and did a solo album called \"Book of Shadows\" and started another band called Black Lable Society, which is also an awesome band.  He\'s one of my favorite artists, he\'s quite talented.   Today has been fairly boring, not much to do but sit on my ass alone and watch TV, Movies, and/or my Computer Screen...  I guess it\'s nice to take a load off everynow and then, but i think i\'d much rather be chillin with my friends, laughing it up.  I even found myself not enjoying Video Games recently, i would play Computer Games in a second, if my comp didn\'t shit out on me every now and then.   I have such a good peice of shit computer.  By that i mean it\'s hardly shit Hardware-wise, but it can\'t play games for shit due to some (i\'m guessing) software problem.  I guess there also might be some problem with my Motherboard, some PIO shit and how it switches to that mode randomly.  Well  fuck you  motherboard, i just know it better not be my harddrive, because if it is i am gonna be pissed.  I fucking hate computers, they are the bane of human existance, and i also believe it will be the downfall of human civilization.   Stupid machines, we rely on them too much and they break down all the time,  ooh how i hate them.  It\'s getting late, and i still have some net surfing to do, gotta look at some new computer cases for the new computer i am about to build...  '),(5366,4178914,'09,August,2004','2004-08-09 00:00:00','   Well this entry is a test more then anything else, so far this system has systematically fucked me over so i will be amazed if this even gets posted without the bitchy system trying to get me to sign in again... which i would have to do over and over again.  Oh don\'t get me wrong, i\'ll be able to post it\'ll just take me for ever to re-write what i write because the cookie system on this site is fucked more then a high school Prom Queen  '),(5367,4008193,'24,July,2004','2004-07-24 00:00:00','   hey all,     &nbsp;   today was an okay day.....i went to a karate class even though i am not suppose to     &nbsp;   it wasn\'t bad at all because it was jus a sword class and even though afterward my foot was huge and swollen it was better than sitting around the house doing nothing     &nbsp;   after sword class i stayed for the belt testing and it was fun..i haven\'t been to one of those in like forever...it\'s kinda sad that i am excited about that ...maybe not ; )     &nbsp;   ANYWAY.....my friends pasted and they have their new red belts with brown stripes i was soo excited for them.&nbsp; so congrats SHANNON and KRISTIN!!!!! WAHOO!!!!!     &nbsp;   after the test i saw my old friend matt who i haven;t seen in like forever...he is SOOOO busy at work that today was his only day off since summer started...so it kinda stunk    &nbsp;  later on kristin came over and i taught her how to play \"tail\" (in polish it sounds kewler but it still means tail)    &nbsp;  she liked the game and i learned dutch blitz and i like that game even though i am not that good at it.    &nbsp;  well.....later on i had dinner which was sooo yummy    tomorrow&nbsp;shannon it flying off to Hawaii and i think that&nbsp;is soo unfair best i wanna go but what can i do&nbsp;    alright i am going now because i am tired ...working tomorrow ewwww &nbsp;   '),(5368,4008193,'23,July,2004','2004-07-23 00:00:00','   TODAY WAS SOOOO FUN!!!!!!   i woke up way to early in the morning but other than that it was fun.   I went out to my friend lizzy\'s house to see my friend zak who moved away but came backfor a visit. it was sad because he DIDNT REMEMBER ME!!! after a while he seemed to remember who i was so that was good.   we all went to pirates cove and then didn\'t stay long...we went to see a girl named karen( i don\'t know her)&nbsp; she worked there.&nbsp;   after pirates cove we went to walgreens to buy some cameras for a remeberance stuff.   we went back to lizzzy;s house and hung out for a while....my mom called on my cell and asked where i was..she said she couldn\'t hear me she i repeatede being a lizzy\'s like a million times then she was like oh ok bye then...affter like an hour my dad called and did the same thing.   after a few hours we ordered pizza and it was fun....then we played spoons ..i was REALLY good or atleast i like to think so..   so then i left and got home and called my best friend shannon...she came over we hung out and then i had to go to work...   some people (costomers) are dumb and they were horrible to deal with today.   after work i went to get shannon becuase she is sleeping over. we went to see I, ROBOT and it was good i dont know what some people are taking about.   ok well that\'s all....i have to go now!!!1   BUH BYE ALL   LYLIMT   '),(5369,4008193,'22,July,2004','2004-07-22 00:00:00','   urlLink     ERRRRR....   I HATE physical theraphy!!!!   I had another new person today and they didn\'t understand my thing either...huh how odd since i have explained it to everyone else and they alwaays seem to understand. well she tried a new deep tissue massage that totally killed and i was really about to go postal on her...i told her it killed and she was well just another 5 minutes and u will be done and OMG it hurt sooooo bad.   so after a ton of pain...i went to help out at the church...less that a week away from the LAVA LAVA ISLAND thing....kinda ready kinda not.   well ...i have nothing else to say i guess..soo   BUH BYE LYLIMT   '),(5370,4008193,'22,July,2004','2004-07-22 00:00:00','   hey,   this is my first entry....kinda wierd   well...just resently i started physical therapy on my right foot.   well, just before i started karate in like 5th grade i had surgery to remove an extra bone becuase it hurt a ton. well......unfortunatly it GREW BACK. didn\'t know that could happen. so it hurts again after like 7 years of no pain.   so we went to doctors and they said surgery is a short term and therapy is a long term solution. so we are giving it a try.   if it don\'t work then i will have surgery and be out for like another 6 weeks of no karate!!!   already bummed about it and feeling really lazy so trying to lift weights at home but it is soooo boring. &nbsp;&nbsp;&nbsp;   i have been trying to get shannon to come and wieght lift with me but she won\'t.   &nbsp;  i have a new summer job. i have been working there for like a month now actually. every one at Culver\'s (where i work) is kewl. one guy kinda freaks me out because he is touchy feely guy and normally i am fine with that if i have know u longer than a week and&nbsp; i only see u every two to three days of that week. so he is weird.(no he is not gay like my friends have asked)   ok well i have to get up early for tomorrow so goodnight all!!!   '),(5371,4008193,'29,July,2004','2004-07-29 00:00:00','   OMGOSH!!!!    &nbsp;  It\'s been awhile since i wrote so...i wrote last monday and now it is thursday.    &nbsp;  so tuesday was alright nothing out of the oridinary really plain and blah... went to church the games were alright and the kids had fun so what was all that mattered....i had work and northing interesting happened.&nbsp;&nbsp;    &nbsp;  i wanted to leave early but i couldn\'t but luckily i didn\'t have to CLOSE.....i was asked like 20 times but turned it down because it was until 11 and then after to close.    &nbsp;  so i went home at 9 and then did nothing else because my feet were hurting me.    &nbsp;  so weds was fun when to church again....games again fun but i sprained my good ankle..so NOW i have two bad feet....left sprain right the surgery physical theraphy thing...my friends are yelling at me becuase i am falling apart.    &nbsp;  so weds night we had a GAME NIGHT and it was fun. we played the 90\'s game and scene it...then some poker and BS but then we taught Amie some \"tail\" in polish it sounds kewler...now normally new people are not that good at the game but she was pretty good!!!! So GO AMIE WAHOOO!!!!!    &nbsp;  then i slept over and woke up almost late for church but don\'t worry i got there ON TIME!!!!    &nbsp;  the games were great today too and then KIDS REALLY REALLY LOVED THE GAME! before they were liek oh great ...games but this time they really did have fun    &nbsp;  so then i came home after going shopping and buying really kewl \"invisible\" playing cards...very kewl    &nbsp;  then came home..read and knitted....    &nbsp;  then&nbsp; am off to sing..we are practicing for the church choir so hope we do good....alright that was alot of talking so i am gunna go..    &nbsp;   &nbsp;  BUH BYE ALL!!!!!!!1    &nbsp;   '),(5372,4008193,'02,August,2004','2004-08-02 00:00:00','   well....it is MONDAY      over the weekend nothing to much happened.....i had work on fri night and then came home and then i had work on sat morning so i was there at 11 til 4 and then when i got home...i was sooo TIRED but my cousin\'s house opening and his daughter\'s b-day party was that night so we went and it was fun but i was still so tired.      we got home at like 11:30 ..which isn\'t late or anything but i stayed up way to late afterward.      so i finally went to bed and then i remembered i had church in the morning for singing so i had to be up at 8 to be there at 8:30 so i was toooo lazy and i felt guilty but not as tired and i skipped the first mass and went to the second one....no one at church was upset at me so it was okay.      when i got home i decided to do some re-decorating so i cleaned out my WHOLE bathroom and BED room....it took forever....i did a BUNCH of shopping and then came home...i had to finish it today but i was sooo tired...      i had to get up for physical therapy at 8 so i was up at 7 which was too early after going to bed at like 3 in the morning...   it lasted longer than normal because we did some new exercises and then some more newer ones that he has never tired on any one else...so i was the test subject....it hurt sooooo much afterwards.      then i had work at 11 because i was nice and took someone else\'s shift like others have done for me when desperately couldn\'t come into work...well i was there until 4 and nothing too interesting happen.....this one co-worker i hate was there today and she was bossy and she just s ERRRRR.....i was refilling stuff and she was playing with the ice cream and not getting anything for any costumers just playing and starring at the ice cream so food came down and i asked her if she could grab it since she wasn\'t bus and everyone else (including myself) were busy and she as like well go right ahead...she is just soooo GGGGRRRRRRRRR!!!!!...but i am not upset :P       so then i came home ever more tired and then finished my room makeover.      after that shanno came over after being in hawaii for a week....we had a BUNCH to catch up on...      ok well i have to go and do stuff so bye all   '),(5373,4008193,'14,August,2004','2004-08-14 00:00:00','   hey  i am in the dells right now visiting family and our newest family member SABRINA!!!  she is soo cute and i love her already  she wieghed almost 7 lbs when she was born...  so far that is it i saw everyone else in my family here and they are all kewl so it was great... i love it here...my cousin let the pool stay open ater becuase i got home later and then 15 minutes before it is normally suppose to close i decide to go swimming and he said he would leave it open for me if i wanted..i was happy and i love him sooo much and thanks if u reading this!!  sometimes i wish we could still live there but my cousins\' talk aobut how they dont like it some of the time and then i just change my mind  um...finished my book \'to have and to hold\'...it was a GREAT book i couldn\'t put it down...also my fav book and everyone should read it \"The Christmas Train\" GREAT AND WONDERFUL book  ok well i am starting color guard soon and i am so scared about who is going to be there  also this weekend i am going to church camp and i CAN\'T wait i am soo excited....only becky, tommy, and i are going but that is kewl becuase we can all \"bond\" or something odd like that lol  got home today...monday....went to work right away but still really tired and tomorrow i have to start work at 11 so i should be going to bed...so me going   '),(5374,4008193,'08,August,2004','2004-08-08 00:00:00','   hey, it has been forever since my last blog  well i have been up to a ton....i am allowed to do karate again so i am sooo excited.  i can\'t wait tomorrow i am going to take two or three sparring classes to try and get myself back up to speed....hopefully...  the week was pretty normal....i went down to see ISU and UIUC and i think i really want to go to UIUC now!!!! it was soo pretty down there and we did some shopping ..not tooooo much but just enough.  then my mom went on this pilgrimage to IN and it was cool...kinda odd just my dad and me since my brother is still in the dells  we went to my uncle\'s house for a suprise happy b-day thing and we were there until 2 30 in the morning and i was soo sleepy when i went to church. whiel at my cousin\'s house though i made a really fun bracelet out of speaker wire..now it sounds weird but it looks so kewl a bunch of people liked it from karate  then i went to the NK party and it was pretty fun....i got to play volley ball and i hadn\'t played in such a  long time...then we looked at pics and i had to head home.  so now i am home and i am sooo tired ...i am reading a book called  Ender\'s Shadow  and it is a great book but u must read  Ender\'s Game  first to get this book more...ok i am off to read some more   '),(5375,4008193,'16,August,2004','2004-08-16 00:00:00','   urlLink    i like this one the best!!! i did make fun of it because i am not sitting on that chair..they are just were like \"hold this\" and then tookt he picture!! but i think it is my fav&nbsp; urlLink   '),(5376,4008193,'16,August,2004','2004-08-16 00:00:00','   urlLink    Sr pics...i was soo scared that day&nbsp; urlLink   '),(5377,3554687,'22,July,2004','2004-07-22 00:00:00','   I must admit, at the bonfire that&nbsp;I hosted tonight I was quite exclusive in who was invited. I figured I would invite my beloved D. Squared and give him the perfect chance to spread his good name throughout our generation. However to my shock and surprise he outright declined my offer. He claimed he had to many important things to attend to, which begs the question \"What is more important than the furthering of your DJ Persona?\" This man no longer has his eyes on the prize, and hopefully getting away from the bogged down environment of Chaska will develop his creative genius even further in this newest adventure. I must say I was disappointed in his choice not to attend this evening and his rating in my mind has led him in a spiraling motion into second behind his own master PvD. However my love for D. Squared will not end so easily. I just hope that after his trip he will once again join up with his Chaska homies and enjoy the good ole days.   '),(5378,3554687,'19,July,2004','2004-07-19 00:00:00','   This Saturday D. Squared will leave on an epic 14 day journey, first to San Diego and then secondly to Auburn. He plans to celebrate capitalism on the San Diego beaches while fondling the local Sea Lions. Maybe Sea Lions like Trance, D. Squared could DJ for them possibly? While at Auburn D. Squared plans to hang with his neo-capitilist friends. But hey, thats what you get for being libertarian. And this just in...... D. Squared has confirmed his sexuality, yes he is indeed straight. I\'m sorry Super Honky, but you can\'t have him.   '),(5379,3554687,'14,July,2004','2004-07-14 00:00:00','   We\'ve come to ponder many a times whether D. Squared is hetero or homo. There are many signs that point in both directions. He once descibed what the beloved PvD was wearing at a concert by this, \"His outfit was fabulous yet simple, as usual for the master of fashion.\" Ok Queer Eye, a little creepy. Another thing he commented on Pvd went like this, \"About three minutes after Paul took the stage, Jessica turned to me and said \"he is so good looking.\" Duh. When he looked up at the crowd for the first time, I swear he smiled right at me.\" HOOOOOOMO, HOOOOOOOOMO. C\'mon D. Squared, you can do better then that. For some reason his friends at Gonzaga thought he would score with 15 ladies in his freshman year of college. Please, we saw the women he was hanging out with. D. Squared, don\'t let your devilish good looks go to waste. But apparently that Jessica girl was his girlfriend. Hmm, maybe D. Squared his bi-sexual. Screw that, he should take up a new love for......animals.   '),(5380,3554687,'13,July,2004','2004-07-13 00:00:00','   Some of the people who now know of the divine being known in the underground trance world as \"DJ D. Squared\" have all sat awake at night, all wondering the answer to the same question. Just who is D. Squared, and when and where can I hear some of his music? Well my esteemed colleagues and I have wondered these things ourselves. Although we have heard hour\'s of D. Squared\'s trance sets, these are but teases of what his DJ mastery holds in store for the future. I will now be the first to demand more out of D. Squared. I demand that he makes a public appearance, and to add more, that it be a full out trance rave. Chaska Minnesota has never seen the likes of a well known Trance DJ entrepreneur as D. Squared. And it is about time that the public should recognize his mastery of all things electronic. Thus I myself shall proclaim that myself and the co-founders of this site will in the future be in contact with D. Squared in order to get him to hold a concert in order to spread his godlike image throughout more young impressionable trance addicts like ourselves. So to all those who this message reaches please contact us by posting comments to this BLOG about your interest in making our dream of a D. Squared extravaganza a reality. We need all of our D. Squared superfans to unite under one banner regardless of race or gender or musical taste and agree to love the wonderful music played by D. Squared.   '),(5381,3554687,'12,July,2004','2004-07-12 00:00:00','   The former home of the King of Rock N\' Roll is now the current home of the King of Trance/Prog./hardhouse/electronica.  DJ D.Squared bought the much celebrated property after his sophmore album, \"Intergalactic Euphoria\", went multi-platinum in 49 countries, staying at No. 1 for a record-breaking 67 weeks in Venezuela.  Upon moving in, DJ D.Squared was quoted as saying that \"The D to the Squared ain\'t never gonna stop preachin\' to the people.\"  Three days later, he filed for bankruptcy.   '),(5382,3554687,'12,July,2004','2004-07-12 00:00:00','   After reading our blog, D. Squared became infatuated with our sophisticated witty humour. He just couldn\'t handle it. He was even kind enough to write  urlLink about us  on Grooves on a Vinyl. What a Playa.   '),(5383,3554687,'11,July,2004','2004-07-11 00:00:00','   I just got done listening to D. Squared\'s hour long set, Just A Touch of Techno. It was alright. For some reason in the beginning some lady kept yelling something about how the revolution will not be on the radio. I dunno what revolution she was talking about, I must ask D. Squared about this. Then for what seems like 40 minutes (probably shorter then that) D. Squared is just pumping the same drum/bass beat (I think it\'s drums and bass) into you. It changes a little bit in between, then go backs to it. Once in a while some guys says \"I know you want me.\" Man, what a ride D. Squared takes you on. Amen Brutha, Amen.   '),(5384,3554687,'11,July,2004','2004-07-11 00:00:00','   In case you were wondering about this websites layout, we decided that this would best fit D. Squared. The neon lights and all would definitely suit his raves that he plays at and attends. Tacky they may seem, but perfect for full blown trance addicts like Mr. D. Squared himself.   '),(5385,3554687,'11,July,2004','2004-07-11 00:00:00','   DJ D.Squared is raw, unadulterated talent.  To all you naysayers who claim that trance isn\'t real music, I only have three words for you:  Paul Van Dyk.  Mastery and skill flow through DJ D.squared like blood would flow through a mortal.  but DJ D.Squared is no mere mortal.  His trance music doesn\'t just justify mankind\'s very existence, but it takes it to a higher plane.  Long live DJ D.Squared.   '),(5386,3554687,'11,July,2004','2004-07-11 00:00:00','   If you want to read some of D. Squared\'s angry venting then go to this blog that he has,  urlLink Grooves on a Vinyl.  It\'s definately worth the read homies.   '),(5387,3554687,'11,July,2004','2004-07-11 00:00:00','   A conversation that I had with D. Squared from a few weeks ago.....  RageCave518: I wish I was quad-lingual RageCave518: that would be cool ElvenNuke: If you had something to say besides \"damn homie,\" maybe...  Muahahaha, true comedy. God I love D. Squared.   '),(5388,3554687,'11,July,2004','2004-07-11 00:00:00','   I am glad I am one of the founders of this website. My love for D. Squared shall flourish for many a years.   '),(5389,3554687,'11,July,2004','2004-07-11 00:00:00','   ElvenNuke: Are you three going to become hardcore capitalist? stabwestward15: hardcore trance addicts is more like it stabwestward15: we are going to be your super fan section ElvenNuke: ah ok stabwestward15: yea yea yea ElvenNuke: who is in my regular fan section? stabwestward15: i dunno..........your parents? ElvenNuke: Hah.... my parents couldn\'t care less about my mad trance skills. stabwestward15: i care about it more than anything stabwestward15: invite me to one of your dj raves sometime stabwestward15: i will bring a cool sign   '),(5390,3554687,'11,July,2004','2004-07-11 00:00:00','   D.Squared has begun throughout the world. All that oppose it shall fall. We love the D. Squared, and this site is devoted to our love for him, as well as other things, such as cream soda, finch, or the dong in the face rap.   '),(5391,1261679,'28,February,2003','2003-02-28 00:00:00','   My intelligent conversation with White Chocolate aka Elise:  elisebee06: Liar! aquafina seaweed: I\'m not a liar! I\'m a truther! elisebee06: see you changed your spelling from liar to lyer.. elisebee06: therfore admitting I am right!! aquafina seaweed: See? a TRUTHER! elisebee06: what?? elisebee06: truther isn\'t a word you crazy! aquafina seaweed: what\'s the opposite of liar? elisebee06: truthful person? aquafina seaweed: I guess, but i like words that end in -er elisebee06: No  elisebee06: no elisebee06: n elisebee06: oo aquafina seaweed: see? you are a Noer! elisebee06: kiss my butt amanda! aquafina seaweed: you kiss my booty! aquafina seaweed: my mocha choca booty! elisebee06: no one wants to kiss your butt amanda:-! elisebee06: thats GROSS! Yeah.  We even addressed the spelling of \"liar.\" Anyway, I just started out this journal.  Who knows?  This might be my last entry. By the way, is my blog green?  If so, that was totally a coincidence, you know, about the whole booger/snot theme.  Sorry if that\'s too gory for you.  I mean that\'s not too gory for me, because I have nightmares about my big-lipped English teacher, Miss  Eye-Corn, eating me up.  I know she really would if she had the chance to.  I just shudder at the thought of it.  Going to English class in fear of being devoured!!  Man!  Why couldn\'t I just have a normal life?!?  '),(5392,1261679,'31,March,2003','2003-03-31 00:00:00','   Hey fellow humans or alien satellite interceptors.  By the way, what’s going on out there in space?  I hope you’re not having a war there also, because the war here, on earth, is very depressing.  It’s inevitable, meaning that if we didn’t have the war in the first place, Saddam Hussein would be a very dangerous, or much more dangerous threat, to mankind.  And all those war protester people make me so mad, because if we didn’t have the war, it would be like sitting back and doing nothing.  Those people are just doing that so they can say they have a hobby, which they really don’t.  They need to get out more, like that Michael Moore guy at the Oscars.  He needs to exercise a bit, not just walking around in circles holding a sign up.  That’s like burning a pound, Mr. Moore!    MUAHAHAHA!! Okay, kids.  I’ve finally decided what I want to be when I grow up!  A VAMPIRE!  No the females are called Vampiresses.  I want to be a Vampiress when I grow up!!  I’ll have long black hair with white highlights and I’ll pin it up like Marge Simpson!  And I’ll have really cool fangs- long and sharp and, um, bloody?  I don’t like blood.  That’s why I don’t want to be a doctor when I grow up, so scratch the vampire thing!  He he.  Oh now I know!  I want to be a person that’s always in the “after” pictures for all those diet and exercise machine campaigns.  That would be so fun!  The fat people get the ugly clothes and no makeup, but the skinnier people get all the pampering!!  And the cool clothes!!  I would also like to win  eight Grammys .  Norah Jones is so lucky and talented!!  I want to be just like her now!!    You know, there’s this thing dog-like thing that my family is housing that is of relation to me.  He’s actually my cousin and he’s one of those people who you really, really don’t want to be associated with.  He’s somebody that you don’t want to say is related to you.  He’s just like a dog too!  Crawling around on the floor, always scratching himself, and-um- looking at porno on my computer??  You don’t know how grossed out I was when I saw this file named “Britney and JLo Together in Bed.”  There were like three pictures deleted of JLo!!   Scary, scary, scary .  I feel very sorry for Ms. Lopez.  I can hear an ice-cream truck in the distance…ahh, Heaven.    Well enough now.  I’m going bouncing.  Why do people say that anyway?  “ Gotta bounce !! LOL!!”  He he.  Web-savvy teenagers are the most fun people to talk to.    PS:  Elise!  We need to make our shirts!!  And I ain\'t Khan\'s or Austin\'s girlfriend!!  '),(5393,1261679,'02,March,2003','2003-03-02 00:00:00','   Sup, kids?!? Yo Frothy Coffee what\'s going on with your man, Legolas?  I hear you guys are getting to be good friends now!  That darn fried blonde girl, \"MB!\"  You know, all you have to do is use a ghetto pickup line!  Proof:  aquafina seaweed: then i think i will start with a ghetto pick up line too! elisebee06: it works!!!!!!!!  And White Chocolate!  You gotta \"lure\" Viggo (not you dan you weirdo who thinks every girl likes you) away from his girlfriend, because I know he would leave her for you!!  Lure!  Like bait on a fishing line, except for the hooking and reeling part--yeah that would kind of hurt. You know what? I should be following my own advice.  PUH!!  Who cares?!?  '),(5394,1261679,'01,March,2003','2003-03-01 00:00:00','   Yet another day in the life of me, Miss Choca.  (My real name is Amanda, by the way.)  Some people call me Mandy Candy.  What the f?!?  It\'s too icky sounding.  You know, like, to bubblegum poppy girly.  >:P!!  Anyway, I\'m supposed to see How to Lose a Guy in Ten Days.  I don\'t even know why I want to see it!  I hate chicky flickies.  You know, like the ones that are so predictable.  It just seems like a \"comic relief-y\" type movie.  But Bethy-Wethy told me that she saw it and it makes you want to have a boyfriend.  I don\'t think I\'ll get a boyfriend, so I\'m taking a big risk in seeing this movie.  After Elise sees this movie, she\'ll be with her Viggo!!  Carolyn will be with her Legolas!!  I will be with Keroppi.  '),(5395,1261679,'27,April,2003','2003-04-27 00:00:00','   Hello everyone!  (All of you nonexistent readers!)  How are you?   Well it has been kind of a long time, because it is now the 27th, and my last entry was on the 18th.  I am very sorry for that.  Anyway... Hmm, my Biology teacher is evil.   Eeeeeeeeeeevil .  No, you haven\'t seen evil until your biology teacher gives you a lab to do at home, and to top it all off-  With a  meterstick.  What the fudgsicle?!?  >:O!!   You know what bugs me?  When musicians don\'t update their tour diaries.  It\'s so annoying because they do have a lot of time, I mean they can skip all of their sleeping hours for just one entry, you know what I mean?  Like photo shoots are really that tiring.  I don\'t think so!!  If I were a rockstar, I wouldn\'t leave my fans hanging, the very few I\'d have.  It would be fun.  I\'m seriously thinking about becoming one.   I lost interest in a lot of things over time, for example : Brandon Boyd, Keroppi, stalking, hygiene... Now I\'m re-discovering my obsessions.  (Not Keroppi, though...No  way!)   He he he Okay now I have to get back to that biology lab.    P. S.- \"But soft!  What light through yonder window breaks?\"  Nice use of iambic pentameter, I must say! ;)  '),(5396,1261679,'14,April,2003','2003-04-14 00:00:00','   I\'m  innie !! Okay, kids!! A lot happened over the weekend.  Let\'s start on Friday, actually.  I saw What A Girl Wants with Elise and Laurel, Bethy, Michelle.  OHHHHHHHHHH myyyy GOOOOOOOOSSSSSSSHHHHHHH!!!  I was smitten!! Oliver James: the cutest actor! He\'s all mine, Elise!!  You can have Bostjan and Sark, but  Oliver\'s mine !  He he!  Well anyway, On Saturday I went to La Grange for the MS-150.  No I didn\'t actually bike, it was just that we had to help the rifers ( I mean riders) for the Healix team.  I\'m doing it next year, though!  I can\'t wait!  That night we stayed in a bed and breakfast called:  \"Granny\'s Bed & Breakfast Inn.\"  Now isn\'t that supposed to hint something there?!?  It was like staying at the  Wunderlich Farms , man!!  They even had the original copy of Animal Farm!!  Scary...  On Sunday I went to visit my sister at her dorm in UT.  Her dorm is so cool!  (Although she has a creepy roommate who could tell you anything about the war...) I can\'t believe she doesn\'t decorate it.  I\'d have posters all over the place, man.  (Of Oliver James!!)  Muahahaha!!!   So all my orchestra friends are bringing dates to the end-of-year banquet.  Well, one is lucky enough to already have one, and the other is hesitant about asking.  I am thinking about asking  Spottie Dottie , but I\'m so scared that he\'ll say no.  I mean, he has all right to say no and I wouldn\'t be mad because there\'s no way he can go with someone else.  But I\'d die if he said no!! So I\'m not taking any risks and I\'m saving my life before it\'s too late.  Unless someone convinces me (with some lovely persuasion!)  or forces me (like threatening me or something) then I will have to leave him alone.  I\'m very tempted, but I am brutally realistic.  If anyone has a good probability statement on his chances of saying yes to me out of 10, I might do it!  Emphasis on the  might , because I\'ve never done anything like that before.    So there you go!  And by the way, you  can  have your cake and eat it too.  (Take that Ashley Olsen!!) I\'m  outie !  '),(5397,1261679,'09,April,2003','2003-04-09 00:00:00','   Shakespeare was a homo.  I bet he has a boyfriend right now in HE double hockey sticks!!  Just kidding, William.  I haven\'t started that Great Expectations english essay.  It\'s due Friday, but I have an orchestra concert tomorrow night.  First chair of the second violins...woohoo.  Inner circle power!    I\'m stupid.  Need I say more?  '),(5398,1261679,'05,April,2003','2003-04-05 00:00:00','   This entry will be about animals today.  I have this friend who wants to hook me up with this frog.  Think this is a good idea?  Remember that story,  The Frog Prince ?  If I kiss him will be turn into a prince, or will I just get a wart and have to live with the fact I dated a frog?  We\'ll never know... Anyway, I have developed a new fear of dogs.  I am so afraid of dogs, especially the one that\'s living with me.  And I don\'t even have a dog.  ;)  Peace, love, and ice cream trucks.  '),(5399,1261679,'29,May,2003','2003-05-29 00:00:00','   I watched Jay Leno last night and they had a little special airing of a live debate between the two most ignorant 18-year-olds in the world.  \"Point: What\'s Your Point?\"  Yeah.  It was mighty funny, I tell ya!  Both of the kids (one was a girl and one was a guy) agreed that the saying \"I traveled around the world,\" was invalid because, according to them, \"you can\'t see the  backside  of California!!\"  And that guy\'s going to TCU!!  It\'s so sad, sad, sad.    I\'ve got a lot of summer goals.  I will list them for you: (1) Lose 10 pounds! (2) Learn how to play guitar. (3) Think of better jokes/comebacks.  (They always come in handy!) (4) Learn more card tricks/games!!  (Fun, fun, fun stuff!!) (5) Stop being so internet dependent.  (6) Grow! (At least 6 inches :D) (7) Try to be smarter, grade-wise, which basically means to adapt new study habits.  (Like, um, yeah    right!! Hehhhehehe) (8) Try to be a better person in general.  :)- I love all the people who care to read my blog.  It really feels good inside. '),(5400,1261679,'28,May,2003','2003-05-28 00:00:00','   Hey, so I\'m at home, basking in my laziness, and also CELEBRATING MY LAST DAY OF SCHOOL!  Fun, fun, fun.  Nothing to worry about for a looong time.  This isn\'t so bad.    You know my worst pet peeve ever is when people step on the back of my shoes.  It fricking bugs me sooo much.  It almost equals PMS, but not quite.  (hehehe)  Especially when you\'re wearing flip-flops.    My taste in music is changing ever so fastly.  I am really very annoyed by Maroon 5 right now, and I don\'t know why!!  I used to be obsessed with them!  Not a big loss though.  I also cannot bring myself to listen to any Incubus, because I just recently found out that Brandon Boyd was bisexual and was too open about it that they lost Dirk Lance (the bassist) because of it.  Scary.  I was obsessed with Brandon Boyd man!  It\'s so scary that I can\'t listen to them anymore.  Oh well...I still have a 50% chance of hooking up with him now.  :D  That was crazy...but Elise also told me that, so me not so crazy!!  I also realized that I don\'t like \"yelly\" music.  Whenever I see a commercial or something for a band I\'m like \"I have to listen to them!!\"  And then after I force myself to like them, I still don\'t.  For example: Taking Back Sunday, The Starting Line, Good Charlotte.  The list goes on, but I can\'t remember it all.  It\'s so blah, if you know what I mean.  I love Phantom Planet, Nikka Costa, John Mayer, Jason Mraz, and OK go very much.  That would be the perfect concert too!!  Well, so, hmm, this is just like an awkward silence, which I\'ve had the opportunity to experience finally...:D  Yeah that was great.  But I guess it\'s for education when you\'re stuck in your next situation, right?  Yeah, that\'s what I\'ll keep telling myself.    I cannot wait until guitar lessons!!  It\'s gonna be so incredibly fun!!  Yay, yay, yay!!  I really wanna be in a band, but I don\'t know.  I have a few friends who said that they wanted to be in a band -me singing- but they haven\'t mentioned anything lately.  (This is your cue!)  Yeah, that\'s my dream...and living in a country-style peaceful house in SoCal, or a penthouse loft in NYC or Chicago.  That would be amazing--having a good husband and a dog---okay sorry I\'m kinda dozing off into daydreaming...:D  Well then, I bid thee farewell with a few notes on the future-- >>I might not update as frequently during June, so entertain yourself with past archives. >My birthday\'s on June 10th...(Hint! Hint!-- Remember Mrs. Metzler?!)   -Mandy Candy- **Ice cream trucks**  '),(5401,1261679,'24,May,2003','2003-05-24 00:00:00','   Hello again everyone!  Well now.  School is almost over!  (Only two more real days.)  Geometry final :P  Stinks.    Okay, I didn\'t have much to say.  Sorry!  Maybe next time it\'ll be more interesting.   -Manda-  PS- Tammy saved my life last night!  \"White People Cards (aka Egyptian Ratscrew)\" and VC :)  that was great!  He he he.  '),(5402,1261679,'21,May,2003','2003-05-21 00:00:00','   It has been almost a month since I have written (typed) last.  Sorry about that.  Yeah.  Limited access to a working computer is not exactly my kind of paradise.  If you have been a consistent reader [Casper], you might know that this dog who is almost 150 lbs and speaks Thai lives with me and looks at porno occasionally.  Stupid dog.  I wish there was a pound where we could send unwanted guests.  Stupid cousin freak thing.    ANYWAY!  I have a new impetus for living!  Jason Mraz\'s music!  (And probably Jason Mraz!)  (Probably not...) (Yeah it\'s because he smokes.)  Useless parenthetical phrases aren\'t so useless after all!  Let\'s see:   \"Remedy (I Won\'t Worry)\"   I saw fireworks from the freeway and behind closed eyes I cannot make them go away Cause you were born on the fourth of july, freedom ring  now something on the surface it stings that something on the surface it kind of makes me nervous who says that you deserve this and what kind of god would serve this? We will cure this dirty old disease if you\'ve got the poison I\'ve got the remedy  the remedy is the experience. It is a dangerous liaison I say the comedy is that its serious. Which is a strange enough new play on words I say the tragedy is how youre gonna spend the rest of your nights with the light on So shine the light on all of your friends because it all amounts to nothing in the end.  I wont worry my life away. I wont worry my life away.  I heard two men talking on the radio in a cross fire kind of new reality show Uncovering the ways to plan the next big attack  they were counting down the days to stab the brother in the be right back after this the unavoidable kiss, where the minty fresh death breath is sure to outlast his catastrophy dance with me, because if you\'ve got the poison, I\'ve got the remedy   the remedy is the experience. It is a dangerous liaison I say the comedy is that its serious. Which is a strange enough new play on words I say the tragedy is how youre gonna spend the rest of your nights with the light on So shine the light on all of your friends because it all amounts to nothing in the end.  I wont worry my life away. I wont worry my life away.  When I fall in love I take my time There\'s no need to hurry when I\'m making up my mind You can turn off the sun but I\'m still gonna shine and I\'ll tell you why Because  the remedy is the experience. It is a dangerous liaison I say the comedy is that its serious. Which is a strange enough new play on words I say the tragedy is how you\'re gonna spend the rest of your nights with the light on So shine the light on all of your friends because it all amounts to nothing in the end.  I wont worry my life away. I wont worry my life away. I wont and I wont and I wont [etc.]  -Download it-  I love love love that song.  I\'m taking guitar lessons in July.  Rediscovering my obsession with wanting to be a musician!  (No \"rockstar\" term anymore, too general.  But musician is even more general.  Oh well  It\'s just that when you tell someone that you want to be a rockstar, they automatically think GODSMACK!  And picture you jumping up and down with an electric Fender in your hands smashing it on the stage while screaming your head off into a nonexistent microphone. Yeah.)  Stupid parenthetical phrases!!  Shoot them.  (But don\'t sue me when you break you computer.)    I think me and Elise\'s (Yes, yes, me and Elise\'s.  I used incorrect English in attempt to not sound weird.  Elise and my...??  That sounds very very gross.  I\'m a stupid parenthetical phrase.)  conversations have gotten a bit more sophisticated.  You should hear us!  Look for me when I get to be 22.  I\'ll have a different name, though.  (Musician name, like Tori Amos.)  Get it? Top future names: 1.  Aquafina Seaweed (...which by the way is my AIM screen name if ya wanna slap me on your buddy list.) 2.  Amanda Mraz 3.  Amanda Boyd  Outie!  '),(5403,1261679,'27,June,2003','2003-06-27 00:00:00','   So here I am again sitting at the computer.  It\'s not actually the same computer I\'ve been generating my previous entries from; it\'s my mom\'s laptop.  Laptops are sooo, incredibly portable.  I mean, they\'re supposed to be but it\'s unbelievable when you actually use one.  But it\'s so slow!!  Maybe because it\'s so small or maybe I downloaded too many songs at one time.  I just watched a creepy CSI.  It was about magicians.  I hate clowns because they do magic, too.  They\'re scary.  I have an irrational fear of clowns, just like Tino from \"The Weekenders!\"  It started when I went to a company picnic for my mom\'s work.  There was this girl who wouldn\'t leave me alone and she had to go and she wanted my number and I said that I would get my mom to give it to her mom, but I didn\'t even know her name, but I would have forgotten it instantly.  Anyway, at the same instant she asked for my number, this clown said \"Can I have it too?\"  Oh boy, how creeped out I was.  At the tender age of 11, I could not tell if the clown was joking or if he was being very inappropriate and wrong.  Yeah.  Or maybe it started when \"Are You Afraid of the Dark\" started and in the intro there was this tall clown statue with a creepy grimace on his face and every time they flashed it, a creepy string bass slid to an E.  (Shuddering.)  I have a feeling this will be a creepy night, because this Daria episode is all about people telling scary stories.  \"I don\'t get it. Why am I almost perfect?  It\'s my eyelids!  They\'re fat!\"  He he he.  That is the best and perfect cartoon.  I remember when it was regularly aired on MTV and I watched (Oh my God, Daria just ended very creepily and I\'m practically home alone! Okay, back to the entry...) it and the movies they made about her like \"Is it Fall Yet?\" Okay, I have no life because I\'m justifying this on my own.  I\'m not some former cheerleader who broke her ankle and got fat and now stays home and eats Homemade Vanilla Blue Bell ice-cream who needs her friends to kick some sense back into her, no that\'s not me.  I don\'t need a reality check because I\'m already in reality!!  He he he.  (Even Stevens is better.)  Doesn\'t Shia LaBeouf remind anybody of Austin Rourke? There\'s just a strange comparison.  Maybe it\'s just me.  Dunno.  Anyway, I\'m actually writing this entry in an e-mail I\'m sending to myself which I\'m going to put shove into my blog later in effort to keep the URL to my blog-site available on my mom\'s laptop because it\'s just wrong for any member of my family to read this kind of stuff, but my big sister reads it regularly and my little sister found out about it and might try to blackmail me with all the info I\'ve put in here.  Okay I better stop there.  I\'ll add a song also, so you can download it and everytime you play, you\'ll think of me and my fear of clowns...   \"When I Get You Alone,\" by Thicke Oooh... Eah!  [Verse 1:] Baby girl, where you at? Got no strings, got men attached Can\'t stop that feelin\' for long no Mmmm You makin\' dogs wanna beg Breaking them off your fancy legs But they make you feel right at home, now  [Hook 1:] (Ooooh) See all these illusions just take us too long (Ooooh) And I want it bad.. Because you walk pretty, Because you talk pretty, \'Cause you make me sick  And I\'m not leavin\', till you\'re leavin\'  [Hook 2:] (Ahhh) Oh I swear there\'s something when she\'s pumpin\', Asking for a raise (Oooh) Well does she want me to carry her home now? (Oooh) So does she want me to buy her things? On my house, on my job On my loot, shoes, my shirt, My crew, my mind, my father\'s last name?  [Chorus:] (Ahhh) When I get you alone (\'lone) When I get you you\'ll know baby (know) When I get you alone (\'lone) When I get you alone now (it\'s all mine)  (Ahhhh....oh) Come on Oh yeah-yeah  [Verse 2:] Baby girl you da shit That makes you my equivalent Well you can keep your toys in the drawer tonight, All right All my dawgs talkin\' fast- Aint you got some photographs? \'Cause you shook that room like a star, now Yes you did, oh  [Hook 1:] (Ooooh) All these intrusions just take us too long (Ooooh) And I want you so bad.. Because you walk steady, Because you talk steady, \'Cause you make me sick  And I\'m not leavin\', till you\'re leavin\'  [Hook 2:] (Ahhh) So I pray to something she aint bluffin\', Rubbin\' up on me (Oooh) Well does she want me to make a vow? Check it (Oooh)Well does she want me to make it now? On my house, on my job On my loot, shoes, my voice, My crew, my mind, my father\'s last name?  [Chorus:] (Ahhh) When I get you alone (\'lone) When I get you you\'ll know baby (know) When I get you alone (\'lone) When I get you alone now (it\'s all mine)  (Ahhhh....oh)  Oh no Get you alone baby Oohh..  Oooh..oh-oh That a girl! You get some  [Hook 1:] (Ahhh) All these intrusions just take us too long (Ahhh) And I want you so bad.. Because you walk pretty, (You, you-you) Because you talk pretty, (You get some) \'Cause you make me sick  And I\'m not leavin\', till you\'re leavin\'  [Hook 2:] (Ahhh) But I pray to something when she\'s pumpin\', Rubbin\' up on me, now (Oooh) Want me to break it down? Check it (Oooh) Well did you want me to make it now? On my house, on my job On my loot, shoes, my voice, My crew, my mind, my father\'s last name?  [Chorus:] (Ahhh) When I get you alone (\'lone) When I get you you\'ll know girl (know) When I get you alone (\'lone) When I get you alone now (it\'s all mine)  (Ahhh) I get you alone (\'lone) When I get you you\'ll know (know) I get you alone (\'lone) When I get you alone (it\'s all mine) Yeah  Yeaaahhh....... (\"Temptation Park\" he he he [aka Templeton Park]) Download it!  Goodbye.  '),(5404,1261679,'10,June,2003','2003-06-10 00:00:00','   I am sorry that I haven\'t updated this blog soon enough for all of you curious people!!  (Because I just know you\'re curious.)  There is a fine line between curiousity and nosiness.  Keep yo nose to ya self.  Wise words.  Anyway...Let\'s see.  I love getting voicemails on my cell phone.  It is very very exciting.  To all you people who know my cell number, please call it and leave an interesting message.  My little ditty dooty that answers when I\'m not there is still in progress.  I kind of messed it up... Too lazy to fix it!!  (Story of my life!)  I left myself three messages.  Now don\'t say that\'s weird too soon, because it is very addictive once you start.  First I told myself to change the ditty dooty (which I did).  You know how you sound so much different when you\'re recorded?  I hate that because I like the sound of my voice.  It\'s not too deep, not too high.  That\'s because it vibrates in your head, making it sound deeper than it really is.  I sound icky recorded!!  The other two messages are of me singing, because I\'ve never heard my singing voice recorded.  \"She is Always On My Mind,\" Phantom Planet and \"Excuse Me Mr.,\" No Doubt.  Lalala.  I sound like a little girl.  It ain\'t fair. :\'(    I am leaving tomorrow!  Yay!  I love flying.  It\'s so fun like Puzzle Fighter 2X!!  And Tekken 3, and Dynasty Warriors 4.  And any RPG game.  Yeah, I\'ve rekindled my relationship with the Playstation, but unfortunately I have the PSone, not the PS2, which really really really makes me want to eat my head.  (But don\'t worry!  That\'s virtually impossible!!)  Hehehe.    My sisters are very nice.  Not too much the Nina, but the Natalie is nice.  I\'m getting her a Game Boy!  The new one that folds, you know?  Yeah, and with an aluminum case, and I want to play Zelda!!  I have a craving for playing Zelda with the creepy owl and the evil chickens and the weird unidentifiable things that make you lose hearts.  That\'s the life.  \" Caught in the Rain ,\" Revis  Is it real, we\'re always the same We\'re almost unknown now Well I was caught in the rain Wasting my time on the ground Waiting the call, of what would you say? Can you come over? Well, I was caught in the rain Wasting my time on the ground.  [Chorus] It was you, who could get me high With whatever you say You tell me something real What we do, it doesn\'t matter now Whatever it takes, for you to stay with me.  Trading thoughts, crossing the room I saw you surrounded I was caught in the rain Wasting my time on the ground Feeling this fall, without an escape I almost let you down I was caught in the rain Wasting my time on the ground.  [Chorus]  It was you, who could get me high With whatever you say You tell me something real What we do, it doesn\'t matter now Well I was caught in the rain You make me feel You make me feel Make me feel  I was caught in the rain Wasting my time on the ground I was caught in the rain Wasting my time on the ground I was caught in the rain Caught in the rain.   That is a great song, but probably their only great song.  Tear.  I think I will include a recommended song in every entry, so the whole world will be listening to good music.  I think that is a good idea.    Well, I guess this is a good stopping point.  My booty is vibrating. That must mean someone\'s just pulled into the driveway!  (My sisters and dad.)  Yay! I\'m out my homies. P.S.- Today\'s my birthday.  '),(5405,1261679,'05,June,2003','2003-06-05 00:00:00','   \" Get Busy ,\" Sean Paul  Shake that thing Miss Kana Kana Shake that thing Miss Annabella Shake that thing ya Donna Donna Jodi and Rebecca  Woman Get busy, Just shake that booty non-stop When the beat drops Just keep swinging it Get jiggy Get crunked up Percolate anything you want to call it Oscillate you hip and don’t take pity Me want fi see you get live ‘pon the riddim when me ride And me lyrics a provide electricity Gal nobody can tell you nuttin’ Can you done know your destiny  Yo sexy ladies want par with us In a the car with us Them nah war with us In da club them want flex with us To get next to us Them cah vex with us From the day me born jah ignite me flame gal a call me name and its me fame Its all good girl turn me on ‘Til a early morn’ Let’s get it on Let’s get it on ‘til a early morn\' Girl it’s all good just turn me on  Woman don’t sweat it, don’t get agitate just gwaan rotate Can anything you want you know you must get it From you name a mentuin Don’t ease the tension just run the program gals wan pet it Just have a good time Gal free up on ya mind caw nobody can dis you man won‘t let it can You a the number one gal Wave you hand Make them see you wedding band  [Chorus] Yo sexy ladies want par with us In a the car with us Them nah war with us In a the club them want flex with us To get next to us Them cah vex with us From the day me born jah ignite me flame gal a call me name and its me fame Its all good girl turn me on ‘Til a early morn’ Let’s get it on Let’s get it on ‘til a early morn\' Girl it’s all good just turn me on  Woman Get busy , Just shake that booty non-stop When the beat drops Just keep swinging it Get jiggy Get crunked up Percolate anything you want to call it Oscillate you hip and don’t take pity Me want fi see you get live ‘pon the riddim when me ride And me lyrics a provide electricity Gal nobody can tell you nuttin’ Can you done know your destiny  [Chorus] Yo sexy ladies want par with us In a the car with us Them nah war with us In a the club them want flex with us To get next to us Them cah vex with us From the day me born jah ignite me flame gal a call me name and its me fame It’s all good girl turn me on ‘til a early morn’ let’s get it on let’s get it on ‘til a early morning girl…it’s all good just turn me on  Yo, Shake that thing  Miss Kana Kana shake that thing Yo, Annabella shake that thing Miss Donna Donna  Yo Miss Jodi yu’r di one and Rebecca shake that thing  Yo shake that thing  Yo Joanna shake that thing Yo Annabella shake that thing Miss Kana Kana   [Chorus 2X] Yo sexy ladies want par with us In a the car with us Them nah war with us In a the club them want flex with us To get next to us Them cah vex with us From the day me born jah ignite me flame gal a call me name and its me fame Its all good girl turn me on ‘Til a early morn’ Let’s get it on Let’s get it on ‘til a early morn\' Girl it’s all good just turn me on  Let’s get it on ‘til a early morn\' Girl it’s all good just turn me on  That is the song of summer \'03.  It is the greatest.    Summer is lagging a bit.  I am leaving for Los Angeles, California in six days.  I am going to have sooo much fun!!  Yeah.  There is no more drama in my life and nothing to write about.  Sorry if this is boring.    You know what song be the best situation?  Let me copy and paste it for you ;) (hehehe): \" Without You ,\" Justincase feat. Michelle Branch It\'s been two weeks Three days long I\'m all alone since you\'ve been gone I can\'t keep myself from asking why Just wanna see my reflection In your eyes  Just know that I\'m with you And what you feel I feel it too  (Chorus)  I\'m feeling lately I\'m going crazy Wondering how I\'ll go on without you I keep on trying Slowly dying Thinking of how I need you here with me  I wear your shirt The scent of you Reminds of the night that we became two There\'s nothing else  That feels the same Everyday there\'s sun, it feels just like rain  Just know that I\'m with you And what you feel I feel it too  (repeat chorus)  If you think it\'s easy for me Well, you\'re wrong Sometimes I\'m weak and baby you know I\'m not that strong  You\'re in New York And I\'m in L.A. There\'s so much to tell you Words I need to say But your phone keeps ringing [ ringing, ringing ]  (repeat chorus)  I need you here with me  How will I go on without you?  It\'s been two weeks Three days long I\'m all alone since you\'ve been gone   I hope that happens to me later in life.  Feel free to download that.  It\'s a bit sappy, but, as I said before, the perfect situation.  I\'d be in New York, an absent boyfriend in LA... Fantasizing future situations is a hobby of mine, unfortunately.    I have another topic to address.  Why can\'t anyone spell the word  definitely ?  NO ONE CAN!!  It\'s a common trait.  My mom even said no one could spell at her workplace either.  I\'m just wondering.  It\'s always spelled  definately, definatly , or  defanitely .  I don\'t know, it\'s just something I\'ve notice and never expressed.    I learned how to play VC (Vietnamese Cards)!!  It is a very easy concept... I wanna play Tammy to see how good I am!!  She\'d start with an 8-card straight or something, then I\'d pass, then she\'d go again with triple 7\'s, then I\'d beat her 7\'s with Kings, then she\'d pass, then I\'d go with a 6-card straight, she\'d pass, a suit (or family), she\'d pass, then I\'d put down a low single card, she\'d put down her 2 of hearts, then I\'d slam it with four 6\'s, then I\'d put my final card down (a nine)... hahaha.  I have it all planned out.  (You better watch out Tammo.)  I\'ve recently become addicted to KleinForums.com.  It\'s all Tammy\'s fault too!  Hahaha, ok.  No one knows who I am on the forums either.  (Except for Dan R.)  It\'s fun!!   I have a feeling someone will ask soon.  I hate giving away my name.  They\'re like, \"Whaa?  That\'s not your name!\" or \"Are you that short girl who always so quiet in class?\" or \"I\'ve heard of you before,\" when I\'ve known her/him for three years.  It\'s depressing yet refreshing.  Yes, both.  First you get depressed, then you get refreshed.  It\'s just like a Mr. Goodbar:  \"First you taste the peanuts, then you taste the chocolate.\"  It\'s actually logically the other way around, but the slogan flows better.    I\'m trying to learn some French before my trip to Montreal, too.  I don\'t know any though, so there was no point in me starting a new paragraph.  (AHH!!! Lawn mowers outside= loud.  Good grief.)  Nothing else to say, so I will leave you satisfied with a farewell and an extra random statement.  Yours Truly,    amanda    PS: Britney Spears cursed us all.  '),(5406,1261679,'04,July,2003','2003-07-04 00:00:00','   It\'s Independence Day!  Most people refer to it as \"The Fourth of July,\" not realizing that the whole point of the holiday is celebrating our independence from England.  Commercialized holidays bug me.  (Valentine\'s Day, St. Patrick\'s Day, Halloween, Christmas [now, in the 21st century]...I mean, come on now.)  I guess I\'m bitter about Valentine\'s Day because nothing ever really happens for me, but Christmas is getting way too butchered nowadays.  I read something that said:  \"Christmas isn\'t just about getting presents and egg nog.  The real background is Christ\'s birth.  CHRISTmas.  Get it?\"  I was like, \"whoa.\"  I think I read that in a Teen magazine.  It\'s funny how ignorant people are now, especially the teenagers.  Oh well.  It\'s really sad, though.  All teenagers are referred to as a specific demographic, fair enough, but when our knowledge is also referred to as common ignorance by adults, it becomes even more depressing.  In Time magazines and such periodicals, there are these statistics claiming that only a very small minority of 13-19 year olds know about what\'s going on in the world.  Talk about misrepresentation.    Anyway, I promise I\'ll stop turning my blog entries into editorials.  ;)   \" The Best of Me ,\" by The Starting Line tell me what you thought about when you were gone  and so alone  the worst is over  you can have the best of me  we got older  but we\'re still young  we never grew out of this feeling that we wont give up   here we lay again  on two separate beds  riding phone lines  you hear a familiar voice  and pictures drawn from memory  we reflect on miscommunications  and misunderstandings  and missing each other too  much to vow to let go   turn our music down  and we whisper  \"say what you\'re thinking right now\"   tell me what you thought about when you were gone  and so alone  the worst is over  you can have the best of me  we got older  but we\'re still young  we never grew out of this feeling that we wont give up   jumping to conclusions  made me fall away from you  I\'m so glad that the truth has brought back together me and you   we\'re sitting on the ground and we whisper  \"say what you\'re thinking out loud\"   tell me what you thought about when you were gone  and so alone  the worst is over  you can have the best of me  we got older  but we\'re still young  we never grew out of this feeling that we wont give up   turn our music down  and we whisper  we\'re sitting on the ground  and we whisper  we turn our music down  we\'re sitting on the ground  the next time I\'m in town  we will kiss girl  we will kiss girl   tell me what you thought about when you were gone  and so alone  the worst is over  you can have the best of me  we got older  but we\'re still young  we never grew out of this feeling that we won\'t  feeling that we can\'t  that were not ready to give up   we got older, but we\'re still young  we never grew out of this feeling that we wont give up   I don\'t really like this kind of music (The so-called punk rock with bands that can\'t be distinguished from each other), but I kept seeing the music video and the song kind of grew on me.  By the way, the situation in the song has no relation in my life whatsoever.  God, I have no life.  He he he.  Goodbye.    PS- For some great insight on life, go to  urlLink Esin\'s blog .  To read about a girl on pseudo-cocaine, go to  urlLink Elise\'s blog .  (Hehehe)  And just go to  urlLink Tammy\'s xanga site .  The layout and the music is just so cool.  '),(5407,1261679,'03,July,2003','2003-07-03 00:00:00','   Do you know what I really hate?  Of course I\'m asking something that cannot be answered by the queried, but it\'s just for emphasis.  I hate leeches.  Yuck.    So...I drank a Coke mindlessly and it is now tomorrow.    \" Love Song For No One ,\" John Mayer  Staying at home alone on a Friday Flat on the floor looking back On old love Or lack thereof After all the crushes are faded And all my wishful thinking was wrong I\'m jaded I hate it   I\'m tired of being alone So hurry up and get here So tired of being alone So hurry up and get here  (can you)  Searching all my days just to find you I\'m not sure who I\'m looking for I\'ll know it When I see you Until then, I\'ll hide in my bedroom just staying up all night just to write A love song for no one   I\'m tired of being alone So hurry up and get here So tired of being alone So hurry up and get here   I could have met you in a sandbox I could have passed you on the sidewalk Could I have missed my chance And watched you walk away?(2 times)   I\'m tired of being alone So hurry up and get here So tired of being alone So hurry up and get here You\'ll be so good You\'ll be so good for me   I\'m done.  So now I will leave you with a random phone number: 713.530.3486  Ha ha.  And still I get no calls!  (Just kidding.  I am quite the shiznit.)  '),(5408,1261679,'13,August,2003','2003-08-13 00:00:00','   This past afternoon I wanted to bake chocolate pretzels, so I stirred and mixed and rolled and kneaded and preheated and refrigerated and popped them in the oven.  So when they were ready, I took the fattening suckers out with oven mits.  THEN, a very terrible thing happened:  the 375 degree cookie sheet slipped a little and I caught it with my bare forearm.  Talk about third-degree burn.  Now I have a two inch long, centimeter-wide grayish welt on my poor, cute, little arm. :\'(  It\'s all good.  It\'ll look cool if it becomes a scar.    Anyway, I just remembered something from my trip to Montreal this past June.  I was at McDonald\'s for breakfast and my family and I got Egg McMuffins and hashbrowns.  When I finish the hashbrown, I don\'t finish the Egg McMuffin, and vice versa.  Well that day, the case was that I didn\'t finish my hashbrown and ate the whole Egg McMuffin.  So I was sitting there at the small table in the semi-spinny chair connected to the table just watching all the odd French-Canadian people walk by.  I actually saw a dark-gothic, loving couple and the guy was wearing a long, black skirt.  I thought that was pretty odd, but I quickly remembered that I was in an odd world.  Anyway, back to the original story.  My parents were renting a car to drive to Quebec City and my big sister was in the bathroom, so I was stuck with my annoying, ill-tempered, premenstrual, little sister.  I saw another person pass by with a lot of bags.  He didn\'t look very clean-shaven, or, rather, clean.  He went to the trash can where I thought he was throwing away all his trash bags.  When I was getting ready to throw the 1/4 hashbrown I didn\'t eat, he came up to me and said, \"Are you going to throw that away,\" and I said I was and he asked to have it.  I was so sad that I finished my sandwich instead of my hashbrown.  It turns out that he was digging inside the trash cans for scraps of food to eat.  I  hate  it when it seems like I don\'t care for the poor.  I don\'t have any money to offer to the poor on the street most of the time, and we just drive by the people who beg for jobs just to get food.  (**Heavy sigh**)  On a lighter subject, I\'ve noticed that when girls have to leave, they say, \"I g2g!\"  But when guys have to leave they say, \"I gtg.\"  Why a \"t?\"  I will never know.  I guess it\'s guy thing.  (This was pointed out by one of my friends, too.)  Now it\'s music time!  \"Thank You,\"  by Led Zeppelin  If the sun refused to shine, I would still be loving you.  When mountains crumble to the sea, there will still be you and me.   Kind woman, I give you my all, Kind woman, nothing more.   Little drops of rain whisper of the pain, tears of loves lost in the days gone by.  My love is strong, with you there is no wrong,  together we shall go until we die. My, my, my.  An inspiration is what you are to me, inspiration, look... see.   And so today, my world it smiles, your hand in mine, we walk the miles,  Thanks to you it will be done, for you to me are the only one.  Happiness, no more be sad, happiness....I\'m glad.  If the sun refused to shine, I would still be loving you.  When mountains crumble to the sea, there will still be you and me.  ____________________________________________________ Reduce your banana intake to one a week.  They can\'t reproduce by themselves and they might go extinct.  -Amanda  (Oh yeah, this burn\'s gonna scar.)  '),(5409,1261679,'10,August,2003','2003-08-10 00:00:00','   Well, I know all of you are wondering about this next topic and will just read this over and over again.   Things that bug me:  -As I have mentioned in a previous post, I just hate it when people step on the back of my shoes or on my jeans.  Ugh!!  I don\'t know why this bugs me so much, but please, everybody, avoid it for the better of the universe. -When people try to sound intelligent by using words that aren\'t usually used in normal language it just bugs me so freaking much.  First of all, they don\'t  use them correctly, and second, they talk loud enough for everyone to hear their wrong diction.   -There was a kid in my English class who always asked me the same question over and over again everyday until he moved out of my class: \"Hey Amanda, what do we do in Biology today?\"  And that\'s all he needed to know from me.  I don\'t know if he was always trying to start a conversation or something, but it really annoyed me so I started telling him that she didn\'t collect the homework that day or \"No, don\'t bring your book today,\" because if that\'s all ya wanna know, eat poo. -I hate it when people do the \"|_|\" thing with their hands at Good Charlotte or Avril Lavigne concerts.  It kills me. -Generalizations kill too.  That whole \"Asians are quiet\" thing in elementary school just scarred me for life.  In my classes I was the \"quiet person,\" instead of \"that awesome girl Amanda,\" which no one would ever say anyway, but, you know, I\'m complaining here.  I mean, I\'m not too shy, but I\'m not \"lol, OMG!\"-outgoing, if you know what I mean. -Pessimistic doctors have no right to be in their profession.  If you know me, you know that I\'ve hated my height (until now), and I\'ve always complained about it.  I\'ve had X-rays and they showed that I had a teeny bit of growth left (which was a half inch) and my doctor kept saying, \"It\'s not that possible for you to grow three inches to reach five feet in time for your growth to stop,\" and I was all :\'( and he was all :D.  I just went again last Wednesday and he asked if I\'d grown and I said \"A half inch,\" and he said, \"The  last   half inch?\"  I was shocked at his negativity, first of all, and second, I was wondering why he asked me a question like  that .  I mean,  hellOoOo,  I didn\'t want to be short!  But I didn\'t care at that time anymore.  People say things like, \"You\'re cute like that,\" or \"You wouldn\'t be you if you weren\'t short, so don\'t change,\" and there are other people who haven\'t seen midgets, or \"little people,\" as they want to be called now, that call me a midget, which I don\'t think is funny or nice.  I don\'t waddle or have improportional limbs!  It\'s not that that\'s a bad thing to do or have, but I\'m just 4\'10,\" man. -That was long...anyway, there\'s one other thing that I can think of that bugs me.  CosmoGirl is a great magazine and yadayadayada, but they print there sad and embarassing and perverted stories with imagery that is too sexually graphic.  No one wants to read  that  much about a rape or a molestation!  It bugs me to think that people actually would.  A guilty pleasure of mine is to take weird psychological tests to find out more about yourself, like this one I took last night: I was asked to pick a color and use three adjectives to describe it.  I picked mauve and said it was different, unclassified, and pretty.  Then I was asked to pick and animal and describe it with another three adjectives.  I picked the Emperor Penguin and said it was small, cute, and adorable.  Results: The description of the color is how you see yourself, so I see myself as different, unclassified, and pretty.  The description of the animal is how others see you or what you think you\'re lacking.  I said the animal was small, cute, and adorable.  Go figure. That one turned out to be believable.    Now I will sing you a song... \" When I Look In Your Eyes \"  When I look in your eyes, I see the wisdom of the world in your eyes  I see the sadness of a thousand goodbyes  When I look in your eyes   And it is no surprise, to see the softness of the moon in your eyes  The gentle sparkle of the stars in your eyes  When I look in your eyes   In your eyes, I see the deepness of the sea  I see the deepness of the love  The love I feel you feel for me   Autumn comes, summer dies  I see the passing of the years in your eyes  And when we part there will be no tears no goodbyes  I\'ll just look into your eyes   Those eyes, so wise  So warm, so real  How I love the world, your eyes reveal ______________________________  Beautiful. -Amanda  '),(5410,1261679,'08,August,2003','2003-08-08 00:00:00','   ¿Qué pasa, jugadoras?  I haven\'t posted in over a month, sorry.  I have just been sooo, incredibly lazy.    So now I\'ll mention a topic and argue about it.  Talentless pop-idol money moguls.  I really don\'t understand why Hilary Duff should go into singing.  I\'m her age and I can sing a hundred times better than she can.  I guess if you want to sing about things you didn\'t write and haven\'t experienced  and  you don\'t want to play any instruments, you should sell yourself short like that.  I guess she has an okay voice, but I really don\'t think she should profit off of it.    -Manda  '),(5411,1261679,'06,September,2003','2003-09-06 00:00:00','   I am so very tired today.  Sleep-deprived people are not the most energetic.  I hate this cycle: going to sleep at midnight because I can\'t stay awake to finish homework, waking up at six to catch the bus thirty minutes later, then falling asleep during chemistry, algebra II, and world history.  Then during the weekend, every thirty minutes I slowly drift off to sleep and wake up with a bruise on my head.  It\'s only the beginning of the freaking school year and man, am I already a living-dead-girl.    Anyway, off to another subject.  I hate it when something has run out-- forever.  Like extinct animals.  I just wish I could see a real-life saber-toothed tiger or maybe a dodo bird.  It\'s really upsetting, and I just think that I\'ve missed out on something really cool.   I don\'t know why, but when someone is really good at something like performing, I think they\'ll never perform to their full potential if they know someone is going to hear.  (Like singing.)  :-\\.    \" Wasting My Hate ,\" Metallica Good day  How you do  When I send a smile to you Don\'t Waste waste your breath,  And I won\'t waste my hate on you  Ain\'t gonna waste my hate Ain\'t gonna waste my hate on you I think I\'ll keep it for myself  Ain\'t gonna kill no more Ain\'t got the time to help you score I think it\'s time you pleased yourself Yourself  Good day  How you do  When I send a smile to you Don\'t Waste waste your breath,  And I won\'t waste my hate on you  And I think your working now You think enough to even raise the brow And to life that tip that you belong Come on  When I see my hate I see my feet I feel that blood that pumps in me Where the hell\'s my mind Don\'t know Don\'t know now  Good day  How you do  When I send a smile to you Don\'t Waste waste your breath,  And I won\'t waste my hate on you  Then I\'ll keep it for myself Hate  Ain\'t gonna waste my hate But I don\'t clean it when they say Better to give then to receive  Ain\'t gonna waste my hate Ain\'t got time to waste my hate on you I think I\'ll keep it all for myself For myself  Good day  How you do  When I send a smile to you Don\'t Waste waste your breath,  And I won\'t waste my hate on you  Think I\'ll keep it for myself Yeah  Hate  ________________________________ This, too, shall pass.  -amanda-  '),(5412,1261679,'21,October,2003','2003-10-21 00:00:00','   Today I came home and I ate a banana and two Kit-kats.  Then I went upstairs and watched Camp Jim.  I was planning to go to sleep at around 4:30 and wake up at 5:30, but while watching Room Raiders I drifted off to sleep.  I never fall asleep in front of the tv.  That is very odd.  I woke up at 7 and watched Gilmore Girls.  Finally Rory goes through something imperfect and real!  Awkward dates: how scary i think of them to be.    I am just having too much \"me\" time because I am finding out so much about myself.  A lot people think of me as \"quiet,\" and not very outspoken.  Ha ha!  Those people are just the ones I never feel like talking to.  And sometimes I just don\'t feel like talking.  It\'s nothing personal.  I enjoy being alone.  I don\'t feel the need to introduce myself to everyone that passes and I certainly don\'t see the use in striving to \"climb the social ladder.\"  I don\'t even like ladders.  I\'m just a natural loner, if that makes sense.  It\'s not depressing either because I do have friends.  It\'s not like I shoo people away when I don\'t feel like talking, you know?  It\'s that simple.  I should go do homework now.  I should also try raising my GPA.  It\'s a frickin 4.4.  That\'s not very good, but oh well.    Have a day.  :-| -Amanda-  '),(5413,1261679,'19,October,2003','2003-10-19 00:00:00','   I am so lazy.  It shows.  I don\'t do a lot of my homework, I don\'t bother studying, I hate doing chores (washing dishes, laundry), and I stay in bed 5 minutes before we have to leave for church and school.  Yeah.  It\'s really sad.  I think I might have gained a pound or two too.  Darn!  I am supposed to be on a diet but it\'s frickin Halloweenish time and my dad bought a buttload of candy.  Kit-kats, Snickers, Starburst, drooling excessively... Maybe this diet will start after October.  Wait- no, cause there\'s Thanksgiving.  After Thanksgiving.  (Whatever.)  Some say procrastination is laziness.  (Like my parents.)  But sometimes during procrastination, you end up being productive, so ha!  Contradicted!    My hair smells really good!  I wish there was some sort of scratch and sniff technology that could be incorporated into this website.  Unfortunately, there is nothing like that so you can\'t smell my hair.  (Like I would want you to!  Yuck!)  I wonder if anyone else could be as excited as I am when a new episode of Degrassi or Radio Free Roscoe is airing.  Channel 298 on satellite:  the best channel!  Hmm.. my favorite shows:  1) Degrassi  2) Radio Free Roscoe 3) Alias 4) The Newlyweds 5) Alias 6) As Told By Ginger 7) Even Stevens  Wow, I find out more and more about myself when I list things.   __________________________________________________ What is up with The Cheetah Girls??   He he he.  -Amandio  '),(5414,1261679,'25,November,2003','2003-11-25 00:00:00','   innocent kiss - you\'re cute and sweet and like it that way    urlLink   What Sign of Affection Are You?     brought to you by  urlLink Quizilla    Well anyway, I was getting a ride home from a friend and he was surprised because I knew the band we were listening to.  Everyone knows The Smiths.  Come on.  \"How Soon is Now?\"  The Smiths  I am the son and the heir of a shyness that is criminally vulgar I am the son and heir of nothing in particular  You shut your mouth how can you say I go about things the wrong way I am human and I need to be loved just like everybody else does  I am the son and the heir of a shyness that is criminally vulgar I am the son and the heir of nothing in particular  You shut your mouth how can you say I go about things the wrong way I am human and I need to be loved just like everybody else does  There\'s a club if you\'d like to go you could meet somebody who really loves you so you go, and you stand on your own and you leave on your own and you go home, and you cry and you want to die  When you say it\'s gonna happen \"now\" well, when exactly do you mean? see I\'ve already waited too long and all my hope is gone  You shut your mouth how can you say I go about things the wrong way I am human and I need to be loved just like everybody else does ____________________________  I am thankful for Spark Notes.  '),(5415,1261679,'21,November,2003','2003-11-21 00:00:00','   I am 13th chair (out of 16) in the All-Region Orchestra.  Joy.  And to think I just got that music on Monday.  I am so frickin proud of my wonderful self.  But the rehearsals, and those to come, are real time-wasting, butt-hurting, tiring, stupid pains.  It would have been so much better spent if I had my prodigiously talented yet jaded friend Tammy.  How her proximity consoles me.  I indulge in solitude.  Does that make me intellectual or antisocial?   ____________________________________________________  (It\'s rhetorical.)  '),(5416,1261679,'19,November,2003','2003-11-19 00:00:00','   Well, right now I am being kept awake while my conscience tells me to finish my pile of homework.  It is literally a pile.  Of homework.  Unfinished.  And while my conscience is whispering, I am updating my blog and thinking of ways to waterproof my dollar-eyeliner.  Figures, considering the rebel I am.    I am supposed to be studying for Spanish, first of all, because I haven\'t even begun to translate the culture section we\'re being tested over.  I barely know how to spot the direct and indirect object pronouns, but cut me some slack here, for I am taught by a squawking parrot from Kentucky.  And she seems to be going through menopause.  Her accent is horrible, man.  I think she loses it by bragging about the years in college she\'s spent studying Spanish in between every sentence.  Wow, I am evil.  But I am also premenstrual at the moment, so cut me some slack.  And in chemistry, my goodness.  What is up with Miz Rushin emphasizing the second syllable of H2O?  Wa ter .  (Mind-boggling.)  Shame on me!  I have done nothing productive since dinner!  (That was a couple of hours ago.)  Time to go give my nightly shower concert.  Or countdown, perhaps.  My bathroom has the best echo. ____________________________________________________________  We\'re having the best weather this week.  '),(5417,1261679,'18,November,2003','2003-11-18 00:00:00','   Afterlife as an Angel by childdoll      Your Name           Astrological Sign           Angel Type    Fallen Angel      Wing Color    Pink      Heavenly Weapon    Axe             Created with  urlLink   quill18  \'s  urlLink  MemeGen  !       Oh darn.  I\'m gonna be evil  after  I die?  What a waste of life.  '),(5418,1261679,'17,November,2003','2003-11-17 00:00:00','   Oh my god.  I hate that stupid commercial for Foley\'s with the walking giftboxes.  They\'re all \"Why is everyone staring?!\"  It\'s because you\'re frickin WALKING, TALKING INANIMATE OBJECTS.  That is just the beginning of the commercialization of holy holidays.   Well, despite my viewing of that horrendous display of \"yuletide humor,\" today was such a beautiful day.  You have never truly lived if you haven\'t splashed around in the Klein High School courtyard.  If it really floods, like  really  floods, the school administration should supply inflatable rafts and kiddie pools.  Oh, PARADISE...  Today, I have decided to be a jazz pianist/vocalist at high-class clubs in New York or Chicago.  I wonder what I\'ll be tomorrow.  A meteorologist, perhaps.    Mad tv was great Saturday night.  I just remembered the part where the girl turned into a werewolf when she had a craving to spend money.  Hahaha.  I know I might sound like a slave for the mainstream but Hilary Duff\'s performance on the AMA\'s was baaaaad.  Her voice track was playing so naturally you\'d think she was gonna lip-sync, but instead she SANG ALONG WITH IT.  You could actually hear her trying to find the notes.  But I\'m sure it was just nerves.    Shoot, it\'s late and I\'m still wide awake with nothing to do or say.  I think it was that three hour nap I took when I got home from school.  Or maybe that Diet Coke I drank after dinner.  Or maybe just the thought of you reading this and the concept of me revealing my every thought to a total stranger...    Haha, WHATEVER.  I\'m hyper and rambling.  Who are you?  I wonder...  I hate using this ellipsis... ____________________ Good night.  Or morning, afternoon, evening.  Whatever your case may be.  I wanna go shopping like a mofo.  AND I\'m out.  '),(5419,1261679,'16,November,2003','2003-11-16 00:00:00','   I don\'t know what\'s wrong with me!  I keep practicing piano like it\'s normal.  On Friday I came home and played for three straight hours.  Haha, whoa that looks scarier written.  I\'m watching a boxing match right now.  Those cuts are nasty.  Hemoglobin and plasma and AIDS and sickle-cells and STD\'s... icky.  I don\'t want to be Leila Ali no more.  I actually did want to be a boxer at a time.    And Esin, come on!  You are NOT gonna turn into  urlLink Juan Moreno !  I really can\'t see THE Esin  running a KFC, and you know you\'re getting your feelings returned! ;)  (Or, at least I know, and everyone knows that my opinion is the only one that matters.  He he he..)  This has been stuck in my head: \"I Think I Love You,\"  The Partridge Family  I\'m sleeping  And right in the middle of a good dream  Like all at once I wake up  From something that keeps knocking at my brain  Before I go insane  I hold my pillow to my head  And spring up in my bed  Screaming out the words I dread:  \"I think I love you!\"   This morning, I woke up with this feeling  I didn\'t know how to deal with  And so I just decided to myself  I\'d hide it to myself  And never talk about it  And didn\'t I go and shout it  When you walked into my room.   I think I love you  So what am I so afraid of?  I\'m afraid that I\'m not sure of  A love there is no cure for  I think I love you  Isn\'t that what life is made of?  Though it worries me to say  I\'ve never felt this way   Believe me  You really don\'t have to worry  I only want to make you happy  And if you say,  hey, go away, I will  But I think better still  I ought to stay around and love you  Do you think I have a case?  Let me ask you to your face:  Do you think you love me?  I think I love you!  _________________________________________ Darn boxing matches are so yucky.  Your lover, Amanda  '),(5420,1261679,'15,November,2003','2003-11-15 00:00:00','   You are Sandra Gibbs!!    urlLink   Which Completely Random Person Are You??     brought to you by  urlLink Quizilla    Man, that is harsh.  '),(5421,1261679,'11,November,2003','2003-11-11 00:00:00','   Hmm... It is exceptionally late right now.  I have not even begun to finish my homework.  Shoot.  Well there\'s always morning!  (In a couple of minutes.  Shoot.)    I\'ve relapsed into this utterly annoying sickness.  I can\'t believe it.  I couldn\'t even remember the feeling.  Maybe it\'s because I\'ve never felt this way before.  '),(5422,1261679,'10,November,2003','2003-11-10 00:00:00','   I wish there was a symbol for a less emphasized exclamation mark.  '),(5423,1261679,'03,November,2003','2003-11-03 00:00:00','   It\'s Homecoming Week at Klein High!    Monday-Nerd Day Tuesday-80\'s Day Wednesday-What I Wanna Be When I Grow Up Thursday-Sweats Day Friday-Hawaiian (Spirit Day)  I don\'t think the people who dressed up for \"Nerd Day\" knew that they were making fun of the actual nerds.  And eagles don\'t sweat.  So I\'m dressing going goth on \"Sweats Day.\"  I actually want to be a pharmacist/doctor/struggling musician when I grow up.  Should I wear a stethoscope and a guitar?  (Heh heh, no.  That would be very sad.)  I\'m not very school spirited.  I\'m going to homecoming, but the night before I\'ll be at home at 7 pm watching Holes.  No, I am not going to the game with my date.  I\'ll skip the mindless cheering and stay away from the those two scary colors.  (Blue and yellow...*shudder*.)    Greek play today: A hilarious mess.  People should\'ve paid to see it.  My role was to kill Prince Davidius.  Every killing needs a battle cry, right?  He he... (\"You will pay with your life for what you have done!  EEEEE!!\" *stabbed* [Fall dead.])  Ahh..haha.  Great stuffing.    I\'m wondering if you can incorporate any html in this website...  '),(5424,1261679,'03,December,2003','2003-12-03 00:00:00','   WHOA  I haven\'t done any homework.  I don\'t know why I call it homework anyway because I always end up doing it the day it\'s due at school.  Hmm.    Post-Lab quiz tomorrow!  Yay!  Another quiz that I can completely and totally guess on! Merheherherher.  (Weird laugh, won\'t happen again.)  I hate having 5 minutes in the morning to get ready.  It sucks.  I wake up, I turn my blaring radio down, and I sit there half asleep and stare at the minutes go by thinking I have forever to primp.  Then I realize I don\'t and jump out of bed and hit my head on my closet door.  It happens everyday.  Every single day.  Believe it.  SoOoOooOoo... (awkward silence.) -_-;; ____________________________ Let\'s hang out again sometime.  I had fun.  '),(5425,1261679,'03,December,2003','2003-12-03 00:00:00','   I didn\'t eat breakfast.  I didn\'t eat lunch.  What does this mean?  I went to Chick-Fil-A afterschool today.  Why are their chicken nuggets so frickin small?    I came home around 4:30 and I watched Room Raiders.  I don\'t know why.  I think I was too lazy to change the channel.  I accidentally turned off the TV, so I left it off because I was too lazy to turn it back on.  I took a two hour nap in which I remember my dad yelling at me from my doorway complaining about my tornado-struck room.  Then I woke up at 6:30 thinking \"Oh my frickin gosh I\'m gonna be late for school!!!\"  but really it was still today.  You should\'ve seen me snap back into realization.  A funny sight.  And now I\'m sitting in front of my computer staring back at this C-3PO Pez dispenser with no candy in it.  It\'s time for procrastination! __________________________________ The Anti-Gravity Boy is so cute!  '),(5426,1261679,'28,January,2004','2004-01-28 00:00:00','   I had so much fun yesterday.  I got a big fat pen from my sister!  It has a little Hello Kitty kitty stuck inside of it.  I don\'t think I could run Sanrio.  I would have never thought of putting the kitty inside the pen.  And making it so fat.  Fat, fat, fat.  That word is worth a million pictures.    You know, I\'ve become really insensitive.  I have nothing to look forward to.  It\'s a feeling without words to describe it because you don\'t feel anything.  Nothing.  But I do have nerve endings, so don\'t come up and start punching me; I\'m talking about the emotional type of feelings.    \"Devil May Care\"  No cares for me I\'m happy as I can be I learn to love and to live Devil may care  No cares and woes Whatever comes later goes That\'s how I\'ll take and I\'ll give Devil may care  When the day is through, I suffer no regrets I know that he who frets, loses the night For only a fool, thinks he can hold back the dawn He was wise to never tries to revise what\'s past and gone  Live love today, let come tomorrow what may,  Don\'t even stop for a sigh, it doesn\'t help if you cry That\'s how I live and I\'ll die Devil may care ____________________________________________ I really really like that song.    Apathetically yours.  '),(5427,1261679,'24,January,2004','2004-01-24 00:00:00','   This is one saying I really can\'t stand:  \"Good girls are bad girls who don\'t get caught [usually followed by some kind of winking emoticon].\"    Good girls are good girls.  Good girls are girls who don\'t do anything to ruin their reputations or make it seem like they are \"bad.\"  And if they do something bad, they are smart enough to cover it up.    Bad girls are just stupid.  '),(5428,1261679,'23,January,2004','2004-01-23 00:00:00','   The Advertising Slogan Generator    TD.cutout { font-family:trebuchet ms,verdana,arial,helvetica,sans-serif; border-style:dotted; border-width:1; border-color:#F1F1F1; } input { font: Arial; font-size: 10pt; color:222222; border-width:1; border-color:222222; border-style:solid; background-color:ffffff;}  ..button { font: Arial; font-size: 8pt; color:222222; border-width:1; border-color:222222; border-style:solid; background-color:ddddbb;}                 The Advertising Slogan  Generator     Updated 13th March 2003 :  Now with  503 slogans .                        Happiness is a Cigar Called Amanda.                   Word :         (or reload for another)        src=\"http://www.textads.biz/getNextAd.jsp?siteref=2f5caae260657105ba701db754047ef1&adNum=1\">           Recently :   amanda      amanda   amanda   mike carafas   emi  '),(5429,1261679,'23,January,2004','2004-01-23 00:00:00','   There\'s a girl in my Algebra 2 class who had a mental breakdown from boredom today.  I empathize with her like I do with little people (a.k.a. dwarves or midgets).  I mean, it\'s the same thing every single day.  I wake up stumbling and tired, go to school, and sit through every class.  I come home and sleep until dinner and stay up late doing homework, which makes me even more tired the next day.  The same exact thing every freaking day, except it gets worse as time passes.  Homework builds up, I get sleepier, the little sister gets more annoying, bed gets less comfortable...  I could go on and on.    So here\'s a seemingly unanswerable question: What cheers me up?  I have a list, don\'t worry. First of all, Frappuccinos.  They taste indescribable.  (Paradox!) Then there are those stores with the most amazing clothes. Playing jazz on piano. Playing Michelle Branch on guitar. Mexican food.  (Not just eating it, but the aroma and the presence of it is just wonderful.) The cute guys that turn my head when they pass me by. The cute guys that turn their heads when  I  pass them by. Compliments.  (The actual ones, excluding the ones about my clothes.) Friends and conversations with them. Nikka Costa, Thicke, Nine Inch Nails, and John Mayer. My pearly clips. The sight of [pretty] guitars. Alias, the show that airs on ABC at 8 pm on Sundays. And last of all, this song:  \" My Happiness ,\" by Powderfinger  I see your shadow on the street now I hear you push through the rusty gate Click of your heels on the concrete Waiting for a knock coming way too late It seems an age since I\'ve seen you Countdown as the weeks trickle into days  So you come in and put your bags down I know there\'s something in the air How can I do this to you right now If you\'re over there when I need you here  My happiness is slowly creeping back Now you\'re at home If it ever starts sinking in It must be when you pack up and go  It seems an age since I\'ve seen you Countdown as the weeks trickle into days I hope that time hasn\'t changed you All I really want is for you to stay  I know I know I know what is inside _______________________________________ I need a break from normality.  '),(5430,1261679,'20,January,2004','2004-01-20 00:00:00','   Hahaha.  That Juicy Fruit Strappleberry commercial cracks me up every time.  What about a flavor that\'s cranberry and apple?  Crappleberry!  Appetizing.  Wow, my hair smells really good.    I\'m really tired.  I got a new cd yesterday!  It\'s funny how a thin little product of illegal labor can make me so happy, as opposed to, well, people.  People fail to cheer me up nowadays.  I think I\'ve come to a reasonable conclusion.  I could be the most innocent and prude girl ever.  However, that was an extraneous statement.  What I mean to say is that teenaged guys suck.  I don\'t know why, but I seem to think that someone perfect will come around.  My perfect guy just completely blows all existing males out of the water.  Waaay out.    So Mr. Perfect Guy would be indescribable, but here are a couple of ideas: -He would have a low voice. -He would be able to sing well.  Really well. -He would be neither perverted nor prude. -He would not be profane. -He would have sensible style. -He would be unique and individual, and also respect that characteristic in others. -He would be faithful. -He would be at least 5\'10\". -He would single me out as someone very special to him. -He would not be conventionally cute, but a deeper kind of cute. -He would love me with all his heart.  And he\'s all I\'m asking for.  If only this was a perfect world.  I might add to that list everyday.    Naptime is about now.    A song: \"Goodbye To You\" by Michelle Branch  Of all the things I\'ve believed in I just want to get it over with Tears form behind my eyes But I do not cry Counting the days that pass me by   I\'ve been searching deep down in my soul Words that I\'m hearing are starting to get old It feels like I\'m starting all over again The last three years were just pretend And I said,   Goodbye to you Goodbye to everything I thought I knew You were the one I loved The one thing that I tried to hold on to   I still get lost in your eyes And it seems that I can\'t live a day without you Closing my eyes and you chase my thoughts away To a place where I am blinded by the light But it\'s not right   Goodbye to you Goodbye to everything I thought I knew You were the one I loved The one thing that I tried to hold on to   And it hurts to want everything and nothing at the same time I want what\'s yours and I want what\'s mine I want you But I\'m not giving in this time   Goodbye to you Goodbye to everything I thought I knew You were the one I loved The one thing that I tried to hold on to The one thing that I tried to hold on to   And when the stars fall I will lie awake You\'re my shooting star _______________________________________________ I\'ll be performing that soon.  Not that you\'ll hear or anything.  '),(5431,1261679,'19,January,2004','2004-01-19 00:00:00','   I hate cute strangers, straight up.  Like when you\'re walking around a store and BOOM:  Mr. You\'ll-Never-See-Me-Again-But-Keep-Checkin-Me-Out bumps into you.  I  hate  that.    How I yearn for my  urlLink green  Fender.  Of course, the parents have to say no to everything.  I also hate that.  I wish love was less used to describe feelings because when you say you love something, you can mean like or adore, but not actually love.  It\'s just so cliché that it has lost its meaning.  '),(5432,1261679,'13,January,2004','2004-01-13 00:00:00','   I have always wanted a trademark animal association with myself.  Well in my math class today, I was reading a biography of Boy George, innocently trying to decipher whether people in these pictures of his entourage were male or female.  Then all of a sudden a very credible person told me that I remind her of an otter.  Reluctant to thank her out of uncertainty of the comment\'s complementary content, I asked if I was like the \"PB&J Otters\" from the Disney Channel.  Fortunately I am not.  I am like a realistic otter.  So there you go!  Amanda the Otter.    So how did you like my avocado pants today?   And what are otters anyway?  '),(5433,1261679,'10,January,2004','2004-01-10 00:00:00','   Bad news guys!  I\'m not getting into Harvard.  I know, depression galore.    My report card is awful though!  I mean, I don\'t think I could\'ve done any worse.  Unless, of course, I actually failed the classes, which some people actually do but it\'s never occurred to me because it still seems almost impossible.    So I got a gummy-spiked yellow lightbulb for Christmas recently.  I put it into one of the light sockets of the sink part of my bathroom which I use a lot and I don\'t want it to burn out.  It\'s really cool looking but I\'ll probably forget that I have it.  But every once in awhile I\'ll look up and remember that I have a really cool lightbulb.  Kinda like the person who gave it to me except he\'s not a lightbulb.  I like my new chemistry class.  I sit by Blake, whom I never really liked, but realized yesterday he\'s not that bad.  \"Closer,\" Nine Inch Nails.  I would post it but it\'s a little inappropriate.  Disturbingly addictive and brilliant song.  You know why whenever you ask me if I have a celebrity crush I say I don\'t have one?  It\'s because I never remember them, and I can\'t have a crush on someone I don\'t know.  You weirdos!  _________________________________________________________ Is it just me or does Clay Aiken look really odd?  He also talks really oddly for a straight guy.  '),(5434,1261679,'22,February,2004','2004-02-22 00:00:00','   There are so many gigantic mosquitoes in my house.  It sucks.    So it has been an uneventful weekend.  I\'m really anxious about next week.  I suck at my solo.  My ensemble sucks.  I haven\'t started my persuasive paper.  But life is still as great as it was before.  I can\'t wait until my life isn\'t centered around school.  '),(5435,1261679,'18,February,2004','2004-02-18 00:00:00','   So last night, or morning (because it was 12:30-something), I was thinking about when I was finally going into REM.  But that thought just keeps you up all night.    \"Why are there so many  Songs about rainbows?\" -  The Rainbow Connection , made famous by Kermit the Frog  That one phrase repeats over and over in my head.  Constantly.  But I love it.  '),(5436,1261679,'16,February,2004','2004-02-16 00:00:00','   Something I\'ve noticed about myself is that I always say the things I know people will want to hear.  Just everything but how I really feel to benefit the told.  I\'ve been doing that for so long that it just comes out subconsciously.  So in a way, I\'m always lying.  =\\.    \"I was afraid to be alone.   Now I\'m scared that\'s how I\'d like to be.\" - November , by Azure Ray  '),(5437,1261679,'14,February,2004','2004-02-14 00:00:00','   I had never seen a dead person up close before yesterday.  I attended a wake for my dad\'s colleague, and I really didn\'t want to see the corpse, but it would\'ve been disrespectful to walk past and not pay my respects.  So for the first time in my whole life, I saw a dead person up close and personal.  I didn\'t even know him, but the stories that were told in memory of him made me cry.  They also made me want to spend more time with my family, because he was a father of three young children and died so suddenly.  (I think I suggested to go to the movies tonight, but I went with my family.)  Then after the wake, I went to this restaurant downtown owned by my dad\'s friend.  He covered the check!  But this creepy girl waited our table.  She laughed at everything: When she dropped the fork on the table, when she quickly  popped her head into the bathroom and left, when we refused to not pay.  The laughing was actually annoying giggles.  And she pronounced \"pudding\" pooding.  Hahaha, it was so funny.  This entry\'s song:  \" Back To You \" by John Mayer  Back to you It always comes around Back to you I tried to forget you I tried to stay away But it\'s too late  I\'m so good at forgetting And I quit ever game I play But forgive me, love I can\'t turn and walk away  Back to you It always comes around Back to you I walk with your shadow I\'m sleeping in my bed With your silhouette  should have smiled in that picture If it\'s the last that I\'ll see of you It\'s the least that you Could not do  Leave the light on I\'ll never give up on you Leave the light on For me too  Back to me I know that it comes Back to me Doesn\'t it scare you Your will is not as strong As it used to be __________________________________________ And now I\'m going to watch  Lost in Translation .  =)  '),(5438,1261679,'11,February,2004','2004-02-11 00:00:00','   I seem to be losing everything I own.  First my most beloved cd, then my pearly clips.  What next?  I will develop an allergic reaction towards coffee.  Now that I thought of that, it\'s going to happen.  Darn, I should stop thinking of worst case scenarios.  Like today in chemistry we were burning peanuts and I was wondering what would happen to me if my sleeve caught on fire.  Well, the flames would spread across my paper-thin shirt and scorch my chest, back, and arms.  The class would sit back in consternation and do nothing because 1+1=11 [to them].  So while they stared, the fire would spread to my face.  Then my smart teacher would finally call for help and I\'d be put out.  For the rest of my life, people would walk past me and stare at my hideous scars.    Unfortunately, my sleeve did not catch on fire.  Only the nut did.  Lucky nut.  '),(5439,1261679,'10,February,2004','2004-02-10 00:00:00','   It has been a mediocre week so far.  Full orchestra lasted short of forever today, but you get that feeling when it finally ends.  Like finding something long lost.  My parents have been overusing the karaoke ever since American Idol started.  My parents do not sing very well.  Wishing for a power outage.  I would be asleep by now if it was yesterday.  Which reminds me of a dream I had.  God, it was weird, and I think it definitely told me something.  But how can a dream be so direct in explaining your problems?  I guess sleeping earlier means clearer dreams.    I wish my blog knew how to play songs.  But sadly, its owner is computer illiterate.  So feel free find your satisfaction in the lyrics.  \" Lightning Crashes ,\" by Live  Lightning crashes, a new mother cries Her placenta falls to the floor The angel opens her eyes The confusion sets in Before the doctor can even close the door  Lightning crashes, an old mother dies Her intentions fall to the floor The angel closes her eyes The confusion that was hers Belongs now, to the baby down the hall  Oh now feel it comin\' back again Like a rollin\' thunder chasing the wind Forces pullin\' from the center of the earth again I can feel it.  Lightning crashes, a new mother cries This moment she\'s been waiting for The angel opens her eyes Pale blue colored iris, Presents the circle And puts the glory out to hide _______________________________ My lyrical interpretation skills are improving.  '),(5440,1261679,'09,February,2004','2004-02-09 00:00:00','   I lost my Hotel Paper cd.  =\'(  That cd is like life support for me.  Oh well.  If it loves me, it will come back to me.  Then again, cds don\'t have feelings.  So I\'m screwed.    Here\'s an awesome song from the wandering cd.  (My guitar song of the moment.)  \" Til I Get Over You ,\" Michelle Branch  Everytime I feel alone I can blame it on you And I do, oh You got me like a loaded gun Golden sun and sky so blue We both know that we want it But we both know you left me no choice You just bring me down So I\'m counting the tears \'til I get over you Sometimes I watch the world go by I wonder what it\'s like To wake up every single day Smile on your face You never tried We both know we can\'t change it But we both know we\'ll just have to face it If only I could give you up But would I want to let you off of this soapbox baby? We both know that we want it But we both know you left me no choice We both know that I\'m not over you I\'m not over you ____________________________________ Please pray that my cd comes back to me! I am ever so sad.  '),(5441,1261679,'08,February,2004','2004-02-08 00:00:00','   It was abnormally hard to stay awake today.  I was nodding off while taking chemistry notes.  \"1000 caloriasfndfncx equaallslaas a kilocalroiasdisjdv.\"  Hey, it makes sense to me.    Caffeine doesn\'t take effect on my anymore.  My sleepiness has finally overpowered that addictive drug.  No more night-mornings.  (Like when you stay up until 12am.)  I\'m gonna miss those.  Diligence does not belong to me anymore.  Oh well.  I had it long enough.  \" Life of Saturdays ,\" Dexter Freebish  I just wanna sail the seven seas and hide behind the sun, and never close my eyes, lay on my back and fantasize of the perfect place to be.  I wanna see the Mona Lisa, climb the pyramids of Giza on a moonlit night, yeah I just wanna fantasize of the perfect place to be.  I just wanna disappear and feel like I\'m free at last, The time is right, my day has come, it\'s gonna start today, my life of saturdays.  I wanna see the seven wonders, feel the rolling thunder of a Spanish sky. Lay on my back and fantasize of the perfect place to be. I wanna swim the deepest ocean, be poetry in motion on a paris night I just wanna fantasize of the perfect place to be.  I just wanna disappear and feel like I\'m free at last, The time is right, my day has come, it\'s gonna start today, my life of saturdays.  Go around the world in eighty ways. Never slow down, never fade away.  I just wanna sail the seven seas and hide behind the sun, and never close my eyes, lay on my back and fantasize of the perfect place to be.  I just wanna disappear and feel like I\'m free at last, The time is right, my day has come, it\'s gonna start today  I just wanna disappear and feel like I\'m free at last, The time is right, my day has come, it\'s gonna start today My life of Saturdays, yea  Oh where will I go Oh where will I go Oh where will I go ______________________________________________ That song makes me so jealous.  I want to live a life of eternal Saturdays too. =( Wow, that Elise is very bored.  30 messages in the past minute!  '),(5442,1261679,'08,February,2004','2004-02-08 00:00:00','   Where did my impetus for living go?  It\'s so sad.  Nowadays I lose it as often as I lose my scrunchies.  I hope these words become important someday.  This is a pretty pointless read.  Sorry for wasting your time, because I know I have been wasting mine.  '),(5443,1261679,'05,February,2004','2004-02-05 00:00:00','   Well I have my heart set on next week being a good week.  I hope it\'s a good week.    The Grammys are on tonight.  It\'s funny though.  When hip-hop is finally recognized, I dislike it.  Oh well.  I guess I won\'t watch this year.  All the annual TV broadcasts get suckier with time.  The last Grammys were good, though.  Music lifts my spirits like no person could ever.  Discovering all this \"old\" music is so refreshing.  I recently burned a cd for myself with a phenomenal compilation.  I have made five copies of the same cd because everybody likes it so freaking much.  Ahhh... I love my music.    \" Hurt ,\" by Nine Inch Nails  I hurt myself today To see if I still feel I focus on the pain The only thing that\'s real The needle tears a hole The old familiar sting I try to kill it all away But I remember everything  What have I become My sweetest friend? Everyone I know Goes away in the end  You could have it all My empire of dirt I will let you down I will make you hurt  I wear this crown of shit Upon my liar\'s chair Full of broken thoughts I cannot repair Beneath the stains of time The feelings disappear You are someone else I am still right here  What have I become My sweetest friend? Everyone I know Goes away in the end You could have it all My empire of dirt I will let you down I will make you hurt If I could start again A million miles away I would keep myself I would find a way. ____________________________ My silence is always to be taken personally.  '),(5444,1261679,'05,February,2004','2004-02-05 00:00:00','   Paper!  Snow!  A GHOST!  '),(5445,1261679,'04,February,2004','2004-02-04 00:00:00','   There is a strange sense of enlightenment in walking home in the cold.  Time goes by so fast, especially if you\'re walking in a straight line the whole time and you don\'t have to think.  The brisk air breathed in clears your mind and makes you happy.  Happy, happy, happy.  People keep petting me.  My velour sweater; my poofy jacket.  First it\'s complimentary, then after a little while it\'s a little disturbing.  Stop the touchy-touchy.  -Manda-  '),(5446,1261679,'16,March,2004','2004-03-16 00:00:00','   I didn\'t get to sleep until 3:00 am last night because I was afraid that someone would chop my head off or something close to that.  I was reading a creepy book.  I type this as my heartbeat suddenly speeds up and as my forehead drips with sweat.  I almost shoved the book in my freezer.  (Yes, it is that creepy.)  So after a fun day at the mall with two of the coolest people, I am exhausted.  I sigh deeply.  It is the good kind of Spring-Breaky exhaustion.  Now it is time to go scarf down some pepperoni pizza.  '),(5447,1261679,'14,March,2004','2004-03-14 00:00:00','   I really don\'t like drama.  Drama is annoying.  (And I mean the exaggeration of emotions kind of drama.)  So keep me oout of your problems and please don\'t complain to me.    I like things said to me with the intention of making me smile, although I\'ll smile either way because the muscular impulse has become reflexive to me.  Doesn\'t everyone like making people happy, not to mention  being  happy?  I adore happiness and I\'ll take however much I can get.  '),(5448,1261679,'09,March,2004','2004-03-09 00:00:00','   Spring Break is only three days away.  A week of paradise after a week of hell (as I predict).  People from New England call their week of paradise \"March Break.\"  Weird, isn\'t it?  Their Cokes are pops and their cools are wickeds.  But our southern dialect could be strange to them too.    So I found that Mrs. Stark is retiring after this year.  Can you imagine?  All those 125 years of Spanish study down the drain for some social security pension.  Or maybe she wants to go on a perpetual cruise.  That would be fun.  I don\'t blame her for choosing living on a cruise ship over teaching español.  I wonder if she can ever hear everyone bad mouthing her while she\'s writing on the board, because I\'m pretty sure everything said is within earshot.  Oh well; she should\'ve gotten the idea by now.  '),(5449,1261679,'07,March,2004','2004-03-07 00:00:00','   Paranoia is the result of bad situations.  I don\'t think people learn from mistakes; they just become more cautious.  Holding yourself in the protagonist position all the time sometimes is selfish, but it is pretty much inevitable.    Blah, blah, blah.   I worry too much about everything.  It\'s about time to whip out my Venezuelan Worry Dolls and put them under my pillow!  '),(5450,1261679,'03,March,2004','2004-03-03 00:00:00','   Suitors do not flatter; they annoy.    I yearn to view a horror film.  Dawn of the Dead.  Next week.  It looks so colorful and creepy.  I need to watch it.  Choppy sentences.  Kill my radio.  I wake up to it blaring in my left ear every morning and accidentally sleep through it, as loud as it is.  The songs haunt me in my sleep...  Silence is easy.  It just becomes me.  You don\'t even know me.  '),(5451,1261679,'28,April,2004','2004-04-28 00:00:00','   TAKS testing week is weird.  Too much Spanish yesterday.  Yicky.  I have a weird craving to speak in Indonesian-sounding gibberish.  But I remain silent.    I have two poems to write for English.  A total of 30 lines, 150 words, 2 or 4 figures of speech, due Friday.  I need inspiration.  Or depression.    Listen to this  urlLink song .  It makes me sway and sigh.  So does the thought of the soon summer.  All we can do is wait.  (Counting down doesn\'t make it come faster!)  '),(5452,1261679,'24,April,2004','2004-04-24 00:00:00','   Sorry about the boring white background.  Well, the background has always been boring, but I have been trying to get this really cool background to work, but everytime I go to the link, it freezes up the computer.  So I\'ll have to settle for this.  It\'s not too bad, actually.  (Oh yeah, this isn\'t important.  Read the below.)  '),(5453,1261679,'17,April,2004','2004-04-17 00:00:00','   Rowdy raindrops run roughly round.  The thumping thunder thanks them, thinking thoughts that threatened thrice.    The on and off shower didn\'t calm or cheer me up, surprisingly.  A lot doesn\'t anymore.  My \"favorite things\" have probably changed.  But how inconveniently timed.  I need a pick-me-up, but I don\'t even know what that is now.  I\'ll have to try some new things to figure everything out.  Last night was so fun.  Taboo and Guesstures are so fun when your closest friends are playing with you.  Then the evil Mr. Midnight crept up behind us and brought everyone home.  Split-ends galore.  Unhealthy hair begs for haircut.  Haircut is not yet accomplished.  Man, I hate my inability to tell people what I want to say.  I know what I want to say, I know when I want to say it, I know how I want it to come out, but when I come face-to-face with a person, I dismiss my prepared statement due to the easiness of its exclusion.  I guess do this to avoid being disliked, but consequently, I depise my stupid self.    \"We Never Change,\" by  Coldplay   I want to live life, and never be cruel And I want to live life, and be good to you, And I want to fly, and never come down, And live my life, and have friends around  We never change do we? No, no. We never learn do we? So I want to live in a wooden house  I want to live life, and always be true And I want to live life, and be good to you, And I want to fly, and never come down, And live my life, and have friends around  We never change do we? No, no. We never learn do we? So I want to live in a wooden house But making more friends would be easy  Oh and I don\'t have a soul to save, Yes and I sin every single day.  We never change do we? We never learn do we?  So I want to live in a wooden house, Making more friends would be easy, I want to live where the sun comes out. ____________________________________ You know what?  Sometimes ignorance  is  bliss.  '),(5454,1261679,'17,April,2004','2004-04-17 00:00:00','   Everything I encounter each day makes me rethink all the things I believe in.  It\'s amazing how quickly I can change my opinions about something or someone from the sight or sound of something accidentally convincing.  Revamping my thoughts isn\'t always a good thing, though, so the frequency of this occurrence somewhat depresses me.  (Lung-deflating sigh.)  \"The Happy Ending,\"  Phantom Planet   There\'s nothing for me here I\'m starting to get old Time just ticks on We\'ve all been taking some comfort in repetition And I\'m left with no way to scratch  What has been itching and I\'m I\'m ready and waiting For the happy ending Now the days go by They pass right through us Our night lives make us so useless Now don\'t take anything to heart We don\'t care We don\'t want to start Well, I\'m leaving for a while I\'ll head in some direction I just don\'t know which one I should be taking And I\'m ready and waiting For the happy ending I\'m begging and pleading For another beating I\'m ready and waiting For the happy ending __________________________________ Inconceivableness is upsetting.  '),(5455,1261679,'10,April,2004','2004-04-10 00:00:00','   Spring, Texas is a sad little dump.  I wonder how everyone came to live here.  I mean, of all places, you know?   My supposed cold, as opposed to allergies, is really getting on my nerves.  The icky, tickly throat drives me crazy.   I have no homework over this religious holiday.  Nothing to worry about.  =).    Nothing to write, either.  Bluh, bluh.  '),(5456,1261679,'07,April,2004','2004-04-07 00:00:00','   I don\'t think I\'m afraid anymore.  I just sleep with light and music on.  That\'s not too bad.  I\'ve been avoiding any type of medicine to cure my coughiness in high hopes that I might develop a fever and miss a day of school.  Nothing\'s happened yet, but I wouldn\'t want to miss tomorrow, so abandoned plan.  My chemistry class makes me so mad.  On Tuesday we had a test and a sub.  The two supposedly smartest people in my class took advantage of this unlikely combination (You know, test + sub) and cheated.  It\'s not fair.  I did my best and did not cheat, which I could have done very easily.  They got high A\'s.  I didn\'t do so hot.  Anger...  Chocolate bunnies!! And plastic eggs!!! And stuffed chicks!!  (What am I getting at? Hmm.)  \"Something is Wrong,\" by  Phantom Planet   Burn me down if that\'s how everything\'s gonna start With the flames from the heat off the look on your face On a day like today could have made out every  Raindrop falling down on your head  Now I fear that I may taste these tears rolling down my face  And realize something is wrong  With the light of the sun and the color of the sky  With you and everyone _______________________________________________ Steal the royalties from that band and download that song!  '),(5457,1261679,'16,May,2004','2004-05-16 00:00:00','   I\'ve been having those I\'m-going-to-Hawaii-in-two-weeks feelings that make me exhale happy breaths every second.  It\'s been a good day.  Perfect dose of rain.  We finally finished our English project!  It\'s so cool, with rainbows and a weird story.  There\'s this blue sky with clouds fabric on our tri-panel/ stage that also makes me sigh happily.  =)  \"Mr. Blue Sky,\" by ELO is stuck in my big head.  La la la...  [Radio announcer:]Warning! Todays forcast calls for blue skies.  Sun is shining in the sky  there aint a cloud in sight  it\'s stopped raining Everybody\'s in the play And don\'t you know it\'s a beautiful new day, Hey-hey Running down the avenue See how the sun shines brightly In the city On the streets where once was pity Mr. Blue Sky is living here today, Hey- Hey  Mr.Blue Sky,  please tell us why,  you had to hide away for so long(So Long), Where did we go wrong  Mr.Blue Sky,  please tell us why,  you had to hide away for so long(So Long), Where did we go wrong  Hey, you, with the pretty face  welcome to the human race A celebration Mr. Blue Sky\'s up there waiting And today is the day we\'ve waited for, Oh-Oh  Mr.Blue Sky,  please tell us why,  you had to hide away for so long(So Long), Where did we go wrong  Hey, there, Mr. Blue, we\'re so pleased to be with you, Look around, see what you do, everybody smiles at you  Hey, there, Mr. Blue, we\'re so pleased to be with you, Look around, see what you do, everybody smiles at you  Mr. Blue Sky, Blue Sky, Blue Sky, Blue Sky, Blue Sky, yi  Mr. Blue, you\'ll get it right, but soon comes Mr. Night, Creepin\' over, now his hand is on your shoulder, Nevermind, I\'ll remember you this, I\'ll remember you this way  Mr.Blue Sky,  please tell us why,  you had to hide away for so long(So Long), Where did we go wrong  Hey, there, Mr. Blue(Sky), we\'re so pleased to be with you(sky), Look around, see what you do(wo-oh), everybody smiles at you  Please turn me over __________________________________________________________ (I like blue skies, too.  And that\'s a lot of unnecessary repetition.)  La.  '),(5458,1261679,'14,May,2004','2004-05-14 00:00:00','   I was just e-mailed two viruses.  The same one twice.  It said \"Find the white rabbit,\" but in The Matrix isn\'t it \"Follow the white rabbit...\"?  What a freak wannabee.  '),(5459,1261679,'12,May,2004','2004-05-12 00:00:00','   -That British lady who hosts The Weakest Link has such an annoying voice.  (Scrunchy nose.)  \"Whose jelly donut ran out of filling?!\"  Her shrill voice combined with her harsh accent makes it so hard to understand.  Ugh.   -I was so lazy when I got home: I watched Oprah.  Man.  It was about transgender children.  There\'s this cute little boy who wants to be a girl, but I think he\'d be cuter as a gay kid than a normal girl.  I wonder what it\'s like to be transgender, what with the uncomfortableness in ones body and such.  (I have no desire to be male, by the way.  Silly cootie-carriers.  Yicky.)  I think Oprah has fascinating hair.  It\'s so poofed out and silky and shiny and wiglike.  You can\'t really see her scalp where her part is, so it could be a wig o_O.  One day I\'ll know for sure.    urlLink       what flavor pocky are you?    urlLink  [c] sugardew    -Yeah, Pocky\'s way too mainstream for me. -My contacts are really drying up.  Silly optometrist guy just popped into my head.  I\'m not very fond of him.  He has pretty eyelashes, though. -I\'m sleepy.  I have to write my whole critical review tonight.  I still haven\'t finished reading the book (Fahrenheit 451).  It\'s not long; I just keep forgetting.  I think it\'s because I have the less attractive cover.  You can totally tell what kind of person I am from the previous statement.  Sigh, I disappoint myself.  \"Parachutes,\" by  Coldplay  In a haze, a stormy haze, I\'ll be around, I\'ll be loving you always, Always  Here I am, and I\'ll take my time, Here I am and I\'ll wait in line Always _______________________________________ Yawn.  '),(5460,1261679,'11,May,2004','2004-05-11 00:00:00','   Hey you readers!  Not many of you have been clicking in.  How do I know?  I don\'t.  Well anyway, sorry about the delay for the long-awaited next entry, which is this very entry!  So, I have realized that no matter how much harder I try to avoid some situations, it\'s inevitable because others, at their own free will, persist.  Stop!!  My Apologies about the vagueness of this paragraph.  Onto another subject:  (Disclaimer- This is rhetorical.)  Why do bad things always happen to me?  All the time, in everything I do, I know something will go wrong.  So I just set my standards lower to reduce the disappointment, but really shouldn\'t because it\'s like a self-fulfulling prophecy.  I do worse than I should and I guess I don\'t really realize that I actually have potential.  (This is like a failure\'s speech.  \"You haven\'t seen the best of me yet!  I have potential and everything!\")  Bluh.  Good news: (If you have not been keeping track and reminded so many times or counting down) Only 8 more days of excruciating pain!  (As in school, the torture chamber?)  Well I guess school isn\'t that bad, but I hate having that worry feeling in the back of my head saying \"You had to do this last night, and now you don\'t have time!  You are stupid and ugly!\"  Stupid.. head.. urgh.  Now I sit here listening to my LaunchCast Radio station worrying about the Spanish test over Future and Conditional!  Those words did not need capitalization.    Rest easy...  :)  “To My Magnanimous Friend”  It is pompous to whimper at a loss And become cumbersome and cross. But this resemblance does not appear,  As your bereavement was not a burden here. Pining for the presence of someone so beloved Confirms your altruistic intentions as in tears shed. And whatever creed may be yours He could be up above or return on all fours So feel free to flush your conscience clear For woe could not mar your tenacious exterior. All left to do is to reminisce and cherish Without lament and regret over the perished. ~5-10-04 __________________________________________ I love you!  '),(5461,1261679,'05,May,2004','2004-05-05 00:00:00','   Tuve una buena noche.  No sé por qué, pues estuve bien.  Mi día no estuve bien.  Alguno gente no entiende como no decir cosas y que otra gente duele.  Este situación es estupido.  No sé el tiempo.  No muy tarde, espero.  Siento perdón que este es en español para no razón.  No entiendo español.  Jí, jí, jí.  Pues, tuve una mañana malo y una noche bien.  ¿Cómo?  No sabría nunca.  Buenas Noches, amigos.  '),(5462,1261679,'30,June,2004','2004-06-30 00:00:00','   ..yes another.  (Entry, that is.)  I am unable to fathom the longevity of this month of June, and it isn\'t even over yet as we are still milking it and scraping the bottom, dreading July:  The First Indicator of the End of Summer.  Or maybe it\'s that sleep deprivation habit I have.  Either way, my stitches have already closed up and I\'ve forgiven the coral reef and  still  June lingers and loiters as if it has not already been handed its Ice-Cream.  (From the Ice-Cream Man in the Truck.)  Then again, June  is  my birthmonth and even  I  would wait for my ice-cream to be given to me.  I like Ice-Cream.  In fact, I think Ice-Cream should be capitalized.  Now I will go back and capitalize every \"Ice-Cream,\" even though you will not be able to tell if I performed as promised, but I will have to joy of the possession of integrity.    Buttered Pecan is my favorite.  The line in bold has been stuck in my head for the past day.  No parallels, unfortunately.  Go dig in the Loony Bin for drama, kid.  \"Don\'t Get Down,\"  by Phantom Planet   I want to live on the moon Never see a human again And when the earth explodes I\'ll light a candle for my dead best friends  I don\'t enjoy this place  My best friends are so mean   I look up to the sky when I am sad Because the planets don\'t get down on me I just want to fall asleep And dream of a place that\'s warm...  '),(5463,1261679,'29,June,2004','2004-06-29 00:00:00','   Same old Driver\'s Ed today.  For some reason I\'m always dead last waiting for someone to pick me up, but I live the closest.  Oh well.  I went compact disc shopping very hyperly and splurged on a very wanted compact disc.  I bought my sister New Found Glory\'s  Catalyst  bitterly, because it was $11.98 out of my birthday money and because I don\'t like New Found Glory at all.  I also purchased some healthy, hearty gummy worms.  Love personified.  Well, not quite.  I have to go to the gym and workout and become even smaller, yet buffer.  My goal is not to gain another tongue-twisting, alliterating, rhyming epithet, because it\'s your problem if you want to call me The Buff, Busy Bodybuilder.  Expect more today; you deserve it.  '),(5464,1261679,'28,June,2004','2004-06-28 00:00:00','   I wonder if there is some superhumanbeing that can actually read people\'s minds.  I wish I could read people\'s minds.  \"Why doesn\'t he like me?\"-- Zap, zap, beep, beep --\"Oohh.  Tear.\"  Well I guess that\'s one of the cons of that superhumanbeing power.  I\'m so subconsciously negative!    Driver\'s Ed is really boring.  As adventurous and youthful Mr. Walker may think he is, he\'s all the more soporific in a slow summer day kind of way.  When I walked into the Safe-Way Driving Centre, I thought I would know plenty more people because it\'s right by Klein and all, but I\'d never seen anybody in that classroom in my whoooole life.  And today I will, once again, step into a strange, new classroom as I have to make up Lesson 6 at 6:30!    You know what?  The desk just vibrated and the ground just shook for 5 seconds due to a strike of lightning in close proximity, and with the odds of being electrocuted while using electrical appliances growing more than 1 in a million, I will have to end this entry now.    Hasta La Vista  '),(5465,1261679,'25,June,2004','2004-06-25 00:00:00','   Would it be weird if I sang like Barry White?  Ha ha, I\'m imagining myself having such a low, promiscuous, seductive voice... hmm.  I just rediscovered my Barry White\'s Greatest Hits CD.  And my Spice Girls CDs.  We all went and go through necessary phases, but I still think those Spice Girls inserted some subliminal, hypnotic messages in their TV specials targeting small girls to buy their CDs and strive to be an altered version of Ginger, Posh, Sporty, Baby, or Scary Spice.  I liked Sporty Spice for some reason.  I was a very stupid 10-year-old.    I say that like I\'m smarter now.  Pshh.  I\'m still as superficial and ignorant as ever.  I\'m thinking of painting my room bright yellow, dark red, or a light salmony melon color.  I think I typed this with the intent of a reader\'s feedback, but I don\'t think y\'all care.  Blah, blah, blah.  I want to be in a band.  '),(5466,1261679,'09,June,2004','2004-06-09 00:00:00','   I don\'t like words that end in \"-ness.\"  It sort of invalidates the word.  I like \"uniquity\" better than \"uniqueness.\"  But things don\'t change just for me, ya know.  I want to live in Honolulu in a house that makes your ears pop on the way to it.  I want to go surfing every weekend and befriend all the nice people there.  But alas, I am stuck living in Spring, Texas.  Poopie.  '),(5467,1261679,'08,June,2004','2004-06-08 00:00:00','   Sigh.  Back to reality.  '),(5468,1261679,'30,July,2004','2004-07-30 00:00:00','   Last night I had a dream that I really believed until I woke up. It was so scary, because sometimes I have dreams that call for eating, but I stop myself thinking \'My pillowcase won\'t taste so good,\' knowing that I\'m dreaming. But this dream was a thriller from beginning to end.  I woke up and randomly went to the doctor\'s to retrieved ultrasound pictures. They were of a pretty big and well-developed baby. Then it hit me- I\'m going to have a baby. But I\'m a virgin, so my mom was asking me who could\'ve raped me, and I came to the conclusion that this mischievous kid in my driver\'s ed class named Ray, or something, did. (He\'s the scariest looking person I\'ve ever seen.) I was afraid that I would have to abort the pregnancy, but being 4 months pregnant that wasn\'t an option, so I proceeded to worry about missing my junior year of high school:  I wouldn\'t be able to go to college... my future!!   Then I was startled awake by my alarm and whispered with a sigh of relief, \"I\'m not pregnant.\"  '),(5469,1261679,'28,July,2004','2004-07-28 00:00:00','   Maybe they do.  But what if the whole \"human\" race as we know it is just a sham?  What if we\'re just two billion bacterial organisms thriving off of a giant, as we know it, animal carcass and we\'ve only created a society out of things we can see?  I\'ll give you somethin\' to think about.  '),(5470,1261679,'23,July,2004','2004-07-23 00:00:00','   So right now, I\'m -title of entry- and listening to Phantom Planet.     &nbsp;   \"Come Away With Me,\" Norah Jones. &nbsp;   &nbsp;     My momma told me that, due nature of the installments of extensions in her hair, Christina Aguilera has developed alopecia.&nbsp; Poor her, but fortunately she\'s beautiful no matter what they say, and words can\'t bring her down, so don\'t you bring her down today!    &nbsp;  Now anyone can comment on my blog!&nbsp; Blueberries and carrots!  '),(5471,1261679,'22,July,2004','2004-07-22 00:00:00','   \"Quintuplets\" is such&nbsp; disappointing show.&nbsp; It\'s hard to be funny- WHEN YOU\'RE NOT!&nbsp; So kinda just, um, stop trying.&nbsp; Consider this advice too, you wonderful person you!    &nbsp;    So I want to see this movie, tomorrow maybe.&nbsp; And The Village next week.&nbsp; \"Never enter the woods.&nbsp; That is where they wait.\"&nbsp;  '),(5472,1261679,'20,July,2004','2004-07-20 00:00:00','   I found my password out for my screen name!&nbsp; Yay!&nbsp; It was really exciting, like finally finding your graphing calculator right before a math test, then figuring out that you have to use the school\'s calculator.&nbsp; But I can\'t see profiles on AIM Express, which is basically all I do online, but I am still online, unaffected.    &nbsp;    Like him.    &nbsp;  I\'m stuck in a computer lab.&nbsp; It used to be cold in here, but I guess the temperature has gone up because there are a bunch of old men in here arguing about the advertisements for University of Houston.&nbsp;  Snore.&nbsp; Rumble in my tummy.     &nbsp;   I\'m starving too.&nbsp; I guess looking at recipes for cherry pies on FoodTV.com doesn\'t help.&nbsp; Nope, it doesn\'t.&nbsp; I really need to eat a medium-rare burger with pickles and tomatoes and cheese and lettuce and mustard right now.&nbsp; Hungry, hungry hippo[patamus].&nbsp; Pretty, pretty princess.&nbsp; I liked those games when I was little.&nbsp; I would still today if I had access to them.&nbsp;  (Never wear the black ring... Wear the crown and you win!) &nbsp; As you can see, the Princess game appealed more to me when I was little (and maybe today) but more likely not today, because the word \"princess\" is reserved for spoiled toddlers, not to mention tacky shirts proclaiming extreme epithets.&nbsp; The European monarchs might as well be teenyboppers or first graders. &nbsp;     &nbsp;  I want a kitty.&nbsp; Oh look!&nbsp; I wish I could own that pretty kitty to the right.&nbsp; (Oops, haha.&nbsp; That\'s, like, me.)  '),(5473,1261679,'19,July,2004','2004-07-19 00:00:00','   Maybe you were.&nbsp; I titled the last entry \"Sigh... babies,\" and I forgot to mention something.&nbsp; Whoopsie.&nbsp; Well there is a commercial or SBC Yahoo whose theme is to connect people around the world.&nbsp; And there is an extremely cute baby that I want to babynap in the commercial.&nbsp; So when you guys are all grown up and you have babies, don\'t ask me to babysit because I might steal&nbsp;her/him if&nbsp;she\'s/he\'s&nbsp;extremely cute.&nbsp; That means if I haven\'t stolen it by the time you&nbsp;get home, yo&nbsp;baby\'s ugleh.    &nbsp;   And here I am listening to variations on \"I\'ve Got Rhythm\" on NPR. &nbsp;&nbsp;   So... John Mayer&nbsp;has really big lips.  &nbsp;&nbsp;  '),(5474,1261679,'18,July,2004','2004-07-18 00:00:00','   Dale Earnhardt Jr.\'s car caught fire during the Grand Prix.&nbsp; Pshh.&nbsp; That\'s what you get for eating so much Fried Chicken.&nbsp;     &nbsp;  \"Well I must say, I\'m solitary and quite merry.\"&nbsp; Wouldn\'t it be cool if girls said that instead of \"I\'m single and men are pigs.\"&nbsp; Pigs are the most intelligent animals, smart ones.&nbsp;&nbsp;I say the former if I\'d have known it another way.&nbsp; Hmpf.  &nbsp;  I love Arrested Development.&nbsp; Maybe sometime next week you might see something worth reading.  '),(5475,1261679,'15,July,2004','2004-07-15 00:00:00','   Sometimes you just wanna scream the f-word.&nbsp; But you can\'t.&nbsp;  &nbsp; But I\'m talking about you, because I\'m happy as can be yes I am. &nbsp; I\'ve got an early drive-time tomorrow, and an appointment at my ghetto dentist right after!&nbsp; I actually brush and floss regularly now.&nbsp; Those dental hygienists are gonna get told tomorrow!&nbsp;  &nbsp; Winky winky KRBE.  '),(5476,1261679,'15,July,2004','2004-07-15 00:00:00','   ...She wakes up.  And stuff from that Britney Spears song from long ago in the Stone Ages.  You know, if I hurt my knee, I\'d still go on tour and sing while sitting instead of flat out cancelling the tour.  What a diva.  So how can I be so awake now but so sleepy just hours ago, but with no refreshing rest in between?  Mind boggling.    My little sister went to the Jessica Simpson concert for free last night.  (A few hours ago.)  So many popstars in this entry.  That\'s my fault.  I still haven\'t remembered the password to my screenname.  I guess a month left of not talking to a lot of people isn\'t much to endure.  Actually, it\'s pretty enjoyable.  And it\'s also pretty sad that AIM is an understood wave of communication.  I guess it was all too common without me knowing.  I think I\'ll stay up until 6AM and then sleep.  DOn\'t ask.  Actually you can\'t.  This picture of me in myprofile makes me so uncomfortable.  For one thing, it\'s not clear at all.  And I look unlike myself.  And vanity swallows my body.  *Sigh... What else can I write about?  I just thought about only having two years left til college and I got a good feeling in my heart/tummy or something.  You know?  It\'s like when you\'re reminded of someone you lOoOoOove.  Catch you later.  '),(5477,1261679,'14,July,2004','2004-07-14 00:00:00','   I was listening to NPR (88.7 KUHF) on the way home yesterday.  The show that was on was about a girl\'s teenage life further complicated by a baby.  She sincerely said she served the all-too-typical story of \"Boy meets girl, girl gets pregnant, boy gets busted for drugs and goes to juvy.\"  Ha ha.  Sure isn\'t typical here.  Airheads taste good.  Except for Mystery White.  What\'s up with Mystery White?  I want a donut.  Dough-nut.    My momma\'s in Chi-town.    Buh Bye.  '),(5478,1261679,'13,July,2004','2004-07-13 00:00:00','   I\'m starting to count Calories.  I just think it\'s fun: You get skinny and good at math.  The first signs of anorexia!    I started thinking about The DaVinci Code last night and I lost some sleepy time.  I don\'t know why I\'m so afraid of that story.  I imagine that Mary Magdalene will chop off my head with an already bloody axe if I close my eyes.  I don\'t know why- I mean, you\'re supposed to be happy that Robert Langdon discovered where the Holy Grail was and it spoke to him and here I am trembling with fear.  Weird.  '),(5479,1261679,'08,July,2004','2004-07-08 00:00:00','   I\'ve got a picture up!  It\'s not really that exciting because I have already posted pictures, then took them off because I felt a little insecure about my face being out there for everyone to see, negating the ultimate anonymity of a blog, although other \"bloggers\" would beg to differ.  I think I\'ve already explained this insecurity.  Moving on.  What is this sudden obsession over wanting to go to a concert soon?  I don\'t know, but I feel like I have to go to one or I\'ll die.  But I won\'t; I\'ll just deteoriate into a breathing blob, but that would be cool.  I\'m not saying that not going to a concert would benefit me; I\'m begging you to take me to a concert!  (Thanks for picking up on the hint!  [Even though you couldn\'t respond anyway.])  '),(5480,1261679,'07,July,2004','2004-07-07 00:00:00','   I can\'t remember the password to my AIM screenname.  I changed it because I have recently developed a suspicion that someone might be signing on and talking to people on my buddy list as me.  And I forgot what I changed it to.  Oh well.  I never go online anyway.  I never talk on the phone either.  I just figure that if I run into a friend somewhere, we were meant to hang out or something, you know?  Why initiate something and plan and plan when you can just run into a fun situation?  That\'s what I think.    So this morning my sister runs into my room as I am combing the Spray-In conditioner I bought instead of a David Sedaris book last week and says, \"Manda!  They\'re giving away John Mayer/Maroon 5 tickets on 104!\"  So naturally I get really excited and call in and wait for 10 minutes on the phone as it rings while I squirm into my pants only to hear, \"Sorry, we already have a winner.\"  Oh well.  Like I was going to win anyway.  I haven\'t been to a concert since May 10, 2003.  Lung-deflating sigh.    These were beautiful until they died.  Like everything else.  '),(5481,1261679,'03,July,2004','2004-07-03 00:00:00','   I hate having obligations.  (Like schedules.)    I wish I could do whatever I wanted to.  Gosh.   My List of Movies I Want To See Soon:  Kill Bill: Vol. 2 Super Size Me Saved! Coffee and Cigarettes Spider-Man 2 Napoleon Dynamite The Village King Arthur  That\'s $52.00 at the movies.  But MiniSkirt + StreetCorner= $0.00.  (Heh, heh.)  PS- I don\'t own a mini skirt.  '),(5482,1261679,'02,July,2004','2004-07-02 00:00:00','   I just woke up.  I think I fell asleep because I was watching a very boring tennis match (Ancic vs. Roddick).  The whole time I was just cursing at Roddick for dumping Mandy Moore.  I hope he loses.  Croatians will soon rule the world!  I posted a picture of myself last night, but I took it off because I felt insecure about putting my face out there for everyone to see.  Rationalizing my insecurity ain\'t helping.  Oh well, I\'ll look for a better one soon enough.  In the meantime, enjoy the music.  '),(5483,1261679,'01,July,2004','2004-07-01 00:00:00','   Awake in a might-as-well-be empty house.  It\'s so quiet that I cannot tell the difference between my computer\'s hums and crickets\' leg rubbing.  Exciting!    I went to Barnes & Noble and spent 2 hours there with my sister.  A cute stranger seemed to be retracing my path at a few turns, or that might\'ve been me vainly protagonizing.   I browsed the classics, chick books, poetry, modern plays, paperback bestsellers, David Sedaris, and memoirs until I realized I would not have enough money to buy Spray-In Conditioner if I bought  Barrel Fever .  So, for the sake of saving my head from alopecia, I sacrificed quality laughter inducing reading time.  Narcissistic.  I can\'t believe I actually failed the test on alcohol in Driver\'s Ed.  Whiskey, reaction time, DWIs, $2000 fines: way too much to be packed into my relaxed summer brain.  I need to sleep!  '),(5484,1261679,'04,August,2004','2004-08-04 00:00:00','   I want this blog to look cool, even if only three people read it.  It would be so much more interesting if the template was more colorful... but not like Alice in CreepyLand.   Is that mushroom blood on her apron?  I don\'t wanna know... Mario Winan style.  So I guess my search for an interesting template continues.  Check back later.  '),(5485,1261679,'04,August,2004','2004-08-04 00:00:00','   I think the people next door are watching something X-rated. Their laughing and clapping is getting annoying.       What\'s it like being in a room facing a slightly tinted window in which people stop in front of to check out their smiles while talking on cell phones then realize startlingly that a little unappreciating girl is on the opposite side seeing everything and irritated? Annoying.    I wish someone would invent a complex enough teleportation system to get me to a mall already. It\'s the 2000\'s, people. Come on.    Relax kitty!     Miz Martin is much less intimidating over the phone, so call her by Friday to make schedule changes, kitty and students Rj- Sw!  '),(5486,4280767,'21,August,2004','2004-08-21 00:00:00','   got my license this morning!! I\'m happy, I feel so independent-ish even though I think I\'m gonna make my mom ride with me in van a bit more before I actually try to take it out by myself. But still. And my license picture isn\'t too bad, I look like an elf! lol...  gotta go DRIVE MYSELF (!!!) to Hobby Lobby!  '),(5487,4280767,'20,August,2004','2004-08-20 00:00:00','   I feel much better now...not so guilty. In another sense, I feel icky, really dizzy and my stomach\'s still kinda upset. but oh well. I can\'t believe I\'m updating this again...it\'s the second day in a row!! of course, it is new....and I don\'t have to worry about anyone reading it....\'spretty cool.  I can\'t wait for school to start, thank goodness for wednesday. Maybe I\'m weird....I just want to be able to have some social contact (although, this is Recluse Lydia I\'m talking about...) and stop hanging around being bored all day every day. Plus I\'ll get to drive, woohoo.  Speaking of driving, I should get my license tomorrow. I mean, I AM going to get my license tomorrow. I guess I should look over my rule-booky-thingy but I\'m really not too nervous, I know the test and I know I\'m a good driver, and I don\'t have to drive the manual for the test. So even though I haven\'t driven for a week or so, I still think I\'ll pass. I\'d better pass, because I refuse to take the MTD and have to walk all the way to that bus stop, it\'s insane. sigh.  I hope my mum gets my clothing allowance taken care of tomorrow so I can get some jeans...I don\'t really have a pair I\'m willing to wear to school. and I did my job, I figured it all out and typed it up and printed it and submitted it for her approval...so technically she should really take care of it. I wouldn\'t mind having some of the rest of thatstuff before school starts, either.  I wonder if she\'s just gonna let me use my credit card for my allowance this time. Even though it\'s not a cash card, but she keeps saying she trusts me, so...it\'s cool though, I feel so powerful with that thing. I think maybe I need to ask her to make \"emergency\" more specific, lol...  once again, I have no idea how this post ended up so long. wow.  '),(5488,4280767,'19,August,2004','2004-08-19 00:00:00','   so it took me forever to find a web address that was \"available\". grr, who is the world (besides me, of course) would want fasterthanaspeedingwart or gagmewithaspork or lollyronismyhero or mucwerqubinwips or any of those other crazy things I came up with!! It was so frustrating!! everyone else knows my inside jokes with myself! *sob* I feel so...so violated!!!  on a different subject, I\'m feeling really horrible. Which I guess is a good thing since I\'m going to to end up lying and saying I threw up and passed out and went to the emergency room, etc etc as an excuse as to why I didn\'t show up at the movie I was supposed to go to with Merry and Aaron today. The truth is...Merry couldn\'t go and I refused to go without her, aka by myself. To those people who told me to just go alone, shut up. I need a buffer, ok, I couldn\'t go by myself, I was already feeling sick just about the idea of going WITH MERRY. And now I feel really guilty but I can\'t feel TOO guilty because I understand why I did it. I know people are gonna be really fed up with me....oh well. I\'m afraid to answer the phone now, I\'m avoiding windows, jumping at any sound....I hate this.  I doubt I\'ll be updating this real often, because I have a livejournal that I really like, and then I also have a xanga, which I don\'t like so much...and now this, too. oh well, I like blogger, maybe I\'ll get around to it.  hmm, now that I\'ve seen so many commercials with Michael Phelps in them during the Olympics, I\'m starting to agree with Grace. He IS hot. If only I didn\'t always have to see him in speedos. *shudder* Women\'s trampoline is on tomorrow, I\'m so excited!! Trampoline rocks.  wow, it\'s amazing how much I\'m rambling. Possibly it\'s because I\'m home alone and hiding, which includes sitting her in the basement with the curtains closed, not on AIM, not on MSN, avoiding windows, not answering the phone. hey!! I\'ve just realized, it\'s the return of Recluse Lydia!! Recluse Lydia is a paranoid delusional freak, who has no social life and at any sign of getting one reverts back to standing people up and hiding in her basement. Recluse Lydia is a jerk and Recluse Lydia is feeling guilty and very ashamed of Recluse Lydia at the moment. Recluse Lydia wishes she could call Lisa but seeing as it\'s a thursday afternoon and Lisa DOES have a social life/boyfriend and Recluse Lydia does not especially want to pay the phone bill resulting from the talk she needs to have with Lisa, it is not exactly likely that Recluse Lydia will get to talk to Lisa any sooner than saturday, whcih is two days away, which mean Recluse Lydia has about 2 days of feeling horrible and not being able to vent on anyone, and will be forced to rant on her online journals, subjecting any of you poor people who are unfortunately reading this to huge paragraphs told by Recluse Lydia in the third person, further proof of her insanity and need to be put into one of those way cool padded rooms. *gasp*  wow, I\'m rambling way too much....think I\'m going to....leave now....not that I know what else I could do...but yeah. If you read all of this, maybe you should go see a psychiatrist...haha.  '),(5489,4115407,'18,August,2004','2004-08-18 00:00:00','   Today freakin sucked, i couldnt get my ears pierced because my dad and step mom was in a accident, nobody was seriously hurt but i was put off for one day. U might think that im being selfish about this but u dont get it, every time i want sumthing, or want to go sum where sumthing always comes up, and im so sick of that. I never get to do or get anythin its like it always happens to me. I mean wut did i do to deserve this, and this is not even about my ears anymore, it never really was im so fed up with this. It has been goin on as long as i can remember which is a while. I dont wanna say im ready to grow up but i am so nobody can stop me from doin sumthing, ill be able to drive so nobody cant say they tired or maybe next week or sum like that. I can do whatever i want when i want oh wow im starting to stress myself out again, im just ready to go home so i can go to school and get out of school and move out of new jersey, im not concered about leaving friends i mean i will keep in touch and visit them, i think the only thing i would have a problem with is that if i still have the boyfriend i have now or get another one ( im not sayin im am, i mean anythin can happen, but i dont want it to anyway) but he is either coming with me or im srry to say its over unless he is like in nevada or sum state close. But ima be real i cant have a long distance relationship, i mean wuts the point of havin a bf when u can only talk to him on the phone, u would never be able to kiss him or hug him or do anything else i mean thats gay thats y i think thta goin out with ppl over the internet is dumb to, srry no offense to anyone, but there is no point. I mean sum ppl probably can but i cant cause its dumb. And u kno wut else is dumb cybering and phone sex, i mean that is the most dumbiest thing. Its stupid there is no point y would u want to want to talk about it or even type about it. I mean if u got the nerve to do that u might as well do it. I mean y would u wanna talk about to make u horney, i mean doesnt that make it worse, i mean im sure there are other ppl just like u, i wouldnt kno im jus usin common sense. And u kno wut elseeverybody blames all the rappers and stuff for having girls in they videos and men and strip clubs and stuff, i mean dont they realize that women are not forced against their will to dance or be on tv half naked its their choice, they make the desicion to degrade ( if u dont kno wut that means look it up) themselves, so all the falut shouldnt be on men, even though they are not in the right either for encouraging it. I mean videos and men have put alot of girls self esteem down. I mean all they talk about is having a \"fat ass\" and all these curves and everything thing else i mean girls are using butt implants as if breast implants wasnt enough, girls feel they have to look a certain way to impress sum body or fit in, like they have to wear degrading clothes for boys to notice you, i mean im not talking about me or nobody else i kno, i dont have to dress half naked for nobody and i dont or look a certain way 4 nobody and im fine ppl dont think im ugly and i really dont care if they think i do or not, nuthin is wrong with me and every girl should feel like that, the only person they should be pleasing is themselves and if they have a boyfriend or husband, but only if they accpet u 4 who u are and are not makin u go through sum surgery to look a certain way or change the way u dress to show more or sum stupid stuff like that, men are such dogs sumtimes, but not all are like that and ull be lucky if u find sumbody thats not. And the bad thing is u cant live with out them unless ur a homosexual ( Faget and lesbian are not nice to say)   '),(5490,4115407,'17,August,2004','2004-08-17 00:00:00','   \"Diary\" Featuring Tony! Toni! TonÃ©!  Lay your head on my pillow'),(5491,4115407,'15,August,2004','2004-08-15 00:00:00','   Today i was bored i thought i was getting my ears pierced but i am tom. so today i was just on the computer and thinkin as u can c i think this is my 3rd blog today, lol. Sometimes i think too much and then the stuff i was really happy with all of a sudden i wasnt cause i thought of all of the bads things that can happen, i always done that, i sumtimes i wish i didnt cause i might can lose sumthing really good. Like with relationships, i thought about got happy and was ok ima do it, then i kept thinkin and was no, jus like with health science tech. i thoguht of all the work and stuff, but i mean i lost an oppertunity. I have to stop sumtimes, cause im doin it this very second, i dont know why, but maybe every time i thought sumthing good was gonna happen, it didnt, sum 1 didnt keep their promise or sum crap like that. I keep thinkin about all the possible bad stuff that can happen 2 me at franklin like i could get stabbed, or even jumped by sum gang but im not scared im just worried i mean if i get shot then wut. Then i keep thinkin about how im not goin 2 the same high school as none of my friends and that they will forget about me cause that happens 2 me alot, they stop calling or inviting me places, but then again i do to. But if i really wanna keep them then i wont. Im worried that since me and my bf arent goin i mean i trust em, but i mean wut if he finds naother girl there and drops me like he did to this other one, if i was her i would have tooken that out on him, but then again i cant take it out on no one, because im sure he didnt want to like the person on purpose, then again i dont kno that either, i wouldnt kno wut would go on in school, i would jus trust the person hope that they woudnt do me like that. But im not a jealous person i dont care as long as hes not cheating i dont care. I do but, no to that point where u can talk to no girl at all. I jus need sum 1 to talk 2 i wanna go home so i can talk 2 my momma ( yea i call her momma, so wut sue me). Im making my self sad lol, but then again i jus laugh it of which i hate doin sumtimes cause ppl dont always take u seriously. Im happy im leavin friday i cant be away from my house for to long, i dont like stayin down here with them except for my aunt or godmom, but i mean i cant do stuff my way, i mean they claim its my \" second house\" but i mean i sumtimes have to ask permission to eat, i mean give me a break, but i think thats because its two babies in the house but that doesnt mean that. And i cant really ell them i dont eat stuff like pork, i dont eat pork cause its not healthy, ( except 4 bacon once in a while, but not often) and i member i told them that and they looked at me like i was crazy, i dont eat hot dogs either, but my mom respects that, if she cooks and i dont like it, she\'ll let me eat sumthin else, but here its different, i dont care im not movin back down here, cause im stayin in nj for high school then im goin 2 califorina, so my dad can keep dreaming. I miss my mommy, i mean she is so easy to talk to, and she is nothin like my dad i, he is so like over protectingmtimes. But if i have any kinda problem i can talk to her, i mean even about my chores and stuff, she says im sneaky sum times cause i dont tell her everything at once, but im not i can be sneaky but in a bad way, like sneakin food in my room way. well i was kinda when i started having a bf, she didnt ask so i didnt tell her, but i guess she got the clue w/e its not big thing. I mean she like gives u all the freedom like a 14 yr. old can have only because she can trust me and she kno i wont do nuthin stupid. And its sad cause i never go no where, i mean there is no 1 2 play with outside, it always my sister goin 2 new brunswick, i couldnt cause i had 2 babysit. And im glad that if i do sumthin or left sumthin that she will be there, like the time when i had left my cap and gown at the lib. me and xianne thought we were goin back but then i went sumwhere and saw my mom to ask her if i could stay, she was like yea, then i was like but my cap and gown is at the lib. so she was like well go back and get it when u pick ur sister up so i said ok. So i got there and i wasnt there i was like omg where is it so i was up and down new brunswick lookin 4 it. I finally c my mom and she was all quiet, and i didnt wanna tell her yet until i called xianne to see if she had mine, well i get home and was like oh ya cap and gown in the house, i was like wut! omg u had me in this car worried when u had it. I was so relieved she always finds a way to get me, and she always fins out wut im doin, i mean she could be an 1 hr. away from me, i mean in a different state, btu she still kno, not the exact details, but she kno sumthin, thats kinda scary. Im not thinkin no more but im still worried, and i need to stop worrying my self about stuff like that cause its nuthin compared to other stuff like me being pregnant or getting a std or disease or sumthin serious like that.   But im gonna listen to my music cause that helps me out.   '),(5492,4115407,'15,August,2004','2004-08-15 00:00:00','   \"My Baby\"(feat. Kanye West)'),(5493,4115407,'15,August,2004','2004-08-15 00:00:00','   Ok, just to clear up things on my last blog, I am obssed or w/e with vampires, and I dont have a problem with the govenor being gay, i mean i dont have a problem with them at all male or female, i have a problem with them not respecting my sexuality and tryning to turn me gay, thats not cool at all, i wont try and make u from not being gay, so dont do it to me. Also alot of people call me mean, i am not mean i just have a opinion like every one else i am not cruel in any way, and people to think that i am, i am also well, im still tryna figure out werid, well i really dont care i mean apparently im not that weird if i have friends and a boyfriend that claims there not weird, lol anyway im not shy either, i hate when ppl say im shy im not shy im just quiet until someone\'s sparks the conversation, as u can see i dont have the greateset social skills but i can really care less. Another reason is that ppl always think im getting smart, but being smart and sarcastic are two different things, i dont get smart and im not mean i just state my opinion and yea so wut i always got sumthing to say i wouldnt be anisa if i didnt. And it pisses me off or when ppl say me and my sister look alike cause we dont at all and we are nothing alike, ppl think im always mean to her, trust me im not, usually the big sister always has the mood swing problems and gets attitudes with the little sister, well were opposite she always like that 2 me with the attitude and mood swings, but i let her it doesnt matter to me as long as she doesnt touch, cause as soon as she does its over and she cant win an argument well not with me, and if she wont do sumthin that i told her to do i tell my mom yeap ima snitch well im not like that, but with her i have to, cause her little friends got her like that thinkin she grown and that she dont gotta listen, so i tell my mom to get my way, cause my mom already kno how she is, and they is no other way for her to stop unless u hit her hard enough to kno that ur not playin, if she pushes me to that point i have to do that, hey she asked for it. one teacher was Ms. Powell she assumed that i was sum quiet little girl to myself, but she assumed and didnt kno me and misunderstood me and thats her fault 4 assuming. And my dad id kinda like that, he\'s not around me long enough to find out how i am so he thinks all this stuff, but still i really dont care. And im glad that im \"weird\" w/e that means and i have karina and shanae to thank, because if i would have hung out with ashley or buddah or any of them any longer they probaly would have drove me crazy with all that crap they talk and then all the back stabbing then look u right in ur face and say im ur friend omg that will drive any 1 crazy. And then my coach, he didnt kno me either, he thought that i was sneaky because i didnt really talk in front of him then with my friends i would and he called me nonchalant, w/e that means i dont feel like looking it up. So as u can c im really misunderstood, oh and after 4 yrs. of negociating ( my spelling is off today srry, cant think) i can finally get my freakin ear pierced, now im not goin another 4yrs. for the top of my ear, nope, im slicing the time in half or sooner, i hope.   Oh any my legs hurt so bad, this working out thing is not cool at all.   '),(5494,4115407,'15,August,2004','2004-08-15 00:00:00','   Omg i am like a horror movie fien this summer usually half the stuff i watched i could never watch but its like a breeze now nuthin scares me or gets me thinking about stuff anyway, i have fell in love with vampires, i mean like i want to be a vampire so bad sumtimes, like if sum 1 would give me a choice i would be one. I mean like usually boys in movies woudnt be cute or anything at all to me, but if theyre a vampire then like all of a sudden they are, like being a vampire is like soo wuts the word um, sexy thats it, i sound like a total idoit right now but i really dont care and no im not sum vampire freak or anythin i jus think that being a vampire is cool, wut is wrong with me, im like turning white no offense, i dont care if i am just because i like to listen to sum rock and pop but that doesnt mean anythin so stop stereotyping me! ok, anyway.   Have u ever ever like seen ur parent kiss, omg that is soooooo nasty even when i was little i would run to the other run, but i mean my dad and step mom is jus like ewww, i mean like ur old now u still cant do the same things i mean especially in front of me, i mean him grabbing her a** was jus so un nessesary, i mean i kno ur married or w/e but i mena old ppl shouldnt be allwoed to do it after a certain age limit, its ok for every 1 under 30, in my opinion i mean, they can kiss or w/e but the stuff they do its like not needed for other ppl to see. And our governer is gay, lol, thats not funny but i mena wut is the world comin 2 i mean seriously.   '),(5495,4115407,'11,August,2004','2004-08-11 00:00:00','   Oh i c how yall are yall can comment on all of karina\'s but u cant do mines, thats ok, ima member that cough* nabor *cough, lol jp   '),(5496,4115407,'10,August,2004','2004-08-10 00:00:00','   For a couple of years i wanted to be a pediatrician, but now, i dont really know i still want to be, maybe because i have to be in school for about another 12 yrs. I mean that is too much, i want well i think i wanna have kids, im not sure about that, but if i do i want to have them in my early 20\'s, and i cant do that if im always in school, but i am not having kids after 30, 30 is the limit. But i think i wanna be open to other things, i wanna find sumthing that i wanna be happy doin. I wanna love to go to work every moring i dont know im just confused.   '),(5497,4115407,'09,August,2004','2004-08-09 00:00:00','   Today i saw my high school schedule, and i was disapointed, they only classes they gave me that i wanted was geometry, but they didnt give me any of my advanced classes that i should have got, im really pissed cause im not about to do the same crap over, and i wanna be on the same level as all my friends i dont wanna be behind, i wanna be ahead of them, but im gonna tell my mom to go up there and fix it before i start school, so it wont be any problems im going with her this time and bringing my new brunswick schedule show where i am suppose to be. Beacuse if i dont get any of my advanced classes or credits, then what was the whole point of me being in TAG, i mean Ms. N said that i doesnt matter what high school i go to i should still get my credits and classes.   '),(5498,4115407,'09,August,2004','2004-08-09 00:00:00','   i\'ve been hearing about alot of relationship provlems lately, and im not talkin about only u karina, i mean i watch how my dad and my stepmom relationship is, i hear about my other friends problems. I mean i hope i never have to deal with a physical abuse relation cause if i do im kickin his butt first then im leaving him lol, but im glad that i have a bf that wont well he said he wouldnt and i believe him, that wont hit me, i mean when i hear that girls have been killed by their boyfriends that pisses my off, cause they shoudnt hace stayed in the abusive relationship to the point where he killed u, im mean u may love him, but he doesnt love u if he is willin 2 kill u, y would u want a man that puts his hands on. There is also the relationships where only the boys want sum, but dont want u for ur personality, i mean a girl can be attractive but dont waste a girl\'s time 4 that, im glad he isnt like that either, i dont think he is, i hope he isnt. Im jus glad i found sum 1 that would treat me right, i mean this summer i think i saw him like twice, but that was because i was doin stuff, and he wasnt here, then i wasnt there, but i think i kno him enough to kno that he wouldnt do that. Even though he can get moody and ignore me but, im sure he doesnt do it on purpose and i dont mind if he does, i mean i do, but i dont it a big deal, i rather him ignore me than yell at me, or abuse me mentally or physically. I hope that it works out, but even if it does during high school will it during college, cause im goin 2 college and im goin 2 california, and i dont think his parents would want him follwong sum girl, but maybe we will move together, i hope, that would be cool, if we did. But im greatful, and i hope that all my friends and family could work out or meet the person who would treat them right.   '),(5499,4115407,'08,August,2004','2004-08-08 00:00:00','   I just saw my new high school and it is huge, i saw it in construction, but i didnt think it was gonna be that big, now that i have saw it, it makes kinda glad that im going to a nice high school like that, and now that i think about it im glad that i am getting away from new brunswick, im sick of seing those same faces and i dont think i wanna c em for the next 4 yrs. i mean after a while all the crap that they do gets on ur nerves, im glad to be going where new ppl is, and the only bad thing is that i wont be with my bf, but maybe thats kinda good i mean not like that but i will work out i hope, i trust him 2 be a good boy with all those other girls around lol, and he can trust me, im not going to do anything he wouldnt like ima good girl, lol, but now that i have seen it im excited and glad that im goin away and its sad because i was just complaining about how gay it was, but im not goin to new brunswick and i need to get over it, i need to make this experince a good 1,  so i wont miss out on nuthin, i wish sum of my friends could come but, idk   '),(5500,4115407,'08,August,2004','2004-08-08 00:00:00','   This blogger thing really works, i got sumthin i can write my ideas or w/e im thinkin on here, anyway lol. There are good things and bad things about being a girl, and it sucks cause there is nuthin really bad about being a boy when u think about. But nooooooo there had to be alot of things bad, i mean girls have to get this dumb period well im gonna be mature and use the proper word \" menstral cycle \" thing which is soooooooo stupid cause it just gets a whole bunch a young girls pregnant cause they think their grown, then they have to push a freakin body out of this tiny hole, i mean see this is all adam and eve\'s fault, if she wouldnt have been dumb to eat that freakin fruit, we girls, women w/e would never have any pain but nooooooooooo she jus had to do it, and dumb adam went along with it, anyway lol, being a girl sux so much, we get like no respect, sum boys just think of us is to cook for them and to clean and to freakin have their kids, which sum dont even take care of, sum dont think that were weak and that we cant certain stuff as them, but if u look at it, we have the babies and thats alot of pain doing that, i bet they couldnt do that now could they. We had to fight for all of our rights, and today we dont even get respect, then we get pregnant at liek 16, were considered a hoe, but nooo they dont pay attention to the boy who went    right along with her to get her pregnant now do they, nope. Even though, we dont get treated right sumtimes, and we have to go through alot of sh**, we still have sum ups, we can get out of alot stuff, we can talk our way into getting anything we want from men, and sum people go easy on us, (even though that still sux), but if u look at it, it is kinda nice knowing that u are bringing another life into the world, i guess? U kno wut it is kool being a girl, ( well kinda, depends on how the way u look at it) we are much smarter than boys i dont care wut nobody says and stronger, cause if they were so smart then they would take care of their freakin kids, and think twice b4 they get sum 1 pregnat, ugh now that i think about it, sum boys are such dogs,  i mean all they want it sex, not all sum, so i happy im not like that. Im sure im not the only girl that has felt like being a gurl sux .   '),(5501,4115407,'06,August,2004','2004-08-06 00:00:00','   when i was  little  i was insecure with my self, but now im not, and im glad to have a name that freakin people couldnt get right for the past six years, and then when i get to a class who can say my name right, they didnt want to call me that but im not gonna get into that. As for me......... im fine with myself. I like to play sports and get dirty but im not a tomboy, i dont like dressing liek a boy, or acting like one, i jus like proving a point that sumtimes girls and better and stronget than boys thats all lol, i still like to go shopping, but im also not a girly girl, i dot go to a pool and sit in it and scream if my hair gets wet, or dress up and wear skin tight jeans when im goin to the movies or go to a game, i mean like its really not that serious u can look nice but sum of the people i used to hang out with took it over the extreme, sumtimes outside of school they would wear sumtimes jeans and shirts sooo tight it didnt even look like they can breathe, anyway, i cant do that i have to able to breathe, and if im not comfortable in it im not wearing it, thats y i never wear dresses( except for sum summer dresses) or wear shoes with heels, im fine with sneakers.... My friends, well associates, the people who automatically put me int their little clique which i really wasnt in, was bad influences but i never got influenced i dont think, no or i would be sum where flirting and always lookin for sum boy, thats all they ever talked about that and how theyre onn their freakin perios d like i really wanna hear that, then when i saw sumthin or make a face they be like wut u didnt get it yet, its not that but i mean who wants to sit there and talk about, i kno i dont, and i mean there like kinda the best hypocrites u will ever meet, im not talking about them im jus tellin the truth, they will sit there and talk about this girl when she is not there and then the next day they smile in her face like it was nuthin, then when she do sumthin, they all say we aint to her, but at sumbody\'s swimming party who were they all over. And they are the same way with sex i mean they talk all nasty liek they experts, but when they find out that sum 1 really did it eww u nasty i would never go that far, or they a hoe, i mena give a break, anyway im jus glad that im away from that, and i didnt wanna hang out with the other girls cause they was all stuck up and to tell the truth, u kno wut no ima be quit, about them cause they just make them self look really stupid not likin me and callin god knows wut over sum dumb boy that dont freakin like u, ( no offense to the dumb boy lmao, jus saying, ur not dumb, dont take it offensivly) i mean sum girls are jus so ignorant, i would never do that to another person jus because he like another girl and not me, u didnt do it on purpose i dont even think u can make a boy like u, well i dont know, but they shoudnt hold grugdes on me they should hold them on him, or be like a mature 14 yr. old and get over it, hey that sounds like a good idea! Sum girls can be soooo immature, and im goin to stop cause im not being right either talking about em but so freakin wut, ok im done. lol   '),(5502,4115407,'04,August,2004','2004-08-04 00:00:00','   This has bascially the worst summer of my life, well as far as i can remember. I can say that my summer is over because as soon as i come back, I start school the next 3 days, so barely have time for school ssummer again, that is a promise, i could never go hang out with my friends because of that badhopping, and to even see my *cough* boyfriend, oh lol jp, ( that makes me laugh everytime i do that) and my friends that i wont get to see as often. I am never babysitting during the  little boy who could have took care of his own self. I will never have a suumer like this again, well i hope i wont cause this one reall sucked, next summer im doing wut a 15 yr. should be doin during the summer, im goin 2 parties, im hangin out everynight, im goin 2 the movies, im going swiming and whatever else i can think about doin that wont( that  EXCLUDES SEX AND DRUGS ) i ahd to put that in bold so i can just let everyone know that so they wont get any ideas. I am so looking forward to the next summer i am going to make that summer the best one i had. I might as well forget about this one and start looking forward to school ( oohhhh god) everybody pray 4 me lol jp.   '),(5503,4115407,'04,August,2004','2004-08-04 00:00:00','   This world has changed so much our socitey is so different and more riskful. Most adults say oh please i was 15 once, yea um that was about 40 yrs. ago, i mean society has changed so much, kids and teens have changed. Now we just dont have to worry about maybe just getting hit by a car, we have to worry about rape, drugs, AIDS, STDs and alot more. Adults dont have the slightest idea wut it feels like to be a teen in this society now, they never walked down the street and had grown men yellin at them, i mean grown men give me a break, sum men are just dogs. They have no idea wut it feels like to go to school have have boys just come up and touch and talk to you however they like, ( just from wut i see, cause if they touch me they kno they getting their a** kicked) they never know how it feels to be presseured to have sex, or to take drugs, they dont even know the conversations half of us teen even be talking about. I just hate when adults try to compare their lives and say they know how we feel, because they dont have the slightest idea of wut goes in school, and after school, if they would just take one day in my shoes or any other teen now, they would give us some slack, well my mom probably wouldnt even let me out of the house again lol, but this world is sooo dangerous, i mean now i see why u should wait till have sex until ur married, no i havent done anything, but my friends have made mistakes have gotten theirself into problems, first of all, u cant even find a boy to trust, boys now jus wanna (excuse me language but this is a real quote) \"hit it\" and go tell their friends. And now girls have to worry about getting pregnant, or catching aids or sum nasty STD..... anyway, its jus my opinion that adults dont have the slighteset idea of wut their talkin about, and i wish they could stop assuming and judging us without even seeing wut we have to go through everyday.   '),(5504,4115407,'03,August,2004','2004-08-03 00:00:00','   Ughhh im goin to the gayiest skool in the world, lol, i am attending Franklin High School for my freshman year, ill dont know how long i will be there but i hope not for long. I want to go to New Brunswick soooo bad but my mom said i had my chance and that was to apply for the Health Science and Tech. school, but my being Anisa i didnt take that oppertunity, but maybe that was a godd choice not attending there anyway.. this means i will habe to start all over, i dont im gonna try and make friends, when ppl talk to me ill respond but im not gonna kill myself to be popular or to be anyone\'s friend. Well at least i\'ll know sum people there hopefully those freakin chipeset haters wont mess with me cause im beautiful, lol jp, and i have nuthin against gay people at all i dont mind i still consider them as people and to be my friends, but i swear and i usually dont swear but if one of those gay girls turn on to me and even freakin touch me im snaapin off, i dont care, im noy like that and i want them to that im not being gay 4 nobody, let em touch me ill be sumwhere, ok nisa its ok, ok breathe, ok anyway, im gonna have to adapt to franklin stuck up ways, and i will never forget new brunswick never im attending all new brunswick games and if they play each other ill be for new brunswick, ill never be a traitor to new brunswick jus to let every one kno..... anyway my friends better keep into with me, and i will do the same with them, anyway jus hope that i\'ll do good here so i can get my scholarship to UCLA yea california is so kool, im movin as soon as a i graduate, i think lol.   '),(5505,4115407,'02,August,2004','2004-08-02 00:00:00','   When I entered TAG sometime happened, I heard, someone had a secret crush on me ( well not for long) but I was just as clueless, then I heard rumors, but I thought no he has a girlfriend, later I found out that he broke up with her over sill little me (lol). Now the next year the secret had spread like a nuclear bomb over some stupid little juice. And I had made some enemies and which I still don\'t really know why. Anyway I found out and to be real I didn\'t know what to do, so when he asked me out as a clueless person would do I said no, i didn\'t even know we never talked at all, so its like the boy down the street who u have never met before coming up to you and saying I like you a lot would you go out with me? So I started talking to him, I started to get to know him a little better, he asked again and I said no, then he told it was because I was afraid of a commitment and a relationship and now that I think about it I was because I have never had a real boyfriend before. So we continued to talk he still continued to like me and I had no control over that. Then there was a last time that he he asked me, I was like no im srry and I know I really pissed him off that time, (im not gonna go into details cause this person doesn\'t even know im writing about him). So we contiued to talk and I continued to convince myself that I didn\'t like him, but after a while when ur hangin out with sum one u cant help it. So the eighth grade dance came up and I convinced myself that I wasn\'t going with him, but then I thought its the least I can do so we went. Then we went to the movies the movie was whack so I fell asleep tthen i woke up and he was on my shoulder i guess it was ok cause he was sleep so i left him alone. Then the next we went to the movies again, and like right in the middle of the movie he just like asked me out. So i guess something persuaded him that i kinda felt sumthing for him cause he dont it again, and if i felt like i said no that i would be acting really stupid, so i said yea, after months and months pf saying i didnt feel for him, i found out that i did, i said hey i never gave him a try so let me. So i was a big girl and said yes and stopped being afraid of a commitment, and nuthin\'s wrong with me im still alive and now i have a boyfriend. (who better still like or im gonna kick his a** , lol, jp) Even though we are not goin to the same school we will be fine i have to get my friend to watch in case he gets any ideas cause i know have girl\'s boyfriends get when they are not around.   '),(5506,4115407,'02,August,2004','2004-08-02 00:00:00','   I was born Anisa Nicole Williams, i was named after my aunt who was mentally retarted.... ...great huh? I am now 14 years and getting ready to attend highschool. When i was younger i really didnt like who i was, i wanted to be somethiing else or to look like something else. I always wondered why couldnt i have soft and long hair like her, or why couldnt i have hazel eyes, or why it was always me to be the tallest or the biggest one in the class. I was never greatful for what i had, i was never greatful for always making good grades. I\' ve done alot of dumb stuff in my short 14 years, i\'ve played around with serious stuff that shouldn\'t be played with, i never took oppertunities that could have helped me later on. I was never able to accpet myself for who i was i always wanted to be that skinny girl on tv, or that girl who was mixed with something else besides being all african American. I always hated my name and think why couldnt i have a simple name like ashley or jessica. But now as i am maturing mentally ive come to realize that my eye color my height, my weight, my name were all fit for me, god made me not to always be the skinniest or to have the least common name, or to have the nappiest hair, and now i can accept that am i am glad to be that i am glad to be \"weird\" or to be quiet, or even to be misunderstood. So like it or not because this is who i am.   '),(5507,3186371,'13,July,2004','2004-07-13 00:00:00','    '),(5508,3186371,'13,July,2004','2004-07-13 00:00:00','   By JEREMY RABKIN WSJ July 13, 2004; Page A14  Maj. Michael Newton, a military lawyer who teaches at West Point, coined a new term earlier this year -- \"lawfare.\" It is the pursuit of strategic aims, the traditional domain of warfare, through aggressive legal maneuvers. Last Friday\'s decision by the International Court of Justice (ICJ), holding Israel\'s security fence in violation of international law, is another milestone in the onward march of \"lawfare.\" The ICJ has now confirmed that \"lawfare\" and warfare can be pursued simultaneously.  The terror war against Israel, launched in the summer of 2000, has by now resulted in the deaths of nearly a thousand Israeli civilians. The security fence, by greatly impeding the movement of would-be terrorists into Israel, has helped to achieve a sharp decline in terror attacks over the past year. Nonetheless, the ICJ admonished that the nations of the world are obligated, not to pressure Palestinians to abandon terrorism, but to pressure Israel to dismantle its security fence.  Most of the Court\'s reasoning, based on arguments advanced by British barristers, is superficially plausible -- so long as one ignores the actual political context of the dispute. Perhaps the Children\'s Rights Convention or the Fourth Geneva Convention do provide arguments against disrupting the free movement of innocent Palestinians. But the arguments are more plausible if one ignores the terror threat to Israeli lives, as the Court essentially does. In concluding that the fence sits on \"occupied territory,\" the Court assumes that the armistice lines of 1949 are Israel\'s final borders, though never accepted as such by Israel\'s neighbors. In concluding that Israel cannot undertake intrusive measures to protect \"illegal settlements,\" the Court assumes that Jews had no claim to return to places, like the Old City of Jerusalem, from which they were forcibly expelled by Arab armies in 1949.  Who now has what rights in what territory is at the heart of the dispute. Stripped of its technicalities, the case here was Palestine v. Israel. And Israel lost by 14-1.  The statute of the ICJ provides that the Court may only decide disputes submitted by states and then only with the consent of the states that are parties to the dispute. This provision follows the practice of the international court established under the League of Nations, of the international arbitration court established by the Hague Peace Conference in 1899, and of all previous international arbitrations. Serious diplomats have always doubted that international questions, involving the highest national interests, could be settled by mere legal reasoning.  The traditional approach to international arbitration would have barred the Court from entering into this most intractable international conflict. Palestine, since it is not a recognized state, cannot sue before the ICJ in its own name. And since Israel assuredly did not consent to let the legality of the fence be decided by the ICJ.  The U.N. General Assembly, however, at the prompting of Arab states asked the Court last December to provide an \"advisory opinion\" on the dispute over the fence. The U.N. Charter allows for this procedure, but only in regard to \"legal issues\" and only in conformity with the overall scheme of the Charter. Here, the General Assembly was effectively asking the ICJ to endorse its own political conclusions, with its resolution describing the fence as a \"wall\" (most of it is, in fact, chain-link construction) on \"occupied territory including East Jerusalem.\" The General Assembly was also seeking to have the ICJ do an end-run around the Security Council, which is supposed to have primary responsibility for resolving threats to peace (while at the same time circumventing the legal requirement that Israel consent to be judged in a case to which it was a party).  The Council had exercised this responsibility only a few weeks earlier, by endorsing a \"road map to peace\" which stipulated that ultimate borders -- and such questions as the ultimate status of Jerusalem -- should be settled between the Palestinians and the Israelis in direct negotiations. The ICJ\'s ruling is likely to prove a genuine obstacle to the step-by-step bargaining process envisioned in the road map.  How can a Palestinian representative, for example, now embrace a peace plan that cedes less than full sovereign control of the Old City of Jerusalem, when the highest international court has declared that Palestinians should have the right to oust all Israeli claims to that place? How can Israel put any trust in international guarantees when even the ICJ seems so indifferent to its life-and-death concerns?  The Court had many legal grounds for refusing to decide this case. And such arguments were pressed on the Court not only by the Israeli government but by the U.S., by Russia, by the European Union and a majority of EU member states. All the sponsors of the \"road map,\" in other words, urged the Court to stay out of this heated political conflict. Many other governments around the world took the same view.  Given the membership of the Court, where states hostile to Israel are plentifully represented, condemnation of the fence was to be expected. Most dismaying, however, was that all five judges from EU states (UK, France, Germany, Netherlands, Slovakia) went along with the majority. Only the dissent of the American judge, Thomas Buergenthal, argued against taking this leap into the heart of the Mideast conflict.  Israel has already announced that it will ignore the Court\'s decision. American officials have vowed to veto any resolution at the Security Council aimed at enforcing it. But European governments, already so eager to distance themselves from Israel, may find it much harder to disclaim what the ICJ -- and all its European judges -- have pronounced to be the legal rights and wrongs of the Mideast conflict. For Yasser Arafat, the decision now looks like an unalloyed boon, tossed to him without extracting any serious effort to help suppress ongoing terrorist activities.  In effect, the ICJ now claims that countries beset by terrorism must ignore terror threats and focus on the Court\'s priorities. It is a dangerous precedent for the U.S., which has often contended for interpretations of its rights, under international law, that a majority of U.N. members might dispute. To those who argue that the U.S. should join the new International Criminal Court, because that new court will be moderated in its rulings by the influence of European members, this ruling of an older U.N. court should be sobering.  The ruling raises still broader questions about the U.N.\'s capacity to contribute to any serious international effort against terrorism. Even U.N. judges, we now see, have other priorities. '),(5509,3186371,'10,May,2004','2004-05-10 00:00:00','   I kind of like this new blogger. Some of my friends do not, and cite their uncontrollable conservatism as a reason. Screw them. '),(5510,3186371,'30,June,2004','2004-06-30 00:00:00','   The Icelandic government made the decision in 2003 to be a part of the Coalition of the Willing, and has been a supporter of the US-led War on Terror and the War in Iraq. The Icelandic opposition, with the \"pacifist\" red-green alliance in front have criticized that decision on numerous grounds (always with the underlying assumption that THIS war is evil because it\'s being waged by American troops - the Palestinian \"struggle\" is not supported by the US, and therefor it is not Evil - but I digress).  One of the attacks on the government is based on the fact that the decision to enter the war wasn\'t presented to AlÃ¾ingi (the parliament) for approval, but that the leaders of the two coalition parties, the prime minister and the foreign minister, made the decision themselves and then acted on it.  This is true. Iceland did practically declare war on another country and the decision was made by two men. This is also a state of affairs which is altogether unacceptable.  The decision to declare war on another country is one of the gravest and most important any government can make, and therefore there should be safeguards against that kind of decision being made lightly.   The awkward fact of the matter is, however, that in the pacifist constitution of Iceland there are no provisions on how to declare war! It is acknowledged in various statutes that Iceland can be at war, just one example is art. 89 in the  urlLink criminal code of 1940 . \"If an Icelandic citizen takes up arms in war against the Icelandic state or its allies, he shall be sentenced to no less than two years imprisonment.   \"He who, when the Icelandic state is at war, or when war is imminent, offers aid and comfort to its enemies, either in word or deed, or in any other way weakens the resolve of the Icelandic state or its allies, shall suffer the same sentence.\" This (loosely translated) article clearly shows that it\'s assumed that Iceland can be legally at war with another country.   Some of my foreign brothers might now ask \"what\'s the big problem here? Just throw in an amendment to the constitution stipulating parliamentary consent to any and all warmongering - that should straighten things out\" - and they would be right.  The problem however lies - as it so often does - with my socialist, \"pacifist\" countrymen. They stubbornly hold to the notion that Iceland is and always shall be neutral and therefore they can\'t bide the thought that the possibility of war be codified in law - even if such codification would doubtlessly lessen the chances of Iceland being at war!  It\'s kind of like the mentally ill person who thinks that \"only sick people eat pills - therefore if I don\'t eat my pills, I\'m not sick\".   I happen to think that the government was both right and courageous to stick with the US and UK in the war - even if we couldn\'t offer much material support. The decision to start a war is however much too important a decicion to leave the process as is. As with all government activities, we need to make it as difficult for the state to do what it wants to do - even when it wants to do things we approve of. '),(5511,3186371,'29,June,2004','2004-06-29 00:00:00','   \"As a test of the state of \"Bush the Nazi\" rhetoric, I went to Google and typed in \"Bush is a Nazi\" and got 420,000 hits, well behind \"Hitler was a Nazi\" (654,000 hits) but then Hitler WAS a Nazi and had a 75-year head start.\"  urlLink John Leo , on Townhall.com. '),(5512,3186371,'28,June,2004','2004-06-28 00:00:00','   The  urlLink Samizdatistas  have this to say about  urlLink James Lileks\'  run-in with a young statist: \"Mr. Lileks, sturdy Midwesterner that he is, believes people should should earn their money. His earnest young interlocutor, following in the sadly well-worn path of Minnesota socialism, thinks money should shower down like manna from heaven.\" OK, when you quote someone quoting someone else - is that metaquoting? '),(5513,3186371,'28,June,2004','2004-06-28 00:00:00','   Analogies and metaphors are often powerful weapons on the battleground of ideas. A vivid metaphor can become a very successful meme and be instrumental in winning, or losing, a given rhetorical battle.  A very effective metaphor used by the anti-smoking lobby here in Iceland is the one that smoking in a restaurant/bar/cafe is like peeing in a public pool. You wouldn\'t like to swim in a pool with \"peeing\" and \"non-peeing\" areas would you - so you shouldn\'t want to have non-smoking areas in restaurants either.  This metaphor is effective because it is at first glance wholly applicable (i.e. corresponds point by point to restaurant smoking) and is extremely vivid - it rouses feelings of disgust in the average person.  It is however inadequate - or at least it leaves out some things which really ought not to be left out.   First is the point that if I own a pool and want to be able to pee in it, and allow others to pee in it - I should be allowed to, right? If there is a sign by the pool which states that this is a wiz-friendly pool, why shouldn\'t people who like that sort of thing be allowed to swim in foreign urine? It isn\'t as if the world is one pool which we all have to swim in - there are millions of pools, large and small, and if some people want to pee in their pools - however disgusting we find the practice - who are we to prohibit that?   It\'s the same with bars and restaurants - there are millions of bars and restaurants, all of them privately owned, and no one is forced to visit any one of them. Some people (the staff) are even paid to spend time in them, but can leave at any time.  The second point is that the only logical argument in favor of banning smoking is that second hand smoke may be harmful. The metaphor, however, only stresses the \"ick\" factor - pee is disgusting and so (the argument goes) is smoke. Pee, however, is not harmful in any way, and swimming in diluted urine might even be healthy (They used the stuff for washing wool in Iceland for centuries - it has certain soap-like attributes). You can\'t seriously advocate that the government ban an activity just because it is \"icky\"!  As I said in the beginning metaphors and analogies can be powerful weapons, and if we are to win the battle of ideas and bury socialism for good, we must be able to use those weapons effectively and also be able to find the flaws in the weapons of the oposition and point them out before the meme takes its\' viral hold on the collective intellect.  I\'m of course all in favor of good libertarian memes taking viral holds where ever they can get their little hands to grip. '),(5514,3186371,'24,June,2004','2004-06-24 00:00:00','   I\'d like to thank  urlLink John Ray  and  urlLink Willy Sutton  for linking to my post on the Icelandic economic progress during the nineties. I\'d also like to congratulate  urlLink Shay  on being the first to comment on a post by me. These three are excellent bloggers and I recommend their blogs to all and sundry.  I also think it\'s kind of cool that John is Australian, Willy is Icelandic and Shay is American. God bless the Internet. '),(5515,3186371,'24,June,2004','2004-06-24 00:00:00','   urlLink    Click for larger image. '),(5516,3186371,'24,June,2004','2004-06-24 00:00:00','   urlLink Veronique de Rugy  writes a rather flattering account of the Icelandic economy and Icelandic economic evolution over the last fifteen years or so, corresponding roughly with the term of our current, but soon to be former, prime minister, DavÃ­Ã° Oddsson. Mr. Oddson gave a  urlLink talk at the American Enterprise Institute  on June 16. this year.  Mr. Oddson is not a libertarian, but a free market conservative of the Scandinavian variety (a little more of an economic socialist than conservatives in the US, and a little more liberal in personal matters than US conservatives). That said, it has to be acknowledged that Iceland has enjoyed almost unprecedented growth under his leadership, and that is largely thanks to the fact that he acted on his promised liberalization of the economy - unlike most politicians in the West.   The situation when he took power in 1991 was bleak, to say the least.   \"Political power infused the entire financial system, which could hardly be distinguished from the political system. Furthermore, the fiscal position at the start of the nineteen-nineties was far from acceptable. The persistent Treasury deficit and the mounting debt that accompanied it were, in my opinion, one of the greatest threats to the Icelandic economy in those years. A recession in the early nineteen-nineties, which was caused among other things when the all-important cod quota had to be cut back by as much as one half, compounded the existing problem.\"  The solution, according to Mr. Oddson, quoting Milton Freedman, was \"freedom\", and lots of it.   \"The freedom of the individual and the freedom of the nation are the foundation of all well-being--spiritual and material alike. The motive behind attacking the Treasury deficit was not only its bad economic consequences. An equally important consideration was that a persistent deficit is a sure-fire recipe for higher taxes in the future, and higher taxes erode the freedom of the individual.\"  The Oddson government cut taxes and public expenses left and right and now, although we still have a nationalized health care and education systems we enjoy a lower tax burden than most Europeans and more economic freedom and growth.   The following is a table of GDP in Iceland from 1990 to 2003, with the GDP of 1990 being 100. It is based on information from the Icelandic Government.  YEAR - GDPINDX 1990 - 100.00 1991 - 100.10 1992 - 96.79 1993 - 97.56 1994 - 101.49 1995 - 101.58 1996 - 106.88 1997 - 111.87 1998 - 118.12 1999 - 123.07 2000 - 129.96 2001 - 133.44 2002 - 132.73 2003 - 138.07   As you can see, the first couple of years under Mr. Oddsson were not pleasant. There was much tightening of belts and growling of stomachs. However in 1996 this austerity started to pay off and in the ensuing 8 years the economy advanced by leaps and bounds.   The following is a table of GDP per capita ( in $US) from 1990 to 2003. PPP adjusted numbers come first after the year, and the second column is the unadjusted GDP per capita.  YEAR - PPP ADJ - UNADJ 1990 - $20,452 - $24,978 1991 - $20,952 - $26,373 1992 - $20,504 - $26,690 1993 - $20,974 - $23,178 1994 - $22,106 - $23,601 1995 - $22,497 - $26,139 1996 - $23,969 - $27,130 1997 - $25,275 - $27,241 1998 - $26,279 - $29,214 1999 - $27,136 - $30,363 2000 - $27,927 - $29,850 2001 - $28,968 - $26,683 2002 - $28,381 - $29,600 2003 - $29,199 - $36,320   This table tells the same story, and shows clearly how well the country is doing now. Of course we could be doing even better with even lower taxes and less regulation. Thankfully Mr. Oddsson has plans:  \"Personal income tax has already been lowered and during the current governmentâ€™s term of office it will be reduced by a further four percentage points. An income tax surcharge on the highest incomes will also be removed, and inheritance tax has been reduced and harmonised at a level which will never exceed 5 percent. In addition, it is planned to reduce value-added tax on food, which will benefit the lowest income groups in particular. I am convinced that these tax cuts will greatly strengthen the Icelandic economy well into the future. And not just the economy. The whole of society flourishes and becomes more diverse when each and every one of us keeps more of what we earn and can dispose of it as we please.\"  As I said in the beginning, Mr. Oddsson is a conservative politician, and I find quite a number of his past actions and present beliefs unpalatable. But if I can\'t get a Libertarian PM, I\'ll settle for a conservative of Mr. Oddsson\'s caliber.  I now declare the love-fest over and will go back to bashing the conservatives for their statist attacks on free speech (the current row in Iceland). '),(5517,3186371,'23,June,2004','2004-06-23 00:00:00','   \"Care for language is more than a concern for purity. When one describes President Bush as a fascist, what words remain for real fascists? When one describes Fallujah as Stalingrad-like, how can we express, in the words that remain to the language, what Stalingrad was like?   \"George Orwell wrote that the English language \"becomes ugly and inaccurate because our thoughts are foolish, but the slovenliness of our language makes it easier for us to have foolish thoughts.\" In taking care with language, we take care of ourselves.\"   urlLink Bret Stephens  in the Wall Street Journal. '),(5518,3186371,'22,June,2004','2004-06-22 00:00:00','   Compared with the major Indexes the Icelandic ICEX-15 index is really in a league of its own. From the start of the year ICEX has risen by more than 40%, while the Norwegian OBX index has risen by a respectable 13%. Other Indexes have not made it above 10%!     The rise has for the large part been driven in huge advances in the stock prises of a couple of large companies,  urlLink Ã–ssur ,  urlLink Marel ,  urlLink KB Bank  and  urlLink BakkavÃ¶r  in particular. Foreign investors have shown considerable interest in Ã–ssur and Marel and the rise in KB Bank and BakkavÃ¶r is fueled by the foreign acquisitions of the two companies, in the Danish investment bank FIH by KB Bank and in the food distributor Geest plc by BakkavÃ¶r.  Analysts are not of the opinion that this is a temporary bubble as the rise in the aforementioned companies is not based on speculation alone but genuine, long-term investment and increase in assets.  Some investors have said that the Icelandic currency, the KrÃ³na, or rather the fact that we have a special currency (as opposed to the EUro) is a hindrance to foreign investment, but the markets seem to prove them wrong. Something to keep in mind when the EUrocrats try to squeeze us into the EU in the future. '),(5519,3186371,'22,June,2004','2004-06-22 00:00:00','   \"The economic boom of the 1990s proved to be partly a financial bubble that we now know began to burst with the stock market break in April 2000. The growth that was real resulted less from any specific policy than from the gridlock between Mr. Clinton and the GOP Congress that kept government from doing much at all after 1994. No period since the 1920s was as free of new federal intrusion.\"    urlLink The Wall Street Journal , giving minarchists everywhere ammunition in the fight for smaller government and real seperation of power. Sadly Iceland is burdened with too much government interference in the affairs of its citizens, and a parliamentary system which all but gives the executive dictatorial power over the legislative branch. That, coupled with complete authority in judicial appointments - parliament has no say in the matter - makes all talk of seperation of power in Iceland a joke. '),(5520,3186371,'20,June,2004','2004-06-20 00:00:00','   The Guardian is probably not the best source for \"big issues\", but it is the only big paper that once in a while reports on Icelandic affairs.  urlLink This piece  on the recent crisis over the media law (which The Guardian calls \"The Baugur Law\") and the ensuing constitutional crisis is factually correct.  As I said earlier it is sad that an otherwise excellent prime minister has decided to soil his legacy with legislation as oppressive and statist as this.   The only argument for the law has been the, admittedly severe, bias in some privately owned media against the government. That a center-right government is actually moving against the press because of the editorial policies of some its members is not libertarian, it\'s not even conservative - it smacks of totalitarianism. It\'s an ugly thing to say about a man who has done more to liberalize the icelandic economy than any other since the independence hero JÃ³n SigurÃ°sson in the 1870\'s - but it\'s not far from the truth. '),(5521,3186371,'20,June,2004','2004-06-20 00:00:00','   \"To describe this film [Farenheit 9/11 by Michael Moore] as dishonest and demagogic would almost be to promote those terms to the level of respectability. To describe this film as a piece of crap would be to run the risk of a discourse that would never again rise above the excremental.\"  Christopher Hitchens in  urlLink Slate . '),(5522,3186371,'20,June,2004','2004-06-20 00:00:00','   \"Scientists have performed successful  urlLink teleportation  on atoms for the first time, the journal Nature reports.\" '),(5523,3186371,'20,June,2004','2004-06-20 00:00:00','   For the last twelve months the Icelandic Icex-15 index has more than doubled. The combined value of all stocks registered with the Icelandic Stock Exchange is more than the 2003 GDP of Iceland! If this isn\'t good news, I don\'t know what is.     When the current prime minister,  urlLink DavÃ­Ã° Oddson , took power in 1991, every analyst was predicting doom and gloom for the future of Iceland. Inflation and unemployment was rife and overreglulation was crushing business everywhere.  Over the last 13 years all banks have been privatized, along with countless other former state-run businesses, taxes have been lowered - especially on business - the national bank has been given its independence. Now the GDP per capita is among the best in the world, only slightly lower than in the States, and considerably higher than in the EU (big surprise).  It\'s a shame the PM is soiling an otherwise incomparable legacy with his recent tantrums and moves against the freedom of speech and of the press. '),(5524,3186371,'20,June,2004','2004-06-20 00:00:00','   Metallica is coming here to Reykjavik and the concert is going to be held, on all days, on the fourth of July. For that reason I find it appropriate to publish the lyrics to one of their best songs, with some links inserted for the readers enjoyment and enlightenment.    Metallica:  urlLink Don\'t Tread on Me .    urlLink Liberty or death , what we so proudly hail Once you provoke her, rattling of her tail Never begins it, never, but once engaged... Never surrenders, showing the fangs of rage   urlLink Donâ€™t tread on me   So be it Threaten no more  urlLink To secure peace is to prepare for war  So be it Settle the score Touch me again for the words that youâ€™ll hear evermore...   urlLink Donâ€™t tread on me   Love it or live it, she with the deadly bite Quick is the blue tongue, forked as lighting strike Shining with brightness, always on surveillance The eyes, they never close, emblem of vigilance   urlLink Donâ€™t tread on me   So be it Threaten no more  urlLink To secure peace is to prepare for war  So be it Settle the score Touch me again for the words that youâ€™ll hear evermore...   urlLink Donâ€™t tread on me   So be it Threaten no more To secure peace is to prepare for war   urlLink Liberty or death , what we so proudly hail Once you provoke her, rattling on her tail  So be it Threaten no more To secure peace is to prepare for war So be it Settle the score Touch me again for the words that youâ€™ll hear evermore...   urlLink Donâ€™t tread on me . '),(5525,3186371,'29,July,2004','2004-07-29 00:00:00','   From the  urlLink Mises Institute : \"\'Need\' now means wanting someone else\'s money. \'Greed\' means wanting to keep your own. And \'Compassion\' is when a politician arranges the transfer.\" Hat-tip:  urlLink Pragmatic Libertarian . '),(5526,3186371,'23,July,2004','2004-07-23 00:00:00','   urlLink Booker Rising  did  urlLink two  articles on  urlLink Iceland  recently and then Molotov  urlLink asked  if I could comment on Icelandic immigration policy. Keep in mind that as a libertarian I would like to have all barriers on immigration and emigration everywhere dropped. Freedom is the answer in this as most other matters. Reality, however, often lets the ideal down.  I would like to write some more about the matter in the future. Recently there was a big discussion in Iceland about immigration as the immigration act was amended to make it more difficult to enter the country. The libertarian society to which I belong was at the forefront in the fight against the changes, but sadly we lost. That post has to wait, however, as my employer has first dibs on my time at the moment.  Sadly it\'s true that there was an unofficial \"no black servicemen\" policy until the mid-sixties. Iceland has no military and relies on the US for protection. The racist policy was never put in writing, but it seems that the demand was made verbally around the time of the signing of the defence treaty in 1951. President Kennedy was instrumental in changing that.  Icelanders are mainly white, although recent genetical studies seem so indicate that we are more celtic than Norse in origin. The Vikings took more Irish slaves with them on the trip than was previously thought.  This homogenized nature of the population has made Iceland somewhat of a promised land for Nazis and skinheads the world over. This is not something that I, or most Icelanders are happy with. God knows there are few people I would less want as neighbors than the Klan.  This is changing however, and changing fast. Numbers of immigrants - almost all of them legal as it is rather difficult to get illegally in here - are rising with every year. The largest numbers come from Denmark and Poland, but South-East-Asia is well represented with immgrants from Thailand, the Philippines, Vietnam and Indonesia becoming more numerous.  Immigration policy is rather strict, and was made more so with the passing of a new immigration law last spring.  There are no quotas on immigration, but you have to have a work permit before you get a residents permit. The work permit is held by the employer, not the immigrant, and that puts the immigrant in a very precarious position.  I imagine that the largest number of Blacks in Iceland are US servicemen. I don\'t know of any special immigration treaties between the US and Iceland, but since the Icelandic constitution does not allow any discrimination based on race or color Black Americans would get the same treatment from the immigration authorities as whites. (The small size of the country has kept transparency high and corruption low).  The small size of the population also means that changes in nationwide attitude are faster than in larger populations. Although racism was probably much more prevalent a couple of decades ago - and there still is racism in Iceland - there is much less of it now. The same goes for other prejudices, for example as regards homosexuality.   On the whole I would think Iceland is a fine country to live in - for Black Americans as well as others. I would think that most people would regard Black Americans as Yanks first and Blacks second (or third, fourth or whatever). Also the more people of different colors and creeds we have here the faster the country is bound to lose its remaining prejudices.  Also I would welcome with open arms any good fiscal conservatives/libertarians of the American School. So get on over here! '),(5527,3186371,'23,July,2004','2004-07-23 00:00:00','   Haven\'t been as productive a blogger lately as I would have wanted but work has been crazy and after ten hours of writing about bonds and the Housing finance fund I have simply been all written out.   People who want to know more about Icelandic society and nature should check out these two websites. On the one hand is  urlLink Iceland Review  - a news website in english - and the other is  urlLink Icelandic Geographic , an annual publication which deals with Icelandic nature and wildlife. '),(5528,3186371,'19,July,2004','2004-07-19 00:00:00','   From  urlLink The Wall Street Journal  (emphasis mine): When it comes to Canadian identity, Fox News Channel is apparently a threat. Al-Jazeera, on the other hand, is just another point of view enriching Canadian culture. That\'s the message sent last week when the Canadian Radio and Television Commission (CRTC) granted the Qatar-based, anti-American al-Jazeera a Canadian license. Fox has wanted into Canada since 1999 but has so far been shut out -- except in Ottawa,  at the Canadian Parliament , which requested and got a Fox feed last August.  [...]  The darkest interpretation of all of this is that Canada\'s liberal political elites find al-Jazeera\'s view of the world more compatible than Fox\'s. The Arab network is always putting the worst face on U.S. policies, while Fox doesn\'t mind suggesting that it is rooting for America to win when it goes to war. We prefer to think better of our neighbors to the north, but it would be nice if they gave Americans the same benefit of the doubt that they give to our enemies. Isn\'t it delicious that the Canadian politburo thinks ordinary Canadians can\'t handle Fox, but the politicians deem themselves able to withstand it\'s dreadful propaganda? This is what happened during prohibition here in Iceland. There were errand-boys employed by the Althingi (parliament) whose main job was to run back and forth between the parliament building and the pharmacy, where they bought pure ethanol for the legislators - who didn\'t consider themselves bound by the same laws as the rest of Iceland.  Not that Icelanders in general considered themselved bound by the law - illegal production of schnapps was enormous, and the traditional Icelandic Schnapps, \"Black Death\", was born during those years.   The funny thing was that a majority voted for prohibition in a referendum in 1915. Another classic case of \"I think it\'s a good idea that X should be banned. I can partake, of course, but it\'s bad for others who don\'t know how to handle it\". '),(5529,3186371,'15,July,2004','2004-07-15 00:00:00','   Amusingly, or incidentally - however you want to put it -  urlLink Roger Bottle  writes in the Telegraph about CPI and the question of wether it is worth anything. With the huge reservation that he IS the manager of Capital Economics, and therefore not a novice when it comes to these matters, I would venture to point out that what he sometimes calls inflation is really changes in purchasing power. That said - the quotes:      \"Since there are no prices for the output of the public sector, estimating this index involves taking the increase in government spending and adjusting that for the \"real\" increase in what public spending buys. And if you think it is easy to measure that then you are a better man than me, Gunga Din.       \"Even measures which restrict themselves to household spending show quite a divergence. The overall household spending deflator, which covers a wider spectrum of household spending than the CPI, shows inflation at 1.8 per cent, not that different from the CPI, while the retail sales deflator, which covers a narrower range of household spending than the CPI, shows inflation at minus 1.1 per cent.\"         urlLink    Click image to enlarge  (image nicked from the Telegraph, but hosted by me - I only steal intellectual property - not bandwidth)     \"Yet widespread scepticism about the CPI cannot entirely be put down to an illusion. When different things are rising at very different rates, what does \"the\" inflation rate mean? The statisticians who compile the CPI attempt to measure the overall increase in prices by weighting the increase of each individual component of the index by its share in overall spending.   \"Of course, very few people will have an expenditure pattern which exactly mirrors this average weighting. Each person effectively experiences a different inflation rate, depending upon their own mix of expenditure. If you are a heavy buyer of clothing and footwear, CDs and household furniture you should not be complaining. On the other hand, if you have children at private school and are a heavy user of hotels and restaurants your own personal inflation rate is in a different parish from the official CPI measure.\"  Interesting stuff this. Hat-tip:  urlLink Prestopundit . '),(5530,3186371,'15,July,2004','2004-07-15 00:00:00','   If, as we suspect, inflation/deflation is a direct derivative of the quantity of money in circulation at any given time, and not a result of rising prices or wages - how do we measure real inflation correctly?    Is there a better (more accurate) method than the current method of calculating CPI\'s based on numbers of arbritralily chosen factors? If so - what is it?    Is it really necessary to measure inflation? Is the obsession with inflation that we have a result of the trust people (mis)placed in the Phillips curve until the eighties? Now that most economists no longer think the Phillips curve is an accurate yardstick in the fight to hold down unemployment - is there any reason for us to measure inflation?    Maybe there is, now that there is a Central Bank that issues money and regulates the quantity thereof - if it is to make (relatively) accurate decisions regarding the quantity of money in circulation it is better off knowing the appriximate levels of inflation.    If there were no one institution or Bank who could issue money - if anyone and everyone could issue money - would we then need to know the levels of inflation? And if we knew - would it do any good?    Answers and comments would be very much appreciated.  -----------------------   From Statistics Iceland:   The Harmonized Indices of Consumer Prices average for member countries of the European Economic Area (EU-countries, Iceland and Norway) was 115.5 points (1996=100) in June, remaining unchanged from May. At the same time, the HICP for Iceland was 128.9 points, 0.5% higher than in the previous month.     The 12-month rate of change of  the HICP was 2.3% for the EEA, 2.4% for the Euro-zone while it was 2.9% for  Iceland.    The highest inflation rate in the EEA in this twelve month period was 8.1% in the Slovak Republic and 7.5% in Hungary. In Finland the consumer prices went down by 0.1% while the inflation rate was 0.9% in Denmark.     urlLink    Click image to enlarge '),(5531,3186371,'15,July,2004','2004-07-15 00:00:00','   David Boxenhorn  urlLink responds  to my post on the CPI. \"Before I go on, it is important to know that in order for the CPI to be meaningful, it has to roughly represent the buying habits of the public. As an example, suppose the price of movies suddenly goes up ten times. Has your buying power decreased? That depends on whether you go to movies or not. If so, by how much? By the percentage of your income that you usually spend on movies. You may note that there is a problem here &#8211; if the prices of movies go up ten times, you will surely substitute some other form of entertainment for movies, so your buying power will be impacted less than it otherwise would be.\" David is surely right here, and that begs the question - how can an arbitrarily chosen CPI give accurate results? As David points out, a rise in the price of one commodity may even mean that I spend less on said commodity in the future. I may feel that 800 kr. is a fair price for a movie ticket and may go to the movies six times a year. If the movie ticket price were to double I certainly would cut down on cinema visits, even possible stop going altogether, and the price hike would therefore have little effect on my purchasing power. \"Getting back to our data, the first thing I noticed was category 02. Do Icelanders really spend 4% of their income on alcohol and tobacco? Almost as much as clothing and footwear? More than health? And what about educational services &#8211; only 0.6%? Something is wrong. Then I realized, of course, health and education are provided &#8220;free&#8221; by the government! But they are not really free, someone is paying for them, but can we know how much? Maybe we should include the amount the government spends on these services per capita?\" David then suggests that taxes should constitute the 13th category in the Icelandic CPI. Not a bad idea at all.   As for the 4% of Alcohol and Tobacco. If you lived in a country with about five hours sunlight during the winter you would need a stiff drink every now and then; also according to  urlLink John Ray , nicotine may have a anti-depressant effect - again with the dark-induced depression.  But seriously, Icelanders may drink a bit too much, but that is, I think, a result of a larger societal problem - Icelanders have difficulties expressing emotion, any emotion, when sober, and alcohol may act as a release - its not a healthy solution, but it is an explanation of sorts.   Update  Of course *slaps head* the monstrous taxes placed by the state on alcohol and tobacco account for a very large part of this. We don\'t neccesarily drink more of the good stuff - we just pay more for it. '),(5532,3186371,'14,July,2004','2004-07-14 00:00:00','   From the  urlLink Central Bank of Iceland .  \"The Central Bank of Iceland has revised the currency weights in the basket that the Bank uses to calculate the exchange rate index of the krÃ³na. Weights are determined on the basis of Icelandâ€™s external trade in 2003. The last revision was made in July 2003. Revised currency weights and the changes from the previous weights are shown in the accompanying table. The revised basket will be used to calculate changes in the exchange rate index as of Friday, July 16, 2004 until the next review at a similar time next year.\"   urlLink    Click to enlarge  Why the Bank does things like this instead of letting the currency just float on an open market I don\'t know. Still it\'s better than to let politicians decide the exchange rate to suit their short term political interests.   Update  Acting on an  urlLink anonymous  tip I went  urlLink here  and read this: \"The exchange rate of the krÃ³na is determined in the interbank market for foreign exchange which is open from 9:15 to 16:00 on weekdays.\" Seems I was wrong. Always good to learn new things. '),(5533,3186371,'14,July,2004','2004-07-14 00:00:00','   From Lord Butler\'s  urlLink report :  \"We conclude also that the statement in President Bush\'s State of the Union Address of 28 January 2003 that \'The British Government has learned that Saddam Hussein recently sought significant quantities of uranium from Africa\' was well-founded.\" Not than anyone will listen - the War For Oil was based on a Lie, right? '),(5534,3186371,'14,July,2004','2004-07-14 00:00:00','   From the  urlLink Washington Institute : \"Ironically, three countries -- India, Saudi Arabia, and Turkey -- condemned Israel at the UN General Assembly and voted to refer the Israeli fence to the ICJ for an advisory opinion even though they had themselves built barriers in areas contested by their neighbors. India is just completing a 460-mile barrier in the contested Kashmir to halt infiltrations supported by Pakistan; within the last two years, Saudi Arabia built a sixty-mile barrier along an undefined border zone with Yemen to halt smuggling of weaponry; and Turkey built a barrier in an area that Syria claims as its own.\" Hat-tip:  urlLink Dissecting Leftism.   Sure, the Israeli government has done and is doing things which are abhorrent to Libertarians everywhere. But, because you see that Israel is clearly not being treated fairly in the UN or in the world press, you can\'t help feeling sympathetic. '),(5535,3186371,'14,July,2004','2004-07-14 00:00:00','   Not really my fight as I\'m not American, but I\'m still glad they didn\'t fudge up the Constitution with this crud. The Senate dealt an election-year defeat Wednesday to a  urlLink constitutional amendment  banning gay marriage, rejecting pleas from President Bush and fellow conservatives that the measure was needed to safeguard an institution that has flourished for thousands of years. '),(5536,3186371,'14,July,2004','2004-07-14 00:00:00','   Willy Sutton was  urlLink wondering  what factors to the CPI.   Being the obsessive-compulsive that I am I went and found out. The results are posted  urlLink here . They are, as always, courtesy of  urlLink Statistics Iceland  (Hagstofan).  I am however not enough of an economist to debate whether the idea of a CPI is a sham or not (I have a faint suspicion that Willy is of the former opinion). '),(5537,3186371,'14,July,2004','2004-07-14 00:00:00','   urlLink Amritas  has it down pat when he says that  barn  is cognate to the word  bear  (as in to bear a child). The word  trÃº , however, is cognate to the words  traust  and  tryggÃ° , meaning  trust  and  loyalty  respectively. They in turn are cognate to old germanic, celtic and slavic words meaning strong, healthy, active etc. Examples are the ancient prussian  druwis  (loyalty), welsh  drut, drud  (strong), ancient slavic  su-dravu  (healthy) and the sanscrit  darunÃ¡  (hard, strict).   So the modern Icelandic word trÃº (religion) originally meant to be loyal to your god. Not necessarily to have faith in him/her.  The old word for religion -  siÃ°ur  - meant, and still means, tradition, behaviour, morality.  SiÃ°aÃ°ur maÃ°ur  means a well behaved, well brought up, moral man. '),(5538,3186371,'13,July,2004','2004-07-13 00:00:00','   What do social conservatives have to say about this? On one hand they are for harsher punishments for everything, up to and including sneezing in public, but on the other they believe in the absolute sanctity of (heterosexual) marriage.   urlLink    Vexing. '),(5539,3186371,'12,July,2004','2004-07-12 00:00:00','   David Boxenhorn writes some very interesting things about the naive Icelandic notion of BarnatrÃº (child\'s faith) and the jewish Talmud in the  urlLink comments section  of his blog. \"Now imagine that the intellectual elite embraced barnatrÃº, and theologians of diverse points of view spent around 500 years discussing it, and reconciling it with chapter and verse under the assumption that the Bible was the source of truth â€“ and that it was all written down, interspersed with â€œtraditional stories and mythsâ€ â€“ youâ€™d get something like an Icelandic Christian version of the Talmud (lit. Learning). Imagine further that over this time a practical consensus was worked out as to the public and social aspects of observance, while leaving private aspects and theology undefined â€“ youâ€™d get an Icelandic version of halakha, Jewish law.  Jewish law is a common-law system, like English law, and unlike Roman law. Instead of going back to the original laws for interpretation, the whole weight of tradition must be taken into account. In this sense Judaism is inherently conservative, as opposed to (at least the Protestant form of) Christianity. But it is also based on the common sense of its practitioners, rather than a literal reading of the original text. In that way, I think, the study of Jewish law is the study of human nature.\" This is very interesting, and may require some digestion before I comment. '),(5540,3186371,'12,July,2004','2004-07-12 00:00:00','   A couple of links regarding the religious demographic structure of the Icelandic nation.  First an  urlLink Excel-file  from the government statistics office detailing the changes in numbers of Icelanders belonging to religious organisations between 1997 and 2002, which should give some indication on which religions are growing and which are in decline. All religious orgs are in the file so get ready for some pathetic numbers (in 2002 there were eleven people in Iceland in the First Baptist Church).  (Side-note: Don\'t you think it\'s lovely that the government collects and stores information like that about the citizens?)  Then, there is this  urlLink doohickey . It\'s all in Icelandic, but it has two more years of information (1996 and 2003). '),(5541,3186371,'12,July,2004','2004-07-12 00:00:00','   The  urlLink Consumer Price Index  compiled in July 2004 is 234.6 points (May 1988=100), 0.47% lower than in the previous month, most analysts had predicted a change of 0.0% to 0.2%.        urlLink CPI   Current Value |  234,6  12 Months Change* |  3,6%  3 Months Change* |  4,6%  Last Month Change* |  -0,5%   The discrepancy is mostly attributable to changes in the system of the State Housing Financing Fund. The index for owner occupied housing went down by 0.6% (effect on the CPI -0.08%). Housing prices rose by 0.7%, but the aforementioned system changes overrode that.  * Annualized rates '),(5542,3186371,'12,July,2004','2004-07-12 00:00:00','   urlLink David Boxenhorn  writes about an  urlLink Icelandic woman  who tells of her conversion to Islam, and about an  urlLink Icelandic website  devoted to Islam. The following is a post adressing these, and other issues raised in correspondance between us.  First, I donâ€™t know the woman in question, and her story isnâ€™t very interesting to one, like me, who isnâ€™t interested in becoming a muslim. What is striking is the extremely partisan anti-Israeli feeling that pops out in a couple of sentences in the letter.   The thing is that Iceland has the same problem as many western countries â€“ rising anti-semitism disguised as criticism of the Israeli government. Talk for more than a couple of minutes with these people (more on the left than the right, but there are lots of anti-Israelites on the right as well), and they start telling you how the Jews control America, they control American TV and Hollywood, they control the IMF and the World Bank etc. etc. This, coupled with the vintage anti-semitic argument that all Jews stick together and plot incessantly, is used to explain away the support the US government sometimes gives to Israel.  I thought I had seen a poll wich measured anti-Israeli feeling among Icelanders, but as I canâ€™t find it I donâ€™t want to quote it â€“ I canâ€™t remember the numbers accurately, but they were quite high.  I am not saying that Iceland is a hotbed of rampant Anti-semitism. Most people donâ€™t think much about Jews in general, but the only time you hear the word â€œJewâ€ on the news is when Israel is discussed. International news here in Iceland go through a double leftist filter â€“ first the leftists abroad who submit the stories and then the leftists here in Iceland who read and write them (I am of course generalizing, there are many dedicated and unbiased journalists here in Iceland, and the liberal takeover of the media hasnâ€™t been as thorough in Iceland as in the US, but TV and Radio is more left-leaning than some newspapers). Therefore the news from Israel is usually biased in favour of the Palestinian cause.   All this said, I am dismayed that an Icelandic woman writes so harshly about Israel and Jews, but I am not surprised. No doubt her stay in Arab countries has fostered anti-Israeli feelings in her, but the seeds were sown here in Iceland.  As to the website  urlLink www.islam.is  it seems to me a very benevolent, tolerant website about the Koran and the Icelandic muslim community.  Icelanders may seem to outsiders a fairly godless bunch â€“ many children born out of wedlock, lots of couples not bothering to get married etc. etc. All this is true, but Icelanders are, if anything, more religious generally than many westerners. The reason many children are born out of wedlock is simply that for quite some time there hasnâ€™t been a stigma attached to being a single mom. Itâ€™s just a part of society, and therefore people arenâ€™t jumping to marry just anyone before the child appears. Iâ€™m sure more people would get married if the tax law didnâ€™t actively discriminate against married people. Young couples with children get more money from the government if they donâ€™t get married.  The Icelandic tradition towards the family and religion is that it is very much your business, and noone elseâ€™s. Gay family members of leading politicians and illegitemate children of politicians may not have been to the liking of a lot of voters, but the general feeling was, and is, that it isnâ€™t something that has anything to do with politics. With religion it translates into the notion that you donâ€™t talk much about it, and you donâ€™t argue with people about their religion, but it doesnâ€™t mean that Icelanders arenâ€™t religious.   There is a term often used here,  barnatrÃº , or  childâ€™s faith . This is the faith that you take with you into adulthood from the time you were a child. This is often a rather generalized version of Christianity, you donâ€™t qoute chapter and verse, but you know your stories and the morals of the faith and you consider yourself a Christian. This is, I think, one reason why many Icelandic Christians have no problems with single mothers or homosexuality â€“ it may contradict some verses in the Bible, but it isnâ€™t in conflict with the more basic, maybe naive,  barnatrÃº .   I, for example, consider myself Christian. I donâ€™t go to church, and I have never read the Bible from cover to cover. My  barnatrÃº  tells me that God isnâ€™t going to punish me for using the reason or conscience that He gave me and I trust those better than most single bible verses.  The  barnatrÃº  seems to me however to be on the decline, and more dogmatic versions of Christianity, and other religions such as Islam, taking it\'s place. As a libertarian I think this isn\'t a good thing, as surrendering reason to authority can never be good, even if the authority is a book.  Of course, when you write about a whole nation, even one as small as Iceland, you are bound to generalize more than you should, so everything written here should be taken with a liberal helping of salt. '),(5543,3186371,'12,July,2004','2004-07-12 00:00:00','   ...just for fun. \"The mutated muscular soldiers of Mordor turned out to be hilariously ineffective fighters, a dozen of them held off by a single dying human. Apparently they made the beasts by crossing Orcs, Goblins and the  French .\" From 50 Reasons why  urlLink LORD OF THE RINGS Sucks ! '),(5544,3186371,'09,July,2004','2004-07-09 00:00:00','   urlLink Crispus  has this to say on recent developments with western companies outsourcing to African countries. \"While still a small trend, it is a growing and good one. African college graduates have access to jobs that pay $200-$500 per week, in countries where most people don\'t make this money in a year. It expands the consumer markets in these countries, which can help countries like USA. Companies get qualified workers for lower costs, which will increase jobs for Africa. Countries like USA can then focus our energies on sectors where we have a competitive advantage - which is innovation. Of course, French unions are getting nervous - which is why France\'s economy will continue to stagnate while others grow, because it doesn\'t focus on what the French do best at the lowest cost.\" This is a useful reminder for us who actually believe in free trade. When the Unions and other lefty pressure groups talk of jobs being lost over here and poor people in other countries being exploited by evil corporations - remember they are not thinking at all about the welfare of those foreign nationals - only about the political power that comes with running a union.  But Crispus, what the French do best is surrender - you can\'t make a living off that. Maybe they could offer their services as references to evil dictators, like an unscrupulous movie-critic who sells blurbs - there could be money in that, and it would all be government run and financed, which should suit the French statists. '),(5545,3186371,'09,July,2004','2004-07-09 00:00:00','   From  urlLink Yahoo News  (emphasis mine). \"A primary school in Wakayama Prefecture will provide electronic tags on students\' belongings that will help parents and teachers monitor their safety on the way to and from school, a local bureau of the telecommunications ministry said Wednesday.  \"The tags --  similar to those used for merchandise at retailers and wholesalers for inventory control  -- will be attached to students\' school bags or nameplates, while tag readers will be installed at the school gate and locations the parents and teachers think could be dangerous.\" This is of course what will happen in Iceland a couple of years after the institution of the car chip. '),(5546,3186371,'08,July,2004','2004-07-08 00:00:00','   I\'ve been wondering what individuals can do when faced with the overwhelming power of the state coupled with the force-multiplier that is modern technology, as in the Orwellian proposals outlined  urlLink here . Should we turn Luddite and smash the plastic tools of computerized oppression?   Nah. I\'m much to attached to the Internet for such drastic action.  We can, however, fight the government on technological terms. Like the proverbial gun who doesn\'t kill anyone - the computerchip doesn\'t spy on anyone, or intrude on anyone\'s privacy. The chip and all the techonlogical paraphernalia can be used for good as well as for evil.    urlLink This  is the system the Icelandic statists are so keen to expand and use in every single car in the Land. Can\'t someone crack the code it\'s software is written in, and be ready with a patch for general distribution when the crap hits the fan?  I am not advocating the vandalism of private property, and if the proposal is not carried through I will of course be against the hacking of the system - which is made by a private company for other private companies. If, however, a derivative of this same system is used by the government to spy on free citizens I have no qualms about people vandalizing the equipment to their hearts\' content - I just would like to be able to vandalize without the cops finding out - hence my plea to skilled hackers and computers geeks everywhere. '),(5547,3186371,'08,July,2004','2004-07-08 00:00:00','   \"Maybe Michael Moore is a man for our time. That\'s a pretty damning indictment of our time.\" Cathy Young in  urlLink Reason Magazine . '),(5548,3186371,'08,July,2004','2004-07-08 00:00:00','   urlLink     The excellent people of  urlLink White Rose  have seen fit to honor me with a link. I thank them and welcome all guests from Thorn. '),(5549,3186371,'08,July,2004','2004-07-08 00:00:00','   The government has just fired a broadside against civil liberties in Iceland. The Minister for Transportation, the Chief of Police in ReykjavÃ­k and the state \"Traffic authority\" have all let lose in just two days.   Yesterday  urlLink this article  was published in the MorgunblaÃ°iÃ° newspaper, for which I work. Excerpts: \"New ways to cut the number of road accidents have to be found, and one possible way is to install computer chips in every car and thereby increase the amount of government monitoring of driving. This was among other things mentioned in a speech by BÃ¶Ã°var Bragason, Chief of Police in ReykjavÃ­k, about the annual report of the office.\"  \'I think there isn\'t enough discussion about these things and that we are struggling with the same things for years or even decades, regarding something which is supposed to improve traffic safety, but doesn\'t.\'   BÃ¶Ã°var said there was a reason to ascertain whether technology couldn\'t be used to greater effect, than it has been, in the fight against road accidents.  \'I want to propose an increase in the number of surveilance cameras on intersections in the city, but I also want a task force to inspect wether technology can be used in the cars themselves. I have the idea, which can easily be implemented, to put a computer chip in every single car. The Police then could stop a given car, connect with the chip and see the way the car has been driven that day, and even before that day.\'\" Those of you who don\'t read english will have to trust me to have translated that properly - I know it sounds incredible.      The statist\'s utopia.  I really don\'t know what to say or do about this. The idea is so odious to me that I would think the negative aspects of it should be obvious to everyone. The Icelandic home secretary (who is also the attorney general) is, however, too much of a Blunkett to do anything about it. I actually think he would find this quite appealing, and would support this sort of legislation.  Today the same paper published four seperate articles. One asking the \"man on the street\" what he thought of the plan; another detailing a massive speeding frenzy in a large town in the north of Iceland - Akureyri (apparently quite a lot of people were cought driving over the limit);  urlLink third  telling the readers that speeding was the major cause of fatal road accidents - and announcing a newly launched government campaing against speeding and the fourth telling of a speech made by transport secretary where he expressed the view that more speed cameras and computer chips in cars were quite nifty ideas (not all articles make it to the online edition).  The first article was a bit depressing, but with a ray of hope. Two out of the four people asked (or whose answers were published) expressed a negative opinion of the proposal while the other two thought it a \"step in the right direction\" and \"a very good thing\". Of course it is a depressing fact that there are people out there who think this is a reasonable way to reduce the number of accidents on the roads, but the ray of hope is that the sensible two were quite a bit younger than the middle-aged statists.  The other three articles were the standard scarmongering fare that seems to accompany these sorts of proposed slashings of our personal liberties. The aim here is to convince us that we live in constant, immediate danger of being killed or mutilated on the roads and that \"something\" has to be done. That \"something\" is not usually something wich will actually have any effect other than to tighten the rope around our necks.  The statist can never be happy as long as individuals have some modicum of freedom of action and travel, hence these proposals. This kind of surveilance system, coupled with a court system which allows for any and all evidence to be submitted in a criminal trial - without regard to how it was obtained (f.ex. illegal wiretaps are admissable), is a brutal attack on the personal liberties of Icelanders.   Armed revolt would seem a better option if it wasn\'t for the draconian gun control laws here. '),(5550,3186371,'08,July,2004','2004-07-08 00:00:00','   \"Laws are spider webs through which the big flies pass and the little ones get caught in.\" HonorÃ© de Balzac '),(5551,3186371,'05,July,2004','2004-07-05 00:00:00','   urlLink This  is just too cool (check out the section number). Hat-tip:  urlLink Catallarchy . '),(5552,3186371,'05,July,2004','2004-07-05 00:00:00','   From the  urlLink Adam Smith Institute Blog : \"There is a new Puritanism about. The original Puritans, you will recall, opposed bear-baiting not because it was cruel to the bears but because it gave pleasure to the people. And today, anything you might do to enjoy yourself - smoking, boozing, and now even eating - incurs a similar injunction from the politicians. Their idea of a perfect society seems to be one in which we all walk sheepishly to work and back, consume a bowl of thin gruel, and then go quietly to bed at nine o\'clock.\" This fall the government of Iceland is set to propose a ban on smoking in bars, cafes and restaurants, just like Tony Blair\'s government is trying to do in Britain. A pox on all statists! '),(5553,3186371,'04,July,2004','2004-07-04 00:00:00','   He that would make his own liberty secure, must guard even his enemy from opposition; for if he violates this duty he establishes a precedent that will reach himself.   - Thomas Paine   I prefer liberty with danger to peace with slavery.   - Author Unknown   Liberty means responsibility.  That is why most men dread it.   - George Bernard Shaw   Many politicians are in the habit of laying it down as a self-evident proposition that no people ought to be free till they are fit to use their freedom.  The maxim is worthy of the fool in the old story who resolved not to go into the water till he had learned to swim.   - Thomas Macaulay   By liberty I mean the assurance that every man shall be protected in doing what he believes is his duty against the influence of authority and majorities, custom and opinion.   - Lord Acton   Always vote for principle, though you may vote alone, and you may cherish the sweetest reflection that your vote is never lost.   - John Quincy Adams   Perhaps the fact that we have seen millions voting themselves into complete dependence on a tyrant has made our generation understand that to choose one\'s government is not necessarily to secure freedom.   - Fredrich August von Hayek   ...if we wish to preserve a free society, it is essential that we recognize that the desirability of a particular object is not sufficient justification for the use of coercion.   - Fredrich August von Hayek   We must plan for freedom, and not only for security, if for no other reason than only freedom can make security more secure.   - Karl Popper   Freedom is never free.   - Author Unknown '),(5554,3186371,'02,July,2004','2004-07-02 00:00:00','   From a  urlLink letter  written in November/December 1846 by Friedrich Engels and addressed to Karl Marx: \"...[T]he Icelander, who still continues to speak exactly the same language as the unwashed Vikings of anno 900, swills whale oil, lives in a mud hut and goes to pieces in any atmosphere that does not reek of rotten fish. I have several times felt tempted to be proud of the fact that I am at least no Dane, nor yet an Icelander, but merely a German.\" Hat-tip:  urlLink Willy Sutton  (Icelandic version). '),(5555,3186371,'01,July,2004','2004-07-01 00:00:00','   I\'ve been thinking about my last couple of posts. I hope I haven\'t given the impression that Iceland is a libertarian paradise, where children play happily with tax-free kites, unfettered by oppressive government.   We still pay 38% income tax.    We still subsidize agriculture by billions of Kronur per year, a staggering amount for such a small economy.    The manufacture of agricultural products, mainly mutton and milk, is governed by a quota system designed to keep the price of said products as high as possible.    The import of dairy, and other agricultural products is severely limited by law and custom duties.    We have a nationalized health system.    We have a nationalized education system, from the age 6 upwards.    Gun ownership for self-defence is illegal - hunting shotguns and rifles only.    Affirmative action in favour of women is the rule here.    We have a state run TV station and two state run radio stations.    We have a state run Theatre, and city-run theatres in Reykjavik and Akureyri (there might be more).    We have a state run Symphony orchestra.    We have a slew of state subsidized sports clubs, football teams and the like.    The state has a monopoly on selling alcohol, including beer and wine.    The police has powers here that would make our American friends sweat blood - just one example: Evindence acquired by illegal means is still admissable in court, the policeman in question still has to face disciplinary action.   That\'s just off the top of my head. And this is the situation after 13 years of government headed by a man who has done more to move the country in the direction of freedom than any other in decades. Imagine how things must have been around 1990!   I ask any Icelandic readers to add to the list in the comments section. '),(5556,3186371,'01,July,2004','2004-07-01 00:00:00','   urlLink This  is a couple of days old, but is still relevant. \"U.S. cinematographers and other film industry workers have asked the Bush administration to take action against Canadian, Australian and other government filmmaking subsidies that they say have lured away tens of thousands of jobs.  [...]  The groups charged the Canadian federal and provincial governments with offering a wide array of subsidies to encourage film and television production in Canada. They also accused Australia of offering \"lavish\" tax breaks and other incentives to entice movie production.  The success of those countries has encouraged many European countries, as well as Brazil,  Iceland , New Zealand and South Africa, to offer similar incentives...\" Hollywood is, of course, famously partisan in its support of the Democratic party, but I think that Bush has acquired such a strong reputation as a protectionist president that everybody, and their uncle, are now about to knock on his door, screaming for ever more protection of \"American jobs\". I also think that as the November elections loom ever closer, Bush will be more and more inclined to give them what they want.  In an interview with the MorgunblaÃ°iÃ° newspaper (where I, incidentally work) the minister for trade and industry, ValgerÃ°ur SverrisdÃ³ttir, said today:  \"I think the Americans should do something about the logs in their eyes before bothering us about any splinters that might be in ours.\" '),(5557,3186371,'01,July,2004','2004-07-01 00:00:00','   John Kerry is no idol of mine but  urlLink this  really is beyond the Pale (emphasis mine). \"A Catholic lawyer has filed  heresy charges  against Sen. John Kerry with the Archdiocese of Boston, accusing the Democratic presidential candidate of bringing \"most serious scandal to the American public\" by receiving Holy Communion as a pro-choice Catholic.\"  Hat-tip:  urlLink Andrew Sullivan . '),(5558,3186371,'01,July,2004','2004-07-01 00:00:00','   The Bank of Iceland raised interest rates yesterday by 0.5% and they now stand at 6.25%. This is the third rate increase since early May when the rates stood at 5.3%. The hike itself did not come as a surprise to analysts, but in general they were expecting a smaller increase than half a percentage point. The Bank of Iceland is expected to raise interest rates again in the immediate future, as inflation is reaching the upper limits set by the bank.    Key interest rates since 1994 (courtesy of the  urlLink Landsbankinn bank ).     Inflation  Consumer Price Index for the last 12 months. The solid yellow line marks the target inflation rate set by the Bank of Iceland, while the dotted lines mark the upper and lower limits set by the bank (courtesy of the  urlLink Ãslandsbanki bank ). '),(5559,3186371,'01,July,2004','2004-07-01 00:00:00','   William Sutton, the mysterious Icelander, kindly pointed me to an  urlLink old post of his . It\'s a must read; abso-bloody-lutely. '),(5560,3186371,'01,July,2004','2004-07-01 00:00:00','   This is a poster made by the guys over at  urlLink catallarchy .  At the moment I am linking straight to the image on their server but when I get home I will move the pic onto my server so as not to leech to much bandwidth.  (Done) - The poster is a spoof of this one.    Update  Took out the link to jiffylux\'s image. Will host it on my server. Also took out the image made by catallarchy as there is another one out there which doesn\'t infringe on any copyrights, as I am informed that this one does. All this will be remedied later today (July 22. 2004). '),(5561,3186371,'01,July,2004','2004-07-01 00:00:00','   Courtesy of  urlLink Vox Day . \"For the social problems: We demand:  a) The quick enactment of a law of the State that sanctions an eight-hour workday for all workers.  b) A minimum wage.  c) The participation of workers\' representatives in the functions of industry commissions.  d) To show the same confidence in the labor unions (that prove to be technically and morally worthy) as is given to industry executives or public servants.  e) The rapid and complete systemization of the railways and of all the transport industries.  f) A necessary modification of the insurance laws to invalidate the minimum retirement age; we propose to lower it from 65 to 55 years of age.\" From the 1919 manifesto of the Italian Fascist party, written by Benito Mussolini. Sounds familiar don\'t it? '),(5562,3767607,'29,June,2004','2004-06-29 00:00:00','   Blogging is new thing to me.  Just thought I\'d take it for a test dive.  '),(5563,3767607,'04,July,2004','2004-07-04 00:00:00','    urlLink http://www.raycomm.com/techwhirl/  TECHWR-L (\"tech-whirl\") is the Internet-based community and resource for technical communicators worldwide. The TECHWR-L discussion list has more than 5,000 direct subscribers and an estimated daily readership of more than 10,000 people worldwide. The TECHWR-L site serves more than two million page views per month and provides a unique combination of articles, features, and services unrivaled by any other Internet-based technical writing resource.    urlLink http://www.useit.com/  Jakob Nielsen\'s (need I say more?) website on \"usable information technology\" .   urlLink http://www.w3.org/  The World Wide Web consortium website. The World Wide Web Consortium (W3C) develops web specifications, guidelines, software, and tools and is a forum for information, commerce, communication, and collective understanding.   '),(5564,3767607,'03,July,2004','2004-07-03 00:00:00','    Research and Writing    urlLink http://www.visualthesaurus.com/index.jsp  Visual Thesarus is a way cool interactive thesaurus. You type in a term, click, and your term is instantly surrounded on-screen by a web of synonyms and antonyms. When you click on one of those surrounding words, it becomes the center of the page and a whole new set of related words pops up. It\'s very intuitive, and a lot more fun than a book thesaurus!     urlLink http://www.landmark-project.com/citation_machine/index.php  This is a solid (but not perfect) citation maker. Just select the type of source document, fill in the template, and Citation Machine creates an MLA and APA citation.  http:// urlLink www.libraryspot.com/  Library Spot is a great starting point for research. Here\'s their tag-line: Simplifying the Search for the Best Library and Reference Resources on the Web. Tons of links to dictionaries, thesauri, reference libraries, etc.    Medical stuff   urlLink http://www.ncbi.nlm.nih.gov/entrez/query.fcgi  If you need to search for real medical research, PubMed is the place to go. Part of the National Library of Medicine, PubMed has millions of citations for biomedical articles from MEDLINE and additional life science journals. It\'s easy to search, and includes links to many sites providing full text articles and other related resources.   '),(5565,3767607,'02,July,2004','2004-07-02 00:00:00','    The I Need to Laugh  group o\' links  You know how some days you just need a break?  Well, here are some of my faves.   urlLink www.davebarry.com  Need I say more?     urlLink http://www.schickele.com/  Ever heard of PDQ Bach?  If you are into classical music and like to laugh a little, check this out.  Professor Peter Schickele is the brain behind the whole concept of PDQ.     urlLink http://www.homestarrunner.com  This is totally inane--but it is about the most hilarious thing on a bad day or when you are extremely sleep deprived.     The  I Need Computer Help  group o\' links  I\'m a novice at the whole computer maintenance thing, but these sites offer a little guidance.   urlLink http://www.Helpwithpcs.com    urlLink http://www.pchell.com      The  I Need to Learn More about Technical Communication  group o\' links  Realistically, this isn\'t one group that most people would be interested in, but I think it\'s the bomb!   Society for Technical Communication   urlLink http://Stc.org    Society for Technical Communication Technical Editing Special Interest Group  urlLink http://www.stcsig.org/te/  (Because my favorite sub-category of technical communication is technical editing)         Technical communication  journal   urlLink www.techcomm-online.org    EServer TC Library: A Cooperative Library for Tech Communicators   urlLink tc.eserver.org  Another great resource with lots of technical editing links   The \"no formal title\" link   urlLink http://www.december.com/info/techcomm  I just found this one, but it has some good stuff on it!   '),(5566,3767607,'02,July,2004','2004-07-02 00:00:00','    Here\'s a list of links of my favorite Tech Writing, Research, and Diversion sites.      Technical Writing    urlLink Council for Programs in Technical and Scientific Communication  I have to admit I don\'t remember ever using this site, but it was on my favorites list and it looks terribly important.    urlLink The Bay Area\'s Editors\' Forum  I have used this one. The forum offers a number of useful resources for editors of the technical or non-technical variety.   urlLink eServer TC Library  Perhaps this site would be better listed under the research header, but I\'m running low on tech writing sites and the TC Library is an immensely useful site, so use it and be happy.   Research   urlLink Bartleby  This site is so cool that I\'m afraid to let the word out.  Unlimited access to tons (I weighed them) of reference, verse, fiction, and nonfiction texts.   urlLink Baywood  Fee-based service offers articles from journals you can\'t get anywhere else.   urlLink Voice of Shuttle  As deep as it is wide, VOS offers a dive into the murky depths of humanities research.     Diversions   urlLink Bloodshoot records  Treat yourself to the latest from the Insurgent Country scene.  Be the first on your block to discover Trailer Bride, Neko Case, and The Waco Brothers.    urlLink Mundane Behavior  Don\'t just live a commonplace life, read scholarly studies about it.   urlLink The Unemployed Philosophers Guild  Fuel up for your next discussion on postmodern angst with a Nietzche \"Will to Power Bar\" or convince your analyst that you really are ready to start healing by giving him/her an Anna Freud doll.   '),(5567,3767607,'02,July,2004','2004-07-02 00:00:00','    XSL (eXtensible Stylesheet Language)   XSL is a new and exciting league of professional American football sweeping the nation.  Take a American football and add a twist of Aussie rules, throw in liberal amounts of lethal explosives, a healthy dose of heavy metal music, scantily clothed women, and lots of alcohol and you have . . . okay, so thatâ€™s the WWF.  XSL has something to do with websites.        Definition provided by Shannon Brown.   What is it?   Extensible Stylesheet Language, or XSL, is a special kind of Extensible Markup Language, or XML. As its name suggests, XSL was drafted initially to give developers specific language tools to augment XML functionality. Neil Bradley, in The XSL Companion, gives the following concise history of XSL to the year 2000:  â€œ[XSL] is a family of standards developed by W3C (the World Wide Web Consortium). These standards emerged out of a proposal for a stylesheet language, submitted in 1997, which was to be called \'XSL\' (eXtensible Stylesheet Language). However, during its gestation, this proposal was eventually pulled apart into three separate standards. The first of these, XPath, defines a mechanism for locating information in XML documents, and it has many other uses beyond its role in formatting documents. The second, XSLT, provides a means for transforming XML documents into other data formats, including (but not limited to) formatting languages. Finally, the term \' XSL \' is now properly used only to name a proposed standard for embedding formatting information in documents using XML elements.â€ (Bradley 2000, iii)  In other words, almost at its inception XSL split into three standards, each with different purposes: Xpath, XSLT, and a third standard that Bradley sees as the proper heir to the larger name â€œXSL.â€ Nonetheless, W3C, the body responsible for overseeing the development of XSL, still treats the XSL family as a single entity in its official definition, with three distinct parts:  â€œXSL is a family of recommendations for defining XML document transformation and presentation. It consists of three parts:  XSL Transformations (XSLT), a language for transforming XML, the XML Path Language (XPath), an expression language used by XSLT to access or refer to parts of an XML document. (XPath is also used by the XML Linking specification), [and] XSL Formatting Objects (XSL-FO) an XML vocabulary for specifying formatting semantics.â€(World Wide Web Consortium 2004, w3.org)  Pairing off the terms from Bradleyâ€™s discussion of XSL to the terms in the W3C definition, weâ€™re left to conclude that the third standard Bradley mentions simply as â€œXSLâ€ is actually â€œXSL Formatting Objects.â€ We should note that there is actually some ambiguity to current usage of the term â€˜XSLâ€™â€”some authors use it strictly to refer only to the entire family, while others appear to use it interchangably with XSL-FO. Nonetheless, given the differences between the three parts of the XSL family, we should take care to keep them straight. So, how do these different parts of XSL work with XML?   Transformation and Presentation   The W3C definitionâ€™s reference to â€œtransformation and presentationâ€ reveals key characteristics of two of the sub-lanaguages in XSL. Elliotte Rusty Harold describes the transformative power of XSL in The XML Bible:  â€œThe Extensible Stylesheet Language (XSL) includes both a transformation language and a formatting language . . . The transformation language provides elements that define rules for how one XML document is transformed into another XML document. The transformed XML document may use the markup and DTD [document-type definition] of the original document, or it may use a completely different set of elements. In particular, it may use the elements defined by the second part of XSL, the formatting objects.â€(Harold 2001, 481)  Hence, in addition to specifying the formatting of the resulting XML content, XSL can also provide a map for turning one XML document into a second, different XML document. XSLT is the language for performing this kind of change. But why would anyone want to do this?  XSLT: An example  Miloslav Nic gives several applications in his XSLT tutorial, available online at zvon.org, which I used as the basis for the attached examples.   Whatâ€™s left in XSL?   G. Ken Holman, in the article What is XLST? he wrote for xml.com, describes the difference between XLST and the rest of XSL:  â€œ[There are] two distinct styling steps: transforming the instance of the XML vocabulary into a new instance according to a vocabulary of rendering semantics; and formatting the instance of the rendering vocabulary in the user agent . . . In order to meet these two distinct processes in a detached (yet related) fashion, the W3C Working Group responsible for the Extensible Stylesheet Language (XSL) split the original drafts of their work into two separate Recommendations: one for transforming information and the other for rendering information.â€ (Holman 2000, xml.com)  Weâ€™ve already discussed the confusion in some of the literature over the entire XSL family and its constituent parts, especially XSL Formatting Objects. Given this confusion, we should try to decipher Holmanâ€™s reference to â€œXSLâ€ as one of the three parts named in the W3C definition, most likely XSL-FO. In The XML Bible (again), Harold describes XSL-FO as sharing some transformative characteristics with CSS. XSL-FO, however, is much more powerful and dynamic.  â€œWhile CSS (Cascading Style Sheets) is primarily intended for use on the Web, XSL-FO is designed for broader use. You should, for instance, be able to write an XSL style sheet that uses formatting objects to lay out an entire printed book. A different style sheet should be able to transform the same XML document into a Web site.â€ (Harold 2001, 571)  XSL-FO, then, is portable to many different formats, which makes it ideal for content managers working in many different media. Of course, this flexibility means that XSL-FO must use its own language for attribute and element commands. Miroslav Nic offers a searchable reference of XSL-FO commands, with plenty of examples, at zvon.org.   And, finally: XPath   Elliotte Rusty Harold and W. Scott Means, in XML in a Nutshell, explain XPath in terms of what it does for XSLT:  â€œXPath is a non-XML language for identifying particular parts of XML documents. XPath lets you write expressions that refer to the first person element in a document, the seventh child element of the third person element, the ID attribute of the first person element whose contents are the string â€œFred Jonesâ€, all xml-stylesheet processing instructions in the documentâ€™s prolog, and so forth. XPath indicates nodes by position, relative position, type, content, and several other criteria. XSLT uses XPath expressions to match and select particular elements in the input document for copying into the output document or further processing . . . String manipulation in XPath lets XSLT perform tasks such as making the title of a chapter uppercase in a headline or extracting the last two digits from a year.â€ (Harold and Means 2002, 154)  So, XPath is the tool used by XSLT for finding and deploying parts of one XML document in another XML document. We saw it at work in the XSLT example without pointing it outâ€”XPath is the standard from the XSL family responsible for pulling the â€œtitleâ€ and â€œauthorâ€ elements from the XML source.   Sources   Bradley, Neil. The XML Companion. London: Addison Wesley, 2000. Portions available online at O\'Reilly.com  Nic, Miroslav. XSLT Tutorial,1999.  â€“â€“â€“â€“â€“. XSL FO reference, 2000.  Harold, Elliotte Rusty. The XML Bible, 2nd edition. New York: Hungry Minds, Inc., 2001. Portions available online at ibiblio.org  Harold, Elliotte Rusty and W. Scott Means. XML in a Nutshell, 2nd edition. Sebastopol, CA: Oâ€™Reilly, 2002. Portions available online at O\'Reilly.  Holman, G. Ken. â€œWhat is XSLT?,â€ Oâ€™Reilly xml.com , 2000.  World Wide Web Consortium. The Extensible Stylesheet Language Family (XSL). w3.com, 2004.   '),(5568,3767607,'02,July,2004','2004-07-02 00:00:00','    WORKFLOW   Iâ€™ll give you three guesses on this one.  How about â€œWorkflow is the flow of workâ€?  Fork please, my job here is done.     Definition provided by Gary Hernandez     Workflow   Last week I read an article from ASTDâ€™s magazine, T&D, which addressed workflow in terms of learning while working.  The article spoke to the efficiencies of employees learning while engrossed in on-the-job activities.  The efficiencies are realized when employees do not have to remove themselves from work to attend learning events about work.  Hence productivity improves and, perhaps, the knowledge is further embedded through kinesthetic means.  When I saw workflow in the list of terms for us to define, I jumped right on it thinking I was at least on article ahead of the game.  Alas, much like granularity, semantics, and life cycle, workflow is yet another term that has a differential meaning in the context of content management than it does in other fields.      Rockley on workflow   â€œWorkflow, as its name implies, is the way tasks flow through a cycle on their way to getting a job doneâ€ (Rockley 228).  According to Rockley, workflow in its simplest form articulates the path from the first task of a work plan to each successive milestone and task through to completion.  Workflow could be illustrated in a traditional flowchart or swimlane diagram.  One would presume a PERT chart and Gantt chart could be used as well. The importance is that players, tasks, and process are clearly delineated.   The players are people and roles that complete each task, the tasks are discreet units of work, and processes indicate the flow and interdependencies of the tasks.  Workflow can also be automated.     â€œWorkflow 101,â€ Jim Miniham  Jim Miniham states that workflow is not a definitive term.  Miniham points out that the term originated in the manufacturing industry where it had organizational implications.  In the manufacturing world workflow referred to the flow of a product through its entire production life span: how long the product spent in each stage of production, how many items moved through a given work location, and even where specific products were at any given time.  In the software industry, workflow came to mean the â€œrouting and distributionâ€ of documents as well as everything that happened to it along the way, including who touched it, how long it spend in what location, and its status at any given time.  With such a wide span of functionality, the word workflow began to be applied to any number of applications that approximated any stage production.  In addition, workflow began to be applied with production-class tools as well.  In this realm workflow expanded its influence from tracking products to being critical in managing the production process.     â€œThe Wild, Wild World of Workflow,â€ Molly W. Joss  Molly Joss, similar to Jim Miniham, indicates that workflow is terminally ubiquitous and can stand to mean any number of things depending on the context.  Postmoderm implications aside, workflow principally means the configuration of work processes.  From this standpoint, all organizations engaged in work are in engaged in workflow.  How one manages workflow is perhaps what people are really talking about when they mention the term.  To help in that management process, of course, is quite a plethora of computer applications.  Joss cites three types of workflow tools available in the graphic arts industry:  Â·         Task-oriented workflow tools which â€œautomate or simplify a specific prepress taskâ€ (62).  Â·         RIP-centric workflow systems for a full system that address the entire workflow paradigm  Â·         Workflow analysis tools to help â€œcapture and calculate data on various prepress processesâ€ (67).    Joss concludes that the ability of workflow applications to improve productivity has given rise to a world of workflow vendors, giving would be organizations wishing to optimize workflow through automation the blessing of choice and curse of choice.    â€œWhy Workflow Works,â€ Robert B. Segal  To the relatively flat discussion of workflow, Robert Segal adds a refreshing twist.  The benefit of automating workflow is, yes, improved productivity through efficient use of resources, but the beneficiary of that productivity is the customer.  Segal notes that â€œWorkflow automation systems are designed to ensure that information is processed efficiently, all activities are monitored and measured and expectations are reportedâ€ (94).  The key for Segal in workflow is the monitoring and reporting functionality that gives the line manager the ability to respond to and correct workflow deficiencies that often directly effect the customers.   With this in mind, the marriage of workflow management with customer insight data produces Customer Relationship Management (CRM).  Segal sees workflow management as the foundation to CRM.           â€œWorkflow,â€ Tom Singer  Tom Singer quotes Workflow Management Coalition as â€œthe automation of a business process, in whole or part, during which documents, information, or tasks are passed from one participant to another for action, according to a set of procedural rulesâ€ (34).  Singer notes that workflow optimization is not only about managing productivity and maximizing efficiencies, it can also help manage maintenance issues.  The point is that maintenance is also driven by workflow processes and hence can reap doublyâ€”1) by improving efficiencies and 2) as maintenance becomes more efficient, the benefits of maintenance can be brought to bear more quickly on the organization.      Conclusion  Workflow principally focuses on maximizing efficiency in the processes embedded in a work plan.  Though originating in manufacturing, it can be applied to any environment which has work processes but perhaps reap more benefit in an automated work stream.  The benefits of workflow management can be felt in productivity, customer service, and even in some collateral areas such as maintenance and training.      WORKS CITED   Adkins, Sam S.  â€œThe Brave New World of Learning,â€ T&D.  June 2003.    Joss, Molly W. â€œThe Wild, Wild World of Workflow,â€ www.electonic-publishing.com.    September 1999.   Miniham, Jim.  â€œWorkflow 101,â€ www.edocmagazine.com. July/August 2003.    Segal, Robert B. â€œWhy Workflow Works,â€ Mortgage Banking. June 1999.    Singer, Tom.  â€œWorkflow,â€ www.plantengineering.com.  September 2002.   '),(5569,3767607,'02,July,2004','2004-07-02 00:00:00','    WIKI  WIKI is not a form of wicker furniture found at Ikea, it has nothing to do with outdoors Hawaiian barbecues, and it is not that threatening dance the New Zealand Rugby team does at the beginning of their games.  To discover exactly what this term is, read below and be fascinated.    Definition provided by Shannon Brown.     What is wiki?   The idea of Wiki, as explained by the wiki.org site, is deceptively simple:  â€œWiki is a piece of server software that allows users to freely create and edit Web page content using any Web browser.â€ 1  Of course, â€œwikiâ€ has come to refer to much more than this server softwareâ€”it has become identified with the activity this server software makes possible.  Ward Cunningham created and christened the first â€œwikiâ€ in 1995, naming it â€œwiki-wiki,â€ which means â€œsuper fastâ€ in Hawaiian, in order to avoid the blander â€œquick-web.â€ 2 The wiki was created so that anyone, regardless of his or her previous experience with computer languages, could contribute to the web publishing community. More importantly, though, the wiki software and wikis themselves were created to be used by everyone, all together, as described in the entry for wiki in the Wikipedia:  â€œAWikiWikiWeb enables documents to be authored collectively in a simple markup language using a web browser . Because most wikis are web-based, the term \'wiki\' is usually sufficient. A single page in a wiki is referred to as a \'wiki page,\' while the entire body of pages, which are usually highly interconnected, is called \'the wiki.\'â€ 3  So, a wiki page, in theory at least, has no single author. Anyone can contribute, everyone on equal footing. The word usage distinction between single pages (wiki pages) and an entire group of pages (the wiki) is significant: the software is designed to support an entire body of interconnected pages, of which individual pages are only parts.   What makes wiki different?   Probably the most unique aspect of wiki, however, is its emphasis on collective authorship. Although several web-publishing methods offer interconnected communities of easy-to-author pages of content, only the wiki opens up the authoring privilege for every page to every visitor. The ISP Webopedia notes wikiâ€™s crucial difference from blogs in this respect:  â€œSimilar to a blog in structure and logic, a wiki allows anyone to edit, delete or modify content that has been placed on the Web site using a browser interface, including the work of previous authors. In contrast, a blog, typically authored by an individual, does not allow visitors to change the original posted material, only add comments to the original content.â€ 4  Likewise, although the wikiâ€™s approach to community ownership might appear to make it a kind of message board, the epowiki site, in a page with text attributed to Randy Karamer, explains the differences between the two publishing methods:  â€œCompared to a plain vanilla message board, a Wiki\'s advantage is in DocumentMode. In a message board, somebody writes, somebody answers, somebody else answers, somebody asks a follow up question, ..... You end up with a whole lot of messages but not an integrated story. On a wiki, somebody can start a wiki page by asking a question. Someone can come along and edit that page to answer the question. Somebody else can come along and edit the answer to cover an additional point, or clarify a point, or whatever. If somebody edits the page by adding a follow-up question, it can be left on that page and then answered. Maybe it makes more sense to move it to its own page.â€ 5  Comparing a wiki to a physical message board reveals how radical the wikiâ€™s approach to authorship really is. Imagine that each person who wants to can not only add new messages to the board, he or she can edit, reorganize, even merge and delete messages that are already there.  Of course, since anyone can edit, reorganize, merge, or delete wiki pages, a wiki is vulnerable to the whims of users with less-than-honorable intents. Vandals can lay waste to months of collaborative work in a matter of hours. Ward Cunningham, the creator of the first wiki, describes this as the wikiâ€™s inherent â€œfragility.â€ 6 The wiki, however, can fight back, as The Humaine Portal (itself a wiki) explains:  â€œVandalism is prevented through an uneditable document history -- if anyone vandalizes the current version of a document, it is easy to restore a previous version and ban that user.â€ 7  This unalterable change history, since it is normally accessible to any visitor, allows law-abiding contributors to a wiki to police their community. Violators can be permanently exiled for their transgressions.   Open, semi-open, private   Naturally, not all wikis are the same. While there are many large wikis that still hold to the original wiki precepts of collective, egalitarian authorship, other communities have leveraged the wiki idea without completely agreeing to the wiki ideals. The Humaine Portal explains a fundamental difference in wiki culture:   â€œAt the core of the wiki principle is the idea that the obstacle for contributing to a document should be lowered. In open wikis, anyone can edit a web page. In semi-open wikis, as the HUMAINE know-how melting pot, users must register before they can contribute.â€ 8  The Wikipedia is perhaps the largest open wiki in existence, with nearly 300,000 articles/entries. 9 On the other end of the spectrum is a category the above quote from the The Humaine Portal wiki ignores: the private wiki, housed on private servers. Obviously, since all thatâ€™s needed to start a wiki is a server, a piece of database server software, and a network of users, the wiki technology holds promise for many different collaborative teams, not just the public at large. Several companies are already offering their wiki services to private groups or individuals, including Altassian and Project Forum among many, many others.   Sources   1 â€œWhat is Wiki?,â€ wiki.org, 2004.  2 â€œWiki History,â€ Portland Pattern Repository Wiki, 2004.  3 â€œWiki,â€ Wikipedia, the Free Encyclopedia, 2004.   4 â€œWiki,â€ ISP Webopedia, 2004.   5 â€œWhat is wiki,â€ epowiki, 2004.   6 Rupley, Sebastian. â€œWhatâ€™s a Wiki?,â€ PC Magazine (May 9, 2003). Available online at pcmag.com.  7 â€œWhat is wiki,â€ The Humaine Portal, 2004.   8 Ibid.  9 â€œMain page,â€ Wikipedia, the Free Encyclopedia, 2004. (as viewed on June 30, 2004).   '),(5570,3767607,'02,July,2004','2004-07-02 00:00:00','    VERSIONING   Okay, isnâ€™t there a grammar rule about not turning nouns into verbs?  This word is clearly out of line, and I am shocked to see it in a technical writing list of terms!  Out of protest I will not provide an intro to this term and, instead, insist that it be struck from the tomes of technical communication forever and anon, amen.   Definition provided by Tim Carter.   What is Versioning?  Versioning, or version control applies to a content management system or strategy. I have found that it is identified in two distinct working environments: technical writing and software development. It is a system or strategy that tracks the revisions made to a document or software code in an environment with multiple writers or programmers. The system keeps control of the document or program so that no two people are working on the same version at the same time, and it can track when documents were â€œchecked outâ€ and when they were â€œchecked inâ€ (source control). In this way, control is maintained, the original documentation is never lost, never incorrectly duplicated, or unnecessary changes made to a down-level document.    How Can I Use It?  As defined by Anne Rockleyâ€™s Managing Enterprise Content, versioning says that each time content is saved, a new version will be created. In the sample unified content life cycles (pg 149), there would be two levels of versioning: draft version and full version. Content would remain in draft version until approved, then it would be fully versioned. Through version control, authors would be able to view previous versions of the content or revert to an earlier version if necessary. Authors and reviewers can see the complete detailed history for each item. The history will indicate who made the change, when the change was made, and the reason for the change. Similarities exist between programmer and writer versioning in that both groups need to maintain control over their documentation to eliminate or reduce lost time and money on extensive projects.    Where Are All These Versions Stored?  According to David Thomas and Andrew Hunt in their book, Pragmatic Version Control at http://www.pragmaticprogrammer.com/starter_kit/vc/whatis.pdf, all versions of documentation in most version control systems are stored in a repository. The repository is a central place that holds the master copy of all versions of your projectâ€™s files. Some systems use a database, some use regular files and still others use a combination of the two. In any case, the repository is the pivotal component in the versioning strategy.   This repository doesnâ€™t simply store the current copy of each of your files, it stores every version of every file that has ever been checked in. If you check out a file, edit it and check it back in, the repository will hold both the original version and the version that contains your changes. Then the versions will be assigned a numbering system. For example, the first version of a file is assigned the revision number 1.1. If a changed version is checked in, that change is given number 1.2. The next changes gets 1.3 and so on.    More Contrasts and Similarities  The TechEncyclopedia at www.techweb.com/encyclopedia narrowly defines versioning (or version control this way: The management of source code, documents, graphics and related files in a large software project. Version-control software provides a database that is used to keep track of the revisions made to a program by all the programmers and developers involved in it. This narrow definition, applied to XML language, is echoed in the Versioning SML Vocabularies by David Orchard at http://www.xml.com/pub/a/2003/12/03/versioning.html. Orchard describes versioning as, â€œadding, deleting, or changing parts of the language.â€  Another similar versioning concept comes from John Gruber in his article, â€œBBEdit Version Controlâ€ at http://www.macworld.com/2003/09/secrets/bbeditversioncontrol. With the CVS (Concurrent Versions System) software, writers (or programmers) can record all changes to a document and you can quickly step back to an earlier version â€“ even if youâ€™ve changed and saved the file many times â€¦ a personal time machine. The CVS is useful for managing any type of document. In fact, a Web site is a perfect example of the sort of multifile project that can benefit from versioning. The difference with this concept is that you can make a mistake, then easily track down and reverse the mistake using an earlier version.  Finally, the application FrameMaker provides another concept for versioning in which two distinct versions of text for two separate products. As you will see, this definition is in stark contrast to earlier definitions. Instead of one product and multiple versions, here we have two products with simultaneously different versions. In â€œDesigning Print Documentationâ€ by David A. McMurrey, Thomas A. Moore and Pam Renwick at http://www.io.com/~tcm/etwr2372/planners/frame/frame556/frame_variables_ver556.html#versioncontrol they say versioning is the ability to make one set of files support two more versions. Their version control â€œenables you to keep everything in one set of files and hide or show a particular version. This is useful if you must maintain documentation for two versions of the same product. This versioning is available only through FrameMaker and its native documentation.   '),(5571,3767607,'02,July,2004','2004-07-02 00:00:00','    VERSION CONTROL   Not be confused with â€œrevisionist history,â€ a popular and ubiquitous form of propaganda used to shape public and political opinion.  Version control is used to drive consistency in the development and revision of texts.    Definition provided by Kathylynn St. Pierre   Dysfunctional Families and Version Control  Kathylynn St. Pierre English 6430 â€“ Summer 2004   Oh many, many years ago When I was twenty-three I was married to a widow Who was pretty as can be This widow had a grown-up daughter Who had hair of red My father fell in love with her And soon the two were wed   This made my dad my son-in-law And changed my very life For my daughter was my mother â€˜Cause she was my fatherâ€™s wife To complicate the matter Though it really brought me joy I soon became the father Of a bouncing baby boy   This little baby then became A brother-In-law to Dad And so became my uncle Though it made me very sad For if he was my uncle Then that also made him brother Of the widowâ€™s grown up daughter Who of course Is my step-mother   My fatherâ€™s wife then had a son Who kept him on the run And he became my grandchild For he was my daughterâ€™s son. My wife is now my motherâ€™s mother And it makes me blue Because although she is my wife Sheâ€™s my grandmother too.   Now if my wife is my grandmother, Then Iâ€™m her grandchild And every time I think of it It nearly drives me wild For now I have become The strangest case you ever saw As husband of my grandma I am my own Granpa.   From â€œIâ€™m My Own Grandpaâ€ by Dwight Latham and Moe Jaffe     The term â€œversion controlâ€ applies to everything from sophisticated technology-driven systems that manage changes in data files (source code for software programs, CAD files, etc.) to the simplest form of file management: the ability to check differences from a previous version, and to provide a layer of assurance that one personâ€™s work isnâ€™t overwritten by someone elseâ€™s.    Some sophisticated engineering systems have file management/control and version control systems built in, others require that separate tools be used. But according to Paul D. Sherriff of Klarity.com,  â€œsoftware change management is not just for development, or QA or the operations and release group. Good documentation and change process must be designed to support the whole organization; it must help everyone to better understand their products and their jobsâ€.   That makes perfect sense. Everyone needs to pay attention to version control issues, and clearly version control doesnâ€™t just apply to one department, group or industry.  The Pragmatic Programmer website says â€œTools to automate change management and control (SCM tools) can only work when implemented together with corporate standards and a well-defined process.â€ While Iâ€™m not sure that the standards are necessary (or realistic) at the corporate level, it is clear that a process must be clearly defined and integrated with every aspect of the documentation process - from file management to quality requirements to approaches to working with outsourced content.    The definitions here apply to publication and documentation departments, where version control often simply means the daily challenge of managing hundreds of files, sometimes with many writers working on them at the same time.     It is clear that there are two ways version control is applied on a daily basis: automated change control systems, and manual systems developed within documentation departments to satisfy their unique needs and circumstances.        Automated change control systems    Â§	Many systems, (Wiki is a popular one) use a database back-end, which records each modification in a log. Scripts within the system access the log and follow the revision history of a file.  The log is really designed for recovery from abuse or accidents that cause widespread damage to the database, and not for retrieving the individual modifications to one file inside that database. In most publication departments, Wiki-type systems are overkill for day-to-day file management.   Â§	Many change control systems are marketed on the â€œtime-machineâ€ premise - you can look at any version of any page and then follow links to check out the entire file as it was at that point in time. Engineers like this because they sometimes follow a design path so far only to find out that their design wonâ€™t work â€“ this way they can â€œgo backâ€ to the point in their development before they veered off onto the â€œwrongâ€ path. Some of these keep a long sequence of revisions available and include a diff feature. Engineers like to see the complete evolution of a file that lets them start alternate evolutions at any point.   Â§	Software Configuration Management (SCM) refers to any collection of processes and tools that are used to effectively manage the development, maintenance, and build processes of software.     Â§       With Concurrent Versions System (CVS), multiple writers can check out the same files and then check them back in. When working in ASCII modes, such as HTML, this works fine as it calculates the differences from the last version and applies them. Depending on your product version and platform, it figures this out quite well. But this method requires excellent communication between team members. If there is a conflict, it stops the writers from checking files back in, and forces them to manually reconcile the differences.   Â§	Professional Version Control Software (PVCS) â€“ PVCS is actually a process management tool, but some technical writing projects are forced to use it because the engineering departments of the projects they document use it.     Michael Krasowski, of The Code Component Developer website says, â€œWithout a well understood process, no tool will solve the problem.â€ With this in mind, it is important to mention that adapting your process to meet an automated version control system is starting the whole process backwards â€“ you need to establish a process that works with your departmentâ€™s situation, and then find a system that meets those needs.    Manual solutions that solve day to day issues     At the risk of making an unfounded generalization, Iâ€™ll admit to thinking that most publication departments find individual and creative ways of dealing with their unique version control issues.   The size of the information product, the size of the team, frequency of releases and publication dates, all play a role in how a publication department may approach version control. But it is clear that a lack of formal guidelines poses a serious problem for publication departments. Even the simplest methods for \"getting\" versus \"checking out\" documents create huge issues that most always must be addressed even before the first page is written.   But publication departments canâ€™t rely on their chosen publishing software to take of their needs. Some publishing programs allow writers to \"get\" a copy of a file in read-only mode and then make it write-able. This overrides the security \"check out\" feature of the system that of course, proves counter-productive later and creates even more confusion.  Where I work, our documentation team was at one time very large (before company cutbacks and layoffs occurred). We write product and system documentation for a large software product, which usually involves at least two or three major releases per year.    Like the publications departments I mentioned earlier, we devised two version control systems that meet our needs.  These are probably the simplest forms of version control and involve processes similar to what I believe other publication departments like ours probably use to meet their needs.   The first process we use involves two major activities: Checking Out and Checking In.   Check Out and Check In  A copy of all files is put on a local shared server, drive, or machine (what The Pragmatic Programmers call a â€œrepositoryâ€) and to work on files, writers must check them out (manually, through a documentation â€œlibrarianâ€, through a manual registry that the writer must fill out each time, or by the writer taking the file out and then â€œlockingâ€ it so that only the writer that checked out the file can check it back in).    This process requires strict adherence to the process itself, but at the same time protects the team from itself - it alerts the rest of the team when someone is working on that file. The last saved version overwrites the previous version, so these alerts are absolutely necessary.     Youâ€™re your own Grandpa    On some releases, all files are divided and weâ€™re each responsible for specific files that we â€œownâ€. Only the writer assigned to a certain file works on that file. We each keep our own copies and deliver them to a special delivery directory when it is time to do documentation build (usually once a week). Then, the build directory becomes our official directory and we are each responsible for our own version control activities. The only time this becomes tricky is when we are working on two or three releases at the same time. In this case, we have to keep three different copies of the same file and make changes that may apply to one release but not another. When we practice this method, Brian Harroff, our project manager, continuously reminds us â€œYouâ€™re your own Grandpaâ€.      So much to consider    The version control business (and it is a business in and of itself) has many facets, and Iâ€™ve explored it very broadly here and have simplified it for purposes of creating a high level definition. But the bottom line to version control, whether you are using an expensive change management system or a manual system your department has devised to meet your needs, is to remember that with a well-defined process and strict adherence to the policies defined in them, being your own Grandpa isnâ€™t all that dysfunctional after all.     References      Â¾Â¾Â¾. 2004. â€œ.Automation Through Process Definitionâ€ AFORA, International. Available http://www.scmbyafora.com .   Â¾Â¾Â¾. 2004. â€œVersion Control Applied to Wiki.â€ Cunningham & Cunningham, Inc. Available http://c2.com .   Grant, Samantha. 2002. â€œChange Management and Software Developmentâ€ E-Commerce Times.   Krasowski, Michael, 2004. â€œSoftware Configuration and Management Using Visual SourceSafe nd VS. NETâ€. Available http://www.devx.com.   The Pragmatic Programmers, LLC, 2004 â€œPragmatic Version Control with CVS.â€ Available http://www.whatis.com .   Sherriff, Paul D. 2004. â€œApproaches to Version Control Software.â€ Klarity. Available http://www.klariti.com .   '),(5572,3767607,'02,July,2004','2004-07-02 00:00:00','    My college roommate got arrested for that once. What a drag, man. That open container law is a bummer. Oh, Open CONTENT, you say? Well, hmmm. I donâ€™t know what the heck that is. Letâ€™s find out in Steve Crowâ€™s lively discussion (and you wonâ€™t even have to hide that drink):   The Basics  The term â€œopen contentâ€ is aptly named; it means pretty much what is sounds like it ought to mean. In 1998, David Wiley of Opencontent.org provided this basic definition, â€œOpen Content is a new and evolving term, but in the strictest sense it refers to content (books, articles, documentation, images, databases, etc.) that is \"freely available for modification, use, and redistribution under a license similar to those used by the Open Source / Free Software community.\"â€  Free Doesnâ€™t Always Mean Free But â€œfreeâ€ doesnâ€™t always mean â€œfree to change it, resell it, etc.â€ The online dictionary Word IQ : â€œOpen content, coined by analogy with open source, describes any kind of creative work (for example, articles, pictures, audio, video, etc.) that is published under a copyright license, or in the public domain, in a format that explicitly allows the copying of the information. One example is the GNU Free Documentation License, which is used by Wikipedia and Nupedia. \"Open content\" is also sometimes used to describe content that can be modified by anyone. Of course, this is not without prior review by other participating parties--but there is no closed group like a commercial encyclopedia publisher which is responsible for all the editing. But not every open content is free in the GNU GPL sense (for instance the Open Directory). Some licenses attempt to maximize the freedom of all potential recipients in the future, while others maximize the freedom of the initial recipient.â€   Free Content vs. (truly) Open Content The Life OpenContent site gives a little more insight into the distinction between Free content and Open content:  Free (or copyable) Content:   â€œThis content is â€œopenâ€ only in the sense that it is accessible. The content may be copied freely, but not changed. It is not really â€œopenâ€ in the sense of â€œOpen Sourceâ€, but it is called â€œopenâ€ by some people. In Software, this would be called â€œFreewareâ€ - Software to which someone holds a copyright (otherwise it would be â€œpublic domainâ€), but allows other people to use the software for free. The fact that â€œfreeâ€ can have two meanings (free beer vs. liberty) is a problem. I.e. the Free Software Foundation, which pioneered some important licenses for both software and content uses â€œfreeâ€ in the sense of â€œlibertyâ€. Maybe it would be better to call this content â€œcopyableâ€.â€   (truly) Open Content :  â€œThis is content that may not only be copied freely, but also modified. It is â€œopenâ€ in the sense that it may be changed. Some other viewpoints consider â€œopenâ€ as â€œFree for certain users or applicationsâ€, i.e. â€œmay be copied freely for personal or academic use.â€â€  Free-definition.com reminds us that â€œopen contentâ€ is not a stand-alone term, but that the â€œopen content movement is yet another variation on the \"open movement\" theme. Others include the open systems movement, open source movement, open information movement, etc. The open content movement seeks the cataloging and/or accumulation of nonproprietary knowledge; particularly machine-readable knowledge about subjects other than computer programming.â€   A Little History Please For those who want a more detailed history of Open Content, Joseph Reagle, Jr. has written an interesting review in M/C: A Journal of Media and Culture. His whole article is about 2,000 words; Iâ€™ve used fewer than 500 of them here.   â€œThe Free Software movement was begun by Richard Stallman at MIT in the 1980s. Previously, computer science operated within the scientific norm of collaboration and information sharing. When Stallman found it difficult to obtain the source code of a troublesome Xerox printer, he feared that the norms of freedom and openness were being challenged by a different, proprietary, conceptualization of information. To challenge this shift he created the GNU Project in 1984 (Stallman 1998), the Free Software Foundation (FSF) in 1985 (Stallman 1996), and the authored the GNU General Public License in 1989.  What is often meant by the term â€œopenâ€ is a generalization from the Free Software, Open Source and open standards movements. Communities marshaling themselves under these banners cooperatively produce, in public view, software, technical standards, or other content that is intended to be widely shared.  In 1991, Linus Torvalds started development of Linux: a UNIX like operating system kernel, the core computer program that mediates between applications and the underlying hardware. While it was not part of the GNU Project, and differed in design philosophy and aspiration from the GNUâ€™s kernel (Hurd), it was released under the GPL. While Stallmanâ€™s stance on â€œfreedomâ€ is more ideological, Torvalds approach is more pragmatic. Furthermore, other projects, such as the Apache web server, and eventually Netscapeâ€™s Mozilla web browser, were being developed in open communities and under similar licenses except that, unlike the GPL, they often permit proprietary derivations. With such a license, a company may take open source software, change it, and include it in their product without releasing their changes back to the community.  The tension between the ideology of free software and its other, additional, benefits led to the concept of Open Source in 1998. The Open Source Initiative (OSI) was founded when, â€œWe realized it was time to dump the confrontational attitude that has been associated with â€˜free softwareâ€™ in the past and sell the idea strictly on the same pragmatic, business-case grounds that motivated Netscape.â€ (OSI 2003) Since the open source label is intended to cover open communities and licenses beyond the GPL, they have developed a meta (more abstract) Open Source Definition (OSI 1997) which defines openness as:  Free redistribution   Accessible source code Permits derived works   Ensures the integrity of the authorâ€™s source code   Prohibits discrimination against persons or groups   Prohibits discrimination against fields of endeavor   Prohibits NDA (Non-Disclosure Agreement) entanglements   Ensures the license must not be specific to a product   Ensures the license must not restrict other software   Ensures the license must be technology-neutral   Substantively, Free Software and Open Source are not that different: the differences are of motivation, personality, and strategy. Given the freedom of these communities, forking (a split of the community where work is taken in a different direction) is common to the development of the software and its communities. One can conceive of Open Source movement as having forked from Free Software movement.â€  Bottom Line  Open content is virtually always free to use and reuse. However, the devil is in the details, so when youâ€™re considering revising and reusing â€œopen contentâ€ for commercial purposes, be sure to read the fine print of the licensing agreement. The GNU Free Documentation License, which is used by Wikipedia and Nupedia, is the industry standard.  Sources: Wiley, David. Opencontent.org. Accessed 29 Jun. 2004  .  Word IQ. Encyclopedia. Accessed 29 Jun. 2004  .   Life OpenContent. Definition. Updated 20 Aug. 2003. Accessed 29 Jun. 2004  . Free-Definition. Accessed 29 Jun. 2004  .  Reagle Jr., Joseph. \"Open Content Communities\" M/C: A Journal of Media and Culture Accessed 29 Jun. 2004  .   '),(5573,3767607,'01,July,2004','2004-07-01 00:00:00','    Not a rock group, not a geology database, but â€œdata about dataâ€. Are we allowed to say â€œdata about dataâ€? Find out in Jeromy Caballeroâ€™s extended definition:   The term metadata is most frequently used in connection with technology. However, its most general definition is simply data about data. Mark Taylor, an editor with the Global Spacial Data Infrastructure Association (GSDI), has said: \"Metadata is the term used to describe the summary information or characteristics of a set of data. This very general definition includes an almost limitless spectrum of possibilities ranging from human-generated textual description of a resource to machine-generated data that may be useful to software applications.\"1  Library Trends offers this definition: \"Metadata is the value-added information that documents the administrative, descriptive, preservation, technical, and usage history and characteristics associated with resources.\"2   In this sense, metadata is anything from the name and address on a stamped envelope to special tags describing a packet of information traversing the Internet. Metadata could even be background information providing context for an article or a story.  Taylor also describes metadata as something that allows knowledge about data to persist from user to user. For example, as people leave an organization, knowledge about certain data might be lost if that data is undocumented.1 But if metadataâ€”additional information describing that dataâ€”is attached to the data, knowledge of the dataâ€™s use and value can be preserved for future users.  Therefore, metadata can be defined as an interoperability enabler, providing an environment in which two systems can work together by making data from each system usable by the other system. The metadata is the common context in which the systems can cooperate and work with the same data. \"The application of metadata is critical in the digital environment,\" observes Library Technology Reports, \"because it allows a digital object or collection to be understood by both machines and humans in ways that promote interoperability.\"3  However, not only does metadata create context for interoperability, but it serves as a transmission/communication enabler that facilitates packaging, transmitting, routing, receiving, interpreting, and applying data over the Internet. Metadata accompanies all the data that crosses the Internet, allowing it to be correctly transmitted and correctly used when it reaches its destination, wherever that destination might be.   Because of the wide variety of data that crosses the Internet, and because its potential destinations are often unknown, the metadata that accompanies data transmitted over the Internet is complex and must account for a large array of possible uses. This metadata is difficult to incorporate into an information system that expects very specific, tailored metadata. \"The metadata required to describe the highly heterogeneous, mixed_media objects on the Internet,\" observes Library Trends, \"is infinitely more complex than simple metadata for resource discovery of textual documents through a library database.\"2 According to InfoWorld, the data that crosses the Internet is being collected by data stores at its destination, but the metadata is often being discarded as unuseable by those data stores. New methods of creating metadata that serves both to transmit the data and to effectively catalogue it in an information management system are the inevitable result.4   This is the type of metadata that \"provides the underlying foundation upon which digital asset management systems rely to provide fast, precise access to relevant resources across networks and between organizations.\"2   Not only do organizations need data to be transmitted, they need data to be efficiently catalogued and easily retrievable. Information Management Journal offers this definition of metadata:  \"Traditionally, the term `metadata\' has been widely used to characterize the descriptive information that will support search and retrieval of both paper and electronic material. Over the past three or four years the use of the term metadata has expanded to include additional information that must be acquired and retained in order to effectively manage electronic records over long period[s] of time, including permanently.\"5  Metadata is a storage and retrieval facilitator that enables organizations and even groups of organizations to catalogue, archive, and access data accurately and effectively in information management systems. Data entered into a database, for example, is accompanied by metadata describing that data in many different ways. This metadata allows the data to be searched and retrieved in a variety of contexts.   Finally, metadata can be defined as a supplemental resource for data.   \"Metadata is needed to describe both external and internal characteristics of the resource,\" explains Library Journal. \"External metadata explains and maintains those critical relationships among different versions of the same content.... Internal metadata allows for the description of the content of the resource at the desired level of granularity.\"6  When a video camera records a segment of footage, for example, it might also record supplemental information such as the time, the pan, angle, amount of light, and so forth. This information is metadata by dint of its accompanying the actual footage and being data about data. In a book, supplemental data is often gathered in an appendix, and data about the overall contents of the book is gathered in an index. In these cases, the data establishes the context in which the metadata has value, not the other way around.      Sources:  1 http://www.gsdi.org/pubs/cookbook/chapter03.html  2 Hunter, Jane L. \"A Survey of Metadata Research for Organizing the Web.\" Library Trends, 00242594, Fall2003, Vol. 52, Issue 2.  3 \"Applications of Metadata.\" Library Technology Reports, Sep/Oct2002, Vol. 38 Issue 5, p60.  4 InfoWorld, 11/10/2003, Vol. 25 Issue 44, p38.  5 Baron, Jason R. \"Recordkeeping in the 21st Century.\" Information Management Journal, 15352897, Jul99, Vol. 33, Issue 3.  6 Thomas, Judith. \"Digital Video, the Final Frontier.\" Library Journal, 03630277, Winter2004 Net Connect, Vol. 129.   '),(5574,3767607,'01,July,2004','2004-07-01 00:00:00','    Ooooh, I know this one! This is where newspaper columnists make millions when their sycophantic drivel gets run and rerun in papers nationwide. Yeah, thatâ€™s what it means in print publication, but what about technopublishing? Not surprisingly, it means about the same thing. Come on â€“ didnâ€™t you read about RSS earlier on this wonderful blog of ours? If you need a behind-the-scenes look at RSS, Tiffaniâ€™s put all the good stuff together in one place. Read on, weedhopperâ€¦     Abstract  The following is a short definition essay of the term â€œsyndication.â€  It is written for the benefit of the non-technically gifted content managers out there, though the author included highly technical passages for lack of a better explanation.  The definition essay is separated into the following divisions: Â·	Introduction to Syndication  Â·	What is Syndication?  Â·	How Does Syndication Work?  Â·	Why is Syndication so Important?  Â·	How Does Syndication Relate to Content Management?    Introduction to Syndication    Understanding syndication is easier if you first understand its origin.  Syndication began as a practice among news reporting agencies and comics artists.  Their columns and comics were syndicated, or licensed and distributed widely to other newspapers for publication (â€œSyndicationâ€).  This process was practiced for many years and is still widely practiced.  However, due to the evolution in computer technology, the term â€œsyndicationâ€ now has dual meaningâ€”one definition for print and one definition for the web.  Henceforth, this short definition will refer only to the web definition of â€œsyndication.â€    What is Syndication? â€œSyndication is a processâ€ (â€œWhat is RSSâ€).  It makes â€œa section of a website available for other sites to useâ€ (â€œweb syndicationâ€).  Syndication â€œis a way to keep updated in real-time of changes to your favorite websitesâ€ (â€œSyndication isâ€).  The process closely resembles the syndication practiced by news agencies, but becomes more complex upon deeper investigation.     Defined more comprehensively, syndication â€œrefers to making feeds available from a site so other people can display an updating list of content from it (â€œweb syndicationâ€).  It â€œis the process of using RSS/Atom for automated updates, another way of getting the information you wantâ€ (â€œWhat is RSSâ€).  RSS and Atom are examples of â€œfeeds,â€ or â€œwrapper[s] for pieces of regularly and sequentially-updated content, be they news articles, weblog posts, a series of photographs, and moreâ€ (â€œWhat is RSSâ€).  These feeds contain the content for the webpages they are linked to, and are continuously updated.  A result of syndication is a webpage that has news stories.  The stories are automatically updated through the feeds programmed into the page.      How Does Syndication Work?    There are two parts to understanding syndication: knowing the process and knowing the formats.    Syndication Process  Simeon Simeonov, lead web architect at Allaire Corporation, states, â€œSyndication covers the process of aggregating and reselling content, data, and services.â€  He further expounds on the complexity of syndication, saying:   . . . XSLT, XSL, and other transformation engines are employed to convert a system working within a narrow set of specifications to an open information clearinghouse that can interoperate with any content and data format. Affiliate management services provide a framework for exposing syndication to users of an application in a controlled, secure, and business-focused manner. This includes affiliate registration, setting up affiliate profiles, designating access permissions, syndication negotiation, and activity reporting. All these services are fully automated and remotable.    Syndication Formats  There are multiple formats of syndication, including ESF, RSS, and Atom.  Currently, RSS (or â€œReally Simple Syndicationâ€) is the most widely used of the formats and is XML based (Liorean).     XML syndication is similar to the process described by Simeonov, however, the description below includes the definition and an example of XML syndication:     XML Syndication is a form of electronic broad-based publishingâ€”meaning â€œdistributed to many people.â€ Publications released in one of the many XML Syndication formats are called â€œfeeds.â€ Visitors to web sites offering feeds can subscribe to the feeds using an application called a Desktop Reader.  The Desktop Reader application periodically checks information in the feeds it subscribes to, looking for new information. When new information is found, the Desktop Reader gets the userâ€™s attention, alerting that new information is available. This behavior is similar to that of an e-mail program, such as Microsoft Outlook. The way XML Syndication functions is completely different from e-mail, but it serves a similar purpose of allowing publishers to distribute the content to a massive audience. However, since feed subscription is controlled by the audience and not the publisher, XML Syndication is incapable of delivering SPAM. If a user receives unwanted information from a feed, he or she can unsubscribe from it without any interference or further â€œharassmentâ€ from the publisher.  Because of this significant difference from e-mail, XML Syndication is quickly becoming the new standard for broad-based electronic publishing. Many of the Desktop Readers available to users are powerful and free to download. (â€œWhat is XMLâ€). This explanation illustrates the importance of syndication in just one way.  But syndication has other justifications.    Why is Syndication so Important?   Syndication is a huge time saver to many organizations, giving them â€œthe ability to deliver accurate, constantly changing content in real-time [which] helps them reach new markets and build brand awareness in their established markets. Syndication also lets organizations build stronger, more meaningful extraprise relationships, fueling improved satisfaction and loyalty of customers and partnersâ€ (Eliason).    In addition, syndication plays an important part in society, as it perpetuates the distribution of up-to-date information, informing readers of cutting-edge world developments.  But syndication is also useful to content managers.    How Does Syndication Relate to Content Management?   On first glance, you might not identify syndication as a tool of content management.  However, one writer expounded on this relationship, clearly demonstrating syndicationâ€™s importance to the people in the world of content management.  Though the following is written in relation to XML syndication, the principle applies to all syndication formats: As the publisher, XML Syndication can help you to reach a significantly wider audience and reduce overhead associated with previous distribution models. This fosters a new and much more enriching environment for organizations already distributing e-mail newsletters and updates.  E-mail requires the exchange of personally identifiable information: the userâ€™s e-mail address. Because of SPAM, users are now reluctant to release their e-mail addresses to anyone they do not know, or will provide your company with a false or unchecked e-mail address. XML Syndication requires no such transfer of personally identifiable information.  Further, e-mail requires that you maintain a list of e-mail addresses representing your distribution list. In addition, the recommended best practice of the e-mail distribution model is to provide users with easily accessible functions to add or remove themselves from your distribution lists. With XML Syndication, publishers are no longer burdened with maintaining any such list.  Even further, e-mail requires that you conduct a transaction with one or many SMTP servers to transmit your content to users. Afterwards, the SMTP server must then transmit the message to each individual user, which is time-consuming with large lists. Services are available to perform this function for e-mail publishers, but these services are a repeating cost. XML Syndication completely alleviates the encumbrance of individualized transmission.(â€œWhat is XMLâ€)   Hopefully you have gained a more extensive knowledge of what syndication is, how it works, and ultimately, how it can contribute to the preservation of customer service relations and the personalization of customer documents.     Works Cited   Eliasson, Daniel. â€œContent Syndication.â€ http://zope.it.bond.edu.au.  June 24, 2002.  June 29, 2004. zope.it.bond.edu.au/projects/sdl/SDL-issues/content_syndication.   Liorean. â€œSyndication, the web, the future, XML and flying pigs.â€ www.web-graphics.com. May 24, 2004. June 29, 2004. http://www.web-graphics.com/mtarchive/001224.php.   Simeonov, Simon. â€œThe Future of Web Architecture.â€ June 29, 2004. http://www.infoloom.com/gcaconfs/WEB/philadelphia99/simonov.HTM#N111   â€œSyndication.â€ Webopedia.com. October 16, 2003.  June 29, 2004. http://webopedia.com/TERM/s/syndication.html  â€œSyndication is.â€mahajaba.com. June 29, 2004. www.muhajabah.com/xml-whatis.htm.   â€œWeb syndication.â€ Encyclopedia.thefreedictionary.com. June 28, 2004. http://encyclopedia.thefreedictionary.com/Web%20syndication. â€œWhat is RSS/XML/Atom/Syndication?â€ www.mezzoblue.com  May 19, 2004. June 28, 2004. http://www.mezzoblue.com/subscribe/about/. â€œWhat is XML Syndication?â€ June 29, 2004. http://www.howdev.com/products/syndicationstudio2004/documentation/source/get%20started/learn%20about%20xml%20syndication/what%20is%20xml%20syndication.htm.   '),(5575,3767607,'01,July,2004','2004-07-01 00:00:00','    Wow! In many ways, single sourcing is the heart and soul of every content management system. This is the strategy (both contextually and technologically) that lets writers take content from a single source and use it in multiple documents. If youâ€™re new to the whole CMS game, this is a great place to start.   Single sourcing refers to reusing content. The STCâ€™s single sourcing website defines single sourcing as \"Using a single document source to generate multiple types of document outputs.\"  Simply Written, a company that provides single sourcing services, gave this definition of single sourcing: \"So what is â€˜single sourcingâ€™? It is an updated approach to creating and maintaining that valuable asset called content or information or whatever you happen to call it. It is using a well-developed document manufacturing process instead of throwing stuff on a static page and sending it out the door. So instead of the â€˜pipelineâ€™ approach where you create a document, you create containers of information that are shared, or not shared, as needed\" (Simply Written, 6).  Companies and academics discuss two main methods of single sourcing: using tools for single sourcing and using databases for single sourcing. Each of these methods moves beyond copying and pasting information into multiple documents; instead, these methods enable writers to create a master source, and from that single source, writers produce multiple documents.   Using Tools for Single Sourcing  Tool advocates recommend using tools, such as FrameMaker + SGML, to create a master document. From the master document, writers can create multiple types of manuals. For example, using FrameMaker a writer could create a single master document that contains information for both PC and Mac users. From the master document, the writer can produce separate HTML manuals for PC users and Mac users; the writer can also create printed brochures for both types of users.  In this type of single sourcing, one document contains the content for a product, and that document can be used to create multiple types of documents (HTML and PDF); however, the content is not stored in a database for constant reuse. Anne Rockley breaks using tools for single sourcing into two levels: identical content/multiple media and stable customized content (Rockley, 190).   Identical Content/Multiple Media  In this type of single sourcing, identical content is used in multiple formats. For example, a main document produces an identical online userâ€™s manual and printed userâ€™s manual.   Stable Customized Content  In this type of single sourcing, a main document stores the content, but the content appears differently in different formats. For example, the same five steps might contain links online and use colored graphics in a printed Quick Start Guide.   Using Databases for Single Sourcing  Database advocates recommend storing chunks of content in a database for constant reuse. Writers create chunks of content, which are stored in a central database. The content can be shared across products and across departments. Writers can link to the content, and when the writers generate documents, the database automatically populates the document with the latest content. Using databases, writers create new content chunks, and organize and shape documents with existing content chunks.  Databases can also directly provide online users with content, creating unique documents for each user. Anne Rockley breaks this type of single sourcing into two levels: dynamic customized content and electronic performance support system (Rockley, 191).   Dynamic Customized Content  This type of single sourcing stores content in a database, and users access this content on the fly. Users find information through assigned user profiles, surveys and forms, or profiling; this user information gets used to â€œpushâ€ information to users.   Electronic Performance Support System (EPSS)  This type of single sourcing is also based on user profiles; however, users receive information only as they need it. The computer recognizes when the user needs information, and provides the appropriate content at the appropriate time.   Conclusion  Single sourcing involves reusing information, and it runs on a continuum. Copy and paste lies on one end of the continuum, using tools such as FrameMaker lies in the middle, and using databases lies on the other end. Each of these techniques (copying, using tools, and using databases) allows writers to take content from a single source and use it in multiple documents.   Works Cited  Finger, Heidi. â€œThe Joy of Single-sourcing.â€ http://www.soltys.ca/techcomm/articles/the_joy_of_single-sourcing.html Rockley, Ann. â€œThe Impact of Single Sourcing and Technology.â€ http://www.rockley.com/articles/Single_Sourcing_and_Technology.pdf Simply Written. http://www.simplywritten.com/library/SSIntro.pdf STC: Single Sourcing. http://www.stcsig.org/ss/ The Center for Information Development Management. â€œMaking a Business Case for Single Sourcing.â€ http://www.infomanagementcenter/pdfs/ssbusinesscase.pdf   '),(5576,3767607,'01,July,2004','2004-07-01 00:00:00','    Iâ€™ve lost a lot of arguments when challenged with the â€œItâ€™s just semanticsâ€ line. But what is semantics supposed to mean? Oh, now I get it â€“ semantics is meaning!! For us CMS groupies, semantics really means that when we name a unit of content within our model, we give it a name that has meaning, This makes it easy for authors to identify exactly what content they should include. One of our teammates, Steven C, has done the digging on semanticsâ€¦   â€œSemantic Informationâ€ is not a techno-ubiquitous term. Searches for â€œsemanticsâ€ and â€œsemantic informationâ€ on three tech sites (techdictionary.com, Cnet.com, and techtutorials.com) all generated zero responses. So, I turned to several print publications for clear definitions.    â€œSemanticâ€ refers to â€œmeaning.â€ Likewise, in technological jargon, â€œsemantic informationâ€ or â€œsemanticsâ€ always relates somehow to the meaning of a word, phrase, symbol, or instruction. I am in no way a technophile, but there seem to me to be at least three uses of the term. The first use has to do with programming instruction; the second relates to search engine methodology; the third reflects the way in which weâ€™ll use the term in this course â€“ semantics refers to the strategy used in naming content units in a unified content strategy.    First, letâ€™s see a basic definition and how the term applies to programming instructions. The 1996 Dictionary of Computing generically defines semantics as â€œthat part of the definition of a language concerned with specifying the meaning or effect of a text that is constructed according to the syntax rules of the language.â€ The 1996 Random House Personal Computer Dictionary takes the definition a bit further, injecting the element of programming instruction: â€œIn linguistics, [semantics is] the study of meanings. In computer science, the term is frequently used to differentiate the meaning of an instruction from its format. The format, which covers the spelling of language components and the rules controlling how components are combined, is called the languageâ€™s syntax.â€    The connection of semantics to programming instructions/language bridged into the 21st century (although this use may be archaic in 2004). The 2000 Encyclopedia of Computer Science uses several pages explaining the detailed instructional codes it refers to as â€œProgramming Language Semantics.â€ This is, to me, an incomprehensible algebraic-looking code of Greek letters, subscripts, and arrows, all apparently designed to tell the computer what to do when it encounters various words and/or phrases. The 2001 Dictionary of Computer Science, Engineering, and Technology continues the line of thought, defining semantics as â€œthe meaning of a string or sequence of toke symbols in some language, as opposed to syntax which describes how symbols may be combined independent of their meaning. The semantics of a programming language are a transformation from programs to answers.â€   Letâ€™s turn to the use of semantics in search engine methodology. Azeem Azhar, a U.K. writer, analyst and consultant on technology and society, published an enlightening article (4/25/03) discussing a Google acquisition. Forgive the lengthy quote, but Azhar explains well the importance of word meaning (semantics) in search methodology:        Google has purchased privately-held information retrieval company, Applied Semantics, a firm dotcom-formerly known as Oingo. It is a big deal because it brings together two competing schools, of which more below.    Information retrieval is the core of all search businesses. It is about creating software that solves a hard question: getting computers to understand human language with all its vagaries. These vagaries include: Â·                    polysemy (words with multiple meanings like DRIVE or SET) Â·        synonymy (different words with similar meanings like AIRPLANE and AIRCRAFT) Â·        multi-word expressions which need to be treated as such (BILL CLINTON) Â·                    errors, typos and poor grammar  For example, a key word search engine would find it hard to distinguish between A RED FISH and A FISH IN THE RED SEA   Broadly speaking there have been two major schools of thought. The first is one I call the statistical school and the second is the semantic. The statistical school held that context could be determined by look at statistical patterns within documents and across documents in a collection. Essentially, they use a variety of techniques to recognise word co-occurrence. So when words like DRIVE, CAR and HIGHWAY are used together frequently, we can make assumptions about the context of those words.   The other approach is the semantic approach. Here knowledge engineers build up a complex network of relationships, an ontology, that relates words together. So a CAR is defined as a type of VEHICLE and identical to the word AUTOMOBILE. A search on the word CAR will also turn up documents with the word AUTOMOBILE in it, even if they donâ€™t mention it. Such semantic networks require a good deal of work and a lot of maintenance to keep them up to date.   So why is this relevant to Google and Applied Semantics? Well Google comes from the statistical school of information retrieval, albeit in a very light way. Currently, Google queries are not parsed very much at all. No stemming is applied, although some word proximity algorithms are used. Google does take advantage of one unique aspect of the Web: the interconnections between documents which provide a context-weighting to documents based on their link popularity. Applied Semantics will add a layer of semantic understanding that Google needs. There AdSense technology is already being used by a raft of web sites to improve targeting based on user-behaviour.    Finally, we turn to how weâ€™ll use the term â€œsemantic informationâ€ in designing unified content strategy. In â€œManaging Enterprise Contentâ€ (2003), Rockley defines semantic information as, â€œa component of an information model; uniquely identifies the content of that element, making it easy for authors to identify exactly what content they should include. Semantic information also enables the identification and reuse of specific content.â€ Used this way (and compared to other meanings), â€œsemantic informationâ€ is actually a straightforward idea. It simply means that as we identify tags that link to content units, we donâ€™t use generic names (e.g.  ). Instead, we use names that have meaning (e.g.  ). According to Rockley (and logically so), these semantic tags are of great benefit to authors engaged in opportunistic reuse of content units. Rockley differentiates semantic tags from metadata; weâ€™ll explore that in Chapter 12.    In summary, â€œsemantic informationâ€ always refers to a word, phrase, symbol, or instruction that is recognized for its meaning, not its form. In the realm of a unified content strategy, â€œsemantic informationâ€ means that we tag content units with meaningful names.     Sources: Dictionary of Computing (4th Ed.). Oxford: Oxford University Press, 1996.    Margolis, Philip E. Random House Personal Computer Dictionary. New York: Random House, Inc., 1996.   Ralston, Anthony, Edwin D. Reilly, and David Hemmendiner, eds. Encyclopedia of Computer Science (4th Ed.). New York: Groveâ€™s Dictionaries, Inc., 2000.   Laplante, Phillip, ed. Dictionary of Computer Science, Engineering, and Technology. Boca Raton, FL: CRC Press, 2001.   Azhar, Azeem. technology, economics, stuff. Website archive 25 Apr. 03, accessed 21 Jun. 04  .   Rockley, Ann. Managing Enterprise Content: A Unified Content Strategy. Indianapolis, IN: Newriders, 2003.   '),(5577,3767607,'01,July,2004','2004-07-01 00:00:00','    Everybody knows that scalability is the rating system used by mountaineers. Or is it a measure of my willingness to check my personal poundage each morning? Nope. For techies like us, this seems like an easy one: scalability basically refers to how well (or not) a computer system can handle increasing loads. But thereâ€™s a lot more to itâ€¦    Definitions    1.	The ease with which a system or component can be modified to fit the problem area (http://www.sei.cmu.edu/str/indexes/glossary/scalability.html).  2.	The capacity of a system to handle increasing load or demand (evolt.org website at http://www.evolt.org/article/Scalability_s_New_Meaning/21/238896/).  3.	The capability of increasing the computing capacity of a website or a computer systemâ€”and the siteâ€™s or systemâ€™s ability to process more operations or transactions in a given period, in particularâ€”by adding more, or more powerful, processor (Computerworld website at http://www.computerworld.com/hardwaretopics/hardware/story/0,10801,46308,00.html)  4.	Scalability is about reducing the time to solution for critical tasks (SAS website at hppt://support.sas.com/rnd/scalability/intro/index.html)  5.	Interface scalability is probably the most relevant to technical communicators.   According to Jeff Freund at CMS Watch (http://www.cmswatch.com/Features/TopicWatch/FeaturedTopci/?feature_id=92) â€œjust as your content management infrastructure must scale to meet your growing web publishing needs, the interfaces to your web content management system must scale to meet the ongoing needs of your organization and editorial team.â€  Content and content management needs are growing at astounding rates. Rather than wait for a problem to occur, businesses should invest in scalable systems that can expand with and accommodate their growing content management needs. In order for content management systems to be â€œscalable,â€ they must be able to handle any type of and high volumes of content. Content includes not only text but digital media (audio & video) and objects in XML applications.     The traditional way to increase scalability is to add large multi-processor servers. A newer way to increase scalability is to add many small servers. Either way, the term â€œscalabilityâ€ usually conjures up images of lots of servers. Even though scalability usually refers to hardware, the data explosion of the information age has forced software developers to make scalability a priority and build applications that can be easily adapted to meet growth demands. Some computer languages (Java, VB, C) are more scalable than others. The reason for focus on hardware rather than software to improve scalability is that it costs less to buy big machines than it does to build efficient programs.        The concept of scalability is further broken down into two types referred to as vertical (scaling up) and horizontal (scaling out). Scaling up is achieved by adding large servers, multi-processors, or even a mainframe computer; and scaling out is achieved by adding many small servers. Vertical scaling (large servers) is most effectively used for back-end databases because database transactions work better when ran from a single applicationâ€”think here of the powerful processing needed to run financial transactions for large banks. Horizontal scaling (small servers) is usually easier and faster to implement than vertical scaling and can be effectively used to run multiple copies of Web or application software because these transactions work when ran from different applications.   Windows new COM+ component architecture that comes with their 2000 operating system handles multiple transactions by balancing the load through distributed computing (multiple servers). It can be set to allow multiple transactions for Web applications or require separate transactions to maintain data integrity in high-volume databases.  	 Whether to scale up with large powerful processors or scale out with small servers depends on how large a companyâ€™s database is and the number of transactions that occur on their web applications. As a rule of thumb, when a companyâ€™s data storage needs exceed four processors, itâ€™s time to consider scaling up.    The task of assessing scalability usually falls to engineers who are responsible for the companyâ€™s technical infrastructure. Engineers assess not only how well their existing applications respond to increased data storage and retrieval demands but how well their applications run with added resources (servers, etc.). Linear scalability means that with fixed resources performance decreases as demand increases, and when server resources increase with demand performance improves. Because people want information on demand, scalability is also about speed.    Many Content Management System (CMS) interfaces begin to break down when content demands increase, but interface scalability problems are harder to identify than technical scalability problems. Web-based interfaces are especially prone to problems. Common web-based interface scalability problems are: limited bulk operation options so that operators are forced to perform the same tasks over and over again, collaboration features that allow usage to grow beyond system scalability, content is difficult to find because itâ€™s not separated into functional units, and lack of behavior-based filtering which would eliminate content hunting problems not solved by permission based filtering. So, to keep editorial teams productive and satisfied, decision makers need to consider interface as well as technical scalability needs.   '),(5578,3767607,'01,July,2004','2004-07-01 00:00:00','    If you donâ€™t already know what RSS is, and you havenâ€™t checked out the definition of â€œsyndicationâ€ yet, drop that mouse and step away from the computer!! RSS is a form of syndication; in fact, it means â€œReally Simple Syndication.â€  (Is that the best acronym youâ€™ve ever seen, or what?!) So if youâ€™re smart, youâ€™ll want to get a good understanding of syndication first. Then come back here and tackle RSS.    RSS â€“ May The Best Content Win   Upfront lingo clarifications Â·	Aggregator = portal  Â·	A â€œfeedâ€ contains a list of items, each identified by a link  Â·	Blog = web log, used to be personal diaries of sort, now business potential is being realized      What is RSS?   RSS is the acronym for Really Simple Syndication.  RSS can increase traffic to a site by gathering and distributing news.  RSS is a format for syndicating news and the content of news-like sites, â€œincluding major news sites like Wired, news-oriented community sites like Slashdot, and personal weblogs. But it\'s not just for news.â€    Pretty much anything that can be broken down into discrete items can be syndicated via RSS: the \"recent changes\" page of a wiki, a changelog of CVS checkins, even the revision history of a book. Once information about each item is in RSS format, an RSS-aware program can check the feed for changes and react to the changes in an appropriate way. http://www.xml.com/pub/a/2002/12/18/dive-into-xml.html      Um . . . come again?  RSS is a means of content distribution.  Itâ€™s a mini database maintaining headlines and descriptions of whatâ€™s new on your site.  It is a form of syndication, hence the name.  If you look at your morning paper, there are articles from various different news sources, AP, Boston Globe, NY Times, etc.  RSS works the same way.  RSS is a format that allows specific content (like headlines) to be added to other sites.  Your specific content (of which you have full control of what or how much) is instantly distributed to other sites.  RSS can sort content by topic, time, date, and other specifics.  RSS allows peoples\' computers to fetch and understand the information, so that all of the lists they\'re interested in can be tracked and personalized for them. It is a format that\'s intended for use by computers on behalf of people, rather than being directly presented to them (like HTML).  To enable this, a Web site will make an RSS feed, or channel, available, just like any other file or resource on the server. Once a feed is available, computers can regularly fetch the file to get the most recent items on the list. Most often, people will do this with an aggregator, a program that manages a number of lists and presents them in a single interface. http://www.mnot.net/rss/tutorial/   Originally, blogs were a primary market for RSS.  Readers would subscribe to a blog, and the RSS would allow the readers to receive up-to-the-minute headlines about the topics that interested them.   When blogs first began, they seemed to be a form of online diaries, but the idea of blogs has really grown.  Now there are blogs on almost every topic, and it is a cheap and easy market savvy businesses are discovering.  The RSS feeds are a wealth of information:  hourly, daily and weekly updates, new information, new products, product updates, company changes, company specifics, and the list keeps going.      How is this relevant to publication management?  RSS is a way of getting content out to the right people at the right time and in the right format.  RSS makes it easier for your users to keep up with the content they want.  It allows immediate notification when something that interests them is available on your site.   With thousands of sites now RSS-enabled and more on the way, RSS has become perhaps the most visible XML success story to date. RSS democratizes news distribution by making everyone a potential news provider. It leverages the Web\'s most valuable asset, content, and makes displaying high-quality relevant news on your site easy. Soon we\'ll see RSS portals with user-rated channels, cool RSS site of the day, build your own topic-specific portal, and highly relevant search engines. A collective weblog would be another intriguing possibility. May the best content win.  http://www.webreference.com/authoring/languages/xml/rss/intro/3.html RSS is not limited to your computer screen.  RSS feeds can also be sent to PDA\'s, cell phones, email ticklers and even voice updates. Email newsletters can easily be automated with RSS. â€œEven more compelling, affiliate networks and partners of like-minded sites (say a collection of Linux sites) can harvest each other\'s RSS feeds, and automatically display the new stories from the other sites in the network, driving more traffic throughout.â€ http://www.webreference.com/authoring/languages/xml/rss/intro/    Technical information that seems very important but is way over my head, so Iâ€™m not even going to attempt to paraphrase it Each RSS channel can contain up to 15 items and is easily parsed using Perl or other open source software. If you want more details on creating RSS files see Jonathan Eisenzopf\'s excellent article in the February issue of Web Techniques. But you don\'t have to worry about the details, we\'ve made it easy to create your own RSS channel with free open source scripts, all Web based. More on these later. Once you\'ve created and validated your RSS text file, register it at the various aggregators, and watch the hits roll in. Any site can now grab and display your feed regularly, driving traffic your way. Update your RSS file, and all the external sites that subscribe to your feed will be automatically updated. What can be easier? (http://www.webreference.com/authoring/languages/xml/rss/intro/)   Here are the basics of RSS (no matter what the version â€“ there are currently 7 versions)  1.	It is XML. This means it must be well-formed, include a prolog and DTD, and all elements must be closed.  2.	The first element in the document is the   element. This includes a mandatory version attribute.  3.	The next element is the   element. This is the main container for all RSS data.  4.	The \"title\" element is the title, either of the entire site (if it\'s at the top) or of the current item (if it\'s within an  ).  5.	The \"link\" element indicates the URL of the Web page that corresponds to the RSS feed, or if it\'s within an  , the URL to that item.  6.	The   element describes the RSS feed or the item.  7.	The   element is the meat of the feed. These are all the headlines (\"title\"), URLs (\"link\") and descriptions that will be in your feed. http://webdesign.about.com/cs/rss/a/aa052603a.htm   Examples Below are just two examples of the thousands of sites who use RSS.  You can tell if a site uses RSS by checking the code.    Moreover http://w.moreover.com/main_site/content/index.html   Moreover asserts the following:    Moreover is proud to offer one of the most comprehensive and diverse range of sources delivering relevant, reliable and timely content;    From more than 6,500 premium websites and 200,000 business value weblogs  In 26 languages, including multiple Asian languages  From 115 countries including provincial and regional sources  Available within minutes of publication  Each source reviewed and ranked by a team of editors  Enhanced with up to 25 individual pieces of metadata    \"Moreover does a sweet and swift job of ferreting out the top stories of the world -- news, research and business sites . . . trumps its competition.\" Time Magazine, Winner, Best of the Web   \"As close as a company can get to having eyes and ears everywhere on the Web\" Wired News    This site is a prime example of RSS use.  Clearly Moreover does not come up with all of this content every singe day.  Whatever they have specifically requested is delivered to them the second it hits the web.     Oâ€™Reilly http://www.oreillynet.com/meerkat/   Another site with clear examples of specific content RSS feed           Works Cited     King, Andrew B.  â€œWebRef and the Future of RSSâ€.  Jupitermedia Corporation. 27 March 2000.  Accessed 21 June 2004. http://www.webreference.com/authoring/languages/xml/rss/intro/3.html   Kyrnin, Jennifer.  â€œRSS â€“ Real Simple Syndicationâ€.  About.com. Accessed 21 June 2004. http://webdesign.about.com/cs/rss/a/aa052603a.htm   Nottingham, Mark.  â€œRSS Tutorialâ€. 22 May 2004.  Accessed 21 June 2004 http://www.mnot.net/rss/tutorial/     Pilgram, Mark.  â€œWhat is RSS?â€.  Oâ€™Rilley xlm.com. 18 Dec. 2002. Accessed 21 June 2004  http://www.xml.com/pub/a/2002/12/18/dive-into-xml.html   Rees, L.C., â€œWhat is XML?â€ Accessed 21 June 2004. http://www.geocities.com/SiliconValley/Peaks/5957/xml.html   '),(5579,3767607,'01,July,2004','2004-07-01 00:00:00','    Some acronyms are awesome uses of our 26 letters (see RSS)! But have you ever seen an acronym that struggles to make sense? Well you have now â€“ PHP. PHP is supposed to stand for â€œPHP: Hypertext Preprocessor.â€ How can an acronym include itself? Too weird â€“ I think whoever came up with this one must have taken a few extra licks on the blow pop. But anyway, PHP is a scripting language that enables dynamic web pages and boosts visitor interactivity. Youâ€™ve got to read more, right?â€¦    Defining PHP  Since it was created in 1994, PHP (a recursive acronym for PHP: Hypertext Preprocessor) has steadily risen in popularity as a scripting language. PHP is an open-source language, meaning that the code used to create it is available for anyone to use and modify, and that the language itself is freely available to download and install.  Youâ€™ve probably encountered many web pages already that have been generated using PHP. You donâ€™t have to know anything about PHP in order to open a .php page in your browser. However, installing PHP and getting it to work with your computer and your website is a gentle initiation into geekdom. Even with a manual to walk you through the process, youâ€™ll still have learned plenty about your computer and your website server by the time you have PHP up and running. And once you have PHP ready to go, you still have to learn how to use it.   PHP is a scripting language.  The main online resource for all things PHP is www.php.net. \"PHP is a widely-used general-purpose scripting language that is especially suited for Web development and can be embedded into HTML,\" says php.net on its front page.1 This is probably the most basic definition of PHP. Most people use PHP to beef up the functionalities and capabilities of their websites. PHP is primarily a scripting language that can be inserted directly into an HTML document.  \"In an HTML document,\" says webopedia.com, \" PHP script...is enclosed within special PHP tags. Because PHP is embedded within tags, the author can jump between HTML and PHP...instead of having to rely on heavy amounts of code to output HTML. And, because PHP is executed on the server, the client cannot view the PHP code.\"2  In English, this means that you can write PHP commands in the midst of the HTML tags in the HTML page you are creating for your website. When a visitor opens your page, however, all they see is a straight HTML page, because your web server already processed the PHP commands into HTML for them. This \"server-side processing\" is different than a scripting language like JavaScript, whose commands are processed by the website visitorâ€™s computer.   PHP enables dynamic web pages.  Server-side processing enables what are known as dynamic web pages. \"A dynamic Web page is a page that interacts with the user, so that each user visiting the page sees customized information. Before the page resolves, the Web server calls PHP to interpret and perform the operations called for in the PHP script.\"3 A dynamic web page built with PHP can display regularly updated information within a consistent layout. Rather than using static HTML tags that canâ€™t be changed without manually opening, saving, and re-uploading the file, PHP commands enable a web page to be created with content pulled from other files placed on the serverâ€“or content generated by the computer itself using PHPâ€™s programming functionality. Visit any large news site, like cnn.com or even news.google.com and you will see dynamic web pages at workâ€“constantly updating content within a consistent layout at one web address. PHP boosts visitor interactivity.  Not only can PHP enable a website to display dynamically generated content, it can allow a visitor to interact with the website through their browser. \"You can use [PHP] to add dynamic features to HTML pages or use it to create entire dynamic sites that generate HTML on the fly in response to user input,\" says viptx.net.4 PHP can be used to simply detect the kind of browser a visitor is using; it can be used to create forms, process a visitorâ€™s input, and place the results back on the screen; it can generate pages filled with information retrieved according to the visitorâ€™s activities and requests on the website. Its power comes through being able to add interactivity to a web page without the visitor having to leave the browser and open some other application.    PHP runs the back end.  Web developers like to talk about something called the \"back end.\" The back end is all the code- and data-processing that happens behind the scenes at a website. PHP hangs around at the back endâ€”taking visitor input and telling the server how to process it into more meaningful data, telling the server how to talk to a database and retrieve data, and telling the server how to generate a new page filled with useful content for the visitor.  \"One of the strongest and most significant features in PHP is its support for a wide range of databases,\" says php.netâ€™s PHP manual.5 PHP works wonderfully with many different databases and is compatible with database querying languages like SQL. PHP can enable a website to become a powerful database interface and provide strong search and retrieval capabilities.  \"PHP has extremely useful text processing features, from the POSIX Extended or Perl regular expressions to parsing XML documents,\" continues php.net.5 This allows PHP to help put data from a website visitor into a form that the server can understand. Dr. Dobbâ€™s Journal extends PHPâ€™s data-handling powers even further: \"With PHP you are not limited to output HTML. PHP\'s abilities includes outputting images, PDF files and even Flash movies (using libswf and Ming) generated on the fly. You can also output easily any text, such as XHTML and any other XML file. PHP can autogenerate these files, and save them in the file system, instead of printing it out, forming a server-side cache for your dynamic content.\"6   PHP can create applications.  Finally, PHP can be used to create entire applications that can be run on browsers or even from the desktop. \"With PHP,\" writes Dr. Dobbâ€™s Journal, \"you develop applications that can be viewed by all HTML-capable browsers.\"6 The php.net manual points to PHP-GTK, an extension of PHP, as a way to create standalone applications that are handled by the userâ€™s computer.5  PHP is a complete scripting language powerful enough to create entire applications, yet versatile enough to enable a range of simple to complex functionality boosts in HTML pages. Because of its power, versatility, compatibility with HTML, and wide availability, it continues to grow in popularity among a wide range of computer users.  1 http://www.php.net 2 http://www.webopedia.com/TERM/P/PHP.html 3 http://www.ameritech_hosting.net/resources/glossary.htm 4 http://www.viptx.net/orientation/glossary.html 5 http://www.php.net/manual/en/intro_whatcando.php 6 Ross, Darryl; Zymaris, Con. \"DB FORMS: PHP, MYSQL, AND PHPLIB.\" Dr. Dobb\'s Journal, August 1, 2000, Vol. 25, Issue 8.   '),(5580,3767607,'01,July,2004','2004-07-01 00:00:00','    Isnâ€™t localization simply the process of translating your documents to whatever language you want? Get an English to Spanish (or Polish, or German, etc.) translator, and voila! Youâ€™ve localized your document. But wait, that approach has caused major problems for some companies:   The Dairy Association\'s huge success with the campaign \"Got Milk?\" prompted them to expand advertising to Mexico. It was soon brought to their attention that the Spanish translation read, \"Are you lactating?\"  Coors put its slogan, \"Turn It Loose,\" into Spanish, where it was read as \"Suffer From Diarrhea.\"  And my personal favorite:  In a Tokyo bar: \"Special today for the ladies with nuts.\"  Still confused? Read Becky Johnstonâ€™s excellent definition that points out the important  differences between localization and mere translation:   Dictionary.com defines localization as a noun meaning to limit to a specific area. In businesses, the term localization usually refers to translating content; in this sense, the content becomes available for a specific area of the world.   Several localization providers do not specify a difference between straight translation and localization. However, LISA, the Localization Industry Standards Association, defines three levels of localization: translation, localization, and globalization/internationalization.  Translation Translation is the act of changing content to a new language such as changing a userâ€™s manual from English to Spanish. ASE&T, a translation company, argues against â€œthe erroneous notion that translation is simply â€˜typing in a foreign language.â€™â€   According to ASE&T, â€œThe company that performs your translation is crafting your image to your foreign customers. If your documentation has taken six months to compile, revise, and edit, it is critical to allocate sufficient time to achieve similar high-quality results in a foreign language. For large-scale projects, short time lines or simultaneous release, the scheduling and management of the translation process are as important as the translation itself.â€  Although translation is concerned with consistently using terms in the same document and with accurately changing to a new language, translation does not alter the original content.  Localization Localization is more involved than translation. Localization involves translating and polishing content, so it is culturally appropriate.   According to the Localization Institute, â€œLocalization is the process of creating or adapting a product to a specific locale, i.e., to the language, cultural context, conventions and market requirements of a specific target market. With a properly localized product a user can interact with this product using his/her own language and cultural conventions. It also means that all user-visible text strings and all user documentation (printed and electronic) use the language and cultural conventions of the user. Finally, the properly localized product meets all regulatory and other requirements of the user\'s country/region.â€  Localized content might be altered from the original content because localized content is designed for a specific region. LISA used the example of the changes that might occur when localizing a software package. â€œOn the software programming side, screen dialog boxes and field lengths may have to be altered; date, time and currency formats changed; delimiters for figures replaced; and icons and colors adapted; to give only a few examples. What is more, in the case of bi-directional languages (such as Arabic and Hebrew) and double-byte character sets (such as those for Chinese, Japanese and Korean), more extensive reprogramming may be required to ensure that localized text and numerals are displayed correctly on the target platforms.â€  Localizing involves changing the content to fit cultural expectations; whereas, translation involves only changing the language of the content. LISA also describes some of the localization changes made to chunks of content: â€œ[T]he color, size, and shape of objects such as coins and notes, taxis, telephones and mailboxes, and buses and ambulances, traditionally vary from country to country. Vehicles may suddenly have to drive on the other side of the road, while dress codes will vary, and symbols take on a new significance. Similarly, mainstream business applications such as address databases and financial accounting packages have to be adapted to the procedures and conventions applicable in their new environments.â€  Localization makes content more accessible than mere translation.  Globalization/Internationalization Globalization involves an overarching strategy for creating a global company. When a company uses globalization techniques, content becomes easier to translate and localize because the content is already culturally neutral. Many consulting companies, such as welocalize, get involved early in the content creation process. According to welocalize, this early interaction makes the content more consistent across languages and speeds up the launch date of localized content.   According to LISA, â€œ[I]nternationalization is the â€˜oppositeâ€™ or forerunner of localization. In other words, it is the process of designing and implementing a product which is as culturally and technically â€˜neutralâ€™ as possible, and which can therefore easily be localized for a specific culture or cultures. This reduces the time and resources required for the localization process, thus saving producers money and improving their time-to-market abroad. As with localization, language, technical and contents issues are involved, with project management and coordination also playing a significant role. Internationalization has now reached the point where major software publishers can release 30 or more different localized versions within a month or two of the original version, a process known as â€˜sim-shipâ€™ (short for â€˜simultaneous shipmentâ€™).â€  Globalized text must still be localized; however, the localization becomes less costly and more efficient. According to the Localization Institute, â€œInternationalization is a way of designing and producing products that can be easily adapted to different locales. This requires extracting all language, country/regional and culturally dependent elements from a product. In other words, the process of developing an application whose feature design and code design do not make assumptions based on a single locale, and whose source code simplifies the creation of different local editions of a program, is called internationalization.â€  Summary Translation means changing contentâ€™s language without changing the content. Localization refers to adapting content, so it becomes applicable to specific regions. Finally, globalization or internationalization refers to writing your initial content so it is easier to translate and localize.  References  ASE&T http://www.asetquality.com/gf.htm  Dictionary.com www.dictionary.com  LISA (Localization Industry Standards Association) http://www.lisa.org/  The Localization Institute http://www.localizationinstitute.com/  welocalize http://www.welocalize.com/english/services/software_localization.html   '),(5581,3767607,'01,July,2004','2004-07-01 00:00:00','    A tech writer friend told me how proud his 93-year old grandma (â€œGrammy Berthaâ€) was when she found out that he had become an architect. He couldnâ€™t bring himself to tell her that he wasnâ€™t the kind of architect she had in mind. Probably just as well. She would not have understood that while a traditional architect designs and directs the construction of buildings, an information architect is designs the organization of a website\'s structure and content, the labeling and categorizing of information and the design of navigation and search systems. Itâ€™s a whole new world, Bertha.      Although the term information architecture has become a commonly used term in the world of web design, its origin stems back to the 1970â€™s. Long before the internet was a household mainstay, Richard Saul Wurman, a trained architect who later became a skilled graphic designer coined the term information architecture. Wurman recognized the similarities between designing a building and the problems of organizing and clearly presenting information. Just as an information organizer or information architect evaluates and arranges information, when designing a building, an architect analyzes the buildingâ€™s use and designs the flow and layout so that it best meets user needs.   Since Wurmanâ€™s conception of the term information architecture, many variations have evolved stemming from a need to organize an ever increasing influx of information presented by the Internet and other communication platforms. The variations fall within three basic categories: the actual structure, the physical act of creating the structure and the community that promotes the structure and practices the act of creating it.  The following definition expressed by Asilomar Institute for Information Architecture (AIfIA) further defines these categories: â€œ1. The structural design of shared information environments. 2. The art and science of organizing and labeling web site, intranets, online communities and software to support usability and findability. 3. An emerging community of practice focused on bringing principles of design and information architecture to the digital landscape.â€(AIfIA)  The Actual Structure  Information architecture is often used to describe the structure or the blue print behind the information. This can best be described by an example taken from Shel Kimenâ€™s definition of information architecture, â€œIn a library, for example, information architecture is a combination of the catalog system and the physical design of the building that holds the books. On the Web, information architecture is a combination of organizing a siteâ€™s content into categories and creating an interface to support those categories.â€ (Kimen)  The Physical Act of Creating the Structure              The occupation of web designer has become more prevalent, and the term information architecture has been adopted to describe what web designers do.  Because of this, â€œthe physical act of creating a structureâ€ definition of information architecture has grown to be the most common use of the term. The act of creating dynamic web sites that guide people through information is the foundation, as both of the definitions below illustrate. Mattie Langenburg refers to information architecture as, â€œtaking content and creating a structure to present that content to an audience.â€ (Cohen)  Additionally, WebWord says, â€œInformation architecture involves the design of organization and navigation systems to help people find and manage information more successfully.â€ (Rhodes)   The Community that Promotes the Structure and Practices the Act of Creating it  Information architecture also refers to the community that is affected, works within and works for the information environment. This community evolves as the value of structure gains visibility with key business members. The global information community serves as a communication link for the information architect to the rest of the business. This global community connects the architect to tools, development opportunities and research while advocating the information architecture basis. Because this definition of community is evolving, few concrete definitions exist. However, considering there are entire organizations and web sites devoted to the advancement of information architecture as a community, it warrants discussion as a definition of information architecture.  Although information architecture carries multiple definitions, the term itself is becoming more common as web sites become more complex and the need for specialized knowledge increases. Information architecture can refer to the actual structure of the information on a web page, while also referring to the roles of the architect in creating the structure and the community that supports it.  References  Asilomar Institute for Information Architecture (AIfIA). http://aifia.org/pg/about_aifia.php  Cohen, Sacha. Becoming an Informational Architect.  http://technology.monster.com/articles/infoarchitect/  Kimen, Shel.10 Questions about Information Architecture. CNet. June 22, 1999. http://builder.com.com/5100-31-5074224.html  Shimple, John. Information Architecture. HotWired. http://hotwired.com/webmonkey/98/28/index0a.html?tw=design   '),(5582,3767607,'01,July,2004','2004-07-01 00:00:00','    An old boss once told me I was â€œextensibleâ€, I was thrilled.  Didnâ€™t that mean he thought I was forward compatible, able to be extended beyond my current abilities? Wow! What a compliment! Later on, I realized he actually meant to say that I was â€œexpendibleâ€.  Oops. Now thatâ€™s a horse of different color. Maybe Iâ€™ll suggest that my old boss read the following definition by Tim Carter (which I find not at all expendable):     Extensibility seems to be a uniform idea. As I researched this term, I didnâ€™t find different definitions. Most applications of the term were the same in nature. What was different were the strategies and the approaches as extensibility was applied to computer programs or coding by various programmers and application producers.  From the TechWeb Encyclopedia, I found this definition which seemed to be consistent in each of the other resources. (http://www.techweb.com/encyclopedia/defineterm?term=extensible)  â€œThe capability of being expanded or customized. For example, with extensible programming languages, programmers can add new control structures, statements or data types.â€  Basically programmers can create the code for applications and operating systems to accept adaptations from third party programmers who wish to extend or diversify the original code. The reasons for extending the capabilities of original code are wide and varied, from adding software to making small â€˜personality tweaksâ€™ that can change from programmer to programmer. It is the strategy for implementing extensibility that differs from party to party.  From his Architecture of the World Wide Web, First Edition Ian Jacobs writes, â€œDesigners can facilitate the transition process by making careful choices about extensibility during the design of a language or protocol specification. Language designers SHOULD provide mechanisms that allow any party to create extensions that do not interfere with conformance to the original specification.â€ This seems to be a holistic approach and one that, in the community of pure programming, exists â€“ flying in the face of free enterprise. For example, Bill Gates and Microsoft do not allow extensions to their software except in a very narrow and controlled capacity because they need to protect the bottom line for their organization. Again from Ian Jacobs, â€œApplication needs determine the most appropriate extension strategy for a specification. As part of defining an extensibility mechanism, a specification should set expectations about agent behavior in the face of unrecognized extensions.â€ These extensibility agents are either allowed at the factory or denied as the strategy of the original programmers dictates.  http://www.w3.org/TR/webarch/#pr-allow-exts  Utilizing another definition and strategy for implementation, Brian Bershad, et al. in the SPIN Operating System show the need for extensibility not only in applications but in the larger environment of operating systems. â€œExtensibility needs to be introduced to the extent that allows applications to safely change the operating system\'s interface and implementation. Extensions allow an application to specialize the underlying operating system in order to achieve a particular level of performance and functionality.â€  Bershad, et al. go on to elaborate on operating system extensibility. â€œThe need for extensibility in operating systems is shown clearly by systems such as MS-DOS, Windows, or the Macintosh Operating System. Although these systems were not designed to be extensible, their weak protection mechanisms have allowed application programmers to directly modify operating system data structures and code [Schulman et al. 92]. While individual applications have benefited from this level of freedom, the lack of safe interfaces to either operating system services or operating system extension services has created system configuration \"chaos\" [Draves 93].  http://216.239.51.104/search?q=cache:riW5fu6hH-gJ:www.cs.washington.edu/research/projects/spin/www/papers/SOSP95/sosp95.ps+extensibility&hl=en  From a global or open perspective, in The Object-Oriented Database System Manifesto Malcolm Atkinson, et al. write that programmers need to write their applications using extensibility â€œâ€¦ in the following sense: there is a means to define new types and there is no distinction in usage between system defined and user defined types. Of course, this requirement strengthens that capability by saying that newly created types must have the same status as existing ones.â€ In other words, programmers should expect extensions, and the programmers doing the extending should respect the original code.  http://www-2.cs.cmu.edu/People/clamen/OODBMS/Manifesto/htManifesto/node10.html  Now contrast that ideal, open society to the mandates and licensing of Microsoft products. For Microsoft Visual Studio, a developer tools suite for Windows and Web applications, consumers can purchase this software to create extensibility in other Microsoft products. In the Microsoft Visual Studio 6.0 documentation, extensibility and the rules for its use are laid down in no uncertain terms: â€œExtensibility is the capacity to extend, or stretch, the functionality of the Microsoft Visual Studio integrated development environment (IDE). The IDE provides you with a programming interface known as the Extensibility object model, a set of powerful, easy-to-understand interfaces for customizing the environment. It allows you to hook into the IDE to create extensions known as add-ins. There are two types of extensibility: built-in and user-designed.   Â·         Built-in extensibility features are already contained in the development environment. They\'re conveniences such as keyboard rebinding, customizable command bars, docked and tab-linked window configurations that you can name and return to at any time, new templates that you can create for the Template wizard, and so forth.  Â·         User-designed extensibility features are constructed with code. They take two different forms: add-ins and wizards.\"   In addition to these rules, Microsoft also lays down a series of best practice ideas for the use of extensibility. Clearly, Microsoft products are extensible; however, the extensibility is controlled by Microsoft and the consumer must pay to achieve it.  http://msdn.microsoft.com/library/default.asp?url=/library/en-us/csext/html/csconExtensibilityOverview.asp  The Object-Oriented Database System Manifesto  Malcolm Atkinson, University of Glasgow; FranÃ§ois Bancilhon, AltaÃ¯r; David DeWitt, University of Wisconsin; Klaus Dittrich, University of Zurich; David Maier, Oregon Graduate Center; Stanley Zdonik, Brown University  SPIN Operating System  Brian N. Bershad Stefan Savage Przemyslaw Pardyak Emin G\"un Sirer Marc E. Fiuczynski David Becker Craig Chambers Susan Eggers  Department of Computer Science and Engineering University of Washington at Seattle  Citationsfrom SPIN Operating System [Schulman et al. 92] Schulman, A., Maxey, D., and Pietrek, M. Undocumented Windows. Addison-Wesley, 1992. [Draves 93] Draves, R. The Case for Run-Time Replaceable Kernel Modules. In Proceedings of the Fourth Workshop on Workstation Operating Systems, pages 160-164, Napa, CA, October 1993.   '),(5583,3767607,'01,July,2004','2004-07-01 00:00:00','    DISTRIBUTED KNOWLEDGE   '),(5584,3767607,'01,July,2004','2004-07-01 00:00:00','    EPPS (ELECRONTIC PERFORMANCE SUPPORT SYSTEM)     And now, ladies and gentlemen, the  Lord of the Blog  is pleased to present . . . the EPPS, the Webâ€™s hit system.  Or, at least itâ€™s one of the systems that can be used electronically, anyway.  An EPSS key features are its electronic nature, immediate accessibility, and. . . hey.  If you want to know more, youâ€™ll just have to read extended definition.       What is an Electronic Performance Support System (EPSS)?  If we simply look at the words â€œElectronic Performance Support Systemâ€ we can begin to form an idea of what it meansâ€”an electronic system that supports performance.  Of course this is almost too simple a definition, and a bit more detail is needed in order to fully understand what EPSS is.    Starting with the root of EPSS, we will begin our discussion with a definition provided by Gloria Grey, the individual given credit for originating EPSS. According to Grey, EPSS is â€œan integrated electronic environment that is available to and easily accessible by each employee and is structured to provide immediate, individualized on-line access to the full range of information, software, guidance, advice, and assistance, data, images, tools, and assessment and monitoring systems to permit job performance with minimal support and intervention by othersâ€ (Sleight, 1993). Iâ€™d like to point out several key words addressed in Greyâ€™s definition because they will be reoccurring themes as we move on to further discuss the definition of EPSS. The words â€œelectronic,â€ â€œimmediateâ€ and â€œpermits job performanceâ€ typify the essential elements of EPSS.   The definition above mentions â€œelectronic;â€ however, the broad term electronic leaves much open to interpretation. The following definition taken from Hyperdictionary differs from that above in that it further defines â€œelectronicâ€ by listing the types of information EPSSâ€™s display such as text, graphical displays, sound and video. EPSS is described as â€œA system that provides electronic task guidance and support to the user at the moment of need. EPSS can provide application help, reference information, guided instructions and/or tutorials, subject matter expert advice and hints on how to perform a task more efficiently. An EPSS can combine various technologies to present the desired information. The information can be in the form of text, graphical displays, sound, and video presentations.â€   Following closely with Greyâ€™s idea of EPSS, the following definitions also explains EPSS as an electronic system that provides the user with instant information needed to efficiently perform a task. According to HPT Quick Reference Guide, â€œThe Electronic version of Performance Support Systems is one of numerous tools of the technology age that provide workers with almost instant access to information vital to do their jobs.â€  Similarly, a source sited in an essay by Tony Mostek defines EPSS as â€œAn integration of artificial intelligence technologies, hypermedia, and computer-based training to produce a system whose components include embedded training, hypermedia help, artificial intelligence based coaches, and adaptable, model-based user interfaces.â€ (Mostek)   Although most of the definitions of EPSS specifically stress the electronic portion, the definition below takes a more generalized approach in describing EPSS in that no mention is made of electronic or technologies used. In fact, the definition is so broad that it could even apply to a paper-based system provided it offered problem solving and learning opportunities immediately available to the user. Another source identified in Mostekâ€™s essay describes EPSS as â€œA human activity system that is able to manipulate large amounts of task related information in order to provide both a problem solving capability and learning opportunities to augment human performance in a job task by providing information and concepts in either a linear or nonlinear way, as and when they are required by the user.â€ (Mostek)   What can we discern from the definitions weâ€™ve discussed? In short, EPSSâ€™s consist of or perform the following: â€¢	Are Predominantly Electronic â€¢	Are Immediately available to the user â€¢	Assist user in performing tasks â€¢	Present information in forms such as software guidance and assistance, data, images, tools, computer based training, subject matter expert advice and assessment and monitoring tools â€¢	Utilize technologies such as hypermedia, text, graphical displays, sound, and video      Resources  Albert Sleight, Deborah, (1993). What is Electronic Performance Support and What Isnâ€™t?. Retrieved June, 29, 2004, from  urlLink http://www.msu.edu/~sleightd/epssyn.html . Electronic Performance Support System: Dictionary Entry and Meaning. Retrieved June 29, 2004, from  urlLink http://hyperdictionary.com/dictionary/Electronic+Performance+Support+System . An HPT Quick Reference Guide. Retrieved June 29, 2004, from  urlLink http://www.greenworks.org/hpt/non_instructional.htm . Mostek, Tony. Electronic Performance Support Systems (EPSS)-Definitions. Retrieved June, 29, 2004, from  urlLink http://www.comet.ucar.edu/~mostek/ep/epdef.htm .   '),(5585,3767607,'01,July,2004','2004-07-01 00:00:00','    CRM (CUSTOMER RELATIONSHIP MANAGEMENT)     Colloquial Raging Musicians? Crazy Rabid Monsters?  Uh . . . no.  But good guess.  CRM stands for Customer Relationship Management and this practice is as vital to many organizations as the Colloquial Raging Musicians are to the underground classical music crowd.  Learn more about the multiple definitions of CRM as you read this extended definition.     I picked this term because I just rolled off a project in which we assessed 10 of our organizational capabilities, one of which was Customer Management Service (CRM).  I discovered that the definition we used was only a sliver of what CRM means to the rest of the planet.  Basically for us, CRM is about help desk for internal customers to deal with their issues so that our field personnel donâ€™t have to can spend more time managing . . . customer relationships!   Okay, true confessions of a graduate student: My first stab at researching CRM entailed typing â€œcrm definitionâ€ into my favorite search engine.  The first hit on the page was CRM Information.com!  CRM Information.com ominously stated, â€œWhat exactly is the definition of Customer Relationship Management.  Ask a dozen professionals, get a dozen different definitions.â€  Further research on my part validated this claim.    Reciprocal relationships  CRM Informaiton.com posits a fairly generic definition which emphasizes the reciprocal relationship between company and customer: â€œCRM is used to learn more about your key customers needs in order to develop a stronger relationship with them.â€  Bland as this definition may be, it pretty much covers the key points.  The site goes on to elaborate that CRM leverages technology to capture data about customers that drive activities aimed at growing and retaining customer base.  The site also highlights the fact that CRM focuses on enhancing long-term relationships with customers as a key strategic impulse.  Although customer-orientation is as old as dirt, CRM adds technology as a key leverage point between the company and the customer.   As Ephraim Schwartz state sin â€œCRM Gets Strategic,â€ â€œit looks as if CRM is coming back with a vengeanceâ€”due in a large part to the enterpriseâ€™s unprecedented rediscovery of that always distasteful creature: the customer.â€     Zachry et al. pick up the reciprocal relationship perspective in â€œThe Changing Face of Technical Communication,â€ indicating that CRM narratives have two recurrent themes 1) â€œestablishing long-term, mutually beneficial relationships between companies and customersâ€ and 2) â€œinnovative uses of database technologies and their corresponding interfacesâ€ (250).     In â€œManaging Customer Relationshipsâ€ by Lisa Cross, the author positions data as the â€œfundamental factorâ€ to CRM and places customers at the heart.  Crossâ€™s simple CRM model, however, states that for a CRM strategy to be successful, the entire company has to â€œadapt a customer-centric philosophy and to change its corporate culture where neededâ€ (52).   The result of such dogged customer-focus will improve customer retention, ease competitor pressures, and differentiate the company in the customerâ€™s eyes through superior customer service (52).      Advantaged relationships  An article from Information Week, â€œCRM Tools Offer Sales-Force Solution,â€ focuses on improved sales as the ultimate end of CRM.  In this case, â€œmutually beneficialâ€ is a bit one-sided in favor of the company.   Here the data on customers helps the sales-force to better manage the sales pipeline.  The real value in CRM technology in this instance is valued in its ability to integrate and crunch vast amounts of customer data quickly and consistently.  If a customer communicates a change in their account to one part of the company, the sales person should be aware of that change when they walk into the customerâ€™s place of business.        Another company-advantaged view on CRM is from Ira Brenskinâ€™s article, â€œRust Belt CRM,â€ which concentrates on streamlining sales efficiency by using customer data to tightly segment accounts.  Specifically, CRMs allow a company to 1) â€œmore accurately forecast needs,â€ 2) better â€œavoid too many low-profit or high-risk bids,â€ and 3) â€œto identify cross-selling opportunities.â€  An interesting result of this skewed take on customer relationship building is the behavior modification factor it imposes on customers.  Brenskin points out that companies can use CRM data to give customers the â€œchance to modify the product design or production requirements to cut costsâ€”or face higher prices.â€      Multi-layered customer-orientation  Sam S Adkins, in â€œThe Brave New World of Learningâ€ published in ASTDâ€™s T&D, highlights the emerging trend of workflow strategies and the technologies that can actualize such strategies.  By workflow strategies, Adkins is gesturing at the world of just-in-time interventions which avoid interrupting work productivity.  In such a world, â€œlearning is experienced as a by-product of real-time collaboration with people and machines in the context of workflowâ€ (31).   Although Adkins speaks specifically to training and development, he does mention CRM several times and underscores wide-breadth implications.  Specifically, CRM is one multiple TLAs (three-letter acronyms) that can help deliver workflow solutions.  In this case, the customer becomes internal to company.  A better-served internal customer can then better serve the external customer.     Summary  All in CRM spans the customer relationship spectrum from reciprocal customer-company relationships, to company advantaged customer relationships, to multi-layered customer-orientation.  In all the instances, data places an intricate role in understanding the customer.      Works Cited (documents attached)     Adkins, Sam S.  â€œThe Brave New World of Learning.â€  T&D. June 2003.   Breskin, Ira.  â€œRust Belt CRM.â€  Computerworld.  December 15, 2003.   CRM Information.com.   urlLink http://www.crminformation.com/   Cross, Lisa.  â€œManaging Customer Relationships.â€  Graphics Arts Monthly. October  2001.   Informationweek. â€œCRM Tools Offer Sales-Force Solutions.â€ August 21, 2000.  Schwartz, Ephraim.  â€œCRM Gets Strategic.â€ Infoworld.com.  April 26, 2004.   Zachry, Mark et al. â€œThe Changing Face of Technical Communication: New Directions for the Field in a New Millennium.â€ SIGDOC â€™01. October 21-24, 2001.   '),(5586,3767607,'01,July,2004','2004-07-01 00:00:00','    CHECK OUT/CHECK IN     No, this does not refer to your state of mind between the time your physics professor begins and ends the lecture.  This extended definition of check out/check in explains this practice of version control in the document production process and the various manners of implementation.    Most version control processes, whether sophisticated or manual, involve some sort of check out, check in arrangement. Some processes require manual intervention from a â€œlibrarianâ€ of some sort; some processes are entirely automated.   It is probably best to make it clear (when applied to documentation projects) that check out, check in is a part of version control, which is a part of content management, which is a part of publishing. Itâ€™s a process within a process, which means that there are many factors that might change the intricacies of the process, and trying to define the process for every one of those factors is impossible. The definitions described here are my best at a high level interpretation of what check out, check in is and how it relates to version control, content management, and publishing.   Documentation teams use check out, check in in several ways. Most of the time, a check out, check in process is necessary when developing documentation (any sort of information product, including books, guides, online help systems, websites, web pages, etc.) in a collaborative environment. That is, individuals on a team working together on the same files or pages.    The following check out, check in approaches may be used this sort of team environment:   â€¢	 You canâ€™t have it until Iâ€™m finished  This sort of system allows only one person to have a particular file at any given time, preventing team members from overwriting each other\'s changes. When someone needs to work in a file, they check it out (this may be a manual or automated process), and while they have it no one else can access it.   â€¢	 I have it right now and you better leave it alone  This approach does not prevent someone from opening a file that is checked out. They may receive a warning before doing so, but some automated check out, check in systems (like the one within the Dreamweaver software) allow others to override the warning and open a file anyway.  One can clearly see the innate problems within this process. The Macromedia website addresses potential problems by stating â€œTo maximize the effectiveness of the Check In/Check Out feature, the collaborative team members must use it properly.â€ Iâ€™d say. A process like this one requires strict adherence to previously set standards and guidelines, otherwise the version control system it is a part of will completely fail.    â€¢	 Letâ€™s share and compare   Some automated check out, check in processes allow two or more people to check out the same file. When the writers check the file back in, the system compares the file, calculates the differences from the last version and applies them. If there is a conflict, it stops the writers from checking files back in, and forces them to manually reconcile the differences. I guess this system would work okay if the two writers were working in different sections of the document, but I can see where it could become a nightmare.   â€¢	 Itâ€™s mine and Iâ€™ll check it in whenever I want  This approach deploys only half of the check out, check in process. Some teams divide files between team members, who in effect â€œownâ€ their individual set of files. Each writer is responsible for his or her own working directory, so two team members working on the same file is never a problem. At designated intervals, the writers then submit their files in to an official directory, build machine - or some such location, where the â€œofficialâ€ files live. This of course can pose its own set of problems, previously discussed in the â€œYouâ€™re your own Grandpaâ€ discussion in my Version Control submission for Content Management Concepts I.   Almost every version control software supplier out will say something like â€œCheck In/Check Out procedures are not by themselves a content management system.â€ (Clavister AB.)  Version control, too, means much more than simply making sure files donâ€™t get overwritten as a result of more than one person working in a particular file. But having good check out, check in procedures is one part of an overall content management system that need not cost thousands of dollars to implement. All it takes is a well-defined process and the teamâ€™s strict adherence to that process.    References     ----. 2004. â€œCheck In/Check Out: managing a Web site in a team environmentâ€. Macromedia.  Available  urlLink http://www.macromedia.com  .   ----. 2004. â€œSharing files between team members.â€ Macromedia.  Available  urlLink http://www.macromedia.com  .   ----. 2003. â€œThe Check Out and Check In Concept.â€ Clavister AB. Available  urlLink http://www.clavister.com  .   '),(5587,3767607,'01,July,2004','2004-07-01 00:00:00','    ASSOCIATIONS (HIERARCHY & TAXONOMY)     â€œIsnâ€™t taxonomy the art of stuffing dead animals?â€ you might ask.  Well, not quite, though your characterization of the word â€œtaxonomyâ€ can serve as an association with the term taxidermy.  In this extended definition by Ms. Alison Leyda Straquadine, associations are discussed in detail from their origins to their present-day well, associations with hierarchy and taxonomy.    There are many definitions of association. Following are the definitions of association I consider most relevant to a discussion of content management.  Something linked in memory or imagination with a thing or person  the process of forming mental connections or bonds between sensations, ideas, or memories Empiricism, a theory that all knowledge originates in experience, started with Aristotle. Aristotle accepted the concept of association (connection between ideas/events) as the process through which experience is converted to knowledge. When people associate ideas and/or events, one triggers the other.   Taking association beyond the knowledge stage, the theory was further developed by philosophers and learning theorists to become one of the two main behavioral learning theories. Association is a behavioral theory because it holds that memory and behavior are stimulated through associative ideas and events.  Association as a connection between ideas and events is related to connectionism, contiguity, and situated learning theories. E Thorndike created the theory of connectionism which holds that we learn by forming associations between stimuli and response. E. Guthrieâ€™s contiguity theory gets into the behavioral domain. Contiguity theory holds that a stimulus accompanied by movement will result trigger the same movement when it recurs. (The stimulus response theory led to Pavlovian research and the behavioral tradition).  James Mill defined contiguity as complex ideas made up of simple sensations that have become association with each other. Because the more we experience contiguous (close together in time or space) events the more strongly we associate with them, contiguous events can change behavior. Consistent with contiguity theory which holds that learning involves the conditioning of behavior, instruction must present specific tasks ( urlLink http://www.educationaau.edu.au/archives/cp/04b.htm ).  J. Lave came up with situated learning theory which holds that we learn through social interaction. The focus of situated learning theory is on â€œlearning by doingâ€  urlLink (http://otec.uoregon.edu/lerning_theory.htm) . On this website, I found Greg Kearsleyâ€™s discussion of Jean Lave particularly relevant to the social interaction and learning that takes place in our online discussions in this class. In it, Kearsley states that learners become involved in a community of practice which embodies certain beliefs and behaviors to be acquired. As the beginner or newcomer moves from the periphery of this community to its center, they become more active and engaged within the culture and hence assume the role of expert or oldtimer. I found Kearsleyâ€™s statement relevant to class community because I can see myself moving from the periphery to the center as I become more familiar and comfortable with how the online seminars work.    Also related to association theory are hierarchy and taxonomy. Maslowâ€™s hierarchy of human needs which evolves from physiological, safety/security/protection, belonging/love/affection, and self-esteem to self-actualization, can be applied to learning theories and teaching strategies. A humanistic orientation to Maslowâ€™s hierarchy holds that as people fulfill each subsequent need on the hierarchy, the are motivated to fulfill th next. Therefore, people who have fulfilled the lower level needs of security and affection are motivated to fulfill the higher level needs of self-esteem and self-actualization through learning.    Taxonomy is defined as the study of the general principles of scientific classification. In Benjamin Bloomâ€™s famous taxonomy of learning domains, he classifies learning into three domains (cognitive/mental, affective/emotional, and psychomotor/physical) which he further divides into hierarchical step staring with the simplest and progressing to the most complex. The cognitive domain begins with knowledge (recall of data) and moves from comprehension (interpretation), application, and analysis, to synthesis, and finally to evaluation. Bloom developed the classification and hierarchy of the cognitive domain, and his buddy David Krathwohl developed the classification and hierarchy of the affective domain. The guy that was supposed to develop the psychomotor domain with Bloom and Krathwohl didnâ€™t do it, but others have developed it since.     In Taxonomy of Educational Objectives: The Classification of Educational Goals, Handbook 1: the Cognitive Domain, Bloom devotes an entire chapter to defining the difference between taxonomy and classifications and discusses the problems with hierarchies. Bloom decided to go with a hierarchy to develop the cognitive domain because each area tended to build on behaviors from the previous class.   Familiarity with the theories of association (connections between ideas), hierarchy (information that builds from previous information) and taxonomy (classification schemas) may become more useful to us as we get into the content reuse, maps and information architecture aspects of our content management plan.   '),(5588,3767607,'01,July,2004','2004-07-01 00:00:00','    ASSET RESPOSITORY    What is an asset repository?  You mean you donâ€™t know?  Well, hereâ€™s the beef.  An asset repository is similar to a bank depositoryâ€”it stores valuables, although in the case of asset repository, it stores valuable information.  These repositories are high class, elaborate, and require the same thing that banks doâ€”lots and lots of money.  Read more about asset repositories in the following extended definition.     â€œAn asset is only an asset when you can find it,  or you know that you have it in the first placeâ€   ï€­Teri Ross, Techexchange.com    An asset repository is rather a combination two elements: digital asset management and asset repository.  Whereas the digital asset management (DAM) is a tool to organize digital media assets for storage and retrieval, an asset repository is where and how the data is stored, typically on â€œa high-end UNIX server, [with] large online storage space and high-speed networksâ€ (Ross).  The DAM application organizes and retrieves the digital media and the asset repository stores it.    Digital Asset Management   Digital Asset Management (DAM) is a means to keep, find and secure files of text, images or sound.  DAM also tracks ownership and rights of the content and implements task management.  Well organized and accessible content results in the ability to quickly find images and data, which systemizes workflow, saving time and money (Brown).   In the computer world . . . there are systems that will pick up after us and remember where we put things. These systems will retrieve items, let us pass them around, and then clean the place up when we\'re through. I refer, of course, to the software and hardware systems that are called, variously, digital media management, media asset management or digital asset management (DAM) (Smith).  Digital asset management is the software and hardware needed for content management.  The location in which content is stored is the asset repository.   Asset (or Content) Repository  Actual â€œasset repositoriesâ€ are generally found in large, high-end asset management systems ($100,000 and up).  Asset repositories are a â€œsecure collection of the source files themselvesâ€.  The image at the webpage below shows the asset repository as a sort of holding tank for content.   urlLink http://www.assetnow.com/index.cfm/4,437,html   A large asset repository performs like a database.  It contains: security, replication, backup safeguards, disaster recovery, referential integrity, centralized data management and a hierarchical storage structure. â€œSuch a system lends itself to workflows that require access security, including management of rights or permissions and access for suppliers or customers. These systems require a high-performance server, a fast network and significant online storage capacityâ€ (Brown).  Asset repositories are sophisticated and expensive systems that contain the actual content within a secure database.  This usually means higher performance software like , â€œhigh-end UNIX servers, formidable online storage, and very high-speed networksâ€.  This allows for â€œsecurity levels, replication, referential integrity, and centralized data management. Also included is the comfort of full hierarchical storage management and disaster recoveryâ€ (Ross).     Solutions based on the asset repository model are ideal when systematizing studios with industrial workflow, managing rights and permissions (such as the intellectual property of either your company or a third party), and structuring global access by employees, contractors, suppliers, partners, and customers (Smith).   Media Catalog vs. Asset Repository  Instead of an asset repository, smaller companies often use media catalogs â€œthat store proxies alone, indexed to the source files\" (Brown).  Media catalogs have proxies (like thumbnails) which are stored in a database indexed by keywords.  The actual source files are under the control of the operating system and are not touched.  Media catalogs are low cost, easy to install and administer and have â€œscalability across multiple divisions of an enterpriseâ€ (Ross).  Media catalogs donâ€™t manage the actual content, so anyone with system access to view, change, move or delete content elements (Ross).     Media Catalog  Entry level solution Desktop or within a workgroup Organizes small collections of images, graphics and text Provides multiple access Cost â€“ under $50,000   Asset Repository  Mid-range solution Client/server technology Wide area networks Manages all assets used in a production job (layouts, images, fonts) Files stored on server, access via DAM system only Check-in/check-out to control versions Stores assets with links for simple retrieval Cost â€“ under $100,00   Upper-End Solutions 3-tier client/server architecture Designed to â€œdrive enterprise on the Internetâ€ Cost â€“ over $100,00 (Smith)   '),(5589,3607694,'20,June,2004','2004-06-20 00:00:00','    I\'m finally seeing the beginning of some brown algae. See, when you first set up a tank, brown algae will begin to grow, and then the brown algae will turn to green algae. Once the brown has turned to green, the tank has been fully cycled, and the water quality is great. I guess the extra fish in there moved the process forward because there are little patches of brown algae sprinkled all over the tank. Which is great because that means the cycle has finally started!   '),(5590,3607694,'19,June,2004','2004-06-19 00:00:00','    Well, it\'s been a week since I put the two chromis in there, two weeks since the damsel. So, I took a sample of my water to the pet store today to get it tested. He said that my cycle hasn\'t even started yet. So strange! But they are small fish in a large tank, and he said it wasn\'t that strange. There wasn\'t any ammonia in there at all...So, I\'m going to take a sample back in another week, and see what it says then. I also don\'t really need to do a water change yet. So, I might wait another week until I do that. The fish store guy also told me that for a tank my size, that a 20% water change every two weeks would be enough. And since I don\'t have any ammonia, I could wait a bit.  And...I got another chromis! I just love those guys. So, now I have three. I was a little worried when I first put him in there because one of the ones already in there (I\'ve named him Bilbo), didn\'t seem to like him. That was almost 8 hours ago though, and they seem fine now. I suppose it was just that he was a new fish and they had to get used to him being there.  So, now I have yellow-tailed damsel(Spike) and three chromis(Iggie, Bilbo, and Gordo). And don\'t forget my piece of live rock!   '),(5591,3607694,'16,June,2004','2004-06-16 00:00:00','    The fish are still doing great. They are all getting along. I kind of didn\'t expect them to just because of the nature of damsels. The chromis sure do like to eat. They\'ve already figured out when I\'m about to put some food in there and they hover around the place where I put it in. And then it only takes them like two seconds to snatch it all up. It\'s a little bit of a problem because it\'s hard for Spike to get any food. I have to distract them to the other side of the tank to keep them from eating all of the food. I\'m going to do my first partial water change sometime this week. I\'m kind of nervous because I don\'t want to mess anything up. They are doing so good! But in order for the cycle to go smoothly and to keep them healthy, I need to do one soon.   '),(5592,3607694,'12,June,2004','2004-06-12 00:00:00','    The fish are doing great! I got up this morning and fed them. They went crazy when I put the food in there. Spike(the damsel) doesn\'t bother the chromis at all. Actually, he hangs out with them a little bit. Speaking of Spike, he was digging around in the rocks last night. Behind this piece of shell, he was picking rocks up in his mouth, swimming out from behind the shell, and spitting them out. He was making himself a little home! I\'ve decided I\'d like to get one more green chromis. I really like them, and I think it would be better if I had a group of them. I might get one next week when I go get my water tested at the fish store.   '),(5593,3607694,'11,June,2004','2004-06-11 00:00:00','    Well, I\'ve been watching my new fish, and it seems like they are going to get along fine. The smaller chromis is following the bigger one around. I\'ve named the smaller one Iggie. :) As for the damsel, I\'ve named him Spike because he has little spike-like things that prick up sometimes. Spike isn\'t messing with the other two, and hopefully he won\'t start to. I haven\'t decided on a name for the bigger chromis yet, but I\'ll think of one. :)   '),(5594,3607694,'11,June,2004','2004-06-11 00:00:00','    Well, I have just started a saltwater aquarium. It\'s a 65 gallon tank that came with an oak stand/cabinet. I got a power filter that hangs over the side, a heater, and a hydrometer. I haven\'t gotten a powerhead yet, but I plan on getting one of those in the next few weeks. I\'m also going to get a protein skimmer at some point. I got some crushed coral for the substrate. Cleaning that stuff was no fun. And even after I cleaned it, when I put the water in the tank, it made everything all cloudy. I got a few pieces of rock(not live rock) and some sea shells. I left it running for a week, made sure the specific gravity was right and all that mess.  So, a week ago, I went to purchase my first fish. I had read up and I thought a couple of damsels would be perfect for cycling my tank. WRONG! One was bullying the other one terribly. He was bigger than the other one, and would constantly chase him, and not let him swim around in the tank. So, I took the bigger one back. And, I just came home with two green chromis, and a piece of live rock. The fish store guy said the live rock would help the cycling process. Right now, the fish are getting along great, especially the two chromis. They love each other. I don\'t know how the yellow-tail damsel is going to get along with them, but it seems like it will be okay.   '),(5595,3607694,'25,July,2004','2004-07-25 00:00:00','    I\'m starting to get really impatient. I really want to get something new to put in my aquarium(besides live rock). I talked to the guy at the fish store, and he said that I really shouldn\'t put anything in there until the cycle in completely finished...and that\'s probably going to be a little while since my tank is so big. I really want some crabs, snails, and shrimp. I don\'t know why, but I just think it would be awesome to have all sorts of little creatures like that living in there. But, unless the cycle is done, they would probably die. They are really sensitive to water quality. And of course, no new fish until it\'s done unless I wanted to take one of mine in to trade it for another. But never! I love those fish I have too much to do that. :) Oh well, hopefully the cycle will be over soon.   '),(5596,3607694,'21,July,2004','2004-07-21 00:00:00','    It\'s almost been a month since I last wrote. Oh well. Things are going well with the tank, although it still hasn\'t cycled completely yet. I\'ve taken a sample to the fish store every week. It\'s in the second stage of the cycle. I\'m starting to see a few specks of green and red algae as well. I\'m really hoping that the cycle will be over soon. I can\'t wait to get my next fish, plus some intvertebrates...some snails and crabs...and a shrimp. Of course, I\'ve done a fish tank cleaning twice now. I wiped down everything with excess salt on it, cleaned the glass of algae, and changed the water. It was kind of tedious, trying to get the water the same temperature as the water in the tank, plus getting it the right specific gravity.  The fish are doing great though. I can\'t wait to get another one.   '),(5597,3607694,'02,August,2004','2004-08-02 00:00:00','    Well, the tank in finally cycled! I\'m so excited. So, Saturday, I went to the fish store and got a few new things for my tank. I got some live rock, of course, because I still need a ton of that. I also got a Camel Shrimp, named Mr. Whiskers. :) He\'s cool, but he doesn\'t wander around much. He likes to hide behind things. I\'ve seen him walking around at night when the light is off. But not much other than that. I also got a new fish! It is a very pretty Scopus Tang, named Sebastian. He\'s beautiful. I have some green algae growing on my live rocks now, and he likes to nibble at it. Anyway, I can\'t believe how great things are going.   '),(5598,2862007,'29,April,2004','2004-04-29 00:00:00','   Today I turned in my thesis.  I never want to write another thesis as long as I live.  5 years of classes down.... 4 finals to go...  '),(5599,2862007,'28,April,2004','2004-04-28 00:00:00','   Good news.  Currently 5 of my 7 vehicles are running.  I think it is a new record.  And all of those 5 are even plated and everything.  What would be a fun interesting thing to do to the Bronco before it goes to the scrapyard?  I\'m up for anything.  '),(5600,2862007,'27,April,2004','2004-04-27 00:00:00','   Hey Family!  Tatoo Day is tentatively scheduled for Sat., June 5. Let us know if you want to be included.  Kari  '),(5601,2862007,'27,April,2004','2004-04-27 00:00:00','   Colin has been named to the Indiana All State Marching Band and will participating in the Indianapolis 500 Festival parade on May 29. The parade begins at Noon for anyone interested in coming to watch.  For those of you who haven\'t heard, Cassi was a member of the Winter Percussion, and they were the 2004 Indiana Percussion Association State Concert Champs!!!   Woo Hoo!!!! :)  '),(5602,2862007,'25,April,2004','2004-04-25 00:00:00','   Here are some pictures of Ariel in Australia.  Nobody in that portion of the Lewis family has bothered to join this blog, so it\'s up to us! :p  '),(5603,2862007,'22,April,2004','2004-04-22 00:00:00','   Hmm... not sure what happened to the comment links.  They seem to have disappeared.  '),(5604,2862007,'22,April,2004','2004-04-22 00:00:00','   Hello again, updates on me...I went to the doctor yesterday and everything is awesome with me.  Although we talked about it and I can\'t ride rollercoasters anymore :(  I can however get a tatoo which is the one thing I wanted out of the appointment.  So family that is in on it with me...now you have to think about what you\'re getting. Betcha didn\'t think I was going to be able to get one did ya? Oh and if I want I can have one celebratory drink on my 21st birthday but I don\'t care about that as much as the tatoo. So I\'m a happy person right now.  Bye  '),(5605,2862007,'22,April,2004','2004-04-22 00:00:00','   Nick requested I post these pics of Alex and Chase, so here they are for your viewing enjoyment!        Also, there doesn\'t seem to be an easy way to post pictures to this, so if you want to have a picture put up, just e-mail it to me and I\'ll get it up here for you!  '),(5606,2862007,'21,April,2004','2004-04-21 00:00:00','   Check out www.cedarpoint.com to see their new Top Thrill Dragster roller coaster. It\'s 420 ft. tall. On the left side of the page, go to live park cams. It should be fun to watch when the park opens. I think we should have a family reunion there. Kari  '),(5607,2862007,'21,April,2004','2004-04-21 00:00:00','   Woot. After much confusion and hard work on the part of Brittany I am able to post on this here family blog. My life is very fast paced as of right now. I go to school and then straight to practice for the musical which I\'m student directing. It\'s fun but has some long hours. I clock in about 13 hours at school a day. Which isn\'t bad considering my personal record for straight hours at the school is 17. Staying at school past midnight for the spring play is great fun. School stinks though. It needs to end before I pull my hair out. Meh. Off to the homework.  '),(5608,2862007,'20,April,2004','2004-04-20 00:00:00','   Hello! I\'m on the blog. News with me is school is still 5 weeks from being done. I don\'t have much work to do that is tough so that\'s nice. I did most of the hard stuff in march and now I\'m just reworking on the same stuff until the end. For now that\'s all but I\'ll keep posting.  '),(5609,2862007,'20,April,2004','2004-04-20 00:00:00','   Hey CassI.  (Blush) it is pretty bad when I spell family names wrong.  Sorry about that!  Way to go with the drums.  '),(5610,2862007,'19,April,2004','2004-04-19 00:00:00','   Have no fear, your Cass-i is here! So yeah, uh family...my news consists of, i\'m in high school as you know, i don\'t think my brain can hold any more information, i detest my english teacher, even more so for teaching me that word...i just go around doing random things of randomness and that concludes me. if anything of remote interest happens to me, i\'ll be sure to post it. the only interesting thing that i have is that my winter percussion thing won state...and that\'s all i got, we\'re doing a community night on the 26th, and drum enchanted evening the 3rd week of may. that is all my life has in it right now. so till next time, this is me, signing off. Much love to the family!  '),(5611,2862007,'17,April,2004','2004-04-17 00:00:00','   Wow, for a Family that really talks alot, we sure are quiet!  Where is everyone?  Kevin, Cole, Elyse, Cassi, Collin, Ariel, Harry etc.... where are you?  '),(5612,2862007,'17,April,2004','2004-04-17 00:00:00','   Congratulations, Brittany!!  This is a great idea! Nothing to really report...just Hello!!  '),(5613,2862007,'17,April,2004','2004-04-17 00:00:00','   Go Brittany!   Hello to all my northern family.  '),(5614,2862007,'16,April,2004','2004-04-16 00:00:00','   For those of you who may not have heard, Brittany was named the Outstanding Graduating Senior of the Business Dept. at Ball State!!!!  I\'m very, very proud of her!!!  Maggie  '),(5615,2862007,'16,April,2004','2004-04-16 00:00:00','   Hey Family Blog!  Just a quick hello and am happy we have this blog thing figured out.  Will try and get Granma Dena and/or Grampa Harry to post something.  Love you all! Kari & Jim  '),(5616,2862007,'16,April,2004','2004-04-16 00:00:00','   Hey Family!  I think this Blog idea is great.  Great idea Brittany! And cheers to Chase!  My 3rd nephew.  I am starting to think it is time for more girls around here.  I am off to try to figure out how to post pictures and stuff.  Be back later.  January  '),(5617,2862007,'15,April,2004','2004-04-15 00:00:00','   Since I don\'t have any news at the moment, I am taking up space by welcoming Chase! (rhyming as I do)  '),(5618,2862007,'15,April,2004','2004-04-15 00:00:00','   This is the new family blog.  I hope everyone can find some use for it, and I hope you come back and post when you have something to say... or even when you don\'t.  '),(5619,2862007,'11,May,2004','2004-05-11 00:00:00','   Woah, looks like they changed around Blogger a bit!  Anyway, I just wanted to put up some pictures from my graduation.  Thanks to all the family who came to see me!  '),(5620,2862007,'04,May,2004','2004-05-04 00:00:00','   Howdy everyone! Not much new here. Chase is 6 weeks tomorrow and the poor boy is fitting another cold, which the 4 of us seem to be passing around to each other. Nick is working hard in his new position, which he really enjoys. I\'ve started to work from home part-time, which I\'ve found to be much more difficult than I imagined. Especially with a sick infant!! Alex is a great big brother and we\'re very proud of him.  Congratulations to Cassi, Colin, Cole, Kevin and Brittany!! (Wow! Lots of good news to be celebrated!)  I\'m sure the boys and I will be up North for a visit soon. We\'ll keep those of you up North updated on that!!  Steph  '),(5621,2862007,'28,June,2004','2004-06-28 00:00:00','   I need Tattoo Day photos!  If I can get a copy of them, I will post them here on this fabulous blog.  '),(5622,2862007,'27,June,2004','2004-06-27 00:00:00','   WELCOME HOME ARIEL from your most excellent adventure! Can\'t wait to hear all about it.  Although I\'m not sure it measures up to our putting the 18,223 layer of paint on the world\'s largest paintball! We did that yesterday to finish up my Karipalooza weekend. And we also saw the World Rosary of Peace made out of bowling balls. Only in Indiana! Love, Uncle Jim and Aunt Kari  '),(5623,2862007,'27,June,2004','2004-06-27 00:00:00','   HAPPY BIRTHDAY ALEX. WE LOVE YOU!!  AUNT MAGGIE, BRITTANY, ELYSE, CASSI, COLIN  '),(5624,2862007,'25,June,2004','2004-06-25 00:00:00','   HAPPY 2ND BIRTHDAY ALEX! HAVE LOTS AND LOTS OF FUN! LOVE, UNCLE JIM, AUNT KARI, COUSINS COLE & KEVIN  '),(5625,2862007,'25,June,2004','2004-06-25 00:00:00','   HAPPY  50TH  BIRTHDAY  KARI!!!!  '),(5626,2862007,'13,June,2004','2004-06-13 00:00:00','   After comparing all of the open house food this year up until now, I have the say the best was at Brittany\'s.  Hooray for pretzel rolls and hot wings.  Kevin  '),(5627,2862007,'08,June,2004','2004-06-08 00:00:00','   We had a great time at Family Tattoo day. The crew at Burning Image is wonderful. Two guys came in on their day off to help out. We were the first family to come in. We\'re going to be featured on their web site. Some young kids were in and Gramma Dena was going back to get hers and one girl yelled \"She\'s getting one?\" (the girl\'s mom wouldn\'t let her get one). We have some awesome tattoos. Hope we\'ll be able to get some pictures up. Kari  '),(5628,2862007,'04,June,2004','2004-06-04 00:00:00','   Tomorrow is Family Tattoo Day! We have nine going. Gramma Dena is really getting into this. She thought about riding over on Kevin\'s motorcycle but the helmet messes up her hair! You can still get in on all the fun. We\'re going at 2 p.m. central.  When we\'re all healed, we\'ll put up the pictures. Love, Kari  '),(5629,2862007,'03,June,2004','2004-06-03 00:00:00','   Have we forgotten about the family blog so quickly?  I don\'t have much to report.  I\'m still on the job hunt, hoping to find something soon.  I hope everyone is doing fine!  '),(5630,2862007,'29,July,2004','2004-07-29 00:00:00','   Hey everyone! I don\'t know who\'s all heard but Cole is moving to Chicago. He got an apartment with a friend from school. He\'ll only be 15 mins. from Dr. Weigel and Children\'s.  We will be empty nesters! We now have a wonderful guest room if anyone would like to visit.  Kari  '),(5631,2862007,'19,July,2004','2004-07-19 00:00:00','   Since Cassi was away when we had tattoo day, she had her own personal tattoo day a few days ago:  '),(5632,2862007,'18,July,2004','2004-07-18 00:00:00','   More Alex and Chase pics:  '),(5633,2862007,'18,July,2004','2004-07-18 00:00:00','   HAPPY BIRTHDAY CHERYL!!! HOPE YOUR DAY IS FABULOUS!  '),(5634,2862007,'12,July,2004','2004-07-12 00:00:00','   TATTOO DAY PHOTOS!!  Sadly, the pictures from Kari\'s camera didn\'t turn out, so we don\'t have all the pictures we\'d hoped for.    Me in the process of being tattoed!    My finished product.    Cole in progress.    Colin in progress.    Colin\'s finished product.    Grandma getting hers done.    Grandma and her artist, Baldo.    Jim\'s finished product.    Kevin\'s halfway done...    Maggie\'s finished product.    Laurie freaking out.  Too bad we don\'t have finished pictures of Kevin\'s, Cole\'s, Kari\'s, Laurie\'s, Grandma\'s, and Rachel\'s, but they\'re all cool!  '),(5635,2862007,'11,July,2004','2004-07-11 00:00:00','   Here are pictures from Kari\'s 50th birthday activities:  World\'s Largest Paintball                                             Rosary of Peace made of bowling balls:               James Dean\'s grave:  '),(5636,2862007,'11,July,2004','2004-07-11 00:00:00','   Here are some Alex and Chase pictures:  '),(5637,2862007,'11,July,2004','2004-07-11 00:00:00','   Oops! Senior moment!  HAPPY BIRTHDAY BRENDA!!!!  Hope you are going to do something that the neighbors will be wondering about!  Love, Your Murray relatives  '),(5638,2862007,'09,July,2004','2004-07-09 00:00:00','   Ariel- I was calling my 50th birthday celebration weekend Karipalooza. Since it was on a Friday, I got a whole weekend to party. Uncle Jim had planned all sorts of stuff and Karipalooza sounded much more fun than 50th birthday. I had already used a fest when I turned 40 (we had fortyfest).  Brittany- We have pictures of the paintball and rosary that a friend had taken from his digital camera and sent to our email.  Can you put them on the Family Blog if I email them to you or do you need it to be on a cd or disk?  Our pictures from tattoo day did not turn out. They all look like they had been taken with a softfocus lens and are all blurry.  We\'re going to a Nascar race today. Jim got free tickets from work and since it\'s in Joliet, only about 1 1/2 hrs. from here, we thought we\'d go. It should be a great people watching day!  Love to all and hope everyone is having a super summer. Kari & Jim  '),(5639,2862007,'09,July,2004','2004-07-09 00:00:00','   I am such an idiot. I wrote a post as a comment to aunt kari\'s post without thinking (obviously). Yes, so... im going to go now and come back with a brain. So... please read comments under kari\'s latest one haha. love, ariel  '),(5640,2862007,'08,July,2004','2004-07-08 00:00:00','   Wow - can you believe Cassi and Colin are 17 now!! Yikes. And this Sunday, Cassi leaves for the Indiana State Police career camp. It\'s just the start of her journey to become a criminal investigator.  I feel so old!!  '),(5641,2862007,'10,August,2004','2004-08-10 00:00:00','   Ralf got a job at Blockbuster, so at least one of us now has an income :)  He also gets 5 free rentals a week, so I predict a lot of movie watching in the future.  Also, Ralf and I drove to Anderson a couple days ago so that I could go to Rax and relive childhood memories, but they were closed on Sundays.  How rude!  When I go back I will write a full report.  LOL.  '),(5642,4213470,'20,August,2004','2004-08-20 00:00:00','   today rocked. life rocks. i\'m feeling optimistic. i cant w8 for my geog test tmrw, amazing rite? i actually studied for it. well thats not excactly why i\'m in a real gd mood... hehe. anyways hi, im the new me. no more vulgarities, no more stealing, no more breaking rulez. the old me\'s real interested 2 c how long i last. actually i\'m interested 2... argh... im turning gay. yes that means i\'m bcoming happy, as in gay=happy.  '),(5643,4213470,'18,August,2004','2004-08-18 00:00:00','   So today was our RE presentation, we had put in lots of effort and were all ready. The lights when out, we shined the spotlights on our presenter, the screen went down... then all of a sudden, the projector turned off (broke down) ! wonder of wonders... aft 6 succesful presentations the projector had decided to sabo us. Down to the main point, i think it was all dalai\'s fault. he used his jinx powers on us, his own group. But even without our slides, we continued. Though we actually cut more than 1/3 of the presentation and presented the rest rather jitterily. But the worst part was definitely that just after we concluded (as in literally, 2 seconds), the projector went on! *sigh* sadness rite? but wat is passed will never come back we arent goin to get another chance, so who cares. Anyway ferdi said to me that he liked our slides when he saw them later, so i guess it will be quite ok.  rite after that was hockey, bull had just found out about the shirts thing... so he screwed us. But i\'m getting quite used to this kinda thing now, i even managed to find some humor in the way he was screwing us! i had to stifle my laughter ever so often. yeah, i guess i\'m turning looney, yay!  '),(5644,4213470,'17,August,2004','2004-08-17 00:00:00','   A few dayz on and Iâ€™m losing interest in â€œmy blogâ€. Haha. School really sucked today, as usual. Anyway zhang mei suo was giving problem again. Its like he planned the test for next Thursday and told us like beginning of last week. Then today, he found out that we are going to be at a camp on Thursday and just changed it to Monday like â€œthatâ€. Its like no freaking time to study. And I need the time cos I take real long to study. Anyways that started me thinking again > my life is Â¼ over, and I havenâ€™t accomplished anything!!! And wat am I doing now??? Iâ€™m studying and putting myself through all the trouble of passing examsâ€¦ and that is useless to me. I wanna be a superstar roadsweeper! So what is the use of studying Chinese? Math? Science? Its just soo exasperating! I dun get to make my own choices, why cant I choose a roadsweeper course and spend my school life slacking like they do in austrailia??? In austrailia the first â€œgradedâ€ test is held when the kids are 18 years old! And that is obviously beneficial to the kidsâ€¦ take ri geps for example, the crÃ¨me de la crÃ¨me or what everâ€¦ of the singaporean students. They take so freaking many test, they are â€œsmartâ€ . But they are all so freaking weird, â€œvegetable brainsâ€ trained to memorize rubbish. Oh yah they are also so cocked up and all! And the moe doesnâ€™t see all this, cos they are made out of this kinda people too! So what do people with different aspirations do? ---- to be continued.  '),(5645,4213470,'16,August,2004','2004-08-16 00:00:00','   Man, I am in trouble again. The t-shirt stealing thing has only recently concluded and I appear once again on the wrong side of law. Mr koh is already really pissed with me, and now some one told him that I threatened aaron about some history test thingy!!! Thatâ€™s soo non-existent. I hope I can resolve this one quickly without getting into any further trouble. If u happen to be the one who told mr koh or maybe you told someone who could have told mr koh, please contact me cause you have made a mistake. Oh yah, if anyone knows who told mr chow about my selling food in class thing please contact me asap. I really wish ri boys had more integrity than this, if a class mate cannot put his trust in youâ€¦ who can?  '),(5646,4213470,'13,August,2004','2004-08-13 00:00:00','   This is my first blog and my first post and i\'m doing it on friday 13th! How jinxed is that...  Anyway, today was our hockey third and fourth placing. So we were dismissed early and met in the canteen, most people werenâ€™t too serious, but i sure wasnâ€™t anyway... i\'ll tell you why later. anyways after warm up we got onto the pitch we played around for awhile. Finally the refs walked to the center of the pitch and the coin was tossed and stuff, the whistle blew. the game had begun, the first half was pretty uneventful though, i got a reverse chop which was on target, but sadly it flew out, besides that the game was pretty even. Second half i picked up my pace, making a few nice dribbles and centering the ball once, then toward the middle of the second half i dribbled past a defender into the D and got a short corner.i dragpushed the ball to chun men who stopped it smoothly and sent me a nice pass to the top D. i took a first timer... the ball hit the goalkeepers pads and went in with a resounding \"bang\". i was filled with joy, bursting with it perhaps, but this wasnâ€™t it, this was not how i used to feel when we won.  This was my first 3rd 4th placing ever, i had always had the good fortune of being in the finals. So why was that not the case this time round? Firstly, of the entire team of 19, 12 were sec1 and only 7 were sec2, in the first team of 11, 7 were sec1 and only 4 were sec2! for the last two years there have only been 1 or 2 sec1s in the first team! Iâ€™d like to congratulate all you sec1s though and I think you guyz played well, it wasnâ€™t your fault that we lost, it was your first major tournament and you guyz were visibly nervous, but it was not your fault. I feel that it was bulls fault. last year he neglected the sec1s causing many to drop out so we had to use many sec1s this year. So very frankly we were weak, people should not have expected us to win anything with our team, infact I was pretty amazed that we managed to make it this far. Iâ€™d like to point out that it hurts alot when people say \"you are the captain of the first hockey team in seven years to not take the championship title\". Hopefully after reading this you have realized that it was not my fault and that the team has actually done \"well\" I mean, next years team is going to be like really strongâ€¦ if you know Iâ€™m talking about you, feel regretful now. Cos I have given my f******g best and it is none of ur f******g buisness, beat it.  As I was saying the 3rd 4th did not matter much to me,  people only remember the best and the lousiest , so I feel that between 3rd and 4th, there is no diff.  '),(5647,4102045,'03,August,2004','2004-08-03 00:00:00','   Children, are you not so happy to see me again?  Children nod entusiastically with gap-toothed smiles.  Wonderful. Let\'s have some fun...  I started a new exercise regime today. Yes, on a Tuesday. Sundays seem to jinx me. Now, I\'m not in it to lose weight (at 5\' 5\'\' and 110 lbs. I can\'t afford to). It was quite invigorating, actually. I\'m hoping I stick to it. I really can\'t afford to get flabby. And we all know, summertime is the easiest time to do that.  I\'m done with being shy. Seriously, it\'s gotten me now where thus far. It is a pointless emotion, only holding me back from my true potential. I mean, I\'m just as pretty, talented, intelligent, witty and all-round great as the next gal (not to sound vain), but I still get all fuddled and muddled around certain people. Gosh... why didn\'t I decide to do this sooner? Hmm.. the possiblities are now endless. Yay!  Well, children.. ta-ta for now. Be good. Remember...  I\'ll be back. Mwahahahahahah- choke*cough*.  Ahem. Right. Bye.  '),(5648,4102045,'01,August,2004','2004-08-01 00:00:00','   Goodnight, ladies and gentlemen. Welcome to another one of kÃ¡mmÃ­\'s \"episodes\". If you will, look to your right and put on the specially designed 3D glasses, which will provide more vivid viewing.  **WARNING: Content not suitable for sane viewers. kÃ¡mmÃ­ is not liable for any altering of mental state caused as a direct result of viewing this episode. Thank you for your understanding, and even if you donÂ´t, tough.**  A strange man is playing guitar and crooning mellow songs on my TV.  Shudder . I really hate folk music. Nostalgia is good, but add an acoustic guitar, and a handful of soulful back-up singers swaying and swooshing... aargh. No way.  I\'ve been thinking about becoming more cultural. I look at my life, and I say \"Kamara, you really lead a sheltered life.\" No Russian novels on my bookcase, no operatic CDs in my collection...  (oh no, he is singing gypsy music!) , no witty intellectual friends on speed-dial whom I can call at 2 o\'clock in the afternoon and discuss the fickleness of humanity, whilst munchin\' on some chocalat. None of that. So I say to myself, \"Kamara, you need to get some culture.\"  \"I know that,\" I reply, \"but what is there to do on a small, tropical island which prides itself on how many half-naked tourists sunbathe on its beaches in a year?\"  \"Aha!\" I cry, for I realise, that there in that simple question, folks, lies the answer to my profound lack of culture. Small, tropical islands are, unfortunately, not known as cultural hotspots. No symphonies, no theatre, no (decent) museum, and a library that apparently abhors the classics (or at least makes them terribly hard to find). What\'s a gal to do?  First thing. High tail it out to some place where there are more people. Statistics show  (and we know I love me some math)  that the more people you have, the more different they will be. And since people dictate culture, we can logically deduce that culture will be diverse. Folks, I cannot wait for college! The opportunity to live a new life, it\'s almost like having a blank slate. I can almost taste it. Sure, living alone will be scary, but for the chance to get some culture and to \"expand my horizons\"  (how clichÃ©) , I\'ll hop the next plane...  Meh. Just had a thought. Suppose there\'s no culture over there either? Suppose people in big cities are even more narrow minded than those on this sunny isle. Lord forbid...  My quest for culture cannot fail. It musn\'t.  To culture! And beyond...  '),(5649,4102045,'01,August,2004','2004-08-01 00:00:00','   The first post. The defining factor in whether or not a journal will survive. Well, maybe not  the  defining factor, but important for sure. It\'s kinda like that first toe-stick into the pool. You know? If the water is too cold, there is no way a herd of drunken buffalos could run you into that pool. The temperature is all-important. Just like this first entry. It\'s a toe-stick thermometor/glass ball, 76.8% guaranteed to predict the life-expectancy of any fledgling blog, within a 20% margin of error. So judging from this, the toe-stick thermometer says ( wiggling toe-stick thermometer for accurate measurement )... ERROR!  This can\'t be good...  '),(5650,4264876,'22,August,2004','2004-08-22 00:00:00','   I AM SOOOOOOOOOOOO BORED!      Little Seymore (aka daniel) is SOOOOOOO CUTE BLESS HIS LITTLE BI SEXUAL BEANIE lol!   I love him aaaaah cuteness!         PEACE OUT   '),(5651,4264876,'21,August,2004','2004-08-21 00:00:00','   Another day of crappyneess not 2 mention the history test that I had nooooo rememberance was WOT so eva!   I LOVE MATTY lol mmk little outburst there mmk umm yeh!   WELL i\'m bored once agen have nuthing 2 talk about as usual and i\'m buggered OH and i\'m goruchy coz there\'s no more white chocolate left and i\'m TIRED N SOOKY:\'(      Luv u all   PEACE   '),(5652,4264876,'21,August,2004','2004-08-21 00:00:00','   She won\'t get off the phone She won\'t leave me alone when I\'m talking to that guy She can be a real nitemare Get\'s me 2 do her hair and says it neva turns out right But When I Put On a Face She can tell coz she knows me so wellllll  Closer Tha My closest friend someone who will be there till the end My sister sister Deeper than the deepest sea No-one loves you like your family My sister sister  (ok hand getting sore stop writing nnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnow!)  This Is 4 Sum1 by the name of strawberry shortcake and i\'ve done a few bad things and i\'d like 2 say sorry (even thought its from a few months back) i still don\'t believe that one is over it yet coz i\'ve been told and i am honeslty sorry and you have 2 admit...I\'ve been a good gil lately..and frankly I enjoy it!   '),(5653,4264876,'20,August,2004','2004-08-20 00:00:00','   After a lovely morning of roll call i walk out into the corridor to find to my surprise a tic tac (the disturbing sence and trail of the one and only...DEMOND BOY AAAAAAAAAAAAAAAAAAH), yes the demond boy had struk the very spot i was about 2 walk on...anyway!     Recess (or waz it lunch I dno i can\'t remember) from being KNOCKED OUT BY A FREAKING BALL hit in the head(well not just a little hit in the head, it was THROWN AT MY HEAD by sum mysterious person i don\'t....know.....WHO but whoeva did they will b prosecuted...once i know wot i\'l do i\'ll tell you) along with the after shock of the DEMOND BOY striking his prey (which was me) with his deadly supply of DA DA DAAAAAA tic tacs!     Well that was mainly my day OH and then there was this one time after school and this sweet little person came up to me and said goodbye and gave me a big hug and a kiss on da cheek and sed ENJOY YOU WEEKEND and 2 my surprise it was ANGEL WOMAN...DEMOND BOYS sidekick!          FIN   '),(5654,4264876,'17,August,2004','2004-08-17 00:00:00','   Well Hey Kidlets,      I thought I\'d start off by welcoming myself and you to my thingo i dno wot they are but they seem koolness!   Well my experience 2day was exciting.....i saw a moonbeam of a life time and u could say it was a very shady experience!      well i\'m boring so peace out rabbits!   '),(5655,4075835,'29,July,2004','2004-07-29 00:00:00','    My cousin Chris sent me his Blog to read and asked me to pass it along, so here it is....  http://www.toomadawg.blogspot.com/  to start mine off, in honor of one of my fave comedians new shows, & the fact I go upstate at least 2 times a year I decided to put a forward that one of my upstate pals sent me (see below) Amy & I miss Tyler!!)  Just a quickiefor now... suppose to go to  a Floor hockey game to night to watch the demons.. (there was no rec. hockey last season they were renovating)  Well I dropped by today @ Murry\'s to see if I coul get a copy of the demons sked, and turns out they\'re palying the team I want to see them beat.. (and PELASE let them beat the crap out of them (or at least #17 on the team.. I despise the guy now.. such an ass...some of you know the deal about him.. to keep things short, had a crush lead me on so now,  I hope Foley or one of the guys breaks his nose tonight!!) anyway.. the renovations made are nice. a cool color blue was used has some tables to have your snacks at. that\'s blue too!  The rink actually has NO marks or at least as far as I saw... well here\'s the forward.. LOL   Jeff Foxworthy on Upstate New York    If you consider it a sport to gather your food by drilling through 36 inches of ice and sitting there all day hoping that the food will swim by,  you might live in Upstate New York.    If you\'re proud that your region makes the national news 96 nights each year because Saranac Lake is the coldest spot in the nation,  you might live in Upstate New York.    If your local Dairy Queen is closed from September through May,  you might live in Upstate New York    If you instinctively walk like a penguin for six months out of the year,  you might live in Upstate New York.    If someone in a store offers you assistance, and they don\'t work there,  you might live in Upstate New York.    If your dad\'s suntan stops at a line curving around the middle of his forehead,  you might live in Upstate New York.    If you have worn shorts and a parka at the same time,  you might live in Upstate New York.    If your town has an equal number of bars and churches,  you might live in Upstate New York.    If you have had a lengthy telephone conversation with someone who dialed a wrong number,  you might live in Upstate New York.    YOU KNOW YOU ARE A TRUE UPSTATE NEW YORKER WHEN:    1. \"Vacation\" means going South past Albany for the weekend.    2. You measure distance in hours.    3. You know several people who have hit a deer more than once.    4. You often switch from \"heat\" to \"A/C\" in the same day and back again.    5. You can drive 65 mph through 2 feet of snow during a raging blizzard, without flinching.    6. You see people wearing camouflage at social events (including weddings).    7. You install security lights on your house and garage and leave both unlocked.    8. You carry jumper cables in your car and your girlfriend / wife knows how to use them.    9. You design your kid\'s Halloween costume to fit over a snowsuit.    10. Driving is better in the winter because the potholes are filled with snow.    11. You know all 4 seasons: almost winter, winter, still winter and road construction.    12. You can identify a southern or eastern accent.    13. Your idea of creative landscaping is a statue of a deer next to your blue spruce.    14. You were unaware that there is a legal drinking age.    15. A brat is something you eat.    16. Your neighbor throws a party to celebrate his new pole shed.    17. You go out to fish fry every Friday.    18. Your 4th of July picnic was moved indoors due to frost.    19. You have more miles on your snow blower than your car.    20. You find 0 degrees \"a little chilly.\"    21. You actually understand these jokes, and you forward them to all your Upstate New York friends.    Have an awesome day,  and know that someone who thinks you\'re great has thought about you today!.. \"And that person was me.\"..... '),(5656,4075835,'13,August,2004','2004-08-13 00:00:00','    (full length will be mailed as soon as I see AvP)              Happy Friday the 13th Hope you are all having/had fun.  Fitting day ya think that 2 legendary killers go head to head? (mm am I having Daja vu? Didn\'t we just do this? Let\'s hope this one isn\'t a disappointment.like Freddy vs. jason was a bit)  Choose your side   Vs. Predator    AVP         anyway... Yesterday (thursday my day off...) woke up and had a bad pain in my neck down through my spine!!!!  Well my Chiropractor sn\'t scheduled to be in the office on thursdays so i was in pain all day. Woke up today the 13th .. (how fitting.. I guess those two above got after me eh!?)  and I htik the pain was worse... so called in tht i would at least be late.... the ciropractor isn\'t in til 9:30 so I walked around the apt. cause I didn\'t want to lay back down.. I had to cradle my neck & back in a spot toactually \'move\' it. So I ended up going back and laying down on the couch was actually able to fall asleep awhile.. and got up a little easier. (love \'my\' couch..) went to Chiro, and he said i probaly slept funny. (duh!) adjusts me and says go home ice back and come back in afternoon for another adjustment!!!!!!   well feeling a little better at 6:54 pm... but my neck still has pain in it and I can only move myhead a little.  So therefore i din\'t go to see AVP today.. I\'ll try saturday. Then send this off..   Opening movie of the week.- Princess Diaries 2: Royal Engagement  *  PrincessDiaries2.com         fun funny & fabu....   for a rated G movie.. (yeah I went to a kiddie movie..)  I enjoyed it. but then I liked the first one. I see it as a Miss. Congeniality jr.  and one of the guys at work said that was so me.. Lol.... I loved the slumber party sesen. now that\'s an idea to do.. now I jsut need my own \'castle\'  and Queen Grandma ROCKS!    Lol... Also like the fact of them showing what would an american girl do if she were arranged.  you\'ve got a  kinda dorky hunk of a duke..& hunk of a next heir to be \'king\'. Hey I got out of my world for awhile and the movie kept me amused and in fantasyland. So I give it a B   even tho the stuff critics; The local paper & Entertainment weekly gave it a C- to a C   but come on critics over anylize everything...  1. does it really matter that the original screenwriter didn\'t return? (or some other misc. thing)  2. Doesn\'t it matter that you had fun watching the movie(s) and  3. that it kept your attention?  4. yes it has to be \'smart\' but the theater laughed at the right moments and all that.   so I think it deserves at at least what I gave it. '),(5657,4052552,'31,July,2004','2004-07-31 00:00:00','    What\'s up people? Nothing for me. Today has been ... sort of eventful. I guess. Got up this morning and did nothing, then ran to Fort Dodge, Iowa to get my haircut and look for cell phone covers for my phone. I didn\'t have luck, so I guess i\'ll just have the plain old default one for awhile. Got home and picked up my pictures from Pamida from NYE. They turned out pretty good. If you wanna see em, just go to  urlLink http://public.fotki.com/rosco2007/  and have fun.          Well that\'s all I can think of to write about today, I\'ll catch ya later.          Rosco. '),(5658,4052552,'30,July,2004','2004-07-30 00:00:00','    Life sucks. I thought breaking up with Tara would affect her more than it would affect me. But, I\'m wrong. As always. I\'m not upset that I broke up with her. I\'m not regretting it. I\'m just a little confused. She just moved on soo quickly. So quickly that if it was a movie the dialogue would look something like:          Me: I don\'t think we should be together.          Tara: But ... but ... I love you.          Me: Goodbye.          -5 Days Later-          Tara: Hey look, there\'s Arik! I love Arik so much.          Somebody: What about Ross?          Tara: Ross who?          So yeah. Arik still thinks Tara loves me, which I doubt. Arik thinks me and Staci are gonna hook up and fall in love. I doubt it. We\'re just gonna stay good friends, but have a little more fun. That\'s not wrong, is it? Eh, I don\'t care.           Tara thinks me and her will be great friends. Which I doubt. I would consider it if she hadn\'t run off with one of my ... i dunno if I can say it but ... best (?) friends. But hey, she\'d feel the same way if I ran off with ... hmm ... Tracy, or Tori. Who I barely know so I wouldn\'t do that. I\'ve come to the conclusion that life sucks. Shit will happen, and I\'m gonna let it. Arik and Tara can have all their fun and shit, and I\'ll just sit back and watch the drama unfold. It\'s bound to happen. And i\'m not the only one that thinks so.          Adios.          Rosco. '),(5659,4052552,'29,July,2004','2004-07-29 00:00:00','    Hey My Fellow Losers:       &nbsp;   It\'s me again and I have exciting news. Heh. Right. No. I was actually just bored and figured I\'d write something in this thing so here I am. I just got back from an interview at Pamida, it was horrible, I suck at those things. Oh well, I have another interview at Fareway at 4 this afternoon, hopefully I\'ll have a job by next week. Woo.        &nbsp;   Damn, I can\'t think of anything to write in this bloody thing so i\'m just gonna ramble. It seems me and Tara and done, forever. Oh, here\'s a shocker, she found someone new before me. That really surprised me, I thought I\'d have a new girlfriend before she\'d have a new boyfriend. Oh well, I guess i\'ll just be single for awhile. That\'s okay, I -love- being single. *rolls eyes* I have a few people on my mind but I know they\'ll say no, but w/e. :-( So if there\'s anyone out there that\'s single, CALL ME! or email me, whatever.       &nbsp;   I\'ll catch all you losers later.        &nbsp;   -Rosco. '),(5660,4052552,'27,July,2004','2004-07-27 00:00:00','    Okay, so here I am, I have a new blog now and I havent updated the other one since May so I guess this is long overdue. For you people that don\'t know who I am, my name is Ross, but people call me Rosco (well atleast the cool people like Christine and Ashley). I\'m 16 years old and I live in Humboldt, Iowa and I don\'t have a life. Oh, and I\'m single...so if there\'s anyone out there that needs a hot boyfriend...keep looking, but whoever wants an average boyfriend, I\'m your man!       &nbsp;   Well this past week was prolly one of the best weeks I\'ve had, or will have. It was the United Church of Christ National Youth Event on the campus if the University of Tennessee in Knoxville, and it was a lot of fun. On Tuesday, we all met at Pilgrim Heights in Tama where we got to know eachother and play games and stuff. On Wednesday we loaded two charter buses and left for Indianapolous, Indiana where we stayed the night and messed around at a Great Times place.        &nbsp;   On Thursday we loaded the buses again and set out for Knoxville, at around 5:00 or so, I don\'t remember when. We were givin the wrong information so we sat around for an hour in front of the wrong dorms. So when we went to eat, they were done serving and we were late to the opening worship and procession of banners, so UT paid for 42 pizzas from Domino\'s since we didnt get to eat.        &nbsp;   On Friday we had our first workshops and stuff. I hung out with Arik, Staci, and Hannah mostly that day. Then after our first workship we had freetime/sack lunch until 2 then we had another workshop. Then we had free time from 4 to 7 p.m.,&nbsp;me Arik,&nbsp;Hannah, and Staci went to the pool, the girls tanned and me and arik sat in an air conditioned&nbsp;building and watched hot&nbsp;girls.&nbsp;Then we had worship and Christine and Ashley sat by me and we quickly became good friends.       &nbsp;   On Saturday we had opening bible study and stuff, most messed around with Ashley, Christine, and Nate, lol. Then I went to a workshop with Ashley and Christine, it was fun... Then we had lunch again, we just hung out in the dorms and ate pizza and listened to music. That\'s when Ashley and Christine crimped Jon\'s hair, it was hilarious, then they made him put on a skirt and a tube top, omg it was great. We didnt have a workshop that afternoon so Ashley, Christine, Nate and I hung for awhile then at the dorms more people joined us and they played cards. Worship again, Ashley and Christine got in trouble for talking, lol, and then there was a Sheltered Reality concert outside our dorms, they\'re awesome...still.       &nbsp;   Sunday, was the same as every other day, hung out with mostly Chirstine and Ashley (since I\'m a big ladies man) and had workshops and shit. Then we had closing worship and we got yelled at by a lady behind us for being too loud and shit, its all good. Then we had closing vespers, the senior\'s last service, a lot of people cried, but I didnt, i hate crying. Then we loaded the bus and got on the road for home at around 12:45 or 1. Me and Ashley sat together, she slept but i couldnt get to sleep since she was laying on top of me the entire time, lol. But thats okay ;-). After the first&nbsp;stop, I finally got some sleep. At a quarter to 5 we arrived back at Pilgrim Heights, we already dropped people off at I-80 truck stop, and Coralville. I didn\'t want to go home because I knew I prolly wouldnt see all the cool people I met for a long time...but oh well, have to get back to my worthless boring life sometime.       &nbsp;   Well that\'s about it...       &nbsp;   -Rosco.        &nbsp;     &nbsp; '),(5661,4052552,'17,August,2004','2004-08-17 00:00:00','    Hey people, what\'s up? Yesterday was my first day working at Fareway. It went alright, I got the hang of everything already, which is good...right? I got there I had more paperwork to fill out, had to watch a couple movies on safety and shit. Then he took me up front and showed me how to sack groceries, like to put all the frozen shit together, and dont put the bread on the bottom, common sense really. Then he showed me how to properly take the groceries out to the people\'s cars... Yeah it was ... really exciting. I worked from 10 till 2 and made like...$24 whoo...money, i don\'t get paid till next Friday tho so that sucks.          Anyways...last night my dad finally decided to tell me that we\'re going to the Nickelback and Puddle of Mudd concert on Thursday at the Iowa State Fair. So I think me and my parents are going down early to like ... do stuff at the fair or something. And I was just talking to Staci and I guess she wants to go so if she can get there and if my parents dont care we\'re gonna go together and that\'d be awesome. I\'d kinda rather go to the concert with Staci than my dad, lol.           So yeah...I have to work tomorrow again...meh...I hope I don\'t have to follow the manager around again, that\'s embarrassing...Anyways, I\'m bored so I think I\'m gonna let you get back to your boring lives...           Adios.      Rosco. '),(5662,4052552,'15,August,2004','2004-08-15 00:00:00','    Hey people, it\'s me and I\'ve returned from a long weekend at Pilgrim Heights. Lemme fill ya in...          Okay well, Friday morning me and my mom left Humboldt at 6:30a and we headed for Madrid because we had to take my grandma to Mercy West in Des Moines for her knee surgery. We eventually got to Mercy at around 9:15a or so, i dunno. They took her in and stuff. Me, my aunt and my mother sat and watched TV in the waiting room for awhile and then we went out to eat to Cheddar\'s and I had a huge bacon cheeseburger, omg, the best. Anyhoo...we went back to the hospital and went into the recovery room and talked to gma for awhile and stuff. She was sick from the anesthesia so we had to wait until 2:30p or so before we could take her back to Madrid.          So we got back to Madrid around 3 or so and we sat around for an hour and me and the mother left for Tama. We got there at around 6:30p or so and we unloaded all our stuff into the lodge and headed to Montour to eat at Rube\'s Steakhouse. Yum...the 16 oz. ribeye is really good. Anyhoo...after that we headed back to the camp and we sat around for awhile and talked and stuff.          The next morning we had breakfast and then the official day to do nothing began. All the adults sat around and talked and shit while me and a couple other kids screwed around by the lake, mostly trying to catch minnows and shit. That night we had a camp fire and the adults sang songs and crap, the little kids had s\'mores and I did nothing. Then me, Marcus and Andrew played foosball and ping-pong for a while and stuff.          This morning we had breakfast, blah blah blah and then we had church. After church me and Andrew went out on the lake in a boat and tried catching turtles unsuccessfully. After lunch we went back out for a couple hours and caught a total of 4 painted turtles, it was pretty sweet. We let em go, of course, and then we all left and here I am filling in all you un-interested people. Next year it doesn\'t look like Marcus and Andrew are gonna get to go so I\'m hoping I\'ll be able to bring Staci maybe since she told me she really wants to go sometime. *shrug*          Anyhoo...tomorrow I start training at Fareway. Ugh. That\'s gonna suck...but atleast it wont be that busy. ... Yeah ... I can\'t think of anything to write so I\'m gonna let you go. I\'ll catch ya later.           Adios.      Rosco. '),(5663,4052552,'09,August,2004','2004-08-09 00:00:00','    WHOOO!!!! I just got back from Fareway and they fucking hired me!!! In your face!! This morning Todd from Fareway called me and asked me if I could come and talk to him today and I did and he hired me. But, eh, he made me run up to the courthouse so the Iowa Workforce or w/e could fill out paperwork and shit, then I had to run it back to Todd then he had to fill something out, I had to take it back to the courthouse they had to make copies and shit and then I had to run some more stuff to Todd. Now he has to send all the paperwork to the headquarters in Boone and stuff and he should hear from them by Friday and stuff. Then I have to get trained and stuff...fun stuff.          Anyhoo...last night I decided to talk to Tara. Not about the shit I\'ve done and stuff, but about her and Jon. Her and Jon, Kait\'s bf, hung out and stuff on Friday night. Supposedly because he doesn\'t know anybody in Hartley except Tara. Then on Saturday they went to a ... dance. Yeah, that kinda pissed Kait off so last night I decided I would try to get juicy details but I couldn\'t. She just told me the same stuff that she told Kait. Then we moved on to other stuff...blah blah blah. Then I got off and Staci called me and she was talkin to me about what her and Tara were talking about. They hate eachother again, surprise surprise. I don\'t care anymore, I\'m not in the middle of it anymore, Arik\'s the bitch now. :-p          So yeah anyways, I\'ll catch y\'all later.           Adios.      Rosco. '),(5664,4052552,'07,August,2004','2004-08-07 00:00:00','    Well I\'m bored and have nothing to do so I guess it\'s rambling time.          Well, first off, Tara deleted her blog so now I don\'t have to read sappy love notes or whatever anymore. But her last entry was hilarious. She was sayin how there\'s 4 people that her mom is pissed off at and stuff and that if we even think of coming close to Hartley, we won\'t hear the end of it from her mother or whatever. Heh. Whoo..another person that hates me...well prolly a whole \'nother -family- that hates me. I\'m always the bad guy. *shrug* I don\'t care, I can go to Hartley if I damn well please, its not like I\'m gonna be anywhere near them when I am up there.          Well this morning I had to get up early and work at a car wash at my church to raise money for our mission trip to Hawaii next summer. It was a lot of work and it sucked, mostly the part where I got soked from a hose. Eh. Oh well, it cooled me off. I think we raised like over $400, that\'s not bad, not bad at all. After that I went out to eat at Chang\'s with Kait and her family since my family had to go to Omaha to move Jen into her apartment. I was thinkin about goin with them to Omaha but I was thinkin about goin up to Hartley to see Ashley but I havent talked to her in over a week so with my luck she\'d be on vacation when I got up there so it wouldn\'t of been worth it.          Anyways, I\'ve been bored all afternoon with nothing to do...stopped watching the Cubs game because they were losing but now they\'re ahead so I guess I might go watch the rest of the game. I\'ll prolly get bored again and I\'ll end up in front of the computer. So yeah...           Adios.      Rosco. '),(5665,4052552,'03,August,2004','2004-08-03 00:00:00','    Hello people,          I\'d first like to address a couple people in my life who I\'ve had hard feelings towards for the past week or so. Tara and Arik. Last night I got into Tara\'s email account and did some stuff which I regret. I deleted all of her email\'s she\'s saved from me, Arik, whoever. I changed her password and her secret question and answer. I also gave Staci the password and she emailed Arik (using Tara\'s email) and told him some stuff...idk...          Then this morning I got into Tara\'s AIM account (it was way too easy) I, personally, didn\'t do anything except sign in once or twice and I didn\'t talk to a soul using her SN. However, I told Staci her password once again and she did talk to somebody. She pretended she was Tara\'s mom and talked to Arik. I guess that pissed off Arik and Tara...          I\'d like to apologize to Tara for hacking her email and AIM, I dunno why I did it, maybe as revenge for hurting me, idk. I\'d also like to apologize for Staci, sometimes she\'s a little -too- evil and...yeah.          Anyways...I talked to Arik earlier, clearified that it wasnt me he was talking to when he thought he was talking to Tara\'s mother. I guess our friendship is a -little- better. It\'s never gonna be like it used to, but it may get better.          Oh yeah, one more thing. Me and Kait are planning on going to Hartley this weekend, Kait wants to see Jon and (dunno why) help him move and stuff. So i figured I could go with her, maybe see Ashley, if I\'m lucky. Idk, the parentals will be in Omaha moving Jenny so I wont have anything better to do all day Saturday.          Welp, I better go.           Adios.      Rosco. '),(5666,4052552,'02,August,2004','2004-08-02 00:00:00','    Yes, life still sucks.   My depression is getting a little better since I\'ve talked to Tracy (Tara\'s friend) about stuff and she was a lot of help. But I guess i\'m starting to get sick. Not physically sick. Just sick of some stuff. Mostly Arik and Tara. It seems that they -think- they are in love but face the facts, people. They\'re 15 years old. Love doesnt exist when you\'re 15. I\'ve learned from that mistake. I just hope this time, they will too.          I\'m also sick of the fact that they have the nerve to say that Staci and I are jealous. Jealous of what? Oh, because they\'re in love. Uhh...No. We\'re not saying all this stuff like that they\'re not gonna last because we\'re jealous. We\'re saying it because it\'s true. People said the same stuff about me and Tara. And they were right. I wish I would\'ve listened earlier.           I can\'t believe I\'m saying this but...I can\'t wait till school starts. Something to take my mind of all this shit that\'s going on in my life. Maybe once school starts I can find a -real- girlfriend. Not a girlfriend that\'s gonna be depressed 24/7 and maybe one thats not gonna run off with my best friend. I can only hope. It\'d be nice to have somebody that\'s a little closer to home. That long distance thing was a mistake. I\'m not gonna try that ever again, unless I find a -perfect- girl that I can trust, one that won\'t come crying to me every 15 minutes because they\'re depressed. Yeah, I know, far fetched dream, right? Oh well. Maybe some people were meant to be alone.          Adios.      Rosco. '),(5667,3970154,'29,July,2004','2004-07-29 00:00:00','   urlLink           Is has left Wellington this morning and I didnt send him to the airport coz I don\'t have the guts to watch another of my friends leave. It\'s already hard to say goodbye but to watch him walk away much more harder. I wish him goodluck in his future undertakings. May he gets what he always wanted. &nbsp;   urlLink               p/s: that\'s my brother at the back sneaking into the pic '),(5668,3970154,'28,July,2004','2004-07-28 00:00:00','   Had dinner with&nbsp;fellow&nbsp;Malaysian students in Wellington jst now and man the&nbsp;dinner was superb! everyone was stuffed man. the food was alot and everyone was getting full by the time the last dish arrived.&nbsp;jz now was also the last dinner with Iskandar coz hes going home tomoro for good *goodbye Iskandar* he gave a speech and everyone was listening to his conversation and Jan as usual as a clown, interrupt his speech making his own&nbsp;jokes hehehe newayz was fun to see all the&nbsp;familiar faces again =)&nbsp;  &nbsp;    urlLink      &nbsp;&nbsp;  Me and friends at Satay Kajang sponsored by MSDNZ. From left: Nysha, Christy,&nbsp;Me, Lya, Naz. Jan, Sita. Sit down from left: Nalein, Zahrin, Yusuf, Adnin, Sean and Faz   urlLink '),(5669,3970154,'26,July,2004','2004-07-26 00:00:00','   Fuh..yesterday was another tiring day of my life.Had the busiest day of the week. 9am-4pm class. I was so tired I didnt even had the guts to visit my dear blog. Sorry blog, had to ignore you for a day but now I am back. Yesterday was fun even tho\' I was so damn tired. I met my Prince Charming [bak kata Jhaja]. Finally met him but the sad news is he\'s bold again.=( He had his hair last week tho and he looked so damn hot!Buat he\'s still the best looking guy evn tho\' he doesn\'t have any hair *laugh* Anyway, I had fun with my class yesterday coz normally I would sleep the whole lecture. But yesterday was the best man! we talked about the indegeous people in Aoateroa/New Zealand. We were asked about who can claim theirself as a Maori and who can claim theirself as being a Maori as well as who can claim&nbsp;theirself as a Pakeha or being a Pakeha.&nbsp;That questions had make me think till this morning because when we had the debate yesterday, the issues weren\'t yet been solved. Being a Maori and a Maori have two different meaning. A Maori means you are the Maori that have a Maori descent from your ancestors but being a Maori means you are associates with the Maori itself [e.g relationship]. But the question is who can claim theirself a Maori or being a Maori? A person who has Maori blood but doesn\'t know the language, the culture what-so-ever still can be known as A Maori but for those who do not have the blood but know the knowledege, are they still consider as Maori or being a Maori? hm..Oh what about being a Pakeha or a Pakeha? Pakeha means you\'re the white people[European] that live/born in New Zealand and you\'re not&nbsp;associate&nbsp;with any of the Maoris.&nbsp;Being&nbsp;a Pakeha means&nbsp;you have connections [e.g relationship, grandparents, parents]&nbsp;with the European people&nbsp;here and do not have any connection with the Maori people. It\'s quite confusing aye? coz the explainations are similiar.&nbsp;But the word Pakeha&nbsp;was given from the Maori people, it means \"Pake-Pakeha\" which means someone with the fair skin.&nbsp;So not all of the people can claims themselves as a Pakeha even though they were born and educate here.     When I had this discussion yesterday, I have been thinking about my own roots as well. My mind had slipped a little bit from what I was supposed to discuss. I was thinking about my roots. My culture. About my own indegeous people in Malaysia. How many of them can claim themself as a malay, chinese, indian and what-so-over when they have forgotten about their own roots? I am not trying to raise any questions or point out finger or anything, but isnt it we have to know our own roots first before knowing other people roots? isn\'t it your identity is the one who shape your life? speaking of the identity, I have met few fellow Malaysian who have been brought here but never ever know their descendant or ancestors back home but have claimed themself as a Malaysian. But when I ask, do you know how to speak malay?, they replied \" why should I learn the language? I\'m not gonna use that language anyway, what is the point of learning it? \". \"but you do claim yourself as a Malaysian but you do not speak the language, then why do you have to say you are malaysian anyway when you think the language is not important to you?isn\'t the language is important too? isn\'t the language is the thing that can identify you where do you come from? then what language you gonna use when you go back home? your grandparents would not want to speak other languages with you&nbsp;except your own language. It\'s funny how these days youngsters do not see how important is the mother tongue to them. The language that shape their culture. Without the languge, outsiders would not know where do you come from. They can claim themselves as Pakeha,Maori, European whatsover but what is important is your culture. That is more valuable than anythingelse. I wouldn\'t claim myself to be Javanese or Chinese or somethingelse, just because I have bits of bloods from&nbsp;each of the categories but I will claim myself as a pure malay because in this these days you will not find any of the purity of the ethnic groups anymore.All of them have been mixed with everything.&nbsp;So what is so proud with your own identity anyway? because that identity is the one that will shape the part of your life. That identity will show where do you come from and where you will be categorize in..and besides there is nothing wrong to know other cultures ae. Is just you have to knw your own roots first coz that is more important..hehehI love my culture heheh [proudly admit]is just in these times like this you tend to think something that you\'ve&nbsp;never thought&nbsp;before..just a thought not to be offensive or something. we live&nbsp;in a free country where thoughts&nbsp;are counted~so do more thinking and who knows&nbsp;you mite learn something from it~~&nbsp; '),(5670,3970154,'26,July,2004','2004-07-26 00:00:00','   It\'s 6.00am in the morning and I am still awake. Awaken by the books that are all over my bed. I was supposed to sleep by now but how can I sleep when the books are all over my bed!!! Yet I am still awake. UNi&nbsp;has just started but I&nbsp;can almost&nbsp;feel the pressure right now.And guess what, I am already under&nbsp;lots of stress even though UNi&nbsp;just started two weeks ago.&nbsp;Just look at my workloads, it\'s unbelievable,. I have 3 journals, 5&nbsp;essays, 5 assingments, 1&nbsp;writing editing, 3 writing exercises to complete and 1 seminar presentation. How was that? Amusing?&nbsp;I&nbsp;don\'t think so!&nbsp;I have to finish all these by the end of October, before UNi close. *long sigh* I hate my subjects sometimes because it&nbsp;\"steals\" me away from my social actiavities [e.g clubs, gossips, friends] but again, I should not be complaining. I should be thankful because I do not any exams after UNi closes. Lucky Me =)     &nbsp;   I called my fren, Chip before and I admit, I am a bit jealous and admire of him. Jealous because he already complete his degree whereas I am still struggling with my degree *long sigh again* He\'s wating for his graduation which takes place next month. *lucky him* Now I am&nbsp;starting to feel the presurre again. All friends have finished their study and have found an excellent job and I am still stuck here completing my degree. But why should I be sad because I remember what Chip told me, he said I inspire him somehow. I\'ve insipires him with my confidence to study in overseas whereas he doesn\' t have the chance to do so. I took it as a compliment. I never actually seen my self as an inspirer so when a person told me that, it\'s a big thing for me. Thanks!     &nbsp;   I remember my friends back home told me that I am lucky because I have the opportunity to study at Overseas University. Different Universities than them. I questioned them, what do u mean different Uni? Isn\'t just an UNi after all except different names and country? They reply, I am in NZ, they are in Msia. That\'s the different. I have the opportunity to study in different education systems. I totally agree about the diff education systems but what I do not agree is that , education will always be an education no matter where you do learn it. Coz the end of the day, the overseas students will come home and serve their country about what they have learned. Is it? No doubt about that. If only they know what is the truth beyond the eyes they see. I know some of the students here can be helpful, cheery, friendly and nice but do not forget about some students who are selfish, pathetic. egoistic and jeolousy???Oh yeah do not forget abt the fees that increasing every year?that\'s sux! I have to pay a large amounts of fees whereas they only pay just a bit from us which I think that\'s no fair. Speaking of the fairness, where is the fairness in the world when you have striked to get&nbsp;the fees decrease&nbsp;but you\'ve been ignore? So fair and square.     &nbsp;   Oh well, I think I better go back to my work. Have to get ready for school * a very long sigh* damn those thinking have distracted me from finishing my work! damn those thinking. Oh Boy, I hope today will be another excitement day of my life. Hope to see my crush today. Haven\'t see him last week. Kinda missing his smile tho\'s. Speaking of crush huh! =) &nbsp; '),(5671,3970154,'25,July,2004','2004-07-25 00:00:00','   urlLink        Ninie, a great friend and amazing human being with a pure soul. When you\'re with her, she puts on this tough exterior but she\'s actually marshamallow soft.   urlLink           I was reading the comment that my friend wrote on friendster. It\'s kinda touchy and emotional coz he has left NZ last saturday.&nbsp; I was feeling sad and lonely after he left. Not because I do not have any friends besides him or anything but because we are more than friends. He has become a brother to me. We shared secrets and gave advices&nbsp; to each other. But whats more important is that, we trust each other. Dat\'s more important than antyhingelse. &nbsp; I\'ve realise after living in Wellington, New Zealand, I have made lots of friends here. I have also made some enemies too :p heheh. Well i guess thats what we call the balance of life. You have to balance your life a little bit.&nbsp;It cannot go perfectly&nbsp;through your journey. It has to be ups and downs moments once in a while. Like&nbsp;Ninie told me, my friends feel comfortable to be around me. They feel ease whenever they around me. I take that as a compliment. For me, when&nbsp;my friends need me to help them, I will try my best to help them as much. Isn\'t&nbsp;that what you are supposed to do, offering a helping hand when they need a help?&nbsp;I guess so. My life has been improving days by days. I have&nbsp;gained lots of experiences by being acquainted with&nbsp;new people. Which I think good basically. Because&nbsp;I&nbsp;have learned how to control my life and not being control by any&nbsp;pathetic humanbeings. I am rebellious when I am being control. I normally do what I want to do at first without want to think what is going to happen&nbsp; and face the future consequences later. Hence,&nbsp;because&nbsp;I think, my heart rules my mind and my body. I always follow my heart. It\'s like I have this strong instinct inside of me. I try to behave like confident, wordly-wise girl who has a mind of her own. But eventually,it is my mind that takes over.  I&nbsp;have watched my friends come and go. After Ninie left last year, it seems that all my friends have continuously following her steps. Recently was Firdaus and soon will be Iskandar and Jan. God knows how sad it can be, seeing fthe riends you always hang around suddenly make a huge departure in your life.&nbsp; But I guess it wonâ€™t be called life, if you keep on receiving happy moments with you. Life is not about equality. Life has never been equaled to me. To me, life is like a tyre. It goes round everyday and it never sticks unless its stop-means youâ€™re dead. Speaking of dead, I just lost my best friend. Ardiana Azrin Iskandar. She was my best friend for 9 years and she died after got hit by a truck. She was on her way to see her wedding dress. Sheâ€™ going to get marry in about 2 weeks. I was shocked to receive the call from her fiancÃ©, Hazri Irwandi. It was like a drop bomb and there was no word to describe how shocked I was itâ€™s because I was talking to her on the phone just few hours before the whole thing happened. I can still hear the way she saying, â€œHey Ike, I am on the way to see my wedding dress. Itâ€™s your favorite color, purple. I cannot wait to see it. My tailor said itâ€™s done. Iâ€™ll call you when I am trying it on ok? Take care girlâ€. There, that was the last sentences she said to me. That was the last ever called I am going to received from her.&nbsp; May she rest in peace..Amin '),(5672,3970154,'24,July,2004','2004-07-24 00:00:00','   urlLink        Went to work this morning and it wasn\'t a bad day for me. Normally I would be complaining and angry towards my customers. Sometimes they are nice but sometimes they can be a pain in the ass. I can see sometimes they look at me without seeing two eyes just because I work at a fast-food restaurant. So they assume people who work at the fast-food restaurant don\'t have the manners or anything. People do work to gain money or even experiences. People who work at this kinda of places do have feeling. Just because they work at this kind of places, doesn\'t mean they are low or anything. They just want to get extra pocket money. In the contrary, even though they treat me like that, I feel I am better than them because I have what they do not have [I think]. I\'ve got money every weeks and excel in my studies and also I know how to treat people nicely. Not just judge them by it\'s cover. Coz I\'ve been there and felt that. So by treating other people they why people deserve, doesn\'t make me a loser. It makes me a winner. Speaking of that, I was watching the movie \"I am Sam\" yesterday. It\'s definitely the best movie so far I\'ve watched. It\'s definitely a movie that will teach you a lesson about love. There is no difference when it comes to love regardless whether you are normaly or disable. Sam who is a developmentally disabled adult who has been singularly raising her daughter , Lucy for all her seven years. Her daughter was taken by the social worker after they discover the father\' disability. I was thinking, how can you question the father\'s disability of taking care her daughter when he has raised her daughther 7 years by himself? Why they do not question about the mother who is perfectly normal has left her daughter behind just because she felt ashamed being pregnant to a disable person? But they question is, how can they took her away from her after 7 years? and questioned the father tons of questions which they already know the answers? For me personally, you cannot judge what you\'ve seen with being there personally. Sometime they do not know what they cannot see beyond the eyes. People can judge randomly but only us who know the truth.     p/s: that\'s the picture of my feet-was taken by me when I was in Uni last week =)&nbsp;&nbsp;  urlLink '),(5673,3970154,'24,July,2004','2004-07-24 00:00:00','   urlLink        WMSO\'s Party&nbsp;  urlLink      &nbsp;[At the back: Jan, Idzwan, Nal, Johan & Kakera. Front: Christy, Oli, Me &amp; Sita]   I attended the WMSOâ€™s party just now and it was fun to see all the familiar faces again. I was so happy to see Nad again. He just arrived from Malaysia last 2-3 days. I was thinking its like Nad arrival is like replacing Firdausâ€™s departing. Firdaus left and Nad is here. Kinda sad to lose Firdaus but I am happy to see Nad again. At least I have someone I can talk to when I am down. Oh yeah, it was fun to see KakEra again. Sheâ€™s finally here. She arrived couple of days before Nad. KakEra is like a big sister to me. With her, I am able to talk anything I want with her. I admit I like to tease her a lot even though she is older than me. At the party, I saw all the pictures that were taken on Bersatu Games. [Bersatu games is a friendly games that unites all the Malaysian students around NZ] Happy to see my friends around New Zealand from Auckland to Christchurch to Dunedin. It reminded me the memories I had with them during the Bersatu Games. But suddenly, my feeling was changed to sad feelings. I realized the Bersatu Games this year was be the last Bersatu Games I will ever joined. I wonâ€™t be here anymore when it will be in Auckland next year. I will be finishing my studies end of the year and attending the graduation next year. After that, I will be packing my bags and heading home. I will be sad to leave New Zealand when my memories have been increasing every year. '),(5674,3970154,'23,July,2004','2004-07-23 00:00:00','   urlLink     All Blacks won the rugby against South Africa. It was a great victory thanks to Doug Howlett who was the hero of the day. He made the last try before the games ends. On the other hand, sad to Carlos Spencer, he lost the last kick. But oh well, All Black already WON anyway even though he lost it. All Blacks-23: South African-21&nbsp;&nbsp; urlLink '),(5675,3970154,'23,July,2004','2004-07-23 00:00:00','   ...throughout the course of my life, I have learned many  things, not just the basics of living, but also the basics  of living with others - but it is that we are not or won\'t  be always alone. We are often in the company of one  another, whether we like it or not. My life has always  been most rewarding to me so far, but this not because the  rest are not. But it is the most memorable I have yet  because I started active socialization. I made new  friends, and learnt how and where to get new friends. I  ran... and I fell... I started to realize that I have not    just have to constantly make new friends.... but I also      have to learn how to keep them..I began to see things  differently, from a more grown up point of view - not just  making and keeping any friends, but true friends. People  whom I can trust all my secrets with, people who can  provide listening ear, a shoulder to lean on, and a heart  to care...whenever you need them there...It was only then  did I realised that true friendship is indeed more  precious than all the diamonds in the world. It reaches  beyond time, and the true miracle is in giving, not  receiving. I also realised that we do not know what we  have got until we lose it, but it is also true that what  we have been missing out until it arrives...Once, I closed  my eyes and recalled all my dear friends faces.I realised  that they all all been like one of the shooting stars we  could seldom see in the New Zealand and Malaysia\'s night sky.      They had streaked across my life and lit up the darkness cos  without friends, our lives would be terribly empty they have    filled up a gaping hole in my life...thanks to all of you,    &nbsp;my friends, for filling up this hole in my life....   urlLink '),(5676,3970154,'18,August,2004','2004-08-18 00:00:00','   Today was a \"sial\" day for me..everything today was so uneased around me. Uni was arite but once i went to work, most of customers either pervert, bitch and unfriendly..damn! i hate that..When I went home, I got my Adsl Modem for broadband[yey] but yet i still cannot use it coz the telekom ppl hasn\'t yet connect my internet even tho they told me i can log on anytime today damn!sial betul lar..damnit! geram then when i was on the internet[yes i was using my dial-up, then suddenly blackout!] gerammnyer.......actually byk lagi bad things happend to me today but i dnt want to mention it coz it already past n besides there is no point of talking it      Ive got my extension for my essay til next week friday yey..good for me but i guess i hav to struggle to finish it coz i hav another to due on the same week oh bugger..the upcoming holiday is not going to be a hols for me coz i have essay but what the heck i have to enjoy it anyways..i am going back and that is much much more fun hehehe      this morning was so damn cold wait a sec, the whole day man..bloddy feeezing...it was haling[hope i spell it rite] and it keep continuosly happen heheh and it was so nice i even took picture of my brother on it..heheh its like snow..i hope it snow. i wan snow hehehe...desperatly need it...:) '),(5677,3970154,'17,August,2004','2004-08-17 00:00:00','   Took me about 20 minutes to log on to the Internet today in Uni ..what the F**k????i am so piss now..i prefer to go home at the mo to check my mails but i cant coz i gotta go to class n go to work later :( damn..oh well      today mite be the last day i update my blog before i hv the access to log on to the net again..il be leaving welly tomoro nite afta skol and taking the plane to Msia on thursday morning yey!!!kewl kewl..i can seem to smell the Msian food now hahhah..missing the food so much now..hehehe cant wait to go home lar.. '),(5678,3970154,'16,August,2004','2004-08-16 00:00:00','   dear blog   I have been so bz for the past couple of dayz that i didnt even bother to check my mails or update my blog..ive been working for 10 hours straight the whole day yesterday and by the time ive got home, i was damn tired. Today my mother told me that we are going back to Malaysia on thursday, This thursday, so malaysia, here I come..yey!!! newayz, even tho I am happy to go home, but i gotta bring my books wit me because I hav 2 essay to due on mondy and tuesday and i got to send it online :( wat a holiday huh??then i hv another one to due after hols open..damnit      I am gonna miss CHCH malam Malaysia that I have promised Kam to go..Sorry liaw..I am going home much2 better aye..Il promise Il come to Visit u guyz aye..:)      I promise I will try to update my blog when I am in KL..even tho I dont have any internet access at home but I will try my best to go online..:) that is a promise that I will keep aye..:)      I am so happy that I could hardly eat at the moment..I was so happy to go home that the very first person I told was Olivia then I called Ninie in KL..I am so gonna Miss the weather in wellington. I am gonna miss the winter in NZ..wait hang on a second.i am not going back for good..i am out frm NZ for the past few dayz..oh bugger!!      i gotta go blog, if not i couldnot finish the essay that I have to due tomoro..its only 300 words..i can finish it in 1/2 hours coz I am expert [bak kat Olivia].. '),(5679,3970154,'13,August,2004','2004-08-13 00:00:00','   Happy Birthday to my dearly friend: Kam Weng Yuen. Today is his big day yey! Hope he has a good day today..  I received a bad news from my mother saying my grandfather passed away yesterday morning [Malaysia time] :â€™( I was shocked but mostly I felt regret coz I didnâ€™t had the chance to visit him when I went home last April. I think that was the most regrettable moment I had in my entire life. I wish I had seen him before so that I donâ€™t have this guilty feeling burden on me now. But someone had told me that thereâ€™s nothing to be regret about because it already happened and I got to move on. I hope my grandfather will forgive all the mistakes I have done and I will always love him and I will never going to forget him. Amin to my grandfather:  Abdul Aziz '),(5680,3970154,'12,August,2004','2004-08-12 00:00:00','   urlLink       Bestfriends are like diamons, they are rare but once you found it, you tend to keep it forever [Olivia Thomas, Me, Christy Ngeow]   urlLink '),(5681,3970154,'11,August,2004','2004-08-11 00:00:00','   *This song was orignally done by Brian McKnight but Brian (Littrell) sang it as his solo in concert.       ONE LAST CRY      My shattered dreams and broken heart Are mending on the shelf I saw you Holding hands Standing close to someone else Now I sit all alone Wishin? all my feeling was gone Gotta get over you Nothing for me to do But have one last cry  One last cry Before I leave it all behind I\'ve gotta put you out of my mind This time Stop living a lie  I know I got to be strong, But around me Life goes on And on And on And on  I\'m gonna dry my eyes Right after I have one last cry  One last cry Before I leave it all behind I\'ve gotta put you out of my mind For the very last time Quit living a lie  I guess I\'m down I guess I\'m down I guess I\'m down To my last cry '),(5682,3970154,'11,August,2004','2004-08-11 00:00:00','   Today was not a bad day for me at all. i wen to class as usual in the morning and the evening. i went to morning class and did my presentation. even though the presentation did not go the way that i wanted [technically, it was my mistake coz i was supposed to tell my tutor if i need a projector, laptop etc beforehand, instead but i didnt know abt it and i didnt tell her that] I did my presentation safely and sound. The end of the presentation,my classmates have to ask quetions or reflects on what i have presented. Well it turned out good and luckily my tutor[well she wasnt my tutor actually. my real tutor was sick so she had to replace her for a while till shes feeling better] saved the day. Oh man! the reading was hard to understand and i dun even know what they wanted and i presented the whole passage to the class instead. I felt so ashamed that I wanted to cover my face with something. So embarrasing man. How do u feel that you  have done your work nicely, clean and presentable but u cannot present it and u had to summarize a reading that you dont even get it even though you have read it for almost gizzilian times? Literally I understand the reading but to summarize it and presente it to the whole class [nah i dont think so!] I had to read the sentence that I highligted one by one. Just imagine how embarrasing was that? I felt like a total idiot. I am in my last trimester of uni. I was not supposed to do it but like in a freshman year, I did it anyway. Take Note:    The most embarrasing moment of this trimester.     Later, I went to see Neene and we walked together to go back to her place[coz i parked my car there at her parking space] I had to man coz there was so bloody hard to find a parking space in Uni these days. It even hard to find car parking for 5 minutes.  I remember the other day that I needed to drop my essay in Uni and took me for about 10 minutes to find a \"5 minutes parking space\". I was so pissed off man. I just needed to drop an essay and it takes less thatn 5 minute to do that. Damn it!    I went to my \"Guidance and Counselling\" class in the evening and it was totally totally a boring class as usual. When i went there, i didnt see any familiar faces that I normally sit with. I had to wait and praying that those familiar faces come today or else I would be die waiting. Luckily my fren turned up. Today we were learning abt this Pyschoanalytic and Self-Helping family what-so-ever. I had no idea what my lecturer was talking about. Me as usual after finished writing those notes on the OHP, started dreaming and floating in the air as if I was somewherelse. Then afterwards we had to watch a video about the Family problems what-so-ever and yet again, I still I had no clue what was the vid about. I was so boring that I ended up creating something on my piece of paper. Gosh, then what\'s the point of going to class then? [the reason why I always  have to go class is  because she gives a good note or I am force to go]    I am now actually taking break from doing my essay [yeah I know..how many essay do I have?]. Well lets see how many do I have to finish by the end of this trimester. I have 3 journals, 5 essays, 5 assignments, 3 writing exercises, 1 portfolio, 1 seminar presentation and 1 writing test. Hmm..hows that sound? This is the reason why I alwayz lock myself up in the house. Its not that I dont want to go out and have fun with my friends but look at the amounts of work I have to finish. Each essay can be 2000+ and each week continously I have to hand in essay, so obviously I have not time to hang out. I know my friends have been asking me \" Ike where are you? how come never heard from you, blah2 blah2?\" Take note guys, its not that i dun want to hang out, I am extremely busy. I have heaps of things to due, that still doesnt count the essay that I hav to rewrite sometimes. Arghh when that happen, I will be totally piss.    Break in another 2 weeks..but still its not going to be a holiday for me..I have 2 to due on the hols and 1 after that * a very long sigh* '),(5683,3970154,'10,August,2004','2004-08-10 00:00:00','   dear blog, it\'s been a while since i\'ve updated my website ;} been so busy lar did not even had time to look at it *silly me* newayz nothing interesting happen to me for the past couple of dayz except today oh bugger! i went to my WRIT 203-Writing for print media class and we had to this peer review thingy and i had to swap my essay with somebodyelse in my class. i swaped my essay with this guy who sat next to me and guess wot? he didnt even do his essay so i said \"well, u can read mine then. when u finished we can review it together\". So i waited for abt 5 minits for him to finish reading it. After a while, he called me and we reviewd the essay together and guess again what happend? he said MY WRITING SUCKS!!!  he qoute\" your writing is bad, ur conclusion n introducation dont flow, u use really bad wordS, ur reading make me confuse, u use repetition, u cannot put present, past n future in one essay, u have a bad structure of essay and blah2\" i was like \"f--k u man!. who do u think you are? u are not even suppose to say those things. those things are mean man!\". While he was saying that, I felt like I want to slap his face and say \"well buddy at least i did mine u didnt even do ur homework\". Stupid Russian!!!! u were suppose to correct my grammar or some shit like that not gave a bad comment on it...or u can say something bad but not as bad as that!! thats mean, horrible, despise and bad bad opinion..damn those people..i really hate him even tho i only spoke to him once and i will not speak to him nor seat next to him again..  oh well shit stuffs happen but u gotta overcome it then..arghh tomoro is my seminar presentation day and i havnt done anything abt it..and i am totally totally going to die..what am i gonna do it. the presentaion is 20 minute..what am i gonna say what am i gonna do damn damn..*sigh* arghhhhhh tidakkkkkkkkkkkkkk!! '),(5684,3970154,'04,August,2004','2004-08-04 00:00:00','   I want to cry but there are not more tears in my eyes     I want to scream but there are no more voices in me     I want to run but I am scare I mite fall     I want to live but I am afraid to live anymore     I want to love but I am scare to be loved again     I want to hide but I am scare people will find me     What I should do?? '),(5685,3970154,'03,August,2004','2004-08-03 00:00:00','   not in the mood to write anything at the moment, but feeling to jot something down..      dear blog, yesterday was another interesting day of mylife. I have met my old classmates even tho one of them cant seem to remember who I was..am..kinda sad but what the f**k, at least had to the chance to talk to him was great enuff..it was so kewl to know at least some of my frens are still breathing...       dear blog, I was supposed to be in class now but I am so sick at the moment. I feel all my body is itching, my legs are sore, my nose has been sneezing, my throat is sore and so on [all the bad conditions when your body is not feeling well]..I have been having a really bad headache for the past couple of months. The same headache that has been bugging my brain and keep me sick all the times. I wish I could tear this brain up and see what is going on inside of my head..har har..but I know it\'s impossible. If  I have done that, I would be dead by now..=)      to sum up, I am getting sick. Face the facts Ike! '),(5686,3970154,'02,August,2004','2004-08-02 00:00:00','   finally..i have finished my essay..wala! but the bad news is my friend had an accident and he wont tell me when did it happen *sob*..well not in the mood to talk abt anything now.. '),(5687,3970154,'01,August,2004','2004-08-01 00:00:00','   I was supposed to be doing my essay now but I don\'t feel like doing it. It\'s DUE tomoro and I am still playing with it, looks like I am going to hand it in on tuesday..f**k, I have another essay to due on tuesday..and I haven\'t quite finish yet! lazy me!I dont\'t know what have happend to me..I can\'t seem to remember what I have read. I have read the same book 5 times and I still cannot understand it.I have to write a written critique based on what I have read but yet, I am stuck here running out of ideas on what to write.Wait..hang on a second, I have done this essay before *sigh* I know how to write a critque but why my mind still shallow..out of ideas??why?why can\'t i understand the context?why? thats one question that I cannot find the answer..=) by the way, I have been listening to this song..instead of doing my essay..          Was it something I didn\'t say?-98 degrees             Spending another night alone     Wondering when I\'m ever gonna see you again     Thinking what I would give to get you back baby     I should have told you how I felt then     Instead I kept it to myself     I let my love go unexpressed\'     Til it was too late, and you walked away          Chorus :Was it something I didn\'t say     When I didn\'t say I love you     Was it words that you\'ve never heard     All those words I should have told you     All those times, all those nights     When I had the chance to     Was it something I didn\'t say     Always assumed that you\'d be there     Couldn\'t foresee the day that you\'d ever be leaving me     How could I let my world slip through my hands baby     I took for granted that you knew     All the love I had for youI guess you never had a clue\'     Til it was too late, you walked away          [Chorus]All the words were in my heart, they went unspoken     Baby now my silent heart, is a heart that\'s broken     Shoulda let you know you\'re the one I needed near me     But I never let you hear me     [Chorus] '),(5688,4173841,'14,August,2004','2004-08-14 00:00:00','   she had enough money to buy a coffin for her sisterâ€™s child who had none for a coffin for her daughter she thinks I had something to do with it she who is the one makes my bed  she lives on (and buys coffins) for almost the amount of money I spend paying off my car every month nearly the amount of the bond less than I give my mom and dad about the same as the life insurance and the new hospital plan she lives on far more, often double, sometimes triple what many of her fellows earn and she laughs    a lot  nomqibeloâ€™s brother does not want to help she paid him back all what he done there in Amsterdam  nomqibelo named Saturday in Swazi maybe gogo knew Martha would need a day of rest she is our Sabbath, our Saturday, our Seventh Day she is St Martha the Patron Saint of Housekeeping, Mikeâ€™s Hair Stylist, Purposeful Plumber, Flowers from the Garden, Electrician Extraordinaire, Chili Boy Goddess Mother, Trampoline with Dental Floss Repairer    who checks her work by jumping   leaping  I owe a fortune on wrong things like water, lights and artists fees God is helping me to be magnificent    Nomqibelo makes me famous   '),(5689,4173841,'13,August,2004','2004-08-13 00:00:00','   he has such hang-ups about sex because he says all women see him like men see women and they just want to fuck him all of him all of them which if you saw those lips those hips is easy to understand and i see him and i want to fuck him too so fuck him   '),(5690,4173841,'13,August,2004','2004-08-13 00:00:00','   before Gito was shot last week I was standing in a queue waiting to pay for subversive literature when the instore radio played an ad so loud it went right through me I could hear how  much the actor was paid to say the words because he didnâ€™t breathe once in the whole 45 seconds it was a lot especially in comparison to most peopleâ€™s money  the ad was encouraging me  to SPEND QUALITY TIME WITH MY FAMILY THIS WEEKED I listened because that is what I felt like if you felt me you would know all I needed to do was  BUY THE VIETNAM WAR PC GAME  and play it WITH MY LOVED ONES THIS WEEKEND VERY LOUDLY  with spit flying out of the corners of his mouth and  THEN EVERYTHING WOULD BE ALRIGHT like Jesus Christ Superstar  started puking in the C N A one day last week  because it smelled like napalm and I canâ€™t stand that smell in the morning disgusting bits, blood, body pieces, came flying out of my mouth, with gob, you could smell the green bile from halfway across the mall pieces of wet vomit made like a projectile, stuck on the plastic stationery oozed down page-a-day diary covers, derms and offal the passion of Christ had nothing on this filth that flowed, out my every orifice, until the C N A succumbed  to my own intensely private chainstore massacre  shook my head found my balance paid for the Noseweek thanked God for my friends and family left the C N A suffocating with snot stood on its head pressed down hard before I left   '),(5691,4173841,'13,August,2004','2004-08-13 00:00:00','   1. Savings fathers\' Life x 3 times at 2am in the morning over three months.  2. Mother threatens to hit father with knobkerrie (South Africanese for Zulu cultural weapon) and runs away to unknown destination for one night during which I manage to embarass the entire family in desperate search. Quite proud of parents\' passion afterwards considering their age and physical condition. Mother approached to explain why she has never apologised to me in whole life. Answer (to be expected) \"Because I am always right.\"  3. Doctors get hold of lover\'s father and operate endlessly for no (all intense and) purpose other than to inflict misery on sweet man. God get\'s hold of lover\'s father who is buried with love and sadness and without great collection of stamps and coins which have all been sold to defray.  4. Own father not breathing so well and smoking like chimney  5. Lover\'s mother-in-law rises from bedridden status (2 years long) to drive car, only to be struck down by unknown forces and then, after recovering from serious medication addiction (as a result of a decision to leave her finally to be with just a morphine patch and love for comfort) opens death\'s door to walk back into world but immediately finds herself with minus one leg and therefore inability to walk and no gees (soft \'g\' - South Africanese for \'heart\' or \'spirit\') anyway.   6. Earning less than have in many years as a result of being over ambitious and seriously fucked over by government department responsible for developing small business who I should have known better (I did know better) are not capable of same.  7. Phone cut off.  8. Water and lights not cut off God knows why.   9. Oh. Lights cut off anyway often because neighbourhood transformer suffering from neurosis and everytime this happens it\'s when I am bleeding like there will never be a menopause and freezing cold and want my bath, my hot bath, my hot bath, my solace. Also, everytime it happens we have to phone to check it\'s the whole neighbourhood in case my neurosis is real and we have actually been cut off because the bill looks like old biltong with green growth.  10. Magician Housekeepers Brilliant Mother Dies. She had a great life but the sadness is deep sad.  11. Great friend musician dies of murder. Great friend musician\'s wife leaves for An Other Country. No comfort.  12. Child requires promised most expensive addictive toy and no way of backing out without appearing completely treacherous so buy on tick and get black mark.  13. Immediate work superior behaves like Most Important Bully On Block on Daily Basis. She does not have a moustache but you can only tell by looking at her.  14. Nobody cares so new puppy constantly gets out and races cars along main road. Gets inevitable hit. Operation. Survives. More moolah required. Thank God our vet is powerfully reminscent of childhood sweetheart so at least get to flirt in the midst of mania.  15. Electric gate has nervous breakdown. Area bad so have to fix or get mauled at front door. Fix. Big - very big - bucks. Week later, wake up at 5am to see three men at bottom gate (open) in dark, talking. Think it\'s our friends - the guys who live at bottom of our garden in The New Improved South Africa so wave and go back to sleep. They are young. We think they must have been partying. One even waves back. In daylight it becomes obvious gate motor has been stolen by deserving fiends. Fix again quick. This time fix so can\'t be stolen. Why didn\'t we do that in first place? Can\'t answer that. Next: Week later, big truck arrives to fetch rubble and reverses out of gate straight into intercom system. Breaks. No money left to fix. Son sings Gladness costs you not a thing and he who is happy is a king. So we sing loud here in the inner city and stick a sign on the intercom next to the fancy electronic gate to say phone us when you get here. Only phone is still cut off so if you don\'t have cell no, you have to leave.  16. Memory fails. Shit flies. Amnesia a blessing.  17. Funny how you can love your life. I went to visit my boba and she asked me how\'s it going. I said \"Boba I am praying for the Messiah\". She said I should start by putting out that cigarette.  I love Pablo Neruda - \"love is short, forgetting is long\". Goodnight Sweet Netwit   '),(5692,4173841,'09,August,2004','2004-08-09 00:00:00','   urlLink    tower of strength&nbsp; urlLink   '),(5693,4173841,'09,August,2004','2004-08-09 00:00:00','   The 12th of August. I decided that seeing as how it is currently obvious that shit is flying in my direction, it would seem most appropriate to accept the fact instead of freaking out at the mess. Therefore I defy the universe: let it fly! I am not going to behave as if I\'m the one living in a urinal. Nor am I of the opinion that I am guilty of some strange karmic manipulation that has purposefully attracted this chaotic compostic. Fuck that. Nevertheless, I am prepared to take 100% responsibility for the outcome so have decided to wrap myself in plastic. Hence the latest pseudonym - Ms Cling Foil. Aka The GladWrap. No shit on my skin. No smell up my nose. No matter how much sewerage. Bit challenging to breathe I must admit but at least I\'m transparent. I fight with love and laughing.   '),(5694,4173841,'09,August,2004','2004-08-09 00:00:00','   I say my hat looks cool hanging on the tower he say shall I take it off? I say nothing he say the reading lamps not working I agree he say thereâ€™s nothing wrong with them look and he switches one on I say you want to fight?   '),(5695,4173841,'08,August,2004','2004-08-08 00:00:00','   I come from a long line of people we escaped into the desert we wandered for forty years we were looking for a promised land  we got commandments we broke we got manna from heaven we ate we got the land of milk and honey weâ€™ve been fighting over it ever since maybe our mom was not at home so much  there is blood in the mud in my Jerusalem my Joburg where Gito died like Israel some motherâ€™s child someoneâ€™s boy maybe two kids they were not the pharaohs maybe brothers it was not Goliath maybe there were three and maybe they werenâ€™t such kids maybe they were older they could have been the philistines someone or more than one someone sure was their mother maybe she was not home so much maybe she was not enough  they shot Gito no one knows what happened but we all know they cut us to the quick blood flowed from our neck the sound of it was not like a fountain or Gito chanting it was the sound of Gitoâ€™s lover, screaming she is the one who wants to think about justice so I think about Gitoâ€™s mom because she is also someoneâ€™s mother she is the mother of the shot child shot by the son of another mother who was not at home or enough  Jerusalem is everywhere you go down President Street Gito knew that and he knew we all know the arbitrary mayhem intelligence officials use those sharp pointer spikes to finger terror three days after they shot Gito one more shot was fired close to our house and Rosie said she saw police cars and someone else got shot but we donâ€™t know him if it was a man If we did we would have heard about it by now  some people get shot and just lie there and noone comes. Not even anyoneâ€™s lover and nooneâ€™s mother. someone else gets it in the neck and the world mourns stricken, while the lone body lies there still, loved ones surround some other spirit. Arbitrary mayhem. Arbitrary love.  Gito was shot everyone is crying and making soup people come from everywhere they are phoning and we are trying to find them in the dream your legs canâ€™t move when itâ€™s your time Erika wants to think about justice but I lost my confidence I will drive up and down President Street Jerusalem late at night looking for my promised land, eating my manna, breaking my commandments but I canâ€™t also be expected to speak crevices in my character where terror creeps in to stop my tongue from turning we canâ€™t hold our faces straight Before Gito died it felt like everything was ok because we werenâ€™t lying in the street shot in the neck dead and waiting for someoneâ€™s mother to come but she was not at home  AND WHERE IS THEIR DAD?   '),(5696,4173841,'08,August,2004','2004-08-08 00:00:00','   We are having smoked salmon in times of hardship and famine. This is our way of defying the water and lights bills. My life is not really a series of one-liners. We have so much death and destruction around us, you need to have a quip with you wherever you go. Hey I love my city. We live in a house with no burglar proof and a view so fuck you if you think you can use this to justify your fear.   '),(5697,4173841,'08,August,2004','2004-08-08 00:00:00','   the paperâ€™s less waxy this side the penâ€™s back in my hand and then, nothing happened no song not even a verse the words stretched out all over this page in gangarese the no-poem was stoned it was full of rocks with sharp corners and barefeet bleeding, heart beating it was saying yes! repeat after me it was saying yes! yes! thatâ€™s what I want to hear more of great new ways to string along an audience without declamation punctuation or even any sign of meaning nor even a mean looking sign none only two telephone interruptus then still it was, my sonnet my kind of sonnet my sonnet shoobeedoowahwah   '),(5698,4173841,'08,August,2004','2004-08-08 00:00:00','   There\'s an alarm going off in the house on the hill. It is not a very loud alarm but it has that madness to it that comes with never having heard the sound before and it\'s half past 8 and it suddenty went off. And we don\'t have an alarm. Unlike most of the rest of the nation who have laptops to blog around in bed with. Was that a split infinitive? Do you know how many times I have been shown how to avoid splitting infinitives? About the same too many times as I have forgotten. The alarm has stopped. As I wrote that it just stopped, like it began. Joburg has its maniacal moments. Often. I\'m sure there are peaceful spots but having lived here all my life, never having left for more than six weeks, ever, (can you sense the self-pity?), I am reasonably sure that they are not close by. The mayhem is so normal now you just keep on typing. Hurtling slowly through Africa.   '),(5699,3291131,'27,May,2004','2004-05-27 00:00:00','    I usually eat a twix at 10:00 am at work. (that means in 20 minutes now) they currently have this thing where u enter on the web the codes that u find on the wrapping paper. when u get 15 codes (hopefully I will before the promotion ends) u get $10 gift certificate at ticketmaster if I understood correctly. a twix is $0.7 at the vending machine -> 15 twix = $10.50 ... that means they re almost free !!!   Update : actually they cost $.60 at the vending machine -> 15 twix for $9 and u get a $10 gift certificate, that s a good deal !   '),(5700,3291131,'26,May,2004','2004-05-26 00:00:00','    urlLink    Here is my new bike.   '),(5701,3291131,'24,May,2004','2004-05-24 00:00:00','    yes, I got a bike today !! I finally choosed the treck 3700. I tried it for like 20 minutes, it seems pretty cool :-))  but I didn t know I would get tired soo quickly, i defenitely need to exercise more often. anyway, I am quite happy.  btw, as expected I didn t get the free helmet, but I managed to get a free kick stand. well it s not much (like $8) but it s symbolic. actually it was just to show roxie that u can always get something free :-)   '),(5702,3291131,'19,May,2004','2004-05-19 00:00:00','    I am about to buy a bike. As always when it comes to spend a substantial amount of money, I spend hours researching the best item or at least the one that fits me (and my budget) the best. Anyway, I just don t know what to buy, I change my mind everyother hour. Should I buy a low end model at target, the best target bike or the cheapest model at a bike store (these propositions were sorted by prices) Should I buy a full suspension mountain bike, or a \'comfort\' MTB would be better for me... My latest thought was to buy a comfort mountain bike at target. this would cost me $115 (on sale, orig price : $135), which is actually the first budget I imagined (after adding the different accessories). advantages : - cheap - more comfortable than a sport model - target -> if I am not satisfied I can return it within 3 months - expensive enough to get some good stuff (I am not an expert, but shimano derailleurs sound cool to me) disadvantages : - target -> who assembled it... if I have a problem, nobody is going to help me there... frame size is unique and not necessarily the best one for me - cheap   so in my mind I think, why buy an expensive bike if I don t bike every single day, and I bike more in the future, I will just upgrade it then. but in an other hand I am thinking : I don t buy a bike every day, this is not going to hurt me to spend some extra bucks for better quality...  so your comments are welcome folks !  ----------------- Update :  as expected I changed my mind again. I am now pretty sure to spend some extra money and buy at a bike store. I don t trust target quality that much and I think frame size is quite important and target bikes are all too small. I tried some bikes at a bike shop, and I can feel the difference just seating on it for 5 seconds. so I will certainly buy a mountain bike that is very capable on road too. I am thinking the trek 3700 ...   '),(5703,3291131,'19,May,2004','2004-05-19 00:00:00','    I just wanted to wish myself an happy birthday   '),(5704,3291131,'17,May,2004','2004-05-17 00:00:00','    I just come back from 10 days vacation. it was pretty hard to wake up this morning.   '),(5705,3291131,'11,May,2004','2004-05-11 00:00:00','    urlLink    Now that s it... it s over... I started watching friends when I was in high school.  It felt really weird to watch the last episode. It reminds me of so many things, ... wow....   '),(5706,3291131,'11,May,2004','2004-05-11 00:00:00','    urlLink    urlLink     I went to see game 2 of the playoff round 2, spurs vs lakers, in San Antonio. tonight is game 4, I wanted to go see game 5 on thursday, but so far it is sold out, but who knows, I succeeded in getting last minute tickets last time ....   '),(5707,3291131,'10,May,2004','2004-05-10 00:00:00','    I finally got my first credit card application accepted today. that s such a good feeling....   '),(5708,3291131,'10,May,2004','2004-05-10 00:00:00','    I read a little bit the help of this website. it seems that u can not protect your posts. So I guess I won t be that private in my posts.  But otherwise this website seems pretty funny, let s give it a chance !   '),(5709,3291131,'10,May,2004','2004-05-10 00:00:00','    Hey Everybody  tired of my livejournal experience, let s try this website, seems pretty cool actually. so let s start with this and we\'ll see how far we can get   '),(5710,3291131,'10,June,2004','2004-06-10 00:00:00','    I have a gmail account !!! (google mail). thank you so much Cynthia.  so far I like it. because it s still on beta program, the number of users is limited and so the access speed is great. but other than that, I like the features a lot !  what I like a lot is the way the interface behaved like it would be a stand alone software : what I mean is hen you look at your inbox, you click on an email and then go back to your inbox, the web pages don t reload each time. so when you have 10 new emails, it goes a lot faster that way. NO POPUP !!  their advertisement system does not bother at all. I like the grouping by conversations too. If u have some friends with who you have an email \'chat\', replying back and forth to the same email all day, it will be shown as one line in your mailbox, not 20... which is cool to still be able to see your other emails. then when you click on the conversation, the display is \'forum like\', which is nicer than the usual \">\" \">>\" ... \">>>>>>>>\"  in front of the beginning of the discussion  and of course, what s great too : 1Gb of storage....  so far so good, I love it !   '),(5711,3291131,'01,June,2004','2004-06-01 00:00:00','    I have a counter on my blogger. Because blogger users pages show up on google searchs, I thought it could be funny to do some experiment. let s write here the most researched words on google and let s see if more people come to visit my page : iraq american idol john kerry cyprus michael jackson fallujah howard stern omarosa wachovia rebecca loos earth day usher kill bill halle berry lindsay lohan the apprentice britney spears sex shrek 2 bachelor avril lavigne dessi espana moore emma watson tony randall   '),(5712,3291131,'20,July,2004','2004-07-20 00:00:00','    I am looking for opponents at scrabble game. I play once in a while, but i unfortunately don t know anybody to really compete with.   '),(5713,3291131,'11,July,2004','2004-07-11 00:00:00','    I recently bought a packof big K cola to try it. I am a fan of pepsi one, diet pepsi and diet coke.  well I had a big deal, like $3 for a 12 cans pack, and it s actually pretty good.  I may give it a chance once in a while   Update : $3 for a 12 packs is not really a deal, I meant 24 cans   '),(5714,3291131,'11,July,2004','2004-07-11 00:00:00','    I have been lazy lately, and so I just realize I haven t updated my blog in a whole month. I guess I need motivation, so leave your comments !! :-)   Some friend asked me why I haven t updated my blogger, that made me happy enough to come back, so I hope u enjoy it !   '),(5715,3291131,'11,July,2004','2004-07-11 00:00:00','    urlLink    these past days, I have been playing with my bread machine, homemade hamburgers, pizza and my latest creation, this kind of apple pie, very good....   '),(5716,3291131,'11,July,2004','2004-07-11 00:00:00','    urlLink    some weeks ago, I was at Ottawa on a work trip. Here is one picture from there, that was actually my hotel...   '),(5717,3413707,'29,May,2004','2004-05-29 00:00:00','   I know that i dont post often, but hey thats because I dont care about any/many of you.  If you are actually reading this. That means you care enough to read all I can say to you is your my best friend. I actually porbably wont do this much so watch out for times when i do because it means one of three things.  A) I am VERY bored  B) I am VERY happy  C)I am VERY pissed  before I leave you I have a message for a specific person, I wont say said persons name, but that person will know who they are.  Here it is.  PEACE  '),(5718,3413707,'22,May,2004','2004-05-22 00:00:00','   I missed bowling last night I am sooooooooooooo pissed.  i really wanted to go  anyway I know that no one but me is reading this so whatever  '),(5719,3413707,'21,May,2004','2004-05-21 00:00:00','   hello all you byotches. I am gus, i have three names, gus, angelis, and yokokurama.  Call me witchever of the three you want I already made a blog but i forgot my old screen name so i made a new one.  I can only post on weekends so there you bag o poop.  I also am not partial to blogs and Will only write about once every one-two weeks so there  '),(5720,3413707,'11,June,2004','2004-06-11 00:00:00','   It must be awesome being a blog administrator.  I mean having the power to remove someone\'s posts and commments at a whim....... anyway I think we (you, the reader and I) need a talk about anger management.  I know some of you have the faint reckoning that i shouldnt teach/preach about anger management, but TOO FREAKING BAD!!!!  I personally think there arer too many vibes being sent my way (SAM THAT MEANS YOU!!!!) but its not all sam there are other people with way too much anger towards others/me.  Well to all of you feeling angry i say BLAH!!!. LEts define anger for a moment...... ah yes   a  n  g  e  r  what exactly is it?  The international dictonary of gus defines it as \"a strong feeling of displeasure\" so to all reading this....... DROP ALL YOUR DISPLEASURE.  be free of anger (especially towards me) and be happy.  I SAID BE HAPPY!!! WHY YOU LITTLE................  '),(5721,3413707,'10,June,2004','2004-06-10 00:00:00','   I apologize for my unstraightforwardness, but I have been ANGRY.  but i am not anymore so you should all be happier now.  What I meant in my last blog is for anyone who read it to list reasons for why they were reading it.  In stupid I mean common folk terms that means write the reasons why you think you are my friend.  But that was when i was angry so dont worry about it now i just thought i would clarify it for ya\'ll.  SHREK 2!  '),(5722,3413707,'10,June,2004','2004-06-10 00:00:00','   from now on i will try to use what bart writes on the board as my titles so GET USED TO IT. i have been edgey this summer and i apologize. but i have one question fo ranyone reading this.  WHAT THE CRAP ARE YOU DOING READING THIS. comment on that list good strong reasons as comments i will sheck constantly to see if anybody has the guts/brainpower to and i realize the irony of my edgeyness  '),(5723,3413707,'05,June,2004','2004-06-05 00:00:00','   School is out. YAY!!!!!   smell that sweet air.  What does it smell like? OK we\'e wafting, we\'re wafting and it smells like SUMMER!!!!!  Now I would like to send a couple of subliminal messages to one person. ahem.....  summer is PEACEful.  it is a time for PEACE.  for anyone reading this doesnt peace mean that all wartime afflictions and effects are dropped? comment on that.  I hope the sublimanee got the message there  '),(5724,3413707,'02,June,2004','2004-06-02 00:00:00','   Chello again.  ANd for the record I know that it is spelled cello, but I spell it chello to make it a hello kinda thing. Well anyway I left early from the \"party\" for two reasons.  my parents made me and CARSON LANGSTAFF or ling-lang as scott made up spilled water on me.  Polly did do some of it, but carson has ben a jerk lately. Carson/ling-lang if you are reading this you have been way too mean lately even before the iincident.  PEACE  '),(5725,3413707,'13,July,2004','2004-07-13 00:00:00','   I apologive to you all for my partially unexplained absence.  for about four days i was in the mountains, then i fell down the mountain.  remember that seen from the simpsons where homer trys to scateboard across the gorge and fails miserably.  well my fall was like that, but on a less steep hill. I landed on some jagged rocks and had to be rushed to the hospital.  and there i have been and now i am free!!! but i still hurt so no plans for me for a while.  i was in that hospital for so long.....   I was on double morphine the first day and i figured out the best thing to do, watch waynes world over and over and over until the morphine runs out.  if you need or want something call me i will be at my dads house for the next three weeks because my mom went to england.  and it has been annoying because she left a day after i got hurt and keeps calling to see if i am alright.  somebody call i am sooooooo bored. until the doctor says i am better, farewell......  '),(5726,3413707,'31,May,2004','2004-05-31 00:00:00','   We are most likely doing crossdressing day, people.  It would be tomorrow.  Please inform your non-BotD friends about it.  (And your BotD friends).  '),(5727,3413707,'30,May,2004','2004-05-30 00:00:00','   IM TOTALLY ALL FOR THE CROSS DRESSING THING BUT..........how are we gonna pull it off? are we just going to come into school in regular uniform and bring a pair of pants and then just exchange with each other (in bettween u cna wear the pants and whatnot). Im all for it, i think we should start a phone chain and circulate the word. each person calls two other people... ok well i gtg, im all for it tho!  '),(5728,3413707,'30,May,2004','2004-05-30 00:00:00','   i am back, i know you all missed me. the cross dressing thing has to happen, i mean thatd be amazing. in case your wondering what i did when i was in new orleans, this is what i did: shopped, went on a few vampire tours (with tony), hung out at burbon street, walked around the french quarter, and had a great time. anyway... i had fun and i know you all missed me.  '),(5729,3413707,'30,May,2004','2004-05-30 00:00:00','   Geez Gabo, tell the whole world....  '),(5730,3413707,'30,May,2004','2004-05-30 00:00:00','   Please can we have a cross-dressing day??? I mean, Taco used to running around his yard wearing a kilt screaming, \"You can take our lives, but you\'ll never take our freedom!!!\" anyway!...  '),(5731,3413707,'30,May,2004','2004-05-30 00:00:00','   Hey everybody, I haven\'t been heard from in a long time cause my internet failed on me, but I\'m back, never fear!!! Oh, and by the way.... FOOT THONGS ROCK!!!!!!!!!!!  '),(5732,3413707,'29,May,2004','2004-05-29 00:00:00','   This is Dougie speaking on Sporko\'s behalf, since she is in New Orleans and must pay for an Internet connection with money out of her own pocket. Getting to the point, Sporko wants people to bring their yearbooks with them to school so that she might have a chance to sign them. I\'m bringing mine, you should too.  '),(5733,3413707,'28,May,2004','2004-05-28 00:00:00','   Unfortunatly, it is unlikely that a \"cross-dressing day\" will happen. The only day it is possible is Tuesday. Monday and Thursday we have off, Wednesday, is the awards ceremony for the atletic awards and honors, (Mrs. Braxton will be taking pictures) and obviously due to commencement we will be unable switch on Friday. (That is of course unless Katie is willing to give up her Versace dress or Cimmie plans on losing a hundred pounds so she can fit in a guy\'s outfit). Nevertheless, we have to do this switch on Tuesday, if at all. This wont work if someone forgets or backs out thinking \"Oh my God, if I cross-dress I won\'t graduate, because they\'ll expell me!\" We need to coordinate plans since our only choice of a day is the next day we have school. We need to figure out something soon. My Idea is the following: In this \'chatterbox,\' write your name if you are   willing  to participate. Maybe we\'ll do it if more than half the class is willing. After that, I\'ll make arrangements with people asking if they can let someone borrow their skirt or pants. I will also support requests so Emily can get in Kyle\'s pants. Below is the chatterbox and I hope this can happen with your support.   Write your First name, and whether or not you are willing to participate.      var chatterbox = \'no\';         urlLink target=\"flooblechatterbox\"  onClick=\"window.open(\'about:blank\',\'flooblechatterbox\',\'toolbar=no,location=no,directories=no,status=no,menubar=no,scrollbars=yes,resizable=no,copyhistory=no,width=480,height=400\');return true;\">  align=\"absbottom\">Talk   in my  urlLink flooble chatterbox, a free javascript chat tag board / shoutbox / tagboard program for your xanga, diary, blogger or weblog  |   urlLink logic puzzles, brainteasers      if (!chatterbox || chatterbox != \'ok\') {  document.write(\'This  urlLink flooble chatterbox  is\'); document.write(\' temporarily unavailable. It will be back up shortly.\'); }  '),(5734,3413707,'27,May,2004','2004-05-27 00:00:00','   urlLink Click here to rate your teachers at Monty G!  This is one of the coolest websites I have ever seen, I encourage everyone to check it out!  '),(5735,3413707,'27,May,2004','2004-05-27 00:00:00','   for some odd reason my mother (yes, my mother) felt the need to discuss the cross-dressing day with me and she thinks it would be funny, but says we should do it (if we do it at all) on Tuesday because wednesday is the moving up ceremony and the acedemic awards thingy so people will be in full \"dress code mode\" and they might not approve...  '),(5736,3413707,'27,May,2004','2004-05-27 00:00:00','   We should totally do a cross dressing day. While I fully understand about the whole uncomfortable to wear a skirt and problem with getting the right sizes things, we should still do it becuase it would be really funny and no one\'s ever done it before.  '),(5737,3413707,'26,May,2004','2004-05-26 00:00:00','   New way of saying Spanish is dumb: L\'Espagnol is deaf!  '),(5738,3413707,'26,May,2004','2004-05-26 00:00:00','   read the title, i am in new orleans having a great time, and it would be totaly awesome if we did have a cross dressing day, but it has to be when i came back - like tuesday or wendessday. oh yes, thank you for making people put titles. and well... blah*  *copyright spork townsend 2004. not permitted to use with out written consent my owner, under penalty of pokey  '),(5739,3413707,'26,May,2004','2004-05-26 00:00:00','   You guys have to invite Ransom into the blog, and this whole cross-dressing thing is not gonna work for me, I\'m all for a good joke but you know you\'ll never get it together and no one wants to see me in a skirt, I mean come on.  '),(5740,3413707,'26,May,2004','2004-05-26 00:00:00','   Does anybody have a Spanish study guide?  '),(5741,3413707,'26,May,2004','2004-05-26 00:00:00','   ok so here\'s my thought: since we have the awards ceremony on wednesday, june 2nd, it would be wise to be dressed in regular uniform for that event. We, however, could switch dress codes on tuesday. That was a good idea tori, with going to the trift store, because it would also be problematic if we didn\'t get our uniforms back in time for the awards ceremony on wednesday (I don\'t know about the rest of you, but i only own one kilt that fits me). If anyone likes this idea, please say so in your next blog  '),(5742,3413707,'26,May,2004','2004-05-26 00:00:00','   From now on, you  have to  include titles.  This is because otherwise, it seems like you named your post something else.  For instance,  A Wrinkle in Time  by Madeleine L\'Engle would be called  It was a dark and stormy....   Doesn\'t work.  That includes you too, Emily Combe.  '),(5743,3413707,'26,May,2004','2004-05-26 00:00:00','   My mom and I actually saw one of these the other day.  '),(5744,3413707,'26,May,2004','2004-05-26 00:00:00','   We could probably also borrow clothes from the thrift shop, because come on, how many of us are really going to fit into each others clothes?  '),(5745,3413707,'25,May,2004','2004-05-25 00:00:00','   Pieter i fully agree, we should totally do that, anyone who is all for switching uniforms put it in your next post. I consider wednesday as the last day of school  '),(5746,3413707,'25,May,2004','2004-05-25 00:00:00','   Anyone who\'s wondering, I made up a name for my phobia, caducus-acer-phoebia, it means destined to die by tree\'s, in latin. Yeah, this is gonna seem weird to most people but I had to post it for the people who will get.  '),(5747,3413707,'25,May,2004','2004-05-25 00:00:00','   Emily Combe you wrote, \"sp\", thats not a full word and is breaking a rule (strike 2), also spelling something so wrong you have no idea what there saying should be illegal, (emensely?!?!?!). Yeah I\'m done.  '),(5748,3413707,'25,May,2004','2004-05-25 00:00:00','   I really want to see  urlLink The Day After Tommorrow  today. It came out 13 hours ago and I was thinking maybe for a cool reward to the end of exams, we could go see it tonight in Frazer. This is not finalized and I don\'t want only like one person showing up so I have included a poll stating if we should go. I will decide whether or not the movie is on based on the popularity of the category \"yes we should do it.\" If it is less than half, than the trip is off (and visa versa). I hope we can do it tonight at 7:00.    urlLink The Day After Tomorrow (PG-13, 124 min.)| 7:00pm |     THE MOVIE IS OFF, NO ONE WILL BE ATTENDING TONIGHT\'S MOVIE  '),(5749,3413707,'25,May,2004','2004-05-25 00:00:00','   This blog is being posted due to the fact that little development has happened recently regarding the last day of school uniform swap. If we all plan to switch uniforms with a certain individual, we must somehow organize something soon. Also what do we consider the last day of school? I think if  everyone  participates, this will be a success. We must plan this soon and pray that there is not a drop-the-dress-code-day on the last day of school.  '),(5750,3413707,'25,May,2004','2004-05-25 00:00:00','   vicksy-i think that technically at the moment we should be called the semi-ex montgomery schoolers b/c we\'re not really part of the school while we still are...if you know what i mean...gahhh...i\'m confusing myself again...  '),(5751,3413707,'25,May,2004','2004-05-25 00:00:00','   i, thats spork, will be gone for a week... i am leaving in about an hour for new orleans. i know you all will miss me (tucker dougie gabo gus i hope you miss me). blah, well thats all i have to say... by the way only about 2 people gave me their money, you all so horrible (you and your bad memories that is). by the way matt vessey, i am liking that idea of my own dictionary. have a good week and i hope the most of you fail, i mean, dont fail your finals, dont want me looking smart and bettering all of you in the finals do we? ok, i am gonna stop typng and go, i dont know, pack. love and all SPORKO  '),(5752,3413707,'25,May,2004','2004-05-25 00:00:00','   someone -- make sure that you change that thingy (yes, \'thingy\' is a word) along the side of the page that says \"a blog for all montgomery schoolers who feel damned\" because unless you\'ve had your head stuck in a toilet for the past three months, you will know that we are graduating soon...we most definitely will not be Montgomery Schoolers then - we will be EX-montgomery schoolers... ((sorry, i had to get that out in the open, it was bothering me))  '),(5753,3413707,'25,May,2004','2004-05-25 00:00:00','   hey guys, its sydney, dougie thanks for inviting me to this blog.  '),(5754,3413707,'25,May,2004','2004-05-25 00:00:00','   I apologize emensely (sp?) for my incomplete sentence---- Tucker is being mean though!!!!! ;(  I HATE EXAMS!  '),(5755,3413707,'25,May,2004','2004-05-25 00:00:00','   In case you want a study guide, I\'ve got one  urlLink here .  '),(5756,3413707,'24,May,2004','2004-05-24 00:00:00','   Does anybody have a study guide for Science, one with the definitions already on it?  If so, can you email it to me at  urlLink m_vssy@yahoo.com  '),(5757,3413707,'24,May,2004','2004-05-24 00:00:00','   urlLink Click the link!  '),(5758,3413707,'24,May,2004','2004-05-24 00:00:00','   it was actually grellow (my bad) and it\'s pieter\'s blog--------duhH?  '),(5759,3413707,'24,May,2004','2004-05-24 00:00:00','   duh taco.. i-princess am rupielover really tucker duh  '),(5760,3413707,'24,May,2004','2004-05-24 00:00:00','   Yea, who is rupielover. Do I even want to know...  '),(5761,3413707,'24,May,2004','2004-05-24 00:00:00','   Does anyone have any idea what Emily is talking about in the post at 4:02?!?! gellow plz?? Hey wait that\'s another broken rules, plz is short for people and that\'s against our rules, off with her head!!!  '),(5762,3413707,'24,May,2004','2004-05-24 00:00:00','   i fully agree katie (is that katie?). well anyways, i cant stand science, it\'s all higgeldy-piggeldy (sp?)  '),(5763,3413707,'24,May,2004','2004-05-24 00:00:00','   Spork-we should definately donate money to a we love dougie fund and unconditionaly love him on days when we are not donating money. haha. a pondering Oogly-Boogly...I must think about this...  '),(5764,3413707,'24,May,2004','2004-05-24 00:00:00','   pieter-----invite me into gellow plz--i think that\'s the name, im not sure  '),(5765,3413707,'24,May,2004','2004-05-24 00:00:00','   Watch out Matt!!! Sometime I\'m just gonna come up behind you and slap you up your side head!!! Watch out... I\'m watching you!!  '),(5766,3413707,'24,May,2004','2004-05-24 00:00:00','   dougie needs cheering up, but he dosent seem to want it at all. what are we going to do? he wants money or uncoditional love. should we all donate money to a \"we love dougie\" fund? or all uncoditionaly love him? leave a comment or put up another post titled DOUGWARD if you have an idea...  from spork  '),(5767,3413707,'24,May,2004','2004-05-24 00:00:00','   Well, it\'s official, (because I\'m saying it,LOL) Wednesday is cheer up Dougie day. In honor of this special day everyone is going to prove their loyalty to Dougie by sharing one funny, or inspirational story about Dougie.  '),(5768,3413707,'24,May,2004','2004-05-24 00:00:00','   Remember what I say in my post I\'m Baaaaaaaaack?  I said that I had not been eaten/stolen by wolves/gypsies, not Bob, and not studying for finals.  I was helping Cimmie to lose weight, remember????  Unfortunately, it didn\'t work.  '),(5769,3413707,'24,May,2004','2004-05-24 00:00:00','   You are indeed right, Sporko. I do have lots of friends. Lots of very good friends without whom I am uncertain as to what my fate would have been. And believe me, I love you all too.  In other news, Tucker created this awesome piece of work, possibly using declassified CIA technology:       It\'s going to be the cover for Ted\'s album when it is released. That should be happening fairly soon, with any luck at all. Oh, and you can click right  urlLink here  if you want to see a real photo of Ted.  '),(5770,3413707,'23,May,2004','2004-05-23 00:00:00','   what are you talkign about? dougie has lots of friends including me, his best friend. he just needs cheering up, we can all cheer him up. i declare wendessday to be \"cheering up dougie day.\" this is a mandority day of cheerieness. dougie, that includes you, be happy, we all love you. by hte way we are also all happy that quiggy is back, was it the qypsies? and last but not least emily i would like a lollypop but i am sorry i hate butterfingers. tho i would realy perfer a kitkat.  '),(5771,3413707,'23,May,2004','2004-05-23 00:00:00','   urlLink Hey U.G.L.Y.  is such a funny web site, you should check it out. The sad thing is I don\'t know whether it\'s a joke or not. They use U.G.L.Y. as an acronym describing the Unique Gifted Lovable You.  '),(5772,3413707,'23,May,2004','2004-05-23 00:00:00','   Everybody, stop worrying about me, all that happened was that I was helping Cimmie lose weight, and we all know how that goes.......  Anyways, I did not die.  '),(5773,3413707,'23,May,2004','2004-05-23 00:00:00','   Taco---why do you hate me so much? what did I ever do to you?  Dougie--- thank you for not kicking me out of this blog (and also feel better)  Rachel--- sorry, I\'m only featuring Butterfinger bars at my house, if you would like one please put up a post saying that you want a Butterfinger bar (we also have lollipops)  Quiggy--- come back  Phatness--- add more posts  Gus--- add more posts  everyone else--- don\'t hate me  P.S. my apologies are not fake  '),(5774,3413707,'22,May,2004','2004-05-22 00:00:00','   aw im sorry Dougie, Taco is right you need a friend! I vote that friday, May 28th be Montgomery School\'s Cheer Up Dougie Day! add a comment onto this blog entry saying \"agreed\" if you agree--------we all love you Dougie  '),(5775,3413707,'22,May,2004','2004-05-22 00:00:00','   I know you\'re gonna hate this Dougie, but i have to post this. This is  one of Dougies away messenges... \"I have nobody. My dad, the last person left in my life who believed in me, just got turned against me by my mom. My sister hates my guts and my mom and my dad think I\'m a slacker. My grades have dropped in school and I\'m going to end the year by disgracing myself. I\'m probably going to do badly on my finals. The ex-marine who used to live here is breaking into our studio to use it and he\'s messing with our dogs. Add in the whole unrequited love and teenage angst, and my life is 100%, unfiltered, premium, grade A shit\" I think this means that we have a friend in need, if everyone is willing to donate something that would cheer our heart-broken friend up, post it, donate to the \"boost-dougies-self-esteem-fund\". Thank You  '),(5776,3413707,'22,May,2004','2004-05-22 00:00:00','   Hi, you know you love  urlLink Apple , but only 1% of computer users use it. Most of the 98% of  urlLink Windows  users say that they don\'t use  urlLink Macs  because of compatibility (where only few of their beloved Windows programs are not usable on Macs.) I don\'t think that was among the things that bothered  urlLink this man  but he was definitely bothered. Don\'t get me wrong that I dislike Apple Computers (because I am about to switch, as I like them more than windows machines), but this is one of the funniest videos I have ever seen. This man speaks of his experience on the Mac very badly, but it is funny because although he \"dislikes Macs\" he still uses them and even edited the video on a Mac.  urlLink Check it out here  , and   enjoy .  '),(5777,3413707,'22,May,2004','2004-05-22 00:00:00','   i need a kit-kat bar. if anyone wants to me nice, they can bring a kit-kat bar into school tommrow and give it to me. thank you and good bye.  '),(5778,3413707,'22,May,2004','2004-05-22 00:00:00','   ::clap::, at least someone is on my side, I have nothing against Emily (well for this argument at least) but I just don\'t see what\'s she\'s adding to this blog, she just screws everything up, breaks all the rules and then does a stupid lie apology and does it again. It\'s really not cool.  '),(5779,3413707,'22,May,2004','2004-05-22 00:00:00','   i take it i have been banned from speaking in hacker, sayign i doubt anyof you can read hacker fluently (knowing a few words is NOT considered fluent.) blah, where did quiggy go anyway, i miss him. and emily the poll that tuck put up was \"mabye later if she breaks more rules\" and \"yes\" i personaly voted yes, you have broken enough rules as it is. 60% says \"mabye later\" its later we shoudl kick you out, youve broken enough rules. 40% thought we should have kicked you out a while ago. quiggy, we need your expert advice on this matter (or your password so i can kick her out.) for now, this is sporko reporting from SBC (sporko brodcasting company)  '),(5780,3413707,'22,May,2004','2004-05-22 00:00:00','   Um... Well... For the time being, yes, but when Quiggy comes back I\'m going to have to get his opinion.  '),(5781,3413707,'22,May,2004','2004-05-22 00:00:00','   can i curse in 1337?  '),(5782,3413707,'22,May,2004','2004-05-22 00:00:00','   It\'s fine. I know, Quiggy did break his own rule, but that was \'cause he had sort of... well, he freaked out, frankly. And we can use mild curse words, if you really want to. And, of course, you can curse as much as you want as long as it\'s in a different language. For instance, I frequently say bugger (although, that\'s just British slang, not a different language). And I am certainly going to miss you too. And you. And you. Heck, I\'ll miss all of you.  In other news, I figured out what was messing the blog up and I fixed it, so everything\'s back to normal, although you might want to check out the sidebar for the updated rules. I did some tweaking.  '),(5783,3413707,'22,May,2004','2004-05-22 00:00:00','   thanks dougie----again, I\'m sorry taco and quiggy and dougie. I do not understand why we cannot use curse words, seeing that quiggy broke his OWN rule by using the word--and i quote \"fucking\" in one of his posts-----again I am sorry, and i just wana keep this blog going as long as possible, I\'m going to miss all you guys soo much next year!  '),(5784,3413707,'22,May,2004','2004-05-22 00:00:00','   Do you think maybe we could keep the fightin\' and the feudin\' to a minimum?  To Tucker: Scroll down. Emily apologized, she had not read your post. I, at least, am not going to kick her out. Also, as you are not an admin you don\'t set the rules, although I would prefer mild swear words (or badly misspelled ones ::cough  urlLink danmmed  cough::) to a constant stream of not-so-mild ones. One other thing is, we don\'t all always spell everything right. I know I don\'t.  To Emily: Tucker is right in that a proper respect for the rules should be maintained. However, as you did apologize, I don\'t see any real problem.  To Quiggy: Where the heck are you, Quiggy!? This blog is succumbing to chaos! There is looting and pillaging in the streets! Somebody, light the Quig-signal!  We need you Quiggy!  '),(5785,3413707,'21,May,2004','2004-05-21 00:00:00','   how do you try to call me stupid and spell it wrong, jeez thats just sad.  '),(5786,3413707,'21,May,2004','2004-05-21 00:00:00','   ok you\'re right about one thing Emily, 66.7 percent voted to give you another chance and now you turn around and swear even more, then do an annoying \"smarty jones\" post, everything i asked you kindly not to do. Well, you just used your last chance, someone please kick Emily out!!! I\'m not trying to be mean but shes breaking every rule and grr....I\'m done.  '),(5787,3413707,'21,May,2004','2004-05-21 00:00:00','   I am sorry i didn\'t not read the post made by tucker at 8:34am, i will stop cursing (even though matt vessey broke his own rule) and I will use my own blog as my personal blog..........i don\'t want to make ppl angry or anything like that   P.S sorry taco  '),(5788,3413707,'21,May,2004','2004-05-21 00:00:00','   i think ur poll is stoopid (matt v broke his own rule of cursing so ya know what, i dont give a fuck....see, it\'s a curse word here are some more :fuck, btich, ass,damn, shit, motherfucker) haha ok ya well if u look at ur lil polly thingy, you will see that it says that 66.66% vote for me to stay in and 33.33% vote to kick me out so hah..............btw SMARTY JONES WAS AWESOME! and i also had a great hroseback riding lesson but that\'s besides the point. hey taco, i thought we had made a truse (spelling!) ok well i gtg byeness  '),(5789,3413707,'21,May,2004','2004-05-21 00:00:00','   I feel special the 100th post is here, and I (its Sporko) an honored to put it up. So I have to think of something meaningful for the 100th post on the blog of the damned. ::puts on trucker thinking hat:: ::thinks:: well, the trucker hat of knowledge hasn\'t helped me at all. phui. Well, this was fun lets do it again sometime. Whoever centered the whole blog (quiggy...) should be brutally beaten my me it looks weird, at least I think it does. P.S. Everone should please put up post titles.  '),(5790,3413707,'21,May,2004','2004-05-21 00:00:00','   &nbsp;&nbsp;&nbsp;&nbsp; Meg, everyone loves him/her but nobody knows her real sexual identity. That was all to change when we did a drive by of Sonoco\'s A-plus. Cimmie handed the slip asking whether she/he preferred Coke or Pepsi which Taylor had recently typed up, and...after much speculation over the whole \"it\'s for a school newspaper\" alabi, Meg filled out the form. The first field was her name which we told him/her was unnecessary to fill out so she left it blank. Right below that field, however was the most important field...the gender question. It asked wheter or not Meg was a girl or if she was a boy. Taylor waited until we left A-plus to read it. And it read FEMALE. I was shocked Meg is/was/will be a female. I was convinced she was a he, but I am starting to face reality and I can\'t wait to see Meg over the summer when I visit A-plus.  Oh and P.S. Emily, USE TITLES!  '),(5791,3413707,'21,May,2004','2004-05-21 00:00:00','   Have you ever had a craving to make your own song? Get a laugh? Eat some Chinese food? No? Well you need to go to this website to get some inspiration and a laugh, but I don\'t think they offer Chinese food. I am not going to ruin any surprises but it sounds pretty cool if you type in the only large field in the website (I wont say what it\'s labeled)\"I found this place not to long ago, This is too much fun. I know what I will do tonight.\" Don\'t ask someone told me to do it and it the result is surprisingly good. You wont regret clicking on  urlLink this link . Enjoy!  '),(5792,3413707,'21,May,2004','2004-05-21 00:00:00','     '),(5793,3413707,'21,May,2004','2004-05-21 00:00:00','   Stop swearing it\'s not funny and it breaks one of the few rules of this blog, everything you post is so annoying!!! Ok, now it\'s mine turn, i say we give Emily one last chace and if she does another \"smarty jones\", swearfest or dissing another blogger blog then we kick her off.  '),(5794,3413707,'21,May,2004','2004-05-21 00:00:00','   shitmotherfucker that sucks sporko!!!!!! ahh mutters *stupid siblings*!!  '),(5795,3413707,'20,May,2004','2004-05-20 00:00:00','   my life is over. over over over. my parent are in boston this weekend, and i needed a ride home from school. so i ask polly, she says yes. we go to her house and just hang out with gabo, but wallball hurt. so then we go bowling, that was fun, before we went i called my brother telling him i would be home a bit late mabye 10. i told him not to worry if i dont get home until then, i told him where i was going and who i was with, everything. the moment i get home my mum has come home early, a day early. so she yelles that she didnt know where i was and how worried she was. and the thing is my brother forgot where i was and who i was with. so i cant do anything tommrow. this deserved a uber blah.  '),(5796,3413707,'20,May,2004','2004-05-20 00:00:00','   i forget what i was gonna write in here-----------IM GOING TO SEE SMARTY JONES (HORSE) WORK OUT TOMORROW AT 8:30 AM, HES DOING A PUBLIC WORK OUT AT PHILIDELPHIA RACTRACK, GET UP AT LIKE 6:30 TEH GATES OPEN AT 8-----------BE THERE!!!!!!!!!!!!!!!  '),(5797,3413707,'20,May,2004','2004-05-20 00:00:00','   hello all i have finally got here it be gus.  I hope you had a good time bowling.  So I would like to say how HORRIBLE finals are.  '),(5798,3413707,'20,May,2004','2004-05-20 00:00:00','   All right, I got rid of the pigs. They were bothering me, too. However, anybody who still wants to see them may do so by clicking  urlLink this link .  '),(5799,3413707,'20,May,2004','2004-05-20 00:00:00','   Get rid of the stupid pigs, it is so annoying!!!!!!!!!!!!!!!!!!!!!  '),(5800,3413707,'20,May,2004','2004-05-20 00:00:00','   I\'m so stupid, hm....dan dan\'s blog, thats interesting ::click::, wow, I can\'t get any stupider Dan Dan+ Blog?!?! Ok, you can\'t count one of the hits on your brainy thingy cuase I don\'t count since I\'m on a team ( I guess, not that I did anything..).  '),(5801,3413707,'20,May,2004','2004-05-20 00:00:00','   ben----you are an innocent little child, come on AIM  '),(5802,3413707,'20,May,2004','2004-05-20 00:00:00','   urlLink Phatness World   Thought everybody would like to know about the existence of this blog.  '),(5803,3413707,'20,May,2004','2004-05-20 00:00:00','   urlLink http://www.blogpoll.com/poll/  '),(5804,3413707,'20,May,2004','2004-05-20 00:00:00','   sorry to interupt all that forgivness, but didnt quiggy say you must speak in ill words on this blog. full words, emily combe full words!   you know i think we should boycott finals, i mean, no one wants to take them so i guess boycotting is the answer. whos with me?  '),(5805,3413707,'20,May,2004','2004-05-20 00:00:00','   This is something of a good-bye post, as I am leaving you all forever. Just kidding. However, I may not be blogging much, if at all, over the next week, due to the parental units coming down hard on Intarweb use. Sorry.  '),(5806,3413707,'19,May,2004','2004-05-19 00:00:00','   IM SORRY TACO! (IF I KNEW HOW TO MAKES THIS BOLDED AND UNDERLINEDED I WOULD, BUT I DONT SO I CANT VER-WELL DO SO------SORRY)  '),(5807,3413707,'19,May,2004','2004-05-19 00:00:00','   hey taco! im sooooooooooo sorry *gives taco hug, then thinks to self and immediatly stops* sry, i wana b ur friend, and sporko, im sry for insultin taco and mattness in sry for insulting ur blog (u too dougie) Im gonna miss you all next year, i lvoe you all! (even you carson) and if we r now alloawed to curse i wouold like to say one thing: fuckdamnbitchasswhoreshitmotherfucker i do not wana go to shipley!  '),(5808,3413707,'19,May,2004','2004-05-19 00:00:00','   Just wondering if Dougie and whoever else is in TED knows when and where we are going to record on Saturday, please post if you know cause I really need to find out.  '),(5809,3413707,'19,May,2004','2004-05-19 00:00:00','   i am not an object tucker, i can not be stolen. but in this case i am so glad you took me away. blah. new danmmed meaning i must check this out. ::opens new window:: ::laughs:: ::feels a strong urge to slap tucker:: its all true except the liking girls and gay guys part. ::requests that urbandictionary delets it:: heh heh heh. your a horrible horrible person tucker. horrible yet so amusing.  '),(5810,3413707,'19,May,2004','2004-05-19 00:00:00','   Did anyone see the new definition of danmmed, its pretty danmmed funny.  '),(5811,3413707,'19,May,2004','2004-05-19 00:00:00','   Dear Emily (needs-a)Comb,  I am posting this just to say that I\'m sorry if I did something that triggered that outbreak where you demanded that I leave this blog. it is in my most heartfelt manner that I ask why you demanded that I leave the blog, other than the fact that I stole sporko from you I can see no reason for you to hate me. Because of that outburst I have decided to kill you....with hugs and caring.  Thanks for your time, Taco  '),(5812,3413707,'19,May,2004','2004-05-19 00:00:00','   Hey someone tell me how make one of those brain-eaty websites. That would be awesome, we should start tomorow.  '),(5813,3413707,'19,May,2004','2004-05-19 00:00:00','   And also, Emily, Tucker has as much right to be here as any one of us. Same goes to you, Carson.  '),(5814,3413707,'19,May,2004','2004-05-19 00:00:00','   quiggy, you broke yor own rule of no swearing, and i salut you for doing it. and why does emily combe hate tucker so much? whats up with that?  '),(5815,3413707,'19,May,2004','2004-05-19 00:00:00','   THIS BLOG WAS DESIGNED TO BE A BLOG FOR ALL MONTGOMERY SCHOOL 8TH GRADERS.  LAST TIME I LOOKED, ALL OF YOU FALL INTO THAT CATEGORY.  IF YOU DON\'T LIKE IT, THEN SUCK IT UP AND MOVE OUT OF THIS BLOG, BECAUSE YOU FUCKING WELL DON\'T DESERVE IT!!!!!!  '),(5816,3413707,'19,May,2004','2004-05-19 00:00:00','   First of all, Dougie, the abbreviations USB and CSB stand for United States of Blog and Confederate States of Blog, not Universal Serial Bus and ...... whatever the other one was.  Tonight, the contest shall begin.  The one most important rule is that you  MAY NOT MENTION THIS CONTEST ON BLOG OF THE DAMNED, NO CHANGING IT IN ANY WAY .  Graaaagh!!!  '),(5817,3413707,'19,May,2004','2004-05-19 00:00:00','   mmh, that was you? i was sure it was dougie. by the way, if anyone is wondering (witch i doubt anyone is) i am on dougies side for the brain eating contest. oh and another thing quiggy, you have to add oogoly boogoly to the urban dictionary. i wont because i am too lazy and it involves too much typing. blah.  '),(5818,3413707,'19,May,2004','2004-05-19 00:00:00','   I can\'t believe you?!?! I came up with gunker not dougie!!! (nothing against dougie but...) I demand you change your blog now or, or.....BLAH!  '),(5819,3413707,'18,May,2004','2004-05-18 00:00:00','   I just read the rules and saw that little rule thing, my bad, just ignore the post and don\'t kick me off!!  '),(5820,3413707,'18,May,2004','2004-05-18 00:00:00','   Hey I\'ll do this brain eating thingy, I call dougies side and Matt has to refresh his site so the hits he already has dont count, also what do the team names stand for?  '),(5821,3413707,'18,May,2004','2004-05-18 00:00:00','   Doug and I will probably be heading the Brain-Eating contest starting tomorrow night.  If you\'re interested, please comment on this post, stating which side you want (USB or CSB), and who you want for your captain.  Tomorrow night is the deadline.  '),(5822,3413707,'18,May,2004','2004-05-18 00:00:00','   blah, well congrats tucker for joining the blog. by the way, i love quiggys idea for the brain eating contest. i claim captain of one of the sides.   \"danmmed it all to hell.\"  '),(5823,3413707,'18,May,2004','2004-05-18 00:00:00','   Hey everybody, it\'s Taco, just saying that i finally got the stupid link to work and I\'m in and blogging!!!  '),(5824,3413707,'18,May,2004','2004-05-18 00:00:00','   I have a new idea, quite possibly the best one yet.  We should split into two teams, the United States of Blog and the Confederate States of Blog.  Each team gets their own blog.  Each team also gets their own Brain-Eating account.  For one week, each team tries to eat more brains than the other team.  You may affect B of D, however the changes must be fixable.  Doug and I would be on seperate teams, and would not make any new administrators or delete old ones.  Vote in the new poll to vote on whether or not this is cool.  '),(5825,3413707,'18,May,2004','2004-05-18 00:00:00','   If you like Bush, don\'t click this urlLink link  '),(5826,3413707,'29,June,2004','2004-06-29 00:00:00','   ...if you\'re desperately bored or have an addiction to voting in polls. In case the previous sentence didn\'t tip you off, the Blog of the Danmmed is proud to bring you another poll! The results of the previous poll indicated that Quiggy and I are evil mutant lobster chickens, which would explain why typing is so difficult. Thank goodness for mail-order helper monkeys!  '),(5827,3413707,'28,June,2004','2004-06-28 00:00:00','   Um... yeah, in case anyone is wondering why i dont answer calls and email, its becuase im going to cali. Oh yeah and when i come back my mom is making me \" be more socail\" and do a bunch of camps so if dougie or gabo or any of the other people in that art camp feel the need to include a struggling cartoonist into your little portait and landscape posse just post the info about the camp, thanks.  '),(5828,3413707,'27,June,2004','2004-06-27 00:00:00','   Ok, Dear Dopugie/ Ben/ Matt/ anyone who can help, How do I post an image on a blog? Gabo  '),(5829,3413707,'26,June,2004','2004-06-26 00:00:00','   Hey Quiggy, could you put up a link for Evan\'s/my blog?  It\'s http://rhythmsection.blogspot.com/.  Thanks  '),(5830,3413707,'26,June,2004','2004-06-26 00:00:00','   Well, seeing as how the reaction seems to be a resounding thumbs down, I will remove the post, but I still think that the song is hilarious.  I sincerely apoligize to anyone who was offended.  If the admins care to warn me, so be it.  Once again, I\'m sorry.  '),(5831,3413707,'25,June,2004','2004-06-25 00:00:00','   That was the greatest concert I\'ve ever been to...ever! Ok, Peter Frampton was amazing on the guitar. For those of you who love the guitar I suggest you listen to one of his songs. But, STYX IS THE GREATEST BAND!!!!!!! I know most of you expected Styx to be old and crippled by now, but it has been more than 4 decades ago since Styx started (so they must be in thier late 40\'s, early 50\'s, or late 50\'s) but they were dancing around up there jumping over the equipment, doing highkicks, and seemed they were in their 30\'s! Their voices even sounded exactly the same from the early years! The person who played the keyboard (and did a lot of the lead singing) is officially my new hero. He\'s also really amusing. Randomly in the middle of the concert he turned to the rest of Styx and said, \"Well, we know you can rock, but can you...rag-time?\" And he started playing rag-time on the keyboard and everyone on stage started square dancing. And all of Styx was steaming hot, with their tight, tight, tight, tight pants, esp. the keyboardist who had the tighest pants of all, and...well, I\'ll just stop there, for all of your sakes...STYX ROCKS!!!!!!!!!  '),(5832,3413707,'25,June,2004','2004-06-25 00:00:00','   Neon (aka Evan), I have a compromise.  If you never try to destroy BotD, I am willing to have you have our members joining your  urlLink blog , so long as you never insult BotD\'s admins (in an admin sense) or try to destroy BotD.  If you agree, send me an invitation to quiggy@gmail.com.  Cool!  '),(5833,3413707,'24,June,2004','2004-06-24 00:00:00','   HI YOU DONT KNOW ME BUT I AM ALAINA, CARSON\'S GIRLFRIEND. AM I SUPPOSED TO BE TYPING STUFF UP ON THS BLOG. OH WELL I DONT CARE. BUT ANYWAYS. I AM ALAINA AND I LOVE CARSON.  '),(5834,3413707,'24,June,2004','2004-06-24 00:00:00','   all i have to say about everything today: its purple-y pink-ish.  =spork=  '),(5835,3413707,'22,June,2004','2004-06-22 00:00:00','     '),(5836,3413707,'22,June,2004','2004-06-22 00:00:00','   jeez people i dont no who neon is!!! im not trying to be mean, just please tell me who it is,i didnt no if he was a complete stranger or w/e, is he the one who posted on the crap on the chatterbox??  '),(5837,3413707,'21,June,2004','2004-06-21 00:00:00','   Tomorrow we are going to see The Terminal starting @ 5:00pm in the Downingtown 16 theater.  Come if you feel the need, invite who you wish.  See you there!  '),(5838,3413707,'21,June,2004','2004-06-21 00:00:00','   Quiggy is not being oppressive; he is discriminating accordingly. So far in its virtual existence, The Blog of the Danmmed has existed as a haven for Montgomery students, now ex-Montgomery students of course, and he is merely sticking by the principles on which this blog was founded. While in the future there may be exceptions to this, you have not made an overly convincing case for yourself. I hope that future animosity between the denizens of the BotD and yourself may be avoided, and I do apologize if the current configuration of the BotD has caused you any distress. However, we are currently unable to fulfill your wish to join, and as your present actions contradict any perceived desire you might have to join, we do not see a possibility of a wish of that nature being fulfilled in the foreseeable future. Once again, I apologize for the inconvenience.  '),(5839,3413707,'21,June,2004','2004-06-21 00:00:00','   i dont no who u are but if you dont/havent gone to montgomery you cant join...there is nothing oppresive about that, and no one things you should join, especially if you keep being annoying on comments and the chatterbox  '),(5840,3413707,'21,June,2004','2004-06-21 00:00:00','   Who is Neon, and HOW THE HECK ARE DOUGIE AND I BEING OPPRESSIVE???  That new rule is already gone, and was simply to try to catch who ever was making fun of people via the Chatterbox.  '),(5841,3413707,'21,June,2004','2004-06-21 00:00:00','   When are you going down to Stone Harbor?  '),(5842,3413707,'21,June,2004','2004-06-21 00:00:00','   I have been pretty bored this summer. I have watched tv and played video games and sometimes gone to the pool. Now that the public schools are out I will actually have someting fun to do.  '),(5843,3413707,'20,June,2004','2004-06-20 00:00:00','   The new rule is disbanded.  '),(5844,3413707,'20,June,2004','2004-06-20 00:00:00','   First of all, I know who wrote the stuff on the Chatterbox.  It was Evan\'s brother, Matt.  Evan and Matt are down in NC with me and that is my IP address.  He used Evan\'s laptop.  Because I revealed the evil (not to say hilarious) prankster, we should not have the chatterbox post/blog post rule.  If that rule stays in effect, I will have to flood the blog again, and remember how irritating that got?  I wil also flood the chatterbox just to get rid of the great prank messages.  '),(5845,3413707,'20,June,2004','2004-06-20 00:00:00','   Just so everyone knows, I am temporarily enforcing a new rule: if you post,  you must also enter a message into the chatterbox   This is twofold: it will help remove the offensive messages, and it will help me catch the doer of these evil doings.   SHOULD YOU NOT LEAVE A MESSAGE, I MAY ASSUME THAT IT IS YOU!!!  '),(5846,3413707,'20,June,2004','2004-06-20 00:00:00','   Attention everybody:  this is a public service announcement.  The chatterbox is completely rude and offensive.  Unless the person responsible comes forward, they will be kicked out when they are discovered.  Should they come forward, they will only recieve one warning.  Thank you.  '),(5847,3413707,'20,June,2004','2004-06-20 00:00:00','   hey guys, this is Evan under Ben\'s username asking if I can join.  \'Twould be quite fun, you guys have quite an operation going here. Anyway, have fun voting and the like!  '),(5848,3413707,'18,June,2004','2004-06-18 00:00:00','   i saw the ted store, awesome but we need some shirts... and do i get any money for it, comsidered i am one of the co-founders of the band and i drew the picture...so yea make some shirts and ill buy. but i should get them for a flat price  '),(5849,3413707,'18,June,2004','2004-06-18 00:00:00','   i saw the ted store, awesome but we need some shirts... and do i get any money for it, comsidered i am one of the co-founders of the band and i drew the picture...so yea make some shirts and ill buy. but i should get them for a flat price  '),(5850,3413707,'18,June,2004','2004-06-18 00:00:00','   Thanks whoever put my personal blog up on the side.  I realise, however, you may have done that a while back but seeing as I am sometimes slow, it is completely understandable.  Thanks again! Kels  '),(5851,3413707,'18,June,2004','2004-06-18 00:00:00','   I just hope none of you are fathers...(especially sporko or emily combe...who would be the father there?...)  '),(5852,3413707,'18,June,2004','2004-06-18 00:00:00','   ok, sad story, the other day i ordered a laptop from dell (graduation present!!::cheer::) it was all good until yesterday when a got an email...\"becuase of a unexpected error your shipment will not arrive till the first of july\" so imthinking ok that sucks but i can live... then i get this \"your order has been lost, please reorder\"... so i call up dell... and i get this \"if your are calling about a lost order please understand that a computer error has occured, please reorder\"...a computer error YOUR A COMPUTER COMPANY!!!! you have millions of employees who make and program computers HOW CAN U HAVE A ERROR!!!! ok, if they cant make computers that work for the people who make them, then why would i want to buy one for me...  '),(5853,3413707,'17,June,2004','2004-06-17 00:00:00','   i have one single question: is every ones summer as lame as mine? i mean i have so much time its getting scary. comment if you actualy have a life, well wait, if you have a life you wouldnt be on the internet. well then comment if you dotn have a life. =sporko= p.s. I WANT TO GO TO THE FEZ!!! but no, my mum is taking me to some other stupid concert. ARRRGH!!! holy unfairness batman.  '),(5854,3413707,'17,June,2004','2004-06-17 00:00:00','   urlLink click here  does anyone else thing that this has all gone a little to far...  '),(5855,3413707,'16,June,2004','2004-06-16 00:00:00','   A new poll is up. I thought it was time since we already decided and acted upon the last one. Vote danmm you, vote!  '),(5856,3413707,'16,June,2004','2004-06-16 00:00:00','   coolest thing happened, ok so i got this new pedal ok...so im playing around with the knobs and i try turning everything to full treble, then suddenly i hear this weird noise so i play around with it and and get rid of most of the feedback then suddenly the noise come throught clear and its a JAPANESE RADIO STATION!!! i dont no how its possible but my new pedal picks up radio stations that arent from anywere near here, i even checked no japanese or chinese radio for a few hundred miles... but i still get it loud and clear, i have proof from my mom and brother they both heard it.  '),(5857,3413707,'15,June,2004','2004-06-15 00:00:00','   Hmm, I just saw Franz Ferdinand (the band, not the historical figure) on Late Night with Conan O\'Brien. They\'re pretty good, even live. I\'ve even decided to buy the album. I love  urlLink iTunes , I don\'t even need to leave the house - pathetic, but convenient.  '),(5858,3413707,'15,June,2004','2004-06-15 00:00:00','   if someone agrees to actualy let me be in ted, i would still say no. anyway i dotn think your high voice would go (at all) well with the music ive heard.  '),(5859,3413707,'15,June,2004','2004-06-15 00:00:00','     '),(5860,3413707,'15,June,2004','2004-06-15 00:00:00','   hey members of TED or w/e ur new name is... can i please please please join ted? i mean matt vessey is now in ted, and he plays the clarinet i wana sing in ted.....yeah thats all i had to say oh wait...one more thing.....i wont be mean to eric or tucker if ur wondering, i kidna was stupid about that during my time at montgomery...i wish we all had been better friends  '),(5861,3413707,'15,June,2004','2004-06-15 00:00:00','   I\'m thinking about getting stuff so that I can play Dungeons and Dragons with people.  If you don\'t know what it is, it\'s the nerdiest pencil and paper game on the face of the Earth, but it\'s fun.  I was thinking that a grout of people would get together at my house every friday or something and mess around and play and whatnot.  It\'s mainly to stay in touch, and if you want to play, just post something.  Happy nerdiness!  '),(5862,3413707,'15,June,2004','2004-06-15 00:00:00','   urlLink'),(5863,3413707,'14,June,2004','2004-06-14 00:00:00','   Yeah about Dougie\'s award: like every other day, Dougie is the greatest person we all know and doesnt get credit for, they forgot to give him the award!!! So I say we all give him a little cheer for winning and a virtual pat on the back...  '),(5864,3413707,'13,June,2004','2004-06-13 00:00:00','   Why was the name changed... I don\'t really mind.  I am just wondering.  Thanks if you should answer.  '),(5865,3413707,'13,June,2004','2004-06-13 00:00:00','   am i the only one who relized that either guiggy or dougie changed the name of the blog?? thank you so much for maiking me feel loved and changing the name from damned to danmmed i feel so, whats that word, special. anyway, thats awesome, thank you. so... are we ever gonna have that pool party, and when? i mean pick someones houes and plan it, cause i can go to either. there is just one big peoblem, the boy/girl sleepover is fine with me but i dotn think its gonna go over well with my parents, gabo and tuck know how paraniod my rents are. but if you set something up ill try to pursuade them. blah!!! love and all,  =sporko=  '),(5866,3413707,'13,June,2004','2004-06-13 00:00:00','   After about a month of trying get the link for the blog of the damn to work I finally was able to join. Oh and by the way I think it would be awesome if we had a pool party at night and I would defintley want to go if i\'m not busy.  '),(5867,3413707,'13,June,2004','2004-06-13 00:00:00','   Doug or Matt, I have a personal blog too if you wanna put it on the side thingy.  my username on here is titanic1fan (don\'t ask because you already know... or you don\'t want to). THANKS  '),(5868,3413707,'12,June,2004','2004-06-12 00:00:00','   Hey Everybody! Thanks so much for votin\' yes to allow me in and those that voted no... thanks anyway.  So anywho the pool thing sounds like fun but I am dreadfully confused.  Who\'s house?  Emily\'s or Tori\'s???  Well the one thing I know is that it won\'t be at my house on account of the fact that if anybody wanted to swim... it would be in my bathtub.  Well best of luck to whoever figures this out.  I will come though... where ever and whenever it may be.  So that is all for now. Cheerio,  Kelsey  '),(5869,3413707,'12,June,2004','2004-06-12 00:00:00','   Blogger now offers free  urlLink audioblogging !  And also, notice anything different about the blog?  '),(5870,3413707,'12,June,2004','2004-06-12 00:00:00','   my rents wouldn\'t care if i had a boy/girl sleepover if everyone else agreed....my pool doesn\'t have changing color lights but it does have lights by teh side of the pool. i actually had a night-time pool party one time....we found a frog...in the pool...it was er, interesting, but fun!anyways, my house is pretty big, i have a living room, a family room, my room, a computer room, and a game room and a guest room in teh attic that could be used, also, i have a large backyard if people wish to camp out. i believe i actually have a few tents. there\'s also other stuff to do.....we haev a huge drum set in the attic of teh garage if anyone wants to play it (no one ever does, sporko u r welcome, i know u like the whole precussion thing). so yeah...call me if ya wana do this....(610) 651-2878.  '),(5871,3413707,'11,June,2004','2004-06-11 00:00:00','   Prospective members of the Blog of the Damned: Comment to this and leave your email address. Specifically I am talking to Kelsey and Phatness, but I suppose others can do it too if they want. And Phatness: I think we\'re just going to have to get on a conference call and go through this step-by-step or something, because we have sent you a  lot  of invites. Oh well. Anyway, I need your email address if you want to join.  '),(5872,3413707,'11,June,2004','2004-06-11 00:00:00','   I\'M GOING TO A STYX AND PETER FRAMPTON CONCERT THE 25TH!!!! By the way, who the hell is Peter Frampton?  '),(5873,3413707,'11,June,2004','2004-06-11 00:00:00','   ::sigh:: i want to see stepford wives, but my mum says i cant see it, she says it got bad reviews and she thinks it looks bad. dont you love the fact that she thinks if she wont like it i wont like it, heh heh so not true. and i think that goes for riddick too, sorry tucker. ::sigh:: good news for gus tho.... i can see shrek2, and now for the bad news, not today. since my dad broke his leg he thinks he should be pampered, so we are talkign him to longwood gardens, and guess who gets to push around the wheelchair? you guessed it yours truly. joy. and now for more good news (bad news for you gus) we cant see shrek2 tommrow either for its my bros birthday party and we are going paintballing (yay!). bad news for dougie, and tucker tho, i am oh so sorry i cant bring anyone, plus i doubt any of your parents are gonna want to sighn the waver required by anyone under 21. heh heh heh its a fun but dangerous sport. so have fun at the movies ::sniff sniff:: with out me... -sporko- p.s. i am free monday!!! w00t!!!  '),(5874,3413707,'11,June,2004','2004-06-11 00:00:00','   Blah, Stepford Wives...hm...I say Chronicles of Riddick... but that\'s just me...  '),(5875,3413707,'11,June,2004','2004-06-11 00:00:00','   tomorrow some people are going to see The Stepford Wives in Frazer @ 4:30pm - come if you feel the need, tell anyone who\'s interested  '),(5876,3413707,'10,June,2004','2004-06-10 00:00:00','   Members of Ted: go to the  urlLink TedBlog  for an important announcement.  '),(5877,3413707,'10,June,2004','2004-06-10 00:00:00','   Here\'s an idea for that pool party: if we were to have a pool party at night, it would probably be best to have a sleep-over because I have a feeling that my parents, at least, would be somewhat less than willing to pick me up at someone\'s house at whatever it may be, 2 AM or whatever. So, as I said, if it were a sleep-over it could go as late as we wanted without any worries.  Just an idea.  '),(5878,3413707,'09,June,2004','2004-06-09 00:00:00','   happy birthday RJ! RJ is my older brother and its his sweet 16. i truly have no idea why i am tellign you this. but its cool, sunday RJ, RJs friends and i are goign paintballing. that ought to be amusing (i am realy quite good at painball w00t!) anyway.. if you want to wish RJ a sweet 16 please leave a comment... hed realy like that... happy birthday RJ! -spork townsend  '),(5879,3413707,'09,June,2004','2004-06-09 00:00:00','   Calling all members of TED, your mission, if you chose to accept it is to get Dougie to get everyone together to record again, soon!! So Dougie, invite them and they will come. I just got a ton of crazy like hard-core garage rock stuff and a hard-core garage rock pedal so I need to record and it\'s gonna sound so cool with like the wailing clarinet in the backround and then erics have to do a solo with my pedal!! It\'s gonna be amazing.  '),(5880,3413707,'09,June,2004','2004-06-09 00:00:00','   righ tnow i am sitting on my couch in my living room watching the west wing, eatign apples and drinking water (i wonder how long i can live on water and appels) i am talkign to my bro about witch movie we are gonna watch. i want to see the boondock saints he wants to see donnie darko (i watched donnie darko a few days ago, thats just too much donnie (no matter how hott he is)) i think we are gonna watch the boondock saints its got more than 230 swears (about 3 a min) its terribly violent and vuglar. anyway, i felt like telling [insert your name here] what i am doing. by the way i am gonna close my poll soon so vote now! (red? purple? green?)  '),(5881,3413707,'08,June,2004','2004-06-08 00:00:00','   I\'M NOT SICK ANYMORE!!! I\'M FREE!!!  '),(5882,3413707,'08,June,2004','2004-06-08 00:00:00','   The Nighttime Pool party is officially ON.  I have yet to send out an invitation (so it will be official) but it is most deffinitely on.  It\'s going to be within the next two weeks (request of spork) and it shall be ultra cool, don\'t drool.  Maybe next friday...i don\'t know.  tell me what you think and then i shall do what I want anyways.   ((tor))  '),(5883,3413707,'08,June,2004','2004-06-08 00:00:00','   Everyone, it is imperitive that you visit www.engrish.com.  Once again, my link thing isn\'t working.  This stuff is great.  Just go.  Now.  '),(5884,3413707,'08,June,2004','2004-06-08 00:00:00','   Gus, I found this picture on Engrish.com.  I thought you might like it.  It\'s right up there it \"Beware of Doug.\"  My link doohickey isn\'t working, so just copy and paste.  http://www.engrish.com/recent_detail.php?imagename=poisonus_gus.jpg&category=Signs/Posters&date=2004-06-04  '),(5885,3413707,'08,June,2004','2004-06-08 00:00:00','   I have a really cool idea.  Sometime this summer we should throw a pool party...at night.  My pool has lights (that change colors:)) so it would be safe, but it\'s have to be really late.  Anyways, tell me what you think because as I said before, I think it\'s a fairly cool thought.((tor))  '),(5886,3413707,'07,June,2004','2004-06-07 00:00:00','   to the whole group of PSLs: (gabo, tuck, sporko, dougie) we are meetign at my house rom 11-4 just to hang out. if you wish the pool is open so bring your bathing suit (if you wish). and er... um... yeah. by the way its me, sporko and we are meeting at my house (aka the paperbag)  '),(5887,3413707,'07,June,2004','2004-06-07 00:00:00','   Sorry again about that, but i\'ve been shut in my room for that past three days, am living on water, and am about to completely spaz.  So, humor me and take this poll. I finally remembered about my appointment at the eye doctor to get some colored contacts because they are cool.  Now I don\'t know which color I should get first. Please vote as I shall do as you bid.  '),(5888,3413707,'06,June,2004','2004-06-06 00:00:00','   Dougie please re-invite me to the TED blog...  '),(5889,3413707,'06,June,2004','2004-06-06 00:00:00','   I know most of you, know all the cool things about google like that it\'s can search telephone numbers and give you addresses and names but did you know that if you type in a math problem it solves it?? Now that is a high quality search engine.  '),(5890,3413707,'06,June,2004','2004-06-06 00:00:00','   Can someone please reinvite me to the TED blog, and make me a administrator!!!  '),(5891,3413707,'06,June,2004','2004-06-06 00:00:00','   It\'s the future of concerts!! I was at the Blink 182/ No Doubt concert, which was seroiuly rockin\' and they had this thingy, were they played like a slower song and they were like \"before many of u were born, in the late 90\'s, during a slow song poeple held up lighters, but now it\'s the techonloy age and everything so in recongion of this ever hold up you\'re cell phones instead\". Cool idea right, well its was awesome, one of the coolest thing\'s I\'ve ever seen. Just had to share that moment, and i got a cool pic but it was too dark and didnt come out right.  '),(5892,3413707,'06,June,2004','2004-06-06 00:00:00','   20 things I have learned Through My eighth Grade Experience: 1. Before you insult someone, make sure they are not standing behind you. You will save yourself much explaining and stuttering and awkwardness this way 2. TRY to remember where you left your bag - they have a tendency to move around 3. learn how to open your locker rather than having someone else do it all year long 4. Never volunteer if you don\'t know what you will be doing 5. Be creative with your excuses 6. continuation of #5: become emotional 7. DON\'T GIVE PEOPLE YOUR PENS (YOU WILL NEVER SEE THEM AGAIN) 8. Visible bra straps are a no-no 9. The same goes for clear dresses 10. Don\'t be downwright mean to anyone without proper contemplation - you might need them to do something for you later 11. When in doubt, employ the phrase \"I didn\'t do it!\" 12. be inconspicuous when chewing illegal gum 13. Always get out of running laps - again, be creative and remember, it\'s all in the timing... 14. Make friends with the lunch people 15. Wear the specified socks  16. EVERYTHING has a loophole - it\'s your job to find them 17. listen to everyones advice and then promptly disregard it (haha ms. V and the box projects) 18. remember that you will be seeing these people almost every day for the next year - be nice to them 19. inside jokes will not seem funny when you tell them to others - hence the name inside jokes. And...last but not least... 20. Stand out  I hope you all feel very knowledgable now. --feel free to post your own knowledge you acquired this year  '),(5893,3413707,'06,June,2004','2004-06-06 00:00:00','   Kelsey, as most of you know, was a former member of the Montgomery School class of \'04.  She is asking for admittance to the blog.  Should we allow members of our class who are no longer members admitance into BotD?  Vote in the sidebar to cast you opinion.  '),(5894,3413707,'06,June,2004','2004-06-06 00:00:00','   Sorry, that was my stomache.  Begging for more water.  I\'ve been living on water lately.  Oh well, at least I have no chance of getting dehydrated.   Anyways... people who have control over these things (sadly not me): FIX THE LINK TO MY BLOG!   PLEASE! it just takes you to a page that says error or something.  I think it\'s because I deleted my other blog and started another one, but it\'s under the same username so it should work...i hope thank you for your time  '),(5895,3413707,'06,June,2004','2004-06-06 00:00:00','   I was trying to put that into my own blog, but I\'m sure you would love to see it anyway...  '),(5896,3413707,'06,June,2004','2004-06-06 00:00:00','   Ok, today we have my favorite music list, stolen straight from the blog of Gabo. (not in order)     THE WHITE STRIPES!! ::  (::insert loud cheering here::) um...well everything, there are only about 2 or 3 white stripes songs i dont love, but i still like all of it. But if I had to choose a favorite it would be the self titled cd. NIRVANA!!: they have this three disk CD thingy, titled Nirvana (hm..how orginial), thats the best one. Franz Ferinand: one of the CD\'s in my stereo now, is there one and only CD, self titled Phantom Planet: yes Carson, I do like some emo, both of Their CD\'s are awesome but the second one is my favorite Jimi Hendrix: Are you expierienced (or the greatest hits, but thats kinda cheating) Led Zepplin: IV and III Crystal Method: greatest techno band ever, um..it\'s the CD with the crab on it, damn, don\'t remember the name     Ramones: again I\'m gonna cheat and say Greatest hits Red Hot Chili Peppers: By the Way Vines: Winning Day\'s Weezer: Pinkerton (yes more emo, but only pinkerton) VON BONDIES!!: amazing band, you have to buy Pawn Shoppe Heart Hives: Veni Vedi Vicuois (expect spelled right) Less than Jake: fine I give up, Hello Rockveiw and Anthem are good, but Pezcore still sucks!! Muse: have no idea what the album is... AC/DC: Back in Black, you HAVE to buy this Modest Mouse: damn the radio for over playing this song, this was MY song, i predicted it would get big and now everyone sings it :: sob:: Black Sabbath: We Sold our Souls for Rock and Roll, best quote ever...\" I was into sex, drugs and rock and roll, was paying for the sex and the drugs\"- Ozzy Bush: Sixteeth Stone, everything zen is a awesome song... Metallica: Saint Anger Green Day: (I\'m cheating again) Greatest Hit\'s Yeah, that\'s a pretty good list but I\'m sure I left someone out...  '),(5897,3413707,'06,June,2004','2004-06-06 00:00:00','   I\'m so lonely...what the hell ever happened to the PSL\'s?  '),(5898,3413707,'05,June,2004','2004-06-05 00:00:00','   urlLink I\'m ahead Dougie!!!  '),(5899,3413707,'05,June,2004','2004-06-05 00:00:00','   i have one question for the makers of AIM, or anyone else who wants to answer me. when i IM someone, there SN is blue and my SN is red, but the colin (the two little dots after the SN) are blue for me and red for them, thats all fine and dandy. but when i put up an away message the colin for in my away message box is gray-ish. whats up with that? [[the food calls for me]] sporko  '),(5900,3413707,'05,June,2004','2004-06-05 00:00:00','   I don\'t know whether or not you need something like this, but if you need a free image hosting site  with hotlinking  (aka it gives you something like \"http://www.place.com/pic.jpg\"), you\'d like  urlLink ApplePics .  Just a thought.  '),(5901,3413707,'05,June,2004','2004-06-05 00:00:00','   i am not realy happy, summer is depressing but whatever i am gonna pretend to be happy, mabbie i can pursuade my self. right well i realy cant think of anything postive... nothings happened, except that in the next week or so i am gonna be dyign my hair black and red, black and purple or black and green i am picking the color on what you vote. if you care to vote on the color check out blog-o-rama. the ted practice was cool, after aboitu 4 hours of practicing and recording we watched spinal tap, when that was over we watched monty python and the holy grail, then i left. we had fun. now i have to go do laundry for i am the only one in the house who knows that colors and whites dont get washed together. [[instert closing here]] sporko  '),(5902,3413707,'05,June,2004','2004-06-05 00:00:00','   This is an awesome story, it\'s the weirdest accident ever... A construction worker in Lancaster, Calif., is   expected to make a full recovery after accidentally being shot in the   head with a nail gun -- six times. Three of the nails penetrated his   brain; one entered his spinal column, but luckily not his spinal cord.   Witnesses say Isidro Mejia, 39, fell from the roof of the house he was   building onto another worker who was using the gun below. \"The guy   holding the nail gun was grabbing to keep from falling, and just   gripped tighter on the trigger,\" said a Los Angeles County sheriff\'s   investigator. \"It\'s a nail gun they use for framing. It can drive nails   through 2-by-4s.\" (Los Angeles Times) ...It\'s the first time L.A. cops   have said it: a Mexican guy really was framed.    urlLink pictures (don\'t worry, just x-rays)  '),(5903,3413707,'05,June,2004','2004-06-05 00:00:00','   Wow, this is really sad...All of my friends are away so instead I\'m spending my time over a celebirty obsession of Craig Nicholls, or as I like to call him CN, (the crazy lead guitarist and singer of The Vines) who looks like he\'s stoned...but..oh well...so does Green Poncho Boy... Anyway, for all of you Vines lovers out there (like myself). Here\'s the most amazing music video ever!!!!!!... Ride...(I\'m lazy so just cut and paste this in the address box yourself)... http://www.nme.com/features/107518.htm  If you want more of my pointless ramblings of CN or the lyric to Ride, go to my blog...you know where it is...  '),(5904,3413707,'05,June,2004','2004-06-05 00:00:00','   Ok, we officially need to get some better posts on here, Ransom posting about how bored he is...come on...hm...we need something good to post about...ok, everyone post a link to there favorite site and say what it is. Yeah, that might work, every so often someone just post something like favorite band or something, then everyone visit the site/listen/whatever the thingy then comment on it...  '),(5905,3413707,'05,June,2004','2004-06-05 00:00:00','   ya im gonna post on blog of the damned about my bordom...ok so if anyone has any ideas of what i can do than please post and answer. and if anyone can guess what i\'m doing while i pass the time until i get a post telling me what to do... i could be:  A: Eating B: watching TV (bonus points if you can guess what im watching) C: Sleeping in an odd place (like on a tv)  D: Helping Mrs. Van Doodles find her poodles  Z: learning my alphebet E: catching somthing (maybe myself) on fire F: still on the computer typeing more possible answers G: or not   post any and all answers to this \"riddle\" that winner gets...i dont know what yet, but somthing...ewwww a mystery....its...mysterious!?!  '),(5906,3413707,'04,June,2004','2004-06-04 00:00:00','   Calling all members of Ted! We should be getting together at my house for practice/recording at around 11 today until as long as you can stay. If you need directions, I suggest you look online for them, because I\'m no good at them. Anyway, I\'ll be on AIM if you need me for anything else.  '),(5907,3413707,'04,June,2004','2004-06-04 00:00:00','   im out of \"go ol\' Monty-G\" lol but ya its kinda cool and kinda sad, but i know this will be posted in blog of the damned like a million times but o well thats too bad. and i know that everyone feels at least a little sad bout\' leaving. (well eccept maybe kenis) i mean wats up with him not being here? i know it was a family thing but really! anywho on with my rant, im sad to see everyone go but wat can i do about it. i mean no one is sad to see me go...(please post and tell me im wrong...or not)....so any ideas on another way to stay in really close touch? i didnt think so. and i said it in my personal blog (ransomsblog.blogspot.com) that at the end-ish of the summer i will be having a pool party for the whole class...so keep ur schedules open kinda. there u go im done...for now ::insert evil laugh::  '),(5908,3413707,'03,June,2004','2004-06-03 00:00:00','   being grown up isn\'t half as fun as growing up. these are the best days of our lives.  '),(5909,3413707,'03,June,2004','2004-06-03 00:00:00','   ok my turn.....wow...its actually over. it\'s actually done. montgomery really did change me guys. it\'s like, my life. these past three years have been the best years of my life and i have a feeling they always will be. i have the best.....and i mean  the  best friends ever and i dont wana let that go. I\'ma miss every single one of you, yeah even you taco *starts crying* i love you... all of you. Thanks for making my life  '),(5910,3413707,'03,June,2004','2004-06-03 00:00:00','   sorry  '),(5911,3413707,'03,June,2004','2004-06-03 00:00:00','   ok my turn.....wow...its actually over. it\'s actually done. montgomery really did change me guys. it\'s like, my life. these past three years have been the best years of my life and i have a feeling they always will be. i have the best.....and i mean  the  best friends ever and i dont wana let that go. I\'ma miss every single one of you, yeah even you taco *starts crying* i love you... all of you. Thanks for making my life  '),(5912,3413707,'03,June,2004','2004-06-03 00:00:00','   Wow, guy\'s you no what, all the crap that everyone say\'s about, missing each other and how you\'ll always remeber. Yeah, well, it\'s time for me to say it, and for the first time I mean it. Only two years at Montgomery, but damn, best two years of my life. We have to keep this blog going, and I\'m gonna start having a bunch of parties with all the Montgomery people\'s. Oh, yeah and cheesy inpirsation chapel speech thingy, you guys remember the barn, thats our class, awesome together but even when we are broken up made into another barn, or furniture, we are still Montgomerian\'s at heart, still a part of the barn. God, I\'m gonna miss this school (::sniff, sniff::).  '),(5913,3413707,'03,June,2004','2004-06-03 00:00:00','   As my present to all of you, I\'m allowing each of you to ask me a question, any question about anything, and I will answer it. It\'s your choice how you want to ask me, you can email me, IM me, or even just ask me in the comments. Oh, and if you really want to know how I feel about graduating, you can just go to my  urlLink blog  and read the latest  urlLink post .  '),(5914,3413707,'03,June,2004','2004-06-03 00:00:00','   Ok, today was graduation, blah blah blah I love you guys with all my heart blah blah blah.  Guess what.  I have news.  I love you people so much, you will never truly be rid of me.  So, this summer and next year at school, don\'t just forget about all of your friends.  You haven\'t seen the last of Tori...  '),(5915,3413707,'03,June,2004','2004-06-03 00:00:00','   that was the best year i have ever had, and no i am not lying to you, i have never had a more meaningfull, fun, amazing, awesome, slinkster cool, time ever. and i know that a lot of you think that too, but i am just stating facts like i allways do. the best parts of the entire year: 1. the first day of school. 2. the first dance (when i met tucker). 3. the malvern dance (not the kiss). 4. the play. 5. the whole (yes whole) canada trip (moaches!) 6. graduation 7. getting hugs from everyone   8. knowing that i will never forget any of you or the stuff we went through, no matter how hard i try (ill only be trying to forget some things) 9. seeing every one together for the last time and loving everything about it all. thank you so much for all the wonderfull and not so wonderfull times we had, it was the most amazing year i have ever had and will stay the best danmmed year ever! -Rachel (AKA sporko)  '),(5916,3413707,'03,June,2004','2004-06-03 00:00:00','   Next time somebody gives you a brown receipt which they claim will defeat a satanic 400-year old cat who has possessed a Chinese businessman, you\'d better darn well take it!  Taken from  urlLink SomethingAwful.com .  '),(5917,3413707,'02,June,2004','2004-06-02 00:00:00','   The new blog is born!!! The address is... http://tacohut.blogspot.com/ so one of you smart folk add my blog to the side bar thingy.  '),(5918,3413707,'02,June,2004','2004-06-02 00:00:00','   So how was everyone\'s last day here at Montgomery?  I had fun.  I did stuff.  I had fun.  Again.  Oh, just just for the record, I\'m not wearing women\'s underwear (suckers).  Crap.  '),(5919,3413707,'01,June,2004','2004-06-01 00:00:00','   actualy that is a realy good CD, i also have it (and have had it for a while). the single \"take me out\" is realy good, and no tuck your not the only one who knows of them. blah, um... yeah, i still want a kit kat bar, i saw one in the superfresh, but i didnt have anymoney and my mum wouldnt let me buy it grr.  '),(5920,3413707,'01,June,2004','2004-06-01 00:00:00','   Your random history lesson of the day.  (Because it pertains to my 20th Century Project).  Franz, also known as Francis, Ferdinand was the Archduke of Austria-Hungary in 1914.  In early July, he and his wife were driving through the town of Sarajevo, Bosnia, when 19 year-old Gavrilo Princip stepped out of the crowd and opened fire.  This event triggered world war 1, one of the bloodiest military conflicts of all time.  '),(5921,3413707,'01,June,2004','2004-06-01 00:00:00','   for some reason the hyperlink on the last post died or something, so heres the address...http://www.franzferdinand.co.uk/  '),(5922,3413707,'01,June,2004','2004-06-01 00:00:00','   FRANZ FERINAND: sure they\'re named after a pyscho military leader, and sure they sing creepy songs about being gay with a guy named Micheal but they rock!! These guys are all like funky retro garage rocky and thats awesome. Some of you may have heard of there single \"take me out\" (::cricket:: maybe not...). anyway you guys have to buy the CD, which is self titled. Oh yeah, there also British, this is kinda crapy site becuase the left out all the music but heres there homepage...  urlLink  '),(5923,3413707,'28,July,2004','2004-07-28 00:00:00','   ok, rj has changed his campagin... and a few people cant make it. so, now...  if you can come to our house tommrow (thursday) from noon to 6 be there. experience dosent matter. &nbsp;call for directions if you need them (6106400605)  spork  '),(5924,3413707,'27,July,2004','2004-07-27 00:00:00','   Tomorrow.  2-8 PM.  My house.  Death Gate campaign.  Be there.  '),(5925,3413707,'26,July,2004','2004-07-26 00:00:00','   If you are coming to my D&D tomorrow (Death Gate campaign), or if you would like to, please go to  urlLink my blog  and vote.  Times would be 2-8 PM.  If you are not on the current list (Doug, Gus, Ben, Gabo), please comment on this post and leave your name and say that you would like to come.  AND VOTE!!!  '),(5926,3413707,'26,July,2004','2004-07-26 00:00:00','   RJ and i have managed to set up our own DnD session, its a new epic level (20 - 50) its called the black pantheon. so far... whos coming: RJ, me, evan, eric tait, and alex&nbsp;d.&nbsp;it ought to be fun.&nbsp; so we are  looking for one more player... must be experienced , may not be a first time player. also... RJ knows the rules for&nbsp;playing gods ::wink wink:: &nbsp;everyone is starting at around level 20. its on thursday at the paper bag (RJ\'s and my house) from noon to six. call if you need directions (6106400605).&nbsp;roll a d20 to see if you can come.  SPORK and RJ  '),(5927,3413707,'26,July,2004','2004-07-26 00:00:00','   How does  Thursday  2-8 sound?  '),(5928,3413707,'24,July,2004','2004-07-24 00:00:00','   D&D at my house Wednesday 2-8. Continuation of the Death Gate campaign.  If you were there last time, feel free to come again.  If not, please ask permission.  Not a comedy campaign (Ben).  BE THERE!  '),(5929,3413707,'24,July,2004','2004-07-24 00:00:00','   ok i messed up the last band prediction post... bumblebeez 81 are not a garage rock band, i was listening to the wrong cd at the time... its actaully the new hives, lol, so yea u have to get that cd if that idea sounds good... bumblebeez 81 is like gorillaz but much dirtyier, lots of distorion guitar and crazy bass beats and semi-rap, semi-rock lyrics... u really have to here them to understand it... so try to pick up the cd if u see it.  '),(5930,3413707,'23,July,2004','2004-07-23 00:00:00','   three awesome bands for you today... Bumblebeez 81- these guys r pretty garage rocky but all there songs r mixed by one of there own who is a pro dj so really cool sounding. Very slickly combining hives and white stripes...these guys have a video that im sure is gonna be playing on mtv very soon so look out for them.   Whirlwind Heat- i gurantee you have never and will never hear of these guys from anyone but me... they are very odd, have thier own flare, a personal favorite of mine...  The Dirtbombs- love LOVE these guys, very cool have two of every intersting, but are still pretty smooth and jazzy, kinda up beat too but still rocking  i have a few other bands that you should check out but arent as good and have no chance at stardom.... -The Soledad Brothers (blues...no one here would like them) -Blanche (smooth as silk)  OK, im done.. goodbye until next time...  '),(5931,3413707,'23,July,2004','2004-07-23 00:00:00','   well this is my not so formal good bye for all you. oh, this is your last chance to see my hair in its slighty pink stage... in about 4 days i am leaving for NY state and i am gonna be dying my hair black and&nbsp;red (or blond) &nbsp;again&nbsp;(because i look good that way). so if you feel the need to talk, see, or hang out with me for the next month or till after school starts, best to do it now. and, oh wait, for all you ex-montGs, i problay wont be seeing you for a long long time ::evil laugh:: ... gonna be busy as hell. so uh... fare thee well. spork.  '),(5932,3413707,'22,July,2004','2004-07-22 00:00:00','   ok the party is a no go, no one commented and i cant do it becasue im not sure anyone will even sdhow. All you had to do was comment and you could come...o except for sporko, who can always bbe counted on for a vague piontless, un-defineing comment. Ok so becasue of you there is no party and if you show up you will be luaghed at, attacked by dogs and locked in a room with keegan. ok so next time you want to come to a party of mine, just follow the frickin\' directions!!  '),(5933,3413707,'20,July,2004','2004-07-20 00:00:00','   Ok poeple my casa is open for a party on friday!! Anyone who wants to come just comment... and you can bring friends, everyone who reads this must tell someone about this, and mention their name in the comment. Please get as many poeple as you can. If you need info call my house or email me at Dont_tell_the_cops@hotmail.com. Oh, yeah details... um.... from like 5 to whenever.... and not sure about food yet but we will PROBABLY feed you. I think thats it but please send this to everyone who you know and hopefully knows me!! ok, this party will only be good if a lot of people show.  '),(5934,3413707,'17,July,2004','2004-07-17 00:00:00','   IIIIIIIIIIIIIII\'M BAAAAAAAAAAAAAAAAAAAAAAAAAACK!!!  '),(5935,3413707,'15,July,2004','2004-07-15 00:00:00','   Don\'t click  urlLink this !  '),(5936,3413707,'14,July,2004','2004-07-14 00:00:00','   dougie you must call me at my dads as soon as possible! HURRY ( in no immediate danger, but for something that we must do)  '),(5937,3413707,'13,July,2004','2004-07-13 00:00:00','   look at the title  '),(5938,3413707,'12,July,2004','2004-07-12 00:00:00','   hello&nbsp; urlLink  '),(5939,3413707,'12,July,2004','2004-07-12 00:00:00','   See the new  urlLink Teen Girl Squad!  '),(5940,3413707,'12,July,2004','2004-07-12 00:00:00','   I really havent posted anything yet and i just wanted to... anyone who has an idea about what i should do with my broken mp3 player (this is the 3rd one of that kinda that has broken through no fault of my own). ya post on here or on my own blog... ransomsblog.blogspot.com. thanks  '),(5941,3413707,'10,July,2004','2004-07-10 00:00:00','   Is it just me or have Ransom and Emily not been  posting the last few weeks. What happened to them?  '),(5942,3413707,'07,July,2004','2004-07-07 00:00:00','   Presenting the brand-new  urlLink danmmed shirt ! Get yours today!  '),(5943,3413707,'06,July,2004','2004-07-06 00:00:00','   Is it just me, or are people neglecting good ol\' BotD, their faithful blog, for the new RhythmSection blog?  Not that there\'s anything wrong with that, but I do sorta have a preference (being a founder and all).  Keep BotD alive!  Post stuff!  '),(5944,3413707,'03,July,2004','2004-07-03 00:00:00','   Back in the age of pirates, there was one ship that ruled the seas.  It was a typical pirate ship, with a typical captain.  The only strange thing was that the captain always wore a red shirt.  One day, one of the crew members went up to the captain and asked, \"Why do you always wear that red shirt?\"  \"Oh that?\" asked the captain.  \"I wear that so that if I ever get shot, you will never see the blood, so that you can keep in fighting the same.\"  And so the guy walked away thinking  OK, that\'s pretty cool.  We\'ve got a pretty great guy for our captain .  Later that day, an enemy ship started approaching.  The pirates got ready for battle, and started sailing towards the enemies.  They shot a few warning shots, which were responded with other shots.  Soon, both sides were almost ready to fight when the captain yelled back to the crew, \"Quick!  Go get my brown pants!\"  '),(5945,3413707,'05,August,2004','2004-08-05 00:00:00','   Dance like hypnotized chickens, my young uns!!!!!!!! Dance...dance...DANCE!!!!!!!!!!!!!!!!!!!!!!!!!!!!  '),(5946,3413707,'04,August,2004','2004-08-04 00:00:00','   ::wave:: good bye for now, ill see you all in a few weeks, for i am leaving for NY in about an hour. (feel the need to make me feel loved? special? not so alone? email me ( urlLink sporko@gmail.com ) or call my mums cell, but i have been forbiden to tell you what that is online, so you gonna have to ask someone at my house, hence you gotta call the house first. so much work, but it pays off, you get to talk to yours truly) for now, the one, the only, spork.  '),(5947,3413707,'03,August,2004','2004-08-03 00:00:00','   ok sop since everyone is doing it i took the daunte inferno tst thingy too...and i did better than Sporko or Flabo....i got limbo hahahaha and i actually took it honestly! suprise! ya so im too lazy and or stupid to post is on the blog but o well i think u can live with the heartbreak.  '),(5948,3413707,'02,August,2004','2004-08-02 00:00:00','   ok, i took the test twice, the first time i got in the 9th level (the worst) the second time i didnt do so bad.... only 5th level.  The Dante\'s Inferno Test has banished you to  the Fifth Level of Hell!   Here is how you matched up against all the levels:     Level    Score       urlLink Purgatory   (Repenting Believers)   Very Low       urlLink Level 1 - Limbo   (Virtuous Non-Believers)   Very Low       urlLink Level 2   (Lustful)   High       urlLink Level 3   (Gluttonous)   Low       urlLink Level 4   (Prodigal and Avaricious)   Low       urlLink Level 5   (Wrathful and Gloomy)   Extreme       urlLink Level 6 - The City of Dis   (Heretics)   Moderate       urlLink Level 7   (Violent)   Very High       urlLink Level 8- the Malebolge   (Fraudulent, Malicious, Panderers)   Extreme       urlLink Level 9 - Cocytus   (Treacherous)   Very High      Take the  urlLink Dante\'s\" Inferno Test  '),(5949,3413707,'02,August,2004','2004-08-02 00:00:00','   I was sent the following email and rather than forwarding it I thought I\'d just post it on the blog -- obviously this only has meaning for some members, so if not just ignore it, sorry to waste your time   --as for everyone else, I know you might have already seen this, but whatever.  PLEASE FORWARD TO OTHER FRIENDS & FAMILY THANKS!!! OCEAN CITY REMEMBERS EMILY 07-08-1990 â€“ 12/11/02 WHEN:   THURSDAY, AUGUST 5, 2004(FAMILY DAY IN OCEAN CITY) WHERE:  GILLIANâ€™S WONDERLAND PIER / OCEAN CITYâ€™S BOARDWALK AT                               6th STREET TIME:   1:00 â€“ 5:00 P.M.    4 HOURS OF FUN! ENJOY ALL AMUSEMENT RIDES FOR ONLY $10.00 BENEFITS:       THE EMILY KOENIG MEMORIAL FUND (Managed by the Philadelphia Foundation â€“ supports vaccine awareness & research and scholarships for children) HELP WIPE OUT MENINGITIS. . . SAVE THE LIFE OF A CHILD CONTACT INFORMATION:     urlLink www.littletreasurebooks.com  /  urlLink www.gillians.com          Email:  Paula Lizzi at  urlLink paula@littletreasurebooks.com                  John Kavchok at  urlLink john@gillians.com  '),(5950,3413707,'02,August,2004','2004-08-02 00:00:00','   Well...here I am....posting something...hmmmm...oh what the heck!  The Dante\'s Inferno Test has banished you to  the Second Level of Hell!   Here is how you matched up against all the levels:     Level    Score       urlLink Purgatory   (Repenting Believers)   Very Low       urlLink Level 1 - Limbo   (Virtuous Non-Believers)   Moderate       urlLink Level 2   (Lustful)   High       urlLink Level 3   (Gluttonous)   Moderate       urlLink Level 4   (Prodigal and Avaricious)   Very Low       urlLink Level 5   (Wrathful and Gloomy)   Moderate       urlLink Level 6 - The City of Dis   (Heretics)   Moderate       urlLink Level 7   (Violent)   Moderate       urlLink Level 8- the Malebolge   (Fraudulent, Malicious, Panderers)   High       urlLink Level 9 - Cocytus   (Treacherous)   Moderate      Take the  urlLink Dante\'s Inferno Test  '),(5951,2813430,'14,May,2004','2004-05-14 00:00:00','   Sitting in solitude, thoughts in plentitude Pondered,over the day, she had a role to play That crossed another\'s way,many in the fray Sailed beyond,she may,they too had a say The line was very fine,started with a dine Invisible though it may seem, they could feel It was an illusion, she explained, Her words, they failed to entertain She brooded over the language, Passed to her down the ages To communicate the right feeling, She dug deep and went down reeling Blinded by the raging conflict, She was afraid she would still inflict More wounds to bring down a friendship All I could do was sit and listen Trying to understand the compassion And smile straight at a solution  '),(5952,2813430,'14,May,2004','2004-05-14 00:00:00','   We wish to see in reality, things in plenty But lose count of the perennial flow, after twenty Day and nights race by swift Still the counts come by drift Somewhere down they become dormant or die Only to stand up to remind our prayers are a lie Still we hold strong and resolve To see them happen and not dissolve Human beings are born to pray and wait Still end up becoming just a bait Sacrifice is supreme, cried the master We console the death of the counts, that seem faster At the back of my mind, it rings loud That I am trying to reach the cloud Still I resolve, to pray with hope and desire As I believe God will never conspire After all, we all have the right to dream As we all unconsciously flow down the stream.  '),(5953,2813430,'14,May,2004','2004-05-14 00:00:00','   Our arrivals were miles apart And destinations, a big question mark But now, together we travel And so many things together we unravel You, as a person I know a lot Well enough to read your mind and thoughts But still puzzled about your personality As it stretches me beyond certain rationality Sometimes soft and emotional as a child Yet many a time, bold decisions coming from your mind I see a true and trusty human being, unmasked and not far In you Dear and I like you for all that you are. Frankly from my heart I admit You are someone special to me, my friend.  '),(5954,2813430,'14,May,2004','2004-05-14 00:00:00','   Tides are always high We all face them with a sigh Still make an effort wide To feel the success lap our side Champion, is the one who fights However dark be the nights Difficulties everyone will feel But experience is only around your heel I admire your brave stride Against every upcoming tide My heart skips a beat, goes blank \'Cos I see you without even a plank Also wrestling with a dangerous shark My tears are not seen as it is already dark My dear friend, I pray for you For courage and vision too Don\'t give up, the world is with you. And as a neighbor I am too.  '),(5955,2813430,'14,May,2004','2004-05-14 00:00:00','   In solitude stood a rooted trunk In highest majesty spearding leaved branches Into an appreciable diameter Its pride was clouded with a strange helpless feeling of loneliness Travelers passed by looked with raised brows Awestruck they were dragged to spend A few hours in its relaxing cool shade Every visitor would make sure to remark He wished if this could be shiftedto his home Nomads too joined the list They all got substitutes None thought about making a house beneath And make \"his majesty\" owned. It was killing The dried trunk was saved and twigs were also taken home They made good firewood and furniture Everyone used it alive and wanted it dead What remained with it always was - Solitude  '),(5956,2813430,'14,May,2004','2004-05-14 00:00:00','   O\' Lady! Reveal not your bosom  And prove, you are beautiful to some Let it talk for itself I will judge myself. A flower naturally says its fragrance And nothing causes it a hindrance Let me not see bright red marks  On your forehead, like hungry sharks To prove your matrimony And disrupt my harmony Let your eyes speak that language And give me the freedom to gauge Have I caged you, with your freedom? And makes you writhe in boredom From your mind, speak out But carefully murmur not By time I am getting older Does that make you bolder? Allow me to feel Still I repeat, do not reveal.  '),(5957,2813430,'14,May,2004','2004-05-14 00:00:00','   Gazing through the sparkling glass window The little boy\'s eyes burrowed He had to be dragged and hopped His eyes were filled with eagerness And his heart with desire They swelled with each passing minute Till finally he got one toy defenite The longer the wait, the more the emotion  So thought the creator Who never went against nature He too made mistakes grave That showed due to a man who was brave For he never knew That they died a natural death too With no desire and eagerness in store He still laughed and made merry And never ever felt sorry Still it was in him mild, he wanted a child Is the creator still churning? Or is the fire in him still burning?  '),(5958,2813430,'14,May,2004','2004-05-14 00:00:00','   He knocked twice No response He turned to leave A wind blew and the door opened a little He peeped inside Sound of water gushing He stepped in Silently sat with a magazine Good cover photo  The gushing stopped Irresistible He got up and peeped in though the gap She was drying her hair A short towel around her He felt his blood gushing He resumed his seat  He is seated behind her  In the classroom He could smell the shampoo His blood gushed He teased her She smiled back He had liked her He could not muster enough courage To tell her â€œI love youâ€  He met a friend And got her address His wait ended when  He got out of the cab At her house And knocked her door Great expectations He forget his family Temporarily  She came out Saw him and stood stunned Forgot that she was still in a towel Moved slowly towards him He stood up Gaping at her â€œHow are you?â€ â€œFineâ€ â€œAnd youâ€ Yeah. Fineâ€  He stretched his hands She dropped into them They remained locked The room was filled with The pungent aroma of Sweat and the same Shampoo His and hers They walked into the inner room  He came out All over His expectations fulfilled He kissed her one last time And bid goodbye She stood there Smiling at him She did not feel any difference It was just another customer But with a difference She did not charge him Because she thought  She owed him He never knew  Her normal day resumed His new day dawned Both their own ways Back to their life Long journey For a moment of ecstasy  And normalcy  That is life. We see only our paths Not otherâ€™s It goes on and on and onâ€¦â€¦  '),(5959,2813430,'14,May,2004','2004-05-14 00:00:00','   Compelled to retreat his path by a well wisher he traced the long forgotten starange way and recognised many familiar objects When flashes crossed his film most strangers became just crossed familiars he dug deep into his storage cells to strike similarities his retreat turned too novel and truth seemed yonder came along one introduction that seemed known till recently but took a unknown turn to a scarry path he feared his long healed wounds had opened he did not bleed becasue he did not have blood or life a strange look fixed on his face his wellwisher saw  not his friend but his ghost the ghost still lives happily and with love  '),(5960,2813430,'14,May,2004','2004-05-14 00:00:00','   I believed I was full For so many years. This â€œfullnessâ€ was accompanied by others Independence and self sufficiency I survived all along â€“ starving. You poured and filled the rest Into a space filled with vacuum Which I never knew existed. I looked back and noticed the drains Which you carefully plugged. As you never tagged any, I contend shamefully I never had a reason to say I never noticed. I take a long silent pride in admitting in whispers Which I believe will reach only your ears That I always subconsciously Leaned on your strong shoulders. Now sitting in a corner Of this lonely strange planet With a few inert objects as companions I think aloud to myself Because perfect communication Is a visible obstacle here. You graduate to become my best half. I pray you hear this too My dear.  '),(5961,2813430,'14,May,2004','2004-05-14 00:00:00','   Like a dried trunk, vast in diameter he stood Full of life, a guardian with a majestic hood Strength, was his strength, they all under stood Moral values and material, he gave them ample food His sharp eyes spoke of one thing, always Righteousness and perfection in every phase They played with glee but well mannered on this firm base  He made sure they never slipped and fell on their face Nothing can cause me a bother. As I am proud to call him my father.  Leaves of green and purple, all in color The best they have seen and felt in demeanor  No compromise whatsoever in grandeur  Can feel the radiance of love in every manner Cool shade ensured with ample breeze around  When it comes to our share it is always more by a pound The call from the heart is never less in sound And leaves a passerby standing and gaping astound Nothing can cause me a bother As I am proud to call her my mother.  A perfect companion since arrival Always teaching the threads of survival With a list of things to be rushed for a burial Strong beside to feel the sand and gravel An eagles eye against possible danger Lacking, in appearance, the tough look of a ranger Full of melody and softness, seen in a singer Taught, never give up with out being a winner Nothing can cause me a bother. As I am proud to call him my brother.  Dedication and confidence shining in the eyes Never knew the term, normally called lies In life, proved a graduate more than twice An association resulting in the term wise Changing the word adamant to sound positive and good Unchallenged in the decision and path, always stood Unflinching and straight in every mood Taught everyone how important it is not to brood Nothing can cause me a bother As I am proud to call her my wife.  For the show, already begun, a late entrant A mind that seemed always transparent Seeing a step ahead was a basic requirement Measuring well with the best temperament Life is never a knot, a complicated ball Always to ensure, they do not fall In calculation though slightly small A fire wall to protect them all Nothing can cause me a bother As by law I am proud to call her my sister.  The angel arrived, long awaited. Cute, cuddly and tender. I call him - my life.  '),(5962,2813430,'14,May,2004','2004-05-14 00:00:00','   I switched off the fan Not because The air with in the four walls was chill or The sound was snatching away traces of my sleep or To save the mounting electricity bills or Due to the fear that the hook holding the fan might snap and bring down the fan on my knees But because I felt the fan was tired of running  Between this and that Views differ, yet both seem so real. Everyone is right when you see through their eyes.  '),(5963,2813430,'14,May,2004','2004-05-14 00:00:00','   Arrest him! The voice loud Failed, he, to make his parents proud Lock him in â€“ the order trailed Clenched fists, he looked frail.  Snatched a chain, for a wile Abuses were heaped in pile Careless of his fathers name Closed his eyes in shame  Punched and beaten on his face Darkness alone, a saving grace Pounded to reveal his identity Silent and mute, his head hung in gravity  All the townsfolk were summoned Identify, your son who sinned Everyone hurried to the cell Anticipation in the air was felt  An elderly single father in the crowd Walked earnestly without a frown Prepared to face the worst He knew it would be his son dearest  Rehearsed several times in his mind Reaction to the truth when he finds Prayed for miracle and courage All alone caught in a barrage  The moment arrived - reality stared Trust, faith and belief lay bared In the darkness for his son he searched To HIM, his prayers never reached  No one can prepare for a reality It comes in all its brutality Escape is non-existent Rest is consistent.  '),(5964,2813430,'14,May,2004','2004-05-14 00:00:00','   Blessed  Shattered and scattered, in ruins not The spirit was always hot Communication ignored and lost The binding was sound and fast  He traveled from far and long Merged as one, like a song Thrilled to see a familiar face Instantly caught in a wild embrace  A kiss of happiness on each cheek Something he will always seek Tears welled in his heart Which always tore him apart  Crouched and crumpled in his large and lonely home He lay counting on the days to come For an instant he closed his eyes All his loved ones rolled like a dice  With deep anguish he missed them all Their strong hands could hold him from a fall He thanked god for the care and blessings He loved the best human beings during the passing  For his worries he saw they shed tears That erased all his worst fears Even time cannot move us apart I love you all from my heart  '),(5965,2813430,'14,May,2004','2004-05-14 00:00:00','   He ran around a circle, imaginary Thoughts pushing him beyond the ordinary Concentrating on a problem, so seemed Till time gave a solution, much in need Stopped then, he all of a sudden Chance to put down his burden Were his efforts a waste, he later thought Could these solutions be bought  All around him everyone raced Each of them getting outpaced Their destinations do they all reach Or end up as a topic to preach The solutions seem dramatic When not looked pragmatic Once achieved they look inane As simple as a window pane  His thought was his problem He carried on as his emblem They said thoughts are god given Every man will one day be forgiven Then this running, a question mark Reality stared at him stark Look back for a while, with feign God is once again playing his game.  Who wins?  '),(5966,2813430,'14,May,2004','2004-05-14 00:00:00','   A lasting impression, deep within She creates, with out a whim A smile stitched on her face Full of life and overflowing grace  Compelled to compare with a kite Flying high on a full moon night Gliding and swaying all her way In her territory having a total say  The softness seen externally Wraps the toughness permanently As the night closes tight She is still seen everywhere in sight  Undiminished vigor and always active I term this boss â€œvery effectiveâ€ Circumstances give birth to friends few I am glad to see a friend in you.  '),(5967,2813430,'14,May,2004','2004-05-14 00:00:00','   In a stroke, numbers change Creatures from deep slumbers wake Of endless activity, boredom and strife To an expectation filled life Another routine phenomenon Consciously blown out of proportion  Resolutions galore everywhere Trying to tie down things gone hay ware Success is never measured But resolutions are always treasured To be repeated the subsequent year Still failures never bringing a tear  Health and hygiene to fitness Resolve to also increase the smartness Education, careers and family All fizzle out in time tamely Have you thought about your society? One that has helped you in your anxieties  Go out and help others around And you too will be crowned  Strangers, colleagues and friends All wanting you for some errands Extend to them your hands and a smile You will build towers of happiness, tile by tile  Have kindness and compassion in your eyes And a spirit to accommodate and be nice Resolve to help everyone around you I am sure the world will be a better place for you.  '),(5968,2813430,'14,May,2004','2004-05-14 00:00:00','   Do you know,  you are being used as a plank of the staircase for people to go up. They depend on you very much  when they are below you. Once they step on you and go up you are forgotten.  But you feel you too have gone higher, to help them even more. Once again they use you,to climb.  Do not feel dejected for being used and  do not break down, for they will need you to climb down too.  Help. That\'s why you are there.  '),(5969,3883177,'31,July,2004','2004-07-31 00:00:00','    Can Seinfeld sue me for using that as a title?  Well, the picture thing is a bust.  I\'ve been helping my wife try to link to Yahoo picture albums at her blog  Poo-tee-wheet  and having no luck at all - seems that Yahoo tries to force you to include a session id or something at the end of the JPG extension in order to force you to go through their site to see the pictures - darn annoying - but I guess it probably saves them the trouble of several billion hits to linked pictures.  So, I reaquainted myself with FTP - for some reason I thought I needed to use CoffeeCup FTP or something like it, but then remembered that I could just the ftp. prefix through IE and treat the site like another directory.  Works great if we do it that way through our web space.  Just slow as heck on a dial-up connection.   '),(5970,3883177,'28,July,2004','2004-07-28 00:00:00','    Probably not the DNC per se, but at least a nice Democrat who was wearing his DNC button. Where ever he is, I\'d like to apologize for not handing him a soda and being a bit friendlier - over the years I\'ve developed a paranoia about anyone who comes to my door as they are invariably either:   1.) religious  2.) selling magazines  3.) soliciting donations under the guise of a charity, but really something else (even if it\'s a PAC with which I agree).   top it off with trying to wash Eryn\'s high chair tray and conduct interroom communication with my wife, and I was a little out of it. He noted that it was really nice to see someone with a John Kerry sign in their yard (I shall get a link out here, I have a picture of said sign) and that he hadn\'t been to a lot of friendly doors. I at first assumed he meant in my neighborhood, but most of my neighbors seem to be immigrants (Asian and Hispanic), duplex dwellers (of which I was once one) and at least one house w/a lesbian couple, people I sort of assume are Democrats (or at least not Bushies). Of course, maybe he started on the far end of my side of the street, that seems to be the white, young crowd in the neighborhood. Or maybe he was referring to Eagan in particular, a far more Republican locale than my previous neighborhood in Richfield, and home of our illustrious Governor, Pawlenty (he\'s a local soccer dad).      While I don\'t expect this to in any way reach the door-to-door Democrat I met yesterday, I\'d like to extend a virtual apology nonetheless - I owe you a beer/soda and some friendly conversation to make up for my distractedness (guess I can send an email over to Democrat central in St. Paul and see if they know who was over here yesterday - we\'ll see how that goes, I\'m a notorious procrastinator, non-follow-up-er, generally the reason I don\'t volunteer myself, although I can beg off at the moment because I have a kid).   '),(5971,3883177,'27,July,2004','2004-07-27 00:00:00','    Well, the email test, sending postings to the blog w/o actually having to log on and use the website, seemed to work fairly well, but I noticed you have to go on and republish the site at some point, otherwise the postings just queue up and wait for you w/o actually showing up on the blog.  So I can blog all day and then suddenly publish half a dozen posts - I\'ve noticed a number of blogs I like seem to operate on that schedule, so I wonder if that\'s what they\'re up to as well.   '),(5972,3883177,'27,July,2004','2004-07-27 00:00:00','    The most amusing piece of email that crossed my desk today was an ongoing conversation with two friends about G.W.\'s second mountain bike tumble.   AP Mountain Biking Story   First of all - who bikes 18 miles in an hour and twenty minutes on a challenging trail?  I\'m not saying it\'s not possible to maintain that speed on a mountain bike, I\'m just saying it doesn\'t seem likely if you\'re really off roading it. I did a spin around Lebannon Hills the other day, about 8.5 miles, including the 4 miles involved in getting there and back, and it took about eighty minutes - when you get to the rocks, the trees near your elbows and the stumps you simply bounce off of, you\'re lucky to go faster than a crawl - and, I\'d like to note, lying on the ground recovering from a fall counts against your average speed.  Second, 1200 calories?  One hour on a bike trainer for me, dripping sweat until I\'m soaking my practice carpet, at two hundred and fifty pounds, burns under a thousand calories. A one hundred and sixty pound individual burns significantly less, about 340 an hour (depends on the speed) according to calorie counters.  Does he have to mark up everything?  I figured the reasoning was, G.W. is carrying the weight of the nation on his shoulders, so if there are 290 million people in the U.S. at one hundred and sixty pounds,each burning 340 calories, that comes out in the 94 billion calorie range (about what we\'re spending on Iraq) - he can use any portion of those.  He chooses to use the calories expenditure of three people losing money because of his tax change/medicare policies, etc. He had 93.9-odd billion calories to spare he could have claimed and somewhere three poor people didn\'t lose weight.  The article itself is hilarious if you read it carefully - $3,000 mountain bike, toe clips (not clipless), the reporter watching G.W. lying there instead of helping him, the medics who must either tail him in a car or also expend 1,200 calories each at 18 MPH, the implication that John Kerry is some sort of dilletante for riding an expensive road bike and that $3,000 mountain bikes are you run of the mill blue collar bikes - I\'m reading a bit into it there, but I don\'t think they\'d deny they were implying it.   He pants hard, emitting low \"hrrr, hrrr, hrrr\" grunts with each stroke of the pedals, his shoulders bobbing up and down.   Yep...good stuff.   '),(5973,3883177,'26,July,2004','2004-07-26 00:00:00','    Did I mention I have to use IE5 because the hard drive on this laptop is nice and small?  So if you see spelling errors, blame Blogger - if I can afford a rich editor for IE5+ and NN for my web project using my personal credit card, then they can certainly afford one for here.  Anyway, didn\'t I promise an autobiography (or autobibliography as I so nicely put it before I edited it)?  I noticed that I can fill in a rather extensive biography in the tool that comes with blogger, so I\'ll do that instead.  Sort of pseudo entry for this evening.  Feel free to peruse my favorite books, movies, etc.  I\'ll try to find a 55K picture of myself on the web tomorrow that can serve as an avatar.   '),(5974,3883177,'23,July,2004','2004-07-23 00:00:00','    Thought I should run a test to see if I can get pictures into my site if they\'re uploaded to my home page.  Here it goes.   '),(5975,3883177,'23,July,2004','2004-07-23 00:00:00','    So I have a nice pentium with a decent modem on it downstairs, and instead I\'m relegated to trying to create posts with this nasty old laptap via a 19.2 connection while sitting at my second hand kitchen table (which used to be the conference table for the Third District Nurses Association, sort of the local union, but not quite).  For a while I thought maybe the problem was my phone lines, but this old piece of  junk never loses its connection and the one down stairs does nothing  but  lose its connection about every three minutes.  I guess I should have known, if you can maintain a telephone conversation, you should be able to maintain a data connection.  That means the only thing it can be is the computer itself and that\'s distressing - no viruses, so there\'s just something intrinsically wrong with it - I thought for a while it might be the graphics card, but heck if I know why a bad graphics card might create an issue with my modem connection.  I know it\'s not the hard drive (ghosted and replaced) or the memory (which all seems to be registering as there, active and healthy), so it has to be something conflicting with something else and redoing all the ports and IRQs that the hardware was associated with didn\'t do anything last time and the system monitor doesn\'t show anything in particular sucking up all the CPU, so I\'m starting to think it\'s just schizophrenic and I should worry about finding a new machine for $450 and use that one as some sort of fileserver for photos. I wouldn\'t feel so bad about it if it weren\'t the fact that the thing has been almost unusable from a gaming perspective and video perspective (for the MPEGs of Eryn) for the last year and half - pretty much since we moved it from the duplex in Richfield.  I could simply clean it completely and start from scratch, but that sounds like a headache beyond imagining.   '),(5976,3883177,'23,July,2004','2004-07-23 00:00:00','    I thought perhaps it was time to officially kick off my blog.&nbsp; Like most bloggers, I don\'t really have a clear idea what it is I want to talk about as a theme, but I\'ll work on it in the first few posts and see if I can\'t come up with something.&nbsp; I could focus on Westlaw books in movies and on television, but that deserves more of a website sort of treatment. So, what kinds of things are going on - what do I want to talk about?&nbsp; Let\'s start by noting that this last weekend I had the twenty year reunion of my high school bicycling group.&nbsp; Like most teenagers, that means that the bicycling took place before we were all old enough to drive - pre-motorized vehicle, mostly pre-opposite sex - ages fourteen and fifteen (two trips) to be precise.&nbsp; One trip to northern Wisconsin (Copper Falls) and one to northern Minnesota (Grandma\'s Marathon).&nbsp; I can\'t say I hadn\'t seen those friends since I was 15, but I certainly hadn\'t seen a number of them since graduation.&nbsp; We (Ben, Mark, Jenny, Jason [Jenny\'s boyfriend from Canada, not an official reunion member], Dwain, Dwain\'s son and myself all biked from Monticello to Rockford (Lake Independence) to commemorate our training ride.&nbsp; Ben was thoughtful enough to recreate the puking incident of one of our original members (at only three miles into the ride), and we all had a pretty good time reminiscing about Jenny\'s wheel rolling off, her third degree sunburn, streaking, baby powder fights and other fun.&nbsp; I\'ll add a few links to the pictures from that weekend when I get them uploaded to Yahoo.    urlLink http://f1.pg.photos.yahoo.com/ph/s_mcvay/slideshow?&.dir=/e7f0&amp;.src=ph   &nbsp;  Tomorrow: An autobiography that I didn\'t feel like making my first post.   '),(5977,3883177,'12,August,2004','2004-08-12 00:00:00','    Hmm...maybe I  should have only included those since the beginning of the year - it\'s starting  to be quite the list.  &nbsp;      &nbsp;         Scotts Books               1       2/25/1994       Mote in God\'s Eye, The       Niven, Larry and Jerry Pournelle       1974       Simon and Schuster       475       8.00       SF&amp;F                   2       3/1/1994       Forlorn Hope, The       Drake, David       1984       TOR       318       1.00       SF&amp;F                   3       3/5/1994       Talismans of Shannara, The       Brooks, Terry       1993       Del Rey       440       3.00       SF&amp;F                   4       3/9/1994       Gripping Hand, The       Niven, Larry and Jerry Pournelle       1993       Pocket Books       413       7.00       SF&amp;F                   5       3/16/1994       Shudder Again       Ed. by Michelle Slung (various authors)       1993       ROC       357       3.00       HORR                   6       3/20/1994       Elementals, The       Llywelyn, Morgan       1993       TOR       369       6.00       SF&amp;F                   7       5/6/1994       Escape From Management Hell       Gilbreath, Robert D.       1993       Berrett Koehler       161       7.00       BUS                   8       5/21/1994       Breakfast of Champions       Vonnegut, Kurt       1973       Dell Books       297       7.00       FICT                   9       5/30/1994       Hobbit, The       Tolkein, J.R.R.       1966       Houghton Mifflin Co.       317       8.00       SF&amp;F                   10       6/4/1994       Nightmares and Dreamscapes       King, Stephen       1993       Viking       816       3.00       HORR                   11       6/12/1994       Against a Dark Background       Banks, Iain M.       1993       Bantam Spectra       515       8.00       SF&amp;F                   12       6/16/1994       Use of Weapons       Banks, Iain M.       1990       Bantam Spectra       389       9.00       SF&amp;F                   13       6/19/1994       Ender\'s Game       Card, Orson Scott       1977       TOR       324       8.00       SF&amp;F                   14       6/24/1994       Speaker for the Dead       Card, Orson Scott       1986       TOR       415       7.00       SF&amp;F                   15       7/3/1994       Armor       Steakley, John       1984       DAW       426       10.00       SF&amp;F                   16       7/5/1994       Magician: Apprentice       Feist, Raymond E.       1982       Bantam Spectra       323       8.00       SF&amp;F                   17       7/7/1994       Magician: Master       Feist, Raymond E.       1982       Bantam Spectra       366       8.00       SF&amp;F                   18       7/9/1994       Silverthorn       Feist, Raymond E.       1985       Bantam Spectra       343       8.00       SF&amp;F                   19       7/11/1994       Darkness at Sethanon, A       Feist, Raymond E.       1986       Bantam Spectra       430       8.00       SF&amp;F                   20       7/23/1994       Real Story, The       Donaldson, Stephen R.       1991       Bantam Spectra       241       3.00       SF&amp;F                   21       7/25/1994       Faerie Tale       Feist, Raymond E.       1988       Bantam Spectra       435       5.00       SF&amp;F                   22       7/27/1994       Golden, The       Shepard, Lucius       1993       Bantam Books       196       2.00       HORR                   23       8/15/1994       Worldwar: In the Balance       Turtledove, Harry       1994       Del Ray Book, Ballantine Books       488       2.00       SF&amp;F                   24       8/31/1994       Snow White, Blood Red       Datlow, Ellen &amp; Terri Windling, Editors       1993       AvoNova: William Morrow &amp; Co., Inc.       413       4.00       SF&amp;F                   25       9/16/1994       Ways of Seeing       Berger, John       1972       British Broadcasting Corporation       166       8.00       SCHOL                   26       9/22/1994       List of 7, The       Frost, Mark       1994       Avon Books       401       9.00       FICT                   27       9/24/1994       Multimind: A New Way of Looking at Human  Behavior       Ornstein, Robert       1986       Anchor Book (Doubleday)       206       5.00       SCHOL                   28       9/26/1994       Anno-Dracula       Newman, Kim       1992       Caroll &amp; Graf Publishers, Inc.       353       4.00       HORR                   29       10/1/1994       Christian &amp; Oriental Philosophy of Art       Coomaraswamy, Ananda K.       1943       Dover Publications, Inc.       146       6.00       SCHOL                   30       10/4/1994       Last Aerie, The       Lumley, Brian       1993       TOR Horror       758       5.00       HORR                   31       10/7/1994       Quantum Reality       Herbert, Nick       1985       Anchor Books (Doubleday)       268       7.00       SCHOL                   32       10/14/1994       Sexual Personae       Paglia, Camille       1990       First Vintage Books       431       3.00       SCHOL                   33       10/16/1994       Without Remorse       Clancy, Tom       1993       Berkley       750       7.00       FICT                   34       10/18/1994       Ficciones       Borges, Jorge Luis       1956       Grove Weidenfeld       174       8.00       SCHOL                   35       10/26/1994       Sum of All Fears, The       Clancy, Tom       1991       Berkley       914       7.00       FICT                   36       11/12/1994       America       Baudrillard, Jean       1986       Verso       129       2.00       SCHOL                   37       11/12/1994       October 70 (magazine)       Various       1994       MIT Press       89       6.00       SCHOL                   38       11/19/1994       Reenchantment of Art, The       Gablik, Suzi       1991       Thames and Hudson       191       6.00       SCHOL                   39       12/2/1994       Consider Phlebas       Banks, Iain M.       1987       Bantam       497       8.00       SF&amp;F                   40       12/21/1994       Druids       Llywelyn, Morgan       1991       Ivy Books       404       8.00       SF&amp;F                   41       1/14/1995       Grails: The Quest of the Dawn       Gillam, Richard (ed.), et. al.       1994       ROC       387       7.00       SF&amp;F                   42       1/20/1995       Taliesen       Lawhead, Stephen R.       1987       Avon Books       486       7.00       SF&amp;F                   43       1/27/1995       Merlin       Lawhead, Stephen R.       1988       Avon Books       447       7.00       SF&amp;F                   44       2/28/1995       Arthur       Lawhead, Stephen R.       1990       Avon Books       446       7.00       SF&amp;F                   45       3/10/1995       Vampires       Steakley, John       1990       ROC Fantasy       357       2.00       SF&amp;F                   46       4/1/1995       Writing for Business       Wilbers, Stephen       1993       Good Writing Press       150       7.00       BUS                   47       4/9/1995       Pendragon       Lawhead, Stephen R.       1994       AvoNova Books       436       7.00       SF&amp;F                   48       4/19/1995       Wild Swans: Three Daughters of China       Chang, Jung       1991       Anchor Books, Doubleday       524       9.00       HIST                   49       4/22/1995       Bone Dance       Bull, Emma       1991       Ace Science Fiction       279       7.00       SF&amp;F                   50       5/3/1995       Tales of the Dervishes: Teaching Stories of the        Sufi Masters       Shah, Idries       1967       E.P. Dutton &amp; Co., Inc.       222       5.00       SCHOL                   51       5/4/1995       Style: Ten Lessons in Clarity and Grace       Williams, Joseph M.       1994       HarperCollinsCollegePublishers: NY, NY       260       8.00       SCHOL                   52       5/7/1995       Congo       Crichton, Michael       1980       Ballantine Books: NY, NY       316       7.00       FICT                   53       5/11/1995       Prose Reader, The: Essays for Thinking, Reading        &amp; Writing       Flachmann, Michael and Kim       1993       Prentice Hall: Englewood Cliffs, NJ       574       6.00       SCHOL                   54       6/5/1995       Finder       Bull, Emma       1994       TOR Fantasy: NY, NY       317       7.00       SF&amp;F                   55       6/9/1995       Ring of Swords       Arnason, Eleanor       1993       Tom Doherty Associates: NY, NY       383       8.00       SF&amp;F                   56       6/22/1995       Republic, The       Plato       -420       Penguin Classics: NY, NY (Sir Desmond Lee, Ed.      1955)       469       8.00       CLAS                   57       6/26/1995       Utopia       More, Sir Saint Thomas       1516       Harlan Davidson: Arlington Hts, IL (HVS Ogden, Ed.        1949)       84       9.00       SCHOL                   58       6/29/1995       Communist Manifesto, The       Marx, Karl and Friedrich Engels       1888       Penguin Classics: Middlesex, England       124       6.00       SCHOL                   59       7/3/1995       We       Zamyatin, Yevgeny       1920       Avon Books: NY, NY (Mira Ginsburg, Ed. 1972).       232       7.00       SF&amp;F                   60       7/5/1995       Shipping News, The       Proulx, E. Annie       1993       Touchstone: NY, NY       337       10.00       FICT                   61       7/10/1995       Brave New World       Huxley, Aldous       1932       Harper Colophon: NY, NY (1965)       199       7.00       SF&amp;F                   62       7/13/1995       Ubik       Dick, Philip K.       1969       Vintage Books: NY, NY (1991)       216       8.00       SF&amp;F                   63       7/15/1995       Hot Zone, The       Preston, Richard       1994       Anchor Book, Doubleday: NY, NY       422       8.00       NONFI                   64       7/17/1995       Left Hand of Darkness, The       LeGuin, Ursula K.       1969       Ace Science Fiction: NY, NY       304       7.00       SF&amp;F                   65       7/20/1995       Creating A New Civilization: The Politics of the        Third Wave       Toffler, Alvin and Heidi (forward by Newt      Gingrich)       1995       Turner Publishing Inc.: Atlanta, GA       112       7.00       BUS                   66       7/24/1995       Popcorn Report, The: The Future of Your Co., World,        Life       Popcorn, Faith       1991       HarperBusiness (HarperCollins Publishers):       100       1.00       BUS                   67       8/12/1995       Alienist, The       Carr, Caleb       1994       Bantam (Random House): NY, NY       600       8.00       FICT                   68       8/16/1995       Swords Trilogy, The (King, Queen, Knight)       Moorcock, Michael       1971       Ace Books (Berkley Publishing): NY, NY (1988)       403       9.00       SF&amp;F                   69       8/19/1995       Chronicles of Corum, The       Moorcock, Michael       1978       A Berkley Book: NY, NY       381       8.00       SF&amp;F                   70       9/7/1995       Debt of Honor       Clancy, Tom       1994       A Berkley Book: NY, NY (Jack Ryan Ltd.      Partnership)       990       6.00       FICT                   71       9/28/1995       Three Stigmata of Palmer Aldrich, The       Dick, Philip K.       1964       Vintage Books: NY, NY (1991)       230       7.00       SF&amp;F                   72       9/29/1995       Creation of Feminist Consciousness, The       Lerner, Gerda       1993       Oxford University Press: NY, NY       393       7.00       SCHOL                   73       10/10/1995       Diamond Age, The       Stephenson, Neal       1995       Bantam Books: NY, NY       419       7.00       SF&amp;F                   74       10/14/1995       Bloodwars       Lumley, Brian       1994       Tom Doherty Associates: NY, NY       760       6.00       HORR                   75       10/21/1995       Middle Ages, The       Bishop, Morris       1968       Houghton Mifflin Co.: Boston, MA       350       9.00       HIST                   76       11/1/1995       Grand Inquisitor, The       Dostoyevsky, Fyodor       1880       Penguin Books: NY, NY (1995)       55       9.00       FICT                   77       12/1/1995       Hastur Cycle, The       Price, Robert M. (editor)       1993       Chaosium, Inc.: Oakland, CA       304       6.00       HORR                   78       12/15/1995       Mysteries of the Worm       Price, Robert M. (editor) (Bloch, Robert)       1993       Chaosium, Inc.: Oakland, CA       263       6.00       HORR                   79       12/16/1995       Man in the High Castle, The       Dick, Philip K.       1962       Vintage Books: NY, NY       259       8.00       SF&amp;F                   80       12/18/1995       Towing Jehovah       Morrow, James       1994       Harcourt, Brace and Co: NY, NY       371       8.00       SF&amp;F                   81       1/19/1996       War for the Oaks       Bull, Emma       1987       ACE Books: NY, NY       309       8.00       SF&amp;F                   82       1/29/1996       Tales of the White Wolf       Kramer, Edward E. &amp; Richard Gilliam, Editors        (Moorcock, Michael)       1994       White Wolf, Inc: Stone Mountain, GA       381       6.00       SF&amp;F                   83       1/30/1996       Book of Iod, The (The Eater of Souls &amp; Other        Tales)       Price, Robert M. (editor) (Henry Kuttner)       1995       Chaosium, Inc.: Oakland, CA       205       6.00       HORR                   84       2/9/1996       Odyssey, The       Homer (trans. by Robert Fitzgerald)              Vintage Classics, Random House: NY, NY (1961)       509       8.00       CLAS                   85       2/12/1996       Zap Gun, The       Dick, Philip K.       1965       Carroll &amp; Graf Publishers, Inc.: NY, NY       258       7.00       SF&amp;F                   86       2/19/1996       Emergency!       Browmn, Mark, M.D.       1996       Villard Books       209       7.00       NONFI                   87       2/23/1996       Job       Unknown (Moses?)              Soncino Press, The, Ltd.: NY, NY (1985)       226       6.00       CLAS                   88       3/1/1996       Odyssey, The       Homer (trans. by Robert Fitzgerald)              Vintage Classics, Random House: NY, NY (1961)       509       8.00       CLAS                   89       3/8/1996       Divine Comedy, The: Inferno       Dante Allegheri (tran. by John D. Sinclair)              Oxford University Press: NY, NY (1939)       432       8.00       CLAS                   90       4/15/1996       Doomsday Book       Willis, Connie       1992       Bantam Books: NY, NY       578       9.00       SF&amp;F                   91       4/30/1996       Aeneid, The       Virgil (trans. by W.F. Jackson Knight)       -19       Penguin Books, Ltd.: Baltimore, MD (1956)       366       8.00       CLAS                   92       5/18/1996       Earthclan: Startide Rising and The Uplift War       Brin, David       1987       Nelson Doubleday, Inc: Garden City, NY (1983/1987 -        two books)       985       8.00       SF&amp;F                   93       5/21/1996       Sundiver       Brin, David       1980       Bantam Books: NY, NY       340       6.00       SF&amp;F                   94       6/8/1996       Heritage Universe, The (Summertide, Divergence,        Transcendence)       Sheffield, Charles       1992       Guild America Books: NY, NY (1990-2)       724       7.00       SF&amp;F                   95       6/13/1996       Good Omens       Gaiman, Neil and Terry Pratchett       1990       Ace: NY, NY       368       7.00       SF&amp;F                   96       6/16/1996       Divine Comedy, The: Purgatorio       Dante Allegheri (tran. by John D. Sinclair)              Oxford University Press: NY, NY (1939)       446       7.00       CLAS                   97       6/16/1996       SeinLanguage       Seinfeld, Jerry       1993       Bantam Books: NY, NY       180       6.00       HUMOR                   98       6/20/1996       Things They Carried, The       O\'Brien, Tim       1990       Penguin Books USA Inc.: NY, NY       273       9.00       FICT                   99       7/15/1996       Aristoi       Williams, Walter Jo       1992       Tor Book, Tom Doherty Associates, Inc: NY, NY       448       7.00       SF&amp;F                   100       7/22/1996       Forbidden Knowledge: The Gap Into Vision       Donaldson, Stephen R.       1991       Bantam Spectra: NY, NY       455       7.00       SF&amp;F                   101       7/30/1996       Dark and Hungry God Arises, A: The Gap Into  Power       Donaldson, Stephen R.       1992       Bantam Books: NY, NY       518       8.00       SF&amp;F                   102       8/1/1996       Mindhunter       Douglas, John and Mark Olshaker       1995       Pocket Books, Simon &amp; Schuster: NY, NY       384       8.00       NONFI                   103       8/12/1996       Chaos and Order: The Gap Into Madness       Donaldson, Stephen R.       1994       Bantam Spectra: NY, NY       674       8.00       SF&amp;F                   104       8/15/1996       Home Brewmaking: The Complete Beginner\'s      Guidebook       Moore, William       1991       Ferment Press: San Leandro, CA       71       7.00       F&amp;D                   105       8/21/1996       Six Messiahs, The       Frost, Mark       1995       Avon Books: NY, NY       423       2.00       FICT                   106       8/22/1996       New Complete Joy of Homebrewing, The       Papazian, Charlie       1984       Avon Books: NY, NY (2nd ed. 1991)       262       8.00       F&amp;D                   107       8/22/1996       Relic, The       Preston, Douglas adn Lincoln Child       1995       Tor Books, Tom Doherty Associates, Inc.: NY, NY       474       7.00       FICT                   108       9/13/1996       Transplanted, The: A History of Immigrants in Urban        America       Bodnar, John       1985       Indiana University Press: Bloomington, IN       294       7.00       SCHOL                   109       9/14/1996       Lost World, The       Crichton, Michael       1995       Ballantine Books: NY, NY       430       8.00       FICT                   110       9/19/1996       Erin\'s Daughters in America       Diner, Hasia R.       1983       John Hopkins University Press: Baltimore, MD       192       7.00       SCHOL                   111       9/21/1996       Treason       Card, Orson Scott       1979       St. Martin\'s Paperbacks: NY, NY       309       7.00       SF&amp;F                   112       9/30/1996       Fugitive From the Cubicle Police: A Dilbert Book       Adams, Scott       1996       Andrews and McMeel: Kansas City, MO       224       9.00       HUMOR                   113       10/1/1996       Whoever Fights Monsters       Ressler, Robert H. and Tom Shachtman       1992       St. Martins Paperbacks: NY, NY       289       8.00       NONFI                   114       10/2/1996       Strangers from a Different Shore       Takaki, Ronald       1989       Penguin Books: NY, NY       570       8.00       SCHOL                   115       10/9/1996       Out of This Furnace       Bell, Thomas       1941       University of Pittsburgh Press: Pittsburgh, PA        (1976)       424       7.00       SCHOL                   116       10/13/1996       Genetic Soldier       Turner, George       1994       Avonova, William Morrow &amp; Co, Inc: New York,      NY       403       5.00       SF&amp;F                   117       10/20/1996       Earth       Brin, David       1990       Bantam Spectra: NY, NY       682       7.00       SF&amp;F                   118       10/25/1996       America Play, The       Parks, Suzan-Lori                     41       8.00       SCHOL                   119       10/25/1996       Becoming Mexican American: Ethnicity, Culture &amp;        Identity in Chicano Los Angeles, 1900-1945       Sanchez, George J.       1993       Oxford University Press: NY, NY       367       7.00       SCHOL                   120       10/25/1996       Elements of Style, from       Parks, Suzan-Lori                     12       7.00       SCHOL                   121       10/28/1996       Making a New Deal: Industrial Workers in Chicago,        1919-1939       Cohen, Lizabeth       1990       Cambridge University Press: NY, NY       526       7.00       SCHOL                   122       11/14/1996       Regulators, The       Bachman, Richard (penname for Stephen King)       1996       Dutton, Penguin Group: NY, NY       475       7.00       HORR                   123       11/18/1996       Man Who Mistook His Wife for a Hat, The: And Other        Clinical Tales       Sacks, Oliver       1985       Summit Books, Simon &amp; Schuster, Inc.: NY, NY       233       8.00       PSYCH                   124       11/21/1996       Race Rebels: Culture, Politics and the Black        Working Class       Kelley, Robin D.G.       1994       The Free Press, A Division of Macmillan, Inc: NY,      NY       351       8.00       HIST                   125       12/4/1996       America: Who Stole the Dream       Barlett, Donald L. &amp; James B. Steele       1996       The Philadelphia Inquirer (World Wide Web      Article)       10       6.00       WWW                   126       12/4/1996       Can We Still Afford to Be a Nation of  Immigrants?       Kennedy, David M.       1996       Atlantic Monthly Magazine       10       7.00       MAGAZ                   127       12/4/1996       Ethnic Myth, The: Race, Ethnicity, and Class in        America       Steinberg, Stephen       1981       Atheneum: NY, NY       277       8.00       HIST                   128       12/4/1996       Immigration Wars, The       Laws, Bart       1996       Z Magazine       9       7.00       MAGAZ                   129       12/4/1996       New Economics of Immigration, The       Borjas, George J.       1996       Atlantic Monthly Magazine       7       7.00       MAGAZ                   130       12/4/1996       West Unique, Not Universal, The       Huntington, Samuel P.       1996       Foreign Affairs Magazine: Nov/Dec. 1996, vol. 75,        no. 6       19       6.00       MAGAZ                   131       12/6/1996       Defense in an Age of Hope       Perry, William J.       1996       Foreign Affairs Magazine: Nov/Dec. 1996, vol. 75,        no. 6       16       8.00       MAGAZ                   132       12/6/1996       Germany\'s New Right       Heilbrunn, Jacob       1996       Foreign Affairs Magazine: Nov/Dec. 1996, vol. 75,        no. 6       19       7.00       MAGAZ                   133       12/15/1996       Eternal Champion, The       Moorcock, Michael       1957       White Wolf Publishing: Clarkston, GA (1994)       484       7.00       SF&amp;F                   134       12/28/1996       Bean Trees, The       Kingsolver, Barbara       1988       Harper &amp; Row: NY, NY       232       8.00       FICT                   135       12/29/1996       Dilbert Principle, The: A Cubicle\'s-Eye View of        Bosses, Meetings, Mgmt Fads, etc.       Adams, Scott       1996       HarperBusiness (HarperCollins Publishers): NY,  NY       336       8.00       BUS                   136       1/18/1997       Sustaining the Illusion: An Empirical Study of        Adult Bulletin Boards in St. Paul/Mpls       Dunker, Marcia L. (unpublished MALS thesis, Hamline        University)       1992       Hamline University (unpublished)       123       5.00       SCHOL                   137       1/22/1997       Beowulf\'s Children       Niven, Larry, Jerry Pournelle and Steven Barnes       1995       TOR (Tom Doherty Associates): NY, NY       493       7.00       SF&amp;F                   138       1/25/1997       In the Lake of the Woods       O\'Brien, Tim       1994       Penguin Books USA, Inc.: NY, NY       303       9.00       FICT                   139       1/27/1997       Shadow\'s End       Tepper, Sheri S.       1994       Bantam Books: NY, NY       454       7.00       SF&amp;F                   140       3/5/1997       Belgarath the Sorcerer       Eddings, David &amp; Leigh       1995       Del Rey Book, Ballentine Books: NY, NY       723       7.00       SF&amp;F                   141       2/25/1997       Foreigner       Cherryh, C.J.       1994       DAW Books, Penguin U.S.A.: NY, NY       428       8.00       SF&amp;F                   142       2/22/1997       Scivias       Hildegard of Bingen       1151       Paulist Press: Mahwah, NJ (1990)       316       7.00       SCHOL                   143       2/15/1997       Timaeus and Critias       Plato       -350       Penguin Classics: NY, NY (Desmond Lee trans.      1977)       167       7.00       SCHOL                   144       2/15/1997       Theogony       Hesiod       -750       Macmillan Publishing Co: NY, NY (Norman O. Brown        trans. 1953)       87       7.00       SCHOL                   145       3/15/1997       Mystical Qabalah, The       Fortune, Dion       1935       Samuel Weiser, Inc. (1984)       311       8.00       SCHOL                   146       6/6/1997       Starship Troopers       Heinlein, Robert A.       1959       Ace Books: NY, NY (1987)       208       7.00       SF&amp;F                   147       5/15/1997       Deathbird Stories       Ellison, Harlan       1983       Bluejay Books, Inc.: NY, NY       297       7.00       SF&amp;F                   148       3/22/1997       Psychology &amp; Alchemy       Jung, CG (tran. by RFC Hull)       1944       Princeton University Press: Princeton, NJ (1993)       264       6.00       SCHOL                   149       3/29/1997       Marriage of Heaven and Hell, The       Blake, William       1789       Oxford University Press: Oxford, England (1975)       54       5.00       SCHOL                   150       3/29/1997       Doctor Faustus (&amp; The History of the Damnable        Life &amp; Deserved Death of Dr. John Faustus)       Marlowe, Christopher       1600       Signet Classic: Penguin Group: NY, NY (1969)       176       8.00       SCHOL                   151       4/5/1997       Book of Lies, The       Crowley, Aleister       1913       Samuel Weiser, Inc: York Beach, ME (1981)       196       5.00       SCHOL                   152       4/12/1997       Giordano Bruno and the Hermetic Tradition       Yates, Frances A.       1964       University of Chicago Press: Chicago, IL (1991)       466       8.00       SCHOL                   153       7/1/1997       Morbid Taste for Bones, A       Peters, Ellis       1977       Mysterious Press Books, Warner Books: NY, NY       197       7.00       MYSTERY                   154       7/4/1997       Geek Love       Dunn, Katherine       1989       Warner Books: NY, NY (1983)       355       8.00       FICT                   155       7/8/1997       Desparation       King, Stephen       1997       Penguin Books USA Inc: NY, NY       547       7.00       HORR                   156       9/13/1997       Hyperion Cantos (Hyperion and Fall of Hyperion)       Simmons, Dan       1990       Guild America Books (DoubleDay): NY, NY       929       7.50       SF&amp;F                   160       10/25/1997       The Dilbert Future: Thriving on Stupidity in the        21st Century       Adams, Scott       1997       HarperBusiness       258       8.00       HUMOR                   161       12/24/1997       The Postman       Brin, David       1985       Bantam Books       321       4.00       SF&amp;F                   162       1/17/1998       Blameless in Abaddon       Morrow, James       1996       Harcourt Brace &amp; Company       404       7.00       SF&amp;F                   163       2/3/1998       Jack and Rochelle: A Holocaust Story of Love and        Resistance       Sutin, Jack and Rochelle (Ed. Lawrence Sutin)       1995       Graywolf Press       218       9.00       NONFI                   164       2/13/1998       The Night in Question       Wolff, Tobias       1996       Vintage Books       206       5.00       FICT                   165       3/2/1998       The Stars My Destination       Bester, Alfred       1956       Vintage Books       258       7.00       SF&amp;F                   166       4/11/1998       Endymion       Simmons, Dan       1996       Bantam Spectra: NY, NY       563       8.00       SF&amp;F                   167       4/4/1998       Out of the Silent Planet       Lewis, C.S.       1996 (1943)       Scribner Paperback Fiction: NY, NY       160       8.00       SF&amp;F                   168       2/1/1998       Otherland       Williams, Tad       1996       DAW (Penguin): NY, NY       780       8.00       SF&amp;F                   169       12/1/1997       HTML: The Definitive Guide       Musciano, Chuck &amp; Bill Kennedy       1996       O\'Reilly &amp; Associates       324       5.00       COMPUTER                   170       4/26/1998       Wizard and Glass: The Dark Tower IV       King, Stephen       1996       Plume (Penguin): NY, NY       672       7.50       SF&amp;F                   171       5/26/1998       Xenocide       Card, Orson Scott       1991       Tom Doherty Associates, Inc.: NY, NY       592       6.00       SF&amp;F                   172       7/13/1998       Angel of Darkness, The       Carr, Caleb       1997       Ballantine Publishing Group: NY, NY       752       8.50       FICT                   173       7/20/1998       Children of the Mind       Card, Orson Scott       1996       Tom Doherty Associates, Inc.: NY, NY       370       5.50       SF&amp;F                   174       7/27/1998       Legacy of Heorot, The       Niven, Larry &amp; Jerry Pournelle and Steven      Barnes       1987       Pocket Books, Simon and Schuster: NY,NY       383       7.50       SF&amp;F                   175       7/30/1998       Reliquary       Preston, Douglas &amp; Lincoln Child       1997       Tom Doherty Associates, Inc.: NY, NY       464       5.00       FICT                   176       8/9/1998       Daughter of the Empire       Wurts, Janny &amp; Raymond Feist       1987       Bantam Books: NY, NY       421       7.50       SF&amp;F                   177       8/16/1998       Servant of the Empire       Wurts, Janny &amp; Raymond Feist       1990       Bantam Books: NY, NY       691       8.00       SF&amp;F                   178       7/16/1998       Visual Basic 5 Certification Exam Guide      (bootcamp)       Merzick, Dan &amp; Scot Hillier       1998       McGraw-Hill: NY, NY       496       8.00       COMPUTER                   179       9/7/1998       Mistress of the Empire       Wurts, Janny &amp; Raymond Feist       1992       Bantam Books: NY, NY       676       7.00       SF&amp;F                   180       9/19/1998       Cider House Rules, The       Irving, John       1985       A Ballantine Book: NY, NY       598       8.50       FICT                   181       10/14/1998       Autobiography of My Mother, The       Kincaid, Jamaica       1996       Plume, Penguin Group: NY, NY       228       4.00       FICT                   182       12/1/1998       Neverwhere       Gaiman, Neil       1996       Avon Books, Inc.: NY, NY       388       8.50       SF&amp;F                   183       12/14/1998       This is the Way the World Ends       Morrow, James       1986       Harcourt, Brace &amp; Company: NY, NY       319       7.00       SF&amp;F                   184       2/2/1999       Biggest Elvis       Kluge, P.F.       1996       Penguin Books, USA, Inc: NY, NY       341       8.50       FICT                   185       4/2/1999       Memoirs of a Geisha       Golden, Arthur       1997       Vintage Contemporaries: NY, NY       434       8.00       FICT                   186       4/11/1999       Chronicles of the Celts       Zaczek, Iain       1996       Collins &amp; Brown Ltd.: London, UK       159       8.00       HIST                   187       5/1/1999       Instance of the Fingerpost, An       Pears, Iain       1998       Berkley Publishing Group: NY, NY       735       9.50       FICT                   188       5/8/1999       Shadows of the Empire (Star Wars)       Perry, Steve       1996       Bantam Books: NY, NY       385       5.50       SF&amp;F                   189       5/15/1999       Accidental Empires: How the Boys of Silicon Valley        Make Their Millions&#8230;       Cringely, Robert X.       1996 (1992)       HarperBusiness, HarperCollins: NY, NY       370       8.50       NONFI                   190       5/22/1999       Fire Upon the Deep, A (Hugo Award)       Vinge, Vernor       1992       Tom Doherty Associates, Inc (TOR): NY, NY       613       9.00       SF&amp;F                   191       5/31/1999       Excession       Banks, Iain M.       1996       Bantam Books: NY, NY       499       8.50       SF&amp;F                   192       6/1/1999       Glensheen\'s Daughter: The Marjorie Congdon Story       Darby Hendry, Sharon       1999       Cable Publishing: Bloomington, MN       332       8.00       NONFI                   193       7/1/1999       To Your Scattered Bodies Go: Book 1 of  Riverworld       Farmer, Philip Jose\'       1971       The Ballantine Publishing Group: NY, NY       220       7.50       SF&amp;F                   194       8/1/1999       Cthulu 2000: A Lovecraftian Anthology       Turner, Jim (ed.)       1995       The Ballantine Publishing Group: NY, NY       398       7.00       SF&amp;F                   195       9/1/1999       Necroscope: the Lost Years       Lumley, Brian       1995       A Tom Doherty Associates Book: NY, NY       593       7.00       SF&amp;F                   196       10/1/1999       Feersum Endjinn       Banks, Iain M.       1996       Bantam Books: NY, NY       309       7.50       SF&amp;F                   197       11/1/1999       Darwinia       Wilson, Robert Charles       1998       A Tom Doherty Associates Book: NY, NY       320       8.50       SF&amp;F                   198       11/18/1999       Game of Thrones, A       Martin, George R.R.       1997       Bantam Spectra: NY, NY       835       8.50       SF&amp;F                   199       12/1/1999       Death in Yellowstone: Accidents &amp; Foolhardiness        in the First National Park       Whittlesey, Lee H.       1995       Roberts Rinehart Publishers: Boulder, CO       276       8.50       NONFI                   200       12/19/1999       Bag of Bones       King, Stephen       1998       Pocket Books: NY, NY       732       8.00       HORR                   201       1/3/2000       River of Blue Fire: Otherland Volume Two       Williams, Tad       1998       DAW Books, Inc: NY, NY       634       7.50       SF&amp;F                   202       2/1/2000       Stardust       Gaiman, Neil       1999       Avon Books, Inc.: NY, NY       336       8.00       SF&amp;F                   203       3/1/2000       Archangel       Harris, Robert       1999       Jove Books: NY, NY       415       7.50       FICT       A very unexceptional ending.            204       3/22/2000       Deepness in the Sky, A       Vinge, Vernor       1999       TOR: NY, NY       774       7.50       SF&amp;F       Not as good as the last book by Vinge.            205       8/5/2000       Tangle Box, The       Brooks, Terry       1994       Del Rey: NY, NY       341       6.00       SF&amp;F                   206       8/1/2000       Final Truth: the Autobiography of a Serial  Killer       Gaskins, Donald &amp; Wilton Earle       1992       Adept: Atlana, GA       222       8.00       NONFI       Very very disturbing - about perhaps the most        prolific mass murderer in U.S. history.            207       5/15/2000       Snowcrash       Stephenson, Neal       1992       Bantam Spectra: NY, NY       468       8.50       SF&amp;F                   208       8/20/2000       Cassini Division, The       MacLeod, Ken       1998       TOR, LLC: NY, NY       305       8.50       SF&amp;F       Warring political ideologies and rampant aliens        that are former humans turned AI.            209       12/1/2000       It\'s Not About the Bike: My Journey Back to Life       Armstrong, Lance (w/Sally Jenkins)       2000       G.P. Putnam\'s Sons: NY, NY       275       8.00       NONFI       Lance\'s autobiography - you get a feel for his        personality from the book.            210       1/1/2001       Dune: House Atreides       Herbert, Brian and Kevin J. Anderson       1999       Bantam Books: NY, NY       604       1.50       SF&amp;F       AWFUL - contrived, clichÃƒÂ©d, predictible - you know        smart people are smart because the authors assure you they\'re smart.            212       2/1/2001       Necroscope: Invaders       Lumley, Brian       1999       TOR: NH, NY       534       5.00       HORR                   213       3/12/2001       Eternal Footman, The       Morrow, James       1999       A Harvest Book, Harcourt, Inc.: NY, NY       359       8.00       SF&amp;F       The third in Morrow\'s trilogy about God\'s death.        The body of the book is good, but the ending is weak (I think his endings        are usually weak).            214       2/14/2001       Only Begotten Daughter       Morrow, James       1990       William Morrow and Company, Inc.       312       8.50       SF&amp;F       God gives the world his only daughter. She\'s born        in a test tube to a man.            215       1/14/2001       Clash of Kings, A       Martin, Geroge R.R.       1999       Bantam Spectra: NY, NY       1009       8.00       SF&amp;F       2nd in the series - very good, I like him better        than Jordan - he focuses on less characters so they have a bit more      depth.            216       3/17/2001       Celestial Matters       Garfinkle, Richard       1996       A Tor Book: NY, NY       348       8.25       SF&amp;F       Greeks and Taoists rule the world and are eternally        at war. One of the Greeks attempts to bring back a piece of the sun to        drop on the Taoist capital.            217       3/26/2001       Stone Canal, The       MacLeod, Ken       1996       A Tor Book: NY, NY       339       7.00       SF&amp;F       A precursor to The Cassini Division - rather        disjointed and hard to follow in places.            218       4/1/2001       Body Bazaar: the Market for Human Tissue in the        Biotechnology Age       Andrews, Lori and Dorothy Nelkin       2001       Crown Publishers: NY, NY       237       6.00       NONFI       About biotechnology tissue patenting and related        medical issues - fairly one-sided, not well written. Could have used a lot        more concrete examples in order to be interesting. Nonetheless, some        enjoyable anecdotes and useful generalized information.            219       4/8/2001       Storm of Swords, A       Martin, George R.R.       2000       Bantam Spectra: NY, NY       973       7.75       SF&amp;F       Not as good as the previous two books in his        series, or else I\'m not as excited with them as I was when I started.        Couple of main characters died in this book, however, which made for an        interesting novel.            220       4/24/2001       Moonseed (Biohazard, Moonseed)       Baxter, Stephen       1998       Harper Prism: NY, NY       662       6.25       SF&amp;F                   221       5/11/2001       Cryptonomicon       Stephenson, Neal       1999       Avon Books: NY, NY       918       9.25       SF&amp;F       Not really Science Fiction and Fantasy, but that\'s        where all the bookstores seem to place it. A great novel about WWII code        encryption tied to a present day search for gold that was hidden using        those codes. A great interweaving of the WWII characters, and their        offspring in the present day - they\'re almost mirrored.            222       5/1/2001       Distributed Applications with Visual Interdev 6.0:        for Exam 70-152       Microsoft       2000       Microsoft Press: Redmond, WA       488       1.00       COMP       No good - if I had studied this and not the        Transcender tests, I\'d have probably flunked the MCP.            223       5/18/2001       Souls in the Great Machine       McMullen, Sean       1999       A Tor Book: NY, NY       448       8.00       SF&amp;F                   224       5/26/2001       Hundred Years War (The): The English in France        1337-1453       Seward, Desmond       1999 (1978)       Penguin Books: NY, NY       282       8.50       HIST       Very nice overview - easy to read.            225       6/2/2001       Wicked: The Life and Times of the Wicked Witch of        the West       Maguire, Gregory       1995       ReganBooks, Harper Colins: NY, NY       406       8.00       SF&amp;F                   226       6/10/2001       Skyroad, The       MacLeod, Ken       1999       Tor Books: NY, NY       291       7.75       SF&amp;F                   227       7/3/2001       Red Lobster, White Trash, and the Blue Lagoon       Queenan, Joe       1998       Hyperion: NY, NY       194       8.50       HUMOR       Last 1/3 was a little slow until it picked up at        the end.            228       8/15/2001       Inversions       Banks, Iain M.       1998       Pocket Books: NY, NY       376       7.25       SF&amp;F                   229       8/18/2001       Perdido Street Station       Mieville, China       2000       Pan Books: London, England       867       9.25       SF&amp;F                   230       8/20/2001       Hinge Factor, The       Durschmied, Erik       1999       Coronet Paperback: London, England       378       5.00       HIST                   231       8/12/2001       Complete Book of Long-Distance Cycling, The       Burke, Edmund R. (Ph.D.) and Ed Pavelka       2000       St. Martin\'s Press, Rodale: U.S.       292       9.50       NONFI                   232       7/28/2001       American Gods, A Novel       Gaiman, Neil       2001       William Morrow: NY, NY       465       8.50       SF&amp;F                   233       9/2/2001       Balsamic Dreams: A Short but Self-Important History        of the Baby Boomer Generation       Queenan, Joe       2001       Henry Holt and Co, LLC: NY, NY       211       8.50       HUMOR       Validated my feelings about overalls. #1 way to        tell if someone is a baby boomer: do they own a copy of Carole King\'s        \"Tapestry\".            234       9/6/2001       Great Australian Urban Myths: Revised Edition (The        Cane Toad High)       Seal, Graham       1995 (2001)       HarperCollinsPublishers: Sydney, Australia       240       4.00       NONFI       Wasn\'t very good because 1.) I\'d already heard most        of the stories elsewhere, there wasn\'t a lot of new information and 2.) I        didn\'t particularly like his writing style, he repeated the same points        over and over and tried to overexplain everything.            235       9/9/2001       Wars of the Roses, The (through the lives of five        men and women of the fifteenth century)       Seward, Desmond       1995       Viking (Penguin Group): NY, NY       379       8.00       HIST       His style is great and easy to read, but he seems        to have some biases, in particular, you get the impression he would have        done most anything to meet Margaret Beaufort (Henry VII\'s mother). His use        of the \"characters\" to drive the book seemed strained in some places by        lack of information (as concerns the more \'common\' characters, like Jane        Shore\'s father).            236       9/24/2001       Tale of Two Cites, A       Dickens, Charles       1859 (1981)       Bantam Books: NY, NY       369       7.75       SCHOL       I liked many of his other books better than this        one. It was interesting to compare the movie (4 hour) and the book,        however.            237       10/1/2001       Giovanni\'s Room       Baldwin, James       1956 (1988)       Laurel Book (Dell Publishing): NY, NY       224       8.25       SCHOL       Writing is absolutely wonderful.            238       10/1/2001       Year\'s Best SF 6 (All New Astonishing Stories)       Hartwell, David G. (Ed.)       2001       Eos (HarperCollinsPublishers): NY, NY       500       8.75       SF&amp;F       Been too long since I read a collection of short SF        stories - some very good writing.            239       10/14/2001       War for the Oaks       Bull, Emma       2001 (1987)       Tom Doherty Associates: NY, NY       332       9.00       SF&amp;F                   240       10/14/2001       Two Girls, Fat and Thin       Gaitskill, Mary       1991       Scribner Paperback, Simon &amp; Schuster: NY, NY       313       5.25       SCHOL                   241       10/19/2001       Harry Potter and the Sorcerer\'s Stone       Rowling, J.K.       1997       Scholastic, Inc: NY, NY       309       9.00       SF&amp;F                   242       10/21/2001       Harry Potter and the Chamber of Secrets       Rowling, J.K.       1999       Scholastic, Inc: NY, NY       341       8.75       SF&amp;F                   243       10/26/2001       Harry Potter and the Prisoner of Azkaban       Rowling, J.K.       1999       Scholastic, Inc: NY, NY       435       8.75       SF&amp;F                   244       10/28/2001       Cambridge       Phillips, Caryl       1991       Borzoi Book, Alfred A. Knopf, Inc.: NY, NY       184       7.50       SF&amp;F       Wonderful stylistically (language), but rather        boring otherwise.            245       11/5/2001       Harry Potter and Goblet of Fire       Rowling, J.K.       2000       Scholastic, Inc: NY, NY       734       9.00       SF&amp;F                   246       11/12/2001       Jazz       Morrison, Toni       1992       Plume (Penguin Books): NY, NY       229       8.75       SCHOL       I\'d give her about a 10 for writing style, but I        didn\'t enjoy the story as much as the writing style.            247       11/19/2001       Communitas       Goodman, Percival and Paul       1960 (1947)       Columbia University Press: NY, NY       280       6.00       SCHOL                   248       11/30/2001       He, She and It       Piercy, Marge       1991       Alfred A. Knopf: NY, NY       446       6.75       SCHOL                   249       1/3/2002       Ape and Essence: A Novel       Huxley, Aldous       1992 (1948)       Elephant Paperbacks, Ivan R. Dee, Publisher:        Chicago, IL       205       6.50       SF&amp;F                   250       1/18/2002       Brave New World       Huxley, Aldous       1998 (1932)       Perennial Classics (Harper): NY, NY       268       8.50       SF&amp;F                   251       2/14/2002       Shadow of a Dark Queen (Book One of the Serpentwar        Saga)       Feist, Raymond E.       1994       Avon Books, HarperCollins Publishers: NY, NY       497       7.00       SF&amp;F                   252       2/21/2002       Rise of a Merchant Prince (Book Two of the        Serpentwar Saga)       Feist, Raymond E.       1995       Avon Books, HarperCollins Publishers: NY, NY       436       6.75       SF&amp;F                   253       2/28/2002       Rage of a Demon King (Volume III of the Serpentwar        Saga)       Feist, Raymond E.       1997       Avon Books, HarperCollins Publishers: NY, NY       572       6.50       SF&amp;F                   254       3/7/2002       Nineteen Eighty-Four       Orwell, George       1992 (1949)       Alfred A. Knopf: NY, NY       326       9.25       SF&amp;F                   255       3/25/2002       Year\'s Best SF 5 (All New Amazing Stories)       Hartwell, David G. (Ed.)       2000       Eos (HarperCollinsPublishers): NY, NY       494       8.75       SF&amp;F       Chipmunks story - no good. Macs story (Terry        Bison), good - about everyone getting their own criminal to kill (Timothy        MacVeigh). Fossil Games by Tom Purdom not really a short story (about 50        pages), but great detail. Had some aspects like my Back Issue story. Chris        Lawson\'s Written in Blood about putting the Koran into the bloodstream was        innovative.            256       4/1/2002       Algernon, Flowers For       Keyes, Daniel                     268       8.50       SF&amp;F                   257       5/2/2002       Shards of a Broken Crown (Volume IV of the        Serpentwar Saga)       Feist, Raymond E.       1998       EOS: NY, NY       512       6.50       SF&amp;F                   258       5/12/2002       Time Ships, The       Baxter, Stephen       1995       EOS: NY, NY       520       9.00       SF&amp;F       Much better than Moonseed - about H.G. Wells\'        character from the Time Machine and his \"looping\" adventure post and        pre-Morlock world.            259       6/12/2002       Darwin\'s Radio       Bear, Greg       1999       Ballentine Books: NY, NY       538       7.00       SF&amp;F       He delves a little on specifics at the expense of        story.            260       6/18/2002       Revelation Space       Reynolds, Alastair       2002       Ace Books: NY, NY       585       9.25       SF&amp;F                   261       6/24/2002       Galveston       Stewart, Sean       2002       Ace Books: NY, NY       454       8.00       SF&amp;F                   262       7/10/2002       Chronliths, The       Wilson, Robert Charles       2001       TOR: NY, NY       315       8.25       SF&amp;F                   263       7/19/2002       Biological Weapons and America\'s Secret War:      Germs       Miller, Judith and Stephen Engelberg and William        Broad       2001       Simon and Schuster: NY, NY       320       8.00       NONFI       382 pages with appendices/etc.            264       7/21/2002       Darwin Awards, The: Evolution in Action       Northcutt, Wendy       2002       Dutton: NY, NY       327       7.25       NONFI       Lots of short little stories about stupid    people            265       8/6/2002       Icarus Hunt, The       Zahn, Timothy       2000       Bantam: NY, NY       457       8.50       SF&amp;F                   266       8/12/2002       Cosmonaut Keep       MacLeod, Ken       2000       Tor: NY, NY       336       6.50       SF&amp;F       Wish it would have gone somewhere more as a        self-contained book            267       8/17/2002       Bones (An Irene Kelly Novel)       Burke, Jan       1999 (2001)       Signet: NY, NY       423       7.00       NONFI                   268       8/19/2002       Choke       Palahniuk, Chuck       2001 (2002)       Anchor Books: NY, NY       293       8.25       FICT       Tied it up well in the end for not knowing where        the hell it was going sometimes            269       9/4/2002       Evil Geniuses in a Nutshell: A User Friendly Guide        to World Domination       Illiad       2000       O\'Reilly: Cambridge, MA       123       5.00       FICT       Cartoons about Unix - truly not very funny, even if        you like Unix            270       11/11/2002       Resurrection Man       Stewart, Sean              Ace Books: NY, NY       248       7.75       SF&amp;F       Not as good as Galveston - but some interesting        ideas that led to Galveston            271       11/13/2002       Darwin Awards II, The: Unnatural Selection       Northcutt, Wendy       2002       Dutton: NY, NY       240       6.50       NONFI       Lots of duplicates with #1 which was very        disappointing - Larry got #1 for Christmas            272       11/15/2002       Social Life of Information, The       Seely Brown, John and Paul Duguid       2000       Harvard Business School Press: Boston, MA       252       8.75       NONFI       Revised many times - a great book about how        information is about sociological acceptance and use, and that it doesn\'t        work like businesses expect it to, but has a life of its own. Started        reading this WOW weekend, so it took a while.            273       12/1/2002       Music of the Spheres, The       Redfern, Elizabeth       2002       Jove Books: NY, NY       484       7.00       FICT       Not nearly as good as other historical serial        killer mysteries I\'ve read - just not much mystery, really - more about        the history and the (amateur) astronomers.            274       12/12/2002       Spooked: Espionage in Corporate America       Penenberg, Adam L. and Marc Barry       2000       Perseus Publishing: Cambridge, MA       179       8.00       NONFI       All about corporate espionage practices - some good        specific instances, but not enough as those made up the most interesting        parts of the book - particularly cases like Avery Denison turning a small        case of espionage to their own espionage benefit.            275       12/28/2002       Dragonbone Chair, The       Williams, Tad       1998       Daw: NY, NY       783       7.25       SF&amp;F       1 in Memory, Sorrow and Thorn            276       1/5/2003       Stone of Farewell       Williams, Tad       1990       Daw: NY, NY       749       7.25       SF&amp;F       2 in Memory Sorrow and Thorn            277       1/14/2003       Intuitionist, The       Whitehead, Colson       1999       Anchor Books: NY, NY       255       8.50       FICT       Great book about a sort of elevator dystopia - very        well written - lot of attention to crafting if not as much to        characterization - definitely trying to say something            278       1/21/2003       Just Like Beauty       Lerner, Lisa       2002       Picador: NY, NY       279       8.75       FICT       Reminded me of the Geeks books Geri gave me about        sideshow folk - about a sort of dystopian future where young women try out        for a disgusting beauty pagent and one contestant falling in love with        another girl            279       1/28/2003       Survivor       Palahniuk, Chuck       1999       Anchor Books: NY, NY       289       8.50       FICT       Didn\'t like it quite as much as Choke, but still        very good. Quirky, like his other stuff - about a death cult survior who        doesn\'t want to die and meets a girl who knows the future of everything        (sort of like God - and there\'s a lot of the point).            280       2/5/2003       Blackie the Talking Cat: and other favorite        judicial opinions       West Publishing (compiled by)       1996       West Publishing: St. Paul, MN (Eagan)       448       8.00       NONFI       All sorts of unique court cases - was an        interesting way to read my first court cases and see what it is my end        users are up to (in general)            281       2/19/2003       Corrections, The       Franzen, Jonathan       2001       Farrar, Straus and Giroux: NY, NY       568       9.50       FICT       Borrowed from Erik Swenson - wonderful -        exceptional writing.            282       2/23/2003       Flatland: A Romance of Many Dimensions       Abbott, Edwin A.       1992 (1884)       Dover: Mineola, NY       83       8.50       SF&amp;F       Amusing            283       2/25/2003       To Green Angel Tower: Part I       Williams, Tad       1993       Daw: NY, NY       815       7.25       SF&amp;F       3 in Memory, Sorrow and Thorn            284       2/27/2003       Question of Evidence, A       Evans, Colins       2003       John Wiley and Sons: Hoboken, NJ       250       6.75       NONFI       Stories about forensic science over the last two        centuries (and some) - writing wasn\'t very good and he didn\'t flush the        stories out enough (or focus on something specific instead), so very        general.            285       3/9/2003       Everything\'s Eventual: 14 Dark Tales       King, Stephen       2002       Pocket Books: NY, NY       583       5.00       FICT       Pretty much unmemorable - only The Man in the Black        Suit was particularly good, and \"The Little Sisters of Eluria\" if you\'ve        read his Gunslinger books.            286       3/14/2003       Infinity Beach       McDevitt, Jack       2000       EOS: NY, NY       510       7.50       SF&amp;F       Was going to give it a 7 until I saw I\'d given        Green Angel Tower a 7.25 - it was better - the only thing I really didn\'t        like was that it had a somewhat weak ending.            287       3/20/2003       Canticle for Leibowitz, A       Miller, Jr., Walter M.       1980 (1959)       Bantam Books: NY, NY       313       9.25       SF&amp;F       Best SciFi book I\'ve read in a while - classic. The        first section was the best - the last section the second best (of the        three).            288       4/5/2003       Looks to Windward       Banks, Iain M.       2002       Pocket Books: NY, NY       483       8.50       SF&amp;F       Good ending - another Culture book.            289       4/11/2003       Deepsix       McDevitt, Jack       2001       EOS: NY, NY       432       6.50       SF&amp;F       Not as good as his other book, some of the stuff        seemed borrowed or rehashed, and his tendency of tying up loose ends with        an extra chapter is annoying, not satisfying. Beowulf\'s Children was        similar in some respects, but much better.            290       3/30/2003       Vacuum Diagrams: Stories of the Xeelee Sequence       Baxter, Stephen       1997       EOS: NY, NY       483       8.00       SF&amp;F       Quite the scope - Phillip K. Dick winner            291       4/3/2003       Shrub: the Short But Happy Political Life of George        W. Bush       Ivins, Molly and Lou Dubose       2002 (2000)       Vintage Books: NY, NY       237       8.75       POL       Very funny            292       4/18/2003       Angelmass       Zahn, Timothy       2001       TOR: NY, NY       531       7.50       SF&amp;F       Really rather good - he\'s a great writer, be his        endings tend to do a lot of wrapping up - definitely better than      McDevitt            293       4/21/2003       Nanny Diaries, The       McLaughlin, Emma and Nicola Kraus       2003       St. Martin\'s Griffin: NY, NY       306       8.00       FICT       Weak ending, good writing, although definitely a        women\'s book            294       5/20/2003       Herding Cats: A Primer for Programmers Who Lead        Programmers       Rainwater, J. Hank       2003       A! Apress: NY, NY       252       6.50       COMP       Could have been much better - a little obsessed        with his pet VB program for managing time.            295       6/7/2003       Mike Nelon\'s Death Rat       Nelson, Michael J.       2003       Harper Entertainment: NY, NY       325       8.50       FICT       Mystery Science Theater 3000 host - the writing        isn\'t very technically sound in the beginning, but it very much grows on        you, and his story and eye for the humorous detail is absolutely    great.            296       6/17/2003       I Have No Mouth &amp; I Must Scream       Ellison, Harlan       1983 (1967)       Ace Books: NY, NY       152       8.00       SF&amp;F       Short stories - the title story is great. Eyes of        Dust is a very good dystopic story.            297       6/22/2003       Sick Puppy       Hiaasen, Carl       1999       Warner Books, Inc: NY, NY       513       8.50       FICT       A funny ecoterrorism book - main character and an        ex-governor try to teach politicians the joys of nature with the help of a        dog and a rhinoceros.            298       7/14/2003       Integral Trees, The       Niven, Larry       1984       Balletine Books: NY, NY       128       7.00       SF&amp;F                   299       8/5/2003       McSweeney\'s Mammoth Treasure of Thrilling Tails       Chabon, Michael (ed.)       2003       Vintage Books (Random House): NY, NY       479       7.75       FICT       Lot of good stories - some show up in the year\'s        best fiction (2003) later.            300       8/21/2003       One More For the Road       Bradbury, Ray       2003       Avon Books: NY, NY       350       3.50       FICT       Short stories and I didn\'t particularly like any of        them, even the ones that were written well.            301       9/5/2003       Chasm City       Reynods, Alastair       2003 (2001)       Ace Books: NY, NY       694       8.25       SF&amp;F       Not nearly as good as Revelation Space - but still        very enjoyable. Would have been better with a stronger ending.            302       9/14/2003       Lies and the Lying Liars Who Tell Them: A Fair and        Balanced Look at the Right       Franken, Al       2003       Dutton (NY?)       377       9.00       NONFICT       Very enjoyable - Al Franken is very funny.            303       10/14/2003       Lies of George W. Bush, The: Mastering the Politics        of Deception       Corn, David       2003       Crown Publishers: NY, NY       337       8.50       NONFICT       Would have ranked it as high as Franken\'s book, but        for the fact that the lack of humor basically gave me a headache because        it just pummels you with fact after fact after fact about administration        misdirections and lies.            304       11/1/2003       Skipping Towards Gomorrah: The Seven Deadly Sins        and the Pursuit of Happiness in America       Savage, Dan       2003       Plume Book (Penguin): NY, NY       302       9.25       NONFICT       Extremely funny look at being tolerant.            305       12/3/2003       Cosmopolis       DeLillo, Don       2003       Scribner: NY, NY       209       8.00       FICT       The writing was WONDERFUL, but the characterization        wasn\'t. It was more as though he had something he wanted to say and the        character (rich .com entrpreneur) just facilitated what he wanted to      say.            306       12/14/2003       Inquisitors\' Manual, The       Antunes, Antonio Lobo (translated by Richard      Zenith)       2003 (1996)       Grove Press: NY, NY       435       8.75       FICT       Look at Portugal under facists and during a preiod        of extreme change - covered from the vantage of literally dozens of        characters associated with a minister who felt he\'d been slighted.        Translation seemed great as it really flowed and was great to read - not        stilted at all.            307       12/20/2003       Definitive Guide to SQL Server Performance        Optimization, The       Jones, Don       2002       realtimepublishers.com       166       10.00       COMP       For what it is, this is absolutely the best        resource on SQL Server I\'ve ever read and, to top off everything else, it        was free. Except for some duplication that probably could have been        avoided, I think I learned more about optimizing and scaling out SQL        Server in the three days I was reading this than in the last 7 years        working with the product. Glad I printed it out rather than just kept the        PDF file.            308       12/28/2003       Harry Potter and the Order of the Phoenix (Book      5)       Rowling, J.K.       2003       Scholastic: NY, NY       870       8.25       SF&amp;F       Not as good as some of the others - I felt it took        a step back from some of the grown-upness of the last one (went back to a        more \"safe\" mode for kids).            309       1/10/2004       Chindi       McDevitt, Jack       2002       Ace Books: NY, NY       403       7.50       SF&amp;F       Humans find sattelites left by a race that just        sends out lots of drones looking for interesting phenomena/life at        sublight speeds.            310       1/18/2004       Wolves of the Calla       King, Stephen       2003       Scribner: NY, NY       714       7.00       SF&amp;F       He actually published some of this as a short story        I read in McSweeney\'s and that had a lot more immediacy to it - this        seemed to tell much of the same story, just spread out over 700 more      pages.            311       4/18/2004       Worse than Watergate: the Secret Presidency of        George W. Bush       Dean, John W.       2004       Little, Brown and Co.: NY, NY       253       7.50       NONFICT                   312       5/18/2004       Dante Club, The       Pearl, Matthew       2003       Random House: NY, NY       380       9.00       FICT       Serial killer who follows the works of Dante            313       5/19/2004       Pardise War, The (Song of Albion, Book I)       Lawhead, Stephen       1991       Lion Publishing Corp.: Batavia, IL       416       8.00       SF&amp;F       Not as good as the first time I read it - you just        can\'t go back all the time :(            314       6/1/2004       Lamb: The Gospel According to Bif, Christ\'s        Childhood Pal       Moore, Christopher       2002       Perennial (HarperCollins): NY, NY       444       9.00       FICT                   315       6/8/2004       Practical Demonkeeping       Moore, Christopher       1992       Perennial (HarperCollins): NY, NY       243       7.50       FICT                   316       6/15/2004       Lust Lizard of Melancholy Cove, The       Moore, Christopher       1999       Perennial (HarperCollins): NY, NY       304       8.50       FICT       Very funny - Practical Demonkeeping wasn\'t as funny        as the other two, but some of the same characters, the same town, etc,        turn up in this book. '),(5978,3883177,'10,August,2004','2004-08-10 00:00:00','    wtf...let me get  this straight - there weren\'t politics in Springsteen\'s music before?&nbsp;  Sure, Pink Cadillac and Cover Me might not exactly be subversive, but Prodigal  Son and No Surrender certainly aren\'t apolitical.&nbsp; And Pawlenty would go to  a show with R.E.M. to see Springsteen before Springsteen said he was choosing  Kerry?&nbsp; yeah, right.&nbsp; Bummer - I guess this means I won\'t be able to  get tickets to Toby Keith because he\'ll be taking my seat.         &nbsp;    Minnesota Governor Tim Pawlenty opened his  weekly radio show Friday with \"Born to Run,\" but followed by saying he was  \"heartbroken\" that Springsteen had gone partisan.     \"I really appreciate his music, but I wish he wouldn\'t interject his music  with politics,\" said Pawlenty, cochairman of Bush\'s reelection campaign in  Minnesota. He said he was \"going to have to miss\" the planned Oct. 5 show in St.  Paul where Springsteen and R.E.M. are slated to play.     http://www.boston.com/news/globe/living/articles/2004/08/10/some_fans_dont_want_musicians_to_play_politics/   '),(5979,3883177,'06,August,2004','2004-08-06 00:00:00','    A co-worker&nbsp; &nbsp;told me that Michael  Moore was the new Leni Riefenstahl.&nbsp; And now I hear that it may have been  the writer of the The Day After Tomorrow (below)?&nbsp; What about Robert  Greenwald (Outfoxed) is he the new Leni Riefenstahl too?&nbsp; What about  Nickolas Perry and Harry Thomason (The Hunting of the President) - can they both  be Leni Riefenstahls, or is one of them Leni when she\'s like younger, and the  other one when she\'s older and all wrinkled looking, which is how I remember  her?&nbsp; Some people (Washingon Post) even compare Verhoven (Star Ship  Troopers and Show Girls) to Riefenstahl, and who can argue with that.&nbsp; That  pretty much makes it a good argument to use, because basically, when all is said  and done, you\'re not comparing anyone to Leni Riefenstahl, you\'re just comparing  everyone\'s movie to Show Girls.&nbsp; Those GOP b*stards are cunning! &nbsp;             &nbsp;      Wall Street Journal Compares The Day After Tomorrow director  to Leni Riefenstahl:          http://mediamatters.org/items/200406070001   &nbsp;   '),(5980,3883177,'03,August,2004','2004-08-03 00:00:00','    So, my father in law  spent his time with my fifteen month old daughter teaching her to say \"Bush\" in  retaliation for the picture we took or her leaning against a John Kerry sign in  our yard.&nbsp; This might have been annoying, except that now every time she  sees a fish, like when we\'re reading \"The Cat in the Hat\", she says \"Bush\" - I  can\'t bring myself to explain the difference, and I\'m not sure I could.&nbsp; On  the other hand, we received a nice card from John Kerry for donating to his  campaign, complete with a postcard sized picture of him on the front - my wife  tells me that Eryn found it on the steps yesterday and hauled it all over the  house chanting \"Daddy\".&nbsp; Either my wife is in trouble (and incredibly  sneaky), Eryn has decided my politics are politics she shares, or Eryn has  greatly misconstrued the amount of hair I have - then again, she never gets to  look at it from the top unless it\'s time for a piggy back  ride.   '),(5981,1629718,'31,March,2004','2004-03-31 00:00:00','   Today all in all is a good day, financial wise i\'m doing..stable.. but then again thats what unexpected car repairs will do to your pocket book, literally fark em up bad. I have $7.50 to last me for gas until friday...which i\'m crossing my fingers that will be enough to get to point A and B.  Taking out a Loan to help things out hasn\'t done much of a lick to make life easier..and it seems all the bills are due on the 15th..yet another nightmare... I\'m content and know that in a few weeks, everything will be straightned out...and money, will be at equal amounts, and maybe, I could survive on more better things.  Lena just called, seems I have to pickup my years supply of contacts today, my eyes are watery and it seems like i\'m wearing the ones I have in my eyes out bad, but it\'ll be nice to see clearly , yet , once again. I head to LA on friday, which I\'m not really all that excited about, but hopefully accomplishing one thing through the whole deal, and thats getting myself finalized and GONE from the X. Tired and sick of the thought of even having anything to do with her, more less taking part of her crazy lifstyle of constant bickering and cheap liquor in her front yard.   When I was growing up families had values, and you would do anything to make them work, but for that family, it seems like an endless task to make another $1 off of someone not wanting to be with your cheap, lame, ass. And collecting some more wellfare. I can\'t wait till she has to work, wait...the State will take care of that as well *shrugs*   Hmm dinner tonight sounds good, think i\'m gonna hammer into some butter noodles or spegetti, misspelt, i suck..yes I know. As well as watch some tv, yank some jams outta the plastic box thats my proclaimed life, and sit and compromise about the situations i\'m fixing to face.   I miss Rosie though, seems like the past few times we\'ve gone places we\'ve always had a blast, from Walmart to Taco Bell...it always seems like we\'re up to no good, and I love that! I miss it! :)   Well i\'m finished yet once more...but for how long.... thats something we\'ll never know....  '),(5982,1629718,'31,March,2004','2004-03-31 00:00:00','   Wow kids, take a look at your local gas pump, or shall i say pimp, it seems now its costing you more locally to fill that tank, then it would tankin your girlfriends mother. Why? Well let us see... due to stupid wars in IRAQ the oil industry has something to bitch and complain about, and who ends up paying more at the pump, you do... Today I heard that Oklahoma is taking to the senate a bill that would raise our gas tax more...(are you friggen kiddin me?), and on top of that Oklahoma is footin and approving to tax 50 cents more per pack of smokes.   Lay off the TAXES, and give us a break... while you drive your nice LEXUS and Caddy\'s. My salute to you Mr Bush, I\'d tank ya in the swamp before i\'d cast another vote.  '),(5983,1629718,'31,March,2004','2004-03-31 00:00:00','   Today\'s been pretty good considering the headlines of things going on around us are crazy, lets take for example what\'s going on in the latest of craze when people see a movie and start acting stupid , stating that the movie changed them, therefore it\'s time to confess that I nuked a SOB and goto jail? Wake UP! Let us look @   urlLink Yahoo Oddly Enough News  And there\'s more where that crap came from.   Do yourself a favor..save the popcorn and 10 bucks it costs to see the damn movie and just head straight to the local jail... it works, and it\'s less expensive, after all we all enjoy prision food, and admitting were stupid right after a movie, that of all people, mel gibson (I do remember all his movies such as lethal weapon, etc, included VULGARITY), would produce. Let us think, its not about \'touching others\' for free..it\'s about money, and if mel was so advertly a clearly christan man, he would be donating at least 3/4 to the churches...wait...I\'m yet to see that....  Well more picking on others later... if you have any comments, by all means feel free to email me admin@wickeddreams.net and don\'t forget your movie pass.  '),(5984,1629718,'28,April,2004','2004-04-28 00:00:00','   Seems like today went great, and i\'m happy, me and jessica are talking and things are working out great for katie bug. I finally got my divorice papers and ruled and judged on the things i wanted and fought for in court. Wow...  I\'m now in college, wow..thought i would never do it, *L* crazy but im in, as pre-law, so gear up cory, its time to kick butt, and i can do it. With the support of everyone helping me through life, thank you. I know sometimes you can\'t change the past, but as for the future i still think and thank everyone from the bottom of my heart.  I\'m checking out a few new hotspots, and soon to be moving to OKC, or dallas, depends on which of the directions i want to go.   Well im gonna close, but hey life is good! 4 Years to go for the BAR!!! LOL  lord help me lmao  -W|CKED  '),(5985,1629718,'12,April,2004','2004-04-12 00:00:00','   urlLink Yahoo! News - Good Goin Goofass! - Air Marshal Leaves Gun in Airport Restroom    Stupidity Meter  (5/5) *****  '),(5986,1629718,'06,April,2004','2004-04-06 00:00:00','   I\'m sure we all do, but why not express yourself and send me an email by clicking  urlLink HERE , and tell me what you think about the War in Iraq...one voice is heard, but i\'m always wanting to hear another too. Express your views  urlLink EMAIL  me! And i\'ll post your comments on the blogger... lets find out if were all Hardcore GI Joe junkies...  -W|CKED  admin@wickeddreams.net cory@engineer.com  '),(5987,1629718,'06,April,2004','2004-04-06 00:00:00','   How many times must we go through this, once again, a few more soldiers blown to pieces, yet another family man not being able to spend his last holidays with his family. Is this war really worth it? Or have we made a mistake just like vietnam...well let us clap for Bush, i mean he did catch Sadam but, um..last time i checked the 9/11 attacks were from Osama...thats wild... And don\'t you think its time to pack the crap up and leave? Before you turn around and find out the body of your relative or friend is being burned and dragged a million miles too?   Latest News: More contractors die... go figure, ...i\'m gonna re-elect this one right away!  -W|CKED  '),(5988,1629718,'06,April,2004','2004-04-06 00:00:00','   I won.....everything I wanted , and now I have my daughter back in my life, I saw her yesterday, shes beautiful, we both cried as we held eachother, and finally, I\'m a dad once again... 50/50.. I can\'t wait to make up for lost times, and make her life, light up like it used to be... I\'m no longer empty anymore.....Thank you to everyone for their support and helping me legally get my daughter back, all i can say without wiping a tear is, paperwork speaks, and standing up for yourself before a judge does too... I love you Katie bugg (-your daa daa).  '),(5989,1629718,'01,April,2004','2004-04-01 00:00:00','   Well we\'ve all seen nothing going to nothing again  urlLink See The Headline  And how were these people to defend themselves when they were the objective to rebuild iraq... Go BUSH!  '),(5990,1629718,'30,May,2004','2004-05-30 00:00:00','   Submit Lesson 1 Project    Your project was successfully uploaded.  Your Project ID is: 655948  Your instructor uses the Project ID to reference your submitted assignment. Please print or write down this ID number, as it is not available elsewhere.  '),(5991,1629718,'15,May,2004','2004-05-15 00:00:00','   A given example of why we should all thank goodness that the election year is up, looks like everyones \"favorite\" good boy president is in hot water and looking down on his security advisors and wishing he woulda never gotten in the crap he\'s in.   I\'ll vote for that one my golly!  urlLink Iraqie Abuse Photo\'s Released    Go bushy go bushy!  '),(5992,1629718,'07,May,2004','2004-05-07 00:00:00','   47 Core Requirements          CM102 5 College Composition I           CM220 5 College Composition II           CS110 5 Human Dynamics  Human Dynamics May Term 2004      CS210 2 Career Development Strategies           IT133 5 Software Applications           MM201 5 College Algebra (or higher)           PA253 5 Legal Ethics           100/200 5 Communications           100/200 5 History           100/200 5 Diversity and Culture             5 Open Elective Requirements          100/200 5 Open Elective             38 Major Requirements          PA101 5 Paralegalism Today  Paralegalism Today May Term 2004      PA110 5 Civil Litigation I           PA112 5 Civil Litigation II           PA130 5 Contracts           PA201 5 Legal Research           PA299 3 Associate\'s Capstone in Paralegal Studies           200 10 Paralegal Elective             90 Total Credits   Holy crapola im in for a ride :P hehehe  '),(5993,1629718,'05,May,2004','2004-05-05 00:00:00','   Wohoo enrolled! Funded and ready for battling the paperwork i\'m now a college student! 2 classes a week :) Sure sure it\'s gonna kick my bum, but hey! I\'m excited! WOOHOOOO!!!!! I\'ll be done in 2 years on Phase I, and 2 more I\'ll be barred in family law for Oklahoma & Iowa. Wow..and people said law was trash, maybe so but its exciting...watch me sprint dis :P   W|CKED  '),(5994,1629718,'03,May,2004','2004-05-03 00:00:00','   urlLink i-Mates.de - Games, Screensaver, Screenmates and more Download Goodies - Froggy  '),(5995,1629718,'03,May,2004','2004-05-03 00:00:00','   urlLink Surveillance Scan II is a free webcam surveillance app!  '),(5996,1629718,'03,May,2004','2004-05-03 00:00:00','   urlLink NoTrax - Secure Web Browser that leaves No Evidence of your Internet Activity on your PC  '),(5997,1629718,'27,June,2004','2004-06-27 00:00:00','   Things that make us wonder at times is the past relationships we had. In my experience as a mad sexoligist in relationships, i often wondered was it good to rank head or rank income? Let us investigate with Wicked..come on kids!  The names have been changed of past hoes to protect the people with the names that are not hoes, hell your not a hoe, am i? nah i\'m a manwhore..but if your a hoe cool...lets move on.  Income vs head is a prospecting idea, venture, and yet a standard of living. If your in a relationship and the situation bases income over getting head, don\'t bitch i suppose, but head is nice...but if she aint given is she giving to your neighbor? Let us ask...  Dr Ruth states... if the biatch dont toss dont get mad make her slide that mad puppy where it dont belong... so why is it that the performace of others rank in our minds? Well my x sucked at doing things..hell I remember one time if i asked for head, she pertained the thought to cleaning the bathroom and abused me, punched me in the eye, and made me cry.  Let us play some debbie gibson to clear our minds, okay from hell this study went to hell when chewy my dog decided to go fark some other puppy and run away with gold chains on his neck, freak ROLEX ! Damn the man! DAMNIT! their goes my subject for pyscological questioning... i was going to study him for the replies on the females perspective..ah fark.   which brings me to the next level is michael jackson a male/female or a tranvestite locked in a mans body... he loves good luvin and you love it too... did he abuse you? If so shame shame... lets sue the Riaa..fark em! They fark you! For Killin sakes DONT share the MOVIES! Fuckit!  Okay im off the subject but we dont have one now, see OU im smart, i sit and utilize my time doing useless things, i\'ll fit fine in your programs and twiddle my dong endlessly... hmmm did you poke today?  Let us all get dropped kicked at resturants in ada that are chinese ones that suck, man whip out the fucken nubchucks and make us hide, and give us free cookies... and charge us! Anally abused i swear! someone hide my ass!  Lemme drive ya to the store in my pimp ride that breaks down 1/2 there and we have to push the car the rest of the way, cuz it goes like that, you love me peoples, friggen admit it, ya freaks.  And if i find another sex bar on another pc i\'m gonna go nuts, wait did you go nuts? hell the popups suck but wack wack! you love it dontcha! lemme install 2 more freakit!   Ok this is useless my study showed that I get laid alot and you dont, i watch happy tree friends and i pimp my car, and i dont have credit, so i rule..end of case!  '),(5998,1629718,'24,June,2004','2004-06-24 00:00:00','   Today has gone good so far, seems like Randy didn\'t show up for work, that stinks, but other then that todays been nice. YAY! I\'ll post more later or somethin...lol  '),(5999,1629718,'23,June,2004','2004-06-23 00:00:00','   Sure we\'ve all had our times for wondering how the hell to get out of bed..but did you know that Amp Energy Drink from Mt Dew is the answer to all your hellish waking up days? If so don\'t be fooled, there\'s a drink even higher up on the tot\'em pole then that for sending your energy to another planet, hell even my fiance\' thinks so.  I would have to rank a few drinks on the market in this order:  1 - Monster Energy (a pounding hellish 2000+ MG of caffine per drink, not more then 3 a day recommended) ***** 5/5 -- Made me wanna climb a friggen wall, then a moutain, then the tv, and then go fix my car when I know nothing about mechanics - Cost $2.20 for 40 Oz Can (Woohoo!)  2 - MT Dew Amp - **** 5/5 - Peaceful Luv Energy Drink - Sure a cheap chip off of monster energy once again, but it does the job, and it\'s 20 cents less! , still makes me do nutty things, hmmm wonder if this is addictive? Cost = $1.99 for 8.4 Oz Can (Rippoff?)  3 - Pure Luv Mini - **** 4/5 - Mini size of pure love, you can score two for like $1.99, not a bad deal, better hose down both of em they seem to sweeten the kick, sadly these had only  75 mg of caffine in each can :(   4 - Red Bull - ** 2/5 - Wtf is this junk? This can\'t even spark my mind, nor send me racing through a cardboard box, I\'v had more energy nibblin on friggen sour patchies with my fiance at midnight half awake. This stuff should be labeled - Kiddie Sooper Shooter Drink and given at McDonald\'s in their kiddie meals. And for $2.XX !! It\'s not worth it. Cost = $1.99 8.4 Oz Can.  Well that concludes my day of rantin on about energy drinks...maybe i\'ll post more later. Maybe I won\'t ..ah what the hell...  W|CKED  '),(6000,1629718,'12,June,2004','2004-06-12 00:00:00','   Gila Monster: mad dog is a penis burglar maddog642000: Get bent loser maddog642000: But your a fudge packer Gila Monster: ok let me hump your face over a stool then Gila Monster: do you have enough hair on the back of your head to wrap around my hand at least once? okbuilder2000: lmao redheadokla: hmmmm i do have alot of hair Gila Monster: can you give a worthy blowjob? redheadokla: i can suck the crome off a bumper Gila Monster: come suck my crome bumped then woman W|CKED-DREAMZ: she claims she got mad chrome skillz W|CKED-DREAMZ: and can define the meaning of D.S.L W|CKED-DREAMZ: haha W|CKED-DREAMZ: Santan\'s such a pimp W|CKED-DREAMZ: Er Santana val418: britney spears  W|CKED-DREAMZ: hey hey now dont juke my jams val !   val418: lol val418: did you get somethin good wicked W|CKED-DREAMZ: yah i got pimpin music val418: out my nose? W|CKED-DREAMZ: yup haha )))***BIGS***(((: all i want for Christmas is a cheese pizza wif natalie Portman ass juice Brittany*~!: ass juice?  Brittany*~!: SIck! ***ENDING OF LOGGING*** W|CKED-LOGGER.MODE3.--CMD SUCCESSFULL-- UNIX REMOTE DISCONNECTED:IP.KILL.SIGNAL.ACCEPTED   /ROOT /CONFIG?A:/sND/ETC/LOG1.TXT *******SENDING TO WWW.BLOGGER.COM*** *******ACCEPTED******* **POSTING**  VNC TERMINAL READY. SUCCESS.  '),(6001,1629718,'10,June,2004','2004-06-10 00:00:00','   Wow what a day, had to fix a PC and drop by EYE-Q and get some phat pay for devoting a full 1/2 hour *L* I love this area, seems like everyones in need of a full fleged computer geek 24/7. Work is going great, classes are going smoothly and i\'m ace\'n the friggen college thing. For once i\'m smart again! Hahaha.   Tonight i\'m kickin back, relaxing and fixing to just enjoy some tv and writing my assignments. Exciting eh? Rosie came down yesterday and me and her had a wonderful time eating at Bandana\'s, gigglin , laughing, and she even had breakfast when she woke up! :O Wow i got groceries *L*   Paid my cable bill, friggen high as hell 130 is to much for cable, but oh well, least it in cludes business cable speeds and i\'m powered up enough to power like 40 pc\'s here.   Finally got me a laptop, for on the road and on trips, its quite tight. And oh man Corpus Christi was beautiful! I love traveling on business trips.   Well nuff blabbing from me... ZAP! Nugga!  '),(6002,1629718,'30,July,2004','2004-07-30 00:00:00','   Today has gone great, though I plan on cleaning the house tonight all in all the day has gone good.  Last night was different, for once I actually went to bed at a decent hour and woke up at 2:30 in the morning, ate some chicken fettucini and then went back to bed, waking up just in time to get my butt to work. LOL. How close things are sometimes scares me, but this week has drug by the toes and it\'s nice to have the weekend off.  I do miss having one day here a week, then one here off as well. But dog gone it, it\'s not like I accomplish anything on the 2 days im off for the weekend, more less the 5 days that I work. LOL I love the car, its so pretty, blue like the sky, its my baby! Finally nice to have something new, and not be Poor! LOL like I would ever be anyways, even if I didn\'t lift a finger, it always seems like money comes to me from friendliness and computers *L*  Summertime is almost over, look at this crap, and i havent even gone swimming but once, or bbq\'d outside but maybe 2 times since the summer.  What a crapper. I think im gonna spend like a weekend and go camping, reguardless of who comes or not, I havent spent anytime doin anything outdoors this summer (insert boo hoo here) *L*  Welp i\'m out ...more blogging later i suppose lol  '),(6003,1629718,'29,July,2004','2004-07-29 00:00:00','   urlLink Skype  '),(6004,1629718,'29,July,2004','2004-07-29 00:00:00','   urlLink BBS Corner - BBS Software  '),(6005,1629718,'26,July,2004','2004-07-26 00:00:00','   I hate the bastard, OK your NOT FUNNY that\'s&nbsp; the reason why your now doing AT&T commercials, because nobody dials down the friggen middle, everyone has a damn cheaper phonecard then your orange haired yellow bellied bastard of a child ass..DIE CARROT TOP! your comedy did why can\'t you??? *kix the tv*   I feel better..  WCKED  '),(6006,1629718,'25,July,2004','2004-07-25 00:00:00','   Finally got all of my school finals done Whew that sucked. Woke up early today and realized I need to go grocery shopping and clean the house, its horrible (eww). Even my fish is like about to roll belly up if i don\'t clean the tank. Der i was wondering why my desk was so wobbly last night and found out it was that i kicked the retention wire thingy out of the bracket. I put it back and it was fine..wohoo.   Im starvin starvin uch... one more hour till lunch, lol mt dew just don\'t fill ya up no more *shakes can* sheesh the nerve of these people. Last night i was freezing damn , i flipped the air on 65 and this morning i was so cold i couldnt move to the thrermostat..crazy!   Well i\'m out for now, i\'ma gettin phunnie lewks.   &nbsp;   urlLink    &nbsp;  WCKED  '),(6007,1629718,'25,July,2004','2004-07-25 00:00:00','   Friggen about time, i decided to get yet another car... threw the old one back and got me a NEW 2004 Nissan Sentra HELL YAH! WOOHOOO!!!!!!!  I\'m happy and off to bed, laters pimps!   WCKED  '),(6008,1629718,'22,July,2004','2004-07-22 00:00:00','   urlLink'),(6009,1629718,'19,July,2004','2004-07-19 00:00:00','   Ah,  &nbsp; So my bordism has kicked in, but hey here\'s a few cool urls.   What OS Are You? -  urlLink http://www.bbspot.com/News/2003/01/os_quiz.php   I turned out to be OS2/Warp Ewww!  What Does Your Phone Number Spell? -  urlLink http://www.phonespelling.com/   Mine was a buncha mumbo jumbo..better luck net time I suppose.  Pop some bubble wrap! -  urlLink http://www.snapbubbles.com/   And you call me insane??  Pick a Quiz oh what the hell! Pick one! -  urlLink http://www.ecrush.com/lovequizzes/index.phtml?trip=261&amp;cobrand=bored   Ok thats enough! I\'m going home damnit! DAMNIT! arg..i hated missing a day of work this week, literally blows arse, low pay sux!   Till later ya cheap bastards!  &nbsp; -WCKED  &nbsp;  &nbsp;  &nbsp;  '),(6010,1629718,'19,July,2004','2004-07-19 00:00:00','   Well kids, here i am blogging uselessly when i should be doing schoolwork, however I find myself more intriged by looking at my blogs then anything else. Mom and dad are visiting either today or tommorrow,&nbsp; and staying for a day. Reason? Antique shops and a lil lookin around. I think i\'ll take em to golden corral for dinner, and they better like it for 10 bux a piece damnit lol.  Arg i gotta get my licence renewed, that sucks... i kinda liked the old way they looked, now im gonna have a bold, awsome, over 21 look on my licence,&nbsp; lol. Beer beer, yes kids, i love beer too. Bah!  Randy went to lunch before me today , damn oh well, i\'m here till 6 anyways...heck of alot more refreshing when you get a late lunch from like 1:30 - 2:30 then 12-1 , to much damn traffic of first things, and for the rest , to much heat... lol  Hmmm what to eat for lunch, i think today it\'ll be Carl\'s Jr\'s or maybe some burger king, lawd knows! I\'m kinda lavish for the taste of the chicken whopper, its just so damn good, but for 5 bucks a meal, it seems to be gettin a lil expensive. Yup br.king it is, i\'ll read the paper too...whit whew how i rock lol.  Well im off peace!  WCKED  &nbsp;  &nbsp;  '),(6011,1629718,'19,July,2004','2004-07-19 00:00:00','   Today is starting out well, I woke up early, grabbed a soda, put on the new clothes rosie bought for me, and man do I look Rico Swave\'! Whit whew! I\'m excited about this weekend wooohooo!!!!! I can\'t wait I can\'t wait. Last weekend was cool looking at cars, however the Mustangs seem way to expensive, uchkies!  I feel way overdressed for work, heck maybe I am maybe I\'m not, never know lol. Well i\'m out for now, maybe i\'ll add more to this in a lil while. The boss is back today, thats comforting but it still doesn\'t make the business any slower like it needs ta be der!   WCKED  '),(6012,1629718,'19,July,2004','2004-07-19 00:00:00','   When you get caught in the rain with no where to run  When you\'re distraught and in pain without anyone  When you keep crying out to be saved  But nobody comes and you feel so far away  That you just can\'t find your way home  You can get there alone  It\'s okay, what you say is  I can make it through the rain  I can stand up once again on my own  And I know that I\'m strong enough to mend  And every time I feel afraid I hold tighter to my faith  And I live one more day and I make it through the rain  And if you keep falling down don\'t you dare give in  You will arise safe and sound, so keep pressing on steadfastly  And you\'ll find what you need to prevail  What you say is   I can make it through the rain  I can stand up once again on my own  And I know that I\'m strong enough to mend  And every time I feel afraid I hold tighter to my faith  And I live one more day and I make it through the rain   And when the rain blows, as shadows grow close don\'t be afraid  There\'s nothing you can\'t face  And should they tell you you\'ll never pull through  Don\'t hesitate, stand tall and say   I can make it through the rain  I can stand up once again on my own  And I know that I\'m strong enough to mend  And every time I feel afraid I hold tighter to my faith  And I live one more day and I make it through the rain   I can make it through the rain  And I live once again  And I live one more day  And I can make it through the rain  (Yes you can)  You will make it through the rain  '),(6013,1629718,'17,July,2004','2004-07-17 00:00:00','   Wohooo here comes a new car oh yah! oh YAH! *does a dance* i think im gonna get a red one, oh wait er a black, oh HAHAHA i dunno! Hell YAH! I\'m excited!!!  &nbsp; WCKED  '),(6014,1629718,'15,July,2004','2004-07-15 00:00:00','   urlLink'),(6015,1629718,'14,July,2004','2004-07-14 00:00:00','   Ok this post didn\'t make much sence, then neither does my odd sence of humor at times. I dunno why i\'m blogging, hell I dunno why I\'m doing anything, I\'m lost, that sucks. Someone build me some lego towers so i can jump off em. Er.. anyways I think im gonna go out to eat for lunch today, and maybe get my haircut sounds cool arg..Captain!   -W|CKED  '),(6016,1629718,'14,July,2004','2004-07-14 00:00:00','   urlLink Welcome to the 419 Eater - Where We Scam Scammers!  '),(6017,1629718,'07,July,2004','2004-07-07 00:00:00','   See the story now!   http://story.news.yahoo.com/news?tmpl=story&cid=638&e=1&u=/nm/20040714/en_nm/leisure_subway_dc  Not only has Subway decided to bring the old fat bastard back that for some reason he could afford 200 dollars in sandwitches a day, he\'s now tappin on fat kids! Yes! You will see more of Subway commercials singing \"your kids a fat bastard, be like jerrod! eat fresh!\"   For someone that should jump off a bridge, I\'d throw him right next to my xwife.  -W|CKED  '),(6018,1629718,'07,July,2004','2004-07-07 00:00:00','   urlLink Cyber Tech Help Support Forums - Help please: is this a video card problem?  '),(6019,1629718,'07,July,2004','2004-07-07 00:00:00','   urlLink System Downloads : Windows XP Video Screensaver /// AnalogX  '),(6020,1629718,'05,August,2004','2004-08-05 00:00:00','   Fark BUSH! He Blows My NaDS!       urlLink     Another joyful addition to my useless sack of ramblin crap on this earth of oh god loviness for the fine thursday of hell almighty! Since it\'s only 4:16 here, I have a few more hours before grabbin my bag of stinkin ass laundry and hope that it cost less then a cheap hooker on a nyc street side.   After all you can only turn your pants inside out 2 times right? or will anyone justify that you can do it another way...  Today I heard the suggestion from a college student here in ADA that when they can\'t afford laundry they do it in thier bathtub? What the hell is this crap? How do you dry it I kindly asked ... Oh easy, you put it on the cieling fan and flip in on high , thats after you secure it with clothespins..wow this cheapskate really took some though into this, maybe i should save him the hellish view of living life as it is and bash him with a glass picture of koolaid on his forehead. After all the is Ada right?  So here goes my fanmail i\'v saved up for a few months, yes todays the day to unveil why not couldn\'t be a better day!    urlLink     Dear Wicked,    After reviewing your website I noticed you put alot of self gradification into how good you are, what about all the rest of us? Why do you take credit for the design of the website when it\'s truly the work of someone else? And why are you tapping on poor people when you know your one as well?  -Jack ( urlLink Jack1@xx.com )  Dear Jack,   After reviewing your email it reminds me of how much better I am at self gradification and how much your religious view of how people are classified as poor and inexperienced relate to beer drinking / guzzlin hobo\'s kicked around by us better people on New York City Streets.  We all need to talk highly about ourselves, even though in due time it seems that you are 10, you will grow up, and as for ripping off websites, sure copyright it, but even elvis can\'t keep free people from sucking his funds dry for old lp\'s. Move on, and why your at it get off comcast, they suck.  -  Dear Wicked,   I thought you cared alot about your past? What happened to your daughter you loved so much and tried for? What about now?  -B (referred by your xwife)    urlLink    Dear B,    I surely how do honoribly shall bow before you and the rest of the world to express the flaws in the judicial system, and how much paperwork can do if someone gets off their fat ass and quits flying trips back and forth to california, one could afford a reasonible meal for the family. Funny as it seems, I do love my daughter, I send her things, and yet through many months of sending items , I haven\'t YET to recieve a colored picture of barney, more less a audio cassette tape or a video. So maybe one wouldn\'t be poor if one got a job? Oh wait! That\'ll never happen! Best wishes! I\'ll be playing my Xbox and watching my big screen tv thinking of you! As for my daughter, I\'ll continue to send things, as for the things i\'ll ever get back like a normal parent should, i\'m still waiting. As for trips it seems like i have to do the 100%. Seems like someone doesn\'t own a car as well. Have a beer on me! Buddy!  -WCKED   urlLink     More to come....  '),(6021,1629718,'03,August,2004','2004-08-03 00:00:00','   Seems like nothing works at times, here i am barely enough money to do laundry, i\'m gonna get a load done so i have some clothes to wear to work tommorrow and something to pack when i goto oklahoma city for the weekend. It just sucks, i\'m friggen broke and i have $10 to go travel on, wow..thats gonna make it far there willy. Just seems like no matter how much i overwork, or try harder the less i stinkin have at the end of the week.  I suppose i should get off my crapper and do something about it, but right now at the time i feel tired after work, and i think its BS to have to work 2 jobs to have another 40 per week if needed, hell maybe i need a roomate, maybe i can find one in the city if i move, till then i\'m stuck here in Ada, with like $10 a week to live off of, hardly anything to do, and a suck ass view of life.  The more and more I look into moving, the more inconvient it seems to my so called \'bs contract\' with this place, and moving back in with my parents is out, as well as my sister to secure a job in okc. Fark i just feel like im doing this all alone, oh yah i am.  Broke sucks.. i think i\'ll make some koolaid and hope that someday equality in a roomate will help me out, till then , since i have 3 months left, i might look into options of another job, less time with people, ah love that idea.  More to bitch about later  =WCKED  '),(6022,1629718,'03,August,2004','2004-08-03 00:00:00','   Todays Wonderful HeadLines Brought II You By'),(6023,3672170,'17,June,2004','2004-06-17 00:00:00','   Your promises cause pain, I can\'t stand the pain, you send me dancing in the rain, yet your love I can\'t gain, always running with a chain, your treatment is insane, in pain I rarely complain, do I only serve to entertain, always drinking champagne, I must get away from the pain, as this relationship was all in vain.  The End. '),(6024,3672170,'17,June,2004','2004-06-17 00:00:00','   Heaven is a whisper away, don\'t wait till the final day, hear all that Jesus has to say, He is the only way.  The End. '),(6025,3672170,'17,June,2004','2004-06-17 00:00:00','   It wasn\'t long ago, when we first met, how I\'ll never forget, your looks stirred passion, always the look of fashion. '),(6026,3672170,'17,June,2004','2004-06-17 00:00:00','   No matter what you say, you are not here to stay. We will be going away, as it is His way. Life is not always so gay, don\'t like your life decay, accept His way.  The End. '),(6027,3672170,'17,June,2004','2004-06-17 00:00:00','   I see the pain in your eyes, your touch makes me cry. I wish I could heal you with my touch, my love for you grows so much. Out time together grows shorter every day, I\'m afraid I\'ll not know what to say. My love for you will never stray, our love will carry us far, far away.  The End. '),(6028,3672170,'17,June,2004','2004-06-17 00:00:00','   I can\'t believe I have been out drinking all day, now I have to figure out what to say. Should I start with \"it started this way\", no, I used that last time. Maybe I should explain it in a rhyme, no, that\'s not worth a dime. Who am I kidding she won\'t let me in anyway, So I\'ll go back to the bar and drink the rest of my pay!  The End. '),(6029,3672170,'17,June,2004','2004-06-17 00:00:00','   The door is locked, curtains are drawn, lights are out, no one will see, no one will know, it\'s just you and me, get ready for the show, but first take it slow, for this could only be a moment.  The End. '),(6030,3672170,'17,June,2004','2004-06-17 00:00:00','   As I search for answers in my mind, I realize I am always left behind. Many doors close leaving me alone, even leaving me without a phone. There is no time to feel sorrow, I can only hope for a better tomorrow.  The End. '),(6031,3672170,'17,June,2004','2004-06-17 00:00:00','   There are many rules in life, they cause us great strife. Are rules a necessary part of life, I guess so just ask my wife!  The End. '),(6032,3672170,'17,June,2004','2004-06-17 00:00:00','   Why Lord, oh why, we sit and only ask why. We should know the answers don\'t just fall from the sky. Just because the sky is blue, should we be too? No we don\'t know it all, so always have faith before you make that call.  The End. '),(6033,3672170,'17,June,2004','2004-06-17 00:00:00','   My how the moon shines brightly tonight, it lights up a dark room, tossing about a mysterious glow, silent fingers of light seem to creep across the room, but there is one object that catches my eye, her lower leg quietly rests along side the sheet, like a caressing touch the moonlight rolls up, towards her inner thigh, with a slow sigh and exhale the sheet raises, quickly the moon light races towards her lower back, exposing the gentle curves of her petite frame, within a blink the clouds cover the light, and now it is time to say good night.  The End. '),(6034,3672170,'17,June,2004','2004-06-17 00:00:00','   Watching rainfall reminds me of tears,  sometimes falling gently, sometimes falling hard, tears and rain fall for many reasons, rain can signal a change of seasons, a little rain falls in everyone\'s life, tears show emotions in time of strife, like rain that always finds away, let the tears fall where they may,  both signaling everything will be ok.  The End. '),(6035,3672170,'17,June,2004','2004-06-17 00:00:00','   It is just another day, one that resembles yesterday. I keep waiting for a change, but the waiting makes me feel strange. Maybe the window above this cage, will disengage. Could this be the day I fly away, or should I just stay? This feeling is turning to fear, maybe I don\'t want to go up into the atmosphere. I could just persevere, for tomorrow will be just another day anyway.  The End. '),(6036,3672170,'01,July,2004','2004-07-01 00:00:00','   It is that time of year again, where all can pretend, to be who we desire, not who we are, as mystery fills the air, everyone\'s let down their hair, some go for scare, looking like a nightmare, others just weren\'t sure what to wear, and did little to prepare, they surely aren\'t millionaires, but no matter the outer wear, their lives are still in despair, as the day ends and the return to solitaire.  The End. '),(6037,734023,'03,June,2004','2004-06-03 00:00:00','   Boredo...     Tired? Grab a blogger...    Yep. I am SO freakin\' bored it ain\'t funny. I mean, I need something to do really really bad. My sis wants to use the comp so I can\'t type much. I think. But, anyway, I\'m gonna be going to the store w/ my mom so I can\'t really type much. So, I need to take a shower. Isn\'t that grosse? It\'s almost 6:00 p.m. and I haven\'t taken a shower yet...(ok...maybe I shudn\'t hav sed that). But, I dunno. I called my buddy P-bear (the nick my friend Alescia made 4 him) and he wudn\'t home. Ya know, I miss skool. At least then, I actually had something to do. Ugh. Now I\'m just sittin\' next to this freaking 1984 edition comp that\'s about to break down. I need a life. I NEED some friends. *sigh I NEED 2 take a shower. For god sakes, I\'m a GIRL! hehe. Well, I wonder if my P-bear called me back. Most likely, he didn\'t. He was out playin\' basketball w/ his cousins. *sigh...everything is gonna be A-O-K (you have 2 read that in letter form 2 get it)         So...gotta go...     Yes...I\'m going to the store with my sis and mom. UGH! Who likes going to the grocery store w/ family? Not me. I\'m tired and cicadas are swarming in my tree in front of the house. I\'ll try to blog lata. Otha than that...I\'m out. Peace 2 the Mid-East. That is, if u live in the Mid-East...uhh...yea.  '),(6038,734023,'03,June,2004','2004-06-03 00:00:00','   Summer fun!  Yes...yes...it\'s about time isn\'t it? Haha. Well, I\'m actually really bored as you can entirely see. It\'s 2:00 in the morning and I felt the need to uncover my bloggy from it\'s ruins and ya know...type. Type, type, type. I think I\'ve had enough of it. I mean, school is enough. Sheesh.    Introduction    I feel the need to introduce myself. Well, since many of you out there in the world don\'t know me. I\'m 13 years old and my nickname is Nichi (nee-chee). You can just call me that for now since you really don\'t need to know my real name. I live in Ohio and yes, it\'s the year of the cicada and they\'re going strong. Now, they\'re asleep. I can\'t hear any of them. My birthday is July 17, 1990 which means I\'m turning 14 next month! Yay! (it\'s really not all that) My sn for aim is  lilnichi717  so you can im me sumtime if you\'d like. I have a website and it\'s called  urlLink Kawaii Kutie . So, please go there sumtime. Right now, I\'m on hiatus but, you can keep posting back with it. Um...if you know html, photoshop, php, etc. I need you. I mean, I need as many ppl as I can get to help me w/ my site. But, I just don\'t know where I can go to get help. So, if you think you can help please aim me.    Scary...     Ok...it\'s really quiet down here. So, I can hear footsteps and things everywhere. I am SO PARANOID right now. I mean, I keep thinking somebody\'s going to rampage through my house and hurt me. Or something...me afraid...me VERY afraid. But, since I have you I\'m not scared anymore. [If ANYTHING happens to me...they did it. Also, I ain\'t never scared. lol ^-^] Well, I\'m gettin\' sleepy (and scared) so I\'m gonna go to sleep. I\'m gonna be gettin\' a new layout for my blog sumday. I dun really like these. I luv ya ppl!  '),(6039,734023,'09,June,2004','2004-06-09 00:00:00','   Music...     Srry bout the non-bloggin\' me...    Hey everybody! I\'m sorry I haven\'t been bloggin\' lately. It\'s just, I really need this book called \"Girl in Hyacinth Blue\" by Susan Vreeland. It\'s my 9th grade summer reading book and guess what? I don\'t have it. And if I don\'t get it soon then I\'m gonna be in scads of trouble...ugh! Somebody please help me! *sigh...poor nichi.     Kawaii Kutie     Ok. I need to pick a layout for  urlLink kawaiikutie.tk . I mean, I have so many possible layouts. Then, the html and the rest of the crap that I need to do for my site to be really good. So, yes, I have been working on it. I would put up some pics showing the so-so layouts I\'ve made but I\'m kinder lazy. hehe. But, we\'ll just see what happens. Ya never know. So...I gotta go cause\' \"Last Comic Standing\" is coming on in a few. I absolutely LOVE that show! It\'s literally hilarious! Love ya peoples! (O.o)  '),(6040,3860353,'25,July,2004','2004-07-25 00:00:00','   Today, I met up with my friend at our usual spot, the neighborhood Starbucks.  She had just come back from running a bunch of errands and cleaning house for her mother.  She was frustrated.  Ultimately, because at the age of 24, she is more financially dependent on her parents than when she was living with them about 6 years ago.  The thing is that she\'s finally decided what she wants to do, and, her parents agreed to pay for everything, so long as she agrees not to work.  Who wouldn\'t pass that up, right?  Well, there is a price, even in that.  For instance, today, I accompanied my friend and her mother to the store.  I was somewhat of a mediator.  You see, since her parents are paying for everything, shopping for interview clothes required my friend\'s mother\'s approval for each outfit selection.  So, rather than accompanying her daughter to the dressing room to view the items she had put on hold, my friend\'s mother made her own outfit selections for her daughter.  This posed a problem, not only because, she was completely ignoring my friend\'s need for individuality but because the clothes she was picking out were designed to appeal to much older women (like, her age).  Not to mention the fact that everything she picked out consisted of a paisley floral print, elastic bands, or enormous shoulder pads; talk about a fashion nightmare!  And, who do you think was expected to support her positions?  Me, of course.  I did my best to both compliment the mother for her taste (ugh), all awhile, sway her decisions in order to spare my friend\'s reputation over wardrobe.  I had to get creative, and it seemed as though I was grabbing for words at times but somehow, I managed to save the day, without providing a forum for argument.  All I know is , tapered pants are not flattering, period-neither are they\'re baloon shaped-upper pant counterparts; shoulder pads are hideous, maybe disgusting even (unless someone is just completely without) -they remind me of Milly Vanilly, the bad part of the eighties, and women trying to be men; Then, there is elastic and paisley prints -enough said.  Its just wrong. '),(6041,3860353,'19,July,2004','2004-07-19 00:00:00','   Thats right, today my bestfriend told me that she has to build a mold of her own face to restore.  In addition, the face needs to exhibit definite trauma -in other words, a tragic death.  Apparently, the molds look pretty real, especially after applying all of those restorative materials.  I just can\'t imagine having to look at myself in that way -talk about messing with your head. Even my morbidly-preoccupied bestfriend found that disturbing.  Apparently, the school keeps restored models of their past students faces; and, as you walk through the hallways, they peer out at you. '),(6042,3860353,'11,July,2004','2004-07-11 00:00:00','   My best friend recently decided upon an occupation -turns out, she\'s going to become a mortician.  Her classes start in August, and, although I\'ve never known anyone to take up such a career, I can\'t say that I would expect her to be so excited.  She has decided to completely immerce herself in the culture.  She began by purchasing the first two seasons of Six Feet Under- which, is good for me; since, its one of my favorite shows.  There is also this other show called Family Plots that she is determined to get her hands on.  Okay, so imagine this: we\'re sitting in this Vietnamese restaurant eating round steak soup when my best friend blurts out; Hey, do you wanna go watch an embalming with me (really it\'s just a video that she has to watch for the school)?  Keep in mind that her voice is as loud as someone who has just discovered that they won the lottery. Now, her goal is to become the HOT mortician.  A couple of weeks ago, we were hanging out and drinking at Friday\'s when this group of older men from Bangledesh approached us.  They started asking us a bunch of questions about what we do and so on when bestfriend proudly proclaims that she is becoming a mortician.  Next thing we knew, the men were backing off.  Well, what do you know, there are some benefits of having an undertaker for a friend.  Although, we haven\'t exactly tested it on guys our own age.  Her biggest fear is that her newly chosen profession will not offer any contributions to her love life. '),(6043,3860353,'10,July,2004','2004-07-10 00:00:00','   Okay, so about 6 years ago, I met this Irish Rugby player.  His name was David Cantley.  Although, I specifically remember him telling me that his name was Piette -the name of some famous rugby, player.  Anyway, I was dancing with my sister at this club called Red Jacket.  It was retro night and, that tune from Night Rider was playing when I noticed the group of young men wearing white polos; on the breast was stitched a green four-leafed clover, surrounded by an orange ribbon-like scroll.   Being that I was completely infactuated with the Irish, I decided to move my dancing toward them.  I began scanning them over when my eyes spotted David.  He was both adorable and handsome, with mesmorizing eyes.  I approached him and asked where he was from.  As we began dancing, I could not help but notice his scent, like some fresh, nautical cologne that sent forth waves of nostalgia, along with 100% pheromone enforcements-thats how good it was!  Next thing I knew, he was asking me for a smooch.  At first I said, no.  Then, I gave him a kiss on the forehead and smiled directly at him.  He begged and begged for real one, so, I gave him a kiss on his cheek, next came an eskimo kiss.  By then, even I couldn\'t resist.  All I can say is, it was the best, most enjoyable kiss, ever.  I think he might have even held my face.  Shortly after that, my sister was motioning for our departure.  The next day, I could not stop thinking about him.  My sister mentioned something of the experience to my parents and, the next day my father arrived home with a pamplet entitled Ullster Rugby.  It turns out that his friend at work was a big-time rugby player himself and had played the Irish Rugby team only a few days before.  That is when I discovered his picture, true identity, age, and team name.  Although he had lied about his name, I had sort of dashed out of the club.  So, when I came into the news online that his team would be playing one last game nearby, I figured it might be worth the drive.  Once I got there, it took a little social maneuvering through his teammates to get to the center of the huddle, where, at last he was, chugging anxiously from a large bottle of water and sweating profusely.  He was surprised that I was able to find him and showed up to the game.  I remember saying something silly like, my father has lots of connections.  Anyway, it wasn\'t a dream come true or anything.  At first, he acted like he didn\'t remember me; I don\'t know if he was just trying to be cool in front of his friends or if he was really that lame but he finally came to remember and was, I\'d say, understandebly shocked.  I did stay to watch the game and when it was over, I think he invited me to come to their hotel.  I made some excuse up not to put myself in that situation.  I also remember him offering to exchange his rugby shirt for my spaghetti strap shirt but only if I agreed to exchange the shirts in this outhouse with him.  Yeah right, I think I said.  Well, there was a bit of sarcasm in it all but lets face it, the boy was a horny little shit. The next moment, his charming side shone through and his friends let us alone to enjoy one last kiss goodbye.  And, that one was just as good as the first -enough said.  Well, it may not seem like much, but I could not stop thinking about this guy for months and, obviously it still stands out among my romantic experiences and relationships up to about six years post occurrence.  Just today, I was reminded of him while looking through some Irish memorabilia and decided to write this.  I decided to try and look him up on the web -just out of curiosity and, I did stumble across a recent photo of him on a rugby site for Northern Ireland.  I thought about trying to contact him to see if he would remember.  Well, I\'m sure he\'s had many of these experience with his rugby tours.  I still think it would be funny to write something like:  You know, I bet your a real shit but you sure do kiss well -thats the only reason I writing you...by the way, will you be coming back here anytime soon?  I\'d hate for you to miss out on that smooch:).   '),(6044,3860353,'28,July,2004','2004-07-28 00:00:00','   So, I\'ve got a second interview tommorrow.  I feel pretty good about it but there is thus other job that I had my eye on.  The thing is, I really kind of like both of them but this other job would allow me to do more traveling.  It would also allow me to relocate to some pretty cool places if I chose to.  Ok, so I went to college, graduated, took a job (one of those corporate smorporate things) -couldn\'t stand it, became interested in a new career field, went back to school, finished, and, am now waiting to become licensed as an esthetician.  But somewhere between the time it took to receive my test date and the actual wait, I became not so enthused about my new career choice.  I don\'t know if it\'s that or the fact that while I was in training, I experienced an array of clients who literally sucked the energy right out of me.  Now, I do want to help others, and consider myself to be a fairly generous person but even I, require a certain amount of space.  I just felt like these people were right in my face, and I guess they kind of were.  Yes, I know there is an adjustment period, time to settle in, time to set the balance with every new career and I do plan to try it out -even if its just part-time for extra money.  The thing is, I want to start building on something.  I want to know that that is what I am supposed to do.  Does anyone ever really know what they are \"supposed\" to do?  I mean, for the longest time I thought I was supposed to become a doctor.  Everything in me said that was what I was meant to do but I didn\'t.  The thought hasn\'t left me, though; one thing at a time, I guess.  I know what I want to do though, what I desire.  I really feel like I am supposed to do those things.  I\'m just not sure what those things have to do with my career.  So, back to the job thing, that\'s what it will be for now -is a job.  Or...you never know -gotta start somewhere.  '),(6045,3860353,'10,July,2004','2004-07-10 00:00:00','   Well, I guess the title makes it fairly obvious.  I\'m in search of a cool career.  I\'ve applied at several places and have had 1 interview so far.  I want to do something creative.  I have a natural, keen sense for fashion and I\'ve always entertained the idea of pursuing a career in fashion -or, should I say that the idea has always entertained me.  Yes, I am a party-girl and I live for excitement.  I want to have a fun job; is that completely unrealistic?  All I know is that I am terrified of working in another corporate setting where rows of cubicles stretch around corners to meet each other and fluorescent lighting provides the only shield against the freezer-like climate.  I have searched high and low, across the internet and in through the doors of my local career center and, have yet make a solid career connection.  Some seminar on \"how to market yourself\" is being held in town this week; I guess I\'ll go to that.  So far I have applied at The Cooper Aerobics Center, Renaissance Laser Hair Removal Centers, MAC cosmetics at Galleria, North Park, and Willowbend, the Spa at Vista Ridge, Some spa in North Dallas, Peace Corps, and even, the Air Force.  I have also applied several places online and have registered for work on several websites including the Texas Workforce site. Since I am waiting to become licensed as an esthetician (which is taking forever), I\'ve made a list of jobs/careers that I see fit for myself at present: Public Relations, Event-Planning, Marketing displays, National or Local Network Intern, National Newspaper Columnist or Intern.  I have a few alternate yet less equitable job ideas.  For instance, I have thought about working at a Starbucks and, there is this really cool, local vineyard that needs help- how romantic is that?  I could just see myself with the wide, brimmed hat, picking fruit from the vines and stomping grapes.  Who knows, maybe they will become really successful and make me a partner.  I could have my own wine; how great would that be?  But, these are just ideas, and I have so many of them.  I need to travel.  I want to create networks of friends and connections.  I want to study other cultures -both wealthy and impoverished.  I want to their languages and teach them all that I know.  And, at the end of each day, I\'d write an ethnographic account of my daily experiences -perhaps, I\'d use a use a blog.  '),(6046,3860353,'29,July,2004','2004-07-29 00:00:00','   rurban- adj. Combining aspects of both rural and urban or suburban life. â€”rurbanite n. â€”rurbanism n  '),(6047,3860353,'28,July,2004','2004-07-28 00:00:00','   Living bandage-  n. A bandage or dressing made from skin cells, particularly cells cultured from a sample of the patient\'s skin.  '),(6048,3860353,'26,July,2004','2004-07-26 00:00:00','   Velocitize- (vuh.LAWS.i.tyz) v. To cause a person to become used to a fast speed. â€”velocitization n.  '),(6049,3860353,'26,July,2004','2004-07-26 00:00:00','   Slackademic- \"A perpetual student who prefers the safety and comfort of academic life over the trials and tribulations of the real world.  Encryptlish-  Encrypted english. Correspondence from a non-native English speaker that, although well-intentioned, makes creative use of English grammar and spelling.\"  -www.wordspy.com  '),(6050,3860353,'23,July,2004','2004-07-23 00:00:00','   Pillow Book: \"A list of elegant things that makes one\'s heart beat faster.\"  -unknown  '),(6051,3860353,'14,July,2004','2004-07-14 00:00:00','   Gamine- 1. An often homeless girl who roams about the streets; an urchin. 2. A girl or woman of impish appeal check out the French Kitty site -new illustrations called Gamine Girls  '),(6052,3860353,'13,July,2004','2004-07-13 00:00:00','   Gamine- 1. An often homeless girl who roams about the streets; an urchin. 2. A girl or woman of impish appeal check out the French Kitty site -new illustrations called Gamine Girls  '),(6053,3860353,'13,July,2004','2004-07-13 00:00:00','   Gamine- 1. An often homeless girl who roams about the streets; an urchin. 2. A girl or woman of impish appeal check out the French Kitty site -new illustrations called Gamine Girls  '),(6054,3860353,'13,July,2004','2004-07-13 00:00:00','   Thaumaturgy- 1. The working of miracles or magic feats. 2. The act or art of performing something wonderful; magic; legerdemain. --T. Warton.  '),(6055,3860353,'12,July,2004','2004-07-12 00:00:00','   Kittenish- Playfully coy and frisky. comments: like me:)  '),(6056,3860353,'11,July,2004','2004-07-11 00:00:00','   Swanny- 1. To declare; swear. 2. Swanlike; as, a swanny glossiness of the neck. --Richardson.  '),(6057,3860353,'10,July,2004','2004-07-10 00:00:00','   Bushed- 1. Extremely tired; exhausted   comments: like, the forth or so defintion but the way I\'m choosing to use it today. 2. A shaggy mass, as of hair. Vulgar Slang. A growth of pubic hair. comments: interesting.  '),(6058,3860353,'09,July,2004','2004-07-09 00:00:00','   Demitasse-  1.	A small cup of strong black coffee or espresso. 2.	The small cup used to serve this drink. p.s. this is what I need at the moment.  '),(6059,3860353,'08,July,2004','2004-07-08 00:00:00','   Osculate-  1. To kiss.  2. Mathematics. To have three or more points coincident with.   3. To come together; contact.  '),(6060,3860353,'07,July,2004','2004-07-07 00:00:00','   Chichi-  Ostentatiously stylish; deliberately chic.  '),(6061,3860353,'07,July,2004','2004-07-07 00:00:00','     '),(6062,3860353,'04,August,2004','2004-08-04 00:00:00','   Expresso- When you express yourself through espresso; caffeine from espresso influences one in a way that allows them to express themselves.                              hehe:) I just made that one up  '),(6063,3860353,'03,August,2004','2004-08-03 00:00:00','   Slugabed- One inclined to stay in bed out of laziness.  from Shakespeare\'s Romeo and Juliet  '),(6064,3860353,'02,August,2004','2004-08-02 00:00:00','   anticipointment- noun. A state of mind produced after having overly-anticipated a particular event or purchase, thus creating unreasonably high expectations, and then been disappointed after the event or purchase has taken place because the expectations were not met.  my comment:  This word describes exactly how I felt after returning home from The Sock Shop the other day -to discover that my long, white socks were not near as long as I thought they were when I bought them.  '),(6065,3860353,'01,August,2004','2004-08-01 00:00:00','   anticipointment- noun. A state of mind produced after having overly-anticipated a particular event or purchase, thus creating unreasonably high expectations, and then been disappointed after the event or purchase has taken place because the expectations were not met.  my comment:  This word describes exactly how I felt after returning home from The Sock Shop the other day -to discover that my long, white socks were not near as long as I thought they were when I bought them.  '),(6066,3860353,'01,August,2004','2004-08-01 00:00:00','   Corplaining- present participle;Griping, grousing, and otherwise complaining about the corporation you work for.  '),(6067,3935450,'31,July,2004','2004-07-31 00:00:00','    It\'s Saturday and I feel so shitty.  This is probably the first Saturday in forever that I haven\'t been shopping.  I\'m misty eyed and jaded.        I spoke to Eric yesterday and the more I think about him the more I lose interest.  I keep hearing him tell me how his opinion of me has changed....(PUKE! ) My thoughts are If you don\'t like it, forget it! He doesn\'t like me anyway.  I think he\'s just interested in sex.  He has potential to be awesome.  He\'s super funny, handsome, but soooo tense at times.  Everytime I say something sarcastically, I have to tell him--JOKE or he takes me seriously.  I guess thats going to go far.       Today I\'m starting induction.  I\'m going to stay under 20 carbs for the next two weeks.  I\'ve been eating poorly the last few days.  Not anymore.  Zippy carbs for Maya :)  Sugar what?       I\'m listening to Ave Maria and I think I\'m going to just cry myself a river.  Time to change it up a notch... lets see...Now, it\'s Sade\'s By Your Side--I think of Kat telling me lastnight that she had been with her husband throughout the entire stay in the hospital...and she wouldn\'t have it any other way.  She loved every minute she got to spend with him.  \"When you\'re cold, I\'ll be there to hold you tight to me\".  Gosh,  I\'m weak, dammit!       I need to stop feeling so sad.  Death is depressing and I feel like its all around me.  There is normally 3 in a row....I\'m wondering who the hell is next????  \"The Wages of Sin is Death\" Kat\'s father, My uncle, has Cancer as well and isn\'t expected to survive much longer either...My heart goes out to her.  But she\'s keeping the faith.  One of my favorite quotes \"God only gives me as much as I can handle, I just wish he didn\'t trust me so much\"   I\'m sure that\'s what Kat\'s feeling this a.m.       I can\'t dwell on this.  I have to think of something more positive.  I think I\'m going to go run and just release some tention.  And after that   I think I\'m going to go shopping.  That always feels good.        So...   Amaya complains a lot.  Mostly about petty shit too.  MEN > LOVE LOST> DAY TO DAY LIFE....whatever it is....I complain a hell of lot!  I rarely am grateful for much anymore.  I think that all started when M decided to walk out of my life.   I\'ve never been bitter.  I am now as a result.  I\'ve got to get over this hump, til then I\'m phucked.       I live a good life.  I do as I please.  I work hard, I travel often, date casually, fuck seriously, and complain often....K.  I made myself laugh.  Seriously, God has been good to me.  I beat AML, I have a great family who loves me, supports me 100% who will do anything for me.  These people are my everything.  And I concentrate so much on random guys when in reality only 2 have been good enough to be apart of my family...Everyone knows that\'s Adam & Aaron.        Adam--24, All American Boy who served in the Navy, Loves his family as much as I do, he\'s so passionate about me.  I\'m convinced that no guy will ever love me as much as he does or did.  He loves me, flaws & all. '),(6068,3935450,'28,July,2004','2004-07-28 00:00:00','    Today has been a real challenge for me.  I woke up to the news that My cousin\'s husband had died almost immeadiately after I was there at the hospital lastnight.  I knew he wouldn\'t live much longer, but he was taken too damn quick.  He\'s out of his misery now.  He\'s no longer suffering.  He was such a nice man.  Always smiling, always friendly.  I wasn\'t  that  close to him, but it hurts to see my family hurting.  I\'m extremely sensitive.  I cry at the drop of a hat.        It\'s still so freaking hard to believe.  He was so young.  Mom always says that God takes his angels first....      Becca, Missy, Eric, Emma & I went to dinner tonight and totally pigged out. tons of dessert, pastas, wines, and a lot of Calamari later we ran into another cousin, Allen, who was in town because of the death in our family...we all just vented to each other and hugged and it felt so good to express all of these feelings that I\'m having...the shock, the sadness, the pain..He left to go visit my cousin, the 30 year old, who is now a freaking widow....Then I asked for our check and our server says that we have none!  Shocked of course, we asked what she meant, so she said that Allen paid for everyone\'s dinner at our table!  K...So nice of him, however, we had a really exspensive dinner.  It was probably nearly somewhere around $200, including gratuity.  So that made me smile.  It was such a surprise and a totally sweet gesture.  It made my evening better.       On a much better note, I got a call from Adam today...He has a bar now, is doing well.  Talking to him though, it makes me miss him--for real.  We probably won\'t be talking again for a while, but we\'ll see.       Scott is doing well. Eric is good too.  Issaac is fantastic....      I\'ve had lots of random conversations with \"T\", he\'s doing well.  He wants Nicole back.  He truly loves this woman.  I think he\'s scared to let go of his pride and just tell her that he wants her and she wants him too, dammit!!!  He emailed me an article about relationships and such, I can\'t bring myself to read it.  He knows I can relate because of all the drama with M.M.   He got wasted lastnight...I\'m sure he\'s hungover now...sick too perhaps.  He\'s so adoreable sometimes.....      It\'s nearly 2 am. I\'m going to go to bed....      update more, later!   *~A~* '),(6069,3935450,'28,July,2004','2004-07-28 00:00:00','    Today has been a real challenge for me.  I woke up to the news that My cousin\'s husband had died almost immeadiately after I was there at the hospital lastnight.  I knew he wouldn\'t live much longer, but he was taken too damn quick.  He\'s out of his misery now.  He\'s no longer suffering.  He was such a nice man.  Always smiling, always friendly.  I wasn\'t  that  close to him, but it hurts to see my family hurting.  I\'m extremely sensitive.  I cry at the drop of a hat.        It\'s still so freaking hard to believe.  He was so young.  Mckground pic-->'),(6070,3935450,'28,July,2004','2004-07-28 00:00:00','    Woman to Woman Encouragement.Someone will always be prettier. They will always be smarter. Their house will be bigger. They will drive a better car. Their children will do better in school. And their husband will fix more things around the house.&nbsp; So let it go, and love you and your circumstances. Think about it. The prettiest woman in the world can have hell in her heart. And the mosthighly favored woman on your job may be unable to have children. And the richest woman you know -- she\'s got the car, the house, the clothes --might be lonely. And the word says if I have not Love, I am nothing.So, again, love you. Love who you are right now and let God be your barometer. Mirror Him. Look in the mirror in the morning and see how much of God you see. He\'s the only standard and even when you come up short; He will not leave you or forsake you. Smile and may God continue to bless you.&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; \"I am too blessed to be stressed and too anointed to be disappointed!\"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; The shortest distance between a problem and a solution is the distance between your knees and the floor.&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; \"The one who kneels to God can stand up to anything.\"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; \"Winners make things happen. Losers let things&nbsp;happen.\"Be Blessed ladies and pass this on to encourage another woman.&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; \"To the world you might be one person, but to one person you just might be the world.\"  &nbsp; --------------------------------------------------------------------------------------------- In other news...Work was great.&nbsp; I went to dinner with Haven & Mathew.&nbsp; He\'s sooooooo sweet.&nbsp; I feel sooooooooooooooo happy when he\'s around.&nbsp;   I called Scott today. I shouldn\'t have.&nbsp; I should have waited to see if he would call me, but silly me had to go and do it first!&nbsp; He said lastnight became lousy.&nbsp; I guess what he was trying to convey to me was the fact that she was a lousy lay.&nbsp;&nbsp;NO?&nbsp;   I\'m going to go for a run...writing that pissed me off.&nbsp; '),(6071,3935450,'27,July,2004','2004-07-27 00:00:00','    Brandi sent me a link to Craigslist, which I read often..it\'s hilarious, this though, I think has made my day.   TO THE GIRL IN THE HOT DRESS  You, looking hot in a black dress. Me riding in a correctional bus. Our eyes met through the metal slats. I was in orange. The driver wouldn\'t stop. He didn\'t understand. If you are willing to wait, I am available to date in march, 2007 (or sooner with good behavior). Please write. '),(6072,3935450,'27,July,2004','2004-07-27 00:00:00','    Today has just been wrong already.&nbsp; I got a sweet email from Issaac.&nbsp; He\'s on US soil now :)&nbsp; Yay!&nbsp; We\'re going to talk tonight.&nbsp; Scott was nursing his classy, drunk ass friend back to sobriety lastnight. Promised to call when he got back to the hotel room, never called.&nbsp; I\'m not pissed.&nbsp; His lost not mine.&nbsp;        &nbsp;   Aaron is being totally ridiculas.&nbsp; I guess he thinks I should spend all of my time sitting on the phone with him.&nbsp; I don\'t know how to tell him that I\'m not feeling the same way he is.&nbsp; Which sucks, he\'s a great guy.&nbsp; I love him dearly.        &nbsp;   I was with chatting with Eric when Aaron called lastnight.&nbsp; Eric is an interesting character.&nbsp; He\'s very attractive. I\'m still feeling him out too.&nbsp; I\'m not sure if we have anything in common.&nbsp; I\'m more family oriented.&nbsp; I love my family, I\'m always with them.&nbsp; He\'s just different I think.&nbsp; Who knows, I barely know him.&nbsp;      It\'s so odd...Scott is so cute and handsome, Eric is very hot, but I\'ll take Scott anyday over Eric....&nbsp;&nbsp;Theres just something about him.&nbsp; He\'s so sweet.&nbsp; Sometimes. When he calls.&nbsp; What I\'m trying to say is that although Eric is probably better looking than Scott, Scott is the one I\'m mostly attracted to.&nbsp;        &nbsp;   I think I\'m going to compile a background on all the guys just because I don\'t want to get them confused!!!&nbsp;        &nbsp;   I have a headache.&nbsp; I need to get laid.&nbsp; Everyday.&nbsp;        &nbsp;     &nbsp;   I think I may go see Batwoman tonight.&nbsp; I want to see it, but do I go alone?&nbsp;        &nbsp;   I read this article this morning....I found it cute.&nbsp;  urlLink http://story.news.yahoo.com/news?tmpl=story&amp;u=/ap/20040728/ap_on_el_pr/cvn_kids_for_kerry        &nbsp;     &nbsp;   Brandi sent me this via email, I have to share!&nbsp;    VACANCY - WOMB FOR RENT **************************** &nbsp;Hi Gentlemen, &nbsp;My name is Maggie. I\'m 26 years old, attractive & sexy. I\'m an aspiring actress and landlady of a very special piece of property. I\'m placing my property on the market. This is a beautiful piece of property, that will comfortably accommodate one nice man, and occasionally a women, if there\'s a party.   TENANT REQUIREMENTS:  1) Male tenant preferred and Female visitors allowed 2) Must be successfully employed, or retired 3) Must be clean cut, honest, and reliable 4) Must be respectful of the Landlady\'s property and treat it gently and with the utmost care. In seeking a tenant the Landlady does not discriminate by age or race.  PROPERTY FEATURES:  1) Very small & tight quarters. Perfect to lay your head down. 2) Manicured and neatly trimmed dark blond shrubbery, leading to a bright pink inviting entrance. 3) The interior is always well kept, clean, and fresh. The landlady always keeps the place nicely scented. 4) The property comes with all original appliances, that were installed 26 years ago. The appliances are in excellent condition, hardly used and well mantained. 5) The property is always dark, damp, and moist, and doesn\'t get much sunlight. However,the few tenants that have occupied the property in the past have really enjoyed it that way. 6) There is a \'WELL\' on the property, that will spout a gusher from time to time. In order to get the benefit &amp; pleasure of the gusher, it will be the tenant\'s sole responsibility to correctly sink a shaft and erect a pump. The pumping must be done vigorously. During the pumping, the tenant must supply and place a couple of his own large stones near the opening to the \'WELL\', thus preventing anyone from falling in accidently. Please note, when removing your shaft & pump you must be careful not to leave the opening any larger then you originally found it. This helps to prevent any unauthorized use or entry into the \'WELL\'. 7) The entire property is free of all diseases and vermin. The tenant must promise to keep it that way. 8) The exterior facade of the property is absolutely beautiful. It features a blond colored roof, freshly touched up, but mostly natural. There are two large round headlights mounted near the top of the facade. The high-beams are pointed straight ahead, and not towards the ground. This lights up the entire driveway making it easier for you to enter the property at night. 9) They always say, \"Location, Location, Location\". This property is centrally located in Manhattan.   HOUSE RULES: &nbsp;1) Please, No Dogs Allowed. But a little Pussy is nice. 2) Party & Entertaining Policy: The walls are strong and in great shape. So, having loud and wild parties, and banging against the walls is allowed &amp; encouraged. Also, if you have a couple of friends that you hang around with, and they\'re \'NUTS\', it\'s okay to invite them. They\'re always the life of the party. 3) All deliveries must come in the front door. There is a very nice rear door, but that\'s ONLY used to take out the trash. 4) You must be able to move in and out, rapidly and repeatedly, on short notice, if necessary. 5) Sorry, no short term transients. You must make a commitment to a long term lease. 6) All prospective tenant\'s must pass a credit and background check, prior to taking possession to the property. 7) The rent is negotiable. Make an offer. The tenant should feel free to make improvements to the property at their own expense. Redecorating & dressing the property up, as well as scenting it with expensive perfumes, is highly recommended.   APPLICATION PROCESS: &nbsp;If you\'re interested in applying for this property, then email your application to the landlady with all your important details. It would greatly help if you would describe the tenant and the size of your personal belongings. Since, the premises are VERY SMALL & TIGHT, you may not be able to fit large personal belongings. Send photo of tenant and personal belongings, if available. Upon receipt of your email, I will immediately send all approved applications, photos of the landlady together with her fine piece of property. If you like what you see in the photos, we can arrange a private showing. At which time you may be asked to leave a deposit on the premises. So please come prepared.  Best Regards, &nbsp;Maggie  P.S. Also for convenience, there is an eating establishment on the premises, for those who enjoy eating at the \'Y\'.     I found it funny :)&nbsp;        &nbsp;   Now I\'m going to go add my Platinum Taffy to my hair & go nuts with the hair dryer.&nbsp;        &nbsp;   I\'ll update more later...       &nbsp;     &nbsp;     &nbsp; '),(6073,3935450,'27,July,2004','2004-07-27 00:00:00','    Has to be the sweetest lady around!&nbsp; I came home from a day of doing nothing to hear her sweet message...\"Amaya, this is Rebecca, your grandmother, as if I couldn\'t distinguish her &nbsp;voice from my sister Rebecca\'s)&nbsp;, Sweetheart, I made your favorite peanut butter pie today with a homemade meringue and I don\'t want the crust sitting long&nbsp; so whenn you get home from work come by and pick it up and have some coffee with us.\" That definetly made my day.&nbsp; My grandmother is too damn sweet.&nbsp; We both share a really bad sweet tooth.&nbsp; I remember growing up as a child and helping her bake in her kitchen, thats when it was fun, licking the spoons and eating homemade frosting with your fingers!&nbsp; Needless to say, I\'m going over tonight after I unwind and I\'m in my pajamas where I\'m super comfy eating peanut butter pie with Grandma.&nbsp; To hell with the carbs, the fat, the&nbsp; sugar intake...pffffffftttt thats all out the door tonight!&nbsp; I think I\'m even going to skip my workout, even!&nbsp;        &nbsp;   I woke up bright & early feeling a bit guilty about&nbsp;lying &nbsp;to someone whom I\'ve developed feelings for.&nbsp; I hate that I did that, but I didn\'t feel like I could answer his question with the truth.&nbsp; As sad as that is...        &nbsp;   My day went rather well.&nbsp; Mike called & woke me up, he says that he\'s lonely living in the hotel...He\'s moving into an apartment in Georgetown on Saturday.&nbsp; He said that Melanie is leaving for Australia on Sunday, by telling me that alone I know that he\'s had contact with her.&nbsp; Everything is a lie with him.&nbsp; Thats so messed up though. The entire situation.&nbsp; I guess I should wave my white flag.&nbsp;        &nbsp;   I spoke to \"T\" lastnight.&nbsp; He\'s so increadiably human, it\'s so easy to just talk to this man.&nbsp;He said something that moved me, I had a moment!!!&nbsp; He said that he was searching for homes and he actually wanted Nicole\'s opinion on the home before he went into contract....he said that he just wants her to be apart of it.&nbsp; How sweet is that?!!!&nbsp; I like how he\'s becoming passionate about getting back&nbsp;with her&nbsp;after he basically called it quits, and now she\'s saying they need time away from each other.&nbsp;&nbsp;I\'m sure if it\'s meant to be, it will be.&nbsp; And this morning, like the idiot I can be sometimes, I call this number that I didn\'t reconize until I heard \"T\"\'s voicemail.&nbsp; I felt like such an idiot!&nbsp;He called back & I explained to him what had happened, he laughed it off, THANK GOD!!! I was mortified.&nbsp;        &nbsp;   I also talked to Scottie-to-Hottie lastnight as well.&nbsp; He was a bit drunk but so damn sweet at the same time.&nbsp; He told me that he wanted to be with me!! Isn\'t that what I\'ve been wanting?&nbsp; I think so :)&nbsp; Ask me tommorrow, that could all change overnight!!!&nbsp; Just kidding.&nbsp; We talked until I got sleepy.&nbsp; He will be home on Thursday.&nbsp; I can\'t wait.&nbsp; He travels so much with work it makes me sick!&nbsp; Thats his job though.&nbsp; I can\'t complain.&nbsp; I travel a lot as well.&nbsp;        &nbsp;    And news of the day---I called Broke Ass&nbsp;today.&nbsp; I can\'t believe I did it.&nbsp; I just picked up my cell and dialed his stupid work number from memory and he answered.&nbsp; The odd thing is, he didn\'t know who&nbsp;I was!!!&nbsp; (WHAT WAS I THINKING)&nbsp;We talked briefly about him and then I said \"I Miss ya\"!!!&nbsp; Then I got this huge punch right in the arm from Haven!&nbsp; She couldn\'t believe I was telling him that.&nbsp; I had to tell him that.&nbsp; I did miss him.&nbsp; I have no excuse.&nbsp; I wasn\'t drunk. I have nothing to blame it on.&nbsp; I\'m ashamed of myself.&nbsp; Haven said that she\'s never seen me so weak before.&nbsp; Scary.&nbsp; I do miss the fucker.&nbsp; He was awesome. It was the freaking Avril song...My Happy Ending.&nbsp; MOVING ON.&nbsp;'),(6074,3935450,'26,July,2004','2004-07-26 00:00:00','    ~Super tired right now, I went to Bible Study. Got my learn on.&nbsp; I thoroughly enjoyed it.&nbsp;        &nbsp;   ~Worked out fiercely, lifting Kennedy from Baby Carrier to Baby Bouncer, to my bed to my floor.&nbsp; Pushing Emma on the swings was tough as well.&nbsp; No workout comes close to being as hard as playing Hide & Go Seek with Isaiah.&nbsp; My workout was intense, overall.&nbsp; More emotionally than physically. No worries.&nbsp; I will go nuts tommorrow.        &nbsp;   ~Scott phoned from Houston, at a bar with his brother.&nbsp; Sooooooo cute.     ~\'T\' is adoreable too.&nbsp; Rocks my socks :)&nbsp;&nbsp; I\'d hit it!!!      ~Emma is saying Puerto Rico all the time.&nbsp; She\'s excited to be going. About as much as I am.&nbsp;   ~Isaiah said I was more of a Velma/Thelma from Scooby Doo.&nbsp; I was offended at first, because She\'s disgusting, has bad taste, dresses poorly and has coke bottle thick glasses--Then he told me I was more like her because I\'m sooooooooo smart!!!&nbsp;&nbsp; I was flattered afterall.&nbsp; In the end I\'m glad, because the blonde guy isn\'t my type at all.&nbsp; Jesuth Chrith, he\'s Thuper Gay!  ~I\'m off to bed.&nbsp; No wet dreams.&nbsp;   &nbsp; Amaya '),(6075,3935450,'25,July,2004','2004-07-25 00:00:00','    I\'m feeling like a bad girl, I didn\'t go to church again this a.m.&nbsp; I didn\'t get up early enough :(&nbsp; But I did take communioun!       &nbsp;   Saturday, Mike & I went to the Zoo.&nbsp; Soo much fun.&nbsp; We\'re going to go back again real soon.       &nbsp;   Lastnight, Michelle, Kellie, Haven, Becca, Trish, Kevin, Stacy, Danny, Ross, Ian, Courtney, Rick, Mason&nbsp; went to a carnival.&nbsp; It was soo much fun being a kid again.&nbsp; I actually rode the Ferris wheel...just like old times.&nbsp; I bought home my cotton candy & candy apples.&nbsp; I was a happy girl.       &nbsp;   Scott called while I was out with everyone so we didn\'t get to talk much.&nbsp; I called at 2 am when I got home but to no avail.       &nbsp;   \"T\" was supposed to call too.&nbsp;&nbsp; He didn\'t!&nbsp; SHAME SHAME!!!        &nbsp;   I\'m going over to Ian\'s & Courtney\'s new crib and I\'m having some beers.&nbsp; Update more later :)&nbsp;        &nbsp;   XOXOXOXOXOXOXOXO       &nbsp;   ~Maya '),(6076,3935450,'23,July,2004','2004-07-23 00:00:00','    Holy HELL!!!       &nbsp;   Voicemail from Mike....       &nbsp;   \"I just moved out of the house, Call me when you get this\"&nbsp;        &nbsp;  I never saw it coming.&nbsp; He\'s been saying that for months though...damn.&nbsp; I\'ll call but what am I going to do?&nbsp; Let him move in?&nbsp; SHeesh! Never.    &nbsp;  Back to my shower.... '),(6077,3935450,'22,July,2004','2004-07-22 00:00:00','    I\'m feeling Avril Lavigne\'s newest song \"Happy Ending\"&nbsp; I\'ve been singing it all day!        &nbsp;   Scott & Brad are leaving town tonight...Back next week.&nbsp; I think I\'ll miss him :)&nbsp; I think it\'s blatant that I\'m crushing on him.&nbsp; He\'s such a good guy.&nbsp; We get along well thus far.&nbsp; Does that mean all the others are out of the picture?&nbsp; Not exactly ;)&nbsp; Not yet anyway.&nbsp;        &nbsp;   So I lost a pair of Jimmy Choo\'s.&nbsp; I\'m soooooo going to die if I don\'t find them. They\'re the cutest black shoes ever!&nbsp; So summer 2004 too.&nbsp; I\'m thinking they\'re in Myrtle Beach :(&nbsp;&nbsp;        &nbsp;   Mom called today, we leave for Puerto Rico on the 29th.&nbsp; I\'m only going to be gone for 5 days, while they\'re staying an entire 10!&nbsp; Bastards.&nbsp; But the gallery would fall apart if I was gone longer than 5 days.&nbsp;        &nbsp;   Tonight is Kellie\'s Birthday.&nbsp; Michelle, Brandi, Lisa, Haven, Kristi, Matt, Trevor, Cameron, Brian, Becca, Courtney, Tara, and the list goes on are all meeting for dinner @ 8:00 @ Milano or Sequoia\'s not sure yet...Then on to a night of celebrating.&nbsp; I won\'t be celebrating much though. Too many carbs in the drinks for my taste.&nbsp;        &nbsp;   Aaron told me he loved me lastnight.&nbsp; How cute is that?!!!&nbsp; He\'s definetly one of my bestest friends forever.&nbsp; Not in love though.&nbsp; I have my eye on Scott :)&nbsp; For right now!       &nbsp;   Oh and Vishal Milino wants to take me to Crete.&nbsp; I\'m about to say yes.&nbsp; He\'s so damn hot.&nbsp; Armani, Versace, Gucci everything, everyday.&nbsp; The man has style, the man has grace.&nbsp; It\'s just too bad you can\'t buy cool.&nbsp;&nbsp; I think I can do a few days in Greece though, even with an uptight man.&nbsp; I could loosen him up though :)&nbsp;        &nbsp;   On to the shower...       &nbsp;   ~*AMAYA*~ '),(6078,3935450,'22,July,2004','2004-07-22 00:00:00','    Today has been pretty uneventful.&nbsp; I just went to the office sporting Juicy Couture, a tank, and a ponytail with my Dior\'s.&nbsp; I was scrubbing!&nbsp; And I didn\'t care either.&nbsp; I signed & notarized some contracts, checked my voicemail and got out of there.&nbsp; The day was too pretty to be stuck inside.&nbsp;     &nbsp;  Scott leaves town tommorrow :(&nbsp; Guess that means that it\'s a girls night out again!&nbsp;Can\'t wait.     &nbsp;  So, I paid bills today, went to the bank and had lunch with Mom.&nbsp; She\'s so awesome.&nbsp; I\'m such a Mama\'s Girl....I don\'t know though, Daddy\'s Girl too!&nbsp; Dammit--I\'m both.    &nbsp;  Daddy\'s Birthday is on August 14th, Mom\'s is August 5th, and we\'re planning a big birthday party for them.&nbsp; The location has been set yet, but we\'re thinking Fairfax Country Club.&nbsp; Thats all still up in the air.&nbsp; But I love planning events and such so this is going to be fun :)    &nbsp;  Tara gets married August 7th,&nbsp; She\'s registered at Neiman\'s & Nordstrom\'s &amp; Tiffany\'s.&nbsp; That\'s going to be fun, buying presents for everyone else! Right.    &nbsp;  Jeremy also gets married on August 7th.&nbsp; It\'s weird how I got invited.&nbsp; I wonder if he wonders if I care.&nbsp; My grandparents are going.&nbsp; My Mom & Dad are going.&nbsp; I think thats rather gay, but to each their own.     &nbsp;  I\'m going to go running and hopefully I\'ll feel better.&nbsp; I\'ve been eating the hell out of Karb Karma Super Fudge Chunk by Ben & Jerry....The God\'s of Ice cream.     &nbsp;  Tommorrow Ian & I are playing tennis and then I\'m getting my roots did.&nbsp; Mario will be the only man that will touch my hair.&nbsp; In a non sexual way!&nbsp; But then again, Mario is hot, love his Croatian accent, mmmmmm! I\'d do him.&nbsp; The guy is sex on legs.&nbsp; His cologne, his mannerisms, his winks, his gestures, his smile, the way he bites his lip, the way he leans in to talk to me and his chest rubs against my tits--I think I want that!!!&nbsp; He\'s always been interested and dammit, I\'m getting laid this weekend!!!!    &nbsp;  K...Me & My horny ass is going to go running...As if I need anymore endorphines :)&nbsp;     &nbsp;    SUPERGIRL! '),(6079,3935450,'21,July,2004','2004-07-21 00:00:00','    I was up too late with Scott lastnight, I\'m super tired.&nbsp; Lisa called at 6:00 am and asked if I could watch Kennedy for few hours, I said sure.&nbsp; I didn\'t realize I had to climb my happy ass out of bed to actually drive 10 mins in my jammies to pick her up!&nbsp; That was nothing compared to having to pump gas in Pajamas and paying for it inside because the machine wouldn\'t read my debit card!!!&nbsp; So Kennedy & I came home and we climbed into my bed and fell asleep.&nbsp; I woke up around 9 ish and she was awake just laying there trying to talk to me!!! SOOOOOOOO CUTE.&nbsp; She didn\'t even cry, she just laid there.&nbsp; I\'m in love with this child already.&nbsp; I think she\'s going to be mine too :)&nbsp;&nbsp;     &nbsp;  I worked out lastnight from 8-9.&nbsp; Ate healthy all day too.&nbsp; My butt is a rock!!! So firm.&nbsp; Definetly a runner\'s butt :)&nbsp; Things are getting back to the way they once were.&nbsp;&nbsp; I realized that the only cd I ever listen to while working out is Britney\'s CD.&nbsp; I love it.&nbsp; It\'s upbeat, it slows down, it\'s awesome.&nbsp; Love her cd even :)&nbsp;     &nbsp;  I just checked my email and Pyscho Kenny emailed me from another email address, thats blocked now too.&nbsp; He\'s really a twisted human being.&nbsp; What part of STOP doesn\'t he get?&nbsp; I\'ll paste the email.&nbsp;     &nbsp;   This may be a product of my newfound conscience, but I don\'t like having people  carrying around animosity toward me {Since when?}. We\'ve been through this a few times {WE have?}, so no, I don\'t expect us to be close or anything like that {Smart man}, I just want to apologize for anything that I said that may have offended you {SICK, HE MAKES ME SICK}. No, I\'m not \"psycho\" {He\'s never proven me wrong} or a \"stalker\" {He\'s attempted all means to contact me even nonstop via phone, what would you call it? Denial perchance?}I was just tryin to make things work with us{There was never an US}. Anyway, as the song  says, \"Don\'t go away mad, just go away.\" So, that\'s what I think we should do{This is about the 5th email about just going away, it\'s never been done yet}. I  feel like I went out on a rather classless note {Classless note --->classless person,&nbsp;does 2+2=4?}, and I felt badly about that{Tear!} Good luck with everything, again{Right}, I apologize for what I said{Haha, uh huh}. Apparently I have a tendancy to lose it here and there {You think? I never noticed!}(although I\'m sure you knew that).&nbsp;&nbsp;       Kenny&nbsp; {STALKER}   So thats the email. Everything in brackets are my thoughts.&nbsp; He\'s a weirdo for sure.&nbsp;   So lastnight was interesting...Aaron called, we talked for a while.&nbsp; I\'m going to hang out with my cutie pants for a few days at the end of the month.&nbsp; I haven\'t seen him in so long.&nbsp; He\'s so adoreable.&nbsp; He\'s the one guy friend that I\'ve had forever that has never changed or has ever said or treated me unfairly or improperly.&nbsp; He\'s my definetion of gentlemen.&nbsp; Great guy.&nbsp;  He likes peanut butter too :)&nbsp; We ate peanut butter together on the phone lastnight!!! Talk about weird conversations--we have those! &nbsp;DORKS!   Scott told me that he\'s interested in me :)&nbsp; (like I didn\'t know)&nbsp; We\'re going to hang out in a few days and actually go on a real date.&nbsp; He had a rough day at work yesterday.&nbsp; The family business thing is hard to do.&nbsp; I\'m glad Mom doesn\'t work with me!&nbsp; We\'d be at each other\'s throats al the time!&nbsp; Anyway, the bad thing about Scott is that work is his life--he\'s going to be leaving town for Houston on Friday won\'t be back until Thursday of next week.&nbsp; I\'m not sure if I like that :/&nbsp;&nbsp;   Issaac is in his transition phase.&nbsp; He\'s going to be back home sometime this week I think :)&nbsp; I can\'t wait.&nbsp; It\'s Top Secret, the actual date.&nbsp; But if my sources are correct, he\'ll be back home on Thursday :)&nbsp;&nbsp; We shall see!&nbsp;   Kennedy is napping now so I\'m going to get my shower while she\'s asleep.&nbsp; I\'m going to be at the office til 6 tonight, then I\'m working out, I told Mike that we could have dinner while the bitch is out of town.&nbsp; We shall see.  Amaya '),(6080,3935450,'20,July,2004','2004-07-20 00:00:00','    My lovely Brandi has gone nuts!!! Here is the email she sent...Loco Chic! &nbsp; &nbsp;  WANTED:  Slick, savvy and charismatic tycoon-to-be with the fortitude to make a fortune with anything from land investment to lemonade production  Amaya,   The Apprentice\'s Donald Trump&nbsp;&nbsp;is doing a nationwide search for the next Apprentice! You must apply or I will apply for you.&nbsp; You\'re &nbsp;Slick, Savvy and Charismatic, Total Tycoon to be to!!&nbsp; You have the look, the brains and the attitude&nbsp;to&nbsp;at the very least make it to the semi-finals&nbsp;&nbsp;so do it already or I will tai chi your ass, girl! Matt said that he is going to make you apply to! Here is the application it\'s a PDF so open it now! I\'ll be getting ideas together for your VHS debut.    &nbsp;   &nbsp;   Now, I\'m convinced that she\'s gone totally nuts or she\'s smoking crack!!! Perhaps both.&nbsp; The questions are super gay too, An Embarrassing moment?&nbsp;God knows I have plenty...A major event in my life?&nbsp; Uhm, losing my virginity?&nbsp;Why am I successful in Business?&nbsp; Cuz I\'m cute dammit!!!&nbsp; KIDDING :) &nbsp;She\'s a dork. It\'s keeps getting worse..Needless to say, I won\'t be applying.        &nbsp;     &nbsp;   So I went to my meeting with Santos yesterday.&nbsp; Nobody told me he was a Greek God!&nbsp; Single too.&nbsp; Gotta luv that.&nbsp; So witty.&nbsp; Attractive.&nbsp; Contract signed :)        &nbsp;   Anyway--I went to Bible Study after I lusted all evening.&nbsp; A lot of people find it strange that I\'m so passionate about learning about something that can be so trivial.&nbsp; I like the idea of having faith in a Higher Being.&nbsp; Jesus is that to me.&nbsp; It\'s not \'cool\' or \'acceptable\' to be religious at my age.&nbsp; But I don\'t care what society thinks.&nbsp; I\'m comfortable with it, I love it and it\'s a passion.&nbsp; I try to read scriptures daily and apply them.&nbsp; There is nothing better than the power of Prayer.&nbsp; I always find myself praying.&nbsp; Waking up, Throughout the day, before eating, before bed.&nbsp; Sometimes my prayers are soo long that I find myself falling asleep!&nbsp; I always give Thanks and note what I\'m thankful for each and everyday.&nbsp;&nbsp; I\'m far from perfect.&nbsp; I\'m a sinner.&nbsp; It\'s human nature to sin.&nbsp; It\'s asking for forgiveness, it\'s trying to not&nbsp;commit the sin anymore--which is soooooooo hard, that will help me overcome temptation.&nbsp;&nbsp; Like this weekend, the way we partied, SO WRONG.&nbsp; I knew better.&nbsp; I regretted it.&nbsp; I didn\'t even go to Church on Sunday.&nbsp; Thats when I definetly should have went.&nbsp; Communioun is another must.&nbsp; I missed out on it again.&nbsp; Everytime I miss it, I feel like I\'m rejecting Christ....so wrong.&nbsp; I will get better.        &nbsp;   After Bible Study, I worked out--Hard too!!!&nbsp; 2 hours of Cardio.&nbsp; I worked on some proposals and then fell asleep atop my bed, no blankets on me at all, thats how exhausted I was.&nbsp; I woke up mid morning to the phone ringing.&nbsp; TMike was out drunk wanting to come over...I hung up on him.&nbsp; He just got back from the Hamptons.&nbsp; It rained all week there :)&nbsp; His tan isn\'t that good!!!&nbsp;        &nbsp;   I am going to have lunch with Dad today, play some tennis, pick Isaiah up from swimming lessons for Becca, then Emma, Isaiah &amp; I are going to the library for story hour and maybe, depending on the time, go to Grandma\'s and swim.&nbsp; I haven\'t worked on the tan in a while...&nbsp; The sun is my addiction.        &nbsp;   I still have to find a Cingular store around me!&nbsp; The website is gay.&nbsp;        &nbsp;    More later :)  &nbsp; &nbsp; ~*Amaya*~     &nbsp;   &nbsp; '),(6081,3935450,'19,July,2004','2004-07-19 00:00:00','    \'There\'s a boy I know'),(6082,3935450,'18,July,2004','2004-07-18 00:00:00','    My quick fix for a Hangover.&nbsp;&nbsp; I don\'t know why I allowed myself to get so wasted.&nbsp; It wasn\'t cool at all.&nbsp; I made so many drunk phone calls.&nbsp; God forbid I said or did anything that offended anyone.&nbsp;        &nbsp;    I didn\'t wake up in time for church.&nbsp; I haven\'t been in so long.&nbsp; I feel terrible about that.&nbsp; I need to get back on track.&nbsp; This destructive behaviour isn\'t me, at all.&nbsp;        &nbsp;    Isaiah and I ordered Fattening Italian food loaded with Carbs. *YAY*&nbsp; Very&nbsp;yummy, but I feel bloated and disgusted by eating it.&nbsp;&nbsp; My lil Isaiah is so cute, we ate his spaggetti&nbsp; a la ~\"Lady and The Tramp\" style...Super cute even.&nbsp; It fot a lil messy but anything to make him smile :)        &nbsp;    Yesterday was scary.&nbsp; We were shopping, Marathon style--Mom, Becca, Haven &amp; Me along with Chandler &amp; Isaiah.&nbsp; Becca, Mom &amp; Haven were still trying on clothes at Banana Republic,&nbsp; I had already selected everything I wanted&nbsp; and paid, Isaiah was restless and getting bored.&nbsp; I offered to take him over to Barnes &amp; Noble to chill with a Sugar-Free Mocha&nbsp;Frap with extra Whip Cream and Chocolate Shavings--Just the way we luv it!&nbsp;&nbsp; So we got our \'Frappies\' and we went to the children\'s reading corner, he was looking for books on Sharks, I was on my cell with Aaron,&nbsp; I told Isaiah I was going to go Pee, I\'ll be right back....        &nbsp;    I come out of the bathroom still on the phone with Aaron, Becca, Mom, Haven, Chandler approach me and ask where Isaiah is....I told them in the children\'s section, Becca takes off running&nbsp; toward that area&nbsp; Mom heads to the front, Haven informs me that There was a code Adam at Banana Republic for the area we were in...Code Adam means there was an attempted abduction.....Haven&nbsp;&amp; I freaked when Becca tells us that he\'s not back there...I ran to the Stationary section, almost in tears myself when this older Man &amp; Woman have Isaiah at the checkout with a clerk....Isaiah was hysterical. Naturally. Then I balled too.&nbsp; He was lost and scared.&nbsp; I felt like the worst Aunt in the world.&nbsp; SO SCARY.&nbsp; He will never leave my sight again.&nbsp; I\'m too carefree, It won\'t happen again especially after hearing about the kid that was almost abducted earlier in the day!&nbsp; How phucked up is that?&nbsp;&nbsp; SICK people.        &nbsp;    David, Zenus,Courtney, Becca, Kristina, Michelle, Haven, Matt, Chris, Danny, Michael and I all went out clubbing lastnight, Courtney, Ian, Becca &amp; I all had dinner together first.&nbsp; Fun Times.&nbsp; I started a tab early in the evening, when we were leaving it was nearly $200.&nbsp; So I basically woke up and pissed $200.&nbsp; THAT MAKES ME SICK.&nbsp; The good part is, As we were getting into the&nbsp; Escalade lastnight, in a puddle of water, I put my manicure in it--even, and I found a $100 bill.&nbsp; I initially thought it was only $10.&nbsp; I was going to take it to the bouncer, My brother grabbed my arm and Bruised the mohucker at that, Told me to not be so stupid and keep the money...I\'m such a nerd...The Bouncer told me to keep it.&nbsp; Courtney told me after we let everyone out at&nbsp;destination,&nbsp;that I told the bouncer that the money was probably Milk Money for someone\'s baby!!!&nbsp;&nbsp; She said I also told him that I can\'t take the money--I have to go to Church tommorrow and the bouncer said \"He\'ll forgive ya, but if he don\'t come see me and I\'ll spank ya\" Then he told me to call him sometime---I have no recollection of any of that being said.&nbsp;&nbsp;However I found a wet $100 bill this morning.&nbsp; So technically I spent $100.00 after God gave me back my money for new shoes!!!        &nbsp;    As for my cute purchases yesterday,&nbsp; I bought a LV Cell Phone Wristlet, it\'s Spring 2004 but thats okay! I\'ll link you        &nbsp;    I Bought the cutest, Sunshine colored Tube dress, A la Juicy Couture to wear to Tara\'s Wedding in August @ Neiman\'s. It\'s an Outdoor wedding so it\'s perfect.&nbsp; I have a pair of Sparkling, Strappy Manolo\'s that work perfectly.&nbsp; Becca Suggested I carry a Dior Clutch....we\'ll see. The best part is...It was under $150!!&nbsp;        &nbsp;    At Bebe where I went nuts, I got the cutest Cami\'s ever!&nbsp;  urlLink http://www.bebe.com/Main/browse_detail.jsp?PRODUCT%3C%3Eprd_id=845524442540644&amp;FOLDER%3C%3Efolder_id=2534374302796658&amp;bmUID=1090185919655&amp;viewmode=page         &nbsp;    I got it in Saphire &amp; Willow! Super sweet &amp; I can wear it to work&nbsp; &amp; out. It would look hot with some Seven Jeans or with a cute white Paris Skirt.        &nbsp;    I also got the prettiest lil lime skirt...it\'s belted...I love the fit and the way it flows.&nbsp; Again I can wear this out or to the office.  urlLink http://www.bebe.com/Main/browse_detail.jsp?PRODUCT%3C%3Eprd_id=845524442537901&amp;FOLDER%3C%3Efolder_id=2534374302796664&amp;bmUID=1090186545557&amp;viewmode=page         &nbsp;    I also got a lot of cute workout tanks &amp; pants    &nbsp;     urlLink http://www.bebe.com/Main/browse_detail.jsp?PRODUCT%3C%3Eprd_id=845524442539022&amp;FOLDER%3C%3Efolder_id=2534374302796687&amp;bmUID=1090186659126&amp;vewmode=page         &nbsp;      &nbsp;    urlLink  http://www.bebe.com/Main/browse_detail.jsp?PRODUCT%3C%3Eprd_id=845524442538577&amp;FOLDER%3C%3Efolder_id=2534374302796691&amp;bmUID=1090186804793&amp;viewmode=page     &nbsp;  At Victoria\'s  &nbsp;  Their swimsuits were on sale, Haven &amp; I went nuts... What we got isn\'t online, but I\'ll post pics in a few days.&nbsp;  &nbsp;  I got this beautiful Floral Mini--TOO CUTE!!!&nbsp; Brandi is going to love this.   urlLink  http://www2.victoriassecret.com/commerce/application/prodDisplay/?namespace=productDisplay&amp;origin=onlineProductDisplay.jsp&amp;event=display&amp;prnbr=MA-172891&amp;cgnbr=OSSALCLOSKI&amp;rfnbr=722&amp;page=1&amp;cgname=OSSALCLOSKI     &nbsp;  I got this outfit for next to nothing!&nbsp; I wanted it to wear to work, it\'s too conservative for playtime...I thought it would look great with my Chanel sunglasses and my blue Coach Hobo Spring 2003 bag    urlLink  http://www2.victoriassecret.com/commerce/application/prodDisplay/?namespace=productDisplay&amp;origin=onlineProductDisplay.jsp&amp;event=display&amp;prnbr=MA-160532&amp;cgnbr=OSSALCLOPAN&amp;rfnbr=725&amp;page=1&amp;cgname=OSSALCLOPAN     &nbsp;  And as if this isn\'t enough to piss you off, This I bought in late early June for   urlLink  M$@#\'s    Birthday, it\'s now only $38!&nbsp; I paid nearly $90!  urlLink  http://www2.victoriassecret.com/commerce/application/prodDisplay/?namespace=productDisplay&amp;origin=onlineProductDisplay.jsp&amp;event=display&amp;prnbr=MA-158555&amp;cgnbr=OSSALSLPBAB&amp;rfnbr=1068&amp;page=4&amp;cgname=OSSALSLPBAB     &nbsp;  Super Cute and Very comfy...    urlLink  http://www2.victoriassecret.com/commerce/application/prodDisplay/?namespace=productDisplay&amp;origin=onlineProductDisplay.jsp&amp;event=display&amp;prnbr=MA-168866&amp;cgname=OSSALSLPPYJ&amp;cgnbr=OSSALSLPPYJ&amp;rfnbr=714&amp;page=1&amp;cgname=OSSALSLPPYJ     &nbsp;  &nbsp;  &nbsp;  I did get a couple of pairs of panties too, but their nothing worth posting ;)&nbsp; Here anyway!  &nbsp;  I\'m going to go and play Volleyball with the boys....Oh and my brothers new Escalade is hot!  &nbsp;  *Amaya* '),(6083,3935450,'17,July,2004','2004-07-17 00:00:00','    ....And that means it\'s a day of shopping! I\'m thinking Pentagon City &amp; Tyson\'s Galleria II.&nbsp;&nbsp; I\'m feeling some Juicy Couture I think.  &nbsp; Dinner was great last night. I brought Emma along.&nbsp; She\'s going to be just like me :)&nbsp; She\'s a big fan of Sushi already!&nbsp; I painted her fingernails and toe nails.&nbsp; She loved it. Almost as much as I did. I took some shots of her playing on my bed yesterday, she loves the camera.&nbsp; Loves to smile all the time--She melts me!  I went running yesterday after work, My neighbor who is a hot Italian&nbsp; (God forbid), I always thought Puerto Rican,&nbsp;asked if he could run with me, I was apprehensive at first, we\'ve only chatted breifly before.&nbsp; He\'s super funny and sexy as all hell.&nbsp; But He\'s married 2 children with another on the way in August.&nbsp; Not shocking to me.&nbsp; This only proves to me what I already knew--I\'m a loser magnet.&nbsp; He told me I could come over and swim anytime I want.&nbsp; \"Genevieve won\'t mind\" RIGHT.&nbsp; I\'m all over that!&nbsp; He\'s educated and a great businessman at that. He owns several fast food chains here, Domino\'s Pizza being one of them.&nbsp; Dale Ernhardt Jr and Michael Waltrip are coming out to his business today to sign autographs and memorabilla.&nbsp; I told him I\'d make a cameo.&nbsp; I probably won\'t.&nbsp;  &nbsp; So, the church that I attended as a child, the one I didn\'t agree with, The pastor\'s wife died. She lost her battle with Lukemia on Wednesday. She lived with it for over 15 years.&nbsp; It\'s hard to believe that she hung on that long.&nbsp; She\'s or she was rather, such a sweet lady. Always polite, proper, and sweet.&nbsp; I remember when I was diagnosed with AML, She came to visit and I was complaining about all of the crap I had to deal with, she studiously said,&nbsp; I have to take 37 different pills at breakfast, lunch &amp; dinner.&nbsp; She said it with a smile.&nbsp; Only grateful to be living at all at that time....Her daughter, Rebecca, whom I graduated with told me that Judy didn\'t want to leave her grandchildren behind, she was having a hard time coping with that fact.&nbsp; I know that she\'s no longer feeling any pain at all. I know that she\'s comfortable and I know she will be with her grandchildren too one day.&nbsp; I know that right now she isn\'t feeling any sorrow at all.&nbsp; I sent a large array of flowers from my family and the Gallery.&nbsp; I\'m actually going to go and pay my respects today before I do any shopping &nbsp; I washed my car this morning, I love it when it\'s all clean and shiny :)&nbsp;&nbsp; It\'s such a super cute car.&nbsp;&nbsp; Isaiah loves it.&nbsp; I love it.&nbsp;  &nbsp; I\'ve been eating poorly.&nbsp; I bought those damn Lemon Pastry Girl Scout Cookies and I can\'t stop eating the bastards!!&nbsp;&nbsp; I\'m going to pour salt all over them so I won\'t eat them....haha &nbsp; I should get going.&nbsp; I\'m sporting some super cute Cropped Jeans by Chloe :)&nbsp; Gotta luv em.&nbsp; I\'m channeling Jessica Simpson today!&nbsp; That is a good thing....everyone knows how super cute she is.&nbsp;&nbsp;  &nbsp; OH and I talked to Issaac via phone lastnight, yeah, I bailed on the girls!&nbsp; I loved hearing his voice.&nbsp; He\'s coming home in less than 2 weeks :)&nbsp; My sexy boy!&nbsp; I can\'t wait really!&nbsp; I\'ll have to add pictures to Fotki of him.&nbsp;&nbsp;  &nbsp; I\'m listening to Beyonce\'s cd and I love it.&nbsp;&nbsp;  &nbsp; Plans for tonight include hitting up Platinum, Dream and wherever else we end up.&nbsp; Chris is getting us a limo....I paid for the last one, I\'m not in on this one! &nbsp; Gotta get going...... &nbsp; &nbsp; *Amaya*&nbsp;  &nbsp; &nbsp; '),(6084,3935450,'16,July,2004','2004-07-16 00:00:00','    \"I&nbsp;can\'t   be elusive with you Honey    Cause It\'s blatant that I\'m feeling you   And it\'s too hard for me to leave abruptly   Cause you\'re the only thing I wanna do   And it\'s just like Honey   When your love comes over me   Oh baby, I\'ve got a dependancy   Always strung out for anotherTaste of your honey\"    &nbsp;   &nbsp; &nbsp; I can\'t stop singing that!&nbsp; I know it\'s so 1999 but I heard it yesterday and I still luv it just the same.&nbsp; &nbsp; I got my tan on yesterday...Luv that. A new addiction that I\'ve found. As if I need another.&nbsp; I\'m going into the office today to get some work done, who knows where the weekend will lead me.&nbsp; Ian-Bradley and I are going to find him his SUV.&nbsp; He wants something that he can go Trail Blazing in....I suggested a Range Rover, we\'ll see.  &nbsp; I\'m going out to dinner with Mom &amp; Dad tonight, Daddy wants to go to his favorite Japanese Steakhouse, as long as I get my sushi, it\'s all good! Then I\'m meeting up with Brandi, Haven, Payton, Courtney and Jules---We\'re going over to Dream, Ozio and Club Daedalus. Should be fun.&nbsp; Issaac\'s going to call tonight from Fallejuah Iraq...I may miss his call :(&nbsp;  &nbsp; Poor Jas is still a bit upset.&nbsp; I hate that he\'s so hurt but I don\'t know what he was expecting...All I can think about is him saying \"Gonna drink some beers, try to catch a buzz, turn my radio up and forget about us.\" I\'ve always been upfront with him...I told him I wasn\'t going to regress. I know he\'ll be alright...We\'ve tried dating. It wasn\'t for us.&nbsp;I won\'t dwell on that.  &nbsp; So, I took Emma to the Circus that came to town lastnight...The workers really do give new meaning to Freaks!&nbsp; It costs 75 cents per person to see a white Tiger and a Panther, I gave the guy a $10 bill, and I told him I was paying for all six of us, (Ian, Isaiah, Becca, Courtney, Emma &amp; Me) he couldn\'t figure out the math! He had to get someone else to give me back my change and he couldn\'t do it either!!! So, I basically told them it was $4.50 I owed them, they owed me $5.50,&nbsp; Then he didn\'t believe me and broke out pen &amp; paper---TOTAL FREAKS! They all looked like freaks too, very deformed. Scary even.  &nbsp; The Babboon had a diaper Rash that was very disgusting...The Bengel Tiger was stir crazy and was pacing back and forth, The llamas Reeked, the Goats were gay, and the place was filthy! Isaiah who loves animals, even noticed how ridiculas it was.&nbsp; You could even get your picture taken while holding a baby lion...Isaiah said No Thanx to that! Again, this is all strange for him. &nbsp; I\'m still undecided about the weekend....Lisa wants to go down to Hotlanta. Not sure if I\'m up for it. Faith is coming into town, I should atleast hang with her for a bit.  &nbsp; I am definetly going to Church on Sunday. I feel like I haven\'t been in a while because of going to Myrtle Beach this past weekend, not getting back until late Monday which means I missed Bible Study too :(&nbsp;&nbsp; BAD BAD MAYA. &nbsp; Scott is going out of town for the weekend with his BROTHER! Sketchy? A lil, possibly because I don\'t know him that well and everyone knows I trust nobody!&nbsp; I have to work on that though... &nbsp; I thought I was right about Micheal too.&nbsp; It\'s still hard to deal with that. I miss him so much sometimes.&nbsp; It drives me nuts.&nbsp; I always question myself too...like why did I allow myself to get hung up on him?&nbsp; Mostly because he fed me tons of lines all the time too. UGHHHHHHHHHHH I need to just let it freaking go!&nbsp;  &nbsp; Anyway, I am going to get to the office get some work done so I don\'t have to go in on Monday and then I\'m going to have myself a fabulous weekend.&nbsp; NO WORRIES :) &nbsp; :) &nbsp; -Amaya &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; '),(6085,3935450,'15,July,2004','2004-07-15 00:00:00','    So I decided to create a new blog. I had too many people knowing my business. It wasn\'t intimate anymore.   About Last Night....  Erik called me very last minute and told me that Nadia\'s son, Chase, was injured in an ATV accident and I have to take over her client which meant going out of town for another 3 days :( I managed to convince Monica to go :) I just got home from South Carolina, I wasn\'t about to go! I was relieved to say the least-Thank God for Monty :)   After I ran my usual 3 miles, I went into Harris Teeter and bought some Ben & Jerry\'s Carb Karma in Half Baked--SO GOOD! While in the checkout line in my workout garb, in walks \'T\' whom I met a few weeks ago. Total stunner. Very handsome, has good values, established career as an Investment Banker and has just gotten out of a relationship with \'Nicole\' who he consistantly tells me that I am \"SO MUCH LIKE HER\", I\'m still not sure if thats a compliment or not. \'We\'re both sweet &amp; soft spoken & gentle\'--I am?!   He walked me out to my Maxx and we stood there and talked for a while when Rick Hilliard who Becca is trying to hook up with, pulled up beside us in the Envoy and he has some 16 year old bimbo affixed with a tattoo and of course the glamourous pack of Cigarettes in her hand. Lots of frizzy hair too--totally cute chic!!! Rick then came over and said hello, but I was pretty pissed at that point, the nerve! He &amp; Becca are trying to hook up and then he brings some Juvie in public with him. Low blow, I told Becca and she said she was done. The good thing is, I know when she says that she\'s done, she\'s done!   \'T\' then followed me back to my house and we talked on my balcony and just chilled until 2 am. We made out like teenagers. He totally turns me on physically, although I\'m not really into him otherwise :( I hate saying that but he\'s kinda blah.   Scott on the other hand has been extremely wonderful and has held my interest since Mannarino. He\'s so funny. Very sarcastic which I love. Very sweet at the same time. He likes the fact that I\'m a \'Good Girl\', although he\'s yet to see my halo!!! The best part is he has his own life aside from mine. And for Scott to help me get over Michael, he\'s doing good! God knows how much I cared for him. Sometimes I wonder how I fell so quickly for him. We talked about doing so much together--We wanted to do Vegas and get married for the pure hell of it! I was his keeper, We were peanut butter & jelly! I know now that he\'s happy and not even I could interfere with that. I always knew that he genuinely cared for her...Oh well. MOVING RIGHT ALONG!!!   My plans for the day include going to Grandma\'s and laying out and swim some. I promised Isaiah that he can go with me and even bring his friend. What am I getting myself into?! I invited Courtney over too we need to catch up on the wedding plans.   Haven &amp; I are going to take her to NYC for the bachlorette party in Mid August. The date is still up in the air. Theres so much to do...Lotus of course is my hot spot. I do like Club NY too. I\'m wet just thinking about it!!! Woot :)   More thoughts later....Onto to packing my beach bag!!! '),(6086,3935450,'17,August,2004','2004-08-17 00:00:00','    Since I updated this.  Since the 11th, a lot has happened, such as Dad\'s Birthday Party that was a total surprise for him.  He loved it.  It was wonderful.  Was fun for all.  I\'ve actually considered taking a position at a Private School here.  After being cornered by my mother and giving me guilt trips about the gallery, I think that has basically been shot down.  Daddy has been supportive and has told me to do as I want and he\'d support me.  I just wish Mom didn\'t have so much power over me...  Greg wanted to come over and spend an evening at my place, a sleepover, if you will....after discussing it with T, I decided that it wasn\'t a good idea after all.  He seems like a great guy and we could definetly hang out, it\'s just that the entire idea was a bit sketchy.   Adam & I have been in contact recently.  It\'s sick, but I really care for this boy.  I need to move on and leave the past where it is.  Thats so old.   Eric & I have been getting along rather well, as much as possible without ever seeing each other.   He had a date tonight with a chick to catch a flick...how cute is that?!  Scott & I haven\'t talked in a while...I don\'t know whats going on with him.  I think ever since he showed me so little interest, I think I lost taste in him.  It\'s sad because initially, everyone, including T, knows how much I was feeling him.  I won\'t give up, just yet though....  Married Mike and I talked a bit yesterday.  He went to Vegas over the weekend and partied.  He\'s heading to Mexico on Thursday too.  Hot trotting for sure.   I did a lot of shopping over the weekend, just enough to make me feel good...I\'m fiending for more trips to the mall.  Since fall is around the corner and thats the time to really tune into the fashionista in me...Can\'t wait :)    Enough for now.  Maya '),(6087,3935450,'11,August,2004','2004-08-11 00:00:00','    Okay, okay..he text messaged me first then I called him!  He\'s not doing well.  Seems like he has a pyscho on his hands.               Went to the office, left early, came home & made dinner...Becca made a Pina Colada cake...the bitch.  How am I supposed to stay away from that?  What was she smoking?  She knows I devour that shtuff.                T & I talked a lot lastnight.  I was being petty and aggravating at the same time.  Poor T!  The shit he puts up with from me!  Gotta luv him for that.                Scott & I had a heart to heart lastnight.  I told him what was up.  He\'s always neglecting me so I had to set him straight.  He acted oblivious, but thats just what guys do.  No sweat off my back.               So I\'m working on a project for Eric.  I\'ve come to realize how much fun Mr Eric can be.  He totally gets my sense of humor now.  Whereas before I always had to break it down for him.  Everything I say & do...he loves.  For example we talked today while I was grocery shopping and I told him I didn\'t remember where I parked my car and that I would have to set off the panic alarm in order to hunt down my car--the guy almost busted a nut.  I like that he gets my sense of humor, however I don\'t want him thinking I\'m a complete idiot.  I just have the beginging stages of Altzheimers!  Okay, so thats not funny.                As for Daddy\'s party everything is taken care of so far.  Sunday is the illumination ceremonies at the Cave, I have to be on hand to read some unimportant and useless info to a crowd of Chinese/Japanese people who don\'t even know what I\'m saying.  Basically it\'s just good PR.               I\'m going to take Emma outside and play for a bit. '),(6088,3935450,'10,August,2004','2004-08-10 00:00:00','    ~ Hung out with J LO & Derek Jeter all day :) ~ Mom says that Emma took my picture down off the coffee table and was yelling for \"Mini\" all day today...I came home, she had her hands on her hip, yelling \"Mini\"--That put a smile on my face. ~ I bought a couple ofT-shirts at Fair Oaks today that says Talk Nerdy To Me &amp; Fo Shizzle! ~ I chilled with Reece, Renee & Rod for a bit...fun times. ~ Got a text message from My Baby.  HOW THE HELL THAT HAPPENED I DON\"T KNOW!!!  T and I spent all night talking about him, and at 12:59 am he left a text of just \'What Up\'--I\'m thrilled.  ~ Scott just called, calling him back shortly. ~ DIdn\'t work out today :(   I SUCK.  I think it\'s time to...GO TO BED :) '),(6089,3935450,'08,August,2004','2004-08-08 00:00:00','    ~* I love weddings!  Especially hot summer weddings where the heat makes your skin glisten and you look radiant under the summer starry night sky!  When you\'re dolled up in the hautest Summer 2004 Bridal Couture in the perfect shade of Lavender. Yep, I live for those, and you can\'t forget about a haute man as your date. The kind that dresses stylishly and follows fashion trends...the kind that dances well to \"TURN ME ON\" by Kevin Lyttle.  The kind that wear your favorite cologne for men without even knowing it\'s your fave :)  Higher by DIOR.  You can\'t go wrong.  All of that combined equates for the perfect August evening.*~          Yesterday started off horribly.  I had to make a cameo at the local ER.  That was total fun, as you can imagine, but I didn\'t mind conversing with the intelligent, Oliver Peoples eyewear wearing doctors whom seem to know my own body better than Oui.  After indulging on a few hits of Morphine, interovenously of course and telling the resident Doctors that I\'m a regular, that I hit the ER a couple times a weekend just for a quick pick-me-up for the Percosets and the Morphine and managing to make them believe me for a quick second even after I told them I\'d already blown all of my veins due to everlasting heroin addiction...which did seem believable due to all of the bruises on my forearm because of my routine morning ritual of shooting a few cc\'s of insulin into my bloodstream, Dr. Gerfin did believe me for a nano-second until I managed to crack my Miss Congeniality 2000 watt smile.  I got so comfortable that I told Dr. Gerfin that I will be back next week too, only to just make him feel me up.  He lost total control over that remark and told me that I really must stop because I was making his side hurt...Of course that only lead me to tell him that I like playing Doctor too!  It was all in the game of fun.  He did a couple of name droppings of places that he frequents....I just may have to remember those.            The wedding was impecable.  Lovely. Great Music. Great Sermon/Vows even with the words \"OBEY\" in the vows....I personally would have to drop. Unless it is to Obey God, not my husband.            The reception was grand too.  Loved the decor, the flowers, the candles, the everything.   Tara was stunning as always. Beautiful Bride :)  Michael was a looker as well.  Total eye candy.            Now, I\'m sipping on a water with lemon.  I didn\'t drink too much lastnight, but I still have a horrible headache, the kind that could make you pull your hair out!  Yeah, not even. Not eva.           I\'m fully dressed in my best Sunday attire.  I\'m going over to my grandmothers with the video camera.  Tara & Michael are showing up at 2:00, Grandma has arranged for them to leave her home in a horse &amp; carriage and head to the Victorian in that.  How cute is that?!  It\'s the perfect afternoon for a carriage ride.            Mom, the lady who told me this morning that I wasn\'t nice--is ready and is standing at the top of my staircase yelling for me.  I feel like a child.  A child with boobs.  A child playing dress up in heels, with boobs.  I need to wear this dress more often.  I suddenly feel fullfilled.            I\'m off to see the wizard... '),(6090,3935450,'07,August,2004','2004-08-07 00:00:00','    Ugh, Holy Hell!  I woke up in excrutiating pain this morning.  Yesterday, the same thing.  It\'s my sides...they\'re killing me.        I have a headache from hell too.  I\'m going to go to the hospital before my hair appt. at 4:00.  I took 4 advil and that\'s eased the pain, on a scale of 1-10, I\'m at a 7 right now, whereas earlier I was an 11.        The wedding is today and this is a great day to feel like all hell.        Update later...probably tommorrow. '),(6091,3935450,'05,August,2004','2004-08-05 00:00:00','    However, I\'m not :)   Today was Mom\'s birthday.  I knew her birthday was August 5th, but it never crossed my mind today.  Dad & Mom showed up early at my house this morning, Dad asked what we had planned for Mom...confused at first until it dawned on me that TA DA!  Today\'s Mom\'s Birthday!!!            Dad was furious, but Becca, Haven, Ian & saved the day.  While Dad took Mom to play golf, I called everyone and made the arrangements. We all had dinner, caught up, and I even had a glass of white wine.  I knew I shouldn\'t simply because of the carbs,           We all chipped in for a Day Spa certificate and a Hermes scarf.  She adored it.  She asked for a freaking salad spinner and thats exactly what dad bought her!  Gay.           In other news,  I toured the Country Club today. Thats where I intend to host Mom & Dad\'s Birthday party.  I just have to cut the guest list in half.  That will be hard. We\'re at 102 now.  Including drinks, the dinner is @ $25 per head...which is sensible considering all the options and it includes the dinner wine, the cocktails, beers etc....Now, the hard part is up to me.  I get to cut the guests.  Fun fun.  I somehow always get stuck with the coolest job.                 I think that Eric is becoming interested in me. He confessed that he read the journal too.  Thats great.  Invade my life, please! Again, Thanx!          Scott called me at 1:00 this a.m.  I didn\'t answer.  I crashed early.  I just want him to persue me a bit.  If he is infact interested I\'ll be able to tell.  His signals are too weird and hard to read.            T & Nicole talked.  All I can say is (BECAUSE HE READS THIS)  is that it\'s about damn time.  They didn\'t get very far.  Except she knows exactly what he\'s feeling and I just hope for his sake that she doesn\'t take his vulrenablity for granted.  I think he\'d give anything to have her back.  Special prayers for T, for sure.           I thought of Broke Ass today, as always.  I realize now that everything that happend was a part of God\'s plan and he intended for us to cross paths.  I just can\'t wait until his face stops haunting me.  I can\'t wait until that day comes that I say Broke Ass who?!!!  I want to get over him so bad.  I know he never thinks about me,  I know that he doesn\'t spend this much time focusing on me so why do I do it for him?  I don\'t want to talk bad about him, he just simply was in love with HER...even after all the shit he said.  Did he really think it though?  Or was he really angry?  Who knows.  Why do I care?          Weekend plans include:  Wedding Rehearsel tommorrow then partying later with Miss B Haven...Saturday, Wedding madness, then ass shaking all night :)  Sunday is an afternoon tea for Tara & Michael.  My grandmother is hosting it.  Should be fun.  Grandma has great taste and always provides the best Raspberry Scones :)  Ooops, CARBS--FUCKERS.            I didn\'t run again today, I\'ve been too busy.  TOMMOROW for sure.            Going to bed now....God Bless.      ~A '),(6092,3935450,'04,August,2004','2004-08-04 00:00:00','    It\'s midnight and I\'m just getting around to updating my journal.  I just got home...      I went to the office today and took care of lots of dreaded paper work.  I was there all alone for most of the day.        I got a Brazillian today.  Josephina hooked me up.  She knows my pootie inside & out.  She don\'t miss a beat!  Whats bad is walking out after the waxing...PAINFUL.       Mom, Dad, Missy, Eric, Emma came over for dinner.  Mom cooked Mexican for everyone.  I had a low carb Mexican night.  Chicken tacos are tasty.  I could think of better foods, but better foods are loaded with carbs.  This will have to work until I\'m satisfied.        I walked through my door this evening and Emma went nuts.  She wouldn\'t let go of me.  Becca said that she was yelling my name all evening :)  Doesn\'t that just make your day??        I\'m tired. Like always.  I think it\'s going to be a shopping day for me tommorrow.  We\'ll see.        Off to dreamland........... '),(6093,3935450,'03,August,2004','2004-08-03 00:00:00','    Today went well.  I didn\'t do much of anything. I did get an increadibly sweet email from someone--Came out of nowhere!  I was touched, to say the least.   Snippet:   So I guess what I\'m trying to say in this message is that I\'m sorry for actinglike a one-dimensional jerk.  I pride myself as someone who can offer more to someone than just one aspect of his life.  I obviously have not proven this toyou so far.  I like you for so many reasons... your charismatic, your sense of humor is amazing, I find you incredibly attractive,you have an ability to surprise me where most people cannot, your sensitive without being over-sensitive and your confidence without coming across like a snob are just a few reasons that come to mind right away.  However, most of all I like you b/c you\'re genuine.  It\'s the most underrated characteristic intoday\'s world.  It\'s a quality that doesn\'t come to you overnight, it developsfrom the time we come into existence and not everyone has it but fortunately,you\'re one that possesses it.      How can I not smile to that?!   I got my tan on today.  My body is warm now.  Feels good.  I met up with Rebecca & Elizabeth for coffee at Starbucks.  I met \'Caleno\' there.  I think he\'s Italian not sure... He gave me his number, but being my brilliant self,  I left the number on the table!  Very rude, I know.  I could have put it to good use, but oh well.   T & I are on day 2 of Zippy carbs.  I\'m going strong.  I haven\'t gotten an update on him yet.  We have to stay in contact, he motivates my ass!   Things to do tommorrow: ~Pick-up dry cleaning ~Return calls, tons even ~Do paper work, contracts, invoices, etc. ~Call Brandi ~Narrow down the dates to Tara\'s wedding...IMPORTANT   I\"m going to bed now!  Thuper Tired :)   Amaya '),(6094,3935450,'02,August,2004','2004-08-02 00:00:00','    Today was the funeral and it goes without saying how devastatingly painful it was for the entire family.  Kat is just nuts. Hearing her scream \"Don\'t leave me, don\'t let them take my baby\" left me sobbing aloud. Her heart is broken.  She\'s a beautiful strong woman and to see her down on her knees and in pain, it kills me.  I pray she\'ll be okay.  She has her boys, Logan & Ethan and they will help her get out of bed every morning and take on the world.  I\'ll keep praying for strength.  There was over 100 people in attendance, there were people standing in the Vestibul (SP), he was loved by many thats for sure.  But finally he\'s no longer in pain and he\'s at peace. If he could come back to earth, I\'m sure he wouldn\'t want to.  Gosh, I can\'t believe he\'s gone though.  I just realized that. He is gone.  Wow. What happens next? Nothing will ever be the same.  What happens at Christmas? Holidays? Just wow.            It all brings back the painful day in January of 2000.  A day I\'ll never forget.  I still remember the music, I still listen to it occasionally.  It keeps me grounded.  I sure do miss him.  Everyday. All day.  Everyday I shed a tear for him and I wonder if I\'ll ever get over that? He\'s watching over us now though...     \"When my life is almost gone, hear my cry, hear my call, hold my hand, l\'est I fall\"            ---------------------------     On to something else...          Damn you, Induction!!!            So, T & I have decided to do a ketonic diet.  We\'re staying under 20 carbs per day. It\'s basically the Atkins Diet.  We\'re going to work out fiercely too :)  200-400 sit ups a day along with a protein shake or 2 ;)            Today has been rather easy, simply because I\'ve been so busy...I did infact have lunch, I\'m such a nerd.  Today has been a complete blur and it\'s definetly been a long one.            I had a big grilled chicken salad with Blue Cheese dressing for dinner and it was rather fullfilling.            I\'m estimating that my total carb intake, WITH lunch is about 15 Carbs.      I\'ve had three 33 oz.s of water.  The water is good for my skin too.  Love water lots.  Gotta have the water.           Things to do tommorrow:     ~Amber\'s Scrapbooking party @ 6:00     ~Quick meeting with Qyweneth @ Allison     ~Mani & Pedi, maybe--no appt yet.     ~A lunch date with my Aunt Carmen          My day isn\'t going to be busy at all, Thank God!          We\'re making the plans for Courtney\'s Bachlorette Party in NYC, Can\'t wait.          This weekend is Tara & Michael\'s wedding.  Our dresses are super cute.  She\'s doing Pastel Colors, I\'m in lavender.  I love my dress too!           The following weekend is Daddy\'s party.  He has no idea yet :)  Total Surprise          I\'ve been dreaming of my wedding since I was 10!  Of course the color theme will be all different shades of Pink...from a Blush to a Mauve to a Coral.  Flowers will be white Roses with some diamonds in the bouquet. I want a summer wedding, on the beach...at Sunset, no shoes, and I want the waves to crash at our feet...You can say I\'m a dreamer, but I\'m not the only one!           Highlight of the day: Seeing all my family! Love my family :)           T emailed me this: \"amaya, you\'re my inspiration\" How sweet is that?  This guy is a total cupcake.  So sweet.           Time to get into my jammies :)               Thanx for all the sweet IM\'s, Emails and Prayers, everybody!  God Bless you all!!!          ~*Amaya *~                    amaya, you\'re my inspiration '),(6095,3935450,'01,August,2004','2004-08-01 00:00:00','    ~ I\'m super happy.  I just landed a huge deal with UPS :)   I don\'t have to work the entire month of August :)  KIDDING.  Sorta.       ~ I went shopping, got some cute skirts and I went nuts at Victoria\'s Secret too.    ~ Had lunch at Qudoba with Justin, Mark & Richard, I even bought....They were so easy to chill with.  So laid back.  Seems we all like the Yankees.  Good guys, can\'t wait to work with them!      ~I talked to Eric, he doesn\'t know who Jimmy Choo is!!!  Never heard of Manolo Blahnik either--what kind of man is this?  Sheesh.        ~I\'m getting ready for the funereal now...it starts at 3:00.  Say a prayer for us.  We need all we can get :)         God Bless & God Keep '),(6096,3935450,'01,August,2004','2004-08-01 00:00:00','    When you don\'t have to watch your family grieve.   Went to Church, had lunch @ Grandma\'s.  Then we all went to Kat\'s and spent a lot of the evening there.   Emma now says \'BATH\' .  She\'s so smart.  I painted her nails again.  She\'s so adoreable.  I love her so much.  Tommorrow is the funereal and I don\'t know what to expect.  Funereals are great to begin with....   Sheesh...  K...on to finish some work, no work for me again until I say... However I have a huge meeting with UPS tommorrow...it\'s so mine already, but we shall see.   Goodnight :) '),(6097,3935450,'01,August,2004','2004-08-01 00:00:00','    I woke up in a great mood.  I\'m on my way out the door to Church :)   Thoughts for the day: \"If somebody loves you, won\'t they always love you?\"       Good luck to T....He\'s going to go see about a girl ;)        I think I\'m going to change my outfit again.... I\'m too sexy for my shirt..too sexy yeah!          XOXO   A to da M to da A to da Y to da A '),(6098,4184856,'17,August,2004','2004-08-17 00:00:00','   I think i may be in love. I\'ve known her since I was a small child and am I  finally  realizing it? No I\'ve had inklings before but this...Let me explain. I was up at Mick Connelly\'s lake house after his son\'s Bar Mitzvah. We were all swimming and being happy and I was sad when she left and explaining is moot beyond this point. Best of friends but one\'s a little bester. '),(6099,4184856,'10,August,2004','2004-08-10 00:00:00','   Just read a couple pages of a book I\'d read before then felt blessedly tired but didn\'t go to sleep. I can never get to sleep these days. I recently had a dream where I would walk down this trail next to a boggy marshy little place to get to open bay water and climb onto the dock. I couldn\'t climb it though because of the nasty spiders hiding underneath. Pity. Also i remember something about a rich house. A large house...maybe. I forget. All of you shut your mouths and fetch me the...go fetch. '),(6100,4184856,'09,August,2004','2004-08-09 00:00:00','   There\'s nothing worth telling! Why can\'t i just spin right back into the world of lies i\'m used to. It\'s hard to be so boring. Oh Well. I\'d better work on my story. '),(6101,4184856,'09,August,2004','2004-08-09 00:00:00','   I\'ve recently written to Hollywood unknown Annelise Van Der Pol. Check it fools.  Anneliese, I am writing to say hello and to ask a favor of you. Iâ€™m not really a fan of â€˜Thatâ€™s So Ravenâ€ but, I do watch the show (if only occasionally and due to a couple of younger sisters in the house). I believe you to be one of the better actors on the show and find you quite attractive as well. Perhaps you should pursue some better acting roles. The show is great and very entertaining, but I would love very much to see you in something serious, something real. Youâ€™re very beautiful, and good at acting, from what I can see. I donâ€™t think finding roles would be a problem. Now, onto the reasoning as to why Iâ€™m really writing. Most of the mail you get is probably all asking for autographs and proclamations of love for the show. And though I would like an autograph, Iâ€™m requesting it, if you would, in a different manner than most. It may sound a tad bizarre, but stick with me. I would like a checkâ€¦for as little as you see fit, something like a quarter or a dime. It is an autograph, technically. However, you wonâ€™t find the amount taken out of your account, as it will simply by mounted and stuck on my wall. To me, the novelty of an autograph/check is insurmountably entertaining. In all probability, this letter with be ignored, as Iâ€™m sure it has been delivered stifled among hundreds of others. Youâ€™re also bound to have preparations for the show, etcetera. But if you do answer this letter, especially with the special sort of autograph I have requested, I will be sure that a kindness of that level does exist in the world, and will praise you for it. Thank you for simply taking some time out of your day and reading this, if nothing else. Much luck on your future endeavors, be they acting related or otherwise. Please and thank you in advance, and hereâ€™s hoping. With love,              Adrien Darcy, age fifteen of Upper Montclair, New Jersey.  P.S.I have noticed online that you charge a fee of $5.00 for an autographed photo. Seeing as this is a bit more elaborate, I have included a sum of $10 as a sort of deposit (should the check should fall into the wrong hands and be cashed) and for any personal inconvenience it will cost you.    Noticed i lied about my name....Well what would you have done? I can\'t have hollywood get wind of me.    I\'m lonely. '),(6102,4184856,'09,August,2004','2004-08-09 00:00:00','   Bon soir mes ami. I\'ve found that living a lie is far more fun than surviving in the world you\'re provided with. TALKING about such a world...well it just turns my stomach. But I\'ll do it. And you are going to be here to watch. '),(6103,3558845,'22,June,2004','2004-06-22 00:00:00','    well, i have an unexpected long break before my next shift at six, so i thought i would write some.  first of all, this blog is not written for eric or anyone else.  it\'s for me.  there will be one paragraph in today\'s blog that includes anything personal.  that\'s it.  the decision that dan and i have made together will not be discussed.  it\'s no one else\'s business.  all i really want to say about the whole situation is that i\'m disappointed.  despite many attempts, i couldn\'t get it through someone\'s head that reactions are not misunderstandings.  the comment was made that i chose someone else because he was in a better financial situation.  how sad that someone that claims to know me so well thinks i\'m so superficial.  ironically, that was one of the main reasons dan wanted to be with me--because i wasn\'t superficial.  i have many other comments to make about a certain post on a certain blog.....but i\'m not going to.  they\'ve all been made before and you can\'t make someone understand who doesn\'t want to.  so it doesn\'t matter anymore.  c\'est la vie.  time to move on.....  i actually get a day off tomorrow.  i was going to get my hair hightlighted, but i don\'t think i should spend the money yet.  i have rent and bills coming up, so i\'m gonna try to do the responsible thing.  i am gonna go jump in a tanning bed for a little while, though.  haven\'t tanned in a few days.  having withdrawals.  i hope everyone has a wonderful night!   '),(6104,3558845,'18,June,2004','2004-06-18 00:00:00','    please read the next blog all the way to the end, even though you might think that it\'s going to hurt, just from reading the first couple of paragraphs.  the rest of it might even make sense.   '),(6105,3558845,'12,June,2004','2004-06-12 00:00:00','    i have alot to talk about.  i actually have the afternoon alone, and i\'m welcoming it with open arms.    i\'m going to be completely honest with you.  i hope you can appreciate that, and not fly off the handle over the first couple of paragraphs.  i have alot more to say after that.  dan and i have been together for about six weeks.  i was afraid to tell you for one very important reason--i didn\'t want to get a phone call in the middle of the night saying that you had done something to yourself.  i\'m still going to be afraid of that.  but now i\'m going to have to trust God to take care of you.  you asked me if i was in love with dan and i said no.  i wasn\'t exactly lying.  i\'m just not sure.  i definitely love him, and i\'m happy being around him, but i\'m not sure if i\'m in love or not.  we have pretty much the same conversations every day--which are about you.  honestly, he\'s not happy about my interaction with you.  you can see why now.  he has alot of insecurity and stress about that, and it\'s affecting our relationship.  i told him last night that, either way, i would lose a friend.  because of his decision not to be around you (which he still stands by, and i have to respect that), if i chose you, i would lose him.  if i chose to stay with him, then i would feel bad for not respecting his wishes, and i would lose you.  i feel trapped and very stressed out.  there is alot of love there, but dan and i have not had the picture-perfect relationship so far.  it\'s really weird--while i crave solitude above anything else right now, i\'m afraid to be alone.  i don\'t depend on you or on dan for my happiness, but i\'m still scared.  dan is constantly reading my face and asking me what i\'m thinking, and most of the time i don\'t want to tell him.  i just want to think and keep it to myself.  but when i say \'nothing\' he automatically thinks that i\'m thinking about you.  then he gets sad and even more stressed out.  and that, in turn, does the same thing to me.  it\'s a vicious cycle.  what stresses me out is the promises i made to you that you\'ve brought up.  i promised when we broke up that i would welcome you back with open arms.  trust me, i did not see everything coming that has happened--dan falling in love with me, your reactions, my feelings toward them--none of it.  that doesn\'t mean that i\'m trying to release myself from my promise.  i\'ve cried on way more than one occasion over it.  i promised you that i would never turn my back on you and that i would always be here for you.  i cry every time i think about that.  i\'m crying now.  because i feel trapped and that any decision i make will end up hurting someone so much.  i\'ve been allowing dan to love me and take care of me, so in a sense i feel guilty because i feel like i replaced you too soon.  i know you\'re probably thinking about that, too, and hurting.  i\'m so sorry.  before you and i were together, i always looked at dan and saw someone that i could be perfectly happy with for the rest of my life.  he was confident, straight-forward, secure with himself, and had his priorities in order.  he also knew exactly what he was looking for and didn\'t depend on anyone else for his happiness.  all of this is what was going through my head the night he told me that he loved me (along with shock).  i\'m afraid that everything that\'s happened has made him different.  i never saw him afraid before.  never saw insecurity in him.  i\'m afraid that it\'s made him a different person, and i\'m afraid that he\'s relying on me too much for his happiness.  he watches me all the time.  he watches my face, and if it changes, even if i\'m not thinking about anything important, he gets worked up because he thinks i\'m upset or that i\'m thinking about you.  and when i try to talk to him about some of these things, he gets defensive.  so i don\'t know what\'s going to happen.  but, just for clarification, we are in no way, shape, or form sleeping together.  dan will stay a virgin as long as he\'s with me, until he gets married.  i don\'t know what else to say.  i hope that i\'ve made the right decision in telling you all of this.  i felt the same way dan did--he felt like he was lying to you by not telling you.  he was just alot braver than me and told you a long time ago how he felt.  he\'s worried about me telling you this because he\'s scared that i\'m going to get the same reaction from you that he got.  i personally think that i trust you more than that.  i know you won\'t ever physically hurt me.  but i\'ve told the absolute truth now.  i\'m not hiding anything, but i hope this blog stays between me and you.  if you need to talk about this some more, we can.    i just feel like i\'m not going to be able to make anyone happy.  on one side i have you telling me how you feel and that hopefully we\'ll be able to get back together, and on the other side i have dan accusing me of never being able to let you go and telling me to move on.  i feel so trapped.  as far as me and you....i don\'t know.  i miss you alot.  i love you, and the idea of you lying in bed and crying and trying so hard to get me back and hurting so much hurts me more than you can imagine.  but i don\'t know how close i can be to you right now because i honestly don\'t know what i want from you.  and i don\'t want to encourage you and give you hope and then realize that it\'s never going to work out.  i don\'t want to hurt you any more than i already have.  the idea of not being able to be there to comfort you at night hurts me.  especially when i have someone with me.  i\'ll admit it\'s probably because you\'re stronger than me.  what i\'m sorrier for more than anything is that my weakness caused you pain.  please forgive me for that.  i just don\'t know what to do.  but i want you to know that you\'re not the only one hurting and crying.  someone else is sharing that with you.  even though i have someone with me, it\'s not someone who can completely understand....it makes me feel very alone.  the only one who can understand is you.  please continue to pray for me, and please be patient.  i love you, and i miss you.  amber   '),(6106,3558845,'12,June,2004','2004-06-12 00:00:00','    i\'m finding it easier to post a new entry on my own blog than to post alot of comments on yours.  i hope you don\'t mind.  i\'m sorry you couldn\'t sleep last night.  i went to bed really early and actually got plenty of sleep.  hopefully you fell asleep and you\'ll get pleny of sleep before work.  i hope you make alot of money tonight, too.  i turned in all my health insurance stuff yesterday and went on a little cleaning frenzy around my house.  gave me a little sense of accomplishment.  i had a really good food-running shift last night, too.  i have the weekend off and my bar and wine class on monday.  take my last test on tuesday and hopefully be on the floor on my own by tuesday night.    \"For God has not given us a spirit of fearfulness, but one of power, love, and sound judgement.\"           ~2 Tim. 1:7  i read that yesterday and thought of you and dan both.  i can\'t think of anything else i really need to talk about at the moment....i\'m sure i\'ll post again later.   '),(6107,3558845,'11,June,2004','2004-06-11 00:00:00','    wow.  haven\'t done this in a long time.  since last summer, anyway.  and we both know how that ended up.  so you\'re the only one who\'s gonna know this address.    first of all, i just wanted to let you know that this isn\'t always going to be a serious blog.  half the reason i wanted to start this was for me--to remind myself of things that i want to personally accomplish.  it isn\'t always going to be about you--when it is, you\'ll know.  but you won\'t always know what i\'m talking about, since i intend to write about pretty much everything that\'s on my mind and everything that\'s happened to me.  i started thinking today about how much i needed to get my life back together.....just little things that i\'ve been putting off because i haven\'t taken any time to myself.  i started trying to fix that situation by cleaning house today.  speaking of which, i finally had to throw my flowers away.  they had just about had it.  there are so many other little things i need to do--i need to stop by the financial aid office at letourneau and talk to jennifer about starting to pay back some of my loans. (i mean, all i need is another monthly bill, right?).  i need to finish studying for my marketing and finance finals so i can take them and be done with that.  all these tests at carino\'s are taking up so much time.  and i\'m really missing my parkeway membership.  truthfully, i\'m missing outback.  i don\'t like the people at carino\'s at all.  people down there ask me how i like it so far, and i have to hesitate and plan my words before i answer them.  i just wish i could get my life back to normal.  i didn\'t miss waiting tables at all during the last two weeks when i didn\'t have a job, but i hate having to stress about not having my own money.  i hate depending on someone else to take care of me.  i know dan didn\'t mind helping me at all and he would do it again in a heartbeat, but i hate feeling like i\'ve lost my independence.  i\'m irritated because at the rate carino\'s is going, it\'ll be next tuesday or wednesday before i\'m on the floor by myself, and it\'s getting to be time for bills to start showing up again.  grrrrr...  i also started thinking about something you had mentioned in one of your posts....it was something i told you first.  if you love something very much, let it go.  if it doesn\'t come back to you, it was never yours to begin with.  if it does come back to you, it\'s yours forever.  i\'ve been encouraging you to let go so things can get better and we can live our own lives seperately, when i\'m the one who\'s having trouble letting go.  i want to apologize for that.  i understand now what you meant when you told me that i was selfish.  i\'m so scared of letting you go because you\'re such a big part of my life and because i\'m scared of what my life would be like if you weren\'t in it.  you\'ve been my best friend for a long time.  i don\'t want to lose you, but at the same time i know we can\'t be together because i know we wouldn\'t be happy.  i just don\'t know what to do.  i guess i\'m just scared and selfish and i need to grow up.  hopefully telling you all of this is the first step, because i\'ve been terrified of telling you what\'s really on my heart.  i\'m scared of your reactions.  the thought of you being sad or upset or depressed and knowing i\'m the cause of it hurts me more than i know how to handle.  i absolutely hate the idea of hurting anyone.  but you told me once that you would rather i tell you the truth, and that you wished that my unwillingness to hurt you wasn\'t a factor, so i\'m trying to give up my way of doing things (which has been a failure and hurt more than one person) and focus on your way of doing things.  i also know that i haven\'t focused on God enough, and that i haven\'t been trusting Him to take care of you or dan or myself, for that matter.  i\'m going to focus more on trying to change that.  it\'s all part of the healing process, i guess.  it\'s a hard thing to accept that none of this (including your feelings and dan\'s feelings) is in my control.  i\'m going to have to focus more on trusting God and doing what He wants me to do.    right now i think he wants me to focus on laundry.  so i better go, since my work clothes are in there.  i have my food running shift tonight, and then i have to come home and start studying for my bar test.  i\'m so happy for you right now.  i\'m so glad that sam did what he did....it really was an answer to prayer.   '),(6108,3558845,'06,June,2004','2004-06-06 00:00:00','    hopefully this works......   '),(6109,3306314,'28,May,2004','2004-05-28 00:00:00','   I havent been on here in awhile..been pretty busy. I guess. But not really like i have a life. hahah obviously not if im sitting here at 12 19 am and on a friday night. I hate my life. all my other COOl friends are probably hooking up in bars. im sitting here just counting the cars. not that cars come down my road very much but if they did i would probably go count them. but instead im sitting here\"on the couch\" watching the the south park reflection on my dave matthews poster. bored off my ass. but high so. its ok. im thinking about watching my Ryan Adams DVD..he is the greatest man...besides ive seen the episode of south park already...not exactly fun...nothing much is very funny these days ..not really sure why..i have grown accustomed to being depressed..which isnt a good sign...but it gets old just putting on this mask thinking everything is ok. but deep down you really want to do is scream your head off...am i practicing good self control or is it that i am slowly hurting myself until i cant fall anymore and lash out irrationaly at the closest person near me at the time of 15 yrs of aggression, resentment, hostility come out..God I hope not. better off to keep it to myself..because most of the time its me at fault not anyone else and it sucks...im tired of being at fault..or thinking im at fault..and even when im probably not at fault i feel like i am because of been raised by a person which did no wrong and I was always at fault no matter what. im 24 yrs old and im still always at fault. on a brighter note Ryan Adams is the greatest.   Go on little girl feet twirl, go and make him smile, go on like the rumbling drums in the march of time and when the father meets the dad it was be something better for the pain. But you will always be the same  --Ryan Adams  Tomorrow is another day Hopefully it wont end this way Im tired of being alone and sad and having to pretend i am glad  Maybe one day it will be a diferrent way I dont know hopefully soon Kind of like a butterfly coming out of a cocoon Pretty as can be One day that will be me.    So ya never know Keep your head up, just go with the flow. Next time you will run the show and dictate how it will go  i dont know rambling on.  '),(6110,3306314,'12,May,2004','2004-05-12 00:00:00','   Well, I doubt anyone will ever read these but im still gonna post em. Dont really know what to write or type for that matter im sure eventually i will become a babbling idiot on here and people will be like what in the world is this kid talking about..my day was pretty good..work was busy but i liked it..makes me feel better and i try and learn stuff to become smarter at my job and a well rounded person what a dork.   Smallville season finale tonight cant wait.. I love the WB good quality family entertainment for the \"US\" generation damnit that should be a slogan for them. anyway i love quotes im going to try and put one on here a day and eventually get to the stuff that ive written in the past..we will see. if i get the balls to do it..  \"Fighting is like running in the special olympics, You win but you still look like a retard\"  '),(6111,3306314,'11,May,2004','2004-05-11 00:00:00','   Well, This is my first day actually being on this thing. Its pretty cool.. My friend Bobby hooked me up. His name is project_mayhem on here check him out. I\'m gonna try and write on here everyday.  I dont know how good i will be at it though..thanks for checking it out though. I hope to get a pic on here soon.  '),(6112,3306314,'16,June,2004','2004-06-16 00:00:00','   Before the curtain finally closes The last scene of the night Ill take my bow  the lights go dim and exit stage left Now you are the only one left  while you are picking up the roses pick up my heart  i think i left it right from the start maybe it was closer to act 3 i dont know who cares which one it could be Its all over  Before the curtain finally closes Ill take my bow As the lights go dim I exit stage left You will be the only one left  I gave you all i had to give You just throw it back in my face Im sorry i wasnt good enough but making you happy was just too tough Im not taking all the blame for this curtain call Im not the only one at fault You had the starring role  As the crowd applauses your performance I take a bow  The lights go dim I exit stage left You are the only one left.  while you are picking up the roses pick up my heart  i think i left it right from the start maybe it was closer to act 3 i dont know who cares which one it could be Its all over  As the crowd applauses your performance I take a bow  The lights go dim I exit stage left You are the only one left.   (this is my TRYING to write)  '),(6113,3394208,'05,August,2004','2004-08-05 00:00:00','   Chapter 1: Unintended Meeting  It was around 12:00 am on April 19, 2004. A rain cloud hung over the city of Makati as I walked these dirty desolate streets... Ayala, where all the well known malls are to be found. Always found to occupy thousands of people shopping, dating, eating, conducting buisness and whatever normal people usually do in their everyday lives... but it was midnight, and I was walking this street all alone, accompanied only by the few cars driving off home and a few bums here and there lying on the street. The lights guide the way for me, Edsa is where I\'m headed. My head drooped lower as the rain poured harder on my back. I was freezing, walking half concious not noticing the scenery around me... what\'s there to see in this city anyway? There are just massive buildings that are found here anyway, buildings used for office, or residential purposes. It\'s a shame I don\'t know too many people living here except for my big brother and a couple of friends I barely know. I stopped by G4\'s lobby and shook myself dry.  I reached deep in my jean\'s back pocket and took out the brand new pack of Marlboro red\'s; which has slightly been wet by the rain; took a stick and lodged it in my lips then gave it a light. It was a sad day for me. I took a deep long puff from my cig and I tried to clear my mind. I was soaked. The street lights blinks at me furiously a couple of meters away from me and all I could hear was the pit-pattering of the rain as it fell on the hard asphalt road. I wasn\'t alone after all on that night, 3 or 4 people were there with me just wanting to escape the rain. One of them was an attractive young woman that appeared to be in here early twenty\'s. She was wearing a black hoody sweater with the words \'I like boys with tattoos\' sewn in the front. And she had a tight dark blue bell bottom pants on. She looked cold; as did the guard who was probably on the grave hour shift; I didn\'t really pay attention to the other people. I just gazed at the street shown in front of me. The rain is actually a beautiful sight, it\'s a shame it doesn\'t feel that way when your under it.  I figure the rain would last longer than usual so I slid down against the wall. Sitting there as I watched nature do it\'s work. I took another puff from my cigarette, turned my head up high and blew a smoke ring at the ceiling of the mall\'s lobby. I was thinking of what my consequences would be once I get home. Would anyone find out I skipped from the house to get hammered? Most likely. I just hoped no one came down to check up on me. Whilst I pondered on, I overheard the woman in her early twenty\'s cuss and fight over the phone. I just stared at her in a dazed way and didn\'t even bother to make out the conversation she was having. She caught me staring at her, and she turned away and moved a bit farther from me. Paranoid bitch. I stared once again at the ceiling. I wondered on what it was made of... was it marble? It looked like marble, too bad it\'s high up for me to go and touch it\'s surface. The guard paced back and forth at his space and rambled on his walkie-talkie. And as for the other people they have gone into a taxi and drove off for home probably... \"Why don\'t I get a taxi\" you might be thinking if you ever met me in this situation. I\'m broke.  I stood up from where I was seated and threw the cigarette I had in between my fingers. The woman wasn\'t on the phone anymore, but she still seemed pretty upset as she reached for a cigarette, it appears that she didn\'t have a lighter, cause she threw a short fit. I came up to her. I whipped out my lighter and offered her a light \"Rough night eh?\" I said as she reached over to light her cigarette. \"Yeah pretty rough night... and you?\" she replied. \"Same here... alright then, I\'ll be back over there.\" I said as I put my lighter back in my back pocket. I didn\'t know why I did that but it can\'t be helped, I already done it. So once again I sat in the place I was at earlier staring once again at the street... it was quite hazy now as the rain poured harder... The woman came over to me, and asked if she could sit with me. I was shocked. You never ever find people like that over here, especially in the streets, sure at the clubs you could, but where I was sitting, never. \"Sure\" I said as I scooched over...  I leaned my head against the wall as she slid down beside me. Once again I found myself staring at the ceiling thinking, that the wall must be made of marble, or if it isn\'t, maybe it\'s black smooth tiles that look like black marble... idunno... My pondering has been broken by the woman\'s voice... \"Ok, so tell me what are you doing here at a time like this?\" She asked me with much ease... \"And my name by the way is Sarah\" She followed up... \"Me? Idunno I figure it would have been a good walk in the rain, broke...\" I said in reply to her question... \"Oh ok... and ur name is?\" Sarah asked showing a bit of dissapointment... \"Would it even matter if I give you my name? I mean we won\'t probably see each other again anywhere anyway...\" I said blurting out a short sarcastic laugh as I shifted my stare at her. \"Fine, be sarcastic... I mean it was you who first opened up...\" Sarah said as she flashed a face of distaste... \"Alright, I\'m Chris... nice to meet you\" I said in a jokingly manner as I reached out my right hand for her to shake... \"Nice to meet you too\" She said as she took my hand and giggled a bit...  \"If you don\'t mind...\" Sarah said \"I was wondering why you\'re broke...\" My God what is this? An interview with the local folk here? I felt like I was in an internet chatroom. \"I\'m not sure I can tell you that information, I mean you have not yet told me your reason on why your out here.\" I said as I looked at the street again.\"Ok fair enough...\" she started \"I got too rowdy at the bar and my friends left me.\" Sarah finished... \"That still doesn\'t explain why you\'re still hanging around here, I mean there are a few unoccupied taxis passing by here, you could\'ve gotten one already and head on home\" I said showing signs of distrust as I took my pack of cigarettes out... \"Ok, if you want to know the whole picture that much I\'ll tell you then...\" she said flashing me that distasteful look again... \" I live at my friends condo, and since I went overboard at the bar earlier they left me there, and the reason I can\'t go home now is I lost my set of keys.\" She finished again... \"Can\'t you just go there and just ring the doorbell constantly then?\" I said in reply as I lit my cigarette... \"You don\'t know how my friends are like... nevermind, you\'ll never understand.\"  The rain continued pouring, I was beginning to think this isn\'t just a normal rain but a typhoon. Sarah continued talking, and I just kept nodding at every question she laid on me, till finally... \"Are you even listening me?\" She said... I just simply said \"Yup\" \"Ok what was I sayin earlier then?\" she asked me \"Uh... you were talking bout ur friends bein this and that and that they\'ll never let you in whatever you do etc... look I get you...\" I told her in reply, she just said \"Ok then... that\'s my story on why I\'m here... and now it\'s your turn...\" I was just thinking whether or not I should go and tell a stranger my situation, but I guess it\'s only right since she seems to have spilled me her story... \"Ok...\" I said as I took a long puff and blew 2 to 4 smoke rings...   \"My reason of being here...\" I started off... \"is cause I overspent at the bar, nothing amazing about it, I just drank and gave all my buddies a treat, the thing is here though, they left me there.\" I took another puff and I noticed Sarah reaching for her own pack of cigs... \"I mean I just went off to the restroom I came out they were just gone... the only thing they left me was a bill of all the extra orders they had...\" \"So how were you able to pay for it?\" Sarah asked... \"I didn\'t, well not all of it, I gave em all the rest of my cash and I told em I\'d pay them back\" I told Sarah as I lit her cigarette.       The rain slowed down, so I stood up and started walking. \"Where are you going?\" Sarah asked me as she too rose from the floor... \"The rain isn\'t pouring hard anymore, so I\'ll continue my walk home... I suggest you go find a taxi and head on home too, or wherever you can sleep tonight\" I said as I turned to answer her question... \"So you\'re gonna leave me w/o saying goodbye or giving me a number?\" Sarah said as she folded her arms... hmm I can\'t believe she\'s sayin these things, is it my charm? is it cause I took the time to talk and listen to her? I don\'t know, but whatever I did made her want to see me again... well in my perception of the whole scenario... \"Ok... sure, why not, here\'s my number...\" I said with a smile, and she noted my number in her phone. \"Thanks\" Sarah said flashing back a smile at me... \"You take care then Sarah, bump into you some other time...\" \"You too...\" Sarah said in reply, at this I turned and faced the cold wet street. I was once again walking home... '),(6114,3394208,'27,May,2004','2004-05-27 00:00:00','   Who was outside watching over a slow dying rat? was it me? I couldn\'t tell... As it was lying down on the hot hard asphalt covered street, it made me wonder... shall i share the same fate? A street cat appeared, watching the rat die in the corner of it\'s eye... already the cat is in it\'s position to pounce at the slow dying rat, and again, it makes me think... Is someone ready and waiting for me to die? Already is someone in his or her position to pounce? Ready to feast upon my soon lifeless body? Damnit! Someone always has to be on top of the food chain... and how i love how it works... Rat at a corner dying... will it die peacefully? Who knows? '),(6115,3394208,'25,May,2004','2004-05-25 00:00:00','   I\'ve walked upon these streets before... damn... why am I here again? I guess now, cause I\'m not walking but running... running towards something or away from something... the rain hasn\'t hit here yet, or has it ceased... I\'ve never been so alone... and I\'ve never been so alive... sunny days... burning... something\'s burning within me might it be my hate again?  or something else is fueling this burn... damnit...  running, such freedom... no one stopping me... no one in front of me... running wherever i want to... this fire burning, happiness... such happiness... hardly ever thinking of the cold dark pit i was in anymore... i\'m not in it anymore, it\'s not gloomy anymore, it\'s just so sunny... What?!? '),(6116,3394208,'23,May,2004','2004-05-23 00:00:00','   Fastballs coming \'round my way Trying to evade each one everyday Every single package comes to me so simple Every step must be quick and nimble Took two steps back and slipped back to the past All the good times and fun didn\'t seem to last I got hit with a hard one, a sharp pain to the head The fastball knocked me out, i thought i was dead I awoke from the ditch which i thought was a grave Not once did i think that i would be saved Feeling as though i\'m going against the tide losing my strength from each pushing stride i finally reached atop the ditch I saw myself on top, there must be a glitch I said to myself \"I threw that fastball\" \"I threw it at you, so I\'d make you fall\" As I listened to each word Id said I realized that only i can mess with my head... '),(6117,3394208,'23,May,2004','2004-05-23 00:00:00','   I seem to space out almost every single day Dont\' seem to care about what comes my way A collapse into a void still and silent as a droid Money comes around I throw it away Everything else I feel the same way A collapse into a void, still and silent space Staring at myself i see a blank, unknowing face The void grows bigger every single day If i continue staying this way '),(6118,3394208,'23,May,2004','2004-05-23 00:00:00','   Been staring at the clocks that seem to stare back Days seem to fade into the back of my mind The sands of my time drag on pulling me throught the ground Memories appear out of nowhere pices scattered in my head Been staring at the clocks that seem to drag on Seconds feel like hours it\'s been to long Can\'t fall asleep, falling fast at the state of insomnia Can\'t open my eyes easily drowning in eternal slumber Been staring at the clocks that warp my mind Time pulls me in losing sense of reality Reality is quickly turning into fantasy once again And fantasy is turning into reality '),(6119,3394208,'23,May,2004','2004-05-23 00:00:00','   Been waiting too long Sitting in my room Move to the city Meet new friends See everything And see nothing at the same time Experience the Big N quickly filling me up Watch me Continue floating around The Big N Brings me down Nothing again I feel it Coming into the room Better things Will surely come my way The Big N Makes you want to stay Everything matters the Same time doesn\'t Big Nothing Big Nothing Big Nothing Another chapter closes '),(6120,3394208,'23,May,2004','2004-05-23 00:00:00','   Sitting at my porch Staring at the star-filled sky Smoking a cigarette Listening to some rock It makes me wonder makes me wonder You doing the same thing? Busy with poetry taking a break  As i look up at the sky today I feel a greater good staring back at me Fills me up and makes me wonder Are my dreams being shared with someone else?  The songs stuck in my head playing over and over Taking sips from my iced tea I\'m pretty laidback Chillin around smoke myself a pack Walking to my room Fall on my bed all feeling sleepy Slip in some CDs volume\'s full blast  I stare at my ceiling Sounds from my system fill the air Filling up fast The sounds forever they last '),(6121,3394208,'23,May,2004','2004-05-23 00:00:00','   Have you ever asked yourself? What part did you play in someone\'s life If what you did had an effect, a consequence... Maybe that person is who they are because of you I have asked myself this question and found that i have denied the negative and fully accepted the positive actions to a person\'s well being,his/her becoming of a real person knowing that denial is a front, i broke it down I tore every brick down with my hands and found many mistakes in the development of many friends and family I wish I could go back, but i can\'t Have you ever asked yourself? '),(6122,3394208,'23,May,2004','2004-05-23 00:00:00','   Sometimes, I just close my eyes and go to sleep Dreaming dreams I know i shouldn\'t keep Where have I gone? Where have I been? All my thoughts lead back to Raine One of the few who made me feel alive again I\'d loved alot and it\'s not the same I\'d give too much and expect nothing in return My rainy days are over Sunny days have come to me again... '),(6123,3394208,'23,May,2004','2004-05-23 00:00:00','   It was a tiring day... or was it? Looking over my shoulder constantly hoping i would bump into someone who would cheer me up... it never happened... I just wandered on, going wherever my feet could carry me... effortlessly i walked on not having a care to where i would be headed... just follow your feet...  no need to look back... just walk on... no matter how tired you are... smiles, left and right... all flashing at me... damn you happy people... I\'m envious of you all... i wonder what makes you smile? even in this  dark and gloomy day... damn you... while i trudge on these cold cold hard roads you mock me with your smiles... is it because you people know i can\'t? how placid must your lives be?  to not have a worry... to not ever draw a frown on your faces... to have all the time to do what you want to do... and how warm it must feel for you who has arms wrapped around you... damn you all... must you torment me with your laughs and giggles... must i always fulminate your actions... am i so garrulous about my hate towards happiness? why do i generate such anger? it must be the weather... everything is so abstruse... or must it be, i choose not to? i ponder on without sense... it is natural for me... i spot a woman, at first overwhelmed with sadness, but a message soon after came, and brought her much happiness... damn you... you are so fortunate... I tire of looking within... so much as in looking out... '),(6124,3394208,'21,May,2004','2004-05-21 00:00:00','   I was at starbucks today, sipping on my cafe mocha... sitting alone at one table, wanting to feel the loneliness looking at couples in there tables conversing... torturing myself, i don\'t understand why i\'m looking  for ways to hurt myself... to bring me much more pain i want so much to numb myself, someone kill me! i scream from within... no one is ever interested in me no one ever wonders where i am... and it\'s good i\'m not important, why should the world stop turning if i ask it to? i love getting ticked at myself for stupid shit i love it when i\'m hated... hate me more cigarette butts on the floor... i need more of you to hate me... i\'m not numb enough... that is one thing i haven\'t expierienced, pure numbness of self being... i want that... i want to reach the point where nothing could bother me even for a day...i believe it to be atainable... i just need your help title of the \"most empty man\" is what i want... '),(6125,3394208,'21,May,2004','2004-05-21 00:00:00','   it has engulfed me once again... much stronger now, the pain and regret... it had swallowed me once... but i had survived... i crawled out from that pit bruised and battered by the constant mockery of my own hate... i\'ve lived among the dark hoping to find a shred of light but alas, i couldn\'t... so i ventured to one pit after another... each and every time the same effect... vowing never to step back in ever again... but this time was different... i fell, not knowing how deep this chasm goes and in falling i have found a shred of peace and serenity thinking that something would break my fall at the end hoping i would not be disappointed at what lies beneath... what i thought came true, there was something there to break my fall but i was falling too fast, to fast that even it couldn\'t take my fall...  and so i lie there, broken beyond repair thinking that what i done was stupid... i fell too far down, too far for me to see the light too far for the light to see me...  darkness... total darkness... i lie there in a fetal position bones shattered, breathing heavily... '),(6126,3394208,'21,May,2004','2004-05-21 00:00:00','   What compells you to remember me? What is it i\'ve done to make you cling onto me like a grapevine to a wall? Is it that hard?  I have dragged you so far... I\'ve cursed this pathetic existence of a man... I\'ve told you straight forward... but still... You remain still...  How far must my anger dive in your ocean? How hard should i be on you? What must i do to drive you away? Forget me... I\'ll be happy if you do... '),(6127,3394208,'21,May,2004','2004-05-21 00:00:00','   feelings... what are they really? especially love... it feels so good but hurts so bad... damn them... why must i fall to such a depth knowing  once i reach the end of the pit i know it\'ll cause so much pain? loneliness... can force you to much extent surpassing ur own capabilities pushing you to walk on... till you meet someone who you espect could take it all  away... but when they can\'t  anger comes into play... rage... yes it kills all other feelings.... rage has blinded me to the point that i feel nothing at all... and when all my anger has been drought... sadness to walk among the dead... that is sadness pure obliteration of one\'s mind and soul Damn them... damn them all! why must one such as i must expierience such a deep level of all these feelings? i have found peace, serenity, comfort with being alone and for many years... only to be shattered by love... To walk among the dead... you will find me... '),(6128,3394208,'21,May,2004','2004-05-21 00:00:00','   I have been staring at the monitor for days Lighting a cigarette after another only standing when the pack is done or only to throw the ashes that overflow my ashtray I can only imagine what my friends are doing whether I am or not in their thoughts \"My own doing will lead to my demise\" I remind myself often just before i sleep As always only my music keeps me company At least i have something to keep me sane Although at times i can only Imagine Who I\'ll be when i see your faces once again I have Changed and it scares me Will you also be the same as the other\'s Who have shrugged me off, cause of the state I\'m in? I can only imagine... '),(6129,3394208,'20,May,2004','2004-05-20 00:00:00','   Those who are checking this shit please leave a comment or something, it\'s good to hear some... and oh yeah... any messengers there, just add me...  ym is  still_fighting_for_nothing@yahoo.com  msn is no_more_dope_for_me@hotmail.com '),(6130,3394208,'20,May,2004','2004-05-20 00:00:00','   I am no coward, escape is temporary It is only a delay to what is to come inevitably which is truth, answers, the end I stand now because i face what comes to me and survive. that is the only way even if standing my ground will be the cause of  my demise I shall not run I will continue standing I will continue fighting for what i believe in whether it be right or wrong Running away is and forever will be the last thing on my mind... '),(6131,3394208,'20,May,2004','2004-05-20 00:00:00','   I was at Puerto Galera with nothing to do I\'ll never forget the date of April 14, 2002 The names Jamie and Ren\'ee ring on that day Sneaking up to their room really late In the afternoon was when i met them They were playing volleyball, till i passed by, butted in The natural thing to do was flirt and ask questions These crazy girls said they came from the state of Washington After an hour of laughs and exageration They showed me signs that they wanted to get it on Slowly their hands crept up my thigh I told them i couldn\'t so Ren\'ee set up place and time Being the guy i am i broke house rules I snuck out from my room to get a taste of those two ho\'s I fucked them both at room A-42 I had me Washington bitches last April 14, 2002 '),(6132,3394208,'20,May,2004','2004-05-20 00:00:00','   People hanging around with nothing to do I light a cigarette after another sitting here waiting for you Grafitti on mirrors done by some local gangs feeling hella stoned, mind like quicksand the bustle of cars and local transportation Writing here feels so fucking stupid God damnit, what to do, what to do drinking, smoking, sitting here for you i light myself a joint, a single stick a blunt, a roach, friends say i should quit A pen, a notebook in my hand Starbucks coffee drinking is grand..... '),(6133,3394208,'20,May,2004','2004-05-20 00:00:00','   I blankly stared at the ceiling for about an hour spinning on this chair listening to cake Visions flashed in my head right there whilst in my seat of power decisions decisions, which one should i take?  I stood up and walked towards the kitchen 4:00 in the morning the clock glares at me Nothing in the refrigerator just left over fried chicken i gaze at the stove, thinking it would cook something for me  Cooking up a meal, damn i\'m too lazy just grab some canned goods and shove it and nuke it Vinna sausage, pork and beans and corned beef too messed up to think, shit just blenderize the shit '),(6134,3394208,'20,May,2004','2004-05-20 00:00:00','   What is there to find? I might be out of my mind These people drive me out of control What do you want from me? I can knock off all your teeth Need medication for this soul  Am I so blind? whatever it is I cannot see Am I so kind? Read the warning label on the bottle please  What are your motives? They\'re pretty precise, planned carefully You\'re screwing around with my lack of judgement What\'s in this disease? That has got me fucked up to my knees You start me up with your own set of keys  Dementia got a rope around my neck Tight, too tight to breath You got me walking back and forth effortlessly The time has come Fot you to step out the door and leave This is just too much, just let me be '),(6135,3394208,'20,May,2004','2004-05-20 00:00:00','   life\'s like a bitch you gotta know where to slap it! '),(6136,3394208,'20,May,2004','2004-05-20 00:00:00','   what is this word? i wonder... BLOG... Butt Loving Omnipotent Gooch? idunno... is it an abbreviation of  some sort? or just a make believe word? or am i just to lazy to find it in the Dictionary? what is it? hmmm i daftly manuever to my pile of books and carefully tear through the pages  blister... blunt... oh shit i think i passed it... blog... there is no such word in the dictionary! so what the fuck is a blogger?!? for this is blogger.com what the fuck is this? shit... blog... what is blog? '),(6137,3394208,'20,May,2004','2004-05-20 00:00:00','   Rain... why must it? To percieve into myself once again  is what happens when the heavens cry... cause what is there to do? locked in your room glaring at the computer monitor showing signs of glee when the words \"1 new message\" flashes on screen boredom... yes it goes hand in hand... and how irritating it is once it hits you... you tend to think of ways to get rid of it but it still remains, damn... why must it be like that? silence... it comes after boredom... or usually before that, but  it\'s more clear after... only it be broken by the continous pounding, or tapping of the rain on the roof or on the ground... why!?!?  Loneliness... Damn fucking rain won\'t stop... it\'s too cold too boring... too quiet... too lonely... shit someone come save me,  call or annoy me... release me from this fucking insanity DAMN FUCKING RAIN!!!!!! WHY MUST IT!!!!!! '),(6138,3394208,'20,May,2004','2004-05-20 00:00:00','   I woke up, i couldn\'t believe it what i had expierienced was just a dream a fictitious feeling in a state of unconciousness forever have i pondered on whether or not i should forever sleep, expieriencing bliss and absolute happiness or to wake up into a cruel and confusing reality dreams... things i know that can only be manifested in my head, in my thoughts...  it is only i who can see them it is only i who can understand them and so it is only i who can understand myself no one else... loneliness i\'ll live alone, i always have And i will be in this state, even after death in my awakening i have found freedom i have found truth and it is beautiful yet tragic at the same time '),(6139,3394208,'20,May,2004','2004-05-20 00:00:00','   It is true that i have expierienced more than i should normally have, but i made this decision, i took this path. I have surpassed my peers not in age, nor in any physical attribute, i have surpassed them in wisdom... i am young but i have lived my life as if i was born in my 20\'s and continued on... i have lived a life enough for 2 generations. although i am at peace with my state of being, it sometimes saddens me that i have not lived my age, that i had not had a chance to be a kid... the simplicity and naivety of a young boy\'s mind and heart for something... i\'ve expierienced much... I am young, but my mind has surpassed juvenility.. i feel old, think to mature... it\'s sad to know that i can never ever regress... but progress... '),(6140,3394208,'20,May,2004','2004-05-20 00:00:00','   At times like this... i\'m glad to have lived and walked in this world Although at times I just can\'t believe it Cause it\'s just so sureal I may not understand this life which hold numerous sublime expieriences but what is there to understand anyway? I am in a state of Halcyon That is all that matters to me right now... '),(6141,3394208,'20,May,2004','2004-05-20 00:00:00','   Could I bum a cig? These motherfuckers have none, so I\'m asking you A solid steel chair would replace The cig i want, cold steel hits my face A rush of blood to my head Lying on the floor, face forward, half dead Why don\'t you go dig a hole! I wake up losing all my self control I gave him a Whack! as he fell i brought a steel chair down on his back As i see blood drip slowly down his face the cig i want, a solid steel chair would replace '),(6142,3394208,'20,May,2004','2004-05-20 00:00:00','   A cure for this itch... is what i need wonder what\'ll happen if i go plant this seed? i\'ve been living in a wheel that continously turns doing things over and over not minding your concerns this is probably the thousandth time that i\'ll be doin this Insanity to me... doing things again and again expecting diff results... that\'s what it is... control yourself from this pestering cycle control yourself from loading the rifle A cure for this itch... is what i need wonder what\'ll happen if i go plant this seed? '),(6143,3394208,'20,May,2004','2004-05-20 00:00:00','   I\'ve been lying in my bed the whole afternoon No phone calls, No messages... I think I might just go out... I\'m craving for something, some excitement perhaps? Only a bolt of lightning can stop me now... Or probably a good friend... I\'m too bored, I feel like a total loser I know i\'m not what i used to be but i think i might just go out... even for awhile Everyone i know keep telling me It\'s for the best... just relax... take a rest But after awhile, this room starts to bore me So i might just go out... i\'m all done thinking Time to put some action into my thoughts '),(6144,3394208,'20,May,2004','2004-05-20 00:00:00','   Well, here i am, waiting for my time to come unravelling these tangled pieces of what i done reality, slowly overwhelming me i\'ve lost many friends cause i didn\'t want to stay  my time has come to be unglued searching memories of times i abused this mind ablur... alone and confused i just know you\'re waiting for me to lose  This addiction... is just too much to handle It laughs at me, as i continously stumble down this life i love... i lived it so fast and now i find myself just living in the past  I try to lean back, but there are thorns on my head i should\'ve never tried knowing that i\'d end up dead And now that I\'m here, after all that i\'ve done Where are my friends? to help me come undone? '),(6145,3394208,'29,June,2004','2004-06-29 00:00:00','   Alright so far all I knew of my blood was philippine, chinese, and hispanic, never have I known I had a drop of jewish blood in me, a drop of an ancient race driven from one domain to another, a race that had endured thousands of years worth of struggle and conflict... This new found knowledge explains alot, my main feature which in being; my nose... My knack of knowing how to get into and out of trouble... alot more... they say knowledge is brought down upon men through text and novelty, and recently noted, knowledge is brought down upon men by the origin of his blood,our paths have already been lain upon us even before our birth... it has been decided by our bloodlines... fascinating this new found truth... '),(6146,3394208,'28,June,2004','2004-06-28 00:00:00','   How many times will they do it  Taking what\'s mine and they tell me that I blew it.  What can you do; who me, yes you.  When bitches get scandeless and full of voodoo.  Catching my breath is the term I use when the world gives up on me.  Can you love or not? The truth finds all of us eventually.   I don\'t mind taking lives I find.  They all shine in my pocket at night.  You wonder why your always alone.  You never fit in you never could find a home,  and I don\'t need to second guess why,  your eaten by the anger and it makes it so hard to....   Love to love to love ya..ooh ooh   And please don\'t let me go, and please don\'t tell me no (who me yes you) '),(6147,3394208,'26,June,2004','2004-06-26 00:00:00','   Feeling feelings of hostility Won\'t someone come bother me Watching these people atop their high rise People glancing at me with their piercing eyes  These world that I have been Complicates the things I believe in And I\'m constantly wondering why Is it that i look up at ths sky  Running down the trcks Still viewing works of art they all try to fool me And I\'m wondering on smoking till the lights go off All around me and I\'m Knocking at your door but much harder than before  The sky is looking dim I say The sky is lit in a beautiful gray Here alone in the Zone Makes me wish i was home '),(6148,3394208,'26,June,2004','2004-06-26 00:00:00','   How I Look Like...&nbsp; urlLink     urlLink '),(6149,3394208,'25,June,2004','2004-06-25 00:00:00','   Me And My Guitar BABY!&nbsp; urlLink     urlLink '),(6150,3394208,'25,June,2004','2004-06-25 00:00:00','   urlLink    This My Room&nbsp; urlLink '),(6151,3394208,'25,June,2004','2004-06-25 00:00:00','   Watch as I strum away,  you\'re fear pain and misery And hear all the soothing word\'s  I sing along the way I pluck, pluck pluck  At these 6 strings rapidly Hitting and missing  Sometime\'s my guitar\'s a gun you say  So I shoot you up with these funky rythms  while I shoot myself a tequila Whoa! Whoa!  While I look around Figuring what makes good sound Comforting really isn\'t it? I sweep your feet off the ground And so you you dance dance dance not minding what\'s around Listen to my radical funky sound Baby my guitar\'s a gun  So I shoot you up with these funky rythms  while I shoot myself a tequila Whoa! Whoa! '),(6152,3394208,'25,June,2004','2004-06-25 00:00:00','   You came in so slow, but your exit was quick And baby I thought that your loving was thick But I guess I was just too unsure about you And it was a bad idea to ever call you my boo-boo I\'ll show you how I want it if you come back my way Get down on your knees and take it the other way  These playgirls think they can mess with my head They like messing with it so take the other one instead  I\'ve been totally wrong about you my boo-boo Tell me slut were the things you ever said to me were true? C\'mon baby baby I\'m waiting for it At first glance never in my head I thought you a trick Alright my boo-boo I\'ll let you go have your way But when you pass me my boo-boo still remember you owe me  These playgirls think they can mess with my head They like messing with it so take the other one instead '),(6153,3394208,'25,June,2004','2004-06-25 00:00:00','   And so say that I\'ll never stray so far away And never again will I ever feel that way The only thing that matters now is me being anywhere but here Cause wherever I roam your presence lingers there  And so my brother we will never separate And our memories of the past won\'t dissipate You may have gone but to me you stay My friend, my family to me you never have lain  I walk on with our memories intact I still haven\'t forgotten our secret pact To me and to all you have always said \"I can\'t, I won\'t, I don\'t fade away\"  My peace to you my cousin it\'s been over a year Anywhere but here cause wherever I roam your presence lingers there '),(6154,3394208,'25,June,2004','2004-06-25 00:00:00','   I\'ve been cruising around looking for my only lady She\'s somebody I aint know, that\'s for sure but I\'ve been down watching the sets Looking and looking till I couldn\'t catch my breath and I spot her resting over by the shade And now it\'s time to go and lay my charm but then a man came up and wrapped her in his arms so I walked out again looking for my buddies, Telling them someone come comfort me and But then I saw this pretty girl shining brightly And all of a sudden I was blinded suddenly by her light so bright and I couldn\'t even fight for her attention please come glance at me oh think you\'d come to my way oh I\'d like to think you\'d come to my way Been dreaming again but my friends have butted in but I couldn\'t even move I was mesmerized wondering if that was a diguise I\'m not smart nor am I wise I thought I\'ve seen it all all the things that bring you awe But I guess I was wrong Oh baby please kill your lights it is far too bright shining too much this can\'t be right I\'m blinded suddenly I\'m suddenly blinded It doesn\'t really matter on how you arrange it but it\'s all the same... '),(6155,3394208,'25,June,2004','2004-06-25 00:00:00','   I\'ve been waiting all this time, soliciting is my one crime  Take a step which way to go, with no permission from a ho  And I live with no disease, my back seats got remedies  I\'ll be part of history, with 15 books on my story  And be a swift like roddigan, equipt with the best selection  Wondering what you use, slipping off the booze cruise  And your focus turns to blur, when you set your eyes on her  As I turn away when she looks at me   And I\'m a wanting watching you walk away  I will try to meet you by the old highway  And if it\'s now I\'m down to slack  With the sun shown accross my back  And I\'m comforted to know That you\'re backstage at my door  Why am I so short of your attention  Having this discussion downtown  No obligation, it\'s a real hard emotion  Now that your gone, I\'m moving right along  Run down the road still viewing sluts, bowcats,  these ho\'s are trying to fool me  No promises you get the one time slam  Seductive sexy capers in the hilton tram  with this abuse of napkins  X-amount of heineken  Waking up\'s so hard to do when I\'m lying next to you  So I\'m knockin at your door, complicates the night before  She says the joke is on me, I said the joke is on her  Mary oh Mary, your still so quite contrary  How could you ever know   And I\'m a wanting watching you walk away  I will try to meet you by the old highway  And if it\'s now I\'m down to slack  With the sun shown accross my back  And I\'m comforted to know That you\'re backstage at my door  Now that you\'re gone... '),(6156,3394208,'24,June,2004','2004-06-24 00:00:00','   gimme some lovin early in the morning i just woke up with this appetite come on, give me love in the morning light  gimme some lovin early in the morning i just woke up with this appetite come on, give me love in the morning light  some people like it in the afternoon i don\'t think that I can wait that soon come on, give me love in the morning light  some people like it in late at night i could wait that long, but i don\'t think i might come on, give me love in the morning light  i\'m going to love you soft and love you sweet i can feel your heartbeat need your attention to get me started right  i\'m going to wake you up with a kiss on the cheek come a little closer, open up to me good morning babe, good morning babe  open your eyes just a little bit good morning sunshine, i can\'t resist dreaming about you babe, all through the night '),(6157,3394208,'24,June,2004','2004-06-24 00:00:00','   Ooh baby nobody does it better Take me away they will never Cause I will forever be free So take it down down down And if you ever leave this town Look for me cause I will forever be free  Oh baby how you love me \"I\'ll love you everyday\" that\'s what you say to me I\'ll give you all my lovin Till the day\'s start dyin Cause I will forever be free  So take it down, down low And promise you won\'t act a hoe Cause I\'ll love you forever baby I\'ll love you forever  Now I can\'t stop dreaming Cause my heart keeps on beating I\'m letting you know '),(6158,3394208,'24,June,2004','2004-06-24 00:00:00','   I\'ve been waiting all this time Procrastination is my only crime Could you be interested in my evergrowing heat To tell you baby I\'m at my sexual peak Knocking at your door, but much harder than before And I\'m a wanting and watching you come my way Oh baby, baby please knock me down I\'m on my way, to your home baby I\'m bound  Take me down down down... I say take me down down down... Tell me why is it that you leak Baby, baby I\'m at my peak...  Wondering about when you will come Baby baby don\'t you tell me your already done I\'m knocking, I\'m knocking once again But baby you left your done with nothing to say Could you be interested in my evergrowing heat To tell you baby I\'m at my sexual peak Won\'t ya stop if you done go low Don\'t ya shutter what i got for you  Take me down down down... I say take me down down down... Tell me why is it that you leak Baby, baby I\'m at my peak...  Oh baby what I did do to you Have always been the dreams of my crew Oh baby what I did do to you Have always been the dreams of my crew '),(6159,3394208,'24,June,2004','2004-06-24 00:00:00','   Oh no- here comes that sun again.  And (that) means another day without you my friend.  And it hurts me to look into the mirror at myself.  And it hurts even more to have to be with somebody else.   And it\'s so hard to do and so easy to say.  But sometimes - sometimes,  you just have to walk away - walk away.   With so many people to love in my life, why do I worry about one?  But you put the happy in my ness, you put the good times into my fun.   And it\'s so hard to do and so easy to say.  But sometimes - sometimes,  you just have to walk away - walk away and head for the door.   We\'ve tried the goodbye so many days.  We walk in the same direction so that we could never stray.  They say if you love somebody than you have got to set them free,  but I would rather be locked to you than live in this pain and misery.  They say time will make all this go away,  but it\'s time that has taken my tomorrows and turned them into yesterdays.  And once again that rising sun is droppin\' on down  And once again, you my friend, are nowhere to be found.   And it\'s so hard to do and so easy to say.  But sometimes, sometimes you just have to walk away, walk away and head for the door.  You just walk away - walk away - walk away.  You just walk away, walk on, turn and head for the door. '),(6160,3394208,'23,June,2004','2004-06-23 00:00:00','   I was at the beach, sitting on the sand When I saw this girl Dark brown eyes, and a dark olive tan She was watching the sets With a surfboard underneath her arm The first moment I saw her I knew she\'d do me some harm  She\'s so good She came out of the water From the best session I\'d seen Man, she had the biggest chi chis I had ever seen She put down her board, and sat down next to me And played a soft ole lonely tune, on her ukulele I don\'t know what had happened But I\'d fallen asleep The next morning I woke up She was sleeping next to me   I fell in love with her I\'m so in love with her I fell in love with her My little Surfer Girl   She got out of bed Waxed up both our boards Grabbed me by the hand and then she took me out the door We surfed all that day long At least eight hours straight Man! I knew she was the one And damn I couldn\'t wait Well I asked if she loved me She said, \"Yeah well I guess\" So I asked her to marry me Dammit, you know she said \"yes\"   I fell in love with her I\'m so in love with her I fell in love with her My little Surfer Girl   I was at the alter In my tuxedo-o-o Waiting for my Surfer Girl Man, where did she go Later on that day A friend to me did tell-ell-ell That she went to Mexico To catch a, to Catch a, to catch a 10 foot swell That Bitch '),(6161,3394208,'23,June,2004','2004-06-23 00:00:00','   Spinning away, so far away Hammered beyond belief shattered pieces of my memory Spinning away, so far away  It began with a bottle of beer walking around with an irresistible veneer I picked up a girl, maybe two they said \"we could do so much with you\" At this i took 3 shots at the bar spotting my buddies at the door that was far They motioned to me and glanced at my female companions and took them to dance  Spinning away, so far away Hammered beyond belief shattered pieces of my memory Spinning away, so far away  My head aches from the alcohol and probably cause of the v\'s I took and demorol I just took a girl by the hand took her to a room and she assumed position Before I moved in she passed out Damn i thought my luck would never run out I got out and met with Horry and we had a session with Sancho and Cherry  Spinning away, so far away Hammered beyond belief shattered pieces of my memory Spinning away, so far away  I woke up no clothes lying in bed all these memories clouded just flooding in my head Grab myself the pack of cigs on the floor just then i heard a familiar voice say \"baby gimme some more\" What? Huh? How? And when? I don\'t remember letting any girls in Cherry there lying in bed talking asleep my god i don\'t remember anything i was hammered beyond belief  Spinning away, so far away Hammered beyond belief shattered pieces of my memory Spinning away, so far away '),(6162,3394208,'22,June,2004','2004-06-22 00:00:00','   I only seem to look at the frontispiece And you always seem so frugal toward me Why don\'t we spend the day out in the sun Drive till we tire of this so-called fun  Far off  from this overwhelming skirmish Like cream like the cream-like cheese fromage frais '),(6163,3394208,'22,June,2004','2004-06-22 00:00:00','   Beauty being sold in heaps Center stage filling with seats You are such a flake You can\'t get none so you\'re buying 2 for 1  Cheap motherfucker has never loved another buy a drink and a hoe fucking and watching the show  Look at that lady by the corner street But she won\'t be bought by the green You are such a fake Cause your beauty can get 2 for 1  Cheap motherfucker has never loved another Get a drink act a hoe fucking and watching the show '),(6164,3394208,'21,June,2004','2004-06-21 00:00:00','   My sweet sugar mama Dance for me won\'t ya? Take me on ya joyride I\'ma take you on front back and side to side  Would ya ever take me on looped around your way baby Ooh baby how you lay it down I wanna love ya everyday  My lady, my baby, my sweet sugar mommy baby crush me Lemme take a hold of you soon i\'ll be servin breakfast for two  Would ya ever take me on looped around your way baby Ooh baby how you lay it down I wanna love ya everyday  Tell me what ya want I\'ll give it to get in your pants Looped around your way baby I wanna love ya everyday '),(6165,3394208,'21,June,2004','2004-06-21 00:00:00','   Sige na nga diyan, magtatagalog na ako... Pangtrip lang to kaya huwag kayong mag-abuso...  Isang simpleng tao naman ako Pero marami na akong pinagdaanan Ngunit hanggang ngayon wala akong naiintindihan  Ano ba ang aking kasalanan Parang ako lamang ang palaging hinahamon Sa lahat ng labanan kahit hindi aking panahon Ang aking isang bilin sa diyos namataas Sanay hindi ako mawatas-watas Dahil marami akong minamahal sa mundong ito Kahit na wala silang pake alam sa buhay ko Sa ngayung araw at hanggang kamatayan Ang buhay ko ay hindi ko nang pagsisisihan   I can\'t really speak tagalog well but i tried Uh hopefully ur comments would dent my evergrowing pride '),(6166,3394208,'21,June,2004','2004-06-21 00:00:00','   To what did i do to recieve this acclamation? When all i have really done was plain procrastination... Am i really simple in a complex way, you nod in agreement When have i reached this social status, is it such a statement? Must i bow to this ever evolving way of perception? These subsequent events fill my mind with much confusion '),(6167,3394208,'19,June,2004','2004-06-19 00:00:00','   Whoopee doo, doopee wah Stuff i say to make you laugh Tell me what more can i say or do What the hell agagagoogoo? Someday, someway Baby, baby don\'t you go stray Slap me silly till i bleed give me some booze, and awholelottaweed  Clean your ears Babyflo Gentle Buds Clean your ears Babyflo Gentle Buds '),(6168,3394208,'19,June,2004','2004-06-19 00:00:00','   Enjoy what you are about to see This is my made for tv movie The things i say they might be true But it depends really upon you  I have walked a mile over land and above sea If you don\'t believe me take a closer look at my feet I have battled many wars, wars that have no ending My body bruised, mind like ice, and a heart crumbling  Down, down, down i go below Wham, bam thank you man, but really, i got to go My head is aching, i mean i really do not know  I drank and i swam like an overbloated fish I took everything as if it was nothing at all like a sushi dish I\'ve rambled to myself and to you on and on But never again, no never again shall I turn myself on  You, you, you i go below taking and making and raping all of you know Overintoxicated i mean i really do not know  I took to the streets, to the streets and i wandered off I can\'t, no i won\'t ever leave again and go just take off Can\'t i see what this life is to me unimportant really But i guess that i have to because this aint no ordinary fantasy '),(6169,3394208,'18,June,2004','2004-06-18 00:00:00','   I haven\'t much to say except you shouldn\'t stay I haven\'t got a clue of what I should do with you And I don\'t wanna play cause it\'ll just ruin the days Of my goin around my world with you  Someone take me away before I get faded for the night cause I really don\'t feel well at all Come just take me away take me away cause I really don\'t feel well at all  So what\'s wrong with you you just look so damn blue And I don\'t wanna fight cause it\'ll just spoil the night And you don\'t wanna play cause it\'s just not your way About you, it\'s just you and a bit of me too  Someone take me away before I get faded for the night cause I really don\'t feel well at all Come just take me away take me away cause I really don\'t feel well at all  I\'m faded I\'m faded I\'m drunk, stoned and hated You feel the same way so I guess you should stay And I guess we could play cause it just aint our day To go wandering off on this cold dark night  Someone take me away before I get faded for the night cause I really don\'t feel well at all Come just take me away take me away cause I really don\'t feel well at all '),(6170,3394208,'17,June,2004','2004-06-17 00:00:00','   She looked me in the eye What have I done wrong Visions of the sky They didn\'t last long These worlds that she has seen \"you will never find me\" Locked above the rise Can I hear us? Can I hear us?  Falling back to the ground Never even made a sound Won\'t you ever leave this town?  4 weeks have gone by Wanted to scream, but he did not try Killing his fears, he challenged he and said  You have no idea what it\'s like What it\'s like to be me...  Unhinging all these screws She says she wants something different No options left to choose He says he wants something different Paying all these dues I wish that something was different Yeah something was different No matter what happens to you today '),(6171,3394208,'11,June,2004','2004-06-11 00:00:00','   I start to pace and pace around listening to my feet shuffle on the floor Dangerous this feeling i got Got myself a bottle of scotch to fill this cup  It\'s 3\'clock and i still feel up It\'s 3\'clock and i still feel up  I run to my room to grab my guitar Holy shit... that felt really far My artistic mind starts kicking in Writing and writing with a face on my grin  It\'s 3\'clock and i still feel up It\'s 3\'clock and i still feel up  I call my bro\'s and we start to jam Looking up and down, left right hands We talk till my names turns to Chrisizzle My nizzles are busy with the cali tie grassizles  It\'s 3\'clock and i still feel up It\'s 3\'clock and i still feel up '),(6172,3394208,'11,June,2004','2004-06-11 00:00:00','   I\'m sure I have done this...  Keeping all the good times forgetting the bad times Choosing a part of my life to tell  And leave the rest in the closet, locked away... But the question here is...  Why I do it? I ask and ask... Why do I do this, when i will be reminded of it Somehow, someway, hell...  Someday it will come out,  And in that time it\'ll be too late... Editing my life... I probably do this to compensate for yours Or maybe, i don\'t even have a life, So I add things that were never even there To keep you near me... just maybe... The thing here is sometimes I\'m not even concious about it... Unconsciously, i\'m already fucking around with you And what\'s weird is, i already am before i even meet you... And at the same time I\'m also getting fucked By Me... Editing my life... '),(6173,3394208,'11,June,2004','2004-06-11 00:00:00','   I fill it up, I fill it up And take a shot from this cup I strum along, I strum along Memorizing the lyrics of this song The watering can with holes on it\'s end Feelings of constipation quikly sink in The watering can with holes on it\'s end Can\'t help it i just fall thru the ceiling I tear them down, I tear them down Pictures of myself, faded and frowned I\'m far too late, I\'m far too late I have no room to suffocate The watering can with holes on it\'s end No use still fighting for nothing The watering can with holes on it\'s end I shall start to shout, no time for pretend '),(6174,3394208,'11,June,2004','2004-06-11 00:00:00','   I had a match She had a lighter I had a flame but she had a fire I was bright but she was much brighter I was high but she was the sky  Oh baby I was bound for mexico Oh baby I was bound to let you go  I don\'t know much  about cinco de mayo I\'m never sure what it\'s all about but I say i want you and you don\'t believe me you say you want me but I\'ve got my doubts  Oh baby I was bound for mexico Oh baby I was bound to let you go  I had a match She had a lighter I had a flame but she had a fire I was bright but she was much brighter I was high but she was the sky  Oh baby I was bound for mexico Oh baby I was bound to let you go '),(6175,3394208,'09,June,2004','2004-06-09 00:00:00','   How you take it down low How you do it so slow Taking your time just grab it I want it all not just a bit  Funky freaky girly don\'t just tease me, don\'t just tease me  How you diggin me, diggin you? The way you love me treating you? Don\'t take your time just grab it Give it to me all, just give it  Took me awhile to go How you do it so so slow Sweet as honey funky freaky girly Don\'t tease me, just don\'t tease me '),(6176,3394208,'09,June,2004','2004-06-09 00:00:00','   Nice and easy, ooh how you please me you got the love that\'s really real Sweet and spicy, c\'mon and touch me you know the way i like to feel  Aint nobody move me, aint nobody move me like you do Girl you keep me up there, when you gettin down you chew Do it do it girl, c\'mon and do it for me yeah  Tantilizing, so appetizing,  you got me running for your love Sweet as honey, ooh how you love me you keep me coming back for more '),(6177,3394208,'09,June,2004','2004-06-09 00:00:00','   Yo- could I get a cold beverage I need some leverage It\'s sunny outside Some lemonade would be nice Or a sprite through the drive through At B.K. Yo I\'ll buy girl a drink But girls got to pay On the front porch I got Some iced tea If you like a taste of tea Then come along with me It\'s martini time Yeah yeah feeling golden Bring your own beverage Just make sure it\'s cold I like cold beverage Give me a frosty mug Filled with A&W If you got ice cream Make it a double scoop Milkshade at the fountain Really good sounding Chocolate egg cream Yo, I\'m champ cherry pounding Caught a chill vibe Orange juice in my ride Wawa\'s to the right They got a beverage inside Dig me a hot coffee Fill it up with ice Watermelons like drink Please fix me a large slice Summertime is cool the heat is getting old Yeah I\'ll get a beverage Just make sure it\'s cold cold I like cold beverage When I\'m fishing Let\'s keep one thing clear The baits over there The brews right here Two six packs and a big bag of ice Didn\'t even catch a bite But the brew tasted nice Back to the bar Strawberry daiquiris and a colada I need a whole lotta them Fruit drinks to catch me a buzz I must tell you I\'m the Cool Aid kid Before you serve my drink Please stick it in the fridge I like cold beverage '),(6178,3394208,'07,June,2004','2004-06-07 00:00:00','   Cooling out, cooling out Waiting for my lady got the special sauce that\'s why she\'s my baby My baby, my baby And of course I\'ll do anything for her Any day she wants to show her honey, show Got a sharp mouth a sharp tongue gimme gimme gimme some Don\'t take to that gimme gimme none You gots to work and deserve it earn it to own it Once you got never ever gonna leave it alone don\'t take wise cracking or Forgetting never been stood up your... cuz she does the stepping  My baby got sauce Your baby ain\'t sweet like mine (She got sauce) Your baby ain\'t sweet like mine (She got sauce) She ain\'t just sweet..  Well that\'s some funky lemonade you got going baby Is it special for me? Did you make it today Cuz I\'ll always stop by if you like me to I\'ll do anything that you could ever want for me to do A kiss for some of this A smile and it\'s done Miss I\'m your mister My sister\'s your sister My mother\'s your mother father\'s your father It all started when I kissed her, I missed her, I did It was love at first sight, yo, though Better when it started I ain\'t broken hearted but baby I need some time, some time alone with the crew To do the stupid things that we used to do Upon this she slapped me With a frying pan Yo, this is love I\'m not gonna leave it  My baby\'s got sauce Your baby ain\'t sweet like mine (She got sauce) Your baby ain\'t sweet like mine (My baby got sauce) She ain\'t that sweet, see...  Baby what I would do for just one look in my direction, I gave you all I had but then you still act mad, You don\'t give me no attention  My baby got sauce and of course as a matter of course She does what she wants to because she\'s the boss Please don\'t tell my friends about the situation \'Cause, like they think I\'m on some extended vacation \'Cause, I don\'t get out to see my boys anymore Only time I leave the house is to go to the store for some milk What she wants what she wants me for What could she want this time I\'m all hers No, she ain\'t even mine, ain\'t even mine Believe me what she do at night i dont know but baby  I could guess it ain\'t right Coz she got me waiting comes in late and I\'m trying not to raise my voice Negotiating it I\'m confident that she\'ll win the argument Shouldn\'t even started it I\'m broken hearted I can\'t be parted y\'all I need a taste So I count my loss...  Dig......My baby got sauce Your baby ain\'t sweet like mine (She got sauce) Your baby ain\'t sweet like mine (My baby got sauce) Your baby ain\'t sweet like miiinnne She got sauce, she got sauce, she got sauce Your baby ain\'t sweet like mine, ain\'t so sweet like mine, ain\'t so sweet like mine Ain\'t so sweet like mine She ain\'t so sweet '),(6179,3394208,'07,June,2004','2004-06-07 00:00:00','   Counting my days, one by one i fall Won\'t you be so kind and be my lapdog? Lady please fall, my lady please fall Curiousity take this and stand tall  Someday hope my days shant be with fog So would you be so kind, and be my lapdog? Still i fight for nothing fascinating For still, my lady won\'t understand my reasoning  For days i\'ve tried and sought for you to be my lapdog... '),(6180,3394208,'02,June,2004','2004-06-02 00:00:00','   An act of pure desire Doing it, I\'ve never been better Pull me in thru your metaphorical State of being, purely subliminal  The parallel universe Adjacent to ours it\'s course Must be some way to cut away From this over-empowering being of state  Cast me in the pit Of your hollow heart remains unlit Just a continous cycle Running around these perfect circles  To be pacified locked within your stare Your beautiful face covered by your hair Run, till you have nothing to fear Purely based on desire, I\'ve never been better '),(6181,3394208,'02,June,2004','2004-06-02 00:00:00','   Wake up in the morning, Clock says half past one  I have no sunglasses As I step into the sun  Thereâ€™s no recollection Of the evil things Iâ€™ve done  My head feels like I mustaâ€™ had some fun   What happened?   Last thing I remember I was chillin at a party  Pinching girlyâ€™s asses I was drinking recklessly  I know I did something Lord what could it be?  Woke up in the morning And all my friends say me   What happened?   What am I doing here? Who is this girl in my bed?  What is this shit on my face? My god, what is that awful smell?   She may be an angel She may be a queen  She might be black, white, american, indian or japanese   Thru a bottle at a bouncer Didnâ€™t think that he was cool  Pissed in someone\'s drink And thru a bike into a pool  Driving down the side walk Like a drunken possessed fool  I broke every single traffic rule.   What happened? What happened?   Letâ€™s go '),(6182,3394208,'02,June,2004','2004-06-02 00:00:00','   To me, coming from you... Friend is a four letter word... And is the only part of the word That I heard... Call me morbid or absurd... But to me... coming from you Friend is a four letter word...  When I go fishing for the words I am wishing you would say to me I am really only praying that  The words that you\'ll be saying like  Betray the way you feel about me... '),(6183,3394208,'02,June,2004','2004-06-02 00:00:00','   Let the day fade, pass time, drink up Lie in bed, looking up, take a sip from the cup The music echoes... echoes in my head Let\'s waste the day together... You and me forever... Serenity and I, in bed, the others are downstairs Echo... in my head again I\'ve never been so alone And I\'ve never been so alive! the sun rises, it\'s light seep through the blinds brush it does on mine and Serenity\'s... It echoes once more... It\'s so distant, can\'t hear it... But it still does... '),(6184,3394208,'02,June,2004','2004-06-02 00:00:00','   Why are you there? locked alone in your hole The anchorite stares he says he lost his goals Callus are his feet his hands, his body scarred \"I was a failure went to high, i fell discarded\" \"Leave me alone\" the anchorite pleads and cries \"Let me sing my song away from all your prying eyes\" '),(6185,3394208,'02,June,2004','2004-06-02 00:00:00','   Suffer did I do so gallantly locked away deep in this belfry Don\'t be a stranger, come if you like But take advantage, of everything in sight Wishing away, doesn\'t do anything Wishing away, gives you nothing Born with a disgruntled abnormality Holding onto your dreams and fantasy You have no idea, say what you want to say Listen i shall, but don\'t think i\'ll stay Wishing away, doesn\'t do anything Wishing away, gives you nothing Wishing away, go ahead waste your time Go keep wishing for something sublime I\'m not saying that i don\'t wish away But don\'t go dwell on your dreams the whole day... '),(6186,3394208,'02,June,2004','2004-06-02 00:00:00','   Looking and looking, still can\'t see, still can\'t find What makes my mind heart and soul tick... The medicated state of mind Things i try to get everything i know of you out of my mind Searching, keep searching till ur completely blind All of my time spent on things such as you, it\'s ardous Bottled up inside this feeling of angst, must read Dr. Seuss To clear my head and spark a fire, what do i have to lose? Completely oblivious to my surroundings, my options to choose Dangling on a fine piece or thread, my head it rests Wondering about the day after come, should i take these tests You cut me up, I\'m wide awake lying on a pile of pests That gnaw at my skin, the medicated state of mind, am i so blessed? ');
INSERT INTO `msgs` VALUES (6187,3394208,'02,June,2004','2004-06-02 00:00:00','   Beginnings and there Endings  By: Chris Sarmiento  Intro  It all began on Sept. 19, 03â€¦ cool late summer breeze blowing as the clock reached 5:00 pm. I was outside the bus terminal waiting for a friend of mine. I remember it so vividlyâ€¦ I was there with my pen and notebook as usual sketching down whatever visions Iâ€™ve been receiving in this deluded head of mine. Itâ€™s been only four minutes but it felt as though the whole day went by and still, I hadnâ€™t put down anything on this old Maui and Sons hardbound notebook. I decided to take awhile off so I split, sent Horry a message through my phone telling him weâ€™ll meet up tomorrow instead. Walking these dirty streets, my head still throbbing from a hangover, I noticed the sun set early I didnâ€™t see the orange filled sky it was just plainâ€¦ darkâ€¦ â€œDamn, Christmas is drawing nearâ€¦â€ I thought to myself. After realizing, I stopped to check my wallet, see if I had enough money to buy gifts for my friends or familyâ€¦ Fucking broke. I only had enough cash to get me homeâ€¦ Then for a moment I thought to myself â€œwhat am I worrying about Christmas isnâ€™t here till three more monthsâ€ I put my wallet back in my back pocket. Walking steadily, trying not to fall asleep. My phone started ringing.  I answered the phone. â€œ Hey! Bro, look I canâ€™t go out man, my car got wrecked last night on the way homeâ€¦â€ It was Horryâ€¦ â€œ I â€˜m really sorry man but there ainâ€™t nothing I can do right now Iâ€™m broke and please bro I just canâ€™t goâ€¦â€ I could tell he was lying, cause I heard the laughter of one or two girls in the backgroundâ€¦ â€œ Look, Horry, bro, I told you to meet me there cause Iâ€™m fucked up right now, didnâ€™t I tell you! Plus you couldâ€™ve just told me you were fucking some hoes you didnâ€™t have to lie to meâ€¦â€ Horry apologized and I told him Iâ€™m going home alone. Iâ€™m one patient guy but if Iâ€™m not feeling well my patience is just a piece of thread that an insect can easily tear. My mind went a blur after that short and meaningless talk over the phone. I started once again on the long way homeâ€¦ then I decided to turn, stop once again and call a friend of mine. As I dialed the numbers on my phone I had a short flashback of what had happened last night.  I was at Cherryâ€™s place at around 6:00 p.m. setting everything up, laser lights, the foam machine, everythingâ€¦ I walked around in her patio telling my underlings where to put whereâ€¦ it was the first time I was asked to host a party, and usually I never hosted parties, but I threw great onesâ€¦ it was all going as planned, we installed a couple of Bose speakers to the living room, set the den as the chillout room, every room was a make out room, and the fucking patio was a great site. It was around 9:00 p.m. I had my first shot of Jose Cuervo, just sitting alone till Cherry sat next to me by the bar. I asked about her expectations for this party. Cherry replied with a smile and a flirtish glare that made me weak in the knees.â€œ So tell me, do you think Iâ€™m missing anything?â€ I asked Cherry while I took another shot of tequila. â€œ I donâ€™t think so, actually Iâ€™d like to say thanks for doing this for me, I couldnâ€™t find anyone else to set the place and host itâ€¦â€ she leaned over and gave me a soft kiss on the lips, I took another shot and strolled off to the gate leaving Cherry by the bar. I didnâ€™t expect her to follow me; I just wanted to enjoy company alone. Iâ€™ve no intentions in fucking a good friend of mineâ€¦ That night Cherry wore a dark red, spaghetti strapped blouse exposing a bust of 38 câ€™s that would make any manâ€™s jaw drop, she just passed by me trying to put fantasies in my head, well some did pop in there. It was around 10:15 the place was swarming with familiar faces and new onesâ€¦ all to which I was surprised of, for they praised me for the set up and the soundsâ€¦  10:15 p.m. all these people just started pouring in I thought fuck Iâ€™m going to get a lot out of this party. I started mingling around, when I saw one of my old girlfriends, Melanie, I remember this girl just loved it up the ass, too bad she turned lesbian after we split upâ€¦ I couldnâ€™t resist so I came up to her, over the loud music I shouted â€œ Hey Melie! Long time no seeâ€¦ howâ€™ve you been?â€ I waited and tried to get her attention once more. After a few more tries I started thinking, fuck she isnâ€™t going to talk to me so I started to walk awayâ€¦ I took only a couple of steps when I felt someone grab my arm. It was Melanie she just smiled and spoke some words that I could barely hear, so I pointed to the glass door leading to the garden. She smiled and nodded, and told her friend to wait inside. As we pushed our way through the crowd to get to the door, I bumped into Horry, he glanced at me and saw Melieâ€¦ then he drew me close to hear him, â€œ Bro, donâ€™t tell me youâ€™re going to start fucking Melie again?â€ he gave me a slap on the shoulder, as I walked away. I could here him shouting over the music saying â€œ at least introduce me to her lezzi Friends Bro!!â€ I just laughedâ€¦ damn Horry is just one fuckinâ€™ horny bro.  We reached the garden, I took my shoes and socks off cause I thought it would be a trip to feel the grass with bare feet. I was surprised that Melie also did the same thing; we walked around till we reached the gazebo. I started the conversation asking her how she has been since three years ago. She just laughed and glared at me for a second â€œ Oh, me Iâ€™ve been doing well lately, Iâ€™ve finally finished my course on Architecture and Iâ€™m planning on moving to Australia soonâ€¦â€ as she continued I couldnâ€™t help but just stare at her. I kind of missed the days I was with her, my hand caressing her soft creamy skin, her long blonde hair and those soft lips that touched mine everyday. Hey she was great company and a great fuck. â€œâ€¦ And then I guess thatâ€™s all I got planned for me, What about you Howâ€™ve you been?â€ she finally stopped. I simply just said. â€œ Me, Iâ€™m doing great, I got my life back somehow. I just canâ€™t help thinking of one thingâ€¦.â€ â€œWhat is that one thing?â€ Melie asked me with curiosity â€œâ€¦what if we were still together, do you think youâ€™d ever become a lesbian?â€ she just laughed at me, I know I just asked a stupid question, and that gave me awayâ€¦ â€œWellâ€¦â€ she replied, â€œ to tell you the truth I was always a lesbian I just wanted to try dick for a change, especially up my assâ€¦ he he...â€ from then on we told whatâ€™s been going on and I was still trying somehow to get in her pants once more, with more bedside humor. Well as the conversation dragged on I noticed that the air had been much cooler and the music started to fade, it was the time that I had to officially start the party.  I told Melanie it was time for me to get the party officially started so we left the gazebo and went separate ways when we reached the glass door leading to the dining hall packed with people. Before I headed to the DJâ€™s booth I stopped by the bar and downed four shots of tequila to drop the fear of talking in front of crowds and I also brought a bottle of beer along with me. The bartender Edong glanced at me telling me to take it easy on the drinks, cause the party barely started and Iâ€™m drinking like a fish. I disregarded the advice and took another bottle. I stood up from the bar ready to make my party opening speech to the people, as I reached the booth I could see Cherry walking through the crowd to the DJâ€™s booth. It was a beautiful sight, well for a drunk like me, imagine a beautiful girl that could easily split these mass of people without pushing them or speaking to them, she was like a walking crowd repellent, or maybe my thoughts were already messed upâ€¦ as she got up there with me I could barely take my eyes off her, I was once again thinking fantasies in my head, but I easily snapped out of it telling myself that itâ€™s the alcohol thatâ€™s making me feel this wayâ€¦ it was already 11:24 I motioned the DJ to kill the music for a secondâ€¦ as he did, I stepped up to the microphoneâ€¦ fuck I was once again surprised as to hear some of the people in the crowd just call my name, when I didnâ€™t even do, or say anything yetâ€¦ I took a sip of my beer and beganâ€¦ â€œ Aiyyo I donâ€™t know how the fuck you guys know me, and why you guys like me, but thatâ€™s coolâ€¦â€ Cherry stepped back a few steps and laughed. â€œ Anyways we all know why we are all here arenâ€™t we?â€ I drank some more from my bottle; I could already feel my face getting red. â€œ We are here to party, to smoke dope, to fuck the police and their motherfucking laws, to get laidâ€¦ but most importantly we are all here for Cherryâ€™s added year in her life letâ€™s give her a holler people!!! And letâ€™s get the fucking party started!!!!â€ these guys just went crazy after that frickin speechâ€¦ Cherry went up to me and gave me a kiss and threw me down the booth along with her. All right that shit was fuckedâ€¦  Lying on the floor with her on top of me, with bits of pieces of glass and food piercing my back, I just noticed how much this girl liked meâ€¦ she stood up and motioned me to walk with her outside. I stood up brushing of the mess from my back; she took my hand and snatched a bottle of Cuervo in the other. I once again found myself walking through this mass of people but now with much ease and with Cherry leading the way, It felt like a dream sequence, from one of my acid tripsâ€¦ as we reached the glass door hand in hand, she moved her lips and started pointing to the wooden bench, assumingly she wanted me to sit with her, and so I nodded. I could tell she was planning to do some crazy shit to me, or was it the alcohol doing the thinking for me again? I slid the glass door with a force I forgot to check, it hit the edges hard almost causing it to shatter; I faced Cherry and flashed a smile of embarrassment. She just flashed me a flirtish smile and didnâ€™t even mind if I almost broke the door. At this point I could already feel the heavy buzz of tequila and beer; and the music was already getting to me. You know I almost lost my mindâ€¦ it was The Chemical Brothers playing, the track was The Test; my head was throbbing harder now pounding each time I heard the bassâ€¦ but now Iâ€™m home and Iâ€™m freeâ€¦ if Cherry wasnâ€™t there Iâ€™d already lost my composure and start fucking up the place. We walked till we reached the bench; she smiled again. I noticed that this girl just loves to smile; her lips always curl up forming a perfect smile. And her eyes always gave off an honest yet deceiving look; that would make a guy wonder whatâ€™s going on in this girlâ€™s headâ€¦  I slid down the bench beside her, twirling the bottle of beer in my right hand. I could feel her eyes watching me, it felt as though the sun was focusing its rays on me keeping me warm, or trying to burn me up. So I glanced back at her. Her long black hair just hypnotized me, the way it fell on her face, making her eyes much more seductive. Trying to battle these thoughts, I started a conversationâ€¦ â€œ Soâ€¦ whatâ€™d you think about my opening speech?â€ she swayed to the left laying her right elbow on the edge of the bench, twirling her hair and was thinking of a response. â€œ Wellâ€¦ you forgot to use what you had written on paper, but that was cool you still kicked it off pretty well.â€ That was rightâ€¦ I did write what I was to say to kick off the party, I mustâ€™ve forgot. I drank what was left in the bottle, and motioned Cherry to pass the bottle of Cuervo. â€œ I thought. Hey itâ€™s best to say what I felt about the party, and it wouldâ€™ve been awkward if I whipped out a piece of paper reading it in front of the peeps.â€ I told her as I reached for the bottle. I was trying to cover up my little mistake. She looked at me and giggled â€œ you are still trying to be cool eh Mr. Iâ€™m The Bombâ€ after I took my shot from the bottle I passed it back to her, and laughed as wellâ€¦ she took a shot and glared at me once again, smiled, and leaned down and gave me another kiss. I once again had thoughts about what was going on, receiving images of her and me in the bedroom stripping off that blouse and caressing those size 38 c cupsâ€¦ how I wish that shit would come trueâ€¦ I stood up abruptly; my phone was ringing, the vibration in my back pocket startled me. I asked Cherry to excuse me while I answer the callâ€¦  I stepped over by the pool where the least noise was. It was Rachel; she has been looking for meâ€¦ â€œ Chris, Where are you?â€ she sounded worried and pissed at the same timeâ€¦ â€œ Rachel, Iâ€™m hosting a party for a friend of mine, remember Cherry? â€¦â€ she raised her voice a tad bit, and I could hear the blender start in the backgroundâ€¦ â€œ Cherry? Yeah, I remember herâ€¦ why didnâ€™t you tell me?â€ She said, â€œ I didâ€¦ didnâ€™t I?â€ I replied trying to get my head straightened out. â€œ Look baby, Iâ€™m sorry it mustâ€™ve slipped my mi-â€œ she cut me off retorting â€œ look Chris if Iâ€™d known I wouldnâ€™t have been looking for you and calling youâ€¦ just be sure you get back home tomorrow, or itâ€™s over!â€ she put the phone down on me. Rachel and I have been together for only a few months, but she feels as though we were already married, I donâ€™t understand why some girls are like that, jealous and always breathing down my neck, always wanting to know what Iâ€™m doing, and easily get pissed off when I mention Iâ€™m with another girl; maybe they know what Iâ€™m up toâ€¦ I shrugged it off, and went back to the bench to look for Cherry. I wanted to forget about the call; it would just ruin my nightâ€¦ I passed the numerous bodies that were lying by the poolside, it reminded me of a scene in one of those teen movies I forgot which one but it was exactly the same; the bodies on the ground drinks everywhere and people making out in the dark corners, loud trance music coming from all sidesâ€¦ I reached the bench but I didnâ€™t see Cherry there instead I saw a couple sipping drinks and having a light conversation. â€œ Fuck it, Iâ€™m going to the bar get myself another drinkâ€ I thought to myself. I once again stepped into the dining room it was less crowded, the time was at around 11:45â€¦ it was easier for me to get to the bar, without all the peeps pushing around no more.   I saw Horry trying to make some moves on some Chinese looking girls, I could see he was doing a good job, the girls were smiling thatâ€™s a good sign, another was they were exchanging wordsâ€¦ I sat away from where Horry was, and asked Edong for a glass of whiskyâ€¦ Edong passed me the bottle and told me to take the whole thing instead, I thanked him and made my way to the living room where I set the beanbags, I thought I could just enjoy my bottle of whisky thereâ€¦ at this time Chicane was being played, Daylight, damn this DJ knows what I listen toâ€¦. my head throbbed harder as I lounged around, I slouched down in one of the bean bags, bottle in hand enjoying the sounds and the buzz I was havingâ€¦ when a familiar face stepped into the living room. It was Sancho in his usual blue Pornstar shirt and khakis, always carrying his pouch, which is most of the time; filled with drugsâ€¦ he passed by and motioned me to follow him. I was back at the bar, Horry already went upstairs with one of the chinay girls, and I ordered a glass of vodka for Sanch. Edong took one look at Sancho and gave him the bottle of vodka, not the glassâ€¦ he did the same thing he did to me to Sancho. â€œ Yo Edong! Why are you giving us the bottles bro when all we want is the glass?â€ I asked Edong in curiosity. â€œ You guys are heavy drinkers, I can tell and donâ€™t worry itâ€™s all paid for, donâ€™t you remember?â€ Edong told me in reply. â€œOh, yeah right.â€ With that being said, Sanch and me made our way to the poolside.  I grabbed a magazine from the coffee table in the living room, and stepped out to the poolside with Sanch. We whipped out the bag of Chronic from his pouch. Beautiful stems and tiny green leaves that gave off a musky smell. We then peeled and shredded the leaves on the magazine as not to waste any, Sancho ate some seeds in the process, he said it would fuck him up earlier, this guy. After we got the leaves set, I whipped out the Rizla from my pocket and rolled and packed it tight with my make-believe rolling machine, which was a hundred peso bill. The ritual was almost complete, we were just waiting for one more elementâ€¦ The Bitches.  Sancho and I had the joint set up we just waited for one more thing and that was Andrew our best friend pimp to hook us upâ€¦ we were conversing about how the world wouldâ€™ve been a better place to live in if people in the government didnâ€™t judge addicts cause of their lifestyles, we are people as well that have a different way of looking at life but still try and maintain the way ordinary people liveâ€¦ we waited for about 10- 15 min and Andrew arrived, he was sweating as hell, and his right hand shook more than usualâ€¦ â€œHey guys, sorry I ran a little late I got confused with the place.â€ He took a seat across me and took the bottle of Vodka from Sanchoâ€™s hand. â€œGo ahead why donâ€™t you? Just grab my bottle and drink awayâ€¦â€ Sancho said sarcastically, as he rose from his seat slowlyâ€¦ â€œ Iâ€™m sorry bro, I just needed a drinkâ€¦â€ he took a shot and gave back the bottle to Sanch. â€œI had a run in with some friends of mine and I couldnâ€™t pay them back so I had to give them what I promised you guys, then I got lost; these streets all look the same dudeâ€¦â€ Andrew said shaking his head, and started reaching for my bottle, I took it away from his reach, saying no through my eyesâ€¦ â€œ Look at thisâ€ I said showing the joint I rolledâ€¦ â€œItâ€™s cool Andrew, I know weâ€™re missing the free fucks you were supposed to bring, but as long as we are here we might as well enjoy the party; do you know that the percentage of hot horny girls in this party is 78%, just look at the crowdâ€¦ for sure weâ€™ll catch some, I know I willâ€ I said as I passed the joint to Sanchoâ€¦ â€œLetâ€™s just smoke this joint and mingle aroundâ€¦â€ Sancho told Andrew. Andrew looked relieved and tried to reach for my bottle once again. â€œFuck dude! Here! The bottles yours, lazy ass fuckâ€ I told Andrew and figured Iâ€™d get myself another bottle and bring someone along for the session. I rose from my seat and told the guys to wait for meâ€¦  I dropped by the bar once more, Cherry was there entertaining some of her guests and Horry snuck behind her and gave her a hug. I whistled for them. Cherry glanced at me and spoke in Horryâ€™s ear, Horry looked at my direction, and I motioned for them to come to meâ€¦ it was 12:30 and my mind wasnâ€™t that straight anymore. They excused themselves from the guests and walked to my direction, each holding a bottle of beer. They took the two seats beside me that were unoccupied. Horry dropped down and sighedâ€¦ â€œBro this night is turning up all right, I got laid again, although I wish it was with youâ€¦â€ she turned in Cherryâ€™s direction. Cherry just gave him the finger and asked me whatâ€™s upâ€¦ â€œ So whatâ€™s up?â€ I leaned over and whispered in her earâ€¦ â€œWell Sancho and I got the stuff ready by the pool I wanted to know if you want a toke as well, you know since you just grew a year olderâ€¦â€ Cherry flashed me her smile again, that meant a yes of courseâ€¦ she told Horry as well, Horry just smiled as if he was already satisfied with his night, and motioned us to go aheadâ€¦ Cherry just gave Horry a sucker punch and pointed out to the direction of Melanie and her groupâ€¦ Horry got the message and split from the bar. I got another bottle of whisky from Edong and stood up with Cherry  Cherry and I then made our way back to the poolside where we saw Sancho and Andrew wrestling on the grassy side by the pool. â€œHey no Fucking Kids allowed in my party!â€ Cherry called out to Sanch and Andrewâ€¦ â€œWell excuse us Miss I just turned 19!â€ Andrew saidâ€¦ the idiot. I walked over there and gave him a punch on the shoulder. They broke their little melee and both fixed themselves up and sat down joint in handâ€¦ we lit that baby upâ€¦ Sancho as usual got the first hit, then I then Andrew. It was Cherryâ€™s first time to smoke dope so she kept asking us how it was likeâ€¦ â€œJust go for it already Girl nothing bad will happen to you. Itâ€™s just like a cigarette except this cigarette here has an attitudeâ€¦â€ Sancho said, â€œYeah a Fuckin attitude is right Bro!â€ Andrew blurted out as he and Sancho gave high fives to each other. â€œJust suck it hard and hold the smoke in and suck some more!â€ I told her, as I took a hit and passed the roach to Cherry, â€œDude! Chris that didnâ€™t come out right broâ€¦â€ Sancho told meâ€¦ I paused; then the guys and me just burst out laughing. While in our laugh trip Cherry took one good long hitâ€¦ her face just crumpled up as she started coughing it out, she still tried to keep her composure, but couldnâ€™t help it. She put on a face that read: Oh Fuck! I Smoked Too Much! We just laughed and she laughed with usâ€¦   It only took us a minute to finish that joint then we just settled by the pool, talking bout things we saw on TVâ€¦ â€œ You know dudes it sucks to be a sea turtleâ€ I told them as I opened up a conversation â€œâ€¦cause one night right, I was switching channels and I saw two male turtles on top of a female rightâ€¦â€ I took a shot of my whisky â€œ anyways, #1, the turtle on top of the female, was trying to fuck her right? But then #2 came along and started eloping #1 while #1 was eloping the femaleâ€¦â€ I said re-enacting the scene with my hands, â€œand as the program progressed the three were being chased by other sea turtles who wanna get a piece of the action as well, the whole program was just these turtles trying to dislodge #1 to fuck the only female sea turtle in the seaâ€¦â€ I told the story laughingâ€¦ â€œShut the fuck up!â€ Sancho said as he threw his empty bottle at meâ€¦ it hurt, but I didnâ€™t mind and continued my storyâ€¦ Cherry stood up and said sheâ€™d hang out by the bar, me Sanch and Andrew just chilled and listened to the musicâ€¦ after awhile I dozed offâ€¦ next thing I knew Andrew had already left the partyâ€¦  Sancho and Horry met up and conversed about what life could be if drugs were legal. Andrew left no trail, everyone just said he left the party with a bottle of vodka and never came back. I left the party to go and buy cigarettes by the corner store; Cherry followed me to the gate acting paranoid telling me not to leave. I told Cherry to stay put in her room; she was tripping, she was asking me to make her a baby, and in return she would cast a spell on my drawings to make them come to life, she smoked too much, hopefully sheâ€™ll snap out of it in an hour or so. On my way to the store I noticed a black van drive by playing music of Fifty Cent. And it skidded after the first street and ran into a post. Neighbors came rushing out. But I didnâ€™t give a shit I just wanted to buy my cigarettes. The time was around 1:00 am and I was lostâ€¦ I didnâ€™t know where I was going anymore. â€œFuck I shouldâ€™ve brought someone along with me.â€ I thought to myself. Iâ€™ve been lost now for 10 min. Iâ€™ve been going round and round and these streets all started to look alike. I decided to go back to the house hoping I could find my way there. Then Sanch called my phone upâ€¦ â€œWhere you bro?â€ he said it with a hint of worry blended with sarcasmâ€¦ â€œBro you wonâ€™t believe this but Iâ€™m lost bro, itâ€™s fucking surrealâ€¦ bro could you drive around and look for me?â€ I asked hoping he took it seriouslyâ€¦ â€œThatâ€™s what I am doing right now just sit tight Iâ€™ll find you alright?â€ with that he put the phone down on me and I checked my pockets for some stray cigarettes that I might still have and sat by the curbâ€¦  So I was there, sitting on the curb next to the Bluebird street sign, waiting for Sanchoâ€¦ I noticed the sky filled with lights, and heard the engines of drag racers from afar; the air was cool and it gently brushed through me once in awhile. I was imagining if I could live life like this, no problems just listening to the world and never getting involved in itâ€™s problems, no one to bug you, no one to tell you what to do, living without conflict, living life just plain, just simple, just peaceful, with a buzz in your head and a joint in your handâ€¦ I was out there, just imaginingâ€¦ it took fifteen minutes for Sancho to find me, from afar I could already hear his music drawing closer, Sancho always loved Telepopmusiksâ€™ Breathe, he played it whenever he had a chance toâ€¦ his blood red gallant stopped in front of me and he opted me to jump inâ€¦ â€œBro kill your sounds and sit with me for awhileâ€¦â€ I asked Sanch in my laidback voiceâ€¦ â€œI just had a vision, a feeling to share with youâ€¦â€ Sancho lowered the volume, and stepped out. He took a long look at me and gave me a whack on the headâ€¦ â€œYou Idiot! Donâ€™t be tripping out just yet broâ€¦ itâ€™s still early.â€ He reached out his left hand and gave me a boost up. I groggily walked to the passenger seat telling Sancho I needed a follow up. Sancho laughed and got in, turned the music up and we drove off back to Cherryâ€™s place.  â€œBro, donâ€™t crash on me nowâ€¦â€ Sancho shook me upâ€¦ I fell asleep againâ€¦ I was just waiting to get back to Cherryâ€™s house so that I could get a drink, and hopefully get lucky in having a fine fuck for the evening. I could already see the gate; it was only a couple of meters awayâ€¦ I shook myself up; it was around 1:55 in the morning. We stopped for a minute, to smoke ourselves a roach. We stepped out and walked to the entrance. Damn, the tall majestic mahogany gate looked like the gates of heavenâ€¦ or hellâ€¦ I couldnâ€™t decide on what should be what no more. The car park inside was still pretty packed but the sounds have moved from trance to hip-hop, it wouldnâ€™t be the same trip for me from the trip I was having earlier that night. We moved in straight to the bar, Cherry was there, I was glad to see she snapped out of it earlier than what I had expected. She welcomed us back with a smile and pointed to the poolside, where Horry was, half naked and red as hellâ€¦ Sancho and I just laughedâ€¦ then Sancho made his way to the poolside, opting me to go with him, I told him Iâ€™d rather chill by the barâ€¦ Edong wasnâ€™t there anymore instead; a guy at around my age was there, mixing some Bloody Marys and Kamikazes. I ask him to pass me a bottle of whisky, which he did without question. â€œHey bartender, what happened to Edong Bro?â€ I asked himâ€¦ â€œOh Edong, he left awhile ago sir, itâ€™s my shift nowâ€¦ by the way my nameâ€™s Rudyâ€ the bartender repliedâ€¦ â€œOk, thenâ€¦ Rudyâ€ I could tell this bartender was a fag, the way that faggot spoke, and how that faggot gave me his name with a faggot-like smileâ€¦ I despise faggots; all kindsâ€¦  I sat beside Cherry, hoping she got her head straightened outâ€¦ â€œSo Chris, I heard you got lost,â€ she said giggling a bitâ€¦ â€œYeah I did. I was going to the corner store, you know just to buy cigs, but I just found myself lost.â€ â€œYou stoner! I donâ€™t think there is a corner store around here,â€ she giggledâ€¦ â€œHey at least I wasnâ€™t trading a fuck for magical drawing spells!â€ I replied laughing a bit. She stopped for a second and askedâ€¦ â€œI did? Did I say such a thing?â€ She just laughed some more, till her strap broke offâ€¦ â€œShit! My strap brokeâ€¦â€ she said trying to hold her blouse together. â€œChris could you take me to my room so that I could change?â€ she looked at me with those eyes againâ€¦ I thought, â€œDamn, take her upstairs to change, or to fuck? I canâ€™t think straight anymoreâ€, I snapped out of it and said yes trying to just keep in mind clean thoughtsâ€¦  I was walking past all the drunken people lying on the floor following Cherry up to her room. Her house was a mess after only a few hours, damn what would you expect to happen in a house party anyways right? I stopped by her bedroom door, and told her Iâ€™d rather wait outside for her, â€œSuit yourself, Iâ€™ll just leave the door unlocked is that ok with you?â€ she said flashing me with a look in her eyesâ€¦ she then gracefully entered her room; and as promised left it unlockedâ€¦ â€œIâ€™m not falling for that shit,â€ I told myself, knowing that I think I know what she had in mind, or was it my mind? After only a few short seconds I just said fuck it and entered her room locking the door behind meâ€¦ she was on the bed as I had imagined, waiting for me to enterâ€¦ â€œShit youâ€™ve planned this whole thing havenâ€™t you? Eh Cherry?â€ I said to her as I paced the roomâ€¦ â€œSo Chris, youâ€™ve known about this and still fell for itâ€¦â€ she smiled at me â€œWhy?â€ Fuck! What was I going to say? I was already thinking of words, till I just saidâ€¦ â€œYou made me horny the whole time at the party, and I already knew you had a thing for me when we first chatted by the barâ€¦â€ I said, slowing down my pacingâ€¦ â€œAnd I just followed the path laid before me so I could get the truth.â€ I said as I walked up to her on the bedâ€¦ â€œSo do you want to know the truth?â€ she said to me flashing that smile and that glint in her eye againâ€¦ I was readyâ€¦  We just fucked each otherâ€™s brains out; it was all just a blur to meâ€¦ Cherry was lying beside me, she had passed outâ€¦ I took my jeans, put them on, gave Cherry a soft kiss on the forehead and went down back to the party, or what was left of itâ€¦ I didnâ€™t bother checking the time anymoreâ€¦ I staggered down the stairs and passed the bar to get myself a drinkâ€¦ I sat down; only a few people now were to be seen, mostly consisting of Cherryâ€™s good friends and a couple of outsiders trying to hook themselves up for the nightâ€¦ Rudy the faggot bartender popped his faggot head from out of nowhere, and ask me for my drinkâ€¦ I just told the faggot to pass me a bottle of whiskyâ€¦ â€œGood night sir?â€ Rudy the faggot bartender asked me, trying to open up a conversationâ€¦ â€œLook here just give me my bottle, Iâ€™m not in the mood to talk to anyone all rightâ€¦â€ I told him sternly. He gave me my bottle and I went around the house to look for Sancho. My first stop in the house was the buffet table at the dining hall; I wanted to chow for a whileâ€¦ there were some wanton noodles, barbecue, and cold beefsteaks, I grabbed myself a plate and took a sample of each. I didnâ€™t mind if the food was cold, I just wanted to fill my stomach upâ€¦ I sat down in one of the chairs by the dining table and commenced my eatingâ€¦ â€œFuckâ€¦ this night was a good nightâ€¦â€ I told myself as I took shots from my bottle. I finished eating and left my plate then and there on the table. Holding my bottle I headed for the first place Sancho would be, the poolsideâ€¦ as I walked to the poolside, I felt my ankle start to ache and my head throb itâ€™s hardest, â€œFUCK!â€ I shouted as I threw the bottle of whisky on the ground. I fell down on my knees clutching my head and trying to shake this pain offâ€¦  Sancho caught me puking my guts offâ€¦ it was the first time this had happened to me, and what made me freak out that it wasnâ€™t just stomach juice but bloodâ€¦ â€œFuck! Chris!â€ I heard Sancho as he was running to help meâ€¦ â€œBro, I guess you messed yourself up this night ehâ€¦â€ Sancho said as he gave me a boost up and guided me to the benchâ€¦ I couldnâ€™t believe what happened, but I asked Sancho to get me another bottle of whisky at the bar, Sancho ran to get me what I orderedâ€¦ I just laid back and lit myself a cigarette while I waited, my buzz disappeared and I needed a follow up, even if it killed meâ€¦  My head was throbbing with pain and my body started numbing up. I could only wait for Sancho to bring me my bottle as I waited at the bench. â€œDamn I overdid myself this night,â€ I told myself as I rocked back and forth in the waitâ€¦ Sancho arrived with my bottle; he brought Horry along with himâ€¦ â€œBro I canâ€™t believe what Sancho told meâ€¦â€ Horry said laughing, I took the bottle from Sanchoâ€™s hand and drankâ€¦ â€œAnd I canâ€™t believe my eyes! You still continue after puking blood!â€ Sancho added as he and Horry stepped backâ€¦ â€œFuck bro, like I told you guys before nothing and I mean nothing can stop my self-destructionâ€¦â€ I told them as I took another shotâ€¦ â€œItâ€™s something that you guys will never understandâ€¦ this disease I have, this fucking Addiction it can never stopâ€¦ so just leave me be for awhile, I need to stabilize myself ok?â€ I told them and they went off.   I lumbered off back in the house into the living room. I dropped into one of the beanbags and lit a cigaretteâ€¦ â€œIf Iâ€™m to die tonight, at least I died doing what I loved mostâ€¦â€ I told myself aloudâ€¦ my mind was a blur with images of my past and I just enjoyed itâ€¦ after a few minutes, I had already finished the bottle of whiskyâ€¦ I stood up and headed to the DJâ€™s booth and told him to splitâ€¦ â€œYour services are no longer required dude, you can go home, or hang if you want ok?â€ I told him as I gave him a slap on his shoulder, no one was here anymore to listen or dance anyway, just the few friends of Cherryâ€¦ â€œDude thanks man, Iâ€™ll be by the bar, go ahead play whatever you wantâ€¦â€ The DJ told me as he left his CD case with me and headed to the barâ€¦ I flagged through his case and took out the Sublime Greatest hits CD and played Badfish, my song to end the nightâ€¦ it was already 3:30 in the morningâ€¦ I roamed around Cherryâ€™s house with Badfish playing. It occurred to me that Bradley Nowell and I had a lot in common, except of course he died years ago from an overdose and Iâ€™m still walkingâ€¦ Our lives were pretty much the same, and I thought I would probably end up the same state heâ€™s in if I continued this a couple more yearsâ€¦ â€œCoolâ€ I told myself as I fell to the floor and drifted off to sleep thinking I wouldnâ€™t wake up anymoreâ€¦  â€œWake up Chris!â€ it was Cherry, she was slapping my face but I couldnâ€™t feel it, in fact I couldnâ€™t hear anything that wellâ€¦ I didnâ€™t even know where the fuck I was alreadyâ€¦ I just stood up and I could see the sun start to shine over the buildings, there wasnâ€™t any sign of Horry, Sancho, or anybodyâ€¦ it was just Cherry and I, but where? Well she was in front of me, moving her lips I couldnâ€™t make out the words she was trying to say, what am I trying to say, I couldnâ€™t even hear themâ€¦ my head was jarred with images and flashbacks, my head felt like it was being worked on with a jackhammerâ€¦ ouchâ€¦ I could already feel my legs but as soon as I did, I fell hard on the floorâ€¦ my legs were twitching and my ears started to ringâ€¦ Cherry helped me up and brought me to the kitchen, she had a hard time cause I kept on tripping over some junk lying on the floorâ€¦   We finally got to the kitchen and she guided me to a chairâ€¦ â€œFuck! My head feels like it just wants to explode!â€ I told Cherry as she walked to the fridge and took a pitcher of water and handed it to meâ€¦ her lips formed the words â€˜drink upâ€™ I still couldnâ€™t hear shitâ€¦ I took the pitcher and drank. Fuck it wasnâ€™t waterâ€¦ It was Gin! â€œWhat are you trying to do girl?â€ I said in my shocked state of mind fully awake and pumping with adrenalineâ€¦ â€œAt least youâ€™re awake nowâ€¦â€ Cherry said giggling a bitâ€¦ yeah it was a good trickâ€¦ â€œNow hand me a pitcher of waterâ€¦ For real this timeâ€¦â€ I said as I tried to stand upâ€¦ â€œSancho and Horry left after you crashedâ€¦â€ Cherry said as she went back to the fridgeâ€¦ â€œThey asked me to tell you to call them up when you woke upâ€¦â€ she came back with a pitcher, with water perhaps, and an empty glassâ€¦ â€œAlrightâ€¦ What time is tit?â€ I said as I groggily reached for the pitcher and glassâ€¦ â€œTit? Did you just ask me â€˜what time is tit?â€ Cherry said laughingâ€¦ â€œWell, Chris the time I know is 7:00 am. In tit time Iâ€™m not so sureâ€¦â€ she replied laughing a bit moreâ€¦ â€œHey! Donâ€™t trip on a guy who just woke up fucked as hell alright?â€ I said laughing a bitâ€¦ Cherry just smiled again and excused herself from the kitchenâ€¦ I finished the pitcher of water and went out to the poolsideâ€¦ I guess Cherry and I werenâ€™t alone after all, Sancho left his pouch of goodies; it was floating on top of the pool. I jumped in to get itâ€¦  1 The phone rang for about a minute then finally, Sancho answered. â€œ Aiyyo! Who this?â€ I could tell he was fucked up as well but I knew this guys in the condition to drive. â€œ This Chris, look dude I need you to pick me up bro, Horry the fuck just ditched my ass for some hoes dude.â€ I could already hear Sancho moving around over the line. â€œ Alright, where you? Iâ€™ll be there in five minutes alright, let me just cleanup this shitâ€¦â€ Sancho started slurring overâ€¦ â€œ Hey Bro, you got buds man?â€ I asked him as I found a spot to sit inâ€¦ â€œ Yeah man, shit Iâ€™m slurring again. Look just tell me where you are Iâ€™ma be there.â€ I gave him the location and put down the phoneâ€¦ it was already 6:30 p.m. and I was ready to crash. I got my busted ass notebook out and scanned through it, once again I started reminiscing bout what happened.   My train of thought was shattered, as a bus came to a screeching halt right in front of me, the bus was a fuckin rusty big pile of metal with wheels and it just stayed put for about five seconds. Then the doors slid open when a conductor popped his head asking where I was headedâ€¦ I told the conductor to buzz off, he continued asking the other people that were on the bend as well, waving his wad of stinking bus tickets in his hand. A couple passengers stepped out. One of which I noticedâ€¦ it was Bart. I called to him but he didnâ€™t pay attention, so I stood up from my place and gave him a slap on his back. He mustâ€™ve been wired with speed cause he reacted faster than a fucking rabbit could and whipped out a knifeâ€¦ â€œ Whoa! You Fucker itâ€™s just me bro!â€ I replied with a bit of strain in my voice. He took a look at me and said. â€œ Bro Fuck sorry man, I thought you were a cop or a snatcher or someoneâ€¦â€ he replied as he put aside his knife in his jacket pocket. â€œ â€¦By the way what are you doing here?â€ I explained to him the situation. In turn I asked him â€œ So what are you doinâ€™ here as well?â€ he shifted the placing of his feet a couple of times and said â€œ HUH?â€ fuck heâ€™s messed up, but I asked the same question again. â€œ Oh me Iâ€™m here cause my place is just down this road, want to chill there and wait for Sancho there?â€ I shook my head declining his invitation, reminding him what happened the last time I was in his place. â€œ Dude, remember three weeks ago? We chilled in your place, me, Sanch, Horry and Andrew, and then while we were stoned we got raided and you split like a frickin jackrabbit, we spent how many nights in jail till Sanchoâ€™s brads pulled some stringsâ€¦â€ He just shook his head and snickered apologizing for what happened. â€œ Bart, you wired man?â€ he shook his head shuddered his shoulders and just continued snickeringâ€¦ at that point I was getting annoyed till he asked me bout my party last night.    Sancho finally arrived there in his blood red Toyota gallant, sounds booming from his speakers. As he stepped out, smoke billowed out. Damn the moving hot box is in action once again. â€œ Aiyyo! Chris, whatâ€™s Barty boy doin here with you?â€ Sancho asked in a disgruntled way. â€œ He got off the bus I forgot he just lived down the street, I couldâ€™ve told you to meet me there, but you remember last timeâ€¦â€ I told Sanch, laughing a bit. â€œYou are right man, next time, Bart Yo! Donâ€™t split like a frickin tadpole when the cops come man.â€ Sancho said laughing his stoned ass off. I walked over to the passenger door bid farewell to Bart and took my notebook and jotted down a short and meaningless poem, Sanch followed a few seconds later closed the door and put the volume down. â€œ Bro put some ganja music in, and tell me where you want to go.â€ I reached for the CD case in the back seat; I always left my music in Sanchoâ€™s car for some strange reasonâ€¦ I took out my reggae CD that contained Bob Marley And Yellowman stuck it in and turned up the volume. â€œ Sanch, bro, I canâ€™t go home in this condition, Rachelâ€™s gonna give me another sermon againâ€¦â€ I said shaking my head. â€œ Letâ€™s just chill around in your placeâ€¦ letâ€™s just pass by 7-eleven and buy some beers.â€ We drove off and Sancho loaded the pipeâ€¦ I took a toke and passed it back to Sanch. Sancho always had the good shit from Amsterdam. I had a good buzz and asked Sancho to wake me up once we reached 7-eleven. I drifted off to sleepâ€¦  Chapter 2: The stop at 7-eleven  We were almost at 7-eleven when Sancho woke me upâ€¦ â€œHey Broâ€¦ Bro! Wake the fuck up weâ€™re near 7-eleven alreadyâ€¦â€ he shook me up I was nudging him off till finally his continuous shaking woke me up. â€œ FUCKIN A! Bro I was sleepinâ€™ manâ€¦â€ He raised his right hand and gave me a light whack on the head, reminding me that I asked him to wake me upâ€¦ â€œ Oh yeah I forgotâ€¦â€ I said laughing still hit from the toke I had. We stepped down the car, after hiding the shit we smoked up in the carâ€¦ I was reminded that I didnâ€™t have any cash so I asked Sanch if I could bum from him. â€œ Bro check your CD caseâ€¦ I think you left a couple of bills theirâ€¦â€ he takes his car keys and tosses them to me. â€œMeet me inside alrightâ€¦â€ he said as he walked to the store. I walked back to the car and opened the trunk, revealing 10 of my CD cases. â€œWhich CD case did he mean?â€ as I surfed through one of the CD cases, a picture fell; it showed me and Sancho back when we were still in grade 1. Bunch of innocent looking kids, itâ€™s hard to imagine that weâ€™d end up where we are right nowâ€¦ I slipped the picture back and found my money on the fifth case. I locked the car and went to the store. Sancho was by the fridge checking out the beers I tossed him back his car keys, and scavenged around for the munchies. I passed through the fields of plastic wrapped preservative filled dried grain products, and settled on 5 packs of Doritos, and I then walked to the counter and paid for the chips and bought two packs of Marlboros. Sancho followed with a twelve pack of san mig lights. We stepped outside with our supplies and headed for the carâ€¦ we sat down and drove off. I rolled down my window feeling the air blow through my face, I reached for the pack of cigarettes and lit myself one. I then took my share of the beer and had a huge gulpâ€¦  Minutes after we left 7-eleven the radio announced that the 7-eleven we were in just got robbed and the clerk was shot 5 times and many were wounded, me and Sancho just sighed and thanked god for our lucky starsâ€¦ half way through my first can we had already reached Sanchoâ€™s village in White Plainsâ€¦ we caught up with Andrew by the gateâ€¦ he was staggering with a deep gash on his left cheek. We asked him to hop in. Andrew, the fuck, always got himself fucked in situations, we told him to chill at Sanchâ€™s place so we could take care of the woundâ€¦ apparently he had a run in with some major underground group last nightâ€¦ that made things clearer for me.  Chapter 3: Andrewâ€™s Issue  Andrew was lying on the couch feeling the gash on his cheek; Sancho was in the bathroom lighting up a jointâ€¦ he was also trying to contact Horry. In situations like these Horry was good helpâ€¦ I was pacing around the room thinking whether I should go home or stay with the guys; I was still wearing the same clothes from last night and I didnâ€™t have a bite to eat; and I meant real food. The time was already 9:00. We cleaned Andrews wound up, we dropped some hydrogen peroxide on his face and Sancho sewn his face up and patched it with a square band-aidâ€¦ Apparently after Andrew left the party he had a run in with some major fraternity he had a beef with two months agoâ€¦   â€œI was drunk as a mother fucker last night I was walking towards my car and next thing I knew a couple of big tough punks from APO pulled a 45â€™ on me telling me to hop in the vanâ€¦â€ Andrew walked towards Sanchoâ€™s fridge and got himself a can of beer and helped himself to a bag of Doritosâ€¦ â€œFuck bro if I only knew I was gonna bump into them I couldâ€™ve brought my shot gun alongâ€¦â€ he finished his beer in 4 huge gulps and grabbed another oneâ€¦ â€œThe funny thing about it was after a few minutes I was in the van, getting my ass whooped and threatened, the driver crashed into a post trying to avoid an idiot standing in the middle of the road.â€ I stood from the bed and grabbed myself a beer as wellâ€¦ â€œWell I guess you should thank me broâ€¦â€ I said laughing a bitâ€¦ Andrew replied asking whyâ€¦ â€œI was there. I thought some kids were just drunk when they crashed their ride they were driving hella fast, so I didnâ€™t pay attention cause I was pretty stonedâ€¦â€ Andrew just stared at me and laughedâ€¦ â€œI guess I should thank you for the state Iâ€™m in as well eh bro?â€ I just chuckled; Andrew was trying to be seriousâ€¦ â€œLook, bro the driver died instantly in the crash, the others were knocked out and I busted my knee and got fuckin cut in the faceâ€¦â€ â€œHey itâ€™s better than lying face down on a pool of blood or your body riddled with bullets; oh wait the meant the same thing didnâ€™t it?â€ I told Andrew chuckling on the last bit of my sentenceâ€¦ Andrew explained to me that he owed these guys some cash and that he has up to next week to pay them twenty grandâ€¦ at this point Horry arrived.  Upon entering the room Horry once again apologized for ditching me at the bus station, he looked tired as mofo. I accepted his apology and told him about Andrewâ€™s situationâ€¦ he looked at Andrew as he lit a cigaretteâ€¦ â€œSo tell me, what did you do to owe these guys so much?â€ Horry asked Andrew in a serious tone of voiceâ€¦ â€œWell Iâ€¦ Well I rammed my V-tec in one of the OGâ€™s Expedition, I promised I was to pay them back, but that was two months agoâ€¦ andâ€¦ I was ashamed to ask cash from you guysâ€¦â€ Andrew said shifting his position on the chair. â€œSo you let it escalate to the point that they had to kidnap youâ€¦ just because you didnâ€™t pay for the carâ€™s bill of repair? Thatâ€™s bullshit! Youâ€™re hiding something from usâ€¦ arenâ€™t you?â€ Sancho stepped in from the bathroom; apparently he heard everything from insideâ€¦ â€œ Cause hey, I wouldnâ€™t go to such an extent to kidnap a guy and threaten him just to pay for a repair bill that wasâ€¦ 2 months ago you say?â€ Sancho paced to his fridge got a bottle of vodka and took a shotâ€¦ â€œSure Iâ€™d remind the guy a couple of times to pay me back, but what kind of guy would do this shit to you if you didnâ€™t do anything else to him? Tell me broâ€¦ tell us what else you did.â€ Sancho said passing the bottle to Andrewâ€¦ I just sat there reached for Sanchoâ€™s pouch, rolled myself a joint and hit the bathroom; I wasnâ€™t in the mood to hear anymore problemsâ€¦  I got out after I finished the joint; Andrew had his hands on his face shaking left and right, spilling tears on his bright red shirt. Sanch and Horry got the truth from him; I didnâ€™t bother asking anymore what happened I was just ready to do what I can doâ€¦ â€œChris, grab the 12 gauge under the sink in the kitchen, you know what to doâ€¦â€ Sancho told meâ€¦ â€œAndrew you stay hereâ€¦â€ Andrew just nodded, then Sancho gave Horry and me a look only we knewâ€¦ we hit the road in Sanchoâ€™s muscle car, a modified Land Cruiser, he had it repainted black and had the windows bulletproofed, as usual had his speaker set up, and believe it or not, he added a NOS system that was under the passenger seat. We reached the gate of White Plains, our windows rolled down, a cig in each handâ€¦ â€œChris, bro I know weâ€™ve done this before, but as usual I need to remind you, cause youâ€™re stonedâ€¦ Againâ€¦â€ Horry told meâ€¦ â€œWeâ€™ll just talk to the guy alright, scare the guy as usual maybe ask a favor from him, you know give Andrew some time. You stay outside and if things go wrong, pop some shells got it?â€ I nodded, trying to focus myself. We were already driving on Edsa by Crossings; I was enjoying my cig and the lights of buildings and cars passing by. By the time we reached Ayala I was still hit hard, I thought I could get a drink once we reach WhereElse! I sat lower in my seat just thinking to myselfâ€¦  â€œFuck Iâ€™m doing this shit againâ€¦â€  We looked for the best parking spot in the parking lot in front of the club; we found the perfect spot after a minute. I had a great view of the entrance, no obstructions I could get a clear hit if anyone who came rushing out the door. Sancho, Horry and I wished each other the best of luck, so that I wouldnâ€™t have to pull the trigger. They stepped out, reminding me again of what I was to do, and walked into the club. The clock on Sanchoâ€™s dashboard flashed 11:00, theyâ€™d be out in 10 minutes or less if everything went well. I went down and had a cigaretteâ€¦   Chapter 4: Waiting for Ten Minutes  I finished my cigarette. My head was buzzing with images of what was going inside the clubâ€¦ Sancho would introduce himself and Horry, then Sanch would say his usual statement to startle whomever the guy was, he would start the conversation firmly, then Horry would mediate the situation if things ever got out of hand. He then would rephrase Sanchoâ€™s words in a more understandable, but meaningful statement. Probably at this point, the guy would have two things in mind, he would either give consideration, or fuck these two guys and do whatever the fuck he wanted to do. If ever the second thought came into mind, I would step in do my partâ€¦ hopefully he would give Andrew the chance, I didnâ€™t have the time to get my shit ruined again, and after all Iâ€™ve gone throughâ€¦ I remember the first few times we did this, give a talk to a trouble maker, we were all young back then and our conversations usually ended with the other guyâ€™s head blown off; of course I did all the dirty workâ€¦ I was never really good with words, Iâ€™m more of the action type of guyâ€¦ although lately our meetings all went smoothly, I didnâ€™t have to blow anymore brains, and we also always kept our end of the bargain, so both ends had good dealsâ€¦ hopefully weâ€™ll get lucky againâ€¦ as I lit another cig, my phone rang. â€œChris, Where are you? Again!â€ it was Rachel, damn this bitchâ€¦ I never get a single day free of this bitchâ€™s complaints.  â€œBaby, I know I was supposed to be home this day, but something came upâ€¦â€ I said still trying to keep my emotions intact. â€œWhat is with you? Youâ€™ve been hanging around Sanch, and Horry all the time; you and those fuckin assholes called friendsâ€¦â€ she said, fuck I couldnâ€™t believe that she was dragging my friends into this shit between usâ€¦ â€œLook here bitch! My friends have nothing to do with us all right?â€ I shouted over the phone trying to make a point. â€œLook donâ€™t be breathing down my neck all the time alright! You wouldnâ€™t, I repeat you wouldnâ€™t understand the shit Iâ€™ve been doing even if I wrote it in black or whiteâ€¦â€ I saidâ€¦ I couldnâ€™t keep myself in check anymoreâ€¦ â€œ The best thing for you to do is get the fuck out of my apartment and erase my number from your phone.â€ I lit myself another cigarette, as I waited for her responseâ€¦ â€œChrisâ€¦ Iâ€™m sorryâ€¦â€ she was sobbing on the other endâ€¦ my heart sank again and I refreshed my thoughts and feelingsâ€¦ â€œLook donâ€™t cry alright, Iâ€™m sorry I shouted ok, you just caught me in a bad timeâ€¦â€ I took a puff from my cig and threw itâ€¦ â€œYeah, I know I havenâ€™t seen you lately but try and understand that Iâ€™ve just went through a great ordeal, I needed some time alone, spend some time with the guysâ€¦ cause shit theyâ€™re the only ones who could screw my head back together again alright?â€ I got in the Land Cruiser and checked the time; five minutes have passedâ€¦ â€œLook I promise you, Iâ€™ll be home tomorrow ok?â€ I said reassuringly in a lower tone. â€œOkâ€¦â€ she replied her sobbing started to fade, and put the phone down on meâ€¦ Damn it!  I got back into the car and checked the time another two minutes have passed, that made a total of seven, shit I hope things were going well inside, I rolled down the window and had a smokeâ€¦ fuck, this day was fucking stressfulâ€¦   Horry called my phone saying it went well, the OG would give Andrew another two months to pay him back for the bill, all he wanted was Andrew to apologize for drugging the OGâ€™s sisters and screwing his ex-girlfriend a year ago. Fuck I was glad I was smoking up in the bathroom when Andrew told Horry and Sancho this; cause if Iâ€™d known I wouldâ€™ve just beat him up then and there whether or not he was a good friend of mineâ€¦ they said that the OG wanted to meet me, I told them I was in the parking lot and I didnâ€™t want to move so they told me to waitâ€¦ â€œOne thing Chris this guy is alone, weâ€™ll be coming out with no one else, once you see him fuck bro try to keep yourself composed alrightâ€¦â€ Horry said to me over the line, I promised him and turned my phone offâ€¦ â€œKeep myself composed? For what? â€ After a minute they were out with the OG, Ricoâ€¦ they walked over to me, I was at the side leaning on the Land Cruiser, then Rico stepped up in front of me laughing, â€œSo this was the guy who was to bust a cap in my assâ€¦â€ Rico said looking up to meâ€¦ god I canâ€™t believe a guy like this was an OG of a major fratâ€¦ this guy was around five foot four inches tall wearing â€œtotoyâ€ hip-hop clothes, I just smiled at the punk tough guy wannabe. â€œYup, Iâ€™m the guyâ€¦â€ I told him battling my thoughts about this fucking shrimp being an OGâ€¦ â€œItâ€™s a good thing you took our terms bro. Iâ€™d hate to see another guy fall down cause of what I might do, you know?â€ I told him I noticed him shiver a bit â€œRico, right? Ricoâ€™s your name? All right nice meeting youâ€¦ hope when we see each other again it wasnâ€™t a product of conflict alright?â€ I motioned Sancho to talk to me. I excused myself from Rico and whispered to Horry to take him back and get me a drink from the clubâ€™s barâ€¦   I waited for them to step back in the clubâ€¦ as they disappeared me and Sancho just burst out laughingâ€¦ â€œA fucking totoy like that is a fucking OG?â€ I said as I struggled to open the door of Sanchoâ€™s rideâ€¦ Sancho just laughed and stepped in as well busting out a cigarette. â€œFuck the dude looked like he was still wearing diapers manâ€¦ â€˜So this is the guy who was to bust a cap in my assâ€™ that tough guy wannabe man!â€ I told Sancho as I lit myself a cigaretteâ€¦ seconds later Horry came out with my beer, hopped in the back seat and we left WhereElse! Horry just started laughing at the situationâ€¦ â€œI canâ€™t believe we were going to shoot down a fucking Kid, I swear to god this kid looked like grade 6â€ Horry saidâ€¦ I drank from my beerâ€¦ Sancho just told Horry to forget about what happened with the totoy and think about what to do with Andrewâ€¦ I drank and just laid backâ€¦  Chapter 5: The Idiot  Me, Sancho and Horry were driving off, back to White Plains, where we had hoped, Andrew was waiting. Passing Edsa once again, only a few cars were on the road accompanied by a bunch of semisâ€¦ â€œBro what should we do about Andrew man?â€ I asked Sancho and Horryâ€¦ â€œI mean, why the fuck did he have to do those things to a fucking shrimp dudeâ€¦ I mean sure the OG was a fucking shrimp, I could tell he was on the verge on crapping his pants when he came up to meâ€¦ Ricoâ€¦ what a fucking stupid ass nameâ€¦ And Andrew is an idiot for picking on oneâ€¦â€ I said venting out some humor and anger at the same timeâ€¦ Sancho and Horry just laughed, â€œHey guys, wanna pass by 7-eleven grab a couple of drinks before we see Andrew again?â€ Horry suggestedâ€¦ â€œSure, why not?â€ Sancho said as he stepped on the gas and made a turnâ€¦ we knew every 7-eleven in this fucking city, every road to take, every nook and cranny just to get to a 7-elevenâ€¦ we finally reached one along Shaw Boulevardâ€¦ They both stepped out. I was starting to doze offâ€¦   â€œChris, get your ass up!â€ Horry said as he opened my doorâ€¦ â€œAlright, alrightâ€¦ Iâ€™m up and walkingâ€¦â€ I said as I got down from the car and walked around to prove myselfâ€¦ â€œSatisfied?â€ Sancho just shook his head and we walked into the storeâ€¦ as usual, Sancho was by the fridge again and I handled the smokes and chowâ€¦ Horry was by the magazine rack reading a PSM magazine that featured Metal Gear Solid 3 on the frontâ€¦ â€œHideo finally came to his senses guysâ€¦â€ Horry said as he browsed through the magazine â€œI canâ€™t believe itâ€™ll be out soonâ€¦â€ Horry dribbled on as we were doing our usual stuff. Horry was the eldest and the youngest in our group. Heâ€™ll be serious in fights and business, but when he turns his back he kind of becomes a 12-year-old kidâ€¦ we finished our routine and headed for White Plainsâ€¦ Sancho whipped his phone out from his pocket and called his house up hoping Andrew would pick it upâ€¦ No one was answeringâ€¦ we figured Andrew crashed after we leftâ€¦ Sancho stepped on the gas anyway just to be sureâ€¦  We arrived at Sanchoâ€™s place, it was around 3:30 AMâ€¦ we lit ourselves a roach in Sanchoâ€™s garage and preceded upstairs with our suppliesâ€¦ on the way upstairs we had noticed bottles of beer and vodka on the floor, Andrew must have drank to forget his worries for awhileâ€¦ as we got to Sanchoâ€™s room we found Andrew on the floor face forward, we had dropped everything we had in our hands and rushed to check Andrewâ€™s conditionâ€¦ â€œBro!â€ we shouted in Andrewâ€™s ear hoping he was just asleep, Sancho took Andrew in his arms and laid him on the sofaâ€¦ we checked his pulseâ€¦ nothing. â€œFuck!â€ Horry shouted as we continued to see if there was any chance we could revive Andrewâ€¦ nothingâ€¦ we were too lateâ€¦   The idiot ODâ€™d on Sanchoâ€™s stash of Ecstasy and Morphine. Sancho was pissedâ€¦ We had a dead friend of ours lying on the sofa and he had to take Sanchoâ€™s stash along with him, wherever he was off toâ€¦ he had broken the syringes that contained adrenaline shots to revive him, I guess he wanted to pass away then and there fearing that we would have a talk with him after we settled our business with the shrimp OGâ€¦ â€œYou Fucking Idiot!â€ Sancho said at the top of his lungs as he was clutching Andrewâ€™s lifeless body in his armsâ€¦ We were all just bummed out finding out one of our good friends have passed awayâ€¦ â€œWhat are we going to say to the parents guys?â€ Horry had said his voice was shaking along with his body. â€œHorry, his parents are deadâ€¦ remember?â€ I reminded Horryâ€¦ come to think of it, after Andrewâ€™s parents have passed away, we were the ones who took care of him; we took care of each otherâ€¦ We were the only family Andrew had, if he ever did consider us as his familyâ€¦   We just sat there, I didnâ€™t bother what was going in their heads, I was busy with getting my head straightened out, I was feeling stoned and having feelings of grief at the same time for what had happened to Andrewâ€¦ that is the worst combination for a guy to have going on in his headâ€¦ â€œHorry, Chris,â€ Sancho said shaking his headâ€¦ â€œGet a Garbage Bag and stuff him in thereâ€¦â€ Sancho said as he started shedding tearsâ€¦ â€œThis is the best we can do for him, pack him up and letâ€™s head to Batangas, he always said heâ€™d love to be buried by the beachsideâ€¦â€ â€œSanchoâ€¦ yeah he did say that.  Well let me get the bagâ€¦â€ I said as I walked out of the room. I had reached the kitchen and opened up the fridge and grabbed me a carton of milk, it was time I drank something else aside from alcoholâ€¦  Chapter 6: Outro  I had grabbed 4 XL sized garbage bags from the kitchen and headed upstairs to where I had left Sancho and Horry with Andrewâ€¦ I still couldnâ€™t believe one of us had fallen, and of all the people, it was the least we had expected; Andrew, the one whose health was pretty much intact, the dude who partied least, the dude who smoked out the least, the dude who basically stayed home frequently, even though he had the cash and the time to go out, he was a clean pimpâ€¦ all the guys thought I would be the first to go; at the rate I was going I thought so as wellâ€¦ Then it would be Sancho and Horry, last wouldâ€™ve been Andrew. Itâ€™s funny come to think of it, what if we had the order wrong?   â€œFuck it Chrisâ€ Sancho had told me, as I stepped into the room â€œLetâ€™s just drive off there now; bring some shovels alongâ€¦ And weâ€™ll pack him then and there insteadâ€¦â€ Sancho had said, â€œI got dibs on the front seatâ€¦â€ Horry blurted out trying to humor the situationâ€¦ it didnâ€™t workâ€¦ we just sat there a few more minutes and Sancho checked the timeâ€¦ â€œItâ€™s 4:15 guysâ€¦ Letâ€™s moveâ€¦â€ Sancho spoke as we all had risen from our seats to carry Andrew downstairsâ€¦ we laid Andrew on the backseat of Sanchoâ€™s gallant, then I went off to grab the shovels, located underneath the sink in the garageâ€¦ â€œI canâ€™t believe we are doing this!â€ I said to the guys as I popped the shovels in the trunkâ€¦ â€œI just canâ€™t believe this shitâ€¦â€ I dribbled on, continuing what I was supposed to say in my headâ€¦ I walked over to the front seat till Sancho spokeâ€¦ â€œChris, you drive, Iâ€™ll spend the trip with Andrew over hereâ€¦â€ Sancho tossed his keys to me and slid into the backseat with Andrew. Horry hopped in the front with meâ€¦   Sancho and Andrew were the closest to each other in our group; cause Sancho was mostly there to give advice and help him out of situations, he was Andrewâ€™s big brother if you looked at it, so it hurt him the most to see his â€˜little brotherâ€™ dead in his house, lying dead in the car now cause of his stash of narcoticsâ€¦ Sancho was blaming himself for leaving Andrew there with all the drugs he could takeâ€¦ we did regret leaving him there, we shouldâ€™ve brought him along with usâ€¦ well there aint no use grieving what has been doneâ€¦ I started the engine and we drove offâ€¦ '),(6188,3394208,'02,June,2004','2004-06-02 00:00:00','   Some people say that being asleep is better than being awakeâ€¦ I used to believe in this concept, who would want to wake up in a harsh reality, when itâ€™s better to live in a state where your dreams come true anyway? Iâ€™ve spent most of my life awake, and what I got were constant problemsâ€¦ wars, conflict, murder, basically we live in a world where people hurt people constantlyâ€¦ we try to blind ourselves from all these with drugs, friends and activities that keeps us from the ultimate truth, which is death, pain and sufferingâ€¦ we eventually hurt people consciously and unconsciouslyâ€¦ but there isnâ€™t anything we can do about itâ€¦ it has dawned on me that without these conflicts, life would be a dull and empty voidâ€¦ we canâ€™t escape the inevitable, we are going to continue hurting, we are going to continue blinding ourselves, and we are going do die eventually in the coming daysâ€¦ at this I would like to say we canâ€™t sleep, live in our dreams and expect everything to go away when we wake up, these things will always be there when we do, so go ahead and sleep, live inside your head and brush these issues off your shoulders, but when you wake up you can only wait for the day to end, to escape this world, this reality, the truth againâ€¦ I know Iâ€™ve gone through a lot and I still wake up cause I know that when after Iâ€™ve gone through these shit Iâ€™ll come out strong and soon after I wonâ€™t feel anything Iâ€™ll be numb from these conflicts and the only thing Iâ€™ll feel is pleasure soon afterâ€¦ '),(6189,3394208,'02,June,2004','2004-06-02 00:00:00','   hmmm... what\'s there to do? i\'m just sitting here typing, watching the day go by... play myself a tune on the guitar, thinking how and why i\'ve ended where i am right now... these letters pop up when i press on a letter at the keyboard showing me that this thing still works... some of these keys were jammed last night... the visualization of my winamp keeps me entertained although the images i see are just bright colors intertwined into each other showing different shapes and such... like red smoke mixing with blue to create this purple hazy dragon... but it changes randomly, different ways of blending the colors moving to the beat of the mp3 files being played... why is it that i have to be so cheap and download pirated music? idunno, i\'m just letting my mind wander off right now... yeah... why do i do this, idunno, maybe cause i love music, and don\'t have the time or money to get my own album?... oh well, music is music it was made for everyone, so i guess yeah it shouldn\'t be sold but it should be open to all those who want it... uh... you may be wondering why i\'m typing here and what am i typing... i\'m bored and i guess its just for the attention... why? cause i need it... don\'t you? doesn\'t everyone need a little attention sometimes? i know i do... hehe, damn, why is it that some people just have this innate desire to do something out of the ordinary just to get noticed? for me it\'s just plain boredom... when i\'m bored i just go crazy... idunno why? what is it that bores me? there are so many fascinating things to be seen and expierienced i\'m just too lazy i guess and being lazy leads to my growth of boredom... why must I wander off into my thoughts? Itâ€™s necessary fo me I guess to know myself a little better, cause idont know who I really am anymoreâ€¦ igotta come up with something to doâ€¦ oh well, i\'ll come up with something probly this... everyday writing The Mind Of The Bored Man yeah first shit oh vell... i gotta split... read this or not i don\'t care, just bored... guys gonna do anything crazy this summer? let me know ayt PEACE! '),(6190,3394208,'26,July,2004','2004-07-26 00:00:00','   Could\'ve sworn I heard you say amen this morning, showing'),(6191,3394208,'23,July,2004','2004-07-23 00:00:00','   In high seas or in low seas   I\'m gonna be your friend,   I\'m gonna be your friend.   In high tide or in low tide,   I\'ll be by your side,   I\'ll be by your side.    &nbsp;  (I heard her praying, praying, praying)I said,    I heard my mother,She was praying (praying, praying, praying)   And the words that she said (the words that she said),   They still linger in my head (lingers in my head),    &nbsp;  She said, \"A child is born in this world,   He needs protection,   God, guide and protect us  ,   When we\'re wrong, please correct us.(when we\'re wrong, correct us).   And stand by me.\" yeah!    &nbsp;  In high seas or in low seas,   I\'m gonna be your friend,   He said, \"I\'m gonna be your friend.\"   And, baby, in high tide or low tide,   I\'ll be by your side,   I\'ll be by your side.    &nbsp;  I said I heard my mother,   She was crying\' (I heard her crying\'), yeah! (crying\', crying\'),   And the tears that she shed (the tears that she shed)   They still linger in my head (lingers in my head)    &nbsp;  She said: \"A child is born in this world,   He needs protection  ,   God, guide and protect us,   When we\'re wrong, (when we\'re wrong), correct us.    &nbsp;  And in high seas or low seas,   I\'m gonna be your friend,   I\'m gonna be your friend.   Said, high tide or low tide,   I\'ll be by your side,   I\'ll be by your side. '),(6192,3394208,'23,July,2004','2004-07-23 00:00:00','   Everyone just listen to the music,&nbsp;just sit back and relax   And don\'t forget that the brews are flowin over here    &nbsp;  Can you hear my guitar crying for your love   And can you feel the bass line move your body   And won\'t you dance to the drumline giving you the beat   And let me fill your minds with my melodic lyrics    &nbsp;  Everyone just listen to the music,&nbsp;just sit back and relax And don\'t forget that the brews are flowin over here    &nbsp;  Are you gonna go meet yourself a stranger   Well if you do remember you gotta be all sweet   Cause these guys are looking for a little action   And these girls are hankering for a little bit of lovin    &nbsp;  Everyone just listen to the music,&nbsp;just sit back and relax   And don\'t forget that the brews are flowin over here    &nbsp;  Or maybe you\'re the chill kind of guy   Chillin by the bar and downing a couple of shots   So just sit back relax and listen to the groove   Cause&nbsp;I also got these girls acting real smooth    &nbsp;  Everyone just listen to the music,&nbsp;just sit back and relax And don\'t forget that the brews are flowin over here    &nbsp;  Relax... Chill... Drink up... This night\'s on us... '),(6193,3394208,'23,July,2004','2004-07-23 00:00:00','   There\'s a place that&nbsp;I know    when people wanna get out of this city    And if you wanna go    I&nbsp;can take you there in a hurry    People walkin round    enjoying the summer days    While the kids are playin    under the sun\'s rays     &nbsp;   All the people give a hell yeah    So just swing everybody swing     &nbsp;   And there\'s a girl that&nbsp;I know    who would go and take all that&nbsp;to&nbsp;me    And if you wanna know    then pack your bags and get ready    Relaxin by the sand    you girlies come over here    While everybody\'s drinking    and smoking up here and there     &nbsp;   All the people give a hell yeah  So just swing everybody swing '),(6194,3394208,'23,July,2004','2004-07-23 00:00:00','   I am being charged with crimes&nbsp;I didn\'t do    Once again these bitches pull their voodoo    Time once again for the red lights to go on    And show you my terrible visions of you    You motherfuckers don\'t even realize    These things&nbsp;I do have always got a reason    But what you accuse me of doing just isn\'t mine    You\'re backstabbing me right before your eyes    I won\'t be burned by your accusations filled with fallacy    I always gotta be the one to blame for every single tragedy    That always has to be something that has befallen upon you    This my song to show the world the fools in you    You don\'t think I know what you\'re all about do you?    Once again you\'re bringing out the demons in me    Don\'t you ever forget I once brought you such great pains    Tell me something... while toying with me, what do you gain?    You plant false information about my reputation    While raising yourself amongst&nbsp;the ranks of this modern society&nbsp;    Don\'t you ever forget you are where you are with my help    Once again you take me for a fool better kick your can    Cause I won\'t be doing it for you this time... kick your own can    Why don\'t you just wither up and die leave me alone    Once again I have to go and do these things to you    Don\'t you ever look at me that way again... you might not be able to the next time '),(6195,3394208,'23,July,2004','2004-07-23 00:00:00','   This&nbsp;man still wanders off into the night   He always has   Because he always let\'s her win the fight   She always has   Friends tell him what he\'s doing isn\'t right   They always have   And they always catch him in a terrible sight   He always has    &nbsp;  Taking tests to prove his innocence   Taking tests to fight for his independence   Taking tests for those with no sense   Taking tests without resistance '),(6196,3394208,'23,July,2004','2004-07-23 00:00:00','   Hello this is Chris your operator    Hi there ma\'am or sir, may I take your orders?    Would you like to try our slice of the day    Why would you wanna try it anyway?    Alright&nbsp;I need to know your location    So I can drive off there with no hesitation    You say you want a pepperoni pizza with some shrooms    Would you want a drink to go with that holmes?    What if I force you to order some dessert?    We got apple pies and shit ice cream, order now and get a shirt    I\'m your pizza guy&nbsp;I can come on over    If I aint over there in 30 minutes, free are your orders    And please don\'t mind if I take a slice    It was a long drive and the rain was cold as ice    And by the way could you give me my tip?    Or could&nbsp;I use the lou?&nbsp;I need to shit    Alright&nbsp;thanks for your hospitality&nbsp;I need to fly    And folks don\'t forget me... I\'m your pizza guy '),(6197,3394208,'23,July,2004','2004-07-23 00:00:00','   I wonder when I\'ll ever see   How much you really mean to me   I\'ve gone spelunking looking for relief   You\'ve been to busy smoking reefs   I guess it\'s time to make a change   Sleep alone tonight never again   Still I\'m gone still I\'m gone   I tire of what&nbsp;I thought was fun    &nbsp;  I\'m tired of you... you tired of me   I\'ve gone spelunking just to see   The friends&nbsp;I know are falling   You\'ve gone away again    &nbsp;  The roof\'s up high and cannot see   The river flows oh so gently   I smoke away this misery   And do my own ways so gallantly   I guess it\'s time to make a change   Sleep alone tomorrow never again   Still I\'m gone still I\'m gone   I\'m tiring from this extra long run    &nbsp;  And when&nbsp;I reach the end   I\'ll laugh and breathe now baby   Cause now&nbsp;I know that I\'m alright   And you\'ll know that your just fine '),(6198,3394208,'23,July,2004','2004-07-23 00:00:00','   The day feeling low minutes drag on   Take out your ganja take out your bong   Roll it up right... that\'s it   Smoke it all up that\'s the shit   Smoke all day, smoke all night   Smoke with me? It\'s alright   Doin this shit all day long   Trippin out, smoking out with the   Ganja Song    &nbsp;  You know how we\'ll feel when it starts to roll   Stuff it real tight then it\'s time to go   We\'ll be Jammin 24/7 or until we pass out   Always got stash especially on beach night-outs   Got enough beers to match our dope   Open the hot box and spread the air with beautiful smoke   That\'s it I\'m cool I\'m already buzzing   I don\'t care if I\'ll feel like shit in the morning   Been doin this&nbsp; so very fucking long   Trippin out smokin out with the   Ganja Song '),(6199,3394208,'22,July,2004','2004-07-22 00:00:00','   Scooby Doo\'s Scooby Snacks  Marlboro reds cigarette packs  Listening to rock, air guitar  I\'m so close and so very far  Jot down notes on cats and dogs  Cold breeze, soft winds and cool fogs  Set of pictures on the wall  Memories I remember them all  Sketches of women and poetry  Visions I see all coming from me  Lying down, standing up  This cup of coffee keeps me up  Radio on with my Bose stereos  Finished Doritos, iced tea and Oreos  CD walkman turned on, eating chocolate mousse  Down some warm milk, and just start to snooze  Time to sleep put out the torch  Donw with my time, out on the porch '),(6200,3394208,'22,July,2004','2004-07-22 00:00:00','   The river god\'s daughter Daphne  Pursued by Apollo\'s undying love, He  Has been struck with an arrow from cupid  So did Daphne, although a god may forbid it  It was because of Apollo\'s foolish heart and mind  What was in Cupid for Apollo to find?  Now Daphne runs far from love fearing to be wed  And Apollo will chase her till the day gods will have been killed  It\'s a beauty wasted since she fears love  But to Apollo it\'s a beautiful sight as Daphne flies like a dove  And as Apollo caught up and Daphne has tired  Daphne calls her father; help is what she asks; for what had transpired  She get\'s her wish but is unable to move an inch  Her feet get rooted, her skin turns to bark and her hair replaced with leaves  But Apollo still loves her even in her present state  Apollo truly in love has no sign of hate  Apollo gazes at Daphne because she\'s everything he sees  Respects and still loves her dearly even as a laurel tree   (my poetic rendition of Daphne and Apollo) '),(6201,3394208,'22,July,2004','2004-07-22 00:00:00','   Are you awake? Or are you still roaming around in your dream?  How do I know if this is reality? And what if this is fantasy?  Could I have been you at a different time?  Is there a truth to death? Or at the time of our death  We are instantly born again with a new life and identity?  How can I tell if you are real? What if these \"Thoughts\" are a message?  A message telling us that everythin\'s not real?  Or is the message a lie? Beguiling our beliefs?  What if we are awake but in truth asleep?  And what if we are asleep but in reality awake?  Thoughts, thoughts, thoughts, thoughts  Never-ending, eternal, defiling our souls  Never let it stop '),(6202,3394208,'22,July,2004','2004-07-22 00:00:00','   Have you ever marvelled at the sight of a supernova?    Newly formed shimmering from the sky up above    Look at what\'s growing within us    Wonders and dreams start overflowing    Stacks of unopened boxes waiting to be exposed    Chests unlocked waiting to be shared with the world    Don\'t throw them away plans and dreams    Pick up what was thought waste and recycle    Spill the pale of overflowing thoughts    Use the knowledge and build your ideas into reality    Our minds are not made to waste others    Create reality from your fantasies '),(6203,3394208,'22,July,2004','2004-07-22 00:00:00','   A magical plant that grows oh so sweet    A plant that grows in cold weather not heat    Long green stems; all packed in a bunch    It often helps me eat dinner, breakfast and lunch    Gives me a vision, no two visions, three or even more    Get\'s me thinking what\'s goin on at the other side of the door    It goes well with pastries, glass, wood plastic and paper    Sometimes with a strong stomach even in between wafers    I sip it, I eat it, I smoke that shit rolled tight    I only need alot to make me feel all right    A magical plant that grows oh so sweet    A plant that grows in cold weather not heat '),(6204,3394208,'22,July,2004','2004-07-22 00:00:00','   I remember driving through long winding roads'),(6205,3394208,'22,July,2004','2004-07-22 00:00:00','   Counting down the time left   Spending my time online   The world can really shine   Money makes the world go round   Money used makes jamming fun   Hangin round my room is grand   It really feels like quicksand   Really makes me unstable   Music up glasses down   Take a shot   Dial on the phone   Get on-line waste my time   Here on my own   The beat gets me down   Smoke a pack of cigarettes   You\'ll soon be starin at a smoke-filled room   Just get away, just run away   Online, dialer, sun rises till sun sets   Finally, I got myself stuck   Staring at the monitor   You need to stop staring   The dialer speeds up time   The dialer slows down time   Click, click, boom   Online livin gets me down   Online livin gets me high   Dialer, dial online now   I need to get a life   I need to start moving   Cut off the dialer '),(6206,3394208,'22,July,2004','2004-07-22 00:00:00','   \"Drive\"'),(6207,3394208,'22,July,2004','2004-07-22 00:00:00','   I was chillin over by the corner store early this afternoon   Alone on the curb with a pack of cigs and a bottle filled with sweet intoxication   I stood up, picked my bag up and headed for the mall   The rain poured harder with each and every single step but still standing tall   Carefully I maneuvered to a nearby coffee shop   Took a seat laid my bag down went to the counter to order myself a cup   Whipped out my journal and thought of something to write   Then had an idea&nbsp; maybe&nbsp;I should start something to fuel this empty mind   And so this man stepped through the shop entrance   Around 5\'9 dark hair chinky eyes and built... what\'s weird was he gave me a glance   In the back of my mind&nbsp;I was thinking he&nbsp;looked familiar   I shrugged it off and finished my cup walked for awhile till&nbsp;I was very far   Then I was forced to stop my walk and forced to listen   Cause&nbsp;I could\'ve sworn I heard some footsteps and heavy breathing   I turned around and there was the guy from the corner store   Cussing at me askin if I remembered him from way before   Before a word of mine was said he was already upfront   And threw me a left hook and kicked me on the nads... what a cunt   I was crawling on the ground and he kicked some more   I struggled to get back on my feet whilst taking the free shots&nbsp;I was giving this fucker   I was pained yes&nbsp;I was pained but&nbsp;I manage to get a hold of his neck   Tighter my right hand squeezed at this punk\'s trachea   He ceased his kicking and punching his objebtive now was to get free and breathe   I landed only 3 left hooks and let him go   Gasping for air this fucker gave me no show   He did have a fighting spirit so once again he charged at me   Spiking me to the ground... what a pathetic attempt to ground me   Once again&nbsp;I had him locked gasping for air his head trashed around in the mud   He valiantly punched at my ribs hoping I\'d dislodge him&nbsp;I thought after this melee&nbsp;I needed a bud   I told him bro I\'ll let you go and you run off I have no beef with you   He replied with ceasing his uncontrollable trashing and tapping my arm   This melee took merely a min to settle... he went off the other side   I stood up took my bag and just watched him walk away just to make sure&nbsp;I wouldn\'t get back stabbed   And so he disappeared around the corner street   Nothing more nothing less... I walked home that day bruised and aching   And jotted this short meaningless melee down '),(6208,3394208,'21,July,2004','2004-07-21 00:00:00','   There\'s gotta be something to talk about now   Gimme an idea of anything like what when and how   I\'ve run out of things to say it\'s unbelievable   Thinking of what to say isn\'t justifiable   I always had stuff to say cause my mind\'s always high   Full of many things but it seems they\'ve emptied... why?   Mental block can\'t think of anything else   I\'m done '),(6209,3394208,'19,July,2004','2004-07-19 00:00:00','   My heavens low but the song\'s the same      Seems I\'m always looking when I hear your name      You\'ve been away but don\'t leave this side      Seen all of your faces you\'ve tried to hide      So I\'m sitting here in a train station      Lord knows&nbsp;I step with no&nbsp;hesitation&nbsp;     Oh I seen through your lies     Still I\'m one still I\'m one to testify and I say        &nbsp;   I\'ll find a way to let you know      hold me close when tradewinds blow      chances are you\'re over me,      I\'ll catch that train eventually       To the questions that you\'re asking      I\'ve only expected everything      These arms of mine are burning      As I see, as I see the pages turning       I\'ll find a way to let you know  Hold me close when tradewinds blow  Chances are you\'re over me,  I\'ll catch that train eventually        &nbsp;   My heavens low but the song\'s the same  Seems I\'m always looking when I hear your name  You\'ve been away but don\'t leave this side  Seen all of your faces you\'ve tried to hide  So I\'m sitting here in a train station  Lord knows&nbsp;I step with no&nbsp;hesitation&nbsp;  Oh I seen through your lies  Still I\'m one still I\'m one to testify and I say        &nbsp;   I\'ll find a way to let you know      Hold me close when tradewinds blow      Chances are you\'re over me,      I\'ll catch that train eventually        &nbsp;   Said I\'m sorry     Said I\'m sorry     I\'m a lover who says he\'s sorry '),(6210,3394208,'19,July,2004','2004-07-19 00:00:00','   Johnny Butt was a man with a real strong   Will to survive'),(6211,3394208,'19,July,2004','2004-07-19 00:00:00','   When you grab a hold of me Tell me that&nbsp; I\'ll never be set&nbsp;free But I\'m a parasite Creep and crawl&nbsp;I step into the night Two pints of booze Tell me are you a badfish too Aint got no money to spend I hope the night would never end Lord knows I\'m weak Won\'t somebody get me off of this reef Baby you\'re a big blue whale Grab the reef when all the diamonds fell I swim but I wish I never learned Water\'s too polluted with germs I dive deep when it\'s ten feet over head Grab the reef underneath my bed Aint got no quarrels with God Aint got no time to grow old Lord knows I\'m weak  Won\'t somebody get me&nbsp;off of this reef Aint got no quarrels with God  Aint got no time to grow old  Lord knows I\'m weak  Won\'t somebody get me&nbsp;off of this reef '),(6212,3394208,'19,July,2004','2004-07-19 00:00:00','   We planned a thing it was no problem at all I didn\'t have a reason to doubt them all Just take it there and hope that it don\'t fall When they go, I\'ll go, cause you\'ll go &nbsp; Just a little way to get some satisfaction And my homies think we need no calculation I settle things my way and head on home When they go, I\'ll go, cause you\'ll go &nbsp; If you don\'t hear me&nbsp;I say we got to get sold Cause all&nbsp;I got in my mind is making me some dough You know that we need this all cause When they go, I\'ll go, cause you\'ll go &nbsp; It\'s all about what, when and where\'s your goal I don\'t wanna get stuck mining fer coal I want to be the guy who\'s&nbsp;running it all When they go, I\'ll go, cause you\'ll go &nbsp; My mind is set on gettin and spendin g\'s Cause remember money makes the world go round My boys you gotta understand money cause They\'ll go, and you\'ll go, cause I\'ll go '),(6213,3394208,'17,July,2004','2004-07-17 00:00:00','   When my girly comes around I just want to hear one sound  I just want her to stay by my side Cause I grow weak with her camel toe slide Why does she do it&nbsp;I don\'t know Just thinking about it gets me high when I\'m low Where\'d you ever get such a thing? The thought of it is just so mind boggling &nbsp; &nbsp; &nbsp; '),(6214,3394208,'15,July,2004','2004-07-15 00:00:00','   Sorry eyes Cut to the&nbsp;bone Make it hard To leave you alone Leave you here Wearing your wounds Waving your guns At&nbsp;somebody new &nbsp; Baby&nbsp;you\'re a lost Baby&nbsp;you\'re a lost Baby&nbsp;you\'re a lost cause &nbsp; There\'s too many people You used to know They see you coming  They see you go They know your secrets And you know theres This town is crazy Nobody cares &nbsp; Baby you\'re a lost  Baby&nbsp;you\'re a lost  Baby&nbsp;you\'re a lost cause &nbsp; I\'m tired of fighting I\'m tired of fighting Fighting for a lost cause &nbsp; There\'s a place where you&nbsp;are goin You aint never been before There\'s no one laughing at your back now  No one standing at your door That\'s what you thought love was for &nbsp; Baby&nbsp;you\'re a lost  Baby&nbsp;you\'re a lost  Baby you\'re a lost cause &nbsp; I\'m tired of fighting  I\'m tired of fighting  Fighting for a lost cause '),(6215,3394208,'15,July,2004','2004-07-15 00:00:00','   close to nothing you never have felt cause no one ever did, so no one ever will '),(6216,3394208,'13,July,2004','2004-07-13 00:00:00','   It was a night to remember And a night to forget Cause we drove away, so far away We were lost and silent Just you and I sitting on the curb I sat and you stared We ran, yes indeed we ran feelin chills from this air These lights all pass us by As I stepped on the gas We laughed and shouted whilst in the car Driving so fast And for an hour we haven\'t said a word Radio did all the talking Rastafarian music played so loudly Ganja kept on burning  We stopped to rest by countless motels And if we ran out of gas And for days we haven\'t said a word But you always knew when I wanted ass We drove for weeks then turned to months We fled We drove till we almost reached a year But this car\'s engine was close to dead '),(6217,3394208,'13,July,2004','2004-07-13 00:00:00','   I don\'t know what I want But I know how to get it You\'re sayin there\'s somethin wrong with me But there isn\'t that i can see What am I supposed to do But I\'ll do it when it comes around My heaven\'s grow ever so ever so This scenery ever never blow My mind is broken down And my body irepairable Funding my state of mentality  Now I know what I want Now I know how to get it Go and say what you want Never will I go forget it And I know what to do So I\'ll be the one coming around My heaven\'s now ever so ever so This scenery I know will never grow old My mind has grown and not being beat My body was mearly in a healing state Funding my state of humanity '),(6218,3394208,'12,July,2004','2004-07-12 00:00:00','   Hey there goes that guy  with the guitar down the road Strumming away his pain  and probably yours with his odes I can\'t wait till he hits the stage  and watch him play His rythm and all his solos and his riffs  make my day No one wants to leave or turn away  cause if they do he\'s gone He, only stays to play  and to be heard a shame he\'s done  Hey baby look at the guitar man he plays so good I wish it wouldn\'t end Hey baby look at the guitar man these six strings get struck, plucked and bend '),(6219,3394208,'07,July,2004','2004-07-07 00:00:00','   I\'ve got new sunglasses and I\'ve got the back stage passes boy on girls with fine asses yeah oh yeah sunshine, girlies, and money what it takes to get me on a roll some call me Mr. Soul  Went to the party the other night I wanna feel what my brain had felt like I pullled the bottle and I start drinking wine I thought that preety soon I would be feelin fine But no I couldnt get high oh no I said I couldnt get high oh no I said I dont know why Threw down the bottle Whipped out the pipe Stuffed it full of goodness Then I gave that shit a light I heffed and I puffed I smoked and i choked I smoked so long until my heart was nearly broke Because I couldnt get high oh no I said I couldnt get high oh no I said I try and try  Threw down the pipe pissed as I could be Gobbled down a mound of some LSD And then I waited and I waited for my body to sing I waited I waited but not a fucking thing Because I couldn\'t get high oh no I said I couldnt get high oh no I said I dont know why '),(6220,3394208,'06,July,2004','2004-07-06 00:00:00','   And some man lives his life for profits alone  That very same man he lives his life all alone  And the road to life yes it goes up and down  Doesnâ€™t really matter as long as the music goes on  Hey mister collie man why donâ€™t you come round no more  Hey mister collie collie man donâ€™t you hear me troubled call  Hey mister collie man look i didnâ€™t come to disturb  You best not be come â€˜round here unless you\'ve got me sensi herb  I never needed any reason for me to say  Through all them troubled times yes me love you any way  And the roughest path you know the rocky rocky road  You know that life and love is a heavy heavy load  Hey mister collie man why donâ€™t you come â€˜round no more  Hey mister collie collie man dont you hear me troubled call  Hey mister collie man look i didnâ€™t come to disturb  You best not be come round here unless you\'ve got me sensi herb  You best not be come round here  Unless you\'ve got me sensi herb '),(6221,3394208,'06,July,2004','2004-07-06 00:00:00','   I been staring at your name Written on this paper it\'s a shame Why\'d you have to go away Why couldn\'t you have stayed I\'ll catch that train soon eventually Oh baby baby I hope you\'ll remember me  Chances are, you already have... Tis a shame, you already have...  And so I stopped by Just once more I would like to flash a smile Just laugh along at the movie And have you by my side in front of the telly Tis a shame it\'ll end eventually Oh baby baby say you\'ll stay  Chances are, you already have... Tis a shame, you already have...  I\'ll find a way to let you know Holding you close when tradewinds blow I should have never hesitate Cause baby baby you\'ve gone away I\'ll catch that train soon eventually Oh baby baby I hope you\'ll remember me  Chances are, you already have... Tis a shame, you already have...  Said I\'m sorry... '),(6222,3394208,'05,July,2004','2004-07-05 00:00:00','   See you walking down the street,  I just want to talk to you.  But you donâ€™t have too much to say,  So Iâ€™ll just be on my way.   cause youâ€™re just too shy, (just too shy, shy girl)...  Just too shy, (just too shy, shy girl),  Aww babe, youâ€™re just too shy (just too shy, shy girl).   Sittin there all by youreself,  talking to nobody (talking to nobody )  your head is bowed,  your eyes look to the ground,  I want to you know you baby.   cause youâ€™re just too shy, (just too shy, shy girl)...  Just too shy, (just too shy, shy girl),  Aww babe, youâ€™re just too shy (just too shy, shy girl).   See you laughin with your friends,  But they do all the talking.  Wonâ€™t you come and walk with me awhile,  I want to you know you better   cause youâ€™re just too shy, (just too shy, shy girl)...  Just too shy, (just too shy, shy girl),  Aww babe, youâ€™re just too shy (just too shy, shy girl). '),(6223,3394208,'04,July,2004','2004-07-04 00:00:00','   urlLink    My Blood-red Desktop&nbsp; urlLink '),(6224,3394208,'04,July,2004','2004-07-04 00:00:00','   urlLink    HAZY DAYS BABY!!!&nbsp; urlLink '),(6225,3394208,'02,July,2004','2004-07-02 00:00:00','   urlLink    My bro Gabs&nbsp; urlLink '),(6226,3394208,'02,July,2004','2004-07-02 00:00:00','   urlLink    My bros Potter and Jaime&nbsp; urlLink '),(6227,3394208,'02,July,2004','2004-07-02 00:00:00','   urlLink    My cousin Pocholo and The Golden Mean&nbsp; urlLink '),(6228,3394208,'02,July,2004','2004-07-02 00:00:00','   urlLink    My Homie Doni&nbsp; urlLink '),(6229,3394208,'02,July,2004','2004-07-02 00:00:00','   A man who always goes with the flow Will never know what\'ll happen if he doesn\'t '),(6230,3394208,'02,July,2004','2004-07-02 00:00:00','   Tell me what you think?  Am I goin down the drain,  In this dirty dirty love sink overflowin\' with the foam of our loving  Cleanin\' these dirty dishes?  Girl I know you aint mine  But baby I\'m yours baby baby I\'m yours...  I\'m tellin all my friends would ya wanna swap me some  Cause I aint in control here,  maybe baby all I need is to be the man  Even for a minute that\'s cool with me...   Girl when I first got ya I had it all, lest I ever knew That I was bein hooked I was hookin you... But you were hookin me to I can\'t even step out to see my boys anymore Cause I got my baby knowin everytime I hit the door Ah shit I broke it off I aint that dumb And tell all my buddies would ya wanna swap me some? '),(6231,3394208,'02,July,2004','2004-07-02 00:00:00','   I believe that a life;  Long lived and empty, is worthless  Short lived and full, is priceless  I also believe that if you\'re living a life  just solely to live it, is grand  But to live and be remembered by all, is sublime '),(6232,3394208,'01,July,2004','2004-07-01 00:00:00','   I aint a playa baby,  Cause nowadays I only play my guitar I\'m just a bit careless now,  But I\'m gonna get ya just to see I\'ll get ya just to see,  What makes you tick and tock And maybe just maybe,  No I know I\'ll get you when we talk Bout nothing that you believe that it must be something Something like love,  But hey,  I\'m gonna get ya just to see...  I aint gonna kick the bucket soon, Not until I get the rundown from you I\'m gonna get ya just to see I\'m gonna get ya just to see '),(6233,3394208,'01,July,2004','2004-07-01 00:00:00','   Do you remember the days? spending it outside by the park swinging on the swings that bring us up high And what about the turles fighting crime in the tube how easily amused we were when they said cowabunga And remember the time when you and I fought about who was who back then and who\'d get the last slice of pizza Watching cartoons in my room and laughing so loud and tripping on the show Mr. Neighborhood  I miss those days back at my old house in that little street  both of us were so close back then  What about the time when we helped each other during tests and examinations, passing and sharing notes And remember the day when you cried cause you were afraid of the dark so I\'d come and bring flashlights along And we\'d play pretend like I was a king and you were a princess and all we had done all day was play snes And hope you remember the day when we fought for us to stay in my little house and how we cried when I moved away  I miss those days back at my old house in that little street  both of us were so close back then  And after many many years I came back to know that you have left your own little house in the street by the park Where we used to meet I still remember those good old days Especially the day we both said \"don\'t forget me\" I wonder how you are and if ever will we ever meet again or we both are still looking at our favorite star And I hope ur still cool and remember never ever turn a fool to those who\'s only desire is to be cruel  I miss those days back at my old house in that little street  You and me were so close back then '),(6234,3394208,'01,July,2004','2004-07-01 00:00:00','   I saw you at the bar I was laughing with some friends I saw you turn your head I was checking you instead I asked you if you would like to dance I really wasnâ€™t trying to get into your pants Not to say that youâ€™re not a beautiful young lady But I drive a chevy nova And I thought you liked mercedes Soon enough though We were allured Bumping and grinding out on the dance floor With the system so loud The vibrations so pulsing You whispered in my ear But I couldnâ€™t hear nothing Steered me over in the corner And kissed me on the neck Thatâ€™s when I decided oh what the heck I put my hands on your hips And I kissed you right back You kissed me like a shark Thatâ€™s sniffing on a snack And then you said \"my boyfriend never kissed me like that\" I said \"boyfriend? you didnâ€™t say he was around\" On this friday friday night When I came and rocked your town  On this friday night When the feelingâ€™s alright Rockin rock steady grooves In the middle of the night Everything is heated up And nobodyâ€™s standing still With my arm around my girl And a hundred dollar bill  The song was over So we headed for a drink And whatta you think She gave another guy a wink Who came over for a kiss And asked the girl \"whoâ€™s this?\" She said \"thatâ€™s G. Love you know G. from special sauce?\" Then he said \"Iâ€™m getting louie kid you better get lost\" I asked the girl \"whoâ€™s louie?\"  And she said \"thatâ€™s my man\" Women women I just canâ€™t understand Then I says \"I donâ€™t want to hear about girl problems And if you got a problem babydoll I can solve them\" But I feel youâ€™re gonna get me in trouble So câ€™mon letâ€™s leave this bar And go out clubbing on the double Girlâ€™s putting me in a trance All because she wants to dance And get down with the man Who brought the sound On this friday friday night When I came and rocked your town  On this friday night When the feelingâ€™s alright Rockin rock steady grooves In the middle of the night Everything is heated up And nobodyâ€™s standing still With my arm around my girl And this hundred dollar bill  I can just tell A girl like youâ€™s a deadend Pressing up against me Talking about \"I got a boyfriend\" Now here he comes looking for you On the dance floor While youâ€™re pushing on my knee Talkinâ€™ â€™bout \"I love you more\" Iâ€™m new in town Bout Iâ€™m no stranger to this Iâ€™m not trying to exist With a hit and run fix I get my loving on the run And I like having fun But I donâ€™t sneak around And I donâ€™t gotta gun Now give me three steps Like the boy lynyrd skynrd Thereâ€™s more drama in this town Than a soap so I figured Maybe Iâ€™ll get your number And maybe I will call And maybe weâ€™ll slip Into the backroom down the hall Or maybe Iâ€™ll get jacked up By your angry boyfriend louie â€™cause his eyes are getting redder While his angerâ€™s busy stewing on the fire The little trick in you brought out On this friday friday night When I came and rocked your town  On this friday night Everything is alright Rock steady grooves In the middle of the night Everything is heated up And nobodyâ€™s standing still With my arm around my girl And this hundred dollar bill '),(6235,3394208,'01,July,2004','2004-07-01 00:00:00','   Baby weâ€™ve been through a lot And times sure have changed Everybodyâ€™s leaving town Looking for better things But I found what I want And I got no problem working I wish to god That my heart would stop hurting Shops closing down Factories outdated People so frustrated My whole life is changing Just trying to find some solid ground A little house with a picket fence Howâ€™s that sound  Love is all I need Love is all I need  Iâ€™ve seen tears that are Falling like leaves from the trees Houses boarded up and they Threw away the keys But this town is where weâ€™re born And this town is where weâ€™re staying Iâ€™ll never ever turn my back On you ever again Come on letâ€™s be together I got some money put aside And if I make it big Would you take a little ride Would you promise to always Stay by my side Woudl you give me true love And have nothing to hide Would you give my your love Would you give me your love Would you give me your love If you want to give me your love babe Itâ€™s alright Itâ€™s alright Itâ€™s alright We can make a change in our life babe Tonight Itâ€™s alright Itâ€™s alright '),(6236,3394208,'04,August,2004','2004-08-04 00:00:00','   Excuse me Mr. do you have the time or are you so important   that it stands still for you  excuse me Mr. lend me your ear or are you not only blind but do you not hear  excuse me Mr. isn\'t that your oil in the sea and the pollution in the air Mr. whose could that be  excuse me Mr. but I\'m a mister too and you\'re givin\' Mr. a bad name Mr. like you  so I\'m taking the Mr. from out in front of your name cause it\'s a Mr. like you that puts the rest of us to shame it\'s a Mr. like you that puts the rest of us to shame  and I\'ve seen enough to know that I\'ve seen too much  excuse me Mr. can\'t you see the children dying you say that you can\'t help them Mr. you\'re not even trying  excuse me Mr. take a look around Mr just look up and you\'ll see it\'s comin\'down  excuse me Mr. but I\'m a mister too and you\'re givin\'Mr a bad name Mr. like you  so I\'m taking the Mr. from out in front of your name cause it\'s a Mr. like you that puts the rest of us to shame it\'s a Mr. like you that puts the rest of us to shame  and I\'ve seen enough to know that I\'ve seen too much  so Mr. when you\'re rattling on heaven\'s gate let me tell you Mr. by then it\'s too late  cause Mr. when you get there they don\'t ask how much you saved all they\'l want to know, Mr.is what you gave  excuse me Mr. but I\'m a mister too and you\'re givin\'Mr a bad name Mr. like you  so I\'m taking the Mr.   from out in front of your name cause it\'s a Mr. like you that puts the rest of us to shame it\'s a Mr. like you   that puts the rest of us to shame '),(6237,3394208,'01,August,2004','2004-08-01 00:00:00','   As I inhale the smoke of boredom  my body shivers   for with boredom comes silence, such deafening silence   And as my hand traces her still and peaceful face   I wonder what she dreams of, or if I am found in her state of unconciousness   The ceiling stares back at me from up above   it still fascinates me looking up and thinking of things looking down upon me   She lay still and silent on the bed wrapped up as she is   still dreaming on and I\'m thinking... thinking of the posibilities   Maybe she\'s forgotten? Maybe she\'s thinking of someone else   who am I to try and control one who has no control over her feelings? Her choices? '),(6238,4216577,'12,August,2004','2004-08-12 00:00:00','    What a weekend!  If any of you have never been to Manitoulin Island, the long weekend in August every year is a great time to visit!  This year, we decided to be a part of the festivities and become vendors.  What an adventure it was! This is an outdoor event by the way.  The first day we set up outside of the Rec Centre in Little Current on a small grass walkway between the drive and the parking lot.  (it was about 9am for me)It turned out to be a great location becuase we REALLY needed the tree that we set up under.  The day was overcast, threatening to break out in a heck of a storm.  Across from us was a clothing vendor and his family who were a great help when my tent refused to go up on its own.  (Its really wonderful to meet such great folks!)  I had a great lady set up beside me (Rayvin from Eclectic Enchantments here in Sudbury) and the two of us rode the bad weather together.  Luckily she brought tarps which were much needed when the storm finally broke through the gray sky and we got dumped on.  But it was a blast! I was selling while holding onto the frame of the tent while the wind DROVE through everything, pieces flying across the grass... some quite far! All you could do was laugh and enjoy the great company of the people who came to our tents supporting our business.  (Thanks!)  The weather finally drove us to close up shop as it was tearing down larger tents beside us and causing a heck of a mess.  But it was a great day!  The next morning I set up quite early on the main street downtown in Little Current.  I had a fabulous view right onto the walkway of the docks looking at the North Channel.  At 8am the temperature had already hit 28 degrees C.  HOT! And no wind!  I brought extra stones with me for the base of my tent just in case!  The main street was blocked off, so we had to unpack our vehicles and move them.  The heat continued to build and so did the energy of the day!  Right on the street, we had the perfect view of the parade going by and later in the evening the fireworks that went off on the shore across from the walkway.  I\'d like to thank EVERYONE who attended the festival and came to my booth for making this show a fabulous success for us!   The streets were jammed with people and I was lucky to be beside a great food vendor that made the most fabulous home made breads and fried bologne! YUM!   By 7pm I was feeling pretty tired.... my family was meeting in Little Current for a dinner and they were sweet to bring me a plate to munch on - Thanks!! A big thankyou to my cousin Peggy, who hosted the dinner (I did not get to thank her properly as I was extremly busy when she showed up)  The night continued on and finally around 11pm we had to close up shop as the fireworks were about to start.    My mom and daughter helped me to unfold our travelling show quickly so that we could get our cars over the swing bridge before the crowds started to head out.  For anyone who has not been to Manitoulin Island - it IS an island and the only way on is via a one lane swing bridge OR the ferry on the south side.   We folded up the shop and headed out over the bridge and STILL had a lot of time to sit on the back of the truck watching the beautiful lights in the sky.  I\'ll be there again next year!! Raven   '),(6239,4216577,'12,August,2004','2004-08-12 00:00:00','    Hi!  You know, sometimes being a vendor in Canada is not easy. I\'ve been looking for a good e-commerce solution so that I can sell in both currencies (US and CA)... Right now I am using a really great service provider that I\'m really satisfied but I can only sell in US dollars..  If you find something you really like on our site and you are from Canada, for now you can give me a call (or email) and we\'ll work out the price and I can ship your item to you. (All the contact info is on our webpage)  I\'m hoping that by mid September this year I\'ll be able to offer Canadian prices but I wont make any promices until I work all the bugs out.  Anyways, I\'m off to make some more creations.. right now the stock is low on the site since we had a fabulous time at Haweater weekend. Maybe I\'ll post about that later on.  Raven   '),(6240,4216577,'12,August,2004','2004-08-12 00:00:00','    Greetings and welcome to our Blog      What\'s a Blog? I know I thought that the first time I saw one of these things... its a place where you can come and visit and read all about Renaissance Market and things that go on here.      It\'s a personal inside look into our business and maybe some of the \"little\" things that go on in our lives from day to day.  Call it a peek inside of the shell you see here on the internet, a place where you can become a part of our day.      Well I hope you enjoy our view behind the magic curtain.      Be well! Raven   '),(6241,2568090,'21,May,2004','2004-05-21 00:00:00','   When I think of nursery school, I think of the blue and red patterned rug we all used to sit in a circle on.  The rug had some sort of a circular chart imprinted on it, one that told the seasons and the months and the days of the week.  There were trees and cars on the rug, and smiling children of all races holding hands within the circle.  To this day, I still visualize the seasons in that large, looming circular shape.  Winter is north, and proceeds clockwise, with fall ending up in the east.  I remember always sitting down squarely in Fall during story hour or nap time. I also remember chasing a boy named Abraham around the playground during one particular playtime.  There were a bunch of us girls, and we unanimously decided to tackle this unassuming, soft-spoken blond boy for the purpose of kissing him.  Abraham was shocked and probably scared silly when he learned of our purpose, and proved to be a tough catch.   My best friends during those two years were Krista Morgan, Dana Atschuler, Sara Bromley, and Tamara Sue Dutton.  Kristaâ€™s motherâ€™s name was Dale and went insane after the divorce.  I remember her only in her alternate persona, Strawberry Shortcake, though ostensibly she only wore that costume for childrenâ€™s birthday parties.  Dana was a bored-looking girl with a shiny brown ponytail who had an older brother I might have been in love with.  She was a great swimmer and I remember us scuttling out of our bathing suits in the concrete locker room of the Newfield Swim Club, eager to dash out to the playground swings after a long day in the water.  Sara Bromley was small and had a heavy dark fringe of hair and wore fashionably baggy sweaters.  We cracked up one night to the point of hysterics at her house while telling the â€œI C-U-Pâ€ joke to our bewildered parents, who had been enjoying a glass of wine in the living room.  Tamara Sue Dutton was my poor friend who lived in a trailer (ok, a small house) with her mother.  She was blond and very pretty and I remember a particular pink ruffled jumper she used to wear, but thatâ€™s it.   Then there were the twins, whom I was fascinated with but didnâ€™t know all that well.  Being twins, I suppose they were pretty self-contained.  One of their names was Lisaâ€”I donâ€™t remember the other.  They wore identical brown shaggy pageboys, navy blue Fair Isle sweaters, and a slow, sleepy expression on their faces.  I remember that one witnessed the event of my tripping and skinning my knee on the big rock face we used to climb over and play on.  Her eyes widened a bit, but she said nothing and didnâ€™t offer to help.  Though the event was memorable in itself, that bloody knee is permanently visible in one of my nursery school portraits, a lone blemish on my chubby white leg, the red patchwork of scratches just north of my Strawberry Shortcake lace-up sneakers (my first since graduating from Velcro). Unlike the skinned knee, the Birthday Party would have receded into oblivion if not for the picture of my twisted, tear-stained face of agony, surrounded by the surviving balloonsâ€”those that had not popped, that is.  (Iâ€™m not sure, come to think of it, why someone would have snapped a photo of that particular horrible moment.  Was it my mother?)  The sound of the exploding balloons had alarmed me so that I panicked and burst into a flood tears.  Unable to finish my cake and â€œget on with it,â€ I had to be consoled by a team of teachers for the rest of the afternoon.  My poor mother had taken the afternoon off from work just to bring those goddamn helium balloons and that cake over to my school, and all for naughtâ€”the â€œhappy birthdayâ€ ended in a fit of hysterical tears instead.  Apparently, the other children were totally unperturbedâ€”delighted evenâ€”by the sound of the balloons popping, and were confused by my horror.  Is that why I came to loathe being the center of attention?  Already in the spotlight by the occasion of my birthday and the impromptu party, my hysteria over the balloons was all the more conspicuous, and subsequently puzzled over.  What a weirdo, they must have thought. There was a strange annual fair at my nursery schoolâ€”an Episcopalian churchâ€”that has muddled in my mind with the memory of the Land Of Make-Believe on â€œMr. Rogersâ€™ Neighborhood.â€  Clad in my summer playclothes, as I wandered through past the ponies and barrels where people would bob for apples, I felt very much as though that woman with the high pitched voice, clutching a puppet, would come over and engage me in a game of Make-Believe.  I remember that there was hay there.  Lots of it, and little prizes like small flip books printed on thin paper, and charms for the ubiquitous charm bracelets that every little girl wore at the time.  That was at the end of nursery school.    Kindergarten was taught by Ms. Ehren, a wild-haired, denim-jumpered Jewish woman who loved children and behaved in an eccentric, child-like way herself.  Ms. Ehren comforted me and told me I was â€œA whole new Emily!â€ when I got a horrible hacksawed haircut that spring.  Ms. Ehren was sympathetic in a way I did not yet understand, as demonstrated by her gentle disciplinary methods when dealing with the unruly black children in our class.  Ms. Ehren was prone to making well-intentioned but utterly misleading declarations like, â€œTomorrow weâ€™re traveling to the Mezzezoic Era, so pack your bags!â€  Telling my parents nothing of what I imagined to be an illicit and unconventional trip, I decided that the Mezzezoic Era was something akin to Florida, where our family visited every winter.  Trembling with excitement the next morning, I shoved a few bathing suits in my backpack before boarding the school bus, frantically trying to imagine how on earth Ms. Ehren was going to get all of us on the plane, to the Mezzezoic Era, and home again before dinnertime.  Would we be away for more than one day?  Would my parents be worried?  Still, I trusted Ms. Ehren and ultimately decided that secrecy of her plan was bestâ€”she would take care of the details. Kindergarten was also memorable for my obsession with the School Store, which conveniently set up shop for one lunch period a week in the hallway right around the corner from Ms. Ehrenâ€™s classroom.  The School Store sold things like colorful pencils with fluorescent grip divots, erasable pens (howâ€™d they do it?), and mechanical pencils in girly shades of light blue and pale pink.  As the School Store schedule was woefully unpredictable, however, I never had enough money for these delicious trinkets.  Look, but donâ€™t touch, said the derisive looks given to me by the hopelessly wise and mature sixth graders who minded the Store.                    to be continued...   '),(6242,3823807,'05,July,2004','2004-07-05 00:00:00','   1. Are you single or not? single :) 2. Do you currently have a crush on someone? no 3. Does your crush know? um... 4. Nicknames: Manda, Mandy, Dander, A-man-duh!, Squishy, Celineiac, no name, that girl obsessed with Celine Dion 5. Eyes: Blue 6. Height: 5\'7ish 7. Siblings: none 8. What are you wearing right now? mahogany tank top, jean capris, my Celine necklace and big hoops 9. Do you have your a/c on? I think so 10. Birthday: sept. 6 1989 11. Where do you live: in a house 12. Sex: female 13. Righty or Lefty: Righty  ON GUYS FOR GIRLS TO FILL OUT (for girls only)  14. Boxers or briefs: boxers 15. Tall or Short: Tall 16. Western, Preppy, or All-American: umm...idk 17. Good or Bad guy: good i guess, but not goody goody 2-shoes 18. Hat or no hat: dont care 19. Ears pierced or not: one ear pierced, that\'s sexy 20. Tan or Pale: tan 21. Dimples or not: dimples are sooo cute! 22. Athlete or Academic: academic, but not too nerdy 23. Accent or not: accent!!!   ON GIRLS FOR GUYS TO FILL OUT (for guys only)  27. Regular underwear or thong: 28. Painted nails or not: 29. Bra or sports bra: 30. Cute n\' mysterious or wild n\' sexy: 31. Dressy or casual: 32. Dark hair or Blonde hair: 33. Long or short hair: 34. Dark or light eyes: 35. Long or short nails: 36. Hat or no hat: 37. Good or bad girl: 38. Hair up or down: 39. Jewelryor none: 40. Tall or short: 41.Curly or straight hair: 42. Tan or fair: 43. Freckles or none: 44. Indoor chick or party chick: 45. Accent or not: shy or outgoing:  EVERYONE  51. Root Beer or Dr. Pepper: hard call... I\'d say Dr. Pepper tho, unless it\'s Sprecher... 52. Mud or Jell-O wrestling: jello wrestling, only bc i like jello better than mud 53. Sunshine or rain: Sunshine 54. Vanilla or Chocolate: depends, but most of the time, chocolate 55. Skiing or snowboarding: skiing 56. Day or night: night! I HATE mornings. 57. Summer or winter: another hard call... I\'ll go with summer cuz that\'s my birthday 58. Pie or cake? gah! too-many-hard-decisions! lol I\'d say cake tho. yummy. 60. Silver or gold: jewelry silver, everything else gold 61. Diamond or Pearl: diamond 62. Sun set or sun rise: sunsets, especially in Hawaii :D besides I\'m not up early enough to see sunrises 63. Have you ever broken/fractered a bone? yup, my wrist in 1st grade 64. What\'s your favorite color(s)? purple mostly, but I like any vibrant or pastel colors pretty much. 65. Do you hate any one: not at the moment (except Mariah Carey... does she count?) 66. What\'s your favorite band? tossup btwn Good Charlotte, Evanescense, and Steriogram 67. Do you have a HUGE crush on someone? not right now 68. Have you ever been in love? ya 69. Ever been hurt by a member of the opposite sex? yes! *cough cough Alex cough cough* 70. Who\'s the loudest friend you have? SASHA!  71. Who\'s the quietest friend: probly Caroline or Kim... maybe Alexandra? ooh! Emily from camp! eh...who knows. 72. Who\'s the craziest friend: ALL my friends are crazy! jeez. 73. Who makes you laugh the most: all of my friends... but probly Chelsea or Alex. 75. Who do you tell EVERYTHING to: Val, Chelsea, and Alex mostly 76. What is your idea of the perfect date? something that involves stargazing and preferrably Celine... but if not, a barefoot walk on the beach in Hawaii or Door County sounds fine to me 77. Name as it appears on your birth certificate: Amanda Lauren Armstrong Rehrauer I think... maybe w/o the Armstrong? who knows. 87. Date that you regularly blow out your candles: what?? my birthday I guess... what kind of a question is that? 80. Pets: Bogy! (golden retriever) 81. Tattoos: no 82. Piercings: ears 83. Birthplace: good ol\' Milwaukee 84. Current Residence: wisconsin 85. Favorite food: too many yummy foods 2 pick from 86. Favorite drink: smoothies, soda, and artificial grape juice. the koolaid kind. 87. Been to Africa? no 88. Been toilet papered by someone? ya, but idk who tho 89. Loved somebody so much it made you cry? ya 90. Been in a car accident? thank god no 91. Favorite WORD(s): hexametaphosphate, methylchloroisothiazolinone (lol I like reading ingredients to stuff when I\'m bored! like syrup and shampoo and stuff.) onomatapia, choisir (it didn\'t say in english!), yayishness 92. Last book you read? idk I don\'t read that much... probly Harley (for like the 15th time lmao) 93. When was your last hospital visit: idk I don\'t go to hospitals much 94. What color is your bedroom carpet? brown! YUK! 95. Who is the last person you got email from before this? sasha   '),(6243,3823807,'04,July,2004','2004-07-04 00:00:00','   OnE HuNdReD AnD OnE QuEsTiOnS >Please fill it out too and send it back!!! Okay, you all know the >routine...copy (not forward) this email and paste it onto a new >email that you will send. Change all the answers so that they apply >to you, then send this to a whole bunch of people you know INCLUDING >the person who sent it to you. The theory is that you will learn a >lot of little facts about your friends. > > (except not really. but still here ya go.) 1) Starting time: 9:38PM 2) Full name: Amanda Lauren Armstrong Rehrauer 3) Nicknames: Manda; Mandy; Dander; A man, duh!; Squishy, Celineiac, that girl obsessed with Celine Dion, no name  4) School: Homestead High School  5) Email: ducky61803@yahoo.com 6) Eyes: blue 7) Height: 5\'7-ish 8) Siblings: does my dog count? 9) Ever been kissed: yes :) 10) Ever cheated or been cheated on?: YES!!! *cough cough alex and val cough cough* 11) What do you want in a relationship?: somebody who cares about me and likes me for me (all together now: awwwwwwww!!) 12) Ever missed school because it was raining?: lol no! but good idea.  13) Kept a secret from everyone?: yup 14) Had an imaginary friend?: I LOVE YOU WENDIE!!! 15) Wanted to hook up with a friend?: *no comment* 16) Cried during a Flick?: yeah- A Walk to Remember 18) Ever thought of animated characters as hot?: um, no. 19) Ever thought your best friend was hot?: *shivers* definetly not. 20) If someone you didnt know were to just ask you out, out of the blue, what would you say?: i depends on if he was hot or not! lol no jk id probly get to know them better. 21) What the first thing about a guy you notice?: the way he acts and the way carries himself  -----------------FAVORITES---------------- >  22) Shampoo: bio from bath and body works and beach blonde 23) Color: i like pretty much anything either bright and vibrant or pastel ]24) Summer/Winter?: summer  25) Online?: um.... if youre asking if i\'m online, then obviously, but otherwise..... ????? 26) Lace, silk or satin?: silk i guess 27) Like anyone: wait r we still doing favorites? what r u talking about?! that doesnt make sense! dont confuse me...its not nice! (in the words of Sasha) 28) Who have you known the longest out of your friends: either Kira or Chelsea 29) Who\'s the loudest: SASHA! 30) Who\'s the shyest: Alex probly 31) Who do you go to for advice?: Val 32) Who do you get along with?: everybody, I\'m pretty neutral ---------IN THE LAST 2 WEEKS HAVE YOU------------  33) Cried: i don\'t think so 34) Been mean: probly but i cant think of anything  35) Acted blonde?: um, duh. 36) Been sarcastic: ya like everyday 38) Talked to someone you have liked?: ya 39) Told someone you loved them?: umm...my mom...does that count? 40) Hugged someone: ya  41) Been depressed?: a little, does not going to the John Mayer concert count? 42) Wished upon a star: yea 43) Laughed until you\'ve cried: no but i\'ve laughed hard 44) Played Truth or Dare: not in the last 2 weeks 45) Watched a sunrise/sunset: yeah, it was really pretty too 46) Went to the beach at night?: no but it sounds like fun  47) Broke up with someone?: no 48) Just sat and thought, looking for answers?: yea, it was quite nice 49) Are you lonely?: yeah  50) Are you happy?: most of the time  51) Are you talking to someone online?: sorta... they\'re talking to me but i\'m away. haha. >-----------DO YOU BELIEVE IN.......------------ >  52) God/Devil: ya 54) The Closet Monster: AHHHHHHH NOT THE CLOSET MONSTER!!!!!!!!  55) Santa Claus?: :) 56) Heaven/hell: yep  57) Superstitions?: yeah, just not telling ppl important stuff otherwise you\'ll jinx it.  ---------ODD QUESTIONS---------  58) Half empty or half full: half full  59) Who named you?: parents  60) Favorite quote: i dunno...i like lots of quotes 61) When was the last time you showered?: earlier today 62) What was the last thing that you said when you were online? sash!! omg i just met a new celine fan!! 64) What is your computer desk made of?: wood and glass (there\'s a glass top on it) 65) What was the last thing that you did: atalked to my aunt and uncle on the phone, they\'re in las vegas right now  66) where do u want to go on your honey moon? someplace hot, like australia or something 67) Who do you want to spend the rest of your life with: someone who cares about me 68) Night or Day: i\'m a night owl but i dont mind the daytime 69) What type of weather?: warm and sunny  70) What did you do last night?: wouldnt you like to know ;-) jk jk jk i watched tv til 11, listened to music and slept on the floor with my dog (but not WITH my dog... lol) cuz he was scared 72) How do YOU eat an Oreo?: twist it and then lick the cream then eat the cookie. or if I have 2 oreos ill twist them and then stick the opposite sides together for a double stuff. 73) All time favorite TV show: degrassi: the next generation or in case you havent heard of that, vh1\'s best week ever or spongebob 74) Dream vehicle? vw bug!! 75) What do you want to be when you grow up: um.......... 76) Favorite music?: anything by Celine Dion or whatever\'s on the radio that isnt rap 77) Favorite band?: if you mean BAND, then Maroon 5... but musician, Celine Dion :-D  78) Favorite food?: greasy fatty goodness!!! and peanut butter toast.  79) Favorite Songs?: anything by Celine Dion and whatever\'s on the radio that isn\'t rap 80) Favorite days of the week?: days where I can sit on my fat ass and do absolutely NOTHING!!! 81) Favorite Animal: GOLDEN-RETRIEVER-MANATEE-PANDA-THINGY!!!! 82) Where would you rather spend your day? somewhere with my friends, like Hawaii :) 83) If you could change your name, what would it be?: idk(val\'s name would be evelyn)(however it is spelled) 84) Do you sleep with a stuffed animal?: I used to... then I found out that I threw them out of the bed.... now I dont/  85) The most stupid thing u ever done: oi too many to name  86. First Son\'s name?: Micheal  87) Husband/ Wife: ?? 88) First daughter?: Chloe Nichelle 89) Silver or Gold?: silver 90) You like scary or funny moviesbetter? both... but i\'d have to say scary 91) On the phone or in person?: depends... if i like them, then in person, but if I don\'t, then on the phone, lol 92) Lust or Love: Love  93) If you could change something about yourself, what would it be?: ai more like what wouldnt I change?! but i\'d hafta say i\'d like to be a little more confident in myself and more in shape. and more attractive to guys cuz i think i have like guy repellent or something. lol. but not really cuz it sucks. 94) Do you have pets: Bogy!!!! 95) What kind of pet do you want! ?: the golden/setter at the OHS 97) Favorite carnival ride: tilt a whirl-ish rides 98) Who sent this to you?: Sasha 99) What do you think about this person?: hyper and random, but fun 100: Favorite baseball team: GO BREWERS!!!!! lol the brewers suck! but i don\'t watch baseball anyway. 101. Time Finished: 10:35 (but I took like 45 min. on the phone)   '),(6244,3823807,'02,July,2004','2004-07-02 00:00:00','   YES! more quiz results!   urlLink   Are you ying or yang?    urlLink   \"Choose\" your RFR character!   '),(6245,3823807,'02,July,2004','2004-07-02 00:00:00','   so here\'s the thing...  John Mayer/Maroon 5 concert tonight and I\'M STILL NOT GOING!!! WAAAAAAAH! k i\'ll stop now. but seriously it SUCKS!!!!  ummmm not much else to say. just I\'m so friggin bored. what else to tell tho... I never really told you much about myself tho. but if ur reading this u probly kno me anyway. so screw that idea.  COUNTDOWNS: John Mayer and Maroon 5 concert: TONIGHT!!!! :\'( Flute camp @ UW Whitewater: July 11th- 8 days Band camp (whoopty freakin do!): August 16th- 48 days 1st day of school (nooooooooooo): September 2nd- exactly 2 months MY 15TH BIRTHDAY: September 6th- 65 days  ummm ya thats all.   '),(6246,3823807,'02,July,2004','2004-07-02 00:00:00','   First blog entry! yay! k thats it.   '),(6247,4235996,'14,August,2004','2004-08-14 00:00:00','   steack knife, caro shark, con job, boot cut dog torn, blood bath, rib cage, soft tail black jake, dope dick, pawn shop, quick pick song bird, main light, cash back, hard top (By the way)  fly away on my ZEPHER (the zepher song)  cant stop addicted to the shindigcop top he says im gunna win big  To be concluded)***********   '),(6248,4235996,'14,August,2004','2004-08-14 00:00:00','   Ok one time I was playing socom and the guy, right, well he came up to my and I turned around and like I shot my 50 cal. And like he got hit right in his chest(me thinking he is dead chucked a nad at another guy not knowing the guy I shot didn\'t die) so bye then I was dead. Hoping my nad dropped before I was killed I waited but nothing happened, then about a minute after I died this guy walked over my dead body I I heard an explosion(I was in my room getting a pen and paper) I ran in and at the top of my TV it said ,GSG-9 MAGNUM(me) fragged ******(don\'t remember) with an m67 and I was like what the fuck. And I won the round for my team. THAT\'S F.U.B.A.R.(Fucked Up beyond All Recognition)   '),(6249,4235996,'14,August,2004','2004-08-14 00:00:00','   Do u ever wonder why where here, why do they but MIGHT CAUSE drowsiness on a sleep medicine bottle? does anyone ever pay attention in health accept for the sex edd part, no really? Don\'t deny it, you know it\'s true. You know all that other crap like \"saving peoples lives\" aren\'t important. In a human mind nothing else matters if the even the WORD sex is involved. That\'s what drives our livens on this earth( well except for troll\'s and jakoffasours but that\'s another story all together) if any one know the importance of a toenail. The world would be a better place.   '),(6250,4235996,'14,August,2004','2004-08-14 00:00:00','   What do I do to ignore them behind me?, do I fallow my instinct blindly?, do I hid my pride from these bad things?, and give in to sad thoughts that are maddening?, do I, sit hear and try to stand or DO I try to catch them red handed?, do I trust some and get fooled by phoniness, or do I trust nobody and live in loneness?, because I cant hold on when IM stretched so thin, I make the right moves but IM lost within, I put on my daily facade but then, I just end up getting hurt again bye my self( my self)  chorus I ask why ( but in my mind I find) I cant rely on my self( my self) I ask why (but in my mind I find) I cant rely on my self  I cant hold on, ( to what I want but IM stretched so thin) its all to much to take in, I cant hold on( to anything watching everything spin) thoughts of failure sinking in  if I turn my back IM defenseless, and to go blindly seams senseless, if I had my pride and let it all go on then they\'ll, take from me till everything is gone, if I let them go ill me out done, but if I try to catch them ill be out run, if IM killed by the questions like the cancer , then ill be buried in silence of the answer By My Self repeat chorus        how do u think, I\'ve lost so much, I\'m so afraid, I\'m out of touch, how do you expect, I will know what to do, when all I know is what u tell me to        don\'t you(know) I cant tell you how to make it (go) and no matter how hard I (try) I cant seam to convince myself (why) I\'m stuck on the out side   '),(6251,4235996,'14,August,2004','2004-08-14 00:00:00','   wtf did i make this for? w/e.   '),(6252,3367343,'17,May,2004','2004-05-17 00:00:00','    Hey everyone!      This is my first time to do this blog thing so I\'m kinda excited. I guess you woulod refer to me as a \"Blog Virgin\" cuz I actually have something to write about. On Friday Night I saw Mean Girls with my sister. It was really good, but Lindsay Lohan\'s boobs were just everywhere! It was gross! There\'s one scene where she\'s wearig this really tight strapless dress and she\'s wearing a red bra under it so you could see her bra, and she runs after the guy she likes in high heels and her boobs are just bouncin\' around... it\'s disgustin\'. But overall it was a good movie.      Friday night I had the wierdest dream. See, in History we\'re watching The Pianist, and it\'s about WWII and the whole thing is just terrible. Anyway, so I had a dream that I lived in Germany during WWII and I was hiding a Jew in my apartment. There\'s a really funny prt of the dream where we\'re sitting on the couch in like the main lobby and he wants to watch the Great Mouse Detective and I\'m really nervous that someone\'ll come in and see him, and for some reason he got really mad at me and stormed off into the room. Wierd, I know! And our apartment was like a bathroom with 2 bath tubs in it and we would have to sleep in the bath tubs...whew glad that\'s over.      I had a good weekend. My sister isn\'t sick anymore, so we got to hang out. On saturday, I thought the choir was going to have a car wash and we went around looking for it in the most likely of places, but we didn\'t find it. So we decided to go to Six Flags. The lines were friggin\' long but we waited in them anyway. They were going fast so it really wasnt a big deal. Anyway, so we were in line for the Batman and there were these really cute boys behind us. My sister started talking to them and when we looked at them, they were twins!! Oh my goodness they were so cute! Their names were Hunter and Tanner and they had a friend with them named Jonathan, who was pretty funny. I liked Tanner the best, but they were all cool guys. I think they all pretty much liked my sister more since she\'s prettier than me and can flirt more easily so she was talking more than I was. They were from Meers, Oklahoma and they were there for their senior trip. Man were they precious. So they had to go and I still am beating myself up for not giving them my phone number!! Why the heck did I not give them our number!? Gees I\'m so stupid! Oh well. So it was really fun hanging out with them. That was my Saturday.      I learned Spanish on Sunday! We\'re going to Mexico for a church mission trip and we have to learn Bible verses and a few simple songs in Spanish. So I went home and was saying the verses out loud and my dad was yelling at me cuz he was trying to watch TV and I was being too loud. Hehe  Well, that\'s all I have for now... tune in again... Same blog time, Same blog channel!   '),(6253,4169661,'08,August,2004','2004-08-08 00:00:00','    I received an email from one of my sisters today, that was very scary.  My nephew, is a police officer in a very nice middle class neighborhood, with a below average crime rate.  He was dispatched to a home, a neighbor had called 911 stating that a man was beating up a woman and was trying to force her into a truck.  When my nephew arrived at the home, he noticed immediately the truck that was reported in the 911 call.  He parked his car and walked over the check out the truck, while doing so, he heard a woman scream & the shattering of glass.  He started running up the drive way of the home, as he turned towards the back of the house, he caught a glimse of a man lunging forward in his direction.  (There was another officer running behind my nephew, but in his shadow away from where the guy was - out of sight.)  The man lunged toward my nephew with a butcher\'s knife in his hand!  My nephew had a split second to react, as the knife was coming towards his face! He turned his head to the right just far enough in that instance to get stabbed in the skull about 1 1/2\" above the hairline over his left eye. The man lunged with such force at my nephew   (My nephew\'s a big boy - 6\'2\" 240#), the impact knocked him to the ground!!  The other officer still didn\'t realized my nephew was stabbed ! My nephew had pepper spray in his hand, which was knocked out of his hand, as he fell to the ground.  My nephew was able to get up quickly and the guy was still in view with the knife. My nephew drew his duty gun and shot the man hitting him twice.  The man died at the scene.  Lucky my nephew didn\'t have too much time to react or he probably would have got the knife in the neck or the chest. The Doctor kept telling him how lucky he was when he was in the ER where they put 3 staples in his head in the deepest part of the stab wound. My nephew has never had to kill anyone on duty. He actually thought he might die because he could not tell how severe the stab would was. He called the dispatcher and told her he was stabbed and then struggled to get to the curb. He was  bleeding  and worried that if he didn\'t get to the curb they might not find him .He  thought he could bleed to death.He said all I kept thinking about was my family and my children.  This man had already severly beaten his ex-girlfriend, when his mother tried to protect the girl he slashed his mother across her face.  There are times when you watch TV, and think; do those types of crimes really happen?  Or, that\'s horrible!  But, when it happens to someone in your family, reality hits home!  I have always been proud of my family members who serve in the military, and those who work in law enforcement (another sister in addition to my nephew).  However, I often try not to let myself think about the dangers too often.  Then there are times like this one, where you just have to be thankful that your family member made it through the situation alive and well, although I\'m sure there will be a recovery period - mentally.  I am not a religious person, but you can bet I will be saying a prayer of thanks this evening!   '),(6254,4169661,'07,August,2004','2004-08-07 00:00:00','    I\'ve been reading about blogs in newspapers and magazine articles lately, and decided to check them out and now to give it a try.  For the first time in a long time, I am sitting by myself ~ bored ~ looking for something interested to do, which is very strange!  Normally, I\'m hanging out with my kids, but they are all out with friends tonight.  Can\'t say I blame them as they are 20, 22 & 23 and in the prime of life!  Who would want to hang out with their Mom at that age- you say?  Well, I\'m not exactly old - 41, and I am fun to be around, plus have great kids. On the other hand, I am married, but my husband and I are extreme opposites, which can be a good thing, but often we spend little time together in the evenings....he currently preparing his boat for a day of fishing on the lake, with our youngest daughter.  (Hopefully she\'s not too hung over in the morning...ha ha...or he may find himself on a afternoon trip!)      So, I\'m sitting her at the computer with three dogs laying at my feet, in the peace and quiet of a big house!  Actually two dogs, and a very rampunctious puppy, who at the moment is sleeping and looking adroable.  Chico is his name, he is an Austrailian Cattle Dog, at least that\'s what the lady at the pound said.  We were told he would be a medium size dog, but I\'m currently thinking he\'ll be BIG!  Oh well, unconditional love from a medium or large dog, it doesn\'t matter!   '),(6255,3482840,'19,July,2004','2004-07-19 00:00:00','   This  is my new blog all about my school and social related life. C.Love Diaries is my AWAY from school life blog. If it happened at school, it\'s here. If it happened over the weekend, it\'s THERE. Get the picture?     NEway,  the title of this blog is \"Let the Gossip and Backstabbing Continue.\" This is so because, HELLO this is High School. I do NOT plan on living it like my mom did. Only one or two friends, who were constantly being threatened if they didn\'t make fun of my mom. Once my mom\'s BF Melda ((i know... weird)) was being threatened that if she didn\'t call my mom a warthog ((since that\'s the worst there was back then)) these other girls would basically kick her @$$. So she says, \"Rosie, I hate to do this to you but you know I love you. WARTHOG!\" And my mom threw her across the floor and into the wall in the cafeteria, which I\'m assuming really hurts when smacked against your skull. OUCH.       Yeah  well, since school hasn\'t started, I\'ll leave you now. Later daze!'),(6256,3482840,'05,August,2004','2004-08-05 00:00:00','   Mrs. Miller is our new Junior/Senior High Principal. Weird since she WAS the special ed. teacher a while back. NEway, now the Wagner formerly known as \"Principal\" is Superintendant.  I wrote Principal Miller a letter to get her to bring the dance team back. I think it\'s pretty convincing, and we put on there that she should scheduale a meeting for us to discuss it. It would be SO cool if she brought it back finally!  In 2 weeks I have 2 pick up my scheduale at the school. We have a new French teacher too! Her name is Mrs. Melha Woods... her hubby Mr. James Woods, is teaching Physics, Chemistry, and Environmental Science. Mrs. Hart is teaching Physical Science, General Biology, AND Health now. Talk about multi-tasking...   '),(6257,3482840,'14,June,2004','2004-06-14 00:00:00','   Well my punishment for everything that happened with me, Rachel, Crolann, and Sammy is that I\'m not allowed on Blogger for the entire month of June. But July 1 u can expect me back! o:) I guess it\'s b/c mom doesn\'t like the language Rachel used on her blog, even though they don\'t do that NEmore. Well I\'ll miss u all! Talk 2 u in July...  ~Byeness!~  '),(6258,3482840,'13,June,2004','2004-06-13 00:00:00','   I am  SOO  bored right now... so I come online and talk 2 ppl I don\'t even know. Sounds gtm. Well that\'s me i guess... is garfield the movie out yet? I think so but I don\'t remember. Haven\'t talked 2 Heather yet. I should call l8tr. MK called this morning but no time 2 talk. g2call her when i get offline. that means in an hour! *lol* Talked 2 Jared friday. It was interesting... especially about the  \"switch\"  thing. honestly idk y he would ask that but it came up so...  no , jared, I don\'t think I would. I still think it was odd that u asked though. oh well, g2g check otha blogz. Byeness!  '),(6259,3482840,'13,June,2004','2004-06-13 00:00:00','   Well I\'m back from Cincy in time 2 post! It was SOO fun, even though I didn\'t win. Some girl who did it last year won. Oh well. I still get a crown and a sash and a trophy. Oh yeah and the life-altering experience of standing on stage with the cool ppl instead of the snobby ones! ::lol:: OMG u know how many pix my mom took?!?! Like, a BILLION!!! So annoying but hey, she loves me. That\'s all I care. ok well I g2g see my other blogs. L8tr days guys and gals. Luv u all!!!  '),(6260,3482840,'11,June,2004','2004-06-11 00:00:00','   \"Cas\" meaning casual. I know what I\'m wearing 4 casual, and I wanna wear the dress from Michelle 4 formal, but my mom says it\'s not formal enough... w/e. so Grams is taking me out 2 c if I can find something else I like. But I don\'t wanna wear a gown and look like Cinderella @ the ball ok?! -_- I refuse... so there... well I\'m leaving 4 Cincy 2morrow as u know... mayB I\'ll call sum of u guys from the hotel! They have 2 phone lines in every room! ^_^ sounds GTM ((good 2 me))!!! Ok well, c u all l8tr!  '),(6261,3482840,'10,June,2004','2004-06-10 00:00:00','   OMG... 2 days until my competition! Uncle Sonny cancelled his shift that weekend so he can come out with us 2 see the show. It\'ll take 4 hours and 7 minutes to get there, which is why that title. U see? Aha! Not so dumb after all! J/k!   Kevin and my mom are at the store. I\'m  SOOO  glad school\'s finally over!!! Miranda\'s phone is broken and Heather has been   ASLEEP ALL DAY!   Erik said she went 2 some thingy late last night... w/e. Candy, ur \"plans\" may not go over on my favor. sry. explain l8tr and not online.  I\'ve learned from past experiences!   Well my hamham is still alive, so is Kevin\'s hamham and mom\'s rat ((who btw is a  girl !)). Neway glad u guys could come read. And, yeah Carolann read my old one so she knows it was my little \"life-story\" blog. heh heh.  And when me & Heather get it worked out, we have a  talk-show blog  4 u guys 2 see! But I won\'t tell u the URL until it\'s ALL ready! When it is it\'ll be posted on  THIS BLOG  so check back in a few days or so. And it\'s on a different username so u can\'t look it up on my profile. So there! U have 2 wait! ::lol:: Ok well this is getting long so, TTYL!  ~Byeness!~  '),(6262,3482840,'30,July,2004','2004-07-30 00:00:00','   I\'m at Carolann\'s right now. I met her Grandma! She\'s a little scary... she was playing with my hair... IDK why but... I met a whole buncha ppl over there @ Carolann\'s party, but I don\'t care to remember all their names. Her uncle\'s cool though.  NEway,  I actually watched an R rated movie, and it wasn\'t as scary as Carolann tried to make me think... It was THIR13EN GHOSTS in case you care at all... And we watched Galerians:Rion 2. It wasn\'t actually all that bad... but the makers of that movie are severely twisted. Either that or they have a LOT of time on their hands. But I DO like the soundtrack... which we r listening 2 right now b/c Carolann has it.  Yeah well.  Ooh we had cheesecake! That was good... and popcorn and chips  ((mild salsa, not 2 hot! *lol*)) And her mommy bought me  Stewarts root beer... mmmm... that\'s the best! Well we rented  Along Came Polly  and  The Perfect Score  so it\'s movie night! Hmmm... that\'s pretty much all 4 now. Catcha l8tr ppl!!!  '),(6263,3482840,'29,July,2004','2004-07-29 00:00:00','   I\'ve been on Yahoo messenger w/ my cousin 4 the past hour playing w/ the webcam. This is SOOO cool!  I\'m going 2 Carolann\'s 2maro. It\'s her b-day. We\'re gonna watch movies n just hang. And prob go online so if we do I\'ll update from there. Well for now I g2g. L8tr!!!  '),(6264,3482840,'28,July,2004','2004-07-28 00:00:00','   IDK why you would care, but that\'s what I\'m eating. Hey, that\'s what is going on right now. So there. We\'re trying 2 hook up the webcam b/c Mandy got 1 recently n wants us 2 use ours. should be cool.   I hope I\'ll be able 2 go 2 Rachel\'s b-day. I mentioned it 2 mom but she never said NEthing. I\'m supposed 2 call Carolann so we can see what\'s goin down 4 HER b-day in 2 days... HAPPY ALMOST BIRTHDAY!!! So I should prolly call her l8tr.  Jared\'s @ band camp... I can\'t help thinking of Amer. Pie.... *lol* I\'m going 2 the Medieval Faire this Sunday after church. Church seems 2 be my fave place now. Heh heh... *lol* We all know why, Andrew &amp; donuts! Mmmm.... custard.... Well that\'s all so I g2g!  '),(6265,3482840,'26,July,2004','2004-07-26 00:00:00','   I just got an invite to Stone Model & Talent Agency\'s auditions next month! And get this, they\'re for modeling, acting, singing, AND dancing, ALL of which I do. How awesome is THAT?!?! I\'d say pretty dang sweet. -^_^- I\'m so excited!  So every1 hope and pray 4 me that I get 2 go, first of all. Then maybe I can be REALLY good at one of 4 things that I love, all of which they\'re looking for. I feel proud of myself.  Well that\'s all 4 now. G2G!!!  '),(6266,3482840,'25,July,2004','2004-07-25 00:00:00','   ... to Grandmother\'s house we go. I had to stay @ Grams\' house last night b/c mom had a show @ Hickory ((DJing)) and Kevin\'s with us b/c his dad was busy and couldn\'t take him. You can pretty much guess they\'re divorced. NEway, I spent yesterday watching TV ((Fairly Oddparents Channel Chasers!)) and playing with my Grandma\'s new kitten. She\'s all black so I wanted to name her Mystery, but Grams can\'t say that right so we named her Misty. She\'s so cute! =^_^=  We\'re skipping bible study @ church this morning b/c mommy has a bad headache. NOT my fault she\'s not used to the bar NEmore... loud, drunk ppl singing... Youch! But I hope we\'re going to service. Church is a lot more fun now. ^_^ But u know y! Well that\'s all 4 now! L8TR!!!  '),(6267,3482840,'23,July,2004','2004-07-23 00:00:00','   Thimen Thez is a duck on those funny cards at the stores. I just read ALL the ones they had where mom was at. This is my fave birthday one: *clears throat* \"I thay to you from all uth ducks... ((inside)) ... getting older really thucks.\" And, \"It\'s your birthday! ((inside)) Get down with your bad thelf!\" Haha... I think he\'s cute. Color me crazy...  I went to the dentist for the firstest time 2day! Yay for me! I got my teeth cleaned. Uggh... one word to describe it. GUNK. It was gross! But she gave me a sticker :) heh heh. And the new Crest Whitening Expressions toothpaste! Mmmm!!!  Mom has bible study over here 2night and we get PIZZA! WooHoo! Nothin like greasy, cheesy, yet still healthy food after a trip to the dentist. But I have 2 go 2 the optomotrist on Monday... *sigh* \"Life Thucks.\" *lol* Well, I g2g surf blogger! L8TR!  '),(6268,3482840,'22,July,2004','2004-07-22 00:00:00','   Life is dull, life is dull, life is dull, dull, dull... There\'s no1 2 talk 2! I\'m gonna go NUTS!!! 25 days until my birthday. Aunt Kim is most likely gonna take me clothes-shopping like she ALWAYS does. Yay! That\'s always cool.   I\'m hungry! I want a bagel. But we\'re out. Cereal\'s good. But we\'re almost out. Cinnamin Raisin TOAST!!! YEAH! Ok I\'m gonna go eat then. ~Byeness!~  '),(6269,3482840,'21,July,2004','2004-07-21 00:00:00','   ~ My mom had an interview @ Holly Hill nursing home 2day for a job up there. If she doesn\'t get that one she already has another job lined up. So THAT\'S cool.  ~ I got rid of all ((well, most of)) my stuffed animals tht were all over my bottom bunk! That\'s means when my friends come over, they\'ll actually have sumwhere 2 SLEEP instead of on the floor! I mean, they could sleep there b4 but there were SOOO many animals there wasn\'t much room... *smacks head* I fixed THAT though!!!   ~ Mitch & John hired my mom back at Hickory doing karaoke once a month just 2 see if they really DID lose a lot of ppl. HINT: Yeah!!!  ~ Ok well there\'s nothin else. G2g read every1 else\'s blogs 2 see what\'s up! L8TR!!!  &nbsp;  '),(6270,3482840,'20,July,2004','2004-07-20 00:00:00','   You know no1 has called me, invited me over, emailed me, been on my blog, NOTHIN\'! HOW RUDE!!! I can\'t wait until my birthday b/c then I get to see Michelle, Liz, Britt, Cortney, and all them. Ppl are just so thoughtful aren\'t they?'),(6271,3482840,'19,July,2004','2004-07-19 00:00:00','   I made a new blog just for school stuff! It\'s    urlLink www.back2school.blogspot.com   THIS blog is for stuff that happens OUTSIDE of school, like summer and weekends. THAT one has everything that happens DURING school. Coolie or what?! G2g! L8tr!  '),(6272,3482840,'18,July,2004','2004-07-18 00:00:00','   1) I got the JoJo CD! And I got the Cheetah Girls Soundtrack w/ karaoke tracks 2. &nbsp; 2) I met Andrew ((the guy from church)). He\'s pretty cute, but he\'s SOOO shy he barely said hi to me. All I know is he\'s 14, going into 8th grade, and he goes 2 West G. &nbsp; 3) My mom got a new cell, and the guy asked if she wanted 2 put a celly 4 me on the account. Unfort, mom said \"That\'s not happening until she\'s allowed to date and that will be a while!\" O_O HELLO! I\'m almost 14, I\'m going into high school, and I always get good grades so it\'s not like I\'m irresponsible... -_- She\'s impossible. She\'ll let me go out w/ KK ((eeeeww...)) but she won\'t let me even SPEAK to Andrew.  &nbsp; So so so SOOOO not fair! It wouldn\'t be a NEW kind of phone, just 1 like her old one. Just simple and CHEAP. But nooooooo... screw it.  '),(6273,3482840,'15,July,2004','2004-07-15 00:00:00','   IDK y but my house is freezing! Mrs. Hasman woke me up b/c she called and I\'m the only 1 home. She wanted 2 ask my mom what is was like getting her belly botton pierced... apparently Mr. H is talking her into it... weird. So yeah. Yesterday I read  Midnight Magic  by Avi. It\'s reallly good. U know the whole, ghostly conspiracies, old world castle w/ hidden passages, secret identities, fate told by tarot cards, a thrill-seeking princess, and a new twist every other paragraph. That\'s my kind of story!   My mom has bible study ppls over here 2night so I have 2 clean the house while mom\'s @ work... *sigh* If only I had a life during the summer. Uggh...  This weekend I\'m going 2 get my JoJo CD, and mayB another 1 but idk what yet. And then the 23rd I have a dentist appt. and an eye-doctor appt. the 26th. Oh how I love the optomotrist... NOT. I hate getting my i\'s dialated!!! It\'s gross...  Well that\'s all 4 now I guess. TTYL!  '),(6274,3482840,'14,July,2004','2004-07-14 00:00:00','   Well, this weekend I\'m getting JoJo\'s CD! WooHoo!!! I\'m so happy! And IDK when, but I\'m gonna get my ears pierced. Too many cute earrings that I can\'t wear ya know? Totally. And my uncle is getting me the Cheetah Girls DVD 4 my birthday! Which is in, 33 days but still. It\'s all good in the hood. Not that I LIVE in the hood but hey! *lol* Y\'all gotta know I know what I mean.   There\'s not very much going on though. I\'m SOOO bored ALL the time. My mom might be getting a job at Holly Hill though so that\'s cool. I found a stuffed hampster that looks just like Smurf! That\'s all 4 now so, L8TR DAZE!!!  '),(6275,3482840,'13,July,2004','2004-07-13 00:00:00','   Mom and I spent ALL of last night splicing every taped spot in our phone lines, b/c the phone wasn\'t working. So we FINALLY got it 2 work and I can go online again! WooHoo! I would seriously DIE w/o the internet... Just WAY too dependant on it I guess. Well like if I had stuff to DO I wouldn\'t miss it. Like if I was at camp or something. But I\'m not and this is all there is to do.   Well idk if Rachel\'s spoken 2 luverboi at all... *cough*Jared*cough* but my mom still doesn\'t want me 2 be friends with Rachel. OR Carolann OR Heather. Basically she wants me to have NO friends in High School, just like SHE went through it with 1 friend... She LOVES my friends, \"just doesn\'t love their behavior.\" OK as if APOLOGY ACCEPTED means nothing? That\'s what friends do. Forgive. But not neccessarily 4get.  '),(6276,3482840,'10,July,2004','2004-07-10 00:00:00','   Ok so this guy I was supposed 2 go out w/... never met, liked him a little but not boyfriend material. He used to call every week, but he hasn\'t called in the past 3 weeks. I then found out that my best friend was gonna go out w/ him. So naturally, I got mad @ her. BIG mistake. I found out from another girl online who lives by this guy, that she has been w/ this guy for 6 months, which is longer than he had even been talking 2 me, let alone my bff. So I tell her, even though I\'m mad at her 4 another reason. So now we\'re both mad as heck at this guy, b/c:   1) He ditched me for my best friend. 2) He cheated on us BOTH w/ the same chick who he\'s been w/ half a year.  Even though I don\'t like him, I never TOLD him that. So as far as he knows, I\'m still at home absolutely \"smitten\" and waiting for his call. Oh I\'m looking forward to his call alright, but I guarantee you, he won\'t find it pleasant AT ALL.  ~ME~  '),(6277,3482840,'09,July,2004','2004-07-09 00:00:00','   Yesterday I cleaned off all my shelves in my room and got rid of all the junk. Now, 6 shelves across my whole wall full of junk, has become 2 shelves only half-full. Pretty good if I do say so myself. Next I have 2 clean out my cabinet... here we go again. But while cleaning, I found an old radio/tape player that records w/ a buit in microphone. :) That\'s perfect 4 me! Well I have work to do.  ~*+*~BYENESS!~*+*~  '),(6278,3482840,'06,July,2004','2004-07-06 00:00:00','   We\'re seriously not doing well on the moving thing. I don\'t wanna go back 2 living w/ grams! I mean, I lived there for 9 years, and HERE only 4 years. I finally got out and into out own house. What are ppl gonna think when they find out I\'m living w/ grams again? I don\'t even wanna think about it.   There\'s not much else going on. I was just @ Heather\'s. We made reallly cute earrings! Not that either of us can wear them but still! Yeah well, I\'m gonna make grilled cheese. Or as the rainbow-folk would say, \'toasted cheese sandwiches\'. *lol* Only kidding! L8tr peeps!  '),(6279,3482840,'04,July,2004','2004-07-04 00:00:00','   Happy 4th of July to all of you in the USA!!!    Since mom\'s cash is running low and the job aps. aren\'t workin out, she\'s actually considering us renting out the house and moving to KIWANIS! Thanx, but slumming is not included in my interest category... NEway, we were just @ my cousin Danny\'s in KL and I saw Jason. He finally cut his hair! That was a shocker.  Mom invited Mimi & Lydia over 4 a cookout 2morrow. Great isn\'t it? Yeah right. Just how I wanted to spend 2morrow after being w/ Lyd ALL weekend. Ah well, nothin I can do about it. So, I g2g. TTYL!  '),(6280,3482840,'03,July,2004','2004-07-03 00:00:00','   All I did like all weekend was play with Carolann and Lydia w/ hula hoops. And I got a dragon statue from mom\'s friend James 4 working hard. More like hardly working! Yeah well it was fun. We got to ride around in the Jeep w/ the top off. That\'s rare actually! It hasn\'t been warm enough yet. Ahhh... well. N2M 2 say... God my life is boring these days. Wait! I DO have something funny. Riding around all weekend w/ the top off the jeep, things kept flying out the windows. U can guess what they were from some clues.  1) It\'s hand-held, can be used with a TV or  radio , and is an aid to laziness. 2) It\'s round, smells good, and can be used in a  car  or in the house.  So can u guess what flew out the windows? I\'ll give you the answers when I have a few guesses. OK, talk 2 u all l8tr then. Byeness!  '),(6281,3482840,'02,July,2004','2004-07-02 00:00:00','   Carolann\'s over. We were up till like 3 which isn\'t bad I guess. But we had to get up at 10. :( Seriously uncool. So NEwayz, yeah I\'m bored. There\'s nothing 2 do except this, and we\'re going 2 my mom\'s store l8tr. Woo and Hoo! Yeah, ok no one cares. Which is ok b/c IDC that much. *sigh* Well then, Carolann wants 2 go on her blog so, Byeness!  ~Cara~  '),(6282,3482840,'01,July,2004','2004-07-01 00:00:00','   That\'s what you guys missed out on while I was banned from blogger! Good times! *tear* It\'s almost as if you\'ve all left me... which you have. I don\'t see evidence of anyone having been on here since I had to leave. Then again, why come back if there\'s nothing new? My point exactly... I know what I mean.  OMG I had such a weird/cool/freaky dream last night. Read about it on my other blog: www.ucap.blogspot.com  And I\'ll TTYL!!! Missed you all SOOO much!  ~Cara~  '),(6283,3482840,'11,August,2004','2004-08-11 00:00:00','   I went with Grams and saw the Princess Diaries 2: Royal Engagement today. It was a great sequel! It had a really cool ending too. Word is they\'re thinking of a third movie!!! YAY!!!       Then we went to KMart so I could get Rachel\'s b-day gifts. Of course I got some of my OWN as well. This IS me we\'re talking about! Grams bought me 2 new pair of shoes. One are white SOHO courts w/ blue stripes on the sides, and the other are navy and pink leather tennies for gym. But when I don\'t HAVE gym they\'re for all the time. FINALLY new shoes!!! I miss shopping SO much.      And then we went 2 eat at Mangia\'s and I saw Eileen. DUH b/c she works there! But w/e. Then I came home and got online. GAH! I almost 4got! KEVIN\'S BACK FROM HIS DAD\'S!!! NOOOO!!!! Friday can\'t come soon enough so I can go to Rachel\'s and get away from him!  '),(6284,3482840,'10,August,2004','2004-08-10 00:00:00','   Ok so we actually went to see A Cinderella Story w/ Hilary Duff and Chad Michael Murray ((who, btw, is SOOO hott!)). But we\'re going 2maro 2 see the Princess Diaries. Julia still doesn\'t think we can do it. And Rachel spent the night Sunday night. She decided to quit volleyball but every1 yelled at her so she\'s gonna break down and join back up. I knew it. She should just tell them it\'s her decision and stand by that.  '),(6285,3482840,'08,August,2004','2004-08-08 00:00:00','   I\'m going with Grams and Rose O. to see the Princess Diaries 2: Royal Engagement. It should be an awesome movie! I\'ve already seen the first 5 minutes of the movie on a Disney 411 Exclusive.  Andrew was at church again. He still ignores me completely and runs out of the building as soon as it\'s over. So I NEVER get a chance to say hi... :( It\'s not cool...  Julia also said that we can\'t bring back Knightettes unless we have a band, and she said we have to raise $1,500 before school starts. No offense, but she\'s crazy! At clinic we used CD\'s and tapes and played them over the PA. O_o? Oh well. If she doesn\'t even wanna TRY, more power to ya Jules. Still luv ya.  Well I\'m chatting w/ some ppl so, I g2g! L8tr!  '),(6286,3482840,'07,August,2004','2004-08-07 00:00:00','   My mom has a karaoke show in Geneva that goes till 11pm tonight, and starts in an hour. I haven\'t seen mom since 10 this morning and I\'m already bored out of my small, fragile mind. So I\'m gonna call Grams and see if she\'ll take me somewhere. I dunno where but, we\'ll see. If Grams was like Carolann\'s nana ((scary...)), I wouldn\'t love her so much.    And Mr. Kuchta\'s family history project was so cool, since I interviewed Grams. I had no idea that she married my Grandpa TWICE, once in Italy and once in the US, or that she was a lunch lady at OUR SCHOOL years and years before her oldest son was a twinkle in Grandpa\'s eye. It was so awesome that she even remembered! After her 3 strokes and loss of memories, we had to teach her how to tie her shoes again! But she could remember stuff from farther back than I ever cared to think about.    She even told me about how my Grandpa worked at Carling\'s Black Label Brewery, but he never drank alcochol at all. And she told me how they got arrested on thier honeymoon because they weren\'t sure if Grams was really a US citizen. Grams is SO cool. Oh, and she\'s an AWESOME cook!  '),(6287,3482840,'06,August,2004','2004-08-06 00:00:00','   Rachel\'s asking ppl if they wanna be on the team, Heather\'s in, and I\'m gonna call Carolann and try to GUILT her into it. *lol* \"All the things we do for you!\" That should work. Now I was thinking for one of the songs, we could use \"One-Girl Revolution\" from the Cadet Kelly Soundtrack. I\'ll see if I can get it. Meanwhile, here\'s the lyrics:                  I wear a disguise'),(6288,3482840,'05,August,2004','2004-08-05 00:00:00','   OK guys here\'s the deal. You know Mrs. Miller is our new principal right? Well, I wrote her a letter to see if she\'d bring Knightettes back since the second levy passed and it\'s not back. I think it\'s really convincing, especially since my mom\'s friend Nicole wants to coach us! I know Heather\'s interested, I doubt Carolann would be but you could do it with us ((!)), and Ray you probably are but I need to know.    When I have the meeting w/ Principal Miller before the new year I want to be able to tell her how many girls would do it. Rach, if you would could you ask Vicky, KC, Julia, Hali, and whoever else if they\'d want to do it? Thanx in advance. I REALLY wanna do this and she won\'t unless she really thinks enough girls would want to. Trust me, Nicole\'s been in dance since before high school and she still dances in competitions. She can do this if we want her to. Everyone loves our moves Ray, let\'s get serious about it! Heather\'s in already, yay Heather!!! I\'ll CALL and ask Carolann.  '),(6289,3482840,'02,August,2004','2004-08-02 00:00:00','   We went to the Medieval Faire yesterday and we brought Carolann. It was a LOT of fun. I saw one of their Jousting Tournaments for the first time, and our champion won. Of course! We got lost in the Labyrinth trying to find the bell, but Carolann & I still beat Kevin at finding everything. Mom bought me an Elven crown, that\'s a single strand w/ a hanging crystal. Mom\'s is the same only hers is two strands woven together w/ a crystal. There was this one guy there at a stand, who dressed his little poodle up as a dragon. He was sooo cute! The dog, not the guy.  Aside from all that, we saw pasrt of the Tortuga Twins\' show ((\"We\'ve upped OUR standards, now UP YOURS!!!\")), some dagger-juggling, a comedy show by Issac Faulkes - Magician and his assistant, and bagpipe players. It was really cool.  Well that\'s all for now. G2G!  '),(6290,3460713,'29,May,2004','2004-05-29 00:00:00','   Lately.. I\'ve been thinking about happiness and what it really means. Whether it\'s something attainable or simply just a state of being. But then the more I think about it, I look back at my life and ask myself whether I really am happy. For a few days, I had been confused. I just wasn\'t sure whether I really was happy, trying to be or plain unhappy.   Luckily, I was able to overcome this big fuss over my own happiness. I thought to myself that I shouldn\'t be thinking of myself all the time. I mean, I need to be at least a little selfless at times and quit thinking of my personal pleasures by trying to look around my environment and see if they\'re also happy. When I said my environment, I didn\'t mean just anyone. I had to start with the people I knew. Is my family happy? They seem okay but you\'ll never know. So I thought about doing something to cheer them up a little. I sent my mom a little sweet quote just to tell her that even though I\'m quite far away, I still do remember her and appreciate everything she\'s done for me. Then I thought about my friends, I had a lot of acquaintances being disguised as friends but to those people whom I can really call \"friends\"... well I can name a few. How are they? Are they alright?! Well, good thing I went bonding with this friend of mine last week. We went to the gym, then we ate at her place. Anyway, I haven\'t seen her in quite a while and it\'s nice to catch up on things. It\'s ironic how opposite the two of us were during the first semester, we sort of had a little indifference because we didn\'t have much in common, but now, we even go to the gym together.   Anyway, I\'ve learned that when we talk about happiness, it doesn\'t have to be you alone. I mean, a bunch of cash and a new cellphone would make me happy.. but I\'ve realized that a day with my \"G\" in Greenhills just browsing around the tiangges is even better.   You see, money or material things could never ever replace happy times with your family, loved one and friends. And now that I\'ve checked on everyone I care about especially my \"G\", I know there may be problems once in a while, but the fact that I\'ve showed them that I\'m just here no matter what and that I will try to make them happy, that also makes me really happy.  I don\'t have much money right now, actually, I\'m more than running out.. but the smile on my face couldn\'t be much bigger because \"G\" is making me so happy these days.   So I would rather cry a thousand times and be happy afterwards with the one I love. Rather than to smile when I want to and cry in the end. I know it\'s not only me who wishes that.   Happiness starts from you, but it doesn\'t have to end there.. it goes around and is being shared by more people.  I hope I made sense. If not, let me make it short.. I really am happy now, and I hope you will be to. Take care.   '),(6291,3460713,'28,May,2004','2004-05-28 00:00:00','   Trust had been such a big issue with the relationship I\'m having right now. You see, when we were just starting, I had already gave my trust to him. Sadly, he had broken my trust. What\'s more sad is, it had happened too many times. It\'s partly my fault because I allowed him to toy with my emotions, and I didn\'t punish him for breaking my trust. But you see, it\'s not that I\'m being such a martyr but a matter of not giving up. I never gave up on him and our relationship no matter how many times he had hurt me. I still tried to look on the bright side and I didn\'t lose my faith in him to change.  Fortunately, I see a big change in him right now. I know it has to be a two-way thing. Actually, I hope he really does trust me as much as I trust him because that\'s going to help a lot. Lately, he had proved to me that he\'s trying everything he can not to hurt me anymore. And I\'m so happy for that. It is indeed a wonderful feeling that I simply could not trade with anything else.  So I keep on praying and hoping that my trust in him will not be put to waste. I see his efforts so I really wish he would keep it up. Because if we only try everything we can not to break each other\'s trust, there wouldn\'t be a problem.  It\'s just so nice to have someone to love, someone to trust. :)   '),(6292,3460713,'27,May,2004','2004-05-27 00:00:00','   Yesterday, I had been robbed while I was doing a little shopping at the \"tiangges\". Its was partly my fault, I have to admit. Because I already knew that the place was dangerous for a girl like me to be roaming around alone. And yet I still went there, alone. So, I had to learn my lesson the hard way.  Though I lost my money, like what my loved one told, it was only \"money\" and not my life. So I had to be thankful that nothing more dangerous had happened to me. As for the love of my life, well he scolded me for being stupid. But I know he means well, I deserve to be scolded. It also means that he really does care for me. I felt so bad that I had lost my money because I also felt so stupid. So I asked him to fetch me, and he came. My savior! Hehe. I felt like a damsel in distress waiting for my knight in shinning armour.  Anyway, I must be really lucky that day because nothing more dreadful happened to me. It was a lesson learned.. I had to pay the price, but as they say, no price is expensive enough because at least I was safe.   '),(6293,3460713,'27,May,2004','2004-05-27 00:00:00','   Something keeps me smiling all day long.. I don\'t know why, maybe I do It doesn\'t matter as long as I\'m smiling... oh baby, I love you It\'s really not a matter of what makes me smile... but who makes me smile It\'s my baby, he makes me so happy.. that I can\'t help but smile  Sure sometimes, he makes me frown a little.. but nevermind He makes me feel a lot of thing.. it\'s coz he\'s always on my mind I know we\'re going to last forever and that really makes me smile We\'re not one of those couples who would only last a while  Baby I\'m so thankful for the smiles you give to me Even though you\'re not around now, I wish you could see That this smile is because of you, you make me smile all day I simply am thankful that you\'re here with me everyday! :)   '),(6294,3460713,'27,May,2004','2004-05-27 00:00:00','   Today, my pride within me was tested with my loved one. I need not relay the whole story, but the incident that happened to us today had proved a lot to us. Not only to our relationship, but also to how much pride I am willing to swallow for my loved one. I\'m not saying that someone was right with what had happened, no one has to win the argument. So I just tried my best to fix things as soon as possible so that it wouldn\'t grow into a bigger argument. I wasn\'t trying to escape the long dreadful argument that was yet to happen, but I was trying to look at things on the brighter side by swallowing my pride and see where the argument might lead. And just like that, things got better. If I had not thought of a more peaceful way of solving things, it really would have gone worse. But that would only be useless if there really was a better way. So as I\'ve said, today my patience and my pride was tested, and I think I aced the test.   '),(6295,3460713,'26,May,2004','2004-05-26 00:00:00','   Relationships need to be conventional. True, everyone might be vying for that stage where in everything seems to be at the right place and time. Yet, wouldn\'t you think it\'d be boring that way? Believe me, if keeping a relationship with someone or being a good partner was so easy, love wouldn\'t be that special. It\'d just be like, mineral water or jello.. no offense to jello lovers.   My point is, when you and your loved one has been given this title of being unconventional, as long as you two still love each other, there is nothing wrong with that. Because it may only mean that you\'re keeping pace with each other in this fast-paced world that we\'re living in.   * I have no idea why I had felt the sudden urge to write about unconventional lovers tonight. Though I\'m still proud to be writing about love. As for my personal relationship with my one and only \"G\", I think our relationship may be a little bit unconventional. But that\'s what makes it more exciting and that\'s what will test us to be stronger each day. *   '),(6296,3460713,'26,May,2004','2004-05-26 00:00:00','   If he wasn\'t here with me right now, I\'d probably be dead. No kidding. Before I met him, my life had been filled with false pretentious and superficial happiness. It was like I was stuck into the hellhole and there was no way out. Yes, my life was sad. Meaningless. It looked fine from the outside, but from the inside, it just wasn\'t good.   Before, love wasn\'t much of an issue. I mean, I have loved before but at such a young age, I had been fooled and hurt too many times that I thought \"true love\" really didn\'t exist anymore. Pretty mushy huh? But when it comes to me being in love, there is no limit to how mushy, dramatic or sensitive I may become.   You see, there are a lot of things I can say about the love that I have right now. One of them is when I consider love as an art. Like a sculpture being made just for me. My love for \"G\" is truly an art. He is my masterpiece. And he saved my life!   '),(6297,3674747,'27,June,2004','2004-06-27 00:00:00','   Proverb of the day: The lifespan of a person is based on the amount things he/her had done during that life.   Haiz...tml the hols end le...Mus study le wo...Feel a little sian but keep thinking that must work hard le...>. WILL  regret. In the hols, i did not done mus things. Everydae is play play play...hahx die le. Got  crazy  with a online game call \"Ragnarok\". I siao liao... play lik crazy. haha. My mum say i was poisoned....... Fine.(i was a little.)haha Anyway... tml can meet you all again...quite delighted... i got a question for you all... you all giv comments hor... Q-- Many ppl say tt China will stand out in the world market and become one for the most powerful country in the comin yrs., Do you agree??  Sorry ar...Very boring rite? Next time find a better one...Buai.   '),(6298,3674747,'19,June,2004','2004-06-19 00:00:00','   Proverb of the day: Everyday is a new beginning and you should spend your time carefully.  Haha... dunno wat to write, so i decided that i can write a proverb everyday...^^. well, todae nothing special happened la... onli playing of games, watch tv and read alittle..=X i think i spent to much on games..haha(dun follow me) den now cannot concentrate on study le...die. Therefore im stopping all my games stufs...(maybe not all ^^) so lata go school won keep on thinking about it... Btw..i think i will not be writing blog everydae...lata my brain juice no more. nuttin to  write la. so pai seh...^^ i will onli be posting the proverbs onli..haha. Buai   '),(6299,3674747,'18,June,2004','2004-06-18 00:00:00','   Haizzz... Dis my first blog, so i dunno whether is it good... Need you comments la...haha Anyway, i write dis cos many of my friend oso got write, so i join in the fun. ^^ Anyway, now very late le...so cannot write the things tt happened 2dae. haha...don get disappointed lor... tml i will write one, relax.  I wan to ask you all who hav done your hw le??? I onli do half way lor... die. Surely rush last minute one. Nvm... used too it le. Anyway, wish you all a good hols & enjoy yourself... cos after dis, it time to prepare for one of our\'s greatest battle!!! Good luck.   '),(6300,3674747,'09,July,2004','2004-07-09 00:00:00','   Proverb of the day: Clever(chong ming) is not equal to smart(zhi hui). The clever guy doesnt mean tt he is smart, but a smart guy is definately clever.  Hihi... Very long nvr rite le...Sorri, (too lazy la)> Anyway... i hav read the yi ren\'s horoscope thingy... \"Scorpio are passionate ppl, they hope for a long lasting love... haha... gald tt im a scorpio... this really say out wat i wan to say...\" (but scorpio are vengeful ppl... erm... there is no vengeance in my dictionary... so you all relax...) ^^   '),(6301,3674747,'07,August,2004','2004-08-07 00:00:00','   >>>>  Think back to the day when you first laid eyes on her, you found yourself charmed by the way she talks, the way she dresses herself to show off her best features and the way she embraces life with her laughter. It seemed as if a mysterious \"chemistry effect\" has suddenly developed to draw you closer to her.       You two then began to meet regularly, and you discover more things that you admire about her, her clever ideas, her healthy values and the way she stands up for you when others doubt you, you find yourself thinking of her not just as a normal friend, but a very good friend.       It is often during this period that a boy and a gal will start thinking of bringing their friendship to another level. After all, the kind of wonderful experience you have between each other can only become even better if it develops into uniqueâ€¦ no one else seems able to replace her in your heartâ€¦ so both of you agree to go steady and work even harder on the relationship, you guys \"graduate\" to become a couple, and are the envy of the sea of singles.       Howeverâ€¦ at some point in your relationship, you forgot how it all began. You start to take your partner for grantedâ€¦ why can\'t she laugh in a more ladylike manner? Why doesn\'t she dress herself more trendily? Why must she assert her views and point out your silly mistakes? Is she really the one for you?       To be fair to yourself and to her, take some time to reflect on your \"love memory\". The \"love memory\' contains all the reasons that you fell in love with her right from Day One. It contains rarely accessed snippets of how your life has changed since meeting and loving her.       Pre-steady days, did you heap compliments on her for the brave way she spoke her mind on bullies and snobs? You probably did. Did you like her unusual fashion sense that makes her stand out from the crowd?       Right-o. So why are you criticizing or finding fault with her now that she is your girlfriend? Because you have forgotten the fundamentals of loveâ€¦ like so many of us. You have forgotten the reasons you admired her during the friendship phaseâ€¦ insteadâ€¦ once you went steadyâ€¦ you put your \"love memory\" in cold storageâ€¦       As her boyfriend, you take up a new set of demands and expectations about her. These new ideas are not necessarily better; they could put your relationship at riskâ€¦       A healthy relationship, like learning to walk properly, follows a step-by-step development; you can\'t possibly become part of a couple if you aren\'t friend in the first place. Ok, I know some of us break the rule and plunge straight into whirlwind courtship, but how many of those couples can go the distance? Couple hood works well when there is something you like about her (and vice versa), and Iâ€™m not referring to merely the physical aspectsâ€¦ dun think too hard huh!       So whenever your relationship hits a rocky path, don\'t give up without checking on your \"love memory\"â€¦ rediscover the reasons why you fell in love with her, and watch your relationship flourish with a newfound vigor.       Love is to listen to the needs of the other party: giving without asking for returnsâ€¦ overlook the faults and find the plus point of the other partyâ€¦ voice out &amp; say sweet things and express your happiness and words of encouragement for the other party, just like me giving kisses to the person I likeâ€¦ kekekeeâ€¦ putting in effort, keep putting in more time to find the world of loveâ€¦   Give comments....^^   '),(6302,3699086,'25,June,2004','2004-06-25 00:00:00','   I would just like to share with everyone the supreme wonder, fun, and excitement of paintball. Deifinitely one of the greatest sports ever created. It\'s fun, you can do it with your friends, it\'s a great adrenaline rush, and YOU GET TO SHOOT PEOPLE.....LEGALLY!!!:p What could be cooler? You should go ty it. Really.  '),(6303,3699086,'22,June,2004','2004-06-22 00:00:00','   I\'ve got a really bum deal with two of my friends. I haven\'t seen one for awhile,  and we\'re really close. A \"thing\", I guess you could say. Well, there\'s this other person here who I\'m into also. They both are into me, so...........HELP! Advice would be appreciated.  '),(6304,3699086,'22,June,2004','2004-06-22 00:00:00','   Alot of people around here have been bugging me to hear this thing that I wrote. I don\'t know if it\'s supposed to be a song or what, but it\'s kinda cool, i guess. If you want to, go ahead and tell me what you think:  I need one thing to make this right,  To stop this never ending fight,   Between what I feel here in my heart,   And what my mind wants to impart.  I only need on little thing, but a wall stands in my way,  What I need I cannot have, not now or any day. I\'ve tried so long to grasp it, but it still exceeds my reach,  If only it were closer, then they wouldn\'t have to preach,  That \"It\'s not worth it, you\'ll never win,  All that\'s left is to quit. Give in.\" But I won\'t stop until it\'s mine,  This one last thing I have to find,  Because just one more thing remains,  To rid myself of depression\'s stains.  But they do not see that I need them now,  More than I ever have. And how? How could I ever make them see,  Make them once again believe,  That what I feel is true, no doubt, but then, why bother to try? Would it not be easier to hide my face, to turn away and cry? Why do I try, why do I strain,   For what I know I can\'t obtain? Why must I sacrifice myself this way,  When what I need just runs away? Perhaps it\'s true, what they all said, That all that\'s left is to turn my head Away from this foolhardy quest? To just take the whole thing off my chest? And now I see, they\'re right of course, and have been all along,  I don\'t need this grief, this torture,  or to write this unsung song.  I\'ll stop this foolish journey now, because what I finally see,   Is I don\'t need this childish feeling to be who I am to be.  I do not need this uncatchable dream,  Although my heart wanted to fool me, it seems,  Into believing that if I did not win, that I would lose a part That still has not been touched in me, deep inside my heart.      Well there you have it. There\'s another four lines at the end, but I have them all scratched out so I can\'t read them. But anyway......ya, hope you enjoyed it.  '),(6305,3699086,'06,August,2004','2004-08-06 00:00:00','   I am a moron. I just admitted my feelings to her. Every time I do this, it always changes things. Even if they say it won\'t, it will. So maybe it\'s me. Maybe I\'m the one who changes. You know, sometimes I just think, \'Is anything ever going to come easy for me? Is anything meant to be for me?\' But apparently not. It looks like I\'m just gonna have to figure everything out and work for everything I get on my own. Not that it hasn\'t always been this way. But with everything that\'s going on right now, I really notice it alot more. And it seems alot worse. I wish I could change my circumstances. I wish I could have my life laid out before, and all I\'d have to do is just follow the directions. Unfortunately, things don\'t come so easy. But maybe it\'s a good thing. I\'ll probably value whatever I gain or lose a lot more because I\'m working so hard for it. If I didn\'t,  I\'d probably feel almost cheated when I reached the end. But that still doesn\'t make it any easier.  '),(6306,3699086,'05,August,2004','2004-08-05 00:00:00','   Well, my  life sucks right now. I\'m thinking of quitting the football team. I can easily be replaced, as I was shown at camp. It would actually work out pretty well, because then I could focus on cross country. Someone I thought I could get into a relationship with is now a no-go. A girl has a huuge crush on me, and I can\'t let her know that I know or that I return those feelings. My parents are really riding me, I just had a poem published, I\'m getting academic acheivment letters from all over the place, and if I make choices and change things, people get hurt. Not including myself. That\'s the worst part. Is I\'m sacrificing everyone else\'s feelings, and not mine. Whether I make choices or not, someone get\'s the short end of the straw. But never me. I never get shafted. I can only shaft other people. Which sucks. I should just quit it all. Everything. Just screw the world.  '),(6307,4152098,'13,August,2004','2004-08-13 00:00:00','   No not the ones on four wheels, even though there are a lot of those. Drivers for computer hardware. Had to setup a wireless network for one of my Aunties. One look at the equipment and alarms started to go off. Linksys router.......and D-Link wireless card (-.-);; I\'ve had plenty of experience with Linksys routers so no problems there. The D-Link DWL-G510 though, started to give me chills.  Install went smoothly enough. Drivers installed perfectly fine, WinXP and Zone Alarm said they picked up a new network, I was even able to get on the \'net. Then I thought I\'d reboot to make sure everything was dandy.........  BAM, S.O.B. couldn\'t pick up the router. Tried to adjust settings on the wireless configuration tool from D-Link, P.O.S., then some settings on the router.....nothing. So I uninstalled the drivers and reinstalled them, started picking up the router again....I already knew where this was going. Reboot, GONE.  So then next logoical step was GOOGLE. Thank goodness for google. Found a forum post on Broadbandreports.com, suggested installing drivers from ASUS\'s wireless card (apparently they share the same wireless chipset), worked like a charm. In closing......  DIE D-LINK DIE '),(6308,4152098,'12,August,2004','2004-08-12 00:00:00','   IT SUCKS...kind of...  More ranting to follow.... '),(6309,4152098,'12,August,2004','2004-08-12 00:00:00','   Sell you stupid Barbarian Sword..............  Oh, hey there. Gotta sell off all my FFXI equipment, so that I can hook up all my LinkShell homies with some Gil. Hopefully everyone will get about 20k\'s worth.  THEN I CAN CANCEL THE DAMN ACCOUNT AND BE FREE. '),(6310,4152098,'12,August,2004','2004-08-12 00:00:00','   I don\'t know what it is these past few months, I can\'t get a good night sleep (-.-).......sigh. Must be the damn heat. I just toss and turn \'til the sun comes up. I can\'t take it anymore, bring on Winter already.....ARGH!!!!!!!!!!  I\'ve been drinking those large cold silky smooth Starbucks Frappuccino in a bottle to stay awake at work. I love my job, but man without that creamy mocha juice, I can\'t make it thru the day.  I DON\'T EVEN LIKE COFFEE FOR GOD\'S SAKE.  p.s. I guess staying up \'til 1-2am in the past 9 months didn\'t help either.  DAMN YOU FINAL FANTASY XI, I QUIT. '),(6311,4152098,'12,August,2004','2004-08-12 00:00:00','   Tired of not getting told straight up if a game sucks? Tired of the game magazines that say they aren\'t influenced by game publishers, yet half of their magazines are filled ads? Tired of celebraties jumping on the video game bandwagon when they know as much about games, as Jessica Simpson knows about tuna?  Well my fellow gaming buds, you\'ve come to the right place!  LET THE RANTING BEGIN! '),(6312,3443396,'03,August,2004','2004-08-03 00:00:00','    Well. This will be my first blog ever. As I create this, I am eating a dinner of salsa and stale tortilla chips. I wish I had some cheese.  There is one week left until the Palm Beach County school system unleashes the force of approximately 176,335 minors upon the unsuspecting populus. Some lucky folks will not spend their summer in this sauna of a state. I, being just that lucky, will be among them.  School is almost over, which means work is beginning; that is, if I can get a job.    Here is some prose I wrote a few years ago. I hope that it is enjoyed.   --------------------------------------------------------   You could fit a dead body in there, you know. Two dead bodies, if you pack them right. Just right for bad Mr. Hesson and Jameson.   I donâ€™t think like that! I never think like that.  Of course not, but they killed your mother. You know they did. And if she isnâ€™t dead already, then she will be soon. Go on. Your fishing knife is in the drawer. Do it. â€œNo.â€ I said. Hessonâ€™s gaze turned to me suddenly. â€œNo what?â€ he asked. My breath caught in my throat, suspending my answer. â€œNo socks. My mom forgot to wash them.â€ I turned and sifted through a pile of clothes.  â€œHurry up.â€ Hesson said. He was young, but his suit gave him years beyond his own. I wondered if he had a family.    Do it! Itâ€™s in your blood.  Kill! Then put him in the suitcase.Remember your father?   Daddy? Yes I remember. I love him.    Do you?   Yes, Iâ€™m just like him.    Then kill if youâ€™re just like him.    No! Iâ€™m not bad.   No, you are not bad. They are bad. They are evil. You love your mommy and  daddy?   Yes.   Kill.   Okay.  â€œ I have to get some underwear out.â€ It was hard to look at his face as I spoke. I did not want to see him. Hesson had kind eyes. It hurt me to think that a man with such kind eyes could kill my mother. â€œOkay. So get it.â€ He said blankly.  â€œWell...â€ My silence blossomed a pink, adolescent embarrassment in my cheeks.  â€œFine.â€ He said, turning. Grateful for the relief from his stare, I opened my top drawer and felt underneath my clothes. I felt around, farther back. My fingers grasped at the cold metal handle and my palm, grateful for itâ€™s cool, slick surface, embraced it.  I flipped it open and slowly turned to face him. His back was still turned. He did not see the knife in my hand. He bobbed his head and hummed to himself. Some song, some song he knew. Some song I, undoubtedly, knew as well.  Kill him. Kill him now!   I donâ€™t want to kill him!   Fine! You donâ€™t have to kill him. Just hurt him. Plunge it in.   Where?   Towards the left side under the shoulder. Itâ€™s all right. Just do it.   But isnâ€™t that-   Do it!    All right.  I raised the knife, ready to counter bone with all the velocity my young hand could gain. Closing my eyes, I brought the knife down. His skin gave in. He had seemed so strong, but now was all too mortal. How softly and smoothly the knife entered his skin, without a noise. It was poetry in motion. A dark, beautiful poetry. He didnâ€™t even let out a gasp. He turned to face me. The look in his eyes was dumb, pleading. His mouth seemed to gulp at the air. His tongue writhed in his mouth and, in his throat, there was a deep, unintelligible noise, a humming. I closed my fist tighter and tighter until I could feel my own blood trickle out from beneath my fingernails. The first man I ever killed was, by far, the most remarkable; his face and eyes were full of poignancy and remorse, his hands reaching in a picture of perfect desperation and humanity. As he fell backwards onto my dresser, the knife plunged deeper into his body. The tip of the blade shone on the other side of his chest. It was small, the knife that killed Hesson. A little silver secret, just between us.   '),(6313,3371255,'17,May,2004','2004-05-17 00:00:00','   I\'m White...and I\'ve been the victim of rascism...hmmm There are these black kids at my school...yeah, they think they are better than all white people...rascist, hypocritical dipshits... This is sad. I know I never owned a slave...and I\'m pretty sure that they were never slaves(if they were...they\'re pretty old for Freshmen...), so their argument MUST be based on  anscestrial heritage... Let ALL know, even before the civil war, only 10% of the white people owned slaves...and none after the war. My family imigrated durring the 50\'s...and by their suposed logic, this statement must make perfect sence. I am part Jewish. I am also part German. Does this mean that my anscestors were Nazis, and treated my anscestors horribly, and  as such I should hate myself for my anscestors treatment of...themselves? No, this makes no sence, and I find that these...honestly I can\'t think of a word to call them... are just trying to get an advantage rom their fore-father\'s strife. And as we all know, the Africans wouldn\'t even think about selling their brethren into slavery... the Europeans just swept over and conquored the place... I hate people sometimes ... they can be stupid, and naustalgic. And another thing, why is it considered rascist for me to say the \"N\" word, but they can call me a \"Cracker\" all they want? GOD!!! It just doesn\'t make ANY SENCE!!! You know, there was a guy who was walking in the Bronx at night with his friend, when two black guys woth knives started walking towards him. His buddy, the smart one, ran. He, however, felt it would be rascist if he ran. His body was found in a dumpster, behind a bar. Funny...  '),(6314,3371255,'17,May,2004','2004-05-17 00:00:00','   I find it Ironic that our society is willing to accept blame for the ills of the world...and from those who poison it! If you have ever been on the west coast of the US, you will find that many of the people actually SIDE WITH THE TERRORISTS! WHAT THE HELL ARE YOU THINKING?????? If I came up to someone, and punched them in the face, would they blame themselves? NO! They would punch me right back! You know there are SERIOUS problems with a society when they feel that they are to blame for EVERY LITTLE THING! 2 1/2 years ago, we had a 2 planes fly into the World Trade Center...killing 3000+ people, and now, all we can think about is a few \"Tortured\" Iraqi soldiers? As I recall, those who died in the Trade Center were not exactly beligerant...they were TRADING. but let\'s not even mention the men and women in the towers, crying out in vain as fire consumes them, or those who jumped out from 80+ stories onto the pavement to escape such a fate... let\'s talk about 3 ENEMY SOLDIERS who were \"discrased\"...and ofcourse they were never planning to bomb our military bases in Iraq, never ever ever ever trying to KILL our troops...My thinking is, if you are going to attempt to attack those who are liberating you, then you had better be ready to face the consequences. Let my rant be a lesson to society... for many of the ills in the world, we are not to blame. We are the ones who prevent the ills of the world...or at the least TRY to.           *cough*France,Germany*cough* Grrr...  '),(6315,3371255,'17,May,2004','2004-05-17 00:00:00','   And so it has begun...I shall soon enough begin to post my problems with society(and there are MANY), and I believe that we shall find we have a jolly good time reading this...and laughing at the idiosincrocies of the world. I bid you all a good night...though I will indeed post again tonight...in about 2 min.  '),(6316,1980443,'30,September,2003','2003-09-30 00:00:00','    Well, my happiest childhood memory has got to be baking muffins and biscuits in my little E-Z Make oven at my grandma\'s house.  Now, I was only 5 at the time so I was like, \"Gwandmaw, can I make some biscuits pwease?\"  And she\'d be like, \"Yeah sure William Alexander Teat!\"  So she\'d make some dough and I\'d play with it, put a little extra flour in it cause I liked the taste of flour, and \"would shoot em in da\' pan\".  After they were done, I\'d eat them.  \"Wow gwandmaw, I did vewy good dis time wight?\" I would say.  \"Yes William ALexander Teat, you did\".  I loved making those biscuits - they were so good!  Even the dough was good!  Everything was good!  I love my childhood, I love you, also.   That\'s how we do things round herrrrre.  I\'m Will - see ya\'ll later.   '),(6317,1980443,'26,September,2003','2003-09-26 00:00:00','    The Room Suddenly Became Quiet when Mr. Harmless said, \"He hates Jesus\".   It all start that day when we left to go to the fall retreat.  We left at about 8:00 am or however early (too early actually).  I was excited to go cause I wanted to see all of my friends from the preceding year.  Well the retreat was going great, we had lunch, chapels, gametimes, freetimes and all sorts of stuff.  The day grew old, and there were gnats everywhere.  Oh man, I hated those gnats.  They were like really small little brothers, who kept trying to get closer and closer and closer to you until you just smacked them or something, then they\'d always come back.  Gnats are like demerits, they\'re everywhere and they don\'t really serve a purpose.  (Just kidding by the way) Anyway, the gnat thing is another story, back to the silenced room.     It was time for the \"Bonfire\" inside of the building.  Mr. Harmless was giving a testimony when he told us about this witnessing oppurtunity the Lord placed in his life.  By the way, I loved all the testimonies, it\'s a great time as a school to grow closer.  Anyway, he\'s a funny guy right? So, he told us that this guy didn\'t have a single tooth in his whole head.  Everyone started to laugh, then right as soon as we started to laugh, he  said that the man hates Jesus Christ.  The room was silent except for a laugh by Christopher Thomas.  It was a good silence though, because it showed that we were concerned about the man.  How cold and unrevealed he was to the truth about Christ.  But the testimony went on and that was the most important thing.  That\'s about it, there\'s been other silencing times, but I couldn\'t think of one.  Love ya Anderson. Ms. Wiebe, that\'s (320 Words).  Oh yeah. That\'s how we do things round herrrrre.  I\'m Will - see ya\'ll later.   '),(6318,1980443,'25,September,2003','2003-09-25 00:00:00','    There is a Soccer team with lots of guys, Coach John\'s, in practice tries to make us die; We dont\' care we still play well, After the games we kinda\' smell; We all have names like Hecter and Amberto, Adam\'s is Rodriguez but I\'m not quite sure though; We will beat Martinsville by 4, Let\'s hope for once that I can get a score; Yeah that\'s right I can never put one in, Even though I don\'t we usually win; Taco Bell here I come, I want something to eat-o, This is Hector, c-ya later, I want a Burrito.    I\'ll try to get a new poem ever week. That\'s how we do things round herrrrre.  I\'m Will - see ya\'ll later.   '),(6319,1980443,'25,September,2003','2003-09-25 00:00:00','    I don\'t have one.  Thank you Just kidding - my imaginary friend has many great qualities.  Even though I can\'t always see him, I know he\'s there watching me.  He has blondish brown hair, greenish-blue eyes, and a very hairy dad.  I love it when he does the robot, and we celebrate goals together.  We always pass before the games, and he got mad at me one time because I didn\'t talk to him after Miranatha.  I had to apologize to his mom and hug his dad. I was sad.  Anyway, now we\'re good friends, even though sometimes we don\'t get along.  I can never really think of one time where we didn\'t get along, but If I do I\'ll put it on here.  All of this stuff is a great friend huh?  Well guess what, this person is Adam Johnson!  Love ya Anderson.   That\'s how we do things round herrrrre.  I\'m Will - see ya\'ll later.   '),(6320,1980443,'24,September,2003','2003-09-24 00:00:00','    My Greatest Fear, huh? Hmm, well I would have to say that first of all I have a ton of fears!  I always sleep with my light on, but I don\'t really fear the dark, I just always sleep with my light on.  Plus I hate it when those hairy spiders bite you, they look nasty and well, I don\'t like anything hairy and nasty crawling on me.  Those are about the only ones I can think of right now - but trust me, there are way more. I Never feared a lot of things like heights, fire, or storms and stuff like that.  So, what exactly is my GREATEST FEAR?  My greatest fear would definitely getting to the judgment seat of Christ and not being written in the book of life!  I know that isn\'t going to happen, but it would be horrific if it were.  I had a dream about that one time, actually.  Anyway, fears are not always bad for you!  There is definitely I good side of fear.  If I didn\'t fear not being written in the book of life - I Wouldn\'t be written in it - cause I just wouldn\'t care!  For real, if I weren\'t afraid of those nasty spiders, I\'d probably have some contagious disease with bumps all over my face.  That would be bad.  So, use your fears for good, and if necessary you can always conquer them by facing them.  Alrighty then, that\'s the end of this journal! And....  That\'s how we do things round herrrrre.  I\'m Will - see ya\'ll later.   '),(6321,1980443,'24,September,2003','2003-09-24 00:00:00','    Ah Yes, the little things! What more can I say about them?  They\'re little.  Simple meaningless parts of life that don\'t ever matter do they?  YEAH RIGHT!  I love the little things.  Everything in and of itself is a little thing.  Anything I do with my friends is a little thing - but I love it, and would never trade it for anything, not even a big thing.  I would like to mention just 10 of my favorite little things: - Calling a friend of mine every night ; ) - Playing soccer with Amberto, Hector (myself), Rodriguez, Sanchez, Moses, Victor X, Chiquita, y Maria who play Friday, and we\'ll beat Martinsville. - Getting an F on my Physics test - Trying to get a job for Saturday at the Purdue game but they didn\'t want me : ( - My cat dying - Nothing else really, I live a boring life. Cause it\'s the little things that make you happy, and without them my life would definately be like a college course class on greek - BORING.    That\'s how we do things round herrrrre.  I\'m Will - see ya\'ll later.   '),(6322,1980443,'24,September,2003','2003-09-24 00:00:00','    You wanna know how Bojangles Biscuits is doing?  It\'s my fantasy football team competing against people such as: Robbie Dupre, Jon Cline, David Cole, Coach Grimes, Jordan Delks and like 6 others.  You can see how I\'m doing at this link right here.  http://exit42b.fanball.com/ff.x42?prg=team&action=select&league=F000000759&team=B&61308  Thanks!   '),(6323,1980443,'24,September,2003','2003-09-24 00:00:00','    Alright, Let\'s hope that this blog works for once!  All my other ones, yeah they didn\'t work!  Thanks for looking at this one though.  This is Hector, Hector Sanchez.   I\'m gonna\' go eat a burrito - Hector   '),(6324,1980443,'21,October,2003','2003-10-21 00:00:00','    Oh My GosH, My HaiR iS iN ThA HouSe! And no, it doesn\'t look like hair on a mouse; It\'s Black, kinda\' brown, and I use this gel, It\'s called crew comb, and has a different smell; I wake up, take a shower, and do my hair, What\'s that underthere.....underwhere? Haha I made you say underwear, cat! Get back or I will attack, and you don\'t want that! Some people think my hair is spiked, I tell them, yeah so what..get a life; That may sound mean but make sure you know, That hair spiked it NOT AN ACT OF REBELLION - YO! That\'s one of the things our school kinda zones, Instead of spiked style, we have to wear Bob Jones! Yeah Bob Jones, the true and eloquent style, The only thing that goes with it is a really cheezy smile : ) But anyway, I\'m not bitter I swear! It\'s just another rule that\'s well...just..there! Seriously y\'all, keep it real in da\' hood And If you want corn rows get your hair did good....   Thanks yous allz (applause) (the bongos bongo all night long)  That\'s how we do things round herrrrre.  I\'m Will - see ya\'ll later.   '),(6325,1980443,'15,October,2003','2003-10-15 00:00:00','    Ah yes, the importance of family in society!  I love it!  That is a most brilliant title I\'ve ever come up with.  Now to talk about it, the importance of it.  Hmmm, well I love my family and it\'s important to me.  I\'m also in society and it\'s also pretty important.  Putting these two together won\'t be too hard.  I\'ll sum it up in about one or two sentences.  If people would depend on their family more in society and have better relationships with them, than I\'m sure alot of different, bad things wouldn\'t happen.  More time with family means more care and love for those who need it.  Our society isn\'t always \"nice\" ya know?  Like I think of the Japenese people, and they have a tremendous reliance on family.  They\'re always really nice to everyone!  They share there proverbs, their rice, even give their daughters to marry someone before they\'re born!  Wow, we need more importance on family in our society.  Thanks yous  That\'s how we do things round herrrrre.  I\'m Will - see ya\'ll later.   '),(6326,1980443,'08,October,2003','2003-10-08 00:00:00','    Anger must be the energy that has not yet found its right channel. Wow, how true is this?  So many other energys are on their right channel.  All the energy like electricity, and light, they seem to be doing good.  But how often have I let my God-given anger, control me.  It definately has not found its right channel.  Anger is used for taking whatever it is you feel like doing, out on the other person.  Anger results in just about anything these days than something positive.  Killing, hurting, slandering, putting down, dividing, and many others are the effects of this energy at its worse.  Anger should be used for solving the problem, SOLVING THE PROBLEM.  Usually it just creates more problems, not even close to solving them.  I can learn from this quote by using my anger next time, to get the problem resolved, and glorifying God in the process!  Jesus did it, so can we.   That\'s how we do things round herrrrre.  I\'m Will - see ya\'ll later.   '),(6327,1980443,'07,October,2003','2003-10-07 00:00:00','    There\'s alot of characterisitcs that are included in a hero.  When someone, does something courageous, or extraordinary that you would never do, or just being a great example of how to live life the way we\'re suppose to.  With all that said, my heros are my parents and sister (sorry Squeeb, you don\'t do much except say WEEEEE).  My dad has always been the most unselfish person I know, and he does it in a way in which he doesn\'t tell me - just the way he acts.  I know so many different things he does different to make his family better.  My mom is always using the Bible - ever since that counselor thing is goin\' on.  She shows me when i\'m wrong - how i\'m wrong - and what I need to do to change it.  And my sister, wow, so many times she\'s encouraged me, I love her so much.  Most nights, my favorite thing to do was to walk in to her room at like 12 or something, while she\'s doing her devotions, and talk to her about anything and everything.  She was the coolest, who always tried to make time for me, well except when she went and did things too cool for me, : ).  But the list doesn\'t stop there, there are so many more people who are my heros.   Jabolini, Jabolini Africa  That\'s how we do things \'round herrrrre.  I\'m Will - see ya\'ll later.   '),(6328,1980443,'02,October,2003','2003-10-02 00:00:00','    If I was a cartoon? Or if I were a cartoon? I\'m not that good at english, but i am in this group where we don\'t have to do homework, but that\'s beside the point.  I would definately be batman or something cool like that.  Man he kicks some major joker..... He fights all the dudes, and he\'s got cat girl for a sidekick, can\'t get much better than that.  Batman rocks the house until it falls down, he wears this black cape thing, and I am so jealous - Let\'s get Yogurt!    That\'s how we do things round herrrrre.  I\'m Will - see ya\'ll later.   '),(6329,1980443,'02,October,2003','2003-10-02 00:00:00','    My Sister, I miss my sister, she\'s away from here, I hope I get to see her sometime this year; Her name is Rachel, and her last name is Teat, She\'s really buff, but she has small feet; She plays volleyball and was ok at basketball, It turns out she\'d rather just shop at the mall : ); But that\'s ok she can shop till she drops, I went with her one time and i about flopped; \"rachel, i didn\'t know you were so fast\" \"c\'mon wilbo, you don\'t want to be last\" Ok, fine I ran around the place, I look of excitement on my sister\'s face; I knew my sister loved shopping, She has never shown a sign of stopping; Right now she talk to me on the internet, This weekend I\'ll talk to her in person, better yet! See you soon, Rach!  That\'s how we do things round herrrrre.  I\'m Will - see ya\'ll later.   '),(6330,1980443,'20,November,2003','2003-11-20 00:00:00','    GO SIGN MY NEWLY ACQUIRED MESSAGE BOARD! THANKS!  The Bongos will BonGO all night long.... Keep it real, I\'m Will   '),(6331,1980443,'27,January,2004','2004-01-27 00:00:00','    My Holiday in January, Well I\'ll start off with letting you know that every single HOLIDAY is a NOSCHOOLDAY as well.  So my holiday, of course, wouldn\'t have school.  It\'d be a Tuesday so you wouldn\'t have to have homework 2 days in a row.  (sorry people who read my blog but this extra credit stuff is for extra credit, not entertainment).  The name of the Holiday is Celebrate your Date DAY!?!?!  Yeah, you\'d go on a date.  That would be nice, except a whole day devoted to a date would kinda\' get lame.  So you\'d celebrate you date by going to skate and then open the gate to find another girl named Kate.  You\'re girlfriend gets mad, kate gets mad, the holiday is ruined, and this is the end of my blog. The Bongos will BonGO all night long.... Keep it real, I\'m Will   '),(6332,1980443,'25,January,2004','2004-01-25 00:00:00','    Well about a couple of days ago, I started to read this new book called - The Purpose Driven Life.  After reading the first two chapters I can honestly say this will definately be a life changer.  I encourage you to get the book, ha, and I haven\'t even finished it yet.  Thank God for a sunday school class who makes me buy these kind of things too, ya know.  Also, with Smite, we have to fill out the journal for each chapter.  After it is all said and done, I can look back at the progress I\'ve made.  The very first sentence in the book is - It\'s not about you.  How true that was to me, I needed to hear that really bad! Let me know you\'re thoughts on my tagboard, let me know what you think of the book.  Thanks  The Bongos will BonGO all night long.... Keep it real, I\'m Will   '),(6333,1980443,'20,January,2004','2004-01-20 00:00:00','    Ok guys, this is going to sound really stupid.  But look at the Title of the blog entry, it\'s kinda\' hard to make this kind of thing exciting. BUY THE FLAG  Buy purchasing this american flag located at the top of every teachers white board, you will certainly recieve more than you are paying for!  At a measely 5 dollars, how could you pass this sort of thing up.  But wait, buy now and you\'ll not get just 1 american flag to go on your whiteboard, you will also recieve 2!  A total value of 10 dollars, all yours for just the price of 5.  Show your patriotism with this beautiful display of american history.  Red, White, and blue matches all sorts of colorful, uh, white boards.  You get the picture, white goes with white.  Do what so many other teachers and ordinary people are doing now and call the number today.  1-800-BUY-FLAG Our flag is our American way of life, make it yours! Must be 18 years of age or older to order.  Cihldren, ask your parents before calling.  So that\'s it, sorry for wasting your time :) (but you read it) The Bongos will BonGO all night long.... Keep it real, I\'m Will   '),(6334,1980443,'13,January,2004','2004-01-13 00:00:00','    I\'ll tell you what happened,    Like the event that happened, not the name of it, but if you know anything about me/my school you\'ll understand.  Day was January 6, and I was so pumped, getting a chance to do something like this!  I thought it\'d be great, the sweetest time of my life.  So as the game began, the nerves were kinda\' getting to me, and I could hardly dribble properly, the floor was hard to get used to but that was no excuse for my play.  And I\'m not talking about my play only, our team was horrible, they know that, I know that, but we just have to live with it now.  Despite an embarrassment at halftime with ourselves and our coach, we still kept playing like Kindergardners play tag, if that made any sense to you. (running around without a purpose is what i\'m trying to say).  The final wistle sound, I CAN STILL HEAR it.  The thought of defeat, especailly to this team (I won\'t mention the reasons) was just unbearable.  It was sad guys!  I tell you, it\'s right up there to the defeat of Heritage Hall when I played for JV 3 years ago.  That\'s how bad it was.    Even when he kissed her, he was lookin\' over starin at her sister (what a jerk huh?) .... Keep it real, I\'m Will   '),(6335,1980443,'09,January,2004','2004-01-09 00:00:00','    Well, I just deleted the whole last Blog I wrote!  Shows how smart I am : ) Anyway, this last week was great I loved Mo G (Morris Gleiser).  If I was ever to become a preacher/pastor I would be sure to speak like he does.  Almost the perfect, balance of humor with message.  I got so many things from the messages this week.  The fact that the prodical son\'s brother was worse than he was is probably the first time i\'ve heard that.  In many ways of my life I was letting little sins that didn\'t in fact seem real to me continue.  They were minimized in my thought, and the Holy Spirit helped me to see that.  To show how much complaining I had been doing,  I caught myself (complaining) 5 times the rest of that actual day!  Spiritual Emphasis week is great guys!  There should be at least 3 of these each year.  It was a great time to come together as a school (which I love so much), and become closer to Christ.  I truly mean that.  The Bongos will BonGO all night long.... Keep it real, I\'m Will   '),(6336,1980443,'23,February,2004','2004-02-23 00:00:00','    Paper or Plastic Ms Wiebe told me I couldn\'t do this really funny commercial that I saw that had to do with paper or plastic, sorry to ruin all you guys fun out there.  Basically the guys needed toilet paper but they bought (alchoholic beverages :) ) instead, lol, when it gets check at the register the lady asks, \'paper or plastic\' of course the boys reply in pain - PAPER! hahaha.   i\'m gonna\' have to go with paper also, you can use it for book covers, and it doesn\'t tear as easily as plastic.  remember this is for extra credit so, you\'ll have to bear with it again. c-ya The Bongos will BonGO all night long.... Keep it real, I\'m Will   '),(6337,1980443,'22,February,2004','2004-02-22 00:00:00','    Everyone,    I have the best sister in the world.  This might sound stupid coming from me, but I love my sister so much.  She came all the way from Cedarville (a good 3 1/2 hour drive or more) to see me play my last home game this weekend.  It was a huge suprise, I didn\'t even see her until I got done w/part of warm-ups.  She rocks, and she knows it.   She laughs at all my dumb jokes, and always takes my side w/the parental feuds.  I love you, rach.  Keep it real, I\'m Will   '),(6338,1980443,'21,February,2004','2004-02-21 00:00:00','    Look, if you have a blog, sign my tagboard and tell me so I can put your blog on my \"links\".  Also, tell me your Instant message name cause I need some, i only have like 25 or something, and i know there are more students that use the AIM service.  Thanks The Bongos will BonGO all night long.... Keep it real, I\'m Will   '),(6339,1980443,'18,February,2004','2004-02-18 00:00:00','    Well lots of funny things happen. I mean, especially on the way to places.  I remember going to a state tournament game with my trusted friend Adam.  Something that I did because I couldn\'t help it cause Adam to throw something cause he couldn\'t help it.  It resulted in diminished playing time for both of us; but the memories were still there.  I can\'t specify what was exactly involved, just me and adam didn\'t start for like 2 minutes, and then he put us in, haha!  It was sworth it let me tell you.  I miss you, Clinton.  Matematicas...  The Bongos will BonGO all night long.... Keep it real, I\'m Will   '),(6340,1980443,'31,March,2004','2004-03-31 00:00:00','    My Favorite Experience While on Break   Well probably my favorite experience will have to be the whole FMT tour, but probably not the traveling or singing!  It was watching \"Manichan\" with Diablo Dos.  This really kinda\' weird dude has us at his house for the couple days we were in New York.  We watched some Christian Music Videos, lol, Really, and then we watched this really disturbing OLD OLD movie about a guy who falls in love with a manichan!  Kinda\' retarded and all, but I will never forget that movie!  There was this one guy named \"HOLLYWOOD!\" he was pretty feminine, and snapped his fingers alot, he kinda\' reminded me of... well you fill in the blanks!  Later Players! The Bongos will BonGO all night long.... Keep it real, I\'m Will   '),(6341,1980443,'25,March,2004','2004-03-25 00:00:00','    Ah yes the FMT tour,   Wow where do I begin.  It was so much fun/work/fun!  Really, i\'ve never had so much fun serving God and being served by God\'s people also.  The two things I really learned on the trip was that God works through people everywhere (I already knew that, It was just really emphasized, so..) and the second thing was that I\'m so thankful for our school!  Some of those kids there were really really jerks!  I mean horrible!  It was like they were bitter to us.  Out of the 7 different school, almost all 7 were pretty rude.  Our school rocks...  Of course, you\'re always thankful for the nerds at the schools cause they\'re the ones that come up and talk to us.    There were a few of those at each school - so everyone who reads this - be nice to guests!  It really helps. I mean, you don\'t have to always reach out to everyone especially in the Christian Lee case.   It\'s really good to be home!    The Bongos will BonGO all night long.... Keep it real, I\'m Will   '),(6342,1980443,'17,March,2004','2004-03-17 00:00:00','    As most of you know,   FCS got a baseball team this year, whoohoo!  I\'ve played baseball all my life, ALL MY LIFE! It\'s definately probably my favorite sport.  God was so gracious to bless us with a team and a place to practice.  Being a professional baseball player has always been one of my dream professions.  TO get up every morning and play a sport, plus make G\'s of G\'s.  What a life huh? Of course only if God wanted me to.  I\'m definately thinking about playing college ball, there\'s still alot to learn.  It\'s always been a passion of mine, my dad really pushed me to love baseball and I\'m glad he did.  My little brother isn\'t quite the same as me but that\'s pointless information that you don\'t need to know.  Later Players The Bongos will BonGO all night long.... Keep it real, I\'m Will   '),(6343,1980443,'08,March,2004','2004-03-08 00:00:00','    That\'s the title for today\'s or should I say this week\'s blog.  What am I putting off? That\'s a good question.  Ms. Barbic, I\'m not sure if she means like a spiritual thing or like something physical like.... uh ... weight.  I\'m not putting off weight anymore so, that\'s taken care of.  I guess I\'m putting off laziness, my grades have suffered this year and it\'s all thanks to laziness.  I know that I procrastinate alot, but sometimes I procrastinate to take care of my procrastination!?! Isn\'t that hard to believe!  Sometimes with all this business the best thing I can do sometimes is to sit down and watch tv.  At least, that\'s what I do alot when I am just being lazy.  I know that this is something I need to work on and am.  Thanks homies.  This girl walked up to me the other day and was like uhhhhh, and I was like WHATEVER! Then this other girl walked up to me and was like uhhh, then I like yeah whatever! The Bongos will BonGO all night long.... Keep it real, I\'m Will   '),(6344,1980443,'02,March,2004','2004-03-02 00:00:00','    What is exactly in a name?   I\'m really not sure why Ms. Barbic had the guts to put this blog entry up.  I know that my name - William - means protector.  I also know that Emmanuel means God with Us.  So, I guess that\'s in a name.  I guess names exemplify who you are.  I mean, Muhammed Ali wouldn\'t be Muhammed Ali if he kept his name Cassius Clay, he\'d just be Cassius Clay.  There\'s also nicknames, like mine is Wilbo, and Jon Ananias\'s is Gumbee, and Larry Wayne Jones is Chipper.  I love nicknames - it\'s all in the name, baby.  If you have a nickname - please sign my comment list and tell me what it is.  Later Players.  The Bongos will BonGO all night long.... Keep it real, I\'m Will   '),(6345,1980443,'27,April,2004','2004-04-27 00:00:00','    Consistency,    My soul problem ever since the beginning!  I mean I know what I have to do, and I want to do it, but I just dont\' always do it!  You guys know what I\'m talking about it - Alot has to do with my motives, and alot has to do with my stupidity.  I\'ve been learning myself alot about consistency, if that made sense!  I\'m sure you all know about the events happening in my life lately, ALOT! I\'ve learned alot too and I\'m thankful for all that.  I\'ve just learned, now is time to get serious about this thing that doesn\'t really exist until you\'re a junior called life!  And guys, we are actually not as mature spiritually as the girls, that\'s sad, and I\'m probably the worst one!  Juniors, I hope we all know that we aren\'t doing the greatest job that we can at being leaders, and I\'m a true believer in that, let\'s help each other, isn\'t it about time we look to our futures, look to what God has in store for our lives instead of our little fun time during school.  That\'s what I\'ve learned.  I\'m not pointing fingers, or trying to preach, but if we want to have a great year, we need to get a hold of this!  I love my class so much, you all rock.    Help ourselves out - Wilbo   '),(6346,1980443,'25,April,2004','2004-04-25 00:00:00','    Hey I didn\'t mean that inlaw thing! ok later playa The Bongos will BonGO all night long.... Keep it real, I\'m Will   '),(6347,1980443,'15,April,2004','2004-04-15 00:00:00','    Should I make a list?   I don\'t know, maybe i\'ll just do it in paragraph form.  Well obviously I want to be married - or maybe not so obviously... Anyway, also I want to at least have started a family with my wife of course... Have a stable job that keeps a steady income..  Be routed in a local church, keep in touch with my sister, and stay away from the in-laws.  There\'s a whole lot of little things that goes on in all of those things I just mentioned, you can figure them out for yourself!  The Bongos will BonGO all night long.... Keep it real, I\'m Will   '),(6348,1980443,'06,April,2004','2004-04-06 00:00:00','    I had an old box, Strong as an ox, Dumb as a box of rocks, Muslims wear smocks....1 2 3 4 Let me bust a rhyme about a box name Weisel, His V12 truck\'s engine only ran on Diesel Grew up in a little town outside of St. Louis, Some little rock told him that he could do it; Now he\'s busting rhymes like no other, Gave alot of his credit to his mother, She said he\'d make it, the money he\'d take it, And all the lyrics are his, he spake it. Next time you see an old box, Rap and rhythm is the flow he talks.  The Bongos will BonGO all night long.... Keep it real, I\'m Will   '),(6349,1980443,'30,May,2004','2004-05-30 00:00:00','    Well here\'s the cool story,   I\'m on my way to meijer, just to test out my car in the rain cause i washed it with RainX so it would\'ve been cool to see, I know the tornadoes are going around and stuff.  So I\'m at meijer and get ready to leave when the sirens go off, and i\'m like oh crap they better not make me stay here. So a lady like saw me so I took off down the isle, then she\'s like SIR SIR!!!! I was like \"oh great\".  She made me go back into this one room where they do training with all the other meijer people, so I sat there for an hour!!  With Arvid Olsen I might add... good times.  The meijer guy comes in and he\'s like you can go home but we don\'t recommend you do, but i left anyway, did\'t think it was that big of a risk.  The cops wouldn\'t let me on 26, so I went through some subdivisions and got home, lol.  There was no one on the road so it was so freaky!  When I got home, I just looked outside of my garage at all the stormy stuff, with the moving trees and the lightning and thunder, and it was a huge reality check with the power of God.  He\'s soooooo powerful!  And this isn\'t even close to all the stuff he can do.  God is soooo huge, and he still cares about each of us!  That\'s wild, and that\'s love.  Just a thought to think about...  So that was a really fun experience, but I\'m really jealous of Jordan Brand cause she actually got to see it!  Ugh, that would\'ve been cool!Also I\'m made cause I didn\'t get to buy the new toothpaste I wanted, lol, oh well!  Summer\'s easy.... Keep it real, I\'m Will   '),(6350,1980443,'26,May,2004','2004-05-26 00:00:00','    Well guys,    Another school year is over - can you believe it?  I love it!  Sure the trig exam and the physics exam were harder than the english section (470 ouch!) of the SAT for me, but it\'s all good.  Summer\'s cool cause there\'s so much stuff you can do. Fun in the sun is coming up like mustard on a bun you can\'t stop it, no not none, so now i\'m done.  You can do fun things as much as you want, and you can also grow closer to God!  That would make it a truly beneficial summer wouldn\'t it?  I mean if this summer was going to mean anything at all in eternity, it would mean us growing closer to God. Keep that at the back of your mind, I know it\'s going to be really hard for me since i\'m so busy, but make this summer a great one!  Sum are hard but summer\'s easy.... Keep it real, I\'m Will   '),(6351,1980443,'11,June,2004','2004-06-11 00:00:00','    Hey Guys,   I just want to tell you all about this new book that Mr Grimes, Me, Robbie Dupre, Andrew Rolley, And BLOE Flora are doing for kinda\' a study every thursday morning.  It\'s called \"It\'s not about me\".  Now, if you know me, I have a tendency to think the opposite!  And I don\'t have a problem with admitting it, just a problem with thinking it and doing it, lol, anyway, it\'s really cool.  It starts off by expaining how we as people in this small small world compared to God\'s Son - ARE THE MOONS!  Jesus is the Sun.  Our whole purpose is to REFLECT, not make our own, light of the sun,  the application is very clear.  We simply need to do exemplify the love that Christ shows us.  Before we where the sun in our lives, but now it\'s Christ, everything revolves around him.  Keep that in mind!  It\'s incredible to think that the whole.... everything that will ever be is by him and for him.   summer\'s easy.... Keep it real, I\'m Will   '),(6352,1980443,'07,July,2004','2004-07-07 00:00:00','    Hey Guys,   I\'m sure most of you have had a chance to hear this song,  it\'s by Casting Crowns, and it\'s called Who Am I.  It is an amazing reminder to me.  It gives me so much hope and joy to know that God cares about us, and also tells us who we belong to, God comforts us and tells us that we are His.  But don\'t take it from me here\'s the lyrics.  Who am I, that the Lord of all the earth Would care to know my name Would care to feel my hurt Who am I, that the Bright and Morning Star Would choose to light the way For my ever wandering heart  Not because of who I am But because of what You\'ve done Not because of what I\'ve done But because of who You\'re   Chorus: I am a flower quickly fading Here today and gone tomorrow A wave tossed in the ocean A vapor in the wind Still You hear me when I\'m calling Lord, You catch me when I\'m falling And You\'ve told me who I am I am Yours, I am Yours  Who Am I, that the eyes that see my sin Would look on me with love and watch me rise again Who Am I, that the voice that calmed the sea Would call out through the rain And calm the storm in me  Pretty Amazing huh?  Keep it real, I\'m Will   '),(6353,1980443,'02,August,2004','2004-08-02 00:00:00','    urlLink    Here\'s us guys, Nathan, Me, Jared, Robbie.  Aren\'t we incredibly rediculously good looking?&nbsp; urlLink   '),(6354,1980443,'02,August,2004','2004-08-02 00:00:00','    urlLink    Booo Yeah, Me and Luke, lol.&nbsp; urlLink   '),(6355,1980443,'02,August,2004','2004-08-02 00:00:00','    urlLink    This was actually one of the good kids!&nbsp; urlLink   '),(6356,1980443,'02,August,2004','2004-08-02 00:00:00','    urlLink    There are 8 foot sharks underneath us just so you know!  Really!&nbsp; urlLink   '),(6357,1980443,'02,August,2004','2004-08-02 00:00:00','    urlLink    Here\'s the Group!&nbsp; urlLink   '),(6358,1980443,'02,August,2004','2004-08-02 00:00:00','    What\'s up Guys!!   It\'s been so long since I\'ve written on this electronic journal.  Well I really want to let you know the things I\'ve learned on SMITE.  You\'ll hear alot of these testimonies I\'m sure, but it\'s just really good to experience this kind of trip.  I won\'t go into detail about all the fun things we did, I will tell you we swam with sharks, I\'ll try to put a picture in, went to Atlantis (that\'s a really expensive hotel, that bro kwame had connections with), and we had a great time running their VBS where there was a fight every night, lol.  Anyway, spiritually: 1.  I need to constantly be in the word of God.  Those people knewtheir bible like the back of their hand, and they applied it.  So many times I read my devotions and forget them the next day, I need to be using my Bible as the tool that God has given me to fight against sin, and give glory to God. 2.  I need to be more vocal with my faith, especially with all my friends, I\'m a senior now and a lot of leadership comes with that, I think we senior guys and girls need to be \"sharpening\" one another just like proverbs says. 3.  Witnessing, How often I don\'t even think about it, if I really am excited about being a Christian, I should tell people the Great News.  Hey this guy who lived, died, for me!  Wouldn\'t that be enough motivation to tell someone? That\'s so important.   I hope you enjoy these pictures, if they load...  Summer\'s almost gone.... Keep it real, I\'m Will   '),(6359,3326628,'13,May,2004','2004-05-13 00:00:00','   Man, I love this show.  The writing is pretty tip top, but the actors and their deadpan humor make the show genius.  Hopefully the American version will do the Britich original justice.  Don\'t count on it with the censors we have to deal with.  '),(6360,3326628,'13,May,2004','2004-05-13 00:00:00','   I just listenend to the new Wilco album that is due out June 22.  It is simply an amazing listen.   Now, this is a band that plays music simply because they have no choice, they love doing it for themselves and their fans.  They release their albums online months before commercial release and still hundreds are lined up on the day of the album\'s release to buy it.  That\'s what you call trust.  '),(6361,3326628,'13,May,2004','2004-05-13 00:00:00','   ACNE!!!  I have had it since the onset of puberty, but very minimal and normal like any other teenager.  Last fall is when it took a turn for the worse.  I think it had to do with the fact that your body changes every 7 years, and my time of change had come - just not the change I had expected.  You think your acne will go away along with your teenage years, but no, my acne got worse, the kind that are deep, painful, and leave red marks for weeks.  I was clueless on how to deal with it.  It wasn\'t a simple matter of hygiene or diet, it was infection and pain - both physiological and physical.  I tried Proactiv, toothpaste, neosporin, everything over the counter for a couple months and finally went to the gynecologist to see if the birth control pill could rid me of this hormonally linked problem.  I started on that, but no changes in my skin had taken place.  I graduated college a couple months later, and went home in the middle of winter depressed, but with an appointment to a dermatologist at the suggestion of my mom.  The derm prescribed me a oral antibiotic and 2 topicals.  I saw some results after 6 weeks, no more deep and painful lesions, but the persistent acne was still there.  I was still was so hopeless and alone.  I really don\'t think people know what persistent acne sufferers go through, unless they have had it themselves.  It has nothing to do with vanity.  It goes so much deeper than that.  Anyways, I came across a website (I have become very proficient in this area) called  urlLink Acne.org  which basically suggests a Benzoyl peroxide regimen.  Well, after regular visits to the site and further frustrations with the doctor\'s medicines, I tried the regimen.  Let me just say, it worked within a week.  It was my miracle.  My face is about 99% clear.  I don\'t know why I hadn\'t tried it earlier.  All I know now is that in some sort of odd way, I am thankful for my experience with acne.  It made me even more humble and more aware of others in similar situations.  Every time I am out and see someone with bad acne, I want to shout out \"Acne.org\".  Suffering isn\'t overrated, but one can not do it forever without going mad.  '),(6362,3326628,'13,May,2004','2004-05-13 00:00:00','   but blood, toil, tears, and sweat.\"  Winston Churchill said this 64 years ago today, as he had just been elected new Prime Minister of England three days prior.  He also inherited a world war, further ensuring his fellow allies that victory would come \"at all costs, victory in spite of terror, victory however long and hard the road may be.\"  Yeah, I know it sounds familiar.  I\'m sure something similar has been said at numerous times throughout the history of wars.  It\'s the only rational way to go about such an event.  I doubt very much that  urlLink Abraham Lincoln  put a casualty maximum or time limit on the Civil War.  He kept fighting.  Victory ensued.  The country was reconstructed, slavery slowly abolished, but not before the 16th president of the United States was shot and killed, along with numerous others.  The first presidential assassination in American history had occurred and was thus the ultimate price to pay for victory.  Perhaps the dramatic reminder of mortality sealed his immortality.  Is death, suffering, and pain necessary for victory, for triumph, for peace?  Does one\'s life have to pass before our eyes before we get it?  I don\'t know.  Clouds rain upon the ground, and we end up with floods  and  the alleviation of famines.  There isn\'t a clear cut argument for or against war.   urlLink $50 billion , however, is nowhere near the cost of one casualty.  What do we have left to win when we have already lost so much.  '),(6363,3326628,'13,May,2004','2004-05-13 00:00:00','   It really isn\'t a question of whether or not to do so, but rather a question of where and on what.  I have blogged every day for the past 23 years in my thoughts, essays, and arguments.  It is now time to put it into a modern medium, a more colorful and legible form.  So here it is, a page dedicated to me, myself, and I, on what I love, even if it will never love me back.  '),(6364,3215937,'27,April,2004','2004-04-27 00:00:00','   I just learned about Blogging last night on the news.  I thought it was interesting. I wanted to do it too.  So, here goes.  I am not sure what to put in here.  Today has been a \"blah\" day.  I\'m ready to go home.  The sun is out and it\'s so nice outside.  It only stays this way for about a month and then it\'s boiling hot outside and you begin to sweat the second you walk out the door.  That\'s Texas for ya.  Or Ya\'ll.  I\'m an accountant, and no, I\'m not boring.  I just happened to do this, not planned or anything.  I do like my job and I\'ve been here for almost 7 years.  Can you tell that I get bored?  This is the perfect solution for that I guess.  I have three dogs and a part time cat.  My cat, Buddy Lee, has been coming and going lately. I think it\'s because it\'s Spring and he\'s not fixed.  I know, shame on me.  If he would stay home long enough, I\'d take him to get fixed.  I\'ve got two Pomeranians and a mut.  I love animals.  I would love to live on a farm someday.  Well, I\'m not sure how stupid I sound so I better stop now.  '),(6365,3215937,'03,May,2004','2004-05-03 00:00:00','   Jack Nicholson once asked \"What if this is as good as it gets?\" and I wonder the same thing.  It seems that we\'re always striving to be smarter, skinnier, funnier, or more popular among our peers.  Am I just going to die trying?  I wonder how much I do for myself and how much I do for others?  My therapist told me, after I unloaded on her about my marriage, that I\'m a victim.  I cringed when she said that.  I don\'t want to be a victim, but I guess I put myself here and if I don\'t do anything about it, then I am really just a glutton for punishment.  God, it\'s sad that I\'m too lazy or scared to make any advancements towards happiness.  I just take things day by day and nothing changes.  The only thing that makes me happy is my son.  My son also drives me up the wall sometimes, but I\'m sure that\'s normal for a toddler.  At least he doens\'t know any better which is more than I can say for my husband.    Well, I keep on trying.  I go to the gym, I try and eat right every now and then.  I lay awake at night and think of how great I looked right before I got pregnant.  I had lost about 30 pounds by running and eating right.  I looked so good.  I was wearing a size 4.  Those jeans, which I refuse to throw out are in my closet now.   I look at them and tell myself that I\'ll wear them again soon.  I wonder where my self control has gone?  I was so good, but now I see a cookie and I fall apart.  If I could just get myself back to where I was.  I was in the \"mode\".  Nothing could stop me.  Nothing could tempt me either.  Geez, I was so good.  It is harder now becuase I have a son and I can\'t just go out the door and run anymore.  I bought a running stroller but somehow it\'s broken and the wheels are flat.  Good excuse, huh?  That thing is a real pain in the ass anyways to run with.     Well, crap, duty calls.  Gotta work.  '),(6366,3215937,'03,May,2004','2004-05-03 00:00:00','   I\'m back.  I\'m also bored.  I have been feeling so numb lately.  I think it\'s the Effexor.  I just feel blah.  I don\'t really have any emotions but anger.  That\'s great for my husband, who I take everything out on.  Things are so great on the homestead.  I\'m trying to take things day by day though.  Some kid bit my son this week.  I know which one too.  A little brat with a shaved head.  That little shit.  He\'s like 6 years old and my son is 16 months.  Talk about a bully.  My poor little man.  If his Gamma knew, she would throw a fit.  Tristen, my son is so clingy right now.  I guess it just goes with his age.  He\'s a Momma\'s boy for sure.  That\'s okay, in a few years he won\'t want me anymore, he\'ll want his Daddy to play sports or other \"manly\" things.   Right now he literally hangs from my shirt while I\'m tyring to do stuff.  I\'m the only one who can rock him to bed and hold him most of the time.  He\'s so sweet though.  He\'s already a little dare-devil. He climbs on everything and gets up on the table.  Every one says he\'s \"all boy\".  Well, my stupid contacts are killing me so I\'m going to check my insurance to see if I\'m elligable to get some more.  Ciao.  '),(6367,3215937,'08,June,2004','2004-06-08 00:00:00','   Wow, I really need to blog more often.  It\'s been a long time.  Well, where to start?  The past two weeks have been very crazy for me.  My life has been flipped upside down and then back around again.  I found out just before Memorial Day that I was pregnant.  My husband and I were very shocked, and then exicted.  My husband swore that it will be a girl.  I thought so too.  I tried to hold off on telling everyone, but within a few days I had told everybody.  I went to the doctor and did all that fun lab work and so far, I was four weeks and everything looked okay.  Four days later I started bleeding.  This didn\'t happen with my first pregnancy so I freaked out.  I kept on bleeding and I called my OB but he was not there and my old OB was on call for him.  Yuck.  All he said was for me to lay down.  I did that.  I knew though, deep down, that it was over.  My husband told me to quit being so negative but it wasn\'t that at all.  A woman knows her body and I knew this wasn\'t right.  The next Tuesday my fears were confirmed.  A miscarriage.  I cried and I cried some more.  I felt that I was to blame.  I felt a relief as horrible as that sounds and then I felt guilt.  I felt a lot of things.  I didn\'t know that one out of every ten pregnancies ends up in miscarriage.  I\'m just glad mine happened so early.  So, me and my husband decided to try again in about 3 months.   At first we weren\'t sure we were ready for another baby, but after this has happened, I guess it makes you want it more.  I had already had my mind wondering about my little girl and what we would name her.  I thought about what I would decorate her room in and I even came across my old baby clothes and a dress that I wore home from the hospital when I was born.  Every day it gets a little better.  If I sit and think about it, I want to cry again but I know I\'ve got to get over it.  My body has been in shock basically.  All the pregnancy symptoms have gone now and that is another part that makes it hard.  I had actually started to \"feel\" pregnant.  I have been spending a lot of time with my son and I\'m so grateful that I have him and he\'s healthy.  I love him so much I can harldy stand it.  He\'s so beautiful and perfect to me.  I really am blessed to have him.     Work... work.  I feel so unappreciated.  I keep telling myself \"just wait until our reviews in August\"... I can\'t wait.  I have so much to say.  I am the fastest and most efficient person in this department and I want the title to back that up.  I was promoted to \"Senior\" accoutant a couple of years ago but my boss\'s boss shot that down.  Probably becuase they don\'t want to pay me more.  I\'ve been here for almost 7 years now and I think I deserve it.  I know that everyone is replaceable but I would like to think of myself as an asset to the company.  My husband resigned from his job.  Normally I would be so mad at him but his boss was a scum ball.  This guy owns the company and treats all his employee and customers like shit.  He does this bait and switch thing where he sells them a very expensive door and then installs a cheap one instead.  He takes people\'s money and doesn\'t order the product.  He\'s on credit hold with several vendors.. I knew my husband shouldn\'t stay there long.  So, now he\'s trying to get his own business going.  Precision Door and Trim.  I helped him pass out flyers door to door th is past weekend.  He\'s been getting calls and has a few jobs lined up.  I hope it works out for him, and us.  It\'s very scary but I see so many people with their own company\'s and they do very well.  Look at his scum bag boss, he\'s a crook and has a huge business.  Oh, and he also punched my husband when he came asking for his last paycheck.  We\'re suing him and taking him to court.  The police came out and seem to be on our side.  They said that this isn\'t the first time they have delt with Bobby.  They want him out of business and so do I.  Well, I better get some work done.  '),(6368,3215937,'12,July,2004','2004-07-12 00:00:00','   Happy Birthday to me.  I am 26 years old today- well, at 6:23pm I will be.  We had a get together for my birthday yesterday at my brother\'s house.  It felt surreal.  It felt like a dream, it always does when the attention is on me.  I woke up with a sore throat yesterday morning.  I went out with my girlfriends Saturday night and had some fun.  I hate being the fat girl though.  I did the Limbo contest and lost.  My friend Sarah with her big ole\' boobies almost won.  She just broke up with her boyfriend a few days ago.  She seemed fine to me though.  She\'s so skinny and cute, she will have no problem finding a transition guy or two.  So, well, I had my first orgasm in like 2 years ago the other night.  Talk about relief.  It wasn\'t a great one, but at this point I\'ll take what I can get.  My husband and I talked seriously about our marriage on Saturday morning and I cried and told him that I can\'t imagine my life being this way forever.  He said things can change, blah blah blah.  Whatever.  My friends called me today at work to say that I should call Shannon for a booty call.  Lord knows I\'ve played out that scenario a million times.  Right now, I guess I\'m just not in the mood.  I\'m not up to my full potential today.  I feel kind of sickly and it\'s Monday.  It\'s so weird to think about cheating on your spouse and not sense any guilt.  I feel like I deserve to.  That probably sounds horrible.  Like I told Jana, Shannon doesn\'t count as cheating because I have slept with him before!  Right?  Is that the rule?  Geez, I\'m awful.  I wonder what I would tell Tristen if I cheated on his daddy?  How could I explain it so that he didn\'t hate me?  I\'m such a chicken shit though, I\'d probably never do anything anyway.  I sit back and think about it.  BUT, if he just happened to be at the same bar as me one night.... Well, we could talk, right?  I am too fat to see him now.  I\'ve got to loose some weight first.  I went to the gym today and ate an Atkins friendly wrap.  I\'m determined to get this lard off.  Well, gonna go now.  '),(6369,3215937,'06,July,2004','2004-07-06 00:00:00','   I don\'t blog nearly enough as I should. I got in a fight with my husband last night.  He told me I was bitter like my mom was when she was married to my dad.  So, I said maybe when I\'m forty-something I\'ll find true happiness like her too.  So he said are you happy?  I said no.  He got up, broke the lamp and went to the couch.  I laid there for a while thinking.  I didn\'t really care to go and get him but I figured I should.  I did because I\'m married and you\'re supposed to do that stuff, right?  Even if you don\'t give a shit.  So I went in there and told him to come to bed.  He said that I\'m weird and one day I\'m fine and then I go and say something like that to him.  Then I asked him why he told me I was bitter?  He didn\'t have an answer.  He thinks that if we ignore our problems, they go away.  We haven\'t said anything else about last night and I bet you anything he thinks it\'s all fine again.  Meanwhile, it\'s another unresolved issue for me.  We went to the lake over the weekend I saw couples loving each other and saying nice things about one another.  I was so jealous and sad.  I don\'t have that.  Am I supposed to suffer through a marriage like that?  I feel that my husband doesn\'t want me sexually and I know it\'s my fault.  I am fat and ugly and don\'t have anything attractive about me.  I\'ve always blamed myself.  I was thinking yesterday about how much I need passion and that stuff and I figure if he won\'t let me get a divorce, then he\'s going to have to let me cheat then.  I can\'t go on for the rest of my life only having sex once every couple of months.  Maybe that makes me selfish but I can\'t help it.  I\'m only human and I don\'t think I\'m asking that much.  My life is miserable sometimes.   I dreamt of walking around a mall looking for any attractive guy I could find.  I just wanted to have sex with someone, anyone.  I wish Brad weren\'t in San Diego...  '),(6370,3215937,'05,August,2004','2004-08-05 00:00:00','   Jeremiah,   I am writing this to you because it seems any time we try to talk, it turns to an argument and I think I am done with those. I\'ve been doing a lot of thinking lately and I keep coming to the same conclusion. We fight so much and we are both so angry and I worry about the effect it has on our son. I am going to move out. I don\'t want to argue about anything and I\'m willing to listen to any requests you might have about Tristen or anything else. I am so afraid of hurting you, but at the same time- you have been hurting me for so long. Maybe you don\'t even realize it, or it\'s a subconscious thing. I keep a lot buried inside and now it\'s festered up and overflowed. I\'ve never felt so worthless in my life. I know that a lot of this is due to me and all my problems and I need to work on those. I have realized though, that I can only work on these on my own. Time apart will help us realize what we both want from life. There\'s no way that Tristen would not benefit from this. Right now he has two agitated and angry parents who yell at each other an at him daily. It hurts me to the core to think about what we might be doing to him. I love him more than life itself and that\'s why this decision is so important. We will always be linked together through Tristen even if we do not live together. I want to be happy, and to find happiness. I cannot find it with you right now. I beg you to understand my decision and to let me do what needs to be done. If I stay, I will only become more bitter and hateful. I can\'t even stand myself these days. I care for you and your feelings and the last thing I want to do is hurt you or take anything away from you. I promise to never do that with Tristen, but as for myself, I must go and sort out my thoughts. They say if you love someone, let them go.       Love, Shelly  '),(6371,3215937,'04,August,2004','2004-08-04 00:00:00','   Omigod, I\'m screaming on the inside. I\'m ripping and tearing and I want to bust loose so bad. Isn\'t it horrible to know exactly what you want but don\'t know how to get it? Screw all this fear! Can\'t something make this easier? I wish he would hit me. That would be the best thing ever, then I could leave and not feel the guilt. Maybe I should provoke it? What am I saying? Am I nuts?  Every time I read an email from Brad, I smile so big and I can\'t even make myself stop. This big, goofy ass smile across my face. I started reading \"The Notebook\" last night and the romance and passion has taken it\'s toll. I want that. God, do I want that. I want to feel love again. I want someone to want me, right here and right now. I want that kind of urgency. I have to realize that I control my destiny, no one else. What am I going to do? I can\'t even afford a divorce I think. I don\'t know. Only if it were uncontested. I bet he would argue over every last detail. He can\'t stop me from moving out though. Where would I go? To my mom\'s? That\'s so far away. Ugh....  '),(6372,3215937,'04,August,2004','2004-08-04 00:00:00','   Call me a hopeless romantic, but I think that I\'m destined to find my true love out there.  I don\'t think the one I\'m with is that at all.  I\'ve made up my mind.  I don\'t want to have another baby.  I\'m not going to.  What I want is to get a seperation.  I want to start my own life over again.  I want my independance again and I don\'t ever want to marry again.  I remember getting into fights in the past where when he threatened to leave I would cry and beg for him to stay.  My heart would freeze and hurt but know.... oh no.  When we fight and threats start, I feel relief.  I feel like, is this it?  Will this be the last fight?  Can we get a divorce now?  I\'m so afraid.  Not of being alone, not of starting over.  I\'m afraid of hurting him and I\'m afraid of what I will do to him.  Not one person can say that Tristen would be better off if we stayed married.  We would both be better parents apart.  I just can\'t do this anymore.  I\'m too young to be in this spot.  He\'ll be home soon and I don\'t want him to read this blog.  Marriage is so stupid and pointless to me.  Why do you want to tie yourself to one person and have to agree with everything for the rest of your life?  I want to be me again.  '),(6373,3215937,'02,August,2004','2004-08-02 00:00:00','   It all started with a dream about a monkey. Thursday night I dreamt that Brad was in town and I met up with him and his friend. We first met up in my room, at my dad\'s house. My sister was sitting there with me and my dresser fell on us but I got out of the way. My sister\'s head was bleeding and stuff and I was so mad at the dresser. I took it outside and threw it on the ground and broke it. I remember that I think I had my pajamas on and make up from the day before. Then it went on to a street in Duncanville. I was standing there with Brad and his friend and I kept kissing him. He was a little stand-off-ish. I was watching someone\'s spider monkey and it kept running and jumping all over the place and I ended up having to put the damn thing in my purse and zipping it up. Brad went home and went to bed and I came there and tried to get him to come back out. It was weird.    I called him on his cell phone on Friday afternoon. He answered but I could tell he was in a car with the windows down. He said he was in a jeep. I could barely hear him but he asked how I was and how Tristen was. I told him that I would call back later when I could hear him better. I went to my softball game after that. Jeremiah went to watch the Tyson fight so after I got Tristen in bed, I snuck outside and called him again. No answer. I left a message. I never heard back. Oh well. I don\'t know if it\'s Brad, or the idea that he represents. Something different. Any guy seems perfect when you live with Jeremiah. He basically called me out on Thursday and said that I\'m a bad mother. He said I don\'t watch Tristen good enough and that he only gets hurt when I am watching him. He can make me feel two inches tall and I think he likes it that way. It\'s wierd because I don\'t see myself with him for the rest of my life, but I don\'t know what to do. I\'m scared... I\'m also torn. One day I want to have another baby and the next I want to get the hell out. He wants to sell our house and build another one. All that does is scare me. I don\'t want anymore comittments. Then I think, I\'m such a looser. How can I encourage Jana, who\'s pregnant, to leave her man when I\'m too chicken shit to do it myself? I guess it\'s easier for me to tell Jana that but I realize how much harder it is to actually do it. I can convince myself that she\'ll be okay and better off here but can she do that for herself? I guess if I could snap my fingers and make anything happen, I would have a rent house that Jana and I could live in. We could raise our kids together and be there when times are tough. You don\'t need a man. You don\'t  have  to be married. When did it get easier to stay and take the abuse than to leave and set yourself free? What do I tell my husband? You irritate the shit out of me and I want to start seeing other people? Yeah, that would go over real good. He almost hit me again the other day. I didn\'t change the cat litter so he took the cat and was trying to leave and take her to the pound. I threw a plastic cup at his truck and he came charging back out and forced me into the house and I was sure he would choke me or something. He restrained himself though. I don\'t need that drama in my life. Not in front of my son either. Woe is me...  '),(6374,3838218,'31,July,2004','2004-07-31 00:00:00','   urlLink    i\'m an angel sent by heaven                                   angel\'s golden rule... be kind and love will rule the world                                                                my magical word is i love you!!!                       happiness is my priority                                        send a helping hand to the needy                              and sing to brighten the lonely moment&nbsp; urlLink  '),(6375,3838218,'29,July,2004','2004-07-29 00:00:00','   urlLink    cutee me&nbsp; urlLink  '),(6376,3838218,'29,July,2004','2004-07-29 00:00:00','   urlLink    THIS WORLD&nbsp; urlLink  '),(6377,3838218,'20,July,2004','2004-07-20 00:00:00','   There\'s something you my give      A friend and stranger too      it seems that when you give it      It\'s given back to you      This gift is worth a million      But doesn\'t cost a dime;      It\'s lasting in effect      But doesn\'t take much time      This simple little gesture      Can make the dayworthwhile      It\'s just as good as sunshine      It\'s what we call a smile.        AA'),(6378,3838218,'20,July,2004','2004-07-20 00:00:00','   TOP TEN THINGS MUST DO:     DRINK MILK BEFORE SLEEP.    SOUND TRIP    WATCH RUSH HOUR AND CHARLIE\'S ANGEL    GET MUCH SLEEP    PLAY BIKE UNTIL IT\'S WORN OUT    EAT CHOCOLATE    MEMORIZE FLAGS AND MAP    ABC\'S AND NUMBERS    GIVE DANYL MILK    PLAY THE SIMS      &nbsp;'),(6379,3838218,'20,July,2004','2004-07-20 00:00:00','   the darkest hour is the best time to see the stars. it is when life seems most harsh and painful that the most wonderful blessings arrive.'),(6380,3838218,'14,July,2004','2004-07-14 00:00:00','   urlLink    spiderman&nbsp; urlLink  '),(6381,3838218,'09,July,2004','2004-07-09 00:00:00','   urlLink    bald head&nbsp; urlLink  '),(6382,3838218,'08,July,2004','2004-07-08 00:00:00','   urlLink    what\'s up???&nbsp; urlLink  '),(6383,3838218,'08,July,2004','2004-07-08 00:00:00','   urlLink    adyn\'s 1st travel.&nbsp; urlLink  '),(6384,3838218,'07,July,2004','2004-07-07 00:00:00','   i really can\'t believe it with my two eyes! imagine i bought a new pants for a price of less than 80% from the original price(giordano) and shirt (???) wow...wow...  i saved more. it gives me euphoria.  that\'s for now.  '),(6385,3838218,'06,July,2004','2004-07-06 00:00:00','   changes maybe good or bad. we have to avoid the bad effects that adversely affects our life, property and environment. modernization through science has both good and bad effects. but if the bad effects supersede the good effects, then these modern changes must stop. we must do our share in maintaining a healthy environment.  i love mother nature.  '),(6386,3838218,'05,July,2004','2004-07-05 00:00:00','   this is my first time to create my new website thru blogspot. take time to tour in my humble creation. it\'s really exciting to have one. i\'m beginning to start a new realization that nothing is permanent in this world. everything... as in anything... at least give me one. confussions...commotion...communication...anything goes. actually,i don\'t know how to start this. it makes me kinda sucks... beginning to have headaches... but it\'s new to me. i don\'t know why i\'m doing this. maybe, for self-satisfaction. right!!??!!!  '),(6387,3838218,'12,August,2004','2004-08-12 00:00:00','   whenever i go around, i used to memorize the landmarks of each place. i have good sense of direction...would you believe, my mom let me memorize the philippine map... and even the streets we\'re going to pass way to our home. imagine, at my age of 3 years old, i opted to sit beside the window of the car just to take glance of the road.   and it\'s great help because everybody is proud of me. one more thing, i\'m always mesmerized of the different flags of the world. everytime before i go to bed, i always take a look on flags. i just don\'t know why...maybe, i want to travel and learn different cultures.  '),(6388,3838218,'09,August,2004','2004-08-09 00:00:00','   i really miss to blog. nwy, i\'m here na. i miss my schooling. i miss mindoro. i miss everybody.  '),(6389,3838218,'01,August,2004','2004-08-01 00:00:00','     '),(6390,3287144,'31,May,2004','2004-05-31 00:00:00','   I\'ve spent over two hours waiting for a telephone call and when this post is finished I shall give up on the waiting and do some  real  work instead of fooling around playing solitare as I have been doing.  In other news, I have decided on a new layout and like I promised, those music, mood and site things have gone, but fear not, thanks to  urlLink iMood  we have a lovely little smilie thing on the side menu so now you know my mood and get a lovely smilie as well - what more could you ask for?  Anyway, its a lovely day today so I think I might go outside and enjoy it, Ciao! '),(6391,3287144,'30,May,2004','2004-05-30 00:00:00','   Quite out of the blue ( mem. find the origin of that phrase) I have decided on several things; 1. To stop doing the Live Journal-esque things at the bottom 2. To try and put some things on the side bar 3. To use \'()\' instead of \'[]\' so much when doing brackets Of little consequence really, but this will be the last post of being in the style of Live Journal. Don\'t cry too much children.  Anyway; to the present, it\'s Sunday afternoon and I have nothing to do. Revision is always there, but that can be tiresome, I don\'t have a job, and haven\'t since I completed my work for  urlLink Hunter Seeds\' website  and I haven\'t been able to buy myself a bass guitar yet.  A glance over today\'s \'world news stories\' showed every one of them to be war-related. Seems that\'s all which is important today is war. Global Warming is of little consequence - in fact, itâ€™s rather good if youâ€™re Russia [warmer climate is better for their agriculture] and you don\'t really care if youâ€™re America [because you ARE America and if it wasn\'t for them there would be a global depression]. Which I think is very thoughtful of the latter: economic depression soon or apocalypse later.  Music: Junior Senior - Move Your Feet Mood: Thumb Twiddling Site:  urlLink http://www.livejournal.com '),(6392,3287144,'29,May,2004','2004-05-29 00:00:00','   Again I am blogging at a silly time in the evening, the time on this post is wrong to be honest - it is in fact the 30th by now but I decided to disguise it at the 29th so it looks nice and that I do one a day. Which I do, even if I am a day behind...  Anyway as today [yesterday] is [was] a Saturday I thought I deserved a lie in but eventually decided that it might be nice to be awake in time for lunch and I was just having a shave when I heard a knock at the door which told me that lunch was ready and there were a plate of spring rolls outside if I wanted them.  What is the point? People will do this to me all of the time, see I am busy and interrupt me with stuff; do I want to eat spring rolls while shaving? No. I\'m holding a sandwich with a receipt in one hand while trying to put money in my wallet with the other; do I look like I need to be served? No. I\'m reading a book; do I want someone to turn the lights off to save energy? No. I could continue on with this stream of stupidity but I won\'t and instead say this; if you ever meet me, whoever you may be, don\'t bother me with similar questions or actions!  Anyway before I do go to bed [which I really want to do] I wish you all good morning.  Good morning.  Music: Coldplay - Spies Mood: Dull Site:  urlLink http://www.neave.com '),(6393,3287144,'28,May,2004','2004-05-28 00:00:00','   Its quite late and I shouldn\'t be doing this as late as it is but hey. Anyway, today was my English Literature exam - an exam where it is by no means an exam on your knowledge of English literature and it isn\'t really an exam on  English  literature; my set text was Of Mice and Men by John Steinbeck; an American. The poems that I looked at were written by an American, a welsh person, an Irish person and a Scottish person. No one technically English was involved in my English exam. But what is it to be English? They all speak the same language, some of their descendents could have quite possibly been English and an \'English\' persons ancestors could quite possibly be Anglo-Saxons or of some other European origin. Who is English?  As it is indeed Friday night [now Saturday morning], I went out for a meal earlier, except on finishing the mile walk to the Indian restaurant, one of our party realized they had forgot something which meant that I had to get from the restaurant, back to the house, and back again - a two mile round trip - all in time for the first course. I did it, partly as they were delayed by a few minutes and partly because I ran a good deal of it. Getting back to the restaurant I arrived in good time to receive my curry. I would like to inform anyone considering doing a two mile run and then going for a meal, to go for something a bit more cooler, like a sandwich, curries just don\'t seem to be quite the right thing.  On the topic of sandwiches I recommend a Safeway Chicken and Bacon \'sub\' [you know, one of those baguette like things which isn\'t], at Â£2.69 its hardly value for money but for a cold and wet May day it worked for me. Speaking of baguettes, Sally\'s Place provides an excellent guide to them. Let them eat cake.  Music: Bo Diddley - Road Runner Mood: Not so bad Site:  urlLink http://www.sallys-place.com/food/single-articles/french_baq.htm '),(6394,3287144,'27,May,2004','2004-05-27 00:00:00','   \"Hi\" \"Hey, how are you?\" \"Fine\" \"Good\" So went the guilty conversation between a pair of builders who were taking a sneaky break from work and me, the son of their employer who appeared to be bunking off school. Neither of us could blame the other and so the conversation went.  Anyway, I\'m not bunking off school as it is still study leave and I\'ve just had the gruelling three hour IT exam, while it is simple it is incessantly dull; the practice exam I had for it back in the winter wasn\'t so bad as half way through a wasp landed on the next persons computer which provided a quick break of humor but alas, today, nothing.  Prior to meeting the builders [who to be fair, deserve as many breaks as they want] and after the exam, I decided on a visit to town where I purchased a portion of fish and chips, I consumed a good deal of them but left one or two because I wasn\'t very hungry. Point is I went to put them in a bin and this lovely old woman decided to comment; \'oooh look, he\'s throwing some of his chips away\'. Yes I know I am, and I dare say your friend can see that as well. Why comment?  To be fair I quite like the way old people are very talkative to strangers, they normally talk to me if I am either alone or with a female friend [never if it is me and a male friend or me with several female friends], and usually they only talk about the weather but its still nice just to think that unlike the rest of our society, they can remain happy and content as well as confident to talk to strangers.  Finally while returning home I had the misfortune to be involved in a drive by in which I was \'shot\' at by someone with a water gun. This has happened twice now, next time I will be sure to do something - what exactly I will do I am not sure of but who ever it is in that blue car with water guns will regret the day they crossed swords with the Don. I mean me...  Music: Hanson - MMMBop Mood: Sort of Stressed Site:  urlLink http://www.hansonline.com '),(6395,3287144,'26,May,2004','2004-05-26 00:00:00','   This was going to be quite a bleak posting but the events of the last five minutes have managed to cheer me to a level above that. Most notably is my newly found cruel hobby of engaging arguments with argumentative people who don\'t have a chance of winning the said argument. It\'s cruel I know but secretly fun.  I had another dream which I can recall again, I don\'t know whether it is the fact that I am recording them that helps me remember them or what but here is last nights; I found myself in a shop with my father, the shop sells a variety of 1930s toys, especially models of cars, boats and trains, the shelves are dusty and the shop keeper wears old clothes and looks tired, I don\'t know why we went in there but when we come out again we have bought the shop. The next day I start work in the shop and decide that it needs to be cleaned and that way it can become successful, I start my attempts at cleaning it but I get tired and put on an old jumper and over the coming days I find myself transforming in to the old shop keeper. Well in other news I bought myself a replacement copy of Dracula [Bram Stoker] as I had lost my original copy, Â£1.50 - a bit of a bargain really. I have decided to re-read it as well, beats revision...  Music: The Rolling Stones - Sympathy for the Devil Mood: Aloof Site:  urlLink http://www.bbc.co.uk/.../... '),(6396,3287144,'25,May,2004','2004-05-25 00:00:00','   My German listening exam was dire, not to the extent of being not great, but a true calamity. I think it was destined to go badly, there were a number of clues but I didn\'t see them until now. I had another dream last night, I can\'t remember it exactly but it went a little like this; I am attending a posh dinner and I go wearing a tuxedo, someone I used to date is also attending and I remember how I still love her. She still hates me. I shrug it off; she never liked me and so nothing changes. In other news I have started re-reading all of Sherlock Holmes books, they are truly magnificent. Sign of Four this evening, brilliant. In other news I\'m listening to the Cuban Overture as composed by George Gershwin, my brother tried to sell it to me as jazz but its quite simply, not the standard style of jazz. But its still good. To be honest I\'m quite envious of any classical composer, it would be such a marvelous skill to be a good classical composer, despite lacking in the same recognition as their pop counterparts. Play on!  Music: George Gershwin - Cuban Overture Mood: Not all that bad despite being a failure. Site:  urlLink http://www.classical.net/music/comp.lst/gershwin.html '),(6397,3287144,'24,May,2004','2004-05-24 00:00:00','   I cook. I iron. I clean. I am the 21st century house husband - well maybe not but that\'s what I have been doing a lot of these past few days. I\'ve spent most of the day losing time to trivial activities [in other words I took a siesta this afternoon] instead of revising for my  politically correct  German test tomorrow.  Its one thing to be hit with buzzwords like \'equal opportunities\' and other social issues during lessons designated for such wastes of time but I think its taking things too far to be able to understand that: Ich halte Gleichberechtigung fÃ¼r sehr wichtig Which roughly means; I think equal opportunities are very important I could also inform you of how to say how to say something is racist or sexist... isn\'t this going too far? What are majority of people going to do with a GCSE in German? Some may go on to A Level, and then maybe take it to degree level but that\'s quite rare, shouldn\'t we focus on things that  will  be of use, not how to talk about people being \'der Gruppendruck\' [peer pressure] or \'die Arbeitslosigkeit\' [unemployment]. Let\'s focus on what will be of use, not what\'s \'politisch korrekt\' [I think you can guess that one].  Oh I also had a dream last night; I am back in school [again] and I\'m feeling quite lost, suddenly someone who I used to be seeing for a while runs up to me and talks to me briefly, she then moves in to kiss me but instead tears the flesh of my face. She runs off and I also go off and see a small year 9 girl who tells me that she ate my face. I respond \'yes, she did eat my face\'. I carry on and no one else notices my absent face. Like I always say; Fetch Freund.  Music: REM - Get Up Mood: Quite jovial Site:  urlLink http://www.online-literature.com/doyle/study_scarlet/ '),(6398,3287144,'23,May,2004','2004-05-23 00:00:00','   Yesterday I neglected to mention that I had a letter from my  urlLink Careers advisor  the letter consisted of some 20 pages of the following; Notes of what I said [just encase I forgot what subjects I am taking at A Level] and careers that match my preferences, some of my favorites include \'Virtual Reality\' and \'Science Fiction\'. I\'m guessing the former is where you have an imaginary job and the latter is where you pretend you work on a space ship. Maybe not.  Also I seem to be under constant attack from beetles looking rather similar to  urlLink May Beetles . While these pose no threat to my existence, I still feel it my duty to dispatch them, except I can\'t bare to kill the things so the result is a twenty minute chase in order to corner the thing and then send it out through the window and hope that it managed to survive the fall. I worry too much.  Anyway, I better do some  more  revision. It\'s all good.  Music: REM - Radio Free Europe [Original Hib-Tone Single] Mood: Relaxed Site:  urlLink http://www.ellor.vispa.com/rat/welcome/my_bit/error/error.html '),(6399,3287144,'22,May,2004','2004-05-22 00:00:00','   This morning I took a trip to Chester, for those who are unfamiliar with Chester, it is a rather charming place which got a bit too popular for its own good. Anyway I bought a few albums and a new jacket and wallet, both fit the idea of being \'smart/casual\', which seems to be my current adopted style. Except for my hair which is just a mess.  My lunch consisted of going in to a small cafÃ© known as \'The Crypt\', a quick fifteen minute lunch extended to an hour due to their incompetence but after a complaint a full refund was obtained which equalled a free lunch for me. Top stuff!  After lunch I noticed an absurd number of fat people roaming the streets, maybe it was as I had just had lunch or they only come out after midday, point is, if your fat, why don\'t you try and make an attempt at hiding it, or more to the point, not advertising, for example, for the seriously obese, a \'crop top\' is  not  a good idea. Wear a lot of black, and keep away from stripes. Worked for me when I was a bit podgy.  Music: Gang of Four - At Home He\'s a Tourist Mood: Stylish Site:  urlLink http://www.quaffersoffers.co.uk '),(6400,3287144,'21,May,2004','2004-05-21 00:00:00','   Today I can say that nothing of any real value happened, my attempts at making lunch were not entirely successful as the custard for the intended pudding went a bit wrong, while it may be easier to do it in the microwave, I recommend you don\'t.  I also watched two hours worth of TV programs with Dan Cruickshank in, very sad I know but has anyone stopped to look at the coat he so often wears? Its rather wonderful in its own way but I don\'t know why. Today my friend read an article that had been written about her but as if she had written it after she had an interview with one of the journalists. The writer managed to infer that she was spoilt. Well said it to be precise which has my friend aggravated beyond measure as it looks as if she is admitting to it. Which she should. Just joking.  Music: Queen - One Vision Mood: Tired Site:  urlLink http://www.spartacus.schoolnet.co.uk/USAmccarthyism.htm '),(6401,3287144,'20,May,2004','2004-05-20 00:00:00','   I had my second maths paper today, that is all I will mention of it and its evil.  After the exam I went in to town to see someone off on the bus and then decided to stop by the record store, during my time there I managed to kick the store clerk [well kind of nudged him with my leg] and bought a Joy Division t-shirt.  Only when I got home did I realize Joy Division never released any official t-shirts [can\'t remember where I heard that but I think its true], but its not too bad; I only know one other person who knows of Joy Division so I doubt they\'re going to tear me down for wearing a fake, assuming they know it is.  Finally my horoscope tells me to look for romance. Hmmm.  Music: Joy Division - Love Will Tear Us Apart Mood: Relaxed! Site:  urlLink http://www.horoscope.com/ '),(6402,3287144,'19,May,2004','2004-05-19 00:00:00','   meÃ‚Â·chanÃ‚Â·ic - noun. A worker skilled in making, using, or repairing machines, vehicles, and tools. Mechanics are great. Well except for the dodgy ones, but in general, rather brilliant. The save us hundreds of pounds and by making things last longer, help the environment. Yes, mechanics are quite vital to our society.  Except if you look a bit further down the dictionary you will see another meaning;  The branch of physics that is concerned with the analysis of the action of forces on matter or material systems  Now I\'m sure mechanics [the second version] in the right hands can be really useful, just it doesn\'t seem that useful to know what the tension is on a piece of string between two objects. Oh and I have the exam tomorrow.  In other news my relative who was given two months last week died last night. I seem remarkably untouched by the whole episode but that might be for the best as I have my GCSEs to concentrate on now.  Finalltoday\'sys link is to The Churchill Insurance mini site about their corporate mascot, also called Churchill, it isn\'t great except for the ability to watch adverts - I recommend the Alien Invasion. Brilliant.  Music: David Bowie - Ziggy Stardust Mood: Not too bad Site:  urlLink http://www.churchill.co.uk/fun/site/outtakes/wm_index.html '),(6403,3287144,'18,May,2004','2004-05-18 00:00:00','   It is my belief that we underestimate the Irish - today I had an Advanced Maths paper and unlike the British exam boards, the Irish didn\'t have a single \'Blank Page\' [which are always labeled \'Blank Page\']. So this meant even the back cover was used - very economical! However I didn\'t know this and thus missed off the final question... Good for the consumption of paper in Ireland, bad for my grade [well 8% of it anyway].  I have also decided that exams are only good for providing an excuse to hug pretty much anyone, a fact I intend to abuse as much as possible over the next few weeks. Also prior to my maths exam a friend came over to \'revise\', instead we locked horns in a match of  urlLink Worms World Party , my friend however has an unnatural ability at the game, however today I beat him in the two matches we played. Seems I can win at something, just nothing useful...  Finally I recommend the link at the bottom highly; strange that something so inexplicably annoying can be made to be mildly amusing by calling it a game and adding music but then again by listening to my MP3 player and thinking life is a bit of a game I guess I have done the same for my mildly amusing life.  Music: T-Rex - Children of the Revolution Mood: Victorious! Site:  urlLink http://www2.b3ta.com/realistic-internet-simulator/ '),(6404,3287144,'17,May,2004','2004-05-17 00:00:00','   So my prom date is ignoring me, I had 2 people in my family diagnosed with cancer last week and I discovered that I am rather poor at Advanced Maths which I\'m studying which leaves me with very little left to hope for.  In other news I had a rather interesting dream last night;   I\'m back in the forth year and I arrive for school in my uniform but everyone else is dressed casually so I go home and change but when I get back casually dressed, everyone is in uniform and each time I change yet remain different to everyone else, no one notices that  I  am different.   Fairly easy to interpret if I say so myself. I\'m fascinated by dreams, they can say so much and often with doing so little. I guess I should thank  urlLink Sigmund Freud  really - just I can\'t see anything sexual in being dressed differently.  Music: Radiohead - Exit Music (For A Film) Mood: Failed. Site:  urlLink http://ferryhalim.com/orisinal/g3/pig.htm '),(6405,3287144,'16,May,2004','2004-05-16 00:00:00','   For some unknown reason I decided to put on a crisp long sleeved white shirt on this morning, it makes me look remarkably smart despite the fact I probably won\'t be leaving the house today.  Rather a dull day, well its quite bright but its been a dull one for me, anyway, as I have nothing good to say today I will leave things as they are. Visit the link at the bottom for enlightenment.  Music: Bauhaus - Silent Hedges Mood: Blissfully unaware Site:  urlLink http://www.mattstuart.com '),(6406,3287144,'15,May,2004','2004-05-15 00:00:00','   I arrived quite late to maths this morning but no one seemed to bothered. 3 Days until the exam... I\'m not really too worried which in itself worries me. My current movie recommendation is for Van Helsing. It\'s not really a great film, its alright but Richard Roxburgh is in it who is rather special in his own little way.  In other news I discovered that \'painting\' your nails with permanent marker is  not  a good idea - I thought it was yesterday while idly colouring them in blue, now 24 hours on I have been able to get most of it off except the bits around the edges which prove to be the most annoying.  Finally my father managed to get a small bargain from Aldi yesterday, a Pocket PC, or a \'Multimedia Pocket PC PDA with GPS Navigation\' to be precise. But like all things from Aldi which seem to good to be true, this little technological master piece isn\'t all that it could be. Think of it like this; just because it is a fraction of the size of a standard PC, doesn\'t mean you\'ll have a fraction of the problems - quite the opposite. Still it has Solitaire on it so I\'m not too bothered plus I have managed to fix number of these problems.  Oh and at the bottom there is a link to do a practice theory exam; 26/35 means I failed!  Music: The Smiths - Panic Mood: Fairly good Site:  urlLink http://www.dsa.gov.uk/mockpaper/theoryintro.htm '),(6407,3287144,'14,May,2004','2004-05-14 00:00:00','   My shirt is signed and I have had more hugs than a bear, yes - today was the end of high school. Except I have to go in tomorrow. Then next Monday, and Tuesday... Well it was the last  official  day. According to my shirt I\'m God and I have enough \'good luck\'s to repel a lot of those cursing chain mails.  So I have spent over 12 years of my life in education now, taken from society and not given anything back. A pleasing thought that. Last night I found myself watching Question Time from St Andrews featuring some 20 something who I announce to be very irritating. On the issue of  urlLink a 14 year old becoming pregnant and having an abortion without the mother\'s knowledge  she said that it showed that the government\'s health and education departments have failed. Now I am not a great supporter of the current British government isn\'t this a little extreme? It\'s like saying that because one crime occurred, the police have failed. This same woman had this annoying habit of calling \'here here\' to anything anti-government - I dare say that if it was announced that Tony Blair \'smells\' she would have called \'here here\' to that as well.  On a more personal note I finally asked someone to the prom and she said yes. Now I need to find my self a tux and everything else that goes with it... She\'s worth it though!    Music: The Lightning Seeds - Sugar Coated Iceberg Mood: Quite happy. Site:  urlLink http://www.faultline.co.uk '),(6408,3287144,'13,May,2004','2004-05-13 00:00:00','   One more day of High School left; I won\'t be sad, quite happy that it is finally the end, High School was never my thing. Today I had my German speaking exam it didn\'t go too bad apart from the following mistakes I made;   &#8729; I eat pizza for breakfast.  &#8729; I don\'t go swimming in the swimming pool; I go swimming in the river Rhine.  &#8729; My father is dead. [He isn\'t]  In a bitter irony though I have since been told that my  grand father has two months to live: On Monday he was admitted to hospital, on Tuesday he was diagnosed with cancer, on Wednesday they operated and now it is Thursday and he has two months to live.  It\'s a common question, but what does one do if they are told they have two months to live? I don\'t really know, but surely if we do  live  then there is nothing to be done, just to carry on with our lives until something does happen? For example by going around the world on holiday, aren\'t we saying that we haven\'t lived and now we are trying to get something back that we lost? Could it not also be said that we are giving up - admitting that we\'re beaten, a doctors judgment our death sentence?  Music: Aerosmith - Jaded Mood: Shocked Site:  urlLink http://www.cancerresearchuk.org/ '),(6409,3287144,'12,May,2004','2004-05-12 00:00:00','   Today was  the  definitive typical day. I mean there was nothing unusual, nothing to cause any emotion in me. Subsequently I am feeling quite tired for having had such an uninspiring day. Tomorrow I have my German oral exam, after calling my examiner a dragon last week (and yes, she did hear me) I am not entirely looking forward to it. On hindsight I think she not really like a dragon - more like  urlLink The Watcher  from Lord of the Rings [I\'m not a big Tolkien fan, just someone bought me the game at Christmas which I felt obliged to play].  In other news a small ray of sunshine hit my inbox this morning with news that Ross Noble has returned from Australia to unveil his new show \'Noodlemeister\'. Which is a good thing but it now makes my t-shirt for the \'Unrealtime\' tour look somewhat dated. If anyone does want that t-shirt they can buy it off the site at the bottom as well as find listings for the new tour.  Music: REM - Low Mood: Alone Site:  urlLink http://www.rossnoble.co.uk/ '),(6410,3287144,'11,May,2004','2004-05-11 00:00:00','   I think I have this sorted out; Careers advisers only interview people as they feel they have failed to get a â€˜realâ€™ career and must validate themselves by interviewing others. Yes, today I was chosen quite randomly to be interviewed. It seems to be a matter of stating the obvious [\'so your good at maths and your taking economics... have you ever thought of being an accountant?\']. Eventually I decided that I would make out that engineering would be the career for me so she wouldn\'t have to continue guessing jobs for me - the plan seemed to work as the interview was concluded fairly quickly after that.  4 Days until the end now; I had two random second or third years [they all look so alike] randomly scream in my face today. Perhaps itâ€™s a hormone thing, I don\'t know but I remember not doing it to year 11s when I was a little second year. They should have more respect for their elders!  Music: The Cure - Boys Don\'t Cry Mood: A bit down. Site:  urlLink http://www.rinkworks.com/bookaminute/b/bronte.eyre.shtml '),(6411,3287144,'10,May,2004','2004-05-10 00:00:00','   So a new \'site\', just it\'s not a site, it\'s a blog. Bwa haha. Anyway, this should be some amusing reading for someone, and if not, it\'ll be a no use to no one.  Anyway, it\'s the beginning of the end today; Monday, the beginning of the final week at high school. I won\'t miss the place, in fact for the good part of this year I have been looking forward to it. Anyway, visit Slowly Downward [see link at the bottom], if you can \'get in\' to it, then you\'ll find it very... interesting. Soon people will say that about this.  Music: U2 - When love comes to town Mood: Quietly optimistic. Site:  urlLink http://www.slowlydownward.com/ '),(6412,3287144,'30,June,2004','2004-06-30 00:00:00','   Well the taster day at sixth form went alright; Economics seems strangely interesting and our head introduced his speech with the great line; \'this is not a speech\'. I have this wonderful plan to sit in the common room and drink coffee out of a thermos in there all day. Yes. I aspire to great levels.  But you can not disagree [double negative, naughty me] that it does sound awfully nice. I just need to find someone who likes coffee to enjoy it such moments with me; rain lashing down on the windows outside while we sit in drinking warm coffee watching the more misfortunate people traipse in while we are both roasty toasty.  Just sounds so very snug and yes, I do need to get out more. '),(6413,3287144,'29,June,2004','2004-06-29 00:00:00','   Another day wasted - well not totally wasted but 90% wasted at least! Anyway, I managed 6 hours with my contact lenses today after spending 20 minutes getting them in. Really need to work on that.  Tomorrow is my \'taster day\' at sixth form, I haven\'t applied to anywhere else so if I don\'t like it then I\'m in for a bit of bumpy ride really. If I do then that\'s excellent. Let us hope for the latter option.  So tomorrow I will be back; back without a Caramac.   [Yes that was a cheap excuse to get a rhyme in but I do have a Caramac at the moment which is lovely.] '),(6414,3287144,'28,June,2004','2004-06-28 00:00:00','   Well today I managed to watch another episode of the original Charlie\'s Angels, they don\'t make them like that anymore, groovy flares and all! Groovy! Anyway, contact lens insertion time is still 15 minutes but I kept them in for over 5 hours today - I\'m going to try for 7 or 8 tomorrow... and I\'m supposed to build up gradually apparently... oh well!  Finally I would like to urge anyone working in a Saturday job, a temporary job, or in actual fact, any job to quit as I still can\'t find myself a job and that would create more vacancies for me to try for. '),(6415,3287144,'27,June,2004','2004-06-27 00:00:00','   Well I have managed to reduce the time it takes to put a contact lens in by 20, to 15 minutes!  How fabulous. Well I\'m still bitter over last night but that\'s more me than anything. Today was spent in the pursuit of nothing which has probably made things worse. Oh well.  Just me. '),(6416,3287144,'26,June,2004','2004-06-26 00:00:00','   I commence with last nights dream as an introduction; I was on a Viking longboat sailing around an Eastern European country, somewhere not dissimilar to Prague in the Czech Republic. In the night sky are three moons, each carrying its own evil grin toward me, along the city streets lie animals in cages, possessed by an evil force as is each building, only the mortar stop it from escaping. I have a pewter staff though, on one side is the Christian Cross engraved; another shows Father Time - I cover the staff in a dark liquid and then this repels the evil; the three moons shrink in to one; the animals return to their normal form and the buildings return to their previous inanimate state. The Longboat continues to sail around the city and gradually the howling of wolves stop as I use my staff to repel evil.   Eventually the longboat makes one final turn and it hits a black wall, the wall doesn\'t smash the boat; instead it splits in half letting us pass through and only then do I realize that its just a fair ground ride. So to today, or this evening more importantly; with less than an hour left my prom date confirmed that she would be going to the prom and subsequently, I got changed and in due course, got to the venue. I played my part as a gentlemen better than ever but she just didn\'t seem interested me. She even left without informing me and so I too left and found myself watching Paul McCartney at Glastonbury, ending his set with Sgt Pepper\'s Lonely Hearts Club Band [the reprise].  We\'re Sergeant Pepper\'s Lonely Hearts Club Band We hope you have enjoyed the show Sergeant Pepper\'s Lonely Hearts Club Band We\'re sorry but it\'s time to go. Sergeant Pepper\'s lonely. Sergeant Pepper\'s lonely. Sergeant Pepper\'s lonely. Sergeant Pepper\'s lonely. Sergeant Pepper\'s Lonely Hearts Club Band We\'d like to thank you once again Sergeant Pepper\'s one and only Lonely Hearts Club Band It\'s getting very near the end Sergeant Pepper\'s lonely Sergeant Pepper\'s lonely Sergeant Pepper\'s Lonely Hearts Club Band. My life has a habit of playing songs to be ironic at me. Oh well, I guess that might be the last time I see her, so is the way of puppy love.  Well good luck to everyone else as I seem to be out of it. I\'ll end this with irony;   We\'re sorry but it\'s time to go. '),(6417,3287144,'25,June,2004','2004-06-25 00:00:00','   *phone rings*  Me: Hello? Caller: Hello, is Mr Pearson there? Me:  *another call for my dad*  No, can I take a message? Caller: Well could you tell Stephen Pearson that he needs to come in about his contact lenses Me:  *ooh thats me*  I\'ll make sure he knows this minute. Good bye Caller: Bye So went the call from my optician, in due course I went to the opticians to collect them; Me: Hello, it\'s Stephen Pearson about some contact lenses Receptionist: What about them? Me: Well I was hoping you could tell me... you call me, me... me brother! This morning about them Receptionist: Oh, well you can take them away now if you\'ve had \'THE TEACH\' Ok so I over emphasise the idea of the THE TEACH but to cut to the chase, basically \'THE TEACH\' is where your told how to put them in and take them out and as I looked around seeing the opticians to be completely empty, I was told that they were quite busy so I have to go back tomorrow for THE TEACH. '),(6418,3287144,'24,June,2004','2004-06-24 00:00:00','   Ironically, I hate students despite being one of them, and this is why;  After being turned down for a job at Safeway, I was on my way to the Job Centre who also had nothing for me, but on the way I met a student; Student: Hi Me: Hi Student: What are you listening to? Me: oh erm, Joy Division Student: Oh I\'ve never heard of them, are they old or new? Me: Old  She carries on with small talk for a while  Student: Anyway, do you know what stress is? Me:  *the evil part of my brain was telling me to tell her to naff off - I\'m that evil*  Yes Student: Well we\'re doing yoga classes for students Me: Well I don\'t really get stressed that often so no, not for me Student: Yes, you look pretty chilled out, well how about a donation? Me: Sorry I\'ve no money on me Anyway, I\'ve tried 3 places for work, nothing so far, I\'ll try two more tomorrow then I will give up - yay! '),(6419,3287144,'22,June,2004','2004-06-22 00:00:00','   Well in order to get money for the things I want, no, need - I must get a job so tomorrow I shall be on the prowl, shouldn\'t be too hard, everything else seems to be for me right now (not like that). '),(6420,3287144,'22,June,2004','2004-06-22 00:00:00','   My room is red! Oh life is so wonderful right now - I am to paint the skirting boards for the rest of the house and in return a bass guitar shall be mine! My life seems to be coming together!  But for a moment I would like to talk about my room; it went all rather smoothly until I had washed all the brushes and then came to packing away the dust sheets which had some wet white paint on from the previous day and my assistant managed to get several white streaks of this on to my fine wall - so back went the dust sheets, had to re-wash the brush... But it is worth it, even if it does look like an Indian restaurant. '),(6421,3287144,'21,June,2004','2004-06-21 00:00:00','   Well today was as predicted, a happy day spent painting listening to Steve Wright; today I only did the ceiling - tomorrow its the walls... Should be interesting...  A few tips for wanna-be decorators;    Paint barefooted, its cheaper to wash feat than buy new socks  If you use linen dust sheets - make sure they are thick enough else you get paint on the carpet like I did  Use a radio that doesn\'t matter if it gets paint on  If you can afford it, use rollers in a disposable fashion - they take years to clean    I will post a picture somewhere to show you the finished result... Until then; Fabius out. '),(6422,3287144,'20,June,2004','2004-06-20 00:00:00','   Well I didn\'t start painting today either. Tomorrow though I will, a day of painting in old clothes listening to BBC Radio 2 echoing about. It sounds strangely good and enticing.  Anyway, I must sleep. '),(6423,3287144,'19,June,2004','2004-06-19 00:00:00','   Well seems my 100 minute appointment was actually two 10 minute ones separated by 80 minutes which left me stranded in town with nothing to do - fortunately I sought refuge in the library where a copy of The Times entertained me, as did the rather comically written sign saying \'Watch out! Thief About!\' further reading informed me that some lost soul is going around there stealing... not really a criminal mastermind who ever they are.  The actual optician was a finer guy as any - had a wonderfully dark sense of humour which I admire in anyone, especially in people such as him. \'We\'ll give you a free trial, come in after a month and we\'ll sort out payment, assuming youâ€™ll make it back!â€™  Sadly I couldn\'t begin decorating today... hopefully tomorrow... '),(6424,3287144,'18,June,2004','2004-06-18 00:00:00','   Well I\'ve been waiting for summer for so long and now its here it is quite frankly, rather dull! I mean I spent all my money yesterday and I don\'t have a job which leaves me with nothing to do as I can\'t afford to do it really as I don\'t want to use money from my guitar/camera fund.  Oh well, enough of my troubles; Today while hopping around channels I discovered on BBC4 was Brian Wilson [of The Beach Boys] doing a live performance of the Pet Sounds album with his new 10 piece band...  wouldn\'t it be nice  if he could still hit those high notes, which he unfortunately can\'t and so I ended up flicking over to the ever-reliable \'Bid Up TV\' - as the name suggests its an auction on TV but it is one of the funniest things ever!  The average presenter is a bit of a flop and thatâ€™s the only reason why they are working there is because they aren\'t really top quality but they do work their hearts out on it all. However the presenters are nothing compared to the audience; despite selling all manner of rubbish, people ALWAYS bid, and this is the odd part; it costs Â£1 to bid, so instead of putting in a safer high bid, at the start people will bid the minimum Â£1. I have  never  seen anything go for Â£1, so why do they think today should be any different?  Poor misguided bidders... speaking of which I hopefully will be bidding on a camera on eBay. Poor misguided me. '),(6425,3287144,'17,June,2004','2004-06-17 00:00:00','   Yes my hair is cut and it doesn\'t look too bad, today I used some gum stuff, itâ€™s odd; neither like wax nor gel but better than both in my view. Anyway took another visit to Chester... seemed less fat people today than last time but they seemed slightly fatter than before. Anyway I was pressurized in to buying a Big Issue type thing today by the seller and my friend, Â£3 for something which was patently worth hardly anything as it was just photocopied.  Lesson; don\'t give in to peer pressure. '),(6426,3287144,'16,June,2004','2004-06-16 00:00:00','   Well I got my hair cut and its now, well cut. Details to follow tomorrow! '),(6427,3287144,'15,June,2004','2004-06-15 00:00:00','   Well today I had my history exam, which is the last major exam really so this posting will instead be dedicated to politics.  Well the EU elections are over and it appears that in general, few people bothered to vote in the elections, and a great deal of those did, voted for parties campaigning against the EU, including my much hated UKIP. I think Kilroy-Silk put why not to vote for them best - when asked what he intended to do in the EU parliament he replied to \'wreck\' it in a news conference yesterday, in which he overshadowed UKIPs leader who was also present.  I don\'t look upon him at all kindly if you hadn\'t noticed, he was forced to resign from his previous job after some racsist comments. And he didn\'t do well in that job to make good television [awful daytime television instead];   urlLink http://www.robertkilroysilk.com   Rant over. Kilroy bad. Green Party good. '),(6428,3287144,'14,June,2004','2004-06-14 00:00:00','   Wow, I had an English exam, something that was dangerously too exciting to me. Maybe because I had a little fun in it;  I linked a poem by John Agard to being similar to a song by Paul McCartney  I quoted lyrics from an REM song  In my original writing, God spoke on the PA  Weak but it got me through the whole tedious experience. Sadly I have a history exam tomorrow which is slightly harder to make your own fun in - I might try something but there isn\'t much room to manoeuvre with the Magna Carta.  Anyway I better be off revising for it, so I\'ll make like a tree [and leave/leaf]... '),(6429,3287144,'13,June,2004','2004-06-13 00:00:00','   So what happened today? I worked a lot and that\'s been about it, I have secured about Â£110 for doing up my room and that\'s going to commence on Friday when my cold blue room turns to a dark red lair with  urlLink bats on the wall . I shall be envied!  Well I still have little time to write here, just a few notes; I am still off finding myself a girlfriend, itâ€™s all too much hassle really at the moment and I am still addicted to the deadly concoction of concentrated fruit drinks warmed to lava-like temperatures. But in all the mayhem, I did have the following thought;  Squirrels. So the grey variety are rather detested but they do have a few trials in their life as I noted this morning that manufactures of bird feeders often make squirrel proof accessories, most commonly \'THE DOME\'.  [see this link  urlLink http://www.duncraft.com/prodimg/1092_200.jpg ]  Anyway, little squirrel comes along, asses the situation - can\'t get at it from anyway than via the top so he climbs a tree, gets on the dome and thwump, he\'s fallen to the ground and he can\'t stand on it. Now we must think how they test them, they probably have a garden where they trial the new bird feeders, now think of the poor resident squirrel, each month or so a new feeder comes along, and each time he has no such luck, that squirrel is going to be getting mighty paranoid, he\'s probably scared of picking up little berries in fear that there is a grand piano suspended above it ready to be dropped on him. Maybe not. Well it looks like I am ranting which means something is bothering me deep down.  Ah well, another long post will arrive on Thursday and then every day after that if your good boys and girls. '),(6430,3287144,'12,June,2004','2004-06-12 00:00:00','   I wrote it out, clicked post and it was lost so today you won\'t be getting a posting. Blame blogger. '),(6431,3287144,'11,June,2004','2004-06-11 00:00:00','   Well I was going to go out tonight; Me: Are you still going tonight? Her: What? Me: You know, tonight? It is this Friday? Her: Oh... No not this one Me: Well see you next Friday there then Her: It\'s not next Friday. Its a month next Friday. Me: Oh. Ah well, time isn\'t my game. Speaking of which I found a nice short break in to which I wrote a really long posting but then I suffered a power cut so instead you get this; a short one. '),(6432,3287144,'10,June,2004','2004-06-10 00:00:00','   Tell me I\'m not dreaming but are we out of time? Yes. I have no time to blog. Again. '),(6433,3287144,'09,June,2004','2004-06-09 00:00:00','   Went in to town today for a pork with stuffing and apple sauce baguette - tasty! Anyway not much to mention other than I had a rather interesting dream last night; I sit with a few friends on a bench next to a man; the man is of 60 years of age and is very thin, his clothes hang off his frail body and his skin is unnaturally white. While I am sat next to him I am conscious that he is evil and planning on stealing something off me so I try to get up and move away except I can\'t. Stuck fast I wait and eventually he gets up and I stand up at the same time, he then goes to embrace me and I reach to protect him from stealing my watch, I hold on tightly to it so that eventually he realises he can\'t get it and subsequently walks off. I then realise that he was Father Time. Looks like this was just a short one. Again. '),(6434,3287144,'08,June,2004','2004-06-08 00:00:00','   Recently I have felt that my postings have got rather dull, well I think I can blame this on the GCSEs (I\'m blaming anything I can on them), so don\'t worry, less than a month left. Wow. Less than a month left. I had not thought of it like that. To be precise its less than two weeks then a history exam a week after, but thats not too much.  Well with that little excuse I\'ll be off for some sleep. It\'s still too hot these days. '),(6435,3287144,'07,June,2004','2004-06-07 00:00:00','   This applies to everything involving chairs. From trains to dinning, we often have to book a chair, a chair is then allotted to us and at the right time we use it. So why on earth can\'t people realise this when it comes to sitting down?  Today I had my geography exam [the irony here is amazing] in which I found someone where I was meant to be sat, after a brief silent argument [mostly hand gestures] they refused to move so I was forced in to sitting else where. So in geography exam they couldn\'t navigate to their seat. What is the world coming to? Short of putting up neon signs pointing out which seat is yours, there isn\'t much else people can do other than say which seat is yours and make sure you know it. If youâ€™re on a train, you can SEE the reserved ticket things on the seat. Why, oh why can\'t people SEE them?  That said geography didn\'t go too badly. Well Thursday is when my next exam so tonight has been my little \'break\' as it were - basically wasting time listening to Talking Heads while complaining to friends about the heat - which is unbearable. REALLY awful. Just think, global warming will give us another ice age, or alternatively we could have a more Mediterranean style climate. Frankly things are hot enough as they are - and too cold in winter so lets not contribute to global warming folks. '),(6436,3287144,'06,June,2004','2004-06-06 00:00:00','   Well my computer slow down problem was solved by... Microsoft [or Micro$oft if you don\'t like them], and it wasn\'t really solved by them but their lovely \'System Restore\' feature. Anyway...  Today I got my hands on a copy of  Practical Photographer  which is a very interesting read, quite inspiring to me really... I just need to get a camera.  Among my revision I also had to happen upon some bhangra songs, they\'re not that bad to be honest, not my thing, but most certainly listenable.  My apologies for this being like a list in awfully short paragraphs, so to end that I have one of my dreams to write up; I find myself as a soldier in a kind of marsh area which we are attacking, after a bitter fight forward I board one amphibious vehicle and advance forward where after climbing what looks to be an embankment, the vehicle becomes a small yellow boat.  I find myself the commander of a fleet of similar boats and we advance on the retreating enemy, about half way across this ocean of sorts, I predicted that they would try to counter attack. In preparation I drew up the rest of my fleet and readied to defend which we successfully did, now they retreated with no hope of winning so I paddle furiously with my hands to another similar dinghy, on getting close the solitary sailor jumps over board and one of my men jumps on to replace him [where he came from I do not know], we then paddle back with the extra boat. Well just another day with just another dream. '),(6437,3287144,'05,June,2004','2004-06-05 00:00:00','   Well today I thought I might do something worthy of note - unlike yesterday - so I went on a little expedition to the book shop and bought a copy of Frankenstein by Mary Shelley as I had finished reading Dracula yesterday. While in town I met some \'friends\' - the sort that are alright really just they can be an embarrassment to themselves at times (if you are one of those people... well lets just hope your not).  Also I had to cook myself lunch and dinner, and on both occasions I cooked twice as much as what was needed which has left me feeling somewhat tired. On a final note this weekend is the 60th anniversary of D-Day; a day when no one is sure whether they are celebrating or \'commemorating\'. Shame really, anyway, over and out. '),(6438,3287144,'04,June,2004','2004-06-04 00:00:00','   Well I did a bit more work today than usual and to be honest, thatâ€™s been about all there has been to my day. I hate doing short postings but this is going to be one. Good Night! '),(6439,3287144,'03,June,2004','2004-06-03 00:00:00','   I have had one of those \'its-a-crap-day-and-I-don\'t-know-why\' days. I mean I got to the bottom of my problem with my computer running slowly - Norton Anti Virus apparently but it has also been so tiresome on me; after a number of things going wrong after my interference I sought consolidation in a friend, I use the term loosely as I have my suspicions to whether she likes me, and in keeping with my suspicions she was as cold as ever toward me.  So what good has come of today? I did minimal work and so now feel naturally unsatisfied with myself and I have a cold and everyoneâ€™s being mean to me and... and... *Cries*. Seriously for a moment though, I always feel so much younger in the sense of being \'immature\' when I feel low like this, I don\'t know why but I do - perhaps it is because when we are children we have more moments when we do feel so insecure and so our experience of being insecure comes mostly from a time when we were immature hence we go back to that immature nature when we are insecure. Maybe. '),(6440,3287144,'02,June,2004','2004-06-02 00:00:00','   Well 3 decisions in 24 fours;  Not buy a bass guitar - buy an acoustic  Buy an SLR camera [the expensive sort]  Get a hair cut   Quite a decisive day really, the second idea stems from the first as with the money saved I could buy a nice camera; take some artistic style photos and post them here for all of your viewing pleasure!  And the hair. Well, quite genuinely it is big but certainly not clever; perhaps if I let it grow longer it might be but it\'s doubtful and it is a mega mess at the moment so best just get it cut.  Also I had a rather long and complex dream last night, consisting of two parts;   Part 1  I am with a group of people I know touring an old tower, at the very top there is the garrison of oversized  urlLink die cast figurines , one of them rides around the room on his horse and eventually we leave. Once outside I find out that they are evil and are currently under siege so I help in destroying them; they are going to come out and attack us and so in preparation we form a line and pass eggs down it until it gets to the final group of people of which I am one, smash them on the ground. I smash two eggs and when they hit the ground they turn out to be made of glass and are hollow, I fear that I might get in trouble for breaking them so I go in to the growing crowd.  As expected the door opens and the besieged come out but instead of being the model soldiers they are tourists who come out with their hands above their heads and carefully walk across the glass and are arrested. As they march out I see one of them dressed up impersonating my dead grandfather. They are evil.   Part 2  I am alone with a friend on the top of a mountain at an abandoned picnic area. Between one peak and another of this mountain there is a Chinese temple which, my friend tells me, contains lions and goes down there to tame the lions. I instead look around and find my father, together we get in a cable car and after calling someone it begins to move quite normally, except when it gets to the first support it shakes violently and stops, after a brief pause it continues but on each support it shakes even more violently, so much so that bottom and sides disappear and I have to hold on to a pole connected to the ceiling. During this life-threatening voyage I compare my plight to that of a soft toy holding on to a balloon, on hindsight I think of this as being very strange.  Eventually the remains of the cable car arrive at the other station and we walk through a cabin and meet a couple who run the cable car, we find out that we are in a foreign land and must get back to Britain, so we all go in their car, trying to get to a port. At one point we come across a river which we must cross, the driver decides to accelerate but when it hits the river it flips over as I had predicted. Somehow we find ourselves on the road again. On the journey to the port we run out of food and occasionally stop to get more but when it looks like we are running low on money as well, we stop at a van which sells food and the man gets out and while I am in the car some distance away, I can see him, he buys himself a large piece of meat and eats it, my mum [who has now also appeared] drives off leaving him stranded with the meat. We carry on driving until an undefined point where it all stopped. Sorry this posting is so long but I wanted to include it all. One last thing - DON\'T vote for the UK Independence Party. I would list my reasons but this post is already too long so just listen to any debate featuring a representative of the UKIP and you\'ll see why. '),(6441,3287144,'01,June,2004','2004-06-01 00:00:00','   My God! What is this? I\'m awake in time to experience a normal day instead of being nocturnal?! Yes, it is all true, due to some fool waking me up at the unearthly hour at 7am. Personally I didn\'t know of its existence until this morning.  Anyway, unlike yesterday where I was waiting for a call, today I am waiting for a parcel and with a bit of luck it should arrive soon containing  urlLink Joss Stone\'s  album.  As it is very early I can also remember my dream from last night; I have got a job in an office for a couple of weeks and it is my first day, I go to the third floor in the lift with a German, I don\'t know much German but I manage to answer his question of \'how did you get here?\' by saying \'by foot\'. I then go around trying to find my office but realize that I haven\'t a clue where it is and so after a while I decide to take the bus back home, except now I have no shoes so I am stuck in this office. While walking around looking for my shoes a man falls down the stairs and opens a cupboard before dying, there I see my shoes and my army jacket [why I don\'t know] but they are taken from me as part of the death before I can claim them. Anyway yesterday I read that they will be adding more domains; soon Star Trek fans might be able to be it www.geeks.klingon etc, well I had a small brainwave, why not move all those dodgy websites which you find yourself on if you type an address wrong to being at a .con address, that way they get to have the exact address and only those who miss type an \'m\' and put \'n\' will find themselves there [which is probably just me]. I know it is flawed in that no one would want to have a .con as no one would want to go to it, but its a nice idea to thing that all those pyramid schemes and \'manhood\' enlargers could be in one place where no one dare goes. '),(6442,3287144,'31,July,2004','2004-07-31 00:00:00','   Wonderously brillaint news readers! Next Sunday I shall be leaving for Cheltenham, then on the Monday I shall voyage 40 miles futher south to Cribbs Causeway where, with luck I shall meet a very old [well not old but I\'ve known her for a long time] and dear friend.  However tomorrow I am going to Yorkshire which isn\'t so good but hey, you give a little you get a little!  Until tomorrow, Fabius. '),(6443,3287144,'30,July,2004','2004-07-30 00:00:00','   It\'s a centuries old problem - how do you effectively totally clean a paint brush?  Well I have a solution - instead of using a diluted solution of paint brush cleaning fluid, simply use it in all its concentrated glory! Anyway must sleep... more painting tomorrow you see... '),(6444,3287144,'29,July,2004','2004-07-29 00:00:00','   Just before I go off to watch \'Hunt for the Red October\' for the third time, accompanied by a bowl of popcorn and an apple [a refreshing combination I assure you] I would like to point out the following;  ANYONE who steals ANY of my lovely little phrases, WILL be punished - I\'ve got an imitation-me going around ending all his conversations with ciao [which I have done for several years now] who I will be dealing with very soon... '),(6445,3287144,'28,July,2004','2004-07-28 00:00:00','   If practice makes perfect... well how long must I practice? To be fair my fingers are hardening up but otherwise my practice seems to be in vain. Sigh.  And I don\'t have a can of white gloss to be able to paint the skirting board in my room... what is life coming to? '),(6446,3287144,'26,July,2004','2004-07-26 00:00:00','   This is more a test than anything - to see if I have totally messed up my blog... I was trying to get the template I mentioned running - well it sort of worked but I wanted to go back to the original and... well now it won\'t do that... '),(6447,3287144,'25,July,2004','2004-07-25 00:00:00','   Yes, I have a nice little template in the works... hopefully it will be finished sometime this coming week. Tomorrow I will be going on a voyage to \'The Peacock\' [pub] in search of employment so I will tell you how I get on tomorrow... '),(6448,3287144,'24,July,2004','2004-07-24 00:00:00','   I Have A Blister! This is not good for the main reason that it limits my bass playing by a whole finger- my right index finger - which is vital. Oh the shame... Nice day otherwise. '),(6449,3287144,'23,July,2004','2004-07-23 00:00:00','   I\'ve got a bass guitar! I\'ve been attempting to play it all afternoon - slowly improving but my fingers are now very sore from the whole experience... so I\'ll go and do the whole thing again tomorrow!  Well first impressions... it will take a while to learn...  Anyway here is to me! '),(6450,3287144,'22,July,2004','2004-07-22 00:00:00','   While strutting about in my Led Zeppelin t-shirt I had the most unusual experience - walking down a small track - wide enough for only one person -&nbsp; a couple in front of me stopped 10 meters ahead to let me pass which was very kind and so I jogged up so they weren\'t waiting long and as I approached I was heraleded by both of them with compliments for my t-shirt.  Other than that brief episode today has been somewhat uneventful - tomorrow though I might be geting a bass guitar! Whoo! '),(6451,3287144,'21,July,2004','2004-07-21 00:00:00','   Am I cursed to never be satisfied with a hair cut? Todays was meant to leave it fairly short and that was about it... now my hair ends by tapering off in the back and isn\'t that short but seems too short... I\'ll attack it with hair gel tomorrow and hopefully some good may come of it.  Anyway it is late so I bid you good night. '),(6452,3287144,'20,July,2004','2004-07-20 00:00:00','   urlLink      A photo [taken by my JVC GC-A50 camera] while walking past this piece of writing in a busy street and wanted to capture the piece without stopping. Hence it is taken while I was walking so at an angle I have purposely neglected to sort out [the photo was taken on holiday a few years ago in another European country - I won\'t say which as I have seen similar statements in other countries as well as Britain]      So why have I posted a picture of a politically incorrect statement to my blog? Because quite frankly I can - I far from agree with it - my beliefs are quite opposite really but I am not going to get to my political beliefs just yet. So here is my piece; should this statement be removed?  &nbsp;'),(6453,3287144,'19,July,2004','2004-07-19 00:00:00','   Ok so this is what happened between this post and my last post;&nbsp;    Having used up my 15 minutes to the full I turned to leave and found that my party had left for some church leaving me to make my way to the station to meet in 30 minutes so I arrived at the station in 5 minutes and discovered another internet cafÃ© so now I have 20 minutes here! Well to add anything more would be pointless as I will have access to my note book when I come to type up my full voyages on here but I will detail my reading so far;     1984&nbsp;- George Orwell  Gulliver\'s Travels - Jonathan Swift  Bones and Silence -&nbsp;Reginald Hill   All 3 are worthy of reading - The first being the best by far obviously but the other two have seen my through this holiday without a loss of sanity... anyway 3 minutes left and it took an awfully long time to publish last time so I will end this posting... here. '),(6454,3287144,'15,July,2004','2004-07-15 00:00:00','   So I get back tomorrow but for your reading delight your all getting a live posting from a small cyber cafe in Italy! The rest I will write up from my pocket note book and then date them as if I had done them live but this will be the only live one. &nbsp; So Italy... well it\'s nice enough - I\'ll be posting my full conclusions on my last post which will be dated as tomorrow but will be posted the day after tomorrow - well it doesn\'t seem to make much sense now but when it is all done it should all figure. &nbsp; To the present - they do cyber cafes as well as anywhere - a QWERTY keyboard is still a QWERTY keyboard (Unlike in the Czech Republic where is a QWERTZ - the only difference but amazingly quite annoying!). The machine I haveto use is a lovely black Dell which I have become quite fond of during the past 10 minutes... anyway time presses - I only bought 15 minutes and that is disappearing quite rapidly! &nbsp; Ciao to all... Fabius! '),(6455,3287144,'15,July,2004','2004-07-15 00:00:00','   urlLink      Where\'s Wally? Or more precisely, where am I? Well click to see the image larger and try and spot me reading from 92m - answer at the bottom of this post  Today was rather wonderful;    While other members of my party chose to go around a few of Bologna\'s older churches, I found myself a quiet square and sat on a wall, learnt against a pillar I read, during my reading an accordion player came along near by and played various parts of various songs - \'fragments of Somewhere Over the Rainbow\' drifted my way at one point while a pair of old Italian men laid out some crisp white wooden chairs for an outdoor concert while the smells of a local bakery filled the air along with the pleasantly warm sun.'),(6456,3287144,'14,July,2004','2004-07-14 00:00:00','   Well it\'s hot but bearable - especially in an air conditioned art gallery in Parma - most amiable to spend a morning there - even if other members in the Party were rushing about as if it were a supermarket and the aim was to complete it as soon as possible... well ciao! '),(6457,3287144,'13,July,2004','2004-07-13 00:00:00','   If you\'ve ever wondered how to get that smile from a waiter or waitress after giving a tip but your too tight not to give a tip... here\'s how you can get one; pay up and leave a few coins on the table as you finish the coffee round, get up, glance at the tip and smile to a member of staff - when they look away you grab the coins and leave. So was the method of paying of payment for a coffee I had this evening [by the way I do tip generously generally - just today someone else was paying].  So the flight to Italy was fine enough - my first meal was a challenge though - my only phrases of Italian being \'per favore\' [I would like], gratzie [thanks] and ciao [hello/bye]... so what meals did I know in Italian? Well McChicken Nuggets is the same in Italian and from there it was fairly easy if I say so myself. So maybe I might develop my ability tomorrow, maybe not. '),(6458,3287144,'12,July,2004','2004-07-12 00:00:00','   Tommorrow I leave but today I must pack but before I go I will tell you this; don\'t eat turnips which doing the tango. Deadly. '),(6459,3287144,'11,July,2004','2004-07-11 00:00:00','   I have reached the stage of holiday planning where I am having to wear old clothes which I won\'t be wanting to take on holiday so today finds me in typical \'townie\' attire by which I mean I am wearing a Reebok t-shirt - I feel so dirty! Actually I do as an unfortunate accident with a carton of orange juice has left me smelling of the stuff quite badly of the stuff.  Anyway I require an early night so I can wake up early for work tomorrow! Away! '),(6460,3287144,'10,July,2004','2004-07-10 00:00:00','   Well instead of going to Chester I stayed at home and watched Sherlock Holmes DVDs while drinking ginger beers... days not working for me are not existing.  In other news I dug up a O/S map remember a dream I had many years ago. I look out of my bed room window and realize that the true love on my life lies out in that very direction. So I go off to find her - I can\'t see it from my window but I soon discover that to reach her I must cross some very steep hills but eventually I find her in a small one roomed hovel and we live happily ever after. Now this dream occurred 5 years ago I estimate but it holds significance for two reasons;  -5 degrees in the direction of my window lies the house of an ex-girlfriend [who happens to live in a bungalow [the closest to a hovel really in terms of size]... I once cycled to her and discovered the route to involve going up [and down] many hills  +10 degrees in the direction my window lives a close friend who lives very far away and subsequently between us lie many hills  While the second is a very vague coincidence I can\'t help to shake off a feeling on the first one may have a little fate in it... just a shame her title is  ex -girlfriend. '),(6461,3287144,'09,July,2004','2004-07-09 00:00:00','   My second to last day of work and I am struck by the fact that once this is over and I return off holiday I will need to find another job. Shame really - Â£25 a day isn\'t too bad for 5 hours work... but like all things it must end. Tomorrow I think I might go out to Chester... again... '),(6462,3287144,'08,July,2004','2004-07-08 00:00:00','   [I borrowed the layout of this off someone else and filled it in myself...]   If I were a month I would be: November If I were a day of the week I would be: Thursday  If I were a direction I would be: West If I were a piece of furniture I would be: A Bookshelf If I were a historical figure I would be: Harold II If I were a liquid I would be: Mercury If I were a stone, I would be: Limestone If I were a tree, I would be: a Willow If I were a bird, I would be: a swallow If I were a flower/plant, I would be: a rose If I were a kind of weather, I would be: thunder If I were a musical instrument, I would be: a bass guitar If I were an animal, I would be: a butterfly If I were a color, I would be: Blue If I were an emotion, I would be: Dark If I were a vegetable, I would be: The humble potato If I were a sound, I would be: Feedback If I were an element, I would be: Copper If I were a car, I would be: a Nissan Micra If I were a song, I would be: \'Sympathy for the Devil\' - The Rolling Stones If I were a movie, I would be directed by: Clint Eastwood If I were a book, I would be written by: Mark Twain If I were a food, I would be: a Crunchie bar If I were a place, I would be: Venice, Italy If I were a material, I would be: Velvet If I were a taste, I would be: Bitter If I were a scent, I would be: Woody If I were a word, I would be: Cantankerous If I were an object, I would be: a studded wrist band If I were a body part I would be: an index finger If I were a facial expression I would be: a glare If I were a subject in school I would be: English Literature If I were a cartoon character I would be: Dogbert If I were a shape I would be a: square '),(6463,3287144,'07,July,2004','2004-07-07 00:00:00','   Wow, I\'ve been secretly enjoying my work - doing websites and being paid for it but this time serious money - Â£75 for 3 days work! But it can be quite busy and I\'ve spent most of this evening doing more work for it to add to it tomorrow, I fear I might get it done faster than expected and so receive a pay cut as I will have been working less hours...  Anyway as I have been at a computer all day and evening I\'ll be off now so this is only a short posting... Ciao! '),(6464,3287144,'06,July,2004','2004-07-06 00:00:00','   Well today I was going to commence work at school... But it turns out I couldn\'t so I start tomorrow... So I went off in to town to find another way to spend my time... And instead spent my money on a copy of Commandos 3 and a David Bowie album - both of which have been keeping my occupied through the afternoon.  During a conversation with someone doing similar work to me I found out what they had done... Lets just say there work will be better than mine will ever be and the other persons was far worse than anything I would ever admit to be mine!  Anyway there is a thunderstorm I want to watch so I\'ll be off now... Bye! '),(6465,3287144,'05,July,2004','2004-07-05 00:00:00','   Well while taking in a bit of jazz I\'ll describe today and tomorrow for you;  Waking up a little later than what I should have I was forced to rush putting in my contact lenses and then got to an eye appointment [again] quarter of an hour late. Fortunately though the opticians where quite empty and so my delay wasn\'t of much importance to anyone except myself.  The rest of the day has been idled away but I have decided to go for this job I was offered tomorrow so in preparation for that I have assembled the following; a thermos flask for coffee [in case it is a cold day], a bottle for water [in case it is hot day] and a copy of George Orwell - Nineteen Eighty Four [in case it is a boring day].  Apparently my payment will be of Â£50 so I will be very happy once I get it and I also get to give someone a little present who I otherwise wouldn\'t see for a long time - all in all - it should be okay if all goes well! '),(6466,3287144,'04,July,2004','2004-07-04 00:00:00','   See title for today\'s events. '),(6467,3287144,'03,July,2004','2004-07-03 00:00:00','   A day of boredom in Shropshire; the peak of entertainment came from locating one of those grabber devices for reaching tins off high shelves, using it I set myself various challenges, pouring a glass of 7 Up for example... Truly thrilling and after I could take as much of that as I could I went back to the people...   After hearing the elder residence of Shropshire comment how much I have grown over the years for about five minutes straight I was driven back to my game of pouring 7 Up. What a great day huh?   Anyway that will be all for now. '),(6468,3287144,'02,July,2004','2004-07-02 00:00:00','   Well I was intending to go to Chester today but unfortunately that got called off so I instead went to Stoke and saw Shrek 2 and noted the following;  British people are becoming more American.  Yes that was all but its quite annoying, applauding a film when it starts... why? Who is going to hear your appreciation and be happy about it... the projectionist? Otherwise it wasn\'t so bad, a song by Counting Crows opening but then it was mostly covers of other songs - Ever Fallen In Love (With Someone You Shouldn\'t \'ve?) by The Buzzcocks for example.  In other news I have been commissioned to work on another website! Yes, from doing websites on genealogy, to selling grass seeds my next project is for... my old secondary school! Well that\'s all I know as someone took the call for me and said I would be here so they should call then... but guess what? I\'m not here tomorrow! So now I have to figure out how to answer a call when I am not there to take it.  Ideas? '),(6469,3287144,'01,July,2004','2004-07-01 00:00:00','   Of the few PC games I do play these days, I only ever play Real Time Strategies [or RTS for those who like abbrevations]; \'Rome wasn\'t built in a day\' is a great expression with which to describe my evening playing a LAN deathmatch, after a couple of hours of building up a super economy [yes, it is unnecessary but I always want to have maxed out stuff on everything] I then attack and have my \'Grand Armee\' kill pretty much anything with two [or four if there are horses] legs which leaves me using cannons to slowly destroy everything made of mortar which they own.  So yes, I wasted today. But I might join some work at home scheme... sounds like a scam though... '),(6470,3287144,'07,August,2004','2004-08-07 00:00:00','   Apparently my friend can\'t make in on Monday now - her parents said no so what money she had saved went on a pair of trainers instead. I feel so utterly lost - most of my friends are away on holiday as well which leaves me a skeleton crew with which to go running to with my problems. I might have to rename my blog \'The Fall and Fall of Mr Stephen Pearson\' from the \'Misadventures\' - but the later also seems appropriate right now.  I was looking forward to seeing her so much it has come as a stronger blow than what it should have - but then again it does mean I am going to be stuck around Cheltenham with nothing to do but contemplate it over the coming days. '),(6471,3287144,'06,August,2004','2004-08-06 00:00:00','   This week really is starting to get me down - all I did today was sit about and watch films; an appuling horror film and \'Trojan War\' - a film following a guys attempts and getting condoms - not  the  Trojan War. Nothing much to say... '),(6472,3287144,'05,August,2004','2004-08-05 00:00:00','   Another day went by and another day was wasted; I did watch the 1939 version of The Hunchback of NotrÃ© Dame which did provide a short window of pleasure. At least I get to look forward to Monday when I meet up with my friend. '),(6473,3287144,'04,August,2004','2004-08-04 00:00:00','   There is one thing I don\'t like about holidays; a lack of routine. While many people dislike the idea of a routine, I quite like it, maybe not a strict one but some general idea as otherwise I find myself wasting every moment of every day doing silly things which don\'t benefit me in anyway, I might as well be drinking poison for the times like these. '),(6474,3287144,'03,August,2004','2004-08-03 00:00:00','   Several things about today;   I\'m thinking of trying my hand at pen related art - well cartoonist and typography to be precise... just an idea at the moment  I\'ve decided I need a second chance with everyone  I need to do something about my hair  I spend an absurd amount of money on various things I don\'t need   Well that\'s all really, I don\'t have a job still and I don\'t have anything to do. At all. '),(6475,3287144,'02,August,2004','2004-08-02 00:00:00','   This is being writen on the 3rd despite my ability to hide it by post dating using blogger... anyway the point is yesterday I decided I would write up the days event the next day, except now I can\'t remember what happened. Except someone played a game of cricket and I learnt a couple new songs for the bass guitar. '),(6476,3287144,'01,August,2004','2004-08-01 00:00:00','   After yesterdays flourish of excitement, today was in grim contrast - a visit to Yorkshire being the major culprit - I do like Yorkshire, the people, the landscape but some times it can seem so awfully grim and today was one such day - a blanket of fog greeted us in the morning and a blanket of clouds saw us away in the evening and in between it was one of those hot but not hot days - hot but without the fun...  Well you may be able to understand what I mean, you might not... Either way it was an exceedingly dull day. '),(6477,828778,'22,May,2004','2004-05-22 00:00:00','   You mean the Massachusetts Supreme Judicial Court didn\'t see this coming?    Mere days after gay marriages started happening here in Massachusetts,  urlLink a wife filed suit  against the medics who treated  her  wife, under a \"loss of consortium\" statute after she developed breast cancer (To review, wife A sues clinic because of alleged malpractice which happened to wife B). This shouldn\'t surprise anyone in the United States: any substantial change to existing law yields lots of tort law cases. For the un-initiated tort law = I sue you for money because I can, regardless of the validity of my case. If I can convince a jury you \"hurt me\", you  gots  to pay. Tort law is good when you get screwed by a guy who cheated you in a contract. It\'s bad when the woman down the street sues you because your kid plays basketball and the noise causes her \"emotional distress.\"  Now I\'m going to lay out my position on gay marriage after what I hope are some fair historical basics. Early in the history of Western Civilization marriage was solely conferred by the Church. Marriage carried with it several \"common law\" restrictions, priviledges, etc. Over time, after the United States was founded and freedom of religion became pervasive, the Church was pushed out of the civil sphere and governments began to confer many different kinds of civil priviledges on a traditionally religious institution. Ergo, not too long ago you applied for a license to get married with the state and then got married by a preacher/religious figure of some kind. The church wouldn\'t recognize your marriage if it wasn\'t done by a priest and the state wouldn\'t recognize your marriage unless it was licensed by the state. However, at some point in our society my guess being the late \'60s early \'70s, the church lost even more of its domain over marriage and people began to be married only by civil judges. A marriage might have lost some of its religious legitimacy, but it was still a marriage in the eye of state and in the eye of the participants. To review: marriage went from purely religious, to both religious and civil, back to mostly civil.  At about the same time, people in the United States began to feel that life was about fullfilling the self. A career to pursue is one that makes you happy, doing some drugs are ok as long as I don\'t hurt anyone, essentially anything that you do that doesn\'t hurt anybody else is ok. So, after the queer community received its social exceptance in the mid-90s, gay people felt like being discriminated against simply being who they are, should be illegal. As long as they hurt no one, they should have the full rights of everyone else. And rightly so, according the U.S. Constitution. However, the big mess got started when the queer community felt it was entitled to marriage like everyone else. It\'s a big mess because marriage in the U.S. is at the rare nexus of both civil and religious authority. Marriage is historically a religious or personal institution, but because of our history it has civil priviledges. I say priviledges because marriage is not a right. Voting is not a right. Free speech is a right. Government is  very limited  in how it can restrict speech, but the state can and has restricted driving, veterans rights, voting, and who can and under what conditions can get married. First cousins can not be married, under-aged youth can not be married, etc. Yes, this reasoning applies to historical but eventually illegal restrictions on inter-racial marriage, but the point here is that marriage has always been a heavily regulated civil priviledge--not a right as many pro-gay marriage advocates argue. But since pro-gay marriage advocates follow the philosophy of self-fullfillment, \"since I grew up dreaming of the day I would fall in love and marry, and because this hurts no one, this should be legal regardless of who I want to marry and love, be they man or woman.\"  And I do not dissagree with them. Gays and Lesbians should be able to marry whomever they like. They should be able to enter into the civil contract that is marriage. Despite the fears of many on the religious Right, this does not necessarily destroy traditional marriage. I also dissagree that gay-marriage puts a big crack in this traditional institution. Let\'s be honest, marriage has already lost most of its gravitas when you could get married in Vegas in a drive-through or when you could get divorced for almost any reason whatsoever as long as it was \"irreconcileable.\" The gay marriage conflict exists because anti-gay marriage advocates believe marriage to be the domain of churches and religion(s) many of whom proscribe against homosexuality. Yet, obviously marriage has existed outside the churches for decades. However, gay-marriage advocates fail to recognize the pandoras\' box they\'ve opened. Marriage law never recognized \"love\" or \"want\" as part of the marriage contract, for the most part only a bride and a groom of age. For gay-marriage advocates they\'ve only argued in favor of homosexual couples be able to recieve the priviledges and protections of civil marriage. But now marriage has lost its gender-contraints and indeed all others since the Mass. Supreme Judicial Court decision has set a precedent of only \"want\" to get married. Two heterosexual widows can now get married to merely enjoy the civil priviledges of marriage. A brother and sister can get married. I can marry my best friend so that we can share dental-insurance coverage or to inherit his college of vindage video games.  Ultimately, as far as U.S. law is concerned, I\'m in favor of gay couples having the priviledges that straight couples enjoy, but that is becaue I see how civil marriage should be seperated from religious marriage and that good families come in many non-traditional forms. As far as social consequences and social policy is concerned, however, I must argue that there  must be  real reasons for why men and women were the traditional marriage  even in cultures that accepted, condoned, and promoted homosexual relationships , such as Ancient Greece. Most if not all successful civilizations in history have employed the female-male pair-bond/marriage paradigm for the production and development of human offspring. (Yes, a rare few have been polygamous or polyandrous and here I define \"successful\" civilizations as having lasted long enough for there to be a significant historical/archeological record.) Lastly, I have grave reservations for the method by which gay-marriage advocates have won their victory. Unlike the civil rights activists of the 60\'s that they desperately compare themselves to, a judicial decision was not followed or preceeded by a winning-over of the majority of the electorate or the Congress.  Brown v. Board of Education  was a landmark case against majority opinion, but it was followed by the Civil Rights Act of \'64. The argument that blacks and whites should be equal under the law gained momentum and won over the masses. So far, gay-marriage has had its support level off or wain in the polls. Lastly, unlike human skin tones, there is a  substantive difference  between the union of two men or two women to that of a man or a woman; only the latter can physically produce a healthy human offspring (with out the aide of modern science.) My dissapproval of gay-marriage as it has been achieved so far stems from the fact that they won over 4 non-elected judges. Not a majority of the elected representatives of Massachusetts, and certainly not the majority of Massachusetts citizens. For such a controversial issue, I think we should have the most Constitutionally legitimate source as the fount of social resolution: if not the electorate as a whole, at least the legislature. Sadly in making their decision, the slim majority of the Supreme Judicial Court in Massachusetts has established itself as a tyrannical body writing rather than adjudicating the law; and as a result the gay-marriage lobby has lost the democratic high-ground.  '),(6478,828778,'21,May,2004','2004-05-21 00:00:00','   urlLink    My friend Mike and I tailgated for the Patriots\' Day Game against the Yankees.&nbsp; urlLink  '),(6479,828778,'21,May,2004','2004-05-21 00:00:00','   Another insight into my feelings on the Red Sox, taken from an email   I don\'t have any specific plans to be in Hanover this august, indeed my only plans are to continue the aforementioned training and watch nearly every single red sox game from now until late october. (I\'ve watched 28/30 so far) I\'ll put it this way: should the red sox win the world series this october---and they just might---my death in marine corps fatigues in some jungle or desert a year from now won\'t be as tragic as it might be, as far as I\'m concerned.  '),(6480,828778,'21,May,2004','2004-05-21 00:00:00','   I\'m trying this again.   Where to begin? Right now I\'m 3 sheets to the wind having spent some quality time with my friend Mr. John Daniel, watching Fight Club. That\'s probably as good an insight as any into my life.  A little less than a year ago, I graduated from college. Here I sit, nearly a year later and I have very little to show for time I\'ve spent since. What I do have is a goal: I\'m planning on getting a commission in the Marine Corps. What\'s funny is that the last time I (tried to) start(ed) a blog, I was applying for a commission in the U.S. Navy. This is the point where I tell you that despite my life-time repudation to the contrary, I really screwed up while I was at college. I managed to get  terribly mediocre grades . Let that be a lesson to you y\'all who are still in school and can do something about your immediate futures.   This reminds me of a story.  It was Spring Break 2002 and I was in Houston Texas, staying at the home of a retired alumn. Over many a glass of scotch and bourbon, of which my host did not partake, Mr. Rice told me how he spent his time in Hanover mostly drunk, singing, and chasing tail. Having graduated with a \"gentleman\'s C\" he went to a large state law school, practiced and when I met him had a huge home, a lovely wife, and now consulted part-time for businesses and his old lawfirm.  I guess what I\'m trying to say is that I graduated with a \"gentleman\'s B-\" and I\'m not going to let that stop me from getting on with my life. That\'s where the marine corps comes in. There, I\'m hoping I can recoup my professional competitiveness and discipline. Yes, I\'m risking my life. Yes, there\'s a chance I can be killed. Yes, I\'m an only child and a soft one at that. But as you\'ll learn, my ideas of good foreign policy and manhood are more in line with the heroes of Republican Rome than they are with your typical college kid. And so we\'ve come full circle. What I did manage to retain at college only confirms my desire to serve.  So, here we go: while I juggle long-term goals of fullfillment, I try to get some sort of short-term foothold on life. Part of me just wants to move the F**** out of my house. I currently weigh in at 245 lbs. I need to weigh about 200lbs for the Corps. I\'ve lost 35lbs so far. It\'s on.  '),(6481,828778,'26,May,2004','2004-05-26 00:00:00','   It\'s not personal, Sonny. It\'s strictly business.   Words of wisdom from one of my favorite movies. I went to see  Still We Believe: The Boston Red Sox Movie  today with my friend Daina, who I hadn\'t seen since a weekend in Times Square back in early \'99. (That weekend is for another post) I had my fears and reservations about this movie, but I was surprised with how well it was done and how enjoyable it was. Rather than follow the team, the documentary actually followed the fans. Something that isn\'t clear even among my favorite sportstalk guys at  urlLink WEEI . The fan who got the most pub from this movie was \"Angry Bill\" a horribly negative and cynical fat guy, and deservedly so.  He is everything a Red Sox fan should not be . Despite allegedly being a fan, he simply had no faith and was more obsessed with his own suffering and a fatalistic view of the team. Daina and I found ourselves yelling \"kill yourself\" at him when he appeared onscreen. Yet if he is the yin, then the yang is what makes Red Sox fans some of the best sportsfans on the planet. They always hope and have faith that one year will be our year. They don\'t dissappear from fenway during a losing season, or after several, and they care, scrutinize, and argue over every move on and off the field.  The documentary itself was well done by my documentary standards. It was far more organized and far less preachy than the good Mr. Michael Moore\'s more critically acclaimed works--a man doesn\'t fail to appear in his own films to interview his subjects and then cut/edit his film to put his subjects\' comments in a less than positive light. Surely it wasn\'t his politics that gave him the edge at Canness...hmmm...Anyways, director Paul Doyle, Jr. weaved the story of a few \"case-study\" Red Sox fans in with the ups and downs of last season, and showed them as they are, with no editorializing, which I feel is the strength of the documentary as a form of film.  '),(6482,828778,'24,May,2004','2004-05-24 00:00:00','   They Asked for the Storm [UPDATED below]  Over at  urlLink [Mis]adventures in the World of Dating , junior contributor Madeleine and her best friend Michelle composed a SHP (Standard Hookup Protocol). My comments were too long to put there, so I decided to post them here. Here were their \"guidelines\":   PUBLIC SERVICE ANNOUNCEMENT. 1. We don\'t really need to make out for an hour. Stick in already. 2. Foreplay does not need to consist of the vertical approach. You ALL know what I mean. A girl\'s body is not a series of checkpoints. I know what you are all thinking. Mouth - 5 minutes. Move to ears. 2 minutes. Move to boobs. 5 minutes. Proceed to oral sex for a far too short period, become bored, stick it in. We propose a more varied route. 3. Overrated: boobs. Underrated: fingering. 4. Sex does not need to last hours. When we are looking skyward and biting our lips, we are NOT having a good time.  5. Invest in a ceiling with tiles, so we can count them while you try for a new world record for longest shag. We\'ll thank you later. 6. If we have condoms on hand, that does not make us sluts. If we ask you which size condom you need, that MIGHT make us sluts, but don\'t let it keep you up at night. 7. Be funny. Sex is supposed to be fun. Forgive random noises. It\'s your fault for insisting on weird positions. If you like a sterile, serious sexual experience, throw on some Dave Matthews, close your eyes, and crawl on top. Of a virgin. Laugh a little more, we like it. 8. Learn how to unhook a bra. We\'re all adults now, for pete\'s sake. 9. Correct order of clothing removal: shirt, pants, bra, underwear. Don\'t take off our bra while our pants are still on - it makes us feel like shirtless construction workers. Socks, you\'re on your own. Just take them off. 10. Porn. We like it. Jenna Jameson preferred.   Well, first of all, this list seems geared towards the DNWRG (Drunken Night with Random Guy) rather than a date that is gonna go somewhere. If that is the case, then it would follow that 1) If a guy\'s drunk/horny enough, then this shouldn\'t be a problem. If he doesn\'t know he\'s just here for tonight, maybe this info would help him know to cut to the chase 2) A concern with variety of foreplay seems inconsistent with \"stick in already.\" If it\'s more oral sex you want, get a boyfriend who has incentive to work down there 3) Guys are only interested in the sex or your boobs. Fingers are for quality time by yourself. Sorry, this is not going to change. 4) Efforts to make sex last longer are a result of one of two things, either a) social pressure from women constaintly complaining of unsatisfying sex before Joe Random Guy gets his or b) it might be the rare drunken functionally erect penis, but hours of pounding isn\'t going to get things done. In this latter case, it\'s a guys responsiblility to take himself out of the game. But in the former, give your man a heads up if you\'re a 3-poke-and-done kinda girl, or if he needs to scrub in. 5) Like 4, the basic problem here is that this springs from the general complaint women have of all interactions with men. We can\'t read your minds. Communicate with your Random Guy. 6) A random guy you\'re about to have sex with has a hang-up that you have condoms? Frankly, he should be more worried about you  not  having condoms. And forget all the anxiety about being judged.  You\'re sluts, the both of you and it\'s ok.  7) If you want fun sex, hookup with a guy who you\'re into because he\'s funny. You\'re not gonna get fun sex with a guy you brought home because of the size of his wallet, ehem, penis. 8) You can\'t be serious. A man can\'t get your bra undone and this kills the momentum? Again, we\'re men. We have good gross motor skills, not fine. Hell, I even worked at Victoria\'s Secret and I have trouble from time to time. Frankly, if a guy can smoothly undue your bra with one hand, then maybe you should be worried he\'s on the job. 9) All of this concern over the order of clothing removal and \"stick in already\" are horribly inconsistent. If we\'re gonna get straight to the sex, we\'re not going to worry about how you feel as the clothing comes off. 10) What the hell is this? You\'ll have to be more explicit, pardon the pun. How does this work? You stumble into his/your apartment, he\'s fumbling with your bra, and you want him to quickly throw on the tv/computer and put on some porn? How the hell is that standard? Also, Jenna, though the benchmark of the industry, is a bit plastic. Check out Avy Scott.   UPDATE   It has come to my attention that Madeleine/Michelle\'s original SHP was not intented to address only random/drunken hookups but all hookups in general, even with girlfriends. That being the case, allow me to adjust my critique as briefly as possible.   1) This has been addressed enough elsewhere, but I believe, from what has been said, is that the length of making out should be tailored to the individual hookup. Feel out her pace. 2) Any man still operating on this particular schedule needs to move beyond the \"baserunning\" of grade/highschool. Ergo, agreed. 3) I guess y\'all will never understand the draw of boobs, so a compromise: we play with the boobs  while we finger . 4/5) This doesn\'t change, there are reasons for why men try to lengthen the time of the experience, the solution isn\'t a blanket statement but more communication from both parties. 6) I stand by my initial statement vis-a-vis a random or consistent non-relationship hookup, but if a girlfriend has condoms, obviously she isn\'t a slut--she\'s just sharing the responsiblility of avoiding disease/pregnancy. (Also let me be clear: I don\'t use the word slut in any pejorative or judgemental sense. I wasn\'t explicit above, I meant both parties are sluts, ie the guy and the girl. I was certainly a slut in college. My point was one shouldn\'t care about being judged. Any fear of social condemnation or moral guilt should be a sign you\'re not truly comfortable with your behavior--and that isn\'t healthy.) 7) In general, I agree. 8) Let\'s keep it simple. Men should do their best, but please appreciate men\'s lack of fine motor skills and that we simply can\'t compete with a female\'s level of experience in this regard.  We don\'t wear bras . 9) My suspicion is that this seems to be the authors\', not necessarily everygirl\'s, preferred order of clothing removal, because I could envision a hot/passionate encounter where the order of it coming off would be irrelevant, no? Unless Madeleine/Michelle are citing female consensus here that I haven\'t seen. 10) Whether a random girl or my girlfriend wants to see porn during/before/after the hookup, I still think this needs more explaining.  Regarding  Anonymous \' comments, men  should care  about the enjoyment of their girl during hooking up.  My critique was made in the most general sense towards the SHP, not any sort of statement or reflection of my own game, nor statements specifically directed at the authors.  I apologize for not making that clear. I refer you to the line in my original commentary, \"this list seems geared towards the DNWRG (Drunken Night with Random Guy) rather than a date that is gonna go somewhere. If that is the case, then it would follow that ...\" [emphasis added]. My assumption was in error, but my thoughts were based on it. Also, I was not as successful as I would have liked to have been in adding some levity to the discussion. I was hoping the line about a guy stumbling into a room, struggling with a bra and rushing to put porn on would have gotten some smiles and maybe a laugh, but I digress. Lastly, on a personal note, I do not think myself a sex god, never have. I\'m merely a student of the game, always looking to compete and to improve--hence my interest in the subject.   Ultimately, the general thrust, ehem, of my commentary was that this SHP is not standard, nor is it any sort of protocol. It reads more like a list of recommendations and pet peeves from the female (and specifically the authors\') point of view--or worse, a quiz from cosmo. For example, my girlfriend would probably flip out if I put on porn during a hookup (she\'s decidedly far more old-fashioned than some of my previous girlfriends who have broached the subject of porn).   I think I should thank Madeleine/Michelle for getting this ball rolling. A collaberative attempt at a SHP that will address both genders is in the works. I will be working on it with Isa over at  urlLink The Domino Effect . Look for it there in a coming post.  '),(6483,828778,'24,May,2004','2004-05-24 00:00:00','   Another stab at the language police   For those of you who missed it, and for those of you who don\'t know how much I hate the idea that we should censor words, I link to  urlLink this article of wisdom .  '),(6484,828778,'24,May,2004','2004-05-24 00:00:00','   This Queer Eye for the Straight Guy crap has to stop   Why the hell does anything on tv that appeals to women and gay men, builds up a lot of media-hype, has to lay claim to some sort of social phenomenon?? I don\'t know of a single man who watches  Queer Eye  or who watched  Sex in the City  and men in my demographic make up the vast majority of tv-viewership! Why am I subjected to some hyper-gay man pitching me a goldfish cracker?   On a deeper level, straight men  used  to be well-groomed and dressed properly...then there was that social change in the late 60s and 70s when everyone became a cool hippie, much more casual and no one wanted to be a square. There is nothing necessarily gay about being on top of your game as a guy, showing some effort about your appearence, hygiene or your home. Oh, and metrosexual is just another term for an overly vain and effeminate guy---what woman wants a man more vain and obsessed with his appearence than she is?    urlLink THEY ARE CRAB PEOPLE. TASTE LIKE CRAB, TALK LIKE PEOPLE .  What happened to manhood? I\'ll post on this again and in depth later.  '),(6485,828778,'24,May,2004','2004-05-24 00:00:00','   I Hate Fake Boobs   Anybody who knows me knows that I am a fan of the female form, but I\'ve recently come to hate  urlLink fake boobs . Some of you women out there might applaud and some guys might wonder why. Well, in general, I prefer a much more natural, simple look in the fairer sex: less makeup, less fancy hair and I\'m a strong proponent of a healthy body shape. Some of the waifs parading around my old college campus worried me, it was definitely a big problem at two prominent national sororities. Anywho, I hate fake boobs because to even a casual observer, they  never  look real. It\'s quite simple, real boobs are made of soft fatty tissue that is  lobed  and shaped as women develop through adolescence. They get bigger and smaller as women go through natural phases. Fake boobs are static, firm plastic  circular  bladders filled with saline. Also, often the women in hollywood who get fake breasts are girls who are in really good shape since they don\'t have the fat for real breasts, so when they get big fake breasts they don\'t have the skin and other tissue to support the monstrosities they\'ve had tacked onto their chests. Even good fake boob jobs don\'t last because the skin begins to ripple or the fake boobs sag or drift in weird directions. Most importantly, fake boobs are a lie. I don\'t walk around on 3-inch platform shoes despite the fact I know some women might find a taller guy more attractive. Be who you are. Sometimes we shouldn\'t do what science might allow us. Like  urlLink Christina Aquilera .[shudder]  '),(6486,828778,'24,May,2004','2004-05-24 00:00:00','   Is American Crime and Punishment Obsessed or is it just Me?   So tonight, after I finished watching  Law and Order: Criminal Intent  followed by newcomer  Crossing Jordan , I was suddenly struck with my tv-viewing obsession, and frankly, I don\'t think I\'m alone. With all of the hype about reality tv, the crime/punishment drama saturation is getting little attention. The  Law and Order  franchise has 2 successful spinoffs from the original, my favorite being  SVU . I think what I find appealing about  SVU  is the ridiculously vile crimes. Sadistic kid doesn\'t just kill, he kills and collects the skulls of his animal and human victims. In comes Dr. Huang and he tells us that the serial killer wouldn\'t throw away the murder weapon because the knife is the embodiment of his penis, a weapon he later uses to kill his high-society mother who has been sexually abusing him continuously since adolescence; it\'s part freak show, part guilty pleasure. Beyond the three  Law and Orders , there\'s  The Shield ,  CSI ,  CSI Miami ,  Crime and Punishment ,  the Practice  (recently cancelled but due to have a spinoff),  JAG  and  Navy NCIS  for the military lovers out there, the senior senator  NYPD Blue ,  Cold Case  and even a  CSI: NY  in the works for next fall.  The basic formula for all of these is approximately the same, some mix of detective work and lawyering: a mysterious/sordid crime leaves evidence that requires some forensic science and then the prosecution and defense lawyers duel it out, a few red herrings or plot twists are tossed in for variety. Flavor with type of crime, or focus on lawyers or cops, an interesting and diverse ensemble cast and BINGO you have a show.  We can\'t get enough of it. As a side effect of all of this, though, I find cop-movies laughable. They don\'t use warrants, rarely give suspects due process, shoot first all the time, totally unbelievable.  Anywho, my current favorite top 5:  1. Law and Order: SVU  2. JAG (Catherine Bell...and a throwback to a less cynical world of \"character\" and \"honor\") 3. Law and Order: Criminal Intent 4. the practice (had staying power despite changes in the ensemble and the final episodes were hysterically funny with \"denny crane\") 5. Crossing Jordan (set in Boston, good combo of forensics and copwork)  '),(6487,828778,'04,June,2004','2004-06-04 00:00:00','   This is your daily \"person of color\" blog   So while scanning nationalreview.com, I came across  urlLink this article by Catherine Seipp  about a panel of bloggers held in LA aimed at representing most of the political spectrum and so naturally, criticism came in this form, \"Never mind the progressives on that panel; how many bloggers of color did you have?\" one of [her] commentators huffed. \"Didn\'t think so.\"  My response varies somewhere between a sigh and jumping up and down yelling with anger. This is just another example of what goes on in all other media and especially on college campuses. Essentially, it is assumed that people of color have different points of view to offer than do people who aren\'t \"of color.\" (aka white people and in some camps, asian people) Now, I don\'t have much of a problem with this position. I don\'t sunburn as much as my pale friends and maybe I got a few more suspicious looks in stores when I was 14. Wouldn\'t it be nice if this is where it stopped?   But it never does. When any people of color do not toe the ideological/political line they\'re  supposed  to toe, the charges of \"betraying their people\" or they\'ve negated their color or worse (house n*gger) come flying in from haters on the left. Colin Powell, Condy Rice, Clarence Thomas are essentially branded as tools of the right-wing. Their achievement is cheapened because they are suspected of receiving favor from \"the man\". (of course achievement derived from affirmative action isn\'t cheapened...oh wait) It\'s as if they\'re not black because they\'re conservative. Black people aren\'t allowed to be conservative? The idiocy here is only exacerbated when applied to the Latino community which includes both traditionally conservative and liberal blocs. The horrible crime being perpetrated here is based not on the assumption above, but on the assumption that belonging to a racial group  requires  that one have certain political/social beliefs. This is preposterous and repugnant. My skin tone does not require me to believe in anything, moreover it is only a component part of my individual identity (those of you who have read my blog since its inception know that I\'m more obviously focused on other things) Race should not be a proxy for thought. Continuing to insist that people must think of themselves as parts or spokesmen of racial groups only serves to undermine the progress we\'ve made in judging the content of character.  '),(6488,828778,'03,June,2004','2004-06-03 00:00:00','   Whence Wah Hoo Wah?   Some of you who read this might be wondering how or why I chose this title for my blog. Ok, some of you aren\'t at all. But I wanted to be absolutely clear. Some of you out there just know it as some verbotten phrase, the reasons for which are vague but unquestioned. For some well-researched background info check this out  urlLink here . Basically, some well-intentioned students created \"wah hoo wah!\" as a chant to foster school spirit back in the late 1800s. Administrators in the 1970s decided it was \"offensive\" and so it is now pretty darn divisive. For those of you unfamiliar with Hanover, in some circles the old College cheer is tantamount to yelling \"N*gger!\" Something is terribly wrong when made-up nonsense words gain this much value as a result of their detractors\' wishes and with no clear logic to back it up. Yes, Dartmouth failed its mission with regard to Indians during its first century. The irony is that detractors of Dartmouth\'s Indian mascot and other traditions would have objected to what Dartmouth was supposed to teach Indians in the first place: namely to give them a Western education and to Christinize them. Added to this is that several publications (from Sports Illustrated to the Dartmouth Review) have done polling research of Native American populations and most are either apathetic about Indian Mascots or support them. The differences don\'t even vary when viewed as seperate populations that live on reservations and those who do not. It is a super-vocal minority and their non-native liberal allies who make all the noise on college campuses and in front of  urlLink Fedex Field . What cooks my goose even more is that many of this folks aren\'t sportsfans. For example, they don\'t really understand my relationship and feelings to the symbol of two red socks hanging side by side. There isn\'t a hierchy between me and the socks. They don\'t serve me and they certainly aren\'t supposed to represent or affect the self-esteem of socks everywhere. Those socks represent a team I follow as passionately as any nationalist of the 20th century. They mean I hate New York City and its team. They mean I live and die by everygame they play. I  love  my socks in red.  So, fast forward to this century and school spirit towards the teams on the field is often apathetic, football crowds don\'t fill Memorial Field and a team has to be making a lot of noise with wins to get noticed. Dartmouth has such a proud athletic tradition and its students are essentially fair-weather fans. Big Green? Gimme a break. No one gets behind a color, albeit a bigger one than average.  My blog is named Wah Hoo Wah because those words mean to me a stab at the eye of intellectual sloth, political censorship, and of course love for my dear old Dartmouth; of days when its teams challenged for national championships (think football games against Army and Notre Dame, baseball, basketball and hockey) and there was pride in our institution for more than what it does for your career or academic reputation.   Come stand up men and shout for Dartmouth. Cheer when the team in green appears. For naught avails the strength of Harvard, When they hear our mighty cheer: WAH HOO WAH HOO WAH!  Fight, fight, fight for Dartmouth, And tear on down the field. Touchdown, Touchdown, Dartmouth, For the old Crimson\'s strength must yield.  '),(6489,828778,'02,June,2004','2004-06-02 00:00:00','   Neglected, Confused, Betrayed?   Dear Blog, I haven\'t forgotten you. A summary of the past few days:  Saturday Night: Party at Alexis\'. Fun was had, stayed in Central Square with Jean. Sunday: I was late to picke up my girlfriend Meghan at Logan in the am, but couldn\'t contain my elation anyways. I probably looked like Homer when he\'s licked a frog or two. This was capped as we watched the Sox win in extras after Schillings No-No bid failed, Foulke blew the save, and then McCarty (McCarty!) jacked a homerun over centerfield to end it. Monday: BBQ\'d for the g/f and family, watched the Sox lose a makeup game. Finally saw  The Big Lebowski ; should have seen in back in the late \'90s. Tuesday: drove a total of 8 hrs to see my uncle at his surprise 60th birthday party.  Meghan leaves tomorrow am. Clearly, this wasn\'t enough after not seeing her for 3 months. These days have just flown by.  '),(6490,828778,'08,June,2004','2004-06-08 00:00:00','   Wisdom on the current NBA finals   From my friend joe:  bearsjoe411: if the lakers get the title, the terrorists win SamAdms03: HAHAHA SamAdms03: well played bearsjoe411: it\'s true.....it shows america to be lazy bearsjoe411: and that laziness pays off  There you have it. Here\'s to hoping the Lakers lose and that Detroit wins it, the team that plays the best defense and looks like they want it more; America\'s Team.  '),(6491,828778,'07,June,2004','2004-06-07 00:00:00','   If the supply of virtue is insufficient in a free society like America, it is almost nonexistent in Islamic societies because coerced virtues are not virtues at all.  â€”Dinesh D\'Souza   Some of you know I\'ve been playing church-leage softball this summer for my church. This has provided me with no end to discussion and reflection. First of all, that it is a church-team seems to be in name only as I\'m one of only 3 church-goers on the team. The other 8 or 9 regular players are all softball ringers, obsessed joe-sixpacks who play for a variety of teams all spring and summer long, buy $300 bats and come to play after installing garage doors or spending their day roofing in the heat. Although I was initially put off by the number of non-church goers on a church team, ironically I guess it serves one interest of the church and the league which is out-reach.  On different level, I\'ve found myself frustrated on the one hand by men who are condescending towards me because I don\'t play as well as they do or hit the ball as long--and yet on the other hand I\'ve also come to enjoy the most redeeming aspect of playing which is the comraderie that has grown between myself and a couple of the guys. Most of the guys enjoy busting my balls and I don\'t mind the majority of it, but part of me bristles at their condescention because they remind me of the jerks I used to play little league with: the quick on the field but slow in the classroom if you know what I mean. I\'m a pretty competitive bastard, but I\'ve never gotten my self-worth from my success in the batter\'s box. To be honest, I also enjoy playing because part of me wants to redeem the chubby, bespectacled 11-year old kid who couldn\'t field a baseball to save his life.  '),(6492,828778,'07,June,2004','2004-06-07 00:00:00','   I scoured the world for more Wah Hoo Wah out there   ...and came across  urlLink this blog by \"Footer\"  who as far I could tell from the text is an Dartmouth Alumn at a law school and apparently a member of Phi Delt. I think he adequately covers a frustration I\'ve had on many an occasion and one I think many of you reading this have probably encountered in your day-to-day lives away from Hanover (or indeed any of you when away from your respective institutions of similar reputation: Yale, Harvard, Stanford, etc).  Essentially, in casual conversation with a new acquaintence, if you come right out mentioning that you went to Dartmouth, it\'s arrogant or snobby...but probably worse if you try to avoid it while telling a college-era story about you and your best friend Joe trying to drink a gallon of milk, knowing full well you\'d be booting and they ask \"where did you go?\", it\'s like you\'ve baited them. Yet it would be sweet to be able to be fired up about Dartmouth like someone who goes to Notre Dame Oklahoma, or USC.   Footer puts it best, \"I guess the answer would be to invite everyone to a night back at [Dartmouth for a night of  our  pong]. They\'d not only see how squalid and disgusting much of the fun was, but the ridiculous extent of the same.\"  '),(6493,828778,'06,June,2004','2004-06-06 00:00:00','   Bling Bling via the Badda-Bing   I have to say I knew  urlLink this  intuitively all along, but thanks to a pair of economists out there, we now know for sure. $50K!?! Holy Cow!  Of course, this means that the inverse is also true...and even better that means when you\'re NOT having sex and doing something else implies opportunity cost. Or something like that.  '),(6494,828778,'06,June,2004','2004-06-06 00:00:00','   Duty  Day   Today, while running into the grocery store for a few quick things, I saw a smallish, elderly man with his wife doing their shopping. I noticed he was wearing a WWII veteran\'s cap and felt immediately and spontaneously compelled to walk right up to him, offer him my hand-shake and say, \"excuse me, Sir, thank you for your service.\" He shook my hand, and thanked me back with a look of surprise. I walked away to get back to my shopping.  Have you thanked a veteran today? How about on Memorial Day? Lately?  I think it is interesting that lately when we talk about John Kerry or Pat Tillman\'s military service we use worlds like courage and sacrifice, especially since most of us think it is inspiring that men with so much to lose (advantages of a yale education, inherited wealth in the former case and lucrative sports career in the latter) would chose to put their lives on the line in military uniform for causes that were and are controversial; the war in Vietnam and the current war on terror lack unified support from everyday Americans.   But what about duty? I think it is more common to use the word duty when discussing, or when they themselves discuss, the hundreds of thousands of American men who served in World War II. Men from all walks of life (jews, christians, italian, japanese, german, english, irish) dropped what they were doing (teaching, pro-baseball, lawyers, doctors, mechanics) after the attack on Pearl Harbor and enlisted. Yes, they had courage. Yes they all made sacrifices and some of them the last full measure--But I believe they were driven mostly by the  duty  to defend their country and to defend freedom around the world from the evil (yes evil!) enemy of Nazi Germany. Maybe WWII vets benefited from a singularly and morally unquestionable moment. There was no \"Why War\" campaign or \"teach-ins\" on college campuses in the spring of \'43: too many students had not returned because they had enlisted. It was stop Germany and Japan or accept them as masters.  It\'s shocking how different it is today. People mock President Bush for using words like \"axis of evil\" or \"evildoers.\" Maybe you find these characterizations the niave thoughts of a dogmatic simpleton, but why do we have to argue that the deliberate, mass-killings of civilians for shock-value and political ends is evil? Is it any less evil than the dileberate, mass-killings of innocents for expediency and racial purity?  Anywho, thank a veteran or current soldier/sailor/marine for their sacrifice sometime. It\'s a rewarding experience.  '),(6495,828778,'19,June,2004','2004-06-19 00:00:00','   F***king Chuck Norris   I\'ve had a fun couple of days house-sitting for my friend Kathy\'s parents and keeping an eye on her 15-year old little brother, Andy. On thursday, we stopped in to catch a little league baseball game. Nobody saw us coming. The parents who go to their son\'s game, zone out and enjoy idle chit-chat with other adults, just mailing it in. We came for the game fired up and decided to root for the home team, the 9-10 year old Astros. We had a blast cheering these kids on, yelling \"quack\" when the \'Stros had runners on and standing up and cheering for full counts, with 2 outs.  On friday night, I took Andy to my softball game. I finally had a solid offensive day, going 3-4, with a double and 2 triples...only more anti-climactic as I only managed to move runners over with 1 out in the bottom of the final inning with a deep fly-out to center. We ended up losing by two runs. The Frank Childhood Redemption tour rolls on monday night.  After the game, we went to dinner where we talked sexual politics and I laid on Andy my collected wisdom on alcohol, like Captain Quint from Jaws. (We\'re gonna need a bigger boat). With kids his age drinking and experimenting younger and younger, I hoped I might answer some of the questions he might have, and help him to avoid making some of the mistakes I made. And as many of you know, I\'ve made my share. At least I avoided telling him about \"boot\'n\'rally\" or \"thunderdome.\"  At midnight, We took in Dodgeball with my best friend the Quan and I\'ll keep my review simple so I won\'t spoil it for y\'all: funny. brief. kept the audience off-balance with fastballs (jokes you see coming) and curveballs (jokes that come out of no-where and catch you lookin\'). Worth the price of admission. Well, sort of, you\'d think at  $9.50 , it would come with complimentary hookers and cocaine...but I digress. OH SH*T. That reminds me, as we walked in the usher was a dead ringer for RICK JAMES. We took some quiet steps into the theatre and started laughing our asses off. We began to exchange quick one-liners, \"What did the five fingers say to the hand? SLAP!\" and \"cocaine is a helluva drug\" when two guys on their way out of the theatre asked us, \"Wait--are you guys talking about the guy who looks like Rick James?\" You can\'t make this stuff up.  '),(6496,828778,'17,June,2004','2004-06-17 00:00:00','   Thoughts while mowing the lawn   Mowing my lawn is a totally underrated experience. I can\'t believe I hated to do it when I was 13. Ok, at 13 I hated pulling the starter cord because I lacked the arm-strength, stepping in my dog\'s poop and being eaten alive by the mosquitos that find harbor underneath our jungle of trees. Nowadays, mowing the lawn only means I focus on steering the tractor, but otherwise have 50 minutes to let my conciousness stream away. I think about girls I crushed on in high school or remember fun times from the past.  Today I realized that a year had passed since I graduated from college and I didn\'t really notice or think about that. (realized that I hadn\'t realized. \"the lesson as always: I\'m an idiot\") June 8th came and went, Dartmouth sent me a \"last chance this fiscal year\" to give money, but I just zombied along like I have for the past 12 months. As the blade kicked up dust everywhere and I swatted strange bugs away from me, it hit me: Somewhere, sometime I lost respect for myself. I guess when you perceive and value your world through a certain lens, you\'re at risk of looking yourself in the mirror and thinking you\'re worthless when you don\'t fit into your schedule/category of success. That wouldn\'t be so bad if it didn\'t rub-off on your family members and friends. Maybe self-deprecating humor is only so funny. Eventually, people might stop thinking its a joke.  There must be some kinda way outa here.  '),(6497,828778,'16,June,2004','2004-06-16 00:00:00','   Reflections    1. For the love of the Game.  I can\'t wait to coach little league baseball. I\'m going to be the best coach ever; I\'m going to buy my kids sweet unis, wooden bats (no way I\'m going to allow metal bats, they\'re dangerous and not old-school), and pizza. I will teach them the fundamentals of the game, but above all to have fun. I\'m gonna be one big goofy kid coaching them and teach them how to get fired up.   2. \'Queer Eye\' Must Die.  I\'ve noticed a cunundrum on the modern man out there: on the one hand, women and others roll their eyes or get sarcastic when a guy does something to affirm his sexuality. It\'s all about a relaxed confidence they say. On the other hand, a guy who doesn\'t avoid making positive mention of his heterosexuality might raise eyebrows or create this annoying  urlLink fear . I say keep it old school: feel comfortable talking about playing/watching sports, but try to avoid mentioning those great shoes you\'re going to buy.  3. The Terrorists Lose  It\'s a clean sweep: stars, history and offense are  out --defense, hard work, and all around teamwork are  in . The Detroit Pistons join the Florida Marlins and my New England Patriots as victors over evil (sloth and entitlement). The big-picture shift will be (hopefully) a market correction on the top salaries. Why pay a Kobe Bryant/Terrell Owens/A-rod $150+ million when you can pay a bunch of guys 60 million and win a title. It\'s a shame the NHL is too late.  '),(6498,828778,'14,June,2004','2004-06-14 00:00:00','   Rears [His, Her, Its] Ugly Head   Over the weekend, a friend I hadn\'t seen in a while asked me if I was going to see  urlLink  Fahrenheit 9/11  ...my answer is filled with apprehension. Once again, Michael Moore reared his ugly head--which as you can tell from the title of this post is one of my favorite expressions. I have a nice mental image of a hulking, dark beast that suddenly shows its even more ominous face, lit by dark sinister eyes and gleaming teeth.   That being said, I have a littany of reservations about seeing this film. First, I\'m fairly certain  it isn\'t a documentary . I am a big fan of the documentary format ever since I started watching basic cable in the early \'90s. Lions, whales, insects and my favorite sharks (esp. Shark Week): hidden cameras documenting elements of nature with British scientists narrating the footage as the animals behaved as they might with no camera present. The main problem with Mr Moore is that he goes in the opposite direction. Having seen  Bowling for Columbine  and  Roger and Me , Moore will probably cut and edit his footage making President Bush look criminally stupid and the aftermath of the 9/11 attack and subsequent wars in Afghanistan and Iraq as the evil imperialist machinations of the big oil/neo-conservative cabal---all narrated with Moore\'s characteristically simple questions that drip with sarcasm and condescention. Now, that might be a valid point of view and Moore\'s point of view, but Moore won\'t be presenting his subject matter in this way. He will use the information, footage and quotes that support his point of view, but this makes it a film of persuation. In earlier eras, this sort of film was called a propaganda film. In essence, Moore makes movies as if he was a man doing research and publishing his findings, method, data and thesis in a journal  urlLink like Science , but instead he\'s just another editorial writer in a major magazine. The most pernicious aspect of this is that the average american who walks into a Michael Moore movie probably doesn\'t realize the need to make this distinction. (You can read Moore  urlLink defend himself here ).  In case I\'ve gotten my point confused, my bone to pick here is that Moore makes movies that represent opinion, whether you agree with his opinion or not--but he represents his opinions heavy handedly as fact. Other good documentary makers tread the line of impartiality much more carefully or at least present some facts and provide alternate theories. Some of you might want to check out  urlLink Biggie and Tupac  by Nick Broomfield. He does some of Moore\'s trade-mark ambush-style interviews, but isn\'t as sarcastic or self-righteous. I think that is what really bothers me most about Moore and his ilk. He\'s so self-righteous, wrapping himself in the cloak of saintly truth-giver, injustice-fighter and matyr. (isn\'t it obvious that gunmakers/GeneralMotors/PresidentBush are nothing but evil and wrong?--but \"they\" don\'t want to let me tell you!!) So, although I might go to see  Fahrenheit 9/11  because Moore\'s film might entertain me, expose me to a different point of view and possibly enrage me, I think Moore illustrates a cancer in the left\'s \"activist\" camp, namely that if the cause is \"right\" then the means are justified: be they propaganda films moonlighting as documentaries or the ridiculous lengths gone to in the crusade against big tobacco.   I also want to address some of the controversy that his film has created. When Disney/Miramax did not want to distribute this film, Moore predictably cried foul. He was a victim of the unholy union of Jeb Bush and Disney\'s interests in Orlando. His allies shrieked, \"censorship! first amendment! freedom of speech!\" Firstly, censorship is by definition an act of restriction by the government or some other organ with authority. A private entity deciding NOT to publish/distribute something is called tact or better judgement. Disney/Miramax did not stop Moore from producing his film, nor would they (or could they) stop him from distributing it on his own. Hey, I\'ve got some pretty crazy ideas about life, it doesn\'t mean if I make a movie about it, a company  has  to distribute it. Disney/Miramax has responsiblilities to its shareholders and arguably to society. Some were critical when Miramax released Kill Bill vol. 1 because its was over-the-top with graphic depictions of violence. Disney could have nixed Kill Bill because of these concerns...indeed its not-uncommon for movie studios to have films re-edited to reach an R-rating or PG-rating. Most companies deliberately avoid overtly supporting any party or candidate in an election year. If Disney were to distribute this film, they would have to fear potentially alienating anyone who finds Moore\'s opinions repugnant (at last count probably somewhere around 49.99% of the US population). Politics today are definitely more contentious and heated than they were a decade ago. Maybe Disney thought it would be irresponsible (both to their shareholders and to society) to mass-distribute a propaganda film chock-full of only one side of the issue. (Now that you\'ve digested some of what I\'ve had to say, watch the trailer  urlLink here )  '),(6499,828778,'24,June,2004','2004-06-24 00:00:00','   Itching like the most minor of sunburns   What I don\'t think has been addressed enough by anyone in the popular media is the philosophical foundations at the root of the conflict between the Right and Left in the body politic. Like nothing since the Civil War, the Right and Left are horribly divided. Have any of you noticed that each side seems to be talking past each other?   The Right believes tax cuts to the private sector have fueled growth because individuals use capital more efficiently to start or invest in businesses which in turn creats jobs. \"Tax cuts for the rich\" is pretty much factually correct, if only because the  poor don\'t pay taxes . The Right believes that there are several macro-economic factors which suggests that the economy is growing. The Right thought in good faith that Iraq had WMD. Dude,  everyone  thought Saddam had WMD. And we have found stuff like Sarin gas and other more mundane stuff, just no nukes or evidence of a big nuke program. That we were wrong about the nuke stuff doesn\'t make it a lie. The Right thought Saddam was part of the bigger war on terror picture, he offered cash payments to the families of Palestinian suicide bombers. The Right was also going by a general foreign policy principle. If the enemy of my enemy is my friend, then both of my enemies are probably (or could soon be) buddying up. A democratic and peaceful Iraq is possible and it\'s worth the lost American soldiers today, rather than the potential loss of thousands or millions of Americans in the future.  The Left believes that the tax cuts have done nothing but line the pockets of the wealthy who have the GOP/GBush in their pocket---especially since the unemployment rate has stayed about the same and even as some jobs have been created, more jobs are lost overseas to outsourcing. The Left believes that no American soldiers should be dying in a foreign land for the Oil interests of the energy industry (also in bed with the GOP and the Bush fam). Bush et al knew all along there was no WMD or at the very least maliciously exaggerated the likelihood. There is no direct link between Saddam and Al Qaeda and there never was one.   What\'s my point? It\'s obvious both sides are interpretting more or less the same information from vastly different philosophies. The Right is predominately pro-capitalism, small government, nationalistic, and in favor of changing the status quo in the Middle East, unilaterally if need be. The Left distrusts business, wants the US to engage the rest of the world through multilateral organizations and that there is a way to peace through negatiation and that war is the very, very last option. I don\'t think either side will be able to win over a substantive majority during this national election since it just isn\'t easy to persuade people away from these core beliefs. The current international, healthcare, economic dilemmas are happening at the very nexus of conflict between these two belief sets. For example, the terrorists won\'t negotiate, but they\'re not easily stopped by the hammer of the American military either.  I worry about the future.  '),(6500,828778,'23,June,2004','2004-06-23 00:00:00','   The Current Summary    Torqueedo0: frank Torque0: please save me from my life Torque0: please find a way for us to move to a relatively urban, relatively anglo locale and make millions   And with that, my friends, my buddy JP sums up both our lives in a nutshell. I\'m working on a whole host of different outs. But to change it up I offer a story. My friend Ben and I stood in line for an hour to buy day-of tickets at Fenway Park\'s gate C. Some promo-guy walked up and down the 100-200 person line to give away Schick razors and Sweedish fish. Sometime around minute 48, three Minnesota fans in full jerseys and hats walked by the line and had this exchange about the sometime hostile atmosphere at The Park:  \"Hey Bob, we picked a  great  day to come to fenway...they\'re givin away razors.\"  '),(6501,828778,'21,June,2004','2004-06-21 00:00:00','   Another Top Ten List!   This one will be debated here and in dark-lit tables or bars for a while. This is my TOP TEN RACIAL/ETHNIC SLURS. Calm down for a second. I, of course, have some sensible criteria for this list. For one, the N-word does not make this list. The N-word just comes with WAY too much baggage, and honestly, I don\'t like it. The slurs on this list come from eras past and don\'t have the same level of history of hate and violence. These are some you might hear from some old coot who\'s clearly still living in the 1930s. These are for the most part anachronistic to the point that many of us have forgotten their origin. Oh and they strike me as funny. Right or wrong pathetic ignorance is funny. These made the list not because they\'re not hateful--because directed at the proper people they clearly are, but they don\'t have enough hate-momentum, if you will, to be taken seriously. Yes, there could be some serious dialogue as to why some are funny, some are not, or why you or I find one of these more or less offensive. But that\'s for a different post. Oh yeah. and  absolutely don\'t take this seriously. at all.  Laugh or don\'t laugh. That is all I ask for.   10. Floppy-Head  This is a great image. You gotta hand it to the guys at South Park for inventing a slur for Canadians out of no-where.   9. Goo-Goo  Some of you have no idea what this is. But it\'s a slur from back during the Spanish-American war when the US fought in the Philipenes.  8. Cracka  An off shoot of \"whitebread\" this one has come into more usage in the last 10 years or so.  7. Wetback  Although some might find the idea of a Mexican swimming across the Rio Grande characterized this way kind of demeaning, it\'s a much better turn then the dudes who try to brave the desert further west on the boarder with New Mexico and Arizona.  6. Camel-Jockey  A fun term for our friends the Arabs. My friend Safadi actually has a picture of himself riding on a camel in Syria, but camels aren\'t nearly as common a mode of transport as we think they are. Oh and they spit.  5. Honky  A shot at the poor rednecks who were just bitter that at least poor black folk had an excuse.  4. Gook  Sounds funny. And now it only implies the North Koreans. And they only have enough money and energy to fuel their dictator\'s porn collection.  3. Kraut  Now you get on your hot dog or braut, but at one time we didn\'t like the Gerry\'s very much at all. Fought them twice, then sent over 5% of our GDP so that they\'d be our friends.  2. Gringo  Mi madre dice este casi todos los dias, que me olvido que \"gringo\" is una palabra irrespetuoso a los Americanos. Ella todavÃ­a llama mi papa un gringo cuando el la molesta.  1. WOP  I hadn\'t really heard this one until I was a junior in high school, but it still gives me the giggles. Firstly, it\'s derived from \"with out papers,\" the achronym of which was stamped on incoming Italian immigrants at Ellis Island. Pre-Fascist Italy apparently didn\'t keep good public records and this was a source of ridicule used by other European immigrant groups. You\'ll remember Hollywood Studio Chief Wallstein from the Godfather, \"I don\'t care how many dago, guinea wop greaseballs come crawling out of the woodwork, Johnny Fontane is not getting in that picture.\" Tom Hagen says, \"I\'m not Italian, I\'m German-Irish.\" Wallstein finishes, \"Well listen here, my Kraut-Mic friend, Johnny Fontane is never gets that movie!\"  '),(6502,828778,'20,June,2004','2004-06-20 00:00:00','   Neo: But if you already know, how can I make a choice?  The Oracle: Because you didn\'t come here to make the choice, you\'ve already made it. You\'re here to try to understand  why  you made it. I thought you\'d have figured that out by now.    Has there been any more simpler statement of the dilemma that faces the person who believes in the omniscient, omnipotent God and yet who derive their sense of being from free will?   I\'ve always felt that  Matrix: Reloaded  was super underrated. It combines some decent philosophy with the best \"chase\" sequence ever. I\'m not kidding. Watch it again. Nothing can hold a candle to Morpheus, Trinity, and the Key Maker on the crowded highway running away from those albino-rasta twins, \"human\" police and agents all at the same time. A few great fights and a sprinkle of existential romance with the fate of mankind all in the balance, what\'s not to like??  '),(6503,828778,'03,July,2004','2004-07-03 00:00:00','   The Good, the Bad, and the Academic   I have a natural suspicion of books that become way too popular too quickly, and sometimes that skepticism pays off. I finally read  The Da Vinci Code , and I have a few responses.  Firstly, it\'s a fun book to read. The cliff-hangers and twists are great. But part of me had to wonder at the history behind the novel, since the author went to lengths to state that much of his book was fact. I, subsequently, approached some of his claims and the general thesis/conspiracy theory of his book with a fair degree of skepticism. For most of his more outrageous claims, you can find an answer from a real academic  urlLink here .   I feel like ultimately, his conspiracy theory plays on the expectations of a crucial core of his readers (to get the ball rolling)...that is to say, a great portion of the book review/media people out there are well aware of the feminist critique of much of the Western Literary Canon, so here comes along a book that decries the whole of orthodox Christianity as a patriarchal kabal to surpress the divine feminine. Cloaked in a mystery novel,  The Da Vinci Code  in many respects is telling a certain portion of the U.S. exactly what it wants to hear, serious research or fair assessments of facts be damned (See Michael Moore\'s F9/11 or Bowling for Columbine).  What I think is most pecular about  the Code  phenomenon is that this book falls under the category of  fiction , yet has so many people giving it more consideration it\'s due.  '),(6504,828778,'02,July,2004','2004-07-02 00:00:00','   Just Get The Man His Car Already   While looking for some levity from my life and the redsox, there\'s this story. Apparently Cadillac  urlLink is a bit late  in getting a promised auto to your hero and mine, Tom Brady. Handsome as a devil, married to a supermodel and deliverer of two Superbowl Championships to New England, but apparently even he has to wait for Cady to scare up a roadster.  '),(6505,828778,'30,June,2004','2004-06-30 00:00:00','   Dead Men Walking, Wearing Red Sox   As some of you might already know, Boston got swept in the Bronx by Yankees this week. As I detailed in a blitz to my buddy Poitras, the Sox lost these three in the most painful ways possible.  Game 1: Blow out, frustrated by many errors Game 2: Battled, Held lead and lost on errors Game 3: Close game back and forth into extras, multiple chances at putting it away and even led going into the bottom of the 13th, 2 strikes 2 out and lost. LOST. This team needs viagra AND levitra.  That being said, I\'m so terribly close to becoming less like a youthful optimistic, faithful, hopeful red sox fan and more like a jadded, old-guard, bitter, \"the season is over july 1st\" red sox fan.  On the brighter side of things, I\'m back off the wagon.   Where is  my  Mrs. Lafave???  '),(6506,828778,'30,June,2004','2004-06-30 00:00:00','   Scott Peterson Must Die and Take  Me  To Smoothy King   I don\'t know why  urlLink this  is a national story at all. But have you noticed that the murder case of Laci Peterson (and the trial of accused husband Scott) has had legs for one and a half years already? Aside from the rather gruesome details of the how the remains of Laci and her (un)born fetus were found, there\'s no real reason the national media is following this so intently. It\'s like they decided this was hi-profile and even though they\'re wrong, they won\'t admit it, so they stubbornly follow it up just not to look like a bunch of vulturous goons. As far as I can tell, it\'s just a garden-variety murder mystery where the spouse of the deceased is the lead suspect. Fodder for L&A: CI,  maybe .  Now  urlLink this  is a story that could leap right over the Peterson trial and  should  have the focus of the national media right through the dog days of summer. A 23-year old teacher in Florida is charged with 2 counts of lewd and lascivious battery and 1 count of lewd and lascivious exhibition with a 14-year old student. I know what you\'re thinking, but this is a  female  teacher having sex with a  male  student. Read all the ridiculous details  urlLink here , courtesy of the good people at thesmokinggun.com. Sadly, there won\'t be much coverage, because the name of the teenager cannot be disclosed because he\'s a minor and she isn\'t an NBA basketball star. But c\'mon! there are soooo many angles to investigate here:  1. Was a crime really committed here? This teacher is hot as hell and probably made the kid\'s adolescence. He was bragging only days afterwards and believe me, he won\'t shut up about it for the rest of his life. He\'ll be telling this story in bar-rooms and locker-rooms for decades.  2. On the other hand, although I\'m quick to question the crime status of hot female teacher having sex with teenage male, aren\'t we as a society more inclided to believe that a male teacher having sex with a teenage female is more clearly a crime? 3. And is this inclination just another indictment on the deep-rooted differences between men and women and how we perceive them? 4. Did this teacher want to get caught? For knowing that it was forbidden, she didn\'t really spend a lot of time worried that it be kept a secret.  For me, it\'s summed up in these few lines from the alleged \"victim\'s\" statement:   \"He was sitting on the couch when she sat next to him and asked him if he wanted to have sex. He told her he did. While he was sitting on the couch, she straddled him facing him and and asked him if he was sure. He said yes.\"   Some implications:  1. Her husband must be hurting; not performing well in bed and his wife cheated on him with a 14-year old. 2. Some of you are thinking that this is really sick and how can I not see that the desparity in age means she took advantage of him. Well, I guess part of me remembers being 14 and I sure as hell would have said yes. Don\'t kid yourselves, this is nearly every 14-year old boy\'s DREAM. Also, if he can\'t give consent at 14, does that mean all of the fooling around we did as 8-graders or freshmen in high school was illegal? Or is it merely the fact that one of the parties is above the \"age of consent\" and the other is not that makes this a crime? How arbitrary is this line? Obviously legally he couldn\'t consent, because we wrote the law believing that he couldn\'t appreciate the severity of the question...but read that quote again, man. Dude knew what was up.  Comments appreciated from all viewpoints.  '),(6507,828778,'29,June,2004','2004-06-29 00:00:00','   More on Moore   Now that his documentary has been given its opening weekend in full distribution, I decided to see what comments have surfaced about Fahrenheit 9/11. Some highlights from the IMDB.com that only further my point that Moore\'s film is not a documentary, but rather propaganda (I\'ll include a note whether it was from a positive or negative review):   \"I\'m not a huge Michael Moore fan. He\'s often too extreme and close-minded, but that doesn\'t mean I wasn\'t up and cheering (philosophically speaking) during his controversial Oscar acceptance speech at the 2003 Oscars. I see his point more times than not, and often agree with him. If he makes up facts, so what?\"  (Positive)   \"If you are looking for a fair and measured evaluation of all of the facts you can skip this movie. This movie is an opinion piece and thus one sided. Criticizing it for being one-sided only proves that you miss the point.\"  (Positive)   \"Moore\'s apparent thesis--that Bush and his cronies are bent on world domination--is laughably absurd. Fahrenheit 9/11 is essentially a long string of insane conspiracy theories babbled by a lunatic you\'d normally expect to see somewhere in a Manhattan subway clinging to a sign with a warning about an impending apocalyptic nightmare. The only people who can truly appreciate a movie like this are those who are most susceptible [to] believe lies, distortions, and inaccuracies--that is, the gullible.\"  (Negative)   \"Fahrenheit 9/11 may be one of the only films I\'ve ever seen where the audience - most of them - began prolonged cheering when the director\'s name appeared in the OPENING sign boards. That reflected, of course, the political leaning of an affluent Westchester audience that was all-white and which filled surrounding parking lots with SUVs and luxury cars (my Honda Civic a small disgrace).\"  (Interesting footnote from a positive review)  I guess I\'ll have to go see it sometime soon, but I might have to sneak in. I don\'t like the idea of paying for propaganda. From the Administration or from Moore, I think it should be free. (I know that last comment doesn\'t really further my thesis that Moore made a propaganda film, but it actually ties into a thesis I have regarding the linkage between fuel demande and terrorism...but that\'s for another post.)  '),(6508,828778,'28,June,2004','2004-06-28 00:00:00','   Strugglin\' with the Skin   I\'ve been crapplin with a strange issue lately. I substituted at my old high school at the onset of spring and I couldn\'t help but notice a huge change since I graduated 5 years ago. The young women are increasingly scantily clad. Thongs said \"hello\" and many a navel-piercing was the mid-point between a bust and very low-rise jean. I have no idea how their male counterparts manage to study or focus on a lecturing teacher.   Naturally, the 14-17 year olds are way out in front on this change in American fashion and the rest of us have to deal with the reprocutions. Over the past 5 years I think it\'s fair to say, young women have been steadily wearing less and less clothing. In started in 1999/2000 with the arrival of the f*ck-mepants. Tight and black. Then came ever-increasing numbers of hip-hugging jeans and f*ckmepants in more colors (tan, white, lightblue, pink) and some with contours (north-south ribbing). With the tighter pants also came more demand for thongs under the guise of the \"seamless look.\" Lastly, we got low and lower-rise jeans and pants.  Up here in the more conservative Northeast (by \"conservative,\" I don\'t mean politics at all, I\'m referring to an oddly Puritan-influence on things like dress, last-call, business on sundays and such), I feel there\'s a temptation to see one of these skin-baring young ladies and smugly think to oneself \"slut.\" I\'ve heard older men or women quip, \"they don\'t respect themselves.\" Even my 20-something female peers are not beyond this temptation as I\'ve overheard many a upperclass Dartmouth woman complain about the \"young freshmen skanks who don\'t wear any clothes\" annoying galavanting about fraternity basements.   Yet I can\'t help but feel that this is all unfair or at the very least terribly wrong-headed. First of all, as many would point out, a woman\'s dress says very little about her proclivity towards promiscious behavior. That is, just because she\'s wearing pants low and tight enough for a man to stop-short and double-take, doesn\'t mean she\'ll necessarily let him take those pants off.  Moreover, why do we here in the US (but moreso the Northeast) rail against this pro-skin trend in the first place? Might it be something about the cultural heritage up here in New England and New York? This sort of thing is pretty standard in places like Miami or Southern California.   The female form is certainly something to appreciate for its beauty, but why does seeing more of it upset create such intrigue? Maybe we can just get used to it or at the very least, enjoy the view.  Oh, and a shoutout to my boy Safadi.  '),(6509,828778,'27,June,2004','2004-06-27 00:00:00','   Off the Team, better luck next year oh and let girls play   Softball should be abolished. How do parents still allow this? The attitude that created softball is essentially, \"women can\'t throw a ball hard over-hand, can\'t hit a baseball, can\'t run 90 ft, can\'t hit a ball farther than 200ft.\" I\'m not arguing that America\'s daughters should be playing baseball in co-ed leagues (but what that be all that bad up until about 10yrs old?), but I am saying they could be playing baseball against each other. I think we owe girls that opportunity. Baseball is good, clean, All-American fun. Softball is condescending...except for maybe old, unathletic men, like me and my softball teammates.  Ok, one big problem I have with some \"sports\" out there is that they aren\'t decided on a quantified basis, but a qualified basis. I don\'t want to restrict how we define sport too much, but I think a clear and objective method for deciding the victor of a contest should be a necessary condition. I love art, but when your contest is decided at least partially on your artistic expression rather than how many tallys are recorded, then we have a sufficient condition for  expulsion . For example, football is decided by which team can score the most number of points in the allowed time. Football = sport. Let me be clear, a team scores in only one of two ways. A football crosses the plane of the endzone, or it splits the uprights. No bonus points for style, as much as the NFL dudes would like to think so. (Sidenote: bring back the throat slash!) Pairs figure skating, on the other hand, is decided by the composite of two scores: technical merit for how hard and successful your routine was and artistic merit. Pairs figure skating = not a sport. Simple.  Ergo  On The Docket for Immediate Banishment:   1. Figure Skating.  \"Ooooh. Stuck the landing, but the French judges tend to dislike the more fancy costumes and the Brahms. That\'s gonna hurt their score, Jane.\"  2. Diving.  Spin about, no splash. YAY!  3. Synchronized swimming.  Wasn\'t this formerly a feature of over-the-top 1950s musicals?  4. Gymnastics.  How do we endorse little girls working their bodies into the ground with zero body fat and judging them for how well they were able to not shift their feet after the pain of the impact from their tumbling? Is NOW so busy attacking the fashion mags that they don\'t see damage this is doing? To be fair, these events provide with hilarious unintentional commentary, \"Wow, look at how she cleanly mounts the apparatus.\"  5. Cheerleading  I like tradition as much if not more than the next guy, but cheerleaders no longer even do that much at football games anymore. I\'ve been to countless college games and only rarely do the cheerleaders actually  lead  cheers anymore. Sure, they occasionally clap their hands and shout out the teams name after it\'s been spelled, but that\'s about it. Mostly they do stunts during timeouts or jump up and down when a team scores. What about all of that intense competitive cheerleading, you ask? Again, it\'s judged by a mixture of technical merit, and artistic expression.   Please notice I\'ve stayed away from arguing about the level of physical exersion and whatnot. That is not my aim here. I\'m also not saying that sport and art should be completely seperate. A diving catch by a centerfielder moving quickly to his left is pretty beautiful, as is the expression of joy on the faces of a team after they\'e won the superbowl...but if your event involves the words like \"routine\" or \"judges,\"  chances are it isn\'t a sport. Why you ask? Because then we\'re talking about pieces of flair. And you  do  want to express yourself, don\'t you?  The Nazis had pieces of flair.  '),(6510,828778,'27,June,2004','2004-06-27 00:00:00','   Bibere Veritas   For some time now, I haven\'t liked wine. The long story involves a bitchy-blonde grad student, a bottle of white, and a party at Phi Delt--though not under the hospitality of the brotherhood, but rather the tenants they allowed while they awaited the end of their purgatory, their temporary de-recognition.  That\'s a story for another time. Tonight, I watched  Big Fish  with my parents. My Mom is always a trip to watch a movie with, since the cultural conventions of viewing film differ drastically from Colombia to the U.S. Let\'s just say my Mom doesn\'t observe the \"no talking\" rule. My Dad is usually stoic, which brings me to my point. Big Fish, brought the three of us to tears. My mom first and obvious, sniffles and covering her face, but what was more surprising was my Dad. I looked over at him and the tears had left behind their paths on his cheeks. This isn\'t unprecented, but for y\'all to understand, if my Mom is the wellspring of emotion, my Dad has been for two decades more often than not the direct opposite. I\'ll even admit that towards the very end of the film, even I felt a bit misty-eyed.  I know that movies like  Return of the King  and  Cold Mountain  usually receive more critical and box-office acclaim, but maybe we should wonder how it is that a movie like  Big Fish  manages to sneak under the radar somewhat more or less, yet it clearly appeals to a much broader spectrum of humanity with a more profound message. Is this the nature of \"classics\" which capture the human condition?  '),(6511,828778,'06,July,2004','2004-07-06 00:00:00','   Positively Laughable   While watching  urlLink Nip/Tuck  tonight, I caught an interesting ad. It opened on a street at night with three dudes approaching a car, the driver is about to unlock the vehicle when he stops short with keys in hand. He along with his crew notice this strange violent red feather dancing in the night air. They follow it to a bus, the doors open and they walk into a sophisticated island-themed fiesta.   All very mysterious and intriguing but it was an ad for Captain Morgan\'s Parrot Bay. First off, it seemed like a \"don\'t drink and drive\" spot, then head-and-shoulder faked with the Forrest Gump routine.  Coconut Rum ad aimed at men? They must be joking.  Any man worth his salt does not drink coconut rum.  And that\'s for a whole host of reasons. Authoritative list of booze to come.  '),(6512,828778,'05,July,2004','2004-07-05 00:00:00','   Absolutely Fascinating AND Funny   A happy belated Independence Day to all who read my blog. While bored and following a series of links from blog to blog, I came across  urlLink  True Porn Clerk Stories  .  From what I gather an educated woman in her late 20s/early30s worked at a video store in Chicago for a while and posted about her observations about porn in general and dealing with the people who rent porn. Her writting is pretty funny, fair and intelligent--with some surprising insights:   And as badly distorted as the female porn body can get, that\'s nothing compared to what happens to the men.  Here\'s what the straight boxes have taught me, though: guys are the ones who care about giant cocks. Not women, guys. Especially the straight ones.  But the more I\'ve worked at the video store, the less I\'m convinced that porn is inherently degrading, and the line between degrading and not gets blurrier.  '),(6513,828778,'13,July,2004','2004-07-13 00:00:00','   Some Inspiration   While I take a break from my girlfriend\'s current visit, allow me to apologize for not having time to update with the regularity you all deserve...but do take a moment to read this:  urlLink this guy deserves a tip of the cap from all of us .  '),(6514,828778,'23,July,2004','2004-07-23 00:00:00','   PONG     &nbsp;&nbsp;&nbsp;&nbsp;  an authoritative history and guide'),(6515,828778,'21,July,2004','2004-07-21 00:00:00','   cocaine is a helluva drug    Simple computer games. Why are they so addictive? Why do they appeal so strongly to my sense of competition.&nbsp;I\'ll never forget my \"gateway\" game....snood. Freshman fall the Blue Dragon found me.&nbsp;Countless hours of my life were lost to Snood. I\'d get back to the dorm after class and I\'d Snood for a good hour or two. If I came home early on the weekends, I\'d snood it up drunk for a bit before passing out. My two roommates and I waged a war of one-up-manship late into spring term. Snood got us all, though. You couldn\'t escape the draw of those damn dumb faces, smiling so innocently--and the sounds! The blup-blup of shooting them upscreen, the celebratory cord you\'d hear when you cleared up a bunch of the bastards. I studied the angles. I thought of it in terms of simple physics and vectors I never really learned about. I played until my top score was beating that of my&nbsp; son of a bitch&nbsp;roommate \"Chip.\" Yeah, I still despise him. No, I won\'t be letting it go. I will have my&nbsp;vengence on his smug, arrogant ass&nbsp;and snood isn\'t even a snowflake of the iceberg that he earned that\'s&nbsp;drifting slowly his way...   uhhh back to the games   After snood played itself out, I managed to stay clean and sober&nbsp;until Poitras offered me a hit of&nbsp;Miniputt&nbsp;my junior winter. Poitras, Singer and I would blitz back and forth with low rounds....25, 24, 22.....until sometime junior spring I played a 18-hole round of&nbsp;18.  The Perfect Game.  Nothing but holes-in-one. I had it down to a science, aiming the little dots at just the right spot, angle and force to get the right number of bounces before it would rattle into that cup. It has&nbsp;rearred it\'s ugly head again just after an also-ran about a helicopter&nbsp;tried to creep its way into this addict\'s day. (Miniputt is now it\'s listed on an MIT site, different from its original Harvard abode.&nbsp; Play it  urlLink here . The Geeks on the other side of the Charles spend too much time trying to out-do each other. And God bless\'em.)  The helicopter game&nbsp;failed to cut mustard as a truly addictive computer game in the Snood/Miniputt mold because your ablility to build up a score is really just determined by the speed of your computer. My newish laptop had me scoring just over 1000pts (about 1000 pts below the competition)&nbsp;until&nbsp;my laptop&nbsp;was running really slowly one day and I scored over 3000 just like that. C\'mon. (play that one  urlLink here .)&nbsp;Another, the dropping penguin swung at by polar bear for distance had a chance, but it wasn\'t readily apparent how to improve your distance. Top of the strike zone or bottom? All to random for me.  I guess it\'s a subtle balancing act of simple gameplay, accessible venues of competition and simple (flash, mouse)&nbsp;interface that make for a truly addicting computer game. Games like Madden, GTA, etc usually don\'t play well on PC because they can\'t compete with the interface of setups like PS2 and Xbox. But if only involves the point and click of a mouse, simple graphics,&nbsp;and a way for me to&nbsp;compete against&nbsp;my friends, I\'m all over it.  &nbsp;  '),(6516,828778,'19,July,2004','2004-07-19 00:00:00','   I\'m Back!     &nbsp;  Dating my girlfriend long-distance has been an interesting challenge.&nbsp;Certainly weeks of dating someone by phone or email&nbsp;is less than ideal, but spending a solid amount of time with my girlfriend here or&nbsp;at her home&nbsp;Arizona&nbsp;can be a wonderfully sustaining and satisfying experience.     &nbsp;  The first few days are a complete bacchanal:&nbsp;a festival&nbsp;of affection, glutonous meals, lots of celebratory drinking, and a saking of other thirsts (wink, wink). We usually mix in a few fun outings to this tourist attraction or that museum, but I think the most satisfying time is the time we spend doing the alternatively mundane: renting a movie, watching bad cable tv&nbsp;or the Red Sox. Check that--the Red Sox are no where near mundane.    &nbsp;  So after being seperated by thousands of miles and hundreds of days, it was rewarding and heartening to enjoy this time with my girlfriend. Unfortunately, that time came at the expense of you, my darling blog. But not to worry. I have many posts stored away in my head to rant about.    &nbsp;  For&nbsp;starters, I continue to follow-up on Michael Moore.  urlLink Here\'s a cutting point-by-point disection of some of&nbsp;Moore\'s arguments . Essentially,&nbsp;Moore makes claims that \"sound right\" but fail under closer scrutiny.  '),(6517,828778,'30,July,2004','2004-07-30 00:00:00','   3rd World Nation-State Roulette!    As some of you know, yours truly recently discovered he passed the Foreign Service Written Exam. This news came as great relief, lately I had been kind on edge about where I was going with my life. My next step towards a fabulous Foreign Service Officer career is the Oral Assessment to be scheduled this september-ish. Although the idea of living abroad for two years at a time sounds absolutely glamorous and enriching, there\'s a catch. Of course there\'s a catch, it\'s the Federal Government. Good news, bad news. Bad news, the first tour generally has to be done at what\'s referred to as a \"hard-ship\" post during mostly consular work (read: stressfully long lines of disgruntled would-be immigrants and mundane forms, stamp visas, etc). These are the sorts of countries where the infrastructure is less than U.S./Euro/Australia standards, ie shaky economy, poor healthcare/education, angry mobs of people who hate the US and like to throw rocks, albiet throw them poorly. Good news: these hardship posts are assessed \"hardship differentials.\" Basically, the government has guidlines that it uses to assign levels of hardship endured by their civilian employees working abroad and they\'re given a percentage of salary on top of the base. For example if you\'re assigned to Kuala Lumpur in Malaysia you get paid an extra 5% of base-salary. If you\'re assigned to Hanoi, Vietnam you get a 25% bonus. (on a base salary of 30K that\'s an extra 7.5K not to mention the housing allowance)  So, given my language predilections (Spanish, French, English or Portugese) and my aversion to cold-weather climates after paying my dues in Hanover, I\'ve assembled my scouting chart of places I\'d like to go, with their differentials listed:  Barbados: Bridgetown -5 Belize: Belize City - 10 Brazil: Sao Paulo - 5 Colombia: Barranquilla - 10, Bogota 5 Cuba: Havana - 20 Ecuador: Guayaquil & Quito - 15 Fiji: Suva - 10 India: Mumbai (Bombay) - 20 Mexico City -15 Pakistan:  Islamabad 20, Karachi 25 Solomon Islands -15 Sri Lanka: Colombo -20 Trinidad and Tobago: Port of Spain -5  The polls are now open, vote now for where I should go. Think of it, y\'all,  Wah Hoo Wah goes GLOBAL . I think my early favorite is Sao Paulo, Brazil. I\'d love to go back.  '),(6518,828778,'29,July,2004','2004-07-29 00:00:00','   Biggie, Give Me One More Chance    Last night while house sitting, I broke into my employer\'s bottle of Jameson\'s Irish Whisky. The inspiration came while I was watching  Lost In Translation  for the first time, which I thoroughly enjoyed---yes, I know what you\'re thinking: Does he really need any sort of incentive to drink whisky in the first place? But seriously, how could I resist? Bill Murray continuously hawking that&nbsp;\"Suntory whiskey for good times....Sontory times.\"  In other news the rabble of the unwashed, the DNC, is in town as you all know. I rode into the city on tuesday night to have dinner with Alexis, the same of [Mis]adventures fame. The T-ride was weirder than usual. After 45 minutes, I was almost flipping out thinking about how these legions of people actually believe government&nbsp;can be&nbsp;a vehicle for social good. Universal Healthcare, more money spent on education, blah blah blah. I can\'t help thinking they want to make us into a more European state, yet how can it be helped when they criticize the current Administration on all issues with Europe as an ideal: \"our foreign policy needs to be more in line with our European allies, we need healthcare like the Europeans, need to tax fuel like the Europeans.\"  All of this leads me two believe two things. Firstly, the Dems are no longer Americans. They don\'t believe in America or think of America as the ideal. Secondly, they\'re absolute  MORONS . The U.S. is tops in economy, defense and ranks very&nbsp;highly in education, standard of living, and opportunity (especially when considered that countries that beat us in education or opportunties for women are dramatically smaller and easier to&nbsp;manage).&nbsp;Basically, in almost any conceivable category for comparison the United States is the best. We\'re Numero Uno. And yet the Dems want us to shift our policies to be more in line with the third best team out there: the EU. What a joke. It would be like the Patriots having won another superbowl deciding: you know what? Defense and inventive offense are good, but maybe the Chiefs have it right. More offensive genius and speed is what we need! (Or worse for me to have to state, the Yankees saying to themselves:&nbsp;Boston and Los Angeles&nbsp;has a point. Let\'s cut our payroll by 70&nbsp;million and compete that&nbsp;way cuz it\'s better because they say so.  You don\'t get to be number one by copying&nbsp;number 3\'s homework, people.&nbsp;  '),(6519,828778,'02,August,2004','2004-08-02 00:00:00','   A Quickie on Global Opinion  Before I get to watch the New Look Red Sox for the first time since saturday\'s trade, I wanted to post quickly on something Mike brought up in the comments section after my post on  urlLink  Lost in Translation /theDNC/basic Democratic Party policy .  As a rebuke of my position that the Dem position that we should be more like Europe is completely crap, Mike brings up opinion polls around that currently show most people elsewhere in the world have unfavorable opinions of President Bush and the U.S. He then concludes \"We no longer have the luxury of being governed by the school of isolationism.\"  Firstly, I don\'t think Mike is really referring to the school of isolationism. In IR, that school argues that the United States policy should be derived from a principle of \"mind your own business.\" It\'s philosophical underpining is drawn from George Washington\'s farewell address that advised the US from becoming involved in the entangling politics and alliances of Europe. Applied today this would mean not becoming involved in anything abroad that doesn\'t directly affect us at home. The US would withdraw from NATO (with no USSR, it has lost its raison d\'etre), withdraw from our mutual defense pacts with Japan/Taiwan/SouthKorea, etc etc. In the context of the challenge of global terrorism, this strikes me as insane BUT there are those out there, serious academics and polititians, that feel this is what we should do, and generally they\'re in the Libertarian camp. They argue that if we didn\'t have so many troops in the middle east or support Israel, Radical Muslims wouldn\'t hate us--and having close to no troops committed abroad would save boatloads of money that we could spend on education and healthcare.  Of course, I think Mike mispoke, I think he\'s referring to Unilateralism, which is what our European critics abroad and the Democratic critics of the Bush Administration at home have been accusing us of since we decided to go to war with Iraq without the formal approval of the UN. Unilateralism as it\'s latin roots clearly implies means the U.S. does what is in our best interest often without consulting or co-operating with other nation states; we go it alone if we want to/have to. Unilateralism lately falls under the banner of the School of Primacy. Primacy is the foreign policy framework where the US does everything it can to maintain it\'s position as the global system leader--biggest economy, military, most influence in NGO\'s.  Citing global popular opinion, I believe Mike and certainly Europeans and the rest of the planet want us to follow the tenents of  Multilateralism . This would mean that we wouldn\'t have gone to war with Iraq because the UN didn\'t give us approval and several countries out there wouldn\'t join GW\'s \"Coalition of the Willing.\"  Multilateralism sounds peachy, it really does. A bunch of countries get together to do something they all agree is the right thing to do and will benefit everyone. Unfortunately, the current global system doesn\'t relate well and global popular opinion forgets this. First, the collected nations of this planet do not share the risks of international terrorism equally. It\'s pretty obvious that the US and Israel face the brunt of the hate of Radical Islamists. Second, the costs of stopping the terrorist threat cannot be shared equally. Put simply, the US spends the most absolute money on defense of any country on earth and it  outspends the next 10 countries on the list combined . So, as much as everyone here in the US would want other countries to supply more troops and supplies to our effort in Iraq, other countries just don\'t have it in the cupboard. The last obstacle to true multilateralism is that global concensus is almost impossible to reach. Terrorism isn\'t the easy call of Nazi Germany where it was relatively clear that Hitler was not going to stop until he took over the world. The US and Western Europe just don\'t have the same interests when it comes to terrorism.  Defending the United States and being it\'s President is not a global popularity contest and it shouldn\'t be. The President, regardless of who he is, is responsible and accountable only to American voters. And this is the disjuncture of the Democratic Party. They argue that having Europe on board would make the war against terror easier, but the bottomline is that most of them don\'t have significant resources to offer to the cause. Ultimately, I don\'t want a popular President or government, I want an effective one.  If you\'re questioning President Bush\'s effectiveness, please consult Mr. Kadafi.  '),(6520,828778,'01,August,2004','2004-08-01 00:00:00','   Nomar No Mas    So, today I\'m writing from Tucson. This is WAH HOO WAH live from Arizona. Apparently, while I was in the air over the midwest, the Red Sox traded away Nomar Garciaparra to the Cubs and involving 3 other teams, got back a SS, 1B and an OF.  I heard about the trade while my plane was taxing about the runway and I checked my voicemail. Over the din of the airplane cabin noise, I could only hear my buddy Joe say \"we got NOMAH\" and that we got a bunch of players names I couldn\'t make out.  From what I can tell, this trade improves our defense at 1B and gets us a speedier outfielder. The improvement isn\'t that profound, that much is obvious. As far as I can tell this trade was about clearing the clubhouse air. After the aborted trade in the offseason and his injury, Nomar didn\'t look happy in the dugout. He was playing like a pro, but you could tell he wasn\'t the same guy. No smiles, not up on the rail with his teammates during that game against hte yankees.  I feel really weird about this in its aftermath. I feel like after a very fun, successful relationship with a girlfriend, we got to the point where it was best we part ways. There had been a spat of bad blood and now it wasn\'t the same. No hard feelings, it\'s not like I hate her or anything, but it\'s clear it\'s over and if you don\'t break up with her now, you know she\'s gonna break up with you soon anyways. Now you\'ll have to see her every now and then, say hi, politely catch up. You still feel like you hope good things happen to her. In a way it\'s the closest thing to an amicable breakup, if there is such a thing. Sure there was some drama and some hurt feelings, but you don\'t harbor ill-will and you\'re happy about the prospects of a fresh start.  For those of you wondering, when I landed in Phoenix it was about 108 degrees and completely dry. In fact, when I left massachusetts it must have been 83 with 60% humidity....and they didn\'t feel that dissimilar.  Also, my first meal before the drive down to tucson was the ever famous  IN AND OUT  burger. It was awesome, friends. Absolutely awesome. I savored every bite of the best rapidly prepared burger in America. It\'s totally equal to the hype produced by southern californians during my freshman year of college.  '),(6521,828778,'10,August,2004','2004-08-10 00:00:00','   Sad, Sad Pigeon Holes and Some Biased Analysis    This afternoon, I got home to check blitz and found myself on the recipient list of this forwarded email from my friend Cara:    Things you have to believe to be a Republican today.      1. Saddam was a good guy when Reagan armed him, a bad guy when Bush\'s daddy made war on him, a good guy when Cheney did business with him,and a bad guy when Bush needed a \"we can\'t find Bin Laden\" diversion.     2. Trade with Cuba is wrong because the country is communist, but trade with China and Vietnam is vital to a spirit of international harmony.     3. The United States should get out of the United Nations; our highest national priority is enforcing U.N. resolutions against Iraq. 4. A woman can\'t be trusted with decisions about her own body, but multi-national corporations can make decisions affecting all mankind without regulation.     5. Jesus loves you and shares your hatred of homosexuals and Hillary Clinton. 6. The best way to improve military morale is to praise the troops in speeches while slashing veterans\' benefits and combat pay. 7. If condoms are kept out of schools, adolescents won\'t have sex.      8. A good way to fight terrorism is to belittle our long-time allies (\"that\'s old Europe\") and then expect their cooperation and money. 9. Providing health care to all Iraqis is sound policy.     Providing health care to all Americans is socialism.     10. A President lying about an extramarital affair is an impeachable offense. A President lying to enlist support for a war in which thousands die is solid defense policy.     11. Government should limit itself to the powers named in the Constitution, which include banning gay marriages and censoring the Internet.     12. Being a drug addict is a moral failing and a crime unless you\'re a conservative radio host. Then it\'s an illness and you need our prayers for your recovery. FEEL FREE to pass this on. Friends don\'t let friends vote Republican.        Someone on the list replied to recipients and offered their response to some of these points, but I wanted to put mine up here to kill two birds with one stone (blog for the night and respond). My responses ended up being kind of involved, so if you choose not to read all the way down, remember this: To vote or be a Republican you don\'t have to believe in any of these things. Most of these things are the stereotype of some crazy far-right Christian fundamentalist, neo-con war-hawk. In fact, this list reminds me about the growing trend in society where people who have a single, strong belief or cause for change such as pro-choice or gay-rights activist tend to be Democrats. Yet, if I was going to say there was one most important thing a Republican should believe in, I would say \"smaller government, strong defense.\" Anyone who thinks that the federal government should be smaller and generally stay out of everyone\'s way should vote Republican. If we differ about gay marriage or environmental stuff, I don\'t care as long as we agree on that phrase. Taxes bad. Markets good.  1. I\'m tempted to forgive this skewed and mistaken analysis of history since it\'s common to neophytes of foreign policy. As a general rule, foreign policy is often a choice between the lesser of two evils, tempered by \"the enemy of my enemy is my friend.\" Since one of the bigger responsibilities (if not  the  biggest) of a sitting government is the sovereignty and security of the state, this often means strange and changing alliances. It doesn\'t matter if the President is a Democrat or a Republican, this is always the calculation of a head of state. So, although this redux on Saddam is technically accurate--it naively ignores the larger picture. It wouldn\'t be any different to say in 1941 we were friends with \"Uncle Joe\" Stalin and then in 1945 he was the main threat to the U.S. So what changed? It\'s obvious that in \'41 we needed the USSR to fight off the greater evil of Hitler and Japan and then with that threat ended, the Communists were the biggest threat. Bottomline, Saddam was a lesser evil than Iran or the Soviets, but once the Cold War ended and the era of global terrorism began, Saddam became persona non grata. Claiming that Saddam was a distraction is just silly. Bringing home Bin Laden\'s head on a plate would be the biggest political prize the Bush administration could claim. Mark my words, Bin Laden dead or alive before November 2 and this election is over.  2. I don\'t know why we still have this embargo with Cuba, but it has nothing to do with free trade with Vietnam and China. Trade with China and Vietnam has to do with global capitalism, no Republican worth his salt would put \"global harmony\" above Econ 1\'s relative advantage. The embargo on Cuba has everything to do with the ethnically Cuban electorate in Florida and if the Democrats want to win Florida\'s votes for real this time they will continue to support the embargo, too.  3. Sigh. Who said we needed to get out of the UN? If anything, this is a libertarian position, since the UN involves the US in too many things not directly (read super-directly) in the interest of the US. Republicans never used the phrase \"highest moral priority\" when describing the need to enforce the Resolution against Iraq. The Bush administration did say that without teeth, UN resolutions were meaningless. For the entire Cold War, the UN was meaningless because of US/Soviet vetos. It\'s meaningless now for the same reason with a different cause: the different interests of the Security Council. Don\'t kid yourselves, France and Germany said \"give peace a chance\" but what they meant was \"we own too much Iraqi (Saddam\'s) debt.\" Either way, this Republican thinks the UN as presently constituted (Security Council of permanent members runs the show) is digging itself a grave next to the League of Nations for largely the same reasons.  4. Many Republicans are not \"pro-life\" just as there are a few Dems who aren\'t \"pro-choice.\" I don\'t even know how this made its way into the agenda, Roe v. Wade is not under any sort of attack at the fundamental level. Moreover, the core of the decision on a legal level is about personal autonomy. Hey, personally I think if you choose to get an abortion, you\'re going to hell. But no more so than someone who lives their life based on the pursuit of greed and sex. Both of these things the government shouldn\'t involve itself in--many Republicans structure their beliefs this way. As to corporate misdeeds, regulations as Democrats usually construe them  are never the answer . I can\'t go into too much detail, but government would be better off crafting market forces to create desired social results: ie, give tax breaks to companies and car owners who build/buy more fuel efficient autos or conversely tax poor fuel efficiency. Given these market constraints, the result would be better than government saying \"you can\'t have this.\"  5. Hillary sucks. I don\'t know how she has the support of feminists or Dems at all. She rode the coattails of her husband to a seat in the Senate of a state she didn\'t live in until the election year and she put up with decades of adultery. How is this not the grossest case of nepotism? How can you point the finger at GDub for being Daddy\'s son and not at Hillary? God doesn\'t hate anyone, but the Judeo-Christian/Muslim God certainly doesn\'t condone homosexual behavior--but again, I as a Republican don\'t think the federal government should involve itself in these matters.  6. I don\'t know much about veteran\'s benefits and how they were changed, but I do know that much of them were probably created after WWII. The size and nature of our military has changed a great deal since, most notably WWII-Vietnam involved drafts, while the current military is entirely professional. \"Cutting benefits\" might not sound nice, but it could be the right thing to do. Cutting spending in general is a Republican trait and I totally support it. I don\'t believe the federal government is a good vehicle for social good, so shrinking it is generally good.  7. There\'s a fine line between educating adolescents about sex and facilitating their sex. Ignoring adolescent sexuality isn\'t the answer, but neither is \"here kid, use this and you\'ll be fine.\"  8.  urlLink Alliances are good things to have, but we shouldn\'t give up on foreign policy goals just to retain them --after all, what good are they if they don\'t help with our problem now? As to \"belittling\" them, the British don\'t feel that way. The French and Germans, Belgians, et al feel belittled but (and this goes for France especially), they\'re just ignoring geopolitical reality. The United States is the global economic and military hegemon. We expect some quid pro quo from NATO/EU because we were in their corner during WWII/ColdWar, but maybe that\'s wrongheaded. Much of the EU doesn\'t face the threat of terrorism like we do, so naturally they\'re not jumping at the idea of sharing the cost. Monetarily, we\'re talking about nickles and dimes, though, when this war on terror is costing some benjamins. Remember that the US outspends the top 2-10 countries combined in absolute money on defense. As far as hard military assests there is little that other countries beyond Britain and Japan can offer the US, but many are offering logistical support on hunting down terrorists, differences on the War in Iraq aside.  9. Providing health care to Iraqis is important because much of their infrastructure was destroyed by Gulf War I/II and bad upkeep by Saddam, without some of their health needs being met it\'s difficult to imagine Iraq forming a healthy democracy and  a democratic Iraq is in our best interest.  As for healthcare for every American, the choice is between the best healthcare money can buy for 95% or mediocre/poor health care for 100%. I choose the former since the latter would involve the biggest bureaucratic, wastful monstrosity the world might ever see. Universal State-led Healthcare is socialism. Socialism is bad. It only works well in places like Sweden or Denmark because their populations are so small and used to donating more than half of their income to the public wellfare.    10. This first line is disengenuous. A President lying about an affair to a grand jury is an impeachable offense. It\'s called purjury; you or I would face a fine or worse if we did the same. Moreover, he had sex with an intern--a more horrid example of abuse of power to sexual ends I can\'t think of. Any CEO would have been GONE. How NOW didn\'t crucify him is beyond me...except that is was the GOP that lead the witch hunt, so that made it ok? In Jan \'03, the CIA, the French, the Israelis, GDub, Kerry... everyone  thought Saddam had WMD. To accuse GDub of lying would be to accuse most governments on earth of the same. A global kabal lied about WMD? Enough all ready! For Bush to have lied about the existence of WMD in Iraq he would have had to have known conclusively  before  troops hit the ground that there wasn\'t any, despite everyone and their mom feeling pretty sure that Saddam did have WMD hidden somewhere. Believing that Bush knew when everyone else didn\'t is just a detail to forget when you want to hate him.  11. Yeah, Republicans generally prefer \"strict constructionist\" interpretations of the Constitution. And the proposed amendment defining marriage as a union of a man and a woman is a response to some judges reading our constitutions quite liberally. But proposing an amendment to the Constitution is fair game, the Dems have proposed a few themselves. I don\'t support censoring the internet. But if we\'re going to through around the \"censhorship\" mud, let\'s take a look at the college campuses around the country that censor lots of speech that \"offends\" people.  12. I don\'t really listen to Rush. I prefer  urlLink Laura Ingraham , a fellow alumnus of the Old College. I don\'t even know what the charge against Republicans is here....did he not get prosecuted? I\'d prosecute his ass. Is the complaint that we should legalize drugs? Being against legalization is definitely not a Republican-only position.  '),(6522,828778,'09,August,2004','2004-08-09 00:00:00','   Awkward    I got back from my annual visit to the eye doctors this afternoon after something odd went down. During my exam, we got to talking about my color blindness and how I feel like it doesn\'t manifests itself in any ways I can point to aside from a few shirts I own that I thought were one shade of one color, but turn out to be a shade of a different color. I said that I guess I tend to enjoy and prefer wearing clothing the color of which I can easily distinguish or clothing with sharp contrasts or bright colors which lead my eye doctor to say  \"that\'s the negro in you.\"  To say I was confused, puzzled and feeling awkward is an understandment. I just held a pause and said...\"uhh what do you mean.\" He nervously, quickly went on to explain how black folks wear bright colors...I was more confused. \"I\'m sorry, do you think I\'m black?,\" I asked him. He mumbled some quick apologies and started to retreat about making a bad joke, conceeding that it wasn\'t exactly PC.  To me, the problem here isn\'t that he made a joke that involved a race/ethnicity. It just wasn\'t funny. I suppose it might have been funny back in the late 80s, early 90s when pop/hip-hop artists started wearing \"African\" color schemes and African-Continent pendants. I felt even more odd because after his joke bombed, he was afraid I was somehow offended. I actually don\'t care about insensitive jokes and stereotypes. They can be darn funny and I think people shouldn\'t censor themselves and especially humor. Funny is funny. What offended me was that his joke just wasn\'t funny.  '),(6523,828778,'09,August,2004','2004-08-09 00:00:00','   Missing the Dry Air, Sunshine and Arizona Love    I\'m back from my vacation in Tucson, AZ and feeling jet-lagged for the first time in years. Fun Fun. Before I post about the highlights of my trip, I wanted to tie up a loose end. I think my last post was kinda wordy and I\'m not even sure if it was inteligible, so freshly culled from  urlLink this  article in the National Review, my sentiments but better written:  \"How many times have we been told that George W. Bush \"squandered\" the good will of the world \"community\" after 9/11? The assumption behind all of this seemed to be that anything which cost America the support of allies like France or Germany was, in effect, too costly. In other words, the means â€” \"strong alliances\" â€” are more important than the ends â€” winning the war on terror, toppling Saddam, and so forth. Listening to these folks, one gets the sense that America\'s greatest foreign-policy triumph was to get sucker-punched on 9/11 because it resulted in sympathetic newspaper headlines in Paris and Berlin.\"  '),(6524,2818788,'23,May,2004','2004-05-23 00:00:00','   \"Sir, do not look at the lady, but look at your arms. She is not yours,  nor shall she ever be.\"   '),(6525,2818788,'23,May,2004','2004-05-23 00:00:00','   \"After all is said and done, a lot more will be said than done.\"   '),(6526,2818788,'23,May,2004','2004-05-23 00:00:00','   \"Otis: Yeah, I wrote my history paper. My dad even said it was good! Yay! Josh: Is he Jewish?\"   '),(6527,2818788,'23,May,2004','2004-05-23 00:00:00','   \"Oh, I see how you are! A one-night stand and then it\'s all over.\"   '),(6528,2818788,'23,May,2004','2004-05-23 00:00:00','   \"Otis: Wanna hear a bad joke? Dani: Sure, why not. Otis: Ok, why do you never see any hooker fish? Dani: Umm, I dont know. Otis: Because they have all been caught! Dani: Uhhhh. Otis: Pretty bad huh? Dani: Yeah, it was horrible. By that I mean it was whoreible. Otis: That made it worse.\"   '),(6529,2818788,'23,May,2004','2004-05-23 00:00:00','   \"Dani: You\'re just jealous that I got to play hookey. That\'s why you aren\'t talking to me. Diana: Hi, good morning, I didn\'t notice you! Dani: ...I bit and kicked you. How could you not notice that?! Diana: No you didn\'t...\"   '),(6530,2818788,'23,May,2004','2004-05-23 00:00:00','   \"Don\'t make me go chibi-pocky-eating-ninja on your ass.\"   '),(6531,2818788,'23,May,2004','2004-05-23 00:00:00','   \"From the name hydrochloric acid, you can determine that there is one hydrogen atom, and one chlorigen atom in the molecule.\"   '),(6532,2818788,'23,May,2004','2004-05-23 00:00:00','   \"Theoretically, people see money on a counter and no one around, they think they\'re being watched. Honesty through paranoia...\"   '),(6533,2818788,'23,May,2004','2004-05-23 00:00:00','   \"Silly Otis-blog-making-person, behold! The Dani is here\"   '),(6534,2818788,'23,May,2004','2004-05-23 00:00:00','   \"Quote me as saying I was mis-quoted.\"   '),(6535,2818788,'23,May,2004','2004-05-23 00:00:00','   \"\'Reality\' is the only word in the English language that should always be used in quotes.\"   '),(6536,2818788,'23,May,2004','2004-05-23 00:00:00','   \" HiEmDosHuevos:  You screwed up my joke DAMN YOU TO HELL!  SelketRune:  .....Fine then, Are you a truck?  HiEmDosHuevos:  No, it\'s too late now.  SelketRune:  Yay!  HiEmDosHuevos:  You\'ve gone a screwed up the whole anatomy of this jest.  HiEmDosHuevos:  I may cry.  SelketRune:  ....The joke has an anatomy?  HiEmDosHuevos:  yes. And you cut off it\'s WANG.\"   '),(6537,2818788,'23,May,2004','2004-05-23 00:00:00','   \" HiEmDosHuevos:  Glah.  HiEmDosHuevos:  Things drive me insane.  HiEmDosHuevos:  Like, y\'know... ZOMBIES.  SelketRune:  ......Why?  HiEmDosHuevos:  They make me want to dance on things. Things that men were not meant to dance upon.\"   '),(6538,2818788,'23,May,2004','2004-05-23 00:00:00','   \"DM: You attempt to strike at A Goblin. You miss. The goblin attempts to strike. It misses.  Player: I try again.  DM: You attempt to strike at A Goblin. You miss. The goblin attempts to strike. It misses.  Player: Again.  *about fifteen minutes of this passes*  DM: You attempt to strike at A Goblin. You miss.... The goblin attempts to strike. It misses......  Player: Ugh.  DM: Suddenly, the goblin decides it is worthless and no longer deserves to live. It throws a suicide roll...  DM: ...And misses. *cry*\"   '),(6539,2818788,'23,May,2004','2004-05-23 00:00:00','   \"Aww.. little children. Let\'s go terrorize them!\"   '),(6540,2818788,'23,May,2004','2004-05-23 00:00:00','   \"Devin: I\'m a fishy.... fishy... fishy... fishy.. Diana: You\'re a stupid fishy is what you are. Devin: Ow.\"   '),(6541,2818788,'23,May,2004','2004-05-23 00:00:00','   \"Homosexuality is never mentioned explicitly by any character save when Gene says that Finnyâ€™s pink shirt makes him look like a â€œfairy.â€ But the relationships in the book are all between boys (we never see them interact with girlfriends or mothers, except for one brief scene with Leperâ€™s mother), and the central relationship contains hints of an almost sexual attraction between Gene and Finny. When the boys go to the beach, for instance, Gene remarks that his friendâ€™s skin â€œradiated a reddish copper glow of tan, his brown hair had been a little bleached by the sun, and I noticed that the tan made his eyes shine with a cool blue-green fire.â€ Although the language is not expressly homoerotic, Geneâ€™s words suggest that the connection between him and Finny is, if not sexual, then at least strongly physical: the boys direct their affection toward each otherâ€™s whole beings, increasing the intensity of their bond.\"   '),(6542,2818788,'23,May,2004','2004-05-23 00:00:00','   \"Okay, I\'ll post that picture for you now.  Thank you! You\'re now my favorite person!  ....And I wasn\'t before?  ...Well...You were my third favorite person....  **AHEM**  I mean, you were always my favorite person!\"   '),(6543,2818788,'23,May,2004','2004-05-23 00:00:00','   \"**We\'re bitching about Naron at ASD yesterday**  Diana: You know, for being so pretty, she sure is bichy Truley: She\'s not even that freaking pretty, her nose doesn\'t match her freakin lips!\"   '),(6544,2818788,'23,May,2004','2004-05-23 00:00:00','   \"Can I still have my Throwing Stars of David?\"   '),(6545,2818788,'23,May,2004','2004-05-23 00:00:00','   \"I\'m going to look old by the time I\'m twenty! I\'m going to have grey hair!  ....Why?  You!\"   '),(6546,2818788,'23,May,2004','2004-05-23 00:00:00','   \"What about those forearms that are held together by a chest? Come along, my dear embryo, and find your tool.\"   '),(6547,2818788,'23,May,2004','2004-05-23 00:00:00','   \"hiemdoshuevos: Hrm. Well it\'s a RP thing where DM puts us in stupid situations, and then we do stupid things. hiemdoshuevos: And then at the end you have to fight the dwarf, and you cannot win against the dwarf. hiemdoshuevos: Anyway. hiemdoshuevos: I chose, this game, to be Vannah White. Don\'t ask me why. I don\'t know. SelketRune: Wait, you were in Vannah White? hiemdoshuevos: No, I chose to RP -as- Vannah White. Like, she was my character. SelketRune: Oooooo! SelketRune: Okay. Continue. hiemdoshuevos: And throughout the entire game I kept collecting letters from things. Whenever I saw something with a letter on it, I took a letter from it. SelketRune: **Snickers** hiemdoshuevos: At the end, I\'m staring down the Dwarf, and I still have the letters W, L, A, and N. hiemdoshuevos: So I use these letters to spell the word \"LAWN\", and cover The Dwarf with grass. He then caught fire. SelketRune: .....And? And?! hiemdoshuevos: He put himself out and then shot me in the face with a shotgun. hiemdoshuevos: Like I said, you can\'t beat The Dwarf.\"   '),(6548,2818788,'23,May,2004','2004-05-23 00:00:00','   \"The cheese is smiting me!\"   '),(6549,2818788,'23,May,2004','2004-05-23 00:00:00','   \"1st Man:Mikes not actually my name it stands for where I used to work.... Microsoft! 2nd Man:Yeah I used to work at NASA. 3rd Man:*sticks hand up in dorky way toward chest* Disney 4th man: AOL JoAnn:Is that why all these women are so slow! \"   '),(6550,3072650,'31,March,2004','2004-03-31 00:00:00','   Ahh, Phantasy Star Online...  one of the funnest online games that I play offline. ... as well as the only one. With all this long-range, short-range, scifi mages, not-pausing-while-in-the-start-menu, one can get addicted. One day, I went into the menu of one of my other games, and tried (sub-consiously) to navigate it with the c-stick. But getting to beat episode I is a good goal, especially since the boss looks like a neon sign gone horribly, horribly evil.  GO BUY IT!!! NOT THE X-BOX VERSION!  x-box = waste of money = wasted money for Windows = wasted money for people that starve babies  Always behind you until you turn away, thusly ninja-ing into the nearest shadows, The Tempo  '),(6551,3072650,'30,March,2004','2004-03-30 00:00:00','   My name is the Tempo, but you can call me the Tempo. Or the Twitch. Blah blah blah about me: I like potato chips. GIVE ME POTATO CHIPS! I have a \'cube. And a PS2. I am NOT going to EVER have an X-box. I hate X-BOX. I hate X-BOx. I hate X-Box. I hate X-box. I hate x-box. I hate X-box. I hate X-Box. I hate X-BOx. I hate X-BOX. ASCII HATRED ART!!!  until someone tells me something to do, Always behind you, until you turn around, thereof I\'ll ninja into the shadows, The Tempo  '),(6552,3072650,'30,April,2004','2004-04-30 00:00:00','   I just got to beta a Gmail account, 1 gig of storage.  That\'s 1000 megs.  That means that I can store more information in my e-mail than I can on my old Win95 machine... saaaad.  So e-mail me, it won\'t matter, because my e-mail will only get full every 3 YEARS!  zero.exe(at)gmail.com  Always behind you until you turn away, thusly ninja-ing into the nearest shadows, The Tempo  '),(6553,3072650,'22,April,2004','2004-04-22 00:00:00','   well... I just bought an archie sonic comic... and quite frankly, I do NOOOT like it. it made sonic sound like a valley ... er ... boy ... or something, and I am pretty sure that sonicteam did not want a green knuckles. Or Eggman\'s first name Julian. Poor guy.  I mean, seriously, you come up with brilliant ideas(heck, you\'ve got an IQ of 300), but every single time some personified animal gets in the way.  AND THEN you, a male, has a feminine name.  Poor, poor eggman.  So, I will not include the archie comics in my studies.  Always behind you until you turn away, thusly ninja-ing into the nearest shadows, The Tempo  PS: Anti-X-box testimony: Today someone just blurted out \"X-box!\" for no good reason.  NO GOOD REASON I TELL YOU!! THEY\'RE BEING BRAINWASHED!! Anyway, right after that, I said \"Sucks.\"  '),(6554,3072650,'21,April,2004','2004-04-21 00:00:00','   Well, I have begun research on sonicology and will thusly post my findings here.  question: When I say sonic, what comes to mind? average joe answer: that one fast food place! sonicologist answer: adventure.  society these days... too obsessed with sport games... BEFORE the dumb -- I mean X box, most of us didn\'t play sport games. Or volleyball games. I don\'t think those were even in existance until the shadow government made that one DOA game... *Sigh*  But before the X-box, we relied on the systems with mascots.  Mario, Sonic, Crash Bandicoot... they were all popular.  But then mr. no-mascot came with all the M-rated-ness and the neuter\'d version of Halo - Bungee didn\'t finish it; mircosoft did.  If the X-box wasn\'t around, alota people would answer hedgehog over fast food.  Always behind you until you turn away, thusly ninja-ing into the nearest shadows, The Tempo  '),(6555,3072650,'30,May,2004','2004-05-30 00:00:00','   At the going away party for my friend, Kylie Jones, at Boondocks, I owned Virtual On.  I lost to no-one. It\'s a giant-robot 1P-2P arcade game, 1 play for 2 tokens, 5$ gives you 10 plays on VO.  There are 10 \"virtualiods\"(VR is the abbreviation, don\'t ask) to choose from.  The 2P game is best 4 out of 9.  My favorite VRs are V.R. Fei-Yen (I ownz joo and teh police and teh government and teh moon with her) and V.R. Viper II.  Now, allow me to share some veteran VO war stories.  It was classic.  Here is Parker, a complete newbie to VO.(I had played it twice before going to boondocks) He challenged be, thinking that he could dethrone me quickly.  I selected a medium VR, and Parker chose a heavy-weapontry specialist.  Parker relied on pure power of weapontry throught all 9 rounds.  It was quite funny to see him launch a bomb when I was on the other side of the field; he \"tried to confuse me with them\" whereas I was oftentimes behind him when he launched them.  Anyway, he won 2 of the 3 rounds he won consecutively, whichin I was developing strategies and the like.  He was overconfident. VERY, very overconfident.      It was 3/2, Parker in the lead.  If he destroyed me once, I would have lost.  I had to destroy him twice.  We started off with Parker throwing his diskus-bomb at me, but I had already dashed to the left, jumped up and fired a few missiles at him.  He fired another diskus-bomb at me, and I took the hit, but fired a napalm bomb at him.  I was near-death, he was 1/4 out of HP.  I switched to ninja tactics.  He kept on firing his heavy projectiles and the like at me, but I was always dashing out of the way.  I mainly fired some grenades at him, and one or two missiles. He lost horribly.    The score was 3 even.  He launched a diskus-bomb at me, but I jumped over it.  MISSILES ARE POWERFUL AT CLOSE RANGE!!!  He fired at me with his heavies, and a few gunshots, but I ran from their blast radius.  I was barely sratched.  One more hit with anything and I\'d of killed him.  Now - this is a very good strategy - as the VR cannot fire until the climax of the jump, normally locate your target and shoot at him with a powerful weapon before the climax of the jump.  He was destroyed midair.     Now, Parker challenged me again. This is ironic, but Kylie came up and messed with the dual joysticks when I was selecting the VR, and I, for the first time, selected Fei-Yen.  Parker selected Dorkas.  There is something very useful about Fei-Yen.  If she takes 2/3 damage, her armor turns gold and her attacks are 150% more powerful.    It is sooo emasculating to loose to a Fei-Yen, mainly beacause as the killing move I usually use the Heart Beam. Which is in the shape of a heart.  I dashed right at him at the begining and passed him, when he tried using a fireball at the start.  I destroyed him 1/4, Alan in the lead.  I\'m too lazy to describe it.  Always behind you until you turn around, thusly ninja-ing into the nearest shadows, The Tempo  '),(6556,3072650,'26,May,2004','2004-05-26 00:00:00','   I am officially un-grounded. I am an official bonsai owner. I am an official web storywriter. And I AM FREE!!!  Always behind you until you turn away, thusly ninja-ing into the nearest shadows, The Tempo  '),(6557,3072650,'14,May,2004','2004-05-14 00:00:00','   Well, on last Tuesday, I got the guts enough to brag about my gmail account[zero.exe(at)gmail.com], and found out that a fellow white belt has one too.  Small world. Too bad Green_ninja didn\'t get one... eh heh heh heh... so next time, I might ask her for her blog URL.  Always behind you until you turn away, thusly ninja-ing into the nearest shadows, The Tempo  '),(6558,3072650,'08,May,2004','2004-05-08 00:00:00','   update on operation chao factory:  thanks to green_ninja, I can now make any chao I want. hip hip horray! ... that is all. TempoGrunge out.  Always behind you until you turn away, thusly ninja-ing into the nearest shadows, The Tempo  '),(6559,3072650,'08,May,2004','2004-05-08 00:00:00','   Me & my friend,  urlLink green_ninja , have been hacking the sonic advance 2 chao garden. We\'ve sucessfully hacked the total ring value and the ring value for the white GBA minigame, and the total ring value APPEARANCE, of which the latter\'s maximum value is 8888.  Green_ninja suggests modding both the appearance and the total value to the same value; I, however, do not. here are the locations in the memory:  03001ab4 = minigame white GBA ring value -- 8 BIT-- 030044d8 = garden ring appearance value --16 BIT-- 03005254 = garden ring value --16 BIT--  Always behind you until you turn away, thusly ninja-ing into the nearest shadows, The Tempo  '),(6560,3072650,'07,May,2004','2004-05-07 00:00:00','   w00t to the power of w00t!! Hooray for shigmuffins!! and other various shouts of glee and general happyness! Why? Because I have just recieved a 16-port hub for FREE, AND 3 network cards, for FREE! ... but I have NO idea what to do with them. curse you, foul irony!  Always behind you until you turn away, thusly ninja-ing into the nearest shadows, The Tempo  '),(6561,3072650,'05,May,2004','2004-05-05 00:00:00','   But alas, for only about 60 hours... but it is GOOD to have the internet & my \'cube back.  Now I\'m going to sign onto battle.net and see if any of my starcraft friends are on. ... Fiend is like Friend without an R... hm...  Always behind you until you turn away, thusly ninja-ing into the nearest shadows, The Tempo  '),(6562,3072650,'05,May,2004','2004-05-05 00:00:00','   Being grounded from all forms of electronics... unable to update... must... find hole in the code... get... ungrounded...  being grounded... really, The Tempo  '),(6563,3072650,'01,May,2004','2004-05-01 00:00:00','   Current music: few tracks ripped off of Sonic CD for the PC  So here I am, putting notes on my life for the first of May on the great expansion known as the internet.  My friend has made a blog in a last-ditch effort to get a gmail beta offer. This is his blog:  urlLink Life as We know it    Ok, now that that\'s down and finished, it is time to talk about how the universe has recognized the word \"shigmuffin\". The word \"shigmuffin\" was created nigh but a few minutes ago, in a simple IM conversation with green_ninja.  He was telling me that Sonic said the \"sh**\" word in the Japanese version of Sonic X. (kudos to whoever animated, produced, and storyline\'d it)  I jamp* to conlusions, and my brain instantly thought of \"shigmuffins\".  Now you will get an exaustive definition and history on that word. Shigmuffin is a variable, as in it can be used as a noun, verb, etc. (i.e.: Look at that shigmuffin go!\", \"Wow... did she just shigmuffin her LUNCH?\") It is not nessisarily an insult or the opposite of an insult, whatever that is. ... And that is all.  Now, on to the final part of this post.    look at me!! I know how to HTML!!    ahem, I am planning on writing a mecha science fiction soon. So there.  Always behind you until you turn away, thusly ninja-ing into the nearest shadows, The Tempo  *jamp: best described as: ran is to run as jamp is to jump  '),(6564,3072650,'30,June,2004','2004-06-30 00:00:00','   Just made a new song, but the unfortunate side of it is that I don\'t know how to post mp3\'s onto my blog. Sad, no?  Always behind you until you turn away, thusly ninja-ing into the nearest shadows, The Tempo  '),(6565,3072650,'04,June,2004','2004-06-04 00:00:00','   Since my last post, don\'t you agree?  James and Ryan have been at California for the last 2 weeks, I\'m going to be graded for a 5th Kyu in Aikido on July 13th, and I had a very good time at the Reniassance(sp?) festival with Kylie.  In my honest opinion, I am going to die at the grading, being as I have no freaking idea what the techniques are.  Oh, and did you know that Kylie\'s PC doesn\'t have a floppy drive? It is evil, very evil.  James says that Blogger gives you a list of people that share the same interest as you so he\'s going to use it as a way to find a girlfriend. Sad, eh? Oh, and I found out that nobody can out-ninja me at night. ... or at least in theory... I will have to test it... In other news, Kylie has invited me to an RPG with Jordan and Joseph, but nobody plans when the sessions happen. It\'s a tad annoying, being that I have to wait a long time and all. Dang, I feel like this needs to be longer. Eh well. I\'ve been wasting my days playing the highly addictive game of Flight Simulator 2004, and I am currently flying around the japanese countryside in a biplane (Curtiss Jenny to be exact, for those who care).  I haven\'t seen a town since I\'ve taken off.   Always behind you until you turn away, thusly ninja-ing into the nearest shadows, The Tempo  '),(6566,3072650,'04,June,2004','2004-06-04 00:00:00','   With James Bergstom (aka Green_ninja) and Kylie Jones (aka ... uh ... Kylie?)  You have just entered room \"Chat 8810893060444337009.\" i Luv 2 Fli13 has entered the room. Psychoticis has entered the room. i Luv 2 Fli13: oooo i Luv 2 Fli13: i told you so Psychoticis: what ZEROdtEXE: *shoots james with anthrax gasket* i Luv 2 Fli13: w00t! i Luv 2 Fli13: i mean... Psychoticis: i\'ve had my anthrax shots too i Luv 2 Fli13: dodge it james! ZEROdtEXE: curses i Luv 2 Fli13: lol ZEROdtEXE: *shoots james\' belly with dana\'s love gasket* ZEROdtEXE: HOHOHOHO ZEROdtEXE: I am EEEEVIL i Luv 2 Fli13: ooo the lowest low Psychoticis: NOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOoooooooooooooooooooooooooooooooooooooooooooooooooooooooo Psychoticis: get it off me Psychoticis: it burns Psychoticis: it buuuuuuurns ZEROdtEXE: eh, get a shot for it i Luv 2 Fli13: haha ZEROdtEXE: should we dial 911, kylie? i Luv 2 Fli13: i laugh i Luv 2 Fli13: hmmm i Luv 2 Fli13: lets hear how much pain he is in i Luv 2 Fli13: .... ZEROdtEXE: ok i Luv 2 Fli13: i dont hear anything i Luv 2 Fli13: lol ZEROdtEXE: how much pain are you in, jameschu? Psychoticis: aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaghhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhh i Luv 2 Fli13: no he is fine ZEROdtEXE: yup Psychoticis: how do i make dana hate me i Luv 2 Fli13: ummmmmm ZEROdtEXE: you CAN\'T! HAHAHAHAHAHAHAHA i Luv 2 Fli13: be really mean to her ZEROdtEXE: kill hampsters? i Luv 2 Fli13: act like a girl with attitude ZEROdtEXE: in public? i Luv 2 Fli13: oooo i Luv 2 Fli13: thats a strech Psychoticis: kill hamsters in public Psychoticis: i like it i Luv 2 Fli13: lol ZEROdtEXE: well ZEROdtEXE: lemme call dana... looks like you need some CPR james... you don\'t have a pulse i Luv 2 Fli13: hahahah! Psychoticis: nooooooooooooooooOOOOOOOOOOOOOO i Luv 2 Fli13: mouth to mouth! Psychoticis: nooooooooooooo i Luv 2 Fli13: ha! thats hilarious Psychoticis: i feel like pedro i Luv 2 Fli13: whom? ZEROdtEXE: hahahah ZEROdtEXE: guy from anime ZEROdtEXE: excel saga i Luv 2 Fli13: my  bad ZEROdtEXE: which I still need to see *hint hint* i Luv 2 Fli13: me 2 *hint hint* Psychoticis: that means we need to split the cost of the DVD Psychoticis: alna Psychoticis: alan ZEROdtEXE: or... ZEROdtEXE: ro.. ZEROdtEXE: or... i Luv 2 Fli13: hehe i Luv 2 Fli13: ill buy! i Luv 2 Fli13: half! ZEROdtEXE: OBTAIN ILLEGAL COPIES!!! i Luv 2 Fli13: if i get to see it Psychoticis: i threw my illegal copie off the roof on accident ZEROdtEXE: *runs in circles* ILLEGAL COPIES!!! ILLEGAL COPIES!!! ZEROdtEXE: you are a sick man james i Luv 2 Fli13: hahaha ZEROdtEXE: a siiiick man Psychoticis: the riaa busted Jon ZEROdtEXE: oh i Luv 2 Fli13: jon ZEROdtEXE: I\'m sorry about your mofia cousin ZEROdtEXE: I... didn\'t know Psychoticis: jon poncelow ZEROdtEXE: kyile, do you know what the RIAA is? i Luv 2 Fli13: nope i Luv 2 Fli13: :-D ZEROdtEXE: it is... the greatest evil of planet earth Psychoticis: it actually was probibly the riaa and the mpaa ZEROdtEXE: it sues 12 year olds i Luv 2 Fli13: .... i Luv 2 Fli13: at least im 14 ZEROdtEXE: and 14 year olds i Luv 2 Fli13: how!?! i Luv 2 Fli13: for wut!?!?! ZEROdtEXE: music Psychoticis: watch the pepsi ad i Luv 2 Fli13: ok i Luv 2 Fli13: its not on ZEROdtEXE: james ZEROdtEXE: ILLGAL COPY OF PEPSI AD!! Psychoticis: ok i Luv 2 Fli13: lol Psychoticis: now your speaking my language ZEROdtEXE: oh ZEROdtEXE: well ZEROdtEXE: james, meet the geekiest girl in junior high ZEROdtEXE: kylie, meed a weird person ZEROdtEXE: *meet i Luv 2 Fli13: lol i Luv 2 Fli13: im a geek! i Luv 2 Fli13: yes! i Luv 2 Fli13: praise ummm i Luv 2 Fli13: wuts the geek god? ZEROdtEXE: antibill gates i Luv 2 Fli13: praise antibill gates! i Luv 2 Fli13: i dont get it i Luv 2 Fli13: if course i Luv 2 Fli13: of* Psychoticis: i got it ZEROdtEXE: copy of pepsi ad? i Luv 2 Fli13: why am i always the last one to know things i Luv 2 Fli13: :-( ZEROdtEXE: because of james ZEROdtEXE: want me to beat him up for it? Psychoticis: i got the pepsi ad i Luv 2 Fli13: well ill get you james i Luv 2 Fli13: sooner or later i Luv 2 Fli13: ya beat him with a light saber! i Luv 2 Fli13: and an ugly apple ZEROdtEXE: I\'ll take note of that i Luv 2 Fli13: awsome ZEROdtEXE: oh oh oh ZEROdtEXE: wait...  You have just entered room \"anit xbox persons.\" ZEROdtEXE: w00t it works Psychoticis has entered the room. Psychoticis: woot ZEROdtEXE: ... i Luv 2 Fli13 has entered the room. i Luv 2 Fli13: yes sur i Luv 2 Fli13: sir* ZEROdtEXE: w00t! Psychoticis: alan you spelled anti wrong ZEROdtEXE: well ZEROdtEXE: you took away an l ZEROdtEXE: so we\'re even i Luv 2 Fli13: here or there i Luv 2 Fli13: or anywhere? ZEROdtEXE: here i Luv 2 Fli13: ok ZEROdtEXE: not there i Luv 2 Fli13: ha! ZEROdtEXE: because here... we are anit xbox persons i Luv 2 Fli13: ur pink alan! i Luv 2 Fli13: haha ZEROdtEXE: james is ZEROdtEXE: as well i Luv 2 Fli13: alan\'s pink i Luv 2 Fli13: and james is 2 ZEROdtEXE: on your PC i Luv 2 Fli13: i laugh i Luv 2 Fli13: well i have to get up early tomorrow i Luv 2 Fli13: im gunna go ZEROdtEXE: noooo i Luv 2 Fli13: good night! ZEROdtEXE: ooooo ZEROdtEXE: why? i Luv 2 Fli13: im going to elitches ZEROdtEXE: oh i Luv 2 Fli13: without u! i Luv 2 Fli13: hahahahahahaha ZEROdtEXE: WELL i Luv 2 Fli13: ouch the pain ZEROdtEXE: bye i Luv 2 Fli13: lol ZEROdtEXE: or something of it i Luv 2 Fli13: im kidding alan i Luv 2 Fli13: love you lots! ZEROdtEXE: hm? i Luv 2 Fli13: bai! ZEROdtEXE: bye i Luv 2 Fli13 has left the room. ZEROdtEXE: poor james Psychoticis: poor me? ZEROdtEXE: poor james Psychoticis: poor you! ZEROdtEXE: poor me? ZEROdtEXE: NO MY BOLDNESS ZEROdtEXE: ok I\'m better ZEROdtEXE: why poor me? Psychoticis: because i have a gmail account ZEROdtEXE: uh... ZEROdtEXE: so do I Psychoticis: which means your no longer better than me email wise ZEROdtEXE: actually, I still am Psychoticis: fo real? ZEROdtEXE: for alas, if it was not for me, you would be nothing ZEROdtEXE: don\'t you remember who invited you? Psychoticis: curses Psychoticis: curses^20 ZEROdtEXE: will james ever find a girl he likes? ZEROdtEXE: tune in next time on ZEROdtEXE: uneligible bachelor Psychoticis: maybe i will find that special someone ZEROdtEXE: but then the question is, will she like you back? Psychoticis: she will be forced to ZEROdtEXE: by law? ZEROdtEXE: well anyway ZEROdtEXE: up for some starporting? Psychoticis: sure ZEROdtEXE: ok see you there  (note to self: beat jame up with light saber and ugly apple)  '),(6567,3072650,'26,July,2004','2004-07-26 00:00:00','   Ok, so it\'s 2:30 AM, and I plan on spending the whole night up. I\'ve already consumed 1 liter of pepsi, and I hope that\'s enough caffiene to keep me up all night.  Anyway, some of my readers might be wondering why I am doing this. Especially since this means all of those hours in front of the PC.  Although my parents will probably make me go to sleep,  I cannot, for \"CURSES, I hath been foiled!\" would be said. And that makes me most unhappy.  But, I digress. The reason for this test of geekness is simple - I am going to my friends house this afternoon (monday). BUT, thursday morning I have to wake up at 7:30. So, in order to get enough rest, I will deprive myself of sleep today/yesterday.  Thusly insuring my enough sleep-ness.  Oh, and this has been a goal of mine for a long time. I also plan on staying 2 days up straight. HAHAHAHAHA!!! HA! BUWHAHAHAH!!!     ...    DANG I need a MMORPG.    Always behind you until you turn away, thusly ninja-ing into the nearest shadows, The Tempo  '),(6568,3072650,'24,July,2004','2004-07-24 00:00:00','   Just alerting people of the current Terror Alert Level ... and acossicating sessame street characters with them ... Elmo is red. Pray we never see elmo.      urlLink         Always behind you until you turn away, thusly ninja-ing into the nearest shadows, The Tempo  '),(6569,3072650,'23,July,2004','2004-07-23 00:00:00','   Or the lack thereof, if you might consider it as. Alotta things have happened lately. For one, I put my RPG PBP forum back up, and ... uh ...    Well school is almosst back. *shoots self* Which is good and bad, but mostly bad. In fact, very bad and a tad good. The good part is that I get to see my high school friends more often, and the very bad part is the teachers, and the fact that most of the 8th grade left this year. Like, half of it.    So here I am, listinging to Japan-a-Radio on my iTunes, talking to Ryan, king of the nordic manslaughterers, and listening to my dog barking at harmless specters. and the King of Nordic Manslaughterers is sending me a picture of the girl he likes and it\'s messin\' with my bandwidth, which is small enough already. And I\'m on a quest to find a copy of \"Godzilla: Destroy all Monsters\" for the GCN, which another one of my friends thinks isn\'t as good as it is. BUT THE GAME HAS GODZILLA 2000  AND  GODZILLA \'90\'s!!! Plus all this AWESOMENESS of japanese monsters. AND DANG THIS IS TAKING UP MY BANDWIDTH!!! but enough of current events. I was going to take a flight in a cessna 152 at 8 AM this morning, but the weather foiled that plot really good, making VFR impossible. So no flying for Alan. *sob sob sob*    Always behind you until you turn away, thusly ninja-ing into the nearest shadows, The Tempo  '),(6570,3712913,'16,August,2004','2004-08-16 00:00:00','    As some may know, school started today. Here\'s the \"lo\' down\" on my current classes (all of which are eighteen weeks, ninety minutes per class).   English II Honors -  This is my \"mandatory\" shitty class. I already know it isn\'t going to be much different than the year before, which is to say this teacher is going to seem nice (and/or cool), but is, in reality, a bitch.   Biology I -  This class seems alright. I don\'t know many of the people in it, and I don\'t expect to become friends with any, but they seem really nice (and some funny), and the teacher is able to deal with anyone in the class very well.   Introduction to Business -  This one class is one I can somewhat look forward to. A few people I know, and some friends are in it. Also I honestly am glad I have it since, as I am always hearing, it is hard to survive in the business world. This class should help me prepare... or something, also the teacher is really calm, almost zen like I suppose.   French I -  Hmm, French. I don\'t know what to say, it seems pretty nice... though I\'m sure it\'ll be pretty hard if I loose track or am absent for a single day. But it\'s good I chose this over Spanish, because (as I found out today) French is the first language of 50 countries or so (Spanish of course being used in less countries around the world).  \"Be kind, for everyone you meet is fighting a hard battle.\"  -- Plato   The bus ride home was sadly, horrible. Our bus is much more crowded this year (by, honestly, twenty or thirty more people) and it is in some places three people per seat. Very uncomfortable, and Christ! Can these people complain (rhetorical question)! I finally told some of these asses to just complain to their parents, who would thus complain to the school board, and things would be switched around and made more comfortable. It happened before (when I was in seventh grade... or maybe it was eighth) and it can happen again.    To chase a feather in the wind...   '),(6571,3712913,'10,August,2004','2004-08-10 00:00:00','    Well, I\'ve deleted my old blog and started this new one. If anyone would like to post on my blog please sign it so I will know who it was, or at least sign it if I know you. Also, I don\'t usually update my website because my host doesn\'t let me post more that about 1.5 bytes (or maybe it\'s megabytes) per image. I might switch to GeoCities though, since they are supposed to have a great deal for  free  hosting. That\'s all for now, I\'ll try to post every three to five days... though I\'m not sure if I\'ll always be able to do that.   '),(6572,3723692,'31,July,2004','2004-07-31 00:00:00','   Him: if you donâ€™t mind can I say something?  Me: yea go ahead.  Him: (Looking at my french beard) if you dont want to keep a beard at least donâ€™t make fun of it.  Me: Thatâ€™s between me and My Allah and I walked out.  After Maghrib prayers, mosque next to office building.  '),(6573,3723692,'30,July,2004','2004-07-30 00:00:00','   So far so Good, was able to avoid sleep, thanks to another marathon meeting. Good thing is we have been able to come with the design of the initial requirements and now comes the approval and after that implementation.&nbsp; Since three days I am listening to EP - Rahguzar and its on constant repeat, at times I get bored of it and close down my winamp but after a while play it again. I think it is the combination of music, vocals and lyrics.  There are threats of terrorist attacks on Mosques and Imam Barghas&nbsp;during jumma time, May Allah keep us all safe and protect us from such acts Ameen.  Adios  '),(6574,3723692,'29,July,2004','2004-07-29 00:00:00','   Ok I wasnâ€™t like that, I never used to scold people on their acts (read annoying acts). Never ever used to get frustrated or agitated with extra work load or pressure inserted in order to get the work done. Tense yea but no signs of displeasure. I would never allow people remarks/comments regarding me get on top of my head and blow me. This isnâ€™t Saad I knew, this isnâ€™t me. Recently I have noticed a change in me, I would easily get annoyed if loaded with extra work and having people constantly inquiring about the status. I would pass angry glance at drivers(specially bikers) &nbsp;making stupid moves on road (I still believe their licenses should be snapped and they should be whipped for their acts). I would make sure to have a talk with the person who either steps on me in bus or blocks my way or just bumps me without caution.&nbsp; And yea I would even yell at myself like I feel right now coz&nbsp;its 2:45 AM and i shud be in bed not here posting argghh. Couple of days back I asked a colleague to ---k off just because twice he called me â€œbadey mianâ€ (I still feel he was rude) but that doesnâ€™t give me right to show him my middle finger. And yea I would show it to any driver who would honk his horn for no reason. Two months back I had again lost my cool, when a colleague was constantly asking me when I will be done with my work and that too 15 minutes after assigning it and we both knew it wont take less than 2-3 hours. This is not me. My GIKI friends would agree ( I hope they do :))   When I came back to Karachi, I used to find it quite silly when people show ed&nbsp; their anger at each other. Then I had this fear that one day I might be one of them and now when I look at myself, I feel my fear is turning into a reality. I am becoming one of them, frustrated of this world, wanting to get things done their own way, having thin temper control line. But I donâ€™t want to be them, I want to be once again ultra cool and be a person who would shrug off any remark, any annoying act with a smile. I want to be Saad again.  '),(6575,3723692,'29,July,2004','2004-07-29 00:00:00','   So they butchered them, Its on the news right now that the both Pakistanis who were&nbsp; kidnapped in Iraq have been killed ruthlessly. This wasnâ€™t supposed to happen, the script wasnâ€™t written for such darkness. Whatever and wherever went wrong might have dire consequences. My heart doesnâ€™t agree to the fact that sons of same ummah (if there is any) are responsible for this crime. It hasnâ€™t sunk in properly yet. But this is what reality is, it might just stop our government from sending troops to Iraq. The sacrifice might save many more innocent lives.  But what about those families? Someoneâ€™s beloved husband, someoneâ€™s dear father, someoneâ€™s brother and someoneâ€™s guardian. Will someone replace those shoulders? Will someone be there to heal their pain.  May Allah give them place in Jannah and provide patience and means of living for the suffered families.  '),(6576,3723692,'27,July,2004','2004-07-27 00:00:00','   Ok if u guys want to take the experience of limopani one step further, try the following recipe. For those who have no idea what limopani is, let me give an intro. Its basically a sachet of lemon juice and it tastes simply great, very much like fresh lime water. So here you go get ready for the ride of making a great limopani slush.  Note: This will make 2 to 3 glasses.   5 cups ice cube , crush it so that you save the blades of the mixture and yourself from wraith of your mom.  1.5 cup water.   3 sachet of limopani .  6 tea spoons sugar  or as you desire.  Place everything in the blender/mixture. Blend until smooth. Pour into glasses and have time of your life ;).  Have fun and do try Adios  &nbsp;  &nbsp;  '),(6577,3723692,'26,July,2004','2004-07-26 00:00:00','   And Yessssssss we have done it, finally have&nbsp;beaten India comprehensively, although we could not get a bonus point and the Indian commentators made sure that they ruin our party by mentioning it again and again and again. We might not reach the finals, but this win was very important. Defeat today and we might have seen the end of Pakistani cricket. But no we are back, although its way too early to say that things might change but at least we have got ourselves a breather. Lets hope this victory brings some consistency and professionalism in our cricket and might transfer some in our daily lives. I have a strong belief ,you all can disagree, Sports especially cricket can play a big role in&nbsp;reviving&nbsp;our Nation. Majority have lost faith in it but it can all come back with few good years and one World cup will add a cherry on top of the cream. Lets hope and pray for this Nation  Adios  '),(6578,3723692,'25,July,2004','2004-07-25 00:00:00','   So far it looks all good 223/3 in 39 overs, quite a good going for Team in Green(that is the official name or will be). Shoaib Malik coming up with a century while batting at number 3 and I am still not convinced that he should bat at this number, he is too vulnerable against swing and seam. Either Inzi or God knows who should bat at this spot. Anyhow still I feel no score is enough against this Indian batting line up, lets just hope that it continues the way its going right now. Till then  Adios  '),(6579,3723692,'23,July,2004','2004-07-23 00:00:00','   Something is not right and I am feeling it quite intensely right now. I want to scream out loud, but I canâ€™t. I donâ€™t want to do any work, but I am working. There is a shit load of work to be done btw(assigned by two different team leaders one after the another). I donâ€™t want to talk to anyone, but I am. I donâ€™t want to listen to any music, but I am and definitely I donâ€™t want to joke around, but guess what I am!!!!&nbsp;  '),(6580,3723692,'22,July,2004','2004-07-22 00:00:00','   Look at Love...'),(6581,3723692,'18,July,2004','2004-07-18 00:00:00','   Sami D(Durrani) aka White Jazz&nbsp;will fly from Karachi tommorow, he will be leaving for Multan and then on 6th of August for his PHD (donâ€™t get wrong guys , he is just 25)studies to Florida. Our friendship started on GIKI chat server. White_jazz(Sami), Acrobat(me) and nightshade(Mansoor, my batchmate) three of us used to keep those channels alive around 2-3 am. From there our friendship got kick started, it was most unusual of combination. I guess many people had difficulty digesting it :). He was known to be from the burger generation of GIKI and you all know me :) (I know what are you saying Awais). But I guess it wasnâ€™t the reason of our friendship :). It was built on pure trust and care. But yea we do have common interest in books and music (and yea his awesome cycle too ;) ) which helped. We both helped each other during tough times in my last year at GIKI.  Sami fell in love with the very lovely Annie and mashAllah say both are engaged now and inshAllah will be getting married pretty soon. And I know what difficulties they both had but still got through.  Last year He got job with Unilever which brought him here to the city of lights. Although rare but whenever we had a get together it was amazing, but as it always happen in metropolitan city thanks to long distances and hectic job we could not meet on regular basis and this is the one regret I will ever have in life.  All the best for the future Sami and May Allah help you. You were an amazing friend in GIKI and after it and I hope it remains all the time. I Thank you.  &nbsp;  Adios  '),(6582,3723692,'18,July,2004','2004-07-18 00:00:00','   Opportunities, At times in life we wait for the opportunities of our own preferred choices. We would wait for them for a long time. During the whole waiting process, quite a few opportunities of lesser or no preference pass by and we let them pass. We do not even give a second thought of riding on those opportunities. And so it happens most of the time when our preferred opportunity doesnâ€™t arrive, we are forced to take the non preferred one. But the difference is we end up losing considerable amount of time. The time in which we could have achieved a lot and won a lot. We could have been lot ahead of&nbsp; what we ever imagined to be. So my advice to you is, donâ€™t always end up waiting for chances of your own liking, if u ever get a secondary or tertiary choice always give second thought to it. Check all the parameters and if its worth it go for it, you never know where you end up.  &nbsp;  Adios  &nbsp;PS: I could have written it quite well, just am in ------g mood right now so this is all I could do. But I believe you guys are intelligent enough to interpret it on your own.  '),(6583,3723692,'17,July,2004','2004-07-17 00:00:00','   I wish I was a neutron bomb, for once I could go off    I wish I was a sacrifice but somehow still lived on'),(6584,3723692,'13,July,2004','2004-07-13 00:00:00','   Barely 20, his seat was hanging from the roof top 5 floors above the ground. Two rubber wires and 2 ropes were holding his chair. Ropes condition indicated that they have undergone maintenance work with couple of knots on each rope. He himself is wearing a safety belt, no helmet and no jacket to minimize any damage. Legs between a sealed rod and chair he rests his feet on the windowâ€™s edge. His face showed an image of a scared man, he wasnâ€™t purely confident of his safety. But perhaps getting the job done as quickly as he could was his utmost priority. With shaking hands he was holding roller in his one hand and trying to balance his seat with another. Whenever he gave an extra push to roll his roller on to the glass his seat would take a big swing and he would struggle to balance it. Started from 8th floor he had come down to fourth his mission was half done. How many times he must have done that before? Risked his life for what we call â€œearningâ€, atleast he is far better than the one who beg for living, far better than those who live on â€œunder table resourcesâ€ and far better than those who feed on the Meager ones.  But is he responsible for his own safety?  What if, with some bad stroke of luck he has an accident? Who will compensate his family? Who will feed his dependants? This is what I call Life. This is what is I call Pakistan, where price of life is maybe of few â€œPaisasâ€ or even lesser. I must say he is Brave Man. He is a Window Cleanerâ€¦  PS: Check out this new link on my side bar it goes  urlLink \"Elena\'s Motorcyle Ride through Chernobyl\" . Its about a journey through the horrors of Chernobyl. Very intresting Read.  '),(6585,3723692,'12,July,2004','2004-07-12 00:00:00','   Sunday started around 16:30 PM. After having breakfast + lunch, tuned star sports to watch Silver stoon F1 Grand prix. My mom is a big fan of Schumi(Michael Schumacher) and shes usually the one who reminds me of the race. She will follow it from start till end,  will discuss taking over maneuvers and keeps me updated if I slip around. She specially enjoys the podium ceremony. I guess I have inherited my sports genes from her. One thing I must say about Schumacher, no matter if it was his 80th win he celebrated it as if it was his first. Its just got to show how much he loves to win and still has hunger to win. And There has never been any doubt about his mental strength, a topic I want to write about.  In evening watched Black Fish perform. They were truly awesome, quite creative with loads of fun. Went with Hammad(my younger bro). Crowd was too western for my liking but I was concentrating solely on the show. At the end, the whole trip costed me quite a few bucks which has thrown my budget out of proportions. But thatâ€™s the story most of the time, last fifteen days are always spent in tight zituation(read situation). Which means I wont be attending next three shows of Black Fish (sniff sniff). One thing for sure Umar was the best of them all and it will require some special stuff to reproduce similar kind of show in his absence. Good luck to them.  Transport was a major issue today YET again and bharams of cab drivers really piss me off now. Yea I will keep whining about it so get used to it. Apart from work working on 2 hobby projects, one I am developing â€œgame of pongâ€ in Visual C++. The game is quite simple but since am doing VC after some time that too GDI so there is a lot to learn. Still havenâ€™t decided on AI(Artificial Intelligence) of the game but will think about it too. My Computer science buddies are welcome to make suggestions. Second is in C#(read C sharp) which involves converting few csv files into a format more suitable for another program, so that my friend can go away with the manual work of cleaning off those files. But its getting kinda irritating now. Lets hope by Tomorrow I make some progress in it. Thatâ€™s about it.  Adios  PS: Right now I am missing  urlLink GIKI  days and life there. Even though Its been 4 years but still cant get out of it. Maybe havent been able to adjust socially after it. No there is more to it , what I really miss is not being around my friends, My hostel room,Ather\'s room,Khokha chai/beef fry/haleem and parathas, water figths,Activities we used to do,late night chappas of assignments,jamming sessions,#roundadbout,#lunacy and loads more. This will take quite a many posts so I will end it here.  '),(6586,3723692,'11,July,2004','2004-07-11 00:00:00','   Saturday ends and Sunday begins, though its 5 AM and I should have been sleeping by now but I donâ€™t feel like. Slept from 6 till 10  reason, was feeling bit feverish, still do. So  urlLink Silent spring  departs. We twice had phone conversations today. He wasnâ€™t happy how the things went about with his mom today., I just hope before leaving it was all patched up. He did sound as if he doesnâ€™t want to come back, but I know he will. His family means a lot to him. I have suggested him to go for MS. He should do MS in film making etc etc. He is hell of a talented bloke.   Awais my  urlLink GIKI  buddy called from Paris today, he is there for a course and damn he is having fun in roller coasters and joy rides of Disney ;). He is a shy guy so there is no possibility of beach ventures * grin*. I wish him all the best for the course I know inshAllah he will do great. Tomorrow I will be attending play by Black Fish, you can read all about them in their community on  urlLink orkut . I am hoping for a good time. Till then  Adios  PS: whats wrong with my sidebar(its more of a footer now) I think its either the post size or number of posts, any idea anyone?  '),(6587,3723692,'10,July,2004','2004-07-10 00:00:00','   Had flu all day long with terrible sneezing. Initially I thought it is one of those sinus allergy stunts which happen regularly to me but now I fear its turning into a regular flu/cold scenario errrr and I donâ€™t want that. Lets pray for the best.   Met  urlLink silent springs  today(after 3-4 months) at pizza hut mid night deal, infact just came back an hour ago. He is off to his trip to Iran, so we decided to meet up. I just hope it proves to be a positive milestone in his life. I am sure he will be posting stuff from there hopefully with pics from his newly bought digi cam. He is a troubled soul and I just wish that when he comes back things take a U turn, inshAllah.   Bus traveling in Karachi is no more less than a severe mental and physical torture. You are pushed , squashed and suffocated to extreme. Not to mention the manner in which these buses are driven which clearly shows worth of your life is not even in pennies. Still I donâ€™t get why we board loaded buses, maybe we know all the buses will be same no matter how long we wait for or I think usually we are quite late with our schedule so we board whatever bus we get. Although new transport schemes have been launched but firstly they have limited routes secondly they have fallen into similar category of â€œthoosoo jitna thoos sakoâ€. So as long as some dramatic steps are taken we will end up being treated as Animals and Animals we are.  Adious  '),(6588,3723692,'08,July,2004','2004-07-08 00:00:00','   Call you up in the middle of the night Like a firefly without a light You were there like a blowtorch burning I was a key that could use a little turning  So tired that I couldnâ€™t even sleep So many secrets I couldnâ€™t keep I promised myself I wouldnâ€™t weep One more promise I couldnâ€™t keep  It seems no one can help me now, Iâ€™m in too deep; thereâ€™s no way out This time I have really led myself astray  Runaway train, never going back Wrong way on a one-way track Seems like I should be getting somewhere Somehow Iâ€™m neither here nor there  Can you help me remember how to smile?  Make it somehow all seem worthwhile How on earth did I get so jaded?  Lifeâ€™s mystery seems so faded  I can go where no one else can go I know what no one else knows Here I am just a-drowninâ€™ in the rain With a ticket for a runaway train  And everything seems cut and dried, Day and night, earth and sky, Somehow I just donâ€™t believe it  Runaway train, never going back Wrong way on a one-way track Seems like I should be getting somewhere Somehow Iâ€™m neither here nor there  Bought a ticket for a runaway train Like a madman laughing at the rain A little out of touch, a little insane Itâ€™s just easier than dealing with the pain  Runaway train, never going back Wrong way on a one-way track Seems like I should be getting somewhere Somehow Iâ€™m neither here nor there  Runaway train, never coming back Runaway train, tearing up the track Runaway train, burning in my veins I run away but it always seems the same  '),(6589,3723692,'07,July,2004','2004-07-07 00:00:00','   Monday ended quite well unexpectedly, I was able to advance to next level in swimming which I was struggling to do so. I know one nothing although I learn sports well(many will disagree with well) I am a slow starter usually take time to get in terms with the standards but once the I get in flow its always fun. Lets hope this experience is a similar one too.  Tuesday started quite bad, missed my colleague who was supposed to pick for office. That was big misunderstanding which resulted in me arriving office late by 20 minutes. I have noticed one thing if this event had occurred 2-3 years ago I would have taken it quite calmly but today I didnâ€™t. I was pissed and it took till lunch hour to get me back to normal. :). Nakhray hogayein hain meray this is bad. I need to be same old Saad again. Maybe these are the effects of living in Karachi, I love this city but sometimes I feel it does take its toll on you. Especially bus traveling Aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa. Itâ€™s a perfect way to mentally torture you, unless you become  Numb to it which unfortunately I havenâ€™t :(.  Anyways got back home wasnâ€™t feeling well so didnâ€™t go to pool and I know I have missed an important session I wasnâ€™t supposed to take break. Khair kul daikhtay hain kiya hota hai. At home I called couple of friends whom I wasnâ€™t in touch for couple of months. I am on a â€œreestablish-old-connectionsâ€ mission. It was good to have a chat with them. But its really tough to stay in touch with every one. There are so many old friends I am not in touch with, really miss those days. Likewise there are people who donâ€™t bother to stay in touch. I mean I know if I donâ€™t stay in touch with them nothing will remain,  should I let go?? Oh well its 3:00 AM will talk on this topic some other day. I just wish I had more time 24 hours donâ€™t seem to be enough anymore.  Adious  '),(6590,3723692,'05,July,2004','2004-07-05 00:00:00','   It seems to me Monday blues has hit me hard. The moment I sat on my PC chair I am feeling low on energy both physically and mentally. Maybe something has got to do with the dinner I had last night at my cousinâ€™s reception hmmmmm sounds fishy.  I had to push myself hard to work, hardly listened to any music, cracked around any jokes. Just lazed around and now I feel like screaming AAAAAAAAAaaaaaaaaaaaaaaaaaaaaaaaaaa. Nope didnâ€™t help much. Even had caffeine but no effect at all, maybe I will rush home early today and have some rest. It just might only be a bad tummy. Sunday as usual was best in boredom thatâ€™s why I sleep on Sundays as much as I can. This Sunday wasnâ€™t different. Woke up at 4 pm (that was 12 hours sleep), worked for 2 hours, and went to Saloon to get my French beard done. Though I slept out the whole time I was there, yes I have this thing of dozing off at odd places. I bet I can even sleep in rickshaw :D. Back home again slept for 40 minutes, got ready for the reception and then it is all the same. But the worst was kept for the last; yes Greece beats Portugal the host nation to win Euro 2004. That was just unbelievable as if Kenya wining the cricket world cup(with full apologies to Kenyan cricket fans). Although Greece played very well but they never deserved to win, their style of football is the most boring anyone could ever imagine. Their strategy was simply too negative â€œscore one goal and then defend with 11 playersâ€. You guys can disagree with me, but I donâ€™t care. I love attacking football and will keep on supporting it. And for me the best team of the tournament was Czechoslovakia. They should have won this one, sigh. Its 4:45, 45 more minutes then am off.  Adious  '),(6591,3723692,'04,July,2004','2004-07-04 00:00:00','   Its 4:35 Am and am just done with the new comments box. Also was able to edit its color scheme not sure if this one is any better(comments please)And also I think I need to reduce the font size of my posts what you guys say? Ok so day ended up with one good news  urlLink Maria Sharapova  won :D (also I got my salary, its party time ;)), Although I couldnâ€™t watch the match. I was in office waiting for my friend to get done with his work  so I could get ride for home, finally did at 8:30. But I managed to check live score on Wimbledonâ€™s official site. It was a goooood match shouldnâ€™t have missed but am used to miss good things in life. Anyhow time to sleep now, I am sure I will be having a nasty headache in morning God help me!!!!  Damn KESC voltage is fluctuating and my PCâ€™s UPS isnâ€™t working either.  PS: Because of the new comments box your old comments are missing I will try to put those in new system, till then have fun with this one. Its 5:00 AM now  '),(6592,3723692,'03,July,2004','2004-07-03 00:00:00','   Looks like its gonna be another boring day, there is nothing planned for the evening yet and doesnâ€™t seem like.Wasnt much work in office today either and thats a big big big surprise. I am still searching for a 3rd party comments box, found couple of vendors but the templates they got simply suck or maybe I am not searching hard enough. Good news I just heard a sec back, funds have transferred yaaaaaaaaaaaaaaaaay. That might just trigger a dinner or something. Btw today its Wimbledon Femaleâ€™s Singles final but I think I will miss fair part of it. And I am supporting  urlLink Maria Sharapova  â€œGo Maria Goâ€. Though it would be sad she will not have her fan cheering her :(.   Now few words in tribute of  urlLink Marlon Brando .  He was a legend and a complete actor. He thrilled millions around the globe with his acting. There can be as good movies as God Father but there wont be any Marlon Brando.  Adious  '),(6593,3723692,'02,July,2004','2004-07-02 00:00:00','   So it all begins, after resisting for so long finally decided to write my own blog. I have had enough of reading your blogs, I tell the world.  Although I must say it will not be as intellectual as silent springâ€™s or as funny as of Virtuous vibes or Zeniaâ€™s or as captivating as that of Crazed Insomniac or Atrophying Sensesâ€™s, but am sure it will be mixture of all, I am good in copy paste you know ;).  Friday has always been a dull and least productive day as far as work is concerned and if you had an hearty lunch and got over stuffed then it gets really hard to go through the rest of the hours. Today isnt any different, nine of us(newly employed, although I am here from 6 months now) had to treat 14 other members of our department. So KFC was decided as a venue. and all went well. The only issue left is, it was my CC which was charged aaaaaaaaaaaaaaaa and I hate recovering money (do I sound like shylock here???) Since coming back to office I am yawning and having ocassional 5 minutes naps :) and writing this.  By the way My name is Saad and I work as a Software Engineer. Work as usual has its ups and downs and to look forward for the day I started a sporting activity but for some odd reason struggling with that which is making me more and more frustrated as days are passing by and now I donâ€™t have anything to look forward for the whole day, yikes. Gtg now  Adious  '),(6594,3723692,'17,August,2004','2004-08-17 00:00:00','   Another burst up at work, not did I shout at anyone I just quietly asked him to go away. My inability to tolerate criticism specially negative remarks and sudden burst of rage might land me in difficult situations. Most of the times I find myself innocent and blamed wrongly for mistakes or lack of quality work and when it gets extended to a certain period I just refuse to cooperate, I am not sure I might be wrong in thinking myself innocent and they might be right or maybe I am right. I am scared that this unprofessional attitude might become permanent part of me and I might develop this as a habit, this can be fatal. I should keep my cool and deal with them some other way some other time, showing anger will not stop them nor will satisfy my soul.  One thing I believe that itâ€™s a gradual process, starts off with a feeling of lack of achievements taking along numerous regrets mixed with anger of making stupid decisions and all it needs a remark or two from people I dislike most and rest is history. So either I need to sort it out myself or may take some anger management therapy, situation isnâ€™t out of hands right now but I want to get rid of now and forever.  '),(6595,3723692,'16,August,2004','2004-08-16 00:00:00','   Was browsing through net when I came across  urlLink this site , it has records of characters killed in fictional dramas. I wonder how he/they have managed to do that and no idea when it was last updated but surely it is a fun site for fiction lovers. Have a go  '),(6596,3723692,'14,August,2004','2004-08-14 00:00:00','   Happy Independence Day to Everyone, May Allah bless this country and us and bring Justice and prosperity to this land of numerous sacrifices. Ameen  '),(6597,3723692,'13,August,2004','2004-08-13 00:00:00','   Love rescue me Come forth and speak to me Raise me up and don\'t let me fall No man is my enemy My own hands imprison me Love rescue me   Many strangers have I met On the road to my regret Many lost who seek to find themselves in me They ask me to reveal The very thoughts they would conceal Love rescue me   And the sun in the sky Makes a shadow of you and I Stretching out as the sun sinks in the sea I\'m here without a name In the palace of my shame Said, love rescue me   In the cold mirror of a glass I see my reflection pass See the dark shades of what I used to be See the purple of her eyes The scarlet of my lies Love rescue me   Yea, though I walk In the valley of shadow Yea, I will fear no evil I have cursed thy rod and staff They no longer comfort me Love rescue me   Sha la la...sha la la la Sha la la la...ha la la... Sha la la la...sha la la la Sha la la la...sha la la Sha la la la...sha la la la Sha la la... I said love, love rescue me   I said love Climb up the mountains, said love I said love, oh my love On the hill of the son I\'m on the eve of a storm And my word you must believe in Oh, I said love, rescue me Oh yeah, oh yeah, oh yeah...   Yeah I\'m here without a name In the palace of my shame I said love rescue me   I\'ve conquered my past The future is here at last I stand at the entrance To a new world I can see The ruins to the right of me Will soon have lost sight of me Love rescue me  '),(6598,3723692,'11,August,2004','2004-08-11 00:00:00','   Sunday was fun, went out of town with relatives. And guess where we went, to the land of VL yes you have guessed it right Balochistan. My uncle is posted at Soumyani( read Soum Yani) beach, and a visit was long due so last Sunday was chosen for it( although I wasnâ€™t happy about it, I get only one day for a looonnnggg sleep and it was ruined). After two hours of bus ride which I believe is the most fun part of any outing, we got to his place. It wasnâ€™t the kind of surrounding I was expecting it to be or the image that was drawn in my mind by different people was slightly more breathtaking. Nevertheless still the scene was a great sight. His house is located on top of the cliff and from the tip one can see sea waves crashing against the shore. And I have heard at night it becomes quite a scene. Anyhow after lunch had little bit of soccer before we all got exhausted, it was good to see I still got my touch ;). Went to beach afterwards and you all know what happens next. But the best part was, there was no one on the shore except us and the poor crabs yessssss we had crabbing. I think we caught about 20 of them.   So on Monday we decided to have the crab party, I asked for a recipe from my office cook who happens to be ex-gymkhana cook and he suggested to me quite a mouth watering one. I wonâ€™t go into the details of how we managed to get eatable stuff out of  the crabs (it was quite a eeewwwww for many). Anyways in the end the fried crabs came out pretty well and tasty ;). One thing which made me feel bad is the way you kill a crab, the only way to do is to boil them dead and that is a very very painful and slow death. May Allah forgive us.  '),(6599,3723692,'07,August,2004','2004-08-07 00:00:00','   Ah weekend(donâ€™t we love it) is here, although its been a painfully slow day(maybe because I am waiting so eagerly for clock to turn pass 6 PM). Major activities of the day consisted of handling adhoc tasks, avoiding sleep ;), reading articles (I will post an interesting link at the end of this post), and did I mention lunch at subway.  Resuming after 40 minutesâ€¦  Ok its 6 pm now yaaaaaaaaay time to rush home, though it doesnâ€™t feel good when u want to run away from office and head home. It gives an impression that something is wrong, u arenâ€™t interested in work anymore or not comfortable with environment. May be it\'s because of a strong inner feeling that I am not doing what I always wanted to do. Anyways time to go now.  Cheers  PS: check out this link  urlLink http://www.popsci.com/popsci/computers/article/0,12543,592948-1,00.html . Its about a simulation game which predicts states of different countries in future. One of its subjects is Pakistan yes our dear land. Do read its fun ;).  Adios  '),(6600,3723692,'06,August,2004','2004-08-06 00:00:00','   Parents are the most precious blessings of Allah, there is simply no replacement of them. Two weeks back a colleague of mine lost her mother because of illness, today she lost her father(nobody exactly know the real reason). This is one of the tragic events one could hear about and to go through it, must be a real pain. Not yet recovered from one loss and another inflicted upon them. May Allah bless her parents and place them in Jannah. Provide all of her family sabar Ameen. Incidents like these refresh our minds about us being mortals and the importance of parents in our lives. So respect your parents no matter how much you disapprove of their decisions/actions. Never ever lose your cool on them(Thatâ€™s for you Saad). If any one of you donâ€™t have one or both of them please pray for their Maghfirat thatâ€™s the best gift you can offer them.  '),(6601,3723692,'01,August,2004','2004-08-01 00:00:00','   I feel so dejected and defeated I feel so lost and humiliated I know its not the end But certainly its not the way I wanted it  '),(6602,3723692,'01,August,2004','2004-08-01 00:00:00','   Having doubts regarding faith, check these sites    urlLink http://www.omeriqbal.com/muslimsparadise.html     urlLink http://www.islamicissues.info/qa.html    urlLink http://islam.tc/ask-imam/index.php        Disclaimer: I will not take responsibility of any answers/views/interpretations expressed in these sites.  '),(6603,4129418,'20,August,2004','2004-08-20 00:00:00','   the hippie and i have noticed that we fall rather naturally into very distinct, traditional gender roles. our friends notice it. and the other night, it was very very apparent that our friends jenn and harry do the same. i mean, we went to their house for dinner, which was grilled by the men, after being prepared by the women. then after dinner--jenn put the baby to bed and i started knitting while the boys retired to the deck for pipe/ceegar smoking, and some scotch drinking on the hippie\'s part. and then jenn came back and we both knitted.  the hippie was telling me later that he and harry were on the porch talking about how cute it was that we were knitting, and how they both get a sense of satisfaction out of us being so girlie. funny--i get a sense of satisfaction out of him doung \"guy things\" as well.  and yes, i am pretty comfortable with this. it is bizarre to me HOW comfortable i am with it. i mean, i have multiple degrees and have always been pretty damned independent. but this last year and a half with the hippie, i have learned that there is a bit of joy in codependence. i LIKE our gender roles. i like that i do most of the cooking and he does most of the heavy lifting. i like that i make the drinks for our friends who are over while he entertains them. i like the whole set-up, and how much i feel like he is taking care of me, while i am taking care of him.  the weird thing is that i feel no less a feminist for all this. i am glad i have choices in things, and that i have had the opportunity to do what i want with my life. i am grateful to the women who came before me for fighting so that i could have those choices. i am glad i have the same rights and the same power as he does. i like that he listens to me and that he respects my brain even while he is grabbing my ass.  i like that this weekend, harry and the hippie are off having a boys weekend, playing D&amp;D, going to hooters, and probably seeing something awful like alien vs. predator (which i will get out of going to see thank god!!!), while jenn and i get to be girls and go on a ninja ikea run.  this is a great time to be a girl. you get to have it all. '),(6604,4129418,'18,August,2004','2004-08-18 00:00:00','   here i was ready to give up on my  urlLink wire things . here i was, all ready to throw in the towel. after months of no activity to speak of on this little business i am trying to run on the side while still frantically job hunting, one of my stores sent me a consignment check today. there it was in the mail, looking suspiciously like a bill or a letter telling me that no one inthe world would ever want to buy my artwork ever again. i was almost afraid to open it.  but i did. and it was a check. a fabulous little sum of money. for five pieces of artwork.  not exactly sweeping the art nation, but hey--it\'s SOMETHING. and damn, that feels good. '),(6605,4129418,'17,August,2004','2004-08-17 00:00:00','   my baby count for 2004 is up to 14. that is 14 babies i can name who were born or are going to be born in 2004. and my 2005 count is starting to add up as well. it is really amazing to me how many people i know who are breeding at the same time. i mean, i know i am the right age for this, but damn people! do you have to do it all at once??!  as a woman about to turn 35, this does have an effect on me. i think thoughts like--am i ever going to have one? the jury is still out on that, but i feel the pressure of my age to make up my mind. i can\'t wait TOO long, or my body will answer that question for me. i feel like hitting 35 means hitting a fertility wall or something, like my ability to get pregnant, should i choose to do so, will be irretrievably impaired just by crossing that one line. and yes, i realize this is slightly irrational.  i also think about the friends who are having babies now, and the ones who already have them--really good friends. the hippie and i talk about the need to have a few baby things in the house for their babies, things like a toy box, a place for them to sleep, a high chair we can keep in the basement and bring out when we need it. and it makes me reaize that if i AM ever going to have a baby--i want it to be able to grow up with these baies all our friends are having. i want our baby to have cool baby friends, too. as a child who was much younger than the children of my parents\' friends, i do not want to do that to our baby, if we do have one. it SUCKED always being the baby, literally 6 years younger than the nearest kid. i want our baby to have that sense of belonging and community that i never had from having a bunch of family friends.  if we have one that is.  like i said, the jury is still out... '),(6606,4129418,'16,August,2004','2004-08-16 00:00:00','   there is not enough room on the entire internet for me to talk about all my mixed feelings on this topic, so i am going to try to sum it up.  right now--fucking hate it. loathe it. dread it every day with every fiber of my being. i go to the gym, i feel self-conscious the ENTIRE time i am there, i count the minutes till my workout is over, and i am exhausted when i get home. half the time it is all i can do not the throw up on the treadmill. it is all i can do to make myself go.  and sometimes i don\'t, and it fills me with self-loathing.  but here is the thing. i need it. i know i need it. i know that this is a change i need to make for the good of me, my body, my future health, the future health of my possible future children, and the happiness of my hippie, too. (not because he judges me, but because he worries about my health.) right now i am heavier than i have ever been, and i feel it all through my body. my energy is lower, my blood pressure is higher--i have to take medication, which then makes me even more sluggish. and if i can get myself to consistently work out, then all that will definitely change.  i know that eventually i will come to like it. i know that eventually i will even look forward to it. i know because that has happened to me in the past. i long for the day when my workout ends with endorphins instead of exhaustion. and i know it will come if i keep it up.  but here is the thing about gyms today. this gym we go to pushes the personal trainer like nobody\'s business. and damn, how the hell can anyone afford one? a woman at the gym actually suggested to me that i get my boyfriend to pay for it when i explained to her that i just could not afford it! holy crap! do i have to pay out hundreds and hundreds of dollars to someone so that i will feel guilty enough about NOT working out that i do it? i have to make myself accountable to another person to follow through with a commitment to myself now???  and the other thing i totally fucking hate is how these trainers are so goal oriented. \"we need to get your body fat percentage down to blah.\" \"what is your goal weight?\" who the hell cares about that!? i know, and anyone who has read anything about anything knows that this is all just simple math. if you put more calories in than you expend, you gain weight, and vice versa. further, losing a massive amount of weight is only possible with a lifetime commitment to a lifestyle change, whether you need to eat less, work out more, or both. i mean, why can\'t i just focus on getting this shit into my daily routine. why do i have to think about my goal weight and how low i need my body fat percentage to be? that shit overwhelms me and makes me want to crawl into a hole and give up and go ahead and die of a heart attack.  screw that.  and man, i know self discipline and denial are so not my strong suits. but damn. i can do this. '),(6607,4129418,'07,August,2004','2004-08-07 00:00:00','   sadly, our romantic weekend getaway was squashed by all the hurricane madness attacking the north carolina beaches. *sigh* no worries--we will reschedule it! but, some clandestine activities on the part of our friend  urlLink steve  were revealed as a result of the cancellation. apparently, he read my previous post, called around, found the number for stardust, where we were planning to have dinner, and called them to pay for our meal as a birthday surprise. how cool is THAT??? steve is bona fide awesome. that is all there is to it.  so saturday, as i sat knitting on the couch while the hippie smoked a ceegar on the porch, steve called me to find out why we were not at the beach. i explained of course. he noted that hurricanes don\'t hit phoenix very often. i concurred. and he sent us to  urlLink sullivan\'s  for a birthday dinner on his tab. let me repeat: steve is awesome.  so we go to dinner, and dinner is great, as it always is at sullivan\'s. the waiter, however, is a patronizing ass, who neglected to ask the two important questions: 1) \"have you been here before?\" and 2) \"are you celebrating anything special?\" he explained every single tiny piece of the menu to us, the ordering process, the sizes of the sides, the sauces we could get on our steaks, the way the meat would be cooked, etc., etc., etc.. wanted to slap him. still kinda want to slap him. oh well--my steak was absolutely fabulous. i will remember it fondly on the actual day of my birthday. :)  thanks, steve! '),(6608,4129418,'07,August,2004','2004-08-07 00:00:00','   yes--i have finally stopped procrastinating and am cleaning out the last room in this apartment that needs to be totally overhauled. the bad thing is that this room has been the catch-all for EVERYTHING since we moved in. and god is a lot of it CRAP! for example, WHY do we have all these bags of bags here??? are we releasing the rednecks within us both in this one tiny room in the house? this shit has GOT to stop! i am trying really hard to stop being such a packrat--throwing things away like crazy.  but dude--i am KEEPING that morgan freeman can.  anyway--here is some shit i found while cleaning today:   things i hate about airports   (in no particular order)   1) TVs--loud, always on 2) HUGE # of people 3) lines 4) overpriced bad food 5) NO CELLPHONE SERVICE!!! 6) lack of electrical outlets 7) never a clock when you need one 8) carrying all my shit everywhere 9) obnoxious airline people 10) too much perfume--tne it DOWN people!!! 11) unwashed passengers 12) NOISE NOISE NOISE!!!  some quotes written on random pieces of paper:      certain organic chemicals of value to me are bound up in your body. kindly die and rot. --jeff freeman 02-23-2001     he was using a method of distraction as a fixer-upper for a lack of personal responsibility. (no idea who said that)     one must have chaos within oneself to be able to give birth to a dancing star. --nietzsche    what you really need to get through some serious cleaning is a girlfriend to park near you with a pile of magazines to make rude comments and hang out with you while you do your thang. snaps to my friend  urlLink sam  for talking to me on the phone through a lot of the madness even though she could not be here in person...  i know sam is a true friend because i can call her \"dumbass\" and she doesn\'t get mad. '),(6609,4129418,'05,August,2004','2004-08-05 00:00:00','   had kind of a rough day yesterday. my hippie is in california at a meeting--a meeting i would almost certainly be attending were it not for the fact that that company laid me off last november. and i have to say--i have been jealous. and angry. and lonely. all week. and no--i don\'t miss working there at all! i have been much happier since i lost that job actually. but i DO miss the people and the interaction and the social aspects of that job. and all the people i miss are gathered in one place for the whole week. and i want to be there too--schmoozing, staying in the nice hotels, drinking the fabulous drinks, passing notes in the boring meetings, and yes, even staying up half the night ranting about the company! but hey--it is a trade-off right?  so to get out of my stupid, partially hormone induced funk, i went to the stained glass store. love being in there--they have so many cool little things that are just inspirational to me. i have been working on a design for the hippie to do this fall, and now i am working on some things for me, too! some friends actually want to hire me to make an enormous panel for them, so i have been working on some rough sketches for it. hate trying to determine how much to charge people, however. i never feel like i am justified asking for what i know a piece is worth, even though i know they are still getting a huge discount having me do it rather than a professional custom glass person. i am in such a state with this job search that i wonder about everything i like to do--could i make a living doing this? stained glass is certainly high on my list of things i like doing right now, so who knows...  in other artistic endeavors ( i always have about 50 things going!), i am taking a knitting class tonight with a friend, and i can\'t believe how excited by it i am. it is SO old school. but then, so is the quilting thing. and the stained glass for that matter. heh--like i keep telling the hippie--i really am just a farm wife at heart. except for the actual farming part--screw that!  fabulous friends from california are arriving today. always wonderful to see them--how i wish they would move home already!!! but like everyone else, they are constrained by the job market. i look for jobs for him almost as much as i look for myself! they are having a baby in the fall, so the drive to get them home is stronger than ever, and not even just for selfish reasons! :)  i would give just about anything to get my friends all in one place! oh well--at least this way i have an excuse to travel.  and speaking of travel, i have made some little getaway plans for me and the hippie. our birthdays are so close together, it just made sense to do something together this year instead of trying for the elaborate presents we can\'t really afford anyway. so--we are going to the beach to stay in a little B&amp;B with a claw footed bathtub,  urlLink the cedars . and we have reservations at our favorite restaurant in the state, Stardust, with our favorite waitress there, Darla, and we will hopefully get to do a little sailing as well. very excited!  and now that i have procrastinated as long as humanly possible, i will go do some work on my workroom--really want to get my own little space set up so that i have a place to draw. but, oh the work to get it set up... '),(6610,4129418,'04,August,2004','2004-08-04 00:00:00','   driving today, after a grey and dreary morning, i was nearly overcome looking at the clouds.  the sky here in the south is usually hazy, but not today--the rain cleared all the haze away and it was blue blue blue--as blue as our bedroom walls... and i realized that although there are many things going wrong in my life today, the fundamentals are alright, and it was really just a gorgeous day.  wanted to write that feeling down.  i have been thinking about starting a blog for some time now, and have been strangely reluctant. the blogs i love to read are the ones where the folks posting are the most real, the ones that make me feel like i know the person writing. and i thought, \"am i ready for that?\" i guess we\'ll see... '),(6611,3802286,'30,June,2004','2004-06-30 00:00:00','   QUEER EYE FOR THE QUEER GIRL: Eve Tushnet defends her outfit and continues last night\'s discussion  urlLink here .  AND GOSH DARN IT, PEOPLE LIKE US: More free publicity on our very first day  urlLink here . I want to be  urlLink Eve  when I grow up.   More blogwatch and updated links to come.  '),(6612,3802286,'29,June,2004','2004-06-29 00:00:00','   The the national media finally caught on to the  urlLink rampant corruption  that pervades Philadelphia\'s mayors office.  All I can say is, it\'s about freakin time.  The  urlLink WSJ  confirms Street\'s \"Pay to Play\", quid-pro-quo politics, funneling bribes through City Treasurer Corey Kemp, who \"wanted four more years to do our thing\".  (Picture a rapper-like thug pronouncing it \"do our thaaang\".  Street\'s goons epitomize race-baiting and political cronyism.    This past fall, I worked heavily on the Sam Katz election campaign.  Republican Sam Katz, a repeat contender and Ed Rendell\'s director of economic development, promised to bring new hope to a city in decay.  Though the cards were heavily stacked against Katz with over 3 to 1 Democrat to Republican ratio, in miraculous fashion Katz stood a fighting chance, early on.  It is common knowledge across the city that John Street has a histroy of shady dealings, malfeasance,  urlLink rote corruption , not to mention administrative ineptitude.  Democratic Mayor John Street, urban slumlord, is Philadelphia\'s modern day  urlLink Marion Berry .   Street\'s policies are keeping Philadelphia down, whereas DC, New York, and Boston are thriving.  Philadelphia, in decline, is losing population each year and hemorraging businesses, in part, thanks to an overbearing wage tax.    Election time rolled around and Katz was within range, when it became national news that the FBI was investigating Street for his suspected criminal involvement.    Then the unthinkable happened.  Race-baiter that he was, Street was able to demagogue the investigation, energizing Philadelphia\'s lockstep minority electorate to turn out for him like lemmings.  He misconstrued a legitimate investigation by the FBI as a racial witchhunt and a heavy-handed Republican power-grab.  What was even more deplorable, were the amount of outside money and political influence that were smuggled into the city to help a crook get re-elected.  Both Clintons, Gore, and and countless other Democratic operatives turned out to stoke the racial flames.  The whole ordeal was an utter travesty, but this renewed media attention and indictment of Street\'s treasurer serve to remind responsible Philadelphians of what they have known all along.  It was a shame that Sam Katz failed to remove this man from office.    Street won\'t stop until Philadelphia is driven into the ground.  Thanks to Street, the city could very well become the next Detroit 15 years from now.  And that\'s not all...Street\'s most egregious sin of all was driving Koch Fellow and Philadelphia native, David J. Healy, ostracized like a Palestinian, out of Philadelphia and across the cold, wintry, and polluted Delaware River, and into refuge in the trash refuses of New Jersey....  '),(6613,3802286,'29,June,2004','2004-06-29 00:00:00','   Knight Ridder is distributing an  urlLink article  about a GAO report which rather bleakly describes the situation in post-war Iraq. The headlines tend to sum it up as something like \"Iraq is now worse off\".  First of all, that is not a very accurate summary of the article itself. To say that Iraq is currently worse off in a few key areas is not at all to conclude that it is worse off  overall ; that holistic determination is quite difficult to make, especially when you\'re working with only a few statistics. The GAO doesn\'t make that conclusion, and neither should the headline writers.   But moreover, it seems to me that some of the statistics the report relies on are badly misused. The report claims, for example, that the Iraqi court system is more clogged now than before the war. Well, I imagine a country wouldn\'t have much of a problem with the courts being clogged if it was, you know, a  totalitarian dictatorship . If the reason the courts are clogged is that we\'re trying to give people due process, then maybe a little court-clogging ain\'t so bad.  I don\'t mean to downplay the bad news contained in the report or to suggest that we have done everything right in Iraq (we certainly haven\'t). I just want to caution against the glib use of statistics in deciding the complicated and important question of whether we\'ve improved the lives of Iraqis.  '),(6614,3802286,'29,June,2004','2004-06-29 00:00:00','   John,  I concede that Ann Coulter is polarizing to people on the opposite side of the political spectrum.  I would add though that it is in part because she has been targeted by the left.  I would also re-emphasize that she is nowhere near as bad as Michael Moore.  As for your charge that she can not persuade, I would just venture the caveat that there are varying degrees of persuasion.  I think she can be persuasive, granted not as persuasive to novices, independents, and liberals, but definitely more so with people who are already somewhat conservative, or who have sort of inclination.    Furthermore, not everything that one writes should be held to some high-falutin standard of persuasion to be considered relevant.  Her writing also informs conservatives of issues, often political.  Note too that informing is another purpose that writing can serve.  Her writing also energizes some conservatives, which conservatives likely consider a favorable effect of her work.  What I see in Ann, on top of her writing ability, that makes her great are two things.   1) Her tenacity and assertiveness, contradict the seeming monopoly    on assertive, politically active women held by liberals.    (Conservative women on the whole are less confrontational than     liberals.  Of course, exceptions abound.)  2) I think conservatives need more emotionally involved people.  I    think that liberals do a better job of crusading for what they    believe.  Liberals also have more political stuff with an     entertainment component, and I agree with you that her work     does have entertainment value for some, myself included.  One criticism, is that I think it was wrong for you to bring up discussion of her weight.  Most people politics aside would agree, she is an attractive woman.  If we are ever going to mention a person\'s apprearance, it should generally be a positive comment, otherwise we should keep it to ourself.  We learned that one in kindergarten.    As for Erica\'s comment, no one ever said/implied that all women should strive to look like her, nor is there a single definition of beauty, nor is beauty solely physical appearance, nor does it count for all that much.  So Erica, I would say that you are wrong to call the unattainble standard of beauty that causes some women to feel self-concious about their appearance, the \"conservative\" standard.  I would say the reverse is true in that \"liberal\" sources i.e Hollywood, women\'s magazines like cosmo, popular music, etc. are much more responsible for making women feel self-concious about their bodies.   urlLink Spin Sisters  is a great book that articulates this point of view very well.  Michelle Hinman will vouch for me on this.  Ann Coulter has spoken regularly for the Young America\'s Foundation, CPAC, and  urlLink Conservative Leadership Seminars  as John mentioned.  She he would not be such a popular speaker, who is repeatedly invited back, if she was an \"idiot\" or a \"nutcase\" as Fabio so defamed her.     Yes Fabio, that is slander.  I am not going to call you names, but you saying that you \"despise\" her is despicably harsh.  '),(6615,3802286,'29,June,2004','2004-06-29 00:00:00','   The previous pseudo-debate surrounding Ann Coulter and her antics is an interesting one. Ann Coulter has become a figure roughly equivalent to Michael Moore in modern politics. While I believe that Ms. Coulter has a higher standard of journalistic integrity than Moore (she is, for the most part, factually correct in her pieces), she polarizes rather than persuades--that is, she is a rallying point for those onthe left who like to point out how heartless and partisan conservatives can be. Whether she is right or not, her rhetorical method of presentation is rude. All the sneering and pessimistic inflexibility that we on the right despise in, for instance, third wave feminists, Ann manages to espouse for the conservative cause; and those who delude themselves on this point are demonstrating that they either agree with these alienating tactics, that they don\'t care for the art of persuasion and for the careful propogation of ideas, or that they are, indeed, solely into Ann\'s work for entertainment value. Like a professional wrestler or Chris Rock, she is interesting (for a while).    Having met Ann in person and having taken pictures with her at a luncheon sponsored by the Clare Booth Luce Society, I can attest to only two personal points (and only two).  1) Ann is terribly underweight. This may be a result of a medical condition, etc. in which case, I have nothing but sympathy for her. However, she is not the kind of woman the Republican Party (or any party) should market as the standard for beauty. To encourage women to starve themselves into a coulter-esque physique is not only degrading and demonstrative of a perverse definition of beauty (one that requires unnatural sacrifice in women), but dangerous and irresponsible. Most women could not look like Ann if they tried--and most men would prefer that they not. A woman with a naturally beautiful and full body is both more historically beautiful (look at pre-Kate-Moss cultures) and more feasible.  2) Ann is a very poor speaker. She does not respond well to questioning and she, generally, reads from a prepared script. To many, rightly or wrongly, this gives the impression of \"idiocy.\" Based on her relative ineptitude in handling unfamiliar topics in front of live audiences, this is neither an outlandish claim nor does it demonstrate some character flaw in the person who would pass-down such a judgment.  Ann is a polarizing figure in American politics, but she is also a figure dwindling in popularity precisely because she alienates most readers and listeneners who haven\'t wholly bought into the belief that all liberals are liars, traitors, and thieves. I know many liberals...their hatred of Ann is understandable. My dislike for her (among other things) is based on her disrespect towards my honest, kind, and warm-hearted liberal friends. To blindly lump a group of people bound loosely by political ideology together and then to lambaste them with such harsh terminology truly is \"idiocy\"--whether the other side is willing to engage in similar tactics or not.  '),(6616,3802286,'29,June,2004','2004-06-29 00:00:00','   urlLink Great article .  Found this at Newmark\'s Door.  '),(6617,3802286,'29,June,2004','2004-06-29 00:00:00','   Saw  urlLink this  over at The Idea-Shop. Prostitutes are an inferior good. As male incomes increase (very few females solicit prostitutes) consumption of prostitutes should decline.  '),(6618,3802286,'29,June,2004','2004-06-29 00:00:00','   Wow...you guys are serious (Scott...did you already have all of this written?).  I am impressed, and here is my addition. If we were real radicals we would be fighting the oppressive  urlLink D.C. cell phone ban that is agitating Radley over at The Agitator . I, for one, as a driving member of the District populace am angered and dismayed. Does anyone know if this ban applies in Arlington and Fall Church or is it only within the city?  Hat tip:  urlLink Me.  '),(6619,3802286,'29,June,2004','2004-06-29 00:00:00','   Are Democrats  urlLink aborting their electorate ?  Among other points in the this articleâ€¦ 1) Democrats have more abortions than Republicans.   2) More abortions mean less Democratic voters, more so, than more    abortions means less Republican voters.  If Democrats want to    abort themselves, I say go right ahead.  (I am pro-choice).     Sense the irony here?  Pun intended.   Here is the kicker conclusion paragraph:  \"As liberals and Democrats fervently seek new voters and supporters through events, fund-raisers, direct mail and every other form of communication available, they achieve results minuscule in comparison to the loss of voters they suffer from their own abortion policies. It is a grim irony lost on them, for which they will pay dearly in elections to come. \"  I\'d say this idea is still up for debate (as Ben pointed out to me), but it is definitely a new way of looking at an issue that has routinely beaten to death and \"sucked\" dry like the brains of tragic, baby victims of partial birth abortion.  '),(6620,3802286,'29,June,2004','2004-06-29 00:00:00','   My conversation today over lunch break with a girl handing out information for fundraising with the DNC:   Girl:  Hi, would you like to help beat Bush in 2004?  Me:  Actually, I\'m a supporter of him, but would you like to explain to me why I shouldn\'t be?  Girl:  I would love to argue with you, but that\'s not my job.  Me:  Okay.   I walk off and buy a banana. 15 minutes later, I walk up and she is just finishing haggling a woman who said she didn\'t have money on her but wanted to support the cause. The woman insisted she worked for a nonprofit and was thus short on funds, and asked said Girl for her website. Girl said she was sorry the woman was low on cash, but couldn\'t she please give them at least half the amount she was thinking about donating? Woman walked off and said good luck.   Enter me.    Me:  Hey, can I ask you a question? If I told you I was really undecided about Bush [I may very well be ... -Ed.], would you argue with me then?  Girl:  No. That\'s not what I\'m here to do.  Me:  So what\'s your job?  Girl:  I am fundraising for the Democratic National Committee.  Me:  Right. So what is the protocol they tell you to follow?  Girl:  I am to raise money from our constituents.  Me:  But wouldn\'t you want to convert more people to your cause if you truly care about Kerry winning? Wouldn\'t you want to persuade me so I would give you money, too?  Girl:  No. If you like Bush, you won\'t give me money, right? So why should I try to convert you?   Affirming smiles mutually exchanged, I tell her to have a nice day and walk off.   More interested in grabbing the money of those already on their side (or even, on the off chance, actually comprehending with their philosophy) than explaining to the rest of the country why they too should care about the Dems? Complaining that Republicans are just out to get money and care nothing for the common people while wheedling money from street passersby (who are clearly not, in any way, the common people)? Sounds about right.  '),(6621,3802286,'29,June,2004','2004-06-29 00:00:00','   I think it was over at MR where I first saw this a while back. Here is a song by Rush called, \"The Trees.\" Rush is well-known to be libertarian in their beliefs. See what they have to say about fixing income inequality below:   There is unrest in the forest,  There is trouble with the trees,  For the maples want more sunlight  And the oaks ignore their pleas.   The trouble with the maples,  (And they\'re quite convinced they\'re right)  They say the oaks are just too lofty  And they grab up all the light.  But the oaks can\'t help their feelings  If they like the way they\'re made.  And they wonder why the maples  Can\'t be happy in their shade.   There is trouble in the forest,  And the creatures all have fled,  As the maples scream \"Oppression!\"  And the oaks just shake their heads   So the maples formed a union  And demanded equal rights.  \"The oaks are just too greedy;  We will make them give us light.\"  Now there\'s no more oak oppression,  For they passed a noble law,  And the trees are all kept equal  By hatchet, axe, and saw.    Well put in my mind.  '),(6622,3802286,'29,June,2004','2004-06-29 00:00:00','   For the record, Saddam tried to buy Uranium.  Gee I wonder what he wanted with that?  urlLink Uranium   \"New Iraq Leader believes Saddam connected to Al Qaeda\":  urlLink Al Qaeda Connection   WMDs check.  Al Qaeda Connection Check.  (Not to mention the extensive laundry list of Saddam\'s other gross war-provoing infractions).  Brokaw is the usual anti-war motivated liberal in this line of questioning:  excerpt:  Allawi: We know that this is an extension to what has happened in New York. And - the war have been taken out to Iraq by the same terrorists.  Saddam was a potential friend and partner and natural ally of terrorism.  Brokaw: Prime minister, I\'m surprised that you would make the connection between 9/11 and the war in Iraq.  The 9/11 commission in America says there is no evidence of a collaborative relationship between Saddam Hussein and those terrorists of al-Qaida.  Allawi: No.  I believe very strongly that Saddam had relations with al-Qaida.  And these relations started in Sudan.  We know Saddam had relationships with a lot of terrorists and international terrorism.  Now, whether he is directly connected to the September - atrocities or not,  I can\'t - vouch for this.  But definitely I know he has connections with extremism and terrorists.  No matter what, most Democrats are going to continue to be in denial and oppose the war, even if subconciously for partisan, election-year reasons.  '),(6623,3802286,'29,June,2004','2004-06-29 00:00:00','   From: Letter to Koch Listserve Tue 6/29/2004 12:27 PM   Subject: [KOCH]     For all the undecided voters out there... If you don\'t like Bush\'s foreign policy you can put our future in the hands of a man who wants to \"grow the spiritual left movement,  obtain peace and mindfulness within ourselves and talk about possible solutions\" ...or we could just blow ourselves up and save the terrorists the trouble.     Yes, this came from the Kerry website:   Yogi Jenn Joos, Oregon   Jenn Joos is a Kundalini Yoga teacher in L.A. and hosts Kerry Fundraisers.   â€œI felt that the yoga community could be a resource to reach out and talk about the important issues in a positive, constructive way. Often when politics were discussed at the yoga center after class, there was frustration at how things were in the world...but no solutions. I saw it as an opportunity to educate ourselves on issues, talk about the possible solutions and not feel so powerless! or hopeless. Kerry is an antidote for both politics-as-usual and a boost for the growing Yoga or â€œSpiritual Leftâ€ movement. Yogis believe in positive change. It starts as peace and mindfulness within ourselves by practicing Yoga, and filters out to help reach out to others in the world around us. We have a wonderful chance to help financially, but also give a message that â€œyoga votersâ€ want peaceful world resolutions, socially conscious programs and resource sustainability for our future.   I say that this is sickeningly hilarious.  '),(6624,3802286,'29,June,2004','2004-06-29 00:00:00','   Koch Fellows,   My good friend from school, Dan Kline has a Protest Warrior Chapter.  It is a great organization that opposes left-wing protests and demonstrations, and does a little counter protestation of its own.     Watching some of the clips and reading a little from  urlLink the website  (and from elsewhere and my personal visits to some of these demonstrations), many people wouldn\'t believe some of the radical tactics that many of these maniacs employ.  When the mob mentality sweeps over them, they are so full of hatred and a wreckless disregard for truth, reason, and couth.   Anyone who wants to is welcome to join our U of Penn chapter for the periodic updates.  There are numerous other towns and colleges with chapters.  You can browse by state.  Just go to  urlLink this link  and click on \"register\".  There are no obligations and if you decide to, I\'m sure that you will be glad that you did.  Pass this on the word to fellow anti-protesters.....  '),(6625,3802286,'29,June,2004','2004-06-29 00:00:00','   Here\'s the  urlLink latest PJ ORourke Article  for anyone interested. (I know a bunch of Koch Fellows went to see him last week.)   I don\'t agree with O\'Rourke\'s premise that Rush Limbaugh is merely are \"preaching to the crowd\".  Rush runs his \"Advanced Institute for Conservative Studies\" and 14.5 million listeners tune in every week.  I know for a fact that he regularly changes hearts and minds, and shapes the opinions of his listeners.  Listeners call-in and give confessionals about him converting them from Democrats quite often.  Michelle Malking writes  urlLink \"In Defense of Conservative Talk Radio\" , and rebuts some of what O\'Rourke was saying in his \"snarky\" article.    I would also note her observation that the troops want Rush.  Certainly not \"Nationalized Public Radio\" or Al \"the Freak\" Franken.  The only preaching that Rush does is through his \"talent on loan from God\".  '),(6626,3802286,'29,June,2004','2004-06-29 00:00:00','   My country right or wrong, right Scott?  You know, there are more than two types of people in this country. There\'s some good and bad in both parties.   And yes, I do think the Republican Party is a collosal failure that is only marginally on the right. And when I mean right, I don\'t mean pro-business, and small government. I mean, pro-guns, pro-specific businesses, pro-religion and sections of government that are smaller.   And no, I am not a Democrat and don\'t support their socialist mantra.  And no, I\'m not a LP sympathizer either. Anarcho-capitalist my ass.  I personally don\'t think that there is much difference between anyone who identifies hard-core with any party, be it Republican, Democrat, Communist, or Nazi.   Just watch where your allegiance lies. Is it with America or is it with the Republican Party?  Just my thoughts...  -Ed Mease  '),(6627,3802286,'29,June,2004','2004-06-29 00:00:00','   Ann Coulter rocks even though the left has so aptly demonized her to the extent that the right has Michael Moore (though characterizations of Moore as such are warranted).  If you are a conservative (as am I), then I highly suggest getting the  urlLink Ann Coulter doll  for your girlfriend.  I got one for mine and she loves it.  Fabio sent some rant to the Koch listserve about \"educating ourselves about the other side.\"  My feeling is that people who publically espouse something like this are most often the people that are least likely to live that way.  I doubt many people have actually read anything by her, so I am inviting you to do so.  Here is her lastest  urlLink debunking the Clinton\'s \"vast right-wing conspiracy charge.\"   Fabio also \"doesn\'t trust anything by Fox News,\" but  urlLink here  is one where she defends us \"crazy\" Fox News viewers, because I know there are a lot of us out there, more so than there are CNN (aka Clinton News Network) viewers.  Can I also just say that it is an outrage that a lot of the DC area networks don\'t even carry Fox News, the most watched name in news.  Last, here is a great piece by Ann that was posted on Town Hall that  urlLink Michelle Hinman so kindly shared  with the group last week.  I liked this excerpt: \"Though Kerry makes liberal ladies\' bosoms heave with his self-advertisements about his Vietnam experience, the Democrats might not want to let Kerry pursue this particular line of argument. According to Thomas Ricks\' book Making the Corps, the vast majority of officers currently serving in the military are conservative Republicans â€“ \"largely comfortable with the views of Rush Limbaugh.\"   Ann has high-level diction and is very entertaining.  '),(6628,3802286,'29,June,2004','2004-06-29 00:00:00','   Libertarians are right to criticize profligate Republican spending, but they go too far, when they put the party at parity with Democrats, who are by far the most reckless spenders.  Koch Fellow, Alex Potapov pointed out the  urlLink ranking by Citizens Against Government Waste . CCAGW\'s Congressional  Ratings evaluate members of Congress how much congressmen spend.   As Alex pointed out, in the 108th Congress the average score for House Republicans was 72%; for the House Democrats, it is 17%. In the Senate, the numbers are 70% and 19% with 100% being the high and 0% (Teddy Kennedy types) being the low.   If you still aren\'t convinced, I would also submit that my organization, The National Taxpayers Union, also does a ranking, called Bill Talley, which can be accessed at www.ntu.org,and of the biggest spenders, Democrats perennially come out on top   Here is a  urlLink great article  that ran in the WSJ influenced by Chris Cox, former Reaganite and Chairman of the Republican Policy Committee in the House.  (He\'s from Cali for you Cali people out there.)   It talks about reform efforts by Republicans to do away with some of the institutional biases toward bigger spending, such as spending being only a \"guideline\", a complicated committee system, \"baselining\", deadlocks at the ends of sessions, and lack of executive recssionary/line-iterm authority.  The article is fair, because although Republicans are better than Democrats on spending, it points out at the end that some Republicans aren\'t as accepting of the reforms.  Check it out.  Republicans aren\'t perfect, but Democrats, sure as hell, are a heck of a lot worse.  '),(6629,3802286,'29,June,2004','2004-06-29 00:00:00','   urlLink The New York Times book review rips Clinton .  (If Clinton can\'t even get a decent review from them, then his book must be pretty retrograde stuff.    urlLink MSNBC  tears into â€œMy Lieâ€ as well.  WSJ makes the case for  urlLink Clinton\'s historical Inconsequentiality    According to  urlLink public opinion , Reagan will be remembered as a better President than \"Slick Willie\" Clinton.   Fred Barnes has a nice piece titled,  urlLink \"The Incredibly Shrinking clinton\"   One low light of his book is that Clinton is  urlLink still not forthright  in his book about the Lewinsky cover-up.  What guile! (This link is from CNN for you Fabio).  The media sure did their darnedest though, even if indirectly to promote that crap.  There was not way that the book would have sold like this, if it had not been for a virtual weeklong media promotion.  '),(6630,3802286,'29,June,2004','2004-06-29 00:00:00','   Democrats will try to influence the election by any means necessary...  they will even try to  urlLink lure voters with free beer .   ...and  urlLink hire convicted felons  to go door-to-door.  In case you are a left-winger like Fabio and \"don\'t trust Fox News\" Here is  urlLink John Fund , a former Cato Intern, on the subject.  Lastly, people like Fabio also probably won\'t read the most professional newspaper in the country (the Wall Street Journal), b/c it is probably too \"right wing\" for them, so here is the morally-relativist, non-judgmental   urlLink AP write-up .  '),(6631,3802286,'29,June,2004','2004-06-29 00:00:00','   Rick Moran, who is a congressman where I\'m living here in Northern Virginia, was just accused of making an anti-semitic remark, and leave it to  urlLink CNN  to report it in the contrapositive.    He said something along the lines of \"if the jews hadn\'t supported it, America would never have gone to war in Iraq\".     I noticed the usual double standard as the media gave it barely any national attention at all.  The media would have lambasted a Republican such as Lott, though they give Democrats on stuff like this (Patti Murray Bin Ladin is a good guy, Hillary\'s bad joke about Arabs working at convenience stores, Al Sharpton\'s Hymie Town, Robert Byrd etc.) a free pass.    urlLink Fox News speculates  that Kerry wants this chump out.   The Jewish population is shifting from predominantly Democratic to a more even split, and stuff like this is part of the reason why.  '),(6632,3802286,'29,June,2004','2004-06-29 00:00:00','   On Eve Tushnet:  In arguing against gay marriage, Eve Tushnet argued that children do better when they have mother-father parents because they learn to relate to both sexes, as opposed to only one.  In fact, throughout her arguments, Tushnet made the assumption that non-same-sex relationships are better because they provide gender balance.  Here\'s what I find to be flawed in that assumption:   1.  Not All Relationships are Good Ones   Not all mother-father households provide a good example of how to relate to the opposite sex, or how to best exemplify attributes of one\'s own sex.  Consider the example set to children wherein the parents are abusive to one another, or, worse yet, are abusive to the children themselves.  I am not saying the overwhelming majority of mother-father relationships are negative, but I think the assumption that all of them are necessarily positive, in terms of providing role models for children, is fallaious.  Parents are not good parents because they have a penis or a vagina; parents are good parents because, regardless of their gender, they exemplify positive character attributes that serve as good examples for their children.  Honesty, love, trust commitment: these attributes are not bound by gender, yet I think most people would agree that they provide some of the best character attributes in parents.  Mother-father households, like heterosexual marriage, cannot claim a monopoly on providing these (among other) positive role model traits/attributes to their children or society.   2.  Children of Single-Parent Households Are People Too   Regardless of what role models children have at home, there is this other thing that all children, regardless of the number of parents they have, experience: the outside world.  In terms of relating to members of the opposite sex or of the same sex, there are a plethora of social situations that provide opportunities for children to relate to members of both sexes, namely, school.  Moreover, I think Ms. Tushnet\'s assumption that children of single-parent households are somehow less socially equipped to be successful or socially normal is based on an outdated notion of normality and success.  Like Megan brought up in her question, the real generation to watch, in terms of how children of single-parent homes fare in society, is ours.  Similarly, the generation of children to watch, in terms of how children of same-sex households fare in society, will be the one right behind us.  Just because something used to be normal, doesn\'t mean it is anymore.  And that would be my two cents on Ms. Tushnet, absent a lesson in What Not to Wear.  Your thoughts?  -Fabs  '),(6633,3802286,'29,June,2004','2004-06-29 00:00:00','   Ann Coulter\'s new book,  How to Talk to A Liberal (If You Have To)  comes out soon.    Since Ms. Coulter is busy telling everybody how to do everything, and casually redefining terms like \"Slander\" and \"Treason\" (to say nothing of the word \"Idiot\"), perhaps she should consider publishing  urlLink a magazine on living .  A gift from last year\'s  urlLink Koch Blog .  '),(6634,3802286,'29,June,2004','2004-06-29 00:00:00','   I can no longer find this article, so I am going to rip it off:    The discovery that affiliation with the Republican Party is genetically determined was announced by NURTURE, causing uproar among traditionalists who believe it is a chosen lifestyle.  Reports of the gene coding for political conservatism, discovered after a decades-long study of quintuplets in Orange County, CA, has sent shock waves through the medical, political, and golfing communities.  Psychologists and psychoanalysts have long believed that Republicans\' unnatural disregard for the poor and frequently unconstitutional tendencies resulted from dysfunctional family dynamics -- a remarkably high percentage of Republicans do have authoritarian domineering fathers and emotionally distant mothers who didn\'t teach them how to be kind and gentle.  Biologists have long suspected that conservatism is inherited. \"After all,\"said one author of the NURTURE article, \"It\'s quite common for a Republican to have a brother or sister who is a Republican.\"  The finding has been greeted with relief by Parents and Friends of Republicans (PFREP), who sometimes blame themselves for the political views of otherwise lovable children, family, and unindicted co-conspirators.  One mother, a longtime Democrat, wept and clapped her hands in ecstasy on hearing of the findings. \"I just knew it was genetic,\" she said, seated with her two sons, both avowed Republicans. \"My boys would never freely choose that lifestyle!\"  When asked what the Republican lifestyle was, she said, \"You can just tell watching their conventions in Houston and San Diego on TV: the flaming xenophobia, flamboyant demagogy, disdain for anyone not rich, you know.\"  Both sons had suspected their Republicanism from an early age but did not confirm it until they were in college, when they became convinced it wasn\'t just a phase they were going through.  The NURTURE article offered no response to the suggestion that the high incidence of Republicanism among siblings could result from their sharing not only genes but also psychological and emotional attitude as products of the same parents and family dynamics.  A remaining mystery is why many Democrats admit to having voted Republican at least once -- or often dream or fantasize about doing so. Polls show that three out of five adult Democrats have had a Republican experience, although most outgrow teenage experimentation with Republicanism.  Some Republicans hail the findings as a step toward eliminating conservophobia. They argue that since Republicans didn\'t \"choose\" their lifestyle any more than someone \"chooses\" to have a ski-jump nose, they shouldn\'t be denied civil rights which other minorities enjoy.  If conservatism is not the result of stinginess or orneriness (typical stereotypes attributed to Republicans) but is something Republicans can\'t help, there\'s no reason why society shouldn\'t tolerate Republicans in the military or even high elected office -- provided they don\'t flaunt their political beliefs.  For many Americans, the discovery opens a window on a different future. In a few years, gene therapy might eradicate Republicanism altogether.   My Take:  This is obviously a poorly written article about a poorly conducted study in a left-leaning magaizne of some kind.  The idea is interesting though.  I have often wondered about what determines political affliation.  I would think that parental affliation (as a proxy for the politcal atmosphere at home...not genetic makeup), income, leanings of  secondary educational institution (i.e. liberal high schools probably encourage liberalism...of course they also attract liberal students), demographics of the area surrounding home,and religion would be decent predictors, but I really do not know.  '),(6635,3802286,'29,June,2004','2004-06-29 00:00:00','   Now that the listserv abusers have been identified, I would like to plead with all of you to post any interesting information or articles you may like to share with us here... on the blog .   As far as I\'m concerned, events are still appropriate for the listserv (feel free to comment!). I really love all of you, but I think I speak for many when I say that my day would improve tenfold if I didn\'t have to unclog my inbox everyday. Thanks.  '),(6636,3802286,'29,June,2004','2004-06-29 00:00:00','   After last week\'s lecture on \"The Role of Government During War and Crisis,\" I came away with the sense that Mr. Higgs\' message was deeply fatalistic. Despite numerous prods on my part at dinner, I retained the sense that his worldview did nothing but describe the world in all its totalitarian gory (glory?), without providing much hope or room for improvement at all. What follows are the substantive excerpts from our email exchange, ripe and ready for your insights and epiphanies.  *  Subj:  Is the system doomed?   You spoke of the way Republicans buy Democratic support for the War on Terrorism via boondoggles to illegitimate special interest groups. You said, too, that this sort of arrangement was inevitable in a two-party system.   My question is what kind of system would you recommend in which a third party *could* sustain itself? Would it be a system of proportional representation with pluralistic coalitions, like that in Israel and Germany? Would it be a form of anarcho-capitalism? How would we break free of the self-perpetuating expansionary government under which we currently suffer and cannot seem to abandon?   * Subj:   Your queries   Dear Diana,   First, I do not believe that the party structure of government (two parties; proportional representation with multiple parties; etc.) makes a critical difference in the process I was describing. So long as public opinion supports a big governmental response to crisis, then the government in power will make that response, paying off whatever opponents occupy positions from which to block it (if those opponents should dare to do so; they too occupy a tricky position, and they can do themselves grave political damage by appearing to be too obstructionist during a national crisis). My main point is that in a crisis, the normal frictions of political give and take are smoothed by an all-around sense that \"something must be done,\" even if getting it done requires cutting some deals that ordinarily would not be cut. Crisis, so to speak, bursts the dam that normally holds back government expansion. Whatever the party setup, those same forces tend to prevail. That\'s why the ratchet effect may be observed in the history of all the economically advanced Western countries, despite their differing political institutions (recall my graph of Sweden\'s experience).   As for a third party, in this country such parties are virtually impossible to sustain except at the utter margins of politics. The two major parties have engineered the system (voting districts, requirements for getting on the ballot, etc.) in a jillion ways to ensure the continuing domination of the two-party cartel. Their mutual profit from this setup helps to account for the countless loot-splitting mechanisms now built into the system.   As for anarcho-capitalism, that is not a genuine option. Only a handful of dreamers and idealists even know what it is, and if it is explained to others, 99 percent of them will reject it out of hand. It\'s what the politicos call a nonstarter.   Hence, we won\'t be \"breaking free.\" The present system bids fair to persist indefinitely in more or less its present form: ebbing and flowing between tyranny so severe that it restricts economic prosperity, on the one hand, and sufficient room for entrepreneurial maneuver to allow economic progress, if only at a slow pace, on the other hand. Extreme collectivism is constrained by economic infeasibility; extreme market reliance is constrained by a statist dominant ideology and its expression via democratic politics. We are locked into a bad situation, though it is not so bad as it might be (and has been in the USSR, China, Cuba, and so forth, where tyranny produced very bad outcomes indeed).   I regret that I can\'t be more hopeful; uplift is not my specialty. Every person on earth, however, can embrace a personal commitment to justice, though the whole world seem dedicated to injustice. Although we have to pay taxes and obey laws (more or less) to avoid physical violence against us by the established rulers, nobody must give affirmative support to the tyranny in which we find ourselves enmeshed. Though we are slaves in body, we need not accept that status in spirit. We ought never to bow down to the bastards.   Robert Higgs  '),(6637,3802286,'29,June,2004','2004-06-29 00:00:00','   In 1729, Edmund Burke noted     \"The only thing necessary for the triumph of evil is for good men to do nothing.\"    We are here to prove him wrong. We are the few, the committed, the mighty, the proud, the strong. We are the Charles G. Koch Summer Fellows of 2004, and this is our forum. Welcome and be warned.  '),(6638,3802286,'30,July,2004','2004-07-30 00:00:00','   Fellows! Revel in the brilliance of you peers!  urlLink Tim Dorsett publishes an article in Peoria  and  urlLink David J. Healy is laying the smack down at the Capital Research Center . Remember, you can check out all the fellow publications at the  urlLink Freewebs site .  '),(6639,3802286,'29,July,2004','2004-07-29 00:00:00','   Last night at the Koch Fellows speaking event with Clinton Bollick of the Alliance for School Choice and founder of the Institute for Justice, I wanted to raise a few key points about the school&nbsp;\" urlLink Voucher Wars \" in America today.   By far, liberals and Demcrats represent the main opposition to voucher plans across the US.&nbsp; I tried to flesh out the following to points during the question and answer:   1) A John Kerry presidency and its consequent power of appointment would be devasting to&nbsp;IJ causes such as school vouchers&nbsp;in courtrooms across the US.&nbsp; (Bollick\'s two favorite justices are Thomas and Scalia.&nbsp; He was able to win support of Kennedy, Rehnquist, and O\'Connor with the liberal justices dissenting in the 5-4 Zelman v. Simmons Harris voucher case.)   2)  urlLink 6% of the Dem top brass at the Democratic National Convention&nbsp; (10% as&nbsp;Bollick pointed out, if one includes the AFT with the NEA) are teachers unions.&nbsp; Along with \"church-state separation\"&nbsp;groups like the ACL(iberal)U and  urlLink People for for the (UN) American way, these are the two biggest blockades to voucher plans .&nbsp; The teachers unions represent an especially entrenched and organized opposition to&nbsp;school vouchers.   I\'m sure that&nbsp;the partisan nature&nbsp;of my questions as always&nbsp;may have made people who have considered themselves Democrats (or Kerry voters)&nbsp;or who have grown up in places/hold views&nbsp;that are more hostile to the Republican label, a little uncomfortable.&nbsp; (That is one reason why we have a disproportionate number of people from the south and who are Republicans in our program as compared to Democrats and even true libertarians)&nbsp; On one hand, making some people a little uncomfortable is what I was trying to do with my questions as always.   I also wanted to take what&nbsp;Bollick was saying and apply it to the&nbsp;political reality of&nbsp;the two party framework.&nbsp;&nbsp; Save my question and a few hints or allusions by Bollick throughout, he really was non partisan, although conservatives and Republicans (although not perfect) have shown themselves to be more supportive of Bollick\'s causes.&nbsp;   Fabio mentioned affirmative action below.&nbsp;  urlLink Libertarians side with conservatives in opposing affirmative action  as evidenced by&nbsp;this Cato article&nbsp; debunking \"Affirmative Action\'s Myths\" that ran&nbsp;on Fox News.&nbsp;&nbsp;  '),(6640,3802286,'29,July,2004','2004-07-29 00:00:00','   Affirmative action: without it, they say that minority students who apply to Americaâ€™s elite colleges and universities will be denied, shut out by a â€œglass ceilingâ€ of prejudice.    At the same time civil rights leaders worry that minority students will be trapped in lesser institutions of higher education, top flight schools like Harvard, Yale and my own Bates College, are scrambling to make sure that minority students are not only attending their schools, but are increasing in visibility.   I never knew it when I came to Bates, but I was accepted to be one of those â€œvisibleâ€ minority students.    An e-mail I received from the admissions office told me that as one of my classâ€™ â€œaffirmative action admitsâ€ I was requested to participate in a study and seminar on diversity at Bates.    I ignored the e-mail, and I kept ignoring the repeat e-mails until I received a call from the office; an appointment was set and I was expected to show.     Though I felt degraded and insulted to be designated an â€œaffirmative action admitâ€ after four hard years of work in high school, I am glad that I went to that meeting.    I realized then and there, that no matter how many programs we create to thrust minority students into higher education, real diversity does not exist in American high education.     When I reported to my survey discussion group, I was asked if I felt that Bates was a diverse place and whether or not I felt comfortable as a minority at Bates.    Did I have any reason to feel uncomfortable?    I hadnâ€™t thought so, though raising the question immediately made me feel that, perhaps, all those smiles and waves as I walked around campus were not polite ways of saying, â€œHi, how are you,â€ but â€œOh, how pleasant, thereâ€™s a little dark one running about.â€     What affirmative action advocates miss is that diversity truly exists when issues of race need not be discussed.    Real diversity is recognizing and awarding the talent and intellect in individuals without considering their race, that as Martin Luther King Jr. said, it is â€œthe content of their characterâ€ that matters, â€œnot the color of their skin.â€    Such vigorous attempts at forcing diversity onto a campus through discussing issues of race only amongst minority students serves only to stigmatize, not integrate, them.         The reality is that the glass ceiling that used to prevent minority students from ascending to the highest ranks of our society no longer exists.    A qualified African-American or Latino student is just as able to get into Harvard or Yale or even Bates if they have distinguished themselves academically as any Caucasian student.    That many black and Latino students arenâ€™t present at these institutions is due more to the glass ceiling that exists in their own mindsâ€”the belief that a private school will reject them or that a private education is completely unaffordable.         Yet, both of these attitudes are simply without merit: academically qualified students will gain acceptance and if finances are an issue, more and more schools offer ridiculous amounts of aid to help pay for the cost of tuition.    The onus on creating diversity is not on the institutions, where attempts, like the one I experienced at Bates, lead to more stigmatization.    The onus on creating real diversity in higher education is on students themselves, regardless of race.     All in all, I consider Bates to be a diverse place; my friends come from all different parts of the country and the world.    My debate partner, for instance, is from Pakistan.    But what makes me appreciate the diversity I experience at Bates is that I can take it for granted: I know that my debate partner is from Pakistan and I am interested to learn about where she comes from, but my respect for her culture need not make an issue of her culture.    I respect her because of our common humanity.    In the interest of true diversity, colleges and universities across America should pay the same respect to minority students: to their humanity, not their race.      urlLink Cross-posted .  '),(6641,3802286,'29,July,2004','2004-07-29 00:00:00','   If you would like to see me rant about the soul-crushing evil of politics,  urlLink go here  and add your own message of disillusioned frustration.  '),(6642,3802286,'29,July,2004','2004-07-29 00:00:00','   Watch Kerry\'s Iraq War Documentary as he flips...flops...and then flips again on his rhetoric regarding the Iraq War.  It is an incredible, must watch, and I can\'t emphasize it enough.   urlLink https://webmail.wharton.upenn.edu/exchweb/bin/redir.asp?URL=http://demsextrememakeover.com/  (\"Kerry\'s Iraq War Documentary\")  I just can\'t get over how firm Kerry\'s support of the war used to be before he became his shameful post-Veitnam self.  (Now again, he supports the war by the way.)   It is all about political expediency and political opportunism.  The Democrats and John Kerry needed something to criticize President Bush on.  It\'s grotesque how they are misrepresenting the facts and misleading the American public Michael Moore-style.   Copy, paste, and forward this, so that others can see Kerry\'s waffling war progression.  '),(6643,3802286,'28,July,2004','2004-07-28 00:00:00','   Last night while watching the DNC Convention, I was more than a little revulsed, when Ron Reagan tried to pigeonhole opposition to/reservations toward stem cell research as&nbsp;originating from squarely from religious peoples\' beliefs, which as he called them- were \"nothing more than beliefs\".&nbsp;   What got me was not so much that&nbsp; urlLink the prodigal son&nbsp;was using junk science to misrepresent the Bush administration\'s position on&nbsp;stem cell research&nbsp; &nbsp;(article written by a Cato guy) or that the only reason that&nbsp;Reagan was speaking was because Democrats were shamefully trying to capitalize on the good name and death of the greatest modern American president; what got me more than a little peeved was half an hour or so later, when the Democrats had an Imam reading from the Holy Koran. Democrats suddenly displayed a grossly contradictory and ostentatious act of religious solemnity like never before.&nbsp; Talk about shameless pandering to Arab voters or \"diversity\" or whatever they were trying to do!&nbsp;&nbsp;In any event, I want to go on record saying that was just about&nbsp;most respect that I have ever seen religion accorded by the Democratic&nbsp;top brass&nbsp;ever!&nbsp;   Just think, would Democrats ever ask an evangelical to come and read from the Bible at the DNC convention? Answer: Never.&nbsp;   Just think, what would happen if the Republicans brought a priest to their convention to give a sermon? Answer: Democrats would have a hissy fit and cry foul.   This basically goes to show that in the post Post 9/11&nbsp;world,&nbsp;Islam has become the unofficial religion of the Democratic party.&nbsp;   On a related note, Yoga is also a \"spirituality\" of the Democratic Party (see also my post from a while back&nbsp;titled  \"Yogis\" Courtesy Michelle Hinman ).&nbsp;&nbsp;Here is what was&nbsp;posted on the Kerry website a while back:&nbsp;   [Yogi Jenn Joos, Oregon  Jenn Joos is a Kundalini Yoga teacher in L.A. and hosts Kerry Fundraisers.   â€œ I felt that the yoga community could be a resource to reach out and talk about the important issues in a positive, constructive way. Often when politics were discussed at the yoga center after class, there was frustration at how things were in the world...but no solutions. I saw it as an opportunity to educate ourselves on issues, talk about the possible solutions and not feel so powerless! or hopeless. Kerry is an antidote for both politics-as-usual and a boost for the growing Yoga or â€œSpiritual Leftâ€ movement. Yogis believe in positive change. It starts as peace and mindfulness within ourselves by practicing Yoga, and filters out to help reach out to others in the world around us. We have a wonderful chance to help financially, but also give a message that â€œyoga votersâ€ want peaceful world resolutions, socially conscious programs and resource sustainability for our future .\"]   I thought these values/expressions of faith by Democrats were just plain goofy and twisted.  '),(6644,3802286,'27,July,2004','2004-07-27 00:00:00','   Where\'s our  urlLink invite ?  '),(6645,3802286,'27,July,2004','2004-07-27 00:00:00','   My organization,  urlLink The National Taxpayers Union, was cited in the&nbsp;Wall Street Journal&nbsp;yesterday . The article points out that Kerry promises higher taxes and a \"grab bag\" of 76 new spending proposals.   A Report by  urlLink NTU  claims that Kerry\'s proposals would increase government spending by $226 billion in his first year in office, about $2,000 per American family, or 10% of the federal budget.&nbsp; The author estimates Kerry\'s proposals at 2 trillion from 2005 to 2014. This is much more spending than President Bush, granted many of the bills that Bush has signed have been appalling costly.&nbsp; (I very much disliked the Medicare and Education Bills and want to see him veto the highway bill or it not get through congress at all.)   The&nbsp;article goes on to point out that in flip-flopping fashion, Kerry is trying to have it both ways by presenting himself to voters&nbsp;as both a free-spending liberal and a deficit hawk.   Here is another great article&nbsp;that points out the phony  urlLink Democratic storyline (talking points) about how&nbsp;the economy has fared from Bill Clinton\'s two terms until present . (On an economic sideote, the&nbsp;consumer confidence level just rose to&nbsp; urlLink its highest level&nbsp;since June 2002 .)  Anway, here are&nbsp;three great points that get blurred as Democrats have&nbsp;attempted to misrepresent Bill Clinton\'s economic legacy:   1) The Republican Revolution in Congress was more responsible for balancing the budget and reforming entitlments (welfare) than Bill Clinton.   2) Clinton presided over the dot com boom.&nbsp; The roaring economy had little if anything&nbsp;to do with his governance and lots to do with the private sector (free markets).&nbsp; (Alan Greespan was a carry over from Reagan).   3) The recession that continued through the beginning of Bush\'s first term has since been revised back to have begun on Clinton\'s watch.   On a final note, Democrats thrive on class rhetoric.&nbsp; I would like to offer some great advice for the insanely rich people such as the Kerry\'s who don\'t think that they pay enough in taxes.  urlLink GIVE IT BACK .&nbsp; In fact, supposedly a Massachusetts law allows them to do just that, but I wouldn\'t expect the Kerry\'s to actually do this anytime soon.  '),(6646,3802286,'27,July,2004','2004-07-27 00:00:00','   I\'ll admit that I have not been the most adept commuter, but is that any reason for the Metro station managers/attendents to be rude to me?&nbsp;   I\'ve tried smiling, calling them by the name on their nametag, asking them how they are, but to no avail.&nbsp; They just suck.&nbsp; I have met only one nice man ALL SUMMER who works at the Farragut West stop in the evenings, but that is it.&nbsp;   Why this tirade on the Koch blog?!&nbsp; For the obvious reason that the Metro customer service sucks because it can.&nbsp; Because it has a tyrannical monopoly on commuter travel that gives it no incentive to cater to us...at all.&nbsp;   I don\'t see how this can be improved though.&nbsp; Competition isn\'t really an option because of the limited number of rails, so privatization is really our only hope.&nbsp; But would that even improve the state affairs that much?&nbsp; Thoughts?  '),(6647,3802286,'27,July,2004','2004-07-27 00:00:00','   A reminder that I will be watching  The Manchurian Candidate  (1963 version) Wednesday night, and all are welcome!  Room B-503, 8PM.    Plz leave partisan name-calling at the door.  :-)  '),(6648,3802286,'26,July,2004','2004-07-26 00:00:00','   Koch Industries, which funds the  urlLink Koch Fellows Summer Program , &nbsp;is a decidely&nbsp;Republican/Libertarian/Conservative&nbsp;Organization.&nbsp; It is also well known that for generations&nbsp;the Koch Family (Fred, Charles, and David), which funds our endeavors,&nbsp;has held conservative/libertarian economic beliefs that put them much closer to Republicans than Democrats.&nbsp;   Fred Koch was a member of the  urlLink John Birch Society , which a liberal magazine such as  The Nation  might call \"ultra right\".&nbsp; (Note too  urlLink in this muckraking, leftist, conspiracy theory of an article&nbsp;from  The Nation &nbsp; how Koch supposedly pumped loads of money into the National Republican Senatorial Committee, helped Bob Dole become Majority Leader, and aided the candidacies of other prominent midwestern Republicans.)   David Koch helped found  urlLink Citizens for a Sound Economy .&nbsp; This is an organization with links to articles by Dick Armey, Jack Kemp, and Bill Bennet&nbsp; on its home page and which has at least two clearly, anti-Kerry/Edwards reports on its home page, placing it decisively in the conservative column. (Recently,&nbsp;CSE has been in hot water for allegedly&nbsp;helping the&nbsp;Nader Campaign get on the ballot in Oregon.)   Clint Bolick, a conservative who the Koch Fellows will be fortunate to hear speak, made his name fighting the conservative cause against affirmative action.&nbsp; Then Koch money enabled him to found the  urlLink Institute for Justice , which has come to be viewed as the conservative equivalent of the ACLU.   Koch has also poured lots of money into the Tax Foundation and the Federalist Society, both decidedly conservative groups as well, and also Cato, though libertarian, is much more conservative than liberal on most issues.   Since 1998 alone,  urlLink Koch Industries has given George Mason (one of the most conservative law schools in the country)&nbsp;$14.4 Million dollars .&nbsp; About half of that money went to the Mercatus Center.&nbsp; Presumably a lot of the rest went to the Insitute for Human Studies which makes our (including mine)&nbsp;Koch Fellowships possible.   Koch Money has aided our current Republican president.&nbsp; When President Bush&nbsp;took office in 2000, his administration created a \"hit list\" 23 anti-free market regulations&nbsp;to scrap.  urlLink 14 of the 23, which were ultimately targeted&nbsp;were reccomended termination by Mercatus .   An Arcticle from CNS News points out that  urlLink Koch Industries is a major GOP donor .&nbsp; Koch donated $332,000 to Republican congressional candidates during the 2000 election cycle, David Koch gave $250,000 to the Republican National Committee between 1999 and 202, and the two&nbsp;gave a total of $6,000 to President Bush\'s 2000 Campaign.   This last article also points out near its end how the Clinton Administration possibly targeted Koch Industries for its political support of Republicans.&nbsp; Because of its support of  urlLink \"the country\'s most prominent conservative and libertarian voices\", as this article calls it , a politically motivated Clinton Adminstration issued a 97-page indictment against Koch Industries in 2000.&nbsp; This came after Democrats in the Senate launched an investigation of Charles Koch in 1996. Luckily, when the  urlLink Bush administration took office in 2000 the charges were promptly settled .   So why am I pointing out all of this about stuff about the vast and intricate connections between Koch money and Consservatives/the Republican Party?   If you ask me, Koch Fellows who call themselves Democrats or say they are registered that way and even more so go on spewing liberal rhetoric and arguments across our&nbsp;blog&nbsp;have a lot of gall. Here they are in a program which&nbsp;places its fellows at organizations, a preponderance of which are conservative, and therefore more Republican than Democratic, and they&nbsp;obviously don\'t hold dear the&nbsp;same&nbsp;values and political philosphies, which the Kochs no doubt hoped that they would be fostering through the munificient contributions.&nbsp;   The Koch family is a generous benefactor to us all and I think that they might be dissappointed to see Koch Fellows supporting liberal orthodoxy and Democratic politicians (such as Fabio\'s lauding of Bill Clinton, etc.) who have been antithetical to Koch principles and even hostile to the Koch Industry.&nbsp; I have read some of the things that have been&nbsp;posted on the blog (and I am not referring to posts by all Koch Fellows) and see it as a big slap in the face to our benefactors at Koch.&nbsp;   Fabio asked me the rhetorical question a while back \"who am [I] to speak for Koch Industries\"?&nbsp; The answer is that I am a beneficiary of their generosity and a like-minded supporter of the same political causes as the Kochs.&nbsp; I am also proud to say that the Koch money that has been spent on me is money much-appreciated, money well spent.&nbsp; I feel good saying this as one who is loyal, not traiterous&nbsp;to Koch\'s many causes.&nbsp;  '),(6649,3802286,'26,July,2004','2004-07-26 00:00:00','   1. &nbsp; At the risk of sounding like a bleeding heart liberal, I do hope that my fellow libertarians checked out  urlLink Bill Clinton\'s speech  last night at the Democratic National Convention.&nbsp; I did, and I almost cried.&nbsp; Almost.&nbsp; It was almost overwhelming to see Clinton speak for the first time in God knows how long.&nbsp; The man is just eloquent, something our current president cannot lay claim to.&nbsp; It makes me wonder about that eternal question of the political elite: Whatever happened to statesmen?&nbsp; What happened to great orators who filled halls with words that rang through the ears of their audience like music?&nbsp; Clinton has that, as does Hillary, and it is that quality that I hope John Kerry and John Edwards can return to the White House.    2. &nbsp; Last night a friend and I had a friendly exchange about free market capitalism and whether or not it would work if it were given the chance.&nbsp; To me, the ideas of von Mises or Hayek and those like him are no different, except in method, than the ideas of Engels or Marx.&nbsp; I recognize that there is a difference between capitalism and communism, but what essentially remains the same is that both systems are based upon ideals-- ideals that humans may strive for, but never actually reach.&nbsp; It\'s akin to what Nietzsche said about perfection, that to reach it impossible, however much we might strive.&nbsp; Humans strive to live up to their ideals but never really do.&nbsp; So, on paper a completely free market system would work, while in practice it most likely, like its communist-ideal counterpart, would not.&nbsp; The problem with economists and academics is that their ideas remain in the realm of ideas, rarely ever actually put to test to see if they stand up to the ultimate scrutiny: the real world.   Cross-posted on urlLink  my blog .  '),(6650,3802286,'26,July,2004','2004-07-26 00:00:00','   Do we really need party conventions?&nbsp; Or are they vestigial anachronisms of politics before mass media and the Internet?&nbsp; Jonah Goldberg writes a  urlLink well-written piece &nbsp;on their obsolescence.  I&nbsp;pick at&nbsp;an earlier Goldberg piece at  urlLink Obernews .  '),(6651,3802286,'26,July,2004','2004-07-26 00:00:00','   Ann Coulter was hired by  USA Today  to write a conservative column about her experiences at the Democratic National Convention.&nbsp; For her first piece, Coulter chose to write  urlLink this , and  urlLink was promptly sacked .  Coulter claims that \" USA TODAY  doesn\'t like my \'tone,\' humor, sarcasm, etc., which raises the intriguing question of why they hired me to write for them.\"&nbsp;   Tell me, since when and on what planet is it funny to make derogatory references to gay men and simply insult everyone with whom you disagree?  '),(6652,3802286,'26,July,2004','2004-07-26 00:00:00','   urlLink Some profoundly insightful analysis on the gay marriage debate.  '),(6653,3802286,'24,July,2004','2004-07-24 00:00:00','   Daniel Okrent has in i urlLink nteresting piece in the  New York Times  on liberal bias in the  New York Times  . Okrewnt admits the bias of the paper, but says that it is \"urban\" not just \"liberal,\" and that the urbanization of the paper is a direct result of the city from whence it sprang. A short excerpt:    I\'ll get to the politics-and-policy issues this fall (I want to watch the campaign coverage before I conclude anything), but for now my concern is the flammable stuff that ignites the right. These are the social issues: gay rights, gun control, abortion and environmental regulation, among others. And if you think The Times plays it down the middle on any of them, you\'ve been reading the paper with your eyes closed.'),(6654,3802286,'24,July,2004','2004-07-24 00:00:00','   More shameless self-promotion. Over at my site I have initiated a new feature:  urlLink Profiles in Liberty and Profiles in The State . My first profile is up and it is  urlLink an evaluation of  Esquire  writer at large, Tom Junod . Comment!  Also, I am sure she will mention this, but  urlLink Sara Hinson has a new weblog called Melete . It is great! Good job Sara. We know you will continue to fight the good fight for liberty.  '),(6655,3802286,'23,July,2004','2004-07-23 00:00:00','   That Republicans are sometimes&nbsp;marginally better at spending restraint than Democrats is not enough, folks.  From  urlLink Radley , a  urlLink list of pork and big-government initiatives from the GOP.    urlLink  &nbsp;  '),(6656,3802286,'22,July,2004','2004-07-22 00:00:00','   I thought I\'d never post on marriage, but that was before I had something wildly unpopular to say about it.&nbsp; On Wednesday,  urlLink Erica posted  about the alleged weakening of the institution of marriage in the Netherlands following&nbsp;the legalization of&nbsp;same-sex marriages,&nbsp;citing the separation of marriage from parenting as the link.&nbsp; Let\'s ignore the debate on the veracity and significance of the story for the moment and stick it in the \"Gay marriage weakens traditional marriage\" pile of argumentation, and assume there\'s at least some truth and some significance to it (that is, there is reason to infer causation, and the political/cultural situation in the Netherlands is sufficiently similar to predict the same results in the US or universally).   So what?     What if gay marriage really does undermine straight marriage?&nbsp; In order to make some sort of judgement about the effects of gay marriage, assuming we knew what those were, we have to answer a few questions.&nbsp; First,&nbsp; how much do we value straight marriage, particularly as evaluated in the survey cited above?&nbsp;   If we\'re judging the \"strength\" of straight marriage by indicators like the number of divorces and births out of wedlock, allow me to offer some social variables that  strengthen  straight marriage.&nbsp; Where there is ess financial independence for women, divorces will be less common.&nbsp; Where women have fewer choices and opportunities for education or work, more will seek security through marriage.&nbsp; Excision certainly cuts down on infidelity.&nbsp; And hey, for a sure-fire way to ensure the \"strength\" of marriage, as evaluated by births out of wedlock, try honor killings and radically enforced sharia ( see also : northern Nigeria).&nbsp; If we assigned an infinite value to stable straight marriage, we wouldn\'t think twice about the necessity and value of the above.&nbsp; At some point, individual, family, and community happiness come before the strict indicators of a stable marriage used in the study.   Secondly, then, we have to ask how much we can afford to weaken marriage.&nbsp; If we could prove, for instance, that legalizing gay marriage would definitely cause exactly one more divorce each year, I think most of us would agree that extending the right to marriage (or the marriage \"franchise\")&nbsp;to&nbsp;the homosexual population&nbsp;would be worth it.   What is missing between the study\'s&nbsp;conclusions (again, assuming truth and significance) and \"gay marriage ought not to be legal\" is something like this:  The legality of gay marriage weakens straight marriage to a point below what is optimal.&nbsp;&nbsp;As such gay marriage also diminishes the individual, family, and community happiness that we hold as ends, to which marriage is a very good but imperfect means.&nbsp;    Moreover, it might be important to demonstrate that the&nbsp;reduction in happiness above is greater than the increase in happiness resulting from lots of gay people finally being able to marry.&nbsp; And while I will abstain from getting all post-modern on my small audience, I should also add that it\'s important to ask what constitutes good parenting, because wedlock certainly isn\'t both necessary and sufficient to that end.   My point is not that undermining or weakening straight marriage is a good thing, nor, more specifically, that more divorces and births out of wedlock are good things.&nbsp; They\'re almost always bad things.&nbsp; But since we don\'t assign and infinite value to mere marriage and births in wedlock, the implications gleaned from the cited study aren\'t nearly enough to merit substantive conclusions.  '),(6657,3802286,'22,July,2004','2004-07-22 00:00:00','   This  urlLink article  summarizes everything I could possibly say about my views on the \"War on Terror.\"   Honestly, I think it\'s pretty fair too. It doesn\'t really attach blame to anyone, just an analysis of US foreign policy in general.  Thoughts?  '),(6658,3802286,'22,July,2004','2004-07-22 00:00:00','   Itâ€™s a word thrown around by those on the Left who donâ€™t want to be called radical or liberal or socialist.    Itâ€™s a word that shields its user from criticism by taking an unarguable position.    It is a fraud, a misrepresentation, and a copout.    Progressive: This is the culprit in need of handicapping.               urlLink Continue reading...  '),(6659,3802286,'21,July,2004','2004-07-21 00:00:00','   After reading Meg\'s post on the Jacobsen story (you remeber, the Syrian musicians the woman thought were terrorists) I was alert to  urlLink this story  on NRO.&nbsp; Turns out, they  were  just musicians on their way to the show.&nbsp; Nothing nefarious here, people.&nbsp; Just thought I\'d share.   CORRECTION:&nbsp; Oops, I was wrong.&nbsp; There is something nefarious.&nbsp; The&nbsp;casino they played at booked  Carrot Top,  too.&nbsp; That\'s just wrong.  '),(6660,3802286,'21,July,2004','2004-07-21 00:00:00','   To take the blog to a happier place -- if only for a minute -- here is the newest from \"The Hill.\"&nbsp;  urlLink DC\'s 50 most beautiful people .&nbsp; Now taking your corrections, criticisms, or suggestions. :)  '),(6661,3802286,'20,July,2004','2004-07-20 00:00:00','   Since we\'ve been talking about whether or not the New York Times is biased, I though I would include this.  urlLink http://www.foxnews.com/story/0,2933,126249,00.html    How come Fox News is the only news source that reported that Sandy Berger took papers out of the National Archives by putting them  in his socks ? I think that\'s a pretty important distinction. I can maybe buy that he walked out with the papers in his pockets unknowingly...maybe. But his socks?  It seems that anyone reading Fox News and reading that he put the papers in his socks would immediately assume that Berger was lying and intentionally trying to remove documents.  Just thought I would present some \"Fair and Balanced\" bias against both the New York Times and FoxNews.   Personally, I really like the Wall Street Journal because it is generally objective.&nbsp;The Financial Times too for that matter. Both are excellent papers. I&nbsp;try not to&nbsp;read NYT or watch Fox News though.  '),(6662,3802286,'20,July,2004','2004-07-20 00:00:00','   Sounds like something out of The Onion.&nbsp; Only it isn\'t.&nbsp; A drunken Russian flight crew beat up a passenger.   Here\'s the  urlLink CNN story .&nbsp; Found this over at Marginal Revolution.&nbsp;  '),(6663,3802286,'20,July,2004','2004-07-20 00:00:00','   I have noticed a distinct bias from the NY Times on how they have chosen to cover the Sandy Berger revelations.&nbsp;    urlLink Check out their article .    urlLink Then check out the Wall&nbsp;Street Journal Story .   It is amazing how differently they chose to frame, report, and write about the same story.   The Times try to play it off as a \"furious\" Republican attack and tries to play up&nbsp;Berger\'s \"tireless service\", even if he could have been devious/incompetent.&nbsp;   It also tries to downplay his connection to the kerry camp.&nbsp; Anytime the mention that he was an advisor they have to use the word \"informal\".&nbsp;   The Times using the word \"mishandling\" to describe stealing/smuggling and destroying highly classified documents, which is a felony, and then doesn\'t tell what he is actually accused of doing- stuffing them in his pants- until near the end of the article. It also relied on statements by Berger\'s lawyer, but gave no statements from investigators (they aren\'t legally allowed to, so those by his lawyer should also be left out).   Also, the article doesn\'t&nbsp;put in quotations&nbsp;the statement about his actions being \"inadvertant\", which is implicity saying that this is stipulated fact, which has not been proven.&nbsp;&nbsp; (There could have been motives).&nbsp; Circumstantial evidence and speculation suggests that there was.   Definitely take a look at the two different ways of looking at the story. By reading each one carefully for yourself.   This is the second story by the Times about this same subject.&nbsp; The last write-up was even worse.&nbsp; It said that Democrats were \"willing to give him the benefit of the doubt\" and \"Republicans were less charitable\".&nbsp; It also said that Republicans \"zero-ed in\" on the case as if to imply purely partisan criticism and it place statements questioning the timing of the revelation much higher in the article and gave them more weight than they should have.   Here is a great chain letter that has been circulating about the differences between the major newspapers.&nbsp; I have gotten it three different times over the last year and a half:   1. The Wall Street Journal is read by the people who run the country.   2. The Washington Post is read by people who think they run the country.   3. The New York Times&nbsp; is read by people who think they should run thecountry, and who are very good at crosswords.   4. USA Today&nbsp; is read by people who think they ought to run the country but don\'t really understand the Washington Post.&nbsp; They do, however, like their statistics shown in pie charts.   5. The Los Angeles Times&nbsp; is read by people who wouldn\'t mind running thecountry, if they could spare the time, and if they didn\'t have to leave LA to do it.   6. The Boston Globe is read by people whose parents used to run the country and did a far superior job of it, thank you very much.   7. The New York Daily News&nbsp; is read by people who aren\'t too sure who\'s running the country, and don\'t really careas long as they can get a seat on the train.   8. The New York Post&nbsp; is read by people who don\'t care who\'s running the country, as long as they do something really scandalous, preferably while intoxicated.   9. The San Francisco Chronicle&nbsp; is read by people who aren\'t sure there is acountry...or that anyone is running it;but whoever it is, they oppose all that they stand for.&nbsp; There are occasional exceptions if the leaders are handicapped minority feminist atheist dwarfs, who also happen to be illegal aliens from ANY country orgalaxy as long as they are Democrats.   10. The Miami Herald is read by people who are running another country butneed the baseball scores.   11. The National Enquirer is read by people trapped in line at the grocery store.   I used to get the NY times emailed to me, but I stopped that a while back when they started running John Kerry ads on everything even what was supposed to be their \"objective news\" pieces.&nbsp; The WSJ wouldn\'t do something tacky like that.  '),(6664,3802286,'20,July,2004','2004-07-20 00:00:00','   If everyone remembers our debate a few weeks ago on the matters of same-sex marriage, you might recall the experiment practiced in the Netherlands touted by Mr. Rauch.&nbsp; The Netherlands is the first country to legalize same-sex marriage, having done so in 2002.&nbsp; Though the mere two years that have passed since the practiced has been permitted, the country is already gathering data as to the effects of this&nbsp;campaign.&nbsp;   A letter from Dutch scholars to other nations\' parliaments notes a drastic decline in Dutch marriages, which they attribute to \"the (legal and social) separation of marriage from parenting. In parliament, advocates and opponents alike agreed that same-sex marriage would pave the way to greater acceptance of alternative forms of cohabitation.\"  This evidence flies in the face of Mr. Rauch\'s postulation that allowing same-sex marriage will halt the trend to alternates to marriage.&nbsp;&nbsp; While the scholars note that it may be too early to trace causation in these matters, they advise \"that enough evidence of marital decline already exists to raise serious concerns about the wisdom of the efforts to deconstruct marriage in its traditional form.\"  However, to support these claims, researches can point to the correlations in Denmark, Norway, and Scandinavia between cohabitation with children and the drastic delines in marriage rates.&nbsp;   Hat Tip:  urlLink National Review  and Eve Tushnet at  urlLink Marriagedebate.com  '),(6665,3802286,'20,July,2004','2004-07-20 00:00:00','   urlLink This article  in Tuesday\'s NYT is the latest in a charming trend I call Canting Capitalism.&nbsp; It opens with a delightful mockery of bobos\' infatuation with fashionable urban poverty and squalor:  &nbsp;   Simone Shubuck didn\'t find working in her Lower East Side studio without a bathroom a bit romantic. Actually, she found it weird that visiting dealers and collectors always declared her bohemian deprivation \"absolutely charming.\" As time passed, Ms. Shubuck, now 34, found herself yearning for prosaic luxuries like a toilet, health insurance, a pension plan.  &nbsp;  This is the same reaction we see when first-worlders visit developing countries to ooh and ahh at the wholesome, natural lifestyles of people without modern conveniences like Peapod, Netflix, plumbing, or vaccinations.&nbsp; Absolutely charming.  &nbsp;  The article goes on to describe the Artist Pension Trust, which also received ink/pixels in  urlLink the Economist  [$], and gets a rough treatment from Tyler Cowen  urlLink here  and  urlLink here .&nbsp; It\'s an interesting scheme, and TC provides a number of valid critiques as to its viability as an investment.&nbsp; Whether or not it\'s the best pension plan doesn\'t interest me for the moment.&nbsp; What does interest me is this aspect:  &nbsp;   \"It\'s a way of taking advantage of the capitalistic nature of the market* and mix in a healthy dose of socialism to create a hybrid form,\" said David A. Ross, the former director of the Whitney Museum of American Art and then the San Francisco Museum of Modern Art, who is the Artist Pension Trust\'s president. .... Socialist inclinations aside, the trust isn\'t meant to be altruistic.  &nbsp;  I\'m intrigued by the marketing of capitalism here.&nbsp;  urlLink Mutual Art , the company behind the Artist Pension Trust, targets a generally market-phobic audience who, as one might glean from the article, wouldn\'t ordinarily invest in any sort of pension scheme.&nbsp; Now, there are those of us who find markets elegant, but there is no shortage of people who find capitalism aesthetically unappealing.&nbsp; Pop culture images of robber barons, destitute Dickensian protagonists, exoticization of underdevelopment, and romanticization of \"starving artists\" reinforce this all over.&nbsp; The capitalist ick factor is manifest in the anti-globalization movement,  urlLink numerous films , and clearly, the art world.&nbsp; What\'s interesting is the market\'s adaptation to this set of anti-market preferences.  &nbsp;  Take [Sub]Urban Outfitters, which siphons money from rich teenagers who shell out to look destitute and bohemian.&nbsp; The current fad in \"ironic\"  urlLink $80 t-shirts  stems from a  past  trend of shopping at thrift stores - which was first a trend out of necessity, and quickly became an exercise in irony.&nbsp; There are countless other examples of people spending money to appropriate the charm of socialism.&nbsp; A more mainstream example is  urlLink socially responsible investing , where investors choose firms based on their contributions to their communities, their treatment of animals, their hiring diversity, or anything other than the firm\'s predicted profitability.&nbsp; This allows investors to engage in the market, without that nasty morning-after feeling of exploiting the worker/land/kittens.&nbsp; I see the Artist Pension Trust as simply the latest idea in making capitalism more palatable to those who find poverty and/or socialism more romantic, but who can\'t deny or resist the benefits of a market system.  &nbsp;  So where\'s this amusing hypocrisy taking us?&nbsp; I predict that, given the prevalence of vague anti-market preferences among  urlLink bobos , the rise of bobo culture will bring about more creative ways to be capitalist without the aftertaste of oppression.&nbsp; And eventually, maybe private enterprise won\'t taste so bad to the cultural elite.  &nbsp;  *Ahem, capitalistic nature of the market?&nbsp; As opposed to what?&nbsp; Is that like the bovinistic nature of the cow?  &nbsp;  [Gratiuitously cross-posted at&nbsp; urlLink Brooke\'s\' blog , and  urlLink my own .&nbsp; When you write less than once a month, you really have to milk it.&nbsp; Bovinistically.]   UPDATE: Will Wilkinson takes me on at the  urlLink Flybottle .  '),(6666,3802286,'20,July,2004','2004-07-20 00:00:00','   I have two questions which I hope will start discussion threads.  1) Are any of you familiar with any really prominent black politicians right now--particularly those who don\'t exactly fit the civil rights era style of black leadership?  2) Did John Edwards make money off the tobacco settlements in any way?  '),(6667,3802286,'20,July,2004','2004-07-20 00:00:00','   Yesterday I commented in Scott\'s post on Arnold\'s \"manly women\" remarks about the importance of rhetoric and diction when in politics. And I thought I\'d provide an example of what I was talking about through a  urlLink rhetorical analysis of Bush\'s last (and hopefully last  ever ) State of the Union address.     Take it as you will, but what a person in power says  does  matter more than what someone on the street says. Schwarzenegger should not have been attacked as sexist or homophobic. Rather, he should have been attacked for being tactless and stupid. The same goes for Dick Cheney: these people are our leaders, representatives to the rest of the country and world. Statesmanship requires that we demand a bit more from our leaders in their public behavior and methods of address.  '),(6668,3802286,'20,July,2004','2004-07-20 00:00:00','   I just finished reading the book  Imperial Hubris: Why the West is Losing the War on Terror &nbsp;written by \"Anonymous\", a current, senior CIA analyst with 20+ years of experience examining transnational terrorism, Afghanistan, South Asia, Al Qaeda (AQ) and Usama Bin Laden (UBL).&nbsp;&nbsp; You can read the  urlLink introduction  and  urlLink excerpts .  Anyone can borrow the book from me.&nbsp; I personally find this book very&nbsp;scary because the analysis&nbsp;is spot on- it challenges many of the assumptions I have held about UBL and AQ and paints a deadly accurate picture of the War on Terror.&nbsp; &nbsp; Mr. Anonymous assesses that AQ is stronger than it was before 9/11 in spite of, and to a large extent,  because of  US conduct in the War on Terror.&nbsp; The US government just flat out does not understand the grave threat that AQ really poses.&nbsp; AQ is not a terrorist organization; UBL is not a terrorist- to label either as such underestimates their power.&nbsp;&nbsp;&nbsp; Instead, AQ is the forefront of a global Islamic  insurgency  being waged against the United States and UBL is more Robert E. Lee than Al Capone (i.e.&nbsp;more a cunning, calculating&nbsp;tactician who inspires fanatical devotion among his men rather than a ruthless criminal).&nbsp;  &nbsp; Unfortunately, US leaders still believe that AQ leaders and cells&nbsp;can be hunted down one-by-one and brought to Texas-style justice, and&nbsp;that&nbsp;the removal of terrorist sponsoring states will sap AQ of its&nbsp;fundamental sources of power (or \"center of gravity\" for you military types).&nbsp; Mr. Anonymous points out that AQ\'s center of gravity comes from Muslims\' perceptions of the need to wage&nbsp;a \"defensive jihad\" against the US, not from its leadership, finances, safe havens, etc...&nbsp; UBL has convinced his followers and fighters, and they are many, that specific US policies- &nbsp; 1.&nbsp; US support for Israel 2.&nbsp; US troops on the Arabian peninsula 3.&nbsp; US occupation of Iraq/Afghanistan 4.&nbsp; US support for Russia, India, China against their own Muslim militants 5.&nbsp; US pressure on Arab energy producers to keep prices low 6.&nbsp; US support for tyrannical Muslim govts. (e.g. Saudis, Pakistan) &nbsp; - directly threaten Islam.&nbsp; Whether WE think those policies are a threat to Islam does not matter; UBL has already convinced thousands (perhaps hundreds of thosands/millions?...unfortunately nobody knows) to take up arms in Islam\'s defense because of them.&nbsp; This goes back to the whole \"Why do they hate us?\" argument.&nbsp; I agree with Mr. Anonymous (and with Ron Paul\'s remarks today), those waging war against us hate us because of what we do, not who we are- UBL\'s statements directly echo this.&nbsp; I will be happy to debate this point in more&nbsp;depth with anyone who wants to challenge it.&nbsp;   So what is the solution, Mr. Anonymous? Well, American policy makers need to develop a strategy which focuses on AQ\'s center of gravity.  Neo-conservative strategy to transform the Middle East into something resembling Western, democratic, secular society only fuels Muslim perceptions that Islam is under attack and strengthens AQ\'s center of gravity.  The US government instead needs to address those 6 policies listed above that enable AQ to motivate Mulims to become \"soldiers of God.\"  If US policy makers (and the public) believe that maintaining these foreign policies is much too important and worth the costs of fighting a global insurgency then the only other alternative is total anihilation- \"killing in large numbers\" combined with a \"Sherman-like razing of infrastructure\" in the Muslim world. Half measures like those taken in Iraq and Afghanistan will not do. The US must essentially kill everyone and destroy everything where AQ might be- that\'s the only way to eliminate them b/c they have seamlessly integrated themselves into dozens of societies around the world.  I share Mr. Anonymous\' views and believe that current US policies are a recipe for continued defeat. Anihilition of the Muslim world is not feasible/wise/moral. AQ has already successfully scattered to the four winds. Is there any opportunity left for a decisive military blow? I fear that nobody knows. It\'s time for the policy makers and the public to have a vigorous debate on the strategy for the war on terror (let\'s face it, there is NO current coherent strategy for addressing a global Islamic insurgency) before it\'s too late to employ any alternatives at all.  '),(6669,3802286,'19,July,2004','2004-07-19 00:00:00','   I have found an mp3 of a discussion about the role of obscene content in the development of the internet: &nbsp;  urlLink download mp3  &nbsp; Download, Listen, and Discuss.  '),(6670,3802286,'19,July,2004','2004-07-19 00:00:00','   urlLink Here , for your reading pleasure is the  urlLink latest publication  from our illustrious assistant diretor.&nbsp; I think he wins the prize for  urlLink best publication  this summer. &nbsp; Congratulations, John.  '),(6671,3802286,'19,July,2004','2004-07-19 00:00:00','   Does&nbsp;a spoonful of sugar make the  urlLink smart pill &nbsp;go down?  &nbsp;  A new area of research in memory-enhancing drugs poses the question of what intelligence is, the standard nature/nurture debate as to how much eugenics is worth pursuing, and how much of it is unproductive and illegitimate. I think to call the use of a drug that enhances your memory something out of  Brave New World  is a bit much -- even now, people use gingko biloba, other memory-enhancing&nbsp; urlLink natural supplements , and even fish oil (in Russia, people&nbsp;consumed this in spoons out of bottles in its pure form as a form of vitamin) to increase their memory. Ultimately, I think, one\'s intelligence is not all inborn or biological but much the sum of one\'s experiences. In the Gardnerian sense, broadly defined, intelligence includes not only rote knowledge but also practical, social, and emotive skills, which are generally not packageable.  '),(6672,3802286,'16,July,2004','2004-07-16 00:00:00','   I view Political Correctness from the left as the single&nbsp;greatest threat to free speech in this nation of ours.  &nbsp;  Today, because  urlLink Gov. Schwarzenegger made a joke  based on his hilarious Saturday Night Live skit by referring to the CA legislature as \"girlie men\", there was a resulting Democratic/leftist outcry.&nbsp; Talk about crying wolf.  &nbsp;  I would also like to share two other instances of politically correct overkill.  &nbsp;  Last fall, I argued with my friend, VP of the College Democrats at Penn,&nbsp;til I was blue in the face that it is perfectly allright to use the phrase \"jipped\" as in \"you got jipped\".&nbsp; He was dead seriously trying to tell me that the phrase is discrimminatory because he claims that it assigns a negative trait to a class of people, gypsies in Europe.&nbsp; (Yeah, I know how can I be friends and actually live with someone like that.)  &nbsp;  Also the U of Pennslvania \" urlLink Water Buffalo Case \" is particularly renown.&nbsp; A guy, who was trying to sleep,&nbsp;yelled \"shut up you water buffalo\" out his dorm window to some loud black sorority sisters who were passing by in the street.&nbsp; They sued for harassment, he was suspended, the case dragged on and on for a long time, and caused the university to enact lots of bullshit programs and waste resources.  &nbsp; On a free speech related noted,  urlLink some evidence has emerged that&nbsp;hands free cell phones could&nbsp;be more dangerous&nbsp;than regular cell phones in automobiles .&nbsp; Researchers claim this is because it is not the act of holding the phone, but rather the act of talking that poses any added risk (I&nbsp;think the risk is overstated), so by encouraging&nbsp;hands free sets in parts of the country, the government is actually encouraging people to talk more and take on&nbsp;more of this perceived risk. &nbsp; This form or nanny-state regulation in highly&nbsp;Democratic areas (DC, NYC, parts of NJ) is very much against libertarian principles, not to mention it curbs free speech by making it more difficult&nbsp;for people to lawfully communicate with each other.&nbsp; John Coleman, I might add,&nbsp;posted about the&nbsp;cell phone ban in DC going&nbsp;into&nbsp;effect a while&nbsp;back. &nbsp;  Libertarians should recognize the threat to free speech that the left so often poses.  (And I haven\'t even gone into \"campaign finance reform\".)  '),(6673,3802286,'16,July,2004','2004-07-16 00:00:00','   If you were one of the lucky few to attend the shooting day at Quantico, you\'re in luck. John Murphy emailed me the  urlLink link  to your pictures.&nbsp; So take a look at the evidence, and decide who you would want with you in a back alley.  '),(6674,3802286,'16,July,2004','2004-07-16 00:00:00','   Chuck Colson has an interesting article at  urlLink Town Hall  titled  urlLink Same-Sex \"Marriage\" and the Economy .&nbsp; Colson examines a Congressional Budget Office Report that claims that gay marriage would actually save the country money, because of the marriage penalty and decreased entitlement benefits for married couples as opposed to singles.&nbsp; I\'m not sure that I buy this 100%, but it is interesting to consider.  &nbsp;  Also on a gay marriage related note, David J. Healy and I watched on  urlLink Cspan  a packed room of minority leaders, as they agitated in favor of the gay marriage ammendment.&nbsp; The leaders&nbsp;became nocticeably idignant, when efforts were made by the media questioners which&nbsp;insinuated that opposing&nbsp;gay marriage&nbsp;is \"discrimminatory\".&nbsp; The leaders were even more indignant, when efforts were made to liken the gay marriage struggle to the civil rights&nbsp;struggle.&nbsp; On a related note, I would point out that I read somewhere that, although whites are roughly evenly split on the issue of gay marriage, only about 16% of blacks support it.&nbsp;&nbsp; urlLink Blacks are almost uniformly&nbsp;opposed to gay marriage .&nbsp;  &nbsp;  This is ironic because black activists, although not monolithic,&nbsp;stereotypically and statistically-speaking take the hard-left side&nbsp;on many&nbsp;issues.&nbsp; This is also relevant, because the NAACP (see my post below)&nbsp;does not support gay marriage.&nbsp; More precisely, it&nbsp;has tried hard to avoid the issue.&nbsp;  &nbsp;  If that isn\'t ironic enough, it is further ironic, because I\'m sure that John Kerry did not speak about gay marriage at the NAACP convention the other night (see Justin\'s post below), nor do \"black leaders\" think about gay marriage, when they claim that president Bush doesn\'t stand with blacks on the issues.&nbsp; At least he does on this one, and as he does on as many issues as he stands with white Americans.  &nbsp;  Colson\'s article also made me think of another interesting question.&nbsp; Granted recognizing homosexual \"marriage\" is a lot about coercing/convincing society to accept homosexuality in general and to legitimize homosexual relationships in the public eye, I am also wondering,&nbsp;if these tax benefits could lead people/interest groups to press for these same monetary benefits to be made available to&nbsp;domestic partners (i.e. roomates etc.) and other social units, if they were to be made available available to \"married\" homosexuals.&nbsp;  &nbsp;  To end on a separate, marriage related note: liberated, man-hating feminists be damned, it seems that more and  urlLink more married women are deciding to keep their husband\'s last name .&nbsp; This bucks a trend that began during the culturkampf of the 1960s.&nbsp;  &nbsp;  I never understood why women could vehemently oppose this, other than if their name was&nbsp;a business or other such asset.&nbsp; I&nbsp;view reluctance to change one\'s last name as trivial in the first place.&nbsp; The practice has become a systematic tradition for names to be passed down through generations and as a way for couples to be symbolically unified&nbsp;by sharing the same last name.&nbsp;  '),(6675,3802286,'15,July,2004','2004-07-15 00:00:00','   Sorry for the posting with little comments lately, but I think everyone needs to read this.  urlLink Terror in the skies (again)?  I think yes. If you do anything today, read this. You don\'t have to believe it, just read.  '),(6676,3802286,'15,July,2004','2004-07-15 00:00:00','   This is an interesting  urlLink news item  I noticed today.        China   is trying to remedy the imbalance of male to female infants by giving welfare checks (US$144) to families that do not have male babies. They are also trying to prevent abortions and infanticide by placing a ban. I think that the welfare check system will be a better policy, because people will always try to work around the system to get around bans.    I wonder if this monetary incentive will create an imbalance of girls in rural areas since poorer families would benefit the most from the increased income. Will this also affect the prostitution industry in China once there are enough girls to go around? I remember hearing before that countries with a higher male:female ratio also have more prostitution.  '),(6677,3802286,'15,July,2004','2004-07-15 00:00:00','   Todayâ€™s  Wall Street Journal  has a front page article on the   Mercatus   Center  .  In   Washington  , Tiny Think Tank Wields Big Stick on Regulation .  '),(6678,3802286,'15,July,2004','2004-07-15 00:00:00','   First they spend our tax money to reduce the number of smokers, and then they subsidize the industry because of \"low demand.\"  urlLink This deal  is absolutely dripping with disgusting regulations. Here\'s a quote to get you interested; &nbsp; \"Until now lawmakers from the tobacco belt have opposed allowing the Food and Drug Administration to regulate tobacco. But they signed off on the new oversight in exchange for a $12 billion, 10-year program to aid tobacco growers struggling for survival in light of lower prices resulting from less smoking and increased imports. \" &nbsp; You know you have a problem when Kennedy is&nbsp;a chief&nbsp;sponsor of the bill, but also when it is called a \"marriage of convenience.\" The tobacco industry is basically a government program now, so whether you partake in the wonderful substance of tobacco or not, you are in for a wild ride. I expect ranting.  '),(6679,3802286,'15,July,2004','2004-07-15 00:00:00','   Yes, that is Kerry giving the \"Black Power\" salute at the NAACP.&nbsp; You\'re not mistaken.  &nbsp;  Because let me tell you how much time Kerry has spent with \"minorities\" in his lifetime.  &nbsp;  All of the three seconds it took to take that picture.  &nbsp;  -Justin  '),(6680,3802286,'14,July,2004','2004-07-14 00:00:00','   Looks like  urlLink Julian \'s cooking up a  urlLink party . You heard it here first.  '),(6681,3802286,'14,July,2004','2004-07-14 00:00:00','   Since some people feel comfortable posting one line schlopp on our blog that takes two seconds and little thought to produce, I will do that here...   Here is a  urlLink great link to conservative/Republican memorabilia and apparel .    RightWing Stuff has funny, Kerry-bashing stuff, all of which would make great gifts, for birthdays etc.  '),(6682,3802286,'14,July,2004','2004-07-14 00:00:00','   urlLink It\'s number 2 this year ...not bad.  '),(6683,3802286,'14,July,2004','2004-07-14 00:00:00','   Iraq may not have had \"large stockpiles\" of WMDs, but Iraq did have very dangerous weapons in the past, and it did used them on Iran, on the Kurds, and in the Gulf War.  Furthermore, Saddam would have loved to get his greasy hands on some, and Iraq didn\'t \"fully cooperate\" with weapons inspectors before the war.   If you doubt Saddam\'s intentions to obtain WMDs consider this.  The charge that Saddam had not sought weapons grade uranium turned out to be false.   urlLink Saddam did attempt to purchase uranium .  The  urlLink false negative about the uranium purchase  originated from  urlLink an op-ed by former Ambassador Joe Wilson  in the New York Times (no suprise).  Lord Butler\'s report also confirmed Saddam\'s attempt to purchase uranium and vindicated Prime Minister Blair (as did the 9/11 commission of President Bush) of \"lying about pre-war intelligence\".   To the NY Times\' credit though, I will link them for an article that they ran the other day by one of its token conservatives, William Saphire.   urlLink Saphire wrote about what he calls the current \"groupthink\" surrounding the war .  Right now, the steady media drum beat is \"he lied\" or \"he misled\" us, he being president Bush and that media groupthink is just plain false.    P.S.   Michelle and I attended a very compelling booksigning this past weekend about the Connection between Iraq and Al Qaeda.    I would be more than happy to discuss it with any interested parties tonight at the party...THE PARTY FOR BUSH, tonight 7 PM at Oakwood.  Hope to see all Bush supporters and all undecideds there.  '),(6684,3802286,'14,July,2004','2004-07-14 00:00:00','   The NAACP has become a defunct organization since its heyday during the civil rights era.  President  urlLink Bush made the correct decision to not speak before them .  Even if race mongers would inevitably try to portray the decision as some \"snub\" against black people, I applaud his tough decision.  NAACP Leaders, Julian Bond and Kwesi Mfume, have spouted some exceedingly venomous remarks about President Bush and Republicans in general that go far beyond the pale.    They have spoken of a \"right wing conspiracy\" against blacks, referred to the \"Taliban wing of the Republican Party\", said that the party \"flies a NAZI swastika next to a confederate flag\", and accused President Bush of metaphorically dragging a little black girls father (James Byrd) by the back of a pickup \"all over again\" for not supporting so-called \"hate-crime\" legislation.  (All crimes should be punished equitably, and President Bush supported the toughest punishment for the crime, the death penalty.)  Education Secretary  urlLink Rod Paige blasted the NAACP today .  President Bush has done so much for blacks and minorities in general (but he doesn\'t look at it that way, the way a Democrat would).  President Bush concentrates on doing things for  AMERICA , not a particular racial group.    Still, President Bush has the most racially and gender diverse cabinet ever, has spent more on aids relief to Africa, increased funding for education through NCLB and denounced the \"soft bigotry of low expectations\" in education, and the list goes on.  Paige, a black man himself, points out many of these things that the president has done and classifies the current (not the NAACP of the civil rights era) as \"Naked Partisans\", which I think is right on the money.  The Bush administration has said that it will reach out to blacks and minorities in other ways.  I doubt that speaking at the NAACP convention really helps any candidate that much other than symbollically, but unfortunately some slanted media accounts of the decision portrayed the decision more as a \"snub\" than justified rejection of an invitation, and the reasoning behind the decision more as \"the NAACP being critical\" of the Administration than \"the NAACP engaging in exceptionally hostile rhetorical attacks and unfounded slanders\" against the Administration.   urlLink Lynn Cheney sought to woo Hispanics  when she spoke this week at the National Hispanic Leadership Conference this week.  This is one tangible act of many more to come of an administration and a party, which genuinely wants to appeal more to hispanics and other minorites.    I am very confident that Republican policies appeal to minorities as much as Democratic ones.  The real test is influencing people to break from tradition and for the Republican Party to better market itself.  The New Democratic strategy of using race politics, fear, demagoguery, etc. to keep minorities in lockstep Democratic voting is equivalent to the much-maligned \"southern strategy\" of the 1960s.  While I am on race issues,  urlLink here is a little bit of Affirmative Action stuff in Arizona . Basically, the school was handing out raises to minorities and women only, and not to white males.    Lastl speaking of affirmative action, here is an article by Marie Gryphon at Cato, which  urlLink debunks the myths surrounding Affirmative Action . First, that since less minorities are admitted into \"elite schools\" that means the process is biased.  Second, that since as many minorities don\'t get into elite schools, their future earnings potential will be significantly diminished.  Both myths are false.  The Republican Party, the party of Lincoln, has become the party of color blind policies and of equal racial rights for all.  '),(6685,3802286,'14,July,2004','2004-07-14 00:00:00','   John Murphy would be so  urlLink proud !  '),(6686,3802286,'14,July,2004','2004-07-14 00:00:00','   I wrote a response to Diana\'s post on the FMA; it is four pages, single-spaced, in length.  For that reason, I\'ve placed it on  urlLink my blog .  Especially if you\'re opposed to homosexual marriage, or have religious objections to it, I would encourage you to read it; it goes by fairly quickly.  '),(6687,3802286,'13,July,2004','2004-07-13 00:00:00','   Everyone knows the FMA is great for religious fundies. That argument\'s easy: marriage is a traditionally heterosexual institution, and gay unions are a different can of worms altogether and should be treated as such. Okay, so we see how they would win if the FMA was passed. But it is also great for libertarians.  I\'ll preface this by saying that most of my argument here is co-opted from a late night conversation with the ever-intriguing  urlLink Eve , but that I really am convinced this is so.   The FMA does not eliminate the \"right\" of gay people to marry. It does, however:  1) Promote the right of children to be raised in married, healthy homes with both genders as role models, and  2) Give gay couples more liberty by preserving the privacy (and free markethood) of their relationships.  The first point is self explanatory, and to agree with it of course you do need to buy that it is harder, for example, for a boy raised in a lesbian home to grow up with a positive understanding of what it is to be a man (see  urlLink Dawn Eden  on this, T-shirts, and cultural malaise). But the second point is something people are talking about a lot less these days.  If marriage is not about children any longer, and if the only reason government got involved in marriage to begin with was to arbitrate in situations where it is not clear what to do with children, a class of people in society not old enough to make rational, consentual decisions on their own accord, then what place does government have in a marriage institution that is no longer primarily about children? We ought preserve the heterosexual marriage as a government-governed entity, because no matter how much contraception straight people use in these newfangled times, they still pop out babies now and again, and, evil of evils, the government does need to step in to deal with that.   But in a marriage that is solely about the two spouses pledging to love one another, why is it the government\'s role to put their stamp of approval on that union?   Any libertarian who wants the government to rule on the validity of the love between two same sex individuals should really be creeped out by the contradiction in his opinion. Our world reeks of weird nanny-state notions when it is no longer enough for people to know they love someone until a higher authority validates that love.   Think about it: why exactly does the government need to know or certify the difference between two same-sex friends (gay or straight) and a romantically involved gay couple? You might say that I would trust a romantic partner more than I would a best friend. But I might, if I\'m single, trust my best friend with my life more than I would even my parents, and thus I would want to get the same power of attorney packages that let her come see me in the hospital, summon me to a nation to which she has immigrated, and all the other non-child-related benefits of marriage of which the gay community is so shamefully deprived. Heck, throw children in there too -- adoption by gay couples is legitimate and regulated by the law even now, right?  Libertarian gay rights activists should really be focusing on making privatized contract law more concrete, universally recognizable, and readily available to two same-sex people who would legitimately seek to share that package of benefits. Since gay couples are not procreating a previously nonexistant third party (aka babies) into their relationship, they need to stop  begging for the government\'s stamp of recognition on their love and start seeking their own marketable solutions to the technical difficulties of contract law that currently hold them back.  '),(6688,3802286,'12,July,2004','2004-07-12 00:00:00','   urlLink This is something new?  I feel like my degree is nothing more than a receipt for a payment everyday I go to school and am forced to take \"cultural and global and historical diversity\" courses to graduate instead of learning anything that may help my career. You can\'t manufacture diversity people.  And yes, the Onion might not be the most credible source. But this one is seriously not too far-fetched. (As opposed to yesterday\'s story about devious rabbits convincing Bush to pass gun regulations. Although...)  '),(6689,3802286,'12,July,2004','2004-07-12 00:00:00','   urlLink Mike Ditka.   Honestly, no commentary is necessary.  '),(6690,3802286,'12,July,2004','2004-07-12 00:00:00','   As most of you know, I was against the war in Iraq since it was first beginning to be discussed in the fall of 2002. To me though, the one thing about the Iraq situation that is worse than going in in the first place is withdrawing too quickly and for the wrong reasons. That being said, I could not fault the  urlLink Philippines  more for withdrawing their whole whooping force of 51 people out of Iraq because a Phillipino citizen is being held captive by Iraqi guerillas.   I\'m sorry, but you made this mess, now you clean it up. The mistake to go into Iraq was made by the United States and the meager coalition we brought with us (for you Fahrenheit 9/11 watchers, The Republic of Toga...The Republic of Iceland), but the decision to pull out now because it is too hard is wrong. Even if it takes us ten years of street fighting like the kind we have been seeing for the past year, every country that joined the coaltion now has the obligation to see it through. One of the worst mistakes in United States history was withdrawing from Vietnam too early because it was \"politically popular.\" The nearly one million Vietnameese people who were consequently slaughtered as a result of our withdrawal have only us to blame.   And all this Philippino withdrawal (and Spain\'s ealier after the Madrid bombing brought the socialists to power) does is legitimize Iraqi guerillas and throw fuel on the fire. If anything, its an added threat to us now, because with Spain and the Philippines demonstrating that if the going gets tough enough, we\'ll get out, there\'s a much greater incentive for a terrorist attack in this country to sway the elections and our policies.   It may seem callous, but it\'s only one hostage\'s life. If I were a hostage, I would hope that my country wouldn\'t be so cowardly as to let my life dictate my government\'s decisions. It\'s one life at the expense of many many more.   Cold amoral utilitarianism? Yeah, I guess so.   -Justin H.  '),(6691,3802286,'12,July,2004','2004-07-12 00:00:00','   In President Bush\'s latest radio address, he outlined his reasoning for supporting the highly discrimintory Federal Marriage Amendment.  Like most of the things Bush says, he\'s wrong.   urlLink And, thanks to Andrew Sullivan, I\'m not the only one to say so.   An excerpt:   \"Then there is his second premise: that allowing gay people to enter into civil marriage would \'fundamentally redefine\' marriage.  In fact, of course, gay couples want to enter into marriage as it is currently defined.  And the fundamntal redefinition of civil marriage to which the president refers occured decades ago--when contraception became widely available, severing the link between procreation and civil marriage...\"   I\'ve had a lot of debates over this issue recently, and I\'ve always been stonewalled in responding to the charge that marriage is about the protection of children.  Whereas I do not disagree that children are better off with married parents (same- or opposite-sex alike), tying what is best for children to an institution classically used to protect property or secure nations against war (see: the Middle Ages) seems preposterous, if not delusional.  When I write or talk about things like gay marriage to those who do not (and, most likely, will not ever) support it, I do so out of the desire to express, in terms of humanity, something that for me and those like me is best not explained in terms of logic and public policy.  Libertarians argue that man is born with certain rights that cannot be stripped from him; of course, this is an ethical and moral argument, one that cannot be made in terms of logic alone.  You have to believe it to be true for it to remain so.  The same, I think, applies to gays and gay marriage.  '),(6692,3802286,'12,July,2004','2004-07-12 00:00:00','   So, I was wondering: As human beings (rather than libertarians, appreciating the contradistinction between the two), what do you guys think of the modern stem cell debate. Primarily this debate revolves around the use of embryonic stem cells, the acquisition of which requires the destruction of a human embryo. A few thoughts:  1) Abortion: At base, it seems that pro-lifers would have to oppose the use of embryonic stem cells. If a life is a life at conception, then murder for body parts would seem inhuman. Am I wrong?  2) Alternatives: There seem to be alternatives to embryonic stem cell research. Michael Fumento has argued that all the research necessary can be accomplished using adult stem cells. Also, the usefulness of stem cell research in curing say, Alzheimers, seems overrated. Thoughts?  3) Utilitarianism: At what point should societal good play in to the debate.  4) Politics: How much of the modern debate is purely political. John Kerry has said he would leave the \"science to the scientists.\" Bush wants ethical oversight. First, does Kerry actually mean it? Second, is that a prudent policy on such a sensitive issue of biotechnology?  I hope you guys have some input....  '),(6693,3802286,'09,July,2004','2004-07-09 00:00:00','   If you didn\'t see on the Koch web page, we\'re the  urlLink blog  of the week per the Adam Smith Institute blog, run by one Alex Singleton, former Koch fellow.  So go check them out.  Hat tip:  urlLink Our web site  '),(6694,3802286,'09,July,2004','2004-07-09 00:00:00','   I\'m more than a little amused at the accusations floating about regarding John Edwards\' supposed lack of qualifications to be vice president.  Perhaps its because I remember another time when the current president was running for office and similar accusations surfaced.  Interestingly enough, these accusations were not thrown around all that much by Democrats, but by members of Bush\'s own party, who suggested that Bush run for lesser office.  Based upon the botched war on terrorism and the complete lack of foreign intelligence cogency, perhaps the original observations on Bush\'s foreign policy experience had weight.  I would observe, also, that Edwards, who sits on the Intelligence committee, would be more versed in public policy in comparison to a governor, no less the governor of a state that historically empowers the Lt. Governor more than the actual governor.  Just some random thoughts.  '),(6695,3802286,'09,July,2004','2004-07-09 00:00:00','   urlLink John Tabin applied for press credentials as a blogger at the DNC  and actually got it....until the Dems realized that he is a frequent contributor to the American Spectator.  I laughed at this, because Dems can be so dumb, and it reminded me of a stunt that a coservative friend of mine at Penn pulled last year.    Eric Rechtschaffen, a member of the Penn College Republicans\' Board, actually got the College Democrats to endorse his candidacy for a student government position by emailing them and using his middle name as his last name.  The Penn College Dems used to hand out their \"endorsements\" like candy, apparently no vetting whatsoever.  (They would then email their Democratic listserve with the names of the people to vote for).  When the Dems discovered their blunder after the election, they unsuccessfully attempted to go before Penn student government to get Eric disqualified.   This is a case of no standard Democrats.  In Mr. Tabin\'s case, this displayed the left-wing bias of bloggers at the DNC.  (One can\'t call the bloggers legitimate \"media\" or even-handed \"journalists\".)  And in Eric\'s case it displayed sloppy administration by college Democrats.  There was also some sloppy administration apparent Tabin\'s case as well.  Anyone want to throw a Kerry party?  We can make it a dud like his candidacy.  '),(6696,3802286,'09,July,2004','2004-07-09 00:00:00','   If they were any more  urlLink intimate,  we might have our first White House romance since Bill and Monica.  '),(6697,3802286,'08,July,2004','2004-07-08 00:00:00','   Wow,  urlLink this  makes me feel safe. An excerpt:  \"Problems at the jail have been growing for years, but they drew more attention last month when an inmate escaped while rapper Clifford Harris, known as T.I., was allowed to use a maximum-security cell, guards and inmates as props and extras for a music video.\"  Sometimes I think stories are made up for the sole purpose of making me laugh. It better be a pretty good video.  '),(6698,3802286,'08,July,2004','2004-07-08 00:00:00','   There is a great article on  urlLink John Kerry and the politics of science  up at Tech Central Station.  urlLink I wonder why I like it so much ....  '),(6699,3802286,'08,July,2004','2004-07-08 00:00:00','   In a startling  urlLink case  in France, a six-month pregnant French-Vietnamese woman, Mrs. Thi-Nho Vo went to a doctor\'s appointment for a routine check-up of her child.  She did not speak French well enough to communicate with the staff at the hospital and was mistaken for another women who had an appointment for a local surgery.  Mrs. Vo\'s gynecologist inadvertently punctured the amniotic sac and had to therapeutically abort her child. Vo took the man to court, charging him with involuntary manslaughter.  The doctor was acquitted on appeal after the court declared Vo\'s child was not a human being.  As a proponent of pro-life, I am astounded and disgusted at this tragic story. For a number of you, I know you are equally revolted at this blatant disregard for both the child\'s life and the woman\'s choice. But I am curious as to how a group of libertarians would respond.  Even beyond the matter of a child\'s murder, this is a serious violation of the sanctity of life. This cannot even be argued as a matter of the woman\'s choice; her choice was robbed as surely as if she were raped in her gynecologist\'s office.    I want to know your reactions.  '),(6700,3802286,'07,July,2004','2004-07-07 00:00:00','   How did John Edwards managed to become the vice presidential nominee with next to no experience?  The  urlLink Weekly Standard\'s Andrew Ferguson  had an amusing, but probably not unfounded hypothesis...  Here is an excerpt:  \"LIKE A FEW OTHER excessively nice-looking human beings, John Edwards is a victim of reverse lookism. Common experience, lately reinforced by the most rigorous scientific research, demonstrates lookism\'s effects: Ugly people never catch a break, while the well-configured among us always receive favored treatment.   In the past, Edwards has undoubtedly been lookism\'s beneficiary. Having no political connections or experience to speak of, he would never have been elected senator from North Carolina, nor been fingered early on as a presidential prospect by the national press corps, if he had had the face, for example, of Dennis Kucinich, who, as a presidential prospect, has been fingered only by himself.\"  This is so sad, but yet so true.  I have read/heard a lot of crap today about Edwards potentially \"appealing to younger voters\" and give the Kerry ticket a \"fresh look\".  This is code for \"John Edwards is a pretty/peppy good looking person and  urlLink he smiles a lot \", so people who judge others partially or even subconciously based on their looks will like him.  I personally think he looks too much like a pretty boy to be considered handsome, but of course, I\'m biased.  At least President Bush can say that he actually chose his Vice President for a legitimate reasons.  Cheney had  TONS  of experience,  urlLink much, much more so than Edwards  does now.  President  urlLink Bush hinted at Edwards\' inexperience today .    (This article also talks about Bush judicial nominees.  John Hasnas, our speaker the other night, told me that he is concerned by liberal obstructionism of conservative judges and Edwards\'preference for activist judges.  Edwards is like Patrick Leahy in the Senate Judiciary, when I watch them on C-span, I want to strangle them both.)  The Drudgereport incorrectly speculated that Kerry might chose Hillary Clinton for his running mate.  This was shown to be tabloid reporting, when Kerry instead picked another white male.   I am keeping track of the  urlLink lack of diversity in Kerry\'s inner circle , because Democrats/liberals are much more quick to criticize Republicans for lack of diversity.   President Bush\'s cabinet and his top advisors are far more diverse than any president before him and far more diverse than Kerry\'s inner circle.  Think Powell, Hughes, Rice, Paige, Whitman, Venneman, Martinez, Gonzalez, Linda Chao, Norm Minetta, etc.  Bush doesn\'t get the credit that he deserves on this front, because he doesn\'t go around flouting the diversity within his administration as much as a Democrat would.   On another subject, the  urlLink economy is booming , despite  urlLink Kerry\'s attempts to distort the facts about the economy .  The AP reports that the \" urlLink Economy Set for [its] Best Growth in 20 Years \".  This will become more and more of an issue, especially if America\'s role in Iraq continues to subside.  Here is a great  urlLink article by a hard-working, small-business owner  who is one of the much-maligned \"rich\", considered undeserving of tax cuts by Democrats.  Every election year without fail, Democrats bad-mouth \"the rich\".  Prepare for more of the same this year, especially now that \"Two Americas\" man (Edwards) is on the trail and ready to play the politics of class.  Kerry was a signatory on a letter to  urlLink meddle with free trade  with China, so libertarians who view free trade favorably should be wary of Kerry/Edwards.  '),(6701,3802286,'06,July,2004','2004-07-06 00:00:00','   urlLink Here  is a great article on one possible determinant of gender wage gaps.     Bottom Line:  The author more or less seems to infer that the wage gap is due, at least in part, to the fact that women are less confident than men.  For highly competitive, high paying jobs, men are more suited for the \"tournament\" style competition.     Also, women are more competitive when men are not around accoring to the study.  So, as Tyler points out, this would agrue for single sex education and prehaps single sex firms.  '),(6702,3802286,'06,July,2004','2004-07-06 00:00:00','   You can barbecue it, boil it, broil it, bake it, sautee it. Dey\'s uh, shrimp-kabobs, shrimp creole, shrimp gumbo. Pan fried, deep fried, stir-fried. There\'s pineapple shrimp, lemon shrimp, coconut shrimp, pepper shrimp, shrimp soup, shrimp stew, shrimp salad, shrimp and potatoes, shrimp burger, shrimp sandwich. That- that\'s about it.   Bubba would turn in his grave to hear the  urlLink news .  Hat Tip:  urlLink Me  (but really  urlLink Cacciaguida )  '),(6703,3802286,'06,July,2004','2004-07-06 00:00:00','   John Kerry screwed-up (or succeeded depending on how you look at it) by picking John Edwards as his Vice Presidential candidate to challenge President George W. Bush.  I can\'t fathom who the GOP could have realistically preferred to Edwards.  John Edwards showed himself to be an inept campaigner during the Democratic primary.  Edwards\' ability to emerge as the runner-up in the Democratic primary said absolutely nothing of his abilty, but everything of the shortcomings of the incredibly weak field that he was up against, including Wes \"dull, zero political experience\" Clark, Al \"race demagogue\" Sharpton, Carol \"I wanna be the 1st black woman president\" Moseley-Braun, Dennis \"whiny socialist\" Kucinich, Howard \" urlLink YAAAAAAAA \" Deaniac, and Dickie \"union tool\" Gephardt.   urlLink Edwards managed to win only one state , South Carolina, by an (all-things-considered) slim margin.  Edwards was born in South Carolina and knew the area better than any of the other candidates, the voters in the area knew him, and he spent much more time and money than any other candidate in the state.  Yet, Edwards still only won by 15 points, and Al Sharpton still managed to win 10% of the vote despite Edwards\' concentrated effort in the state!    Not only did the Democratic primary demonstrate Edwards\' shoddy campaigning ability, but his battle with the other Democrats also rendered him damaged goods.  Kerry and Edwards even used to openly criticize each other before, in usual flip-flop fashion, Kerry turned around today and named Edwards his runnning mate.  What\'s more, the RNC already has compiled a  urlLink massive stockpile of quotes and ammunition  to use in its campaign against the Kerry/Edwards two-headed monster.  The press loves to look at potential vice presidential candidates in terms of what \"added dimension\" they can bring to the ticket, or in other words, how potential vice presidents compliment the presidential candidate.    In 2000, Cheney was a wise choice for President Bush.  Cheney (former representative, former defense secretary, etc.) added tons of legislative experience and foreign policy knowledge to the Bush ticket.  Cheney is also an effective debator, as evidenced by his skilled debate with candidate Lieberman in 2000, is great at rallying the Republican base, and is  urlLink skilled on the political offensive , allowing Bush to remain somewhat above the fray.    Edwards, on the other hand, brings nothing to the Kerry ticket except a pretty face, and by that, I just mean that he hasn\'t yet needed/elected botox injections.  Edwards is young and inexperienced.  Edwards\' little foreign policy knowledge showed through in the primary as he rountinely answered foreign-policy-related questions in vague, rhetorical terms.  Edwards\' experience in government (not that being a lifelong, opportunist, career politician like John Kerry is a good thing) consists of one measly Senate term.    Furthermore, Edwards does next to nothing to moderate ideologically the Kerry ticket.  Edwards has a lifetime  urlLink American Conservative Union   urlLink rating of 12  on a scale of 0 - 100, which is barely better than  urlLink Kerry\'s dismal score of 5 .  An ineffective legislator, Edwards has never introduced a bill that has passed Congress, and his signature bill, which he fought hardest for, was his so-called \"Patient\'s Bill of Rights.\" It was a give-away to his bosom-buddy, trial-lawyer friends -- as if Edwards needed any more reason to be labeled a slavish litigator.    In his article \" urlLink Southern Identity \", Tom Curry is wrong to imply that John Edwards, joining the race as a Senator from North Carolina, will have any meaningful significance on the race at all. In 2000,  urlLink Bush won North Carolina  over Al Gore by a handsome 13 percentage points.  With such a cushion, even with Kerry\'s Edwards-aided candidacy in North Carolina, Bush is still the presumptive favorite to win the state.  Whereas Iowa Gov. Tom Vilsack could have locked up his swing state, Gephardt could have given Kerry an edge in Missouri, Richardson in New Mexico, and Graham in Florida, Edwards brings much, much less.  Sure, Edwards speaks with more southern twang than GW, but he might as well be from Mars, because he does little to shore-up a single red or swing state for Kerry.  John Edwards is at least as oppositional to free trade as Kerry, and Edwards made his way as a parasitic trial lawyer, feeding off other peoples\' money and extorting companies for exorbitant settlements along the way.  Overall, he is a virtual bonanza for Republicans, thirsty to criticize his liberal policies.  The Kerry Campaign didn\'t need another John. It already had one who married a wealthy ketchup heiress, and one \'john\' is plenty.  '),(6704,3802286,'02,July,2004','2004-07-02 00:00:00','   This is one of the  urlLink best pieces  I have encountered since I started reading blogs in the fall.  I read this a while back and really enjoyed it, so I had to post it.  It is written by a guy who runs  urlLink Selling Waves.com .  Here is a glimpse, but you really should read the whole thing, despite its length; you will definitely be smarter for taking the time.   It is my contention that many people allow the laws of government to become a surrogate morality. Instead of synthesizing their own moral code, people are willing to allow the law to determine what is moral and what is not. Furthermore, the illusions created by democratic government reinforce this confusion of laws and morals, making it virtually impossible for many people to conceive of a moral society existing without the rule of law.  '),(6705,3802286,'02,July,2004','2004-07-02 00:00:00','   The revolution will be no re-run, brothers The revolution will be  urlLink live .  '),(6706,3802286,'02,July,2004','2004-07-02 00:00:00','   urlLink Noise Pollution?  Oh no!!!  The poor little animals that have to endure a loud noise so that a snowmobile can rescue an injured person lost in the wilderness in the middle of winter.   The poor little bunnies and such...  '),(6707,3802286,'02,July,2004','2004-07-02 00:00:00','   Think of it. The  urlLink wedding  of the century. I know I\'d perk up to hear the news.  '),(6708,3802286,'02,July,2004','2004-07-02 00:00:00','   Democratic Presidential Nominee John F. Kerry chose his running mate early this morning. And it looks to be Edwards, his former rival in the Democratic primaries.  Thoughts, comments, lamentations?  '),(6709,3802286,'02,July,2004','2004-07-02 00:00:00','   Check out  urlLink this article  that I found over at  urlLink Newmark\'s Door .   Italy recently began offering 1,000 euros (roughly $1,200) to every woman who has a second child. As a blogger friend of mine named Shonk who blogs at  urlLink Selling Waves  pointed out, this is pretty interesting given the Catholic stance on sex/birth control.  '),(6710,3802286,'01,July,2004','2004-07-01 00:00:00','   I remember last fall, when the anti-Vietnam War protest pictures of John Kerry with \"Hanoi\" Jane (Fonda) came out and the interviews of Kerry accusing his fellow  urlLink Vietnam veterans  of war crimes first aired.  A week later, led by DNC Chairman Terry McAuliffe, the Kerry campaign went back on its word not to make President Bush\'s national guard service an issue. They did so presumably in retaliation for the heat that Kerry was taking for his anti-war activity.    The Kerry campaign started the drum beat, Democrats howled and screamed, and finally under intense media pressure, President Bush agreed to  urlLink release his National Guard records  in an effort to quell the politically-motivated demands for \"transparency\".  The Democratic trasparency/smear campaign was in full force again last week as Democratic lawyers sued Jack Ryan, Illinois Senate candidate, to release his divorce records.  The lawyers got their wish, and there was a subsequent media feeding frenzy.  Despite his denials of the allegations, Ryan  urlLink resigned his senate bid , putting Republican hopes of retaining the seat vacated by Pete Fitzgerald in peril.  Fast forward to this week.  John Kerry, we learn, has some sealed divorce records of his own, but in waffle-house fashion, Kerry  urlLink has resisted opening  them up and holding himself to the same standards to which he held Bush\'s Guard records.    What\'s good for the geese should be good for the gander.  Kerry\'s ex-wife is reported to have been battling serious depression and contemplating suicide around the time of their separation.  Following the Democratic standard, the American public should have the right to know.  What could John Kerry be hiding from the American Public?  Maybe nothing, but this is an example of Democratic hypocrisy (for you Fabio).   How about some equity by the media lynch mob? Bush did it. Ryan did it. Now it\'s John Kerry\'s turn....  '),(6711,3802286,'01,July,2004','2004-07-01 00:00:00','   Will no one else join the  urlLink debate ?  '),(6712,3802286,'01,July,2004','2004-07-01 00:00:00','   Here\'s  urlLink the link  to the story. Here\'s a blurb:   The Canadian Refugee Board has denied asylum to a Mexican homosexual because he is not \"visibly effeminate\" and therefore not vulnerable to persecution.  '),(6713,3802286,'01,July,2004','2004-07-01 00:00:00','   urlLink Eve Tushnet on what a gay couple  can  do.   In response:  Ms. Tushnet raises two very interesting points with regard to being a gay person and the possibilities in a life without the possibility of marriage.  First, Ms. Tushnet argues that some people find themselves to be less homosexual than they once thought, that they go on to marry members of the opposite sex and live happy, normal lives. The idea of relative levels of sexual orientation is nothing new; Alfred Kinsey first reported the phenomenon in the mid 20th century. Ms. Tushnet, however, leaves out the vital logical extension of her argument, which, in my view, makes it moot.  Kinsey\'s report on male (and later, female) sexuality supports Ms. Tushnet\'s claim. Kinsey posited that sexuality existed on a scale, that no person could, in reality, claim to be 100 percent either homosexual or heterosexual. So, whereas some people may, over time, find themselves to be less homosexual, there are a proportional number of heterosexuals, who, over time, may find themselves to be more homosexual. This explains why later in life, after establishing families and lives, come out as being gay or lesbian; the sliding scale of sexuality works both ways.  So, why does this make Ms. Tushnet\'s argument moot? Well, there are two reasons, the first a little more obvious than the second. First, if people discover opposite or same sex attraction later on in life, it makes moot the idea that, somehow, people will necessarily discover themselves to be heterosexual (or, at least, less homosexual) later on in life. People will find the opposite as well; so what are they to do? Ms. Tushnet may argue that they should sublimate their desires; I will, later, argue against this idea.  Moreover, in discussing those who are \"less homosexualâ€ or \"less heterosexual\" than they previously thought, I think that Ms. Tushnet is referring to those who would, on the Kinsey scale, fall into the middle ground of sexual uncertainty. The truth is, according to Kinsey, that most people have a fairly consistent interpretation of their sexual orientation; by relative example, they do not end up starting as Yankee fans, then switch, mid-life, to being Boston Red Sox fans. One need only note the increasing numbers of young people coming out as gay or lesbian at earlier and earlier ages to take notice of the fact that people, by and large, are hardwired, if not at birth then very early on, with a distinct placement on the Kinsey scale.  But, perhaps more importantly, if sexuality is a sliding scale in both directions, then there is no logical reason to preference one direction on the scale with rights and privileges over the other.  The fact is, both sides of the scale have the potential to produce long-lasting, deep emotional commitments that enrich not only the lives of children (natural or adopted), but also the people with whom they share their lives (friends, co-workers, community members, family, etc). The potential is the same; it is only the form that is different.  The oft-heard argument, however, is that the major value of marriage in society is that it produces children who, later in life, produce more children. I do not disagree that the production of children is important and that it should be encouraged. Insofar, however, as it should be encouraged, a society based upon ideas of liberty must also recognize the equal rights of persons who do not wish to have children, and must recognize their desires as equally legitimate. If we recognize the rights of heterosexuals to marry and not produce offspring, why, when an astounding 28% of gay couples are already raising children, do we deny homosexuals the right to marry and not produce (or raise) offspring?  There is no coherent, logical argument that can defend such a double standard.  Ms. Tushnetâ€™s second argument centers on the idea that gay persons, unable to marry, should devote their lives to something greater than themselves, that they should channel their passion into something besides forming a relationship. This is, I think, the most shameful example of arguing for second-class citizenship for gay couples. So, left without the ability to marry and thereby secure multitudes of legal protections, gays should drop sodomy, pick up art supplies and paint a better Sistine Chapel? This argument is a call, literally, for gays to revel in the sublime bliss of ignorance.    But that aside, I have three reasons why sublimation and transmutation is misguided and ultimately doesnâ€™t fully replace what a codified relationship offers.  First, I want to respond to this idea of sublimation versus repression. The difference between the two is, I think, is in word pronunciation and not practice. â€œDeep abiding friendship, great art, and personal sanctityâ€ sounds great at first. But extending Ms. Tushnetâ€™s argument means recognizing that oneâ€™s â€œdeep abiding friendshipâ€ will still leave someone alone at the end of the night; love, for the sublimated person, is forever stunted and incomplete. â€œGreat artâ€ may provide an escape, but it is necessarily vain and selfish, always in pursuit of that quality of humanity only touched upon by the connection and love shared by two individuals. â€œPersonal sanctityâ€ may sound sublime, but it may as well be loneliness; it is relationships that make a person transcend their own existence and, in doing so, gives them the ability to reflect on what, in their lives, actually gives them personal sanctity. In short, the idea of sublimation, to me, does not seem a whole substitute for the truly sublime experience of loving and relating to another person.  Second, and more practically, should someone have to sublimate his or her sexual desire for â€œsomething greater?â€ I raise this question for two reasons. First, it is blatantly discriminatory that gays, and only gays, would be made to purge themselves of sexual desire or, by extension, sexual expression.  But, secondly, in the spirit of liberty, categorically denying gays any choice besides sublimation (through banning access to marriage and legal legitimacy) violates that spirit of individual liberty on which the country was founded.  If there is no practical reason to restrict someoneâ€™s liberty, then we should not allow it.  Third, and finally, I want to relate the idea of sublimating sexual desire to ignoring a big pink elephant in oneâ€™s living room: ignoring it, or channeling oneâ€™s energy into working around it while doing nothing about it, does not make the elephant go away. Sexual desire is part and parcel of what makes us human and it cannot simply be sublimated. Ms. Tushnet argues that historically, people dealt with sexual desire through sublimation. However, evidence abounds that even though no one saw expressions of homosexual desire, throughout history they have existed. The red tie, the carnation, the green pocket square: all of these were ways that those with homosexual desire would signal to others that they, on the sly, wanted to express a sexuality considered abnormal. Expressions of homosexuality run right through history with as much consistency as religious fervor and the fundamental institutions of Western civilization. That desire was not sublimated; it existed under the radar of the rest of society.  That desire is no longer under the radar. It exists without apology or shame and with varying levels of acceptance amongst a large cadre of Western society. In overcoming sublimation, homosexual desire has reached its most sublime: it has extended beyond itself in an attempt to create something greater and more beautiful than desire.  It is trying to create love.  '),(6714,3802286,'01,July,2004','2004-07-01 00:00:00','   urlLink This  should end the debate as to whether or not the Republican Party is paralyzed by the religious right. The Bush/Cheney team is seeking church rosters from \"conservative Protestant denominations\" to encourage church members to vote in November for Bush. I think that\'s a pretty strong indication that the Bush ticket is indebted to conservative Church leadership, especially in the South.  I don\'t know about you, but personally I\'m disgusted. I have no beef with religion, in fact I think it\'s essential. But involving it in politics?  You know, I really liked what Tom Palmer said when he lectured to us a month ago. He basically argued that libertarianism is not a consequence of Christianity or the focus on the individual of Protestantism, but the unique history of the West and the antagonistic relationship between competing sources of authority (Church, state, guilds, etc.). If anything, our beliefs are a fortunate historical accident and not some \"faith-based initiative.\"  -Justin H.  '),(6715,3802286,'01,July,2004','2004-07-01 00:00:00','   Please excuse what seems like a non sequitor, but I recently took a class on the Chinese economy and have continued to follow it in the news....   Someone has to let the Chinese government in on what the West already knowsâ€¦ Their ugly-little-duckling of an economy may have spunk, growth, and ambition, but it sure isnâ€™t yet the graceful swan of an economy that it hopes to be. Thus, I found the Chinese Foreign Ministry spokeswoman to be humorously out of line when she asked the EU this week to reconsider Chinaâ€™s market status.  She later told the China Daily that the EUâ€™s decision was â€œunfair.â€   Spokeswoman Zhang Qiyue alleged Tuesday that Chinaâ€™s status as a market economy is a â€œmatter of factâ€ but why then is her countryâ€™s economy one of the most unpredictable and risky so-called markets competing in the world today?  Iâ€™ll give it to Zhang that many of the most inefficient aspects of Chinese central planning have been eliminated since 1978 (albeit it slowly and reluctantly), but the jury is still out on whether Chinaâ€™s economy will overheat in the coming years.  The remaining central planning elements of the Chinese economy are definite contributors to this imminent threat.  The EU cited four specific reasons for turning down Chinese officialâ€™s pleas.  First of all, China needs to reduce its overall state influence in the market by freeing up prices, etc.  Next, they need to increase corporate governance in order to quell corruption and increase compliance with accounting law.  China also needs to improve the integrity of its bankruptcy and property right laws as well.  Finally, they need to liberate their financial sector by opening up the Chinese capital market.  You donâ€™t need to be an expert on the Chinese economy or its government policies in order to see the difference between a market economy and the Chinese.  In fact, you donâ€™t even need to major in economics.  I learned in my high school economics class the basic components for a market economy.  One of which was a strong legal system to ensure and uphold private property rights. Right now, corruption in the courts and government officials does not guarantee anyone their rights especially to intellectual property which has been making headlines in China recently.  Another key element to the market system is a transparent financial sector which accurately gauges prices and has  interest rates, amongst other things.  I would be curious to hear what Zhang would say about Chinaâ€™s central ownership of the banking system, its repeated bailouts of state-owned enterprises, and alarmingly high non-performing loan rate.  Zhang did allege that the market is predominantly in control by telling the Associated Press that â€œThe private sector makes up â€˜a considerable part of the national economyâ€™â€ and that â€œâ€˜Some statistics show that the prices of 90 percent of goods are set by the market.â€™\'\' I think the operative words there are â€œconsiderableâ€ and â€œsome.â€  It depends on who you ask and how they spin it.  But as far as the West is concerned China may have potential, but it is still the ugly duckling of the world economy and no one wants to claim it.    I would love to see China walk the walk instead of just quacking the quack, if you know what I\'m saying.  I do think that they are making progress, but they\'re not quite there yet.  I don\'t profess to be an expert, so I welcome ya\'ll\'s opinions...Let me know what you think.  '),(6716,3802286,'01,July,2004','2004-07-01 00:00:00','   CNN totally called NBC out this morning by showing a clip of Katie Couric playing badminton yesterday morning while every other news station in America was playing Saddam\'s appearance in court.  Really badminton?  What were they thinking?!  '),(6717,3802286,'01,July,2004','2004-07-01 00:00:00','   Check out  urlLink this article  on trade (a.k.a. offshoring/outscorcing).  '),(6718,3802286,'01,July,2004','2004-07-01 00:00:00','   Last night, John Coleman listened to me give my schtick about the basic difference between liberals and conservatives, loosely defined, on topics of religion in politics.    That difference is simple. One side sees religion as a fundamentally good force in the world, while the other sees it as something that must be shielded against as if it poses a threat.   urlLink \"Pilgrim\'s Progress: John Kerry\'s dubious approach to religion\"  by Steven Waldman (Mike, any relation?) deserves a very careful read.  The article touches on some key points about the constituents of the Republican and Democratic parties.   Excerpt:  \"In general, most Republicans and most Democrats are pretty religious. The stark differences are at the extremes of each party, and, as so often is the case, the big question is whether the extremes will define the party as a whole. Most Republicans aren\'t conservative fundamentalists, although it sometimes seems that way given the proclivities of the leadership. And the Democrats have their own version of that same dilemma, and it\'s affecting the most important arena there isâ€”this year\'s presidential race: Will Kerry\'s Democrats act like the Party of Secularists even if they aren\'t?\"  I would contest that secularists have much greater power over the Democratic party than fundamentalists do over the Republican party. Although people who believe in biblical teachings more often support Republican policies, the support behind those policies is not rooted in religion. Take any official Republican policy and pose the question \"does the party advocate this because some Bible-beater told it to?\" The answer is patently and repeatedly no. The overarching rationality for Republican policies is reason and not religion. Yes this is even true for our favorite abortion-related wedge issues and gay marriage. (This is not to say that people can\'t disagree with that reason on their own terms). In sum, saying the Republican party is controlled by religion is like saying that the Heritage Foundation is a church, Fox News its scripture, and Rush Limbaugh its priest.  On the other hand, if you ask Democrats why they oppose allowing religious nonprofits, which serve a secular purposes such as feeding the homeless in environments which are required to offer a secular alternative, to compete on an level playing field for government grant money, they will give you some murky answer about well, \"I just don\'t think that government money should go to churches.\" In saying so, they pronounce this last word with a certain disdain. Furthermore, opposition to indirect funding of religious, private schools has been the second largest obstacle to voucher programs, next to teacher\'s unions. In sum, Democratic policy is more controlled by the secularist left than Republican policy is by the \"religious right\".  Labeling one side or the other is unproductive, anyway. We should focus instead on policy and not on pigeonholing one side as \"the religious right\" or the other \"the anti-religious left.\"  Back to the article...  Another interesting point raised in this article is that the African American voting block is much more \"religious\" than many people from the so-called \"religious right\".  A great line from the article was, \"if John Kerry is uncomfortable with religion, then he\'s uncomfortable with Americans.\"  John Kerry is the type of man who will use his religion (or lack thereof) any which way he can to get elected, so I don\'t doubt that he will figure out how to say the right things.    While I am on a church-state related issue, check out  urlLink my tribute to Reagan , which was published in Campus Philly. I juxtaposed Reagan\'s state-run funeral with the attack on the Pledge of Allegiance by the ACLU and atheist Michael Newdow. If something frivolous like \"under God\" makes the Pledge unconstitutional, then state-run funerals for presidents with clergy giving eulogies must be too.    In case anyone is wondering, I don\'t consider myself a \"religious\" person. I haven\'t gone to church or said a prayer in a while, but I do believe in God.  '),(6719,3802286,'01,July,2004','2004-07-01 00:00:00','   Happy birthday to John Coleman.  Big number 23!! Last year of the early twenties.  Enjoy it while you can. Next year, you\'re in your mid-twenties and inching ever closer to middle age.  So make it a good one!  '),(6720,3802286,'01,July,2004','2004-07-01 00:00:00','   With all of the faith in religion talk that went on yesterday, I thought I\'d bring up another inspiration point...our faith in the libertarian way.  urlLink Julian Sanchez   brings up a really good point in his latest blog. He asks, do we believe in low taxes and small government because we are libertarians? Or are we libertarians because we believe in all of those things, for what Julian calls \"independent reasons?\" I think the Koch Fellows are a perfect example of this. Often, we reach the same conclusions, but for very different reasons. That only enriches our debate. So bring on the drama Koch-heads...  '),(6721,3802286,'01,July,2004','2004-07-01 00:00:00','   Check  urlLink this  out.  I can\'t believe this is true. I saw a link to this over at Vice Squad. Here\'s a glimpse:   The bill would punish anyone caught wearing low-riding pants with a fine of as much as $500 or as many as six months in jail, or both. â€œIâ€™m sick of seeing it,â€ said Shepherd, a first-term legislator. â€œThe communityâ€™s outraged. And if parents canâ€™t do their job, if parents canâ€™t regulate what their children wear, then there should be a law.   This is real funny, but it is also real scary.  Maybe we should just have a country-wide dress code.  '),(6722,3802286,'01,July,2004','2004-07-01 00:00:00','   Fabio, the masquerade is over.  Come out and fess up to what you are: a liberal.  Stop trying to fake everyone out and claim that you are basically a libertarian, because next to nothing you write exudes libertarian ideals.  As for anyone who actually toiled through Fabio\'s posts, be sure to read my post below on \"abortion\" and \"gay marriage\" and the 2004 presidential election.  Fabio does not argue any specifics but rather pontificates in vague generalities before taking the leap of faith to tell people that they should vote for John \"Whiskey Dick\" Kerry.  My post, on the other hand, expresses the nuances of two specific issues.  Fabio denouncing the \"relgious right\" only demonstrates Fabio\'s anxiety toward anything religious.  The Republican party is not controlled by the \"religious right\" anyway, but instead by the \"rational right\" that rejects the irrationality of the left.  As for the influence of religion on politics, any sect of Christianity splits more evenly Republican and Democratic than does Judaism, but does Fabio translate that constituency into a license to call them the \"religous left\"?  No, he sure does not.  His \"corporate welfare\" charge is unsubstantiated garbage and claims that the Democrats are \"open\" are false.  But since he did little to reinforce his remarks, I will save those topics for another day.  '),(6723,3802286,'01,July,2004','2004-07-01 00:00:00','   While I agree with Fabio\'s \"pragmatic\" approach to libertarianism, I could not disagree more that the Democrats are about laisez-faire social policy. That\'s like saying that Republicans are for laisez-faire capitalism. Maybe in some abstract ideal, that\'s what the two parties stand for (at least that\'s their mantra) but the reality is quite different.   I wouldn\'t call the support for Teacher\'s Unions and the failed state of education in this country a laisez-faire approach to social issues.   I wouldn\'t call increasing welfare and Medicare laisez-faire. To me that translates explicity to Democrats taking a  specific  stance on a social issue. And for that matter, civil rights. What right does the government have to insist that a private indivuidual must hire or admit a token percentage of nth degree ethnicities to paint a picture of diversity? That sounds like taking a stance and making a subjective decision on what to and what not to do regarding social issues. Civil rights is not about laisez-faire.  If the question is about gay rights (which I think is where many libertarians identify with the Democratic Party) I would offer the critique that working within a government that defines a social issue by increasing the number of people involved in that issue (hence allowing gay as well as straights to marry) is not in any way keeping the government out of defining morality. If anything, gay marriage, at least the way the Democratic Party has approached the issue, is explicity defining morality no differently then how Republicnas are in their call for a Marriage Constitutional amendment. It\'s just the reverse conclusion. I believe, and I think any libertarian would agree, that marriage is a contract between two people with NO role for a government.  So I would offer the opposite opinion that voting for a Democrat is better than voting for a Republican. I can\'t make my mind up yet on the election (at the moment I am leaning towards Kerry because of the whole CATO argument of divided government), but I still identify with the Republican Party more at least in principle because I think that it is impossible to ever have social liberty without economic liberty. If the government controls the economy, it will inevitably control mass opinion. And conversely, if the government has no role in the economy, inevitably the freedom and individualism afforded in the market place will translate into a freedom of ideas and morality in social issues.  Therefore, I respectfully dissent.  '),(6724,3802286,'01,July,2004','2004-07-01 00:00:00','   The  urlLink latest ...  '),(6725,3802286,'01,July,2004','2004-07-01 00:00:00','   1. Vote Democratic    I\'ve long said that I am a libertarian in a pragmatic sense.  That is, I generally see things from a libertarian point of view, but recognize that political realities often obstruct full libertarianism from flowering.  It is for this reason that I would not vote for the Libertarian party in an election; it\'s impractical because (though I hate to say it) a libertarian president simply will never happen in the current political reality.  So, in the current political reality, what is a libertarian to do?    I say, vote Democrat.    Ultimately, libertarian fiscal policy is not inline with either party, nor is it in line with the fiscal politics of a majority of the nation.  Though Republicans claim to be the party of fiscal responsibility (and apparently spend less than Democrats) the reality is both parties spend fairly irresponsibly; where their spending differs is, I think, measured solely on where the money goes.  Here\'s a quick translation of spending:  Republicans: $20 million Airline bailout Democrats: However-many-millions on Welfare  Put that in the Spending-Translator and you get:  Republicans: Corporate welfare Democrats: Social Welfare  I will not presume to make a judgment as to which one is more or less important, or more or less harmful.  I will say, however, that, though this is a grand oversimplification, there is a lot of truth to it.  Moreover, though Democrats may spend more empirically, generally their policies effect a greater number of people (because they are, after all, social entitlements, not corporate ones), proportionally, I am willing to posit that the numbers balance out a little more evenly.  Socially, however, the libertarian laissaz-faire attitude is far more similar to the social openness espoused by Democrats than it is to the religiously-based conservatism of the right.  Democrats tend to take things a bit far, fringing rational arguments about social liberty with arguments about entitlements; that is a problem for another day.  But, ultimately,  la societe laissez-faire  exists with more purity and freedom in the social politics of the Democratic party.   2.  Raid Nader   Going off what I\'ve already said about pragmatism, I see no reason for third party voting, particularly for Nader.  I feel that every vote cast is a vote not thrown away, even if the candidate is never elected.  But how I  feel  is about as politically relevant to electoral politics as Anna Nicole Smith is to a drug-free existence.  Indeed, when it comes to practical matters of change, it is, in my opinion, best to side with one of the major political parties.    Often, I find that the problem with the extreme left or extreme right is not that they don\'t have good ideas, it\'s that they don\'t know how to market them; they want too much, too soon.  For people like Nader, I think the best way to effect change in the direction of their politics is to work from within a particular party, to use incrementalism to push for small-change-now, instead of getting no-change-now.   3.  Kill TV   On a completely different topic, here\'s a little something on everyone\'s favorite idiot box, once again arguing that  urlLink TV Sucks .  '),(6726,3802286,'01,July,2004','2004-07-01 00:00:00','   In Shoaf\'s latest blog, he vastly oversimplifies the social issues that are in play during the 2004 presidential election.  This is a common mistake that the general population makes all too often.  There are many more social issues at stake than soley \"abortion\" and \"gay marriage\".  If you are deliberating between Bush and Kerry on social issues alone, you also have to ask if you want to have more nanny-state, environmentalist-activist social regulations, such as the cell phone bans which Coleman pointed-out and toilets that won\'t flush, and you also have to ask whether you want to be subjected to racial preferences, a radical secularist agenda, and \"social justice\" policies out the waazooo.  Kerry is a nightmare on all of these fronts.     Now to digress to the main point of this blog.  The terms \"abortion\" and \"gay marriage\" are ambiguously thrown around much, too much, to do justice to the actual, nuanced debate on more specific issues that exists today.  Where we stand on \"abortion\" and where we stand on \"gay marriage\", would more appropriately be considered where we stand on \"abortion-related policies\" and where we stand on \"gay marriage issues\".  If what you mean by \"abortion\" is overturning Roe, then people need to realize that even with Bush in office, Roe will not realistically be overturned in near future for a number of reasons.  First let\'s have a little historical background.  Justice Blackmun, a Nixon appointee, wrote Roe.  O\'Connor, Souter, and Kennedy all Republican appointees vote to uphold it.  A significant percentage of registered Republicans and Republican officials are pro-choice.  The bottom line is that Republican appointees, and Republicans in general are far more pro-choice on the whole than what\'s written in the Republican platform and just because Bush might get a chance to make some appointments, doesn\'t mean that his appointments will overturn Roe.   If Bush does get the chance, the most oft-speculated as the next to retire is Chief Justice William Rehnquist.  Rhenquist is the most staunch, pro-life presence on the court, and Bush would have difficulty replacing and would be loath to do so for obvious political reasons. Political concerns, along with Generational and demographic ones, not to mention fervent Senate filibustering, would be exceedingly prohibitive to appointing an equivalent justice.   White House Counsel (Bush\'s top legal advisor at present) Alberto Gonzalez, who demonstrated pro-choice inklings while presiding on the Texas Supreme Court, is widely considered to be Bush\'s top choice to fill any court vacancy, and become the first Hispanic Supreme Court Justice.  Ironically, Bush could very well appoint someone who could acutally strengthen Roe!   Next in line to become Chairman of the Senate Judiciary is Republican R.I.N.O, Senator Arlen Spector from Pennsylvania, who is unabashedly pro-life.  The Chairman and the committee have more influence over nominations than Bush, himself.  This means that, if they don\'t approve of a Bush appointment, then the appointment is pretty well screwed as evidenced by Miguel Estrada, Charles Pickering, Janice Brown, and the list goes on.  Democrats and \"pro-choicers\" have been applying an ideological litmus test and there is no way that they would let too many, if any, justices through to the Supreme Court, who would upset the balance of the court on this fragile issue to the breaking point.  People need to realize, ABORTION AINT GOING ANYWHERE!!!   Not Only is Roe not in danger, but the court as a body has repeatedly indicated that it is not looking to readdress Roe, regardless of what individual justices think about the abortion.    Again, the bottom line is that the right to an abortion will not be endangered by Republicans or President Bush any time soon, so pro-choiceers can vote for Bush and everyone else can be free to abort as many unwanted preganancies as they so desire.   Why do you you ask then, does John Shoaf and so many other Americans make the mistake of thinking that the basic right to an abortion is in jeopardy?  The answer is the pro-abortion lobby is much, much, more heavily funded and heavily staffed (as compared to the pro-life lobby), and is more widely accepted (though not by any means mainstream).  The pro-abortion lobby and pro-choicers in the media rountinely use scare tactics to cow pro-choicers into voting for Democrats, even though \"Abortion\" is not really at issue.  Without the pro-abortion lobby, Democrats are without one of their most effective lobbies, and without misinforming the public that abortion is constantly \"under-siege\", the pro-abortion lobby is out of a job.  As for what is really at issue, I support everything Bush has done so far reagarding abortion, which could be considered \"pro-life\" initiatives.  He signed the bill to ban 3rd-trimester, partial-birth abortions, the Unborn Victims of Violent Crimes Act to increase the penalties for those who commit crimes against pregnant mothers, has vowed not to employ a litmus tests (as in a pro-life one) on judicial appointments, has upheld the principle that the government shouldn\'t fund abortions, and has supported a loose form of parental notification for minors.  To the glee of diehard \"pro-lifers\", polls indicate that a majority of Americans also agree with these common sense \"abortion restrictions\", as they are sometimes called.  So in conclusion on the topic of abortion, whether you are \"for\" or \"against\" abortion is not the question you should be asking yourselves.  George Bush is the more logical vote for the vast majority of voters, pro-lifers and common sense pro-choicers alike.  I am for Bush\'s \"abortion-related policies\" fervently over Kerry\'s.  Moving on to Gay Marriage.  I take Jonathon Rauch\'s point of view-that the country\'s stand on gay marriage should not be coerced by an activist court, but rather left-up to the states and more democratic legislatures and referendums.  As I see it, there are two main issues at stake right now 1) Activist Courts 2) The Proposed Constitutional Ammendment.    The constitutional ammendment is not a threat, because it doesn\'t stand a chance of being ratified.  Any serious political analyst would agree.  Rather, the ammendment is merely reactionary, posturing, a direct reaction to the activist judges in Massachusetts.    If the ammendment did pass, which it won\'t, then it wouldn\'t \"ban\" gay marriage either.  It would merely enunciated that nothing in the Constitution can be interpreted to necessarily allow for gay marriage.  Then, it would again be left-up to states to decide the issue, but this time without the threat of coercive legislation from the bench.  Activist judges on the other hand are threatening, and not just on gay marriage, but also on a whole other wide array of issues.  If we elect Kerry, he will appoint activist judges and that would be very bad.  In conclusion, Elect Bush for two reasons, among many others:  1) He is the logical vote for \"rational\" pro-choicers, like myself.    2) He is better on gay marriage, even if you are accepting of it long term like I am, but realize that you don\'t want to have the issue dictated to us by some rogue judges in Massachusettes and realize that a new constitutional ammendment to ban it is pie in the sky.  If you have read this far, I thank you and I hope that it gives you new, more informed way of viewing the issues of abortion and gay marriage as they relate to the presidential election.  Vote George Bush in 2004!!  '),(6727,3802286,'01,July,2004','2004-07-01 00:00:00','   If there has ever been a testament to the propagandist filth that is Michael Moore,  urlLink this is it.   So are we supposed to applaud China for letting its guard down for two seconds and importing an American documentary? I mean that\'s like saying that we should be proud that China is allowing its citizens to read Western books, when all those books include are Das Kapital and the Communist Manifesto.   Not really an improvement.   But still, better than  urlLink Iraq.  '),(6728,3802286,'01,July,2004','2004-07-01 00:00:00','   Earlier in the year I was set on voting for Gary Nolan, who was without question the best libertarian option.  Then the party made a foolish nomination, and I started to rethink things.  Once I calmed down and settled in a somewhat rational state, I started thinking about how foolish it would be to vote libertarian. The only way it makes since to vote libertarian is if one thinks a libertarian candidate is better than all the other candidates (i.e. that the primary candidates are EQUALLY worse than the Libertarian option).  Voting should be an expected value calculation...just like anything else. So, since I do not think Bush and Kerry are equally bad, I should probably vote for one of them because my vote will \"count more\" (this presumes that I have decided to vote and it ignores the futility/irrationality of voting arguments which I find quite persuasive). So, as a libertarian, I approve of Bush\'s tax policy and most of his labor policies. But I prefer Kerry\'s stance on gay marriage and abortion (in part because I do not think it is a government issue and in part because I think gays and women should have these rights...I know that is contradictory). It really comes down to evaluating whether economic policy or social policy is more important (I realize the two are often related and that candidates deviate from what you would expect their economic and social policy to be...but I ignore that for now because I am an economist and I can assume things). If I care more about the economy and about getting a good job, I should vote for Bush. If I care more about a women\'s right to choose and a gay\'s right to marry, I should vote Kerry. I suppose this all means I should vote for Bush because I am not gay and I do not anticipate getting an abortion or causing someone else to. Yet I am still very unsatisfied about the whole thing. Maybe I will write-in someone obscure (perhaps the great  urlLink Chris Sabo )...that might be the most fun of all, especially considering that my vote is not worth much anyway. I have a friend who votes for Hayek, and apparently, Michael Jordan gets about 6000 votes ever year.  '),(6729,3802286,'01,July,2004','2004-07-01 00:00:00','   Not to redirect the flow of traffic (or destroy my libertarian credentials), but here is an  urlLink interesting post on the Virginia ACLU\'s decision to fight the state ban on teen nudist colonies .   Also,  urlLink love me .  '),(6730,3802286,'01,July,2004','2004-07-01 00:00:00','   Finally I can get the respect  urlLink I deserve.     I just hope I made the cut.  '),(6731,3802286,'01,July,2004','2004-07-01 00:00:00','   The  Washington Post  ran an  urlLink interesting article  this morning about a debate between Nader and Dean on whether or not Nader should run. Now, I know most of you donâ€™t care to hear either of these men speak, unless they are overshadowed by your own angry, belligerent commentary, but I think it might be interesting for those people who are wondering about the viability of a third-party candidate. I, for one, am skeptical that the American majoritarian voting system can ever support a successful third party â€“ and this is especially relevant for those of you who are thinking about voting for Michael Badnarik. Is voting for a third party actually going to make a difference (or change someoneâ€™s mind), or are you simply throwing away your vote?  '),(6732,3802286,'03,August,2004','2004-08-03 00:00:00','   More shameless self-promotion reinforced by an ethos-grabbing Alex Potapov citation. Tonight I got incredibly fed up with  urlLink the idea of universal healthcare and blogged about it . Read the whole thing or this excerpt:    As I sat in the lower room of Vidalia restaurant in Washington, D.C. and watched Al Sharpton rant over closed captioning on CNN recently, I realized that something has gone terribly wrong with the Founders vision of America, with our grasp of logic, and with the basic language through which we Americans communicate our ideals. At some point in his speech (ironically enough, sometime before he quipped \"read my lips, \'Our votes are not for sale\'\"), Sharpton made reference to the fact that universal healthcare was part of the American Dream--you know, that hope-filled policy that sent thousands floating to Ellis Island more than a century ago. That dream espoused by the Founders in the Declaration of Independence. You know, the American Dream--universal healthcare.  As I pondered Sharpton\'s turn of phrase, it struck me how nonsensical the phrase universal healthcare is--particularly as it is seen not only as a rhetorical tool, but a right. When asked why he is voting for George W. Bush in November, my friend Alex is fond of quipping, \"Because universal healthcare is not a right.\" And while he is, in some ways, being funny, there is a lot of truth in that simple statement. Universal healthcare is not a right.   Also,  urlLink check out Ben Affleck . Jerk.  '),(6733,3802286,'02,August,2004','2004-08-02 00:00:00','   For a summer intern and small town native, this past Friday, marked my first encounter with the HOV lanes of the DC metropolitan area.   After a full weekâ€™s work, eager to return home to spend time with my girlfriend, I hurried by metro rail to my northern Virginia apartment and readied myself for the drive, two hours north to Pennsylvania.  By the time I showered, packed, and bid my roommates good weekend, I was on the road around 5:30 P.M., hoping to make a long trip short.  Minutes later, I was abruptly stopped behind at least a dozen cars pulled-off to the side of the on-ramp along I-66 west.  Their occupants looked impatient and anxious to get where they were going, possibly to reach family or loved ones as in my case.    Wondering what the hold-up was all about, I rolled down my window and questioned a heavy set man sitting in a red car next to me.  He informed me about the HOV lanes and the general rule for route 66.  Single occupant drivers, I discovered, were not permitted on the highway during certain times of the day.    According to other commuters, tickets are routinely issued in a draconian fashion for non compliance.  Exorbitant fines of up to $5,000, provide law enforcement with a windfall of added revenue-raising opportunity and an incentive to be especially strict.  Commuters assured me that police look to ticket as many single-passenger drivers as possible and do not allow the slightest grace period for early drivers.  I even heard a horror-story rumor about police pulling over a pregnant woman, who had gone into labor and was trying to reach the hospital.  Inconvenienced, but still relatively unperturbed, I lined-up with the rest of the traffic.    Seeing the world through a political/public policy prism as I often do, I noticed a hint of irony in the situation.  This form of government regulation was unintentionally creating a form of the very problem it had set-out to prevent.  By limiting the free-flow of traffic on I-66, a defacto roadblock had created a new traffic logjam, exacerbating the traffic situation.    As if congestion wasnâ€™t enough, my fellow commuters sat waiting with their air conditioners cranked-up and their engines idling.  I quipped to myself that this could likely raise the ire of many of the same people most prone to over-regulate highway usage.  Because idled engines meant wasted fuel, that meant, gaspâ€¦greater fuel emissions, than if people had been allowed to proceed freely.  Somewhere environmentalists were probably losing sleep over all this.    I continued to do my best to avoid the growing mental irritation from the unexpected delay.  Nevertheless, as an active person, I live by the maxim- time is money.  But this is not always quite so easy, especially, when oneâ€™s intent is getting somewhere to see a loved one.    In this case though, it wasnâ€™t solely a factor of the opportunity cost of my wasted time.  Wasted gas from our air conditioners and idled car engines cost money too, meaning we, the idled commuters, would pay for the unfortunate layover at the pump.  Thatâ€™s ok though, I continued to reassure myself, Iâ€™m only going to be waiting here for about 25 minutes, so it shouldnâ€™t cost me all that muchâ€¦or should it?  Gas prices vary dramatically across the nation.  More â€œliberalâ€, typically urban centers and neighboring, more-densely-populated, wealthier, suburbs can experience astronomical costs of living relative to the rest of the country.    Prices at the pump are no different in that they are often noticeably higher in these â€œbluerâ€ areas as a result of higher gas taxes, inflation, environmentalist obstruction to new oil refineries, and more stringent, some might say overbearing emissions standards, which require â€œcleanerâ€ burning blends and create gas shortages.    Much to my chagrin, Washington D.C. is a prime example of this phenomenon.  I was reminded of the drastic price disparity, twenty-five cents per gallon to be exact, between my hometown in Carlisle, Pennsylvania (30 minutes west of Harrisburg) and Washington, D.C.  Too stubborn to absorb the additional gas charge and a tiny bit disgusted by it all, I frugally shut-off my engine, resigning myself to bake in the beltway sun.    	Unlike Amtrak, to the extent that the Washington Metro is able to remain financially solvent without raising taxes or fares for the third consecutive year (which is improbable) and even though the Washington Metro is subsidized in large part by federal monies raised on the backs of millions of hardworking Americans across the country who will never benefit from MetroRail, I felt glad to have it available for my usual daily commute.    	Americans outside Washington DC (and New York, Boston, and similar urban areas for that matter) are more likely to own their own cars and more free to drive themselves wherever they want, when they want, maybe even by themselves, if they want.  Beltway-types have ceded a degree of individual freedom to drive to work in the morning and in general, to transport themselves as they please.  For some commuters, the Metro system must not only be viewed as a convenient option, but also as the dominant option, which indirectly crowds out other options, such as driving a car to work.    Furthermore, people outside these places donâ€™t have to contend with crippling parking taxes (18 % in New York), massive expanses of no parking areas and parking meters, other parking fees, higher taxes on automobile purchases, higher car insurance, greater-regulation and tax-induced higher gas prices, severe traffic congestion, and other obstacles that make owning and utilizing oneâ€™s own car and individual mobility much more problematic and costly in these more liberal areas.  This HOV example is one illustration of some of the unintended consequences that government intervention in the publicâ€™s transportation systems, and the free flow of goods and capital, can have; although other examples abound.  In certain areas, speed limits are unnecessarily restrictive.  In New Jersey, drivers are not permitted to pump their own gas, because it as seen as a â€œsafety hazardâ€ (donâ€™t ask me to try to explain this one).    New Yorkers and Democrats in Congress (and most recently the District of Columbia) have sought to outlaw cell phone use, while driving.  Never mind that the risk of cell phone use while driving is negligible, that other driving related activities are more risky, that reckless driving is already illegal, precluding the need for an additional statute.  Never mind that such a law would be unenforceable, hurting only the most law-abiding among us, or would become another facile, revenue-raising opportunity for hyper-vigilant law enforcement.    New research by the National Highway Traffic Safety Administration even suggests that  urlLink requiring hands free sets in lieu of regular cell phones is as dangerous, if not more so to drivers .  This is because the study shows that the danger (negligible though it may be) lies not in the act of holding the phone, but rather in the distraction from talking on it, effectively negating the difference between the two modicums.  The study goes on to posit that the hands free movement might have actually put more people at risk by inadvertently encouraging more talking while driving.     The overall point of this article is that, the stealthy hand of a concentrated government creeps a little further over Americaâ€™s automotive liberty each and every year.  So next time you are driving, hopefully somewhere far away from a â€œfreewayâ€ in Washington, take a moment and revel in just how free youâ€™ve got it.  '),(6734,3802286,'02,August,2004','2004-08-02 00:00:00','   Think Republicans are less blameless for the onset of socialism? Walter Williams, one of George Mason University\'s premier libertarians, doesn\'t think so.   urlLink Check out his article here.  '),(6735,3802286,'02,August,2004','2004-08-02 00:00:00','   Even a Badnarik voter like me was distressed at the whitewashing of the post-Democratic convention coverage.  As many junkies know (and as many news sources reported up to the convention), presidential candidates often get a \"bounce\" in the polls after their official nomination, usually around 10 percent or so.  Every poll conducted after the Democratic convention has Kerry with a bounce anywhere between zero and 4 points -- a disaster, it seems, historically.  After all, even Michael Dukakis had a convention bounce.  But that\'s not what I read in the papers yesterday and today.  The articles almost uniformly crow about Kerry\'s minor bounce, with almost no mention that it is a woeful underperformance, historically.  USA Today, which actually ran an article reporting \"no bounce for Kerry,\" (their poll had Kerry down a point after the convention) blamed the \"intensely partisan\" race rather than Kerry and his platform, or his awful convention speech.  Imagine if President Bush lost ground after the Republican convention; what would the headlines say then?  I rarely point a finger at the \"liberal media,\" but in this case bad news for John Kerry is being reported differently than if it was bad news for George Bush.  '),(6736,3802286,'02,August,2004','2004-08-02 00:00:00','   Moving to a more urgent issue for discussion...Did anyone go to see \"The Village,\" by M. Shyamalan, this past weekend?  I went with John Coleman and Sara on Friday and the best part of the movie was hearing the two of them huddle together like school children in a tornado bunker, whimpering, \"Oh my god....\" and \"Ohhhh Sh#t.....,\" every two minutes, in reaction to the \"suspense\" on-screen.  (To be fair to both of them, we did suffer this huge-ass ceiling fan blowing cold air onto us, giving us goose bumps before the movie even started.  However, my knowledge of some of John\'s living habits prevented me from seeking comfort in his bosom.)  Anyway, what should we make of M. Shyamalan and his movies?  Is this guy really a genius for manipulating audiences with simple and straightforward plot-twists that are carefully hidden until his chosen moment of revelation?  I won\'t spoil the secret behind \"The Village,\" but it\'s really not that clever and leaves you feeling like you\'ve been jipped out of a good-ol\' TERROR-FRIGHT, which is presumably what we\'ve all come for.  Beyond that, his movies are slow (meaning boring and without the sophistication/realism that sometimes compensate in other flicks), the adults talk and behave like straight-up weirdoes, the children are imbued with uncanny powers of wisdom and act like geriatrics, and everything else is unlike anything I recognize from real-life or other forms of art.  In short, there\'s no terror, no laughs, no realistic-profundity, and nothing else that I would deem \"entertaining\".   The only thing to take away from these movies is an appreciation for how M. Shyama plays everyone like a fool (without entertaining!) and how dumb we are to pay money to be on the receiving end.   --Tim Dorsett  '),(6737,3802286,'02,August,2004','2004-08-02 00:00:00','   In continuing with my rant about Metro from last week...  If you didn\'t see this ridiculous Post  urlLink article  from last week, check it out. A lot of debate has been spurred on the Post\'s Opinion pages because of it.   It just deepens my belief that Metro needs to privatize part if not all of its operations.  I mean, when nice restaurants require ties and jackets, some may complain that it is stuffy, but everyone recognizes that it is the ownerâ€™s prerogative to set such standards and they must comply in order to be served. But when the Washington Metro Area Transit Authority asks riders to please refrain from eating or drinking, people like Stephanie Willet see no harm in finishing their candy bar before swiping their card to ride the train (see original  urlLink article ). They certainly do not expect to be taken to jail for finishing their mouthful and even disposing of the trash in the proper receptacle!  Whatâ€™s the difference in the two scenarios? It is the tragedy of the commons. Public goods are taken for granted and thus not treated with the same level of respect that private property usually commands. In a private establishment, you would expect the manager to either ask Ms. Willet nicely to please refrain from bringing food down the escalator in the future or maybe even denied her use of the train that day. However, when the state owns the metro station, they have to create their own police authority, the Metro Transit Police, to force customers like Ms. Willet to comply with rules and respect the public property.  As WMATAâ€™s budget problems and customer dissatisfaction increases and instances like Ms. Willetâ€™s become more frequent (see the Postâ€™s July 9 archived article, Metro Expects More Deficits), consideration should be given to the privatization of the Metro system. The unprecedented negotiation of a private contract with Bechtel Corp. and Washington Group International is definitely a step in the right direction, but the WMATA was hesitant to contract even for this job. Maintenance and management services could also be contracted out in order to improve metro services. Privatization would not only save money but also increase customer service standards by giving both customers and employees the incentives to behave respectfully.  '),(6738,3802286,'02,August,2004','2004-08-02 00:00:00','   \" urlLink Democrats hail working stiffs, but party on corporations\' tab \" is a great article, which points out how that the lavish parties attended by delegates at the DNC Convention are in stark contradiction to the Dems\' supercharged, populist rhetoric. To some extent, the Koch Fellows are \"partying\" extravagantly and enjoying a great experience in DC thanks to Republicanesque money from the Koch Foundation.  Just today, I was informed by the Chair of the Young Conservatives of Pennsylvania that a Democratic legislator in PA is trying to enact legislation to ban smoking in bars (similar to what was done in other liberal strongholds such as NYC and how Dems support regressive \"sin\" taxes on cigarettes) for over-inflated \"health/safety concerns\".  A similar thing happened this past year at the University of Pennsylvania. Its more liberal/Democratic student government, banned smoking in dorms on campus.  Conservatives, more so than liberals, support smokers\' rights. Business interests (largely Republican) want to be able to cater to customer preferences and the tobacco industry (which supports Republicans), doesn\'t want to be unfairly targeted at the expense of basic individual freedom and personal liberty, which is really what this issue is all about.  From a state-run liquor control board in PA, which is dominated by Democratic labor unions and which causes the prices of liquor to be much higher and less available than in neighboring states, to dance clubs that have had to be shut down because of reckless lawsuits from more likely-to-be Democratic trial lawyers, to fireworks (think fourth of July) which are much more regulated, if not banned, in liberal northeastern states, than they are southern ones, such as Alabama and South Carolina, to liberal attacks against university Greek life as being discriminatory, misogynist, and non-inclusive; on more than just these issues, the conservative view is the more party-friendly position.  This is somewhat ironic, because liberals are generally stereotyped as being bigger partiers/more fun-loving than \"up-tight\" conservatives. Though liberals may be bigger partiers on the whole, I would seriously contest on a whole range of issues that conservative policies are more party friendly.  Younger, party-concerned voters (not just the political ones) should realize that the Republican Party is the best party in town.  urlLink  '),(6739,3802286,'02,August,2004','2004-08-02 00:00:00','   urlLink Interesting article  in the Christian Science Monitor about the gender gap experienced by many American colleges and universities.    \"It\'s wrong to put your thumb on the scale for a male applicant when you\'ve got a better-qualified female, but I know the colleges are doing it because they\'ve told me,\" says Thomas Mortenson, a policy analyst at the Center for the Study of Opportunity and Higher Education in Washington.  A Midwestern college president who asked not to be named agreed, saying that some colleges he knows well admit men at higher rates even though \"women are usually the stronger candidates.\" ....  \"We could possibly enroll an entire class of women - our pool of women applicants is that strong,\" he says. \"We could not enroll a whole class of men, because we don\'t have enough strong male candidates.\"    Admissions offices are struggling for balance, and many resort to raising acceptance rates for male candidates in order to reduce the gender gap. I\'m interested to see how this affects affirmative action supporters\' arguments that AA is necessary for diversity\'s sake. Clearly, few supporters of AA will rush to ensure that more men are placed in top colleges. Maybe this will reveal the diversity question as the distraction it is, and force the debate to focus on the real question: whether AA is a good means to address past or present systemic discrimination.  By the way, this is a very serious problem at  urlLink my school .  '),(6740,3888166,'29,July,2004','2004-07-29 00:00:00','   in case u were wondering maggie just bit me in the head it hurt. according to the little kidz at my camp i am a mongoose eating squirrel.  '),(6741,3888166,'28,July,2004','2004-07-28 00:00:00','   Hi! guess wut? i\'m a cit (councler in training) at a little kids camp! sounds fun? it sux i hav never had so many fuc*ing hairs pulled out of my fu*king head in one f*cking day (i\'m sorry about the language im just a little pissed off right now, could ya tell?=) andyway i\'m actally happy cuz Mariah\'s coming home today and i haven\'t seen her in forever so yay!!! if ur reading this mariah -hi! welcome bak- anyway like the new photos i added?  p.s. if anyone needs help putting fotos on their blog post a comment and beleive me its really easy but its really hard to figure out ( i spent like 2 hours just trying to figure it  out) anyhoo i hav to go but i luv u all and i\'ll talk to you later! p.s. maggie, i loved ur bible stories(j-dawg\'s the best)  '),(6742,3888166,'24,July,2004','2004-07-24 00:00:00','   urlLink        urlLink sexymaggie     Originally uploaded by  urlLink Bordness .     this is maggie (isn\'t she sexy?) she\'s posing with her ping pong paddle I\'m sorry if you didn\'t want me to post this maggie if u want me to take it off write a comment  '),(6743,3888166,'24,July,2004','2004-07-24 00:00:00','   urlLink        urlLink maxine     Originally uploaded by  urlLink Bordness .     This is a pic of me pondering about who knows wut  '),(6744,3888166,'24,July,2004','2004-07-24 00:00:00','   hi it\'s maxine(duh) i was just chillin\' (with the villans) at home and decided to go on my blog. Maggie if you\'re reading this i just wanted to tell you that.... cocaine\'s(however u spell it) a hell of a drug. yay! i said it, anyway tell more people to read my site cuz its cool and soon i\'m going to post a lot more funny pics on it(and they WILL work this time!!) I\'m bored and if you are too you can call me or e-mail me at pinkyflamingoe@yahoo.com (yes it\'s a weird e-mail but there weren\'t any good ones that weren\'t already taken) if u do we can have a deep and interesting conversation or we can just talk about random crap(that\'s always fun) peace out syonora ttfn ttyl adios aloha bye buh-bye good-bye oh nevermind  '),(6745,3888166,'18,July,2004','2004-07-18 00:00:00','   urlLink        urlLink maxine2tif     Originally uploaded by  urlLink Bordness .  '),(6746,3888166,'13,July,2004','2004-07-13 00:00:00','   urlLink     			 				 			 urlLink catie2 			 			Originally uploaded by 			 urlLink Bordness .  '),(6747,3888166,'13,July,2004','2004-07-13 00:00:00','   A horse is a horse, of course, of course,  And no one can talk to a horse of course  That is, of course, unless the horse is the famous Mister Ed.    Go right to the source and ask the horse  Heâ€™ll give you the answer that youâ€™ll endorse.  Heâ€™s always on a steady course.  Talk to Mister Ed.    People yakkity yak a streak and waste your time of day  But Mr. Ed will never speak unless he has something to say    A horse is a horse, of course, of course,  And this oneâ€™ll talk â€˜til his voice is hoarse.  You never heard of a talking horse?    Well listen to this: \"I\'m Mister Ed.\"  '),(6748,3888166,'13,July,2004','2004-07-13 00:00:00','   Why isnt anyone interested in my pie club?! I am so hurt! You idiots dont know what you\'re missing. I am going to go cry in a corner now and theres nothing you can do about it!!! (except join my pie club, hint hint)  '),(6749,3888166,'13,July,2004','2004-07-13 00:00:00','   Hi if you read my first post you probably realized that i have a pie club if you\'re interested in joining this club or attending one of these events read on..... The Pie Club was started by Earny Green in 1989. Since then he has sadly moved on to other things and gave the ownership to me. The Pie club meets every Sunday from 4-6 p.m. at wherehouse 50 on Peir 50 (it\'s near the baseball feild) We have many discussions and do other interesting things related to pie. If you would like to join it costs $7 and $3 for every month you are part of it. We also have special pie club events which are open to the public. These events are $5 for non-members and free for members. Here are some of the events and their dates. July 18th Pie History Day( a presentation will be made on the history of pie) July 25th-Pie Eating Contest August 20th-Pie themed birthday party for our special friend maggie August 27th-Learn how to make 5 different kinds of pies Note anyone planning to join or attend any of these events must contact me before hand -Maxine  '),(6750,3888166,'11,July,2004','2004-07-11 00:00:00','   Wow this blog thing is kinda cool, u can write stuff and people can read it. Sally why dont you allow comments on yours it makes me want to......(wait i\'m thinking).......to........ Eat a bannana, Ha!!!! beat that non - comment allowing person!!!! whats that you say? thats not evil? well for your information i hate bannanas!!!! thats why I got my friend fred the monkey to sit on them. and now the world is happy except for me becuz I cant write comments on sally\'s posts! poooooooooooooopppppppppppp!!!!!!  '),(6751,3888166,'10,July,2004','2004-07-10 00:00:00','   beep!  '),(6752,3888166,'09,July,2004','2004-07-09 00:00:00','   See, I ate food today and it was quite good. sooooooo...... this is fun........I like pressing the button that putts dots on the screen...... I think I spelled my title wrong oh well, isn\'t this interesting? I went to Europe for those of u hoo didn\'t know!!!!!!Hey, its Ariana here to spuce things up because Maxine doesnt know what to say!!!!!!Well okay back to Maxine.... I hope that little commentary inspired you to spruce up boreing stuff.(oops spelled it wrong, oh well) If you like pies u can join my pie club its quite fun, we eat pies not that that has anything to do with pies or anything. okay ariana has somthing to say-Okay Maxine I say SCREW YOUR PIE CLUB!!!!!!!!!!but that wasnt very nice.....what is a pie club anyway. Before we can say that it is bad we need to understand it more deeply. Nuclear wars are started by the word \"hate\" so I suggest you never say that you HATE Maxine\'s pie club OR ELSE (there will be a nuclear war)!!!!- back to maxine- I hope you enjoyed arianas \"speach\" about maxine\'s non-existent pie club  '),(6753,3549714,'21,June,2004','2004-06-21 00:00:00','    She woke up suddenly in the midst of the nite due to an unknown fear ...there was darkness all around. it was dark ,dark enough for her to hope for any ray of light...her eyes wide open, struggling hard to see but in vain...the darkness was deep, piercing into her soul making her feel like crying and calling out loud for help but it was in vain seemed like all the screams and cries somehow lost their jingle as she opened her mouth...her vocals failed ,but somehow tears made there way silently, moistening her face and adding to her vulnerability...the gloomy and mysterious winds gave her a sickening feeling ..she was soaked with perspiration...fear was gradually creeping into her soul making her feel entangled into a web of darkness , lightly but tightly woven around her...... The locks of darkness were as tightly woven as that of her life, where she was breathing and living with a feeling of being caught up helplessly and off guard, wanted to come out of it..get over it, but things got worst at even a slight move,it was as if the strands tightened around her, hurting her badly, making her cry ,but the depth of darkness was deep enough to absorb her vails and cries...beam of light was nowhere...darkness had taken over the glow and shine of hope and lightening from her life...she felt miserable........ but suddenly a realization struck..\"Therz always light behind every dark nite\"....and so definitely somehow , someday the shadows of the dark will be taken over by all the brightness and joy this life has to offer her...some day the strands with b let loose to give her the freedom to breath and radiate the smiles of happiness and contentment.This realization led her to the new dimensions of life ,now even in the blackness of mid nite she could c the far and distant sun rising above to eat up the shades of the dark sky ,announcing a better tomorrow and bringing a smile of satisfaction to her lips...she closed her eyes and just in an instant was lost into the valley of sweet dreams.   '),(6754,3549714,'15,June,2004','2004-06-15 00:00:00','    Just today only a realization struck me.........somehow I have lost track of life. Every second is passing by and i have just stopped worrying ab8 the preciousness of every instance that i am losing.The sand of time is just slipping out of my fingers giving me the insight of being a looser!!! I am, what i didnt expect out of life..what i didnt want to be..but now since i am what i am then i have to give the best to what i am and make sure to c how good can i be. I need more commitment towards life, more commitment towards what i do, more seriousness in my work, in thoughtsâ€¦â€¦running away from life,from responsibilities, from realization  â€¦is not like me!!! Once certain things need to be done they are to be done with all the best that I have to offer and I need to enjoy every bit of itâ€¦.how can I just keep on convincing myself that whatever is happening is not what I want and still passing my precious days doing the sameâ€¦.either I must quit or I must end up doing betterâ€¦enjoy ur work,ur life and b the best..coz this is the key to success and success will have to come to u as ur rewardâ€¦so better get going!!!   '),(6755,3549714,'07,June,2004','2004-06-07 00:00:00','    She looked in the mirror and the image it reflected really took her by surprise...somehow it reflected a beautiful girl with sparkling eyes reflecting all the dreams and passions of tomorrow,shadows of a beautiful past....hopes of future........ahhhh!!! life was wonderful...just looking at the reflection made her shiver with excitement... She was an ordinary and plain looking girl but a very strange feeling tranformed her into a princess........ luv is beautiful  now she wanted to sing ,songs of happiness..songs of passion....even the dull colored flowers in the pot nearby, looked more bright..just like her smile..a smile that suddenly could brighten someone\'s day,give way to light into the darkness...even she didnt know what actually made her feel more real.....what was it that gave her the realization of the existence of a beautiful soul within her....it was  HIS LOVE  that shone thru her eyes...reflected thru her smile....and danced around her making her feel like walking on water and touching the stars.. Some say luv is bliss........a blessing that is granted to those who are special..she was a special girl with a sensitive heart that could beat for the poor ,that could cry for those who were in pain...she deserved to love and be loved!! Just his thoughts brought contentment to her heart and smiles to her lips..she was in  LOVE!! He never looked into her eyes to say that they were deep enough for him to feel like getting drowned....nor he ever parised her looks ,he never said that he loved her..or wanted her badly enough..but some feelings are strange enough just like some strange people who some how enter into our lives unknowingly..just the thoughts of them make us beam and feel wanted.She was struck by the same romantic fantasy...he was no where but every where she looked...he never was there, but the feel of his presence in her life made her feel alive..love is strange.... rite!! she never saw him,never met him ,but still every night she slept with his memories shut down behind her eyes....spent the whole nite dreaming about him......HIM!!...........who was virtually living in the computer she owned...who was a virtual existence that was always alive in her chat window when ever she logged on...who was always all ears to listen to her...to feel her,to make her life as beautiful as ever...talk about the miracles of internet....it makes u alive when u r at the verge of loosing urself as an entity..makes u feel beautiful when u start feeling that u have lost the whole beauty of ur soul and ur mind...this is the place where ur mind lives and let u b what u want to b ..let u fall in lov with whom to want to fall for...life is beautiful here.Thanks to all the virtual souls!!!!   '),(6756,3549714,'06,June,2004','2004-06-06 00:00:00','    I was on my way to work today..where i saw a young pregnant woman running after cars asking for money..she was begging!! That was the instant in my life where i realized how cruel and inhuman this society has become...That woman is at the verge of a new stage..ready to welcome a new life...future of this society!!!a life that deserves all the better things like love and respect...and c what the woman has to offer to that new ray of hope......she is all set to bring that life into the world and teach him/her how to beg...how to lie...how to rob...how to kill one\'s prestige and cry for help..how to run away from respectable work and look for easy money..how to ask for money and from whom...these r the lessons that the child is gonna get from the mother who has an eye for the rich...has a nack of begging expertly... I am not blaming that woman...its not her fault..who knows what led her to this stage..was it poverty? was it desperation?was it illness? or was it her hubby running after her with a big stick threatening to kill her if she wont come back home with enough money for him to buy his Cigarettes...what was it that led her do it..will remain unknown to me atleast who just crossed that womens path unknowingly just by chance..but still just on looking at her many questions arose in my mind...... Every day we talk ab8 moving towards development...towards betterment..spend our lives reading bks...talking highly ab8 the days to come, a bright future..but the question is are we ready yet to face TOMORROW?.....are we prepared enough to offer our next generation a better life?....next generation.... The child of a beggar will learn to beg..will never know what a SCHOOL is, will never know what self respect is,will never know how to repect others...will only know how to beg coz that is the lesson his mother is giving him even from before his birth... i wonder, where lov has gone........lov for poor.....lov for humans..lov for children...lov for our nation... If this is the way we keep on treating with the generations to come..we are being crual to ourselves...as a nation..and as humans. God knows where lov has gone.....but it needs to be brought back for a better tomorrow..for our generations to come..so that they may have there share of love and respect which they may spread to the whole world in return and represent pakistan.   '),(6757,3549714,'05,June,2004','2004-06-05 00:00:00','    Once known as \"The city of lights\".........karachi, has now completely transformed into a city that has lost hope ,it has plunged into darkness..........darkness of despair.....hopelessness...pain!!! The day started as a normal bright, sunny day...no one had a slight idea that its going to end up as deady as possible..... It was time for ISHA prayers,all the mosque\'s cried a voice of calling from God those who could hear it went straight ahead to say there prayers...the prayers started with the name of God!!............and then there was a loud BAAAAAAAAAAAAAAAAAANG!!!a loud noise that shook the building of the MOSQUE...all the lights suddenly went into darkness.......darkness that was frightening......that was killing......has something happened??.......was it a tremor??an earthquake??.........did the earth shake?........is the sky preparing to explode!! has the sky fallen down??....what was it????????....... IT WAS A BOMB BLAST!!! ............where?? in the MOSQUE!!....uhhhhh....how many people died?.....1 ...........2.........no its 12!!......ohhhh it\'s 15.......eventually its 20.......and hundreds hurt ..wounded badly........   the BOMB exploded in the mosque just when the prayer started!!!!    Death took over the city that was known for its brightness and life!!.......death of innocent people..cries of there loved ones....whispers of anger and hatred and protest are all over the place now.....insecurity is all over.......prayers are left unanswered or what...??? The darkness was deadly........there was blood all over the place...those who had gone to pray were killed for reasons unknown....or may b for being The faithful!!! Death took over the city..........questions that rose in the minds of people was....WHY??why were these innocent people killed.......why did they deserve to have a death like this......why were they chosen to die in that manner.......but who has the answers?? KARACHI, once had peace and happiness reflecting thru the lights sparkling out of the eyes of its people......and now those sparkles have been replaced by fear...fear of death.....fear of being burnt alive....when will this massacre end?.....when will the common man be safe enough in the city to dream of a healthy and safe life...........When will a common man look forward to a safe and secure day for  himself and his children.........who has the answers?? when death takes over...LIFE RESIGNS!!! life has always been playing hide and seek with death, sometimes its life that wins and other r the times when death takes over but now is the time when  LIFE HAS RESIGNED ...........it has been loosing enough to continue with the battle......when today in the mosque a small child lost her father,  life lost a battle  .......when an 18 year old in the mosque was desperately trying to breath but failed .. it was life again that failed ......life lost when a young energetic Doctor was burnt alive in the mosque leaving behind a beautiful wife having all the hopes of a happy life ahead ....it wasnt the Dr that died it was life that lost again......and now its time! life has resigned..and death has taken over the city KARACHI......once known to b the city of lights!!!........ when death reigns....... fear mounts and insecurity is the order of the day........ Lets hope that this city, dear enough to a lot of people doesn\'t end up being the city of deaths!!!   '),(6758,3549714,'02,July,2004','2004-07-02 00:00:00','    Sometimes it feels as if we spend half of our lives standing in a queue, â€œwaitingâ€â€¦its hard to wait for anything may it b something or someone!! Standing at the end of a long queue with a waiting status is killingâ€¦ We all are somehow or the other in a waiting phaseâ€¦waiting for hard times to endâ€¦waiting for another day to start, waiting for the sun to set and give way to a peaceful night. Waiting for all the struggles of life to end, waiting for a better tomorrowâ€¦!!! Like the whole world out there I am standing in the same queue since ages, the strange part is that the queue keeps on growing delaying my turn to attain my share of happiness and this time the queue is lined up in front of Gods distribution cell. Theirs a fear that keeps on haunting me, what if when after standing in this wait queue for so long when my turn comes up and I reach the window of Godâ€™s distribution cell of â€œhappinessâ€â€¦â€¦.what if I find a label there â€œ store out of stock â€“ CLOSED!!!â€    Plz God!!  Donâ€™t do this to me!!!   '),(6759,3806796,'30,June,2004','2004-06-30 00:00:00','   i\'m going to be sending out my cd on monday. you guys better like this cd because i\'ve had to spend about 30 bucks just to get cd\'s and cases, print the sleeves and buy the envolopes and stamps to mail this stuff out. crazy.  anyway, what\'s asher\'s address? i have everyone else\'s but his so if he\'s going to get a cd, i need his address.  '),(6760,3806796,'30,June,2004','2004-06-30 00:00:00','   I\'m here in Nashville with Dave *(this is Asher writing).  We agree that, regardless of burner accessibility, coordination, and/or speech impairment, tardiness on CDs is unacceptable.   If you got flat feet, you can\'t be in the army.  (Or a wooden eye, says Dave......) ?  Also, Dave has chronic body odor.  This is a great quote i just heard...  \"It\'s nothing personal.  It\'s just music.\" -D.M. Woodall, 2003, in conversation (referring to his own brother\'s dilinquency)  '),(6761,3806796,'30,June,2004','2004-06-30 00:00:00','   I\'m voting that Mike Woodall be put on Afterburners probation.  I know this sounds rediculous, and it has never been done before but if I end up taking another turn(I was the first person to go) before he ever puts out a mix then I vote he be put \"on hold\".  That means he will not be recieving anymore cd\'s untill he puts out.  So let\'s say that if your late a third month then you get probation, that\'s more than fair.  He\'s missed one month and is most likely NOT going to make it for Dec. 1st(Matt, that\'s you) so he\'s got untill New Years to produce or he\'s cut off. Now, of course, I\'ll need a majority vote on this issue...so....Yay or Nay? Sorry I\'ve gotta come down hard like this but that last email was my breaking point.  '),(6762,3806796,'30,June,2004','2004-06-30 00:00:00','   can\'t wait to hear it ash.  anyone mind getting me an ipod for a wedding present?  '),(6763,3806796,'30,June,2004','2004-06-30 00:00:00','   When a family member is put on probation, only a family member can choose if the delinquent family member is to remain on probabtion or be \"subsituted\" by another equal or more compatable family member.  (Since it is only fair that it will be the active family member who has to impart the ruling of dissmissal to the deliquent party.) 000000000000000000000section C/09er  My subsitution will be the admission of the first female in the club: Esther Bumbier (to go  9/1/04) 699 Ponce De Leon Apt. 319 Atlanta, GA 30308  '),(6764,3806796,'30,June,2004','2004-06-30 00:00:00','   yeah, scott. i\'d say you nailed it with #4. in my opinion, that was one of the disc\'s strengths - it offered some work from a few bands/artists that I\'ve been meaning to check out but haven\'t been motivated enough to (i.e. - jurassic 5, ryan adams). so thanks to this disc, i\'ve been more appropriately introduced to said acts. i think the J5 song is probably my favorite on the disc, which is unusual b/c i don\'t normally choose hip hop over rock, but in this case i do. it was better than anything else on the mix, at least of the stuff that was new to me. death cab and postal service still reign. i only had one real complaint about the mix. the whole time listening to it i was looking forward to hearing a coldplay song i\'d never heard and then it turned out to be \"help is around the corner\", a tune which i have been quite familiar with for awhile now. still a great song, i had just got my hopes up for something different and then they were kind  of popped and flew around the room in loops and esses, leaving nothing but a wake of farting noises behind before dropping to the floor motionless and completely deflated. ok, that was a bit overdramatized. i\'m just joe king. i am still having a hard time deciding if the mix is more nashville or chapel hill, though...i think the latter. speaking of the nash, i\'m heading home for the holidays this week (tues-sun) to visit the old man and the rest of my kin.  i\'ll miss you all very...very much while i\'m gone.  so, anyway, when i get back, i should be expecting no.7 already from matt bubar. am i right or am i right? yeah, i thought so. anyway, good job, asher. hey asher...one more thing. i was looking at your acrylic portraits, which are very good, i must say. but i came upon one that looked very familiar by the title, \"Jessica\". it looks just like a girl that went to my high school. it wouldn\'t happen to be a girl named Jessica Gallup, would it? just curious. she dated my best friend for a few months senior year. anyway, that would be weird if it was. it would be weird if it wasn\'t, too, cos it looks just like her. i know, guys. i\'m abusing the afterburners team blog. veering off topic. sorry.  '),(6765,3806796,'30,June,2004','2004-06-30 00:00:00','   1 - you immediately listen to the first track twice. 2 - you recognize a few songs that you love, and look forward to re-discovering 3 - you recognize bands that you like, but songs that you\'ve never heard of 4 - you recognize bands that you\'ve \"been meaning to listen to\" 5 - you don\'t recognize all of the bands on the list, all the while thinking that you were the authority on most independent music  most of you know that asher and i got off to a rocky start with him offering up ways to evangelize me (if you didn\'t know that now you do) - ha.  anyway, i was unsure how well we would recover from that situation, but now i realize that we have so much in common.    nice work asher.  are you going to DW\'s wedding?  if so i hope to chill with you and knock a few back and possibly have some good talks.  nice work on the ABCD!  '),(6766,3806796,'30,June,2004','2004-06-30 00:00:00','   I just got Asher\'s cd in the mail today.  Nice Job Asher!  Although I do have 1 question:  track 17 you have listed as Richard Ashcroft, I\'ve seen both The Verve and Richard(solo) live and that track sure does sound like The Verve.  It\'s really no big deal either way but I was just wondering if you knew for sure if it\'s just Richard.  I must say it\'s a good mix.  '),(6767,3806796,'30,June,2004','2004-06-30 00:00:00','   Hey Scott, I(Ben) like The Gloaming!  Not MJ. And I would have to agree with MJ on the single issue.  There\'s no way I\'d plunk down $8 for a song I already have, a remix of another song I already have and maybe one new song.  It\'s insulting to the fans.  I still have quite a few single from when The Verve (if they\'d only get back together!) was making music and they are quality singles, with great b-sides.  You just dont see that much anymore.  Maybe that has something to do with Napster and Kazaa, because the last time I bought a single might have been before I discovered Napster.  Or maybe bands just stopped making quality singles because they knew fans would just download the stuff for free since it\'s only like 1 or 2 new songs?  '),(6768,3806796,'30,June,2004','2004-06-30 00:00:00','   hmmmmm ... you guys are pretty hard on GTS.  i did not think it was all that bad of a choice, although i would agree that 2+2 should be a released single eventually.  mass appeal, etc ... i kind of like the song.  but something that MJ said:  he could not get away from the gloaming.  i feel the same way about backdrifters.  when thom comes into the song with the \"ahh ahh ahh aaaaaaaaa\"  it makes me want to scream and dance every time.  '),(6769,3806796,'30,June,2004','2004-06-30 00:00:00','   You may actually be the last one to see things. I never saw it on MTV - MTV doesn\'t show music videos except during TRL, and I don\'t stay up late enough to see videos that aren\'t hip hop on MTV2 - but I do remember seeing this video somewhere at least a month or two ago. I think I found it posted on the internet somewhere...probably on Heavy.com. So, anyway, don\'t get excited about seeing this video get any attention. It\'s probably already missed its chance.  One other thing; I\'d also like to add that I\'ve never liked the single anyway (in general).  I think a band should only release a single if it\'s a non-album track that didn\'t fit on the album or something. The idea of me going and buying a disc for almost the same price as the album I bought, just for a song that I\'ve already got on a full album and maybe two b-sides (one being a live recording and one being a demo recording of another song on the album) is ridiculous to me. I used to justify it when bands like Oasis and the Verve were actually putting quality b-sides on them that I could listen to. In fact, Oasis\' \"Whatever\" single is a perfect example of what I\'m talking about. That\'s what singles should be like. New song and three new b-sides (some call this an EP, but in the UK I guess it\'s still called a single). Everyone loved it. If bands would do that, I might actually buy singles. Even if one or two of the songs turn out to be mediocre, at least it\'s new and fresh to my ears. So to hell with singles!  I actually read somewhere that Thom Yorke said he\'s unsure if Radiohead will do the album thing anymore. He\'d like to release only EPs from here on out, because putting together an album has put way too much pressure on them and become quite a daunting thing. Hmmm...  '),(6770,3806796,'30,June,2004','2004-06-30 00:00:00','   Yes I would have to agree with the last couple of posts about Radiohead.  However, Am I the only person that thinks The Gloaming is an amazing song?  I\'m not saying it should be a single or anything but damn!  I love the way the lyrics are done, repeating itself like that, and sonicly speaking it just lovely.  When that base-line comes in...it\'s like {in Neo\'s voice} Woooh.  It\'s hard for me to say it\'s my favorite song on the album but it\'s the only song that I just can\'t get away from.  Love the video although it kinda reminds me of a modern day version of that video from the 80\'s for \"Money For Nothing\".  Remember that?  \"Money for nothin and yur chicks for free\"  '),(6771,3806796,'30,June,2004','2004-06-30 00:00:00','   I\'d have to agree with Matt that \"2+2=5\" and \"Where I End and You Begin\" would\'ve made better picks for a single, if only because they\'re more accessible songs that could appeal to more people. That\'s why I agree with that statement. And if I had to choose between the two, it would be  W.I.E.A.Y.B. . But I also feel that it\'s good when a band that has such great attention from the media releases something that doesn\'t fit into the mold that MTV and popular radio use to filter everything through. Radiohead have the ability to stretch the boundaries a little, and they\'re utilizing it - maybe paving the way for more bands with styles and sounds that don\'t have such broad appeal to see some success, too. Might not see it immediately, but...who knows. The video\'s pretty cool, technically speaking. The idea doesn\'t interest me all that much, but I thought it was executed really well. It\'s really amazing to see things like that done. It makes me feel pretty bad about myself, though, since I\'m soooo not even close to being able to do stuff like that. Thanks again, Radiohead.  '),(6772,3806796,'30,June,2004','2004-06-30 00:00:00','   i thought that the first single, \"there there\" was a good choice. as for \"go to sleep\", i would have picked almost any other song on httt other than that. but if i were to chose the next single, it would have either been \"2+2=5\" or \"where i end and you begin\".   on another topic, i have the go to sleep single and the import go to sleep single and i thought the b-sides on that were way better than the b-sides for there there.   oh and i thought the video was kick ass :)  '),(6773,3806796,'30,June,2004','2004-06-30 00:00:00','   i have no MTV, no cable, and do not listen to the radio.  the person that turned me on to this suspected that i already knew about it, but i was amazed!  is this all over MTV and the radio yet?  does radiohead have another single in mainstream rotation?    go  urlLink here  and click on the radiohead story.  you might not see the almost hidden link at the side that says \"to see the full video click here\" you definately want to click that.  (don\'t read the story until you watch the video).  and please:  wait for work for the high speed access if need be.  let\'s get some posts going here at the afterburners blog.  topics:   radiohead mainstream singles:  good or bad? this choice for a single:  good or bad? the video:  kicks ass, or i suck ass so bad that i don\'t like it :-)  '),(6774,3806796,'30,June,2004','2004-06-30 00:00:00','   Thanks Dave! I would have done it myself if I was there.  '),(6775,3806796,'30,June,2004','2004-06-30 00:00:00','   yeah yeah yeah, well, you will all be happy to know that i caught up with dave woodall last night and he gave a punch in the arm, adding that he\'s sure you\'ll all probably thank him for that.  '),(6776,3806796,'30,June,2004','2004-06-30 00:00:00','   Hey Jonze, I was just bust\'n yur balls on that last email.  We\'re still boys.....but you started it!  '),(6777,3806796,'30,June,2004','2004-06-30 00:00:00','   i motion that there should be no more afterburners mass emails.  all posting should be done on this site!  fun fun.  '),(6778,3806796,'30,June,2004','2004-06-30 00:00:00','   This blog is officially open to the entire Afterburners crew....this should get interesting! -Ben  '),(6779,3806796,'30,June,2004','2004-06-30 00:00:00','   since mike posted his, i thought i\'d share mine since i already made it a few days ago and posted it on my lj. so here\'s  urlLink my top 20 this year  along with some other stuff.  ps. mike, some of the pictures on yours didn\'t show up for me. i could only see albums 19-25.  '),(6780,3806796,'30,June,2004','2004-06-30 00:00:00','   What\'s up, ma niggaz. Check out my  urlLink BEST ALBUMS OF 2003 LIST ...J.T., Britney, 50, y\'know, all the good stuff from the 2-0-0-TrÃ©. Yo. Word.  '),(6781,3806796,'30,June,2004','2004-06-30 00:00:00','   \"music.activisim\"         urlLink  downhill battle  .  is a website devoted to \"music activism\".  among other cool things they have a pretty good  urlLink interview  with robb nansel of  urlLink saddle creek  records.  '),(6782,3806796,'30,June,2004','2004-06-30 00:00:00','   urlLink This  is a little lesson on what a crappy present is and what to do if it happens to you. thanks Matt.  '),(6783,3806796,'30,June,2004','2004-06-30 00:00:00','   the whole album is pretty near decent my friend.  '),(6784,3806796,'30,June,2004','2004-06-30 00:00:00','   Found it!  It\'s called \"Are you gonna be my girl\" and it by Jet.  '),(6785,3806796,'30,June,2004','2004-06-30 00:00:00','   Cool, Jet it is.  Might have to throw a little of that on the ABCD#8. thanks, Scott  '),(6786,3806796,'30,June,2004','2004-06-30 00:00:00','   if it\'s the \"stroke-ish\" sounding song, then it\'s \" urlLink jet \" (that link will take you to the website with the flash intro, and if my memory serves me correctly, then it\'s the same one from the commercial). -sg  '),(6787,3806796,'30,June,2004','2004-06-30 00:00:00','   Does anyone know the name of the song that\'s on the one i-pod commercial with the yellow background?  It\'s not the Sean Paul song or the techno song if that helps.  It\'s more of a rock/punk song. thanks.  '),(6788,3806796,'30,June,2004','2004-06-30 00:00:00','   So you all know, I saw Mike Woodall this past weekend and he said he\'s not ready to go for Jan. 1st so I\'ll be jumping him in line and he\'ll move to Feb. 1st.  So keep in mind, this will activate Mike\'s probation starting on new years.  If you still want to send him a cd that\'s up to you.  Once he sends a mix out he\'ll be off probation and free to recieve abcd\'s from everyone.  On a side note:  I had the opportunity to see Mike in green tights as he pretended to be Peter Pan.  It was quite a sight and it is sure to leave me  urlLink scarred  for life.  If I can find a picture it will no doubt be posted on this blog. peace  '),(6789,3806796,'30,June,2004','2004-06-30 00:00:00','   Shameless self-promotion{shakes his head in  urlLink disgust }.  '),(6790,3806796,'30,June,2004','2004-06-30 00:00:00','   dudes, ABD here. if you go to google and type in \"churchxclothes\" both my blog and jones comes up. how would you all like to say a little somethin\' in the internet world about it and your site will come up. win win for both of us.  '),(6791,3806796,'30,June,2004','2004-06-30 00:00:00','   yeah, just wait \'til you hear \"Our Nation Of Thoughtlessness\".  DYK:  Two fifths of ONOT are Bubar!  '),(6792,3806796,'30,June,2004','2004-06-30 00:00:00','   looking forward to it. for i can say i know not one of those artists.  '),(6793,3806796,'30,June,2004','2004-06-30 00:00:00','   Number 7 is now out and the track listing is to the right.  I just got my copy of it yesterday and it\'s rock\'n me oh so nice.  Good job Matt.  '),(6794,3806796,'30,June,2004','2004-06-30 00:00:00','   urlLink Sigur RÃ³s news .  '),(6795,3806796,'30,June,2004','2004-06-30 00:00:00','   Dave...I seem to remember somebody saying that they only wanted 1 CD a month, so they could fully absorb it. Am I remembering wrong??? Ben, on the other hand, shared the same feelings as I had about that issue. That\'s why he gets a bonus.  Too bad, dude. :(  Hey, I\'m just kidding. But the amount of music I go through, I couldn\'t possibly afford to mail CDs out to everyone everytime I found something good. So let\'s all share. Perhaps you could try hitting up Ben and maybe he\'d be kind enough to send you a copy.  '),(6796,3806796,'30,June,2004','2004-06-30 00:00:00','   mike, can\'t you at least hook all us air fans up with the goods? esther\'s info is on the way.  '),(6797,3806796,'30,June,2004','2004-06-30 00:00:00','   Mike:  Thank you so much my friend, I am looking fwd to it with much anticipation! Dave:  Welcome to marriagdome(is that even a word?), I\'m sure Scott will have no problem switching with you, he keeps telling me about all these great tunes he has planned for when it\'s his turn next.  So what\'s the deal with Esther?  Is she really joining or were you just thinking about asking her?  '),(6798,3806796,'30,June,2004','2004-06-30 00:00:00','   Hey, Ben. I got a copy of the new AiR CD all packaged up for you with postage and everything. So it\'ll be in the mail today. Enjoy.  '),(6799,3806796,'30,June,2004','2004-06-30 00:00:00','   Marriage has thrown me behind a bit. i don\'t know if i can pull of a 2/1 date with still trying to get established. scott can you and i switch.  you all know how it is....late night sex and all. well except for matt and asher... he he he matt and ben, i have a 5 disc changer w/ both your discs in right now, still getting the initial paruze. long live LIAM! you guys got my new address?  2515 camino del mar #3 del mar, CA 92014  yea, it\'s touch living 100 yards from the beach. i\'m constantly awoken at night by those crashing waves on the shore. still have to get used to them.  '),(6800,3806796,'30,June,2004','2004-06-30 00:00:00','   yeah i\'ve heard of phantom planet. they\'re alright. i\'ve only heard one song though which is california. i personally wouldn\'t buy their cd.  as for the free mp3\'s, i would rather win free soda\'s so i can rot my teeth out more and because i can download those same mp3\'s on kazaa. and if your worried about using kazaa, try  urlLink mute . nobody knows what your downloading or sharing and there\'s no annoying ads. very simple and basic.  '),(6801,3806796,'30,June,2004','2004-06-30 00:00:00','   back to the music...anybody else heard of phantom planet? i\'m thinking of picking up their newest disc. i just read this review that this character,  urlLink \"Ursus the White\" , wrote about it. sounds like it might be worth ten bucks.  '),(6802,3806796,'30,June,2004','2004-06-30 00:00:00','   Oh I\'d have to agree 100% with you Mike.  I\'ll occationally drink a Coke every now and then but other than that soda is not really a part of my diet and isn\'t worth buying for the chance of getting a free song.  But, you know I\'ll be scamming bottle caps from the people that do drink it....ignorant fools, drink up!  '),(6803,3806796,'30,June,2004','2004-06-30 00:00:00','   Yeah, drink lots of Pepsi...so that you can get ONE free mp3 download from the iTunes music store. Like, whoop-de-freaking-doo. I\'m sick of these companies thinking we\'re all idiots and that we can\'t see right through these pitiful attempts at product promotion. As if I\'m going to start buying cases of Pepsi so that I can win \"100 Million Songs\". There\'s a small chance I could win just one. Anyway, I just quit drinking soda. If I drink even a sip, I have to pay Matt Pamer $20.00. So I don\'t think there\'s a single free mp3 worth that, at least for me. Now, I must go. I must return to downloading my songs for free...later.  '),(6804,3806796,'30,June,2004','2004-06-30 00:00:00','   Oh great, now I\'ll be drinking soda.  Why cant it be on water bottles or Chocolate Silk!!!  Those bastards!  '),(6805,3806796,'30,June,2004','2004-06-30 00:00:00','   well that was easy:  urlLink http://biz.yahoo.com/prnews/031016/sfth062_1.html  '),(6806,3806796,'30,June,2004','2004-06-30 00:00:00','   I saw this on Apple\'s web site today but theirs no link on the image and no info on it.  Does anyone know what this is all about.  You can see for yourself at http://www.apple.com/itunes/, it\'s at the bottum of the page on the left.  '),(6807,3806796,'30,June,2004','2004-06-30 00:00:00','   It\'s not a rumor, actually. It\'s true. They developed their own language that they call \"hopelandic\" - some sort of fusion between English and Icelandic. I\'m sure they didn\'t come up with an ENTIRE language. My guess is that they just took their songs, looked at the lyrical content of the songs in both Icelandic and English and kind of combined them and manipulated them however they wanted to get them to sound \"just right\" alongside their wonderful, wonderful music...and then, yes, probably sang the same lines multiple times on ( ). Anyway, whatever it is, it works really well and DOES end up sounding quite beautiful. Everything they\'ve done astounds me. And to have made such gorgeous music at such a young age...it\'s pretty amazing. I hope you have all heard 2000\'s \"Agaetis Byrjun\", too.  '),(6808,3806796,'30,June,2004','2004-06-30 00:00:00','   Yeah I noticed that too about ().  It sounds like they came up with a chorus line and then sang it over and over again, changing the music obviously.  But I also heard that they developed their own language/code that everything is sung in, it\'s english and it\'s not icelandic(isn\'t that were they\'re from?).  Has anyone else heard that rumor? That (album) is so good though, it doesn\'t matter what language it\'s in!  '),(6809,3806796,'30,June,2004','2004-06-30 00:00:00','   i thought abcd 8 was pretty good. although i\'ve heard a few of the songs already, the rest was new to me. andy zipf was in circadian rhythm, right? i also thought there was a good mixture of songs in it.  also, if anyone is interested. i stumbled upon this band today called  urlLink godspeed you! black emperor . these guys are like sigur-ros a little bit except some of they\'re songs are twice as long. but i think it\'s great. they have a lot of mp3 samples on their website. i\'m not exactly sure, but i think it\'s all instrumental.   and while i\'m talking about sigur-ros, is it just me or on the album - (), it sounds like he says the same thing over and over again on every song. either way it doesn\'t matter, i still like this album. just curious if anyone else thought that too.  '),(6810,3806796,'30,June,2004','2004-06-30 00:00:00','   Just wondering if you all have gotten a chance to listen to ABCD#8.  What do you all think?  Likes? Dislikes?  In hind-sight, I have come to realize that there is quite a few radio-songs and/or MTV songs on the cd.  If that really is the case I had no idea.  I really do not listen to the radio and haven\'t seen a minute of MTV in the last year or so.  So I apolagize for any excess mainstream stuff, I didn\'t mean for that.  Hope you are enjoying it.  '),(6811,3806796,'30,June,2004','2004-06-30 00:00:00','   In addition to the Best of 2003 List compiled by Pitchfork Media that I already directed you towards, here\'s some other best-of-2k3 lists you might be interested in checking out:   urlLink Almost Cool\'s Top 20 Albums of 2003   urlLink Opuszine\'s Favorite Albums of 2003   urlLink PopMatters\' Best Music of 2003   urlLink Tiny Mix Tapes\' Top Twenty Five Albums of 2003   urlLink Delusions Of Adequacy\'s Top 10 Lists of 2003   urlLink Onion A.V. Club\'s Best Albums of 2003 List   urlLink Metacritic\'s Highest Rated CDs of 2003 (and a spattering of other critics\' Top 10s)   urlLink Silent Uproar\'s Staff Top 10\'s (under the \"Featured\" section on right side of page)  '),(6812,3806796,'30,June,2004','2004-06-30 00:00:00','   I think his vocals sounds ALOT like The Violent Femmes\' vocals.  Maybe that\'s why I dont mind it so much.  Maybe, secretly, I\'m pretending it\'s The Violent Femmes\' new album...that rocks! But I totally agree with you about OUT HUD.  '),(6813,3806796,'30,June,2004','2004-06-30 00:00:00','   I can\'t figure it out. Everything around me is telling me to like the Rapture - that they\'re just, like, wicked awesome and can make you get up and shake yer arse around once or twice. And I do dig their music. It totally makes you wanna do that. Except I can\'t get past Luke Jenner\'s vocals, so I don\'t think I\'ll ever like them that much. Maybe I just have to keep listening and get used to it so it doesn\'t bother me. Does anybody else think his vocals are just, like, way way obnoxious? It\'s just caterwauling, in my opinion. If I could just get past it, then I might like the Rapture a lot. But for the time being, it just ain\'t happenin\' for me. Sorry. And #1 on  urlLink Pitchfork\'s Best Albums of 2003 ??? I don\'t know about that one...     But I do know about Out Hud. Now that\'s music that will make you get out yo seat and shake it. And someone in the band thought it best to not ruin the songs by injecting wretched howling at the moon. Seriously, everybody needs Out Hud\'s \"S.T.R.E.E.T. D.A.D.\"  '),(6814,3806796,'30,June,2004','2004-06-30 00:00:00','   i\'m just now listening to the rapture - echoes and i just got the new strokes today from a friend.  '),(6815,3806796,'30,June,2004','2004-06-30 00:00:00','   yep, 2003 was pretty cool.  for example:  i\'m just now listening to the shins: chutes to narrow.  gosh!  where was i all year, it\'s such a cool album.  i love the way the album (the beginning of the first song) starts.  '),(6816,3806796,'30,June,2004','2004-06-30 00:00:00','   Is it me or was 2003 an amazing year for music?  It just seems like 03 had so much to offer in the way of new, quality music.  I\'m still playing catch-up.  '),(6817,3806796,'30,June,2004','2004-06-30 00:00:00','   Since Mike Woodall is no longer with us, that means everyone will be moved up a month.  So, Dave, you\'ll be going next(Feb. 1st)...and so on and so forth.  '),(6818,3806796,'30,June,2004','2004-06-30 00:00:00','   The last time we had a relative of Dave\'s in Afterburners we never got a mix from them EVER, so before we all agree to have Esther join our little cult let\'s make sure she has a computer and a cd burner, or at least 24/7 access to one.  That\'s all.  Other than that, I\'m all for her joining!  '),(6819,3806796,'30,June,2004','2004-06-30 00:00:00','   i would love to have a female representation in AB!  I love that idea!  I\'m all for it.   PS:  who in their right mind does not have the police album:  synchronicity  in the top 50?  '),(6820,3806796,'30,June,2004','2004-06-30 00:00:00','   I\'m down with that.  Esther seems like she knows what\'s up.  Good call Dave.  '),(6821,3806796,'30,June,2004','2004-06-30 00:00:00','   dudes, since mike woodall is obviously not a dedicated AB, i need to substitue him for another family member. one that knows music and can add to the infamous well of music wisdom. any copies mike wants, i\'ll give them to him. sorry mike, you understand. my vote is my cousin  esther bumbier.  she should go 9/1  '),(6822,3806796,'30,June,2004','2004-06-30 00:00:00','   I\'ll agree to disagree with you about Sting.  Except for his new album, I dont like it all that much and the single is kinda gay but the album before that (Brand New Day)...that\'s a great album, i dont care what you say.  But whatever, your too cool for Sting.  '),(6823,3806796,'30,June,2004','2004-06-30 00:00:00','   Man, I think I struck Ben in his soft spot. Anyway, about Sting...the Police were good. I would agree with you there. I have, in fact, owned a Police record (no longer have it, though). So, so much for that hasty judgement. Anyway, they did have some really great songs that I enjoy[ed]. But let me clarify what I was trying to say. Sting sucks  now , as a solo musician. There\'s no doubt about it. Did you not hear his latest single? It\'s horrible! What is he trying to do??? It sort of makes me wonder about all that \"talent\" he has. Hearing this crap he spews out now, I gotta believe maybe Sting owes a lot of credit to the other guys in the Police. I don\'t know. Don\'t get me wrong, I\'m not saying he has  no  talent. It\'s just been steadily depleted over the years...a lot - a  whole  lot! It\'s that, or else he\'s totally surrendered it to those around him in the studio and production room. And where\'s the integrity in that?  I could not disagree with you more about Sting having \"more talent in one finger than 1/2 the people in [my] iPod.\"  '),(6824,3806796,'30,June,2004','2004-06-30 00:00:00','   What!  dude, I\'m standing my ground on Sting!   He definitely does NOT suck!  Yeah maybe this last album isn\'t as good as his others but the guy is a world class musician and song writer.  How can you deny that?  Let me guess, you dont own a single album of his, do you?  I\'ll even count the Police, anyhting?  Ok, then you can\'t say shit about Sting.  Yeah so Sting might not be the \"cool\" or \"hip\" music act that he once was but he\'s still got more talent in one finger than half the people in your i-pod! C\'mon, somebody get my back on this!  '),(6825,3806796,'30,June,2004','2004-06-30 00:00:00','   Yeah, yeah, yeah. Sting, Bryan Adams, Seal, Shmeal - whatever...they all suck. Hey, I\'m glad you had a good time, though. Really, I am. In fact, I\'m jealous [chortle, choke]. Well, maybe, I\'ll get lucky and being able to at least see one of them live...on Oprah - Aw, snap!  '),(6826,3806796,'30,June,2004','2004-06-30 00:00:00','   That might have been funny if I had just seen Sting in concert but since I saw Bryan Adams it\'s not!  Actually I\'ve seen Sting in concert(a few years back) and it was great!  In fact, I\'d go see him again!  Bryan Adams, on the other hand, was a New Year\'s Eve show that I was given free tickets to...so why wouldn\'t I have gone?  And leading up to the show I was making fun of it cuz he\'s kinda laim but I left the show saying \"hey, he\'s pretty good live and I was familiar with just about evey song, that was fun\".  I checked my musical pride/taste at the door and had a good time despite the performer. So SHUT UP JONES!!!!  '),(6827,3806796,'30,June,2004','2004-06-30 00:00:00','   Ben Bubar enjoys Sting in concert! This would have made my list for the funniest things of 2003 if I hadn\'t found out about it until 2004. I\'ll have to remember this one when I\'m reflecting on this year next December. Oh...good one, Ben.  '),(6828,3806796,'30,June,2004','2004-06-30 00:00:00','   I just mailed off ABCD#8 this morning, and opened the link to it\'s track listing to the right.  Track #14 is for all the O.C. fans out there, although I\'m not one of them. Happy New Year you guys!  '),(6829,3806796,'30,June,2004','2004-06-30 00:00:00','   okay dudes, the california country soundtrack is ready.  i\'m sending it to you tomorrow.  '),(6830,3806796,'30,June,2004','2004-06-30 00:00:00','   OK My shit is ready and I gotta wait 3 months. I made mine now cos I\'m seeking new employment and I don\'t know if I\'ll have the capabilities there or not. Hopefully we don\'t add any new members between now and may.  Cheers, Josh  '),(6831,3806796,'30,June,2004','2004-06-30 00:00:00','   \"I feel like a fag...\" - Josh  '),(6832,3806796,'30,June,2004','2004-06-30 00:00:00','   Dave send me an invite to your church clothes thing and call me about that stuff. I finally got #9, glorious. I like the outkast. Some of the rest of the stuff, I don\'t know. I guess everyone else in the club has like a girlfriend or something but I feel like a fag driving around downtown with buddies listening to whimpery sad stuff. I guess my musical tastes are a little more upbeat and when I do get down on like sad time music its usually like the white stripes or brmc. I don\'t get sad,  I get rocked out. It was nice to hear vaseline again though. It would also be cool to try and never have the same band as anyone else has had until we get into like the third year or something. I\'m gonna end up being able to put together my afterburners cds and compile the entire streets, pearl jam, radiohead, and ryan adams albums.  '),(6833,3806796,'30,June,2004','2004-06-30 00:00:00','   I\'m all for it as long as i get the ability to pass if i\'m on the slower side of aquiring music. unlike you mike, i\'m not boiling over w/ new stuff constantly.  '),(6834,3806796,'30,June,2004','2004-06-30 00:00:00','   I feel what your saying Mike.  The waiting does suck and I do feel like things are at a good place with AB so why change it.  Let\'s get more input on this because it should probably be voted on or at least discussed, right?    Oh and Mike you actually before Josh, not after.  '),(6835,3806796,'30,June,2004','2004-06-30 00:00:00','   no no no no no! nobody\'s cuttin\' ahead of me again. i\'m tired of waiting. i\'m up after josh and that\'s it. and to tell you all the truth, i really don\'t want to add anyone to the club anymore. i feel very strongly about that. i feel if we add anymore people, we\'re gonna lose our grip. i already can\'t stand waiting six months between mixes. i\'m telling you. you guys are already missing out on so much from me. let\'s cut it off, alright?  '),(6836,3806796,'30,June,2004','2004-06-30 00:00:00','   Dave, how many times do we need to send out addresses?  Let me guess, you lost them in the move.  Well, because we love you so much we\'ll go ahead and send them to you AGAIN....but dont go losing them this time.  Since Esther is joining the team officially(right?), we\'ll need to get all the needed info and decide where she\'ll be placed in line.  Does she just fall in at the end of the line?  Or, does she go right away, like after Dave?  What do you all think?  '),(6837,3806796,'30,June,2004','2004-06-30 00:00:00','   i need everyone\'s addresses, i\'m going to get esthers too.  '),(6838,3806796,'30,June,2004','2004-06-30 00:00:00','   Josh Woodyard 264 Altamonte Bay Club Circle #106  Altamonte Springs, FL 32701  '),(6839,3806796,'30,June,2004','2004-06-30 00:00:00','   joshy where you been?!?!?!?  '),(6840,3806796,'30,June,2004','2004-06-30 00:00:00','   i think i\'m on the dang blog children  '),(6841,3806796,'30,June,2004','2004-06-30 00:00:00','   i\'m going through the emotions of the typical \"former college athlete.\" you know how when they \"were great\" and they could excercise more than what a real life schedule allows them to do.  the reason AB was founded, was b/c i was loosing my musical edge w/ the ALLbusiness life. I\'ve recruited the finest men i can find to help a brutha keep his edge. so don\'t be suprised if I don\'t contribute the same quality  magical mike,  sexy scott, brillian ben, magnificant matt, Juicy Josh, or Ashy ash bring to the round table.  '),(6842,3806796,'30,June,2004','2004-06-30 00:00:00','   You did not just bring up CRD! That\'s amazing you even remember that!  Too funny.... ....Google\'s got nothing on CRD, that would\'ve been great to find an old picture.  '),(6843,3806796,'30,June,2004','2004-06-30 00:00:00','   yeah, dave. thanks to afterburners 2 we all know you\'re a wigga, too - what with your john reuben CDs and all. ouch! CRD, christian rapper dennis, baby - playing at the upper room friday nite!!!  '),(6844,3806796,'30,June,2004','2004-06-30 00:00:00','   urlLink ...here\'s why.   This is stupid, why even report this?  '),(6845,3806796,'30,June,2004','2004-06-30 00:00:00','   ummm yea,  I didn\'t realize i was cruzin\' with such gangsta\'shiz\'nit yo. i got some mutha\'grubbin\' homies tappin a case that they gave me on the A to the bumpin\' B site. don\'t be crampin\' the cause cause a nigga\'can\'t stop the funk or the hype beats that be blowin\' up this place. can you feel it my fellow homies? let our true cuLaz represent.  '),(6846,3806796,'30,June,2004','2004-06-30 00:00:00','   1.  Thank Mike Jonze for the rap mix he\'s sending. 2.  Tell people to get over to illegal-art.org and d-load those Grey album tracks, they\'re so dope! 3.  Eat Marty\'s jambalayaÂ  this friday at the pub. 4.  \"I got 99 problems but a bitch ain\'t one, if yur hav\'n girl problems I feel bad for ya son, hit me!\"  '),(6847,3806796,'30,June,2004','2004-06-30 00:00:00','   ben, no problem, dude. i\'ll try to put together a package of all those artists\' CDs. i\'m not promising, but i\'ll try to have it out to you by the end of the week - sort of a \"thank you\" for the outkast and mates of state full length\'s. i just got broken social scene and the walkmen\'s  bows + arrows  on 12\" vinyl yesterday. it\'s so sweet, dude!  '),(6848,3806796,'30,June,2004','2004-06-30 00:00:00','   dj danger mouse (london) has made a mix tape entirely comprised of jay z\'s black album, and the beatles white album.  it\'s called , aptly enough:  the \"grey album\".  it was released and then promptly  urlLink yanked  due to copyright concerns (dj danger mouse did not have the rights to sample either album).  you can illegally download the tracks  urlLink here .    but not for long.  it\'s fantastic.  '),(6849,3806796,'30,June,2004','2004-06-30 00:00:00','   Can somebody, anybody, make me a mix of said rap artists from the post below?  Dizzee Rascal rules and I\'m really curious as to what the others below sound like.  Is Kane West any good? I\'m thinking maybe Mike or Scott or Matt could do the honors of the mix, if it\'s not too much of a problem. Let me know, thanks!  '),(6850,3806796,'30,June,2004','2004-06-30 00:00:00','   Scott, if you liked Sage Francis...here\'s some others you should check out: cLOUDDEAD, Aesop Rock, Why?, King Geedorah, DoseOne, Buck 65, Atmosphere - just to name a few. Trust me...it\'s not for everyone, but if you liked SF, then it might be worth your time to look into some of those names. I probably dig cLOUDDEAD and Why? the most out of those I listed. But most would think it\'s pretty weird. The others are more beat driven, where those two can drift into the abstract. Some more underground hip/hop that\'s good but hasn\'t had the \"emo rap\" tag slapped on it: any of MF Doom\'s incarnations (Madlib, especially though), Mr. Lif, El-P...hmm, that\'s all I can think of right now. Out of the mainstream stuff, I\'m really liking Kanye West a lot, and anything that the Neptunes do is magic. It sounds like I\'m a big rap listener, but I\'m not. I just get in that mode every once and awhile, where some beats & rhymes is just what I need. This stuff\'s what I turn to to satisfy that craving.  '),(6851,3806796,'30,June,2004','2004-06-30 00:00:00','   ben posted that picture of me, and i might have to kill him for it.  also:  matt had a track ABCD 8 called \"crack pipes\" by sage francis.  i\'ve now got the whole album, and it\'s fantastic - s.f. is a spoken word guru - turned - hip/hop emo kid.  so cool.  thanks for the heads up matt.  '),(6852,3806796,'30,June,2004','2004-06-30 00:00:00','   Dave, get with the program and sign up for Photo Bucket!  '),(6853,3806796,'30,June,2004','2004-06-30 00:00:00','   ohhh,  if Only i can be so cool. scott, i forgot how hot you are.  '),(6854,3806796,'30,June,2004','2004-06-30 00:00:00','     '),(6855,3806796,'30,June,2004','2004-06-30 00:00:00','   urlLink photo bucket  is a good hosting site. it even gives you the code so all you do is copy and paste it.   in other news, i\'m starting another cd mix club with some of my friends. it\'s going to be called the letter bomb club. should be good times.  '),(6856,3806796,'30,June,2004','2004-06-30 00:00:00','   The same way you post pics on any blog site or website for that matter.  You need a host....A place to upload the images to.  Mike, maybe you could direct Dave to one of the many free image hosting sites on the web. Dave, Once the image is uploaded to the host you need to write the code that links it from the host to the website or blog entry in this case.  This( ) is a basic line of code for an image, only thing you would do is put the image\'s name between the quotes and remove the space between  Hey I\'m a real novice at this so if someone else has a better solution for Dave that\'s great.  '),(6857,3806796,'30,June,2004','2004-06-30 00:00:00','     '),(6858,3806796,'30,June,2004','2004-06-30 00:00:00','   I love that album.  But it took me alot longer to say that for \"Human Condition\", I dont know why, maybe I was expecting something different from him musicly, so I let it sit for a while.  I\'ve since then rediscovered it\'s greatness.  That album is so spiritual and honest, it\'s not even funny!  It\'s a beautiful thing to hear a man cry out to God and even question him and himself.  I think it\'s been an exciting journey to watch so far.  That is, if you\'ve been following him since The Verve days, he\'s a different man now for sure!  Although, every time he comes out with something new, part of me is wishing for a song like \"The Rolling People\" and those wild Nick McCabe guitar riffs. \"Dear God, please let Richard and Nick get past their differences and reform The Verve, Amen\".  '),(6859,3806796,'30,June,2004','2004-06-30 00:00:00','   okay, I am just wondering, is anyone else still listening to the brilliant Ashcroft second? I can\'t take it out of my changer. I hope this site isn\'t just feeding music consumption. we all know he\'s searching, track 5: \"If there\'s a God, well can you hear me now, i am crying out, hoping you know.\" \"i DOn\'t know if my prayers are received, the ones i\'ve sent.\"  '),(6860,3806796,'30,June,2004','2004-06-30 00:00:00','   i think this is the only disadvantage to living on the west coast is that i have to wait longer for my discs. can\'t wait to hear that shiz yo! ben i tried calling this week, but i only had VA\'s cell #. my cell is down so i don\'t have your home number. ps. how do i post pictures? i want to post the new car I got this weekend, a Lexus LS 400...YEA BOYEEE!!!! big pimpin\' what what!  '),(6861,3806796,'30,June,2004','2004-06-30 00:00:00','   1.  I\'m lov\'n those Ryan Adams tunes!  I\'m going to get that cd for sure.  Nice choice Scott. 2.   urlLink \"She Dont Use Jelly\" on Amazon  3.  I liked the Sting vs. Sean Paul performance last night too.  They played Roxanne the same way(w/o Sean Paul of course) when I saw him a few weeks ago, reggae style.  Very cool. 4.  Good mix Scott, sorry if it sounded like I was coming down on it.  Oh and I like that Streets song too.  '),(6862,3806796,'30,June,2004','2004-06-30 00:00:00','   1 - ben:  did you ever think that maybe, just maybe, you were the first to get it, and thus should not ruin the suprise for everyone else? hmmmmm  1.5 - yeah, jack white blew the walls off the place.  i can\'t believe bb king (handing out the next award right after the performance) did not mention it.  1.75 - what about sting and sean pual?  i thought that was pretty dope - but maybe i\'m the only dancehall regae fan in AB, that\'s understandable.  MIKE JONES STAY OUT OF THIS ONE ;-)  2 - gheeze matt, of course you can get it all for free.  i\'m not sure how much of that i actually bought.  i just \"discovered\" it on itunes. like i discovered the exclusive ep\'s from siguor ros this week.  3 - there goes my TFlips history award.  i thought it was new.  are you sure it\'s the same songs / same EP? or is it old songs re-done.  have you seen cool hand luke? and why did i hear all of the buzz about plastic jesus recently (i think matt also told me about that track - which is one of the reasons i started listening to it. )  PS:  go and get the japanese version of \"yoshimi battles the pink robots\"  -it\'s mas bueno.  3.19 - bowtie.  whoops for some strange reason i thought \"happy v-times day\" was on ben\'s, and bowtie on mine but i got mixed up.  4 - was there anything you liked on ABCD nine ;-)  '),(6863,3806796,'30,June,2004','2004-06-30 00:00:00','   Hey Scott, I like your packaging, I almost threw it out without even looking at it closely.  I thought it was Apple trying to get me like AOL does with their cd\'s.  Hey Woodski, your next!  And dont go putting Outkast\'s \"Bowtie\" on yours too!!!  '),(6864,3806796,'30,June,2004','2004-06-30 00:00:00','   what\'s up with exclusive tracks on itunes? i went to the store and saw that the darkness has two exclusive track on there. so then i thought, i\'d rather get my music for free. so i went on kazaa and one of the songs was on there. downloaded it, same song and it sounds good.   yeah, real exclusive.  '),(6865,3806796,'30,June,2004','2004-06-30 00:00:00','   \"within a week, they were doing more interviews than they could handle, and the orgiastic british press had declared the group to be a better live show than the return of jesus christ. \"  from the  urlLink dallas observer  via  urlLink tb  '),(6866,3806796,'30,June,2004','2004-06-30 00:00:00','   Hey, did anyone else see Jack White blow it up last night at the Grammys?  I mean, he tore shit up on the guitar!!!  That was an amazing display of guitar geniusness.  They should have just given him an award for that performance alone.  '),(6867,3806796,'30,June,2004','2004-06-30 00:00:00','   I\'m pretty sure \"She Dont Use Jelly\" & \"Plastic Jesus\" is NOT exclusive to iTunes.  As a matter of fact, I remember hearing \"SDUJ\" like 10 years ago.  I think it was probably the first Flaming Lips song I\'d ever heard.  Yeah, both of those songs are on the cd single that came out in \'93.  '),(6868,3806796,'30,June,2004','2004-06-30 00:00:00','   dudes who is on IM? throw me on yours. churchclothes@hotmail.com  '),(6869,3806796,'30,June,2004','2004-06-30 00:00:00','   ben, can you make a link for this?  thanks.   urlLink abcd nine  '),(6870,3806796,'30,June,2004','2004-06-30 00:00:00','   Woodski,  checked out the page and I\'ll have to get some more info on some of those  bands.  I just mailed out my first disc today.  We\'re called the \"Disc  Junkies.\"  Very much kin to the Afterburners.  Grab some silky waves in  Diego and tell Danae that I say what it is.  I\'m heading to Snowbird this  weekend.  Trying to get some fresh snow as Denver doesn\'t have too much.   I\'ll stop by the Artic Circle for you and then chase it down with some Hogi  Yogi.  Hayes  '),(6871,3806796,'30,June,2004','2004-06-30 00:00:00','   Ursus says about Elbow\'s new album:  \"#9 is probably the catchiest song on the record, â€œButtons and Zipsâ€. It begins with a really cool guitar progression and a few doo-doo-doo-doo-dooâ€™s. Iâ€™m already sold, but by the time the chorus sneaks in so naturally, saloon piano and handclaps in tow, Iâ€™m absolutely won over. This is the melody that I find coming to mind hours and days later, stuck in my head in a good way\".  Thanks Mike, you SOB.  I swear, if it was not all wrapped up and postage paid, i would immediately open up ABCD9 and re-burn all of the cd\'s without that track on them!  grrrrr...  ABCD9 is wrapped up and sitting on my desk.  I am waiting for josh\'s address.  someone email it to me (scottgravatt@yahoo.com).  In the morning, after i take an unexpected trip to the dentist for my daughter ( urlLink picture ) i will be at the post office, and they will be in the mail.  2 things.    1 -  look for it in the mail.  i mean really look, you might miss it :-) 2 - i have a tracklisting complete with links, commentary and liner notes online, but i don\'t want to reveal the link until you get the CD.  so i\'ll post it monday / tuesday am, or give it to you when you email me for it, whichever comes first.  enjoy!  '),(6872,3806796,'30,June,2004','2004-06-30 00:00:00','   If asher didn\'t already send me his disc for a road tri[, i would be in trouble all the way back to disc 6. what did i get myself into with you guys? you all are SERIOUS! I not worthy to the mass quality of music coming my way. i\'m feeling like i have to set aside some serious listening time to know 7 and 8. wow guys, very impressed.  '),(6873,3806796,'30,June,2004','2004-06-30 00:00:00','   8  '),(6874,3806796,'30,June,2004','2004-06-30 00:00:00','   yo,   i\'m done with ABCD9.  my plan is to wrap up the casing and the notes, etc... tomorrow am, and then mail it off on thursday.  you all should see it within a week after that.  sorry for the delay, remember - up until a week ago, i thought this was a march project for me :-)  hmmm ... should i put 8 or 9 sting songs on this one?  '),(6875,3806796,'30,June,2004','2004-06-30 00:00:00','   Alright, boys. After a weekend toiling away, I finally narrowed down the field to 18 tracks and an hour and nineteen minutes of bonafide brilliance. I created the blog for the  urlLink tracklisting , but I don\'t have access to the template for this page. As such, I can\'t add the link to the rest of them. So Ben, if you could do me the honors, I\'d appreciate it. Following suit, the link is: http://ABCD11.blogspot.com.  CDs should ship tomorrow, so be expecting them in your mailboxes soon. Enjoy.  '),(6876,3806796,'30,June,2004','2004-06-30 00:00:00','   Hey guys. I just got fired from my job yesterday and then, last night, my house burnt down to the ground along with everything inside it...including my computer. So I don\'t think I\'ll be able to step up to bat for Afterburners 11...\'cause I don\'t have a computer...even though I\'m on one now. Can someone else go next and give me another month to come up with another excuse? Thanks. Oh, by the way, I\'m kidding. 11 will be undergoing construction this weekend.  '),(6877,3806796,'30,June,2004','2004-06-30 00:00:00','   You can mix Jay-Z\'s \"The Black Album\" with just about anything these days.  Now it\'s being mixed with Weezer\'s \"The Blue Album\".  Check it out  urlLink here .  '),(6878,3806796,'30,June,2004','2004-06-30 00:00:00','   I got ABCD#10.....no other comments at this time.  '),(6879,3806796,'30,June,2004','2004-06-30 00:00:00','   check the latest music related entry on  urlLink gravattville  '),(6880,3806796,'30,June,2004','2004-06-30 00:00:00','   sweet.  this \"team blog\" is really working out.  i\'d like to propose a future afterburners get together.  like a short weekend centered around a music festival.  the winter music confrence here in miami, or what\'s the name of that one in new york, or SBSX in texas, or even sundance.  that would be so sweet if we could all meet up and get some nice tickets.    it\'s cool how you can get to know someone by a mix CD.    i\'ve got some sweet hook-up tickets to the Strokes tomorow night.  yeah!! yeah!! (the last time she told me they were hook-up tickets\" it was like 8th row center radiohead), so i\'m expecting something nice.  '),(6881,3806796,'30,June,2004','2004-06-30 00:00:00','   I still haven\'t gotten my ABCD#10!  '),(6882,3806796,'30,June,2004','2004-06-30 00:00:00','   fellow \'burners,  i am now officially registered as a blogger.  #10 was superb.  super.  supa.  i love nada surf.  never heard them before.  always thought their band name was clever, but not anything that reliable friends put into my hands.  i now question the validity of several friendships.  wait, no.  but i DO like nada surf.  very nice.  kent is always a favourite, and \"quiet heart\" is one of my favourite songs.  the rest of the tunes are cool, too, flowing together nicely with my faves.   i could have sworn that the turin brakes song was a ween track...  dude, but no.  aieet,  asher  '),(6883,3806796,'30,June,2004','2004-06-30 00:00:00','   I live in downtown orlando now folks:  911 N. Orange Ave. #250 Orlando, FL 32801  '),(6884,3806796,'30,June,2004','2004-06-30 00:00:00','   as anticpated, my attempts to impress the omnicient Jones failed. I should just bow out.  that braddigan was great when i was stuck in my room all day last fall hacking away at cxc and freelance work. i was rarely out of the apt. much longer than to see the sunset. so i needed some upbeat good music. that hit the spot for me. can\'t say i have any idea what\'s going on w/ the O.C.  '),(6885,3806796,'30,June,2004','2004-06-30 00:00:00','   Got mine last week - Friday, I think. You may have cause for concern, Bubar. It\'s alright, Dave. The mix, I mean. I\'m always up for some Spoon. I was surprised to find that on it. I didn\'t know you\'d heard of them. But, yeah, they\'re great. I urge anyone that likes the Spoon songs to search out more of their stuff. \"Girls Can Tell\" and \"Series of Sneaks\" are my two favorite albums of theirs. Also glad to see some Nada Surf on the disc. They\'re pretty cool, even though they won\'t ever get out of the wake of \"Popular\" no matter how hard they try. There was a great moment on last week\'s episode of \"The O.C.\" at the end where Anna was leaving to go back to Pittsburgh and Seth didn\'t want her to go and they were playing Nada Surf in the background. It was awesome. A cover of \"If You Leave\" by Orchestral Maneuvres in the Dark. I seriously almost cried. It was heartwrenching. Anyway, next time less Braddigan. And more John Ruebens.  '),(6886,3806796,'30,June,2004','2004-06-30 00:00:00','   Dude, I still havn\'t gotten my abcd#10!  '),(6887,3806796,'30,June,2004','2004-06-30 00:00:00','   the gary allen song was a last minute entry. Danae was listening to the cd and I overheard this song. it was so funny i had to add it. It starts off sounding so \"spiritually\" encouraging, but listen to the words especially him talking toward the end.  '),(6888,3806796,'30,June,2004','2004-06-30 00:00:00','   Dave, I moved your track listing over to the right with the rest of them. Looks good, cant wait to hear it.  '),(6889,3806796,'30,June,2004','2004-06-30 00:00:00','   do you guys have that \"blog button\" on your browser menu bar?  the post pictures Question was supposed to be on the cxcsamples site, and i accidently put it on the ab site. i\'m serious, i\'m not putting any non-music related questions on the ab site ever ever ever again. i am still very truly sorry for my dilution of the almighty blog. :)  '),(6890,3806796,'30,June,2004','2004-06-30 00:00:00','   Oh my gosh! SOMEbody ain\'t listenin\'. refer to the post regarding this matter further down the screen (Wednesday, Feb.11) for directions about posting pictures...just raggin\' on you, dave. \'cause it\'s so easy. OH!  '),(6891,3806796,'30,June,2004','2004-06-30 00:00:00','   ben, how do you post pictures?  '),(6892,3806796,'30,June,2004','2004-06-30 00:00:00','   dang, talk about reprimanditation. didn\'t realize we run such a tight ship here. :)  '),(6893,3806796,'30,June,2004','2004-06-30 00:00:00','   Dave can you not use the Afterburners blog for other business?  It\'s all good, but that\'s why the other blog was made.  And there\'s always email....which is always faster.  I mean, what if your artists friends dont check this blog for like a week?  It\'s all in love, but let\'s just keep this blog for the Music and AB\'s related stuff.  '),(6894,3806796,'30,June,2004','2004-06-30 00:00:00','   guys, kind of strange that all my artist friends are also in AB, so I\'m just going to use this medium to promote the other. go to cxcsamples.blogspot.com for a message. Ben did/will invite you so you can post artwork there. we need 50 logos by April 15.  '),(6895,3806796,'30,June,2004','2004-06-30 00:00:00','   i just got \"jam pack\" in the mail today... the additional 2000 sounds upgrade... i\'m not sleeping for weeks!!!!  '),(6896,3806796,'30,June,2004','2004-06-30 00:00:00','   i was going to make this a complete suprise but i\'ve realized that it\'s better i let it out now.  ABCD#17 will be composed entirely by myself. i have been using garage band 3 months now and i\'m hooked!!!! i would love to see the evolution of AB into everyone contributing their OWN music. if you\'re skeptical that we\'ll all put out the \"same stuff\", you\'re wrong. this program is as deep as photoshop is. it will go as deep as your creativity will take you. to quote my good friend hayes:: \"it\'s better to play in a bad band, than listen to a good one.\" i\'m hoping this gives you all time to think and prepare for your next disc. i have 3 songs done, hoping for 10 by 10/1/04 who\'s with me!!!  '),(6897,3806796,'30,June,2004','2004-06-30 00:00:00','   I like this discussion - here are my top three from each disc in the Afterburners catalog, thus far:   v.1 : 3) Ashcroft - The Miracle, 2) Radiohead - Where I End..., and 1) Herbie Hancock - Rock It   v.2 : 3) South - Here On In, 2) Seahorses - Love is the Law, and 1) John Reubens - Hindsight (purely for it\'s tremendous humor value - I can now drop this one whenever I want to blast Dave\'s mixmaking skills)   v.3 : 3) Bright Eyes - One Foot In Front of the Other, 2) Postal Service - Sleeping In, and 1) Aqualung - Strange & Beautiful (I Put A Spell On You)   v.4 : Impossible to pick just 3 - no filler ANYWHERE!   v.5 : 3) Death In Vegas - Scorpio Rising, 2) Modern English - I Melt With You, and 1) Primal Scream - Autobahn 66   v.6 : 3) Jurassic 5 - Concrete Schoolyard, 2) Death Cab - I Was A Kaleidoscope, and 1) Frou Frou - Hear Me Out (SO GOOD in the \"Garden State\" trailer!)   v.7 : 3) The Faint - Total Job, 2) The Darkness - I Believe In A Thing Called Love, and 1) Cat Power - He War   v.8 : 3) Outkast - Bowtie, 2) Travis - Midlife Krysis (perhaps a reference to Canadian CCM\'ers \"The Kry\"? - eerrrrr, I think not!), and 1) Chemical Brothers - Get Yourself High   v.9 : 3) Elbow - Buttons and Zips, 2) The Shins - Young Pilgrim, and 1) Polyphonic Spree - Light and Day (Reach For The Sun)   v.10 : 3) ----, 2) The Verve - Two Step, and 1) Spoon - Don\'t Let It Get You Down   v.11 : Once again, too many to narrow down the field. Exquisite compilation! Who is responsible for this one!  I\'m out!  '),(6898,3806796,'30,June,2004','2004-06-30 00:00:00','   ABCD#6: 1. Ben Folds: Annie Waits 2. Jurassic 5: Concrete Schoolyard 3. Venus Hum: Wordless May  ABCD#7: 1. Sage Francis: Crack Pipes 2. Mates Of State: Everyone Needs An Editor 3. Death Cab For  Cutie: We Look Like Giants  ABCD#8: (yeah i know it\'s mine, so what) 1. The Rapture: House Of Jealous Lovers 2. Over The Rhine: Long Lost Brother 3. U2(Johnny Cash): The Wanderer  ABCD#9: 1. Ryan Adams:  Love Is Hell 2. The Polyphonic Spree: Light & Day 3. Dizzee Rascal: Jus\' a Rascal  ABCD#10: 1. Turin Brakes: Blue Hour 2. The Verve: Two Step 3. Nada Surf: Inside Of Love  ABCD#11: 1. The Cure: Halo 2. Waybe Everett: Chalk 3. Franz Ferdinand: Take Me Out  '),(6899,3806796,'30,June,2004','2004-06-30 00:00:00','   i thought i would list my three favorite songs (that i hadn\'t heard already before) from each of the abcd\'s that i\'ve gotten.  #6 1. Bjork: \"It In Our Hands\" 2. Frou Frou: \"Hear Me Out\" 3. Ben Folds: \"Annie Waits\"  #8 1. Ours: \"Red Colored Stars\" 2. U2 w/Johnny Cash: \"The Wanderer\" 3. Travis: \"Somewhere Else\"  #9 1. The Flaming Lips - (plastic \"jesus\") [porch version] 2. Dizzee Rascal - jus\' a rascal 3. Johnny Cash - redemption song  #10 1. Nada surf-blizzard of 77 2. spoon-don\'t let it get you down  3. nada surf- inside of love  #11 1. Beck - Everybody\'s Gotta Learn Sometimes 2. Franz Ferdinand - Take Me Out 3. Schneider TM - Reality Check  what\'s everyone else\'s?  '),(6900,3806796,'30,June,2004','2004-06-30 00:00:00','   urlLink It  looks pretty easy to use.  You should pop into an Apple store and play around with it.  '),(6901,3806796,'30,June,2004','2004-06-30 00:00:00','   anyone had a chance to try it yet?  '),(6902,3806796,'30,June,2004','2004-06-30 00:00:00','   PJ Harvey has been added to the bill for Lollapalooza \'04!  That\'s it, I\'m in!  '),(6903,3806796,'30,June,2004','2004-06-30 00:00:00','   anyone on limewire or downloading programs similar should get any and all breezeblock sets. they feature all the best djs and I\'ve been listening to the kid koala and hexstatic ones ad nauseum.  cheers  '),(6904,3806796,'30,June,2004','2004-06-30 00:00:00','   so i made that cd like a month ago cos i though i wouldn\'t have burning capabilities again for a while but i still can burn and i got loads of new crap since i made that old one so everyone\'s gettin both. I gotta go type up th tracklists and make a cover at home tonight then the pidgeon wil be away.  '),(6905,3806796,'30,June,2004','2004-06-30 00:00:00','   You ever get ABCD#11 in the mail?  I can burn you one if you still need. lemme know.  '),(6906,3806796,'30,June,2004','2004-06-30 00:00:00','   MM is great, I got the album this past weekend and it\'s so fun to listen to.  Makes me even more excited for Lollapalooza \'04.  '),(6907,3806796,'30,June,2004','2004-06-30 00:00:00','   urlLink modest.mouse :   i\'m so digging this new album!  ps:  the  urlLink video  for the first single is so dope.  '),(6908,3806796,'30,June,2004','2004-06-30 00:00:00','   Scott, Hey, man. Hope that CD gets to you. If worst comes to worst, just let me know if it never comes and I\'ll send out another one. I printed extra copies of the mix\'s artwork and can always burn another copy of the CD easy. So, that\'s it. Later.  '),(6909,3806796,'30,June,2004','2004-06-30 00:00:00','   i\'d love to give my input, but i\'m still waiting on my copy!  my mail has been screwed up for years, and it\'s pretty often that i do not receive important mail.  and the worst part:  it won\'t even come  back to the sender!  i hope i get this tomorrow.  '),(6910,3806796,'30,June,2004','2004-06-30 00:00:00','   http://www.enpack.com/ Sorry i don\'t know how to make it a link.  '),(6911,3806796,'30,June,2004','2004-06-30 00:00:00','   Guys, it\'s great to see the positive response to my mix. I had such a hard time whiddling it down to that. I had more I wanted to share. Thought about doin\' a double-disc, but nixed it. Anyway, I\'m psyched that you\'re all enjoying it so much. I admit I have been enjoying it very much myself. Now to field some of your questions/comments: Ben: the Wayne Everett disc. Ordered it straight from his label,  urlLink Northern Records , last year. It\'s a quality album. Course, I\'ve been a sucker for all things Prayer Chain and Lassie Foundation-related for years. All those guys have the Midas touch. Matt: I didn\'t notice the similarity to  Under The Milky Way , but now that you mention it, I can sorta hear it in my head. I\'ll have to go back and listen to it to fully realize what you\'re talking about. That\'s cool...cos that\'s a great song. You know what would be really sweet would be if he transitioned into it when performing the song live. Anyway, cool. Josh: The Cure do rule. Join the Dots box set, baby. B-sides, b-sides, and more b-sides. It\'s a treasure trove. Well, the set has got it\'s fair share of less than mediocre songs, but it\'s worth the price for the good ones. Disc 3 is the best. This Twilight Garden, Play, Halo, The Big Hand...great tunes. I\'m convinced the Cure were the best thing - musically - to come out of the 80s. The best thing, musical or non-musical, to come out of the 80s was a two-way tie between Snake Eyes and Storm Shadow. Yeah, you all know what I\'m talking about. Dave: Yes, I can be commissioned for a comp. Call me. We\'ll talk numbers sometime. I\'m seeing six figures...  '),(6912,3806796,'30,June,2004','2004-06-30 00:00:00','   Hey Jonze, where\'d ya get that Wayne Everett cd?  Not even Amazon carries it.  I really enjoyed hearing that song, takes me back to the days of listening to Mercury(love that album).  '),(6913,3806796,'30,June,2004','2004-06-30 00:00:00','   yea jones you\'re on a roll homie!. NICE WORK!  the only remedy for critical feedback for abcdx is to come back with such a solid disc. :) way to put your money where you mouth is. I might sponser you to make my next disc.  '),(6914,3806796,'30,June,2004','2004-06-30 00:00:00','   yeah, i agree. the mix is really good and i haven\'t heard of a lot of the bands on there.   one thing i thought was interesting was the song \"reality check\" by schneider tm has the same beat to it as the song \"under the milky way\" by the church. i was singing under the milky way to it. i thought it was kind of funny.  '),(6915,3806796,'30,June,2004','2004-06-30 00:00:00','   Lov\'n the mix man!  Must have listened to it like a dozen times this weekend. I love how there\'s so many artists on it that I\'ve never even heard of, and some that I\'ve heard of but haven\'t had a chance to listen to yet. Nice job!  '),(6916,3806796,'30,June,2004','2004-06-30 00:00:00','   jonze your cd is excellent this time. that\'s my favorite cure song. my sister got it as a b side on the high single when i was like in 8th grade. nice to hear it again. surfing on a rocket is my favorite track from talkie walkie also and i didn\'t even know there was a new phoenix album out now i have to go to virgin. i guess i\'ll send mine around like early mayish, but i might have to tweak it now and give all the old copies to people around here.  cheers  '),(6917,3806796,'30,June,2004','2004-06-30 00:00:00','   josh did my copy make it to the west coast?  '),(6918,3806796,'30,June,2004','2004-06-30 00:00:00','   ok, i am having a serious mail problem.  write this down and send all ABCD\'s to:  Island Water Sports Scott Gravatt 1985 NE 2nd St Deerfield Beach, FL 33441  Thanks!  '),(6919,3806796,'30,June,2004','2004-06-30 00:00:00','   ok, i am in the throes of track decision now...  how can a 700 mb CD seem so small?  some highlights to look forward to... -a great new song by a chick from Louisiana called \"Toxic.\" -an olde school ballad to \"Iesha.\" -and a slightly explicit version of \"Friends are Friends Forever.\"  rawk on. get ready to get kataaah!-ed again...    lades,  ash  '),(6920,3806796,'30,June,2004','2004-06-30 00:00:00','   I couldn\'t possibly have heard him(Josh) right when he said that abcd#12 was late(3 weeks) because he was working on the art/liner notes.  So I take it back, I have no idea why he was late.  It doesn\'t matter now anyway.  The mix on the one disc labled abcd#12is great, i really like it.  Love the new UNKLE stuff on there.  The other disc, which I\'m assuming is the older one, is classic Josh. Nice job compiling the songs....btw, a cd insert is 4.75 inches square.  '),(6921,3806796,'30,June,2004','2004-06-30 00:00:00','   Josh(ua), you must really want us to hear that song.  It\'s made it onto both of the abcd mixes(#5 & #12).  Is there maybe some sort of hidden meassage in the lyrics that we\'re supposed to hear?  Oh, i know, it your favorite song, right? Agh, i\'m just bust\'n yur ballz! Good on ya to give us 2 discs this time around, your a good man.  '),(6922,3806796,'30,June,2004','2004-06-30 00:00:00','   check the sidebar:  it\'s really simple, but it just takes a little work to get it there.  asher and josh(ua) still need to fill in the profile (that\'s only important b/c it\'s got a space for a current email address, in case someone needs it, for a mailing address or something)  other than that, don\'t use images in your posts, unless they are smaller than 400px wide, post your tracklistings as you mail them out, ben and i can do the rest.  please leave suggestions / opinions is the comments box of this post.  peace!  '),(6923,3806796,'30,June,2004','2004-06-30 00:00:00','   MAY 2004 Mixed by Josh Woodyard  1)UNKLE â€“ The Reign 2)Air â€“ Brakes On 3)The Beatles â€“ Flying 4)DJ Shadow â€“ Six Days (Mos Def Remix) 5)Groove Armada â€“ Join Hands 6)Hampton Hawes Trio â€“ Blues For Jacques 7)DJ Shadow â€“ The Number Song (Cut Chemist Remix) 8)Kevin Shields â€“ Are You Awake 9)Lee Perry â€“ Police And Thieves 10)My Bloody Valentine â€“ Untitled 11)50 Cent â€“ In Da Club (NIN Remix) jk 12)UNKLE - The Knock On Effect (featuring Noel Gallagher & Mike D) 13)Radiohead - Climbing Up the Walls (Cinematic Orchestra Remix) 14)Spiritualized - Lay Back in The Sun 15)Beta Band - I Know 16)Jesus and Mary Chain - Head On 17)The Roots - The Seed 2.0 18)Travis - Unbelievers 19)UNKLE - In A State        BONUS DISC: \"Hafta Burn\'em\" 1)Method Man â€“ Bring the Pain (chemical brothers remix) 2)Outkast â€“ Roses 3)Isley Brothers â€“ Itâ€™s Your Thing 4)Primal Scream â€“ Movinâ€™ On Up 5)Death In Vegas â€“ GBH 6)The Freestylers â€“ Phenomenon One 7)Primal Scream â€“ Star 8)Toots and the Maytals â€“ 54-46 Was My Number 9)Black Rebel Motorcycle Club â€“ Kill The U.S. Government 10)Beck â€“ Fuckinâ€™ With My Head 11)Black Rebel Motorcycle Club â€“ In Like The Roses 12)Primal Scream â€“ Medication 13)Evan Dando â€“ Live Forever 14)David Holmes â€“ Hey Lisa 15)Oasis and Friends â€“ Fade Away 16)Ted Nugent â€“ Stranglehold 17)Charlatans â€“ Nine Acre Court  '),(6924,3806796,'30,June,2004','2004-06-30 00:00:00','   this site is having a little work done on the template. Thanks, Management  '),(6925,3806796,'30,June,2004','2004-06-30 00:00:00','   I just got an email from Josh, he says he has 2 discs to send out and that it took him this long because he was working on the liner notes and art work.  But he says they are done.  Better late than never.  He needs everone\'s addresses though.  So email him with you address. I\'m out.  '),(6926,3806796,'30,June,2004','2004-06-30 00:00:00','   Are you going to finish the changes to this blog?  Like the cd track listings in the side bar(i only see abcd 1), and the names of the members are all out of order.  Before they had dates next to each person\'s name as to when they were going next. Does anyone remember who\'s going after Josh?  It was whoever was going in June.  I know I was August.  Maybe it was Asher.  '),(6927,3806796,'30,June,2004','2004-06-30 00:00:00','   hey Josh, i got your abcd #12!  Not bad... I had no idea you were into Hillary Duff and Josh Groban though. Anyone else get there\'s yet?  '),(6928,3806796,'30,June,2004','2004-06-30 00:00:00','   some neat things happening at blogger, and this website gets an instant *free* update.   a couple of highlights:  1 - comments!  hell yeah!  no more adding new posts just to respond to the old ones.    2 - profile!  the next time you log in please create your profile.  make sure under the section for your home page *do not out the afterburners site*  put your website.  if you don\'t have one leave the website questions blank, but please fill out your profile, and keep your email address current.    3 - one of the coolest features is that now blogger keeps every single post that you make on it\'s own archive page - this means that from now on, when you send out a mix, it is your job to log in to the website, and post the tracklist.  - cause then ben and i will be using that post to keep a permanent archive on the sidebar.  that also means that later this week, you will notice a string of posts with just old tracklistings.  that is so i can start to create the permanent links for them off to the left.  please leave suggestions in the comment box!  i am totally open.  thanks!  '),(6929,3806796,'30,June,2004','2004-06-30 00:00:00','   Has anyone gotten ABCD#12 yet?  '),(6930,3806796,'30,June,2004','2004-06-30 00:00:00','   Hey Josh, how\'s #12 coming? Will we be seeing it soon?  '),(6931,3806796,'30,June,2004','2004-06-30 00:00:00','   MAY 2003 Mixed by Ben Bubar  1. Herbie Hancock: \"Rock It\" 2. Pearl Jam: \"You Are\" 3. Pearl Jam: \"Small Town\"(Live in WPB, FL. 4/11/03) 4. Radiohead: \"Where I End And You Begin\" 5. Radiohead: \"Punch Up At A Wedding\" 6. Radiohead: \"Go To Sleep\" 7. Pete Yorn: \"Carlos (Don\'t Let It Go To Your Head)\" 8. Pete Yorn: \"When You See The Light\" 9. Pete Yorn: \"So Much Work\" 10. Jason Mraz: \"Curb-Side Prophet\" 11. Jason Mraz: \"The Remedy\" 12. Jason Mraz: \"You And I Both\" 13. The Bad Plus: \"Love Is The Answer\" 14. Richard Ashcroft: \"The Miracle\" 15. The Jayhawks: \"All The Right Reasons\" 16. The Jayhawks: \"Stumbling Through The Dark\" 17. John Mayer: Covering Radiohead\'s \"Kid-A\"  '),(6932,3806796,'30,June,2004','2004-06-30 00:00:00','   APRIL 2004 Mixed by Mike Jonze  01 thecure.halo 02 franzferdinand.takemeout 03 thefrenchkicks.closetomodern 04 duraluxe.allright 05 wayneeverett.chalk 06 radiohead.fog 07 beck.everybodysgottalearnsometimes 08 holgerczukay.fragrance 09 air.surfinonarocket 10 phoenix.everythingiseverything 11 tahiti80.alovefromouterspace 12 duraluxe.hitsohard(winsomeversion) 13 lalipuna.lefthanded 14 schneidertm.realitycheck 15 telefontelaviv.myweekbeatsyouryear 16 bfleischmann.passby 17 thecansecos.thissmalldisaster 18 halloweenalaska.youreit  '),(6933,3806796,'30,June,2004','2004-06-30 00:00:00','   MARCH 2004 \"CALIFORNIAXCOUNTRY\"  Mixed by Dave Woodall  1. Braddigan- Fallin\'  2. Nada surf-blizzard of 77  3.spoon-don\'t let it get you down  4. death on wednesday - (track 5 from the ep, songs to ___ to.)  5.braddigan - pura vida  6.kent - king is dead  7. verve- two step  8.Gary Allen - what would willy do  9. Turin Brakes - blue hour  10. mason Jennings - living in the moment  11. Tait- electric avenue  12. spoon - back to the life  13. Death on wednesday - sympathy  14. Kent - quiet heart  15. nada surf- inside of love  16. turin brakes - self help  17. slightly stoopid - sweet honey  18. braddigan - The Crossing  19. samples - little silver ring  20. death on wednesday. (hidden track 6 from ep.)  '),(6934,3806796,'30,June,2004','2004-06-30 00:00:00','   FEBRUARY 2004 Mixed by Scott Gravatt  1 - jus\' a rascal - dizzee rascal from the album \"boy in a corner\".  i heard about dizzee last year from a friend\'s blog that lives in the uk, and is pretty hip in the club scene.  i would have loved to be the first one to introduce you to him, but alas, everyone seems to have beat me to the punch.  he\'s shown up in every major music magazine and website as of this month.  damn.  i was almost cool  2 - kill the band - ours from the album \"precious\".  this is not the best song on the album, it\'s just one of my favorites.  i chose the first track off the album, \'cause ben (abcd8) chose the last track.  if you like these, you should hear the stuff in between.  3 - young pilgrim - the shins from the album \"chutes to narrow\".  this should have made it to my favorites of 2003 list, but i didn\'t get to it in time.  4 - this is not - blonde redhead from the album \"melody of certain damaged lemons\".  ok here it is.  this is the track to impress your friends with.  just mosey up to the bar, say hi to your friends, and say \"how about that blonde redhead album, huh?  then you can say \"what? you havenâ€™t heard of blonde redhead?  let me burn you the album\"  instant indie-music idol status for all.  5 - she don\'t use jelly - the flaming lips from the itunes exclusive ep of the same name.  i\'ve been into some of the exclusive stuff on itunes, so this is the first of a few.  6 - light and day/reach for the sun - the polyphonic spree from the album \"the beginning stages of\".  i included this track, on the off chance that some of you have committed the unforgivable sin of not listening to this.  tps is probably the most original sound in a decade, and besides:  it just makes you all warm and happy inside.  people say that the live shows are like a really good religious experience.  7 - give me my lighter back - the streets from the exclusive itunes ep of the same name.  8 - bowtie - outkast from speakerboxxx / the love below (the album that you will be tired of hearing about after the grammy\'s).  you got to hand it to them, this is what good american hip-hop is supposed to be - pushing limits and boundaries, evolving, and creativity.  9 - ******** (plastic \"jesus\") [porch version] - the flaming lips from the itunes exclusive ep \"she don\'t use jelly\".  i heard this story about how wayne (tfl) gave jack white (twstripes) some sort of dashboard jesus that lights up, and then wrote a song about it.  i disregarded the story.  i have since seen the movie \"cool hand luke\", and realize that this song is a cover of a song in that movie.  more than that;  the movie is not the original song.  apparently it is an old negro spiritual song that came to represent rebellion against oppression.  i\'ve now watched the movie about 25 times, and i have all kinds of elaborate theories about cool hand luke, and the gospels that i would love to share with you all sometimes.  the gist of the theory is that the movie is one big allegory for the gospel of luke.  anyway, see it. and here\'s the song.  10 - love is hell  - ryan adams from the ep \"love is hell volume 1\".  love is hell.  11 - leaves (edit)  - ours from the album \"precious\".  here\'s a song from the middle just incase you didn\'t believe me.  12 - mary\'s in india - dido from the album \"white flag\".  there are more obvious choices from the album, but this song sticks with me b/c of the lyrics.  give it some thought, listen to the hook at the end, and go ahead and feel sad, it\'s ok.  13 - wonderwall - ryan adams from the ep \"love is hell volume1\".  a word to the wise:  skip ryan adams latest album \"llor dna kcor\" that\'s rock and roll backwards btw.  these ep\'s (love is hell volume1 and love is hell volume2) were released at the same time as that album, and while they are getting absolutely no press play, they are 10X the album that \"llor dna kcor\" is.  take for example this version of the popular oasis song.  14 - 11:11 - rufus wainright from the album \"want one\".  i vowed never to listen to him after that shrek debacle.  but i caved in, and i am glad i got over my homophobia.  i really like it, i don\'t care what anyone else says.  this song has the hooks, the gay references, the pretty music, and the message.  15 - buttons and zips - elbow from the album \"cast of thousands\".    16 - avalanche - ryan adams from the ep \"love is hell volume 1\"   17 - political science - randy newman from the album \"the randy newman songbook volume1\"  i was introduced to randy newman from all of the soundtracks from my kids dvd\'s.  he\'s done complete soundtracks for toy story, monsters inc. and a bunch of other stuff, including a lot of songs that are used as introductions to tv shows.  he has an unmistakable voice and ragtime (i thinks it\'s called ragtime) style.  nice to listen to.  so i bought the album not realizing that he has this uncanny ability to tell a story in a song.  i love the whole album.  he covers themes such as racism, marxism, and prejudice.  i chose this track b/c it\'s so appropriate to where america is right now.  regardless of what you think about gwb, iraq, israel, etc.  the truth of the matter is that americans come across as arrogant know-it-all rednecks, and i wish it was otherwise.  18 - somewhere over the rainbow / what a wonderful world. - israel kamakawiwo\'ole from the album \"facing future\".  i guess i\'m a new-age nerd for owning anything classified as \"world music\", but after you listen to the undisputed master of the steel guitar genre you will understand.  while you are listening to this, try to imagine a 500 pound hawaiian in traditional garb with this beautiful voice.  I listened to this on repeat about 5x in a row one day.  19 -  i shall be released - jeff buckley  from the double disc \"live at the sin\'e\" deluxe edition.  fyi:  this is an old hymn, it is not his song.  does anyone else find it uncanny that jeff buckley sang about his death so often?  this song made me cry the first time i listened to it on my ipod sitting on my back porch well past midnight.  20 - redemption song - johnny cash with joe strummer  from the 5 disc special posthumous \"selections\" album.  johnny cash, joe strummer, bob marley.  i don\'t need to justify this one do i?  and it\'s such a fitting way to close a mix cd.  '),(6935,3806796,'30,June,2004','2004-06-30 00:00:00','   JANUARY 2004 Mixed by Ben Bubar    1. The Rapture: \"House Of Jealous Lovers\"  2. The Chemical Brothers: \"Get Yourself High\"  3. Jet: \"Are You Gonna Be My Girl\"  4. Outkast: \"Bowtie\"  5. Travis: \"Somewhere Else\"  6. Over The Rhine: \"Long Lost Brother\"  7. Ours: \"Red Colored Stars\"  8. Mates Of States: \"A Control Group\"  9. Bobby Byrd: \"Do\'n The Do\"  10. K-os: \"Freeze\"  11. Isaac Hayes: \"Run Fay Run\"  12. N.E.R.D.: \"Rockstar(Jason Nevins Remix)\"  13. U2 w/Johnny Cash: \"The Wanderer\"  14. Paloalto: \"Breathe In\"  15. Kinky: \"Mas\"  16. Outkast: \"Dracula\'s Wedding (Featuring Kelis)\"  17. Andy Zipf: \"Open Up\"  18. Travis: \"Mid-Life Krysis\"  19. Over The Rhine: \"Nobody Number One\"  '),(6936,3806796,'30,June,2004','2004-06-30 00:00:00','   DECEMBER 2003 Mixed by Matt Bubar  1. Cursive: \"Excerpts From Various Notes Strewn Around The Bedroom Of April Connolly, Feb.24, 1997\" 2. Mates Of State: \"Everyone Needs An Editor\" 3. Jimmy Eat World: \"No Sensitivity\" 4. Eastern Youth: \"The World Shatters and Resinates In My Ears\" 5. Death Cab For Cutie: \"We Looked Like Giants\" 6. The Darkness: \"I Believe In A Thing Called Love\" 7. Our Nation Of Thoughtlessness: \"Everyday\" 8. Cat Power: \"He War\" 9. Reggie & The Full Effect: \"Happy V-Day\" 10. Pygmy: \"Nous Vetement D\'Hiver Sont Beaux\" 11. Rilo Kiley: \"Pictures Of Success\" 12. Sage Francis: \"Crack Pipes\" 13. Cursive: \"Sierra\" 14. The Faint: \"Total Job\" 15. Les Savy Fav: \"Crawling Can Be Beautiful\" 16. A Jealousy Issue: \"Who Crucified The Chaperone?\" 17. Q & Not U: \"Recreation Myth\" 18. All State Champion: \"Asleep At The Wheel\" 19. Lisa Germano: \"From A Shell(It\'s The Buzz)\"  '),(6937,3806796,'30,June,2004','2004-06-30 00:00:00','   NOVEMBER 2003 \"KATAAH!!!\" Mixed by Asher Wood  1. The Brown Derbies: \"Karma Police\"(acapella) 2. The Flaming Lips: \"Fight Test\" 3. Jurassic Five: \"Concrete Schoolyard\" 4. Aimee Mann: \"Stupid Thing\"(acoustic) 5. Neil Finn: \"Drugs Don\'t Work/Better Be Home Soon\" 6. Kent: \"Socker\" 7. Frou Frou: \"Hear Me Out\" 8. Ryan Adams: \"Luminol\" 9. Bjork: \"It In Our Hands\" 10. Jump Little Children: \"Cathedrals\" 11. Deathcab For Cutie: \"I Was A Kaleidoscope\" 12. The Flaming Lips: \"Yoshimi Battles The Pink Robots\" 13. Postal Service: \"Such Great Heights\" 14. Venus Hum: \"Wordless May\" 15. MC Solaar: \"Caroline\" 16. Ben Folds: \"Annie Waits\" 17. Richard Ashcroft: \"Lucky Man\"(live) 18. Coldplay: \"Come On\" 19. Deathcab For Cutie: \"405\"(acoustic)  '),(6938,3806796,'30,June,2004','2004-06-30 00:00:00','   SEPTEMBER 2003 Mixed by Josh Woodyard  1. The Minutemen: \"Corona\" 2. Baad English: \"Melt With You\" 3. ELO: \"Don\'t Bring Me Down\" 4. Velvet Underground: \"Sweet Jane\" 5. The Warlocks: \"Baby Blue\" 6. The Warlocks: \"The Dope Feels Good\" 7. Oasis: \"You\'ve Got The Heart Of A Star\" 8. Catalan FC: \"Respect Is Burning\" 9. Primal Scream: \"Autobahn 66\" 10. The Streets: \"Lets Push Things Forward\" 11. RJD2: \"Good Times Roll pt.2\" 12. RJD2: \"Ghostwriter\" 13. Death In Vegas: \"Scorpio Rising\" 14. David Holmes and The Free Association: \"Pushin A Broom\" 15. David Holmes and The Free Association: \"Paper Underwear\" 16. Primal Scream: \"Star\"(Live @ The 9:30 Club with Ben) 17. Avatars Of Dub: \"3kdub\" 18. Verve: \"South Pacific\"(Live @ The 9:30 Club with Circadian...and Ben)  '),(6939,3806796,'30,June,2004','2004-06-30 00:00:00','   AUGUST 2003 Mixed by Mike Jonze  1. Talib Kweli: \"Get By\" 2. Four Tet: \"As Serious As Your Life\" 3. Prefuse 73: \"Choking You\" 4. Out Hud: \"Dad, There\'s A Little Phrase Called Too Much Information\" 5. Mirwais: \"Disco Science\" 6. Tricky: \"Anti-Matter\" 7. Manitoba: \"Crayon\" 8. Mouse On Mars: \"Rompatraille\" 9. Lali Puna: \"Bi-Pet\" 10. M83: \"Run Into Flowers\" 11. Mellowdrone: \"Fashionably Uninvited\" 12. Broken Social Scene: \"Stars & Sons\" 13. Kings Of Convenience: \"I Don\'t Know What I Can Save You From\"(Royksopp mix) 14. Telefon Tel Aviv: \"Introductory Nomenclature\" 15. Dntel: \"Umbrella\" 16. Four Tet: \"My Angel Rocks Me Back & Forth\" 17. The Notwist: \"Consequence\"  '),(6940,3806796,'30,June,2004','2004-06-30 00:00:00','   JULY 2003 Mixed by Scott Gravatt  1. Radiohead: \"2+2=5\" 2. The White Stripes: \"There\'s No Home For You Here\" 3. Cursive: \"Art Is Hard\" 4. Hot Hot Heat: \"Oh Goddamnit!\" 5. The Streets: \"Let\'s Push Things Forward\" 6. Zwan: \"Lyric\" 7. The Postal Service: \"The District Sleeps Tonight\" 8. The Postal Service: \"Sleeping In\" 9. The Flaming Lips: \"Yoshimi Battles The Pink Robots, Part 1\" 10. Bright Eyes: \"Bowl Of Oranges\" 11. Bright Eyes: \"One Foot In Front Of The Other\" 12. Andy Stochansky: \"Stutter\" 13. The New Pornographers: \"From Blown Speakers\" 14. Sondre Lerche: \"Dead Passengers\" 15. Sondre Lerche: \"Suffused With Love\" 16. Aqualung: \"Strange & Beautiful\" 17. Aqualung: \"Good Times Gonna Come\" 18. Jack Johnson: \"Gone\"-Live 19. Christopher O\'Riley: \"Fake Plastic Trees\"  '),(6941,3806796,'30,June,2004','2004-06-30 00:00:00','   JUNE 2003 Mixed by Dave Woodall  1. Talvin Singh: \"Traveller (Rid Loco Remix)\" 2. Leftfield: \"Dusted\" 3. Switchfoot: \"This Is Your Life\" 4. Blur: \"Look Inside America\" 5. Pearl Jam: \"Grievance\" 6. Seahorses: \"Love Is The Law\" 7. Charlatans UK: \"Frinck\" 8. Charlatans UK: \"Your Skies Are Mine\" 9. Elastica: \"2:1\" 10. U2: \"Stay(Far Away, So Close)\"(Live Boston, \'01) 11. South: \"Here On In\" 12. Switchfoot: \"Beautiful Letdown\" 13. Leftfield: \"Chant Of A Poor Man\" 14. Jon Ruben: \"Hindsight\" 15. South: \"...\" 16. Blur: \"Country Sad Ballad Man\" 17. Kid Loco: \"Cosmic Supernatural\" 18. Bach: \"Suite No.1 In G Major: I. Prelude  '),(6942,3806796,'30,June,2004','2004-06-30 00:00:00','   fellows of the post-fuego,  all apologies for being out of the loop.  i just got back in town from a few weeks of being out of town on a sculpture installation.  could not get online while gone.  and definitely could not get to my computer to make the CDs.  the final draft is compiled too...  tragic.  anyway, there are no excuses, i should\'ve had this done in may... but... dude but no.  again, this is apology.  also, this is announcement.  this thursday, i compile the copies and mail them out fri or sat.  for real though.  thanks for patience.  and no worries for lack of it.  you all deserve more than my slackitude.  and hopefully, my mix will deliver the goodz.    talk more soon,  asher    \"most art is revenge on the world anyway.\" -thom yorke  '),(6943,3806796,'30,June,2004','2004-06-30 00:00:00','   Jones took it up a notch and i couldn\'t be left in the dust. since i don\'t have admin privs to post this on my abcdx, can someone do it for me? thanks ps. my wife took this picture on our honeymoon. look out ben!  '),(6944,3806796,'30,June,2004','2004-06-30 00:00:00','   so remember when i said that i wanted to replace brother w/ my cousin? well now i have her address. Esther 699 poncedelon ave #319 Atlanta, GA 30308  she\'s got the skills and will have that female touch.  also, I met a good buddy of Tim russo\'s this weekend. Tim was working at a record label, loves music, the whole nine yards and he introduced me to Tyler. Tim introduced him as \"the guy i get all my music from\" so we started talking ABCD and he laughed and said he already does it, but not in a club. just one offs to good friends. so i asked him if he\'d like to join. he said he has about 2,100 songs in his ipod. he replied to my email today:  >>i checked out the afterburners cd club page. there where a lot of bands i knew of and a lot i had not. i\'d be stoked to jump into the mix and share my passion for music as well.  so what do you think? He\'d bring in some more west coast flavor.  '),(6945,3806796,'30,June,2004','2004-06-30 00:00:00','   okay maybe i spoke a little to soon. i got all hyped up on my garageband and i\'m still excited about it but i got kind of busy to keep on that project. so i\'ll probably be putting out a combo of sorts. i mean i don\'t want to be late seeing the intensity of this crowd.  '),(6946,3806796,'30,June,2004','2004-06-30 00:00:00','   Asher gets lapped by Matt. C\'mon!   What ever happened to June 1st? I dont mean to bitch but it\'s 20 days late...and counting! 2 months in a row of this shit, this is crazy!  I\'m going to go check the mailbox, and I wont be surprised when it\'s just bills and credit card applications.  Dough!  '),(6947,3806796,'30,June,2004','2004-06-30 00:00:00','   Asher + Matt, Has the June 2004 edition of Afterburners gone out in the mail yet? I just gotta get on your case, because as a founding member, I cannot encourage a spirit of complacency by sitting idly by. Last month\'s was late (by a whole month). And Matt Bubar, how\'s \"14\" coming? I\'ll be expecting yours the first week of July, too. Just cause the last couple have been late doesn\'t mean all the others get pushed back. I\'m not saying yours was going to be, by any means. Just wanted to clear that up in case there was any question. I hope I don\'t sound like a total dick. Just wanted to remind everyone. I expect the best out of each and everyone one of you. Don\'t disappoint me. This hurts me far more than it hurts you. I take this club VERY seriously. I hope that you will, too. If not, I\'ll get the dobermans on your asses.  Jonze  '),(6948,3806796,'30,June,2004','2004-06-30 00:00:00','   the cd covers stopped showing up for me. is that for everyone?  what do you think about theafterburners.net for $6.95 a year?  '),(6949,3806796,'30,June,2004','2004-06-30 00:00:00','   i just back from being on tour with a friends band of mine so i\'ve been gone for a month.   i got abcd 12 but i only recieved the bonus cd. i didn\'t get the actual cd.   did everyone else get both cd\'s?  '),(6950,3806796,'30,June,2004','2004-06-30 00:00:00','   -----Original Message----- From: scottgravatt@yahoo.com [mailto:scottgravatt@yahoo.com]  Sent: Saturday, June 05, 2004 8:43 AM To: domainoffer@reflectmedia.com Subject: Domain Offer - theafterburners.com  scott gravatt scottgravatt@yahoo.com  theafterburners.com Offer: $$1.00 Listed Price: $15000 Sender\'s IP: 66.176.93.6  is this domain for sale, and what is an acceptable offer?   -----Reply Message-----  From: Worldwide Media, Inc, [mailto:mostwanted@mostwanteddomains.com]  Sent: Saturday, June 05, 2004 2:04pm Subject: Domain Offer - theafterburners.com  This domain is priced at $3,500  Very truly yours,   Worldwide Media, Inc,   http://www.mostwanteddomains.com  '),(6951,3806796,'30,June,2004','2004-06-30 00:00:00','   Josh, great tunes man. some oldies and newbies. loving it! lovin\' those classics we listened to in highschool that i always wanted for my own. now i got! mike, SWEET job on the artwork!!! now it motivates me to add some killer visuals to the mix.  i told Ben a few months back how blown away i am at your alls depth of musical knowledge. it\'s INCREDIBLE!!! i have just been at a point to start listening and getting to know from 8 - up and i\'m loving them!!! thanks for keeping me in the loop.  '),(6952,3806796,'30,June,2004','2004-06-30 00:00:00','   Hey everyone, I hope you don\'t mind but I took the liberty of utilizing my administrative priveleges on this blog. I added artwork for all of the Afterburners discs that came with artwork. I had them all scanned in for my iTunes artwork window anyway, so I thought \"what the hell.\" I stuck them in the posts that Scott did last month and then i changed the  links in the template to go to those posts rather than the individual blogspots they were at before. I didn\'t like how you clicked on one of the \"previous volume\" links and then you were taken to a page that looked totally different from the new Afterburners Blog page. So now it\'s all uniform. Like I said, I hope nobody minded me doing that. Check it out and let me know if you had problems with it. It was just bugging me, so I had to fix it. By the way, no artwork for volumes 5 (josh) and 10 (dave).  '),(6953,3806796,'30,June,2004','2004-06-30 00:00:00','   I was just thinking...I can\'t believe I won\'t be making another Afterburners mix until November. You guys have no idea what you\'re missing out on. I\'ve been made to feel like I\'m not allowed to send you stuff unless it\'s my turn. And for fear of being ostracized (sp?), I just tuck my tail between my legs and wait.  '),(6954,3806796,'30,June,2004','2004-06-30 00:00:00','   i mailed all that stuff, everone let me know wether or not you got it so i can send another.  cheers  '),(6955,3806796,'30,June,2004','2004-06-30 00:00:00','   I am just wondering if Esther is really in or not.  She still hasn\'t posted or signed up to post on the site here.  Just wanted to know if she\'s committed or not.  I\'ll hold her\'s until I hear otherwise.  \"Esther are you there?  Just wave you hand.\"  And I\'ll also be holding Asher\'s mix untill I recieve #13.  Sorry dude, but 2 months is too long.  For real tho\'.  Other than that, ABCD-XV went out today so be on the lookout.  '),(6956,3806796,'30,June,2004','2004-06-30 00:00:00','   Asher is sending me the one cd he burned and i\'m burning/sending it for all when i receive it. i\'ll let  you all know when it goes out. peace out.  '),(6957,3806796,'30,June,2004','2004-06-30 00:00:00','   How about that fri\'n header, boys! You like? I thought so.  '),(6958,3806796,'30,June,2004','2004-06-30 00:00:00','   I thought our team blog needed a little sprucing up. What do you guys think of the new color scheme/design???  '),(6959,3806796,'30,June,2004','2004-06-30 00:00:00','   Hey Ben, Just wanted to confirm if Vol.15 will be on our doorsteps August 1st. Please tell me it\'s true! BTW, check your e-mail.  '),(6960,3806796,'30,June,2004','2004-06-30 00:00:00','   doubters, believers, naysayers, himhawers, g-force, brethren, and betrayed,  &nbsp; i am full of muchos apologies for tardiness, and a desire to explain and display all the art that has been created in the last month at the expense of the timeliness of this schaweet CD.&nbsp; but, i will save you all from my verbose stumblings and shameless self promotions, and will just gird up my loins and dish out a \"thirtEEn.\"&nbsp; technically it\'s not CD #13, but i guess the 14th floor in buildings is also technically something else...&nbsp; that begs another question:&nbsp; with 12 apostles, does that make jesus #13, or #1?&nbsp;  &nbsp; anyway, i will publish the track listing in a couple days, giving this time to mail out...&nbsp; thanks again, to peepz with patience.&nbsp; i resolve to live, with all my might, while i do live. &nbsp; more soon.&nbsp;  for real tho\', &nbsp; ash &nbsp; &nbsp;  '),(6961,3806796,'30,June,2004','2004-06-30 00:00:00','   Hey, guys. Guess what... It looks like a copy of Afterburners 15 has leaked and found its way into the filesharing networks of the world. I found it yesterday on Soulseek. Ben, I wouldn\'t trust your friends so much and I\'d keep an eye on that computer of yours! Anyway, if anyone wants an advance copy, let me know and I\'ll pull some strings.  '),(6962,3806796,'30,June,2004','2004-06-30 00:00:00','   Okay, guys. I\'ve sent both Esther and Tyler a formal introduction into ABCD (in the form of an invitation to join this ol\' blog here). Let\'s greet them warmly...  '),(6963,3806796,'30,June,2004','2004-06-30 00:00:00','   Didn\'t we decide a while back that each member is only allowed to invite one new member into the club? I invited Matt a while back and Dave didn\'t you invite someone already?  So, what I\'m say\'n is, do we all agree that it\'s ok for a member to add multiple people at once?  I\'m still not sure if this is fair.  I mean, I can think of plenty of people, that I know, that would be great for this club, but do i add them? No, because I\'ve already used my invite.  I can\'t remember 100% but didn\'t Dave use his invite on Asher? Dont mind me, I\'m just being a legalistic pain in the ass!  '),(6964,3806796,'30,June,2004','2004-06-30 00:00:00','   I\'m down with either them going in march/april or a 14.5  Tyler Ward 349 fulvia street encinitas, ca 92024 his email is: tyler@competitor.com  esther  699 poncedelon ave #319 Atlanta, GA 30308 her email is: esthergen@hotmail.com  '),(6965,3806796,'30,June,2004','2004-06-30 00:00:00','   guys remember having one person a month. to show you these people are legit i\'ll have them send out abcd 14.5.  if they can\'t do that then they can\'t join. how\'s that sound?  are we down or what?  '),(6966,3806796,'30,June,2004','2004-06-30 00:00:00','   My  urlLink prediction  came true, Asher has been lapped by Matt.  '),(6967,3806796,'30,June,2004','2004-06-30 00:00:00','   JULY 2004 Mixed by Matt Bubar  1. Le Tigre - Deceptacon 2. Modest Mouse - Float On 3. The Distillers - Drain the Blood 4. Sleeping at Last - Ghost 5. Azure Ray - New Resolution (Postal Service mix) 6. Sleater-Kinney - I Wanna Be Your Joey Ramone 7. Yeah Yeah Yeahs - Maps (acoustic) 8. Emiliana Torrini - To Be Free 9. The Sleepy Jackson - Good Dancers 10. Smashing Pumpkins - Said Sadly 11. MÃºm - Summer Make Good 12. Xiu Xiu - Crank Heart 13. The Darkness - Street Spirit (Fade Out) 14. Copeland - She Changes Your Mind 15. Modest Mouse - The View 16. Masters of the Hemisphere - Rules of Life 17. The Envy Corps - Prisoners of War 18. Mary Timony - Blood Tree 19. Desaparecidos - The Happiest Place On Earth 20. Sondre Lerche - Two Way Monologue 21. Stereolab - La Demeure  '),(6968,3806796,'30,June,2004','2004-06-30 00:00:00','   look for it in your mailboxes tuesday, july 6th.  ben and scott, i\'ll give yours to you whenever i see you.  that is all.  '),(6969,3806796,'30,June,2004','2004-06-30 00:00:00','   i don\'t really mind if people are late. I love the afterburners club but i know we all have all kinds of other neat stuff happening. so if anyone\'s late or wahtever i understand.  '),(6970,3806796,'30,June,2004','2004-06-30 00:00:00','   AUGUST 2004 Mixed by Ben Bubar    1. Jamiroquai- Canned Heat   2. Radiohead- Where Bluebirds Fly   3. Midwest Product- When Doves Cry   4. Robi Draco Rosa- Heaven Can Wait   5. The Spirit That Guides Us- Real Life Motion Picture   6. Loretta Lynn- Portland Oregon(with Jack White)   7. Lamb- Angelica   8. Beneath Autumn Sky- Winter Solstice   9. Mazzy Star- Fade Into You 10. Feist- Let It Die 11. The Bad Plus- Film 12. Michael Flume- Agolele 13. Lamb- Please 14. Robi Draco Rosa- Solitary Man 15. The Polyphonic Spree-Section 18(Everything Starts At The Seam) 16. The Polyphonic Spree-Section 19(When The Fool Becomes A King) 17. Pedro The Lion- Metal Heart 18. Wilco- Kicking Television  '),(6971,3764220,'28,June,2004','2004-06-28 00:00:00','    Yoz..Today is the first day of school for secondary and primary students wow.. Cool.. haha.. This morning I woke up at 6 and message him ask him to WAKE UP but guess what. Haha.. He wake up only at 6.15 and the rush here and there and I waited for him at the MRT station for hmm.. About 5 mins or 10?? haha.. He really very cute .. Like small boy like that.. His hairstyle really make him look like primary school kid but very cute I like.. haha.. cause like that then can show that he\'s small haha.. No la.. His character is cute so his outlook must suit his character in order to make him have the correct looks wad right.. Today got ERM.. 2 homework only.. A. Maths and Social studies.. haha.. '),(6972,3764220,'26,June,2004','2004-06-26 00:00:00','    2.00pm... and I have another 16hrs and 30 mins before I can see him... I have not seen him for about 46 hrs already... I last seen him was Friday at about 4.00pm like that when we just came back from sentosa... Saturday,which was yesterday, he could not come out as his mother was not at home and he needs to look after his siblings.. And today I could not go out as it\'s Sunday which is suppose to be my family day and we will be leaving to our grandma hose later around 2.30pm.. haiz... Missing him like hell man.. What is he doing now..?? Sleeping?? Playing games?? He would only be doing these 2 things as he is lazy to do others.. haha.. This is his habit man... Just can\'t imagine if really we 2 are going to last till I go overseas to Australia to study.. What is life going to be man.. I\'m not going to live like that.. How can I keep in contact with him.. How am I going to see him.. haiz.. That really makes me sad.. haiz.. But Just not to think about it man right??.. haha... '),(6973,3764220,'01,July,2004','2004-07-01 00:00:00','    Haiz.. skipped 2 days to write my jornal.. haha.. haiz.. although it\'s onl 2 days but guess wad?? Lots of things happen in this 2 days..  29/6 : My parents found out that i have a boyfriend outside.. That night, i talk to my mum and she told me that my boyfriend\'s mother actually called her up to ask if she knows about it and even tell her lots of things.. My parents were very furious therefore rush out to my school to fetch me back from my volleyball training.. i was scolded all the way even on the car or after alighting.. Their mouth was like wa machine gun man.. can\'t stop..That night my handphone was confiscated and return he next morning.. But one thing was that my dad would bring me to school and after my volleyball training..  30/6 : Yesterday, when i reached school it was like wa very scared that i would be late as the sky was like bright and i\'m still at home.. My first period was PE and that my mum says that i called her but i didn\'t.. I was having PE for the first 5 periods as today was the health fair which we were practicing for so long juz for today.. During the first 2 period, he was also having PE and it was like a terrible thing as the day before which was the day we were found out that we were caught. His face and mine were like charcoal black.. We were unhappy at all.. I went up to the hall to find him get him out and talk to him and yet we finally came to a conclusion and he has been following it ever since today.. cool isn\'t it..  1/7 :  Today, haha fun day man.. only study for 3 periods of POA and tt\'s all... fun isn\'t it... ahha... love it man... Hope everyday like today... '),(6974,3764220,'06,August,2004','2004-08-06 00:00:00','    Got about a month never come in and write le.. Now?? guess where am i?? haha at the woodlands library the Reference area.. Sianz.... In here for about 3 hours... Then in this months.. a lot of things happened... july 1st was the last day i write this \"dairy\" Guess where i write during this past 1 month.. In my student handbook.. haha... Among these 4 weeks.. Guess which incident i never forget?? Happen very recent... Real recent until cannot forget.. The day where by my stead and my mum chat.. and last tuesday when i and my stead have a nice chat.. hor dear.. haha.. Guess u nv read this ba.. so nvm lor juz put for nice lor.. haha.. okay.. tt\'s all... '),(6975,3950869,'29,July,2004','2004-07-29 00:00:00','   I got back Sunday afternoon.  It was a fun wk for me at least.  Not too sure if my brother and sister had as much fun as I do, but they normally don\'t like hanging out with the family up north.   Monday I put my two wks in at Wendy\'s.  Yup-next wk is my last wk of being employed.  Two days after I came back home Ashley decides to be stupid-runs off again.  Of course the consequences get worse.  Now she has one of those tracking devices that are worn on the ankle. Her life is on a downward spiral.  But last night we found her and my mom dropped her off at the police station and they had court today(that\'s when she got the anklet thingy).    Today was a pretty good day for me.  Nothing too extraordinary.  Woke up at 8:00(which is a suprise, especially with no alarm), took my sweet time to get ready for the day, then went up to my old alma marter(sp?).  I saw some of my old band buddies-it was nice.  Then Kelly and I went to La Fiesta to eat.  After all that I went to Biloxi.    The doctors scared the crap out of me.  They made me think that I was going to have to get two or three shots. HA ha.  Tomorrow night my mom and me are going out shopping for dorm stuff. Yee haw. '),(6976,3950869,'23,July,2004','2004-07-23 00:00:00','   I got back Sunday afternoon.  It was a fun wk for me at least.  Not too sure if my brother and sister had as much fun as I do, but they normally don\'t like hanging out with the family up north.   Monday I put my two wks in at Wendy\'s.  Yup-next wk is my last wk of being employed.  Two days after I came back home Ashley decides to be stupid-runs off again.  Of course the consequences get worse.  Now she has one of those tracking devices that are worn on the ankle. Her life is on a downward spiral.  But last night we found her and my mom dropped her off at the police station and they had court today(that\'s when she got the anklet thingy).    Today was a pretty good day for me.  Nothing too extraordinary.  Woke up at 8:00(which is a suprise, especially with no alarm), took my sweet time to get ready for the day, then went up to my old alma marter(sp?).  I saw some of my old band buddies-it was nice.  Then Kelly and I went to La Fiesta to eat.  After all that I went to Biloxi.    The doctors scared the crap out of me.  They made me think that I was going to have to get two or three shots. HA ha.  Tomorrow night my mom and me are going out shopping for dorm stuff. Yee haw. '),(6977,3950869,'23,July,2004','2004-07-23 00:00:00','   Serendipity means a fortunate mistake or accident...Today while in New York City we went by there(inspired by watching the movie \"Serendipity\").&nbsp;It\'s an ice cream shop/restaurant. &nbsp;Although we didn\'t get to eat there b/c the wait was too long.&nbsp; So I took&nbsp;a picture of the sign.&nbsp; But let me tell you the beginning of the story...ha&nbsp; WE GOT LOST ON THE SUBWAY..thanks to my lovely sister.&nbsp; So we were there we were going the total opposite direction in which we were suppose to be going for about an hour and a half.&nbsp;&nbsp;&nbsp; Once we got off it was raining-ack.&nbsp; Rain, Umbrellas, alot of people, and New York City do NOT MIX WELL.&nbsp; So after this whole experience you can imagine how cranky and irritated I was.&nbsp;&nbsp;    &nbsp;  Yesterday we went to Six Flags Great Adventure.&nbsp; It was fun- I rode most of the roller coasters and convinced my sister to do so as well.&nbsp; So now she loves them too.&nbsp; Tomorrow is the reunion.&nbsp; I\'m not sure how I feel about it.&nbsp; Hopefully it won\'t rain.&nbsp; All I know is that there are a few people coming from Ireland, France, and Austrailia so I\'m excited about meeting them.&nbsp; Not much left to say...ciao '),(6978,3950869,'21,July,2004','2004-07-21 00:00:00','   Yesterday we left Penntucky and went to my Nana\'s house.&nbsp; My little cousin Sean was there-I felt like a little kid again...I played with bubbles(tried to eat one or two), put together a spider man puzzle, and made little figures out of playdough.&nbsp; Later on that night we went out to the boardwalk.&nbsp; Sean and I went on this upside down ride-which was great!&nbsp; Then he and I went rock climbing.&nbsp; It was fun.&nbsp; The houses that I saw though were gorgeous.&nbsp;&nbsp; Yea-and they only run for a million or more dollars.&nbsp; I love it up here-I\'ve always loved the north.&nbsp; Today we planned on going to the beach, but instead we\'ll be going out to the lake.&nbsp; That\'s fine with me I guess.&nbsp;  byee '),(6979,3950869,'19,July,2004','2004-07-19 00:00:00','   It was nice to be able to get to sleep in today(got up about 11:30).&nbsp; A couple hours later we left to go out to eat at Fuddpuckers-then to the park for Christopher.&nbsp; But all of us \"grown-ups\" played as well.&nbsp; Man I miss being a kid.&nbsp; Even though I still kinda am.&nbsp; Then we went to downtown Bethlehem, PA.&nbsp; I liked the environment of everything.&nbsp; After all of that I was ready for nap.&nbsp; Then my cousin Katie came over to Beth\'s for dinner.&nbsp; Yummy-it was baked Ziti(spelling?).&nbsp; Very cheesy.&nbsp; Right now I\'m listening to the Gin Blossoms, I was actually suppose to go to their concert, but my sister waited like an idiot to get the tickets and then they were sold out.&nbsp;  &nbsp; Tomorrow, we\'re suppose to be heading out to my Nana\'s-back in New Jersey. audios '),(6980,3950869,'18,July,2004','2004-07-18 00:00:00','   This morning my brother woke me up, singing.&nbsp; Yea at like 9 in the morning, I wasn\'t too happy.&nbsp; Last night, I caught fire flies for the first time ever.&nbsp; It was cool, I\'ve never really seen them up close.&nbsp; One lit up in my hand-i love the little things.&nbsp;&nbsp;Later we met up in this little diner with my Aunt Kathy, Conor, and Ryan.&nbsp;&nbsp; Then we went bowling.&nbsp; The first time we played I won with a 98!&nbsp; Then my sister got a little competitive and then she won by one point.&nbsp; I had a score of 102(hey I guess I should mention we had the little gutter liners :) ) ha ha.&nbsp;  &nbsp;  Later on this week we plan on going into the city, Six Flags, The boardwalk, and then of course the reason why we came- the family reunion...i wonder if anybody will be able to guess who I am.&nbsp; I\'m at Beth\'s house right now, earlier Uncle Tom, Darren and Clare, and Uncle John were here.&nbsp; This is going to be short-  Au revoir. '),(6981,3950869,'17,July,2004','2004-07-17 00:00:00','   I didn\'t get to sleep until about 3:00 in the morning and then woke up at 7:30.&nbsp; Miki worried me because I thought we were going to be late for our flight.&nbsp; But of course, we got there way before time.&nbsp; There was a mixup with the tickets Miki and Christopher had, so their luggage was \"temporarily\" lost.&nbsp; Hahaha.&nbsp; After we got picked up by Uncle Tom and Aunt Kathy we went on over to my Aunt Ann Marie\'s.&nbsp;&nbsp;&nbsp;It was nice-I saw my Uncle Frank and Aunt Eileen as well.&nbsp; Then we had this nice little barbecue...tasty, I think I can officially say, I like shrimp! Tomorrow we\'ll be going over to my cousin Beth\'s house and stay there for a day or two.&nbsp; I think we\'re also going to the boardwalk.&nbsp; That\'s all I have to say for now, I feel like I\'m ignoring my hostsssss.  &nbsp; Goodnight YouS Guys. '),(6982,3950869,'17,July,2004','2004-07-17 00:00:00','   Hmm...I already have a journal type thing online already on xanga.&nbsp; But I feel like this one is going to be different in a way.&nbsp; Tomorrow I\'ll be leaving for New York and Jersey.&nbsp; I love it up there-the feeling is amazing.&nbsp; So different from the coast...well at least to me that is.&nbsp; Not so sure I could live in a big city like that though.  It\'s 11:07 right now and Christopher, my nephew, is on the couch asleep-and here I am typing for myself, sipping fruit juice, and listening to Journey\'s faithfully.&nbsp; Music-I love it.&nbsp; All varieties, whatever catches my ear, whatever catches my emotions...country, alternative, soft rock, latin, jazz, the genres go on and on.&nbsp; Ha-it\'s 11:11 right now...reminds me of Puja in class.&nbsp; She never missed making a wish on that thing.&nbsp; Occasionally I would wish, just for the fun of it too.&nbsp; It\'s exciting about going off to college, but I still have longing for the past.&nbsp; I kinda want to go back one more year-not possible I know. &nbsp; It\'s a mystery. But they still try to see  Why something good could hurt so bad... One love feeds the fire. one heart burns desire. Wonder whose crying now? ----Journey. &nbsp; Night. '),(6983,3950869,'03,August,2004','2004-08-03 00:00:00','   You build me up You knock me down Provoke a smile And make me frown You are the queen of runaround You know it\'s true You chew me up And spit me out Enjoy the taste I leave in your mouth You look at me I look at you Neither of us know what to do There may not Be another way to your heart So I guess I\'d better find a new way in I shiver when I hear your name Think about you but it\'s not the same I won\'t be satisfied \'til I\'m under your skin Immobilized by the thought of you Paralyzed by the sight of you Hypnotized by the words you say. Not true but I believe them anyway'),(6984,3321866,'29,May,2004','2004-05-29 00:00:00','    So yesterday was fun and such and I went to James\'s house right after school and he taught Steph and (somewhat me) the guitar. Me being the impatient one didn\'t quite listen so I just randomly played. Then we went to Claytons and had fun there and ate and oddly we all ordered the same thing...The waitress was great cause she put up with Chris\'s antics and we had fun. We walked Steph home and then found a shopping cart and tried to lewer CJ away from Emo ball. We failed so we went to video land and rented  \"Better Off Dead\"  and that was fun and Chris and James meet my grandparents and saw my little brother amuse them and my room which is cool I guess and Steph came and was scared by Chris and James. That was fun. Today my mom says \"your shoes are too small\" so we bought some Converse at the exchange that are navy blue and are cool and I also got some cool shirts that my dad owned when he was young and they look like something Gordo would wear so there cool. I\'m going to wear them on Tuesday :). This is my life....Cool. '),(6985,3321866,'27,May,2004','2004-05-27 00:00:00','    Spanish project due and my hair is cut and i got an extra day cause i said it didnt work but it did  :)  http://www.xanga.com/home.aspx?user=Executive_IMP  look at it on my xanga '),(6986,3321866,'26,May,2004','2004-05-26 00:00:00','    Spanish project due tomorrow and all needed is editing and woo hoo its pritty much boaring but ok. All ours is is a birthday at a resturant and a commercial of Albertsons. Unlike Collins which is a super good awsome gun fight matrix pariody movie thats cool ours is simple and...simple. All the editing realy needs is a slide show tibit of Albertsons stuff. I really want to see the finishing product on Collins though :)  On another note yester day was my moms birthday and also (drum roll please) my half birthday!!! That means i can get my permit (Technically) and i just realized that when Ryan Jones said \"Dont you know you can get your permit now?\" and then i write. The first thing that comes to mind is \"Road Rage\" woo hoo. At our neighbors house we went to eat and for my moms B-Day and we started a conversation about road rage and the whole time i was listening i thought of \"Road Rage Brad\". I heard stories of Mr. Couture and his road rage and then the conversation stoped due to the stupidity of Kenny (8th grade friend) who climed well tried to over a turned over coucn on its end and it fell and he almost hit tons of Priceless pottery and Mr Couture said \"Geise Kenny, why dont you just take a bat and hit the pottery!\" It was great! Then we all realized that it was 10:30 so we left.  Im writing a story of Link and his adventures just for fun and its pretty good so far even though its one paage long...its good  adios....i think '),(6987,3321866,'24,May,2004','2004-05-24 00:00:00','    Well \"inner bitch\" was Hilary and she was talking to me on the phone well now I\'m typing with the monitor off monitor off and short term memory are bad together and this post may seem weird (not wyrd like fate ) and lala in world history with Worthington and she has no clue and Hilary says she is sorry to all and I forgive Jon and I tell him not to do that again and listen to my advice sometimes and don\'t kill the people who help you cause they will kill back and spam :) Ryan Jones who is sitting next to me is sleeping while I\'m on this computer typing and did I mention Hillary is sorry reading Hilary is sorry a lot and is or was sad and ya you catch my drift.....6 minutes left so bye for now ill fix it later if you care..... '),(6988,3321866,'22,May,2004','2004-05-22 00:00:00','    my grandparents are here and moo im stuck '),(6989,3321866,'19,May,2004','2004-05-19 00:00:00','    9.27022% - Geekish Tendencies  Spring fair fun and my toe bandage is gone!!!!  =) '),(6990,3321866,'18,May,2004','2004-05-18 00:00:00','    Which Neglected Mario Character Are You?     http://www.smbhq.com/nc/nctest.html  Toad is like in every game woot woot '),(6991,3321866,'17,May,2004','2004-05-17 00:00:00','    readminification succeeded. '),(6992,3321866,'16,May,2004','2004-05-16 00:00:00','    Jon is being a blog tyrant and I don\'t want him as an admin cause WELL HE ASKS ME IF HE CAN BE ONE AFTER ONE COMMENT AND COLLIN ON THE OTHER HAND HAS COMMENTED ON EVERYTHING SO HE IS ONE AND NO JON YOU DIDN\'T ASK WELL ENOUGH AND DON\'T YOU KNOW THAT COLLIN CAN DO THAT AND YOUR TO BUSY TO DO THAT CAUSE YOUR CAUGHT UP IN YOUR COSTUME THING AND FOR THE PEOPLE WHO ARE LISTENING AND NOT REALLY SURE WHAT\'S GOING ON WELL JON IS RIGHT BEHIND ME AND DURING THIS WHOLE WRITING I ACCIDENTALLY PRESSED CAPS LOCK AND IM TO LAZY TO TYPE AGAIN. Ha um so no hard feelings Jon but one helper is enough and if you commented more than just that once then you would have had a fighting chance and such so.....School is dumb. Hunger pains.  Bye. '),(6993,3321866,'16,May,2004','2004-05-16 00:00:00','    Collin is in the administrator status cause he\'s smart with this stuff and is the only one who\'s smart with this stuff who\'s posted.  If there are any complaints well.....TO BAD!!!  and folks...lookward to the near future in video games for   Metroid Prime 2 Echos   Nintendo DS   The Legend Of Zelda Minish Cap  and  The Legend Of Zelda  unknown name '),(6994,3321866,'14,May,2004','2004-05-14 00:00:00','    my sphere of influence grows daily '),(6995,3321866,'14,May,2004','2004-05-14 00:00:00','    This is the future and collin already put this up but its so awesome that I\'m putting it up.  http://games.slashdot.org/games/04/05/11/217221.shtml?tid=127&tid=186&tid=207&tid=213  I also realized wait Imp there\'s a spell check!!! My bad spelling days are done on posts and not on comments so.....Moo.  http://www.nintendo.com/gamemini?gameid=54610f14-1826-4d46-9981-8f72874aee2e&rp=1 '),(6996,3321866,'14,May,2004','2004-05-14 00:00:00','    hahahahahaha steph and the worm was fun and on tape and phils uhh wrong turn slap and smack of head to ground was great and fun and everyone at Rep/Java night was Great. '),(6997,3321866,'14,May,2004','2004-05-14 00:00:00','    lalalalalalalala school is mean '),(6998,3321866,'13,May,2004','2004-05-13 00:00:00','    Well I was thinking about using this so I am now and I put the pick of link and I made a different one but it wont go on so and I made my IM thing as:   I_Have_No_IM_Cause_It_Makes_My_Computer_Mad  and the blog so screwed it self up so I took it off. For those of you who don\'t know that really isn\'t my IM and it really is RynImp but as I said it makes my computer mad well it does. Add some comments on how to make links if anyone who knows how knows how who views this por favor.(for those who are Spanish stupid out there POR FAVOR means PLEASE) '),(6999,3321866,'12,May,2004','2004-05-12 00:00:00','    I did this cause you need one to post to others so yah don\'t count on any updates<3Kristen Emata wrote the heart cause were in health in the library '),(7000,3321866,'29,June,2004','2004-06-29 00:00:00','    Isaac will be hard! But with help it will be done!  On...Star Trek day we should all plan out stuff and for the people who don\'t come we\'ll link them in some how. If people want to know when the comic-con is I\'ll be posting (in MM style) when it is in every post if you haven\'t noticed. Tomorrow we should do some planning to and stuff so someone call me and stuff...I\'m of to look for pictures of Isaac to get a costume idea for its design so...*POOF* '),(7001,3321866,'28,June,2004','2004-06-28 00:00:00','    Do The Poll!  I need help with my costume and picking which person to be and we should all go on one day and do costumes and schedule stuff! '),(7002,3321866,'26,June,2004','2004-06-26 00:00:00','    Comic-Con (Majora Mask style)  Choice A: 1up shirt Choice B: Luigi/Mario Choice C: Shadow Link Choice D: Open to suggestions cause I need them DESPARETLY!!!!!!!!!! '),(7003,3321866,'22,June,2004','2004-06-22 00:00:00','    Woo my form came in the mail thing and I\'m set for the convention! I missed the water fight due to my working at Camp Coronado and Hilary calling me there. Tomorrow I\'m not going cause its summer school and at five my job so...COOL! Steph if you read this then I need you to get me my shirts back to me I left at your house! I\'ve already been told to bring it twice so...Blaah. Poor Fado and Laruto  :*( tear to my eye about there fate. The times this summer I\'m free are every Wednesday this summer after 10:00am till 4:30pm and anything after 4:30pm on any other day and fully free on weekends! Comic-Con! I don\'t know who or what to be so I\'m open to suggestions from all...Jon make it sane ;) '),(7004,3321866,'16,June,2004','2004-06-16 00:00:00','    Four Swords during finals rocks cause it does and shadow battle relieves anger! '),(7005,3321866,'14,June,2004','2004-06-14 00:00:00','    Help it\'s being mean!(That means you Collin!)   Edit:  I win, as always '),(7006,3321866,'13,June,2004','2004-06-13 00:00:00','    Worst sequel ever! 1) Sceens were changed that weren\'t suppose to. 2) Main Characters were changed that aren\'t suppose to. 3) the story was skeweredd and put back to geather wrong.  Dementors were cool though! '),(7007,3321866,'10,June,2004','2004-06-10 00:00:00','    As on the last post  Zelda Four Swords Adventures  came and Jeremy got it and...It\'s one of those games...Were...It can\'t be bad. This game is awesome and can\'t go wrong! Force gems should  be ruppies but they aren\'t money so maybe they shouldn\'t. I always am the HERO OF DARKNESS and its kind of cool that I won a game with that title once. Tingles tower has a new meaning to all you Wind Waker fans out there and its this awesome mini game place. So far we have 2 out of 12 I think and the 2 we have are still really fun and they\'re suppose to get even better along the way! Hammer Tag as James saw on the web site is the next one and just the name makes it good. Were stuck on Death Mountain level and I think Collin Nik Jeremy and I all hate the bomos medallion part cause \"I was trying to protect him!\";) part is pure evil and that\'s only the first level of 3. More coverage on game in future.                                             urlLink '),(7008,3321866,'07,June,2004','2004-06-07 00:00:00','    The Legend of Zelda Four Swords Adventures   Jeremy,Collin,Nick,and I are(after Jeremy gets it)going to be parteners in this game which is wicked awsome and im the blue Link and james there is no orange tunic  There is a Green ,Blue ,Red ,Purple ,Black ,White ,Gold or Yellow ,and the Great Denti Mask tunic.  they should so change the purple one to orange though.  SHADOW LINK IS GOIN DOWN!!!  I need Help loading the pictures up cause there are like 40 of them :O '),(7009,3321866,'01,June,2004','2004-06-01 00:00:00','    This is the totals of all the stuff I got or is happening in the future due to what I did on the weekend:   SHIRTS: 5   SHOES: 2   HOURS PLAYING VIDEOGAMES: 24   MONEY AND FUTURE MONEY: $66+$8 an hour   TIME SPENT DOING HOMEWORK: .......right homework   in that I included my job at the yacht club and the shirt I got from Hilary which is awesome and I love you a lot for that Hilary :D  I\'m a Samurai!!!  WOO HOO FUN.....IM LOST!!! '),(7010,3321866,'31,July,2004','2004-07-31 00:00:00','    I hate beating Zelda games cause they end...Now I have to wait for the next two... '),(7011,3321866,'26,July,2004','2004-07-26 00:00:00','    Comic-Con was fun...   Things bought:     I Stole The Triforce Shirt  Don\'t Make Me Go Zelda On You Shirt  Link and Epona Figure  Transformers Comic #1  Vending Machine Grounds Keeper Willie   Food   Fun... &nbsp; '),(7012,3321866,'19,July,2004','2004-07-19 00:00:00','    Night&nbsp;of The Second Day      &nbsp;&nbsp;&nbsp; 48 Hours Remain        &nbsp;    play some Majoras Mask and you\'ll get it Jon...Can\'t beat evil archery '),(7013,3321866,'18,July,2004','2004-07-18 00:00:00','    Night of The First Day      &nbsp;&nbsp; 60 Hours Remain    &nbsp; '),(7014,3321866,'10,July,2004','2004-07-10 00:00:00','    Dawn of The First&nbsp;Day    &nbsp;&nbsp;&nbsp; 72 Hours Remain       &nbsp;   Been playing\' Majora\'s Mask...Wow that game is really really  really  short... '),(7015,3321866,'10,July,2004','2004-07-10 00:00:00','    Four Swords is Over... urlLink The Knights Of Hyrule  Will always stay alive in the Green, Red, Blue, and Purple...:*(    Night Of The 11th Day 241 Hours Remain '),(7016,3321866,'06,July,2004','2004-07-06 00:00:00','    I\'M GOING TO BUY A SWORD!!!  Comment on how much money you\'ll are bringing to the comic-con.  I will probably bring 100-150 dollars and...Comment on how much you\'ll are gonna bring.   NIGHT OF THE 15TH DAY (almost)343 HOURS REMAIN '),(7017,3321866,'04,July,2004','2004-07-04 00:00:00','    THESE NEXT FEW SENTENCES PAST THIS LINE ARE RANDOM!!!  ******************************************LINE***************************************  Crime Scene  Man lying face down with blood trailing out of back. Two cops look over the body to  find the cause of death. The Chief comes by to find the results.  CHIEF: Did you find the cause of death?  Cops look up at chief.  COP: We believe it\'s the knife in his back sir.  ******************************************LINE***************************************  END OF RANDOMNESS!!!  WOOHOO! Staying\' up is fun and watching  [Adult Swim]  is cool cause I get to see show like  Wolves Rain  and  Cowboy Bebop   and after watching those shows I realize that the DVD\'s of Bebop will be at Comic-con and that\'s a great and funny and cool show. Silent Hill is Fuckin\' scary to watch or play! Chris, Alec, and I played it at his house after a game of Mario Party 1 and *shivers Pictures image of a bloody basket ball hoop with a dogs head underneath.* Awesomely Scary! As the title says woke up at 2:00 P.M. so I\'m up...Still and board. Tomorrow I guess I\'ll roam until I find people. If I were  Mario  for Comic-Con.                               Still need for costume:                                         1)Overalls                                         2)Gloves                                         3)2 Big Yellow Pins                                         4)Brown Boots                                         5)Hat                                         6)Mustache                                         7)Red Shirt                                         8)Help  I might as well not be any thing since its  Day 17 405 Hours Remain  (and Collin, In Majoras Mask this message comes up at 6am and pm of each change where  it becomes night and day but my systems different)If I were to be Isaac I would need a lot of help from Collin and people who have seen him in a game cause his looks hard to do. I might just be picture guy.   Do the new poll.    ALSO  Check this picture of  urlLink Isaac . Isaac now that I see looks a whole lot easier than I thought so I might do both so on Parade Day I will consult wit my people. Remember these words I tell you.    urlLink Blades Will Bleed. Shields Will Shatter. But As The Light Fades...Will The Hero Rise Again? Or Will Darkness Reign?   Joy and Rapture:D '),(7018,3321866,'02,July,2004','2004-07-02 00:00:00','    Spiderman 2...Poor Doc...  anywho  Maria ,Sabrina ,Arial ,Collin ,Philip and I say it and it was great and I forgot to buy online so I...\"walked right in\". Sadly people these days are losing there math skills ;) I give you 5 you let past 6! $11 in food I think will cover it.   Night of the 18th day 433 Hours remain '),(7019,3785845,'30,June,2004','2004-06-30 00:00:00','   Just got bac and mann was it hot out!! wooo! cant wait to go to cottage and swim now...thought i was going today but IM NOT ...but tomrrow i am...ahh wonderland was hell with the heat i was wearing my light blue jeans and blue top...lines were hell...than i waited for my mom for 45mins! In those 45mins there was a old dude parked at the drop off sittin on the hill takin pix and proably videos of random people...sickoz these days! Than some random car comes in front of me and the ppl in it start making out...was lonly and hottt soo very hotttt! well im off to take a nap ciaoz! '),(7020,3785845,'29,June,2004','2004-06-29 00:00:00','   yup i knew it was going to be a bored day but not that its really late it doesnt matter because too hyper to think about that lala yumm cereal honey nutt yummmm gooddd well bac to boredom...lalala wooo phil has kool squarie things on his jounral! exticin!! yayy well im goign to go have fun with that now..wait i say we all whip the honey bee\'s ass just because he\'s a stupid smiling bug! he will die some day... '),(7021,3785845,'28,June,2004','2004-06-28 00:00:00','   ughh i am soo annoyed with this 1 specical person...she came to my school only a year and a half after i left she\'s a nice person and all but she\'s ebrassin to be around...becasue she came she ruined our friendship all i do is bitch at her when she\'s around. I bet if she still was at her old school i wouldnt be so mad cuz i wouldnt have to see her every fuckin day...she\'s would always be around me and my friends at lunch in class EVERYWHERE never with any1 else becasue she has annoied every1 to the point where u just want to kill her. THank god its summer :D:D but still she can still connect me with the phone..ahhh well i can always hope she\'ll find friends and give me space!!! ughh well im going to try and not be pissed off even other she has only been at my school for half a year!!! and i already hope she leaves (which is unlikly) or dies or ahhhh just soo mad and AHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHH  ok im leavin before i say stuff i\'ll regret!! laterz! '),(7022,3785845,'28,June,2004','2004-06-28 00:00:00','   Once again my day is ruined because of the weather...now i get to stay home and be lazy..woo in pjs all day :D Plannin to find SOMTHING to eat eventaully...havnt left my room yet *yawns* Need some music to wake me up...ughh what a boring day ahead! '),(7023,3785845,'28,June,2004','2004-06-28 00:00:00','   Today i got bac from my cottage which wasnt too exticin just of the crappy weather!! grr Went to see \"white chicks\" with katie one of the millions of cousins i have lol it was pretty funny once it got started! Another funny part to my day was Katie runnin into her ex at the theater and her complainin about his hair and how he looks less hot when he wasnt in the first place haha...  I sware she should be a stylist :P After the movies i heard some random bangin from none other than shannon and nikki who happen to have been at the movies too! We say hi than wen to tim horton\'s. Got some food waitied for katie\'s mom (nothing exticin there)...just got in car when her mom came...lmao i was TRYING to get this peachjuice can into the holder but noo it didnt want to go!! stupid mr can!! and apparetly katie\'s grandma named her dish washer..Hazel and she took it really seariously was a bit werid but what do u expect my crazy italian family!! woo wrote alot there feel like typing *type type* lmao aww almost 2 and philie had to go *sigh* probaly go to bed soon after i check out more of this journal thing! '),(7024,3785845,'28,June,2004','2004-06-28 00:00:00','   Well this is my first posting on this journal thingie...dunno where to beingin wish i made this sooner had things to worry and talk about but now its all dead and gone with...*sigh* not going anywhere not doing anything nope nothing exticing not now anyways doesnt seem like anything will but mayb me luck will change? pshhtt right *rolls eyes* well im going bac to my music soo laterz! '),(7025,3785845,'19,July,2004','2004-07-19 00:00:00','   Do u ever wonder who u are? Sometimes i do...Who is megan d\'angelo? I feel like i can change at anytime...thats not how it should be...so i sit here confused if i can change so easily than who am i really? no one has answers so i will continue to think in this light purple flowered box room untill i find myself...dunno when or how but i will...just have to...i think this unknowing is the key to a few things...like why i havnt had a boyfriend...if i dunno who i am i have nothing..nothing... '),(7026,3785845,'15,July,2004','2004-07-15 00:00:00','   Wow long time since ive wrote in this! Nothing has happened or changed...my life is boringggg. I ment go to warped tour with channie, sammie and her bf and later meet up with  laruen and kate. ughh only if my mom lets me go actully i wouldnt have to tell her i got a ride there and bac just say at wonderland or somthin...but still gotta make 48 bucks in the next few days. Dam i wish i had a job...*sigh* mayb next summer...Well i\'m off to listen to some musio :D laterrrzzz '),(7027,3785845,'07,July,2004','2004-07-07 00:00:00','   Tomrrow Me and Katie are suppose to meet with paul, johnny and 2 random girls they know at wonderland :S. I think its going to be just a bit werid since every1 thinks paul is an idiot when me and katie both think johnny is hot lol. Watch it be like a soap opera! haha Mayb it wont be so bad johnny is a funny kid and katie gets along with every1. I just go with the flow. Well I hope this works out since i been just bitchin at paul the last few days cuz he was being stupid lol and i havnt talked to johnny in a while. Jeeez i got to stop thinkin so much... '),(7028,3785845,'05,July,2004','2004-07-05 00:00:00','   Woke up at 10, had an hour to get ready for wonderland..ended up wearing plain blue top and my black pants with my little sis\'s  shoes because mine were at left at my cottage ooppss:P Me and Channie walked around for about an hour, went on dragonfire 5 times in a ow than found nadine and heather. Nadine\'s hair looks soo nice now:)(black with green tips) Me and Chantel were countin all niggers lol and came up with 156. \"THERE TAKING OVER THE WORLD\" channie says :P We\'re planin to put them in a box. HAHA Chantel asked the guy workin there if he was tryin to posionin heather. and he said it was for her hahaah he also said his name was billy when he was like packi soo no one beveiled him. Ughh i felt so left out today  hangin out with 3 bis. One wantin the other two. Channie did talk to me but she was like just hittin me and being mean most of the time lol but hey u get use to it i guess....ughh this is making me think of other depressing crap i really dont want to think about im out!   -hating my life right now- '),(7029,3785845,'04,July,2004','2004-07-04 00:00:00','   Why do people have to be so fuckin anooyin? Ugh paul was complaining about not havin a gf AGAIN! If he really wants one that fuckin bad why doesnt he ask some1 \"dont want to be rejected\" they get to know some first! If he doesnt meet people or ask anybody out what u think is going to happenn? what is happenin now for him NOTHING!!!! I asked if he wanted me to help him find one this week \"no point\" How is there no point!! How else u going to meet some1 is u dont get out there! I gave him advice and all he could say was \"i know\" \"lol\" \"yup\" He doesnt deserve a girl if he gets one. He just complains and bitches like a girl on pms i sware! UGHHH There is only so much i can do to help some1 and people dont appreciate me going out of my way to help them. People like that can fuckin die! GRRRR Alrighty well laterz '),(7030,3785845,'04,July,2004','2004-07-04 00:00:00','   Well, i\'m just listenin to the radio, hangin out  in my sexy white pants :) Love them hehe. Got nice outfit on but no where to go, which ya sux! Get to wear it to wonderland tomrrow if the weather works for me which would be nice :) *hint hint* ahhh got to blow dry now CIAOZ! '),(7031,3785845,'04,July,2004','2004-07-04 00:00:00','   ahhh bored out of my mind!!! I really want to go home and try out my new cell :D:D Finally got one and i cant even use it! Hopefully be home monday...tuesday at the latest. Also Dad fixed my webcam thing finally got to use it. Great new stuff and cant even use it! Its torture i say! Omg haha a fire was set by this cottage near by, My mom thinks it was the kids in the forest on their dirt bikes. And she\'s now all worried they\'ll come back and set a fire at night and we\'ll all die. Well she went crazy in her thoughts i just sat and listened to my music hehe. Well I think chicken is done so write ya later -Boredom......wanting, needing, cant...- '),(7032,3478510,'01,June,2004','2004-06-01 00:00:00','    hhehhehehe sdasdasass   '),(7033,3478510,'29,June,2004','2004-06-29 00:00:00','   well..whats had gotten into me?..for no reason i was quite frustrated by the surrounding atmosphere...some of the chinese  O level`s answer was revealed...and i could only get 25/35..perphap i was frustrated over this matter...because my aim was to get a2 or a1 and i putted in a lot of effort and my expectation for this paper was high...and yet 25/35 pull down my moral..sigh....everything could onli depend on my listening comprehension and oral examination....                      frustrated today!!!! '),(7034,3478510,'27,June,2004','2004-06-27 00:00:00','   hmmm today woke up at 8.20a.m for a math tuiton as usual...the lesson today was boring and i couldnt pay attention and find it boring..but time passes fast lesson end...  well today didnt touch much on my homework...and i didnt know how i spend my day today... time passes fast...and i found out that there were many things which i have not accomplish....and i was pretty concern about that....actually before school holidays,i did planned to study in the holidays,but everything was undone... i neglected my studies....i hope that school opening would be a warning to me that o`level was around the corner..  i finally tidy up my computer room....it was messy since the school closed and the messy was left unattended for months...there was a lot of dust recollected around the desk since last time..well....it was goin to midnight now....i`am going to bath and have a early rest for tomorrow school....thats is!!!!!!!!!! '),(7035,3478510,'26,June,2004','2004-06-26 00:00:00','   ...todae not a tiring day throu...today woke up at 12 due to last night catching the match of england vs portuagal...is was disappointing as i thought owen score in the first goal , it would secure the whole match..but who knows portuagal score and it went to extra time....extra time both score and it led to penalty...beckham took the firsr penalty but who knows that he miss it and england lose...i was so disappointed as he miss two penalty in the cup....but everything since fine after i woke up= )....well today i was not satistfy with my performance as i never accomplish the task of finishing my a math paper=(...but i hope that by tomorrow i can finish all the a math paper...and i`am getting a hair cut too for school re-open..well tonight channel U telecast the show \"my sassy girl\"...it was my favourite show as i longed to having such a gf like the quan zhi xian..haha.....the characteristic was my ideal= )....and its early morning now...and i`am goin to catch the next euro 2004 match france vs greece..thats all for today...sweet dream... '),(7036,3478510,'24,June,2004','2004-06-24 00:00:00','   hee!!today was a  happy day for me...its been a long time since i been engaging in water activites(bout 3 months)....and with my best frenz cheeeewwww weeee(siew wee) , my twin korkor and my doggie frenz...(alex)haha...our destination was jurong swimming complex....although everything was not going as planned(some of them(frenz) nv turned up), but four of us did enjoy ourselves..at first i thought that siew wee wont turn up as he was busying doing his D&T project, but he did and we met at hougang mall(1430)....hee! And so after his arrival,we were on our way to the destination...we had much laughter on the way as we were crapping around..haha...we reached the complex at around 4`o`clock...it was a bad start for me;when i was  enjoying the ride of the blue slide,accident occurs.  .  .  . the float capsized and i bang my head although its a bit painful....after this incident,i was afraid of taking the blue slide again as i had nv capsize before....but i brave up my courage and tried it again..this time everything was well and it didn`t capsize and it relief my nerve=)!!hee.. we had much fun time at the wave pool and dizzy river....we were trying for diff pose to get onto the  float and the action was funny too...haha...time past quickly and is was 7.45p.m...we had our bath and went for our dinner at kovan heartland mall and siew wee went home himself...upon reaching heartland mall,we were searching for neoprint machine(me, me korkor n alex)..and to my amazement,heartland mall had change...there were much more shops as compare to last time with only EPbookstore and best denki..indeed we found the shops...the shop had many neoprint machine....we taken our pics and was disappointed as some of the pics was not nice...but i liked one of those as it was very nice...haha....kawayi...and so me and me korkor had our late dinner at the hawker centre and doggie return home....after the dinner it was still early and so me and kor had our finger exercise at superbowl...haha....its been a long time seen we went there ...and its quite enjoying.......after a long day.we return home....and home sweet home...=D..that`s my today:24th June 2004 '),(7037,4320505,'23,August,2004','2004-08-23 00:00:00','   In my family practice clinic, I have a lot of good days as my nurse and I sift through the miliau of ambulatory illnesses that cross my path.  Occasionally, we have a day (or a string of days) where we feel like asking, \"Do you want fries with that?\"  Let me explain.  In this culture of fast food and 7-11s, that has been the mantra of customer service.  People are used to ordering in a loud voice thru a drive through speakerbox.  They are everywhere, pharmacies, banks, dry cleaning, and even the big red drive-thru beer barn on the seedy side of my town.  In the old fashioned world of the office of a family practice doctor, we are still clinging to the exercise of face to face communication.  Most of my patients love this type of interaction, especially the older, retired people who clutch their bags of medication and Medicare cards hoping to win the jackpot of longevity thru pills.    There are a few, however, who make my nurse feel like a waitress and me, a board-certified, yada yada, physician as a short order cook.  Is this really the kind of relationship you want with some people who have the potential to really help you or hurt you?  If I was a lazy person, I could just write out whatever drugs people felt they needed.  In fact, I could just install a vending machine and people would just insert their credit cards and then push a button.  I could be a kazillionaire.  But then, what a waste of education and many years and late nights of studying pharmacology, physiology, biochemistry, and other useless medical school classes it would be.    These fast food medicine consumers will tire of my long explantions about how this particular antibiotic they want (after seeing on that catchy commercial on TV) will not kill the bacteria that is currently infecting their prostate.  They will move on to other doctors, physician assistants, nurse practitioners, urgent care centers, emergency rooms and other providers that will be most happy to give them what they want and supersize their order for only a dollar more.   '),(7038,4320505,'23,August,2004','2004-08-23 00:00:00','   I am a slow starter, last one to get the CD player type, and I am going for the Blog.  So here goes...  While I was watching the Olympics the other day, one of the trivia facts thrown in my the usually inane sportscasters was the average income in the Ukraine was $76 a month.  That really stuck in my brain and I am not sure that is even right.  But it made me think about how rich materially we Americans are, yet so often we think we lack so much.  A single person making $780 a month in my town qualifies for indigent assistance at our local hospital.  Considering how much medical care can cost, that is probably appropirate.  Ukraines are taxed at 80% and get free health care, but apparently you have to bring your own cotton swabs and gauze.  As I watched more Olympics, I noticed some of the atheletes representing foreign countries were actually American (living here).   I further understood the love/ hate realtionship for Americans around the world.  We are one of the few countries in the world that people want to come to live and work in.  We are not a country that many people leave or that a government has to fence in or shoot at when escaping.  But what I think Americans often need to be reminded of is that material wealth doesn\'t necessarily promise happiness.  I saw it in the faces of a Ukranian gymnast who did a really great routine, joy.  I just wished that they could have looked more well-nourished like the Americans.  '),(7039,3068590,'29,March,2004','2004-03-29 00:00:00','   Based on the recent response to Richard Clarke, and the reaction of people towards Bush with regards to Iraq and the war on terrorism. It is safe to say that future presidents will pass on this style of controversy. Let the world rot like Clinton did, and avoid the Bush Sr. / Jr style of political demise. After all who needs sound leadership?  '),(7040,3068590,'29,March,2004','2004-03-29 00:00:00','   I will attempt to import my blog from .mac to blogger, if not it can be read at this url. (click on title)  '),(7041,3068590,'30,April,2004','2004-04-30 00:00:00','   I must be psychic.  Speaking about the management issues in oil and gas industry in the prior blog, it is clear that the related accounting issues are systemic throughout all industries, or just in Canada.  Stating their is a difference in the way the Canadians and Americans approach their business problems, this problem is likely Canadian only. The Enron\'s and WorldComs prove the American desire to start over is both constructive and healing. Whereas Canadians like to relive the same car crash over and over again.  Therefore based on that it might be wise to hone up on our methods of politely answering the phone. Houston will be calling.  '),(7042,3068590,'27,April,2004','2004-04-27 00:00:00','   The Canadian oil and gas industry leaves the door open for a further extension of the branch plant management style that is becoming prevalent in the oil and gas industry. My concern is for the method of management that occured in Calgary up to 1983 was a branch plant management style where international companies owned and operated the industry. This mindset was eliminated by the dynamic and entreprenurial spirit that is prevelant in Calgary and I fear that the branch plant method may soon return.   Why? Its really quite simple, we have continued to prove that we are having significant difficulty in maintaining our deliverability and reserve base. The ability to drill wells on the prairies is constrained by a number of resource limits that impede the work that is required. When the work is completed the ability to estimate and deliver on the basis of what it will cost is poor and deteriorating quickly. This is particularly difficult in the oil sands where cost over runs are approaching 100% on multi-billion dollar projects.   A secondary issue and one that is the focus of my thesis is the reporting and accountability capability in the industry. My opinion regarding reporting is that it is desperately constrained. The ability to \"report\" to the various regulatory and statutory requirements is constrained by the systems in play. Other industries can use applications such as SAP to meet the revised Sarbanes Oxley and CICA requirements, however, SAP in oil and gas, in many instances, can not even report gross and net on operations and capital.   I have a couple of pointed questions regarding the attitudes of those international producers that have taken over Canadian independents in the past few years. Believing they had cherry picked the best of industry for themselves; did they subsequently regret their acquisitions on the basis that, with the key talent moving on, the remaining management was inadequate to run the company? Hence moving the management to other locations?   Those companies or assets that were of no interest have found a home in the royalty trusts where the objectives are limited to the further exploitation of the reserves and the explicit renunciation of any and all exploration.  This is not what the industry needs in the long term.   With only a few real independent left, the time for those remaining, should head the saying that we reap what we sow.  '),(7043,3068590,'26,April,2004','2004-04-26 00:00:00','   Since everyone is leaving Iraq based on the terrorist threats to their countries. (Spain). Now would be a good time for Paul Martin the fearless prime minister of Canada stand up and do what he said he would. A very novel and unique idea for a Canadian prime minister.  He said that he wanted to repair the damage made by his predecessor and patch relations with the United States, and specifically George Bush.   Well bigger opportunities don\'t come along that often Mr. Martin. Call up George and state that you would be very pleased to send the other 5 troops of Canadian military to Iraq to fight in the War on Terrorism.   Just a thought, however, I think that Mr. Martin will no doubt be creating his next crisis to divert Canadians waning attention span.   (No slight against the 10 outstanding Military personnel in Canada. They have done extremely well on a poor budget.)  '),(7044,3068590,'25,April,2004','2004-04-25 00:00:00','   Its Formula Michael, Not Formula Juan.  Line up behind all those that were bumped in line before you. And stop criticizing the son of Max Mosely\'s nightmare.   You should know by now that you can not beat Michael.  '),(7045,3068590,'24,April,2004','2004-04-24 00:00:00','   Google\'s IPO is scheduled soon. Why they wanted to do it was a mystery to me but I now find that the company had to complete the standard of public reporting anyways. This requirement was invoked due to the probable size of the market cap is in excess of some predetermined number of billions. Google has some proprietary data that they do not want in the public domain regarding their financial situation.  Nonetheless, parallels to the Netscape and .com IPO\'s has begun with the expectation that Google\'s value will exceed $25 billion on its IPO. Far greater, by a substantial margin, then any other valuation. The average investor will be wary of the market price of google, however, if microsoft can top out at $400 billion, I would suggest that Google may be worth at least that.   Given the lead start they have over their competition, evaluating their use of the technology, and their pipeline of products. The only concern Google has is the speed of acceptability of their application of technologies. (Gmail for instance.) This company is a thinker and an innovator, and that in addition to the billions in cash from the IPO, will be the reason that the internet and technology continue to morph into its potential.   It certainly is going to be an interesting ride, and start the countdown for the demise of Microsoft. (My Microsoft countdown started 15 years ago.)  '),(7046,3068590,'24,April,2004','2004-04-24 00:00:00','   I only wish.  Seems the powers that be at Ferrari are reading this blog. Recently Valentino Rossi was invited to test a Ferarri at Fiarino. Completing 20 laps he came within 3 seconds of the lap record. Not bad for a guy who doesn\'t drive a car.   On a related note, Max \"the black shirt\" Mosely has decided to propose major revisions to the Formula one rule book. Since Max created the monster that is known as the six time world champion Michael Schumacher. I wonder what monster Max will create with these new rules. It is pleasing to note that McLaren\'s Ron Dennis has given up responding in any form or fashion when Max speaks. Keep up the good work Mr. Dennis, Max is out of his mind which is nothing new.  Watching the races from ten years ago, specifically the Monaco Grand Prix, this being the first race of the calendar after the death of Ayrton Senna, shows the quality of the talent in 94 and 95. No wonder Michael won, no Nigel Mansell, No Alain Prost, and No Senna. Instead of coming in fourth at best, he just needed a road car to bring him home the win. No disrespect to De Cearsaris or Martin Brundle intended.  '),(7047,3068590,'20,April,2004','2004-04-20 00:00:00','   Alan Greenspan looks like he may have been able to achieve what it is that he set out to achieve so long ago. The years of stable prices has created the opportunity for an extremely efficient market economy. An efficiency that is realized on a global basis. The ability of the market to respond is truly remarkable, and required. There are a number of challenges ahead, and it seems that Mr. Greenspan has prepared us well for this future.   On Tuesday in his testimony to the Senate, Mr. Greenspan noted the level of sophistication that the market was using to hedge their risks and leverage the assets under management. And when you think of the changes in the means and methods of financial instruments, the analytical tools and understanding, not at the higher academic levels, but contained with the common public understanding, its no wonder that the market looks very rosy to Mr. Greenspan.  '),(7048,3068590,'20,April,2004','2004-04-20 00:00:00','   Apple announced a number of products and alliances that place their video production and editing in the High Definition era. But only 64 bit apps.  Of particular interest is the alliance with Panasonic. Seems like Apple has prepared a method to secure the intellectual property rights of the movie creator and ensured that the technology respects that any copying will only be done in an authorized manner. I\'ll bet Pixar would never be interested in that. Doing a deal with Panasonic is a first for both companies and proves that Apple can establish standards with any electronic manufacturer. Somebody should think about doing that with Music, wait Jobs did but he only implemented it several years ago.  Ho hum another day that Jobs locks up another major future business, while all the others run around trying to get into the music business.  Jobs needs to understand that talking big is more important than products, services and strategic positioning years in advance of the rest of the manufacturers.  On a related manner the announcement of Storage Area Network services will be available in the fall of 2004 shows a distinct lack of understanding that Apple is a PC company and not a super computer vendor. (Virginia must have been a fluke!) Besides Intel is only now planning commercial versions of 64 bit processors with Microsoft announcing their 64 bit offering in 2005. Why does jobs not understand that delivering 64 bit processors, operating systems and applications will never win over customers. Just announce the stuff is coming, thats all that the consumer wants!  I hope that the press continues with their field day in making Apple look like a bunch of idiots, the press needs the company. It certainly will be a long time before Jobs realizes that execution is more important then strategy, at least so the saying goes.  In an unrelated manner, keep serving lunch Jobs, the press and your competitors don\'t seem to get it.  '),(7049,3068590,'19,April,2004','2004-04-19 00:00:00','   There was an emotional and deserved pause by the legend Valentino Rossi in Welcome South Africa this weekend.  Rare would we ever see such a display. Dominant in the Honda for the past 4 years, Rossi chose to leave to take up his bitter rivals drive of a Yamaha. Although a competitive offering, the budget and scope of Yamaha was sizeably smaller then Honda\'s effort. Rossi\'s rivals stated that if they had the Honda they would also win, so Rossi obliged them.  To take the pole was a surprise.  To win consecutive races on different machines has never been done.   To win consecutive championships on different machines has never been done, and only done on different machines in a career once before.  This is what a true champion thinks and does. The otherwise impossible.  Michael Schumacher is nowhere close to attaining any of the respect that he claims with Rossi now in the picture. Schumacher should have long ago swapped rides with either Mika Hakkinen or Jacques Villinueve, or even join them at the same team.   As the prior blog entry reflects, Rossi\'s championship makeup is established. It is time for Schumacher to begin to better manage his career before he becomes a brief anomaly in the history books.  '),(7050,3068590,'13,April,2004','2004-04-13 00:00:00','   Valentino Rossi gets back on his bike (Yamaha) this weekend and starts to demonstrate his championship attitudes and capabilities. I hope that both Max Mosely and Bernie Ecclestone, past and current presidents of the Michael Schumacher fan club, note that Valentino is a champion and what they have been passing off as a \"champion\" in Michael Schumacher starts to fall apart.  I only hope that its not to late for Ferrari to recover from the inevitable decline of Michael \'mania\' induced by Jean Todt. (Michaels sycophant boss)  To a spectacular season of demonstration of daring, skill, personality and the championship attitude of Mr. Rossi.  '),(7051,3068590,'12,April,2004','2004-04-12 00:00:00','   Microsoft appears to have a use for the money that they have generated. And like all things Microsoft it is counterproductive. Recent speculation around the motivation in the Sun settlement and other settlements has revealed a possible new strategy.  This may also be considered innovative by Redmond however, the application is not.  The use of Patents to eliminate the use of open source code in the corporate environment appears to be the strategy they are laying out. Retroactively applying for patents on existing patents, the old embrace and extend with a legal twist, attempts to lock out any non-licensed software from interacting with any licensed Microsoft software operating in a client corporation.   Legal technicalities aside, the money used in a court of law in this fashion will only seek to \"embrace and extend\" the threat and effect of Microsoft \"innovations\" in the marketplace. The solution that Microsoft has suggested is that don\'t run any non-licensed software in the corporation.  '),(7052,3068590,'10,April,2004','2004-04-10 00:00:00','   Was that not something that Jackie Gleason was asked?  The comedy routine continues with the Ralph in question, Schumacher, states that the Toyota team are in discussions with him for the next year.   Reading my prior blog, you\'ll notice that Ralph does not deserve a civilian license let alone a Williams.   It now appears that Ralph is out of the Williams for next year as their does not seem to be any major announcement of his signing.   Toyota would pay for the Schumacher name and that is Ralph\'s best interest. Williams have no signed drivers for 2004, and yet don\'t seem to be the least bit concerned about it.   As we start the 2004 season we can clearly see the pretenders and the contenders, and really only the current stable of champions of Jacque and Mika Hakkinen, but not Michael Schumacher and Juan Pablo Montoya and Kimi Raikonen are capable of securing a championship. Mclaren have their drivers, Ferarri have theirs and Williams is sitting back and cherry picking the real talent.   Lets hear it for Sir Frank for continuing to be the smartest guy on the grid.  '),(7053,3068590,'05,April,2004','2004-04-05 00:00:00','   This article marks the decline of the PC as we now know it.  Long live the Mac.  If 700 PC\'s can only muster 180 Gflops, and 1100 G5\'s in Virginia are the third fastest supercomputer in the world at 10.28 Tflops.  Then note the PC\'s death occurred on April 5, 2004. Just don\'t tell anyone who has a PC.  '),(7054,3068590,'04,April,2004','2004-04-04 00:00:00','   Now is the time to go on the record regarding this years F1 champion, and a few other rambling thoughts.   The champion will be either Kimi Raikonen, with the second alternative being Juan Pablo Montoya.  Constructors championship will go to McLaren.  I hope that Jenson Button recognizes the efforts of one Jacques Villienueve in his current success. I am certain that Dave Richards and Button will realize the former champion efforts by the end of this year. They will look back at the end of the year and wonder how they were so successful at the beginning and how did they tail off.  Time for Patrick Head and Sir Frank to seriously consider the replacement of Ralph. The man does not deserve a license, let alone a Williams. Time to get Jacque back and make some hay with the BMW boys early.  '),(7055,3068590,'03,April,2004','2004-04-03 00:00:00','   In my past blog I wrote about the strong analogy that could be drawn regarding the extinction of the Buffalo and the current situation in energy. (Click on the title to be taken to the old site)  I wanted to add that in Canada we have a historical site called \"head smashed in Buffalo jump.\" The site documents the native method of herding the Buffalo over cliffs.  I was thinking that there may also be a strong analogy between the Buffalo (again) and the energy industry (again). The fact that the commodity prices have risen so substantially has created a new operating environment for oil and gas companies. However, it is also just as remarkable that the companies have not changed any of their thinking regarding the approach to oil and gas.  '),(7056,3068590,'03,April,2004','2004-04-03 00:00:00','   I don\'t recall where but I remember reading that the Chinese are having difficulty in developing pipelines to fuel their energy needs. The ability to meet the demand for energy within China will not be met by China. The need to import energy is well established when we consider that China is now the second largest importer.  The inability to secure the pipeline and deliverability for its domestic needs, the Chinese feel susceptible to an oil embargo from the Pacific Ocean. They are therefore developing their first, and certainly not last, aircraft carrier. Now that is a bold energy policy.  '),(7057,3068590,'28,May,2004','2004-05-28 00:00:00','   This week saw two important developments in the technology world.  Microsoft announced that a major component of longhorn\'s 2007 release will be a major initiative towards super computing. This would fully exploit the 64 bit technology of long horn and be the typical Microsoft type of wonder.  Secondly the denial that Apple orchestrated towards its users in allowing the concept of a virus attacking their system did not muster the appropriately deemed result.  That this is a theoretical possibility, and it is important to think what could happen if, that no one was attacked, and indeed no virus has been written, does not accelerate the concept to the full blown risk that the PC people have as a result of turning on their machines.   Anyway, Jobs just keeps delivering products that are fast, powerful and safe, which is proof he just doesn\'t get it, when all you have to do is promote garbage 4 years before the time you expect to deliver to really earn the care and attention of the user community.  '),(7058,3068590,'23,May,2004','2004-05-23 00:00:00','   Michael Schumacher\'s car betrayed him by heading directly into the wall. Either that or Juan Pablo Montoya\'s \"car\" betrayed him by not attempting to avoid Michaels Ferarri.  Maybe next time the six time world champion will be able to muster the ability to overcome the special attributes of his car.  '),(7059,3068590,'20,May,2004','2004-05-20 00:00:00','   Step two of my 2003 prediction is about to be undertaken.   Ralph Schumacher has reportedly signed with Toyota, and good riddance.  The road is now open for two seats to be filled at Williams in 2005.   With such limited talent available, Villenueve is the only proven talent remaining.   The machinations of Patrick Head are only to reinforce the prior blog entry, that the M. Schumacher sycophants Eccelstone and Mosely pay for the Canadians substantial $35 million annual salary.  '),(7060,3068590,'19,May,2004','2004-05-19 00:00:00','   It should be clear at that time that the \"technology revolution\" started on May 10, 2004, and a major re-configuration of the development market will occur.   The event that I am speaking of is the Apple Worldwide Developer Conference and Steve Jobs will be using this forum to make some major announcements. The first is the announcement of \"Tiger\" there next version of the operating system.   As I like to do, I like to make predictions of the types of products will be made.   Further details of these, and why they are important will be posted to this blog before the 28th.     - Redevelopment of the entire IBM suite of development products on Apple\'s Platform. I previously predicted this and look forward to the day that it actually occurs. Requires IBM to port their tools to Apple\'s G5 processor, aka the IBM 970 processor. Apple\'s contribution to this will be the release of their API to Cocoa and WebObjects to IBM.     - This offers enterprise customers an impenetrable security model for any developments made on this homogenized platform.     - Updates to the Xserve to 4 - way processors. This will be the outer limit of the 1U server, and the beginning of volume production of these units.     - Updates to Xraid, and clusters, with the addition of a \"blade\" processor based on G4.  Further elaborations on these points, particularly the \"security\" model will be made prior to the 6/28 deadline, barring the silencing of this point by Apple or IBM.  '),(7061,3068590,'18,May,2004','2004-05-18 00:00:00','   The title will lead you to the glitch that caused the Terabyte.  There is always the future.  '),(7062,3068590,'18,May,2004','2004-05-18 00:00:00','   My Gmail states that my Gigabyte is now a Terabyte.  If only this were true. I noticed that some other services had bumped their disk space to 1 GB in response to gmail, however, a TB would be awful nice.  Not much I could not do with that.  '),(7063,3068590,'14,May,2004','2004-05-14 00:00:00','   Buzz is the real innovator and visionary in the Canadian economy.  By effectively killing Air Canada, he has effectively solved the energy demand issues, and increased the wear and tear on personal vehicles, sustaining the Canadian Auto Worker for another two generations.   All without any explicit government support. Who says you can\'t teach an old dog new tricks.  '),(7064,3068590,'14,May,2004','2004-05-14 00:00:00','   Some needed debate on the oil price situation.  This attempt is being run at this blog site and at the econolog as the encouragement to discuss this point was made. (Click the title to go to the econolog site.)  The facts are what is causing the Americans to believe there is a conspiracy in oil supplies.   Attempts to put across three important facts generates significant resistance, these facts are.   - There are no alternatives to the use of oil and natural gas.   - There are finite supplies.   - The American economy requires energy in its current form for the next 5 years to sustain its advanced economy.   These facts should also be augmented by the consideration that:  The world is a global economy where supply and demand are international on critical commodities.   The American has significant competitive weaknesses.    - Significant reliance on imported energy.    - High labor costs in comparison to China and India.  The use of imported oil by the Chinese is now second only to the U.S.    - The Chinese are building Aircraft carriers.    - India and China have communist governments.     - The Chinese economy is operated for the long term, the Americans need to meet the next quarter numbers.     - The Chinese economy is one tenth the size of the American economy, and the same size as the Canadian economy yet uses 3 times as much fuel to create the same GNP as the Canadians. (Values are approximate.)    - There is a new political alliance between the Russian, Chinese and Arab states on the supply and demand of oil.     - There are now beliefs that the new political alliance can develop their resource without American technology. The dependence on foreign technology hurts their economy.    - The Chinese strategic competitive advantage is their low labor costs and large population. (And a disproportionate number of men.)  The American mind set that Cheap oil fuels the ribbons of SUV\'s is the most ridiculous abuse of resources since their abuse of the Buffalo.  This tendency for short term thinking is coming home to roost in the new Global economy.  I would assert that the Chinese are purchasing five to ten times as much fuel as they need to \"stock\" their own strategic reserve to be used as a competitive weapon against the \"west\".   The United States needs to wake up and learn some really tough lessons. Its a competitive global marketplace.  '),(7065,3068590,'13,May,2004','2004-05-13 00:00:00','   while at the Starbucks. An order for a half coffee and half cocoa? Was this a new kind of refreshment? I did not see the desire to drink that type of mix but to each their own.  After I thought about it, I began to think that maybe the individual had a significant trend identified. So I did some checking.  The high end developer is using Apple\'s PowerBook for its Unix underpinnings and Java integration. Key to the evangelist is the explicit and overt support for the hardware by Dr. James Gosling, the founder and key contributor to the Java Programming Language.  What Dr. Gosling uses the PowerBook for is also the \"Mac\" only development environment known as Cocoa. Cocoa is the product name that Apple provided for the development environment acquired when they bought Steve Jobs from Nexxt.  I also use Eclipse and Netbeans, but use Cocoa and purchased WebObjects from Apple. Cocoa is also used as the primary development environment by my kids. Yes it is that sophisticated and capable that Gosling and my kids are now using the same development environments. My oldest (13) uses Cocoa with a group of like minded kids in California and Kentucky through their iSight webcams.   The point is that I subscribe to Cocoa, WebObjects, Eclipse and Netbeans environment mailing lists. (Please review how I use Gmail in this blog.) The volume of WebObjects and Cocoa based mail is significant. In most cases almost 5 to 1 of the other environments. Is that good or a reflection of more problems. The analysis shows that Cocoa based questions and answered are being debated by academic and scientific based individuals. (Apple quietly announced a science based market drive that is reaping big rewards for the firm.) The doctors and scientists are solving and debating some severe technical issues. They are not debating minor points.   Although Eclipse and Netbeans are young and starting the troubles and difficulties of these environments don\'t compare to the robustness and maturity of the 10+ heritage of Cocoa and WebObjects, but they are still crawling. It will be interesting to see if anyone suggests that Jobs should take Apple out of the development and science based areas as well. I would expect they will catch on to this and make that recommendation in as little as two years from now.  '),(7066,3068590,'12,May,2004','2004-05-12 00:00:00','   The limit of collaborations.  Collaboration is the future of how the economy and businesses operate in a fundamental manner.  I have always subscribed to the theory that I would put out all my ideas at once and let them develop as they should through the process. Feeling that their are no limits to what should be stated and just continue to push the thought forward.  I now define the limit as the point where the idea is too over the top, or just nuts.  So I apologize to all of my victims for the over the top methods of my collaborations. Lets see if this limit has any effect.  '),(7067,3068590,'12,May,2004','2004-05-12 00:00:00','   Anyone interested in analyzing the ideal political model should review Formula One and the powers that be in that sport.   Bernie Ecclestones ability to influence and affect change through political manipulations are a case study on there own.  The natural foil for Ecclestones political desires is Max Mosely. Son of the \"black shirts\" Mosely and a political animal of his fathers and Ecclestones class.  The current problem is the sports declining appeal due to the dominance of Ferarri. What is required is a spark for the sport to draw the fans back and that is one Jacques Villeneuve, and although \"Bernie\" and \"Max\" fought the young Canadian at every turn, they now realize the value of his honest and direct nature.  Enter Sir Frank Williams who understands drivers best and the fact that Jacque was his last world champion and best opportunity for 2005. Jacques testing in France this week leaves one believing that the universe will align and the Canadian will drive for Williams soon.  Enter Patrick Head, partner in the Williams organization and technical genius. States the Canadian was in France not on the invitation of the team.  So Patrick was does this mean? I believe that the sport needs Jacque so badly, and the solution to Williams driver line-up is Jacque, that Sir Frank and Patrick feel they can poach Jacque and have the powers that be contract and pay the Canadian for his services in F1.  Bernie and Max have acquired Billion dollar fortunes through their alleged ownership of Formula One. Sir Frank and Patrick Head, who like other teams, do most of the real work in Formula One, is a modest man of maybe $100 million in net worth.  If Bernie and Max want Jacque in the sport they can pay for him, and the \"league\" will pay for his services. After all what is that saying \"let the end user pay\".   Watch out Bernie, your slipping, and by the way how come the Queen has not called yet to provide you with that much coveted knighthood?  '),(7068,3068590,'12,May,2004','2004-05-12 00:00:00','   Not to steal a famous saying and denigrate in the context of a commercial point, but...  Every two years I drive across the Alberta Montana border (twice) to support the Arthur P. Cox Memorial Ryder Cup Team. (We still hold a perfect record.)  I was thinking ten years in advance and marveling at the size and number of Nuclear reactors that have been buit on the Montana side of the Alberta border.  It is the only practical means for North America to secure its long term energy supply. (The reactors fueling the process of heavy oil extraction.)  '),(7069,3068590,'11,May,2004','2004-05-11 00:00:00','   Comments regarding the role and influence of Apple in the marketplace are making an interesting turn in their tone and substance.  As opposed to declaring that Apple is on a death march, they are encouraging the company to abandon the Macintosh to pursue the iPod and give up on enterprise computing.  In the world of rampant vested interests (which is today) Steve Jobs appears to me to be one of only a handful of strategic innovators.   Strategic innovation is the only method that can sustain a firm in this new market environment. Vested interests are the counter weight to the innovator.   Keep up the excellent work Mr. Jobs, they\'ll get it some day.  Oh, the comment about the enterprise will be certainly be those individuals with vested interests next biggest mistake.  '),(7070,3068590,'11,May,2004','2004-05-11 00:00:00','   The beheading of an innocent American civilian should put this debate in the proper context. The Al Queda and Iraqi prisoners do not deserve our concern. They are criminals and abused their system for 25 years to advance their personal causes. These people are scum and nothing makes that more evident then the fact that they think today\'s actions are in response to the prison incidents.   Wake up, this is war and they are the enemy.  '),(7071,3068590,'10,May,2004','2004-05-10 00:00:00','   These url\'s and this blog are documentation of an issue that I discovered in my Master\'s thesis and is the risk associated with the use of IBM\'s Workplace toolset. This product has a very \"dark\" side and companies should be warned about the threat. Contact me for further discussion if you would like.   This issue is particularly of concern due to the fact that today, IBM announced the Workplace product and its intention to eliminate Microsoft Office.    urlLink Computerworld    urlLink Techworld    urlLink Financial Times    Technology has now stepped across the threshold of being either benign and inert, or positive. To that which is either positive or destructive to a corporation. My thesis documents how these technologies have the capability of causing significant damage to a corporations data integrity. That this threat is an unintended outcome as a result of IBM dealing with the eight-hundred pound Gorilla known as \"Microsoft\".   This toolset wants Microsoft out of the corporate marketplace, and I would think that the majority of the corporate world would agree. However the unintended consequences of technology today, whether IBM or otherwise, brings the context more in line with a game of snakes and ladders where the board game is played in the offices of technologies customers.  '),(7072,3068590,'09,May,2004','2004-05-09 00:00:00','   Significant changes have occurred this week that will have an impact in the next few years.  The most prominent news was Intel\'s announcement to ditch the next generation P4 desktop and Xeon processor development. Sighting the heat problems they stated that to resolve the problem was too great compared to the potential results. The solution is the ability to run dual processors in one chip design. Now I may be wrong but the IBM 970 and Macintosh G5 both carry 215 separate instruction units, I am certain that they will warmly welcome the future Intel competition. Just as the Model T is not used in transportation today the Intel processor will cease to be a viable commercial solution.  Microsoft has noted difficulty with their secure computing environment and has stripped some of the functionality from the 2006 Longhorn release. This environment in response to the viral and associated security issues that Microsoft is experiencing. As I have stated before, building the worlds largest structure out of toothpicks will lead to problems. The toothpick analogy to the Microsoft technology I think is valid. Now I may be wrong but Microsoft was using this environment to manage the intellectual property rights of MP3 copyright holders.   Apple announced \"Tiger\" there next iteration of the Mac OS X. This following on the heals of their SAN, Xserve G5, iTunes, New Powerbook models, Final Cut Pro, Motion etc. It would be unfair for me to compare these two environments because the Mac is such a poor second cousin of the Wintel duopoly, and no true computer user would use a Mac.  '),(7073,3068590,'07,May,2004','2004-05-07 00:00:00','   As an avid Apple user and proponent, I enjoy watching the fits of the Microsoft Intel world struggle with the viruses, security and performance related issues. Add the heat issues that Intel is stating are the primary reason for the next generation P4 demise, to the never ending deadline for Microsoft Longhorn, one begins to see and appreciate the reasons why the Macintosh platform should be avoided at all costs.  If after 8:15 in the morning your machine was running without a hitch, that would require that you do a full days work. Everyone knows it is much more enjoyable to chase down phantom security holes, fiddle with the \"dip\" switches and unexpected performance failures as your full time profession.  '),(7074,3068590,'05,May,2004','2004-05-05 00:00:00','   Trust the googlers to come up with Gmail.  It is truly a dream of a system and is highly recommended. The concern about privacy is *%(#$. The ability to have the advertising of the vendors that you may be interested in, fed to you without any effort will only be interested in consumers that are looking for new and innovative quality products. Those interested in *$%&# will keep by the Microsoft product suite.  One of the ways I use email is to get the most current information about vendor\'s product by aggregating their various mailing lists. As time passes, the ability to search the recieved mailing lists provides the means of determining the answer to questions that are not generally known by anyone.   The ability to search provides a knowledge base that does not otherwise exist. As you add additional vendors, and add Google\'s search capability, you begin to have a technical knowledge base that begins to be amongst the best sources of technical information anywhere.  '),(7075,3068590,'05,May,2004','2004-05-05 00:00:00','   I seem to be out of step with the majority of thinking regarding the Iraq war in general, but particularly on this topic.  It would appear to me to be acceptable for the American soldiers to be treating the Iraqi\'s in the manner that they are.  Although I realize it is inconsistent with the Geneva convention regarding war criminals, it is better then what the American soldiers should have done to those people. (just shoot them next time, and skip the problems). And I am certain they are being treated far better then they were \"dishing\" it out over the past 25 years to their victims. Those being the innocent women and children that did not follow the rule of law under So Damned Insane.  These alleged humans, the Iraqi prisoners, are the primary reason that a dictator has the ability to keep power for their own purposes.  '),(7076,3068590,'05,May,2004','2004-05-05 00:00:00','   An interesting dichotomy is being identified and discussed regarding the energy prices and the energy industries reaction to the shortfall.  Firstly review of the ISM manufacturing and non-manufacturing indexes for the month of April 2004 reflects that the American economy is under a massive increase in GDP. There are various pricing pressures that put the pricing index under a severe strain, registering 88 for the month. The review also reflects that although there are inflationary pressures, there appears to be a trend for increases in productivity to meet the constraints. That although the inflation pressures are impressive, the productivity is offsetting the trend and mitigating the impact.   The pricing situation discussed previously, has an odd anomaly in the long term energy prices are not accurately reflecting the shortages in the future. The energy industry uses these long term prices to determine and allocate their capital budgets to the greatest return. Costing exploration and development on the economics of $28 oil is fundamentally different then the current prices of $40. Not until the long term prices reflect the current prices will there be any reaction by the industry in allocating funds to riskier projects, or to increase their overall capital exposure.   Based on this dichotomy in energy pricing, the need to ask if a revision of the pricing model to better reflect the supply and demand in the long term is required?   I would also postulate the need for the higher energy prices is only beginning to be reflected in $40 oil. The future of cheap energy solutions is over and will never be seen by this generation again. With the increasing productivity of the U.S. economy, and the level of economic activity increasing the demand on energy, a further doubling and tripling of prices may yet occur.  '),(7077,3068590,'02,May,2004','2004-05-02 00:00:00','   Everybodys favorite topic these days is the cost of energy. People are surprised that the oil and gas companies are charging so much for what consumers have taken for granted, and abused, for so long. Consumers should grab a Starbucks latte and start thinking seriously about this topic.  One of the more critical areas that is a source of my frustration is the short term thinking that goes into the pricing of the commodities. My frustration was that commodity pricing was essentially based on the availability of product as reflected in the current inventory, general deliverability and storage. I felt that this has been short sighted and needs to be revised. The point I discovered is that the pricing assumption is not necessarily incorrect. The inventory numbers are supported through the assumption that all oil and gas reserves on a worldwide basis have been reported as increasing year over year. The oil and gas companies continued to find more then they produced, therefore the situation was constantly improving. This tacit understanding was not flawed but a reasonable assumption.  The issue today is the reserves of the oil and gas companies are beginning to be questioned as to their validity. Trained in management I always found reserve engineering to be somewhat of a black art as opposed to a science. These reserves, as Mr.Mattheww Simmons states, are known exactly what they are when they are gone, much in the same way an actuarial knows when the subject has died. Everything else is an educated guess.  Unfortunately the oil and gas companies needed to optimize their reserves in the past 15 years based on the stock market demands for performance, and the consumer markets demand for energy. Deliverability became the issue more then the reserves. Again, as Simmons states, there has been a limited amount of application of known technologies to prove the reserves were not there. This has created a situation where some producers are writing down the value of reserves. The situation is well documented at Shell and recently was noted in El Paso\'s reserve write down of 40%. It appears to me that the latter instance creating little or no market response.  Another issue that will amplify the scope of the write-downs is the fact that reserves are booked as financial assets on the oil and gas companys financial statements. The statements are valued on the prices realized; less capital required to access them and the operating cost to recover them.  The issue should be defined by the fact that the majority of the gas reserves recorded when there was an abundant gas reserve, but at the low gas prices of around $1.50. When the reserves are booked there is an implicit need to reflect the commercial value, and would have probably been valued at their net of around $0.25.  At the current gas prices of $5.95 the net should therefore be around +/- $2.50. It is inappropriate for a company to restate their reserves upwards to reflect their revised market value, but accounting rules require the strict recording of the lower of cost or market value. The question then should be asked by those that are drinking Starbucks coffee is:  	Why are Starbucks prices going up to cover the costs of milk? 	Why are gas reserves being written down on producers balance sheet?  I can assure those that are able to answer the second question will begin to understand the real value of energy very soon.  '),(7078,3068590,'01,May,2004','2004-05-01 00:00:00','   Seems to be talking to one Canadian Jacques Villenueve.  Frank is either leveraging Jacques desire to drive a championship car, for little in salary, to make Ralph Schumacher cave on his ridiculous salary expectations, or legitimately negotiating with Jacque.  As mentioned before, Frank is the key poacher of high quality talent that is championship grade at low costs. Villeneuve is probably going to join the ranks of the other greats who signed on with Frank to be with a racer. The other greats all went on to championship status soon after signing with Frank. Nigel Mansell, Alain Prost and unfortunately Ayrton Senna were all grateful to be driving Franks cars.  '),(7079,3068590,'09,June,2004','2004-06-09 00:00:00','   Seems that the powers that be at Apple are feeling the markets response to them is strong. Product announcements are coming fast and furious with the 28th of June being the ultimate crescendo.  '),(7080,3068590,'26,July,2004','2004-07-26 00:00:00','   I have always subscribed to a theory that the value contained within the North American marketplace was subject to greater risks that the average consumer / investor could not comprehend. The ability to move from the consumer to the investment environment would create winners and losers and these would lead to significant opportunity or destruction of value.   This destruction, in my opinion, has commenced with the markets slow and insidious decline in value. There will be no overnight reconciliation, only death by a billion pin pricks over a number of years. Only when the average consumer can perceive the losses that have occurred in this environment, will the death be realized.  The value declines will come in a number of methods that were assumed to be the basic fundamental components of western living or lifestyles. They include personal values, securities, and pensions.   The home values and personal possessions are the premier concern of the Western consumer. These are attributed far greater value then they currently can and will realize, therefore the value associated to these has already diminished. The value of homes will be reconciled during these coming years and exacerbate the capital deflation in the western world.   Securities are scary. Particularly mutual funds. When money is managed through the explicit renunciation of accountability, problems occur. Consistently the accountability deficiencies provides greater opportunities to mitigate the accounting for these deficiencies. As time passes the deficiencies only become epidemic and are manifest in the \"owners\" losses.   Bonds and stocks are provided much of the liquidity and support through the consumers and not through the investment in capital goods. Therefore the decline in those will occur in this time frame as well. If the value of the U.S. dollar declines, then these problems will become difficult for many to overcome.   Pensions are a direct result of the health and viability of the marketplace and the companies that employ the majority of the employees. Here I ask if the large organization, with its inherent hierarchy and associated bureaucracy can survive the natural force of self forming teams. Based on my direct research in this area, the answer is an unequivocal no.  Death by a billion pin pricks provides real opportunities to those that are able to use this decline. The only requirement is the capacity to use the tools afforded all in the technologies, the very advanced education of each individual, and investments in capital goods. The analogy I would use would be to ensure that you have a producing asset stuffed in the mattress.  '),(7081,3068590,'23,July,2004','2004-07-23 00:00:00','   Although the Sarbanes Oxeley legislation is based on a definitive need. And the political environment appears to have struck a measured and reasonable solution to the moral and ethical decline in corporate America. The scope of this corporate issue is significant, with its implications bringing into question many fundamental assumptions that the markets are based upon. Nonetheless this legislation is having some internal political dynamics that were unexpected and undefined when the legislation was passed.   The unintended consequences are that a distinct division in the accountability within the organization has formed. With the CEO\'s and CFO\'s having signed all of the financial data in terms of the responsibility and accountability of its accuracy, the individuals responsible have been identified and are limited, in my opinion, in there ability to defend their actions in the future. The consequences of a breach are probable jail time. The evidence necessary to trigger the conviction being the signature of the CEO and CFO on the document.  The remedial methods at the CEO and CFO\'s disposal is an increased role of the internal audit controls and the increased emphasis on the annual audit.  Where are the directors in this process? Are they provided a free pass? The organization certainly has attained a pass, in that they have now been stripped of the responsibility and accountability of the accuracy and integrity of the corporate data that they prepared. That does not suggest they are involved in some criminal conspiracy, however, their diligence in identifying and reporting any inappropriate actions may be diminished in the legislations focus on the key principles in the organization.   When this issue is added to the decline in loyalty of staff within organizations, the focus of employees becomes more personal with their pensions, benefits and income taking priority over any and all organizational issues. I would also extend this to include individuals who have budgetary control are affecting the long term perspective of the organizations development and operations.   This long term issue is manifest through the commitment of the organization in any long term improvement. These committments and objectives are and should be the premier concern of those with budgetary control at this time of globalization. These employees are not that interested in engaging in difficult or controversial organizational tasks, and undertaking any risks that could affect their compensation. Therefore organizations are undertaking limited, if any, long term strategic or operational changes.   The identification and solution to this issue has not appeared in the market at this time. What the ultimate resolution will be is undetermined and is of great concern. This concern stems from the probable amendments to the legislation being impossible due to the fact that any change would be sending the wrong message.   If the enforcement of Sarbanes Oxeley is a political issue, does that impute that enforcement is a direct accountability of the president of the United States?  '),(7082,3068590,'16,July,2004','2004-07-16 00:00:00','   With globalization, supply and demand becomes the method of allocating resources and establishing prices.  There appears to be an advanced market in all resources when the global market place is considered. Allocations of resources and pricing are completed in a rapid manner, and effect all areas of production and consumption.   Competitive threats and opportunities are global in there source and disposition as well. With the effects of innovation and abundancies of resources felt immediately.  Or are they. My concern is that the standard of living experienced in North America and other advanced societies appears to be disproportionately higher then the standard of living realized in other regions of the world. In this efficent market, the Chinese producer has a competitive advantage over their more advanced competition in that they can access lower labor costs more easily.  If the supply of labor in China and India are as they are, does this not affect the labor market on an international basis? Is the standard of living the measure of comparitive advantage of an individuals productivity in the global market?  It therefore should be asked, how will the North American standard of living be sustained with the incremental supply of populations from China and India. Is the supply of human resources in a global market, not realized by the global market?  '),(7083,3068590,'03,July,2004','2004-07-03 00:00:00','   Six time champion. Now that means something.  Another prediction based on the results of the opening prologue. Lance will easily win his sixth championship.   No one, even looks close to touching him. The build in time over the next few weeks may become significant.  '),(7084,3068590,'03,July,2004','2004-07-03 00:00:00','   The speed at which many of the societal changes are happening are truly fascinating. The results of these changes may or may not be seen or realized by many before it is too late.   It would seem that many of these changes are affecting life in North America more then the rest of the world. With the rest of the world gaining at the determinant to the North American society.   An example is the Canadian election. The ability for the country to instill a government such as the liberals, headed by a criminal group of unaccountable and allegedly unknowing administration, boggles the mind. The head of this group, our now famous Prime Minister is a tax evader and drug smuggler. This so that an unproven honest guy doesn\'t muck it up.  Certainly interesting times, you reap what you sow.  '),(7085,3785275,'29,June,2004','2004-06-29 00:00:00','    Men are the reason all of us are here.I mean yes they made us BUT  us women usually get stuck raising those kids.SO when a man says its the mothers that make a man the way he is well thats a lie.Most women dont know how to raise a man so if the man would take care of his kids the way he should maybe we would have less GAY men and more REAL men.Dont get me wrong not all men are bad but the majority id say 80-20 are.They get the girl pregnant and say \"oh i dont want her anymore\" then leave her high and dry.Then when she slap child support on his sorry ass he wanna act a fool.Then be gettin jealous when she find a man better than him that wanna take care of his kids.But see me I just want a man to be with me and except my kids.He aint gotta take care of them because they aint his,and they shouldnt call no other man daddy but at the same time why not when they real daddy aint doin shit but hoein around and talkin shit to u about who u with.So men if u want us to raise a strong man get ur asses in the picture and HELP.Of course it takes a man to raise a man and if u aint there for ur kids u aint a man to begin with.ANd by that I mean get a job support ur kids love aint gonna buy diapers and clothes and food.Love aint gonna put a roof over they head.Yall talk shit about a woman being on welfare well if the man was doing his job we wouldnt have to be.       Then u have these sorry ass men that think its fun to put they hands on a woman and they say they love u baby that aint love.Thats  a pussy.\" I love u thats why I hit you\" What the fuck is that.I know all this from experience.Its horrible and degrading to the good men out there.Then while he putting them hands on u he also out there sticking his piece in every woman that he meet bringing u home STDs and shit then try to blame u for giving it to him.HA  thats what i mean - men are no good.No matter what race men are worthless.All they good for is a quick nut and satisfying a womans needs and if u want they there to mate.Thats it then u can raise that baby alone like u gonna be doing anyway.       For the good brothas out there i mean no harm towards u but damn  show these other brothas how to treat a woman!! All well I wont say all, most women are goddesses and we deserve to be treated like queens.There are some scandolous women out there too i know a few myself but thats the ones that deserve the sorry men.What about us that treat our men like kings  we deserve the same in return but most of the time we dont get it.So fellas if yall got a good woman hold onto her and ladies if yall got a good man hold on to him. '),(7086,3785275,'28,June,2004','2004-06-28 00:00:00','    LOVE????? What exactly is the meaning of love?Does anyone really know? I mean it happens and in my eyes almost ALWAYS ends in heartache,pain and misery.Who wants to feel all that.I damn sho dont.But I find myself falling again and again and enduring that pain over and over?WHY?  I ask myself that so many times but it never comes to me?Dont get me wrong it can be good and make u feel incredible and overwhelming happiness but I just wonder where is the man fa me?Is there even one out there?Im sho there is somewhere BUT will I be able to love that man?ANd if I do is he going to cause me just as much pain as the others? I guess the only real answer is just put my heart out there again and give it a shot.ANd for the meaning of love well i still dont have answer for that. '),(7087,3785275,'14,August,2004','2004-08-14 00:00:00','    ITS BEEN ALMOST 3 YEARS AND YET THE MAN/BOY  THAT HURT ME THE MOST I STILL LOVE I STILL WANT.I DONT GET IT I DONT WANT TO FEEL THIS I HATE HIM I CANT STAND HIM BUT I LOVE HIM SOO MUCH.MY HEART HAS TAKEN SO MUCH FROM HIM.WHY  CAN SOMEONE TELL ME WHY I LOVE THIS MAN THE WAY I DO.A LOVE THIS DEEP IS SUPPOSED TO BE HAPPY AND BLESSED.NOT CRAZY AND DRAMATIC.TEARS IS ALL I KNOW ALL I FEEL IS PAIN.I WANT HAPPINESS WITH THIS MAN AND I KNOW ITS POSSIBLE.WHY AM I WAITING ON GOD TO THAT RIGHT THING AND MAKE HIM GROW UP AND BE A MAN.I MEAN HAVE YOU EVER LOVED SO ONE SO DEEP IT JUST WONT GO AWAY NO MATTER.YALL BOTH MOVE ON BUT STILL U LOVE HIM.I HAVE NEVER FELT LIKE THIS FOR ANY MAN ITS A GOOD FEELING BUT AT THE SAME TIME A VERY BAD SICK FEELING THAT I WANT TO GO AWAY.IF HE LOVES U THE WAY HE CLAIMS WHY WONT HE JUST DO IT.WHAT IS IT THAT THE OTHER FEMALES HAVE THAT U DONT.ALL THIS HAS BEEN KILLING ME INSIDE.IVE GAINED WEIGHT  BECAUSE I JUST DONT CARE.MY DEPRESSION CONSISTS OF EATING  A LOT AND JUST CRYING.LOVE ISNT SUPPOSED TO BE LIKE THIS.OH LORD UP ABOVE ALL I ASK IS TO HELP A SISTA OUT AND MAKE MY LOVE GROW UP AND CHILL.IIGHT YALL IM OUT. '),(7088,3785275,'10,August,2004','2004-08-10 00:00:00','    Why ask why? Good question? In this world today u pretty much have to ask that question WHY? Why is there still prejudice and racism out there.Closed minded people is one of the answers.All the answeres well we might never know.Why does love hurt when its supposed to be good? Just a simple thing that we all go thru and heartbreak unfortunatley follows it most of the time but its life.Why everytime theres a birth we have a death? Who knows that one.Why  ask why?This is just another rough draft.IM practicing guys so be patient with me. '),(7089,3407290,'31,May,2004','2004-05-31 00:00:00','   Back home after a fabulous weekend away with my girlfriend, Angie... I ache in places I didn\'t know I had! ;)  On Friday I didn\'t leave work until 4am on Saturday morning. One of the programs that was meant to take an hour to run (at the most) ended up taking seven and a half hours! Had to wait for it to finish... well, what can you do. Got home at around 5am and saw the sun rise on the way. It was very romantic, which put me in a nice frame of mind for the weekend. Plus a late night meant I\'d be able to stay awake later on Saturday!  The hotel was great; a lovely room, and a shower that had so much pressure you could use it to strip paint! Makes a change from what you usually get in hotels. Had planned to go for a drink when Angie arrived... but that all went out the window when she appeared. As I\'m a gentleman, that\'s all I\'m prepared to say; if you want juicy stuff, go browse for porn! lol  We had such a good time... it\'s great being in love. :D  Steve ;)  '),(7090,3407290,'26,May,2004','2004-05-26 00:00:00','   Ahhh... just booked for my first weekend away with the girlfriend... can\'t wait! We\'re going away somewhere so we can be together with no distractions, just me and her, spending lots of time in each other\'s company. I\'m going to try my best to make it a wonderfully romantic weekend. I hope I can, because I love her so much.   Steve ;)  '),(7091,3407290,'24,May,2004','2004-05-24 00:00:00','   Ok, I thought I\'d put a few of my favourite links and stuff on here for you all to check out.   Acme.com  This site has a lot of cool stuff including a car number plate maker, and a love hearts maker (those little sweets?) and you can customise the messages on them. Use them as signatures in forums, etc. My personal favourite is the  Dymo Tape labelmaker .   Mousebreaker.com  and  Miniclip.com  Two sites packed with great Flash games to play, and totally free too. I love the  Table Tennis  game, even though it\'s pretty tough!   DTI Keeper  I always had a problem when I resized my PC screen for one reason or another, and when installing new drivers, because it invariably screwed up the layout of all my icons on the desktop. But I discovered this fantastic little shareware item, made by  DragonOne Software . This nifty little application memorises desktop layouts for you and saves them, so you can restore them to where they were quickly and easily. I\'ve found it to be invaluable.   Enjoy!  Steve ;)  '),(7092,3407290,'24,May,2004','2004-05-24 00:00:00','   Yep, it\'s Monday again... well, at least I\'m working from home today, and don\'t have to face the traffic. But month end is almost here, which means long hours in the office :-(  Had a pretty good weekend though. Spent most of it either on the phone or on MSN to my girlfriend... she continues to amaze me with the way she makes me feel. It\'s soooo good, can\'t wait for us to be together again.  Arrrgh just remembered a job I have to do this afternoon, preparing reports for upload to the intranet... possibly the single worst part of my job. Long process on a system which continually crashes... who bought this shite software anyway? Oh yes, that\'s right... the \"bosses\" (who don\'t have to use it, and don\'t have to sit in the office until 4am each month like we do, trying to close all the subsystems)  Ah well. Pay day tomorrow :-D  Steve ;)  '),(7093,3407290,'23,May,2004','2004-05-23 00:00:00','   Wow! What a race! Fantastic stuff, at last we see some proper F1 for the first time this season!  This was always gonna be an incident packed race. The first start was aborted after the parade lap, when Olivier Panis stalled his Toyota on the grid (out of position). On the second parade lap, he did it again, and was pushed into the pit lane in time so that the race could begin.  The lights went out, and the Renaults made perfect starts, with Fernando Alonso passing Jenson Button into Ste Devote, but the fastest start was made by Button\'s teammate Takuma Sato... he stormed into 4th place, nudging front wheels with Michael Schumacher in the process. But all was not well with Sato\'s BAR... you could see on the first couple of laps, there were wisps of smoke coming from the Honda engine... was it just an overfilled oil tank? or something more serious?  The answer came before too long. With a queue of cars behind him heading into Tabac, the BAR Honda let go in spectacular fashion, churning out thick white smoke. Sato got the car off the racing line, and Kimi Raikkonen shot past him. But this area of the track was now enveloped in smoke, and visibility was zero... and the inevitable happened. Giancarlo Fisichella came flying round at about 90mph straight into the back of David Coulthard. The Sauber flipped over, narrowly missing a couple of marshals, and Coulthard lost his rear wing, ending his race. Fisichella was unhurt, thankfully.  All this meant that the safety car was brought out, bunching up the field. After five laps the racing resumed... the Renaults maintained their positions, keeping Button, Raikkonen and Michael Schumacher behind them. Juan Pablo Montoya managed to steal 6th from Rubens Barrichello at the restart.  At the first round of pitstops, Button was first in, followed by the two Renaults and the McLaren. Michael Schumacher now had a clear track in front of him, and he used it to good effect, putting in some stunning laps, before stopping a few laps later. This put him up to thrid, right behind the two Renaults.  But Alonso decided to try and lap Ralf Schumacher on the outside line through the tunnel... a bad move, which saw him understeer into the barrier, wrecking his car and his chance of a win. The safety car was brought out once more while the debris was cleared. Trulli and Button immediately pitted to take advantage, but Michael Schumacher was not so lucky; he was caught behind the safety car and couldn\'t make a pitstop without compromising his position. This was a major blow to his chances of a win.  Then a bizarre incident ended his chances completely... while still under the safety car, they came into the tunnel. Michael accelerated and them braked hard, presumably to warm up his brakes - Montoya, who was behind him and seemed to be going way too fast, had to brake and swerve right to avoid hitting him. He came alongside, and nudged Michael, who slid into a barrier, losing his nosecone and his left front wheel.   So, who was going to win? Only Trulli and Button looked like they had any chance, and the last 15 laps or so were fascinating... Button was 8 seconds behind, but closing, and was right on Trulli\'s rear wing on the last lap. But Trulli had driven a masterful race, and wasn\'t about to make a mistake. He coolly cruised to victory in Monaco, the first Italian to do so in 22 years. It was also his first Grand Prix win, and even though Michael Schumacher had retired from the race, this result was not in doubt, as Trulli had maintained the upper hand from lap 1.   Button was second, and Rubens Barrichello, who had a rear suspension problem on his car, limped home in a distant 3rd place, ahead of Montoya\'s Williams, Felipe Massa\'s Sauber, and Cristiano Da Matta\'s Toyota. Seventh place, and two valuable points for Jordan, went to Nick Heidfeld, and the final point for 8th place when to Olivier Panis in the other Toyota.  Congratulations to Renault and Jarno Trulli on a brilliant weekend. All in all, a fantastic race!  Steve ;)  '),(7094,3407290,'22,May,2004','2004-05-22 00:00:00','   Quite looking forward to this race... with Michael off the front row, and two fast-starting Renaults ahead of him, it could be an end to his dominance this season. I know my girlfriend will kill me for saying this, but I hope he does badly... ;)  FA Cup yesterday didn\'t go too well... 3-0 to Man United in the end. Felt sorry for Millwall, the result didn\'t really reflect how well they played. And one of the goals was offside. C\'est la vie.  My web browsing continues... check out  http://www.themakeupgallery.info/  - this site is great, you can see the amazing transformations people go through to play parts on TV and in movies. Fascinating stuff.  Steve ;)  '),(7095,3407290,'22,May,2004','2004-05-22 00:00:00','   Woke up this morning just as the phone rang... it was my girlfriend. I can\'t think of a nicer way to wake up... oh no, hold on, I can... but she\'d be involved in that, too ;-)  F1 qualifying has just finished, Michael is 4th on the grid... the two Renaults and Jenson\'s BAR are ahead of him. I suspect that Jenson is carrying more fuel as well, as he was quite slow in sector 2. Maybe Michael won\'t beat Mansell\'s record by winning the first six races after all... fingers crossed.  20 mins to go until the FA Cup Final between Millwall and Man United. I have to support Millwall, for a couple of reasons... everyone hates Man United, and Millwall are the underdogs... and as a Chelsea fan, I would love to see Dennis Wise raising the trophy. And Alex Feguson looking miserable... always good! :-D  Come on Millwall!!  Steve ;)  '),(7096,3407290,'21,May,2004','2004-05-21 00:00:00','   Into gaming? You might wanna check out  Hamsterserver.com , we\'re a gaming group and we have a server that runs various online games. We\'re all hamsters. Get on the forums and give us a shout... we don\'t bite :-)  It\'s Friday at last... It\'s great working from home. I don\'t have to face the usual 45 mile drive back from Gatwick (and of course on a Friday it\'s always worse, as the roads I take to get home are the same ones everyone uses to head for France...)  Still not sure what to do this weekend. Might nip to the Vauxhall garage in the morning and get a new external air temperature sensor for my Cavalier... it\'s been showing the temperature as \"--.--\" ever since I got the car. It would be nice to know what the real temperature is!  I just read back what I typed... GOD how boring! I really must get a life...  Steve ;)  '),(7097,3407290,'21,May,2004','2004-05-21 00:00:00','   Yep, I\'ve been bitten. I\'ve got the lurrrve bug! :-)  There\'s a girl who I\'ve known for a couple of years now, and recently it\'s become more than just friendship... she\'s absolutely gorgeous, and I think the world of her. She\'s so beautiful and sexy, and she affects me in ways that I\'ve never felt before... it\'s just incredible!  She knows exactly how I feel... and she has strong feelings for me too. We\'ve been out on dates a couple of times, and it\'s gone  really  well. But for various reasons the situation is a little complex, so we need to take things slowly.   But at the moment, I\'m about as happy in my life as I\'ve ever been.  Steve ;)  '),(7098,3407290,'21,May,2004','2004-05-21 00:00:00','   Ok, I seem to have set things up properly now... didn\'t take too long. It\'s a breeze! if only everything in life was this simple... *sigh*  Aaaanyway... my plans for the weekend? Well, none really. Unless someone lovely calls and says she needs to see me! ;-)  The F1 race at Monaco is on, but I expect it to be boring. This season seems to be a bit lacking, and Michael Schumacher seems to be running away with the title. In fact, bookmakers Paddy Power have paid out already to people who bet on MS to win the title this year! It\'s mad. Come on Bernie, get the sport sorted so it\'s worth watching again...   God, this sounds all so pessimistic! aaargh! Time for some good news. :-)  My flat in Aberdeen is under offer! At last. So, providing the valuation comes up to the Â£43k offered, it\'ll be sold. I can\'t see it coming in below that amount. Fingers crossed... should free up a sizeable chunk of my income that I\'ve been paying out on it the last few months!  That\'s it for today. Tune in later for more info! Will something exciting happen? Is the world gonna explode? Probably not. But tune in anyway... lol  Steve ;)  '),(7099,3407290,'21,May,2004','2004-05-21 00:00:00','   Well, this is my blog... and this is my first post to it. Still gotta work out how to get it all set up properly, layout etc... so I\'d better get creative! here goes...  Steve ;)  '),(7100,4188274,'12,August,2004','2004-08-12 00:00:00','   Editors\' meeting went fairly well today. It was productive, at least. I\'m still letting myself get weak and giving in to Trevor, though, and Fred called me on it. It just seems easier to let him have his way sometimes, you know? Gotta be strong, Brooke, gotta be strong.       You know what? Family is a beautiful thing. It truly is. If you don\'t spend time with your family, you are missing out. They have wonderful stories to tell! I helped my aunt move tonight, so I got some quality time in with her and my grandma. I love them so much. My grandma told me about the days when she used to go to TWU, back when they were VERY strict (ever see Mona Lisa Smile? JUST like that! strict curfews, no guys on campus, dresses all the time, etc.). And then my aunt (she\'s 31) told me about when SHE went to TWU, and you could wear pjs to class and it was co-ed (not that many guys go there), and the guidelines weren\'t nearly as strict.. it was so interesting! Of course, now I\'m exhausted because of all the labor from moving, but it was worth it.       [current]     [mood]  upbeat, but incredibly worn out    [attire]  UIL Regionals t-shirt and purple NYU pj pants    [song]  \"Quiet Please, There\'s a Lady Onstage\" from  The Boy From Oz     [wish]  that I didn\'t have to clean tomorrow    [convos]  Kristine on AIM    [thoughts]  man, I\'m really sleepy    [future plans]  Ranger game tomorrow night with my dad and Rhonda! woot!  '),(7101,4188274,'10,August,2004','2004-08-10 00:00:00','   First post of my new blog! I got tired of Xanga. I like this better, it allows more room for creativity, in my opinion. Then again, who am I to talk about creativity? Kristine did my HTML, lol. (Which I love, btw.. thanks girl!)       Let\'s see, news: My drivers ed stuff came in the mail today (I\'ve been waiting for it for 8 months now), so I get my permit in a couple of days! Woot! I remember when I was little, and I always imagined getting my permit on the morning of my sixteenth birthday... yeah so much for that, lol. Hey, better late than never, right?      I had a bit of a nervous breakdown last night. It wasn\'t very pretty, but if you are one of those who I talked to about it, I\'d just like to report that things are going more smoothly now. I may be loquacious (SAT word) most of the time, but it\'s different when I\'m with my mom. I can be very blunt, and I let my mouth run away from me. That\'s something I need to work on. New School Year resolution: be a better daughter.       I guess that\'s it for me this evening. Ciao.       [current]     [mood]  somewhat ambitious, but tired    [attire]  Renaissance t-shirt and blue plaid pj pants    [song]  \"St. Elmo\'s Fire (Man in Motion)\" by John Parr    [wish]  a hot fudge brownie sundae with peppermint ice cream from Braum\'s    [convos]  Kristine, KD, and Sean on AIM    [thoughts]  I miss Ben...    [future plans]  editors\' meeting tomorrow at noon  '),(7102,4188274,'10,August,2004','2004-08-10 00:00:00','   Jo Kristine Lauderdale, how I love thee   I wanted a new pretty blog, and you gave that to me   Your big and friendly heart had a vision, I could tell And your expert knowledge of codes and scripts wrote me HTML   I was lost and you rescued me   Jo Kristine Lauderdale, how I love thee.  '),(7103,3505165,'31,May,2004','2004-05-31 00:00:00','   I just sorta thought that it would be appropriate to include a blog about how absolutely great lindsay is.         Okay thats about it...............  '),(7104,3505165,'31,May,2004','2004-05-31 00:00:00','   Freedom- the ability to do what you want, when you want. But are we really free?  Of course not.........dont be silly. If we were free would we have any laws. Would we have to have prisons. Would we have any government at all.  To be totally free, we would have to rid ourselves of all establishments of government, which is an anarchy.  Now i dont know about you guys out there, but I personally prefer to have it so that a killer can  NOT  come into my house at night and kill me.  Sooooooooooo............I guess what I\'m really trying to say is:  \"All you poser anarchists out there are really annoying and have no idea what you are saying.\"  '),(7105,3505165,'31,May,2004','2004-05-31 00:00:00','   hey everyone,  How ya all doin\'   I just started this blog, isnt it cool. I think so. What\'s that. You think it stinks. Well you know what, I dont care what u think, ur ugly, so there.  Im not paranoid............he he he...............  '),(7106,3505165,'25,June,2004','2004-06-25 00:00:00','   I had a Physical today. Lots of fun; nurses attaching things to my arms, doctor hitting my knees with his thingy, and shining lights into my eyes.  When we get screened for scoliosis in school each year, they always say that I have it and suggest that I have my doctor look at it. Well, this time, the doctor told me to do all the regular stuff (bend down, touch your toes, stand up straight and spin around in circles) and he said that I have mild scoliosis but it is due to the fact that one of my legs is shorter than the other. Then he called my dad in, and was pointing out how my hips are tilted, and the line at the fold of skin at the bottom of my knee-pit (that\'s the back of your knee for those of you without my specialized vocabulary) is at slightly different heights on each leg becuase of the difference in lengths.  Well, now at least I can say that I am one of the few people I know that has legs of different lengths.   I FEEL SOOOO SPECIAL   This is \"Different-Leg-Length Boy\" signing off.  '),(7107,3505165,'21,June,2004','2004-06-21 00:00:00','   I had Summer Baseball practice today, for 2 hours. It was so pointless, becuase I got the ball probably only 2 times that entire scrimmage. And then I stuck out every single time I got up to bat, which was only twice, and that is pretty sad, because we were facing a thirteen-year old team.  And i missed the fake, 8th grade girls softball all-star game.  *Tears and sadness*  I really felt bad for Chris today, and I\'ve decided that Kelly is on my list of people who I strongly dislike, for the time being. She just should not have said yes if she did not like him in the first place.   Arrrrgggghhhhhh!!!!!!!   Oh and Chris and I have only told one person who Conor likes. But it was not even really my fault. Chris was really the one who told Maria.  Okay fine, so it was partly my fault, but that doesn\'t really matter.  Don\'t get lost now.  '),(7108,3505165,'19,June,2004','2004-06-19 00:00:00','   The 8th Grade Social was last night.  It was pretty fun. It was cool seeing all the girls wearing nice dresses and things and all the boys wearing nice shirts. And I got a lot of practice doing my Chris/Indianna Jones/Harrison Ford half-smiley thing. I think i look pretty good doing it. Hahaha.  Loula and the other peoples kept trying to get me and Lindsay to dance together, and when we finally were, Mr. Phelan came over and told us that we needed to separate a little and not be so close to each other.  They had free slush, I only had 5 of them.  At the end, as the last songs were playing, some of the girls started crying. But by the time we were heading out, I think pretty much every single girl was in tears. Brianna was practically hysterical. When I first saw and heard her crying, I thought she was just pretending becuase she was just bawling. Then she came over and gave me a hug and said that she would really miss me, and I realized that she as acutally seriously crying.  I really just dont understand girls. When we were standing outside, me and ian and billy were standing together, surrounded by all these crying girls, and we were just talking about how we would be seeing each other in two whole days. I can sort of understand it if people start crying on Wednesday, when some people are actually going away, adn we might never really see them again, but not on a Friday when we still have 3 days of school left.  But, girls are just like that. Sometimes I just do something that I find totally fine, and then I realize that girls want to be doing something else instead. Like, I think it was the last song, and they announced that if you still wanted a slush you should go and get one, so I asked Lindsay if she wanted a slush, and she was said sure. So when went and got slush (I got two) and were eating them and I then Lindsay said something like, \"Peter, can we please just finish this last dance. I\'m a girl and care about this mushy stuff.\"  and i didnt even realize that a girl would want to do that, probably partly bucause I really like slush and and wouldnt miss a chance to get some more for free.  Now this might make no sense to anyone, bucause I just went back and read it and it barely makes sense to me, but I think you might sort of get what I am talking about.  This is Chicken Man, signing off.  ps. All the girls looked really pretty in dresses, except for the weird uglyish ones, but I wont get into that.  pps. Michelle invented the best dance move ever, the \"Daddy, where are you\" move.  ppps. Oh and Chris came up with one of the worst dance things ever, you all stand in a line and turn your arms around each other.  pppps. Congo lines are fun, especially during songs that shouldnt have them in the first place  ppppps. Lindsay looked really pretty in her skirt and shirt.  '),(7109,3505165,'14,June,2004','2004-06-14 00:00:00','   Hey guys,  Well school gets out in.....uhhh........oh yeah.... 6 and a half days!!!!!!!!!!   yay!!!!!!!!  I can not wait until summer, I am so bored of school. But then again, I always am so I guess that\'s no big change.  If you could be any condiment in the world what would it be?????  hmmm.... I think I\'d choose guacomole, \'cause I\'m probably the least likely to get eatin that way.  If you could be any meat dish, what would it be?  I would be a hot dog, becuase who knows what really in it.  If you could be any amount of money what would it be?  $42.00  '),(7110,3505165,'07,June,2004','2004-06-07 00:00:00','   Ronald Reagan died this weekend.  I was driving in my car with my dad to Luke\'s house to hang out and we were listenting to a thing about him on NPR. It was many different sound clips of him, making speeches, making announcements and whatnot.  And all through the whole saddening, tear-bringing thing, the only thing I could think about was about how long they must of had this ready for when he finally kicked the bucket.  Now, does being so, well, I don\'t really know......I guess.......cynical, make me a bad person. Maybe being so cynical is just my way of overcoming the sadness that I should be feeling since of of our most famous presidents has just died. Or maybe the sadness just doesn\'t affect me because I wasn\'t alive while he was president.  I remember seeing footage of JFK\'s assassination, and remember thinking how sad it must of been to be alive for that. That\'s one of those events where you will always remember what was happening that day when you found out.  Like, I can\'t really remember what happened for the first part of the day, but I can recall like it was yesterday what happened on 9/11 when we found out about the terrorist attacks. There were rumors going through the school that something had happened in New York, because I guess that some teachers had let it slip, or had shown their students what was happening on TV. But at the very end of the day in homeroom, I remember they made a short, puzzling announcement to the effect of:  \"Earlier today some planes crashed into buildings in New York and Washington D.C.\"  And that was pretty much it. They gave us no information, absolutely nothing besides the facts that some planes had crashed. That pretty much left us to make our own ideas. I can remember Jimmy Fahey saying on the bus that it was probably some terrorists who hated the U.S., and I was saying that he was wrong and that it would turn out to be some crazy Americans who didn\'t like the government or something.  I don\'t really think that any of us understood the full extent of what had happened that day until we got home and just sat there, in front of the TV, for hours, just watching the news coverage.  They kept showing the planes crashing into the towers again and again, non-stop, just imprinting it into your brain.  The next day in Writing class, I remember that Ms. Connell made us write in our journals our reactions to what had happened the day before. All I was able to write was pretty much the fact that 2 planes had crashed into the World Trade Towers, one into the Pentagon, and one that might of been on it\'s way to the White House or Camp David. But I distinctly rememeber that Harry Edwards wrote a poem about it. I can\'t remember how any of it went except for the very last two lines. But those lines will probably stick with me for the rest of my life. It went:    Yesterday there was no fun, Yesterday was 9-1-1     I don\'t know why it was that that snagged in my brain, but it was, and that it was I will always remember.  '),(7111,3505165,'01,June,2004','2004-06-01 00:00:00','   what is the use of adults they teach you to hate they force you to wait they tell you what to think  what is the use of school it teaches you little it makes you quibble it creates violent killers  what is the use of teachers they yell at your mistakes they force you to take each and every one of their opinions  what is the use of society it gives you pointless goals makes you pay all the tolls and destroys all individualism  what is the use of it all its not use to me that is clear to see hermits live the best life of all  '),(7112,3505165,'01,June,2004','2004-06-01 00:00:00','   What is the point of school?  Okay fine.....I can sort of see that school is supposed to teach you stuff that you can learn later on, but I learn practically nothing in the majority of my classes. For instance, who didn\'t already know that volcanoes spew out lava, which is pretty much the extent of what Ms. Bertrand teaches us. That class is soooooooo boring. And what about math class; they claim that they level the classes, but it is really just split into the stupid kids, adn everyone else. Why not just have an actual class with actual smart kids who actually know stuff. Mrs. Mikels has to give me these stupid worksheets so that \"I am challenged in class.\"  Sorry about my ranting but I would just like to point out the fact that  I HAVE LEARNED PRACTICALLY NOTHING IN MANY OF MY CLASSES AND IT\'S BORING AS HECK .    p.s. how come we are the only class who hasn\'t even gotten to the point in Root\'s where his leg gets cut off. We never really watch it anymore.  '),(7113,3505165,'23,July,2004','2004-07-23 00:00:00','   If you\'ve read the DaVinci Code (which pretty much everyone who reads this has) then try this. It\'s cool and interesting. It\'s a webquesty thing. Just do what it says, it\'s fun. Really.  Just click  urlLink here .  And if you have a copy of the DaVinci Code somewhere around you, or within a couple of yards or so, try this. It\'s a bit less fun and more hardish. But it\'s pretty good.  To go there  urlLink click here, or else.......      P.S. As you may be able to see...I just found out how to do the linky things...they\'re fun...really...I\'m serious.   urlLink See, they\'re all cool and blue and awesome    P.P.S. Linds, click  ^  on  ^  the  ^  link  ^  above.....I think i should use some of those sometime.  ; )  '),(7114,3505165,'19,July,2004','2004-07-19 00:00:00','   What Makes You Sexy?  by  urlLink  eva71        Name/NickName           Gender           Sexy Body Part Is    Nothing      Special Talents Are    Absolutely Nothing             Created with the ORIGINAL  urlLink  MemeGen  !  '),(7115,3505165,'12,July,2004','2004-07-12 00:00:00','   Beatles - She Loves You                 She loves you, yeah, yeah, yeah She loves you, yeah, yeah, yeah She loves you, yeah, yeah, yeah, yeah  You think you lost your love, When I saw her yesterday. Itï¿½s you sheï¿½s thinking of And she told me what to say. She says she loves you And you know that canï¿½t be bad. Yes, she loves you And you know you should be glad. ooh!  She said you were to know That she almost lost her mind. And now she says she knows Youï¿½re not the hurting kind. She says she loves you And you know that canï¿½t be bad. Yes, she loves you And you know you should be glad. ooh!  She loves you, yeah, yeah, yeah She loves you, yeah, yeah, yeah And with a love like that You know you should be glad.  And now itï¿½s up to you, I think itï¿½s only fair, If I should hurt you, too, Apologize to her Because she loves you And you know that canï¿½t be bad. Yes, she loves you And you know you should be glad. ooh!  She loves you, yeah, yeah, yeah She loves you, yeah, yeah, yeah And with a love like that You know you should be glad.    Hope you liked it!!!!  '),(7116,3505165,'12,July,2004','2004-07-12 00:00:00','   What Makes You Sexy?  by  urlLink  eva71        Name/NickName           Gender           Sexy Body Part Is    Nothing      Special Talents Are    Looking Innocent             Created with the ORIGINAL  urlLink  MemeGen  !  '),(7117,3759312,'30,June,2004','2004-06-30 00:00:00','   sigh..wat a disappointment =( went Maju Camp for IPPT today. guess wat? haha..lol, i fail. slacked for too long, neva train, should have guessed my failure for today. anyway, the PTI encik there was fun and frenly with his long cane, was a master warrant, chat with me in hokkien. not like those evil PTI i have came across. hahaha RT coming soon, wish me luck ba. >_< '),(7118,3759312,'29,June,2004','2004-06-29 00:00:00','   both my online games, gunbound å¦å…‹å®è´ and fairyland ç«¥è¯ server got hacked. sianz man... 2 days already... still cannot play. haiz... time to search for new online games, or maybe not, cuz i got tonnes of PS2 games haven touch or completed yet. arc of lad, resident evil voronica and fatal frame 2 all play half way, =( no time sia!! '),(7119,3759312,'26,June,2004','2004-06-26 00:00:00','   yum yum..the hainanese chicken rice at pricep street was delicious. they also have other dishes. some of my fav are beancurd hotplate, fried porkchop with potato, kanggong with balajang and not forgetting my fried cai po egg. shinz recommended, go have a try, is at pricep street right till the end, chin chin eating place =) '),(7120,3759312,'26,June,2004','2004-06-26 00:00:00','   sigh =( ..went to dreamcars show yesterday at expo..kind of disappointed. not much cars really. and the best thing, my fren digi cam went out of batt...and not much really lots of race queen as they say in newspaper >_< sad sad man.. '),(7121,3759312,'15,July,2004','2004-07-15 00:00:00','   juz started on my new online game darkeden 2 days ago..yes vampire for sure ^^ time to wipe out all the slayers. but i\'m weak haha lol.. me newbie vampire only. yipee..gotten my 2 teeth rank today. sigh.. 3 teeth rank seem so far away... arg.. '),(7122,3759312,'13,July,2004','2004-07-13 00:00:00','   my com has been down for 5 days due to serious viruses. damn tat donk and korgo virus!! arg..my com has been formatted and clear of viruses now. yipee, can surf back chinese site already, and no more irriating script prob. ^^ woohoo.. '),(7123,3759312,'04,July,2004','2004-07-04 00:00:00','   i juz celebrate my 24th birthday yesterday. it was a fun and enjoyable day. thanks for all the friends i know tat came to celebrate with me for the pass 2 days. ^^ sigh.. sad man, i\'m older by 1 yr again. old man le.. hohoho.. '),(7124,3759312,'16,August,2004','2004-08-16 00:00:00','   2 days ago was liling bday, we went out and have fun, but the director aka fatimah neva come. kind of disappointed. there are some weird problems happening, on a girl, i shall not say it out. the 2 person who were with me, u know who u r, maybe we can help to slove this girl problem. seeing her with lots of probs juz make mine heart pain =( '),(7125,3759312,'07,August,2004','2004-08-07 00:00:00','   nowadays been busy with work, hardly got time for my darkeden game. i\'m have been stuck in lvl 81 for quite some time. still got a long way to become baronet...sigh...=( can someone give me some free lvl..lol '),(7126,3759312,'07,August,2004','2004-08-07 00:00:00','   awww.. how did i do it?? my highest score was only 117. wth is wrong with me?? sigh, lack of training..bloody heck still can\'t understand why...so ashame, my friend who juz started on bowling today..even trash me in 2nd match. zzzz.... '),(7127,3759312,'03,August,2004','2004-08-03 00:00:00','   went to watch the village with my pri sch fren liling and shijie, erm..wat to say..not bad lah...but too boring loh, cuz the show was actually abt a lie made by the village elders. and guess wat?  somewhere near the ending, when one of the ranger went to take medicine, the ranger superior was actually mr bean.. o_O or did i see wrongly?? anyway looking forward for more horror movies ^^. '),(7128,3759312,'03,August,2004','2004-08-03 00:00:00','   wow so happy..get to see my pri sch friends yesterday. qiuyan, liling, siyun, siling and shijie. all still look the same no change ^^. so nice..we went back to our pri sch front gate, sit down and chit chat there..but too bad its closed, we can\'t get in and take photos. well..quite nice ^-^ '),(7129,4287218,'22,August,2004','2004-08-22 00:00:00','   And my life continues to stay full of weirdness. Last night I was pretty bored, I layed on the couch watching tv with my laptop on my lap. Then I got another email from CC. I am glad we are still friends. I missed him and Kenyon more than I realized. Kenyon can crawl already and say dada. I missed all that =(. In the end I fell asleep in bed around 3 am, while watching a movie. I kept waking up in the middle of the night feeling sick, running to the bathroom throwing up. I hardly slept again.  Today is the big day though, I hope the recording goes well for the radioshow. The intro\'s I got from Napalm are soooo tight! I am so happy with them. Chamelleon is making an intro for me also. He is the best. Tried to stay up to have it done for me, but he was too sleepy, did make me a shout out I can throw in the show. His song got the nr 9 spot on the top 10 request list on Royal Radio. This morning, he told me something, nobody ever told me before, not even a loved one. That meant a lot to me. He is such a good person, and I am so happy I got someone to share the same passion for music with. He always genuinely smiles with me whenever I am all excited about some music stuff. I defenitely have to shout him out tonight at the show.  I think I am done preparing everything for the show tonight so I might catch me a nap before its time to bounce. Dont forget to bring the PG cd\'s. Mah homies are doin da damn thang in NYC! So proud of them also.  Music was, is and will always remain my first love. It always makes me happy, keeps me happy, never disappoints me or lets me down. I love it. And thats an understatement. =) '),(7130,4287218,'21,August,2004','2004-08-21 00:00:00','   I just woke up. I actually slept for a change. I was watching a movie last night but I fell asleep, that hasnt happened to me in a while. Me and Devon had a huge fight just now on the phone. I hung up. Then a few later, he hit me back tellin me he sorry for hurting me and all. We had a real good talk about how this situation is. We kinda both figured out we probably dont match. I dont know, I guess its done. Blah.  The last thing I want is another insecurity about something. I already have enough insecurities to deal with, especially my health now. I have to go back to the hospital on Sept. 2nd. Do another biopsy to see if the lower tissue has cells that are effected too. Let\'s hope not. A few years ago I went through the same thing, alone. Now I still go through this, Alone. Nobody really cares, or maybe dont understand when I tell them. So I just dont tell them anymore. It just hurts more when you know that people know, but there is no support really. I am used to dealing with this alone so it doesnt really make a difference.  Funny how right now, after a good night sleep, my state of mind is pretty peaceful though. I have an important radioshow tomorrow so I am glad I am focused on what\'s important at the moment. I have to record the show and send it to Juize FM. They are interested in my (our, DJ Dynamite and myself) show. Let\'s hope they like the platform and decide to put the show on their station. That would be half a dream come true for me. Lord knows I worked hard enough for it...... '),(7131,4287218,'20,August,2004','2004-08-20 00:00:00','   I don\'t know what happened anymore. Last night I was real frustrated, usually a good night\'s sleep helps, but sleep is something I rarely get to enjoy nowadays. I set up this blog stuff to write stuff off of me, normally I start writing songs, but it seems like I am \"singing\" the same old songs over and over, so I try a different approach to deal with these feelings. Who knows, this might be a new therapeutic way of healing or maybe I should say \"improving\" the quality of my life. *shrugs*  Last night before I went to sleep (tried to that is) I talked to Tarik, a good friend. Last time we saw eachother was last month in Brooklyn. Now he moved from DC to St Maarten and spent his birthday last sunday on the beach there. I was so jealous. I am glad he had fun though. I told him about all the good things and bad things in my life going on right now. Its such an emotional rollercoaster really. Its like I can\'t even enjoy the good things and simply be \"happy\" about it cause the bad things are downing it so bad. So much, I don\'t even know how to feel anymore these days.  I always been good at business stuff, but personal stuff......blah. I just bury myself in the work I got and hopefully I dont have to deal with a personal life this way. Work keeps me happy right now. Music too as always. I just hope my health doesnt get in the way of the things I love to do. '),(7132,2977760,'30,July,2004','2004-07-30 00:00:00','   *yawn*  Can I borrow someones bed?  I don\'t want to sleep here tonight.      BETH '),(7133,2977760,'29,July,2004','2004-07-29 00:00:00','    '),(7134,2977760,'29,July,2004','2004-07-29 00:00:00','   I am absolutely and completely bored. I actually hope Beth sends me 10 boxes of JollyJohn\'s free condoms. Atleast then I could enjoy some humor. Oh My God. My mom rented another movie and left it here. Guess when it\'s do? Today. Guess what time it is? 7:01 p.m. I have to rush upstairs to grandma and tell her we need to hurry to take this movie back. I  hate  when my mom does this. Grandma\'s going to be pissed! If mom went to work Tuesday in the shape she was in I bet a dollar they fired her. It just discusts me to no end. Kill Bill is a good movie. I want to see Kill Bill volume 2 now...  &nbsp; '),(7135,2977760,'28,July,2004','2004-07-28 00:00:00','   As I lay in my bed with the covers all around me, comfortable as can be and drifting off to sleep once again my kind mother comes barreling into the house. \"You had better be ready! I mean it!\" I jump up angrily disturbed from my cozy little half-sleep and she continues with her rant\"Where are you? I mean it! You had better be ready!\" As you can imagine I am confused as hell about what the hell she\'s going on about so I&nbsp;ask &nbsp;irritably \"What the hell are you talking about?\" And she respondes \"Your dentist appointment, why the hell did grandma come over yesterday if she wasn\'t going to tell you when your appointment was?\" So, I get up somewhat forcebly thinking to myself \'she asked you when my dentist appointment was and you responded \"August 28th\", so, either grandma didn\'t hear you right,which I seriously doubt, or, you were so high you didn\'t know what month it was. My bet goes on the latter.\' Although, I didn\'t say anything of that sort out loud because I was way too tired to&nbsp;start an argument. I got up and got dressed and grabbed my tooth brush and comb and shoes and did all that in the car on my way to the dentist. I do not recommend brushing your teeth in the car, by the way I just&nbsp;remembered I left my tooth brush in the car. Damnit. That\'s cool cause the dentist&nbsp;gave a new tooth brush, some floss, and a little tobe of Crest toothpaste :-D I feel happy, I got presents today. Anyways, I got to the&nbsp;dentist and sat in the scary dentist chair and looked at all the scary dentist equipment and kept thinking \'Seneika is behind this.&nbsp;Seneika has got these dentists under her paycheck. They are going to do all kinds of evil Nazi experiments on me. They hate Jews... Wait, I\'m not jewish.\'&nbsp;I did think&nbsp;Seneika was paying the Nazi dentists under the table, all ... I did think they were&nbsp;going to do&nbsp;mad-evil-scientific-torture experiments on me and then I got to thinking of evil and evil reminded me of mr.McDonnel and McDonnel reminded me of a lesson to do with the Holocaust he had taught.. Or it may have been mrs.Beck-I don\'t know the point is I was reminded of the Holocaust and then I got scared cause I started thinking the dentists were Nazis and see how my mind wonders? It just got worse from there. At first she stuck some plastic thing in my mouth and arranged it in it way that when I had to bite down on it(I was required to bite down on it)it hurt and stabbed my cheek and tongue, I think she was taking x-rays but the whole while I kept thinking\'Krissy!! Come save me! Make Andrew come get me! Beth! G! Chelsea I want out of this torture seat!\'. Then she cleaned my teeth. Okay, the only bad part about that is the paste stuff gets on my lips and dries them out and I don\'t have any chapstick cause I\'m too poor to buy chapstick. After that she used some kind of sharp metal object that she referred to as a metal toothpick to scrap my teeth, I haven\'t the foggiest idea why, but it scared me cause it doesn\'t feel right to have a metal object scrap between and across your teeth. After that she flossed my teeth. That was the weirdest thing. I\'m not used to having peoples fingers all up in my mouth and she was ALL up in my mouth. Okay, last thing she did was fill this blue thing that you put in your mouth and bite down on with this purple flouride crap just for my teeth. That stuff wasn\'t very good and I had to bite down on it for about 10 minutes and you know, after awhile your jaw gets tired and the sucky-thing wasn\'t suckying in the place it should\'ve because I had drool runnin\' down my&nbsp;chin&nbsp;. Nasty.&nbsp;Again I thought \'save me all my friends!\' and actually&nbsp;made a noise but&nbsp;nobody heard because I had that thing in my mouth, plus noone was in there at that time.&nbsp;Then after she took out the nasty blue thing I waited for the doctor cause the x-rays of my teeth had gotten back. I waited in there for quit awhile but I didn\'t care because the dentist woman let me play with the sucky-thing. Doctor came and told me my wisdom teeth are coming in but that was the least of my worries. I have scrabble teeth. If you do not know what that is I will tell you even though I don\'t know what it is but I\'ll tell you what\'s wrong with my teeth. My teeth are fairly straight it\'s not like I need braces or anything but I have ONE tooth that\'s kind of crooked, not very much but you cam tell. I thought \'oh-well big deal it\'s just ONE crooked tooth, that\'s okay\' but the doctor said it wasn\'t crooked. Meaning it would have been totally straight IF it wasn\'t for my one canine baby tooth right beside it. I have ONE baby tooth left. One. And it seems as though the big tooth that should have been grown out by now is growing on the inside and pushing against the one tooth and that\'s why that tooth is crooked. I have to have surgery. I have to go to Charlottsville to get an ortho consult so the doctor here can operate. What&nbsp;he wants to do is take out the baby tooth and then pull out the grown up tooth that should be there in the first place in it\'s place. I\'m not going to be missing a tooth or anything&nbsp;so everythings cool. I just haven\'t been in surgery for years and I\'m kind of nervous about it. &nbsp; '),(7136,2977760,'27,July,2004','2004-07-27 00:00:00','   Hey People,     &nbsp;  How is everyone doing?&nbsp; Good here.&nbsp; I\'m really bored.&nbsp; Its 3.30 in the afternoon and everyone I know is sleeping but me.&nbsp; and shelly.&nbsp; Shelly isn\'t sleeping.&nbsp; I tried to sleep but every fricken person in Harrisonburg decided to walk on my porch.&nbsp; Which, by the way, if you try to sleep in the living room and anyone makes a noise on the porch you\'ll here it like its RIGHT BESIDE YOU.&nbsp; No fun.    &nbsp;  I\'m cranky and I\'ve read all my library books.   BETH '),(7137,2977760,'27,July,2004','2004-07-27 00:00:00','   I was reading one a blog and came across an entry that had another link, so, I clicked on it. It was site for a little 5 month old&nbsp;baby girl, it was titled Save Julia. She has a rare disease called SCID. She needs a bone marrow transplant and her recovery depends a lot on what hospital she is at and that depends on how much money they have. I wish to give some money, though I don\'t have much at all($12) but I thought we might raise some together. It doesn\'t have to be a lot, just a little. For more about this go to  urlLink http://jerusalemgypsy.blogspot.com  and then go to&nbsp;an entry titled The Millionaire Next Door and scroll to one of the bottom paragraphs and find the orange thing&nbsp;   -This &nbsp;.   It\'s called This and is orange. Thank You. '),(7138,2977760,'26,July,2004','2004-07-26 00:00:00','   As I was reading I came to a paragraph in my book that especially caught my attention and made me go \"ewww\"(the good ewww, remember the bad ew is \"ew\" the good ew is \"ewww\"). From the book  Resurrection  by Tolstoy is the paragraph talked of in the first sentence that made me go \"ewww\"-    One of the most widespread superstitions is that every man has his own special definite qualitites: that he is kind, cruel, wise, stupid, energetic, apathetic, and so on. Men are not like that. We may say of a man that he is&nbsp; more often kind than cruel, more often wise than stupid, more often energetic than apathetic, or the reverse; but it would not be true to say of one man that he is kind and wise, of another that he is bad and stupid. And yet we always classify mankind in this way. And this is false. Men are like rivers: the water is the same in one and all; but every river is narrow here, more rapid there, here slower, there broader, now clear, now dull, now cold, now warm. It is the same with men. Every man bears in&nbsp;himself the germs of every human quality; but sometimes one quality manifests itself, sometimes another, and the man often becomes unlike himself, while still remaining the same man.    To &nbsp;say that I agree&nbsp; may well be &nbsp; hypocritical. I do not care. I agree. I think that a lot of people feel this way but still classify someone as being stupid or kind and I think that that\'s okay as long as they realize(this includes me)that there\'s more to them than&nbsp;just stupid or just kind and I think that a lot of people know&nbsp;that. I just like the fact that Tolstoy brought it up and&nbsp;wrote about it because it seems like such a&nbsp;small thing that loads of&nbsp;people know in there&nbsp;minds but never bring it up because it seems like everyone knows this.&nbsp;I could be completely ignorant &nbsp;and have no idea what I\'m talking about, though. That be the case- I\'m sorry if I have offended anyone. '),(7139,2977760,'25,July,2004','2004-07-25 00:00:00','   Hehe I said pizzel.    &nbsp;  Today went alright even though the realization of how far behind christian school put me, keeps clicking in.&nbsp; My dad is taking me to talk to the counselors tomorrow about my classes.&nbsp; I\'m nervous.&nbsp; I know they\'re going to have to look at my grades from last year and I shouldn\'t have failed.&nbsp; I passed everything with flying colors and I know there is a reason why they won\'t let me have my end of the year exams back.&nbsp; I know what you\'re thinking:&nbsp; Man is she paranoid.&nbsp; Typically I\'d laugh and say \'You\'re right, what am I doing?\' but christian school is screwed up in too many ways and its definately a possibility that they failed me because I wasn\'t coming back.&nbsp; The teachers would constantly ask me \'are you coming back?\' no.&nbsp; I\'m not you crappers, and you wanna know why biotch?&nbsp; Guess.    &nbsp;  Anyways, I don\'t feel like talking about this because I get mad beyond all recognition and then shut myself in my room for days.    &nbsp;  Catch you all tomorrow,    BETH '),(7140,2977760,'24,July,2004','2004-07-24 00:00:00','   Recent statments have been made about one of   OUR   dorks by a   fellow   dork. It so appears that a certain Dorkerer is   NOT   the Dork of   guitars  . Rather, he is the Dork of all music. So, I hereby declare(with the approval of all Dorks) that&nbsp;Dorkerer&nbsp;now become Dorkerer, the Dork of all music.   Please send your coments of approval   IF   you approve of this sudden change. Many thanks.   I had fun yesterday playing with the giant frisbee. I don\'t know if it\'s actually a frisbee, it\'s sort of like an oval balloon type thing that you throw as a frisbee but it isn\'t made of hard plastic.... I\'ll have to find out it\'s name... It amuses me. '),(7141,2977760,'24,July,2004','2004-07-24 00:00:00','   Hey everyone, I hope you all had a great day.&nbsp; I know I did     &nbsp;  You know what people?&nbsp; I love life (no I\'m not on any drugs, despite what you may think)&nbsp; I don\'t know why I love life.&nbsp; I just do.&nbsp; You guys do too, trust me you do...&nbsp;     &nbsp;  I had an oreo milkshake tonight     &nbsp;  I didn\'t see any good looking guys today.&nbsp; I saw some mediocre looking guys and some old men.&nbsp; They didn\'t suit my fancy.    &nbsp;   I like big butts and I cannot lie...     &nbsp;  I\'m turned in all my vacation cameras today at Walmart.&nbsp; Yay!&nbsp; I tried to take some really good pictures but I don\'t know how they\'d turn out because it was a disposable camera...&nbsp; very low quality.&nbsp;     &nbsp;  Tomorrow=Church.&nbsp; Wish me luck.&nbsp; Its like a freakin\' battlefield out there.&nbsp; Bodies torn apart and ragged left for dead in the feilds of sorrow...&nbsp; Okay so maybe it\'s not that extreeme...&nbsp; but you know, its not heaven either.&nbsp; Huh... isn\'t that interesting... church isn\'t heaven.&nbsp; Okay, I have to go.&nbsp;     &nbsp;  Bed time for Bethy.    &nbsp;   You otha brothers can\'t deny, when a girl walks in with an ittybitty waste and a round thing in your face...       &nbsp; '),(7142,2977760,'23,July,2004','2004-07-23 00:00:00','   This was very weird...I just took thisquiz thing somewhere, I don\'t even remember how I got it but it was a quiz to tell me what kind of self mutilator I was. My results-\"You are a cutter, you can be moody as hell and probably don\'t even know how your feeling half the time. Despite this, you pull yourself together with a lot of strength and aren\'t to be messed with\".      weird...      Am I supposed to feel proud?       If draging a big needle over my arm is supposed to be&nbsp;my inner strength, I would hate to&nbsp;see my inner weakness.... &nbsp;  &nbsp; '),(7143,2977760,'23,July,2004','2004-07-23 00:00:00','   Hello! how are you!?    -That\'s \'hello! how are!?\' in webding. Hehehe.    I want to talk about something and yet I find myself in a position where I have nothing to say. I\'ve been reading this book   Nana by Emile Zola  &nbsp;and I really really like it. I recommend it . I want to go to&nbsp;JC Penneys and shop but I don\'t think&nbsp;I\'ll be able to because noone can take me because my grandpa had a stroke and he\'s not allowed to drive and grandma\'s at work.&nbsp;The other day my mom asked me to dry&nbsp;off the dishes and I ignored her [partly because she asked me and partly because I would be doing them beside her. I know that\'s horrible but I can\'t stand doing things she asks me to. Also, the other day she asked me to&nbsp;go to the pool and the only reason I said no&nbsp;was because she would be wearing a bikini, her bikin\'s are more revealing than mine like she wears the kind that looks like a bra and underwear. If you have ever seen my mother&nbsp;you would understand why that is&nbsp;a VERY bad idea.&nbsp; '),(7144,2977760,'22,July,2004','2004-07-22 00:00:00','   I\'m back all my wonderful friends.    &nbsp;  I wrote about vacation in my other blog, so check it out if you are feeling bored. ( urlLink Http://www.angelfire.com/freak2/twizzledazzle/bloggy )    &nbsp;   &nbsp;  It was weird that when I checked my email that Chelsea inquired about which converses she should get, because my mom and I were talking about school clothes this week and she said \"I am trying to figure out a way to get the money for those converses you want.\"&nbsp; I love my mom, she\'s so rad.&nbsp; I don\'t love her because she wants to buy me things, I love her because while I was screwed up last year... she never gave up on me and even though I did a million things to hurt her she made sure that I came back as Me.&nbsp;     &nbsp;  Anyways, back to the converse dilemma.&nbsp; I can\'t decide whether I should get green, red or classic navy.&nbsp; If I can get them at all.&nbsp; I can always dream, right?&nbsp; I\'m obsessed with Kelly Green.&nbsp; My favorite shirt was Kelly Green and I found it at goodwill.&nbsp; On vacation I bought a kelly green watch.&nbsp; So, typically I would go for the green.&nbsp; BUT....&nbsp; RED.... red is so...&nbsp; Evil.&nbsp; It says \"Theres a little bit of hell on my feet.\"&nbsp; We all like the dangerous side of red.&nbsp; OH NAVY.&nbsp; Navvvyyyy is so... classic.&nbsp; You can  never  go wrong with classic.&nbsp; Rush is classic, Queen, AC/DC, Black Sabbath,&nbsp; Zepp....&nbsp;&nbsp; its all so good and  classic. &nbsp; What do you guys think....    &nbsp;  Now that I have killed you with my spear of boredom I will leave you to die in your computer chair.   BETH '),(7145,2977760,'21,July,2004','2004-07-21 00:00:00','   Well I\'m going to leave in about 2 hours, so I won\'t see/ talk to you until I get back, and that will be on the 31st.&nbsp; So I will talk to you all later, and I\'ll miss you all while I\'m away.&nbsp;&nbsp; ( And I\'ll take mental notes Shell!!)....L@8r &nbsp; '),(7146,2977760,'21,July,2004','2004-07-21 00:00:00','   Bubbles may very well be the most exciting things there is-well besides sex and I should really not talk about sex because I\'m inexperienced with that and I\'m rambeling,what was I saying?Oh yes! Bubbles! I love \'em! I was blowing bubbles with Luke(my 4 year old cousin) yesterday on the porch and grandma brought Maddy over(my 1 and 2-3 mons old cousin) and I started to blow them and she got this look on her face like :-o, I do it too when I\'m excited about something and there\'s more then one thing I want to do because I\'m so excited so I try to make myself be in more then one place and as you know that doesn\'t turn out too well. But anyways she got that look on her face and started running towards the bubbles and poking at them and popping them and it was so cute!    :sigh: The good times.   I love my cousins and I love bubbles! '),(7147,2977760,'19,July,2004','2004-07-19 00:00:00','   I want to meet a guy! I really want a guy!!!! :whine: I\'m lonely. I want a guy that doesn\'t think my sense of humor is weird or bad... I want a guy that cares about me... I want a guy... I want to meet someone that\'s kind of like me, and I know that sounds weird but I want to get along with him and I know that opposites attract and that is really true for a lot of people including me but I want a guy that isn\'t totally opposite,you know what I mean?... I want a freakin\' man.. '),(7148,2977760,'16,July,2004','2004-07-16 00:00:00','   i am still cruelly being forced to spend time with my stupid, ugly, disgusting dad. in a cold climate. we went to the beach today and there was this guy lying right across from me and right as i was about to take a big gulp of ice tea i noticed him staring at me. i lowered the can. stared right back at him. and shrugged my shoulders. he smiled. and i took my big gulp. and started reading my book again. and the asshole stared at me the rest of the time. it was okay the first time... and he wasn\'t even remotely bad-looking. but even his friend told him to stop staring. that has to be saying something. &nbsp; ANYways. i dunno what else to put here. &nbsp; hope you all are having a better time than i am... '),(7149,2977760,'16,July,2004','2004-07-16 00:00:00','   hey everyone!!!! I might not be on for the next couple days so don\'t worry, I\'m at Kissy\'s grandma\'s house. One of my friend got bit by something or she was having an allergic reaction to something cuz one side of her face swelled up a little and her eye got puffy and the awelled part is rock hard so she\'s going to the doctor today. She hates docoters but I\'m glad she\'s going cuz we need to find out how to treat it.  Yesterday I stepped on a slug... Krissy said I would have the feel of the slug under my foot for the next three days... I think she\'s right.... '),(7150,2977760,'13,July,2004','2004-07-13 00:00:00','   I starting to see that boys suck.  Okay not all of them, but some of them.  Maybe it\'s just me, but I hate the way they leave you hangin\'.  And you try not to think about it (whatever \"it\" maybe), but you end up doing it anyway. (Did that come out right?)  So than you wonder why can\'t the be more understanding about things and try to see something from your P.O.V.  But do they? no.  And why do they get so mad when things don\'t go their way???   Anyone w/ me on this?? '),(7151,2977760,'12,July,2004','2004-07-12 00:00:00','   \"It is a tidal wave of children who have lost one or more of their parents,\"Carol Bellamy, the executive director of the U.N\'s childrens agency UNICEF, told Reuters. \"Fifteen million globally\"she says\"It has the possibility of destablizing societies quite dramatically\".-Patricia Reaney,Reuters. AIDS has tacken a parent or two from 15 million people.  15 million people.  It goes without saying it has increased the number of orphans. The 15th International AIDS Conference released a report that stated an estimated 18.4 million children would lose atleast one parent by 2010 to the disease.  People are more interested in finding a way to control the disease and finding a cure and I\'m not saying they shouldn\'t be but orphaned children because of Aids are being somewhat neglected and aren\'t getting as much attnetion they need(and deserve)and that pisses me off. '),(7152,2977760,'12,July,2004','2004-07-12 00:00:00','   I hope I didn\'t sound bitter or unhappy for all my dorks in the below entry because I\'m really not. I hope that Beth has unforgetable fun despite the trip being created because of various evil reasons. I hope Chelsea has a great and knowledge-filled time on her cruise . I hope G\'s remaining time in Turkey is exciting and eventful(the non-boring and good kind of events). I hope Krissy\'s stay at her grandmothers in Elkton is peaceful and relaxing.                        *HOPE YOUR ALL WELL* '),(7153,2977760,'12,July,2004','2004-07-12 00:00:00','   Dorkenya is gone on her trip for various evil reasons, Dorketta is in Turkey, Dorkster is getting ready to go on a cruise, Dorkerella is in Elkton and I, Dorkengerine, am stuck in Harrisonburg,sweet ol\' H\'burg. I am so lonely;(. The only company I have is the television. I\'ve become rather interested in The West Wing. Why? I\'m not exactly sure. I think it\'s because I really want to know things about politics, I hate being ignorant in stuff like that, I hate how ignorant I am. Atleast I\'m not ignorant to the fact that I am ignorant. Anyways, I really like The West Wing but I don\'t fully understand what they talk about and that\'s partly becuase they talk to fast about important things. I don\'t know I\'m just rambeling probably, nobody really cares about my ignorance... But since there\'s noone here for me to do stuff with I\'m stuck home watching soap operas and reruns of Dawsons Creek, The West Wing, Buffy the Vampire Slayer and Angel.  :sigh: Woe is me... And on that mockingly dramatic note- '),(7154,2977760,'12,July,2004','2004-07-12 00:00:00','   I wish i could watch I love the 90s. I LOVED the 90s... I have to spend at least a week with my dad and today I called my mom to tell her something and she started randomly yelling at me ...  this is turning out to be a reaally crappy summer. and i was so looking forward to it.   i\'m staying at this medieval place now... so don\'t expect me online again for at least 5 days...  as the SECOND DAY with my DAD ends.... '),(7155,2977760,'12,July,2004','2004-07-12 00:00:00','   If you have not seen The Blue Collar Comedy Tour you must absolutely beyond a doubt see it. It is hilarious. I don\'t care if your not country or redneck you HAVE to see it. It\'s a fix for depression too. Here\'s a little tid bit out from the \'you may be a redneck if-\' \"if someone says you\'ve got something in your teeth and you take them out to see what it is,you may be a redneck\" \"if your idea of fastfood is hitting a deer with your truck at 65 miles an hour you may be a redneck\" \"if your momma says your lying through your tooth, you may be a redneck\" And here\'s a bit from \'here\'s your sign\'(retard sign i guess) \"the airport has lost my luggage and you go to the lost luggage place and the person working there says \"has your flight landed yet?\" and i reply\" no they just dropped me off by parachute. There\'s your sign\" \"I went shopping and went up to pay for them,i dropped them on the counter by the cash register and the lady working there goes \"you buyin those?\" i replied \"no i\'m gonna steal \'em i\'m just lettin\' you see \'em first. There\'s your sign\" It\'s great you\'ve gotta see it. '),(7156,2977760,'11,July,2004','2004-07-11 00:00:00','   OMFG.I\'m going to fucking cry because the motherfucking hairdresser does not know how to fucking cut any goddamn hair. UGH!!!!!!!!!!!!!!!!!!!!!!!! I\'m gonna cry;-(:whine: :whine:....... '),(7157,2977760,'11,July,2004','2004-07-11 00:00:00','   I\'m getting my hair cut today! YAY!!! I want to get a trim and get sweep away bangs and if the hairdresser does it right this time it should look really good.  :sigh: My mom is taking me.... I have to tsepnd some quality time with the wicked witch from the west wacked out on drugs..... Oh-well. '),(7158,2977760,'11,July,2004','2004-07-11 00:00:00','   ...Be happy! JK :-D  Anyway, don\'t worry shell, I got your comments and your\'s too G.  If you\'re wondering shell, when you post comments they\'re not going to show directly on the screen.  They\'ll appear on the link that says \"comments\".  I hope that didn\'t sound mean.  I didn\'t want to to, I was just telling you incase u forgot or didn\'t know... '),(7159,2977760,'10,July,2004','2004-07-10 00:00:00','   hello everyone I was just seeing if the blog will view this post.... '),(7160,2977760,'10,July,2004','2004-07-10 00:00:00','   for some reason this blog is not publishing my comments.....i wrote one to chelsea on this blog and another on her other one and then i wrote g one on this blog and another on her other blog...and it won\'t even let me on beth\'s blog....i\'m confused.... '),(7161,2977760,'10,July,2004','2004-07-10 00:00:00','   Hey all dorks. I\'m still desperately hanging on to my mom but... pleeeeeeease don\'t make me go. we just went to a beauty salon thingie... and wow... waxing has to be the worst form of punishment. but maybe it helped a bit because my mom was talking to this person who was like... \'oh, she\'s BEAUTIFUL...\' :makes face: i don\'t see it. anyways... oh shellliii help me. i dont wanna goooooooooo...........   - desperate dorketta '),(7162,2977760,'10,July,2004','2004-07-10 00:00:00','   I really want a tattoo.  So I figure that I\'ll get one when I\'m 18.  But I think one time my mom said that I\'m not going to get one as long as I\'m living under the same roof.  But I\'ve talk to her before about getting one in front of her and she never said anything bad about it.  So I dunno.  But anyway, when I\'m 18 I\'ll be a senior and - That\'s perfect!!! I\'ll get my little black star and then I\'ll get a blue streak to remember all my friends.  So u all should come w/ me when I get that, so it will be extra special.  And why does everyone look and me when I say that I\'m going to get a tattoo and think \"YOU!!!????... YEAH RIGHT!!\"... '),(7163,2977760,'09,July,2004','2004-07-09 00:00:00','   I FINALLY DONE IT!!! I\'M ON HERE.  YA GO ME.  I FIGURED IT OUT ALL BY MYSELF (SORTA) JK, but really, I\'d like to thank all the people that got me here.  shel, beth, and finally G.  Thank you... '),(7164,2977760,'09,July,2004','2004-07-09 00:00:00','   Hey all my dorks! I am going to miss you Dorkenya :snog: I hope you find a way to have fun during your...um trip for various evil reasons. Dorketta I\'ll help anyway I can when your staying with your dad,I mean I hope I can help-You\'ve helped me so much with my mom by just being there. Of course I\'m here for your if you need me, your my Dorketta! What makes people think they have superiority over anyone else? I\'m talking about racists and sexiests. What makes them think they have a certain \"specialness\" that say someone of the another gender or someone of a different color could not have or does not have? This is just how I feel, skin color is not a reason to dislike or hate someone. Hate is what kills people which just leads to more hate and eventually that brings along pain in however order you wanna put that. It never solves anything and never brings love,kindness or happiness just more hate and pain and evilness over and over and it never stops. '),(7165,2977760,'09,July,2004','2004-07-09 00:00:00','   I taught my cousin Maddie(15 mons old)to dance in the rain today:-D. Omg I\'m soo happy, she was so cute out there in her little pink tee-shirt and diaper dancing in the rain. I just hope she doesn\'t get a cold or anything. I feel special:-D. '),(7166,2977760,'09,July,2004','2004-07-09 00:00:00','   Hey everyone.  Yes... it is saturday and I am awake BEFORE one oclock in the afternoon.  Whats wrong with Beth, one might ask themselves.  Well... I am very excited because my sister is taking me out this morning.  Usually she is gone with her boyfriend by this time in the morning (9:59 AM) and she does not come home until midnite or after.    Anyways, other than the fact that I\'m spitting flummery at you... I\'m actually quite happy this morning.  I\'m wearing my \"Llamas and Pirates and Cheese OH MY\" shirt.  This one guy came up and hugged me for wearing it.  MWAHAA he was good looking too.  I found my toe ring.    Obviously you guys can tell I\'m bored and I just WANT TO POST SOMETHING WITH ANYTHING IN IT.  Oh... I won\'t be here for a couple weeks for various evil reasons.  Don\'t worry I\'m not dead.  Wait a minute... I MIGHT BE A VAMPIRE WITH VAMPIRE NAILS.  My sisters leg hurts.  MWAHAA.  Catch you later you guys. DorkenyaBETH '),(7167,2977760,'09,July,2004','2004-07-09 00:00:00','   Hello, all dorks. today i dyed my hair this really dark brown - its like black and i got it layered and i got my bangs... i think my hair looks pretty now - just the rest of me looks crappy. im with my mummy in the capital of turkey and my DAD is forcing me to spend a week with him at this other place and i REALLY, REALLY don\'t wanna go... i haven\'t gone to the isolated beach so I will be able to get online... so you guys are gonna have to help me through this...   i visited my old elementary school and my old house today. i\'m going on a steady downward spiral.   YEEEEEEEEEES dorkenya itd be awesome if i could spend spring break with all you dorks and my mom said i might be able to come to the USA early since there\'s no place for us on a plane on august 15. hanging out with you guys - if i can, that is - is the only thing i have to look forward to.  CIAO... oh sorry about the long boring blog... i\'m BORED... '),(7168,2977760,'09,July,2004','2004-07-09 00:00:00','   ...okay never mind the post that was telling you to sign your name b/c it already says who wrote it,i just didn\'t see it because the print was so small... '),(7169,2977760,'09,July,2004','2004-07-09 00:00:00','   I\'m in love with you all my little chicas.   Penis.   Sorry--I\'m REALLY tired and can\'t move my fingers very well... OH! YAY!  RUSH IS ON THE RADIO!  Hell... fluffing... yea.  DorkenyaBETH '),(7170,2977760,'09,July,2004','2004-07-09 00:00:00','   I(Dorkengerine) wrote the last 2 entries called \'England\' and \'I am a dork and a member of the very special dork family\' and G(Dorketta) wrote the first entry. Don\'t forget to sign what entry you have written so nobody gets confused.  -Dorkengerine- '),(7171,2977760,'09,July,2004','2004-07-09 00:00:00','   One of these days I\'m going to go to England with my dorks. We are going to show up the old British women who think Americans are unable to speak proper English. We are going to spit at all the british girls and steal their hott english boyfriends. We are going to ride the humungous ferris wheel and look upon the stars. We are going to dance in the rain and laugh intill we have to take Dorkkenya to the hospital because she forgot to breath between the bursts of laughter. We\'re going to walk into the shops that are so obviously made for just the rich and wealthy and sneer at the employees as we try on their expensive clothing. I am going to England with my dorks and find the dark prince of my twisted fantasies. I even wrote a \"cheese poem\" for him. can\'t I find a place somewhere? a place for me that I can have in someones eyes that love me too have his soul,could I? and call it mine and mine alone? a place protected from groping hands; icecycle fingers threatening a chill a warm ocean draft,I dream a key to unlock the secrets of his masked eyes such sweet and sour seductiveness what key better then my heart? ready to unfold and unfurl all dark ready to stop hiding in myself and find a pleasant burrow inside of him  and for him to find his place in me like a lone hotel abandoned,haunted and dark I wait he\'ll walk in and turn the light on and start a newness in me like freshly baked pumpkin pie sitting warm beside the turkey; safe from November wind and with a place '),(7172,2977760,'09,July,2004','2004-07-09 00:00:00','   Well, I don\'t have much to say right now but except I really really want to go to England...... '),(7173,2977760,'09,July,2004','2004-07-09 00:00:00','   THE DORKS...  Rachelli - Dorkengerine (Dork of Tangerines) G - Dorketta (Dork of Sleeping Beauty) Krissy - Dorkerella (Dork of Cinderella) Beth - Dorkenya (Dork of Kenya)  Chelsea - Dorkster (Dork of Monsters) David and Andrew - Dorker and Dorkerer (Dork os Guitars and King of Dorks) '),(7174,2977760,'12,August,2004','2004-08-12 00:00:00','   My fucking grandpa is cancelling the internet. I don\'t know when. If you want to contact me you\'ll have to either call me or write me a letter-Phone # (540)433-9620, cell phone #(540)435-1850. Address-708 Stuart st. Ugh, why does he get the computer and then all of a sudden take it away? Why get it in the first place, man, it pisses me off. I need it not only to talk to my friends but for school. In Loughrans class if we didn\'t type our writings(poems and stories and projects) we got points taken off.  Who\'s to say another teacher won\'t be like that? I also need it to research things. Talking is not the only thing I freakin\' do on the damn computer! Ugh, he pisses me off.   Oh, and after awhile he\'s cancelling the cell phone too.   Damn him. '),(7175,2977760,'11,August,2004','2004-08-11 00:00:00','   I\'m jealous of you G. Why can\'t are school have a photograph glass??? This sucks...!!!   ( I\'m not mad at you, I\'m just jealous) '),(7176,2977760,'11,August,2004','2004-08-11 00:00:00','   I\'m wanting another and newer SN for AIM.  I\'ve got some ideals, but I want to know what you think I should put??  OHH, I just thought of this, what about Dorkster8??? '),(7177,2977760,'10,August,2004','2004-08-10 00:00:00','   Hey G, I noticed that you put  urlLink Anna Karenina  as one of your favorite books?  Is it really good?  What\'s it about???? '),(7178,2977760,'09,August,2004','2004-08-09 00:00:00','   LOL, I\'ve FINALLY posted the 1st part of my vacation stroy.  If you don\'t know the link, just click on this :   urlLink www.theotherdorkster.blogspot.com '),(7179,2977760,'09,August,2004','2004-08-09 00:00:00','   I think I could be one of the happiest people in the world!!! I got Loughran again for english, so I\'m happy.  Who do you have shell and beth?  G, I wish you hadn\'t moved!!!! '),(7180,2977760,'09,August,2004','2004-08-09 00:00:00','   Today was really an OK day. I went and got signed up for school. I made sure I knew where all my classes were. Grandma(yes, I made grandma come with me, I wasn\'t about to go with just my mom)met some people at school, Dale and Virginia, they\'re janitors and since grandma knows the kind of work, she decided to say hello, it\'s a way of saying thank you, somehow. But I always so \"hi\" or smile when I see Dale in the halls, he\'s really nice and so is his wife(Virginia). It\'s so cute, they always where matching shirts. Didn\'t see anyone I knew. I managed to get my mom away from the PTSA(whew! that was a close one)so she didn\'t sign herself up for any colunteer crap. If my mom was not a drug addict I would\'ve cared less how many things she sined up to do, but she is. The last time she signed up for that kind of stuff was in elementary school. It went as follows: She came into my third grade class to help us paint Christmas knomes or \"Holiday\" knomes. I don\'t remember if she was high or not but I think she just coming off of something. Anyways, you were supposed to make two and paint all of it one color(red or green) and leave a little circle for it\'s face. I forgot to leave a circle. Big mistake. My mom flies off the handle screaming at me at the top of her lungs-\"YOU STUPID! FOLLOW THE FUCKIN\' DIRECTIONS!\" blah blah blah and so on and so on. I look around the room at all the faces looking at us. I start bawling and run into the bookbag place(you know the place that was like a closet that you slide the door open and put your coats and bookbag? that thing). And I refused to come out. It was sooo embarrassing. Mrs. Lile, one of my favorite teachers by the way, finally got me to come out and I can just remember everyone s eyes on me and how hot my face was. So, you can see my desperation to avoid her volunteering at all cost.    I had a nice little wrestling match with Luke and Logan upstairs on the floor. Haven\'t had one of those since.... I can\'t remember, but it was kinda like a way to lose myself. I was happy. It felt nice. Me and Maddy played peekabo. :-D. She would put her hands over her eyes and I would go-\"Where\'s Maddy?\" in my stupid little kid voice and she would put her hands down real fast and I would act surprised and go \"AHH! There you are!\" and she would giggle and do it again. I love my cousins. Cousins are grand. '),(7181,2977760,'07,August,2004','2004-08-07 00:00:00','   I just finished watching The Blue Collar Comedy Tour, and you\'re right shell, it is funny!!! '),(7182,2977760,'07,August,2004','2004-08-07 00:00:00','   I went to Hillandell park today with the kids(Logan and Luke) and read while they played on the playground thing. There was a cute little girl down there, she was blind. It was sad because she was feeling around everywhere so she would know where to go.    You know there worlds going to end when you see a country singer wearin\' a rap shirt.   Gentry of Montgomery Gentry had on a G-Unit shirt... Odd.   Odd rhymes with... Wait, what? '),(7183,2977760,'06,August,2004','2004-08-06 00:00:00','   I am excited. I got advanced profiency on my World History SOL\'s. I only missed 5 questions.    On my Math SOL\'s I wasn\'t so lucky. I got 39 out of 50 which is pass/profiency.   I\'m not tryng to brag or anything, I\'m just happy.   :-D. '),(7184,2977760,'04,August,2004','2004-08-04 00:00:00','   Today is a very pretty day. I am happy to be alive today. The sky is extra clear. It feels like September or October. There\'s a cold breeze. It feels so beautiful, cold but I\'ve always liked the cold when the sun is shining and the sky is so clear... Like it is today.    I got up at 7:50 a.m today. Why? You ask. My grandma walks in as I\'m trying to fall back asleep and informs me she\'s going on a walk. I mumbeled \"... okay...\" and she pauses for a second ..\"would you like to go?\". This is the strange part, I said yes. So, I get up and find that it\'s very cold so I have a good excuse to wear my new jacket without looking like an idiot. We had a nice little walk. I like the weather today, even though when I got back and took a shower it was colder when I got out then usual. I\'m not use to that. One of the things I\'m not looking forward to when it gets cold. '),(7185,2977760,'04,August,2004','2004-08-04 00:00:00','   Apparently I\'m not going to be a social worker. Apparently I\'m not going to college. Apparently I have no common sense. Apparently I\'m not going to amount to anything. Apparently I\'m some weirdo hermit that never does anything. And apparently I\'m going to end up like my mom. Says grandma and grandpa. Well thanks a fuckin lot for having faith in me O\'Family of mine. '),(7186,2977760,'03,August,2004','2004-08-03 00:00:00','   I\'m bored. I can\'t think of anything to research...    I was watching a Red Hot Chili Peppers video and the lead singer looked kinda hott with a mohawk.. He\'s probably old now but I don\'t care. My ears are hurting.. These head phones used to not hurt... Odd.   Odd rhymes with ... I don\'t know.    Hey! Dorkster! Guess what I got you? I got you a/the/an- I\'m not saying! HaHaHa! Was that as fun for you as it was for me?   I wish grapes were hot pink and this color pink. Then they would taste like bubblegum. I wish I saw a neon green sky. I swear I\'m not on anything.   (Not that I don\'t like grapes the way they are(not discriminating against grapes)I just want to eat a hot pink grape that tastes like bubblegum) '),(7187,2977760,'02,August,2004','2004-08-02 00:00:00','   Life just isn\'t what it used to be.      But what it will be is going to be the best. '),(7188,2977760,'01,August,2004','2004-08-01 00:00:00','   :Sigh:      I went to CiCi\'s the other day with my grandma,pappap,aunt,Luke and Maddy(cousins), and my mom even though I didn\'t eat but one piece of pizza and I didn\'t want to go at all but since grandma thinks I have problems with eating, I went. Needless to say my mom manipulated this dining arrangement to spend time with me. I went to look around at Kohl\'s because I got $10 from the yard sale and they were having a sale. She tried to buy me clothes because I guess she thinks she can buy me back. Noooo, she can\'t. I was going to get a pair of jeans that were on sale and a care bear shirt and a cordiroy coat because it was 50% off and my old one was kind of shrinking at the sleeves and it\'s sort of difficult to find a cordiroy jacket that\'s 50% off. That\'s what I was going to get and I would be done with shopping for school clothes but she wanted to get this sweater/shirt for me that was cute but it was $14.99 and in my opinion that was still too high and I knew what she was taking me shopping for. I didn\'t want her to get it so I begged her to put it down and she wouldn\'t and it pissed me off because I knew I was going to end up feeling guilty. Sure enough. I feel guilty. I HATE when she does that. I love when I get new stuff but I hate what else it costs me. '),(7189,3866526,'22,July,2004','2004-07-22 00:00:00','   It has now been two straight days of switching between Belle and Sebastian\'s \"Tigermilk\" and Van Halen\'s \"1984\" as the only records that I listen to in my Sony Psyc! Discman.  Weird Musical Taste or Bi-Polar Disorder?&nbsp; You decide!   '),(7190,3866526,'12,July,2004','2004-07-12 00:00:00','   Just go see Anchorman: The Legend Of Ron Burgundy.  When I\'m not giggling, I\'m speechless.  I haven\'t laughed that hard in the theater since the original Naked Gun.  That\'s all.   '),(7191,3866526,'09,July,2004','2004-07-09 00:00:00','   Having lived in Allston for a year, I realize that I will never, ever, never, ever, never come close to not feeling screwed by the B-Line.  But, unlike my hopeless, beloved Red Sox, I finally got my small victory.  I got on a very crowded train to go home last night.  It was one of those deals where all the out-of-towners were standing in the first car when there was an open football field in the back.  So I struggle and fight my way through these idiots to freedom ... only to find myself surrounded by about 85 really loud high school kids in Sox shirts.  They\'re screaming, fighting, throwing gummi bears and generally acting like complete neanderthals.  I figured this would go on for the whole ride, and I\'d have to use a paint-chipper to scrape the grime and tinnitus off.  (These are the things that happen on the B-Line - loud and raucous kids, shouting - it\'s like the official Short Bus of the city.)  Anyway, after one stop, one of the kids comes over to me.  Uh-oh.  KID:  \"Hey, nice watch.\"  Then he threw a gummi bear at me and it stuck to my knee.  Soooo ... just to show them who\'s boss in this town, (are you ready?  This is what I did...) I put on my Sony Psyc! Discman, and I listened to \"Get Me Away From Here I\'m Dying\" by Belle and Sebastian 12 times in a row! Without interruptions!  Ha ha!!!!  (I know, I know.  I can\'t believe it either.)  Take that, B-Line!   '),(7192,3866526,'08,July,2004','2004-07-08 00:00:00','   When you read my astoundingly awesome blog, please keep in mind that even though it is written in the voice of ESPN reporter Sal Paolantonio, that  it is still meant to be read in the voice of Dave Coulier doing an impression of Popeye .  (I can\'t stress this enough.  This astoundingly awesome blog works two ways, people.  Snap to it!)  Also, all punctuation sounds like the sound effect from America\'s Funniest Home Videos when some jackass gets hit in the junk.  Enjoy, everybody!  Blog, blog, blogamundo!  Do-do-do-do-do....   '),(7193,3866526,'07,July,2004','2004-07-07 00:00:00','   Ah, jokes.  Eh?    I actually quit smoking on December 22, 2001 at 4:35 pm EST while eating a cheese sandwich and watching the Saved By The Bell episode where the gang heads to the mall to get U2 tickets.   See?  I remember the exact date, time and Saved By The Bell episode, therefore it\'s true.  Rule of three wins again!  Yay comedy!  Now, if you\'ll excuse me, I\'m off to ebay to check on my birkenstocks.   '),(7194,3866526,'07,July,2004','2004-07-07 00:00:00','   2 Reasons:  1. I just quit smoking.  2. I just quit not blogging.  Please help me before I buy sandals.   '),(7195,3866526,'02,August,2004','2004-08-02 00:00:00','   I think I\'ve had it with eggplant.  No more eggplant.   '),(7196,4116584,'20,August,2004','2004-08-20 00:00:00','   Michigan was cool, the only thing i didnt like about it was my cousin, Marie, she is so spoiled.....they had a carnival up there by her church, and i had asked her if she wanted to go, cause i could take her, then she said she wanted to wait for her mom, but then her mom didnt feel like goin that day so Marie didnt wanna go either, that really pissed me off......i dont mind goin up there, but if she just wasnt so spoiled.....Sunday they had a parade and Marie was on one of the floats, i had a chance to go, but i didnt want to.  so one of the floats stops because a little kid fell off and got his foot ran over....my Aunt Ruth told me to go check it out to make sure it wasnt Marie, i saw no point b/c Marie\'s float was 5 floats behind this one....so i check and its a little boy, but he was fine he started walkin around that same night....so about my trip, i liked it, i saw a lot of hot girls up there and one of them gave me her number!   '),(7197,4116584,'06,August,2004','2004-08-06 00:00:00','   Yay, i get to go to Michigan to see my aunt and her annoying daughter, she\'s an only child and u can expect her to b spoiled and she will be.  Its for these Maritime days where they have a big carnival and its really fun.  My next post will be about what happened up there, peese and much love!!   '),(7198,4116584,'03,August,2004','2004-08-03 00:00:00','   Well, when i think about it, im the kind of person who wants to do a lot with my friends, but sometimes, my friends dont want to do a lot with me, but no matter how hard they try, they cant hide it from me.  But hey, dont feel bad for me, as long as i have GOD and myself, no one can stop me from doing what i want to do.  and when people talk about me, it hurts of course, i wouldnt be human if it didnt hurt.  sometimes i think about what would\'ve happened if i didnt make it when i passed out or i didnt make it through a lot of the stuff i have to go through.  some advice, dont ever think like me and dont ever think ur me!   '),(7199,4116584,'02,August,2004','2004-08-02 00:00:00','   Today was actually kinda boring, i pretty much watched TV like any other kid during the summer, but then i realized i had to get something done so i did it, then i said a rosary for my brother, Dan, and then watched more TV, i know, not an exciting day, but hey, im just starting out!!  For Our Lady of Angels:  Readings for Next Sunday August 8, 2004  1st reading-Your Ancestors waited in faith and courage for God\'s primsed summons (Wisdom 18:6-9)  2nd reading-Because of his faithful obedience to God\'s call, Abraham\'s descendants number as the stars (Hebrews 11:1-2, 8-19)  Gospel-Keep watch and be prepared, for you do not know the day or the hour when the Son of Man will appear (Luke 12:32-48)   '),(7200,3992573,'29,July,2004','2004-07-29 00:00:00','   ok so i just don\'t understand men. it\'s later on in the day because i just haven\'t had a chance to blog today, and it\'s really not because i was busy working because to be honest i haven\'t done a damn thing all day & ya know what? i don\'t give a shit! it\'s just that kind of day.       so last night was fine, didn\'t end up going to dinner with meta because he had to do some shit with the DNC. and i really couldn\'t get mad over that because for the 1st 2 years that we were together he didn\'t work so the fact that he\'s actually working now is a big deal so i can\'t get mad when he\'s doing work related stuff. so i just went home &amp; cooked supper & chilled there.        so today comes around. he calls me at like 12:30 we talk for like 10 minutes &amp; then he tells me that he has to go to the boston pops concert that he was supposed to be going to last night but didn\'t go because it was rained out & postponed until tonight. so i was like oh well when did you find out that it was postponed? and he tells me that he found out when he met up with his boys last night. so i was like YOUR BOYS? meaning your friends were there &amp; that\'s why you blew me off? and i then went into a frantic rage.....well he shuts me up & asks me why the hell would he lie about something that i could so easily find out about? how the hell am i supposed to know? so he says whatever then you wanna go tonight? and i was like no i don\'t wanna go because i am not about to be fit into your schedule. and he was like ya know what here i am trying to make this better, trying to find time to spend with you &amp; what the hell are you doing? yelling at me because i chose the wrong words, it wasn\'t my friends, it was my boss, kirsh, their mothers & my mom....OH. so in the meantime i email choclate kisses &amp; tell her to hurry up & check via if the concert was really cancelled last night so she emails back saying that she doesnt understand....UGH. i am trying to find out if this arguing is worth it &amp; she doesn\'t know.....AHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHH       so anyhow, i continue to argue & he was like ya know what instead of me being rude i\'m just gonna say bye &amp; call me back later because i don\'t want to fight with you over nothing. so i was like whatever maybe i won\'t call you back at all & he says fine &amp; hangs up....it\'s not 2 1/2 hours later & he\'s ignoring my calls. oh did i forget to mention that after i hung up the phone &amp; started ranting to chocolate did she tell me that she heard that the concert was cancelled this a.m. on the radio HELLO! omg he wasn\'t even lying. i yelled & i screamed &amp; he wasn\'t eveb lying. pick your battles natty....damn it. now i\'m trying to call him & he won\'t answer because i rearranged all of my plans for tonight so that i could go to dinner with him &amp; he\'s not answering.       so i\'m gonna keep calling him....we will see what happens tomorrow.          &nbsp;    HASTA MANANA!   '),(7201,3992573,'28,July,2004','2004-07-28 00:00:00','   well here i am again! just another day in my world...welcome.       yesterday was ok i guess. worked overtime which always helps with the bills that just seem to keep piling up! mom & dad really need to get off their ass\'s about opening a new store or something needs to happen to that dip shit that thinks that he can run our store so that i can have $ in my pocket again! i hate him. i have begun my plot though so it now is only a matter of time before he cracks. and good old meta is helpin me out so i\'ll be ok. it\'s just so damn hard for me not to have $ like i used to, so readily available to me at any &amp; all times ya know? but it will be ok. i know that they will open something else up, but i want that store back. it is literally a cash cow & why would anyone so willingly give that up i still haven\'t figured out!        but anyhow, i left here &amp; went over j\'s for a bit. he forgot his key in my bag on saturday night & needed it. then i went home &amp; cleaned up a bit. made dinner (hamburgers & hot dogs) ate dinner, gave the baby a bath because f &amp; p went to home depot to buy some sheet rock, watched some t.v. & then went to bed.        yeah i know too much excitement for 1 day...try &amp; control yourself.       oh i forgot to mention big I called me 3 times yesterday. guess the brother is on shook status...GOOD. that\'s where he needs to be. because honestly i not only have 1 foot out the door, my louis vuitton is right behind it.       supposed to be going out to dinner with meta tonight, we will see what happens with that.          &nbsp;    hasta manana!!          &nbsp;    : )   '),(7202,3992573,'27,July,2004','2004-07-27 00:00:00','   good morning everyone!! here i am in another day. i am not that tired today which is a good thing, funny how that extra 10 minutes from the snooze actually can make you feel so much better! but then after that p ends up being in the bathroom so i have to either knock on the door & get annoyed by him, or just go into the hall bathroom. which is the route that i have taken for like 2 weeks now. i just moved the toothbrush, toothpaste &amp; bruch in there....whatever. it\'s way too early to argue with him about that.        so last night i saw meta. i picked him up at his mothers house & we were going to go out to dinner. well i go &amp; pick him up & he tells me that he needs to go by his grandparents house to say bye to his family that came here from FL. so i said that was fine &amp; then he starts telling me that i have to go inside with him because i need to meet these people. so i asked the simple question....what am i meeting them as? i mean come on now i\'m not his girlfriend so what exactly will i be meeting them for? what is the relevance? well that started an argument. so i pull up to the house & he went in &amp; i waited in the car. no sooner did the door close behind him did his grandmother come out & literally yell at me to go inside. so i call J as i\'m doing this because, well just because i had to tell her about the craziness that was occuring. so i go inside &amp; we hung out for a while & then i finally said to him ok are we leaving so we can go &amp; eat? and he\'s like yeah i all ready told them that we had reservations so we\'re straight to go. we didn\'t have reservations but that\'s not really the point. we haven\'t really seen eachother since the weekend before the 4th & we wanted to spend some time together, damn i have to explain myself. anyhow, so we get up to leave &amp; she freaks out asking us why we were going & we\'re we too good to sit down &amp; eat with them & blah blah blah. so he asks me if it was that big of a deal if we ate there &amp; i of course said hell yes. i wanted to go out somewhere where we could be one on one, but now we were going to be stuck in PR all night! and then she starts, i never see my grandson, he\'s never here & when he\'s here he\'s with you yada yada yada. now as she\'s saying this all that was going through my mind was well shit i never see him either so i guess we both have points. so finally he tells me that if we stay there to eat then we\'ll go out on wednesday night. whatever. so i go outside to move the car because it was in the middle of the road. well they live on a frigen dead end with 4 houses on it. so i went to find a space, oh i found one....on the next street over.&nbsp;oh did i forget to mention that as i was closing the&nbsp;door she asks me oh is the restaurant closing? is it the last night that you guys could go there? nice.&nbsp;so i go back inside & we started eating &amp; then the grilling came. and i\'m not talking about steaks......so when are you 2 going to get engaged? when are you going to move in together? when are you going to have a child? HELLO! CAN WE ESTABLISH A RELATIONSHIP HERE FIRST & THEN MAYBE WE CAN TALK ABOUT MARRIAGE? UGHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHH.        so we finish over there &amp; we\'re driving back home & he says to me thank you so much for doing that babe, you didn\'t have to deal with that. and thank you for helping them pick up. lemme explain that. we all finished eating &amp; automatically all the guys got up & walked into the living room, well he grabs my hand to go with them but all the ladies started helping with the cleaning. so what the hell was i supposed to do? go in there &amp; sit with all the guys? yeah that would have made me look really good. so i helped out. not that it took long at all because there were literally like 10 of us doing it. but i guess that made a difference in some way, ok. so i go to drop him off at his mother\'s he asked if i wanted to come in & after making out for a while we both decided that it just wasn\'t a good idea because he really can\'t do anything yet. he\'s in a frigen sling &amp; a brace across his back. not to mention the stiches aren\'t even healed yet. so yeah going inside not a good idea. so he goes upstairs i go home & pass the hell out because it\'s midnight by then and oh yeah i\'m up at 5.       so all &amp; all it was a good night.       now let\'s see what happens tomorrow night.          &nbsp;    : )   '),(7203,3992573,'26,July,2004','2004-07-26 00:00:00','   ok so here i am...again. it\'s monday again & i realized today that i need to be getting to sleep at like 8 if i want to feel refreshed. but then again, i got like 10 hours of sleep on friday night &amp; was still tired. probably the cold i am trying to kick.&nbsp; and yes i know what your thinking, how can i have a cold in the middle of the summer, but that\'s what happens to me, i get sick in the summer. even though meta doesnt want to understand or believe that, whatever!      anyhow, the weekend was good, tiring but good. friday i left this place & went home, had a wax appointment at 7 with f then went to dinner with the fam. left there &amp; went over j\'s for a bit & decided that i should have gone straight home &amp; to bed, i was home by 12 & in bed by 1201. woke up at like 1030 sat morning &amp; cleaned the hell out of the entire house. talked to meta for 2 hours about absolutely nothing other than the fact that i haven\'t seen him in way too long. i miss the big head. then f came home & we ran some errands. went back home &amp; prepared for yesterday (p\'s party) then went over joe\'s. we had chinese & hung out. he was bartending at RED that night so he left i picked a up from work &amp; we went out. stopped at RED & got all drunk for free then walked down to octaine, drank some more, then on the rocks for reasons which i still dont know. i hate that place &amp; everytime i go there i realize more & more why! it\'s just a bunch of euro trash.....wait sorry wanna be euro trash. there is a difference. then we went home ate something &amp; all passed out.      sunday morning i went back home early to help with the rest of the party preperations. we finished all the food, put up the decorations & touched up everything. people started coming at 2. p, dad &amp; the boys got home around 330 & he was so surprised. his whole family came &amp; we just had a really really nice day. after everyone left at like 9 he was opening his gifts & he got all teary eyed &amp; was telling me how shocked he was that everyone came & that so many people care about him. ya know as much as he\'s a pain in my ass he\'s a great guy, wonderful heart. i hope some day that i will find a man who cares half as much about me as he does to my sister &amp; their daughter. i can\'t believe i said that because he gets on my nerves so much, but boil it all down & he is a great person with an even better heart.     and that\'s how it\'s all ending for the weekend.     supposedly meta is coming back home tonight (we\'ll see what actually ends up happening) he was being really&nbsp;good on the phone today. only a few add moments with the t.v. and radio because my baby just can not concentrate on more than 1 thing at a time.&nbsp; although he did tell me that i was his responsibility, which i didn\'t get....oh well.       &nbsp;   hasta manana!       &nbsp;   : )   ');
INSERT INTO `msgs` VALUES (7204,3992573,'23,July,2004','2004-07-23 00:00:00','   ya ya ya i know i know, its ridiculously early, but hey this is what time i get into work, so this is the time that your going to have to deal with reading this! anyhow, last night was o.k. i got out of here went by yeya f\'s house, quick lesson in greek, yeya is greek for grandmother, yeah i\'m greek. anyhow, i left there & went home, it was nice because p was the only one home with the baby. mom &amp; f went to shop for sunday (p\'s surprise party) and who knows where dad went. but saw that i wasn\'t feeling well & was good enough to not leave the baby with me because i really didn\'t want to be holding her because i\'m sick &amp; i don\'t want to get her sick. ok quicky about the living situation. mom, dad, f (sister), p (brother in law) & z (their baby) ya we all live together. we just built an addition onto the top of our house where they will evetually live, when it\'s done &amp; at this point it doesn\'t seem like it will ever be but he\'s shooting for the fall so we\'ll see. it\'s a great big happy damn family so don\'t judge it unless you know! i love them all don\'t get me wrong, but damn sometime i don\'t feel like talking, i don\'t feel like entertaining everyone, but if i\'m quiet for like an hour oh what\'s the matter? are you upset? sometimes i am, but most of the time it\'s just because i don\'t want to frigen talk! but the situation is just for now. i have been saving my pennies & want to move out on my own very soon. i have been ready for a while but i don\'t want to end up moving out &amp; struggling, i\'m just being smart right now about it. because once i move out i don\'t want to go back! so anyways, they all came back home after a while & brought thea (aunt that\'s visiting from greece for theos wedding next weekend) theo (my mothers brother) we had dinner &amp;amp; then jo called. i went & picked him up &amp; he came over for dinner. jo is my best friend. jo is gay & i love him to death. he &amp; i were \"dating\" for a few months until he came out to me & told me that he was bisexual, i knew don\'t get me wrong, but the whole dating thing quickly ended at that point &amp; we\'ve been best friends ever since! well he came over we had dinner & coffee &amp; that was it. it was ok, but i was just so tired & in so much pain that i didn\'t want anyone there, but they were there so i made the best of it.      so anyhow, now the weekend is here, thank god! i felt that this week would really never ever end. it\'s just been hard not feeling well, working, going for that damn mri yesterday which had to have been the worst experience ever! i just can\'t wait to not do anything for once. but i can almost bet my life on the fact that i will be bothered to help f with the party because well as much as i love my sister she is just incapable. you will quickly learn to know this as the days progress.      until monday.....        &nbsp;    : )   '),(7205,3992573,'22,July,2004','2004-07-22 00:00:00','   ok so here it is, my second blog..yeah i know dont get so excited all at once. anyhow, i just realized that i havent told anyone about me! well here\'s my deal, i have no \"man\" i have a meta (and for those of you who don\'t know, thats ok you\'ll have to deal with it. he & i have been \"together\" for 4 years in october known eachother for 7. i have no clue where it will go or when it will end but for now its good. i love him &amp; all, but he just confuses the hell out of me so i\'m not worrying about labeling us anymore, we are what we are & what that is i have no idea. and then i have my pl. now for those of you who are really into this blogging thing then you will know what that is, for those of you who don\'t pl stands for prison love.&nbsp;mine has been down for 5 years now with about 2 more to go. it sucks. but that\'s ok because i make my girl come with me to ease the pressure a little bit. but to be totally honest with you i am really starting to get sick of it. he thinks that he knows everything & the reality of it is is that if he did know anything he wouldn\'t be where he is as we speak! but he is &amp; that\'s the reality. but it\'s just starting to get on my nerves. plus the more that i think about it the more i realize that i would have no life with him when he gets out so what the hell is the point? i need things & his whole perspective is that why would you need anything other than food, water &amp; shelter? oh ok african rasta king, my bad, i must have been the one who had 15 pairs of tims & separate socks that went with each one of those. and i was the one that had the&nbsp;5 avirex leathers that went with the 5 pairs of sneakers that you owned just for the coats...oh lets not even try to get into the sneakers, sneakers for every day of the week, every outfit, season, playing ball, playing ball outside, playing ball inside, playing ball outside/inside in the winter/summer. and trust when i tell you that these were all different sneakers...but i\'m materialistic? oh my bad!      ok my bad, i just took like a 4 hour break from this & just realized that i had to finish it!! that\'s just the kind of person that i am though so deal with it.      anyhow, i think that i have wasted enough of my time writing about big I.      j &amp; i just went to the po (privacy office) & we each got a pl (prison letter) and i have been releived of my duties for this saturday (halelujia just sang out through everywhere) supposedly my beloved has a special visit. whatever, better for me.      anyhow, i guess that i should work now, all is well, julias not having to take a ride in the accord to the you know where to do the killings, i dont have to go the pl on saturday oh &amp; FYI meta still hasnt called me......OH        &nbsp;    LOVE YOU ALL : )   '),(7206,3992573,'21,July,2004','2004-07-21 00:00:00','   ok so lets see if this damn thing works, considering that they got me hooked to it now & i felt the absolute need to share my life with people that i don\'t know!!        hola to everyone who is going to read this.       peace love &amp; hairgrease          &nbsp;    : )   '),(7207,3992573,'10,August,2004','2004-08-10 00:00:00','   yeah yeah yeah i know, it\'s been a few days since i\'ve blogged, but hey i actually have work to do sometimes when i\'m at work, so i didn\'t get a chance to do it, stop harrasing me!       anyhow, what\'s happened since the last time i blogged hmmmmmmmm. oh ok we\'ll start with friday. i came here (obviously) left here & went home. then i picked j up &amp; we got take out & ate &amp; hung out at my house for a bit then i took him home. i swear i got back home at like 10 & went straight to bed i was so damn tired for some reason. then saturday i went to go see big I. we had a really good visit for once actually. i think that we\'re both coming to the terms that this is actually going to end at some point &amp; i think that we\'re actually both ok with that. i mean honestly though how long was this really going to last? we both knew that it would eventually come to a hault. but i didn\'t think that it would be so easily done ya know? i mean i really thought that i cared about this kid, but now looking back on it, it was just something that i was so used to doing that made me keep going. i hate change, everyone who knows me knows that. but i think that this will actually be good. so i\'m most likely going to continue to go up there for a while until i feel that i\'m ready for a clean break. because i know myself was too well & i know that if i try to cut it right at the seams right now i\'ll fall back to him quickly. so i need to pace myself. slowly but surely it will end. so then i came back home &amp; the sister asked me to watch the baby so that they could go to the movies. sure why the hell not. no actually i didn\'t mind at all because i knew that i wasn\'t going anywhere for a while so i didn\'t care. so they came home i went over j\'s house & we had dinner. then i picked a up from work, we went back home got ready &amp; went out. we had a good time actually. we met s & m (ha i just realized those initials!) at the sole for drinks then her &amp; i went to the hot spot. we had fun. then chocolate calls me & tells me the drama with her man (yes again) i don\'t honestly know what the hell is wrong with that man. i swear i just wish that she would let me talk to him &amp; make him open his eyes to her. he\'s so damn dumb. whatever though. then on sunday we went to the movies, the italian festivle then to dinner. so all in all it was a really nice weekend.        yesterday was insane here for some reason too. i don\'t know why. actually i do know why, it\'s because monday\'s are just like that, but yesterday it seemed like every 10 minutes someone was at my desk. hey i can\'t complain though at least i have a job which is so much more than a lot of people can say. and it\'s a plus that i like my job that i have too!       so mom & dad are going to greece too for a month. they just came back last night from the cape from the weekend &amp; all that their waiting for is their new passports to come in so that they can get their tickets & all that stuff. dad is a little pissed because he wanted to be there when the olympics are going to be there but what can you do? like i said, being in greece after the olympics is better than not being there at all! i\'m happy that their going to go though. they need it. also, we\'re waiting for the business to come back about the hot dog place. i swear i am praying for that place. not as much as i\'m praying for the other dip shit to fall flat on his face with the pizza place, but we need something trust me. so everyone reading this, say a silent prayer for us that this goes through.       anyhow, i guess i should do something here now.               HASTA MANANA!              : )   '),(7208,3992573,'06,August,2004','2004-08-06 00:00:00','   ok so we finally went to dinner last night! meta & i that is. it was nice actually, well other than the fact that we had another \"talk.\" it\'s so weird between us like that. i mean we\'ll be fine for as long as we are &amp; then i get thoughts & need to talk. and i know that it pisses him off cause last night he swore up &amp; down that he was never, ever going to have a talk like that with me ever again unless we\'re discussing when we\'re gonna have kids....um hello? kids? hold up baby your not even my man, but now we\'re talking about kids? oh. but all in all it was very nice. i like being with him, no matter how annoyed he gets me. i still like being there with him. he makes me feel safe. almost like nothing can go wrong. and i know that he cares about me, i know that for a fact. but the thing is is that he says that he can\'t commit himself right now because he\'ll end up cheating & he doesn\'t wanna do that. oh so this way you can do whatever the fuck you want &amp; not feel guilty? hmm must be nice! then he freaks out, no that\'s not what i mean. so i said look whether you wanna admit it or not, you & i are more than just friends, i know that relationships freak him out, i know that, but like it or not, this is what it is. and he tells me that he doesn\'t want to think about that. ok then don\'t think about it. so i ask him if he has been with anyone else &amp; he starts laughing. hi i\'m a joke now? so i asked him why he was laughing & he tells me that he doesn\'t even talk to other girls why would he be with anyone else? and furthermore, if he wanted to be with someone in any way why wouldn\'t he just be with me? ok are you asking me or telling me? but whatever. we had a nice time. i want to spend more time with him but it\'s just so hard. he\'s not living here &amp; all. damn getting older sucks. having to deal with shit like this is gross. plus i want him to go to law school which he is going to do next september. he\'ll be around for 1 year & then he\'ll transfer to a better school where he can major in the type of law that he wants to practice.  i want him to go through with his plan & not back track. so i guess that i just need to suck this up for now. and in the mean time keep doing what i\'ve been doing. i love him too much to let him go. plus, as sick as it may seem, i actually think that, sorry, i know that i could spend the rest of my life with him. omg i can\'t even believe that i just said that. but him &amp; i started off as friends, so as a friend i need to be there for him....that\'s my decision, mature huh? yeah i know.       and the best part of the evening...doing the things behind a fucking tree behind circuit city....WHY! hell after 4 martinis i think i would have done him in the damn booth!              anyhow, i\'m out. have a great weekend, cause i know i will! plus white girl is coming back from vacation on monday, so we\'ll all be 1 big happy family here again....yeah right.              HASTA LUEGO!!   '),(7209,3992573,'04,August,2004','2004-08-04 00:00:00','   ok so why do we really need to be at this place again? there are so many other things that i could be doing right now....like sleeping. but no. instead i\'m here trying to figure a lot of shit out. i don\'t even know where my mind is lately, but i have been really trying to figure shit out & it\'s really not as easy as it seems like at all! ok so the 1st thing is that i want to be with someone. i don\'t know who that person is right now at this time or anything like that, but i know that i am now ready to be with someone. i want to get married, i want to have children, i want to have a normal life ya know? but then the more that i think about it the more i realize that i don\'t know if that\'s what i really want. because ok look at this, white girl has been with her man for what like 9 years? has 2 kids with him, lays next to him on a nightly (well unless either of them is on the couch for the night) but the point is, she still doesn\'t trust him. and that sucks. i mean she is 1 of the nicest people that i know in this fucked up world, and she doesn\'t have piece of mind. i mean deep down i know that she knows that he would never do anything to hurt her. but the bottom line is that she is basically married to someone that she doesn\'t trust. is that how i\'m gonna end up? and i gonna end up lonely &amp; afraid? god why.   and then on the other hand we have chocolate kisses. damn it she\'s my girl but i wanna take that big ass man & slap him so fucking hard that his rose bush giving mother feels it! i just wanna be like hello asshole do you have any idea what you have with her? damn both of them for that matter, they have no clue what they have in these women!   ugh i don\'t know. maybe it\'s just my mood today, but my eye won\'t stop twitching &amp; these thoughts won\'t get out of my head.   i need to go to dinner with meta so he can make me feel better. maybe it\'s also because i haven\'t done the things in like a month & it\'s starting to get to me. i just need something or someone to change my thoughts to better ones. and he can usually do that......   let\'s pray that it happens.      HASTA MANANA!      : )   '),(7210,3992573,'03,August,2004','2004-08-03 00:00:00','   ok so i was really going to do more yesterday, but here i was actually doing my work when frigen spoon comes over to introduce the new girl to me, yeah hi, then she tells her that she\'ll be sitting with me for the rest of the day..shit. i totally forgot that C said that she would be sitting with me. damn it. i didn\'t even prepare anything for this bitch to do. now i\'m gonna have her crawled up my ass all day. wtf. so whatever. she comes back & sits down, yeah b get comfortable. i explain everything that i do to a girl who has no clue about anything that i\'m saying. then she starts. she tells me about her fiance, his mother who they live with, the fiance\'s mothers fiance that lives there too. how she got locked out of the house last week, how her fiance &amp; her had a 10 year anniversary party for themselves. how the mother\'s fiance is a crackhead. all damn day people. this went on all fucking day. i was so happy when 12 came along because she went to lunch. i would have gone with her but couldn\'t stand the thought of listening to more. i don\'t know you chica, chill the hell out with telling me your life story! so she goes to lunch, comes back & then goes to training until 3...oh good i leave at 4, we\'re just gonna have her sit with you until you leave, yeah that\'s exactly what i wanted. ugh.       anyhow, chocolate kisses had a horrible day too. she basically got told that she doesn\'t do her job well & that she\'s too slow. oh and that her &amp; i have a conspiracy towards private ryan. ok # 1 bitch we can like/dislike any frigen person that we want. # 2 shut the fuck up. leave us the hell alone & let us do our work whatever way we get it done! nah instead their crawled so damn far up her ass that they can\'t even see daylight. let it go. so now we\'re back to not being able to talk anymore at work. it\'s happening again. so that\'s fine. we don\'t talk when we\'re here. we talk after work &amp; on breaks. fine. ugh.      so i leave here all annoyed & was going to the gym &amp; a calls me. oh i missed you so much blah blah blah. so i was like look whatever man, all that i know is that you guys dicked me. so if you want to be freinds with julia, we are not chillin with her when i am around. you can do whatever you wanna do with her, but minus me. she says fine. fine. so we go to the gym. on the treadmill great fun, daron walks by. FYI daron is, well, was my 1st, 1st everything. i love him. we were together for 4 years & it never actually \"ended\" we just stopped talking. but i would still meta his ass. no bullshit. so i say hi to him. well as we\'re walking away from the treads he was talking to some guy on another one &amp; says hey n i wanted to talk to you about something. so i stopped & was like right now? and he\'s like oh nah not unless your leaving right now. and i was like nope i have a couple of machines to do still so he\'s like oh ok then i\'ll catch up with you. and i said ok &amp; walked away. so i\'m on another machine & he comes over to me &amp; asks me why i keep winking at him because he doesn\'t know how to take it. i started laughing & asked him how he wanted to take it &amp; he was like oh i don\'t know. so i said look don\'t go down a road that you don\'t wanna go down. and he was like if i didn\'t want to go down that road i wouldn\'t be talking to you right now. OH BITCH. so i asked him what he wanted to talk to me about & he tells me that he wanted to get my number to see if i wasn\'t busy at some point this week. so i said fine &amp; gave him my cell #. OH BITCH. so he was like well are you busy tonight? and i had to say yeah because i was watchin the baby for f & p. but i told him to definately give me a call if he could get out of the house this week you know baby mom\'s &amp; all. and he was like if you give me a reason to leave the house then i would. oh. so i started laughing & just told him to call me. ok i\'m flirting. i\'m flirting with daron b. oh. good. now i will die. i\'m sorry did either of us not get enough ass beatings when we were together that we have to go back to eachother? wow. it\'s so weird he has like this magnetic pull with me &amp; i just don\'t know why. but this is the way that i see it. if a man can see your ass at the gym, therefore at your worst time, and still try & holla, he\'s not goin anywhere. but then again he\'s seen my actual ass so i can\'t really say anything about that! so let\'s see if he has the balls to call me. cause i will answer.      and my poor meta wants to try & do dinner this week again \"if i don\'t act up\" oh i act up now. ok i\'m just gonna let that go because i don\'t want to start something that was my fault initially. but he\'s taking the fact that i apologized about something he is taking to his head. so he keeps bringing it up. let it go baby. i will say this honestly though. i do miss his ass. i just want to be with him for a while to get my fix &amp; then i\'ll be straight. i just miss him. yeah i know i sound corny but whatever.      now i am going day by day because mom & dad are very serious about going to greece for 3 weeks WOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO   HOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO      ugh i will blog more tomorrow i gotta work.      hasta manana       : )   '),(7211,3992573,'02,August,2004','2004-08-02 00:00:00','   ok so it\'s monday again. yeah i know i didn\'t get to blog on friday but i had way too much work to do. it was so weird last week, it was like i was working my ass off but didn\'t seem to be getting anything done! but it\'s a new week, it\'s going to be a good one too, i can feel it!   so the weekend....well friday night i went to go too Big I. chocolate kisses was supposed to come with me to see Big Cherb but she lost her i.d. & even though i told her to still come because they would have let her in regardless she didn\'t want to risk coming up there &amp; having to sit in the waiting room & waiting for me. even though i told her that they would still let her in she didn\'t want to listen to me. and low &amp; behold when i got up there to co that was there told me that it was more than ok as long as she was with me she would have been fine.....OH. anyhow, the visit was ok i guess. it\'s really weird actually. as much as i am mad at him, as aggravated whatever i just can\'t stay mad at the kid for some reason. and just as i was thinking that he said the exact thing to me! it was really weird. but i stayed for a while, left there went home hung out there for a while then went to bed. saturday was theo\'s wedding so i was up at 7 to go get my nails done. then i went home showered & went to get my hair done. in the meantime jes tells me that julia invited a &amp; j to the beach house for the weekend. hmmmm that\'s interesting no one said anything to me about the beach, nice. so i go get my hair done & ask j if he\'s going &amp; he says i don\'t know maybe, i just need to get away....oh because doing hair is a stressful job, yeah i get that. so whatever. we go to the wedding & it was ok i guess. it just didn\'t seem like a wedding reception for some reason. i don\'t know. j never showed up some bullshit excuse that he didn\'t have a car. i swear the kid tells so many lies that he probably doesn\'t even know his own name. so after the wedding i go home change &amp; went over there. julia was there that stupid jealous bitch & j, a hadn\'t gotten out of work yet. so i didn\'t even look at her 1 time the whole time i was there. so we\'re all talking &amp; shit. a comes home & sits down for like 10 minutes &amp; then says oh i need to go take a shower. so she does, comes back sits on the couch in her towel for like another 10 minutes & then asks julia so what\'s going on? and she says to her, my shit is in the car i have a full tank of gas so we\'re just waiting for you. to which a responds, oh i\'ve had my stuff packed since this afternoon. you stupid bitch. when i talked to you this afternoon you told me that you were stuck in the middle &amp; that you didn\'t know what you should do & blah blah blah. oh but you were so confused about going or not going because i wasn\'t invited that you just packed your shit? whatever man. so on that note i got up, went to the bathroom went back in there grabbed my bag &amp; said ok guys i\'ll talk to you later & left. i didn\'t ask them when they were coming back or anything. fuck them. she\'s gonna be so damn jealous that she won\'t invite me to go the beach house? ugh whatever. i\'m not gonna get aggravated this early in the morning. so after that i just went home &amp; went to bed.    yesterday i went food shopping with f & we cooked dinner &amp; prepared dinner for tonight. we ate rack of lamb yesterday & tonight it\'s spagetti &amp; meatballs. tomorrow eggplant parm. we never set things up like this but hey maybe it\'s a good thing that we do!   now let\'s see what happens today with the assholes.   i\'m just happy that i have true friends that don\'t talk behind my back. and if i seriously have to spell it out cause y\'all don\'t know chocolate kisses & white girl good lookin out.      well signing off for now kids.      hasta manana       ; )   '),(7212,3446654,'25,May,2004','2004-05-25 00:00:00','   Alot of you have already been informed of my \"Baby Seal Quotiant\", surely you didn\'t think I would start a blog and NOT include the Baby Seal Quotiant. For the initiates, let me explain what it is, and from whence it came.   The Baby Seal Quotiant started as a theory based on principles that clubbing baby seals could be a more effective and thorough use of bandwidth than what I am currently using. I actually have a screenshot of me downloading a PDF at 75 BYTES a second. That\'s right boys and girls, a blazing speed of 75 BYTES a second. I thought to myself \"Boy, I could probably just call the guy and have him transate the entire document into binary over the phone whilst I type it in and re-compile it faster than the way I am recieving it now..\"  This led me to think of common ways (and some not so common) of transmitting data rather than plain ol\' wires and radio and stuff. So here it is...enjoy...  PART 1   Take into consideration that all these are dependant on something being able to accurately measure the changes etc that I\'ve listed. I have a million of these in my head, but Ill just start with the most \"down to earth\" ones. Well...that last one isn\'t so down to earth but it\'s horribly funny...    1. A dead bug smashed on the visible side of each tire rotating on 4 cars traveling at approximate speeds of 135.5 KM/H, recording at each full cycle of rotation marked by the revolutions of the dead bug would be a more effective transmitter of data than I am using.  2. A cloud passing in the sky that makes 601 tiny changes in it\'s exact shape over the duration of one second would be transmitting data faster than me.  3. A heavy rainstorm falling on a tin roof is almost enough for me to begin thinking about a game of Counter-Strike. But if you want math, you can figure that it would only take a piece of tin about 26 feet square, at 2 raindrops per square inch per second to exceed my current speeds by a full 6 percent.  4. A laser pen, shot through a snow storm, going through an average rate of 1 snowflake per inch and measured only 1 time a second, would only take approx 50 feet of laser to accomplish speeds comparable to my current speeds. If it was snowing really hard, and I got an average of 2 laser penetrated snowflakes per inch, you could cut that down to 25 feet. 25 foot of laser in a heavy snowstorm sends data more quickly than my connection....  5. If you dropped a plate on the floor, and upon impact it broke within a half or second or so (pretty reasonable), the plate would only have to break into 301 individually different pieces, moving continuasly for at least one second, to be faster than my connection to load theonion.com  6. A semi dense carpet of grass, with lets say 8 blades of grass in each square inch. The wind blows. The movements of the blades of grass in a 6 foot 4 inch square area could match my speeds.  Here\'s the kicker..  7. Approximately 200 men simultaneously clubbing 200 baby seals at 3 hits per second (they really hate baby seals) would be transmitting data via the screams of dying seals at the same rate I am now at. If one guy gets in an extra hit during that 1 second, then the clubbing of baby seals is a more effective internet connection than mine...  Nice eh? I like that last one in particular. I\'m going to start a new bandwidth measure called \"The Baby Seal Quotiant\". It\'s a measurement of bandwidth based on how many baby seals one would have to have clubbed in one second at 3 hits a second to match the bandwidth your seeing. To HELL with megabytes, etc. Baby Seal Quotiant is something more people can identify with....  PART 2:  Ok...I\'m putting two guys per baby seal, and upping the hits to 4 per second per seal. That makes 2 seals a byte. Nice and easy number to work with. Of course if we took into measurment the amount of blood splatter in conjuction with dying seal bleats, we could probably work 2 seals up to a kilobyte pretty easily...it would only take 256 individual splatters of blood, per hit, per seal to make a kilobyte. I got it...the blood could be the actual data packets, and the bleats are header info... LOL ok cool. We\'ll go with the blood dispersion method. If so, then 2 seals equals a  kilobyte ..so...  2,048 beaten seals equals 1 meg  (2 men at 2 hits each per seal per second)   So the conversion is set at:  1Mbps = 2MBSps (that\'s MegaBeaten Seals Per Second)  the reverse is  1MBSps = 512Kbps  for every 1MBSps, there\'s also a 512bps header (the leftover un-used dying seal bleats). We\'ll refer to the seal bleat header as the SB data channel. Since it\'s comprised of a totally seperate transmission medium (using sound waves rather than blood dispersion patterns) it\'s use is highly configurable, not too mention wireless. There\'s even a system failure warning that could be incorporated. When the primary transmission method falls below acceptable levels (IE the seal just isn\'t bleeding as much as is needed for effective transmission), the SB data channel can still be utilized until the Baby Seal goes into complete system shutdown (IE the baby seal is dead)...  I think this could go somewhere..clubbing baby seals will be in the vogue again!  '),(7213,3446654,'24,May,2004','2004-05-24 00:00:00','   Well, I\'ve done it. I started a blog. Though I\'ve always mulled the idea of a place to post my insanely meaningless thoughts and ideas, I always had the gut feeling I would be following the tide of most other bloggers...mainly leading to old, un-used blogs. The blog scene seems to be like fishing for credit card numbers from people\'s trash. You have to dig through a bunch of dirty diapers and clumps of used cat litter before you hit the jackpot and start going crazy on E-Bay.  I have to be honest. I signed up here at Blogger.com in hopes of getting a gmail account from Google. My buddy got a gmail account becuase he has a blog here. This whole thing of folks getting upset about privacy issues confuses me. If a google bot wants to scan my e-mails for targeted ad\'s, it\'s fine with me. Maybe it will help cut down on the \"Live and Work in the USA, apply for you US Visa now!!!\" or \"Shoot the mad Kangaroo for a chance to win  \"  So there you have it. Currently I have many ideas floating about I can put on this blog, but we\'ll see if it continues to be updated. Only time will tell I suppose...  '),(7214,3659651,'31,July,2004','2004-07-31 00:00:00','   urlLink        urlLink stuff is everywhere     Originally uploaded by  urlLink and_rew .     A vorpal mass of swirling chaos slowly begins to spin into neater pockets of relatedness, lists lengthen and are checked slowly off. Contents of emptied drawers are slowly sorted and compacted. Emotions and plans and uncertainties slowly pass into one: it is almost time, I will soon be sent forth. Slowly! Slowly, I am getting ready for the next phase of my life... '),(7215,3659651,'11,July,2004','2004-07-11 00:00:00','   this blog coming soon!!! '),(7216,3659651,'12,August,2004','2004-08-12 00:00:00','   Well that thing where you start to realize you\'re actually leaving is actually happening to me now. We went to eat Chinese Food last night. I love that in America, when you want to enjoy some American food before you leave, you go eat at an Italian place, a Mexican place, an Australian place. And Whataburger (several times).  Today I will say bye to many people and spend this evening doing the last bits of packing. Dad is going to grill something for dinner.  Anyway, I got up kinda early so that I could start getting a bit closer to German time. I am going to eat some Honey Nut Cheerios. Maybe I should put a box of those in my suitcase. I keep thinking \"I should bring this with me\". Also that cookie cake Becca and Mom made. They need to make me some of those to take with me. But the suitcases are packed, and stuffed quite full! So I hope everyone enjoys all their AMERICAN STUFF a little bit extra on my behalf this year.  Breakfast time. '),(7217,3659651,'10,August,2004','2004-08-10 00:00:00','   urlLink        urlLink my room and our homeshrine     Originally uploaded by  urlLink and_rew .     Can I put our living room, couches, homeshrine, and room into my suitcase? '),(7218,3659651,'10,August,2004','2004-08-10 00:00:00','   urlLink        urlLink missing the sound effects     Originally uploaded by  urlLink and_rew .     Can I put Zach and the playmobiles in my suitcase? '),(7219,3659651,'10,August,2004','2004-08-10 00:00:00','   urlLink        urlLink Weirdo and Birthday Girl with Family     Originally uploaded by  urlLink and_rew .     this is just a really swell picture of some really swell people who agreed to let me take a picture next to them. '),(7220,3659651,'09,August,2004','2004-08-09 00:00:00','   urlLink        urlLink WE LOVE GRANDMA MINNIE!!!!!!!     Originally uploaded by  urlLink and_rew .     Having a bit of birthday cake, made by Becca. Do they have cookie cakes in Germany? Saurkraut cakes? '),(7221,3659651,'09,August,2004','2004-08-09 00:00:00','   urlLink        urlLink mom packs, i take pictures     Originally uploaded by  urlLink and_rew .     we\'re working on it. '),(7222,3659651,'09,August,2004','2004-08-09 00:00:00','   It was super nice to see everyone who came. Thanks for coming, people who came. I am lucky to have so many really good people to support me and see me off. And bring me presents.  Below are many photographs of the weekend. There are many more at http://flickr.com/photos/andrew_/. I did not post a certain picture here, because Dad didnt want me to spill the beans... about why the beans tasted so good! Everything else tasted really good, and not because it was spilled all over the floor (oops! sorry Dad), and I hope everyone had as much to eat, and as much fun, as I did.  At this moment there are stacks of clothes and a few other piles of things assembled upstairs in the computer/school room. Things are coming together, but there\'s a lot more to do to be ready for Friday. And its still just beginning to dawn on me that I will soon be in SchÃ¶nstatt. Everyone being here at our house made me realize more than ever that home is with the people you love and are related to. So going to Schoenstatt will be the biggest homecoming of my life. So right now the main part I\'m worried about is getting from Frankfurt to Koblenz, and then being sure that I have enough underwear and Advil to last a year!!! '),(7223,3659651,'09,August,2004','2004-08-09 00:00:00','   urlLink        urlLink Molly, Polly, St. Paul, and Me     Originally uploaded by  urlLink and_rew .     Circle the one who doesn\'t belong!  Hint:  The weirdo in the middle who still has to go earn his halo. (ME) '),(7224,3659651,'09,August,2004','2004-08-09 00:00:00','   urlLink        urlLink Too Much Hair     Originally uploaded by  urlLink and_rew .     My godson and his godfather. (This is actually the best of about seven similar photos.) '),(7225,3659651,'09,August,2004','2004-08-09 00:00:00','   urlLink        urlLink Joey and Weirdo     Originally uploaded by  urlLink and_rew .     Bye Joey! Bye everyone! Smile and say SAURKRAUT! '),(7226,3659651,'09,August,2004','2004-08-09 00:00:00','   urlLink        urlLink Irina likes this kind of beer a lot     Originally uploaded by  urlLink and_rew .     Yes, Mr and Mrs Rakalina, the girl with the Miller Lite is your daughter. '),(7227,3659651,'09,August,2004','2004-08-09 00:00:00','   urlLink        urlLink Setting up     Originally uploaded by  urlLink and_rew .     Lovely decorations! '),(7228,3659651,'09,August,2004','2004-08-09 00:00:00','   urlLink        urlLink People get here     Originally uploaded by  urlLink and_rew .     there were lots of nice people at my party. '),(7229,3659651,'09,August,2004','2004-08-09 00:00:00','   urlLink        urlLink Esteemed Guests     Originally uploaded by  urlLink and_rew .     Nice party. :-) '),(7230,3659651,'09,August,2004','2004-08-09 00:00:00','   urlLink        urlLink Flashy Room     Originally uploaded by  urlLink and_rew .     Some of the wonderful young people in attendance. '),(7231,3659651,'09,August,2004','2004-08-09 00:00:00','   urlLink        urlLink I need a Haircut     Originally uploaded by  urlLink and_rew .     Nice new blanket. '),(7232,3451166,'28,May,2004','2004-05-28 00:00:00','   got im bored.yo steve today mrs v was a beatch.but whatever.class was boring after u left.r u gonna bring your box to school next week? i  hope so cus i want to beat the hell out of u in halo.jk. but seriously....                                                                  thats about all i got peace   '),(7233,3451166,'27,May,2004','2004-05-27 00:00:00','   give me money tick-and tell ur friends to do the same--i need 500,000 in unmarked bills in the dumpster behind my house by 6am tomarrow morning--and if i go sifting thru that pile of shit and i dont find my loot, ur fucking dead, every last one of you.  ill hunt u down one by one, skin you alive and smear rock salt into every square inch of your bloody staggering carcasses...  j/k! i luv u guys!!   '),(7234,3451166,'27,May,2004','2004-05-27 00:00:00','   not duck? what the hells goin on, im phreakin out--meh ive got evol intent tonite but tomarrow i mite be able to chill--i dunno where tho...gotta go fix my car now, peace!   '),(7235,3451166,'27,May,2004','2004-05-27 00:00:00','   nah that wasnt duck, gary. and please, if you need something(money, ideas, whatever) please dont hesitate to ask me. i will help you out in whatever way i can.. as for a place to stay, im sure that if you dont have a place to stay, my mom can rent out my room at her house for you. he has had many people stayin at her house before, and as long as your willing to help with th rent every month and not eat all their food, i dont doubt that she will let you live there. and i havent spent ONE SECOND in the realm of the cock so dont even fuck with me you ass-faced jew-banger! lol. im workin tonight and saturday morning, but after that, if you want to do something, give me a call.  -Â§Ã‰LÃ‰k7Ã¥-   '),(7236,3451166,'27,May,2004','2004-05-27 00:00:00','   You do know me silly pants!!! And as of getting married, #1 Although we have met, I am sorry to say that I don\'t know you  that  well... # 2 perhaps in the future, if you are not threatened by comments such as \"mr wheat, marry my girlfriend and i will eat you.\" then we might be able to consider it.  As of now, I think the answer is no. Sorry. Good news all, tomorrow is Friday, and then that means 3 days of no schools! :oD!! PEACE!   '),(7237,3451166,'26,May,2004','2004-05-26 00:00:00','   sorry bra, i was confused--good for u tho, hope that whole vagina thing works out for ya, with all of your experiences lying in the realm of cock. now was that duck trying to talk to me? honestly i dunno what the fuck hes squakin about. so life\'s ruff without the barn--just yesterday i put a 2D rpg in my ps2--lunar silver star story--fantastic game, but only tolerable in boredom. the thing is i have so much work to do--getting my car fixed,getting a place to live, packing all my shit, but i lack motivation. and its not like i have anyone to push me. ive always given off this image of self-sufficiency, so now when i want help i cant ask for it because it dissassembles everything ive worked for. i always thought myself to be one of foresight, and in a very shallow way i am--i often kno what to do in the morning if i want to achieve happiness later in the day but thats about as far ahead as i can look.  o well, elixirs tonite and tomarrow evol intents playing at element so everythings cool.   '),(7238,3451166,'26,May,2004','2004-05-26 00:00:00','   dude. jenny is my girlfriend. and i can assure you, she HAS a vagina :P   '),(7239,3451166,'26,May,2004','2004-05-26 00:00:00','   i mean geek o wait gee.. there we go... so im dont even have cable because im a hick from nowhere and i can not see chapelle even if he was as visable of your leg hair. and your smack talk about Steve-o man we will kung fo fight in peru.... we will take it up in the middle of the street when you want to talk smack.. steve needs the comp so i got to bounce   '),(7240,3451166,'26,May,2004','2004-05-26 00:00:00','   mr wheat, marry my girlfriend and i will eat you. and yes jenny, your are my bitch:) sigh. i hate what the ny state educational system has to offer. i am busting my little white ass to get a 70 in math class. that sucks. and gee...dont mind dutko, he knows not what he says. i dont think he can handle such a merciless slew of insults that you issued me. Â§Ã‰LÃ‰k7Ã¥!   '),(7241,3451166,'26,May,2004','2004-05-26 00:00:00','   ...seriously, im desperate and lonely...stop laughing! im dying inside!...eh, forget it. what have i gotten myself into? as if having my own weblog, where i could seek pity from complete strangers with equally pathetic lives wasn\'t sad enough, here i am battling wits with the only person ive ever known to have taken a wet plunger to the face, warding off the inane babblings of a chappelle-posessed half-wit with the voice of a woodland fowl, and proposing marriage to a someone ive never met who may or may not have a vagina...if this were a glass of fruit-juice we would be the soaking pulp at the bottom, and i the fattest chunk.   '),(7242,3451166,'25,May,2004','2004-05-25 00:00:00','   Hi Y\'all. I\'m Jenny, not to be confused with Steve\'s Bitch. (lets see how well that goes over) Anyways, I was forced to delete my first intentional entry because someone didnt like that, not to name names... And to who ever Duck-O is, you should obviously talk to me, because I am apparently the only one in this \"interactive shit\" that has a vagina. :o)  Peace   P.S.  Hows that for a title? Sucka   '),(7243,3451166,'25,May,2004','2004-05-25 00:00:00','   What Is ALL THIS JIB-ER-JAB-ER ABOUT????? WHO SHOULD I TALK TO??? YOU??? IM RICK JAMES BE-OUCH   '),(7244,3451166,'25,May,2004','2004-05-25 00:00:00','   sigh. once again i have been dumbfounded by gary wheat\'s endless slur of words that i dont understand. fuck you. thats all i can say. i must tell you, that when i read the line about the plasma grenade, i started laughing so hard that i began to cry. that had to be the FUNNIEST fuckin thing iv ever heard in my short, plotless life. thank you sir. and may the better man win. my balls should drop in a week or so.  i present the award to gary wheat, for the funniest fuckin thing iv ever heard.   -Â§Ã‰LÃ‰k7Ã¥-   '),(7245,3451166,'25,May,2004','2004-05-25 00:00:00','   well see how tuff u r when ur lying face-first in a puddle of ur own grey matter, staring at the concrete thru ur bloody tears. maybe when u stop using leetspeak i mite take u seriously--until then ur nuthing to me but a sniveling, dickless nub who wouldnt no a plasma grenade if it was burning blue on his snot-ridden facemask, u swashbucklin, shit-knuckled son of a fuck--u wanna halo match--call me when ur balls drop  o btw: congrats on the road test, friend   '),(7246,3451166,'24,May,2004','2004-05-24 00:00:00','   new template. new look. easier to use. mmhmm. :)   (i schooled chris and steve-the- german at halo at the same time:P) and yes, gary wheat, i will indeed school you. i will step with stealth and glee throughout the gravestones of the hanging of high. i will don my triangular cloak...and i will beat you into the floor with two deafening swats from my gas operated, magnetic-fed MA5B assault rifle. thats right. i am going to Rox0r jo0, and im going to make you feel like a small, helpless newbie.       -Â§elekta-   '),(7247,3451166,'24,May,2004','2004-05-24 00:00:00','   yeah so im ben. congrats to steve for passing your road test. i should get goin on that huh? this blog thing is pretty cool i must add.so yeah im real behind in my drawing class and my other art class so id better go work on that shit. peace my friends   '),(7248,3451166,'24,May,2004','2004-05-24 00:00:00','   ILL TRY THIS OUT FOR THE SAKE OF IT>... WHAT TO DO NOONE KNOWSS IM OUT YO   '),(7249,3451166,'24,May,2004','2004-05-24 00:00:00','   whoopdi-shit. i passed my drivers test. i guess i dont hate the dmv as much as i though i did lol. anyways. i should be getting a car soon. that way other people dont need to cart my ass all over the place. i got a job too. yyup, thats right. im moving up in the world. i got a job at a restaurant just down the street from me. i make 5.25 an hour under the table with at least 25 in tips every time i work. i am a happy happy boy. and when gee got busted, a picture of me smoking a cigarette was on his dads camera. lol, i am anxiously awaiting a bitch-session the next time i venture into the depths of avon. oh well. -Â§Ã‰LÃ‰k7Ã¥-   '),(7250,3451166,'24,May,2004','2004-05-24 00:00:00','   So my parents found out im not the godsend they once thought, but a raving, smoking drunk. So now im a stranger in my own house and i hate it. I need to get out quick. I need a real life. So much of my stress comes from my parents, but once im out so much of it will come from money--or my lack thereof.  All i want rite now is to make some decent money and pursue my DJing and music production. But its hard to work on the extra things in life when all the important things are falling apart.   '),(7251,3451166,'24,May,2004','2004-05-24 00:00:00','   the blog is growing so nicely! sweetness. everyone update a lot and it will be tons of fun. oh, and towels are the only things seperating us from the yugoslavians. peace.   '),(7252,3451166,'24,May,2004','2004-05-24 00:00:00','   My name is Gary, they call me gee. Blogs are for saps and ive been forced by one such sap to come to terms with my own sappiness. So here i am. Ill post something when i want pity--until then, cheers.   '),(7253,3451166,'24,May,2004','2004-05-24 00:00:00','   hi. this is kyle. my original post was \"fuck yo blog nigga\" but administration(selekta) had a problem with this. this is my formal introduction, so hi and bye. i will post frequently when possible.   '),(7254,3451166,'24,May,2004','2004-05-24 00:00:00','   hello, this is Â§Ã‰LÃ‰k7Ã¥.(steve mccorry). attention friends of me. this is a blog that i have created for the purpose of multi-user ranting and raving. i will give a select few of you the passoword so that you may add your thoughts or feelings. Rather than each of us having an individual site for this shit i have decided that a forum style blog would be a nice way to share thoughts and ideas. if you have any questions or concerns please contact me at theresbeerhere@hotmail.com   thank you. -Â§Ã‰LÃ‰k7Ã¥-   '),(7255,3451166,'04,June,2004','2004-06-04 00:00:00','   http://www.kingsofchaos.com/recruit.php?uniqid=9asj8s26 go to that shit and help me out and play too if u want to   '),(7256,3451166,'02,June,2004','2004-06-02 00:00:00','   my site is up...it has a new domain name also...so, the new url is www.sketchball.tk   i put up the super cool forum thingy also...check it out. -Â§Ã‰LÃ‰k7Ã¥-   '),(7257,3451166,'02,June,2004','2004-06-02 00:00:00','   yeah, the site is gonna be awesome. a friend of mine(the php programmer) is now making a shout box. a shout box is a section of the site where there is a text bar to type in, you type..add it...and it pops up in the chat. it keeps on re-updating itself everytime you refresh. it is going to be very similar to the degolburg square in lotgd.net  this should all be done by tomorrow or the day after. me and a friend have been slaving over the flash nav bar on the site. it looks fuckin sweet. and gary, the ideas(as far as the comic itself goes)remain the same.except i liked your idea of making it a community comic. there is still the process of creating characters and a story-line. if you have any ideas, i leave this partly up to you. i would like to keep tick and gee though. things are looking pretty good for the comic this time. i now have people involved who know what the hell they are doing. and better news...my friend murph(the php dude) is letting me house his server at my place. so not only will we have a kick-ass site...it will be on our own server with our own band-width. no more of that angelfire shit. good times. -Â§elekta-   '),(7258,3451166,'01,June,2004','2004-06-01 00:00:00','   Thanks Gary, I appreciate your sympathy. =)   '),(7259,3451166,'01,June,2004','2004-06-01 00:00:00','   id like to get back into it--but ill need some time to come up with something--post some ideas for me to work with and ill see what i can do--now that im not in school i should be able to find sometime for something like that--jes lemme no where u wanna go with it, alrite?   '),(7260,3451166,'01,June,2004','2004-06-01 00:00:00','   ah yes. i can see that this blog has grown substantially. i want to thank everyone for posting so frequently. that aside.  most of you know that i had a pitiful excuse for a webcomic up and running a little while ago. well...its down, and a new comic is in the making. foolish? yes. but this time,i have help. plus i know someone who is fluent in php prgramming. so, there will be a forum on the site not so different from blogger. members will be able to update posts and profiles, and comment on that days comic. i have gained more knowledge on how to build web-sites, and how to plan a biweekly comic. if anyone would like to help, please email me.  the new site url is www.angelfire.com/ill/selekta     it will be moved to a domain once the page templates are finished. gary, if you are still interested in drawing or writing for this comic, please, for the love of god, step forward and offer your mad skills to me. arigato.   '),(7261,3451166,'01,June,2004','2004-06-01 00:00:00','   but that was long ago...now they fester and lay waste to the only good--even now they plague our fair jenny...damn you ouchies--damn you...   '),(7262,3451166,'31,May,2004','2004-05-31 00:00:00','   Man, there is nothing wrong with sleeping until 5:30, I wish  I  slept until 5:30 today, but noooo school got the better of me.  Oh well Friday I\'ll sleep till 5:30.  WOOT FOR GETTING YOUR WISDOM TEETH OUT TWO WEEKS BEFORE SCHOOL GETS OUT!   PS. My scalp is sun burned and it is ouchie!  PPS My lip has an ouchie on it! Like a gigundo blister or something on it!! =(   '),(7263,3451166,'31,May,2004','2004-05-31 00:00:00','   yea its 5:30 pm and i jes woke up...somethings wrong with me   '),(7264,3451166,'31,May,2004','2004-05-31 00:00:00','   Alright, I didnt post that at 9:26 am, it was definately 12:28 pm. This thing is messed Steve   '),(7265,3451166,'31,May,2004','2004-05-31 00:00:00','   If sleep is for the weak, then I must be the weakest person ever, because I slept from 5pm on Friday to 6am on Saturday, and then was still tired.  I dont know what time I went to bed on Saturday, but I think I woke up around 9:30? And then last night I went to bed at like 12:30 only to be dragged out of bed at 9:30-ish again. AND TRUST ME IF I DONT HAVE SCHOOL TODAY THIS DAY IS DEVOUTED TO SLEEP! Seriously, I think that should have nap time at school, like they did in like preschool and kindergarten, Im sure it would make alot of people happier, I know I would be.   '),(7266,3451166,'31,May,2004','2004-05-31 00:00:00','   i cant feel my legs...im goin to bed   '),(7267,3451166,'11,June,2004','2004-06-11 00:00:00','   ur blog sux, get over it (emoticon)   '),(7268,3451166,'09,June,2004','2004-06-09 00:00:00','   cuz i havent seen a soul here in over a week! c\'mon people. lets here some bitching and complaining!   '),(7269,3451166,'07,June,2004','2004-06-07 00:00:00','   hey guys. i realy dont post as often as i should. but anyway steve you have to joun me in this game.click this shithttp://www.kingsofchaos.com/recruit.php?uniqid=9asj8s26 or copy and past the link   '),(7270,3451166,'14,June,2004','2004-06-14 00:00:00','   always a pleasure indeed gary. i dont see you hardly at all anymore. i dont know gee. haha. i am driving now, so there will be no excuse not to chill just a little more often! and jenny, what phrase did mr. wheat steal from you? good to see people actually posting. that way i dont have to post about people not posting, as gary so elequntly put it. ariht folks. LUNCH-TIME!! peÃ¥ce. -Â§Ã‰LÃ‰k7Ã¥-   '),(7271,3451166,'13,June,2004','2004-06-13 00:00:00','   Alright, Gary we need to talk... YOU CANT GO STEALING MY PHRASES! I know we are buddies and all but, lets keep that between ourselves, aight? And thats all I have to say about that.  Oh and it was nice to see you yesterday. Always a pleasure. :o)   '),(7272,3451166,'13,June,2004','2004-06-13 00:00:00','   I have been thouroughly enjoying my wheels. 95 Ford Taurus, not the fastest of the beasts..but it gets me where i want to go. I chilled with Gee, Jenny, Chris, and Testa yesterday. When i got up, i didnt think that i would be able to see all of these people. alas, i did. idk whats going on for today, but if anyone wants to chill, hit ther cell. -Â§Ã‰LÃ‰k7Ã¥-   '),(7273,4296269,'21,August,2004','2004-08-21 00:00:00','    It\'s been 10 days and counting since I last had a taste of my current addiction. I\'m not hooked on drugs or alcohol or even nicotine, no this girl has a taste for something far more sinister - something that digs its claws in and won\'t let go - something that sends you soaring like an Eddy Gordo Rio Special, eats you alive like Liu Kang\'s Dragon Fatality, slams you face first into the ground like Tina\'s Death Valley Bomb, and leaves you begging for more. I am a video game addict, there - I\'ve said it - it\'s out - no turning back now.      When I left home I wasn\'t sure how long I\'d be gone - family emergencies rarely come with a reservation or a schedule to follow - but I was confident that I could go without my game for a couple of days. I mean, I\'d lived 26 years without City of Heroes - surely I could make it 3 days more. Two days in I was taking it like a champ, but as the days turned into weeks it all started to go south. Hands shaking, cold sweats, downloading inferior net games like Mah Jong and Solitare and trying to get my fix. Normally I would have brought a console with me - I have plenty to choose from - but my jumbled mind left them all behind and stranded me here, cold turkey, trying desperately to hold on a few days longer until I can return to my life of heroism in Paragon City. Somewhere out there is a mental institution with my name on it.      Reading : GameStar new fledgling zine with a great preview of the new Silent Hill and Half Life, 1492, Neil Gaiman\'s perfect comic series, and Lord of the Rings as a single book not a trilogy as it was meant to be. I decided I had to read it again before the third extended edition was released at Christmas.       Watching:  The Amazing Race, addictive and highly entertaining - really the only reality TV worth wasting your time or your brain cells on. I\'m glad those brothers are gone, they gave not just our country but our species a bad name. Elephant, incredible and beautiful in it\'s simplicity. You want to know what your kids go through in the hell holes we call High School, it\'s all right there.       Playing:  City of Heroes, of course - community is the thing, find a good group of people with your style and enjoy the ride, Final Fantasy X-2 - I just want to get my 100% so I can feel like I\'ve gotten my money\'s worth, and Animal Crossing - yes, I know it\'s a kids game but it\'s damn addictive, you try to start your town then put it down!       Listening:  Muse - Absolution, if you\'re ready to be angry about what\'s happening to your country, to your planet, to your world (and if you\'re not, then you haven\'t been paying attention) then this is the album for you.   '),(7274,3831350,'28,July,2004','2004-07-28 00:00:00','    Canada\'s Premiers met today to discuss with the Federal government how to disseminate and spend Health Care dollars.   These dollars are much needed in an obviously taxed national health care system.   Much to the Federal government\'s chagrin, it seems that self-serving provincial premiers are trying to dictate how the money should be spent.&nbsp;   Frankly, there\'s a lot of humour in this don\'t you think?&nbsp; Dalton McGuinty, friend of all Ontarians, wants Ottawa to stay out of his way while he fixes health care.&nbsp; Well Dalton, ol\' buddy ol\' pal, I think you want to the money to actually pay for ALL of the improvements to our health care system.&nbsp; Hopefully if you come back with a big bone in tow, you\'ll actually be able to shorten some of those waiting lists you promised when you slapped us with a new health tax.&nbsp; Skimming from the cookie jar no doubt, the Ontario NPD have already been suggesting that funds from that tax are being applied to other government costs.&nbsp; Like&nbsp;a shell game with tax payer dollars.&nbsp; Or better yet, it\'s an illusion.&nbsp; Increase taxes, let the Feds pay for better health care and HE will take credit.   You may have also heard about the new user fees and toll roads from which Dalton and his team want to build a better infrastructure.&nbsp; Don\'t count on it.&nbsp; If you watch Dalton\'s history it\'s quite simple.&nbsp;   Step 1:&nbsp; Float an idea people really hate.  Step 2:&nbsp; Blame somebody else for creating the problem  Step 3:&nbsp; Implement an idea what people hate less than idea suggested in #1  Step 4:&nbsp; Sigh of relief when Ontarian are thankful that it wasn\'t as bad as it could have been.   Don\'t they use that those types of skills in 1984?&nbsp; Double-Think or something like that?   I guess we truly do get the government we deserve! '),(7275,3831350,'22,July,2004','2004-07-22 00:00:00','    Well it looks like the Liberal government of Canada wants to decriminilize pot again.  Do you think Paul Martin, Prime Minister of this great country of Canada is concerned about smoking pot?   Frankly, I don\'t.   I don\'t know much about politics but something tells me that this is a threat to the other parties (in which he shares his minority government) that they better play nice.&nbsp; The Conservatives can\'t support decriminaliztion of pot, however, they can also not afford to have government dissolve.   So, looks like Paul Martin\'s new slogan should be, when the going gets tough, the tough, the tough threaten to commit suicide.&nbsp; Silly thing is, it just may work.&nbsp; Any time the Conservatives won\'t play nice with Martin he just needs to show his \"dope card\" and threaten to play it.&nbsp; Parliament will dissolve and, low and behold, Canadians are in the throws of a new election with Martin hoping the Conservatives won\'t be able to win yet again.   And, surprise surprise, government will dissolve on a social issue.&nbsp; Something the conservatives don\'t have a good track record with.   Smart people in the Martin camp.&nbsp; Let\'s just hope they pay as much attention to running the country as they do in trying to win the next election.   &nbsp;   &nbsp; '),(7276,3831350,'21,July,2004','2004-07-21 00:00:00','    Mornings are pretty normal similar and sort of blend into each other. First off, my wife gets up, what seems like, several hours before me to the sound of the alarm.  Then, at around 6:45 she and Sebastian (a.k.a. Fat Cat) come and wake me with a pat on the leg and a meow.&nbsp; Sometimes they interchange roles and Sebastian will touch my leg and my Wife will Meow but that\'s rare and is usually fixed with an extra cup of coffee.  I love this morning ritual so much.&nbsp; I know it\'s not a big deal but there\'s just something extra special being woken up by those you love.&nbsp; Being woken up by a caress and not a siren.&nbsp;   Now, weekends is a different story.&nbsp; We both try to sleep in on weekends.&nbsp; Sebastian, the cat, does not.&nbsp; He has an internal clock that reminds him of his loving duty (I say with much sarcasm).&nbsp; He gets pretty demanding for his \"morning petting\".  The morning with Sebastian doesn\'t usually end there.&nbsp; Once I hope out of the shower he is crying at the door.&nbsp; It wants in.  He sits up on the toilet lid as I take care of the morning duties like shaving and gelling of hair.&nbsp; He usually tries to rub up against my legs and purrs quite a bit too.  Anyway, time for my shower...The cat wants up on the toilet....  &nbsp; '),(7277,3831350,'17,July,2004','2004-07-17 00:00:00','    Well, I made it through another week at the office.  This week, overall, was pretty average but dragged on way too long.  &nbsp;  I received hope and excitement that things may get better though.&nbsp; Lately, I feel that I\'m full of energy.&nbsp;&nbsp; I don\'t know why but I feel something is going to happen.&nbsp; There\'s anticipation and change in the air.  &nbsp;  I\'ve come to realize that state-of-mind is half the battle in life.&nbsp;&nbsp;  &nbsp;  All I need to figure out is how to bottle the good feelings and sell it to   urlLink  Gillette   .&nbsp; This state-of-mind truly is  the best a man can get .&nbsp;   I arrived at work by way of&nbsp;the subway, as usual, riding it to my stop in downtown Toronto.&nbsp; Instead of taking the escalator, I took the stairs.&nbsp; One battle won in the life long war against&nbsp;the sedentary liftstyle of the cube-dweller consuming e-mail and Doritos in one singular&nbsp;motion of the hand.&nbsp; Small gesture, walking up the stairs, perhaps, but I just made a dent.&nbsp; A small, but positive, dent.&nbsp; I felt good.  &nbsp;  I arrived at the office, no nasty voice-mail.&nbsp; No urgent e-mail either.&nbsp;&nbsp; Just another&nbsp;Friday thank God.&nbsp;&nbsp;&nbsp;7 hours and&nbsp;45 minutes!&nbsp;&nbsp;I proceed to&nbsp;brew a pot of coffee.&nbsp;    Then I read that there\'s possibly life discovered on Mars, supposed to be \"microbes\" (either that or it\'s a volcano).&nbsp; Life sounds better, let\'s go with that.&nbsp;  &nbsp;  The 2 robots, Spirit and Opportunity apparently cost several hundred of millions of dollars and they only put dentist drills and cameras so we can see brilliant black and white photography of dust on scraped rocks and landscapes that&nbsp;resemble the desert of Nevada.&nbsp; Thanks for the \"Opportunity\".&nbsp;&nbsp; Your \"Spirit\" of discovering life on Mars&nbsp;is overwhelming to me.&nbsp;   Don\'t get me wrong, I\'m a big supporter of the space program but this makes me sad.&nbsp; Nasa\'s had it rough lately and a big discovery sure would bring a smile to the faces of&nbsp;30 or so scientists who can make sense of blobby pictures of goo (excluding pictures from Hubble) &nbsp;that they always show on Discovery.&nbsp; Everyone talks like they understand how to make sense of some of the crap.&nbsp; Microbes from Mars don\'t impress me that much.&nbsp; Microbes are on my subway too.&nbsp; You can catch the flu from them.&nbsp; Isn\'t West Nile a microbe?&nbsp; I once caught pink eye from the subway. Since Mars is red, would Pink Eye from&nbsp;Mars be called Blue Eye?&nbsp;  &nbsp;  Then I read that Martha Steward is going to prison.&nbsp; Oh well, on news that this mess is finally over investors rallied around her stock&nbsp;sent&nbsp;it into an upswing and she made something like 90 Million USD today.&nbsp;&nbsp; I\'d go to jail for 5 months in a minimum security prison to make 90 Million US.&nbsp; Poor Martha!  &nbsp;  Noon arrives and I take out my sandwich and turn on the video feed from the Big Brother 5 house.&nbsp; I realize that I don\'t really much like the Big Brother TV show.&nbsp; I\'m enthralled with the web feed.&nbsp; I realize it\'s voyeuristic but that\'s not it.&nbsp; I think to myself, \"it\'s validation of the ordinary and that\'s why it\'s beautiful.\"  &nbsp;  The other day for example, one of the players woke up, flexed his muscles like Hulk Hogan,  grimaced &amp; farted .&nbsp; I smiled to myself.&nbsp; My buttercup leans over to me and says \"You do the same thing every morning.\"&nbsp; I keep smiling, pondering what genetic advancement would cause this behavior what would have made it such an attractive trait that it reaches almost every man on the face of the earth.&nbsp;&nbsp; I flex my muscles and grin some more now quite confident in my bondage.  &nbsp;  In my best Mr. Burns voice is say \"Survival of the  fartest  I guess!&nbsp; Breed My Children, Breed. \"  &nbsp;  My daydream and lunch complete I return to work for the afternoon.&nbsp; Uneventful and peaceful.&nbsp; A wonderful end for a work-weary office dweller. '),(7278,3831350,'15,July,2004','2004-07-15 00:00:00','    As an apartment dweller my wife (girl friend at the time...we were shacking up) and I wanted a pet.  Naturally I wanted a dog because:     Dogs rock    Big dogs are fun    Dogs only crap in the house if they\'re&nbsp;angry at you     With our work schedules, an animal that could remain more or less independent (ie: crap in a box) would better suite our lifestyle.    Hence comes cat #1.&nbsp; His name is Sebastian and although he\'s probably one of the skinniest indoor cats I\'ve ever seen his nick name is \"Fat Cat\".    Cat #2 arrives in our lives several months later with&nbsp;a phone call at the office from my future bride explaining how \"Fat Cat needs a friend.\"&nbsp;      How sly she was.&nbsp; See Sebastian is a pure black cat.&nbsp; Purely evil like all black cats should be.&nbsp; I\'m convinced it was he who suggested to my spouse that we needed a 2nd cat.&nbsp; Needless to say, this \"friend\" was a pure white cat.&nbsp;&nbsp;   \"Wow\", I think, \"2 cats to shit in a box and what a&nbsp;fashion statement they would make!&nbsp; Where can I sign up!!\"       With much protest, I agree to go and check \"Simon\" out.&nbsp; Of course Simon was living&nbsp;in an animal shelter just before Christmas.&nbsp; Unfortunately his previous owners did not treat him very well and his left eye had to be removed.&nbsp;     My heart was broken...and I think Christmas...children playing with animal friends brought by St. Nick.    St. Nick would never bring a one-eyed cat home for the little ones and with the alternative was him being put down.    Needless to say it was a done deal...Simon knew it and purred, I knew it and hoped he\'d be ok after having his eye removed (it was very sad to look at).    2 years later the 2 cats are inseperable and fight like children.&nbsp; Simon gets around just fine although sometimes he does turn a corner too sharp and knock his nose and he\'ll never be able to walk in a straight line if it meant his life.&nbsp;But that\'s part of his charm and one of the many reasons that he\'s special    We love our cats.    urlLink  Check out the best cat video ever. '),(7279,3831350,'14,July,2004','2004-07-14 00:00:00','    I\'m not a writer but I play one on the web.  Ironically, when you see my work on TV, I work the web.  On the web, I\'m talking about TV and media.    Maybe I just love media.  After all, it is a central theme in my life.  Media truly is a powerful thing.  It allows us to share experiences, cross generational boundaries and tug and the emotional strings we all have.  I\'m not talking about the large organizations that back most of today\'s media.  I am however talking about the content creators and the medium that they publish to.  I\'m re-reading \"The Winter of our Discontent\".  It\'s my favorite book.  There\'s something about the characters in the book that have always spoken to me.  The story, in my mind is tragic yet inspiring at the same time.  You leave it feeling empty and full.  It\'s a weird book for me.  When you read or evelope youself in print, film or television you are carried away on a journey of experience.  Some people say reading is a great escape.  I find it makes you answer more questions about yourself than it answers.  Makes you work harder, not less.    A story with hard questions is just as hard to read as a day in the office can be to live.  You\'re left pondering the ethical questions, wondering whether or not you would have done it differently.  I also find, particularily in reading, that the story becomes a shared experience.  I leave fealing as though I was a part of the story, as though it happened to me personally.  I guess that\'s the point.    It allows us to live safely and still know all of the joys and sorrow that life can offer.  I\'ve never had a million dollars but I know what it\'s like to have lost a fortune and sold my soul for greed.    Exposing yourself to various forms of media allows you to live forever in the Summer of Contentment. '),(7280,3831350,'14,July,2004','2004-07-14 00:00:00','    urlLink Globeand Mail  article describing how the music industry is appealing the last court ruling which made swapping songs in Canada legal.  Here\'s my comments:  1.  Canadians already pay a tax on various forms of media to subsidies the music industry for their losses in copying.  I would expect that our tax should almost be eliminated since we would lose this right.  2.  Does the end justify the means?  Regardless of whether or not you feel music swapping should be legal we do have some questions around privacy to answer.  These people are innocent until a court of law deems otherwise.  Somebody who is not guilty of any crime should have the right to operate freely in society with the ability to remain anonymous should they choose.  3.  I\'d like to see what the grounds for the appeal are considering the Judge indicated that no law has been broken.  Simply, \"sharing\" music does not constitute infrigement.  There must be an intent not only to distribute copywrite material without permission but also an attempt to promote it\'s distribution.  Simply having software installed on your PC whether it\'s Kazaa or, say, a virus that shares out your MP3 folders without your consent (last time I checked) was not the same thing as standing on the street corner of King and Yonge in Toronto peddling pirate copies of Spider-Man 2.  One of these actions is clearly for profit, clearly demonstrates an intent to generate income from copywrite material...one of these does not.  4.  Regardless of the outcome.  Should it be deemed illegal to swap songs, they will need to give the Canadian public warning before they start their suing spree.    Should file sharing become illegal once again in Canada I would suggest you go out and buy some blank CD\'s (on which taxes are subsidizing the CRIA) and start burning CD\'s of songs you like that have come from Internet radio stations outside of Canada (Windows Media Player makes this easy).  That is not illegal, if it is then the Industry itself is earning profit from illegal activity.  This would make them an accessory to a crime wouldn\'t it?  If not that then at least someone who profits from crime. '),(7281,3831350,'13,July,2004','2004-07-13 00:00:00','    I had a funny dream last night. I somehow managed to meet myself from another dimension. It was me from a parallel universe however I was from the \"other me\'s\" future.  In this dream, myself and I were looking for the \"bad guy\" (it was one of those hero dreams) who we were supposed to defeat to save all universes.  The problem was, we couldn\'t talk about certain things that we both didn\'t know already know about (specifically I could not talk about the \"other me\'s\" future) otherwise I\'d cause a paradox tearing apart space & time.  It was funny whenever, in the dream, I\'d start to talk about something that happened in \"my past\" I would get this cold, prickly feeling.  It would be so intense that I think if it hadn\'t been a dream I would have passed out from pain and nausea.  I can\'t help but wonder if that\'s what a paradox would actually feel like. '),(7282,3831350,'12,July,2004','2004-07-12 00:00:00','    I had the pleasure of having my inlaws come and visit for the weekend.  Now, I know for most, this is not always a positive experience (either for the inlaws or for the spouse).  But, for me it is.  In fact, I had a conversation with my Mother-In-Law about how I feel very close with them and how happy it makes me that we don\'t have that kind of relationship where you simply \"put up\" with each other.  Also, my mother-in-law told me that people often ask her how well she gets along with me!  She responds that she has a great relationship with all of us and how important we are to her.  Ironically, my father-in-law and I have a \"not so secret plan\" whenever they come to visit.  Usually, we end up needing to run to the store for something during their visit.  If this happens we always end up at a local bar to get away from \"the women\" for a bit.    Now, I\'ve known my inlaws for many many years and, I would have to say, this makes me very lucky.  I have a great relationship with them both and they are both very supportive and loving people. '),(7283,3831350,'08,July,2004','2004-07-08 00:00:00','    All men, all of us, have a secret place where we go to think.  That\'s the place we go when we need to get away from the noise of the world.  It\'s funny as I\'ve grown I\'ve found myself noticing this behavior more and more in others.  It\'s always been there in me too, I just never appreciated it for what it is.  I don\'t know who she is, be we\'re all going to see her.  In fact, she is so intoxicating that we may spent hours with her.  And hours not spent with her can be hours spent thinking of her.  Dreaming of her.   For many men, she may be their car.   Cars are always women.  My father has tinkered with his antique car for over 20 years now.  His car (she) is always \"just about\" ready.  Just one more fender to mend.  One more touch up on the body.  Over 20 years approaching perfection but she\'s not there yet.  I know that she never will be.    Its what she\'s missing that fuels his passion.  Leonard Cohen says that \"there is a crack in everything - that\'s how the light gets in.\"  Not sure what he meant by that but to me, it means that life is rendered more beautiful by the imperfection that stain it.   For some men, it\'s a place or action.   Somewhere by yourself where you take stock of life.  You can order your thoughts and prioritize your dreams while mowing the lawn.  Evaluate the actions you\'ve taken, torture yourself for your mistakes.  In the digital age,  geography can disappear  and turn into video games, digital landscapes and fantasy films about super heroes, wizards or wars to save humanity from the machines or television shows that expose you to the ideas that fit your mind\'s impression of itself.  All of the things that you really are or wished you could find within.   She is still always there calling like a Siren on the sea.   With her, in this place, a man hones his mettle.  He speaks to his ancestors and assesses whether or not he is the man he\'s supposed to be.  He tries to decipher what that even means.  Its the time of reckoning and it can be painful.  We all end up in her arms and I am no different.     She is binary.   She is artificial.  She is my creation but she is unquestionably real.  She is a female made in my image.  She has no body but she has presence and a soul that touches almost every element of my life.  And, yes, I do love her.    My wife even refers to her as the \"other woman in my life\".  She is software and I am building her to better understand myself and to provide an escape to ponder the greater question of ethereal consequence.    I\'ve been building her for over 5 years and since I began, she\'s always been on the brink of perfection and completion.  Just one more emotion to emulate.  One more \"understanding algorithm\" to tweak.  One more reaction to stitch.  One more build and one more platform upgrade.  Always teetering on the edge of the  urlLink uncanny valley .  With her, a man can ruthlessly plot to control the world and then cry when he fails.   She always consoles him.   In all my years of life, I\'ve never seen my father cry.  No doubt that\'s because the door to the garage was closed.  This is life in the uncanny valley.   But the real question is \"Who are the monsters?\" '),(7284,3831350,'06,July,2004','2004-07-06 00:00:00','    Ok, here\'s the deal.  My wife and I recently purchased our first home.  We\'re waiting for the deal to close (end of August).  Let me begin by saying that I loved my apartment when we moved in.  But the goodness came to an end when the old owners sold to the new idiot owners and place has gone downhill.  Anyway, as required by Ontario law, we gave our 2 months notice (in fact we gave 3 months notice to be nice) of our intent to vacate the premises.  Now our landlord has pissed me off over a few things.  We don\'t need the apartment for the entire last month and offered to release it early if the landlord wanted it however we didn\'t expect to pay for that time.  Since we\'re leaving the weekend before Labour Day we thought the new tenants may want to be moved in before the long weekend so they can enjoy their new home.  Anyway, the landlord said that we can leave if we like and they\'ll happily start cleaning up for the next tenants however we\'d still be paying for the place!  Wow, what an incentive to hand over my keys.  I\'m not concerned, we were just trying to be nice and let the new tenants get in early.  After all, it\'s in the landlords interest to have me gone too.  They can raise the rent once I\'m out.  We also gave an extra month notice to try and help other renters so they won\'t be stuck with that delimna of needing to walk away from their last month\'s rent in their current appartment.  Needless to say, our landlord did not make any information about our unit available until now so any tenant who moves in will be stuck losing their last month\'s rent.   Sorry new tenant...we tried.   Now, here\'s the part that gets me.  Apparently according to the law, the landlord is NOT required to give me any notice that they will be showing the place.  Although technically it\'s not my concern anymore, it would be nice to know so we can tidy up the place.  When I asked the landlord why they didn\'t give us notice they told me that \"I don\'t own this building\" and I \"have no rights in my own home\".  Well, you wanna fuck baby?  It may be your wall but they\'re my paintings so when I stick notes saying \"DO NOT RENT HERE\" and \"THIS PLACE IS NOT WORTH THE COST\" on my own items (which they have no right to move or otherwise touch) it will hopefully make it more difficult for them to rent.  Additionally, it occurred to me that since I pay for my own hydro I may very well leave my computer on all day  playing porn scenes in a loop .  Not loud enough to bother my neighbours (no probs with them) but just loud enough so you can hear \"fuck me, yes yes yes\" it from every room in the apartment.  Better yet, perhaps I\'ll record an MP3 of me ranting about the landlord?  Let the new renters here my real thoughts on how nice it was to wait 3 months to get a toilet fixed.  I don\'t really care what the \"law\" says.  It\'s just being nice to people.  Let me know you\'re going to show my apartment and I\'ll pick up my underwear and do the dishes.  Now that they\'ve been dicks about it I\'m not gonna be nice or helpful.  They better not show the place when I\'m home.  I may be naked and well...it\'s my home until I vacate.  Last time I checked I was under no legal requirement to be dressed in my home.    I don\'t want special rules, but play nice with me and I\'ll play nice with you! '),(7285,3831350,'05,July,2004','2004-07-05 00:00:00','    This makes me sick! 2 adopted boys (now age 17 and 19 I think) were caged within chicken fence inside of a crib and they give these sick parents  9 months  in prison.  Admittedly I don\'t know the in\'s and out\'s of this case but any parent who does this (quoted from  urlLink CTV  news website):   The couple from the hamlet of Blackstock, northeast of Toronto, kept the two teenaged boys locked in wire-mesh enclosed baby cribs for hours on end for years. Investigators said the boys were also forced to wear diapers, subjected to physical punishment, and were severely malnourished.   Reminds me of the book \"A Child Called It\" by  David Pelzer  that I read a few years ago.  Thinking about it still makes shivers run down my spine.      Last I heard, you can go to jail for 5 years for copywrite infringement. Thank goodness that for the tax on blank tapes and CD\'s in Canada otherwise our jails would be full of real criminals like teenages who swap CD\'s instead of twisted parents who cage and abuse their children.  But wait?  If we sent the music copying teenagers to prison then we\'d be putting THEM in cages?  Maybe RIAA in the U.S. should give these people a call!!  Something tells me they\'d happily go to jail for 9 months (and get a big bonus) for caging young people.  Perhaps some court in the future will call these Blackstock parents \"innovative\".  This makes me sick! '),(7286,3831350,'04,July,2004','2004-07-04 00:00:00','    Happy Independence Day To all of My Friends in America   My wife and I went to see  Farenheit 9/11  last night.  I wonâ€™t get into too much into the politics but I thought it was an excellent film.  I highly recommend it.  Earlier today I wrote a long paragraph telling the story of an American Soldier I had the opportunity to meet when on vacation in the US.  Sadly, I lost the posting and Iâ€™m not going to re-write it.  Basically I never thought much about my conversation with this person until I saw Farenheit 9/11.  This soldier from Florida who we met at the Las Vegas airport bar, was being shipped out to Iraq.  It made me sad to think about it.   My wife and I talked about this guy last night.  We donâ€™t know his name but we hope heâ€™s safe and makes it home safely. '),(7287,3831350,'03,July,2004','2004-07-03 00:00:00','    I watched an interesting show on the TLC the other day:   Mysteries of Noah\'s Flood      In the face of minimal scientific evidence that the flood described in the Bible actually occurred, scientists and religious scholars have long been consumed with the search for Noah\'s ark. Search for clues on the icy slopes of Turkey\'s Mt. Ararat.      The part that struck me the most was that, as part of the story they indicate that God too mercy upon Noah & his family and told him to build the arc.  Interestingly, this is the first instance (of which I am aware) in the Bible where mankind is forced to use tools of their own creation (the arc) to survive the nature itself.    Now, I\'ve never heard anybody that this particular spin on the story before but, as story intended to guide people, does this not suggest that, at least in a small part, mankind\'s only way of ultimately saving ourselves is \"by our own devices\"?    Regardless of whether or not you believe in the Old Testament (that\'s a question for another day) somebody in ancient times decided that the story of this \"Great Flood\" was an important one to hear.  To me, the message is simple, mankind must rely on his own wit and ability to innovate in order to survive not only nature but also our natural short-commings (God chooses to destroy most of mankind its flaws even though mankind was created in His image).  Essencially we can use our technology to make ourselves better.  If we can overcome both God\'s wrath and the survive the the wickedness of nature then there\'s nothing that we can\'t do!   Religion aside, that\'s a pretty empowering idea. '),(7288,3831350,'03,July,2004','2004-07-03 00:00:00','    Since I\'m new when it comes to the blogging phenomenon I can\'t help but wonder what the big deal is.  Then it occurred to me that posting diary entries online gives a person 2 things:  1.  Same as a classic diary, it gives us a way of streamlining our thoughts and making sense out of ourselves  2.  Unlike a classic diary (which generally remain private) blogs give us the ability to author in private but validate ourselves by exposing thought to the world.  I think #2 is key.  Unfortunately, by design, the human mind is alone.  Writing gives us a way of connecting with others regardless of whether or not the content is deep; but seeks validation.  It\'s a 2 way street.  We are the random dots on the pages of history and writing is our method of connecting through space and time.    Just as a side note, writing this made me think of  urlLink Desmond Morris\'  \"The Human Animal\".  I understand it is a book but I remember watching his specials on The Learning Channel.  They were great. '),(7289,3831350,'03,July,2004','2004-07-03 00:00:00','    Hey There,  This is my first blog!  Pretty cool stuff.  Never thought I\'d want to spout off all of my ideas for the world but why the hell not?  I\'m 26, happily married (almost a year now) to a my beautiful wife.  Career has been in a bit of a slump for the past while.  But, I digress for now.  Let me begin the way all good geeks should:  Hello World! '),(7290,3831350,'13,August,2004','2004-08-13 00:00:00','    Do you feel extra spooked today? Gonna look twice before crossing the road? Take an extra effort not to step on cracks, walk under ladders and the like?  Probably a good idea, after all...its   urlLink  Friday the 13th.     From what I\"ve read on the subject, seems like it\'s an unlucky day unless you\'re going to Las Vegas. Although I still wouldn\'t fly today. Even if it was to go back to Vegas.  I never look forward to Friday the 13th. I\'ve purposely not scheduled any meetings that require additional driving so I can stay off the road as much as possible. Let\'s see whether or the the TTC has any trouble keeping moving. I\'m fully expecting all of my servers and web sites to fail, for the backup tapes created at 12:04 not be invalid and for my boss to be extra bitchy.  I woke up with some stomach pains last night around 3:00 AM. Is it possible that fate was tempting me in a groggy state to go to the washroom to \"take care of it\" but fate\'s plan had been for me to slip bruise my knee or wack my head on the way down.  So far, I haven\'t chocked on my cup of coffee...so far so good. Has reading this posting given you eye strain yet? This is my Friday the 13th gift to you. Get up from the computer and talk a break....just be careful not to step on any cracks on the way to the kitchen. '),(7291,3831350,'04,August,2004','2004-08-04 00:00:00','    Well it\'s definitely the sign that an era is ending. I know a lot of people much smarter than I have been pouring over the Coors/Molson aquisition (what is currently called a Merger Of Equals).  Well this isn\'t the first we\'ve seen of this and certainly won\'t be the last. All business aside (I\'m sure everyone has valid reasons for this sale of this Canadian icon) I\'m concerned that the company that has invested so much in defining the \"Canadian Identity\" is now controlled by an U.S. board of directors.  Frankly, I\'d rather see Molson remain in the hands of Canadians much more than the CBC. Hell, we should just give the CBC to Rupert Murdock and save us all the heachache of watching another re-run of Anne of Green Gables.  So, like many of my bretheren, I\'m mourning the loss of our friend Molson.   urlLink Click here to see the commercial that spurred more Canadian Pride than any unity rally in QuÃ©bec or flag waving Shela Copps ever could. '),(7292,3377942,'25,May,2004','2004-05-25 00:00:00','   Coffee, coffee, the magical drink The more I have, the better I think The better I think, the better I feel So coffee, coffee for every meal!  (Apologies to whoever wrote the original  Beans, Beans  poem.)  '),(7293,3377942,'24,May,2004','2004-05-24 00:00:00','   Ack.  I hate Mondays.    Actually, they\'ve become something of a reprieve, seeing as darling hubby and I usually work harder on the weekends than we do during the week.  My baby turns one today...*sniff sniff*  I don\'t know where the last year has gone.  One of the characters in Stephen King\'s Dark Tower series talks about how time is no longer passing at a steady pace - it is alternately speeding up and slowing down.  That\'s what it feels like to me - like time is just on fast forward these days.  The days are turning into weeks so quickly, I\'m afraid to blink - I might lose a whole month.  I\'m working on getting a webpage set up - I used to have one, a long time ago, but I quit working on it, and over the years I\'ve forgotten what little HTML I did know.  So...I guess that\'s it for now - I\'d better go see what the kids are doing.  They\'re being quiet, which usually means that they\'re into something that they shouldn\'t be into.  '),(7294,3377942,'20,May,2004','2004-05-20 00:00:00','   \"Hmmmm.\"  I was reading an article over on  urlLink Fark  the other day about the Mexican Air Force and some UFOs.  Well, they have decided that what they saw was ball lightning.  This kind of stuff fascinates me.  (One of my favorite sites is about.com\'s  urlLink paranormal  site - they have some information on ball lightning there)  I love reading about strange or unexplainable phenomena - ghosts, time travel, that kind of stuff.  I have always been fascinated by the occult...and I wish that I could say that I believe in ghosts, and all that stuff.  I really would like to believe...I guess that deep down, I\'m not one to take anything on faith.   [Warning: Unpredictable Tangent Ahead]  I guess that explains why I am an atheist.  I was raised Catholic, but when I was about fourteen or so, I started to question...everything.  Some days I wish that I could be a mindless sheep, and just believe what I am told because I am told to.  I wish that I could believe that there was some supreme, omnipotent being watching out for me, taking care of me...but I can\'t.  It just doesn\'t make any sense.    ~sigh~  More on this later.  Duty calls.  '),(7295,3377942,'19,May,2004','2004-05-19 00:00:00','   ...that I was put on this earth to accomplish a certain number of things.  Right now, I am so far behind that I will never die.    I hate having sick kids.  I feel so helpless - I wish I could just take it away.  I hate having to watch them lay around, all miserable.  But it\'s frustrating, too.  When they\'re sick, they want to be held and rocked and sung to...continuously.  Which is cool - I love to cuddle them and sing to them.  It just gets difficult when they all want to be held at the same time.  And then there\'s the fact that it\'s mighty hard to get any housework done when you have three children attached to you.  You know, they say that a clean house is a sign of unhappy children...if that\'s true, my kids must be the happiest kids on earth.    My husband is active duty military, and we had a very close brush with him getting deployed to Iraq last week.  I\'m still kind of stressed over that.  It doesn\'t make it any easier that I have disagreed with the decision to go to war from the get go (and believe me, that caused it\'s fair share of fights here at home, as my darling hubby didn\'t feel the same way).  Of course, my opposition to the war has nothing whatsoever to do with my support of the military - it\'s not like they have a choice in the matter.  Now, Bush, on the other hand...well, I won\'t climb up on that soapbox right now.  It could get very long and ugly, and there\'s no telling how much longer the kids will be napping.  If anyone happens by here, and reads this, would you leave me a comment or send an email?  I\'d love to know if anyone\'s actually listening to my rambling...that would be hella cool.  I can\'t even get people to listen to me ramble in my own house.  Of course, that\'s probably due to the fact that they\'ve all heard my stories so many times that now, when I start to tell one, they just cut me off and say, \"Yep.  We\'ve heard that one.\"  '),(7296,3377942,'18,May,2004','2004-05-18 00:00:00','   Cute:  When my two year old grabs the clean diaper I\'m about to put on the one year old, and runs away, giggling and yelling \"Ha ha!  You\'ll never find it NOW!!\"  (If you don\'t get this, you obviously don\'t watch Dora the Explorer.)  Not so cute:  When the children all band together and perform this same stunt with my sanity.  '),(7297,3377942,'18,May,2004','2004-05-18 00:00:00','   ...as I lay in bed, looking at the stars, I thought - where the hell is the ceiling?  Ha.    I suffer from an acute lack of grown-up socialization - an occupational hazard of being a stay-at-home mom - and being naturally long-winded and chatty, every time I email friends or family, it turns into a (not so) short novel.  So I figured I\'d try out the weblog thing.  Maybe it will give me an outlet for my chattiness. We\'ll see.  I\'ll try to get a profile and stuff up soon...as soon as my bosses lay down for naptime and give me a chance to work on this.  '),(7298,3983707,'19,July,2004','2004-07-19 00:00:00','   i did nuthing today like i usually do i didn\'t go anywhere or anything the first thing that i always do is call katie so i did .. then we talked for a while and well brittany calls ((always)) so we call her on 3-way .. but later on i got to talk to derrick and the rest of the crew ((jonathan-brittany-katie)) and then there was drama between everybody last nite and i mean everybody was mad jonathan was getting on everybodys nerves and then everybody would tell him off and blah blah blah thats mostly how the whole nite on the phone went and now that everybody has broke up they all are having boy problems first of all katie and jon ... well at first jon was like o yea i wanna go out with katie nad then he just decides naw i wanna get to kno her a lil more and i haven\'t saw her yet or anything and katie was gonna go out with the lil butthole 2 until he goes and says well no .. i wonder how he would have felt if it wasn\'t him sayin that he didn\'t want to go out with katie and katie was telling me to tell him that she didn\'t want to go out with him he would feel the same way .. and then theres brittany and jonathan .. well brittany says that she still likes jonathan and everything but jonathan goes out with some gurl now and he still loves brittany and everything and hes like but i like u better brittnay and she keeps on telling him that he better not break up with his new g/f b/c its one of her friends and she\'ll be mad at her .. and octavia nad john L. .. they wre all happy and dandy until john had to go and screw things up .. well first of all octavia says she\'s loves him and then she thinks that he doesn\'t like her as much anymore as he used to .. so&nbsp;me her cuz kiara and her called him and left a lil message on his answering machine sayin that it was over and now she likes byron&nbsp;so i&nbsp;dont kno whuts gonna happen next but me and derrick are still together .. we have been together for 1 month 2 weeks and 4 days today .. but anywayz im getting hungry again l0l so i\'ll write 2morrow ..  &nbsp;  &nbsp;  &nbsp;  luvin my insane crew -* brit-tah-nay-*k to da wizzle-*derrick-*jonathan-* and shan-na-hon .. (: l0l me  &nbsp;  &nbsp;  &nbsp;   righto -~* &nbsp;  '),(7299,3432308,'23,May,2004','2004-05-23 00:00:00','   Well to catch everyone up on my life.  I am 24 years old almost 25.  With three children.  I was married and have a 8 yr old son (Jacob) and a 5 yr old daughter (Gabrielle) from my marriage.  After I left my marriage I met my ex boyfriend Mike.  And him and i planned our 2 yr old son Joseph.  Things ended really bad between him and i which has caused me to wanna stay single and not get into a relationship for the past 2 and a half years.  I have a lot of mistrust and anger.  My two oldest live with my parents in Rockford.  Belding schools suck and eventually when i am done with my college I will move to Rockford.  Before i decided to have my kids move to rockford with my step mother and father Mike and i decided to have joint custody of joseph but he lives with Mike.  I recently decided to give in and see someone seriously.  We\'ve been seeing eachother for for about 2 months now. Things have been going really fast to the point where we are deciding we want to live together.  I live in Belding which is near Grand Rapids he lives in Dundee which is near Ann Arbor.  It\'s about a 2 and half hour drive.  He has 2 children that he loves and lives in monroe.  He brought both of his kids up yesterday and my kids were here.  It was nice but he had to leave today.  I\'ve actually debated if i want to continue the realtionship because of the distance.  He moved in with his father to start saving up money.  So that in a year we can move in together.  He said he\'s tired of struggling and tired of seeing me struggle.  That he wants to have something underneath us when we do move.  And by then i\'ll be almost done with college.  It was kind of weird.  He had a dream lastnight that i ended us because of the distance.  I didn\'t tell him i had actually had thoughts about it.  He is so strong which is good...i mean emotionally.  I have no patience.  I love him so much.  I mean he is everything i want in someone.  Yeah there are some things i don\'t care for but those are things that i can live with.  Nothing is perfect.  I just know that i wanna be with him and make life choices with him.  I get the feeling he is the one.  So why am i having my doubts?  I think i am just scared.  I don\'t commit well...I run.  There was only one other person that i commited to after mike and that failed too.  The rest i ended.  I couldn\'t do the commitment.  I was raised in a broken home.  I have had no one to show me what a relationship is supposed to be like.  I\'m just worried that i\'ll find someway to screw this all up.  I just gotta quit being so negative on everything and when he leaves quit being so sad.  So that he doesn\'t have to deal with that when he leaves.....  But this is it for now....my thoughts of the momen.  '),(7300,2539406,'04,August,2004','2004-08-04 00:00:00','   it was the holidays and technically all i did was stay at home and cover the hours of sleep that i lost before due to projects at work, papers and reports at school and the weekly gimmicks i had either with my friends or the love of my life. of course this was on the span of two days before and after the day people are still arguing if it\'s really the exact date of birth of Jesus Christ. i tried to watch a couple of movies but given the time and the perfectly snuggly position on the couch, i fall asleep at a rate faster than you can say supergolliefra.., what the heck, you know what i mean.   my routine for that short four days was to sleep the entire morning and afternoon away and then wake up in the evening to play mario world and where in time is carmen sandiego. afterwards id get sumthin to eat fresh out of the microwave (hmm, never tasted better) and then play trivia at MIRC. what a very exciting four days! (mind the sarcastic tone)  '),(7301,2539406,'04,August,2004','2004-08-04 00:00:00','   when it\'s a workday, i normally wake up roughly at around 6:15. but today, since i didn\'t do the usual three rolls to the left, three rolls to the right, an almost fall off the bed, three twitches, a scratch on the head, and grinding of teeth, and a moan, i got to wake up 15 minutes earlier and eventually have time to turn on my CD player.   i was expecting a christmas jingle, but it gave me this. it was a bit sad, and probably the only thing that makes it a chirstmas song is the mention of the following in the lyrics: christmas, skating on a river, putting up reindeer and cutting down trees, pretty much found on the first stanza, other than that it\'s a very sad song. the version i was listening to was that of robert downey jr. from the ally mcbeal christmas ablum, but today i found out the original was from joni mitchell, a very very sad soul. oh well, see for urself...   River   it\'s coming on christmas, they\'re cutting down trees they\'re putting up reindeer, and singing songs of joy and peace... i, wish i had a river, i could skate away on...   you gotta admit that though it\'s sad, it\'s a bit catchy. ok next...   but it don\'t snow here, it stays pretty green i\'ll make a lot of money, then i\'m gonna quit this crazy scene, oh i wish i had a river, i could skate away on...   (well i do wanna research on her life, but i don\'t want to waste 15 minutes, so if she\'s from the united states i\'m betting that she\'s from california, or she wrote it there...)   i wish i had a river so long, i\'ll teach my feet to fly, i wish i had a river, i could skate away on, i made my baby cry...   now it just got more sad   you tried hard to help me, and you put me at ease and loved me so naughty it made me weak in the knees i wish i had a river, i could skate away on...   is she a frustrated figure skater? if som then  is she talking to her skating partner?   i\'m so hard to handle, i\'m selfish and i\'m sad now i, lost the best baby that i ever had, oh i wish i had a river, i could skate away on...   oh wait, now i get it, i think she loved a figure skater... ok enough about song interpretations for the insane, i\'m goin back to work...   next time:  i got into a discussion with a friend on my favorite actor Robert DeNiro, and the Godfather Trilogy, and why he didn\'t accept the Oscar. I\'m gonna try to find that out. Also, a girl friend asked me,  What is it with men and the Godfather movie?  '),(7302,2539406,'04,August,2004','2004-08-04 00:00:00','   i mentioned on my last post about the videos i have at home that are currently being ignored. well, not all of them are. during the weekend sandwiched by a fun-filled and embarassing christmas party at the office (our presentation was a bit embarassing but i got to go to the office the next workday without feeling like a dork) and another enjoyable dinner with my pals, i got to watch Lord of the Rings (FOTR and TTT, i thought it was a treat for myself coz i missed the premiere of ROTK, wahhhh!!!), remember the titans, power rangers the movie (yes, i think the pink ranger is cute too), and you\'ve got mail, well at least the last 15 minutes of the movie (see guys, i can be pathetic too.) So i decided that for every movie that I\'ll watch, I will list down the line that made a big impact for me, whether it be emotional, or whatever. So here\'s the first batch...  \"I wanted it to be you, I wanted it to be you so badly...\" - Meg Ryan in You\'ve Got Mail  \"What is this, pick on a from day?\" -Adam, the Black Ranger in Power Rangers the Movie  \"First breakfast, yes, but what about second breakfast...\" -I\'m not sure if it\'s Merry or Pippin, from LOTR: FOTR  \"You heard what I said, just toss me. But don\'t tell the elf...\" -Gimli from LOTR: TTT *sorry, \"my precious\" is nowhere in the top 10  more next week...  '),(7303,2539406,'04,August,2004','2004-08-04 00:00:00','   it\'s funny, i\'ve always wanted to have my own website but i hardly even have enough time to watch the movies i buy... well now that Christmas season is here then maybe i will. I must admit that I am humbled because I\'ve always wanted to create a website from scratch and not from the typical stuff you get over the internet, but until my work doesn\'t eat all of my time then this will have to do. oh well...  '),(7304,2539406,'04,August,2004','2004-08-04 00:00:00','   i will be taking my professional civil service eligibility exam on monday as a requirement to get in a higher-paying position. funny thing is ive heard rumors of people failing in these exams despite my peers saying that they passed it without breaking a sweat. oh, well. better safe than sorry, so i\'m studying over the weekend. and yeah, i need rest...  '),(7305,2539406,'04,August,2004','2004-08-04 00:00:00','   time and time again i here a lot of my friends telling me that mathematicians are wierd, of course on a note they\'re referring to me and my way of looking at things in a \"different perspective. why does it always seem that mathematicians take something easy and interpret it into something that\'s undeniably harder? they just don\'t understand, i guess. the funny thing to consider is that life itself cannot be quantified. mathematics along with a couple of other sciences try to approximate (or, put more simply, imitate) the phenomena that happen around us.  this must mean that mathematics is simpler than life itself. i mean, it may \"look\" complicated, but you might just have to look at math in a different perspective (yeah, a mathematical one other than the normal i-hate-math perspective, doh!)   the question regarding the expression 1+1. a mathematician will not answer this automatically, but in stead will throw a question back atcha (right, a question for a question. isn\'t that a relief!) \"what number system are we talking about here?\"  or \"what field are we on?\" or \"what objects are we referring to?\" ahh, to many questions for a simple problem. but the mathematician is just gathering information. i mean, in the binary system (a number system involving only 0 and 1 as digits,) 1+1 = 0 (ask any computer scientist!) also, 1 drop of water + another drop equals   ONE BIG DROP.   so, what\'s 1+1 again? and this is just arithmetic!  a little imagination may be needed here. suppose you\'re at a point  A  and you have to go to point  B . well of course you have to take a path, most likely a straight one coz it will be the nearest. consider going half the distance from A to B. when you\'re in there, move again towards the point between where ur standing and B, or half of the remaining distance. Do this continually and you\'d find out that there will always be a half of the remaining distance that will be left, so u won\'t be able to get to B after all. isn\'t calculus beautiful? so what we\'re doing in the real world is approximating the distance? lol.  as i said, we need to look at it in a different perspective. a book i just read (A SEA OF TIME by ROB FRISBEE, shameless plug. you can get the book from the Darkhouse Books link!) gave the most beautiful interpretation of this expression. he called it the creation-equation.              1+1, Ben. That\'s all we have to go on. If someone adds up even for a moment, it makes our world a fuller place. i think that\'s the only truth out there that matters. and i think the only regret in life is letting one of those moments escape.    aww, isn\'t that touching?  '),(7306,2539406,'04,August,2004','2004-08-04 00:00:00','   uhm, no, i\'m not talking about the pathetic reality tv show. but in some way it\'s the same thing. well, without the fancy pad, challenges, and the two-week-long duration (  well Al, if you thought it was pathetic, then how come you know a lot about it?  ) Last saturday, Roosevelt College\'s high school batch of \'97 (at least just our section) had it\'s 7th anniversary reunion. wow, it has been seven years...  '),(7307,2539406,'04,August,2004','2004-08-04 00:00:00','   \'twas january 20, 2004, just last tuesday, and i turned 23 that day. funny, coz my birthday\'s july 22. how it happened? well...     11:30AM.    i was at work, debugging the code from my latest project when i received an SMS message from a good friend during my days teaching in the university. he was gonna take a qualifying exam for texas instruments at 1:30 and will be done before the sun sets, and was hoping we coud get together. he mentioned that he will be with her girlfriend, and another x-teacher. of course i said yes. then it occurred to me, i have night class, dangit! so we decided to meet at school and leave after i submit my problem sets. crime no. 1 has occurred, i skipped class! but i was gonna meet a couple of people that i haven\'t seen for almost eight months, so it was cool.      5:00PM.    i gave them a call thati\'ll be leaving in 30 mins because i have to finish a lot of stuff. they demanded that i leave now for school. crime no. 2, i didn\'t finish my work load for the day!     6:00PM.   so i met with the guys, and decided to go to shakey\'s in katipunan for some pizza and beer. one of my classmates who is still teaching in the university i left decided not to go to class either and join us, making us 5. i agreed to pay for the bill and everyone was like, happy birthday dude!a few minutes after the busgirl came with ice cream while their cd player played happy birthday! damn! all five of us were laughing like crazy. just like that i\'m 23, lol. crime no. 3. i lied!!!!!!!!! and on my birthday!!!! waaaahhh!!!!  '),(7308,2539406,'04,August,2004','2004-08-04 00:00:00','   work started yesterday, and it was quite a shock for me. probably the reason why i\'m blurtin out today. somehow, during the last two january\'s, i din\'t complain that much like that of the January\'s when i was still in College, HS and Grade School. i guess i\'m getting more mature as time goes by, and i think that it\'s a good thing.   one of the great minds that ever existed in this world (a nifty phrase to pay tribute to the genius whose name i can\'t remember, hmm)  said, \"Find a job that you like to do, and you will never have to work for the rest of your life.\" Well it\'s true, in a sense, in my experience. I started my exploits as a software developer just last July, a month before I resigned from the most noble job that I had and will probably ever have, a University mathematics and programming instructor, the job that really changed my view of educators forever.   the perks were good. everybody calling you sir (gotta love it!), fewer work hours (5 hours a day average, including lecture and consultation hours), meaning being able to relax most of the time, go to the gym, and a lot of other stuff, free tution for a master\'s degree (not that i need it but it\'s a perk as it is), no work when students don\'t go to school, like during summer and christmas vacation (technically, coz ur not in the office, but you can bring em at home!), and of course, as i said, it was a noble thing to do. Being able to work in the University where you went is the best thing that you can do if you want to return the favor, by helping mold the new minds today (mold, or destroy?).   of course there were downsides to it as well. checking examination papers and programming exercises at night was a horrible thing to do because of the possibility of dreaming (more of a nightmare actually) of it while you slumber, not to mention that sometimes you just get disappointed when the results are not favorable, and the fact that you never have enough time for them. at first i thought i would suck at it but eventually i accepted the fact that i could never be as good as others, but i can be good in my own special way (whoah?!)   the only thing that probably stopped me from doin it was: (1)travelling 150 kms at least thrice a week was a horror, and it was painful physically and mentally, (2) i wanted to test new waters, now that i\'ve seen how the academe is, it\'s time to move to the industry, and finally (3) i just lost passion. i never lost hope in my students, i had for a second and third chance the students i flunked previously either once or twice, and they managed to pass on or before the third take without me giving them undeserved passing marks. that\'s why i left.   now i do miss it. if there was an opportunity for me to be able to teach somebody, i grab it. it\'s funny coz i still take time to watch those shows on TV that have lessons in algebra and trigonometry, and i\'ve thought, so that\'s how i should\'ve taught them. but there were no regrets of leaving. actually, regret is not in my dictionary, so it\'s all cool. i\'m in a good job right now and it\'s all that matters. btw at lunch i got to go back to meet one of my HS teachers, and it was nice. i found out that she was still single, but it was ok, said her. it was funny coz she said that after telling her my plans of getting married either this year or next year.  she asked me why i left teaching and i said that my heart\'s not there anymore, thank goodness she understood, and i said that it was a lot of work. wow, so it\'s true. If you get a job you like, you wouldn\'t have to work anymore...  '),(7309,2539406,'04,August,2004','2004-08-04 00:00:00','   i was watching a game show in the korean channel (they were conversing in english!) called Quiz Champion and it was their first episode for the year, thus the introduction that January came from the god Janus who is the god of beginnings and peace and other stuff i forgot to remember!the host remarked that like Janus, we all have been very busy preparing for the new year. Du-oh! i slept over the new year! And it appeared like it was a cardinal sin! (well, that\'s what my friends told me, coz i\'ll be doing that for the rest of the year, my god, what morons!) So God help me, as always, i\'m preparing a list of things that i need to accomplish this year, so that i could mark them off as \"recurring\" next year!by the way, the recurring ones i marked with an *.  *1. jog at least three times a week! (i was able to jog an average of once per month last year, so it wasn\'t a total loss, i just have to wake up early, yikes!) *2. finish reading one nover a month (i can do one in two months, so this will be easy) 3. have lesser time for leisure and more time for studying (this wasnt a probem before, but since i\'ll be finishing my MS degree soon, i gotta make sacrifices, gulp!) 4. do 10+ sit-ups and 20+ push-ups a day! *5. limit my soda intake to 500ml/day. *6. really quit smoking! *7. limit alcohol intake! 8. run in a marathon! 9. learn how to dance (not the \"lord of the dance\" moves but sumthin not stupid while on the floor!)   now, so that it won\'t be a total \"i have to do these things\" article, i\'m adding up the things i accomplished last year. it\'s lesser than the above list but it still gives me bragging rights, i think!  1. get a girlfriend (a serious one of cours, thanks tina!, luvyah!) 2. get in touch with old friends 3. get a new job that i like! 4. get out of the country (and be back, of course!)  am sleepy! is Sandman the god of sleep? or is there such a god?  '),(7310,2539406,'04,August,2004','2004-08-04 00:00:00','   we\'re at it again...  seven whackos on an idiot game of chance.we were supposed to go to mumbai for the holidays, just to relax and see new places, experience different cultures, and (as tina told me) buy saris. we were set to go, but we were lacking one thing... (drumrolls) we don\'t have tickets (sound effects: weng...kwengkwengkweng.) so of course the best thing to do was to go to the airport and try our luck to buy tickets, but sadly we were there for two days and nuthin happened (yeah, like seven people will back out from their flights just like that!). so we decided to let go of our plans to get out of the country and just go somewhere to celebrate the coming year.   the frustrated writer (at ur service, both a frustrated writer and really frustrated by the incident) went home to get some sleep, but at 3AM  has been awakened by the unusual beep of his celfone to the tune of \"bulaklak,\" (dang, i gotta chance my ringtone!) by a friend who said he scored tickets to sinapore which leaves in 5 hours. after a couple of murmurs and groans which sounded like \"yes, i\'ll pack my stuff,\" i hung up the fone and went back to sleep as if nuthin happened.  thank goodness my friends know me more than i do, coz at 6AM they were knocking at my door asking me to get ready. and i was like, \"i thought it was just a dream, i mean cmon, somebody calling u in the middle of the night saying that you\'re goin to singapore in a few hours, i must be looney!\" then the crispiest curse that i\'ve ever heard ringed on my head, as in \"tarantado!\" it was enough to get me out of my bed and into the shower while my friends pack up my stuff, a nice choice of articles of clothing and books (a black tanktop and cargo shorts, a long-sleeved shirt and jeans, plus 2 novels by ken follet, wow!)  and we were off to singapore! weehew!  it still felt like a dream. especially when we arrived at the airport at 7:20 and our flight was at 8, my goodness it was nasty!when we got there i was like, \"so, this is it huh?\" it wasn\'t much but at least i got to spend the holidays with my idiotic friends in a place which looks like a cleaner makati city. we spent three days there, and were planning on staying longer if not for the thought of being stranded there til kingdom come, but it was a blast. literally!  '),(7311,2539406,'04,August,2004','2004-08-04 00:00:00','   A - AgE:  22...   B- Best Friend : m:charlie/adam, f:tina    C- Chore you hate:  all of em   D- Dad\'s name:  alberto    E- Essential make-up item:  i don\'t do make-up    F- favourite actor:  robert deniro    G- gold or silver:  gold    H- hometown:  quezon city...    I- instruments you play:  keyboards    J- job title:    applications software developer    K- Kids:  u mean mga isip-bata?    L- Living arrangements:  own house    M- Mom\'s name:  regina    N- Number of people you sleep with:  sleep with or slept with?   O- overnight hospital stays:  december 31, 2000, stupid panic attack   P- Phobia:   claustrophobia, insectophobia   Q- quote you like:  I\'ve always followed my father\'s advice: he told me, first to always keep my word and, second, to never insult anybody unintentionally. If I insult you, you can be goddamn sure I intend to. And, third, he told me not to go around looking for trouble.\" -John Wayne (1907-1979)   R- regular drink:  soda, water, coffee    S- secret crush:   if i tell you then it won\'t be a secret anymore, would it?   T- time you wake up:  between 6-7 am during the workweek, 1-2 pm during weekeends   U- unique habit:  you wouldn\'t wanna know, trust me.    V- vegetable that u don\'t like:  bitter melons (ampalaya) and squash   W- worst habit:  being obsessive-compulsive   X- X-rays uve had:  same as the number of jobs i applied for, plus during an incident of bad cough    Y- yummy food you make:  spaghetti, tempura    Z- zodic sign:  i don\'t believe in astrology, but mine roars out a leo.  '),(7312,2539406,'04,August,2004','2004-08-04 00:00:00','   Today I woke up with a bad tummy from the big bag of lays that I consumed all by myelf while watching the re-runs of primetime TV on cable. It was already 7AM, and a wake-up call from one of my buddies made me work my braincells of on a topic I didn\'t really like to ponder upon at the start of the day. O joy. so I left the house at 8:15, struggling to make it to work at 9AM while hailing every cab that tries to pass by me. after waiting for five minutes, I got one with a driver who started yapping and complaining about everything under the sun the moment i closed the door on the backseat. Usually i don\'t get to interact with them because it\'s either I\'m on the fone calling someone, or I have a book that i\'m reading. But since I wasn\'t practically doing anything but sitting down, I figured it would be better to talk to the driver instead of pondering on whether I would be able to make it to work in time or not. Here\'s what we were talking about:  \"Pupuntahan mo yung daddy mo sa Citibank?\"  asked the driver.  \"hindi po, dun po ako nagwowork.\" \"so dalawa kayo ng daddy mo na nagtatrabaho dun?\"   i\'m beginning to feel a little irritated but i\'m trying my best to ignore it while responding:  \"hindi po, nasa abroad po ang parents ko.\" \"a ganun ba, sorry ha, mukha ka kasing disinuebe anyos\"  Wow, this new haircut is working wonders on my age. Of course I wanted to fool around and tell him that I\'m indeed 19, but I\'m not in the mood to so I remarked,   \"hindi po, 22 na po ako...\" \"bakit ayaw mo mag-abroad? bata ka pa naman pala, mas madaming pera dun.\" \"tinatapos ko po kasi yung master\'s degree ko\" \"kayo talagang mga bata kayo, puro kayo aral. ang dapat nyo lang matutunan e kung pano kumita ng maraming pera!   At the back of mind I was like, well that\'s why I\'m studying you dingbat, coz I want to earn more money. As I looked out the cab window I realized I was still somewhere in Boni Ave, dang, I still have a couple of minutes with this guy. Just makes you want to ride a bus. I\'m never waking up late again!  '),(7313,2539406,'04,August,2004','2004-08-04 00:00:00','   today i woke up with the promise of opening the CCNA study guide i borrowed from my officemate. and it was an issue i was willing to argue about with one of my friends. \"just stick to programming, you\'re good at it. besides, networking is just a textbook thing\" is what he said. doh! i for my rebuttal, i pointed out to him that it\'s just me trying to make a difference by learning stuff that i think will be handy later, you know, for work, because i\'m on the MIS department of our company. then it hit me, here i go again trying to learn a new thing when i\'m stuck with stuff to do like 4 more papers to finish for my master\'s degree, the project improvement of the IMS i\'m developing for work, plus a couple of things. after that thought, i opened the book anyway. hey jack, u\'r master of nuthin!  '),(7314,2539406,'04,August,2004','2004-08-04 00:00:00','   1. IF YOU WERE TO CHOOSE YOUR OWN DEATH, HOW WOULD YOU LIKE TO DIE?   a quick and painless one    2. WHAT WOULD BE THE FIRST THING THAT YOU WOULD DO IF U WAKE UP ONE MORNING AND FIND OUT THAT YOU\'RE GENDER HAS CHANGED?   remove my clothes in front of a mirror, hehehe   3. IF YOU WERE GIVEN THE CHANCE TO BECOME INVISIBLE, WHAT WOULD BE THE FIRST THING THAT YOU WOULD DO?   become a superhero   4. WHAT IS YOUR FAVORITE CHILDHOOD MEMORY?   there\'s so many, infact, they\'re so many i can\'t even list one   5. WHAT WAS THE VERY FIRST MOVIE THAT YOU SAW?   one of those tagalog movies that i call pathetic now, i think it starred vic sotto   6. DO YOU EVER DREAM OF BEING A POLITICIAN?   despite the carreer tests showing that i\'m fit to be a politician, i would never want to be one.   7. WHAT WAS THE WEIRDEST DREAM THAT YOU HAD?   that i was a power ranger.    8. DO YOU BELIEVE IN GOD?  yes, and also the devil   9. WHAT WOULD YOU DO IF YOU WERE STUCK IN AN  ELEVATOR ALONE FOR AN HOUR?   sit still, conserve energy, then die, or i could just probably CALL FOR HELP!!!   10. DO YOU BELEIVE THAT FINISHING STUDIES IS THE KEY TO SUCCESS?   would that fact that i\'m finishing my master\'s degree be enough to answer this question?   11. DO YOU ENJOY WATCHING OR LISTENING TO GHOST  STORIES?   i enjoy watching them, but after the story it becomes a different scenario   12. HAVE YOU HAD ANY PARANORMAL EXPERIENCES?  i dont want to talk about it   13. WHAT\'S YOUR NEW YEAR\'S RESOLUTION?  lose weight, for one.  urlLink click for the rest of \'em    14. WHAT IS YOUR GREATEST FEAR?   things that crawl    IS IT OKAY FOR YOU TO FLIRT WITH ANOTHER GUY/GIRL THOUGH YOU ALREADY ARE COMMITTED WITH SOMEONE ELSE?   i find nothing wrong with it as long as you know your bounds   17. WHAT IS YOUR LEAST FAVORITE PLACE IN THE WORLD?   under the bed, and don\'t ask why.   18. DO YOU HAVE A STAGE FRIGHT?   my hand used to shake when i\'m talking in front of alot of people, but not anymore thansk to my previous job as a teacher.   19. HAVE YOU EVER IMAGINED YOURSELF KILLING ANYONE?   *gives a mischievous grin*   20. HAVE YOU EVER STOLE ANYTHING IN YOUR WHOLE LIFE?   yeah   21. DO YOU HAVE ANY CELEBRITY FRIENDS? NAME ALL...  well, i\'m friends with ROB FRISBEE, 1/2 of the winners of Amazing Race 1, though it\'s just by e-mail. BTW his books are great, you should buy them from his  urlLink website    22. HAVE YOU ENGANGED YOURSELF IN ANY BUSINESS  BEFORE?   does monkey-business count here?   23. ARE YOU COMPUTER LITERATE?   i\'m an applications software developer!   24. WHAT BORES YOU?   somebody boring   25. IF OFFERED A MILLION PESO, WOULD YOU RUN  NAKED IN YOUR STREET?   i\'m not really that desperate for money   26. WHAT IS YOUR FAVORITE RIDE AT ENCHANTED KINGDOM?   the carousel? nah, its rialto   27. HAVE YOU EVER BEEN CONFINED IN A HOSPITAL?  WHY?   new year\'s eve 2000, a panic attack.   28. ONCE IN YOUR LIFE, DID YOU EVER DREAM OF  BECOMING A NUN OR A PRIEST?   my parents will laugh at the idea. and that will be the biggest joke i could tell my friends   29. WHAT IS YOUR CURRENT FAVORITE COMMERCIAL?    coke commercial on the jeepney, i think. well, that was the last one that i saw last nigh while watching TV   30. WHAT IS THE MOST IMPORTANT LESSON THAT  YOU\'VE LEARNED IN YOUR LIFE?   that when a door of opportunity is closed, another one is opened.   31. WATS UR GREATEST DREAM?  to dream, the impossible dream..... to light, the unlightable hope....  '),(7315,2539406,'04,August,2004','2004-08-04 00:00:00','   today, being a tuesday, is one of the days where i try my socks off to get to work before 8AM, because i have night class and can\'t afford to miss any lessons anymore. now the only way to beat traffic is to ride a jeepney at Buendia instead of taking the bus straight to Ayala Ave, because, one, it\'s much faster, and two, it\'s nearer, opposed to the fact that i don\'t really enjoy riding a jeepney. funny thing, while i was getting money for the fare, i realized that my coins were one peso short, so i had to take out my wallet and notice that the smallest denomination i had was a 100-peso bill (and jeepfare costs only 4 pesos, right!) the driver, still pissed, returned the favor by giving change in coins! 18 5-peso and 6 1-peso coins. dang, and to think that i didn\'t bring a bag today...  '),(7316,2539406,'04,August,2004','2004-08-04 00:00:00','   Al, you\'re a Politician!      Your personality is actually determined by two personality sub-types - your primary, or dominant sub-type, and your secondary sub-type. You are a Politician which means you are a Success / Golden Your primary sub-type is defined by \"Success\" characteristics and your secondary sub-type is defined by \"Golden\" characteristics.   That means you\'re forceful, innovative, and popular with the masses. Chances are you aggressively pursue your goals, and you enjoy impressing people. You\'ve always been surrounded by friends and loved ones and have a natural sense of showmanship.  How do we know all this? How do we know that you\'re perfect for any work that requires personal interaction? How could we have divined that you would never say anything that offends anyone â€” at least to their face?  Because while you were taking the test, you answered four different types of questions â€” questions that measured confidence, apprehension, willingness to take risks, and your focus on experience versus appearance â€” the primary traits that determine your personality. Based on your responses, we determined your personality type, Politician.  And that\'s just scratching the surface.     this is what i got after taking the  ULTIMATE PERSONALITY TEST  from   emode.com . dang, can\'t I be a sky diver or a secret agent???  '),(7317,2539406,'04,August,2004','2004-08-04 00:00:00','   it was already 4PM and i was hesitant to go to school yesterday, but i felt the need to go because i haven\'t attended class for three straight weeks already, well due to liesure and work conflicts (hey, what would you prioritize, huh?) so anyway i got to school at around 5:30PM, despite our class starting at 5PM, only to find out that the class was cancelled, dangit! next time i should trust my instincts more. so since it was still early, i decided to see brother bear.  so after buying the ticket and being in-line to buy some food, a woman barged out from the cinema and talked to probably her boyfriend who was in front of me, telling him that they were able to get seats for at the back row, however while they were going to get to the seats, two girls came in from the back and placed their bags on the seats they wanted, so they have to sit at the third row from the front. yikes! well, it\'s not my fault that i was able to eavesdrop on their conversation, i was bored in line waiting for my turn to order my food, and she talks loud. but i must say that i do feel sorry for them, and i felt sick of the people who did that to them. oh well.  now as i entered the cinema, i was able to find two available seats at the back row. so i took the seat and while i was watching the movie, i realized that one seat away from me were two girls with backpacks, so i safely deduced that those were the girls that i kinda hated for the misfortune of four other people. on the other hand those if not for them then maybe i would be sitting some place nearer to the cinema, so in a way i should be thankful. the moral lesson of the story, try not to eavesdrop on other people\'s conversations so as not to make life complicated!  '),(7318,2539406,'04,August,2004','2004-08-04 00:00:00','   8:05 AM. funny, i\'m still here at home,  when i\'m supposed to be at work!  it\'s a bad case of tummy upset, and now i\'d be able to finish getting the requirements i need for the \"promotion.\" and i get to watch american idol and survivor, booyah!   ps. i think rudy will be booted off the island on this ep  '),(7319,2539406,'04,August,2004','2004-08-04 00:00:00','   it\'s 1 AM here. obviously mr. sandman hasn\'t dropped by yet. come to think of it, maybe he did drop by but i\'m just too \"excited\" to sleep. (pardon the term for a lack of any appropriate word to describe the sensation i\'m feeling right now) i\'ll be taking my civil service eligibility exam today and i need my rest. but apparently i\'m still in front of the computer blabbering while playing trivia at MIRC. oh well...  '),(7320,2539406,'04,August,2004','2004-08-04 00:00:00','   Part One: Basics 1. NAME:  Aldwin Chester Palomo   2. SEX:   not after i get married (yeah, right!)   3. BIRTHDATE:   July 22, 1981   4. WHERE DO YOU LIVE:   QC, Phils   5. HEIGHT:  5\'6\"   6. SHOE SIZE:  8 1/2   7. FIRST CRUSH:  i think it was joanna   8. PARENTS:  Alberto and Regina   9. RIGHTY OR LEFTY:  lefty   10. WHATS YOUR SIGN:  I don\'t believe in astrology   11. WHATS THE FIRST IMPRESSION OF YOU:  mysterious   12. EVER BEEN IN LOVE:  of course!   13. PUPPY LOVE:  well, i had a puppy once...   14. HAVE A TATTOO:  not those permanent ones    Part Two: Foods 1. CHOCOLATE MILK OR HOT CHOCOLATE:  hot choco   2. MCDONALDS OR BURGER KING:  whopper all the way!   3. COKE OR PEPSI:  doesn\'t matter   4. ROOT BEER OR DR. PEPPER:  rootbeer   5. TEA OR COFFEE:  definitely coffee   6. MILK/DARK/WHITE CHOCALATE:  white choco   7. VANILLA OR CHOCOLATE:  choco   8. CAKE OR PIE:  cake    Part Three: Pick one 1. MARRY PERFECT FRIEND OR PERFECT LOVER:  lover   2. CATS OR DOGS:  man\'s best friend   3. 1 PILLOW OR 2:  2+   4. W/ OR W/O ICE CUBES:  w/o if i have a cold   5. TOP OR BOTTOM :  what for?   6. WINTER/SPRING/SUMMER/FALL:  summer, baby!   7. SKIING OR BOARDING:  well, we don\'t have snow here!   8. BIKING OR BLADING:  blading   9. ROCK/RAP/R&B/PUNK/ALT/SKA/TECHNO/SWING:  night, coz i luv to party!   11. GLOVES OR MITTENS:  gloves, BOXING GLOVES!   12. DRESSED OR UNDRESSED:  depends on the occasion   13. BUNK OR WATER BED:  water bed   14.MAKE PLANS OR GO ALONG:  go with the flow!   15. TRUTH OR DARE:  dare!   16. MTV OR VH1:   MTV rules!   17. OCEAN OR POOL:  ocean, i love its smell!   18. SHOWERS OR BATHS:  long baths   19. LOVE OR LUST:  depends on my mood   20. SILVER OR GOLD:  gold   21. DIAMONDS OR PEARLS:  neither?   22. PENCIL OR PEN:  pen   23. SUNRISE OR SUNSETS:  the beautiful sunset   24. BLIND OR DEAF:  i\'f rather be deaf    Part four: Favorites 1. COLOR:  black, blue, orange   2. NUMBER:  3 or 4   3. SUBJECT:  anything but history or home ec   4. CLOTHING BRAND:  rl   5. SHOE BRAND:  DMs   6. SPORT TO PLAY:  badminton, tennis   7. DRINK:  oj   8. ANIMAL:  uhm, a frog?   9. HOLIDAY:  new year\'s day   10. FAVORITE LINE FROM A MOVIE:  \"I wanted it to be you. I wanted it to be you so badly\" from You\'ve got Mail   11. BAND:  U2, BoyzIIMen   12. MOVIE:  lotr, the godfather   13. PLACE TO CHILL:  a bar   14. FLOWER:  i don\'t like flowers    Part five: Future 1. HOW MANY KIDS:  3   2. COLLEGE:  did it   3. JOB:  owner of a software company   4a. WHAT DOES YOUR IDEAL GF/BF LOOK LIKE:  sweet, smart, understanding, patient, lovely, understanding, interesting, understanding...   4b. HOW DO THEY ACT:  uhm, normal?    Part Six: Would you? 1. MOVE ANYWHERE:  yeah, after wasting time on this i\'ll leave my \'puter and get sumthin to eat, in the kitchen   2. DYE YOUR HAIR:  never considered it yet   3. MEET ONE FAMOUS PERSON:  yeah, robert deniro or michaelangelo   4. LIVE WITH ONE PERSON FOR THE REST OF YOUR LIFE:  probably not with just one...   5. NAME ONE PERSON YOU LOVE:  my mum   6. NAME ONE THING THAT EMBARRASSES YOU:  cant think of any   7. DO YOU LIKE SCHOOL?  am taking my master\'s, nuff said   8. DO YOU LIKE TALKING ON THE PHONE?  for long, long hours   9. a. DO YOU HAVE YOUR OWN LINE?  i have my own celfone!      b. CAN WE HAVE YOUR NUMBER/EMAIL?  aleck_palomo@edsamail.com.ph   10. DO YOU LIKE TO DANCE?  no!   11. ARE YOU SCARED TO ASK SOMEONE OUT?  definitely not   12. HAVE YOU EVER GONE SKINNY DIPPING?  it\'s in my list to do   13. HAVE YOU EVER THOUGHT YOU WERE GONNA DIE?  yeah, when i had my first panic attack   14. DO YOU SLEEP WITH STUFFED ANIMALS?  whisphers (yes)   15. HAVE YOU EVER BROKEN/FRACTURED A BONE?  not yet   16. DO YOU HAVE ANY PIERCINGS?  i used to have one in my right ear   17. DO YOU WEAR BRACES?  nope   18. DO YOU CONSIDER YOURSELF A GOOD LISTENER?  my mind tends to wander alot   19. CAN YOU SWIM?  yeah, dogstyle baby, lol   21. DO YOU THINK CHEERLEADING IS A SPORT?  no, especially after i saw \"bring it on\"   22. HAVE YOU EVER STOLEN ANYTHING?  are you a cop or sumthin?   23. WHAT\'S ON YOUR CEILING?  glow-in-the-dark stars and planets ans spaceships and...   24.WHAT\'S YOUR WORST INJURY EVER?  broken heart, sob   25. WHAT\'S THE HARDEST THING ABOUT GROWING UP?  fitting into your clothes   26. DO YOU BELIEVE IN LOVE AT FIRST SIGHT?  no!  '),(7321,2539406,'04,August,2004','2004-08-04 00:00:00','   a new record, although i\'m not proud of this one. i\'ve been under the influence of alcohol for five straight days now. a couple more and maybe i\'ll die. too bad it\'s not a slow and painful one like i\'ve always wanted. hey, i can\'t believe i\'m still capable of writing on this, uhm, thing while my head spins around. dang, i just hope i\'m pressing the right keys...  '),(7322,2539406,'04,August,2004','2004-08-04 00:00:00','   i\'m wasted as usual, and i cried myself to sleep last night. my girlfriend hasn\'t called yet and i\'m stuck at home. my friends have been buggin me all morning to leave the house and get some fresh air. i didn\'t feel like going out so i politely declined to their invitation. the next thing i know, adam came into my room with a pail of water and emptied it all on my bed. with freakin me on it.it was ok, i thought, my bed\'s all wet with tears anyway.  we decided to go to baguio, just so they would shut up, of course i gave them some serious whacking first. hey, maybe going out would be a good idea after all. besides, where would i sleep tonight?  '),(7323,2539406,'04,August,2004','2004-08-04 00:00:00','   1. To deliver up to an enemy; to be a traitor to. 2. To prove faithless, to  to disappoint : to betray a trust   you know what\'s funny? how can a negative word like  betray  be a feeling induced by a loved-one? I mean, why do we feel betrayed whenever a loved-one tries to do say a certain something to you, but you wouldn\'t feel anything if a complete stranger would do exactly the same thing? why am i so freaking emotional now? can someone please whack me on the head or sumthin?  '),(7324,2539406,'04,August,2004','2004-08-04 00:00:00','   American Idol 3 was officially over in my book when the person I\'m pulling for to win, namely Matt Rogers, was booted off tonight. It was a bit wierd that I was able to relate to the guy and I thought that he did have a good voice. I hope the guy who looks like a  hobbit  wins, or maybe the girl who has red hair and sings power of love. Dang, first, Josh didn\'t win AM2 (no offense to Ruben, but I did like Kim Locke more than him) and now this. How depressing. At least the two hot filipinas are still in the show.  Menawhile, on camp Pearl Islands, a recap of the first 21 days. Reality TV this week sucked! I gotta grab a bottle of beer. Dayyynnnng!!!  '),(7325,2539406,'04,August,2004','2004-08-04 00:00:00','   After revealing something I should\'ve revealed months ago to my girlfriend, It is the first time that I\'ve heard her use one-word sentences to answer my questions. My officemates mentioned that she probably still is in a state of shock, and I\'m just waiting for the right time to ask her how she\'s doing. Man, why can\'t girls be more like guys that can say whatever is on their mind?  '),(7326,2539406,'04,August,2004','2004-08-04 00:00:00','   Just got in from a 12-hour ride from Bicol. I attended a wedding which I thought would be nothing more than whale-shark watching. I waited for my officemate to leave first before I type to my heart\'s content. Suddenly I\'m too tired to tell the story. Seriously.  '),(7327,2539406,'04,August,2004','2004-08-04 00:00:00','   i kept blabbering about my computer not working only to find out that the plug wasn\'t secured tightly on the extension cord. some computer geek i am.  credit to tbrc for the fitting title  '),(7328,2539406,'04,August,2004','2004-08-04 00:00:00','   An SMS message from an ex-UP teacher mentioned a tax refund that I can get, so I decided to pay my Alma Mater a visit, and collect it\'s dues. It would also be a good time to visit my old officemates and spend time with them, before going back to the toms of papers I need to finish. I spent two nights there and as I was leaving, I couldn\'t help but feel sad. I sure did miss that place, and the memories are still there. As I jumped in the bus on the way home, I banged my head three times on the railing of the chair in front of me, while asking myself why I ever considered resigning in the first place.  '),(7329,2539406,'04,August,2004','2004-08-04 00:00:00','   Why is it that when you tell yourself to stop thinking about things, that unconsciously your mind tends to remind you of it?  '),(7330,2539406,'04,August,2004','2004-08-04 00:00:00','   Today is the start of 12-day study leave to be able to finish the requirements for my Master\'s degree. And what better way to start it than to wake up at 2:30pm and go to the University Library. Oh joy.  '),(7331,2539406,'04,August,2004','2004-08-04 00:00:00','   I compare myself to a snail with regards to making relationships, may it be with acquaintances, friends, generally everyone I meet on a day-to-day basis.  You may see me as the mysterious, quiet guy who likes to be in the comfort of his shell, until I get at ease and start crawling out of it\'s protective zone into the unfamiliar surroundings. Well I guess there\'s nothing wrong with leaving your comfort zone, but the problem with me is that once I leave that shell, I tend to be careless. Meaning, when you sum it all up,  it\'s me who gets hurt badly.   They call it  emotional attachment . We are social animals by nature, depending on other people for things such as trust, encouragement and love. It\'s a simple equation if you ask me. The moment you start sharing yourself to others, you give them a piece of what you have. At the same time you build trust and become more comfortable with that person as probably a companion in some of the routines you do everyday, or a confidant for problems that you can\'t keep to yourself.   It\'s awkward that when you don\'t see that familiar face, you feel a bit empty. It\'s like, the taste of coffee just isn\'t the same without the talks that you have with your friend, or that the long walks seem like a waste of time without the one you love wasting that time with you. I mean, it\'s okay to attach yourself to people because, as they say, no man is an island, but  when should you stop , to avoid the hurt?   When should you stop sharing things to your confidants? Does the equation have to be balanced, meaning should the amount you share and the amount the other party shares be equal? Why does it hurt when the other party doesn\'t meet the expectations you set for them as friends or loved ones? Do you have to set expectations?  I compare myself to a snail with regards to making relationships. Day by day as I share myself to other people, I let myself succumb to the hurts that are waiting to attack my soft exterior. When the pain of a broken relationship sinks in, it\'s the same time that I crawl back to my shell, my stronghold, where I lick the wounds of betrayal.  I would feel strong, but somehow I feel that I lost part of me, a part that I won\'t be able to get back.  '),(7332,2539406,'04,August,2004','2004-08-04 00:00:00','   i would definitely think twice when i ask somebody his/her profession and they reply \"AUDITOR.\" dang, i may even scream in terror with my hands up like i did while the word \"AUDIT\" was being mentioned over and over again. and yeah, i won\'t be able to look at a \"BALANCE SHEET\" the same way again. AAAAAAAHHHHHHH!!!!!!  (no pun intended for any auditors out there!!)  '),(7333,2539406,'04,August,2004','2004-08-04 00:00:00','   i took a couple of tests found in the internet via  urlLink allthetests.com  , ansd here\'se what i found out about myself.  1. my IQ is 137, and that i\'m highly intelligent. ooh.   2. after taking the personality quiz, it remarked:  You are a real  soprano.  you got a bad temper and you demand respect. You understand the value of money and want the best of the best   3. i got 0 points on the metrosexual quiz, therefore i\'m not metrosexual. it was a debate i had with one of my friends (and, coincidently, with an officemate), after saying that i don\'t dare to take time shopping, or sumthin like that. actually, i took the test without even trying to figure out what it means first, but i realized that i don\'t really have to know what it really means coz i\'m not really one.  4. now, as i  ask myself if i have a malicious mind, this is what the test said:  With your 50 % you are: You\'re trying to be insane, malicious, whatever you wanna call it but you can only cut it as TOUGH, I\'m afraid, but after reading this, if you feel like hunting me down and killing me, you\'re insane like me. But don\'t try it, I can kill you a million times before you can switch your pc off............*mwa ha ha* .  5. ok, now i\'m asking some vain questions. after taking the \"are you sexy\" quiz, it said that i am pretty sexy, and with some work, i could be really good. There are some people sexier than i am, though. (damnit).  6. am i whacked?  .You are sort of whack! That\'s normal! Have fun with your doggy- woggy and scream \"woo- hoo!\"    7. and finally, i asked simon cowell if i can be the next american idol, and he said that he\'s heard better subway musicians that me. aww, that\'s gotta hurt.  '),(7334,2539406,'04,August,2004','2004-08-04 00:00:00','   yesterday, me and an officemate ate at lydia\'s lechon for dinner. being the health buff that i am now (well, i\'m trying to, and for starters i have to watch what i eat) i passed on the roast pig to order their second specialty which is called paella, a mix of fried rice, green peas, crabs, roast pig, hotdogs, pork barbecue and mussels on a big platter. i decided to order it because i was, kinda hungry and i never thought it would be that big of a plate, i mean, cmon, false advertising on what restaurants serve? to my surprise it was the same size, and i couldn\'t finish the plate.   my officemate, while laughing out loud, offered his \"condolence\" with the remark \"maraming batang nagugutom sa pilipinas.\" i could only reply \"tama na po, hindi na mauulit,\" sounding like a kid being spanked for doing something bad. \"o ano, kakain ka pa ng paella sa lydia\'s lechon?\" asked my officemate. again, i had to remark the same words. :o  '),(7335,2539406,'04,August,2004','2004-08-04 00:00:00','   it\'s 12:53 AM and my tummy is growling because of the big bag of lays that i ate while watching a tagalog movie that i happened to like (btw it\'s entitled keka, highly recommended if you know tagalog.) hey wait a minute, wasn\'t that line used in one of my posts before? oh yes it was. dangit, i still haven\'t learned, apparently. i better stop munching on big bags of chips when i\' watching movies. oh well, can you blame me? btw i\'m still contemplating on the issue regarding the nationwide, uhm, PUV strike that will happen later on. in case you don\'t know what that means, well, vehicles won\'t be gracing the steets of the metro, well, not only on the metro (since it\'s supposed to be nationwide), and i\'m not sure if there\'s work. oh well, whatever happens, i need to get my sleep. i\'ll think about that later in the morning when i wake up. for now, i gotta run somewhere and take a crap. booyah!  ps. there\'ll probably be more posts entitled \"nonsense\" from now on. it\'s gonna be like, well, what i\'ll call when i can\'t think of anything to sum up what i have written, which is most likely, well, nonsensical.  '),(7336,2539406,'04,August,2004','2004-08-04 00:00:00','   Before goin out to drink, me and a workmate stopped by a home improvement center.  We feasted our eyes with some things that we wanted to buy, including this lazyboy with a massager, while thinking about living on a penthouse of a condominium and how we were gonna decorate it. Somehow it felt a bit wierd, coz I\'ve never entered a home improvement shop with a friend before.  '),(7337,2539406,'04,August,2004','2004-08-04 00:00:00','   ...on a wednesday? now that\'s gotta be the wierdest idea. but i gotta admit that i\'m looking forward to it. dang, it\'s true, I AM WIERD!  '),(7338,2539406,'04,August,2004','2004-08-04 00:00:00','   Today I drove with my mum to the airport because she needs to go back to Kagoshima for wok. I guess it\'s true, that you tend to neglect people if you see them often, but you you\'ll never know how much you really love them until they\'re gone. I love you mom, too bad you couldn\'t read this. Although, in some wierd way, I know that you already know that.  '),(7339,2539406,'04,August,2004','2004-08-04 00:00:00','   Danny, one of my best friends dropped by earlier today, saying he\'ll be staying in Manila for a couple of days, which is really cool. I must admit that I am really happy to see an old friend after all that has happened for the past weeks. Danny has always been the sensitive guy who knows what to say at the right time, and thus a tribute is given to him here in my humble blog. To Danny, a true friend, a great companion.  '),(7340,2539406,'04,August,2004','2004-08-04 00:00:00','   Why do I always get emotional when things don\'t go my way? Why do I feel easily betrayed? Why do I make a big deal out of what people say? Why can\'t I just grow up?  '),(7341,2539406,'04,August,2004','2004-08-04 00:00:00','   I have to admit that things at work weren\'t really looking up the way I wanted it to be. As the days pass, especially with the recent incident with my girlfriend, plus a couple of other stuff, I feel as though I want to quit the job I have and move to the United States, just to get away from it all. Today however, things here at work have gotten better (at least I hope so.) I presented my recently developed software application to a higher-up. Thankfully, she liked it and said that with a couple of adjustments, it would fit their needs. She then asked me why I wasn\'t a permanent employee yet. I explained everything and she said that hopefully come June 1, I\'ll get the permanent tenure that I deserved. She even suggested that I go through the selection board come afternoon so as to get it done with. Too bad it was already 6PM and everybody needed to go home, so I guess my interview will be done either next week or the week after. Wow, is this really happening?  '),(7342,2539406,'04,August,2004','2004-08-04 00:00:00','   Last night, an SMS from a workmate while he was waiting for me on a sofa:  \"Pare take your time\"  plus a smiley. It took me more than twenty minutes to meet him where he was standing (or sitting.) He asked me how it was, and I told him  it was fine. In fact, it was too fine . It was so fine I wasn\'t able to go to work today. I did take my time. Now those three little words would have a whole different meaning.  '),(7343,2539406,'04,August,2004','2004-08-04 00:00:00','   typical worknight, leave work before 8 and go to Glorietta to grab a bite to eat before goin home. sometimes a couple of bottles of beer so as to ensure a goodnight\'s sleep and a summary of the days events. well, tonight was no different, except for a rather funny incident.  after ordering a zinger from KFC, i pulled away from the counter to look for my workmate, hoping that he already found us a vacant table because it was a full house that night.i treid my best to squeeze my way into the flooding people going against me while trying to find my companion, but after two minutes of time without avail, i pulled out my celfone and dialled my workmate\'s number.   he answered to find out where i was, while trying to figure out his directions. by the time i got to the table, the call was still active and we looked stupid talking to each other over our celfones, wehen in fact we were just inches away. it was kinda okay, though,because the two ladies sitting next to us were, uhm, laughing like crazy. i remarked,  at least we were able to entertain a couple of ladies, hehehe.   good thing they weren\'t really that attractive, or else i wouldn\'t have the strength to stop myself from flirting again. i mean, cmon, you gotta admit that it was a good icebreaker! (right, hit on a woman who laughed at your expense)  '),(7344,2539406,'04,August,2004','2004-08-04 00:00:00','   one word,  karaoke!  here\'s a summary of the songs (at least the ones i could remember) that i sang while supposedly on vacation during the lent:   First Set: The Love Songs  Amazed Three Times A Lady She Believes in Me  I\'ll Be - Edwin Mccain All Or Noting I Don\'t Wanna Miss A Thing (i nailed this one, a perfect 100)   Set Two, can there be any other songs more pathetic than these?  Just Once Desperado  (another 100!) One Last Cry You Won\'t See Me Crying Bakit Ngayon Ka Lang Put Your head on my Shoulders (?) Unchained Melody (??)   Third Set, singin til the break of dawn  Liquid Dreams Fame Bikining Itim This Guy\'s in love with you, pare (whammy!!!) She Bangs  With that, a remark, sadly I couldn\'t remember from who, but i\'m 100% positive that it\'s for me:  Now I know why they call it Black Saturday.  '),(7345,2539406,'04,August,2004','2004-08-04 00:00:00','   People\'s reactions after I say  I want to die a slow and painful death!   an officemate:  babae lang yun!  another officemate  sige, uumpisahan na namin ngayon  while he continues to pinch me to death, doh! a friend:  sige pare, inom lang tayo, gwark!  another friend:  pucha pare, humihina ka na sa babae  yet another friend:  if i will ask you to pay me one peso everytime you say that, i would have enough money to buy myself a recorder and listen to you say that over and over again. heck, i\'d even have extra to buy a blank tape.  '),(7346,2539406,'04,August,2004','2004-08-04 00:00:00','   nine straight days, whoah! i\'m a drunkard!  '),(7347,2539406,'04,August,2004','2004-08-04 00:00:00','   it\'s the start of the month, dayng, i\'m back at work!  '),(7348,2539406,'04,August,2004','2004-08-04 00:00:00','   We now continue our coverage of Alec\'s journey to his IS Officer position come June 1. Last time, the final interview. Earlier today, and hopefully up to tomorrow only, the requirements.   Today, I decided not to go to work so I can take my medical examination. I was with two friends They volunteered to hold me while the nurses take blood from me using needles. It was too embarassing. Finally when the results came out, I felt a sigh of relief. The IS Officer position is as good as a few more spaces within my reach.   The only thing left is the tons of forms that I have to fill out, handwritten. Oh well, so close, yet so far.  '),(7349,2539406,'04,August,2004','2004-08-04 00:00:00','   Well technically the interview with the President for the IS Officer position is tomorrow, however it was moved today, in the morning, where I was basically caught off-guard. It took like less than 5 minutes, the whole time I was sweating bullets, despite the airconditioning working perfectly. At last, it was done. And it\'s official. I just can\'t wait for June 1 to come. Now I have to finish my requirements. Oh, the dreaded medical examination.  '),(7350,2539406,'04,August,2004','2004-08-04 00:00:00','   Ah, it\'s election time once again in the Republic of the Philippines. Yet another power struggle that I pretty much don\'t want my hands on. Not today, not in the past, and definitely not in the near future. (Yep, you guessed it, I\'M NOT VOTING!!!) But I am kinda glad that it\'s a national holiday so I got a 3-day vacation from work. I never got to watch survivor\'s live feed, but I was able to know that Amber won the All-Stars version, and that Boston Rob proposed to her on the reunion show, which she accepted.   It was a bit disappointing for me because I wanted Boston Rob to win, well for a couple of reasons:  1. He was a great strategist. I mean, really. He did play the game exceptionally well. People may think that what he did was awful, but it\'s a game of deception, backstabbing and broken relationships. It\'s the nature of the game. Do everything you can to win the big money.  2. I also thought that Amber just basically rode on Rob\'s coattails. Geez, her only strategy was to flirt with Rob. But I gotta admit that she is cute. Too bad she\'s getting married :(  3. And finally, being the chauvinist that I am, of course I would want a guy to win the All-Stars. Too bad that\'s not gonna happen, unless another All-Star season comes in.   Well now that another conniving, cutthroat season of survivor has come to a close, I\'ll just have to wait for the 5th installment of the Amazing Race coming in on July. I just hope that this is not as disappointing as Flo winning in season 3 or the faggots winning the 4th one, or lese that\'ll be it for me and reality TV.   Just a thought, what if the five presidential candidates, instead of the normal elections, be marooned somewhere and play a game Survivor style? wouldn\'t that be nice, huh? Alec, out!  '),(7351,2539406,'04,August,2004','2004-08-04 00:00:00','   It\'s just that it has been a very long day and I have a feeling that it won\'t be ending soon. Today, (well technically it was supposedly yesterday, bcoz it\'s way passed midnight, meaning it\'s another day, dang, you know what I mean, right?) the selection board convened to \"interview\" me, and you are now looking at the new Information Systems Officer effective June 1, woohoo!  Sadly, what happened tonight basically broke my heart. I\'ll just go vague on this coz I really don\'t want to make a big deal about it. I just hated the fact that I\'m supposed to be happy, but instead because of some people who I consider friends, this night would be dreaded. This makes you even wonder who your real friends are. I hate being the emotional guy, but I guess I have to deal with this. It just sucks that you try to get out of your way just to make favors for people, but sadly the same people end up hurting you and they don\'t seem to care about it.  '),(7352,2539406,'04,August,2004','2004-08-04 00:00:00','   Just tryin out the new default skins. This one really looks nice, don\'t you think so? The problem is, it can hold only a fraction of the entries I had before. Might as well, I\'m getting tired of the old stuff anyway.  '),(7353,2539406,'04,August,2004','2004-08-04 00:00:00','   A scence in one of our offices while I was troubleshooting a computer. My workmates in the department call it  answering a service call , but somehow I don\'t really like the sound of it, so I\'ll go as it is. This was when one of my co-employees started a conversation:  Employee 1:  Marunong ka rin pala nyan? Akala ko programmer ka?  Me        :  Medyo, pero nag-aaral lang din po ako mag-troubleshoot.          Employee 2:  All around yan! Ganyan talaga sa MIS.  Me        :  Ganyan po talaga ang buhay!          Employee 3:  New graduate ka ba?          Me        :  Hindi po!          Employee 3:  Last Year?          Me        :  Actually pagraduate na nga po ako ng Master\'s ko e.           Well, the conversation with the Employee 3 doesn\'t have anything to do with what I\'m driving at, but it sure is nice to be mistaken for a new graduate. This has to mean that I look like 19, since I left college when I was 19. hehehehe. Hey! Who put the airconditioning at supercold?  '),(7354,2539406,'04,August,2004','2004-08-04 00:00:00','   Blogging recently has not been exciting for me. I have been struggling to type at least something worthwhile in what seems to be a page of blacks and blues. But the so-called \"writer\'s block\" shouldn\'t be blamed by itself. I mean, there are lots of things to write about, seriously. But I am trying, for once, to refrain from making this blog the output of my state of being an emotional basketcase.   A song entitled \"Don\'t Cry Out Loud\" made me think. I really don\'t know who sang it, but Diana Degarmo sang it during the finale of American Idol 3, and it just clicked. Another song popularized by Diana during this season of AI is \"I\'m Still Standing\" by Elton John. Somehow, that did fit too. So from now on, I\'ll try to do a balancing act. Lots of inner-self checking on my part, and again, patience, which I know I don\'t have. Of course it\'s worth a shot anyway.   Alec with a thick skin, now that\'s kinda new. Can I manage? Wait, I think the more important question is, will the people around me be able to manage it?  '),(7355,2539406,'04,August,2004','2004-08-04 00:00:00','   Amazing race is coming on July 6, and rumors are that the teams passed by the Philippines. Dang, where the heck was I?  Initially I don\'t like any of the teams. Looks like nobody stands out. It\'s not like season 4 with the Clowns and the Gay couple standing out. Wait a minute, come to think of it, there\'s no gay couple in this season. I guess they got fed up with the Chip\'n\'Dales last season.   I\'m buckling my seatbelt, I know it\'s been worth the wait.  '),(7356,2539406,'04,August,2004','2004-08-04 00:00:00','   Today I parted with my palm pilot. It was a hard thing to do, but it was for the best.  Oh yeah, I do have one. I acquired it almost two years ago and it was good for keeping lectures while I was teaching in the University. But now, my celfone doubles as an organizer also, so I don\'t have any use for it, except the games. At least i know that it\'s in the hands of a person who can use it to the fullest potential (what kind of crap is this!)  '),(7357,2539406,'04,August,2004','2004-08-04 00:00:00','   Oh, the long wait is over. I assumed the IS Officer position today. And guess what, no celebration. I\'m at home in front of my computer typing in my blog. Yippee!  Well, it was kinda a choice I made. I wanted to be alone. Plus, something disappointing happened today. I\'d rather not talk about it, I guess I\'ll just let this one pass. Maybe if I\'ll sleep over it, then I\'ll forget about it in the morning. Yweah, maybe.  '),(7358,2539406,'04,August,2004','2004-08-04 00:00:00','   It\'s been  a week in, and 3 gym sessions have passed after my first compliment, saying that  my   tummy is getting smaller . Three days of hard work paid off. Somehow, my aching joints stopped aching for about 10 milliseconds. Hey, at least it did stop. And this is a person that doesn\'t know that I\'m going to gym, mind you. Come to think of it, I have been silent about this. I\'ll save the bragging later when it takes full effect. Right now, the only thing I have to do is  not  get tired of working out, or the hefty sum I paid for it would be nothing but a faded memory, including my wanting to lose weight and tone my body.  As freebies for signing up, Gold\'s gym gave me an official tumbler and shirt, plus a cap to follow. Surprisingly, they gave me a  size S .  Then I thought,  when was the last time I wore a size S ? It went as far as sophomore high, I think. The shirt is orange in color and has the gym\'s logo near the bottom. I kinda like it, however I haven\'t tried it on. I kinda want the whole  I\'m gonna lose weight  thing goin, so I\'ll probably wear it once my tummy looks  in shape . Now if that isn\'t motivation, I dunno what is...  '),(7359,2539406,'04,August,2004','2004-08-04 00:00:00','   Ryan, one of my close friends, gave me a Star In A Million CD, yep, the filipino version of American Idol. My first impression was,  damn, buddy, do you think I\'m jologs?  I wanted to throw it away the moment I get a chance to, however I ended up playing it. It was Ok, I thought. But it can never replace the American Idol 3 Soul Classics album that I have near my component.  '),(7360,2539406,'04,August,2004','2004-08-04 00:00:00','   The blogger turned a year old today. And as always, I try to look back at the year that just passed and see if there is any improvement. Sadly, everytime I do this, I fall asleep, even before I figure one out. Today is no exception, so I figured I should try again next year. ;)  On a lighter note, a look on today\'s events is appropriate. Nuthin special, the usual 12MN -3AM is busy with calls from friends greeting me a happy birthday, and as always I try to answer all of em. A call from my parents during lunchtime while I was having lunch with my workmates. Nuthin unusual except I used to hear something like \"Happy Birthday, hope your well.\" This year however it was something like \"Happy Birthday, don\'t drink too much, stop exhausting yourself from work...\"  Today is a double celebration BTW, besides my birthday, it\'s my \"I\'m thankful for the promotion thingie\" so a feast was in order at work, well, it was just a small one. Everybody was like, \"You\'re just 23? wow. I can\'t believe you\'re that young...\" Yeah. I thought, me either. Then a department meeting. Then dinner with a friend. Part of the discussion went to marriage, which I figured later was a bit scary. In between well-wishers greeting me for a good year, parents telling me to stay low on the booze, and finding a better half, I can\'t help but want a break from all of this. This is too much, I thought. And to think that I still have work tomorrow.  Footnote:  Blog antidated to my exact birthdate and time.  '),(7361,2539406,'04,August,2004','2004-08-04 00:00:00','   Today has gotta be the happiest day for the MIS department of our corporation. Wait, lemme rephrase that... Today is the start of our better days, wooohooo!!! Before I sound crazy or cynical, the reason behind our celebration is that finally, our IT Committee, which basically in a more practical sense fried us alive everytime the meeting came to order, is now DISSOLVED. That\'s right, no more IT Comm, as per an office order prepared by oneof my workmates, and upon recommendation of the CEO of the corporation.  Just a footnote, while my workmate was preparing the office order, we were figuring out another word to replace DISSOLVED. I wanted to use INCINERATED, however we really didn\'t want to sound mean about it, inasmuch being ecstatic, duh! I guess we\'ll be singing \"Happy\" again. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. Can I say that again? Well, I guess nobody will stop me so...  No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM. No more IT COMM.  Ok, i\'ll stop now!!!  '),(7362,2539406,'04,August,2004','2004-08-04 00:00:00','   Friday night at Padi\'s Point Cubao. After downing a couple of bottles, a friend refered to the ladies dancing in front of the live band as \"bottles of beverage.\" He called one a  cervesa negra  because of the obviously dark complexion, and  1.5 Liter Coke  to a fat-assed girl,   just to name a few. I for one wasn\'t able to make a comment, because I was lauging so hard.  Dancing bottles, now that would be something nice to write about...  '),(7363,2539406,'04,August,2004','2004-08-04 00:00:00','   Today, the choice was, go to work or watch the UAAP. What, work on a saturday? Of course you know what my choice was. So there we were, reminiscing the college times. IT was all good, except for the fact that we lost more than 20 points. I was so disappointed I wished I just went to work.  '),(7364,2539406,'04,August,2004','2004-08-04 00:00:00','   July 6 - 7, 2004 was team building day for the corporation. I, for one, tried my best to be one with its goal (hey! you gotta give me credit for trying, huh!) I really coudn\'t say that I liked everything about it, but I guess when you sum it all up, the positive definitely outweighed the negative ones. The highlights of the event were:  1. Meeting other people in the corporation. It\'s funny that I see them more than once every week for the past eleven or less months, but it\'s only now that I got to know some of them. I realized that my world doesn\'t have to evolve around a certain circle.  2. The team challenges were Ok, and me being a tenacious competitor was shown in a lot of occassions.   3. There was this farewell thing when eaveryone was given white of yellow candles. We were asked to give the white candles to people who we want to say  thank you  too, while the yellow candles were given to people who we want to say  i\'m sorry  to. I was a bit surprised, coz I got a lot of candles, nothing that I ever expected. I hafta admit that after all that\'s happened, I felt special. Plus, it made me feel that I am part of the family. At last!  '),(7365,2539406,'04,August,2004','2004-08-04 00:00:00','   It has been one month in the new job, so I think an update on how I\'m doing should be proper. So far, except for a few \"TEAM\" issues, everything\'s working out fine. Of course it\'s still too early to make assumptions for long-term, but pretty much, at this rate, I\'m okay, surprisingly. There\'s too much to say, however I\'m pressed for time. I have to go somwhere far tomorrow but I haven\'t packed my bags yet.  '),(7366,2539406,'04,August,2004','2004-08-04 00:00:00','   Just when we thought that UP was gonna get it\'s first every win for UAAP season 67,  WE FREAKIN LOST AGAIN, by 1 FREAKIN POINT!!!! And to think that UP was commanding a big lead all throughout the game against DLSU,  the archers got the lead for the first time when the clock timed less than three minutes left in the quarter. It was a disappointing loss.  It was more sad coz I kinda thought that something like this might happen.  '),(7367,3416335,'20,June,2004','2004-06-20 00:00:00','   I just had a thought......          Reality is fake, I woke up last night thinking I was stuck in a dream, that everything was around me like a really dense cloud, I could fumble about, but nothing could be found.        But that\'s how it is in life, isn\'t it? When everything\'s clouded and you don\'t know where you\'re going, a light comes along and clears the clouds away. Either that or it rains, and while the rain may seem dreary for a while, you know that it leads to something clear and awesome after it\'s done.       I enjoy this journal, it lets me put my thoughts and my feelings down, whereas if I say what I want to say, it never comes out in the way it can go down if I am writing. I feel some kind of a lure to writing, it\'s provoking me to put a deeper thought, a deeper feeling down on the paper(or screen). I\'m thankful to this site for allowing us to have this opportunity for free because this writing is a hell-of-a-lot better than my handwriting.        Lately I\'ve been viewing sleep as an obstacle now, whereas I viewed it as an escape before. I don\'t want to go to sleep. Last night I didn\'t want to get in the car, but I did and I hated it. I want to continue what we had last night Mollie, I don\'t think it is anything more than a teenage relationship, but I figure what\'s life without little of that? I hope it can continue, and, one more thing, I\'m paying next time again! :)        I guess all things do really work out for good. I was mad last November when my parents wouldn\'t let me take drivers ed, but all these months later I realize that if i hadn\'t of taken drivers ed last week, would this have happened? All I\'m saying that is sometimes we look at the little things in life that are bad, when really everything else is good. We need to focus on the good, while taking care of the bad, because in doing so I believe that we all can be better people.        Now on the other side of things.........you\'re all dumb :)  '),(7368,3416335,'19,June,2004','2004-06-19 00:00:00','   all i know is all i need and all i show is what i bleed i need a cover to prove what i will want to remove all that i\'ve been through  is all that is not true.  Make the pain pure just wash it away it\'s been your best lure now end it today make it stop here you\'re causing pain still what i fear when will it wane.........  that is the result of 30 seconds boredom. i know the meter\'s off but why change it? i\'m too lazy.  '),(7369,3416335,'18,June,2004','2004-06-18 00:00:00','   Lo and behold, drivers education yields something useful.        Boredom is instilled in my mind, i feel the urge to post. I don\'t know what i\'m going to say or even what i\'m saying at this point, i just know my personalities are ok temporarily. Emphasis on temporarily. I\'m trying to find words to say, to present what is reality, to show what isn\'t tangible. I\'m here for a reason, but what it is lately i sometimes wonder.......I\'m not going for what I believe and i think that may be a good thing, since all I believe is all that\'s holding me back. I wish I had turrets so i could let out how I feel in an endless stream of words not allowed, and not be persecuted for it. I am going against my friend\'s best wishes, not doing what I know is right, but who cares? Who cares what is right and what is wrong? Who cares what will happen? I know one thing and one thing only: I\'m temporarily satisfied. Sometimes you wish time had a rollback and distances weren\'t so far, that you could jump a mile in one step and not worry about it anymore, but instead you have distance. Distance means time, time means more to cover, and blood, sweat, and tears. Reality is what we make it, and we make it not real sometimes. this is the way it is. I have accepted this and I hope you do too.  '),(7370,3416335,'17,June,2004','2004-06-17 00:00:00','   Yes i gave in and it worked. yay.  '),(7371,3416335,'16,June,2004','2004-06-16 00:00:00','   Oh drivers ed that doth so comely teach and deny my inner flesh\'s lust of driving fast and getting high to wonder why someday i\'ll drive in fields of concrete,oh so grey and wonder how i may die today with druggies, speedbumps, and trains i see that it will become part of me. I wish for this freedom so my inner sanctum refined i wish to eat a lot and mostly drive.  '),(7372,3416335,'16,June,2004','2004-06-16 00:00:00','   urlLink'),(7373,3416335,'16,June,2004','2004-06-16 00:00:00','   urlLink'),(7374,3416335,'16,June,2004','2004-06-16 00:00:00','   Recently i\'ve been a little insane. Just slightly. My multiple personality is taking over and it\'s an ugly thing, it is...really. One\'s name is jimbo, he\'s the crazy, outgoing one, one\'s name is jimmy, he\'s the one that\'s laid back and doesn\'t care, and the last one is jim, the one that\'s never sure and is always pissed about something. I\'d rather have Jimbo around b/c he takes the risks, but i know that jim is lurking in the backround saying \"wtf would you do that for?! you\'ve been through this before!! Don\'t kid yourself!\" Meanwhile jimmy\'s just saying chill out and smoke a joint. But i definitely won\'t do that. Also i have this other personality that is named Larry and he\'s a Mexican that just randomly shouts curse words at eskimos.        Yah, i missed newfound glory and that sucks, but i don\'t care, it was funny watching my teacher get pissed.......really.       I\'m still undecided, what should i do? Let my decision push people away, or maybe it will help and lead to getting hurt again in the end. Post ur comments, advice plz.       Now i leave you for today with this \"Don\'t kill the monkey before it eats the banana, because then you don\'t get to eat the banana.\"  '),(7375,3416335,'16,June,2004','2004-06-16 00:00:00','   Yah well, mollie and liz got me into this but hey, livejournal.com was too confusing and i like this better cause it\'s easy. So get over it. Anyway, a little about me.........       I\'m a weird kid. Not that weird is always a bad thing, but even a good thing can be bad when it\'s messed up. I don\'t think that many people understand that, but it works out, just think about it. I tend to wear everything I know and feel on my sleeve, and it creates a lot of problems, in life, in love, and in other shiznit. It makes people trust me, and yet not feel comfortable at the same time. I\'ve never really had problems making friends, or being made fun of despite my size and looks. I never really understood that fact. A lot of times in my life i\'m viewed as \"the kid that\'s there for you, the good friend\" and sometimes that gets old. I don\'t want to be the kid who\'s \"friendship people don\'t want to mess up\" and \"let\'s stay friends\" shit.        Thanx to mollie and liz for hanging out with me at drivers ed, which i have newly taken up and it\'s fun because of that. Sam, ur cool and so are u Colton. Yes these are my driver\'s ed buddies. Now that i have u guy\'s sn\'s and liz\'s phone #, life could be a lot more interesting........j/k. Now all i need is mollie\'s #............       It sucks.....I should be going to a concert tomorrow night..but i\'m not. Driver\'s ed got in the way of that. I\'m missing out on newfound glory for a bunch of useful information that will probably save my life.........good trade (in a way).       Well that\'s it for now. I\'ll write lata.  '),(7376,3416335,'11,July,2004','2004-07-11 00:00:00','   Rip my life away all  i know today, Pain is at fault now for losing my way. I am numb with trust and lost in my lust, Hating what I am, And falling to dust.  '),(7377,3416335,'05,July,2004','2004-07-05 00:00:00','     '),(7378,3416335,'05,July,2004','2004-07-05 00:00:00','   Yes here I am once again after a long time, finally blogging once more. New things in my life....Well let\'s see...       I\'ve been working like a Mexican pack horse lately at work, not that I don\'t like work so don\'t get me wrong, but the fact is it\'s an adventure everday. Today I dropped a huge-ass dock piling on my foot and it only weighed like 600 pounds. No big deal. Hurts like hell, but at least it\'s not broken. Wow I\'m actually talking to Beth right now and we\'re not flipping out, we\'re being cool with each other. What brought about this sudden change in our hatred toward each other? Who knows, this is just the way it turned out and I\'m willing to go along with it. I suppose I\'ll have to get used to it if she\'s going to the same school as me next year. I think there\'s a time to grow up, a time to forget the past, and a time to start new and pretend that the past never existed and everything is new. New to the fact that past ENEMIES can become friends again, and new to the fact that we\'re willing.        Mollie\'s now in Long Island still, and I miss her. I visited her house and we walked in the woods for a while (during which time I was raped repeatedly by massive swarms of satanic, bloodthirsty, hell-hounds known as mosquitoes) and had a lot of fun while she flipped out about all the mosquitoes on my back and proceded to slap me repeatedly on the back and neck and shoulders and head. You would\'ve thought that she was trying to kill me. Then she felt all bad because she brought me into the woods and said it was her fault that I got bit a million times. Whatever, I dont\' get it. Mollie I would take a million mosquitoe bites if it means spending time with you.       Here\'s a tibit that I just thought up:  You\'re a pain in the ass so shut the fuck up.       Not in the poetic mood today. Oh well, if you\'re offended do this: Read the tidbit one more time.  '),(7379,3354517,'22,June,2004','2004-06-22 00:00:00','   the cup is not half empty as pessimists say as far as he\'s sees nothings left in the cup a whole cup full of nothing for him to induldge since the voice of ambition has long since been shut up    aaaaaaaaaaaaaaaaah.  okay papa just called. made me feel slightly better tt shes nt all awkward or anyth. cos she saw.            if i slit my wrists wad wud happen.  no, okay first thing wud anyone care. wud they do an autopsy?   //just.want.the.pain.to.go.away.  if i made two cuts above my eyes maybe the pain wud seep out from there. maybe all the hurt would just wash away. maybe if i drowned myself or something     probably haram to think all these things, yes?  yes.   '),(7380,3354517,'22,June,2004','2004-06-22 00:00:00','   cant. do this. on my own.  dont feel like im here anymore. like someone took all the semangat outta me.  and still in spite of all this or because of it  or something this is what my life would be like without him or my friends init. really really empty. and it could be like that. anytime :(   i.cant.do.this.  if u get balasan does it mean u ahvent been forgiven?    try. climb trees and live somewhere else. for once running wont do any good. sadness will follow u anywhere. i could. anytime.  :\'(  is it stupid? crying cos u could lose anyone at anytime. him at anytime. stupid stupid stupid. and i dont want to. and maybe i just messed it up . i dont know :( somehow i dun think it wud be as big a tragedy for him. why am i thinking this at all?  i know why.  not on my own.  and the people in the next flat can probably see me all like this staring at teh screen.  maybe this is balasan for yelling at li. i dont know.  maybe running away =wud- help at some level. at least maybe if i didnt see hear or feel anything familiar or anyth from here it wudnt hurt so much. why do we go thru life just flitting thru people? forgetting people or wanting to forget. does it not mean anything :( am i never gonna see you again?    what s the point. self pity? loathe that. but these are just thoughts in my head, am i not entitled to let them out?  why let them out? so they can mess up other people? i think my thoughts would be greyish blue like the colour of bruises. floaty but they sting when they hit you.    i dont wanna be your mistake  not dreaming now. its just a tragedy.   '),(7381,3354517,'22,June,2004','2004-06-22 00:00:00','   *thishasbeenareallytoughmorning.  -hmhm. now i dunno where to run (:  negative happiness? perhaps! hahaha. wheeness. oh this is so balasan. for yesterday. wonder if anyth else will come out of it. hopefully not! hmhm. (: so ure running. and without anyone to love you. and without anywhere to go. and without direction. and with like some unsolvable problems on ur back. den where to go? (:  ahaha. crap.  die die die.   '),(7382,3354517,'17,July,2004','2004-07-17 00:00:00','   mab:ili!!! wud hav liked to hav been there..ahaha [: mosh... dunno hw to explain to matin i feel like being all havoc again n forgetting :X [li]:Maab! Ya damn fun next time we all go gigs together k? Ya that\'s why i just suddenly miss being havoc and what. Sigh. After o levels! I\'ll be free! mab:yesness next gig kiter pergiiii!! (:(: korang nampak nadia or not? k after o levels havoc balik ah? *sticks out pinkie promis?    anden after i typed tt. i terswitch to another window n saw sth n the first thing tt popped intomy head was tt he was probably the reason. for changing and stuff. *dum de dum*   '),(7383,3354517,'17,July,2004','2004-07-17 00:00:00','   i dont wanna listen i wanna go to sleep i wanna go away i wanna run i need something else i dun know what i need so i cant run cos i couldbe running to it maybe i just need to be alone maybe i need soemone else maybe i should find a sexy skinhead AHAHAA sorry beautiful bald boys dun exist or maybe they doooo. hahaha ahhh k. i wanna scream run and hide NO DUN HIDE! then wad to do okay il just slack and maybe ill find something in me. mmhmmmmmm   '),(7384,3819668,'01,July,2004','2004-07-01 00:00:00','    Beleive me, if you really apply yourself to this business, you will earn a lot!! I just came back from our regional conference and one lady has sold over $6,500 dollars worth of retail this month!! That is $3250 in her pocket because once you hit $6,000 in a month you make HALF the profit!! So that right there proves how much product people are willing to buy.  I am just going to run through the basics with you:  First off, I make 40% commission on all the merchandise I sell. Most of my parties average $500-$800 so I make around $200-$320 per party and I have about 5 parties a month. So I take home $1000-$1600 per month! My parties are around 4 hours with drive time factored in so that equals $50-$80 dollars per hour just from parties!! Think about this, I get paid to party and talk about sex!  I also make a commission on what the people who have signed up under me sell! Eventually, I will be making 5% commission on every girl under me, every girl under them and every girl under them!!! THAT is where the REAL money starts to flow in. For example:  You sign 15 girls up; they each sign 2 girls up; and they each sign 2 girls up. You then have a total of 105 girls in your \"down line\" Lets say they each have one party in a month that has total sales of $500. Since you would make 5% off each of their sales and their sales together are $84,000 you would get a check for $4,200!!! Thats per month!   Now that may not seem like a lot but take into consideration that MOST parties are $500 dollars in sales if not a lot more and women usually do more than one party a month. So letâ€™s look at their next month:  They each had 3 parties (still low) averaging $500 per party. Now their total sales are $157,500 and your 5% check is now worth $7,875!!! Remember, that is not including the money you actually WORKED for at YOUR parties!! So, so far in two months you have made $2,000-$3200 on your parties and $12,075 on your down lines sales, for a total of $14,075-$15,275!!! Pretty good deal huh?  To get the most out of your business you want to invest 10 hours a week for marketing and parties so about 40 hours a month. Divide that all out and you are making $351.87-$381.87 dollars PER HOUR!!!!!   So, the real question is, since I have shown you exactly how much money can be made in this business are you ready to start now or do you have further questions?  If you know now that this is the business for you, just go to my website and fill out an online application and we can go from there!   I can\'t wait to get you started!! Have a passionate day!  \"Passion Parties is the place to be!\" http://whatsyourpassion.yourpassionconsultant.com '),(7385,4188004,'09,August,2004','2004-08-09 00:00:00','    well, dang, its been a while since i wrote... so yea. im going to write! so, this weekend was RenFaire. it was fun. i had a good time. besides the throwing of ice cubes down my shirt and the water, and almost getting thrown in a big metal bin of ice cold water while wearing a white shirt, it was great. i actually had a lot of fun. last weekend was Rainier. that was great fun. i had a really good time. we only took the A and a couple of blankets. oh, and some cheese and summer sausage. it was good. i got to have a nice long chat with justin. my mom didnt go, and my dad left sat night, so i was there all by my onesie. it was pretty great. i had a good time. other than that, i am a mean, messy person supposedly. and now, im just waiting for the day that justin says oh well b***h, i cant stand you anymore, get the h*ll away from me and never talk to me again. i mean, chris screwed with my mind so much, i cant even trust the one person i truly care about now. and justin doesnt even know the half of it. i cant trust him because of chris, and i dont want to tell him because i dont want to know what he is going to say about it. i dont think i would be able to breathe without having justin in my life, but then again, if i am so screwed up that no matter what i try i cant trust the one person i actually care about besides my family... since they are family and all.... but yea.... so i am freaking out. i dont want things to go south, but then again, its almost like inadvertantly i am asking them to. its crazy. i mean, for a few months i was told how i need to lose weight and blah blah blah. and then, im so ugly that he would rather f**k my brother then me. i mean, dang, if you feel that way, whyd you ask me out in the first place you f****r??? i mean, sh*t. thats crap. whatever. and then come to find out, he had another girlfriend the whole time?? WTF is up with that?!? whatever. and they were engaged like less than a week or something like that after we broke up. d*mn thats crap. if they wanna play games, then f**k them. i couldnt even go out of my house without fear that i was gonna die. what kind of crap is that??? then, thank god, he left, and then EMily had to keep bringing it up! WTF like just to rub it in or something... i dont know. whatever. he can go f**k his GF whenever he wants, as long as i never have to talk to him again. i mean, he followed me around for 2 months if i ever went into town and he saw me. i was deathly afraid for my life. what kind of crap is that?? then he tried to get a job in the mall?? WTH is his problem?? thank god he is gone! i mean, i look at justin, and i beleive what he says, but at the same time... i have some little voice in the back of my head saying \"hes a guy, dont trust him. youre just gonna get f***ed over again! dont beleive it!\" i mean, i dont thik he would do half the crap if any of what chris did to me, but still. i cant even trust aynone without thinking oh, is he gonna pull the same shit chris did?? dang, ive rambled on forever... but i guess it has to come out sometime. its the end of the process.. and then that bastard thinks he can go to my friends and parae sh*t in front of them and they are going to run to me??? i dont think so jerk. now i think youre stupid. i dont know what i saw in you in the first place.. oh wait, i didnt see anything, i was just looking for a way to get out of my house for the summer! i hated you even when we were going out you a$$hole! so guess what... go f*ck your boyfriend in virginia or wherever it is you are... and stay the h*ll away from me and my friends! i am quite happy where i am so leave me the hell alone! and to my dear friend who never reads this that keeps bringing him up... DO NOT TELL ME ANY MORE SHIT ABOUT HIM! I DONT WANT TO HEAR IT! QUIT OPENING UP OLD SHIT I DONT WANT BROUGHT UP! ITS NOT ANYTHING THAT I EVER WANT TO HEAR AGAIN! HES BAD NEW S FOR ANYONE HE CROSSES... SO LEAVE ME ALONE ABOUT HIM!!! OH GOD I THINK I AM GOING CRAZY!!! SOMEONE SHOOT ME AND GET IT OVER WITH! I dont think anyone needs to go through this shit. not at my hands. what if i just died?? would that solve me problems?? or in my next life.. .would i have the same ones??? probably... who knows. i mean, i cant do anything without thinking, is this some huge game justins playing on me if it has anything to do with him. i dont know. i really hope not... i dont want to get played again. it was bad enough the first time. if i get screwed over this time, i will wind up in a funny farm. litterally... whatever. i think i have ranted enough. there is no reason to poison my already black heart anymore talking about that a$$hole and my problems. whatever. i think i am crazy. how can someone live their life like me??? i mean, i am officially underweight now... and i love it. i think im sexy now. i mean, before, i told people like a typical model that i am 105... but, i was back up to like 115. now, i jumped on a scale this morning... and im down to 106!! i am so happy, even though i know i shouldnt be. justin is going to kill when he reads this. im going to need a million years of therapy for all this sh*t. whatever. i am done. justin, i love you, even if i cant trust you yet... i honestly am trying. if my mom finds out any of this, i will kill everyone i know who has the link to this site. got it??? good. '),(7386,669116,'01,June,2002','2002-06-01 00:00:00','    Whew, I\'ve just spent my first day in Colorado after exactly 40 hours on the train...and I am happy.  A few words to the wise about the train...    1)  Don\'t take the train if you feel you have a schedule to keep.  The train will NEVER arrive ANYWHERE on time.  It is not for people who want to hurry up and get there so they can start having fun.  I don\'t know anybody like that, so I think we\'re in good shape, but just a precautionary note.    2)  Don\'t take the train if you like paying less than $15 dollars for a meal.  Or just bring your own food, because you will be broke whenever you get to wherever you\'e going if you eat on the train.    3)  Don\'t take the train if you have a problem seeing other peoples\' asses.  There is some kind of nation-wide mooning ritual known as the \"Amtrak Salute,\" and the land-bound just LOVE to introduce us unsuspecting passengers to it.      4)  Don\'t take the train if you are scared of meeting people.  They will put you at a table in the dining car with a bunch of strangers and you will be expected to deal.  Now, I LOVED this, but some are horrified at stuff like that, so that\'s just another warning.  Okay, so all that out of the way, let me just say that I had so much fun!!!!!!!!!   For anybody who\'s never been on the train, there are three basic places you can go.  You can stay at your seat like a nerd (but there are people to meet there too, so that\'s cool), you can go to the observation car where the seats face the windows and you can watch the world go by, or you can go to the dining car.  You can also go to the bathroom, of course, or the smoking car, or the snack bar/lounge downstairs from the observation car, but those are kind of secondary thoughts.  Going any combination of these places will undoubtably allow you to meet a whole bunch of cool people who you will most likely never see again, but it\'s an awesome way to spend the time.  Actually, I take that back because I met a woman on the train that had met her husband (of nine years) on the train, so how\'s that?  Just a few things I have to mention in case any readers happen to ever take the train as far as Reno (from the Bay Area).  In Sacramento the train picks up a couple of volunteers from the train museum there.  These volunteers do narration all the way through the Sierras.  They tell a lot of interesting stories and, overall, it\'s pretty cool (though kind of refreshing, too, when they get off the train in Reno and leave you to formulate your own thoughts about what you\'re seeing).  I was a little concerned, I have to say, when I heard the narrator say that the chinese workers that helped to build the railroads were paid as much as the white workers.  This was a total lie!  It is well known that there was a white wage, a black wage, a mexican wage, and a chinese wage, and you can all guess who got the most.  He DID mention that the chinese workers bought their own food, but it was made to sound like they did this by choice because surviving on rice alone improved their efficiency or something like that.  The other weird thing was, when we went through Donner Pass, he was telling the story of the Donner Party.  At the end of the tale, he said, \"in the end, of the 89 people who started out, only 47 survived the winter.\"  He cunningly left out any mention of a little thing know as  cannibalism , which I was admitedly impressed with.  The glossing-over was unmatched.  Do they think that we just couldn\'t handle it if we knew what really went on?  I mean, it\'s not like these things are secrets, and it just works to ruin their credibility.  This would be a great time to recommend James Loewen\'s   Lies My Teacher Told Me   and Howard Zinn\'s   A People\'s History of the United States  , both great myth-dispelling reads, and user-friendly to boot.  So, now that I\'m back on unmoving ground, I have to start thinking about a job and the fact of it is incredibly depressing to me.  How I would LOVE to just relax and talk to my parents and my dog and write and read and lie around in the sun.  Oh well, que sera sera.  Just one last note, cuz my pops wants to use the computer now...I read my horoscope today in the Denver weekly publication that\'s like San Jose\'s \"Metro.\"  These were the first couple of lines:  \"I think it\'s high time you dreamed up a few new vices.  The old ones barely tweak your guilty conscience any more, and they certainly don\'t pack the educational punch they once had.\"  So don\'t blame me when I take up crack-smoking and whatever else you can let your mind imagine.  :)   '),(7387,669116,'30,May,2002','2002-05-30 00:00:00','    Ok, so now I\'m (finally) about to embark on my trip to Colorado.  I\'m for the most part packed, ready for two days on the train, and just kind of tired.  I\'ve got lots on my mind too, so I\'m looking forward to the fresh Rocky Mountain air.  John Denver is just so rad!!!   He makes me feel happy.  And that\'s all I have to say about that...I\'ll be around in a couple of days...   '),(7388,669116,'28,May,2002','2002-05-28 00:00:00','    Would anybody believe me if I said that I\'m giving up on love?  You probably shouldn\'t, because I doubt it\'s true, and I know it sounds ridiculous at the age of 24, but jeez louise!!!  I know that everytime one decides that they are going to open themselves up to love, they also open themselves up to the pain that comes when the nature of that love relationship changes.  So, no room to bitch and moan, right?  Right.  Sometimes I wish that people would just be evil like they are in the movies, so we wouldn\'t feel so bad about breaking up with them.  But I don\'t believe in evil people...or at least I\'ve never met any, so life is not so easy, then, is it?  Does anybody have any comments on the idea that what somebody thinks of \"The Pina Colada\" song says a lot about their take on love?  If you think that it\'s romantic when they guy discovers that it was actually his own woman that placed the personal ad he was responding to, does that make you a sucker?  I mean, they WERE both gonna cheat on each other, right?  Yes, but, but, it\'s just so sweet, you know, how he   never knew   that she like Pina Coladas and all that stuff.  Maybe I AM a hopeless romantic despite the fact that I just threatened to swear off love 5 minutes ago.  Something weird happened to me yesterday.  I was walking down San Salvadore on my way to South First Street Billiards, and I (think I) was propositioned...you know, as in, as if I were a prostitute?  The weirdest thing is that it was the second time this has happened to me in as many months!!!  I was walking down the street and this man walking behind me said \"excuse me\" or something close to that (he didn\'t seem to speak much English).  I could not make out what he was saying, but I\'m pretty sure I heard the words \"price\" and \"pussy\" (!!!!!).  He had the most devilish grin on his face and seemed to really be enjoying my baffled reaction.  Now, I was wearing blue men\'s jeans and an long-sleeved blouse and I most definitely wasn\'t hanging around on any street corners or making eyes at anybody.  So I have no idea about all that.  The other time this happened to me I was waiting for the bus after work.  This car was pulling into the parking lot and on his way by the driver yelled, \"How much?\"  Wha?  That time I was wearing a dress and heels, so I   guess   I might be able to see how I would look like an unlikely bus-rider and how I might be suspected of harboring some sort of alternate motive, but c\'mon.  Can\'t a girl wear a dress and stand at a bus stop in peace these days?  I\'m really pretty troubled by these events.  I mean, I wear your occasional low-cut blouse, but prostitute?!  That\'s pretty insulting.  I\'m only sharing to help get over the pain.  :)   '),(7389,669116,'26,May,2002','2002-05-26 00:00:00','    Ok so, as promised, I spent five (count them, FIVE) hours in front of the T.V. today watching TLC.  It was a Memorial Day \"Trading Spaces\" marathon, and I got sucked in!  I justified this to myself by noting that I wasn\'t just watching T.V., but watching T.V. in between doing laundry and drinking beers, which were the   real   tasks at-hand.    For anybody that doesn\'t know (because I didn\'t before today), Trading Spaces is a show where two sets of neighbors get $1,000 and the help of a professional designer with which to set about completely re-making one room of their neighbor\'s house.  This happens in the span of one weekend, and the hour-long show condenses everything so we see only the most interesting parts and don\'t have to actually watch them do too much of the work.  Now, don\'t misunderstand me...I do not think that this show is so interesting that I just HAD to talk about it here.  It\'s just that there was one episode that got me thinking about something that I think is really cool.  The show travels all over the country to find these neighbors.  They aren\'t necessarily married couples, but they often are unless they\'re mothers and daughters or something like that.  So anyway, there was one show on the marathon today that took place in Oakland.  The first set of neighbors was a married couple, an African-American woman and a white man.  Their neighbors were a lesbian couple with two small children.  Watching this (and knowing that people all over the country were watching it) made me feel strangely (and immensely) proud.  I mean, there are parts of this country where either one of those couples would be absolutely unacceptable.  And here, in California, they are neighbors and they are friends.    There is a line in the movie L.A. Story (one of my favorite movies of all time) where the English reporter makes a comment about how the people in L.A. seem to have their own set of rules, their own little world, really, and that nobody is looking to the outside for approval or reassurance that what they\'re doing is okay.  I think that is true of California, in general, which is why it would be so hard to ever leave here.  When you grow up here, you take things for granted...diversity, open-mindedness, liberal political climate, sunny weather.    I remember flying into Cleveland once to visit my parents when they were living there.  I was walking through the airport and I started to get this really weird feeling.  I was racking my brain, trying to figure out what it was that was making things seems out of place.  At the end of a long walk to the end of terminal and toward the baggage claim, it dawned on me.  There were so many WHITE people!!!!  Everywhere I turned, white people in all directions.  Now, this is not to criticize the white people, or to say that they shouldn\'t be there, all congregated together in Ohio like that.  It\'s just to say that when a California native is suddenly thrust into the midst of a homogenous scene, things feel a little off.  I get this feeling a bit when I go to visit my parents in Colorado, too, where they now live.  When I mention it, they swear to me that the Hispanic population there is growing in leaps and bounds.  I suppose I believe them, but I think I\'d have to let it grow for a few more years before I\'d really feel comfortable moving there (which is something I\'ve considered).  I would never want to do the disservice of raising my future children in a place where they don\'t get to experience the kind of diversity that I grew up with.  And I don\'t have any illusions about this...it\'s not that I think that California doesn\'t have racism problems or that hate crimes are less frequent here.  It\'s just that I would want to give my kids a fighting chance at being able to look around them and see people for who they are, with less overt separations.  Ya\'ll dig?  Oh well, it\'s late, and though I\'m not exactly tired, I\'m not exactly functioning all that well either.  I think I\'ll call it a night.   '),(7390,669116,'26,May,2002','2002-05-26 00:00:00','    The end of day of a melancholy Day One of single life has me feeling grateful for my sweet roommate Lisa.  It\'s not just her spunky disposition that brings me a bit of comic relief...but she made me an absolutely awesome mixed tape when she thought I was about to embark on a five-day road trip - one of those rare instances where the tape-giver doesn\'t just give you a tape full of songs they like, but one full of songs they think that  you\'d  like.  So I\'m sitting here listening to Diana Krall and the bongo wonders of Perez Prado, Dead Can Dance, etc., etc, stuff I dig.  Now would be a good time to mention the CD that my friend JD burned for me the other day as well.  Man oh man...we had been talking early 80\'s country and the fact that the Oak Ridge Boys and Alabama and the Nitty Gritty Dirt Band just kicked the shit out of anything Nashville has to offer these days (and both of our Mom\'s seem to harbor some morbid fascination with Anne Murray)...so my CD came complete with Kenny Roger\'s \"The Gambler\" (that  is  Kenny Rogers, right?), and John Anderson\'s \"Swingin\',\" a song that, for some reason has been burned into my memory as a piece of sheer brilliance since the tender age of five.  Go figure.  I\'m sitting on my bed wondering what on earth to do with myself given the fact that I have no finals to study for (not that I would have been studying even if I did), no job to go to, an aching feeling in the back of my eyes that prevents reading, and can barely stifle a dry heave at the thought of watching T.V. on a Sunday night (Simpsons being over and all).  This is all I wish for (free time) during the nine months of school, and now I can\'t figure out which way to go.  Whatta dork.    I went to visit my friend Dean today, and I\'ve finally found somebody who agrees with my claim that sci-fi paperbacks are men\'s answer to the romance novel.  Since the fact that he works at a bookstore allowed for such a luxury, he went to the trouble of finding a sci-fi novel and a romance novel that had nearly identical cover art - each with a warrior-looking man holding a heaving-breasted damsel in distress in his arms...or something like that.  Awesome.    I\'m trying to prepare for my big train ride out to Colorado, but I can\'t seem to get motivated.  It probably has something to do with the fact that some of the things I have to do are to go to my storage facility, the bank and the post office...none of which can be done on Sunday (or Memorial Day, for that matter).  So I think tomorrow I\'ll abuse myself with TLC programming...A Makeover Story, A Dating Story, A Wedding Story, in that order.  Or maybe I\'ll just take a long long, long long drive, and think about all the reasons why life is wonderful while listening to my groovy new musical gifts.  I\'m hoping to wake up with motivation and joy...wish me luck.   '),(7391,669116,'26,May,2002','2002-05-26 00:00:00','    Hmm, trying something new in the form of an online journal.  I was inspired by my friends Murphy and Dean, whose rambling thoughts I have saved to my favorite places and whose sites I visit when I feel like getting out of my own head (for once, dangnabit!).  I thought, at the very least, this could be some summer fun, and a chance for me to be expressive in a way that is more accountable than the likes of my (for-my-eyes-only) journal, and yet less accountable than the for-grade work I turned in to Maio\'s creative writing class.  I hope you\'ll enjoy, as well as forgive any spelling and grammatical errors.  I never said I was an English major because I was interested in  that  sort of thing.  I\'ll try to form nice little bite-sized morsel-like paragraphs for any of those personality types who might be reading this (and you know who you are), but I\'m not making any promises.  The sun is shining, I think I\'ll go splendor in it for awhile.   '),(7392,669116,'07,June,2002','2002-06-07 00:00:00','    What a fun day!!!!  I woke up this morning without a care in the world (despite the fact that I owe a few different credit card companies a whole bunch of money and I still have no job to speak of).  That will probably change soon...I\'ve decided to go back to my old overnight stocking position at Target.  Mindless work is just the way to go sometimes.    It\'s been a Father/Daughter afternoon.  Pops came home early from work...we were gonna go to some boring banquet for something rather (I was supposed to step in for the Moms, who left for California this morning for a weekend getaway), but I convinced him that we should stay home and watch the Laker game instead (it wasn\'t too hard), so I\'m killing time in anxious anticipation.  We decided to go to Boston Market for dinner (because my Ma has a problem with mass-produced chicken products, and this was our chance to grub on some for a night), and my Dad did this thing there that was SO typical of my Dad, and cute in that way that it can only be when you\'re no longer in middle school and embarrassed by this sort of thing.  We were ordering, and I ordered this meal that comes with chicken and two sides.  My Dad ordered chicken and  three  sides, and when he saw the plate the guy was gonna put it on, he goes, \"when you order three sides are the plates just divided into smaller sections?\"  (Pops LOVES food!  :)  He said it in a really serious way, cuz he was, and he sounded just a slight bit upset.  The kid who was shlepping the grub looked confused, and my Dad added, \"because they looked like the same sized plate.\"  The funny thing is that I was thinking the exact same thing, I just don\'t think I would have said anything about it.  I was asking myself what my Dad was expecting to happen from the exchange (like, the kid couldn\'t possibly change the size of the plate right then), and then the kid said, \"I could give you more of everything,\" in this way that sounded like he was totally scared.  I forget sometimes that my Dad is a 6\'4\" 220-pound intimidating figure to those who don\'t know whatta gentle goof he really is.  So I guess he got what he wanted out of the deal.  The kid rung us up and asked if we wanted drinks in this way like he was expecting my Dad to reach over the counter and grab him by the collar, screaming that \"NO, we don\'t want any damned drinks!!!!!\"  But we totally wanted drinks, and we just smiled politely.  Sometimes it\'s nice to eat with men because they don\'t mind if you\'re more interested in your chicken than conversation.  Don\'t get me wrong, I love talking and joking with my Dad, but it\'s a very low-pressure sort of relationship.  And to explain the chicken part...I finally realized why I\'ve always liked that \"Swingin\'\" song so much (the one I mentioned about two weeks ago).  It\'s because of the line where John Anderson says, \"Mama\'s in the kitchen cuttin\' chicken up to fry.\"  When I was a little girl, fried chicked was my favorite food in the whole world.  Every birthday I wanted to go to this restaurant near Anaheim, California called the Big Yellow House.  It was a chicken place made out of this old victorian-style house, and I just LOVED it.  There\'s a picture of me with a big plate FULL of chicken bones sitting in front of me on my 5th birthday.  Of course, everybody had been using the plate, but it was a funny picture.  Come to think about it, my parents really seemed to dig on that kind of humor.  My Dad used to always fall asleep watching T.V. in the evenings (still does, actually).  One time, he was lying on the floor on his back in the living room, and my Mom took all these Twinkies out of their wrappers and put the wrappers on his chest and all over the floor around him and then took a picture.  It was like he passed out cuz of all the Twinkies or something.  Those wacky Konrads!  So anyway, I think that when I was little and heard the part about the Mama cuttin\' up the chicken, I just let my mind wander and imagined that I would get in on some of that.  Cheap cheap, cheap cheap thrills.  Anyway, it\'s about time for the game, and my Dadio\'s saving a spot for me on the couch.  Go Lakers!!!   '),(7393,669116,'05,June,2002','2002-06-05 00:00:00','    Mmmm, remember PeeWee\'s Playhouse?  Of course you do.  Well, I\'ve got a word for the day.  Today\'s word is........drumroll....... LOVE  (I hope you\'re all screaming).  Why would I pick such a HUGE little word?  Because I feel like it, and because I\'m presently trying to understand some previously misunderstood facets of this HUGE little word.  My thoughts and my desire to understand were prompted by an article I read this afternoon titled, \"Is There a Better Way to Love?\"  I didn\'t plan on getting much out of this article, seeing as I glanced ahead and saw that it was only three pages long (what can you say in three pages?!), but I decided to go ahead and give it a chance to change my way of thinking (you can say a lot in three pages).  The article was written by a man who was interviewing the Buddhist teacher Joseph Goldstein, founder if the Insight Meditation Society.  What he has to offer on the subject is the idea that love and attachment are two separate concepts completely, which cannot possibly co-exist.  Goldstein asserts that love is made of a giving energy, and that attachment is based in the act of holding.  He says, \"We\'ve been bamboozled into thinking that attachment gives us something, provides some security for love.  But it doesn\'t enhance love in any way.  How could holding ever enhance giving?  Instead of creating a relationship of beauty, attachment creates a bond of fear.\"   Ok, this is something that I\'ve been hearing about for years and years, but I can honestly say that it\'s never sunk in until recently.  I think I finally understand the idea of \"if you love something, let it go.\"  I mean, I understood it in the literal sense...if you love your girlfriend, and she goes far far away for college, but she comes back in four years and wants to be together again, well, then that certaining seems like love (if it\'s meant to be, it\'ll come back to you).  The part that I think I hadn\'t gotten until now, is that this concept is true even if the test is not letting somebody go in the physical sense.  You have to let go even as the person is sitting right beside you.  I am reminded of something that somebody once said to me about love.  They said, \"I don\'t want to be with somebody because I have to, because I don\'t think I could live without them.  In fact, I want to know that I could live without them, that way I\'ll know we\'re together simply because we lift each other up...not because we would sink without each other.\"  Ok, maybe it wasn\'t said in that many words, but the sentiment remains unchanged, and I think the statement really coincides with the article.  I want to relax in love.  I don\'t want to be jealous, scared, or coast through things with my defenses up.  I think that\'s a BIG part of the problems that people experience in relationships...defenses.  When I say coast, it\'s because I think that having defenses up keeps people from experiencing the depths of love, it\'s like coasting over the surface of the water instead of just diving in.  And I mean love in every area.  Wouldn\'t it be nice if we could just see our loved ones, feel them and what it\'s like to love and be loved by them, and then just relax instead of trying to control every last little thing?  I think so...I\'m gonna try letting go a little bit, see where that takes me.   '),(7394,669116,'04,June,2002','2002-06-04 00:00:00','    Ok, I don\'t know why I do this...I guess because I like to torture myself...but I watched Politically Incorrect last night.  For anybody who\'s never seen it, it is a show where four (sometimes qualified) guests join sarcastic and (sometimes) informed Bill Maher for a half hour of debate about politics and current affairs.  Though I hadn\'t seen the show in a long time, it has stayed true to format, usually devolving into a (s)he-who-shouts-the-loudest competition, with the unimaginative audience going along with whoever that loud person might turn out to be.  Last night, David Crosby (musician now billing himself as \"political activist\"), and Kennedy (former MTV V-Jay now billing herself as \"author\") joined a really smart female author from England and some republican strategist who was going bald WAY too early (you decide if the two descriptions of him are in any way related).  Now, it\'s a shame that I didn\'t catch the English woman\'s name, because she was by far the most educated, most articulate, and most sensible person on the show (in my book).  Anyway, there were a few debates that ensued that I wanted to comment on.  First, the topic for the majority of the show was nuclear (or, \"nucular,\" if you\'re the president) power and what we can do to keep Pakistan and India from acting on their threats to use nuclear power against each other and, in the process, blowing us all to smithereens (Don\'t you just love that phrase?  What does it MEAN?!).  Anyway, at first I thought that Crosby was gonna be cool, not only because he made really good music once upon a time, but because, after all these years, he still hadn\'t cut his hippy hair (sound like anybody you know?).  When the subject was transporting America\'s nuclear waste from the 130 separate sites where it is presently housed to the middle of one big mountain (which I think is in Nevada), he had the sensible answer...he and the British woman both thought it ridiculous, given the accident rate of any moving vehicles on America\'s roads, and considering the little matter of the mountain resting on, like, 23 different fault lines.  Ok, points there Davy.    Things got horrible, however, when the British woman brought up the idea of peace, and started questioning why the hell we have to go around KILLING each other, in that silly way that women do.  She brought up the 5,000 civilian casualties that we caused in Afghanistan when we decided to flex our muscles and show the Taliban who was boss.  She was troubled by the fact that nobody was talking about this, and hinting at the idea that the U.S. Army was, perhaps, a bit hasty.  My thought, as she was saying this, is that 5,000 is MORE than died in the World Trade Center.  I mean, you can\'t compare numbers, really, when it comes to innocent human lives lost, but it is shocking to me that it is easy for so many Americans to just blow it off, because, hey, they killed our people FIRST.  Kennedy chimed in with, \"oh don\'t worry, those 5,000 were just orphans and old ladies.\"  It would have been funny, if it weren\'t so fucking unfunny.  Of course, in response to the woman\'s comments, Bill Maher made fun of her and accused her of being idealistic, not to mention unrealistic and, dammit, British.  To my shock and dismay, Crosby was right there with Bill.  He was laughing at the lady and when Bill said, \"People say that violence doesn\'t solve anything, while I\'d say that violence solves just about EVERYthing,\" (with an annoying little smirk on his face), Crosby joined in with, \"Yeah, and there is a good kind of violence, which is the kind that we use.\"  Yeah, I\'m sure that\'s not EXACTLY what those terrorists were thinking when they flew planes into buildings in the name of their faith.  Can\'t they SEE that it\'s all violence, no matter what side of the line you\'re born on and what religion you practice and what you look like and blah blah blah??!!!  At one point, I just got really, really mad.  I thought...men (and I am referring to men, specifically) just keep dreaming up new and more efficient ways of killing each other, and WHY???!!!  Now we not only have our own \"enemies\" to worry about, but we have to worry about Pakistan and India and their conflicts and the fact that they both have absolutely NO problem dropping the first bomb/launching the first missile, anything you can imagine.  But you know what, as mad as I got, I didn\'t want to go out and KILL anybody.  I\'m going to sound like a male-basher, but that\'s not it...goodness knows I love men.  I\'m not saying that all men are violent and want to solve problems by killing people.  But honestly, how often to women go around harming each other, relatively?  And would we ever  dream  of developing the most deadly weapons that human beings are capable of creating, in order to ensure our place at the head of the proverbial table?  I don\'t think so.  So what I say to all you weapon-developing men out there is...you got us into this, now you get us out.  Some advice might be needed from the women\'s camp, so I\'m offering this...why don\'t you try TALKING to each other.  I know you\'ll say I\'m also idealistic.  And I am.   '),(7395,669116,'02,June,2002','2002-06-02 00:00:00','    Mmm, and the Lakers are going to the Championships again.  Knowing this helps me get over the fact that it seems Rick Fox will never grow his beautiful curls back  :(  Oh well.  Nothing too ground-breaking has happened to me today.  I haven\'t started looking for a job, and I didn\'t even take a bath until about noon.  I probably wouldn\'t have gotten dressed at all if it weren\'t for the fact that I went out to dinner with my parents and an acquaintance of theirs.  Yes yes, I know it\'s summer and all, but the whole point of me being out here is so that I can make some money and pay of some bills.  Somehow, though, time spent with my Ma sitting on the couch and talking about boys seems so much more important for the time being.  I promise I\'ll start looking tomorrow.  I was noticing something horrible at dinner.  We went to this restaurant that serves a hunk of bread to each table of people while they\'re waiting for food.  They serve it with this yummy cinnamon butter that is just awesome but I\'m sure you can guess that I haven\'t gotten to the horrible part yet.  I noticed that all these people were leaving and there was a ton of bread left on the table just waiting to be thrown away.  My friend was mentioning the other day how important bread is in Russia, and how having bread at a meal (or just bread  as  a meal) is a mark of survival.  I\'m sure this is true in many parts of the world.  And then here we go, in typical American fashion, just throwing it away.  I know I sound like a Mom...\"Starving children in China\" and the whole bit.  But you know what?  Sometimes our Moms are just right.   '),(7396,669116,'14,June,2002','2002-06-14 00:00:00','    Aha!  And my wish for mindless and menial work at Target is realized!!!  I\'m a happy little clam these days.  Boy, those clams sure ARE happy, aren\'t they?  I was asked to come in last night at 10:00 for orientation.  It was all going so well...the loss protection lady cut her normal 11/2 presentation down to 10 minutes just because she knows that, moreso than any other shift, overnight workers don\'t give a CRAP.  Plus, I think she just wanted to get home.  My manager cut back his presentation as well, and the whole thing looked like it would be wrapped up in about two hours.  Then the whole shift went on break, so he made the four of us new-hires go too, and after that they had this cheesy \"huddle\" thing where everybody gathered and listened to all the new info.  The manager took that opportunity to introduce my cohorts and I to the crowd (about 50 people), made me wave and everything.  I couldn\'t believe how many people still worked there from before (3 1/2 years ago)!  The difference is that now there are a whole bunch of Hispanic workers there, so all of the things mentioned in the huddle were also translated and the whole thing took twice as long.  It was awesome, though.  I\'m so excited about having the chance to speak Spanish everynight.  At one point I had to translate a video about bloodbourne pathogens to the two Spanish-speaking new-hires.  Talk about challenges!  Anyway, after the huddle, I was waiting outside this room to get my ID picture taken, and I heard somebody say \"Hi Kisa.\"  I turned to see this guy named Brian who used to be the manager of the shift when I worked there before, but who now apparently just changes displays or something.  Anyway, the funny thing to note about this guy is that when I worked there before I had the HUGEST crush on him.  It used to drive me crazy, couldn\'t sleep, all that jazz.  So, like the smarty I am, on my last day, I decided I needed to TELL him about the crush, made a total fool out of myself getting all red-faced and stuff.  He was like, wondering why I chose my last day to mention it, all this...but he never gave me the impression that he returned the feelings.  No big deal, I was about to leave for sunny California!  I got over it.  So now, 3 1/2 years later and the guy has to STILL be working there.  When I saw him yesterday, I was thinking \'what on earth did I ever see in this person?!\'  He is one of those people (and as I hinted at before, the majority of the night crew is like this) who just seems to hate life or something.  Super quite, super disillusioned, smoker, plays paintball for kicks, has shaved head - looks like a skin head, has worked at the same job he can\'t stand for over five years.  Sounds dreamy, right?  Then I realized that he was actually just my type, for the time when I was interested in him.  I don\'t understand this...I LOVE life, I\'m not disillusioned, I hate to sit in corners and scowl and be lazy and I don\'t fancy the idea of pelting my friends with little balls of paint.  Opposites attract, I guess they say.  But I\'m done with all that.  I used to be attracted to the shy boys because I thought they must have some really interesting thoughts going on in their heads.  It took me dating three of them before I realized that they\'re really NOT thinking anything interesting at all, which is why they aren\'t saying anything.  Ok, so maybe that\'s a harsh stereotype.  I\'m sure that there are plenty of interesting shy boys out there (I know quite a few interesting shy girls), but I just don\'t want to date them.  People who wear their emotions on their sleeve...that\'s what I like.  No guessing, no always wondering if they\'re REALLY having fun or not.  Yes, I lost interest in the shy boys a while ago.    So now I\'m faced with the task of trying to work with somebody whom you used to like, and behaving in a way that makes it clear that you are no longer interested.  Cuz don\'t you just HATE when you know that somebody thinks that you like them and they\'re like trying to avoid you and stuff and you want to tell them (or convince them) that you don\'t really like them but you know that if you do it\'s gonna seem like you really DO like them?!  That sucks.  The store\'s pretty big, though, I should have no problem avoiding contact.  So, after that run-in, we had to go back for more training...this STUPID interractive CD-Rom about safety that lasted an entire hour!!!  There was this little robot that you had to make do all this safe stuff like mop up spills, help a guy who had fallen and whose head a huge box of detergent had landed on (picture this in cheesy, primary-colored, low-tech graphics), and put out a fire that some little kid had started in the automotive aisle.  The thing took way too long and was so, so boring and sleep-inducing at 2:00 in the morning.  The best part was when they introduced the codes for the intercom...code yellow = missing child, code red = fire, code green = somebody hurt.  The robot would get on the phone and say three times \"code yellow, garden, code yellow, garden, code yellow, garden\" in this total robot voice that I decided I\'m gonna use if I ever have to make such an announcement.  No influxion, just the robot facts.  Then they\'ll know just how well I was paying attention during my orientation.  I hate when execs approach techies and ask them to develop interesting training tools about subjects that could never possibly be interesting.  A whole HOUR of that stupid robot!!!!!!!  Anyway, it WAS pretty funny, I got a good kick out of the ridiculousness.  I\'ve gotta get ready for work now.  I promise I won\'t talk about work all the time...unless something really exciting like a code green happens.   '),(7397,669116,'12,June,2002','2002-06-12 00:00:00','    If any of you readers don\'t believe that romantic love can last, I would like to offer Exhibit A...My Parents, who are an extraordinary example to the contrary.  Today my parents celebrated their 26th Wedding Anniversary, and I have to say that they are, by far, the most romantic couple I know.  They went out yesterday night because they wanted to go the the Botanical Gardens, which aren\'t open today, and planned on having a picnic there.  The undying sweetie, my Dad planned for some friends of his to surprise my Mom and asked them to set up a beautiful spread of decadent food and dessert, flowers, and candles, which were waiting for them when they got there.  They spent three hours in the gardens, sitting among the flowers and talking.  And that is the thing that is amazing to me about my parents...after 26 years, they never run out of things to talk about.  My Mom has told me that sometimes they will go to bed, but before they know it, it\'s like 2:00 in the morning and they\'re still awake, talking to each other.  As I mentioned, my Mom was out of town last weekend.  On the second day she was gone, I noticed a little pad of paper sitting on the kitchen counter that had my Dad\'s handwriting on it.  I started reading down what seemed like a list of the day\'s events, and I figured out that he was keeping a list of things he wanted to remember to tell my Mom about when she called that night.  I teased him about it, and he said it was true...he wanted her to know everything, didn\'t want to leave anything out.  How cute is THAT???!!!!  My parents still hold hands when they walk around the mall or go to outdoor festivals.  My Dad still opens car doors for my Mom.  They are there in support of each other whenever one of them gets involved in a project or organization that is important to them.  They still listen attentively to each other and laugh in all the funny places when one of them is telling a story to a new acquaintance....even though they\'ve each heard each others\' stories hundreds of times before.  My parents still sneak a card in the suitcase, when one of them is going out of town.  My Mom still sports a smile at my Dad\'s corny jokes (mostly), and my Dad loves nothing more than to watch my Mom when she\'s laughing so hard she\'s about to pee.  They are each others\' best friends, and they call each other four times a day while they\'re apart.  When I\'ve reached uninspiring places in past relationships, I used to try to tell myself that that was just the normal progression...that love fades and that sooner or later, people get realistic.  My parents just keep on shattering this self-lie for me.  Don\'t get me wrong...it\'s not like they haven\'t had challenges in their marriage.  The difference is that they saw them as just that...not endings, but new opportunities for their love to grow.  They are each others\' biggest fans, each others\' eternal support and intimate friendship.  When they look at each other, there is still a spark in their eyes.  My parents have taught me many, many things over the years.  The most important thing they\'ve taught me, however, is about love, because they know how to be in love better than anybody I\'ve ever known, and I thank them for it.  Happy Anniversary, Ma and Pa.   '),(7398,669116,'11,June,2002','2002-06-11 00:00:00','    Mmmm...peeing into cups.  What a strange thing.  Actually, the peeing into the cup part is not the weird thing.  It\'s the carrying the cup out of the bathroom, handing it to someone else, and watching them pour it into a smaller cup that is kind of goofy.    I had to take a drug test to work at Target.  Can you believe that?  I didn\'t have to do that when I worked as the stranger danger talk girl, going into different elementary and middle schools everyday, left to  influence youth !!!!.  But Target, evidentally, is real concerned.  Oh well, it\'s kind of a nice feeling to take a test that you already KNOW you\'re gonna pass...kind of reassuring.  They\'ll get the results in a couple of days, and then I\'ll have my old, manual labor job back.  Lifting and climbing and opening and stocking and throwing stuff down from high places from 10:00pm to 6:00am.  I can\'t think of any sort of work I\'d rather be doing right now.  I have made a startling and disturbing discovery during my time here, and that is that COLORADO DRIVERS ARE THE  WORST  DRIVERS EVER!!!!!!  First I thought that Arizona drivers were bad...and they were, especially the ones in golf carts.  Then I thought Cleveland drivers were bad...and they were, nobody could cut you off and then slow down with greater skill than they could.  But Colorado drivers, it\'s like they have all the horrible elements rolled into one.  They are slow, but that\'s okay by me...I\'m not usually in a hurry to get anywhere.  They do the cutting-off-slowing-down thing.  Yeah, that\'s bad.  They have no clue how to merge, and THAT is really annoying.  But the worst part is the red-light running.  I swear it happens at every single stop light, every single time.  It occurred to me the other day, that the reason they do this is because they can.  I haven\'t seen a police car in the entire two weeks that I\'ve been here, and that is just so WEIRD for somebody coming from San Jose.  It lends weight to the theory that people will do what they can get away with.  And that\'s incredibly depressing to me because I hate that theory!!!  I always wanted to believe that people do what\'s right because it\'s the right thing to do, not because they are afraid of the fallout it they DON\'T do it.  Poopie poopie!!!  People from other parts of the country are horrified by California drivers (which I think comes mainly from the fact that California drivers drive so  fast  )  It is my contention, however, that while California drivers are assholes, they are  considerate  assholes, if that makes any sense.  I mean, they will cut you off, yes they will.  But California drivers KNOW how to cut you off.  They pull in front of you and then they get the hell out of Dodge.  And they DEFINITELY are the best mergers I have seen in all my travels.  Plus, they can parallel park, they signal with more frequency than others (believe it or not), and they hardly ever dare to drive in the carpool lane unless they are qualified to do so.  My parents say that it\'s getting worse out here, and I hope it doesn\'t continue on that trend, but only time will tell.  All the wildfires are still burning, and the skies are still pretty yucky-looking, but apparently the winds are working in the favor of the fire fighters now.  They discovered that the largest fire (which is only 5% contained!) was started by an illegal campfire.  That is really hard to even think about, so I won\'t.  Anyway, aside from peeing in cups, bad drivers, and uncontrollable wildfires, things are peachy keen here in Colorado.  I\'m employed, I\'m well-rested, I\'m in the company of my sweet little dog and my swell parents, and I STILL have all my limbs!!!!!  I\'m really excited about that last one.   '),(7399,669116,'09,June,2002','2002-06-09 00:00:00','    Last night I dreamed I was babysitting Nicole\'s baby.  For any of you who don\'t know Nicole, she\'s my best friend (whom I\'ve know since eigth grade), and she\'s really, really pregnant.  It was a cool dream, complete with the sensation of holding the baby against my shoulder, and the smell of baby head.  In my dream, the baby was really chubby, which was surprising because Nicole is tiny, and I don\'t think the baby will be heavier than six pounds in real life.  Anyway, it was SO awesome, and I can\'t wait to be an Aunt!!!  Today\'s kind of a sad day in Colorado.  It seems that half the state is ablaze, and sooty clouds of air have moved in over the entire valley (referred to as the \"Front Range), which makes it seem like a giant storm is brewing.  It doesn\'t help that everything is making me want to cry today.  I\'m not sure why that is, I think I just miss the daily human interraction I left behind in San Jose.  I know it\'s good for me to get away and do some soul-searching, but it\'s just not really my style, makes me feel naked and abandoned.  The good news, however, is that I went for a drive yesterday and found a beautiful park with a view of mountains and jutting rocks just perfect for figuring your life out among.  I walked along the trails and sat down for a while and just watched and listened.  Have you ever just sat, closed your eyes, and tried to pick out every single separate sound that you could hear?  It\'s incredibly meditative.  I had a nice, relaxing run of it until I opened my eyes and realized that the entire ground around my booty was moving.  Ants in my sandals and on my feet, I decided to get up and walk again.  If any of you want to come visit me in Colorado, I\'ll take you there (with a blanket).  You will not be disappointed.  Yesterday my Dad rented a Harley and went out riding with some of his friends from work.  When he got back he and I went for our own ride, and man oh man, it made me miss riding!!  This morning we went out again, and as we were about to take this little curve, my Dad broke harder than was necessary for what seemed like no reason whatsoever, and we headed into the curve really slowly.  As we were rounding the curve, a HUGE-ass dog walked into the road, and my Dad had to go into the other lane to avoid hitting it.  When we were pulling away, he said, \"I don\'t know why I broke like that, I guess I just sensed something.\"  More proof for my theory that riding motorcycles puts one in-touch with their intuition.  It\'s something about the open air that gets your mind wandering into all these places that you didn\'t even know existed, and you are all of a sudden the smartest person you know.  I was riding on the back of my ex-boyfriend\'s bike when it occurred to me that he was a lying meanie.  I was on the back of his bike when I realized that he was cheating on me.  And I was riding a bike that I myself had rented when I figured out who he was cheating on me with.  I guess, in a way, I sensed a big-ass dog of my own.  But anyway, riding is like that.  Riding around in the hills yesterday made me think of the trip that he and I took out here (from San Jose) a few years ago.  Thinking about it now, I can\'t believe we made it all this way.  His starter broke in Yosemite (the very first day) and from there on out we were parking on hills and push-starting it, and leaving the engine running when we got gas.  Now, ex-boyfriend was not a big fan of anything that detracted from the sleak and powerful look of his Fat Boy.  So a back-rest was not happening, and a nice, cushy, comfortable seat for me was out of the question.  I ended up riding all the way out here on what looked like 1/6 of a tire, and it was about as soft.  On the last day, about 30 miles away from our final destination, we pulled off the road in the Rockies for a little break.  We parked on a hill, but I guess we didn\'t park up high enough, or else the bike was just SICK and tired of going, because it would NOT start.  We pushed and pushed and I don\'t know if you\'ve ever tried to push-start a Harley on flat ground, but let me tell you it was HARD, and I started to cry after it didn\'t start on the 7th attempt.  Characteristic of ex-boyfriend, he chose to yell at me rather than sympathize.  He said, \"What are you doing crying?!  Crying is not going to get it started!!!!\" (in a thick, German accent).  He had a point, and his lack of TLC pissed me off...so much so that I put my everything into that last push, and the thing finally kicked up.  I\'m sure, now, that he did it on purpose (pissed me off), and it worked.  All in all, though, I wouldn\'t have minded sitting on the side of the road and pouting a bit.  I had earned the right, had never complained once about my swollen ass and the fact that I was just TIRED of riding behind his big, stupid head.  Sorry, I just had to get that out.    So anyway, the bike has been returned, and I\'ve decided that I want to buy one some day.  What an incredible thing...to have the wind all around you and to be able to smell all the smells and  feel  the curves of the road.  If you\'ve never ridden (even on the back), you should give it a try.  Somehow, in some way, it\'s like being connected with nature, and we could all use a little more of that (especially us San Jose residents).  And just to clarify, part of what\'s making me want to cry about everything today has nothing to do with being sad.  Do you ever want to cry because everything is just so  beautiful ?  If not, then I\'m sure that sounds weird, but I think that everybody I know could identify with this feeling.  Because I still haven\'t been officially hired anywhere and have a lot of time on my hands, I was watching A Baby Story on T.V. earlier today.  I don\'t normally watch that show as I am years away from wanting a baby of my own, and in that state of mind, all the babies on the show come out looking the same.  Today was different, though.  When the featured Mama\'s baby came out and the doctor handed him to her, she kept saying, \"Oh, my beautiful baby, oh my beautiful baby, thank you God, thank you God.\"  It was the \'thank you God\' that got me, got me all choked-up and watery-eyed, thinking about what an amazing thing it really is.  And what a gift.  In that way, I envy Nicole.  Her body is doing the most incredible thing the human body can do...creating new life...a WHOLE NEW LIFE - where there was NOTHING before.  It blows my mind, and she is going to be a great, great mother.  I can only hope to know that kind of miracle some day.   '),(7400,669116,'19,June,2002','2002-06-19 00:00:00','    I WAS planning on sitting down to try to re-write my entry from the other night, but that\'s all gone now.  Basically, I had been talking about how strange it was for me to realize that all my co-workers are married with children, even though most of them are about my age.  Then I was talking about divorce and blah blah blah.  I want to talk about something different now, but relationships all the same.  I got a letter from a friend today in which he was mentioning that women are basically phonie balonies when they say that they want sensitive men.  He was pointing out that he is a sensitive man and good listener, to boot, but that women (possibly as a result of those qualities in him) end up thinking about him in strictly friendship terms.  He has decided to treat women the way they say they don\'t want to be treated, and see how far that gets him (I\'m thinking and hoping that last part was said kind of tongue-in-cheek).  Now, I have to say that I am no exception to the women-go-for-assholes phenomenon.  I can\'t explain this, I have no idea where it comes from and why we do it.  Hmm, but that\'s not really true because I know it has a lot to do with self-esteem.  I guess our choices in men are reflections of how we view ourselves and what we think we deserve.  So my advice to my friend would be to seek out women who have a stronger sense of self.  They are the ones who will appreciate his finer points.  I\'d like to think that I\'m past going for jerky boys.  I mean, right now I wouldn\'t dream of dating another one, which I think has to do with me having grown up and seen enough to know what will make me happy in the long run.  Time will tell.  The other part of this is that I think the jerky men (and women) are all about low self-esteem as well.  It takes a strong person to be able to be with an AMAZING person and still feel like you are okay, or that you are deserving of that person.  So it makes sense that those with serious self-doubts will choose to put-down the people that they are in relationships with.  As long as you are able to keep them doubting themselves, they will be afraid to leave you.  So I think that those with a stronger sense of self will automatically want to be with somebody who helps them to shine, rather than blocks their light from escaping.   And in this area I know I\'m getting better.  I want to glow, dangnabbit!  This is my rather simplistic take on things, but it\'s as much as I know about all this (or as much as I feel like writing about it anyway).  I only know cool people (aren\'t all people cool?), so it\'s my hope that all my friends will end up with people who appreciate them and treat them like the cool people that they are.  Life is SO SO SO SO short.  No need to waste it with people who bring us down...we are better off alone.   '),(7401,669116,'17,June,2002','2002-06-17 00:00:00','    ARRRRRGGHHHHHHHHH!!!!!!!!!!!!!!!!!!!!!!  I just spent an entire hour writing, and then my connection was lost and it ALL went away!!!!!!!!!  It was brilliant, too.  :)  This is why I still write snail-mail.  Dammit!!!  I\'ll try to reconstuct it when I regain my foolish sense of false security in technology.   '),(7402,669116,'28,June,2002','2002-06-28 00:00:00','    Golly Gee, it\'s been awhile!  I think I\'ve grown scared of writing in this thing after The Incident (when my everything got erased).  I\'m like the kid who just discovered that hot stoves hurt...I\'ll never have quite the same level of confidence.  But here goes....  I want to talk about music.  I need some help.  Part of my job at Target is to stock all the music and movies and books that come in.  I\'m realizing that there are A LOT of movies that I haven\'t seen, but that doesn\'t bother me so much because most of them look pretty stupid and I can\'t  believe  the things that people will spend 16 bucks for the DVD version of.  But anyway, this was supposed to be about music.  I\'m also noticing that there are a lots of bands and artists out now that I\'ve never heard of, which I know is my fault for not listening to the radio or watching Mtv.  The radio part is due to the fact that I can\'t handle any radio stations because of the repetition of crappy songs and the million commercials. The Mtv thing is basically the same...I mean, do they even show videos anymore?  Anytime I flip past it\'s like the Real World 56 or a bunch of people dancing on the beach to some lame somebody rather on stage.  Sooooo, the other night I turned on Mtv and there was show on called \"Spankin\' New,\" on which they show videos of the latest artists or an old artist\'s latest album.  I thought \'cool\' because I figured I would get to hear a lot of those people whose CD\'s I\'m stocking everynight and wondering who the hell they are.  And I was right.  I saw videos from Ashanti, B2B, Our Lady Peace, Dirty Vegas, Paulina Rubio, I don\'t know, a bunch of other groups that I can\'t remember.  And the reason I can\'t remember them is that they ALL SUCK!!!!!!!!!!!!  I can\'t believe the kind of stuff that gets serious air-play.  I know at least 7 musicians in my own private life that write better music, I seriously can\'t believe it.  So, the part about me needing help is where I ask you readers to send me some good recommendations.  I\'m definitely out of the loop since I can\'t seem to move past U2 and R.E.M and Tori Amos...you know the tried and true.  I think the most recent artists I\'ve been into were Coldplay, India.Arie and Badly Drawn Boy - and they\'re not even recent anymore.  I mean, I know there\'s good new stuff out there, I just don\'t know where to find it, cuz it ain\'t happenin\' at Target.  Even if it\'s an old artists that you think I might like....please help me!!!!  My e-mail address is  scoobisnac2000@yahoo.com .  I would LOVE to hear from you!!!!!  And there\'s another thing that kind of bothered me.  In the hour and a half or so that I was watching Mtv, there was only one song that I thought interesting and, I don\'t know, is catchy too dorky of a word?  And guess what it was!!  Freakin\' Eminem (Ahhhhhhh!!!).  Some song called \"Without Me\" where he talks about how awesome he is and how the controversy he causes on Mtv is what keeps it interesting and how everyone\'s just bored to tears without him and all this stuff.  Ok, so it wasn\'t ground-breaking or anything.  I agree with him about how controversy keeps things interesting, and it\'s hard to disagree with him that he says that he\'s just trying to be himself and that people want to censor him and everything.  I guess I just can\'t dig on his causes.  I mean, homophobic misogyny just isn\'t my thing, you dig?  So I guess it bothered me that I could dig his song.  I mean, don\'t you hate it when you can\'t stand a person\'s politics but you like their art?  I feel like I\'m betraying myself or something.  And I\'m wrestling, too, with how I feel about his politics.  I mean, it\'s obvious that I don\'t agree with him, but if I just go around disliking him and talking smack about him, aren\'t I just as bad as, say, the religious right, who I wish would just chill out and let people live?  I think I\'m asking for help in this area, too.  I think it\'s an interesting discussion and I would love to hear what other people have to say about it.  So write to me!   '),(7403,669116,'01,July,2002','2002-07-01 00:00:00','    Ok, first I just want to thank everybody who has responded to my plea...I will be publishing the responses soon, but they are still trickling in so I want to hold off a little.  Music....what amazing power it holds!!!!  I had happiest day since I got here yesterday because I finally got out and listened to some live music.  There is an eight-week long jazz thing here in the summers and every Sunday night they have a different group playing for an hour and a half on the bank of this little lake in a park downtown.  I knew that going would put me in the right frame of mind for the whole week, so I made my way down there.  Now, as I\'m sure many of you know, few things can enhance the live musical experience like a little somethin\' somethin\' to wet the whistle.  So I made a stop at the grocery store before I went.  I sold out, though, because, as I was a young girl there by myself, I felt too goofy to bring a 40 ounce like I would have liked :), and ended up with the more socially acceptable Bartles and Jaymes Hard Raspberry Lemonade.  I know, I never expected to see the words \"Bartles and Jaymes\" and \"hard\" on the same label, either.  It was for the best that I drank mild wine-coolers, though, because I had to go to work after the concert last night.  So, Hard Raspberry Lemonade and stationary in-hand, I made my way down to the lake.  The Sunday evening concerts seem to be a place for jazz afficionados to have a listen, and young non-jazz afficionados to see and be seen...the mix is interesting, and everybody brings food and makes a picnic of the whole thing.  So, walking to find a spot, I was letting myself enjoy all the smells that were changing and mixing as I made my way along the cement path...barbeque....cigarette smoke....pachouli (?)....and I found a spot right by the lake.  There were families on rented paddle boats coming around to listen, dogs and kids running around everywhere, and loners like myself taking it all in.  The kids were so freakin\' cute - a bunch of them jumped onto the steps near the large gazebo that was serving as the stage, and they were moving and shaking in that free and spaztic way that only kids are smart enough to know is not only okay, but a pretty good to spend an evening.  I watched this one little girl wander onto another family\'s blanket and ask them for a piece of the chicken they were eating...only little kids (about four years old) can pull that sort of thing off.  They complied, and the girl made a wry little face when she took a bite and discovered it was spicy.  She just grabbed her little pink water bottle and wandered around with chicken in one hand and water in the other...and she was very happy.  That\'s the other thing about children---they really know how to adjust.  So, just as I suspected, the majestic peacefullness of the jazz made me happy and ready for work...live music should be a part of my daily experience.   '),(7404,669116,'13,July,2002','2002-07-13 00:00:00','    Mmmmm, menial jobs.  You know how when you go to the grocery store (especially if you go early in the morning), the cans and boxes are all lined up and you can tell that some poor soul actually spent their time pulling all that stuff forward for the asthetic pleasure of you the shopper?  Well that\'s me...some poor soul.  Yesterday morning I got my first taste of \"zoning,\" and lemme tell YOU...what a rockin\' good time THAT was!!!!!  Did I mention that the Target I work at has a grocery store attached to it?  As I was sitting on this little milk crate stacking the tuna, the soup, the olives, and hosts of products I have never heard of before and certainly would never have noticed if I hadn\'t been meticulously stacking them on top of one another, I was thinking that I can\'t believe that they pay people to do this.  And all because some marketing person somewhere along the line did some kind of survey and declared that people will buy more if they perceive, I don\'t know, AMAZING selection, cleanliness, FULL SHELVES.  I\'ve thought about this.  I mean, I know a nice and clean store makes for a pleasurable shopping experience, but sometimes I think I\'d be more likely to buy something if it seemed like they were almost out of it, like there was a run on it or something so I\'d better get it while it\'s hot.  You dig?  The other thing I was marvelling at last night was wondering how big a kick marketing execs must get out of the fact that we, lemming Americans that we are, will pretty much buy whatever they tell us to buy.  I know this is true.  I stock the books every night, and I notice that the majority of the books that have to be replenished are the paperback bestsellers that are on an endcap for the world to see.  C\'mon, you can\'t think of ANY other book to buy than the latest soft porn from Nora Roberts?!  Actually, I don\'t even know if there\'s sex in her books, so please don\'t write me horrible letters if she\'s one of your favorite authors and I have it all wrong.  Also, there\'s that DVD thing that I mentioned before.  Every single day we run low on the supply of about 50 copies of \"Black Hawk Down\" on display.  Was is THAT much fun watching the American soldiers blow away the evil enemy?  And someone, I don\'t know who, has been buying \"Shallow Hal\" on DVD.  Yes, it\'s true.  The store I work in is absolutely HUGE.  I heard last night that it is the 6th most profitable Target in the country.  I was walking around there last night, stocking an insane selection of back-to-school supplies, when I thought about the fact that I don\'t have any of the 100,000,000 things for sale there (back-to-school section and beyond).  Seriously, I know I have a trash can from Target that they still sell there, and there are definitely SOME CD\'s there that I have, but other than that there\'s nothing.  More importantly, I don\'t  need  any of that stuff!!!!  And neither do you!  From now on I\'m spending my money on rent, school, food, CD\'s, and traveling.  I\'ll buy some new (used) clothes every now and then, but I certainly don\'t need a new,  better  set of tweezers, a mini-paper shredder for my desktop, a police officer-shaped cookie jar that tells me to step away from the cookies when I lift his head, or a little hamster that twirls num-chucks while singing \"Everybody Was Kung-Fu Fighting.\"  Okay, so maybe I do need that last one.  One must be free to indulge ONCE in awhile.  My friend Murphy has a blog that contemplates art, economic systems, authors, the cycle of life, and all things beautiful and ugly in the world.  I contemplate canned tuna and singing hamsters.    I can\'t WAIT for school to start again!!!!!!!   '),(7405,669116,'11,July,2002','2002-07-11 00:00:00','    Ok, I think I\'m about to risk sounding preachy, but I\'ve got important stuff to say, so please read up!!  :)  I don\'t know how many of you have these sorts of days, or how many of you would believe that I have these sorts of days...but yesterday was one of those days....those days where you feel like the worst person that ever lived and that you have nothing good to offer and that you should just forget about it cuz it\'s too hard.  Ever feel like that?  I went through a year-long period just like that when I was about 18.  Now I only get that feeling every once in a great while, and I know more now than I did then, so it doesn\'t last long.  But anyway, all I wanted to offer about that is this...when you feel that way, please just go to bed with the hope that the next day, or the day after that or the day after that, it will feel better.  Because eventually it will.    About six months after I had finally climbed out of my depression, I took a trip to Mexico with three good gal friends.  It wasn\'t really a vacation, we were there to help out where we could with what was a somewhat dangerous situation in Chiapas, but even at that we experienced all the amazing things (and more) that one does when they take an extended vacation.  I have a vivid memory of riding on a rickety old bus back from a little village, sitting next to kids and old people and chickens, looking out the window and crying a cry of absolute joy.  I was thinking how happy I was that I never gave up on life, back when things seemed so very unbearable.  18 months is a long time to wait for a sign that there is still more to be experienced, more worth hanging around for, but it was well worth it.  Last night wasn\'t nearly so dramatic.  I just went to bed feeling sad, and woke up feeling happy.  I know it\'s not always that simple or instantaneous (for any of you who have ever been through a depression), but it\'s a cycle all the same.  Things will ALWAYS get better.  There are ALWAYS things worth living for.  I know I sound really serious, and I suppose I am.  But I\'m calm.  This is one of those things that I just know, and I want to share.  I also want to advise everyone not to run from the rain, the next time it\'s raining.  I spent the afternoon outside in a hailstorm yesterday, and it made me feel young and free and like there were things I\'ve been missing out on that I don\'t want to miss out on anymore.  \"Gather ye rosebuds while ye may...\" I have to repeat this again.  Life is SOOOOO short.  Please go out there and enjoy it, and please invite me along if you think of something incredible, even if it\'s just a sunset.   '),(7406,669116,'10,July,2002','2002-07-10 00:00:00','    Ok!  The new music issue has finally arrived.  Thank you to everyone who responded.  As for those of you who didn\'t, I\'ll just assume you\'re listening to O-Town :)  I was disappointed to find that a lot of people feel the way that I do - that music that\'s out now is just not that good.  I have REAL faith that there is good music out there, I\'m just gonna have to go to alternative sources to find it.  I mean, that\'s where all the good EVERYthing usually is, anyway, right?  I will definitely mention anything ground-breaking that I run into here on this journal.  In the meantime, here are the responses, in the order in which they were received :)   Personally I think music out now sucks.  If I ever turn the radio on,  I find myself listening to 91.1.  The jazz station.  Why?  Because they express themselves through the music.  Not by what they say but what they do on the instruments.  I love it.  I like to turn 91.1 up when there is a car next to me \"bumpin tight ass shit foo\" rap, pop, and so called rock.   This is from Richard, or Meatball, as I prefer to call him.  I couldn\'t agree more about the jazz, which is why I\'m coming back to San Jo in time for the jazz festival (Aug 9-11).  Be there or be square.  I was telling him that, as far as jazz musicians expressing themselves though the music, I find jazz to be some of the most organic sound available for that reason.  How talented, to be able to say something without saying a word!   I wish I could help you out with your search for new music but I still consider bands like Counting Crows, Alice in Chains and Pearl Jam to be the new wave.  I guess Karl agrees with me about tried and true.  Counting Crows just released a new album, by the way, for those of you who also dig them.       My contribution to current music would be David Gray and Dar Williams.  Oh yeah..Linkin Park is pretty cool..It\'s fun to hear someone screaming again...And the lyrics are quite intelligent. Nelly Furtado is awesome.  This is from Murphy, whose super smart, funny, and entertaining online journal inspired me to start my own.  David Gray....mmmmmmmmmmmmmm.  Good stuff.  Anyone else out there familiar with Dar Williams?  My curiosity has been piqued.     Don\'t feel bad about being out of the music loop. I keep up with much of what\'s out, but find it all so uninspiring that I don\'t take it very serious nor spend my hard earned money buying it.  Though a few things I\'ve picked up of late our India Arie (which you mentioned) Bilal and of course Maxwell.  Most everything else I listen to is older things that Steve has given me like Santana, The Doors, The Cure, and Hendrix.   This quote is from Paul \"Pablo Honey\" (Steve is his roommate).  I haven\'t listened to Bilal or Maxwell, but I\'ll give \'em a whirl.  As for Santana, The Doors, The Cure and Hendrix...well, yeah....of course.  :)    WAR - Mos Def- A Tribe Called Quest- Method Man- more Bob marley- Jill Scott- Lauryn Hill- Ella Fitzgerald- Billie Holiday- Common- Ani DiFranco- The Roots- Black Eyed Peas- Nina Simone- Erykah Badu- and don\'t forget who told you about India Arie!!!    All these wonderful suggestions are from Freeda, who yes, I have to declare, introduced me to India Arie and made my life that much happier!!  There are definitely some great artists on this list.  I have to check out Mos Def, Common, The Roots and Black Eyed Peas, but I can speak for the rest and second her on her suggestions.   Patty Griffin is a gem, as is Jennifer Knapp. I listen to a lot more Christian music these days, as it\'s hard for me--like you--to actually see the artistic merit in a lot of new bands. I still listen to mostly U2, Counting Crows, Dylan (who\'s in concert at my school in August!!!), Taylor, my man Paul Simon (who played here last year!!!), Morrison, REM (although I can\'t seem to chill to anything other than Automatic), DM Band on occassion and Pearl Jam (who will always be cool). Of course, there\'s Leonard Cohen and Louis Armstrong. I like newer people like David Gray somewhat, and Shakira\'s song, \"Underneath Your Clothes\" is really sweet.  In high School, Renee and I used to bump down the streets of Phoenix in her 1982 BMW blasting The Pixies and REM.  Mmmm, sweet memories.  I have to reinforce her U2, Bob Dylan and Leonard Cohen references.  That\'s Van Morrison and James Taylor, by the way, if you were wondering.  And as for REM, please do check out the Monster album (it grows on you) and my personal favorite, New Adventures in Hi-Fi.  Well, kids, that\'s it.  If you think of anything else, let me know.  My brother was just out here, and he had Wilco with him, who my friend JD also introduced me to, along with Jack Johnson.  Both come highly recommended.  I also forgot to mention Beta Band and Mary J. Blige, and JD and I have decided that Norah Jones deserves some further listening.  As for my brother\'s visit...it was awesome, as usual.  I didn\'t write that whole time because I was trying not to be such a computer junkie while he was here.  It\'s so hard though, not to become addicted to technology.  Just as long as I never get a cell phone I think I will be okay.  Oh!!!  That reminds me.  My family and I went to that jazz in the park again on Sunday, which was just as rad as last week, with a Dixieland band this time.  The kids were shakin\' it again, along with this old and *clearly altered* man with a cane and a little duffel bag that seemed to REALLY be enjoying himself.  He was just moving non-stop, and then he would hold this cell phone up to his ear while he was dancing (though it was obvious that he wasn\'t really talking on the phone).  We were all watching him and digging him for awhile, and my bro said, \"I\'m jealous of his moves.\"  And I said that I was jealous he got so many phone calls.  Sometimes, I seriously think it would be fun to be a crackhead.  I\'m all for alternate perceptions of reality, and nobody else in that park was feeling as free to physically enjoy the music.  I\'m kidding, of course, about the crack part.  But I do wish more people were so unihibited (starting with me).  I know this song is kind of cheesy, but LeeAnn Womack\'s \"I hope you dance\" has a really good message.  Next week I\'m getting out there and get my groove on with that guy!   '),(7407,669116,'17,July,2002','2002-07-17 00:00:00','    It occurs to me that, if I think about it, I can\'t really think of anything that I do as a routine.  The most I can say for myself along those lines is that I brush my teeth everyday.  Other than that I can think of nothing that I do daily or even weekly, or that I\'ve done as a habit over a period of time...which is why I am so intrigued by and appreciative of the creatures of habit in the world.    Three and a half years ago, when I was working at Target the first time, there was man that I worked with (actually, I didn\'t really work with him...he and this other man (neither of whom I ever heard uttering a single word) worked sort of independently, changing displays and signs and stuff like that).  Every single night....I mean EVERY SINGLE night, he wore this american flag handkerchief around his head.  This was before american flags as clothing were  ALL THE RAGE .  Every night...same handkerchief tied in the same way around his seemingly dirty and certainly unkempt hair.  Hey, guess who\'s still working there are still wearing that SAME american flag?!  Yes!!!!!!!!!!  It looks a little more war-torn, and little more faded, but the same handkerchief nonetheless.  I find myself wondering about this man with some degree of frequency.  I know he\'s married, and I wonder what his wife thinks about it.  I wonder if it really is the same one or if he has a whole drawer-full.  I wonder if he will ever wake up and just want something different.  One morning I had just passed this man in an aisle and I started wondering about him, as usual.  It was after 8 o\'clock in the morning, so there were already customers in the store, and I was walking through the toys section.  I walked by the hot wheels aisle and was nearly stopped in my tracks.  Another creature of habit from my previous Target life!!!!!!!  The last time I worked there the store used to open at 7 instead of 8.  We used to get the biggest kick every morning watching for the Hotwheel Men.  Every morning at 7 o\'clock sharp, the doors would open and in would rush a group of five men walking, no trotting together toward the hot wheels.  They were collectors, and they came everyday looking for brand new cars.  This group of men was entertaining to watch...not only because of the seeming ridiculousness of their task, but because of the dedication and seriousness of purpose with which they carried it out.  They were always in such a hurry, and I even saw them run sometimes to get there faster.  And the thing I could never understand was who it was they were trying to beat there.  They were the only ones looking at the hotwheels, and they were all friends, or connected in some way by this hotwheel thing...so were they racing  each other  there?  The other funny thing with these men them was the physical look about them....a motley crew of people that you would NEVER put together in a group of people if you had a thousand tries.  And it was my favorite man that I saw the other morning - a super tall and super thin African-American man with Wranglers, cowboys boots and a cowboy hat....still....same outfit, same aisle, looking for new hotwheels.  I LOVE these people.  I think the world needs these kinds of people.  They are the ones who notice when things are out of order, who keep things in-check.  Hotwheel man and his cohorts keep thousands of people gainfully employed - designing, producing, packaging, transporting, stocking and selling hotwheels.  Seriously, where would the world be without people like that?    I think I could take a lesson from these people.  I mean, I don\'t ever want to get all weird about my routines and feel like I will have serious problems if something doesn\'t go EXACTLY LIKE IT ALWAYS HAS.  (You should have seen the looks on the Hotwheel Mens\' faces one day during Christmas when we were just plain OUT of them.  They looked at each other with puzzled expressions and didn\'t know what to do with themselves.)  But that said, I\'ve always admired people who had things that were that important to them...exercizing, gardening, playing the guitar, I don\'t know, things like that that they love to do and feel a little bit off about if they can\'t do them for some reason.  I try new things, try new hobbies and stuff like that, but nothing seems to stick.  I\'ve always felt a little bit lacking about that...like I lack passion about the things I like to do because, obviously, I don\'t like to do them THAT much because often sooner than later, I stop doing them altogether.  To be honest, I\'m surprised I\'ve been doing this journal thing as long as I have (I know, two months doesn\'t seem like a lot to most people, but believe me, it\'s the sort of thing I would normally have quit doing by now...so it\'s actually good to know that people are reading it, it inspires me), and I worry a bit about waking up one morning and not caring about writing in it anymore.  I feel like I want to work on not letting things go so easily, not being so quick to move onto the next thing...in some way, if I can do that, it will feel like I\'ve grown up a bit.  It is said that Geminis get bored easily, and I don\'t know if it\'s a Gemini thing or not, but I am certainly easily bored.  And I just think that\'s lame.  There is nothing so uninteresting that it doesn\'t warrant a little effort and attention.  I guess only time will tell.  I want to be a creature of habit...just one, just one really good habit that brings me joy and makes people look at me funny :)  I\'ll keep you up-to-date.   '),(7408,669116,'15,July,2002','2002-07-15 00:00:00','    I have a new person to love!!!!!!  My best friend Nicole just had her baby, and life looks more beautiful than ever.  A tiny little girl, 6 pounds and 3 ounces...I can\'t WAIT to meet her!!!  It was funny, I mean, I\'ve never been in this situation before - where somebody I\'ve known since adolesence makes that huge leap into the next phase of life, and I didn\'t know how it would affect my outlook.  I heard somebody at work the other day saying that having a baby makes you look at EVERYthing differently and I imagine that\'s true, cuz this baby isn\'t even mine and I feel that way.  When I talked to Nicole I couldn\'t help it, I started crying because I felt joy and amazement and admiration and excitement all at the same time.  I just still can\'t get over the fact that her and Raul made an ALL NEW PERSON!!!!!  I know I mentioned this in an earlier journal, but I feel like I can\'t say it enough or articulately enough to really express how I feel about the whole thing.  My best friend......a Mom....the most important job in all the world (next to being a Dad, of course).  I already love the little tyke and I have yet to see a picture.  I could hear her making little baby sounds in the background while we were talking.  Apparently she was asleep but was just sending out little signals or yawning or something, I don\'t know but it was cute.   As I was driving to work the night that I found out about the baby, I was still in a state of awe.  Listening to Van Morrison\'s \"Moondance\" CD just brought out the intensity of emotions...\"She give me love, love, love, love, crazy love.\"  I was thinking of the song in reference to my new honorary niece.  I know kids do that for their parents (and all the people that love them, actually).  They just really know about love and unbridled joy and wonder and telling people what you really think and saying stuff that makes you laugh so hard when they\'re not even trying to be funny.  Crazy love.  I want to send out my sincerest congratulations to the new parents!!!!!!  Auntie Kisa can\'t wait to spoil her rotten!!!!!  :)   '),(7409,669116,'02,August,2002','2002-08-02 00:00:00','    Ahem.  I saw something interesting tonight.  I was waiting at a red light when I witnessed four grown men standing on the street corner, holding signs that said \"Stop Immigration!!!!\"  One of the signs said \"Help us take back OUR country\" on the bottom, and another of the signs said \"Non-whites are RUINING our land!!!!!!!!\"  Okay, I know I don\'t even really need to go into this any further....all the thoughts that ran through my head, all the questions I had for them, of all the ridiculous things!!!!  But I found something interesting in my reaction to these men.  I was baffled, a bit speechless (imagine that) and awestruck, but the one thing I didn\'t feel was anger.  I felt it just briefly, but it didn\'t stay with me like it would have in the past.  In the past I would have let the incident eat away at me until it reached the very core of my soul and maybe I\'d get sick to my stomach or something.  Tonight, I just thought \"huh.\"    I wonder what Malcolm X would have to say about that.  I know that anger was the driving force behind a lot of important social change that has transpired over the past couple of centuries (or millenia).  Anger has served its purpose to an extent, and it will continue to, I just don\'t know if I want to live with it, personally.  I\'ve been reading a lot over the past couple of days about a graduate program in psychology that has sparked my interest and that came to mind while I was gauging my reaction to these men.  The program is called Contemplative Psychotherapy, and it is offered at a Buddhist school here in Boulder.  All the programs at this (1,000 student) school are rooted in Buddhist idealogy (including their writing program, which was founded by Allen Ginsberg and is called the \"Jack Kerouac School of Disembodied Poetics\"), and meditation is a part of graduate studies every single semester.  What interests me about this program is that it lends a spiritual hand to otherwise cold and clinical western psychology, and its participants are encouraged to find peace and presence in the moment in their own lives, as a way to understand people and the world around them.  There is so much to it that it\'s hard to describe it all right now, but there were a few things that I was reading about in some of the program\'s literature that I thought were interesting, and which are giving me a different perspective on people and their (sometimes unfathomable) behavior.    One Buddhist monk (Chogyam Trungpa Rinpoche) is describing his first experiences with the western world (in England) and how he reconciled them with his eastern background.  He writes, \"Buddhist psychology is based on the notion that human beings are fundamentally good.  Their most basic qualities are positive ones:  openness, intelligence, and warmth....Coming from a tradition that stresses human goodness, it was something of a shock for me to encounter the Western tradition of Original Sin.  When I was at Oxford University, I studied Western religions and philoshical traditions with interest, and found the notion of original sin quite pervasive.\"  In this particular case, the Rinpoche was talking about how the idea applies to western psychology, which sets up the hierarchical arrangement of sick patient and healthy doctor.  The sick patient has done something wrong and the doctor is trying to \"fix\" the problem.  The Buddhist idea is that people are fundamentally healthy by nature, and that through various forms of meditation, etc. people can find their way back to health, having temporarily departed from it (the reasons for this departure are not considered unimportant, but regression therapy and things like this would never be the focal point in this kind of therapy).    I know it seems I\'ve gotten off the subject, and I probably have, but please indulge me.  I often think about how we are products (to an extent) of our environment.  Sometimes I look at guests of, say, the Jerry Springer Show and think how I shouldn\'t judge them because their experience could easily have been mine, had I been born where they were born, into families, financial, social, or educational situations like theirs.  Even saying that sounds like a sort of judgement to me.  It\'s hard to think of different without considering words like \"better\" or \"preferrable,\" you know what I mean?  Anyway, the Buddhist idea of inherent goodness, rather than guilt as a result of the actions of Adam and Eve and those that crucified Jesus, fits well with my outlook on things.  These men on the street corner...they are not evil (is there really such a thing?).  They just feel cheated out of the things they really wanted in their lives, and they are looking for something to blame their disappointments on.    Another central tenet of Buddhist theology is compassion.  I like that.  I think we can accomplish a lot more in the world by using compassion, than anger and war.  I know there are MILLIONS of people out there that would disagree with me (they would say that Osama bin Laden is evil, Hitler, Rush Limbaugh, etc. and that they should be destroyed), but I just can\'t let go of the idea that peace is possible, when people look at the world with loving eyes.  We spend so much time blaming.  So much time feeling guilty, judging and being judged....looking around at other people and wondering why, damnit, they couldn\'t just be and act more like US!!!!!  What a waste of time.  What a lonely and depressing way to pass a lifetime.  I mean, those men could have been spending a nice, quiet evening at home with their families playing Scrabble or something.  Instead they\'re standing on a street corner feeling livid and inciting anger in others.  And think about all the anger.  They riled up the people that agree with them, and pissed off those that don\'t.  One thing I can say for them....they made people think, and I\'ve decided once and for all that thinking is a good thing :)  Ahhh.  Breathe deep.  On a lighter note...I got a great little bit of trivia from my friend JD the other day.  It seems that there is a Walt Whitman Mall (yes, I said Mall) in Long Island!!!!!!!!!!!!  I have to quote him here...\"In Long Island there is a WALT WHITMAN MALL!!!! If that don\'t beat all huh?? I can just see that planning meeting.... \"What can we think of that will make people want to buy stuff?? I know... Walt Whitman hated materialism so why not name the epitome of materialism after him!!!  And we will call it Walt Whitman Mall!!\" (picture the guy moving his hand in an arch with that far-off look in his eye)\"  Awesome.  Here\'s a site:   http://www.shopsimon.com/smt/servlet/SMTMall?mid=106&pn=ENTRY&rs=0    :)  Peace and love and joy to you all.   '),(7410,669116,'31,July,2002','2002-07-31 00:00:00','    Ha, so speaking of things that I don\'t stick to doing!  I know, I\'ve been far far away for a long long time.  But I haven\'t just been sleeping, if that counts for anything.  I\'ve been thinking...just not sharing.  It started the weekend that my new niece came into the world.  I don\'t remember ever having felt that sense of wonder in my adult life.  Few things came close in my childhood or adolescence either, come to think of it.  The idea that Nicole offered an entire person as a contribution to the world continues to blow my mind.  But the following week I got the news that the brother of an old friend had passed away - the very same day.  The same day that one family and circle of friends was experiencing unparalleled joy, another family and circle of friends began to feel the profundity of loss.  Was it one soul making room for another?  I don\'t know how to explain these things.  Chad, who passed away, was close to my age, and he passed away as a result of a number of medical complications that came from having become partially paralized in an accident years ago.  His brother Marc was telling me about the funeral service, which was attended by hundreds of people.  From everything I\'ve heard, Chad had an incredible lust for life before the accident, which left him in a state of depression that he never really recovered from.  I met him after his accident, and I remember him as a charasmatic and gentle soul, but Marc has told me over the years that something was missing after the accident and it never really returned.  I wonder what it was that Chad came to understand at the moment of his passing.  I know that he (possibly unknowingly) taught all the people around him about living, which is a beautiful gift.  But why him?  I think he probably had that thought a lot in his quiet moments...why him?    Sometimes I am amazed at the depth of suffering that the human spirit can endure.  I can\'t imagine what it was like for Chad\'s parents to bury their son, for his brother and sister to say goodbye to him and know that their lives would never, ever be the same.  I guess life really is a series of waves - waves of joy, sadness, wonder, loss, love, fear, beauty.........big, undulating waves that bring us to the highest of highs and the lowest of lows.  It reminds me of that scene at the end of the movie \"Parenthood,\" where the grandma says something about roller coasters...she\'s making a life analogy, but Steve Martin is not in the mood and he makes some sarcastic comment, to which his wife (Mary Steenberger) answers, \"I happen to like the roller coaster.\"  It\'s like Joni Mitchell\'s song \"The Circle Game\":  \"And the seasons, they go round and round and the painted ponies go up and down, we\'re captive on the carousel of time...\"  It\'s like Khalil Gibran wrote in \"The Prophet\":  \"When you weep, know that you are weeping for that which has been your delight.\"  It is only in the greatest of joys that we can know the worst of sorrows.  Arrivals and Departures.  Round and round and up and down.   '),(7411,669116,'06,August,2002','2002-08-06 00:00:00','    Packed, mentally and (almost) emotionally detached from Colorado, ready to come home.  I\'ll be on the train as of 9:00 tomorrow morning, that\'s about 11 hours away, and I am quietly content.  The beginning of the rest of my life awaits me...I can\'t wait to get started......   '),(7412,669116,'04,August,2002','2002-08-04 00:00:00','    Apparently I\'m not the only one exploring the world of Buddhist thought...and I\'m glad to hear it.  My intelligent friend and neighbor Paul has reminded me that, along with compassion and a basic belief in inherent good, Buddhist idealogy emphasizes patience.  I hope he won\'t mind if I include his own words...\"I think more than just compassion though we also need patience (another tenet of Buddhist thought).  We are all in such a hurry to change the world that we sometimes forget that it takes time for others to grasp the meaning of what were talking about or even allow them to explain why they are the way they are.  Rather than be patient and allow people to adjust over time we become frustrated and immediately discount their abilities to change.  What then happens is these people feel misunderstood, frustrated themselves, and then angry which is then projected onto those that have tried to help.  Having the best intentions, however lacking the necessary patience, those who wish to help end up causing more strife.\"    It\'s a thoughtful, important and interesting point.  I was thinking about what he said about letting others have the chance to explain themselves.  It occurred to me that - so often - when we ask questions of people who we know think and act really differently from ourselves, we aren\'t truly interested in the answers.  We ask loaded questions and, beyond even anticipating the responses, we go as far as to be thinking ahead to OUR responses to their responses.  Headway will never be made in this fashion.  For a spirit of change to really take hold, we need to ask questions and then be open to hearing the responses...really HEARING them,  listening  to them.  That is my challenge for myself.  My radiant and articulate friend Laurie also had some thoughts on the subject (with the Psych degree to back them up)....\"Their [the Buddhist school\'s] take on Western Psychology (actually, through my cross-cultural studies, I’ve found that 99% of all psychology is STILL Western, actually, mostly AMERICAN WESTERN, and change is happening, but quite slow) is right on the mark, which is why most people don’t like going to see a “shrink,” since we look to them to control and fix us, as opposed to sitting next to us and guiding us.\"  Guides...isn\'t that what we all want?  Spiritual guides and human mentors...I mean, we ALL have something left to learn (some of us have lots and lots and LOTS to learn, starting with myself, of course).  Soooo much more yet to be figured out...never stop wondering, never stop learning.  I learned something about myself the other day - a very important something.  I went out to breakfast with a few co-workers the other morning, after work.  This was a goodbye sort of breakfast as it was one of my last nights at work and the friends that were there wouldn\'t be working on my actual last night.  So, as I was pulling away from the parking lot, after a few hugs and the exchange of some \"ahh, I like you, I\'m gonna miss you\" sentiments(only in Spanish), I found that I was crying.  Like, I really WAS gonna miss these people.  I know it\'s only been two months, but you can make some nice connections in two months...there are many, many warm and amazing people out there.  And here\'s what I learned about myself:  I\'m  tired  of moving on.  I used to feel like I had turned my heart off.  I was so used to moving around the country and starting and quitting jobs and stuff, that I became de-sensitized to loss.  I didn\'t miss people.  I mean, I did from time to time, and there are people that I\'ve chosen to keep in touch with over the years, but I didn\'t let the emotion really take hold in me, didn\'t feel the pain of it.   I think that\'s horrible.  I think it shows a lack of understanding of human value and the importance of meaningful relationships (friendships and otherwise).  I\'m not like that anymore.  The older I get, the more it hurts...just planting little seeds, sending down little baby roots, and then ripping everything up and moving on.  I\'m ready to settle in some more, which is partly why I can\'t wait to get back to San Jose.  I\'ve finally let myself call a place home, and been there long enough to where I can run into people I know walking down the street downtown.  While I do want to come out to Colorado for graduate school eventually, I know it\'s gonna be really hard to leave San Jose (I can hear the snickers of all you people who have wanted nothing more than to get the hell out of Dodge, but I\'ve been a lot of places, and I can say that, for me, San Jose holds a unique kind of charm, despite its urban sprawl and techie MADNESS).  No more uprooting all the time.  At some point, moving on is nothing more than just empty searching (for me at least).  In the past, I\'ve prided myself on not being afraid of change.  But is being afraid of NOT changing any better?  I feel like the proverbial bachelor...afraid to commit.  I\'m not talking about relationships here, but that thing I mentioned before - fear of sticking with something and seeing where it can take me, which includes just STAYING PUT for a little while!!!!  We\'ll see where this road leads....   '),(7413,669116,'24,August,2002','2002-08-24 00:00:00','    Sometimes I have moments during which I have the distinct feeling that I could love the entire world.  I’m not talking about a hippie kind of we’re-all-connected-I-have-love-in-my-heart-for-everybody thing (although I mean that, too).  I mean that I think I could fall in love with every single person I meet.  How to explain this?  I know it’s probably sounding like a drug-induced sort of feeling, but it’s not.  It’s actually usually a music-induced thing…live music in particular.  I remember watching all the listeners and dancers at the Salsa stage during the jazz festival a few weeks ago and thinking to myself that this was the most beautiful group of individuals that had ever been gathered in one place.  The thing is that it’s the most beautiful group of people EVERY SINGLE TIME, no matter who the people are, what kind of music it is, or where it’s all taking place.  So, that said, I think it must have to do with the joy to be found in watching people let go and allow themselves the pleasure of abandon (“Let your mind go and your body will follow”).  I was trying to describe the feeling to my friend while we watched the salsa dancers.  It’s like the music and the dancing reaches a sort of fevered frenzy, and when I look around it seems that every movement, every smile, the little beads of sweat forming on the surface of somebody’s sun-browned arm – it’s all frozen in time – little snapshots of euphoria, and there’s me standing or dancing nearby, trying to soak it all in and remember every single detail for future reference.    It’s at those moments when all I can see are awe-inspiring scenes everywhere, in every direction I turn:  ‘that man has eyes a color that strikes me to my very soul.’  ‘the way that woman moves in her hip-hugging skirt and midriff top makes me feel like, if I touched her (even just accidentally brushed against her briefly), I would melt at her feet.’  Thoughts like these go through me head, and I find myself wishing against all hope that the music would never end and that there were, indeed, time enough in one person’s life to love everybody up close.  I want to hear all their stories.  I want to know them all and touch them all and be invited to their families’ houses for Thanksgiving.  I want to read their diaries and look at their bookshelves and ask them all about their third grade teachers.  I want to know what they’re most afraid of, and if it’s something of this world I want to go out and conquer it with them.  “Sky diving?  Let’s go next Saturday…we’ll have ‘em drop us somewhere over the Salinas Valley and we’ll see the amazing hills that Steinbeck spoke of, only as birds do, instead of humans.”  I think this all sounds kind of selfish…me here, wanting it all.  But it’s just that there is so much to love out there, and what seems like an unfair amount of time in which to love it.  If the universal waitress asked me what I was having, I\'d answer that I wanted seven courses of life and an extra side of childhood (for good measure), plus a glass full of water from the Fountain of Youth to drink.  Somehow, though, I know I\'d still be hungry.   '),(7414,669116,'20,August,2002','2002-08-20 00:00:00','    What is with these overcast mornings?  I feel like I\'m living in L.A. again.  And that\'s a bad, bad feeling.  Not really.  I actually feel pretty good, which I think is partially due to the fact that I returned to old stomping grounds last night...Cafecito\'s Open Mic...the grooviest place to be on a Monday or Thursday night.  For anybody who doesn\'t live in San Jose or hasn\'t been to this place, I\'ll break it down.  There are two different house bands that play at these open mics (both jazzy/funky), and they play in-between hearing from various musicians and poets (and sometimes freestyle hip-hop artists), all of whom are very talented.  It used to be free, but most of the people there are students and I think they weren\'t buying any coffee or something because now it costs two dollars, but it is well worth the money.  Going there is definitely a natural high (and necessarily, too...they don\'t sell alcohol, which is beautiful, because you get to see how amazing and expressive people can be all by themselves - without anything helping them reach another state of consciousness but their own talents).  The performers are all very socially conscious and have so much positive energy to share that it always contagious.  I\'d been away from there all summer (obviously), but I also didn\'t make it all last semester because in my mind I was too busy.  But that\'s bullshit.  I could have made it if I would have just made the effort.  And I pledge to this semester, because I think that if I don\'t allow myself that sort of outlet, I\'m gonna go crazy.  I just mapped out my schedule like a high school freshman because, between six classes and seven 1 1/2 hour sections of tutoring in the school\'s writing center (my new job), I know I\'m gonna have a hard time remembering where I\'m supposed to be at any given time.  I feel it...I can actually feel the freedom dripping out of me :)  Ahhhhhhhhhhhhhhh......................................   '),(7415,669116,'19,August,2002','2002-08-19 00:00:00','    Okay, so the last time I started doing this I experienced one of those horrible computer-crashing-out-of-nowhere things and lost it all (again!), but I\'m trying my luck now because I wore a summery dress today in order to make the sun come out from behind ugly grey clouds out and it worked, so I figure I\'m magic today and magic people don\'t have computer crashes (do they?).    It\'s been roughly a week and a half since I returned to San Jose and I\'m oh so happy to be back!  I realized I missed ALL of it...all of the lovely, as well as the yucky (including, but not limited to:  endless passing of cars all night right outside my window, many of them with thumping bass, pigeons (which I don\'t think exist in Colorado) and the ridiculousness of signs reading things like \"Historic House Moving\" as a euphemism for the mayor\'s heartless uprooting of really old houses with long-time residents in order to make room for the new city hall which will ensure his legacy).  Yes, the city is a beautiful thing - as was the Jazz Festival (Salsa Stage, most especially), and the GREAT show I saw the other night at Plant 51 - an Ozomatli-esque group of fine, talented young men who call themselves the B-Side Players.  But there was business to take care of, too.  I got a job in the writing center at school, which sounds like it will be challenging, and challenging is good.  However, beyond the job, three weeks worth of laundry to contend with today, and the business of hanging out lazily in coffee shops catching up with friends, I\'ve been exquisitely, thankfully, and pleasantly idle.    One can never be quite so idle without feeling a slight bit of guilt (if the \"one\" is me, anyway), so I decided I had to get some good reading in before school starts and I have to read what THEY tell me to read.  I decided to start with Camus\' \"The Stranger,\" which had shamefully sat unnoticed on my bookshelf for who knows how long.  You know how you grow up your whole life hearing about books and movies that are supposed to be the Be All End All of books or movies....and then you finally get the chance to watch, say \"Gone with the Wind\" and you\'re all, \"Wha?\" \"What the hell is so great about THIS movie, book, etc.?!\"  You know what that\'s like?    Well, \"The Stranger\" is nothing like that.    This is a book that has earned its place on the imaginery list of must-read books that might not necessarily be in THE CANON, per se, but everyone knows that you\'re supposed to have read them (The Canon, what a bunch of phony-balonieness that is, by the way).  For any of you who haven\'t read it, this book is said to epitomize existential philosophy and all that jazz.  I would normally be opposed to a work of fiction that attempts to present such a huge philosophical or spiritual idea (\"The Celestine Prophecy\" drove me nuts for that reason!) because if it\'s something as important as a philosophy about life, I\'d prefer for authors to just come out and say what they mean, you dig?  But this is the exception, and I\'m not sure why, maybe just because it\'s so well-written.    Check this out: (the speaker is getting used to his new life of incarceration after having committed a murder) \"Afterwards my only thoughts were those of a prisoner.  I waited for the daily walk, which I took in the courtyard, or for a visit from my lawyer.  The rest of the time I managed pretty well.  At the time, I often thought that if I had had to live in the trunk of a dead tree, with nothing to do but look up at the sky flowering overhead, little by little I would have gotten used to it.  I would have waited for birds to fly by or clouds to mingle, just as here I waited to see my lawyer\'s ties [the lawyer wears ugly neckties] and just as, in another world, I used to wait patiently until Saturday to hold Marie\'s body in my arms.\"  I love that passage.  Anyway, what\'s interesting about the story is that the plight of poor Meursault (the story\'s main character) is presented in such a way that you can hardly blame him for his actions (for which he is ultimately sentenced to death), and you find yourself thinking like an existentialist (sympathizer, at the very least).  In reading, I could see how society was constantly trying to impose a morality on the character based on the (unproven) idea that there is a god and that life holds some kind of deep meaning (if only in-reference to an afterlife).  It\'s interesting, really.  Meursault was only being honest, and yet nobody would believe that he put his mother in a nursing home or was able to kick up an affair with a woman the day after his mother\'s death, simply because he felt they no longer had anything to say to each other (he and his mother) and that he was ready to move on in his life.  Only evil people behave with such utter absence of humanity, so they said.    I would like to think that I would have given young Meursault the benefit of the doubt, if I had known him in real life.  But I knew a man just like that once, only he was named George, and really, I just thought George was an asshole.  I would like to have the chance to know George now, though, because I would like to think that I could let him live his life and think in his way and not internalize everthing (which is usually what the problem is - internalizing - when people can\'t just let others live and have their opinions without taking issue with everything).  And not that I advocate rampant murder or anything, but there are degrees of personal freedom, free of judgement, that I think humans should be entitled to.  Yeah, so, reading.  That\'s been good.  I\'ve since nuzzled up with Kerouac\'s \"Big Sur\" for the second time around...if only for the pure pleasure of reading frenzied words put together in the strangest and yet most satisfying of ways.    One more week until school starts...I\'m trying to soak it all in and slowly savor it...lick, instead of bite my way to the center of the Tootsie Roll Pop.   '),(7416,669116,'13,June,2003','2003-06-13 00:00:00','    Ok, so I just wrote a half hour\'s worth of my continuing adventures, and it all went away--disappeared into the netherworlds of cyberia.     Poopie Poopie.  I\'ll do it again soon, but I don\'t have the wherewithal right now, so I\'ll just give a brief summary:  Still in Rome.  Saw the Collosseum  Saw Vatican City  (mmm, saw Michaelangelo\'s \"Pieta,\" which made me cry)  Saw a free concert of Two Tenors outside the Collosseum last night, which will be aired on PBS later this month  Leaving today, destination yet-to-be-determined  Am the happiest girl in the whole world :)  I wish you all well on the homefront!   '),(7417,669116,'10,June,2003','2003-06-10 00:00:00','    Where to start?  At the beginning...  One of the last things I said to my friends at the bus station in San Jose was that I was tired and hoped I didn\'t end up sitting next to a \"talker.\"  Well, in perfect universal irony, the young man who sat next to me pulled out a little device on which he typed a message to me: \"Hi, I\'m Nick.  I hope it\'s okay that I sit here.  I\'m deaf...have you ever used one of these devices [referring to the thing he was using to type]?\"  I should have been more specific; should have said \"communicator,\" but all in all it was good.  We talked on that thing, and two hours later, I made an attempt at sleep, but it seems I can\'t sleep on moving vehicles these days.  So I spent a few days with my brother in Hollywood, which was great.  Last Friday night he was honored on the field at Dodger Stadium for his perfect usher attendance (and his dashing good looks).  I arrived roughly 8 seconds too late to see his 30-foot mug displayed on the Diamond Vision screen, but it was cool all the same.    After a few days of playing with him and his Super Cool group of friends, I set out for my adventure.  Sunday night: no sleep on an overnight flight to New York.  Monday: eleven hours wandering in midtown Manhattan.  Nice.  Monday night: no sleep on an overnight flight to Rome.    First lesson in Italian men: don\'t trust Italian men who warn you about Italian men.  I was talking to this man while waiting to board.  He lives in San Diego and was returing to Rome to visit his parents.  Well, that was his story anyway.  He was warning me about the men here (I\'m writing from Rome) and how they will tell you anything and blah blah blah.  As we boarded, he was telling me about the perfume he bought for his mother.  A few minutes later, he brought up the perfume again (one called \'Paris\'), and when I told him that I didn\'t know anything about perfume, he said, verbatim, \" I don\'t usually buy the new one, but sometime my wife, I mean, no my wife, my mother...she like to...\"  I didn\'t hear the rest because I had walked away at that point.  When somebody lies to you within the first 4 and a half minutes of meeting, I figure it\'s a good idea to move on.  So.  Rome.  This morning I thought I lost my purse first thing upon arriving.  I was ready (sleep-deprived and a little nervous anyway) to pack up and head home.  Turns out I left it on the plane.    So I\'m staying then :)  It\'s true about the Vespas...those and other kinds of scooters are everywhere.  It\'s pretty charming.   It\'s true about the Italian men, though I haven\'t found them particularly threatening or anything.  And Rome is just teeming with handsome fellas.  But then I\'ve always liked the dark, curly-haired types anyway.  I had a long non-conversation with an elderly German man outside a cafe over my first cafe latte (which was everything I dreamed--the latte, not the non-conversation).  Lesson learned (I don\'t speak German and he didn\'t speak English, Spanish, or French...neither of us speak Italian): European phrasebook is next to useless.  Now, it\'s time to get some sleep.  I\'m curious about the hotel in which I\'m staying.  It\'s called \"Bed and Breakfast\" (the hostels were full).  The man showed me the room and told me about the keys and the bathroom, and when he was fixin\' to leave, I asked him how to go about breakfast.  His reply was \"Oh...we don\'t serve breakfast; it\'s just called \"Bed and Breakfast.\"  I repeated:  \"You\'re called \"Bed and Breakfast\" but you don\'t serve breakfast?!\", which he answered with an apologetic shrug and a promise to bring me some coffee and juice in the morning.  At this point I\'m sort of wondering if it really is a hotel at all :)  As far as I can tell, mine is the only room there.  Hmmm.  In summary, so far, so good.  But Rome is expensive.  I think I\'ll be leaving for some quieter, less touristy realms tomorrow.  Hope everything is great on the homefront.   '),(7418,669116,'16,June,2003','2003-06-16 00:00:00','    Last I wrote I had just been to the Collosseum and Vatican City.  Three things I can say about those places:    1) Breathtaking (seriously, I haven\'t breathed in like 5 days).   2) Big (really really really big).   3) Germans!  (Soooooo many Germans).  I\'m summarizing of course.  I\'ll just have to share the pictures and tales in detail when I get home.  Where is home, anyway?  I was thinking about this.  I have no place to live, no job, nothing tying me to anywhere.  Maybe I\'ll flip a coin and pick a city when I get back to the States.  But anyway, back to Italy...  I was gonna scrap my plans to see Malta the day I left Rome, but then I got an e-mail from my friend Nick, who was down there, so I decided to check it out.    Well, that was the easy part.    First, I decided to buy a train ticket to Reggio di Calabria, in the very south of Italy.  It seemed like a port city and was about as close to Malta as I could get.  There were many trains leaving, but most of them arrived late at night.  So I decided to take an overnight train and get there in the morning.  Leaving me with 10 hours more in Rome.  Oh, but first things first---I couldn\'t use my ATM Visa card at the train station.  Turns out there was a fraud alert placed on it.  Apparently a sudden 2-Euro purchase at a grocery store in Rome causes some alarm.  So I got that straightened out first.  I ran into a tourist office employee named Amadeo, who offered to show me some more sights.  I\'m pretty sure this was not part of his job; I suspect he may have been hitting on me under the guise of \"helpful tourist office employee,\" but he wasn\'t imposing or shifty or anything.  So I took a crazy busride with him to the Fountain at Trevi, The Spanish Steps, a charming little square called Campo di Fiori, and the Pantheon.    I really don\'t think I can describe accurately how enchanting it felt to be amidst so much history.  So, so, so beautiful.  Like I said, we\'ll have to talk over a slide show in the future.  But right across from the Pantheon, in the same little square, sits a McDonald\'s.  Sick.  Really sick.  But I peaked in at the menu and saw a colorful picture of a salad with ripe red  Roma tomatoes and chunks of Mozzarella cheese on it.  When in Rome...I guess.  Amadeo (along with every guidebook I\'d consulted) warned me not to fall asleep on the train if I wanted to have my luggage in the morning.  Ok.  I lucked out an ended up in a car with 4 natives of Reggio di Calabria--one young girl, one crazy old man (I\'m not just saying that--even the other Italians exchanged \"this old guy is crazy\" looks.  I had to pay close attention to these things because I was now entering the region of Italian-only speakers--body language is everything.  That and cherades.), and one young married couple.  None of them seemed eager to rob anybody, and I was sitting far from the passageway (where the thieves roam, I suppose), so I felt safe.    The young married couple was in a fight all night, though, which made me a little sad and uncomfortable.  See, it wasn\'t a vocal, let\'s-have-it-out-and-be-done-with-it fight.  I mean, it couldn\'t be: there were four other people in the room.  It was a slow, quiet, awkward, man-touches-knee-of-woman (apologizing and speaking in hushed tones), woman-gives-man-the-cold-shoulder-and-faces-the-window, man-gets-up-and-paces, woman-pretends-she\'s-asleep, man-comes-back-to-seat-and-tries-again, woman-gives-the-evil-eye, etc., etc. kind of fight.  They were friends again by the time we stepped off the train, but still, I was feeling for both of them for the first 6 hours of the ride.  So then.  7:00 a.m.  Saturday.  Reggio di Calabria.  I asked a woman at the train station how to get to Malta (figuring people there must go to Malta all the time), and she sold me another train ticket.  I figured this train must go to the port (the ticket was only 1Euro).    I was wrong.  When I figured out I was wrong (luckily, before I actually boarded the train--turns out there\'s a very nearby town called Malta), I asked her how people get to Malta, the country.  She referred me to the travel agency.    Closed.  It was Saturday--who would be traveling or making plans to travel on a Saturday?  Lesson 1 regarding the south of Italy.   But Reggio was quiet and wonderful, so I wouldn\'t have minded staying at all.  I did want to get to the port, though, just to know what my options were.  And here\'s where it gets interesting...  I found the bus to the port, which looks nothing like the kind of port where anything other that huge containers of things would be shipped from.  I certainly didn\'t see any family members waiting around to greet arriving passengers.  It was something like Long Beach Harbor, but smaller.  There I was referred to a dim, dingy little office where the man in charge of \"Sea Malta\" was sitting, smoking a cigar.  I know, it sounds like a movie, but this whole sequence felt that way for me, so it\'s fitting.  He told me that \"Sea Malta\" was a cargo ship, and that I could maybe take it to Malta, but that the priority passengers were Maltese truck drivers and the space was limited.  Don\'t freak out, Ma.  I promise it all turned out well.  He said to come back at 1:00 in the afternoon to find out if there was room for me.  The other option was to take a passenger ship the next day that arrived in Malta at 2 in the morning.  Whatta dilemma!  I spent a few hours roaming around and trying to communicate with the incredibly friendly people of Reggio di Calabria.  Then back to the port.  The man said that there was room, and he took my 55 Euros (roughly 75 dollars or so), along with my passport.  He made a copy of the passport, but when I asked for it back, he said I\'d get it back later.  I squinted and thought about it, and when no female intuition alarms went off, I said \"ok,\" and left.  The boat was leaving at 7, and he said to be back by 5.  So I went back into town and drank some sort of coconut drink, wrote a quick e-mail to Nick, and went back to the port.    At 3:30 I went back to the office, which was then closed.  I wasn\'t that surprise, since he had told me to go directly to the ship, but this was a crucial sort of moment.    Here I was, walking along this port with my big backpack, sweating like a madwoman.  These were my thoughts:  Random American girl, really out-of-place in a quiet cargo port. Extreme south of Italy. Speaks no Italian. Gave a stranger her passport and 55 Euros. Trusts that the man really does have something to do with this ship. Trusts the ship really is going to Malta that evening. Trusts they\'ll let her on the ship. Trusts nothing bad will happen to her on the ship. Trusts she\'ll get her passport back.  Needless to say, I felt a little apprehensive.  When I got to the ship, there were nothing but men standing around all over the place.  They looked at me like I was a leprechaun.  I suppose I seemed a little out-of-place to them, too.  I picked the one who looked closest to my age and asked him what to do (hand motioned, really) if I was supposed to take that ship.  He pointed to the steps and told me to get aboard.    \"Okaaaaay.  Grazie.\" (Said, turning slowly toward the ship with raised eyebrow)  Imagine my relief when, walking up the rickety steps, I noticed a 40\'s-something blonde woman on board.  I spoke to her and her husband, and all my fears dissolved instantly.  She was from Sweden and her husband from Malta, and their 18-year-old son was there, too.  She said they take the boat all the time, and that it\'s very safe and I\'d have my own sleeper room with a lock and the use of a shower, dinner provided, and breakfast in the morning.  It was all true.    There were only 3 actual truckers on board, and they were the friendliest people I\'d met so far.  We all (the truckers and the family and I) had dinner together (pasta, the best food I\'d eaten since I arrived), drank wine, and told stories (they all spoke English--English is one of the two official languages of Malta), and I had the soundest sleep I\'d had in a week.  It was a random, rare, and beautiful experience.    The Mediterranean, glowing under the light of a near-full moon--just imagine.    And it turns out they needed to hold onto the passport for the Maltese customs agents to have a gander at when we arrived.  So, like I said... Hi.  I\'m in Malta.  And more on Malta later.  For now I\'ll just say that everybody should try to make it here at some point.  It\'s a gem of a place, and a pretty well-kept secret (except from the Germans--they seem to have found out about it :)    I\'m off to collect my laundry, and then to the beach.  Greetings to all my friends and family.  I miss and love you all.  And I\'m still the happiest girl in the world :)   '),(7419,669116,'16,June,2003','2003-06-16 00:00:00','    Hi.  I\'m in Malta.  I had to say it like that because reporting this news reminds me of that scene in Wayne\'s World where Wayne and Garth are playing with the blue screen: \"Hi..I\'m in New York.  I\'ve got a gun, let\'s go to a Broadway show.\"  Then they\'re in Texas and I think Hawaii and, all of a sudden, Delaware.  And Wayne says, \"Hi.  I\'m in Delware.\"  Now, I\'m not saying it reminds me of that because Malta is boring or anything (quite the contrary--it\'s amazing), but just because who knows anything about Malta?  Really, I mean, I didn\'t know anything about it before...Maltese falcon?  Maltese is a kind of dog?  What else?  I was talking to somebody here who used to play on Malta\'s national futbol team, and he said that when the team was playing in New York, the chant that the U.S. fans liked was \"Where the hell is Malta?!\"  Anyway, I\'ll get back to Malta later...first, some back-tracking.  I\'m going to post this first, though, just to be safe.   '),(7420,669116,'25,June,2003','2003-06-25 00:00:00','    As promised, I would like to invite you into the world of the Savoy Guesthouse.  First, a bit about the house itself.  There are three floors and something like 14 rooms.  My room is on the second floor, where it\'s warm, but not as warm as the third floor, so I lucked out.  There is a little rooftop terrrace where I can go at night and look at the stars above or the people below.  I can also hear, for example, the sounds of neighbors\' domestic disputes, and a strange little bird warble I\'ve never heard before.  I\'ve yet to see the actual bird, and I think it\'s more interesting that way.  My room has tile flooring, a double bed, and a window that overlooks a busy street, where there\'s a crosswalk just perfect for getting killed while crossing (the crosswalk is right at a curve in the street, where cars come a-flyin\').  Mine is also one of the few rooms with its own shower, for which I sacrificed having a kitchenette.  The kitchenette would have been nice, but I\'ve been grateful for the shower since I realized that a pervy old male resident named Tony likes to check out (up close and personally) the laundry of the women who use the rooftop clothesline; washing and drying my clothes in my private shower has afforded a nice alternative.  There is a common room where people watch T.V.... about 6 channels in Italian, 2 or 3 in Maltese, and one sometimes showing American series (series\'? serieses?) in English.    The plumbing, due to the building\'s age, is, um, a bit of a gamble.  Everytime I flush and the toilet actually works, I feel like I\'ve won the lottery.  Also due to this temperamental plumbing, I never know exactly what I\'m going to find in the toilet, courtesy of the last visitor, who may not have been so lucky with the plumbing lottery.  I\'ve had a few frights in this situation, unable to  imagine  what said non-flusher might have consumed or imbibed the night before to have produced what was left in the toilet for me to find.  The window in my room is best left open at night (as the afternoon sun makes my room pretty stuffy in the late afternoon), but leaving it open means letting in the mosquitos, and I have a very bad relationship with mosquitos.  Last week, after declaring an all-out war with one particularly robust pest, I witnessed the emergence of my alter-ego, which was quite disturbing.  See, I don\'t usually kill bugs.  I\'m more one for ushering them out, or ignoring them.  But you know, with mosquitos it\'s an \"us or them\" sort of thing--\"eat [or \"kill\" in this case] or be eaten.\"  So I chose \"kill.\"  Now, this mosquito had an admirably determined will to live; I had thought I killed it twice, when I saw it land on my bedspread, on my leg.  I was just about to turn the light out and go to sleep, so I knew that this was the Moment of Truth.  He would have had his way with me all night long.  So I moved my hand slowly toward him and threw my hand down in a quick and violent SLAP!!  Then (and this is where my dark side exposed itself) I actually uttered the words...  \"YES!!!  DIE, M@therf*cker!!!\"  Mygoonus!  Then, in what I\'m sure was my universal lesson for the day, I became aware of a warm and throbbing pain coming from my leg.  See, I had forgotten, in my focusing on KILLING, that I had a fresh sunburn, and I had struck my leg hard enough to keep me awake with the sore, radiating effects of the slap for another 25 minutes.  Tisk, tisk, tisk.  We\'ll work on ushering the little buggers out from now on, eh?  But anyway, enough about my room and my evening death-related activities.  I want to introduce you to the players...the Savoy Guesthouse Horror Picture Show.  I use \"Horror\" in jest.  There is nothing terrible about my fellow residents (well, most of them, anyway), but they are indeed interesting.  First, I\'ll start with the house Mamas--two sisters who run the place and who keep everyone reasonably in-line.  The first is Josephine, a buxom sort of Mother Hen with a wicked dry sense of humor.  It took my a while to plug into this humor, and a few times I really thought she was upset with me.  For example, the other day I was talking with my friend in the common room, and I turned the T.V. off because nobody was watching it (or so I thought).  Josephine emerged from her little office, where she had been talking on the phone, and said, \"WHO turned off my T.V.?!!!!\"  I said, \"Oh, I\'m sorry.  I did.  I didn\'t think anybody was watching it.\"  She replied with, \"Well,  I  was watching it.\"  Then, looking directly at me with raised eyebrow, she said, \"It was an AMERICAN film,\" and walked off in a huff.  A couple minutes later I realized she was joking when my friend pointed out that she had actually been watching the news.  But despite her sometimes ambiguous comments, she is clearly a caring woman who always remembers to give messages, never goes snooping through rooms, always remembers to bring a fresh towel, and likes to ask about the plans you may have mentioned having for the previous evening...\"How was the jazz?\"...\"Did you have fun on the island of Gozo?\"  Her sister, Lina, is a bit more military-like in nature and is less likely to chew the fat.  In the afternoons, she sits watching T.V. with her boyfriend Charles, a red-faced Maltese man who\'s been living in San Bruno, California for 44 years or so.  He comes to visit her every summer and watch T.V.  They are both cordial and pleasant enough, though definitely more distant than Josephine.  I would never refer to Lina as a \"Mother Hen.\"  But both sisters stir up a mean carafe of instant coffee in the mornings--my choice meal after I fast grew tired of the standard breakfast fare...corn flakes and milk, orange juice, and bread with marmalade.  I bought my own wheat bisquit sort of things (called \"digestives\"(?)) and usually have a few of those as well.  Incidentally, I wanted to say a bit about the milk here in Malta because I promise it\'s unlike anything you\'ve ever experienced.  Okay, you know how thick all-fat, Vitamin D milk is in the States?  Well multiply that by two, at least.  The milk here crawls down your throat at a pace that might make you feel like gagging, or chasing it with something like cranberry juice to clear the canals.  But if it\'s cold enough (which is rare), it can be especially satisfying and take the place of a three-course meal.  But we were talking about the residents...  There\'s Iris, a young Chinese student studying English here.  She has a wonderfully soft and most melodic voice, and when I told her this, she said that she\'d studied voice (as in, singing) in China and had trained her voice to be that way.  Cool.  I asked if she would mind giving me a little concert sometime, which she answered with the words, \"Yes, I will sing to make you feel happy.\"  Isn\'t that awesome?  There are two other young Chinese students who are also studying English but are loathe to actually speak it.  The only conversation I had with one of them concerned American movies and basketball, during which the young man (Da) declared his favorites to be \"Bread PEE\" (Brad Pitt), \"SUHdra Block\" (Sandra Bullock), and \"SHUH-cle OH null\" (Shaquille O\'Neal).  Please understand I\'m not making fun of his accent, just trying to give an idea of why I had to strain and stretch my ears in my desire to have a conversation with him.  He was nice, but I fear the fact that the two share a room and never leave the guesthouse is keeping them from learning much English while they\'re here.  Da told me that they must pass Standard English exams next month in order to be admitted to the University of Malta.  Hmmm.  Oh, one other thing he told me was that, growing up in China, one of the few American series he saw was \"Growing Pains.\"  It took a while to determine this was the show he was talking about, finally becoming clear when he mentioned a \'Michael\' who was always getting into trouble.  Turns out Da\'s dad had always wanted a son more interesting, less shy than Da himself, and continually asked Da why he couldn\'t have been more like Michael from \"Growing Pains\" and at least do something  interesting  once in a while.  That broke my heart.  There\'s an Egyptian man named Mario who brews beer in his room and wears a near-constant scowl on his face.  The only time he ever spoke to me, the conversation went like this:   Me : Hello  Mario : Hello [ heading up the stairs and then retreating to add to this initial response ].  Are you Jewish?  Me : Me? [ a line from \"The Princess Bride\" coming to mind: \"Do you always begin conversations this way?\"  ]  Mario : Yes, are you Jewish?  Me : Do you mean religion or race?  Mario : Religion  Me : No  Mario : But your parents, are  they  Jewish?  Me : No, I was raised Catholic  Mario : Hmm. [turning, without another word, toward his room]  So.  That\'s Mario, as far as I know him.  There\'s an older British man and a young Russian girl, neither of whom I\'ve seen enough to have learned their names or anything about their situations.  There\'s an older Maltese woman named Anne, whose habits and the incredibly lived-in state of whose room would make you believe she\'s been living in the same guesthouse, piling the same tangerine marmalade on her bread in the morning, for the past 37 1/2 years.  In the afternoons, she can be found in the foyer, where she serves as the house\'s first line-of-defense, or last barrier-to-your-exit, depending on if you\'re coming or going:  \"Where are you going?\"  \"To the beach?  Which one?\"  \"When will you come back?\"  \"Who are you going with?\"  \"Where did you go?\"  \"What did you do there?\"   This drives some of the guests absolutely crazy, and I think Anne is reason #1 for Nick\'s refering to the Savoy as a \"halfway house,\" but the woman is harmless and actually very sweet.  She\'s just a bit inquisitive, which I think is merely due to the fact that she is, I think, somewhat physically unable to leave the house and so lives vicariously through the other residents.  I felt bad the other day because I was sitting near Anne in the foyer when I happened to glance down at her feet and caught sight of her breathtakingly overgrown toenails.  I, anticipating that maybe she didn\'t own a pair of nailclippers and thinking she might like to borrow mine, embarked on this conversation:   Me : Anne, I noticed your toenails are a little long.  Do you...  Anne :  I know, I know.  I\'m sorry.  I was gonna cut them last night and then I forgot.  Me :  No, no, I was just wondering if you needed to borrow some clippers.  Anne : No, I have them.  I\'m sorry.  I\'ll cut them tonight.  I will cut them tonight.  Me :  I don\'t care, I mean, it doesn\'t bother me.  I was just gonna offer you mine.  Anne :  No, I have them, but it\'s just a little difficult for me to cut them.  So then I was thinking maybe she needed help cutting them, and the conversation continued:   Me : Oh, do you need help?  I mean, I could help you if you wanted.  Anne :  It\'s okay, and I promise, I\'ll cut them tonight.  I\'ll cut them tonight.  I felt really bad at this point because I could see that Anne couldn\'t understand that I was just concerned.  She thought that I was disgusted by her toenails and was trying to hint/guilt her into cutting them.  The more I said, the more I felt that she was misunderstanding me and that I was making her feel bad.  So, I dropped the toenail subject.  Now, I\'ve neglected to mention yet that Iris (the young Chinese girl) had been sitting in the foyer also, but she wasn\'t listening to us; she had been watching T.V.  I\'m not kidding...five minutes later, during a commercial, she gingerly glanced around the room, looked casually at the two of us, caught sight of Anne\'s feet, and said (in the harsh and unapologetic way that only English-as-a-second-language speakers do accidentally), \"Are you going to cut your toenails?!\"  Oh my.  The whole conversation started all over again, Anne becoming more and more convinced, I\'m sure, that her toenails were an affront to the civilized world.    The next morning, I walked into the breakfast room and the first thing Anne said to me was, \"I forgot to cut my toenails again, but I\'ll cut them tonight.  I\'ll cut them tonight,\" shaking her head absolutely in betrayal of the fact that she was not only trying to convince me of this intent, but to convince herself as well.  And that\'s Anne.  Then there are Tony and Eddie, both elderly long-term residents who hover on the edge of eviction every single day.  See, Josephine\'s problem with Eddie is that he begins drinking at 11:00 every morning and doesn\'t stop until he\'s consumed an entire bottle of Auld Lang Syne Scotch Whiskey.  His only defense to this beef is to point out that Tony begins drinking at 5:00 in the morning and doesn\'t stop until he\'s consumed  two  entire bottles of Auld Lang Syne Scotch Whiskey.  Both arguments are true, and the entire hallway near these men\'s rooms reeks terribly of sour alcohol breath.  Tony is Maltese, but he lived 44 years in England, where he was married to a woman from India.  Tony fled the country when his wife died and her son (from a previous marriage) tried to kill him for attempting to incinerate her body in compliance with her wishes.  So he brought her dead body and all to Malta and has been living in something like exile for the past who knows how long, in the Savoy Guesthouse.  Now, this lamentable tale might cause you to take pity on the old man, but nay, I say.  You should hear the rest first.  You might remember my having referred to Tony earlier as the \"pervy old man\" who checks out women\'s clothing on the clothesline.  Well that\'s true of him.  He is also given to offering hourly work to female guesthouse residents.  And he sees absolutely NOTHING wrong with this.  On the day he offered this kind of work to me (as if it were some kind of favor), and I became angry and yelled for the first time in I have no idea how long, he asked, very innocently, wherein lied the problem.  \"It\'s good money,\" he reasoned, \"Ten Lira per hour.\"    Ten Lira is roughly 27 American dollars.  Eddie is alternately disgusted with Tony and is his best friend.  That afternoon, after Tony had been asked to leave the room, Eddie denounced him as a scoundrel, a truly reprehensible human being, a waste of good oxygen.  Moments later, Tony reappeared in the doorway, having forgotten his glass of whiskey.  And when he left, Eddie said, \"Ok, I\'ll come see you later.\"  I throw my hands up.  Eddie is from Pakistan, where he was forced out because of an inability to find work, having suffered from prejudice associated with his being a Christian.  He claims to be a palm reader, and has offered to read my palm but failed to deliver on this promise so far.  Until now, he\'s only shared with me that my lucky day is Wednesday, and that things I begin on Wednesday will have good outcomes.  Good to know.  Eddie also writes jokes (or joke, as far as I know--so far he\'s shared the same one with me three times).  The example:       A bartender closes his bar and goes home to sleep.  A little while later he\'s awakened by the phone, and the man on the other line asks when he\'ll be reopening in the morning, to which he answers, \"Sometime around 10:00 am.\"  A couple of hours later, the man calls back, wanting to know EXACTLY when he\'ll be opening, and the bartender repeats, \"I told you!  Around 10:00.\"  By the third call, the bartender is irate, and when the man asks again for the exact time, he answers with, \"You bloody bastard!  You\'ve ruined my night!  If you\'re so bloody fond of drinking, why don\'t you buy a case of liquor and drink it at home?!\"  The voice on the other end answers, \"Look man, I don\'t drink.  I promise.  I just came into your bar to use the restroom and now I\'m locked inside until you reopen.\"    Now, it\'s difficult to find that joke funny more than once.  But watching Eddie\'s face--the way he cracks up at his own cleverness--makes it worth waiting for the punchline.  The other day Eddie tried to pay his rent and Josephine told him to keep his money because she was kicking him out.  In a panic, he called Tony downstairs so the two together, whiskey bottles in hands, could brainstorm a solution to this new problem.  After a couple of hours of careful discussion and plotting, their answer was to send Tony in to speak to Josephine.  Slowly, cunningly, he crept into her office, his lines planned.  She took one look at him and said only these two words: \"You\'re next.\"  Ouch.  But it\'s not true.  They\'ll probably both die there.  I think she just likes to scare them into compliance...she\'s told me how happy she\'s been with the silence and the cleanliness of rooms she has observed since she made her threats.   The one guest I\'ve yet to discuss, Saad, refers to them as Tom and Jerry, but we can\'t really figure out who\'s who.  Saad is a young man from Morocco who\'s here to find work.  He\'s full of insane tales about Morocco and what he refers to as \"magic.\"  He also talks a lot about \"phantoms,\" which he pronounces \"FAWN-tomes,\" in a way that makes them sound truly scary and almost believable.  He claims this magic has been responsible (in his eye-witness experience) for: * A young man spelling an American woman into falling in love with him, her taking him back to the States where they still live. * His mortal enemy coming to his house, weeping, and waiting with his mother for three hours until Saad came home, at which time the enemy apologized for all past transgressions. * Same said enemy coming to and wondering what the hell she was doing in his house, after Saad asked the spell-casting third party to remove the magic. * A young newlywed being unable to do the deed on the wedding night, thanks to the magic performed by the scores of young bachlorettes in the village who were jealous and angered at his finally getting hitched. * Same said newlywed impregnating his wife when Saad\'s magical grandmother intervened with her White Magic, on his behalf.  * A young man who moved to Libya to work being sometimes given to intense and overwhelming desires to return home to his girlfriend, who made a magic paper and tied it to the leaves of a tree so that the magic would make his mind change with the wind.  !!!   It\'s funny, though...if you listen to him tell these stories, they sound true.  Do you think that magic is merely a matter of believing in it?  It seems to me that might be a contributing factor.  All in all, these characters have made the stay entertaining and well worth the whopping fifteen bucks a night I\'m paying to stay there :)  And that, my friends, is the Savoy Guesthouse.  Ya\'ll stop by sometime, y\'hear?   '),(7421,669116,'22,June,2003','2003-06-22 00:00:00','    Patience, my friends...I fear this entry will be loooooong.  I\'ve had the details floating around in my mind for exactly one week.  Upon arriving at the port in Malta (pronounced MOLE-dtuh by the locals--go ahead and say it aloud; I know you want to), I took a bus to my guesthouse, which sits atop a curvy hill, overlooking, among other things, the nearby Mediterranean Sea), I called Nick to let him know I was here and in good spirits, and he came right over.  Fifteen seconds of greetings later, he asked me if I was ready to go to church.  CHURCH?!    I checked my face and clothing for signs of blantant heathenery, but he explained that it was the first stop on our tour--he had just learned more than a bit about the history of the building and was eager to make me his first victim, er, audience.  And Saint John\'s Cathedral, in Valetta, was worth the journey.  There are 350 valiant knights buried below the marble floor, and the ornate details of the interior (which is well belied by the unassuming exterior) rivaled only Saint Peter\'s in Vatican City.  Okay, indulge me...I have a limited frame-of-reference.   Though I am no longer religious, I have to say that there is something very comforting in being able to attend a mass anywhere in the world and know when to stand, when to sit, and when to sit quietly and brace yourself for a long-winded homily...  In Maltese!  Needless to say, I didn\'t understand a single word, but the enchanting voices of the choir members (a real choir, which is difficult to find at Catholic churches in the States) reached down and took a firm grip on the innermost knot in the bottom of my soul (I had been wondering where it was (the bottom of my soul, that is...it\'s good to know).  Then, a walking tour.  We checked out the ENORMOUS fortifications built, uh, a very long time ago (sorry Nick, you\'re a better guide than I am a student) to keep the Turks out.  Damned Turks!  Always trying to sneak in somewhere.    When you walk up and down the narrow streets of Malta, your shoes make the sound that shoes make in British documentaries.  Do you know the ones I\'m referring to?  In these documentaries, there\'s some old man, always walking UP a hill for some reason.  So between heavy breaths, he says something like (think high British pronunciation), \"For the Egyptians, mumification ensured the soul\'s being kept in-tact during the afterlife,\" and all the while, he continues to walk and his shoes, on the limestone/sand combination, make a cool kind of crunching sound.  Do you have the slightest idea what I\'m talking about?  Well, the streets of Malta make that sound.  When talking to myself as I walk (one of my favorite passtimes), I always have to affect a British accent; it seems so apropos.  But, of course, one needn\'t walk everywhere.  There are busses...if you dare, mwah ha ha.  Well, actually, the busses themselves seem entirely safe.  The problem is me.  See, they drive on the left here, and I\'ve been nearly killed at least a half-dozen times.  Guess my learning curve is less-than-optimal.    The busses are relic \"gifts\" from England, many of them dating from sometime in the 1950\'s, chrome fenders and all.  Their interiors are best described using a phrase I don\'t think I\'ve ever uttered in my life: A real hoot.  There are religious stickers plastered all over the front dashboards and window panelings (many featuring images of a glowing Virgin Mary and the like).  Some of the more modern-looking stickers read things like \"Think GOD,\" or \"Jesus Loves Me,\" but the most intriguing one so far read \"I Love Safely.\"  Now, I don\'t know if this was a typo, meant to say \"I Love Safety,\" or if the driver of the bus was especially careful about the way in which he loves and wanted to make sure all his passengers knew this about him (in my feeble understanding, public transportation is handled in a private-contracted sort of way--each driver owning and (sometimes) caring for his own vehicle), but anyway that\'s what it said.  Which brings me to the language discussion.  As I wrote before, English is one of the two official languages here (three cheers for British Imperialism and the Knights of Saint John), but that doesn\'t remove the possibility of encountering some idiosyncratic discrepancies.    My favorite sign so far reads \"Andrew\'s Beefy Crunch.\"  I\'ve never seen the establishment open, but I\'m eager to find out what exactly is sold there.    Another, written on the side of a mini-bus, read, \"Stuff Your Jealousy.\"  Okay, I\'ll get right on that.    A shirt worn by a local man: \"International Delivery.\"    The Maltese speak with a charming accent that sounds like a combination of the accents you would here from native Arabic and native Spanish speakers, which is fitting considering the island\'s location.  The worst thing I can say about Malta (which is a relatively welcoming place, otherwise) is about the service, which is probably the worst I\'ve ever encountered (seriously, you thought Original Joe\'s was bad!)  I don\'t know if this is owing to a general and overwhelming annoyance with tourists or the effects of the weather (oppressive sun with a healthy dose of wet--wet air, not actual rain), but SOMEthing has crawled under the skin of every person in the service industry here (and, I suspect, laid eggs).    An example: The other day I was sitting at a cafe, writing in a notebook, when my server (an elderly and sour-faced man) came to take my plate away.  I realized, as he was taking it, that my notebook had been resting in the ketchup on the plate\'s edge.  I made an \"Aww\" kind of face...as in, \"how unfortunate,\" which he misinterpreted as my assigning some kind of blame to him and to which he responded by pointing an extended forefinger in my face, followed with the words, \"It\'s YOUR fault!!\"  Hot Damn!  But, as cultural differences can be charming, I\'ve found eliciting a smile from a server to be a challenging and, when successful, gratifying way to pass the time.  And speaking of food (because we kind of were), I\'ll share what I\'ve learned of it so far.  Well, they have rabbit (stewed), rabbit (fried), and rabbit (baked).  The end.  Kidding of course, though rabbit is a favorite when it comes to local cuisine.  Unfortunately, it seems the Maltese (probably having something to do with the British occupation I referred to earlier) suffer from a bit of an identity crisis.  It is difficult to determine what true Maltese culture is all about.  So, my choice vegetarian option is chips and egg (how much more British can you get?), followed by couscous and pita bread from the Istanbul Kebab joint near my guesthouse (I guess some of those Turks snuck through after all).  And then, there\'s (surprise, surprise) McDonald\'s, Burger King, and Pizza Hut.  The notable menu item in a Maltese McDonald\'s: salad with fresh tuna or shrimp.  And Italian gelatto has meandered its way here, as well...  There IS a God!  And now for the best thing I can report about Malta: It is--by far--the safest place I have ever been.  Locals and tourists alike muse about the fact that anybody (man, woman, child) can walk down the dark, deserted streets at any time of the night and feel completely at-ease.  Do you think it has anything to do with the absence of a National Rifle Association?  Naaaaaa.  Really, though.  It\'s a nice change from the dagger-like sideways glances I had grown accustomed to throwing during my stay in Rome.  Oh, and while I\'m thinking about Rome...I wanted to remove any remaining doubt about my being a bona fide freak from the minds of my friends and family members, by making a side-by-side comparison of....drumroll please....  Body Odor!!  Now, the overcrowded busses and Metro of Rome afforded me with countless opportunities to (connoisseur that I am) take a whiff, sampling the smells of Italy (or, Italians).  Assessment: pleasant, in that horse maneur kind of way.  You know what I mean?  How B.O. can sometimes be not-so-bad?  My friend Nicole (thanks to lessons learned in her Human Sexuality class) would point out that our attractions to another\'s sweat smell has something to do with the likelihood of that person\'s immune system, coupled with our own, producing offspring with healthy immune systems (ain\'t that cool?).  Assessment of Maltese B.O.: more offensive.  Is it the proximity to the salty sea air?  I don\'t know, but the smell is very bitter, most unwelcoming.  So, my belief in this immune system thing has been thrown into doubt.  I mean, it couldn\'t be that I would produce healthy offspring with EVERY single Roman...afterall, the women have this smell, too.  And there\'s no way my coupling with any native of this island would result in weak little sicklings.  Or is there?    But anyway.  I fear I haven\'t really given a good description of this place.  Where to go from here?  On the boat here, I was told that Paceville (PAH-chuh-vill), the local hot spot here, offered the best nightlife to be found in the modern world.  Though I\'m not really one for \"nightlife,\" I figured I\'d check it out, especially after I learned there was a Salsa club.  After days of loafing around in the sun, I was eager to expend some energy.  So.  Paceville.  Paceville is a nightmare.  Well, I suppose that if I were 17, or drunk, or eager to spend too much money to go to a club exactly like any club I could find in the United States, it would be quite a thrill.  But since I\'m none of the above...  I went to the Salsa Club (Fuego, it\'s called), and was happy when I entered and heard a merengue song, followed by a salsa.  But then, suddenly, the tides turned, and the next 6 songs were (sometimes) in Spanish, but with nothing even close to a salsa beat anywhere to be found.  Most were just overlaid with a thump, thump, thumping just perfect for losing your mind to.  I guess that\'s the idea.  Oh, and a resident of the guesthouse where I\'m staying had kept mentioning to me a song that had the words \"Life is life\" as the main chorus.  \"Life is life?,\" I asked, \"What does that MEAN? That doesn\'t mean anything.\"  I heard that song that night...\"Life is life,\" and apparently it\'s quite popular, because after the \"Life is life\" part, the DJ would turn down the volume and, in one loud voice, the entire contents of the club would yell \"La la la la la,\" the rest of the chorus.  Is this a local phenomenon, or is my ignorance, like a poorly hemmed slip, showing again?  Please, if anybody knows anything about this, do set me straight.  Ah, but last night, in a wonderful stroke of luck, I discovered the local jazz club.  It sits underground in a sultry, red little room that seems like some sort of a secret, and like--among the patrons--there should be real live beatniks or maybe Boris and/or Natasha. Sadly, they only have jazz on Thursdays, but I\'ll be sure to be there.  As for daylife, I have only the beach to offer explanation of, but the beach is enough.  The great thing I can say about the beaches in Malta is that all are welcomed with open arms...big\'uns, lil\'uns, all types, all sizes, all in bikinis (well, the women anyway).  They are far less body-conscious here, and it was incredibly refreshing to see old and young, thin and not-so-thin, tawny bronze and ghostly white and beet red, all enjoying the sun as it should be enjoyed.  And the Maltese children, splashing and sand-castle-building and running and screaming, are especially nice to watch.  Though eating and entertainment are comparably priced to the U.S., accomodations are cheap, so I\'ve decided to stay a while before heading to Spain.  My purchases so far include the obligatory sarong (only $6 here, compared to the $15-20 you\'d pay at, say, the Jazz Festival) a sun hat (which I bought from the Maltese version of a dollar store--awesome!), and a traditional henna hand tatoo.   Despite bad service and yucky nightclubs, life here has treated me well.  Next time I write, I will tell all about the cast of characters at the Savoy Guest House, a motley crew of folks so strange at times that, in hearing the reports, Nick (in the classist and superior manner for which he is known) has taken to calling \"Halfway House\" rather than \"Guest House\" residents.  Our differing attitudes about what is \"uncivilized\" and what is simply interesting and quaint, has resulted in a general happiness (in both of us, I think) about the fact that our respective hotels are located on different parts of the island.  :)  La La La La La.  I wish you all well.   '),(7422,669116,'04,July,2003','2003-07-04 00:00:00','    Okay, I just had to make one amendment to my last entry...  It wasn´t the peeling out that finally put things in Malta over the top.  It was the Mormons.  Mormons on Malta!!!!!    But Barcelona, My Oh My!  I arrived yesterday and have decided it is the most beautiful city I´ve ever seen.  Yesterday I saw a Dali exhibit, which gave me a little peek into the brain of certainly one of the most imaginative and eccentric figures of the 20th Century.  You know, he had a real knack for depicting avant garde uses for human genitals in drawings.  Just kidding.  I mean, he did, but the exhibit was much more than that.  Today I went to a garden designed by Gaudi, the famous architect.  Another imaginative fellow.  There were a house and a few structures in the park that felt like something out of Smurf Village.  Well anyway it was dreamy and wonderful.  But last night, I experienced the best moment of my trip thus far.  I was walking through this little alleyway that passes by a really big, really amazing church, and there was a small opening where a solo cellist was playing (with a portable CD player nearby, playing a single piano as accomaniment).  The night was perfect (cool and a little cloudy), and the music floated out from the cello, danced mournfully through the air, and approached me with a vivid sort of incredible sort of wondrous sort of hesitance that was haunting and inviting at the same time.   I decided to sit down for a while.  And my reward for staying came in the form of ¨Ave Maria,¨about six songs later.  That song gets me EVERY SINGLE time, but most especially last night, alone in a little alleyway on a perfect night in Barcelona, Spain.  I decided that that moment alone--even if it would have been the *only* memorable moment during the entire trip--was worth the journey.  There are truly amazing street performers here.  Street performing seems less of a way for people to subsist than a venue for really talented people to showcase that talent and make a decent living.  The charcoal and pencil portraits the artists do here, in an hour´s time, represent the kind of work they would be paid REALLY GOOD money to produce in the States.  And I´ve heard all kinds of brilliant musicians playing all kinds of instruments.  It´s difficult to retain the will to continue walking most of the time with all the entertainment to be experienced on the way.  And guess where I´ll be heading the day after tomorrow!!  Guess!!  Guess, I said!!!!  Well, I won´t give it all away (since it seems you reFUSE to guess), but I´ll just say that it has a little something to do with trampling, a little something to do with red, and a lot to do with ludicrous, idiot tourists looking for an adrenaline rush.  I´ve yet to decide if I´ll be running :)   '),(7423,669116,'02,July,2003','2003-07-02 00:00:00','    Godforsaken Island!!!   Yes, yes, the island fever has set in.  I think what finally broke me was the shudders that run up and down my spine each time I hear the sound of a car peeling out.  And this is a sound I hear quite often here, at least 10 times/day.  Where are they all  going ?  I mean, what\'s the big hurry?!  I think the sound of a car peeling out is my least favorite sound in the whole world.    I\'m oh so happy to be leaving for Spain (Barcelona) tomorrow.  But my last few days here have been really, really wonderful and full of the kind of solitude I had hoped to find when I left for Europe without a travel plan.  The exception to my solitude came in the form of Nick (whom I hadn\'t seen much at all during my time here) and a German fellow by the name of Martin who was staying in Nick\'s guesthouse.  On Nick\'s last night here, the three of us went to a little bar and then headed down to Paceville (Fuego, specifically)because Martin had been there on Friday and learned that they play real Salsa and Merengue from 3-5 in the morning or so.  Well, that didn\'t exactly turn out to be true on a Monday night, but the three of us were having a good time, Nick dancing sprightly with the goofy, slightly opened-mouth smile he wears when he\'s very drunk.  At something like 4 o\'clock in the morning, it began to seem to me that the entire dancefloor was covered with men--horny, eager, squinty-eyed men who seemed not to have seen a female in years of something like solitary confinement.  So, lucky for me, Nick and Martin were hovering nearby and were quick to intercede when they sensed I was uncomfortable.  At one point they nearly fought with a man who decided that 15 seconds was enough time spent dancing with a woman before it was okay to grab her ass.  They were the most valiant of knights in a sea of slime.  After Fuego we decided to take a sunrise skinny dip (my first ever) in the Mediterranean.  Wonderful.  Martin laughed (well not really, though I sensed he was laughing internally) at me in my prudish Americaness (demonstrated by my unwillingness to remove clothing until actually in the water), but for goodness sakes, the sun was already up, and people were walking their dogs on the nearby strand.  Plus, Malta is a pretty conservative, and I\'m pretty sure we were breaking the law, so best to be subtle about it, no?  Today I learned that Saad (from the guesthouse) was detained for an hour and a half by the police yesterday because \"he resembled a man who was wanted for stabbing a Maltese woman last week.\"  See, if you talk to the Maltese, they will tell you that the only problems with crime here come from the Arab population.  And to the Maltese, all Arabs are the same.  The Arab population here suffers a kind of racism that I think could be rivaled by the United States maybe 60 years ago or so.  Terrible.  Saad said he eventually saw a photo of the wanted man in question, and that the only resemblance the two of them bore was their shared black hair and dark eyes.  Again, terrible.  But at least they didn\'t beat him like he said they were beating some of the other men (all Arab, of course) being held while he was there.  More of the terrible.  I\'ll be heading to the beach one last time this afternoon.  There is an oh-so-groovy little cabana near the local rock beach, where the owner (a curly-haired old hippy with eyes that travel in two different directions) plays music just wonderful to frolic in the water and bathe in the sun to...Bob Marley, Pink Floyd, Radiohead, Grateful Dead, U2.  Yesterday, I felt spontaneous tears spring forth upon hearing Marley\'s live version of \"No Woman, No Cry.\"  That song ALWAYS gets me, but I don\'t think it ever made me cry before.  Where would we be on this earth without the comfort, the joy, the inspiration, and the sometimes heart-wrenching effects of music?      I\'m waxing a little dreamy right now.  I\'m having one of those moments where I just feel incredibly fortunate to be alive.  And I\'m really looking forward to Spain; I feel her reaching her fingers invitingly in my direction.   '),(7424,669116,'09,July,2003','2003-07-09 00:00:00','    Well, forget the cello.  The cello was cool, but the Running of the Bulls was *****SUPER COOL*****  I mean really super duper amazing...the biggest party and the best time of my life.  I can\'t write many details right now because  1) I\'m in Paris and have slightly more interesting things to do (I just had to get online to book a hostel in New York) and  2) As the parenthetical comments in #1 imply, I\'ll be home soon enough anyway and can talk about it in-person.  That\'s right...I\'ll be back as of Sunday, just 5 short weeks later.  Yes, this is a big time bummer.  But I don\'t have any money left, and the bills and students loans I have back home are looming menacingly overhead.  I figure it would be a bit nicer to go traveling again when I can get those things under control and not be worrying about it.  But Pamplona...let me just say a bit about this.  It was inSANE!!!!  Sooooo many people.  Sooooo much fun.  Has anybody seen photos of the big square where they kick off the festival every year?  You know, the sea-of-red-handkerchiefs-and-near-rabid-with-enthusiasm-Spaniards?  Well, I happened upon that square quite by accident and found myself in the heart of the madness.  And it was a very invigorating sort of madness.  And of course there were the people jumping off that monument (mostly tourists, mostly Aussie tourists at that--the Aussies are ridiculously crazy and great!) and the Sangria a-flowin\' and sleeping by the bank of a little river for a half hour total in two nights and the 14 hour busride from Barcelona to Paris afterward (three days, yes three days without a shower covered in the champagne and sangria and beer and eggs that landed on me during the celebrations).  Paris has been quite calm, relatively.  I\'m staying in a cute little hostel in the Montmartre area of the city.  Montmartre is French for \"peep show.\"  Just kidding, but there are a TON of peepshows around, though the area is nice otherwise.  I saw about 600 monuments yesterday and am about to head out for Round 2 today.  Paris is beautiful and ugly at the same time.  Kind of like every major city...it\'s got its attractions mixed with the comfortable littering tendencies of the locals.  I can\'t think of a place I\'d rather be at this moment...except maybe Pamplona :)  I leave tomorrow and will be in New York until Sunday, then back to L.A.  :(   '),(7425,669116,'15,July,2003','2003-07-15 00:00:00','    So I was in kind of a hurry the last couple of times I wrote, and I fear I may have left out some of the juicy stuff.  I’m in Hollywood right now, hanging out with my Bro and trying not to tie up his phone line too much.  See, I’m trying my hand at the job market here in L.A., but it seems there’s no such thing as classifieds in the newspapers these days.  So I’m doing a lot of looking online, and he’s raised eyebrows at me more than once today alone.  Sorry, Bro.  But anyway, I think I have to backtrack a little to Pamplona and the bulls and all that.  The thing is that I didn’t actually see any running of any bulls, which is probably a good part of the reason that the experience was such a pleasant one.  I think that if I had witnessed anybody getting mangled I probably wouldn’t feel so happy about the whole thing.  But I was there for the opening festivities, and, like I said, the Aussies made it an incredibly exciting experience.  I don’t know what it is about those people.  It’s as though being an adventurer comes along with Australian citizenship or something.  All the Aussies I met abroad were traveling for 6 months or more (some with surfboards under their arms).  When I told them that I would be gone for roughly 5 weeks, most of them asked me questions along the lines of, “Why even bother?”  They’ve explained to me that Australia is such an isolated country, and it’s so expensive to leave, that they’re gone for long periods of time in order to make it worth the trip.  So I already told about the smelly bus ride (as in,  I  was smelly because of the caked on eggs and various beverages), and I already told that I visited beaucoup monuments during my stay in Paris, which was brief but pleasant.  I could tell you about my stay with my old pal Renee in the Hamptons, but that was just soooooo mellow that I fear I could lull anybody into a pleasant, dreamy siesta just by talking about it.    I’ll just say that Renee’s house is The Cutest and she was The Most Gracious and Entertaining Hostess and the other friends she had staying (a friend of hers from kindergarten and two Swedes) with her for the weekend provided excellent conversation and the whole thing was wonderfully refreshing.  Wonderfully relaxing after the whirlwind travels.  That is despite the fact that her septic tank burst (or something like that—I’m pretty ignorant about plumbing) while I was there and she spent about 7 hours dealing with getting it pumped by an Emergency Septic Tank Pumping Company.  I actually welcomed the waiting-at-home-for-the-guy-to-come-and-do-his-thing thing.  That was part of what made the stay so relaxing.  So that leaves me with Hollywood.  Oh, but first lemme just advise all readers to avoid Charles de Gaulle airport in Paris like the plague.  I arrived 3 ½ hours early and STILL I was running to catch my plane.  Nightmare.  Part of the thing was that the security inspectors were moving sooooo sloooooowly.  I kept getting frustrated while I was watching them from the line because it seemed that they were just chit-chatting with all the passengers; it was like they were flirting or something because they had these smirks on their faces and all the passengers were smiling all goofily while they talked.  ‘What IS this?,’ I thought.  But when I got to my guy, I realized the people were smiling and laughing because the inspectors’ accents were so strong (accentuated by the fact that they were speaking through smirks for whatever reason) that it was nearly impossible to decipher what they were asking.  Example:   Inspector:  What do you do to leave?  Me:  What?  Inspector:   What do you do for to leave?   (I’m thinking  well, I packed my bag and checked out of my hostel and then took the Metro to the RER train to the airport shuttle and now I’m here.   But I know that’s not what he means.  I just can’t figure out what he DOES mean.  I answer just to answer something…)   Me:  Well, I packed my bag and checked out of my hostel and then took the Metro…  Inspector:  To  leave !!!  What do you do for to   leave  ???!!!!!  Me:  Oh, you mean what do I do for a living?  Inspector:  Yes, for to leave.  Me:  I’m a student. (I lie to make it less complicated)  Inspector:  And who pack your bag?  And on and on.  But the funny thing was that, at first, the way he asked questions was so unassuming that it almost seemed as though he was just making pleasant conversation, and that he was really interested in my trip, which is the other reason I think all the passengers were smiling.  Because he said, “Where did you go on your holiday?” and I thought he was just passing time while he did the official business of checking my passport.  So I said, “Well, I was in Italy and then I went down to Malta, oh and then I was in Barcelona and Pamplona and then I came up here to Paris.’ I was about to pull out my pictures and tell him the story of Sea Malta when he started in with the “What do you do for to leave?” business.  The whole thing was just kind of goofy but cute somehow.  Not cute enough to overshadow the rest of the de Gaulle nightmare, though, so be warned.  And Hollywood.    It’s only my first day, but I’m just returning to old stomping grounds anyway, so I feel at home.  Except that I think there’s a limit to how much one can feel at home in Hollywood.  There is definitely an anonymity about the place.  And yes, everyone here is trying to make it.  And yes, the smog is yucky poo poo.  And yes it’s dirty and somehow sad.  But something has me drawn here, so we’ll see what happens.  The best thing that happened to me today was that a human sandwich spoke to me.  There was this sign twirler dressed like a big sandwich standing on the street corner outside of Subway.  When I walked by, he spoke.  It seemed to me that that sort of thing shouldn’t be allowed—like talking sandwiches are scarier than they could ever be effective promotional agents, but that’s just me:   Sandwich:  Helloooooooo.  Me:  laughter  Sandwich:  I know why you’re laughing.  Me:  A talking sandwich is pretty funny.  Sandwich:  Do you want something to drink?  Silliness.  I had just had something to drink, so I walked away without answering, but now I’m curious.  I mean, did the sandwich have beverages tucked between his folds of lettuce?  Was he giving away free drink coupons with subway sandwich purchases?    Maybe I’ll go by tomorrow to kill to boredom that will inevitably come while I wait for responses to my employment queries.    Bottom lines are that I’m back on American soil and safe and mostly happy save for the sad that has come with the end of my fantastic voyage.  I’m currently looking into jobs with travel opportunities.  The travel bug dies hard like Malta mosquitoes, it seems.  Love to you all.   '),(7426,669116,'02,January,2004','2004-01-02 00:00:00','    At the kind request of one of my international readers (ok, my only international reader), I\'ve decided to take up my blog again and write for the simple satisfaction of it while trying to convince some-body/company/entity to pay me for my efforts.    So, a lot has happened since I last wrote--since, for example, I forewent the hare-brained idea of making a living in Los Angeles.  Den of despair!  I shouldn\'t say that.  I know plenty of people who are, in fact, very happy living in the city of would-be angels; I just didn\'t happen to count myself among them.  As was customary in recent years, I made sure to be in San Jose for the annual jazz festival.  My return to L.A. was to happen the following Tuesday, but when that afternoon arrived I found myself physically incapable (or more likely, simply unwilling) to board the lighting-fast, Greyhound bus that was to take me there.  San Jose is indeed the place for me.  I pitched a tent in the living room of my dear friends, Nicole and Raul (and Maya too) and set out to find gainful employment, which I found (well, employment anyway) without too much delay.  Two months later I moved into a groovy (umm) flat, we\'ll call it, with my best gal Kelsi, and that brings me about to present.  I\'m still looking for the job I went into student loan debt for...the one that would make all the late-night term papers worth the trouble (and that is, of course, the most pessimistic viewpoint...the education WAS worth the effort in its own right...but it will be infinitely MORE worth it when I have the means to begin  paying off  those student loans, you dig?)  In the meantime I\'ve been back at an old haunt--Barnes and Noble--which is among the most pleasant environments one might find when it comes to working retail.  It\'s especially nice to be working in a Barnes and Noble in one of the most liberal areas of the country, where requests for Al Franken\'s  Lies and the Lying Liars Who Tell Them  outnumber those for Bill O\'Reilly\'s  Who\'s Looking Out for You?  by about 25-0 and the Evil Anne Coulter\'s book sits in heaping mounds, growing dusty,  under  a display table.  And that\'s the short of it, for now.  I\'m going to start writing more consistently, sharing some of my favorite public transportation tales of horror, along with other daily observations and embarrassing moments and good news as well.  Hope all is well--whoever you are, reading this, and wherever you might be.   '),(7427,669116,'09,January,2004','2004-01-09 00:00:00','    “ This check cashing place stoled [sic] $85 from me .”  So read the message painstakingly markered on a sandwich board worn by the one-man picket line I saw downtown yesterday.  I didn’t stop to get the details; I figured he must have a legitimate beef if he’s willing to stand in the cold and explain his plight to complacent passers-by.  And yet I wondered how much that “stoled” was taking away from the serious reception of his message.  Or was it just me?  See, I’ve been thinking a lot about language lately.  There’s a sign outside a furniture store near my place of work that advertises a free entry into a drawing for a  harly davisons motorclcye , free with  every purchased .  I have to say there’s something about those and other errors on that sign that make me leery of shopping at that store.  A little part of me fears I won’t be able to communicate with the people who work there.  I don’t like that part of me.  What’s that statistic say?  That something like 90% of all messages are transmitted and received non-verbally?  I’m not sure how They calculate these things in clear statistics like that, but it certainly feels true.  So why should poor grammar be such a roadblock for me?  I’ve just begun the process of getting credentialed to teach adult English as a Second Language.  This seems a good fit for me—I love language and (just as I enjoy learning other languages) will enjoy helping others communicate better in English.  It’s something I’ve been doing informally for a while anyway, and I think what motivates me is my own awareness of sounding like a mentally challenged third-grader when speaking either of my two foreign languages.  Have you ever had a relationship (even a brief one—a.k.a. a fling) in another language?  There’s something very frustrating about wanting to say, perhaps, “There’s this look you have that’s very arresting in nature.  I feel somewhat off-balance when you look at me like that—like you know something I don’t and would regret knowing if I did,” and instead saying something about the equivalent of “Your face makes me sad.”  So, last summer when I was in Malta and a German acquaintance said things like, “We are having these [zeez] meetings all the [zuh] time in the [zuh] mornings,” I would (at his request, of course, that I help him with his English) explain how and why “we have meetings every morning” is a sufficient and more common way of expressing the same idea.  But then, there are infinitely many ways to convey a message, and aren\'t the idiosyncratic expressions of second-language learners what make their speech so captivating?  Really, is “your face makes me sad”  that  far off base?   I’ve traveled a little distance, however, from my original point, which was that a few letters can make a big difference in whether or not one’s true intention will be effectively communicated.  This idea hit home one day when I was in high school and my Mom, who’s really named Carmelita Konrad, received a letter addressed to Carnekuta Jibrad (say it aloud in a mean voice—it sounds really intimidating).  It seems whoever typed the envelope had his or her right hand over to the left by one key [what does your name become?  I’m Jusa Jibrad or Losa Lpmrad, depending on in which direction I’m off].  If I were into New Year’s resolutions, I’d resolve not to equate poor spelling and grammar with bad ideas or a lack of intelligence.  It’s not like I do this, ultimately, now—but I’m tempted sometimes and I don’t like that.  Even if grammatical skills were a true measure of one’s intelligence, it would be, at most, one type of intelligence out of many.  Would Nietzsche have needed the ability to spell Nietzsche in order to prove the worth of his ideas?  Kahlil Gibran?  What of those who don’t even communicate their ideas in words—or more interestingly—their art, their emotions, their music?  It’s a process, though.  So if you write me an e-mail that doesn’t contain a single period and says things like “whose comming to you’re party besides myself?” and it takes me a day or two to recover and respond, please forgive me.  You can laugh at my ill-fated attempts at baking, my inability to stay in one key while singing, the fact that I can\'t do a single pull-up, or my complete incapacity to “get” double entendres all you want.  I’ll understand.   '),(7428,669116,'04,January,2004','2004-01-04 00:00:00','    When Kelsi and I decided to find a place together, we had a pretty substantial list of requirements.  I won\'t bore with ALL the details, but we were looking for things like good natural light, a kitchen that opened to other living spaces, a bit of space separating the bedrooms, etc., etc.  Our new place meets nearly every requirement, with two exceptions.  I\'d wanted laundry on the premises and she\'d wanted a fireplace.  We settled for a bi-weekly jaunt to our most favorite laundrymat and employing the two central heating systems in our apartment.  All was good, except, that is, one thing...  ...see, there was this pigeon shit to contend with.    Honestly, I didn\'t really feel the bird shit represented a force to contend with, per se.  But my gal Kelsi--when she gets an idea in her head...  :)  The thing is that the view from our little dining area is less-than-breathtaking.  The many-windowed wall faces a large, Victorian house next door whose adjacent wall just happens(ed) to be covered in the droppings of the two birds who\'d made a love nest of the tiny stoop above.  Before we moved in, Kelsi mentioned to our soon-to-be landlord and landlady (on separate occasions) that she\'d like to have an improved view, and she asked for their suggestions.    The landlady suggested we talk to the owner of the house next door and see if she\'d mind our cleaning it.  Logical.  The landlord was slightly more, shall I say, proactive (?).  In addition to bringing over a large ladder and some kind of high-powered cleaning agent, he (how to say this delicately?) saw to it that the pigeons, well, uh, met with an accident.    Horrible!  I know.  If I\'d known that was going to be part of his solution, I would never have let Kelsi mention it to him.    But anyway, we were then free to clean the droppings ourselves.  There was some downtime though...it rained or we didn\'t have the time or it was dark or we didn\'t have the cleaning stuff yet.  It was during this downtime that I came to understand just how big a stick in Kelsi\'s spokes this bird shit thing was.  There was more than one occassion when I walked out of the hallway to find her gazing disconcertedly out the window at the poo.  I\'d laugh--I couldn\'t help it; she reminded me of the grocer in \"What\'s Eating Gilbert Grape\" who was always driving into the Walmart-like, small business-crushing Food Mart\'s parking lot and staring up at the big sign, no doubt wondering where he went wrong, what he did to deserve this lot in life.  It would take her a second to realize why I was laughing, why I began to fear she was being overtaken with thoughts of these...damned...pigeons (clenched fist, clenched teeth)!!!!!  I\'d been thankful for the post-poning rain and for my absense over the holidays because, as you can imagine, scraping bird droppings off high-up places atop rickety ladders is, while ranking high, just not The Highest on my list of fun ways to pass the time.  But finally, the rain cleared, the sun came out, we had the cleaning stuff and the time and (50% of us anyway) the drive.  I\'d kept thinking the problem would take care of itself (an inclination I\'d like to blame--as I blame so many of my other undesirable qualities--on my Gemini-ness), but alas, problems don\'t generally work that way.  [Incidentally, when one group of birds meets with an accident, others are happy to move in--just so you and YOUR sniper landlord know]  The day of reckoning came last Friday.  Even as the day was upon us, I tried to delay.   I took an extra long bath in the morning. I mentioned that maybe, for who knows what reason, the following day might be an even BETTER day to clean bird poop! I took to a bit of pouting, pointing out how it was my first day off in 7 and blah blah blah.  But that didn\'t last long.  After all, who really wants to look at bird droppings when eating?    I was about 8 steps up the ladder when I realized this was one of those things I\'d be telling Kelsi\'s future children about: \"And yet  another  thing your mama somehow convinced me was a good idea...\"  See, I have a serious ladder phobia.  I discovered this only  after  I was hired to work at Pier 1 Imports about 9 years ago, where climbing to the very tops of ladders and flinging huge, off-balanced floor pillows onto and off of shelves overhead is just part of the daily grind.  Four years after that and about 10 jobs later I graduated to even taller ladders at Target, flinging even heavier objects like King-sized comforters.  Strangely enough, however, ladder phobias don\'t really go away.  So I did wonder how far I\'d have to fall to ensure my spot in next year\'s \"Darwin Awards\" book.  I can hear Kelsi\'s quote now: \"She was reaching over...see, there was this especially menacing little chunk, a bit yellowish with a big gloppy hanging off... I asked her could she please grab that one while she was up there and the next thing I knew...\"   Obviously, however, I survived the ordeal.  \'Cept the rain starting down in sheets just as we were folding up the ladders.  I was soaked through and through by the time I\'d rolled up the neighbor\'s hose and made quick waste of the newly bio-hazarded scrubbing agents.  I giggled up the back steps, tugged my red, dripping wet Chuck Taylors off my feet, and thanked the universe for yet another ridiculous memory to tuck into the folds of my joyful life.  And I\'ve never quite enjoyed a homecooked meal so much as the blueberry pancakes we shared while gazing out the breakfast nook window at our new, gleaming and glowing, dookie-free wall.  Sometimes, you just gotta clear away the shit.   '),(7429,669116,'17,January,2004','2004-01-17 00:00:00','    Last month I took the CBEST (California Basic Education Skills Test) as a preliminary step in attaining my adult teaching credential.  The CBEST isn’t known for putting its takers up to challenging mental feats of strength (it’s not like taking the GRE or the Bar Exam).  Mostly, it’s just another hoop to jump through—a barrier between the school districts and the mobs of motley folk who would gladly take $110/day to substitute teach or $30/hour to teach ESL, if it were simply a matter of having the desire.  The CBEST had been a pain in my ass for a few months.  I’d signed up to take it once, paying $20 more than the initial fee because I’d signed up late.  Then I missed that exam because of car-borrowing schedule conflicts, which sucked.  Two months later (when it was finally offered again) I signed up early enough to pay only the regular $40 the second time around, and made sure to have my travel plans well-outlined.  That’s a total of $100 I paid in order for the state to tell me, on an official piece of (not just paper)  stock  paper, that I can convert fractions to percentages and identify the closest synonyms to obscure words like “strange,” or “ambiguous” if asked to do so.  The reading section was easy, the math a fun little blast from the past (I realized I hadn’t done any math—other than standard arithmetic—in ten! years), and then came the essays.  Now, nobody likes an in-class essay (let alone 2!), and despite what y’all might think, English majors are no exception.  It might be slightly easier for us to whip out lengthy, super cheese responses to the types of generic questions asked in these standardized tests, but it’s no less obnoxious.  The first question had something to do with whether or not we were currently living in something that could be considered “the best of times and the worst of times.”  I don’t recall much about my response to that one except the main point being: we have a great deal of potential to be living in “the best of times” with scientific advances and what not, but that such potential was rendered null and void when advances were used for less-than-noble causes or when the worthy causes (i.e. treatment for HIV patients, or even basic healthcare) were like pipedreams for the majority of the county’s citizens—due to outrageous costs, etc, etc.  Anyway, the second question was a bit more tangible, though no less irritating to consider: “What is the single best piece of advice you’ve ever received?”  For a stagnating 4 or so minutes, the only pieces of advice I could remember having received were, while incredibly practical, not the sort of responses you want to hand in to state-administered essay topics.  The first piece of advice was good for around-the-house matters: “Righty tighty, lefty loosey,” from my friend JD.  If you’re unfamiliar with this phrase, it has to do with the direction for turning screws in order to produce the desired effect; I’ve used it many times in the year and a half or so since I learned it.  Incidentally, this is close to what an ex-boyfriend deemed the best piece of advice he’d ever received from his father, which was: “Only an idiot screws a screw in too tightly.”  Neither of those would have made much of an essay.  The second piece of advice I’ve considered invaluable came from my brother, about 6 years ago: “Beer before liquor, never been sicker…liquor before beer, never fear.”  It’s not like I’m a big drinker, which is precisely why—come to think of it—a compact, rhyming, easily-remembered-even-when-already-a-little-bit-tipsy phrase like that has come in so handy—the two-line poem of sorts has more than made up for my lack of experience or marginal street smarts.  How many toilet-hugging incidents were shrewdly sidestepped?  How many “sick” days carefully preserved for bona fide illnesses?  How many embarrassed apologies never uttered?  Now, I should note here that there’s room for mention of wine in that little ditty (a point the owner of the wine bar downtown, my friend Nick, and any unfortunate witnesses to an unfortunate incident last winter will verify).  Should it be considered liquor?  Can it be safely mixed with  any thing?  And what about warm alcoholic beverages like Sake, which just seem doomed for disaster by virtue of their being warm?  It’s not that I couldn’t have written an essay about this advice from my brother…look, I’m writing about it right now.  It’s just that I shouldn’t.  And I didn’t.  What I did was make up a piece of advice that supposedly came from my Dad: “Never fail to recognize the lessons you can learn in the midst of a so-called failure,” or something like that.  It’s exactly like something my Dad  would  say.  My Mom, too, for that matter.  And it\'s good advice.  But it’s also like something you’d hear in an after school special or even from the Evil Dr. Phil.  No matter how good the advice might be, it’s just not all that captivating to write about.  Is it so impossible to imagine that giving test-takers topics that are truly interesting would perhaps produce a better example of their best writing?  Wouldn’t you have some pretty exciting responses to a question regarding the best and worst things about your first kiss?  How about the shittiest day you remember?  The time you walked out on a job and what brought you to that breaking point?  What are the three most terrible (and yet true) things your ex-girlfriend could/would say about you?  I suppose there are some censorship rules prohibiting the state from printing words like “shittiest” on standardized texts, but maybe it’s time we relax these puritan tendencies a bit.  With the new topics, they wouldn’t have to pay people so much to sit around and read these boring essays all damned day.  It would be like a privilege—something they could get Masters students to do for next-to-nothing.  Ok, maybe that’s exaggerating, but they sure would be easier to write.  And they’d be more reflective of our personalities—isn’t that the most useful information anyway?  What if some man’s shittiest day was the one that landed him in jail on charges of 2nd degree murder and racketeering?  It might be a slightly better gauge of his appropriate place teaching children than some piece of advice his grandma handed down when he was 11.  I guess that’s what the background checks are for, eh?  Just allow me my fantasies for now.  I happen to believe standardized tests shouldn’t HAVE to be less fun than an enema.   '),(7430,669116,'16,January,2004','2004-01-16 00:00:00','    It had been a peaceful bus ride until then—a group of elementary school kids tend to pull one’s attention away from even the most mundane reading (and this was far from—a recommendation from a friend…Chalmers Johnson’s “Sorrows of Empire”…most excellent and disturbing critique of the current administration).  I was a little annoyed at the disturbance.  But once I decided the reading was a wash, I was happy to sit there and eavesdrop.  Kids are so damned cute without even trying.  There were about six of them—third graders, I’m thinking.  Loud, loud, loud.  There was an elderly man sitting next to two of them, and I initially thought he was lamenting his deplorable choice of seating, but it turns out he was quite taken.  The chaperone had told this one little boy that he should tie his shoes so he didn’t trip.  The kid made a face and pretty much ignored her.  Then the old man said:  “You may not want to, but it’ll make her happy.  That’s the reason men do things.  There, I’ve let you in on one of the secrets.”  BOY #1:  Then you get her heart?  MAN: (laughing) Hey!  That’s the other secret!  BOY #2:  I already knew  both  those secrets.  I like a girl.  She hates me.  GIRL #1: Yeah, and her name’s A- man -da!  (laughter all around)  BOY #2:  So?     He looked down at the ground then and smiled a super shy smile revealing two dimples.  So stinkin’ cute!!  A conversation ensued regarding what these boys knew about women.   KID #1:  Actually, I learned another lesson: girls can talk a realllllly long time.  KID #2: Yeah, and they like to yell.   I thought maybe they were talking about their mothers until Kid #1 chimed in again   KID #1:  When I don’t listen to my sisters, I’m a REAL man…I think.   Apparently, the women these boys know just love to scream and fight, cuz   KID #2: The girl I like?  She wants to  kill  me.  I was laughing so hard.  Kill him?  Goodness!  They certainly knew they had a captive audience, so I think this group was perhaps more animated than usual, but then maybe not.  If you never spend any time around children, I can tell you it’s worth the strain on your patience (if you\'re not-so-kid-inclined) just to hear how they interact and try to remember having that kind of outlook.  Things are pretty clear:  either a girl likes you back (not bloody likely, in the 3rd grade), or she wants to kill you.  Makes things pretty easy, no?  Can you imagine such conversations in the grown-up world?   John:  Hey what about that chic from marketing…did you ask her out?  Rick:  Naw.  I haven’t.  John:  Well, what’s your take?  Does she like you?  Rick:  No, actually, she wants to kill me.  John:  Tough luck, man.  I’m calling for all-out, across the board regression, just to simplify matters.   P.S.  The kid  actually tripped  when he got off on account of not tying his shoelace.  If men would just listen to women...   '),(7431,669116,'24,January,2004','2004-01-24 00:00:00','    On the bus a couple of evenings ago, I was kept entertained by a man and woman sitting at the front of the bus and talking (though they were seated right across the aisle from one another) as if there were miles between them.  They seemed to know each other only marginally, through mutual halfway house friends (from what I could gather—and it’s not eavesdropping if the people are talking that loud).  Regardless, when the woman discovered they were headed to the SAME halfway house just then (him because he lived there, her because her friend lived there), she sidled up next to him.    “Well, then I’ll sit next to you, since we’re going in the same direction,” she said.  Then, nudging him, she added, “Hey there, Mister…Bubble…Gum…Ring.  Huh huh.”    He giggled knowingly then, and the two of them sat there staring  straight  into each other’s eyes, “huh huh-ing” in a knowing and, then, surprisingly familiar way.    I say surprising because, at that point, she put her arm around him (he responded by warning her he was sick) and then he said, “Charlene, right?”    And she said, “ Dor -een.”    And he said, “I was close enough.”  Which she apparently agreed with, because when he asked, she agreed to stop with him for a soda at Kentucky Fried Chicken on the way to where they were going.  I was rapt the entire way home, listening for any clue into what “Mister Bubble Gum Ring” might have been referring to.  Did he wear such a ring?  Had he once given her one?  What  is  a bubble gum ring, anyway?  It sounded kinky, the way she said it, but “kinky” didn’t really seem an applicable term for either party.  Who can say about those types of things, though?  I love the bus.  A few years ago, I was sitting on the 22 (Eastridge Mall to Palo Alto Caltrain Station) when a man and woman (not together but having apparently struck up a conversation at the bus stop) boarded, both with chocolate rings around their mouths.  She was carrying a big, Costco-sized container of chocolate muffins, which she had, by all outward appearances, been kind enough to share.  [I’d seen the woman before.  We’d once been sitting at the same bus stop and she told me about having worked twenty-some years at the now-defunct Del Monte canning plant in San Jose before deciding to take an office job.  She advised me against taking such a job (secretary), because “the men, you know, they only want one thing.”  I’d been facing forward, and when it became clear she wasn’t going to continue that sentence, I looked at her to find her making that finger-sliding-in-and-out-of-a-hole-made-by-the-other-thumb-and-forefinger motion while watching me with a raised eyebrow.  Ohmigod!]  Anyway, the man seemed about her age (mid- to late-50’s), and wasn’t so much talking with her as listening to her.    She talked about things I can’t remember now, and then stood abruptly when she realized we had reached her stop.  Though the man had said absolutely nothing the entire way, when she stood, she shook his hand and said, “I wish you well.  You seem like a lovely person.”  His response was, “I am.”    That’s it.  No “thank you.”  No humble “puh-shaw” and dismissive waving-off.  Just a simple, “I am.”  What I thought was so great about this transaction was that “lovely” would rank about 9,987 on a list of 10,000 adjectives I would have thought of in describing this man.  He was very large and very stinky with a scraggly beard and crooked teeth and, you’ll remember, a big chocolate ring around his mouth.    But that was only on the outside.  Whatever it was he knew he had inside of him—whatever it was that made him lovely—she had seen it.  He knew it was there and she had glimpsed it.  And I thought that was one of the most beautiful things I’d ever witnessed.  Imagine how different the world would be if everyone could honestly describe himself or herself as “lovely,” could accept a compliment that soberly and know it held within it some truth.  Or imagine if we all went around commenting on each other’s loveliness and sharing our chocolate muffins.  I think I had a vision of my dream world that day.  Who’d have thought it would happen on the 22?  I must always strive to keep my eyes and ears wide open.  There are a million transpirings to witness and overhear in a day, and some of them--even on the Urinemobile of all Urinemobiles, the 22 bus--give me the kind of renewed hope that every person needs now and then to remind him or her of what\'s awe-inspiring, what\'s comical, what\'s worth it, in this world.   '),(7432,669116,'31,January,2004','2004-01-31 00:00:00','    About 6 years ago I made a deliberate trip to a bookstore on Superbowl Sunday.  I claimed to be in search of intelligent life, figuring if there were romantic possibilities lurking about the greater metropolitan Denver area (where I was living at the time), they might best be discovered during this crucial moment of unbridled American machismo.   Basically, it occurred to me that anyone away from the TV at that moment (especially one whose alternate locale was the bookstore—Tattered Cover—a thriving independent bookstore, at that) was someone worth getting to know.  Oh, the idealism of youth.  The self-righteousness, perhaps(?).  I didn’t meet anyone interesting that day.  And anyway, how much more worthy was  my  cause?  Instead of being in the bookstore seeking to expand my web of knowledge, I was in there on the make!  What a phony baloney.  (I’m reminded of a book I saw the other day called  When God Winks on Love , subtitled  Let the Power of Coincidence Lead You to Love .  All I could think was this hypothetical testimonial from readers-turned-devotees to the author’s cause: “The coincidence I masterminded worked out beautifully.”  Ridiculous.)  I know what part of my problem was at the time:  I had just returned to the States from an extended stay in Mexico, where my new awareness of sub-par living conditions greatly soured my outlook on American wastefulness and decadent self-indulgence.  To me, the Superbowl and everything that touched it was evidence of a spoiled rotten population with too much time on its hands and its priorities way, way off-base.  And I still feel that way sometimes.  And I don’t plan to watch tomorrow’s game or wager any money on it or even bother to check the score in the next day’s newspaper.  But that’s mainly because the Raiders aren’t playing this year (and that’s assuming I’m kind enough to term what they did last year “playing.”  Other words that came to mind are “floundering,” [are you guys] “kidding?!,” and, simply, “Sad.  So sad.”).  It’s not that I care particularly about football (certainly not in the way I care about baseball, not in the way I have elaborate fantasies about meeting the Dodgers’ Paul LoDuca), but I am no longer opposed to the idea of sitting down for an afternoon and enjoying the be-all, end-all, American sporting Super Event.  Please indulge my desire to share a few reasons for the change of heart.  I’ve thought a lot in recent years about the Superbowl in relation to the World Cup.  I happened to be in Mexico when the World Cup was here in the United States, and I was struck by the serious devotion to fútbol (soccer) I found among the people there.  One day when Mexico was playing Germany (causing me, in my mixed-raced Beanerschnitzel-ness, great inner-conflict), I noticed that the town where I was staying was teeming with young people who would otherwise have been studying at the time.  I stopped a young boy and asked, “Why aren’t you all in school?  Is there some sort of holiday or something?” to which he answered, “Mexico’s playing,” with a look on his face that said precisely, ‘Like, duh.’  My friend Kelsi was somewhat disgusted by this fanaticism, sparking a great debate between us.  I surprised myself by presenting the argument that soccer, while not being the most intellectual of pursuits, was a wonderfully unifying activity.  It was the only thing I could think of that so great a percentage of people worldwide took very seriously and agreed upon: The importance of the sport’s outcomes are agreed upon.  The rules are agreed upon.  That grown men should put their machismo on hold, flop, and cry to their heart’s content if it means drawing the attention of the ref and, ultimately, a yellow card, is absolutely agreed upon.  To me soccer is the innocuous version of war.  (Excepting, of course, the unfortunate incidents of murder and mass-suffocation that have resulted from the games more fervent, more overzealous elements—to everything, there are extremes).  Countries from all over the world come together once every four years and settle their scores in a forum where the laws are set, the game is timed (though loosely sometimes), and everyone has the chance to cover his privates before any serious damage can be done.  Why can’t men always play by the rules in this way?  American football is just a microcosm of this worldwide phenomenon.  It’s a less-exciting, heavily padded, far-too-oft-stalled version of soccer, but a version all the same.  Black Hole Raider fans excluded, grown men come together to play or to cheer for their favorite teams in a normal, healthy way.  We need not kill each other to prove our superiority.  Why not just run faster, block better, or have a person on our team whose precise aim and strong “gun” means an extra six points rather than an extra person dead?  And there’s another thing.  In all my former pooh-poohing of the Big Game, what I had neglected to consider were the humans who comprised the teams.  I had forgotten that, behind every facemask and (behind that) every snarl of threat and audible declaration of dominance, is a man with a dream—a man who has, no doubt, been working day after day and year after year for the chance to be, in this exact moment, on this very field.  Who am I to criticize that?  Put aside the announcers, the commercials, the millions of dollars exchanging hands, and you’ve got about 100 men who are about to experience the best goddamned day of their lives.  And yeah, I support that.   P.S.  Please send any comments to  scoobisnac2000@yahoo.com   '),(7433,669116,'27,January,2004','2004-01-27 00:00:00','    It seems we’ve created a colossal monster.  Or, to be more exact, we’ve created millions of lesser monsters that, when considered individually, are really just singular off-shoots of (maybe roughly the size of the fingernail of) one, big, sweaty, ugly, snaggle-toothed beast: The Customer.  And how did we do this?  Well that’s easy.  Someone, somewhere, in some fluorescent lit conference room came up with the phrase that would become the bane of retail and restaurant workers from then until somebody comes up with a truer (and yet equally profitable) phrase…the phrase, of course, being “The customer is always right.”  Whoever came up with that phrase couldn’t possibly have been in contact with any  actual  customers, or at least not the kind who hang around bookstores.  I know what you all are thinking (I used to think it myself)… The bookstore—what a mellow place to work.  What a nice environment.  Books.  What could be better than working around books all day?   And to be honest, that’s mostly true.  But there’s a definite, lurking danger hidden below, around, and inside words like “mostly”—that danger being what “mostly” doesn’t include.  Some examples:  1)  Golf Club Guy , who’s been an everyday regular since the first time I worked at Barnes and Noble, about 3 years ago.  He earned his nickname by bringing (you guessed it) his golf clubs to the store with him and practicing his swing in the parking lot.  As far as I know, the man has never purchased anything more than a cup of coffee, and yet he still feels it’s his right to remove an entire shelf of books from the poetry section and use the then-empty shelf as his personal coffee table.  He comes to the store in the evenings, but when he was a morning customer, he was joined by the likes of:  2)  Motorcycle Man , who is also still a regular and who earned his name because of the dirt bike attached to the back of his truck.  The truck, it seems, is his bedroom, while Barnes and Noble, he seems to think, is his living room.  Then there’s  3)  Obsessive-Compulsive Lady Part 1 , who used to grab a stack of books, sit on the floor in the children’s section, and tear tiny pieces off the corners of every page, creating a sea of tiny white scraps all around her.  She was soon asked to leave and was replaced by  4)  Obsessive-Compulsive Lady Part 2 , who comes in every morning and has to move a table in the children’s section over about 1 ½ inches before she can comfortably sit and do whatever the hell it is she does all day.  5)  Que Buena , a Hispanic man who hangs around the magazine section and, when a girl he likes walks by, stares her down and says in a thick accent, “Oh…My…God” real slowly, the words dragging out and morphing together in one long, creepy mantra.  6) Thank goodness, not a regular, is the man who was sitting in (again) the children’s section of all logical places, and masturbating while reading (naturally) a math book.  Yes, I said masturbating.  Yes, I said a math book.  Somehow, this escaped the notice of not only all the employees working, but the plainclothes security men as well.  He was only discovered, in fact, when the fruits of his labor landed (yes landed) on the leg of a woman sitting nearby, who screamed and went home to change, sanitize, and tell her husband.  She returned shortly thereafter to tell the manager, who was amazed to find the man still seated right where she’d left him, still reading the same sexy math book.  He was shocked to be asked to leave and insisted on a full explanation of what was wrong with his being there.  7) There was the man who was caught stealing but who managed to escape security losing only his backpack.  They opened it to see what all he’d taken and also discovered hypodermic needles and a small stash of what they figured must be heroin.  He returned a little while later, asking for his backpack, and ended up going into a full-on seizure in, of course, the children’s section.  It’s a wonder there are still parents willing to take their kids there.  Luckily, I was not there to witness those last two incidents.  Once, an employee found these romantic items in a lady’s bathroom stall:  an empty half-bottle of champagne, an empty box of stolen Godiva chocolates, and a pair of lady’s panties.  If  that  doesn’t make it into the Date Hall of Fame!  These are clearly the extreme examples and not a fair representation of the knowledge-hungry folks who comprise most of our clientele.  And it’s not even them who prompted me to write this.  It was the woman who called the other day with a four-book long list of requests.  The first book, we had in the store.  The second and third I happily ordered for her, but the fourth was one we couldn’t get because there weren’t any in our warehouses and the publisher information was not provided.  Her: What do you mean you can’t get it?! Me: Well, it’s not available from any of our warehouses. Her: Can’t you call the publisher?! Me: Sometimes we can, but this publisher’s information is not provided.  It’s possible you’d have better luck at Borders(?) Her: What’s Borders? Me: Another bookstore, like ours. Her: Can you call them, then? Me: You want me to call Borders? Her:  Yeah, call them.    I should say here that her request might have stood a chance of being honored if she weren’t so rude.   Me:  I’m sorry, we’re not really in the practice of making business for our competitor.  And this is where the non sequitur of all non sequiturs came:   Her:  Can’t you call them for me?!  I have a bad back!  I think the fact that most customers are familiar with this “Customer is always right” business just works to perpetuate their spoiled, demanding behavior.  A person should be able to work in the service industry and still retain a decent amount of integrity, no?  Just ask the manager who was lucky enough to have to clean up what didn’t land on the leg of sexy math man’s poor victim:  Some customers are not only not right: They’re just wrong.  Just really, really, really wrong.   '),(7434,669116,'07,February,2004','2004-02-07 00:00:00','    Refund checks are the best.  A tax refund check is the paramount variety—mine usually hovers around the $300 range and comes exactly when I need some extra cash (ok, that’s always, but still, they come when I  Most Especially  need some extra cash).  But there are other types of refund checks, of course.  Utility companies are big on these, particularly if the customer discontinues service.  They don’t normally exceed, say, 10 bucks, and it can feel pretty silly to cash or deposit one, but they’re nice all the same.  A couple days ago I received a refund check from our utility company, PG&E.  I could tell from the envelope that it was a check, which is always exciting, but I tried not to get my hopes up, knowing I couldn’t have overpaid them by much (we’ve been using the heaters with wild abandon).  Still though, I tore open the envelope and sought out the amount, picturing what I could buy (even if only a cup of coffee) with my surprise money.  Surprise indeed!  5 cents.  They sent me a check for 5 cents.  They spent  37 cents  to send me a check for 5 cents.  My check was signed by one K.M Harvey, CFO and Treasurer of PG&E, as well as Mr. Michael Donnelly, Assistant Treasurer.  I was flattered they both took the time.  First I just threw the check away, but shortly thereafter my friend Nick came by and, in telling him about it, I realized the situation was too ridiculous to simply let pass.  I fished the check out of the trash and began brainstorming with him: how best to handle this newfound wealth?  I would definitely be going to the bank with the check, but which action to take?  I could simply deposit it.  Or would it be funnier to cash it?  Nick thought I should cash it and walk away from the desk smiling, whistling, and flipping my (hopefully) shiny new nickel merrily in the air.  I could also deposit a portion of it and sign for cash back in the presence of the teller.  Nick said I should proudly declare, “Yes, and I’ll take two of your finest pennies, please,” pointing to the teller and winking with the word “finest.”  I could sign it over to Nick.  How about this:  I let him strong-arm me into the bank lobby, making a big show of it.  He shoves me against the counter, clearly forcing me to fill out the necessary paperwork and capturing the attention of the guard, who decides he’d better keep an eye on us.  He grabs me by the elbow very ostentatiously and leads me to the teller, where I give her his deposit slip and my 5-cent check, making a sad face and scowling at Nick like some little sister who just had her chocolate chip cookie bullied out of her.  5 cents!  There’s a notice on the bottom of the check that reads “Void after 90 days.”  How about I take it to the bank on the 91st day and, when they refuse to cash or deposit it, I dramatically throw my hands in the air yelling, “this is  bull shit!”  I have nothing more to say about that really.   '),(7435,669116,'28,February,2004','2004-02-28 00:00:00','    “Modern man lives in a state of low-grade vitality.  Though generally he does not suffer deeply, he also knows little of true creative living.  Instead of it, he has become an anxious automaton.  His world offers him vast opportunities for enrichment and enjoyment, and yet he wanders around aimlessly, not really knowing what he wants and completely unable, therefore, to figure out how to get it.  He does not approach the adventure of living with either excitement or zest.  He seems to feel that the time for fun, for pleasure, for growing and learning, is childhood and youth, and he abdicates life itself when he reaches “maturity.”  He goes through a lot of motions, but the expression on his face indicates his lack of any real interest in what he is doing.  He is usually either poker-faced, bored, aloof, or irritated.  He seems to have lost all spontaneity…He spends endless time trying either to recapture the past or to mold the future.  His present activities are merely bothersome chores he has to get out of the way.”  ~Fritz Perls  from  \"The Gestalt Approach & Eye Witness to Therapy\"  Kelsi read this quote to me from a book she’s studying as part of her Masters degree in Holistic Psychology.  It struck me because it scares the hell out of me in the way that only true (and truly disturbing) ideas can.  Who would want to live this way?  And yet who could protest the truth in the statement?  How to avoid the humdrum, the mundane, the mediocre, the settling for what we can convince ourselves is good enough?  I told Kelsi that if I had significant interest, I would invent a new kind of therapy called “Deathbed Therapy,” whose principle tenet is so basic that one need not even include an actual therapist.  It’s helped me confront fears and do things that seemed somehow beyond my capacity.  It goes like this:  When you think of something you want in life (long-term or immediate) but that seems scary, you just think, ‘If I’m lucky enough to take a graceful, deathbed exit from this life, and I’m lying there reflecting, is this gonna be one of those things that I think about NOT having done and go, “Stupid!  Why didn’t I just  do  that?!  And now there’s no going back and what was I so afraid of anyway?…In a few hours I’ll be dead and who cares?!’”  In the past, this exercise has enriched my life in little ways, from allowing me to have a damned good time while making a fool of myself on a dance floor, to bigger ways, like encouraging me to pour my heart out in a secret admirer-like love letter or taking a trip abroad by myself.  And I don’t regret those things at all.  In fact, I’ve never regretted the things I HAVE done.  Kelsi told me they already invented that kind of therapy: it’s called Existential Psychotherapy.  Fair enough.  I don’t mind sharing proprietary rights with the likes of Irvin Yalom.  Really, though…isn’t it that simple?  Can’t it be?  I know we’ve all heard this a hundred times, but that doesn’t make it less true; it simply makes us less open to hearing it because the message has become trite: Life is Short.  As proof, I offer that my 10-year high school reunion is next year already.  My friends are all getting married and having kids.  I refuse to call this a quarterlife crisis.    It’s a recommitment to the idea of living my life aware, awake, fully, and joyfully.  I want to say the silly thing and make the crazy decision and see the world and not think that because someone or something is different, that someone or something should be avoided, made fun of.  Think of the opportunities for fun and growth in opening up one’s mind to new experiences and different people.  Think how much time we’ve already wasted!  I offer this NON-rhetorical question to myself and anyone interested:  What is the thing I (you) want out of this life, right now?  What stands in the way of my (your) doing that thing, having that thing, making that thing happen, right now?  …  That’s not a good excuse.  100 years,  max .  Why  wouldn’t  I make the most of it?  I remember something an older man once told Kelsi while they shared a seat during a trip on a Greyhound bus:  “Might as well make a good go of it [life], otherwise you’re just turnin’ good food into shit.”   '),(7436,669116,'27,April,2004','2004-04-27 00:00:00','    Those who know me well would agree that my moments of patriotism are few and far between.  Most frequently, I am struck with American pride when I happen to pass by the Civic Center just before the start of the monthly naturalization ceremony.  Hundreds of new citizens and their families will be waiting in a long line, dressed in clothing often representing their native countries, there to celebrate the reward of years of work, and waiting, and work, and waiting.  It\'s a reminder to me that many sacrifice much for the freedoms our country offers...it\'s a reminder to me that I have a great deal for which to be thankful.  And I do consider myself blessed.  I suppose it makes sense that encounters with foreign-born people could bring forth patriotism.  And yesterday I had one such encounter.  I was working at the bookstore when a man walked up to me and said, \"So, do you sell a lot of books here?\"  I wasn\'t sure if he meant \"here\" as in the bookstore, or \"here\" as in the bargain section, where I was standing when he asked.  So I asked, \"do you mean \'here\' as in....\"  \"Here.  This bookstore,\" he said.  \"Well, yeah.  We sell quite a few books here.\"  \"Oh.  That is very surprising to me.  I am from Europe, and it is my understanding that people in the United States don\'t read...they only watch T.V.\"  I should say that his tone was not at all surprised.  It was simply sarcastic...like he\'d really been looking forward to having the chance to say this to me [to  somebody , some American, anyway].  \"No,\" I said.  \"A lot of people read here.\"  Again.  \"Really?  Yes because this is such a big store, such a selection.  I am very surprised you sell books.  In Europe, we all believe that you don\'t read books in America.  You only watch T.V.\"  Didn\'t he just say that?  Was it that great a line it warranted repeating?  I was surprised to find myself feeling a little angry.  \"Well, when you go back to Europe you can tell them people  actually  read in the United States...Not everybody in this country is as dumb as our president.\"  \"Yes.  I didn\'t think you read here.\"  What the?  The thing is, I can\'t completely blame him.  The fact is that a lot of people DON\'T read here.  And a lot of people do a lot of T.V. watching in Europe, too.  It is frustrating to me that Americans have such a bad reputation to overcome.  And yes, I understand its origins and the commonly encountered nationalism--coupled with ignorance--that keep the reputation alive.  But a stereotype is a stereotype, and it\'s annoying no matter where it comes from.   '),(7437,669116,'15,May,2004','2004-05-15 00:00:00','    It had to happen sooner or later.  There comes a day in every woman\'s life when she realizes she needs a new bra.  And that, my friends, is a sad, sad day.  It\'s not about the *idea* of a new bra.  Of course, new things are nice.  But the buying of the bra...now that\'s definitely less fun than a barrel of monkeys.  Trying on bras is second only to trying on bathing suits in the world of torturous activities in which women willfully engage.  I guess I was feeling masochistic yesterday.  So I battled the Valley Fair parking lot at 4:30 on a Friday afternoon (and won!  I found a spot right near the entrance.  Sure, it was the entrance on the complete opposite end of the mall from where I wanted to be, but it was a spot near the entrance nonetheless).  I dodged various obstacles including strollers, midgets, and people trying to sell me cell phones and then stopped in awe of the task in store for me when I came upon it--shining like some kind of half-angelic, half-evil beacon on the second floor of the mall--Victoria\'s Secret, all pink-and-white-and-blacklike.  \"You\'re on a mission,\" I told myself, and mustered up a facial expression that read \'I know exactly what I\'m looking for and where I\'m going to find it.  You may try to speak with me if you wish, but you\'ll be wasting your breath.  And no, I don\'t need a Victoria\'s Secret charge card.\'  So, either the sales *associate* was not skilled in face-reading, or my confidence diminished somewhere between the threshold and the first display containing underthings with no easily identifiable purpose or mode of operation.  And I\'m leaning toward the latter.  \"Hi!\" she said.  \"May I help you find something?!\"  I had a few doubts about whether or not she actually could, which surfaced when I noticed her  own  boobies were only partially contained and in danger of Janet Jackson-ing at any moment.  But I reasoned she was probably aiming for that look and acquiesced.  \"Yeah.  I\'m looking for a strapless bra.\"  \"Ok great!  We have a number of options for you.  Walk with me.\"  This was already bad news.  I didn\'t really  want  options.  I wanted there to be a standard-issue bra made for my purpose and I wanted it to be perfect.  But what could I do?  I followed her to a display.  \"Now,\" she said, and looked directly, directly at my breasts.  \"What size would you be, then?\"  \"Well it really depends on the bra.\"  \"Shall we measure you?\"  And with that, she pulled a tape measure from a holster and slung it around my back.  \'Whoah there, Little Missy!,\' I thought, a favorite line of my friend Wolfgang\'s echoing in my mind.    My inner puritan had a say as well: shouldn\'t we be behind closed doors (or at least somewhere other than in front of a huge, windowed store front in a busy mall) for this sort of activity?  \"Ok,\" she said.  \"We have this bra, which has a nice gel shaper in the cups.\"  I touched the cup and it moved.  It squished!  Bras shouldn\'t squish: boobies squish!  \"Uh.  No, thank you.  Do you have one without padding?\"  \"Ok.  Why don\'t I just take you to the dressing room?  All the bras should be there and then we won\'t have to walk around and find them all separately.\"  This seemed like a good idea until I came to realize that \"all the bras\" included only two other models, and they weren\'t, in fact, in the dressing room.  I stood there and watched while she rifled through a number of little boxes looking for my size in the two remaining strapless models.  When she saw they weren\'t there, she called an invisible fellow sales person using a little hands-free cell phone attachment kind of thing and requested the bras.  This would seem like a cool trick, except that she didn\'t seem to know how to use the thing (there was some button she was forgetting to push) and had to repeat everything.  Additionally, she kept dropping the little speaker part down her blouse and had to fish it back out every few seconds.  We waited until the invisible sales person showed herself with only one bra in her hand.  I began to think I just might have had an easier time on my own.  Did all this hullabaloo  really  save us the \"trouble\" of walking around to find TWO bras on our own?  \"Here you go,\" my helper said, \"this is the [fill in the blank with something sounding like \"Angel Sleak Body\" or something like that] bra.\"  For a second I thought she was going to accompany me into the dressing room and I had to ask myself whether or not I was entirely opposed to the idea.  Despite all the superfluous fuss and the gaping holes in what\'s clearly trying to be excellent customer service, there was something sexy about having another woman take such a keen interest in my breasts.  At that point we were allies, united in the goal of getting me the proper support.  Maybe she\'d want to see the once-impossible dream through to its fruition (?)  But she had other hapless souls to attend to.  I was now all alone in my battle.  And that bra...well, that bra beat me.  Once on, it looked like something from Madonna\'s \"Vogue\" video and felt like something you\'d force onto violent mental patients to keep them at bay.  I got dressed and emerged from the dressing room, downtrodden.  There was another boobie-bulging associate there, happy to take the reins.  \"How\'d that work out for you?\" she asked.  \"Um.  Do you have a strapless bra that isn\'t pointy or that doesn\'t make me feel like my boobs are sitting on a shelf?\" I asked.  \"Oh yeah, let me bring you the [whatever] bra...it\'s much better.\"  And she did.  And it was.  She even walked me through the incredibly complicated 5-in-one convertible strap options.  She told me I can have straps, one strap (those tacky 80\'s looking blouse-styles), halter top strap (nice feature, by the way), strapless, or racerback.    I didn\'t know anything about racerbacks ...didn\'t think I owned anything with a racerback ...am not given to racing.    But whatever it is--this racerback thing--my new bra can handle it!  So yeah, I bought it.  At the register I encountered a pleasant woman who made friendly conversation and who even winked a sweet wink at me when I passed her later in the mall.  If I hadn\'t heard her call another female employee by the name \"Mystique,\" I might have thought Victoria\'s Secret was a perfectly normal place, after all.  But it\'s not.  And you shouldn\'t go there unless you really really need a bra with racerback options.  I have to say, though, walking around the mall afterward with that little pink-and-white handle bag containing my new 5-in-one non-shelf-like, non-pointy bra, I felt a sense of accomplishment.  I felt I\'d made my breasts proud.  And yes, I hope it\'s a long, long time before I have the need to go there again.   '),(7438,669116,'23,May,2004','2004-05-23 00:00:00','    Recently, I’ve come to believe that few things are better than being witness to the realization of a dream.  I don’t mean learning that your best friend’s best sex dream just became a reality, although I suppose that would be cool too.  A couple of days ago I received a graduation announcement and attached letter (on  Yale  letterhead) from my friend Tamika (one of the true, brilliant goddesses I’ve had the fortune to know), who’s graduating tomorrow with her degree in Classics.  In the letter, she reminded me of a day a couple of years ago when we were at work and a cheesy classical music favorites CD was playing; “Pomp & Circumstance” came on, and she told me then that she was considering returning to Yale (where she’d been going before a short hiatus) to complete her degree.  And now here she is, cap and gown, “Pomp & Circumstance” and all.  The letter made me all kinds of teary, the good kind.  To know that a person I care about has worked so diligently toward accomplishing a goal, has been recognized for that hard work, and is now about to feast on the fruits—well I’ll say honestly that I can think of few better things in this life.  And there’s this, too: I can’t wait to see what she does next.  I feel that way about all the people I know, and things are very  good  in the lives of the people around me.  My Ma is set to become a certified Feng Shui practitioner in a couple of months and loves what she does.  Her and my Pa (who is now fully recovered from a hip-replacement surgery (!) he underwent a few months ago and is working and playing as hard as ever) just completed the landscaping—done entirely by the sweat of their own brows—of their gorgeous new home in Colorado.  They are both very involved in their community and I’ve never seen them happier.  My brother is now working on a television show down in SoCal, just moved into a nice apartment, and his co-ed softball team (the Isotopes, ‘Topes for short) is in first place ;)  Kelsi is chugging right along with her Master’s degree in Psychology, she’s written some great new songs recently, is going to Mexico for a month this summer and get her Spanish on, and is fixin’ to start her field placement position in the fall.  I am awash in admiration.  Nicole returned to school this semester and is writing again, which is a great thing for the world at large, by the way.  My gal\'s got stories that  need  to be told, and told well.  Her perfectly chosen metaphors and incredible knack for creating mood are enough to penetrate the heart and psyche of any reader.  Also, she just learned she\'ll be able to graduate as early as next Spring with a degree that I swear was designed just for her: it\'s called Creative Arts.  Awesome!  JD will walk this week (he graduated last winter) and just told me he’s going to Europe for the summer—Greece and Germany and wherever else the wind may carry him…weeeeeeeeee.  My friend Rocky is graduating with his Master’s in Philosophy and is moving back to Boulder, CO next month, a long-time dream of his.  My friends Dave (the one I accidentally knifed at work [who’s healed very nicely, by the way—good job Dave]) and Ramon are both graduating from SJSU this week.   I just learned my friend Brent is starting school this fall (art school in San Francisco, which he’s been talking about for years and at which he will absolutely flourish—he’s remarkably talented).  My sweet and lovely friend Sharon is now halfway done with her Master’s in Library Science, AND she recently chose a wedding dress (which is a wondrous feat in the world of wedding planning, I’m told).  Her roommate and our mutual friend Maribel is graduating this week with her Master’s in Education and will begin shopping for a Ph.D program soon.  The future Dr. Rotsko (my friend Nick) just had his 2nd paper published in an academic journal of philosophy and also completed his 5th straight semester as a President’s Scholar (4.0).  My friend Don from high school just finished his first year in the MBA program at the University of Michigan and accepted an internship on Wall Street (!) for this summer.  And I don’t know if this is a bigger accomplishment for me or for my beautiful honorary niece, Maya (about to turn 2), but she just learned to say (with some, ahem, assistance), and I quote: “Dodgers…awesome!”  It’s the cutest and truest thing ever to cross the lips of a toddler.  She’s a genius!  There’s a lot to celebrate these days.  I know, I know.  One should not live one’s life through the accomplishments of others.  And it’s not that.  What I mean by listing all this is to say that my being able to share in this good new is, I consider, a true gift in my life.  These are the kinds of things (and forgive me, my sappy language) that warm my heart.  I think it’s a somewhat commonly held belief that only relatively few people  really  change the world for the better—the Martin Luther Kings, the Mother Teresas, the Eric Gagnes :).  But I would disagree with that.  I believe each person—setting out to do what it burns within her or him to do—in so doing makes the world a beautiful place.  Imagine all those around you, whose lives touch yours—well, imagine them all truly fulfilled, truly happy.  That, in my mind, is a better world.  I want to take this opportunity to congratulate all these people and anyone else who’s got great things going on that I haven’t been nosey enough to learn about yet.  I consider myself grateful to know you and admire you, and I wish you—always—only the very best.  P.S.  If you\'ve never heard of Eric Gagne, I suggest a Google search :)  P.P.S.  If you\'ve got some good news to share, please write and tell me about it (you can access my e-mail by clicking on \"view my complete profile\").  I\'d love to hear all about it.  If it asks you to register, just make up a name or e-mail...I don\'t think it\'s too painful, but I haven\'t tried it so I wouldn\'t take my word for it.   '),(7439,669116,'03,June,2004','2004-06-03 00:00:00','    It’s Spring Cleaning time.  I’m speaking for myself, first off.  The other day I used a barbeque & Trivial Pursuit party I was hostessing as a good excuse to give my room and bathroom the old once over.  Except it was more like a twice or thrice over.  I even ended up re-painting a nothing-colored dresser that had long sat sadly in the corner of my room begging for life.  I gave it pink—bright magenta pink—which shut it up real good.  The cleaning felt great; seeing such a tangible change was inspiring.  But that’s not the kind of cleaning I was really referring to.  Spring also brings about a different kind of cleaning: one that’s at once more subtle and more palpable.  Spring, it seems, is a time for break-ups.  All around me, my friends are losing their honeys.  Some of these relationships have been years in the making.  Others have spanned a few precious months.  But when it comes to breaking up, time spent together is not necessarily relative to pain experienced when it’s officially over.   (Except it doesn’t seem to me that many relationships truly are officially over these days.  It’s like getting back together is the new breaking up.  But anyway…).   What is it about this time of year?  Is it really some kind of unconscious clearing out?  I know it’s painful to think about it that way, especially for those who have felt the sting of an empty bed lately, but I can’t help but think the witnessing of Springtime rebirth all around can rub off on the humans a bit:  What’s blooming? What needs re-seeding? What’s longing to be pruned? What is—beyond any doubt—dead?  I once broke up with a boyfriend around this time of year, but I think that actually had more to do with my birthday coming.  Just like I once broke up with a boyfriend soon after my grandfather died.  Just like I was once broken up with just before Christmas.  I think events like this trigger an inner Gallup Poll that asks (and asks loudly), “Is this where I want to be in my life?”  “Is this the person I want to be celebrating with now?”  My friend just told me about watching the full moon with her man last night and thinking how she’d watched the full moon on a previous occasion with somebody she felt less-than-thrilled about moon gazing with.  The same thing happened to me before.  I was on a date when the man said we should go for a drive and look at the moon, and I was so happy when driving circumstances prevented it from happening.  The voice inside me was screaming, “I don’t   want   to look at the moon with you!”    Or maybe it was, “I don’t want to look at the moon with   you  .”  And that date was a first and a last.  A full moon is nothing to be trifled with.  The moon, you see, should only be shared casually with coincidental bystanders, or passionately with good, good friends or the honey of your heart’s desire.  And when the full moon or the brilliance of Spring or a birthday or a major holiday comes around and the honey you’re with is not of your heart’s desire…well, that’s a tough day.  That’s a real tough day for everybody involved.  I want to wish a personal rebirth to all those who are aching right now.  I wish there were something I could do to make it feel better, but I know I can’t.  A breakup without mourning could only come after a relationship that was insignificant to being with.  I know yours were not of that category and so I hope your healing processes are thorough and—ultimately—refreshing.  Kahlil Gibran wrote, “when you experience sorrow, know that you are weeping for that which has been your delight.”  May you take this time for personal growth and, in the end, find someone with whom you can relax, someone who wants nothing more than to watch the moon with you.   '),(7440,669116,'31,May,2004','2004-05-31 00:00:00','    Much of the time, in the world of management (I\'m learning), there are way too many things to do.  But on rare occasions (such as tonight) you  find  things to do.  Or somebody else finds them for you.  Tonight, my boss decided that the gum needed to be scraped off the hardwood and tile floors around our store.  My first response when I heard this, was, \"There\'s  gum  on the hardwood and tile floors?\"  She looked at me like I was a leprechaun.    “Yeah,” she said.    This is how good managers are: they notice the little things.  So I grabbed a blade and set to scraping.  Scraping gum off any kind of flooring turns out to be—not surprisingly—quite a humbling experience.  But there are a few things that can take the humbling exercise up a couple of notches.  Like when a little girl (7 years-old or so) standing nearby a particularly gummy spot I was working on asked me what I was doing; said, “ewwww,” when I told her; then scratched her head and said, “it looks like you’re ruining the floor with that blade.”  Thank you.  “Yeah,” I said, “but we don’t have anything else to get it off with.”  Then her friend piped in and, pointing to a big glob of gum a few feet away, said, “There’s some more over there.  Are you gonna get that?”  Thanks.  Thanks again.  There’s something very intimate about digging into a months-old glob of gum with a box cutter blade and revisiting the smell of spearmint that had been locked in below the surface and trampled upon by a million anonymous feet.  “Wow,” I thought when the aroma hit me, “this was in some stranger’s mouth.  And now I get to scrape it up.”  It’s a kind of intimate you could very well live a lifetime without experiencing and be, methinks, just fine.  There are some tasks that I never give a second thought...things that are taken care of invisibly, things that just get done somehow.  I was thinking about this when we did inventory the other night.  The freaky people from RGIS were there (and I only say “freaky\" because it always creeps me out how there are like 1,500 of them who come swarming into wherever they’re working that night and just type numbers into little computers all night like  Night of the Living Dead: The Accountants Return .  And they hardly say a word.  If you’ve ever worked in retail during inventory, you probably know what I’m talking about.  Really, it’s a little scary).  Well, I realized that people don’t EVER give a thought to those people unless they’re working somewhere and it happens to be inventory night.  And then the RGIS people are seen as the enemy because they’ve invaded the store and who the hell wants to be there for inventory anyway?  The RGIS people are always being yelled at to rescan a section or to stop making a mess or to just be faster (if that were humanly possible).  Plus, they always work overnight and have to go on break at the same time like some kind of zombie herd.   Now THAT is a thankless job.  And then there\'s the work that janitors do every single day.  And I do mean every single day; the janitors at my store work 7 days/week.  They clean the store in the morning and then have to return in the afternoon to restock the bathroom supplies.  Sometimes they’ll be all dressed up because they had to leave a party or some other event just to come make sure the customers at Barnes & Noble would be able to comfortably wipe their asses.  These are customers, by the way, who can’t even manage to keep gum in their mouths.  I walked into the restroom the other day and a co-worker who was on her way out warned me not to go into the first stall.  ‘What’s in the 1st stall?” I thought.  But I knew I didn’t want to know what was in the 1st stall because whatever it was, it promised to be gross.  I wondered how many other people had walked in and walked right out of that same stall because they didn’t want to deal with whatever was waiting in there.  Then I wondered absently when it would be taken care of.  And I remembered: the janitor women would be in at 7:00 the next morning.  They would take care of it.  They would unclog the toilet or wipe the crap off the seat or pick the bloody tampon off the floor.   They  would do it.  That thought—thinking of it in those terms—makes me want to cry.    How could this be okay?  How could we have agreed to this setup, collectively?     And I think it’s kind of disturbing that scraping gum off a floor is an “experience” for me—something I can examine from afar like a detached anthropologist and have the leisure to write a blog about.  \' Hmmm…how did that make me feel? \'  And for others, it’s a living.  It’s what you do to make ends meet.  I wish I could give a big hug of gratitude to the people who work tirelessly in thankless jobs that are truly service oriented…like, the Ultimate kind of service.  I want them to know that there are people who see them, who appreciate them, who know they are so many heros working so many miracles, day after day.   '),(7441,669116,'09,June,2004','2004-06-09 00:00:00','    Well, just because I opened my mouth the last time, I had to stop by real quick to say that last night\'s Game 2 of the Finals was one of the best games I\'ve ever seen, of  any  sport.  Kobe\'s 3-pointer with 2.3 seconds on the clock to tie the game...priceless.  And how \'bout that Luke Walton?  I don\'t know about any of you sports fans, but I thought he was the sole source of Laker energy for a good part of the game.  Eight assists!  Ok, and that\'s all I have to say about that.   '),(7442,669116,'07,June,2004','2004-06-07 00:00:00','    When my brother came to visit me in San Jose last month, we took a little trip to the drugstore on the corner, where he found the best pair of sunglasses he\'d ever seen.  They hung lonely and faded on a revolving rack that had clearly been displaying the same glasses for at least 19 years.  The pair he zeroed in on were fluorescent yellow—faded fluorescent yellow—with a strange sort of ventilation device on the sides.  They supported a fortress of lens that would prevent even the most determined ray of light from penetrating through to the wearer’s eyes.  But the best feature, by far, was a faded tag attached to the center of the frames that read “Future Vision” in a 1980’s Tron-like computer font.  He had to have them.  “Hey bro, you  did  just have a birthday,” I said.  “Put them in my cart.”  This was before I flipped over the Future Vision tag and discovered that inflation had somehow retroactively caught up with these babies: their  original  price tag read “$12.99.”  We nearly peed our pants laughing.  “For those?!” Kris asked.  “You’ve got to be kidding.”  We went back and forth.  I said I didn’t mind paying that, that I just wanted to see him happy in the way that only these beauties could make him.  He thought $4.00, the price on a nearby (albeit less exciting) pair was more like it.  So we took both pairs to the front of the store and pleaded our case.  The counter man looked at me, looked at my brother, looked and the glasses, looked at my brother, and said, “$4.00.”  Kris beamed a victory smile and then turned to take the comparison glasses back to their rack.  The counter man mused in his thick Middle Eastern accent, “He looks very happy now.”  “You have no idea,” I said.  Kris wore those glasses for the weekend and then carefully packed them up for his return to Hollywood, where people apparently share his aesthetic sensibilities.  He’s been standing on street corners and had cars pull over, their drivers asking where they could get a pair just like his.  “Sorry,” he says, “you’ll have to go to San Jose.”  Recently, he was wearing his Future Visions on a street in Hollywood when a homeless man approached him and asked him if Kris wanted to buy the sunglasses  he  had to offer.  It seems the man knew his audience because my bro snatched up this pair (brown monster shades with the word \"sport\" written in gold on the side), which immediately took top billing in his sunglass wardrobe.  Well, I’m in Colorado right now visiting my parents.  They flew Kris and I both in, and Kris was kind enough to bring both pairs to share with us.  You’ll find us all modeling them below.  Please don’t be jealous.  I know it must be heartbreaking to realize there will likely never be another pair quite like the Future Visions, or another homeless man offering quite the same pair (for 50 cents) in YOUR neck of the woods, but I hope these pictures will be enough to bring you at least some of the joy we’ve experienced here this week.   '),(7443,669116,'06,June,2004','2004-06-06 00:00:00','    urlLink    The Original \"Future Visions&nbsp; urlLink   '),(7444,669116,'06,June,2004','2004-06-06 00:00:00','    urlLink    I Always Thought Sunglasses Should Cover at Least 50% of the Face&nbsp; urlLink   '),(7445,669116,'06,June,2004','2004-06-06 00:00:00','    urlLink    My Mom Comes Away Looking Remarkably Normal.  Seems Future Visions Were MADE for Her! &nbsp; urlLink   '),(7446,669116,'06,June,2004','2004-06-06 00:00:00','    urlLink    My Dad\'s Nose Never Looked So Small&nbsp; urlLink   '),(7447,669116,'06,June,2004','2004-06-06 00:00:00','    urlLink    Fall Line 2020&nbsp; urlLink   '),(7448,669116,'06,June,2004','2004-06-06 00:00:00','    I was thinking maybe it would be nice if the Lakers actually  showed up  for the next game of the Finals.   '),(7449,669116,'06,June,2004','2004-06-06 00:00:00','    I just want to add something to my last post.  I think I was insensitive in comparing relationship endings to Spring Cleaning.  In doing so, I did not mean to imply that the people involved were in any way comparable to things that should be thrown away and forgotten about.  What I mean to say is that, sometimes, assessing things leads to the conclusion that something in one\'s life isn\'t working, perhaps for  either  party (important to note)--and SOMEbody has to do the breaking up.  When I said that somethings were \"dead,\" I did not mean to say the people involved were no longer of any use, just that relationships themselves can grow stagnant.  I hope that was the way my words were taken, but in the event that they weren\'t taken this way, I apologize for my insensitivity and for any part of what I said that came across as flippant or trivializing.  I\'ve been through a breakup that rendered me very much lost and in despair, so I know what it\'s like...and it\'s no trivial thing.  To anybody whose feelings I hurt, please accept my most sincere apologies.   '),(7450,669116,'13,June,2004','2004-06-13 00:00:00','    When I was a young girl, I had a prayer I would say every night.  It started with the standard little kid prayer: “Now I lay me down to sleep…” and then changed a bit, I think because my parents didn’t want me to get freaked out by the whole “if I should die before I wake” business.  Well after that part, there was a little add-on wherein I gave shout outs to all the people I knew.  I still remember it, verbatim:  “God bless Mommy and Daddy  and Kris and Kisa [don\'t know why the talking about myself in the 3rd person; this was pre-Elmo]  and Nana and Tata [grandparents on the Mexican side]  and Oma and Opa [grandparents on the German side]  and all our aunts and uncles and cousins and friends,   and Kris and his spots ,  in Jesus’ name, amen.\"   Kris and his spots?   As you might know, Kris is my brother.  “Spots” was the kid term my parents taught us to describe Kris’  vitiligo  [vih-dill-EYE-go], technically defined as \"a skin condition resulting from loss of pigment, which produces white patches.\"  Kris\' vitiligo appeared on his hands, elbows, and knees, and the spots would alternately grow and diminish as he got older.  The doctors couldn’t predict whether they would ever stabilize, so the possibility of their growing to cover a significant portion of his skin was very real.  Thinking about that prayer now, I realize it would be difficult for anyone who overheard it to know whether we were praying for the disappearance of the spots or for their continued growth and power.  It almost sounds like we were hoping his skin would, everyday, look less and less like its original shade (described affectionately by my mom as “poodoo brown”) and more and more like the shade of my own (which she calls “stock white”).  But we knew.  We definitely knew.    The spots were evil. They made him stand out. The kids at school teased him.  The spots were a kid’s worst nightmare.  So imagine my surprise when the spots earned a new place of merit in our family just the other evening.  Indulge me while I provide a little background information.  Kris is currently working as a production assistant on a Lifetime television show called “Merge.”   Production Assistant  is a very broad title that could refer to anything from actual technical assistance to the picking up of the director’s dry cleaning.  Kris’ position falls somewhere in-between the two.  “Merge” is a quote/unquote reality show in which a couple that has recently married moves into a new place together.  The designers on the show decide which of their things can stay and go, and they do a makeover on the new space once everything is in place.    Production assistants on this show work as movers.  They’re the ones who put the couple’s things where the designer decides they should go.  Now, there are a few production assistants working on any given episode, and sometimes the camera shots are so quick it’s difficult to tell who’s who (because even the movers, in this case, get airtime).  My mom had taped last week’s show, and we sat down and watched it together when Kris and I were out visiting last week.  We were all sitting in front of the t.v. on-edge, hoping to catch a glimpse of my bro on the tube.  Sometimes we could see him very clearly and up close, and sometimes it was more of a challenge.  Once, it was up for debate because only the person’s arm was in view.   Was that him or wasn’t it?   Kris solved the mystery with this line:  “No, that’s me.  Look, you can see my vitiligo.”  “Huh,” we answered, collectively.  \"We’ll be darned.  That  is  your vitiligo.”    We joked a little about the vigor with which he reported the clue:  Look, look, it’s my condition.  It’s my disorder.  It’s my defect.  It’s  ME !!   It’s me.  The moment caused me to pause, thinking how far we’d come from hoping the “spots” would just go away.  Just go away!  And to be honest, they haven’t been an issue for Kris for a good 15 years.  In fact, he’s always had a strong sense of self and I suspect the condition was always more troubling for my mom than for Kris.  He was her child, and this was something that had the potential to cause him heartache.  Kids can be so cruel.  And yet even as he’s long past being self-conscious, to hear my brother so proudly identify himself using the condition we once saw as so threatening we tried to pray it away  every single night ...  ...that was kind of cool.  It’s interesting the way things that make us different can change from enemy traits to ally traits in the course of a few years.  I used to hate being so tall (particularly in middle school and high school, where Amazon women aren’t in real high demand).  I also disliked my name because it was weird and people couldn’t pronounce it (Keisha, Keeza, Kissa?).  Now I’m happy to be an Amazon woman with a weird name, partially  because  those things make me different.  Oh my god, I’ve lapsed into After School Special mode again.  Forgive me, but I really feel this way.  I hope that if I ever have a kid with three arms or an extra face in the middle of his back…well, I hope he or she can learn to embrace that trait and, eventually, use it to his or her advantage: “Can I lend an extra hand?”  “I know my back was turned officer, but believe me, I  did  see the robber, and he went  thataway !”  I mean really, who wants normal?  How boring.  And so yeah, God bless Kris and his spots.   '),(7451,669116,'24,June,2004','2004-06-24 00:00:00','    For anyone who doesn’t know, besides my job at Barnes & Noble, I have a part-time job teaching adult ESL (English as a second language) to employees of Whole Foods.  It’s a pretty sweet deal for the students because they don’t have to pay for the classes or even leave the store.  And some of them truly appreciate it and work hard, showing up for every single twice/week class and studying enthusiastically when they’re at home.  Francisco is not one of those students.  Now, he definitely  wants  to learn, especially because he’s been working at Whole Foods for roughly 15 years and can’t be promoted any further until his English skills improve.  And he tries.  But he has some deeply ingrained bad English-speaking habits that are incredibly difficult to break and, well, he doesn’t do his homework.  Anyway, I didn’t bring him up so I could publish his progress report, but I wanted to give a little background.  So, the other day Francisco was telling the story of a party a friend once threw him when he was working at a Whole Foods in Palo Alto.  This is how he tells it:  “She make me a party and she rent a, it’s a, how you call?  Like a when you have wedding?”  “A reception hall?”  I ask.  “Yes, well, and we have dancing and carne asada.  I make the salads and also some soup.  We have DJ he play the música, and it was the best party in my life: water for EVERYBODY!”  He says this last part while spreading his arms wide to illustrate the amount of people who enjoyed this particular aspect of the party.  His eyes are all lit up, like he is just  so proud  for having pulled off water for everybody.  “ Water ?!” I ask.  “What do you mean?  You just served water?  “Yes,” he says, still beaming his self-congratulatory smile, “for  everybody .”  I was laughing, but I was also trying to figure out what was so special about this water and his ability to procure it for everybody.  Was it a drought year?  Was it that they were extra hot & sweaty from dancing and he was happy there was enough for everyone?  And who serves water at a party anyway?  His classmates were all laughing at him until he clarified that he was talking about  aguas , fresh drinks they make in Mexico using water and other things like rice or fruits (horchata, tamarindo, those sorts of drinks).  He was proud because he made the drinks, and apparently they were a big hit.  Oh, he hates alcohol and wouldn’t allow any of his guests to drink it, so I guess he figured he’d better have a good alternative.  Somehow, despite the fact that he speaks English a lot better than my beginning students, Francisco’s particular brand of miscommunication always makes me have to try (often horribly unsuccessfully) and stifle a giggle.  The other day he was talking about how he bought a house in a town called Madera and was going to be renting it out.  It’s a big house, but he’s renting part of it to a family of three for only $300/month.  We were all intrigued, wondering how he’d make the payments if he had to pay rent here, too, and he was charging so little.  “Oh,” he said.  “There’s another man, he wanna stay there.  I gonna rent him a room for $300/month and 3 horses.”  “Three horses?” asked Rosi, another student.  “Yes, $300 a month and 3 horses.”  “What do you mean, you’re charging him 3 horses per month?” I asked.  “Yes, 3 horses.”  “Horses are very expensive,” said Rosi.  An image came to mind of a future snapshot of Francisco standing outside his home in Madera, surrounded by hundreds of horses, which have been breeding and making him a rich man.  In the snapshot, he’s wearing his same ultra-happy smile and spreading his arms to show off the fruits of his ingenuity.  I had to come to and get to the bottom of this.  “What are you gonna do with all those horses?” I asked.  \"How can he afford to give you 3 horses per month?!\"  “Whachyou mean?\" Francisco asked.  \"It’s just 3 horses.”  “Oh, 3 horses total?”  “Yeah, he keep 3 horses there.  I charge him $100/month each horse.”  “Oh!”  The collective light bulb went on and we all relaxed a little, now that we didn’t have to go on and worry where Francisco would get the money to feed 3 new horses every month.    I love my English classes.  And trying to help my students through their struggles with this language makes me appreciate it more, not in spite of all of its idiosyncrasies, but because of them.  Many concepts I teach require tangential explanations of the exceptions to the rules, the connotations if you say the thing the wrong way, the 16 alternate meanings (do you know how many ways you can use the words “pick up” together?), or the 6 words that either look just like it and mean something different, or sound just like it but are spelled differently and mean something else entirely.  You know, it feels good to know something well.  And it’s FUN to meditate on a question a student asks and think, ‘how  do  we use that?  Would that be more or less polite?  And why?\'   And also, my students know Spanish with an intimacy that my 2nd language learner skills will never allow me.  I think it must feel good for them to know something well, too.  I can see it in their faces when I ask them how to say something in Spanish, or when I say something wrong and they have to correct me.  There’s a feeling of accomplishment when you can confidently say, “I know this.  No, I KNOW this.  And I can help you.”  And having something to teach to people who want to learn is rather addictive.  So I’m going to grad school in the Spring and learn me some more. :)   '),(7452,669116,'07,July,2004','2004-07-07 00:00:00','    We got the news on a Tuesday.   I came to work at noon and was stopped in the doorway by my store manager, Tony.   “You know those display changes I told you about?” he asked in his slight Texas drawl.   “Yeah, what about ‘em?”   “I want you to put that all on hold for a coupla weeks,” he said. “We got somethin’ much bigger to work on now.”   I looked at him, anticipating the news, and he motioned for me to follow him to the managers’ office. While I followed, I tried to imagine what could be “much bigger” than completely rearranging three full sections of the store.   “Sit down,” he said.   I complied, then watched him smirk for a few seconds before I gave in. “Tony, what is it?” I asked.   He smiled for a moment more, then came out with it: “Bill Clinton,” he said.   “What?!”   “Bill Clinton.”   \"...is coming to our store?”   \"...is coming to our store.”   “When?”   “The 29th.”   “Of  this  month?”   “Of this month.”   I was back on my feet by that time. “No  WAY ,” I protested, a full smile taking over the bottom half of my face.   He took in my reaction for a moment, seeming to enjoy my bubbling-over-ness.   “Way,” he nodded.   And that’s how it started.   ***   Just under a week later, there were customers lined up outside the front doors to be the first to purchase Clinton’s memoir,  My Life .   The press (and, hence, the public) had gotten wind of our news five days earlier, and all the days in between had been spent—more than anything else—answering phone calls about the event.   We were given little information, but that didn’t stop customers from asking any possible question you could (never) dream up and expecting not only answers, but the answers they wanted to hear:   “I live in Missouri, and so I can’t make it to the signing. But is there any way you could send me a signed copy of the book?”  Sure thing, Ma’am. In addition to the 1,000 books he signs that day for customers who will have been camping out for upwards of 35 hours, eating nothing but 7-11 food and pooping in Porta Potties, as well as the 1,000 he signs earlier that day for customers at another bookstore in Berkeley, we’ll make sure he signs just 1 more—specially—for the woman in Missouri who couldn’t make it.    “If I get my book signed, but then I read it and don’t like it, may I still return it?”  This woman had never heard of eBay(?) And another thing: does that mean she always reads through books first and returns them if she \"doesn\'t like them?\" How cheap and shifty is that?!    An interesting phenomenon occurred.&nbsp; It seemed that, when it came to an event like this, people had never been more proud to be handicapped. All of a sudden, everyone was an invalid:   “I just had a surgery [no mention of what type of surgery; it could have been oral surgery for all we knew]. Will there be a special line for me?”   And not only that,  bona fide  handicapped people became a precious commodity:   “My Mom’s in a wheelchair, and I have to push it. Can we get in the front of the line?”   But the worst was the man who came out with this one after his 45 minutes of attempting to finagle special treatment proved fruitless:   “Well, my 9-year old son had cancer before. Will there be a special line for him?”   “You say he had cancer  before ?” I asked. “We...I mean...can he stand in a line? Can he  stand ?”   “Well, yes. But, I mean, he’s still...you know, I mean, he’s kind of sick.”   People will stop at nothing. Just when that kid thought he’d defeated cancer, just when he was trying to put the near-death experience behind him, his dad jumped at the chance to (if this is at all possible) relapse him himself. Cheesh.   But aside from all the phone calls (which were actually kind of fun to answer when the callers were excited and not trying to convince us they were special), there was a lot of work to be done.   A lot of work.   There were walls to be painted, carpets to be cleaned, displays to be merchandized, shelves to be dusted, books to be shelved, café items to be ordered, and everything to be everythinged.   It’s not that we feared Bill Clinton would run his forefinger over our dusty shelves, clicking his tongue and vowing never to come back to our filthy store. But having a former president as a guest tends to attract, well, everybody. All of a sudden we had managers from all over the district pledging their “help” for the big day. And the bigwigs…the bigwigs were coming out of the woodwork. With all those Suits on the way, the store had to be at what our manager—who’s sometimes given to hyperbole—called, officially, 112½ %.   And so we painted and we cleaned and we merchandized and we dusted and we shelved and we ordered and we everythinged.   Twice, for good measure.   ***   As the Big Day came closer, our collective nerves grew shorter. We all wanted to delete the words “line,” “book,” and “sorry Sir, one-legged people still have to stand in the initial line for a wristband, just like everybody else” from our vocabularies.   And some people began to show rabid enthusiasm. I heard from Tim—a self-proclaimed loyal customer who is apparently at our store “all the time” (though I’d never seen him)—no less than 5 times in the days leading up to the event. I think he figured a personal relationship with a manager would earn him the necessary ins.   And then, of course, there was Babette.   My mistake with Babette was to happen to be the manager on duty the first time she called. “This is Kisa, how may I help you?” I’d asked.   “Hi Kisa. My name is Babette and I just got out of the hospital. I got out 1 day earlier than I was supposed to, actually, just so I could make it to the Bill Clinton signing.”   ‘Great,’ I thought, ‘here we go.’   But it turned out Babette wasn’t asking for special treatment. She was just sharing this particular detail of her life in the same way she would eventually share  many  particular details of her life with me during her bi-hourly cell phone update calls from the road.   Babette, you see, was driving up from Texas.   I’d given many verbal sets of directions to the store that week, but directions to the store from Texas?   “I’m just outside of L.A.,” she said, during a phone call Saturday afternoon. “Is there a line yet?” (Note here, the event wasn’t to happen until Tuesday night).   “No, Babette. There’s still no line.”   “Yes, but have you overheard anybody talking about camping out?”   “Well, yes. I know some people plan to.”   “Well, are they there yet?”   “Uh, no. There’s no line yet.”   “Do you know when they’re planning to begin camping out?”   “No, I really don’t know for sure.”   “Well, I’ll get there some time tomorrow morning. I’m going to get a hotel outside of San José. Or do you think I should get one near you? I mean, is there one with windows that overlook your store so I’ll be able to know the second the line starts to form?”   And on, and on.   She was asking for me by name when she called thereafter. “ There’s  my favorite bookseller,” she’d say, when I picked up the phone.   “Hi, Babette,” I’d choke out.   “I’ve talked to you so many times,” she said. “I can’t wait to meet you.”   “Yeah…me too.”   ***   The magic phone call came late Sunday night, and shockingly, it wasn’t from Babette.   “Hi,” a young man’s voice said. “My name’s Rick, and, uh, my friends and I are a few hours outside of San José right now. Is there a line yet?”   “No,” I said. “Not yet.”   “Well, do you mind if we camp out?”   “Go for it,” I said. “It looks like you’ll be Mr. Line Starter Man.”   “Cool,” he said. “We’ll be there soon.”   And so Rick came.   And shortly thereafter, Babette came.   And after that, they just kept coming.   ***   The morning of the event, I spent an inexcusable amount time getting dressed. I wasn’t exactly banking on meeting Mr. Clinton, but just in case,  just in case …did I think he’d like the black slacks and jacket or the dark blue with pin stripes? Or should I wear a dress?   For the record, it was the dark blue with pin stripes.   When I arrived at work at 11:30 in the morning, there were people lined up all the way around the back of the parking lot. This was—by all accounts—a significant change from earlier in the morning, when the line had formed an entire loop around the store, continuing all the way down the street neighboring it, stopping somewhere near the freeway underpass.   Those people had changed their minds about standing in line and were now inside the store, forming mini-mob scenes around each individual manager, screaming like mental patients who’d just found out Soylent Green was (indeed) people.   I should have known the day wasn’t going well when the first thing I said to my boss was, “nice bullhorn.”   From all accounts I could gather, some kind of mayhem had broken out around 4:00 a.m., and&nbsp;all the commotion was a bit much for our&nbsp;rent-a-cops to deal with it. Somehow, people jumped over a fence and cut in line and all these people who thought getting there at 2:00 a.m. would be plenty to ensure them a signing, got—ultimately—swept out of the line.   And they were pissed.   There was a smallish coup, during which&nbsp;some people returned their Bill Clinton books and cheered each other as they did so. I don’t know that I was particularly affected by these people, who must not have been huge fans of the former president if they had no interest in his book beyond his signing it. But it was stressful, nevertheless.   And I understood their frustration. Some of them were very polite and pleaded their cases in civilized manners that made sense.  But there were the others who just wanted to yell at somebody, and for them, it was hard to muster sympathy.   Some people truly seemed to think meeting Mr. Clinton was not a privilege, but a right, which I found very confusing. How could he have possibly signed the books of all the people who thought he was cool? There just wasn’t enough time in the day.   My favorite customer was one who brought his walker in to illustrate the extent to which he was not qualified to stand in line. And he wasn’t among the fortunate to receive the necessary wristband. He was chewing out a completely innocent employee whom we’d borrowed from a nearby store, and he wouldn’t let up.   “Excuse me, Sir, may I help you with something?”   “Yeah. Are you a manager?”   “Yes.”   “Well, I just want you to know that this is a  black eye  on the face of Barnes &amp; Noble. And this is just the beginning. It’s going to get much,  much  worse.” He stared at me unblinkingly, waiting for me to react in shock and horror and a million ass-kissing apologies.   I guess my, “I’m sorry you weren’t able to get a wristband, sir, and I thank you for your feedback” wasn’t really what he was looking for, so he went on:   “And you should know that my son works for ABC news, channel 7.”   “Okay.”   And he just kept staring at me. I had to excuse myself after a few rounds of this because there was a fire to put out at a cash register. He came over there after a few minutes and interrupted a conversation I was having with another customer.   “This is a bad,  bad  thing you’ve done for San José,\" he said, shaking his head dramatically. \"A bad,  bad  thing.”   “Okay, thank you sir. Is there anything else I can help you with?”   “And just so you know, my son is Michael Something, ABC news, channel 7. Okay?!”   “Okay. And yeah, I actually heard you when you said that earlier.”   I had to wonder how many times this man had been unhappy in the past and thrown around the name of his son, as if his son were just around every corner, waiting for somebody to&nbsp;screw his dad over and ready to launch a full-scale investigation.   The man shuffled off with his walker, in search (I can only imagine) of another person in a suit to whom he could drop Michael Something’s name.   Shortly after that I was called to the café for reasons I can’t remember. At the counter stood an elderly female customer in a huge, aqua-colored muumuu with big, tropical purple flowers on it. She looked at me, read my name tag, and said, “You’re Kisa?!”   “Yes,” I said. “And you must be Babette.”   “How’d you know?”   “I don’t know. Just a guess.”   “I got my wristband,” she said, shaking her fist in the air and beaming the happiest smile I’d seen all day.   “Congratulations,” I said, and truly meant it. And for the first time since my arrival hours earlier, I felt good.   ***   There were, following, about 5 hours of relative calm. I took the opportunity to walk around outside and meet—among others— Rick, Mr. Line Starter Man, a friendly young guy who got his 15 minutes of fame on all four local news shows, including the Spanish station.   At that point, the fans were as mellow as people who’ve been sitting in the sun for two days  should  be, reading and eating and playing guitar and cards and Gameboy, and talking. They brought their lawn chairs, their umbrellas, their water bottles, their children, and—most of all—their Anti-Bush t-shirts. There was a like-minded happiness hovering in the air and, below that layer, a nervous sort of anticipation that can only come when people are about to experience one of the most memorable moments of their lives.   The feeling was there for me, too. I know  I’ll  never forget the trickle of unknown liquids dripping down my leg as Kenneth (a fellow manager) and I carried bags and bags of heaping garbage to the dumpsters around the corner of the store. “It builds character,” he said, when a woman nearby made a comment about the yuckiness of the work in which he and I were engaged.   ‘Amen,’ I thought—a little manual labor is good for the soul.   ***   About four hours before Mr. Clinton’s scheduled arrival, the Secret Service arrived. They looked just exactly like you’d imagine they’d look, wearing near-identical grayish-brown suits and, yes (some of them), sunglasses (they were standing outside for much of the time). They all looked to be in their early thirties, handsome, capable, and the most serious bunch of dudes you’ve ever seen.   They walked the perimeter, peeked under things, set up barriers, chatted with us, and mostly stood around looking like the bad asses they were.   The SJPD brought in the German shepherds in the early evening, and they sniffed every sniffable thing in the entire store. We watched, and ate, and watched, and waited.   ***   At 8:00 p.m., I took to my post, an area near some shelves that had been cleared away for the placement of bags. People would have to check in their belongings, be wanded by the Secret Service, and go forth to meet Mr. Clinton with nothing in their hands but their soon-to-be-very-valuable copies of his memoir.   We weren’t sure which entrance he’d chose, so my cohorts, Beum, Traci, and I, tried to relax and chat while watching for a glimpse of the familiar gray hair and feeling around for a tiny waft of the air of presidential-ness.   A loud cheering coming from the customers in the café, who were seated a good bit higher than the floor level and could see better, alerted us that the moment had arrived. Our District Manager, Greg, walked by and said, “guys...we’re rolling.” And he gave a definitive nod at the end.   We’re rolling.   The words never sounded less cheesy and more crucial than they did at that moment.    Oh my God, we’re rolling. Whatever that means, guys, we’re doing it. Here we go…    The line started moving from the front, and then all of a sudden, they were there: the fans, already done with the experience and ready to collect their bags.   Bill Clinton, you see, is a very fast book signer.   But the people didn’t seem to mind at all. They emerged flushed and joyful and some of them in tears. The women swooned. The men were proud. The kids had no idea what the hell was going on, but they were excited, too.   I kept trying to catch a peek in between two curtains that would separate every now and then. But there were hoards of people standing in the store behind me (behind rope barriers) trying to do the same thing, so the Secret Service man and woman standing in front of the curtains repeatedly drew them shut.   Then, he did it. The Secret Service man looked at me and gave a little wave. “Come here,” it said.   I walked over to him.   “You wanna see?” he asked.   I felt like an adolescent boy who’d just been offered a gander at his friend’s sister’s boobies.   “Yeah,” I said, my face—I’m sure—glowing all kinds of pure, Technicolor girliness.   He drew back the curtain a little. “Go ahead,” he said, and smiled a smile that said this was one of his favorite parts of the job: helping make a person feel that kind of yay.   In front of a black curtain, underneath bright semi-spotlights, stood my president, the one I’d voted for the year I was first eligible, the one I hadn’t quite let go of yet.   He was smiling and handsome and attentive, never looking up from the people immediately in front of him, whose books he was signing and whose hands he shook.   It was so hard not to gawk, but I pulled myself away; there were, after all, bags to be put away and then retrieved. But a moment later, I had another go.   A Secret Service man came out from behind the curtain and asked a nearby co-worker, Ellen, whether there was “anything we could do about the music?”   “What do you mean?” she asked. “He wants it turned down?”   “No,” he said, sounding offended at the idea. \"He wants it turned  up . We’re rocking back here.”   She looked at me.   “Well,” I said, “there’s a volume control, but it’s in the manager’s office. I’d turn it up for you, but I’m not allowed back there.”   “ I’ll  take you back there,” he said. And there was that same Secret Service man smile. Like he just knew he was making my night.   We walked behind the curtain and I tried to maintain conversation with him while thinking, ‘I’m in the same [albeit large and sprawling] room with Bill Clinton.’   I know this sounds silly, but it’s difficult to describe the level of energy and excitement in the room. It was more than I’d ever imagined it would be.   “He loves Motown,” the Secret Service man (Cory) said.   “Oh, you mean he chose this mix?”   “Absolutely. It’s his favorite.”   Which just endeared him to me even more.   Once the music was turned up, the mood in the store went from jovial to ecstatic. Everyone was laughing and dancing and singing (because  everybody  knows the words to Motown hits). I laughed when I had to go retrieve bags, which were collected in the area near where the Secret Service men searched people; it was funny to see these Most Serious Men in the Whole World barking orders like “spread your arms!,” and “turn around!,” and “put that down!” with this super jumping Motown music in the background. I don’t know where they find people like that, but I’m glad they do, because somebody has to keep a straight face when the rest of us are feeling silly and there are people to be searched. I mean, I’m sure that comes in handy.   We went through this routine for about an hour and a half: get the bag, give it a number, set it on the shelf, oh! here comes the person, “how’d it go? was it worth it?,” “yes, it was amazing,” get the number, grab the bag, give it back, see ya later.   There was an impatient mob of people outside who had the idea they might get their books signed if Mr. Clinton decided to stick around after the first 1,000 people had passed through. At one point they started chanting, “ Please sign our books, please sign our books!!!”  with so much force, it ended up sounding like the rudest polite request ever made, rhythmically, to the beat of angry fists. Those chanters put a bit of a damper on the moods of those leaving (otherwise) happy—the lucky ones. I thought they were going to get violent until the bullhorn came out again, asking them to shut the hell up and leave us all in peace...only, you know, in a more diplomatic way. Surprisingly, they mostly complied.   And when the last few customers trickled through the line and collected their things, a quiet settled in among us.   We weren’t sure what to do next, so we all began to congregate near the bag check and wait for some kind of sign. There we were, about 60 employees looking around at each other and the Secret Service men and hardly saying a word. I felt like I was in church, only some kind of newfangled, fun one where they blasted Motown music and the president popped in for a visit every now and then.   And then, even the music stopped.   “Okay,” Tony practically whispered, “let’s everyone get in a single-file line.”   We did so, still not quite sure what would happen next.   Then the curtains parted and we noiselessly filed into the room and took places on the media staging that had been set up directly in front of the signing table. Mr. Clinton didn’t look up at us then. He stood talking quietly with an older couple that seemed to be friends of his. We watched while they chatted, laughing every now and then like talking with the leader of the free world for eight years running was the most everyday thing they do.   And we still weren’t saying a thing, just watching as if through a two-way mirror, seeing something we weren’t supposed to see.   And then he shook hands with the couple, said “goodbye,” and turned to us.   “Is this the class portrait?” he asked.   We laughed nervously. He looked down at the place that had been cleared for him to sit, surrounded by young women, while the photographer took his light readings.   “Ladies,” he said, in his Arkansas twang, “I’m not sure you want to be seen with me like this,” he joked. And he sat.   One shot, two, three for good measure: we smiled the smiles of our lives.   The pictures done, he turned and began shaking hands. I couldn’t hear what anybody was saying to him, and I wasn’t sure I knew what to say myself.   When my moment came, I managed a simple, “thank you,” and tried to freeze-dry in my mind the two seconds when I held in my hand the hand of a man who’d held in his hand the hands of hundreds of country leaders and diplomats, the hands of millions of fellow Americans, the hand of John F. Kennedy himself.   When he was done shaking, he looked around. “I always thought I’d like to work in a bookstore,” he said. “Now, we’re in the children’s section? It’s huge. It’s real nice.” He walked around admiring the displays, and I thought of Sharon—the children’s department lead—all her hard work being taken in and appreciated by one of the most important figures of our century, indeed, our entire country’s history, and I was joyful.   ***   The night didn’t end like that, though. There were René and Kelly, two women who had been working away and hadn’t even realized we were all posing for pictures and meeting Mr. President. And they were, of course, both in tears.   So that was terrible to witness...the loss of a moment that could never been regained or done over.   And there were all the displays to put back in order, the line posts to be carted off, the food to be cleared, the signing area to be disassembled, the receiving room to be rearranged, and the closing numbers to be run.   And then we’d go.   Except, not quite. Tony issued one final directive as he drove away from his 20-hour day, preparing to start all over again a mere 6 hours later. “Kenneth,” he said, “let’s get this trash in the parking lot taken care of.”   It didn’t sound so bad at first. When I heard, I told Kenneth I’d be out in a minute to help. And when I saw what awaited us there, I wanted to run back inside and hole myself up in the deepest, indeed even the creepiest (where nobody would look for me), darkest corner I could find.   Thousands of people sitting in the sun for two days, not wanting to visit a trash can and accidentally lose a place in line, produce staggering amounts of waste.   Staggering.   We both took our coats off, rounded up our fellow victims, grabbed trash bags and gloves, and set to stooping and sweeping and picking up between thumb and forefinger some of the nastiest who-knows-what type of remnants imaginable. There were diapers and Popsicle sticks and burger wrappers and newspapers and Coke cans and cigarette butts; everything was sticky, and everything stunk. It dripped on us and gooed on us and clung to us.   At one point I stopped and looked up. The lights in the parking lot were low, so all I could see were the silhouettes of my coworkers outlined as they performed a task I’m sure they never dreamed would be par for the course, working in a bookstore. But I felt—in that moment—immensely proud. We’d just been through, I think, the toughest week and a half ever experienced at our store. And here we were, still giddy enough from our encounter to hunker down and do some grunge work.   And I looked down at my hand. The hand that had so warmly welcomed that of an historic figure an hour earlier was now reaching into the deep recesses of garbage piles and emerging covered in filth.   And I thought, that’s just how life is.   It’s a precious and disgusting mix of real and fantastical, of discouraging and invigorating, of the things you have to get done, and the elevated experiences that make all of it worth it.   I went home and settled into the most satisfying bath of my life, and half an hour later I watched while the drain carried away the sweat and the grime, and the magic of the evening.   '),(7453,669116,'17,July,2004','2004-07-17 00:00:00','    Last week I was teaching a lesson to my English classes, during which I introduced a lot of new vocabulary.&nbsp; My students were studying a drawing of a scene at a park and trying to name all the English words they could.  &nbsp;  They all wanted to know what the things floating in the pond were called.  &nbsp;  “Oh, those are “ducks,”” I said.  &nbsp;  “Hmmm,” said Rosi, “Hay veces que el pato toma mucha agua, y veces que ni agua bebe.”  &nbsp;  I looked at her.&nbsp; “There are times when the duck drinks a lot of water, and times when it doesn’t drink at all?”  &nbsp;  She nodded a little doubtfully, then repeated the phrase.  &nbsp;  “What the hell does that mean?”&nbsp; I asked.  &nbsp;  She started laughing, along with my other students.&nbsp; “I don’t know,” she said, “it’s a dicho [a saying].”  &nbsp;  Alfonso piped up:&nbsp; “I’ve never heard that one&nbsp;before.”  &nbsp;  “Yeah,” Rosi said, “me neither, but my friend said it once like it was famous.”  &nbsp;  \"Your friend,\"&nbsp;he said, \"she is Mexican?\"  &nbsp;  \"Yeah.\"  &nbsp;  \"Funny,\" he said, \"I never heard it in&nbsp;Mexico.\"  &nbsp;  “Well, is that even  true ?” I asked.&nbsp; “I don’t even think that’s true.&nbsp; I mean, I don’t know that much about ducks.&nbsp; But even if it is true, is that different from any other living thing?&nbsp; Why ducks?”  &nbsp;  I likened it to “feast of famine,” or “all or nothing,” and figured it was one of those things that just don’t translate.  &nbsp;  But translation problems aside, I’ve never been good at adages, or clichés.&nbsp; Sometimes it’s a comprehension problem, like with “A penny saved is a penny earned.”&nbsp; What is that supposed to mean?&nbsp; Does it suggest that if you were able to save the penny, then you must have deserved it?&nbsp; Or does it have to do with interest, as in, your penny will become two pennies--one saved and one earned--with time?&nbsp; I never gleaned much from that one.  &nbsp;  But most of the time, it’s a simple inability to get it right, and I think that has to do with hanging out with Kelsi for 12 years or so.&nbsp; My gal Kelsi, you see, is rather creative in the cliché department (paradoxical as that may seem).  &nbsp;  It’s not that she’s  trying  to be different, but the real ones just don’t stick with her.  &nbsp;  One time, she was talking about how she was excited about something she didn’t want to be that excited about, in case it didn’t pan out in the end.&nbsp; “Yeah,” she said, “I would like for it to happen, of course, but I’m just not gonna put all my chickens over there.”  &nbsp;  Our friend Nicole and I looked at each other quizzically.&nbsp; “All your chickens over there?”&nbsp; we asked.  &nbsp;  What she meant, clearly, was that she wasn’t going to put all her eggs in one basket.  &nbsp;  That one, I could deal with.&nbsp; You know, maybe she didn’t want to put all her chickens in one chicken-retainer-area because what if that one burned down and all the chickens went with it?&nbsp; Makes sense.  &nbsp;  On another occasion, she advised a friend of ours to just “let dead dogs die.”  &nbsp;  Well, yeah.&nbsp; I’d agree with that one.  &nbsp;  We’ve both had problems with those two damned birds.&nbsp; Should you not hide your light under two birds in a bushel?&nbsp; Or you can kill 2, but&nbsp;they\'re worth more if&nbsp;they\'re alive, in your hand?  &nbsp;  Why is it “Can’t see the forest for the trees?”&nbsp; Why not “through the trees”?  &nbsp;  Can anyone really take him or herself seriously when throwing around phrases like this?  &nbsp;  The worst, though…the one that always gets me…  &nbsp;  The other night I was playfully reprimanding a coworker who was being mean to his boss, who’d just sprung for his dinner.&nbsp; “Hey,” I said, “don’t bite a gift horse in the face!”  &nbsp;  Or was that “look at the hand of the horse that feeds you”?&nbsp; “Kick a gift horse in the knee?”  &nbsp;  I say we erase all those silly clichés from our minds.&nbsp; The reason they don’t stick is that they bear little relevance to our experiences today.&nbsp; Who even knows what a gift horse is?&nbsp; Would you recognize one if you met one?&nbsp; What is that--a horse somebody gave you, right?  &nbsp;  How many people are out there killing birds (my landlord aside)?&nbsp; And when’s the last time you were loading dozens of eggs into one basket (indeed, ALL the eggs you had!), then thought better of it?  &nbsp;  Kelsi had an interesting point: she said she can’t remember the clichés because she’s too creative; her mind won’t let her recycle the same hackneyed phrase over and over.&nbsp; And I think there’s something to it.&nbsp; I also feel lucky to be privy to her word incarnations.&nbsp; They make me stop and think.  &nbsp;  Hmm, maybe an apple a day  is  worth a pound of cure.  &nbsp;  Maybe I  should  nip it at the heels.   Maybe I shouldn’t give an inch…they’ll take all my chickens.   '),(7454,669116,'19,July,2004','2004-07-19 00:00:00','    I have pretty big feet (size 9), but my mouth is apparently more than adequately large enough to accommodate them.   I’ve set a precedent in this department.   I don’t do the kind of thing where I accidentally ask a non-pregnant woman when she’s due.  I know better than that.  My particular brand of foot-in-mouth disease has to do with inadvertently spewing double entendres and not identifying the potential for the 2nd entendre until it’s too late.   Some previous examples:   I have a friend named Renee who will be the first to tell you that she can’t hold a note to save her life.  She gets this from her Dad, whom we used to force to sing computer karaoke versions of songs like “Camp Town Races” and “My Bonnie Lies Over the Ocean” in his thick, Indian accent.  Now, admittedly, this wasn’t fair.  He’d never even heard these songs before—so he had nary a tune in mind to imitate—but even if he had, I promise he wouldn’t have been able to emulate them, and later we still would have giggled into our pillows while we listened to the surreptitiously obtained audio tape of his dying-frog-like musical stylings.   Well, anyway, Renee is like that, too.   But there is one song,  one song  she can somehow manage to sing and sound like a human: \"Amazing Grace.\"  If you asked her about this, she’d probably say it was divine intervention…God will arm you with whatever it takes to get you to sing His praises.   One night years ago, we were driving in her car, and she wanted to sing.  So, naturally, we had to sing \"Amazing Grace,\" which we did, and I was surprised at how well it went.  There were times when we—quite accidentally—fell into harmonies.  I mentioned this afterward.   “That sounded really good,” I said.  “Especially me.”   Renee is, above all her other good qualities, polite and humble, so she was rather shocked by my expressed, over-the-top conceit.   “Gosh,” she said, “that’s pretty egotistical of you.”   “What do you mean?”  I asked, honestly clueless.   “Sounded pretty good, especially me?”   “Ha ha ha!  Ohmigod!” was my response.  “No, that’s not what I meant!  I meant that note, when we sang, “that saved a wretch like  me ”…the “me” sounded really nice with our harmonies.”   I was happy the thing was cleared up, but I will never be able to erase from my mind the retroactive, 10-second-long embarrassment I felt at the idea of my being such a braggart.  That’s 10 seconds during which my best friend at the time thought I was no longer the person she thought I was, but, instead, somebody who would say (and mean) something like, “That sounded really good, especially me.”   I want that 10 seconds back.            About 6 years ago, I was living with my friend, Nicole, in Flagstaff, Arizona.  We were fixin’ to travel to the south of Mexico, where we would volunteer as human rights observers.  I’d sold my car and planned to use the money to live in Mexico (as had my friend Kelsi) and was working part-time to make extra money.  Nicole didn’t have a car to sell, so she had to work much more than we did.  And jobs in a college town like Flagstaff are very hard to come by.   She took a job as a chambermaid (Don’t you just love that word?  It sounds so naughty).  It was good work (which is to say it was work, period), except for the fact that she spent hours and hours alone everyday with nothing but her cleaning supplies and her walkman, which played a steady stream of Tori Amos, Liz Phair, PJ Harvey, and Fiona Apple.  See, Nicole had just ended a relationship with her first boyfriend, and she now had more than enough time during which to drown herself in mood music and to think about the relationship upside-down and inside-out and any other way it could possibly be thought about.  This was not a good thing.   But I digress.   In the evenings, Nicole would come home, melt into a chair, and tell me about her day.  Then we’d move onto girl talk, laugh and be silly for a while, then call Kelsi and take off in the 1972, magenta VW bug that was our one remaining vehicle to share, and find something fun to do.   In the midst of one of these pre-adventure conversations, Nicole came out with this one:   “Oh hey, I meant to tell you, I found my first condom today.”   I went immediately to a place you never want to go to in reference to your friend, and my response was, “Eww, you  SAVED  it?!”   She looked at me like you’d look at a person who just mentioned he’d had his eyeball pierced (on purpose).   “What?!”   Then she realized what I’d thought she meant, and she pictured me, picturing her, poring over little trinkets, ticket stubs, photos, letters, and other special mementos in a little secret box and coming across—in the midst of it all—the first condom she’d ever used, all crusty and stinky and memorable-like.   The scary thing is that’s  exactly  what I’d pictured, but I’d gone a little further, imagining her picking it up, rubbing it oh-so-lightly against her cheek, and thinking fond, sexy thoughts about her ex.   When I realized what she’d really meant (that she found a used condom in a hotel room she was cleaning that day), and she realized what I meant, we peed our pants (or, I should say, we thought it was kind of funny).   Now that I think about it, that was more a foot-in-mouth moment for Nicole than for me.  Or maybe it was one, each: ‘How could Kisa think that?!,’ she wondered.   ‘How could I think that?!’ I wondered, in turn.   And we each had 10 seconds of life we wished we could just say “erase, erase” to, and start over again.   ***   All of this was meant to be background information.  Jeez I’m longwinded these days.   Which brings me to these days…two days ago, to be exact.   I’d been working backup on a register for a little while, when an older man shuffled up to my station and set a magazine and book on the counter.   The magazine caught my eye—the most recent issue of  Esquire .  Donald Trump is on the cover with a bunch of huge, bling-blingie, hip-hop-style gold chains around his neck.   I picked up the items to ring them up, then made a rather lame attempt at small talk:   “ That’s  an interesting cover,” I said.   When the man didn’t respond, I looked up at him for a moment and smiled.  He was staring at me with a slight scowl.   ‘Best drop the small talk,’ I thought.   I asked the obligatory questions, “Do you have a Barnes & Noble Membership?  No?  Have you heard about the program?  Would you like your receipt in the bag?”   He grunted his responses in my general direction and I began to wonder, ‘What gives?’   I wondered until I picked up his items to put them in the bag.  It was then I noticed the cover of his book.   It was some kind of erotica, and the slightly smoky cover featured a ripped-stomached young man with an orgasmic look on his clean-cut face, as he was in the throes of an oral sexual act with (I’m assuming, based on the look of the back of the person’s head) another young man.   So, on the shopping list that day: The latest issue of  Esquire , and a little gay erotica.  Which (of course) not that there’s anything wrong with that.   Except I flashed back to the beginning of the transaction and my innocent little, “now  that’s  an interesting cover.”   And the silence, then the scowl, then the grunting…it all began to make all-too-perfect sense.   How to salvage this moment?  This misunderstanding?   My weak effort:  “You know, that picture of Donald Trump…I never saw him with necklaces like that before,” I offered, a tiny question mark hovering at the end.    What was that supposed to mean?  I mean, really.  Like when I mentioned the interesting nature of the cover, I meant that I just hadn’t seen Donald Trump in necklaces  quite like that  before?  Which—come to think of it— is  what I meant.  But is that even worthy of comment?  How many types of necklaces  had  I seen Donald Trump photographed wearing?  Or how many pictures had I seen of Donald Trump at all?  What’s more, who gives a damn?   The man looked at me with a face that said, “uh…yeah.”   That, I learned, is just one of the things wrong with making small talk.  From now on, I’m working through transactions in silence…speak only when spoken to, and then, only if what’s spoken is interesting enough to warrant a response.  This includes questions.   I’m kidding, of course.    Suffering from foot-in-mouth disease does make life a bit more interesting.  But here’s the thing: In considering the 10 second’s worth of misunderstanding, I always cringe, even years later, at the thought of being so grossly mis-taken.  And what really horrifies me is the thought that I’ve been misunderstood in similar ways countless other times, of which I never did become aware.  How many other people are out there thinking I’m an egotistical perv and a homophobe to boot?   Very disturbing, and, yes, a little funny.    I’m teetering out here on this limb.  Tell me this happens to you, too.   '),(7455,669116,'28,July,2004','2004-07-28 00:00:00','    Last year, while on my backpacking trip in Europe, I posted occasional blogs with the intention of highlighting certain adventures and letting all my peeps know I was alive and well (and tan even!), without having to send out individual e-mails.&nbsp; Internet cafés are expensive.   What I didn’t realize until after I returned was that I had a secret reader, who was following along like so many Metro ticket stubs at the bottom of my backpack, observing my every move.&nbsp; He said he was living vicariously through the blog versions of me and my meanderings.   This, it would seem, was creepy.&nbsp; Only it wasn’t, because the reader was Doug, Kelsi’s dad and my long-time dad-away-from-dad.   “Hey,” he said, when he came to visit in January, “why didn’t you have a link on your blog where people could contribute money to your cause?” (I’d gone without a plan and come home when I ran out of money).&nbsp; “Yeah,” he said, “I would have paid to keep you over there so I could keep reading about it.”   “Damn!”&nbsp; I said.&nbsp; “Now you tell me.”   I thought it was an excellent idea, if a long shot.&nbsp; And I also thought it cool that he’d been reading.&nbsp; Having an audience is nice.&nbsp; During that same conversation, he asked why I’d stopped writing and suggested that I might want to continue posting entries even though I was home from my trip—some people were (believe it or not) interested, he offered.   ‘Hmm,’ I thought.   ‘Hmmm’ turned into a post about cleaning bird shit off the neighbor’s house.   Doug’s been reading my entries and feeding back ever since.&nbsp; So it didn’t surprise me when he mentioned my blog while I drove him, Kelsi, and myself to Gordon Biersch last Friday night.&nbsp; Kelsi made a joke about the vulnerability of living with a blogger.&nbsp; “Hey!” she said, forefinger pointed straight at my ear from the passenger’s seat, “and don’t you write about this in your blog!”   “Yeah, don’t you write about us,” Doug joined in, a smile smiling through in his voice.   I told them I thought their optings-out funny because I’ve had the opposite request also, like when my friend Dave Marquez mildly chastised me for not writing about a group Trivial Pursuit match we’d both participated in, which he found plenty worthy of a blog post.&nbsp; I’d responded with something like a shrug at the time.&nbsp; I thought one post about the event (his, on his “Diary of a Poor Sport” blog) was probably sufficient.   Similarly, I told them, my Dad once hinted none-too-subtly that his then-recent hip replacement surgery would make for fun-filled reading.   The drive to the restaurant was short, so the conversation pretty much ended there.   We had to stand and wait a while for a table, and when our magic blinker finally buzzed and blinked, signaling the readiness of our table (I love that moment...I always feel like I won something), we were sat just the perfect distance from  Quasimodal Quartet , the jazz band we’d come there to see.    QQ  features a few former members of San José State’s English Department (which is why I knew of the group), and the assembled listeners included a sprinkling of acquaintances—fellow book nerds and wordy folk.   Let me repeat here that these people were  acquaintances .   After we’d been sitting there a little while, one band member (who was sitting out that particular night) approached the table.   “Hey Kisa, howya doin’?” he asked.   “Oh hey, Vince.&nbsp; Just fine.&nbsp; Have you met my roommate, Kelsi?&nbsp; Kelsi, Vince, Vince, Kelsi.”   “Oh, so you’re the roommate,” Vince said, referring to the fact that I’d mentioned Kelsi to him in some previous context.&nbsp; Each presented a hand for the shaking.   “Nice to meet you,” they jinxed.   I’m not exactly sure what happened next, but I suspect Doug misinterpreted Vince’s seeming familiarity with Kelsi for the mark of a man who’d read all about her.   Doug placed a hand on Kelsi’s shoulder and said, “yeah, she’s the one in the blog,” his eyes sparkling the sparkle those of any proud father would.   I’m not sure Vince heard him, because he didn’t reply specifically except to nod and turn to Doug.   “And this is Doug,” I said.&nbsp; “He’s Kelsi’s dad, out visiting from Phoenix.”   Doug extended his hand while smiling, his other hand touching his own chest apologetically: “ I’m  not in the blog yet,” he said, “but, hey, you never know.”&nbsp; He laughed, demurring.   Vince gave another frozen-smiled nod, this one a little more apprehensive than the last.   He had absolutely no idea what Doug was talking about.   I stifled my laughter while Vince and I chatted for a few minutes, then turned to Doug when he walked away.   “He doesn’t read my blog, Doug.&nbsp; I’m not sure he even knows what that&nbsp;is.”   “Ah, well,” he shrugged it off, and we went on taking in the jazz.   &nbsp;   Doug’s presumptions were working on this main level:   1)&nbsp;&nbsp;Everyone who has ever met me is reading my&nbsp;writing religiously,   and this sub-level:   &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; 1a)&nbsp;&nbsp;Each die-hard reader is interested in and following the stories of the action’s principle players.   I was touched by this, his vote of confidence.&nbsp; He was doing his job as my dad-away-from-dad in that moment: the encouraging, the proud, the one who says, \"That\'s my girl.\"   There are plenty of people who—despite their friendship and/or family relation—find it difficult to be supportive, for myriad different reasons.   “Nay.&nbsp; Nay,” they say.   And then there are people who are just always, always on your side, rooting in your corner.   I had the extreme good fortune to be raised in an immediate family that doubles as a cheer squad.&nbsp; “My name is [insert first and last name here] and there’s nobody in the world better than me,” is a phrase my family members often make each other utter when they sense self-doubt or sadness on the horizon.   We say this with tongues most definitely in cheek, but the sentiment behind the forced, Anthony Robbins-like self-talk remains true.&nbsp; Nobody in our family feels bad about himself or herself on another’s watch.   What a beautiful thing.   With this kind of background, I was naturally affected by Doug’s idea that others should be as consistently supportive and interested in my&nbsp;writing&nbsp;as he’s been.&nbsp; It was a nice, familiar feeling that made me all-over-again grateful for the swell folk that have become my chosen family…the family I’ve assembled to complement the one into which I was born.   Life has been good to me.&nbsp;  People  have been good to me.   So, I offer this as minimum payment on my Debt of Gratitude:   Thank you, Doug, for&nbsp;all to read.   '),(7456,669116,'03,August,2004','2004-08-03 00:00:00','    I’d read about these situations before. A few weeks ago, I had the fun job of going through the managers’ log and reviewing entries from the past 8 months or so; during that time, the managers saw fit to write about things like employee absences, tardies, equipment malfunctions, and incidents involving thieves, masturbators, and other characters who attracted the attention of the authorities.  Relative to some of the tales I’ve read, my management stint has been a breeze, and last night started as no exception to the rule. We had plenty of employees working, and they were in good spirits, to boot. Sales were strong, the place was clean, and the displays were looking good for today’s purported visit from our regional manager (which never happened—that is so damned frustrating!).  In many ways, it was business as usual.  At about 9:00 p.m., one of our regular customers approached me with some sort of strategic plan worked out in his mind, which he began to outline for me,  en media res , in his strong, Pakistani accent. “You know,” he said, “he’s not doing anything wrong, so I’m sure you can’t legally ask him to leave, but there is one route you could pursue.”  “Who?” I asked. “He who?”  “The man. He’s sitting in the chair next to mine. He isn’t bothering anybody, but he  is  taking cigarette butts and putting them in the magazines,” he said. “You could get him on that.”  “Well, what man? Where are you sitting?”  “Over here, in the children’s section,” he said.  If you’ve read any of my previous posts—namely the one about the customers who are not only not right, but very, very wrong—you’ll remember that any hint of trouble coming from a man in the children’s department usually means bad news a-brewin’.  The customer (Casey) began to lead me to the area, then stopped just out of earshot of the man in question. “Oh yes, the thing is,” he added, leaning in to whisper to me: “he smells like  piddle .”  The word seemed to denote a scent that would be far cuter than it could ever be offensive, but when I approached the subject of our discussions, I realized “piddle” was probably not exactly an accurate choice.  “He smells like dirty ash pit” might have been a bit closer to the truth. And it’s important, I think, to differentiate between objectionable odors. “Piddle” would have been unbearable. “Dirty ash pit” was not SO bad.  Casey didn’t seem all that put-off, either. He remained seated next to cigarette butt man for the remainder of the night. I suppose he just wanted to warn me in case some of the  other  customers might be driven away; our regulars like to think of themselves as little helpers…some of them regularly report strange goings-in and even see fit to give talkings-to to pervy ne’er-do-wells.  I thanked Casey for his tip and approached the man, who was clearly transient. I’m not one to kick people out simply because they’re dirty, or even because they smell (aside from the fact that there aren’t really legal grounds for this, my inner hippy says, “Well, as long as they’re not hurting or truly disturbing anybody—you know, to each his own.”)  “Excuse me, sir,” I began. “Have you purchased that magazine?”  He didn’t respond verbally. He just made a no-ish kind of noise.  “Ok, well then would you mind not putting your cigarette butts all over the cover?…you’re going to ruin it.”  Now, people ruin our magazines all the time; that wasn’t really my concern. The gross thing is that we’re talking a heaping pile of hodgepodge butts collected (I imagine) from ashtrays and trashcans and parking lots the city over. They were all different lengths and brands and—like any cigarette butts do—they just STUNK.  He gave half a nod, scooped them up, and put them back in the old Camel Lights box sitting next to him on the armchair.  “Thank you,” I said, and walked away.    It wasn’t until a couple of hours later, when we were fixing to close, that smokey man and I were reacquainted. It seemed he had passed out on the chair and completely missed both our closing announcements.  Casey the Regular, stood by and watched while I said, then said louder, then yelled, “Excuse me sir?  Sir ?  SIR ?”  He opened his eyes.  “You’re going to have to leave now. We’re closed.”  The man managed a little nod, then dozed off.  “No,” I said. “I mean now. We’re closed  now .”  He nodded again, this time with eyes closed, then let his head fall to the cushion.  Hmm. What to do?  I summoned Darrell, our big, Clark Kenty head cashier/bouncer (you think we don’t need one, think again).  Darrell yelled using his outside voice, and this time, the man didn’t bat an eyelash. Darrell (a former cop) said, “call the police.”  I really hated to do that. But I wasn’t about to sling the man over my shoulder and carry him out, either. The dispatcher asked if the man needed medical attention, and I said I didn’t know.  “If you don’t know,” he responded, “then it’s my job to send it. I’m dispatching it now.”  ‘Okay then,’ I thought, and man-sat while the booksellers straightened the store all around us.  The medical attention arrived 15 minutes later, making me happy for the fact that it hadn’t been a  true  emergency. I heard the sirens and waited for the ambulance to pull around to the entrance.  But no, this was no ambulance. No. This was…drumroll please….  The Fire truck!  Yay firemen!  I was too tired to be as excited as I might have been otherwise, but I will say that had I known it would be (not one, but two) fire trucks (each loaded to capacity with hulking heroes) sent in the event that the man needed medical attention, I would have exaggerated his condition, which truly just seemed to me like ¡Xtreme Sleepiness! (say it in monster truck rally voice).  The ambulance was there, too, and behind it, two police cars. It was overkill, to say the least, but it’s nice to have some excitement around.  The man perked up enough to identify himself as Bill, a 50-year-old diabetic.  Then he  really  perked up when they stuck him on the gurney. He threw his hands behind his head in super relaxation mode, and even smiled as they wheeled him out the doors.  Another night, another emergency call at Barnes & Noble.    I joke it about somewhat now, but it was a bit heart wrenching to watch this man—who was obviously having a tough time with things—be so needy and so unkempt and so unable to care for himself.  I thought about it a lot after work last night, and I was thankful for the systems we have in place in this country to help people. I mean, this complete stranger was left in our hands. Ok, clearly I can’t do much for him myself, not being a nurse or anything. But I can call other complete strangers, and they will come take care of him. I don’t have to know them; they don’t have to know him; nevertheless, the man gets the attention he needs.  Lest I take flak from any of my Republican friends a-lurking, I will say this much:  Yes, I know these services cost taxpayers money.  Yes, I know there are plenty of people who take advantage of the services of the state and become discouraged from pulling themselves up by the bootstraps.  No, I wasn’t all that surprised to see the same man, less than 24 hours later, hospital bracelet and all, plop himself into the very same armchair (!) as I was leaving work this afternoon.  At that point, I had to laugh a little.  We all do what it takes to get by.  Like the smoker who flicks his butts into the parking lot and says, “There are people who are paid to clean this…I’m giving them job security,” I say this: let the man sleep a while. Fires are rare, and how many times can you wash the fire truck to keep busy? It’s good for us to see our neighborhood civil servants out and about and do-gooding every now and then.   '),(7457,2753052,'31,July,2004','2004-07-31 00:00:00','   Just, y\'know, so I could say I work at the Quikie Mart. Not that I think it\'s a great career choice or anything, it\'d just be cool, someone says,  \"I\'m an industrial engineer\"   And then you say \"Oh, that\'s great, I work at the Quikie Mart for $6.50 an hour, I used to only make minimum wage, but I\'ve been workin\' for quite a while now. The boss says I could even get up to $7.00 if I work on my catfood displaying skills..\"   The look un their face...priceless no? '),(7458,2753052,'21,July,2004','2004-07-21 00:00:00','   Lindsey recently posted this to her journal...thing. \"I think, day by day, life gets progessively worse. And sometimes I wonder, \"Y\'know...maybe it\'s just the little depression stage all teenagers get...It\'s not my fault I\'m so fucked up...\" But what if it is my fault? I think I drive people away so I can be alone. I don\'t mean to. I don\'t WANT to; who likes to be alone? I don\'t know...It is so hard to live when you feel completely isolated. I can\'t think of anyone I know that feels like this...Maybe I do and they just didn\'t care enough to say anything. Right now, I think the only thing that keeps me from \"giving in\" would be Adi. It sounds weird, I know. My life depending on someone who lives hundreds of miles away. But I think, no, I know, she\'s the only person I told everything. My bestest friend, and I promised her that we\'d meet up and get drunk together, have a hella fun time. I don\'t want to break my promise, but if shit doesn\'t improve, what can I do?\" Yeah, there\'s more after that, but this is the part about ME, damnit. Now that my bout of selfishness is over...I must say... .....I love you too Lindzeh. I am so looking forward to holding eachother\'s hair up whilst we barf up everything we\'ve eaten from the past decade. Or something like that. '),(7459,2753052,'19,July,2004','2004-07-19 00:00:00','   Lord have mercy..such a good mood. If like, my way of life, did not completely outlaw dancing, I\'d be dancing around right now. Aaaaall is right with the world. Meh friend Lindzeh is back from exile and oooh, I\'m so happy. Yayness. '),(7460,2753052,'19,July,2004','2004-07-19 00:00:00','   You don\'t want to hear me sing. Trust me. Heather nearly dies at the sound of me lalaing. Anyway, on with what I\'d intended to talk about. Today, I had a rather good day. Got up around 6 pm, talked to Simon without fighting, had steak. All around fulfilling day.  I can\'t wait until I have money. I have soo many albums I want to buy: Franz Ferdinand Incubus-A Crow Left of the Murder Avril Lavigne-Under My Skin Blink-182...I don\'t really know the name, it\'s the one with Down on it.  So yeah, tha\'s about it for now. '),(7461,2753052,'13,July,2004','2004-07-13 00:00:00','   I\'ve considered everything, and right now the only thing that seems acheivable is crying... That friend, the one that put me on ignore, I got on a friend of mine\'s e-mail and sent him a message, here was his reply:  Tha\'s low...tricking me into reading an email from you..well I was gonna take you off ignore today..but I don\'t think I will. Hope you\'re having as much fun as I am. toodles...  He\'s..having fun..without me... I screw up everything, I really do. I don\'t want or deserve to walk this earth anymore. '),(7462,2753052,'12,July,2004','2004-07-12 00:00:00','   urlLink Learn How to do Self-Hypnosis  Hot damn! I do a google to find the calories in an orange, and this is what comes up! Well...a site with a link to a site with a link to this site, anyway. The audio part of it I thought was a bunch of crap, but that spinny wheel was one hell of an acid trip....quite a treat if you\'re not epileptic. '),(7463,2753052,'12,July,2004','2004-07-12 00:00:00','   Shit I Like: Pudding, pens, blank books, black, Heather, Laura, Simon, Shayla, Lindsey, Yahoo, condoms, the music of Franz Ferdinand, ditto System of a Down and Evanescence, White Oleander, the Georgia Nicolson Series, boiled chicken, England, air conditioning, irritating my parental unit, Amy Lee (lesbotronic moment), Scorpios, Juila Roberts (totally innocent), the art of Ursula Vernon, male hands.  Things That Should Be Obliterated Or Are Otherwise Disliked: Said parental unit, spanish rice or anything remotely casseroley, blonde hair, shaving bumps, stretch marks, pit hair, shedding of the hair, Jessica Orlikowski, Dale Swafford, R&B, Usher, Usher\'s nose, black men in white high tops, people who refuse to accept that they are white, people who tan, when your pubic hair bunches up and makes you look like you have a penis...tbc. '),(7464,2753052,'11,July,2004','2004-07-11 00:00:00','   Well that\'s fine and dandy for the bird, but I have bigger problems. My best friend has me on ignore (on yahoo). It is, mostly, my fault. But it\'s also his because he just didn\'t  get it.  Then again, I\'m not entirely sure what it was I wanted him to get. Perhaps it\'s for the best...but in the meantime...I\'m bored and I miss my monkey... '),(7465,2753052,'08,July,2004','2004-07-08 00:00:00','   My freezer/fridge...thing...it\'s haunted. I\'m not kidding. .... .... IT BREATHES LIKE DARTH VADER!! What if it\'s been in the food? Oh god...spectral hands fondling my food... '),(7466,2753052,'06,July,2004','2004-07-06 00:00:00','   Ok, so maybe they haven\'t been stolen. Maybe I\'m just out of clean ones. But still, it sucks big style. I have no underwear! .......  .......  ......Ok so maybe I do have the white granny panties at the bottom of the dresser...but I\'m SO not going to wear them...really...really really..I mean it. o.o '),(7467,2753052,'04,July,2004','2004-07-04 00:00:00','   Since when has it been the Fourth of July?! Why wasn\'t I informed?!?! Gah>< I\'ve been jipped out of my fattening Fourth of July snack foods! This is so not fair! '),(7468,2753052,'02,July,2004','2004-07-02 00:00:00','   I had intended to not go to sleep until tonight..but I poo\'d out around 9:30 a.m. or so. I did manage to drag myself out of bed seven hours later. I am still so tired...  In spite of the lack of sleep, I am in a rather good mood. Don\'t really know why...maybe it\'s my new bras...hmm... '),(7469,2753052,'01,July,2004','2004-07-01 00:00:00','   Apparently, dyeing your hair does not lessen the effects of blondeness. Around eight yesterday evening, I drank two cans of doubleshot. I still haven\'t been to bed. I\'m waiting for mum to go to work so I can go take a shower. Unfortunately, that\'s in two hours.  Ah well, my nocturnalization was bound to happen sometime, and at least now it happened quite gradually, unlike last summer\'s excursions. Besides to take a shower I will stay couped up in my room, for fear of bad TV infomercials. Best,  Adi with the bags under her eyes. '),(7470,2753052,'01,July,2004','2004-07-01 00:00:00','   So, being in an extremely good mood today, I have made myself some iced coffee. It is quite good, and I\'ve found coldness makes my lips look bigger (yay).   Also, I have put the first mood postit of the summer. It reads: be nice. I had one on the other computer too, but it said lighten up. Surprisingly, the postits tend to work. ...I really need a life... '),(7471,2753052,'01,July,2004','2004-07-01 00:00:00','   So, I got up about five seconds ago and I\'ve already been hit with bad news. My friend failed two very important exams. Granted, it doesn\'t really, y\'know, affect me, but it\'s still bad news.  In other news, I have $40 in my wallet just BEGGING to be spent...but at least half is going to be going towards new speakers. Woe is me. I need some new reading material, badly. I\'ve read White Oleander (amazing book might I add) at least three times now. Gah, I need a life. '),(7472,2753052,'01,July,2004','2004-07-01 00:00:00','   There are nine things very wrong with my life:(in no particular order) 1.I have my grandmother\'s nose. 2.I am fat, but not fat fat, just annoyingly plump. 3.I become easily obsessed. 4.I live in a town named for an animal whose name happens to be slang for the female genitalia. 5.I was once known as Meatloaf. 6.I have a cat whose favorite food is headphones. 7.I cannot abide said cat seeing me naked. 8.I have had a four month case of writer\'s block. 9.I will live in a box because of number seven. '),(7473,2753052,'24,July,2004','2004-07-24 00:00:00','    Are you sure it was a chick? Maybe it was a crazed fat guy? '),(7474,2753052,'23,July,2004','2004-07-23 00:00:00','    Omg. Okay. So like, we\'re riding home from Kim\'s....and like, we\'re almost there when we see this black car coming for us.    &nbsp;  Neither me or my mom pay any attention to it, right? Well....this chick jumps up (It was a convertible) WITH NO SHIRT ON!! O_O I saw her groodies!! I didn\'t want to!! >___<> '),(7475,2753052,'23,July,2004','2004-07-23 00:00:00','    Hard nipples...ow ow ow... '),(7476,2753052,'23,July,2004','2004-07-23 00:00:00','    You\'d expect this to be a sad post wouldn\'t ya? Nah, that just wouldn\'t be right.   I have friends who care about meh, even if that rat bastard doesn\'t.   *settles down with happy orange mix drink of death, and an eggroll, curtesy of Lindsey*   I have the fan on, it keeps givin me chills....I like^^   My nipples are hard...my ugly, ugly nipples...   *sings louie louie, not thinking how odd it is that I actually know the words*   Three nights..and days..I sail the sea   Think of girl, oh constantly   On that ship, I dream she there (oh boy, wet dreams!)   Aaaaanyway... I\'m gonna shift on out now. '),(7477,2753052,'23,July,2004','2004-07-23 00:00:00','    -dances around, eggroll in hand- ^___^ I\'m in heaven...La la la dee deee dooo...    &nbsp;  Let\'s see....Eggrolls, Spicy Chikun, Fried Shrimp, Shrimp fried rice....   CHECK!!!!    &nbsp;   &nbsp;   &nbsp;   &nbsp;  -toddles off, munching happily-    &nbsp;   &nbsp; '),(7478,2753052,'23,July,2004','2004-07-23 00:00:00','    Sooo much i could tell you about, bu i have to make this quick, it seems the lap dweller has replaced me already -sniff- ;_; Congrats Laura, i all yours...  No Adi i wasn\'t ignoring you, i was tired, comes from working too much bu getting paid lots ^^ Yay, bu right now i have a headache and a Rob behind me...  Uhm, yeah, today at works was great, much bonding with people there, water fight with Anna, constant pocking from Joanne, things are going good ^^ I actually fit in somewhere -sniff- =)  Uhm, yeah, thas it for now, downing lots of Smirnoff as we speak ^^ Post later maybe...  Monkeh Out '),(7479,2753052,'23,July,2004','2004-07-23 00:00:00','    What happens if you don\'t dream.....or rarely dream? o.o AM I WEIRD?! ;-;     &nbsp;  Be back later. '),(7480,2753052,'23,July,2004','2004-07-23 00:00:00','    Hi people. Last night I had the freakiest dream ever it was too weird for words. Did anyone ever see that movie the exocist? Well I had this dream that I lived in the house with the girl from that and the devil kept trying to get her so her mother had to rush her out of the house and she had to go in the sewer to try to be safe from him haha. My older sister lived in the house with me and we were scared that the devil was going to try to get us so we kept staying in rooms together. She was getting ready for bed and I had to follow her all around the house so I wouldn\'t be alone so the devil couldn\'t get me or her. I dragged all my stuffed animals around like a little kid too haha.&nbsp;It was scary cuz everything was all dark and we both kept randomly hearing growling noises haha. What a weird mind I haveeee.   Then I had another dream that it was my ex boyfriend\'s birthday party and me and Erin were invited and we went and ignored him the whole time and didn\'t get him a present and fucked around with other people there. How do I have a dream about satan and then a birthday party ignoring someone? Blehness.   Then I woke up&nbsp;of course and was up at 6:30 in the morning&nbsp;=(&nbsp;   Congrats on sleeping pattern&nbsp;Adi =D   Bai&nbsp; '),(7481,2753052,'23,July,2004','2004-07-23 00:00:00','    I finally fixed my sleep cycle. Goddamnit, I am the goddamn greatest. I am the goddamn champion. I\'m a goddamn goddess for god\'s sake. Goddamnit. '),(7482,2753052,'23,July,2004','2004-07-23 00:00:00','    I just went to IHOP and ate pancakes for dinner =D My sister and Dad ordered chocolate milkshakes while we were eating this guy walked in who looked like my ex boyfriend and i was wondering what it would be like if it was him and i grabbed my dad\'s chocolate milkshake and dumped it over his head.&nbsp;Oh revenge&nbsp;would be so sweet...hey that\'s a pun too! I really not mean, his big girlfriend with big boobs jumped on me the other day&nbsp;at the carnival =( I didn\'t do a thing either. Boys suck, we need more men in the world. *nods*    &nbsp;  Sorry you had to get shots =( I don\'t like them either. Also we can share simon\'s lap if you want though our combined weight will probably exceed 200 lbs and kill him&nbsp;=*(&nbsp;Wellllll ppls time to go sit on my butt and watch hours of tv. This is such a productive night.     &nbsp;  Buh bais!!! &nbsp;&nbsp;&nbsp; '),(7483,2753052,'22,July,2004','2004-07-22 00:00:00','    And muh blog\'s not gay!!     &nbsp;   &nbsp;  >_> Dernit... '),(7484,2753052,'22,July,2004','2004-07-22 00:00:00','    -turns up Eighteen Visions\'&nbsp;&nbsp;\"Tower of Snakes\" as loud as it\'ll go-    &nbsp;  Assanine font....>___>;;;    &nbsp;   &nbsp;  -looks at Simon. Sees this new Laura personage. Scoffs. >_>; Turns up nose and slides onto MarkEH\'s lap- Nu, none of chu know le Mark. Just someone to keep me entertained while....while.....I don\'t know. Anyways.    &nbsp;  Today sucked like, mucho-grande ass. Had to get two shots in both arms. BOTH! What\'s wrong with poking me twice in one arm? Jeesus, disable both of them, why don\'t you? I only have two. Sheesh. >_    &nbsp;  So, Aye-chee....^_____^ I so Japanese-d you it\'s not even funny. No, not funny....but uber-coot!!! Teh heh. Oooh. The  Real  Dracula is on the boob tube. Lovleh. YUSH!!! I shall make a sn one day, and it shall be entitled the following: Elizabeth_Bathory. Know who she is? No?? Then get a textbook.....yuh turd. >_____>!!!!    &nbsp;   &nbsp;  (Don\'t like the pink? TOO BAD?! =D) -now listens to Atreyu\'s \"Demonology and Heartache\"- '),(7485,2753052,'22,July,2004','2004-07-22 00:00:00','    One takes note that no splendiferous Adi was mentioned...   Btw, are you (Simon)&nbsp;not logging into yahoo or just ignoring me? '),(7486,2753052,'22,July,2004','2004-07-22 00:00:00','    It\'s jealous cuz i have a Laura and a Lap Dweller...Least i think i still have a Lap Dweller? =( '),(7487,2753052,'22,July,2004','2004-07-22 00:00:00','    Perhaps the blog just doesn\'t like you? lol '),(7488,2753052,'22,July,2004','2004-07-22 00:00:00','    I just spent ages posting and for some reason it won\'t display the text, screw writing that again... '),(7489,2753052,'22,July,2004','2004-07-22 00:00:00','    Today my Dad comes home I think *cries* he going back to california monday though!!!!! yay!    Is this color orange or red? I always get those&nbsp;two confused hehe '),(7490,2753052,'22,July,2004','2004-07-22 00:00:00','    Ok, not literally (thank god) but I did just get out of the shower, I feel pretteeee oh so pretteee. In other news I am extremely bored, and I have a funny little pain in my toe that will not go away.    And Simon, if you disown the Lap Dweller for the Laura...I\'ll...I\'ll...well then you suck. *nods* '),(7491,2753052,'22,July,2004','2004-07-22 00:00:00','    Simon speled \"sausages\" rong les al laff at his typoessssss =D lmao! hehehehe (In case your wondering ;p that was a joke and I really don\'t spell THATTTTTTTTT bad also i would never make fun of the cute lil monkeh *pet pet pet* hehe =D)     &nbsp;   I just went somewhere with my sister and auntie. They are meanies and wouldn\'t answer my questions so to be obnoxious i repeated myself 50000 times til they answered it.     &nbsp;   Me: WAS IT BASED ON A TRUE STORY WAS IT BASED ON A TRUE STORY WAS IT BASED ON A TRUE STORY WAS IT BASED ON A TRUE STORY WAS IT BASED ON A TRUEEE STORY!    My Sister: YES YOU ANNOYING FUCK!     &nbsp;   Yay mission accomplished.     &nbsp;   I feel fat I ate alot I can\'t stand *falls on the floooooor* bleh hehehe 850 lbs. here i come!     Something really funny just happened I was in the car and my auntie had an ice cream cone and i went to steal a lick and the car went forward so my face got launched into it then i had to use wet naps to clean&nbsp;up my face and i&nbsp;almost choked i was laughing so hard =D i also got yelled at but that\'s ok&nbsp; &nbsp;    &nbsp;   In this book my auntie read a prostitute was talking about sex and how the man stuck his eel in her cave. LMAO!!!!!!!!!     &nbsp;   I so easily amused. I the greatest cuz I posted 4 times in one day k bai '),(7492,2753052,'21,July,2004','2004-07-21 00:00:00','    I dun want this to get all soppy or anything, dun wanna annoy the rest of you ^^;  -pets Laura and hugs- You can keep friendships, like 3 years a memory loss and we still going strong =D  On another note, who has the carrier? Did you pack it Lindz? Cuz i think i forgot? ^^;   Monkeh Out '),(7493,2753052,'21,July,2004','2004-07-21 00:00:00','    Hi all I am sad cuz realized that I can\'t keep friendships =( sucks for me and no one cares ;.;     &nbsp;   I learned how to change text color today this is cool&nbsp;I really am 17 i\'m&nbsp;just really stupid it\'s not&nbsp;my fault though x.x     &nbsp;   Blah sorry this post is sooooooooooo&nbsp;gay I\'m going to mope and eat ice cream so I can&nbsp;move from 117 lbs to 850, shrink&nbsp;from 5\'5 to&nbsp;5\'0,&nbsp;and join the fatties on Maury so&nbsp;everyone&nbsp;instantly&nbsp;luvs me since American&nbsp;seems to&nbsp;have tons of sympathy for HUGE HUGE HUGE HUGE HUGEEEE HUGEEENESSSSSSS   &nbsp;   I like pink &nbsp;    Bai people    &nbsp;&nbsp;  &nbsp; '),(7494,2753052,'21,July,2004','2004-07-21 00:00:00','    *Growls at Lindsey*   Mine mine mine mine mine mine mine mine mineeeeeeeeeeeeee mine imine mine mine mine mine mine mine mine mine mine mine mine mine mine mine mine mine mine mine mine mine mine mine mine mine mine mine mine   he and his lap mine!   *giggles*   I want a pet pet =( even though I get them lots anyway =D &nbsp;   ~meeeeeeeeeee   P.S. ppls! This site needs more penises! '),(7495,2753052,'21,July,2004','2004-07-21 00:00:00','    -gnuh!- Ok, be gentle with the lap...It\'s not a toy o_O  Watching Laura on cam, gonna watch Spiderman 2 in a sec with Sarah and Jeff, keeping this short, wordage not functioning, need chocolate...  Of course&nbsp;I didn\'t forget about teh lap dweller ^^ -pets Lindseh- But...Uhm...You appear to have competition o_O;;  Monkeh Out '),(7496,2753052,'21,July,2004','2004-07-21 00:00:00','    Adi is soooo cool wanna know why cuz she helped me fix my cam not even simon could fix and he a computer geeeeeenious!!!!!!!!!!! =D Just kidding Simonnn   ~meeee '),(7497,2753052,'21,July,2004','2004-07-21 00:00:00','    Well....sort of...I\'m rebelling against Peptobismol-a-go-go Land with my blue font^^ But chu go ahead and keep things pink Tiffers, I\'m sure Rob would\'ve wanted it that way. *fits of giggles* ahem... '),(7498,2753052,'21,July,2004','2004-07-21 00:00:00','    Hi I\'m Laura I was invited by Adi =D I have driver\'s Ed this morning which really sucks ass and I\'m tired. K, before I go I would just like to say:  *Sits in Simon\'s Lap and clings to him*  IT\"S MINEEEEEEEEEEEEEEEEEEEEEEEE!  =) baibai! '),(7499,2753052,'21,July,2004','2004-07-21 00:00:00','    I\'m in heaven...La la la la...Piiiinkyness all around...!!    &nbsp;   &nbsp;  I sure hope Simon didn\'t forget his ever-faithful lap dweller....Eh...EH?!     &nbsp;   &nbsp;  Yep. I\'ve been forsaken. ;-; '),(7500,2753052,'21,July,2004','2004-07-21 00:00:00','    Oh ho ho ho! Lookit this! I CAN REBEL!!! '),(7501,2753052,'21,July,2004','2004-07-21 00:00:00','    Is what you should\'ve named this place, maya luboyv. I\'m inviting Heather to be here too, cos y\'know, the more the merrier, right? Right, damnit. '),(7502,2753052,'21,July,2004','2004-07-21 00:00:00','    So yeah, i hear the other blog got destroyed? That sucked...-looks around- I suppose this will do for a new one...Maybe...Is there a coke machine anywhere? Hmm...OMG I have options! O_O They\'ll regret giving me choice ^^  Ohhh, chunkeh =D   Awww, this is crap ^^  &nbsp; Hmm, look at me, i\'m sunburnt =D   ...I can put pictures here...Mwaha...We\'ll toy with that another day =D Right now i gotta get back to Laura before i go bed, toodles ^^ Here\'s hoping for lots more bloggage ^^ '),(7503,2753052,'21,July,2004','2004-07-21 00:00:00','    Pink?! Of all colours, you picked PINK? Lord..*do dooo da dooo, da dooo* (pink panther theme) '),(7504,2753052,'31,July,2004','2004-07-31 00:00:00','    Hi peoples,      I went to my Merri\'s bday party, slept over her house and saw the village and then the next day went to her block party. *cough/sneeze/coughcough* I was sick the whole time with my bad cold and kept taking medicine every 4 hours can\'t be good. *coughcough* I ate ice cream cake and smeared it over my face and at the block party i face painted little kids and they were oh so cute and this one girl took my hand and showed me something and told me she wanted me to paint her as a donkey but then i said i couldn\'t do that so she said i could do a panda on her and then the girl after her said she wanted to be just like me cuz erin painted my face as a kitty so we painted her as a kitty aw they were so cute. Eric was there and he was fucking annoying like last time so I gave his brother a water balloon and had him hit him with it for me it was cool. Me and merri and everyone had a huge water balloon/italian ice fight and i got covered and soaking wet. I was kidding around and I put a water balloon down my shirt and said i had a third boob and then it popped right in my shirt. I was drencheddd. *sneezesneezecoughhh* x.x bleh I got sunburned a little cuz i didn\'t put on sunscreen boohoo. oh well it doesn\'t hurt too bad. We played DDR at like 1 am and i\'m very bad at it i kept getting E\'s on light mode and D\'s on easy. I suck and can\'t dance on arrows. oh well. The most fucking hilarious thing happened at the block party. These guys who were only one year older than me were wasteddd drunk and they got in the middle of the dance floor, took the microphone from the DJ, made him put on Backstreet boys \"I want it that way\" and started singing it at the top of their lungs. They were dancing sooo gay to it and we went over and watched them and we were on the ground laughing our asses off. It was so funny. Rory was at the block party and I don\'t really like her cuz she acts like a man and at one point she even put joey in a headlock he\'s only 10...that\'s stupid. I dealt with her though.      Simon why you always make posts about girls at work and stuff. Just wondering...   I surrounded by guys in school and none of them like me and everyone always tells me i\'m pretty so i don\'t get it. I don\'t think it has anything to do with the way you look i think it\'s cuz your quiet and shy, i\'m quiet and shy too. People just leave quiet people alone because they are intimidated by them. People would much rather be around a person who is loud and talks alot you know. You can\'t change your personality though just let things happen instead of wondering why cuz if you always wonder then you\'ll drive yourself insane. There isn\'t anything wrong with you you just haven\'t found right person yet who likes a quiet personality. Don\'t worry about it there are tons of girls in the world. *coughcoughxlikexmexcoughcoughsneeze* But you know like if that girl does like you then you should go for it you just have to find some confidence it doesn\'t have to be alot just a little...      *sighcougohcoughcoughcough!*      After Merri\'s block party i went to my cousin\'s grad party for 10 mins and i found out i have these cousins who are insane. This kid is like 7 and he kept grabbing my boobs and my sister\'s boobs and walking up to us and saying wanna hear the most annoying sound in the world and squealing really loudly. Ugg it was annoying he was cute though in his own weird way haha.      AHHH I CAN\'T TAKE THIS SICKNESS ANYMORE! *crawls into bed and pulls covers over head*      bai people '),(7505,2753052,'30,July,2004','2004-07-30 00:00:00','    Well, if you want a serious answer, there are a number of possibilities. The way you carry yourself. You\'re just too shy. Or, more likely, it\'s that women do not generally look for a boyfriend at a clothes shop. Moving on, I know she\'s grounded, but I miss my Lindsey. And where is Laura? o.o Usually she\'s got about a million posts a day...tha\'s what she\'s here for, dernit. In conclusion I would just like to say that I am extremely bored, and will welcome any sort of entertainment, including naked photos of your pets. XD '),(7506,2753052,'30,July,2004','2004-07-30 00:00:00','    Oh...My...God...  Not one single post since I was last here? What is the world comming too? o_o;  Uhm, yeah, i\'m tired, i should be in bed now, work tomorrow and all...Just watched Cyote Ugly ^^  Wow, when did i get so boring? lol...-sigh- Oh well, post something interesting to give me ideas =D  Oh yeah, just a random thought i\'ve been having...Day in and day out, I am surrounded by literally hundreds of girls, some English, some foreign, bu what i find interesting is how none of them, not one, take any interest in me at all...Can anyone explain? Is it my face? I know i dun smell bad o_O; Yeah, well, that was a randomness thingy of the head place...I\'ll shut up and let you get back to your happy lil lives ^^  Monkeh Out  P.S. Yes, the d in invisdible was intentional... '),(7507,2753052,'30,July,2004','2004-07-30 00:00:00','    I blame the gay second rate blog that hates me...Stupid thing, not showing my posts...Ah well, tis not like you\'re missing anything interesting...  Sooo tired, today started crap, dumped in the stock room to work my ass off, woot...>>; It soon got better, after lunch I was dumped on first floor tills, folding womens underwear is a lot more fun than it sounds =D I did something realllyyyy funny, bu i forgot what it was >>; Ah well, it\'ll come back to me next week ^^ Anyway, yeah, fun cuz me and Anna had a pen war =D Stealthy pen stealing and the likes ^^ Also I think one of the new girls that started today likes me, I could be wrong, very very wrong...I shall investigate further ^^  I have such a sad life >>; ANYWHO! I\'m off to play Manhunt, a game that got banned yesterday ^^ I do so love random violence =D Damn you all for not being here =(  Monkeh Out  P.S. Aint me posts boring? Mwahaha ^^ '),(7508,2753052,'29,July,2004','2004-07-29 00:00:00','    Today was interesting. I went to my friend merri\'s house and was tortured by 10 yr olds (aka merri\'s brother joey and his stupid friend eric).&nbsp;Eric screamed across the yard to his 14 yr old brother that i don\'t even know&nbsp;(staying at neighbor\'s house) that i was a hooker and would give him a dance in my underwear, also that i would show him my boobs and vagina. How lovely just about the whole neighborhood heard. So then I told Eric that i loved him&nbsp;to freak him out. He said sorry&nbsp;i don\'t feel the same way or some&nbsp;shit i really freaked him out it was awesome.&nbsp;To get them back even more&nbsp;for this me and merri made them popcorn and drenched it in hot sauce, but it backfired as they ate it and loved it. Actually they didn\'t even taste&nbsp;the hot stuff&nbsp;they thought it was blood but we told them it wasn\'t. uggg. So then I squirted the&nbsp;lovely being with water from the kitchen sink but that didn\'t do anything either. Then merri\'s mom left so we&nbsp;had to watch the assholes and they decided to be obnoxious and annoying. They took a stick from inside the house and tried to hit us with&nbsp;it so then i attempted to grab it from them and banged my hip into the wall ow ow ow hugeeee bruise.&nbsp;Then they followed us downstairs and started talking about their imaginary dragon named Leo that was going to eat us. Merri\'s&nbsp;mom also told them that they had to stay outside while she was gone so we were getting pretty pissed that they didn\'t listen. Then we locked them out of the house and went downstairs to play a game. They started ringing the doorbell 500000 times driving the dog insane and banging on the windows. So we\'re&nbsp;playing the game saying&nbsp;kay they will eventually stop and instead after a while we hear more pounding on windows and then a very loud crash. I screamed oh fuck and we looked at the window and it was smashed. Eric decided to geniously hit the window with a baseball bat to get our attention.&nbsp;Then he ran home. Joey was really upset and kept saying sorry a million times even though it wasn\'t all his fault and merri&nbsp;said she would tell their dad and take the blame for the whole thing. I told her she should tell her mom the truth though so then she did when she got home and her mom said she didn\'t want her taking the blame for it. So now everyone is hoping that the father doesn\'t see the window&nbsp;for about another week or joey won\'t get a birthday party and the block party on saturday will be cancelled. uggness what a productive day. Joey breaks things all the time with his friends and his parents are getting really pissed off about it. He needs to be more careful. His friends also need to not think about my boobs and call me a&nbsp;hooker haha. I normally love little kids but i seriously think after this experience if i have a son who is&nbsp;so bad that he doesn\'t know baseball bats&nbsp;smash windows and screams about&nbsp;vaginas to the neighborhood&nbsp;i will kill myself. &nbsp;   Simon your posts smell because they didn\'t post what you wrote! eeeep!&nbsp;I still don\'t know what a cat flap is. My Dad and Aunt sell stuff on ebay they are fucking obsessed. Go buy music shit and guitar&nbsp;bags from them people then you can&nbsp;get a package&nbsp;wrapped by me since i\'m their personal slave.&nbsp;  P.S. I just made the 69th post in this blog =D i so proud '),(7509,2753052,'28,July,2004','2004-07-28 00:00:00','    I\'m at Heather\'s, she\'s eating soup. She wanted me to post because the last three posts have been hers. We\'re going to see the notebook tonight. I fixed her webcam...well, sort of fixed, really I just turned it on. I also showed her how to burn a cd, but actually she didn\'t have any music to burn. So yeah, more fun than a barrel full of monkies here. '),(7510,2753052,'28,July,2004','2004-07-28 00:00:00','    My webcam works!  buh bye PS It would be thanks to Adi by the way. '),(7511,2753052,'28,July,2004','2004-07-28 00:00:00','    Whoops. Sorry guys. That last post was supposed to go in my personal blog. Guess the dye went to my head. hehe *blushes* Ok. Buh byes '),(7512,2753052,'28,July,2004','2004-07-28 00:00:00','    I\'m sitting in my pajamas and I look like shit. I\'m not sure what that has to do with anything but it\'s something to talk about.  I have a hair appointment in an hour. I should really put some makeup on before I go. I don\'t think scaring old innocent ladies is a thing that I\'m willing to do today. Tomorrow may be a good candidate for that though... Adi\'s coming over tonight.  OMG! I\'m going to see the Notebook again tonight! I love Ryan Gossling. Must touch... *streches out arm* Neyeh. No use. So as i was saying before I so rudely interuppted myself...Adi is coming over tonight. Hopefully she will fix my webcam. It\'s starting to drive me crazy. Damn thing. I just want to throw it against the freaking wall! Calm down, Heather. *breathes* Ok. So now I have to go do my makeup and get ready for perfecting my oh so hopeless locks (which i personally really do love) by dying them blond. Well some of them. It\'s only hightlights. Not the whole shabang. I\'m out. '),(7513,2753052,'28,July,2004','2004-07-28 00:00:00','    BUT HOW CAN A CAT HAVE A FLAP!!! *searches stuffed kitty* My kitty looks just like real kitty and she doesn\'t have a FLAP on her anywhere! How can a kitty have a flap its weirdness. Kitties cannot have flapssssss -_- oy the confusiness!     My head hurts   &nbsp; '),(7514,2753052,'28,July,2004','2004-07-28 00:00:00','    Catflap- n [cat-flap]: A flap, of, for or pertaining to a mammal of the feline persuasion.  Ok, now that we have that out of the way. Well, in the words of Forrest Gump, that\'s all I have to say about that. '),(7515,2753052,'27,July,2004','2004-07-27 00:00:00','    Ok. I\'m gonna totally agree with Laura here. What the hell is a catflap? Hmmm. *sits and ponders* Well, I can\'t figure it out. A little help please! '),(7516,2753052,'27,July,2004','2004-07-27 00:00:00','    I\'ve been taking notice of everyone\'s posts i have i have i have i swearrrrrr =D me me me ME! Heather peed outside, Lindsey is getting eggrollssss, and Adi\'s grandma is weird. yeyness i am all powerful.    &nbsp;  K now for more important matters.    &nbsp;  Dog...Egg...chunks... I get those...buttttt CATFLAP!? WTH IS THAT?...Biscuit...Phone...People...Ticktock...Mouse...Googly Eyed Aliens...Fishsticks...Crab cakes... WHAT THE FUCK IS A Marmaduke! And&nbsp;Tire Iron!? o.o good god i\'m clueless.     &nbsp;  K baiiii '),(7517,2753052,'27,July,2004','2004-07-27 00:00:00','    Good lord. And I thought being punished for a bit of a week was bad. =/ Yeah, that means I won\'t be post a whole lot for a bit. Eee, going to New Orleans this Saturday. ^____^ Yay for French Quarter and New Orleans Museum of Art! Dee dee dee...    &nbsp;   &nbsp;   &nbsp;  OH! Aye-chee....O___O Guess what. Le Denise is going on a lunch date. To where? Kim\'s! Guess what she\'s bringin\' back?    &nbsp;   &nbsp;  .....Guessed yet?    &nbsp;   &nbsp;   &nbsp;  ...............EGGROLLS!!! O.O;;; '),(7518,2753052,'27,July,2004','2004-07-27 00:00:00','    Well, I drove to the other fricking side of the state to go camping with the French Mob (aka: my family). I peed in the woods. Then I licked some grass. Yeah, that\'s right, I licked some grass. Got a problem with it? That\'s really all I can rememeber. Oh, I slept on a pier too. Exciting huh? Toodles '),(7519,2753052,'27,July,2004','2004-07-27 00:00:00','    Know what I hate? Fucking perverted guys who think they are all that. They piss me off so fucking badly. Today I went to the ice cream place to see if the damn manager was in and there were these stupid spanish construction workers there and i asked if the manager was in and they said no and then this really obnoxious guy was like, \"But I am\" and all of them started laughing like HAHAHAHA HOLY FUCK I WANTED TO KICK HIM IN THE&nbsp; BALLS! I was already in a bad mood because my younger sister is so mean and I was looking forward to shopping with her and my aunt tomorrow and then my sister decided she didn\'t want me to come. I didn\'t even do anything to her she just said no i don\'t want laura there. Thanks. I\'m so mad....  *breathes* '),(7520,2753052,'27,July,2004','2004-07-27 00:00:00','    *Crawls into bed with the monkey* I\'m sick and need to spread my germs around to everyone =D yeyness      &nbsp;  Your grandma sounds scary adi my grandma is scary too. Whenever I go over there she talks about how sexy my legs are and then gives me tonssss of food. If she offers me a snack and I say okay then she gives me half a cake practically and I eat five bites and done with it and then feel bad x.x and if I say I\'m not hungry then she gives me an entire cake almost!!! She insaneeee she also always calls my auntie and makes her&nbsp;be her slave and go to the store when it\'s way outta her way to get her stupid things. She smokes alot, can\'t drive, and can\'t breathe. I guess she pretty nice though she just weird. She really fat and&nbsp;a while ago I saw a piccie of her from when she was 20 and she was stick skinny and really really pretty so I said something like, \"Wow grandma used to be alot skinnier than she is now\"&nbsp;to my aunt and&nbsp;my grandma was in the room eeeeeeep. I so stupid.&nbsp;She just&nbsp;laughed though and said I was&nbsp;a \"cheeky wee monkey\" hehe old people\'s insults are funny.&nbsp;    &nbsp;  Today I went to price club and my aunt actually bought food that didn\'t have fat and grease as the main ingredient. Yay because if I have to eat fast food for dinner one more night in a row I&nbsp;am gonna puke. I&nbsp;very sick and feel all achy and tired.&nbsp;=( *pulls&nbsp;covers over head*    &nbsp;   baiiiii &nbsp; '),(7521,2753052,'26,July,2004','2004-07-26 00:00:00','    I\'ve been gone all afternoon spending quality time with my grandma. That would be great, except for she considers quality time to be shopping at save-a-lot, which tends to attract retarded, disabled, maimed, and otherwise unsightly people. To try and speed things up, I&nbsp;offered to go ahead and get certain things, but she was like \"No, there\'s specific&nbsp;kinda I need...and you know how your grandpa is..\" And THEN, in the meed section,&nbsp;she holds this..this...THING, in my face...and I didn\'t get a very&nbsp;good look at it, but I thought it was a penis...turned out it was a pigs foot > ....... ....... Our endeavors were unsuccessful.&nbsp;  So we went to ANOTHER hardware store, where she argued in a very loud way&nbsp;with a very hot hardware-store guy (I LOVE the guys that work in hardware stores..they\'re so manly...besides the old thick-glasses guy anyway) about the weed b gone...When she walked away I whispered \"Forgive her, she\'s old..\" He smiled, buh that was all. Have I mentioned I realleh dun look my best today either? I\'ve lost weight and don\'t have any clothes that fit..and my hair was wet..and it was humid out so the usually lovely locks went insane...so yeah, definitely no boyscouting today. '),(7522,2753052,'26,July,2004','2004-07-26 00:00:00','    Blah work...Blah emails...Blah Studieing...Blah posting...  -crawls into bed and snuggles under the sheets-  Monkeh Out '),(7523,2753052,'26,July,2004','2004-07-26 00:00:00','    Lindsey, the new face looks like a picasso painting ^^ And uhm..yeah...I Aaanyway, life is pretty boring...and a bit crap as well. I can\'t go to England now, even if I get a scholarship, because mum found out she stops getting the lovely checks from the government if I\'m not here -_- So yeah, tha\'s it. '),(7524,2753052,'26,July,2004','2004-07-26 00:00:00','    O_____O I got my finger stuck in it, but that\'s okay!! If you turn it over, the candy comes out!! ^____________^!!    &nbsp;   &nbsp;   &nbsp;  I\'ve been wanting that candy since Easter!!!&nbsp; -rolls a needle around in her mouth- Hmm...Okay, I\'m done. -teeters off- '),(7525,2753052,'26,July,2004','2004-07-26 00:00:00','    Uh....I brought blogger up a long time ago.....I know there was something I wanted to post.....    &nbsp;   &nbsp;   &nbsp;   &nbsp;   &nbsp;   &nbsp;   &nbsp;   &nbsp;  I can\'t remember!! ;-; Help. -clings to Adi, needishingly- (New word. -_\' New face, as well!! ) '),(7526,2753052,'26,July,2004','2004-07-26 00:00:00','    We all be one big happy blog family    &nbsp;   I went to Krispey Kreme in Driver\'s Ed it was muy fun. In the course of ummm 4 days I have had Dominoe\'s pizza, all american cheeseburger, quizno\'s, and now krispey kreme. My aunt needs to stop buying me take out for dinner&nbsp;or I\'m going to turn into a blimp o.o what a lazy woman she neva&nbsp;cooks&nbsp;I guess it\'s ok cuz i don\'t really eat alot and I don\'t look any fatter&nbsp;whoooo =D     &nbsp;   I\'m sick I keep coughing and my nose is running and my throat hurts I caught my sister\'s cold =( it sucks big balls I going shopping with my auntie today maybe she buy some stuff that doesn\'t have fat as the main ingredient yey     &nbsp;   baiiiiii '),(7527,2753052,'26,July,2004','2004-07-26 00:00:00','    Is it sad that Simon has Laura and Laura has Simon...buh...Lindsey has moi dnq and I&nbsp;hath Lindseh dnq? Cos I mean, tha\'s basically how it be... And also, Heather hath no one, I think that is sad *nods* Well, no one&nbsp;on ze blog anyways. Hmm... '),(7528,2753052,'26,July,2004','2004-07-26 00:00:00','    Good god....*_________* -rolls around in the feathery comfort that is, Bleeding Through \"On Wings Of Lead\"     &nbsp;   &nbsp;   &nbsp;  Damnit, there was a time when I despised that song...>____>;;; What\'s gotten into muh? -gives everyone a random hug-     &nbsp;   &nbsp;   &nbsp;  ;-; EH! '),(7529,2753052,'26,July,2004','2004-07-26 00:00:00','    My computer sucks and is gayyy =(     &nbsp;   I\'m trying to get a job at Coldstones...yey! Ice cream for meee. Actually...me making ice cream for other people...but it\'s still pretty cool. I downloaded eve six song adi it was cool. hehe     &nbsp;   I\'m watching Full House what a gay show. Whoever made this was mentally challenged.     &nbsp;   yay for lapdances!      &nbsp;  You know i go to therapy why you make fun it\'s not my fault i can\'t remember stuff =( tons people go therapy it doesn\'t mean&nbsp;crazy      unggg x.x      You getting yahoo lindsey i have yahoo too my sn is melanchiah =D      buh baisss     &nbsp; '),(7530,2753052,'25,July,2004','2004-07-25 00:00:00','    Geez, i was only gone for one night, not even that! Too many posts >  Yay for lap dances! Yay for teh Lindz getting yahoo! Yay for something else! Yay for adi comming to England! Ow i just bumped my head! -puff, wheez- Ok...Now onto me...Uhm, yeah great, nothing interesting goin on with me... '),(7531,2753052,'25,July,2004','2004-07-25 00:00:00','    One, Two, Three....who should I give a lap dance toooooooo     &nbsp;   *Gives Simon Lap Dance*     &nbsp;   Whoo whoo!!!!     No one allowed to open eyes!     P.S. My legs hurt from giving lap dances to chairs =D '),(7532,2753052,'25,July,2004','2004-07-25 00:00:00','    -whines. Struggles to see. Finally sees. O____O Puts Adi\'s hand back over eyes-    &nbsp;   &nbsp;   &nbsp;   &nbsp;   &nbsp;  ;-; Time to start scheduling some therapy classes.... '),(7533,2753052,'25,July,2004','2004-07-25 00:00:00','    O.O *covers poor Lindsey\'s lil chibi eyes* How wrong... '),(7534,2753052,'25,July,2004','2004-07-25 00:00:00','    hahaha this song is so cool and fucked up&nbsp;everyone download it    &nbsp;  Pussy Liquor~Rob Zombie =D     &nbsp;  ...makes me wanna give a LAP DANCE o.o *gives one to the chair*    &nbsp;  One, Two, Three who should I killlllllllllllllll =D '),(7535,2753052,'25,July,2004','2004-07-25 00:00:00','    Is a cool song. Even though it\'s instrumental and sixtiesish. You should go download it, except for Simon who\'ll prolly hear it when he visits me at my Scotland sheep farm. '),(7536,2753052,'25,July,2004','2004-07-25 00:00:00','    I\'m about to cry this sucks    &nbsp;  *hugs simon* here hug for you    &nbsp;  =( i sad     my font is yellow to blind everyone '),(7537,2753052,'25,July,2004','2004-07-25 00:00:00','    -schniffles. Looks to Simon and Laura. Schniffles some. Looks to Adi and her foreign exchange. Schniffles more.-    &nbsp;   &nbsp;  ...............I\'m all alone!!! ;-;     &nbsp;  Corseting by Atreyu    &nbsp;  Just swallow the pill and think of me no longer. Just let go and take yourself out before I kill you too. Aren\'t you tired of fucking me over and over? You were the last person I wanted in my sights. And my heart honestly breaks when I think of you. I understand now what I love you means- It\'s doing the right thing no matter the consequence. I\'m tying you up, using the nicest lace, trying to kill you softly. Trying to erase your face. I\'d redecorate the walls with your inner thoughts, but I\'m afraid it\'s the wrong shade of red. But I have these sadistic urges and I don\'t want to take it out on you. Right now you\'re the only one who understands my plight. Right now you\'re the only reason I can\'t sleep through the night.    &nbsp;  -sniffles- ;-; I need a guy or something. '),(7538,2753052,'25,July,2004','2004-07-25 00:00:00','    I like big blogs and I cannot lie...*dances around*  Moving swiftly along. Adi\'s gonna be in England next year prolly. For le whole year. Dunno what part though. Foreign exchange is a beautiful thing. '),(7539,2753052,'25,July,2004','2004-07-25 00:00:00','    -squeals. Rolls around- Wait. Wtf. This is orange. .....>_>!!!!!!    &nbsp;   &nbsp;   &nbsp;  Closest thing to pink that I\'ll get, I guess. ;-; So anyways...-kicks at the ground and shuffles about- Adi said she\'d be on, and I don\'t see her, and I think I\'m gonna cry now!!! ;-;    &nbsp; '),(7540,2753052,'25,July,2004','2004-07-25 00:00:00','    -hugs Lindzeh back- I do so love your abnormal good moods ^^  &nbsp; LOOK I\'M GREEN! O_O;    &nbsp;  Monkeh Out '),(7541,2753052,'25,July,2004','2004-07-25 00:00:00','    -throws self around- ^________^ Rave-ish music is teh shit.    &nbsp;   &nbsp;  Wheee....soo...let\'s see...-hugs Simon, being all nice and such- n.n; I\'m in an abnormally good mood today!! '),(7542,2753052,'25,July,2004','2004-07-25 00:00:00','    Yeah, so uhm, Robs all upset and annoyed and i have him all week, yay for me...Works going good, had yet more fun with Jo ^^ I hit her ass with a belt, now she\'s out to get me o_o; She electrocuted me with her finger? o_O; That was freaky and tingled...  Uhm, yeah, not much else to talk about really...Getting haircut Tuesday, i be looking smoooooth ^^  First few rubs are free, the rest we can work something out :D  Ok i\'m tired, work is tiring, so tired...tired tired tired...Yeah, you get the idea...My ankle hurts, think i twisted it or something? Hmm...Anyone wanna hug me? I want a hug =(  Monkeh Out '),(7543,2753052,'25,July,2004','2004-07-25 00:00:00','    I\'m so tired I overslept blehhhhh no one is answering my ims on yahoo o.o maybe it\'s broken? I hope not. =(     &nbsp;  This coffee tastes really bad my Dad doesn\'t know how to make it at all I want my Auntie =(     &nbsp;  *pouts* '),(7544,2753052,'07,August,2004','2004-08-07 00:00:00','    I haven\'t posted for a while, mainly cuz i think i\'m boring and i work too much and last night i got drunk ^^;  When the Primark Crew hit the town, you better watch out...  Me and Jeff almost got kicked outta teh gaybar ^^; That barwench had it in for me...Bitch >>;  Anywho, i had a great time ^^ I felt so bad this morning, bu now i fine again! =D  And you so did not miss me or my lap Lindzeh...Bu you still welcome to use it ^^  -gives latest addition to the Carrier....A SATTELITE DISH! Mini one of course ^^- '),(7545,2753052,'05,August,2004','2004-08-05 00:00:00','    Mmmm.....-gnaws on a frozen Zero bar- *___________*;; Heeeee. So, let\'s see. Oooh, Can\'t wait for the My Chemical Romance- \"Three Cheers for Sweet Revenge\" cd to come out!! -much squealing-         So yeah...Adi\'s gone...and I\'m left alone with fatty sugary food. '),(7546,2753052,'05,August,2004','2004-08-05 00:00:00','    Where is everyone? I all alone! =( '),(7547,2753052,'05,August,2004','2004-08-05 00:00:00','    Everyone should take a look at my crazy super drunk day:   urlLink www.picturetrail.com/melanchiah   Enjoyyyyy =D '),(7548,2753052,'05,August,2004','2004-08-05 00:00:00','    God I love southpark. Moving on, I\'M GOING SHOPPING! And the best part is? All my clothes are too BIG and not too small. Yay^^. Whoa..cheerful moment... Better go before it gets worse. '),(7549,2753052,'04,August,2004','2004-08-04 00:00:00','    ..How am I taking out the fun? ...Cuz I don\'t feel threatened by someone sitting on his lap? I don\'t need to prove anything I trust him and he respects me =)      I\'m so tired and people are slamming doors in this house and fighting bleh. This girl who i used to be best friends with in february but then we didn\'t talk for a while and now we\'re friends again, her boyfriend broke up with her and she\'s really upset. So this morning i talked to her and i was on my way to the beach so i said i\'d talk to her tonight and i promised to be online and everything and i said maybe i\'d call her if my aunt wasn\'t hoging phone and all this other stuff. So I was so fucking tired tonight and i wanted to go to bed so badly but i left aim on and forced myself to stay awake just so i could talk to her and i waited for her to get back from work and everything like i said i would just so we could talk so it gets to be like 10 and then she comes online and puts an away message on \"I don\'t want to talk to anyone.\" People so selfish sometimes. I was falling asleep on my bed waiting for her and I was so tired and all i wanted to do was sleep but i somehow stayed awake and then she goes and does that. It\'s so stupid. I\'m always giving my whole life to other people and then i don\'t get same back. I would die for anyone that i care about and i just feel alot of the time they wouldn\'t do same for me. oh well. the end.       Today was really fun but tiring we made a huge hole in the sand and buried merri\'s little brother joey in it and then shaped him like a mermaid. He is becoming a huge pervert when we were making his boobs up he was like \"BUILD THEM UP TO HEAVEN!!!!!!!!!!!!!\" lmao. I got annoying white jellyfish and sand all over my ass and legs it was great. We made joey\'s boobs so big that little kids were coming over to see what we were doing and their parents were turning them away. Joey was all over me and erin too hugging us and kissing our hands lmao. Then in the car joey decided to tell me who he liked and he whispered in my ear that he liked this girl emily and i said a little too loudly \"Emily what? Corbet?\" or something like that and merri and his mom heard so he got really pissed at me cuz he immature 10 yr old of course and he wouldn\'t talk to me and told me i wasn\'t his friend anymore lol. Then he started pinching me 20 times, he gave me \"sunburn\" on my arm which is like when you twist the person\'s skin it\'s so painful and i was screaming it hurt so bad, he threw this ball at my back, he said tonsss of things that i won\'t repeat, and he was being an obnoxious brat -_-. I said I was sorry like 5000000 times and felt really bad but he just kept doing more mean stuff. So then he told me that the only way i could make it up to him was to give a strip show on a soccer field in front of everyone and HE had to be there or date his friend eric who broke a window. I said i wasn\'t doing either. Then he said that to make it up to him i could scream at the men outside that i thought they were sexy so i did that and i screamed \"I THINK YOU\'RE SEXY\" at tons of guys from the car when we passed them. Then he decided that i skipped some men so he still wouldn\'t forgive me -_-. Later on his mom told me that joey probably made up that name and that she wasn\'t even a real girl and he was just doing all that stuff to make me feel really bad. Joey said she was a real girl and I hurt his feelings and all this other stuff ugggg. I bet she really is a pretend girl and he did all that stuff to me for no reason lol. So now he\'s mad at me but he\'ll probably forget all about it in a few days cuz he\'s a weird kid. I\'m 17 and shouldn\'t care about a dumb 10 yr old but I just feel bad because i don\'t like to make people mad or feel bad or anything even if it is a 10 yr old. whatever.       K now i\'m talking to that girl and trying to help her and i feel like a moron cuz i can\'t.      baiii people this was a long post    P.S. I wrote this last night and put it in the blog but the blog fucked it up so i had to copy it and put it in now... '),(7550,2753052,'04,August,2004','2004-08-04 00:00:00','    You\'ll all be thrilled to know that I\'m back! Ok, or not. But yeah, mhmm. Not much to talk about.  Me and Adi are going shopping in Detroit this weekend! Alright, nothing more to say. Toodles '),(7551,2753052,'04,August,2004','2004-08-04 00:00:00','    Anyway... So yeah, Molly Ringwald is hot, or rather used to be, and I\'m going to Detroit on Friday and will be back Sunday afternoon, with lots of clothes that (woot!) actually fit. And no, I haven\'t outgrown my old ones, they are too big, thank you very much. P.S. My boobs grew! P.P.S. Laura\'s taking out all the fun...and if chu get upset and say I\'m mean I\'ll beat you with a salmon! XD '),(7552,2753052,'03,August,2004','2004-08-03 00:00:00','    You may have his lap but I have his heart...  so I don\'t care...  This day has been really tiring i wish i stayed home. '),(7553,2753052,'03,August,2004','2004-08-03 00:00:00','    All who knew he should come and pay yer respects...On a lighter note (FOR THE FOURTH FUCKING TIME....)            -squeals and jumps on Simon\'s lap and claws furiously, marking her territory. Drags the Lindzeh carrier over- ;-; I meeshed chu!!!!! >_ '),(7554,2753052,'03,August,2004','2004-08-03 00:00:00','    What, pray tell, is wrong with black? Black is cool, black is slimming. So meh '),(7555,2753052,'03,August,2004','2004-08-03 00:00:00','    Hi everyone      Because i\'m so cool i posted a comment on every single post in every single archive in this blog. My coolness is amazing      Be back later tonighttttt after the beachhh =) '),(7556,2753052,'03,August,2004','2004-08-03 00:00:00','    Today I\'m going to the beach in my yellow bikini yay. I think it might rain though because the sky looks really gross x.x I hope it doesn\'t. Erin got stitches in her head because work door hit her in the head so we can\'t go swimming. =( poor erin. I\'m going with merri too.  Bai allllll '),(7557,2753052,'03,August,2004','2004-08-03 00:00:00','    .......>_____> SCREW YOU!!!         Jeesus christ, it only took me  5 minutes to fix the goddamn font...            o.o; Well. I\'m back. Maybe. >_>: For a little while. Sorta..            -rolls off- '),(7558,2753052,'02,August,2004','2004-08-02 00:00:00','    Like a certain friend of mine who shant be named, she spits it into her bfs mouth. ...... ...... ...... Lunch anyone? ^_______^   P.S. Could someone please explain the phrase \"the real McCoy\" to me? I think I get what it means but like, who  is  McCoy? Is there fake McCoy as well? It\'s confusing. '),(7559,2753052,'01,August,2004','2004-08-01 00:00:00','    Most girls don\'t suck they swallow. lol I\'m so friggin sick...and I have driving today. He\'s going to yell at me for coughing again. I have to try really really hard not to.       K bai peopleeeee '),(7560,2753052,'01,August,2004','2004-08-01 00:00:00','    So according to Adi the problem is i\'m shy yet Laura says thats not a problem? Hmm...  I think girls just suck -nods-  Maybe i should turn gay? I\'m much better at pulling guys ^^  Anyway, there, i\'m done...Be happy, be fruitful, be green...  Monkeh Out '),(7561,2753052,'01,August,2004','2004-08-01 00:00:00','    I had to delete one of Simon\'s posts, because for some reason it was screwing up the template, making all the links and archives and such go down to the bottom.  In case you\'re wondering, it was the catflap one. '),(7562,2753052,'10,August,2004','2004-08-10 00:00:00','    i got this really cool sweatshirt when i went shopping with Adi and I have it on right now. I just realized that it has thumbholes! I\'m so excited. How cool is that?!  Oh, and yes adi, you do look female. Congrats! lol '),(7563,2753052,'09,August,2004','2004-08-09 00:00:00','    Eventually you will all see the coolness that is my First Day of School Outfit. I look female! I look good! I look....fifteen! Granted, a very short fifteen...but still fifteen. '),(7564,2753052,'09,August,2004','2004-08-09 00:00:00','    I\'m not telling you people anything about going out, cuz i usually get moaned at for talking about what i did with the girls and rolling around on the floor with Jeff, so nyeh...      Working too much, luckily getting paid Firday, woots for me! ^^ Watched \"The Girl Next Door\" today, that was pretty cool, going to see I, Robot tomorrow, that looks cool ^^      So yeah, according to Viv, one of me supervisors, i\'m about as useful as a chocolate teapot ^^ Sad thing is i agree...lol, ah well, describes me pretty well i think ^^      Hey everybody, i suck! ^^      Monkeh Out '),(7565,2753052,'08,August,2004','2004-08-08 00:00:00','    Awww. Adi, so wrong! > Bad images in my mind.... I must leave before you further corrupt me! '),(7566,2753052,'08,August,2004','2004-08-08 00:00:00','    But, moving on... Lindsey, I don\'t think we should call it putting them to death (the guys who kiss drew barrymore) we should call it  putting them out of their misery .  Pick me up now, I need you so bad... For some reason, that line of Down makes me horny...no, I mean  horny...really horny.  Exscuse me for a few minutes....XD '),(7567,2753052,'08,August,2004','2004-08-08 00:00:00','    Stay little, demmit!! >_   -growls. Snarls- >___>; Damn font. Anyways. -squeals at satellite dish. Drags in a leather couch and a popcorn machine- Oh, a Spongebob marathon? ALRIGHT THEN! O___O         Not much happened this weekend....The freakin\' AC broke....in August....in Louisiana. That\'s like, death right there for anyone who\'s never been here. x_x Well...Ugh, Christina is watching Charlie\'s Angels 2 or whatever...Whoever kisses Drew Barrymore should be put to death.         Weeeell, Michael\'s on and James is ignoring me. ^____^ Yeah, life is good when people hate you.          \"Hate can be a positive emotion when it forces you to better yourself.\" - Atreyu, \"You Eclipsed By Me\" '),(7568,2753052,'08,August,2004','2004-08-08 00:00:00','    haha. You went to a gay bar. hehe. *snorts* Hmm. I\'ve never been to one. lol Well, toodle la '),(7569,2753052,'08,August,2004','2004-08-08 00:00:00','    Oooooh! Simon went to a gay bar? Get hit on by any hot guys? hahaha '),(7570,4128797,'14,August,2004','2004-08-14 00:00:00','   Alien vs. Predator seems like an idiot proof concept to me.   Take bad ass villain one and bad ass villain two, thrown in human cannon fodder, much chaos and action ensures.   A toddler could nearly write this movie.  Hell this concept is perfect for the typical movie writing formula of concept first, movie momentâ€™s seconds, finally the story.    Here concept is alien vs. predator, movie moments is much action occurs where two beat crap out of each other while the fodder dies, the story would be what ever takes you from action point to action point.  Real simple.  Idiot proof.  Sadly Alien vs. Predator (directed and written by Paul W.S. Anderson) is proof that nothing is idiot-proof.   Somehow, someway, the director/writer screwed up this concept.   The story is crap but who cares.  The human fodder crap, but again who cares.   One thing that sucks is 20 minutes was spent establishing the fodder.  Time wasted away from the action.   Here is the problem.  What action?  Where is the Alien vs. Predator?  If looking for that, you are not going to find a whole lot.  There are maybe two action sequences in this movie.   One is a bad ass CGI flash back; the other was a decent sequence that quickly turned stupid.     Maybe you are thinking that at least get to see the Aliens and the Predator do their own bad ass things.  Well, kind of, sometimes.  Mostly the Predators come across as ugly humans.  Another words they are incompetent.  The Aliens act mostly Alien like except they sure do grow up fast and make use of the acid way to much.  Also the Alien Queenâ€¦yeah suddenly here she is T-Rex size and acts just like one too.   Basically Anderson stripped away all the coolness of these characters.  They are pale shadows of the original creations.  The only reason I can think this occurred is the usual Hollywood arrogance.  In this case Anderson saw the movies as a kid like 15 years ago and declared himself expert.  Through that haze of memory, a lot of decisions at least track.  After all how bad ass can a Predator is is got beat up by Governor Schwarzenegger (thus incompetent) or that Alien Queen got killed by Weaver (thus the T-Rex Alien is born).  Itâ€™s all I got as explanation because nothing else makes sense.  About the only good thing in this movie are the CGI moments.  Thatâ€™s the only time this movie was good when it was obvious some CGI team did the sequence instead of that retarded director.  Also the Alien and Predator costumes etc are top notch.  Those teams deserve a pat on the back.  Basically this movie is Joel Schumacher level fuck up.   A franchise has been utterly demolished under this imbecileâ€™s hand (how these incompetent directors continue to get work baffles me).   I could go further but letâ€™s put it this way; the Predator only killed one Alien.  Thatâ€™s it.  The rest where just blown up.  That alone should indicate how little â€œvs.â€ occurred in this movie. '),(7571,4128797,'08,August,2004','2004-08-08 00:00:00','   Bush Protestors Turned Away From Event  Aug 5, 2004 6:23 am US/Central Mankato, Minn. (AP) Two young supporters of DFL congressional candidate Leigh Pomeroy were turned away from President Bush\'s quarry rally in Mankato yesterday.   Nick Burkhardt and Matt Klaber of Mankato initially were denied rally tickets after making unfavorable comments about the president while waiting in line for three hours. They later were given tickets, but when they got off the shuttle bus at the quarry they were told they couldn\'t go in.   They agreed to leave, but a Mankato West High School teacher who was accompanying them says he also was asked to leave when he tried to defend the boys.   Global geography teacher Jim Walz said he wanted to stay and was told by a Bush official that he would be arrested and escorted out if he made any attempt to protest during the rally.   Walz says if appearances by the President of the United States are treated as private political events, then cities are right to bill the campaign for the expenses of local law enforcement.   Source: http://wcco.com/siteSearch/local_story_218072505.html  ---------- Sigh. It saddens me how much freedom people are willing to give up for \"safety\". Even sadder that people that should know better don\'t have a problem with it. It started with protest only areas (for safety of course  ) now its spread to support agreements and risk of arrest for even COMMENTING negativily about the president. Sometimes slippery slope logic doesn\'t apply, but when it comes to rights, it does. As the Bush Administration has proven over and over and over and over again for the last 4 years.  Seriously both sides of the aisle should have serious problems with this.   Not even allowed to comment negativily. I think thats sick and any candidate or person, even Bush acolytes, that doesn\'t condemn it is un-American and a traitor. Screw Bush\'s and Coulters definition, theirs is limited to the Republican party. '),(7572,4128797,'06,August,2004','2004-08-06 00:00:00','   Funny site.  Sure many have seen it but still worth reading. http://www.adamchance.com/funny.htm  Never understood cybering.  Maybe between people that know each other and will meet and be able to live the little online fantasy, but strangers doing it who will never ever meet.  Just don\'t get it. '),(7573,4128797,'05,August,2004','2004-08-05 00:00:00','   Just got back from watching The Village.  Was bored silly.  Once again M. Night Schmalyan (schmalyn? bah can\'t spell it and too lazy to look) crafted a technically good movie that would be good for directing classes in college...but not a particularly interesting or even entertaining movie.  You have most of the ingredients, good cast, acting, plot, set and so forth.  Its just the actually story itself is boring.  Also the movie is just so damn sloooooooow.   You thought Signs developed slowly...this one is practically glacial in pace.    As for the twist...its ok.  Really not bad, more creative then what 99.9% of other Hollywood writers would have come up with...except you really just don\'t care.  The problem is you really don\'t care about the characters and even if you did the twist effects them but doesn\'t really involve them personally.  So you are \"oh okay /shrug\" instead of \"ooh ahhh\".    In the end though this movie suffers from the failure of room number one that any movie should aspire to - entertain the damn audience.  This movie was simply boring.  My rating for this movie: wait for cable...better yet wait till free on network TV couple of years from now. '),(7574,4128797,'03,August,2004','2004-08-03 00:00:00','   Hello, welcome to my erasend blog. First time done this stuff so could be a disaster but hopefully not. Couple of things. I tend to make alot of grammatical mistakes cause I frankly I am lazy. I just type what i am thinking that moment and post. Sadly that is probably a bad habit and am trying to break it with only so-so success.  The point of my blog is to spout about whatever I feel like. Can\'t guarentee it will mean a thing to anybody but hopefully those that read it will not feel like they wasted their time.  So anyway...welcome to erasend. '),(7575,4189257,'22,August,2004','2004-08-22 00:00:00','   I am the harm that you inflict  I am your brilliance and frustration Iâ€™m the nuclear bombs if they\'re to hit I am your immaturity and your indignance I am your misfits and your praised I am your doubt and your conviction I am your charity and your rape I am your grasping and expectations  I see you averting your glances I see you cheering on the war I see you ignoring your children And I love you still, and I love you still  I am your joy and your regret I am your fury and your elation I am your yearning and your sweat I am your faithless and your religion  I see you altering history I see you abusing the land I see you and your selective amnesia And I love you still, and I love you still  I am your tragedy and your fortune I am your crisis and delight I am your prophets and your profits I am your art I am your vice I am your death and your decision I am your passion and your plight I am your sickness and convalescence I am your weapons and your life  I see you holding your grudges I see you gunning them down I see you silencing your sisters And I love you still, and I love you still I see you lie to you country I see you forcing them out I see you blaming each other And I love you still, and I love you still  '),(7576,4189257,'12,August,2004','2004-08-12 00:00:00','   This is a very short story I wrote a couple years ago and just recently edited again.  Thought people might like to read it.  ***  So then what happens is: There\'s nothing. And I mean nothing. Tending bar Friday nights at Belluci\'s across the way from the college; trying to get the Honda started in the wind chill; watching the Leafs lose to Jersey on TSN.  I said to Bellucci, not that there was anything in it for me, but I said to Bellucci something like: \"Why don\'t you change the name of this broke-dick place and get some college kids in here to drink beer? Get some music in the box instead of this Kim Mitchell trash. Put a pinball machine or something in the corner. Show something pay-per-view for Chrisakes! They got thirty thousand kids across the street looking for a pitcher of Becks and a Radiohead record and you got this third string Anne Murray knockoff in the corner with a fingers-on-the-chalkboard voice.\"  So then, when Bellucci hears enough, he goes: \"Take a hike,\" which means get gone fast and don\'t come back. When you know Bellucci three minutes, you know how to read between the lines.  \"Pack it in your kazoo,\" is my reply as I unhook my poncho from the nub and hit the door. With Bellucci, it\'s safer to read insults from long distance, or refrain altogether. Big as he is, you can\'t tell where the fat leaves off and the muscles begin.  \"I\'m leavin\',\" I said as I stopped by the open door.  He turned. His face fell slack. His jaw dropped and his mouth sunk open in mock surprise. \"T\'row a party for yourself.\"  That was the last time I saw him - until later when I had to ask him for my job back.  --------------------------------------------------------------------------------  Weinrhoder got rich somehow, trading fat-backs and rat-tails on the commodity exchange. He\'d buy them when nobody wanted them and sell the rights when the scarcity set in. He sits around the bar a lot and sucks on limes that I float for him in a glass of gin. He says this is the only country in the world where you can get on a train and leave town without checking in with the authorities. \"It\'s still possible to become a millionaire here,\" he says, \"as long as you don\'t file any taxes.\" He bores the shit out of me really, except that I figure if a dumbass like him can make it, there\'s still some hope for me.  Zeidel is a weight lifter for Barbell City, down three blocks. He\'ll show you how to ride the stationary bicycle and how to row on the rowing machine. They pay him for this with a cardboard, computerized check from the home office, which I take from him on Friday nights and turn it into about twenty bottles of Lite beer from Miller, and then I give him the change. He told me how, although he is Jewish himself, he questions the Jew. \"Too much emphasis on the affairs of the mind,\" he says, \"to the neglect of the body. A healthy body and a healthy mind are one and the same. To neglect the body is to squander a valuable inheritance. Kill the body and you kill the head.\"  He told me the only thing he hates more than a phony intellectual is an underdeveloped pair of biceps or a man who didn\'t know where his next meal was coming from.  \"Damn right,\" I told him, pretending to be listening. \"I think you touched on something there.\" He tries to be friendly but on Friday nights, I started wearing long-sleeved shirts.  \"What some of these stick-necked intellectuals need is a good beating,\" he said to me one night. I\'m not an intellectual but when I\'m working behind the bar, and everybody gets really drunko, I could be mistaken for one, and I\'m truly on the cusp of being a stick-neck, so now I hang down the other end of the bar when he comes in, visiting only when he bangs his empties on the hardwood.  Craven owns the copy-shop across the street. \"I\'m sick of dealing with the public,\" he says. \"I\'m sick of all the mealy-mouthed mutts who want something for nothing. One day I\'ll set a match to the whole frieken\' thing. I don\'t care if the whole goddamned block burns down.\"  Loughlin drives one of those beverage trucks. One night when Rev. Jesse Jackson comes on TV, Loughlin takes his hi-ball glass and heaves it at the screen and cracks the goddamned thing. Everybody ducks like it was a bullet until they realize it\'s just Loughlin again. \"I hate that sonofabitch,\" he says. \"Don\'t worry, I\'ll pay for the damages.\"  Mr. Gwynne is a banker. Branch manager. Steps in, has a few. Brings in a girl friend from the bank now and then because everbody in here could care less. Told me his wife goes to Colorado in the winter and Miami in the summer because she loves the intensity of the seasons. I used to give him the odd short beer like on the house, when Bellucci was gone of course. We talked politics: Pine Ridge, El Salvador, third world debt relief, the usual BS. I\'d Soften him up, let him sound important. Then I bang on him for a loan. All I wanted was half decent car, not a loft in Yorkville.  \"You think I could get me a car loan?\" I ask him. \"I can get it anywhere, but I figure, you\'re a customer, I\'ll throw you the business.\"  \"Oh, no, no,\" he says. \"I don\'t handle that stuff, I don\'t do anything under half a million, and even then, only to corporations, business loans.\"  I told him, \"Loan me half a million, I\'ll give you back four hundred ninety-five right there at the table.\" I only need a used shit-box that starts up in the winter.  \"Wish I could, my boy,\" he says. \"It\'s just not my department.\"  Later, I see him downtown one day, squeezing his fat ass out of an twelve-year-old Chevvie and I realize he\'s broker than I am. From then on, he pays for his own drinks and I give him the cold shoulder. One Friday night when Zeidel is good and drunked-up, I\'ll whisper in his ear that that guy Gwynne down the end of the bar thinks you\'re a fag.  Stephen floats in on weekends to listen to the fingers-on-the-chalkboard voice. He\'s into interior design and I think he\'s got his eye on Freddie the piano player. Stephen wears those tight slacks and a Byron short, close at the waist, and billowy in the sleeve. One Friday he asks me if I want to go on a Sunday picnic with him in his restored MG. I told Bellucci: \"Get rid of these gayblades or pretty soon they\'ll be dancing naked on the bar. And that includes Freddie.\"  Bellucci says Freddie is an artist and he comes cheap. Stephen drinks Marguerites and pays cash. \"Besides, he\'s a fashion designer, that don\'t make him a queer.\"  \"Maybe,\" I tell him, \"he\'ll write this place up in Better Bars and Gardens.\"  Bellucci teaches some sort of real estate course two nights a week at some community centre up the road. He snuck in there through somebody he knows when nobody was looking. I guess it proves he does have that degree from Seneca like he says, but it forebodes bad times for his students. If the kids\' parents ever for a look at Bellucci\'s expertise in real estate, they\'d send their punks into the army for training, like in the old days.  I can only imagine this pile of mashed potatoes teaching somebody about floating mortgages, prime rates, equity-income ratios, etcetera, when he not only doesn\'t own a piece of real estate himself after forty-six years on the earth, but is even four months behind in his rent on the bar, and (he brags about this part) he hasn\'t filed a tax return since \'76. So much for higher education.  --------------------------------------------------------------------------------  So Bellucci takes off for Florida in February like every other bar-owner in the world. He goes down there to Lauderdale with his Bermuda shorts and thirty pairs of mid-calve socks, the shoes on his feet, three pairs of underwear, one blue Izod shirt, one green Izod shirt, one \"Bellucci\'s Bar and Lounge\" tee shirt and a green, plastic-peak half-hat with a band around the back.  He meets the other bar owners from New York, Boston, and Philadelphia and they bounce around the cocktail lounges with their shorts and their socks and their brown shoes. They throw around tips like a gang of Babe Ruths. The biggest tipper in the world was not John D. Rockefeller, he was a bar-owner with a load on. When he\'s up and around and sober, he won\'t give you the fuzz off a tennis ball, but put a shot of Jack Daniels in front of him with a beer chaser and he\'ll donate his liver and eyeballs to science. He\'ll leave a day\'s pay on the bar on the way to the next joint. Don\'t ask me why, that\'s just the way it is. It\'s tradition and the ginmill jockey who thinks he can do it differently is either new at the job or he will soon find himself a new occupation. It becomes clear to everybody that he doesn\'t belong.  Anyway, Bellucci does the wings of man and here I am, Undersecretary for Barroom Affairs, left in charge. Face the Nation. Shit-face the nation. The Honda\'s fuel pump is shot, and the exhaust system finally rotted its way into Japanese heaven. It\'s parked up at Yonge and Finch with a flat left rear. I think it has finally lost its will to live. It shows no interest and fails to recognize me. The headlights and grill are covered with frozen mud and somebody snapped off the aerial. One windshield wiper is in the up position, the other is gone altogether. Bleak city is the ditty. At this point, a smart guy realizes that bad as it is, it could be worse. You never know when there might be a new Terror Alert.  So, while Bellucci is gone, I devise a little merchandising ploy to put the place on the map. The method I\'m not prepared to reveal because I\'m truthfully thinking of taking this concept public and franchising it again and again around the world. International. Five days didn\'t go by from its inception to the point where there were lines of people literally out into the street trying to get in. Even the regulars, Weinrhoder, Zeidel, Craven, they can\'t get in anymore. It\'s too packed.  When Bellucci comes back from sun city, he can\'t believe his Italian eyeballs. He shoulders his way in, bull that he is, by bogarting his way past the people in front of the line. He muscles his way to the bar and ends up with his head between these two cougers who do Amaretto and are here almost constantly since about two days after my innovation, my master-stroke. Donald Trump would probably like to talk to me about this.  Bellucci, with his arms around these two old babes, leans up real close to the bar where he can get my attention. His eyes are wide and white, his nostrils flaring with excitement.  \"Kid,\" he hollers, \"this is incredible. There must be two hundred people crammed in here and another two hundred and fifty outside tryin\' to get in. It\'s a zoo. I love it!\"  \"Go home,\" I tell him, loudly. \"Give me one more week and I\'ll have this place floating. I haven\'t even started with my marketing. This place will be more famous than the Lido. 426-5050? Don\'t make me lose my chain of thought.\"  \"No, no,\" he hollers. \"You do it. You do it, kid.\"  \"I can\'t have nobody bothering me,\" I tell him. \"You go home. Come back in a week.\"  \"Right,\" he says. \"Right, I\'m goin\'.\" He looks square at the two women, one to the other. He\'s got a grin on his face like the cat who swallowed the canary.  \"Are you ladies enjoying yourselves?\" he shouts.  \"Oh, yes,\" says one. \"We adore this place.\"  \"Great, great,\" squeals Bellucci. \"I love it.\"  People are clamoring up and down the bar for drinks. Others are fighting their way from ten deep to get to the mahogany. \"This is the owner,\" I say to the ladies.  \"How nice,\" says one, the smaller one with glasses. \"We\'re here all the time, every chance we get. Your bartender is a wonderful young man.\"  \"The best,\" screams Bellucci. \"The best!\" He\'s smiling like a whale, his lower jaw open, his bottom teeth exposed, his tonsils dancing like Astaire and Rogers.  \"We\'ve been coming here for over a week,\" says the gray-haired, fatter one. \"And he hasn\'t charged us for a drink yet.\"  --------------------------------------------------------------------------------  So then what happens is: There\'s nothing. And I mean nothing. I said to Bellucci, not that there was anything in it for me, but I said to Bellucci something like: \"Get bent.\" This is after he fires me. I told him too, don\'t think I\'m going to do the Billy Martin thing where you hire me back and forth every time you get in a jam. Once I leave, that\'s it.\"  \"Take a royal hike,\" say Bellucci. \"I\'ll put my size 11 so far up your ass, the inside of your head will smell like leather.\"  That\'s the way he feels about it. Enough said. Some guys are afraid to spend a little money in the interest of promoting their business. Johnson\'s Floor Wax spends about twenty million a year getting their name in front of the public. Bellucci\'s brain curdles at the thought of a few shots that don\'t immediately translate into coin of the realm. That is why Johnson\'s are waxing the shit out of the western world and Bellucci is all out of roach spray.  --------------------------------------------------------------------------------  Three days into the off-season and Zeidel and Loughlin catch up to me. \"Let\'s go,\" they say. \"We\'re takin\' you out for a few pops.\"  \"Oh, what\'s the use,\" I say. \"Everything falls apart. That\'s the way it goes. It\'s entropy. It\'s Einstein. It\'s the quantum factor.\"  \"Your ass is the quantum factor,\" says Zeidel. \"All of a sudden you\'re a smart guy?\"  \"Hell no,\" says I, lying. \"It\'s just that everything I got is broke, including me.\"  \"Hey,\" says Loughlin, \"you\'re with us, you don\'t go to your pocket. There\'s people waiting for us.\"  Before you know it, we\'re across from the college and I\'m being dragged into Bellucci\'s. Craven is there in the corner. With Weinrhoder. And Gwynne. The big tub of whaleshit, the owner himself, is there, sitting at the same table, the wrong way on a chair with his eggplant forearms crossed over the top. He\'s lecturing them like he\'s John Kenneth Galbraith.  \"Here he is,\" says Weinrhoder.  \"Bring him over here,\" says Craven.  \"Sit him in the chair,\" goes Bellucci.  \"We thought we was a friend of yours,\" says Zeidel from behind, where he\'s got his claws dug into my trapezoids.  \"You are,\" I say.  \"You ran out on us,\" says Gwynne. \"I told you I can\'t give you a loan. It\'s the bank, it\'s not me.\"  \"I didn\'t run out on nobody,\" I say. \"I was three feet in front of a size 11. That\'s not running out. That\'s self preservation!\"  \"So what are you gonna do?\" asks Loughlin.  \"Live in my car. Lose a lot of weight. Get my goddamned MBA and get done with it. Don\'t worry about me. Two years and I\'m on Bay Street.\"  \"Bay Street, your ass,\" says Zeidel.  \"It\'s like this,\" says Craven. \"Mr. Bellucci and us, we got an understanding. You don\'t work here, we don\'t drink here. And that\'s not Woody Allen talking. Who\'s talking is five right-down-to-the-wire alcoholics. Mr. Bellucci here, he sees you in a new light. He wants to give you your job back. Like a manager\'s assistant while you go to school.\"  I look at Bellucci square-on. \"That right, fat-man?\"  He hesitates five seconds to get his Italian temper back in the box. His eyes open up wide and he starts to chew on his tongue. \"I\'d like to have you back, kid. It\'s not that these guys spend a few hundred a week in here, neither. You got good ideas; you just get carried away a little. Just ask me before you do something crazy next time. By the way, if you do get to Bay Street, try calling the boss a fat man, see if he\'s as nice as me.\"  \"You guys are all right,\" I say. \"It\'s nice to be wanted.\"  And it is. Yeah, it can hold you back from a lot of things you should be doing, but when you add up all the columns, what the hell else is there in this short, narrow little life that makes a lot more sense than that? What?  '),(7577,4189257,'11,August,2004','2004-08-11 00:00:00','   Let me establish two points of information before I tell this story. First you should know that for the summer of 2004 I have a gig as a glorified janitor at an wealthy suburban Church. You know the kind - with either mint green or fushia upholstery and accents. This gives me a unique persective on North American Evangelicalism because I get to see what rich Christians consider garbage.  Secondly, I am desperately in love with Sacramental Theology. Basically that means I like my Christianity soaked and dripping with liturgy, ritual, sybolism and tradition. Because of this, I have a very high view of the significance of the Eucharist (Communion as we usually call it to avoid sounding too Catholic). While I know the casual manner in which these acts are treated by many evangelical churches is for the most part harmless and well intentioned, we\'ve gotten out of control when people begin to detest communion because it means they have to stay an extra 15 minutes in the service on Sunday morning and that the roast might go dry or they\'ll miss their Tee-time. I do my best not to be critical (at least not mean spirited) and to derive as much meaning as I can from the way we do Communion, while supplementing with the occasional trip to placate myself when the need arises. Of course, sadly, I can\'t participate in the Catholic Eucharist, but just seeing other people do it is usually enough to satisfy me.  These things established, imagine my horror when I walked in to change the garbage bins in the Church on Monday morning after a Communion Sunday. Trays and trays full of unused stoned wheat thins - the crackers broken for the sins of the world - the body of Christ, given for you - just dumped in the trash. I\'ll admit I crossed myself and considered doing my duty in line with Canonical practice by eating the reserve, but there were just too many.  I tell this story with a bit of tongue in cheek. I know the Zwinglian theology behind it and that the Alliance Church comes by it innocently. I don\'t think Jesus will be too mad about it in the long run. But, shouldn\'t it give us pause to rethink what we are doing when we participate in the Lord\'s Supper? We don\'t all have to believe in transubstantiation, but shouldn\'t there be a middle ground somewhere? Surely, if anything, this one act should be given some special consideration and not just put on the same level as the extra bulletins we printed being run through the paper shredder. Christ gave us the Sacrament of his body and blood for a reason. It is special. It is a source of connection to the grace of God that we don\'t recieve by other means. If it isn\'t, why do we do it? Just because Jesus said we have to? I hope our love of Christ is bigger than that.  Just some good natured thoughts - not a Temple clearing.     Peace be with you.  '),(7578,4189257,'10,August,2004','2004-08-10 00:00:00','   While the event you are about to hear described took place several weeks ago now, rest assured it is definately blog worthy. It has most of the unbelievable elements of the opening scene of a cheesy porn movie, and yet took place in the real life of a mild mannered budding theologian on his way home from Church.  I was riding home on the subway after a night of work at Bayview Glen Alliance Church in North Toronto. As is frequently the case, I was making use of the 20 minutes of travel time by reading a good book. On this occasion it was the venerable German Jesuit Herr Karl Rahner, PhD DPhil. As my eyes scanned a section dealing with transcendental revelation, my ears were perked by a female voice in the seat next to me saying \"are you going out tonight?\". I looked up to see that this girl, rather attractive in her early 20s with a slight accent, was accompanied by two others of similar description. Thinking they must be speaking to someone else, I glanced back down at my page. Again the voice, but this time \"hey.. we\'re just trying to be nice.. people in Toronto are so rude\". She\'s talking to me? \"Oh um.. sorry.. I didn\'t realize you were talking to me.. no, i\'m going home.. I just finished work\". One of the companions, the one on the left, chimed in for the first time, \"where do you live?\". \"Downtown.. near the UofT\" I answered. This time it was the companion on the right\'s turn, \"do you go there?\". \"I do, yes\". Back to the first girl in the middle again, \"what do you study\". At this point I\'m beginning to feel like I\'m being interviewed for something, and I rarely perform well in interviews. Also, I often dread the question \"what do you study\" because when my answer is \"theology\" it usually means a very awkward few minutes of conversation ahead of me. This occasion was no different - in fact, it will no doubt go on record as the most memorable and uncomfortable of all. A few questions go back and forth and it turns out that these three young ladies are Catholic girls from Montreal. They \"have their beliefs\" but \"don\'t follow them that closely\". It becomes Obvious that the extent of their understanding of what it means to study theology consists of \"so you don\'t get to have sex\". I tell them I\'m not going to be a priest so I am allowed, but it really is too late - these girls have something to play with now and they decide to run with it. \"So, are you a virgin?\" they ask. I know, I\'m supposed to be ashamed of it but I answered them truthfully anyway, \"yes, I am\". The ladies are in full stride now, making a game of it - lets get the little virgin priest boy hot under the collar. As if it wasn\'t bad enough, their next question puts it over the top. \"Are you gay?\". My answer is a fairly confident \"no\", but they persist. \"Well, if you\'ve never f*cked a girl, how do you know you\'re not gay?\". I attempt to mumble out something about sexual orientation being defined by more than the act of intercourse, but I wasn\'t in top debating form at the time due to the fact that my jaw was still on the floor. Then, the coup de gras, \"we have condoms with us... and we\'re headed to a party at this hotel on Yonge and Bloor... you should come join us\". Fortunately, on this occasion at least, I have a pretty low self esteem. This allows me to assume that there was no way these girls are serious and that they are just doing this to mess with me and amuse themselves so I manage to turn down their proposition relatively cool and unshaken.  I spent the final two stops before my destination trying to pretend to read my theology text while the girls whispered that I probably thought they were the devil. I don\'t know who they were, but they\'re still \"on the loose\" - with all puns intended - so keep your wits about you boys.  '),(7579,4189257,'10,August,2004','2004-08-10 00:00:00','   On the advice of a couple good friends who have already embarked on the Blog enterprise, I have decided to jump into the fray. I remember the first time I heard of the idea of Blogging a year or so ago on some late afternoon \'get net savvy\' type TV show, my first thought was \"now all those pseudo-intellectual type guys who are wannabe writers will go around saying they\'re material has been published as a way to pick up girls\". Now while i\'ve never had much luck enticing the ladies with anything I do or say, I certainly am a pseudo-intellectual, so I guess I at least half fit the bill.  I study God for a living. (I know, I know, I can\'t believe they pay people to do that shit either but its true). Thus, many of my postings will probably be of an ecclesial nature: Either friendly ranting against my semi-FUNdamentalist Christian upbringing, or smitten remarks about post Vatican II Roman Catholicism. I used to fancy myself as one who had his ear to the ground on politics, art, music, and literature so there may be some of that too from time to time. And, I bleed the copper and blue of the Edmonton Oilers so I will rise and fall with them like the tides of my life (hockey and Christianity are intimately connected... ask me about it sometime). But, for my first entry, I thought it would be appropriate to recall a few things about the two aforementioned friends who inspired me to do this: Jeff C. M.D., and Sir Robbie J.  Jeff is my second longest serving friend. Our parents knew each other when the two of us were but toddlers. His Father is a gynocologist, his mother a benevolant dictator. While Jeff has 4 other siblings, I myself am an only child. They say only children develop fraternal bonds with one or two close friends. I would say Jeff is one of these. Although we\'ve been geographically quite far apart now for 7-8 years, and neither of us are very good at keeping in touch, we always pick right up where we left off. In the future I will call on him to perform my coronary bipass surgery or blast away my kidney stones. In return, I don\'t know, maybe I\'ll baptize his kids into the One Holy Catholic and Apostolic Church. It all evens out, and we will always remain friends.  Robbie I have known I believe since the end of the 7th grade. I remeber playing lunchtime pickup basketball against him on the outside courts at Hillcrest Jr. High (home of the Mustangs), and we were both the biggest guys on our teams. I don\'t think I ever told him this but it was Pastor Paul Silcock who suggested Robbie and I become friends because we wanted to get him to come out to the Beulah Youth Group (damn Christians and their prostelytizing!). In spite of this, we did end up getting along very well and would go on to spend many an afternoon in Robbie\'s basement playing Super Nintendo, messing with the mind of our computer teacher Mr. Campsall, interacting with our special ed. friends EJ, Roy, and Keith, and trying to avoid being yelled at to get our \"heads out of our asses\" by our Football coach Mr. Belmont. I\'ve always appreciated Robbie\'s adventerous side, his diversity of tastes, his ability to work hard, and his very evident concern for others.  So, thank you gents, and I\'ll see you in the days to come.  '),(7580,3384726,'26,May,2004','2004-05-26 00:00:00','   well so manisha\'s mum\'s the one with all the contacts. lets see if that works. and if it dosnt then what the heck? maybe losing a year wont seem that bad as it sounds.... i mean it is but then we\'ll all be in the same class.gah! its too complicated. but im gonna make sure that we\'re all gonna be in the same class. but thats if manisha\'s mum cant help out.  so i went to dhruv(ims) today. lets see how they can help. im kinda prepared to lose a year anyhow. lets see. me, manav, kartik anand, and i forgot the names of the others who flunked. damn i seem to be taking my failure a bit too casually i think. i mean when niddhi got an 82% in the 10th, she was crying like as though she\'d just been raped or sumthin... she was crying soooooo badly man. eventually she landed up doin commerce.   well so much for her. heck im the one whose supposed to be crying....!!! '),(7581,3384726,'24,May,2004','2004-05-24 00:00:00','   so i went to nationals today. i attended the Bsc.It orientation. i also found out that they do grant provisional admission too for the people giving the compartment. at least thats what the head of the B.Sc It department told me when i asked her. so maybe then i wont lose a year.  so i reached home and told my mother about it. and she went like \"you know it was destined that you would fail\" and i was like huh? what the heck are you saying? and what the heck anyway? and she went on about how she spoke to some astrologer when i was a kid and he told her that i would fail. well he didnt exactly mention anything about losing a year so lucky for me. i guess that provisional admission thing is gonna come in handy now.... heck astrologer where are you....? i need some help here!! '),(7582,3384726,'23,May,2004','2004-05-23 00:00:00','   well it would be stupid to start cursing anyone (or even God) when you know yr damned board results. ok so i flunked in math. so i shouldnt curse God. in fact i should be greatful coz i dont hav to repeat a year like some others who flunked in 2 subjects. maybe it was their fate? maybe it was my fate too? but their fate was just a little more twisted. anyway i hope that everything goes well for them. and for everyone else too. i got an 84% in english, 50% in chem, 53% in phy and 55% in bio, and a 25% in math. and thats pretty swell (all exepting the math) coz the way i studied for my boards (last minute).. it was all so..damn fast!!! like i always say, i was born intellegent but education runied me! screwed me rather '),(7583,3384726,'22,May,2004','2004-05-22 00:00:00','   Dear God, if yr there, please hear me out. this is me talking to you  well actually this post is meant only for God to read. so God if your there please hear me out. i just wanna say thankyou for all that you\'ve done for me. i dunno what all you\'ve done coz you\'ve probably done a lot anyway (ie if you really exist... or whoever you really are, whatever you really are) actually i wanted to ask you why you like being \"closed\" up. it kinda confuses us humans about who or what you really are. but then i suppose you like making things complicating and un-understandable to us. you know i had this friend who claimed he could see the dead and their spirits. he said he could communicate with them and gain a lot of knowledge from them. so one day i asked him about you. he told me that when he asked the spirits about you, they showed him a \"rock\" in his vision. he said that he never really understood why they showed him that rock. and he dosnt know if that rock was you. you know people pray to you in many forms. i dunno why but sometimes i have this feeling that yr actually human. just like us, but 9000 years ahead. probably just my fictionary imagination again. so then im back to square one. i dont really know who are or what you are or where yr from and what you do. '),(7584,3384726,'21,May,2004','2004-05-21 00:00:00','   well todays news is good. results arnt tomorrow, they\'re on the 23rd. heck but then im supposed to be scared right? somethings gone wrong with me i think. why the heck am i not scared? even manav isnt scared. heck poor bunny....., hes like petrified! anyway. i met dhwani and jyotika at the juhu ccd today. damn jyotika\'s like changed so much with that new hair-do. she got them cut DAMN short. and dhwani got rid of her spex... lenses and all. she looked wicked man... you know the \"sexy-wicked\" kind with the up-cut eye brows and all!  well then i met manav later in the evening. he\'d come from a movie. scary movie 3. which i was suppoed to have supposedly watched few days ago. lol! bet you theres no one who can fib better than me! '),(7585,3384726,'20,May,2004','2004-05-20 00:00:00','   Dear God, if yr there, please hear me out. this is me talking to you  well God, i dont know if your really there or not. i dont know if your real or just fiction. i dont know where you exist if you exist. but many people believ in you. i dont know why they believe in you. i dont know why i believe in you when i always have this feeling that you dont exist \"here\". its kind of complicating to explain by typing. ive never really felt that you were \"there\". but people tole me that if i believe in you, then you\'ll believe in me and make my every wish come true. ive never believed in miracles. probably because ive never seen or experienced one. but i dunno why i keep preaching about the same. ok there i go complicating things again. but yr \"God\", so you\'ll understand.... right? ive seen so many things since 7th. they\'ve been good and bad. but then again at that time i had only seen thigns. in 11th, you just made me feel those things. and experience was a lot. i learned a lot.  you know God, if your really there, i just wanna say thanks. ive learned a lot. i learned to love, to care, to share... and to do so many more things. most of all ive learned to live and not just merely exist. but then theres still so much more to go thru right? and deep down i knwo that ill live by and thru it. gnite !! '),(7586,3384726,'19,May,2004','2004-05-19 00:00:00','   well come to think of it ive come a long way.... everythings like changed sooooo much since 10th grade man. but then what the heck? its supposedly supposed to be for the best right? isnt that what everyone says anyway? well the old days were fun! lots of fun... well now im back from dubai (not to mention i dont wanna ever go back there again). was unpacking and all today and guess what i found.... my 12th grade exam answer sheets : 15/50 in english, 21/70 in physics, 9/70 in chem, 23/70 in bio. dint find the math paper (who cares bout it anyway) im gona be postin all these marksheets up... ive never been prouder of failing in english...! shit thats a bloody miracle. till today i really dont know how i managed that. i just didnt feel like writing the paper! '),(7587,3384726,'19,May,2004','2004-05-19 00:00:00','   and lets see what all did i do today. nothin. hadta sit home and lie to my pop that everyones doing their aieee paper today so that i can peacefully slip out of the house tomorrow. the aieee exam is actually tomorrow and my folks dunno... so i don havta really bother about them callin up my pals coz my pals are actually gonna be writing the aieee paper while i go and deliver the goodies...! oh man im gonna be free tomorrow! '),(7588,3384726,'16,May,2004','2004-05-16 00:00:00','   hey you know bunny and me decided that when we\'re 24 and if we dont have dates we\'ll marry each other... quick and funny thinkin eh? heck we\'re still 16... hes 17... il be 17 too in a few weeks.. heheheh  all the nonsense we talk about when we\'re teenagers. i think when im 24 im gonna laugh reading this blog...!! and hey yea, bunny\'s gonna laugh too(but not if we really get married)  (OMG WHAT AM I THINKING!!) hahaha '),(7589,3384726,'18,June,2004','2004-06-18 00:00:00','   hell why did God create rain anyway? rain sucks... it makes me bloody well wet and stinky and smelly(yuck) and spoils my hair and spoils my bag and spoils my pants and spoils my shoes(my first love) and most of all SPOILS MY WIND CHEATER! and eventually spoils my mood. being drenched is like probably one of the worst things that can happen to anyone. especially if someone of that \"anyone\" happens to like be me. yesterday was the third bloody time i got bloody well drenched for no bloody reason. hmph someone really does hate me.... waaaahh! im never getting outta my place till the bloody monsoons end...!! damn it... bah...! '),(7590,3384726,'16,June,2004','2004-06-16 00:00:00','   why the hell do you half the people comment anonymously? or however you spell the damned word? how the heck am i supposed to knw whose posting what and when? what the heck i hate all you people for this. ummm, err... well not all....! '),(7591,3384726,'12,June,2004','2004-06-12 00:00:00','   And i finally saw potter part 3 yesterday (globus). ooooooooooooooooou groovy as it was, the book was still better. again there was no \"bastard\" sitting behind me. shit man. mals gets all the luck. so thats how it got kinda boring too.  and afterwards we went into globus. and my friend wanted to try on this tee. so he went into the changing room and wore the tee and came back out (leaving his own tee inside the changing room) and was checking himself out in the mirror long enough for someone else to go into the same changing room. the dog who went into the changing room bloody well refused to come out for like a whole half hour man. WTF was that person doing in there? so my friend was stuck wearing the globus tee. and i was mad enough to ask him to try out the swimming trunks as well. shamless rather???. yeah ok im shamless all right. more than all the other girls too. but what the heck? i love being shamless anyhow. lol. yeah so i was fuming coz i wanted to go home and that dog or whatever in the changing room was refusing to get \"his\" ass outta there. ask me how i know that the person in the changing room was a \"he\"? well coz after i knocked (and later banged)the door for the idiot inside to get his bloody ass out and got no reply... i shamlessly bent down and looked under the door...!! i looked under to see if there was someone really inside or if the door was just plain stuck!!  and GUESS WHAT I SAW.....!!! there was someone inside. and he was trying on the same trunks i asked my friend to try on. now obviously a girl wont try on trunks right? '),(7592,3384726,'10,June,2004','2004-06-10 00:00:00','   gawd talk about moderniation. ok i dont know if the correct word is modernization anyhow. i remember once shea was complaining about the fact that girls like jatinder (yeah i know its a guys name)and all had these boyfriends and all in the states and stuff and how \"normal\" gals like us didnt. well that was kinda wierd. heck a boyfriend in the states of all places... WHAT THE HECK?  and to top it all on ym wa home today i see these 2 beggars on the road making out. oh heck whats the world coming too? and it just wasnt a pretty sight. hell so now i do believe the fact that these guys get inspired from movies. SHIT...!!! cow...!!! '),(7593,3384726,'09,June,2004','2004-06-09 00:00:00','   ok the doc\'s diagnosed me with this some shitty mental disease or illness rather called clematis or clemitis  or some shit. the symptoms are laziness, drowziness, unattentiveness, draggyness, dreamyness, living-in-the-future-ness, sloppyness, dope-i-ness, blah blah... this ness and that ness. gah! old fashioned doc. i mean which kid aint got this DEADLY DISEASE\'s (according to wht the doc calls it)symptoms. i mean the doc was like petrified and went on yakking like as though i had aids or something!! '),(7594,3384726,'08,June,2004','2004-06-08 00:00:00','   i kinda lost i (my head) yesterday for some reason. guess my normality kinda uplifted itself. ok... whatever that means! just didnt know where i was running too...!or running on? or running about? ok ok so someone* reading this knows im referring about that person. btw call me mean but it would have been better if you fainted when i was around(hehe) oh screw what am i thinking about anyway? damn ive become naughty*. so much for testing my edges (limits or whatever you call them). ive really lost it (my head) i think. anyway screw it...no wait dont screw it. no screw it. no dump it. wait whats with me anyway? maybe its the clematis or clemitis or whatever the doc calls it. but then i think maybe....&nbsp;  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'),(7595,3384726,'08,June,2004','2004-06-08 00:00:00','   woah..! talk about growing older when it dosnt seem so. everyone still says that i look like a lil kid. yea maybe thats why i still wanna go paddle boating. i dont care if you dont like it but i do... and many other people do too. so now basically all i gotta do is hope and pray that the club gets electrified back normally and the pool\'s gonna be clean. the shallow end of the pool ie. i cant afford to drown on my bday. or maybe not on my bday but the day i celebrate my bday... which probably isn\'t gonna be today anyhow. maybe i should\'ve taken those swimming classs in muscat seriously. and here nikhil thinks im thinking in air. well then ballz to him. hey... theres gonna be a carnival at the club too...! but thats gonna be in sept. talk about my luck..... hmph. anyway... time for me to screw outta my place...!! laterzzzzz! '),(7596,3384726,'07,June,2004','2004-06-07 00:00:00','   dats the movie i saw last night. and funny enough i didnt ask anyone to translate it for me. jeez sometimes i wonder how i watched it anyhow. unlike mals\'s case there was no \"bastard\" (read stupid mal\'s yet again dated on the 5th of june) sitting behing me. damn and i was getting bored. coz at least if there was a \"bastard\" sitting behind me i could have done some tp. probably asked for his #. but then damn my luck. lol!!! ive lost it again! anyhow and the movie went on with the usual \"the almost (not-at-all)desperate guy in his efforts to woo d babes\" guess its another pic thats inspired me apart from city of angels(suggested by that ****) and a few songs.... well not a few, a lot. and apparently one of the tum\'s has an accent now. reported by whizkid. gawd talk about people changing now. been wonderin whats been up with ryan and his \"tum\"............ '),(7597,3384726,'05,June,2004','2004-06-05 00:00:00','   i dont get these guys man. one day they say that they like you. and the next day they say that they love you. and yet again the next day they say that they dont have time to meet you. whose gonna understand them? they just zap us into that stupid trance of theirs and then woo us like this and then woo us like that... probably expecting that we\'ll fall in love with them more that we aldredy are(if we are... not unless we really are) and one day they\'ll finally decide that we suck (the BASTARDS). gawd who needs guys anyway? why did god create them anyhow? to confuse the insane ones like me even more than we aldredy are? i think i need prayers and here god\'s thinkin of screwing me emotionally? what the heck? bahhh...!!! wont suceed this time...!! gakkkkk!! '),(7598,3384726,'05,June,2004','2004-06-05 00:00:00','   ok so mal\'s saw the potter movie. and in her blog she wrote about this \"bastard\" (who apparently turned out to be some korny guy) who was sitting behind her and jabbering on his cell throught out thw whole movie. she \"bastard-ified\" him the title coz she claimed she couldn\'t watch the pic peacefully because of him. and when i asked her how the dude looked she claimed him to be sum wannabe chick. dont get my wrong but i call all the hot guys chicks. the gal mentality.  to all the guys reading this, if ya\'ll dont happen to approve to my new vocab then i suggest y\'ll consume the estrogen hormone tabs. yall probably need em to unscrew yr minds and think straight like a gal. hey! yal mite grow boobs with estrogen too! ita the female hormone remember? lol! i must me high...!  It\'s Like, I don\'t care about nothin man, roll another blunt, Yea (ohh ohh ohh), La da da da da da La, Da Daaa, La da da da, La da da da, La da da daaa I was gonna clean my room until I got high I gonna get up and find the broom but then I got high my room is still messed up and I know why (why man?) yea heyy, - cause I got high  La da da da da da da da da I was gonna go to class before I got high I coulda cheated and I coulda passed but I got high (La da da da da da da da da) I am taking it next semester and I know why, (why man?) yea heyy,  - cause I got high  La da da da da da da da da I was gonna go to work but then I got high I just got a new promotion but I got high now I\'m selling dope and I know why (why man?) yea heayy, - cause I got high La da da da da da da da da I wasn\'t gonna run from the cops but I was high I was gonna pull right over and stop but I was high La da da da da da da da da Now I am a paraplegic and i know why (why man?) yea heayy, - because I got high La da da da da da da da da I was gonna pay my car note until I got high I wasn\'t gonna gamble on the boat but then I got high now the tow truck is pulling away and I know why (why man?) yea heyy, - because I got high  La da da da da da da da da I\'m gonna stop singing this song because I\'m high I\'m singing this whole thing wrong because I\'m high and if I dont sell one copy I know why (why man?) yea heyy, - cause I\'m high  La da da da da da, La da da da, Shoop shooby doo wop. Get jiggy wit it, skibbidy bee bop diddy do wahhh - cause I\'m high  (hey where the cluck at cuz) '),(7599,3384726,'04,June,2004','2004-06-04 00:00:00','   i think i should have actually changed my title name. it should be rather something like \"BECCY THE BAD BOY SLAPPER\" coz i anyway cant slay guys. but i got an amazing record of slapping them. we can start of with mr k.surapaneni who i slapped twice in cirs and i forgot why. and i slapped dat ass oneek and that chut hursh. pretty neat huh? i got crazy guts i suppose. and im only 16 years old! lol! im so proud of me.  i wanna round off my slapping list to 5 by adding bunny to it too. he broke my sharky keychain. and hes so shameless enough to blame the maid for it. hmph. but then i cant slap him man.... hes not a bad guy. hes chooooooo chweet...!! that maid of his only must have broken it. lol! chowwie bunn.. '),(7600,3384726,'02,June,2004','2004-06-02 00:00:00','   met bunn today after gap\'s left. hadto ditch alok and vinay for a while when i went to meet him but they didnt seem to mind anyhow.  so we (bunn and me) spun around for a while and came to mac again. bunn wanted fries so i decided to play waitress for a while. luckily those bloody kids who grab food (and probably even yr boobs if they\'re different) wern\'t there to screw my mood up. (parcelled it).   im so proud of the indian beggars. they seem to have learned some manners finally. thank god! god really answered my prayers and changed those ill mannered things overnight '),(7601,3384726,'02,June,2004','2004-06-02 00:00:00','   well so much for the charity donations. screw i hate beggars man.. they totally suck! esp the kid beggars. all they know is how to piss me off. i dont care if they\'re poor or not. today i like came out of mac with fries and a coke... and those bloody kids just grabbed the stuff from my hands. those shamelass assholes. if they asked me nicely or even stood in front of me i would have given em the stuff man. but whats with the grabbing bit? fuck next time ill take a cop with me when i go. the pandu gandu guy can at least ram those screwballz away.  and then those chuck\'s..they\'re horrible too. we all got off at scruz station. and then they come and ram uss...!! bastardsss!!! '),(7602,3384726,'02,June,2004','2004-06-02 00:00:00','   well then the hsc results caome out today. ashish got a 67% anuj got a 71% bunny got a 68% chirag kinda screwed up totally in 3 subjects(38%) his babe divya got a 62% (i dont exacly remember) his cousin nikhil did pretty well though (forgot his marks) and nikhils babe malvika (ex-icse student) got a 65% and then i banged into neeta when she was going to get her marks. and then i havnt heard from bharti and sagar about thier marks. asif\'s also totally screwed up. at least he thought he did but i dont know his marks then. arshad told me amrita\'s marks but i kinda forgot them. and yeah! today is arshad\'s b\'day. its also ankit maroo\'s b\'day and what all did i do today? wake up at 8 and sit down to do math for 2hrs and then got ready and ran outta the house. went straight to bandra. collectd the bmm forms for rajni and then accompanied gurpanna and nikhil and alok to mitibai... then alok and gurpanna ran home and nikhil and me went back to bandra. nikhil went off to pi and i was waitin @ subway to meet chirag when he told me his disastourous news. and then we went to jatc where we met nikhil(his cousin) and malvika and divya. it was my 2nd time meeting both the girls. and then was supposed to meet bunny too but i decided that it would be \"healthy\" if i didnt. so i told him that i wasnt in bandra and that i was at the andheri mac. oh man talk about lying. thats like the first time ive lied to him. but i think its for the better in a way '),(7603,3384726,'02,June,2004','2004-06-02 00:00:00','   ok so this it what mals writes about me in her blog. i mean how childish can she get (i know yr reading this mals but yr childish anyhow)  Tuesday, June 01, 2004 stupid rebecca...  I was having a conversation with my friend (enemy???) rebecca (msn messenger) and thanks to her, I\'ve started to (temporarily) think and remember stuff....drain bramage!!...brain damage...whatever.....(I usually don\'t think so much)....she thinks I\'m going to do whatever she says ....perhaps she hasn\'t realised that I\'m not her paid slave or anyone\'s slave for that matter....I\'m not going to, in any way do \"B.Sc.it.\" or whatever she wants me to do (to give her company)....I\'ve made a decision to do either: 1) nothing at all 2) become a mass murderer  3) somehow morph into a vampire or werewolf 4) aid lucifer (bless the devil!)  5) destroy the world 6) if none of the above are possible (or if I\'m bored) I shall be an artist,author or both. These vocations are chosen on the basis of what I am capable of doing (which is..ummm...very insignificant)...I\'d prefer the first and the last one but since I believe in destiny, life shall take its own course. But whatever the case, I am most certainly going to watch the Harry Potter movie... and if possible, Van Helsing too.  P.S.: Rebecca, I still stand by what I had said earlier about  that sms.It really is very \"mommy-ish\"....   so thats whats mals wrote about me. i put my heart and soul into figuring out what kinda msg i was gonna msg *** about on his birthday.... it took me like 1 whole week to figure out what to send him and stupid mals thinks its \"mommy-ish\" mals yr an idiot! lol! '),(7604,3384726,'27,July,2004','2004-07-27 00:00:00','   My aren\'t you cool! You have no shame, or much to hide!&nbsp; You share your brave self with the web world! Your pretty fun to read! You blog baby blog!   urlLink   What kind of blogger am I?     brought to you by  urlLink '),(7605,3384726,'26,July,2004','2004-07-26 00:00:00','   been picked up from the BT page 6   for all geminites: Their undecided thinkers. They can intellectualise and fabricate great fantasies but generally lack the vigour to follow it up.   i am?  i can?  did i?  i do?  i dont?  i what?  i when?  i what again????? '),(7606,3384726,'24,July,2004','2004-07-24 00:00:00','   Is it a sin, to love again  I cannot win this pain\'s not fair  And yes I climbed the highest mountain  And yes I painted you a pretty sky  Now you say its over  Forgot to tell you I am sorry  Never had the time to lie  Now its like a sad old story  Why do lovers always cry  I never had the time to worry  Never had the time to try  Now you say its over  Is it a sin, to love again  You broke me in  This won\'t repair  And yes I fell for you so madly deeply  And yes I want to hold you by my side  Now you say its over  Forgot to tell...  And yes I want to tell you I am angry  And yes I need to look you in the eye  Now you say its over  Forgot to tell...&nbsp;'),(7607,3384726,'20,July,2004','2004-07-20 00:00:00','   As I look into your eyes I see the sunrise  The light behind your face helps me realise'),(7608,3384726,'12,July,2004','2004-07-12 00:00:00','   shameless enough and yet i did it once again. bunked my tuts and ran off to watch laws of attraction. and i also learned a few \"laws\" myself. i ill start yapping of by describin these two \"bastards\". well so i was on my usual boy-watching spree and there was this one feller, with a blue tee with yellow checks, and this other feller with a blue tee which had black sleeves. and it so happened that the feller(blue tee with yellow checks) happened to kick butt and the other sucked! so while i was busy yapping about the blue hunk to my pals (unfortunatly forgot to specify which blue \"hunk\", the yummy or the yucky)it so happened that the yucky happned to overhear me yapping about some blue \"hunk\" and the dog thought i was yapping aout him and well all bloody red in face and oooohhhh forget it. ill rest my case.   BOTTOM LINE: ALL BOYS ARE BASTARDS!!!! '),(7609,3384726,'11,July,2004','2004-07-11 00:00:00','   Don\'t be aroused by my confession unless you don\'t give a good goddamn about redemption. I know Christ is coming and so am I. You would too if the sexy devil caught your eye, he\'ll suck you dry...Soon you\'ll cry to be back in his bosom to do it again. He\'ll make you weep and moan and cry to be back in his bosom to do it again...   Pray? Shall I go blind? Pray? Cos nobody ever survives!  Praying to stay in your arms just until I can die a little longer  Saviors and saints, devils and demons alike. He\'ll eat you alive!   Jesus is risen it\'s no surprise even he would martyr his momma to ride to hell between those thighs... The pressure is building at the base of my spine. If I gotta sin to see him again then I\'m gonna lie and lie and lie. He\'ll make you cry! I\'ll sell my soul to be back in your bosom and gladly help. Please suck me dry and still you\'ll cry to be back in&nbsp;his bosom! Do it again!   My pulse has been rising, my temples are burning, the pressure is so overwhelming and building. So steady and heavy I\'m ready to blow. He waits, what is He waiting for? '),(7610,882528,'03,July,2004','2004-07-03 00:00:00','   Tonight has been slightly more productive than the last...  A few days ago I was infected with some nasty spyware that no program nor my own tinkering could remove so, being paranoid about spyware, I reformatted my computer. Unfortunately, I couldn\'t find my copy of Windows 2000 Professional, so I had to use Windows XP Home instead. I got sick of that pretty fast. Anyway, tonight I found my CD and reformatted once more, this time putting 2000 Professional back on, so things are back to normal.  Other than that, tonight I watched a movie called \"Waking the Dead\" on  urlLink The Space Channel \'s weekly Friday Frightmare. The movie is anything but frightening however, it had me in tears reaching for the Kleenex box. Definately worth watching. It got me thinking about my relationship with Amelia and I\'ve come to the decision to stay with her and attempt the dinner. If things don\'t improve afterwards then it\'ll have to end, but I have to try.  Also, tonight we had a thunderstorm, the first big one of the summer. It was great, I opened my windows and turned out my lights to watch it. I love the sound of the rain and watching the lightning, though it would have been much more enjoyable had I someone to cuddle with at the time.  It\'s pretty late now... I\'ve been staying up later and later each night, causing me to sleep in more and more. I\'m getting into my nocturnal habit again as I usually do in the summer. I\'ve also been neglecting MSN far too much the past couple days. I\'ve been ignoring Kat, pretty much the only person that ever wants to talk to me, and I feel bad about that.  I guess I\'ll wrap this up here then. '),(7611,882528,'02,July,2004','2004-07-02 00:00:00','   Well last night was a new low point...   Canada Day, the much celebrated day of the year and I spent it doing absolutely nothing. While everyone else was out with their friends, their significant others, I sat here alone. My girlfriend already had plans and it\'s not like any of my other few friends would ever think to invite me to hang out. Around 1:00 in the morning when a friend came online telling me about how great her night was, I had enough and went to bed, sleeping in until 2:00 PM. At least I got the house to myself, my mom was with her boyfriend the whole night.   I\'ve spent most of today just moping around the house, being generally unproductive. Watching TV, playing games, ignoring MSN. I havn\'t been feeling very social lately, I\'ve just been immersing myself in Freelancer and trying to forget about everything else. Not working as of yet.   Last night I was talking with Kat and came to the decision I was going to dump Amelia, but the more I think about it the more I have second thoughts. I\'m not sure why... Perhaps because then I\'ll be left with nothing, perhaps there\'s still hope in me it\'ll all work out. I havn\'t even talked to her in two days, I have no idea how she feels. '),(7612,882528,'01,July,2004','2004-07-01 00:00:00','   Another day, another post. Today will hopefully be more eventful. My mom is going to some event with her boyfriend and won\'t be back until tomorrow so as soon as she leaves I\'ll have the house to myself. A popular person would probably do something like throw a party, but of course that\'s not me.   Amelia and I talked yesterday for a bit... She suggested we end the relationship... I was kind of shocked. I mean if she supposedly loves me, why would she even bring something like that up? I don\'t know what to think anymore... This is just so frustrating. The same thing keeps happening over and over again in all my relationships, my girlfriend never has enough time for me. A day or two a week, is that too much to ask? I just want someone I can cuddle with... Someone I can care for without always missing them... But it always ends up the same.   I\'m thinking of cancelling the dinner, though the plans havn\'t even been made yet. At earliest it would be sometime next week... That\'s another week during which I can\'t see Mia. I keep telling myself that all I have to do is wait a little longer and it\'ll all be okay... But I did that exact same thing before summer started and things only got worse. How long am I supposed to keep waiting? '),(7613,882528,'30,June,2004','2004-06-30 00:00:00','   Well I talked to Mia, she can\'t go to the movies... No big surprise there. Soonest I can see her is next week sometime at the dinner. Tomorrow she\'s going to a Canada Day parade. She\'s working though, so I won\'t be able to see her. '),(7614,882528,'29,June,2004','2004-06-29 00:00:00','   Hey there. I actually missed posting yesterday, mainly because I never thought about it. Nothing exciting happened anyway though, I played Freelancer pretty much all day.   I still havn\'t gone to the movies with Mia, nor heard from her at all... I asked her on Monday if we could see a movie and she said not that night, but probably tomorrow or Wednesday she could. Well now it\'s Wednesday... No phone call, no message on MSN, nothing. But isn\'t this how it always is for me? The forgotten boyfriend, brushed aside for more important things. I know she can\'t really help it, but will things really change after I meet her parents? Things were supposed to change when summer came. I was supposed to be able to see her almost every day, all day, but that never happened.  Lately I\'ve stopped saying those three words, \"I love you\", to Mia. I realized I never meant them in the first place and only started saying them in response to when she said them to me so it wouldn\'t be awkward. I care about her, but how am I supposed to love someone I can never see? How can I believe she really loves me? I can\'t and am certainly not willing to make that kind of attachment yet the way things are. If things don\'t change after the dinner... Well, this is her final chance.   Anyway... Tomorrow is Canada Day I think... July 1st right? Well, Ally is visiting from Ft. Macmurray for about 15 days starting tomorrow. She\'s a good friend I\'ve known for a while and it\'ll be nice to finally meet her. We both want to have at least one date before she has to leave, maybe more depending on how things go. It won\'t be anything serious of course, I\'m still in a relationship with Mia, but we\'re close friends... it would be nice to get to cuddle...  More to come... whenever. '),(7615,882528,'29,June,2004','2004-06-29 00:00:00','   Another late night. There will be quite a few of those now that it\'s summer, not like I have any plans.   I saw the movie and it was good, though I already knew it would be having seen it before. There was a VERY annoying kid in the row of seats behind us who talked through the whole movie despite his mother\'s attempts at quieting him. At any point of even minor humour in the film he\'d laugh like  urlLink Elmer Fudd , it was so annoying. The kid sounded like he was 14 but he acted like a 6 year old, I wanted to stand up, turn around, and tell him off so badly. But of course, that\'s not the type of person I am.   When I got home I order pizza for supper. I was supposed to call Amelia but I was a bit nervous about it... I\'m terribly shy on the phone so I don\'t like using it and was also nervous that her mom would answer. She\'s always the one to call me. I tried her cell rather than her house phone but got no answer and gave up with that.  I\'ve been playing Freelancer for most of the rest of the night, and chatting with Kat of course. I talk to her so much more than anyone else, she\'s always there for me. Natalie came on briefly, another friend of mine in the USA, but barely long enough for us to exchange hello\'s. I\'ve known her for a couple years and we\'re close friends, though we\'ve kind of drifted apart lately. Kat\'s started to take her place, though Natalie still knows quite a bit more about me than anyone else.  That\'s about all for now I guess. Goodnight. '),(7616,882528,'28,June,2004','2004-06-28 00:00:00','   Back again, I\'m probably going to see a movie in a little bit. My mom wants to see  urlLink The Chronicles of Riddick  and invited me along so I\'m going. I\'ve seen it already but it\'s good so I don\'t mind a repeat.   I\'ve been talking to Kat most of the afternoon, now Amelia. I might be going to a movie with Mia tomorrow, she\'s not sure yet. I had to bug about it though...&nbsp;Makes me feel sort of like I\'m annoying her. I just miss her so much I want to see her whenever I can, but it\'s difficult...   Anyway, my mom is home now so I better go get ready to go to the theatre. '),(7617,882528,'27,June,2004','2004-06-27 00:00:00','   Not much happening today as predicted. I slept in until about 11 and as usual immediately went to the computer. I\'ve got chores today, lots of chores... Right now I\'m talking to Kat, a friend of mine in the US. We met a little while ago and tell each other pretty much everything.   Amelia came online last night before bed and we talked for a while, it was really nice. I suggested an interesting idea to her and she accepted. Being a gamer, naturally I\'m doing that 90% of the time so I figured it would be fun to have her join in. I offered Freelancer to her and she\'s going to try it out. It\'ll be a great way for us to bond more, though perhaps create a little stress from friendly competition.  Back to MSN... '),(7618,882528,'27,June,2004','2004-06-27 00:00:00','   Well it\'s nearing the end of the day and I figured I\'d do a final post. I\'m going to try and keep this up to date rather than neglecting it like I\'ve done with all my other web logs.   I played some Freelancer, watched some TV, the same old stuff. Amelia called around 2\'ish and we talked for a while, it was nice hearing her voice again, I really miss her. She was supposed to come online after cleaning her room, but after waiting for 7 hours I don\'t think that\'ll happen. Tomorrow will likely be much of the same. I\'m looking forward to a time when I can see Amelia more often, hopefully it\'ll be possible after the dinner. We might be going to see a movie this week if she\'s allowed to go.  That\'s all, back to the TV and then to bed. '),(7619,882528,'27,June,2004','2004-06-27 00:00:00','   *Yawn* Awake once more. My mom\'s home now, she had gone to church. She\'s doing it because her current boyfriend is religious and what not. Today\'s plans most likely include chatting on MSN and playing  urlLink Freelancer . Possibly time for a meal somewhere in there. Now that it\'s summer I don\'t have much reason to leave the house anymore so things get pretty boring. I\'m contemplating taking up coding again.  There\'s no set date yet, but I\'ll probably be having dinner with Amelia\'s parents sometime in the near future. Amelia and I are sort of in a relationship... We were but I broke up with her because it looked like I wouldn\'t get to see her all summer. It was a rushed decision and I regret it. Her parents aren\'t too enthusiastic about her dating, moreso if they knew I\'m over 3 years older than her, so that poses scheduling problems... Her mom suggested that if she invited me to dinner I might get her approval, and thus she\'d be able to date. The problem is I have to appear at most 15, but I\'m up to it.  Sometime this week, or perhaps next depending on when I feel like it, I\'ll be going to see  urlLink Spiderman 2 . I liked the first one, mainly for the love vs. duty theme, and the second looks much better. After that I\'ll probably see  urlLink I, Robot , It looks pretty cool.  Time for some gaming. '),(7620,882528,'27,June,2004','2004-06-27 00:00:00','   Hmm, not a lot to put down now, just breaking it in. I\'m probably going back to bed for a bit... So happy school\'s out, though I\'d be home anyway since it\'s Sunday. I\'ve been meaning to start another one of these for a while and finally decided to. More to come later. '),(7621,882528,'09,July,2004','2004-07-09 00:00:00','   Today is my brother\'s birthday, and I\'m pleased to see it\'s not any more eventful than mine was. No cake with candles, no gift giving session, no fancy dinner. Just like mine. Oh wait, actually I forgot the part where he\'s out right now partying and getting drunk with all his friends... I guess that\'s one thing I never got to do, lacking any friends that gave a damn.  Moving right along... Tomorrow is the dreaded BBQ. I already made it quite clear to my mom that I don\'t want to go and that I\'ll be miserable and antisocial. She\'s worried it might rain, well I\'ll have my own little storm cloud with me in any case. I hate how she always makes me go to these things, I don\'t know nor do I  want  to know her friends.  Today was pretty boring. Slept in until about 1 I think, had some pretty nice dreams though I forget most of them. I spent the afternoon on my computer as usual and I talked with Kat for a couple hours. Tonight got a bit more eventful. My brother bought himself an X Box so I got to try it out. I\'m not a big fan of consoles though, I prefer the computer.  One of my ex\'s decided tonight that after a month or so of not talking to me she\'d come online and start telling me about her new boyfriend, after assuring herself that my love life is in shambles. She must get kicks out of it or something, whilst making me feel even worse than I already do.   Still not sure what to do about Amelia... I keep changing my mind. I want it to work, but in the meantime I\'m being an awful boyfriend. I\'m afraid she\'s just going to dump me, but I feel like I deserve it anyway. I\'m trying to make plans again with Ally... I know it\'s wrong, I know exactly how things will go if I see her again and it\'s not fair to Amelia. I just hate being alone all the time... '),(7622,882528,'08,July,2004','2004-07-08 00:00:00','   Yesterday was quite eventuful... and confusing.   Ally and I made plans the day before to hang out. I got up early around 7 AM and went to meet her at noon. We came back here and things got pretty cozy... Lots of cuddling and kissing... and then at 6 we went to see Spiderman 2. Afterwards my brother drove her home, her and I cuddling in the back seat, and I got home around 11. This sounds all good but I\'m still kind of in a relationship with Amelia... Ally will only be in town until the 21st so there won\'t be anything long term but it\'s still&nbsp;cheating...  Amelia and I talked a bit after I got home and she told me the dinner has been put off for another week, so once again I have to wait. I told her about how I was losing my feelings for her, having been apart so long, and she was surprised at that. She told me she still had feelings for me. I told her about the date because I felt guilty, though I didn\'t go into details... I\'m not sure how she feels about that.   I might be seeing Ally again on Sunday... This isn\'t fair to Amelia, I\'m not sure what to do... I want a relationship with Amelia but she keeps me waiting and in the mean time I want someone I can be with... '),(7623,882528,'06,July,2004','2004-07-06 00:00:00','   Been a couple days, I know, I\'ve been slacking.  Empire Earth has been keeping me busy the past little while, though I\'m pretty much done with it now. After a few 6 hour games it\'s lost its appeal pretty quickly. Probably for the best, I\'ve still been ignoring MSN more than I should be. Sorry Kat!  I talked to Amelia about 2 days ago and things seem pretty much back to normal. We solidified our plans for this week, though it\'s still undecided whether it will be on Thursday or Friday. After coming out of my slight bout of depression I realized I had almost let go of the one thing that still makes me happy, my relationship with Amelia, and I\'m now determined more than ever to hold on to it.  Tomorrow, or to be more accurate since it\'s past midnight, today, my brother is flying in from Toronto. He\'ll be staying a week for his birthday which is on Friday. I havn\'t gotten him anything, I\'ll probably just give him cash for booze, same as last year. He\'s always in need of that.  Ally\'s in town now, she\'ll be here practically all month. We might go see a movie together this weekend, I might finally get to see Spiderman 2. Money will be kind of tight this week. I told Ally it would be my treat, so there goes $20 for her and myself at the movie. I\'ll probably give my brother $50, and I want to buy Amelia a little something for when I visit her place so there\'s another $20. Considering I\'m living off a small allowance, that\'s a lot for me to spend. Oh well, my mom owes me $100 come this Friday...  I should really be going to bed now, I have to get up in 4 hours to get ready for when we go to pick up Chris from the airport. We\'ll be taking a white Hummer my mom rented for the day, a present for Chris though she\'s still undecided as to whether or not she\'ll let him drive it.   So yeah... Bed. Soon anyway, I\'m still a bit restless. More things need doing first. '),(7624,882528,'17,July,2004','2004-07-17 00:00:00','   Saturday... Just another boring day. Not much happened yesterday, just socialized online a bit and played some games. I slept in until about 2 PM, not intentionally, so of course I was up all night wide awake.  &nbsp;  I got the chance to talk to Karli a few times... She\'s so amazing. I\'m starting to feel like she\'s a bit out of my league though... Me, the reclusive, antisocial computer geek, going on a date with an intelligient, beautiful, popular girl... Oh, and did I mention she\'s a millionaire? Her father is anyway, he owns some video company. I don\'t even have a job, it\'s a bit intimidating... Suddenly treating her to a movie isn\'t that special anymore. I have no idea how this\'ll work out, she seems like a really active, social person, which is the opposite of me. I\'ll give it a shot though, I really like her and just can\'t pass up this oppertunity. Anyway, she\'s probably on her way back to Edmonton now, she gets home today.  &nbsp;  No set date yet as to when we\'ll meet yet. I think I want to wait a bit longer, maybe a week. I just don\'t want it to turn out like my date with Bree a while ago... I met her, we liked each other, and within a few days we went to the movies. It ended so badly, I was really shy and barely said anything the whole time because I was too nervous around her. Not like she tried to help at all though... She brought a friend who she talked to constantly, ignoring me anyway. I hope she doesn\'t mind waiting a bit longer. I really like her, I just don\'t want to screw this up...  &nbsp;  Today promises to be as boring as every other day, moreso since Karli will probably be hanging out with her friends tonight so I won\'t get to talk to her. The land lords are coming by today, anytime now actually, to inspect the house. Make sure we havn\'t trashed the place, aren\'t growing marijuana plants or housing illegal immigrants, that kind of thing. So of course my mom had me cleaning the entire place in preperation...  &nbsp;  Aside from that, there\'s not much to say. I\'m kind of sick today, I threw up a little while ago. Not pleasant at all. I\'ll probably take a nap or something once the land lords come and go. I guess I brought it upon myself though, drinking that awful Kahlua last night. It was about 4 AM and I was restless in bed, so I decided to pour myself a glass of my mom\'s booze. That stuff is so horrible, I don\'t know why she likes it. It got the job done though; once it started kicking in I had no problem dozing off. I didn\'t even wake up about 3 hours later when my mom put Maxx on my bed (the puppy). So yeah, not something I\'ll be doing again soon. I hate alcohol, but sometimes it helps with things... I\'ve never actually been drunk, but I\'ve been tipsy twice now and it\'s not that impressive. '),(7625,882528,'15,July,2004','2004-07-15 00:00:00','   I\'m pretty sure Amelia just dumped me... and I deserve it of course. She asked about my date with Ally, I had told her of it once before just not the details. She wasn\'t too happy with what I had to tell her... So she yelled at me, blocked me, and now I feel miserable. I\'m going to keep that date with Karli and see where that goes. I just want somebody I can be with damnit... '),(7626,882528,'13,July,2004','2004-07-13 00:00:00','   Not much news except the thing with Karli seems to be a bust. I was checking out the comments on her profile and apparently she has another guy waiting for her to get back so they can get together... They were both flirting back and forth a lot, suggestions of cuddling and more... I guess I shouldn\'t be surprised, she\'s 14 afterall. I\'m going to call off the date. '),(7627,882528,'13,July,2004','2004-07-13 00:00:00','   The spaces between posts are growing... I keep forgetting about this thing. Anyway, a bit has happened since the last one...   I met a new girl only a couple hours after that entry, her name is Karli. She made a post on the Nexus forums looking for a certain type of guy and I replied. She sent me a comment and we\'ve been talking since. She seems to be interested in me, but I\'m being cautiously optimistic... She wouldn\'t be the first 14 year old to break my heart. She seems okay though and I\'m starting to really like her. She\'s visiting relatives in Seattle until Saturday, so we might meet when she gets back. Don\'t get me wrong, there\'s nothing I\'d like more right now than to have someone I could cuddle with, especially someone as cute as she, just it\'s never that simple... I\'m after a long term relationship, not a date or two, so I\'m going to get to know her better first.  Aside from that there\'s not much that\'s noteworthy... the BBQ was long and boring, and so was the rest of my weekend. My brother is returning to Toronto tomorrow evening, thankfully. Dinner plans with Amelia are still up in the air, she can\'t decide when she\'ll be available. I\'ll give her another week...   Ally had to go back to Ft. Macmurray early, she sprained her wrist and her father wanted her back because of that or something. Kind of cruel how the best date I\'ve ever had was with her and now fate has made it so there will be no second.  That\'s all... '),(7628,882528,'22,July,2004','2004-07-22 00:00:00','   It\'s thursday evening now and I\'m depressed to say that nothing new has happened since my last post almost a week ago. Eat, sleep, chat... That\'s been the usual lately.  I talked with Karli a few times more, though less in the last couple days. I think perhaps she\'s losing interest in me. Probably for the best if she is, I\'m a horrible boyfriend anyway. Right now she\'s off on another trip.  I also talked with Amelia once or twice since the break up... She\'s still mad at me, as I would expect her to be. I told Ally today that I had been already dating someone during the time we spent together. She says she\'s not mad, and claims&nbsp;she still wants to see me again but she\'s just busy.  So here I am, bored and alone as usual. It seems I\'m pretty much left with nothing now. I craved to be with someone so much that&nbsp;I wrecked my relationship with Amelia for a single day with Ally, and now I probably won\'t see either again. I guess it doesn\'t matter, I wasn\'t happy with Amelia anyway... I had much more fun that day with Ally than I had ever had with Amelia, and in that one day I spent more time with Ally than my combined time with Amelia. That\'s how a real relationship should be... Too bad it only lasted one day... '),(7629,882528,'27,July,2004','2004-07-27 00:00:00','   Best date ever. Goodnight. '),(7630,882528,'25,July,2004','2004-07-25 00:00:00','   Amelia and I are back together! For once, I am happy. I thought about it and figured out what I wanted, and that\'s her. I asked her if we could try again... and she said yes, if I\'m willing to try hard enough to make this work. So tomorrow her and I are going to K-Days together! I\'m so happy... I get to hold her again. One thing I really want to do tomorrow is go on the ferris wheel with her when it gets nice and dark out so we can cuddle... *sigh* '),(7631,882528,'25,July,2004','2004-07-25 00:00:00','   This weekend has been good I guess, in that the entire time I\'ve had the house to myself. Gets pretty lonely though.  I\'ve been thinking constantly about Amelia lately...&nbsp;I\'m not sure what I want anymore... I thought I wanted Amelia back, and so I tried that in a way... I told her I missed her and wished we were still together... She invited me to come along with her to K-Days tomorrow, as a friend... But then I got scared and told her nevermind. I could go, I could see her again... I want to do more than see her though. To be there as her friend, that close to her and not able to hold her hand, or kiss her lips, or put my arms around her, would kill me inside. Even if I could, even if we got back together, it would probably end up the same. I\'m not even sure if I really have feelings for her or if I\'m just desperate. I\'m just afraid of getting hurt again, or hurting her more than I already have.  Karli has kind of left the picture. Since she left Edmonton again a few days ago I havn\'t heard from her. She\'s probably forgotten about me already. Ally won\'t see me again, but I don\'t care, I don\'t want to see her anymore.  So... I\'m living the single life again. No girlfriend, nobody interested in me. Seems pretty bleak. Bree came on MSN yesterday looking for anyone to make out with and I actually asked her about it. Then&nbsp;I told her nevermind. Then I asked again. Yes, I\'m definately desperate.&nbsp;I just want somebody to hold... '),(7632,882528,'06,August,2004','2004-08-06 00:00:00','   Well it\'s final, there\'s no chance of ever getting back Amelia.  She sent me an email today in reply to the one I sent her a few days ago. She said I wasn\'t being honest with her by hiding that I was unhappy with our relationship and that I had led her on by pretending not to be. Well I\'m a little confused here... because in my mind, someone who sticks with you when things are rough is someone who truely cares about you. Should I have cut and run at the first sign of trouble? I really cared about her so I put up with it. My mistake. But on the other hand, I couldn\'t put up with it enough not to cheat, so I guess I\'m not that person. I guess this is just karma catching up with me.  The night\'s been pretty good aside from that actually. It\'s kind of nice to have it finalized like that, I can stop worrying. She made it clear she never wants to talk to me again. After our exchange of emails, I decided to contact an old friend, Stacey. I don\'t know why... I added her, then deleted her, then added her once more before getting the courage to message her. I guess I just needed to talk to a friend, and nobody else was on. Well I\'m glad I messaged her, we had a nice conversation. Stacey is the second girl I\'ve ever dated, we went out for a month before seperating. It was my decision, and now that I look back on it, it seems pretty stupid, but things are back to normal now and it\'s nice to have her back as a friend.  The only other thing of note is that I helped interpret for two people on an IRC channel. I can\'t take all the credit though, I had the help of  urlLink Babel Fish . Now I\'m talking to Natalie, apparently she had a great night too.  So off I go... '),(7633,882528,'05,August,2004','2004-08-05 00:00:00','   Today started out rough but got better.  The morning sucked, simple as that. After I pulled myself together I tried giving Amelia a call, unfortunately her cell was off. Then I remembered that last week she volunteered at the farm for a few days and got home Thursday night, so maybe she did the same thing this week... Which makes me wonder if she got any of my messages at all. If I can get the courage again, I\'ll try and give her a call tomorrow. What\'s the worst that can happen, she refuses to talk to me? It wouldn\'t be much of a change.  Besides that my evening was dull, but kind of nice. I just relaxed and watched TV, cuddled with my puppy Maxx for a while, and surfed  urlLink Nexopia . Last night I posted two new pictures to my profile ( urlLink 1 ,  urlLink 2 ) so I\'ve been eager to see what kind of response they\'ll get. Nothing big yet, but Bree decided to talk to me today after weeks of silence. Hmm, I wonder why? Heheh. I\'m not interested, but it\'s nice to know the pictures have that effect.  Hmm, what else... Oh yeah, I trimmed my contact list yesterday. I decided I don\'t like knowing lots of people, so I removed all but my closest friends. I keep a pretty tight social circle... I went from 30+ to 5. Kat, Natalie, Joey, Patrick, and Amelia. Kat and Natalie are my closest friends, they live in the US. Joey lives in Floriday, we used to work on a game project together, and Patrick is a buddy of mine in BC. Depending on how things go with Amelia, that list might be getting smaller very soon.  That\'s about it. I should probably go to bed now, I\'ve been trying to get my schedule somewhat back on track. The past two nights I\'ve gone to bed after my mom leaves for work in the morning at 6:30 AM. Oh yeah, I get some cash tomorrow so I might go see a movie this weekend. I\'ll most likely be going alone; aside from Amelia there\'s nobody I want to go with.  Off to bed... '),(7634,882528,'05,August,2004','2004-08-05 00:00:00','   Things just get worse and worse...  I woke up crying this morning from a dream I had about Amelia. I had dreampt that another guy had gone out with her and was bragging about how he\'d used her. I just can\'t deal with this, I can\'t stop thinking about her and every time I do I start to cry... I want her back so badly, I miss her so much, but she won\'t even talk to me... I don\'t know what to do anymore. I want to call her so badly... But I sent her a final text message saying I\'d stop bugging her... I don\'t want to become stalkerish, I just want to talk to her again so much... I think I might love her... '),(7635,882528,'03,August,2004','2004-08-03 00:00:00','   Things have been especially dull and depressing lately.  Ally messaged me this morning asking for my profile name, so I re-added her to my contact list. Not like I have to worry about cheating anymore. Turns out she has a new boyfriend.  I talked to Karli today briefly before getting brushed off again. If she had any less interest in me she\'d be ignoring me outright. I guess I didn\'t respond fast enough to her interest so she moved on, I was just her flavour of the week. She\'s lost any trust I once thought of giving her.  I sent Amelia a text message earlier today, no reply. I still havn\'t heard from her since she broke up with me over MSN. I find it kind of unfair how when I dumped her, she demanded I do it in person, yet now that our roles are reversed she won\'t do that for me.  So obviously things aren\'t going well for me... Everything keeps reminding me of Amelia and all I want to do is just cry and cry... it hurts so much losing her... For once I was actually happy and I was starting to loosen up around her, then she went and destroyed that. I thought I almost loved her. '),(7636,882528,'01,August,2004','2004-08-01 00:00:00','   Girls suck.  So I had the best date I\'ve ever had that night with Amelia at K-Days. I thought things were back to normal, and even that things might work out well after all. Apparently Karli still had interest in me. Determined not to cheat again, I told her about my relationship with Amelia, immediately losing that interest. I even deleted every girl on my contact list within a 500km radius, including Ally whom I havn\'t spoken to since. But what\'s it matter, I got the girl back right?  Two days ago Amelia dumped me. Again.  She came online so I greeted her, and got a pretty dull response. I asked if she wasn\'t feeling talkative, and then being affirmed I asked why. She told me she had read this journal, and copy and pasted a block of text for me to read.    \"So here I am, bored and alone as usual. It seems I\'m pretty much left with nothing now. I craved to be with someone so much that I wrecked my relationship with Amelia for a single day with Ally, and now I probably won\'t see either again. I guess it doesn\'t matter, I wasn\'t happy with Amelia anyway... I had much more fun that day with Ally than I had ever had with Amelia, and in that one day I spent more time with Ally than my combined time with Amelia. That\'s how a real relationship should be... Too bad it only lasted one day...\"     - Thursday, July 22, 2004    Now let me explain a few things. July 22, 7 days after being dumped by Amelia, I made this post. I said that I wasn\'t happy with Amelia, and that I had much more fun with Ally, which up until K-Days was absolutely true. But 3 days later we got back together, the next day we had an awesome date, and suddenly I was happy again. Does that matter? No. She told me I had \"led her on\" and that this was my third strike, then said goodbye and blocked me. I sent her an email to which I\'ve had no reply, she hasn\'t spoken to me since. '),(7637,882528,'10,August,2004','2004-08-10 00:00:00','   It\'s been so lonely lately... I feel like I\'ll be stuck single again forever. I\'m down to about 2 friends who talk to me - Natalie and Kat. I\'ve given up on Karli, she went on vacation for the rest of the month, and Ally and I don\'t talk anymore. I got bored and decided to try and make some new friends so I browsed Nexus for a while, posting comments on some girl\'s profiles. No replies... '),(7638,4325889,'24,August,2004','2004-08-24 00:00:00','   STARTING  After arranging the `Setup\'.exe, run ragnarok.exe. When you run `Ragnarok\', the new patch will be downloaded automatically and after the patching, you will able to read Level Up! Inc.â€™s official notice window that includes updates and changes in the game for the week.You can log on and choose the set of servers you want to play in. You can then log-in with your account which you registered on the Level Up website. (Take care of your password! Donâ€™t entrust it with anyone to avoid getting a hacked account. Also watch out for the keylogging programs!You will then prompted to choose which server to play (if you chose first set, then it would be Chaos, Loki, Iris and Lydia; the second server set includes Fenrir, Odin and Sarah). When you successfully log in to the server, you can see 3 empty slots for characters.Choose one empty slot to make a character. Meanwhile there are 36 Hairstyles and 8 Hair colors you can choose for Characters. Sooner or later, it will combine into 72 available types with clothes and other equipments. On the `Make\' section, you can arrange the beginning Status of Character as you want. The details of each status are the following STR(Strength) Increase Hit Point and the Maximum Weight which you can carry on. AGI(Agility) Increase Dodge and Attack Speed. VIT(Vital) Increase Defence Rate, Vitality and HP recovery. INT(Intelligence) Increase Magical Hit Point and Magical Defence Rate. DEX(Dexterity) Increase Accuracy Rate and Weapon Damage. LUK(Luck) Affect in Critical Attack,Recovery from Abnormal Status or other Various Stuffs. If you fully understand the effects of status points from the beginning, you could create a powerful and distinctive character for your class. Furthermore your initial status will greatly affect your character as it reaches 2nd job class.   '),(7639,4325889,'24,August,2004','2004-08-24 00:00:00','   CHATTING BOX  In the left side of the box, you can type a character name that you want to whisper to. Or just Click the button next to this box to see the list of previous names that youâ€™ve whispered. You can contact Friends easily this way. In the right side of the box, you can type the message you want to send. The left button on the right side handles either Public message or Private message for party or Guild, the other button on the other hand, lets you handle the size of Message history window (or you could just press F10).   '),(7640,4325889,'24,August,2004','2004-08-24 00:00:00','   BASIC INFORMATION WINDOW  All of character information is shown on this window briefly. Also, you can view particular functions on this Window. Job Level is each job class level after Novice level 9. Every time your Job level goes up, you can get one skill point. The Gauge bar beside each Level Stands for your Experience % which you\'ve gained so far. If it\'s full by 100%, the character gets one level up then the Gauge Bar will reset to 0% again for the next level up. Weight stands for \'the Present weight / Maximum weight \' that your character is carrying. When the present weight is over than 50% of maximum weight, It will be shown in Red colour. From that moment , It is impossible to regenerate HP/SP Naturally. Or when the present weight is over than 90%, The character wonâ€™t be able to attack any monsters. You\'d better store the excess items in Kafra\'s warehouse if you don\'t want to be killed by monsters without even the chance to fight back. For reference, the Basic Information Window is never wiped out and always remains as a Smallest window.'),(7641,4325889,'24,August,2004','2004-08-24 00:00:00','   STATUS WINDOW   You can see the character\'s present status on this Window. When you bring your cursor over each stat, you can see the full name for each status. Every time when the character gets a level up, he will gain an amount of status points depending on your base level, You can distribute them into 6 Status whatever you want. On the Right Side of Status Window, you can see Several Abbreviations of Actual Effect. (i.e-ATK or DEF). They show Actual benefits which the Character can receive from Present stats (Left Value) and bonus points (Right Value) added from equipment. When the character levels up, Blue button appears on the bottom of right side of the monitor screen. Click it to open the status window. Click the blue Arrow to spend the points you get. You can upgrade stats as you want. Unfortunately `Tendency\' is not implemented yet.'),(7642,4325889,'24,August,2004','2004-08-24 00:00:00','   SKILL WINDOW   You can see Present skills. Every time the character\'s job level is up, you gain a skill point. You will see blue button in the bottom left corner of your monitor screen. You can check skill information on each skill by right clicking of your mouse on it. Passive skill means you don\'t have to spend your SP to use it. it is always in effect so you donâ€™t have to activate it manually. You can check the section related to job classes. Also, it is possible to assign skill points on your Hot key window â€˜F12â€™ as well as Items. You can use your favorite skills quickly and easily this way.'),(7643,4325889,'24,August,2004','2004-08-24 00:00:00','   ITEM WINDOW   It Shows your inventory.Item window consists of 3 parts,the equipments you are wearing are not shown on the item window. Just Right Click on a item to see its description and a picture. And it is possible to change any equipment by Dragging &amp; Dropping it onto your equipment window, or using the shortcut `ALT+Q\'. You can also double click the item to equip. You can register 9 items or skills from F1 to F9 in the hotkey window.'),(7644,4325889,'24,August,2004','2004-08-24 00:00:00','   CHAT ROOM WINDOW  At novice level 4 and above, you can make your own chat room with `Chat\' Button on the Basic Window. Even though You can\'t move while Chatting, item trading is available.   '),(7645,4325889,'24,August,2004','2004-08-24 00:00:00','   MINI-MAP WINDOW  With \'map\' button, you can use mini-map in various ways. It can be translucent, opaque and invisible. Also with the +, - button, is possible to enlarge and reduce. Red spot on the mini-map is a loading point to the next map. Player\'s location is shown like a white arrow. The direction of a white arrow shows which way the player facing. Colored arrows are the location of your party mates on the same map.   '),(7646,4325889,'24,August,2004','2004-08-24 00:00:00','   PARTY WINDOW  You can see all party members who are online and their present location. Party members who are on the same map as you, are shown on the mini-map, Party window will be updated in the future patches so that it will act as messenger function like party members log-in. Also it will be upgraded like G-Friend that can connect between Ragnarok and outside.   '),(7647,4325889,'24,August,2004','2004-08-24 00:00:00','   EQUIPMENT WINDOW   It Shows the character\'s equipped items. Each equipped item can be right clicked for the description and picture. You can equip and remove items by dragging and dropping between the equipment and inventory windows. Equipment can also be registered on a hot key so you can change your weapon easily.'),(7648,4325889,'24,August,2004','2004-08-24 00:00:00','   OPTION WINDOW   This window controls the BGM and Sound Effect and turns them on/off. You can also select window skins that you could download in the net.'),(7649,4325889,'24,August,2004','2004-08-24 00:00:00','   SHORTCUT WINDOW  Keys From \'F1\'to \'F9\' are Hot keys for each item registered in Hot key window.   '),(7650,4325889,'24,August,2004','2004-08-24 00:00:00','   TRADING ITEM WINDOW   When you right click on another player, and chose â€œAsk _____ for a dealâ€, \'Item Trade Window\' Pops up. You can trade your Item/Zeny with that player. After both players agree to trade with each other, drag items or zeny from your inventory to the trade window. Also you can View items by right clicking on them. Trade will not be able to succeed as long as you don\'t press `OK\' Button. (Tip: Always check and recheck your trades to make sure that you are getting a fair trade).'),(7651,4325889,'24,August,2004','2004-08-24 00:00:00','   THE OTHER INTERFACES   ESC This is the function key, it calls the exit to window command or return to the game Window. F1~F9 Shortcut keys for each item registered on Hot key window. F10 to Adjust the size of chat window F11 Clears all windows except the basic information window and  the chatting box. F12 to Call Hot key window Insert The hot key for sit or stand command. Also you can type \'/sit or /stand\' on the chatting Box. alt+home Ground cursor on / off on / off alt+end The hot key for character Health and Spell Point gauge on / off .You\'d better turn on all the time to check your HP and SP. page up& down to Scroll Text in interfaces and the chatting box, Mouse whirl will also scroll   '),(7652,4325889,'24,August,2004','2004-08-24 00:00:00','   CAMERA ZOOMING INTERFACE  When you drag mouse from right to left with the right mouse button held down, he camera rotates 360\'. Holding down the `Ctrl\' key and the right mouse button while dragging the mouse up and down, makes that camera Zoom in and Zoom out toward the character. Holding down the \'Shift\' key and the right mouse button while dragging the mouse up and down, controls the angle of the Camera for fixed range. You are all set up! Good luck on your journey and we wish you success on your quests as you venture in the kingdom of Rune-Midgard!   '),(7653,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    beba and i&nbsp; urlLink '),(7654,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    my beeeeeeeeba&nbsp; urlLink '),(7655,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    my real pet IDEAFIX!!!!!!!&nbsp; urlLink '),(7656,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    this is my ciberpet at www.neopets.com!&nbsp; urlLink '),(7657,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    ME BELOVED STUFF MY ANIMITA&nbsp; urlLink '),(7658,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    STUFF ON MY CUCHITRILL&nbsp; urlLink '),(7659,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    MY BEDROOM&nbsp; urlLink '),(7660,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    I LOVE THIS FOOOOOOOOOOOOD I CAN KILL FOR THIS!&nbsp; urlLink '),(7661,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    this is my ocio (hobby)!!! the rakuraku dinokun cyberpet im a fan of ciberpets! lame that he die every week jajajaja ^^;;; why do I always pick the lion if i have like 33 pets to choose? @_@?Â¿;; &nbsp; urlLink '),(7662,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    en el barrio red la zanahoria chick&nbsp; urlLink '),(7663,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    hey carix zanaharoria chick cool view!&nbsp; urlLink '),(7664,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    para de gozar ella en amsterdam la llevay zanahoria!&nbsp; urlLink '),(7665,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    shin seung hun at concert !!!&nbsp; urlLink '),(7666,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    more hip hop gear&nbsp; urlLink '),(7667,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    shinta my character&nbsp; urlLink '),(7668,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    my art work&nbsp; urlLink '),(7669,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    my teacher at the museum and my classmates&nbsp; urlLink '),(7670,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    my pal jazz&nbsp; urlLink '),(7671,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    near home santiago&nbsp; urlLink '),(7672,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    la caro and her bfriend at spain!&nbsp; urlLink '),(7673,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    CERTAMEN GRAFICA MIRA QUIEN YTA ALLI!&nbsp; urlLink '),(7674,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    YO EN LA U&nbsp; urlLink '),(7675,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    GRAFICA&nbsp; urlLink '),(7676,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    HIP HOP STYLE AGGAIN&nbsp; urlLink '),(7677,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    YO EN PAINBALL&nbsp; urlLink '),(7678,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    THIS IS A NICE PIC THANKS MIKE!&nbsp; urlLink '),(7679,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    MY GOD CAROLINA U ARE THE BEST THANKS SO MUCH ZANAHORIA TEAM!&nbsp; urlLink '),(7680,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    COOL AINT I? JAJAJAJAJ&nbsp; urlLink '),(7681,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    MY SPONGEBOBSQUAREPANTS CELLPHONE&nbsp; urlLink '),(7682,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    HIP HOP GEAR AT MOMS BEDROOM&nbsp; urlLink '),(7683,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    THIRSTY? JAJAJAJA NEVA AGGAIN PLAYING MAN!!!!!!!&nbsp; urlLink '),(7684,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    SOME OF MY URBAN ART WORK OK AINT THAT COOL BUT......DRAFIITTY IS HARD MAN!&nbsp; urlLink '),(7685,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    ONDA SHIN SEUN HUN PERFIL!&nbsp; urlLink '),(7686,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    SO CUTIE!!!&nbsp; urlLink '),(7687,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    VIC GO GO GO GOOOOO!&nbsp; urlLink '),(7688,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    HEY MAN LETS GO TU PUMP IT UP&nbsp; urlLink '),(7689,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    she at the cellphone come on!!! shes a buisinessgirl!!&nbsp; urlLink '),(7690,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    JAJAJAJA QUE EXTRA!&nbsp; urlLink '),(7691,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    mi pal reventada!&nbsp; urlLink '),(7692,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    jajajaja playing?&nbsp; urlLink '),(7693,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    chuke cheese funny stuff&nbsp; urlLink '),(7694,3768397,'27,June,2004','2004-06-27 00:00:00','    urlLink    thats me!&nbsp; urlLink '),(7695,3768397,'27,June,2004','2004-06-27 00:00:00','    HEYA THIS IS SOME PERSONAL PICS ABOUT  FRIENDS LIVE CHILE STUFF  SO U CAN SEE SOME OF IT '),(7696,3768397,'28,June,2004','2004-06-28 00:00:00','    urlLink    victoriaï¿½s coussins in Metri&nbsp; urlLink '),(7697,3768397,'28,June,2004','2004-06-28 00:00:00','    urlLink    all the magick of the chilean south at  ptto montt&nbsp; urlLink '),(7698,3768397,'28,June,2004','2004-06-28 00:00:00','    urlLink    ESTRELLA THE CUTIE HIP HOP GEAR BABY EVER AT METRI&nbsp; urlLink '),(7699,3768397,'28,June,2004','2004-06-28 00:00:00','    urlLink    my kpop shinseunghun & seungyoo&nbsp; urlLink '),(7700,3768397,'28,June,2004','2004-06-28 00:00:00','    urlLink    THE KODAK MOMENT AT MIKES GANGSTA PLACE! GREAT MOMENT MIKE!MY SING ALONG BUDDY&nbsp; urlLink '),(7701,3768397,'28,June,2004','2004-06-28 00:00:00','    urlLink    THA KILLA GANSTA LOOK WITH MY CHINGU MIKE&nbsp; urlLink '),(7702,3768397,'28,June,2004','2004-06-28 00:00:00','    urlLink    MY BEST MY ONE THE ONLY!!! MIKE!!!&nbsp; urlLink '),(7703,3768397,'29,June,2004','2004-06-29 00:00:00','    hey guys well a tiny test to see if this is uploading '),(7704,3768397,'29,June,2004','2004-06-29 00:00:00','    urlLink    asi que ociosa caro no? SI TENGO UN POST IT SOY CULPABLE t_t&nbsp; urlLink '),(7705,3768397,'29,June,2004','2004-06-29 00:00:00','    urlLink    no se ocupar msn!!!! caro enseÃ±andome msn for dummies jajajaja&nbsp; urlLink '),(7706,3768397,'29,June,2004','2004-06-29 00:00:00','    urlLink    grafic desighn work&nbsp; urlLink '),(7707,3768397,'29,June,2004','2004-06-29 00:00:00','    urlLink    grafic desighn my work&nbsp; urlLink '),(7708,3768397,'29,June,2004','2004-06-29 00:00:00','    urlLink    my work of grafic design&nbsp; urlLink '),(7709,3768397,'29,June,2004','2004-06-29 00:00:00','    urlLink    Our boulevard with an a Miami structure of Mall and for tiiiiny ignorants YES WE HAVE FRIDAYS, STURBUCKS, BOWLING, MOVIES, HOOTERS, RUBBY THUESDAYS, KENTUKY MAC DONALDS, CHUKE CHEESE AND ALL THAT USA STUFF chile is tiny but aint get safe from the american Idol&nbsp; urlLink '),(7710,3768397,'29,June,2004','2004-06-29 00:00:00','    urlLink    Santiago the capital that shows tiny spaces of a jungle of cristal....marriot hotel, new one and great service, meanwhile they didnt let me in at the sheratton hotel cause i was on a hip hop gear and i couldnï¿½t sketch the people without a permission, this people treat me like a princess when i get there to draw, and also on hiatt but thats another story&nbsp; urlLink '),(7711,3768397,'30,June,2004','2004-06-30 00:00:00','    urlLink    MET THE ONE, THE ONLY, PUSSS IN BOOOTS, SHIIIIIII&nbsp; urlLink '),(7712,3768397,'05,July,2004','2004-07-05 00:00:00','    urlLink    robando estilos de ropa este me gusto para los 26 oh si gracias jjonabebe thanks!! ^^ ;;;;;&nbsp; urlLink '),(7713,3768397,'05,July,2004','2004-07-05 00:00:00','    urlLink    este estilo me gusta!!!&nbsp; urlLink '),(7714,3768397,'17,July,2004','2004-07-17 00:00:00','    urlLink    now my fix looks amazing!&nbsp; urlLink '),(7715,3768397,'17,July,2004','2004-07-17 00:00:00','    urlLink    chilean nature at my campus&nbsp; urlLink '),(7716,3768397,'17,July,2004','2004-07-17 00:00:00','    urlLink    why hoon hates pictures? I really dunno! but i hate the dam kid next to him!&nbsp; urlLink '),(7717,3768397,'17,July,2004','2004-07-17 00:00:00','    urlLink    ice cream with borah&nbsp; urlLink '),(7718,3768397,'17,July,2004','2004-07-17 00:00:00','    urlLink    the kids at the grass bear&nbsp; urlLink '),(7719,3768397,'17,July,2004','2004-07-17 00:00:00','    urlLink    this child is very social person and funloving her name is Hyun jin&nbsp; urlLink '),(7720,3768397,'17,July,2004','2004-07-17 00:00:00','    urlLink    this is hoon jin very clever kid and my student at english class&nbsp; urlLink '),(7721,3768397,'17,July,2004','2004-07-17 00:00:00','    urlLink    my babysitting day! fist time i do it everithing went oko doki ! this is borah at the legoland show&nbsp; urlLink '),(7722,3768397,'18,July,2004','2004-07-18 00:00:00','    in vacations but tired man!        working at vacations really tireds you but i have the hopes to relax a while      during the week        &nbsp; '),(7723,3768397,'19,July,2004','2004-07-19 00:00:00','    urlLink    some drafity of mine cool isn it?&nbsp; urlLink '),(7724,3768397,'24,July,2004','2004-07-24 00:00:00','    urlLink    See this is was i been talking about I look like a vampire this is the stupiest face I ever had an a picture, worst it never took that dam thing till this one i look ugly&nbsp; urlLink '),(7725,3768397,'24,July,2004','2004-07-24 00:00:00','    urlLink    trying to make a draw in motion&nbsp; urlLink '),(7726,3768397,'24,July,2004','2004-07-24 00:00:00','    urlLink    Back back finally back!&nbsp; urlLink '),(7727,3768397,'24,July,2004','2004-07-24 00:00:00','    urlLink    I hate my camera really&nbsp; urlLink '),(7728,3768397,'24,July,2004','2004-07-24 00:00:00','    urlLink    nail bed god for the back&nbsp; urlLink '),(7729,3768397,'24,July,2004','2004-07-24 00:00:00','    urlLink    look at this poor shark with tha get me aoutta here look&nbsp; urlLink '),(7730,3768397,'24,July,2004','2004-07-24 00:00:00','    urlLink    the aquarium: 2 words: tiny, bored&nbsp; urlLink '),(7731,3768397,'24,July,2004','2004-07-24 00:00:00','    urlLink    hong jin and his art work at the coffe shop&nbsp; urlLink '),(7732,3768397,'24,July,2004','2004-07-24 00:00:00','    urlLink    this ball floats at the sky&nbsp; urlLink '),(7733,3768397,'24,July,2004','2004-07-24 00:00:00','    urlLink    and gaby tried to put it on the sky also&nbsp; urlLink '),(7734,3768397,'24,July,2004','2004-07-24 00:00:00','    urlLink    hong jin trying to change the water path&nbsp; urlLink '),(7735,3768397,'24,July,2004','2004-07-24 00:00:00','    urlLink    aggain!&nbsp; urlLink '),(7736,3768397,'24,July,2004','2004-07-24 00:00:00','    urlLink    this was interesting a tornado on the water lame the camera didnt help me&nbsp; urlLink '),(7737,3768397,'24,July,2004','2004-07-24 00:00:00','    urlLink    hong jin also tried to&nbsp; urlLink '),(7738,3768397,'24,July,2004','2004-07-24 00:00:00','    urlLink    creating a giant bubble&nbsp; urlLink '),(7739,3768397,'24,July,2004','2004-07-24 00:00:00','    urlLink    I think of the interactive musseum this is the funner stuff the giant bubbles&nbsp; urlLink '),(7740,3768397,'24,July,2004','2004-07-24 00:00:00','    urlLink    Gaby with the magnets&nbsp; urlLink '),(7741,3768397,'24,July,2004','2004-07-24 00:00:00','    urlLink    my student Gaby&nbsp; urlLink '),(7742,3768397,'24,July,2004','2004-07-24 00:00:00','    urlLink    hong jin my student at the mim museum this weekend hes at the theater place&nbsp; urlLink '),(7743,3768397,'07,August,2004','2004-08-07 00:00:00','    DAMMAN VACATIONS ARE OVER !!!     OK LETS PRAY TO PASS ALL'),(7744,3768397,'07,August,2004','2004-08-07 00:00:00','    urlLink    good pic ^^&nbsp; urlLink '),(7745,3768397,'07,August,2004','2004-08-07 00:00:00','    urlLink    are u gonna call me? &nbsp; urlLink '),(7746,3768397,'07,August,2004','2004-08-07 00:00:00','    urlLink    you wanna a cellphone ? hehe call me :P&nbsp; urlLink '),(7747,3768397,'07,August,2004','2004-08-07 00:00:00','    urlLink    rainbowbrite the oldies!&nbsp; urlLink '),(7748,3768397,'07,August,2004','2004-08-07 00:00:00','    urlLink    grafitty in stgo chile&nbsp; urlLink '),(7749,3768397,'07,August,2004','2004-08-07 00:00:00','    urlLink    more shin the concert poster ad&nbsp; urlLink '),(7750,3768397,'07,August,2004','2004-08-07 00:00:00','    urlLink    things from the past sha la la lalala sha la la la la la la!!!!&nbsp; urlLink '),(7751,3768397,'07,August,2004','2004-08-07 00:00:00','    urlLink    my favorite anime from the 2000&nbsp; urlLink '),(7752,3768397,'07,August,2004','2004-08-07 00:00:00','    urlLink    yo my alter ego&nbsp; urlLink '),(7753,3768397,'07,August,2004','2004-08-07 00:00:00','    urlLink    shin seun hun at the concert of this year T_T i wish i could been there&nbsp; urlLink '),(7754,3768397,'07,August,2004','2004-08-07 00:00:00','    urlLink    tseo taii  i think sorry if im wrong&nbsp; urlLink '),(7755,3768397,'07,August,2004','2004-08-07 00:00:00','    urlLink    shin seung hun my favorite singer&nbsp; urlLink '),(7756,3768397,'07,August,2004','2004-08-07 00:00:00','    urlLink    everybody loves hello kitty me too&nbsp; urlLink '),(7757,3768397,'14,August,2004','2004-08-14 00:00:00','    AHORA SI ESO FUE EL COMISTRAJO!&nbsp; urlLink '),(7758,3768397,'14,August,2004','2004-08-14 00:00:00','    urlLink    no coments&nbsp; urlLink '),(7759,3768397,'14,August,2004','2004-08-14 00:00:00','    urlLink    :)&nbsp; urlLink '),(7760,3768397,'14,August,2004','2004-08-14 00:00:00','    urlLink    so....&nbsp; urlLink '),(7761,3768397,'14,August,2004','2004-08-14 00:00:00','    BUENO ESTAS SON UNA DE LAS  FOTOS DEL COMISTRAJO AHORA LO QUE NO SE VIO!&nbsp; urlLink '),(7762,3768397,'14,August,2004','2004-08-14 00:00:00','    urlLink    THA ADVERTISMENT MASS&nbsp; urlLink '),(7763,3768397,'14,August,2004','2004-08-14 00:00:00','    urlLink    YA BROTHA PINKY MAN FUNNY YALL!&nbsp; urlLink '),(7764,3768397,'14,August,2004','2004-08-14 00:00:00','    urlLink    THA BROTHA GANSTA LOOK (I LOOK UGLYYYY t_t)&nbsp; urlLink '),(7765,3768397,'14,August,2004','2004-08-14 00:00:00','    urlLink    ACA FELIPE DE EXTRA CON CARLOS DIAS....EN LA NUEVA TELESERIE&nbsp; urlLink '),(7766,3768397,'14,August,2004','2004-08-14 00:00:00','    urlLink    ME PUEDO DAR UN LUJO ALUMNOS? JAJAJAJA&nbsp; urlLink '),(7767,3000240,'19,March,2004','2004-03-19 00:00:00','   Wooo finally I have a free house, and do you know what this really cool guy does when it happens, he SLEEPS!! God if I had a pound for every time I\'ve done something boring when the chance came to do something even remotely funky, I guess I could pay for Roman Abramovich to kiss my ass. Anyway, I\'ve had someone mock me today for not including them in this blog, or well at least the past about what townies are acceptable in society from my standpoint, namely a girl from college named Jess. So I\'ve decided this whole thing needs a list of acceptable orgasmic townies, so here they are:  Ian Griffiths Jessica Diggle Lauren Coulten Maryanne Cook Ali Ross (your a closet townie and you know it) and Daniel Kesslar (an in the closet wannabe townie)  One day I will all make you realise anyway, what fools you are for your ignorance in true music, the English Language and in (all except Ali) the greatest football team is. Anyhoo I\'m of for a korma, a bath and msn so I\'ll see you bubbers later . . . xBuBzx  '),(7768,3000240,'18,March,2004','2004-03-18 00:00:00','   Well today didn\'t go as badly as I thought it would, made all the more better, by a consistently constant barrage of abuse I hurled at my dearly beloved mathematics teacher - Jon Hart. I was desperate, and I mean of the highest most amazingly vital necessity, to take a piss (if not I assure you there would have been a puddle on the floor if I had held the tide in me any longer), and he tried to grab hold of me literally, and have one of these chats he seems to solely give to myself and NO ONE ELSE!!. So I replied \"SHUT UP YOU IDIOT\" to which he replied \"WHAT!\", as I ran as fast as I could to the nearest lavatory. He tried (the poor thing bless him) in the lesson there on in, but thus, it was no use to himself, as I continued to mock and down right dissolve his self esteem right to the lessons end. It was a fabuloussite to behold . . . xBuBzx  '),(7769,3000240,'17,March,2004','2004-03-17 00:00:00','   Here\'s an interesting article from my favorite nine inch nails website   urlLink www.sickamongthepure.com  about a responce to a previous article made in the last months edition of the website made, by an overly concerned fellow NIN fan; and how everyone is capitalist in the music business these days. If you\'ve got the patience to read it, then in my opinion its greatly worth it as an honest insight into popular music today:      Everyone is a Capitalist in the Music Biz: But They\'re Not All In It For the Cash    The term immediately hurled at a band that has become popular is the evil little title of \"sell out\".  This term is so easily shoveled out when a single from a long time underground success becomes a regular on the radio.  Now, though it may be fun to say mean bits about bands we don\'t like, what happens when it\'s a band you truly respect?  What if you still like the new stuff?  [Gasp]  Does that mean your band has dragged you down into the realm of \"crappy music\"?  Worse yet, what if you hate the new single from your long time favorite band?  Have they sold their soul for a bit of fame? NO!  I was perusing the \"From the Masses\" section of this magazine a while back and came across a very, well shall we say honest, reaction to an article regarding the song \"Deep\".  Well, I went back through the article that spawned such a very \"emotional\" response from a reader.  The article was a writer\'s opinion on the apparent lack of musicianship on the track.  Now, I will admit I did not think that \"Deep\" was Reznor\'s finest hour but, it had a certain simple quality I was not used to from Nine Inch Nails.  It got me to thinking about the intention behind the song.  Did Reznor really lend his music to an album just to cash in?  Not necessarily.   Yes, this track was a departure from what we have come to know and love about Nine Inch Nails, but I think it is that very thing which leads to Nine Inch Nails to be such an interesting band.  The sound they have produced has never been the same regurgitation, over and over, again re-packaged and re-released (for reference please see: Creed.)  They have not found a set formula for each album and become stagnant with the comfort of a working system of creation (see also: Limp Bizkit.).  I think variation and a constant evolution is essential to the progression of music.  Not every facet of the evolution process is going to be pretty.  All Darwinist\'s, I think, would agree that some of our post monkey, pre-modern human stages were a little less than beautiful. My point being, yes \"Deep\" did not exhibit the usual musical power previously wielded by Reznor.  However, this does not mean \"Deep\" is the end all be all of Nine Inch Nails.  It was not meant to be the culmination of the efforts of those involved with the project.  It was a song that Reznor was likely comfortable with releasing because he felt it would accomplish the one thing he set out to do: make music.  Not every song is going to be a paramount piece of work.  If there were a way to accomplish that, the music world would never change.  Music is a system of trial and error.  If everything had been completely sussed out then the production of new music would be redundant and, well, pointless.   It is ridiculous to fault a musician for making a song you don\'t like.  For most people good music is not something easily defined.  We will like music depending on how it connects with us.  We all have a song we may not admit to liking in public but secretly do nonetheless because it holds some sort of special memory or emotion for us.  The goal of creating good music is to create music that envelops the listener in a world of pure emotion through the shaping of sound waves and the manipulation of visual images projected from the writer, either through lyrics or through pure musical expression.  I do think \"Deep\" accomplished that, though maybe, as I said, not to the same level as, let\'s say, \"Something I Can Never Have\".  Reznor probably just wanted to vent and it came out as \"Deep\".   The reader who responded to the article, Mr. Self Destruct, also pointed out a very honest fact: \"The song \"Deep\" is simply just a spur-of-the-moment-to-make-some-fucking-money-song.  If it took you five years every time to record a brilliant album, I\'m pretty sure you\'d want to release a bullshit pop song every once in a while too.  What use is a mind made for making music if you can\'t afford to eat Ramen noodles?\"  We all often forget music is a business as well.  You have to have money to be able to sustain yourself for five years so you can put together a fantastic album. While I do not agree \"Deep\" was a \"spur-of-the-moment-to-make-some-fucking-money-song\", I do think that putting it on the Lara Croft: Tomb Raider soundtrack may have been a way to generate a bit of revenue.  But also look at the greater scheme of things.  By putting the song on a more popular venue Reznor was able to share his feelings with people who may have NEVER dreamt of even listening to Nine Inch Nails.  Yes, is was shocking to have a NIN song share an album with a Missy Elliot song but does doing this invalidate all Trent Reznor has done in the past?  Does this mean he now only wants to make money?  Is Tapeworm only going to be a CD of \"Deep\" remixed 12 \"new\" ways?  No.   As far as the claim that the release of halos with re releases of songs we have already heard were done for the mere purpose of making money, again look at the bigger picture. Most of the halos with re-releases on them weren\'t released in America.  The \"We\'re In This Together\" singles (which is my guess as to which albums Curtis Westman was referring to) were released in Sweden, again exposing a new fan base to the music of Nine Inch Nails.   This isn\'t to say there are not bands out there that do in fact make music solely for the money and fame aspects of the endeavor, but just because a band makes a song that might be more palatable to a new realm of fans does not mean the band has sold out or is now just in it for the money.  That only happens when a band finds a sound that works and beats the particular horse until it bleeds a painful death (refer again to: Limp Bizkit).  Innovation and variation have to be balanced with a business sense.  Though Reznor may not have sold his soul to Carson Daley or whatever tool hosts TRL now, he did take advantage of MTV when he saw it could get his music to the masses.  He sat and talked with his share of VJs.  The one constant, though, has been everything he has produced has been a new experience for him.  If it leads a parallel with your life or mindset and you further enjoy the song, great, if not you\'ll just have to see what lies over the next hill.  Music is not a level path easily navigated.  It is rather a varied plethora of artistic expression and as long the music is constantly following that path; it is a brilliant artist behind the music.  It is only when music finds a comfortable spot on the path and sets up shop there that makes for awful mind-numbing music.  So yes, there are people making money with music.  Music can be a product used for financial gain, just as a painter may do a commissioned piece.  But there are also those who just wish to capitalize on the chance to share a glimpse of their emotion at the time.  Andy Worhol painted portraits with the same pattern on the same size canvas for money to fund his studio, but the portraits were also being produced on the same size canvas because he planned to put them together to form a giant mosaic made up of all the people he had painted.  He had a greater vision.  I think that just because Reznor made something with a more pop sensibility, it does not make him a sell-out; there is still a library of work to come.  \"Deep\" may just be the small piece, the single tile, in a much larger mosaic of artistic expression, a single step upon a very long road we have not come to the end of yet.            xBuBzx  '),(7770,3000240,'17,March,2004','2004-03-17 00:00:00','   I woke up rather well today (well being where I don\'t oversleep), Got all my college stuff together, and I turned on my phone, only to open a message accusing me of spreading bull about one of my oldest friends *who going to remain nameless hence forth* in Esher. I replied \"so what have I been saying then?\"  So abruptly he told me it was to do with personal information to do with him from which I\'ve denied to the fullest thing I can deny stuff. From previous fallouts that I\'ve been through with this type of circumstance, and knowing this bloke for a fair time, I wouldn\'t blame him if he decided to despise the ground in which I walked on. I\'ll be the 1st to admit that, in the past, I\'ve done some God awful sh-it things, which I\'d hasten to add that anyone who reads this blog isn\'t aware of, and won\'t be in the foreseeable future. Still though Since I\'ve gone to college, I\'ve grown up immeasurably, and this isn\'t me being egotistical, I have been told by quite a few people, on quite a few occasions in the recent past. Also virtually no-one in Esher knows this guy who I\'m talking about anyway (who will be referred to as erm hmmmm . . . Mr X), and the people who do I hate with a passion and don\'t talk to anyway. I have nothing to spread about Mr X anyway, so why would people be so callust and as cold as they are at the moment even just to spite me.   Anyway this hasn\'t been the best of start to the day, as I\'m sure your well aware of, but the day is still young, and there\'s a lot of things to be cheerful about I guess. Meh so I\'ll tell you later about the day. xBuBzx  '),(7771,3000240,'17,March,2004','2004-03-17 00:00:00','   Ok I\'m really soberly subdued, finding it particularly difficult to concentrate on anything other than this blog at the moment, and not work, and yes I know, that obviously it should be the reverse, I know, but work I find is an impossible chore at this time during the day, or any time of the day for that matter. Jesus knows I\'d like an admirable ethic in relation to working, I\'ve been praying on my knees for decades (3 to be exact), but it never seems to be able to pay of. So I\'ve come to the two most sensible conclusions I can think of:  God must really hate me, or i\'m just a lazy bugger, Personally I think both hypothesis are quite plausable.    I\'m listening to a weird trippy poppy/techno/rock kind of song by the dissociatives and called the thinking in reverse, and from my opinion, its quite a fresh outlook on how to try to construct a decent song. Any of you who are bothered to listen go to  urlLink http://www.thedissociatives.com/music/audio/audio.htm . Wait for it to kick in it takes a while for the initial intro to finish.  Anyhoo, Ian and myself have decided to go house hunting in the little hamlet of Mitcham this Easter. It\'s becoming increasingly evident to myself that I can\'t stand living with my parents (an obvious cliche for people my age as I\'m sure you\'re all too well aware of) however, I\'m really quite excited by the prospect with living with completely different people, in utterly unfamilier circumstances. I know where I live is comfortable, well of, where people love and respect me (I hope), and money is never an issue really. Life however I know is one great big lord-of-the-rings-est journey and I\'m more than curious and ready to find out the difficult circumstances that lie ahead, whatever they may be. As the saying goes (from the great Homer Simpson): \"I WANT TO LIVE MARGE, I WANT TO LIVE!\". xBuBzx urlLink  '),(7772,3000240,'16,March,2004','2004-03-16 00:00:00','   God I\'m at college bored and about to go to philosophy. WOW THIS IS HOW POINTLESS MY LIFE HAS BECOME - DOING A SUBJECT I DON\'T LIKE ANYMORE; WHICH WILL NEVER GET ME ANYWHERE ANYWAY xBuBzx  '),(7773,3000240,'16,March,2004','2004-03-16 00:00:00','   urlLink http://www.rotten.com/library/language/the-finger/shocker/    urlLink http://poetry.rotten.com/fingerhead/    urlLink http://www.rotten.com/library/medicine/bodily-functions/nursing/goat-suckling/    urlLink http://poetry.rotten.com/chopper/index1.html   xBuBzx  '),(7774,3000240,'15,March,2004','2004-03-15 00:00:00','   I thought slowly as time drifts by, Iâ€™d need sleep less and less. I mean why waste you life with unconsciousness unless you actually desperately need it. There\'s other far more interesting and important schools of thought to consider i.e. breathing.   Anyhow, I\'m also finding out how wrong I was as I found for yet another \"overslept\" excuse given to my old decrepit excuse for parents, for me not turning up to college yet again. Oh well This seems to affect my college work (I am infact a self-confessed lazy asshole) and I fail to hardly do any work which means I have to work outside the confides of the college, as I recently got god awful mock exam marks in both sociology and psychology, and as for philosophy . . . Well I haven\'t even done it yet gulps. Well I feel kind of null anyway not just because of the sleep; I just really miss my friends like Anna I guess and a whole load of other issues which, most probably are left deep in my subconscious somewhere.  I had my nephew over today, and he was being an obnoxious fucker as usual, so I stuck him in the oven and cooked him (lol only joking), but it\'s been a boring day nether the less. Mum is annoyed at me again, I guess her ego got bruised again or something, no idea why though. Anyhoo meh i\'m bored. So \'m gunna try to define youth into strict categories later. Keep my mind sane . . . xBuBzx  '),(7775,3000240,'15,March,2004','2004-03-15 00:00:00','   Okies this an article from one of my favourite sites in the world www.sickamongthepure.com (don\'t bother visiting it if your not a nine inch nails fan like myself). Otherwise enjoy the article:    Letter from a disgruntled human being to the state of everything     If I have to sit on hold for one more minute, I will, I swear, lose my mind.  It won\'t be pretty.  Not only will you not allow me to speak with your manager, you also will not satisfactorily rectify my situation, make good on your promise to me.  And it isn\'t because doing so would be terribly difficult, or even against policy; you simply won\'t, because you know you don\'t have to. Isn\'t that the wonderful thing about free will?  You don\'t have to do one iota more than you wish to, and yet there is nothing to prevent you from complaining about it for time untold.  You don\'t have to have an opinion about anything, or a foundation for the ones you do have, yet you are allowed to fault others for theirs.  You never have to do one kind act for another human being your entire life, yet you can request, nay, demand it, of others.  All of the laws and rituals we have in place to maintain order; in the end they really have very little to do with justice, don\'t they? So if I have to sit here at this insane gathering, and listen to one more of you prattle on about your back pain, your indigestion, and your headaches, while you are stuffing your faces with french fries, sitting in front of your computer and sucking back coffee and pop, I will, in fact, lose it altogether.  Another promising mind disappears into the chaos of evermore.  Another scream lost in the white noise of industry.  I must say, Everything, while I\'m still capable of saying it: you\'ve lost control of your state.  You\'ve succumbed to what you are. I am writing you this letter to tell you I want my subscription cancelled immediately.  I no longer want to turn on my television to watch the news and see these pitiful levels of \"corruption\" in my government, in their government, in all of them.  I no longer want to read excuses from my favourite musicians for not putting out another album, nor do I want to watch people\'s freaky fans demand things from other people they were lucky to get to begin with.  I no longer want to hear these well-groomed masses reiterating over and over again just how successful they are, in the hopes they themselves will believe it.  I no longer want to see one nation even try to justify going to war with another in this day and age.  I don\'t want to see your dominion looking to another source of power rather than working for themselves.  I would also like my money back for the time I did subscribe to your organization.  However, I realize this will be next to impossible, for I\'ve seen the way your people keep their books, and since they can barely differentiate between a fur coat, a tank, a taxpayer, and a nourishing meal, I don\'t expect they would be even capable of finding my file, much less gleaning any worthwhile information from it if they did. Furthermore, I would like an all-expenses paid one-way trip from The State of Everything to Somewhere Better, or, if that deems too expensive, Somewhere Else.  I am told in Somewhere Better, people don\'t destroy every single thing they touch.  I have heard they treat people with the reverence we give our possessions, that the people who have devoted their lives to leading them have done so because they are good and intelligent, and because they are capable of making their world a better place. Even in Somewhere Else, the laws are very clear on what is Granted, so clear in fact that nothing is ever taken for it.  Because they have been able to make this important distinction, the people there are no longer mired in the religious or secular teachings of the State of Everything, whose greatest of failed motives were to instill an appreciation for that very big difference and act accordingly.  The people there aren\'t paralyzed by our inability to think or act of our own accord without first placing someone or something clearly in the path of our progress, whom we can then blame and receive reparations from.  In Somewhere Else, people are smart.  Which is more than I can say for your customer service team.    Sincerely,  Disgruntled Human Being.   xBuBzx  '),(7776,3000240,'14,March,2004','2004-03-14 00:00:00','   I promised you people some background information on myself which some random people emailed me asking me for, when they strangely stumbled on this site during their boredom, and I hasten to tell you and them, I almost forgot so I apologise. Well; whoever you are who reads this more than perculiar site, this is a short history lesson in my self-procclaimed bubzification . . . My full name is Matthew Edwin Jack Smith, I\'m 16 (17 in 3 weeks yay), I live in West Molesey, Surrey in the UK. Im a rather disproportionately shaped looking bloke (ie ugly to non politically correct/normal folk). I go to Esher College which is about a mile away from Hampton Court where I study sociology, philosophy and psychology at as-level. I hope to be a counselor one day. My interests include sport (Chelsea fc especially) and alternative industrial, metal music. I have 1 brother (Andrew 30) and one sister (Joanne 32) and a five year old nephew called Harry. My parents (John and Carol) are both 57 and should be retiring (and die slowly in my opinion) this year. Any hoo that\'s all I really have to say for myself. Anything else you want know you\'d have to ask to me personally. Anyone wants to get in contact with me my msn is matthewsmith157@hotmail.com xBuBzx  '),(7777,3000240,'14,March,2004','2004-03-14 00:00:00','   The Teenage Sell-out    There you walk, With your self-governing esteem held high, Flicking your clichÃ© bleached hair, Drawing the wondering eyes of the evanescent primitive male, Masticating your gum as quickly as the inhalation from the cheap tobacco. Your vocabulary makes quantum physics, As straightforward as subtraction, Whilst your whole feeble-minded body, Yields to the nauseating tackiness of the whore image, In which you embrace upon yourself with blinded faith.  Ah your stubborn nature is the comedic notion to my ears. The limited thought perspective, Has transparently corrupted the link between your words and brain. The ineptitude of your moral standards,  Would indeed turn Plato in his grave.  Thank God that you exist in the parabola of your own universe, At least the clear minded ones can view your glass galaxy, Each time they flow past your laughable â€œindividualityâ€œ, With righteous pity.  xBuBzx  '),(7778,3000240,'14,March,2004','2004-03-14 00:00:00','   College is a funny thing. When there\'s nothing to (as there invariably isn\'t) I sit and people watch usually from the confides of my music in the LRC. I see people who used to have their rebellious streak fade to dust and turn to the sellouts I like to call townies. Now I\'m not saying all townies are bad people (take my friend Ian for example) but I\'d say, whilst yes I\'m horribly generalising here, that most are shallow, big headed, narrow minded individuals who have a warped view of what the modern day English language. What\'s more many few see to evade the evil clutches of the repo men or seem to navigate there way around a decent roofing company (bad attempt at a joke I know). I can\'t say i\'d like to force my views upon others on this matter, but surely there is another way to live out the \"most important years of your life\" for sure.  Anyhoo, they seemed to have found a new planet today, the 1st for 73 years. I\'m sad I know I view this as an important day in my life, but it\'s amazing what science finds nowadays. College was ok but I\'m increasingly finding even with people for compay at lunch, there\'s fuck all to do most of the time, unless you sleep on the couches, or go to Wimbledon or Surbiton at lunch, which can cost loads . . .  '),(7779,3000240,'14,March,2004','2004-03-14 00:00:00','   Omost of the time as I get bored from time to time, i\'m gunna blog some literature in here always entitled literature and this is my 1st entry. Hope you enjoy it:  One Of Them  Oh how I wish to be someone else: To walk along the corridor with my head held high To be looked up to on my own actions To have my time completely filled. To experience what I have always naturally craved- Not to be patronised. That is what I have always longed for.  Once I was close, but never close enough, I was amongst them, smiled and laughed with them But never really there. Then one day, the whispers came Apparently I was not â€œallâ€ man And with that, all smiles stopped Striking me down like a punch to the solaplexus  From which I never got up.  Having sold my secret soul to Damien Time, growth, life came to a hollow stop I existed inside the hell of my own Pandoraâ€™s Box. As my pain grew, their anger intensified; Small demons poked their fiery sticks through my bars As if they were never going to let me leave- Even heavenly intervention could not break their walls.  But over time, I was flung into memory Forever seen as an obsolete â€œfaggotâ€ at the back of the class. From time to time, their brimstone torched my body Until its casing hardened, formed a cast Inside which died my longing to be one of them.  '),(7780,3000240,'14,March,2004','2004-03-14 00:00:00','   This has been one odd weekend. After the rave, I\'ve just been thinking and contemplating about everything I can at the moment, like where in the hell is my life going. Im glad for the solidarity and consistently after school, but now I\'m just waiting for something big to happen once again. I\'m still in some parallel random sorting thought zone, trying to sort my whole group of friends into groups, and you know what, I think I\'m almost there at last.  Anyhoo, was a relatively average day at work, although I\'m getting increasingly pissed of with my new \"boss\" Becky, she\'s just become incredibly patronising and now should officially die - slowly - in her own filth. Apart from that it was ok and I came home to the great news that Manure (Man U) had lost 4-1 to City. Wrestlemania tonight (yer I know I\'m sad but who cares). I guess I\'ll post later on about more important stuff when anything happens to come to light in this ever increasingly warped mind of mine - BuBz  '),(7781,3000240,'14,March,2004','2004-03-14 00:00:00','   After scowering the internet for years, and taking so much form it, I guess its about time I gave a little back - so this is an introduction I guess. I\'ll give you all the relavent information about myself tomorrow when I get home from work tomorrow . . .  BuBz  '),(7782,3000240,'26,March,2004','2004-03-26 00:00:00','   I think it\'s going to be a rather kick ass weekend again as my weekends are slowly but surely becoming of late, as I\'m of to a rave with my friend Anna and going to an obscure but funky film festival on Saturday. I think also I\'m gathering some decent work ethic. I\'ve almost done my coursework for maths and I\'m beginning to work quite hard in the three as-levels I\'m taking, well all be it not at home and only in the college library, but hey it\'s a start at least, and is ten times better than what I could have managed say a month ago. Well have a good weekend whoever you are, I\'m of to do some psychology - whoooo me I\'m working.  xBuBzx  '),(7783,3000240,'25,March,2004','2004-03-25 00:00:00','   I\'d like to know who on Earth is talking behind my back or should I say spreading bullshit to \"Mr X\" (read back to Thursday March 18th to know who he is) at college, this whole thing is bugging me. If you know me and are reading this, tell me now and have some fucking dignity about yourself, and tell me your doing it and save yourself the embarrassment of me being angry enough to castrate you. You have 1 week . . . xBuBzx  '),(7784,3000240,'24,March,2004','2004-03-24 00:00:00','   Article:  Suicide    This is why I value every day and every second of breathing now.  I\'m so happy at college, I realise that. Happy as I am now, and as inappropriate as it is to post this on a public site, it shivers me to remember what my mental state was like a mere 8 months ago. Imagine what it is like to walk into a room and fall victim to the unrepentant and systematic torment that is the locusts of your class \"mates\"; and just maybe you could see a glimmer of what my life was like. You hasten and begin to think that the people you inhabit school are the archetype of humanity, forgetting the fresh breeze of multicultural, such is the rigor and stiffness of the regime and people, that is Halliford.  I was young in every sense of the word there. Maybe I was never naturally inclined to be a private school boy, or maybe I\'m just stubborn I\'m never quite sure, but I hated the 6 years I spent in the abyss there. I have my limits, and I\'m not going to ever tell too much information about my implosion into excrement within the dying embers of my time there, however if you would like a clue: I trusted people I shouldn\'t, and provoked flames in people who had been looking for the excuse against me for years. I can\'t say I\'m completely fault-ridden I\'ve done some and made some God awful catastrophe\'s, but even in my mind, what I went through was largely unjust. I\'ve learnt not to dwell on the past, but I\'ll never relieve myself of the bitter taste that pollutes my saliva whenever the thoughts of them come to mind.  I came very close to ending it all. Maybe none of you realise it and never will, but luck, my nephews envisaged crying face, and the bottom of my mental stability saved me. Standing on top of the block of flats near the precinct, thinking jump and realising that it can\'t get lower than being on top of here where so many other people have jumped from in the past.  I built a fortress and I became rather bedridden the preceding summer, but at least there was a genuine look at short term hope. I went to Esher and the rest is history. It\'s not worth saying any more, I just get resentful and bitter at everything. The whole thing taught me that you have to be happy through life and keep your head up whatever your going through there is always the horrible cliche of light at the end of the tunnel.  xBuBzx  '),(7785,3000240,'24,March,2004','2004-03-24 00:00:00','   There is a match for the ages as it would be said in pro-wrestling tonight as the 1st leg of the 2003/04 Champions league quarter finals is played between the 2 English clubs left in the competition - Chelsea and Arsenal at Stamford Bridge. This game is absolutely huge from my point of view, as, quite usually from my prospective, I\'m outspoken as I feel this will be the absolute making of Ranieri as Chelsea manager. If he, like I hope to God he will, lead us to our first victory over them since 99\' in the League Cup, then he will be well and truly be taking us up another level, however if we lose over the tie then I\'d rightly say his time is up. The guy has with the team during in the course of the season, despite the constant close-knit victories, only a handful of absolutely decent performances (i.e. Newcastle at home), the rest of the time we were lucky and at the right place at the right time. True, I believe whoever leads the team next season, we will improve on this one, but shouldn\'t a manager who can play attacking football and still better the results Ranieri\'s getting be in charge? His style, baffling tactics, lack of understanding what even his best team is, surely stifling the gelling of the team, and causing already reported public unrest within squad. I have thought however he is a decent man, with morals and dignity, especially throughout the constant barrage of pressure he\'s faced this season, and I do want him to do well, but oh well at the moment we can get any manager in the world who\'s mind would be swayed by money (which is practically all of them) so why stay with Ranieri, a pretty average-good manager, just for the fact of being just and that there\'s no absolute must-have reason to get rid of him, when we can get a world class manager i.e Arsene Wenger clone, which Ranieri I\'m afraid to say isn\'t. Roman Abramovich as I\'m sure your all well aware of didn\'t get to the top of the business world by not being absolutely ruthless and in the words of old Santa-Claus Ken Bates, he is an absolutely brilliant business man, so what chance does Ranieri have anyway?  Anyway, we have been beaten by them for years. I can remember stretching back to junior school that I used to come home on a Sunday from one occasion or another, that we had - inevitably - been defeated well by them once again. My theory (like the one\'s why Tottenham can\'t beat us nor we beating Blackburn at home) is that every team has a style that never well and truly changes over the years, and we are a lesser form of Arsenals, and always have. Any team who plays the same style or one less efficient than Arsenals 99/100 will be beaten, and Chelsea have never been the 1/100 via fate and God I guess. Arsenal are without a doubt, like AC Milan at the moment, awesome. If they had a decent Right back, which (like I think they will in the summer if they lure Hasem Trabelsi from our grasp when his contract runs out at Ajax) they\'d be perfect. The only way in which we will beat them will either be a 0-0 at the Bridge and a 1-1 at Highbury, which lets face it is highly unlikely considering that Arsenal will score tonight whatever defence we put out, they always do, and can shut out shop at Highbury, which they\'ve never really had to do too much, but they would defend there by attacking and inevitably score, which is by far the best form of defence; or by nullifying Viera and dominating the midfield, which lets face it is the hardest job in football. We\'ll have to be on top of our game no doubt about it, and they\'d have to have a rare off day. I live in pessimism and hope rather than expectation, and lets face it tonight, most Chelsea fans are after the 3 2-1 defeats to them already this season. Despite also Arsenal showing signs of fatigue, I think they\'re an intelligent enough team to know that once they recently annihilated Bolton et al recently, to let of the pedal to reserve energy for the 4 major ties that lay ahead of them, and beside they have a well-documented history of being able to raise their game and staying calm when the going gets tough.  Fear will hold us prisoner, hope will let Arsenal score. In my opinion it\'s a no win situation for Chelsea, but wouldn\'t it be sweet if . . . . . . . . . make sure you watch it tonight, whatever happens it will be one hell of a match.  xBuBzx  '),(7786,3000240,'23,March,2004','2004-03-23 00:00:00','   I\'ve been at home all day, chewing the fat mainly, pretty average by anyone\'s standards. I have a mountain to work to do but I can never really get round to doing it, takes too much thinking and my beloved blog is far more interesting. It\'s Chelsea v Arsenal in the quarter finals of the champions league, and we haven\'t got a snowballs chance in hell even over 2 legs, even I know that deep down, but come on Chelsea lets kick some gooner arse for once.  Also discovered I\'ve inspired the 1st person I know (Jon from work) to begin a blog, so it\'s like having a blog child I guess, or inspiring one, meh I\'m weird I know. Thought of me moving out are still as intense as ever, I still have to think about the whole situation a lot more though which is something I\'m gradually getting better at after years of mental-to-action effect, if you understand anything of what that meant . . .  xBuBzx  '),(7787,3000240,'22,March,2004','2004-03-22 00:00:00','   Somewhere Down The Barrel    SOMEWHERE FROM THE MORTAR  THEY SAVED YOU FROM FAR AWAY LABELS ETERNITY MEANS NOTHING BUT A SIGN ON A BRIDGE  WHEN EVERYONEâ€™S A STRANGER THEY CALL FOR YOU BY YOUR FIRST NAME AND LAUGH AT JOKES YOU ONCE MADE WHEN YOU WERE A KID  SOMEWHERE DOWN THE BARREL LIES A BULLET THAT I CANâ€™T KEEP AND IN THESE WATERS Iâ€™M WADING FOR A REASON AFTER ALL, ITâ€™S IN MY HEAD Iâ€™M NOT A SLAVE TO A DESPERATE LUST AND IN THESE WATERS Iâ€™M WAITING FOR A REASON AFTER ALLâ€¦  SCREAMING LIMITATIONS FALL SILENT ON NEW YORK CORNERS A TERRORISTâ€™S A PRISONER, AND A TOURIST A THIEF WHEN PAINTINGS SEEM LIKE BARGAINS BUT THEYâ€™RE NOTHING BUT WALLPAPER Iâ€™M PLAGUED BY SMALL TOWN FASCISTS LIKE A RASH ON MY SKIN  SOMEWHERE DOWN THE BARREL LIES A BULLET THAT I CANâ€™T KEEP AND Iâ€™VE BEEN SEARCHING FOR SOMETHING MORE THAN DISTANCE AFTER ALL, ITâ€™S IN MY HEAD Iâ€™M NOT A SLAVE TO AN UNBORN TRUST AND IN THESE WATERS Iâ€™M WAITING FOR A REASON AFTER ALLâ€¦.AFTER ALLâ€¦.AFTER ALLâ€¦ NAH NAH NAH NAH NAHâ€¦â€¦â€¦â€¦â€¦  xBuBzx  '),(7788,3000240,'22,March,2004','2004-03-22 00:00:00','   Well whether it be the over exposure to them, or this silly little traffic lights party, I\'ve had it, and now I\'m focused here on venting my anger on the most vigorous of diseases also known as townies. Well first I guess a definition - or better yet, let me link to you to a survey of definitions  urlLink http://www.urbandictionary.com/define.php?term=townies&f=1 . Well apart from the fact that my college is infested with the little fuckers, the sell-out craze as I like to call it (when townies in an area increases) is in full throttle there. I have a messed little theory that pre-townie, the devil (ie older teens who are already townies) tempt them just like Jesus was in the desert, and the majority of the time they give in (the girls especially think they look more attractive putting themselves across as good time girls rather than people with any sense of morals or values).   Anyway at college at the moment there is this party coming up, (which I think my friend Rachael is going to i\'m not sure) and as always the townies have made it an R\'n\'B infected event, with no sense of real or decent music involved I\'m afraid, but they seem to be ecstatic about the whole cherade, and are weeks into there preparations for it. Anyway, I\'ll discuss this issue tomorrow in more detail . . . xBuBzx  '),(7789,3000240,'21,March,2004','2004-03-21 00:00:00','   So the anti-humans (who I will refer to from here on in as my parents) returned form they\'re pilgrimage to see the oracle (who shall be known as my dear grandmother) all bitchy and with a fresh incentive to cause grief upon myself and strangely but characteristically not to others. Nehoo, I went to Roehampton university for there open day, and I was quite pleasantly surprised at how big and unlike college, clean the whole area. I mean I\'m quite surprised that in greater London there\'s actually a large a space that there is there, I mean there\'s parks and ducks and everything to awaken the senses for when you become bored with the pressure of university work.   I went to this odd party in Molesey on Saturday night, and being in Molesey there were quite a few people inevitably from work present (work being Tesco\'s for people who don\'t know) and I\'m yet to make up my mind whether it was good or bad. I guess when I heard my friend say \"if your initial reaction to a party isn\'t good, then it never was\", he was most probably right, and anyway the guys whos party it was was an adultaring up his own arsehole anyway, so that just about has tipped the balance into the bad side of the court.  I hope to get tickets for Chelsea next week, just heard some more have been made available, but it seems doubtful, but hopefully it will all come of anyway I\'ll post later. xBuBzx  '),(7790,3000240,'03,April,2004','2004-04-03 00:00:00','   Oh great, let me jump for joy its my day in two days.  I think something must be horribly wrong.  I really can\'t stand the whole thing this year, it all seems pretty pointless, and I guess it\'s going to be this way until I\'m 80 and in a wheelchair whilst valuing every day I continue to breath.  The individual who had been backstabbing me to \"Mr X\" never owned up, so I must take it that they were rather cowardly and quite falsely angry at me for some reason. So whoever you are you anully retentive fool, I hope you die slowly in your own filth as the devil castrate\'s you.  The silent treatment I\'m handing out to my Mother is working a charm, as, after only three days she has cracked already. Little does she know of how long I can extend this epic feud with her until she eventually breaks well and truly; more so than she\'s ever known. I know it may seem harsh of me for doing such a thing against my own mother, however I feel quite betrayed by what she\'s done, which Is a bit too personal to post on the internet even for me, but believe me when I say you\'d feel just as angry if it had happened to yourself. So I came to the conclusion after such a brilliantly monumental start, that I might use these mind games to its full effect for a little while longer - my next plan being to refuse to see her at all during my birthday, which I know to most of you doesn\'t seem like a harsh thing, but considering how much I know she values spending the day with me, and the fact that there\'s no college, this opportunity seems priceless. Oh how I love being cold . . . xBuBzx  '),(7791,3000240,'01,April,2004','2004-04-01 00:00:00','   Why is it that when I talk to my brother that I should feel like I\'m talking to a complete stranger (and I don\'t mean in the metaphorical sense). I mean I\'ve know the man for 17 years on Monday, and it really worries me to say I feel I know him, and I would feel comfortable talking to him as I would if a random hobo walked up to me on the street. I\'m not saying he\'s a horrible person, he\'s a nice guy and an awesome father, but we never really seemed to click as people. Oh well clash of personalities I guess.  Nehoo I went to a looong philosophy lecture today at Goldalming college, and it felt like hell honestly, all apart from the Aprils fools joke that happened at the beginning.   I\'m knackered though so I\'ll see you later xBuBzx  '),(7792,3000240,'31,March,2004','2004-03-31 00:00:00','   I\'ve had enough from Philosophy, my Christian friends and strangely the other day Daniel Kesselaar and so I feel I have to clear a subject up in my mind - religion.  I believe when you die you die. Simple as that. People will say \"but you have to believe in something otherwise life isn\'t really worth living\" but it is what I believe because I feel you can\'t just believe in anything purely on blind faith. Like for say in a court of law you have to prove an incident happened within all reasonable doubt, and that in my opinion is exactly what you should apply for religion as well.  There hasn\'t (beside a few debatable cult religious books which attracted many followers, due to the fairy tale and \"overly accurate\" story lines of miracles and coincidences which according to them couldn\'t have happened without God\'s \"divine intervention\" e.g. the Koran and bible) ever been any conclusive evidence for God\'s existence, despite many fair and well put arguments i.e. ontological and teleological.  I don\'t think I\'ve got the time to go into the suffering issue, however you must admit it\'s almost a solid and firm argument against the Deity (God).  So I might as well leave you with a question - if we found intelligent life on another planet (which is probably likely one day considering how absolutely huge the Universe is) - what would you religious believers think then?  xBuBzx  '),(7793,3000240,'30,March,2004','2004-03-30 00:00:00','   I swear every little situation that is of the smallest nerve I imagine, when I\'m doing it, that I\'m writing the whole situation in a novel or something similar. Odd. Very odd. Oh well I guess my mind is just expanding its horizons which is good considering writing is what I want to do in my life. I mean why would I write this blog everyday if it wasn\'t for the fact that I like writing quite a bit? xBuBzx  '),(7794,3000240,'30,March,2004','2004-03-30 00:00:00','   Wow they over react so easily. I over slept again as I always seem to do on a Tuesday and she came home to me in my PJ\'s on the couch and she freaked. We then had a massive war with her using mind games and me throwing grapefruit, and now in the end I\'ll be the one in the dog house probably by Dad so if I don\'t happen to post on here for the next few days, it\'s because I\'ve been banned form the computer et al. Mum does this all the time, using her \"trump card\" as my sister once called it to invoke sympathy. It was her and Dad which made me decide about quitting education two weeks ago, but I think now I might be going to turn on that decision. It would all mean I\'d get to move out quicker and move in with Ian, which would be pretty cool considering what I live with in this house. Oh well *sigh* xBuBzx  '),(7795,3000240,'29,March,2004','2004-03-29 00:00:00','   I\'m counting down not just the seven days left until my birthday, but also the five short one\'s until the two week Easter break, which will mean an enormous amount of overtime at work which will hopefully help to sooth my already stretched bank balance. Also it will (although unlike what my Dad thinks, if I take the overtime at work) the opportunity to work quite a bit towards my exams, which are only a few weeks away. Judging from my profiling last week, I need to work my ass of to get the c\'s I really need. Oh well, I must be getting on with this work with some Psychology, so ciao for now. xBuBzx  '),(7796,3000240,'28,March,2004','2004-03-28 00:00:00','   Friday was one gooood looong stress reliever of a night at the electric ballroom in Camden at \"sin city\" with Anna, Jade and there respective lovers who seem like cool people. Went funkily although it seemed to last forever in there. Anyway when we left it was about 4 in the morning and we all drifted around Camden (not as dangerous as I had thought it would have been) until daylight and the first buses came. Anyway got to Hammersmith about 7 and then got home via the tubes and whats not. Once I slept it all of I called Anna about the obscure film festival we were planning to go to on the Saturday night, but she never turned up, and that\'s when my bad luck started. Well it\'s a lot easier for me to list it so her you go:  1. Couldn\'t get change a 10 pound note to call Anna from anyone 2. When I eventually did the phone booth didn\'t work 3. Train delayed back to Waterloo 4. When it started I learnt I was on the wrong train 5. Then cunningly (I thought) I\'d get of at bank and get on the waterloo and city line but as soon as I got to the platform I learnt it was shut. 6. Got to Kennignton but missed train just and had to wait 3 quarters of an hour for the next one 7. Got to Waterloo, and called Anna but she didn\'t pick up 8. Had to wait for 3 quarters of an hour for the train to Kingston 9. When I eventually got to Kingston I had to wait for an hour for the 411 10. Eventually got home to a bollocking from my parents.  Hooray gave me new hope for suicide.  xBuBzx  '),(7797,3000240,'08,April,2004','2004-04-08 00:00:00','   I\'ve found to my discomfort that the only thing in life I crave I food. Forget drugs, alcohol and tobacco, food is the curse of my life. It\'s going to surely get exceedingly worse as the Easter holiday approaches, as the capitalist advertising temptation also known as Easter comes into full throttle. God knows how I\'m going to survive this Friday, Sunday and Monday at work with the mountain of Easter eggs at my disposal *sighs*. I guess my will power will have to be on full alert and my eyesight somewhat temporarily afflicted with some tropical infection to survive the war against the flab - but I\'m determined to do it.  Oh and by the way . . .  thanks Jen for the birthday mention, made me feel all warm inside, even if the hot chocolate at the time didn\'t.  xBuBzx  '),(7798,3000240,'07,April,2004','2004-04-07 00:00:00','   Well I\'ve spent the last few days in Mitcham at my sisters flat, trying to get away from it all for a few days holiday I guess, so I\'m sorry I haven\'t posted much recently, however I promise I\'ll change that in the indefinite future.  So where to begin - first of it wasn\'t such a bad birthday after all I saw the Christ movie and I have to say for such an overtold story, such as the crucifixion, it was done wonderfully. The fact they used subtitles, only added to the reality of the whole situation, and every character (apart from unfortunately Jesus I hasten to say) was cast to perfection. Maybe Mel Gibson could have tried it although I doubt the results would have been greatly different. I saw the movie with Jo and her mate Wendy who is the funniest large lady in the world - bar none. The morning was spent with the continuation of the minimal dialogue epic between me and my mother, with Harry there to ease the tension somewhat. The presents were pretty average for my birthday, not that I\'m complaining mind you, I\'m always pretty contempt with any present anyone ever gives me.  Anyway the day after was most definitely, more than previous days I\'d used the example, bitter-sweet. My sister woke up and I was left there watching Harry Potter for a couple of hours. The as a delivery came, I forgot to putt the lock on the latch and it slammed shut behind me and I was left there with only my baggy PJ\'s on for a couple of hours. The once I managed to gather the strength and knocked on some one\'s door to use the phone. I was greeted by a lovely large Caribbean lady (whose name sadly evades me) who gave me tea and somewhere to warm my frost bitten feet. Once Mum had come with the keys I hyped myself up for the football. I take back that big speech I made just a couple of weeks ago about how we couldn\'t beat Arsenal, well, WE DID IT, WE BEAT ARSENAL DAMMIT, and it was well and truly brilliant in every sense of the word. Now we face Monaco in the semis, after unbelievably (well not really considering what there defense is like) they beat Real Madrid, and I really now can see a Chelsea v Milan final, which we would have an outside, but decent chance of winning *dreams*. Well I\'ll see you lot later then . . . xBuBzx  '),(7799,3000240,'15,April,2004','2004-04-15 00:00:00','   Tender is the night Lying by your side Tender is the touch Of someone that you love too much Tender is the day The demons go away Lord I need to find Someone who can heal my mind  Come on, come on, come on Get through it Come on, come on, come on Loveâ€™s the greatest thing Come on, come on, come on Get through it Come on, come on, come on Loveâ€™s the greatest thing That we have Iâ€™m waiting for that feeling Iâ€™m waiting for that feeling Waiting for that feeling to come  Oh my baby Oh my baby Oh why Oh my X2  Tender is the ghost The ghost I love the most Hiding from the sun Waiting for the night to come Tender is my heart For screwing up my life Lord I need to find Someone who can heal my mind  Come on, come on, come on Get through it Come on, come on, come on Loveâ€™s the greatest thing Come on, come on, come on Get through it Come on, come on, come on Loveâ€™s the greatest thing That we have Iâ€™m waiting for that feeling Iâ€™m waiting for that feeling Waiting for that feeling to come  Oh my baby Oh my baby Oh why Oh my X2  Come on, come on, come on Get through it Come on, come on, come on Loveâ€™s the greatest thing Come on, come on, come on Get through it Come on, come on, come on Loveâ€™s the greatest thing That we have Iâ€™m waiting for that feeling Iâ€™m waiting for that feeling Waiting for that feeling to come  Oh my baby Oh my baby Oh why Oh my  Tender is the night Lying by your side Tender is the touch Of someone that you love too much Tender is my heart you know For screwing up my life Oh lord I need to find Someone who can heal my mind  Come on, come on, come on Get through it Come on, come on, come on Loveâ€™s the greatest thing Come on, come on, come on Get through it Come on, come on, come on Loveâ€™s the greatest thing That we have Iâ€™m waiting for that feeling Iâ€™m waiting for that feeling Waiting for that feeling to come  Oh my baby Oh my baby Oh why Oh my    xBuBzx  '),(7800,3000240,'14,April,2004','2004-04-14 00:00:00','   *sigh* why can\'t I  ever  revise . . . . xBuBzx  '),(7801,3000240,'13,April,2004','2004-04-13 00:00:00','   Q. What do you call a townie in a box?  A. Innit.   Q. What do you call a townie in a filing cabinet? A. Sorted.   Q. How can you tell a female townie virgin? A. She can run faster than her brothers.   Q. What do you call a townie on a bike? A. A thief   Q. What do you call a townie in a car? A. Arrested   Q. What do you call a townie waiting in a bus shelter? A. At a party.  Q. What do u call a townie alive? A. Fucking lucky!  Q. What do you call a 30 year old townie?  A. Failed.  Q. What do you call a 12 year old townie girl? A. Pregnant.  Q. What do you call a townie girl without any children? A. Under the age of 5.  Q. What do you call a townie in a skatepark? A. Lost.  Q. Why did the townie cross the road? A. To punch someone for absolutely no reason whatsoever.  Q. What\'s a townies favorite car? A. One without an alarm.  Q. Why did the Townie stare at the carton of Orange Juice? A. Because it had \'concentrate\' on it.  Q. What do you say to a townie in a suit? A. Will the defendant please stand.  Q. What did the little townie say to the bigger townie? A. Can you get served?  Q. Why do townies constantly rev their engines? A. So they don\'t cut out.  Q. What do you call a townie in a jar of honey? A. Sweet!  Q. What do you call a townie in a coffin? A. A damn good reason to kill another.  Q. What do u say to a boy racer when he is circling? A. Are you lost?  Q. Why did the townie shag the chicken? A. He couldn\'t find any other breast meat.  Q. What do you call two dead townies? A. A good start to the day.  Q. What do townies use as protection during sex? A. A bus shelter.  Q. What do you say to a townie with a job? A. Can I have a Big Mac please?  Q. What is a townies favorite ice cream? A. Mint!  Q. What\'s the difference between a dying townie and an onion? A. Onions make you cry.  Q. What have townie girls got in common with turtles? A. When they\'re on their back they\'re fucked.  Q. What happens to a thought in a townies head? A. It dies of loneliness.  Q. How does a townie girl turn the lights off after sex? A. She closes the car door.  Q. How many townies does it take to change a light bulb? A. 5. 1 to put it in, the other 4 to tell him \"innit, innit, innit, innit\".  Q. What do you call a 30 year old townie? A. Dunno. Ask her 17 year old son.  Q. What do you call a townie in hell? A. Wicked.  Q. What do you call a townie in a fridge? A. Chillin\'.  Q. What do you do if you run a townie over? A. Slip it into reverse just to make sure.  Q. What do you call a townie in an iron box? A. Saphe.  Q. What do you do if you shoot a townie? A. Reload.  Q. What do a war veteran and a used townie condom have in common? A. They both live to fight another day.  Q. What do you call a townie girl and a townie boy in a phone box? A. Fucking innit.  Q. Why did that townie go to prison? A. He got caught stealing pens to apply for his dole cheques with  oh and townies just so you know spitting, smoking and getting 13 years olds pregnant does not make YOU LOOK HARD.  xBuBzx  '),(7802,3000240,'12,April,2004','2004-04-12 00:00:00','   My parents randomly left me in the house all of a sudden on Saturday, and a note left on the kitchen table said: Gone to Wales, be back Tuesday ciao. So it\'s felt really cold (yes despite the heater being on) and lonely in my little semi-detached recently and the odd occasion I have ventured outside of the local area, has been a complete and utter disaster, so all I can do is eat, which is destroying my supposed diet *sigh* xBuBzx  '),(7803,3000240,'12,April,2004','2004-04-12 00:00:00','   There\'s nothing to do and Easter break has been shitty *sigh* xBuBzx  '),(7804,3000240,'23,April,2004','2004-04-23 00:00:00','   Lyrics to \"thinking in reverse\" by the dissociatives    NOW THAT EVERYTHING\'S IN PLACE, YOU\'RE LINING UP THE PINS AND NEEDLES DOWN THE LONGEST TRAIL,  YOUÃ‚Â’VE GOT THE ANSWER BUT IT LASTS  AS LONG AS YOU CAN SMILE BUT PRETTY SOON YOUR CHEEKS WILL HURT  I\'M NOT TRYING TO MAKE THINGS WORSE  BUT YOU\'RE THINKING IN REVERSE NOW THAT EVERYTHING IS NEVER TOO MUCHA  TAKE THE STANDARD LINE AND TURN  IT BACK UPON ITSELF TO SEE WHAT HURT IS ONCE YOU\'VE FELT COZ WE\'VE BEEN NUMB LIKE CORRIDORS  AND DRY LIKE TUMBLEWEED,  THIS CHURCHÃ‚  AND I GOT IT ALRIGHT.. TO SEE WHAT I CAN SEE IS TAKING OVER ME.  xBuBzx  '),(7805,3000240,'22,April,2004','2004-04-22 00:00:00','   Escape . . .    It\'s astounding how much of our world is based upon avoiding our problems.  You would think that in our fast-paced and efficiency-driven society, solving problems and achieving goals quickly and efficiently would be one of our top priorities.  On the contrary, it seems that following through with our thoughts to their final completion has a very low priority for most modern people.  This difficulty appears in the workplace, where there are so many layers of management that any problem simply exists forever in the pool of unfinished tasks, shunted from executive to executive, until eventually it is a moot point. But primarily it appears in people\'s day-to-day lives, where they avoid dealing with unfinished thoughts, unpleasant ideas, and unwanted consequences.   As I look inside my mind, I\'m unsure what to call what I see.  In a childish way I wish to give it a childish label:  \"A Swirling Vortex of Doom!!!\"  However juvenile the term may be, I believe it has some merit.  My mind is definitely a swirling vortex.  Unfinished thoughts, ideas, and emotions tumble on without end.  Nothing is ever resolved.  In our modern and busy life, you never have time to actually resolve an issue: everything is given a Band-Aid-style fix, and then you move on to the next item.  Unfortunately, every unfinished thought stays in your poor mind, polluting your consciousness and buzzing with unfulfilled desires.  It might be easy to say you can go about your everyday life, not really needing to have a truly clean and fresh mind.  You might want to leave such luxuries to philosophers or other thinkers, since you believe your life has not been adversely affected by these stray thoughts.  However, I know a good many people who suffer from some of the symptoms of having too many unfinished thoughts floating around their mind.  An inability to fall, or stay, asleep is a prominent one.  I\'ve heard many people complaining on buses, subways, or restaurants, about having been unable to sleep last night because they couldn\'t clear their mind, or being woken up in the middle of the night and being unable to get back to sleep for one or two hours, due to some persistent thought they couldn\'t dismiss.  One major cause of this?  Your mind is trying to resolve these contentious and polluting issues, while you refuse to give it the chance to do so.  It is unable to achieve the necessary level of clarity required for sleep.  A lack of focus in the workplace is another common symptom of having an occupied mind.  How many times have you been sitting at work, trying to focus on your daily tasks, when your mind keeps wandering off towards other concerns?  It is extremely difficult to dedicate your undivided attention to a topic when your subconscious mind wants to give your attention to other thoughts.  However, most frequently, you simply try to brush aside those other subconscious thoughts for later consideration, focusing solely on your work-related tasks.  You have deemed clearing your mind of unwanted ideas to be unworthy of your time, even though dealing with those extraneous thoughts before turning to the task at hand will lead to greater productivity and a more efficient use of your time.  Modern society seems to encourage leaving thoughts unresolved.  People, on some subconscious level, know they need to have clear minds.  They know in order to truly relax, they need to be working with an empty slate.  However, most people don\'t know how to achieve this state.  So, they turn to the answers society provides: drugs, alcohol, sex, vacations.  These classic things are the goals of most men\'s recreation.  But you look at these goals, and what they do for your mind.  Do they help you deal with the issues at hand?  No.  Do they help permanently push them out of your mind?  Not a chance.  What they do is temporarily alleviate the stresses on your soul.  They help you forget about the woes of everyday life; they donâ€™t actually deal with them, but merely allow you to put them out of your mind for a very short time.  There are more permanent and productive alternatives out there.  Meditation, talking out your problems with friends, or analytically dealing with these issues, are just a few of the simpler and more available ones.  And upper management is slowly latching onto these ideas, with therapists and psychologists abounding to listen to your problems, many mini-retreats with courses on how to solve problems and achieve your goals, and ever popular self-help books offering to teach you how to center yourself and calm your mind.  But there are still millions of middlemen out there who haven\'t worked a problem through to satisfactory completion in many years.  The only time they ever achieve silence of the mind is while on a beach in the Caribbean or drunk out of their minds.  I think people need to deal with their problems.  It\'s a hard process to begin, but once begun, the clarity you\'ve achieved will help you deal with more and more problems, until eventually you\'re walking through everyday life with the same relaxed mentality as most people only achieve in a drunken stupor.  People need to focus their lives less on drink and on vacation and focus their lives more on personal development of the mind, and dealing with their own internal issues.  This may not be the key to eternal happiness, but it may be a helpful and simple step towards that elusive goal. xBuBzx  '),(7806,3000240,'21,April,2004','2004-04-21 00:00:00','   Meh went to a gig last night with Jen and had a good time of it, if well not fantastical. The bands were generally of a good quality (well maybe not the 1st one) and got another 1 booked for 5th of may at half moon to see that arrogant prick Rowan - *sigh* xBuBzx  '),(7807,3000240,'21,April,2004','2004-04-21 00:00:00','   Iraq - If you happy and you know it parody    If you cannot find Osama, bomb Iraq.  If the markets are a drama, bomb Iraq.  If the terrorists are frisky,  Pakistan is looking shifty,  North Korea is too risky,  Bomb Iraq.    If we have no allies with us, bomb Iraq.  If we think that someone\'s dissed us, bomb Iraq.  So to hell with the inspections,  Let\'s look tough for the elections,  Close your mind and take directions,  Bomb Iraq.   It\'s pre-emptive non-aggression, bomb Iraq.  To prevent this mass destruction, bomb Iraq.  They\'ve got weapons we can\'t see,  And that\'s all the proof we need,  If they\'re not there, they must be there,  Bomb Iraq.    If you never were elected, bomb Iraq.  If your mood is quite dejected, bomb Iraq.  If you think Saddam\'s gone mad,  With the weapons that he had,  And he tried to kill your dad,  Bomb Iraq.   If corporate fraud is growin\', bomb Iraq.  If your ties to it are showin\', bomb Iraq.  If your politics are sleazy,  And hiding that ain\'t easy,  And your manhood\'s getting queasy,  Bomb Iraq.    Fall in line and follow orders, bomb Iraq.  For our might knows not our borders, bomb Iraq.  Disagree? We\'ll call it treason,  Let\'s make war not love this season,  Even if we have no reason,  Bomb Iraq.  xBuBzx  '),(7808,3000240,'21,April,2004','2004-04-21 00:00:00','   Looking threw my blog, and thinking about anything in particular I\'ve never discussed on here, and it came to me that I\'ve never really talked about the whole love epic in great detail. So this is my theory in my current thinking climate (which let it be known changes on a daily basis).  I see many really quite attractive people at college. There\'s only a few I\'d safely say that I fancy as such. Take one person for example, they\'re not pwahhhh and they don\'t like to flaunt what they don\'t genuinely think they have, which is a refreshing feature for me to find attraction in, as it safely delves away from what I\'d usually be attracted to. The person too is also quiet and anyone could tell, if they met them, that they sway away from attention full stop. They are really \"pretty\" or \"cute\" and go around they\'re daily ramblings with a certain elegance and style mainly associated with the majority of people the same age as them. This is something I have always really quite respected in a person.  Shame my end of the bargain will never be nearly up to scratch. Shame. It\'s not like they\'re someone I can\'t stop talking to about absolutely everything, and they don\'t particularly blow my mind as such, however it would have be nice to have something stable, or just something for that matter. It\'s been so long since, if ever, that I\'ve had an exclusive relationship with anyone. Sad I know, but esteem, confidence and attraction et al is never anything or will never be something that I\'ll have in abundance, so I\'ll just have to lie in my cave, contempt with my lot, until I drag myself to reach the phone book looking for a decent counselor to \"re-enlighten me\" and give me a \"re-birth\" in life.  There must be a million people out there just like me (my sister most definitely included there . . . . . . HA), waiting for just something at least, and I guess we find at least someone who is the great love of your existence, but God its a long time waiting, a very long time indeed. xBuBzx  '),(7809,3000240,'20,April,2004','2004-04-20 00:00:00','   What in hell were Chelsea doing last night. We were comfortable,  very  comfortable at 1-1 and we had to throw it away. I heard on one of the fan sites that we got \"caught out\" and I think this is true in many senses. 3-1 isn\'t the worst result in the world, but God dammit, we have play out of our Caracas\' to win it in the second leg. It can be done, it\'s possible most certainly, and still we are only 90 minutes away from a date in Germany. I pray to God that we get there, and lets have hope instead of cliche pessimism. xBuBzx  '),(7810,3000240,'20,April,2004','2004-04-20 00:00:00','   Good God the CHAMPIONS LEAGUE SEMI -FINALS!!! So how on Earth (and no disrespect to Chelsea but . . .) has the European Cup become so easy to win. Oh well who fucking cares. Well we play Monaco tonight which can only be described as huge. I feel that this match will be significant and ponders a couple of questions for a couple of reasons: isn\'t this just a little above the station for Chelsea, but we will well and truly never have a better chance of lifting the trophy than we will do in foreseeable evolutionary future. It\'s a shame that Gallas and Duff are out otherwise I think Ambrosio apart we would have has the strongest 11 available. I think we will be able to do it, but our defense needs to be on top of its game to stop Europe\'s most prolific goal scoring machine of the year. If we sneak a 1-1 or a 0-1 or even a 2-1 to Monaco I\'ll be happy, just as long as we score . . . May God be with us tonight. xBuBzx  '),(7811,3000240,'19,April,2004','2004-04-19 00:00:00','   So I\'ve finally dragged myself back to the  shit-hole  of a place that, I love to digress to your good selves about the mishaps and novelties about life - yes you guessed it in one . . . . the LRC at college. Well I fully understand that I failed to posted anything remotely worthwhile during Easter and this was for a number of reasons:  1. I was ill 2. I couldn\'t be bothered 3. Chewing the fat seemed a more than viable alternative 4. There was a wart on my shin which needed annihilating with a hefty pair of garden shears.    Anyhoo   I did get the sense of incredible loneliness and so I felt it was justly time for an ounce of soul searching and a slow but solid ponder about the view you get when you stand on the edge of cliffs or high storey flats and how it would truly feel to kick the bucket.  OOOOOOOOOh also I think I\'ve really opened up about something, which sadly I won\'t go into, but has helped enormously, and I cried, REALLY cried away about everything, probably because it was a more genuine emotion than I usually have over things ( If you understand me on that one, your a better man than I, I apologise for its waffling ). However the whole thing has kind of helped me sort a few lingering problems and stamp some labels on unclear things (or maybe get me nearer finally doing them), and now I\'m sure that I\'ll have a sudden new zest for things and I\'ll hopefully be able to take on things more full on, without the usual side-tracking I feel I have to do. So I promise a lot more wonderful pointless posting which I can well and truly excel at. Yay xBuBzx  '),(7812,3000240,'18,April,2004','2004-04-18 00:00:00','   Dammit I hate College, why do I have to go back tomorrow . . . xBuBzx  '),(7813,3000240,'30,April,2004','2004-04-30 00:00:00','   Here\'s part one of a net biography of my favorite band Nine Inch Nails:    Swimming In The Haze: Part One - Nine Inch Nails And The Pretty Hate Machine Years    Trent Reznor was born Michael Trent Reznor on May 17, 1965, in the sleepy rural town of New Castle, Pennsylvania.  Reznor\'s sister, Tera, was born in 1971, and was to be the last child of Mike and Nancy Reznor.  Reznor\'s parents soon split up.  Tera went to live with her mother, while Trent was shipped off to be raised by his maternal grandparents in the tiny hamlet of Mercer, Pennsylvania.  By Reznor\'s own description, Mercer was \"a nice little picturesque one-horse, one McDonald\'s kind of town.\"  By all accounts, Trent Reznor was the typical American youngster raised in the typical American fashion, participating in Boy Scouts, building model airplanes, and skate boarding.  His grandparents would also later put him through the typical childhood nightmare of piano lessons.  Reznor\'s piano teacher, Rita Beglin, has allegedly compared his playing to the likes of Harry Connick, Jr., though this \'fact\' was difficult to verify over the Internet.  In the later high school years, Reznor has been described as being clean cut, handsome, and popular.  Reznor was also in the high school jazz and marching bands, where he played both tenor sax and keyboards.  He was also voted \'Best in Drama\' by his peers and classmates.  Actually, Reznor was the absolute modicum of popularity he would later rebel so violently against in the not-so-distant future.  After graduation, Trent Reznor moved to Meadville, Pennsylvania, where he attended Allegheny College and majored in Computer Engineering.  After one year of spinning his wheels, Reznor soon realized that programming computers for the rest of his life wasn\'t his calling.  So, Reznor unknowingly engaged in the tired cliche of dropping out to pursue a career in music, and relocated to Cleveland, Ohio, to chase the rock and roll dream.  During the early and mid-1980s, Trent Reznor was involved in several small-time, Flock of Seagulls-type bands, including The Exotic Birds (where he would meet his future drummer, Chris Vrenna) and 1000 Homo DJs, which involved Al Jourgensen of future Ministry fame.  Said bands were a continual disappointment to Reznor, however, and he found himself leaving these acts more often than not, his as yet unknown creative outlet left unfulfilled.  During these turbulent times, young Reznor found work programming synthesizers and MIDI in a small Cleveland studio, and it was here that Nine Inch Nails first began to take shape.  A twenty-something Reznor toiled by day at the controls of the studio, but was given free reign at night by the studio owner.  Reznor slowly taught himself the complicated MIDI applications that would allow him to sample, mix, and develop various sounds.  Being free of the typical collaborating (see \'meddling\') band mates, Reznor took his time and immersed himself utterly in his work, preferring the isolation.  He was free to express his feelings through an electronic format that was soothing and grating, soft and hard, quiet and raucous.  Being only an amateur musician at the time, Trent Reznor approached his work from the listeners point of view, and developed a work that allowed the listener to hear someone else express their innermost emotions and pain, but in an entirely unique and unusual way.  The result was a collection of raw, turbulent and brutally honest tracks.  The origin of the name Nine Inch Nails has been widely speculated over the years.  Some argue it refers to the length of a coffin nail, while more imaginative souls liken it to the size of Trent Reznor\'s phallus.  Rather boringly, however, it has no profound meaning.  After tossing around 200 monikers for several weeks, Reznor found that Nine Inch Nails kept ringing in his head.  He has also quipped that it looked great in print, was easy to abbreviate, and \"it\'s a lot cooler name than Trent Reznor.\"  \"It\'s a curse trying to come up with band names,\" says Reznor.  It was about this time that Trent Reznor began to shop his demo out to some smaller independent labels.  The 80\'s were coming to a close, and the music industry was undergoing a subtle shift in trends.  In the summer of 1988, Reznor cautiously sent out about ten, three-track demos to small, independent labels, simply looking for a single 12-inch deal so that he might hone his craft.  He was not yet fully aware of what Nine Inch Nails was, or what direction it was going to take, and he was reluctant to get involved with a label that would want to smooth out his sound into a polished, more palatable brand of pop.  So it was with surprise and some trepidation no doubt, when Reznor received offers from virtually every one of the labels he had approached.  It was at this point that Reznor realized he was on to something.  Trent Reznor settled on TVT records, a record label more commonly known to market albums consisting of TV jingles rather than avant-garde, pseudo-Industrial music.  Nevertheless, Reznor immersed himself in the deal, working with a number of producers on what would be Nine Inch Nails\' debut album.  One notable producer included Adrian Sherwood (Depeche Mode, Ministry, KMFDM), who mixed the lead single \"Down in It\" for the as-yet unreleased Pretty Hate Machine.  Sherwood and Reznor communicated solely over the phone and Sherwood actually mixed the track in London.  The two would never actually meet face to face.  It was this mix that would find its way onto Pretty Hate Machine, as opposed to Reznor\'s original version, despite Reznor\'s protests.  While Reznor appreciated Sherwood\'s mix, it was radically different from his own version, and he protested in vain with TVT to have it kept off the album, but the label would not relent.  This was perhaps the first sign of sour times ahead for TVT and Nine Inch Nails.  Trent Reznor also worked with the legendary Flood, whom he had originally wanted to produce Pretty Hate Machine in its entirety, but due to scheduling constraints and prior commitments, this had become impossible.  \"I liked Flood because he\'s the opposite of Sherwood. He\'s very transparent. You don\'t say, \'Hey, that sounds like Flood,\'\" says Reznor of the legendary producer.  Flood, of Depeche Mode and U2\'s Achtung Baby fame, was nevertheless able to contribute in a small way.  Several other producers were also added to a growing list, including John Fryer (Love & Rockets, Cocteau Twins) and Keith LeBlanc (Tackhead, Maffia).  It was time that Nine Inch Nails took their act on the road.  Trent Reznor was faced with an obstacle, however.  Because he had recorded Pretty Hate Machine entirely alone (\"Nine Inch Nails is Trent Reznor\" the album jacket flatly proclaims), Reznor had to assemble a cast of characters to bring Nine Inch Nails\' music alive on stage.  So, Reznor gathered some musicians about him, and ambitiously played a few dates opening for the legendary Skinny Puppy.  It was an intimidating time by Reznor\'s own account.  He felt they weren\'t nearly as intense live as he wanted them, so after the small tour, he ushered his live band out the door, and collected a new crew around him.  He intentionally looked for young, malleable musicians who would be less likely to voice their artistic opinion and more likely to bow to young Reznor\'s direction.  And so, Reznor was joined by Chris Vrenna on drums, Richard Patrick (whose brother is Robert Patrick, who played the T-1000 on Terminator 2: Judgement Day) on guitar, and David Haymes on keyboards.  In the spring of 1989, Nine Inch Nails embarked on a tour in support of Pretty Hate Machine, opening for the Jesus and Mary Chain, and quickly followed it up with a spring tour opening for Peter Murphy.  While NIN were moderately well received for fans of Jesus and Mary Chain, Murphy fans were not so forgiving.  Playing small theatres and college auditoriums, the steroid-pumping jocks and Bauhaus-fanatics did not take well to NIN and their angst-ridden set.  It was not until an Atlanta show when Reznor pelted the audience with cold pizza before they actually received some semblance of acceptance.  After throwing corn starch on the audience, smashing a guitar, toppling the drum set, and storming offstage, Reznor was amazed to find the gathered Georgia crowd loved them.  \"Forget the music, as soon as they got abused, that was it.  We outsold Peter Murphy on T-shirts,\" Reznor has said about the show.  Nine Inch Nails would also go on to release their first music video at about this time, the lead track \"Down in It.\"  Produced by H-Gun, the Chicago group who were also responsible for Al Jourgensen-related projects like Ministry and the Revolting Cocks, the video did not initially pass MTV standards and was edited.  Towards the end of the video, Trent Reznor is simply lying on the ground, covered in corn starch, supposedly dead.  One does not see how Reznor \'dies,\' nor are there any gruesome details indicating how he might have died, but it was still too much for MTV.  \"That implies suicide, and we can\'t have that on MTV,\" Reznor has said sarcastically about the cut-job on the video, \"but we can have Cher\'s naked ass.\"  With the breaking success of Pretty Hate Machine, supported by the albeit edited video rotation on MTV, as well as the underground success of their second single and video, \"Head Like A Hole,\" Nine Inch Nails were beginning to gain steam and gain some notoriety.  So much so, in fact, that NIN were approached by Perry Farrell of Jane\'s Addition, who was putting together a very eclectic cast of characters for a festival tour labeled Lollapalooza.  Lollapalooza was the brainchild of Farrell, and consisted of a touring package of a wide variety of musical styles, while at the same time offering a venue and voice for political debate.  And so it was that Nine Inch Nails, now joined by James Woolley on keyboards, were squeezed into the middle of a seven-band bill, immediately followed by Ice-T and preceding Living Color.  Nine Inch Nails proved to be a wildcard on the tour, and touring with Lollapalooza undoubtedly set the main foundation for Nine Inch Nails\' huge success.  Fans who had never heard of NIN became immediately enraptured when a corn-starch coated Trent Reznor took the stage, smashing guitars, keyboards, and band mates alike.  Rather than banking the pay checks that they were receiving for their performances, Nine Inch Nails instead put aside only a small kitty and invested the rest into ten guitars per performance.  Ten guitars scheduled strictly for destruction.  And so, each night when Nine Inch Nails took the stage before an audience of people who had never heard of them, Reznor would do his best to destroy each and everyone said instruments, whether it be by smashing them over the stage or his own band mates.  James Woolley found this out the hard way at the performance in San Francisco, California, when a piece of keyboard, destroyed by Reznor, flew up and hit him in the head, drawing a torrent of blood.  Oblivious, Reznor played/thrashed on, until a stage hand alerted him to the fact that medics were on the way to tend to the injured Woolley.  Guitarist Richard Patrick suffered a similar fate in Boston, Massachusetts, where Reznor tackled him over some monitors, leaving Patrick writhing in agony.  This was just after Reznor had soaked the amps, cables, and monitors in water, of course.  Fans at Lollapalooza had no idea what they were witnessing, but they damn well knew an exciting act when they saw it.  Nine Inch Nails\' and Trent Reznor\'s antics went so far as to draw the attention of then mega-rock star Axl Rose, frontman for Guns N\' Roses.  He invited Reznor and mates to open for them on the European leg of their tour, including an opening slot at the legendary Wembley Stadium.  NIN accepted, though they would grow to regret it.  Reznor went full-tilt in their first performance, \"but by the second one my tail was between my legs,\" Reznor says.  Greasy-haired, mullet-wearing metal heads apparently could not digest Nine Inch Nails so-called \'synth\' sound, and Reznor and company found themselves ducking cheap whiskey and beer bottles thrown at NIN in appreciation.  \"It\'s one of those things when you think it might be a funny idea then you realize, when you\'re up there, maybe this ain\'t so great,\" said Reznor.  Needless to say, Reznor turned down Rose\'s offer to open for the final American leg of the Gunners tour.  After nearly three years of touring, Trent Reznor and Nine Inch Nails finally decided to take a break.  Reznor wanted to begin recording anew, working solely with Flood, but he had come to realize that Nine Inch Nails\' relationship with TVT Records head Steve Gottlieb and the label was at an impasse.  The next several years would find Reznor engaged in a bitter legal dispute to free himself from the constraint of the absolute antithesis of artistic freedom, while at the same time recording new, even more angst-ridden music in secret.  Nine Inch Nails had been born and unleashed upon an unsuspecting world mired in musical mediocrity; but at the same time, their future was shaky at best. xBuBzx  '),(7814,3000240,'29,April,2004','2004-04-29 00:00:00','   My   Esteem   Problem   Part One   When I was little, and from what I can remember, I really quite hated myself. I used to think that If I didn\'t exist that: \"who the fuck is gunna care about me?\". I think I must have only been around 6 or 7 at the time and I\'ve only just in the last week or so managed to figure out that it was this and other factors that has caused my problems through all stages of school and life I have been through. This could have been down to a couple of things I guess:  1. No siblings of my own age, 2. How hard it was for me to make friends in general when I was little, 3. The spoiling, lack of true discipline by Mum and Dad, 4. Or I could just be making a rash over judgment on the whole thing, and it could all be in my head.  I tend (or am trying to do more so) to disagree strongly with the latter as It has no relevance or evidence, and personally I think that the answer could be a mixture of all the other 3 alternatives on offer as they are all reliable, viable et al to everything about me. Both my sibling are over 30 and left home when I was 6 and the time they were living at home, weren\'t here much at all anyway. Being as my parents are quite  old  for parents aswell, I was quite lonely at an early age, and never really encouraged to go out and make friends much either. Because of this perhaps, or possibly because of some obscure biological reason, making friends was never really an easy thing for me. Sure I had some. I wasn\'t Billy no friends definitely, but I had a nack for over-reacting and being a selfish brat.    Part Two Tomorrow   xBuBzx  '),(7815,3000240,'28,April,2004','2004-04-28 00:00:00','   Ian\'s random sayings . . .    God I hate Charlie, It befuddles me, alright maaaaaaaaaaaate, awww BuBz maaaaate, God I hate Aiden, Dan is errr, a little bit, errrr sly, OH MATTY!!  I guess I\'ll put more here when I can think of them . . .  P.s. I dedicate this post the God that is Ian xBuBzx  '),(7816,3000240,'28,April,2004','2004-04-28 00:00:00','   So finally, after two quarreling, mouth ulcer infested, blood skewing years of this mechanical animal inhabiting the large majority or my mouth against its will, suddenly, with a mighty desperation it must be said, let go of its reign in a Saddam-Hussein-esque fashion.  Good riddance I most definitely say. When the bloody thing actually came out and the concrete shaver had done its job, my tentative tongue checked the damage and possible remains. I must say that the thing did its job ok. It\'s got more of a \"semi-circular\" feel to it than ever before and eating the ultimate orgasmic cookies from the canteen was a taste, sight, well just general experience to behold for the ages (in my view at least).On bad side I can see thus far is that I\'m going to have to start brushing regularly again, but that\'s not worries for what I\'ve got back - a fully natural, non-war infested set of teeth, WOOOO!!!  So nehoo, I was in quite a lot of boredom last night, and as the Net seemed to be less than the time filling source that it usually is, I sat down and watched a program about ANGRY PISSED OF CHEF Gordon Ramsey, and watch him try to turn a young, awful, obnoxious chef from Yorkshire into something that can produce half edible food, and the program made for pretty good viewing to be honest, helped somewhat by my liking all things where the central character gets his finger yanked out of his own arse.  The Universities also came out to play yesterday at college and just for the hell of it, I managed to pick up certain prospectus\' on Oxford, Cambridge and Thames Valley et al. The day, although on the very long side, came out to be very useful and some things  were really interesting, like the gap year presentation, and most things there made me open my eyes to the advantages of them in the long term. xBuBzx  '),(7817,3000240,'27,April,2004','2004-04-27 00:00:00','   Well today was a good day and I think things are looking up on the romance side, well just some things are looking up that\'s all, and I definitely am getting a clearer picture of who I like for sure. God knows how much I\'d like someone on my life. It was sunny as well until the monsoon hit Molesey with a vengeance, and it was quite a sight to behold with the lightening striking something not too far away from our house. Ciao xBuBzx  '),(7818,3000240,'26,April,2004','2004-04-26 00:00:00','   God and now the  ginga mother of mine  isn\'t going to pick me up from college anymore because I do something  EVERYONE ELSE DOES . Evil parasite . . . xBuBzx  '),(7819,3000240,'26,April,2004','2004-04-26 00:00:00','   Korn - Seed    Every day it gets a little harder, can\'t seem to get away. I remember there\'s a certain place, a place I wish I\'d stay. I feel so lost within - pressured, I\'m headed for that day. Just one thought in my head, really. Do I need this fame?   Every time, god damn, I look at my son (seed), I see something I can\'t be. Beautiful and care free, that\'s how I used to be. Like some god damn fucking freak, I\'m so pressured, I\'m so weak, Something takes a hold of me, something I can\'t believe.  I lay in bed at night and wonder, should I go on this way? It\'s the only thing I really got for now, and it\'s called fame.   Every time, god damn, I look at my son (seed), I see something I can\'t be. Beautiful and care free, that\'s how I used to be. Like some god damn fucking freak, I\'m so pressured, I\'m so worried, something takes a hold of me, something I can\'t believe.  So I see this face so innocent and fine... and so fine. So I see this face and I realize it\'s mine.   I feel the rattle...   So I see this face so innocent and fine... and so fine. So I see this face and I realize it\'s mine.   I feel the rattle...   Every time, god damn, I look at my son (seed), I see something I can\'t be. Beautiful and care free, that\'s how I used to be. Like some god damn fucking freak, I\'m so pressured, I\'m so worried, something takes a hold of me, something I can\'t believe.  Like some god damn fucking freak! xBuBzx  '),(7820,3000240,'07,May,2004','2004-05-07 00:00:00','   Even though its been one hell of a week already, God thank weekends. Should be extended by an extra day no question. Well of to a party in Molesey with Jen, Rach and possibly Dan. Should be good, last time I went to a party there I came home with hickeys and I can vividly remember the elaborate pulling between Rachael and Frances (Never laughed so much in my life apart from when Paul tried to chat up a 90 year old woman in the swan once I suppose). Saturday is still a blur really apart from Chelsea playing Manure at Old Trafford. Quite confident that we will be pissed of and lay the final nail in United\'s season. Well a good end to the week in football terms would be gladly welcome.   Have fun wHeReVeR yOuR gOiNg. xBuBzx  '),(7821,3000240,'06,May,2004','2004-05-06 00:00:00','   After yesterdays abysmal event (which we shall never again utter) I\'m feeling in a relaxed mood and quite tranquil. Maybe because there\'s no agitation or nerves (well except tonight maybe as the mighty Tooting and Mitcham are in some obscure cup final which I\'m going to with Ian - COME ON YOU TOOTS YEH!!) about football anymore and the fact I\'ve decided that my Christian theories are fairly useless and so I\'ll stop relying on signs from God (read into what I would have done if Chelsea beat Monaco last night) and just go for things on my own terms and accord. Hopefully it will work; hopefully.  Was also thinking as you may have gathered I tend to do it quite a fair bit, to try to sway every open minded music lovers, if they don\'t already, onto Nine Inch Nails. I copied a while back, one almighty CD with 12 of there best tracks on them and it tingles my spine every time I listen to it. Just marvelous. Ciao children. xBuBzx  '),(7822,3000240,'05,May,2004','2004-05-05 00:00:00','   We lost but, I think on the whole I knew we were going to lose form the start after what happened in the 1st leg. Anyway got home and was in a thoroughly depressed and suicidal mood (like every other Chelsea fan) but got speaking to Jen and I managed to utter some laughter (Jen often seems to make me laugh). So of to bed now to decipher the next phase of the Chelsea world domination plan. Hoorah. xBuBzx  '),(7823,3000240,'05,May,2004','2004-05-05 00:00:00','   We can, we can really do this tonight. Early goal and it\'s all dandy as a friend of mine would say. Knowing Ranieri, we will trust the injured players, or hopefully not as I hope to God. As sad as this looks as well, this is what I think the team will (or should) be tonight:                                                              Cudicini  Johnson                 Terry                                                       Gallas                                    Bridge         Geremi                   Parker                                                                                                  Cole                                                                                            Lampard                                 Gudjohnson                                                                                            Crespo   I think we will snatch it, but overall I\'m not sure. It can be done and if we play to our maximum (like I hope and expect), we\'ll do it. I\'m probably going to the Shed bar at Stamford Bridge and be very drunk, so I won\'t remember if we loose. Hoooorahhh. I also plan to call quite a few people strangely and have a wonderful drink fuelled celebration over the wireless with them. I\'ve also decided that (as odd as this sounds also) if we win, it will be a sign, and that I should take a few more risks than I usually do on a few things. Don\'t know why but I\'ve decided to stop hating myself and appreciate life and stop presuming that others will not be there for me and wouldn\'t take a chance with me.  Oh and revise also lol.  I was thinking last night also, how above virtually anything, football rallies the emotions. Forget life, depression et al; a game where 22 men try to kick a small synthetic ball into a rectangular goal can evoke the strongest of emotions in people. Chelsea is my only religion, and when times haven\'t been so good for me I\'ve always known, however well they\'re doing, that I can rely on them unquestionably for some relief. This is why I love them so dearly as I do. The love I have for them is like the one you give to your children, and as over-exaggerated as this sounds, it\'s the only way I can describe it, and in the most truest form. xBuBzx  '),(7824,3000240,'04,May,2004','2004-05-04 00:00:00','   I had a truly memorable and marvelous argument with someone on the Heatham boards, something which I try to steer clear of to be honest, but it was against someone with gladly a more than adequate knowledge of how to express his language, so I was more than glad to be part of it :  urlLink Easily the best net argument of my life  (my nickname being matty . . . . of course (-: ).  Anyway I had a loooooong sociology conference at College and I\'m getting more and more nervous about the match on Wednesday as the match ebbs ever closer *sigh*.  Also, you must agree the Weather is having one almighty period isn\'t it? God rain one minute, sunshine the next, what the hells next - snow with sun and tornados in the mix? xBuBzx  '),(7825,3000240,'03,May,2004','2004-05-03 00:00:00','   Swimming In The Haze: Part Two - The Broken Phase    Here\'s part two of a net biography of my favorite band Nine Inch Nails:  Rock superstardom: the tired cliche that is the very crux and irony that confronts rock stars on the verge of greatness. To write, compose, and perform music that defies the establishment in order to gain greater notoriety and eventually become part of said establishment which you must intrinsically scorn, this is the vicious circle that despite their greatest efforts, all stars and starlets with even a scintilla of a conscience must struggle to come to terms with. So it was in 1991 when Trent Reznor found himself precariously balanced on the edge of greatness.  Having recently released arguably the biggest selling Industrial album of all time (for the time), Pretty Hate Machine, and having wrapped up the hugely successful stint on the now legendary Lollapalooza tour, Reznor found himself wondering what the next move was to take his fledgling Nine Inch Nails to higher levels, despite the fact that they were already hurtling along with new momentum.  Prior to embarking on the Lollapalooza tour, Reznor had made it clear to TVT Records, his first record label which released NIN\'s debut album and subsequent singles, that Nine Inch Nails would not release another record under their watch.  Their meddling with Reznor\'s musical artistry on Pretty Hate Machine had enraged Reznor so much that he could no longer fathom working with anyone who would compromise the quality of his art without his say in the matter.  In addition, Pretty Hate Machine\'s success had made Reznor a producer in demand, and TVT resented and often attempted to thwart Reznor\'s numerous side-projects with bands such as Pigface and 1000 Homo DJs.  TVT, smelling the green resulting from the critical acclaim and widespread success of Nine Inch Nails\' electric performances on Lollapalooza, was not ready to allow their budding star to walk away so easily, despite the fact that then TVT Records chief Steve Gottlieb admitted to causing the rift.  \"Our relationship got dysfunctional.  I obviously let Trent down.  We\'re talking about the guy who wrote the lyric, \"I\'d rather die, than give you control!\"  John Malm, Jr., Reznor\'s long-time manager, has said of that period, \"Trent would have mowed lawns rather than make another record for TVT.\"  And so began a several years-long legal battle that is nearly legendary in the recording industry.  The chess match and legalmaneuveringg caused by Nine Inch Nails attempting to break away from TVT forced Reznor to actually record in secret.  Along with renowned producer Flood, Reznor fled to Miami Beach, among other locations, where he and Flood worked guardedly and secretly on the follow-up to Pretty Hate Machine.  Reznor had become bitterly jaded and disillusioned with the music industry, and admits to adopting a \"snotty, elitist mentality, you\'re not cool enough to like my band, don\'t buy my records.\"  He \"wanted to make a \'fuck you\' record\" to prove that he wasn\'t a sell-out, and also to reflect his general animosity towards life during the TVT fiasco.  Two years after it had begun, Nine Inch Nails\' feud with TVT Records concluded with Jimmy Iovine, president of Interscope Records, actually buying out TVT in order to acquire NIN.  At first, Reznor was furious, seeing Nine Inch Nails\' sale akin to being traded on a slave market, and has referenced it as such in past interviews.  Part of the deal, however, was the set-up of Reznor\'s own label independent of Interscope, to be called Nothing, that would allow Reznor and his Nine Inch Nails total artistic freedom.  Nothing Records\' inaugural release was Broken, a harsh and angry six-track mini-record that was the result of Reznor\'s clandestine recording sessions during Nine Inch Nails\' break from TVT Records.  While Reznor expected the switch to a major label would be perceived as a sell-out, thus alienating much of NIN\'s fans, Broken, which was intended as nothing more than a middle finger to pretty much everyone, garnered critical and commercial success.  So much so, that Broken was awarded no less than two Grammy\'s, to the surprise of both the recording industry in general and Reznor himself.  Broken was immediately followed up by Fixed, an allegorical indication of his transition from one record label to another.  Fixed was another six-track mini-album, which was comprised of the re-mixed and re-worked tracks from Broken.  Reznor took full advantage of his newfound freedom to work with other artists on Fixed, and Peter Christopherson of Coil and J.G. Thirwell, a.k.a. Clint Ruin, contributed their own unique interpretations of Nine Inch Nails\' songs.  Fixed would mark the first Nine Inch Nails re-mix album, and this fondness for releasing re-mixed and varied interpretations, both by himself and other contributors, would continue throughout Reznor\'s career.  While Reznor\'s new Nothing Records set up official shop in Cleveland, Reznor found himself spending more and more time in New Orleans, a city fast becoming a retreat for extreme rockers looking to get off the radar for a brief respite.  Being so far removed from the West Coast, Hollywood intensity and fast-paced, urban sprawl of New York, New Orleans was a place where up and coming rock stars could go relatively unnoticed.  After a period of recuperation and reflection, getting used to the new freedom being enjoyed under Interscope, and the satisfaction of knowing the commercial success from Broken and Fixed, made on his own terms, Reznor was ready to begin work on the next, full-length Nine Inch Nails record.  He had a lot to live up to.  Nine Inch Nails\' raucous tours, eclectic sound, and maelstrom of rumours and interesting \'pickles\' that Reznor seemed to always land himself in had spawned a devout, cult-like following of fans.  Pretty Hate Machine was fast-approaching more than a million copies in sales in the USA, and the album was now consistently garnering favourable review and acclaim across the pond in the UK.  Having a solid idea for the direction of Nine Inch Nails, Reznor went shopping for a location in which he could hibernate while giving birth to his next musical creation.  After checking out several properties in the Los Angeles region, Reznor selected a big, beautiful home with a spectacular view of Los Angeles, and with a very reasonable price tag.  Reznor signed the deal and began to move in.  It would later be spectacularly revealed that said house was actually the location of the famous Charles Manson murders where Sharon Tate, wife of director Roman Polanski (Rosemary\'s Baby), and four guests were gruesomely butchered in a bizarre, cult act the likes of which shocked America as badly as Hiroshima or President John F. Kennedy\'s assassination.  After finding out about his new, albeit temporary residence\'s history, Reznor remained unfazed, and elected to remain, even warming up to the place for the added, spookily cool aura it cast over recording sessions.  At least this is what we are led to believe.  This amazing revelation would have to be based on the assumption that Reznor, his agents, friends, and family did not clue into the history of said property during the course of acquiring it.  In addition, Trent Reznor would have been four years old when the Manson murders occurred.  Again, this was a murder/media event that was totally unprecedented in American history, and was the focal point of the media for months, years, and now even decades after the fact.  Further, the trials went on for years, and given ReznorÃ‚Â’s documented and rather obvious penchant for an interest in all things spooky, the odds are very good that Reznor knew exactly what property he was buying.  But why hide that fact?  First of all, perhaps no one denied anything.  The media would undoubtedly have attempted to contact Reznor\'s management or Reznor himself, but a lack of confirmation would have been at the mercy of the journalist\'s creative whim, and simply added to the larger myth.  Additionally, perhaps Reznor simply found it easier not having to explain why someone could possibly have an interest in such a place from a personal point of view, as the media would undoubtedly have twisted and perceived the explanation as nothing more than a publicity stunt; better to remain mysterious and dark, albeit ignorant, rather than an attention-seeking wannabe.  Rumours also dictate that the walls of one particular room still bore the infamous blood stained word \"pig\" scrawled on it.  This is simply not true.  While the word \"pig\" was written in blood, it was not inside the room that Reznor would in fact make the brains of his recording operation, and affectionately nickname Le Pig, or even in the house at all, for that matter.  In actuality, the word \"pig\" was on the front door of the house.  And even putting that fact aside, is it really plausible to believe that after all the investigation and trials, and subsequent renovations of the house to restore it to a condition suitable for sale, that the word \"pig\" would be left, in blood no less, on the wall/door of the house?  Highly unlikely, dear reader.  But it sure makes for more interesting journalism, doesn\'t it?  So, while Reznor claimed that the house\'s history did not bother him, it undoubtedly affected the mood and atmosphere of the recording sessions.  Reznor again exercised his freedom under Interscope and Nothing by calling on select musicians.  Notwithstanding his now staple producer, Flood, Reznor also called on Adrian Belew, the guitar legend behind many David Bowie masterpieces.  Stephen Perkins, drummer for Jane\'s Addiction laid down some percussion, and Alan Moulder was brought in to lend his considerable expertise at the mixing board.  Reznor recognized that a pattern had begun to emerge with Nine Inch Nails, and he \"didn\'t want to box NIN into a corner, where everything had to be faster and harder than the last record.\"  Instead, Reznor wanted the next album \"to be a departure from Broken...one ultra-fast chunk of death...This time I wanted to make an album that went in ten different directions, but was all united somehow.\"  Reznor had decided on somewhat of a concept album, where the record would have a unifying theme.  Where Pretty Hate Machine laid bare raw emotion, and where Broken and Fixed focused specifically on anger, Reznor turned inwards on himself once again, and dug deeper into his own psyche.  The theme of the record would follow the descent of an individual through their own inner revelations as they strip away every aspect of their very selves, and scrutinize the basic moral code that makes up society as a whole today: love, hate, religion, relationships, sex, violence, and vices as seen under a psychological microscope.  Examining every aspect to the finest detail, Reznor wrought a bleak, wallowing descent into introspection that would become the title for Nine Inch Nails\' newest opus: The Downward Spiral.    Never had Reznor delved so deeply into his personal life, and The Downward Spiral represents a thorough purging for Reznor, to the point where Reznor has openly admitted he \"out-bleaked\" even himself and hopes to never again reach that point of despair.  Rather than the all-out aural assault that was Broken, The Downward Spiral was a haunting project that deftly combined texture and mood to evoke nerve-wrenching emotion.  Always one to admire full-bodied albums such as Pink Floyd\'s The Wall, Reznor attempted to create a record that required the listener sit down and absorb the album front to back, several times, before they began to come to the revelations behind the seeming caterwaul of sound.  He admired albums that slowly revealed themselves, and in the long run contributed to one\'s \"musical vocabulary.\"  Reznor also recognized that he may aled the very un-radio friendly lyric \"I want to fuck you like an animal\" would, surprisingly, become their biggest single ever.  The track, edited to a friendlier, profanity-free version, received heavy rotation on the airwaves and huge underground success in clubs all over the world.  The accompanying video, directed by auteur Mark Romanek, closely paralleled the shocking content of \"Closer\" and hit the viewer hard with images of decapitated pigs\' heads, crucified monkeys, naked flesh, and Reznor spinning maniacally out of control.  An extended version, \"Closer to God\" would later be released featuring various interpretations and some additional tracks.  Nine Inch Nails were a permanent success.  Trent Reznor had toiled and worked his way up from gopher at a local recording studio to a mainstream artist garnering critical acclaim from both established and newfound fans, and critics the world over.  The Downward Spiral represented a pinnacle of achievement in both Reznor\'s life, and in the history of music.  By looking into himself and laying bare all the deep and darkest secrets of his soul, Reznor had inadvertently connected with millions of others who exist on a common emotional plane.  The connection had to once again become physical.  After a three-year hiatus from touring, it was time for Trent Reznor to assemble hiser\", a haunting piece of music that featured the very un-radio friendly lyric \"I want to fuck you like an animal\" would, surprisingly, become their biggest single ever.  The track, edited to a friendlier, profanity-free version, received heavy rotation on the airwaves and huge underground success in clubs all over the world.  The accompanying video, directed by auteur Mark Romanek, closely paralleled the shocking content of \"Closer\" and hit the viewer hard with images of decapitated pigs\' heads, crucified monkeys, naked flesh, and Reznor spinning maniacally out of control.  An extended version, \"Closer to God\" would later be released featuring various interpretations and some additional tracks.  Nine Inch Nails were a permanent success.  Trent Reznor had toiled and worked his way up from gopher at a local recording studio to a mainstream artist garnering critical acclaim from both established and newfound fans, and critics the world over.  The Downward Spiral represented a pinnacle of achievement in both Reznor\'s life, and in the history of music.  By looking into himself and laying bare all the deep and darkest secrets of his soul, Reznor had inadvertently connected with millions of others who exist on a common emotional plane.  The connection had to once again become physical.  After a three-year hiatus from touring, it was time for Trent Reznor to assemble his Nine Inch Nails and embark on a tour in support of his hallmark album.  Garnering a great deal of their initial attention and subsequent reputation from the live Nine Inch Nails experience, and now boasting a massive new fan base, Reznor\'s next challenge would be delivering that now legendary stage show to Nine Inch Nails\' new and rabid fans worldwide. xBuBzx  '),(7826,3000240,'03,May,2004','2004-05-03 00:00:00','   The Ego, The Id, And the Ã†nema    The primary role of art has always been not just as an outlet for the artist, but to comment on society, its people, and its values.  Whether said art comments on the society in which it is displayed or an alternate society in order to bring enlightenment, artists typically have something to say, and their art is their voice.  Music is certainly no exception to this rule, and the fundamental tenet of the more alternative vein of music which we, dear reader, typically align ourselves with, is most definitely to comment on society.  But what is the basis for this commentary?  Without question, psychology.  It is the basis of the artist\'s voice, and psychology often works not only in predictable ways, but also in lesser known ways.  Most of us who have opposable thumbs and the ability to read have heard of Dr. Sigmund Freud, who postulated and theorized quite literally the basis of most of modern-day psychology as we know it.  One core theory of Freud\'s concerns the Id, the Superego, and the Ego.  These widely accepted psychological models are more prevalent in modern-day music than one might realize, as they together are more or less present in each and every one of us.  Freud tells us the Id is the primitive unconscious consisting of fundamental drives.  The Id is irrational, impulsive, and seeks immediate gratification with no consideration for what is realistically possible, socially desirable, or morally acceptable.  It is governed by the pleasure principle â€“ the unregulated search for gratification, especially sexual, physical, and emotional pleasures â€“ to be experienced here and now without concern for the consequences.  The Id is self-centered, and looks inward only.  Certainly, much of modern day music has sexual, physical, and emotional overtones, but most of the modern, bubble-gum pop music heard on Top 40 radio lacks the intensity that one would imagine is involved in \'primitive\' and \'impulsive\' decisions.  More alternative artists such as Nine Inch Nails, however, certainly embody the Id, and go so far as to embrace it.  Nine Inch Nails\' abrasive and eclectic sound, searing vocals, and indisputably raw emotion, both recorded and live, most definitely make tangible Freudâ€™s theory of the Id.  And much of the lyrical content of Nine Inch Nails is certainly inwardly focused, centering on one\'s journey through self-exploration.  With the yin, there must always be a yang.  Freud also postulated the Superego, or storehouse of one\'s values, including moral attitudes learned from society.  The Superego is one\'s conscience, consisting of that inner voice of \'oughts\' and \'should nots\'.  The Superego also includes the Ego Ideal â€“ an individual\'s view of the kind of person one should strive to become.  The Superego is often in conflict with Id â€“ while the Id wants to do what feels good, the Superego operates on the morality principle and insists on doing what is right.  The Superego looks not towards the inner self, as the Id does, but instead focuses on the outer world and its morality.  The Superego is most definitely present in standard pop music, but to a much lesser extent in the more hardcore veins.  While Britney Spears is happy to package her ass for sale on the cover of her records, yet wouldn\'t be caught dead in the nude anywhere, at anytime (despite what those alluring pop-up ads tell you), Raymond Watts (of Pig/Watts/KMFDM fame) goes the extra mile and pulls his cock out onstage (I know...sadly, I was front row and witness to it).  One could certainly attribute Watts\' dick tricks to the Id\'s victory over the Superego, and Britney\'s reluctance to bare all, despite the fact 90% has been bared already, indicates instead the triumph of the Superego.  This unconscious factor theorized by Freud is present in everyone, but in artists such as Britney, it is simply more dominant.  To round out Freud\'s theory nicely, he also theorized the Ego â€“ the reality-based aspect that arbitrates conflict between Id impulses and Superego demands.  The Ego, Freud says, consist of one\'s beliefs about the causes and consequences of behaviour.  Governed by the reality principle, which puts reasonable choices before pleasurable demands, the Ego arranges compromise between the conflicting Id and Superego.  Thus, the Ego is the arbitrator, or seeker of common ground.  Marilyn Manson, anyone?  While Nine Inch Nails\' typically brutal, all-out aural assault is influenced primarily by the Id, and Britney Spears unwillingness to shows us every square inch of her breasts is controlled by the Superego, Manson\'s music adheres more to the Ego.  Marilyn Manson take their role as artists seriously, and as such have a lot to say about society.  But rather than simply focus on the pleasure-based principles (rap\'s continuous yearning for \'booty\'), or rather than play it safe by performing easily digestible, zero-to-say music (pop\'s continuous yearning for um...erm...bright lights, tight pants, and a Pepsi sponsorship?), Marilyn Manson analyze society, and comment on it by providing solutions, albeit rather extreme and unusual alternatives (\"light a candle for the sinners / set the world on fire\").  Thus, Manson embrace their Ego, rather than letting their Id or Superego dominate, and as a result enjoy more controversy and media attention than artists who allow their Id and Superego to hold the reins.  Beyond Freud\'s basic structure of personality, several post-Freudian descendants expanded on his basic theories of the unconscious.  One such psychologist is Carl Jung, a student of Freud.  He greatly expanded on the conception of the unconscious.  For Jung, the unconscious was not limited to an individual\'s life experiences, but was also filled with fundamental psychological truths shared by the entire human race.  Jung believed the concept of collective unconscious predisposes us all to react to certain stimuli in the same way and consists of archetypes associated with an instinctive tendency to feel and think about it or experience it in a special way.  These archetypes, known as animus in the case of male archetypes, and anima in the case of female archetypes, are primitive symbolic representations of a particular experience or object.  And these archetypes, both male and female, are present in each and every one of us, and guide us daily in our actions, says Jung.  Now these are relatively big words and theories for the general population, and you would have a hard time convincing this author of evidence of this concept to a great degree in Christina Aguilera\'s latest offering.  Attend a Tool concert, however, and it\'s a completely different story.  Toolâ€™s heavy, prog-rock sound, laced with complex lyrics, undulating imagery, and surreal art suggests just the opposite: these are well-read believers in psychology and Jung\'s basic tenets, and their live show is evidence of this even more than their deliciously intricate records.  Large screens typically overhanging the stage continually bombard the audience with imagery naturally encouraging the audience to unconsciously reach deeper into the collective unconscious in seek of wisdom and fundamental truths.  Spoken word clips from Timothy Leary and other avant-garde artists emulate the collective unconscious.  And might I remind the reader of one of Tool\'s more popular records, Ã†nima?  A cross-breed of the psychological term \'anima\' and the considerably less pleasant (at least for most) \'enema\', Ã†nima is Toolâ€™s humorous play on words that parallels the animaâ€™s continual presence in all of us in conjunction with a collective mindfuck.  While each and every individual on the planet have in theory all of the aforementioned psychological factors at work in their subconscious, the artists described above simply lean towards one or another more directly.  Yet they all are present in our thought processes, the Id and Superego engaged in an eternal tug-of-war, while the Ego maintains balance.  And hovering above all of these processes, says Jung, are the male and female archetypes in every one of us guiding us all in a similar fashion.  So, the next time you\'re spinning a disc, look past the glitzy album cover and the latex encased body, and inwardly challenge the musician to speak to you on these basic psychological planes.  You might be surprised at what you hear, or perhaps what you don\'t. xBuBzx  '),(7827,3000240,'15,May,2004','2004-05-15 00:00:00','   Well the Links are all sorted. Danke to Jen.  Well I\'ll post something meaningful when I can be bothered. Having a bad weekend so far by all accounts. Erm but well done to Jon for winning the random Microsoft prize, well deserved for all those endless hours at the computer as a pre-teen hehe. xBuBzx  '),(7828,3000240,'14,May,2004','2004-05-14 00:00:00','   Im really bored and I woke up too early, *sigh*  '),(7829,3000240,'14,May,2004','2004-05-14 00:00:00','   Well maybe later . . . xBuBzx  '),(7830,3000240,'13,May,2004','2004-05-13 00:00:00','   Not much to say - so won\'t say much, i.e. anything xBuBzx  '),(7831,3000240,'12,May,2004','2004-05-12 00:00:00','   Interesting new feature on blog world I found (as most of you who have a blog probably have) is the link box you can put at the top. So I decided to treat you wonderful people to a piece of renaissance art, which can be magnificently viewed upon clicking the link on the \"Oh there your mates!!!\" part (another of Ian\'s random sayings encase your wondering). Thinking of Ian as well, he said he would start a Blog of his own soon which I look forward to in mesmeric anticipation.  I\'m yet to find out how to do the links to interesting and wonderful links although hopefully tonight they\'ll be sorted (Jen  WILL  help/do that for me [Danke LOVE YOU!!! lol]).  Nehoo Been an average so so sort of day except this morning when I was having my eyes tested (6 month check up, and nothing to do with the *ahem*  mowing down of a child incident ) and discovered these amazing set of geeky old retro type of glasses (yer you know they\'re cool   really  ) were - wait for it - Ã‚Â£120!!!!!!!! They\'re begged and swayed me though with they\'re powerful methods of persuasion (well just a cookie really) and now I await the ultimate-cool retro geek eyesight YAY.  Besides that I\'m loudly annoying the townies (always one of my favorite past times with my excessively loud Metallica and my even more excessive ignorance to there complaints. Hoorah: now where\'s some food I\'m hungry. Well must be of to Psychology soon enough so ciao for now my little one\'s. xBuBzx  '),(7832,3000240,'11,May,2004','2004-05-11 00:00:00','   Well all in all I haven\'t quite figured out how to work out the new blog layout - but I guess I\'ll figure it out eventually.  Well it was a really shitty type of week which gradually got worse as the week went on and on. Never the less - this is not last week and I really think things are on the up. I have a spontaneous confidence streak to do with my esteem (I\'ll write the second part to  the esteem thing  soon I promise) and I reckon, despite some odd setbacks that it\'s going rather well and better than expected. Maybe (in my warped imagination) it will lead to some kinkual shenanigans if that makes any sense with this one person I must have loved and liked, on and of for God knows how many years. Well I hope so anyway, as does the ensuing draw which will be evidently present. OOh I\'m definetely going to write some erotica about that one mwahahahahahahaha. Ciao xBuBzx  '),(7833,3000240,'11,May,2004','2004-05-11 00:00:00','   Blog has a new format *hides* xBuBzx  '),(7834,3000240,'22,May,2004','2004-05-22 00:00:00','   urlLink    Advice for prisoners . . .  Posted by  urlLink Hello  '),(7835,3000240,'21,May,2004','2004-05-21 00:00:00','   urlLink    Advice for pathetic teenage boys, when they don\'t know how to deal with there \"fluids\" properly.  '),(7836,3000240,'20,May,2004','2004-05-20 00:00:00','   Yesterday I was in a spontaneous (crazy) sort of way so I decided to castrate my hair (on my head  not  my  pubic area  *rolls eyes*). I guess I was in need of something to just \"spark\" or \"happen\", and so I decided my hair was the perfect compromise. Nethertheless, it\'s amazing what unwanted attention the \"anti-follicle\" receive. As much as people like to rub and poke my head, I found disturbingly, a new strange (and hopefully not mental) side to my friend Suzanne. SHE DREW A NAZI SIGN ON MY TEMPLE |-:. So I\'m not sure quite what to make of that or any of the other strange behaviour I have come across today (will get considering It\'s only  half past FUCKING 1 )  Neho, I think I\'ve officially got fed up of the love torment friends decide that I actually like to hear, about their girlfriends and boyfriends \"  problems  \". It\'s hard enough being a permanently celibate as I am, without having to hear other people\'s un-gratitude about the fact, despite the problems et al, that they actually have someone in there life that is actually prepared to call them their own. Maybe, I don\'t know, they should decide to think the whole relationship through before deciding to interrupt my sad but contempt bond with the official Chelsea management *sigh*. Oh well I know they don\'t have bad intentions, well  most  of them anyway, so, well, hmmm, I\'m of to the shitter.... xBuBzx  '),(7837,3000240,'20,May,2004','2004-05-20 00:00:00','   Well now I (or Jen) figured out how to add pictures to crazy blog world, I will enter stage two of my ultimate blog plan - nausea for the masses mwahaha. Well If you\'ve ever seen what pictures I keep (non-pornographic I assure you) on my PC then you will know what oddities I\'ll be displaying in the coming month. Think of the last two as a tame introduction mon ami.  Neho, it\'s the oracle\'s 91st birthday on the weekend, which means somebody going to have to either buy 91 candles (heaven above) or one hell of a present (which will probably be me judging on the family\'s prejudices against the young members of it, namely me and Harry exploitation). Poor little Ethel (who solely endorses the Granny Smith apples lol) won\'t know what to do with all the generosity and gifts. Will she   drown   (literally) in them? Will she think due to her ever (  bless her  ) deteriorating brain, think that the gift rap paper is  in fact  the present? Wow it sounds like a roller coaster ride of emotional shocks and shenanigans; and I  only  wish I was there to see it that\'s all *rolls eyes*.  Happy birthday Gran, and have a good day. xBuBzx  '),(7838,3000240,'19,May,2004','2004-05-19 00:00:00','   urlLink    OOOh Arghhhhhh!!!!  '),(7839,3000240,'19,May,2004','2004-05-19 00:00:00','   urlLink    aww how my dad will look in not too long *sigh*  '),(7840,3000240,'18,May,2004','2004-05-18 00:00:00','   Well seeing as I haven\'t really posted anything on here on any real literal meaning to do with the daily ramblings of my life recently, I thought it just time I gave a visit to you poor neglected souls.  Life\'s really been in tandem with the weather recently, and not much could make me feel any more contempt at things. I hit the ground running exam wise, and I\'ve finally got the will power to believe I can achieve things this summer i.e. lose weight, work vast quantities of overtime, and gain some confidence. Getting into revision really started this whole thing of, and now, safe in the knowledge that I\'ll be of to Cardiff to witness the great war between Ian and Dan (Palace and West Ham to you and I respectively), everything\'s just feels like yayish excitement.  I\'m just really keen to do well this year, and explore hopefully new cultures et al, however, that may not be possible considering the furthest I\'ll probably get from home is work. Nevertheless, this won\'t dampen my spirits.  Hell I even feel like doing a jog some-time soon. Oh no wait . . . The sofa is still vastly more appealing *sigh*. Hmmmm wonder what\'s on the box tonight? xBuBzx  '),(7841,3000240,'18,May,2004','2004-05-18 00:00:00','   Nah Nah Nah Nah, Nah Nah Nah Nah, Hey Hey Hey GOODBYE!!! Nah Nah Nah Nah, Nah Nah Nah Nah, Hey Hey Hey GOODBYE!!! Nah Nah Nah Nah, Nah Nah Nah Nah, Hey Hey Hey GOODBYE!!! Nah Nah Nah Nah, Nah Nah Nah Nah, Hey Hey Hey GOODBYE!!! Nah Nah Nah Nah, Nah Nah Nah Nah, Hey Hey Hey GOODBYE!!! Nah Nah Nah Nah, Nah Nah Nah Nah, Hey Hey Hey GOODBYE!!! Nah Nah Nah Nah, Nah Nah Nah Nah, Hey Hey Hey GOODBYE!!! Nah Nah Nah Nah, Nah Nah Nah Nah, Hey Hey Hey GOODBYE!!! Nah Nah Nah Nah, Nah Nah Nah Nah, Hey Hey Hey GOODBYE!!! Nah Nah Nah Nah, Nah Nah Nah Nah, Hey Hey Hey GOODBYE!!! Nah Nah Nah Nah, Nah Nah Nah Nah, Hey Hey Hey GOODBYE!!! Nah Nah Nah Nah, Nah Nah Nah Nah, Hey Hey Hey GOODBYE!!! Nah Nah Nah Nah, Nah Nah Nah Nah, Hey Hey Hey GOODBYE!!! Nah Nah Nah Nah, Nah Nah Nah Nah, Hey Hey Hey GOODBYE!!! Nah Nah Nah Nah, Nah Nah Nah Nah, Hey Hey Hey GOODBYE!!!  '),(7842,3000240,'17,May,2004','2004-05-17 00:00:00','   Yes Jenny you finally will become a supermarket whore like myself (-: xBuBzx  '),(7843,3000240,'16,May,2004','2004-05-16 00:00:00','   I wouldn\'t say bitter/sweet weekend as such but it was more like norm/sensual if that awakens any sense in your mind of what I\'m waffling on about.  Friday basically I slept.  Rare  Very rare. Not many people know it but I\'m an insomniac. So it was a good thing really.  Saturday was beautiful though. I stayed over at Annas (heya my gorgeous German darling of a scandal) and watched this kick ass film called basketball diaries with Leonardo Di Caprio, who I know gets much flack for his lack of presence, but God he played it well. Anyway after that spent the time until about 5 in the morning talking, hugging and \"foot fondling\" (yay we inventned it) and I told her I loved her, which I still do, right to this day. I\'d die for that girl if I had to, and I just wish one day she was mine. Maybe one day.  nehoo - just a message to the rival hell like enemy blog from Mortlake - I\'ll get you back dammit grrrrr with a wonderfully god liek plague of net locusts onto your blog (-:  '),(7844,3000240,'16,May,2004','2004-05-16 00:00:00','   \"I do apologise about the over rash generalisations there, but it was fun in all senses, and I stand by what I say. It\'s not your right either to say who can post what and what about. Even when Dean was being an ass on this site, he still had the right to speak his mind, and no one could deny him that.   I realize I\'m not Gods gift, unlike yourself \"luv\", and If I thought I was then I\'d be a million times more different to what I am.  Now I\'m not saying that I totally agree with either being a townie or a grundger on here, but I would not call grundgers cowards, just because they won\'t fight other people. Fighting has never been a way to earn respect, its only ruling by fear and intimidation, and as townies are for no reasons what so ever picking on grundgers etc, what is there crime by speaking there mind on here about it and for sake of argument even if it was, would that be a worse crime than violence against anyone?   Surely if your as open minded and non-judgmental as your coming across there, then you would have realized what I was saying by now was true, and you would now give up arguing with your pitiful grammar, go away, and of this site.\"  '),(7845,3000240,'29,May,2004','2004-05-29 00:00:00','   A PERFECT CIRCLE LYRICS  \"The Outsider\"    Help me if you can It\'s just that this, this is not the way I\'m wired So could you please,  Help me understand why You\'ve given in to all these Reckless dark desires  You\'re lying to yourself again Suicidal imbecile Think about it, put it on the faultline What\'ll it take to get it through to you precious Over this. Why do you wanna throw it away like this Such a mess. I don\'t want to watch you.  Disconnect and self destruct one bullet at a time What\'s your rush now, everyone will have his day to die  Medicated, drama queen, picture perfect, numb belligerence Narcissistic, drama queen, craving fame and all its decadence  Lying through your teeth again Suicidal imbecile Think about it, put it on the fautline What\'ll it take to get it through to you precious Go with this, why do you wanna throw it away like this Such a mess, I don\'t wanna watch you...  Disconnect and self destruct one bullet at a time What\'s your rush now, everyone will have his day to die  They were right about you They were right about you  Lying to my face again Suicidal imbecile Think about it put it on the fautline What\'ll it take to get it through to you precious Over this, why do you wanna throw it away like this Such a mess, come to this, come to this  Disconnect and self destruct, one bullet at a time What\'s your hurry, everyone will have his day to die If you choose to pull the trigger, should your drama prove sincere, Do it somewhere far away from here. xBuBzx  '),(7846,3000240,'28,May,2004','2004-05-28 00:00:00','   Mother, Dad and the oracle left in the first part of the mass family excursion to Dorset. All that will be left will be me, Big Brother, food, revision (well attempting), and Laura who I guess will be just down in dear old Brixton shagging her lover. So I\'m on my own, with no Cardiff in site and just a little pub treck before even the great Ian goes on holiday also. So I really might from Monday force myself into philosophy revision. I know I\'ve basically fucked up my other two subjects so I\'ve really just got to just pass this subject if I have any chance of going to Uni next year *sigh* xBuBzx  '),(7847,3000240,'26,May,2004','2004-05-26 00:00:00','   urlLink    The townies from Esher would say . . .  '),(7848,3000240,'26,May,2004','2004-05-26 00:00:00','   Part 3 of the history of my favorite band - nine inch nails    Swimming in the haze: spirals, games, and soundtracks    Growth.  It is a basic principle in life.  All things grow and mature as they span their life from birth to infancy, adolescence, maturity, old age, and ultimately the final absolute in life, death.  So it is with artists, and those inclined to the musical side of the artistic element.  And so it was with Nine Inch Nails in 1994, who found themselves at the adolescent stage of their musical lifetime.  The past several years had been tumultuous for Trent Reznor and his Nine Inch Nails, and consisted of a great deal of firsts for them.  Nine Inch Nails had signed their first record deal, released their first album, toured as a live show that turned more than a few heads, won several Grammy\'s, and successfully parted ways with the aforementioned label, TVT, in favour of more artistic freedom under Interscope.  Reznor had also formed his own personal label, Nothing Records, and caused yet more controversy when he purchased and moved into the infamous Tate mansion, site of the infamous Charlie Manson murders, to record Nine Inch Nails\' next studio album.  By March of 1994, Trent Reznor and Nine Inch Nails had completed work on their latest record, The Downward Spiral, and released Halo 9, \"Closer\".  The single and accompanying video by director Mark Romanek began to gain notoriety from an audience of fans who had been caught up in the tempestuous maelstrom that was the Grunge rock movement.  For the past several years, while Reznor remained holed up in \"Pig\" Studios in Beverly Hills, quite literally amidst a genuine American nightmare (or at least the site of), grunge had taken sway as the reigning King of Genres, and had eradicated virtually every hair metal band in the world.  Pearl Jam, Soundgarden, and of course Nirvana held court now, and very few alternative (no pun intended) genres were holding steady against the Grunge movement.  And so, Nine Inch Nails was reborn into the media mainstream.  Three years had passed since the release of their last full album, Pretty Hate Machine, and while they had made an impression in the early stages of Grunge\'s uprising on tours such as Lollapalooza, most music fans\' attentions were focused on the Alternative movement.  But few could easily dismiss the stark sights and sounds that were \"Closer\".  Pulsating beats, warbling effects and no-holds barred lyrics were justification for the great deal of attention the single garnered.  Critics began to simultaneously praise its brilliance and dismiss Nine Inch Nails and Trent Reznor fort \'sell-outs\' the more popular the single became.  In the UK, top BBC Radio One DJ Bruno Brookes accidentally (yah, right) played the full, unedited album version of \"Closer\", rather than the typical, edited version distributed to radio stations free of obscenities.  Thousands of British listeners were treated to \"I want to fuck you like an animal\" over the airwaves, and thousands jammed the switchboards at the BBC lodging their complaints.  Reznor followed up on the success of the single with \"Closer to God\", an additional single that featured mixes of several other tracks including \"Heresy\", \"March of the Pigs\", and \"Memorabilia\", a song originally recorded by Soft Cell.  And just as quickly as Grunge fans were taking notice of Nine Inch Nails\' return to rock, their limited attention spans were abruptly yanked bank to Grunge, and Trent Reznor found himself standing in the shadows yet again.  On April 8, 1994, the world was stunned to hear that Nirvana founder and frontman, the Godfather of Grunge, Kurt Cobain had taken his life the evening before with a shotgun.  Cobain had long since been painted the poster boy for the troubled, drug-addled rock star, and he had finally succumbed to the stress subjected upon him by the media, his art, drugs, and other various elements.  Trent Reznor, who had been often targeted as this same persona given his penchant for the darker side in life, did not succumb to these pressures, but rather poured them out through his art, much like Cobain had, yet had lived to tell the tale. In many ways, Cobain and Reznor were alike and several parallels can be drawn betwixt them.  Both came from broken homes.  Both had been raised in Small Town, USA.  Both had expressed obvious dissatisfaction with society.  And both felt pain and expressed it more acutely than the average bi-pedaled Homo sapiens.  Nevertheless, Reznor had survived, and with the grunge movement in an uncertain place with their leaders Nirvana dethroned, Nine Inch Nails embarked on what would be the most important tour of the band\'s career. From the spring through the late summer of 1994, Nine Inch Nails toured relentlessly throughout North America and Europe on their Self Destruct tour.  It would be their first major, headlining tour, and Reznor and company, now consisting of Robin Finck, Charlie Clouser, Danny Lohner, and Chris Vrenna, were making a serious impression.  The Downward Spiral was garnering critical acclaim, \"Closer\" had become a regular in radio and video rotation, and Nine Inch Nails were being championed as rock\'s future saviours as the Grunge movement did not stumble, so much as it wavered uncertainly, as if hesitant as to where to go next.  On August 13, 1994, Nine Inch Nails took the stage at Woodstock \'94, alongside top acts such as Henry Rollins, Red Hot Chili Peppers, Green Day, and Crosby, Stills, and Nash, to name but a few.  Nine Inch Nails\' bombastic, frenzied, and mud-soaked performance nearly stole the show, as it were, despite the fact that there were many acts to follow.  The crowd that had assembled for Woodstock \'94 were not expecting the electric hellfire that was Nine Inch Nails\' live show, and the performance was heralded as nothing short of phenomenal.  Nine Inch Nails had officially arrived.  After a brief two-week break, Reznor took his Nine Inch Nails out across North America on part two of the Self Destruct tour.  Prior to setting out on tour in the spring, Trent Reznor had relatively recently discovered a band by the name of Marilyn Manson in Florida, and immediately took advantage of his newly formed Nothing label by signing them.  After producing their premier release, Portrait of an American Family, and in an ingenious move afforded him by his newfound freedom under Interscope, Reznor elected to take Marilyn Manson along as the supporting act on the Self Destruct tour, as well as a circus freak show known as The Jim Rose Circus.  Fans of Nine Inch Nails\' hair-raising performances were also treated to brutal displays and horrific stunts by The Jim Rose Circus, and as well as a whole new genre of rock that they simply could not classify in Marilyn Manson.  Mansonâ€™s antics would become infamous, and the combined power of the opening acts fed the voracious craze that was rising up as more and more people were being converted into Nine Inch Nails fans.  As the year progressed and Nine Inch Nails were becoming a household rock name, Trent Reznor found himself with various opportunities that he had never been privy to before.  Tori Amos invited Reznor to sing back-up vocals on \"Past the Mission\" off her Under the Pink album.  He accepted and invited her to the Tate mansion for the recording process.  They quickly struck up a friendship and rumoured romance that would later mellow for widely debated reasons, though Courtney Love, widow to Kurt Cobain, has received the most popular vote as the cause of their split, and Reznor himself has indicated as such, though he has never specified.  Reznor was also invited to contribute a track to the soundtrack for the movie The Crow.  Again exercising newfound freedom muscles he didn\'t realize he possessed, Reznor consented and laid down another cover in the form of \"Dead Souls\", originally recorded by Joy Division.  Joy Division\'s lead singer Ian Curtis was yet another rock n\' roll suicide, and the connection was downright eerie when Brandon Lee, only son to famous martial artist Bruce Lee, died making The Crow in a stunt gone horribly wrong, and on his first major motion picture, no less.  Trent Reznor was inadvertently (or ingeniously) falling under a dark and sombre umbrella that would come to envelop his career and mark his persona. Early in 1995, Nine Inch Nails would again set out in support of The Downward Spiral, joined again by The Jim Rose Circus, but this time the short leg featured another newly signed act to the Nothing label, Pop Will Eat Itself, as Marilyn Manson had embarked on smaller solo shows and set to work on future material.  Reznor was approached for another project heretofore foreign to his artistic meanderings, when Oliver Stone asked him to produce the soundtrack for the upcoming film Natural Born Killers.  Reznor took the offer as a welcome relief from touring, and he and the production crew packed up studio, flew to Europe in early 1995, and holed up to assemble the soundtrack.  The album would feature a very eclectic line-up, including the doomsday musings of Leonard Cohen, angst-heavy L7, snappy and smooth rappings of Dr. Dre and Snoop Doggy Dogg, Jane\'s Addiction, Patti Smith, and Bob Dylan, among others, thus demonstrating Reznor\'s wide-ranging tastes and fearlessness when manipulating and combining various genres.  Nine Inch Nails also contributed several tracks to the effort, including a version of \"Something I can Never Have\" from Pretty Hate Machine, and a new song entitled \"Burn\", which paired with the violent, rage-infested theme of the movie perfectly.  Natural Born Killers, the film, was a smash for a number of reasons, not the least of which was director auteur Oliver Stone\'s name stamped on it.  In addition, the apocalyptic, serial killer ramblings had been adapted from a Quentin Tarantino screenplay, and it now had the sonic backdrop put together by the deft touch of Reznor.  After receiving critical acclaim and praise yet again, Reznor said of the experience that he would \"...like to do a real soundtrack.  I\'m interested in composing, whereas basically this was just editing.  I made a little souvenir of the movie but I don\'t really feel I\'ve created anything.  That, I would like to do... if we ever manage to stop touring.\"  Nine Inch Nails would go on to Australia, where they played just a handful of festival shows, before returning to the US.  Those who would accuse Trent Reznor of resting on his laurels during that period of Nine Inch Nails\' life would be horribly mistaken.  Immediately after returning from Australia, Reznor set to work producing Marilyn Manson\'s remix EP, Smells Like Children, which featured a cover of the Eurhythmics\' \"Sweet Dreams (Are Made of This)\".  The single hit the airwaves in 1995 and would arguably be Manson\'s biggest track both then and now, and it catapulted Marilyn Manson into the spotlight once and for all.  Reznor also set to work remixing and reworking his own material from The Downward Spiral, and these prolific remix sessions would result in Halo Ten, Further Down the Spiral, also released in 1995.  Reznor further displayed his uncanny production and mixing prowess by creating completely reinterpreted songs.  His sessions were so prolific, in fact, that Halo Ten was followed closely by Halo Ten v2, the lesser known little brother, containing a slightly different line-up than its predecessor and alternative mixes. After months occupied by these various side projects, Reznor was presented with a truly golden opportunity.  His success thus far had earned him the attention of one David Bowie, who extended an invitation to Nine Inch Nails to open during a fall tour of North America.  Needless to say, Reznor readily accepted.  So it was in the fall of 1995 that Nine Inch Nails once more struck out on tour, this time with David Bowie, giving them exposure to entirely new generations of fans that had not yet heard of them.  Nine Inch Nails\' now renowned stage show, including the grand finale of \"Hurt\", which would occasionally feature Bowie on stage alongside Trent Reznor singing alternative choruses, garnered Nine Inch Nails respect and awe in places that Reznor had likely never dared dream of believing possible.  Nine Inch Nails would then follow this highly successful stint with yet another few months of touring North America on their own, this time opting for smaller venues and playing with Helmut.  Finally, in December of 1995, several months shy of nearly two solid years of copious touring and production, Trent Reznor and Nine Inch Nails concluded their support of The Downward Spiral and took a break. Of course, \'taking a break\' seems to be a relative expression to Trent Reznor.  Building upon the incredible momentum of Marilyn Manson\'s newfound success with their supporting performances of Nine Inch Nails and the \"Sweet Dreams\" single, Reznor and Marilyn Manson retreated to Reznor\'s new home at Nothing Studios, which was housed in New Orleans, Louisiana. Reznor had purchased an old mansion, located just blocks away from vampire author extraordinaire, Anne Rice, and refurbished the property to house both his own personal studio and his new residence.  His former \"Pig\" Studio and the Tate mansion would eventually be demolished.  Reznor retained the front door of the Tate mansion, which once bore the infamous \"pig\" scrawled by the real Manson cult in blood, as a souvenir.  Nearly twelve months involved Reznor and Marilyn Manson himself forging a friendship and a professional team in the production of Manson\'s upcoming album.  Their recording sessions are the stuff of rock and roll legend, with tales rumours of rampant drug use, excessive alcohol abuse, and other tales of debauchery.  The result, however, was nothing short of sensational.  The fall of 1996 saw the release of Marilyn Manson\'s sophomore studio album, Antichrist Superstar, and much like Nine Inch Nails\' second album, it would mark the dawning of a new era for Marilyn Manson, both the band and the man.  Reznor\'s technical expertise was plainly evident on Superstar, and the singles from the album were both astounding and plentiful, including \"The Beautiful People\" and \"Tourniquet\".  Later that same year, Reznor was yet again approached by a director to work his magic on a soundtrack.  This time around it was David Lynch, who offered Reznor the production job on his upcoming film, Lost Highway.  Trent Reznor accepted, and gathered an eclectic, but viable collection of music for the piece, including newfound friend David Bowie, Angelo Badalamenti (Twin Peaks\' theme), The Smashing Pumpkins, and the German-born Rammstein.  Nine Inch Nails, of course, contributed three tracks to the album.  \"The Perfect Drug\" was the primary of the three, and the obligatory Halo (Eleven) was also released containing five alternate versions of the song.  Reznor would again team up with \"Closer\" video director Mark Romanek for a vivid and image-heavy, visceral video version of \"The Perfect Drug\".  Also included on Lost Highway were the tracks \"Videodrones; Questions\" and \"Driver Down\", both of which were not earmarked as Nine Inch Nails songs, but rather as \"Trent Reznor\", thereby marking the first time Reznor had ever \'released\' a track under his personal name, as opposed to his band\'s name, despite the fact that most of us know that \"Trent Reznor is Nine Inch Nails\".  Marilyn Manson, now at the height of his fame thanks in large part to Trent Reznor, also contributed several tracks, including a bloodcurdling cover of Screamin\' Jay Hawkin\'s \"I Put a Spell On You\" and an original track, \"Apple of Sodom\".  Marilyn Manson himself and Manson\'s bassist, Twiggy Ramirez, even appear in cameos in Lost Highway as porn stars.  This was also the first indication of a divide between Trent Reznor and Marilyn Manson himself and would foreshadow events yet to unfold.  Manson had allegedly wanted to produce the album, and was apparently livid that Lynch had handed the reins to Reznor.  Marilyn Manson went on to a sensational and controversy-laden world tour in support of Antichrist Superstar, and would later move on to Hollywood as the rift between Reznor and Manson grew.  Reznor continued on his workaholic regimen, and yet again got his fingers into unexplored territory when he was tapped to do the soundtrack for a video game.  Reznor would spend months in isolation writing the mood music to Quake, the sequel to the immensely popular game, Doom.  \"I would do something like that mainly because it\'s a hobby of mine, I appreciate the technology, and it\'s fun to work outside Nine Inch Nails once in a while.  Quake was fun because they didn\'t want hard-rock goofy music going through the game,\" explains Reznor, \"it was all about atmosphere at the time.\"  Quake\'s creator, John Carmack tipped his hat to Reznor\'s efforts by placing the NIN logo in the video game itself in the form of a logo imprinted on ammunition boxes.    In 1997, Trent Reznor again teamed up with David Bowie, this time to remix Bowie\'s single \"I\'m Afraid of Americans\" from Bowie\'s album, Earthling.  Reznor, along with the now staple production crew of Charlie Clouser, Keith Hillebrandt, Dave \"Rave\" Ogilvie, and Danny Lohner, set to work interpreting Bowie\'s music.  Again thinking outside of the box that the media would love so dearly to paint for him, Reznor called upon rapper Ice Cube to lend his vocal styling to one version of the single.  Reznor would also be featured prominently as the antagonist in the video for \"I\'m Afraid of Americans\", ostensibly appearing as said \"American\", replete with military style jacket and menacing glare, who stalks David Bowie through city streets throughout the video.  Thinking his production efforts for others at a halt, Reznor began the process of gathering his thoughts and artistic subconsciousness in preparation for Nine Inch Nails\' third studio album.  His work had yet to get off the ground, however, for Reznor was to receive yet one more call from an infamous musician requesting his talents.  Rob Halford, former lead screamer for \'80s metal giant Judas Priest and now openly gay, happened to be in New Orleans for Mardi Gras in 1997 with some friends, and when Reznor\'s studio was pointed out to him, he quite literally showed up knocking on Reznor\'s doorstep, requesting that he lend his ear to a demo Halford had.  Reznor accepted the demo, and after listening to it several months later, called Halford up and offered to produce and release the project on his Nothing label.  Reznor \"told me [Halford] that he had been listening to the music and he had a vision.  He could hear them in a different way.  And could we take them and break them down and build them up again, with his interpretation?\"  Halford agreed, as he \"knew that this was somebody who was a visionary, an incredible talent, and he was in his own exclusive world.\"  The result was 1998\'s Voyeurs by Halford\'s new band, called simply Two.  While the album did not garner critical acclaim or much commercial success, many hail it as yet another production masterpiece from Trent Reznor, who once more called upon Dave \"Rave\" Ogilvie to aid him at the mixing desk. Reznor had had enough of production and side projects that had too long distracted him from work dedicated solely to Nine Inch Nails.  Aside from the few songs recorded for various soundtracks, Nine Inch Nails had not fashioned any new material, and it was nearly four years since the release of The Downward Spiral, more than eclipsing the typical two-year window between most musical acts\' studio releases.  The time had come for Reznor to honestly buckle down and begin work on Nine Inch Nails\' next record.  Having learned in several years what many artists do not in their lifetimes, Reznor began the arduous process of applying this newfound knowledge and success to the task of recording Nine Inch Nails music once again.  The progression would be long and extensive, and Nine Inch Nails fans waited impatiently for their hero to come forth with new material.  Their patience would be tested indeed, as it took Reznor nearly two more years to create his next record, a double-album opus that would attempt to sate the now ravenous appetite of his fans. xBuBzx  '),(7849,3000240,'25,May,2004','2004-05-25 00:00:00','   I never, EVER want to hear what my sociology exam result is, I think if I did, I\'d go in to some psychotic fit. Anyway less said, soonest mended, as the dear oracle would say.  Anyway, I had a dream last night, I guess well because of what I was talking about with Justine after Troy about her life being one big mess until she found Jesus. I said fair enough and that it was good for her to vent her emotions at something if it meant that her mental well-being would remain sound, because I know If I went through what she went through, well from what I can vaguely anticipate, I doubt my brain could stomach it and just carry on as usual, and this is coming from a social fuck-up who\'s achieved more mistakes than is countable. I also guess that that\'s why I want to be part of her life, to understand these things and just to share and un-repress some things lurking around in the deepest unconscious, in both her and me. She\'s also really fun to be with as far, funny, un-repetitive and spontaneous (something which I adore in magnitude).  I\'m still though trying to get my head around Troy. Was it good or bad I\'m not sure. Usually after films I can tell straight of whether it was worth the ÃƒÂ‚Ã‚Â£5 or not, but I\'m swaying towards a thumbs up at the moment, my only concern was that Brad Pitt was cast wrong, maybe an actor with more of a prescence would have been suitable, although you can\'t really fault his effort in it all, he played it to his best ability and fair enough to him. Orlando Bloom was cast so so but also played it well and as for the rest of the actors/actresses (eg Helen of Troy), they all played it superbly. The only reason I\'m undecided about the film is the start and the length of it all. At 2 3/4 hours, It\'s alot to take in at one go and a 2nd viewing wouldn\'t go a miss. The start seemed to have a lack of a background somewhat with the ideas of Hericles being this awesome fantastico warrior man dude, just coming straight into it like a newcome dictator without much history explained to you, and it was evident on many other features about the film. Oh well maybe I think about this stuff too much, but not really a bad effort on the part of the film-makers. Ciao for now. xBuBzx  '),(7850,3000240,'25,May,2004','2004-05-25 00:00:00','   Short blog today I\'m afraid, but I\'ve just asked someone out by text, and the waiting for a reply is killing me *sigh* xBuBzx  '),(7851,3000240,'24,May,2004','2004-05-24 00:00:00','   Granny Smith came down to Molesey for the 1st time in 2 years yesterday and the instant she laid eyes upon me, she growled \"what the fuck have you done with your hair you stupid bugger\" which burst me into hysterics. What also is great about her is that she beats the living crap out of my dad (argument wise) with such truthful grace, which must be astronomically hard for a 91 year old. Anyway I\'ll explain more later . . . xBuBzx  '),(7852,3000240,'23,May,2004','2004-05-23 00:00:00','   urlLink    ...  Posted by  urlLink Hello     urlLink    Ahh the good old days . . .with hair and metallic glasses *sigh*. Oh and Harry trying to look all mean.  Posted by  urlLink Hello  '),(7853,3000240,'23,May,2004','2004-05-23 00:00:00','   For a Sister, I can hardly say that me and Joanne are particularly close. However it doesn\'t mean that I give less of a toss about her than say what identical twin sisters would about each other. So, tonight at 10, with me and my Mum minding our own business watching CSI (a rarity considering I hate CSI and \"TV with Mother activities\")I hear that she had collapsed from a rather sordid night of, from what I\'m told \"alcoholic blood transfusion\" I got quite concerned for  the old lump .  Ahh well, after much transparent activity, me and Mother decided that Morden accident and emergency was far too much of a distance to reasonably give a 2nd thought about, as we had came to the conclusion that: 1. She shouldn\'t be drinking at that age 2. The police cell was a far more viable alternative for her.  Anyway been a boring weekend so far, apart from I met someone new who I like sort of, and I might ask her out sometime soon I guess. Well I hope I get the courage anyway. Ciao. xBuBzx  '),(7854,3000240,'05,June,2004','2004-06-05 00:00:00','   urlLink    A message to the blonds of Esher\'s GCSE redo-maths class . . .  '),(7855,3000240,'03,June,2004','2004-06-03 00:00:00','   urlLink    mwahaha  '),(7856,3000240,'02,June,2004','2004-06-02 00:00:00','   I must be going through a backward 1 step then forward 2 steps kind of day or week, or if you want to be mischievous, life.  No; life would be something only depressed 14 year olds do (i.e. me a couple of years ago).  I really can\'t be bothered to clean out the flea ridden litter until the motivation returns - which suffice to say shouldn\'t be too quick in its glorious return.  Weekend is shaping up to be funky - I\'m meeting this ultra-cool girl and we\'re going to SEGAWORLD woo!!!!! I always remember that place from when I was little and it was so awesome that I thought it due time for a revisit, and if all doesn\'t go to plan, then we\'ll be of to the pub. I really like this one, she\'s so unlike her age lol - if that makes any sense. On Saturday I\'m going for a 8-3 shift at work and then traxlover has a big super-extreme happy/hippy party planned which she\'s asked me to grace with a tenner for the lovely drug bloke. He seems so sweet with the guns and death threats and all I bet he dishes out with regular consistency to all his \"chums\". Aha me and my warped soap box of complex opinionated bullshit. xBuBzx  '),(7857,3000240,'02,June,2004','2004-06-02 00:00:00','   I dare not really go downstairs that much, well it\'s filthy dammit!  They all come back tomorrow, and I\'ve already received the warning from Jo that it\'s coming, all 6 of them back with a vengeance. Hoorah. Well I\'ll tell you more later about it, too tired to post anything relevant at the moment. Ciao xBuBzx  '),(7858,3000240,'02,June,2004','2004-06-02 00:00:00','   urlLink    ahh tv in its finest . . .  '),(7859,3000240,'01,June,2004','2004-06-01 00:00:00','   If you know me fantastically well or not, the fact whether you know or judge me is really quite irrelevant, I\'ve decided from this weekend that I\'m going to sort out the whole sexuality issue once and for all. It\'s been manifesting itself (for want of a better term) for near on 6 years with boyfriends and girlfriends, flings and other stuff. I know I won\'t truly know what I am until I\'m say 20, but I can\'t stand the constant change of opinion, so I have to deal with this whole thing once and for all. Think of it as a never ending isle of sweets with unsupervised children, and waiting for them to hence reach a conclusion of true nirvana, and you understand my longitudinal dilema.  I don\'t think I love a girl I thought I did once. I guess if the weekend had a moral - that would be it. Well really it\'s been a 3 day weekend with the bank holiday and all, saw Anna and others, went to green fair and saw loads of folk I haven\'t seen in ages like Nikki and Tom and Bronwen. Was fun. Today however I\'ve done a 7-4 shift and tomorrow I have a 9-6 shit planned YAY.  Well I guess I\'ll just have to trust a lot more people with the whole sexuality stuff. Well the one poor sole who from what I gather regularly visits this blog is in for a strange literary sensation tonight when finishing this sentence which I\'m going to make you suffer by not stopping haha askdfjaskfafshaf *hides in embarrassment* xBuBzx  '),(7860,3000240,'12,June,2004','2004-06-12 00:00:00','   Not even properly Saturday yet and this 7 day weekend marathon preview to the 6 week/3 months/42 day weekend - is well under swing.  Thursday saw me try to get a grip on my deepest and darkest fear at the Greyhorse, and upon my immediate arrival, hearing the initial \"QUEER\" chant, brought me back to what I know and love - ignoring Ewan. If there is such a staunch ravage fucker as him, I have, and probably will never find one. This all bought me into deep, inhumanely pleasurable fantasies of a cigarette facial dis-figuation of the young chap, and his petite manhood. Alas I managed to control my urges enough to amuse myself on the image of the refusal of entry of over half of the parasites, and to enjoy an old acquaintance - Rowan\'s extremely funky quartet of funk and rock to its maximum. I also have to say that if Law ever decides to return for good, then this would be a greatly appreciated thing - good luck with Rachael and life down under in Fiery Dorset, and I can only hope you appreciate this overly korny paraphrase. Jen was also good for being there, I had to close the whole book on that school, and well The book was well and truly singed as far as I\'m concerned, and her being there helped no end, whether the whole ting was done indirectly or not, I\'ve never before managed to make friends with these kind of non-gothic wonders, before, and actually realizing how funky they all are, since joining Esher. Hell I even managed a half decent conversation with Chris Hodgson for crying out loud.  Anyway, tonight I went to another rock show with Jen, Law and Dan this time, in Richmond to see a band with such an exceptionally gorgeous lead singer, one who even Jens eyes were firmly planted upon. She can really sing, and after seeing her and the band, If only in a few gigs of late, they will definitely go somewhere if they keep working hard and go the way they are at the moment.  Tomorrow I\'m of to Camden to meet with Anna, and more importantly Jo, of which I\'m spine razor tingling nervous over. I guess if I just keep it cool and not drink or take too much and just do what I also am extremely good at, which is keeping quiet around large crowds, and so therefore lowering the scientific chance that I will make a twat out of myself, all should be dandy. Ah well, It\'s quarter to 3 now so I have by all accounts been speaking to you at this time for far a far too considerable time compared to a normal person, So I bid you - adjeu. xBuBzx  '),(7861,3000240,'10,June,2004','2004-06-10 00:00:00','   urlLink    Recent card to the oracle . . .  '),(7862,3000240,'08,June,2004','2004-06-08 00:00:00','   urlLink    The recent advertising campaign ahead of benefit reductions in Feltham.  '),(7863,3000240,'08,June,2004','2004-06-08 00:00:00','   Work was for once, of a bearable standard, with Becky not being that sadistic (although her traits will not die!). Hell even Katherine seemed quite funky for the most part, which I can\'t say is the norm for her from what I\'ve known, and I\'d love to get to know both of the better, but its just the way, that not only them, but myself come across towards each other which is the possible reason for the feeling of nausea we all experience when we encounter the joint presence of each other. I\'m sure they\'re lovely and all, but anyway . . .  God how I was scared at 2 am this morning when I thought it was the philosophy exam at 9, and to my utter relief, and utter adoration of Emaan from here on in - it was only the 3rd attempt to get a GCSE in mathematics and nothing of a more sinister persuasion.  I\'m getting quite girlishly excited about seeing Jo on Saturday now, (with mentions from my friends of bout\'s of giggling and skipping amongst other things lol) and I am NOT going to ruing the 1st opportunity with a girl in 2 years out of esteem naivete. I like her. She\'s so funky, and anyone who talks to her would utter the same warm friendly gestures as I would about her. Anyway - I\'m fairly tired now, so I must be of. Ciao. xBuBzx  '),(7864,3000240,'07,June,2004','2004-06-07 00:00:00','   Well the last few weeks have been really good, I\'ve a lover thingy which I\'m really excited about. Jen\'s party went well, apart from some (un-foreseen to myself) unfortunate adultery on a guy I know from college. I only have the 3 philosophy exams of which are of any quaint concern. So I must be of to START to revise for them for tomorrow morning! Lol. xBuBzx  '),(7865,3000240,'06,June,2004','2004-06-06 00:00:00','   urlLink    Another mesage to those most intellectual of blond girls, who love the feeling of a black 20 year old townietaking them against the wall and giving them a good seeing too.  '),(7866,3000240,'19,June,2004','2004-06-19 00:00:00','   I feel quite the fool at the moment, and in some sense utterly betrayed by myself and Jo about how the last few days have turned out.   Myself because well, I put all my eggs in her basket so to speak. I know I shouldn\'t have done, but it\'s been so long, I guess anyone would have done the same wouldn\'t they?   Why though would she have done certain things after the things and excuses she\'s used? What exactly is it about me that has suddenly make her act like this all of a sudden. I\'m beginning to understand why some men can never understand women. Us men do indeed have our flaws, but women are hardly innocent parties in it all.  Well, I guess there\'s a lot more time and people out there who are more suitable.   One thing that bothers me deeply is not to do with Jo, but I gave up men for quite a while because I lost faith in them, and now after the events of the last few days, I seem to have lost faith in women as well.   So who am I supposed to trust?   My friends have had a sordid history of letting me down so to speak, and I could tell by Frances\' actions that she was hardly going to see that Argentinean film out of the pleasure of my company, merely as pay back for me doing a few of those shifts. I happened to have figured that out just before I turned into central avenue bus stop to see her waiting there, \"tired\". She\'s is sweet, my gosh yes, but I couldn\'t bare to look at her because I knew that the whole evening was just fake. Anna for instance, called me from Spain of all places, not to ask how I was, but to pathetically beg for me to add to her phone credit, because she couldn\'t get of her backside to find a job of her own. I mean I love Frances and Anna to pieces, but they are not my true friends.  I even called up the one person today who I really (like Jen) thought that about a year ago I would never befriend his kind of people, Ian, to see if I could drown my pitiful sorrows with him. Alas even he couldn\'t show me how much of a \"best friend\" he claimed to be but a fortnight ago. No, I can\'t take things out on old E, the thought of him just makes me giggle, and I know he\'s just about the most genuine person I know anyway.  No one ever seems to make a great effort to see more of me because they enjoy my presence, they never have, and they never will; I accept this yes, but what other contenders are there?  I still don\'t trust either of my parents after the coming-outgate, where once upon a time, my Dad was my hero, and I don\'t see or care enough of my siblings to even try with them. Harry is too young, and making new friends at college wouldn\'t be worth it as I figure they\'ll all just be replicas of the one\'s I have at the moment.  Bah, maybe I\'m just too cynical, and my trust issues after Halliford were damaged irreversibly I guess, but I hate this waiting for a mutual loving companion, even at 17. My wish for a while now has been to settle down permanently, to properly help me recover from school. The parasites still haunt my dreams, which goes to show that they\'ll never leave me properly, so I put 2 + 2 together and a companion made 4. When will the confidence gamble ever pay of? xBuBzx  '),(7867,3000240,'18,June,2004','2004-06-18 00:00:00','   urlLink    The oracle recently gave me this picture in outrage, however I couldn\'t stop giggling . . .  '),(7868,3000240,'18,June,2004','2004-06-18 00:00:00','   I post to you at this most vivid and peaceful time, to show you to the extent to which insomnia has griped my epicenter over the past year.  I haven\'t slept properly from earlier than 11 at night until past half 8 in the morning for longer than I can remember, and things gradually seem to be taking me down with it. Bad news, motivation and concentration all seem to have suffered somewhat, to which I am seriously considering going to the doctor about. I doubt it is total insomnia, just maybe a partial nocturnal situation, a fear of the night per say.  So I won\'t drag this out upon you poor souls any longer, I just thought I might address this situation to what I feel is the truest thing in my life at the moment - this blog. I don\'t think it\'s all that sad to be honest, my ego needs a good kick up the posterior to get it going. xBuBzx  '),(7869,3000240,'18,June,2004','2004-06-18 00:00:00','   If there were no blond teenage girls in the world - would there be half as much stereotypical comments made about women in 20th century as there has been?  Just a thought . . .  Anyway, It\'s Friday night and I have absolutely nothing planned except writing offensive comments in an e-forum and maybe contributing to one if I feel especially productive. Gives me a good chance to get a bath and get the driver on my ignamous parasite of a home PC sorted. Well not  me  technically but John from work. However I figure that the ÃƒÂ‚Ã‚Â£10 payment should buy the claiming rights to that one and make me look and feel just grand.  College feels particularly strange at the moment - especially now I\'m a 6.2 whereas last week I was just a mere tiny 6.1 midget. The place too seems petite in comparison, and I guess that could be to do with the fact that, half the people aren\'t here; possibly because they\'ve realised they\'re too blond et al for college, or, in my opinion, they have been abducted or eaten by some mass government conspiracy. Well this is how much time I have on my hands these days that I actually formulate these premises/ denoumons .  Thankfully England, or Rooney beat Switzerland last night and so we\'re still in the competition thus far. Next up if the green room with Ian on Monday for the Croatia game (wherever the \"green room\" is I have no Idea, It\'s Ian\'s fault like most things in life). I guess we\'ll win narrowly and France will beat the Switz by even less and be given far more credit than us for there \"champion spirit\". Wasn\'t the world cup 2 years ago, and the limited victories over minnow nations in the Euro qualifiers enough indication, that this team, despite its great individual talent, has:  1. Got and over-rated manager, who will do worse than Pleat did at Spurs next season. 2. Become two years past its sell-by date?  Oh well, what do I know, I\'m only 17 and small and my opinions are worthless compared to people who are so unbelievably over-paid to play, manage, and analyze this game. I guess so. Has anyone noticed however, how absolutely vile the ITV football commentators were last night, messers Tlysdley and Robson? One, Robson (as my Dad would say) is the most boring man on Earth at the best of times when in the 50\'s; and Tylsdley . . . *huffs and scowls vigorously* Tylsdley just one doesn\'t shut up, and his over-emphasised cliche\'s and the mention of Manchester United on over 25 occasions in every match (including the Botswana v Zimbabwe womens friendly reserve match more recently) is just placing in some mental trauma for the 120 minutes I am forced to sit through his \"commentary\".  Anyway, enough of my ranting, I can see (yes through my wonderful internet remote viewing) that you are asleep by now, so I hope you, all have a funky weekend, and I most certainly will on Saturday with Jo Jo hehe. Ciao mon ami. xBuBzx  '),(7870,3000240,'17,June,2004','2004-06-17 00:00:00','   Here\'s the final part of the history of my favorite band Nine Inch Nails:    Swimming In The Haze: Evolution and the consummate perfectionist    Evolution, a fundamental tenet of the world as we know it, affects but is not necessarily grasped by everyone and everything.  Plants and animals continually adapt and evolve to meet the changing environment and all of the chemicals and pollution we pump into it daily, or they perish.  Popular media and art must evolve to mirror, embrace, or comment on our current state of social relevance is vital to art\'s very survival, or a disdainful audience will surely banish the unwanted art to velvet canvas and garage sale grab bags.  Darwin would have us believe even we as human beings have evolved from the most basic of life forms initially springing up from nothing at some point in the very distant past, or else our simian cousins might be surfing the Internet in our place, happily munching on a banana.  And as you no doubt have witnessed, dear reader, Nine Inch Nails, thankfully, has grasped the concept of evolution, though not so easily.  While many artists evolve as easily as one changes their clothing and thus remain in the public eye (see: David Bowie), and still others refuse to evolve and thus perish (see: hair metal), some artists struggle with altering their mindset in an effort to continue down the path of making genuinely meaningful and relevant music.  Following the enormous success of The Downward Spiral, its several singles and shocking videos that catapulted Nine Inch Nails into the mainstream (if their touring on Lollapallooza hadn\'t already), all of the live shows, soundtrack production, and umpteen interviews later, Trent Reznor found himself at a curious place.  Reznor does not pull punches when he describes Nine Inch Nails and himself following their last round of touring in support of The Downward Spiral, as a \"bloated, stupid thing that had become a parody of itself.\"  Reznor was becoming the very thing he had so long rallied against...just \"another retarded rock guy.\"  After viewing videotapes of himself and his band, and taking a personal inventory as a person, Reznor realized he \"was shitty to people, shitty to old friends and new people [he] met.\"  He had become what he had never wanted to be when he set out to be successful musician.  \"It didn\'t work,\" Reznor says of his state at the time.  \"I could pretend I fit into that role onstage every night Ã‚Â– the antics and nonsense.  But alone I had a kind of bottoming out.  I was losing friends; I had the woman who raised me die.  Nothing seemed to matter.  The last thing I wanted to do was write an album.\"  And so Reznor was forced to take stock of his values, self, and the project called Nine Inch Nails, in an attempt to evolve and move forward.  In 1997, when Trent Reznor began working on Nine Inch Nails\' third studio album The Fragile, he lost the grandmother who had raised him.  He had recently produced three Marilyn Manson albums, personally guiding the shock rocker to stardom and the top of the charts, only to then be snubbed by the jealousy of the Reverend Manson not long after.  Reznor found himself alone, without close family and friends, and ashamed of what he had become.  So the early stages of The Fragile and the writing involved did not come easily.  \"I just didn\'t want to sit down with a notebook and see what was in there,\" Reznor says.  \"I didn\'t want to uncap the fuckin\' cauldron and see what came out.\"  After speaking with several friends, including producer Rick Rubin, Reznor temporarily left New Orleans in favour of a cozy house at one of America\'s most visually stunning locations Ã‚Â– Big Sur, California.  With its majestic cliffs and awe-inspiring views, advisors touted the magical quality of the scenery, which might help lift Reznor from the depths of his depression and break his self-imposed writer\'s block.  \"That was terrifying,\" Reznor says of his time there.  \"I went insane and almost killed myself sitting at a piano trying to write like Tom Petty does.  I think Petty\'s great.  I also realize that when I sit down and do that it starts to sound like Billy Joel\'s The Stranger.\"  The several months spent there alone with a piano proved largely unfruitful for the troubled musician, but the venture was not a total write-off.  Of the one solitary track that Big Sur yielded and would eventually find itself onto The Fragile, Left\'s \"La Mer\", Reznor remembers \"sitting down and playing the piano and thinking \'How did I ever forget that this is what brought me joy?\'  How the fuck did that get lost in the mess?  How did I let that happen?\'  It\'s not doing interviews, it\'s not fucking live shows, not backstage passes and bank accounts of all that shit doesn\'t matter.  I did this because I love music.\"  Which is precisely the type of musician Reznor is Ã‚Â– a true composer.  While many artists, be they musically inclined or in favour of other artistic mediums, fail to cope and wrestle with their inner demons, Reznor would overcome his by simply making music. Several storied musicians helped Reznor with his writer\'s block, and coached him on how to battle the funk he was in.  David Bowie offered his advice even before Reznor fell into his dark slump, while Nine Inch Nails was still opening for Bowie in the late nineties.  \"We had long talks when we were on tour together.  He said to me at one point...\'try writing in the third person, because you\'ll find that you\'ll get yourself out of a hole that I was once in.\'\"  While Reznor wrestled with this suggestion, finding it difficult to do, he also considered the relevancy of how big Nine Inch Nails had become, and dealt with the personal demons that came with going from a cool, underground band to being well-known and successful, which possibly alienated many fans who didn\'t think such a band would be cool anymore.  \"I could relate to that,\" Reznor comments.  \"I didn\'t go to see the Jesus and Mary Chain after they got big.  So, I had to ask myself whether I should be trying to make my music more inaccessible, or should I just do what I think is true to myself.\"  It was actually Bono of U2 who offered the sage advice \"fuck those people.  Don\'t ever think that what you do is so elite that it\'s only for certain people.\"  And so Reznor soldiered on, the core of his very self slowly evolving. Reznor accepted the fact he had become emotionally compromised, and set out to do something about it.  He flirted with psychiatry (now that would make for an interesting file to read), but quickly found it unsuccessful.  Reznor also flirted with medication, sampling Effexor and Paxil, but discovered antidepressants and Nine Inch Nails did not exactly go hand in hand. \"I was in an abnormally positive mood all the time,\" Reznor states.  \"Everything was like, \'It\'s okay.  We\'ll get to that.  Whatever.\'  Part of what drives my personality got removed.\"   In the fall of 1997, Reznor decided the best course of action for battling his depression was to throw himself wholeheartedly into writing and recording The Fragile.  Returning to New Orleans and Nothing Studios, Reznor and producer Alan Moulder retreated to the confines of the Nothing compound for nearly two years.  Remaining almost totally aloof from the outside world, stepping out of the shelter of Nothing Studios only occasionally for the odd CD or to acquire an instrument, Reznor remained happily removed and ambivalent to the goings-on of the everyday world.  Instead, he turned his energies inward and simply tried to put all of the \"mouldy, decaying\" sounds in his head on tape.  Reznor poured himself into his music.  While many artists watch the clock, keeping tally of the costly hours involved studio time; Reznor knew no such troubles, situated as he was in his very own studio.  So he could focus on the music, without the typical distractions that might hinder the creative process.  And Reznor wasn\'t quick to dismiss the opportunity.  \"...We would try things that I would\'ve taken short cuts to arrive at in the past,\" Reznor says of the long hours spent in the studio.  \"Like, instead of a background noise off a movie which was what Downward Spiral was all about we\'d go find the original place and mic it.  Instead of a sample of a cello, it was like, \'Let\'s borrow one from down the street and see what happens if you put the wrong strings on it.  Whoa!  Check this sound out!\'  We had space and time to try that kinda stuff.\" As Reznor and Moulder neared completion of the album, they found themselves with over forty tracks, and the revelation the music they had made in their self-imposed exile from the outside world might not be at all like anything the public currently considered palatable.  In an industry churning out bubblegum, midriff-baring pop music, The Fragile, Reznor realised, did not resemble anything being produced at the time.  Further, Reznor and Moulder both believed that after being holed up with the same music for such an excruciatingly long recording process, objectivity had been compromised.  \"...\'La Mer\' was one of our favourites, and it was gonna be the first song on disc one.  Then it was \'Nah, let\'s make it the first song on disc two.\'  Then \'Nah, let\'s make it the first single!\'  \'No, wait!  Let\'s call the album La Mer!\'  And you approach a song like that with reverence you cherish it so much, you\'re not being objective about where it should go,\" Reznor stated as they came to the end of The Fragile\'s sessions.  Sensing yet another tailspin into countless obsessive-compulsive tangents, the duo put in a call to producer Bob Ezrin, mastermind behind one of Reznor\'s long-time favourites, Pink Floyd\'s The Wall.  After convincing the legendary producer to come and spend a week in New Orleans, Ezrin disappeared into the depths of Nothing Studios and three days later, he emerged with reams of notes.  In a very businesslike manner, he proceeded to tell Reznor what the weak spots on the album were, the highlights, and basically, the type of record Reznor had just made.  Excitedly, Reznor put on the compilation made by Ezrin, and found it to be...\"terrible!  Halfway through the first side I was daydreaming, I was looking out the window,\" as he says of the first attempt at finalizing The Fragile.  Needless to say, it wasn\'t easy for Reznor to tell someone like Bob Ezrin that he had failed and his track order just did not suffice.  Further, Reznor began to doubt himself, wondering if the past two years spent in isolation had all been for naught.  Ezrin took the feedback in stride, however, and returned to the drawing board.  On the final day of Ezrin\'s visit, Reznor came into the studio in the morning to find some new discs sitting on his desk.  Of the morning, Reznor says \"I wasn\'t particularly in a bad mood, but I sure wasn\'t in a good mood.  And I wasn\'t very optimistic about what I was gonna hear.  So I sat down right here and I put it on.  I didn\'t have a list of songs, so I didn\'t know what was coming.  And halfway through the first CD was when it hit me, when I started crying.  And I was thinking, \'if it ends now, it\'d be perfect.\'  Then I immediately put the second CD on thinking, \'Well, this can\'t be good because all of my favourite songs are on the first CD.  I uhhhh, think what else could be left?\'  But it finished and I was stunned like, \'Fuckin-A!  Alright!\' I couldnt believe it!  I had goosebumps!\"  Poker-faced, Reznor called his last minute hope into his office, and had Ezrin sit down on his huge leather couch.  With a straight face, Reznor told him that he had done it, and Ezrin replied simply, \"I know.\"  Reznor responded: \"Well, why did you fucking well wait \'til the last day then, smartass?\"  And they quickly embraced, for it was good.  It was very good, indeed.  In the fall of 1999, Nine Inch Nails and their new album, The Fragile, were quickly back in the limelight.  Debuting at Number One on the Billboard charts, Halo Fourteen moved 250,000 units in its debut week, before beginning to slip back down the charts.  Accompanied by the first single, \"We\'re In This Together\", The Fragile was a huge success with fans of all respects, but the album never quite achieved the success that its predecessor, 1994\'s The Downward Spiral did.  Speculation looked to several reasons, including The Fragile\'s length, or its not-so-radio friendly singles such as \"Into the Void\", and \"Starfuckers, Inc.\"  Still, The Fragile was heralded as sonic mastery, and everyone clamoured to lift its mastermind, Trent Reznor, to their shoulders as the saviour of music.  Numerous publications put Reznor\'s scowling visage back on their covers, including Rolling Stone and Raygun, and Spin magazine went so far as to rank The Fragile as \"Album of the Year\".  While The Fragile did not enjoy the commercial success of the typical pop thoroughfare, the musical community happily embraced it, injected as they were with the newfound sense of hope and vitality that came with Reznor\'s double-album opus. The next step was a logical one.  Tour.  And so, slightly evolved and still inwardly smarting from his behaviour during The Downward Spiral touring years, Trent Reznor gathered together his long-time band mates, plus or minus a few members.  Keyboardist Charlie Clouser and bassist Danny Lohner returned to the NIN camp, and Robin Finck was re-called from his brief stint with Axl Rose and Guns N\' Roses\' long-awaited Chinese Democracy.  Former drummer Chris Vrenna had parted ways with Nine Inch Nails in favour of his new personal project, Tweaker, so in his stead Reznor recruited the relatively unknown Jerome Dillon.  After four years away from the stage, Reznor wisely planned several warm-up gigs in and around the world, including locations such as New Orleans, New York City, the Bahamas, as well as a memorable Grammy performance.  After a few private and small-audience sessions, Nine Inch Nails were back in shape, practiced on the new material, and they set out to assault the world with their unique slice of sonic boom.  Trent Reznor and Nine Inch Nails would go on to tour the world once over, including an Eastern global tour dubbed Fragility v1.0 and a Western hemisphere stint entitled, logically, Fragility v2.0.  As per usual, the visual act that was Nine Inch Nails was almost ridiculously impressive, and the tour was consistently ranked as one of the best live shows of the year, and even received said nod from Rolling Stone.  Reznor had taken the angst, intensity, and spectacle of Nine Inch Nails\' live trademark, and further morphed it into a more eye-enticing display.  Fragility became the live manifestation of the Nine Inch Nails experience, and where the music had evolved and taken on a more mature and sombre attitude, so too, did the live concert.  Three large LCD displays majestically dropped from above the band early on in the show, and distorted, beautifully simple images flashed across the screen behind a slightly more subdued band.  While hardcore favourites such as \"Head Like A Hole\" and \"Wish\", as well as newcomers like \"Starfuckers, Inc.\", continued to whip the crowd and band into a frenzy, more restrained pieces such as \"La Mer\" and \"The Great Below\" were thrown into the mix to provide a more all-encompassing musical experience that truly ran the gamut of emotion.  The Fragile would see several singles and videos released, though only two would actually become official Halos.  Halo Thirteen, the single for \"The Day the World Went Away\", predecessor to the actual release of The Fragile, saw the ravenous appetite of Nine Inch Nails fans whetted by just three tracks.  Two tracks were actually pulled directly from the upcoming record, including the title track and \"Starfuckers, Inc.\".  Unusually, the release carried only one \"Quiet\" mix of \"The Day the World Went Away\".  Halo Fifteen, released after The Fragile, was the single for \"We\'re In This Together\", and numerous versions of the Halo were released internationally.  Some contained further remixes of \"The Day the World Went Away\", while still others carried two bonus tracks previously available only on the vinyl release of The Fragile, \"10 Miles High\" and \"The New Flesh\".  After the Fragility tour, as is tradition, Reznor returned to the studio and revisited his masterpiece, and emerged with Halo Sixteen, Things Falling Apart.  The obligatory Nine Inch Nails remix album saw Reznor hand over the reins primarily to those close to him, rather than complete all the reworking himself.  Old collaborators such as Keith Hillebrandt and Dave Ogilvie were recruited yet again, and band mates Charlie Clouser and Danny Lohner also contributed, among others.  A remix of Gary Numan\'s \"Metal\" was also thrown in as a tidbit of new material.  This hugely successful period of Nine Inch Nails life also brought with it Trent Reznor\'s official foray into the Internet.  Both nin.com and nothingrecords.com were launched, not to mention various other sites, such as objectmerch.com, carrying Nothing\'s official merchandise, and sub-sites for NIN releases, including andallthatcouldhavebeen.com.  Always the consummate perfectionist, Reznor\'s web sites were visually stunning and completely engrossing, and many hidden links and gems were cleverly disguised therein.  Always committed to the fans, Reznor also allowed the download of exclusive, Internet-only available material, including remixed tracks, live sequences from the Fragility tours, and pics.  In short, the Internet versions of Nine Inch Nails and Nothing Records were just what you expected from Trent Reznor. Following on the success of the Fragility tours, Reznor readied for the release of the first live material available from Nine Inch Nails.  Whilst touring, Reznor had had the foresight to plan the release of the material, and Reznor and crew routinely handed out camcorders to crew and fans in an effort to obtain more raw and true to life live experiences that aren\'t typically captured by the highly-polished, sweeping cameras of big budget live music.  Upon returning to New Orleans, Reznor and team set to work at their trusty Macs to put together and attempt to capture the live Nine Inch Nails experience on both CD and DVD.  To be the first DVD released by Nine Inch Nails, Trent Reznor wanted to ensure that if it was going to be done, it would be done right.  Halo Seventeen, And All That Could Have  Been, was just that.  After various delays, it was released in early 2002 to much fanfare.  Both the CD and DVD release truly captured the essence of Nine Inch Nails live, and further reinforced their reputation as possibly one of the world\'s greatest live bands.  Chock full of hidden \'Easter eggs\', including unreleased videos and the live segment showing Trent Reznor and estranged friend Marilyn Manson reuniting on stage in triumphant renditions of \"The Beautiful People\" and \"Starfuckers, Inc.\", And All The Could Have Been truly showcased the magnificence of the live performances given by Nine Inch Nails and Trent Reznor.  As is typical with Nine Inch Nails, several different versions of Halo Seventeen were released, including a deluxe CD package containing a bonus disc entitled Still.  Also previously available individually via nin.com, Still featured a very stripped-down, raw musical version of the Nine Inch Nails product.  Months earlier, whilst on the North American tour of Fragility, Reznor and Nails stopped into a Chicago radio station and played an unplugged set to a very small and no doubt dumbfounded audience.  While the CRC Sessions, as they have so been dubbed, were never officially released, Reznor did revisit the concept following the tour, and Still is the result.  Free of the countless samples and harsh electronica that defines Nine Inch Nails\' sound, Still featured just Reznor with a piano and guitar, along with several band mates equally as unplugged.  Haunting, raw, and viscerally beautiful, Still took many fans by surprise, showcasing the musical talent of a man often hidden behind a wall of cacophonous sound.  And now, years later, Reznor has retreated yet again into oblivion.  Since the release of And All That Could Have Been in 2002, Nine Inch Nails has not released anything, nor has much been seen or heard from the NIN camp.  Rampant speculation and the always churning rumour mill that is the NINternet have led to suppositions as to possibilities for the future of Trent Reznor and Nine Inch Nails.  For years, the Tapeworm project, involving any number of musicians including Maynard James Keenan of Tool and A Perfect Circle fame, Atticus Ross of 12 Rounds, Zach de la Rocha formerly of Rage Against the Machine, and Trent Reznor himself, continues to remain mired in obscurity and legal battles between the various artists\' record companies.  While band mates of Reznor have moved on to side projects, such as Danny Lohner\'s recent production of the soundtrack to the film Underworld, nothing seems to be stemming from Trent Reznor\'s corner.  Nin.com has remained virtually unchanged for several years now, with the exception of very muddled, mysterious images and quotes loosely centered primarily on Ursula K. Le Guin\'s novel The Lathe of Heaven, that slowly appeared (and subsequently disappeared) over the course of several weeks in the fall of 2003, until nin.com finally closed its \'doors\' several months ago in favour of a single graphic that prevents entry to the site.  Nothing Records has been equally unproductive, with little to no releases from long overdue artists like 12 Rounds, and the website, previously interactive and relatively interesting, has advertised the same upcoming releases for, quite literally, years.  What will the future of Nine Inch Nails hold?  With Reznor fast approaching his forties, he is not the young man full of piss and vinegar, unbridled hatred and animosity, and confusion and self-loathing that he was nearly two decades ago when he threw his hat into the musical ring.  Reznor will soon be negotiating middle-age, and with that comes a whole new host of inner speculation and soul searching.  The typical collection of thoughts on society, religion, life and death,  blah, blah, blah, undoubtedly still exist in Reznor\'s innermost being, but a new faction of concerns have no doubt surfaced, including love, family, and the possibility of children.  How will this new, older and wiser Trent Reznor interpret the world?  And more importantly, how will he choose to comment on it via the medium of music?  While unfounded rumours continue to spread, magazine articles become fewer and far between, and all but the most dedicated fan sites dwindle into obscurity, NIN fans now in the know as to typical periods of downtime between releases yet again patiently await the next album to, ahem, bleed through. xBuBzx  '),(7871,3000240,'15,June,2004','2004-06-15 00:00:00','   Things with Jo are going pretty swell and slowish which are really good. I\'m glad I\'ve got something stable finally - and we\'re of to the Imax cinema near Waterloo this weekend for some grade A geek eye wear movie viewing and Grade A* company with each other I guess if that makes any sense.  Anyway, It\'s England v Switzerland tonight, and I\'m of to Wimbledon probably with Kate for and hour and a half of getting to know her/whilst drinking/innocent flirting with her I guess before some very graphic and audible swearing at the plasma screen - oh YAY. IF we lose this game, anything at all like the last game then I may go into some psychotic manic depression and alloy a knife to my wrist. Surprisingly, I haven\'t got bored of any of the football games yet - must be an indication as to what the state of my social life these day . . .  Well I\'m back at college at the moment, and It\'s a \"back to work\" day, just some PC nonsense. Been ok I guess and It\'s no work and no pressure of exams on my mind (all be it apart from the last maths one). This week should be ok, nothing major is planned too much apart from seeing Jo which is more major I guess apart from an average movie trip. Speaking of movie trips - I went to see a film called imagining Argentina with the amazingly sweet Frances from work. It was so much better that I thought it would be, and Antonio Banderas played it really well. Anyway must be of. xBuBzx  '),(7872,3000240,'15,June,2004','2004-06-15 00:00:00','   urlLink    They\'re happy - do you know why? That\'s right: BuBz has made more than 100 posts for you poor with standing people, and I\'ll do a 100 blog post review tonight - AS I\'M SO SAD!!!  '),(7873,3000240,'14,June,2004','2004-06-14 00:00:00','   urlLink    Sad I know - but the original Thomas the tank engine courted by my nephew on Saturday. Damm him how did he get it before me . . . )-:  '),(7874,3000240,'14,June,2004','2004-06-14 00:00:00','   Well any need for MAJOR revision is not thankfully at an end. Well It\'s not like it I did much anyway. Philosophy was three hours of laughing at amazing suspense at how little I actually know about philosophy. Half of the time I was staring at Ed beside me -happily completing the 6 pages for each paper - and I was happy too, staring at him was lovely considering just how awesomely cute he is (-:  Anyway, Saturday went well. Well I say well but this poor guy I know got stabbed in the back literally by Somalians - and he looked like the guy you can see if you go to the link in my name. I felt really sorry. Mind you if you throw huge black townies into Camden Lock - it is highly unlikely that you would come out in the long run un-scaved. Anyway the guy is now in hospital as far as I know and won\'t die at least . . .  Jo seemed funky, quite lively; but funky . . . . . . . . . . . erm oooh weathers nice hehe. Well if exurberance is your forte, then she\'s your type of lady. Ahh what am I saying, she\'s lovely really, and I can\'t wait to come see her again. I\'m back into my favorite blogging place at the moment with the nicest keyboard your ever going to meet, and so I\'m not making half as many errors whilst typing to you here. Hmmmmm maybe I could thieve it perhaps. It\'s just so soft and cleansing on my fingers, which could I guess be used for a nasty analogy if you have my two track mind - taboo\'s and sexual orientation (more than just one track eh Tammy).  Oh my as well, that football match almost destroyed my mental wellbeing yesterday. God how can anyone (never mind England) lose it after 92 minutes I\'ll never know. Well it\'s Switzerland on Thursday which we basically have to win to stay in it. Hopefully it will be a walk in the park but you never know with these things.  Not much really to do for the rest of the week except do some overtime if any were available to me, so I guess I\'ll just have to attempt a diet or something to keep my mind in full lubrication as it were. Well that or write. Or figure out what I\'m going to do with the rest of my life once I discover the trio of E\'s in August. Well my dream career would being well paid to do this blog everyday. If only blogger cared enough to give me the chance - God knows how I\'d love to entertain/piss of the masses. As my good friend Ian would say, it would be pretty \"cool\" - with an uninterested Daria type facial expression.  Currently the LRC is bare empty (for the love of good I couldn\'t think of a non-townieish term other than bare - I do apologize sincerely) with most people just on here to stare at the porn because they can\'t afford a home PC i.e pikeys  Anyway, unless something comes up in the next 10 seconds remotely meaningful to write, I\'ll have to say ciao. Oh darn it I can\'t think of diddle squat *sigh* xBuBzx  '),(7875,3000240,'26,June,2004','2004-06-26 00:00:00','   She really should have just held on to the cap in that mank chamber of a vagina in 1986, and not have let go so easily to the ogre\'s advances. Well my justified view anyway.  I try my hardest to give that child some discipline, anyone in there right mind would see he gets his posterior kissed on a regular, guilty conscience\'s basis. Also that he, even for a 5 year old, is one of the most cocky, self righteous, obnoxious, temperamental drama queens walking around.   So the argument started with me doing what regular people do and show Harry up and tell him of in the  proper  way, and I get \"disciplined\" for it. They then come steaming up when I was first attempting to manuscript this blog, and of we went. Ones really brought from deep inside the closet were thrown in all directions, me with my mind games, them with there pitiful double act of \"official ownership regulations Ltd\" and it boiled down to both parties telling each other where the other could stick it, if they didn\'t like the opposing proposition. All got rather tense, violence planned, but, fortunately for my mental well being and my psychotic nature with that cliche hypocrite, it failed to materialize.  Anyway, I need some rest-bite from that tomfoolery, so I\'ll probably be of to see Shrek earlier than planned at Wimbledon with Ian. He really makes me put things into prospectus, well them anyway, and I need a good rat about things and he is the ultimate agony aunt. God bless his soul. xBuBzx  '),(7876,3000240,'25,June,2004','2004-06-25 00:00:00','   urlLink    coming over soon (me hopes) . . .  '),(7877,3000240,'24,June,2004','2004-06-24 00:00:00','   England blew out on the important occasion yet again last night, 6-5 on penalties to the hosts Portugal. I dare not go into too much detail, I\'ll just ruin that new pacemaker of mine into extreme anguish, as well as my syntax. Just, well, I don\'t think the goal shouldn\'t have been allowed, and I can\'t wait for Shaun Wright-Phillips to take over the reigns at right midfield when David Beckham retires, he was abysmal, and his 2 penalty misses in the competition would, if have been taken properly, have seen us through.  Anyway, quite a few people are out of the local area at the moment, including Thomas and Jen and Frances and Anna to name a few. So tonight I guess I\'ll just be at home watching the start of Glastonbury on TV (whilst trying to spot gorgeous Thomas hehe) and mulling over who Chelsea should buy in the remaining close season now that the players are available to talk transfers, as the come hither in drips and drabs. My bet is Rooney and Joaquin, and if not Joaquin like expected, then Shaun Wright-Phillips.  Apart from those couple of things or so, there hasn\'t really been much going on at my life at the moment, well for this week anyway; I had enough time last to make the blog even more professional with even more links (well this happening after I destroyed the HTML language yet again trying to add extortionist links in my crazy way). Now my plan is just to pester Blogger until they allow free funky \"splash introductions\" like the one\'s you get for accomplished rock bands or the Chelsea FC web page (click link in my name). Ooooh I guess I do have another plan to do this weekend, to go and see Ian and get pissed, and think long and hard about coursework for all three subjects (bar philosophy who can\'t get the finger out and just  GIVE  it to us) and finally to see whether I can broaden the style/vocabulary and general literary verity in the blog writing, and so I envisage quite a few more literature issues coming, one\'s which will be a just and deserved carry on from the  Swimming In The Haze  series. Well this and in addition too - official Chelsea management game 2003/2004, and convincing the football faithful that (yes this guy\'s name being mentioned again) Shaun Wright-Phillips is worth buying as a right-midfielder. Sure to be a roller-coaster ride *sobs* . . . xBuBzx  '),(7878,3000240,'24,June,2004','2004-06-24 00:00:00','   With my hand nerves ever deteriorating, I came to contemplate some rather twisted demeaning denoumons on the till last evening. The majority included how to destroy Katherine and Becky in one swift execution, the other, which was the far more horrific of the two, involved me extensively analyzing what I would do If the shingles in my wrist got worse and I developed a motor neuroneish disorder, or worse, a limb castration. You can see how my life\'s priorities are just  swell  at the present time.  Anyway on a lighter note, I\'ve changed the format in the blog in the past week, with the help of Jen, to, in my opinion, give it a touch more professional appeal. It\'s more defined, colorful, with more links than before. Well it is my eyes only, even if they are more corrupted than patients on prosaic . . .  Finally got round to purchasing a new Â£30 (ooh BuBz really splashed out there hasn\'t he) \"JanSport\" hiker/skater bag. It\'s something I always quite craved when I was around the age of 14, and, I just decided to see whether those dreams of a \"funkier look\" would be more justified even at a later age ofadolescencee.  The new NIN album \"bleedthrough\" has been playing on my mind somewhat. It came upon me during an Eminem sing-along near the couches (a.k.a the sofa crew), and I\'m really desperate for Trent to show pop music for what it really is - meh I can\'t be bothered to go into it now; only 3 years until I die slowly from terminal illness, and not nearly enough time to divulge into great detail. Well, must be of now for Psychology so . . . xBuBzx  '),(7879,3000240,'23,June,2004','2004-06-23 00:00:00','   urlLink    The absolute gobsmackingly fittest person on Earth . . .  '),(7880,3000240,'22,June,2004','2004-06-22 00:00:00','   urlLink    Harry trying to look cute with mother and Pauline . . .  '),(7881,3000240,'21,June,2004','2004-06-21 00:00:00','   Something came up to me today, of I felt it more like, which has been manifesting itself over the past week; and it\'s rather worrying to be honest.  When I clench my fist or strecth my fingers to a certain parabola, the seem to lose all feeling literally. I mean it\'s scary. It\'s been getting progressively more apparent and hence far more noticeable and of a far greater degree of concern has succumb due to it. I don\'t really ever like going to the doctors about things, but if things keep going narrating themselves in this consistent fashion, then I may have no other alternative.  What worries me most is that, today especially, if someone would have asked me to carry something, I would most probably have dropped it, or not even clench it in the first place. At least it\'s something not typical I need worry about now, and not tiresome emotional issues. Good to be back *gulps* xBuBzx  '),(7882,3000240,'21,June,2004','2004-06-21 00:00:00','   Celebrity \"Woodstock\" Deathmatch     This time at Woodstock . . .  Johnny Gomez: Ladies and Gentlemen, Nick Diamond and myself are right in front of the set and we\'re live at Woodstock where a scheduled Celebrity Deathmatch is to take place anytime now between Nine Inch Nails\' Reznor Reznor and limpbizkit\'s frontman Durst Durst.  Nick Diamond: No one\'s performing at the moment on the stage, but the fans are all pumped up for the fight.  It looks like Reznor and Durst are going to determine everything tonight.  And we can hear the music being played... It\'s \"Hot Dog\".  And Durst comes out from the side and the crowd goes wild.  Apparently, he looks exhausted from a performance he might have had earlier, but he\'s never too tired for a good fight!  Johnny: What\'s this we\'re hearing?  A techno tune, only increasing in sound... it\'s \"Starfuckers, Inc.\"!  And that\'s Reznor Reznor.  The people are fanatical, roaring \"Nine-Inch-Nails! Nine-Inch-Nails!\"   Nick: We\'ve come to know that this is his first public appearance in almost four years!  Johnny: Both of them onstage together for the very first time!  The lights are flashing and the music still audible, but where\'s Mills Lane?  Oh, there goes the referee right by us, and he\'s getting up onstage.   Mills Lane: Now you two, you obviously know the rules, so look around you and take notice of all the instruments lying around.  Ready or not, let\'s get it on!  : And it has begun.  Nick, this going to awesome, I haven\'t scene a Deathmatch venue of this kind before!  Nick: Yeah.  This sure is going to be one good fight, and because of their experience onstage they\'ll know exactly how to use the tools!  It\'s time to go stage side!  Reznor, wearing his trademark \'Evil Inside\' t-shirt and khaki shorts, walks around the stage; it looks like he\'s thinking about something.  Maybe it\'s nostalgia thatâ€™s getting to him.  Or maybe he\'s just familiarizing himself with the set for more serious reasons.  On the other side, Durst seems to be signing autographs?!  But now he sees Reznor glaring at him so he goes towards Reznor with a mic.  Durst: Reznor \'Nine Inch Nails\' Reznor, we finally meet to end this stupid feud of yours. And it\'s time to show everyone who\'s the best man at Interscope Records!  Reznor takes the mic away from Durst and kicks him in the midsection.  Reznor: You wanna fuck me like an animal, eh?  You like to burn me on the inside?  You like to think that I\'m a perfect drug!?  Reznor tosses the mic aside, picks Durst up and throws him across the stage and into the drum kit.  Nick: There\'s some action!  Reznor\'s infamous habit of thrashing his own instruments onstage may very well prove useful tonight!  But look at Durst... he\'s up already but what\'s he taking out there...it\'s some sort of a slimy substance and it\'s star shaped?  Durst throws the object at Reznor and it automatically sticks onto Reznor\'s face.  With Reznor\'s face now covered with that slimy object, he can\'t see anything and he\'s screaming and running around in circles.  Johnny: Oh my God, it\'s a poisonous starfish!  What a move by Durst!  Nick: What\'s Reznor feeling right now, Johnny?  Johnny: Well, it appears that Reznor is currently in great pain due to the venom released right on his face by the starfish.  With great swelling and redden areas of skin, the one-man-band member requires serious medical attention right now if he\'s to even survive!  Reznor is found falling over instruments and scraping his face against the stage amplifiers.  Durst, on the other hand, starts taunting and posing before the crowd in triumph.  Nick: Look at Reznor...he\'s got the starfish off of his face!  Reznor can be seen holding the remains of the starfish.  But his face looks quite ugly, or rather it\'s just got the pink slime and/or colour of the obliterated starfish; Reznor\'s still trying to gain visual consciousness.  Most of the crowd are in awe, but some are chanting unknown verses of various limpbizkit songs while cheering for Durst.  Johnny: What is Reznor thinking?  He should just go and push Durst off the stage!  Reznor takes out a double barrelled gun.  By the time Durst gets back to Reznor, Reznor is found pointing the gun at Durst, and loading the gun with large nails.  Nick: What\'s that? It\'s a huge gun and look what he\'s putting in it!  Johnny: Apparently they are nine inch nails, quite literally, and he\'s holding a nail gun!  And the nails have some sort of sticky material all over them.  Nick: Oh, he must have used those nails to scrape off that starfish!  Johnny: As messy as the match has already been, with these two it seems like there can only be room for more!  Reznor, having loaded the double barreled machine of a gun, starts aiming all over Durst\'s body while Durst, with his legs crossed, tries to move in different directions only to find out that there\'s no way to escape Reznor onstage.  Johnny: Reznor can play on like this throughout the night; he\'s got Durst scared straight.  But where\'s Mills, Nick?  Nick: Good question, where is he?!  Johnny: There he is, right at the end of Durst\'s side of the stage... and he\'s lying unconscious!  Out of nowhere, Richard Patrick appears, jumps on stage, and kicks Reznor in the back.  Reznor falls but, luckily, is still holding the gun.  Patrick kicks the gun away from Reznor and takes it into his possession.  Now holding the gun, he points to the centre of the stage.  Patrick: You and your stupid sets!  It\'s not even worth taking a picture!  And Patrick starts shooting nails at the instruments.  The nails only ricochet and deflect back onto him.  He\'s fairly injured but not out, yet.  Nick: Johnny, this is quite the occurrence.  But who\'s going to benefit from all of this!  Johnny: I don\'t know, but look over there! It\'s limpbizkit\'s ex-guitarist, Wes Borland!  Borland jumps onto the stage from amongst the crowd, holding a guitar of his own make.  He slams the guitar on Durst\'s head, who had been staring at Borland in disbelief.  Borland then grabs Durst and drags him off the side of the stage.  Reznor, now up, runs over and past Patrick\'s body and jumps at Borland.  Reznor: Durst\'s mine!  Borland takes a shot at Reznor with the guitar, but misses.  And in return Reznor quickly picks up a keyboard that was fortunately laying next to him and breaks it over Borland\'s head.  Johnny: Oh, did you hear the crowd as Borland\'s head exploded?  Nick: Indeed, Johnny... that was nasty!  With Borland out of the scene, Reznor drags Durst to the centre where Patrick is apparently stuck to the floor.  Patrick: Ah, Reznor!  Help me out man!  Reznor throws Durst over Patrick\'s body, pins Durst down with his leg, and picks up the nail gun.  Nick: Reznor just won\'t answer Richard Patrick\'s screams for help!  Filter\'s singer is in serious pain!  Reznor points the nail gun at the two helpless opponents and shoots, only to find that the gun lacks ammo.  Reznor: Damn it, Richard! You wasted all the nails...  Johnny: Nick look, Mills is up and he\'s going straight at Reznor!  Mills: Now what\'s going on here? You wanna finish off Durst with that gun?  Johnny: And the crowd goes silent...  Mills: I\'ll allow it!  From behind referee Mills, a box of nails comes sliding over to Reznor.  Reznor starts staring at Mills, trying to look around him to determine where the nails came from.  And out of nowhere appears Marilyn Manson.  The crowd grows even louder.  Johnny: What else could a crowd ask for?  They\'ve seen it all!  Nick: Look at Reznor smile and nod at Manson.  Reznor reloads the gun and holds it up.  Durst: Reznor you...!  Before Durst can even complete his sentence, Reznor fires shots at Durst, and even finishes off Patrick.  Reznor hands the gun over to Mills.  Mills: Hmm... nice gun.  And the winner is... Trent Reznor!  With \"Starfuckers, Inc.\" playing again, Reznor and Manson leave the stage and walk off into the crowd.  Johnny: What a night this has been.  With interferences from three other artists and Mills unknowingly falling unconscious, this has truly been a whole different Deathmatch altogether!  Reznor was deservedly victorious at the end.  Nick: Yes, Johnny.  It looks like Reznor can start concentrating on making music again.  And who knows, maybe Reznor and Manson are headed to Reznor\'s recording studio right now!  Johnny: Who knows?  But we do know that this concludes a rather incredible edition of Celebrity Deathmatch. Live from Woodstock, I\'m Johnny Gomez.  Nick: And I\'m Nick Diamond.  Johnny: Good fight...good night! xBuBzx  '),(7883,3000240,'21,June,2004','2004-06-21 00:00:00','   College seems fun now I\'m a big old 6.2. Not like anyone else being in my shoes feels different either but . . .  England was really good last night, at the home of football (Stamford bridge of course) with me and Ian making the shed bar\'s faithful howl in laughter on regular occasions, without my knowledge it must be said (I had downed several large beverages by this point) was grand. It was great to hear them say \"FUCK OF TO THE BOG MATE THEY SCORE WHEN YOUR NOT HERE\" where upon everyone attuned into raptuous laughter.  Now we play Portugal, with even more booze flowing between me and Ian, at hopefully, a considerably larger volume of noise, even compared to last night.  Speaking of Ian, form what he tells me last night, He and Tracy (his girlfriend no less) and a camp fellow by the name of Jon are going to move into a little apartment in Mitcham somewhere, and Ian wants me to come stay there, to back up the presence of \"hetro masculinity\" (why can\'t he open his eyes and just see I\'m hardly that stereotype am I?!?!). This has all come about due to Tracy\'s family are moving to Brighton, and I\'ve wanted to meet her and talk to her about young Ian for quite a while now (as has everyone Ian, if you\'d just trust us poor folk). So I\'m caught up at a crossroads to see whether I want to grow up overnight (in Mitcham of all places), or to just stay in the tried and trusted Molesey. I just have to mull intensely for the next few days me thinks.  Thomas seems to be getting more and more adorable by the day, and I think I really like this one, so I might play it not as casual with him from now on. I really want it to work,  even  God must know this by now.  There is sweet FA to do for the next half an hour until 2 and a half hours of super duper Keith man philosopher takes us on another adventure of deep thought and provocation, so I guess I\'ll just read the times until then. Ciao. xBuBzx  '),(7884,3000240,'21,June,2004','2004-06-21 00:00:00','   Like wow I\'m in love with him already . . . xBuBzx  '),(7885,3000240,'21,June,2004','2004-06-21 00:00:00','   I would like to ass that if you have anything to say to myself, that the comments area aern\'t allowed for it, especially when you post under annonymous et al. Msn or nothing, got it.  I\'m the site owner, and if you disagree with my rules, then tough shit little ones. xBuBzx  '),(7886,3000240,'20,June,2004','2004-06-20 00:00:00','   My head is not that much dislike the poor soul in my title link . . .  I had the first controversial on this blog late last night, from Erd (Jo\'s \"boyfriend\"), which I was greatly amused over. I mean firstly, who calls there child Erd, and secondly, who doesn\'t run to the nearest deed-poll station when they finally realize how stupid it is. Anyway, Jo from what she told me, despises this guys carcass anyway, which makes it all the more surprising that he is still so into her, you\'d have thought he would have got the message by now but hey, what are you going to do, he is a guy after all, just innately stupid.  On a lighter not, the second week of the block time table has started, with me losing the third copy of the timetable, and coming in 2 hours early for everything, which isn\'t a great start to the week by all admission, but It\'s England tonight with Ian, Dan and Ali (maybe) in \"the blue room\" opposite the shed bar in Chelsea. I\'m looking forward to it to be honest after such a dreadful weekend.  Well the weekend wasn\'t that bad anyway, I\'ve been talking to this extremely attractive guy in every sense called Thomas (not Tom thank God, that name conjures bad memories of Halliford), who is friends of Pippa, who I just want to say \"awww\" and \"hehe\" like a giggling school girl, every time I think about him, as well as fuck him senseless, but I guess that goes without saying when you meet a new guy.  It\'s psychology for like the first time in a month at quarter past one, which should give me a good chance to catch up with Messrs, Amicia, Sarah and Suzanne etc which should be cool considering I need a long chat with them about men and Jo and how much Steve looks like a chipmunk every time we see him. LRC seems to be bursting with people, so I\'ll be of, I guess I\'ll just mull over a broad sheet, or whatever the smaller version of them are called these days. xBuBzx  '),(7887,3000240,'20,June,2004','2004-06-20 00:00:00','   urlLink    Advertising campaign in preparation for next years cannabis legalisation . . .  '),(7888,3000240,'20,June,2004','2004-06-20 00:00:00','   Nickelback - \"Figured You Out\"    I like your pants around your feet And I like the dirt that\'s on your knees And I like the way you still say please While you\'re looking up at me You\'re like my favourite damn disease  And I love the places that we go And I love the people that you know And I love the way you can\'t say no Too many long lines in a row I love the powder on your nose  And now I know who you are It wasn\'t that hard, just to figure you out (Now I did, you wonder why) And now I know who you are It wasn\'t that hard, just to figure you out (Now I did, you wonder why)  I like the freckles on your chest And I like the way you like me best And I like the way you\'re not impressed While you put me to the test I like the white stains on your dress  And I love the way you pass the check And I love the good times that you wreck And I love your lack of self-respect While you passed out on the deck I love my hands around your neck  And now I know who you are It wasn\'t that hard, just to figure you out (Now I did, you wonder why) And now I know who you are It wasn\'t that hard, just to figure you out (Now I did, you wonder why) (Why not before, you never tried) (Gone for good, and this is it)  I like your pants around your feet And I like the dirt that\'s on your knees And I like the way you still say please While you\'re looking up at me You\'re like my favourite damn disease  And I hate the places that we go And I hate the people that you know And I hate the way you can\'t say no Too many long lines in a row I hate the powder on your nose  And now I know who you are It wasn\'t that hard, just to figure you out (Now I did, you wonder why) And now I know who you are It wasn\'t that hard, just to figure you out (Now I did, you wonder why) (Why not before, you never tried) (Gone for good, and this is it) xBuBzx  '),(7889,3000240,'20,June,2004','2004-06-20 00:00:00','   Work was quite bearable for a fathers day and the evident attachment of paternal ass kissing that all children must endure amidst these most tormenting of hours.  I actually discovered a friend who fitted into none of the several categories I put forward to you yesterday - Rachel. Although she has the same flaws as the others, just in different ways. I think she knows like Frances and Anna that she is indeed one of the stunningly beautiful people and that could have twisted her imagination, not that she or anyone else would be consciously aware of it if they were in her situation. She just never seems to be up for going out, or ever asks even whether we could meet for a drink or something outside college and work. Well I decided after a long and hard think during till service that it\'s not worth really trusting anything, Or D.T.A as one famous wrestler would say.  I also got thinking as I often do in my day dreaming there, about what qualities would I like to have perfected by the time I\'m as old as the oracle, and well I still need to think about it for a while. Meh, I\'ll tell all when the football\'s finished xBuBzx  '),(7890,3000240,'03,July,2004','2004-07-03 00:00:00','   urlLink    Quite . . .  '),(7891,3000240,'01,July,2004','2004-07-01 00:00:00','   Rock\'s endangered species - why alternative needs your help    \"...bowing out early is one of the best things that can happen to a Cultural Icon.  He or she will be deemed \"Voice of a Generation,\" and grief-stricken fans will hold candlelight vigils and anniversary memorials for years to come.\"  Above:  Michael Abernethy, \"The Real Significance of Kurt Cobain \"  Every April, the floodgates are opened; millions of enduring Nirvana (especially Kurt Cobain) devotees privately and publicly express their equally enduring gratitude for the late pioneer of not only the grunge genre but an altogether unique approach to musicianship, rock performance, and dare I say, life itself.  But the clichÃ©s have been reported since 1994 (perhaps earlier), and each year the response is the same: Kurt Cobain was indeed an icon, a visionary, a heroin addict, and a scrawny 20-something-year-old.  Where does that leave the passionately loyal, if diminished, audience still bent on maintaining his legacy?  According to Michael Abernethy, author of the aforementioned PopMatters column, \"The Real Significance of Kurt Cobain,\" a cycle of icons will continue with each generation: \"[L]egacies change over time.  When one icon passes, someone else steps in to fill the void, even as music buyers of one generation grow up, and the next generation finds someone new to represent their particular concerns.\"  The unfortunate reality is that Nirvana was part of a finite generation, one that would either fade with the band\'s unexpected dissolution or the natural passage of time and fickle music fans.  Nevertheless, the grunge era was indeed cut short.  Now, only one day of the year is specifically dedicated to an artist and human being who struggled, survived, and even died on a daily basis.  Why have we disengaged ourselves from his life?  Is it the fragmentation of the current late teens to mid-20s generation, the supposed need to fit oneself into a \"distinct\" community, whether it be based on fashion, music, or crack?  Is it the decayed state of alternative music, molded in part by the increasingly popular hipster sub-culture of independent music?  Or is it just laziness on the part of young eyes and ears, an audience corrupted by a bite-size attention span and a lust for whatever might be cool?  A disease usually has multiple symptoms, and this diagnosis is no different, but my particular concerns of alternative decay, unconscious fans, and divisive sub-culturalism are the primary ailments.  In a post-MACHINA II interview, Billy Corgan of the Smashing Pumpkins, Zwan, and now just himself laments the loss of the alternative community: \"The idea of the integrity, the idea of what it means to be in a band, the idea that you have to walk right and be right because you got to stand up not just for yourself but for your community â€” that\'s all gone, thatâ€™s just out the window.\"  In the fallout of such noteworthy acts as the Pumpkins, Nirvana, Rage Against the Machine, and Soundgarden, a trend within the alternative circle cropped up as certain members of the aforementioned bands joined together to continue the march: A Perfect Circle and Audioslave are two of the recent media-termed supergroups to include hallmark musicians.  Additionally, current acts Radiohead and Tool continue to create.  With such bands recording and touring on a relatively consistent basis, one would consider modern (or rather postmodern) music to be more than just alive and breathing.  Yet, it is undeniable that Audioslave is no where near the significance (yet?) of its two predecessors.  Dave Grohl and the Foo Fighters may be a solid enough group for popular listeners, but there is no doubt that his previous role casts far too big a shadow upon any future projects.  And the Pixies\' recent touring and newly released material?  While the nostalgic sensibilities of 80s and 90s alt-fans are already roused, what is the band returning to?  Certainly not the musical climate of their prime years.  Same goes for Fountain of Youth drinkers, Sonic Youth, who are impressively looming upon twenty albums.  Nevertheless, Sonic Nurse is preparing us for their approaching finale.  It comes down to an observation of the past two decades and, more importantly, the past few years.  Yes, Morrissey\'s You Are the Quarry is incredible, but much like Sonic Youth, the Smiths vocalist is no sapling (he\'s hitting 50 in a few years).  Although it is not too late â€” nor is it ever â€” to begin buying past records of vital artists, the surrounding community is no longer together.  When Tool releases their next album in 2005, it will be one of the few important reminders of just how high alternative music was a decade ago and how low it has sunk since.  Alternative music\'s slow decay cannot be simply attributed to aging artists however; the other side of the equation â€” the audience â€” is an equally necessary component.  Enter cultural fragmentation: musical genres have long served as tangible indicators of personality, as well as carvers of popular trends.  MTV\'s video countdowns are essentially indoctrinations of cool, the popular music the popular people are downlo... or rather buying, since online file-sharing is an extinct beast (sure, RIAA). Even the sub-culture of hipster-independent music is being dug up by MTV\'s Advance Warning; the underground caverns are seeing much more light these days (Pitchfork Media receives 80,000 visitors a day and 500,000 a month).  Hip music listeners have discovered that TRL attendance is not as necessary as owning that indie debut no one else has even heard of.  Knowledge is cool, right?  As much as sub-cultures shift and re-align, fragmentation and division will continue to permeate youth generations.  Specifically, independent music has long been regarded as an inherently exclusive club, its members pompous know-it-alls of those bands far outside the mainstream.  After all, it is cool to own The Notwist\'s Neon Golden â€” that is, until your neighbors start listening as well.  Is this musical superiority complex justified for those ardent fans who do research a band\'s cryptic metaphors, memorize the lyrics, attend multiple shows on the same tour (and perhaps help load the band\'s amplifiers into what could be called a piece of abstract metal art with wheels), and attempt to actually understand and utilize the aforementioned band as a potential source of inspiration for their own lifestyles, whether it be in painting, politics, or just moseying around?  The distinction lies in both the artist and the audience, and this is where alternative begins to shine more brilliantly than any rock genre.  Nevermind.  The commercial apex of Nirvana\'s career.  On one hand, an expansive opportunity for millions of people to experience what the grunge underground had already discovered since the band\'s 1989 debut, Bleach.  On another, the boiling point of Nirvana\'s publicity and mainstream popularity, fueled by a single millions moshed to but barely understood.  But who really cared anyway?  Kurt Cobain was onstage smashing his guitar and huddling into a corner!  Cool, right?  There is performance art and there is suffering; both appeal to an audience, but only one stops when the lights return.  An audience\'s interaction with an artist\'s output is an emotive exchange between the two parties, one determining the extent of the artist\'s significance and the depth to which the audience attempts to honestly see the artist.  It is easy to simply idolize and adore, perhaps too easy.  The current realization of fandom is based on a hierarchy, the artist situated above the audience as opposed to on an equal plain, an equal level of exchange.  Imagine if the original intention of a fan was not to be entertained, roused, or disturbed but simply inspired to tangibly apply the energy of the art to the fanâ€™s own work.  If a generation\'s premier artists are considered links in a larger historical chain, the audience will realize its role in perpetuating that chain.  Alternative music has historically constructed itself as a sequence of diverse links, from British post-punk to American guitar rock.  There\'s a signature style to alternative, an awareness of the importance and responsibility invested by being in a band and writing certain songs.  Ian Curtis knew it, Kurt Cobain knew it, and Maynard James Keenan knows it today.  While none of those artists stuck to their underground scenes, each carried an undeniable significance; independent is by no means a determinant of alternative.  A dozen vague notions can be used to describe the worth of alternative, but it simply boils down to an altogether different way of approaching music: not just as a career, or even a lifestyle; just about every popular band claims to \"do it all for the music.\"  It is the depth to which one will go to achieve an ordinary life amidst extraordinary circumstances: fan idolization, millions of dollars, and tours across the globe.  It is the sincerity and consistency of communication one is willing to make with his/her audience â€” an exchange as opposed to a lecture.  It is art-centric as opposed to artist-centric or even audience-centric.  It is art for art\'s sake, a form characterized by distinct aesthetics and particular ideals â€” music as a tool of consciousness, a method of living, playing, and working â€” and those ideals are what define alternative as a music genre, a mindset, and a worldview.  Where has independent music gone wrong then?  As with many genres â€” including alternative â€” images of the typical artist and fan emerge.  A list of ways to be \"alternative\" is compiled.  It becomes more of a societal decision than a personal one, and cultural division inevitably erupts over who is truly \"alternative.\"  The same thing has happened to gothic and punk ethics; the question of sincerity is raised, and no one is willing to compromise.  While punk is historically rooted in its earliest pioneers, independent music essentially has no roots; it is literally defined by the band being on an independent label.  It is only through modern characterizations that indie music has taken on cutting edge connotations and rebel art aesthetics: \"Yes, we are on an independent label, and because of that we are inherently better than that major label artist.\"  It is part of the often fragile youth resistance to corporate control and multinationals; not only is it difficult for any kid to grow up in America without tapping into the capitalist vein, but it is ridiculous to claim that independent labels somehow fight the already-present conglomerates (offer frequently more liberating contracts, yes, but not fight).  Because \"indie\" has become a misguided synonym for \"brilliant\" to many similarly misguided listeners, alternative music is at a loss: is it possible to express that integrity discussed by Billy Corgan when on a major label?  Obviously so, and assuming a band takes complete creative control â€” which it should â€” over its music, artwork, and tours, the label has nothing to do with a band\'s ability.  While independent labels are important caretakers of emerging artists, the unfortunate disadvantage is one of access.  The indie underground is an infamously private sector, and many fans are simply unaware of promising bands working towards greater exposure.  The fight should not be one of fan access but rather artistic control under a major label; if a Capitol Records band â€” such as Radiohead â€” can retain total authority over its output, their statement can be made to a broader audience.  True, the fan base will increase, inevitably pooling in trendy fans who appreciate a band (take MTV\'s alternative pet, Coldplay) as long as it is hip, but indie music as an exclusive genre as opposed to a stepping stone is only a divisive example of generational fragmentation.  Many of the premier alt-rock artists of the 80s and 90s were by no means indie bands throughout their careers.  They often began in underground scenes, drawing from their respective cities, but an honest band understands the natural diffusion of music, the locomotive quality of expanding art and not restricting it to an audience based solely on geographic or socio-economic access.  Before I step off my soapbox, I want to clarify something: I by no means hate independent music or its role within popular culture.  I just know that in 1993 Siamese Dream was released, and I had no idea.  Yes, I was eight years old, but still... there\'s something remarkably sad knowing that I was alive when Billy Corgan was striving for that fabled integrity, but I was just a grade-schooler, unaware and uncaring.  Kurt Cobain died the following year, and once again, I was in a classroom at St. Michael School, unaware and uncaring.  I have no doubt that alternative music meant something more during the late 80s to mid 90s, and I am hoping that the clichÃ© of history repeating itself does in fact hold some truth.  Alternative, much like punk, is an ethic all by itself, a lens of perception, a way of thinking, and a general understanding.  It is not just a music genre, and my hope is that more bands understand that broader connotation of rock n\' roll.  I sincerely believe it is worth it. xBuBzx  '),(7892,3000240,'30,June,2004','2004-06-30 00:00:00','   I\'m a little less drained than I was last time I blogged, but I\'m still struggling to keep the eye lids from ceasing up permanently, so I\'ll keep it short and sweet. It\'s the last day of college and I\'ve just had the most nightmarishly horrible key skills presentation (God knows how you can call it a presentation when Steve is interrupting you on a second-ly basis) but I managed to come out in one piece so now I\'m happy and looking forward to Thomas, summer and a bit less one way criticism from Dad about work. Oh, but alas, that can\'t be since Mother decided to do a \"Rachael\" (and no I don\'t mean turn into a closet lesbian) and cry over the littlest thing, which obviously, as everyone knows,  WILL  be my fault, even if it was because she can\'t handle menopause, like an achoholic lightweight can\'t handle 10 pints in a minute.  I presume that this will be my last ever blog from the LRC, which is sad considering I actually like the solitude in here, if not the slowness of the broadband.  Eraser  is gently screaming through my ear lobes, so I guess It\'s of to read what Giles Smith has to say in  The Times , or was it the guy with the beard? I\'m not sure; anyway, many things to discuss tonight once my brain is in a farther wider frame of contemplation xBuBzx  '),(7893,3000240,'30,June,2004','2004-06-30 00:00:00','   Nine Inch Nails - Starfuckers, INC    My god sits in the back of the limousine, my god comes in a wrapper of cellophane, my god pouts on the cover of the magazine, my god\'s a shallow little bitch trying to make the scene.  I have arrived and this time you should believe the hype, I listened to everyone now I know that everyone was right, I\'ll be there for you as long as it works for me, I play a game, it\'s called insincerity.  Starfuckers, starfuckers, starfuckers inc, starfuckers.  I am every fucking thing and just a little more, I sold my soul but don\'t you dare call me a whore, and when I suck you off not a drop will go to waste, it\'s really not so bad you know once you get past the taste, yeah (asskisser).  Starfuckers, starfuckers, starfuckers inc, starfuckers.  All our pain, how did we ever get by without you?, you\'re so vain, I bet you think this song is about you, don\'t you?, don\'t you?, don\'t you?, don\'t you?.  Now I belong I\'m one of the chosen ones, now I belong I\'m one of the beautiful ones. xBuBzx  '),(7894,3000240,'30,June,2004','2004-06-30 00:00:00','   This weekend has been harsh, both physically and emotionally, and If I wasn\'t talking to you right now doing my key skills presentation for sociology tomorrow, then I\'d be having another one of those weird dreams which was most probably based on the spiders I read about that American soldiers seem to find every day in IRAQ apparently, of soldiers waking up to find two seperate limbs missing froma deadly humungous spider. Well update tomrrow. xBuBzx  '),(7895,3000240,'29,June,2004','2004-06-29 00:00:00','   Damm it, I\'m going to sit back, kick it, open a can of the manly AMSTEL BEER (yes only dear Ian would know what I mean by that) and relax, safe in the knowledge that I don\'t ever have to participate in Descartes Dualism or Plato\'s analogies ever again. One bad thing would be that I don\'t get to spend my lessons thinking about the gorgeous Anna beside me *meh*. However now I guess I\'m really stuck as what to do with the rest of my life. I guess I\'ll just have to wait and see how the results turn out.  Just learnt I\'m of to the Greyhorse to see Law twang his tunes - and to give Jenny a hug after her epic *find me a bloke day* turned out to be amazing after all seemed lost. Ah brings a smile to your face . . . xBuBzx  '),(7896,3000240,'28,June,2004','2004-06-28 00:00:00','   Nine Inch Nails: We\'re In This Together    I\'ve become impossible, holding on to when, when everything seemed to matter more, the two of us, all used and beaten up, watching fate as it flow down the path we, have chose.  You and me, we\'re in this together now, none of them can stop us now, we will make it through somehow, you and me, if the world should break in two, until the very end of me, until the very end of you.  Awake to the sound as they peel apart the skin, they pick and they pull, trying to get their fingers in, well they\'ve got to kill what we\'ve found, well they\'ve got to hate what we fear, well they\'ve got to make it go away, well they\'ve got to make it disappear.  The farther I fall I\'m beside you, as lost as I get I will find you, the deeper the wound I\'m inside you, for ever and ever I\'m a part of.  You and me, we\'re in this together now, none of them can stop us now, we will make it through somehow, you and me, if the world should break in two, until the very end of me, until the very end of you.  All that we were is gone we have to hold on, all that we were is gone we have to hold on, when all our hope is gone we have to hold on, all that we were is gone but we can hold on.  You and me, we\'re in this together now, none of them can stop us now, we will make it through somehow, you and me, even after everything, you\'re the queen and I\'m the king, nothing else means anything. xBuBzx  '),(7897,3000240,'28,June,2004','2004-06-28 00:00:00','   I\'ve decided to say screw you all and fuck fellow students a-level ethic and just blast NIN into my ears until I\'m  told  and not glared at in  annoyance  by other small-minded townies for my \"dirty grunger music\" (I don\'t think they took a preference to  Eraser  by  NIN . Well I say that actions and more specifically, websites speak louder than words on these underbelly delinquents of British teenage society, so brilliantly portrayed in this  urlLink link  (just watch, read, listen and laugh).  Anyway, my gig whore returned late last night thank God, and enough said really on how I began to feel amidst the dying embers of the weekend, after I returned from viewing Muse\'s stunning performance of the pyramid stage at Glastonbury, too see my little traxlover online (-:.  Feeling have definitely cooled back at home, we\'re basically talking on a normal playing feel, although it will never be happy families incarnate there, that\'s for sure.  I\'m supposed to have this sociology key skills and coursework done and dusted by Thursday morning, however it all seems quite pointless doing any of it really considering I won\'t even be there next year anyway. Well I\'m of for a drink . . . xBuBzx  '),(7898,3000240,'27,June,2004','2004-06-27 00:00:00','   Lacuna Coil - Swamped    When you\'re taught through feelings.  Destiny flying high above, all I know is that you can realize it.  Destiny who cares, as it turns around, and I know that it descends down on me.  It\'s just another day, the shame is gone, hard to believe, that I\'ve let it go.  Destiny can\'t replace my life, Scary shadows of my past, are alive.  Destiny who cares, as it turns around, and I know that it descends, with a smile.  It\'s just another day, the shame is gone, it\'s hard to believe, that I\'ve let it go away,  It\'s just a melody, it bleeds in me, hard to believe, that I\'ve let it go. xBuBzx  '),(7899,3000240,'27,June,2004','2004-06-27 00:00:00','   urlLink    Ha Ha look what Mother did to the car last week . . .  '),(7900,3000240,'27,June,2004','2004-06-27 00:00:00','   I took the step up in my manual self-IT education today and completely assembled this computer from scratch after the Mother got in a hissy fit and  tried  to take away my PC. I mean, what on God\'s green Earth would I do if the internet suddenly disappeared one day, write a book, talk to myself - or worse a sock-puppet?  I am due to go to a gig with Jen tomorrow, but I\'m not sure since she would have only just returned from the business trip to Paris to engage in the mass cocaine transportation shenanigan. Alas I may just be stuck in doors again, much like this weekend, after a huge blow out from a wide verity of friends and an argument with Mother; you would have thought by now that my spirits would be down and grounded into minuscule fractions, but I just put things into prospectus and everything seemed a little calm and more relaxed/peaceful, which was a good self-emotion/reaction if I can repeat it again on countless occasions coming up over the next year.  Ian is asking me for a deposit of $250 if I want in on the place in Mitcham, and I\'ve been thinking long and hard at work and afterwards If It\'s a good idea to take it or not, and at the moment I\'m quite keen, but it really depends on a number of factors and seeing whether they fall through or anything, which I\'ll discuss in detail tomorrow probably.    God I want my gig whore back *sulks*  xBuBzx  '),(7901,3000240,'10,July,2004','2004-07-10 00:00:00','   10.30 am - The polyphonic trumpet ring-tone on my god awful Samsung goes with \"Kate\" buzzing on and of.   \"Oh God\" I said. \"I\'m so sorry I\'ve got a throbbing headache and I\'ve only just woken up - I\'m really sorry.\"  Then with my eardrums barely awake, they shudder at the Canning town accent, reminiscent of the girls that Barron hangs around with - \"uurr ly-in\".  \"I\'m really sorry Kate, I\'ll try and make it up to yo . . . \" - the line crackles and she has hung up in her arrogant - I\'m fitter than you because I\'ve got died blond hair, and fake gold jewelry and Â£1.20 foundation with classy vocabulary in everything I say, and I\'m of to suck Michaels cock so he\'ll beat up this boy who\'s been giving me \"beef\" kind of way.  So the previous evening, Dad had come marching in to the study proclaiming \"BE HERE AND WE\'LL HAVE LUNCH AT ONE!\", which obviously bought a large grin upon his several chin-ridden face.  So anyway, last night was rather a bore, with the German being ill, It was spent in, and a extreme early night, especially for me of 7 o\'clock. Although I did awake suddenly as I got a call from an old friend - Lauren, who no is living with a new bad boy and thankfully out of that shithole of a home of hers, and so should sometime meet up with her somewhere.  Tonight should be better however, as the German is feeling rather better and we plan to go mass drinking for the first official day of summer. HOORAH! xBuBzx  '),(7902,3000240,'09,July,2004','2004-07-09 00:00:00','   urlLink    Mother - as you are reading this - and as you and I are both fully aware that this is the only true glimpse you\'ll ever have to knowing me properly as you so strangley still crave to - and this despite my clearly stated opinions about you and your \"philosophy\" - this is Dads and my own analogy of just how well you came across today, amidst another pety rant. Well you have had practise before havent you . . .  '),(7903,3000240,'08,July,2004','2004-07-08 00:00:00','   I can safely say I\'ve grasped the notion of \"dead end job\" and all of it\'s little PC additives, which in many aspects intensify the unbearability of it. However the Covent Gardenesque area is quite contemporary and most definitely my cup of tea, even if the accent and taste in clothes shop of my partner in crime, Kate, isn\'t as such.  So why dead end you say? Well, frankly I\'ve never been a huge fanatic of the the postal slave labour trade, however there have been occasional moments of joy - such as serving tea and biscuits to the junior school teachers, and the staff all seem rather friendly and polite and in full awareness of there 200+ IQs, which can come in handy if they had a mischievous side to them including money laundering and transportation of illegal immigrants and alike. Alas they are all law abiding citizens, and lovely one\'s really. However, if there has been some advantage to the whole experience, which I value the most, is seeing how I can improve my overall poetry (well sex poetry) and general literary skills I reckon.  I feel I\'ve gotten to know Kate somewhat, and I\'ve changed my opinion on her somewhat from the stagnate blond townie from TL at Esher, to someone who has somewhere, an inner beauty about her. Well I did say somewhere - didn\'t I? Other things which come to mind about the place - I shall most definitely rant/bitch upon tonight.  Since days have lasted from 10/10:30 - 5:30/6 respectively, maintaining a consistent social/blogging life hasn\'t been the easiest of things, however I did manage to live up to my promise for Rowan, and see his ever improving band and band support (even though half of them are pretentious in the closet twats) -  Lunatic Crash  at the  Half Moon  in Putney. The drum/guitar/vocal-screeching solos seem to be gathering improving speed, and volume and, well class to be honest. If there\'s anyone from school who\'s going to make it in the music industry, it will be Rowan, and I should suspect, in Lunatic Crash.  I discovered, as did Erd, what Jo is really like a couple of days ago, so we are deciding to give as good as we got, and raise the volume on the compliment so to speak - I\'ll update you on that later.  I have no idea what is going on between me and Thomas, and to my joy (and probably his as well) neither do I care as much as the days haul on, I really hate things ceasing before they\'ve even really begun which is apparently what seems to be happening again, which strangely enough, the title of this webpage should sum up in an instant. My friends like E and more recently Jen are far more important to me, especially at the moment, and God knows I want to know these lovelies until I have hair growing from my ear lobes. Sentimental I\'m sure. It was quite fascinating to see, from my point of view, when seeing both the friends I\'ve got since joining Esher, and the Halliford patheticos, and really seeing how much more I value the new regime in my life. Take for instance talking to Richard. Now conversation has never been that boys strong point, much like myself, but seeing his weak feeble attempt at humor amidst his \" Nawberry \" and his lackluster conversational skills (worse than when he was 11 for crying out loud), I could finally, once and for all (well close enough but not 100% there I guess) put those 5 years to rest. Lets see when the time passes, that night really tipped the scales between recovery and happiness. xBuBzx  '),(7904,3000240,'06,July,2004','2004-07-06 00:00:00','   urlLink    Recent frustration car carving by Father . . .  '),(7905,3000240,'05,July,2004','2004-07-05 00:00:00','   Default - Deny    Today I woke up and you were gone  The whole day wondering what I did wrong  It\'s like I\'m falling from a mountaintop  My heart keeps pounding and it won\'t stop   Can you see this hell I\'m living  I\'m not giving up   Will you crawl to me  Will you fall with me  I\'ll never crawl to you  I\'ve done it all for you   Well don\'t deny  The hand that feeds you needs you  Oh god I\'d die to try to  Finally please you   There goes a piece of me  Will I cease to be  I\'ve never lied to you  Fought bled and died for you   Well don\'t deny  The hand that takes you breaks you  Oh god I\'d die to try to  Finally please you   There she goes and I\'m on the ground, I\'m on the ground, I\'m on the   Well don\'t deny  The hand that feeds you needs you  Oh god I\'d die to try to  Finally please you   Will you crawl to me  Will you fall with me  I\'ve never crawled to you  I\'ve done it all for you xBuBzx  '),(7906,3000240,'04,July,2004','2004-07-04 00:00:00','   I promised that I\'d return and post something LoVeLy about squabbo hunny, and I will, however a small rant may just be needed before breath is drawn.  I have the feeling, or I know that I\'m really nervous about telling Ian whenever I see him sometime this week, and you know what  it is  so I shaln\'t divulge any more upon you poor soles.  Well, where to begin?  Ah yes, when Dan and  \"Wadan\"  Jess Law and Ali decided to stay at mine that strange night. It was all dandy really and I managed to have a proper chat with jess for a bit, well I say proper but it was more of an information bitch about stuff I\'ve never really opened up to her about before. Then the \"white wannabe rudies\" she so glamorously denounced them all (except Law) decided it was a grand ol\'plan to hide my keys, which enticed Father into all sorts of emotionally charged cliches (blaming myself for it a.s.a.p) and so I had to foot the Â£7 for there faults as human beings. It\'s not just the Â£7, it\'s the principle of it, and Dan decided against re-embersing me with the sterling, electing instead, to use the sly-alternative and blame me for a Kessalar self-inflicted loss of a couple of quid.  Anyway the day later was the final exam in Maths which went rather swell, and then to Sutton with E to search for jobs what\'shats not. We counted an average - slag per 10 seconds, which curiously was backed in the claim by a kind gentleman who was maticulous in his examination of the McDonalds branch in the high street and came to pretty much the same conclusion we arrived at, which was rather fun to discover. Clink  urlLink here  to discover more.  The Friday night wasn\'t all that great, Thomas was in the end, busy with his kittens and so, me and Anna waited to greet her new lover; BB5 Stuart wannabe Carl to arrive at Paddington, no later than 2 AND A HALf HOURS LATE! So no sin city and certainly no threesome. I ended up In Barbara\'s bed (Anna\'s Mum) which was quite a shock at 10 the following morning. It was however nice to see a certain  Chris Tarrent  at the breakfast table, discussing all thing television and alike with Barbara, and as it turned out, he was a rather down-to-earth bloke from start to finish.  Saturday was a night for me and Ian to engage in our extremely unique conversation, in which absolutely no-one apart from me and Ian would ever appreciate fully, which is just swell for the time being, not when, or if I ever move to SW8. We then upon leaving the German, saw Shrek 2, which despite it\'s brilliant prequel, was an outstanding film, easily outstanding its elder from start to finish.  Sunday involved me thinking of Thomas (oh shit I\'ll just do the post about him tomorrow evening  I SWEAR ) and strangely of school, and blank empty insomniac periods of null.  The poetry society beckons for this weeks adventure . . . xBuBzx  '),(7907,3000240,'04,July,2004','2004-07-04 00:00:00','   This week has drained my mental capacity and motivation for writing in this blog, so I\'ll just leave it a bit short.  As I\'ve been recovering somewhat in all areas, I tend to become terribly reflective and subdued, so I\'ve spent most of my time thinking of Thomas, and how I can tell Ian that I\'m bi. It\'s really important I just get it out of the way, I hate being so late out of the closet it just ripes me apart. Perhaps school would have been bearable if I did it in year 10 and not have given two flying fu . . . well you get the picture. I Promise as God is my whitness I\'ll post something meaningful and lengthy about him tomorrow - he deserves it because he\'s been overly nice and lovely to me, and I want him badly now. xBuBzx  '),(7908,3000240,'04,July,2004','2004-07-04 00:00:00','   A man is lying in a hospital bed with an oxygen mask over his mouth. A young nurse enters his room to sponge his face and hands. \"Nurse,\" he mumbles from behind the mask, \"Are my testicles black?\" Embarrassed, the young nurse replies, \"I don\'t know Mr. Jones, I\'m only here to wash your face and hands.\" He struggles again to ask, \"Nurse, are my testicles black?\" Again the nurse replies, \"I can\'t tell. I\'m only here to wash your face and hands.\" The head nurse walks by and sees the man getting a little distraught, so she marches over to find out what\'s wrong. \"Nurse,\" he mumbles, \"are my testicles black?\" The head nurse whips back the sheets, pulls down the man\'s pajamas, moves his penis out of the way, has a good look, pulls up the pajamas and says, \"There\'s nothing wrong with them.\" Frustrated at this, the man pulls off his oxygen mask and asks again, \"I said, are my test results back?!\" xBuBzx  '),(7909,3000240,'16,July,2004','2004-07-16 00:00:00','   urlLink    Pwahness lol . . .  '),(7910,3000240,'14,July,2004','2004-07-14 00:00:00','   BuBz Diet Update  : today was OK, however I did feel quite bored with nothing to do at home, however waking at 3 in the afternoon did the whol thing a world of good. I had a big soup and 2 glasses of fizzy flavoured water and a chili from Mother and afterwards a yogurt and several microwaved pieced of plain brown bread (which actually tasted rather better than you might expect).  I felt rather out of energy today, and I hither to guess that it was a combination of a lack of binge eating, and work last night. Nothing else radical to report, apart from I\'m trying to think of what filsm to bring to Jen and myself film whoring tomorrow evening. I fancy something epic/horror fuelled. This could involve that Sam Neill film I saw once when I was half asleep which was just AWESOME (so after this a half hours research to get the title and availability of this beauty). Wel that and fight club and american history x and HELLRAISER!!! MwAh. xBuBzx  '),(7911,3000240,'13,July,2004','2004-07-13 00:00:00','   Metallica - Fade To black     For Fred and erm Jen and well because I\'ve remembered how much I love this song...  Life it seems, will fade away  Drifting further every day  Getting lost within myself  Nothing matters no one else  I have lost the will to live  Simply nothing more to give  There is nothing more for me  Need the end to set me free   Things are not what they used to be  Missing one inside of me  Deathly lost, this can\'t be real  Cannot stand this hell I feel  Emptiness is filing me  To the point of agony  Growing darkness taking dawn  I was me, but now He\'s gone   No one but me can save myself, but it to late  Now I can\'t think, think why I should even try  Yesterday seems as though it never existed  Death Greets me warm, now I will just say good-bye    Metallica - No Leaf Clover    And it feels right this time On this crash course we\'re in the big time Pay no mind to the distant thunder Beauty fills his head with wonder, boy....  Says it feels right this time Turn around, found new high lights Good day to be alive Sir Good day to be alive, he said.....  Then it comes to be that the soothing light at the end of your tunnel is just a freight train coming your way Then it comes to be that the soothing light at the end of your tunnel is just the freight train coming your way  Don\'t it feel right like this All the pieces fall to his wish Suck up for that quick reward boy Suck up for that quick reward they said.....  Then it comes to be that the soothing light at the end of your tunnel is just a freight train coming your way Then it comes to be that the soothing light at the end of your tunnel is just the freight train coming your way...... It\'s coming your way It\'s coming your way...... Here comes  Yeah, Then it comes to be that the soothing light at the end of your tunnel Is just a freight train coming your way.... Yeah... Then it comes to be, Yeah... Then it comes to be, Yeah... Then it comes to be, Yeah... Then it comes to be, Yeah... Then it comes to be, Yeah... Then it comes to be, Yeah... Then it comes to be, Yeah... Then it comes to be, Yeah.    Red Hot Chili Peppers - Higher Ground    People Keep on learnin\' Soldiers Keep on warrin\' World, Keep on turnin\' Cause it won\'t be too long. Powers Keep on lyin\', While your people Keep on dyin\' World, Keep on turnin\', Cause it won\'t be too long.  I\'m so darn glad He let me try it again, \'Cause my last time on earth I lived a whole world of sin. I\'m so glad that I know more than I knew then. Gonna keep on tryin\' till I reach the highest ground.  Teachers, Keep on teachin\' Preachers, Keep on preachin\', World, keep on turnin\', \'Cause it won\'t be too long. Oh, no Lovers, Keep on lovin\' While believers Keep on believin\'. Sleepers, Just stop sleepin\' \'Cause it won\'t be too long. Oh, no!  I\'m so darn glad He let me try it again, \'Cause my last time on earth I lived a whole world of sin. I\'m so glad that I know more than I knew then. Gonna keep on tryin\' till I reach the highest ground.  (Spoken) An\' Stevie knows that, uh, no-body\'s gonna bring me down. Till I reach the highest ground. (Spoken) \'cause me \'n\' Stevie, see, we\'re gonna be a sailin\' on the funky sound Till I reach the highest ground. (Spoken) Bustin\' out, an I\'ll break you out, \'cause I\'m sailin\' on. Till I reach the highest ground (Spoken)Just, uh, sailin\' on sailin\' on the higher ground Till I reach the highest ground  xBuBzx  '),(7912,3000240,'13,July,2004','2004-07-13 00:00:00','   urlLink    What Stefan most probably thinks whilst managing the kiosk at work - the sick pervert . . .   urlLink    This was a saying someone used last week at the poetry society - in these EXACT WORDS honestly.   urlLink    What I\'d like to do to Mother.  '),(7913,3000240,'13,July,2004','2004-07-13 00:00:00','   After months of self-aggressive-negotiations, well about a years worth of gradual weight gain, it\'s time to get back to something in the last 2 years I have excelled at - starvatio-anorexomanialysing (and what a word you must agree). I was talking with Tony (the resident Tesco security guard) about my swelling balloon tummy, and I have to lose this waist-line and appetite and overall consumption - it\'s not fair on me, Mother and Father, or my bank balance for that matter.  So, this will be my devilish concoct, well in a draft stage at least, commencing as of 7pm yesterday evening once wise old Tony gave me his personalised diet  plan pack , amidst some personal input I must admit:  - Nothing fizzy (Tony) - Nothing fatty at tesco breaks (Me) - Cycle to work this summer (Tony) - No sugary/salty food (Tony) - More worthwhile occupying activities i.e drugs (E\'s especially) and work (Me obviously) - Attempt to use household workout gear like the rowing machine (Me) - 3 meals a day (Tony and Me).  So there done, this blog will end if I fail it . . .  A punishment of the highest order.  On another topic, I\'m ready for a holiday - the easy part there - saying it. Now going on it will be a completely separate issue. I want somewhere sunny to get a tan and alike, a weekend or a few days in Europe is something I\'ve done and done again and again with Mother and Father, so Ian\'s timeshare could come in handy. Although I feel as though there isn\'t nearly enough time to get a proper holiday organised, so I will have to make do with Reading again as my solitary vacation. Alas, despite 3 years since my last holiday, something will work itself out nearer Christmas time. I mean I have plenty on my plate here to keep me occupied and busy until the time where I\'m in absolute desperation for a plane trip.  Work was OKish tonight I guess, although Becky\'s claws still hadn\'t smoothed of completely, I still loved mentally kicking her ass in my mind, when the customers passing through. You really should see the world I delve of into when I\'m there. Makes Heaven look like a shrinking violet. xBuBzx  '),(7914,3000240,'13,July,2004','2004-07-13 00:00:00','   Soulfly - Prophecy    I see the Red Sea part in front of me, I see the desert clouds bleed above me, I\'m with the prophets on the final destiny, We\'ll fight the heathens and the ghost enemy.  This is the prophecy, This is the prophecy.  I see the money beggars on the last temptation, We\'re marching now through the devastation, I\'m with the prophets on a mission of salvation Cities falling, nations falling nations.  This is the prophecy, This is the prophecy.  I see the new millennium tribal warfare, I see them blind sheep going nowhere, The new exodus, just like they said, I see the crown of thorns, you know I was there.  This is the prophecy, This is the prophecy.  I see the times in this armaggideon, I see the flag of the conquering lion, I must reach the mountains of Zion, I\'m with the prophets and the prophecy goes on.  This is the prophecy, This is the prophecy.  New millennium tribal war, New millennium tribal war, New millennium tribal war, New millennium tribal war. xBuBzx  '),(7915,3000240,'12,July,2004','2004-07-12 00:00:00','   urlLink    Exactly . . .  '),(7916,3000240,'12,July,2004','2004-07-12 00:00:00','   Words, phrases and overall sentences may be dramaticised by your truly for self-comical purposes, and hopefully to piss you of.     Saturday 2.00pm    The evil Samsung vibrates with \"Anna Home\" shining into my eyes.  \"Hewwo my little gorgeous German, what\'s up?\" \"I have a dilemma - I need some help\" \"Sure what\'s up?\" \"Well (in a vulnerable considerate voice), could you for a start not keep calling me German please, I\'m not even fully German, only 3 quarters, you know that\" \"I\'ll take it into consideration; anyway, what  really  bothering you my dear\" About ten seconds worth of awkward silence followed, eventually with it being broken by - \"Jade\'s stuck in Marseilles with no money - can she have 140 euros from  your  debit card Mat, you\'d be really helping her out.\" \"Well I don\'t have a debit card for a start\" \"Well can you get your Dads, and like not let him know?\" My eyes begin to roll and my lips start a long drawn out \"PFFFF\" down the phone to her to say \"you know I can\'t do that, or lend her 140 euros anyway, it\'s her fault she\'s got herself into that situation anyway; anyho, do you want to go clubbing tonight? \"Sure!\"  So we did.  Up on arriving at Highbury and Isslington,  had a drunken hobo keep calling me fat Matty Matty fat, to which people really just wanted to there and then spit at this parasite of society, and so I stumped a menthol cigarette into his hand, which felt devilishly fabulous. Then me Anna, Fluffy, and  the guy who lived above the station at Highbury and Isslington  went to Slimelight and clubbed it hard electronic Goth style hard for 9 hours straight.  Anyway *giggles to himself* beside that most of the night sped by as I mixed the industrial and dark wave room and Monsters inc, in the cinema hall, which even, at my great fucked stage of the night, was pretty fucking cool damn it.  I\'ve really missed the days of trying to sneak into that place at 14, through the back door, without the security noticing. It\'s really got better in the last few years, and for once, I felt at home with a cute guy near me, a non-judgmental/remotely townie crowd surrounding that with me in my old Goth clothes, cherished immensely.  I guess I feel quite sentimental at this moment, so I guess I must, If I don\'t ay it enough that I love you Anna so much, no matter how much I tease you or come of so weird, I could just hold you in my arms until sunrise everyday I was alive hehe.   Anyway, I was knackered by 7:30, so I headed back for a 2 hour journey to home . . .  Only to get ready for a 6 HOUR DRAG AT WORK, where Becky sniffed out my fatigue as soon as her fat arse walked through the automatic doors and made sure, with much more precision than usual, to punish with an ECOH warning (each customer offered help bullshit/PC thing i.e pointless, pathetic pompus notion) and to which, she stressed, meant that if I received a further two, I would have to be re-trained; (most probably by her no less) which (she forgets) would be the easiest 4 hours paid work of my life. Stupid bitch. xBuBzx  '),(7917,3000240,'24,July,2004','2004-07-24 00:00:00','   This has been a good day, a great ay infact, and these don\'t come along often, especially for me. If my syntax goes AWOL, then I do apologise, but several beautiful instances have occurred and have affected this \"literature\" immensely.  WELL, I\'m too tire now so I\'ll just edit this later - ciao. xBuBzx  '),(7918,3000240,'22,July,2004','2004-07-22 00:00:00','   Spiderman 2 was rather better than I thought, and with Amy falling asleep in my arms, it felt rather like comfort viewing.  However, I was feeling most pleased with how the day was going until Mother pranced in after a short conversation with Andrew on the phone. Now despite me not having seen Harry for a month, the little tiger bless him has been complaining to his school teacher that I have been hitting him in the last few days, which brought a whale of tear drenched emotion from Mothers eyes for near on 45 minutes. Why 45 minutes you say; well because everytime we delve into some heated exchange of \"words\", everything seems to go pearshaped and digress into other areas of my life, one\'s which should have been (in her mind) put to bed. She often suggests the notion that I should see a counselor, and if not that a psychiatrist, and if not that a psychologist, and if not that, well heavens above, Father *sarcastic gulp*.  Well with my daily ranting over, it\'s been a good day to be honest, met a swell guy called Wayne who was Amy\'s friend and had a childish/fun day in Kingston, and tomorrow I travel to see the German, and I think I might ask her out now or once she\'s single. I don\'t mean by any stretch of the imagination, to sound cocky or arrogant, but I think for sure she\'d say yes considering I know her really well these days; and 3 quarters of the time I can complete her sentences. Well wooopdedoo I\'ll digress my life to you all later on . . . xBuBzx  '),(7919,3000240,'21,July,2004','2004-07-21 00:00:00','   Why you ask, am I going to somewhere nice, without such as much as a physical location? Well mon ami (quite a fitting phrase for this paragraph at least) apart from it\'s similar audible qualities, which are true to life and to the eyes from what the very metaphorical travel agent \"Susan\" said, the seaside town in the south of France is where I will be relaxing, sun bathing and writing some predictable angst ridden articles/poetry/scorn which will hopefully re-juvinate me to my full senses, and a browner shade of skin.    BuBz Diet Update   Going fairly well. It\'s travelling at a much slower pace than last time but, I feel far more comfortable, and in less need to whitness its effect, so this will definetely be managable for the next few months at least. Next update in the coming days, where I will speed things up, a tad, even for this slowish flab burn( Sexual thoughts are obviously screaming there way through you at this minute I\'m sure).  God I miss the old people strangely. I miss the fight of fearing for my mental stability at 8:15am every time I paraded into B2 to *cough - queer* chants, which repetitised (yes word of the day) as the 45 minutes approaching Crosby\'s large backside waddled itself in between the swingdoor ledged, only for the returning door to give it the much needed whack it needed to make the full transition from hall to classroom. It\'s the first time I can safely say I\'ve missed that place, in that place I had aggression, an attitude, a moral high ground, and a worthwhile war to fight. Yes, I know; it\'s hard to win the war, (which in the end I didn\'t coincidentally as the vast majority of you who read this blog are aware) when it\'s 49/1 against you, but I loved being the  true  outspoken open minded martyr. I could go on, even more than I am at the moment, but it seems I may have repressed some thing after spending the 6 days sleeping without wake I did, which strangely, many of you like to call a coma, where it SHOULD have been called a mental holiday, to erm, MENTAL FLORIDA! I wish I would have stumbled across the world of blog in say, year 10, not only to have got an A in English the following year, but just for some help, I used to be able to handle a great deal of stress and trauma on my plate with easiness as its middle name, but since year 11, I swell up at the sight of pressure. In many senses what happened there won\'t ever be repaired completely, but like I stated yesterday, I\'m ready to find inner peace, and a fit man to boot, (if I can).   I\'m not a strong person. The chance of me being one died last January. So if strength is something I\'ll never posses in adundance, then maybe dignity, pride, an adequate esteem and a lovely companion for eternity would possibly fill the void? Well something has to or I\'ll never get anywhere in life.   I don\'t expect anyone to ever comment on these things, or even read there way through all of it, this \"thing\" is a self-help system, but just writing about it helps, little by little. There\'s not much more to say on things in school which I\'d like people to have open access to, even my old or newish friends, So I\'ll just leave you beautiful people pondering what it is,  or  gossiping on what an \"attention seeking gossiping queen\" I am. Whilst you do that, I shall earn money from Tesco. Ciao  xBuBzx  '),(7920,3000240,'20,July,2004','2004-07-20 00:00:00','   Keane - Bedshaped    Many\'s the time I ran with you down, The rainy roads of your old town, Many the lives we lived in each day And buried altogether, Don\'t laugh at me, Don\'t look away.  You\'ll follow me back, With the sun in your eyes, And on your own, Bedshaped, And legs of stone, You\'ll knock on my door, And up we\'ll go, In white light, I don\'t think so, But what do I know?, What do I know?, I know!  I know you think I\'m holding you down, And I\'ve fallen by the wayside now, And I don\'t understand the same things as you, But I do.   Don\'t laugh at me, Don\'t look away .  You\'ll follow me back With the sun in your eyes, And on your own, Bedshaped, And legs of stone, You\'ll knock on my door, And up we\'ll go, In white light, I don\'t think so, But what do I know?, What do I know?, I know!  And up we\'ll go, In white light, I don\'t think so, But what do I know?, What do I know?, I know! xBuBzx  '),(7921,3000240,'20,July,2004','2004-07-20 00:00:00','   Well with Jen finding her inner peace, I\'m ready to find mine.   Although mine is a whole other kettle/ocean of fish in it\'s own right. God at least that girl can get her a stable guy, and quite rightly, she\'s an amazing girl in all aspects.   Last week went rather well, I got in some overtime and saw a good gig in view to the left, who are by far the best local band I\'ve witnessed, one which the gig whore will have to accompany me to next time; and had an enlightening chatter with a skinhead about \"the streets\", and no, I don\'t mean Mike Skinner.   Well this weeks going well, me and Father calmed down after our heated phone conversation, and we most probably will be setting of on the 31st somewhere sunny FINALLY!   The other thing I\'ve been really thinking of in  till time  is what would happen If I got a terminal illness, and I reckon I\'d be ok with it, well more than others would be, and that is saying a lot about others appreciation for a wanker such as myself. Personally I\'d choose the cancer option because it always has that lovely  right to withdraw if you early enough  clause in it. My whole family seems to have been stricken upon at some point in there life, so it wouldn\'t be a great shame If I got it anyway. I doubt anyone would kick up a great fuss about me anyway. Well it would piss me of more if people I haven\'t seen say fro a year from the Kingston/Halliford days started  \"mourning\"  me, and saying how much I meant to them and visa versa; like  you  would have made the effort to stay in touch if you cared wouldn\'t you - Anisha et al.  Ian\'s work interview went swimmingly from what I gathered, and so now he will probably be designing pickey buildings in a place  near you  at a close point in the future. Hoorah, I bet everyone will appreciate the dildo shaped buildings and cock shaped rooves, as it drugingly becomes the archetypical  fashion  for 2004. This all calmed him down from yesterday, where Dan once again showed his true colors with \"oh I\'m too knackered to take part in football in 5 hours after my devastating 4 hour shift I\'ve just come of\" routine. Some people will never change their  slyness . xBuBzx  '),(7922,3000240,'20,July,2004','2004-07-20 00:00:00','   All I want is a 2 week holiday. &nbsp; Ok that\'s not available, a weeks holiday? &nbsp; No; only 4 full days half board? &nbsp; No, Ah yes that\'s because of an incompetant father. &nbsp; See it really gets under my skin when freeloaders get holiday without having to pay a dime. Take for example Rachael in Minorca or Frances in New York - both god damn free!&nbsp;I\'ve worked my arse of for the chance of some sun for the first time in 3 years, and for what? &nbsp; \"Oh Matthew, your Father will pay . . . GOD DAMMIT THIS IS MY HOLIDAY WHICH YOU KNOW I\'VE SAVED UP FOR, STOP KISSING MY ARSE YOU IGNARAMOUS!\" &nbsp; Well I\'m ready to get back in to the writing groove again, but after work mon ami. I shall think of something riveting whilst serving the oldies for 4 hours tonight. Ciao.  xBuBzx  '),(7923,3000240,'18,July,2004','2004-07-18 00:00:00','   Goodness, the people at blogger sure do like to test the odd frightful soul for the limits of there sanity - with all the mod cons and expansional net gimmicks flurrying in on greater validity and numericy. Enough with the commenting on the corporate influence -   this   is what they intended amidst their devious ploying:  gossip and more gossip  within there gospel.  &nbsp;  Why however, am I talking to you&nbsp;upon such an early hour, within a vast&nbsp;dry writing season personally speaking. Well, sleeping never really got back to basics, so here I am.  &nbsp;    BuBz Diet Update      Moderate/good since our last dialogue, need to curb the nocturnal munching somewhat, and I reckon that with the array of cuisine shrinking by the day, this diet won\'t be as hard as I presumed it would have been.    &nbsp;  I think in the last few days, I must have pataken in more high/hippy/with a large transparent townie front drugs session. I\'ve done bong, week, skunk and after a while of doing it, as usual with activities I join in with when I gather a certain confidence in myself, I take a huge liking to it, and so that\'s why at 3am on Saturday, I could just barely catch a glimmer of coud 9 as it sailed past Jen\'s living room window. Anyway, I shall reveal more of recent happenings at a more excessive/mentally accesable&nbsp;period. Ciao . . .'),(7924,3000240,'18,July,2004','2004-07-18 00:00:00','   urlLink    The plaque commemorating \"Thomas The Tank Engine\".  '),(7925,3000240,'30,July,2004','2004-07-30 00:00:00','   I won\'t be blogging for an entire week, which in terms of this site, is a pretty distant period in between.  So I shall try to conjure up something remotely worthwhile to ponder over in the short forthcoming absence we shall encounter.  Hmm, well it appears hard just to create something earth shattering at the first attempt, amidst the first minute - let me think for a while.  I have been thinking of Reading, and more directly, the people associated with going, well the old people. Rarely I think about them, but I wish I saw them more. If your by some freak coincidence reading any of this, and it applies to you in any significant relevance, then I hope you appreciate the sentiment.  Anyway, the kick ass polo shirt I paid homage to yesterday has been BuBzified - meaning it now has embraced sweat alloyed into it, but still, or even more so it\'s funky and always will be.  As for the entertainment I promised -  urlLink enjoy  xBuBzx  '),(7926,3000240,'29,July,2004','2004-07-29 00:00:00','   Caught back up with the Blues in the lovely town of Pittsburg so let\'s get started straight away! 25317 on hand for this one with the total looking evenly split once again. No place for Mutu tonight, or Tiago. Couldn\'t tell if it was Tiago or Carvalho running sprints after the game (think it was Tiago) but clearly they need a bit more fitness.  You\'ve by now seen the lineup, with much the same as the second half of the Celtic game. Kez and GudJ had a go at their chemistry; Cole, Parker, Maka, and Geremi in the middle; Terry and Johnson (again) in central defense with Baba and Ferriera out wide. Once again absent was any presence on the left of midfield, especially going forward. Cole was left to roam and gave Roma quite a time, as he popped up here, there and everywhere. He went forward any chance given, and hassled any white shirt stupid enough to hold on to the ball too long. A tenacious bulldog would be a great description. The first goal was due to his persistence as he battled the ball free, dropped it off to Geremi who returned it through for Joe to score a quick low shot ... 1-0 Chelsea. Geremi\'s involvement was bound to be there as he was excellent throughout; great switch-field passing, two excellent threaded, weighted passes through to Kez, and the one-two with Cole all showed Geremi what appears to be JM\'s choice for right midfield. He\'s played both games on the right and tonight was better than against Celtic. Nothing from the left of midfield however and it seems that when Robben or Duff are not in the side, the play is dominated from the other side of the field. Not one cross, as Cole moved in too many other circles in different areas of the pitch. GJ ran into space on the left a couple of times, but wasn\'t able to cause any trouble for Roma. He went missing for quite a while also and probably didn\'t have his best game. Parker and Maka were both solid but not spectacular and both appeared to defend more than attack. Parker linked very well but passed forward much more than press up into anything remotely offensive. I found the two in central midfield an odd mix frankly, and were it not for Cole, it may have remained 0-0.   All the back four were solid. JT was outstanding, Ferriera quite comfortable, and Johnson and Baba played very good as well. Good forward movement from both the right and left. Up front, like I said, GJ was not a factor, but Kezman was every bit the 30+ goals the Dutch league got out of him last season. While the calls for offsides were numerous, I think the total in the first half alone was five, what can you expect against the Italian\'s defense. They absolutely had no attack and marked our movements fairly well. They had the off-side trap down, that much is for sure. But Kez is very scappy, and very persistent. He does appear to be a bit of a loose cannon also - lots of arm waving at the calls, hasseling around the ball, and moving constantly when off it. He is going to be a handful for any team we face, especially the Arse.   Half-time switches saw Gallas for Johnson and Bridge for Baba to make what more than likely looks like our back four. I see no change in our record shut-outs or goals against. Solid and sure. Not to many set pieces to contend with tonight so the clean sheet should be a boost for sure. JT in particular was fearless and everybit the Captain. Every bit!   Cole made room for Robben on the left, Lamps and Smertin took over duties in the center, and Geremi remained on the right. I can see this also as our starting midfield, especially as Duff works into his fitness. Smertin has really impressed and no one should be afraid of him in place of Maka. He is so controlled and his passing is spot on. Maka tends to run a lot, assasin style, winning balls and breaking up moves forward by the opposition - basically what a good DM should do. Smertin does the same, but then in a flash has linked up with the front for a quick move forward. I didn\'t rate him last season, but he is on so far. Lamps again went missing for spells, threw a few errant passes out, and all and all did not make his presence known like we know he can.   Up front for the second half was what I believe many have been waiting to see - the partnership of Kezman and Drogba and I for one was impressed with what I saw. They played off each other very very well. While both had great movement and were looking to always score from a midfield assist, they also worked off each other well. There movement is what I enjoyed the best and will just give fits to so many teams. Robben\'s run down the left produced the second goal as he pulled it back at the last possible moment - Kezman was there to crash it home. The third was from Lamp\'s hard low shot which couldn\'t be handled by Roma\'s keeper, bounced out about a step or so only to have Drogba come crashing in to poke home. Very quick onto the ball he was - a true poacher at work!!   But before that we had the sending off of Kezman. To be fair, he was trying to standup for a teammate, which I like. However, he could have let the ref do his thing and Dacourt would have walked and Chelsea could have gone on to score 5 or 6 goals. Also to be fair, I think Kezman was rushing Dacourt to simply get in his face but as he charged him, Dacourt responded defensively with a punch, Kezman got locked up with him, and the rest is history. Both were ejected and probably rightly so. I hope the red card doesn\'t effect any Premiership games. Robben limped off a few minutes after and Duff came on. Again, many balls down the left, as they were with Robben. Duff also had a number of cheeky little moves and deft touches. He had a sweet one touch-back heel to Lampard and looked everybit ready for the season to start. He only played 20 minutes but it was fun to watch. Again, Robben or Duff is the hardest question to answer for me.   All and all Chelsea were pure class and out played Roma from the start. I think they had maybe two shots on goal, created very little chances throughout, and never threatened. Totti was the only bright light for them, and then only barely. He went off after 60 minutes with nothing to show for an average night\'s work. There was something like 8 yelllow cards, compared to just 1 against Celtic, with some nasty tackles from both sides, Dacourt\'s being the absolute worse. The second half 11 could be the one and I for one could live with it. Believe Cudicini is going the full 90 minutes against Milan. I think it\'s his spot to lose but Cech sure is not making it very easy for him. This could be the season lads!   MoM - Geremi/Cole first half - Drogba for the second. SongoM - Uh, crickets chirping maybe? No atmosphere sadly to say and the whole game could do with some singing.  WhineroM - Totti for sure. If he spent as much time running as he did complaining and putting his hair behind his ears, he might impress. Never a real factor but hey, HE thinks the ponytail looks good!  Wild man from the hills oM - THE KRAZED KEZMAN! I thought he was gonna pull a Cantona flying drop kick on Dacourt for a second there. No kidding.  FansoM - My two mates who joined me for their first ever soccer match!!! I explained a few things along the way but I think CFC has two more converts. My missionary work is done for the day. xBuBzx  '),(7927,3000240,'29,July,2004','2004-07-29 00:00:00','   Yes, this time tomorrow, I shall be packing for the lovely resort of Nice in southern France.   Can\'t wait - enough said.   I\'ve just purchased this phonomenal polo shirt for Â£5 from the  Officers Club  in Walton, which I\'m deeply proud of, and another coup in the transfer market (even though everything from the  Officers Club  is basically a coup considering there is a 70% sale virtually 365 days a year) for my ever improving fashion taste (mwah I sound so obnoxious there don\'t I).   So my preminition/plan for the next couple of days is:   Preparing my clothes and . . .   A&nbsp;new CD for the trip to keep me occupied.    Andrew, Claire and Harry are on there way to Spain anyway tomorrow (damn them they got there before me and Father). I&nbsp;viewed the temperature for Nice and Spain andvarious other&nbsp;destinations&nbsp;for the forthcoming&nbsp;5 days, and it\'s around 40 degrees&nbsp;God damnit; the first \" heatwave \" of the summer for Europe is approaching anyway, and August has never really been a great time of year to go on holiday anywhere in Europe -&nbsp;far too hot, as on my return I suspect&nbsp;I\'ll&nbsp;be a&nbsp;re-invigorated, gangsteresque BLACK MAN; If it gets any hotter at least.   Anyway, I was looking at the brochure for New York, and my&nbsp;goodness there\'s so many things I want to do in such little time when I\'m there, like go to the WWE cafe,&nbsp; Times Square  and  Central Park  and if I play my cards right,&nbsp;the  sex and the city tour  lol, to name just a few.   I\'ve downloaded all these great songs for a new CD, primarily&nbsp;for the plane trip&nbsp;and the daily 9 hour&nbsp;beach session. My personal favorite in the previous 5 minutes having been  Chocolate  by  Snow Patrol  and  Scooby Snacks  by the&nbsp; Fun Loving Criminals . It\'s looking very indie,&nbsp;however very good; the most convincing, mature, and consistent record/album I\'ve copied to date, and I\'ve copied around 18 in my shot&nbsp;year and a half&nbsp;of illegal download. Half of which I\'ve lost here&nbsp;and there.   Thing are on the up as far as&nbsp;I can&nbsp;reckoned with. All it would take would be to discover some lovely young&nbsp;study in southern France and fuck about with him. Well maybe that\'s pushing it, but yes, indeed, things are swell and swelling and smelling of VICTORY!!!  xBuBzx  '),(7928,3000240,'28,July,2004','2004-07-28 00:00:00','   Gay Erotic Story - The Pastors Son    I knew I was gay when I was seven years old. That was five years ago. I had experimented with this boy that lived in my apartment complex back then and liked it. Word got around and I became the neighborhood whore, sucking every boy in town and satisfying everyone\'s dark desires.      So, anyhow, I\'m twelve years old, and have had more sex in my life than most thirty year olds. Not that I\'m complaining, of course. I just want you to understand me a little bit, a little prelude to the story about me and the pastor\'s son.      Okay. That\'s enough. Anyways, me, my dad, and my sister all moved into this new neighborhood and set out finding a new church. We stumbled upon this little tiny church that\'s services were held in the banquet hall of the pastor\'s Italian restaurant. It wasn\'t very big, and there wasn\'t a big congregation.      But it had David.      David was Pastor Tommy\'s son. He was the third of four boys born to Tommy and Kelly DeMario, and he was a year younger than I was. He had dirty blonde hair and brown hair. He had a smooth look to him that kept my dick hard all during church service. And the fact that he was supposed to be so holy drove me wild! I couldn\'t wait to get an opportunity to screw with him, if I could.      So, me and David became good friends and called each other on the phone in between the Sunday and Wednesday church services. Nothing much was said about sex, because he didn\'t know too much about it, unlike me, but I didn\'t want to come right out and flirt. That would scare him off.      But, the first sleepover we had changed everything.      He came over with his overnight bag on Saturday. He was scheduled to spend the day with us, sleep over, and then go to church with us the following day, when we\'d give him back to his parents.      David and I were too young to go out by ourselves, to my dad took all of us to a movie. Earlier in the week, I had discovered my dad\'s secret porn stash that had about three movies in it. I already had a VCR in my bedroom because I liked to watch old wrestling videos and my dad hated it, so let me have my own VCR.      So, when we were walking around the mall that the movie theatre was in, my dad and my sister walked ahead while me and David hung back and had our own conversation about wrestling. When I was certain we were out of earshot, I turned to my friend.      \"I found my dad\'s porn videos. Wanna watch them tonight?\"      \"Uh, okay,\" he replied. \"We won\'t get caught, will we?\"      \"No. When my dad\'s asleep, there\'s no waking him up. And Gloria is about the same. Plus she don\'t come into my room anyhow. So we\'re safe.\" So we agreed to stay up all night watching pornos.      It was about midnight when everyone in the house was asleep. When we were certain that everyone else was in la la land, David and I snuck out of my room and into the kitchen. Very quietly I got a chair and climbed up to the top of the cabinets above the stove. I grabbed the plastic bag of porn very slowly so the bag didn\'t rustle and wake my dad up, because he slept in the living room of our two bedroom apartment. When I had the bag and was safely on the ground, we put the chair bag and silently ran back to my bedroom and closed and locked the door.      I turned on the light long enough to see the titles of the movies. We decided to watch the longer one first. So, we sat back and watched the hour long film. It had different scenes of men and women fucking, and one scene of lesbians. I only got hard because I knew that David was horny, and plus I liked looking at the dicks going in and out of the pussies and mouths.      When that movie was finished, I glanced at David\'s crotch and sure enough, there was a bulge there. I didn\'t want to make a move just yet, I wanted him to be really horny before I tried anything. The second film we watched was about twenty minutes and was all Chinese lesbians masturbating themselves and fucking their pussies with dildos.      That movie finished and I asked David if he was hard. \"Yeah. Very.\"      \"Want to jack off?\"      \"What\'s that?\" he asked me.      I couldn\'t believe that at eleven years old he didn\'t know how to masturbate, but then I realized the he was the pastor\'s son and shit. \"It\'s a thing you do with your dick to make it feel really good,\" I told him. \"You take it in your hand and move your hand up and down really fast. Wanna try?\"      \"I don\'t want you to see my dick, though. That\'s gay.\"      \"It\'s okay, David,\" I assured the innocent boy. \"Boys jack off in front of boys all the time.\"      He reluctantly agreed to do it. So, I pulled down the shorts I wore to bed and took my pulsating penis out of my briefs. He did the same. He kept his eyes glued to me as I started to jack. Then he took his own dick in his hands an mimicked me.      We just laid there and jerked off, watching each other. He suddenly stopped doing it and sighed. I thought he had a dry orgasm, but I didn\'t see him shake or anything.      \"What happened, David?\"      \"I\'m too tired to jack off anymore, Billy.\"      \"I\'ll do it for you.\"      I reached over and took his hard cock in my hand. He reached down and slapped my hand off of it.      \"Don\'t touch my dick!\"      \"It\'s okay, David,\" I said. \"Nothing wrong with jacking another boy off. I do it all the time. It feels good when another person does it instead of you. Here, jack me off and I\'ll jack you. Trust me, you\'ll like it.\"      So, he reached over with his right hand and clumsily started to jack me. I used my left hand to jerk his big dick. He started to get into it and so did I. Every second it felt better and better. What made it feel the best was knowing that such a hot boy was doing this to me. I never have been with a boy as attractive as David DeMario before, and haven\'t since. He was gorgeous, and my hand was wrapped around his circumcised holy meat!      Again, the pastor\'s son stopped. He announced that it was boring. So, not wanting to offend him, I stopped and we trailed off into silence. I lay there, counting the minutes, wondering when would be a good time to go further. After five minutes, I decided it was as good a time as any.      \"What do you think dicks taste like, David?\"      \"I don\'t know.\"      \"Have you ever tried to lick your own or suck it or something?\" I asked him.      \"Yeah, once or twice. I couldn\'t reach it.\"      \"Well, you can lick or suck or do whatever to mine as long as I can do it to you.\"      \"Would that be gay?\" he asked me. \"Isn\'t a sin to do this?\"      \"No, and no. We\'re just having fun. God likes fun, right?\"      \"I guess so,\" he answered.      \"Can I taste yours first, please?\"      \"I guess,\" he said.      So, I got down on the floor and kneeled on it. My face was hanging over the bed, right near David\'s crotch. His dick was soft, but mine was rock hard. I darted my tongue out and licked David\'s head. I felt the boy shudder. I ran my tongue all around his delicious cock head and started to flick it. I felt and saw his dick stiffen. I licked down the length of his semi-hard boy meat and was now at his smooth balls. I took one in my mouth and started to suck on it. I could feel his dick against my cheek, very warm because of all the blood that rushed to his cock.      I rubbed my cheek against this part of his anatomy while I rubbed his balls. I reached my face over and my mouth accepted David\'s large (well, large to me back then) cock. I took it into my mouth, deep throating that bad boy and sucking for all I was worth. I was sucking so hard that my ears actually popped from the suction. I was having a ball sucking that sexy as fuck boy off.      He had to kill the mood by talking. \"You\'ve done this before, haven\'t you, Billy?\"      I made a sound that passed for a yes.      \"I can tell. So, you are gay, then?\"      Again, I made the noise. I would not take my lips off of this dick. He\'d have to kill me first!      \"That\'s okay. I like this a lot, and I really want to know what dick tastes like now. Any way you can keep sucking my dick while I sucked yours?\"      So, I showed him the ol\' 69 position. We were both laying on my king size bed, my dick in his mouth, his dick in mine. I was sucking him like I\'d never suck a cock again. He was sucking me almost like he was afraid to be doing it.      I took his dick out of my mouth long enough just to ask a quick question. \"You like how dick tastes?\"      \"I like how your dick tastes, Billy,\" he answered me. \"I never sucked another dick, so I can\'t judge them all. But I do like yours, and it does feel so good. But, it feels like I\'m about to pee.\"      I knew he was close, that suddenly that made me close. Now, I just started to make cum. I had been making it for about a month or so. So, after a minute or two of intense cock sucking, I suddenly filled David\'s mouth with cum. He didn\'t know what to do at first, but had no choice but to swallow. But, it put him over the edge, because I felt him have his first dry orgasm. He was shaking so hard, and he looked so confused, yet pleased.      When our orgasms slowed down and my cum was devoured, David looked at me.      \"Was that white stuff sperm?\"      \"Cum, David,\" I told him. \"Same stuff, but everyone calls it cum. Good, wasn\'t it?\"      \"Yeah. Kinda sweet, but kinda not. Hot and thick. Just good.\"      \"Now, don\'t tell Pastor Tommy about this. You don\'t want a whooping do you?\"      \"No,\" he said. \"But I do want to sleep over again sometime. I liked this a lot.\"      \"So did I, David. So did I.\" xBuBzx  '),(7929,3000240,'28,July,2004','2004-07-28 00:00:00','   I have just finished watching an enthralling film with the ever so Russian looking Johnny Depp looking GORGEOUS in  The Ninth Gate .  Anyway the day has been quite alright as days go, and as far as days go in the last couple of weeks, its weather refuses to deteorate.  According to Mother and Father, I can being a friend along at new years to  NEW YORK  damnit! I should love to bring Anna or IAN! Yes Ian and myself kicking the yanks arse would be a site to behold.  Anyway that\'s it as far as today went.  Tomorrow I\'m of to the movies with Megan and to get this top I\'ve wanted for ages (God if I don\'t sound gay there, then I never will *sigh*) - should be fun. xBuBzx  '),(7930,3000240,'28,July,2004','2004-07-28 00:00:00','   urlLink    Anyway . . .  '),(7931,3000240,'26,July,2004','2004-07-26 00:00:00','   Yes, no epic yet I\'m afraid (I swear, I\'m never traditionally so failatic) as it\'s only half past one in the morning; however this is just a mere tiny update on GoThIcA LiFe amidst its never ending empirically fallicle eternity.  Last week, before  The German  went of to Alicante for a fortnight, we became closer than ever. We were in the student room of her house (which strangely after so many stayovers, was a room that completely evaded my attention) and be-fittingly for this new room; well we found a new experience to our friendship so to say. After the fondling, kissing, hugging and outright naughtiness, and I was too scared/dubious to try anymore, and I knew I didn\'t trust myself around her for the foreseeable evening, and so I insisted I disappeared.    That  however wasn\'t before I learnt I might be meeting the worlds cuuuutest guy - yes that\'s right - CHARLIE HUNAMM!!! Yes Barbara is fixing a new series with 2 of the cast including blondie and I may be able to come along for the \"ride\".  After this, Jen put the iceing on the weekly cake, with the fact that I will not be seeing Anna until the  13th Slimelight Townie vs Goth  edition, easily more bareable with the news that she had made contact with Skittles AT LAST, and tonight she had a KICK ARSE haircut - lovely to behold.  Anyway in the last few days I\'ve been to the peel twice and the cinema several times. All films - Spiderman 2, Around The World In 80 Days, and well with the exception of the kids orientated  Thunderbirds , were all fabulous, and the two gigs involving my evil twin, or evil twin, or whatever they are both pronnounced, were wonderful. This even though I hardly saw either of them, as Dan, Ali and tonight Baldwin convinced me that the strip bar behind the peel was far greater value for the evening.  Well naked women must have some appeal, but I haven\'t really seen it a great deal recently, oh Jesus that probably makes me queer doesn\'t it. Fuck it\'s not like anyone who ever reads this blog ever gives a shit and bothers to ever bring this or any subject up with me to help me with my criminalistic issues, so fuck it; I couldn\'t care less.   Lastly - Ian needs emotional support in the coming weeks/months after some truly traumatic turn of events in his life, and I refuse him a day without checking up o E-Man\'s mental stability, even in Nice I\'ll try to manage the feat somehow. It would be rather inappropriate to spill the beans on this blog about the recent happenings with Ian, I mean if he wants to tell you what\'s happened, then get in contact and friendship with the great man himself, but if you ever read this Ian, then you know I\'m here for you whatever. xBuBzx  '),(7932,3000240,'26,July,2004','2004-07-26 00:00:00','   Much adoo about nothing has been parading its way through life in the last couple of days, and therefore blogging has been reduced to a minimum, however soon therewillbe an epic I assure you - many regards: BuBz xBuBzx  '),(7933,4238152,'15,August,2004','2004-08-15 00:00:00','   Musings At Anderson By: Nofing  Of late, Iâ€™ve come to notice how seriously the school values their studentâ€™s opinions. This is especially so for the less positive, less pleasant and well, cruder remarks. Not that Iâ€™ve failed to notice over my past three years at Anderson but what Mr. Ang mentioned just recently made me sit up and wonder what really is their problem or our problem? In school, there are only two sorts of bright sparks â€“ those at 3/7 and those at 3/1. Those in between are just  in â€“ betweens . See, those at 3/7 are nice, respectable students who do their work and keep to themselves. Those at the other side of the rainbow however, are a darker shade of  black . They are bright no doubt and self â€“ opinioned. They are the people PM Lee wants (donâ€™t wait to be invited to tea instead take theâ€¦). Unfortunately, the same cannot be said for our brethren Dr. Chee Soon Juan. Ironically the situation is also the same at Anderson. However, the issue Mr. Ang raised was of similar nature. A secondary 2 student had been voicing his strong opinions on his online blog. Not that I am praising him but certainly there must be room for creativity and self â€“ reflection? Of course, the nature and content of his online entries must be taken into consideration. To what extent was it crude and slanderous, which at the same time is very unbecoming of Andersonians. Certainly voicing out oneâ€™s feelings isnâ€™t wrong or else esteemed writers like Mr. Humphrey and Mr. Brown would have been taken to task? However, cracking blue jokes and touching on controversial issues like what an extremely unwise Sheikh Haikel did would get you into unwanted trouble. As much as I wish to contend with the fact that airing oneâ€™s opinions is perfectly fine as long as one toes the line, it is disappointing to note that Mr. Ang also found usage of vulgarities and dialects of unprintable nature hard to swallow. I am one would enjoys the beauty of languages. The less refined bits are part and parcel of it not that it is encouraged but like the finer points of any language, it allows for better expression. I feel that instead of condemning these words and phrases alike, we must learn to accept them and realize that at times, it is that special word that bridges the gap. Of course, Mr. Angâ€™s cause for concern is understandable. Singaporeans are a refined lot like their British counter â€“ parts where the f word is used without so much as a bat of the eyelid. Dialects are an art. They represent our forefatherâ€™s culture. Their way of life, which at that time was tough â€“ going, and their manner of speech was likewise influenced. Many a time, I find it strange that people take pleasure in back â€“ biting others. I do not stand for that. Instead, we must take pride in our words and actions. Others should not determine our lives; neither should someone else dictate our words and actions. We live in a society of freedom of speech where as far as Iâ€™m concerned is as fair as doing a â€œCheeâ€. I find it all contradictory â€“ indeed there is allowance for freedom of speech but like the â€œCheeâ€ incident, but like the governments stand on cults, it is fine as long as it does not put the countryâ€™s welfare/interest/â€¦ at risk. When I say taking pleasure in back â€“ biting others, I am of course referring to Andersonâ€™s latest case of defamatory remarks. Obviously, the culprit must have heard of Ericâ€™s misdemeanors. The sad thing is, this guy hid behind a mask. By posting his comments online, he has placed fellow writer Eric and myself in bad light. Indeed we have committed offences of similar nature but we â€œclaimed the creditâ€ didnâ€™t we? Though uncommon, his isnâ€™t the first case that Iâ€™ve heard of. Its has happened on our level (sec 3) too but of course being kind and gracious and out of respect for the parties involved, I shall not give names. I feel that Mr. Angâ€™s take on studentâ€™s opinions is a little over the limit. Indeed, like the government, he has the schoolâ€™s welfare at heart. But what about us â€“ the student body? Has the school as a whole ever considered the plight we writers are in? We have no medium though which we can express ourselves, not online much less on paper. So, tell me please, where does that leave us? We are merely giving our 2 cents worth and also humoring our friends. Of course this shouldnâ€™t be done at the expense of someone else say a teacher in this case. But certainly just some harmless comments about the teacher wouldnâ€™t be in the red? Certainly, we will not make such statements without evidence but if there is cause and justification for it, why not? What the sec 2 boy did was definitely out of the question because from what I know, he made hurtful comments about teachersâ€™ and classmatesâ€™ physical appearances. Obviously this greenhorn has lots to learn about being witty yet within limits or should I say lessons on maturity? I feel what he did was way under the belt. To discredit people is already putting one foot in the grave. To hide behind a mask, nailing the coffin. Perhaps, it is time for change, a revolution, and a better tomorrow as our beloved politicians love to proclaim with much enthusiasm though I believe it less and less each day. For starters, lets have a private online forum where only true â€“ blue Anderson writers would be allowed. These might be a source of comfort for Mr. Ang because at the very least, our more meaningful remarks would be kept within ourselves. An interesting point to note is that every time such an incident crops up, Mr. Ang never fails to make mention of it but he only notifies us of the problem. What about his proposed solution? For three good years Iâ€™ve heard nothing but his â€˜pressing issuesâ€™ even in sec 1 when I was none the wiser, he didnâ€™t have any long â€“ term solution. To spell it out simply, the school on its part is doing nothing than providing short â€“ term solutions but in the long run, how effective would such a method be? From the way I see it, their methods are more of deterrence to would be  offenders  and nothing more. Such a tried and tested method would soon work off all its magic. At times, I really have this urge to grab my trusty pen and work my  magic  but my run â€“ in with the school or more specifically Mr. Ang has  deterred  me. As you can see, all is not lost. The schoolâ€™s tried and tested method still holds and hopefully so until I graduate. And when that time comes, I would say,â€ Au revoir Anderson! Never fret, itâ€™s never too late to teach an  old dog  new tricks.â€   Nofing is a sec 3 student would takes Mr. Angâ€™s opinions to heart ever since his run â€“ in with him a long time ago and has been faithfully chronically the more meaningful happenings at Anderson.   '),(7934,4238152,'17,August,2004','2004-08-17 00:00:00','   By: Nofing  Teachers have an inferiority complex. Itâ€™s very obvious â€“ take a look at good old Mr. Ang. His catch phrase though I only heard it once, was â€œI know I look like a fool talking to you people on stage.â€ Real fascinating. And much recently, the most  respectable  Mdm. Neo joined those ranks too. Her catch phrase, â€œI know that even before I start (speaking) some of you will call me â€˜ siao â€™ but I have to do so for the sake of the class.â€ Sounds like a true blue PAP supporter eh? Now before I digress into PAP politics, let me make mention of Mdm. Neoâ€™s most  commendable  speech all of which happened within the four walls with a hole in class 3 / 2. The bell was just about to go. I was as usual waiting to bid farewell being the most courteous student that I am. Today had been a most meaningful Geography lesson. Why? Because Shawn was seating with us and Bo Hui had apparently skipped the first lesson, lost count of time and didnâ€™t even turn up for school, which, is a good thing. Shawn had served very well as my  personal entertainer  but in Mdm Neoâ€™s undisputed opinion, he was a personal disaster. And disaster there was! Mdm. Neo took it t herself to berate the two of us. Quite simply, she stood at the OHP as she always does (because no one will think of having her near themselves) and for the ninth time looked as if though she was on the verge of an emotional breakdown, placed her marker down, glared at us (Rather contradictory, beats me too. But Mdm. Neo is most unpredictable at times) and started her little speech. â€œI donâ€™t wish to say this but there are a few students who are very disruptive. They choose not to or cannot fir into society and form their own group or a delinquent subculture. They value things which society doesnâ€™t value a lot.â€ Pauses to look at Shawn and I then draws some weird diagram on the OHP before continuing, â€œThings like rudeness and defiance,â€ pauses again to look at us. And so her speech goes, punctured with sporadic pauses now and then and also injected with Shawnâ€™s cynical remarks and our laughter. Little is known as to why Mdm. Neo embarked on her highly controversial and questionable speech. Some kind souls who were most touched by her speech said that she was sincere for once. Not very credible but itâ€™ll do. Shawn on the other hand had an entirely different opinion. â€œSincerity will not melt stone.â€ That sounds more like my idea of the  catch phrase  of the year! Indeed Mdm. Neo must have been a sorry sight. But I beg to differ. One, we (Shawn and I) have come to a conclusion â€“ That must have been a publicity stunt which further proves her mental instability. We donâ€™t mean that in a bad way. All we want to say is Mdm. Neo seemed rather out of touch today like as if though she was leaving us (which I hope she would and also leave for good) and that speech was some sort of farewell speech. Two, we wish to thank her for her kindness and concern but could she please wake up to the notion of actions speak louder than words and help us ace our Geography instead? Or must she in what is supposed to be thoughtful and sincere insult us. Not that she came hard on but she played it cool with nearly 80% of her little speech directed at us. We applaud her for her subtleness but hope that as soon as  sows  learn to fly, so would she learn the true meaning of sincerity. This refers to being forthright and to the point. For example, Shawn and I were discussing the different ways to shut Mdm. Neo up. Suggestions raised included throwing the pen knife at her, flinging paper balls and last but not least, walking out of the class. Unfortunately, we were mentally sound and ran through those ideas once more and the verdict was â€“ lets just sit here enjoy the air con and annoy Mdm. Neo. This might sound very passive in comparism to the aforementioned ideas and we canâ€™t agree more. For now, we shall follow by example â€“ that is to say learn from Mdm. Neo. Sheâ€™s not the only one who can do things subtlety. We shall show her our appreciation by sitting together during her future lessons and see how long itâ€™ll take before she yells at us and flings the OHP over (Now you know why she stands next to the OHP and not outside the class). Bearing that in mind, as good OP supporters, we shall declare, â€œThat has been long enoughâ€ I know, I know, youâ€™re referring to my article.  Nofing is still unable to get over the incident, having been influenced by Mdm. Neoâ€™s mental instability; he now takes to drawing a flying sow wearing a cloak bearing the alphabets NEO on it. Of course we all know heâ€™s a big fan of the matrix.   '),(7935,4238152,'18,August,2004','2004-08-18 00:00:00','   By: Nofing  Just today, Ms. Yoong my English teacher for two years posed an interesting question to the class, â€œWhat does marriage mean to you?â€ I, on one of these rare occasions was at a loss for words. So much more when asked to express my opinion to the entire class? Having thought over this issue carefully, I, sad to say still have not come up with a satisfactory answer. When I stood in front of the class, the first thing that came rushing into my mind was, â€œHow am I supposed to make my answer  unconventional ?â€ I thought and thought but nothing not even suggestions from fellow classmates could do the trick. This ranged from, â€œMarriage is the legal consent to have carnal knowledge of someone,â€ to â€œMarriage equals to free sex.â€ In a nutshell, this boils down to a naughty three â€“ letter word â€“ sex. Indeed in this more conservative Asian society of ours, sex is seldom an appropriate topic at the dinner table. However, to equate marriage to sex does seem to be a rather flawed equation. Iâ€™m sure that by reading the aforementioned opinions from  well â€“ meaning  friends, one would doubt the company I mix with. However, before you present the verdict, lets be honest. At the tender age of fifteen, which hot â€“ blooded male would resist the thought of sex; much less a lifetime of free sex? In comparism, if such a question was to be answered by the opposite sex, Iâ€™m sure the response given would have been entirely different. This might (Iâ€™m just guessing) range from, â€œ A sense of security,â€ to â€œeverlasting love.â€ Honestly speaking, which girls wouldnâ€™t want to dwell in such a fantasy, which is seldom possible in this fickle, demanding world? Having explored the possible responses from both genders, one would come to a conclusion that the meaning of marriage goes beyond moral issues but to fundamental ones like the biological make up of men and women alike. I do not claim to be an expert in this area but as far as Iâ€™m concerned, men being more  primitive , will definitely wish to satisfy basic needs before moving on to more complex needs and wants which require greater intellect which is why some men stop at brothels having barely satisfied their basic needs. Women, on the other hand are more sentimental. Apparently, this includes their basic instinct. I mean, itâ€™s the way the human body is designed â€“ do first, talk later. Which is why women would rather be loved emotionally than to have it expressed physically. However, this cannot be said for all women where there are exceptions of women leaving a marriage because the men cannot satisfy their basic needs. Then again, the same can be said for men where the few exceptions also wish to be loved emotionally and the physical aspect need not necessarily come into the picture. My purpose of writing this article isnâ€™t to find the meaning of marriage but to discuss its effectiveness as an â€˜Oâ€™ level examination question. Indeed this was Tuesdayâ€™s â€˜Oâ€™ level English oral examination question. I believe that the purpose of the GCE â€˜Oâ€™ level examinations is meant to test the proficiency of students in various areas of study. However, such a question when probed further goes beyond superficial issues like sex. It requires much thinking before answering not because it takes time to find a satisfactory answer which the examiner will appreciate but also to see if it conflicts with ones beliefs. Being a Christian by faith, I find it hard to provide an  unconventional  answer for such a question. In fact, to come up with a  decent  answer would prove to be a difficult task what more unconventional one! As a Christian, there are many  laws  so to say governing marriage. It is seen as a sacred process, which every Christian should go through only once in a lifetime. Which is why providing answers relating to â€˜one night standsâ€™ and answers of similar nature would not be right. On the contrary, it is stated that men and women are made for each other but to the extent of free sex? That might be testing the limits. I do realize that what I mention here seems to be rather contradictory with regard to my actions. However, one must realize that what I say outside does not go to the point of conflicting with my faith. No matter how â€˜ out of boundsâ€™  my statements may have been, one must realize that it only flirts around the surface, teasing it lightly and that is all. But to answer such a question, even at the best of my ability, my response is as follows. Marriage is the unity of two imperfect humans who regardless of their genders, can live together in harmony and accommodate each otherâ€™s flaws and appreciate them for who they are and not what they are or what they can provide. Marriage is more than an act of love but also a lifetime commitment. In the event of either party being unfaithful, instead of walking the easier route of divorcing, they should take the difficult but permanent route, which should involve professional counseling for both parties and the support of their families. Marriage is another stage in life, which further emphasizes the need for compromise. On a larger scale, this can be put to practice between countries where compromise amongst the parities involved can prevent conflict, which might lead to war. Therefore, marriage is not just an act of love but also a valuable lesson in life. Now, even as a Christian, I must occasionally  forget  myself and therefore the other type of response which if I donâ€™t make mention of, I would certainly live to regret. When two humans in the heat of the moment tie the knot, this is called a shot â€“ gun marriage. Superstars who include Britney Spears have acted out such scenarios. Of course when such an unfortunate incident arises, it will definitely lead to a divorce. However, there is a way around this. It is known as a one â€“ nightstand for the fickle minded and a long â€“ time (platonic) relationship. In both instances, the concept of legality is given a miss and this, might give the couple(s) involved a false sense of security. Whatever the case, it beats polygamy, which as you people should know is against the law. In conclusion, marriage is a must for law â€“ breakers (committing polygamy) who want to have the best of both worlds â€“ sex and breaking the law.     Nofing is still confused and is trying hard to learn a valuable lesson in life (compromise) without getting married. Obviously, he needs to find a balance between the secular world and his religion, which more often than not is impossible.   '),(7936,4238152,'18,August,2004','2004-08-18 00:00:00','   By: Nofing  Iâ€™ve learnt the meaning of standing up for your beliefs. It all happened today. When I had a â€˜baptism of fireâ€™ so to say. It made me realize one thing â€“ being with bad eggs will cause you to turn bad as well. Itâ€™s just so sad that I had to lose friends along the way. But as the old saying goes (or is it my  saying ) gone with the bad in with the good. Whatever the case, Iâ€™ve found out that Iâ€™m still good â€“ At least where my beliefs are concerned. I donâ€™t stand for calling others â€˜fatherless bastardsâ€™; nor disgracing my family or myself; nor being so thick â€“ skinned to the extent of having everyone walk out on you. I had better stop here before things get personal. But as you can see the person in question is none other than Ron. The stinkingâ€¦ â€¦you knowâ€¦ â€¦ Iâ€™m amazed that heâ€™s still around and still acting like as though heâ€™s so cool. I find it getting more and more difficult to write things in perspective but I must try me very best before I disgrace the name of my family and myself. Now looking at things in perspective: Ron to me is a like a pool of water. Heâ€™s always the way people shape him and I will always stand by that. Oh yeah, and did I mention heâ€™s a bully? I canâ€™t help but question his credibility regarding his parentage. You see, humans have something known s a guilt conscience and will pick on others who are in the same boat because, like a smoke â€“ screen, this would draw attention away from themselves. And Ron keeps on picking on Bo Hui and Wai Keat. Why? For the record, Bo Huiâ€™s father died of cancer when Bo Hui was young and Wai Keat has a medical problem. I am not specially mentioning this two extremely pleasant souls to spite them or anything but surely one would see the irony of this all. Having been a friend of Ron for some time, I now know the reason behind his being retained and Iâ€™m none too surprised. In fact, Iâ€™m grateful for him because he has walked the beaten path and I now know how to avoid the same  ending  as him. With every passing word, I am feeling more and more ashamed of myself. Instances of Ron and his unfair treatment towards others have come flooding back. Connections regarding a case, which I missed out, have joined back to complete the jigsaw. I would like to take this opportunity to apologize to those whom I may have mistreated or whom I did not offer help to when Ron unfairly treated them. I also urge all there people to come forward and speak up against such injustice. To put it across bluntly, I do not want the level to be dominated by a  retainee  and his  sidekicks . The sad thing is, many of his so â€“ called â€˜sidekicksâ€™ are in fact fair weather friends. I admit that I was indeed one of them but having straightened my thinking; I can now safely watch Ron destroy himself from afar. Enough of Ron; lets move on to further examine the different characters who â€˜hang outâ€™ with him.  Losers  No hard feelings please; this group of people are the misfits of society and to escape from being picked on by others, they join a group for temporary protection. I pity this group of people the most because their situation is akin to that of choosing between the devil and the deep blue sea.  Manipulators  Like I said earlier, Ron is like a pool of water. This group of people has a measurable intellect and uses this to manipulate Ron for their personal gains. I personally despise this group of people the most. I mean Ron is retained and thatâ€™s no need to further emphasize  this  by making use of him. Then again, he deserves it.  Man on the Street  The man on the street is innocent but finds Ron and company interesting. He or she will â€˜hang outâ€™ with them for a while until he loses interest and leaves the group. Now that the various groups of people who might be associated with Ron have been listed, I must as Mdm. Tang once said, â€œIf you canâ€™t be a solution, donâ€™t be a problem.â€ I shall now attempt to find ways and means to end Ronâ€™s tyranny or more likely the manipulators tyranny. Like I said, the people Ron mixes with define him. So I suggest we place Ron amongst a group of monkeys. As the saying goes, â€œMonkey see; monkey do.â€ In this case, theres nothing for the monkeys to see or do so theyâ€™ll just pass time scratching each other. (Wait! How can they scratch each other if no one teaches them to? But Iâ€™m sure Ron can teach them. After all  monkey see; monkey do .) For the losers, Iâ€™m at a loss for words. They are so pitiful yet at the same time they arenâ€™t exactly the ones whom you would want to interact with usually. So, I propose we send them for an Adam Khoo course and see if they can change their believes. After all, Iâ€™m a loser so are you. Wait a minute that sounds a little wrong doesnâ€™t it? For the manipulators, what can I say? I mean they constantly prey on the inferior lot of people. Therefore, the probable solution would be to isolate them on a far away island like Bali where theyâ€™ll be bombed by the terrorists who apparently are already manipulated by someone else. Like they always say getting a taste of your own medicine. And last but not least, the man on the street. Nothing new here; theyâ€™ll just look around, maybe be part of the group for a period of time before leaving. No harm done but while he or she is at not, why not try manipulating Ron? Its jolly good fun you know? Iâ€™ve said my piece, Iâ€™ve lost a friendship I wished I never had and Iâ€™m glad for that. After all, one should see the way Ronâ€™s  ex  â€“ classmates treat him â€“ with such distaste like as though they had seen something awful like a  garbage bag , which leaves a foul aftertaste.  Nofing has not gotten over this and swears to fight it out with Ron to the end. Unfortunately, the after effects of sublime influence have made themselves known and he now spends his time giving new meanings to the acronym RON. Examples include, â€˜Rotten Old Nigerâ€™ and â€˜Retained Overaged Nitwitâ€™.   '),(7937,4238152,'22,August,2004','2004-08-22 00:00:00','   By: Nofing  We all know of nations fighting over territory. Over  landmarks  like  Pedra Branca , which further extends a nationâ€™s territory. But classes experiencing border crisis? This unusual topic has been tackled by my fellow writer Eric and Iâ€™m here to present yet another case of border crisis. Take a bow Eric. The temporary classroom block if not good for nothing still serves a purpose â€“ rattling the chipboard walls and see how long itâ€™ll be before the partition drops. After all, the whole structure is  temporary  isnâ€™t it? Perhaps more light should be shed on this issue of border crisis. As far as I know, it originated from Anderson, when students who sat near these partitions grew bored of the lesson (which is quite often, ands thatâ€™s saying something) they started throwing their weight against it. This caused the partition to rock violently and thereby disrupting the neighboring classâ€™s lesson. It has proven to be great fun when one side bangs and the other returns the favor. And as humans evolved for the sake of evolution, so has the banging of these partitions. As in order with this yearâ€™s theme â€“ magical, enterprising and wow, Andersonians have taken it to heart. A splendid example would be the mutilation of the partitions. Major Chong is a busy man but recent events have made him busier. You see, holes in these  so â€“ called  walls are a  so â€“ called  eyesore whereas in my opinion, makes the school look more like the  darker  side of LA. Indeed, such a scene is not appealing to everyone but  compromises  must be made right? So, instead of making holes in the wall, students have taken to removing the wall altogether. This in itself is magical â€“ how often do you get to witness such a ceremony? Enterprising â€“ if the school ever needs construction works to rebuild the school, here are some ideal ones. And wow â€“ I mean, wowâ€¦literally; students are really getting the hand of multitasking arenâ€™t they? Now, removing the wall isnâ€™t too difficult. In fact, its part of the cause and effect cycle of banging it. The partition itself is only meant to allocate the space to each class and determine the classâ€™s boundaries. Unlike bomb shelters, these walls are unable to withstand vigorous banging or rocking. And of course, when the glue (can you believe it? The school is so  enterprising . Temporary, really!) is torn apart from the wall, which has already began, the wall will have to fall due to an unstable equilibrium and gravitational forces. And being enterprising students, they must make the best of the fallen wall. So theyâ€™ll see which side has a disliked teacher/student and have the wall fall towards that direction. So far, there have been no incidents of such nature and I sincerely hope so since Iâ€™m sitting next to the wall but as a good student, I would now like to share some of there magical techniques on banging the wall, which will certainly make you, go â€œwowâ€.  Beginnerâ€™s guide to banging the wall  The good â€˜ole wayâ€¦  Simply ram your elbow  ala  a rugby tackle at the wall and marvel at Newtonâ€™s laws and the conservation of energy. Note, this method no matter how macho it makes you feels is ill â€“ advised, as you will feel great discomfort in your elbow the following day.  Knock, knock, whoâ€™s there?  This method was developed by the normal stream of 2002 last year during reading sessions in the hall where they repeatedly slammed their reading materials on the floor much to the supervising teacherâ€™s annoyance. This method has proven to be a hit among passive students who only want to make noise without having any fatalities. Like the normal stream have done, just take a hard object with a decent surface area of preferably 100cm2 and slam it against the wall repeatedly. Guaranteed satisfaction, which will make the DC, that follows more bearable. Last but not leastâ€¦  The retaineeâ€™s methodâ€¦  Yes, yes as a tribute to Ron who started this method, those who wish to follow suit neednâ€™t worry about being retained; obviously, that requires more  skill . This method involves shouting vulgarities at the wall in the belief that sound travels through hard objects faster. Obviously, the nut head hasnâ€™t been doing much reading up ( retained ?) otherwise he would have realized that the wall is make up of chipboard and sponge which doesnâ€™t propagate sound but on the contrary absorbs it. Not to be defeated, there is a plan B. This involves throwing or banging (or a combination of both) either a basketball or chair at the wall. The result would be the students at the other side of the wall retaliate in kind or ask Mr. Ang to witness the event.  A few words of advise before I end this piece and proceed to fine tuning my methods of wall banging, one must realize that at the end of the day, the school is where you receive your education and not where you bang walls. Banging of walls or fellow students should preferably be done outside school.  Nofing has learnt his lesson and has stopped banging the wall. Unfortunately, those nut heads led by Ron on the other side of the wall have not learnt their lesson. He feels that those losers should go bang their heads against a wall or bang someone. She bangs, she bangsâ€¦   '),(7938,3490316,'31,May,2004','2004-05-31 00:00:00','   Well today is my Friday, sorta. I have off tomorrow so that is good, I can get some rest. These new 12 hour shifts at work are a little tiring, but they pay off, cause we have a 3 day weekend every other weekend. I will just be glad to lay around and accomplish nothing but laziness tomorrow. Though there are several productive things I could be doing, staying home helps me to save money. Seeing as whenever I go out I end up spending some amount of money I should have saved.  I will most likely lay around and write some in my book. I have been flooded with ideas lately so I am going to take advantage of it.  /End Blog for now. '),(7939,3490316,'30,May,2004','2004-05-30 00:00:00','   OK this is a very touchy subject for me and at some point I reach this point where I have to vent, and since other people can do it so publicly then why can\'t I.  Why war you ask? There are plenty of reasons sadly enough many think that this war is merely a distraction. I can further you more on that, it is not. I worked in the military, in a field where I handled and decoded much knowledge of what exactly America is in threat of, and why persay there is this war. No I can not speak of what I knew then and the knowledge that I obtained from my training and the job I did for my entire military career. But I will tell you this, threats on our freedom? You bet your little ass there are some threats. I can guaran-damn-tee it!  You can call me all the names you want. You can put me in the category in which you put all the other self-righteous, dimwitted, deluded fools. But know this as I speak from experience: Our VERY NATIONS LIBERTY IS AT STAKE HERE!   Lets see about under imminent threat of nukes, oh wait~EVERY DAMN DAY. Every day we are in danger of losing our freedom. Maybe not right away, but why not break us down bit by bit, city by city, friend to family member. Slowly but surely bring the great nation to her knees.  There is yes a hairy and not so attractive maggot that DID pose a very serious threat to this nation. You weren\'t there, you are the general public here is a secret, there is ALOT the friggin \"general public\" does NOT know. Oh and probably for your own damn good, as I can only imagine the sick state of panic you would all be in if you really knew. Does this maggot pose a threat to us anymore, him personally no, his contacts and pupils, you better damn well believe it.  Then there are those of you who say we should not support those that protect our freedom, that we should protect our nation from our own assfucks that govern her. News flash, I don\'t see in office. You should fit right in as some of them can relate, I will agree there are a few government officials that could pull their head out of their asses. I wont argue with you there.  Oh and those soldiers that would be put to death for treason had they choose not to take part in the war, here\'s another news flash: When you sign up, you expect it. Wether it benefits America or not, its what we do. We fight for freedom, America\'s above all else, but freedom is why we give our very lives. Think about it, from the past to the present, without America where in the hell would the world be? No really answer me it might actually humor me.  Am I saying that I am right on all aspects of this post, no. Am I making a point, maybe. But we are all graced with the ability to voice our freedom of speech, something else we defend might I remind you. So this is what I have to say to all those of you who simply look for every friggin damn excuse to bitch, whine and spit about how shitty a descion this war is, its my 2 cents of freedom of speech. Take it how you want it.  If Uncle Sam called me tomorrow off inactive status I would gladly go, because I know what I would be fighting for. I guess its a different side of the tracks and maybe that is why our views are different.   Do I think it\'s time to bring our troops home, YES. But there is obviously a reason why they are not coming home, and I bet it is for a worthy cause.  /Rant off  Bonnie Rae '),(7940,3490316,'30,May,2004','2004-05-30 00:00:00','   Well hello bloggers! Today is Sunday the second day of the three day weekend that everyone has but I am working all three days, lovely! Course it pays off because next weekend I will have a 4 day weekend! WOOT! And my mother and little sister are coming to visit so there will be many funfilled things planned I am sure!  Its funny because on my way to work I can think of so many things to blog about, but once I sit down in front of the PC my mind goes blank. Anyhoo I did get an email from a very good friend of mine today~ Woot YAY Ryan! I haven\'t talked to him in gosh months so I was very excited to see that he had sent me an email. /Wave Ryan~ He is one of my old Everquest buddies. I haven\'t been able to play Everquest in like a year, but I will have my apartment soon in which I will be able to pull my poor boxed up pc out of its hibernation state and actually put it to good use again. YAY! Course then I am sure I will have some pc problem that I can\'t fix myself and I will have to call the \"God of PC fixitness\" Shane (my bestest friend way over in California) to have him walk me through trouble shooting. /Hugs Shane. What would I do without you!  I haven\'t been able to write lately in either of my books, it really bites but I always go through these phases. Some days pass and its an endless reel of writers blockage, then all of a sudden something will pop in my head and I cant get the words on paper fast enough. I prefer to write all my works by hand first, then transpose them over to type. I am weird, I feel like writing comes from the heart and soul, I feel more connected to the piece knowing that it came straight from within me and released on paper. I have always thought, \"The true writer/author is the one with the ink stained hands\", but that\'s me. I know that there are very talented writers that only type, I guess it is what ever floats your boat. =D  I will close for now, may be back later~ I leave you with a simple question though, ask yourself......  Does change set you free?  I heard that in a song today and it really made me think, how bout you?  Bonnie Rae '),(7941,3490316,'29,May,2004','2004-05-29 00:00:00','   Hello all! I am Bonnie Rae, I am 24 years old and reside in central Florida. I am originally from California around the San Diego area. I served 4 years in the US Air Force and was stationed at Biloxi, MS the entire time. Sad thing is I actually joined the military to travel, but I stayed at the same base my entire first term and it was really well boring to say the least. Don\'t get me wrong my job was exciting but Mississippi was rather a bit of a culture shock compared to busy California! My whole family moved to Florida so when I got out I moved here as well.  Right now I currently work as a police dispatcher, which is rather interesting. It is definitely something different everyday! It also gives me an insight as to what police officers go through and how they handle situations. In the spring of 2005 I plan on going through the police academy, and hope to become a K9 cop with in a couple of years. Though all this may change as I have just recently switched my major in college to International Business. My life long dream is to travel the world, exciting!  I am a avid writer and love all things fantasy based. I am currently working on two novels, one of which is a series. Will I ever get them published, I have no idea. I also love poetry and write daily in my personal journal!  that\'s all I will bore you with at the moment as all this will be in my profile! Looking forward to blogging with the best of ya! '),(7942,3490316,'28,June,2004','2004-06-28 00:00:00','   Well I am off again for another two days. I have a serious doctors appointment tomorrow that I am a little nervous about. I am kinda hoping it will be one of my last ones for a good while! But alas, no one can tell the future!  Sorry I have not been able to keep up regularly with my entries, but once this stress of doctors and what not is gone I am sure it will release my mind! So tootles for now but see ya real soon!  Love Bonnie Rae '),(7943,3490316,'23,June,2004','2004-06-23 00:00:00','   My Lifetime Love, My One True Friend By: Bonnie Rae  She laid her head upon his chest. A lifetime of mortal love, undying passion now comes to rest.  He holds her closer to him, Watches the wind unsettle her golden hair. Embracing a lifetime, Letting go just doesnâ€™t seem fair.  The first time they touched was all on fates chance. Taking his rough hands, she led him in their first dance.  That captivating smile, Embedded in his mind. That comforting laugh Emerald eyes so kind.  A lifetime ago they pledged their love and a kiss, under the willow tree. He holds back stinging tears, for now he must set her, his truest love, free.  Her breaths now grow quieter still. His heart breaks for she makes it whole. How can he live without her? Who now will have his soul?  â€œWe were meant for each other,â€ he whispers soft. â€œNot in this lifetime my darling love,â€ she weakly replies. Tears fall from his eyes and he holds her dear. â€œIâ€™ve always loved you,â€ he finally cries.  She lastly speaks, â€œDo not weep my love for we shall meet again, This is only one lifetime and surely not the end.â€  You never say goodbye, But Iâ€™ll see you again. You were my lifetime love, You are my forever one true friend.  This poem was inspired by two characters in one of my novels I am diligently working on finishing. Just thought I would share with you all!  Bonnie Rae '),(7944,3490316,'23,June,2004','2004-06-23 00:00:00','   My dearest friend, 	I donâ€™t even know how to begin, to tell you the tale of my aching heart. The treachery that in lies my web tangled soul. You came to me at a very hard time in my life, and somehow helped me to see a light. But I tossed you away like cold trash in the morning light. I was a fool, and we have gotten past that; or so it seems. But I still sit all my days and wonder why I canâ€™t explain what you mean to me, not even to my own self. You know every inch of my soul, every beat of my heart. You have my laugh memorized and would be able to find my smile in a sea of transparent illusions of human beings. You whisper to me in the wind even though you are so far away, always there when life seems so hard. Black tears fall endlessly, not even waterproof mascara can hide this pain. I lay and stare at the velvet night, so many stars. The moon witnesses my endless tortured tears; I can only wonder if the moon can understand. Wrapped in the dark of midnight, swallowed up by my own confusion or maybe itâ€™s an inner madness I shall never beat. I donâ€™t think I am crazy, nor does the moon. I think I just canâ€™t understand why my heart beats so heavily for you. Its not fear, or I would come running free. Itâ€™s not remembrance of others, or there would be no reason for this letter at all. Itâ€™s just simply unexplainable at this moment to all and especially me. Have you always been with me? In all my lives? Is that why I feel so close to you, so connect straight in my soul? You are innocents in me that is pure and real, that I know because that much I can feel; with heart and soul. Forgive me for this letter, but I have no other ways to express this confusion of my mind. No way to outlet the stinging tears that feel like newly sharpened knives. You are what you are to me, and at least every time I think of you a smile. At least I have that little blissful piece to keep me going. I sigh as I truly do not even know how to end this letter, this plea, and again wonder how did it even start?  Always with love Me '),(7945,3490316,'22,June,2004','2004-06-22 00:00:00','   Scorn  Something fathoming From deep within Something burrowing Wrapped in sin  Taking over my body Controlling me Entangled in deadly vines Eyes pierced, I cannot see  Blood falls, deep crimson From a reopened wound Fed toxic poison, vile From his â€œsilverâ€ spoon   Now I lie dying, slow and raw Skinned pierced by sour thorns When I am reborn, remember still The horrible wrath, of a womanâ€™s scorn  Bonnie Rae  Just thought I would share. Dont know where it came from the words just flowed into my head. '),(7946,3490316,'18,June,2004','2004-06-18 00:00:00','   I am off soon for my three day weekend, YAY. Though I am broke and will be doing nothing but renting movies and working on my book! But before I go I leave you with this. Do you remember your first love? After reading a novel by Nicholas Sparks, I thought I would quote one of his writings:  \"The first time you fall in love, it changes your life forever and no matter how hard you try the feeling never goes away. Your first love, no matter what you do, stays with you forever.\" (Nicholas Sparks, Author)  That could not be more true!  Bonnie Rae '),(7947,3490316,'17,June,2004','2004-06-17 00:00:00','   I am so tired today! I am working the day shift and I only received 4 hours of real sleep last night! Due to the fact that I was in pain half the night.  Well my doctors appointment went, ok. Course I am left with antibiotics and of course no pain medicine that I can take while working, dang it. So believe it or not I am sitting here at work with a heating pad, is that pathetic or what. Anyhoo what can one do?  Lets see, not much else has been going on. Just taking it easy as the doctor ordered. This weekend is my 3 day weekend so I am excited about that, in which I will be doing more resting I am sure. YAY for getting to sleep in, ALOT. Even though I sleep in like everyday cause I am not do for work until 2pm each  day, but still. Sleeping in and not having to get up and go anywhere is a lot better~ trust me!  Bonnie Rae '),(7948,3490316,'14,June,2004','2004-06-14 00:00:00','   The air is warm and saturated with the smell of newly cooked bread, banana perhaps. Tall blades of grass glide amongst my legs as the plush earth caresses my bare toes. There are no prickles or thorns amongst this ground, so I am free to roam with naked feet. No worries, only the love for life. An endless summer day in the perfect world, of this dream. Running freely to the shore, racing the seagulls in their epic flight. Pausing only momentarily to collect a shell or two. Dress wet now from playing in the emerald ocean, suddenly I realize I am not alone. The white mare, again now for the fourth day, comes to pay me a visit. She keeps her distance but acknowledges that I am there. She is my friend and I hers. We have an unspoken bond, both curious but unafraid. I watch her frolic to and fro from the waves and up the sand salted banks of the pearl white shore. I bid her good day as I have now for this fourth time, just as the last, leaving her some carrots atop the same rock as the days before. Running free so fulfilled, nothing here is vacant in me. Racing the wind I tumble to the soft ground, the grass catches me. Laughing I look to the sky, so blue, I canâ€™t stop smiling. '),(7949,3490316,'10,June,2004','2004-06-10 00:00:00','   I know I haven\'t updated in a few days, forgive me. I have been dealing with alot and working on top of it. But I am off tomorrow so hopefully I can get some rest!  My mother and little sister have gone back home, so I was saddened some by that. But we had a blast while they were here!  Life seems to pummel on. As do many things within my life, but lately I have been feeling as if I am standing back watching my life take place without me. Its wierd I know, but to me it makes perfect sense.   La de da~  Bonnie Rae '),(7950,3490316,'10,June,2004','2004-06-10 00:00:00','   Tonight and the Rest of My Life By: Nina Gordon  Down to the earth I fell with dripping wings Heavy things wonÂ’t fly And the sky might catch on fire And burn the axis of the world thatÂ’s why I prefer a sunless sky To the glittering and stinging in my eyes And I feel so light This is all I want to feel tonight And I feel so light Tonight and the rest of my life Gleaming in the dark sea IÂ’m as light as air floating there breathlessly When the dream dissolves I open up my eyes I realize that Everything is shoreless sea A weightlessness is passing over me  This is a wonderful song, and it is how I am feeling at the moment. Such is life, I guess. I was off for four days back to work today then off two more. This week is a breeze or so one would think. I am happy because my mom and my little sister are visiting, and we have had a blast. But my mom can tell I seem worried about something. I guess moms know best. I just have a lot of complicated things going on right now and sometimes life can seem so bleak. I tend to live worrying about making other people happy always sacrificing my own happiness. I hide it well I guess but sometimes I really am falling apart at the seams.  Both my best friends live tons of miles away from me. I try to talk to my boyfriend and sometimes he really is comforting. I know I have a ton of family and friends that love me, but why do I feel so damn empty, so alone. I guess that is a famous question asked by many, one that can never truly be answered. For one can only answer their own questions, those questions that haunt us day in and day out in our soul all the days of our lives.  I guess that is when faith steps in, I feel the need to pray. I will be the first to admit I am not as into my faith and religion as one should be, but I believe in the Lord. Sometimes he is the bestest friend one can have. I also can feel Brandon\'s presence around me, he is caressing me with his warm spirit.  Some people go through life and never know true love, I have loved and felt true love. Its the one blessing I am forever thankful for...  Bonnie Rae '),(7951,3490316,'04,June,2004','2004-06-04 00:00:00','   Wooot! Today is FRIDAY! TGIF!!! And I am off for FOUR DAYS! SO excited, can\'t you tell?  Ok so there is this song that I love, by Evanesence, of course. Because they are the best, which on that note I am excited because I get to see them in concert July 27, 2004, so sweet! OK but anyhoo~ So there is this song by them that I just love and it reminds me of two characters in my book and their undying friendship, William and Erina. They are my most favorite story line out of all the storylines that I write, maybe because I have a friendship similar to theirs. In tribute to them and the friendship that I have I want to share this song\'s words with you!  Will you forgive me again. I don\'t know what I said, but I didn\'t mean to hurt you. I heard the words come out. I felt that I would die, it hurt me so much to hurt you.  Then you noticed me. You\'re not shouting anymore, you\'re silently broken. I\'d give anything now, to kill those words for you. Each time I say something I regret, I cry, I don\'t want to lose you. But somehow I know that you will never leave me.  Because you were made for me. Somehow I\'ll make you see, how happy you make me. I can\'t live this life without you by my side. I need you to survive. So stay with me.  You look into my eyes and I am screaming inside that, I\'m sorry.  Song By -Evanesence-  Could they be anymore talented? What a beautiful song and so beautifully written, you can feel the passion in it. Fits the mood I am in, and the friendships I talked of earlier. You know who you are ;P!  I think this is a good way to end the week and take off for a great FOUR day weekend! Blog ya later loves!  Bonnie Rae '),(7952,3490316,'02,June,2004','2004-06-02 00:00:00','   So I have been sitting around yesterday and today just thinking. I keep asking myself what it is that would truly make me happy career wise, and I have always wanted to do something where I make a difference in this world. What I would love to do is be a zoologist, and travel the world studying and educating people on Big Cats. But the rest of my college tution for that is way to expensive and seems so far out of reach.  So hence the wanting to become a K9 police officer, I still get to work with animals and make a difference. But my dream is to study aboard in Africa, and aren\'t we told all our lives to do the best that we can to fullfill our dreams? I am 24 now, I am getting old, not fast but it is happening. I dont want to run out of time on my dreams and I have yet to feel that I have fullfilled any of them. You just know when something is your calling, because it calls out to you the minute you see it. Everytime I see a big cat, whether in a zoo, magazine or tv~ I get that feeling in the pit of my stomach. That jolt from my soul. I know this might sound pretty lame, but its true. And its how I am feeling at the moment.  What does one do? The practical? Or take whatever leap and bound to catch the dream? What would you do?  ~Bonnie Rae~ '),(7953,3490316,'01,June,2004','2004-06-01 00:00:00','   WOW, new month already. Where have the days gone? June in Florida, ewww HOT! It is like 90 something degrees in the shade, thats sick! If it does not rain today I think I will frequent the pool! Need to work on my tan!  I am off today, YAY. ~Happy Dance~ Bring on the laziness!! I will blog ya later!  ~Bonnie Rae~ '),(7954,3490316,'31,July,2004','2004-07-31 00:00:00','   Well, we put Jeff\'s cat to sleep today. It was sad and I really felt bad for him, but it was really for the best. We buried him in the backyard in a really nice spot in the garden area. Jeff said he wanted him buried there cause I guess that is where they have some of Jeff\'s mom\'s things. Course we didn\'t plan on hitting every damn root from every damn tree in the yard, but with patience and a really nice axe (yes we actually had to axe the shit out of some of those damn roots) we buried Maxwell. He lived a long and happy 17 years of life, and is certianly in a better place then all of us now.      Anyhoo after that whole deal I went back to bed and slept for about another 5 hours, since I only got about 3 hours sleep last night.      The slut biznitch never called back, so we are guessing whatever she wanted can\'t be that imporant. Jeff is getting the number changed Monday - good thing too!      Anyhoo I think I will venture to the kitchen and see if I have any cheez it\'s left. They were Jeff\'s but I kinda took them over and told him to go and get his own damn box! *evil grin*      ~Bonnie Rae~ '),(7955,3490316,'30,July,2004','2004-07-30 00:00:00','   Everytime   By: Britney Spears      (hush I like her!)      Notice me Take my hand Why are we Strangers when Our love is strong Why carry on without me? Everytime I try to fly I fall without my wings I feel so small I guess I need you baby And everytime I see you in my dreams I see your face, you\'re haunting me I guess I need you baby I make believe That you are here It\'s the only way I see clear What have I done You seem to move on easy Everytime I try to fly I fall without my wings I feel so small I guess I need you baby Everytime I see you in my dreams I see your face, you\'re haunting me I guess I need you baby I may have made it rain Please forgive me My weakness caused you pain And this song is my sorry At night I pray That soon your face Will fade away Everytime I try to fly I fall without my wings I feel so small I guess I need you baby And everytime I see you in my dreams I see your face, you\'re haunting me I guess I need you baby '),(7956,3490316,'30,July,2004','2004-07-30 00:00:00','   I HAVE COME TO THE CONCLUSION THAT ALL MEN SUCK! With the exception of the slight few that actually exist and give a shit about what women feel (which is like almost an extinct animal at this point).  Anyhoo so I am sitting at work - doing what I do there, working. And my cell phone rings so naturally because we are not busy I answer it. Low and F***ing behold it is this same slut ass two bit hoe that has called the cell before asking for my boyfriend. Now, yes I use his phone, and I know he has not talked to this girl in well over a year. BUT - still it pisses me off. This is the same slut that called us at 2am one morning in fact waking us both from sleep. And anyone who knows me KNOWS that for me to fall asleep and actually stay asleep is a rare thing, so you can imagine how pissed I was. I asked him to please (as long as I am using the phone) to get the number changed, so I no longer would have to deal with stuff like this and he agreed. Course he never did it, laziness on his part. Laziness on my part for not making sure he followed through with getting the number changed but I since thought that this problem was solved and would not have to hear from this hooch again. That was until I so innocently answered the phone earlier this evening, innocents was shattered and in stepped my fiery Irish temper.  I was calm to her (for the most part, an extreme effort on my part - my lip was probably on the verge of bleeding from biting down on it so damn hard) but held a mature \"F*** YOU\" kind of attitude so she would get the point! She addressed she was just an old friend, YA RIGHT, to which I informed her I knew better. She then said I had nothing to worry about cause she was engaged. Now lets think about this ladies....  I KNOW I HAVE NOTHING TO WORRY ABOUT YOU SLUT BIZNITCH BECAUSE HE IS WITH ME AND NOT YOU BECAUSE YOU WERE NOTHING BUT PLAIN ASS TO HIM, YEP THAT\'S IT PLAIN I NEED TO GET OFF QUICK PIECE OF ASS! I know this for it has been confirmed, by even my best friend who knew him back when she worked with him when he mention this c*nt slut.  Now, I know that I am slightly contradicting myself through my anger, by stating that I am not worried. Now what I mean by that is,  I am not worried . I do not worry that she is any threat to me. And I trust Jeff, but as I discussed with a friend earlier (while venting I should say) it\'s women that can\'t be trusted. This hoe called once at 2am, you don\'t call some one at 2am (especially when you were merely that to him, a HOE) just to say \"hey\". Can we say BOOTY CALL. So there ya have it, she is a slut ass hoe that can\'t be trusted. You women out there that have had similar situations know what I am talking about. Oh and what \"engaged\" woman calls another mans (one she used to sleep with) cell phone? How dumb is this bitch!?!  Anyhow now that I have vented - I feel somewhat better. I told this bitch she can call back on the cell tomorrow when Jeff is with me (in a polite manner) and gave him one and only one option:  \"You tell her when she calls you have no need to converse with her or have contact with her anymore what so ever. That you are involved with some one and would appreciate it if she never again tried to contact you.\"  If he can not do that well then I told him I would just pack my shit and be out! I don\'t need this B.S.  La de f***ing da!  /Rant off /Vent over  ~Bonnie Rae~ '),(7957,3490316,'29,July,2004','2004-07-29 00:00:00','   Sooo Jeff\'s cat is like on the verge of death and he is tearing himself up about whether or not he wants to put him to sleep. I hate that he has to make such a hard choice but something needs to be done. This cat is 17 years old and Jeff\'s mother got him the cat, so I can see where he is emotionally attached to this cat - in a sense another way he holds on to the memory of his mother. It is an active memory, alive; and to lose it will only make him feel as though he is losing another piece of what little \"live\" memories he has of his mother. I can\'t imagine losing my mother so I can\'t feel his pain. I can only stand on the side lines and watch him get tackled by his endless pain. He is dealing with his mother\'s death much more healthier now then he ever has, but I ache for him. When you love some one so much you yearn to take their pain from them. No matter how much it hurts, on top of all your other pain; you want theirs as well. He means so much to me.     &nbsp;   Tomorrow is another day -     &nbsp;   ~Bonnie Rae~ '),(7958,3490316,'29,July,2004','2004-07-29 00:00:00','   Well life goes on, whether we are ready to accept certain things or not. The past 6 months have been and up and down non stop hell of a ride rollercoaster.&nbsp; I have been staying at Jeff\'s alot more lately, which is something I like. I like being with him, falling asleep next to him, having tell me he loves me first thing in the morning. But on the downside I have his family drama to deal with. Which lately has been less with his dad being away visiting his girlfriend, but his grandmother is a non stop deal. She lives in the apartment attached to the house and though a very nice old woman, she is very needy and I think her needs have been overlooked for sometime. But with everything going on in my life at the moment it is hard to have to deal with her  every  day. I like her and love that she see\'s me as some one who cares, but dang! I need to be able to breathe!     &nbsp;  Well I am off to get ready for work. I am sure if I have some time I will post more later!    &nbsp;  Bonnie Rae '),(7959,3490316,'28,July,2004','2004-07-28 00:00:00','   There are those that can see.   Then there are those that can really see... '),(7960,3490316,'27,July,2004','2004-07-27 00:00:00','   ~*CHANGE*~   ultimately I believe that:   &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Change, sets us free.   ~Bonnie Rae~ ');
INSERT INTO `msgs` VALUES (7961,3490316,'26,July,2004','2004-07-26 00:00:00','   I had a muse this eve! Hope you like:   &nbsp;  Drifter  By: Bonnie Rae   Visitor amongst the shadows  Silent friend, you come and you go  You never hang about long  Just enough to reassure and know   Drifter amongst the soul  Come and stay awhile  Sit at this mad tea party  It just might make you smile.   Come and drink some tea, letâ€™s talk.  We are just evil racing thoughts  ~Yes you have those too  We are childhood fears and forgotten what nots.   You do not visit us enough  Let us toy within your mind  Let us weaken the lasting strength  Break down that empty swine.   We come to you in the dark of night  Interrupt peaceful dreams  Awaken you with frigid sweat  Ultimately never letting you be.   If you would just stay awhile  Take an unforgiving chance  Look us in the eye, and;  Amongst the shadows dance.   You will see we are not so bad  So stop trying to push us past  Live within yourself  Make this change forever last.   &nbsp;   ~Bonnie Rae~ '),(7962,3490316,'24,July,2004','2004-07-24 00:00:00','   I need to step back inside myself, be what I have become.  Lock away the girl who see\'s. Swallow the pain and move on.  Bonnie Rae '),(7963,3490316,'22,July,2004','2004-07-22 00:00:00','   Lets see, I am not throwing up - thats a good thing, considering thats all I did last night! It was so bad that I continued to throw up from 11pm last night till 7am this morning! My damn stomach muscles ache cause I yacked so much!  Anyhoo - not that any of you are remotelly interested in me telling you my lovely \"throw up\" stories!  Soooo lets see, yesterday was Jeff\'s 21st Birthday - YAY Jeff! Course I was to damn sick to even celebrate with him, but he know\'s I love him.  Good thing is I get off work at 10pm this evening (woot only an hour to go) and I will be going straight home and crawling in bed, with my Mooshu (my kitty cat).  Already then - Blog ya later!  Bonnie Rae '),(7964,3490316,'21,July,2004','2004-07-21 00:00:00','   The title of this blog is that of a song a very good and special friend of mine introduced me to. In just listening to it, I remember times we spent together and goofy things we\'ve said and done. I remember the love we shared for one another and the angry end that we were dealt.   It is funny how at one moment of the day you can be in your own dark place. Then one person just makes the darkness all fall away, without even trying! I am sitting here with a smile I can\'t seem to wipe from my face, and it feels so good. He came into my life like a shooting star graces the night\'s velvet sky, entering quickly but everlasting in my memory; from the beauty that encompasses his free spirit and soul. I didn\'t realize how much I missed him until this evening and I am so glad I got to once again here his voice!   I am so glad for him and that he came into my life when he did. I am glad I was allowed to get close to him and I am glad he taught me how to laugh again. He is a special part of me, we had something very unique and still do. We both want more then this world has to offer and we both refuse to settle for less. It\'s a common ground we share and to one another souly understand.   There was this time after my divorce (which was hell and we won\'t get into) that I went to visit him. I will never forget he took me to this park, and it was so beautiful. We played around like two little kids free from the imprisoned chains of school on an eternal trip into the perfect summer with &nbsp;an endless&nbsp;playground. I was at peace, far away from my normal hectic hell I live in, I was at peace. I felt free and not a worry in the world. As humans I think we all look for that feeling at some time or another in our lives, but rarely do we ever find it. I was lucky to have felt it, and feel it still every time I am reminded of him, talk to him or simply just get an email from him. He is so real to me and what we had was real, it was true and for the first time in a long time&nbsp;I felt as if I was living. I say  was  as if he is in past tense to me, but that is so wrong. For after talking to him again this evening I realized that he is real, we are real.   I could go on and on, but I will close with this, words from the song he introduced me to, a song I feel we very much relate to. Thank you, for being you and for finding me and showing me all that you do. I look forward to this friendship, hey ya never know ;-)   \"I was thinking its a sign, that the freckles in our eyes are mirror images.  And when we kissed they perfectly aligned.  And I have to speculate that God himself did make us into corresponding shapes like puzzle pieces from the clay.  And true, it may seem like a stretch, but its thoughts like this that catch my troubled head when you are away and I am missing you to death.   They will see us waving from such great heights.  \"Come down,\" they will say.  But everything looks perfect from far away.  \"Come down now.\"  But we\'ll stay. \" &nbsp;~Lyrics by Postal Service~   That song makes so much sense, cause we kinda have always had our \"own little world\" that other people just don\'t understand. I guess the important thing is that we understand it.   Life is short, but it is even shorter if you never choose to live  in  it.   ~Bonnie Rae~   &nbsp; '),(7965,3490316,'20,July,2004','2004-07-20 00:00:00','   What is life? Do any of us really know? Subdued by this monotoned reality, walking effortlessly in this paradox of a suffocating world. What the hell propels us to keep moving? Is it because we are forced to? We have no other choice but to wake up and face yet another day? Have you ever just been driving along listening to whatever music fancies you at the time and suddenly just thought \"What the hell am I doing?\" Everything that you attempt to grasp as making sense falls swiftly away and&nbsp; you are left feeling emptier then the day before.   I had one of these moments today on my way to work. I was listening to a CD that a friend made me, one I favor because it is one of my lasting memories of us. But I was listening to it and I watched as two beautiful white birds flew past me, almost as if in slow motion. Like something you see out of the movies. I felt tears swell in my eyes but they never fell. I thought for a brief moment maybe it was a sign, but then struggled to analyze in what kind of sign. I have pretty much given up on those things. I have an intuition like no other, but fight to ignore it. I am blessed with a sense that most are not, and though as a child it scared me, I have grown to appreciated it as an adult. But in the recent years I have slowly taught myself to black it out, push it away. I have seen things, I have seen truth but somewhere along the way I have lost faith. Faith in all that ever made me feel safe.   Maybe I put myself in my own darkness. No, but I choose to continuously revisit it. I don\'t know how I can feel safe there being surrounded by such emptiness? Maybe it is not safety I feel, maybe this darkness is all I know.&nbsp;Maybe it is not a safety blanket but a velvet cloak of&nbsp;vacant shadows that I am wearing. I have known this feeling and not in just this life. Above all I have always known this feeling.   I miss him for he was very much a part of me, but I pushed him far away; just like the ones before. I am shrouded in a mystery that even I can not cut through. I have stopped trying to figure out how to step out of this fog, for it is eternal. I have realized that it is not an object to overcome, it is in me. It is who I was, who I am and who I will always be.&nbsp; If you can\'t accept me, you were never ment to; that much has been written amongst the stars. But if you get it, who I am, then you are a part of me; eternally.&nbsp;   *~Bonnie Rae~* '),(7966,3490316,'16,July,2004','2004-07-16 00:00:00','   Let me just start with \'yawn\'!!!! It\'s 0338 am and I am sitting at work hopelessly watching the clock. I am working 7pm to 7am shift, covering for another coworker. I hate the all nighter shift, cause around 3am my body thinks that it is time to go to bed, since that is when I am normally crawling into my bed from my own shift! Bring on the COFFEE! If I round it off, only 3 more hours to go!!!!!! Oh but it seems so far away! Then when I get home, the cable guy is due at the house between 8am and 10am to get my pc online! All I am gonna want to do when I get home is SLEEP! Not deal with cable connections and all that garb!     &nbsp;   This week alone is hectic for me! I work 4 days straight instead of the normal, 2 on 2 off. I am working 2 of those nights as 7pm to 7am, then the other two are my normal hours but I am still going to be tired as hell, cause when I am supposed to be sleeping I will be awake and vice versa! My body&nbsp;schedule is going to be totally off balance this week! But I have Friday and Saturday off so I can go out of town, that will be nice!     &nbsp;   Jeff and I have to take his dad to the airport this Friday, he will be going out of town to WI to visit&nbsp;his girlfriend for about a month! It will be kinda nice cause Jeff and I will have the house to ourselves for that long, and it will be good to see how we get along. Good test for when we get a place, or should I say \"if\" and \"when\" we get our place. But anyhoo we have to drop him off at Orlando International, so we are going to just stay the night in good old Orlando. And since Jeff turns 21 this Wednesday, we are going to celebrate then; since I have to work all night the night of his b-day!     &nbsp;   I am looking forward to this weekend. It\'s always nice to get out of town, even if it is only for 1 night. We love to play miniature golf in Orlando, they have some really kick butt neat courses!&nbsp; I should be getting my digital camera soon, hence I will finally be able to post some pics!     &nbsp;   Well I have to go and get another HUGE cup of COFFEE! Only 3 more hours to go, I CAN DO IT!!!!!!!!     &nbsp;   ~Bonnie Rae~ '),(7967,3490316,'16,July,2004','2004-07-16 00:00:00','   Well another doctors appointment has come and gone and still there is no positive answers I have been given. I am tired of shrugged shoulders and confused looks. I am human, and I am real - I can see you and hear you talking about me and all the tests; stop acting as if I am not in the room!      &nbsp;   Today is my Friday then I am off for 3 days - YAY! But next week I have one hell of a schedule!&nbsp; We (Jeff and I) are supposed to go to the springs this weekend and go canoeing with some of our friends. I am not one for paddling in a canoe, but I am game to get some sun and swim in the springs! As long as their are no alligators - those things suck! *shivers at the thought*     &nbsp;   I am excited to get to go see Nikki this October! I really miss her alot - we have been best friends for so many years now - and living in different states really sucks! I miss big spending at the mall, beach hopping in the \"maro\", Los Tres Amigos (only the best mexican food restaurant in the world), Pina Coladas and Corona\'s poolside then zzzzzz\'s. Gosh so many memories! OMG and who can forget&nbsp;Shmad and \"Danger&nbsp;Zone\"!!!&nbsp;The best friend I will forever have - I would do anything for ya sis!&nbsp; Looking forward to seeing ya!!!     &nbsp;   ~Bonnie Rae~     &nbsp;  &nbsp; '),(7968,3490316,'13,July,2004','2004-07-13 00:00:00','   Well our Evanesence tickets came yesterday - I am stoked about that! Now all I have to do is wait till the 27th then its off to the concert!  I will be off for a few days - more doctors appointments tomorrow, really not looking forward to it. I just wish they would figure out what\'s wrong already, and if there is something wrong - can it be fixed!  I am tired today and in slight pain. The pain comes and goes, the medicine doesn\'t really help so I just have to tolerate it. But I am looking forward to going home and getting what little sleep I can before my appointment tomorrow. I am more looking forward to after the appointment when I can just close the door and rest, without being bothered - or having to answer all sorts of questions. Ahh sleep my favorite pastime!  Whatever darkness be laid in front of me, let there be light in the deepest depth of thy soul. Let it fight when I have given up, and let it be victorious!  Bonnie Rae '),(7969,3490316,'11,July,2004','2004-07-11 00:00:00','   Faith By: Bonnie Rae  Waiting for the heavens On this rainy day For the angels to descend And kiss this hurt away  In this mist that binds me Shadows swirling fast Lost souls linger about decaying  I canâ€™t remember the first sin from the last  I tripped and I have fallen Upon this earth rotted ground The smell of death is putrid and copious Please angels let me be found  I didnâ€™t mean to come here Iâ€™ve somehow lost my way Please hear my prayers and save me Donâ€™t chastise me to forever stay  The vapors of death have consumed me My eyes, I canâ€™t fight to stay awake My body is now lethargic My soul is feeble and willing to break  A soft whisper in my dead ear, â€œAwake my child and play amongst the stars Let the clouds caress you and release you from hellâ€™s barsâ€  I can hear virtuous music From somewhere deep within Itâ€™s then I realize the realm of heaven Has saved me and released me from my sin.  Have faith '),(7970,3490316,'07,July,2004','2004-07-07 00:00:00','   So I have been off the past two days and in that time frame I have done alot of thinking. The one thing I think I have finally come to realize is that love finds you when you least expect. And that love can be in many forms, and is never perfect. If love was perfect then life would be boring.  In the past two months I have been back and forth with doctors and still not knowing if I can ever naturally have children, which is a huge damper on things. I have lost a good friend of mine, or some one whom I thought was a good friend of mine. He so kindly informed me he could no longer be my friend, how lovely and what great timing might I add. I have internally been dealing with my real father and his entire side of his family in the pressure of meeting my 9 year old half sister, whom knows nothing about me and I know nothing about her. That and I have not talked to my father since I was 3 years old. He never really gave a shits worth, and personally that is how I feel about him. Bitter anger, how lovely. Anyhoo - just shit I have been dealing with.   And in all this I have been in a new relationship, well not to new as we have been together now about 8 months. Not long but long enough to really start to get to know some one. I battle with alot of inner demons, many that cause me to have negative outlook\'s on relationships. But these past two days I realized that the prince charming I have always been looking for, might not be some white knight in shining armor that rides in and saves the day. He might be a common man with common thoughts, who is average in his own way. And it may be as simple as him grabbing your arm and not letting go of you, when he has realized what you and this relationship means to him. He isn\'t perfect and never will be, but he loves me and its as simple as looking in his eyes.  I have loved and I have lost. I have been scared and I have been torn to shreds, but only to be put back together again. Did he help fix it, yes. But the best thing he did was help me to allow me to fix it on my own. He may have helped me pick up some of the pieces, and find some of the ones that were lost; but he handed them over to me and let me fix them. He let me put them back together, there is no greater feeling then that. I still have a long way to go on my road to healing, as we all do. But I know now that perfect love doesn\'t exist, because real love you have to work at it and I am so glad I have learned that if you work hard enough it all pays off in the end. He might not be perfect, but neither am I. And I might get mad and scream and yell at him, but he is patient and wipes the tears when I am ready to cry. I am thankful for him and for what he doesn\'t even know that he has taught me.  Love does change us forever, and I would like to believe always for the better, because we are always learning from it!  Bonnie Rae '),(7971,3490316,'07,July,2004','2004-07-07 00:00:00','   Sooo yesterday I was all mad at my car because it was acting up on my way to work. Naturally when I learned it was supposed to rain and storm horribly I cracked a joke about maybe I would get lucky and a tree would land on my car and total it, thus then I could just get a new one. Well the tree didn\'t land on my car but two hours after I has made that comment a tree (a very large one at that) landed on my boyfriends car =/. I felt so bad for him since he LOVES his car.  luckily it did not total his car, just left very LARGE and DEEP scratches all over the car and some dents. Also broke the seal on his back window, all of witch can be fixed. It was his neighbors tree so her home owners insurance has to pay for what his auto insurance company won\'t. Needless to say he gets a nice brand new paint job out of the deal and doesn\'t have to pay any of his deductibles. Anyhoo it was a lesson in being careful for what one wishes for, or rather I should say jokes about.  Bonnie Rae '),(7972,3490316,'06,July,2004','2004-07-06 00:00:00','    '),(7973,3490316,'02,July,2004','2004-07-02 00:00:00','   What would you do if one of your bestest friends told you he felt more for you then just a friendship, and that he can no longer remain your friend because he can only have you all or nothing?     I have a best friend, whom I love beyond belief. He is one of the greatest friends I have ever had and the fear of losing him tears at me everyday. But I know how he feels for me and has felt, I try so hard to respect his feelings; as he does mine. But we live on the complete opposites of the USA, for one. I know \"true love\" knows no boundaries, but still. Now there is this issue that he can\'t just be my \"friend\" anymore, its tearing at him in and out everyday. I know what this means and it kills me. I don\'t know why I can\'t feel the same way on the same level with him, because he probably is my soulmate. HOW THE HELL DID WE WIND UP LIKE THIS!!! You are my best friend, you know me - every inch of my soul. I would do anything for you but now you are telling me you can\'t do it anymore.    Am I angry, no. Am I hurt? Of course. I love you, I always have, but I can\'t love you the way you love me. Believe me if I could I would not be sitting here at the moment. Do I want to lose you, hell no. But if I have to in order for you to be able to live happily every day for the rest of your life, I would make that ultimate sacrifice for you! You have sacrificed so much for me, and I know that. I know you don\'t want this, to have to let go and walk away from me. I know you, you don\'t want this. But I understand you when you say that if I can\'t return the love you long from me, you have to. I could never be angry with you, but I will not lie and say that this doesn\'t hurt. Its like being told you are dead, only to learn you are alive and I can\'t ever speak to you, to share your life with you again.    You said you have never had to act so brash in selfishness like this before, and I know. But you were my friend, my best friend. You weren\'t some made up being, you were real. You were real to me. And now I don\'t have you anymore, or at least left with the feeling because I don\'t feel for you what you feel for me on a romantic level. It hurts to love some one, and to especially to love some one from afar - I know this. But trust me when I say it hurts more to lose them, when you could have always had them somehow in your life. I loved and I lost, God did not take him from me, God only called him home; and I still sit to this day after all these years and wonder if he knew how much I loved and cared for him! I walked away from him, because when you love some one sometimes you have to let them go. If I would have know he was going to die a week later I would have never ever walked away. But we can\'t tell the future, that\'s why we live to take risks.    I understand what you have to do and why, my beloved friend, but know that walking away is harder then remaining. Because you are forever filled with wonder and what if\'s, and only God knows when you will see or hear from that person again. I know this because I live it everyday of my life. I only warn you because I wouldn\'t wish this torment on anyone. You\'ve never loved if you\'ve never lost. If I have to lose you I will forever be pained, but know that if you ever need me I will forever be here for you. No matter when and no matter what. I will never cut myself away from another friend again, because I would rather have you in my life then to not have you at all. You do what you have to my sweet friend, but know I am always here. '),(7974,3490316,'02,July,2004','2004-07-02 00:00:00','   I changed my blog template, this one is very nice I thought! More to come later!!  ~Bonnie Rae~ '),(7975,3490316,'12,August,2004','2004-08-12 00:00:00','   You came into my life. We danced a happy tune, but then the music died and it all ended so soon... '),(7976,3490316,'09,August,2004','2004-08-09 00:00:00','   Fall from Grace By: Bonnie Rae  I found my old ballet shoes Amongst boxes of aged dust Comforting was their site Again for dance, my feet did lust  So melancholy is this day For bloodstains I can not wash away.  Lost in a sea of infinite doubt Tormented by memories of you Confused as of why I love Hands that cause this black and blue.  Echoed piano music hauntingly plays along, As my feet and body whimsically sing this song.  Out in the rain I purge my sin Flanking the edge of an empty pool Dancing silent the with closed eyes Never remembering that Iâ€™m a fool.  Little melodies in my head Remembrances of things you said.  The rain beats down upon my skin I am soaked, sodden with a dayâ€™s grief But I feel no more pain, no more strife. This happiness from my dance I know is brief.  No one can touch me; I am as light as air. Dancing endlessly; free and with out a care.  Wind and rain now thrashing upon me, Tree branches blow viciously past. I care not, for I am not on this earth I am dancing amongst the stars at last.  â€œCome back downâ€, I hear you say, â€œPlease donâ€™t throw this all away.â€  Alas I never knew the plight I took I just wanted the dance to take me away From this hopeless life, and all the pain But you caught me and alas begged me to stay  Blonde hair now crimson, blood stained is my skirt of sheer lace. I am sorry love, I must have hit my head; dying instantly as I fell from grace.             I have no idea where this stuff comes from! But alas I share it with you, the world! Enjoy~      Bonnie Rae '),(7977,3490316,'09,August,2004','2004-08-09 00:00:00','   At some point in all our lives, we realize we have a calling. Mine came to me at a very young age when I was a child. I knew, I have always known what I wanted to do. But life isn\'t perfect and sometimes you don\'t get to take the path you wish you could, for various reasons! This is why I have not fulfilled my dream, for all those \"various\" reasons. And then you fall into that slump of, \"well I have put it off for so long and so much has happened, I just don\'t know that it is possible anymore.\"   Remember that saying: \"Where there is a will, there is a way\"?   Well that saying was renewed in me this evening while working. I don\'t know how I am going to get the ball rolling, but I am going to start researching what I am going to need to do, and equipment that I might need.      I must remember that my soul and my spirit are to free to ever be tied down. Some one told me once that I was living a pipe dream - and that none of what I wanted would ever happen. News flash for you - Up until now I realized I haven\'t been living at all. I have surrendered to what life has given me, but life gives so much more once you open your eyes. No longer will I choose to live blind when I was blessed with the ability of sight. This day I step out of the shadows, this day my life begins anew. This day my spirit and my soul are once again set free...      The path that I have chosen is a hard one. Relationships will have strain because of it, but if the persons involved can withstand it, then so be it. If they can not, as hard as it is I am sorry but I have to spread my wings and fly. I have always searched for love, when deep down I have always known that my \"true love\" would be my work, my calling. All other love would be emotional and the test of times.      I know what I want. I don\'t know that it will happen, but I have to be fair to myself and at least try. It\'s not like me to give up so easily, and I was foolish to just surrender. Here I am world - brace yourself for here I come!      I shall see you always my darling, for our souls forever dance amongst the stars; as our spirits fleet around with world indulging in our dreams. Today I am thankful, moreso then the days before.      Dreams can come true, but at a price. You shall gain and you shall sacrifice, for dreams are hard work!      ~Bonnie Rae~ '),(7978,3490316,'08,August,2004','2004-08-08 00:00:00','   I will not make this a plea~      But I have to ask...   Now that you are without me, all complete contact as you asked; are you happy? Because I miss you miserably. I know that you have gone away, but I can still feel you. And I know that you can still feel me. How is this doing either of our souls justice?      ...I just had to ask, though I am sure I will receive no response. I won\'t allow myself to stop loving you, no matter how far you turn from me; you taught me that remember. You were the only one that could open my eyes and I bet you didn\'t even know you did. I love you.      Bonnie Rae '),(7979,3490316,'07,August,2004','2004-08-07 00:00:00','   ***NO OFFENSE TO THOSE WHO ARE MORMON***      oK! I am content and comfortable in my Catholic faith! If you knock on my door, I tell you this and then you still try to push you \"Book of Mormon\" on me, you are gonna get a door slammed in your face. I am sorry, I am not trying to offend anyone by any means; but those damn people have alot of damn nerve! People ***HATE**** it when you knock on their doors and try to push you religion! So STOP DOINING IT! And please if you are going to percede, then at least pay attention to the \"NO SOLICITING\" sign. It will save you door slam in the face!      /Rant off~      Bonnie Rae '),(7980,3490316,'07,August,2004','2004-08-07 00:00:00','   I am just a normal girl, well depends on what your definition of normal is~      I am standing still as the world explodes behind me, but I do not flinch! I hold my ground and clench my fists. I shout out and end it in laughter. The world crashes down and I slip away to play amongst Jupiter. I watch the planets realign but decide dancing amongst the stars is alot more fun!      Back in reality, Oh darn. I am just a girl, I love to chew my bubble gum and wear sandals. I hate McDonald\'s (yes it is true, by lack of choice I will result in eating there but ewww), but can\'t live without Taco Bell. I love to wear blue mascara and paint my toenails. I am 24, but I still love my video games!      I write poetry and other works. It is my serenity, just like my evening or morning run (depends on when I work). I am a peacemaker, but a drama queen just the same. I love to fall in love and give it all I got! A hopeless romantic is in my bloodline! I hate spiders and love kitty cats! Mooshu my Siamese is evil to everyone but me, he is my bestest little guy!      I am just a girl at heart, a woman if you please.      I am complicated simplicity!!!!!      ~Bonnie Rae~ '),(7981,3490316,'07,August,2004','2004-08-07 00:00:00','   I wrote this poem, inspired from my novel I am currently working on.       Seduction By: Bonnie Rae  Help me to step into something A world of enchanted bliss Seduce me with your stare Allure me with your kiss  Take me unto you Touch my soul so deep Whimsical is this dream Into my virgin mind you creep  Let me savor this lust That so many wildly speak Let me believe the pleasure That so many desirably seek.  Stroke my body delicately As if it were your last The ending of your tragedy Corrupt by your savage past.  Compel me into insanity With your rhythmic love Make me cry out endlessly To the saturated stars above  I am just a pawn for you But a desire none the less You have to ravage me My breasts ache for your caress  I am innocents for you Something lost long ago Within me is a fiery burning But this you already know  Help me to step into something A world of enchanted bliss Seduce me with your stare Allure me with your kissâ€¦â€¦         Oh la la!!!!!      ~Bonnie Rae~ '),(7982,3490316,'05,August,2004','2004-08-05 00:00:00','   Exile, endless exile is where I live. It\'s not so bad once you get used to it and once it takes over well then no one can touch you.      I live in a world of endless bullshit. From normal people right down to the sick and insane, this world is fucked up. Some how I have been able to see through the crime, rundown towns and bullshit filled people. Some where along the way I was given a gift, and it helps me to see the world still and clear. When it rains I see the sun. And when it is sunny, well then I can feel the thunder, frigid rain drops amongst my bones. But the axis this world spins on is small. I feel one day the world is just going to slip, and fade away to an unseen universe of endless black. I am going to laugh all the way there.      Before they were junkies, well they were innocents. Before there was fear there was contentment. Rare as it sounds it has happened. Before there was grass and trees of jade, there was only savage dirt. Before the end there was always a way of begining, but ulimately the end beats us. Its what we fear and it kills us slowly.      I am screaming inside so loud, but on the outside I have a quiet smile. You\'ll never see it coming. When I come crashing down, just like the world~      Think about it~      Bonnie Rae '),(7983,3490316,'04,August,2004','2004-08-04 00:00:00','   I read a fellow poets poem today - and it reminded me of my bastard of a father! How he abused both me and my mother.  Funny that even though I was only 4, I remember he held a shot gun to me and my mother and tried to kill us. It is because of him I hate the dark. She was the mother of his child, and his own child. How can you kill them - or even think it. Amazing what drugs do to some one. Can I forgive, I don\'t know. I have no pleasant memory of the man and as far as I am concerned at this moment I still desire no contact with him. He\'s changed my ass~ to bad he can\'t \"change\" my internal and external scars!!      /Rant off      ~Bonnie Rae~ '),(7984,3490316,'03,August,2004','2004-08-03 00:00:00','   This is a song by Jessica Simpson, the words remind me of a very special friend I had.      \"You Don\'t Have To Let Go\"   I don\'t need your strength anymore cause you\'ve made me strong   You may not see the woman in me    That you dreamed of holding me in you arms    All the days that you gave   All the moments you\'ve saved me,   Praying for my life Sacrificed,   just to make me who I am on my own   You don\'t have to let go   You don\'t say it   But it\'s in your eyes   All the fears of good-bye   But I can promise   You\'ll always have a place, and a way to my heart   All the days that you gave   All the moments you\'ve saved me,   Praying for my life   Sacrificed, just to make me who I am on my own   You don\'t have to let goI can live   Cause you lived for me   And I can love   Because you loved me.   All the days that you gave   All the moments you\'ve saved me,   Praying for my life   Sacrificed, just to make me who I am on my own   You don\'t have to let go   ~Jessica Simpson~      I don\'t know if you read this blog anymore, but if you do well you know who you are, and I hope you know.      Bonnie Rae '),(7985,3490316,'02,August,2004','2004-08-02 00:00:00','   TURN SIGNALS WERE PUT IN YOUR CAR FOR A REASON PEOPLE, USE THEM PLEASE!   I can not stress this enough people! Ever since I have moved to Florida I have noticed that down here I guess people are just all psychic or something. We should all just know what direction you are going to turn when you slam on your breaks with no damn warning only to come to a complete stop forcing me to come to a complete stop only to turn onto a street in which you had plenty of time to slow down to the proper speed and use your signal so I had some sense of warning. I just don\'t get it, people down here just don\'t think. ARG, anyhoo just a pet peeve of mine!      Bonnie Rae '),(7986,3490316,'02,August,2004','2004-08-02 00:00:00','   So today was not that bad of a day, or should I say yesterday seeing as even though I am still at work it is a new day. Anyhoo - the day was not that bad. Work was interesting when I first got here, but has it\'s dry moments as does any job. Later on in the evening a co-worker and friend of mine and I worked up a funny little \"fake\" news cast, mocking incidents that have happened here at work. It was funny as heck and we got a damn good laugh at it! It\'s amazing the crazy stuff one can come up with when you have to much time on your hands~      I have about 20 minutes left and I can feel it. I am tired and not looking forward to the drive home. I am hoping that when Jeff and I find a house, it is closer to where I work. Even if it just knocks off 30mins of the drive. We shall see, now that we are going to start looking for a house to rent we are going to be really picky. Not too picky, but we don\'t want some dump, ya know.    I haven\'t been able to work on my novel lately, do to all the dang drama and stress that I have had to deal with from Jeff\'s family. I am hoping that since I am going to be staying at my house for the next 2 days I can buckle down with my cd player and knock out some chapters.     And on the note of \"drama and stress\", I will be having lunch with Jeff\'s cousin Sarah on Thursday. We used to get along very well until Jeff moved out of her house and back into his dad\'s, then we just seemed to be testy towards one another. But we need to get it (whatever it is) out on the table and see about working through it, because we both love Jeff and he is important to both of us. I am his girlfriend (and the way he has hinted and talked the last 2 months, possibly fiance \"maybe\" around Christmas time) and she is his family - in order to get along I really think we need to get past all these differences we both feel is there. Who knows, I just hope we can come to a common ground so there are no awkward feelings when we are at family events. At least I am trying, I made the call - asked her to lunch and she accepted, which shows me she at least wants to attempt at fixing things.    Ok I am tired and need to put my cases in the book because if I forget again I think Larry will knock me upside the head with it.    ~Bonnie Rae~ '),(7987,3490316,'02,August,2004','2004-08-02 00:00:00','   On one of the poetry forums I belong to we have a quotes section, one of the members posted this one and it just hit me slap in the face. It is so true, so very true..     \"To make the right choices in life, you have to get in touch with your soul. To do this, you need to experience solitude, which most people are afraid of, because in the silence you hear the truth and know the solutions\" Deepak Chopra    ~Bonnie Rae~ '),(7988,3490316,'01,August,2004','2004-08-01 00:00:00','   Well today is another day. I woke up and realized that I had a ton of shit I have to do! Most of which I would have taken care of yesterday, but then we had to get the cat put to sleep and alas my life was again put on hold to service that of another. So now I have a ton of shit that I could have gotten done yesterday that HAS to get done today. Now I know it sounds selfish for me to say \"my life was put on hold\", cause the poor guy had to get his cat put down. OK but for most of you that know the situation. know that I give alot more in this relationship that I am in then the other party. And lately that has really been eating at me. Most of you would say, \"well then change it,\" but it is not that easy.      I just got done reading a friend on mines blog and he had a really good point! Men are hard to change, they are very much set in their ways and getting them to change is like literally pulling their teeth. Boy when I read that I was saying \"Amen\" to myself! As much as Jeff changes with everyday, he seems to fall back to the same cycle - and allows me to pick up all the damn pieces for him. I usually find myself swallowing my anger or pain and not telling him when something is bothering me. Now this is not healthy. I should rely on him as much as he does me, but I just feel that compared to his problems mine pretty much amount to nothing. I do everything for him and he does so little for me. But that is partially my fault, because I have allowed him to act this way. So in reality most of my grief is my fault. Alas the endless circle of bullshit my life seems to be filled with!      I really dont want to do any of this crap I have to do today. I would rather go to the movies like I had planned with my friends, but with picking up a new schedule this week ~ I would really just like to get as much of it done as possible. SO later on I can relax (if that is at all possible) and get ready for the next couple of weeks.      I have come to the conclusion that I need to get a second job though, or find a  reliable  person for a roomate. Because relying on Jeff to get his shit together is just getting old. One minute it is one thing and then the next minute it is completely divergent. I love him and would rather live with him, but in reality I need to get back on my own and fast. So needless to say I will be stopping by my old work today to see if I can talk with my old manager. Not sure if there is a way I can work getting a second job, because my full time job is complicated when it comes to hours, but I have to try. I can\'t just keep living under other peoples roofs, its not who I am and its killing me!      Alrighty I am off to start the day of hecticism! More later I am sure!      ~Bonnie Rae~ '),(7989,2428393,'27,April,2004','2004-04-27 00:00:00','   everytime i have a crush, i\'m always hoping for something that might not come true. i always end up daydreaming about him, i can\'t concentrate on the things i\'m doing. sometimes, i convience myself that the guy i like might like somebody else, but i still like him.  maybe the word  martyr  is next to my name. i\'m always waiting for the right guy but i always falling with the wrong guy. a friend e-mailed me about what those the selection their mean, the question # 8 asked to choose a # between 1-10..i choose 8..and it says thare that\'s the number i will fall in love before i get married..that\'s so weird. ok, i admit it..i easily fall in love with a guy that is really sweet,nice, good-looking and accepts me for who i am.. gtg..i\'ve got some work to do..  '),(7990,2428393,'24,April,2004','2004-04-24 00:00:00','   after the work i\'m doing everyday..i still find the chance to sitback and relax to a hot and not that comfortable room..but slow music in the background. in that hot and not comfortable room, i realize lots of things. like being alone in a valentine\'s day in my 4-teenage years.  yup, i always say \"i hate valentines day, it\'s just for the couples who love each other.\" ok, it \'s a big lie the night i realized it..we should celebrate the valentines day that we really love and care for like our families and true friends. the next time that valentines day will come, i\'ll invite my whole barkada to have fun.  every entry i publish here would be my feelings and realizations in my hot and not that comfortable room.. :) i hope it can help you to realize things too..  '),(7991,2428393,'24,April,2004','2004-04-24 00:00:00','   i\'ll be writing what i really feel here in this blog..even iff it\'s mushy or very corny..you don\'t care \'coz it\'s my feeling..  '),(7992,2428393,'28,May,2004','2004-05-28 00:00:00','   one of my friends told me that their is a big change to her attitude. i think she changed after what happened tonight. and not only her but all of them. and i\'ll be honest, i\'ve changed..a lot! honestly, i\'m afraid to tell them what i really feel about our friendship. i\'m always hidding what i feel, one of my friend told me that  i might comit suicide  if i don\'t tell anyone the problems i\'m keeping to myself. i\'ll not comit suicide, and i don\'t want to hurt myself.  anyways, i\'ll change for the better.  '),(7993,3678942,'18,June,2004','2004-06-18 00:00:00','    \\/\\/ell, I\'m out of my first year in high school.  I didn\'t take any hard classes as a freshman though. here\'s my grades from the classes I had Health-       A Photography-  B- Physical Ed.- A+ Science-      B English 9-    A- Algebra 1-    A See, they weren\'t that hard, and I frankly did little homework.  I still worked hard though, before I didn\'t do ANYTHING and I was a D student, but now I feel like I did much better.  Next years is going to be harder too, and I\'ll have less time because I\'m joining more curricular activities. I allready know my schedule (well not the order my classes are going to be in, but I know what classes I\'m taken)  Us History Biology Geometry physical Ed 11 French 1 English 10  Then I formed a sport for extra curricular activity so I have a sport for every season. At the beginning of school I\'ll take Cross Country when Cross Country is over then winter will be here and I\'ll be taking Nordic Ski-Cross Country then when that is over I\'ll be taking Track (like I took this year)  Well other than that my freshman year is over and in a couple of months I\'ll be starting school as a sophmore (yikes!!!).  My first week for the summer I jogged like 10 miles a day and visited my grandma at my dads house and did a few piano recital\'s. Which reminds me I should go practice piano. (I\'ll do that right when I\'m done typing in this blog) Then my second week for this summer I wanted to help daddy, so I removed all these sticks from the backyard which took ALL DAY!  I dragged tons of heavy logs that were twice my size and you bet I developed some arm muscle from that!  Well I also developed something else from the logs.  POISON OAK (T_______T) it\'s pretty bad too.  It\'s all over my body, yes even on my chest! The worst part is it\'s on my chest and my right side of my face. On my right side of my face it\'s swollen up and red and bubbly, and oh it\'s horrible.  I\'m too scared too go outside and do my four mile jog (since I\'m at my dads house I have not much places to jog too) but since I get beeps from jogging without poison oak on my face, I don\'t even want to know peoples reactions when they see a girl jogging which looks like a big red bubbly faced mutant.  They probably do something worse than honking I\'ll tell you that. (T____T) so I\'m not going out in public or jogging. I might as well just do my power work outs like-1000 crunchies, and 100 leg lifts.  That\'ll work (T____T) but it\'s not the same.     Most the time I\'m either playing DDR Max 2 or DDR Max 1, working on heavy mode, to pass the time. That even annoys my grandma and aunt, so I\'m usually hesitant to play it, but HEY! I can\'t stay on the computer all day, can I?      I should also shower, I\'m all icky (as you need to know, I\'m so sure).  I might as well aisle do a couple of my fanfics here, that\'ll work ^^! (gosh) I just realized, I wonder what Don (my piano teacher) is going to think when he sees me this all poison oaky and stuff. Today I got yelled at by my dad and aunt because I\'m not eating right.  I\'m only eating at lunch, but when I eat lunch I totally binge.  Like yesterday I had 2 nectarine, 2 peaches, 3 handful of grapes, 1 sandwich (with cheese/meat/mustard/jelly), 2 bowls of oats (it\'s actually good, I think I developed a taste for it). Yeah, I totally binge. That\'s like four meals in one, which leads me full for the rest of the day. I\'m not really into the snacking on little meals every two hours, which my mom does because she joined this thing called \'Curves\' which tells you to do that to help maintain your metabolism and yadda yadda yadda to help you lose weight. I already know what I\'m going to do today I\'m going to A-take a shower B-Practice Piano for a while C-Play DDR MAX and DDR MAX2  D-Do Crunchies and leg lifts E-Brush My teeth and hair *of course* F-Have Dinner (I\'ll just not eat anything till dinner) G-Go on the internet and do my live journal H-Go To Bed  Interesting, right? '),(7994,3678942,'19,June,2004','2004-06-19 00:00:00','    Well nothing new to say, I still got bad poison oak but the swelling went down. In fact, the only swelling happens to be part of my cheek and my ear.  Other than that it\'s all good.  OH! I had a heroic dream last night. It was about I\'ll just start at the beginning.    It was at my moms house and I guess she was hosting a party or something because a lot of people were there.  I mean, a lot!  I\'ll list the people. A fat lady--she was black skin and she was very nice to talk too.  She was enjoying the punch at the party.  The Phipps--all of the Phipps it seems from Dallon, john, Ammon, Laura, (oh Barbara wasn\'t there, probably still off in college), Aaron, Sarah, Don, Jason, (Jessica wasn\'t there either, perhaps she was at her house which seems miles away), Mrs. Phipps, and Mr. Phipps were there included---they were all enjoying their time too (odd everyone seems to be enjoying their time at my MOMS house).  There was also Sabrina and her parents their (mom and dad).  Jessalyn and her dad (I wonder why she\'s at my moms house, I actually wonder why she\'s even at my house).  Tabitha was there but her dad wasn\'t so I guess she came with Sabby same with Tiffany.  There was this banker too (which I\'m guessing is my moms new boyfriend).  Plus a lot of kids ( I mean A LOT)tons of little kids, I guess this was a party for the little kids, because there was a lot, I guess a baby-sitting-party.  Weird.   Well anyway the little kids I suppose had parents too but they weren\'t there I guess it was a babysitting party, and well everyone was having fun talking.  The fat black skin woman was the funniest one I remember in my dream, she actually cracked out some jokes. Then we heard a loud bang and we were all outside in the backyard. I was with Dallon (a three year old) and talking with him while walking.  He was talking about a pokemon adventure or something like he always does.   Well we all heard the loud bang so I let Dallon play with a flower while I go and see what the heck it was, everyone stopped talking and in a few seconds Sabrina comes running full speed in the house screaming, \"TIGERS ARE IN THE FRONT YARD\" and we instantly hear a \"RAWRRRRR\" which made everyone scream and run in the front yard.  I ran back to get Dallon and picked him up (oof he\'s heavy) and ran straight indoors too.  We all looked in our front window to the front yard and saw our front yard was packed with tigers and they were chewing on the banker... (not a pretty site).  Everyone was basically screaming or the little kids were crying.   That\'s when I realized our screams and crying was a bit too loud, in fact it was attracting the tigers they were looking at the window and one actually pawed at it! So I realized something, and locked the door/windows/everything!!! Everyone was in a panic.  Sarah looked around too and yelled out at everyone, \"You all need to be quiet, the Tigers hear us, nobody make a sound.\" which we all did.   Then I grabbed the phone and dialed 9-1-1 and exclaimed to them that there was tigers out front of my yard and they started laughing so I gave the phone to everyone in the room, and in the end believe me, they believed me and asked us if we could get out of the house in any safe way possible? And I went, \"No-there\'s no safe place\" then they say, \"we\'ll be right there with help.\" and I\'m all, \"I hope to god so!\" (sorry for saying Gods name in vain in front of you Laura and Mrs.  Phipps..And why am I apologizing .. I didn\'t apologize in the dream >   Well so I put the phone down and say, \"they\'re coming, meanwhile we all have to be quiet like Sarah said.\" so we all were then we heard a car come up. We looked out the front window and saw it was moms car (oh yeah, I didn\'t mention she wasn\'t at the party).  She opened the car and walked out as if not seeing the cars. that\'s when I opened window and screamed at mom to get back in the car, which she instantly did...but this attracted a tiger and it swiped at me.. I closed the window just in time.   Well mom drove the car in the garage and then she ran too the back door (this is brave of her, I\'m amazed) and unlocked the back door with her key and came in and said, \"what the f**** is going on\" (Mrs. phipps and the rest of the phipps didn\'t seem to mind this because they were all pretty scared.) I went, \"I don\'t know mom\" but ding-a-ling-ling I had an Idea.  I then told her, \"Mom I need you to do me a favor.\" and I started whispering in her ear.  When I was done whispering in her ear, mom nodded (this is like the best we ever got along).   Well, I then stood on our kitchen table and said in a calm voice, \"everybody listen up!  When I dialed 9-1-1, they said to leave the house in any safe way possible. Which we are going to do.  First I want all the little kids to go with mom first.  Now everyone bring the little kids foreward.  I want each parent to memorize three kids.  (their was only ten kids total).  Well all the ten kids stepped foreward and there I grabbed five kids and went, \"okay none of you can talk when you leave the back door.\" and one little girl went, \"your going to feed us to the tigers?\" and I went, \"nonono.. I\'m saving you from the tigers, unless you want to be tiger food you have to be very quiet.  When you reach a car I want you all to get into the car, I don\'t care how much little room there is either. None of you can complain. (these kids were all around the age of five or six so they understood me) I also want you kids to try not to cry. Your all going to safety.\" then mom and the kids left the back door and went to the garage and mom opened the door and all the kids climbed in with one in the front. Mom hit the pedal and they were out of the garage.   \"Okay now we all need to wait till mom comes back. WE all should be very quiet\" I said and then I sat down and sighed. Then the black lady (i don\'t know what her problem was) opened up her cell phone and started whispering in it, and everyone over heard, \"hey nancy... What up? well... hear me on this, I\'m gonna die today.  I\'m gonna die today just Like I foresighted. There\'s Tigers out in the front, yeah that right Tigers. This is so exciting I\'m gonna die\" I was thinking in my head \'oh no your not\'. but you know I bet everyone felt like they were gonna die. Well we waited a while then we saw the car come up the driveway from the front window. That\'s where I grabbed the rest of the five kids and then I looked over at the rest of the crowd. \"Laura, Sarah\" they both looked up and I beckoned them to come step over to me.   When they did I told them that they were going with the kids and that they had to be careful and watch over them and that my mom isn\'t entering this building she\'s staying in the car. When they both nodded in understandment they lead the kids out the back door and I guess they were safely in the car because the car left the building.   We waited again for a while and this time the car came back where I then beckoned Aaron to come foreward, Sabby, Tabby, Tiff, and Jessalyn. They all went through the back door. and they made it safely through.  Well I guess we were waiting but while mom was driving with aaron, Sabby, Tabby, Tiff, and Jessalyn.  A cop came and pulled mom to the side of the road for speeding and driving with too many people in the vehicle. That\'s where they all explained Tigers were taking over and they were leaving the building. The cop didn\'t believe and told him to show him if it was true and so mom had to drive back with everyone and when the cop saw this his exclamation was, \"holy mother of God\"  Mom drove away and the cop guy pulled up in the garage and took out a gun where he entered the back door where I jumped up and went, \"9-1-1 people?\" and he went, \"no that aint me.\"  then I blinked and went, \"who you?\" and he went, \"just a cop ready to help you people\" and I nodded and told him that he could help by also being a transportation. He also nodded and here I gathered Mr. Phipps, Mrs. Phipps, Jessalyns dad, Jason, Don, and Jasons girlfriend.  They all packed themselves into the cops car and left, but before they left Mrs. Phipps grabbed ahold of me and said, \"I can\'t go, I didn\'t see Dallon leave with the kids and there was eleven kids total at the party\" I blinked and went, \"GO I\'ll make sure we don\'t leave without him\" and Mrs. Phipps believed me and left.   Where mom came by and I ecsorted sabbys mom and dad out and helped them in the car because they are a bit (on the heavy side when it comes to weight, no offense).  Then I went back inside where the black lady was lauging her head off. I looked at her and went, \"what\'s so funny?\" she then say, \"look they scratching at your front door!\" I looked over and sure enough the Tigers were banging at the front door. I kind of gulped. and went, \"that\'s not funny\".  She laughed and say it was. OH there was something I forgot to add. well a lot of things (hey dreams are hard to remember). Back when the kids were loading the tigers were banging on the door, so I made the black lady sit down and guard the door for she weighed like 400 pounds and she was pleased to do it too, so don\'t say I put the black people first cuz I don\'t! I love black people... even if they don\'t love me.     What I also didn\'t mention was that the Tigers every time everyone went to the back door, I would distract them by going to the window and hollering at them, this was very scary to do.    Well the tigers were very well enough scratching at the front door, and they were getting through. Thank God the police man came and then mom, I told the black lady to go with mom. The police man I told him to stay. Well the black lady went with mom and then the police man asked him why I told him to stay.      That\'s when I told him he had a gun and could help me search the house for Dallon, well we searched and couldn\'t find him, that\'s when I looked from my moms room screen door and saw Dallon right there being bitten by a baby Tiger. There I grabbed the police mans sleeve and we both went out moms back screen door from her room where I kicked the baby tiger and grabbed screaming Dallon.  This however made the baby tiger scream. I just said this to the cop, \"run\" which we did. We ran straight for the police car.   Where I guess the tigers scratched through because right when I got to the car door I saw a tiger crash into the back door from the inside of the house. I was hypervhenilating and Dallon was crying his head off.  When I put Dallon to the side I pulled off some of my pants cotton and put it around Dallons arm (where the baby tiger took a bite).  The police man got the car out and we were both heading away when we saw some cop cars coming up, \"oh finally they come.\" I say so I tell the cop to go up to another cop car. So he does and we roll the windows down and I tell them that I exited all the people out and they just need to get rid of the tigers where they tell us back that they have people from the zoo coming.   Then while we were just about to leave the block I tell the cop guy to turn around and he ask why? and I go, \"we forgot about the neighbors\" the cop understands at once and drives back to the neightbor where I run over to the neighbor and tell them to call everyone at the block (for they know everyones number) which they did and told them to evacute their houses. Where I gave them directions to go to (the Phipps house, it\'s close by but far away from the Tigers)\" So next thing you hear is people sneaking out to their cars with their children slamming the doors and leaving, with tigers rawring and people from the zoo trying to capture them.      Then I grabbed the cops sleeve and went, \"let\'s go\" Dallon was still in the car.  Well the cop was driving to the phipps house when I told him we shouldn\'t go there, we should go straight to the hospital. Which we did.   When we got to the hospital I stayed with Dallon while they were giving him immediate aid, the cut was deep, and I also spoke to the cop. went something like this.  \"thank you so much, you don\'t know how much you helped\" (me) \"you helped some too you know\" (cop) \"na\'ah, but if it wasn\'t for you everyone here would have to wait longer and seeing that door bust down.. we made it just in time, tahnks to you. You should be rewarded or something\" (me) \"if it wasnt\' for you everyone would be dead, your neighbors would be still in their house, and dallon would be worse than harmed from that baby tiger\" (cop) \"wow\" (me) \"yes, wow indeed\" (cop) then I started talking about his job and how was it and paper work and then I asked about his age \"how old are you? you like you are in your twenties\" (me) \"I\'m twenty straight up\" (cop) \"arn\'t cops supposed to be in their thirties and fourties and fat and not good looking?\" (me) \"now your being steorotype\" (cop) -=laughter=- \"sorry\" (me) \"no problem\" (cop is laughing too) So then it\'s like midnight and Dallon has a broken arm and thirteen stitches. So I pick Dallon up and carry him to the cop car where he drives us to the Phipps house. Later on all the tigers have left and everyone went back to their house. The cop looked at me and said this, \"next time it\'ll be bears\" and he kissed me on the forehead and left that was my dream.... Well today will be a normal day as usual. '),(7995,3729846,'23,June,2004','2004-06-23 00:00:00','   I am creating an E-Magazine called \"The Skoop\". I am waiting for more people to write articles for me.  I am not looking for one subject in particular, I want everything and anything from Football (Soccer), Music, Art, Websites, What you have been up to, Fasion anything goes.  If you are interested then feel free to get in contact with me.  Peace   '),(7996,3729846,'23,June,2004','2004-06-23 00:00:00','   I have been busy these last couple of weeks making and editing and designing websites!  First of all, I am trying to put the finishing touches on my Beastie Boys website, that will take a couple of weeks, so i\'ll keep you updated.  I was also asked if I could create a mini-site for Drama at my college to help promote this play that we are doing called \"Work, Leisure, Life and Death\". I have created a template and finished most of the pages, but I have got to now put the information on the pages now which will take about 2 hours.  I also have another site in the pipeline promoting myself. I decided that I wanted to create a site for myself so that I can put up a couple of promo songs for my album that I am creating.  I have other websites that I would like to make in the future, so keep posted for further info.  Peace   '),(7997,3729846,'23,June,2004','2004-06-23 00:00:00','   Welcome to my blogger y\'all! Check out the latest happenings and thoughts of the King ;-)   '),(7998,3729846,'18,August,2004','2004-08-18 00:00:00','   I haven\'t posted on here for AGES, but thats all about to change. Ok this is whats been happning since the last time I posted. I closed \'The Skoop\' E-Zine after just 1 issue, the 1st issue was amazing, really good and interesting, but the reason I closed it down was because there was very little interest in it. I MAY start it up again, but we\'ll see what happens. If I do decide to bring it back, then i\'ll let y\'all know on my website.  Tomorrow (18th August) i will HOPEFULLY have the \'Music\' page on my website open, Ive produced 3 tracks and will get them up on the site for you to enjoy.  Im looking for people to join up to the message board, it\'s takin a little time for the word to spread, so the message board is a little bear right now, but im sure that will all change very soon. I will be adding LOADS of more topics to it, and im also gonna be starting competitions so look forward to that. - If you are interested in being a Moderator for the site then please get in contact with me.  Peace (update tomorrow y\'all)   '),(7999,4210703,'22,August,2004','2004-08-22 00:00:00','   I don\'t know how they do it.  Writers are notorious for being drinkers.  To me, there\'s nothing more difficult than writing while drunk or hungover.  I\'m experiencing the latter right about now.  I\'m feeling like a machine that\'s out of order, and I\'ll be writing about the latest dysfunction in my life.  Weird.  My friends have had to endure listening to my endless bitching about how hard it\'s been for me to meet guys for the past eight months.  Yesterday I met up with a guy, and it was a disaster, much like every experience with guys this year.  He was sweet, very cute, and I enjoyed listening to him talk.  The problem was that I couldn\'t be fully engaged in the conversation.  I just didn\'t know what to say.  I was totally zoning out, and my mind drifted to random crap that had nothing to do with our conversation.  All I wanted to do was lay down and rest.  I wish it were perfectly normal to just say to someone, \"Hey, I know we just met, and I know I\'m the one who suggested hanging out, but is there anyway we could save conversation for later and just lay down and stare at the clouds?\"   Therein lies the problem I guess.  I\'m too stressed out and focused right now to be dating.  Plus I\'m continuously working through issues I\'ve had with my ex-boyfriend.  But the process of working on my goals and working through past-relationship shit has never felt lonelier nor more difficult.  I suppose once I get through it  all I\'ll be a stronger person, but when you\'re thick in the middle of the process, the end seems nowhere in sight.    When I\'m having a hard time practicing for the LSAT, I wonder, Gee, if I had someone to care about and vice versa, maybe I\'d be having an easier time focusing instead of feeling like this ghost of a person.  It bothers me that I\'ve assigned the most meaning to my life in terms of how well I do on a stupid fucking test .  I have a lot around people around me that I\'m grateful for, and little by little I\'m recognizing things about myself that make me appreciate life and give me meaning, but for some reason, I feel it necesssary to share it all with someone.  Sometimes I think what I have is too much for me to handle alone.  I\'ve been stuck on thinking what the next sentence should be for quite some time and tie things together because right now, everything is rather disjointed.  The only thought that occurred to me to let words keep flowing is that maybe, just maybe, I\'m so worried about the possibility of failing that if it does occur, I\'ll feel so broken and low that I won\'t be able to depend on myself to cope with it.  I\'d have my friends, but it\'s not the same as having someone you love there by your side.  God, I wonder if people who grew up in healthy families exeperience this kind of bullshiit.  All of this sounds like trite, cliched crap.  But you know, it\'s the kind of stuff that helps uncrowd my mind.    I think I can study now...  '),(8000,4210703,'20,August,2004','2004-08-20 00:00:00','   When I was in the Creative Writing program at SFSU, I had to do a ton of writing exercises each week.  It got the point where I felt like I was out of ideas completely.  Or in other instances, I\'d stare at my computer screen, and whatever ideas I\'d have in my head earlier in the day no longer seemed fresh and interesting when I tried to flesh them out.    That\'s how I feel now.  I\'ve been trying to think of something good to say in my next post, but feared it would only come off as, \"whoa is me, my life is a bitch.\"  I\'d have damn good reason to think that though.  I might explain that in a bit.    But this thought was particularly striking to me today.  I was pretty desperate to clasp on to a dramatic idea/fantasy in my head to literally force upon myself a fresh perspective of life.  It sort of worked.  While fighting off a very strong wave of post-lunch food coma at work, I got it in my head to finally call my doctor about my x-ray results.  Earlier in the week I went to the doctor because I\'ve been getting short of breath and have been having trouble breathing in general.  For someone at 26 years-old, that\'s pretty unusual.  So I got it checked out.    The receptionist who answered my call kind of weirded me out.  After she looked up my information, her warm disposition shifted to clinical efficiency when she asked,  \"what number do you prefer to be reached at for a private message?\"  I held my already shallow breath out of surprise.  What was this about, I thought.  I gave her my number slowly, hoping she\'d detect the concern she raised in me, and then she abruptly told me the doctor will get back to me and hung up.   When I put the phone down, I thought, God, what if this is something serious?  I wasn\'t prepared for this.  And then a sick idea popped in my head: So what?  So what if it serious?  I got to thinking that if my days were numbered, maybe I\'d start appreciating life a bit more.  And I was!  I was sitting in my cubicle thinking about how short life is, and in my case, how it may be even shorter.  The papers around me, the keyboard I was touching, the office walls and other cubicles all looked beautiful to me.  Every thing I felt bad about in my life seemed so trivial.    I don\'t know how the suddenly appreciative perspective wore off, but it did, and it disappeared rather instantly.  It was like I walked through a fog and into light, and when I turned back to look at the fog, it wasn\'t there anymore.  I could see everything as being just as shitty as I remembered it being always.  The doctor never called me back.  I still have no idea what answer to expect about my x-ray results.  But I\'ll tell ya- I\'m not too worried about it.  Even if I get bad news, I know that things might not change about how I see life.  Why should it have to?    It\'s a sad thing to cling on to ideas of hope over and over again just to let it bring you down.  I realize that the stuff I\'m saying is all a reflection of stuff going on in my life.  If things were going better, I\'d have a better attitude.  All the time, there will be people who say having a better attitude makes thing go better in life.  Those people are either heavily medicated, criminally stupid, or have so much sex that they have no time to actually engage and be part of the workings of society.    Yeah, rather than chronicle the mundane details of all that ails me, I enjoyed bitching about the larger picture.  '),(8001,4210703,'16,August,2004','2004-08-16 00:00:00','   I called my oldest sister to wish her a happy birthday.  This was a big deal.  I don\'t keep in touch with my family that much.  I\'m not mad at them.  We\'re on speaking terms.  It\'s just that they\'re all far older than me, so they all feel like cousins than actual siblings.  But I suppose from their perspective, I\'ll always be their little brother.  For me, I\'m just not big into family, and I keep my distance.  Why bother getting closer now?  Plus the family just keeps getting bigger every year.  There\'s just no way I can keep up with everyone.    Well, I sort of did some catching up with sis and with her husband.  I didn\'t quite tell them, \"Oh, by the way, I\'m a big homo.\"  It\'s not like they don\'t already know.  What\'s my personal life got to do with them?    A lot, apparently.    My brother-in-law told me about how my sister keeps telling everybody how much of a striking similarity there is between her oldest son and me.  She\'s definitely qualified to make an accurate assesment because in a way, she\'s the one who went to my parent-teacher conferences, informational meetings, and other chaperoned events since my mom didn\'t (and still doesn\'t) speak English and my dad was (and is still) always working.  She spent a lot of time with me in a parental kind of way.  But just the basics of course.  She had her own life to live too while leaving more motherly duties for my actual mom to take of.    Despite all she\'s done, I\'ve always sort of been jealous of my sister.  She\'s smart, pretty, and married well.  She\'s got a good life, and now so do her three kids.  Although, her oldest son, who I think is about 12 years old (see how much I keep up with my family!), I\'m not sure will have it so easy.  While I was still in NY, I sort of got to watch him grow up.  He was a little brat, and he was difficult to deal with because he\'s also very bright.  Smart kids are the worst to be around.  He didn\'t endear himself to me.  As he got older, I became aware that I didn\'t really like my nephew because I noticed the many similarities we shared.  He was a bit of a sissy.  He liked pink, didn\'t like playing sports, was overly emotional, always wanted to be around Mom and always had to be the center of attention.  When I visited home recently after not seeing him for over a year, the only thing that changed about him was his height.    The kid has a bright future.  He\'s got parents who can be more involved in his life than mine ever were.  His family is more well-off, and I hate to say it, but the little brat may end up being smarter and better-looking than me.  But he\'s going to have endure so much torture in school for being who he is, and he may lose so much of what makes him stand out as an exceptional person.  Sure, people of all walks of life had to endure junior high and high school trauma.  But there\'s something absolutely brutal about growing up in a Republican environment right outside of NYC, like Long Island, where he is.  The bullies have something more tangible to compare you with.  They see first-hand all the \"freaks\" their parents keep them sheltered from whenever they make a trip into the city.  Worse yet, they\'ve got the smarts, braun, and money to try to make your life a living hell if you\'re not like them.  My nephew is in for a horrible, depressing journey to adulthood.  I think about the distance I keep from my family, and how I don\'t share my personal life with them.  Sometimes I think that if I can help spare my nephew  just one ounce of grief that I had to endure while growing up, maybe he\'ll have less pain than what I\'ve had to live with.  Maybe he\'ll never have to feel washed-up at a young age.  No matter what I do, it won\'t be easy for him.    One way I want to help him out is to follow my dream.  I always wanted to be successful when I grew up: be a famous actor, have a few books published, or just be rich and happy.  Silly measures of success to some people, I know, but that\'s what I wanted.  Each time I think about how much shit I had to put up with just to be happy with my life, I realize how much I survived and how strong I have been.  And now I deal with the arduous task of trying to get into a top law school and becoming a successful lawyer who will also one day write fiction books too, and it can all be too overwhelming for me to handle- almost to the point where I want to quit.  I guess I should keep working towards my goal so that my nephew has someone to look up to.  Maybe my success will get him through the days of his self-discovery and give him hope that one day, everything will be all right.  '),(8002,4210703,'14,August,2004','2004-08-14 00:00:00','   Today I saw a lesbian couple get off the bus.  The one with the shaved head walked beside her lover with her head in the air like she was entranced by angels singing to them about their love.  She smiled.  Her face was peaceful.  If someone took a snapshot of her, cropped out everything and everyone around her and just presented her image out of context, they\'d think she took a shot of morphine.  She looked like an idiot.  It made me think, \"Good fucking, god.  Did I ever look that stupid when I had a boyfriend?\"  And instantly I remembered, Yes.  I have seen pictures in which I have a ridiculous ear-to-ear grin on my face while standing next to my special someone.  That smiling, happy creature seems like a stranger to me now.  I\'ve officially been single for too long (or maybe finally, long enough to have clarity) because I\'ve come to the conclusion that couples are truly, deeply odd.  I\'ve forgotten how people become couples, why they become couples, what makes them stay couples- it\'s a foreign country to me again.  This one-time serial monogamist is now completely baffled by the concept of relationships.  Brings me to my next point: Logic.  For anyone who has taken or will take the LSAT, they will discover it is a very difficult test with no prior education in formal logic.  It is especially difficult if you have lived your life by your own  means of logic, and not the standardized methods used by others use to rationalize opinions, beliefs and arguments.  I came to the conclusion that for the 26 years of my life, I\'ve been living as an insane person, simply going around saying (in paraphrase), \"This is so, because I\'ve decreed it so.\"  I always humored myself with that line of thinking, but it\'s not so funny when you\'re trying to do well on a test that measures how well you apply logic principles.    Where exactly is this going?  My decision to become a lawyer is either sucking the creative life-blood out of me that I regretfully neglected, or it\'s helping me to become a well-adjusted adult that sets goals and actually takes steps to accomplish them.  I\'m starting to think being creative and being well-adjusted are mutually exclusive.  Before this idea of becoming lawyer-John, I had love, I read tons of books, I wrote, and I had fun even with having to endure extreme poverty.  Logical thinking is not part of who I am.  It is eating my soul alive and turning me into someone I don\'t know anymore.  I have a different way of thinking that sometimes makes me feel like someone else\'s brain has been transplanted into my skull.  And the strangest thing of all- despite this drastic transformation I\'m undergoing, I can almost say that I\'m happy to be going through it.  I just don\'t know how the journey will end.  What if I don\'t become a writer nor a laywer?  What if keep going through life just being a peon?    There\'s no hidden message here of save yourself, follow your dream and all that bullshit.  Sometimes things \"are \"just because that\'s the way they \"are\".  It\'s not inherently good, or bad.  It\'s what you make of it.  Right now, life for me now is like a giant pile of manuer.  Someone recently told me that in some regions all over the world, manuer is mixed with dirt and clay to build beautiful feats of architecture.    Reminds me of the wisest thing I\'ve read in my life.  I went to the bathroom at this Indian resturant.  Right above the toilet paper roll, someone scribbled: \"Nothing is True Everything is Permissible.\"  Just like that.  No punctuation.  I tried to make sense of it all in my head, and then I was like, \"Holy Shit!  That\'s just like on the LSAT.\"  Assume whatever you want to be true.  Knock yourself out.  It doesn\'t really matter unless you want it to.  Relationships, LSAT questions, writing- these things are all fighting for title of what matters most.  In my convoluted way, I laid out some stuff I think about.  I bet it was a real bitch to make sense of all that l said.  Now you know how I feel.  '),(8003,4210703,'12,August,2004','2004-08-12 00:00:00','   So I put the hummus away, stopped surfing online, and am trying to quiet the thoughts of how beautiful people make me sick. It\'s not that I\'m bitter or jealous of the exceptionally genetically gifted. I just don\'t like them- sort of how I don\'t like pate. It\'s just gross to me, and I start to gag.  They look so happy. When I see happy, perky people in general, I want to put a pin to their head and let the air out. People probably say the same thing about me when I\'m out and having a really good time with my friends, but I don\'t care really. To exercise my right to judge, I must be prepared for judgments upon me as well. (Though if I hear what someone says and it doesn\'t please me, a serious ass-whooping would be involved.)  I\'m imagining there\'s a lot of happy people today. The California Supreme Court nullified all 4,000+ same sex marriages that took place this year in my fair city. That really bothers me. Gay marriage made so many people in this country unhappy. And now its remnants are being dismantled. All we\'ll have is a memory that gay people once had their love legally recognized by the city of San Fancisco, but no formal, substantitive documentation to prove it.  It just looks bad. A huge civil rights movement was just a flash in the pan. And now what? I\'ll have to endure ignorant assholes on the radio, in the news or in the elevator talking about how a wrong has been made right. I swear, for the next few days, when I hear people discuss gay marriage, I\'ll feel like I\'m eating nothing but pate for breakfast, lunch and dinner.  And then the govenor of New Jersey resigned today because he had an extramarital affair with a man. I wonder if it occurred to the majority of people that maybe he wouldn\'t have been married to a woman in the first place if people were more accepting of homosexuality. What a mess bigotry creates.  Anyway, I\'m done with my ranting for the day. Time for a nap before the tranny rock show tonight.  '),(8004,4210703,'12,August,2004','2004-08-12 00:00:00','   I should be getting to bed, but each time I stare back at my big empty thing, I keep thinking that there is so much I haven\'t done in the day, and that there is so much yet to do. And so if I were limited to one metaphor for my life, it would be an empty bed at night- a bed that should have my body\'s indentations, but doesn\'t. It doesn\'t even call for me. It sits there, like me- flat and demanding, but without specifically asking for attention, almost denying that it\'s asking for attention when it\'s really begging for it.  It sounds so dreary, the way I put things, but that\'s just how I am. I figure that if I live my pain loudly enough, one day it will do something. Maybe I\'ll reach someone who can make it stop. Maybe prettier words will come together and become something useful, like a story or book. To some extent I\'ve put those thoughts behind me. But as I get older, and as I make decisions that lead me further away from my creative ambitions, I feel some need to reconnect with language and put these words online. It\'s like there\'s the spirit of a writer inside me who\'s slowly dying, and these are his last words- sort of like Ted Hughes\' ode to Sylvia Plath. Hence the blog.  As I write, I feel disappointed with what\'s coming out. I wanted something cooler, more moving, more beautifully said, like what comes out during the rare moments I sit down at my computer or at a notepad and try to tell a story. The shocking thing about getting older for me is how useless the struggle for perfection really is. It\'s a losing battle I wish I never took up in the first place. But then I find myself joining in the fight at the most inopportune times no matter how much I try to let go of attaining perfection. Point is: I\'ll say lots of things on this here thingy. They can\'t all be epiphanies, tragedies, comedies, enlightening political insights, etc. It\'ll just be my life- my highly imperfect, flawed life that I don\'t enjoy as often as I should.  '),(8005,3739942,'27,June,2004','2004-06-27 00:00:00','    this weekend was hilarious  I can\'t believe i made it out alive. theres so much to tell and i dont have enough energy to type so i\'ll just type the main points...-- we ended up smelling like fish poop.. and we had sand in our panties.. YEAH! that is priceless.. how did we know? cuz it itched. haha. read my xanga for more gts\'== www.xanga.com/asalways01 '),(8006,3739942,'24,June,2004','2004-06-24 00:00:00','    Tonight*  I was selfishly ditched for a pack of FM girls. I seriously thought that I had figured out who my true friends are, then I get ditched for someone like her. That\'s basically insane and it pisses me off a lot. I told Ashley about it and It\'s not like I can miss the game so I guess I\'m going by myself. ::Gasp:: I never go anywhere by myself, so it\'ll be a learning experience. I got a letter from Eric today. :) The thought of him just makes me smile and giggle. I\'ve been starting to wonder why people are starting crap about us, but then I think i\'ve figured it out. Jealousy. I know Tiffany Engle wanted to jump his bones last week @ volleyball so that\'s probably where some of the stuff is coming from.   FridayÂ¤  I have to pack for a long essential weekend with the girls. There\'s a party @ Samanthas then its off to the sleepover routine. I need to take a nap friday because heaven knows I\'m always the first one out. Maybe not if I get to talk on the phone with Eric til odd hours of the morning. ::giggle:: So much fun.  Saturday MorningÂ¤  We leave for Celina for a weekend full of 3 girls, a cabin, and probably no electricity. I will cradle my cell phone. I love to camp, don\'t get me wrong... but boredom kills. We\'ll be around Celina Saturday night so I was trying to plan something out w/ Angela to go to the bowling alley there or something. Eric said it was pretty cool... and eric knows everything. haha. I was just thinking... Eric hates FM.. and who ditched me today? FM kids.. And ditching Amy.. just IS NOT cool no matter what is going on. So I\'ll be gone all weekend w/ the girls. I will be home Sunday night probably. Then I have to live through another week till i get to see my baby. later on kids.. Â¤Amy '),(8007,1778579,'18,August,2003','2003-08-18 00:00:00','    without the jukebox. The jukebox that Elaine said would \"...win the prize for the worst selection of songs.\" Oh, but me made do. There really is not much that can top eating pizza with friends and listening to \"www.memory\" by Alan Jackson. Let\'s see, there were nine of us there: Ben, Elaine, Chalupa, Rupp, TLoft, Ben, Lydia, Tim, and I. The pizza was good too. We got three larges, and the guy confused the cheese one with the pepperoni one, so at first there was a mad rush for the \"pepperoni\" one, forcing those who wanted only cheese to resort to scraping meat from the other pizzas.  Even the drive was fun. The place is about 25 minutes away, so we rocked out to 80\'s songs in the car, yelling the words at the tops of our lungs and playing highway games with the cars around us.  Overall, it was a great reunion of friends, and if we were all vets from \'Nam we probably would have talked about war memories, but we\'re not, so we didn\'t.   '),(8008,1778579,'15,August,2003','2003-08-15 00:00:00','    \"Won\'t be long before the kids return.\" I hear it all the time. Taylor University will re-commence classes in just over two weeks. The \"kids\" talked about here are the college students. And I\'m one of them. Interestingly enough, I find myself saying the same thing. \"The kids will be back soon.\" As a 22 year old senior who took a year off before coming to school, I guess I am now one of the oldest ones here, but still. When will we grow up? Is there a point in life when we will stop being kids, or stop looking on those a bit younger than us as kids?  Probably not. Parents say to their youngest, \"You\'ll always be our baby.\" And I guess that\'s the way life goes. To someone, somewhere, I\'m young by comparison. You are too. I think that\'s a good thing. Take it to heart, we all need to be something of a kid sometimes.  A speck of life is lost when someone grows up too much to have a sense of humor. Will you be a kid today?   '),(8009,1778579,'14,August,2003','2003-08-14 00:00:00','    because I talked to my Dad on the phone. It\'s his birthday, you know. He\'s 56 now. And that\'s not all - he lives in Cameroon. Hold on though, it gets even better. I called a cell phone that he has been carrying around for while now but he has never made or received a phone call on it up until now. He told me it was a great first call to receive.  I sent him an email last night (with the time change he got it on the correct day), jokingly asking him how it feels to be 39.  He sent back an email saying I was off by a year - 39 is NEXT year. He brought it up on the phone and I jokingly acted embarrassed and said I hoped I didn\'t make him feel too old. We had a good private little laugh.  I talked to my dad on the phone today, so now I can smile.   '),(8010,1778579,'13,August,2003','2003-08-13 00:00:00','    What is the job of a computer security expert? Is it not to keep the computers on their network SECURE from viruses and worms by applying security patches? So what is really going on in a company that gets totally shut down by the latest computer worm? You would think that having three weeks of advance notice before a new worm comes out would be enough time to apply the fix that came with the notice. But apparently it\'s not. The coffee must just be too appealing, the conversations too involving to actually get work done at work.  No wonder people are having to work longer hours more days per week. Alas, such is life.  The funny part is that the Blaster Worm is going to attack the Windows Update site on Saturday...   '),(8011,1778579,'12,August,2003','2003-08-12 00:00:00','    A friend of mine got me into this, and though I am not half the writer she is, I might as well give it a shot. Besides, what have I got to lose?   '),(8012,1778579,'19,December,2003','2003-12-19 00:00:00','    Another semester is over. I have now been at Taylor for 3.5 years. Where does the time go? All I have left is j-term and second semester, and then...  Well, that\'s the problem. I have no idea what I\'ll be doing or even where I\'ll be come June. Kinda scary, but kind of exciting to know that it\'ll be a fresh start.  Tomorrow I head down to Indy, and then from there Andrew and I will be heading out to Massachusetts to spend the holidays with Abby and Dan. Should be a good time. It\'ll be a good change of scenery, that much is certain. But just like time does, it will pass and I\'ll be back here once more.  That\'ll be good too though. I\'m beginning to realize the importance of relationships, and I will try not to take the rest of the time I have here for granted. Oh, if only I had figured that out sooner! Anyways, I have friends I need to get in touch with and no one ever reads this anyhow, so I\'ll stop.   '),(8013,1778579,'27,February,2004','2004-02-27 00:00:00','    And this time I think I might try to keep it that way for a bit...  It was actually a pretty tricky problem, due to the fact the IE has a terrible rendering engine. It just goes to show that Microsoft thinks it\'s too good to be standards compliant. As a matter of fact, you should try using a different browser. In my opinion,  urlLink Mozilla Firefox  is the best browser out there, and it is much more standards compliant than IE is. Not only that, but it has built-in popup blocking, tabbed browsing, and oh so much more. So get it. Now.  Enough of one rant, time for another. Thanks to our IT department here at Taylor, I had to reinstall everything (including windows) on my computer last night. They thought it would be good to do a forcible install of their corporate anti-virus software (Symantec) on all computers joined to the domain, over the network. Well, I was running Norton Anti-Virus at the time, and Symantec makes it clear on their website that the two products can have devastating results if run together. Such was the position IT put me in. I couldn\'t even boot Windows into Safe Mode because my computer was so screwed up. Well, now I\'ve disabled all Administrative shares that they took advantage of to accomplish this, so I should be safe now.  Woohoo for IT doing stupid things.   '),(8014,1778579,'25,February,2004','2004-02-25 00:00:00','    My prayer is that this movie would be a huge witness for Jesus Christ, and that non-Christians and Christians alike would realize what Jesus endured for us. I believe that the movie has the capability of making a huge difference in our world, and I pray that all of Satan\'s attempts to lessen the impact of the movie will fail.  I\'m going to see it at 10:20 tonight, along with some 600 other Taylor students. I\'m excited, anxious, nervous - all at the same time. So in the mean time, seeing as I can\'t get it out of my head anyhow, I figured I\'d mention it here. You should go, and take a friend along.   '),(8015,1778579,'24,February,2004','2004-02-24 00:00:00','    Here I am sitting in my room, and I should be doing homework. But, I\'m not. There\'s only so much I can do at once. Right now I\'ve been working on my re-vamped  urlLink website , finally updating this blog, talking to friends on IM, and thinking about life. And there\'s still so much I need to do. I need to find a job for when I graduate in May, for instance. I could be working on projects for classes. I should probably be sleeping. Oh well. I guess with my plate so full all I can do is keep on eating.   '),(8016,1778579,'28,March,2004','2004-03-28 00:00:00','    Well, I am halfway through my last semester in college. We\'ve had seven weeks, we have seven more. Well, eight if count finals, but they don\'t really count. Time just keeps on moving. Spring break is over now and classes start tomorrow. Let\'s see, did I get done what I wanted to over break? Hmmm. Project for class? Check. Work on my website? Check. Find a job for after graduation? Still looking....  Have I mentioned how hard it is to find a job in the Dallas area when I\'m up in Indiana? Yeah, not having much luck. Hopefully something will come up in say, oh, the next eight weeks or so! Other than the need to find a job, things are going well. I\'ve been more in contact recently with friends from Cameroon / Massachusetts than I have been in a long time.  So things are good. Now I just need a girlfriend...   '),(8017,1778579,'26,March,2004','2004-03-26 00:00:00','    After spending the first half of the week down in Indy, I came back up to Upland on Wednesday. After putting of my project even more, I finally sat down and worked on it today. Started before lunch and worked pretty much until supper time. The good thing is, the project is pretty much done! I still need to come up with some thorough tests for it, but it\'s basically there.  Not only that, I put my resume in at a place in Indy in case something doesn\'t come up right away, and hopefully that will give me something to fall back on. I haven\'t heard from them yet, but I have a couple of friends from Taylor who work there so chances are a lot higher there than other places. So we\'ll see...  Tonight I spent the evening with a friend watching the Xavier/Texas March Madness game. It\'s amazing to me how I can get into sports so much when I don\'t even really care about the teams that are playing. Xavier being the underdog, we chose to cheer for them, and it was a close game the whole way. By the end I was standing up five feet from the TV with white knuckles. Xavier played hard and ended up with the win.  One other thing - Happy Birthday Karyn! Hope it was a special day for you!   '),(8018,1778579,'21,March,2004','2004-03-21 00:00:00','    and it feels really good. Right now I\'m at Andrew\'s place, and all he has is a dial-up connection. Not only that, but he has a laptop and he takes it with him to work. I brought my computer with me, but I don\'t have a modem. Doh! So during the day, there really isn\'t all that much I CAN be doing.  Rather than feel like I\'m wasting a full week sitting around disconnected from the world, I\'ve decided to only stay down here in Indy through Tuesday. On Wednesday I\'ll head back up to Upland and stay with a friend off campus for the rest of the week. He has a cable modem, so I can hook into that and once again get some stuff done.  To Do (once I\'m back in Upland): * Finish my Java project for Language Structures * Work on my website * Keep looking for a job in the Dallas area   '),(8019,1778579,'20,March,2004','2004-03-20 00:00:00','    Well, it\'s spring break. I really should be more excited than I am right now. But that\'s ok. It\'s 1:30 in the morning, and I got a lot of packing done tonight. So many people asked me today if I was moving out or something because they saw me packing up almost all of my stuff.  So...here\'s the explanation for all who want to know:  I\'m graduating in May. One of my best friends is coming to my graduation, and leaving here with me to drive down to Indy. Seeing as I have quite a bit of stuff, I\'m taking as much of it as I can down to my brother\'s house over break, so that I will have only what I need for the rest of the year left up here. That way at the end of the year I will be able to fit Adam, me, my stuff, AND his stuff all in my little car. Otherwise I would have to leave more stuff behind.  So no, I\'m not moving out, I\'m just thinking ahead.  The end.   '),(8020,1778579,'16,March,2004','2004-03-16 00:00:00','    Let\'s see - it\'s Tuesday night. Spring break is next week, and I have oh so much to do before the end of the week. A project for Operating Systems consisting of adding a function to the kernel. My partner and I have a good start, but we\'re stuck at the moment so it\'ll have to wait until after we can ask our questions tomorrow in class. In Language Structures I have to create an interpreter for a language called \"Jay\". A lot of code is given to us to get us going, but I still haven\'t started the part I have to do. Doh. Both projects are due Friday. Oh yeah, I also have a test on Friday.  Not to mention things apart from school. Seeing as I\'ll be done here in May and I have to take all my stuff from this place to the next is my little car, I also need to pack up a bunch of stuff and take it down to my brother\'s place over spring break. That way, all I\'ll have left here are things I need for the rest of the year, and space won\'t be as much of a problem.  Tomorrow I will hopefully get my car fixed. We found out it has a leaking brake cylinder, and I don\'t really feel like driving down to Indy before getting it fixed. I\'m getting help from a friend, and I don\'t think he\'ll charge me too much for the labor. I already bought the part - it was only $12. Hopefully that will fix the problem.  So what am I doing? Talking on IM and procrastinating! Obviously.   '),(8021,1778579,'15,March,2004','2004-03-15 00:00:00','    So, due to my total lack of imagination I took a name suggestion from a friend.  Henceforth,  my snake shall be known as Hunter.  Let it be so.   '),(8022,1778579,'10,March,2004','2004-03-10 00:00:00','    Prayer and Praise with IWU (another local Christian college) had just gotten out and I was heading back to my room. As I was approaching my door, I heard a phone ringing, and it seemed to be coming from my room. Odd, I thought. My phone doesn\'t normally ring. But, it\'s Sunday, so maybe Abby is calling.  Much to my surprise, it was a friend from back in Massachusetts. Now granted, we send emails back and forth, and talk on Instant Messenger now and again, but a phone call was a wonderful surprise. Karyn is one of the few people that I actually keep in touch with from my time in MA, one of the few that actually took the effort to make sure I was still alive over the past few years even though I\'m horrible at returning the favor.  So here we were on the phone, and it was just neat. We caught up a bit more on life, remembered old times, and just had an overall good conversation.  Thank you God for all your blessings on me!   '),(8023,1778579,'10,March,2004','2004-03-10 00:00:00','    because I got a snake! It\'s a Columbian Red Tail Boa, and it\'s awesome. It\'s only 2-3 months old, and I\'ve had it 10 days now. It\'s still really small - only about 2 feet long. I\'m not sure if it\'s a male or a female, but at this point that doesn\'t really matter. I can find out later.  It hasn\'t eaten yet, but that should come with time. Someone on a forum suggested leaving it alone for a couple of weeks to let it get acclimated to its new home. Won\'t be easy to do, but I guess I\'ll try. I think I don\'t quite have it warm enough either, so I\'m gonna try to get a heat pad as soon as I can. I\'ve been told not to worry though. Time is all it needs.  I still need to come up with a name though, so let me know if you have any brilliant ideas!   '),(8024,1778579,'29,April,2004','2004-04-29 00:00:00','    (When this began) I had nothing to say And I get lost in the nothingness inside of me (I was confused) And I let it all out to find That Iâ€™m not the only person with these things in mind (Inside of me) But all the vacancy the words revealed Is the only real thing that Iâ€™ve got left to feel (Nothing to lose) Just stuck; hollow and alone And the fault is my own, and the fault is my own  [Chorus] I wanna heal, I wanna feel what I thought was never real I wanna let go of the pain Iâ€™ve held so long (Erase all the pain till itâ€™s gone) I wanna heal, I wanna feel like Iâ€™m close to something real I wanna find something Iâ€™ve wanted all along Somewhere I belong  And Iâ€™ve got nothing to say I canâ€™t believe I didnâ€™t fall right down on my face (I was confused) Looking everywhere only to find That itâ€™s not the way I had imagined it all in my mind (So what am I) What do I have but negativity â€™Cause I canâ€™t justify the way, everyone is looking at me (Nothing to lose) Nothing to gain; hollow and alone And the fault is my own, and the fault is my own  [Repeat Chorus]  I will never know myself until I do this on my own And I will never feel anything else, until my wounds are healed I will never be anything till I break away from me I will break away, I\'ll find myself today  [Repeat Chorus]  I wanna heal, I wanna feel like Iâ€™m somewhere I belong I wanna heal, I wanna feel like Iâ€™m somewhere I belong Somewhere I belong   '),(8025,1778579,'28,April,2004','2004-04-28 00:00:00','    Today in chapel I heard a really good message about returning to my first love, Jesus Christ. It was a call to refocus on what is really important - my relationship with Jesus Christ. If the relationship is there, everything else will follow. The speaker spoke from Revelation 2, where John addresses the Church in Ephesus. The Church is commended for living upright and holy lives, rejecting evil. They are condemned, however, for forsaking their first love.  This was used to show that striving to be holy should not be our main focus. We need to focus on our relationship with Jesus. Keep doing what you\'re doing, but make sure the priorities are correct.  And on another note...   Hebrews 3:12-13   \"See to it, brothers, that none of you has a sinful, unbelieving heart that turns away from the living God. But encourage one another daily, as long as it is called Today, so that none of you may be hardened by sin\'s deceitfulness.\"   '),(8026,1778579,'28,April,2004','2004-04-28 00:00:00','    Why can\'t I breathe whenever I think about you? Why can\'t I speak whenever I talk about you?  Such a great notion, the idea of falling in love...  Too bad in the song it involves cheating!   '),(8027,1778579,'27,April,2004','2004-04-27 00:00:00','    Delight yourself in the LORD &nbsp;&nbsp;&nbsp;and he will give you the desires of &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;your heart.  Commit your way to the LORD; &nbsp;&nbsp;&nbsp;trust in him and he will do this:  He will make your righteousness &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;shine like the dawn, &nbsp;&nbsp;&nbsp;the justice of your cause like the &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;noonday sun.  Be still before the LORD and wait &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;patiently for him;   '),(8028,1778579,'26,April,2004','2004-04-26 00:00:00','    I\'m more in contact with friends than I have been in a long time. Things are going really well, and yet I have such a poor outlook on life. I\'m constantly thinking that none of my plans will work out in the end. I have this fear that I\'ll end up a single and alone, living a pointless life of going to work and coming home to nothing.  I have so many blessings in my life. I\'ve gone to a great Christian school for the past four years, and I\'m leaving with a good education. I have an awesome family, and friends that really care about me. I have Jesus Christ. I had an incredible experience growing up in Africa. I talk to friends on the phone for hours and it\'s awesome.  Even things for next year a looking good. I don\'t have a job lined up, but I\'m hoping to be near my best friend - and we haven\'t lived near each other for the past five years. What more could I really want?  Why do I feel the way I do?   '),(8029,1778579,'25,April,2004','2004-04-25 00:00:00','    But it\'s not, and it never will be. I try to look on things with a good attitude, but it\'s not always easy. I\'m here at Taylor almost done, and I\'m totally mixed about leaving. I can\'t believe that my life as I have known it (non-stop education) is coming to an end - and I\'m super excited about that. But it means I have to say bye to people too. And that\'s never easy.  I feel that as I leave this place I should have some sort of direction for my life. But I\'m just floating right now. I\'m about to turn 23, I\'m single, I don\'t have a job or a definite place to live. My friends live in either direction across the country and talking on the phone every now and again just isn\'t enough to make it feel as though they\'re here with me.  It\'s a time of change and transition and decision making, and every choice I make closes off a hundred other possibilities. Which is the path for me?  The Road Not Taken ~Robert Frost  Two roads diverged in a yellow wood, And sorry I could not travel both And be one traveler, long I stood And looked down one as far as I could To where it bent in the undergrowth;   Then took the other, as just as fair, And having perhaps the better claim, Because it was grassy and wanted wear; Though as for that the passing there Had worn them really about the same,   And both that morning equally lay In leaves no step had trodden black. Oh, I kept the first for another day! Yet knowing how way leads on to way, I doubted if I should ever come back.   I shall be telling this with a sigh Somewhere ages and ages hence: Two roads diverged in a wood, and Iâ€” I took the one less traveled by, And that has made all the difference.   '),(8030,1778579,'20,April,2004','2004-04-20 00:00:00','    So here I am, sitting down in the dungeon (the CSS dept. here at Taylor), waiting for my 2:00 lab to start. Right now I\'m listening to Linkin Park, Reanimation. It\'s a neat album.  My ankle: It\'s getting better slowly. The color is fading now, and I\'m walking around fine. It\'s still tender and it hurts when I stretch it, but there is definite improvement. It was a pretty bad sprain - it turned purple from my foot to a third of the way up my leg. A word of advice...try to not land on a basketball next time you\'re out shooting around...  School: Things are starting to wind down. I have a total of four more projects and three chem labs. One of the projects is coming along well, and we\'ll keep banging away on it tonight. One of my labs is today, so come the end of the week, I\'ll have three projects and two labs left. Not to mention only four more weeks of school. Wow.  Job: I don\'t have a job lined up, but I\'m not too anxious about it. My parents will be in the states for about a month this summer, and mid-July we are heading down to Dallas for a wedding. The more that I think about it, I think it would be good for me to spend some time with my parents. So I\'m considering hanging out with them, going down to the wedding, and staying in Dallas after the wedding. Seems like an ideal time for me to move. If I don\'t have a job lined up at that point, at least I\'ll be in the area trying to find a job instead of from a distance.  Chip: Five weeks from graduation, and he just started dating a freshman here. We\'ll see how that goes...   '),(8031,1778579,'17,April,2004','2004-04-17 00:00:00','    \"Runaway Train\" - Soul Asylum  Call you up in the middle of the night Like a firefly without a light You were there like a blowtorch burning I was a key that could use a little turning  So tired that I couldn\'t even sleep So many secrets I couldn\'t keep I promised myself I wouldn\'t weep One more promise I couldn\'t keep  It seems no one can help me now, I\'m in too deep; there\'s no way out This time I have really led myself astray  Runaway train, never going back Wrong way on a one-way track Seems like I should be getting somewhere Somehow I\'m neither here nor there  Can you help me remember how to smile? Make it somehow all seem worthwhile How on earth did I get so jaded? Life\'s mystery seems so faded  I can go where no one else can go I know what no one else knows Here I am just a-drownin\' in the rain With a ticket for a runaway train  And everything seems cut and dried, Day and night, earth and sky, Somehow I just don\'t believe it  Runaway train, never going back Wrong way on a one-way track Seems like I should be getting somewhere Somehow I\'m neither here nor there  Bought a ticket for a runaway train Like a madman laughing at the rain A little out of touch, a little insane It\'s just easier than dealing with the pain  Runaway train, never going back Wrong way on a one-way track Seems like I should be getting somewhere Somehow I\'m neither here nor there  Runaway train, never coming back Runaway train, tearing up the track Runaway train, burning in my veins I run away but it always seems the same   '),(8032,1778579,'13,April,2004','2004-04-13 00:00:00','    Spring Break is over. Easter Break is over. We are now in the final stretch. No more breaks. Just five weeks of classes, four days of finals, and then graduation.  Time is flying.  I\'m ready to go. Rather than look at this last little stretch as something I don\'t want to do, I\'m trying my best to take advantage of the time I have left here. So many people that I may never see again.  \"First Day Of The Rest Of Our Lives\" - MxPx  Not a lot right now makes sense to me And it wonâ€™t go quietly Not a lot right now makes sense to me And it wonâ€™t sit patiently  Iâ€™m gonna chase my dreams and catch up to them Iâ€™m gonna find you some how, some way, somewhere, some day  First day of the rest of our lives, I miss you already Last time I saw that look in your eyes, I miss you already First day of the rest of our lives, I miss you already Last time I saw that look in your eyes, I miss you already  I donâ€™t want to spend the rest of my life alone Whereâ€™s life taking me? I donâ€™t want to spend the rest of my life alone Always traveling  Iâ€™m gonna chase my dreams and catch up to them Iâ€™m gonna find you some how, some way, somewhere, some day  First day of the rest of our lives, I miss you already Last time I saw that look in your eyes, I miss you already First day of the rest of our lives, I miss you already Last time I saw that look in your eyes, I miss you already  I wanna love my job I wanna love my life But most of all, I wanna fall in love  First day of the rest of our lives, I miss you already Last time I saw that look in your eyes, I miss you already First day of the rest of our lives, I miss you already Last time I saw that look in your eyes, I miss you already   '),(8033,1778579,'08,April,2004','2004-04-08 00:00:00','    So after having complained about my  left  foot yesterday, this is what happens today. A friend and I went to the gym to shoot around and the ball got wedged between the rim and the backboard. After running, jumping, and hitting the ball a number of times it was still wedged securely in its place. A couple other guys were on the other end of the court also shooting around, and seeing our dilemma, they rolled their ball toward us so we could use it to knock our ball back down. Problem is, I was still in the jumping process...  One of the guys yelled \"Ball!\" to warn me, but it was too late. I came down on the ball with my right foot, and as the ball rolled under my weight, so did my ankle. Needless to say, I lay on the floor for a few minutes in pain before braving the standing up process.  All said and done, I\'ve rolled my ankle worse in the past and it got better then, so I\'m pretty sure it\'ll get better this time too. Funny thing is, my  left  foot hurts because I\'ve been playing sports regularly, so now that my  right  ankle is trashed I won\'t be playing sports, so my  left  foot should get better...  It\'s an interesting world.  And now I\'m off to bowling, where I hope my ankle doesn\'t interfere too much...   '),(8034,1778579,'07,April,2004','2004-04-07 00:00:00','    Ok, so this is what happened. Three years ago I was riding my bike to work and this idiot in a Jeep Grand Cherokee backed out from a little corner convenience store without looking both directions. To avoid being hit, I turned my handlebars all the way to the right to veer around him. Due to attempting such a sharp turn, I was about to fall over, so I put my left foot on the ground to stop myself.  Well, I didn\'t fall over, and the guy didn\'t hit me full out, but his front left wheel (engine end of the vehicle, mind you) thumped over my left foot. Being stupid like I am, I put off going to the doctor even though my foot was killing me. I limped around for a month or so before it felt somewhat better, but even then it still hurt most of the time. Finally around two months after the incident, I went to get it checked out.  At that point it was too late to tell from the x-ray what had happened, so the doctor just supposed that my foot had been pretty much flattened against the road, stretching things that aren\'t meant to be stretched and possibly breaking things. Well, he gave me an arch brace and told me to wear it for a while. I\'m not sure that it really helped at all.  Well now, almost three years later, it\'s really been hurting again. I find myself limping as I walk around campus. I can\'t pinpoint the pain, it\'s just somewhere  inside  my foot. Dang it. Why didn\'t I just go to the doctor when it happened??? So really, it\'s not my foot that is stupid, but me.  Well, hopefully it will just go away and get better. Somehow I doubt it.   '),(8035,1778579,'06,April,2004','2004-04-06 00:00:00','    I\'m really looking forward to getting out of this place. As much as I\'ll miss some of the people here, I can\'t wait to be done. A fresh start somewhere new with new people will be good. I have no clue what I\'ll be doing, but at least it won\'t be what I\'m doing now. Every day now is just like the last, and it\'s really starting to drain me.  Classes - well, I hate school. I always have. High school, college, whatever. Sure it\'s nice being around people my age and hanging out, but I despise doing work for classes. And I don\'t really hang out anyway. I pretty much just sit around in my room. I have six projects left between two classes, and four labs left in chemistry. I have absolutely no motivation to do them, but I will, just like always.  Things will be fine, I just needed to share some of my apatheticism with you all. Maybe tomorrow will be better, but this is how I feel right now. I don\'t care and I just want to be done.   '),(8036,1778579,'01,April,2004','2004-04-01 00:00:00','    I walked out of the dorm the other day and who should walk by but the president of Taylor University, Dr. Gyertson. He gave me a big smile and asked how I was. I responded with the typical, \"Things are going well, how about you?\" At this point he said the most interesting thing - \"Spring is sprung, the grass is ris.\" By the time he had finished saying it, we were already past eachother, already moving back into our own busy worlds. But somehow, what he said has stuck with me. It makes me smile and almost chuckle each time I think about it.  And oh how thankful I am that spring has finally arrived. The grass is green once again and birds have resumed their songs. The weather is warming, if still too slowly for my likes, but it is warming none-the less.  On another note, I got to play volleyball yesterday, and get to play again later today! I\'ve worked with the women\'s volleyball team here since I was a freshman, doing line judging and scoreboard / substitutions. During the spring, the girls call out six or so guys to scrimmage with them, give them a challenge. I happen to be one of the few to get called, and it is always a blast. Volleyball is most definitely my favorite sport, as long as it is competitive.  And now I\'m off to other things.   '),(8037,1778579,'29,May,2004','2004-05-29 00:00:00','    I\'m sitting at Nathan\'s place, writing one last post before I head out. For the next week I will be fairly disconnected - no internet connection.  From here I am heading down to Indy to hang with my brother, and then I am heading out to MA on Friday. I\'m totally excited about getting out there, but I\'m not looking forward to the drive and I still don\'t have a job...  So my life is in God\'s hands. I may not know, but he does:  Jeremiah 29:11 \"For I know the plans I have for you,\" declares the LORD, \"plans to prosper you and not to harm you, plans to give you hope and a future.\"   '),(8038,1778579,'28,May,2004','2004-05-28 00:00:00','    When the rain comes it seems that everyone has gone away When the night falls you wonder if you shouldn\'t find someplace To run and hide Escape the pain But hiding\'s such a lonely thing to do  I can\'t stop the rain From falling down on you again I can\'t stop the rain But I will hold you \'til it goes away  When the rain comes you blame it on the things that you have done When the storm fades you know that rain must fall on everyone So Rest awhile It\'ll be alright No one loves you like I do  I can\'t stop the rain From falling down on you again I can\'t stop the rain But I will hold you  I can\'t stop the rain From falling down on you again I can\'t stop the rain But I will hold you \'til it goes away  When the rain comes I will hold you   '),(8039,1778579,'26,May,2004','2004-05-26 00:00:00','    Nathan just left for PA. He won\'t be back till Monday, but I\'ll be gone from here tomorrow, so I won\'t see him again. He\'s been a great friend the past few years.  And just for his sake, a quote from Dumb and Dumber....  \"I hate goodbyes!!!\"  Well, maybe I\'ll see him again before too long. Never know.   '),(8040,1778579,'26,May,2004','2004-05-26 00:00:00','    Well, I\'m sitting here with absolutely nothing clever, funny, or even interesting.  One thing I can say is that Linkin Park\'s Meteora album is as dismal as it ever was. If you want some real good low down \'I hate my life\' kind of music, it\'s a good choice. I think they must have had some serious issues when they wrote that album.   '),(8041,1778579,'25,May,2004','2004-05-25 00:00:00','    I was just looking at quotes of the day online, and this is the first one I found:  \"Blessed is he who expects nothing, for he shall never be disappointed.\"  - Alexander Pope  It\'s interesting because I just posted about expectations a few minutes ago, and I was talking with my sister about disappointment just the other night. I was telling her that I just take life as it comes to me, and I try not to let things get me down too much. I said that maybe it\'s because I have somewhat of a pessimistic view of life, so when things don\'t work out I get up and move on. I prepare myself for the worst, and then when it happens it doesn\'t hurt as badly. When I said that, she told me that had nothing to do with pessimism, but that I look on life realistically. Sure it\'s good to hope for the best, but in reality sometimes things don\'t work out - and that\'s ok.  It\'s almost like being a Christian: Pray as if it all depends on God (because it does), but work as if it all depends on you.   '),(8042,1778579,'24,May,2004','2004-05-24 00:00:00','    I\'ve been thinking over the past few days, and I\'ve decided that while assumptions lead to problems, expectations do not.  It is wrong to assume something about someone else, especially to the point of making decisions based on that assumption. This comes to mind in relation to emotions, and it makes me think of Proverbs 4:23, \"Above all else, guard your heart, for it is the wellspring of life.\"  On the other hand, if you  know  something, and there are no assumptions overlying the issue, it is altogether proper to have expectations. Expectations are things that you  are able \"to consider probable or certain,\" according to Merriam-Webster Online. When you base things on what you  know  rather than what you  assume , you are approaching the issue correctly.  And that was just something I\'ve been thinking about.   '),(8043,1778579,'24,May,2004','2004-05-24 00:00:00','    I have successfully completed my time in school at Taylor University.  Graduation weekend was a blast. Thank you so much for travelling out here Adam and Karyn, even though plans didn\'t come together until the last minute! And thank you Aron and Amy for driving out here as well.  Graduation day was beautiful. Due to the nice weather, graduation was held outside in the football stadium, so there was plenty of seating. Adam, Karyn, Kelly, Andrew, Aron and Amy, as well as most of the Yoder clan were there to see me walk across the stage. From there we went out to eat at Ivanhoes, and after that Adam, Karyn, Kelly and I drove down to Indy. We met up with Andrew and the five of us had a fun time hanging out for the afternoon and watching a movie at Andrew\'s place that night.  Sunday was full of goodbyes as Adam and Karyn flew separate directions, and Kelly was supposed to but got delayed. She ended up flying out just after lunch today.  So now I am back up at Taylor, where I will be working through Friday. From here I will spend a few days with Andrew down in Indy, and then I\'m off to MA...   '),(8044,1778579,'21,May,2004','2004-05-21 00:00:00','    Well, I\'m moving out now. Time for me to take my computer down. Hopefully I will continue to update this once I have an internet connection again...   '),(8045,1778579,'20,May,2004','2004-05-20 00:00:00','    As plans often do, plans changed. My friends from Massachusetts who have been planning on driving out here for my graduation just decided last night that they\'re not going to after all.  All but one: Karyn is still coming! She decided to fly instead of drive, which in my opinion is a much better idea anyhow. So she\'ll fly into Indy tomorrow around noon, and then she\'ll fly out again on Sunday afternoon. I\'m so amazed that she\'s still coming out here even though the others decided the trip wasn\'t practical (and time off work was not easy to come by, apparently).  It worked out really well because Andrew (who lives in Indy) is driving up here tomorrow afternoon, so he\'s going to pick her up at the airport before he comes up. Then on Sunday, I\'m already making a couple of trips to the airport, and so it works out perfectly.  In other news: I\'M DONE WITH COLLEGE!!!!   '),(8046,1778579,'18,May,2004','2004-05-18 00:00:00','    Adam has arrived safely. He flew into Indy last night around 7:40, I picked him up around 7:55, and after a stop for food we arrived back up at Taylor around 9:35. In the car on the way back here from the airport, we realized that it\'s been 11 months since we saw eachother last, but now that we\'re together it feels like it was just last month. It\'s incredible how good friendships aren\'t lost due to time and distance. Out of sight out of mind is not always the case after all...  It\'s really cool to have him around, and to let him see where I\'ve been living for the past 4 years. It\'s crazy because he was denied leave two times before finally having it approved on the third try. So at first it was looking like he was coming, then it was looking like he wasn\'t. When he finally told me he had gotten leave, I still didn\'t get my hopes too high - I was waiting till he stepped off the plane and was here in Indiana.  Well, now he\'s here. And the rest of my family / friends will get here on Friday. It\'s an exciting time, and yet I\'m still sad about leaving this place.  And the rest of you, I CAN\'T WAIT for you to get here! :-)   '),(8047,1778579,'17,May,2004','2004-05-17 00:00:00','    Well, I had to do one last check, but I appear to be safe. It turns out that if I don\'t show up to any of my finals I will still pass all of my classes. Now that doesn\'t mean that I won\'t show up, or that I won\'t study, but it means I really don\'t have to worry about anything. My graduation is secure.  Tonight Adam flies into Indy. He\'ll be getting in around 7:40, so I\'ll be heading down that direction around 6:30. That should give him enough time to get his baggage and hopefully be waiting outside so I just have to drive up. If he\'s not there yet, I guess I\'ll dish out the cash for parking, but hopefully I can avoid that.  My last computer science work will be done later today. I have my Language Structures final from 1-3. It\'s hard to believe.   '),(8048,1778579,'16,May,2004','2004-05-16 00:00:00','    Time is so short. I\'ve been trying to take advantage of every waking moment, to the point on increasing my waking moments and minimizing my sleeping ones. Going on 3-5 hours of sleep a night starts to drag after a while.  My floor had a pick-a-date on Friday night. We went down to Indianapolis to get Cinnabuns and Orange Julius\'. Well, my brother called about an hour before we left and invited me to go to King\'s Island the next day with him and a few of the music groups from his school. Seeing as I would already be down in Indy, it would work perfectly for me to just stay down there after the pick-a-date and go with Andrew the next day. The pick-a-date ended up being a lot of fun, and we even go somewhat locked in the mall - it took us a long time before we found an open exit leading to our parking garage. After a great time with friends, I sat in the car by myself driving to my brother\'s house.  I felt so alone.  I got to Andrew\'s house (he was out), and sat around in the silence of loneliness, realizing that my time with friends at Taylor is coming to a close. I got something of a taste of how apartment life after college will be. I started feeling really down, so I read my Bible for a bit and then just went to bed.  Saturday was a lot of fun. It rained all day at King\'s Island, but there was no lightning so the rides were still running. After a wet and cold day there, a long trip home on the bus, and a drive back up to Taylor from Andrew\'s house, I got back to my room around 3 AM.  After 5ish hours of sleep, I got up and went to church - for the last time where I\'ve been going for the past 3.5 years. Now I\'m sitting in my room, still realizing that my time here is short. I need to sleep, but I can\'t waste my time. I\'m sad, excited, depressed, and exhausted all at the same time.  I think maybe I\'ll go for a walk to take my mind off of my world.   '),(8049,1778579,'14,May,2004','2004-05-14 00:00:00','    ...are finished. Need I say more?   '),(8050,1778579,'13,May,2004','2004-05-13 00:00:00','    I\'ve been working a on project all day, but it\'s ok. As I was sitting down in the dungeon this afternoon, I realized that this is my last day of homework. So really, it doesn\'t matter that I\'ve been working all day and I will continue to work tonight, because I am almost done.  It\'s interesting because that realization has given me almost a sense of motivation - so I don\'t mind the fact that there\'s still more to be done. Plus, this way I can say I did indeed work until the very end. And it is all for the glory of God.   Philippians 3:13b-14  \"But one thing I do: Forgetting what is behind and straining for what is ahead, I press on toward the goal to win the prize for which God has called me heavenward in Christ Jesus.\"   '),(8051,1778579,'11,May,2004','2004-05-11 00:00:00','    \"But seek first his kingdom and his righteousness, and all these things will be given to you as well.\"  This is such an encouraging verse in times of uncertainty. And how much more uncertain can you be when you\'re leaving college, moving onto something totally new - while not knowing what that next something is. God commands us to not worry about tomorrow, and to incessantly seek him. May this verse provide encouragement to you today!   '),(8052,1778579,'10,May,2004','2004-05-10 00:00:00','    Yesterday was Monday. It was a good day.  It was my last Monday in college. Well, at least for classes. I have a final next Monday, but that\'s not the same.  And the great thing is that I didn\'t even go to class today. We had a CSS luncheon today in honor of us graduating seniors, and so my first two classes were cancelled. I had a class at 2:00, but it was chemistry, so I decided not to go.  So yeah. No classes. A great phone conversation. No real homework.  A good day.   '),(8053,1778579,'10,May,2004','2004-05-10 00:00:00','    I figured it out. I\'ve been thinking this whole time that I don\'t have any connections to help me find a job, but I was wrong. Sure, I may not have too many connections at the moment, but I\'m graduating with a great group of people - and they will become my connections.  Many of them have jobs lined up, and if I\'m still looking for a job six months from now than chances are I can call one of them up and get some help. At that point they will be established enough to put my resume in the hands of their superiors and to say nice things about me. So yeah, it\'s not the end of the world after all!  And that\'s the end of the story.   '),(8054,1778579,'10,May,2004','2004-05-10 00:00:00','    Some Caedmon\'s Call lyrics...  Well this day\'s been crazy But everything\'s happened on schedule From the rain and the cold To the drink that I spilled on my shirt \'Cause You knew how You\'d save me Before I fell dead in the garden And You knew this day Long before You made me out of dirt  And You know the plans that You have for me And You can\'t plan the end and not plan the means And so I suppose I just need some peace Just to get me to sleep.   '),(8055,1778579,'09,May,2004','2004-05-09 00:00:00','    I tried my best, but I failed to cheer up a friend who\'s feeling down. Maybe I should just stay out of other peoples\' lives.  I hope you feel better eventually.   '),(8056,1778579,'08,May,2004','2004-05-08 00:00:00','    \"Everybody becomes boring at one point or another....\"  How dismal is that?   '),(8057,1778579,'08,May,2004','2004-05-08 00:00:00','    After having posted about my gmail account, my hit count jumped up. So I checked it out, and apparently people were doing searches for gmail, trying to figure out how to get an account.  So let me make it clear right now: I cannot refer you. I stumbled on the offer due to my blog, and I have friends who did the same. Maybe the same will happen to you, maybe it won\'t.  But don\'t worry - it will eventually go live and you can then all be happy!   '),(8058,1778579,'08,May,2004','2004-05-08 00:00:00','    It\'s funny how things work out sometimes.  I\'ve been following Gmail for a while now, google\'s new email service. At the moment it\'s in beta mode, and is not publicly available. Well, I found out that another guy on the floor had managed to get an account somehow, so I asked him how it went. He told me that he had been referred by another friend who had gotten one through blogger.com, google-owned blogging service.  Well, if you are reading this than you know that I also have a blog through blogger.com. So I\'m thinkin\', hmm, I didn\'t get any offer like that. I told Telman that if he ever got a chance to refer someone, could he please refer me? He told me that he had already promised that to another friend if the chance presented itself. So I talked to Ben, the guy he had promised to refer, hoping to get somewhere in the string of referrals, with no idea if Gmail would even continue the referring process that long. Well, Ben already had someone else in line, so my hope was pretty much shot.  That was at lunch today. After lunch I came back to my room, went to the blogger homepage, and right there was an offer for me to sign up with Gmail. It said that because I was an active blogger I was being given this opportunity. I couldn\'t believe it. So I signed up, wondering if this was really happening. And now I have a Gmail account, so you can all write to me there!  The address is adam.griffis[at]gmail.com.   '),(8059,1778579,'07,May,2004','2004-05-07 00:00:00','    \"If there were no God, there would be no Atheists.\"   '),(8060,1778579,'07,May,2004','2004-05-07 00:00:00','    I am part of the fellowship of the unashamed. I have Holy Spirit power. The die has been cast. I have stepped over the line. The decision has been made, I am a disciple of His. I will not look back, let up, slow down, back away, or be still.  My past is redeemed, my present makes sense, my future is secure. I am finished and done with low living, sight walking, small planning, smooth knees, colorless dreams, tamed visions, worldly talking, cheap giving, and dwarfed goals.  I no longer need preeminence, prosperity, position, promotions, plaudits, or popularity. I do not have to be right, first, tops, recognized, praised, regarded, or rewarded. I now live by faith, lean on His presence, walk by patience, am uplifted by prayer, and labor by power.  My face is set, my gait is fast, my goal is heaven, my road is narrow, my way rough, my companions few, my guide reliable, my mission clear. I cannot be bought, compromised, detoured, lured away, turned back, deluded, or delayed. I will not flinch in the face of sacrifice, hesitate in the presence of the adversary, ponder at the pool of popularity, or meander in the maze of mediocrity.  I will not give up, shut up, let up, until I have stayed up, stored up, prayed up, paid up, preached up for the cause of Christ. I am a disciple of Jesus. I must go till he comes, give till I drop, preach till all know, and work till He stops me. And when He comes for His own He will have no problems recognizing meâ€¦my colors will be clear.  ~ Author Unknown   '),(8061,1778579,'07,May,2004','2004-05-07 00:00:00','    And I only have one week of classes left.  I calculated it out, and I have 9 hours 45 minutes left in class (minus exam week), and that\'s if I don\'t skip any more...  One of the guys I\'m graduating with applied with Google, and he had his phone interview yesterday. At the end of the interview, the guy talking to him asked if he could make it up to Chicago for a face-to-face interview today at 3 PM. Chris said, \"I think I can make that work,\" or something to that effect, and the guy said, \"That\'s the right answer!\" So hopefully that interview went well for him today. Google would be such a great place to work.  On another note, I FINALLY got an oil change today. It\'s been way too long timewise, even though I was still under 3000 miles. So now I feel much better about my car - and that\'s a good thing.   '),(8062,1778579,'05,May,2004','2004-05-05 00:00:00','    Well, as most of you know, I own a beautiful Columbian Red-Tail Boa. It\'s only a few months old, and after a worrisome 6-7 weeks, it finally ate it\'s first mouse. Since then it\'s still been picky about eating, but it has eaten one other time. So I\'m hoping that it will continue eating and will live to a nice old age.  The thing is, its life will not be with me. I\'ve decided that it will be too much hassle to take it with me down to Indianapolis, out the Massachusetts, back out here, and down to Dallas (if that\'s where indeed I end up). It would be very stressful on the snake, which is especially bad due to its eating habits. So, I\'ve decided to give it to Jonnie. He lives on my floor and had two of them, one of which regrettably died last Friday to what could have been pneumonia. So he\'s down to one, but at the end of the year he\'ll have two again.  So it will be goodbye. I guess if I every really want one once I\'m settled somewhere, I\'ll get another.   '),(8063,1778579,'03,May,2004','2004-05-03 00:00:00','    Well, today I had my last test in my college career. Sure, I still have finals, but that\'s not the same. I also had my last chemistry lab, which was extra credit anyways. So now I really am close to being done. I have a quiz on Friday, a project due the last day of classes (May 14), and finals. That\'s it. So now I\'m getting nervous again because the reality of being done with school and being out in the _real_ world is starting to hit. I\'m not worried about working full time - I\'ve done that before - but I\'m worried about where I\'m going to end up and where I\'ll find a job.  Update on graduation: I have some awesome friends from Massachusetts who are driving all the way out here (it sucks it\'s only for that one day), my best friend is hoping/planning on being here, and my brother and his wife will be here. I\'m excited because my friends from all over the place will get to meet eachother. Who would have thought that three different sections of my life - Cameroon, Massachusetts, and Taylor - would collide in such a wonderful way!  Tonight was a fun night. It started off with a tribunal up on broho, the floor above us. You wouldn\'t understand even if I tried to explain it, so I won\'t try. Then we had some fun here on Sammy II, with one guy trying to eat a gallon of cottage cheese in an hour. So yeah, he didn\'t make it. Good thing we put down plastic and had a trash can next to him! To end the festivities, we watched a video of the Sammy II Titans (intramural football team) win the championship in 2002. So yeah, a good time.   '),(8064,1778579,'03,May,2004','2004-05-03 00:00:00','    I don\'t have caller ID, so I have no idea who calls me unless they leave a message. For instance tonight, someone called, but they didn\'t leave a message. My answering machine blinks AS IF I have a message, but it ends up just being a click as the unknown person hangs up.  So please, if you happen to call, at least let me know who it is so I can call you back!   '),(8065,1778579,'01,May,2004','2004-05-01 00:00:00','    ...and I don\'t really know why.  I guess it\'s a good thing though. Maybe it\'ll come in handy for something.   '),(8066,1778579,'28,June,2004','2004-06-28 00:00:00','    Hope dangles on a string Like slow spinning redemption Winding in and winding out The shine of which has caught my eye  And roped me in So mesmerizing, so hypnotizing I am captivated  [Chorus] I am Vindicated I am selfish I am wrong I am right I swear I\'m right I swear I knew it all along  And I am flawed But I am cleaning up so well I am seeing in me now the things you swore you saw yourself  So clear Like the diamond in your ring Cut to mirror your intentions Oversized and overwhelmed The shine of which has caught my eye And rendered me so isolated, so motivated I am certain now that  [Chorus]  So turn Up the corners of your lips Part them and feel my finger tips Trace the moment, fall forever Defense is paper thin Just one touch and I\'d be in Too deep now to ever swim against the current So let me slip away [3x] So let me slip against the current and let me slip away [4x]  [Chorus]  Slight hope It dangles on a string Like slow spinning redemption...   '),(8067,1778579,'21,June,2004','2004-06-21 00:00:00','    As a Christian, my relationships with friends should stem from my relationship with Christ. I should be focused first on him, with him as my guide in all my other relationships. Then, because I am \"right with God,\" I should be able to show Christ\'s love to others and be wise in my relations with friends.  Why is it that so often I turn that upside down?  Instead of focusing first on Christ, I focus on my other relationships. I keep God in the picture, but so often he\'s somewhere in the background. Then things get all messed up. I have a good time with friends, and things are going well - and I praise God because of it. But when things _aren\'t_ going so well, my relationship with Christ fades as well. It\'s as if my emotions at the time dictate my relationship with God - well, it was a bad day, so God obviously doesn\'t care about me.  How wrong is that??  Oh Lord, help me to put you foremost in my life!   '),(8068,1778579,'20,June,2004','2004-06-20 00:00:00','    So I have this really cool group of friends out here that I hang out with all the time. We have Bible Study on Wednesday nights, and then we hang out during the weekend, and also sometimes weeknights. So it\'s fun.  Well, we decided on Wednesday night that Saturday we would go to the beach. The weather looked a bit intimidating Saturday morning, but it really cleared up that afternoon. A word to the wise (or should it be the \"unwise\" in this situation??) - put on sunblock! Yeah, me being me, I didn\'t. Laying out at the beach for three hours in the middle of the day without sunblock is not the smartest thing I could possibly have done...  But oh well, it doesn\'t really hurt. The only place that really is bothering me is behind my kneecaps. It even hurts to straighten my legs. Doh.  It\'s all good though - I\'ll be in pain for a few days, and then I\'ll go to the beach again!!!   '),(8069,1778579,'12,June,2004','2004-06-12 00:00:00','    My parents arrived safe and sound (or is it safely and soundly...) in the States yesterday afternoon. My sister talked to my mom on the phone, and apparently the trip went well save for one lost piece of luggage. And even that should be here later today, so it\'s all good.  They, along with Aron and Amy, will be coming down to the chapel tomorrow. It\'s weird because I haven\'t seen them in a year, and yet it feels like so much less time than that. It\'s amazing what growing up can do to ya!  My dad has testing on Monday, and hopefully that will show what his next phase of treatment should / will be.  God is good. God is awesome.  And from Jess:  James 1:12 - Blessed is the man who perseveres...   '),(8070,1778579,'11,June,2004','2004-06-11 00:00:00','    I\'m excited because I was walking through Kohl\'s yesterday to buy a white dress shirt and a black tie, and I just happened to walk by the shoe section. I saw a pair I really liked, noticed they were on sale, and low and behold they had my size!  So I have new shoes now. My old ones lasted me just about a year, but the holes in the bottom didn\'t help keep the water out, so it was time to move on...  On a totally not-shoes-related subject, my parents should be in the States now! They should have flown in just after noon today, going up to Aron and Amy\'s place from the airport. On Sunday they will be heading down here, and that will be nice because I haven\'t seen them in a year.  Actually, come to think of it, my parents ARE related to my shoes - they bought my last pair the last time I saw them. It\'s been the same amount of time since I saw them that I got new shoes.  I knew my ramblings weren\'t  totally  random!!!   '),(8071,1778579,'10,June,2004','2004-06-10 00:00:00','    ...all that much to talk about.  Same old, same old. I\'m looking for a job, and my resume has been submitted.  I hung out with friends last night at Wednesday Night Bible Study. The best part in my opinion was when we were sitting around a fire singing. It\'s so easy for me to get caught up in life and to worry about things, and it was good to get my focus aimed back where it should be.  I need to keep coming back to what truly matters - God. Besides, this life is like the blink of an eye, so some of the small things just don\'t matter. Life is just life, and I\'m taking it as it comes to me.  For those of you who are like me, and who tend to get down or worry about life, here are a few verses to encourage you:  Matthew 6:27 Who of you by worrying can add a single hour to his life?  Matthew 6:33-34 But seek first his kingdom and his righteousness, and all these things will be given to you as well. Therefore do not worry about tomorrow, for tomorrow will worry about itself. Each day has enough trouble of its own.  And one of my favorites:  Psalm 37:4 Delight yourself in the LORD and he will give you the desires of your heart.   '),(8072,1778579,'07,June,2004','2004-06-07 00:00:00','    The speaker at my graduation said that \"life is what happens while you are making plans.\" And it\'s so true.  At the moment, there\'s been a rain delay, but the game\'s not cancelled. It\'s still too early to tell, but rain is never good, so I don\'t have much hope that the game will be played after all.  Besides, maybe the game was never supposed to happen anyhow.  Proverbs 4:23 Above all else, guard your heart, for it is the wellspring of life.   '),(8073,1778579,'02,June,2004','2004-06-02 00:00:00','    ...is finally starting to thin out.  Finish college classes and homework -  CHECK  Make it through finals week -  CHECK  College graduation ceremony -  CHECK  Have a blast with friends grad weekend -  CHECK  Hang with Andrew through my bday -  CHECK  June 4, drive out to MA -  CHECK  Get moved in with Dan and Abby -  CHECK  Find a job  As you can see, everything is done except that last thing. Yes, it\'s crucial, but at the same time it\'s all I really have left to get done at the moment. Once I have a job, I won\'t be broke all the time anymore. I can\'t freaking wait.   '),(8074,1778579,'02,June,2004','2004-06-02 00:00:00','    And If I don\'t make it known that I\'ve loved you all along Just like sunny days that We ignore because We\'re all dumb and jaded And I hope to God I figure out What\'s wrong   '),(8075,1778579,'02,June,2004','2004-06-02 00:00:00','    And things are going fairly well.  My car broke down in the middle of PA on the drive out here. I stopped at a gas station, filled up, went inside and paid, and when I came back outside my engine coolant was all over the ground instead of in my radiator. Yay for old cars!  But God was looking out for me. At this point in the trip, exits were 10-15 miles apart, and I just happened to stop at the ONE gas station that had a truck garage in the back. I got my car over there, they checked it out, ordered a new radiator, and 4 hours later I was back on the road. I ended up getting to Dan and Abby\'s place at 11:30 at night, so even with the 4 hour delay it wasn\'t so bad. Good thing I started the trip at 4 in the morning...  So anyhow, I\'m here, and I\'ve already had some time to hang out with friends, and things are going fairly well. I\'ve had some kind of crazy sinus headache or something for about a week now, but I picked up some drugs (advil wasn\'t touching it) and hopefully I\'ll be better in a few days. We\'ll see - it\'s been a real pain so far.  On another note, I\'ve started talking to people here and hopefully the connections I have will lead to a job before toooo long. I was thinking about not starting a job until August, but now I\'m thinking that once I find a job, I\'ll start whenever. Hopefully something will open up soon.  I\'m out. It\'s been a while since my last post, but I still plan on keeping this updated.   '),(8076,1778579,'02,June,2004','2004-06-02 00:00:00','    I\'m 23 years old now. Wow. I\'m gettin\' up there. Soon I won\'t even remember WHEN my birthday is...  So here I am in a little coffee shop spending a bit of time on the internet. I\'m living at Andrew\'s place in Indy right now, and he doesn\'t have an internet connection. I must say that it\'s quite the change going from a dedicated connection at Taylor to having nothing at all. I mean, I hadn\'t even checked my email in a few days...  But here I am now, and I\'ll be heading out to MA on Friday morning, bright and early. I\'m hoping to leave between 3:00 and 3:30 in the morning. That way by the time breakfast comes around I\'m already a good way into the trip.  Well, I need to go - I think someone else may need to use this computer now...  I\'ll update this again once I\'m out in MA!!!!   '),(8077,1778579,'27,July,2004','2004-07-27 00:00:00','    Work is going well so far. I don\'t have my office yet - it\'s full of file cabinets at the moment. And I don\'t have my computer yet - they just ordered it today. But it\'s ok. Right now I\'m sharing an office with another guy who started two months ago, and they hooked me up with a nice and slow PIII laptop so I can still do stuff until my new machine (a desktop) comes in.    Apart from that, life is pretty much the same. I\'m still amazed at how quickly things happened out here, both the bad AND the good.    Yadda, yadda, yadda - it\'s just linkin park lyrics. I like the song.    \"My{Dsmbr\" (feat. Mickey P., Kelli Ali)    This is my December  This is my time of the year  This is my December  This is all so clear    This is my December  This is my snow-covered home  This is my December  This is me alone    And I  Just wish that I didnâ€™t feel  Like there was something I missed  And I  Take back all the things I said  To make you feel like that  And I  Just wish that I didn\'t feel  Like there was something I missed  And I  Take back all the things I said to you    And Iâ€™d give it all away  Just to have somewhere to go to  Give it all away  To have someone to come home to    This is my December  These are my snow-covered trees  This is me pretending  This is all I need   '),(8078,1778579,'22,July,2004','2004-07-22 00:00:00','    The tunnel has been dark for me. My life has been totally up in the air.    Finally, there is some light. I GOT A JOB!!!!    This all transpired in just over a week. I got an email from a company called Anteon Corporation saying they wanted me to interview with them - and the interview happened two days later. That was on Thursday. On Friday I left for Mike and Teresa\'s wedding in Dallas, and we got back Sunday evening. On Tuesday Anteon  called saying they wanted a copy of my transcript from Taylor. On Wednesday they told me that everything was looking good and that I would hear back in the morning.    Well on Thursday morning I didn\'t hear anything before I left for Starbucks (my part-time job). When I got home there was a message on the machine from Anteon saying they were ready to extend me an offer. Well, a bit later I got in touch with them to get the details, and it all looks good to me.    The actual offer should arrive in the mail tomorrow. I need to look it over, sign it, and take it with me to Anteon on Monday morning. Yep, that soon. I\'m sooo excited.    Thank you Lord! Now if you could just help me out in other areas as well.... :-)   '),(8079,1778579,'18,July,2004','2004-07-18 00:00:00','    A month and a half ago when I came out to Massachusetts I had something to look forward to. Something to be excited about. Something to make me want to be here.    And then I went to Dallas for the wedding. It was a good weekend away.    But now I\'m back again.    Only this time, I had absolutely nothing to look forward to.   '),(8080,1778579,'13,July,2004','2004-07-13 00:00:00','    This weekend will be really good. Dan, Abby and I are flying down to Dallas for Mike Nobles\' wedding. We\'ll meet up with lots of friends from Cameroon, and it will be a nice reunion.  And it will be good to get away. Good to have something of a break from this place.  And who knows - maybe I\'ll really like it down there and reconsider moving down to the Dallas area. The only way that will happen though is if I find a job down there first. I don\'t feel like moving into another uncontrollable situation. Besides, I really have it good up here in NE with free rent and family around.   '),(8081,1778579,'12,July,2004','2004-07-12 00:00:00','    If the weather clears enough (somewhat doubtful), my sister, one other friend, and I are considering going to the Providence Water Fires. It should be a good time. It\'ll be a nice small group. I\'m tired of big group activities, and it seems to be hard to do things in smaller groups around here.  If you invite a couple people to something and other people don\'t get invited, it becomes this big old issue of \"why wasn\'t I invited???\" Grow up a bit. It\'s ok to hang out with just a few people sometimes. We\'re not in high school anymore.  Anyhow, it\'s supposed to rain for the next four days or so. Maybe the weather decided to imitate my mood.   '),(8082,1778579,'12,July,2004','2004-07-12 00:00:00','    I guess now I\'m just something of a skeptic because it\'s the safest way to be. I\'ll keep living life, and I\'ll be fine in the end. I always am. But it\'s another layer to break through later. Another piece of protection from getting hurt yet again sometime down the road. Thing is, it\'s my own fault. I let myself get into a situation that was risky. We\'re friends, and maybe one day we\'ll be good friends again. Only time will tell.  Select lyrics from \"Don\'t Stay\" Linkin Park, Meteora  Sometimes I need to remember just to breathe Sometimes I need you to stay away from me Sometimes Iâ€™m in disbelief I didnâ€™t know Somehow I need you to go  [Chorus:] Donâ€™t stay Forget our memories Forget our possibilities What you were changing me into Just give me myself back and Donâ€™t stay Forget our memories Forget our possibilities Take all your faithlessness with you Just give me myself back and Donâ€™t stay  Sometimes I feel like I trusted you too well Sometimes I just feel like screaming at myself Sometimes Iâ€™m in disbelief I didnâ€™t know Somehow I need to be alone   '),(8083,1778579,'12,July,2004','2004-07-12 00:00:00','    So sometimes I\'m stupid and I do dumb things.  I\'ve worked to get past that, and tonight I really tried to be _normal_ and to make an effort not to be stupid in how I acted.  But apparently it didn\'t work, and I got a dose of my own medicine.  Right now, I hate my life. Maybe tomorrow things will be better. But right now...  I just hate my life.   '),(8084,1778579,'02,July,2004','2004-07-02 00:00:00','    I was ok last night. At least that\'s what I kept telling myself.  I woke up this morning feeling sick inside.  Scared that all the conversations, all the emails, all the hours on the phone  All the fun times shared together  Were for naught.  Hoping beyond hope that this is just another bumb in the road.  Oh Lord, what am I to do?   '),(8085,1778579,'10,August,2004','2004-08-10 00:00:00','    I\'m sitting at work, and at the moment I don\'t really have anything to do. My boss is on vacation and the work that he left for me to do is as done as I can get it at the moment.  And he\'s gone for two more weeks.  Oh well. Such is life. I know that once he\'s back and things pick up I\'ll be busy all the time, so I don\'t want to take \"slow\" time for granted.  New in my life? Yeah right, like anything actually happens in my life. Get up. Go to work. Come home. Maybe hang out. Go to bed.  At least it\'s a pretty set schedule. I like knowing my schedule.  Other than that - same old, same old. I still get confused.   '),(8086,1778579,'03,August,2004','2004-08-03 00:00:00','    Congratulations to my best bud  Adam Shideler  for picking up the rank of Corporal in the Marine Corps!  I\'m proud of you, man. Stay strong for these two final months.   '),(8087,1778579,'01,August,2004','2004-08-01 00:00:00','    \"For You To Notice...\" - Dashboard Confessional  I\'m starting to fashion an idea in my head where I would impress you with every single word I said. Would come out insightful or brave or smooth or charming and you\'d want to call me And I would be there every time you\'d need me I\'d be there every time... But for now I\'ll look so longingly waiting... For you to want me, for you to need me, for you to notice me   '),(8088,1778579,'01,August,2004','2004-08-01 00:00:00','    My sister and a couple of friends threw a surprise party for me last night, celebrating the fact that I found a job! I honestly had no idea that it was planned, so that was cool.  Between good friends, ice cream cake, and stoli all around - it was a fun night.  So thanks guys! You\'re all awesome.   '),(8089,3670354,'22,June,2004','2004-06-22 00:00:00','    I need to shoot a gun...I\'m going through withdrawls.  I\'m comin to montucky in a week and it couldn\'t come sooner...I just need to breath that clean montana air...see Clarissa and the rest of my friends...and by rest of my friends I probably just mean my good \'ol redneck friends, everyone else is a dick, so, yeah.  See ya in a week guys.   '),(8090,3670354,'21,June,2004','2004-06-21 00:00:00','    Which also sucked.  I have no idea what I\'m waiting for in this grand new city of Spokompton, but whatever it is hasn\'t showed up yet.  If only I had initiative and motivation to go do something....   '),(8091,3670354,'28,July,2004','2004-07-28 00:00:00','    Here\'s something really cool to me.&nbsp; When living in polson, I\'ve never benched more than 160lbs.&nbsp; Two months into spokane, and now I can do 210.&nbsp; Hmm, strange coincedence?&nbsp; Oh, yeah, I\'m also in a relationship with the woman of my dreams.&nbsp; THAT\'s pretty cool.&nbsp; I\'m healthier, happier, and I have half of a clean respectable house to myself with a parent that I actually like.&nbsp; My friend Jenny moved here too and same thing.&nbsp; She goes to school for 9 freakin hours a week, has her own place and a great relationship going for her, not with that bastard dan.&nbsp; Polson SUCKS MY ASS!&nbsp; I wish Clarissa wasn\'t there.&nbsp; She should be here.&nbsp; Then everything would be perfect.   '),(8092,3670354,'25,July,2004','2004-07-25 00:00:00','    with Clarissa because it\'s so hard to get her to express herself....I don\'t know if what anything I say or do is really sinking in, or if I\'m doing anything wrong, or if I\'m doing something right for that matter.&nbsp; She seems to think that my opinion of her is so fragile that to tell me what\'s up would really set it off....but....she hasn\'t understood that it won\'t...that that\'s what I really want....I want to feel like a working functioning couple that feels open enough to share these sort of things with eachother...and....I can\'t screw up with her....not her...she is balance to me, she is all that\'s beauty, she\'s earth air fire and water to me, and I want her to live without all those horrid barriers around her soul....I know I\'m so far from perfect...but damn......you\'re just too good to be true....   '),(8093,3670354,'14,July,2004','2004-07-14 00:00:00','    I just posted this and then it didn\'t go up...oh well...Went to Montana last week...it went by way too quickly though...Georgia just happened to  move the day after I got there so I never even saw her.....and other than that it was the same old bullshit...everyone always trying to get drunk or get high...pretty irritating..... On the plus side I stayed with Clarissa\'s family for most of the visit and got to ride back with them...and..it was really awesome...I just felt happy + calm about everything when I was around her....and trying to visualize the future...and how I can be a good man for her.....spending that time with her though really did make me happy I\'m alive...for the first time....   '),(8094,3670354,'08,August,2004','2004-08-08 00:00:00','    My birthday always ends up being one of the shittiest days of the whole year somehow(cept last year) and this one was no exception...just a usual day...only I got to hang out with my little brother for the majority of it (yay.) and got to go to a shitty as minor league baseball game with dad, david, and grandpa.  More joy, my grandpa is just a who-knows-how fuckin old version of my asshole dad.  I remembered why I never liked him.  It\'s ok though right?  I\'m a year closer to death now, so that\'s one more behind me....*sigh*  the crazies won\'t leave me alone recently....well that\'s enough bitching for one week   '),(8095,3670354,'03,August,2004','2004-08-03 00:00:00','    I just want to cut everyone\'s head off.  Well, everyone here, at least.  No one is on my side, and talking to them is like watching another version of them drop a noose around your neck.  I don\'t trust any of the bastards here, cept my guitar teacher.  Well shit, he\'s 23 and lives like 45 blocks straight uphill.  I just feel like hating myself, and everything in my surroundings recently, and it\'s getting a little tiring....the \'ol, ultra-negative bi polar seems to be kicking in......can\'t be THAT close to winter   '),(8096,3693380,'24,July,2004','2004-07-24 00:00:00','   Check this forum out   http://www.aimoo.com/forum/freeboard.cfm?id=602411  '),(8097,3693380,'19,July,2004','2004-07-19 00:00:00','   Hi heres a guestbook       urlLink    &nbsp; urlLink'),(8098,3693380,'14,July,2004','2004-07-14 00:00:00','   Wow, I just got a phone call from Tiz, my agent and he wants me to do a film with stars such as Billy Crystal, Reese Witherspoon,Keanu Reeves,Jude Law, Drew Barrymore and Arnold Schwarzenegger, haha nice mix huh?  Anyway the film is about a girl(me of course) who gets lost and cant find her way back home because she was on the way to Russia with her family to see her dad who was in the mafia but he quit and ever since he quit people wanted to track him down and so the family decide that the only chance to visit him would be now, so to go undetected they have to find someone with fake passports who can make them some, then hitch rides(on the way encountering several funny personalities) and then when they get there they find that their father (Arnold Schwarzenegger) had turnes into a Yoga teacher for the stars. Well, while they are there people started to get suspicious and the family had to split up, The girl (me) starts on the journey to their meeting point in China in Bejiing, she then meets a guy and falls in love with him but  he turns out to be one of the people trying to track her down but doesnt know she is the one he has to track down. And then it goes to the journey oft the other family members and they get caught before she does and they get locked in a flat. Then the girl finds the town of Bejiing with the help of her guy (bad)  and he realises that shes the one hes trying to catch but since they are in love he helps her to get his family out of the mafia buisness. Hahaha Anyway seems great. I gotta go have my hair and nails done and a nice massage before i record my next hit single.This will be my last song as i am retiring and going into Movies full-time. So i gotta go but bye for now  Ella  '),(8099,3693380,'14,July,2004','2004-07-14 00:00:00','   Where was I? Lets see, ah yes me with my perfect lifestyle. Well today I won the beauty pageant in Germany, no suprises there. and we have just been on a beautiful Island in Thailand, there were boys,sun,cocktails,the beautiful sea and of course my wonderful family, i just bought a whole new wardrobe and when i came home there were jelly beans everywhere, it must have been my oh so perfect boyfriend who will always love me and doesnt change his mind about me.  Does anyone have any suggestions for more of a perfect lifestyle??? Its really hard once you mentioned what u like.  '),(8100,3693380,'14,July,2004','2004-07-14 00:00:00','   My name is Ella Sprinkles and I am 18 years old, i have the perfect life. Yes a perfect life and here i am top of the class, pretty, great at the arts and i am already a famous actress/singer and i dont have to work for it, i do it when i want to. Ahhhh dreamy i know, and yes i have a perfect boyfriend, with his dreamy eyes, soft lips and gorgeous skin. His personality has to be 10/10. He is everything I ever wanted: Confident but not cocky,polite,cute, clever but not freakishly, hot as it can get, only loves me and doesnt change his feelings towards me because he loves me all the time and noone else. He\'s mine. What else is perfect about my life, ahh yes my family, no fights ever parents get along and dont fight, my brothers and sisters are sweethearts, they are so polite and are always kind and considerate. I love it here. Whatever i want i get in a special machine which i type in what i want and costs me nothing (free gift from my friend who is a famous inventor). I better go, im going shopping :D  Bye Bye  '),(8101,3522772,'09,June,2004','2004-06-09 00:00:00','   Sigh...time realyy flies so fast...without me realising, a week has passed and I haven\'t thrashed out anything in here. Have been busy with work..meetings here and there...really killing me softly...but told maself not to get any negative further...self-control...patience....those are the keys in achieving happiness..(dunno how true actually)... By the way..what shit am I typing? Craps and nothing but craps...   Sometimes I wonder...why am I fated to meet someone so nice but I can\'t have him eventually? Is it called  fate when I was the one who initiated to get to know him? whereas I can jolly well stop myself from chatting... but to stop chatting after I got hooked to it for 10 years?- how to, man? People come and go in our life..but some nice ones are there to stay...linger in your life....make you feel appreciated that you form part of their lives... but at the end of the day, one special person that stands out from the crowd is still not YOURS!!!!!!!   Recently...encountered someone from the net..obviously he appeared the suspicious sort..I couldn\'t care much..but he never fail to make my day with his sweet thoughtful, funny SMSes...somehow..I began to open up to get to know him better... he indeed is a nice person...friendly...funny...and the best part is...he is a true gentleman!!!!. When we first met...I never thought that he would turn up..you know lah..those blind dates..never seems to turn out rite...but he is an exception..Indeed he still makes a gd impression up till this very moment. We had a supper at Fisherman\'s Village..with live fire shows...nice calamaris...and we practically talk anything under the sun....ooopsss...moon I should say...   '),(8102,3522772,'04,June,2004','2004-06-04 00:00:00','   urlLink    phewweeeeeeee....mission accomplished..hehehe...done a para-sail in Penang  Posted by  urlLink Hello   '),(8103,3522772,'02,June,2004','2004-06-02 00:00:00','   Today was a tiring day...not bcoz of work but I\'m tired of sleeping...Been sleeping eversince I got back from ma 24 hours duty...just woke up 2 hours ago...watched t.v and stuffs..and now I am really bored..No frens are online to chat with....practically nothing to do at home...just a stranger living together..someone who is just difficult to communicate with...and the worst part is..I am missing someone....very damn badly....it\'s been long since I last saw him...wonder...wonder...wonder...when can I see him again???????                  Tomorrow gotta work again...at least that can keep me occupied...or else my mind will start to linger.....sigh...   '),(8104,3522772,'02,June,2004','2004-06-02 00:00:00','   Phew....here\'s ma opening blog....finally get to create one for maself.... Today\'s Vesak Day...everyones not working except for me and ma colleagues..Not much excitement here though....just killing our tym doing one surgery after the other...UUURGGGHH...can\'t wait for tomorrow to come..and the next tomorrow..and so on....and so forth....   '),(8105,3522772,'11,July,2004','2004-07-11 00:00:00','   Its Sunday already....been meeting Y for the past 2 nytes....glad we meet up..damn...missed him so much... he is the coolest guy I\'ve never met before..but deep inside no one knows..... n deep inside I know that he\'s such a nice and wonderful person to be wif too...... nuthing really exciting happened today....such a dull Sunday.......  I watched a hindi movie from a vcd...taught me abt SELFISHLESSNESS.... chim... but means a lot... quite a gd show...hindi movies never fail to make us learn something outta their movies, man..... that selfishlessness character is hard to cultivate...but its a really noble thing to do....   '),(8106,3522772,'09,July,2004','2004-07-09 00:00:00','   These few days have been having sleepless nytes...been in the net..chatting with Y till the wee hours of the morning...it was worthwhile though....hehehehe...I am contented enough just to be around him...as the days go by, I beginning to realise that he is such a romantic guy...someone who is very sensitive to your needs and feelings...and I really appreciate that...  Fizal took leave for the whole of this week..... Can\'t wait to go out and watch movie with Y next Friday...gonna watch King Arthur. We planned to watch it at Great world city...hopefully I am able to give him that surprise dat I have planned...   '),(8107,3522772,'03,July,2004','2004-07-03 00:00:00','   Cinta Mencintai Kita - Nurul & Ajai   Tidak pernah ku sangka Rupanya kau lebih ku cinta  Dari hati yang putih Ku hadir untukmu sayang  Aku turut merasa gelora Yang menghempas jiwa Biar lama ku tunggu Ku pasti kau tiba  Kau selalu menduga, menguji Kejujuran di kamar hati Membuatku dihimpit oleh keresahan Ohh...ohh... Kita memendam rasa dan bermimpi  Terbang ke langit Langit yang tinggi Bertemankan awan indah memutih  Sambut tanganku sayang Bawaku ke syurga impian Tiada lagi curiga Oh...Yeah Kaulah yang ku cinta  Kasihku kasihi dia Cinta mencintai kita Kasih ku kasihi dia  Ohh...Ohh...Ohh...   '),(8108,3522772,'03,July,2004','2004-07-03 00:00:00','   Haiya blogggieee... its been a decade since I last pen down sumthing....A lot of things happened throughout June..my aunt got married... Euro 2004 started.. went out with close frens...meetings and stuff...and not forgetting my gruelling 24 hours shifts that I have to do once a week...  I went to the \'Pinocchio on Ice\' show last Sunday..a very entertaining show...worth the money..hehehehe...I brought ma little cuz along...he enjoyed it too...it was a whole lot of fun especially with 2 of my craziieee colleagues; Angela & Kalai...  The most happening thing that ever happened in my life was to be able to get close with a male chatter that has all along been around in my life for the past one year or so... Let\'s just name him \'Y\'... Y has always been known as someone very funny...mature...man of little words...always deep in tots... but overall a very fun company to be ard with...i have liked him for quite a while...eversince he was posted to Vietnam for the Sea Games last Nov...I managed to make some Hari Raya cookies for him to bring along to Vietnam... we met quite often for karaoke sessions...supper with my bestest fren..Laila.. he came over for my Hari Raya Openhouse...I still remembered he gave me a blue silk cloth from Vietnam as a souvenier...dats really sweet and thoughtful of him....there\'s so much to say about him...its gonna last an endless blog...hehehhee..the whole point is..I am glad that we can talk openly abt things...share problems, views together...and he is very gd in giving constructive feedbacks...and mind you....he really is damn cool lor....  No doubt the guy that often gave me sweet Smses everyday....is nice too..but this \'Y\' guy is more real.....coz I see him more often than other guys..... its not a physical kinda attraction that I have for him but I guess its more of his personality that I adore...I know I can\'t have him but by just being around him...makes my day.... nuthing more that I cld ask for... beggars cannot be choosers ..I supposed...hehehehehehehehe.....   '),(8109,3572381,'20,June,2004','2004-06-20 00:00:00','    Today I am feeling good. And it is unusual. Everyone has their bad days right? Well I am having one of those good ones.  I don\'t feel the weight of the world on my shoulders today. I don\'t feel like hiding under the covers today. I feel like going for a nice walk.  Now underneath this inexplicable feeling of peace and hope is a curiousity of Why? What did I do (or not do) to feel this way? Was it something I ate? Did I do something to somehow flood my brain with endorphins?  I think I will just enjoy it while it lasts. I don\'t want to give you the impression that anyone ever notices my dark moods. When I am around people I can actually force myself into a sunny mood most of the time.  I did a bit of praying yesterday. I have been noticing some things that tend to help me have more good days. Praying daily seems to help. Exercise seems to help.  One thing I am realizing is that being self absorbed is very unhealthy. And I think as a society we are all way too self absorbed. If you think about it isn\'t being self absorbed just another way of focusing on what we don\'t have and how to get it? Or a session of self pity for the things we will never get?  Now I am sitting here in my dark room writing a blog instead of going out on that walk...   '),(8110,3572381,'18,June,2004','2004-06-18 00:00:00','    So I\'m not sure if the rest of the country is aware of this but we are all supposed to have a \'Go\' bag packed and ready to go, in case of disaster.   urlLink http://www.nyc.gov/html/oem/html/readynewyork/prep_gobag.html   Nothing like packing a go bag to brighten your day eh? They have all these seemingly logical items you should pack, like water, batteries etc.  I guess \'disaster\' is a polite way of saying \"Oh shit! al-Quaeda(sic) did somethin\' stupid again\". Where we are supposed to go and how is unclear. I guess they will send a bus to pick us up and put us all up at the Trump Casino in Atlantic City or something.  As any New Yorker would tell you, within 10 minutes of any kind of announcement the entire city would be in state of gridlock. The entire city is in various stages of gridlock daily without a disaster happening for God\'s sake!  I am sure there are dozens of guys in think tanks trying to figure out where these retards are planning to strike next. I have a couple of theories on how to catch these turds before they do any more damage.    1) Set up a terror attack staging area.   Maybe if we set up this place where these dudes can blow themselves up without hurting others that would help. Personally you can set it up anywhere between NY and LA AFAIC. Maybe put it in the desert, that would make everyone happy. Set up a TV network to monitor the area 24/7 so we can watch these Nimrods blow themselves to smithereens during halftime or whatever. Hey, maybe they can even coordinate some sort of spectacular halftime show, blow themselves up in a domino fashion or something! They might even consult Paula Abdul on some dance moves...   2) Sting operation:   Here\'s a way to get rid of some of the really dopey ones (i.e. shoe-icide bomber Richard Reid). Set up a (fake) public speaking engagement for Osama bin Laden. Just make sure there is a warning included in the announcement in coded form (pig Latin, ubby-dubby or something) that it is a sting operation, so our own cerebrally challenged don\'t show up my accident. Maybe a good place to do this would be New Jersey in Secaucus or Newark.   Once all these guys are locked in the arena we could do all sorts of stuff to them. One might be tempted to simply nuke the arena. That would be too easy. I say we put on a show for them so horrible, so inconceivably painful that they just start to chew thru their wrists to escape the agony of the experience.  Start off with playing all the \'Cop Rock\" episodes on the Jumbotrons. Next a four hour concert by world famous singer  urlLink \"Wing\"   If they make it thru that it\'s time to bring in Jesse Jackson, Al Sharpton and Richard Simmons to Dance to the oldies..  Anyone left alive after that will be subject to idea #3:   Idea #3:   New York and indeed the whole world has plenty of exercise machines. All the work done to operate these machines is currently wasted. Why not connect up some generators and rechargable batteries to all of them and network all that lost energy potential? So when you do that 30 minutes on the Stairmaster, not only are you shedding unwanted pounds but are storing electrical energy to be used to fry these terrorists!   '),(8111,3572381,'13,June,2004','2004-06-13 00:00:00','    I think we are all born with certain tools for living, and some of us are lacking essentials. Like a sense of other\'s feelings. Society is so wrapped up in themselves and their needs and pleasures that they can\'t see past their own noses.  Some people can\'t sleep at night. My mom used to say that was because they had something on their concience keeping them awake. I wonder how much truth is in that?  When my mind wanders in the twilight time before I sleep I seem to relive all my past mistakes. I have to \'change the subject\' in my mind. Why do I keep reliving my past mistakes?  And why do we place so much importance on other people\'s looks? I am guilty of it, we all are to a certain extent. We all get old. Our bodies wrinkle and stretch and generally wear out. Men are more visually driven they say. Is it really what we decide is beautiful that drives this concept of beauty? Or how we want others to percieve the beauty of our significant others, or ourselves for that matter.  I just had a strange thought. What if we could choose between a mate that looked gorgeous to others but ugly to us, Or a mate that looked beautiful to us and ugly to others. Which would you pick?  As far as mates go I think my ability to please and satisfy her should be most important. It seems that many women are not satisfied with status quo. They always need a relationship to be progressing on some level. I\'m not talking about marriage. I mean being married to a loving husband, having a good job, maybe a kid or two. They look for the next thing. Women always seem to need a goal. Maybe a bigger house? Maybe do the dishes a few more times a week? I think we all need to learn to aprreciate what we have and not look at what other\'s may have.  Women, If you have a man that loves you, takes care of you, is faithful to you and  doesn\'t beat you, you are better off than 90% of the other women out there. And the sooner you realize it the better. If your man possesses all these qualities I want you to go out and buy HIM flowers today.   '),(8112,3572381,'09,June,2004','2004-06-09 00:00:00','    I don\'t know if there will be a part two, I am just covering my bases. :)   Jackie Onassis:   I met Jackie when I was 13 or so. I used to walk her dogs. She had an irish setter (I think it was) which was given to President Kennedy by the people of Ireland This dog was very old at the time. She also had what I think was a maltese. I was introduced to Mrs. Onassis and she was very nice as you would expect. Mrs. Onassis didn\'t spend much time shooting the breeze with me, so no real dirt I can dish. My dad on the other hand swears she would curse like a sailor when she got mad.  I also became somewhat friendly with the secret service guys that were protecting the kids. They had an office in the building where she (and I) lived. In those days the elevators had an operator. When any of the family came on the elevator a special button was pressed to alert the secret service. There were a few incidents where \'John-John\' would duck his secret service by using the service elevator. On one particular day he did this and was mugged of his bicycle in Central Park. It was in the papers. Which reminds me. I was given one of John-Johns old bikes by Mrs. Onassis. I loved it! It had shock absorbers which made the ride nice and smooth. I had never seen a bicycle with shock absorbers before. There is a picture that exists of John-John and Mrs. O where he is riding this bike. How cool!   Billy Joel   Yes, I briefly met my idol! Back in the mid \'80\'s I was a recording engineer. I went for an interview at a recording studio on 42nd street (Chelsea? Clinton or something) Billy was there working on an album called \"The Bridge\" Unfortunately I couldn\'t sum up the courage to have a conversation. I just said hello and sort of basked in the moment. I have seen him on a few other occasions and I saw the hint of recognition when he saw me. Billy is my piano mentor, albeit from afar. His playing totally turned me on to rock and roll piano. I hope to meet you again and chat sometime Billy. :)  I also met the following people during my recording studio days: Debby Harry, Bill Cosby, Roberta Flack, James Ingram, Grover Washington Jr., The original SNL band and many others who\'s names currently escape me.   '),(8113,3572381,'07,June,2004','2004-06-07 00:00:00','    The \'Blog\' phenominon (sic?)is rather new. I know it is supposed to be about current events (I guess it is), but I am writing about my past for now. I would like to use this to collect thoughts on my past and on who I am. I thing it is good psychotherapy to actually write out who I am.   I think on some level I imagine a beautiful princess will read this and be intrigued by me and fall hopelessly in love. Or maybe when I get to know someone I can direct them here to \'out\' myself. Who knows?  I like to think of myself as a very kind and loving person, don\'t we all? But I know I can be a real ass at times as well. :p  It will probably take 100 posts for anyone to even begin to \'get\' me., But I have to start somewhere!  The question I seem to be asking myself at the moment is \"Am I a simple guy, or a complicated guy?\" I think it best you decide this for yourself.  I think if I were to attempt to sum up my feelings on myself I would say I like me, but there is room for improvement.   '),(8114,3572381,'07,June,2004','2004-06-07 00:00:00','    This is a tough one in a way. Trying to take a hard look at yourself can be painful. And the ego will kick in and make some excuses for sure. So this is who I am (in my own perception) at this moment.  I promise you that I can me quite funny and light hearted, but it seems the shit is rising to the top first in my blogging experience. This is feeling very theraputic for me releasing all this sad stuff. I hope someone gets off on it on some level. I should mention I have a few glasses of wine in me as I write this. I will try to be as balanced as possible.  I think I am a nice person. I really like to listen to people, I am not one of those that nods their heads waiting for you to stop talking so I can say something else, totally ignoring what you\'ve just spoken. I am a listener, you have my ear if you are making any kind of sense or have a point. As long as you aren\'t babbling.  I think I am fairly intelligent. I am not as smart as I think I am though. I say this because on some things I can be so perceptive yet on others I haven\'t got a clue. For instance I am extremely sarcastic, yet a lot of other people\'s sarcasm flies right over my head.  I love my parents. There, I\'ve said it. Just like any parent/sibling relationship I have my gripes, but overall they have been supportive and loving thruout my life. God bless you.  At this point you may have noticed I haven\'t gone on about degrees I hold, or how much money I make. I have not been steadily employed since 2000. But that is not why I don\'t brag. I don\'t brag because I don\'t define myself by my work. I didn\'t when I was making 50K a year and I don\'t now.  I am chronically curious. This goes hand in hand with the listening aspect. Even when I have absolutely decided on something, I will ask someone\'s opinion. Quite often their opinion will either reinforce mine in a totally new way, or challenge it in a way I never expected. I have opinions. But if my opinions are enlightened by someone else, I don\'t have the slightest problem with changing them. I see this as a great strength.  Some people ask why. I say why not. Is the glass half full or half empty? I am more concerned with exactly how much is in the glass.   '),(8115,3572381,'07,June,2004','2004-06-07 00:00:00','    I mentioned a few of my good points in my last blog, so before anyone complains I will try to actually take a hard look at my bad points.  I am self absorbed. I guess admitting to being being self absorbed is kinda redundant seeing as I am bothering to write this stuff in a blog for the world to see but there it is...  I am lazy. If I had my say I would sleep 23 hours a day. My dreams seemed to have shattered My way of escaping reality is Zzzzz. Legal and free. The other hour I would be surrounded by people hanging on my every word and worshipping me.  I am judgemental. I tend to judge people on the way they talk and the company they keep. Is that wrong?  I am a megalomanic. I think on some level I am not better than anyone else, but uniquely gifted.   I feel shortchanged. I live in a world where people are starving, beaten, and homeless, yet I feel shortchanged for not having a wife that looks like Jennifer Aniston and living on the beach in Malibu.  I don\'t think of other\'s feelings like I should. I need to work on exercising the golden rule in my life.  I\'m sure there is more, that\'s all for now.   '),(8116,3572381,'07,June,2004','2004-06-07 00:00:00','    I suppose I should dedicate this to all the women that have broken my heart.  From my earliest days of puberty I have been rejected by the women/girls I have truely loved. I know that sounds \'oh poor me\' but that is how I see it.  I should start off by saying I was very naive (and still am to a certain degree) when it comes to relating to women. I somehow got it into my head when I was young that I was unattractive to the opposite sex and was very withdrawn and shy. Upon reflection I realize this was not true at all. It probably is true now because I have not taken care of myself and am 280 pounds of blubber. Yet I still have a cute face.   To all you women that say all men want is sex, I want to tell you it is not true. For a time I wanted and pursued meaningless sex that made me feel empty inside. I only pursued it because the women I loved and respected rejected me. I got some satisfaction from it emotionally on the conquering aspect of it. \"I talked a women into taking off her clothes and spreading her legs whoo hoo!\"  I crossed a line in my mind that I wouldn\'t only have sex with someone I loved, that I would have sex for it\'s own sake.  The night Jennifer told me she wanted to be \'just friends\' I called this woman around the block that had been pursuing me and released all my pent up sexual frustration on her.  The sad thing is this other woman really loved me, and I just used her as a sperm receptacle. Sperm that was meant for Jennifer, the one I loved with every fiber of my being.  Jennifer, when I professed my love for her, told me it was impossible to love someone you haven\'t slept with.   Not to say Jennifer was my first love, she was only the icing on the cake. My experience with Jennifer was what drove me to meaningless sexual relationships.  I think my first love was Laura. I was 13 or so. When you are so young the slightest sideways glance can crush your heart. I somehow from the deepest reserves of my being summoned up the courage to ask her to skate with me at the roller rink.  Me: Would you like to skate with me? Laura (goddess): no.  That single word so casually spoken crushed months of longing and admiration and masturbation dedicated to her. That single word was spoken by her as if I had asked to take a shit on her forehead.  Anyway I have actually loved very few women and the ones I did rejected me. I would like to take this moment to apologise to all the women I used for sex. It seems a few of the ones I didn\'t love really did love me.   '),(8117,3572381,'07,June,2004','2004-06-07 00:00:00','    Yes, I say follower of Christ instead of Christian or \'born again\'. Unfortunately there are many who claim to be Christians but aren\'t walking the walk, or a least trying to. Some have gone wayward on this path. Some just say they are on it, but aren\'t.  It is easy actually to tell if you are on the right path, we have a guidebook. When we do something wrong against God it is called sin. If you are doing something that the book says is a no-no, you are sinning, it\'s that simple  If this seems strange in the context of my previous discussions of loveless sex and other things I have done it is easily explainable.  Before my decision to follow Christ I did things that were clearly against the teachings of Christ. The difference is that I now do not do those things, and if I slip up and do them I ask for forgiveness. You can\'t fool God. He goes by your heart. He knows your thoughts. If you have done something wrong against God and are sorry for it, all you need to do is ask Him to forgive you and He will!  I think the thing that turns off many non-Christians is the concept that they will be held accountable for their actions. NOBODY LIKES TO BE TOLD WHAT TO DO! Even if it\'s in their best interests.  Another thing I think turns off non-Christians is the belief that Christians think they are better or more pious than non-Christians. The true follower of Christ thinks he is better than no one.  I can best demonstrate my thoughts on this with an analogy. Let\'s say we are all in an AA meeting. Some of us are recovering alcoholics, some of us are thinking we may have a problem and still others are in denial that they have a problem. The true Christian is like the recovering alcoholic in as much as he realizes he has a problem, and he is trying to live a sin-free (alcohol free) life. Just as the alcoholic is never \'cured\' of wanting a drink, the Christian is never cured of his desire to sin.   Sorry if I got a little \'preachy\' there. I just wanted to clearly present a few of my thoughts on Christianity.   '),(8118,3572381,'07,June,2004','2004-06-07 00:00:00','    In my family, the truth was always carefully hidden. Not overtly snobbish, but a subtle snide remark or a glance that was a little too long betrays the superior attitude that lies beneath.  My brother is a drug addict.  I don\'t mean he smokes a little pot, I mean he sticks needles into his arm, steals anything from anyone etc. He was homeless for a time. He is now in prison for the second time (that I know of).  My brother\'s \'condition\' is a family secret. I have never seen my dad cry until the last few years.  At this point the story can take several paths. I can tell you some horrible stories, tell you how we are dealing (or not dealing) with this family crisis etc. I think I will tell you how my brother got on this path.  To me there is a line that we cross in our mind when we make a decision to do something. Also I believe we are born with tendencies to certain types of behavior.  A thought crosses our mind, and we either act on it or don\'t. Sometimes these thoughts can be rather bizzare, at least to some people.  One day my brother somehow got it into his mind to smoke pot. Now I know there are plenty of people that \'cross that line\' and never go any farther. My brother crossed the line into drug use. Once there he continued to cross lines, but I think the biggest leap was into drugs or altering his reality. I believe once you have decided that altering your reality using chemicals is OK jumping to a different chemical is a small step.  That is why I believe that pot is a gateway drug. To phrase it differently, how many of you have gone from being totally straight to say smoking crack or shooting heroin?  Anyway he started on pot and ended up on heroin.   '),(8119,3572381,'07,June,2004','2004-06-07 00:00:00','    Did that get your attention? Good.  I have a question for the dozen or so people that may read this. Let\'s imagine I have just won a fantastic amount of money in the lottery. You find out about it thru one of the many full page ads I have taken out in every major American newspaper. It reads something like this:  ---------------------------------------------------------------------  \"I have decided that I have no real need for this much money, and I would like to spend $20,000,000 of my winnings on something that will benefit mankind. Therefore I am now accepting plans for the dispersement of these funds to a deserving cause.\"  \"The person or persons that I decide have the best suggestion for the cause this money should be dispursed to may choose to either accept a one million dollar award, or forfeit the award in which case I will donate $25,000,000 to their chosen cause.\"  ---------------------------------------------------------------------  Two part question:  1) What would you want the money to be spent on? 2) Would you turn down the one million dollars to see an extra 5 million go to your chosen cause?   '),(8120,3293519,'11,May,2004','2004-05-11 00:00:00','   My god, I\'m so disappointed. Yesterday I was posting my first message to my blog announcing the MR2 Spyder Journal, and just yesterday my Check Engine Light (CEL) turned on. I ran to google and searched for some knowledge. Can you believe that a loose gas cap can cause that little light to turn on!? Sadly it wasn\'t my case. Some other, say cheaters, only reset the Engine Computer Unit (ECU) to turn off that mind boggling (not blogging) light. For those who rely on this technique, you must know that the CEL is there for a purpose, not only to mess with you. If you turn it off, your car will keep running with a simple problem that can become a major threat. If you are able to live with that fact, there are 2 ways to reset the ECU.   The simpler one:  just disconnect the car battery and wait 10 seconds until all the car memory is cleared. The trade of this technique is that all memory is reset, even your car radio stations presets and car clock if any.  The not simple but not hard one:  Every car has a couple of fuse boxes. One inside your car, and one (or more) under the hood, near the engine. Open the box and look for a big fuse with maybe the words ECU on the top and unplug it; wait for 10 sec for the memory can be entirely erased then re-plug the fuse.  Either way, the light will turn off. If you are facing a real problem with your car, as soon as you turn it on the light will be lit again. And there\'s where the real problem begins. What\'s even worst is that you can get charged up to $80 dollars just to connect your car to the computer that check the error code stored at the ECU. That\'s a real shame but if you love your car nothing is too expensive.  Well people that\'s all for now. Hope to get back soon.  '),(8121,3293519,'10,May,2004','2004-05-10 00:00:00','   Hello people, this is my first and maybe the only blog I\'ve created. Just wanted to give it a try to this blogxperience. Hope you find this little corner of the internet interesting... I say Hope cause definitely it isn\'t :)  My main theme here will be my car since I don\'t know anyother thing in my life that could be appealing to others. I\'m not a big car enthusiast but an average one. My car is a Blue 2000 Toyota MR2 Spyder, with black top. I\'m starting to enhance it so maybe this blog could serve as a guide to those kids that doesn\'t know where to start with their cars. All comments are welcome, but be kind.  About me? Well, I almost say goodbye without introducing myself. My name is Raul Ortiz, live in Toa Alta, Puerto Rico (yeah, we have cars! And computers with internet also.. hehe.) I\'m 25, with a bachelor degree in Computer Sciences. I work for a consulting firm and currently running with two contracts... By the way, I\'m still at work, supposed to be doing some important tasks. For complete details you can check my Profile section.  There ya\' go... hey.. don\'t fall asleep, I\'m almost finishing.  Ok, you can go and read someone else\'s blog.  Singing out... Raul a.k.a IncubusKnight  '),(8122,3646377,'15,June,2004','2004-06-15 00:00:00','   I didnt do much  urlLink today , so i still dont know what to link here.... '),(8123,3646377,'15,June,2004','2004-06-15 00:00:00','   I left work at midnight and came back at 10am, just to find  urlLink this sick site  for you. What the fuck is their malfunction?? How idiotic and retarded can these guys be??!!  On more serious issues,  urlLink Make Trade Fair  is an interesting website I plan on reading through. A reminder to open one\'s eyes every once in a while.  And if you still beleive what the media feeds us, you should browse through  urlLink this site .  And stop beleiving  urlLink everything you heard in school .  For the couple of idiots I know (and the millions I don\'t - and don\'t want to) who don\'t know how to speak correctly, please  urlLink memorize this shit  (My father\'s wife, my mother\'s aunt). And for god\'s sakes drop the condescending tone of voice!!! You are nobody to be condescending with me. Better yet, shut the FUCK UP!!!! with people like you,  urlLink even this starts to make sense .    For the ultimate gaming experience, dump your Xbox and PS2, and check out the new urlLink  iTari .  Some  urlLink seriously good design here .  This is a great idea to help you deal with  urlLink telemarketers  and  urlLink ugly chicks .  In case you have some extra cash... pass me some!! ....then  urlLink buy one of these .  The DEA tells us how to start a new carreer  urlLink growing opium poppies and processing it into heroin !! Make sure you tell your kids!!  This is  urlLink just too gay .  Why wait for  urlLink Gmail , when you can get your 1Gb e-mail account at  urlLink Walla.com !   Farenheit 9/11  urlLink defense  , and the film\'s  urlLink unanswered questions . I\'m still waiting to see it.   urlLink A nice gallery of funny pictures .   urlLink Somebody\'s got a point here .   urlLink They all are fucking morons !   Thats all I could gather today. Ill be back soon. '),(8124,3646377,'15,June,2004','2004-06-15 00:00:00','   ... to actually start this blog. It\'s been a while since I\'d told some of you I\'d be posting those vacation and party pictures...  Sorry I\'m late. I still haven\'t downloaded all of the pictures, so I guess we\'ll have to wait some more.  But in the meantime, hope I can make you spend some time in some of my favourite pages.  Some may be  urlLink funny , others may be  urlLink stupid  or  urlLink interesting , others  urlLink serious . Just browse around and see if you are interested. I\'ll appreciate any comments. Thank You.  If you\'re feeling  urlLink nosstalgic for your long-lost childhood tv shows ...  urlLink buy a piece of them ! Gotta love these  urlLink crazy commercials . Play  urlLink Lemmings  online. '),(8125,3646377,'15,June,2004','2004-06-15 00:00:00','   Ja! espero que no les moleste. Drop me a line saying what you think of it....  So anyway... today\'s a very fucking long day. And all around me, all I see is  urlLink STUPID people ! I mean, shit is  urlLink NASTY  around here. Sometimes you dont even want to get out of bed...  Or at least tell  urlLink them what you think about them !!!!!  Think about  urlLink this guy , and all the fucked up things he\'s done.   Ma  urlLink imbolnaveste ! Ce sa mai spun de tipi  urlLink astea , care omuara pe  urlLink astea  pentru  urlLink asta .      Va rugam, o    urlLink Pauza!!!!!!!!!    Am mai descoperit ca si Romani sunt  urlLink printre cei mai nebuni  pe aici...    urlLink Vorbesc serios !  Chiar ca ma doare sa spun, de e greu sa surizi prea des.  Alte exemple:    urlLink Berlusconi    urlLink these hipocrites    urlLink And these jerks    urlLink More stupid gringos    urlLink these idiots    urlLink And these    I\'m too tired... bye.... '),(8126,4122943,'03,August,2004','2004-08-03 00:00:00','   hahahahahahaha... today netball very tiring coz got coach oredi.. coach oso volunteer coach lai de.. den lalalalalala.. nth more to write..  '),(8127,4122943,'03,August,2004','2004-08-03 00:00:00','   urlLink    oso at orchard&nbsp; urlLink  '),(8128,4122943,'03,August,2004','2004-08-03 00:00:00','   urlLink    me at orchard on my bdae..&nbsp; urlLink  '),(8129,4122943,'03,August,2004','2004-08-03 00:00:00','   today.. after two days of not being sent home, wei liang send me home..lolx...  '),(8130,4122943,'03,August,2004','2004-08-03 00:00:00','   SIANZZZZZZZZZZZ!!!!!!!!!!!!!!!!!!!!!!!!!!!!  '),(8131,4122943,'03,August,2004','2004-08-03 00:00:00','   today no mood to write.. i\'m juz too angry..  '),(8132,4122943,'03,August,2004','2004-08-03 00:00:00','   haiz.. so boring today.. alot of things to settle seh.. e-fair summore.. haiz... so de buzy onez.. i summore leader of decoration.. haiz.. need to do mascot summore.. haiz.. dunno when wan to finish.. haiz.. still need to decorate de stall.. wakau.. eeeee... so angry.. haiz.. wad to do.. juz do wad i need to do lorz...  '),(8133,4122943,'03,August,2004','2004-08-03 00:00:00','   urlLink    my cute cute doggie doggie.. hee.. i lyk..&nbsp; urlLink  '),(8134,4122943,'03,August,2004','2004-08-03 00:00:00','   haiz.. mi today very sianz sia.. any teacher come in oso scold n scold.. kanina.. haiz.. me n my foul mouth.. haiz.. mi so bad luck go to toOpid claz.. all teacher oso noe my claz so UNMOTIVATED n not united.. tats wad all teacher in my claz says.. haiz.. idiot.. de chinese bois lorz.. lyk wad sia.. dun cooperate.. den not mature sia.. haiz.. oni a few mature thinking lorx.. de rest lyk S**T ar.. heck care.. Kanina.. haiz.. slipped my tongue.. lolx.. tats all ba..  '),(8135,4122943,'03,August,2004','2004-08-03 00:00:00','   haiz.. i always start my blog wit a haiz.. haiz.. i dunno y my life is sooo DAMN bored.. wan to go out but lazy... nv go out very sianz.. watch tv oso nth to watch.. i hate my life.. but de oni thing i like bout my life is my dear.. lucky to haf him.. or else dunno wad to do.. haiz.. today i wake up, i bath, i eat, i on com, i chat wit dear, dear offline, i off com, den watch farking tv, den dear call ask me online, den on com, den chat wit dear, now writing blog....boring isnt it??? haiz.. lucky got computer or else DAMN sianz...lolx.. nth more to write.. buaiz..  '),(8136,4122943,'03,August,2004','2004-08-03 00:00:00','   today ish saturday n it is so damn BORING... if still got ndp, sure today go practice onez.. haiz.. sux ar today.. oni amirah come my hse do project n den after tat chat wit dear n den NOTHING.. eat, watch tv.. so bored until nth to tok about...haiz...  '),(8137,4122943,'03,August,2004','2004-08-03 00:00:00','   haiz.. today my de bdae.. hee.. so happy.. but horz.. its a friday de 13th.. lolx.. bad luck rite.. juz kana catch ny doris ho.. fer tying hair oni... she jealous mar.. she no hair to tie wad.. lolx.. hee.. den still need to bring home de toOpid ndp bag.. n horz.. i get bags n bags of present.. lolx.. den go home wit a lot of things lyk wad sia.. so sianz.. haiz.. den after tat change go town take pics n eat oni.. hahax.. quite fun.. den go colour court see my dear.. hehex..  '),(8138,4122943,'03,August,2004','2004-08-03 00:00:00','   haiz.. today arh.. i horx.. dunno larx.. go follow my fren go jp.. go in one shop to another non-stop... eat eat n eat.. haiz.. wah.. my leg very damn tired and pain sia.. haiz.. come home straight away lie down on bed.. hahax.. den take a nap.. den on com lorx.. yeah.. tmr my bdae.. hehex.. tmr i 14 yrs old horz.. dun playÂ².. hahax.. haiz.. sianz.. buaiz..  '),(8139,4122943,'03,August,2004','2004-08-03 00:00:00','   Moi De NaMe iSh :  aTiKa...(keep it real)...      Moi De NiCkz iSh :  PiXiE Lee... (cute ar)...-_-\"    Moi De bDaE iSh :  13.08.1990...(still young horz)...      Moi De bOi ish :  Wei Liang...(my dear)...   Moi De FwEnZ iSh :   Mira, SinYi, Jas, ZhaO Ling n MoReZ...     Moi De SkOoL iSh :  Boon Lay SecOndAry SkoOl mY dEaReSt...(eeeww)    Moi HoBbIeS iSh :  mSn, sUrF NeT, sLeEp...zZz....      Moi GoDsIs iSh :  FioNa a.K.a ZhU...(oink, oink)...hehex...    Moi GoDmOm iSh :  SinYi, ZhAo LinG, NaZirah...      Moi GoDtWiN SiS :  AmiRaH ba..(we loOk alMost tHe SamE).. hahax...    Moi GoDbRo iSh :  JoNaThaN a.K.a MeOw...      Moi GoDgRanDpa :  Xun HoNg...    Moi GoDgranDma :   SiLinG...     Moi GoDUncLez :  NicHoLas n KeNnY...     Me IsH a Sec 2 gEr naMe aTiKa a.K.a PiXiE.... mE luRvE mOi dEaR berI muCh... LoLx... wO hAtE PpL wHu wO hAte...LoLx... wO LuRvE sCiEnCe n MaThs CoZ wo ToP maR... :p wO LyK tO haNg OuT WiT fWenZ n Jk Jk wiT TheM... wO LuRvE NDP pRAcTiCe N WiLL MisS tHe A.Is wHeN iTs OvEr..:\'(    E-mAiL mE @: pixielee_90@hotmail.com i ThiNK taTs EnoUgh OreDi... i ToK sO mUcH OneZ... =D  '),(8140,4122943,'03,August,2004','2004-08-03 00:00:00','   haiz.. todae no whr to go den i go acc zhao ling n wang min go polyclinic take mc coz they nv go chinese orchestra.. lolx.. wait so long lyk siao ar.. FARK ar de doctor.. so long.. den still need to wait get the medicine.. haiz.. den after tat go eat near polyclinic thr... eat until cannot finish eating.. i eat so slow.. sob.. sob..   den go colour court.. hehex.. i lyk.. hee.. =D  '),(8141,4122943,'03,August,2004','2004-08-03 00:00:00','   urlLink    moi dear n sam tan&nbsp; urlLink  '),(8142,4122943,'03,August,2004','2004-08-03 00:00:00','   urlLink    aLL dE FaNtAsTiC A.Is..&nbsp; urlLink  '),(8143,4122943,'03,August,2004','2004-08-03 00:00:00','   urlLink    mOi clAsZ AI..&nbsp; urlLink  '),(8144,4122943,'03,August,2004','2004-08-03 00:00:00','   urlLink    NDP rulez.. i wont ferget tis experience..&nbsp; urlLink  '),(8145,4122943,'03,August,2004','2004-08-03 00:00:00','   ytd was fun.. we were the first to perform.. de best part of it was de party.. i n moi frenz went crazy there.. we were out of our mind.. we jump n jump n shout n shout until our schoolmates cant bother bout us.. we continued doing it until we went dizzy.. we stop n rest fer a while n then we went wild again.. we took h2O can drink n we drink a lil bit n den pour it on each other.. .lolx.. and then de atmosphere went boring again until one of moi frenz asked us to fill up de empty can with de dirty water from in front of de stage n den pour it on our claz AI.. he was wet very very wet n accidentally drank de water.. YUCKS.. he den took revenge by rubbing gel glittering stuff on moi head.. i chased him around de stadium but could not catch him so i stopped n wait fer him.. once we saw him we wet him again wit de disgusting water n siew chen drew sumthing on de AI\'s face wit a parmanent marker but he could rub it off... after all de running, catching n crazy-ing, i felt lyk going to de loo.. so we went.. but when we went back to de stadium, our sch left.. haiz..so we went back to de indoor stadium ourselves.. there we took our snacks.. damn, it was so crowded.. te siang(think spell lyk tis) n wei liang  took one box of de oxygenated water.. haiz.. n cannot managed to bring all of it back.. NAUGHTY...  '),(8146,4122943,'03,August,2004','2004-08-03 00:00:00','   wah.. so fast ndp oredi sia... but horz.. i need to perform lehz.. tis will be the last time i see de A.Is n Mr Chong... whu cares bout Mr Chong... lolx... haiz... it will be boring w/o NDP practices... haiz.. feel lyk crying sia.. i will nv ferget this experience.. haiz.. later need to perform n see fireworks again.. YESH...   i scared i will make mistakes there.. summore live on tv sia... if i do de wrong movements then it will be obvious sia.. summore moi family all watching.. den later got party.. YEAH.. i cant wait to go.. i think i am de most luckiest ger in de whole world to be participating in NDP.. moi school is lucky too to have the opportunity.. i wont disgrace moi school later.. i will do moi best ba..  '),(8147,4122943,'03,August,2004','2004-08-03 00:00:00','   no mood to write fer today.. wats terrible ish tat I CANT SEE MOI DE DEARZ.... =(   miss him alotz..  '),(8148,4122943,'03,August,2004','2004-08-03 00:00:00','   urlLink    moi dear.. so cute...hehex...muackz...&nbsp; urlLink  '),(8149,4122943,'03,August,2004','2004-08-03 00:00:00','   today ar.. oni go sin yi hse do toOpid project fer the toOpid azlin... we(sinyi,me,jasmine,mira,zhaoling) meet at the control station and then walk to sinyi hse walk past colour court whr i saw moi dear playing bball... soOoOo happy... den we go to mcdonald n take away food go to sinyi\'s hse eat and then do project... den after tat i n jasmine go home together... before we go home, i go to colour court se moi dear again and then go straight home... hehex.. wad a long day... now i waiting fer moi dear come back from colour court chat in msn... lata i cannot use com den very bad luck lorx... haiz...  '),(8150,4122943,'03,August,2004','2004-08-03 00:00:00','   urlLink    me n moi fwenz..&nbsp; urlLink  '),(8151,4122943,'03,August,2004','2004-08-03 00:00:00','   urlLink    went to wildÂ² wet play tis breath-taking game... &nbsp; urlLink  '),(8152,4122943,'03,August,2004','2004-08-03 00:00:00','   today went to wildÂ² wet.... walau... the queue very de long sia... queue fer about 1 hr u noe... den horz.. still need to queue fer the locker... haiz... but the games all quite fun lorx... oni the toOpid QUEUE soOoOo damn LoOoOoNg!!!!!   i think i become more darker oredi.. haiz... nvm lorx...  i got such loyal stead no need to worry... lolx... hope so... juz now go wit amirah n sin yi... amirah wore a pink dress... sin yi i think wore a pink t-shirt n a skirt... i went there two times oredi.. so sianz... then we pay the entrance at children price u noe.. the person tot we all children sia... nvm lorx.. pay cheapÂ² better.. hahax.. den after tat go makan at the foodcourt there.. very cheapÂ² onez.. n not bad lorx.. after swimming hungry lyk wan to die lidat.. den after tat go home... i wan to walk pass colour court but sin yi dun wan to accompany me... :\'(    toOpiD sin yi... now i\'m waiting fer moi dear online.. but still havent online... haiz... he ar.. dunno go plae bball or flirt... hehex.. lolx..  i think he now going home ba.. cant wait to chat wit him... lolx... miss him alot u noe..  '),(8153,4122943,'03,August,2004','2004-08-03 00:00:00','   urlLink    PuSs iN bOoTs.. Moi De FaVouRiTe cReAtuRe.. SoOooOo CuTe.. hahaXx...&nbsp; urlLink  '),(8154,4122943,'03,August,2004','2004-08-03 00:00:00','   todae is ok in school lorx... first period reading.. wakau!! reading sux man... den after tat is the farking yuana lesson.. hate the lesson man... but lucky today go com lab... got air con.. hehex... den after tat ish sciance... god SPA test ... wakau!! moi mind went blank sia... heck care lar... fail oso fail lorx.. whu in this world care... den after tat recess... WoOhOo... can see moi dear lar... hehex... den after recess ish english... tat toOpid azLin lesson... BORING!!! she ask us to do wad newsletter lar... tOoPiD... den after tat ish history.. noT baD loRx Ms Mok teaCh VerY tHe GoOd oNEz... hehex... Den AfTer Tat LiTeRAtUre... fun ar... ms AnG teach veRy de gOoDiE... i ToP fer LiterAtuRe tEst.. Fer The fIrSt tIme In Moi WhoLe LiFE.. gOoOoOoD... hahax.... MUACKZ>>>  '),(8155,4122943,'03,August,2004','2004-08-03 00:00:00','   urlLink    aTiKa a.K.a Pixie....&nbsp; urlLink  '),(8156,4122943,'03,August,2004','2004-08-03 00:00:00','   Moi De NaMe iSh :   aTiKa...(keep it real)...   Moi De NiCkz iSh :   PiXiE Lee... (cute ar)...-_-\"   Moi De bDaE iSh :   13.08.1990...(still young horz)...   Moi De bOi ish :   Wei Liang...(my dear)...    Moi De FwEnZ iSh :  Mira, SinYi, Jas, ZhaO Ling n MoReZ...   Moi De SkOoL iSh :   Boon Lay SecOndAry SkoOl mY dEaReSt...(eeeww)    Moi HoBbIeS iSh :  mSn, sUrF NeT, sLeEp...zZz....   Moi GoDsIs iSh :   FioNa a.K.a ZhU...(oink, oink)...hehex...    Moi GoDmOm iSh :  SinYi, ZhAo LinG, NaZirah...   Moi GoDtWiN SiS :   AmiRaH ba..(we loOk alMost tHe SamE).. hahax...   MoI GoD SiS iSh    :   KeLyN   MoI GoD dEaR        :  WoOf WoOf    Moi GoDbRo iSh :  JoNaThaN a.K.a MeOw...   Moi GoDgRanDpa :   Xun HoNg...    Moi GoDgranDma :  SiLinG...   Moi GoDUncLez :   NicHoLas n KeNnY...      Me IsH a Sec 2 gEr naMe aTiKa a.K.a PiXiE.... mE luRvE mOi dEaR berI muCh... LoLx... wO hAtE PpL wHu wO hAte...LoLx... wO LuRvE sCiEnCe n MaThs CoZ wo ToP maR... :p   wO LyK tO haNg OuT WiT fWenZ n Jk Jk wiT TheM... wO LuRvE NDP pRAcTiCe N WiLL MisS tHe A.Is wHeN iTs OvEr..:\'(    E-mAiL mE @:   urlLink pixielee_90@hotmail.com    i ThiNK taTs EnoUgh OreDi... i ToK sO mUcH OneZ... =D  '),(8157,4220994,'17,August,2004','2004-08-17 00:00:00','   Another day goes by, life deals out more crap and I am sick and tired of it! Well at least I am more normal then this lady: urlLink http://www.users.globalnet.co.uk/~author/ . Not that I am in any position to judge anybody else. But I do think that believing Garth Brooks is your past lover is a little strange. I don\'t think I believe in reincarnation, but who really knows what lies beyond our earthbound selves. As a matter of fact if reincarnation is true, I don\'t want to know it. To know that everything I have experienced, every lesson I have learned, every emotion I feel will be washed away into a new life is heartbreaking. I want to keep my memories, all the little \"life moments\" I have tucked away inside me. And forget me ever being excited about doing this whole life thing over again. Once is tiring enough! I do not even pretend to know what is on the \"other side\" but I believe it is better then here and now. Do you believe in reincarnation? '),(8158,4220994,'16,August,2004','2004-08-16 00:00:00','   As a nurse, I am faced with death or dying at least weekly. This closeness to death has always made me curious to what death brings. I won\'t get into any religous beleifs here, but I will say that the study of death (thanatology) intrigues me. I have always felt like I can  feel  death. I am on this depressing subject because of an  experience today. On the way to work I passed a very bad accident, I felt death all around me. I have felt anxious since then. Living in a small town word of mouth travels faster then the newspaper has chance to report it, and I heard that the man in one truck did die. Which has contributed to my uneasiness that I have felt since passing the accident. I often brush off the \"feeling of death\" and tell myself that I am nuts, but hearing that there was a death makes me uncomfortable with knowing that maybe I did  feel  it. It reminded me of that scene in The Sixth Sense, where the boy and the mom are behind traffic at a car accident and the little boy says, \"mom someone died in the accident\" and the mother asks him why he says that and he says \"she is standing by the window\". ok now that I have myself all freaked out and all of you out there telling yourselves how much of a freak I am, here is something kinda funny. My sister sent me this link for The-Link-O-the-Day:  urlLink Geroge Bush\'s Blog '),(8159,4220994,'13,August,2004','2004-08-13 00:00:00','   To continue one of my favorite aspects of my old blog...My link of the day. Most usually these are dumb links to awful websites, funny websites or just plain odd. Todays link is pretty amusing to me. If your in the market for a  used  blow up doll, then too bad you missed the auction! A friend sent this one to me.  Link-O-The_Day urlLink eBay item 5501862293 (Ends Jun-20-04 06:42:00 PDT) - Realdoll by Abyss Creations Body-5/Head-4 '),(8160,4220994,'13,August,2004','2004-08-13 00:00:00','   After many frustrating hours of being ready to throw the laptop out of the window, I finally got my site working with CSS. I love how CSS makes everything so clean and prevents all the extra typing of tables.   Two more days and my little brother leaves for college. I can\'t believe he is old enough to go off to college. I know I will miss him terrible. From the time he was very little I have always mothered him and spoiled him. So now he is going to run off into the big scary world without me to protect him. It is depressing. I will cry I am sure :0)  Which brings me to my next rant. I overheard someone tell someone else, \"Well my mother always told me that happiness doesn\'t come to you, you have to go out and find your own happiness.\" Now exactly what moron finds this to be great wisdom? Where exactly is this happiness lurking that one can go find it? I would like to know. As a long time sufferer of major depression I am going to be pissed if all along I could have went and found happiness lurking under a bush somewhere. What do you think, can you just go out and find happiness?? (and yes I know this is just a figure of speech, just wondered if anybody else found it as absurd as me) '),(8161,4220994,'13,August,2004','2004-08-13 00:00:00','   Well first it was, livejournal, then Angelfire, then Xanga, now back to blogger.  I hated livejournal, I could never quite get used to posting in it and the look was boring, Angelfire just plain sucks so don\'t waste your time with that. Xanga is ok if you wanna pay for it. Blogger seems to be the most user friendly and I love that i can click on my handy little google toolbar to post. Yay! I love google. '),(8162,4075064,'31,July,2004','2004-07-31 00:00:00','    Whether or not you love John Kerry or hate him is an entirely different argument. Those who love him didn\'t care whether his speech was perfect or whether he fell flat on his face. Those who don\'t support him aren\'t going to vote for him anyways, so they probable weren\'t watching.  No, the only people that his speech at the Democratic National Convention in Boston applied to were those who were in the middle and undecided on who to vote for come the election. Many have claimed that John Kerry needed to reach out to his listeners and prove to them he was capable of winning their vote. I believe he easily completed this task. I have heard numerous John Kerry speeches, none of which I was able to keep from falling asleep. He may be a brilliant man, but if he is unable to connect to his audience, it really doesn\'t matter.  For the first time I heard John Kerry actually sound as if he cared about becoming President and he also didn\'t act like an Al Gore clone. The enthusiasm in his voice should have been enough to convince those who questioned whether or not he was right for the job from a leadership and caring perspective.  Many of the issues discussed will remain a question mark as to whether you agree with President Bush or Senator Kerry. But unfortunately, while such issues should be the main concern on American\'s minds come election time, I believe the main concern is whether or not they are likeable characters. Hence, prior to Kerry\'s speech on Thursday in Boston, the edge easily goes to President Bush, even though he sounds like a five year old child when in front of the podium.  And although I don\'t think such traits are that important when voting for President, I do believe there are a ton of people out there who do feel that is the most important issue to discuss.  So the question I raise to the millions of those reading this is: How important is being a \"nice guy\" when you consider who to vote for for President?   '),(8163,4075064,'30,July,2004','2004-07-30 00:00:00','    How ridiculous is the Major League Baseball economy when the same teams year after year have to sell their good players to the usual contenders for prospects and some cash?     Did anyone realize the MLS (Major League Soccer) held their all-star game today on national television?     Does anyone really care?     I also saw that the WNBA was televising a game today on television     How is that league still in existence?     Brett Hull\'s agent is in discussions with Mario Lemieux and the organization about joining the Penguins for the final years of his great career.  How awesome would it be for him to break Gordie Howe\'s all-time goal record in a Penguins uniform?     I don\'t know about any of you, but football can\'t start up fast enough.     I\'m so happy to see Aramis Ramirez (third baseman for the Chicago Cubs and previous Pittsburgh Pirate until the Bucs traded him before the deadline last season) struggling with the Cubs.  Oh wait, he hit three HR\'s yesterday giving him 21 on the season.  Don\'t the Pirates have a major hole at third base?  I wonder where we could find one?  Thats not to say that newcomer Ty Wiggington can\'t fill the void, but Ramirez should still be in a Pirate uniform     The New York Yankees will be in the World Series again.  Yeah, there sure is parity in Major League Baseball!     How long before the NFL starts up again?   '),(8164,4075064,'30,July,2004','2004-07-30 00:00:00','    Let\'s analyze this objectively....  If the Pirates do not trade pitcher Kris Benson by tomorrow\'s deadline, the chances of them signing him long term after the season is over is slim to none and slim just walked out the back door.  Hence, Mr. Benson will hit the free agent market come next year and like many other average ball players he will get a juicy offer from one or two idiotic teams.   If the Pirates trade him before the deadline, they may (or may not) receive a quality prospect with an average MLB ready position player to go along in the deal.  And while one never knows how good or bad a prospect will become, there is a chance that the player(s) could help the Pirates in other areas of need.  Let\'s face it, the Pirates have a slew of pitching prospects and this is the one are they can actually afford to give up.  They need position players...mainly at third, first and left field (depending on where you want to put Craig Wilson)  Giving away Kris Benson is not like giving away Roger Clemens in his prime either.  Benson has been an average pitcher at best who seems to have been injured moreso than healthy during his breif MLB career.  Benson is a .500 pitcher who\'s ERA hovers near the 5.00 mark, which is not good.  So even though he has pitched extremely well for the Pirates of late, there is no way he is worth his current salary of over 5 million per year.  That salary is sure to increase due to the stupid baseball economy and the Pirates just can\'t afford to pay a guy who might end up getting better in 2 or 3 more years while paying tons of money for him during his \"progressions\".  There is also the notion around the city of Pittsburgh that the Pirates have an outside chance of making the wild card this year, and should hold onto him for the stretch run even if they were to lose him in the offseason.  This might be the most laughable idea I\'ve heard in a long time.  The Pirates are not going to the playoffs.  They may be 6 or 7 games back from a wild card berth, but they also have half of the national league to jump over....it ain\'t gonna happen!  If the Pirates hold on to Benson, they are sure to lose him.  And if they trade him, its a roll of the dice as to what they may get in return.  I don\'t know about you, but I\'ll take my chances rolling the dice with the ladder.   '),(8165,4075064,'30,July,2004','2004-07-30 00:00:00','    Mostly this blog will contain sports opinions and political analyis.    Before the day is done an analysis of the John Kerry speech in Boston last night will be up on my site for your viewing pleasure.      Get ready for the real super genius of our time to  speak his mind!   '),(8166,4075064,'17,August,2004','2004-08-17 00:00:00','    My sweetheart Ann Coulter was on TV last night blaming Democrats for the existence of all terrorists, hurricanes, tornadoes, hell......you know the drill.  I love this lady. I love how she might be the dumbest person on the face of this earth and yet she still talks as if she knows stuff. How awesome is that?  How cool is it that whenever she was asked a question about an untrue Bush ad seen on TV that bashes Kerry, she dodges the question like no other and says, \"you want to talk about ads, lets talk about Kerry\'s ads\". I absolutely love that stuff. I mean after all, she was asked a question and ever so sneakily, she didn\'t answer it.  The thing that was so cool about what I saw on, surprise surprise, Fox News was that Alan Colmes let Coulter dodge the question and stumbled onto his next topic. Only Colmes could do this so well.  How about following up and making her answer the damn question you freaking moron? Let\'s see....he could have said, \"No, I asked you a question about the Bush ad, lets continue with that\" or \"eat shit and die Ann, I asked you a question bio eeitch...answer it!\". But I suppose those questions were so complex and hard to come up with during an interview that it never popped into Colmes\' miniscule brain.  So Coulter refused to answer the question that no conservative wanted to anyways, and she got away with it. Can you blame her? If she is allowed to dodge questions so easily, why not do it? There is nothing wrong with Coulter being a complete hypocrite and an asswad...oh wait.  Now that I think about it, I love Fox News. You can\'t get this type of comedy elsewhere. Why watch CNN and get \"actual\" news and stuff? I can always read the Pittsburgh Post Gazette for that crap. I want lying and hypocrisy when I watch television. Because at night when I watch Fox News there is nothing I like to do more than to bash my head against the wall in frustration.  Yes, ladies and gents, my two heroes were sitting next to each other last night...Sean Hannity and Ann Coulter. We should all have watched and enjoyed it, because it was a special moment on TV.  Ann Coulter....EAT A BOWL!   '),(8167,4075064,'16,August,2004','2004-08-16 00:00:00','    Tommy Maddox is the starter for the Pittsburgh Steelers this year.  There is nothing, other than injury, that can happen in the preseason to give the nod to either the Steelers first round draft pick or Brian St. Pierre.  However, the future sure looked bright on Saturday night in Detroit.  When Ben Roethlisberger (a name i will get sick of typing shortly) entered the game in the 2nd quarter Saturday, people were looking for good things from the youngter.  Instead, they got to see great things.  Big Ben (see?) showed the poise of a veteran QB with the ability to elude the rush while still finding open receivers down field.  I loved what I saw from him.  Not only does he have a cannon for an arm, but it is also very accurate.  The trait that I believe will separate him from other QB\'s in the years to come is his scrambling ability.  No, he doesn\'t have McNabb speed, but a QB in this league doesn\'t need to be that fast.  A QB must be able to KNOW when to run and be able to make the quick moves in the pocket to buy time.  Roethlisberger does this extremely well.  Also, he has the uncanny ability to throw on the run with amazing accuracy, something we all saw from his days at Miami (OH).   Now, I\'m not going to jump to conclusions on whether or not he is a hall of fame QB after one preseason game, but it sure is nice to see that the future quarterback of the Pittsburgh Steelers looked tremendous in his debut.    Game notes:  Ben Roethlisberger went 8-13 for 84 yards and a touchdown pass to Antwaan Randel El.  Brian St. Pierre finished the game at QB.  Tommy Maddox also went 4 of 5 for 52 yards and looked very sharp.   '),(8168,4075064,'11,August,2004','2004-08-11 00:00:00','    Like the title?  I was watching Hannity and Colmes on The Fox News Channel last night, a station that pretends to be fair and balanced (when you have to claim to be fair and balanced....something doesn\'t add up), and I saw something that made me want to sock Sean Hannity repeatedly until he shuts the hell up.  I\'m all for having an objective view from both parties on a subject, but Sean Hannity is a perfect example of how not to be \"fair and balanced\" as he states they are every friggin\' day.  They had a Fox News Exclusive (OH MY!) about how this idiot John O\'Neill wrote a book on John Kerry\'s lies in Vietnam and how he claims to have 60 Veterans claiming he lied as well. His whole purpose: To not have John Kerry elected.  We\'ll get back to this...but what I want to focus on is how Hannity ran the interview. Keep in mind that they had a Kerry advisor on the show later to get a \"fair and balanced\" view from the other side as well. So Hannity starts lobbing up softballs while this O\'Neill jagoff keeps knocking them out of the park. First off, this O\'Neill moron is a Republican who has hated on Kerry for 30 years (so already his claims are partisan as hell and thus very questionable). Hannity makes the claim that only 9 vets support Kerry, and that 60 vets (as O\'Neill so eloquently puts it) think otherwise. So Hannity concludes that you have a 60 vet vs. 9 vet comparison and hence anyone in their right mind can conclude that Kerry is a liar.  One problem assbag...the 9 vets on Kerry\'s side have publicly stated how Kerry is aVietnam hero and how the numerous awards he has won reflect such. O\'Neill\'s buddies have never gone public. And in fact, Hannity said the book \"seems very interesting\" that O\'Neill wrote. So the dick hasn\'t even read the book (which probably contains more twists than a damn roller coaster) written by O\'Neill, and he is already coming to conclusions about Kerry. Typical Sean Hannity. He takes a quote about something and twists it so far out of proportion to get his point across. Isn\'t that illegal or something? The stuff he quotes in his show are stuff that he hasn\'t even read fully yet or understands the meaning of, but he just gets so excited when he THINKS he read something important that he uses it and completely takes it out of context. Thats something that a 10 year old does. How does this ass get away with that crap while Colmes sits there with his thumb up his ass (nice choice of a liberal counterpart by the way...I wonder who picked that gem, was it maybe Sean Hannity? Or was it the producer of Fox News who got rid of all the damn liberals once he took over the channel?)  Sorry about that tangent I went off on there...I just hate that guy so much.  Anyways, so when they go to a commerical break Hannity and Colmes brings in a guy from the Kerry squad to give their side. When they come back...Hannity says something like this..  \"The point I want to make here is that none of us sitting at this table right now know what happened over there...the only thing we can do is listen to those who were over there (like O\'Neill)...so thanks for coming to our show though, talk to you later\".  Honestly, because this guy they chose to come on the show was never in Vietnam, he was never given a chance to talk because \"you don\'t know because you weren\'t over there\". What the hell? Sean Hannity, not to mention, continually interrupted this guy from the Kerry side and he was never able to get his point across while Mr. O\'Neill was given all the time in the world to bash the hell out of Kerry. Sound fair and balanced to you? I hate Sean Hannity.  On to my other point...  Do you think the conservatives will ever quit this shit on Kerry\'s Vietnam record?  How hypocritical is it when their President is a damn draft dodger, yet they have the balls to criticize a guy who served for his country.  They claim he won the purple heart because he accidently hurt himself in battle, not from incoming fire.  Whatever...they sound like 6 year old girls saying that he wasn\'t such a cool hero over there after all.  How low will these dicks stoop?  And I wonder why they put so much emphasis on this?  Could it be because they have nothing on this guy and they want to just start a leak on something and drag it on throughout the entire election?    I bet Kerry didn\'t rescue anyone either, I bet he just drove his boat through the water and accidently bumped into a fallen soldier...and I bet this guy got into the boat himself, without Kerry even seeing.  Which doesn\'t make Kerry a hero at all.  And I bet Kerry wasn\'t even on the boat.  In fact, was he even in Vietnam?  Is this guy running for office?  Is he even American?  He hates America...typical liberal!  I hate John Kerry                                                                                                        Anyone from Fox News                                                                               Have I mentioned that I hate Sean Hannity?   '),(8169,4075064,'06,August,2004','2004-08-06 00:00:00','    I realize the millions of those who read my posts will be upset to hear that I will be going away this weekend and that my award-winning site will not have any updates until Monday.  Hopefully, you folks can find other sites that can hold you over until that time.  may i suggest:  www.espn.com www.pittsburghsteelers.com www.stacked.patrickstack.com  Peace out!   '),(8170,4075064,'04,August,2004','2004-08-04 00:00:00','    urlLink annabenson.net   '),(8171,4075064,'02,August,2004','2004-08-02 00:00:00','    Baltimore  The reigning AFC North Champion Baltimore Ravens are in prime position to win the title again. Ray Lewis and the intimidating Raven defense is back again to torment QB\'s. While the offense hasn\'t seen any significant upgrades, Jamaal Lewis is one of the best RB\'s in the league and the key to an effective offensive attack is a stellar ground game. The Ravens will have this and should win the division title again this year. Oh, and if Kyle Boller happens to fall on his face, good ol\' Kordell Stewart will take over as their QB. I hope Kordell returns to Pittsburgh as a starter for the Ravens.  Prediction: 10-6   Pittsburgh  There is no reason to believe the Steelers will finish second in this division, but I also can\'t see the Steelers finishing 6-10 like they did a year ago. The defense should be much improved via the return of defensive coordinator Dick Lebeau. Lebeau is sure to send the blitz packages with frequency due to their lack of ability to cover WR\'s with the weak secondary they still have. If they don\'t blitz effectively, they will not succeed. But Lebeau will have them ready and seems up for the challenge. Tommy Maddox needs to cut down on the INT\'s (didn\'t we hear this last year) because with a suspect defense and questionable ground attack, the Steelers can ill-afford turnovers this season. We saw what happened with that last year. First round phenom Ben Roethlisberger won\'t be ready this year so his impact won\'t be felt until another season or two. The offensive line will be better this season but with still so any questions to answer, a Wild Card berth is the best they can do this year, and that still may be stretching it. Look for a tough fight between The Steelers and Bengals for the second place finish here.  Prediction: 9-7   Cincinnati  The one main concern this year for the Bengals is whether or not Carson Palmer can step in and be the QB they drafted first overall two years ago. But on top of that Jon Kitna had a decent season last year and his demotion raises questions as to whether or not it was the right thing to do. After all, Cincinnati fell one game short of making the postseason last year for the first time in forever. Marvin Lewis\'s defense will be good enough to win some games on their own, but how many? Rudi Johnson takes over full responsibilities at RB because of Corey Dillon\'s departure to New England. We\'ll see how good he is now. I still think this team is one year away from great things, so a tie with the Steelers for second place seems like the most logical position for them this year with the tiebreaker going to Pittsburgh.  Prediction: 9-7   Cleveland  Cleveland fans can expect a long season this year. Sure, they signed Jeff Garcia at QB. But it doesn\'t mean anything if you have no offensive line to protect him or any type of a running game to speak of, so it\'s tough to predict the impact he will have on this team. As of today, first round pick Kellen Winslow Jr. was still unsigned, and if he doesn\'t get to camp soon, the Browns won\'t have their only major threat in the passing game. Butch Davis seems too into himself and his philosophies that don\'t work in the NFL and until he changes (or he is fired), the Browns will continue to lose. The defensive secondary is OK, but the defensive line couldn\'t contain any running back last year, and that\'s never a good thing. I could be way off on this one, but they\'ll be shooting for a high pick next year.  Prediction: 5-11      Overall this division didn\'t get much better than last year, which leaves the door open for a mediocre team to perhaps slip into a Wild Card spot. But still look for this division to only get one team into the playoffs like last year (Baltimore).   '),(8172,4075064,'02,August,2004','2004-08-02 00:00:00','    The Rooney\'s and their organization have always extended their coaches\' contracts with two years left on the previous deal. It\'s been like that since Chuck Noll took the helm in the late 60s. It\'s a way of showing their trust with the coaching staff of the Pittsburgh Steelers. And while when they followed suit this past week by extending Coach Bill Cowher\'s contract through the year 2007, this time I feel they made a mistake.  With a team coming off of a 6-10 season and missing the playoffs for the 4th time in 6 seasons, isn\'t it a bit premature to lengthen someone\'s contract? The city went ballistic when the Steelers renewed Kordell Stewart\'s contract when no one was sure how good or bad of a QB he was (and we all know how that turned out).  Why should Bill Cowher get a free pass on this one? His motivational skills which he used so well in the early 90s seem to be getting old and his antics on the sidelines just seem idiotic anymore. While one can\'t dismiss the fact that he had this team in the AFC Championship game a few short years ago, one also cannot turn their head away from the horrid season they had last year. In short, the negatives have outweighed the positives the last six years, and in the NFL (which is a \'what have you done for me lately\' league) that shouldn\'t be tolerated. Bill Cowher deserves this year to prove that he still has it as a coach in this league, but he doesn\'t deserve to have his job solidified after the way they played last year. He still had two years left on his previous deal. What was the rush this time?  A 6-10 season should never be rewarded and his contract should not have been renewed, even if that means going against the policies the Rooney\'s have used the past 40 years.   '),(8173,1902121,'19,June,2004','2004-06-19 00:00:00','   Man, the people at Old Navy really need to start beating the sweat market laborerers that make their clothes a little bit harder. The quality of clothing there is such shit.   Ahhh, that\'s right ladies and gentlemen I\'m back. I had to delete the old blog for reasons I won\'t get into here. I was debating on whether to start another one or not, thinking that maybe my blogging days were behind me. But I couldn\'t stay away. That, and apparently my friends (my loyal fans) really enjoyed the voyeuristic glimpse inside my brain. Scumbags.   So here\'s the new and improved blog. Henceforth I shall be known on the internet as \"Nuggets\". I\'ll ask that if you want to link to me you use my pen name Nuggets or whatever name you want, as long as it\'s not something people can google and find me with. Ok, that\'s all for now. '),(8174,1902121,'28,July,2004','2004-07-28 00:00:00','   That fucking joke. Obviously made by some fucking asshole failure who \"quit\" smoking two weeks ago, but just came back from his mid-morning cigarette break. So yeah, in case you couldnâ€™t tell by my sunny disposition, Iâ€™m trying to quit. I am down to only smoking when I drink â€“ which is proving to be the hardest part to give up. I went and got drinks last night and didnâ€™t light up. Which Iâ€™m proud of. But at the same time, Iâ€™m sitting here in my assy cube completely stressed out and craving one like mad. I was hoping that by making this post itâ€™d help with the frustration. But like those stupid anti-smoking commercials they show on TV, itâ€™s only making me want to go smoke one that much more. HELP!! '),(8175,1902121,'27,July,2004','2004-07-27 00:00:00','   Just wanted to congratulate my favorite local bar,  urlLink the Tir Na Nog , for being voted the best neighborhood bar in Somerville by the Improper Bostonian. I love you guys way more than I should. See you tonight.     Thanks to my friend JBar for pointing out to me that my local bar won such an award. I guess I drink too much to ever notice that kind of thing myself. '),(8176,1902121,'26,July,2004','2004-07-26 00:00:00','   You walk with your signs Hatred for both candidates Why do you not bathe? '),(8177,1902121,'24,July,2004','2004-07-24 00:00:00','   Iâ€™ve had so many people tell me that Iâ€™m too pessimistic: friends, boyfriends, ex-boyfriends, family members, strangers, old people, the list goes on. But recently I realized that Iâ€™m not pessimistic â€“ that I, in fact, am the definition of an optimistâ€¦ provided you redefine what it means to be an optimist slightly.   Case in point: my friend Marso and I were recently discussing a news story about a man who went into a Romanian doctorâ€™s office for some routine testicle surgery and came out with his penis sliced into three pieces ( urlLink I\'m not kidding ). Marso, being a man, was horrified. I quickly pointed out that this might actually turn out good for the man. Perhaps he will have his penis sewn back together (and on), and heâ€™ll begin his new career as a porn star, a la John Wayne Bobbit (Optimistic Viewpoint No. 1). Marso was unconvinced. He felt that sewing on a penis that was still whole was conceivable, but he doubted that reattaching one that had been cut into three pieces would be as easy. Again my silver-lining, can-do attitude took over and I pointed out that it wasnâ€™t like the manâ€™s penis had gone through a paper shredder (Optimistic Viewpoint No. 2).   So you see? I quite obviously heard about the manâ€™s situation, found a way it could be good, and also pointed out how it couldâ€™ve been worse, thus putting a positive spin on an a negative situation. I\'m practically like sunshine I\'m so god damned optimisitc. Put that in your pipe and smoke it ex-boyfriends. '),(8178,1902121,'22,July,2004','2004-07-22 00:00:00','   My friend recently gave me a book for my birthday. Yes, I know my birthday is in April, but this friend has never really been good at getting things done on time. But I digress...  The book is  Wigfield: The Can-Do Town That Just May Not  by Amy Sedaris, Paul Dinello, and Stephen Colbert and let me just say, it is hilarious. For those of you that don\'t know who the authors are, they are comedy writers. And they are hilarious.  The story is about Russell Hokes who decides to leave his job painting highway lines shortly after he is fired. Despite being unqualified and even more uneducated, he becomes a writer. His first assignment is to write a novel about small town America. That is how he happens upon Wigfield. Wigfield is less a town and more a seedy stretch of road off the highway filled with strip clubs and used auto parts shops. In looking for the salt of the earth for the subjects of his novel, Russell finds mostly low-lifes and morons...the majority of whom live in their cars and motor homes.  But Wigfield isn\'t just a town with strippers Hokes can ogle, it\'s also a town that is in danger of being flooded should the State Representative succeed in his plans to tear down the nearby Bulkwaller Dam. It\'s also a town that might only be posing as one simply so its \"residents\" can collect federal disaster relief. Hokes hopes that his book will help perserve the little slice of Americana pie that is Wigfield. But more importantly, he hopes that his novel will prove to be the cash cow that will launch him into a life of easy living and easy women.  The book is filled with the most inane metaphors and I often had to re-read sentences to figure out what they were trying to say. It\'s so poorly written that it is pure comic genius. This is not some stirring piece of literature and if you\'re looking for an intelligent, thought provoking book - then look further. However, if you want to sit in an airport bar drinking your bloody mary and looking insane as you laugh out loud, then  Wigfield  is for you. '),(8179,1902121,'20,July,2004','2004-07-20 00:00:00','   I had a meeting with a client this morning. I was sort of nervous because we were meeting so that the client could see the treatment system that we installed on his property. Itâ€™s my first treatment system that Iâ€™ve designed and seen installed so itâ€™s kind of my baby. And since I doubt my own intelligence, I kind of feel like I must have done something wrong somewhere along the way and I was scared the client was going to call me on it (whatever the â€œitâ€ was).  &nbsp;  ANYWAY, I get out there early to set up, make sure the system is operating properly, etc. He arrives and I show him the components of the system. He asks some questions and I actually know the answers. I show him all the neat things the system can do. The meeting ends and he seems impressed. I get into my car feeling confident, intelligent, and all-around fucking fantastic. I get into my car and grab my seatbelt, reveling in my own powers of awesome. As Iâ€™m putting on my seatbelt I look down at my shirt.  &nbsp; My shirt, with the letters sewed on - my shirt, with the extra piece of fabric behind those letters so that the stitching has something to hold on to â€“ my shirt, which I had put on all by myselfâ€¦was on inside out. Inside out. The entire god damned time. '),(8180,1902121,'17,July,2004','2004-07-17 00:00:00','   My boyfriend and I have been dating for two years. About a year ago, we started talking about marriage. But that\'s it...just talking. He says he wants to marry me, but when I quiz him about a time frame, he freezes, telling me to \"believe in him and to trust him.\" How long is long enough? Do I resign myself to wait and wait? I really want to marry him, he knows this, but I am growing frustrated and resentful. What should I do?   &nbsp;  Well, I think you need to first ask yourself the question: \"Why do I really want to marry this guy?\" Generally, the women I know that are so eager to get married feel that way because of one of four reasons: 1) theyâ€™re fat, 2) theyâ€™re old, 3) theyâ€™re ugly, or 4) theyâ€™re insane.&nbsp; Now, if your reasoning is due to options 1, 2, or 3 (or perhaps a combination of all three â€“ yikes!), well you should probably put down your recent issue of  Bridal  magazine, stop eating so many&nbsp;hohos, and pull the brakes on the old marry town express and just be grateful that you have someone to love you at all.  &nbsp;  Now, if itâ€™s option 4 (and a great many women fall into this category) you should probably just continue the endless nagging to get married because, like most women, youâ€™re just too insane to fix. It\'s ok though. I\'m sure all good marriages begin with nagging and pressure to make a life decision before one person is really ready. Just for kicks, make sure to ramp up the pressure â€“ threaten to leave, accuse him of infidelity, compare him to the boyfriends and husbands of your friends â€“ men generally respond well to this kind of behavior. Just think, soon you could be  **married** &nbsp;creating a quiet, tranquil, loving home just like Martha Stewart would do - if&nbsp;she weren\'t in jail...or a total fraud.&nbsp; '),(8181,1902121,'15,July,2004','2004-07-15 00:00:00','   I have to be honest. I really wish, more often than not, that 95% of the people I interact with on any given day would just die. Right in front of me. Without any reason or warning. Just grab their own throats, make that gaspy sound that happens when you lose your breath, and just stop fucking breathing.  Love, Nuggets '),(8182,1902121,'12,July,2004','2004-07-12 00:00:00','    '),(8183,1902121,'11,July,2004','2004-07-11 00:00:00','   Dear Wippitronic, My girlfriendâ€™s birthday is coming up and I donâ€™t know what to give her. I want her to never forget the evening. Do you have any ideas?   Well personally, I think a gift of cash is the most caring, intimate thing that a man can give to the girl he loves. But the presentation of the gift is the most important. Maybe take her out on the night of her birthday and get her really drunk. Go back to her place, have ridiculously loud and passionate sex, then leave your little cash bundle of love on the bedside table and sneak out early in the morning, unnoticed by her. Sheâ€™ll wake up feeling special - a truly important and significant person in your life. I can guarantee that sheâ€™ll never let you forget that evening. '),(8184,1902121,'12,August,2004','2004-08-12 00:00:00','   I finally got a new computer. I\'ve vowed that my new machine will be used solely for the internet and its associated dowloading of porn. But does anyone know how to erase all the shit that\'s on my old computer? The only way I can think of is to submerse it in water. '),(8185,1902121,'10,August,2004','2004-08-10 00:00:00','   When I was in college I worked in a Mexican restaurant. For two long years I worked there. I was a good waitress in that I got the job done, but the part where you chat with the guests â€“ you know, the personable part â€“ I was pretty bad at. So anyway, one day I was in my usual â€œI hate my jobâ€ mood and this couple came in. For whatever reason they asked me what kind of cheese Chihuahua cheese was. With my best â€œNo duhâ€ attitude I answered, â€œItâ€™s cheese made from the milk of a lactating Chihuahua,â€ and walked away. Well, I guess they didnâ€™t think that was funny or maybe they just remembered they werenâ€™t hungry, but when I came back to their table they were gone. Just like that.  So yesterday I was eating my Mexican food lunch that was covered with Chihuahua cheese and I had a flashback to this little story and I thought, â€œDamn, it is kind of gross to think of eating something made from the breast milk of a tiny little dog.â€ But whatever, I still ate it.  My point is this: I need to get back into the service industry - someone with people skills like mine simply should not let them go to waste. '),(8186,1902121,'10,August,2004','2004-08-10 00:00:00','   I have been reluctant to blog about this subject, but I just canâ€™t hold back any longer. My cat is so god damned, fucking cute. Honestly, I love her. Donâ€™t get all jealous. It has taken us a while to get to where we are. When I first got her I was all lovey dovey because she was new. Then after a while the newness wore off and I was just kind of complacent with her. Then I began to downright hate her because she started doing things like â€œpukingâ€ and â€œpissing on my bedâ€. Then she stopped doing those things as often and we began to realize how much we had in common - like our love for sleeping, doritos, and bacon. Then I also realized that I was in total control and that she relied on me for everything and that I could fuck with her as much as I wanted and there was nothing she could do - whoever had her before me had her claws removed. So now, five years later, I love her so much that itâ€™s kind of disturbing - even for me. Anyway, I just thought you guys should know. Me \'n\' Fatty 4-eva...or at least until she dies...or I have to make the decision to kill her. Yeah, 4-eva until then. '),(8187,1902121,'10,August,2004','2004-08-10 00:00:00','   My friend K over at the Waahmbulance  urlLink is such a dick. '),(8188,1902121,'06,August,2004','2004-08-06 00:00:00','   Has anyone else ever seen the show  urlLink  Miracle Pets  ? Oh my god, it is awesome. Itâ€™s on PAX, I think - the pseudo jesusy channel - the one that shows reruns of  Touched by an Angel  and other equally ridiculous touching dramas. I donâ€™t know. I think the premise is something like these people are really big losers and a whole lot of shit is going bad in their lives and then they get a pet, or they remember they have a pet, and suddenly everything is goldy sunlight again.   I watched this one episode recently where this family had some sort of giant pet turtle â€“ although it didnâ€™t live in the house or anything. I think that the turtle just happened into their yard and because their yard was really big, and because turtles move about a mile a year, they claimed it as their pet. Well I guess one day they realized that their â€œpet turtleâ€ had finally crossed their giant lawn and was now gone. And then, for whatever reason, there was a lot of snow. And then the dad got brain cancerâ€¦BRAIN CANCERâ€¦and was having trouble getting better. And then the turtle came back and suddenly the guy was healthy again. And then the narrator â€“ the dad from Growing Pains â€“ comes on and says that the turtle was the reason dad got better.  Holy shit.  Miracle Pets.  I love this show. '),(8189,1902121,'05,August,2004','2004-08-05 00:00:00','   I don\'t think there is any possible way that in the history of all human kind there has existed another person on earth that has craved a margarita more than I am craving one right now.  F work. '),(8190,1902121,'03,August,2004','2004-08-03 00:00:00','   But my friend Katie is incredible. Check out  urlLink her brotherâ€™s site  where you can download a sampling of her talent.  It Hurt So Bad  is one of my favorites to hear her perform and the studio recording is just as good...just FYI. '),(8191,1902121,'03,August,2004','2004-08-03 00:00:00','   I went to the dentist this morning. It was just a routine check-up so the dental hygienist was the one performing the work. As Iâ€™m sitting there, her hands all up in my mouth, spit-sucking machine sucking intently, the dental hygienist looks me directly in the eyes, mid-tooth scrape, and asks, â€œWhatâ€™s the difference between tofu and sushi?â€   Even the spit-sucker stopped for a second at the sheer absurdity and randomness of the question. And coming from the other room you could hear the laughter of the dentist and the other dental hygienist â€“ laughing because they knew the question was dumb â€“ laughing because they knew her hands were in my mouth and I was scared â€“ laughing because this dentistâ€™s office obviously hits the nitrous tank a little too hard after hours.  And you know, as if that wasnâ€™t enoughâ€¦after I made it through the dental hygienist portion of the appointment (amazingly, unscathed), the dentist came in to give his two cents on my oral hygiene. He tells me my teeth look impeccable and to keep up the good work. Then, before I can get out of the chair, he places his hand gently on my shoulder and  kisses me on the forehead!   At this point Iâ€™m thinking that maybe Iâ€™m on some hidden camera show or some sort of reality TV show where they test how much weird shit one person can possibly take. But nope. I get out of the chair, walk into the waiting area, and the receptionist says, â€œOK, so weâ€™ll see you in six months!â€  The rational side of me, the side that doesnâ€™t like idiots in positions where they can inflict pain, the side that doesnâ€™t like to be kissed on the forehead by random fucking strangers, that\'s the side that knew I shouldâ€™ve just walked out and never, ever set foot in that place again. But the other side, the side that loves a good story said, â€œYup. See ya in February,â€ and walked out the door. '),(8192,1902121,'02,August,2004','2004-08-02 00:00:00','   Iâ€™m not some peace-loving, â€œwhy canâ€™t everyone get alongâ€-asking, patchouli-stink-leaving, hippy girl who thinks that the solution to world peace is getting everyone to love each other. In fact, the word I use most often when describing people is suck. But know what? My misanthropic tendencies donâ€™t stop me from having manners. I usually say please, I always say thank you, and if someone takes the time to smile and say hello to me, I will almost always return the greeting.   I was having a discussion with a friend yesterday about why I feel so isolated. In part, he placed the blame on the city I live in. I grew up in Boston so I know how it is. I know the people here are atypically guarded. But I decided to see just how bad it really was. I set out on a mission yesterday. I walked from Union Square in Somerville to Central Square in Cambridge and then up and down Mass Ave â€“ round trip I probably walked about 5 miles. During the entire walk I tried to make eye contact with people. I had a smile on my face the whole time â€“ trying to get the same out of at least one person. At best, I got a handful of half smiles, but most people just turned their heads in order to avoid the eye contact.   So what gives Cambridge? Is it really that hard to return a smile? It doesnâ€™t take much energy. Iâ€™m not looking to start a conversation. No time is wasted. You have nothing to lose. When I was in Honolulu I returned a smile from a stranger on the street and several hours later we were skinny dipping on Waikiki Beach. It was one of the most impulsive, fantastic nights of my life (and no, there was no sex involved), and it was all initiated with a smile.   I know this is a new concept up here so Iâ€™m going to give the people of Cambridge some slack. But Iâ€™m going to keep doing this walk until I get a better response. Iâ€™m going to make this city more polite if it kills me. So if you see some asshole girl with a giant grin on her face staring at you â€“ donâ€™t be afraid. Just fucking smile back. I swear it wonâ€™t kill you. Who knows? You might even enjoy it. '),(8193,1987893,'25,September,2003','2003-09-25 00:00:00','   First attempt at blogging, not sure if this will ever see the light of day. I am hoping to join the world of knitting blogs and keep records of my knitting projects, plus occasional other topics. Can\'t wait to start uploading pictures of my work!  Here\'s a photo(and link to full size photo) of the Mexicali Baby Ole sweater I knit in July (from Knitters 71)    urlLink '),(8194,1987893,'28,October,2003','2003-10-28 00:00:00','   work actively  in progress:   - Socks for him in Wildfoote \"Desert Grass\".  1.75 socks completed! - Cotton/Wool chennille vest - needs zipper basted then sewn in.  Had the most frustrating time this weekend with \"Better Mousetrap\" socks for him.  I had started a pair, oh, a year ago.  Thanks to the strip construction and poor gauge calculation, the first one came out 2 inches too long in the toe.   So I started again.. about a year ago.  And that sock was coming out too small in width.  So I gave up for a while.  This past weekend I picked up the 2nd attempt socks (khaki with grape stripes), added an extra row to the width, sewed them up and..  it\'s 1 inch too long.  OK, fine.  I think I remember adding some extra stitches to make sock long enough for his big feet.. so that was a mistake.  So I went back to the pattern & made a reverse sock (grape with khaki stripes) sticking exactly to the pattern.  Only I didn\'t check gauge and I was using the wrong size needles - 5\'s when I should have used 4\'s - and the damn thing came out EVEN LONGER than the first one and was already too wide when it was 3/4ths finished.  I\'m furious with the socks and myself for not measuring.  Luckily they are small and I can rip them out & start again if I want to, but I\'m more thinking what functional hats this Plymouth Encore wool/nylon blend would make.     So that\'s the end of mousetrap socks for me.  Such a shame, because the first two pairs I made came out beautifully.  I\'ve worn holes in my own pair 3 times already (note to self - straight merino makes soft socks but they won\'t last!) and the pair I gave away was also quite successful.  It\'s just socks for him that are CURSED.  For a corrective emotional experience, I decided to go back to regular old socks.  So I picked up another set of not-quite-finished socks for him and got back to work.  It\'s been a while since I made socks for grown ups.  It\'s so fun to try the sock on in progress with all the needles poking everywhere.  I\'m pretty sure I\'ll finish these socks today or tomorrow, woo hoo!  Maybe this will lift the curse. '),(8195,1987893,'23,October,2003','2003-10-23 00:00:00','   Huzzah!  Granny\'s shawl is off the needles.  Started on September 17th, finished October 22nd.  Unblocked, it measures 22\" x 54\".  I finished late last night and then wove in ends today.   And promptly wore it to the theater to see Sondheim\'s \"A Little Night Music.\"  It\'s going to be a little tough to part with this lovely object.  We took some photos tonight that I\'ll post in the AM.     Update 12-May-2004:  ... here is the photo, finally!   urlLink      Needing something quick with no thinking involved for the rest of the day, I picked up an unfinished Better Mousetrap sock from last year and worked on adding 1 more ridge to each side to make sure it will wrap around his feet.  The Plymouth Encore worsted weight yarn feels enourmous and rough in my hands after 5 weeks of luscious Kid Silk Haze.  That mohair/silk is so airy and soft.  I\'m even having trouble getting used the feel Heaven, which is mohair without the silk.   Notes from the past - Sat Oct 20   I was swatching Heaven over the weekend (again, picture tomorrow) and I am pretty unhappy with the color choice I made.  It was a last minute purchase - saw a sample shawl, needed to knit it, grabbed a color I thought was mostly purples, and didn\'t discover until this weekend that it is really a lot more pink than I thought.  Both pale pink and very bright fuscia pink.  Plus turquoise highlights.    Update 12-May-2004: Here\'s a photo of the swatch of nasty pink Heaven    urlLink      I know there will turn out to be someone in the universe who wants this color, but it isn\'t me.  And I\'m pretty sure it isn\'t Mom, who was going to get the completed shawl.  And the pisser of it - this is a $60 oops.   And it\'s already been through the ball winder, so I can\'t really take it back.    Since I\'m stuck with it... I did a swatch to test the colors using the shawl pattern from the Purl book from Sally Melville.  I like the stitch - 4 rows garter with small needles, then 2 rows stockinette with larger needles- easy to make and interesting to look at.  I started using 10\'s & 6\'s, like she suggestes, but switched to 8\'s and 4\'s and finally to 7\'s and 3\'s.  The 3\'s really felt right for the yarn, so that\'s probably what I\'ll stick with for the next swatch.  I\'m not totally keen on the way it looks with this variegated yarn - I think the runs of each color are too long for stockinette, even when it\'s just 2 rows at a time.  I\'m going to give it one more shot in a garter stitch - the free pattern inside the ball band - and if that doesn\'t make me like the color, I\'ll have to find someone who likes these colors somehow.  Maybe a shop would let me post a \"take this yarn off my hands\" note. '),(8196,1987893,'21,October,2003','2003-10-21 00:00:00','   Pewter Lace Shawl from Interweave Knits is coming off the needles tonight - I\'ve cast off half of one short edge & have around 650 stitches left to cast off.  Then it\'s done.  So tonight\'s task is planning.  What next, what next?  Time for inventory.  Half-finished stuff - red & white Dale of Norway baby sweater needs a hem and (ack) cut armholes (2 - 3 hrs) - pink basketweave baby sweater needs buttons & a mailing label for South Africa (30 min.) - chenille vest needs zipper basted, then sewn in. (3 hrs, if you count time for messing up & ripping out) - Ragg \"Birkensocks\" need 1 foot completed (7 hrs) - Wildefoote desert socks need heel turned & foot completed (7 hrs?) - Better mousetrap socks need 1 more row each side and seams (2 hrs)  Projects planned/yarn bough in the last 2 months - Charlotte\'s Web shawl in 5 colors of Koigu (for me) - Seed stitch T-shirt in Colinette Giotto (for me) - Blue socks with pink stripes (for Siobhan) - Mexicali Baby Ole sweater in Fortissima Colori (for MM\'s baby) - 1 giant skein Heaven lace yarn, was supposed to be a shawl for Mom.  Hate the color! - Hat in Noro\'s Kureyon (for who knows) - Baby jacket in Noro\'s Kureyon (for who knows)  Languishing projects - not started. - Rowan Denim aran sweater (for him) - Celtic knot hat (for sister-in-law)  I bet there are more at home.  I\'ll have to update this list later.  For now it\'s enough to have an overview of the things I might start.  I\'m definitely feeling like I owe some of those half-done pieces some attention, especially the Dale baby sweater & the socks for him '),(8197,1987893,'28,May,2004','2004-05-28 00:00:00','   I figure the back plus the left front means I am half done with Swan.  Photos:    urlLink     urlLink     The front is now off the needles & I\'ve done 5 rows of the right front.  I\'m sort of dreading the sleeves, since my row gauge on this project is way off. I ended up knitting 22 rows for the armhole on the front.. so when it comes to sleeves, I\'ll probably have some calculating to do to figure out how wide the sleeves need to be.  Progress on whales: nil.  I knit the 4 rows of little peaky waves and a few rows beyond that.. but then figured out that I should have had 5 rows of waves and had made a couple other mistakes.  So I tinked back all the way to the end of the blue.  I\'m hoping to finish Swan this weekend & get back to whales.  Today\'s humor: I was up at 12:15 am last night, watching Aqua Teen Hunger Force.  This ad comes on:  \"Do you hit the snooze alarm in the morning?  Feel like you\'re dragging by mid-afternoon?  Need a second cup of coffee to get going?  Well...\" and I\'m thinking... GO TO BED, YOU IDIOT.  \"... you need this fancy pants matress.\".  Talk about hitting your target audience.  Yep, they\'re exhausted all right.  Yep, they\'re too dumb to go to bed.  They might just buy your matress!  In a rare moment of strength, I used the power of laughter to tear myself away from the shiny pictures.  I shut the TV off & went to bed. '),(8198,1987893,'21,May,2004','2004-05-21 00:00:00','   In my aspect as Easily Distracted Woman, I\'ve dropped all other projects and have been working exclusively on the Myosotis cardigan for Julie B.  Cast on Sunday and finished the blue bit with the whales in yesterday morning.  I showed my progress photo to a couple friends and they actually recognized the grey blobs as whales!  Huzzah!  Photos:   urlLink     urlLink     I found a web site which describes the colors associated with French color names for  urlLink silk embroidery thread  and discovered that Myosotis (which means forget-me-not) is used a color name French.  It\'s apparently a \"pale dusty blue\".  And in the magazine, that\'s the color they used.  I\'ve chosen a slightly darker blue, so the pattern name is now completely divorced from the actual object.  Oh well!  Wednesday night I stopped by  urlLink Knitting Arts  to pick up an interesting yarn to make string ties for the Florella baby set and get some advice on my 2 stalled Colinette projects.  I picked Tweedy Cotton Classic from Tahki in rasperry/pale pink/chartreuse/cream/lilac tweed.  I hope it will add just the right amount of jazz to the solid color sweater set.  All the ends are woven in and we\'re planning to be at home tonight, so maybe it will finally get washed - a week after coming off the needles.  Baby Jocelyn was due early this week, so I\'d better get a move on!  On a side note... I\'m becoming rather nervous about my punctuation.  I\'ve read about half of \" urlLink Eats, Shoots, & Leaves \" -  a surprisingly fun book about the history & foibles of punctuation.  Trouble is, some of her negative examples, like \"No dogs please,\" did not even phase me a bit.  (Punctuated like this, it\'s a statement.  And patently false, as many dogs DO please.  It\'s what they do best!)    Trouble is, it makes me think about each comma, dash, & my favorite, the elipsis.  I\'m sure it will wear off in a few months, but it\'s interesting to become aware of something you were totally taking for granted.  Like buying a new car & discovering that there are HUNDREDS of your car on the road.  I now see punctuation everywhere! '),(8199,1987893,'15,May,2004','2004-05-15 00:00:00','   Look, titles!  I\'ve been doing some remedial blog reading to get the lay of the land.  Looked a bit at Rob\'s blackdog blog, where his enthusiausm for yarn & projects must doing wonders for his sales, and Wendyknits and Bonne Marie.  Rob & Matt\'s Charlotte\'s Web color combinations made me want to haul my own Charlotte\'s Web colors out of the closet & get knitting.. but I\'ve got to work on Swan, Myosotis & Bill\'s sweater for a while.  The knitting blogs were inspiring and a bit daunting at the same time.  Queer Joe\'s knitting blog particularly made me ask myself, am I adding anything to the craft here?  And my kntting group at work is sort of falling apart... bleah.  Discovered yesterday that my row gauge on the Swan is waaaay off - like, I should have 16 or so inches at this point, and I\'ve got 9.  I could just keep knitting straight, but that would change the line of th garment from a nice neat trapezoid to a trapezoid with a box on top... so that will have to be ripped out entirely (I could save the last 4 rows..but why bother?) and started over.  It was only 2 days of work.  :-((  We\'re off on a birding adventure today, so I\'ll have plenty of time in the car for unknitting & reknitting.  And I\'d better get going if I want to see them birdies! '),(8200,1987893,'13,May,2004','2004-05-13 00:00:00','   Hazy but beautiful day today.  No knitting this morning, but I am working on the seam of the lilac bootie for Wendy.  Still fussing about whether to have ribbon ties or crochet chain ties for the ensemble.   Tagliatelli Swan  Got started on the \"bed jacket\" project last night.  Knit about 10 rows, ripped back 6 and re-did them with the decreases which make the a-line styling for the body.  I started using just one ball of Tagliatelli, but I was reading the yarn label last night with its stern warnings to use 2 balls at once to smooth out the variations in hand-dyed skeins.  Will add the 2nd one tonight.  Here\'s my progress so far:   urlLink     As I was knitting last night, we watched the end of Angela\'s Ashes and all of Miyazaki\'s Castle in the Sky.  Sent them both back to Netflix this morning, but I\'m feeling a little bad about having knit thru the beautiful animation of Castle in the Sky.  May have to make a new rule not to knit during anime. '),(8201,1987893,'12,May,2004','2004-05-12 00:00:00','   I\'m starting again with my not-yet-fascinating knitting blog.  New plan: start simple.  I read about 9 months of Jerry-the-old-fart\'s blog,  urlLink When Knitting was a Manly Art , and was totally fascinated.  So I\'m back on the horse.  I\'m not yet whether I have the stamina to sustain a blog, but I really want a record of my knitting life, so here goes.   Knitting  With the last stitches done this afternoon, I am almost finished with a layette called  urlLink Florella  for Wendy\'s impending baby.  Found the pattern online with some other  urlLink antique knitting patterns .  Apparently it\'s from a 1943 pattern book.  I\'ve done the whole set in light lilac Dale Baby Ull.  Remaining work is the seam along the bottom & back of one bootie and finding/making strings for 2 booties, 1 hat and the sacque.  Wendy\'s baby is due sometime this week or next.  Update 13-May-2004:  Here\'s yesterday\'s photo of the mostly finished layette.    urlLink     This pattern was very annoyingly written, with lots of K7, K2, K3tog... without much explaination of what was going on.  I had to make a lot of adjustments for a gauge of 7 sts/inch instead of 8, so I spent a long while with a calculator figuring out how to adjust the decreases & pattern motifs.  It was a combination of math and winging it, heavy on the winging it.  Overall, I\'m satisfied with the results.  There are no finished measurements given in the original pattern, but if my math is right, my version is actually the same size as the 7 sts/inch version would have been.  It\'s supposed to fit a 6 month size.  Remains to be seen if it will fit Wendy\'s baby.   Plans  I\'ve, um, been to the yarn shop rather a lot lately. But mostly with very concrete plans in mind.  Here\'s what\'s in the hopper - without counting unfinished objects:  - For Kelli, a cheer-up bed jacket while on bed rest.  Bought some Tagliatelli yarn in  urlLink Morocco  and a pattern for a short-sleeved cardigan called  urlLink Swan .  This is the first in line because I think she needs it now.  Mom says I have to modify the neckline because it\'s too droopy.  I think a little crochet will firm it up.  - For Julie, a baby outfit (hat and cardigan) by August.  Picked design #2 called  urlLink Myosotis , from Anny Blatt\'s \"Layette\", Hors-serie No. 8,  with grey whales in a blue background.  The name is weird, because as far as I can tell, Myosotis means forget-me-not.  - For Bill, a Rowan Denim gansey in the darker blue  urlLink Nashville .  This project has been simmering in the idea stage for about 6 years and he\'s getting antsy  - For me, Fog Chaser jacket from The Knitter\'s Stash with 4 different yarns from Mountain Colors, mostly color  urlLink Wilderness '),(8202,1987893,'25,June,2004','2004-06-25 00:00:00','   Last night we saw Tom Stoppard\'s  Arcadia  at  urlLink theatreworks  in Mountain View.  It\'s deep & multilayered, like all the Stoppard I\'ve seen so far, but also beautiful and emotional.  I was just weeping at the end.  Boo hoo.  Great, great text and very well performed.  ... which explains why my knitting progress is minimal.  I\'ve done scattered work in the last few days to clean up the many, many WIP/doomed projects:     Charlotte  is still lying fallow until I decide if I\'m going to knit another repeat of the pattern.  I sewed the one finished sleeve onto the  Swan  cardigan to check the fit on random models at work.  I\'ve got 6\" of 18 on the 2nd sleeve.  Finished the  turtleneck  of Wendy\'s sweater, just seams left  Crocheted 4 string ties for the  Florella  baby set.   I\'m itching to get Swan, the turtleneck & Florella finished, washed & shipped.  I need to reduce the doomed objects quotient (DOQ) so I can feel OK about moving on to the ASA socks.  :-)  Who thinks I can have them all done by Sunday night? '),(8203,1987893,'23,June,2004','2004-06-23 00:00:00','   Charlotte\'s web : cast off last night, but the finished length (unstretched) is only 22\".  I am supposed to have 24. I am thinking of un-casting-off and knitting a few more rows, maybe even another repeat of the pattern.  While I am stewing over this decision...  I have been looking into \"A Step Above\" (ASA) entrelac socks from Knitter\'s 75.  So.  In a fit of \"must have them\", I bought yarn for these socks on Sunday.  But the yarn I bought is not showing the pattern well.  I am knittng a test triangle and it\'s just not looking good.  I think the best yarn for these socks would have color run lengths of about 6 inches at a time.  I have a bunch of yarns at 20-30-50 inch run lengths and this new cotton yarn which alternates 6\" of purple with 1\" of red.  The color is also muddied by the use of a white thread to make this a tweed yarn.  So I\'m thinking about a quick run to the yarn shop... though of course Charlotte & the unfinished bedjacket, baby sweater and other baby sweater are calling.  And so we come to discipline.  I will knit with what I have tonight.  Tomorrow maybe I\'ll go to the yarn shop.  Today my friend Frank sent me a copy of the \"Person Paper on Purity in Language\" by William Satire - a satirical essay written by Douglas Hofstadter and first published in 1983.  Read it & let me know what you think:  urlLink http://angelfire.lycos.com/rebellion/personpaper '),(8204,1987893,'18,June,2004','2004-06-18 00:00:00','   On Sunday,  urlLink Janice  had a copy of the Summer 2004 Knitters #75 at S-n-B.  I really liked the patterns for using self-striping sock yarn in entrelac, especially the  urlLink Step Above  socks!  I have some glaring bright yellow & red Lorna\'s Laces sock yarn.  I\'ve never done (or much cared for) entrelac, but this is quite a cool design.  Also yesterday I bought a copy of  urlLink INKnitters , which I haven\'t even cracked yet, because of a latent desire to do shadow knitting.  I\'ve been looking for motifs other than the alien head in the Stitch-n-Bitch book.  I\'d never even heard of this magazine until yesterday.  Weird.  Progress on Charlotte was essentially negative yesterday.  Knit 2 rows in the evening, figured out I\'d repeated a row too often and had to un-knit everything I\'d done.  Boring!  But I\'m back to where I started now, so onward & upwards. '),(8205,1987893,'17,June,2004','2004-06-17 00:00:00','   10 knitters showed up last night to what we think was the first  urlLink knitting meetup  in Santa Clara County.  Several of the SJSU SNB\'ers were there and some new folks as well:  Laura, Christina, Janice, Jenn, Mackenzie, Ivy, Jerri, Jessica, Liz & me.  I remembered to grab nametags before the event, so we all got a fair shot at learning the other knitters\' names.  The Barefoot Coffee Roasters was a very comfortable place and the manager was very accommodating to the meet up group, including printing a table tent for us and finding us a more comfortable location.  Unfortunately, Wednesday is Open Mike Night and the music was a hair too loud for pleasant conversation.  We\'re investigating other venues for the next meetup.  Most everyone said they would come again.  Next chance:  July 21.  And I made some progress on Charlotte.  Just 30 rows left!  The last color is integrated - a subtle purple/lavendar/brown/ruby combination barely visible at the point of the triangle:    urlLink     Wendy told me last night she thought the \"color development\" was perfect, so maybe I\'ll get brave enough to cut the yarns & stop thinking about ripping it out.  In thanks for her kind comments, I worked on the turtleneck of her long-overdue Christmas present.  I did 2.5 inches last night, so half an inch more and that sucker is off the needles.  I\'d love to finish it this week and wash it when I\'m washing Charlotte.  And the baby Florella set.  And Rebecca stopped by last night with yarn she bought for her first project - a Beach Beanie hat from Knitters 74. She had a yummy Anny Blatt angora in orange/turquoise/raspberry/brown shades.  I think it will make a great hat.  She did the cast on and most of the first row before turning into a pumpkin.  I am *sooo* happy that she is plunging into knitting.  It\'s great to have something that draws us together regularly, hurray!  I found the blog of one of the SJSU stitch-n-bitch participants, Janice.  I totally admire her dedication, she\'s able to write daily!  Some day I\'ll be that cool.  Here\'s here stuff:  http://knitflix.blogspot.com/  She also posted an interested link to the SJSU SNB group today -  urlLink How to become an expert knitter , an online book.  I\'ve read the intro & chapter 1, so far all great advice on how to select, knit & finish a great-looking garment.  Video report: Last night we watched 3 episodes of Magic Shopping Arcade Abenobashi.  It reminds me strongly of FLCL, which we loved.  Moving the action from one version of the shopping center to another allows them to make fun of something new each episode.  So far, we\'ve mocked D&D-based video games and Giant Robot Anime.  Next episode looks to be set in China... poking fun at kung-fu movies? '),(8206,1987893,'14,June,2004','2004-06-14 00:00:00','   Wow, this has been a fun two weeks!  I taught a friend to knit on June 2nd and she\'s totally taking off - got a knitting book and taught herself a few stitch patterns based on having learned knit, purl and knit to together from me. So encouraging after the week of doomed projects.  The yahoo  urlLink Knitalong  group is doing shawls, so I pulled my 5 skeins of Koigu KPPPM and the Charlotte\'s Web pattern out of the stash and started on Friday.  I knit on our way to Half Moon Bay Friday night, in the car Saturday and at the  urlLink SJSU Stitch-n-Bitch  on Sunday.  This is a good antitidote to the doomed projects, since gauge doesn\'t matter as much and it moves very quickly in the beginning.  Progress photo:   urlLink     I met Emily, Laura, Janice and Jenny at the SNB group in Campbell on Sunday. It was really awesome to finally meet some other knitters?  some very pretty projects in progress.  Now I have to quit writing & pay attention to My Neighbor Totoro, which is new to me and looks gorgeous.  More soon, as Charlotte progreses. '),(8207,1987893,'01,June,2004','2004-06-01 00:00:00','   Existential crisis day.  Swan: I thought I was knitting with size 15 needles, but now I\'m not sure.  They might be 13\'s.  And the sleeves are coming out at a very different gauge than the body (should be 7 sts/4 in - now coming out 8.5 sts/4 in).  So the sleeves are narrow and are taking much more yarn than called for in the pattern.  Also, the gal I was making it for delivered her baby on Sunday and is no longer on bed-rest.  I\'ve got the body plus one sleeve finished.  I think I have enough yarn to finish the 2nd sleeve.  So.. should I keep going?  Myosotis: The whales are still stuck at the water line.  Some of the intarsia on the first bit looks wobbly.  Should I keep going?  Wendy\'s mohair T-neck in Tahki might not be doomed, just stalled.  Needs another 3.5 inches of neck and seams to be finished.  Maybe that\'s what I\'ll take on the road Thursday when we drive to Ashland for the weekend.  Knitting class at work today: just me and Angel, the newest knitter.  10 other knitters did not show up.  I think I made a huge mistake by not having projects & yarn picked out for them at the beginning of the classes.  I\'m not sure how to get the energy back.  Should I keep going?  Bleah.  Movies this week:  American Splendor. Underworld.  Men with Guns. '),(8208,1987893,'29,July,2004','2004-07-29 00:00:00','   This news actually belongs to Saturday, but I was not entirely cogent then.  I have recieved the nicest thank you note from a recipient of a knitted object and I just have to share.   I finally managed to box up Florella and get it ready to send right before the Uni High reunion, but didn\'t mail it until Monday, July 19th.  Already on Saturday, in my mailbox was the thank you note from Wendy, which I am transcribing here:  \"Dear Mary, We just love the beautiful outfit that you have created.  The pearly buttons were such a nice touch.  The color and multicolor tie goes great to gether.  Thank you so much.  Jocelyn is so lucky!  Best, Wendy\"  And what makes this the winning entry in the \"Best gift recipient ever\" contest? - Speed!  Gift sent 7/19, thank in hand 7/24 - Encloses a photo of adorable baby wearing outfit - Noticed the buttons! - Noticed the ties (any idea how long I agonized over tie color? how many yarn shops visited?) - Includes birth announcement card with full baby details  In general, no one ever *asks* for the baby things I send them and who the heck knows if they like them.  But this is the first time I\'ve ever gotten a thank you like this.  And the first time in like 10 years that I\'ve received a photo of a baby wearing the knitted object.   Having this show up in the mail Saturday was a big boost to my spirits.  Wendy, you totally rock! '),(8209,1987893,'26,July,2004','2004-07-26 00:00:00','   Last night I finished reknitting the sleeves of the February toddler cardigan & wove in all the ends again. When I am finally able to leave the house, I can deliver a finished object right away!  (Photo soon.  I think all will agree the new sleeves look much more normal.)  It was good to have a small, easy-to-complete project to knock off.  My little stack of \'things I did while I was recovering\' is growing.  I finished \"24 Hours\" by Margaret Mahy yesterday morning and added it to the stack.  Also watched Donnie Darko on Tuesday night and again Wednesday with the commentary.  Into the pile with it!  Next up on the video parade is Man on the Moon, which arrived from Netflix in like early June.  Everything must go!  More thank you\'s to people who have helped out while I\'m recovering:  Mika, for my Wednesday morning visit and lunch. Rebecca, Cecelia, and Wendy for company last night.  Pat and Bill for great music to recover by.  And Adam (**ADAM!**), king of the kitchen, for dinner and cleanup last night.  Kevina and the gang for a lovely lunch today and all the well-wishes.  Thank you all! '),(8210,1987893,'26,July,2004','2004-07-26 00:00:00','   Had a great  urlLink meet-up  last Wednesday night at Blue Rock Shoot in Saratoga.. gave me a chance to pick up yarn for the next baby sweater which I don\'t yet yave a name for.  Watching Janice at the meet up I got a better handle on using the circularness of the 2 circulars to keep the yarn from tangling & understand the progress of the socks better.    Surgery Thursday morning went pretty well, overnight stay in the hospital was fine by me and I came home Friday.  Since then I\'ve been lumping on the sofa -- something I\'ll probably be doing for another 2 weeks.  I\'ve knit approximately 4 rounds today and that\'s pretty good, considering how spacey I am.  I\'ve totally appreciated the support of my pit crew for the last 5 days.  People have brought dinner, come over to babysit me when Bill is out, sent flowers, called on the phone and just been great.  Thanks to Bill (you\'re the best), Mom, Mom & Dad W., John & Susan, Pat & Wendy, Jennifer and Pastor Richard.  I\'m feeling much better! '),(8211,1987893,'21,July,2004','2004-07-21 00:00:00','   With lots of tips and encouragement from  urlLink Janice  from the Campbell Stitch-n-Bitch group, I am making great progress on 2 socks, 2 circulars.  Using Ingrid\'s  urlLink Double Trouble  sock pattern, I cast on both socks last Friday in Champaign then spent the rest of my visit and the plane ride home working on the feet.   	 Saturday:&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Monday night   urlLink   &nbsp  urlLink      By Monday evening, I had turned 1 heel and Tuesday I turned the 2nd.  I\'ve knit a few rows even now and am ready to start the ribbing.  Project: Double Trouble 2 socks, 2 circs Yarn: Regia 6-fadig Crazy Colors in color  urlLink Dschungel   Needles: Size 1 circulars (Addi turbo and Susan Bates) Start Date: July 16, 2004 Notes:  The yarn was a gift from Mom about 2 years ago.  The color did *not* grab me, so I thought it was OK to sacrifice it to experimental socks.  I dig it much more now that I can see the striping in the sock  Things I learned from Janice for this project: - Use the provisional cast on to start the toes - Short row toe and heel - Use size 1 needles (or 1 size less than normally used with the yarn) to get a firmer fabric - 3X1 rib looks great - 2 socks, 2 circs is easier than it sounds '),(8212,1987893,'18,July,2004','2004-07-18 00:00:00','   Â  '),(8213,1987893,'12,July,2004','2004-07-12 00:00:00','   I just found out Jeanette and Jen are  urlLink selling  the fruits of their knitting as \"Fuzzy Flops\".  Their knit-decorated flip-flops are made with fun yarns that really jazz up your standard flop, plus good-luck charms for your feet.  I love seeing all the ways knitters make use of their talents to decorate the world. I hope they make a mint selling their work! '),(8214,1987893,'12,August,2004','2004-08-12 00:00:00','   I did it, I did it!  The following objects are ready for wash today:     Wendy\'s Reverse Turtleneck   Swan cardigan/bedjacket for Kelli ( urlLink photo )   Baby Molly\'s doll blanket/Feather &amp; Fan Washcloth   Baby Nolan\'s Early Arrivals preemie sweater  ( urlLink photo )   Sample illusion knitting versions of the NC State logo for Stacie   Phew!  I started Turtle in stinkin\' January!  And Swan has been completely done & ready for seams since end of June.  I\'m so pleased to have these things all finished up & read to go.  I\'ll post a photo after I\'ve done the actual washing.  I love to see all the knitted objects drying together.  Double Trouble socks are *not* finished, sadly.  Talking with  urlLink Janice  on Tuesday, I decided I should use a different turnaround for the cuffs.  I tried a purl turning row.. that didn\'t turn outward, it wanted to turn inward.  R-r-r-rip.  So I tried a knit turning row.  That made purl bumps right on top of the cuff.  Not cute.  So I stalled out a while until Janice came up with an idea from something she\'d knit before... change from k3p1 to p3k1 rip about half an inch or so before you want the turn down to start.  No turning row.  Just turn the cuff down after establishing the inverse of the rib parttern.  The change from rib A to rib B is hidden under the cuff.  So this morning I am going to r-r-r-rip the socks back about 25 rows and implement the new plan.  So exciting to be finishing things, working thru the stuck places.  I\'m feeling ready to tackle a couple new things or maybe even finish up some evil old things.  Here are a few items I would like to tackle next:   Things already started     Anny Blatt Myosotis cardigan, started in May 2004.  Half the  urlLink back  is done & baby is due this week - eek!   Colinette Giotto tunic for me... a doomed object having been started twice already.  I\'m worried I\'ve scrunched the yarn & it won\'t be pretty anymore, so I\'ve been putting off making this.. but I kind of want to wear it Saturday so it\'s either put up or shut up time for this project   Rowan Denim OXO gansey for my dearest ( urlLink swatch ).  40 cm of  urlLink boring stockinette  finished, now it\'s time for the patterns   Pursification of the Devine Mitred Squares     Things calling my name      Oat Couture  urlLink Crocus Cardigan  for Baby Clara (due Oct 16)   urlLink  Tilt , from knitty in Noro Silk Garden, the tuquoise, pink, purple color   Doggie sweater.  Been meaning to make one for about a year and the  urlLink Knitalong  Yahoo group is currently considering companion animal knits.  Brought hom Judith L. Schwartz\'s  urlLink Dogs in Knits  from the library and I think I have enough leftover Jaimeson\'s Soft Shetland to make an Icelandic yoked doggie sweater without a trip to the store    urlLink Entrelac Socks  from the Summer 2004 Knitters.  I\'ve got the  urlLink yarn  balled up & ready to go.    And so the debate begins..  what to knit, what to knit!  It\'s so freeing to have finished those lingering sweaters, I feel like I could start a million new projects!  Bwahahahaha! '),(8215,1987893,'09,August,2004','2004-08-09 00:00:00','   I\'m trying to clear out some nearly-finished projects before I go back to work.  First, Baby Nolan\'s preemie cardigan \"Early Arrivals\".  It is the replacement for the little green cardigan which came out to be doll-sized.  On Thursday I finished the seams and buttons on the cardigan.  I am not totally happy with the join at the upper right neck where the knitted-in button band was carried around the neck and grafted.  But since this truly is a temporary sweater for 3-5 lbs, I\'m moving on.  Photo:    urlLink      Next, the Double Trouble socks.  Friday I knit, unknit and reknit the DTs, thinking I could have them done by Sunday.  But my dreamed-of turned down cuffs were causing trouble.  So they are still on the needles.  At Sunday\'s  urlLink SnB , I watched  urlLink Jeni  doing tubular cast off to finish her Double Troubles - a lovely cotton in pinks & orang.  I was a little jealous & thought about giving up, but i\'m gonna stick with it a little longer and see if I can make the cuffs I dream of.  Finally, 2 sweaters whose time has come - the Reverse Stockinette Turtleneck and the Swan \"bed jacket\" cardigan.  Each has been finished for months... except the seams.  Saturday I worked on Swan, but had to take out one of the sleeves I\'d already sewn in (grr!) and Swan had to got to a safe house for a while so I wouldn\'t murder it.  Better luck on Sunday with the Turtle - from one inset sleeve and no underarm seams to completed object in an afternoon and evening.  I think it is now safe to return to Swan and try again.  If I finish those seams, tomororw is wash day for Swan, Turtleneck, Early Arrival, Feather and Fan washcloth/baby doll blankie, and my illusion knitting samples.  Wish me luck!  It seems I am reluctant to let these finished objects really be finished and go on to their new homes. '),(8216,1987893,'03,August,2004','2004-08-03 00:00:00','   OK, so maybe it wasn\'t so sensible, but Saturday as my first outing post-surgery, I took a class with Valentina Devine through  urlLink Knitting Arts .   Wendy was aslo taking the class (not Wendy mom-of-jocelyn, Wendy wednesday-jazz-salon pal)   We started bright and early at 9 am Saturday.  By noon, I was very floppy and worn out - forgot to take the pain meds in the morning! - but a little lunch perked me up and I stuck with it.  The knitting classroom space is just gorgeous, overlooking the trees of Saratoga\'s Wildwood Park.  Crows, chickadees, acorn woodpeckers, Stellar\'s jay\'s and juncos were heard all day while we knit.  Ms Devine presented Creative Knitting - knitting garter stitch in random or semi-random color selections, changing directions counter-clockwise around the square as the mood strikes you, and finishing the piece with surface embelishment.  My square used only \"leftover\" yarn  (Valentina says \"Not leftovers! These are your palette of colors.  No artist calls the tubes of paint not in use \"leftovers.\") in worsted weights.    I felt a little drab, since most of the students used novelty yarns and much brighter color choices than my Jaimeson\'s worsted.  I nearly left the square unfinished, but thought I\'d better try out the surface embellishments before I forgot everything I\'d learned.  I like it better with the decorations on.  Here it is: 		  urlLink '),(8217,3527113,'10,June,2004','2004-06-10 00:00:00','    CNN  urlLink reports  today that brand name drug prices have increased 3.4% (3 times the rate of inflation) since Bush enacted his new medicare policies.  The new cards do not make price savings certain for consumers, allowing drug card companies to modify their \"discounts\" any time they please in order to maximize profits.  The most interesting point of the article states:   AARP researchers looked at the wholesale price of 197 brand-name prescription drugs most used by older adults.  When manufacturers charge wholesalers more, \"this added cost is generally passed on as a similar percent change in the retail price to most prescription purchasers,\" the report said.   It would seem that in an attempt to privatize medicare responsibilities, the Bush administration has actually increased the cost of prescription drugs for seniors, a key demographic in its constituency.  Additionally, the Center for American Progress recently  urlLink  discovered  an unsettling connection  between the major drug companies involved in the discount program and large campaign contributions to Bush.  It seems that 20 of the 73 (roughly one third) have been involved in fraud charges.  Scary!   '),(8218,3527113,'10,June,2004','2004-06-10 00:00:00','    My good friend jble sent me  urlLink this link  today.  Even if it isnt a real company, the humor remains.   '),(8219,3527113,'10,June,2004','2004-06-10 00:00:00','    I encourage people to read  urlLink this article  about how using a word to describe a current state of power does not mean what people think it does.   '),(8220,3527113,'10,June,2004','2004-06-10 00:00:00','    I decided to go for a slightly less offensive blog template.  The green was hurting my eyes, and probably yours as well.   '),(8221,3527113,'10,June,2004','2004-06-10 00:00:00','    I went to see the much anticipated  Fahrenheit 911  last night with a friend of mine.  To say the theatre was packed is a serious understatement. It\'s the first time I remember waiting in long lines to get into a movie at Kips Bay.  Michael Moore bigger than X2 or Matrix?  Crazy.  The movie is basically Moore\'s book  Dude! Where\'s My Country  (which I read) put onto a film reel.  If it was Moore\'s intention to lash out at Bush and take every stutter, every incomplete answer, every \"bushism\" and string them together into a seemingly endless 2+ hour movie experience, he achieved it.  The movie is not bad, but it does at times lack substantiated proof to support the claims providing the foundation.  This, by itself, does not reveal much, especially since the very nature of a secret is that it remain a secret.  Like Pearl Harbor or the Kennedy assassination,  most of the secrets Moore tries to examine in his film will never see the light of day.  It is clear to me, however, that Bush is simply a cog in a huge political machine.  Instead, I believe Moore is simply trying to put as much information as possible out there to increase awareness before the election.  It\'s like he\'s saying \"look at the coincidence that a failed pipeline plan through Turkmenistan was immediately resurrected after Bush invaded Afghanistan.  Need more?  Is it just another coincidence that Halliburton has been given millions of dollars in defense contracts?  Still not enough?  How about the planes that were arranged to fly the Bin Laden family out of the States right after 9/11 that were authorized at the highest levels of government?\"  Although Moore definitely has a statement to make, I think he is trying to encourage people to look at the information that we do have; to see Bush\'s endgame.  Although I do find many of his coincidences a little too coincidental, I am not going to take the movie as gospel.  As someone who is not intending to vote for Bush in November, the movie is not going to be the single issue I take with me to the voting booth.  Still, I wonder what effect this movie will have on the people voting sitting on the fence.  Or the ones voting for Nader.   '),(8222,3527113,'10,June,2004','2004-06-10 00:00:00','    On Columbine...   Slate recently posted  urlLink this article  describing how a group of well respected psychologists and criminal investigators assembled 30 miles outside DC soon after the shootings.  They concluded that Eric Harris, one of the two shooters, fit the textbook definition of a sociopath.    Having done my minor in psychology, I\'m almost ashamed to admit that I never considered the connection.  Still, not having seen all the evidence myself, there was really no way I could form an educated opinion on the subject.  Here\'s what I\'m wondering: why do we have to come up with some sophisticated (sounding?) psychobabble to explain the hateful thoughts of a scorned boy who was angry over feeling powerless? Although Eric exhibited a complete disregard for human empathy by killing several of his classmates, is he really a sociopath?  Could it be that he just wants to be the one in control; to have the kind of power the bullies at school had?  In one of his many audio recordings with Dylan Klebold (the other shooter), Eric says:    My parents are the best fucking parents I have ever known. My dad is great. I wish I was a fucking sociopath so I didn\'t have any remorse, but I do ... This is going to tear them apart. They will never forget it. I really am sorry about all this.  It fucking sucks to do this to them.   Now, on the surface, this seems to display a genuine sense of remorse for his plans.  Below the surface, however, I get the feeling that it\'s insincere, with the exception of the reference to sociopaths.  All this means is that while he\'s manipulating his parents into thinking he cares, he is really taking the opportunity to say what he really wants to be.  What if he idealizes the power that he feels sociopaths possess on the world?  Isn\'t it possible that he is just trying to manipulate his parents into thinking that he really  is  sorry for what he is planning?  He wrote in his journal:   I lie a lot.  Almost constantly, and to everybody, just to keep my own ass out of the water. Let\'s see, what are some of the big lies I told? Yeah I stopped smoking. For doing it, not for getting caught. No I haven\'t been making more bombs. No I wouldn\'t do that. And countless other ones.      So which is it, Eric?  Are you really sorry or is this just another manipulation?  The fact that you care enough to consider how your parents would feel, even if it\'s just another form of manipulation, shows that you care on some level.  Right there, we\'ve betrayed the textbook definition of sociopath.  If you need yet more evidence, consider this:  Eric had to attend youth counseling after being caught breaking into a van and stealing things within it.  As part of completing the program, an essay was required to tell the administrators what had been learned and achieved.  Eric had this to write:   I am happy to say that with the help of this class, and several other diversion-related experiences, I do want to try to control my anger.   In his private journal, however, he said:   Isn\'t America supposed to be the land of the free? How come, if I\'m free, I can\'t deprive a stupid fucking dumbshit from his possessions if he leaves them sitting in the front seat of his fucking van out in plain sight and in the middle of fucking nowhere on a Frifuckingday night. NATURAL SELECTION. Fucker should be shot.   Here\'s what I think.  Lying is just another form of power.  Psychologists call this phenomenon \"duping delight\".  Everything about Eric indicates to me that he was just a powerless kid who allowed his rage to consume him.  He frequently quoted Hitler and even planned the shootings for what would have been his 110th birthday; he desired the power and therefore idolized the man.  It\'s not that he completely disregarded human empathy, it\'s just that his thirst for power and attention completely devoured him.  Over time, his hate and need for power were fuelled by a symbiotic relationship with his growing ego and superiority complex until it completely boiled over with the deadly shootings at Columbine high school.  Like Hitler, Eric Harris took his own life.  Isn\'t a bit suspicious that he left behind such a detailed paper trail describing the very hate underpinning the framework for his plans?   He knew that his words would not only incite people into demonizing him as the powerful monster he wanted to be, but would also provide the bloody, indelible ink for the history pages.  It\'s all seems too deliberate.  Eric got what he wanted.  People think he is a monster.  I think he was a scared, frightened, powerless kid who\'s inability to function in society destroyed the lives of countless people around him, nothing more.   '),(8223,3527113,'10,June,2004','2004-06-10 00:00:00','    Last Saturday, as I was preparing for my trip to London, I was listening to NPR when they announced that Reagan had died.  I was too young to notice the soaring 14% fixed mortgage rates or the debt ballooning 300% to  urlLink 2.6 trillion dollars  in the 8 years he was President, but I do remember the spirit that surrounded him.  There is no doubt about it, Reagan was a charming man.  It\'s hard to forget the laughter that erupted from the press whenever he was in front of a microphone.  And, unlike our current President, Reagan at times embraced his humility by  urlLink admitting  when certain high profile mistakes were made.  Despite this, as President of the United States, the buck stops with the executive.  Disavowing any knowledge of illegal covert operations does not excuse the fact that they occured in the first place.  Being charming is simply not enough.  Personally, I do not understand why many people hold Reagan in such high regard; the statistics collected during his presidency simply do not support the credit he receives.  Many argue that Reagan was the most popular president since WW2, but a close  urlLink inspection  of approval poll data simply does not support this claim.  Others argue that he is responsible for the booming economy of the 80s, but with a 11.5% unemployment rate in 1983, it also does not seem to be true.  Perhaps it was his effective drug policies that curbed cocaine use in the 80s; nope, not that either.  Reagan did have great hair, though.  Gotta give him kudos for that.   '),(8224,3527113,'10,June,2004','2004-06-10 00:00:00','    Anyone who lives in New York City (or any urban city, I suppose) knows what a fitness craze a lot of people are on right now.  I\'ve never been to a place where there are so many hunky guys and extremely hot girls   everywhere .  It all seems to fit consistently considering this city is Superficial Central.  Carrie\'s walk down the platform on HBO\'s  Sex And The City  inspired thousands of New Yorkers to drop what they were doing, rush to the nearest gym, shower, then high tail it to the nearest designer clothes store.  It\'s not at all uncommon to see people sporting Armani, Prada, Donna Karan, and Gucci gear taking their trash out.  Carrie just poured a little more fuel on an already burning fire.  I mean, we\'re talking highly combustible and incendiary stuff here.  The growing popularity of reality shows like The Swan indicate a disturbing explosion of interest in \"making over\" our bodies.  Fox doesn\'t mind the interest at all.  According to the Nielsen ratings, The Swan consistently pulled in roughly 11 million viewers each week, following their other hit show, American Idol, which did significantly better at 28 million.  Plastic surgeons don\'t seem to mind the hype either.  US News and World Report And World Reports estimates that Americans spent close to 9 billion dollars on plastic and dermatologic surgery last year alone.  Let me say that again: 9 billion dollars.  I wonder how much that Hooked On Phonics company made last year.  Are they even still in business?   I am simultaneously fascinated and frightened by this trend.  Who hasn\'t thought to him or herself, \"I dont like this part of my body.  I would love to change it.\"  Well, now you can.  Lips not puffy and alluring enough?  No problem, here\'s some botulinum toxin.   And what about those extra pounds you packed on last Thanksgiving?  We can suck those right out of you.    Still, despite the exhilleration that comes along with instant gratification, some prefer to stay fit the old fashioned way, the gym for 50 minutes, 3 times a week.  That said, I do understand the entire psychology surrounding plastic surgery.  I just wonder what\'s next?  Gene manipulation for receding hair lines?  It\'s just a bit too Gattaca for me.   '),(8225,3527113,'10,June,2004','2004-06-10 00:00:00','    A few days ago, I posted a  urlLink blog  on here about Michael Moore and censorship.  I still maintain that the republican controlled executive and legislative arms of government fervently try to  suppress  opinions that dissent from their collective agenda.  In spite of this, I wanted to make sure that there was some substantiated fact to support my claims.  Then I happened across  urlLink this article .  In it, Daschle asked the Bush whitehouse to explain the extra $600 billion dollars it was requesting for defense, driving the total cost of the defense budget up to $4.7 trillion dollars for defense over the next 10 years.  He states, \"if we\'re going to commit those resources, and if we\'re going to commit those men, those women, those lives, then I think we\'ve got to ask the questions that are required of us.\"    Of course, just asking the question was all that was required to get the conservatives fuming.  Senator Trent Lott cried: \"How dare Senator Daschle criticize President Bush while we are fighting our war on terrorism?\"  The article points out that Lott\'s selective memory is preventing him from recalling his own comments about President Clinton\'s policy in the Persian Gulf.  At the time, Lott stated,   \"real patriots have a responsibility to question policies they believe to be wrong.\"    I guess that responsibility is only truly felt when your party doesnt control the executive, eh Trent?  As if Lott\'s rebuke wasn\'t enough, other republicans hopped on the \"burn Daschle burn\" bandwagon.  Ads ran in South Dakota, Daschle\'s state, showing pictures of Saddam with the caption \"Why is America buying 725,000 barrels of oil a day from this man?\" Next to a picture of Daschle with the caption \"Because this man won\'t let America drill for oil at home.\"  Oh, I see, caring about the environment is now tantamount to aiding terrorists.     Getting back to my original point: criticizing people who raise questions about highly partisan issues not only undermines our ability to exist in a free thinking society, but sets a dangerous precedent to democracy.  People have a right to express their opinions (I think it says something about that in the first amendment), without being labeled \"disgusting\" by Tom DeLay.   '),(8226,3527113,'10,June,2004','2004-06-10 00:00:00','    Alanis, please stop making music.  I beg of you.  Enough of your   really   jagged little pills.   '),(8227,3527113,'10,June,2004','2004-06-10 00:00:00','    I\'m having a really difficult time deciding if Jessica Simpson is completely brainless.  Or, perhaps more frightening, that she\'s not.  Jess is stirring up news lately.  The popstar antidiva is mad at an event promoter, 4 Dog Promotions, for cancelling a product placement endorsement at a race in Puerto Rico last October.  Apparently, Jess was supposed to wear a shirt with \"MOTOROCK\" plasered on it.  It\'s an auto racing company, that personally, I\'ve never heard of -- that doesn\'t say much considering there\'s lots of stuff I dont know about cars, much less cars and  racing.  Still, it\'s not as if anyone would have noticed her endorsement between being blinded by the lights reflecting off the Moutain Dew and MSNBC cars whizzing by and putting mustard on a hotdog.  140k just for wearing a tshirt.  Score!    Now, this is not all that interesting.  The promotions guys didnt pay her, she has every right to be upset.  The question I have is: why do we care?    America has become socially obsessed with the ridiculous.  Jessica Simpson, launched to serious stardom with her hit show The Newlyweds on MTV.  Roughly 11 million people tuned in weekly to see Jessica and Nick unintentionally question the calculus of conflict in everyday life (relatively speaking anyway).  You might remember Jessica asking Nick \"is this chicken, what I have, or is this fish? I know it\'s tuna, but it says \'Chicken by the Sea.\'\"  I mean, how great is that?  It\'s one of those times when you actually catch yourself questioning if it really  is in fact  tunafish, then quickly realize that you\'ve allowed someone to bait you into something utterly ridiculous.    Curses! (in mock Gargamel voice)  I\'m not sure I buy it.  I think there is a deliberateness to Jessica Simpson, which is reinforced and intensified by the fact that she epitomizes every blond stereotype we know we shouldn\'t have.  While watching, we ask with ourselves.  \"is she really that stupid? what a blond.  but i know i shouldn\'t have stereotypes about blonds, so i shouldn\'t believe she\'s really that clueless.  but then again, what if she really is a total buffy?  does that mean i think all blonds are buffies?  wait no, that can\'t be right\" ad infinitum ad nauseum....  And what is Jessica doing right as we ask ourselves these questions?  Well, she\'s only sitting on a pile of cash.  She even has a fragrance line coming out soon.  Hey, Glo worked for JLo (rhyming completely unintentional).  It has to work for her as well, right?  Maybe not...   '),(8228,3527113,'10,June,2004','2004-06-10 00:00:00','    urlLink McCartney tried heroin, preferred cocaine   I can\'t tell if I\'m amazed that someone thought it was news worthy or that someone actually thought we\'d care.   '),(8229,3527113,'10,June,2004','2004-06-10 00:00:00','    How ridiculous is Usher?  I mean seriously.  I\'ve always liked his music, even back to his very catchy first hit single \"You Make Me Wanna\".  Recently (Friday to be exact), I was listening to Internet Radio on iTunes, which if you dont use, you should, when I heard his latest single Confessions, Part II.  Now, I only can take Usher so seriously as an artist, so I dont usually listen to his lyrics.  On Friday, however, I did.    Ok, rewind.    Usher was on Ellen Degeneres\' show a week or two ago.  He performed the smash single Yeah.  I was impressed.  It was coreographed nicely and Usher was wearing nice gear, but in listening to the interview afterward, I lost all respect for him.  Here\'s what happened:  Ellen comments about how much she loves watching him dance (and seriously, he can really dance), and he responds that he dances for her and that (get this!) when he takes off his shirt and dances on stage, he\'s thinkin of her.  Ok, Usher, I\'m not sure if anyone told you, or if you were still in diapers (you\'re all of 12, right?) that Ellen is gay.  Not only is it inappropriate and disrespectful to make prude and lascivious comments to women, but where do you come off doing it in your blinged out diamond chains and very expensive lip moisturizer in front of millions of people?  If you\'re going to prove that you don\'t think, at least try to minimize the exposure.    This is where I realized it.  We created Usher.  He\'s barely in his 20\'s, and he\'s livin the fabulous life that we\'re paying for.  Now, don\'t get it twisted here.  I definitely do not have a problem with people who are successful.  I do, however, take issue with people, especially entertainers, who seem to forget the people who bought the cds or bought movie tickets.  Usher has an  entire smooth operator image going on, but last I checked, Sade sang the song.  Ok, fast forward.  I hear Usher\'s new single on the radio.   In the middle of the song, he starts to talk in a \"look at me, I\'m sexy\" voice and here is what he says:  This by far is the hardest thing I think I\'ve ever had to do To tell you, the woman I love That I\'m having a baby by a woman that I barely even know I hope you can accept the fact that I\'m man enough to tell you this And hopefully you\'ll give me another chance This ain\'t about my career This ain\'t about my life It\'s about us Please  Ok, does anyone else see the problem with this?  \"I hope you can accept the fact that I\'m man enough to tell you this\".  Nevermind the fact that I cheated on you.  Nevermind that I obviously didnt use a condom.  You should accept the fact that  I am man enough to tell you  that I\'m a lying, cheating, bastard.    Usher, are you for real?   '),(8230,3527113,'10,June,2004','2004-06-10 00:00:00','    I have to admit, I\'m at a loss for words over all the hype the new Michael Moore movie is stirring in political circles.  His new movie, Fahrenheit 9/11 is apparently simultaneously provocative and chilling, so much so that Miramax, or more accurately, Disney refused to release it.  Why did they do this?  It\'s been reported that Michael Eisner, a man already in danger of losing his cushy executive job, personally blocked its release.  It was apparently too anti-dubya for Disney to justify its release, possibly risking millions of dollars in tax incentives in Florida.  Nevermind that it won Cannes film festival awards, this film is so dangerous that it had to be pulled.    Another thing, I don\'t know about you, but I for one am sick to death of hearing about the dissension (all of its forms) in Florida.  Between the much disputed 2000 ballot counting to Noelle Bush\'s (Jeb\'s daughter) predilection, and subsequent arrests, for anti-anxiety prescription drugs, I think the American public has had its fill.  Enough already!  *wipes sweat from forehead*  Anyway, back to the point.  I am beyond amazed at the preponderance of political censorship currently happening in the media.  This is not Michael Moore\'s first brush with censorship.  HIs first book, Stupid White Men, was blocked for months by Regan Books after the 9/11 terrorist attacks.  Regan Books decided that Moore\'s book was too controversial and critical of President Bush to be released.  I must admit, I\'m not all that surprised by this.  Regan Books is a subsidiary of Harper Collins, which is in turn a subsidiary of News Corp.  What else does News Corp own?  You guessed it: Fox New.  Who owns all of this?  Rupert Murdoch.  Enough said there.  It just seems to me that people or groups of people who openly criticize Bush\'s leadership (or lack thereof) are labeled immediately as unpatriotic.  Everyday, I listen to Al Franken\'s show on Air America.  He routinely gets calls from people who tell him to start supporting our President because it\'s the right thing to do in times of war.  Despite the sour, lip puckering lemon that was sold to the American public, people still believe we should be supporting this meaningless search to find WMD\'s.  Hey Bush, why not use some of that 200 million dollar election campaign money you\'ve hoarded to pay for some of the war?  Alternatively, how about using some of the 14.9 million dollars you earned from stealing land when you owned part of the Texas Rangers?  It kills me that given this gross injustice, the Bush administration actually has the audacity to use its position to not only create a war, but also censor those opposed to its agenda.  Our grandkids wont be able to pay off this war for \"democracy\".   '),(8231,3527113,'10,June,2004','2004-06-10 00:00:00','    So my nifty friend  urlLink Yaron  turned me on to the idea of blogging.  Normally, I would think that blogging is for those 12 year old girls wishing to keep a modern day diary, but after reading some of Yaron\'s delightfully (or not) conservative views on the world, I decided to chime in with my two cents.  If you think I\'m too liberal or on the left, well, I really just dont care.  This is for me.  All for me.  A little about me.  I\'m a wall street drone.  I help develop derivatives software for a bank.  If you dont know what derivatives are, you totally rock -- I dont think anyone should have to know what they are. I enjoy working out, but have started to take it far less seriously. I love Whoppers from Burger King and can finish off a pint of ice cream on my own. I love putting on clean white socks and parking myself in front of the tv. There are constantly dishes in my sink that need to be done. I will let laundry pile up for 3 weeks and realize there will have to be a day where I wear old underwear and socks waiting for it to finish. I frequently forget where I am in the middle of a sentence, get frustrated that I cant remember what I was saying, then quickly change the subject to cover it up. I am constantly listening to music, mostly on my ipod, and singing along out of tune. It annoys me that all the yuppies in the world now own ipods and dont bother to change the distinctive white headphones to make it less obvious -- when did ipods become status symbols? Oh, people at work who try to use business buzzwords (leverage, utilize, weapons of mass destruction, and the like) in every sentence they say really irritate me. I like tv shows that make fun of white trash. I actually like tv shows that make fun of anything. I love the idea of being in love, looking at the manhattan skyline from the 59th street bridge at night, and especially walking down the street sharing a soda with a friend (the ones who dont mind my backwash anyway!). Im a gemini, so I\'m moody (by default), but intensely loyal to the people who can put up with my shit and call me on it. I find the color blue soothing, but it has to be the right shade. I really enjoy listening to Smash Mouth and singing along. I love people who can finish my sentences with the exact same words I would have chosen.....oh, and Semi-Charmed Life is still one of the greatest songs ever.  So, that\'s me.   '),(8232,3527113,'10,June,2004','2004-06-10 00:00:00','    CNN  urlLink reported  today that Jeffrey Jones, movie star from the clasic cinematic nuggets Ferris Bueller\'s Day Off and Howard The Duck was charged July of last year to probation after employing a 14 year old boy to pose for \"sexually explicit\" photos.  He was sentenced to 5 years probation and was ordered to register as a sex offender.  Problem is, he never registered, and now he\'s back in the news.  I read this and felt completely nauseus.  My first thought was, why was he only charged with 5 years probation?  Furthermore, why do the courts leave it up to the very people charged with crimes to register as sex offenders at all?  Then, why do we give any grace period whatsoever?  After leaving the sentencing, he should have made a direct line for the youth offenders division to fill out forms.  The very purpose of these laws is to make other people aware that laws have been broken.    I can understand Jones\' reluctance to register.  He\'s the guy from Ferris Bueller, one of the greatest teen movies ever produced; helping kids everywhere understand that licking your palms is a surefire way to play hookie.  How cool would it be to have such a inexhaustibly determined principal trying to catch you?  In spite of all this, it was in the news last year that Jones had been arrested, and knowing the media\'s complete lack of compunction for reporting on sex offenders (especially famous ones), I would think he\'d want to avoid more bad press and just get the registration thing over with.  He could have moved on with his life.  I dont think he should have gotten probation at all.  I think adults who ask 14 year old anythings to pose for explicit photographs should be put away somewhere where we can study and rehabilitate them.    Maybe I\'m not so liberal afterall, eh?   '),(8233,3527113,'10,June,2004','2004-06-10 00:00:00','    I came across this little blurb today, which I thought was kinda neat.  It puts an objective spin on the irrationality that can occur in relationships.  A Buddhist teacher once explained that there are three stages in the character development of human beings: dependent, independent and contributive. Unfortunately, most people are completely unaware of the third, the contributive (or interdpendent), state of life. For them, there are only two options, independence or dependence. Independence,the stand-alone self, can be a happy state because we are in control, a necessary condition of happiness. The strong confident self however, can easily become arrogant and isolated. But arrogance and companionship don\'t mix well. It is all too likely that an arrogant person willl be unable to sustain fulfilling relationships. Instead, those relationships will most often end in conflict and strife. The alternative for most is dependent (or codependent) relationships. People give respect and love, but not freely; strings are attached. This is the let\'s make a deal approach to relationships, \"I\'ll love you as long as you give me what I need.\" Life in this sort of relationship can only be an emotional roller coaster, climbing to exhilarating highs and plunging into desperate lows. This is because your happiness is dependent upon another\'s behavior--upon his validation of your worthiness of being loved. Happiness in any situation can not be acchieved with out a sense of control. Depending upon another to validate that we are worthy of love gives that person control over our emotions and our self-esteem. We have given up our power.   '),(8234,3527113,'10,June,2004','2004-06-10 00:00:00','    Coca Cola is a threat to national security.  Yes folks, today, CNN  urlLink reports  that the military has warned all personnel to be aware of the new Coca Cola cans being issued as part of a summer promotion.  These highly dangerous cans are equipped with GPS chips, which as we all know can track movement and eavesdrop on private conversations.  When does this bullshit end?  As the article points out, it\'s no different than the CIA making all that fuss over Furbies years ago because they could repeat phrases.  What exactly do they want us to believe?  That the Coke execs are sitting in some dark room, dimly lit only by cigars, marvelling in their ability to infiltrate and extract information from the most secure military installations in the country?  Wait, maybe Coke isn\'t limiting itself to GPS chips.  Maybe these execs are adding mind controlling drugs to their secret formula.  Think of the security threat if powerful people at Coke could actually control politicians.  They could control the world.  The possibilities are boundless.   '),(8235,3527113,'10,June,2004','2004-06-10 00:00:00','    To all the people who don\'t think Clinton accomplished anything during his 8 year term, I urge you to look at  urlLink this .    I don\'t think Clinton needs to defend his accomplishments.  His record does.   '),(8236,2876677,'29,February,2004','2004-02-29 00:00:00','   I think I may go crazy.  My computer is next to my parents and my brother is on it but meanwhile he is also talking on the phone with John.  Nice kid I\'ve talked to him but he\'s not the brightest star in the sky.  So listening to \"dude...no way dude...DUUUDDDEE...that\'s crazy dude\" is like getting run over by a car over and over and over and over.  I dont know how much more idiocy I can be inundated with.  I have to take the SAT soon and I feel all wisdom, knowledge, et cetera being drilled out of my head.  From what I hear from my brother and his friend-on-the-phone the mullet may be resurfacing as something attractive/cool.  *bangs head againt sharp object*   College is coming up soon.  I am praying for a cool, intelligent, friendly, mature Christian as a roomate.  I think I seriously may die if I listen to this conversation any longer.  DUDE!!!!  HAH HAH HAH  THATS CRAZY!!! WOAH!!!  SERIOUSLY NO WAY!!!  I feel like my ears are on crack...  I should write a book on this or film him for TV.  I bet people would buy.  Primetime or PPV  I dunno.  I wonder if there is a circus I could pay to kidnap him.  DUDE MY PARENTS SET UP THE COMP TO KICK ME OFF AT 9.  THATS SOOOO DUMB KID!  (DAD HOW COME DUSTIN DOESNT HAVE TO GET OFF?)  ummm  he is 17.  *hey what is this on my hand-oh my brain has actually started dripping out my ears*  so i was like...oh my word...dude...no way...seriously...dude...craziness...i know...wow kid...did you know in brazil they wear speedos...dude i know its crazy...well i wont contaminate you any more   '),(8237,2876677,'17,February,2004','2004-02-17 00:00:00','   urlLink Join the War   '),(8238,2876677,'16,February,2004','2004-02-16 00:00:00','   i am so bored... but i am making progress on bob jones so that is awesome!  i sold my games for OVER $400!  WOO-HOO!   '),(8239,2876677,'31,March,2004','2004-03-31 00:00:00','   Note to Self: Block Jenna from commenting.   '),(8240,2876677,'30,March,2004','2004-03-30 00:00:00','   Irritation.  The school implemented a new talking rule today.  Good for them.  I really don\'t care because I have the ability to get serious when it is serious time.  My parents are upset because Seth got kicked out.  He is the retard.  They think it is unfair the first kid gets warned and the second kid gets punished.  Maybe so.  I don\'t care.  It is doubtful my class will ever have to deal with it.  Even if we do, I think I can handle a zero in a few classes for day.  Who cares.  Life is life.  Move on.  Blah blah.  Maybe there should be a blanket warning at the beginning to all kids and then the first kid gets kicked out.  I don\'t care.   '),(8241,2876677,'30,March,2004','2004-03-30 00:00:00','   Who am I?  Who are you?  Is my past real?  Am I real?  Am I just a figment of a greater being\'s imagination?  Are you real?  Are you a figment of my imagination?  Is the physical real?  Am I just a figment of my imagination?  Is my body a projection of my soul?  Does reality exist in this world or in my observation of it?  What is time?  Is history real?   '),(8242,2876677,'29,March,2004','2004-03-29 00:00:00','   I think I am falling...not in love.  I am falling into the early stages of wanting a relationship with someone I have a friendship with.  However, I have not yet decided if I am going to pursue it.  Am I mature enough to be involved in a relationship?  Can I acurately judge my own maturity?  Is she mature enough to enter a relationship?  Should I enter a relationship so close to leaving for college?  What does God want?  Life is so confusing.   '),(8243,2876677,'29,March,2004','2004-03-29 00:00:00','   I am irritated.  Alot.  I went to a \"Christian\" bookstore *cough*in K-mart plaza*cough* and I spent about 45 minutes trying to find something Christian.  I would have left and gone to Barnes and Noble but the only reason I was there in the first place was because I had a gift certificate.  It was a time and a half.  I went up and down the two aisles that applied to me.  There was one aisle of Bibles.  I use the term loosely.  Then next aisle was concordances, and reference books and such like.  Then there were two aisles of Charismatic and an aisle of Christian fiction.  Then a Catholic aisle and a miscellaneous aisle.  No Baptist aisle.  No good books written 50 years ago aisle.  I saw quite a few books that I saw and really wanted but then I picked them up and it would say \"Translated in to modern English\" or \"Bible references updated to NIV or NASV\".  It was irritating.  There were no books older than 1995.  It was insanely irritating.  I finally bought a concordance.  Then I went to check out.  The girl was dressed rather interestingly...  Her chest was about to fall out of her shirt.  Her pants were so low and tight I could see her underwear(not just a little bit).  Her shirt was also so high I could see her navel ring.  We won\'t even mention the music.  Long story.  Good times.   '),(8244,2876677,'28,March,2004','2004-03-28 00:00:00','   Pre-teens are REALLY starting to annoy me.  I go to the mall and there are all these little kids BY THEMSELVES.  Maybe I am sheltered or overprotected, but I was never just dropped at the mall to \"hang out\".  Actually, I wasn\'t even dropped off to shop until I got my first job.  My parents theory was if I was spending their money it would not just be given to me to wander around the mall by myself and shop at 12 years old.  All these little girls, with no clothes on walking with their backs arched(if you GMD) looking for boyfriends.  All these little boys, dressed up like they thugs with the baggy pants and the bling bling, limping around trying to attract a girl.  Little girls, find a healthy young boy that can walk right whose pants fit.  Little boys, get a girl who wants to get married and keep her body for the two of you.  You may find it attractive when you\'re in the mall holding hands walking, but if she is wearing some low rise pants and a halter top she is advertising that she is availible for another guy.  I don\'t know if this is the case at all malls but whenever Kevin and I go to Searstown, all these little girls are walking around with thongs on the outside of their pants trying to get noticed. Gonna get noticed by a rapist is whats gonna happen.  Do their parents drop them off like that?  Do they put them on when they get to the mall?  Parents, when you pick your kids up from the mall, ground them.  It is safe to assume that since you were enough of an idiot to drop your little girl or boy off at the mall for 4 hours, they inherited the idiocy.   '),(8245,2876677,'27,March,2004','2004-03-27 00:00:00','   Wow.  Fun times.  I just went bowling and I am so tired.  It was good.  It was just Emi, Jenna, and I until 7:30 and then Jess showed up.  I had gotten two lanes cuz I thought more people were coming but then they didn\'t show so we had twice the fun.  Then at 8 I brought it down to one lane because it was just the four of us.  We had some good times.  Meanwhile, I am trying to get ahold of J Trinkle and Miss Vasel who are not answering.  At 8, I decide they just got lost(it is Miss Vasel) and stop calling.  At 8:30, guess who shows up?  Yep.  The Lost Girls.  So all six of us crowded on one lane.  Then the stupid computer thing wouldn\'t stop messing up when we added them in.  On the lane next to us, these college kids had showed up and were getting a little friendly.  Especially when someone got a strike.  I wanted to give them the Yellow Pages and help em find a room cuz they were in the wrong place.  Mr. and Mrs.Gomez couldn\'t come...:(  She wasn\'t feeling to well.  Pregnancy can do that to ya.  Been there. done that. never again.  Hope ya feel better.  Thank you for having your baby \"legitimately\".  Alot of couples have opted out of that.  Kids get really messed up that way.  Mom M,W,F. Dad T,Th,Sat.  Sunday is a coin flip.  It can be fun.  So remember kids:\"Fun\" now, makes \"trouble\" later.  A little wait will help you appreciate.  I ALMOST RAN OVER A CAT!  Ok I was on my way home and I was on this street before my street.  I don\'t know the name.  Well anyways I saw this gray thing slinking across the road and decided to slow down cuz I thought it was a squirrel and last time I ran over an animal, the sound made me want to throw up.  I was really glad when I slowed down because I found out it was a cat and I don\'t want to kill some little kids pet.  That would be sad.   '),(8246,2876677,'26,March,2004','2004-03-26 00:00:00','   I really need to get out of this school.  It is so not helping my growth.  The words said in the hallway, the songs sung, the movies that are talked about are all definitely not what God has for my ears.  It is upsetting.  Everyone in my school is so liberal and they just keep drifting.  Socioeconomically, they are somewhat conservative, but on a lifestyle or faith basis they are following the world from a distance.  A distance that seems to shorten as time goes by.  I pray that the Holy Spirit will make some changes in the school as a whole.  Spiritual edification is not even thought of.  It would be awful to think that maybe the movies are wrong.  Maybe I should study it out in the Bible and pray about it.  Instead, I will bask in my sin and call the person who would make such a claim a backwards idiot stuck in the 1800s.  Definitely what \"Christians\" should say about each other.   '),(8247,2876677,'26,March,2004','2004-03-26 00:00:00','   SAT...i am so scared...i studied. for a little bit...a while ago...   '),(8248,2876677,'26,March,2004','2004-03-26 00:00:00','   hi its richard.  umm i umm i understand if you never want to talk to me again but i don\'t umm i can\'t see you anymore...umm bye  funniest thing ever...some guy just broke up with my machine...i am so glad i screen my calls...it would have totally been ruined if i answered it.  poor girl is never gonna know she got dumped.  I was crying...I just wanted to pick up the phone and laugh at him.  i almost fell out of my chair.  wow was that good.   '),(8249,2876677,'25,March,2004','2004-03-25 00:00:00','   SAT in 14 hours and I am scared...   '),(8250,2876677,'25,March,2004','2004-03-25 00:00:00','   Well today went absolutely awesome...Bowling is going to rock...*cough cough*it is me and girls...and mr.gomez...party!  woohoo.  sorry I don\'t want to offend anyone going but...yay.  Thank you all who are going but it just isnt the same.  i have no male friends at all. well ok kevin and ben.  but i see kevin once a month max. and then we go shopping which is just not a cool guy activity.  and ben is cool but i see him once every three months when i go to faca.  Kevin and Ben are both losers.  well ben not so much as kevin but still a little.  NOW Kevin is a huge loser.  it just so happens that his BEST friend is holding a party to celebrate a jewish holiday and kevin doesn\'t even ask for the day off.  and then on the tuesday before kevin tells him oh sorry can\'t go i hafta work.  THE TUESDAY BEFORE!!!  I am so upset.  everywhere i have ever gone my class and church has been mostly girls(for my age group).  i knew it was going to happen...i told someone only girls were coming and they were like ooh mackdaddy.  Not really.  i would SO much rather have guys.  girls are just blah blah blah hes going out with her and blah blah blah you straightened your hair ooohh blah blah blah   '),(8251,2876677,'25,March,2004','2004-03-25 00:00:00','   Physical education.  Some days it is good.  Some days it is not.  I guess it all depends on whether or not my body feels like learning.  There are other factors though too.  Amount of sunlight, temperature, wind chill, heat index, how the classes before went, what we are playing, and the other people on my team are just a few.  Today started out alright, drifted up, came down, shot skyward, and plummeted back down again.  I guess I just don\'t understand people.  People don\'t understand me though so I guess the feeling is mutual.  Speaking of being misunderstood, I am sick of people saying \"You just don\'t understand\" or \"It doesn\'t matter what I look like.  It\'s what on the inside that counts.\"  I do understand. You are shallow.  What is on the outside does mater.  Generally, it reflects the inside.  You are probably not the exception.  Whoah!  That was a little off the beaten path. Where was I?  Oh right!  I just don\'t understand people.  I am wicked competitive.  I HATE losing!  I used to be the worst loser ever.  Then in second grade, I grew up.  Hasn\'t happened yet in my gym class.  I was a fairly poor winner too until probably ninth grade.  I have mostly grew out of that though.  Anyways, PE can be fun.  People get way too aggressive.  I mean \"play hard\"!  Hoo-rah!  Inside of the rules though.  Punching the ground, flipping benches, and throwing students generally are not in the rules.  I am 99%  sure that putting up your middle finger when the teacher turns his back is not in the rules.  Swearing probably goes right along with that but that is more widespread than just PE.  And then there is the making fun of people.  I find it comical when a un-named 400 pound sophmore throws a temper tantrum and hits stuff and swears and lays on the ground holding himself and over-exagerates his injuries.  I am sorry but I really want to inflict actual pain on the pansy when he does that.  Not trying to get pity, I would like to mention that I however ripped through three layers of clothing and my knees, tore open my elbow, received a huge gash to my lower back, and scraped my palms today.  I did not lay on the ground and try to get the gym class to surround me and ask if I was ok.  I got up...then walked over to the side and quietly sat down because my knee hurt.  I feel bad for the seniors.  Sure they say absolutely idiotic stuff but they deal with a lot of unwarranted aggression.  The sophmores, who in general are...the coolest kids I know *cough**nose grows*, give the seniors alot of verbal and physical abuse.  The seniors for the most part don\'t have much to say back.  I admire the amount of verbal crap they take.  I see them tense up, and then let it roll off.  Not too many people do that today.  I mean after so much, problems escalate but hey what can you do?  BTW me falling I know that must sound funny because it also involves me being \"active\".  Here is the deal yo.  I was bored in gym.  I didn\'t want to participate because I have no skill and can\'t outrun your grandmother.  Yes I meant the one with the walker.  So I just kept walking and then Mr.Doan would walk up and tag me and I would smile and walk back and tap the dumpster and then I would walk back and do it again.  And again.  And again.  Well then something unexpected happened.  I walked through and picked up a cone.  Then I took off as fast as I could.  Again, this is not to fast.  Well I am running and running, trying to get back to the dumpster as fast as I could and I am looking backward to see if I am about to be tagged.  Sidenote:  Me running is dangerous.  Me walking and looking backward is dangerous.  Me running and looking back is a volatile mix.  On the my erratic journey across the blacktop, SO CLOSE TO THE DUMPSTER, my feet had a little disagreement about something(I couldn\'t hear too well cuz my ears are so far away and I didn\'t want to stop and ask at the time), and then the blacktop and I had some one-on-one time.  I ripped through my second favorite pair of pants.  No not the gym pants.  I was wearing these awesome pajamas under my pants as I always do and now I have holes in my knees and this wonderful spot where the blood dried.  mmm...That was a joy to pull from my leg.  Then I did my homework.  Mr.Gomez I put my heart into that paper.  Actually, just some blood from my wrist.  It was an awkward place for a band-aid so I just decided to bleed on the paper.  Not too much.  Just enough to remember me by.  I got the package in the mail.  It said \"STRAIGHT TALK\" on the front in cool letters on the front and my first thought was \"I don\'t remember joining any supremacy groups.\"  As it turns out, it was just a little dvd on BJU and \"the facts\" about campus life.  funny stuff...   '),(8252,2876677,'24,March,2004','2004-03-24 00:00:00','   Today was good.  I decided when I woke up not to let anything get me down.  It didn\'t work but I still had a good day.  I really feel lighter with the burden of bitterness off my shoulders.  It might have something to do with fact I have stopped eating though too.  I got a compliment on my writing today.  I was told that even though the person doesn\'t agree with my point-of-view that my journal was well written.  How cool is that?  Now its not as if this compliment came from an English major but still it was nice.  Today was good.  I got caught up on some work, shot at the breeze, and sat bored out of my mind in study hall.  I really wish I could leave...more often than I do.  It is boring.  I have nothing to do except listen to the seniors.  Let me tell you firsthand, as a survivor of the mind-numbing, hearing about car engines and freshman for 50 minutes is not fun.  Independence is what I need.  I don\'t know.  I guess I can wait although it is pretty boring being seventeen.  Not much different than the boredom of sixteen.  I didn\'t accrue any rights or priviliges.  Just a day.  It will fade away.  So will everything else though I guess.  Everything reaches a peak and then just fades away.  Your body, intelligence, wealth, and some other stuff too I guess.  Vanity, vanity, all is vanity.  You know what\'s nice?   Holding someone real close and feeling them breathe.  Feeling their chest rise and fall.  Feeling their heart beat next to yours.  Not that I would know. ;)  I guess what I am trying to say is   '),(8253,2876677,'24,March,2004','2004-03-24 00:00:00','   Life is lonely when you have to re-establish trust.  I feel like I can\'t trust anybody now actually.  When one person stabs you, it makes you wonder if everyone else is getting ready to.   '),(8254,2876677,'23,March,2004','2004-03-23 00:00:00','   Well wisdom kicked in and I deleted a post.  If you read it then you read it, but if you didn\'t than you dont need to know content.  Needless to say it was not appropriate of the Christian I am striving to be.  I really need to learn to cool it and turn the other cheek.  Definitely going to be an area I need to work on.   '),(8255,2876677,'20,March,2004','2004-03-20 00:00:00','   The coolest thing ever would have to be...intelligent conversation.  I love talking to people and piquing their brains for information.  Questions!  I love those.  I like answering questions with a question.  I like being answered with a question.  Its fun.  I like getting out a three-hundred pound concordance and searching for verses.  It is definitely much easier to open up a Bible program and have all the references pop up on my screen for almost any given topic, but it just isnt the same.  Books are just fun.  Not to get off the subject too much (although it is my blog so I can do what I want...), my library is driving me insane.  If a person from 200 years ago were to go to a present day library I think they would be appalled.  25 rows of DVDs compared to 10 rows of books.  This is very irritating.  Libraries today are filled with the likes of  Scary Movie , brand new computers with flat panel monitors, and  Seventeen  magazine.  All this worthless depravity and no Paul Bunyan or Soren Kirkegaard.  I will find you hard-pressed to find any book that does not have pictures, or a liberal slant, or a book written over 20 years ago or one about God from anytime.  However Harry Potter is always in stock.  Why they even have curtains around some computers for \"personal viewing\".  How wonderful.  Where is the education in a place supposed to house our knowledge?  We as a people are going downhill.  People can\'t read a two page paper but they can sit through Return of the King.  No one knows how to use their imaginations these days...  I hate it.  My kids are being read CS Lewis to bed and are listening to Bach.  People are just dumb.  I have been dumbed down to some extent.  You definitely have...   '),(8256,2876677,'20,March,2004','2004-03-20 00:00:00','   Well its been about...three days since I last wrote to my journal.  Long time.  The only reason I am writing now is because I can\'t sleep.  I have been up since two.  I had the WORST dream I have ever had!  I am not exaggerating.  It frightened me.  It was not a horror dream.  It was a normal day...with twists.  Awful, gut-wrenching twists, that tore my soul in two.  It was like a premonition of today.  I am really hoping the dream is not foreshadowing what is to come.  If it is I will throw up.  I spent from 2:00-3:30 thinking about it.  At that time, I thought it was the yesterday, and I was sickened and just could not go to school.  At about 3:30, I realized that it could not have been a Monday in my dream and that it was \"just a dream\".  I spent the next half hour or so praising God that these people I had dreamed about had not said the awful things they said and done the awful things I thought they had done.  I was so thankful because in my dream I was shocked these people were doing all these things.  People I respect, People I trust, People I look up to were all doing horrible sin and everyone was participating.  It seemed like everyone was ignoring me and no one cared what I was saying about God.  I left in disgust.  Only one person left to follow me to see why I just walked out as if he could not understand why.  I told him I could not believe he had just said the things he talked about.  I could not believe he had sunken to the levels of my peers.  With sadness in his eyes, he left.  I rejoiced thinking he had repented.  I went back to talk to him more only to see him back doing the same things.  He looked at me with hate and told me never to come back .  It killed my heart.  The dream was so intense so real.  I pray it doesn not happen.  It has already been four and a half hours and it stills feels real and personal.  My hero feels shattered and broken.  I do not know if I can face him.  I do not know if I can go and look at him and think of him the same as before.  It was not him.  I know it was not him.  I know it was a dream.  I know he did not truly say the words echoing in my head.  I know he did not do the things seared into my memory, flooding back across the screens of my eyes every time I close them.  The suffering was so intense, the debauchery so real, the hate so painful-it was real to me.  I hope I can get over this dream.  I know everyone was awful but for some reason even in the dream he is the only one that seemed unnatural.  The only one I truly could not believe was sinning so grossly and profanely.  The one who initiated it.  I pray I can concentrate today, but my heart still hurts from my wicked imaginations.  I wish I could wipe this dream away.  Only I know that the man I saw in my dream is real.  He is inside my hero, inside you,...inside me.   '),(8257,2876677,'20,March,2004','2004-03-20 00:00:00','   I was downloading the newest version of Quicktime about...2 hours ago and decided to watch a movie clip from online.  Well I chose Setve Job\'s Macworld Expo keynote speech...Wow it was two hours long...Very charismatic though...made me want a G5...during the keynote he had a few of people come on stage to talk about their product.  The only one I can remember is the Microsoft Office 2004 lady because she was SOOOO drunk.  I mean she was swaying ALOT and just looked plastered.  I couldn\'t believe it.  If she was sober, and that was Microsoft\'s best, they have some problems...Well at least I got some work done while I watched.  I learned what teraflop means too.  Cool word.  Tro, teraflop,... this week had some good \"t\" words.   '),(8258,2876677,'19,March,2004','2004-03-19 00:00:00','   i am getting frustrated with people making up stuff.  they see 10 seconds of one incident and make up 10 minutes worth of gossip.  Seriously, my business is my business.  Noone has a life so they talk about everyone else.  I am getting sick of \"Christian\" school.  In math class, if you want to say \"This isn\'t going to help me in real life.\", go ahead, you are probably wrong, but go ahead.  In Bible class, however, I just do not understand how my \"Christian\" peers can talk and not pay any attention.  ITS THE BIBLE!!!  It important for our spiritual growth!!!  We are learning how to intelligently defend our faith!  ISN\'T THAT IMPORTANT TO ANYONE???  *takes deep breath*  I need to better obey the rules.  I have made a genuine effort in dress code and I need to read the handbook and try the other stuff.  Lead by example...and then never come back.   '),(8259,2876677,'19,March,2004','2004-03-19 00:00:00','   I am so sick of dealing with people.  everyone is SO tiring to deal with.  the infighting, the backbiting, high school is so cutthroat.  how much more would we learn if the drama was left outside of school.  if school was school and dating/home life was left there.   '),(8260,2876677,'18,March,2004','2004-03-18 00:00:00','   i feel wicked lonely...and i don\'t know why...its this unfathomable and inexcapable lonliness...i don\'t think i could be more lonely if everyone on the earth died...i don\'t think i could be less lonely if my best friend showed up right now...its like an invisible wall seperating me from humanity...boy do i hate being a guy...seriously in the locker room today the talk was just absolutely disgusting...learned a new word actually...wonderful! bet i will use it often...i just wanted to tear into everyone of them and tell them what perverts they all are.  doubtful they would care... i really need to be away from all people...seriously EVERYONE IS JUST WICKED myself included.  will tomorrow be better? was today bad? better than yesterday???  what does the future hold?  a wife? a family? a job? celibacy? homelessness?  drug addiction? illegitimate child? christian ministry? opulence? poverty? intelligence? stupidity? a life-changing accident? death of a loved one? country ruled by a dictator? OH TO BE IN HEAVEN.  jenna said \"tro\" in her prom prospectus and wow isnt that the hottest word ever.  definitely at least a word of the week.  Don\'t let drowning ruin your summer.  Paddle smart from the start.  I want to go to NYC and bask in my lonliness amidst a swell of humanity.   i dunno.   i am just exhausted...   '),(8261,2876677,'17,March,2004','2004-03-17 00:00:00','   I was outside making sure my dog didn\'t run off while she did her business and I really wish she would just run away...or die.  Seriously she is like 90 something years old and she hasnt died.  Actually she is somewhere in between 98 and 105 according to my calculator.  She really is just holding on.  LET GO AND DIE!!!  well anyways that wasnt even close to where i had intended to go with this blog...I just thought it would be romantic to be kidnapped...Like if I was in college and my g/f had my friends like jump me and take my to her...that would be suprising and romantic...now i dont think it would be romantic vice versa and i would never want to put my g/f through thinking she was being kidnapped. anyways...just my wierd little thought for the day...   '),(8262,2876677,'13,March,2004','2004-03-13 00:00:00','   i am going to go crazy...i need a giant stress ball...either that or just pop seths head...if i hear \"kid\", \"dude\" or any other idiotic phrase I really am going to kill him...its just dumb...almost every person i have ever heard say \"kid\" is saying it to someone older...seth says it quite often to me and its annoying...THESE LITTLE LOVE TRIANGLES ARE DRIVING ME INSANE!!! i mean seriously every day all these little kids are breaking up and the little girls cry for 2 hours until they have a new boyfriend for a day and a half...seriously...if they are making out in 8th and 9th grade what do people think they are going to be doing in 11th and 12th grade...seriously parents...get some sense...spank those little retards...ground em whatever...they all think they are little adults with big attitudes...seriously someone heard a baby crying and said they didnt wanna have to deal with that so they didnt want kids...i am cool with the baby part but i dont want to have kids cuz of the idiots they grow up to be...   '),(8263,2876677,'13,March,2004','2004-03-13 00:00:00','   DOCTORS LIE!!!  well i had to get eyedrops from the doctor yesterday and boy are they fun!!!  she tells me they may cause slight burning or irritation...apparently it would be undoctorly to tell a patient that they have found a way to bottle hellfire and that i am supposed to \"make a pouch(with my lower eyelid), drop some in, close my eye, allow eye to dissolve. repeat on other eye.\"   '),(8264,2876677,'11,March,2004','2004-03-11 00:00:00','   lust...only way to describe it...banana republic clothes are sooo hott...i need a shopping spree...mmmm  urlLink Hott  Shirt   '),(8265,2876677,'10,March,2004','2004-03-10 00:00:00','   Pink Panther...Pink eye...well...i found out this evening that the reason i have been so sick is that i have pink-eye...which is apparently highly contagious...i hope noone at school go it...especially hannah or j. trinkle since i was in their cars...sorry if you did...i feel wicked bad...although...there is a group of people i would mind infecting...but as a general rule i dont go near them...too bad...well more cough syrup for me   '),(8266,2876677,'10,March,2004','2004-03-10 00:00:00','   mmm...ahhh...so excited...two words...cough syrup...   '),(8267,2876677,'10,March,2004','2004-03-10 00:00:00','   Things I want for my birthday(not in order of importance) 01. A clearer face 02. 40 more pounds(of me) 03. Hair that has more body and doesn\'t friz(actually Tom Welling or Evan Farmer\'s hair would do) 04. A pony(not really) 05. A 1995 Red VW Jetta III GLX VR6 with all the options 06. A wife 07. A really nice gun 08. A wife who can shoot a shotgun one handed and not fall down(HOTT!) 09. A rotweiler 10. A nice athmene 11. A personal library with a lot of books 12. Warm weather all the time 13. Everyone in the world to be happy 14. Everyone in the world to have a personal relationship with Jesus Christ 15. ...umm-i don\'t really want many material things. i like clothes though 16. a constitutional ammendment banning homosexual marriage(don\'t get me started) 17. people to not be quite as dumb as they are 18. less government regulations 19. a flat tax for everyone 20. israel to be free from oppression and to kick out those stupid muslims   '),(8268,2876677,'09,March,2004','2004-03-09 00:00:00','   Well I have heard way to many guys say \"I\'ve never cried\" or \"I haven\'t cried since I was a little kid\".  Frankly I hope its a lie but if its not then I don\'t see why you think not crying EVER makes you stronger or more masculine.  I mean I don\'t cry often but I do cry when things happen.  Like when my one of my best friend\'s sister dies then I cry.  Or when somebody who I really care about tells me they have had sex before marriage and have a STD, then I cry.  I just isn\'t worth it to hold it in.  Life is way too long to hold it all in every time something goes wrong.  Life is way too short to let it hold me back.  Crying just makes things better.  If you hold all the tear\'s in, you are drowning your soul.  I really wish I wasn\'t sick today.  I missed school.  I am missing church.  I MISSED MRS.BODANZA\'S PIZZA!!!  Shout out to Jess Booth.  You are a nice kid.  Sometimes I wonder how our school got so lucky.  Sorry for ever being mean to ya. Remember:It\'s ok to cry.(well not about everything cuz then your just a pansy.)   '),(8269,2876677,'09,March,2004','2004-03-09 00:00:00','   i missed pizza day...how depressing...pizza day is only once every other week AND I MISSED IT!!  i was so excited about it too...tear...i really have to get her recipe...its the best   '),(8270,2876677,'09,March,2004','2004-03-09 00:00:00','   Wow!  I am...actually I don\'t know how I feel...not really happy not sad...just in the middle...Well today went well I guess...but I have not written about Sunday yet...My pastor prayed for me specifically in Sunday morning service that the Lord would help me with all I am trying to do...it was the most moving thing ever...I cried...then Sunday night all these families came up to me and said they were praying for me in their devotions and I wanted to cry again...i love my church...everyone is so nice...Well today I woke up with no voice and a wicked sore throat...well i tried to speak and i could speak it just hurt alot and my class told me that my voice sounder better and deeper and more manly...well i guess that is sort of a compliment but my voice is going to go back and what then???   '),(8271,2876677,'07,March,2004','2004-03-07 00:00:00','   Spring is in the air Everywhere I look around Spring is in the air In every sight and every sound.  Such a beautiful day.  I feel like doing a load of laundry and hanging it outside.  It smells so nice.  I have not been alive for very many springs but I love them so much.  I  could be 1000 years old and never have seen Spring too many times.  Bathing in the sunlight walking in the rain.  Spring is so beautiful.  After Winter\'s long painful travail Spring has finally been born.  It makes life seem so whimsical and carefree and everyone seems happier.  I know I feel rejuvenated.  I turned in an application to Old Navy and I really hope I get a job but if it is not the Lord\'s will then it is not the Lord\'s will.  Let it be...  I think I may be bipolar.  I dunno but the weather really does have an effect on my attitude.  *Shouts for joy*   '),(8272,2876677,'06,March,2004','2004-03-06 00:00:00','   Well I went to get a job at 9:00 on a Saturday night...what an experience! i went to PacSun and my friend said that they usually start hiring in May(I love having the inside scoop) so I went to the Gap where a woman I highly suspect to be pregnant told me that I had to be 18 so I moved on to Old Navy where I impressed the manager with how well dressed I was and he gave me an application.  Well tomorrow in between services I need to go get Diana\'s info for a reference.  Guess what!!!  Gap and Old Navy=SAME COMPANY!!!  (same discount) YES!!! Barnes & Noble discount too and Banana Republic but it is doubtful i will be able to shop there with a discount lol.  !!!At pacsun there was this rack that said 9.99 on it so I asked if everything on it was 9.99 cuz it was wicked nice stuff that i have wanted all winter and she said YES!!!  wow unfortunately my excursion that was meant to earn money ended up with me buying a rack of clothing at PacSun but how can you pass up 9.99 for really nice sweaters, jackets, dress shirts and long-sleeve tees??  The answer:YOU CANT!!!   '),(8273,2876677,'06,March,2004','2004-03-06 00:00:00','   Well It is now official EVERY senior in my school is \"with\" or greatly desires to be \"with\"  a freshman or sophmore (except for tim...i think...but he lives on a farm if you GMD).  Not to sound judgmental but EWWW!!  SERIOUSLY!!!  DISGUSTING!!!  THERE HAS TO BE SOMEONE OUT THERE YOUR AGE!!! well thats it for now...think happy thoughts02   '),(8274,2876677,'06,March,2004','2004-03-06 00:00:00','   Mmm...Today was absolutely stunning...I went outside early this morning and the birds were singing and it was absolutely beautiful.  Amazing!!!  I don\'t know if you know but spring is my favorite time of year.  Well I need to find a job because Bedford is just not fun to drive to or in.  People are so rude.  I mean how much quicker are you going to get somewhere by cutting me off, or honking your horn at someone, or swearing or anything.  Even if you do get there quicker did the ends justify the means?  Why cant everyone just be happy?  I am totally inconsistent with my formatting.  I love that.  Well anyways I am going to apply at Old Navy, and the Gap and hopefully PacSun but I am not quite sure how my mom will feel about that...  I really need clothes and an employee discount wouldn\'t hurt the cause.  Well today was absolutely the most beautiful day since a long time ago.  I drove with my windows down (and the heat on because an unnamed passenger thought it was too chilly and that he might catch a cold.)  mmm...fresh spring air...bare feet in the wet grass...delightful...well i am off to get work. urlLink Hint Hint   '),(8275,2876677,'05,March,2004','2004-03-05 00:00:00','   Apparently lack of sleep is not only dangerous when you are driving but also when you are home alone.  I fell asleep while I was doing Economics and I had a dream I was stabbed to death while I was sleeping.  What a comfort...Well now I am really tired and I can\'t go to sleep because I am here by myself and I just predicted own death.  Abraham Lincoln had a dream about dying and look what happened to him.  Also being stabbed to death is in my opinion one of the worst ways to die.  I would much rather have the blood sucked out of my neck(Hi Hannah).  Apparently \"some people\" find vampirism abnormal but  \"unnamed others\" find it wicked hott but now is not the time or place to discuss such things.  I need a rotweiler or a gun or both.  I need a big  guard with \"MOM\" tatooed on his shoulder.  HYMNS!!!  That will help my nerves!!!  HYMNS GRAPEJUICE AND A GOOD BOOK!!!  unfortunately  i am too afraid of the man in the hooded robe who stabs me repeatedly in the back of the head to actually go get my book or juice so I will just sing \"Come Thou Fount\".  my house is way too dark and creepy...I wish Jesus would bodily come forth or an angel...pretty much anybody.  You know what would be nice... if Mr.President were to knock on my door and offer Secret Service protection(from myself).  Actually if someone were to knock on that door I would have a heart attack and die and save my murderer the trouble...What a coward...i can wait till i am old enough to own a gun cuz in moments like this i need it.  i think getting married will help because then i will have another human to talk to and also the protection instinct will kick in...sorta like when a mom lifts a car off her kid.  how does the kid get under a parked car anyways?  where is the driver to this vehicle?  something to think about...   '),(8276,2876677,'04,March,2004','2004-03-04 00:00:00','   I lied...well not on purpose...but anyways i said a few days ago that everybody has enough time just not the right priorities.  well come to find out i dont have enough time...i have school then homeschool then work then friends...sad life...well now i would like to add a couple ministries and i dont know where to fit them...so thankful i dont have a g/f cuz she would probably have to go...heartless but i would have to have divine intervention for me to believe that God wants me to be in a relationship right now even if someone did want me...well actually people have shown interest i just had higher standards so they dont count...well anyways back to my non-life...i think i may have to drop work(or sleep) which means i would also have to drop my phone and eat the contract but that would mean that i would have to drop my car insurance too and i need that to get to school and any ministry or anywhere AND i would have to empty my savings for books and stuff cuz my parents want me to buy my own books for responsiblity AND I WOULD MISS MY PHONE SOOOO MUCH!!!well actually now that i have been thinking about it i guess i do have time just not priorities...i could be working instead of writing and i could only go online to talk to BJU and i could sever any and all human relationships i have...but i would prefer that God would hold the sun for me.   '),(8277,2876677,'04,March,2004','2004-03-04 00:00:00','   I am informally teaching myself XHTML.  wow do I feel dorky! everytime I right-click a link in firefox i check out the html sourcing and memorize it.  right now i can make a link and name it, italicize words, and make a mailto address with subject (cf. Contest YAY! link).  well now i gotta go back to my beloved, the keynesian theory of economics.  oh how happy she makes me.  we will be up till 11 getting to know each other better (inserts sarcastic yay!)   '),(8278,2876677,'03,March,2004','2004-03-03 00:00:00','   Everybody loves some good competition, especially free-market economies(just putting some new knowledge to work).  Well anyways my brain was overloaded so I decided to write a \'lil entry.  This blog needs a name.   Happy Flowers  just isn\'t milking the goat if you GMD.  I know it was  Dead Souls   but some people aren\'t well acquainted with the Russian classics and thought it was a bit morbid.   urlLink Send your name   If your blog name is chosen, your name will be put in a hat with 100 of mine and if it is pulled you may be eligible to recieve a nice unspoken passing thought from me.  Good luck and may the best win.  Few will enter, I will win.   '),(8279,2876677,'03,March,2004','2004-03-03 00:00:00','   6:30 Wake up 8:07 Arrive at school (Theoretically) 8:10-3:00 Learn, Argue about banquet, olympic day etc., try to read during study hall 3:30 get home do TCCS homework,  start on PACE and snack 7:30 take a break and check on Kings of Chaos 8:00 go back to PACE 9:30 read bible, read current book, study sat 12:00 roll over and die   '),(8280,2876677,'02,March,2004','2004-03-02 00:00:00','   I would just like to praise the Lord Jesus!  Not only has he given me salvation and the opportunity to live in the most free country in the world, he allowed copy-paste to be invented!!!  Shout out to Mr.Gomez Thanks for teaching me how to use Excel!!!  It only took me like 30 mins to do my pre-cal homework thanx to the good old folks at Microsoft!!! What a blessing!!!...unfortunately i did not get all my other 3000 projects done...I think i have given up studying for the SAT and ACT  I just need to relax...if college is anything like this I think I am going to die...I need a girlfriend!  WHERE ARE YOU???   '),(8281,2876677,'02,March,2004','2004-03-02 00:00:00','   Well I think I might pass out!  I am writing my research paper, doing pre-cal h/w, and i just got this lil poem and even though the end is doctrinally unsound I thought the rest applied to my life so I posted it.  I knelt to pray but not for long,  I had too much to do.  I had to hurry and get to work  For bills would soon be due.  So I knelt and said a hurried prayer,  And jumped up off my knees.  My Christian duty was now done  My soul could rest at ease.  All day long I had no time  To spread a word of cheer.  No time to speak of Christ to friends,  They\'d laugh at me I\'d fear.  No time, no time, too much to do,  That was my constant cry,  No time to give to souls in need  But at last the time, the time to die.  I went before the Lord, I came,  I stood with downcast eyes.  For in his hands God held a book;  It was the book of life.  God looked into his book and said  \'Your name I cannot find.  I once was going to write it down...  But never found the time.\'   '),(8282,2876677,'02,March,2004','2004-03-02 00:00:00','   Just signed on for a minute to check my mail, write a \'lil entry, and GO.  life is way too busy.  i get up @ 6:30 go to school and get home between 3:10-4:00 depending on if I have any errands then I do TCCS homework, then I do a PACE (which are a lot more work than they seem), eat dinner, CRASH for 45 mins, read a book and wish I had money to buy new clothes and a Jetta.  I have a feeling I am going to be pretty much poor for the rest of my life.  well its not whatcha got its whatcha ...(insert happy maxim here)...well church starts soon so i guess that is all for now   '),(8283,2876677,'02,March,2004','2004-03-02 00:00:00','   100 Things I Like (not i any particular order) 1.honesty 2.a girl who I can just sit next to and we can both have our own books and co-read 3.a girl who dresses attractively but modestly 4.a girl who is athletic or semi-athletic actually I guess it doesnt really matter 5.a girl who weighs 106ish-160ish depending on height/build 6.a girl who is 5\'2 to 5\'8(subject to change with my height) 7.a feminine young lady 8.a girl who says excuse me after she burps 9.a girl who doesnt belch 10.a girl who makes me laugh (but not because she is dumb but because she is witty) 11.a girl who wants to serve God 12.a girl who is compassionate 13.a girl who believes love is not blind to age but the older you get the older the person you like can be 14.a girl who is a year younger than me or the same year (older just seems wierd and too young is like     pedophile-ish) 15.a girl who likes dorks 16.a girl who is smart (she doesnt have to be a astrophysicist just not an idiot) 17.a girl who likes the \'95 Jetta GLX VR6 18.a girl who thinks with her head and not with her emotions all the time 19.kindness 20.a girl who recycles 21.a girl who is not all like buy me this oh and that too ooh thats pretty too 22.a girl who shaves her armpits and legs but does not show everyone...or anyone her pits or legs 23.a girl who is optimistically realistic but not unreally optimistic or pessimisstic 24.a girl who prefers vanilla to chocolate 25.a girl who is not like eww a bug and jumps on a desk and screams at the top of her lungs until someone kills it 26.a girl who likes to talk but also to listen 27.a girl who wants the first guy she ever kisses to be her husband 28.a girl who likes jewelry and dresses nicely but she is not over decorated 29.a girl who wants to stay home and raise her kids 30.a girl who gets along with her family 31.a girl who likes being suprised 32.a girl who likes the city 33.a girl who likes the country 34.a girl who knows where she stands 35.a girl that likes boys 36.a girl who likes cooking 37.a girl who likes being cooked for 38.a girl who has longish hair (like neck,or shoulder not knees) 39.a girl who is going places 40.a girl who is going in  my direction 41.a girl who lets the boy initiate(doesnt ask him out or to marry her) 42.a girl who has self confidence 43.a girl who builds herself and others up but is not arrogant 44.a girl who speaks softly but carries a big stick 45.a girl who is mature 46.a girl who likes fun stuff 47.a girl who is innocent 48.a girl who LOVES monopoly 49.a girl who has read  Concept of Anxiety (ok maybe thats a lil much but at least a girl who has read  Have We No Rights? ) 50.a girl who likes wearing skirts 51.Vanilla Ice Cream 52.swiss cake rolls 53.pepperoni pizza 54.cherry coke 55.licorice 56.chocolate-chipless cookies 57.peanut butter 58.lucky charms 59.grape juice 60.pajama pants 61.my bed 62.my computer 63.high speed internet 64.my cell phone 65.my mom 66.summer 67.spring 68.getting presents 69.giving presents 70.reading books 71.writing web blogs 72. urlLink Kings of Chaos  73.(some types of)free stuff 74.getting my fingers done 75.snow 76.skydiving 77.tubing 78.walking 79.being alone 80.being with people 81.jumping rope, four square, hide-and-seek(j/k...sorta) 81.jews 82.chewing bubble gum 83.picking out what i am going to wear 84.buying clothes 85.driving (nice vehicles) 86.eating 87.taking hot showers 88.jumping into freezing cold water 89.teaching 90.learning 91.painting 92.wiring an outlet 93.hanging a door 94.minigolf 95.10 pin(bowling) 96. urlLink Five Finger Filet (sick I know) 97.9 ball 98.gin rummy(j/k) 98.practical jokes 99.looking at someone else and wonder what they are thinking 100.capitalism   '),(8284,2876677,'01,March,2004','2004-03-01 00:00:00','   Hey!  Wow I guess quite a few people have read my \'lil journal!  That makes my day.  Well actually God made my day.  And I was already happy before that because today in English my contact fell out and so during lunch I called my mom to ask her to bring me one so I could see driving home AND she said my PACEs had just arrived when I called.  I am such a dork.  I am so excited about homeschool.  How often do you hear a kid say that???? Not often.  I am very irritated.  As you know if you read through all my pet peeves (apparently a difficult task), you would know that I absolutely ABHOR being told GOOD JOB in PE.  well anyways after dodgeball in PE today all the guys said GOOD JOB Dustin!  It was NOT a good job!!!  I was a total klutz.  Now on a day I actually participate I do not mind Nice try or something to that effect but NEVER can anything I EVER do in PE be considered a GOOD JOB!!!  Oh almost forgot that I have to make a list of 100 things I like. I think I will make that a seperate post.  BTW Jess you can find my web journal at http://triquetta17.blogspot.com   '),(8285,2876677,'01,March,2004','2004-03-01 00:00:00','   HOLY COW!!!  I AM SUCH A JERK!!!  I almost cried.  I did something irreprehensible.  I am so embarrassed.  I made a young lady come after me.  I feel so bad.  Ok Well here is what happened...I was at BJU this October and it just so happens I walked up and down the aisle on the bus and made friends with everybody.  Seriously...I made my presence known...which is very unlike me...usually I am quiet and laid back when I go somewhere new but on the BJU trip I EXPLODED...well anyway back to this young lady...One of the people I made friends with was a young woman Jaspa.  Well anyways we were talking on the way home and we exchange addresses and she gave me a picture of herself and I am not in the habit of keeping pictures of myself on myself so I gave her my student ID.  well on the back her picture she wrote the following message \"Rusty it was awesome to meet you and get to know you.  You\'re a pretty cool dude (and a gentleman!)\" and then I PROMISED TO WRITE TO HER!!!  well anyways we went home and I put her pic w/ all the other pics I have and everytime I went by it I thought of my promise.  but i am lazy and a sloth and from what I have found did a very \"guyish\" thing (apparently most men do not like writing letters either).  TODAY HOWEVER i got a letter in the mail from guess who.  thats right JASPA BOUCHARD!  I felt SOOO bad. to make matters worse i open the letter and she says that she has NEVER pursued a young man IN ANY WAY NOT EVEN IN A LETTER!!!  I have caused a young woman to break her principles and I am mildly flattered/embarrassed.  WHAT A JERK!!! well anyways thats it for now   '),(8286,2876677,'30,April,2004','2004-04-30 00:00:00','   urlLink One Year Anniversary  Bush Flash Dot Com-Your major source for all your propaganda needs.  I thought this was an interesting bit of film.  Just so everyone knows, I am against war.  War entails having people kill people.  I am against murder. \"Thou shall not kill\" does not have a wartime clause.  Turn the other cheek.  \"What about protecting our rights?\" you ask.  We have no rights.  Jesus was killed and he did not fight back.  He did not allow his disciples to fight back.  So a dictatorship might not be the best political-economic system.  That is alright.  I am against trying to pacify madmen.  I am not a pacifist.  Non-violent resistance.  If they win, they win.  So we have no rights.  Do we need the freedoms to be Christian?  No.  So quality of life is not the best in Communism.  The church flourishes under great persecution.  Think church at Smyrna.  Night.   '),(8287,2876677,'29,April,2004','2004-04-29 00:00:00','   Well today I went to school more preppy than I normally do(I felt very \"casual CEO\").  Well anyways it was a good day at school.  I got home and changed before my grandparents(who are staying with us this weekend) saw me.  I decided to go \"Hipster Scout\" this evening.  Needless to say, the reaction to me dressed this way was quite interesting.  No one understands why I like to dress this way.  Not even me.  Some days I like prep*  Some days I like Hipster Scout.    *By prep, I do not mean Abercrombie, AE, or Aero.  I mean nice clothing.  Like Gap, and Banana Republic and Structure and occasionally Old Navy has some nice clothes.  When I say prep, I mean nice classy looking clothing.  Not faded ripped up shorts and tees with \"Abercrombie\" written across the front in permanent marker.   '),(8288,2876677,'29,April,2004','2004-04-29 00:00:00','   I have decided that the guy I marry  has  to be somewhat athletic. I know it might be tough to get someone to marry me...but at least give me partial athleticism. Just for a second, I was watching this Dad play baseball with his two boys tonight. It is sad to think that I could throw a baseball better than this man. One of his kids was as tall as my waist, and he wound up and lobed it right into his fathers glove like a champ. Pop on the other hand, managed to hit his child 3 times with the baseball...as I was walking by. I had no idea how long they were there, but all I know is that those boys took a beating. Good intentions had become hazardous.  So yeah. Athletic. I don\'t think I want my little boys/girls coming to me all bruised and swollen...\"Honey, what happened??\" \"Oh nothing Mom, we were just playing catch with Dad.\" So first date...you must show me that you can throw a baseball, football, and shoot a three pointer. For the kids, of course. And besides, I would like a little competition.   '),(8289,2876677,'28,April,2004','2004-04-28 00:00:00','   I broke the internet.  I was trying to find out my body fat percentage online.  Every site I went to said error.  Apparently, according to the only one that \"worked\", I have -30.8% body fat.  It said being 5\'11 and 105(roughly) lbs. and having a 28\" waist line was unhealthly.  Hey, I may be a little on the thin side but at least my stomach is flat.  Just kidding, I would rather have a little pudge(VERY LITTLE) than nothing at all.  I swear I eat all the time.  6 meals a day is what my dietician recommended.  6 meals a day and 4 sets of 10 repititions for each muscle group.  Yeah, the eating six times a day is hard enough to fit in, but working out is just not my thing.  So I will wither away and die before I graduate.  So sad.  C\'est la vie.   '),(8290,2876677,'28,April,2004','2004-04-28 00:00:00','   First off, BOOYAH!  I got hired at Staple\'s.  Now on to my post.  I was reading an article on the Presidential race in a periodical I receive and I was suprised at how non-biased it was.  It was amazering.  It gave links to both President Bush and Senator Kerry\'s websites so a voter could be informed by the candidates themselves.  It did not try to drag down either candidate or build either candidate up.  It just emphasized the importance of voting and being politically active.  I was astounded.  Usually every journalist tries to influence his readers nowadays, instead of just presenting the news without his beliefs.  That periodical was  The Collegian  which is put out by BJU.  I was amazed.  Props to BJU for teaching their journalism students to report the issues fair and balanced.  BTW  I am not a journalist and I have an opinion.  Vote for Alan Keyes 2004.  Holla!  *leaves*   '),(8291,2876677,'28,April,2004','2004-04-28 00:00:00','   Today I had an interview with an Italian crime lord.  I have been looking for employment with organized crime but was not expecting to find it at the place I applied.  Actually, I had a list of questions to ask me interviewer when we were talking.  I did not ask them.  It was a one sided story.  He asked me no questions.  I told him no lies.  He was scary.  He made references to \"The Family\", alot of violence, and for some reason he also used the term \"blood\" to describe things of the color red.  He said and I quote \"The customers and the employees are my family when they are inside this store.  If you don\'t take care of a customer, I will take of you.  You see this picture.  These my two little girls.  This store puts food on their table.  You treat the customers bad, they don\'t come back.  They don\'t come back, these doors close.  These doors close, no food for my little girls.  No food for my little girls, I come after you.\"  I am dead serious.  Then he went on about how I need to get an education.  How I shouldn\'t have to live this life forever.  Then some more violent quotes and a little bit of language.  It was fun.  I don\'t think an interviewer has ever sworn before.  The office was dimly lit and added to the effects.  All in all it was just something else to tell my kids.   '),(8292,2876677,'27,April,2004','2004-04-27 00:00:00','   1. First Name: Jenna 2. Were you named after anyone?: No ma\'am 3. School: What? no. 4. Last CD you bought: HA! I don\'t buy CD\'s. I wait til someone gets them as gifts for m...yes I am  that  cheap. I prefer to call myself thrifty though. 5. Last movie you saw in Theatre: Ok. Half of...Ella Enchanted. LONG STORY. Not planned. Don\'t ask. 6. Last movie you rented: Umm...American Outlaws 7. Favorite subject in school: Lunch. or English. No definatly lunch. 8. Least Favorite Subject: MATH 9. Do you actually like math?: no... 10. Who are your friends?: errr...my friends are...my friends. WHAT DO YOU WANT FROM ME? 11. Favorite Actress: humm it would have to be a girl..that acts...oh yes, Dakota Fanning.  12. Favorite Actor: Guys are kinda...I dunno.  13. Favorite Movie: Matilda! Or Iron Will. Oh the joys of Disney. 14. Favorite TV show: The Cosby Show! I am so old school...**hands you jello** 15. How many people are on your buddy list?: 121 16. How long are you in the shower?: 6 min. THATS RIGHT. Usually around 10 though.  17. What\'s your favorite place to go on vacation?: mmm. vacation. It totally doesn\'t matter...If there is a place where my entire family got along, that would become my favorite spot. Otherwise it is my bed. :o) 18. Favorite horror movie: Deep Impact. I know you are thinking that that is not a horror movie....but yes, yes it is. 19. Guys with or without hats: Hats. I like hats. 20. Favorite Food: Mashed potatoes! and popsicles...crap made myself hungry. 21. Favorite color: pinkblueyellow.  22. Worst Fear: Dissapoining everyone. 23. Best feeling in the world: Having someone play with your hair. sounds sketchy but really...it is nice. 24. Worst feeling in the world: Feeling helpless.  What do you think of.... 25. Love at first sight?: can\'t knock it. **shrugs**  26. Abortion: Murder. End. 27. Teenage smoking: smoking in general..I do not condone. **lights one up** I make acceptions for myself, just like my D.A.R.E teacher and model cop. 28. Eating disorder: plague. 29. Rap: DIE. 30. Marilyn Manson: very smart dude. scary though. 31. Boy Bands: gooey. 32. Death: It is appointed to a man once to die, and after this the judgement. 33. Jerry Springer: has no life. 34. Rape: no words to describe how i feel about that. 35. Suicide: sad 36. God: Savior 37. Do you have a job?: I wanna be a personal coach and get rich insecure clients that I can con into paying me way too well. I am evil. No I do not have a job. 38. What school do you go to?: The school of me. I teach me. Its a fun class, you should come. 39. Do you believe in heaven and hell?: Not like it matters if one believes or not, but yes. 40. Do you believe in Angels?: Yep. Angels don\' have wings though. 41. Do you believe in Aliens?: yeah...if you think I am nuts talk to me. 42. Do you believe in yourself?: I can do nothing of myself. 43. DO you want to get married?: Yeah baby 44. What do you do that makes your friends mad?: I try to be annoying. Sometimes. People just think is funny though... 45. Do you have any stuffed animals?: in my closet. 46. Do you have any bad habits?: I like to procrastinate. It is more of a hobbie...no no, a skill. 47. Where is your favorite hang out?: chuckie cheese 48. Pepsi or Coke?: PESSI 49. Leno or Letterman?: Both of them are too ugly and retarded to lok at. 50. MTV or VH1?: ewwy.  51. What are you thinking at this moment?: about SAT\'s 52. Are you a vegetarian?: No. I like Beef Cookies way to much. Pass the cow. 53. Favorite kinds of clothes: the ones that you wear...OOHH OHHH PJ\'s!!! 54. The last TV show you watched: 5 min. of Gilmore Girls with Torg. T.V. is boring. I left.  Have you ever..... 55. Been on a plane?: haha, yeah. 56. Went swimming in the ocean?: I *heart* the ocean. and yes. 57. Cried in public?: yep. In france at a mcdonalds...that was the best one. 58. Let a friend cry on your shoulder?: sure have. haha, most recently because of V. DEATH TO V. 59. Gone to church?: Uh huh. 60. Climbed a tree?: Yup. Best theropy there is...unless youre dustin and cant do anything besides walk...if even that. 61. Watched Dawsons Creek?: haha yeah. no one can act on that show. they drive me nuts. 62. Gone skiing?: YEP!  63. Gone all night with out sleep?: night, day, night. BEAT THAT. 64. Kissed a guy/girl?: COOTIES! >. 65. Ate Sushi: raw nemo? no...hes too cute 66. Wanted to die?: Yeah. Cannot tell a lie. 67. Met a celebrity?: Hello, I  am  a celeb. 68. Met the president?: nope. Good thing to, cuz i would lay down the law about his \'white house\' 69. Thought you were in love?: yep.  Who doesnt love garfeild? 70. Made prank phone calls?: no, I am not that stupid. 71. Snuck out of your house?: umm no. 72. Robbed someones house?: Yeah...my...neighbors on 114 hollis street. HAHA. 73. Broken the law/which ones?: yep. I speededed.  74. Been beaten up?: sure have. black eye at 6 months. 75. Beaten someone up?: I cannot disclose this information 76. Skipped school?: Dude, yes. I SKIPPED TODAY. hows that? 77. Mad enough to pee in your pants: hahahaha...**gets mad** **pees pants** how rediculous. 78. Stabbed in the back by a friend?: yep 79. Stabbed someone in the back?: nope. come be my friend. 80. Had an imaginary friend?: speaking of which...Collin Clement Keenedoola.  81. Cried during a movie?: yeah. because of Dakoda 82. Had a crush on a teacher?: yuck. that would be all...yucky. 83. Been on stage?: sure 84. Cut your own hair?: darn tootin\' 85. Won anything?: yup. everyone hands out trophies. i wanna possicle.   '),(8293,2876677,'26,April,2004','2004-04-26 00:00:00','   1. First Name: Dustin 2. Were you named after anyone?: Not to my knowledge 3. School: Yes please 4. Last CD you bought: I dunno.  Most of my CDs were burnt or gifts. 5. Last movie you saw in Theatre: LOTR: Return of the King 6. Last movie you rented: ...it has been years 7. Favorite subject in school: history 8. Least Favorite Subject: i don\'t really have a  least  favorite 9. Do you actually like math?: yes\'m 10. Who are your friends?: a what? 11. Favorite Actress: dunno 12. Favorite Actor: Don\'t have one 13. Favorite Movie: none 14. Favorite TV show: When I used to watch TV it was  While You Were Out  or  What Not To Wear  15. How many people are on your buddy list?: barely any 16. How long are you in the shower?: 15-30 mins 17. What\'s your favorite place to go on vacation?: the South 18. Favorite horror movie: Don\'t watch them 19. Guys with or without hats: I don\'t care 20. Favorite Food: pizza 21. Favorite color: black 22. Worst Fear: Not to sound macho...but nothing 23. Best feeling in the world: a hug from someone you love 24. Worst feeling in the world: knowing people don\'t care about doing right  What do you think of.... 25. Love at first sight?: doubtful 26. Abortion: murder 27. Teenage smoking: smoking is bad at any age 28. Eating disorder: ...no comment 29. Rap: ugh 30. Marilyn Manson: freak 31. Boy Bands: fad 32. Death: will happen to all of us 33. Jerry Springer: Sick 34. Rape: bad 35. Suicide: your own choice 36. God: the Holy Trinity 37. Do you have a job?: ...maybe soon 38. What school do you go to?: dont stalk me 39. Do you believe in heaven and hell?: yes 40. Do you believe in Angels?: yes 41. Do you believe in Aliens?: ...i dunno 42. Do you believe in yourself?: existentially???  43. DO you want to get married?: yes 44. What do you do that makes your friends mad?: I hope I don\'t make them mad 45. Do you have any stuffed animals?: ...a few 46. Do you have any bad habits?: procrastination 47. Where is your favorite hang out?: ...wherever 48. Pepsi or Coke?: water or grape juice 49. Leno or Letterman?: Ingraham 50. MTV or VH1?: neither 51. What are you thinking at this moment?: i should read a book 52. Are you a vegetarian?: sorta 53. Favorite kinds of clothes: most things from Banana Republic or PacSun...although I made a friend at the mall the other day who said Urban Fashion was a cool place 54. The last TV show you watched: While You Were Out...there was a whole day marathon before I renounced TV  Have you ever..... 55. Been on a plane?: yes 56. Went swimming in the ocean?: yes 57. Cried in public?: yes 58. Let a friend cry on your shoulder?: yes 59. Gone to church?: yes 60. Climbed a tree?: yes 61. Watched Dawsons Creek?: unfortunately...i have seen a part 62. Gone skiing?: no 63. Gone all night with out sleep?: yes 64. Kissed a guy/girl?: no 65. Ate Sushi: no 66. Wanted to die?: yes 67. Met a celebrity?: yes...beverly mitchell...although 7th Heaven actors really aren\'t celebrities 68. Met the president?: unfortunately no 69. Thought you were in love?: yes...i was...and am...with my best friend LYLAS! 70. Made prank phone calls?: yes 71. Snuck out of your house?: yes 72. Robbed someones house?: um...no 73. Broken the law/which ones?: yes...the second law of thermodynamics 74. Been beaten up?: no...almost...but no 75. Beaten someone up?: no 76. Skipped school?: I wish...soon 77. Mad enough to pee in your pants: why would you pee your pants out of anger???? it just doesn\'t make sense 78. Stabbed in the back by a friend?: yes 79. Stabbed someone in the back?: no 80. Had an imaginary friend?: yes 81. Cried during a movie?: no 82. Had a crush on a teacher?: oh yeah...oh wait NO!  that\'s skanky 83. Been on stage?: yes 84. Cut your own hair?: yes 85. Won anything?: a Nobel Peace Prize   '),(8294,2876677,'26,April,2004','2004-04-26 00:00:00','   Someone lied to me.  Someone who has always been seemingly upfront about everything.  It was a very inconsequential thing to lie about.  This person is intelligent enough that I would have thought they would have realized I was intelligent enough to know that they were lying.  Trust=broken.  Also, in other news, today was the day.  Everyday for the past two weeks, those who know me well have been betting on when my mania would end.  Usually it lasts for 1-6 days followed by a long pessimistic depression.  Today, the glass ceiling shattered.  The floor went out too.  I really am upset and it all started with one comment I heard from someone I would not expect to hear it from.  Hopefully, the happy thoughts come back quickly and stay for a long time.  Until then, the glass is not even half-empty.   '),(8295,2876677,'26,April,2004','2004-04-26 00:00:00','   Well, I for one am very glad a certain someone got rid of the background music to his blog. Who\'s with me?  In other news, I have decided that as soon as my kid can walk, they are going to be potty trained. This  will  happen. I had a fun little experience with a 3 year old this past Sunday that I never want repeated. Ever. **sterilizes hands again**  Humm, so did you hear they are making a new 50 dollar bill? Yeah, I know. Looks like I\'ll have a more difficult time printin\' em out. Darn those new hues. This makes my life so much more stressful. How will I pay for my TrippleX clothes? The skinny girls get all the deals. **shrugs** Recently I purchased a parachute, and might I say what a wise investment. They are colorful, durable, a little tight but modest enough, and they serve a dual purpose. I am currently unemployed, but I think modeling could prove to be an excellent career choice for me sometime in the near future. Everyone needs some wholesome Jenna: A picture of perfect health. Perhaps I could enter a pageant for a few bucks? I would need money to spend on a personal coach though. So if you would like to sponsor me speak up. Personal coaches are the best, you pay them to fluff your ego and fill you with a false sense of pride. Whoa, *idea*, I could totally pull that one off. I mean getting paid to sit down, eat and lie? Who could pass that up?....ahhh, so many jobs for the Jenna, so little time.   Speaking of time, look how much I just wasted yours. Comment if you luff me.   '),(8296,2876677,'25,April,2004','2004-04-25 00:00:00','   You kill me.   Every day.  Your decisions hurt.  Watching you walk down that road tears at my heart.  I hate myself for letting you do this to me.  You have never made one right decision since I met you.  Yet I stand at the end of the road.  Waiting.  Standing.  Dying.  Wishing you would let me help you.  For the foolishness of my heart turn around.  I love you.  Please don\'t do this anymore.  I wish I could do something-anything-to turn you around.  I pray to God.  I cry while you ruin your life and the Plan that should have been.  Vulnerant omnia, ultima necat.   '),(8297,2876677,'25,April,2004','2004-04-25 00:00:00','   The only person I hate is me because I know what a horrible person I am.  However, so many people drive me insane.  Stupid girls and their need to be \"loved.\"  They will do almost anything to be in a relationship and to have a guy say \"I love you.\"  Apparently, being insanely jealous is not a sign of unstable insecurity but rather just another way for a guy to show how much he loves a certain girl.  Girls are also way too emotional.  Everything is about being \"fair\" and \"love\".  Wrong!  Life is not fair.  Get over it.  Stupid guys and their need to \"meet people.\"  Yeah...how about you get over yourself and your desire to have sex with every girl you see.  While you are at it, pull your pants up and find a girl who knows what clothes are.  Oh yeah, wear a t-shirt...one with sleeves, I have no desire to see you stomach or your armpits.  Guys...I really just don\'t understand.  I have no desire to do a majority of the \"guy\" things my friends do.  I have no desire to belch or fart on anyone.  I have no desire to be belched or farted on.  Can someone explain to me how guys think?  I also have no desire to hear your problems.  Well, unless you are willing to listen to advice.  If you are going to be dumb again, then don\'t talk to me.  Ever.  Also, if you make a girl who is my friend cry, unless she was in a mood at the time, do not expect me to be chummy with you.  I also have no desire to hear about what you and your girlfriend/boyfriend did last night...or ever.  I have no desire to see you and your girlfriend/boyfriend \"in action\".  Get a room.  If you had a bad relationship experience with someone in high school, which you will, do not go back to them two days later after you spent six hours crying to me about it.  They have not matured.  Obviously, you have not.  I can be mean.  If you are dumb and I say something kind a few times do not expect me to do so indefinitely.  Eventually, I will blow up in your face.  Depending on how dumb the mistake, it may take several months to several years.  Eventually, I will tell you in a less than tactful way how I feel about you and the stupid decisions you repeatedly make.  I make exceptions for those who are unsaved.  That is it.  Otherwise, grow a little.   '),(8298,2876677,'24,April,2004','2004-04-24 00:00:00','   \"One can\'t complain. I have my friends. Someone spoke to me only yesterday.\"-Eeyore  I think Eeyore is an inspiration to us all.  Or at least he should be.  His short, yet deep sayings have an impact on us all.  It may be the fact the I am overtired or that someone at Dunkin Donuts enjoys slipping a little something extra in my croissant but everything is really funny...and blurry.  The room hasn\'t stopped spinning since I got home.   '),(8299,2876677,'23,April,2004','2004-04-23 00:00:00','   Some people think evil is on of the most compelling arguments against the existence of God.  Actually, it is one of the best arguments in favor of the existence of God.  What is the definition of \"evil?\"  Merriam-Webster defines \"evil\" as \"morally bad or wrong; wicked.\"  Evil means that things are not as they should be.  They have departed from good.  That would mean there is a standard for goodness.  Scholar C.S. Lewis said \"I wouldn\'t have known what crooked was unless I knew what straight was.\"  In order for somebody to complain about the problem of evil, there must be some transcendent standard of good that it departs from.  That can only be true if God exists.  Can we have moral laws without God?  Atheism, or materialism, teaches that only material things exist.  Moral laws are immaterial and they exist.  Good and evil are both immaterial.  Neither is a physical object one can empircally observe.  Sidenote:  In order to be an atheist, one can neither love nor hate.  In essence, one can neither believe nor disbelieve anything.  \"Conciousness\" is just random firings of synapses causing action.  Everything happens by chance and nothing can be predicted because everything happened by chance.  We have no free will.  If my synapses fire and I shoot you, it is not my fault.  Technically, I don\'t exist because \"I\" am just nerve cells and chemicals.  The death penalty has no purpose because it is not really a penalty.  There is no suffering.  There is nothing to fear in death.  The collection of nerve cells stop firing.  That is all.  PS  This blog is not real.  It is a figment of your imagination.  \"I\" use the term imagination loosely since you do not really \"imagine.\"  I have no free will and had no choice but to author this blog.  So if \"you\" disagree, I would apologize but I cannot choose to.  Random chance has prohibited me from it.   '),(8300,2876677,'22,April,2004','2004-04-22 00:00:00','   I had really good intentions last night.  My plan was to go out, come back, read my book, and write a report.  I went out and came back...  I usually don\'t go to bed until 2 and am still wide awake when I go to bed but last night for some reason I was wicked tired and couldn\'t concentrate on reading.  So I crashed...and today I suffer for it.  I did manage to get a lot of good books on the effects of nuclear radiation.  ...I skimmed one.  Whopops.  What can ya do?  Well I should go.  Last night was fun even though I didn\'t do what I was supposed to.  Actually, it was probably fun  because  I didn\'t do what I was supposed to.  PacSun was fun...to say the least.  What a good time...and a memory to tell my kids.  I hope my kids won\'t do wierd things for money.   '),(8301,2876677,'21,April,2004','2004-04-21 00:00:00','   A well-known local speaker gave a little speech on acceptance.  It got me thinking.  I used to change to be accepted.  Now my goal is not for my group to accept me, but to find a group that accepts me.  Clarification: I have decided not to conform to be accepted, but to find a group that accepts \"me.\"  Yeah.  Well anyways, today is definitely a manic day.  I was as bit crazy during study hall.  It was fun.  I think I am going to go to the library and put this energy toward my research paper.  Bye.   '),(8302,2876677,'20,April,2004','2004-04-20 00:00:00','   There is alot of food here.  Actually, I have been eating all day.  Someone actually commented on it.  She said, \"Everytime I have seen you today, you have been putting food to your mouth.\"  So true.  What\'s the problem with that?  Breakfast-Cereal Bar.  Brunch-Croissant.  Lunch-Pizza(quarter).  Early Afternoon Snack-Animal Crackers.  Mid-Afternoon Snack-Ice Cream.  Late Afternoon Snack(Now)-Pizza(whole).  I can\'t wait for dinner.  I\'m a glutton.  So what?   '),(8303,2876677,'20,April,2004','2004-04-20 00:00:00','   No food...in our house. No food, no food. no food.  I am so hungry.  I am also a charity case, so bring me some eats.   '),(8304,2876677,'20,April,2004','2004-04-20 00:00:00','   Well, the race to beat Jenna for best looking blog has been a long one.  I joke of course.  However, in all seriousness it has been quite irritating doing all the little html thingies on my own just guessing what they are.  Sunday night I probably spent two hours getting the banners to look right.  Then last night I spent around an hour adding a \"Blog\'s I Read\" section and a counter to the bottom of the page.  Tonight I added a \"Contact Me\" link and for some reason the font would not change so if it wants to be a thorn in my side then I will just leave it.  Well, not for long, because it is driving me insane.  I just need to figure out how to fix it.  I was told I had no life because I spent so much time on my blog.  Just so you know, the average American watches at least three hours of TV a day.  I do not.  Is it worse to spend three hours watching TV or typing and problem solving?  Something to think about.  Also several people have told me recently I should work out.  I say to you, bah humbug.  I need to gain weight first.  Which I am trying very hard to do.  So after I gain 40-60 pounds...I  may  consider serious physical activity on a daily basis.  Don\'t quote me on that.   '),(8305,2876677,'19,April,2004','2004-04-19 00:00:00','   I take no part in, and do not endorse the advertisement of the \"Cat and Girl\" comic strip. My co-blogger seems to think it is...funny. However I, along with the rest of the population do not understand this type of humor.  I claim no responsibility for the so called \'humor\' you will see on this blog in the form of comic strips.   *Edited for spelling by Site Administrator*   '),(8306,2876677,'18,April,2004','2004-04-18 00:00:00','   The term \"gay\" is widely overused/misused and does not sound anywhere near as intelligent as some of its alternatives.  Also, objects and events do not have a sexuality or personality.  If used in reference to a person, it does not usually have any meaning nowaday.  If used in remark to sexual preference, generally the person referred to is not \"gay\".  Actually the term \"gay\", does not actually have anything to do with one\'s sexuality, rather it refers to a joyous or happy mood, and was stolen by the homosexuals to make them sound more personable and well happy.  Some useful alternatives are  alert, animate, animated, blithe, blithesome, bouncy, brash, carefree, cheerful, cheery, chipper, chirpy, confident, convivial, festive, forward, frivolous, frolicsome, fun-loving, gamesome, glad, gleeful, hilarious, insouciant, jocund, jolly, jovial, joyful, joyous, keen, light-hearted, lively, merry, mirthful, playful, presuming, pushy, rollicking, self-assertive, sparkling, spirited, sportive, sprightly, sunny, vivacious, wild, zippy  if referring to one\'s mood.  If referring to sexuality, try  homosexual, homophile, inverted, lesbian, queer .  If in need of a derogatory term try...well actually I have decided it would not be good to help you be unkind.   '),(8307,2876677,'17,April,2004','2004-04-17 00:00:00','   urlLink Cat and Girl  is quite possibly the funniest site ever.  I love it.  Although...it takes someone who is very bourgeois to understand it.  I am getting the best bumper sticker ever from them.  YAY.   '),(8308,2876677,'17,April,2004','2004-04-17 00:00:00','   I was listening to Kiss 108 yesterday and I heard this awesome love song. \"All of You is more than enough for all of me For every thirst and every need You satisfy me with Your love And all I have in You is more than enough  You are my supply My breath of life And still more awesome than I know You are my reward worth living for And still more awesome than I know  All of You is more than enough for all of me For every thirst and every need You satisfy me with Your love And all I have in You is more than enough  You’re my sacrifice  Of greatest price  And still more awesome than I know  You are everything You\'re my coming king And still more awesome than I know   More than all I want More than all I need You are more than enough for me More than all I know More than all I can say You are more than enough for me\"  Just kidding, I don\'t listen to secular radio (unless you count laura ingraham).  Actually, I heard this song in a church.  The people in church thought it was about Jesus.  If it was on the radio(I don\'t know if it was), people would think boy/girl.  Who is to say which is the right interpretation.  All I am going to say is that a song to God/about God should at least mention him.  This is not winning anyone to the Lord(which is supposedly the purpose of \"Christian\" artists getting their music played on MTV and such like but I would like to point out that music is not for evangelization but for worship.)  Also this song is not evangelizing anyone.  Its a boy/girl love song.  Christians fall for so much dumb stuff.  Time magazine said \"The message of  Friends , in other words, is that there is no normal anymore and that Americans--at least the plurality needed to make a sitcom No. 1--accept that.\"  So many Christians watch  Friends  because its \"funny\" or \"entertaining\" but they \"don\'t support it\'s beliefs\".  By watching a show, you say you are fine with what goes on in it.  Time realizes that.  Why can\'t Christians?   '),(8309,2876677,'17,April,2004','2004-04-17 00:00:00','   Well Friday on Jenna\'s roof, the outdoors were beautiful and I thought I should sleep outside tomorrow night.  Then Saturday night, Jenna says \"torg and i are sleeping outside tonight\" apparently she also had that idea out on her roof.  So I thought well if Jenna is doing it too, then it must be a good idea.  Wrong.  Note to self:  If Jenna ever says \"I am doing...\", don\'t do it.  Sleeping outside in April in dumb, and now I am sick.   '),(8310,2876677,'17,April,2004','2004-04-17 00:00:00','   Well last night was wicked fun.  I was bored so I called up a friend and had that friend call a friend while I called another friend to see what we could do.  Fortunately, no one had any plans (probably due to the lack of a social life...at least in my case).  Well Emi couldn\'t go anywhere because of family game night so we had a guy\'s night which I suggest we do regularly.  I had fun and I think they did.  Stu and I went over to Jenna\'s.  I drove by her house once cuz I didn\'t want to be first and then came back because his station wagon was there.  So I go in and he isn\'t there!  It is his brother.  Apparently they don\'t communicate and did not realize they were going to the same place at the same time...  Well anyways I was the first to arrive so Jenna, the torg and I played pool.  It was a good thing no one was watching.  Then Stu showed up and we walked to Dunkin Donuts.  Then we walked back.  (We got food while we were there.)  Then we went on to Jenna\'s roof and Stu gave an astronomy lesson.  Very informative.  Then we played pool.  I don\'t remember who won or if anyone won.  I think we may have stopped or something.  Then we played go fish.  Good times.  Then stu left and we took some personality quizes.  Then I left and got home late.  Like breaking the law late.  After midnight late.  Mum didn\'t care I was late which was cool \'cause usually every second counts after I am late.  Then this morning she wakes me up at some ungodly hour and asks if I am going on the bike hike.  I am like I was out past midnight, I have a wicked hangover and YOU WANT TO KNOW IF I AM GOING ON THE BIKE HIKE!!!  I think she realized that was a no.  Then my car broke today.  So sad.  It is really broked.  It is at the garage or something.  Meanwhile I am being nagged to buy a car but I am waiting till I find the right car.  Oh wait I did...and they said no you dont need a new car the one you have runs fine.  *scowls*   '),(8311,2876677,'16,April,2004','2004-04-16 00:00:00','   Most people like you feel at ease in situations requiring verbal ability, attention to detail, and linear, analytical ability. Whether you know it or not, you are a much stronger written communicator than many, able to get your ideas across better than others.  It\'s also likely that you are methodical and efficient at many things that you do. You could also be good at math, particularly algebra, which is based on very strict rules that make sense to your logical mind.   You\'ve got a healthy love of life and a killer instinct. Chances are you hate rules, and don\'t plan on getting old. You\'re flexible and friendly on top of being innovative and smart. In short, people respect you.  You\'re extroverted and love being in the spotlight. Your mantra is \"work hard, play hard\" — you like to live life to the limit.  Your intelligence seems to be the trait that you most value in yourself of the five main traits that affect confidence. You also seem to appreciate this characteristic in others.   '),(8312,2876677,'15,April,2004','2004-04-15 00:00:00','   Hey!  Its summer(almost).  You know what that means right?  Yup.  Take off your clothes.  It is way too uncomfortable to wear at least loose pants and modest tee-shirt.  So guys take off yours shirts and throw on some long shorts that you wear so low it looks like you are pants.  Ladies, put on some short-shorts, or some really tight capris that are really low and a really tight top that show your stomach and has a low neckline.  Its gonna be a party.  Spring break.  Woo-hoo!   '),(8313,2876677,'15,April,2004','2004-04-15 00:00:00','   Wow I love my license.  Freedom is good.  Like I got to go to Worcester to FACA then to Target then to Dunkin Donuts then back to FACA then to Pizza Hut in Leominster.  Good times.  I managed to be back for curfew.  For once...  I was worried \'cause I was cutting it close but I made it.  People at FACA are so wicked friendly.  You have no idea how many new people I met.  And the people I have met before all come up and talk to me.  I haven\'t been there since like January and they all recognize me.  The youth pastor thought I was from Texas.  I have no idea why...  Anyways...the reason I went wasn\'t there...so sad...  I wish things were different...my life is going to be an interesting one...i am definitely determined to make an impact...negatively or positively is something only time will tell.  i want to get married...tommorrow...i already have a wife picked out...and a house...and a job...   '),(8314,2876677,'15,April,2004','2004-04-15 00:00:00','   ...my dreams lead to scurvy   '),(8315,2876677,'14,April,2004','2004-04-14 00:00:00','   I am so tired.  I need a vacation.  Or rest.  Or alone time.  All three would be nice.  I have no idea what is going on.    I am so confused and tired of it all.  I can\'t think straight.  I gave up caring so long ago.  Unfortunately, it has turned into widespread apathy...  Everyone is so messed up.  I did have one minor epiphany today.  I used to be that way.  Oh  Whose Line?  was a  little innapropriate last night...  DERR!!!  It always is.  Sometimes(rarely) just a little, most of the time it is just filled with perverseness.  Why even put it on that channel?  Hello???  It is going to be bad.  Think about it.  Wow I was a retard.  Unfortunately a majority of people still are...   '),(8316,2876677,'14,April,2004','2004-04-14 00:00:00','   tonight...mmm...wicked excited...i hope things go wonderfully or awfully...that way i will be more sure...because right now i definitely am confused...  BTW if you are going to comment, be man enough to use your real name.  or just don\'t comment at all.  if what you are saying cannot be ascribed to you, don\'t say it.   '),(8317,2876677,'14,April,2004','2004-04-14 00:00:00','   my mom is right over my shoulder...slightly irritating...greatly impeding my creative process...it is not very easy to think when someone is staring at your work.  i love you jenna.  oh wait.  never mind i must have been confuzzled.  i love you my reader.  which at the time is me.  hi self.  leave a comment.  any comment at all would be good.   '),(8318,2876677,'12,April,2004','2004-04-12 00:00:00','   New Blog! Come visit me on my new site...http://rosey287.blogspot.com/  I\'ll still be writing in here, because Dustin is cool, and I wanna be a part of his...blogging. Anyway feel free to click and read. Have fun with the commenting! :op  **cough** my blog is better...and it has **gasp** colors!!!   '),(8319,2876677,'07,April,2004','2004-04-07 00:00:00','   Age as a reflection of emotionally impactful events:  Wow!  Alot happened this weekend.  Not like \"I was wicked busy this weekend\" (I was though) but like several big things happened.  I had a while on the trip home for introspection.  I realized (or at least now believe) that it is not the passing of time itself that makes one older, but rather the events that occur with the passing of time that change us.  I am probably not the first to have this revelation but it was at least original to me.  I had a lot of major trials this weekend. A few I passed.  Many I failed.  I have learned from all of them though.  Where I will go depends on if I apply what I learned.   '),(8320,2876677,'07,April,2004','2004-04-07 00:00:00','   Wow.  The past two days have been crazy.  We will stick to school(good crazy) and not talk about the bad crazy stuff.  Word of advice.  Never sneak out of your home through your window.  Bad idea kids.  School!!!  OK The past two days I have been a servant in 5th grade and 2nd Grade.  Both have cool teachers.  Both have generally cool kids.  I am kidnapping Luke, a second grader, the coolest little kid in the world.  I said to Mrs. Palacios \"That kid is so cool.  I wish he were my little brother.\"  She said \"I KNOW!  He is the cutest little kid ever.  I want a son just like him.\"  Me too.  I felt bad for neglecting the other kids a little but he was cool.  There was this one kid, no names, who was clingy and annoying all day.  I thought it was pretty cool that no matter what I did I was cool just because I was 17.  Respect is nice.  I like little kids.  They say so much random stuff all day.  I am excited about being a parent.   '),(8321,2876677,'06,April,2004','2004-04-06 00:00:00','   Kids today (and the rare adults that act like kids) are just...off. Like my tennis coach, who is 53 going on 16, and uses phrases like, \"duuuuude\", \"sweet man\" and \"whoooaa, awesome!\" He shafted Evan and I today. Didn\'t show up to the lesson. I wasted time, gas, and money. He should seriously think about growing up, this is not the only time it has happened. Jerkmo.  Anyway about \'kids\'. There is something wrong with at least half of the kids ages 5-15, then after that the percentage increases. I was walking through a store today...and this kid was walking towards me next to his mother. Now this kid was \'normal looking\', about 14. And by normal I mean black sweatshirt, knee-waisted oversized pants, and in dire need of a haircut and bath (I didn\'t smell him from where I was, appearances can be a great indication however.) So as he and his dear mother were walking, he started to growl. Growl like a rabid pit bull. I was like, whaaat? --as I move a little faster.  He was sane, if that\'s what you\'re thinking, but yes, he was still in fact, growling. I was sketched out and tried to forget it. But then I realized I needed to go back the other way to find my mum, and as I turned around, I caught another glimpse at the nice young teenage boy. He had taken off his ratty baseball hat, wound up, and smacked his mother in the back of the head with it. Hard. And then again. I got scared and ran away. That is the kind of kid who you can share a cupcake with at lunch, then at the next meal he pulls a gun on ya.  I blame it on TV. And drugs. Well, weed and power rangers/pokemon.    Let me hear ya holla, street balla! (leave a comment, basketball player)   '),(8322,2876677,'05,April,2004','2004-04-05 00:00:00','   I have been reading this really good blog for quite a while now and it has really got me thinking.  The name of it is  urlLink Hot Abercrombie Chick  which is funny because stereotypically it would be thought of as a dumb liberal blonde\'s blog, but it is a conservative college girl who is very articulate.  I love reading what she has to say.  I need to get more involved politically.  I can\'t vote but I can speak.  People need to here the issues.  Correctly. The media is leftist.  The public schools are biased toward liberalism.  The government is socialist.  Kids are being raised to not read.  Just watch TV and listen to what they have to say.  I mean if they made it to TV, they must be right.  Right???   If anyone else wants to get involved email me.  I am going to think of ways to get out there.  Make the issue known.  Intelligently.  Too often, conservatives just go out and get made to look like idiots by people who confuse themselves.  *Speaks out*   '),(8323,2876677,'04,April,2004','2004-04-04 00:00:00','   I am totally unstable.  Every day I am totally unhinged but some I am bitter and upset with everything.  Often for no reason.  The next day everything is crazy awesome.  No matter what.  I could get hit by a car and be excited about it.  I am wierd...  Comments anyone?  Maybe even a rare elusive entry from Jenna???   '),(8324,2876677,'04,April,2004','2004-04-04 00:00:00','   No one \"won\" or \"lost\".  That makes some feel less important than others.  Anyways also I was right on some points and and my co-bloggstress was right on others.  We will not argue about who was right about which points.  Just know that I love my Deep Dish.  Have a good night.  Don\'t break another bed.  That was embarrassing last time you did it.  I felt your fall at my house.  Sleep in.  Catch up on what you missed.  I was truly frightened this morning.  Much love...   '),(8325,2876677,'03,April,2004','2004-04-03 00:00:00','   *I sorry*  1.You are right.  I will be more clear next time. 2. I am glad you don\'t have an overactive bladder. 3. I still believe you said Walmart.  Although I may be wrong. 4. YOU DO NOT OWN MCDONALD\'S!!! 5. I don\'t care that you can beat me up.  Actually it is attractive. 6. I can\'t believe you don\'t know what Crispy Creme is. We are SO taking a road trip. 7. Did you go back and change what you wrote? 8. Really, it isn\'t?  *is stunned*  OWW GET AWAY FROM ME WITH THE TASER!!!  ~am I allowed to eat the pizza?   '),(8326,2876677,'03,April,2004','2004-04-03 00:00:00','   Let\'s run this, one LAST time.  1. You implied it. Be more careful next time. 2. Deal with that fact that I can hold my liquid. 3. You lie, I said I liked Big and Tall. Don\'t put false words in my mouth. 4. I own McDonalds. 5. Get over the fact that I could beat you up. 6. What is this Crispy Cream? I do not lie. 7. I SAID IN THAT PARTICULAR SENTENCE. Next time, try reading. 8. YOUR WEIGHT ISNT ON YOUR LICENSE.   -lock yourself in a room with a pizza. dont come out. ever.   '),(8327,2876677,'03,April,2004','2004-04-03 00:00:00','   Rebuttal- 1. Since when did I say you drank only rum, idiot.  I mentioned both rum and the bubbly.  I do not recall implicitly stating  rum is synonymous with bubbly . 2. You are always drunk.  You have gotten used to it. 3. I vividly recall you saying \"I love Walmart\".  \"It is an awesome store.\"blah blah blah 4. McDonald\'s does not deliver and last time I checked your name was not Jim Cantalupo. 5. You still beat men. 6.  LIAR!!  I know you have a Crispy Creme every minute on the minute. 7.  All as a noun- He gave his all.  as a pronoun- You are all I need.  as an adjective-   All the windows are open  as an adverb- a room painted all white  8. You lied about your weight on your license.  So how can anything else you say be trusted?  BTW I did not speak, I typed. Much love.   '),(8328,2876677,'03,April,2004','2004-04-03 00:00:00','   Let me throw you back into reality.   1. Rum is not bubbly. Idiot. 2. If I were drunk at the time of my post: muy tipong woiulf l0oke lifae thids 3. I do not frequent \'Wal-Mart\'.  4. As I have previously stated, I own McDonald\'s. They come to me. 5. I do not have a husband to beat on. 6. I have never been to Crispy Cream. 7. The \"all\" in the sentence, \"The boys are all tired.\" is an ADJECTIVE. Sit down, I am right. And good luck on your SAT\'s.  8. My statements are verifiable, yours are ridiculous.   -Just don\'t speak anymore Dustin, you\'re embarrassing yourself.   '),(8329,2876677,'03,April,2004','2004-04-03 00:00:00','   Sorry Fred.  I didn\'t mean to misquote you but I have just recieved info that it is actually, \"Der Wille Zur Macht\".  I guess The Will To Power makes a little more sense than To Will The Power.  Well it is too late now and I dunno I guess it has a nice ring to it so I am leaving it.   '),(8330,2876677,'03,April,2004','2004-04-03 00:00:00','   Not to throw mud or lower myself to my un-named co-blogger\'s standard BUT I ask you \"Can you really believe this girl?\"  For instance, DID YOU KNOW that this young lady starts out the day with rum.  Jcarlos287: breakin\'out the bubbly triquetta16: i wanted to talk to your parents tonite about the way you dress but apparently you all are skipping church to get drunk Jcarlos287: alright patty...i need to go get drunk now triquetta16: just dont drive home like that Jcarlos287: we have a designated driver patty Jcarlos287: and some people find my fat hanging over my pants attractive Jcarlos287: so whatever  Can you really trust a girl like this?  How do you know she wasn\'t drunk when she wrote her post?  Also on the obeisity issue, drop a pant size or ten before you try to talk about my fat.  Seriously,  if you have never met \"her\", I would encourage you not too.  If you would like to see \"her\" go to Wal-mart.  She will be in the Plus size section(Large sizes  for girls who need to drop a few)...or McDonald\'s.  It is not a sight you will forget.  She also is a husband batterer.  She can become extremely violent if you try to take her Crispy Creme.  She denies this of course.  The first step in the healing process is admitting you have a problem.  Also \"unnamed girl\", all is a noun, pronoun, adjective and adverb.  Leave comments about your feelings toward my co-blogstress.   '),(8331,2876677,'03,April,2004','2004-04-03 00:00:00','   Readers:  Should we be worried? My co-blogger (and I won\'t mention any names for privacy sake) is starting to develope early stages of domestic violence/condecending attitude/uncoltrolled temper/feirce rage without instigation/adult male immaturity. On top of that, he eats too much. Example:  Jcarlos287: why do you threaten me so? triquetta16: i dont recall threating triquetta16: i recollect an imperative command  and later...  triquetta16: i want something to read Jcarlos287: grab one of those ten pound books you love triquetta16: *grabs book* triquetta16: *throws at jenna* triquetta16: *walks away*   Can we help him? Or is it too late... We need massive amounts of time to change his entire personality and put him on a diet. Is this even possible though...I dunno. He loves his PIE. This is too much for me to deal with right now...please help. **leaves**   '),(8332,2876677,'03,April,2004','2004-04-03 00:00:00','   Sidetrack:  I like being able to blog from anywhere.  It is wicked convenient.  I have found a spectator sport I enjoy spectating!! X-games.  I love the skateborders, and bicyclists.  I think the stuff they can do is crazy cool.  It is so much cooler than shooting a ball in a basket.  I mean throwing your skateboard and jumping and flipping and landing on it requires crazy skill.  I am definitely going to Germany to watch the German version of the Xgames-Circle of Balance.  It is crazy how good these kids are.   '),(8333,2876677,'02,April,2004','2004-04-02 00:00:00','   Well I really have some interesting acquaintances.  As I was talking to one, he suggested that monogamy is a marriage ruiner.  WOAH!  Stop the conversation.  I was like WHAT??  Apparently, men were just not made for one spouse and causing them to be with only one woman is going to cause psychological problems and will cause a rift in the marriage because the man\'s needs are not being met.  HOLD UP!  I said so You think we should not get married and just sleep with whoever.  He said No no no.  There are certain benefits to getting married but a man needs a woman on the side.  I said So you advocate for polygamy?  Him-No.  A man needs one steady woman and a woman on the side with whom he has no emotional attachment that changes on a regular basis.  Me-SAY WHAT!  So you think there is just this pool of women floating out there that don\'t want to get married?  They just enjoy being tossed from man to man.  Him-No.  You are totally distorting what I am saying.  They don\'t have to be unmarried.  They could have a husband and a boyfriend too.  Me- So you wouldn\'t mind if your wife cheated on you?  Him-Um...err...I guess not.  Me-YOU GUESS NOT!!!  You were not even thinking about the emotional side of a relationship are you?  YOUR PLAN DOES NOT WORK.  IDIOT!(I didn\'t say idiot.  It isn\'t generally a debate winner.)  Him-Well a man being constricted to one woman doesn\'t work.  It causes resentment toward your spouse.  Me-How?  What did she do?  Him-Well not being able to  express [emphasis mine] yourself causes emotional blockage.  Me-What the heck is emotional blockage?  Him-Well being with only one woman causes emotional blockage.  So a man needs--Me(cuts him off)-I did not ask for the cause.  I asked for a definition.  Stop repeating yourself.  Get back to me when your views are better defined.(end conversation)  GRR!!!  I hate idiots.  They are just so prevalent.  And they have no idea what they are saying.  Just dumb excuses for acting on base impulses.   '),(8334,2876677,'02,April,2004','2004-04-02 00:00:00','   Chow-food Ciao-hello/goodbye   '),(8335,2876677,'02,April,2004','2004-04-02 00:00:00','   Wow.  ACT is about 100 times easier than SAT.  It all depends on POV I guess.  Some people I tested with did not think the time limit on the ACT was long enough.  It was short and I did take most of the time in each section but the questions were much easier.  The SAT is very wordy and also very biased.  Also, on the SAT, I always had the tendency to answer the question from a conservative POV but the I realized the author is liberal and went with the liberal interpretation of a passage.  ACT had very little bias in my opinion(or at least a conservative one) and it was a lot more in the language of the common people.  I mean here and there a question would try to trip me up but if I read the question it wasn\'t that hard.  Math was a bit different.  I forgot a calculator so I had to do all the work the long way so I just guessed on the last ten when I only had a minute left.  I definitely \"spanked the ACT\" though.  I can\'t wait to get my score.   '),(8336,2876677,'02,April,2004','2004-04-02 00:00:00','   Dustin, here is the lengthy blog you wanted. So deal with it, you asked.  Sooo boreded. Someone needs to come over my house and play with me. I had some friends, once, but they cancelled. NOTE: This is not in any way appreciated. When you make plans to do something with a stinkin\' person, do it. Otherwise, don\'t commit kids- just say no.  And   how   do people not know what Plinko is? What is this madness? Its PLINKO. Every good American citizen should know this. yes, they should. It is a ridiculous, stupid, awesome game. I *heart* plinko.   I seriously have nothing to do, **eats a spoonful of frosting**  and we have no food in our house. It is sad. People come over and offer us food stamps. Well I mean they would...if they came over...instead if canceling...and no, I am not bitter.  Ok so I was just told that I \"jive.\" Umm..what? I have absolutly no clue what that means, or how I do it. So if ya know, give me a holla in the comments, yo.  By the way, when I am bored I am not funny. You do the math.   '),(8337,2876677,'01,April,2004','2004-04-01 00:00:00','   I can barely breathe. I\'ve lost my will.  This morning school was canceled because of flooding.  I thought Alright.  Then I went back to bed.  At 10:30 Miss Schnakenberg called.  She wanted me to come do banquet stuff.  I just decided to go with the flow.  I definitely could have used my day off for better things(ACT, homeschool, research paper...) but whatever.  I went and had a pretty non-OK day.  I started out with feelingsof general apathy but that progressed to emptiness.  I really am sick of the façade.  I don\'t care about the banquet.  I don\'t really care about the school.  Its not like I have any friends at school.  I don\'t hang out with anyone from school.  I just go.  Everyone in my class wants me to go to the banquet.  Why?  Why should I go?  I don\'t care.  I really wish I went to public school.  The moral atmosphere is just the same.  Sure maybe some actions are subdued but a worldly attitude is still at the core of my school\'s heart.  They may not have sex in the closets at school but but if these same kids went to public they would not be any different than anyone else.  If my whole class went to public school there is not one person in my current class I would talk to.  I am pretty sure the feeling is the same.  I don\'t really care.  I am almost out.   '),(8338,2876677,'01,April,2004','2004-04-01 00:00:00','   I went shopping.  For food.  It was fun.  I bought alot.  Actually, I wandered around the produce section for a while looking at the variations of vegetables.  I had no idea.  I just walked around looking dumb wondering \"What am I going to do?\"  Then Miss Vasel showed up.  What a blessing.  She helped me pick out onions and peppers and salad mix.  Then we went shopping for other stuff.  We picked the prettiest boxes and bottles.  It is the only way to shop.  Fortunately, the good stuff was the cheap stuff.  I saved $13.00 with my Shaw\'s card.  That was exciting.  I pretty much emptied out every section I went to.  There is no more chicken, pineapple, or oriental salad mix.  So you should wait until tommorrow morning to go to Shaw\'s.   '),(8339,2876677,'30,May,2004','2004-05-30 00:00:00','   I am buying two computers soon.  Well, one by the end of the summer and then one by next summer.  One is an 1.25Ghz PowerPC G4 eMac with 1 GB of 1GB SDRAM DDR333, a 17\" CRT screen with a Superdrive(DVD-R/CD-RW), 80 GB ultra ATA hard drive, and an Airport Extreme card.  Then next summer, for college, I am getting a 14\" iBook.  Sooo excited.  Nerdy too.  But whatever.  C\'est la vie.   '),(8340,2876677,'29,May,2004','2004-05-29 00:00:00','   triquetta16: Memorial Day is a celebration of murderers who were brainwashed to think that their cause was worth killing other people over triquetta16: so I will not be watching any parades or planting any flags triquetta16: and then the people are brainwashed into thinking these people are heroes triquetta16: jumping on a grenade to save people is heroic, killing people to promote capitalism and democracy is not Jcarlos287: you are against all war? triquetta16: unless Jesus Christ comes down and says go to war against this country then yes I am against war Jcarlos287: so, back in the day when the israelites fought....that was wrong? triquetta16: we made be blessed but America is not God\'s chosen nation and God has not told us that it was our purpose to go into other countries and kill them to promote freedom triquetta16: no.  israel was told directly by God, do this, exactly this way Jcarlos287: uh huh triquetta16: Israel was fulfilling God\'s command to get rid of filthy sinful nations triquetta16: we are just killing people who might be a threat to our defense Jcarlos287: question: If you walked in your house and someone was attacking someone you loved and the only way to stop them was killing them, what would you do? triquetta16: and then in retrospect we say \"oh look how we helped all those civilians\" but at the expense of hundreds of thousands of lives triquetta16: i would not kill them Jcarlos287: liar triquetta16: i would not Jcarlos287: remind me never ever to chill with you in a bad neighborhood triquetta16: i would get in there way and try to stop them...but someone loves that person too and who am i to kill him over my loved one triquetta16: God loves them both equally Jcarlos287: like i said, make sure your wife and family know that triquetta16: taser, and tranq. gun would stop him just as efficiently as butcher knife or real gun triquetta16: without killing him Jcarlos287: oh because everyone had a tazor or a tranq. gun on hand triquetta16: well intead of buying one that uses bullets by one that uses tranq. darts Jcarlos287: pretend you are in a situation where it was either the attacker that dieds, or your loved one triquetta16: it is not like tazers are that expensive either triquetta16: i am not God triquetta16: i refuse to play God Jcarlos287: i refuse to watch while someone attacks my child triquetta16: however if my loved one was saved I would chose the attacker because chances are if I killed him, I would be sending him to hell Jcarlos287: Guess what? God put you in control of your childs life, to protect it no matter what triquetta16: not something i want to spend the rest of my life thinking about triquetta16: (forcefully sending someone to hell) triquetta16: what do you think about that triquetta16: would you choose a saved loved one or an unsaved attacker? Jcarlos287: i would protect the thing God had entrusted me to triquetta16: and send someone to hell to save the physical body of someone who would be going to Jesus\' side Jcarlos287: number one: who is to say my child is saved? Number two: who is to say this attacker will be saved? triquetta16: this is all ethical debate. triquetta16: i don\'t believe in ethics.  i believe in morals. triquetta16: Thou shalt not kill triquetta16: one of my moral rules...given to me by God triquetta16: i do not feel the need to argue situational ethics since i am a moral absolutist.  i had thought you were too Jcarlos287: eth·ic    ( P )  Pronunciation Key  (thk) n. A set of principles of right conduct.  A theory or a system of moral values Jcarlos287: the two are practically synonymous triquetta16: Situational ethics refers to a particular view of ethics, in which absolute standards are considered less important than the requirements of a particular situation. triquetta16: The standards used may, therefore, vary from one situation to another, and may even contradict one another. This view of ethics is similar to moral relativism, and is contradictory to moral universalism, and moral absolutism.  triquetta16: Moral absolutism—the belief or theory that there are absolute standards against which moral questions can be judged—suggests that morals are not determined by societal or situational influences.  triquetta16: not one of the ten commandments has an instance when breaking it is acceptable Jcarlos287: do you believe in capital punishment? triquetta16: the two are polar opposites triquetta16: if God smites a person for murder then yes.  however, the executioner is just as guilty of murder as the condemned Jcarlos287: eye for an eye Jcarlos287: i believe that is in the Bible Jcarlos287: no? Jcarlos287: God smites the person with an executioner triquetta16: Ye have heard that it hath been said, An eye for an eye, and a tooth for a tooth:   But I say unto you, That ye resist not evil: but whosoever shall smite thee on thy right cheek, turn to him the other also.   triquetta16: direct quote from jesus triquetta16: Ye have heard that it hath been said, Thou shalt love thy neighbour, and hate thine enemy.   But I say unto you, Love your enemies, bless them that curse you, do good to them that hate you, and pray for them which despitefully use you, and persecute you;  That ye may be the children of your Father which is in heaven: for he maketh his sun to rise on the evil and on the good, and sendeth rain on the just and on the unjust.  triquetta16: resist not evil.  do not kill someone trying to kill you.  do not cower.  but do not kill Jcarlos287: what if in the act of \"not cowering\" you killed the attacker? triquetta16: i mean by \"do not cower\" stand firm, tall and fearless.  show them you have nothing to fear in death but that you are unwilling to kill Jcarlos287: you can do that in less then 10 seconds? what are you going to do? make a speech? triquetta16: no.  posture resistance and the look in a person\'s eyes can say alot about fear Jcarlos287: there is a man attacking your wife. he is 6\'5 and jacked. you are in an ally. you think youre gonna stop him? triquetta16: sending someone to hell is irreversible, by not killing them i am simply giving God the chance to lead them to him triquetta16: no. i hope she would stand firm with me and we would die together Jcarlos287: did you know God gives up on people? he gives them over to their wickedness and no longer tries to reach them triquetta16: FYI, Ted Bundy, serial-rapist/murderer is now saved and in heaven Jcarlos287: good for teddy triquetta16: he raped tortured and killed almost a hundred women.  he was caught.  Jesus did not give up on him.  One of the victim\'s families kept witnessig to him and he was savef triquetta16: saved* Jcarlos287: give up was a wrong term Jcarlos287: but still triquetta16: still after all that man did he got saved triquetta16: God loves all his creation Jcarlos287: good Jcarlos287: \"it repenteth me that i have made them\" triquetta16: which is a greater witness killing someone or letting them kill you Jcarlos287: you cant say that without saying the other triquetta16: do we go to the same church? Jcarlos287: yes triquetta16: i believe all the verses you are using for crime and punishment are old covenant.  we are learning in church that the new covenant is SO much better.  all the verses i have used are new covenant Jcarlos287: i used two verses Jcarlos287: one was old covenent true triquetta16: old covenant verses Jcarlos287: does the other still stand true? triquetta16: both were Jcarlos287: I KNOW triquetta16: neither do Jcarlos287: but latter still applies triquetta16: that murderer is just as wicked as i, except one thing i have jesus Jcarlos287: i agree Jcarlos287: bottom line: i would lay down my life to save my friends. that is Biblical. I would die to save them, but I would not have them die, that is not their choice Jcarlos287: i told you i agree with that Jcarlos287: and i would protect that which God have given me stewardship over triquetta16: it is my choice not to kill.  it is their choice to murder.  if my friend wants to murder to protect themselves so be it triquetta16: i refuse Jcarlos287: i am saying there is a line between murder and defence triquetta16: nope triquetta16: thou shalt not kill.  PERIOD triquetta16: statement ends there Jcarlos287: i will have to get back to you on that triquetta16:i think generally most people are so egocentric that they would kill the one the aggressor triquetta16: i still feel very strongly about my beliefs Jcarlos287: as i do triquetta16: also think what a testimony cassie bernal was at columbine triquetta16: she didnt lunge at the kid with the gun.  she took the bullets.  she stood tall and unwavering. Jcarlos287: i would like you to realize that that is completely irrelevant triquetta16: i also feel the new testament completely backs up my view triquetta16: one NT verse about it being alright to kill the agressor is all i need triquetta16: it is not irrelevant Jcarlos287: YES it is triquetta16: why do you think so? Jcarlos287: because Jcarlos287: he wsnt attacking her friends Jcarlos287: hah Jcarlos287: as i recal he pointed the gun AT HER Jcarlos287: we discussed this already triquetta16: you really cant think of anything good can you. Jcarlos287: i would die in that case too Jcarlos287: DO YOU REMEMBER THE STORY triquetta16: killing to save a friend is no more noble then killing to save yourself Jcarlos287: he didnt give her a chance to answer before he shot her triquetta16: yes he did triquetta16: she said \"Yes. I believe.\" Jcarlos287: do you get my point? triquetta16:  i have to leave though so we can debate later...btw find a NT verse that says defending yourself is ok and murdering to protect a loved one is acceptable triquetta16: no you have no point Jcarlos287: bye   '),(8341,2876677,'29,May,2004','2004-05-29 00:00:00','   To the fat sloth of a worthless piece of flesh and a waste of oxygen sitting in front of the computer(AKA-Pat):  Some people seem to think is kosher to deny that they are wrong repeatedly after I have shown them what is correct.  urlLink Swallow your pride Patty.   Other then that, I do not appreciate you stealing my cool stuff from my posts on my blog. It shows little to no originality on your part. Perhaps I would be fine with it if you credited myself with bringing it to your attention. But you have not. What a shock.  Not directed to you>> I feel the need to vent majorly: I am really really sick of someone and I can wait till they take off and hopefully grow up. If I hear them complain or whine or falsely accuse me of one more thing I might drive a spike through their forehead.   Have a nice day.   '),(8342,2876677,'29,May,2004','2004-05-29 00:00:00','   D     Delicious       U     Useful       S     Strange       T     Twisted       I     Ideal       N     Nutty       Name / Username:       urlLink Name Acronym Generator  From  urlLink Go-Quiz.com     Wow this is cramped.   '),(8343,2876677,'29,May,2004','2004-05-29 00:00:00','   Any more comments about Patty\'s use of grammar/choice of spelling will be deleted.  If I am not feeling lazy.  So anyways, tennis is cool.  I want to learn how to play(well) because tonight was wicked fun.  Then I spent forever looking for shorts, but I couldn\'t find any I cared for.  I did buy three new shirts though.  Not anything I need.  I have no idea where to put all my t-shirts.  I never have enough dress pants/shirts but that is just because I can never find ones I like...or they are $90.  Shorts either.  Well, in seasons past I have never had trouble finding shorts but this year I just don\'t like the styles.  I would just wear pants all the time but that can get hot and they don\'t allow for great range of motion for tennis which I now love and need to learn how to play well.  Volleying is definitely my thing although returning looks like it could be fun.  I.  Heart.  Tennis.   My reading list Hebrews Hamlet CS Lewis Anthology Through Gates of Splendor Living Thoughts of Kirkegaard(try to finish...for the hundredth time)   '),(8344,2876677,'25,May,2004','2004-05-25 00:00:00','   I am tired.  I need shorts for summer.  Luckily, I got a job and am now \"rolling in the dough\".  Seriously, my job pays wicked well.  In other news, I am excited about next school year.  ALOT!  I will miss a few people.  Miss Warren looked like she was going to cry when I left.  I love Miss Warren.  She is like a second mom to me.  She would be a crazy awesome mother.  She doesn\'t have any kids though so that stinks because I know she really wants one.  C\'est la vie.  Comment.  now.   '),(8345,2876677,'25,May,2004','2004-05-25 00:00:00','   urlLink    This was a healthy human baby before it was torturously murdered inside its mother.  I have no idea how any normal human being could perform such a \"procedure\" repeatedly.  It is absolutely sickening.  These women who talk about their \"rights\" disgust me.  That baby\'s face looks cute even though it was horribly multilated.  Some selfish woman had her beautiful child killed so she could lead an uninhibited lifestyle.  I am outraged.   '),(8346,2876677,'24,May,2004','2004-05-24 00:00:00','   I LOVE THE RAIN!  Oh my word.  You have no idea.  Seriously...we need to go walking sometime(I don\'t jog...).  Everyone thinks I am wierd, but I love just going outside when it is raining and just sitting and let it pour down on me.  Or to go for a walk and jump in all the puddles and just enjoy God\'s creation.  well anyways...one of my teachers stopped over.  she is nice.  so anyways...i am making a technology proposal for an organization...I wonder if a minor can be a technology officer in an organization.  well anyways...   '),(8347,2876677,'21,May,2004','2004-05-21 00:00:00','   Humm...so here is my thought: Anyone who comes up to you and says, \"Let\'s go to Ozzfest.\" should be shot. But umm..not much is up. Seeing as I cannot use AIM, and it is indeed raining, I have very little to do. Well actually I could read or something, but let\'s be serious. I really want to go play in the rain. No one likes to though. I have no one to go jump in puddles with me. I remember when I was little, I dragged my brother and sister out in the rain with me and we jumped in puddles and got totally soaked. Apparently my mum didn\'t appreciate that much, because when we got inside I got a spankin\'. Other then the pain at the very end it was a good time. I would do it again. It\'s no fun alone though. A few weeks back I went for a nice 3 mile run in the pouring rain. Best run of my life. I love the rain. What can I say? I like being crazy and stupid and fun sometimes; just let it go, run in the rain, forget everything. Someday I\'ll teach someone to smell the rain, to smell the seasons. Have you ever smelt the rain? You haven\'t live till you have. Someday that someone will remind me to slow down and remember to smell the seasons or stand in the rain for a moment, stopping time. Someday. Until then, I am sitting here writing because people seem to enjoy the idiot box so much it renders them useless.   The best things in life are free. It\'s the simple things people miss everyday that make me happy.   '),(8348,2876677,'21,May,2004','2004-05-21 00:00:00','   No one cares about anything good or godly...or so it seems to me.  I just wish someone would say \"Let\'s go witnessing\" or something...instead of \"Let\'s go to Ozzfest\" or something.  so anyways...what\'s up jenna?   '),(8349,2876677,'21,May,2004','2004-05-21 00:00:00','   \"I try to carefully weigh the message against the objectionable content. Even if the movie has a good message, I won\'t watch it if there\'s too much swearing or dirty language, gratuitous violence, or any sex whatsoever. You have to decide where the line falls for you. \"  1.”I try carefully to weigh the message against the objectionable content.”  Christianity teaches that salvation is by faith not by works.  God does not measure our good works against our bad works on a scale.  One sin tarnishes every good thing we could do and it damns us.  “A little leaven leavens the whole lump.”  Christians today, however, do not apply this spiritual truth.  As a whole, we try to justify what we watch by seeing if the “overall message” is good.  One curse, one act of violence, one bit of nudity or sexual innuendo is enough to damn a film.   Any of the film’s “good works” are nullified by any sin in the film.  Our holy God will not allow sin into his presence because it would tarnish His glory.  We should not willingly set sin before our eyes and ears.  Movies have a way of putting sin in a good light.  Homosexuality, murder, pre-marital sex, extra-marital sex are all examples of things the movies make look good.  No consequences.  Do what feels good.   2. ”Even if the movie has a good message, I won\'t watch it if there\'s too much swearing or dirty language, gratuitous violence, or any sex whatsoever.”  Too much swearing?  How much is too much?  In our family, it used to be one swear was too much and we changed the show.  Then it was decided that every show has one swear nowadays and so it was upped to two.  Then three.  Then four.  It is a progression.  Following the world from a distance is still following the world.  Are you, a finite human, capable of making a decision about how much sin is acceptable in one’s life?  Coincidentally, God says in the Bible we are not to allow sin in our lives.  So the call isn’t actually up to us.  “Gratuitous violence.”  Hmmm…What is gratuitous violence?  Well “Gratuitous violence is violence that is unnecessary or that does not portray wartime violence or violence that actually occurred (Jesus’ crucifixion)” Ok…People have sex all the time.  Is that acceptable?  Can I watch a portrayal of a married couple having sex?  OF COURSE NOT!  It dulls your mind and exposes you to things you should not be exposed to.  The same with violence.  It has been proven by studies that people who watch movies portraying ANY type of violence or listening to music with violent lyrics are more likely to commit violent crimes.   3. “You have to decide where the line falls for you. \"  This implies that there are no absolute standards.  You can decide for yourself what is acceptable.  There is no absolute moral standard regarding movies.  If we say that we can decide what is acceptable, we are taking God out of the picture.      It is doubtful anyone has taken the time to read this entire article.  If someone did, it is unlikely they will change.  Most Christians today could be (and are) slapped in the face with God’s truth and would not do anything different just because they enjoy their sin.  C’est la vie.  I will pray that anyone who reads this might be positively affected.   '),(8350,2876677,'19,May,2004','2004-05-19 00:00:00','   I am so messed up.  And confused.  About myself.  Someone once said \"We barely knew each other, we didn\'t know ourselves.\"  So true.  I have no idea why I am this way.  I often wonder \"Why does [insert name] think that way?\"  New question.  Why do I think this way.  Not necessarily \"Why do I think about that particular thing?\" but rather \"Why do my thought processes occur this way rather than another way?\"  I wish I could control how and what I thought about.   '),(8351,2876677,'18,May,2004','2004-05-18 00:00:00','   The term \"gentleman\", in the true meaning of the word, has nothing to do with manners.  It has to do with being of the \"gentry\", or noble birth.  The term was twisted to mean the manners which were expected of nobility.  Technically, a gentleman is someone of high birth not good manners so no mattter how well-mannered I am, I can never be a gentleman.  A gentleman can actually be one of the worst mannered people ever.  Such is true with \"sportsmanship\" also.  Sportmanship is not a list of noble qualities in an athlete but rather the way athletes act.  The way athletes act is going downhill.  i.e. Intramurals.  Shoethrowing, water bottle throwing/smashing, throwing/kicking the ball across the field would all be considered poor qualities in a losing sportsman or even a sportsman who may be winning but does not agree with a referee\'s call.  The \"must win\" attitude can be frustrating too.  I paid $25 to play in every game, not to be able to say \"My team won every game\".  Also, if you have the minumum amount of required players, then you should not be able to draft from other teams for an extra player.  Especially when someone from your team is the one who spends the majority of the time sitting out and that person from the other team plays in your game and then their own teams.  Retarded.   '),(8352,2876677,'15,May,2004','2004-05-15 00:00:00','   I really do think Google is doing a  urlLink good thing . I mean sure you can have your little right to disagree with him and whatnot...but making shirts and memorabilia that mock his speech and looks is just kind of immature. Go ahead, dislike him and the way he leads and the decisions he\'s made...but lets all be adults here. Google is trying to take us as Americans slightly out of elementary school and I respect that.   On another not so pleasant note, I despise someone. They actually make me sick, but they are in my heart and prayers at the same time. Apparently this person thinks I will get a kick out of them mocking every aspect of my life. ie. my house, faith, friends, opinions, politics, and even my own future. Seriously, they think I\'ll find it funny. Thankfully God has held my tongue on many occasions. Otherwise I would forcefully verbally crush them into crud. This person went from being a life long friend to...I can\'t even describe it. It hurts though. Good friends are really, really hard to come by. I hope that anyone who is willing to be a decent friend to me will find that I will stick with them til the end. In this case though, I\'ve been tossed aside, thrown into a stereotype, and then not even considered. They bring me down. I\'ll be there if they ever need me, and I\'ll let them know that, but for now...  Time to sever a friendship.   '),(8353,2876677,'15,May,2004','2004-05-15 00:00:00','   People demand freedom of speech as a compensation for the freedom of thought which they seldom use.  Everyone one wants to be able to \"speak their mind\" or \"express themselves without fear of reprisal\".  No ones thoughts are orignal.  People are just shallow mouthpieces as saying the same thing in different ways.  No one thinks of new ideas.  They just say what the \"cool kids\" say.  I should note that \"cool kids\" does not mean I am referring to teen in this post ANYWHERE.  This is about the global populus as a whole.  Everyone goes with the flow.  Even the \"anti-flow\" or those who \"rage against the machine\" are just another cog in the machine blindly stumbling down the highway to hell.  The truth is that individualism is a myth.  Someone somewhere does things just like you.  The more people there are, the less unique any one person is.  You may be a deviation from my system of belief but you are not original or alone in doing so.  Don\'t fool your self.  The belief in free will does not allude to original thought.  Have a nice day. Sincerely, Your fellow puppet   '),(8354,2876677,'11,May,2004','2004-05-11 00:00:00','   Ugh...I have to work at Staples today @ 10.  At least it is not that early.  It still bites though because I have been there long enough for the whole \"Yay! This is fun\" feeling to wear off but not long enough to subconciously associate myself with Staples.  In other news, due to circumstances out of my control(an authority over me has done something) I am totally out of the Lord\'s will.  Sorta.  I know that I am to obey authority \"in the Lord\" but I do not think they made the right choice and I really cant do anything without the authority\'s help.  So I am stuck, feeling awful about not being where I am supposed to be, praying the authority will see what I am saying, and trying to make the best of what I have right now.  So if you wanted to pray for me I would appreciate it.  In other news, this is the closest I have ever been to turning in a research paper...I finished it at 5:00 yesterday...It was due at 12:30...c\'est la vie.  at least I am working the time down.  In 6th grade, I turned one in three months late. My parents wouldnt even let me open Christmas presents until like February.  It was a time.  Well my pizza is ready. *leaves*   '),(8355,2876677,'11,May,2004','2004-05-11 00:00:00','   I work up at 4:30...stupid birds...Then I made something to eat and fixed my blog.  Now I am cooking macaroni and posting.  It has been a long time.  I am getting a little sick of people.  Thus, I have found lyrics(slightly modified) that fit how I feel about how everyone in my school treats each other.   I am the son And the heir Of a shyness that is criminally vulgar I am the son and heir Of nothing in particular  You shut your mouth How can you say I go about things the wrong way ? I am Human and I need to be loved Just like everybody else does  I am the son And the heir Of a shyness that is criminally vulgar I am the son and heir Oh, of nothing in particular  You shut your mouth How can you say I go about things the wrong way ? I am human and I need to be loved Just like everybody else does  There\'s a place, if you\'d like to go You could meet somebody who really loves you So you go, and you stand on your own And you leave on your own And you go home, and you cry And you want to die  When you say it\'s gonna happen \"now\" Well, when exactly do you mean? See I\'ve already waited too long And all my hope is gone  You shut your mouth How can you say I go about things the wrong way? I am human and I need to be loved Just like everybody else does   Also, sidenote: I am also out of this wretched school in 13.5 days.  No more listening to idiot Christians talk about Ludacris and Eminem.  \"My Band\" and \"Milkshake\" ARE NOT APPROPRIATE!!!!!!   Oy vay.  AND THE BACKSTABBING!  Oh my word.  Two faced has a new definition.  Its not like public school is any better but hey I expect it from the unsaved.  I also think they will notice a difference between me and them.  Also, there should be a lot more people in my class.  These past three years have been really different from what I am used to.  I can\'t take my peers shallow stupidity much longer.  14 days of peace.  That is all I want.  And my food.  *leaves*   '),(8356,2876677,'10,May,2004','2004-05-10 00:00:00','   What a stylish new design. This, I like. I mean it screams \"cool blog.\" Also, I am glad to see the \'Cat and Girl\' comic has been left behind. :o) I never understood it anyway. Blogger has a fun new layout also. Whoo! Spring fever!  P.S. Are we putting comments back in?   '),(8357,2876677,'05,May,2004','2004-05-05 00:00:00','   Well today has been a busy one.  I have been home all of about 30 minutes since 7:50 this morning.  Today, my dad locked himself out of his car so he had his co-worker, Tamika, drive him home.  She is totally awesome.  She is from the South.  Actually, she was raised for most of her childhood in the Chesapeke Bay area like myself.  I have talked to her a few times before but tonight we talked and she is cool.  My dad gave her a tour of the house and she was all compliments.  I felt wicked ungrateful as she went through my house. \"Three bedrooms!\"  \"Two computers\"  \"THREE FLOORS!!!\"  \"A LAUNDRY CHUTE!!!THAT\'S SO COOL!!\"  \"You have your own computer!?!  I love this desk!!!\"  She was so nice...and she is wicked funny too.  What a blast.  Then at church tonight a \"guy\" was saw my phone on belt and was like \"Oooh.  An LG picture phone.  How much did you pay for that?\"  So I told him.  He got out his LG phone and said \"Guess how much?\"  I said \"Free?\"  He smugly replied \"Yep.  It works great too.  Its not like I need a color phone.\"  It irritated me.  I do not try to be showy or ostentatious.  I never try to thrust what I have in other people\'s faces.  I actually try to hide or downplay what I have.  I enjoy what I have and I am blessed but I do not try to show it off.  I have a cell phone.  I don\'t have a purse.  I wear it on my belt.  I worked hard for my phone.  I am not to trying to flash around what I have...but I have it.  And I don\'t like people who are...ergh!? what\'s that word...grr...\"stupid will have to do...I don\'t like people who are stupid and get upset because I have nice things or I have more than they had when they were 17.  I worked at McDonald\'s for 2 years and now I work at Staples.  It isn\'t like I was handed nice clothes or my car or my phone or my computer.  I worked hard and I got em.  I don\'t try to flaunt but I shouldn\'t have to hide what I have so other people don\'t get upset.  Why can\'t people be like Tamika and be happy for me instead of like \"him.\"  Well, I need to go  Write my paper   Read and revise the UN charter   Read Hamlet   Finish organizing  Affinity    Write a mission statement for my charity   Think of a wicked good name for it   Find someone to put the graphic in my head on paper   Find support for the charity   Do lots of other cool stuff  *toodles*   '),(8358,2876677,'02,May,2004','2004-05-02 00:00:00','   I like making references to literature or music in my blog, especially when it is an obscure work of art.  I dunno...It makes me feel...cool...intelligent.  I still haven\'t thought of a title for this post.  Hopefully, it is a good one.  No one has commented since Jenna and PP left.  Come back...I need readers...I also need my first paycheck...it is as good as spent...so anyways...i stole a copy of hamlet today...i think i will go read it...i hear it is a good read.   '),(8359,2876677,'02,May,2004','2004-05-02 00:00:00','   I am going to put two songs on here.  Comment on which you think is \"Christian.\"   Song #1  I\'m not a perfect person There\'s many things I wish I didn\'t do But I continue learning I never meant to do those things to you And so I have to say before I go That I just want you to know  I\'ve found a reason for me To change who I used to be A reason to start over new and the reason is you  I\'m sorry that I hurt you It\'s something I must live with everyday And all the pain I put you through I wish that I could take it all away And be the one who catches all your tears Thats why i need you to hear  I\'ve found a reason for me To change who I used to be A reason to start over new and the reason is You [x4]  I\'m not a perfect person I never meant to do those things to you And so I have to say before I go That I just want you to know  I\'ve found a reason for me To change who I used to be A reason to start over new and the reason is you  I\'ve found a reason to show A side of me you didn\'t know A reason for all that I do And the reason is you    Song #2  All of you is more than enough for all of me For every thirst and every need You satisfy me with your love And all I have in you is more than enough  You are my supply My breath of life And still more awesome than I know You are my reward worth living for And still more awesome than I know  All of you is more than enough for all of me For every thirst and every need You satisfy me with your love And all I have in you is more than enough  You’re my sacrifice  Of greatest price  And still more awesome than I know  You are everything And still more awesome than I know   More than all I want More than all I need You are more than enough for me More than all I know More than all I can say You are more than enough for me    A girl in my class had the audacity to ask my teacher to play the \"un-Christian\" song.  I was shocked.  Hello, I go to a Christian school.  One would think if a kid was listening to the wrong music they wouldn\'t ask the teacher to listen to it in class.  I was bored so I looked up the lyrics to the song and then I put up the lyrics to it and to a song the teacher considers \"Christian\" because I thought the lack of difference was funny.  No one on the street who heard both songs would ever tell the difference.  No wonder the kids in my school listen to secular music.  Even they don\'t see the big difference.  There isn\'t one.   '),(8360,2876677,'01,May,2004','2004-05-01 00:00:00','   The Bill of \"Rights\" can bite me.  That\'s right.  We have no rights.  If God blesses our nation with freedom, then so be it.  However, I do not advocate going around killing people to obtain what I believe to better for me or even my country or even the world.  Also in response to a statement from my co-blogstress, \"We do not need persecution.  We need revival.\"  Persecution leads to revival.  It strengthens our faith.  It revives our spirit.  Feel free to write back.   '),(8361,2876677,'01,May,2004','2004-05-01 00:00:00','   Questions to my co-blogger and anyone else that feels like adding their 2.5 cents. I love hay-pennies. Why do they no longer exist? Back to my little point...  You are against war. \"Thou shalt not murder.\" -What exactly does this statement entail?  All  war? Let me ask you a simple question. Will God command one to sin? Now let me answer. No, He will not. He cannot tempt any man with sin. So then, when God commanded the Israelites to attack various and numerous peoples, was it sin?  Was it murder? I understand those were different times/circumstances. War is war. So if you mean all war then we should talk. Perhaps that is something to think about.  I\'d rather not discuss the current war too much. Makes me sick you see.  You also have an interesting outlook on persecution. -My thoughts. We don\'t need rights. We don\'t deserve rights, or freedoms of any sort. What we do deserve is death. However, Christ died to free us. Free us from sin, yes. But He also died to give us a life so that we might live for him. You claim the church flourishes under persecution. I agree. Christ is worth every bit of torture imagined and more. You say you do not desire freedoms. But freedoms are what we in America take for granted, because we have known nothing else. None of us understand what it means to be truly persecuted. Until we have been stripped of everything we have it isn\'t fair to those that are persecuted to say that we want what they have. In reality no one wants to be tortured. It is not on anyone\'s to-do list. Kids do not say, \"I want to be tortured when I grow up.\" It is horribly sad that the Church grows more under communism and whatever else then it does when it is given freedom. America and every other religiously free country is a pit of lazy lukewarm Christians who eat their meals and forget to pray for the persecuted, forget to witness to their neighbor, and who do not attend church except for when it is convenient. Their Bible collects dust while over in China, each single page is treasured and books are collected after months. But do you recall the Pilgrims? What was their reason for starting up this fine country? Oh yeah, I do believe it was for freedom of religion. They were being persecuted. They did not like it. They moved. God was with them, and established America. Freedom of religion.  Also, I draw your attention to the Great awakening. Thousands were saved. Why? They were not being persecuted and were allowed to practice. Try that with the commies.  We do not need persecution. We need revival.  Sending missionaries is also a feat when they are not allowed into countries. Sending them requires freedoms. You do not need freedom to serve Christ, but God has set up these freedoms for the furtherance of His Word.   Again, all we deserve is death. Nothing is needed to serve the one who died in our place.    Freedom is a blessing.   '),(8362,2876677,'01,May,2004','2004-05-01 00:00:00','   Wow.  Some people think it is wierd I have my life planned out fairly well.  Well I have a plan and whether it happens or not is totally not in my hands.  I actually know a young girl who has her life planned out to a much greater extent and she has for quite some time apparently. I just have a list of qualities I am looking for in a wife, a general idea of what I want to do for a career, the names, gender and birth order of my 4 children.  Actually I have three of the names picked out(the boys\') and a name suggestion for a daughter but I will leave that one up to my wife.  My only real desire is that it stay with the Gaelic theme I have.  Woohoo!  I love my heritage.  Well I mean the nations I am from, not necessarily my individual ancestors.  Anyways, if you see any honest Baptist girls who love the Lord and have vampire fangs, give me a ring.  Or tell them to go to BJU.  All in all, the moral of the story is \"It is not bad to have a plan.\"  Don\'t procrastinate.  Illegal narcotics are bad.  Buckle up.   '),(8363,2876677,'01,May,2004','2004-05-01 00:00:00','   Someone who should know be better but doesn\'t because he is a retard said something extremely irritating.  \"I don\'t know why you don\'t want to go to a ball game.\"  Umm...you have known me for a very long time.  I have NEVER ever enjoyed watching soccer, hockey, basketball, baseball, football, and pretty much any other mainstream sport.  Unless someone I know and like is playing I do not enjoy watching sports.  I enjoy watching people I can identify with.  I don\'t identify with those sports.  If you want to take me to a classical symphony, awesome.  If you want to take me to an art gallery, great.  I even like watching extreme sports.  For some reason I identify with those athletes more than other athletes.  I love skateboarding, BMX biking, surfing, snowboarding, sky diving, etc.  People are always telling me I want to grow up to fast.  I like being young, I just want people I identify with.  Generally, they are older than me.  The age gap between me and \"them\" is lessening.  Hopefully, I don\'t die of boredom before I get there.   '),(8364,2876677,'01,May,2004','2004-05-01 00:00:00','   Legion  Enveloped in a sentiment, a sound that rushes over me. Engage an impulse to pretend I have a faith as pure. Not forgetting what it means to dream. Indulging everything. Entertaining thoughts that I\'ve the strength of those I yearn to be. Cheers and tribute greet the saviours. Reckless thoughts survive. Anachronistic and impulsive.  And what will happen? Will I dream? I am too scared to close my eyes. For a second please hold me. None can change in me these things that I believe. But I don\'t know what happens now. I am too scared to close my eyes.  © VNV Nation   '),(8365,2876677,'29,June,2004','2004-06-29 00:00:00','   Anyway. Patty is a loser. Patty is a loser. loser patty is a patty  loser er er est.   End.   '),(8366,2876677,'27,June,2004','2004-06-27 00:00:00','   urlLink'),(8367,2876677,'24,June,2004','2004-06-24 00:00:00','      '),(8368,2876677,'23,June,2004','2004-06-23 00:00:00','   urlLink'),(8369,2876677,'22,June,2004','2004-06-22 00:00:00','   urlLink'),(8370,2876677,'22,June,2004','2004-06-22 00:00:00','   urlLink'),(8371,2876677,'22,June,2004','2004-06-22 00:00:00','   urlLink'),(8372,2876677,'17,June,2004','2004-06-17 00:00:00','   At your ten year high school reunion...  by  urlLink  robbiewriter        Your school name           Your name           Your job will be    President of a small country      You will be worth    $666,563      Everyone will think you    are really hot             Created with the ORIGINAL  urlLink  MemeGen  !   '),(8373,2876677,'17,June,2004','2004-06-17 00:00:00','   urlLink'),(8374,2876677,'16,June,2004','2004-06-16 00:00:00','   urlLink'),(8375,2876677,'16,June,2004','2004-06-16 00:00:00','   urlLink'),(8376,2876677,'16,June,2004','2004-06-16 00:00:00','   urlLink'),(8377,2876677,'15,June,2004','2004-06-15 00:00:00','   urlLink'),(8378,2876677,'11,June,2004','2004-06-11 00:00:00','   Julie, its official.  I have decided to judge you and at the same (also  first) time, I will know personally attack you.  You are an idiot.  Wow.  I stand in awe of how dumb you are.  That was THE SINGLE MOST IDIOTIC THING I have ever heard(Maybe I exaggerate a little...but not much.)  I know(actually hope) it wasn\'t your attention span that caused you to miss this but in the ohh...SECOND LINE I apologized to Seth.  I can handle the fact I upset people.  I am at peace with that.  Everyone upsets someone at some time.  You have to learn that you can\'t please everyone.  You are the upset one.  It is your problem to deal with not mine.  When did I blame?  There is a word for you...If I could spell it I would use it.  nuch schlepp!(Thanks grams.)   '),(8379,2876677,'09,June,2004','2004-06-09 00:00:00','   If your name is not Seth, Julie, or Jenna, it is unnessecary for you to read this post.  Right off the bat, I apologize for saying you are anti-Semetic, Seth. Secondly, it would not hurt one of you to start reading entire sentences or paragraphs.  I, at the top of a comment, direct it to a certain person, and you respond to me as if I said it to someone else.   Instance #1 of misinterpretation \"Dustin- I took offense to the comment you made to Julie regarding her heritage. Not only was it sarcastic, but you called her...arrogant?\"(From Jenna)  I said \"I love your need to point out your heritage.  I thought I would too.  I have decided to be proud(not arrogant) of where I come from.\"  I did not call Julie arrogant.  I said I was not being arrogant due to the fact that \"arrogant\" seems to be your catchphrase for me.  Instance #2 \"**cough** Back off Jules.\"(from Jenna)  When did I directly attack or fight with Julie?  Intance #3 \"As for Dustin: It seems that you forget what you write and then refuse to back yourself up, so I\'m going to try to help you out. TO START, you wrote, \" I have decided to be proud(not arrogant) of where I come from.\" I am quite proud of my heritage, believe it or not. If you question my values on my own family history, you will want to take it back... be VERY aware of that.\"(from Julie)  When THE HECK have I EVER questioned your values?  You jumped into the middle of a disagreement between Seth and I and act as though I personally attacked you before I even knew you existed.  Instance #4 \"You said you were holding yourself back to an extreme extent. Really, what do you have to say to someone you have never met and tried to outwit? What do you have to say to me that you are holding back so darn much? Because if you believe your balls have finally dropped, and that you have the right to say I am arrogant about my past, just try telling me what your holding back.\"(from Julie)  LEARN TO FREAKING READ!  I did not say I was holding back about anything you said.  I was holding back on Jenna\'s need to jump in attack me and be very two-faced.  ALSO, I NEVER CALLED YOU ARROGANT!?!  Instance #5 \"You also claimed that Seth was anti-semitic. If you look at Seth\'s journal under the date - 5/19/04, you said: \" I do not appreciate your anti-Semitic idiocy.\" Either you have extreme short-term memory or you are just ignorant, because he is my best friend and we also dated for two years.\"(Julie)  I did not forget saying this.  I never said I didn\'t say it.  Sidenote:5/17... It took me forever to realize there is no 5/19...  Instance #5 \"you about doing the same EXACT thing to Seth, and now you are doing it to Rosey and Annie as well. Just quit trying to be a higher authority.\"(Julie)  WHO IS ANNIE?  WHEN have I EVER mentioned her?  Instance #6 \"you attempted to wave your little IQ around to impress me. every mannerism of yours that i have experienced is dripping with arrogance.\"  HOLD UP!  When have I tried to wave my IQ around.  Day 1 I acknowledged your IQ was higher than mine.  I am SO far from arrogant.  Don\'t talk about judging people you don\'t know and then call me arrogant.  It is hypocritical.  ALL THREE OF YOU misinterpret EVERYTHING I say and call me arrogant.  There are so many more...I have things to do though.(I bet that is probably  arrogant )  Anyways, Julie I have never said anything bad about you...so other than me calling Seth an anti-Semite...I have no clue what your beef is.  I would suggest you move on and forget about me as my importance is miniscule in the scheme of things.  Jenna, words cannot even express...  Seth, I would also suggest you find a way to let go of your bad feelings and lighten your load.  Also, I miss my spot on the blog.  I felt loved.  ;)  I wish things had not gone this way or this far.  You are pretty cool.  Julie, I don\'t know you.  At all.  I do not try to say ANYTHING about your beliefs because of that.  I would appreciate since you are older and hopefully more mature, if you could do the same.  Bottom line is I DON\'T MATTER.  Stop wigging out over everything I say.  If you have a comment, I have a comment link.   '),(8380,2876677,'09,June,2004','2004-06-09 00:00:00','   OK!  I am no longer saying the pledge in its original form.  It upsets me and I believe it to be idol worship(Pledging my allegiance to a flag) \"I pledge allegiance to the United States of America and the ideals for which it was founded upon, one nation, under God, indivisible, with liberty and justice for all.\"  Also the Christian flag... \"I pledge allegiance to Christ, my Saviour, and His kingdom, Who, through His crucifixion and ressurrection, provided life and liberty for all who believe.\"  Another thing that irritates me is people\'s obsession with proper observance and respect to A FLAG!  ...Meanwhile, the Holy Bible is under the dirty laundry...maybe.  Also, people stand for the national anthem but no one stands for Scripture reading.  Irritating...   '),(8381,2876677,'07,June,2004','2004-06-07 00:00:00','   urlLink    I miss you and it\'s killing me.   '),(8382,2876677,'02,June,2004','2004-06-02 00:00:00','   If I get invited to another leadership conference, I will flip.  Every one that I get, is just another fancy letter saying \"You\'re smart.  We want you at our conference.  It is 6 days long and costs about the same as a year of tuiton at school.  We offer scholarships, but only to minorities.  Sorry, I don\'t know why we sent this to you.  Well, at least you can save the $50 invitation.  Oh, you can tell colleges you were invited to this conference too.  Oh, you were already accepted to college?  Nevermind.\"  I would love to go to one of these conferences, especially the lastest one, the Presidential Leadership Conference.  Basically, I get to choose the Republican or Democatic Presidential thing in New York or Boston, respectively, and watch and even participate a little bit.  Sadly, I do not have this kind of cash.  If you know any billionaires looking to help a poor white kid, let me know.  Seriously, let me know.   '),(8383,2876677,'31,July,2004','2004-07-31 00:00:00','   urlLink'),(8384,2876677,'25,July,2004','2004-07-25 00:00:00','   urlLink'),(8385,2876677,'24,July,2004','2004-07-24 00:00:00','   Your love is... by  urlLink  ChibiMarronchan        Your name is...           Your kiss is...    breath taking      Your hugs are...    warm      Your eyes...    sparkle like the stars      Your touch is...    the only thing I desire      Your smell is...    beautiful      Your smile is...    entrancing      Your love is...    eternal              urlLink  Quiz created with MemeGen  !   '),(8386,2876677,'17,July,2004','2004-07-17 00:00:00','   So...umm...yeah.  I am going on a missions trip for a week.  I am sure it\'s gonna be fun times...yeah.  I might Audioblog once or twice if I need to vent.  Hopefully, I won\'t have to but living that close to all those people...something is bound to irritate me.  So have a good week kids.  See ya on the other side of...something*mumbles**trips as he walks away*   '),(8387,2876677,'17,July,2004','2004-07-17 00:00:00','   urlLink'),(8388,2876677,'08,July,2004','2004-07-08 00:00:00','   urlLink'),(8389,2876677,'02,July,2004','2004-07-02 00:00:00','   '),(8390,2876677,'02,July,2004','2004-07-02 00:00:00','   '),(8391,2876677,'12,August,2004','2004-08-12 00:00:00','   urlLink    I am really sick.  Great times.  I left work early yesterday because I felt so sick but I am going to try again today.  Hopefully I die.  Smile.   '),(8392,2876677,'07,August,2004','2004-08-07 00:00:00','   I hate when I type a really long, really funny(in my opinion) post and the I accidentally hit something and lose it all.  I never end up retyping it.  The gist of tonight\'s post was \"I need a girlfriend\".  Email me at triquetta17@gmail.com   Give your name, age, things that interest you, and reasons I should ask you out.   '),(8393,2876677,'04,August,2004','2004-08-04 00:00:00','   I am SO sick of trying to put band-aids on relationships that are gushing blood.  Seriously, I don\'t mind listening and going in between every once and a while but I AM NOT A COUNSELOR!  I can see a dead relationship a month before the couple breaks up and I am tired of trying to help them.   '),(8394,2876677,'03,August,2004','2004-08-03 00:00:00','   Divide and Conquer.   '),(8395,2876677,'02,August,2004','2004-08-02 00:00:00','   The only thing I want in life.   '),(8396,4238006,'22,August,2004','2004-08-22 00:00:00','   School is back in YAY!!! Well my week was pretty uninventful. I went Bowling on Friday and I also went to a Chruch Lockin. It was pretty fun, we played Mafia and a game called Underground Chruch. It\'s relly hard for me to explain in typing, but trust me when i say it\'s like hide and go seak on crack...it\'s a sweet game. Well thats about all I got now, i\'ll try to do something more interresting later on in this week. '),(8397,4238006,'15,August,2004','2004-08-15 00:00:00','   * Beware this is a long post*   Well I guess not that I have yet to get any sleep I\'ll go ahead and tell you guys about my adventures on Wensday through Friday. Ok it all started out with my friend Andrew calling and telling me that there was a Halo party at my other friend Tom\'s house (we do this ever so ofter, this was different because we had 2 X-Boxes). Well after a few hours went bye I called my other friend whow as going Nathan and asked what the hold up was. It seems my friend can\'t count and doesn\'t know the difference between 10 and 7. Well at about 10 or 9:30 after Nathan called this other guy i know who i guess i could call a friend now Walter comes and picks me up. It takes some time but we make it to Tom\'s house, and on the way invite another friend Wyatt. We get to the house to relize that 1.) we have the wrong A.V. cables and 2.) Microsoft has to squeeze every penny out of you they can so we need to get a X-Box ethernet cable. Well this takes us to a closed Big-K mart, and a Super Wallmart...and as you can guess, the Super Wallmart doesn\'t have the freaking cord we need. Well we call Wyatt, or he calls us...I can\'t remember, and we decide to meet them at the Wallmart because it\'s closer to them. As we wait we (Nathan, Tom, and I) decide to get some food while Walter plays Deer Hunter. We end up stocking up with a box of 120 pizza rolls for like 9 bucks, 4 pizza\'s that are $.98 a peice, 2 sodas that are 2 for $1 and a pack of swiss rolls that are $1.09...yeah we were loaded. So we end up meetin Wyatt and Andrew...and it\'s about 11 or so at this point in time. And we go off on our way back to Tom\'s house with our newly aquired food. We start playing X-Box for a while and we play some Super Smash Brothers Melee. After this Nathan, Tom, and I start to prepare our food, when suddenly the bejesus is scared out of us by Casey, Heather, and Luran (yes more friends). We at first think we mass hullicinated and tell Andrew, Walter, and Wyatt what we saw. We then get a call saying that they want to meet us at a bowling alley...we at first decline, but a few of us put up a fight and it is decided that we go...oh and it is 1 am now. Well, we get outside and notice that Walter\'s car has been wrapped with plastic wrap...and as we then find out they put crisco inside the door handles. At this point we\'re pumped...except Tom who has decided to go to sleep (lightweight)...and decide to plan a counter attack. For this talk we get 1.) Flour 2.) Soap 3.) A Super Soaker (anti-personale) 4.) Eggs. Well we decide that we are going to flower the bejesus out of their cars and make it stick with the liquid soap we have...the eggs serve no use in the first attack and we end up buying more. Back to the present, we figure out that they are at a Waffle House and meet them there before we go to the bowling alley, we play it cool and act as if nothing had happened...and we notice their insecurity. We both go together to get gas and after Walter refules his car they decide to go back to the house at which they are staying and switch cars. Wyatt then come up with a brilliant plan of going to the house, waiting for them to leave and striking at the cars that are left over. This is facilitated by Andrew getting us balls lost around their neighborhood, how does this work out you say?? Well as we finally know where we are going we see their newly aquired car turn out onto the road we\'re using. We immedantly pull over and turn our lights off. What then follows is a tatical strike that totally demolished their cars. They are covered in flour and soap, not to mention pizza (i forgot to mention we brought a pizza with us) which was stuffed under the door handles. We go bowling and from what I gather they struck at us in the parking lot taking a 2/3 stance...oh and they still at this point have no idea what we\'ve done. They get us good though when we get back to Tom\'s they pelt us with water balloons...very inaccuratly though. This all takes place bewteen 1 and 6 in the morning. The next night is proceded by another ninja esque strike where Walter and I meet with them to discuss the rules. We have our firends Maggie, Paul, and Wyatt bail and we use a cover story of them getting lost and we need to go look for them. What then happens is us taking them around the building and having to dart for Walters car due to our perfectly mastered strike, the retelation the plan is ineffectual when we manage to bipass the crappy road block they have created, pick up the crew and move back to the foward base (Nathan\'s house tonight)...oh the strike was a flowering and us using the window markers you see on car windows to write I think it was Cunt all over their car. As it so happens Walter leaves at 3 in the morning thus they have no car to strike at...but as it seems they manage to hit Maggie\'s car by writing explatives over her car. I\'m not sure if the war was continued on Friday...but I don\'t think it was. All in all I would like to conclude that we won the major battles and our percision got us the war...but if nothing else we tied. It was a fair fight and one I think would be fun to continue later on. '),(8398,4238006,'15,August,2004','2004-08-15 00:00:00','   Well, I\'ve finally done it....I have a blog. I though I\'d get one so I could write down all the remenial tasks I do a day, cause I know everyone is interrested in my sweet ass life. I like the name, but I\'m suprised it wasn\'t taken...like everything else I tried to choose for my screen name and web adress. I\'ll leave all you people out there to figure it\'s reference out...and once I get to it.. and if I can figure it out I\'ll try to post the random pictures I\'ve taken out of boredom. I\'m sure if you\'ve come here you have my AIM screen name...as it\'s the only refuge for the link, but if you don\'t my moniker is Killerfurball. Ok for those of you who know nothing about me my real name is Michael and I live in Nashville, TN...please no hick jokes. I like a lot of different forms of music...such as Rock, Rap, some Techno, and....thats it really. Oh I have a complete distaste for Country music. I like video games, they are a good drain of my time, and I find them very entertaining, especially the RPG\'s, anything with an emersive story can usually pull me in....though I never realy liked the FF series. Ok well to the actual reason I created this blog....I like to write...even though I have yet to write 4 AP US History essays due in...well...1 day. I have a few links everyone need to check out (I\'ll have even more once my friend gets on AIM and I can copy and paste them to this) www.seizurerobots.com oh, please don\'t go to that if your an epileptic...I don\'t want anyone having a seizure  urlLink www.lowwatermusic.com  I like this band...they have 3 mp3\'s for download on their site and 3 videoes. Just because I know everyone is dieing to know what I am doing at this ungodly hour...I seem to be watching a Snoop Dogg video on MTV2, I think I might just switch it back to Old School though...thats a good movie, I suggest everyone go see AVP...on the premsis that it\'s Aliens vs Preaditor...I mean that alone warrents a good movie (pfft), actually it wasn\'t as bad as I though it was going to be...least it wasn\'t Van Helsing, god I can\'t believe I paid 8 bucks to see that movie. Well, I guess I\'ll post more stuff in this blog later today...I\'m going to allow comments but please guys/gals keep it down. '),(8399,3505043,'30,June,2004','2004-06-30 00:00:00','   ... and the minutes scurry away. The time is coming that I should leave, and I know that with it, it carries no regrets. However, I am starting to doubt the effectivity of this move, how will it make me forget that which lives in my heart? Is it possible to fall out of love when the love you\'ve fallen INTO is more real and deeper than the greatest love ever immortalized in page and pen? Could I really forget the best moments in my life? Would it be fair, considering that I do not forget the worst moments all too easily?  You still look at me behind beady glasses and perhaps this steely wall that keeps your feeling locked away, hidden out of my sight. However I can still smell you, I can still see you, I can still feel you, and even though it might just be the carcass, the shell into which love was encased almost a year ago, I still love it, the way you cherish the gift\'s wrapping even though the gift is gone.   Nonetheless, the fact that love shines not through anymore, does not give your eyes that perfect glow  ... it breaks my heart and shatters the pieces into the finest dust. I yearn for a love that doesn\'t live anymore, I asked and pleaded once, and even though I plead no more, my heart still yearns and cries out for the soulmate you once were. Our past has been encased and pasted into the pages behind me, I cannot reach out and relive, too certain have you made it that there\'s no way to return to that ecstatic state of mind, body and soul, thet merge of two into the perfect ONE that made both parts... perfect ... just as you are.   ...yet again... for Heri.  '),(8400,3505043,'23,June,2004','2004-06-23 00:00:00','   ... or how I want to get my groove back, however people keep tugging me in different direction.  And friends are not friends when they impose themselves on you without consideration of previous plans and/or the fatigue you\'re already feeling with existing obligations. This is pissing me off so bad, I can\'t even function right.   Let\'s see, let\'s make this call right here (J.J.)...  ... and a voicemail, I get, which is fine cuz I DON\'T impose myself on my friends. What is UP with this girl? Why do other females find this even remotely normal?!  Am I the abnormal one here? :-(  '),(8401,3505043,'17,June,2004','2004-06-17 00:00:00','   You can\'t help it, you can\'t evade it: a piece of you will always live in me. Even if I never carried you child, even if I was never married to you, even if I was never an official pair to you ... I loved you immensely, and to this day the feeling carries an echo through my heart.  Forever will I always be reminded of the very few nights we spent together, hearts and limbs entwined in a soul-embrace, drenched in the salty darkness of an evening by the beach ... or a kiss under the rain. Whenever I get the feeling of a dark glass puncturing my heart, you will be there. That piece of you... bleeding the eternal love that once lived in me ... and still remains.   .. for Heri  '),(8402,3505043,'10,June,2004','2004-06-10 00:00:00','   Aunque suene un poco desfasado, inclusive ilÃ³gico y hasta injusto, he tomado una decisiÃ³n: no envolverme en situaciones a medio resolver. Me he entregado varias veces en cuerpo (no en alma) a aquellos a quienes mi corazÃ³n estimÃ³ como inofensivos. Uno no decide de quiÃ©n enamorarse, no obstante el mismo corazÃ³n posee un subconsciente (aparentemente) q puede distinguir entre aquellos q nos pueden daÃ±ar mediante la intensidad y aquellos que no.  Y en contra de toda racionalidad, los que nos pueden daÃ±ar son los mismos a los cuales podemos amar con toda nuestra alma. MÃ¡s miedo me da la persona a la cual le noto la capacidad de hacerme amarlo intensamente, porque sÃ© q tambien puede hacerme un daÃ±o al mismo grado de intensidad. Por ello opto no entrar en relaciones donde no me puedan ofrecer lo justo y viceversa. Ahora mismo, hoy por hoy y como estÃ¡n las cosas, yo no puedo ofrecerle nada justo a quien desee estar conmigo. Yo me voy. Y sÃ© que la Ãºnica persona q demuestra la capacidad de enamorarme como una tonta tampoco me puede ofrecer lo justo a mi. Por ende: nada va. Solo la ilusiÃ³n de poder disfrutar de su mirada, de su voz, de su energÃ­a y su compaÃ±Ã­a... hasta que llegue el dÃ­a en q me tenga que ir. Luego, q la vida decida si podemos reunirnos, o no.  '),(8403,3505043,'04,June,2004','2004-06-04 00:00:00','   :-/  I know it\'s not love, I know the emotion is mainly visceral and even a bit whimsical. However, it is stronger than what I\'ve felt for some time now (I think?). Maybe because I\'m moving on, maybe subconsciously I am looking for something to cling to, to make \"staying put\" safe and easy. Or maybe it is true attraction. Why even question it? His gaze burns my insides and melts me, corrodes me, makes me feel immense and small all in a single quiver of an eyelash.    This doesn\'t point to a \"falling-in-love\" scenario, maybe. Or maybe it does... \"uncertainty excites me\" (oh so true.  Coming from a different country, with a different attitude, different making from a different mold ... makes me shiver with delight (however the circumstances make me more anxious than anything). Nothing has happenned, and at least in my mind, whilst things remain the same, nothing will.  But in my mind his name echoes ...  '),(8404,3505043,'12,August,2004','2004-08-12 00:00:00','   ... or why the concept of this journal has defeated itself in purpose. The name in itself says it all:  Tangled Emotions . Whereas I found myself in a situation of stormy standstill, like the sea: calm on the outside but broiling with life underneath it\'s mirrory calmness... now I find myself in a totally different place:  a total calmness, utter bliss . A place I had never been before ... and it concerns me that at some point I won\'t have all that much to talk about (at least where my love life is concerned) because it will all be so ... alright... makes me happy but then defeats the initial purpose of this cybernetic space ... which means I shouldn\'t have destined it to be ONLY about my love life ... which means I\'m gonna have to either change its name or tangle my emotions in respect to other areas in my life (which would be ... unnatural ... although interesting)....   Just had to question and re-state the purpose of this journal:  no longer to gripe about my love life or lack thereof... just another journal...  '),(8405,3505043,'05,August,2004','2004-08-05 00:00:00','   ... But in the best sense possible. An utmost tranquility, serenity, that cannot be described by mere words. I may not be taking flight anymore, but the most important part of me has: my heart. At last my heart feels free to love, devoid of desperation and insecurities, it loves like a calm sea... washing over, slowly advancing, enormous, sometimes ominous, sometimes inviting, but inevitable and always there. For the first time in perhaps more than 10 years, this process takes yet again place in my heart. A process that surely and steadily burns set into my heart, changes my world, makes it more real, more colorful, more ... THERE!      Blessed be the moment our lips joined...  '),(8406,3993797,'25,July,2004','2004-07-25 00:00:00','   Todays a new day. &nbsp;  A better day. Today my family lef to go to NJ, sad but I think Dan and I are goin down for the weekend- a little weekend getaway. But, then again theres a party going on that Ki and I want to go too and we\'ve been waiting for one all summer and the one time I have the chance theres a good possibility I won\'t be there. Today turned out to be a pretty great day. I called Dan around 10 and we talked and he asked me that around New years we can go down to Florida to see his aunt since she invited his family down he wanted me to come along. Sounds like a great idea to me and I would love to go. Also, Dan and I are going to rent a cabin for a weekend and spend sometime together later in august. Im really excited for that. Now, back to my day- After that he wanted to get together with me and spend the day together so I was like hell ya. But we both still had to take a shower and I had to go grocery shopping for my grandpa first, we ended up gettin together around 11:15 or so and hung out till 10. He taughted me or helped me perfect my learning how to drive a standard, which I got pretty good at, I would give myself an A for the day! ::pats my back::. Well, now I am home alone and scared and lonely, but Sean is keeping me amused and happy for now, until its time for him or I to go to bed :(. He really cracks me up! We really have good times/conversations-There never really dull or ever dull moments with us. We\'re just two cherry tomatos, wait we dont like tomatos, we\'re just two nuts in a sack! Muah, yes that is it!!    Yes. Well. I suppose I have written my daily post, so I will let you know more tomorrow as my week home alone continues ( as music plays in the background: DUN DUN DUN)   Readers Voice: Next time on Home Alone VVXIIXI, Jaimie...(Well we dont really know because its not tomorrow) ::whispering in the background to the producer:: \"what am I suppose to say now, we havent even filmed tomorrows epipsode?..producer yelling back-JUST FUCKING MAKE SOMETHING UP YOU DOUCHBAGE THATS WHAT WERE PAYING YOU FOR!   ...thank and and thats all the time we have for today people.    &nbsp;  Jaim    &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;   I Love You Dan '),(8407,3993797,'04,August,2004','2004-08-04 00:00:00','   Do we all have one soulmate? One person who we will spend the rest of our life with? One person who we will love unconditionally even of their faults? One person that will be by your side? One person who you fall in love with all over again each time you see them? One person who is   your everything?         It\'s weird that I\'ve found love so young. Is this it? Is he the \'one\'? Is he my soulmate? Is he the one I will spend the rest of my life with? He is the person I love unconditionally even of his faults. Is he the one that be by my side? He is the person I fall in love with all over again each time I see him.   He is my everything.         Is it possible to be in love at this age? Falling in love is possible at any age. Never puppy love to me. It\'s either  Lust  Or   Love.   I fell in love at 14 and 2 1/2 years later here I am, in love more than I was before. Falling more and more, deeper and deeper in love with him. Knowing all along he\'s right there with me, behind me in everything I do. Loving me for who I am.  This is love .      Love has thrown obstacles that we\'ve only defeated, coming out stronger than when we went in. I can only hope for more defeats and a stronger relationship.   Together we are everything  .      The time  we spend is priceless .  He is priceless .   He is the  one I fall in love with over again each time I see him  . He is the  person I love unconditionally even of his faults .  He is my best friend .                                                                                 He is my everything.                                                                                                  I Love You.          Jaimie          -Dan, you aree my everything. You are the one I love and will forever. Thank you for everything you have done for me. :)     I love everything about you. You still ROCK my World :) (Even when we\'re down and dirty) lol. No more dirtyness, thats all for us!     Love You.          Muuuahhh. '),(8408,3848892,'22,July,2004','2004-07-22 00:00:00','   Yesterday mornin i got up around10:30,i got ready for work 12-4,it eally wasnt that bad.After i got off o went straigh to MY baby\'s house and we sat&nbsp; around lookin at pics for a while and then we loaded a table up to take to nese\'s and we stayed at her grandma\'s house for&nbsp;a while,then when we was leavin....n/m,not goin there....anyway,we went to wal-mart to get her somn for her pics to hang and we went back to her house and it took at least10 min.&nbsp;to hang that bitch up lol.we started watchin Jeepers Creepers 2 about 7 then abot when it was over sissy came over and we all got in the hot tub and had a VIRGIN strawberry daquri hehe.i had to leave at 10:30 but we didnt get in til 10 cause of some reason ,but jess called my mom and talked her into lettin me stay til 11:30 *i love u baby!!* we stayed in and goofed off for a while then it was about time for me to leave.When i got home last night i callled her and talked to her&nbsp; for a couple of min. and then i went to bed   This mornin got up around 11,got ready to go get my paycheck came back home and i\'m sitti here now watch in a old race on speed vision heh.but i\'ll be back later 2night to type more MAYBE...bye!!   '),(8409,3848892,'19,July,2004','2004-07-19 00:00:00','   Hello people...today was a lil fun actually...mom woke me up about 10:15 cause she was goin to h;ville, and she wanted me to watch mrs. miko\'s dog *she is from japan* anyway,when i got back home me and my cousin *bernie mac hehe* ordered a pizza and when we finished eatin i called up corey and see if he wanted to come swimmin so i went and got him and we came back and started goofin off...about an hour later Tristan and James heard came by and James started gettin into it with corey\'s younger bro. *he is 3!* and so he started swimmin with us and i told tristan to go get some shorts so he could..we waaited about 15 min. so peeked my head in and he didnt answer so i went inside to find him laying on the bed asleep lol. about 3:30 tristan took james to geesepond to drop him off to see his girl *which one? lol. and he came back and we went to ______to get my baby her surprise heh. *ur gonna love it! kisses!* after that we came back home and he had to leave so here i am now chillin like villin lol. but i MIGHT be back later tonight to blog more..but i\'ll talk to ya\'ll later..be good! bye   '),(8410,3848892,'16,July,2004','2004-07-16 00:00:00','   This mornin i got up around 10:15 i think,i felt like crap so i layed on the couch til i had to go to work from 2-8,that really sucked,plus i was sick as a frickin dawg so,yea,i had real fun...but after i got off work i came straight home and TRYED to eat somn but i couldnt hold it down and then Jess called me about 9:30 and i just now got off the phone cause i need to go to bed so i can sleep some of this sickness off,so i will talk to ya\'ll tommorow...bye! goodnight!&nbsp;   '),(8411,3848892,'15,July,2004','2004-07-15 00:00:00','   Well people ..its been a while since i have blogged *duhh* lol alot has happened i could type it down but u know how frickin long that would take?!?&nbsp; but anyway..today was a lil boring got up about 6:15 JESS called about 5:50 or somn like that to tell me that they were about to leave cause they were goin on a cruise *SIGH!* i wont get to see here til tues! *what am i gonna do!!*anyway,i went to hardees and ate,then went over to the stadium and waited on everyone else to get there,we only had to run 1 mile *coach page kicks ass lol* after that i went to go get my check from foodworld *blah* it was ok but i need more than that and i\'m not gettin 2 damn jobs,nooooooo sir! i\'m to lazy for that shit lol.I got home around 8 8:15 and i tryed callin jess but couldnt get ahold of her so i went back to bed til&nbsp; about 12:30,then i called her again and she said that she was roaming,we talked for about an hour or so,and then she had to go cause they were goin to eat somewhere..so me and my cousin went to play wiffle ball for a while and then went swimmin,i got sunburned but he didnt *now aint that some b/s lol* when we got done i had to go get mom her medicine and then we got back i did somn but i cant remember heh. 9:30 roled around and my baby called and we talked&nbsp; about an hour cause my cuz he wanted to go to DQ,but they were closed so we went to wacArnods lol! *whew* mcdonalds then when we got back i called her back but she had to go to bed so i only got to talk to her about 1 1/2 *sigh* but its all good...but i\'m bout to go so i\'ll talk to ya\'ll later..be good! goodnight!! bye   '),(8412,3848892,'08,July,2004','2004-07-08 00:00:00','   Well this mornin i got up around 10 10:30,i called Jess to wake her up couldnt do that,its not possible to wake her up in the mornin,but its all good,she needed sleep *i love u baby* anyway,i took cory home around 11,and then i went to go get my paycheck,it weas only $48 *wtc*.after i did that i came back home and talked to sissy for a while and i called Jess and talked for a while and then i had to take mom some food at work.then i came back home to talk to my baby,sis,tristan and a few other people..oh yea i said hey to val heh *gag* oops did i say that.then i had to go unload a trailer for jim,it wasnt that bad,but then when i got home i called jess,we talked  for like an hour,then i had to go load a lawn mower *which reminds me i have to mow tommorow,jess,dont let me forget*and then when i got back i called her agin and its been 1hour 29 min HEH. but i didnt get to see my baby today *SIGH* but i WILL get to see her tommorow! tonight is krazy *GET\'EM BRITT*lol ,but now i am talkin to Jess,sissy and britt..but i\'m bout to get off here cause my *fangers* are gettin sleepy,tired lol..but i will talk to ya\'ll tmmorow..be good! good night!! i love u baby!!   '),(8413,3848892,'07,July,2004','2004-07-07 00:00:00','   This mornin i got up around 11:30,i had to be at work at 12! so i got up took a 5 min shower,then i rushed to get to work and i was still 10 min early *how good am i* lol j/k.my baby came and seen me at work to day TWICE the first time she came to get some hamburger meat,and a lil bit of the blood got on her she was like *EEEEK* lol.The 2nd time she came in with mack and john boy lol.*Mack ur haircut really ain that bad,u just think it is* but after i got off work i went home to take a shower,and i got on here for a while and tal to a few people,i talked to *sissy* for a while then i got ready cause we were goin to pizza hut ,so i got ready and i went up to momm\'s work and waited on jess to call cause i thought i was meetin her at adrienne\'s,so she calls me and i didnt hear it the first time so i called back and she didn hear her\'s and then she called me back and told me that they were on their way to pizza hut * i was fused* lol..so i met them at pizza hut and we stayed there a while and talked and ate,then we left and when we got out to the car Jess gave me 2 cd\'s for me to burn johnathin and a Dale jr sticker to put on my car ;).I followed Jess and sissy to alicia\'s house and we waited on brett to get there,and then wehen he got there we left and alicia rode with me cause we had family bussiness to discuse *shhh*..when we got to sissy\'s house we sat there for a while and goofed off,then eric came * man do i hate that kid*lol...then i had to leave cause jess had to leave so i took brett home and when i got here mom gave me a speech about *drinkin* wtc!..and tommorow i get a new phone!! cause they said they were gonna quit makin phones like i have cause the cord kept breakin *go figure*lol...but anyway,here i am now and i\'m bout to get off so i will talk to ya\'ll later,be good! goodnight!   '),(8414,3848892,'06,July,2004','2004-07-06 00:00:00','   This mornin i got up around 9:30,Jess sent me a txt around 9 i belive, i dont know,neway,she said that she was goin to get some carpet and some other stuff i cant remember what it was but she said she would call when she got home..well 3:45 roles by and she calls and tells me she is home and that i can come over and then her mom tells her that she has to go pay bills and that she is goin to Nese\'s to stay...so i left the house around 4:30 i think and got up there about 4:55....when i got there,it was pourin down just a lil lol  and we started to watch \"Detour\" *that was the gayest movie*  but anyway,we watch a lil bit of that and Jess\'s g-ma called and said that the chilli was done *yummmy*.I ate two chilli dawgs and Jess ate a bowl of chillie and a dawg hehe. Then when we got through Nese and Kelly decided they were gonn watch it to. LMAO! Nese was gettin scurred,i couldnt help but laugh and kelly wasnt makin it even better lol *whew* but after the movie was over,we watched Brother Bear,thats actually a pretty good movie *ya* but i had to leave about 10 *sigh*..and hear i am now and i\'m bout to quit,but i\'ll be back tommorow to write mor..ya\'ll be good! goodnight! i love jess!!!!   '),(8415,3848892,'05,July,2004','2004-07-05 00:00:00','   Well to start off,this is my first time doin a blog *duhh* lol. But neway,this mornin i got up around 10 10:30 i think,Jess sent me a txt this mornin *aww! i love her to death!* and then i took a shower,watched a lil bit of t.v.,got on here to talk to my sis*ADRIENNE* i found a sister like 2 days ago lol.After i did that,i went to wash my car,then i went to go help jim move some machine stuff over at the industrial park,and that took about 2 hours,it sucked to,i got ALOT of oil in my hair and all over my cloths,i looked black and proud lol! j/k.I didnt really talk much to my baby today *SIGH* but she had to do a bunch of work today cause here mom was in a bad mood today...but hopefully tommorow we will get to do somn,but right now i\'m talkin to my sis and val,but i\'ll be back tommorow to write some more!! goodnight everyone the night is still young,but i\'m actually sleepy lol bye bye!!   '),(8416,3848892,'15,August,2004','2004-08-15 00:00:00','   Well people,its been exiting,boring,sad,weird,and...krazy...i\'m writing my final blog cause i\'m lazy and fat lol nahh,i\'m just not really into it anymore,jess and adrienne got me into it,it was fun at first then i started reading some crap from other peoples blog that made me so mad and i wouldnt write for a while but thats a different story....its so hard to say goodbye,but not to this,its like a virus on my puter*heh* i\'m deleting it,anyway,,i\'m gonna tell u what happened today...this mornin i got up around 9 or so,got somn to eat,went to foodworld to get some stuff so i could make my famous cheese dip,*secret*anyway...i came back home fixed it,the race came on i watched it for a while but i fell asleep until about the last few laps and tony stewart won*he shit all over himself didnt he britt lmao*...when that was over,cory called me up  so i went down there then we chilled at mack\'s for a while then i left there about 7...i called mom cause i wanted to go see my baby...so i went over there and sissy and christina was there..so we just sat around and talked for a while i had to leave about 9..i got home she  called me cause she was handling some bussiness but if it dont work i\'ma have to do somn i havent done since the 5th grade...but i guess that all i\'m writing for this final blog i wanna give some shootouts that have left comments...    Jess-Baby,we have some krazy moment and some sad moments but we have worked them out...but i love ui with all my heart and not matter what the situation i\'m keepin my kool and wanting u to have a good time..we have grown to know each other pretty good but there is still stuff that we dont know about each other yet cause it hasnt happened,confusing i know,but u know i will always be here for u and i hope that u will do the same...I Love U so much!!!   Mack-whats up man! nothin has changed,no dramma between us,everything has been pretty kool,except with the whole thing with u and chris but i wont  get into that lol its been fun,but its my time to go*i feel like i\'m writing my last letter* later man! we need to hit the gym!!    Adrienne-SISSY! heh wow,its like we have known each other for a couple of years but its really only been 2 months lol thank u so much for being here when i needed some one to talk to and just being a REAL good friend  and i\'m really thankful u have come into to my life*as a friend ;)*and its gonna continue to be a GOOD relation ship..and i want u to show this to tristan this cause every word i aid on here pertains to him to..bye bye sis!!!    Val-Well,were do i start lol me and u have been at each others throat at one time and that was just insane but its all over now,its all good...u need to quit gettin in trouble so damn much! lol ur gonna be grounded til ur 25 one day i\'m serious hehe but anyway its been great,but i\'ll ttyl ight! bye!!       Well,there ya go,my last blog....goodbye everyone,i\'m goin to another world now i\'ll call u when i get there,but i may not cause i might be roaming lol c ya\'ll later! bye!!!!!   '),(8417,3848892,'09,August,2004','2004-08-09 00:00:00','   Well people,its been exiting,boring,sad,weird,and...krazy...i\'m writing my final blog cause i\'m lazy and fat lol nahh,i\'m just not really into it anymore,jess and adrienne got me into it,it was fun at first then i started reading some crap from other peoples blog that made me so mad and i wouldnt write for a while but thats a different story....its so hard to say goodbye,but not to this,its like a virus on my puter*heh* i\'m deleting it,anyway,,i\'m gonna tell u what happened today...this mornin i got up around 9 or so,got somn to eat,went to foodworld to get some stuff so i could make my famous cheese dip,*secret*anyway...i came back home fixed it,the race came on i watched it for a while but i fell asleep until about the last few laps and tony stewart won*he shit all over himself didnt he britt lmao*...when that was over,cory called me up  so i went down there then we chilled at mack\'s for a while then i left there about 7...i called mom cause i wanted to go see my baby...so i went over there and sissy and christina was there..so we just sat around and talked for a while i had to leave about 9..i got home she  called me cause she was handling some bussiness but if it dont work i\'ma have to do somn i havent done since the 5th grade...but i guess that all i\'m writing for this final blog i wanna give some shootouts that have left comments...    Jess-Baby,we have some krazy moment and some sad moments but we have worked them out...but i love ui with all my heart and not matter what the situation i\'m keepin my kool and wanting u to have a good time..we have grown to know each other pretty good but there is still stuff that we dont know about each other yet cause it hasnt happened,confusing i know,but u know i will always be here for u and i hope that u will do the same...I Love U so much!!!   Mack-whats up man! nothin has changed,no dramma between us,everything has been pretty kool,except with the whole thing with u and chris but i wont  get into that lol its been fun,but its my time to go*i feel like i\'m writing my last letter* later man! we need to hit the gym!!    Adrienne-SISSY! heh wow,its like we have known each other for a couple of years but its really only been 2 months lol thank u so much for being here when i needed some one to talk to and just being a REAL good friend  and i\'m really thankful u have come into to my life*as a friend ;)*and its gonna continue to be a GOOD relation ship..and i want u to show this to tristan this cause every word i aid on here pertains to him to..bye bye sis!!!    Val-Well,were do i start lol me and u have been at each others throat at one time and that was just insane but its all over now,its all good...u need to quit gettin in trouble so damn much! lol ur gonna be grounded til ur 25 one day i\'m serious hehe but anyway its been great,but i\'ll ttyl ight! bye!!       Well,there ya go,my last blog....goodbye everyone,i\'m goin to another world now i\'ll call u when i get there,but i may not cause i might be roaming lol c ya\'ll later! bye!!!!!   '),(8418,3848892,'09,August,2004','2004-08-09 00:00:00','   This mornin got up on my own for the first time lol went to pick up blake and then we got to the school at7:20 or somn like that*i locked my damn keys in my car lol* anyway jess came over and barely talked to me then she went to go talk to some people and jon,so i said fuck it,got in the school she ran off with val so i looked for jessie and her and she wanted me to meet this girl from DAR she was kool,but i wouldnt hook up with her lol but she seemed kool,me and jessie talked until the bell rang and then i walked with jess around the school or w/e...my classes are pretty kool,i like them,but they could be alot warmer cause it was cold as hell in some rooms..after school i had to go to football,it sucked but we need work...got home,had to go to wal-mart,got down the rode,mom called and wanted me to come back so she could go,and i had told adrienne i would meet her up there and i was like 45 min. late *sorry sissy* but we talked for a while about what happened but thats all i\'m sayin..got home tryied on the cloths and here i am now sittin here bored and pissed and watchin wrestilin...bye bye! beee good everyone!.....I HAVE LUNCH WITH SISSY!!!!!!   '),(8419,3848892,'08,August,2004','2004-08-08 00:00:00','   This mornin got up,neck was stiff as hell,didnt feel to good..mom needed me to go get somn to eat,came back mowed the yard,when i was finished i came in to play NCAA 2005 for a coup;e of min.I took corey\'s lawnmower but i went to mack\'s house for a while and played football for a while...i think that game should have went down as the greatest game ever dont u think mack lol i was winning in the first half 21-10,3rd half he came back and got ahead of me by 2,then i returned one and now the score was28-23,4th quarter roled around and he was ahead by 1 cause he made the 2 point conversion with 6 seconds,he kicked it off and i ran it out of bounds at the 45 or 50 to stop the clock and i threw the hail mary and won it with 3 seconds to go lol...ANYWAY,,i came back home and tristan was here so he stayed and ate with us and he left about 8:30 or so and i talked to jess on here for a while lol i talked to sissy for a bit and so here i am now bored as crap and enjoying my last day of summer,kinda heh..well i guess i\'ll talk to everbody later bye!be good! and goodnight!   '),(8420,3848892,'01,August,2004','2004-08-01 00:00:00','   This mornin i got up around 11 or so waited around the house playin NCAA and LOTR with bernie until about 1:30.Sissy and Tristan came over to swim for a while *omg* me and adrienne were sittin outside talkin and tristan was in the pool laying on the float and from the way adrienne saw it he was drowning cause he had one arm on the float and the other one in the water and his head halfway in the water lol she was scurred BAD but i guess that just showes how much love was there today ;).after they leftmom came home and me and bernie started doin dishes and all sorts of  crap like that.around 9:30 me and bern went to meet mack at the gym to work out but mack didnt *lazy bastard*when we got finished we went to sonic to get a slushy and rode the square twice cause damn quinton stole my drink,fat m\'fer lol...well i\'m get off here but i may be back to type more i\'ll ttyl bye!! i love  u baby!!!   '),(8421,3599769,'30,June,2004','2004-06-30 00:00:00','   okay, okay so I\'m dramifying everything, but still... I\'m new to this html crap and get really figure too much out. First of all, how the hell do i get Haloscan to work? I\'ve been trying since yesterday afternoon. No luck, I\'ve looked everywhere for help. So now I\'m reaching out to anyone who reads and might help. Next, I have this sideblog thingy, but when I saw it on another site it looked really cool, and it just doesn\'t on mine. How do I box it in and make it bold or more defined? I know I have more questions but for now if anyone can please help me out with these, I\'d appreciate it a thoasund flowers. Oh, yeah one more thing, how the hell do i set up sound on here? I\'ve been to some blogs where you open up the page and music is playing!!! How is that done? or is that out of my league? Thanks.  '),(8422,3599769,'29,June,2004','2004-06-29 00:00:00','   May 2001:  Lonely girl meets lonely boy. Boy falls in love with girl, then slowly girl falls in love with boy. They laugh, share, communicate on such a high level it doesn\'t seem real. He is the boy of my dreams even with his bad habits and faults.    Present Time:  Boy has questioned what he really wants out of life. May never want to get married, doesn\'t think its fair to girl. Girl wants to make up her own mind about that, not let someone choose for her. boy wonders if he really loves girl or just doesn\'t want her to be with anyone else. Girl thinks he caught the \"boy\" virus, the one thats fatal to his feelings.(It must be going around) Girl wants to work things out slowly. Boy has her heart in his power. Girl provokes his theories and gets inside his shell every now and then. Girl remains lonely and sad.    No, I\'m not sure starting all over will make it work. I don\'t know what I\'m really expecting, all i know is that I\'m not ready to lay down just yet. And somewhere inside, I don\'t think he is ready for me to either. I blog all this cause i used up all my whine to my friends minutes. But in the end i think I just want to be satisfied in knowing I did all I could for me to be happy with him. Even if it only makes me feel better.  '),(8423,3599769,'29,June,2004','2004-06-29 00:00:00','   Ooh, I\'m so sad. Although God has sent down the rain to cheer me up, I can\'t help it. I want to throw a raging fit. You know, the kind where you kick and scream like you did when your mom refused to buy you a candy at the store when you were like 5. *sniff* Although I can\'t cry yet, I need to be a home for that first, I wish I could just feel peace. And believe I\'ve been looking everywhere for it. I\'m trying really hard not to be consumed. I think peace just might be playing a morbid game of hide and seek. It\'s kicking my ass in this game too. I used to feel so wonderful, peaceful, and just plain happy. Knowing that after being at work all day we could come and do nothing. Or just cuddle, or go to eat and have drinks or go to the movies, or rent one, or again just do nothing together. ee, thats the greatest part of being a \"grown up\" you can do whatever you want when you get home from work. Not like before when your parents said no to everything because it was a school night. I wish I could just....ARHHHHHH, why am I so whiny. Well it is my blog and I will whine if I want to. I annoy myself. I just want go back to May of 2001 when me and you firsst started and start all over again. I just want to feel realxed and peacful. I just want to stop whining now.  '),(8424,3599769,'28,June,2004','2004-06-28 00:00:00','   What is the DEAL?!?! C\'mon now, why can\'t I have a #1 Fan? I provide #1 fanness to someone. Why can\'t I ever be on the recieving end? (wah, wah) Okay I take some solace in knowing that I am in full \"pity for myself\" mode. So with that being clarified, let the ranting begin.   Says in her whiny, child baby voice:   Why can\'t I have you too make me see the silver lining in everything. Why can\'t I know that I don\'t have to always be the one that thinks or to work things out on my own. Why can\'t you do it #1 Fan? Where are you? For someone to over indulge me in gifts, know that they have bought me too much and spent way too much money. Or indulged me in  sweet nothings? And what the fuck is this guy looking at? -sorry. Or to call me and remind how wonderful my everythings are and to be so damm excited to call me you can hardly contain urself. ARGHHHH!! I\'m just feel really frustrated right now. I want a #1 fan, and not in the stalking sense either. See if I had never beed adored at one point or another in my life by you, I wouldnt know what the hell I was missing. But you showed me what it was like and now its gone. So #1 fan, wherever you are, wherever you escaped to, come back and find me. I\'ll be waiting.  '),(8425,3599769,'24,June,2004','2004-06-24 00:00:00','   Oh my God! They are coming!!! I can\'t wait. I wonder if I even be able to go. Money is limited right now. I\'ll have to fly up there or drive 5 hours and with the cost of gas it might be cheaper and easier to fly. Well see. I\'m gonna go look for tickets....  '),(8426,3599769,'24,June,2004','2004-06-24 00:00:00','   Okay, so today and yesterday have been clouded with schizophrenia. Not to mention that I now have a new illness to nurse: ADD. I mean how many mental illness can this girl get at once. While these are only self diagnosed, by a future psychologist I feel the need to have an intervention with myself. I mean one minute sad, one happy, one hopeless, one full of life and hope, one suicidal, one loving life, one angry, one non-chalant. SOMEone save me from me. So while yesterday was day three of BOOTCAMP for the brokenhearted I\'ve kept my word, one phone call per day. And I have successfully have not initiated an outing. But yesterday I did see him and had Starbucks. It was just like old times until the rain clouds rained on my silent parade. And like Gollum, I had one evil little me beside me telling me to tell him to leave him nice and good, he doesn\'t love you. And the calmer version of me saying Nope, I am not a quitter and this is what I CHOOSE to fight for. Until of course I wash my hands clean of him. Which I might never do. God, help me. I\'ve been reading LUCKY by Alice Seebold, just finished actually I had read Lovely Bones and once I like an author I keep reading what they write. Lucky is a tragic story bringing me back to where my reality meets real reality. Telling myself to quit being selfish and realize that your problems are zilch in scale to those who have been victims of crime. Or those with real recognizable problems. So to everyone whose heart aches like mine and to those whose heartache I will never understand-a moment of silence.  '),(8427,3599769,'22,June,2004','2004-06-22 00:00:00','   To Julia: Thanks for reading these random, yet psychotic words and maybe even enjoying them. ;),  To Johnny: Your words are inspiring, now if only I can get him to see it the way that you do... To Karina: Yes, breakups suck, and your right ice cream does help. So does vodka, lol.  For anyone else who reads these insanity pleas, and for those of you who\'ve read them, I just wanted to say the reason why we broke up is because of no reason. I know that doesn\'t make all that much sense and thats why hurt so much. Noone cheated on anyone or what not. At least not that I know of. But I always ask everyone: \"Do you think a breakup is harder when there is no reason or rather if there was one?\" My irrational answer to that is that to me breaking up for no specific reason is so much harder cause I dont have an outside person to blame. All we can do is point fingers at each other. And yes,I want to blame someone else for the downfall of my relationship because God forbid I did anything wrong.( making my shocked face, lol) My relationship to him was one that just fit together everywhere. You know how you can get when your in a relationship and everything seems just right? Thats how it was for me and him and it lasted that way for 2.5 years. Until all of a sudden we were fighting over the smallest things and it was hard just to be in the same room. We shared many of things together and those memories are etched in my head as if it were stone. So as of last night I started myself on a new round of BOOT CAMP. In this round, ladies and gentlemen I will not call him but only once a day and if he knows that I\'ve called and didnt get my call then it is up to him to call me back. But oh lets just say I call and block my number and he doesn\'t know its me it does\'nt count. What?, I was only kidding about blocking my number. wink, wink. LOL. So yesterday I called no answer so I left a message, and he called back within the hour. Talked for a couple of mins, and then bye. But then he called me later on at night before his beat started and we talked for 30 mins. Woohoo! 30 mins!!! My phone has a timer. And then this morning he called to wake me up and told me to drive carefully since last night produced some major storms and was still going. God, I love him for acting like he cares. So thank you God for letting him think of me and acting on it.  '),(8428,3599769,'21,June,2004','2004-06-21 00:00:00','   1. To get married, have a home and a family to love me and to love.  2. To have a second chance with Mr.Rich.  3. To help raise my munchkins, like I did before.  4. To be 100% happy with my surroundings.  5. To become a stronger person. 6. To get my degree, no matter how long it takes. 7. To never take life, or anyone/anything for granted. 8. To act my age.  9. To not stress out about things that are beyond my control. 10. To learn to just be happy.  I don\'t know if anyone has read these posts besides me. I don\'t even know how to find my own posts half the time. But if I\'m lucky and someone does run across this page, please note: I am not as ungrateful, or self absorbed as my rantings may make me seem. I\'ve been re-reading what I have posted and realize how psycho i may sound. And how obsessed with Mr. Rich I may be. But not so, these are just things that fly into my head and sometimes consume me. He is in my life yes, but not the way I want him to be. Mostly I\'m just justifying these rants to myself. But thats okay. Everyone needs a little reassurance that they are not as crazy as they seem every now and then even if it is coming from themselves. Great. Not only do I already talk to myself I\'m writing to myself. So much for reassuring myself, lol. Note to self: Consider yourself CONSUMED.  '),(8429,3599769,'18,June,2004','2004-06-18 00:00:00','   Thank you Christine for fixing my screwup. In case you read this.  So how come its okay for you to go out and then call me and tell me what you\'ve done after you\'ve done it but i go and have a couple of drinks at like 6 and i get the attitude. Ahh, you do care! lol. Anyways hearing the surprise in your voice that I actually left my apartment, it was priceless. And to think I wasnt even being mean when i said it either. How you didnt notice I was 2 drinks shy of drunkiness, I\'ll never know. So i decided you didnt need to know until today.Someone be proud of me, I\'VE ACTUALLY LEFT THE APARTMENT!Jenn knows this guy who\'s like 2 8and has been in the same slumped I\'ve been in, and get this its been two years for him! Lord, I dont know if i can stop myself from killing myself if it lasts that long. Maybe its time for the good time prescription as well as the good time drug. Ah, fuck it who cares. All i know is that i have an earache from hell and I\'m slowly dying from fever. what a way to start the damm weekend.  '),(8430,3599769,'17,June,2004','2004-06-17 00:00:00','   I know that I am a selfish person. I know that I have done bad things in my life that have hurt other people. I know that iI sometimes take things for granted. But please Lord extend me some compassion. I need your help. i can\'t function anymore. I\'m soo tired of being sad. and hurt, and confused. Please Lord give me a chance to regain what i once had. Please Lord give us a second chance. I promise to be a better person, an diI know for the last 6 months I\'ve tried really hard to be. Please Lord, help me. I miss my munchkins and my lovely boy. I need them more than anything and iI dont want ot live without them in my life. So if you can please find the time amongst the other million prayers (and i truly believe that there are more important ones than mine) to help heal my heart and intervene between me and him and help us regain our strength to be the couple we once wore. thank you Lord for all the blessings that I have recieved in my life, and for teaching me the biggest lesson: never take anything/anyone for granted. Thank you, Lord.  '),(8431,3599769,'15,June,2004','2004-06-15 00:00:00','   today sucks, everything is stupid. last night i thought me and lovely boy would go out. i mean he did ask for me to find out what time hooters closed. then only to find out he is on his way to sherlocks with the booking gang. ugh! no we didnt have specfic plans. but damm, this sucks. i wonder what its gonna take for me to confront him on all of this?? i actually never will most likely. he broke up with me and broke my heart and thats when i decided to not speak of it to him again. why cant i just be him? i think hes very lucky to have someone like me--who loves him  too  much. i want someone to love me too much.i want to be him. i want to not think. period. no thinking. i wnat to not feel sad at certain times of days or cry because of certain songs. i want ot not feel depressed all the time or cry while i type this. i want to be on the recieving end of phone calls not the person calling. okay, i\'ll give u that u do call me. but i call cause i miss u terribly and want to know whts new in ur life, to play catch up and pretend as if im still apart of it all. u call and ur not sad ur happy. is  it all fake? are u really sad? do u really miss me? my guess is no. theres no use in false hope. or any kind of hope for that matter. Please God read this lonely girls blog and help me make something happen for the good of us both.  '),(8432,3599769,'14,June,2004','2004-06-14 00:00:00','   well being sad really sucks. on sat. night the girls came over and blamed me for not going out with them. geez, as if thier sat. night depends on me. i now they just want to get me out of the house. but i just dont feel like it. i just want ot stay home and hide. but they stayed anyway. we did stupid girl stuff and it was okay. i miss him so much. i had a very sleepless night, and i can barely keep my head up right now. i dont know whats wrong with me. i just want to wake up and feel the same peace i used to feel, wake up and just be happy. i know my ultimate happiness does not rely on him but right now being with him will do me a world of good. i just want to love him and have him love me back. Please God give us a second chance. Please God. give me a shot at this one more time. i haven\'t been this sad in awhile. im true-blue sad today. or country music sad. maybe its the lack of sleep too. whatever it is i hope it doesn\'t last. i just want to be with him. in his arms. sleeping, dreaming and smelling him when we wake up. i just want him. Please God Please God.  '),(8433,3599769,'11,June,2004','2004-06-11 00:00:00','   so this is a letter to you: i miss you. i love you. days like this or anyday triggers up all sorts of memories. u just called right now. ahhhh. i wish i could just call u back 100 times. hmmm, maybe i can...j/k. i wish we could go back in time 3 yrs ago to when this all started just so that i can get the wonderful feeling again. or i wish we were at the mayan riveria again or vegas or the grand canyon. remember how beautiful it all was? how much fun it was? im so confused, i just want to scream to you everytime u call that i love you but i know that tactic wont work. so im being sly. trying to act like i dont care much. more for you than me. i want to just feed you all the wonderful times but you know i think that only works on girls. see if this were reversed and u forced fed me memories i would break down and cry and quite possibly give in. lucky you. but its you. and gettting to ur core is so much harder. i thought i seeped into your shell and i did but now im back on the outside and all u have for me now is a soft spot. and maybe i keep you at ur weakest which is something u greatly fear. its okay. ur still my monkeyboy. and i still love you.  '),(8434,3599769,'10,June,2004','2004-06-10 00:00:00','   well its God early in the morning. I had a good time last night, had a couple of drinks, went down memory lane, thought about how my life sucks and then back again to i have a good life. Which in all honestly i really do. i feel as if someway thinking that aloud is gonna make it go away. lets just knock on wood for good measure. well i called him at 3am, i know he\'s awake cause he\'s working and he did answer but then seem to get annoyed with me when i said my tummy hurt which it seems to be doing alot lately at 3 am. subconciously making it hurt in the middle of the night???? maybe. the body is a wonderous thing. my pysche is creeping up on me and maybe i really want it to hurt so that i can have a reason to call. damsel in distress syndrome. or not cause my stomach just gave me a kick for questioning its intentions. what to do, what to do. i miss him. if by some unknown power i can telepathically send u a message its this: call me, i miss you. lets get back together. there now thats coming from a true pathetic. i wish it would hurry up and just rain. thunder, lightning, the works please. i hate work. but i gotta make that money, honey. i swear that im schizo. one minute sad.  one minute happy, one minute confused, and then back to sad again. hmm, i think i need to concentrate on being happy. oh, but i\'m so good at being sad now. k, gotta go cause im annoying myself and prolly anyone else who will read this.  '),(8435,3599769,'10,June,2004','2004-06-10 00:00:00','   well it does work!!!yeah!!! well enough about that. why is everyone breaking up? are all guys mentals? lets see there are a total of four close friends i know that are going/gone through recent breakups. why must this crap continue to happen. and im talking long ass relationships too. doesnt anyone care anymore? and dont get me wrong we have all accepted our blame in the scheme of things but i guess it still sucks to be the dumpee. anyways do people even get to read this? if so please respond to me just so that your out there. thanks.  '),(8436,3599769,'10,June,2004','2004-06-10 00:00:00','   well only 15 mins of work left. noone to talk too. i cant believe that j.lo got married again. i want to get married. poor her, she was just looking for love. arent we all? dont anwser that. all i know is that i am, actually i found it and its slipping away. i keep trying for the life of me to hang on to it. i dotn know maybe im pathetic maybe just a hopeless romantic. who thinks love solves all. is that really so bad? hmmm, i wonder? i dont believe im a bad person. just one with lots of hope and love. if only he could see me-me, not just everything that once was or whatever it was that went bad. anyways woohoo!!! 4 mins to go. i wish it was friday already. time to go home, pay bills, drink a 40, and watch primetime tv. well minus the 40. im a girly girl~~~ and struggle with sitting on my hands to not call my lovely boy. im not obsessed, really. ;)  '),(8437,3599769,'10,June,2004','2004-06-10 00:00:00','   i just wanted to get into the whole \"blogging\" experiance. i wonder if this is free? it didnt say a fee anywhere... well im at work knocking off cause i want to go which i still need 2 more hours to go. i hate everyone. everyone is stupid. or at least everyone i know. i hate love. ah ha! thats an oxymoron. i think. lets see if this mammajamma works. here i go.....  '),(8438,3599769,'30,July,2004','2004-07-30 00:00:00','   Is it bad that even though I rarely work, I still got a quarter raise? Should I feel bad? Cause I really don\'t.   Yesterday was extremely uneventful, no surprise. Watched a bunch of useless t.v., gabbed on the phone. Very woe is me.   Tonights mission is to clean up the apt. I keep saying that but haven\'t actually gotten around to doing it. What fun. On a Friday nite. Ah, Phooey. I wish that workday time would whizz by, and then slow a little when I drive home. Traffic\'s gonna be a bitch.   My horoscope (all 50 of them that I read on a daily basis) say I need to stop reminiscing and focus on the now. Pfft. Easy for them to say, but I will try fervently to take this advice. So heres to looking at the now.&nbsp; And trying to appreciate it.   Why is it that noone\'s blogging incessently? don\'t you know that I live to look at other peoples blogs? c\'mion bloggers BLOG. If for anything, for this lonely girl to say busy without letting her mind wander into the past.  '),(8439,3599769,'29,July,2004','2004-07-29 00:00:00','   Last nite was pretty interesting. Me and Notorious DEST went to have max 2 beers. Don\'t I ever learn my lesson when I\'m with her? And besides don\'t ever tell 2 lonely girls that 6 count beer buckets are only 10.00 for the next 5 mins.   Then I suddenly remember how much I want to have sex. Almost like an awakening. I was very tempted to drunk dial lovely boy and ask/beg him to come over after work.&nbsp; But drunk dials weren\'t my humilation of choice last night. That was Notorious D\'s.   Mine was telling anyone within earshot how \"I just wanted to wear a ring on my left hand that meant something, dammit.\" That sentence stirred up all sorts of convos. I swear 8 beers and loneliness later I was unbelievably feeling pretty good. Must have been the beer. So I stumble home, and decide to drunk dial all my friends, which was just as fun.   Today my stomach WILL NOT stop rumbling. It just keeps gurgling.&nbsp;I don\'t like using the restroom here at work&nbsp; because it sucks. So I can\'t wait&nbsp;to go home. *sigh*&nbsp;  Elizabeth, I want to give you my email but I want to only post it up for a min. since it is both my first and last name. Do you have yahoo messenger? If so let me know.&nbsp; If not, Comment on here and I will post my email and leave it for a bit and then remove it. Thanks.  '),(8440,3599769,'28,July,2004','2004-07-28 00:00:00','   Can anyone please tell me how to add a list of sites to my blog. I would like to make a list/link of all my fav blogs, but can\'t seem to make it work. If anyone can please help it would be greatly appreciated.  '),(8441,3599769,'28,July,2004','2004-07-28 00:00:00','   I haven\'t read a good book in about 2 weeks now. I\'ve been too distracted to say the least. But reading is the one thing that takes me away from my reality and into someone else\'s fiction. So reading it is! i just need to find the right book.   I wonder what will be on T.V tonight. Nothing good prolly. I wish someone would come whisk me away to A Makeover Story. Or better yet, I will invent a new reality show and it will have the same concept except it will make you over from the inside. You will see various shrinks, try different drugs, attend self-help seminars. And you won\'t be able to look in the mirror for the next three months. I\'m sorry but those are the rules....  Its almost lunch time! And I\'m starvin\'.  '),(8442,3599769,'27,July,2004','2004-07-27 00:00:00','   It might make me feel better. It might not.  Damm crank yankers. Quit calling.   Who knew it would still be hurting this bad? I keep wasting precious time on being sad instead of being happy. Realizing you have a problem is the first step...  I feel so ALONE. I wish I could make him see where I\'m coming from. Yeah, right. If I would have known that this was gonna happen maybe I would have never allowed myself to give so freely. I want it back now.   I\'m so tired of feeling like this all the time. I\'m playing \"psych myself out\" games, in order to restore some kind of mental clarity. There aren\'t working yet.&nbsp; So if anyone has any new games for me to play please let me know.   I wonder if I have bad karma? I swear I try to follow the golden rule.   Woh, Woh, Woh.  '),(8443,3599769,'23,July,2004','2004-07-23 00:00:00','   51. My fav num is 27. 52. I love celebrity gossip.  53. I need to be needed.  54. I want to be with someone who loves me more than I love them. Selfish? See, I told ya. 55. This list is making me feel better. 56. I like to drive around, looking at houses. Pretending that I live in them.  57. I love to travel.  58. I want to live in Greece.  59. I\'m 5\'1 and weigh 117 lbs.  60. I wish I was taller.  61. I usually judge guys by thier shoes. It doesnt matter if they are ratty or not, they just have to \"go\" with you.  62. I can be pretty vain.  63. If I don\'t wear makeup I think I look dead.  64. Someone I love told me \"theres noone better than you\". But I\'m still alone.  65. I\'m LONELY.  66. I don\'t believe in finding someone else just to make me not lonely.  67. I need to be touched.  68. I leave the tv on all night. 69. I\'m addicted to watching reality shows.  70. I\'m want to get married. But we have to have a long engagement first.  '),(8444,3599769,'22,July,2004','2004-07-22 00:00:00','   Everyone has a list going on. So I\'m just gonna make one too.   1. I hate hair. Loose hairs. If its near my food, I swear I will vomit or start gagging real bad.  2.I was an over achiever in high school and after one year in college, I decided I needed a break- it took me two years to go back.  3. I make up things all the time. (words, songs, feelings) 4. I never thought I would be co-dependent.  5. But I am. 6. I\'m afraid of going to fun places. Its weird. I always know of fun places to go, I like to do research on whats opening and whats new but I now rarely ever go. Refer to #4. 7. I don\'t have alot of girlfriends.  8. My apt. is a total mess.  9. I never ever take out the trash. 10. I don\'t feel comfortable around people, and sometimes not even myself.  11. I think everything that happens is a sign.  12. I\'m starting to learn that they are just coinsidences. 13. I don\'t make enough money, but I\'m not doing anything about it either. 14. I\'m the oldest of 5 children, the youngest being 4. 15. If I really get to know you, I can prolly read everything about you. At least&nbsp; I have been able to in the past.  16. I\'m extremely selfish, unless I love you.  17. I compromise too much.  18. I\'m in love, and don\'t want to be anymore. 19. I feel like my \"real life\" should have started already. 20. It hasn\'t started cause I have that mentality most likely.  21. I feel like a stranger.  22. I\'m afraid to make decisions. Any kind.  23. I hate CHANGE.  24. I\'m obessive-compulsive. 25. I need to figure out what is going to make me happy.  26. This last year has been the worst in my life.  27. I wish I had a dog. 28. I thought I knew what I was doing, only to find out I\'ve been lying to myself all along.  29. I cry all the time.  30. I used to laugh all the time.  31. I\'m a shell of what I used to be.  32. I wish/pray goodness on everyone. 33. I have massive anxiety attacks.  34. I\'m constantly paranoid.  35. I don\'t do drugs. 36. But I smoke.  37. I like to drink, but always feel so guilty the next morning.  38. I want to start all over again.  39. I don\'t do any work all day, and my desk is a mess.  40. I used to be the \"lucky\" one. 41. It took me moving out at 18 for me and mom to really have a relationship. 42. I will never move back home.  43. I give advice, but never seem to take it. But I greatly appreciate it.  44. I wish more people would visit my blog and comment.  45. I wish I would have went away to college. 46. I need to wash my car. 47. I\'m on the road to recovery. A long road.  48. I haven\'t had that \"excited\" feeling in my stomach about anything in about a year. 49.&nbsp;I whine too much. 50. Sometimes I think about dying. But I\'m seriously not suicidal. Does that make sense?  '),(8445,3599769,'22,July,2004','2004-07-22 00:00:00','   what the hell happened to the comments? And no more options? (fonts, colors?)  '),(8446,3599769,'21,July,2004','2004-07-21 00:00:00','   Non-girlfriend, Bunny who stole it from some other Bloggers.   Ask me any five questions.   Bring it on.   &nbsp;  '),(8447,3599769,'21,July,2004','2004-07-21 00:00:00','   in Psychic abilities? Or in curanderos? (folk healers)   I\'m thinking about visiting one...  '),(8448,3599769,'19,July,2004','2004-07-19 00:00:00','   So here I am. Blogging. Again.  &nbsp; I just need to keep busy doing something. Besides sending him pyshic messages to call me. Which don\'t seem to be working. I hate working! I hate that everyone I know is working and actually doing work. I feel like sleeping.  &nbsp; Only an hour and a half to go. I wish I had a book to read.&nbsp; I just bit off all my nails, and they feel gross now. Not very attractive. Next comes the lip-biting. And then the inside cheek-biting. &nbsp; I\'m gonna be 24 in two months! I feel like a loser. I really thought by this time things would be different for me. I\'m just the&nbsp;marrying kind of girl. I believe in  love  and it seems to be one of my  biggest downfalls. Why is that the word \" love \" and&nbsp;\" hell \" conjure up the same color? &nbsp;Isn\'t that weird, someone at least got that straight.  &nbsp; I believe in alot of stupid things. Good for nuthin\' kind of things.&nbsp;  &nbsp; &nbsp;  '),(8449,3599769,'15,July,2004','2004-07-15 00:00:00','   Let\'s see, today I feel like shit. Less so, than yesterday. So that\'s a start right?&nbsp; i\'m in a pretty pissy mood. I have noone to talk to on the phone, seeing as how everyone actually works at work.&nbsp; Not me. I come here to blog.  &nbsp; I wonder why girls can always conjure up the worst scenario possible. If your not one of these girls ~congrats. &nbsp;In my head right now I imagine that my lovely boy is in the arms of some ho\'. Truth or not, its killing me. I also can picture him taking her to out to eat, I can even imagine his lips moving as he talks to her. What the fuck is up with that? Have I gone completly insane? &nbsp; I\'m serious, if it were\'nt for blogging, I\'d be dialing his number 100 times trying to figure out if any of my evil thoughts are true. So instead I can on here to type the words that are floating around in my head.  &nbsp; I also imagine that he has taken her to the house that we lived in together. She\'s walking around on MY hard wood floors. The ones I insisted we have installed. Shes sitting on my toliet. Eating from my kitchen. I know, I know, there no longer mine. I also fear that she is there with MY munchkins.&nbsp; Talking and playing with them. May she burn in  hell.   &nbsp; I think I smoked too much today already. I can\'t help it. Mostly it just gives me something to do as I pace around. I feel like my teeth are going to rot and fall out. I hate just about everyone.  Mostly I just hate myself. No use in blaming innocent people.  &nbsp; I miss my old life. I miss him.  &nbsp;  '),(8450,3599769,'15,July,2004','2004-07-15 00:00:00','   Can I at least get some quality T.v?  &nbsp; I did nothing all weekend but hang out in front of the tube. And there was nothing to watch. Give a lonely girl a break. Considering that my funds are down to zero, I spent my last dinero on ciggys. And a beer. I swear, I\'m no better than the whino out on the corner.  &nbsp; I\'m trying to feel better, but at this rate it may take a couple of years. If Timmy is God, then Timmy hear my prayers and get your ass into gear! I\'ve been waiting for a non-conventional miracle, and believe me buddy I\'m due.  &nbsp; &nbsp; &nbsp;  '),(8451,3599769,'15,July,2004','2004-07-15 00:00:00','   1. Try and be positive 2. Try and smile at more people, instead of thinking \"Fuck Off\" in my head. 3. Start wearing makeup again 4. Start using my hairbrush, dryer, and straghtner again too.  5. Get some sun. 6. PSych myself out whenever \"evil lovely thoughts\" infest and try and conquer my head.  7. Try and not fuckin\' drink everyday. I\'m such a drunkass.  8. Sloooooowwww down on smoking. Damm, girl! 9. Realize that it is ultimately his lost (which he has already admitted) and that there is nothing I can do to change his mind.  10. Try and laugh more than 5 times a day.    Tommorrow I will take it all back, but here\'s written proof that I once thought these sane, rational thoughts.  '),(8452,3599769,'14,July,2004','2004-07-14 00:00:00','   I was a truly happy girl. I used to be on top of the fuckin\' world. What happened? I just reread all my previous blogs and I\'m dying. Inside I\'m just dying. I still cannot believe that I feel this way, that I am weak enough for someone to trigger all this. Our breakup, in almost every true sense is like a divorce. Except I don\'t get to see my munchkins. Or my house.   Yesterday he said, \"In my mind you represent everything that I want, stability.\" Thanks alot fucko, I can clearly remember about 3 yrs ago when he asked me what I was looking for, that was my exact answer. See thats why I didn\'t ever think I\'d be in this situation~we were looking for the same thing. And then he had to go and fuck it all up.  I asked him when was the defining moment (if any) when he wasn\'t in love with me anyore. He said there isn\'t one. So I asked well then why did you fall out of love with me?  His reply was \"i\'m stupid, and I don\'t know what I\'m doing.\"  Once again, thanks alot fucko.  Someone just please come and take this pain all away. I\'m so tired of hurting.  '),(8453,3599769,'14,July,2004','2004-07-14 00:00:00','   How much more pathetic can I get? Seriously. I want to wake up and feel so free, free from all of this. I can\'t even stop crying as I type this. And I just want to go home.   I\'m such a whiner.  '),(8454,3599769,'14,July,2004','2004-07-14 00:00:00','   Besides obsessing.   Last night me and bro and his friends drank ourselves silly. Again. What the hell is wron gwith me? I once again, drank and smoked too much. I feel like shit this morning. And on top of all that my lovely boy called again last night.   But it wasn\'t really good news. So I guess its just truly over and its time to move on. But why don\'t I want to. I feel like throwing a screaming, crying fit.   Why do I hate feel like I hate everyone? Why do I hold on? To nothing?  '),(8455,3599769,'14,July,2004','2004-07-14 00:00:00','   Let\'s see in my pocket I have a sample of a low-grade Xanax pill. But do I take it? I\'m afraid. What if makes me sick or worse, more anxiety-sticken? I really want to take it , to see if it helps.   I don\'t want to learn any more of \"life\'s little lessons\". Is there any way to stop them? *sigh* I\'m just being a whiny child baby. And also I\'m feeling sorry for myself. Again.  '),(8456,3599769,'13,July,2004','2004-07-13 00:00:00','   I feel like playing practical jokes on everyone today. And the weird part is I think I just feel like doing it because I feel like being mean. Isn\'t that some form of mental illness?   Well so much for positive thinking. It\'s not like I want to hurt anyone or anything, I just want to play jokes but kinda of mean ones, for my entertainment purposes. I\'m prolly gonna go to hell for thinking that way, either that or get some kind of bad karma sent my way.   I hate work, I had the option to leave yesterday but I never did instead I just stayed on the phone all afternoon doing nothing. But now I wish I could go home and sleep today. I\'m just tired. But thats prolly due to the depression.  I\'m hungry, I want to eat movie nachos. And popcorn. How weird. But I have this huge craving for it.    Why would I want to be with someone who didn\'t want to be with me?  '),(8457,3599769,'13,July,2004','2004-07-13 00:00:00','   I copied whats on my Sideblog from Olivia @ notsopinkanymore.blogspot. Sorry if I knew how to make links I would have linked you. Thank you.   This is my new mantra.  '),(8458,3599769,'12,July,2004','2004-07-12 00:00:00','   Here I am blogging, to see if this will help ward off an attack. I\'m typing faster as if its gonna catch with me.  I\'m trying to pinpoint what initiated it. wait, I can\'t reallly keep a straight thought. I think it has alot to do with not facing the truth. I don\'t think I\'ve faced the truth alot here recently. A year ago, my highschool sweetheart died and I still havenn\'t ever really think about him being gone. I kinda blocked it out. And now these things with my lovely boy have happened, and somewhere I have this stupid hope that releases in small doses. You know, the kind where he all of a sudden comes crying back. What the hell is wrong with me? Oh well. Maybe after my dr.s appt. things will be looking up and maybe these feelings well subside somehow, some way.  '),(8459,3599769,'12,July,2004','2004-07-12 00:00:00','   Since my boss won\'t let me leave right away, I\'ve decided I\'m going to blog. Again. Ha. I told her tha I wasn\'t doing any work. She\'s paying for nothing which isnt a bad thing, but i just really feel like I need to go home. i\'m tired and I just want to watch tv, and not think. But if I just sit here my mind just wanders off, and then I want to start crying again. I\'m ready for this to end. I think I might be able to convince myself that I\'m gonna let him go. Or at least try and fake it. Hmmm, since I don\'t really have a choice I guess we\'ll see.   I used to always tell \"Think positive\" and positive things will happen. So I\'m going to try real hard to be more positive from now on.  '),(8460,3599769,'12,July,2004','2004-07-12 00:00:00','   I\'m gonna blogging all day until today is over cause I\'m borderline gonna flip out and close shop and go home.   I called my mom crying just a bit ago. Poor her, she was fast asleep. She asked me whats wrong, and I told her the same thing. She said you need to let go, i know you don\'t want too but you need too. Don\'t let him get to the point to where he hates you. I think thats the best advice my mom\'s ever given me. Then she broke down and cried. She said she knows I think its the end of the world, but it will get better. We all loved him, but you need to let him go. Trying to make him feel something that he doesn\'t isn\'t gonna help the situation. In the end it will all work out, you watch. I\'ve never wanted to believe my mom so much before.   So God, if you read blogs please stop by and read mine. I pray that everything works out for all of us. I pray that someday things do get better and that I will reflect back and still think that all this pain was worth it. I pray for everyone who has been there for me. I pray that I will somehow manage to get my feelings in check, and learn that not everything goes the way that you think it will.  '),(8461,3599769,'12,July,2004','2004-07-12 00:00:00','   This morning I feel like shit.  Why do I keep thinking that drinking will make me feel better? Or that smoking would? Not only am I physically sick, Iam still emotionally fucked up. I made myself throw up yesterday just to get the room to stop spinning. Mornings have always been the hardest for me. I haven\'t slept a peaceful night in a really long time. I\'m talking at least a year. Mornings are the saddest, I wake up and realize my life is still as fucked up as it was when I went to sleep the night before. I think me and my lovely boy are through. I told him yesterday while I was buzzing that whateva girl he\'s talking to is gonna fuck him over. He stayed real quiet, cause I\'m usually the one the encourages him not the one that tears him down, then he said thanks and said he\'ll talk to me later. Which of course, my drunkass could not accept. So I called back but no answer. So then I called his mom, and she repeatdly reminded me that no boy is worth this much pain. How come even thought I know that, it still hurts like hell and I can\'t make it stop. I think the mental ward is gonna come looking for me too, I remember vaguely telling her that I just wanted to die. I\'m sure that didn\'t sound too healthy.    What the Fuck am I doing?!?!  '),(8462,3599769,'12,July,2004','2004-07-12 00:00:00','   This blog is repetitious and is filled with my feelings, but after this weekend I needed an outlet to say what\'s on my mind.  Take them from my head and my heart and keep them. I don\'t want them anymore. I wish that I had never met you, even though I\'m glad I did. I wish that I would have never kept any of my promises so I wouldn\'t hold you to the ones that you made to me. I wish that in my head there wasn\'t someone new in your life, even though in reality there really might be. I wish I didn\'t care so damm much about my munchkins. That I would lay my life down for them. I wish that I could open the top of my head, and with a scapel scrape the bits and pieces that you have infested. I wish my favorite part of the day wasn\'t connected to you in any way. I wish that I could stop saying \"I can\'t believe that this is happening to me\". I hope that I can in the near future say those exact words because everything is so great. I wish that I could be the one who doesn\'t feel the same anymore. I wish that you would have never have said \"This is it, this is what its all about.\" They meant so much to me. I wish those words were true. I wish I never had you on a pedestal as high as the one your on, cause no matter how high I jump I can\'t knock you off. So you can take all the memories and choke on them and realize that that\'s how I\'ve been feeling. I know what we had is over and facing that reality doesn\'t really make me feel any better. I\'m not stupid, or naive. I know that we must/will go on. I just never imagined it would be without you in my life. You took so much so I\'d like it returned. And in return I will gladly give you our memories. So yes, I still feel the need and want to be with you because I love you not because I don\'t think I can\'t live without you. I know that I can live without you, I just think life would be a lot more fun with you and my munchkins. I hope you find the happiness you are looking for, the one you couldn\'t find with me anymore.  '),(8463,3599769,'09,July,2004','2004-07-09 00:00:00','   But he may also be my savior. Today is move-in day for him. Move-in to my apt. complex that is. How funny. He is 19 yrs old and is the cockiest lil\' boy I know. See, me and him were never really close, and now all of a sudden he is going to live a couple of doors away from me. (I moved to my grma\'s house when he was in 5th grade) He just called me, seems all of a sudden I\'m invited to his party tonight at the new place and seeing as I\'m the only one legal enough to buy the drinks if I\'d mind stopping at the liquor store and picking him up something. Go figure!  But thats okay, he doesn\'t know the wrath of his big sis\'s depression. LOL. Late nights at his place to keep me busy should be fun.  '),(8464,3599769,'08,July,2004','2004-07-08 00:00:00','   I wish I was irresponsible. Like truly, really irresponsible. I want to spend all my bill money. I want to stay up all night and drink, and then  sleep all day.  '),(8465,3599769,'08,July,2004','2004-07-08 00:00:00','   Me and my sistor (its gotta have the  o )  connected. She was about 4 yrs old.   I get into my mom\'s car and my dream was still so fresh in my mind, I keep thinking about how I had dreamt about feeding human legs (detached from their bodies) to a hungry shark. Like it was an attraction at SeaWorld or something. Weird. But then my sis looks at me and says \"I had a dream about a shark yesterday.\" I completely freaked out! I hadn\'t even told a soul about my dream, I had just woken up about an hour before she came. She just looks at me and then tells me to put my seatbelt on. Now she\'s the most petite 7 yr old I know and such a ham!  Last night I dreamt about a shark and a whale. And a fire that destroyed the tank that they lived in, leaving them only in shallow pools of water. I\'m standing there freaking out thinking one of these massive creatures is going to swallow me whole.   My derangement has followed me into my sub-conscious world. Now, the sleep Gods cant\'t cut me a break. What next?  '),(8466,3599769,'07,July,2004','2004-07-07 00:00:00','   I looked at the clock, exactly 3 I knew that he would be getting ready to leave for work and it\'s his first day in gang. I wanted to call him so badly, but I called Jenn instead to keep me busy. Line 2 rings, I get off the line with Jenn. It\'s his voice. I already know but I asked who it was anyways.   You: Busy? Me: Uh,nuh. YOu: I can\'t hear you. Me: I said no. You: Do you still have the pic of me in full uniform? Me: Yeah. (Its still posted at my desk dumbass) You: Do you think I can have it? Me: (biting my lip) You can have whateva you want. You: I just want that pic, in case I die I\'d like to have a good pic of me. (his job safety just went down about 500+ points so very much a possibly) Me: Not funny. Well if you die, I\'ll make sure they have that pic.  You: I just want to frame it and put it in the house. Me: No. YOu: Please. ME: No. (He already knows I\'m gonna give it to him) You: Yep, first day today. I\'m gonna bring home a cruiser soon. My uniform is really hot I\'m sweating and I\'m in inside. I\'m gonna have to show you my uniform sometime.  Me: Uh, huh. Well, I\'m very happy things worked out so well for you.  You: (sullen) Yeah. Oh, I need to put my contacts on, call you back.  Me: yep.    click  '),(8467,3599769,'06,July,2004','2004-07-06 00:00:00','   Okay lets play the guess my future game. I have nothing to do at work, or rather I refuse to do anything at work so I thought I would post once more. Keep those comments coming...I\'m depending on you guys.   1. Lovely boy will... a)flip out when he figures that I wont always be there b)won\'t care, and will never call again c)is already seeing someone else so it doesn\'t matter  2. Will I... a)really move on b)take anti-depressents c)continue to drink and smoke in a empty attempt to make myself feel better   3. Chances of us getting back together... a)don\'t hold your breath b)only if like non-girlfriend says about him taking meds c)most likey, but then again maybe not   4. If I see him out with another girl will I ... a)throw a drink in his face b)make sure he sees that I see him and shake my head in disgust c)run out crying   5. Will I get to talk to him... a)in the next 5 days b)never c)in a month or longer   Why are boys such Namby Pambies? I used to call him that when I thought he was acting cowardly, he used to practilly spit nails in my direction. Ha! You Namby Pambie. Oh, you know you can\'t touch a damm boys ego.  '),(8468,3599769,'06,July,2004','2004-07-06 00:00:00','   Isn\'t that a wonderful combo. I\'m serious, last night i decided I was gonna go to the poolhall/bar up the road from me for a beer by myself. Next thing i know a man is sitting next to me, talking about life and marriage and such and its like 845pm. I got there like at 530pm. So it turns out he lives in my complex, the thing is I think I told him which apt. I\'m in and now I\'m scared. I didn\'t in any way feel threatned but I acted so irresponsibly. I never even go to bars by myself, much less talk to some stranger. So all night I kept thinking to myself he\'s gonna knock at my door and I\'m gonna freak. Not to mention, I\'m so freakin\' paranoid about everything, and that just escalated it. So now I feel like shit, I look like shit. So then I get to sleep like at 12am, and wake wide awake at 230am and stay awake until 5am. I\'m freakin\' out majorly. I called him last night too, and he was getting a fitting for his uniforms, small talk for like 3 mins. but that was it. I know, I\'m not gonna call him anymore. Nothing much happened with his mom at lunch, I stayed away from bringing him up but she volunteered some info on what he did yesterday and such but nothing enlightening. I swear I\'m gonna need anti-depressents soon cause I don\'t feel like myself at all anymore. I feel like I\'m invisible and I feel downright scared.  '),(8469,3599769,'06,July,2004','2004-07-06 00:00:00','   Today. Fuck!   It\'s the he\'s found someone else, you can do better even though I\'m talking about my own son speech.  '),(8470,3599769,'02,July,2004','2004-07-02 00:00:00','   Step right up, ladies and gentleman.   The one the only, The Whipping Girl.  No, you cannot whip her, but please feel free to snowball her with your spew of hateful words. C\'mon go ahead, she\'s already at her lowest point in her life, nothing you can say now well destroy her.   ( Crying )  You: We need to talk. Me: (heart drops) Okay. You: i really thing we should put some distance in between us. Me: (knowing thats how breakups usually work, but doesn\'t want to believe it) Why? You: Because, I\'m selfish and I only keep contact with you so that your not with anyone else. Me: Duh. You: I\'m sorry. ( my reasons are somewhat the same) Me: Why? I\'d like a explanation. You: Because I don\'t feel the same way about you anymore.  Me: I need more justification, what does that mean? What are u looking for? How could you pretend like I never existed? Why are u making me invisible? I hate you. You: I\'m not making you invisible, I just have no intentions of being with you and its not fair to you or me. And if I knew what was making me do this I would make it stop.   Me: Are you fuckin\' serious, all I wanted was a fighting chance. And you just want me to say okay so that you can have enough validation to be with new people. And if you think that my munchkins have stopped loving me or thinking about me this whole time your wrong. WRONG. You: (breaking down, sniffling) I know that. I just, I\'m sorry. I\'ll call you back. Me: Please don\'t let this be the last time I talk to you. Please.  You: No i promise you it won\'t.   Me: crying, hysterically. Wondering who the hell up there thinks it funny to make me cry.  I know that I took some easy shots, and more was said, but thats pretty much the whole of it. So, please someone I know that breakups are supposed to be like this and are supposed to hurt, and that they make movies about this all the time but damm. I hate being the launch pad. Next girl he meets he will have realized that he does want all the things that we once had, and he will marry her.    So for all the boys out there, who treated that one girl real good but decided to let her go, you suck. and you never would know how good life with her could be because you didn\'t try.    Once he told me that my best talent was that I was a survivor. Up yours.  '),(8471,3599769,'02,July,2004','2004-07-02 00:00:00','   So today, I\'m thanking God a thousand flowers that\'s its Friday and Monday is a holiday and I get paid for it!!!! So if that\'s not something worth jumping up and down for then I don\'t know what is. I\'m mad busy today at work but I don\'t care cause I\'m having Chinese buffet for lunch with Rich\'s mom. BONUS!!! And another bonus is he will be back in town today. U were gone too long hunny bunny. Ahh, at least I can still baby talk him on here.Last night me and Jenn attempted to go to the gym but they wanted to charge her 15.00 for one day!! Shame on you Gold\'s Gym! So instead we went back to my apt. and did beginners yoga from the on-demand channels. It actually soothed my aching back but today\'s the pain is back. I wish it would have kept raining over here but the sun keeps coming out.    Dammit, I just peeled my cuticle and its bleeding a red river. Especially when I do this... ouch.Maybe I\'ll bleed to death and then everyone will be sad. And they will say all the customary stuff about me when I die. Like, I was such a giving person that I would give up the shirt that I was wearing if you needed it. (In real life, it depends whether the shirt is name brand or not in which I case may have paid too much for it and I\'d still like to wear it.) Or how I was so thoughtful. (Did I ever mentioned selfish in my profile?) Or how I had my whole life ahead of me and that I was someday going to become the PSYCHOLOGIST I wanted to be cause I\'d like to help unyielding persons. (tru dat!) Or how i was loved by many, okay, okay considering family and close friends I would believe it if they added this in my eulogy. Well I guess today that is not a possibility cause I\'ve already stopped bleeding. Planning my death comments will have wait, hopefully wait a realllll long time. Oh, I just remember don\'t you hate running into x\'s? Well Jenn ran into one of mine, well actually his sis, the girly is prego (me and her were really close, but due to her asshole bro I cut off all contact with thier whole fam so I wouldn\'t keep getting sucked back in) turns out he now lives with his gf and is leaving to the navy for 5 yrs. what a jip! I was his first serious relationship gf and now I bet he\'s this great guy and is prolly gonna marry this girl. C\'mon who waits 5 yrs for u to come back unless your gonna get married? Once again, I\'ve been shafted. Who cares he was a fuckin\' loser anyway. Green-eyed monster, anyone? But he really was a loser. So whateva.  '),(8472,3599769,'01,July,2004','2004-07-01 00:00:00','   How can I post pictures on here? I know about that HELLO program, but does that only work with pictures I have saved in a folder? If, so I cant remeber how I got those pics in the first place. Also, how do I create a link list or something that shows other blogs that i have been reading?  If anyone can help, thanks.  '),(8473,3599769,'01,July,2004','2004-07-01 00:00:00','   I\'m feeling too frustrated to do anything today. My workload is too big, and i just feel like breaking dowm. I hate feeling so melancholy. I was fine, but then I had to cancel a client who was\'nt qualify so now I\'m back to square 1. I just wish I could fly to Greece, where in my mind nothing bad happens. Or to the make belive world that Legend takes place. All those posies flying around and unicorns playing. Minus the part where everything turns dark, because everyone knows your not supposed to touch the unicorn. I dont\'t even know what I\'m rambling on about now. I just wish everything was my definition of normal.  '),(8474,3599769,'13,August,2004','2004-08-13 00:00:00','   Its Friday the 13th. Is anyone out there superstitious? I am, but not the whole step on a crack break your mothers back kind just the knock on wood kind.  Surprise, surprise. Lovelyboy showed up last night on his new toy (motorcycle) asked me if I wanted to go for a ride. I said sure, be right back. I needed to put on something motorcycle-ish. I think my white see through wife-beater shirt with a black bra clearly showing and tight Gap jeans did the trick. Seeing his expression only gratified my semi-slutiness. Ha! He knows I hate motorcycles, on my 19th bday I actually saw two people get thrown from one and the girl died. Ever since then I\'m afraid to drive near them much less get on one. But the other day I did and it was fun. And when he picked me up on it yesterday he asked if I was afraid and I said not anymore. I don\'t have to talk while we ride and its so relaxing. Minus the part where we turn and I feel like I\'m gonna get thrown into the street. Also, I think he likes the part where I have to hold on to him for dear life. Who knows with him, I seem to be wrong a lot. Anyways that was a quick update. Oh, and I\'m proud cause when he brought me back he said okay, well I need to go all I said was k\' bye be careful. He did give a couple of good hugs and stared into my eyes a couple of times, but only Lord knows if he does that to the other girl as well. I\'m just gonna pretend its only for me. Pfft. whatever. (I also didn\'t call him, like I would have before, see, I\'m really trying this time.)  Anyway, its Friday and the weekend is here once more. Thank God. Although I hate Sundays. What is everyone up to this weekend? I think I\'m going to the movies (another improvement) to see \"Little Black Book\" or something girly like that. I don\'t know. Not much to say really just bored and thought I\'d try and post something instead of reading everyone else\'s. More later.  '),(8475,3599769,'12,August,2004','2004-08-12 00:00:00','   Pfft. Whatever. That\'s what my online fortune cookie said.  Haven\'t I made it clear that I DO NOT like change. Of any sorts. Unless its changing my clothes. This week already has been such a trying week. Happy to say I\'m still here.  So I found out that he is talking to someone new. I just asked him, told him I wanted to hear the truth from him. He said yes, but not like how I think. Whatever, what am I supposed to think. So somehow him actually admitting to this put things in a little more prespective. I don\'t know why, I guess I just wanted him to say it so I can stop deeming him a liar. He then preceeded to tell me that I was the most important person in his life, but that he wasn\'t the right one for me. Thanks. For some odd reason I thought that was my decision to make. So thats it. For now, why lie? I have not called him or anything since that convo. And he was the one to call me in the first place, he wanted to show me his new toy.  So I guess everything comes to an end. And somehow I don\'t feel all that rotten. I  still hurt, but I think the fact that he has told me the truth somehow let me me mourn in a different way. I have no hope left for us. I just know that someday he will think of me. I will try and distance myself as much as possible but I am weak so I know I will succumb sooner or later. Hopefully way later.  As for blogging, it seems to be my highlight everyday. Well, reading others that is. I always want to post but then I just don\'t do it. I guess I felt like I was whining too much and the masses where gonna come and tell me to quit being a baby already. But I\'m throwing those thoughts out. If I need to vent I will. Off to start a new and somewhat better me. I know that I am capable and worthy of having a love so strong that it could withhold anything. I just need to stop pushing and stop looking.   Whatever, I\'m just glad that I am making it one day at a time.  '),(8476,3599769,'09,August,2004','2004-08-09 00:00:00','   Apparently I do. And not by choice.  Friday night: My bro and I decided to have too much too drink, again. I started off right, with shots of Rasberry vodka and somehow ended up with Budweiser. WTF? Anyways I played bathroom patroll for my bro since he kept throwing up every 5 mins and his gf was nowhere to be seen. I stumbled home and passed out.  Oh, and no drunken phone calls. Woohoo!  Saturday: Did alot of driving around. For no reason. Couldn\'t sit still, but didnt really want to talk to anyone. So I just drove and drove. So depressing. But I did manage to visit my friend at the cemetery. It was nice talking to him, although stupid inconsiderate me forgot that his dad, (who died a year later, almost to the date) was buried right next to him. I didnt even think to talk to him too.  So after I left and realized what I didn\'t do I felt like shit all over. That was only the second time I have gone to his gravesite in the year and half since his death. The first time wasn\'t even at the time of his burial. So I decided to turn in early, but at 10pm I hear someone banging on my door and get this: I grabbed my ciggys as if it were a gun. Oh my God, when I realized I had done that I couldn\'t stop laughing. It turns out it was my bro, crying. I thought something really bad had happened. It turns out him and his gf broke up and I\'m talking this boy was screaming crying, all huffy and couldn\'t breathe. I asked why he was so upset, they were always breaking up. He said that she told him she didnt feel the same way and that it just wasnt gonna work. (Sound familar?) All he kept mumbling was \"This isn\'t supposed to be like this, if it doesnt work for a reason, okay then. But its not gonna end because we just didnt want to try. I am not gonna fail at this.\" Once again, sound familiar? I wonder if it runs in my family? Is it contagious?  Sunday: I went to church, and couldnt stop thinking about how cold I was. After that had lunch with my aunt, then went home. Tried to nap but didnt succeed. Then my bro went and picked up my sistors and we tooked them swimming and ate pizza and watched movies. It was fun, but reminded me terribly of my munckins. They just got home on sat. from a 2 wk stint with thier mom. Lovely boy was off on sun. and prolly spent some quality time with them, time that I miss dearly.  Sunday night finally closed and I was thankful. Mornings and Sundays are the worst for me.  Thank God the mornings almost over.  '),(8477,3599769,'06,August,2004','2004-08-06 00:00:00','   So today has not been a very good day. I have retorted (once again) to self pity. I have also in my head conjure up the thought that my lovely boy will go out tonight with another girl. Tonight is no different than any other except that his schedule does not permit him to have a Friday nite off like tonight. So therefore in my head he will be doing fun things with another Girl.  I know that I am crazy and today only magnifies my craziness. I\'ve even planned there outing for them. I don\'t think I\'m gonna make it through the night. Why do I do this to myself? Even though I know that it doesn\'t necessarily need to be a Friday for him to take a girl out but what better night than tonight. I haven\'t talked to him since the episode on Tuesday.  Oh, God or Timmy whoever can answer prayers faster, PLEASE, PLEASE let him see the errors in his ways. Okay, okay a little dramatic and maybe there not all his errors. But dammit, I \'m willing to give it another shot. (no surprise)  HE DOESN\'T LOVE YOU ANYMORE. I realize that, but it doesn\'t seem to make things on my end any better. I wish I was addicted to something else besides him. I know I sound pathetic, I just couldn\'t help blogging once more on why my life is shit.  Self- pity session ends now.  \"Nobody loves you when your 23....\"  Okay now.  '),(8478,3599769,'04,August,2004','2004-08-04 00:00:00','   Another wonderful idea from the NG herself.  1. I blame the Marlboro company for my inevitable short life. 2. I blame my father for blessing me with co-dependency. 3. I blame my mother for keeping it real. 4. I blame lovely boy for everything. 5. I blame love for treating me like shit. 6.I blame The Cure for keeping me depressed. (Not really, seeeing as how I lOVE them) 7. I blame fate for cursing me. 8. I blame my boss for my bad work ethics. 9. I blame my car for needing an oil change. 10. I blame the sun for not enough Raineydays. 11. I blame Non-Girlfriend for her wonderful site, and inspiring me to start my own. 12. I blame blogging for my compulsiveness. 13. I blame hope, fear, love for the person that I am today. 14. I blame my bro for non-stop drinking. 15. I blame Notrious DES for my rowdy behavior while drunk. 16. I blame Jenn for my thoughts on contemplating my own death. (Ha!) 17. I blame society for allowing me to grow up in a cruel, cruel world. 18. I blame boredom for browsing blogs every 10 mins. 19. I blame money for making me work. 20. I blame Mexican food for keeping small love handles on my waist line.   I will so definitely have more to blame in upcoming posts.  '),(8479,3599769,'03,August,2004','2004-08-03 00:00:00','   I need everyone\'s help. Today after 5pm I will be meeting with lovely boy. Mostly I just needed to see him. I need to hear his voice. I need to know what to say. I know all the things I\'d like to say but I also know I need to keep some diginity.  I\'d like to say that I love him more than life. That I need him. I need to be around him to feel him. But I know that I shouldn\'t. It wouldn\'t make the least difference, I\'m sure. I need him to know how I feel without going insane. Its eating me alive.  So please anyone point me in the right direction. I know that my relationship is over with him, and that I feel the loss more than he. I just need the right words to get my feelings across. Any and all comments would help. Thank you.  '),(8480,3304415,'31,May,2004','2004-05-31 00:00:00','   Well saturday night was good and bad at the same time.  Noel talked about drinking and what the Bible has to say about it. It was really good.  But our band sucked for some reason, especially me.  We just sucked.  I hope it doesn\'t last.  '),(8481,3304415,'28,May,2004','2004-05-28 00:00:00','     '),(8482,3304415,'27,May,2004','2004-05-27 00:00:00','   Dead end streets are supposed to be peaceful right?  You move to a dead end street to get away from the traffic, to have at least one side without neighbors.  Living on a dead end street isn\'t a place for work.  It\'s not a place for \"ministry.\"  It\'s a place to relax to recharge before going back to the emotionally draining life of ministry.  Or is it?  There is a portable basketball hoop across the street from my house.  My new house.  My first house.  Everyday there are a bunch of neighborhood kids who play hoops there all day long.  But they don\'t just play in the street, they play in our driveway.  Their bikes are always in my yard.  That\'s an insurance liability isn\'t it?  There is barely enough space for me to back out of my driveway as it is, without a portable basketball hoop, with dozens of bricks piled on the back of it, just sitting in the middle of the street.    Sometimes I feel like taking out my paintball gun.  Maybe even putting on my camouflage face paint that makes me look like a crazy army ranger dude.  I wouldn\'t shoot anyone. I would just walk around my yard with my paintball gun.  Or maybe set up practice targets in the front lawn. I would turn up the speed as high as it goes and cuss under my breath, but loud enough for anyone out in the street could hear me, as the paintballs clang against my trash can lid target.    This would drive the street kids away. The kids who think they are PDiddy.  The kids who think they are Alan Iverson.  The kids who talk trash and swear at each other.  The kids who rest their basketballs on my car.     Maybe if I play Wilco all the time and loud enough, they would get sick of hearing alternative-country and leave.  It would have been the equivalent of me watching TNN with my grandma as a kid.  She knew how to get me outside.  Turn on Randy Travis and the Grand Ole\' Oprey and I\'m outta there.     So what do I do?  The exact opposite of what I really wanted to do.  I play some 2 on 1 with them.  I\'m dunking (seven foot hoop), blocking shots (4 foot kids), and actually enjoying my time with them.  I let them cheat and win (because I\'m nice), but not without letting them know who\'s got skills.  And the next day as I\'m trying to put my new grill together with incoherent instructions.  They beg me to play some more.  I told them \"some other time.\"  But I was amazed at how much they really wanted me to play.  And I got to thinking.  Most of these kids probably don\'t have fathers.  The kid across the street lives with his mom and sisters.  The other kids are down here everyday.  There\'s nothing better at home.  I wonder how many of them are escaping from conflict.  Escaping to a world where their family isn\'t yelling at them. Where there\'s an adult who actually plays basketball with them.    I\'m not saying I want to be a father figure to these kids by any means.  In fact I don\'t want them to get too comfortable in my driveway.  But I know that there is some healing that needs to be done in the kids lives.  They need some kind of acceptance from me.  They will eventually find it somewhere, I just hope it\'s in the right place. And maybe in the meantime I can help.   My father died when I was four, so God used people to fill in the gaps ever since.  I went to \"Father/Son\" campouts and fishing trips with charitable men in our church. Then, my mother remarried when I was about 8.  My step-dad has become a beacon of wisdom in my life.  And in between there were times when I cried, but I knew that my times were in my heavenly father\'s hands.     Shortly after my dad died, I asked my mom \"so who is my dad now?\"   Fighting back tears she answered \"God has promised to be your father.\"  And he was.   So maybe I am supposed to be ministering to these kids.  Maybe just a little sacrifice of comfort and time will pay off.  Who knows?  There\'s always the chance that they won\'t even remember me when the grow up.  But maybe they will.  Maybe I can leave a good taste in their mouths about God.    Maybe these kids are the modern fatherless.  Ps 68:5  \"A father to the fatherless, a defender of widows,   is God in his holy dwelling.\" Ps 82:3 \"Vindicate the weak and fatherless; Do justice to the afflicted and destitute. Ps 146:9 \"The Lord protects the strangers; He supports the fatherless and the widow...\"  '),(8483,3304415,'26,May,2004','2004-05-26 00:00:00','   check out www.danprice.net  '),(8484,3304415,'25,May,2004','2004-05-25 00:00:00','   By \"Post-Intellectualism\" I don\'t mean \"Anti-Intellectualism\"  I mean \"Beyond\" or \"intellectualism as a beginning\" or something like that.   anyway...  I feel that my faith in many ways is getting simpler.  I\'m 26 now and am in a new place in life.  I have been married three years, have a son, have a house, work in a church.  I am getting older. I\'m no longer in the perpetual stage of youth that I thought I once was. College for me was my time to figure out some sort of shape that my faith would take, my own template maybe.  I was forced to consider my Gnostic/American view of God and Christianity and deal with the fact that my views were changing and becoming my own despite what my parents, pastors, and professors thought.   I surveyed the Bible and connected the historical dots of what I thought (and for the most part still do think) what God\'s message to us is.    One aspect that I always believed in was community and the body of Christ.  I always believed that it was God\'s purpose for us to reach the world together as communities, that the world would know because of our love. And there have certainly been times where I\'ve experienced that. And I think I\'m learning more about this all the time. Dan and Andie Hannah were examples of that to me, we lived together for about 8 months.  Dan and I would spend time at the local community coffeeshop and talk to the regulars there, discuss books, life, beliefs etc....  Some months after we left Columbus, a couple leaders of the church were discussion whether our way of relating to people was really a witness to people.  We weren\'t direct enough they were saying...and just as they were leaving, the owner Patti (not a Christian) came over to ask about how Suz and I were doing up in Michigan and said \"those guys and Dan and Andie were the best pictures of Jesus that I\'ve ever seen...because they didn\'t stuff Him down my throat...\"    You see Patti is one of those loving, community-minded people who does the closest thing to church, without including God.  She has something energetic going on almost every night of the week at the coffee shop,  writing groups, amateur stand up comedy, and live music.  The culture there is made up of former loners who find a place that is safe.  She accepts everyone and just by being herself you want to hang out at Three Cups Coffee everyday (which we did).  But she was one of those who had tried going to church in her younger days, only to be bashed and bruised.  So now she will not be manipulated by religion.  She will not be sold on a bridge diagram.  Pattie will not be convicted watching TBS.  A cool, trendy flyer, given to her by a cool, trendy person will not speak to her.  But I think friendship does.  I think the test of time and commitment will.  I think our youth who volunteered to set up and tear down for charity events at her coffee shop did.  I even think the pastor\'s willingness to change and understand what was \"fake\" or \"unathentic\" about passing out those trendy invite cards, his humility spoke to Pattie.    I\'m experiencing community in new ways here all the time. -when Jen and Eric call us and say \"we have a free night\" we want to help you move/take care of the baby whatever\" -when Melissa asks me how much I\'m working this week so she can help Suzanne when I\'m gone -When Brent and Kate (and 13 other people!) show up to help us move even though we don\'t really know each other that well -When I walk out of a meeting to find my band cleaning the inside of my car because I\'m such a complete slob -When Travis is real and challenging with me....when we spur each other on/sharpen each other with out thoughts -When Erin makes a four hour trip to be with Suz as she is in labor and she can only stay for a day... -When the Sunday band changes the way they do things, just because I say that it\'s a good idea -When Noel pushes me to lead -When Suzanne pushes me to lead -When Chaz buys an outlet for our dryer without us asking -When I talk to Scott about once every 2 or 3 months, but we know we are brothers. -When our mothers take a week off of work to help us with the new kid  I can keep going...there are so many people. I\'ve wondered how in the world people have kids, buy a house...,just live, without being a part of such a community. And it makes me want to bring that uncommon love to others. There are Patties everywhere.  People who are trying, they just don\'t have God.  They might have an ugly picture of God because of the way Christians have treated them or each other. It\'s our job to change things.  '),(8485,3304415,'21,May,2004','2004-05-21 00:00:00','   Ok so  urlLink rumors   are that Andy Kaufman has been alive and just returned on his 20th anniversary of his death.  Does anyone know anything about this?  '),(8486,3304415,'20,May,2004','2004-05-20 00:00:00','   More pics of the new dude on the picture page...  '),(8487,3304415,'20,May,2004','2004-05-20 00:00:00','   Ok, I\'m reading a really good book called Blue Like Jazz by Donald Miller.  It\'s a creative non-fiction book that tells of his jouney of doubt and faith and it is beautiful.  It reminds me of   urlLink  Anne Lamott\'s    book \"Traveling Mercies\" which falls into my cannon of life changing manuscripts.  The writing is minimalist. Sentences are structured almost like J.D. Salinger\'s great American novel.  But the book is more honest and thoughtful than any other book on faith that I have read.  Check his page out at   urlLink www. bluelikejazz.com.   Oh, and he likes Wilco...  '),(8488,3304415,'16,May,2004','2004-05-16 00:00:00','   Ok, I guess it\'s growing up time. In a way at least. I have to say that I can\'t stand the thought of people viewing me as old.  I know couples that start having kids, buying a house etc...and suddenly they are old.  They don\'t hang out at pubs with their friends.  They don\'t go to concerts.  They don\'t play poker until 3 in the morning.  The are just old.  So I think I\'m just going to buy this house, take care of my kid and wife, and hang on to \"life\" or \"youth\" or whatever it is that I\'m trying to hang on to.  But I really do think it\'s important.  People don\'t want to be around \"old people.\" And by that I don\'t mean old in an age-specific way. I know people in their 50\'s and 60\'s who don\'t seem too old.  I have to admit though, that after hearing the kid cry last night, I could use a nap...  '),(8489,3304415,'14,May,2004','2004-05-14 00:00:00','   So we closed on our house today. We seriously killed about 80 trees with all the papers we had to sign.  We can move in at any time now though.  So if you live in Michigan and want to help me, let me know!  '),(8490,3304415,'11,May,2004','2004-05-11 00:00:00','   I\'m serious, if you listen enough, you\'ll like them.  '),(8491,3304415,'11,May,2004','2004-05-11 00:00:00','   urlLink Ok, so here\'s a link to a pic of Jaden and Me as a baby.  Don\'t worry, he looks a lot better than he does in the pic.  I\'ll let you know when I add more.  '),(8492,3304415,'11,May,2004','2004-05-11 00:00:00','   When I get some time, I will update my friends list to include all the people I love who are blogging away.  But for now, you will have to be satisfied with the few that I have listed.  And again, I will put some pics up as soon as I get my camera back from the Kranzos.  Later guys!  '),(8493,3304415,'11,May,2004','2004-05-11 00:00:00','   just making sure that this is working here!  '),(8494,3304415,'29,June,2004','2004-06-29 00:00:00','   I\'m speaking this weekend at Riverview on God\'s Will.  And I\'ve been thinking a lot about how people want to know God\'s will, they want a sign, or an impression that tells this to do this or that etc...  It seems that we are looking for supernatural guidance, when so much supernatural guidance has already been given (see \"the Good Book\"). We want to touch heaven, we want a miricle right now before our eyes.  But that got me to thinking that maybe our world is all backwards.  I mean, if you think about it, the miracle has happened.  That God created the world and that things work (realizing the FALL).  Maybe this is the magic. Maybe when God does do something supernatural, that is going back to the normative state, pre-creation.  The miracle stops? for a second while God intervenes.  I\'m not sure where my thoughts are on this exactly right now.  I\'m still thinking about it a lot.   Here are some thoughts from the book Orthadoxy, by G.K. Chesterton.   â€œA child kicks his legs rhythmically through excess, not absence of life.  Because children have abounding vitality, because they are in spirit fierce and free, therefore they want things repeated and unchanged.  They always say â€œDo it againâ€; and the grown-up person does it again until he is nearly dead. For grown-up people are not strong enough to exult in monotony. But perhaps God is strong enough to exult in monotony.  It is possible that God says every morning, â€œDo it againâ€ to the sun; and every evening, â€œDo it againâ€ to the moon.  It may not be automatic necessity that makes all daisies alike; it may be that God makes every daisy separately, but has never got tired of making them.  IT may be that He has the eternal appetite of infancy; for we have sinned and grown old, and our Father is younger than we.  The repetition in nature may not be a mere recurrence; it may be a theatrical encore.  Heaven may encore the bird who laid an egg...â€  '),(8495,3304415,'28,June,2004','2004-06-28 00:00:00','   Well I just want to post real quick before I work on this week\'s message.  Last night we got some visitors from New Song in Columbus, the Harger girls.  It was really good to have them around.  They are an awesome family.  Our small group came over also, we are reading Blue Like Jazz.  We got into an interesting discussion about Sin Nature vs. Adam\'s sin being a snowball effect.  I love it that we don\'t always agree on things. It makes for really good conversations and leads to a reexamination of belief.    Jaden slept six hours straight last night, which was extraordinary.  I fed him at nine and then talked to him and he smiled and laughed at me for about half an hour.  It\'s pretty cool.  '),(8496,3304415,'25,June,2004','2004-06-25 00:00:00','   I got the ichin\' to create some music last night. But I didn\'t have any finished songs, so I just wrote a song while recording.  I laid down the acoustic track, and then I spat out some vocals/lyrics (i found God in the nick of time, he was about to knock me down...) and then some experimental electric guitars and finally some white noise.    I find that the new stuff that I\'m writing might be less accessible to a broad audience, but there wasn\'t much of an audience to begin with and I really like the experimentation.  It takes 24 hours to be \"approved\" on Soundclick.com.  But tomorrow you should check it out and tell me what you think.  Just click on \"My Music\" to the left. Remember, I wrote it while recording. One take, so don\'t expect too much.  '),(8497,3304415,'22,June,2004','2004-06-22 00:00:00','   So this is my attempt to appease the insurance man.  Three steps is a long way down.  '),(8498,3304415,'22,June,2004','2004-06-22 00:00:00','   Good News is...I just bought  urlLink the most eagarly awaited, weird album of the year    The bad news is that (read comments in previous posts) there will be no Lollapalooza which really sucks.    Oh and more bad news...I\'m sick with a cold.    and one more good thing...I just got my new  urlLink Wilco shirt .  '),(8499,3304415,'21,June,2004','2004-06-21 00:00:00','   So I\'m listening to a mix of Van Morrison stuff and his song  urlLink Madame Joy  is one of the most soulful songs I\'ve ever heard.  It\'s weird because most people have never heard of it, it wasn\'t even released on a studio album, just an album of songs that never made it. The album is called The Philosopher\'s Stone. And it\'s my second most favorite Van album (I  have 15 of his) next to Astral Weeks.  '),(8500,3304415,'20,June,2004','2004-06-20 00:00:00','   Suz and I just bought tix along with Cisco and his girl Jen.  Here\'s the line up. String Cheese Incident Flaming Lips Wilco Gomez Polyphonic Spree TV for Radio  here\'s a  urlLink link   if you want to come.  '),(8501,3304415,'18,June,2004','2004-06-18 00:00:00','   That\'s what our night was like last night.    Jaden was still sick and couldn\'t keep his food down on top of a cough and runnning a fever.  So we called up the doc and he told us to take him to the ER.    The Emergency Room is a sick place.  In the room that they put us in there were blood stains all over one of the seats and it looked like a room that I attempted to clean.  We got there at 10pm and they finally sent us home at 5am.  Crazy night.  There were stabbings, attacks, police inquiries, gross coughing and weezing, drunk talk, among other things.  All the staff were very nice to us though.  I think they were relieved to have a patient who wasn\'t a victim of armed violence.  He\'s a baby so he gets some points that way also.  Anyway, his fever went down and he started drinking and feeling better.  But there is something in his blood called CRP which I think stands for CRAP (thanks Noel) and we had to bring him in to the doctor\'s this morning to have that checked out and we will have to take him in next week just so they can keep an eye on him.    We are tired. I\'m out.  '),(8502,3304415,'17,June,2004','2004-06-17 00:00:00','   Do you ever wish that you could live ten lives, all in different places?  I\'m not up for moving every year (even though that\'s been the case the last couple of years) but if I had unlimited time and circumstances, there are a few places that I wouldn\'t mind trying out.   In one life I would be a local shopkeeper in some ancient European town.  It would be a family business that I inherited from my family and one that I would pass on to my children when I died. Business would never be great.  I wouldn\'t even think about franchising.  But it would be enough to get by on.  It would be a town near the sea where I would meet my friends every morning at some little known restaurant. We\'d sit on barstools and talk about the comercialization of the fishing industry over bitter coffee and toast.  I would take walks with my family out to the country. The hills would glisten and greet us in the late afternoon as our kids would run on ahead and their rolling shadows would whisper goodnight on the way home.  Our lives would be simple.  There would be less plastic and more stone and cherry wood, more earth and less electricity.  We would have a garden that we would eat from.  We\'d have warm bread and beef stew often, and trips to the candy store would be less common and more special.  We wouldn\'t travel much, just content to find all there is to know about our immediate surroundings.  '),(8503,3304415,'16,June,2004','2004-06-16 00:00:00','   101.3=Jaden\'s body temp. right now.  Yeah, he was up all night screaming and coughing. No sleep for us.  But we did take him to the doc today and got him some medicine.    Before we had a kid everyone told us \"your life is going to change\" like it was some huge mystery or something.  It\'s not, I\'ll tell you right now how your life changes.  You don\'t sleep very much and you have a little kid that you love so much for no reason at all and you spend all of your time pushing buttons trying to figure out why he\'s crying.  Pretty simple.  '),(8504,3304415,'11,June,2004','2004-06-11 00:00:00','   I posted on my Xanga site about why we named our kid Jaden Catcher, but I thought I\'d reexplain it since I\'ve been getting questions about it.     First of all, yes I played baseball. Second, no we didn\'t name him \"Catcher\" because he has strong knees and a knack for calling signals.  Cather is from J.D. Salinger\'s book \"Catcher in the Rye.\" You probably had to read it in high School.  Holden Caufield, the protaganist of the book, explains to his sister what he wants to be when he grows up.         \"You know that song \'if a body catch a body coming through the rye\'.? I\'d like-\"        \"It\'s \'if a body meet a body coming throught they rye\'!\" old Phoebe said. \"It\'s a poem by Robert Burns.\"       \"I know it\'s a poem by Robert Burns.\"       She was right though. It is \'if a body meet a body coming through the rye.\" I didn\'t know it then, though.            \"I thought it was \'if a body catch a body,\'\" I said. \"Anyway, I keep pictureing all these little kids playing some game in this big field of        rye and all. Thousands of little kids, and nobody\'s around--nobody big, i mean--except me.  And I\"m standing on the edge of some crazy cliff. What I have to do, I have to catch everybody if they start to go over the cliff--I mean if they\'re running and they don\'t look where they\'re going I have to come out from somewheere and catch them.  That\'s all I\'d do all day. I\'d just be the catcher in the rye and all. I know it\'s crazy, but that\'s the only thing I\'d like to be.\"  That passage just get\'s me going because I feel like that is kinda our job as Christ followers.  I know so many people who sacrifice for others and I want my life to be like that and i know that Suz and I want our son to be like that.  We are to be concerned about what really matters...People!  I think Holden is right so many times in the book where he calls out the \"phonieness\" of striving for happiness in material things.    So that\'s why.  '),(8505,3304415,'09,June,2004','2004-06-09 00:00:00','   Thanks to Travis for sending this cartoon to me.  For anyone who has taken the time to listen to my rant about cows milk, this is almost verbatim.  '),(8506,3304415,'09,June,2004','2004-06-09 00:00:00','     '),(8507,3304415,'08,June,2004','2004-06-08 00:00:00','   Well I slept well last night and still no symptoms, so I\'m kinda worried.   Noel made this for me:  '),(8508,3304415,'08,June,2004','2004-06-08 00:00:00','   Ok, now don\'t freak out here, but I\'m reasonably confident that I have contracted West Nile Virus.     So I went Mountain Biking with  urlLink Noel  this afternoon and we found out that it was us against God\'s green earth.  Man vs. Nature.  Battle of the species.  Specifically I almost ran over a snake.  Noel almost ran over a rabbit and then a baby deer, fawn I think you\'d call it.  Then we stopped and just stared at the deer with was about ten feet away.  Fun.    But, I think the world world was mad at us so it sent a flury of Mesquitos (from the Latin  Musca  meaning \"little fly\"-in case you don\'t know, they like to suck human blood and infect us with various diseases)  to attack us and we lost the battle.  I\'ve already had Chicken Pox, so I know that these ichy, red bumps are just poison injected by our new enemy the \"little fly.\"    How do I know that I\'m infected?  I guess I don\'t for sure.  But I do have some of the symtoms that I found on a  urlLink credible website .     The symptoms read like the following :  \"Many people infected with West Nile virus have no symptoms and do not get sick or have only mild symptoms. When infection does cause illness, symptoms will usually appear within two to 15 days. The extent and severity of symptoms vary widely from person to person.\"  I love you all if these end up being my last days.    Use Deet  '),(8509,3304415,'07,June,2004','2004-06-07 00:00:00','   urlLink Here  is a link to a sermon about Christ in relationship by Don Miller the author of Blue Like Jazz.  It\'s pretty long.  I\'m about halfway through and the beginning of it is a bit slow, but if you appreciate Shakespeare, then you might like his main metaphor here.  '),(8510,3304415,'05,June,2004','2004-06-05 00:00:00','   If anyone wants to check out some of my songs go  urlLink HERE  .  I have three songs on there right now and I\'ll have two more up by tomorrow.  chicka chicka boom boom.  '),(8511,3304415,'03,June,2004','2004-06-03 00:00:00','   Our church is too big for our building.  Our building that looks like a resort with our faux stone front with our comfy couches and neutral colored walls.  We need some art in here for one thing.    Today we have a 4 hour meeting about our building.  Everyone is cramped. There are only two offices in this building etc...  But for some reason I find myself not caring too much about it today.  I\'m fine with whatever they decide.  I mean, I would care if they turned it into the Lansing version of WillowCreek or some baptist church.   In fact I\'ve spent two years in college on a ministry team.  In that span of time I spent time in about 200 baptist churches.   I grew up in a church that met in a school for 15 years.  I went to a mega-church in Akron for  a couple years.  And I\'ve always found people there that have been \"church\" to me.  The building has had so little to do with my faith that  it\'s really hard to want to spend time thinking about this.    Of course, I know I\'m just tired and cranky and selfish.  Because to be kind to people we should be able to fit them into a building where they can worship together.  '),(8512,3304415,'03,June,2004','2004-06-03 00:00:00','   So the weekend of July 4th I\'m teaching on God\'s Will for you life.  It\'s a topic that I\'ve talked about once before, but it was just one teaching in a series on God\'s will (at Newsong).  Dan Hanna has been inspirational to me in the pursuit of finding solid ground on this issue as has a book called \"Decision Making and the Will of God\" by Dr. Garry Friesen.    But the greatest wake up call was when three people claimed to know God\'s will and in one week each made him look like a liar.  Here are the situations: 1.  This one really great guy found himself in a bit of trouble with the law over a minor mistermeaner charge.  The week before his court date he passed a church sign that read \"God Will Take Care Of It For You.\"  And he felt that this was inner leading of the Holy Spirit saying \"you won\'t get punished with jail time for this.\" He then refused the help of an attorney and consequently spent ten days in the county jail. 2. A wife of a man in our church thought that God was speaking to her.  Her husband was out of work and really wanted a job at a certain business.  His wife told him not to apply to any other places because she knew that it was God\'s will that he work there.  He didn\'t get the job and they were left discouraged and feeling abandoned. 3.  A single mom, who was a new believer, decided that it was God\'s will that she start dating this guy who was not a Christ follower in order to \"save him.\"  We knew this was a bad decision and told her so.  In the end he just ended up being a jerk.  He isn\'t saved and she isn\'t dating him now.  I just think it\'s dangerous to think of God\'s will in this way, as a target that we are supposed to hit.  I\'m more inclined as I read through scripture that God has already given us his moral will (the Bible) and that is what we are supposed to follow.  Beyond that we have freedom to choose.  I think though it\'s easy to wave responsibility off, like a mad hornet, and want God to lead you in every \"important\" non-moral decision that we have to make.    The argument is that the Holy Spirit\'s whole job is to LEAD US INTO TRUTH.  Which I agree with.  It\'s just HOW the Holy Spirit leads.  It seems that each time scripture talks about \"being led by the spirit\" or \'walking in the Sprit\" etc... that it relates in some way to God\'s moral will, the bible, the law whatever... But doesn\'t seem to be pointing toward making big decisions about which wife you should marry.  Of course there are a couple of exceptions to this like when God TOLD Hosea to marry Gomer. But it seems that when God speaks in this way he is really clear that it is Him that is speaking. You aren\'t blindfolded throwing darts at a moving tartget which happens to be God\'s will.  Sometimes we mix up IMPRESSIONS with GOD SPEAKING.  They are different.  And you will know for sure when God speaks to you (if he ever does in that way).   OK, thoughts anyone?  What do you think the leading of the Holy Spirit is?  Bring it on.  '),(8513,3304415,'02,June,2004','2004-06-02 00:00:00','   ok, so you should really check out  urlLink  Noel\'s teaching on drinking .  It\'s as good as a Newcastle.  '),(8514,3304415,'31,July,2004','2004-07-31 00:00:00','   .flickr-photo {	border: solid 2px #000000;}.flickr-yourcomment {}.flickr-frame {	text-align: left;	padding: 3px;}.flickr-caption {	font: 75%;/*	color: #666666; */	margin-top: 0px;}.flickr-buddyicon {	margin-right:5px; 	vertical-align:middle;	border: solid 1px;}.flickr-postedby {	font: 75%;}  	 urlLink    	  urlLink JadenBasket , originally uploaded by  urlLink groovingdan .'),(8515,3304415,'30,July,2004','2004-07-30 00:00:00','   .flickr-photo {	border: solid 2px #000000;}.flickr-yourcomment {}.flickr-frame {	text-align: left;	padding: 3px;}.flickr-caption {	font: 75%;/*	color: #666666; */	margin-top: 0px;}.flickr-buddyicon {	margin-right:5px; 	vertical-align:middle;	border: solid 1px;}.flickr-postedby {	font: 75%;}  	 urlLink    	  urlLink Dan and Jaden Pillow , originally uploaded by  urlLink groovingdan .'),(8516,3304415,'30,July,2004','2004-07-30 00:00:00','   .flickr-photo {	border: solid 2px #000000;}.flickr-yourcomment {}.flickr-frame {	text-align: left;	padding: 3px;}.flickr-caption {	font: 75%;/*	color: #666666; */	margin-top: 0px;}.flickr-buddyicon {	margin-right:5px; 	vertical-align:middle;	border: solid 1px;}.flickr-postedby {	font: 75%;}  	 urlLink    	  urlLink DanJadenScott , originally uploaded by  urlLink groovingdan .'),(8517,3304415,'30,July,2004','2004-07-30 00:00:00','   so i lost my pictures that i had on yesterday.  appearently Flickr sucks.  '),(8518,3304415,'30,July,2004','2004-07-30 00:00:00','   This is Dan and Andi Hanna.  The are leaving for Austria next week.  They still need some support raised and the didn\'t ask me to this or anything, but if anyone wants to support them, go to their  urlLink blog .  They are doing some awesome stuff.  Aren\'t they, like, the hottest missionaries you\'ve ever seen?  '),(8519,3304415,'29,July,2004','2004-07-29 00:00:00','   I feel like the last few entries have just been thrown up there.  I hate careless blogging, but at the same time I want to keep my site current.  Oh, well.  I\'ll get some energy here soon and actually put some effort into my writing. In the meanwhile...  The last night in Columbus we had one of the best talks ever with Shane and Erin.  We were up \'till three.  It was just amazing. I won\'t try to relive it here, but I hope I change because of that conversation.  Which brings me to the purpose of blogging, the purpose of writing about our lives.  I have been thinking about this lately and have come to the conclusion that it gives meaning to our lives.  Words do that.  Something happens in our lives and we don\'t really think of it as a story, but then we write it down and becomes more meaningful.  We go back to those stories over and over again. And reliving it is sometimes better than the original experience.  Any thoughts?  '),(8520,3304415,'26,July,2004','2004-07-26 00:00:00','   So, I finally found a wireless internet spot here in Columbus.  We are taking our time coming back, visiting our good friends.  Here are a few quick thoughts on the past week and a half.  1.  Noel and Tom Short both spoke at LT.  I thought this would be weird, because they communicate in completly different ways. But it was really good.  They both seemed to tone down the edge of their extremes a bit for the sake of the students.  At least that was my impression.  It was a very honoring process.    2.  On the way down saw Scott and tammy they are leaving for seminary at Regent in Vancouver in a few weeks... We stopped in Virginia and visited Dan and Andrea Hanna (go to hannasojourn.com) they are leaving for Vienna, Austria to do missionary work there, next week.  And we are now visiting with Shane and Erin and Andy and Jamie.  A + J just moved to Columbus to Virginia this past week.  And Shane and Erin just moved from Reynoldsburg to Clintonville.  So there is a lot of moving going on. I want to really spend time time and write about each of these people, i might tag-team with suz, but they each mean so much to us, that I really want to get it down in writing.    well, we are heading back wednesday. that\'s all I have for now.  '),(8521,3304415,'17,July,2004','2004-07-17 00:00:00','   Well, we made it safely to the beach on thursday after a great time with our friends Scott and Tami and the Silvers.  I am just sitting in the Green Room backstage.  Kyle and Matt are taking naps.  Christine and Sean aren\'t here yet, sean might not be here until five, so it will be a close call. But I am ready and excited.  The only problem i have is an infected toe.  It\'s not green or anything. Just red and puffy. Gross.  '),(8522,3304415,'12,July,2004','2004-07-12 00:00:00','   here\'s my to do list for today: 1. Deposit Check 2. Mow Lawn 3. Oil change 4. Clean car 5. Organize LT Music 6. Practice with the LT band 7. Practice the Electric Guitar 8. Call Bobby and Laura for directions 9. Call Scott and tell him when we will be there 10. Find my camera cord 11. buy a new stoller? 12. Buy guitar strings  if you read through this, then you are a loser.  just like me.  '),(8523,3304415,'11,July,2004','2004-07-11 00:00:00','   Friday I played darts with Travis and Joey at Moriarty\'s cool little dive pub in Lansing.  Santa Clause was there throwing strikes, keeping the tightest groupings.    Saturday we moved Eric and Jen into the commune known as The Kranzo\'s.   Played my two new songs saturday.  Fun.  Decent.  Shortstop is my calling.   Prayer with our small group was wonderful.  roadtrip to the southeast starts tuesday @ 6am.  roadtrip and southeast can look like road trip and south east.  out.  '),(8524,3304415,'08,July,2004','2004-07-08 00:00:00','   The second thing I\'ve been thinking about is evangelism and the law. I\'m actually reading portions of  urlLink Ray Comfort\'s  book (co-written with kurt cameron) about this topic.  It actually has some great points.  It talks about how the law is a tool to use to show people that we need God.  And for the most part I think that they are loving in their approach, though irrelevant in style.  The way of the master website, once entered askes the question \"are you a Christian?\" Yes or No.  If you hit no, they explain the law and God\'s love, trying to get you to make a decision right there.  The thing that I have a problem with is just using the law and the law alone.  This book is almost touted as \"a new way to preach the gospel\" as if Ray Comfort had written the book of Romans himself.  Paul has talked much about the law and it\'s purpose to show us our need for Christ.  I have to also applaud the honesty of motive for including Cameron in the book.  They say right out, \"we thought that this book is important and knew that if we had his name on it, we would have more potential readers.\"    The problem is when we are out of balence.  1st Peter talks in many places, exhorting us to \"live such good lives among the pagans that though they persecute you, they give thanks to God on the day of his return...\" and to live our lives in a way that people would have nothing bad to say about us.  Here\'s the rub.  We live in a culture that mocks Chrsitianity.  People have an innacurate view of God and of what Christianity should be, partly because of Christianity itself!  Our style is old and looks like propaganda.  In many places the \"war\" is with culture and not the devil.  We have placed differerent sins as \"worse\" all the meanwhile allienating those who Christ loved.  So our methods are important. We need to, in some instances change the way we talk to people.  Maybe it will take years of \"living such good lives\" before they will be ready to respond to God\'s law and love.  I know that just putting liturature out there, just preaching on the streets, just using media to communicate our message is not authentic to them and is probably turning people away.  We need to ask ourselves if we are getting in God\'s way by being blind to the love aspect and not just preaching the law.  '),(8525,3304415,'07,July,2004','2004-07-07 00:00:00','   Ok, so much to talk about and I can tell you right now, this is going to be snippets of my thought. But it\'s all about theology.  Two important responses (at least to my face) about my message this week.  1.  \"thanks, so much I\'m going through life changes and have wanted direction on God\'s will...this has helped me.... 2. (from an actual email)  \"Hi Dan,    We met after your sermon on God\'s will on 07/03/04.  We then discussed that one aspect of His will is to hear Him (John 10:3,27) that is to listen to His voice.  If you would like we can setup a time to meet sometime after July 9th.      If you and others would like to learn how to communicate with the Lord I would love to meet with you and whoever else is interested.  It would take a few days to teach people.  The best option would be to meet once a week on friday or thursday for an hour, preferably friday.    Most of what will be discussed will be; pertinent Scripture in learing to discern His voice, so you can actually listen to what Daddy God has to say about you, His loving and encouraging thoughts; revelation/interpretation of Scripture; and the application of Scripture to enable you to do what the Bible tells you to do.    A short analogy (that I gave you on saturday) follows.  How would you grow up and function in society if your parents could not communicate to you?  Take away snail mail, email, sign language, lip reading, and spoken language.  It would be very difficult to function in society.  Since we live in the spiritual realm, how can we do the Lord\'s will if you don\'t communicate with Him?  He hears us but it would be nice to hear His voice on a daily basis.  He already speaks to us all the time but it would be nice if we would listen to and recognize that it is Him speaking.\"  not quite on the same page here... when I talked to him after church he was talking about how God\'s gives him poems songs, words, etc... Basically \"channeling\" God.  '),(8526,3304415,'07,July,2004','2004-07-07 00:00:00','   Well, I am really excited about LT.  Our band should be pretty good, I think. I\'ve been writing some new songs and Sean (and the rest of the band) likes them, so that\'s good. you want your band to like your songs.  We are leaving on tuesday and visiting some of our best friends, Scott and Tammy, in North Carolina, then we will hit LT and on the way back we are going to see Dan and Andie in Virginia and then hopefully Shane and Erin, and Andy and Jami and the Newsong crew in Columbus.  Can\'t wait.  I\'ll be tan.  '),(8527,3304415,'04,July,2004','2004-07-04 00:00:00','   Well, I\'m all rested up from speaking this weekend.  It went as good as can be expected.  Saturday night\'s delivery was a bit rough, but things went pretty well today.  No one stood up and yelled \"heratic\" or anything like that.  So that\'s good.  Davin and Chey are letting us borrow an air conditioner!  No more muggy, hot, sleepless nights.  Very happy about that.  We installed it and then I took a three and a half hour nap. It felt like ten minutes.    My dad and brothers and I have been setting off roman candles, bottle rockets, screamers and so on with the neighbors.  And it\'s still going on.  Nothing like celebrating Independance Day by defing state law.  '),(8528,3304415,'02,July,2004','2004-07-02 00:00:00','   So not much to write about, just been practicing my message and getting the house ready for this weekend. My family is coming into town. I\'m so glad.  My dad and brothers haven\'t seen the kid since the day after he was born.  He\'s changed so much now.  He snuggles. He laughs. It\'s cool.  '),(8529,3304415,'06,August,2004','2004-08-06 00:00:00','     '),(8530,3304415,'04,August,2004','2004-08-04 00:00:00','   Here is a great passage from \"Decision Making and the Will of God by Garry Freisen.   I think it\'s possible that \"God\'s inward peace leading us\" just may be an unbiblical lie that many people (and I have believed it in the past) believe.  I haven\'t really seen anyone back it us scripturally though, besides taking a verse out of context.  I\'d love some response.  I know this is a touchy subject for people and don\'t want to claim to have the corner on truth, but I would really like for someone to show me why we can make decisions based on some inner feeling of peace.    Chapter 8 \"since peace is a fruit of the Spirit (Galatians 5:22), it is viewed by many as one of the means at His disposal to communicate His will to the believer.      This concept is  based primarily on the exhortation of Colossians 3:15 \"And let the peace of Christ rul in your hearts...\" The explanation is that this peace functions as an umpire that \"calls\" each decision in question.  So long as the believer is living within the center of God\'s will, he experiences and inner quietness of heart \"which surpasses all comprehension\" (Philippians 4:7).  Through this peace the Umpire is calling \"Safe!\" But if the Christian begins to move in the wrong direction, he experiences increasing restlessness and inner anxiety--an indication that he is about to step \"out of bounds.\"     It is not difficult to see the logic of this explanation, and it corresponds beautifully with the other aspects of the traditional view of guidance.  Unfortunately, it dos not correspond with the rest of the passage of which it is a part, and we suspect that the apostle Paul would be surprised to learn that his statement to the Colossians was being explained in such a way.  For he was not writing about God\'s provision for decision making, but rather a moral lifestyle that is consistent with the believer\'s position in Christ.     The statement \"let the peace of Christ rule in your hearts\" is part of a paragraph that runs from verses 12 through 17 in Colossians 3.  That paragraph begins with an exhortation for believers to \"put on\" certain virtues: compassion, kindness, humility, gentleness, patience, forbearance, forgiveness, and above all, love, \"which is the perfect bond of unity\"(3:14).  These virtues are to replace discarded vices such as immorality, greed, anger, abusive speech, lying, and the like (3:5-11)...Rather Christians are to be characterized by unity--a unity that will grow spontaneously as Christians put on the garments of love (3:12-14), submit to the peace of Christ (3:15), share the word of Christ (3:16), and do all things with thanksgiving in the name of the Lord Jesus (3:17).  '),(8531,3304415,'04,August,2004','2004-08-04 00:00:00','   Looks like we\'ll be hitting up the jazz fest in Lansing this weekend.  T.S. Monk (Thelonious\'s son) is going to be playing.  '),(8532,3304415,'03,August,2004','2004-08-03 00:00:00','   Ah..today is my day off and I\'m so glad.  Yesterday was a tiring day, we met with some potential architechs all day long... It was good i think they had some great ideas and seemed to be willing to work with our budget etc... so well see where it goes.    So Suz and I are hanging out tonight.  It is so crazy when you have a kid, it\'s harder to get time to hang out with each other, but we are determined to today!  Ah...and I found a store (Goodrich\'s) that has  urlLink Sweet Walter Red Wine   which we used to drink with Walt and Andy and Jami.  Goodrich\'s also has  urlLink Dogfish Head  which Shane and Andy introduced me to in Columbus.  Good stuff.  '),(8533,3304415,'03,August,2004','2004-08-03 00:00:00','   Well it\'s late and I can\'t seem to get to sleep because I drank some coffee earlier.  I \"had the talk\" with the Saturday bands tonight, that is I told them about the upcoming changes (although i leaked it out a month ago).  I\'m basically making the music thing a one year committment and having try-outs once a year.  This will be to keep things fresh and to bring more people into ministries/community at Riverview as well as giving us a wider talent base.  The teams responded so well. They (at least said) thought it was a good idea and seem to be behind us all the way.  It\'s tough when you get into music stuff.  It\'s such a personal thing.  And with praise teams you have to balence out the talent issue with the heart issue and style and preferences etc... especially with a large church.  I get emails from people almost every week, asking to be in the band.  So it will be interesting to see how everything goes. Pray for me!  '),(8534,3304415,'02,August,2004','2004-08-02 00:00:00','   Ok, so thanks to  urlLink Micah  for the picture.  It\'s Sean, Christine, Me, Kyle and Matt.  '),(8535,1111991,'30,July,2004','2004-07-30 00:00:00','   Good evening. It\'s day one. I can\'t say I ever thought I\'d do this again, but it\'s an opportunity and as good a time as any to pour out my feelings. Craig left this morning. We were up rather late on our last night together, after he tucked Mia in and explained that \"Daddy\'s going on a trip tomorrow\". The alarm was set for 3 AM and when it went off it seemed like 5 minutes after I drifted off.. fighting it all the way.   He showered, put the rest of his stuff together, and we kissed and hugged and I cried and he left. It was 3:49 AM. I cried for a little while after he left and then I was stuffed up, so I sat up in bed and did the Sunday Times crossword for a while until I felt tired enough to go back to sleep for a while. At 6:30 the alarm went off for me, and I showered and got ready. Mia woke up, as happy as ever, and I got her ready. I left Craig a voice mail; I knew he wasn\'t taking off until 10:20, so he was still only down the road at Miramar. He called me a few minutes later and I teared up again as I let Mia tell him hello and that she was watching Zooboomafoo or whatever it was.   I took her to daycare, she sang \"Good Morning to You\" and I know how blessed I am to have her company and her giggle. It\'s going to get me through this, just like the love of my family, my friends & my husband. It\'s just hard to think about this seemingly vast amount of time right now on the first night without him. I picked her up, and on the way home, as always, she asked, \"Go home and see Daddy now?\"   \"No, baby. Daddy\'s on a trip, he\'s at work, remember?\"  \"Oh, Daddy go on a trip tomorrow.\" Yes, Mia. Today IS tomorrow. I\'m still reeling at the fact that she\'s finally remembering the things we talk to her about... but I\'m also sure that my little OCD creature of habit will ask if we\'re going home to see Daddy every time I pick her up for a while now.   He flew to Seattle, called me from there at 1:30 PM PST. He was leaving for BWI in a half hour, and I told him that Mia talked about him, and that I picked up the car from the base at lunchtime. Poor thing has 29 hours of travel ahead of him, that is horrendous. He was going to BWI, then Frankfurt and then Kuwait. From there he\'s going to Iraq, to his home for the next seven months on a base outside Fallujah. He did promise me he wouldn\'t try to be a hero. I know that if for no other reason than his love for our daughter, he\'ll stay safe and come home to us. I know he\'s an incredible Marine and I respect and admire him so much.  So tonight I tucked my little Pookie in, and she helped me say a prayer for Daddy... and sang him Happy Birthday. Again. I think it might be Craig\'s birthday for the next few weeks, but that\'s ok. It makes me laugh. Craig called from BWI at 7:30 PST and told me that the next time I will hear from him is on Monday or Tuesday. At least this time I\'ll hear from him regularly via email & phone. That will be MUCH better than last year. *Sigh* last year... I can\'t believe that he had to go there again.   I\'ll be in here to post from time to time, and I promise it won\'t always be heartreaking. It\'s just the first day. I\'m sure we have lots of fun ahead of us in the weeks & months to come, it\'s just a hard day to get through. Each day is one less. I\'ll be back in here tomorrow night at some point. Good night.   '),(8536,1111991,'13,August,2004','2004-08-13 00:00:00','   He called this morning. A helicopter went down earlier this week, and their communications were cut off. Which does make sense... but my imagination runs away from me when i\'m worried.   Poor guy.. they\'re rationing water, which means only using bottled water, which means only using water to drink. Not to shower. In 110-120 degree weather. That is probably one stanky-ass place right now! Hopefully they get more water soon... those poor people on that base, it must be awful. Anyway, he\'s SUPER busy, which is great as it always makes the time go faster.   He needs new boots, so I bought them today & will mail them out. I have another box going with a bunch of magazines, a couple of articles, and a Costco container of Jelly Bellies (there\'s a tease for you, baby!... I know Craig reads this).  So in an email today and on the phone too, he said, \"No news is good news. If something bad happens, you will find out RIGHT away... so if you hear nothing, everything is fine.\" I guess that\'s as comforting as it gets for me right now, huh?  I\'m thrilled that it\'s Friday, especially since I have a nice busy weekend ahead! Tonight Les & Wendi are coming over for opening ceremonies. Tomorrow we have a fun birthday party to attend and then we\'ll go to Meesh & Bojan\'s to see their newly remodeled bathrooms (and them, of course). Sunday a ZOO trip with Anne & Ellie (my boss and her 3-year-old daughter, who Mia has already spent some time with...) and then hopefully a nap. then Dana\'s coming over! It\'ll fly by, but hopefully Mia will just remain happy & in a good mood (and a LISTENING mood) all weekend.   Ok, time to go home and make this yummy yakitori chicken recipe I found in the Weber cookbook... we\'ll see how that goes!   '),(8537,1111991,'10,August,2004','2004-08-10 00:00:00','   Well, I haven\'t heard anything from him since Wednesday morning his time, and it\'s now Friday night his time. At what point do I worry? I\'m up at 5:30 today because I keep checking to hear that nice little mail sound and see that it\'s from him. But instead I get a disappointing \"thunk\". Nothing.  Logic tells me that if his power was out or Internet was down he couldn\'t email me... but he\'s been CALLING. I know they cut off communications when Marines are dying because they don\'t want that kind of news coming from other people via their communications home. Understandable... but unfortunately Marines have been dying too much lately, and this huge protest in Najaf is only one on a laundry list of battles going on there that is SO frustratingly not covered in the media! ARGH. He\'s spoiled me, that\'s the problem... phone calls and/or emails every day, so when this gap happens I worry. It\'s not like last year when no news was GOOD news during the war.  Anyway, if you get the chance, read the Salon.com article \"War? What war?\". Email me if you can\'t get to it, I\'ll send you a copy... it details the American media\'s continuing lack of war coverage.  Ok, I need to shower. What I need really is some communication from my husband. Maybe today.   '),(8538,1111991,'10,August,2004','2004-08-10 00:00:00','   Just when I\'ve had a few bad days, Monday (of all days) comes and Mia is back in my heart. Of course I don\'t stop loving her to pieces, but some days are inherently better than others.  FIrst of all, I spoke to Craig this morning. Thankfully this time he paged me after I wasn\'t at my desk! Well, jeez, it\'s not like I can call him back! Anyway, he sounds good... ready for the turnover & to settle in to his permanent digs. Right now he has to wait for the guy he\'s replacing to leave before he can have his space. It\'s still hot... I felt horrible after I said it was hot here, realizing that was a ridiculous thing to say. (It is, for San Diego standards... but as in many cases, I\'m sure he\'d trade me) He told me that the \"Swiss Steak\" they had last night could have been camel..haha. But I got to speak with him for a while, tell him about Mia\'s weekend more, and ask what he thought of the pictures. He wisely stated, as I\'d also previously heard, that often the \"threes\" are worse than the \"twos\". Sometimes in the middle of fighting with your 2-year-old you don\'t look at the bigger picture; EVERYONE deals with this to some extent. Kids can be rotten sometimes, but man does the good far outweigh the bad. And so he told me that he could look at 100 Mia pictures each day and still look at more. Of course, I know that... I can\'t imagine being so far away from her for so long. Baby, she misses you too... I promise.  Today I went to Costco with the full intention of buying soy milk for Mia and a couple of birthday presents for kids whose parties we will be attending. Then I saw the PUPPETS! They had these adorable glove puppets, there were 4 in each set; one was ocean creatures, one was jungle, one was farm animals and one was zoo animals or something like that. $6.89 a set. Yes, you read correctly. So what\'s a good Jewish mother to do? I bought the ocean creatures and the jungle animals! I have no will power when it comes to getting stuff for Mia. I also got her the Eric Carle book  The Very Busy Spider ; same dude who did  The Very Hungry Caterpillar . Or, \"Caddapidda\" in Mia-speak.   So I was very excited to share the book, but figure I\'ll save the puppets for a desperate moment! So I picked up Mia, and thankfully she had a normal poop today (anyone reading this knows my woes I think, if you care at all and I haven\'t told you, let me know and I can fill you in), she greeted me with the usual enthusiasm, and Miss Mary said her day was great. I had her snack in the car, she climbed in her seat without hesitation, probably due to the fact that her alphabet book was waiting for her. She tells me what letter all the words start with, and the other way around.   \"What starts with \'C\', Mommy?\" \"Tell me, baby.\" \"\'Camel on the Ceiling!\'\"  I think it took her all of a week to memorize that book.. and it\'s the WHOLE alphabet!  She LOVES books. More than ever. I can\'t explain how much she loves her books, she brings them in the car, and when I do NOT have her favorite handful of them along, she gets so upset. So, I explained that the  Are You My Mother  book was at home, and I\'d read it as SOON as we got there. That seemed to satisfy her. Then we got home, and I made her some chicken while I cleaned for a bit (tomorrow I\'m getting the living/dining room steam cleaned and then going out at night, so it should be neat for Meesh who\'s babysitting), and then while she ate she was looking at books. Surprise. I folded laundry, put some more in, and realized it was 6:40 already. I hadn\'t showed her the new book! I told her, \"Mia, Mommy got you a new BOOK today!\" \"ooooooooh!!\" \"Want me to read it to you?\" \"YEAH! OK... read it to ME?\" She ran to where I was standing, and I showed her the book. \"What\'s this on the cover?\" \"A SPIDERman.\" \"Just a spider, Mia.\" \"A SPIDER!\" \"Ok, let me get my water, and we\'ll read the book.\" \"Mommy, I go poopoo, need change diaper.\" Hm. She already did that today. As I got closer, I realized she was not kidding... \"Ok, let\'s change your diaper, and THEN I\'ll read the book.\" She very hesitantly agreed, insisting on bringing two smaller board books with her, two favorites;  Are You My Mother  and  Mr. Brown Can Moo . She was standing by the changing table, and then she handed me the books. I put them on the side of the dresser, picked her up & changed her. Much protest involved due to the sand that came off her previously shod feet and went into her face while I hoisted her legs, poor baby! So I finished to exclamations of,  \"OK! Read BOOK! OK!\" I sat on the couch, grabbed the \"Spiderman\" book, and she came to sit - but no, she didn\'t. She stopped. \"OOPS! Be right back,\" she said. She RAN back to her room, grabbed the two books that I\'d left there, came running back with them, put them on the couch next to her, and then sat to listen. SHE WAS LINING UP FUTURE BOOKS TO READ! Not to mention the memory of what had just happened... amazing. Brain development right before my eyes. I laughed and hugged her. She said, \"Tiss?\" (kiss) and I did. \"Hugs?\" and i did. And I was SO smart, that when I got to the last book, I said,  \"Ok, after THIS one, it\'s bath time.\" Small whine. \"What happens after this book, Mia?\" \".... it\'s baf-time.\" RIGHT! awesome, it worked! And sure enough, as soon as it was over, she HAPPIILY marched into the bath! Bath was great, she\'s finally clean, and I stand her up on the table to put her PJ shirt on. Across the room she spots  Where the Wild Things Are . This is another new favorite, and it\'s the nice hard-cover version from my brother & his wife that she got for her 1st birthday. I\'ve only recently been courageous enough to let her touch it, she\'s very good though. \"MOMMY! Read THAT book??!?!\" I turned around, I\'d forgotten I\'d even brought it IN there, and she had to crane her neck to see it... wow. \"OK.\" So we walked out, I gave her milk, explaining that I\'d read it AFTER milk and tooth-brushing. I said, \"One book, just this one, and then bedtime.... what happens after I read this book, Mia?\" \"Bed time.\" YESSSSSS!!!!!! And what a fabulous bedtime it was. Amazing day. More tomorrow if I get the time. This novel was for YOU, baby!!   '),(8539,1111991,'09,August,2004','2004-08-09 00:00:00','   Sorry, gang (or anyone who\'s reading this) it HAS been a while!   Let\'s see... Craig is doing well. The Gunny he\'s replacing is leaving soon, and then he\'ll have the place to himself. This includes (and according to Craig, he\'s \"living like a Colonel\") his own twin bed, shower & nice big cement walls to decorate! So if you\'re going to snail mail anything to him, let me know so that you can send something appropriate.  I\'m not as sad for right now... but Mia\'s definitely starting to notice that he\'s gone. She was a disciplinary nightmare for most of the weekend, and usually the team effort we have going is a huge help. From the moment I fell for him, I always knew Craig & I were meant for each other. But I had no idea what a great parenting team we would be! He is an amazing backup for comforting, discipline, FUN, anything that Mia needs. He is missed in that department for sure.   She occasionally will hit me now, that\'s not accepted; and I feel AWFUL because I just take her wrist in my hand and sternly say, \"NO hitting! Hitting hurts mommy!\" and she begins to cry her eyes out. Not the reaction I\'m looking for, you know? I mean of course she\'s going to feel bad when I\'m angry, but I want her to feel more apologetic and LEARN from it, not break down in a sad fit of bawling. I guess it\'s the age. She\'s still only so-so with sharing too, but that will come. I just need her to LISTEN to me! I feel like a skipping record, \"Mia, don\'t touch. Mia, put that down. Mia, put it down. Mia, come here and do this, put that down. Do you HEAR me? Put it DOWN!\" If we\'re in public, people must think I have a screw loose... I try to make her repeat what I say, but she won\'t. I just don\'t understand how someone THAT smart, and with that accute of a sense of hearing (she hears a plane, helicopter, truck, crying baby, anything like Radar from M*A*S*H - before any other human ears hear it) can NOT listen. But yet if it\'s something she doesn\'t CARE to hear (mostly things out of MY mouth), she tunes out. Welcome to selective hearing, Mom. This will be in place for at least 16 more years.  Ok, after the bad, the good; I got her a PRECIOUS pair of pink, orange & white flip-flops. (available for viewing at  urlLink The Bowman Web Page ) they\'re Reef brand, and they were only $7. (I checked.. they\'re $14 at Nordies! HA!) So I bought them because I could NOT resist, and I wasn\'t so sure about the size. I left the tags on, and as we arrive at a restaurant to meet Meesh for a happy hour drink after work on Friday, I put one of them up to the bottom of Mia\'s shoe she was wearing. They were perfect. Mia spotted it as I held it for all of 3 seconds, yanks off her shoe & sock and says, \"I want to WEAR!\" pointing at the flip-flops!  \"You want to wear these now?\" \"OK!\" (If you have MET Mia, you know about her love for \"OK!\")  So, I helped her take off her other shoe & sock, and cut off the tags on the shoes. I helped her to put them on, she became and instant pro, and I said, \"Those are so pretty!\" She looked up, mouth in a wide-open smile, and said, \"LOOK at my TOES!\" I said, \"Those are \'flip-flops\'.\"  \"FLEEP Flops!\" She showed Michelle 40 times, she kept telling the waitress about them, \"Look at my FLEEP flops. Look at my TOES!\" it was classic. She woke up on Saturday morning and the moment I opened her door she said, \"I want to wear FLEEP flops!\" So obviously there was a huge breakdown when I wanted her to wear ANY other shoes, but she got over it. The scene is so darn CUTE!  We did have our Sunday morning at Starbucks. She wanted to take 4 books with her, and she\'s always playing with stuff and pretending it\'s a backpack, so I had an idea. I got out the backpack we bought for her for school and said, \"want to take them in this?\" \"OK!\" I helped her put them in, and she carried them to Starbucks, proud as a peacock. Now she needs to master how to unzip and zip it, but that\'s another time...   Ok, time to go. I hope you\'re enjoying the pictures & the stories, everyone. This weekend we have a birthday party & a zoo trip in store! Also a visit with Auntie Dana. Wendi tells me there\'s a community pool with a water park.. maybe we can find time to hit that soon too.   '),(8540,1111991,'04,August,2004','2004-08-04 00:00:00','   It\'s now Wednesday, it\'s been almost a week. Even though Mia & I have colds, I\'m in much better spirits. I know he\'s safe, we\'ve talked on the phone a couple of times, and I\'m making plans. I spoke with Ayelet the other night, we had a great conversation. Also spoke with my mom who I think worries that I\'m depressed; I open my email each day to find an e-card from her wishing me a great day. It\'s sweet... I\'m sure she has no idea what to do.   Just spoke with Lauren for a while tonight, which was great. It\'s nice to wrap my mind around what\'s going on in someone else\'s life, even if only for an hour or so. She\'s doing well, sounds happy... is one of many of my friends that is doing work on her house at the moment. Construction abounds out here! We made plans to see each other in a couple of weeks. This weekend hopefully I\'ll spend time with Les & Wendi, Meesh & Bojan, and I\'m sure I\'ll just be making sure that Mia & I are healthy again. The weather is amazing, it\'s cooled off quite a bit. I\'ve turned off the A/C and opened the windows.  Work is keeping me REALLY busy. It\'s going well, though; I heard today that we hooked Assemblywoman Christine Kehoe to speak at our Political Action Committee golf tournament fundraiser. That is a nice little coup.   I\'m also planning to see Bebel GIlberto (daughter of very famous Joao - see  Girl From Ipanema ) at the Belly Up toward the end of August. A few people from work are leaving after next week so there\'s a going away party. Hopefully someone can sit for Mia so that I can go.  Mia\'s doing so great... she seems like she understands somehow. We went to the store today and she sat in the cart (I should specify that it\'s a cart with a seat for kids, not just the regular cart) and behaved so nicely. She held a box of Spongebob Squarepants mac & cheese the whole time, telling me about the jellyfish & starfish on the box. Others would pass us in the aisles and smile at her and give me the \"She is ADORABLE\" look. I can\'t say I disagree. It seems like I\'m barely treading water in being able to avoid tantrums or curtail ones that have began, and she\'s getting smarter and smarter so those tactics are ceasing to be effective. Her memory is improving at an alarming rate, as is her speech ability. It\'s getting much tougher to fool or outsmart her. I guess it\'s a learning experience for both of us! The funny thing is, sometimes it works to my advantage, and i don\'t want to let on... For example, tonight, we got home from Vons and I had a bunch of bags in the trunk. As I struggled since I only wanted to make one trip, she stood looking and started shouting \"I want to carry! I want to carry!\"  \"You want to carry bags with me?\" \"yeah.\" GREAT! I handed her a couple of lighter bags, and she brought them inside so happily you\'d think I\'d let her have candy or something! Keeping that one in mind...hehe.  OK, it\'s sleep time. I\'m fighting this cold and need my rest. Good night.   '),(8541,1111991,'01,August,2004','2004-08-01 00:00:00','   Craig called me late this morning, he\'s there. 130 degrees in Kuwait... I thought with all that flying time, maybe they screwed up and took him to a base on Mars instead. He sounded fine, he was soaking with sweat, and definitely happy the plane part was over. He did say that the Gunny he\'s replacing has things pretty hooked up out there. I expect we\'ll all get pictures soon. If you care, the time difference is 11 hours (PST, anyway), so when he called me at 10:15 AM or so, it was 9:15 PM there. So you east coast people can do your own math..hehe.  Had a good day with Mia and of course I regret not having the camera with me. We played on a great playground near Les & Wendi\'s, and Mia had a ball. I\'m pretty sure Jake did too... he called Wendi \"mama\" for the first time. Well, actually, the second, but the first time you don\'t really believe it so with babies you wait until it happens again before you get excited. Mia was pretty good, all things considered. She REALLY had fun at the pool. She was previously very apprehensive about being in the pool. In my opinon, she saw Jake\'s lack of fear and said, \"screw this... if he can go in there, I can go in there,\" and proceeded to jump into my arms from the edge of the pool. Repeatedly. Hm! Guess now I can worry about her around bodies of water... swim lessons next summer for SURE!   I appreciate all of my friends here, your comments, your prayers, positive thoughts.. and WILL take up each and every one of you on your offers! Mia and I WILL travel a bit, I\'m buying her a Wiggles sleeping bag that\'s attached to an air mattress; I\'ll bet you all that she\'ll love that thing more than her bed! And I know that now & then even a movie or a dinner out with grown-ups will be very therapeutic. I\'m looking forward to it, every bit of it passes time. And I know that Mia is so much better with familiar and fun friends, that\'s such a huge help to me. (Julie, I mentioned your name in the car yesterday with Stuart, and she fully started talking about you! it was too cute.)   Fabulous conversation with Julie LIttle-Nickson tonight.. she apparently was put in charge of getting a gift for Michael Moore and gave it to him at a cocktail reception at the DNC in Boston last week and he CRIED. She had quite an incredible time with him. Sounds like she enjoyed the convention very much. She & her hubby are headed to the wine regions of France on Saturday, it sounds incredible.   Ok, it\'s 10:15, start of a new month, a new work week... and one more week closer to my guy returning. Good night, all.   '),(8542,1111991,'01,August,2004','2004-08-01 00:00:00','   I was woken up by my daughter stomping on the wall that connects to my room! Talk about a rude awakening. Ah the fun of the \"big girl bed\".  Lumpy... I\'m not a closet  OC  watcher, I\'m a FULL-ON out of the closet  OC  watcher! I love that show, and I fully know it\'s a guilty pleasure. Another few favs:  Six Feet Under  (actually, ANY HBO show, it\'s just that  Six Feet Under  is on right now),  Big Brother, Amazing Race , and coming up.. the Olympics. That will be a great time-filler! In the fall there will be lots of good new stuff too. I\'m a HUGE  CSI  fan, so the New York one is coming and that\'ll be great. I\'m sure there will be more diversions.  Mia\'s watching Pooh right now at 7:15 AM on a Sunday, and I\'m in here wiping the sleep out of my eyes. Yesterday I went to the Carlsbad outlet mall with \"Uncle Stu\" who Mia had a complete blast with. She really is much easier now that she\'s friendly with our friends and can stand to be away from me for a short while. However, I didn\'t bring her stroller so much of the day was spent repeating myself like some kind of idiot.... \"Mia, come here.  Mia, come here. Don\'t touch that. Don\'t touch that. Don\'t touch that! I said come HERE!\" She usually listens.. there was too much freedom yesterday I think. And Stuart, thank you for being so incredibly patient and amazing with her. For a change, I not only got clothes for Mia, I actually got some pants for myself! Then we did the Jamba Juice thing, and they made Mia her own mini smoothie cup and she sat at the table with us, lifting her glass and saying \"CHEERS!\" while we clinked cups, just like her father showed her last time we had Jamba Juice. That kid is developing one hell of a memory.  Just remembered... I\'ve also been watching Last Comic Standing for a while, and I had a dream last night that I had to do a routine on stage. I could probably do it, but in the dream I didn\'t have enough prep time so I chickened out. Strange.  Well, the \"boss\" calls.. time to make a waffle and get myself ready for a Starbucks trip & the Sunday Times puzzle. I may see Wendi & Jake later, we\'ll see.   '),(8543,3348160,'30,May,2004','2004-05-30 00:00:00','    I found these tests, unfortunately, the html codes are to long to post in the column so I\'ll post it here.  These are my test results.  Get yours at  urlLink  Fuali.com        I AM 32% INTERNET ADDICT!'),(8544,3348160,'29,May,2004','2004-05-29 00:00:00','    urlLink    Yo, this is Joel\'s Ebonics page you can reach it  urlLink  here  or on the left side of my page under \"fun sites\". It is packed with....da crazist muah-fuckin sheit yo\' a$$ can handle son, ah don\' make ma bitch slap ya a$$  &nbsp; urlLink   '),(8545,3348160,'28,May,2004','2004-05-28 00:00:00','    Jenny Jen Jen I found this and thought of you.  I\'m sure you already know it, but here ya go!    urlLink  80\'s Tees   '),(8546,3348160,'26,May,2004','2004-05-26 00:00:00','    Bravenet provides an execellent amount of upgrades to add to your website.  It is totally free, and every option you choose, you can edit, change colors, font, styles, add text....etc.  So CLICK ON THE BIG BRAVENET BANNER AT THE TOP OF THIS PAGE!! Thanks.   '),(8547,3348160,'26,May,2004','2004-05-26 00:00:00','    Ok so I added a few things today.    First, at the top of the page you will see the icon that says \"tell a friend\" well it does exactly what it says, so spread my shit around for collective ideas.    Second, I added a poll all the way down on the left side menu.  Right now there are 3 polls, everytime you visit the site the poll changes to a new one of my choice, that I made up....SO ANSWER THEM!!    Third, and the final for today, I password protected my Daily Log link.  If I know you, I will give you a user name and password to get into it.  You will see the login box one the left hand side menu.  Also there is a new link under \"funny sites\" or fun sites whatever the hell it is, check it out, it is funny as hell.  LEAVE SUGGESTIONS FOR POLLS IN THE COMMENTS SECTION OF THIS POST!!!!   '),(8548,3348160,'26,May,2004','2004-05-26 00:00:00','    urlLink    Yes these are urinals at a firehouse in Newportville, Bucks County.  These are the same size as a toilet!  I was extremely under the influence while taking these, but they are truely amazing!&nbsp; urlLink   '),(8549,3348160,'24,May,2004','2004-05-24 00:00:00','    urlLink    The drunk \'Pimp of the Year\'&nbsp; urlLink   '),(8550,3348160,'24,May,2004','2004-05-24 00:00:00','    urlLink    ahhhh drunken weddings&nbsp; urlLink   '),(8551,3348160,'24,May,2004','2004-05-24 00:00:00','    urlLink    More of me and my woman&nbsp; urlLink   '),(8552,3348160,'24,May,2004','2004-05-24 00:00:00','    urlLink    My beautiful woman and I at Citizens Bank Park&nbsp; urlLink   '),(8553,3348160,'24,May,2004','2004-05-24 00:00:00','    urlLink    my rock!&nbsp; urlLink   '),(8554,3348160,'24,May,2004','2004-05-24 00:00:00','    urlLink    me!&nbsp; urlLink   '),(8555,3348160,'24,May,2004','2004-05-24 00:00:00','    Ok, since the Flyers are done, I have to come up with new shit.  Well I can touch on the Phils but not to much.  There will be a link called bizzarre on here like tomorrow or the next.  It will be good shit! Make sure you check it out!   '),(8556,3348160,'23,May,2004','2004-05-23 00:00:00','    I must say even though the Flyers fell to the Lighting in game 7 it was definately one of the best regular seasons, and playoffs I have seen in years.  I am not going to review the game because it is just to painful for me.  Lets just say that there was no quit until the end.   This team reminded me of the Sixers team a few years back that went to the finals against the Lakers.  They to were a special team with all the starters and bench being battered, and bruised, injured, basically half dead.  Even though they lost to the Lakers in the end, they never quit, complained, or gave up hope.  Unfortunately the team will probably be a quarter dismantled in the off season due to salary issues, but hey, there may even not be a next season with the possible lock-out coming after the playoffs.  I enjoyed this season, this playoffs, the coach, the players, and especially Keith Primeau.  I wish that the entire team could stay together for another year.  This season should be one of the most remembered season for a time to come.  So I bid my hockey season farewell!  GO FLAMES!!!   '),(8557,3348160,'21,May,2004','2004-05-21 00:00:00','    Has anyone else ever watched  urlLink  Philadelphia Flyers  games on stations such as:   urlLink  ABC 6  or either of the  urlLink  ESPN  channels?  Well if you have then you will know exactly what I am about to talk about.  If you do not, then do not read this because you suck and are not worthy of reading, nor understanding this. :)  I used to enjoy watching hockey on ESPN with Barry Melrose, Steve Levy, Darren Pang, and Mike Emmerick, but during the entire regular season, and (especially) the playoffs, none of these \"hockey guru\'s\" have anything good to say about the way the flyers perform during games.  I am sick of listening to these assholes, and I\'ll use last night as an example, saying things like:  \"The Lighting are so good at this, and that, and the goaltender made/makes amazing saves, he is a hall-of-famer....and so on\".  This is not just this series, it was like this last serious, and the series before.  If any of these guys made comments about the flyers like they do about the opposing team, the world would come to an end. I may be biased in the way I think about the other teams, but I am certainly not deaf!    After I was done smashing my TV into a million pieces while thinking of Barry and the crew, I still cannot calm down about this.  Everyone knows how all of the  urlLink  NHL  has had it out for the Flyers ever since the \"Broad Street Bullies\" days.  This especially includes the referees and their lack of vision.   Next time you watch the Flyers, mark down how many good things they say about the flyers versus the other team.  You will be stunned!  Ok the \"why they should not\" list:  1) Do not know anything about the team, player, or organizations, yet are allowed to talk, and act like they do.  2) Can never talk about how good the flyers actually are.  3) Only point out what the flyers did wrong, even if they are winning!  .....there are more, I am getting flustered thinking about this!   '),(8558,3348160,'18,May,2004','2004-05-18 00:00:00','    Wow!  Thats about all you can say about Thrusday nights Game 6 against the Tampa Bay Lighting in Philadelphia.  Once again Philly\'s Captain  urlLink  Keith Primeau  has proven to be the  urlLink NHL\'s  best leader.  I do not know how I made it through the entire game without having a heart attack.  Within minutes Tampa Bay\'s Vincent Lecavlier stunned the hometown crown with a disgusting, I mean sick, slap shot right past  urlLink  Robert Esche .  It came off a bad turnover by Dennia Seidenburg, but it was only his second game back, and you stick him in one of the best series in NHL history, mistakes happen.    A few minutes later, one the best players in the NHL,  urlLink  Keith Primeau  , surged up center ice, taking two men with him,   urlLink  Primeau  kicked the puck to a wide open  urlLink  Simon Gagne  who tapped it out of mid air to tie the game.   urlLink  Gagne  who has not had the best of offensive numbers this year, showed us all how much patience, skill, and nerves of steel he has.  Once again Primes continued to be the reason why the Flyers got on the board.  .....Jumping ahead.....  Ok so Ex-Flyer Rusty Fedotenko smoked Esche up high twice.  Lecavlier fired a second monster slapper that Esche got a piece of, but not enough.   Sami Kappanen, one of the best two-way players in the NHL, took a slapper from Khabibulin\'s right side that stuck to the ice like glue and slipped in the bread basket of the ever twitching goaltender (anyone ever notice that?)  Next, el-capiton, made an almost impossible play to bury his second goal of the evening.  A shot from the point that Khabibulin made a right pad save on, drifted towards Primes who (with his right skate) kicked it through the crease, to himself to bury it home with a 1:49 seconds to go in the 3rd, to tie it at 4, and keep the Flyboys alive in their hunt for a game 7.  As Primes said in a post game interview \"If I had to score two goals to tie, I was going to do it.\"  .....Overtime.....  So the start of OT comes, and the Flyers owned it!  From the start that Flyers played hard-nosed, hit everything it site, let no puck touch a Tampa player, unload on Khabibulin from everywhere,  no-holds-bar hockey.  Well, it paid off!  Once again the \"new\" Primeau line (Primes, Gags, and JR) solved Tampa\'s defense, and the \"Boulin Wall\".  Nice work along the wall, passing, led to a wide open net for Gagne\'s second goal of the night, which for the second time in the game, came from a Primeau pass.  WE SURVIVE!  An exciting game, if you missed it, SHAME ON YOU!  If we play like we did in the 3rd, and OT period in game 7, there will be no possible way Tampa can beat us.  We also need to improve on those small little things that Tampa feeds of off (and makes you pay for them) called turnovers.  Esche did not exactly have his best game last night, but neither goalie has really had a good series.    There is no quit in this team.  They will not let the opportunity that is there for them to take pass by.  This may be the only chance for a few of the players to win, let alone go to the Stanley Cup Finals, and the one person who will not them forget that, or let it pass them by is their fearless leader.  Saturday night, Game 7, this is the year of the Flyer.   '),(8559,3348160,'18,May,2004','2004-05-18 00:00:00','    Before tonight the series between the  urlLink  San Jose Sharks  and the   urlLink  Calgary Flames  was tied 2-2, with neither team winning a home game yet!  On the half flip side you have the  urlLink  Philadelphia Flyers  and the  urlLink  Tampa Bay Lighting  who each split games at home ....Whats that you say? Home Ice advantage who?    To be continued............sorry to tired   '),(8560,3348160,'17,May,2004','2004-05-17 00:00:00','    urlLink'),(8561,3348160,'17,May,2004','2004-05-17 00:00:00','    After having an unbelieveable road trip, a losing streak had shown is ugly head to the Phillies.        I cannot speak to much for the first loss (since I only saw a small portion) to the Rockies besides the fact that Byrd and Rollins did not have the best of days at the plate.  Ledee and Utley both had three run homeruns.      Trying to continue his four game win streak, since getting off to a shakey 0-2 start, was Kevin Millwood.  Millwood was pretty much roughed up right from the start.     Was it not sweet to be in first place for a day?!?        Today\'s loss to the Rockies was a little more ugly than yesterdays.  After being up 6-0 by the end of the third inning, the Phils saw nothing but goose eggs from there on.      Randy Wolf had another strong outting as he went 6.1 innings, giving up 7 hits, but only 2 earned runs, with 8k\'s.      Cormier was lit up for 3 runs on 4 hits on 13 pitches! Thats right 13 pitches!  He did not make it through the inning before being replaced.      Hitting wise the Phils continued to hit well during the road trip, at least the 3 through 8 spot which combined for 12 hits.  Pat the Bat went 2-5 with 2 singles to continue hitting in the .300\'s.  Chase Utley who was having a horrible road trip until yesterdays 3-run-homerun day, went 2-4 with 3 RBI\'s.  Rollins & Byrd batted goose eggs today.      So we come to the bottom of the ninth.  With one man one Vinny Castilla walks up to the plate, and walks away with a 2-run walk off homerun to give the Phils their second straight loss.      This one hurt a little more then yesterdays, but look at the bright side......we are only 1.5 games out of first.   '),(8562,3348160,'16,May,2004','2004-05-16 00:00:00','    Ok, so I missed the Flyers, and Phillies game today.  However, I watched plenty of ESPN, and Comcast Sports Net to be in the know.  I even listen to the 129th Preakness Stakes on my way home from the airport!  And to be honest with you, it was actually quite exciting to listen to.  But lets get back to the Broad Street Bullies (Flyers link located on the bottom left of this site).  As we all know ABC hosted today\'s game.  To my disappointment of course, because they do not replay the game.  I only have one question.  Could Keith Primeau become, or already be one of the greatest captains in hockey history?  If the faithful have been watching all this year\'s playoff games we saw him pulverize most of the New Jersey Devil players until they did not even show up to play anymore.  Then Toronto, after a fairly slow start for himself, Hitchcock challenged Primeau to step it up in game 5, in a closed door meeting between himself and Prims.  The result, a man with all systems set on \"kill\" and to top it off he pulls a hat trick out from his sleeve.  Motivating the rest of the team to score on almost every shot they took.  The rest of the series was history!  We now come to Tampa Bay.  The first game, everyone was flat.  The slaughter that took place in game 2 looked like the game 5 against Toronto, with the Captain flying around like a crackhead on speed.  Game 3, very disappointing, only one person showed up to play that game, and that was Primeau.  For being a guy who missed a handful of games this year, suffered a concussion right before playoffs start, he flew around the ice, hit everything in site, and gave the flyers hope (26 seconds of it)of a comeback win in game 3 that was immediately squashed due to the shotty play of the entire rest of the team.  Primeau has bought into the systems that Hitchcock sells, as well as the rest of the team, and even Roenick is coming around to Hitch.  The team has bought into Primeau as their leader.  So he does not have the greatest offesive numbers, but that doesn\'t make you a leader cause you can score.  It\'s showing up to play everyday, leading by example, and being a veteran leader that even the veterans look up to.    Ah the Phillies.  Has this recent road trip been a preview of what everyone, prior to the preseason, was expecting.  Again, I missed today\'s 16-5 slaughter of the Rockies, but I have not missed any other games on this West Coast swing.  Rollins is beginning to hit, and hit well at that!  Pitching is beginning to look like a staff that could possibly defeat a team of the Yankees, Red Sox, Braves type calibour.  Abreu is coming around.  Burrell continues to look like the Burrell of two years ago.  Besides a few injuries that continue to plague some of the starters (Wagner & Thome), and a lot of recent errors, this team is starting to get me excited.  To get excellent reviews of the Phillies games check out  urlLink Phillies Blog .  Mike does excellent, in-depth reviews of games, stats, players!  Now the 129th Preakness Stakes.  I was able to listen to the race on 1060am on my way home yesterday afternoon.  All this week I wondered why people were making such a big deal about horse racing, and Philly own Smarty Jones.  I am the type that does not like to see animals used for money, but let me tell you it was quite exciting listen to the race.  And it was also a record setting race (11.5 horses ahead) the longest space between 1st and 2nd place in Preakness history.  So maybe I will watch the Belmont Stakes in June.......after all we may just have two parades coming down Broad Street in June ;)  Can Philly get anymore Euphoric?   '),(8563,3348160,'15,May,2004','2004-05-15 00:00:00','    Ok, so I decided to dabble (spelling anyone?) in HTML, and website creation, so I went with the blogspot first.  Actually the are two reason behind my maddness for creating this space.  First, that exactly what I wanted.  My own little \"space\" (hence The warehouse) on the internet.  Second, my love is going to Italy for two weeks this afternoon.  Its not the first time that she is going to be in a foreign country while being with me, and I will be talking to her, but am what you call a \"worry-body\", and how can I give a full days review of my life in a few minutes (see Daily Log link at the bottom of this page).  So yeah there are more reasons.  I would like my little space to become a place where people actually stop in here everyday to check out the daily topic of conversation, sports, news, music and comment on it.  Whoa I\'m tired! I\'m off in a half hour.  Almost time to go the airport.  I am deathly afraid of planes....this will be the longest 13 hours of my life.  Honey, I will miss you dearly!   '),(8564,3348160,'19,June,2004','2004-06-19 00:00:00','    Ok I posted a new forum topic, and since nobody has yet to reply to anything, it is a post of how extremely pissed off I am that people feel the need to cut peoples heads off in the Middle East.  Contribute damnit!   '),(8565,3348160,'14,June,2004','2004-06-14 00:00:00','    For those who truly know me, know that I am a history buff.  You can ask me any question, and I\'ll always have an answer.  I posted a few pics of a recent day trip I took to Philly for some well needed relaxation below.   '),(8566,3348160,'12,June,2004','2004-06-12 00:00:00','    urlLink    The Philadelphia Constitution Center.&nbsp; urlLink   '),(8567,3348160,'12,June,2004','2004-06-12 00:00:00','    urlLink    Ben Franklin Bridge heading into New Jersey.&nbsp; urlLink   '),(8568,3348160,'12,June,2004','2004-06-12 00:00:00','    urlLink    A view of City Hall from inside the Constitution Center.&nbsp; urlLink   '),(8569,3348160,'12,June,2004','2004-06-12 00:00:00','    urlLink    One of the first court rooms in the history of america!&nbsp; urlLink   '),(8570,3348160,'12,June,2004','2004-06-12 00:00:00','    urlLink    Benjamin & Deborah Franklin\'s resting site.  I tooks this on a recent day I spent down in Philly site seeing.&nbsp; urlLink   '),(8571,3348160,'12,June,2004','2004-06-12 00:00:00','    urlLink    How true is this?!?!&nbsp; urlLink   '),(8572,3348160,'12,June,2004','2004-06-12 00:00:00','    urlLink    if you dont know what that is, you should be shot.  Test pic since Hello seems to have some issues they need to work out&nbsp; urlLink   '),(8573,3348160,'12,June,2004','2004-06-12 00:00:00','    Due to a mass brain fart on my part my previous \"Warehouse\" fell into the abyss.  Check out all the new features on the right hand side of the page.    One of the best features is the \"Forum\" where anyone can post topics that can be seen and replied to by anyone who pleases, so please contribute.    The \"calendar\" shows my daily schedule, b-days, events, and anything else that someone wants me to add. So let me know what your b-day is, or anything else of importance that is going on so I can post it.  The new \"photo album\" that bravenet offers is a great way to post/view photos.  I can even allow users to post photos in the album, so email me if you would like access.(sszerlagjr@verizon.net)  A new version of an online poll is also new here.  If you have any poll questions you would like to see posted, again, please email me.  So enjoy the new blog, contribute, and pass it around to your friends.  Thanks.   '),(8574,3348160,'05,June,2004','2004-06-05 00:00:00','    Ok instead of doing my \"everyones two cents\" blog I decided to add a forum to this site.  Click on the \"Forum\" icon at the top of the screen and start posting.  There are no limits to what you can post and everyone and anyone can reply/add to the topic.  The only thing I ask is that you spread the link to this blog around as much as possible.   '),(8575,3348160,'05,June,2004','2004-06-05 00:00:00','    I added a new link under \"fun sites\".  It is an Encarta page that you can see what occured on \"This day in history\".   '),(8576,3348160,'05,June,2004','2004-06-05 00:00:00','    As you all know today is the 60th anivarsary of  urlLink  D-Day, the allied invasion at Normandy, France  which was the beginning of the liberation of France from the Nazi\'s, as well as the rest of Europe.  Thousands of american and allied soldiers gave their lives on this day, on what was the largest land/sea/air invasion in history.      If you are planning on visiting the new WWII memorial in Washington D.C. my mother was just down there and said it was beautiful.  You can get a good preview of the site  urlLink  here .    urlLink  The National D-Day Memorial Foundation .  WHAT ELSE HAPPENED ON  urlLink  THIS DAY IN HISTORY?   '),(8577,3348160,'01,June,2004','2004-06-01 00:00:00','    To much of the dismay of police departments across the United States another  urlLink  police officer from PA was killed  in the line of duty, in the early hours of saturday morning.  Officer Michael Wise, of the Reading Police Department, was shot in the head while sitting in an unmarked patrol vehicle, with two other officers watching a possible suspect from a murder earlier that evening.  Pennsylvania is currently #1 on list for  urlLink  officers killed in the line of duty  in 2004, and in the top ten for officers killed throughout history.  Support the  urlLink  National Law Enforcement Officers Memorial Fund Inc.  by donating.   '),(8578,3348160,'01,June,2004','2004-06-01 00:00:00','    Everyone should be thankful for the ultimate sacrifice that the men and women of the military gave, and continue to give, for all of our freedom.  So I will leave everyone with a quote that they should really think about whenever they say fighting war is wrong or unnecessary.    \"War is an ugly thing, but NOT the ugliest of things.  The decayed and degraded state of moral and patriotic feeling which thinks that nothing is worth war is MUCH worse.  The person who has nothing for which he is willing to fight, nothing which is more important then his own personal safety, is a miserable creature, and has NO chance of being free, unless MADE or KEPT so by the exertions of BETTER men then himself\" -John Stuart Hill  Don\'t forget why you are all still alive, in our free country.   '),(8579,3348160,'26,July,2004','2004-07-26 00:00:00','    Ahhhhh I finally got the radio up!&nbsp; There are, or should be, tons of tracks on the radio.&nbsp; Skip through the songs just like you would on a normal cd player.&nbsp; Enjoy!'),(8580,3348160,'24,July,2004','2004-07-24 00:00:00','    A new forum topic has been posted.&nbsp;  The topic is the  urlLink    9/11 report     .&nbsp; If you have read, or started to read it contribute.'),(8581,3348160,'22,July,2004','2004-07-22 00:00:00','    Ok, so I am sure that everyone has gotten a pop-up for the \"get 10,000 smilies for free.\"&nbsp; Well, last night as I was playing around on the interent I decided to check it out, and let me tell you it is a   MUST   download.&nbsp; The only down-fall to it, and it wasn\'t a down-fall to me, was that when you download the service it adds an extra toolbar to your browser.&nbsp; Actually, the toolbar that you get with the program makes it 10 times easier to use the program.&nbsp; The program not only comes with   10,000+ smilies  , but it also allows you to  add backgrounds to your email , add a  personalized signature with picture  (yours or theirs) with up to 4 lines of text.&nbsp; You can also add a custom stamp to your email that you can add text to also.&nbsp;&nbsp; There is  10,000+ screensavers ,  and wallpaper  for your desktop to choose from.&nbsp;  I ended up playing with the program for about 3 hours last night and could not even go through all of the great options.&nbsp;    This is a must download and I highly recommend it!     Oh, another down-fall that I failed to mention is that it only works with certain email programs (for the backgrounds and signatures).&nbsp; However, the company updates the email programs monthly so if yours is not on there now, it may very well be soon!   If you are interested  in this service I made a   link to it on the right hand of my blog  , about halfway down.&nbsp;     THIS IS A MUST DOWNLOAD!!!   '),(8582,3348160,'16,July,2004','2004-07-16 00:00:00','    I have added a  urlLink Google  toolbar at the top of the page. It allows you to search the world wide web, or The Warehouse only. Enjoy!   '),(8583,3348160,'15,July,2004','2004-07-15 00:00:00','    A new forum topic has been posted.   '),(8584,3348160,'14,July,2004','2004-07-14 00:00:00','    Former Eagles middle line backer, once feared by those who faced him, two time pro-bowler, and team leader  urlLink Jeremiah Trotter  has returned to the Eagles for, at least, one year pending a physical sometime this month.&nbsp;  urlLink Philadelphia Eagles  coach Andy Reid noted that He and Trotter had put behind the heated arguement that occured during a contract dispute two years ago.&nbsp;  &nbsp;  I for one am looking foward to seeing Trotter back on the team.&nbsp; However, I hope this does not cause any internal problems since  urlLink Mark Simoneau &nbsp;is the current starting middle line backer.&nbsp; Hopefully, the leadership that Trotter once brought to Philadelphia every Sunday has not disappeared.   '),(8585,3348160,'14,July,2004','2004-07-14 00:00:00','    A new forum topic has been posted.  Please read and contribute.  Thanks.   '),(8586,3348160,'12,July,2004','2004-07-12 00:00:00','    I added a \"On this day in History....\" fun link to the blog.  You can read the daily \"what happened\" on the day.  You can also see birthdays.  Here is what happened in history on my birthday, August 5th:  1305: Scottish hero Sir William Wallace, who led the Scottish resistance to an English invasion in 1298, is captured near Glasgow by the English. He will later be executed for treason.   1583: English explorer Sir Humphrey Gilbert founds the first English colony in North America, near Saint John\'s, Newfoundland. The colonists soon return to England, however.   1912: The Progressive Party, also known as the Bull Moose Party, chooses former president Theodore Roosevelt, who led the group\'s break from the Republican Party, as their presidential candidate.   1962: Movie star Marilyn Monroe is found dead of a barbiturate overdose at her home in Los Angeles, California.   1963: The United States, the former Union of Soviet Socialist Republics, and the United Kingdom sign a nuclear test ban treaty. Ninety-six other nations sign the pact, but France declines.  Here are some people I share a birthday with:  Mary Beard, historian and feminist (1876)  Naum Gabo, sculptor (1890)  John Huston, film director and actor (1906)  Neil Armstrong, astronaut (1930)  Wassily Leontief, economist (1906)  Patrick Ewing, basketball player (1962)  Enjoy!   '),(8587,3348160,'10,July,2004','2004-07-10 00:00:00','    Obviously you can tell it\'s been a while since I posted anything at all.  So here is a brief update.  I have been working two jobs, taking two classes, and interviewing for police departments.    I still post things in the forum from time to time so check in there for new stuff.  However, only one person has replied to them so I may dump the idea, even though you can post your own topics in my forum, only one person has done so thus far.  I will have some things up soon, and my other website will be up soon, which will have a lot more things since you cannot add to to much to the blogs.  As soon as things settle down I shall return.   '),(8588,3348160,'06,August,2004','2004-08-06 00:00:00','    Ah things have just been settling down for me over the past week or so. I am still working two jobs; one at the Hatboro Police Department, the second at Temple University. Recently, actually it was this past Wednesday, I found out that I was to be hired this upcoming Monday at Bristol Borough Police Deparment as a police officer. Obviously, I was quite excited when I found this out because I have been trying to get a police officer job for about a year now. If anyone does not know what it is like to get a cop job it is a very long and selective process with many interviews and test (polygraphs, physicals, etc.). Ok, so before I go on the Chief of the department told me himself over the telephone that I was to be hired, and sworn in on Monday.    Now here is the part where it feels like to be shit on  .  Last night the Detective for Bristol Borough Police Department calls me on the telephone and requests my presence at the station at 11AM this morning. I was thinking \"great, time to sign some paperwork, and go over everything before Monday) but  NO!  this was not the case at all.  Conversation:  Det:   Stephen close the door   . (A nasty tone)   Det:   Sit down  .  Me:  Ok .  Det:   We have a problem  .  Me:  What\'s that .  Det:   You lied on your original application  .  Me:  Pardon me .  Det:   You stated that you have never served in the military before, however, when you took the polygraph test, you claimed to have been discharged from the United States Navy on the premis of having asthma.    Me:  Thats correct. I was discharged from bootcamp after having two asthma attacks .  Det:   Why did you answer \"no\" to this question that asks if you have ever been discharged from the military for medical reasons.    Me:  Well, correct me if I am wrong sir, that question appears in a section where it states \"do you claim veterans preference points\", in turn I answered no to this questions due to the lack of active duty time.   Det:   Ok, well you are wrong. This question appears before the veterans preference points question.    Me:  May I see my application? (Det hands me the application) Oh, ok, I appeared to have misunderstood the question.   Det:   How could you possibly misunderstand the question!?!? It clearly states what it is asking for. You lied, and if you lie we eliminate you from the hiring process.     (Keep in mind I was told on Wednesday night that I was to be sworn in on Monday night, and I had taken a polygraph test over 3 weeks ago, which I  passed  with flying colors ).  Me:  Oh, sorry, just made an honest mistake. I thought the question came under the section under the veterans points category thats why I did not answer the questions/put \"no\" for various military questions.   Det:   Well we see this as you are lying. And if you are going to lie about one thing, this is usually a precurser to lying on police reports, and to other officers.      Me:    What!   I have written over a thousand police reports at my present job, and have never, and would never falsify a police report. I have disclosed everything in my past, and present on the polygraph test which, you told me yourself that I was truthful. One of the questions was \"Did I lie on my original application for employment\" which I answered \"no\" and the test showed I was not being decietful.   Det:   Well, we are are sorry Stephen, but the Chief and I have already discussed this and we decided not to pursue you anymore for an employee. However, you may reapply after six months.    Me:  Sir, excuse me for being blunt but why the hell would I lie about being discharged from the military for asthma? I disclosed information to you about my past drug use, and all of the crimes I have committed in the past. I have nothing to hide from you now, nor have I lied to you or anyone else during this entire process. I made an honest mistake. I have included my military record on over 10 other job applications. I disclosed it when I went to the academy. For what reason do you think I would hide this from   this   department?   Det:   I am sorry, our policy is that if you lie once during this whole process you are done. You are done. I am however going to have one more meeting with the chief before we make our final decision, but I do not think the outcome will be in your favor. Thank you for coming in, and good luck with the rest of your career.    Me:  Ok. Thank you .  This meeting ruined my day, week, months. This was very unfair.  I will post the final result.   This is what it feels like to be shit on .   '),(8589,3356970,'31,May,2004','2004-05-31 00:00:00','   Are you a Borg? Do you want to be a Borg? What would becoming a Borg do for you?  We are all to some extent becoming Borg. Some of us more so and faster then others but in they end we will assimilate you. Some of us depend on our technology to survive, some to keep from getting lost others to keep from getting bored. (Bored, not Borged!) one of my friends at work who would not consider himself Borg has 2 cell phones a beeper and a PDA. Is he Borg? Maybe not. I have my PPC, cell (Bluetooth connection both always on) my TENS and my GPS (not yet a Bluetooth model but soon, I will give the old one to a friend when I get the new one.) Not only that but I am developing my own HUD and will probably convert that to an camera/HUD as soon as I can. some people have technology but use it little, others of us depend on them. I go camping and take solar chargers for my computer in my drybag.  So what do you have that may or may not make you Borg? Are you?  '),(8590,3356970,'23,May,2004','2004-05-23 00:00:00','   I just watched Truely win the Monaco F1 and then left where I was watching it, jumped into a vehicle with 300 HP and this is not a good combination. Smoking tires and sliding corners I decided that running around town at 60+ might attract attention and I pulled it in a bit.  Of course being borg I knew the speed i was going and where i was due to my always there GPS unit. Borg never get lost, we use electronic bread crumbs. LLAP  '),(8591,3356970,'22,May,2004','2004-05-22 00:00:00','   My sister-in-law got a virus on her computer today. Well it\'s been there a week or so but I took the time to ID it today. I got to thinking. What if my newly connected computerized life support system (I use wearable medical devices that I have connected together and to my pocket PC to make controlling them easier and well the PPC is wireless to the net all the time) What if they got a computer virus, or a virus that infected my PPC made them go haywire. Not a pretty sight believe me. I have an antivirus software installed on all my connection points but not on the PPC; not enough mem to put one. One more thing for a Borg to think about. LLAP  '),(8592,3356970,'22,May,2004','2004-05-22 00:00:00','   I am sitting here typing into my computer when I could be just dictating it, except the voice control for computers is just not there yet. I would like to do this with my pocket PC but it\'s not even ready to take commands yet.There is the new Voice Command for Pocket PC but with only 128 meg of ram and no hard drive having enough storage space for the software and the tables and pattern analysis software is not going to fit. On the PC you can  do some with Microsofts new voice command SDK. But the recognition rate is lower then the rate on my 10 year old Mac. For now the part of a borg that listens is still the biological part.  '),(8593,3356970,'16,May,2004','2004-05-16 00:00:00','   Yes you watch Star Trek and you know what a Borg is. Are there any actually living wth us? Not the TV type but there are a ton of people who i consider to be real Borg.  The word Borg is a shortened form of the word cyborg which is a just a person with non-organic parts. These parts must augment the persons life in some way such as a person with a pacemaker is a borg because they have an embedded device that keeps them alive. (If that is not an augmentation of their life them what is?) I am a borg because of my use of electronic devoces to control pain. I use a tens machine all the time. I also have other devices I use all the time such as my pocket PC and other devices in my Personal Area Network (PAN). I believe there are a ton of us out there and we need to gather the collective via the internet and hopefully we can do this by using blogs and IM. \\\\//_ LLAP .||.  '),(8594,3356970,'16,May,2004','2004-05-16 00:00:00','   Hi, I am in the process of moving my blog here from another site.  I am not moving all of the posts here but will start fresh and hopefully I can get a better more mature audience here.  '),(8595,3356970,'26,June,2004','2004-06-26 00:00:00','   Yet again another on-line attack by a group from Russia. This time a triple whammy on the Microsoft front. It starts with an unknown IIS (The microsoft web server) security hole that allowed the group into a ton of internet sites to add code to the sites web pages. This added code downloads a trojan program onto your machine if you are using Internet Explorer (the Microsoft web browser) and this Trojan is so new none of the virus checkers catch it yet.  So advice people, move away from IIS if you are hosting web pages because this is a zero hour attack on you If you are just someone browsing the web move to a browser that is not based on the Microsoft browser or better yet to a non Microsoft Operating System.  This will not guarantee that in the future you will not be attacked again just that right now the bad guys are only paying attention to the Microsoft world.  ysrd  '),(8596,3356970,'13,June,2004','2004-06-13 00:00:00','   Well it was a blowout, and YSRD is again the world champ. We had everytool you can think of and all major firewalls and OS combos you can imagine and Every root password of the target servers and all contestants falling to the champ.   the contest is based on the hacker being able to get into the system without getting caught by any honey traps and IDS software and get the holy grail of hacking, the root password. This involves finding and exploiting any hole in the system and then cracking the native mode root or administrator password and ten announcing their ability by changing a file or web page on the server and doing so while the owner is on the machine and possibly seeing you do the damage. Secondary is protecting your own system while you do this to every other machine on the network. With 6 native servers on the network (the targets) and from 4 to 20 people connected at once trying to do the same thing to you. Winner take all, getting all root passwords is an automatic win.   This year we added wireless and Bluetooth segments to the network and had people using Mac, Intel, Palm and PPC systems. It was a ball and one of the only safe hacking forums left with all contestants and participants signing wavers to allow hack attacks while connected to the network. (All legal and fun.) The main goal is to show that no system is completely secure and to show the participants (mostly 500\'s employees and administrators) how they can be attacked and to possibly show them how to defend themselves and protect their IP.  Even Ysrd learned a trick or two and everyone went away with something they could apply back at base.  I had a lot of fun as did everyone who attended. Best newcomer was a 13 year old girl (accompanied but not helped by her dad) and she shows a lot of promise as well as an enthusiasm and focus that many of her older peers could learn from.  Best firewall is a new comer that is not yet released, brought by it\'s creator in an attempt to stop the champ. Watch for him to incorporate what he learned and make a remarkable firewall that can stop anything thrown at it. Running on a new OS that will be released at the same time keep an eye out for \"the Moat(tm)\" to be released in early fall.  New Ideas for next year (or earlier if we can get together again) are tunnel and VPN hacking. Happy hacking but keep your white hats on,  YSRD  '),(8597,3356970,'10,June,2004','2004-06-10 00:00:00','   It\'s that time of year again, and we are setting up the newest and greatest of Windows, Linux and Macintosh firewalls, ids and day after pills to see how fast they will fall, it hackfest 2k4. Hey isn\'t this what having your own network is all about? You thought it was to keep people out of your machine. Hell no, you want to keep someone out you don\'t connect (that\'s what we do, all the important stuff is never connected.)  But for the duration of the fest the network will not be connected to the internet to keep prying eyes away from the festivities. Nope we are not even leaving a beacon up if you want in you have to be here by 9:00 am and BYOB.  We have 5 servers up and running on the network already and we are setting up more. Guests are to bring their own computers and cables. Don\'t bring anything you can\'t afford to lose guests computers are open targets as are the servers. Points awarded for speed and finesse; anyone getting all root passwords scores an automatic win. Judges decision final, prizes as always: the admiration and acclaim of the best of the best of your peers. \\\\//_  LLAP .||.  '),(8598,3356970,'09,June,2004','2004-06-09 00:00:00','   OK you\'ve just got hit by a virus or youv\'e been hacked or someone in your company has just deleted all f your files or , well your not sure are you. Want to be sure? Today i find on the net  a set of tools that can help you answer those questions (and more) and they all come on a single cd that you don\'t need to install to use. pretty good eh?  Yup another Knoppix disk is out and thids time it\'s the security tools distribution.  If you have to watch over a file server that might get attacked or if you are a web master or a security firm in the making this is the tools you have been looking for. There are all kinds of tools from intrusion detection (IDS) to forensic analysis of a dea hard drive. And it\'s all Open Source. Right, GPL mostly and free for anyone to use. Yes people can use these tools to cause trouble but if you now how to use them you can stop them and catch them as well. Each set of tools of a given type is grouped together in the directory structure as well as the menu structre in the windo manager. Very cool layout and every grouped directoy has an RTFM directory so you can learn how to use the tools included. Well thought out and based on top of one of the best bootable distributions out there it\'s one tool I will keep in my tool chest. ysrd  '),(8599,3356970,'06,June,2004','2004-06-06 00:00:00','   We sort of lucked out the other day. We went garage saling last week (not sailing, saling) and I picked up some remote Control toys including Malcolm the robot. Not much really just a motorized base with remote control and a blow up robot balloon on top, but with the addition of the remote controlled car and other parts we have the ability to create something more. His moving about is handled quit well with his own wheeled base for what we want to do with him but we will have to add a more useful bot on top with more brains then just hot air. (Some of our foliations are getting by with the latter but this is the real world not politic so the blowup is probably going to go away unless we can make it a bit shorter and add stuff to it. I am looking TA using an old Pentium 133 as his onboard brains with sometype of wireless networking like bluetooth to connect him to the closest network or even my PAN. Next thing is to figure out what we want him to do so that we can then program him. I have a robotic arm here from an old project that I will be adding to him. One arm for now, might leave one of the blowup arms in place as well as the head and pat of the body.  ysrd  '),(8600,3356970,'07,July,2004','2004-07-07 00:00:00','   You may have or may not have noticed that i do not bash Microsoft. I am a Linux user but I believe in using the best tool for the job and there are times that is (ahh don\'t make me say it) Windows. I use a windows powered Pocket PC for my always at the ready and always in my pocket computer. I also use their compiler and IDE for writing apps for the PPC. I use Java and Linux for development but lately I have been using WebSphere Studio on a Windows 2k box for java dev. So when I hear that MS is putting out a bunch of free lite versions of their development products I take a look. I have argued for a long time now that if the basic tools for developing for one platform (ppc) are free why can\'t a basic set of dev be available for another (desktop Windows) at least for the guy sitting at home about to have the next big idea. If putting the development tools into his hands means your platform is the one the next big thing sits on then your platform can see clear benefits. So here\'s what i have so far. The sql server and web page tool are already available. They are cut down free versions of the real thing. The others (vb.net, c#.net, c++.net, j#.net are all coming soon if not already. These are for \" hobbyists, enthusiasts, and novices\"  At the moment they are only in Beta 1 and are free to download but they say you are not licensed to distribute the results. The web page indicated that in the beta 2 stage there will be some type of license to distribute. Anytime something is free it is a good thing but with the you can\'t distribute thing in palce I am suggesting people who have studio to stay with it, if you want to start to learn how to develope then the beta one may be ok for you to try out programming (use the j# to start is what I suggest) and then move to c# and c++ If you alreadt program in VB moving to the express version will leave you wanting. Another thing I noticed is that they are trying toi keep the sizes down so if you are using an older computer this may be what you need. Only 50 meg HDD space required and the other specs are loose as well.  ysrd  '),(8601,3356970,'04,July,2004','2004-07-04 00:00:00','   What kind of Borg are you? I consider myself to be a wireless Borg but I do have to hook up so to speak once in a while to do large data transfers and recharge batteries. What the heck am I talking about? Well you see as a Borg I have wearable electronics including my Pocket PC, cell phone and medical devices. I wear a tens machine when I need to and have the PPC and cell with me at all  times. I am \"connected\" able to read my email, surf the web, view information, work on work and all that stuff wirelessly by using the celphone and my pocketPC wirelessly. The two have Bluetooth so they can talk, I have a ton of music on the ppc as well as the other stuff (including the PADD and a rudimentary tricorder.) All of these machines use batteries, which I have to recharge.  Some people are wired borgs, they have to carry a power supply or be plugged in to communicate. These include the older but still prevalent laptop borgs and people on kidney dialysis. Don\'t get me wrong I do have a laptop as well as dektop computer systems but the computer I always carry is my PPC.  How does life as a borg make my life better then the normal non-borg varient human? I use my technology to augment my human tools. I can\'t think faster by using it but I do have more information to make decisions on. I am able to control the pain from my medical problems. I am able to remeber things I couldn\'t before (such as people names and phone numbers.) Soon I will be able to see better with the edition of a heads up display and computer augmentation of what i see.  With the increase in computing power and the ability to connet to my home computer network I can use the collective power of those computers from anywhere I can connect by bluetooth or cell. Life is good.  ysrd  '),(8602,3356970,'03,July,2004','2004-07-03 00:00:00','   here we are sitting, me and my son and we are listening to our fav songs, Thousand foot crutch and SCC.  I spotted a book on my shelf, \"The Formal Logic of Emotion\". Nathan found his fav book his colouring book from SK. We are looking at the remains of a robot as well and pondering it\'s future. Another fin saturday night.  '),(8603,3356970,'02,July,2004','2004-07-02 00:00:00','   After looking at the latest virus code and at the fact it was a triple threat (used an unknown and unpatched hole in IIS two holes in IE, one unknown the other unpatched and was a completely new and undetected Trojan) is it time for people to finally understand that staying with Microsoft is a loosing proposition? How many people finally got the idea that since IE is the most used it is the most targeted? The question though is of the 15% of the internet that downloaded Moz or one of it\'s variants are going to stay with it? And how many will take the plunge and switch to a Mac or Linux? And will the virus writers move with the people?  I for one think that most people will stay with IE Windows and BG. That will be enough for the virus writers to stay with MS and the rest of us can get on with our lives and making money by fixing the computers of the 95ers.  Another thing, I know I\'m from Canada but does that mean that the Microsoft executive emails have to be in French? I don\'t speak French, I speak C, C++, java, php, VB (and Asp but don\'t tell anyone), Pascal, Ada and Fortran but not French.  YSRD  '),(8604,3356970,'06,August,2004','2004-08-06 00:00:00','   My neighbour\'s kid is getting married and because we are so close to them we are up to our ears in it. The latest thing is the photog called and won\'t be able to get there so the borg to the rescue. With the latest eye cam and a good connection we will be recording the wedding to disk across the net (as well as using the 35s and hi 8 vid.) Got to go get film and batteries. I use 35 still because even with my best camera you don\'t get the clean picture you do with analog.  ysrd  '),(8605,3356970,'02,August,2004','2004-08-02 00:00:00','   My freind gave me an old broken lcd screen for a security system. i am trying to fix it and then see what I can attach it to. I am thinking some sort of display for a robot. it\'s 12 volt, and stabdard tv signal. (NTSC) Should be a fun project.  '),(8606,4038842,'27,July,2004','2004-07-27 00:00:00','   urlLink THIS &nbsp;IS THE DAILY RESULT OF THE GENERAL OPINIONS WRITTEN DOWN HERE... '),(8607,4038842,'18,August,2004','2004-08-18 00:00:00','   Ever wondered why there is no real joy in this life? Why every happiness is only a short break  in this great stream of misery? THERE IS NOTHING!!! FUCKING GET IT! '),(8608,4038842,'28,July,2004','2004-07-28 00:00:00','   Wednesday, 28.07.2004 Seems like it`s getting a little better. A hint of summer has come this morning - the sun shines, makes it a little warmer as yesterday und enables me to produce enough serotonine not to feel as bad as yesterday. Work is slightly more fun today. Hope this is not only a little grain against the big grey rock of the other days. I wish I could have fun again... Strange to be a biomechanic organism, dependent on external factors like small water drops between me and the sun. I don`t want to be a machine! '),(8609,4038842,'27,July,2004','2004-07-27 00:00:00','   urlLink EVERYTHING IS NOTHING  is the place where I will try to put down general statements about being miserable - don`t even bother to check it out, haha... '),(8610,4038842,'26,July,2004','2004-07-26 00:00:00','   Tuesday, 27.07.2004  Weather is still not as it should be. Cloudy sky, low temperature, grey atmosphere...  Have to set up and try to think of any action that might fuel me despite of these negative conditions - something to look forward to, like a good meal, like getting drunk, like meeting someone...  But again, there is just coffee in here to raise my blood pressure which is near comatose. Life slowed down to an almost unrecognizable frequency...  I want summer, I want sun, I want warmth, I want laughter, I want life!!!! '),(8611,4038842,'26,July,2004','2004-07-26 00:00:00','   Monday, 26.07.04 Monday sucks anyway - weekend`s over, been to short, back to work... Today`s even worse - cloudy and foggy weather, low temperature, rain, depressive atmosphere, pressing me down to earth, into the earth, makes me sink into the ground. Feeling pushed into the chair, unable to make conversation, unable to follow other`s thoughts, hanging on to mine which circle around warmth, the darkness and peace of sleep, eternal sleep. Getting tired and sleepy. Coffee won`t help it, doesn`t wake up, just makes nervous.  Sitting in the office, having to read through websites like  urlLink http://www.dvd-player.pl &nbsp;- it is the job to register those... counting the minutes until this day is over and I`m able to sink into dreams of a summer that hasn`t come yet and probably won`t come this year. Let`s see what tomorrow brings.. '),(8612,4038842,'18,August,2004','2004-08-18 00:00:00','   Guess why there hasn`t been an update during the last two weeks... The sun was shining and I was trying to catch as much as possible of the sun`s positive energy. Lying around in the park, swimming in the lake, sitting in the evening sun, having some beers... life could go on like this FOREVER! Unfortunately, the clouds are coming back these days. Seems like summer is over this year. Back to normality, back to miserability, back to everything which is... ...nothing... '),(8613,4038842,'03,August,2004','2004-08-03 00:00:00','   Got up much too early due to heat, cutting sunlight and droning noises from the construction site on the street below... too little sleep can be a drug, too... Then read two mags/newspaper, both with very personal content. First, a review of my band`s first demo - and not a bad one! Feels good to read someone else is having fun with my art. Second a portrait of a lobbyist in Berlin who works in a job that I was applying for half a year ago. If I was now there... less problems, I guess... who knows... again, I feel like a leaf in a storm, being blown in any direction without almost any influence where the ground is touched. '),(8614,4038842,'02,August,2004','2004-08-02 00:00:00','   Sun and summer have finally arrived. Blue sky instead of grey clouds. Endorphines are set free. The skin burns and to this moment, it feels good. As good as the sweat from the heat smells... Coming to life, growing like a plant, raising towards the energyproviding mythical power in the sky, the golden burning goddess. Again, how weak must a human be to be so dependent from his surroundings? '),(8615,3779963,'28,July,2004','2004-07-28 00:00:00','    I could stay awake just to hear you breathing  &nbsp;Watch you smile while you are sleeping   While you\'re far away dreaming   I could spend my life in this sweet surrender   I could stay lost in this moment forever   Every moment spent with you is a moment   I treasure Don\'t want to close my eyes   I don\'t want to fall asleep Cause   I\'d miss you baby   And I don\'t want to miss a thing  &nbsp;Cause even when I dream of you   The sweetest dream will never do   I\'d still miss you baby   And I don\'t want to miss a thing   Lying close to you feeling your heart beating   And I\'m wondering what you\'re dreaming   Wondering if it\'s me you\'re seeing   Then I kiss your eyes And thank   God we\'re together   I just want to stay with you in this moment forever   Forever and ever   Don\'t want to close my eyes   I don\'t want to fall asleep   Cause I\'d miss you baby   And I don\'t want to miss a thing   Cause even when I dream of you   The sweetest dream will never do   I\'d still miss you baby   And I don\'t want to miss a thing   I don\'t want to miss one smile   I don\'t want to miss one kiss   I just want to be with you   Right here with you, just like this   I just want to hold you close   Feel your heart so close to mine   And just stay here in this moment   For all the rest of time   Don\'t want to close my eyes   I don\'t want to fall asleep   Cause I\'d miss you baby   And I don\'t want to miss a thing   Cause even when I dream of you   The sweetest dream will never do   I\'d still miss you baby   And I don\'t want to miss a thing   Don\'t want to close my eyes   I don\'t want to fall asleep   I don\'t want to miss a thing '),(8616,3779963,'23,July,2004','2004-07-23 00:00:00','    Sound of many TVs Being Smashed) I don\'t need no arms around me   And I don\'t need no drugs to calm me   I have seen the writing on the wall   Don\'t think I need anything at all No.   Don\'t think I need anything at all   All in all it was all just bricks in the wall   All in all you were all just bricks in the wall '),(8617,3779963,'23,July,2004','2004-07-23 00:00:00','    We don\'t need no education  We don\'t need no thought control  No dark sarcasm in the classroom  Teacher, leave them kids alone  Hey, Teacher, leave them kids alone  All in all it\'s just another brick in the wall  All in all you\'re just another brick in the wall   We don\'t need no education  We don\'t need no thought control  No dark sarcasm in the classroom  Teachers, leave them kids alone  Hey, Teacher, leave us kids alone  All in all you\'re just another brick in the wall  All in all you\'re just another brick in the wall (all repeating) &nbsp;[Teacher:] \"Wrong, Do it again.\"  \"If you don\'t eat yer meat, you can\'t have any pudding.  How can you have any pudding if you don\'t eat your meat?\" &nbsp;\"You! Yes, you behind the bike sheds, stand still laddy!\" '),(8618,3779963,'23,July,2004','2004-07-23 00:00:00','    Daddy\'s flown across the ocean  Leaving just a memory &nbsp;A snapshot in the family album  Daddy, what else did you leave for me?  Daddy, whatcha leave behind for me?  All in all it was just a brick in the wall  All in all it was all just bricks in the wall  &nbsp;\"One, two, three, (whistle), hey\" &nbsp;[Teacher:] \"You! Yes, you! Stand still laddy!\" '),(8619,3779963,'23,July,2004','2004-07-23 00:00:00','    ______________  Momma loves her baby  And Daddy loves you too  And the sea may look warm to ya, Babe  And the sky may look blue  Oooooo Babe  Oooooooo Baby Blue  Oooooo Oooooo Babe  If you should go skating  On the thin ice of modern life  Dragging behind you the silent reproach  Of a million tear-stained eyes  Don\'t be surprised when a crack in the ice  Appears under your feet  You slip out of your depth and out of your mind  With your fear flowing out behind you  As you claw the thin ice '),(8620,3779963,'23,July,2004','2004-07-23 00:00:00','    _______________  \"...we came in?\"  So you thought you might like to  Go to the show  To feel the warm thrill of confusion  That space cadet glow  Tell me is something eluding you, Sunshine?  Is this not what you expected to see?  If you wanna find out what\'s behind these cold eyes  You\'ll just have to claw your way through this disguise  [Pink:] \"Lights!\" &nbsp;\"Turn on the Sound Effects\"  \"Action\" &nbsp;\"Drop It. Drop It On \'Em\" &nbsp;\"Drop it On Them!\" '),(8621,3779963,'23,July,2004','2004-07-23 00:00:00','    _____________________ SITTING HERE WASTED AND WOUNDED AT THIS OLD PIANOTRYING HARD TO CAPTURE THE MOMENT THIS MORNING I DON\'T KNOW\'CAUSE THIS BOTTLE OF VODKA\'S STILL LODGED IN MY HEADAND THIS BLONDE GAVE ME NIGHTMARESI THINK THAT SHE\'S STILL IN MY BEDAS I DREAM ABOUT MOVIES THEY WON\'T MAKE OF ME WHEN I\'M DEADWITH AN IRONCLAD FIST, I WAKE UP AND FRENCH KISS THE MORNINGWHILE SOME MARCHING BAND KEEPS ITS OWN BEAT IN MY HEAD WHILE WERE TALKINGABOUT ALL OF THE THINGS THAT I LONG TO BELIEVEABOUT LOVE, THE TRUTH AND WHAT YOU MEAN TO MEAND THE TRUTH IS, BABY, YOU\'RE ALL THAT I NEEDCHORUS: I WANT TO LAY YOU DOWN ON A BED OF ROSESFOR TONITE I SLEEP ON A BED OF NAILSI WANT TO BE JUST AS CLOSE AS THE HOLY GHOST ISAND LAY YOU DOWN ON A BED OF ROSESWELL I\'M SO FAR AWAY THAT EACH STEP THAT I TAKE IS ON MY WAY HOMEA KING\'S RANSOM IN DIMES I\'D GIVE EACH NIGHT JUST TO SEE THROUGH THIS PAYPHONESTILL I RUN OUT OF TIME OR IT\'S HARD TO GET THROUGH\'TIL THE BIRD ON THE WIRE FLIES ME HOME TO YOUI\'LL JUST CLOSE MY EYES AND WHISPER, BABY, BLIND LOVE IS TRUECHORUS: I WANT TO LAY YOU DOWN ON A BED OF ROSESFOR TONITE I SLEEP ON A BED OF NAILSI WANT TO BE JUST AS CLOSE AS THE HOLY GHOST ISAND LAY YOU DOWN ON A BED OF ROSESTHE HOTEL BAR HANGOVER WHISKEY\'S GONE DRYTHE BARKEEPER\'S WIG\'S CROOKED AND SHE\'S GIVING ME THE EYEI MIGHT HAVE SAID YES BUT I LAUGHED SO HARD I THINK I DIEDSOLOWHEN YOU CLOSE YOUR EYES KNOW THAT I\'M THINKING ABOUT YOUWHILE MY MISTRESS SHE CALLS ME TO STAND IN HER SPOTLIGHT AGAINTONITE I WON\'T BE ALONE BUT YOU KNOW THAT DON\'T MEAN I\'M NOT LONELYI\'VE GOT NOTHING TO PROVE FOR IT\'S YOU THAT I\'D DIE TO DEFENDCHORUS: I WANT TO LAY YOU DOWN ON A BED OF ROSESFOR TONITE I SLEEP ON A BED OF NAILSI WANT TO BE JUST AS CLOSE AS THE HOLY GHOST ISAND LAY YOU DOWN ON A BED OF ROSES '),(8622,3779963,'12,July,2004','2004-07-12 00:00:00','    Everybody\'s waiting for something to happen  Everybody\'s waiting for something to see  Lunatics waiting for bigger diadters  Everyone\'s waiting for news on TV   Winding lives at the end of the spiral  Waiting dictators with their next big thrill  Everyone\'s looking but no one is listening  Everybody wants to be in at the kill   I wait for the signs, they tell me true  I see the signs of the end of time   Everyone\'s searching but nothing\'s revealing  Everyone\'s looking for the reason why  Everyone\'s hoping for life everafter  Everyone\'s looking at death from the sky   Everyone\'s nightmares are going to happen  Everybody\'s ripping the mask from their eyes  Everyone\'s praying but no one\'s believing  Everyone\'s heroes tell everyone\'s lies   So I watch and I wait  And I pray for an awnser  An end to the strife and the world\'s misery  But the end never came  And we\'re digging the graves  And we\'re loading the guns for the kill   Can the end be at hand?  Is the face in the sand?  Future memory of our tragedy   Can the end be at hand?  Is the face in the sand?  Future memory of our tragedy '),(8623,3779963,'05,July,2004','2004-07-05 00:00:00','    Button your lip. Don\'t let the shield slip. Take a fresh grip on your bullet proof mask. And if they try to break down your disguise with theirquestions You can hide, hide, hide, \"I\'ll tell you what, I\'ll give you three blacks, and play you forfive ...\" \"Ta! You was unlucky there son\" \"Time gentleman!\" Behind paranoid eyes.  You put on our brave face and slip over the road for a jar. Fixing your grin as you casually lean on the bar, Laughing too loud at the rest of the world With the boys in the crowd You hide, hide, hide, Behind petrified eyes.  You believed in their stories of fame, fortune and glory. Now you\'re lost in a haze of alchohol soft middle age The pie in the sky turned out to be miles too high. And you hide, hide, hide, Behind brown and mild eyes. '),(8624,3779963,'05,July,2004','2004-07-05 00:00:00','    I\'m alone yeah I don\'t know if I can face the night I\'m in tears and the cryin that i do is for you I want your love let\'s break the wall between us Don\'t make it tough, I\'ll put away my pride Enough\'s enough, I\'ve suffered and I\'ve seen the light  You\'re my angel come and save me tonight You\'re my angel come and make it alright You\'re my angel come and save me tonight  Don\'t know what I\'m gonna do about this feeling inside Yes, it\'s true loneliness took me for a ride Without your love, I\'m nothin but a beggar Without your love, a dog without a bone What can I do, I\'m sleepin in this bed alone  You\'re the reason I live You\'re the reason I die You\'re the reason I give when I break down and cry Don\'t need no reason why Baby, baby, baby '),(8625,3779963,'05,July,2004','2004-07-05 00:00:00','    You gotta find a way Yeah, I can\'t wait another day Ain\'t nothin\' gonna change If we stay around here I gotta do what it takes Because it\'s all in our hands We all make mistakes Yeah, but it\'s never to late to start again (yeah yeah) Take another breath And say another prayer  Then fly away from here To anywhere Yeah, I don\'t care If we just fly away from here Our hopes and dreams Are out there somewhere Won\'t let time pass us by We\'ll just fly  If this life Gets any harder now It ain\'t no nevermind Ya got me by your side And anytime you want (fly fly fly) Yeah we catch a train And find a better place Yeah, \'cause we won\'t let nothin\' or no one keep gettin\' us down maybe you and I Could pack our bags and hit the sky  Then fly away from To anywhere Yeah, I don\'t care If we just fly away from here And our hopes and dreams Are out there somewhere I won\'t let time pass us by We\'ll just fly  Do you see a bluer sky now You could have a better ride now Open your eyes  Because no one here can ever stop us They can try but we won\'t let them No way-ay-ay-ay-yeah-ah  maybe you and I Could pack our bags and say goodbye  Then fly away from here To anywhere No honey I don\'t care If we just fly away from here And our hopes and dreams Are out there somewhere Fly away from here (ya ya ya) Yeah anywhere Now honey, I don\'t I don\'t I don\'t fly (yeah)  We just fly (fly away) '),(8626,3779963,'05,July,2004','2004-07-05 00:00:00','    Humidity could be felt on the walls  Touched with the palm and used to scare  They used to sweat, they used to stink  Everything swamped and hot  But in the corner, Laying on a bed  A cold piece, made to stay alive  Trapped within its body   It could not think anymore  Thoughts of times of sanity  The world was isolated  Where the sun would salute him  And the night was violent  Fear and guilt  Invade the corners of the room  Pain was felt constantly  They keep on destroying   It could be seen through the window  The eye of disgust and scorn  When you hear the laugh of a madman  That\'s about to die  To suffer alone in disgrace  His hate is his own  Always hating being alive  Sarcastic Existence '),(8627,3779963,'05,July,2004','2004-07-05 00:00:00','    ================  I got my head, but my head is unraveling Can\'t keep control, can\'t keep track of where it\'s traveling I got my heart but my heart is no good And you\'re the only one that\'s understood  I come along but I don\'t know where you\'re taking me I shouldn\'t go but you\'re reaching, dragging, shaking me Turn off the sun, pull the stars from the sky The more I give to you, the more I die  And I want you And I want you And I want you And I want you  You are the perfect drug, the perfect drug, the perfect drug You are the perfect drug, the perfect drug, the perfect drug  You make me hard, when I\'m all soft inside I see the truth, when I\'m all stupid eyed The arrow goes straight through my heart Without you everything just falls apart  My blood wants to say hello to you My feelings want to get inside of you My soul is so afraid to realize Every little word is a lack of me  And I want you And I want you And I want you And I want you  You are the perfect drug, the perfect drug, the perfect drug You are the perfect drug, the perfect drug, the perfect drug You are the perfect drug, the perfect drug, the perfect drug You are the perfect drug, the perfect drug, the perfect drug You are the perfect drug, the perfect drug, the perfect drug You are the perfect drug, the perfect drug, the perfect drug  You are the perfect drug, the perfect drug, the perfect drug You are the perfect drug, the perfect drug, the perfect drug You are the perfect drug, the drug, the perfect drug  Take me, if you want Take me, if you want Take me, if you want Take me, if you want  Without you, without you everything falls apart Without you, it\'s not as much fun to pick up the pieces Without you, without you everything falls apart Without you, it\'s not as much fun to pick up the pieces It\'s not as much fun to pick up the pieces It\'s not as much fun to pick up the pieces Without you, without you everything falls apart Without you, it\'s not as much fun to pick up the pieces '),(8628,3779963,'05,July,2004','2004-07-05 00:00:00','    Tomaste el vaso aquel aquÃ©l que no debes tomar saliste a caminar y decidiste irla a buscar mil veces y una mÃ¡s juraste no volverla a ver la fiesta terminÃ³ ya no podÃ©s tenerte en pie esta noche es hora de que pienses en cambiar el tiempo pasa pronto y todo tiene su final pasa, pasa, pasa, pÃ¡same un vaso mÃ¡s volvamos caminando pero elijamos el lugar  Noche de calor en la ciudad ella te dejÃ³ y todo sigue igual.  Mil veces y una mÃ¡s juraste que ibas a crecer y solo una vez mÃ¡s creciste hasta morir sin ver la fiesta terminÃ³ no te pudiste controlar mil golpes y uno mÃ¡s la noche te volviÃ³ a pegar esta noche es hora de que pienses en cambiar el tiempo pasa pronto y todo tiene su final pasa, pasa, pasa, pÃ¡same un vaso mÃ¡s volvamos caminando pero elijamos el lugar  Noche de calor en la ciudad ella te dejÃ³ y todo sigue igual. quisieras volver el tiempo atrÃ¡s pero lo que vuelve es esta noche y nada mÃ¡s.  Esta noche es hora de que pienses en cambiar el tiempo pasa pronto y todo tiene su final pasa, pasa, pasa, pÃ¡same un vaso mÃ¡s volvamos caminando pero elijamos el lugar '),(8629,3779963,'01,July,2004','2004-07-01 00:00:00','    No se bien que dÃ­a es hoy solo se que te vi salir y en cinco minutos perdÃ­ las letras para hablarte a vos  Yo se que no tengo palabras y nunca las voy a tener por eso aprovecho esta noche ya ves, estoy solo otra vez  Por eso aprovecho esta noche tal vez lo puedas entender que no me importa poner las letras solo me importa mi mujer  MaÃ±ana cuando te levantes y pienses lo que dije ayer ay viejo, en este juego a mi siempre me toca perder, siempre  siempre habrÃ¡ vasos vacÃ­os con agua de la ciudad la nuestra es agua de rÃ­o mezclada con mar levanta los brazos mujer y ponte esta noche a bailar que la nuestra es agua de rÃ­o mezclada con mar  A ti te quiero decir no te preocupes mi amor que yo te voy a entender que yo te voy a querer  Siempre habrÃ¡..... '),(8630,3779963,'01,July,2004','2004-07-01 00:00:00','    Maybe your luck has changed Settle down Maybe I\\\'m just deranged And on the rebound Maybe love was the thing Holding me back from all Maybe I\\\'m just the thing To break my own fall Take the rope to my heart and fall You may just be the last before you See the black tangled heart fall Maybe departure\\\'s good Makes room for more Start to mass produce For a chance to ignore Maybe you\\\'ll kill yourself Before I get a turn Maybe I\\\'ll fall in love And never learn Take the rope to my heart and fall You may just be the last before you See the black tangled heart fall Take the rope to my heart and fall You may just be the last before you See the black tangled heart fall '),(8631,3779963,'01,July,2004','2004-07-01 00:00:00','    Is everything a baited hook? And are there locks on all doors? If you\'re looking for an open book, Look no further, I am yours.  We\'ll behave like animals, swing from tree to tree We can do anything that turns you up and sets you free.  You\'re an exception to the rule You\'re a bonafide rarity. You\'re all I ever wanted; Southern girl, could you want me?  So come outside and walk with me We\'ll try each other on, see if we fit And with our roots become a tree To shade what we make under it.  We\'ll behave like animals, swing from tree to tree We can do anything that turns you up and sets you free.  You\'re an exception to the rule You\'re a bonafide rarity You\'re all I ever wanted;  Southern girl, could you want me? '),(8632,3779963,'07,August,2004','2004-08-07 00:00:00','    _____________  Me quieren agitar Me incitan a gritar. Soy como una roca, Palabras no me tocan Adentro hay un volcan Que pronto va a estallar. Yo quiero estar tranquilo.  Es mi situacon Una desolacion, Soy como un lamento, Lamento boliviano.  Que un dia empezo Y no va a terminar Y a nadie hace da~o.  Y yo estoy aqui Borracho y loco Y mi corazon idiota Siempre brillara Y yo te amare, Te amare por siempre. Nena no te peines en la cama Que los viajantes se van a atrasar.  Y yo estoy aqui Borracho y loco Y mi corazon idiota Y yo te amare, Te amare por siempre. Nena no te peines en la cama Que los viajantes se van a atrasar '),(8633,3779963,'05,August,2004','2004-08-05 00:00:00','    ____________  Lost feelings for you await for me in my dreams  Underneath me everything is disappear   I will neverEscape your dreams  Lie to me, hate me forever you will lie and   I will fall I canÂ´t look in your face only parts of you  You call my soul and I will never feel  Await for me in your place (3x)   Underneath me everything is disappear  Following.........my insticts  All my feelings disappears in YOU   Lost feelings for you await for me in my dreams  Underneath me everything is disappear   I will never escape your dreams  Lie to me hate me forever   I canÂ´t look in your face only parts of you  You call my soul and I will never feel  Await for me in your place I canÂ´t look in your face    Await for me in your place (2x) '),(8634,2726395,'31,January,2004','2004-01-31 00:00:00','    Thorn in my side: Repressed anger was perfectly okay with me  Psychologists agree that the problem with most people is repressed anger. They behave a certain way because they\'re repressing something. I never thought I had any repressed anger, that I was perhaps the only person in the world who was incapable of repressing anything.  And I was good with that.  Now this whole thing with my family has dredged up all sorts of things from my past--stuff I was long past, stuff I can\'t do anything about anyway, stuff I can\'t change.  Yeah, yeah, my parents decided to relive their youth the minute they decided I was old enough to watch my sisters, which was when I was eleven, and went off and played bingo six nights a week. Yeah, I never got to go out with my friends because I always had to take care of my sisters, yeah, I never had any money because I couldn\'t babysit for money, because I had to watch my sisters for free and didn\'t get an allowance.  Yeah, this lasted until I was eighteen.  At the time it didn\'t seem all that bad, after all, I shuttled my sisters off to bed and stayed up and watched whatever I wanted on TV and slammed myself into bed two minutes before my parents came in the door.  I got to stay up and write and imagine and plan and pretend I was on Broadway. It was okay.  But now I have one psychopathic sister and one so self-absorbed that she shoved off her pyschopathic sister\'s fake suicide attempt onto someone else because she couldn\'t be bothered.   And a counselor who, in pointing out that Heidi\'s psychopathia is not genetic, but a by-product of failing to bond with absent parents, made me very damned angry.  And it\'s coming out everywhere.  I\'m glad that my kid isn\'t going to go off to a life of criminal sociopathy because of anything I might genetically pass on to him, and that I am not going to someday lose my mind and start stealing social security numbers. That is a huge relief.  But it never once occurred to me that when my parents left me to care for my sisters they were five and seven years old.  I turned out okay because I had a childhood with them. My sisters did not.  How could these people have been so blind?    All right, so not everyone who experiences a \"disruption in the bonding process\" during early childhood becomes a dangerous psychopath, but almost all psychopaths have had that disruption. I look at my little five year old son, and I think, \'how could they leave a five year old in the care of an eleven year old every night?\'  This sweet child, how could I even consider leaving the house every single night?! I don\'t want to miss a moment!  So out comes the anger. And lots of it. I can barely stand to watch my parents shower my son with affection and attention they denied their own children all of a sudden.  I am not jealous, I am angry. Where was all of this all those years ago? Maybe I would have had sisters like other people. Maybe Heidi could have become the lawyer she should have been and Nancy wouldn\'t consider taking my phone call a courtesy.    And I can\'t do a thing about it. I should just get over it. The fact that it\'s all coming out now is making me angry. The fact that I can\'t just put it nicely back the way one can never refold a map the right way makes me angry. Being angry at all makes me angry.  I find myself angry at innocent people. At Mark because he...well, for no reason, really. At dear Thomas because he still won\'t tell me why he chose to try to keep his engagement a secret from me when it was so obvious anyway. At his fiancee Nancy, who has chosen to hate me regardless of a year and a half of concentrated effort on my part to befriend her and earn her trust that I\'m not sleeping with her man.   At my work, who legitimately deserves it, but I\'ve been able to control it because of the whole they-give-me-money-for-not-much-work thing.   So much anger, all at once! And all for something that didn\'t seem so bad in the \'70s!  I got the duvet cover on, by the way. I am very proud of that.  And the owner of my company actually told me he liked the latest thing I wrote and I was to inform comma-less Harriet and machine-writer Robin that they were to use my copy.  So there is some good to outweigh the bad.  But I really still thing there\'s a lot to be said for repressed anger. '),(8635,2726395,'26,January,2004','2004-01-26 00:00:00','    Thorn in my side:  Human decency, or, lack thereof, also, duvet cover installation instructions, or, lack thereof  Has anyone ever tried to install a duvet cover? It even looked like it would be hard to do in the store--imagine trying to stuff a pillowcase that\'s thirty times larger than you are.  But I thought I\'d try it.  There were a few unexpected surprises. For one, I really thought that the thing would be sewn up on three sides, open on one, with some sort of zipper or buttons or something for closing up the patient once you got the thing installed.  You can imagine my shock when I discovered there was only a tiny opening at one end with two buttons on it--far too small to fit the duvet in without ripping it to shreds. Mark and I looked at each other, and then Mark wordlessly left the bedroom.  I thought I would turn to the Internet for installation instructions, since the box helpfully came with none.  All I could find were humorous pieces about how difficult it is to install a duvet cover, but nothing in them which shed any light on how to get a 9 x 9 duvet into a 3-foot opening.  I look at all the stores, I look at the pictures of duvets with covers on them, and I think, \"somebody knows something...but they\'re not talking.\"  I imagine a secret society of insiders who have The Knowledge.  Nobody\'s sharing. I bet the Fab 5 know, I bet Kai can whip one of those puppies on in two seconds flat. But alas, I am female and have a sense of style, so they would have no interest in my case.  And I think, is there no more human decency?  Speaking of human decency, or lack thereof, my psychopathic quasi-suicidal sister got outed, and not in a good way. She planned the whole \"suicide attempt\" in order to bilk my parents out of money. Terrible thing to discover about one\'s own family, but it\'s better than not finding out.  (See post one for sister\'s suicide saga.) My parents are having a hard time with the news, but I think they do believe it.  So I didn\'t save anybody\'s life the night of the 14th. I\'ve spent the last two weeks trying to convince my parents not to hand over their life savings to a psychopath, and have been met with every sort of reaction from suspicion that my motive in keeping them from being bilked is for my own financial gain, accusations that I am being a prima donna because my youngest sister is the focus of attention, that I was trying to prove my parents were stupid, anything except that I was trying to protect them from falling into the grasp of a diagnosed psychopath so they wouldn\'t end up completely impovershed in their old age. My only motivation. Protect parents from predator.  The only  reason I even tried to prevent my sister from killing herself in the first place is so that she wouldn\'t hurt my parents, and because it was the Right Thing to Do.   You know what I think?  This, coupled with a couple of other incidents with some other people, has led me to believe that human decency is so rare that people cannot even recognize it anymore.  Everyone sees conspiracies in everything. I guess you could say it was my own lack of faith in human decency which led me to uncover my sister\'s plot. Even I couldn\'t take a suicide attempt at face value! But I have recently been told that gestures of friendship I have made towards someone were met with suspicion and revulsion--\"why would she compliment my two-dollar Target earrings? Why does she notice my outfits? What\'s she trying to get? Why do they always invite us to things?\"  This person was incapable of recognizing overtures of friendship, because she automatically assumed there was an ulterior motive.  What is WITH people these days?   Yes, I notice two dollar earrings, if I like them. I don\'t withhold a compliment if I suspect something wasn\'t expensive, what a thing to say!  I compliment people when I genuinely like what they\'re doing or wearing, because I know I enjoy compliments myself, and would like to be able to give that feeling to someone else. If I think someone looks slobby and can\'t seem to understand color matching or personal hygene, I shut up.   I make overtures of friendship towards people I like because I like them.  I steer clear of people I don\'t like. And because if I waited for people to be friendly with me I\'d end up living alone with six dozen dachshunds.  It is so sad to me that people cannot even recognize who is a friend and who is a foe anymore.      I understand, it\'s true that there are psychopaths in this world who are friendly and nice and are out to bilk every last dime out of you (\"cashews,\" my friend Dave calls them--nuts who are trying to get cash out of you) and you have to be careful in this world, but I think it may be possible that there are other people out there who also believe that someone just genuinely wants nothing from you but to be your friend, and to have a freind in return.  Isn\'t there? '),(8636,2726395,'22,January,2004','2004-01-22 00:00:00','    Thorn in my side:   Hail Britannia  I am a lifelong Anglophile. The first chance I had to get out of this country, I headed straight for London. When I went on my honeymoon, there was no question. Back to England. I am proud to have been raised to drink tea with milk and raise my pinky. I speak and translate British-English.  I can do no fewer than six different British dialects. But since I came to work for this company, which appealed to me because of its British sub-persona, my tastes are becoming more continental.  A stoic, bespeckled 27 year old by the name of Harriet, and Robin, her dead-on lookalike for Veruca Salt\'s father in \"Charlie and the Chocolate Factory\" boss are the cause.  Harriet has never heard of a comma. Robin is the only human being I know who can write copy that reads like it was computer generated.  And they don\'t seem to know how to say \"Hello\" in under 25 pages. These people are *righteous* about their incomplete, run-on sentences, too It is *their* language, they should be the authorities on how to use it.  Sometimes their sentences are so convoluted even they have lost track of what they were supposed to be talking about.  It is my job to edit them into something that resembles readability. Not that it matters. I ship boxloads of punctuation over to the UK, and they never seem to open the package.  Final printed material comes back to us that is exactly the same as it was before I spent hours untangling it.  Our company\'s material is off the Fleish-Kincaid Readability Scale, and there\'s nothing I can do about it.  Everyone here knows that no one will read this stuff. It\'s unreadable. The collateral material that is going into my portfolio looks like the Rosetta Stone, only somewhat less understandable.  I am essentially editing and re-writing, literally, for my health. Hail Britannia.  To make it worse, I realized the other day that I couldn\'t even enjoy the latest Rowan Atkinson movie. And I have always adored Rowan Atkinson, unless he got naked and showed his skinny hairy behind on worldwide TV.  Now whenever I talk to Robin or Harriet, images of Rowan Atkinson\'s hairy skinny behind pops into my mind. I tried to replace it with Hugh Grant, but he\'s too skinny, too.  The point is, I shouldn\'t be thinking about famous British backsides every time I talk to my colleagues. But they talk like they write, and so my mind wanders.  I have comma-less Harriet and Veruca Salt\'s dad to thank for that.  It\'s not \"whilst,\" Harriet. You\'re writing high tech collateral material, not iambic pentameter verse. Not that you bloody know what Iambic Pentameter Verse is. Maybe I\'ll switch to coffee for awhile.  So now my father is not speaking to my husband, because my husband Mark defended my position about their open-checkbook policy towards my psychopathic sister. Or, at least we think he\'s not talking to him. It\'s hard to tell with my dad. He could just be being really quiet.    Luckily my friend and highly patient German teacher Thomas imparted a bit of wisdom, albiet from an unexpected source. \"Your story is nothing new,\" he told me, in his softly accented, quiet, unassuming way. \"It\'s in the Bible. A few times.\"  Thomas is a religious man, but not fanatically so. He wasn\'t trying to give me a lesson in religion or suggesting that I turn to God for answers. He was just stating a fact, when we met for quesadillas and something, anything, with alcohol in it.  \"The Good Child is often overlooked and scorned in favor of the Evil Child.\"  I thought about this. I know my Biblical commentary. I\'ve read everything Jonathan Kirsch ever put on paper, and most of the stuff in the bibliographies. I thought about Jacob and Esau, and then how Jacob pulled the same stunt on his own kids.  Thomas was right. Thomas usually is. Unfortunately, the Bible is a bit short on information about how to cope with it, but I think maybe the point is there: This is nothing new. I am not alone. There\'s nothing much we can do about our parents. It\'s hard to watch, but I can\'t do anything about it. Not unlike watching a train wreck. And my sister is the biggest train wreck of them all.  But it\'s nice to know: the overlooked Good Child usually is okay, except for a couple of biblical daughters who got needlessly sacrificed. And it\'s happened before and it will happen again.  I\'m not alone historically, and I am not alone literally, either. Thanks, Thomas. A few words go a long way. Harriet and Robin could take a lesson from you.  I can look at the bright side. At least my parents aren\'t British. '),(8637,2726395,'21,January,2004','2004-01-21 00:00:00','    Wednesday, January 21, 2004.  Thorn in my side: My family.    Welcome to my blog, my pyschopathic sister tried to kill herself a week ago tonight, I saved her life and today I am the family\'s pariah.  Still with me?  All right, if you can take that, you can take anything else I might write.    Somehow I felt I ought to be sharing all this, and what is going on. Maybe it\'s a cautionary tale, maybe it\'s interesting, maybe its a comment on what society has become, maybe I\'m just whining.  At any rate, welcome to Rosenthorns.  So, let\'s get back to this sister\'s attempted suicide thing.  My sister is a psychopath. And I don\'t mean that in the colloquial sense, the way we used to use the word \"terrorist\" to describe an energetic two-year-old, but don\'t anymore.  I mean, she is a severely ill psychopathic woman.  She has caused my entire family nothing but pain for all of her adult life.  She stole from my parents, lied to everyone she ever knew, cost me every old friend I ever had with whom she came into contact, stole my SSN and took my name, ruined my credit, cost me at least two major corporate jobs when I failed the background check, and cost me thousands of dollars in additional fees when I tried to buy a house. She poisoned my dogs.  She has never had a long term relationship with another living soul, but has created some very imaginative boyfriends. Real boyfriends can be identified by their restraining orders against her. This is not a colloquial psychopath, this is a clinical psychopath.  Last Wednesday night, I got a call from someone I never heard of. \"Regarding your sister,\" was all the message said. I almost didn\'t call. \"God,\" I thought. \"she ripped someone else off and is now coming after her family for the money.\"  But I did call. It was the Right Thing to Do. And Pam always does the Right Thing.  The voice on the other end of the phone told me that my sister has been threatening suicide, and earmarked the next day as the day she would do it.  My other sister had told this person that our sister is a pathological liar and that she should ignore the messages. However, if she felt that she needed to, she should call up our sister Pam, because Pam always Does the Right Thing.  I told this person that there was at least a 60% chance that my sister was lying, but because the next day was our parents\' anniversary, I thought it would be best to at least attempt follow through. After two harrowing hours on the phone with a completely inept San Francisco Police Department, they broke into her house and found her chock full of Valium with a plastic bag over her head and two suicide notes.  Apparently nobody in our family can say anything in less than two pages.  She was \"detained\" and then promptly lost by the SFPD for another two hours, while I tried to track her. The ineptiude of SFPD is epic, but that\'s another Thorn in My Side.  Anyway, now it\'s a week later. My parents, who never were particularly good decision makers, have come rushing to her aid with open arms and checkbook. They mortgaged their home and are paying her way through everything while my sister continues to lie to them.  They are supporting her rather extravagant lifestyle while she gets \"help,\" though most psychiatrists agree that psychopaths are beyond the help of their profession.  They have practically canonized her, constantly speaking in glowing terms about how she is going to get job training, how she is going to make good on all her debts, how she is going to get better and has such a bright future.  I raised a point. \"Um,\" I said, \"Can you afford to be supporting her to the tune of $2000 a month?  Is there a reason she has to stay alone in her three-bedroom apartment in North Beach with a view of the bay that costs more than my home mortgage?  Do you have any time limit for this open-checkbook policy? Do you have any expectations of her?\"  I explained to them there was a difference between helping and enabling, and that they were not doing anybody any good by supporting her lifestyle so that she can go out and lie, cheat and steal some more with no end in sight. I explained that help would entail a plan, with an end-date, a budget based on what they could afford, and a list of expectations.  My parents angrily told me that they would pay for \"as long as it takes,\" informed me that I have no compassion, and told my sister that she should not talk to me because I am \"hysterical.\"  There\'s another aspect to this little tale.  I have a five year old son.  My husband and I really feel that a five year old really shouldn\'t be subjected to a psychopath who has already shown she is dangerous to herself and to others, and let\'s not forget that she poisoned our dogs, not once but twice, with what two sets of vets think was some kind of detergent. My parents insist my sister is of no danger, and that I am being horrible. The words \"Child Endangerment\" seem to mean nothing to them.  Meanwhile, some of my other friends have been very critical of my actions in preventing her suicide. Here in San Francisco, we are very politically correct, you see, and we are allowed to do with our bodies what we please. I am apparently a fascist because I prevented my sister from doing what she pleased with her body.  Still others have criticized me for saving her because she is so useless, she\'s caused nothing but pain, she wasn\'t worth saving, and now I\'ve caused my family more because I had to be a hero.  I am not a hero. I did not save my sister for her sake. I saved my sister so it wouldn\'t hurt my parents. There\'s an irony, huh? I am not a willing victim, as someone suggested. I did the Right Thing. I know I did.  But damn it all, it is awfully lonely. '),(8638,2726395,'18,February,2004','2004-02-18 00:00:00','    The Waiting Game  I have a lot of faults, but the biggest among them is impatience. This is particularly egregious because I also have no patience for other people who are impatient. I\'m always telling people, \"calm down, there\'s nothing you can do about it,\" or, \"take control of it and make a plan and follow through, don\'t just wait around for something to happen,\" or the one I\'ve been repeating a lot lately to the various male friends in my life: \"No, you aren\'t having a midlife crisis, you have a lousy job. They aren\'t mutually exclusive.\"    But when it\'s my turn to wait, there is nothing that can console me. And right now I am doing a lot of waiting. I\'m up for a really good job that is so perfect for me you\'d think God heard my prayers. It is in Fremont, it is in my industry, my boss would be in Germany,so I would have to use my German, especially when I traveled there on the company dime, and I would be working while all my superiors were asleep nine hours away. It  just doesn\'t get any better than that, except that it also pays considerably more than I\'m making right now.  They\'ve told me I am the candidate. They have told me they will be sending me an offer letter. Twice now. And so far, no letter has appeared.  The offer was to be contigent upon passing a background check, anyway, which won\'t be complete until probably tomorrow, or Friday. But I wanted to have at least something on paper that says \"you liked me. You really liked me.\"  I am a theatre geek, I need instant gratification, instant approval. It\'s why I hated radio and TV.  It doesn\'t help to know that if I got this letter, there\'s nobody here to even give notice to until next week.  But I am impatient. As my sister Nancy wrote in her often family-bashing blog, the sign of a true Depper is to be riddled with anxiety, spitting adreneline and desperate for someone other than myself to blame. That sums it up pretty well, except that I am not blaming anyone else. I am impatient because I am impatient. (It suddenly strikes me as funny that Nancy, in her vitriolic attacks upon her own family, points out that she is desperate for someone else to blame because she is a Depper; effectively blaming someone else for the fact she is desperate for someone else to blame.)  I wait also for Thomas, suddenly terribly paranoid and acting like a trapped animal, I can almost see him pacing in his cage. I want to tell him this, but it would do no good. Thomas builds his own cages and he is very good at it. He\'s also really good at tossing the keys beyond his own reach. That\'s his neurosis. But I wait for him to calm down and be normal again, wait for him to return to the sweet and gentle man I learned to love like a brother. Wondering if he ever will, or if I can get used to him being as touchy as he is about absolutely everything. And he\'s probably waiting for me to calm down and get this stupid letter from this company that is dangling it before me like a piece of carrot cake just out of my reach.    I hate waiting. '),(8639,2726395,'11,February,2004','2004-02-11 00:00:00','    Psychopathia--The Gift that keeps on giving  I should have seen it coming. After all, Mark\'s mother has done it twice, and she\'s a psychopath, too. That\'s something Mark and I have in common--we are genetically related to psychopaths who insist their problems are our fault. And we look at each other in amazement at the sheer gall of it all.  When Mark\'s mother sent the letter five years ago informing Mark that she still wanted a relationship with him despite the fact that he was a horrible person and she a hapless victim, Mark ignored it. He started ignoring her soon after she was committed for trying to kill him when he was a child. Back then they had places for people like that.  So I think I will do the same from the letter I got from my sister today, which informed me that I am a very sick woman in desperate need of psychiatric help and medication, for the sake of my husband and innocent son. She wrote with such dramatic compassion that Mark couldn\'t help read it to me with his best Hal J. Todd Oral Interp style--a reference only San Jose State Theatre grads will get, but I don\'t guess anyone\'s reading this, anyway.    Good God, she\'s good. Manipulative, well spoken, and can really convince you that the problem is really you. I could actually fall for this stuff, I can really see how people fell, hook, line and sinker.  She\'s better than Mark\'s mother, though she doesn\'t know how to use a comma, either. Mark\'s mother is a piker compared to my sister Heidi.  I think I\'ll keep the letter as an example of just how manipulative a true psychopath can be.  I don\'t think I\'m angry anymore. I don\'t think I\'m anything. I tried to muster up some emotion about getting this letter, but it rang false. My provincial parents are just raw meat for this bloodsucker, and I can\'t do anything about it. I just can\'t worry about it.  No, instead I need to worry about other people\'s lives. Dear Thomas, you dear sweet man, you\'ve really managed to mess up your life, haven\'t you? It\'s painful for me because I\'m helpless to stop you; it\'s like you\'ve traveled back in time, you know what\'s going to happen to the Titanic but insist on boarding it anyway. You\'re a mass of raw nerves and I can only even recognize portions of your personality.  I wish I could help you. It frustrates me that I can\'t. All I can do is wait and hope you don\'t get swallowed up completely.    This entry is not much for structure, but I\'m really tired and don\'t feel like being terribly witty or literary right now.  But at least the score is Psychopaths-0, Pam -1. '),(8640,2726395,'02,February,2004','2004-02-02 00:00:00','    In praise of Geese and their contribution to middle class sleep cycles  It started with a duvet cover. And I lay upon it, and I said, That is Good.  So I added a Baffle Box Goose Down Featherbed for the full sleeping-inside-a-marshmallow experience. And it was Good.  Mark did not admit the goodness of all these goose feathers, but he was taking his time getting out of bed.    Friends added to our newfound fascination. Frank (the father of Jacob\'s best friend Franklin), an enormous, shaven-headed man with an infectious smile and a personality that fills a room from the doorway, looked me straight in the eye. \"You gotta get 300 thread count sheets,\" he confided. \"You gotta pay  attention  to thread count.\" He placed his emphasis on the word \"attention.\"    \"We\'re very into thread count,\" his wife Lori added, confidently. Oh, my God, they were right.  I was sleeping better, even through all this strife.  On Saturday, it was Mark who saw them first. Goose down pillows. I wasn\'t even going to say anything, after all we\'ve spent on geese and thread counts. But there it was, right before my eyes: Mark was *testing* these pillows right in the middle of Costco.  He was literally  leaning over the bin  and putting his head on them.  They were the same maker as our featherbed, and they were $39 each, for the King-sized.  The same ones at Linens n\' Things were $119, which were a few more Things than I wanted to spend on Linens.  You guessed it. They went in the cart.  I fell asleep last night trying to imagine how many goose feathers I was currently sleeping on. How many geese.  How do they get these goose feathers, do they pluck them and let the geese grow new ones?  Why geese instead of ducks, or, seagulls?  I drifted off into a marshmallow sleep, where I proceeded in my latest series of anxiety dreams about innocently doing something that makes my friend Thomas think I\'m a raving lunatic.  All right, so goose feathers can\'t cure nightmares. But they\'re a dream to sleep on. '),(8641,2726395,'29,March,2004','2004-03-29 00:00:00','    Dear Thomas,  I am writing this to you here because I can no longer write to you for real. But over the course of the past two years, I have become so accustomed to the practice of beginning each day with typing \"Dear Thomas,\" that my fingers ache if they don\'t feel that pattern.  There is so much I would like to say to you. But I can never say them, and you don\'t want to hear them. And I know that you will never find this blog, so I can write them here and at least they\'ll be expressed.  Thomas, I dreamed about you before I ever met you or knew you existed. I dreamed that one day I would have a dear friend with whom I could walk and talk, meet for coffee, exchange e-mails, share secrets, have fun. I didn\'t want a partner, or a lover, or an affair. I wanted a friend. And then there you were, and you made me so incredibly happy. I felt so much joy in your company because it meant that my fondest, deepest wish had finally come true. . But who knew that it would so abruptly end?  I guess it ended four months ago, but it\'s taken me this long to realize that it wasn\'t a mood, it wasn\'t a phase, that it was permanent.  And now you have the nerve to say it wasn\'t personal.  How can shattering my heart  not  be personal? And without even so much as the slightest tender mercy of an explanation.  Do you think you are the only one who can feel pain, Thomas?   Yes, of course I know that you are dropping me because it\'s  easier  to drop me than to fight with Nancy about it.  And yes I know that it\'s symptomatic of a larger problem.  It has been so maddening to watch you over these past few months, gaining weight at an alarming rate, complaining about your terrible relationship; you must have been thinking if you put a ring on her finger, it would stop the fighting. If you dropped a dear friend, maybe  that  would stop the fighting. If you dropped all your hobbies and lived only for her, maybe  that  would stop the fighting. And you must now think, If you buy a house, that will stop the fighting. If you marry and have children, that will stop the fighting. You would do anything to stop the fighting, but nothing works, does it, Thomas? In fact, it gets worse. And you resent it more because nothing you have given up gives you the payback you want. You have chosen, stubbornly, to live this life, which can have no end except for lonlieness and destruction. It has already begun, and it began with me. Or perhaps earlier, I don\'t know. Except that I do know you have isolated yourself from everything that does not meet her approval, which is most everything you ever enjoyed.  And because I love you so much, it kills me to know that this is the fate to which you have sealed yourself.  To be terrified to step out of her view for more than two minutes, clutching your cellphone for dear life, waiting for her to call you to tell you what you must do next.  Such a brilliant mind, Thomas, such a loving and dear heart, completely controlled and dominated, completely without independent thought. Or if you do still have them, you must surpress them, for fear they might cause another fight.  But you know what? I would accept you this way. I would silently watch you become alarmingly heavier and heavier, I would dance at your wedding,  I would continue to love you and support you and listen to you anyway, because that\'s what good friends do.  I would continue to be the one beacon in your life, but you won\'t let me. I thought you had the ability to recognize a true and stable friend. But your ability to recognize anything has become so clouded with the desperation to end the fighting by doing anything at all, no matter the cost, except the only sensible thing you  can  do, that you don\'t see loyalty, you don\'t see heartbreak, you don\'t accept love or caring.  You don\'t even have a sense of your own self worth. Once you asked me plaintively, \"Do you think I am that weak? Do you think I am that low?\"  The answer is yes to the weak, no to the low. Because, my dear Dr. Thomas Loecherbach, I will love you until the day I die. I will always be your friend. I do not allow anyone to say anything negative about you in my presence, even when they are outraged at you, or are simply trying to make me feel good by painting you in a negative light. I don\'t let them do it. I think you are a wonderful man, and it has been a privilege to know you. I will never desert you, even though you will not allow me in your life. I am resolved to love a ghost. Your fate is a life of abject misery, a life you have chosen for yourself. Mine is to love a man who will soon be a distant memory.  And this started out as a dream coming true. How bloody ironic.  I miss you, dear Thomas. '),(8642,2726395,'01,March,2004','2004-03-01 00:00:00','    Sadness. Traurigkeit.  It has been a few days since my last post, but since there is no one chomping at the bit to read my whining I suppose deadlines hardly matter.  I am home alone today, which is about the worst place I could possibly be, because \"alone\" and I don\'t always get along.    I got the job with Force, which is good news, provided we don\'t all get laid off in three months when Solectron sells it. I start on the 15th. I gave notice last Tuesday afternoon.  By Wednesday morning, a flash flood had come and destroyed most of Adax.  Now some would say this is some kind of karma happening, some physicalization of a clean sweep, a washing away of pain and frustration, and, in the literary sense, I suppose it is.  But in the physical sense it is very different.  I only wanted to leave my company. I didn\'t want to see it destroyed. I didn\'t want to see my co-workers walking around in a daze as years of work floated in the lake. I didn\'t want to see it torn apart; I didn\'t even want to see the horrible crazy owner sitting, staring into space because two days after he buried his father in an Adax teeshirt, Adax was under a foot and a half of water. Nobody really wishes harm.  But see all those things I did. And it makes me sad.  I\'ve been off of work since I made it out of the flood on Wednesday morning, spending money as though I had no budget, and tomorrow I\'m supposed to go back to work. I can\'t imagine why. My furniture is not usable, my files are pulp. I\'m not sure why I\'m wanted there, except maybe to remind me of the karma.  But surely there can be no karma in my estranged friendship with Thomas. I do not believe I did anything that would actually make me deserve the pain I\'m in now, and I\'ve been doing some serious self-examination here.  I\'ve done nothing but provide that man friendship. I\'ve done nothing to cause his fiancee harm, though the only reason I don\'t like her is because she rejected any possibility that I might be a friendly sort.  But here it is.  How does one ever know if a friendship is real?  At what point do they say, \"Okay, this person isn\'t going to suddenly change personalities and decide I am no longer worth anything in their lives. This person isn\'t after anything, this person isn\'t lying to me, doesn\'t want to cheat me, and only wants to share my life as a friend\"?  And then, after you\'ve reached the point of acceptance, that they\'re not going to do any of those things, what happens when they do?  Thomas is there, and then he\'s not. And sometimes he\'s there and he\'s not at the same time.  Thomas is cold and can inspire great intimidation. Thomas is a warm, wonderful and loving man who earned a piece of my heart with his quiet dedication and daily e-mails.  Thomas opened up to me. Thomas damned me for knowing about him. Thomas won\'t answer my e-mails unless they\'re school related. Thomas invites me to coffee and encourages me to keep writing e-mails that aren\'t school related.  Thomas comes with me to see Mark\'s play and even comes to our party. Thomas pretends past conversations never took place, Thomas won\'t discuss the play.  Thomas has no time to get together, Thomas goes to political events twice a week. Thomas has time for a walk in the afternoon. Thomas forgets to tell me that he doesn\'t have to take a day off without pay every two weeks now.  Thomas is friendly and nice. Thomas is cold and heartless.  All at once, sometimes sequentially, and all of it new.  Thomas marked me down.   The e-mails that say how much I mean to him--are they null and void?  Why should I put up with this? Thomas asks me to be patient. I wonder what I am being patient for, exactly. But I can deny what was once my dear friend Thomas Loecherbach nothing.  Once you\'ve determined that this person is a real friend who would never kill you slowly with little tiny stabs, then bandage them, then stab you again, you start to love them. You find a safe place in them. And when the knife pokes start coming, you don\'t believe it\'s real. You don\'t believe it will last. You don\'t understand what happened. And you think if you can stick it out, if you can survive this, you\'ll have your friend back.  I wonder how long one should believe this?    My fingers so automatically type \"Thomas\" because they have every day for two years, that they like the familiarity of the sequence. I loved the familiarity of the friendship. And I want it back. Is this my karma for something else? Do I deserve this? '),(8643,2726395,'23,April,2004','2004-04-23 00:00:00','    I think it\'s just about done. I don\'t understand it. But as I am waiting to find out if Thomas is about to go into a real estate deal he doesn\'t want, or if he\'s just been swindled out of $10K by a crooked real estate agent, i just discovered something quite by accident (well, the search wasn\'t by accident, but actually finding something was) that absolutely floored me.  Nancy--Thomas\' fiancee, who has told him she has three engagement rings prior--was divorced in Connecticutt in 2001. She was already involved with Thomas at the time.  Back in the days when Thomas used to talk to me, he only spoke of her previous engagements, and his reticence to become a fourth. He never mentioned a previous marraige, even when I asked him. That means he was lying to me back then (for what purpose?) or...he doesn\'t know.  More likely he does know and didn\'t want to tell a married woman that he had been involved with a married woman.  Perhaps he even caused the divorce.  That sure answers a lot of questions, doesn\'t it?  Total openness one day and the angry secrecy the next. Or even now within minutes.  I haven\'t seen Thomas now for three weeks. In two weeks, I may never see him again. He won\'t meet with me. And I cannot be president under those circumstances.  And he won\'t discuss it.  Of course, he is involved in what is sounding more and more like a shady real estate deal right now, which is a lot of pressure, I know, plus his car broke down earlier in the week and he\'s relying on Nancy for transportation, and he\'s been positively ghastly to deal with, so now is not the time to bring it up, so I stay silent. But now I know why he is buying the house (if he bought it) under his own name. Because he knows he is marrying a woman who cheated on her husband. Probably separated, though. Because he knows that the relationship is bad to begin with. Because he knows that 90% of all second marraiges fail. My heart aches for him.  It hurts, to lose this friend. It\'s an ache that won\'t go away. And now to know the truth--and to keep it silent--but still not know what any of it has to do with me, is agonizing. I know he is no saint. I know what he\'s done now. But I love him all the same. He rejects my offers of help on one hand and then turns to me for it with the other.   I really love this man, even though he has been nothing but horrible to me for almost six months now. He was wonderful. He earned my love.  Then inexplicably turned against me. With no feeling. With no mercy. Why should I love someone like that?  And worse--now I have to be president of the German School. and work closely with Thomas, which will be quite a trick for someone who won\'t even speak to me, who is someone who once pretended to love me for reasons that are also unknown to me.  Good God, Thomas. Talk to me. Have some mercy. I am not owed an explanation, but I am worthy of one. I have been a good friend to you. Why are you doing this to me? '),(8644,2726395,'05,April,2004','2004-04-05 00:00:00','    Dear Thomas,  I wonder how you are this morning.  I wonder why you called me twice on Saturday. What did you want? I did not call you back. I didn\'t want to get upset when I hoped you were calling to ask me to coffee after the teachers\' meeting and it would turn out to be some inane question about the school or to ask me what I thought of today\'s class.  What do you care what I thought of today\'s class. If you valued my thoughts you would still be my friend.   You know what? I have spent a lot of time lying on my face these past two days, because of a back injury. And it occurred to me in that time, that all you ever needed to do was say six words to me that would have lessened my suffering. \"Pam, I\'m sorry I hurt you.\"  But you never did. You never even acknowledged the possibility that you may have hurt me.   And then it occurred to me that you might not even realize that you broke my heart. You will not even allow me ten minutes to talk.  But you continue to send German lessons for free. And you called me twice on Saturday. Why? Do you enjoy repeatedly stabbing me? Either be my friend or leave me alone.   I intend to put a stop to the German lessons. You  will not accept my money and you don\'t want my friendship, and the last thing I want is to be in debt to a former friend.  I\'d rather just pay someone.  Besides, I don\'t want you to teach me German unless we are also friends. And we are not.  Or rather you are not.  If you have any original thoughts left, Thomas, if you are still even remotely capable of thinking for yourself, please look beyond your own nose for a moment and see that you hurt me.  See that you sent me over 5000 e-mails in the past two years, telling me how much I meant to you, gaining my trust, gaining my husband\'s trust, and that we both believed you were a true friend to me.  And then you suddenly dropped me without so much as a word of explanation or warning, and then, as if that wasn\'t painful enough, had to drive the knife in me one more time by telling me it wasn\'t personal.  Oh, so it was a business decision. You dismissed me like a servant. That\'s comforting.  Thomas, for God\'s sake, open your eyes. You broke my heart. You earned my trust and then you abused it. You are capable of doing such things. Other people besides you can have feelings.  I remember how you said, talking about a girlfriend you had, whom you really loved, who left you without explanation or warning, leaving you devastated for years, \"No one could know my pain.\"  Well, Thomas, you\'re wrong. Someone can. '),(8645,2726395,'02,April,2004','2004-04-02 00:00:00','    Dear Thomas,  Why do I always ask your permission now to do things? Why do I apologize for mentioning that I have injured my back? Why do I ask your permission to talk to you? The answer is simple, and I would tell you, if you\'d only let me. So I will tell you here.  Thomas, I have never in my life seen such profound changes to a person in such a short time as I have seen with you over these past few months. These changes have affected me, whether you meant them to or not.  I am simply trying to figure out how or if I fit into your life with these changes in you that have made you almost unrecognizable to me.  Almost, but alas, not enough. You are not a stranger to me, though I wish you were. You seem aware of the changes yourself, but seem either blind or indifferent to the effect they have on the people who love you. I know your mirror does not lie to you about the physical changes in you--surely you cannot have missed the extra fifty or so pounds you\'ve put on.  You are terribly, profoundly unhappy. You are despairing. And my broken heart goes out to you. I can only imagine what your parents are going through.  I know why these changes have happened. It is simple. You are swirling, circling the drain, as the paramedics so charmingly called it. You are in the unenviable position of being madly, passionately and obsessively in love with a woman you cannot make happy. Nothing you do makes her happy. As I wrote before, you have given up everything and everyone you ever cared about, in the logical thought that if you removed any potential causes of fights,the fights would stop, and then she\'d be happy. And then you\'d be happy. It is a logical thought path, except for the one fatal flaw--it doesn\'t work. For one thing, no one is responsible for anyone else\'s happiness, and for another, tying your own happiness into your ability to make someone else happy is a recipe for abject failure. So you are both desperately unhappy. I wouldn\'t take it personally, Thomas. Nothing you can do will ever make her happy. It\'s just the way she is, it has nothing whatsoever to do with you. After all, you aren\'t her fourth fiance for no reason.  If I could, I\'d give you advice. I would say, \"stop worrying about making her happy. You can\'t do it and neither can anyone else. Actually, we are all responsible for our own happiness. If marrying her and buying her a house and all that makes YOU happy, then do it. If dropping me as a friend makes YOU happy, do it. But if you give up the things and people you love in order to try to make someone else happy, you will fail, and then you will be even unhappier, because then you will be left with nothing.  So do what makes   yourself   happy, and let her do the same.   But, since you have decided to change the parameters of our friendship without informing me what the new parameters are, you leave me only two choices. The smarter choice would be to tell you to go to Hell, or rather, go home, because it\'s one and the same.  The idiotic choice is to see where and if I fit into these new parameters.  I am taking the idiot\'s path.    So I ask questions. Drop sonar. I have learned that it is okay to talk about German school with you. I have learned that it is okay to talk about home buying with you. I have learned that you want to continue to teach me German without pay,(but I think Nancy will one day soon convince you that I am using you.) I have learned that it is not okay to meet in person. It is not okay to even  ask  to meet in person. It is okay to discuss what is going on in my life, but it is not okay to discuss what is going on in yours.  It is okay to send each other some political stuff, but that I should not expect a conversation to arise from it.  I have learned that it is all right for you to put me down as an emergency contact should anything happen to you, but it is not all right for me to ask what I am supposed to do should (God forbid) such an emergency actually occur. This is the new world order, as I see it, from my annoying questions to you about what I can and cannot do now.  Friends do not put such one sided and bizarre restrictions on each other. So the conclusion is obvious. You are not my friend. But the problem is, I am still yours. And will always be. So I take the crumbs. As soon as I know which crumbs I am allowed to pick up. And the only way to find out is to ask. I am an  idiot,  you see. I don\'t know any better. I choose to put myself through this agony. Just as you choose to put yourself through yours. The difference is, I am choosing to do it because it is all I have left of you, and having you as part of my life once made me very happy. You are choosing to do it because it\'s all you have left in this horrific relationship of yours--the obsession to make someone else happy. '),(8646,2726395,'19,May,2004','2004-05-19 00:00:00','    What is with me, anyway? I just spent a week in Munich, and came home with a German flu that \"wasn\'t covered\" by the American flu shot, that is gripping me with the same ferocity with which Thomas grips me and won\'t let me go.  Or rather, I can\'t seem to untangle myself.  Everywhere in Germany that I went, Thomas was there. He was in every sign, every overheard conversation, every cup of coffee, every television show. I couldn\'t shake him. I wanted to shut out the sound of the German language from my ears, just to give me a break from every sound whispering \"Thomas, Thomas, Thomas.\"  And I know, deep in my heart, that things can never again be the same. Now I know the whole truth. Now I have that missing puzzle piece, my opinion of this man has totally changed. Now I understand him. Now I have my deliverance, of fully understanding what happened.    I understand now.  Once upon a time, Thomas loved an American woman who left him without a word of explanation or warning.  Heartbroken, Thomas shaved off his beard and moved to America and vowed that no one would ever hurt him again. And he\'s made an art of it: he has repeatedly gotten involved with women who aren\'t really available to him. Married, separated, obese, career-obsessed.  He\'s made them close, intimate friends, sometimes more than friends. Then he systematically dropped them without a word of explanation or warning, breaking their hearts...our hearts...before his could be broken again, doing to these women what was once done to him.  One in the series of unavailable women was Nancy. A married woman who must have shocked the hell out of him by suddenly becoming available and moving to California. That wasn\'t supposed to happen. She was supposed to get dropped with the rest. But she had a secret weapon or two. One, she saw what Thomas did. And she saw that the more these women in his life cried, the more they asked him to come back, the more they tried to understand, the more careless and distant he became.  But if these women told him to go to hell, to get out of their lives, that he was a selfish asshole, he became contrite and sweet and pliant. He begged them to come back.  And so Nancy began to threaten. Weekly. She was moving out. She would fight him. She would threaten to leave, to move back to Connecticutt, and he would give her whatever she wanted.  And plus, she had the goods on him--he was an adulterer. He would cheat with a married woman.    But now Nancy\'s game is up--she let it be known that she fears divorce above all things, and of course now we know why--because she\'s a divorcee herself.  So the minute Nancy marries Thomas, he will lose interest in her, because the threat is gone.  That is really sick. The whole situation, my part in it as one of the stupid women who really thought they had a friend, Thomas\' intolerable cruelty, Nancy\'s manipulation...it\'s horrible. And I just watched him do it to someone else.  One of the women who told him to go to hell was a woman named Melly, who ignored his phone calls and e-mails for close to two years, as he begged her to resume the friendship after he cut her off.  Finally after two years, she broke the silence to tell him that she was getting married. They spoke, and she invited him to her wedding, which is this coming Sunday.  She told him how much it would mean to her if he was there. And of course you know what he did. He refused her.  This is a horrible man, once cruelly treated and now repeating those actions done to him over and over.    I know it can never go back to the way it was, because it never WAS the way it was. I understand it all now. But it seemed so real, and I think it was, I think the feelings were genuine when they happened, but I couldn\'t see the inevitable coming.  And I miss it so much. It\'s hard to let go.   On Tuesday he was over, borrowing some car seats I promised him in February, but I was too sick to see him. I could have come out, but I really didn\'t want to. Instead I watched him through the window slats as he talked to Mark. They look so much alike. Conversing outside his car. Smalltalk. Both with other things on their minds.  And he grips me. From a distance. He is everywhere. He is in the trip I can no longer tell him about, he is in the phone that does not ring, he is in this language that he put in my head, he is in the stories, he is in the truth, he is everywhere. Is there no place I can go that he is not?  Do I want him gone? Or just the pain? The pain of being hurt, the pain of being stupid, the pain of trusting someone who was absolutely not to be trusted?  What do I want now?  I don\'t know. But as more time passes the less I feel for him and the more I feel for the past. A past that cannot continue into a future.  And I feel empty. Or, really, emptied. From giving love to someone who didn\'t deserve it. '),(8647,2726395,'29,June,2004','2004-06-29 00:00:00','    I have been overcome with a complete revulsion. The thought that I will see Thomas tonight..in my home..is too much to bear.  He has made his newfound hatred of me apparent, finding fault in everything I say, finding something to get offended about in the most innocuous of statements, purposely withholding any friendliness whatsoever and keeping his mail strictly school related.  I stopped sending him anything at all about what I am up to, because he never acknowledges it, and never says anything at all about himself, even if I ask such top secret, highly invasive questions like, \"So when do you get back from Istanbul?\"  It is clear, his transformation is complete. He hates me because Nancy hates me, and it is so much easier to let her think for him than to have his own opinion that might be different from hers. And Nancy hates me for no reason at all.  These aren\'t nice people. I guess I have come to this conclusion. I feel a little sick to my stomach at the realization, because I still have such warm and pleasant memories of a sweet and loving friend, of good conversations, quiet sunsets, and quiet devotion.  And it all evaporated, and only the memories remain.  Now there is this really ugly person left--judgmental, hateful, spiteful..and tonight I have to welcome him into my house, into my HOME, when all I want to do is forget he ever existed, to wipe those memories of two years of happiness.  The sad thing is, I don\'t hate him back. I don\'t like him very much right now, and probably will never like him again, but...I love the man. That just doesn\'t turn off. It would be so much easier if I could loathe and despise him. But I just don\'t.  Now if I could just figure a way to get him out of my life forever. But that would mean leaving the German School, and leaving a lot of people high and dry. Thomas would not think twice about doing it; in Nancy\'s Machiavellian thinking, it doesn\'t matter how much you hurt other people as long as you get what you want. And that\'s Thomas\' philosophy now, too, apparently.  But it isn\'t mine. I haven\'t changed a bit. I still hold to my principles, no matter how painful. I don\'t hate. I don\'t quit. And I believe that everyone gets what they deserve.  And I know what Thomas will get.  That would bring other people a delicious pleasure, knowing that his retribution is a life with a shrew he himself didn\'t want, forcibly cut off from having any close friends, bowing to her every will.  But it brings me only deep sadness and no pleasure at all. Schadenfreude is not in my character, I guess. '),(8648,2726395,'11,June,2004','2004-06-11 00:00:00','    Almost a month since my last post. And I am still gripped. I can\'t get away. Thomas isn\'t even in my life anymore. He\'s not the little voice on my shoulder, he\'s not my gentle companion watching summer sunsets, he\'s not in my e-mail Inbox, he\'s not moving in the same circles. Yet there he is. He isn\'t worth this constant melancholy. My time ran out and he coldly moved on, without care of my feelings, and having none of his own.  I thought we had worked it out about three weeks ago, but Nancy is really putting the squeeze on him. There was a dinner last Saturday we were all supposed to attend, but Nancy decided she didn\'t want to go, and Thomas was noticibly uncomfortable there alone.  But he goes to things alone all the time.  The reason Nancy didn\'t go is because I was going to be there. And Nancy has taken off the mask and now openly despises me. She is going to make it really difficult for Thomas to do anything that might involve me. And it is working. Since she finds fault in everything I do or say, there is no hope that that can ever change. And she must complain--because now all of a sudden Thomas is finding fault in everything I do or say, and now it\'s to the point where we can barely even talk to each other anymore. Now Thomas thinks of me the same way that Nancy does. I have no fight with this woman, I did nothing to her, but she destroyed the friendship by making me not worth the trouble to Thomas.  Does he truly feel nothing? Does he really feel no loss of me?  I feel his loss everywhere. I don\'t know how I am going to get through the school year, with him at my side as my partner in running the German School. My guess is that he will bolt very early. Maybe this will be merciful.  Even my passion for German is tainted, cooled, and I cannot let that happen. I was interested in German long before I met Thomas. But I am starting to care about it less and less.  In fact, part of my melancholy is in NOT having anything beyond family and job to care about. No goals, no ambition to reach, just day to day living.  I would almost rather fight with Thomas than have him be absent from my life.  But I don\'t get that choice.  Traveled to Germany last month, am going to Chicago in a week or so for Supercomm. Maybe he is not in Chicago. \"Why did I have to meet you, love you, why can\'t I rid you from my mind?\"... Was the two years of real bliss worth these last six months of agony?  At least I have the memories, the e-mails. But I guess the answer is no, it wasn\'t. And that\'s really sad.  If you\'re out there, Thomas, and you read this somehow, (I don\'t know how you can, unless this somehow turns up in a Google search attached to your name, or my name, somewhere, and right now, you don\'t turn up there) you have to know that you walked away in disgust from someone who really loved you. Perhaps that was the biggest mistake I ever made in my life, trusting you, loving you. Believing you when you said how important I was to you. I heard your tales of all these lost female friends of yours, all of whom seemed to suddenly become demanding of you after a couple of years and then finally, irrationally, decided to drop your friendship. I knew even then that it was a pattern, that the one common denominator in all these women who suddenly became irrational and demanding was YOU, but I was so arrogant, I thought that I would be different. I was just stupid and naieve, what did I know of men like you, who do such things to people, maybe even not aware of doing it? But I so believed in you, in an improbable but friendship that brought me happiness, I was blinded to the obvious. '),(8649,2726395,'29,July,2004','2004-07-29 00:00:00','    Now this is stupid. This week has been so strange so far. It started with my caterer at V urlLink ienna Bakery  telling me he couldn\'t understand my secrecy about the event, it\'s all over the Internet that there is to be  urlLink an acquisition of Force by Motorola.  I told him, \"look, it\'s okay if the Internet tells you, but if I tell you, I\'m sharing a jail cell with Martha Stewart!\"  Next thing you know I\'m buying platters and doilies and balloon arches and renting big platters and samovars and pushing really hard to get checks written (Motorola is on Oracle and Force is on SAP and they don\'t have compatible cost center numbers, but everyone is really good at passing the buck around here.) This is especially easy since we discovered that at the moment, nobody actually has a boss.  Meanwhile my sister Nancy, who caused so much grief over the years to my family, keeps a blog here too. I follow it once in awhile because it is usually so damned funny. She may be a heartless bitch when it comes to the family (she refuses to meet Jacob, for example, because she \"doesn\'t feel ready to be an Aunt\") but she\'s a good writer. Her entry introduced me to one of the  urlLink funniest websites  I\'ve seen in years, and I was up half the night laughing until I cried. It\'s not politically correct, but sometimes you just have to have a sense of humor. (Look, I drive a Prius, I recycle, I say \"African American\" even if I am not totally sure the descendant of African to whom I am referring is actually American or not, I vote Democratic, I was moved to tears by Barak Obama\'s speech at the Convention on Tuesday night, so how DARE anyone call me politically incorrect because I find the butchering of the English language on foreign-made products hysterically funny. I\'m not berating anyone, I just think the end product is funny. And, by the way, while I am on this rant, using a pronoun to describe a group of people is GRAMATICALLY correct, it has nothing to do with POLITICAL correctness. They, they they.)  There, I feel better about that. I guess I\'m really stressed about,well, the concept that I have to put on a tremendous party for the people who will probably lay me off when it\'s done, no matter what a good job I do. It\' s a little bit like having to arrange my own funeral because there\'s nobody in the family competent enough to do it. The weird thing, is, I want to do a really good job. I want to knock their socks off. I want to stand there serenely in a beige suit with a painted silk scarf and have made it look like I can do the equivalent to a wedding for 100 guests in a week all the time. They\'ll never see my callouses, or, with luck, they\'ll never see my insomniac stress ravings on my blog, maybe my eye won\'t twitch. Maybe I should try to get some sleep. '),(8650,2726395,'26,July,2004','2004-07-26 00:00:00','    Insomnia.  It\'s the middle of the night. That\'s not even the worst part. I am having a stress and panic attack.&nbsp; I am in the middle of preparing for what is surely the largest and most important event that I am personally in charge of in my life--the events surrounding Force\'s acquisition by Motorola--and I have this now.&nbsp; Finally someone is giving me oodles of money and telling me, \"you don\'t even work for us yet, but here\'s a few thousand bucks, go out and create one hell of a rollicking good time for 100 people. You have 48 hours. GO.\" And unsaid, \"we\'re watching. If you do a good job, maybe you\'ll get to keep your job. And then again, maybe not.\"  I feel like I\'m on \"The Apprentice.\"&nbsp; You never saw  their  midnight panic attacks.&nbsp; Maybe it was just me.  I\'ve been reading. 3 of the 4 Dan Brown books, even though I realized halfway through the second one that all the books are exactly alike; Nancy Drew for adult conspiracy theorists. So I thought I\'d try something more imaginative--I never did read the Harry Potter series, and I thought it was about time.  But the monkey on my back won\'t go away. Thomas seems almost like a fictional character to me now, someone I just read about in a book, a character that you really thought for most of the book was a really good guy but turned out to be not so wonderful in the end. A character that you read about and think, \'I never would want to meet that guy, he\'s too dangerous, a potent mixture of sweetness and vulnerability, intellect and drone, selfish bastard and unfeeling cad wrapped in a delicious sugar coat.\'&nbsp; But he will soon be back. He will soon be able to hurt me again. I have to somehow cut off that possibility. I have to find the strength within myself not to ever fall into his trap again, to keep hating him and his evil girlfriend/fiancee/wife/whatever, because if I ever stop, he can hurt me really badly.&nbsp; And I think the only way I can do that is to get away from the German school. And there is really no way out.   No wonder I am having a panic attack. I am a mouse in a maze, and I cannot see which path will lead me away from the traps.&nbsp; Do a good job, maybe get laid off anyway. Take the presidency, Thomas can hurt me. Don\'t take the presidency, I\'m no better than he is, a promise-breaker, a committmentphobe, and, worst, an outcast there.&nbsp;   I feel my super burrito from La Pinata sitting on my stomach like a rock.&nbsp; I  so  need advice. '),(8651,2726395,'20,July,2004','2004-07-20 00:00:00','    The original title for this blog, way back in January, was \"Rosen Thorns,\" about the thorns in this Rosen\'s side.&nbsp; It seems this whole year, now it\'s more than half over, has been about one particular thorn.&nbsp; I have cried almost every night about it. I have been to therapy about it. I tried to travel away from it. Massage it out of my system. Work it out. Hate it out. But nothing is working, and now I am getting angry.&nbsp; I have always been resilient. I got over losing my sisters, my parents\' idiocy, a car accident, becoming president of the German School. I have gotten over losing friends to death, to childbirth (Pam has a kid? Woops, she can\'t hang out with  us  anymore!) to moving away.&nbsp; But I cannot seem to get over Thomas Loecherbach. I have been taken down by a mild-mannered German engineer who wears short sleeve dress shirts on purpose and actually uses the front pocket.&nbsp;  &nbsp; I try to be philisophical about it. I mean, look, it was bound to happen.&nbsp; I knew it was over that night in the Spanish restaurant.&nbsp; When Nancy showed off her engagement ring, and I thought about how three weeks earlier Thomas was telling me he was coming to the end of his rope and was seriously considering ending it, I knew that someone I loved very much had either just made the biggest mistake of his life, or, someone I loved very much had spent the better part of two years lying to me.&nbsp; Either way, I knew I was no longer of value to Thomas.&nbsp;Or maybe I never was.&nbsp;And to find out about it that way wounded me in a way that is taking a very long time to heal. &nbsp; I wish I could explain all this to him. But you know, what\'s the point. I keep telling myself, as I repeat these imaginary conversations,  he doesn\'t care about you. He wouldn\'t understand anyway. He already hates everything you haven\'t even said yet, what makes you think he\'d listen to you now? Pam, he doesn\'t care about you.    &nbsp; If I were a different sort of person, I might be vengeful. I might break my promise of secrecy. I might let him know that I know Nancy was married when she got involved with Thomas. Hell, when I think about that now, it\'s so obvious.&nbsp; But no, no matter how many lies I\'ve been told, no matter what promises have been forgotten, no matter how shabbily he treats someone who genuinely cares for him, I will not betray a trust.&nbsp;  &nbsp; Now I just want this thorn to go away. He\'s in Istanbul right now and I am still haunted. He\'s everywhere.&nbsp; In September I am going to have to see him every week. Right now I can work hard and teach myself to hate him, tell myself he\'s a lying, self-centered adulterer, but as soon as I see him, it melts away. Then I have to work up building those defenses again.&nbsp; I really, really just want the thorn in my side to go away forever, so I never ever have to think about him anymore.&nbsp; I will love him forever. But I just don\'t want to think about it. '),(8652,2726395,'08,August,2004','2004-08-08 00:00:00','    Well, I  guess  it was a success. I thought everything looked really pretty, and I achieved  my  goals. Whether or not anyone else cared less is debatable.  You know, that\'s the thing about me--I like feedback. I like to to hear some applause if I do a good job, I like to be criticized if I\'m off-base. At least I know where I stand.  On Monday everyone sort of moved like sheep, not really sure if the message was \"Welcome to Motorola, you\'re fired, go eat a Danish\" or not.  People really were as glum as the day.  I guess I can\'t blame them. At least for the week that I was putting the party together,  I was able to push it to the back of my mind that Motorola doesn\'t really need me much after the end of the month.  Others probably aren\'t so lucky.  So there was no feedback at all, not a thanks, not a \'good job,\' not a \'that really sucked.\' Nothing.  But at least I feel in my mind I did a good job, and I have pictures to prove it.  On Wednesday, Thomas called me. You could have knocked me over with a feather. We talked for two and a half hours. (Yeah, Nancy\'s in Ecuador.) It was our first \"normal\" conversation in the year 2004, and it almost made me think I have a friend in him again. But then, I can\'t let myself think that, or I will be setting myself up to be hurt again.  After all, I still don\'t know if he is married, and I do know that he would not have called me at all if Nancy\'s watchful eye was not two time zones and three thousand miles away.  I must be wary, I cannot get my hopes up, because they WILL get dashed. I have to work on having no expectations.  I have been interviewing for jobs, just in case. I had one last week, for a place called Immersion Technologies.  Their products were somewhere in the middle of a high wow factor (adding the sense of touch to the computer experience) and a \"My God, this is the stupidest thing I ever heard.\"  To top it off, I instantly hated the hiring manager.  I mean at first glance.  She was this stuck-up, standard-Macy\'s-career-blouse-issued woman with a turned up nose and permanently pursed lips who never smiled and who kept insisting (without a smile) that her job was to help make the computing experience  fun .  I wanted to ask so desperately, \"Okay, so now your phone is on buzz mode, and instead of getting one single buzz when it rings, you get two pulses from someone you know. So now you know who you can\'t talk to at the moment. This is fun....how?\"  When she went into virtual slapping matches and tactile emoticons (a heartbeat pulse for \"I love you\") I almost ended the interview, but when she asked me if I had a typical German personality, that was it. I would not work there if it was the last place on earth. It has \"dotcom\" written all over it. It reminded me of interviews in 1999 where everyone kept asking me, \"Have you seen our game room?\" Which of course translated, in 1999 speak, into \"You are never leaving this building.\"  Thank you, NEXT! '),(8653,4055499,'01,August,2004','2004-08-01 00:00:00','    I was noticing today how whenever there is any fuss, hummingbirds just rise up and float away, like that sweet nectar is nothing to them next to the uncomplicated virtue of silence, like theyâ€™d rather just be alone in the world without the mess of noise, other than the clapping of their own small wings, waiting patiently for their next chance at sweetness.  This must be why theyâ€™re so small.  Today I was walking around the city, trying to find the law library.  In Indiana the buildings are so close together and situated on these very neat and even planes with each other.  There arenâ€™t people on the streets or sidewalks that arenâ€™t students, visitors, or professors of the University, but here there are thousands of people going about their every day lives, totally separate from students.  It occurred to me that I had been on the same street as the law library so many times before.  Last winter my mother and I went to see a show at The Playhouse Square, which happens to be right next to the law library.  We were running through this thick December rain to try to find a restaurant to eat at before the show and we accidentally witnessed the lighting of the Christmas tree.  We stood there for a moment, pretending like we had meant to see it, like we planned it and it was special even though it was something that we would never really do. Still, it was the surprise of it all, the mystery of being caught in an unintended moment.  Not knowing something was there and then becoming a part of the history of it.  So, seeing the law library was a surprise, but I didnâ€™t feel anything that magical.  In fact, with everyone walking around wearing business suits, or tracksuits, or baseball caps looking for the early opening hot dog vender, I felt pretty unremarkable.  This must be a familiar feeling to many students downtown, because it seems like there are extra steps taken to make students feel special in their education.  For example today I received my university ID card.  The man at the desk told me that I could do â€œanythingâ€ with the card.  Anything? I thought, but didnâ€™t say it because I thought I would hurt his feelings.  But now I have this green card in my wallet.  It is in the third position behind my Driverâ€™s License and my debit card, so maybe it is special.  But then, when I was walking outside back to my car, I saw a group of about ten students who attended the College of Court Reporting.  They all had these great book bags with this golden shimmering seal that said â€œCollege of Court Reporting.â€  At that moment I kind of wished that I were going there, so that I could get the bag.  But instead of making a fuss, I just rose up and floated away.   '),(8654,4055499,'01,August,2004','2004-08-01 00:00:00','    when i was running this morning i was thinking how odd it is to pass by houses as people are just starting the day.  everyone is so locked into the becoming of the day, the routine of it.  i feel like such an intruder on their lives.  for example this morning there was a woman wearing a blueberry nightgown holding a long yellow rope which ended on a collar of a golden retriever.  she just looked at me as if i were a spy and i smiled letting her know that i didn\'t care that her dog was barking at me, that if my dog was there he\'d bark at her too.  after that i got to a man who was walking out to the mailbox to fetch his newspaper.  it was almost eight in the morning and everyone was speeding down Lake road to get to work on time.  still, he didn\'t mind that he only had his boxers on.  he was that confident, that comfortable i thought.  but when i got closer, he turned around and pretended to look at the lake, at the storm coming in.  it was such an awkward thing to do, and he knew it.  but he wasn\'t ready for me to say \"goodmorning\" or to catch a wave coming from me, so he did it.  it wasn\'t part of his routine.  anyway maybe i don\'t feel like a spy as much as i feel guilty for feeling so mellow in those moments.  if it were the other way around i would probably turn to the lake as well.  maybe that\'s why some people are runners.  maybe that\'s why i am.   '),(8655,4055499,'01,August,2004','2004-08-01 00:00:00','    and that blog was a beautiful blog&nbsp;     '),(8656,4055499,'01,August,2004','2004-08-01 00:00:00','    I have been away from Ohio for six years and finally moved back this summer.  Every day that I was away, I longingly remembered one thing about this little part of the world.  Itâ€™s not something that inspires the illustrious painters of nature or the romantic bards that croon of the greatest of lakes.  Itâ€™s the gray, cloud-covered waters; the sky, whispering a secret of rain, like at any moment, real silver may fall.  Today was the first day that I found the weather again and I am so happy for it.  August and mysteriously cool, the temperature low enough for me to wear a tee shirt and a fleece.  But itâ€™s the water that really gets under my skin something wickedly happy.  There could be nothing and everything out there and I would never know.  And I donâ€™t need too.  The clouds are so low that the little white caps forming in the coming storm from Canada look like they might just grab them and pull them into some secret chamber.  And the sky between the clouds is a permanent muted pink, even though the sun is no nearer to setting.  All of the kingdom of things feels this weather I think, because I can hear them buzzing violently outside in their bushes and stalks, stems and blades.  All this while, the fattest lake gull I have ever seen sits fixed on the top of the dock light.  Every minute or so he barks at the waves, daring them to knock him from his perch, spit him on his back.  Itâ€™s been over three hours and heâ€™s still there.  Heâ€™s an Ohio bird for sure.  He wonâ€™t move away for anything.   '),(8657,4055499,'01,August,2004','2004-08-01 00:00:00','    My house sounds like a broken cat.  I know it looks like I meant to write â€˜carâ€™ and then fell victim to a typo, but I meant cat.  There is a strange noise that is always just outside of my dining room window.  I canâ€™t figure it out and it grates at me all day long while Iâ€™m writing.  Sometimes I think it is the ghost of dead wood.  This house is old.  Built in the 20s when Ohio was full of industry from the lake and coal from the electric rail cars.  But something is purring outside, a miserable sigh.  Like it wants me to find something out about this place that I didnâ€™t know about before today, or the day before, or a week ago when I first heard it.  The crickets have started now, their dress rehearsal for evening.  The broken cat has left, for now.  Perhaps it is soothed by the metronomic high pitched thud of the crickets.  Maybe not.  Today when I was on my run, I was nearly assassinated by a wasp.  I must have looked crazed, running in the bicycle lane, swatting at my back with my hands.  And the faster that I ran, the more quickly my swats swirled around my head, and I eventually tangled myself completely in my earphone wires.  But I didnâ€™t slow down.  The wasp followed me and I didnâ€™t slow down, I just ignored her.  After about three minutes, I was far enough away from the scent of her hive and she went back to her flowers where she felt safe.  I am so glad that Fall is coming with all of its teeth.  The crickets have stopped and whatever swinging broken thing is back.  I should stop writing now and move away from the window, so I can ignore the sound for another day.   '),(8658,4055499,'01,August,2004','2004-08-01 00:00:00','    Today the lake is high.  The waves come right over the pier and slam into the neighborâ€™s patio with such lonely vigor â€“ because theyâ€™ve been cheated out of their right to curl and fall on land rather than cement.  Today I saw a woman digging in the ground.  From far away it looked like she was making a giant gray cake, mixing something and then smoothing the tops with her large shovel.  But when I got closer to her I realized that she was actually evening what she dug so that she could lay a sidewalk.  I thought to myself, why canâ€™t I lay a sidewalk?  because I couldnâ€™t.  I donâ€™t have the patience for that kind of evenness.  And anyway I would rather there be a little hill or two in it, which wouldnâ€™t make anyone happy but me and some kid who just got his first skateboard.  Maybe weâ€™d be enough.   '),(8659,4055499,'01,August,2004','2004-08-01 00:00:00','    After six last night, the sun was still unrelenting.  It was insistent, remorseless heat slapping my left cheek, as I, equally unwavering, sat atop the breaker wall and watched the sun sink like silk to the West.  There were two pelicans floating in the water.  They let the waves play with their little bodies in a way that made them look wooden.  And then, they just flew away.  A white gull came and they leapt out of the water, their long wings like robes, vestments of conviction, following her in wide circles like a Queen.  I couldnâ€™t help but smile to watch them in the water.  They seemed wholly, absolutely amazed that they were able float.  I wish I could have stayed out there and watched them disappear into the broken light of the West, but the sun was more patient that I, and the heat was getting me.  The sound of the coughing water, the lake doveâ€™s high-toned screams, the small bumblebees making their love, all of the smells of what was alive and dead or alive again in the lake.  These things were small enough for me to find amazing.  After a long night of dreaming of them, I woke to the Lake barking under the kick of lightning.  Clouds, waves, rains.  But no pelicans.  And no sun.   '),(8660,4055499,'01,August,2004','2004-08-01 00:00:00','    i\'ve been slipping back into my dreams in the morning when the road shudders underneath my feet.  even when i\'m running and listening to some nonsense from idiot disc jockeys corporate radio hires to ensure the image of young twenty somethings as dumb fucks with barely an appreciation for a beat to back up the one coming from their chests.  even then.  and the dreams come and suddenly that song \"lucky man\" from the verve comes too and the river of air is everywhere.  even in the heat.  even there on that gravel road that is slowly destroying my knees.  because i\'ve been thinking about how i say goodbye.  it\'s never more appropriate to muse over such a topic than when i\'m running.  or maybe it\'s because the light is just coming and despite the traffic pushing me off the road, i only hear the lake.  recently i\'ve been dreaming of past loves, lost friends.  and there is always something birdlike about them.  and when i say birdlike i mean that there is something mythical.  one person was a giant, another had the strength of athena.  and they were all coming back into my life to say hello.  people i had all run from.  those that received a sudden goodbye.  and i\'ve created this odyssey but without any heroes.  and whiles i\'m running i start feeling so full of fire, so full of fatal flaws.  still i keep running because all goodbyes should be sudden, i think.  or maybe i just run because it\'s all i know.   '),(8661,3454992,'30,May,2004','2004-05-30 00:00:00','   Over on Frank Smith&#8217;s blog I found  urlLink this post  on Peter Laughner&#8217;s review of Lou Reed&#8217;s 1976 album &#8220;Coney Island Baby&#8221;. I agree with Frank that the review is, perhaps, &#8220;the   greatest piece of rock criticism ever written&#8221;, but, unlike Frank, Coney Island Baby is   not   one of my favourite Lou albums. I did, however, recently purchase on eBay a magazine advertisement from the launch of the album, but only because I think the title track is one of Lou&#8217;s best songs which is, as  urlLink Mark Deming from AllMusic  has said, &#8220;a song about love and regret that was as sincere and heart-tugging as anything the man has ever recorded&#8221;. The last line of the song and album, &quot;You know I\'d give the whole thing up for you,&quot; grabs me by the short and curlies every time.              &nbsp;         Back to Laughner&#8217;s review. He suggests at the end, that if Lou was serious about the last line, he should do exactly that and stop making records. I&#8217;m personally glad he didn&#8217;t, as much of my inspiration today comes from Lou&#8217;s songs and his approach to his craft in the last 40 years. Perhaps, although I&#8217;m sure Lou would deny it, reviews like this one and the catalogue of reviews from Lester Bangs, may have fired Lou to keep going when many of his contemporaries either died, burned out, or started writing pop songs? Although I find much of his work hit and miss, when he hits, he really hits. It&#8217;s worth the wait. I would love to write songs with that honesty, that sense of poetry.'),(8662,3454992,'30,May,2004','2004-05-30 00:00:00','   For those of you paying attention (and I know that is a limited audience at this juncture), this is my fourth (or fifth? Even I&#8217;ve lost  interest  count) blog URL in two years. The reason why I moved from  urlLink http://blogs.msdn.com/cameronreilly  will become apparent in future posts. In the meantime, please be patient with my holding pattern&#8230;'),(8663,3454992,'02,June,2004','2004-06-02 00:00:00','   v\\:* {behavior:url(#default#VML);} o\\:* {behavior:url(#default#VML);} w\\:* {behavior:url(#default#VML);} shape {behavior:url(#default#VML);}          st1\\:*{behavior:url(#default#ieooui) }                           &nbsp;         What Buzz thinks about Skype. I was the friend on the other end of the line.     J               &nbsp;'),(8664,3454992,'02,June,2004','2004-06-02 00:00:00','   v\\:* {behavior:url(#default#VML);} o\\:* {behavior:url(#default#VML);} w\\:* {behavior:url(#default#VML);} shape {behavior:url(#default#VML);}                st1\\:*{behavior:url(#default#ieooui) }st2\\:*{behavior:url(#default#ieooui) }                          I tried to buy this new   Seth   Godin   book from Amazon, but it isn&#8217;t available formatted for Microsoft Reader. I hate that.            &nbsp;'),(8665,3454992,'02,June,2004','2004-06-02 00:00:00','   v\\:* {behavior:url(#default#VML);} o\\:* {behavior:url(#default#VML);} w\\:* {behavior:url(#default#VML);} shape {behavior:url(#default#VML);}                           &nbsp;          &nbsp;'),(8666,3962812,'31,July,2004','2004-07-31 00:00:00','   I probably should have included this in my last entry, but I completly forgot. Now, it seems a big story has arisen in the news later. A 17 yr old, bludgeons a 14 yr old to death with a hammer. Now, what could possibly be the reason for this? Apparently it\'s because he played a video game. That\'s right folks, peoples stupidity has finally reached an all-time high, I seriously fucking can\'t believe these idiots are trying to pin somebodies death COMPLETLY on a video game.  I could possibly understand if he was slightly influenced by a violent game, but I do not think a video game MADE HIM KILL SOMEBODY. As you can tell, I feel very strongly on this matter and think that the victims parents are fucking cretins. They\'re looking for somebody to blame over their son\'s death and aren\'t looking at the cause.  I mean, there\'s much worse movies, that show people killing people etc, surely all of these are responsible for killing people as well? Surely all these gorey films make people think \'I KNOW LETS GO KILL SOMEBODY!\', Seriously fuck off if you believe this. People have the right to make their own decisions and aren\'t forced to do anything. A slight influence may have come from the game, but a game does not make somebody want to kill someone.   Obviously the 17 yr old has mental problems and it seems the victims family want to overlook these and blame a video game. How suiting that the prosthetic world we live in want to blame some of the prosthetic goods we place all of our values upon.  I seriously think the world would be a better place if 99% of the people were dead. There is far too much idiocy in the world today.  Peace  xxxx '),(8667,3962812,'23,July,2004','2004-07-23 00:00:00','   Yep I\'m back, so here we go again.  I\'ve heard NOTHING from Angela since the rose incident. Disappointing, disheartening and depressing...wow I should start writing, wait...I do...and that\'s shit too.  So yeah, enter mega \'can not be arsed with life\' mood. Because right now, guess what? That\'s how I feel. I don\'t understand why Angela aint e-mailed me or txt me :( Not even saying \'Hi\', nothing. Well do I care? Yeah I do, I don\'t know why, I don\'t know why I\'m beating myself up over it, looking in the mirror and thinking \'What the fuck\'s wrong with me?\' but I am. I seem to have taken about 100 steps in the wrong direction.  I feel inadequate and like a reject. I find it harder and harder to believe that there is \'somebody out there for me\', which I keep getting told. So please excuse me whilst I go puke in disgust at the appearance of my own face.  Moody? Fuck yeah...Reason to be? Fuck yeah....Reason to , be>/b>? No....So why do I carry on? Maybe there is still a part of me that hopes everything\'s going to work out. Though, how I\'ve managed to have even the smallest part of me being positive, I don\'t know.  Signing off.  Sir Fuckupalot '),(8668,3962812,'23,July,2004','2004-07-23 00:00:00','   Woooo lets be happy, wait lets not. Seen as though I feel like shit right now. Yeah so Angela came on MSN and she obviously hadn\'t received the rose, then whilst I was talking to her it came.  She seemed happy to receive it but couldn\'t understand why I sent it, I told her it was because I liked her and even sent her the poem I\'d written.   Take this rose I offer, A token of my feelings, A mere declaration from the heart,  Do not fear any thorns, For I will be your defence, To any harm which may arise,  If you take my hand, The path may be winding, And the road may be long, But I promise, I\'ll never leave you, Always by your side, Forever more...   She said \'aww\' lots and thanked me for the rose, but that was about it. I seriously don\'t know what to think. I had to go so didn\'t get any chance to speak to her and I won\'t be able to get online for a week. So meh, I dunno what\'s gonna happen.  Take Care All, I\'ll post again in a week.  XXXX '),(8669,3962812,'22,July,2004','2004-07-22 00:00:00','   Yep. So what did I do today....hmmm...went round Alex\'s, played on the Gamecube, came home. Oh yeah and had a driving lesson this morning...fun. So yeah what an exciting life I lead at the moment.   Tommorrow is Judgement Day heh, the rose is being sent tommorrow morning. So, could be good or bad. I\'m hoping for the best so I can be happy again heh. I really haven\'t got much more to say. I\'m sure I will have tommorrow.  Take Care All  xxx '),(8670,3962812,'20,July,2004','2004-07-20 00:00:00','   Right, this is going to pretty brief because right now I\'m tired. I went round Alex\'s today, played computer games, did a bit of Backyard Wrestling, got knackered and came home. Was nice for a change heh, am going round again tommorrow and Mark\'s going to be there wooo :)  But yeah, that\'s nothing compared to what I\'ve done today. I\'ve thought long and hard and well Ness is a nice girl and all I really just find her attractive, I hardly evern know her. Though I don\'t know Angela that well I\'ve spent quite a bit of time with her already and well, the first full day I spent with her was one of the best of my life and I wasn\'t even going out with her. She just seems to cheer me up, whether I get just a text off her, or whether I speak to her on MSN, my mood just seems to change.  So I stayed good to my own word, I\'ve ordered a rose for her, not a shitty one either. The most expensive one I could find heh, it\'s getting delivered to her house Friday morning. Â£25.75 including delivery, money that I really can\'t afford. I wrote on the card \'I was going to get you a white rose, but I knew you preffered red. Take Care xxx\' her mate Laraine is going to be there to make sure she\'s in and gets it, and possibly hint that it was me that sent it lol.  I got worried though, I was speaking to Angela on MSN and she said she might not be in Fri morning, so I quickly rang Laraine and she said that she\'d make sure Angela was in because she\'d be there with her. Heh, I felt better when she said that, I got really stressed when I thought she wasn\'t going to be in heh. Anyways, I\'m glad that\'s sorted now, I just hope she realizes that it\'s not just about the fact that I\'ve spent money on her, it\'s what this symbolizes. I really do think I\'m falling in love with her, and this is just a token of my feelings I have for her right now.  This could really fuck me up bad, if even after this and I\'ve given the poem I\'ve written to her, she still only sees me as a friend I\'m going to feel shitty. Whether she\'s nice about it or not it\'ll hurt, I\'m just hoping for the best at the moment and we\'ll see what happens. Who knows? For once it might work out, but maybe that\'s just wishful thinking. I\'m going to go soon, my hands are tired because of how fast I\'ve typed this and I\'m tired myself anyways heh. She said she\'d be back on MSN soon, I might just wait for her, have a quick chat then go.  Take Care All xxxxx '),(8671,3962812,'19,July,2004','2004-07-19 00:00:00','   In the morning I am empty Of every thing I used to be I write away the feelings  So no one will ever read  If I had all of the answers I could pour salt on my wound I could fill this room with sorrow So no one would ever see  I want to wake up and drink from your river I want reach out and tear down the sky To know I\'m alive yeah and I want learn how to feel without bleeding I want to learn how to live by your side You\'re no longer mine  Then why We got to swim against the tide to each other Send me on the side of the river Then why  If I offered you a picture And the paint is yet to dry Could you live with just a whisper A flower is meant to die  If I had all of the answers I could pour salt on my wound I could fill this room with sorrow So no one would ever see  I want to wake up and drink from your river I want to reach out and tear down the sky To know I\'m alive I want learn how to feel without bleeding I want to learn how to live by your side To know that your mine  Then why We got to swim against the tide to each other Send me on the side of the river  Then why We got to swim against the tide to each other Send me on the side of the river Then why   The lyrics to the song I\'m listening to right now. Breed 77 - The River, at the moment this song is making me feel slighly better. I don\'t know why but it seems to ease the pain. I feel a lot better about the whole picture thing now, the initial shock of it effected me more than anything I think, but I\'m trying my best not to think about it now.  I\'ve spoke to a few people and shown them this, all seem to give me pretty useful advice. Though one quote sticks in my mind. Nat said \'tbh i don\'t think you\'re gonna be happy with yourself til you\'ve found someone you click with\' and I think she\'s probably right. I think maybe it\'s a fault in my personality, this longing to have somebody by my side, to have somebody to confide in, to comfort me, to love...  I say fault because most people my age love their life. They have it easy at home, they don\'t have to work and go out and pull random people. I see it as a fault because although it may be a good thing that I want commitment so badly, people my age don\'t see it like that. Girls my age go for older guys if they\'re looking for commitment and slightly younger girls don\'t want commitment. That leaves me with nothing. Heh.  Another song that seems to have a lot of meaning to me is \'Vermillion - Part 2\' by Slipknot. Practically all Slipknot songs mean something to me and I can relate to them, but this one, as it stands means the most to me.   She seemed dressed in all of me, stretched across my shame.  All the torment and the pain  Leaked through and covered me  I\'d do anything to have her to myself  Just to have her for myself  Now I don\'t know what to do, I don\'t know what to do when she makes me sad.   She is everything to me  The unrequited dream  A song that no one sings  The unattainable, Shes a myth that I have to believe in  All I need to make it real is one more reason  I don\'t know what to do, I don\'t know what to do when she makes me sad.   But I won\'t let this build up inside of me  I won\'t let this build up inside of me  I won\'t let this build up inside of me  I won\'t let this build up inside of me   A catch in my throat choke Torn into pieces I won\'t, NO! I don\'t wanna be this...  But I won\'t let this build up inside of me  I won\'t let this build up inside of me  I won\'t let this build up inside of me  I won\'t let this build up inside of me   She isn\'t real  I can\'t make her real  She isn\'t real  I can\'t make her real   Basically to me, it represents that perfect girl that I could be with, yet all that she is is an mere illusion, fake hope to cling onto that will never become real. It\'s slightly depressing, but strangely comforting as well, knowing Corey\'s felt like this, yet now it\'s all turned good for him in his life, he has a beautiful wife and a child. Maybe there is still hope for me yet.  As you may or may not know I owe my life to Slipknot. Call me pathetic, call me whatever the fuck you want, but it\'s true. I started listening to the first album round about when I got depressed the worst. It helped me out of it and gave me something to cling to, a life-line if you will. I related to the angst Corey felt towards the world and the idiots that populate it. Not only did I end up loving the music it, it meant more to me than any other band ever could. It gave me something to aspire to, something to look up to. Nine men whom had all probably been through a helluva lot of shit just like myself, yet they were still standing saying \'Fuck You\' to the people that had hurt them. Just by typing this I feel stronger, just by thinking this through in my head I\'m feeling better. When people call Slipknot I get pissed off, not because they don\'t like the music, but because they don\'t understand. Music isn\'t just noise to me, especially not Slipknot.   My final thought for this post, try looking deeper into things. Whether it be when your partner tells you they love you, whether it be the music you listen to, or whether it be that friend who\'s always there for you. Look deeper and realize what it\'s all about, try and find a meaning to everything, because in most cases there is one.  Your partner probably does love you and wants to be with you forever, take that bond and do something about it. Take meaning\'s from the lyrics you hear from your music, actually listen to what the band are trying to say with it. Don\'t just dismiss it as noise. And finally look at your friend who\'s always there. Why are they always there for you? Maybe they care about you more than you realize...  I may feel slighlty better now, but probably not for long, I\'m still feeling generally shitty so yeah...  Take Care All xxxx '),(8672,3962812,'18,July,2004','2004-07-18 00:00:00','   I sit here now writing this tears are trying so hard to run down my cheeks to freedom. The wound in my heart I thought was closed has just re-opened.   I just went to find an old cd for Ness, to check if it was the one she\'d lost and said she could have it. When I looked under my cd cases I saw the picture. It\'d been there for over a year and a half, but still it looked perfect. Me and Adele sitting there together. Me looking completly different to what I am now. So happy, not a care in the world.  I still remember now how she\'d really not wanted it taken, but I begged her to cause I wanted something to put in my wallet so I\'d always be able to see her when I was feeling down. Seeing that picture brought back so many memories, all good, apart from the last one.  The one where I split up with her over an argument, a stupid little argument that meant nothing. I threw away everything because of that. The only girl I ever truely loved and I threw it away. I don\'t think I\'ll ever forgive myself for that, I had everything yet I was so stupid and young and careless just to throw away my happiness.  A few months ago I asked her if there was a chance of getting back together. I found out that there wasn\'t going to be any second chances for me cause of circumstances, I had well and truely fucked things up.  Now after every single relationship afterwards has lasted no longer than three weeks I have to wonder if it\'s me. If I\'m the one who\'s fucking things up, am I driving my g.f\'s to cheat on me? Am I driving them away? There have been one or two again where I\'ve finished people for no reason and I don\'t even know why, Nat, for instant. I finished her for no reason at all....I still regret that as well but there\'s nothing I can do about that now.  Maybe I\'m just destined to fuck up when I\'ve got a good thing, all I seem to have is regrets of stupid decisions I\'ve made. Maybe I won\'t be happy because of who I am and the way I am. I\'ve fucked up a couple of good relationships up to now with no reason to, so how many more am I going to end up fucking up? '),(8673,3962812,'18,July,2004','2004-07-18 00:00:00','   So yeah, I had to get up early this morning cause of the decorator coming round. I\'m going to be left on my own with him later when my Dad goes hospital. He\'s got something wrong with one of his eyes, or something, yeah so I hope that gets sorted. This morning I played Day Of Defeat for three hours, I swear I\'m getting addicted to that game, in the past three days I must have played it for about 12 hours, that is surely not good lol...But at least I\'m getting better at it :)  As I sit here and write this I\'m wondering about last night, about Angela, about a lot of things really. Ness just signed on, she\'s like really amazing and unfortunately just a friend. I\'ve been talking to her on MSN a lot lately and she\'s been having problems with her b.f. He treats her like shit but cause she loves him and believes that he loves her she\'s still with him. She said to me earlier today that if he upests her again she\'s going to finish him. If I were her I would have finished him a long time ago, but yeah I\'m trying my best not to interfere where it\'s not my business so I\'ve just given impartial advice to her, though it\'s been tempting to just fuck that and say \'Dump him, go out with me I\'d treat ya right\', cause I sure as hell would. She\'s one of the sweetest people you could ever meet and to boot she\'s incredibly good looking, thus why a guy like me wouldn\'t stand a chance with her.  She\'s one of those girls who could have any guy she wanted, but she doesn\'t realize it, which makes her even nicer. A girl who\'s pretty but not vain, qualities that usually match nowadays, pretty girls seem to think they can treat people like shit, this really pisses me off at college the most, they way they look down on anybody their own age and think they\'re too good for us, that really gets to me.  But yeah, Ness has made comments about my pics and that I\'m good looking, but I\'m not sure if she\'s just being friendly. She was sending me pics of her mates before asking me if I liked them or not, cause she was gonna set me up with one of them, but then she said I\'d send ya a pic of myself, but I\'ve got a b.f. This struck me as really weird, she\'s obviously thought about me as more than a mate, but if that will ever come into reality or not I don\'t know.  She\'s said she\'d split up with her boyfriend before, but it didn\'t happen. I can\'t really see them splitting up and I wouldn\'t want her to split up with him over me, cause I don\'t split up couples just to make myself happy. I\'ve tried my best not to flirt with her, tho I have said some suggestive things, but I think she just thought I was being friendly. As long she\'s happy I\'ll try and be happy for her, but I am really jealous of her b.f. He doesn\'t realize what an awesome g.f he\'s got and I seriously hope he gets a reality check and changes how he is, because as it stands nearly every time I\'ve spoke to her she\'s been upset because of him. I may like her, but she\'s also a really good mate and I don\'t like seeing her upest as she doesn\'t like seeing me upset either.  Yet, in the back of my mind I feel guilty just for even taking a liking to Ness when I\'m sort of pursuing Angela. I feel as though I shouldn\'t be thinking about any other girls other than Angela because of how much I like her, yet I\'m starting to like Ness just as much. The problem is I\'m not going to have anybody because even if I do get a chance with either of them I\'ll probably fuck it up, I\'m good at that.  A few weeks back one of my mates \'Nat\' from Law told me her mate really liked me. Well, I as over the moon cause I liked her mate over a year ago but she had a b.f, so I stayed away. I started speaking to her again and we got pretty close, spent some time together at the college fair and I thought everything was going really well.  At this point I\'d told her I liked her, but apparently she already knew off a load of other people :-\\, I said whatever happens I hope this doesn\'t ruin our friendship. After the fair she txt me saying. \'See today was cool wasn\'t it? We\'re still mates...\' and I just knew then that she didn\'t like me. I txt her asking if we\'d ever be more than just mates and she said no, in no uncertain terms. I seriously don\'t know how I managed to fuck that up when she supposedly really liked me.   So yeah, that was the most recent downfall I faced, and for now I\'m going to stop writing. I\'m listening to Adema - Pain Inside.  Take Care All xxxx '),(8674,3962812,'18,July,2004','2004-07-18 00:00:00','   Yes I\'m writing this because of him, I was going to try starting my novel that I\'ve planned out but that will have to wait till tommorrow now. Well I guess I should start with how I\'m feeling in my life right now and what I want from life. Yes this is going to be sooo intresting   So yeah, feel free to stop reading at any point, I still don\'t even know why I\'m writing this.  At this moment in time in my life I seem pretty happy, though I\'ve still got low confidence due to my past and the way people have treated me. I think I\'m slowly starting to find out that I can be more confident as I try new things and do things I\'d never have normally done. At last, I seem to have found my indentity, I know who I am now, what I am. I seem to understand myself completly and know why I feel down when I feel down and stupid things that most of you probably don\'t even think about. But most of all I feel reasonably happy with myself. Something that has haunted me for a long time, all my failure with girls, all the dicks treating me like shit, I put that down to me. I put that down to the fact that there was something wrong with me, thus why I failed with girls and why people treated me different. But now I finally realize, yes I am different to most people, yes I\'m not your average guy. But I don\'t want to be you \'average guy\', you\'re average guy sleeps around with as many girls as he can using them for fun. I don\'t want to be that man. I\'m glad that I\'m one of the only people my age who actually has morals and standards and doesn\'t just go out to \'get laid\'. Though I\'ll admit I made a mistake nearly a year ago, but since then I\'ve been stronger then over on my philosophy over sex.   I really am going to wait for somebody I truely love before I do anything again. My hormones have tried to tell me otherwise, but my heart disagrees. Now onto the topic of girls, I\'ve had a lot of shitty luck with girls, what with them using me/cheating on me and what not and yeah I\'ve had a lot of depression due to it. But I seem to be meeting the right kind of people now and getting to know that there are some genuine girls out there who don\'t set out just to fuck guys around.  At the moment I seem to have my heart set on one girl even though my mind\'s saying no. I seriously don\'t think I have a chance with her, but I\'m going to try anyways. One of my mates just summed it up perfectly on MSN, \'What have I got to lose?\', the fact is I\'ve got nothing to lose and everything to gain. She\'s such aa awesome girl and I get on really well with her, not only is she absolutely beautiful, but she\'s the nicest person to speak to as well. I feel as though I\'m getting on really well with her, but there\'s a lot of interest in her. I don\'t see how I can surpass the efforts of other lads, whom I see to be far superior to myself. Though I\'m going to try and who knows I might just get lucky and she might just see me for who I am. A kind, compasionate guy who just wants to love somebody and for them to love him back. I don\'t just want a quick shag, or a quick anything. I want a relationship with somebody whom I know it\'s going to work out with.  I met her back on Tuesday of last week. She sat in front of me and my mate Rob on the coach to Alton Towers. Rob knew her mate so we all got talking and we ended up hanging around all day. I can honestly say that was one of the best days of my life and I really didn\'t want it to end. Everything just seemed so right, the only thing I felt awkard with was when I had to say bye to her. I just felt as though I wanted to hug her or kiss her goodbye, and to say that I\'d be around if she wanted to chat or whatever, but it didn\'t happen. I just said bye and away she went.  During the day it was mainly four of us, Rob, Lorraine, Me and of course Angela :), but at one point Rob and Lorraine wanted to go on a ride that we didn\'t and we ended up having a pretty deep conversation. At one point she said I had very nice eyes, I didn\'t even realize she\'d said this until a few seconds after it had happened and I\'m pretty sure I didn\'t comment on it as I completly bottled it. I so wished I\'d returned the compliment and just said \'So have you\' maybe that would have made all the difference, maybe not. But I\'m going to have to leave that in the past now, she gave me her phone number then, so she obviously wanted to stay in touch with me, whether it be as a friend or maybe more. I hope it was for more than just friendship because this is the first time in a while I\'ve liked a girl this much and also been given the opportunity to spend so much time with one.   So yeah, I\'ve just finished talking to her on MSN, Juvie was sorta helping me on what to say and stuff and we talked about alsorts of things, including her star sign and mine and what flowers she liked. I found out she liked pink/red roses, and I told her I like white because they stand out from the crowd yet are still beautiful, this was no word of a lie and is actually how I feel. I\'m going to try and get her address off Lorraine so I can send a white rose to her house when I get paid on friday. I\'m going to write on the card. \'I know you like pink/red roses, but I like white\' and I hope she realizes who it\'s from.  I\'m gonna wrap this up now for today, cause am getting tired and it\'s getting later, I\'ve been typing for 20 minutes straight now so maybe I should leave it heh, my fingers are aching :-\\ lol, anyways I\'ll probs update tommorrow night, hopefully will have spoke to Angela again by then.  Take Care all...  xxxx '),(8675,3962812,'18,August,2004','2004-08-18 00:00:00','   What did we do today lets seeâ€¦  I got up, listened to Moyles on the Radio as always. Went on my computer for a bit chatted to a few different people. Went off and had my two hour driving lesson at 12 which went alright. Came back and had a kick around with Chris Goodwin. Came home and then found out BT are still having problems with their broadband service which is pissing me off a lot.  Other than that Iâ€™m in a good mood again, Iâ€™ll probably see a load of people from college tomorrow when I hope I havenâ€™t done too bad in my exams. I hope I get the results I wantâ€¦but thereâ€™s no point worrying which is what a lot of people seem to be doing.  Roughly about three weeks before I go back to college, I canâ€™t wait. All the new people to meet plus get to see all my mates again. I realize how much I really do miss college now that Iâ€™ve been away from it for so long. Oh well Iâ€™m hoping these next few weeks are going to fly by.  My boss rung my up the other day asking if Iâ€™d work Friday and Sunday. I said Iâ€™d do Friday and wasnâ€™t sure about Sunday. So as it stands Iâ€™m set to be working Thursday, Friday and Saturday, but I want to swap Saturday for Sunday because I want to watch Utd beat Norwich and I want to go out with the family to the pub as well seen as though weâ€™re having a little family outing like we do every once in a while.  So yeah busy times ahead of me, Iâ€™m going meeting Dale at about 12pm tomorrow and then weâ€™re off to get our results. Heâ€™ll have aced everything as always heh. I hope my broadband is fixed soon because itâ€™s really starting to annoy me now.   Take Care All xxxxxxxxx '),(8676,3962812,'17,August,2004','2004-08-17 00:00:00','   Yeah it\'s been a while since I wrote in here. In that short space of time I\'ve...  - Had a g.f, realized I didn\'t know her so broke up with her to save hurting her further down the line.  - Had a hair cut.  - Gone on a diet.  - Bought some weights.  - Decided to get some muscle on me.  - Bought a new football.  - Finished reading Silence Of The Lambs and started reading Hannibal.  - Watched the film adaption of Red Dragon, very good I must say. That has now become my favourite book.  As it stands I feel great about myself and I\'ve ordered a new t-shirt http://www.engrish-store.com/dangertshirt.html woo woo lol. So yeah, let the good times role. Exam results on Thursday. Driving lesson tommorrow.  Take Care All xxxx '),(8677,3962812,'09,August,2004','2004-08-09 00:00:00','   Finished reading Red Dragon. Tis a good book.  As always I feel like shit.  So I might as well just end it here.  I\'m doing nothing again.  What a great life.  Fuck :( '),(8678,3962812,'04,August,2004','2004-08-04 00:00:00','   Yeah well here\'s a little more detailed look at what happened on Monday.  Me and Dale got on the bus took fucking ages to get to Manchester which takes the piss and it\'s st00pid cause the train line\'s being worked on so yeah. We spend an hour or so wandering around and run into Nicola Simms who then starts hanging round with us. Then she finds a mate so there\'s four of us. We spent up until about 5ish/6ish like that and it was cool.  Then A.J and a few other people turned up, Nicola and her mate went and said they\'d ring me cause they\'d be back in a bit. They never came back. Probs because Nicola and Guy had a bit of a thing and don\'t speak to each other now but meh...So yeah, so that left me with Guy, Bauer, A.J, Laurie, Dale and Ruby - Bauer\'s g.f. I hardly hang around with any of these guys, so none of em really bothered talking to me. Some other people Mel and Del the only two who\'s names I caught hung around as well, and basically I started feeling left out. I get on alright with A.J and Dale when he\'s on his own. Then Tom Allen came and he was just obssessed with going seeing Mel and Del. At this point it was about half seven and all they were planning on doing was hanging around Urbis. So fucking boring in my opinion especially when nobody gives a fuck if I\'m there or not, so I ended up going home with Laurie, or at least getting the bus with him. He made a bit of an effort to speak to me then which was cool but yeah basically he didn\'t really want to, I could tell it all seemed forced.  So that brings me onto yesterday...I did absolutely nothing other than sit at home all day. What fucking fun that was...not...  Today I went round Alex\'s and messed about on Gamecube with him and Mark, that was alright. Got home same old, on the net and for a few reasons I feel really shitty right now, more shitty than normal. What\'s the saying, one step forward two steps back? That\'s what I feel keeps happening to me and at the moment I\'m about 100,000 steps away from being happy heh.  On a slightly brighter note I think I\'ve found somebody I like who\'s a mate whom I\'ve never really looked at in that way. Though I\'m sure she\'s never looked at me in that way either. So basically I asked her if she fancied coming out, in a group thing of course, I didn\'t want to make it too obvious...she goes my college and I hang about with her quite a bit. She\'s a really nice girl and I know if she likes me she wouldn\'t hurt me or whatever, but once again I\'m going for a girl whom probably only sees me as a friend so I\'ll probably end up getting hurt. I slipped in my number and asked her to text me if she sorts anything about a group of us going out, smooth eh? But yeah. I\'m hoping maybe I\'ll get a bit of luck with her, who knows.  I gotta keep trying, because eventually I\'ll find somebody who likes a loser as a b.f heh.  Take Care All xxx '),(8679,3962812,'02,August,2004','2004-08-02 00:00:00','   Yep nothing happened hahaha, how did ya manage to guess that? There was plenty of opporunities, I guess I need to work on my self confidence a bit more :-\\ heh, but yeah I finished typing up my short story now wooo so that\'s good, I sent it to a few people and they like it so that\'s good. I need to start writing my next one now whilst the idea\'s fresh in my head.  On other news, I sent a random person a message on faceparty and it turns out it\'s somebody who lives round the corner from me. Scary :| lol...  So erm yeah, nothing of interest to report.  I bid thee good-day. '),(8680,3962812,'02,August,2004','2004-08-02 00:00:00','   Yep so this is going to be pretty short probably seen as though it\'s 9AM!!! But yeah, today should be erm interesting :) I\'m going into Manc town with Dale, we\'re gonna go chill at Urbis and stuff, Dales words were \'We gonna go stare at the rock chicks from afar?\' LOL, so yeah basically off into Manc we go and we\'ll see what happens. Hopefully some good fortune after the horrid few weeks I\'ve had recently and I\'ll try and forget this nasty mess with Angela.  So I bet you can\'t wait for me to come back now can you? Cause you want to know what happened don\'t you? Well, I haven\'t been yet...  Anyways  Take Care All xxxx '),(8681,3962812,'01,August,2004','2004-08-01 00:00:00','   Well, no, not really. Today = Worst Day ever. I find out off Angela\'s mate she has a b.f now. No wonder she didn\'t txt me or e-mail me.   My Mum\'s dead ill again so she\'s in bed right now.  Yet, I\'ve got to be positive or I\'m going to have a fucking break down. I need to try and stay positive. Fuck girls, seriously I can\'t be arsed anymore, I\'m just gonna get messed around if I like anybody so I just won\'t bother.  I\'ll write more tommorrow.  For now, take this thought with you...  \"You can try as hard as you like, but in the end, it won\'t matter...\"  Profoundness from my mouth.  Take Care xxx '),(8682,4060980,'19,August,2004','2004-08-19 00:00:00','   sian...today sucks...once i reach sch n go into the class, my teacher return us the java term test paper le..i failed, even though it was juz 15% of the overall..but the problem is i still failed ma..somemore i only get 30over marks nia..naBei sian..after that is our eng fund lab le..today we learnt new stuffs..kan diff.but luckily got our sgt to teach mi n marcus..den agar agar noe how to do le...after that i still teach mon how to do..followed by shaoWei..but shaoWei\'s that one i dunNo wat problem..bo bian..have to ask sgt to help out again..lol..haiyo..y ENG.FUND so diff de..wtf..shit..we had lunch after that..den go for the ah pui lesson..tok cock one..boring like wat..den after the ah pui lesson, we skipped uncle lok\'s lesson, den we go library to play...play until very noisy..haha...that was enjoyable..we laugh like mad..after that i very moody liao..dunNo y...oso dunNo wat i thinking abt..acc mon home that time wif shaOwei..mon keep asking mi wat happen?..y u so sian?...come back ah suan?..usually when she ask all this i will tell her wat happen one..but this time dunNo y i dunNo how to answer her..i juz kept quiet, sorry hor mon..n i was there daydreaming.maybe thinking of y i get those poor results or maybe something else i dunNo...after sending mon home i was toking to shaowEi,trying to say out y i so moody..ya..i oso got nth much to say actually..dunNo lei..i ki siao already le..mood swing ba...ya...that\'s all for the day...yup yup..   '),(8683,4060980,'16,August,2004','2004-08-16 00:00:00','   LOl...today meet my friend to go sch...on the way there he told mi that girl patch back wif her ex le..Lol..CONGRATULATE TO THE TWO...her ex really look like horse lo..really..dun believe got chance i let uall see..the face mo pan one..kan disgusting..i oso dunNo y they will patch back la..but i oso dun wish to think so much le.let them be ba...hahaha...at nite got nth to do..i dunNo wat to write lei..tonite boo lei boo lei..LOl..tok abt ytd ba..meet marcus in the afterNoon go play poolactually wan to ask shaowei one..but he stilll playing soccer so bo bian lo..he cannot come..sian..so i go out wif marcus only..after that go his hse take something...den he take out one hard liquor again..haha..den i drink a bit lo..at first ask marcus to pour a bit enough le..coz mixing wif coke one..but when i drink a bit..BO BEE lei..not nice..so i ask marcus to pour more...hahaha..after i drink a bit, i told of wat happen last thursday..LOL..so i tell myself i cannOt drink so fast..but after a while, mon called le..n she say reaching le..so have to ta the rest of the cup..LOL...so meet mon le...go eat yoshinoya..dunNo y i get two cups of coke for free..haha..i see mon de leg full of mosquitoes\' bites man..sian..TO MON>>sorry hor..hai ni kana so many mosquitoes\' bites...sorry sorry..today get back my e-maths n eng fund results..maths still ok la..but my eng fund sucks lo..get 55marks..sian..multiplexier la..bcoz of that..my 14marks fly away lo..sian..naBei..dun write le..i wan to sleep le..Gd nite ppl.   '),(8684,4060980,'14,August,2004','2004-08-14 00:00:00','   TOday i didn\'t nth at home...morning i went to play bball wif my friend...den got two stupids come n play wif us...they played until very \"smelly\"..put their legs to trip mi..WAN MI TO DIE MEH??..lol...after that mi n my friend finished them in less than 3mins..lol..we played until very fast den they followed until like idiots like that.after the game, they keep saying:\"they very fast hor they very fast hor\"..lol..kan cute lei..after that when i reach home i kan sian..dunNo y lei..i can\'t stay at home one...like got needle poking mi..naBei..y todae so sian lei..izzt bcoz of the msg thingy lei..who knows..k la..i ask ppl who is reading this..if u msg someone abt some important stuffs den half way through he or she dun wan to reply liao...wat does this mean??.i dunNo..stupid... 7plus in the evening le..my friend call mi n ask mi to play bball again...siao bo...morning played le..nOw still go n play..can die..lol...go to pcc played 2on2 again wif two guys..wa lau...i block one guy..he used his head to knock my nose...he like two years nv shampoo his hair le lo..his hair nOt very smelly, not kan smelly, oso Not kan gei pua smelly, but juz SMELLY!!!!!!!!!!!..shit...i nearly fainted there..lol.. haiyo todae no mood to write my blog lei..oso dunNo wat to say, juz that i m kan sian.y ah?..i ask this the second time le....mood swing meh?..can\'t be wad.ehhh..maybe hor...maybe maybe..shit..i ki siao already...lastly, let mi talk some sense, i really think that fate play an important role in our lives...agree????.   '),(8685,4060980,'13,August,2004','2004-08-13 00:00:00','   hahaha...ytd i went to my class bbq..lol...at first we reach there, mi marcus shaoWei yuki n aH pok went to play bball..play untiL kan funnY..ah pok really like bull like that..hahaha.after that NA MA came n ask us back to set fire n start the bbq as it is already 7plus..ya..ah pok set the fire until he cried man...hahaha..actually No la, is bcoz of the smoke that irritated his eye.lol.mi n shaoWei were setting up the other fire too to cook our maggie mee..Not bad actually..juz like last time what i do during npcc..lol..we started to bbq..the fire was damn big n kan hot, i think if i stood there longer, my stomach would have been cooked.hahaha...den i cook the chicken satay(not enough skills la)..so some of the satay is still raw.den after i ask ah pok to try one, he kao pei kao bu say the satay still raw de..LOL..hahha.after that we started to drink JIM BEAM(sorry if i spell it wrongly) wif coke..at first i drink it was still perfectly alrite..but after i drink 8cups of it..the kick slowly slowly comes..lol....n i was somehow drunk..ber lu ber lu de.shaoWei n Mi vomit mAN...hahaha..luckily i got one super duper gd bro, mon to take care of mi..hahaha..XIE XIE NI!!!!!!!!hahaha...around 3 plus, i can play volleyball liao, i still go jogging wif marcus..two rounds around the small field.hahaha..run until like old dog man..after that got nth to do liao..kan sian..so i ask mon go walk walk...it\'s  already 4plus in the morning, and the thing that make mi surprised is mon isn\'t tired at all n is as lame as be4..LOL..hahaha..ytd she still think of some new phrases...IT IS ALL OVER!!! rite mon???.haha..den when we go back to the pit it is already 6 like that le..time to go home liao...hahaha...marcus very gd..he took cab den long pang mi n shaoWei to the bus stop to take bus...hahaha..when i reach home i was already half dead...lol...that\'s all....IT IS ALL OVER..if mon see this, she will say pls dun imitate mi...LOL..   '),(8686,4060980,'11,August,2004','2004-08-11 00:00:00','   Today i did nth at home.kan sian...now is around 5o\'clock la.weather Not very hot, not kan hot also, is kan gei pua hot...na bei...three weeks past the matter liao, she still dun wan to forget abt it.WAT SHE WANS?!?!?!?!?!?!..think it is abt time to forget abt everything le..forget everything....YES!!!!!!!!..naBEi.ytd mi shaoWei n maRCUs go n sing that stupid KBOX.one mistake>>>FORGET TO WAIT FOR MON DEN GO TOGETHER.soRRy hor mon...lol.. kboX super ex man.they tot we cai tou ah.one cup of LEMON HONEY for 5bucks.it is juz made of honey, plain water n two slices of lemon..THAT\'S ALL..wa lau....i see the bill until i chua sai man...lol..sing until my heart cracking le.sians....after tat we go home le. when i reach home, my mum said keep spending spending money...nag nag nag until i buay lun..but really la..this few days spend quite a lot of money.must budget le..ACTually today meeting SIAO EH to buy solid fuel one..but he say his leg pain..so nv meet up in the end sian...hahaha..if shaoWEi see this he will say,\"MAI KIO WA SIAO EH\"..hahaha..kan funny.buay lun.den  ytd we still sing the TIAN SHANG REN JIAN by my bro,fei yu qing.na bei..that stupid marcus say i look like fei ge...hahaha..wat an insult to mi man.but nvm..i forgive him. anyway fei yu qing the face kan funny.stupid curry pok head..hahaha.wa lau..my com siao one lei sometimes can connect to internet sometimes canNot..juz Now wan to connect den say wat canNot dial..i CHUA SAI man..tot got wat problem again..heng ah..after i tried a few more times can le.lol.....aiyo..very sian lei..dunNO wat to type..dun write liao la..sian...   '),(8687,4060980,'04,August,2004','2004-08-04 00:00:00','   hi   '),(8688,3829312,'03,July,2004','2004-07-03 00:00:00','   First of all, let\'s get the quibble about spelling out of the way.  On  urlLink  Black Table  , where the article first appeared, it\'s \"whimpster\".  In  urlLink  Bust  , where it was recently reprinted, it\'s become \"wimpster\".  I\'m going with \"wimpster\" because that \"h\" makes no sense.  Besides, I think I should have right to spell it how I want, being one and all.  I mean, I\'m a guy who reads  Bust .  And I worry it\'s too establishment.  I prefer  Bitch , because it doesn\'t waste as much space on makeup.  Yeah, I\'m a wimpster.  Manipulative, passive-aggressive, \"emo\".  I display all of the warning signs.  I have even said, and I quote, \"I never ask girls out because I don\'t want to sound like some sexist asshole.\"  And I meant it.  And not to lay blame here, but let\'s try to get to the root of exactly where my ilk come from.  I submit that our reflexive crippling PC-ness is in fact borne of a positive intent.  I\'m hip to the whole feminism thing, and I think I can be reasonably certain that I follow most of the major arguments, even when I run into words like \"reify\".  Being probably more than a little neurotic, I and people like me read about the systematic and often unconscious oppression of women and minorities and we\'re shocked.  This can\'t be us.  We\'re not bad guys.  So we try our damnedest to cleanse our thoughts and actions of suspect behavior.  Like the whole slut/stud double standard, which probably winds up being the crux of our problem.  See, one thing I think we wimpsters have going for us is that we\'re open to the possibility of having friends, pals, chums of the opposite sex without reading too much into it.  The danger comes when it becomes apparent that it would be really fucking swell if there were some way to read a little bit more into it.  The \"Ducky Scenario\", if you will.  We won\'t say anything, because it will corrupt a completely worthwhile friendship.  Put it this way:  We\'re your passive-aggressive non-boyfriend.  It starts out with a lotta beers, lotta laughs, lotta mixtapery.  No pressure, no interest.  Then pow the male hormones kick in and we realize you\'re not only one of the guys, you\'re uh, totally, uh, hot.  And stuff.  This is where things start to get unstable.  Basically the goal is to determine if you\'re interested in having sexual intercourse with us (\" like  us like us\") in a way that wouldn\'t jeopardize our friendship if it turned out you don\'t.  Admit it:  having one of your friends walk up to you one day and make the kind of confession that typically only figures in bad romance novels and the lives of middle schoolers is reasonably traumatic.  So we don\'t do anything.  We try to suss out, based on your throwaway comments about popular culture and local architecture, what you might be thinking.  Once something finally does happen, because we\'ve nurtured this atmosphere of ambiguity around the whole thing, we can\'t actually figure out why you\'re doing it with us (NB:  All American men in their mid-20s invariably think to themselves, \"whoa.  I\'m totally  doing it  right now\" in a Butt-head voice during sex).  So we keep up with the noncommital \"friend\" act while you get bored and cheesed off with the whole sordid affair.  The sex remains rote and PC for the duration because we\'re still operating from the premise that if we fuck up once, it\'s off.  So here\'s my invincible cure-all solution.  I\'ve examined it and it\'s devoid of double-standard, completely germane to use by either side of both heterosexual and homosexual platonic standoffs.  If the first party entertains or suspects they might entertain sexual designs on the second party, heretofore a friend, the first party shall forthwith enquire of the second whether or not the second party would be interested in screwing, often.  If the second party is indeed interested, what happens next is no business of mine.  If not, however, and this is where the important part comes in, the second party must, after declining, agree to completely forget the preceding exchange and the five minutes of stammering that preceded it.  You will note that this arrangement has a lot in common with what happens if you get really fucking drunk hanging out with someone you have a crush on, only the blackout part is mutual and by fiat, and ideally nobody will tell you what you did the next day.  '),(8689,3900039,'28,July,2004','2004-07-28 00:00:00','   Gabe is on a plan right now sleeping on his way to Australia. He comes home on my birthday now, I\'m glad. He asked me if I could handle not hearing his voice for 18 days... how corny can this boy get?  We talked on the phone for a little too long last night, and we had quite a messed up conversation. It was a good one. I\'m supposedly getting presents, we\'ll see if that happens. I still haven\'t even gotten pictures!  Collin come homes on Sunday, I\'m really glad. I need somewhere to live part time.  I\'m at Ryan\'s and he\'s fighting with me over the computer, he\'s a punk. He\'s trying to teach himself quitar and so he \"needs\" the computer to look up \"lessons\". whatever..... its my turn. he is in deep trouble now... he just tickled the crap out of me so he could listen to like two seconds of a song. Oh he is so dead later. And he is reading every word of this...and I don\'t care.  Work was a lot better today. My art friend Aaron Smith works there too, so I was glad to have nice person to work with. He got me a little bit more excited for school \'cause our art class is just gonna ROCK! I\'ll have Corynn and Jaime and Lance and Aaron and Shannon and Emily, and who knows, maybe some new juniors! And we can all be art nerds together! Oh I love it! But no more Patrick Orr to stare at all period...I don\'t know if I\'ll be able to handle that yet.... I might actually have to work instead of flaunt and flirt! AHHH!!! That\'s sooo hard for me haha  Only about a month left of summer...That really really really sucks. But I guess I will get to see everyone all the time. The school district is still thinking about whether we are going to have closed campus or not. They already decided we have two lunches and they are only 35 minutes.....I\'m so mad. I hate the stupid people who decided that we were too irresponsible to be free at lunch. The kids who skip will find a way to skip either way, stupid administrators! Even if you lock us up and put guards at all exits and surround us with barbed wire, the kids who want to leave will find a way. Besides, the kids who don\'t want to be there, kind have a negative effect on the kids that do want to be there. It\'s so much more enjoyable without someone complaining and distracting, and just sitting there doing nothing and voicing how much they don\'t care. And on top of it all, afterr two years of patiently  not having parking and going to a school that looks like a war zone(construction), I think I deserve to be allowed to go to lunch...  On a lighter note, I have Friday off so I\'m going boating with the Lovemarks! That\'s always a good time. hehe   '),(8690,3900039,'27,July,2004','2004-07-27 00:00:00','   urlLink     I hope you have a wonderful time in New York and around. I\'ll miss you tons. Be safe, have fun, don\'t die. Remember, if you are bleeding, I\'m not there to take you to the Emergency Room, so you will have to manage it on your own! HA! Don\'t worry, I can hold the fort down while you are gone and keep the little ones(Tina, Court, and Justine) from quareling too much. Well hopefully I will do better than I did when you went to Costa Rica... hehe. THAT was a nightmare. I love you tons, have a good time!   '),(8691,3900039,'27,July,2004','2004-07-27 00:00:00','   so i had my first day at  linus oakes  today, and i hated every minute of it. all the girls i work with hate me. it sucks... but at least the  old people  were nice to me. the scabs fro mthe blisters i got&nbsp;from walking so much in portland came off....they look so gross.&nbsp;i have  friday  off luckily. OH and i had to go get a tb test and i had an alergic reaction to it so i have this red thing on my arm! its gross....they thot i might actually have tuberculois! it was scary.  gabe  leaves for  australia  tomorrow :( we had a good long  talk  last night. he seriously is so freakin\'  perfect  it scares the crap out of me. i hate it. we\'ve figured the next time we are going to see eachother wont be until  october ....thats so long... i think for my  birthday , which is in less than a month(august 16), i\'m going to go out to  dinner  with all my really  good friends (carol, justine, christina, court, jenna, collin, and ryan). if anyone has suggestions of good places to eat, make them now, cause i dont care. then afterwards have even more people over to make  homemade coldstone (jenna can be the example haha) and just hang out and swim or whatever. i still don\'t know what I want, any suggestions? we have to have my  party  early tho, im thinking the 12th cause i have it off of work.  im done.   '),(8692,3900039,'25,July,2004','2004-07-25 00:00:00','   This has been my favorite song lately....the other day I had it on repeat for over 6 hours... don\'t ask...    &nbsp;   Crash Into Me &nbsp;&nbsp;&nbsp;&nbsp; Dave Matthews Band   You\'ve got your ball, you\'ve got your chain Tied to me tight, tie me up again  Who\'s got their claws in you my friend?   Into your heart I\'ll beat again  Sweet like candy to my soul  Sweet you rock, and sweet you roll   Lost for you, I\'m so lost for you    Oh, and you come crash into me   And I come into you  And I come into you In a boy\'s dream In a boy\'s dream   Touch your lips just so I know  In your eyes, love, it glows so  I\'m bare-boned and crazy... for you    Oh, and you come crash into me   Baby, and I come into you  In a boy\'s dream In a boy\'s dream   And if I\'ve gone overboard then I\'m begging you   To forgive me in my haste   When I\'m holding you so girl, close to me   Oh and you come crash into me, yeah Baby, and I come into you Hike up your skirt a little more And show the world to me Hike up your skirt a little more And show your world to me In a boy\'s dream In a boy\'s dream  Oh, I watch you there through the window and I stare at you  You wear nothing, but you wear it so well   Tied up and twisted, the way I\'d like to be   For you, for me  Come crash into me Come crash into me, yeah Crash into me...Crash into me...  You know,  I\'m the king of the castle,You\'re the dirty rascal  Crash into me  Please crash into me, baby...  Oh, no no no... Yes, I see the waves Come and crash into me  See the waves come and crash into me  Crash into me   '),(8693,3900039,'24,July,2004','2004-07-24 00:00:00','   So today Courtney and I went to Collin\'s house to swim. I got there first so I just let myself in and everything \'cause I assumed he still was out of town. But I noticed shoes and a pillow and some odd placed things....so I kept calling him and there was no answer, so I went downstairs to his room and there was the boy and his dog fast asleep. It was one of the all time cutest things I\'ve ever seen. He woke up shortly and I switched his laundry for him. Court and I swam and when we came back in Quentin had arrived. Then Court and I went to get ice cream(and to see Bryan oh so cute). Christina showed up at Collin\'s while we were gone and called to ask for ice cream too and it was a bad mess....we had so much ice cream all over the car, just a huge sticky melting mess, and then we didn\'t get Quentin any and I was afraid he would be upset, but he seemed fine(we are on good terms, I\'m glad). Kaitlin also came over and Collin had a mini-reunion with all us. Then I sent him back off on his way to Portland with Mindy to catch a flight to Colorado for a family reunion. I seriously haven\'t seen him hardly at all, but he will be back for good in a week. Thank goodness... I\'m having major Collin withdrawls....which is probably why I\'m at his house with out him and we will prolly have a party here tonight. well not a party...more like a get together with the gang to swim and have a fire haha! So I\'m going to jump in the pool now.   '),(8694,3900039,'24,July,2004','2004-07-24 00:00:00','   so I had a blast up in portland with the girls. We went shopping and dancing and came up with some awesome ideas for next year\'s yearbook. It will be great. I have these huge sores on my feet from walking so much. I guess i had some blisters and they popped and now they are just disgusting. But right now, I feel very unhappy, unloved, unwanted, and uncared for. I hate my family, I hate my life, I hate myself, and lately it seems I hate my friends. I\'m getting the praying-I-don\'t-wake-up feeling again, and I desperately wish it would go away. I\'m sucking at pretending today, and everyone I\'ve talked to has noticed.   '),(8695,3900039,'22,July,2004','2004-07-22 00:00:00','   I\'m in Portland at the moment, and we convinced Tovey to get on to the teachers\' account so we could blog and check email :D today was pretty uneventful minus some arguing among us yearbookers. we are working really hard on this book and we want it to look perfect and some people are just getting in the way of that haha just joking tomorrow is a fun day, and I\'ll probably get to see gabe! WHOO HOO!!! we are going shopping downtown. Anyway, someone else\'s turn on the computer... I miss you Carol and Jenna and Ryan and Collin!   '),(8696,3900039,'19,July,2004','2004-07-19 00:00:00','   1. WHAT IS YOUR FULL NAME?     Nicole Renee Gates                     2. WHAT COLOR PANTS ARE YOU WEARING?     Grey capris                   3. WHAT ARE YOU LISTENING TO RIGHT NOW?      dashbooad confessionals \"vindicated\"                   4. WHAT WAS THE LAST THING YOU ATE?      chicken nuggets        5. IF YOU WERE A CRAYON, WHAT COLOR WOULD YOU BE?     magenta, says ryan                     6 . HOW IS THE WEATHER RIGHT NOW?     its been a pretty good day, its late afternoon sunny                   7.LAST PERSON YOU TALKED TO ON THE PHONE?      alexander                    8. THE FIRST THING YOU NOTICE ABOUT THE OPPOSITE SEX?      hair                   9. DO YOU LIKE THE PERSON WHO SENT YOU THIS?      well i copied it off someone else\'s blog.....so......the person i copied it from has been a jackass lately  so no             10. HOW ARE YOU TODAY?     good, i am spending time with my friend ryan             11. FAVORITE DRINK?     grape juice        12. FAVORITE ALCOHOLIC DRINK?     i don\'t drink, but screwdrivers are very good....i only tried once                         13. FAVORITE SPORT?      cheerleading is not a sport. i like to watch lacrosse lol        14. WHAT IS YOUR HAIR COLOR?      dark brown        15. EYE COLOR?     Dark Brown.                     16. DO YOU WEAR CONTACTS?      yes        17. SIBLINGS?       yes, a sister(14), a brother (8 on thursday), and another sister(5)                     18. FAVORITE MONTH?      august, my birthday, summer, i love it        19. FAVORITE FOOD?      macaroni and cheese        20. LAST MOVIE YOU WATCHED?      butterfly effect                     21. FAVORITE DAY?       friday during the school year, during the summer, any day my friends have off work        22. ARE YOU TOO SHY TO ASK SOMEONE OUT?      i havent had to ask anyone in a long time, it just sorta happens, there is no asking        23. SUMMER OR WINTER?      summer, i hate being cold               24. HUGS OR KISSES?      hugs, im  a very cuddly affectionate person          25. RELATIONSHIPS OR ONE NIGHT STANDS?  relationships             26. CHOCOLATE OR VANILLA?     vanilla                            30. WHAT BOOKS ARE YOU READING?     i just finished  \"a brave new world\" it was creepy, i need to start one soon for english, i also want to read summer sisters                   31. WHAT DID YOU DO LAST NIGHT?     slept and talked on the phone                   32. FAVORITE SMELLS?      David Robertson\'s house..... idk.... carol\'s perfume                    33. WHO/WHAT INSPIRES YOU?     i dont get inspired easily                     34. SALTED OR BUTTERED POPCORN?      buttered                   35. FAVORITE CAR?      i really want a mazda tribute, but i love jennas jeep                   36. FAVORITE FLOWER?      peach colored roses        37. HOW MANY KEYS ON YOUR KEY RING?      2                     38. CAN YOU JUGGLE?      no, not enough eye hand cordination, thats why i was a cheerleader, none required        39. RED OR WHITE WINE?      i hate wine        40. WHAT DID YOU DO FOR YOUR LAST BIRTHDAY?     had the girls over                  41. DO YOU OWN A DONOR CARD?     no        42. WHAT IS THE FIRST THING YOU THINK OF WHEN YOU WAKE UP IN THE MORNING?      how tired i am   '),(8697,3900039,'18,July,2004','2004-07-18 00:00:00','   Ryan, the world\'s biggest computer geek ever, just accused me of being too much of a blogger....even though he is the one who got me started blogging in the first place. this is ridiculous! he freakin\' sends audioblogs from his phone to his blog when he is out of town!!!!!! i just have a lot to say so mine are long so get over it ryan! (the real deal: im spending too much time blogging and not talking to him...he envies my blog because i give it more attention than i do him.) this is outragesous!  &nbsp;   '),(8698,3900039,'18,July,2004','2004-07-18 00:00:00','   Carol put a song on her blog about missing people who are no longer in your life. I recently had a conversation with someone I had not talked to in a long time and it made me realize how much I truly miss him, and&nbsp;missed missing him. So this is a sorry to him and the others that have somehow over time become to be&nbsp;a past and not&nbsp;a present part of my life...   Scott:  Thank you so much for caring about me and praying for me when we were younger. Tears come to my eyes just remembering the times in my life when you would pray for me and bring hope back into my eyes. You are an amazing person in every way and I hope your future remains as bright as you make it out to be. You are so optimistic and I love that about you. You have not changed in the year that we havent had an indepth conversation, and I hope you never do. I wish you luck on your wish to become a pastor; you would make a great one, and I would know as you were my own little pastor(Jeremiah 29:11, I still have it memorized). I am so sorry that high school and my own self conciousness and busyness has drawn us apart from eachother. I really wish things had worked out differently. You will always be special to me.   Ryan:  Although I have already repaired our friendship and been forgiven for the time I did ignore you and let what other people think get in the way of us, I want to say sorry again. There was a whole two years between our strong friendship and I want you to know that I feel very badly&nbsp; for making you \'non-existant\' during that time and I never will try to rid you from my life. It obviously didn\'t work the first time, so I\'m sure there is a reason why you are in my life and I will just sit back and let myself enjoy having such a fun great person in my life. I love you, Ry, and look forward to getting closer to my best friend, if that\'s even possible.   Tiffani:  We\'ve been through a lot, and I\'m glad I had someone there. I know you and I will always be best friends and we will always love eachother, but I also know there are things we just can\'t change and that there will always be a odd distance between us. I am sorry I could not be there for as much as you needed and I know you tried your hardest. I do not blame you for anything Fani and I love you dearly. Life is already separating us, and I hate it, but I don\'t know how to stop it. I just pray that we will always know how special our friendship was, and that it\'s not our faults that things didn\'t go our way. God has his reasons. I will never forget you.   Baily:  I\'m sorry that I never took the time to get to know you, and that i judged you tremendously during the short time I did know you. I am sorry that I have much anger still and no one to take it out on but you. I know I shouldn\'t blame you but it\'s difficult not to. I know you would never have intended&nbsp;to hurt anyone you loved, but it is hard to look back&nbsp;at everything Tiff went through and blame no one.   Moyer:  I know we have never talked about it, so I guess I\'m just assuming what made us grow apart, but I also know my assumption has a lot to back it up. I am sorry for what I did to you, and I know it was wrong and I regret it more than you can ever understand. I\'m sure college also took it\'s toll on our friendship, but I know that Tyler and I were probably a bigger reason. I was a fool, and I can only look back on it now and apologize to you.   Quentin:  I took a long time, but I am glad both of us have finally grown up and gotten over all the cruel things we did to eachother. There are many things I regret, but many things about our relationship and history that have helped me to grow and mature. I am sorry for the process of doing so. I know things will never be the same and as hard as we try we will never be as close of friends as we once were and that there are many things I did that you were undeserving of, and for that I apologize.   Jacob:  Thank you for everything, including forcing me to realize I could live without you and I didn\'t need to depend so much on only one person. The end of our friendship made me realize how many more people I had in my life to love. I wish you well in your life. I am trying my hardest to move on and I\'m sorry for needing to know more from you than you want to give. I know things will never be the same, and that makes it hard to forgive.   Zach:  I do not know how much I hurt you, or if I even did, but I know some of the things I did to you were wrong and for that I apologize. You hurt me also, but I have moved on, and I hope you have done the same. There was a time in my life that you were important, and I\'m sorry that it is no more. Thank you for the fun times we had. I learned alot from our relationship.   Sierra, Autumn, Summer, Jennie, Jamie, Haley:  You are all my most prominent grade school memories. I\'m sorry that as we continued with school and grew up we lost contact and our friendships. I wish you all well in life.  Brittany:  I wish I had tried harder to save our friendship back in seventh grade. You were a very important part of my life when I was young and I apologize for throwing that away.   Ricky:  You hurt me tremendously during the short time we were friends, but I want to let you know I have moved on and that I only wish you happiness and hope you are safe and well. I\'m sorry that circumstances at the time only made us hurt eachother. I\'m sorry that I will regret every memory I have of you for the rest of my life, and I wish I had been strong enough to change that.  Most of these people will never read these, but no one has any clue how much more of a relief it is just to get it out somehow.   &nbsp;  &nbsp;   '),(8699,3900039,'17,July,2004','2004-07-17 00:00:00','   &nbsp; I said I was going to put a picture of my dear friend Carol on here so I might as well write a whole post on her.&nbsp;Carol and I have know eachother a lot longer than we have been friends. Our mom\'s were teachers at the same school, so we had some contact with eachother. She thought I was a tomboy(cause I was with boys all the time...) and I thought she was a snob. In eighth grade we became friends...I don\'t know how, lots of mutual friends I guess, and she realized I was just a flirt and not a tomboy, ha.  She is one of the most inspiriational people I have ever met. She has stood as such a great example for me so many times. She has helped me through some of the darkest times in my life, and been there every step of the way, no matter how difficult or stubborn I got, and I can\'t say that for very many people. I\'ve never found anything to complain about or get annoyed with about Carol, and you have to admit that\'s rare when it comes to your friends because you are with them so often you have to have the little tiny things that drive you crazy.    I love you Carol, and I know you know that I would do anything and everything possible for you if you ever needed it, but I want to say it anyway. You just flat out are a wonderful amazing human being and I can\'t begin to express how much I care for you and your wellbeing and happiness. I do not know where I would be with out you. You once had to tell me to hang on and that things would get better because you would make them. At the time all I needed to hear was that. So hang in there, friend, things will get better, and I will make them. I love you.     urlLink   '),(8700,3900039,'17,July,2004','2004-07-17 00:00:00','   This song is way too perfectly worded... SugarCult- Pretty Girl(The Way)   \" Pretty girl is suffering  while he confesses everything  Pretty soon she\'ll figure out what his intentions were about   And that\'s what you get for falling again    You can never get him out of your head    She\'s beautiful as usual with bruises on her ego  and  Her killer instinct tells her to  beware of evil men   And that\'s what you get for falling again  You can never get him out of your head   It\'s the way that he makes you feel   It\'s the way that he kisses you   It\'s the way that he makes you fall in love   It\'s the way that he makes you cry   It\'s the way that he\'s in your mind   It\'s the way that he&nbsp;makes you fall in love\"  &nbsp;  I barely know this person and he is all I think about. I want it to stop, but I can\'t make it. I don\'t do this, I don\'t fall in love, especially not when it\'s impossible, yet I find myself searching for every possible way to talk to him or see him. It\'s crazy, and it\'s driving me crazy. I tell myself it won\'t work and I should give up and stop, but then the what ifs always come back, and he isn\'t exactly helping my efforts. So far the only red flag is the distance, and I\'ve already made my excuse to ignore that... and for some reason I can\'t ignore all the reasons in the back of my mind that are making me strongly consider it... I\'m crazy... but then something will stand out to me that makes me think&nbsp;I\'m not...  A long time ago, my youth pastor taught a lesson on marriage and dating and all that. It was very good and I remember much of it(sadly I didn\'t think the same at the time). He had us make a list of the qualities we wanted our future spouses to have. Of course some of them were very superficial and immature(as I was in junior high), but most were not. I switched rooms with my parents a while ago and went through tons of old stuff(i\'m a keeper..not a throwing away kind of person), and low and behold I found this list... and the part that scares me the most, this person has a lot of the qualities I listed as a naive thirteen year old. Whether that means at thirteen I knew what I wanted or at almost seventeen I\'m still imature and materialistic, I do not know, but I hope the former. It just creeps me out... I don\'t like the idea that I might be becoming a stupid teenage girl who is infatuated to the point she thinks she might actually... I can\'t even say it... bottom line: I\'m crazy.   '),(8701,3900039,'16,July,2004','2004-07-16 00:00:00','   Nickelback- Feelin\' Way Too Damn Good    I missed you so much that I begged you  to fly and see me  You must\'ve broke down \'cause&nbsp;you finally said that you would  But now that you\'re here  I just feel like I\'m constantly dreaming   \'Cause&nbsp;something\'s gotta go wrong \'cause I\'m feelin\' way too damn good  &nbsp;  For 48 hours I don\'t think that we left my hotel room  Should show you the sights \'cause I\'m sure that I said that I would  We gotta make love just  one last time in the shower   Well something\'s gotta go wrong&nbsp;\'cause I\'m feelin\' way too damn good  &nbsp;  And it\'s like,  every time I turn around I fall in love   And find my heart face down and where it lands is where it should  This time it\'s like the two of us should probably start to fight   \'Cause&nbsp;something\'s gotta go wrong&nbsp;\'cause I\'m feeling way too dam good   &nbsp;Oh, feelin\' way too damn good  &nbsp;  Sometimes I think best if  left in the memory   It\'s  better kept inside than left for good   Lookin\' back each time they tried to tell me  Well something\'s gotta go wrong \'cause I\'m feelin\' way too damn good  &nbsp;  &nbsp;  &nbsp;   '),(8702,3900039,'16,July,2004','2004-07-16 00:00:00','   I\'m babysitting a little girl named Sydney right now. She is sometimes a very fun sweet little child, but 75% of the time she is a hellion. Especially when I make her go to bed, oh god its terrible. Anyway, she is on her computer right now playing Dora the Explorer(who she thinks I look exactly like...) so I am on her dad\'s doing nothing else but blogging. I found this on some other kid\'s blog that also likes SugarCult, my music favorite of the week.... Last cigarette: never, I\'ve been around it plenty but I\'ve never had the urge to Last car ride: drove from my house to Sydney\'s house out on North Bank Road about six hours ago Last kiss: well I wasnt kissed back the other night&nbsp;so that doesn\'t count so.... a while ago Last good cry: ha two days ago, when I lost my license Last library book checked out:Two weeks ago, A Brave New World, very good but very creepy, I had to read it for my AP Government class Last movie seen: in theaters->Shrek 2 with Bryan, very cute(goes for the boy and the movie)&nbsp;&nbsp; at home -> Butterfly Effect, VERY CREEPY Last book read: A Brave New World Last cuss word uttered:&nbsp;shit(I went too fast around a corner on my way out here and almost went off the edge) Last beverage drank: apple juice Last food consumed: fish sticks...its what we had for lunch Last crush: Gabe from Vancouver, and I\'m going to see him next weekend YES Last phone call: called Gabe last night at about 10, he was in a movie, he called back at about 12 and I was doped up again on pm painkillers and barely remember anything we said, he laughed at me alot.... Last TV show watched: oh man, okay I\'m babysitting...we watched Blue\'s Clues Last time showered: 7 am this morning and it felt so good...i just got back from camping Last shoes worn: My Old Navy Zebra Flip flops Last cd played: SugarCult mix Ryan burned for me Last downloaded: Hello, a program that lets me put the pictures on my blog Last annoyance: Collin\'s grumpiness at camping, Jenna\'s laughing at 2 am when camping, waking up unable to move this morning, take your pic Last disappointment: Sydney wouldn\'t let me be the princess when we were playing make believe... Last soda drank: man....this was a few days ago...i have no clue Last key used: my car key  Last sleep: i got home from camping at about 2 pm, i slept from then to about 9 when i made some phone calls, then went back to bed until 12 when i made another, then i got up at 7 Last person u saw: Sydney Last sexual fantasy: uhhhh creepy, blogs are not the place to talk about sex Last weird encounter:&nbsp;Sage,&nbsp;Sydney\'s german shepard puppy,&nbsp;choked on a dog bone a few hours ago.... I\'m scared of dogs and I didn\'t know what to do haha &nbsp;Last ice cream eaten: tasted Christina\'s and Collin\'s cold stone the other night.... Last time wanting to die: i actually don\'t know...a few months ago maybe  Last time hugged: all my friends hugged me goodbye after camping yesterday, collin, carol, jenna, justine, tyler and then pam, in that order Last lipstick used: I havea dark complexion and lip stick does not look good on me, i only use clear gloss and chapstick and i used it this morning Last underwear worn: pink striped string bikini  Last bra worn: plain white Last shirt worn: light blue tank top with srunchy sides idk Last time dancing: thursday in jenna\'s jeep \"your top was always down and the sun was always shining bright\" Last poster looked at: the one I\'m making for Alex, its a Johnny Cash one... Last web page visited: my Hotmail Inbox &nbsp; Well that was fun!   '),(8703,3900039,'16,July,2004','2004-07-16 00:00:00','   ...and it scares the crap out of me. To think this time in only a year(which isnt as long as it seems, considering how fast all of high school went) my friends and I will be saying goodbye and living on our own. I might have to actually learn how to fold my own burritos... The other day I lost my license and had to go to the DMV, my most unfavorite place, and get a new one. My parents also mentioned my registration was up. They wrote me a blank check and told me to go, and I started bawling and wanted them to come with me, which is something I never do. I am used to doing things on my own, and generally prefer it. I acted like a baby and refused to go without them holding my hand. I don\'t want to grow up yet... I love just getting to sleep in when I want and call my friends and go float or randomly go camping. I love&nbsp;being able to spend someone else\'s money and the biggest consequence be having to return whatever I bought. I love only having to make easy decisions like what I want to wear today or&nbsp;whether I\'m going to do my homework or not, not decisions like what bills to pay with this paycheck or which groceries are necessary or how I\'m going to pay for college or who I\'m going to marry or where I\'m going to live. Right now, it feels like my friends are all I have, and I know in a year, I might have to start all over in that department. And it frightens me... As badly as I want to be on my own and have a house and a family and a career, I am scared of the process of getting there. I have an issue with happiness I think... I\'ve gotten to the point where I don\'t really believe in it, just the idea that I can create my own little happiness world where I can pretend everything works. But now I\'m afraid that if I really were happy or I got happiness out of anything, I wouldn\'t be able to recongize it...and I\'d push it away like I\'ve done so many times before.   '),(8704,3900039,'16,July,2004','2004-07-16 00:00:00','   So yesterday I went floating the river with some of my closest friends...Carol, Justine, Jenna, Collin, Tyler..and well Collin\'s brother Jonathon&nbsp;went too. &nbsp;Anyways, we had an awesome time. I haven\'t had so much fun in a while....well minus the back hurting and hives and rock burns&nbsp;and all that... Afterwards we went camping with Justine\'s Mom(remember her?) at Whistler\'s Bend. Courtney and Christina, my other two best girl friends, came out and hung out with us for a while too but they had to work in the morning so they couldnt stay the night. It has been soo long since we were all together at once. So much laughing(well in Jenna\'s case, wheezy and snorting)&nbsp;and talking and rambling and nonsense and worrying(mr. grumpy gills...you know who you are). I seriously love you guys so much I just can\'t even explain! We have to make the most of this last year we have all together... We should mutliply our memories by a gizillion! You\'re the best I could ever ask for! What will I do without you when we leave for college?   '),(8705,3900039,'14,July,2004','2004-07-14 00:00:00','   urlLink     Today I went to the farwell party for a very special person, Moritz Julian Jungmann, a German exchange student. It didn\'t really hit me until I went to hung him goodbye and it\'s wierd to think I might never see him again. Hopefully, that doesn\'t happen. He is an awesome, hilarious person and I love him dearly and will miss him tons. The picture is of Carol, Christina, Mo and Mark at New Years. Come back and visit Mo, and if I\'m ever in Germany, I better see you! We have so many great memories... New Years at Justine\'s, getting pulled over by your house(thats the best one! HA), your belly, dancing, helping you write essays with proper grammar, so so many... I\'m thankful to have had you in my life :D   '),(8706,3900039,'12,July,2004','2004-07-12 00:00:00','   urlLink    Welcome home, Collin! I want to see you first thing tomorrow after my job interview, so hopefully you are awake by then, cause I\'m coming over whether you like it or not! I am so glad you are home safe and that you had a good time. You\'re one of my best friends and you better see me a lot these five days you are home! I know you  obviously  love Mexico more than me, but it\'s something I can live with :D   '),(8707,3900039,'11,July,2004','2004-07-11 00:00:00','   urlLink    WE LOVE YOU SO MUCH KAT! Great job at trials. For everyone else: MY FRIEND IS RANKED 25TH IN THE NATION IN THE 200 M BACKSTROKE!!!!!! And thats out of only 46 other girls who  qualified ! You rock hun and I hope to see you soon.   '),(8708,3900039,'11,July,2004','2004-07-11 00:00:00','   So I hung out with Lance today and we watch Butterfly Effect, it was really really creepy, but overall I think I liked it. We also went to the river and threw rocks in the water to get eachother wet, it was pretty funny. But while I was dropping him off at his house, I got a phone call...FROM LINUS OAKES ABOUT A JOB INTERVIEW!!! YES!! WHOO HOO!!! It\'s been soo long and I haven\'t heard from anyone I\'ve applied to, so this is very good. :D and my dear friend Collin comes home tomorrow, well kinda late tonight, but early tomorrow morning is more like it. He\'s been in Mexico for a long long long time and I missed him alot so I\'m very glad he will be home...for a whole five days...whoo hoo....no. anyways, I\'m hanging out with my computer geek ryan and he is telling me what to type about and he suggested this story....the snake from hell. So while Lance and i were exploring at the river, I stepped on a snake and it jumped out and striked at me and hissed and i freaked out and was hysterical of course....it was bad. And this wasnt no gardner snake, it was LONG AND BROWN! I dont know what kind...which only scares me more. Ryan thinks it was an anaconda...I doubt it :D He is my massively best friend and he just wants to get me to freak out cause im so gullible but its not working this time hehe anyway, hope everyone is having a good day. Okay he just said \"thats it? no life lesson?\" sorry I\'m no jacob bentley. BUt here, kids...dont do drugs, don\'t drink alcohol, and don\'t have sex. thats all folks!(happy ryan?)   '),(8709,3900039,'11,July,2004','2004-07-11 00:00:00','   So last night I played bunko, and it was tons of fun. Mom(aka Pam aka St.joe\'s Youth Pastor aka Justine\'s mom), Justine, Carol, Alex, Seth, Emily, some Chris kid I just met, Trevor, Jenna and I all went out to Applebees afterwards to eat(obviously). Gabe kept text messaging me and I was getting a little too giddy and Carol had to tell me to turn my cell phone noise off haha...whoops. Sorry Carol! I also had quite a blonde moment when Seth, who was sitting right next to me, called me and I talked to him...\"Who is this?\" \"Seth\" \"who?\" \"SETH\" \"I dont know any Seths\" okay it wasnt that bad but you get the picture. I just found out last night that I get to go on the yearbook retreat for sure!!! I\'m really excited! Hopefully Gabe and his friends will go to the \"quest\"(name of the underage club) with us. He hasn\'t sent me the pictures yet grrr... Lance also called and I am hanging out with him today! Should be fun...he is my little art friend, we always had good times in art, so I\'m sure we will today. My cell phone went off soo many times...normally I get excited but I was almost irritated. Carol compared me to a monkey again because I was climbing over chairs to get out to talk on the phone. Well maybe you look like a monkey too!!! no, carol is goregous and she is my bestest friend, i will post a picture of her later hehe. Anyway, that was my life last night... Yes, Ryan, I am trying to make this blog as boring as possible! HAHAHAHA   '),(8710,3900039,'11,July,2004','2004-07-11 00:00:00','   urlLink    SO MUCH FUN TO BE HAD! Yearbook retreat ladies(and boy) is so so soon... We are going to have a blast! Just getting all the girls pumped before we go! Night clubbin\' and shopping and annoying the crap out of Tov, what more could we want? I can hardly wait.   '),(8711,3900039,'11,July,2004','2004-07-11 00:00:00','   Ryan is going to babysit me while I start out on this blog(see comment on first post). Guess that means I might actually have to follow through with it.   '),(8712,3900039,'11,July,2004','2004-07-11 00:00:00','   urlLink    Good Luck at Trials tomorrow Katty! I love you so much! Make us proud, have fun! \"Just keep swimming, just keep swimming...\" You\'re awesome, KT! To anyone else who reads this: this is a good luck note to my dear friend Kathryn Thompson. She will be swimming the 200 meter backstroke tomorrow in the Olympic Trials In California! She\'s amazing. Someday, if not this year, she is gonna follow in Big Bro Chris\' footsteps!   '),(8713,3900039,'11,July,2004','2004-07-11 00:00:00','   So, my parents will finally come home today YEAH! They just called me a few minutes ago. I hate being home alone at night. It\'s been scary, but I\'ve survived. My good friend Justine also just made a blog... I\'m beginning to think Jacob is a trendsetter... before you know it, everyone will be blogging like crazy, but I hope not. I wouldn\'t be able to keep up. So tonight I\'m going to go play bunko with my friends at St. Joe\'s. Should be fun, I love that game. I might stay the night at Heather\'s, but plans for that have not been finalized. Gabe, my new friend from Washington, is supposed to be emailing pictures today of us up there at Forth of July. They are supposed to be super good...we will see. I hope I get to see him soon.   '),(8714,3900039,'11,July,2004','2004-07-11 00:00:00','   So, I have this friend, Ryan, and he\'s the world\'s biggest computer nerd ever, and sadly I let him corrupt my mind into thinking something like having a blog,as sad and pathetic as I think they are, could be fun. So here we go...I have no clue what I am going to put on here.   '),(8715,3900039,'15,August,2004','2004-08-15 00:00:00','   I spend a lot of time  here . This is my  second  family. We are watching the  Olympics . Girls in tight spandex suits doing  the impossible  on little bars and beams, its  crazy . I just said goodbye to Christina, she is going to  camp . My boss is an  idiot , but because he is an idiot, I get tomorrow off work. And  tomorrow  is my  birthday !!!!!!!!!    go go go go shorty, it\'s your birthday, we\'re gonna party like its your birthday, gonna sip (pop) like it\'s your birthday, and you know we dont (care) cause it\'s your birthday'),(8716,3900039,'12,August,2004','2004-08-12 00:00:00','   I\'m at  ryan\'s , he just fixed my  laptop . my baby is back! YES! thanks ry, you are the  best . i can\'t  breathe  very well right now, its wierd. im like hypervenilating ha. we named ryan\'s  fisheys , but one is dying as we speak. we are watching him carefully. :( its sad,  spazy mcgee  was/is my favorite! at least we still have  jonah  and um  blow (aka leonard). We will get new ones for ryan\'s birthday. fish are easily replaceable. but  spazy just came back from the dead!  he is freaking out, its like every ten minutes he  dies  and then ten minutes later he\'s  alive ! i went  swimming  today with justin and his friend marco, it was fun. i threw a  rock  at justin cause he pushed me in. it was the same spot that i went with lance, but this time i wasnt attacked by any  snakes ! (phew) but some fish tried to bite me. my  birthday  is so soon. for anyone who reads this, my house  friday,august 20th, around 8pm , we\'re having a little get together with  ice cream  and  brownies  and  homemade coldstone  and all that good stuff. so  you are invited ! but call me first, so i dont have a million unexpected friends... just got word from ryan: Spazy McGee has officially finally passed away. :\'(   '),(8717,3900039,'11,August,2004','2004-08-11 00:00:00','   The other day Ryan and I attempted to take some pictures with his camera, these are the results....                                Not so good.....but heck, we tried! Ryan\'s eyes were having problems adjusting to the flash, and i just went a little overboard with my smile. Ever since I got my braces off, I\'ve had this uncontrolled smile and I just freak out.   '),(8718,3900039,'10,August,2004','2004-08-10 00:00:00','   Love Soon           John Mayer   I know you\'ve been sworn  I read your complaint  You\'re needing someone older  And though I\'ve been warned to live day by day   There\'s something taking over  Did you expect to kiss me one time While looking at me with the same eyes ever again So come on and face it  So come on and face it   It\'s time that we say it    You can cross the line whenever you want to   I\'m calling it love soon   Close your mind and waste some time if you have to  I\'m calling it love soon  It\'s not about you now   It\'s what we are   You\'re mother complains that you need a man  You haven\'t mentioned me yet  And all of your friends don\'t know who I am  I\'ve been your best kept secret   I understand I wasn\'t part of the plan  A dollar short, a minute early But I am your man  Let\'s bypass the bullshit and move on because  The minute hand moves faster than you think it does   And by no fault of yours and by no fault of mine   The bottom line is lyring in the bed that we\'ve been playing in tonight   '),(8719,3900039,'06,August,2004','2004-08-06 00:00:00','   Well now that he\'s been on the other side of the world for a while, it\'s been a lot easier to not think about him as much, or to think about the \"situation\" in a negative way, and to think of someone else in a positive way. Another impossible one.... I hate this.   '),(8720,3900039,'06,August,2004','2004-08-06 00:00:00','   you said you wanted pictures... this is the best i can do for now, they arent very recent, and i am really bored, but im sure you don\'t care. i hope you are having a blast in australia! don\'t do anything too stupid... haha! you\'re the greatest!  urlLink    urlLink     urlLink   '),(8721,3900039,'04,August,2004','2004-08-04 00:00:00','   urlLink    I\'m getting excited again for school....   '),(8722,3900039,'03,August,2004','2004-08-03 00:00:00','   Yellowcard       Ocean Avenue     There\'s a place off Ocean Avenue Where I used to sit and talk with you  We were both 16 and it felt so right   Sleeping all day, staying up all night  Staying up all night  There\'s a place on the corner of Cherry Street We would walk on the beach in our bare feet We were both 18 and it felt so right  Sleeping all day, staying up all night  Staying up all night   If I could find you now things would get better   We could leave this town and run forever   I know somewhere, somehow we\'ll be together   Let your waves crash down on me and take me away    There\'s a piece of you that\'s here with me   It\'s everywhere I go, it\'s everything I see   When I sleep, I dream and it gets me by   I can make believe that you\'re here tonight  That you\'re here tonight  If I could find you now things would get better We could leave this town and run forever  I know somewhere, somehow we\'ll be together  Let your waves crash down on me and take me away  I remember the look in your eyes When I told you that this was goodbye You were begging me not tonight Not here, not now  We\'re looking up at the same night sky   And keep pretending the sun will not rise   We\'ll be together for one more night   Somewhere, somehow    If I could find you now things would get better   We could leave this town and run forever  I know somewhere, somehow we\'ll be together Let your waves crash down on me and take me away   '),(8723,3900039,'03,August,2004','2004-08-03 00:00:00','   Only 13 more days until I\'m seventeen freakin\' years old!!! I\'m really excited. For my birthday I\'ve decided to ask for: scrapbook stuff(for a senior scrapbook), vases and boxes to decoupage(for myself for once), pretty much anything from Michael\'s craft store, and for Gabe to come down for a few days. Okay wishful thinking on the last one, but I\'m allowed to dream. Gabe sent me an email :D :D he misses me, I\'m excited. Anyway, my friend Carol is in New York, and the terror alert there is red now... and I\'m a little paranoid. Kinda creepy, one friend is across the world and the other across the country and we have a red terror alert, not good. But my friend Collin is now home safe and sound for good. I\'ve spent the past two nights doing nothing at his house, it\'s always good. Music in the park is tonight and I\'m going with the girls. Alex maybe too. I\'m not really in the mood to go yet tho... Hopefully later, we will see. The scabs on my feet are all better, I\'m really glad. I have some beautiful scars instead tho. Justine got pictures back from the yearbook retreat and Grafitti. They are really cute! I\'ll post them when I get a chance. I love my friends, we are so cute! HA, Jonathon(collin\'s brother) told us to use better adjectives...We said cute every two seconds about every picture... it was bad. So we went with adorable and pretty and  goregous and fun and gross and ugly and I think we threw in whorish a few times... I watched Kill Bill with the SCs.... bloodiest movie ever, it was disgusting! And then it just freaking ends with no real ending!!!! grrr.... stupid movies I\'m done.   '),(8724,3375539,'29,May,2004','2004-05-29 00:00:00','    hmm yea.. i think i better say something out her.. haha.. yea.. my piggy~ n its... IRENE.. ah~ oh why her?? dunno leh.. somehow i got dis feeling dat shes really interested in me.. yea.. cos she whole day sms me de.. even on small things.. n hey.. she touched me.. haha.. cos she like saw me kinda sian one morning.. n when i was at lecture.. i received her sms.. it goes like this.. \" dun be so sian..SMILE..must be glas u see a cute girl in de morning..cheer up!\" well hey.. she only just got a glance of me lor as i walked pass her.. ah.. dats kinda sweet i tot.. haha.. n den yea.. we begin to exchange sms.. n dunno y.. somehow i will bump into her in sch more often these days.. yea.. i really wonder if she likes me.. cos its she who initiated all de smses.. n oso she always approach me when she saw me online.. den its like she will always inform me bout de things dats gonna happen in PS before hand.. n she took de initiative to call me.. to my house.. haha.. but not much.. chat a lil.. but mostly on PS stuff oso.. was it dat i too sensitive? i somehow thinks she cares alot bout me.. or was it dat she wans me to be de president or sth.. trying to noe me more?? hmm dunno leh.. shes making me thinking of her so often now.. yea.. shes J2, a yr older den me.. but does it really matter? i think shes not dat bad looking la.. n shes v successful to me.. can do well in sports studies n even leadership roles.. a v fine gal.. let fate decide if we r together ba.. haha.. cos in my mind i can still remem SL..  i am greatful to her.. for all her support..n i do like her.. :0 '),(8725,3375539,'24,May,2004','2004-05-24 00:00:00','    oh yea.. its hols le.. FINALLY~ but haiz.. dats it.. we all gonna die.. its the last hols dat every1 will have b4 they will ever stop mugging again.. yea.. it suck.,, its not a hols lor.. every1 is like still going back to sch for makeup lessons n or camps.. n den sumore.. alot hw! so many tutorials to catch up on.. its like shit lor..im still feeling so strssful even tho its hols! Den last few weeks have to go chiong study,, for the common test.. i think i better work hard n study manz.. my grade suckED.. n i so dissapointed in myself over maths.. I GOT A \"F\"~ could u believe it? suck man.. i think i am 1 of de lousiest in class.. dunno y leh..My fren whom i hang arnd wif got ALL As lor..wth.. I got BCOF.. all my grades.. B for chem, C for bio(amazingly) , O for phys, den ...dun tok bout it le.. haiz.. i gotta buck up man.. Hols oso need to increase my flights bookings man.. if not lose out.. haha.. but damn shitty.. i feel so stressed when i fly.. gotta study my theory.. n all those stress every1 gives me is making me crazy.. all asking \"have u got ur license?\" dun be crazy la.. dey think its so easy to get meh????? haiz.. but i will endure n work hard de.. trust me.. even tho i got de feeling i cant get into ppl course.. i cannot give up.. still got chance lor..ahha.. anyway.. tho i got alot to say.. but hmm... cut short.. no use spending time saying all these.. actions speaks louder than words! '),(8726,3375539,'24,May,2004','2004-05-24 00:00:00','    woo wow yes.. i slacked today.. forget about doing any hw, n haha nxt wk no tests.. yea~ dis afternoon, like i said at 2.30 am, i have to fly.. hmm haha.. sian.. cant seem to remem my flight checks leh.. but oso i nv put in de effort.. slpt again jus now..haha.. de instructor says more improvement again.. sianz la..haha.. but on my return journey to hm.. i saw funny things.. haha.. first i saw \"paul\", de guy frm \"under one roof\", haha.. he changed so much.. he now got thick moustache n goatie.. yah.. haha so funny.. den at de bball court, i saw a lao ah peh playing bball..haha.. he still can bounce de ball well.. but when he shoots v funny leh.. but amazingly.. in 10 shots, he still can score 1 la..haha... anyway.. gd day.. '),(8727,3375539,'23,May,2004','2004-05-23 00:00:00','    wahaha~ its 2.30am now lorz.. dunno wat happen to me.. de more stress i am, de more i sleep.. haha..yea dats wat happened.. i slept at 9 just now mah.. den woke up at 2.. ahha.. now my biological clock going crazy le.. i always cant say i am deproved of my sleep.. cos i always sleep alot de.. after doing go essay on \"why sleep\", haha i learnt to sleep even more~ :P.. but slping quite sucky oso.. i hate it when i slp too much~ haha..sometimes slp so long oso no use de.. end up u getting more tired.. n oso dun even have a nice dream.. wats de point? slping eats up de time too..haiz..but jc life got shortage of time leh..n den increase in workload.. so perhaps i shuld sacrifice some of my slping time? yea?? haha.. so ive concluded.. (for tonight :P).. i shall study my flight theory.. haha tmr late morning gotta fly~ weeee~ dunno y oso.. flying supposed to be fun and enjoying.. but dey damn sad de.. wan instructor to gif a grade at de end of de flight.. so cannot dun study..T_T.. last of all.. a brief summary of wat happen today.. my gp test i only got time write 3 content paras, maths test can only like do half de qnts i bet i gonna fail them.. but heck care.. i must begin to get numb with failing.. haha '),(8728,3375539,'21,May,2004','2004-05-21 00:00:00','    wth.. abit cock up today.. morning chem SPA dunno y leh.. sudenly cannot think, den cannot solve de skill c problem.. but heck la.. i m beginning to find myself weary n tired.. theres no joy..BAH~ hate de life im having now.. but somehow wat i tot was dis was de kinda life i wanna live..aiyah i toking crap la..need some sense knocking man.. i going bonkers.. tmr still got gp and maths test.. face it.. its hard even to love urself.. when u dun enjoy wat ur doing.. :( '),(8729,3375539,'18,May,2004','2004-05-18 00:00:00','    ah... guess wat? i am still doing my games proposal for P.S...or photographic society(Pang.Sai :P)haha.. sian man.. but aiyah nvm la.. dats life.. i think i wasnt in de right mind yesterday.. got headache leh.. maybe becos of de vitamin pills i taking..dunno eat got wat use leh..humph.. cos yesterday hor.. got p.s. interview.. haha.. i dunno wat i said..but i think i said i just wan to be member.. haha.. but not true actually..hehe.. who dun wan to be exco.. where got ppl so stupid.. i only modest nia mah.. ahah.. but nvm la.. Irene ( the chio p.s president) say still got chance.. perform in de p.s camp.. yah.. i hope i will la..But anyway.. got kinda stress dis days again.. cos i saturday got GP test and Maths Test.. which i haven study..:( Den tomorrow got chem SPA, hope i dun cock up man.. lastly, i oso wanna pray i dun tio flights on saturday lor.. if not i will die like hell lo.. haiz.. sumore today i menstruate again~! yikes..(take note** menstruate = bleeding of de asshole.. HAHA)is it becos of de pills again?! hmmm... today oso not a very happy day la.. but at least i noe i passed my overall bio common tests by..... 0.5 marks(33.5/60) i failed de 2nd test by half mark.. lucky man.. i didnt study for it.. n i still manage to score.. haha..I Tian Zi Gao la..^.^ '),(8730,3375539,'18,May,2004','2004-05-18 00:00:00','    hehehehe~ i dunnowat i\'m supposed to do.. haha.. nv tot i would be so free to start a blog man.. hmm, well started becos someone reminded me dat she had 1..(^.^) haha.. dun have much to type about.. actually i dun quite like ppl to see my \"diary\".. tot it wuld be like no privacy.. but then.. nvm la.. at least now i noe here can be a place where i would say those XXXXX stuff n scold XXXXX words when i m angry or upset.. hopefully it improves my english too..ahaha.. i damn gei kiang de la.. everything oso wan.. so kiasu.. but no choice.. Blame The Society~ haha~ dats it.. no time la.. haha '),(8731,3375539,'30,June,2004','2004-06-30 00:00:00','    haha.. came here to show dat im a slacker.. :p yea.. shucks mans.. de maths test v hard leh.. den i dunno y.. how come de rest like find it normal.. HUmPH..i cant do like almost half de paper.. sianz.. den i scared i got alot careless mistakes.. yea.. i estimate bout 40/70.. hope not too much mistakes.. or i will fail again~! ARGH.. i dun wan.. haiz.. den me begin to worried liao.. if continue lidat den how to be promoted?? i wan retain 4 As leh.. haiz.. nvm.. dun care.. i think long term plan needed.. i finish wif exams den plan.. haha.. hmm i think finish exam still got alot hols hw gotta do! yikes.. haha..forget it.. tmr got chm test.. i PRAY its not as hard lor.. haiz.. dn gif me those long long structured.. especially on those chem bonding etc.. yea.. i will die.. but still haiz.. y maths so tough?? i read back on my pri sch testimonial.. dey wrote.. \"KianHong has a exceptional capability in maths.. he should furthur develop on this talent..\" -_-\"\" haiz.. wth..now i like dying wif maths.. lucky i now myself.. nv go take F maths. if not die until siao.. but last nite oso slack.. watch alot tv.. haha.. top fun.. got jolin.. den 9 pm show got de TianSong go dig his ear till bleed.. den de gal confess to Hu YiDao.. den 11pm got Lee Kuan Yew interview.. haha.. all dis crap i noe lo..sianz.. im such a slacker.. den i keep awake until 3 am.. den slp wake at 5am.. den study.. till 6.30am.. den go slp until 9am. haha.. de pig tot i was awake all nite.. crazy.. haiz.. now wat im gonna do?? hmm..do short term stuff first.. better go study chem la.. bye~ '),(8732,3375539,'29,June,2004','2004-06-29 00:00:00','    haha...sad to say.. but im a serios slacker.. haha.. me still come online despite de exams.. hehe.. yea.. today got bio den gp.. haha.. both die la.. Bio siao liao lo.. de essay qts 20 marks den i dunno how to write lo.. ppl all write long long.. me short short.. den de gp essay i crap de.. haha.. all de examples i fake out lor.. aha.. think de marks come back oso fake de.. hehe..yupz..hmm.. i tot tonite go netherlands match.. but dun hav.. haha.. nvm.. me v tired.. still go slp la.. '),(8733,3375539,'28,June,2004','2004-06-28 00:00:00','    hmm.. to day last day of hols le.. haiz.. so fast.. rly dunno wat i hav done for de last 4 wks.. rly nth lor.. haiz.. haven even play.. sux.. haiz..nvm..dun care le.. jus go ba.. tmr is bio n gp exam le.. haha.. confirm fail ah.. or.. i will fail all subjects actually.. cos i rly dunno wat i have been studying lor.. i jus simply read thru.. n den information comes out from my brain de next minute.. haha.. nth inside leh.. nvm la.. i noe i will work hard after de exams..haha.. hope so.. yea.. so nxt wk exams.. hmm.. not gg to come online le.. yah.. take care every1.. yea.. yesterday netherlands won.. haha..yippie.. yea.. win 5 - 4 by penalty.. they used to lost n kena kicked out of euro n world cup due to penalties.. but they won now... haha.. heard dey hav trained hard on it.. den de e van der sar so pro.. haha.. so calm.. den when he saved de ball.. wah.. he stared at de player lor.. haha.. so scary.. but damn cool..yes.. so nxt match is netherlands vs portugal.. haha.. hope portugal lose lo.. yea.. lose in their home ground.. den they lose face.. haha.. cos netherlands rly alot stars lor.. even if any of them injured, dey still got pro players on de bench lor.. yah.. look at de no. of strikers dey hav.. besides nistelrooy dey still got 3 others on de bench lor.. haha.. yup.. hope dey win.. den jus now my mp3 player played trademark-only love.. ahah.. den part of the lyrics was..\"tt\'s sth only love can do\".. haha.. den i dream lor.. dream of rly wat is dat sth only love will do.. den i tot of sth in like 5 secs.. haha..dat is walking some1 home..yea.. i guess most frens wuldnt purposely see u home wuld dey? ahha..yupz.. i noe theres many more other thgs.. but den dun mention le.. haha.. waste time.. i go study la.. still got alot to go.. -_- '),(8734,3375539,'27,June,2004','2004-06-27 00:00:00','    hey.. thks u gals for de concern.. but rly.. im alrite.. haha.. y all say.. dun think too much.. duh.. N netherlands will win tonite~! '),(8735,3375539,'26,June,2004','2004-06-26 00:00:00','    hmm.. me out of yfc liaoz.. duh..haha..seems rather fast..left jus lidat.. leaving without anything..sad lorz.. haiz.. now like dunno wat to do le.. suddenly v empty..nvm la.. perhaps used to it after awhile.. im pretty glad.. cos until now everybody like supporting me to leave yfc..haha.. but still.. like to thks SL.. shes been there cheering all de while.. haha.. really glad.. den seems like pig finally understood.. she oso encouraged me last nite.. n den consoled me after i told her i quit.. yea.. thks.. den to all my other frens whom i hav consulted.. thks again.. de decision of quitting is jus made by me.. i nv tell my parents nor my siblings yet lor.. dun intend to tell them.. until dey found out themselves la.. see.. they dun even noe a thg bout me.. haha.. my mama de worst.. whole day at hm wif me.. but nv seems interested in knowing anythg bout me.. all de while.. is me study on myself alone.. she nv nags me to do hw.. haha.. dats y la.. but still.. her method seems to work..haha.. cos she has produced talents like my sis n me :P haha.. (my bro can dun mention..such a failure..so old still play com games all day).. yea.. so de news of me quitting yfc still not many ppl noe yet la.. haha.. i think my 2 other aj frens in my course oso quitting..haha.. sianz.. den left 3 other guy in de next course.. wish dem gd luck la.. haha..if dey quit oso.. den WAH.. AJ YFC gonna destroyed in our hands.. haha.. well reminded me.. i hope AJ PS not destroyed by me lo.. haiz.. i dunno how to lead den dey still choose me.. all crazy de la.. haha.. lastly.. i ought to mention my flight instructor in yfc.. yea.. its a SHE.. haha.. de only female pilot instructor in yfc..ya.. so cool.. n shes like a mother..haha.. not fierce de lor.. den she will always gif philosophy lessons.. ahah.. den today 1 is of course de longest la.. cos she wun have any chance to say anymore.. yea.. n honestly.. there wasnt much chance for her to share her wisdom.. cos theres rly even no time for these in yfc.. haha.. yea.. n so she taught me a few lessons in life la.. but then.. haha.. dun rly remem dem well.. :P haha.. but then today 1 i still remember la.. she told me she just started to go for chinese lessons again..haha..even at her ermm quite old age.. yea..she say she might even consider staying in china for a yr to help her in chinese.. yea.. so that she can communicate with her daughter in chinese.. ahha.. ya.. so shes telling me.. theres no boundaries to learning la.. den she say tho she dunno much chinese words.. but she says de word \"JUE\" \"è§‰\" or AWARENESS is v impt.. haha.. yea.. den she say with it.. u can simply master everything.. yea.. n i think its true.. haha.. den she tell me how to master dis skill of Awareness.. its to COUNT THE NO. OF BREATHS... yea.. u wuld realise.. how easy it is for u to get distracted.. haha.. n yes.. i realize.. its impossible to count it.. it takes de whole life.. haha.. so well jus practise next time.. perhaps while u are running?? u can count de no. of breaths u took.. haha.. ya.. crazy lor.. well..hmm can try still.. cos i believe her.. awareness is important..ok..alrite.. dats it.. its de end of my jc aviation career.. rly hope to have another opportunity de nxt time after i finish studies.. yea.. i will miss yfc.. it marks de end of a chapter.. haiz.. i think i better end it wif a smile :) '),(8736,3375539,'26,June,2004','2004-06-26 00:00:00','    hmm..bless me..i tio flight tmr.. hmm.. might be de last.. im scared.. yes definitely.. rly wish it wasnt de last.. and wats more.. haha.. im gonna sacrifice time to study for flight rather den my exams.. yea.. i haf to gif it a last chance.. i promised myself.. yea.. so i will do it.. gotta be confident..i lacked dat.. hmm relax.. haha.. argh.. rly god bless..   forgot mention yesterday.. ahah.. went wdlands to pass my photos for competition.. i ask dat pig accompany me go together.. den she dun wan.. cos of de stupid spare tyre reason.. yea.. den nvm lo.. i go alone.. she go alone.. she like wan to avoid me lo.. haha.. den when i left wdlands for hm.. i stopped at amk mrt.. and so coincidental.. shes at amk mrt oso.. she was jus about to pass de gate.. while me oso.. and in opposite direction.. haha.. was it fated? lol.. but who cares.. its jus coincidental.. '),(8737,3375539,'24,June,2004','2004-06-24 00:00:00','    aAHH.. Lose money.. sianz.. shuld hav known.. me bad bad bad luck dis yr.. so suay still wan go bet soccer.. in de end england lose.. sianz.. haiz.. den sumore they waste soo much time.. hav to settle wif penalties.. den de beckham cuckoo head.. idiot la.. den de coach oso stupid.. after de 3rd min goal.. dey turned so defensive.. cuckoooooo.. playing offense is de best defense lor.. cuckoooo.. hmm..den dis morning cant rly study.. sneezing n den keep falling aslp.. ya.. haiz.. den previous few nites.. pig odd odd de..she noes i like SL la..haha.. cos i gave a string of nos... 744546 769 4 5683 968.. haha.. cos unhappy wif pig ma..nvm.. den she begin to blabber nonsense.. wat dun wan be spare tyre.. haha.. dunno wat to do oso la.. let her be.. i free le den settle wif her.. haha.. stupid gal.. keep thinking she v clever.. '),(8738,3375539,'24,June,2004','2004-06-24 00:00:00','    hmm.. nv blog last few nites..well..today nv rly study.. cos went out.. go chinatown collect pics.. den go bugis collect my repaired phone.. yupz..den i saw a familiar shadow at bugis.. haha.. ITS AMIN..AHH.. yah.. hes smoking..n..perhaps more shuai n cool..haha..so u noe whos he??? my pri sch fren.. perhaps not fren.. cos hes CRAZY.. sat beside him for almost a yr i guess.. no other guy could endure him de.. haha.. hes disgusting.. he licks ppls hand.. den likes to copy my work.. haha.. its scary.. but then thks to him.. if not i think i wun b so mature.. its a \"nice\" experience anyway.. haha.. today slack manz.. cos me take videos of myself..crazy me.. haha.. v ugly.. n SL says its spastic.. lol.. quite true la.. bah.. take for fun de.. cos boh liaoz ma.. den listened to songs.. yep.. keeps me relax lo..haha.. den i love my mei~~see leng.. cos she sayang me.. haha..thks for her concern.. n hope her leg recovers too.. n SL leg oso.. seems quite long le..haha... okiez..hmm.. me gg off..perhaps chiong.. i hope so.. dun fall aslp like yest..:P haha '),(8739,3375539,'21,June,2004','2004-06-21 00:00:00','    omg.. im fucking suay.. fuck fuck fuck..im sry but i only got dis word..not dat i love to swear.. but its JUs FUCK~ my fone spoil.. for no apparent reason..hmm.. dis whole yr lidat lor.. i got nth to say manz..if god wans to play me out.. go on.. i v tired le.. i dun wan to fight back anymore.. fuck me up man.. cmon.. I fucking hope tmr will be a fucking gd day..n yes. i still fucking sad when i see ppl catching up on flights.. fuck fuck.. '),(8740,3375539,'17,June,2004','2004-06-17 00:00:00','    dunno y so fan again.. didnt go out take pics cos my fren john kena grounded.. duh.. den make me lose all de enthusiasm.. yea.. so sian now.. wanted to really take gd pictures n send in for competition de.. but haiz.. now i make do wif wat i took yesterday at chinatown la.. veryvery sian.. cos i really cant study~! haiz haiz.. i wanna lock myself up in de room.. but then.. i think i still cant concentrate.. firstly cos of yfc.. haiz.. den de pig v irritating la..dun wan tok bout her..if u wan compare.. shilin really understands me more.. haha.. sianz.. even now de bak zhang taste so bland to me.. dun feel like eating le.. sumore de stupid euro2004.. all fixed match lor.. all de bookies fault.. dey make my netherlands lose.. wtf.. SIAN ah.. God  Bless Me Leh.. at least make me concentrate on my studies can? HAIZ '),(8741,3375539,'17,June,2004','2004-06-17 00:00:00','    hmm..gotta think again.. sucks.. ya.. i think i come to dis conclusion..dat if i didnt perform well nxt flight.. i will quit yfc.. no point lor.. im having a terrible time.. i ought to stop my gei kiangness.. haha.. yupz.. my grades r dropping.. den always get worried wif de theory of yfc.. sucks.. i rly think its i no talent la.. den i learn things v slow.. mayb rly not meant for me.. but dun worry.. i will still work hard for it in future.. i not gg to giv up so easily lor.. yah.. actually i rly am sad to hav to make dis kinda decision.. but its like i dun wan de lor.. but i think i hav to.. many ppl advice me to oso.. afterall.. i think i still got PS to backup rite n as a pres... ought to put effort in it ma.. haha.. i dun hav much choice la.. have to quit yfc for my own benefit..   hmm oh no.. i think my geikiangness going to act again.. rly feel like joining Overseas Community Service Expedition - Journey of Sharez 04. yupz.. only 25 lucky youths gg to get selected.. n hmm.. was i 1 of them? cos its like i knew it thru eugene ng.. de chem 1st 3 mths tcher.. he sent email to onli 4 person.. n i was 1 of them.. wow.. ahha.. was i lucky or unlucky.. my heart like itchy to go leh.. haha.. i think i decide abit 1st la..haha..  den lastly.. shilin..these few days didnt rly tok to ya.. will miss ya de lor..n rly.. wanna thks u.. cos its true dat sumhow.. i feel dat no one encourages me like de way u did..no one supports me de way u did.. no one cheers me like de way u did.. no one consoles me de way u did..a BIG THANKS.. tho u did nth much but say a simple \"jia you\".. but its like i appreciate it v much.. cos sumtimes..other ppl dun even bother to say.. yupz.. i think u got fever now.. fall sick le. rest well.. hope to see u lively again..take care.. god bless ya.. i\'m praying for ya.. '),(8742,3375539,'16,June,2004','2004-06-16 00:00:00','    great.. i lost my sense of direction.. dunno wat im supposed to do.. cant i jus lie in de bed all day? duh~ going down to sch later do PW.. fuk it.. i jus wanna crap my way thru.. perhaps later after pw i might jus as well go for a jog.. need some relieve.. argh.. my gei kiangness make me call up yfc and booked my whole wed n fri.. well i do hope to get flights so dat i dun lag too much.. but if i get flight i oso get stressed.. cos i noe i got alot of theory to read up on.. fuk it.. de world is spinning.. i think i needa meditate.. if i dun chiong my studies i might as well end up in hell..fuk off.. '),(8743,3375539,'15,June,2004','2004-06-15 00:00:00','    yah.. sad..cos today nv go out..dey cancelled de outing.. sad.. cos today v lonely.. sad.. cos early afternoon msg shilin..she nv reply.. sad.. cos i msg shilin at nite..but sumhow i got dis feeling she didnt seem to be in a gd mood..sad..cos pig isnt online.. sad.. cos every1 started studying..sad.. cos i got alot of things nv do.. sad.. cos pw is another headache.. sad.. cos i think my yfc starting to die off le..more than 2 wks nv fly..seems to hav a flightophobia..sad..cos my fren leaving sg le.. sad.. sad.. sad.. T_T '),(8744,3375539,'14,June,2004','2004-06-14 00:00:00','    æœ‰ä¸€å¤©ï¼Œå°æ˜Žçœ‹åˆ°å°ä¿å¾ˆå°å¿ƒåœ°æŠŠä¸€ä¸ª æ€ªç¯® æ”¶èµ·æ¥ å°æ˜Žé—®å€’â€œ åˆšä½ æ‹¿ çš„ä»€ä¹ˆç¯®ï¼Ÿâ€ å°ä¿ç­”åˆ°â€œ æ´—ç±³ç¯® ï¼â€ â€œ æ´—ç±³ç¯® ï¼Ÿâ€ â€œå¯¹å•Šï¼ç”¨è¿™æ´—ç±³ï¼Œ è¶…çº§ç™½  è¶…çº§å¹²å‡€ï¼â€ â€œçœŸçš„å—ï¼Ÿè¿™ æ€ªç¯® çœŸçš„è¿™ä¹ˆåŽ‰å®³ï¼Ÿä»€ä¹ˆç±³éƒ½èƒ½æ´—å—ï¼Ÿâ€ â€œå“Žå‘€ï¼Œå¤§ç±³å°ç±³ç™½ç±³é»‘ç±³ï¼Œ æ´—ç±³ç¯® éƒ½èƒ½æ´—ï¼â€ â€œè¿™ ç¯®å« ä»€ä¹ˆç¯®â€ â€œè¿™ ç¯®å«  æ´—ç±³ç¯® !â€ '),(8745,3375539,'14,June,2004','2004-06-14 00:00:00','    yupz.. after addressing to u.. den i tell u wat happened last nite lor..i went rest at ard 2 rite? yupz.. den rly buay tahan.. so i wake up n switch off de com..haha..den orh orh kun le..promised to msg pig at 2 but i didnt..ha.. den i woke up at 4.. Oopx..ahha.. didnt even tell yc,ks,ken,alvin,jing dey all i not gg.. hahaz.. but de first thing i do when i wake up was to msg pig.. yupz.. shes awake.. haha.. amazing..den she became my live telecast of de match.. at first was england1-france0, but ever since i told her france stronger n dat beckham sux.. it all came true.. haha.. de v last min..beckham missed his penalty n zidane scored.. den zidane scored de penalty again.. haha. den france won.. so tyco.. all last min..but still she didnt complain dat england lost..haha.. yupz.. i think she oso not a crazy fan la.. so who cares whether dey win notx..yupz.. so went to slp arnd 5.. btw.. i rly notice me here dese few blocks v noisy leh.. soo many ppl watching soccer den de \"ahs n ohs\" all come out.. haha.. interesting.. '),(8746,3375539,'14,June,2004','2004-06-14 00:00:00','    yupz.. jus wanted to say sry.. rly feel ashamed..i rly shudnt possibly do dis to u.. i noe la.. u dun wan accept me.. i oso bo bian mahz.. dats y i can so transparent tell u everything..but definitely.. ur a wonderful gal..dats y i still continue missing u..ya.. jus dun stop me..nth wrong de ma.. u jus continue stay cool n funky can le.. see u happy i oso happy mahz '),(8747,3375539,'14,June,2004','2004-06-14 00:00:00','    now is 1 am.. yahz.. tired manz.. jus now pig keep msging me lor.. yahz.. den she keep on bhb..den crap alot la.. haha.. ya..den she say she wanna join photo competition..yupz..think i shuld join oso.. no harm trying de mah.. sekali i win.. den hiak hiak.. u ppl throw face ah.. yes.. den she say if i win den i wun be No. 3.. upgrade.. lame.. today i did nth much los.. do de chinese at IVLE again.. yupz.. i so called chiong finish le ba.. de rest i dun intend do.. den.. i oso spend time playing com n use adobe to edit some old pics.. yupz.. kinda into photog dese days.. but still.. i miss my yfc leh.. i haven fly for 2 wks.. kinda worried.. but anyway.. i cant fly well lorz.. haiz.. nvm la.. live some happy days 1st.. haha.. den yupz at nite when hougang mall eat wif my parents.. dey quite old le leh.. so i mus accompany dem la.. my bro n sis always not arnd de.. den yea.. shilin is back.. haha.. she must have suffered during de famine camp.. say she become v dirty..den stinko.. haha.. oso 30 hrs onli eat 2 chocs.. congrats ya.. u made it.. u survived..haha.. great achievement ya.. but i think to me chicken feet la.. i these days keep eating oily food n those fast fd lorz..  think become fat again.. current weight 57 1 kg lesser den starting of yr.. haha.. mus work till 55 kg la.. den i think its most ideal.. since pig got rank ah.. i think i mus act act a lil oso ya..haha.. but rly.. shilin n pig both my no.1s..hehe.. cos pig is sweet lor.. den shilin is vv nice oso.. later i gg to S11 watch soccer.. france vs england.. muahaha.. so cool.. guys gathering..hope i dun fall aslp ah.. hahaz.. den tmr got class 10 outing.. can see my frens again!! ya.. den b4 meeting dem mus go sch.. cos we prepared a shirt for grace.. de gal in my pw grp.. whose leaving sg for a scholarship in UK.. sad sad sad.. i lost a talent in my pw grp.. n lost a cute fren oso la.. yes.. shes cute cos shes always interested by de hdb flats n stuff.. hmm ya.. soo tired.. alot of things.. ya.. take care.. i mean i shuld take care.. '),(8748,3375539,'13,June,2004','2004-06-13 00:00:00','    \'sup manz~ haha.. today montage.. yupz.. still dunno wats de meaning actually.. haha.. but nvm rite? yea.. kinda stupid wan leh.. takes up me 1 whole day.. den find myself didnt get much things.. except abit bonding wif J2s and andrew(exps senior).... den onli end up getting de vinegar smell.. haha.. yupz.. yea? my dear shilin..dat irene rank me 3 sia.. bhb gal..yea.. she rank 2 other J2 guys b4 me yea... haha.. so har.. sry.. nobody wan me still :P yah..so jus now montage.. she was sitting in between de 2 guys.. LOL..nv seen a gal like her..so duh~ u noe.. can onli imagine guys flirting rite..den she like zuo yong you bao..haha..den go toilet rite.. she mus go wif them de.. yah.. so yucks..den wonder when my turn .. haha.. but no chance la.. me 3rd niaz rite? haha.. so i was msging some other gal throuhout de seminar.. yupz.. i msging yee rae.. long time no msg her le.. cos lost her no. den kangli(oso in PS) tell me she in VJ PS mah.. haha.. so msg her ask her whether shes here not.. yupz..den she not here.. so continue Liao Luo Gan Qing.. haha.. yupz..den yea.. shilin not around tonite.. Lonely lorz.. haha.. yupx miss her..shes at famine camp..dunno wat she doing now.. maybe feeling v hungry.. haha.. yesh.. so dats it for today.. tata '),(8749,3375539,'12,June,2004','2004-06-12 00:00:00','    just add in 1 more comment.. pig msg me juz now to apologise.. yupz.. thks.. haha.. at least make me noe u got think of me ya.. :) '),(8750,3375539,'04,June,2004','2004-06-04 00:00:00','    yupz.. i wrote alot below ya? haha.. i didnt noe.. today me kinda talkative perhaps.. den yahz.. i wanna mention.. i really like being honest le.. its becoming a virtue dat i like.. yesh.. I mus say the quote i write myself\' Heart Tells The Truth.. I really wish i can need not say anything but felt through the heart\' yea.. cos its like.. cmon la.. wun u jus feel it? hmm yea.. den yah.. reminds me of pri sch again.. de honesty bk mark i got.. ( earned by returning money,wallets to general office) .. yupz i still remember.. on it wrote \' Honesty Is The Best Policy\" yea.. its so true.. haha.. dunno y only now den realize.. ya.. cmon.. lets be honest.. i dun like secrets k? haha.. or cmon let me feel ppls hearts..oso hor.. haha.. after i came back frm camp den dunno y.. got more ppl ask me wats \" Amor Ipsemet\" yep.. its a latin phrase i go translate frm english my self.. yupz Amor Ipsemet= Love me n Myself.. haha.. i not ze lian kuang.. jus dat stil remem wat SL say.. \"If u cant even love urself, how can u expect others to love u or love others?\" yea.. it really struck me.. i remem i heard it b4 la.. but its like she said it at a right time, and by right person.. haha.. ya..so \"amor ipsemet\" is my motto of de yr~ so i cant blame myself for all my decisions.. my gei kiangness.haha..ya..den i mus say la.. i got this motto..my lifelong motto..tot of it ever since i was CYCERSE.. dis name i got..haha.. yupz.. its \" To be A Gentlemen, Thats What I Am\" yea.. haha.. so gals if i am not a gentelmen must tell me yea? i will change de lor.. i wan be gentlemen ! so lastly.. must stress today happy ya? pls i dun wan any1 to spoil my day! '),(8751,3375539,'04,June,2004','2004-06-04 00:00:00','    hahaz.. finally sum1 see my blog for de first time..haha.. *Blush*.. yea..its u ask for it de.. hehe.. den i dunno y i jus gif u read lorz.. i got nth to hide.. yea.. wo shi Zheng Ren Jun Zi k? yupz.. den soo long nv write something le.. more den a wk le.. yah..so i write sth now.. yupz.. start with today first la.. i v happy lor.. actually morning wake up not v happy cos yesterday nite la.. dat pig la.. dunno wats wrong wif her.. upset me as well.. hmm yea.. but late morning i okie le.. quite happy le.. hahz.. cos i write my friendster thingy.. yepz.. hahaz..suddenly me so onz.. i always lazy to write lorz..haha.. den i write liao.. got some memories come back mahz.. so v happy lorz.. haha.. okiez.. den for de past wk.. well where am i??? yah.. hmm think i was busy preparing de games bah.. yea.. put in effort lorz.. ut dat pig dun wan recognize it..didnt even say thks.. v angry lor.. den i hurt my leg..blue-black lorz.. den she dun even express concern..:( haha..den dat weilun care for me instead..ahah.. guess he rly dun seem like a guy to me lorz.. soo sissy.. or gentle as de pig put it.. yupz.. den yea.. i got PS camp.. ahhaz.. first day damn shitty lorz.. cos my dry games didnt turn out well la... but dun blame me lehz.. i dun wan to de rite? den its like i got try my best lorz.. if de games not fun ah.. den u all shuld tell me earlier mah..den i can change or sth rite? haiz.. den hor hey.. everything is i do de ley..my games aic do dunno wat shit laz.. i mus instruct her den she will do de lor..yah for everything i do.. pls appreciate lehz.. haiz.. hurt my feelings.. dats y 2nd day early morning so quiet.. yah.. dunno if de pig noticed not..yea..den de previous nite.. cant slp lorz.. firstly is cos de J2s and de pig nv slp la.. dey discussing thngs den i think is a scavenger hunt game la.. cos de amazing race cancelled? yupz dey settle de game in 1 nite..haha but wah laoz.. dey 4 brains i 1 brain.. not fair.. den scavenger hunt oso sth like my camera lu kou mah.. anyway..i think my game more creative yea? haha.. den yupz i slept at 4.45 bah arnd there.. yupz slp 1++ hr first day.. haha.. den 2nd day i begin to get anxious.. cos better do well dun b like 1st day..yupz its wet games 2nd day..den in e end turn out quite well bah..at least i can see every1 played n enjoy.. yahz.. so happy.. but de games i tot sibeh lame lehz.. ahha.. but seems like every1 likes lame stuffs. like my Qian Bian wen Da.. seems quite a success.. didnt noe some of my jokes can be lame to them as well yea..tot they might heard of them b4..den its like i dun rly noe how to tell jokes.. i must think den can remem..u suddenly ask me say 1 i cannot de.. haha..yahz.. success.. so 2nd nite yupz.. dunno since when.. yea.. de j2s suddenly like hinting me.. keep toking bout flying duhz~ yupz so crazy de.. haha..yea.. den i think i got de hint liao lor.. i must be going to becum de president of PS..duhz.. haha.. but of cos mahz.. i soooo clever.. plus soo capable de.. nobody else suits better rite? nobody to choose anyway..jc1s so little.. but i dun mind la.. anything lor.. be pres den pres la.. haah.. i dun look it as a real prestigious job or sth.. yea. perhaps gonna get more stress.. yupz den 2nd nite de tcher ask to see me.. duhz.. den sianz la.. i tot shes gonna ask me quit yfc or sth.. but never lorz! wah so happy.. ahah.. yea.. i jus told de truth.. manz.. i beginning to like being honest..it always seems like de rite thing to do.. yea.. so sumtimes i honest dun blame me lehz.. once again APPRECIATE it manz.. yah.. hahaz.. den yah.. nth much le la.. i become president lorz.. duhz.. quite sianz.. haha.. den come back frm camp.. now nth much to say.. jus say.. i slp quite a bit.. den i eat curry for de pass few days.. den msg pig.. den its like suddenly alot to say. den when she online ah..dunno y leh.. like different feeling.. think shes not in a gd mood bah..den i quite quiet lorz.. she oso. hmm dun care la.. yupz.. pass few days didnt really msg SL yah.. haha.. frm her blog .. wah i quite happy she missed me haha... yupz trust me..tho i nv msg u.. i rly got think of u and miss u still haah.. yupz.. anyway.. i think write sibeh lot liaoz.. haha.. i stop le bah.. lastly.. i must pray i will be able to work well wif my ps jc1s. yupz.. and jc2s help me lehz.. '),(8752,3375539,'04,June,2004','2004-06-04 00:00:00','    yep.. i dun care le.. keep thinking makes my head spin.. i  just gonna do my part.. i wun gif a damn even if things jus dun go out smoothly.. yep.. refering to yfc.. hmm.. i am pretty stressed out by it.. yea.. forget it.. if really kena kick out den too bad lor.. anyway.. first wk of hols so sian.. everyday still go back sch.. and theres work to be done.. the 2nd wk gonna be so scary.. haha.. cos got PS camp.. den de games until now still like v messy.. so much work to be done de lor.. haiz.. but dun care la.. god will bless de.. hee..haha.. me n piggy now meet up alot lor.. haha.. at least i noe she did miss me de..:P today i lunch wif her lor.. yup.. jus me n her..;) den i walk her home again.. jus like me yihting n her did on tues.. yea.. but now its me n her alone..haha.. wat else u wan me say? but me n her haven develop dat kinda relationship hor.. dun get wrong idea.. '),(8753,3375539,'31,July,2004','2004-07-31 00:00:00','    ~Life is a battle between faith and reason in which each feeds upon the other, drawing sustenance from it and destroying it~     . tired. when does it stop.   ...     ..     .     .     .     .     .     .    .   .  . .   the road to nowhere..      life is a journey not a destination    there is no ultimate point of arrival, it is the road, now and forever    finite man probing infinity, finding his way endlessly, helplessly '),(8754,3375539,'29,July,2004','2004-07-29 00:00:00','    hey yoz.. guess wat.. 5 mins ago.. i had a sudden stomach cramp.. haha.. dunno y leh.. not luff too hard.. dunno.. haha.. maybe due to PE yesterday?? haha.. im such a weakling.. lol..  den haha..so excitied!! my pri sch fren dat i nv meet for like 4 yrs msg me suddenly!! ah.. its cos his coming for tmr harmo concert ma.. haha.. yea... den he got extra tix so ask me whether i wan not.. haha.. of cos i hav la.. i frm aj ma.. will support lor.. tho i dun like de sch alot.. but i support my frens there ma.. ya.. haha.. rly v exciting leh.. haha.. i miss him.. haha.. orh.. well so for de pass few days.. it wasnt v exciting at all.. lol.. cos busy wif work still..everyday go home onli 1 thg to do lidat.. write my gp essays.. haha.. i still got 2 more to go lor.. continue later la. haha..yah.. so sianz.. lor.. yesterday i buay tahan leh.. haha.. i slp at 8.30 lor..haha.. sry.. i lidat de la.. pig lor.. den ps oso v stress de leh.. alot of thgs to do leh.. cos dey erm quite low efficiency in my eyes la.. ask them do do so slowly de.. den i look thru de pdg photos mah.. overall accpetable la.. but some ah.. kaoz eh.. i dunno how to face ppl lor.. die leh.. so got problems in&nbsp;erm edit.. duh.. haha.. but dunno la.. shuld be still pretty alrite la.. everythg.. ya.. i hope so... cos nxt wk handover le.. den mus plan stupid farewell party for them ma.. haha.. i dunno dun care.. jus a simple 1.. den still got de photo exhibition to settle lor.. den still need to do some admin work on ps structure n discuss bout de details of our operations wif de exco la.. so actually nxt wk oso quite busy~! ah.. haiz.. expected de la..nvm.. endure.. :P but hmm.. bad signs.. i erm.. \"Menstruate\" again~! eh.. maybe some ppl will noe dis term haha.. it means i ass bleed la.. yah.. alot blood ah.. die.. my mum say maybe got Zhi Chang or sth.. eh.. dun wan so scary.. i think its indigestion la.. eh? haha.. so sry for de gross part..  wuld like to say sry again to dearest SL.. i was being insensitive.. yea.. but im sure u not dat petty de rite? :P haha.. thks for forgiving me la.. but i oso not v happy when u like hav such a bad temper.. ya.. den eh.. dun hurt me wif those harsh words can?? like \"scram off\" even lulu oso think u fierce lor.. hehe.. so rly mus control ur temper k? i hope i can help la.. n i noe ur trying hard to.. :) yoho.. tmr n de nxt is v busy oso.. especially sat ah.. think i frm morning 7am to 11 pm de earliest lor haiz.. cos got cip thingy n mus do pw oso.. den de chi tcher oso think she gonna bug me.. she always bug me.. keep calling me leh.. i not her bf or sth.. duh.. she calls me at nite.. aftnn in lessons.. even cca she call me.. ah.. crazy.. i onli chi rep leh.. she in thinking committee de ma.. den oso i/c of many chi thgs... so got any thgs events organised she ask me oso lor.. sianz.. haiz.. nvm..  gtg liao le.. so late.. chatted wif ppl enuff&nbsp;i guess :P haha.. no SL tonite.. :( hehe '),(8755,3375539,'24,July,2004','2004-07-24 00:00:00','    ahah.. every1 seems to be gg to de dinner.. but im not.. hehe.. so paiseh.. cos no J1s seems to be supporting de J2s in their photo exhibition there oso.. haha.. dun lidat la.. haiz.. i dun wanna go oso.. kinda sianz... at suntec lor.. anyway.. last wk was crap.. everyday so late.. well seems like de nxt wk oso de same.. ahha.. nvm.. i will survive.. irene says president mus noe how to distribute work.. ahah.. okie lor.. i noe la.. i noe my style well de lor.. hahaz.. anyway.. today its her bday.. n im not anywhere near her haha.. happy bday pig.. sharing her presents wif a fren.. but im not giving it to her.. my fren is.. haha..  juz now hav mentorship programme.. haha.. n i was posted to deyi sec lor.. haha.. irene came frm there de.. anyway.. look forward to it.. n have fun wif de kids.. haha.. den i saw SL at de amk bus interchange.. haha.. its still u lor.. ^.^ nice to see u.. but u seemed a little shock.. n erm.. u seem to be uninterested or some sort..ahaha.. nvm.. den i came back home n explored my com.. found some pics.. yea.. of my class last yr.. miss them ya.. shall question lynn n bekky some time for my pics.. yea.. i remem got 1 pic i hav wif all de 4e6 gals n im de only guy rite?? haha.. id like to show it to samuel seah.. see wat he says.. LOL.. ahah..  kk.. im gg to mug.. yep.. Jia You Ba me.. 5 gp essays to go..~ '),(8756,3375539,'21,July,2004','2004-07-21 00:00:00','    urlLink    7/9 is in AJ~! haha.. '),(8757,3375539,'21,July,2004','2004-07-21 00:00:00','    urlLink    Found a pic of u somewhere in mi com.. lol '),(8758,3375539,'21,July,2004','2004-07-21 00:00:00','    urlLink    my 4e6 class gals.. de one in de middle is de cutest :P '),(8759,3375539,'21,July,2004','2004-07-21 00:00:00','    urlLink    this is for u lulu.. ur shuai ge erm..doing sth serious.. '),(8760,3375539,'21,July,2004','2004-07-21 00:00:00','    aight.. suc... dying.. once again.. alot of thgs to do... again.. im tired.. sumtimes feel like crying.. especially when ppl dun understand u.. sumore ppl scold u.. reprimands u.. dey cant feel wat i feel.. they dun like to see thgs frm my point of view.. dats v sad.. i need a hug badly...T_T im like doing everythg lor.. ok so wat if im pres.. duh.. i dun havta do everythg rite?? but seems like im supposed to noe everythg..suck.. de ps J1s asked me alot qtns which i cant answer cos i dunno as well.. n den dey shift all de blames to me.. alrite.. i can accept dat.. forget it.. but im not superman.. boohoo.. haiz.. okiez.. im gonna do things at my pace.. fuk it.. i still feel like crying..boohoo.. every nite im troubled.. cos those stupid classes.. dey keep wanting to change de pdg phototaking schedule.. n de shittiest thg is no one knews bout it except me.. boo hoo.. y am i taking all dis Wei Qu.. haiz.. stupid pig.. hate her.. boo hoo.. :(  :) im giving dis smile to u.. ppl who cares.. thks.. i haven die yet.. i wun.. not so easily.. arhhhhhhhhhhh '),(8761,3375539,'20,July,2004','2004-07-20 00:00:00','    yoho.. today im talkative.. yes.. \"ppl,listen to me\" quoted frm mr ari\'s lecture.. lol.. he always say dat..haha.. haiz.. dunno start frm where.. now quite late leh.. hey.. haiz.. i jus tok bit by bit.. miss out anythg dun care le la.. today not a v happy day leh for&nbsp;academic.. nv do phys hw.. haiz.. den de YIP sux.. hes eyeing on me le.. can feel it.. den he scold me nv bring tys.. wan me bring 2 nxt lesson as usual.. duh.. den during maths.. i cant solve de maths problem lor.. haiz.. den ended up teaching somebody 3d trigo hor..*hints*(LULU)  erm.. den today went for interview for de ang mo kio family service centre.. for mentorship programme de la..ya.. ok.. its still quite smooth.. do interview till sianz le.. haha.. den chatted wif trixy n sutha..(dey my classmates k SL^^, dun tsk tsk)&nbsp;dey went interview wif me together.. yea..&nbsp; den dey doing rangoli for competition.. noe wats dat? de\" floor painting\" made on using de salt,sandetc.. during deepavali.. ya.. haha.. i noe lor.. so i haolian n gif suggestions.. so proud of it.. :P haha.. den trixy say i got alot fans.. as in lulu u all lor.. wah lao.. wat fans.. sounds v nan ting leh.. cos she jealous la.. mus be.. cos u ppl call me \"hao ren\" even de siti thaharah oso ask me wats dat.. haha.. well im gd wat.. :P den.. haiya.. dey waited for me finish interview when i say.. sui bian when dey ask wan to wait for me not.. haha.. ok la.. so i treat them Gum Chia Zui.. haha.. (pst..finally find an excuse treat them k) haha.. im a gd guy rite? haha.. see so gentleman.. :P dun choke k..? hehe..  ok la.. my headache like cuming back le.. haha.. i go slp liaoz.. miss alot ppl lor.. yea.. tho i v busy.. i still miss.. no need mention names la.. sec sch ppl.. pri sch oso got.. all la.. haha.. but specifically.. some la.. haha.. kk.. im gonna do well for tmr.. as in de ps stuff.. alot thgs to settle.. de tcher is pleased at de moment.. lol.. as in she say\"gd, u have done a gd job so far\" duh.. of cos la.. :P haha.. kinda bhb today.. haha.. okie..took 1 hr to write??&nbsp;haha.. chatting&nbsp;dats y.. type v slowly.. &nbsp;nitez.. ^^ '),(8762,3375539,'17,July,2004','2004-07-17 00:00:00','    alrite pretty slack for today.. haha.. yupz.. my plan was type de PW graphs n analysts dis morning.. den i would go school print it.. den i meet wenbin in sch.. den go for flag day together.. den we went for lunch at BK..(mushrm swiss double meal+onion rings) den went to watch Brotherhood.. den came out collect for like 15 mins.. erm business was bad..&nbsp;de place was saturated wif so many others volunteers..sumore de ppl in yishun is rahter hostile.. haha.. yupz.. den we return the tins wif like&nbsp;$5 worth of coins.. hehe..&nbsp;den i went to amk mac do PW.. i reach early.. so start doing my evaluation for de PW research.. den we spend time doing alot of writings.. evaluating thgs 1 after another.. argh.. but finally finished.. yippie.. den i ate Mac for dinner.. (Double Mcspicy Burger Meal).. n i flirt wif de \"waitress\"..haha.. or was she jus frenly? she approached me using her palmtop.. den i said.. y so slow.. haha.. den she say she first time do ma.. ahha.. den i jus say ok.. den i waited.. n she serve me.. haha.. den i said.. wheres my straw.. den she kept saying sry.. haha.. cos she first time doing.. haha.. i say nvm.. den i went to de counter go get curry sauce.. den she said.. wat did i forgot.. haha.. den i say no la.. take some curry sauce only..den she say sry again..den i say nvm.. haha.. den i saw her name tag.. yuki.. i called.. den she say ah? den i replied call for fun.. she jus luff n went off.. hehe..&nbsp;n den eat halfway.. i got stomachache..argh.. bak tor tia.. so hmm.. wat was de cause? mushroom or mcspicy?? after my PW.. i walked alot.. i saw my 135 leave.. so i decided to walk to de nxt bus stop.. when i reach de nxt.. de bus haven come.. so went on to the nxt.. den de nxt.. den de nxt.. all the way till i was at townsville primary there!! wah kaoz.. i was almost completing my walk away from AMK.. duh.. stupid man.. so i sang along as i walk.. n theres only dis 2 songs i noe dats linked to \"walk\".. ah.. den so i think again.. hai.. am i walking or running away from my problems??  movie review : brotherhood... no...SL cannot watch this.. so bloody.... too violent for her le.. haha..&nbsp;n the part where de bros reconciled at e end was v touching.. will sad till cry de.. i had to tried hard not to cry lo.. haha.. it was not a bad movie.. consider alot of the war scenes which was quite well taken.. yupz.. it features Korean war shuld help in Social studies.. or now.. GP.. haha.. so.. tmr.. is study day.. haha.. gonna finish up my tutorials.. yesh.. n gotta call my ps mates.. double check thgs.. ensure that they are ready for nxt weeks pdg phototaking.. i got no idea.. but sumhow i think thgs will go wrong.. argh... god bless.. let dis be a success.. &nbsp; '),(8763,3375539,'14,July,2004','2004-07-14 00:00:00','    nope.. cant plan.. my plan not working at all.. duhz.. i onli got 1 solution.. to find back my confidence.. where\'ve u gone? dun go quiet.. theres not even a trace &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; . \"I may be small, but my confidence has made me twice your size\" KH 01:15 * wheres dat air of confidence * '),(8764,3375539,'14,July,2004','2004-07-14 00:00:00','    noe wat? im becoming very frustrated over this fucking life.. fuck off manz.. i typed my blog jus now.. n den life sux for me till it went unpublished.. well.. dat was fuck.. fucking stupid.. this kinda thgs oso wan fuck me up.. forget it..  -_-  *_*  \'~\'&nbsp;  `~`  ;[&nbsp;  #_#  or perhaps not considered.. end..  im going crazy.. all de stress makes me suffer from sleep problems, migraine and indigestion.. my biological clock has gone cuckoo... i cant control it.. im sry.. i sleep from 8.30 - 11.30pm for de pass 3 days.. woke up to do work.. den slp at 2.30 - 5.30am.. after i ate dinner.. i cant help it but lie dead on my bed.. *BuMpH* dats de sound i heard when i \"lie\" down at my bed and perhaps.. dats y i had migraine n indigestion... stomach *rumbles* argh...  im dying.. can any1 do CPR? &nbsp; i\'v informed de pdg reps on the phototaking.. and.. guess i was too sianz to get nervous of using de mic.. -_- bio spa was a f*i*u*r*e, haha.. i spent much time doing de expt, n took my time to plot de table.. gosh..its skill c lor.. no idea wat i doing.. ended up no time for drawing conclusions.. i was crazy.. its not as if i dunno how to write them.. but no time..T_T den maths test yesterday was.. muahahahah.. i no how to do none of them.. *clap* i shuld think of how to fry my big fat \"egg\" PW was crazy.. we didnt tot of de plotting of graphs for our surveys.. now we got soo much variables.. we got alot of problems.. alot of thgs to be done..&nbsp;can i&nbsp;do it?.. haiz.... &nbsp; no use grumbling.. *piacks* im useless.. its hard to amor ipsemet.. sobs ....... .... ... .. . . . . . . . . . . . . . &nbsp; '),(8765,3375539,'14,July,2004','2004-07-14 00:00:00','    head hurts.. argh~! whole day im using it.. n i will still have to use it within de next 72 hours.. still gonna be busy.. haiz..still im thinking.. thinking of u.................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................wow.. full stops r cool.. '),(8766,3375539,'12,July,2004','2004-07-12 00:00:00','    hohoho.. lotsa thgs to do again..haha.. PW got alot alot thgs to do leh.. how ah.. haiz.. my grp onli now den Ling Shi Bao Fuo Jiao.. hehe.. all the minutes nv write lor.. so now mus fake everythg out.. haha.. i think our minutes now like quite dramatic.. each person say 1 thg.. haha.. den now getting de momentum back again le.. my grp begin to luff n have fun le.. haha.. den my other grps discussion today abit off track.. haha.. keep luffing..dunno wats wrong.. we got 3 txtbks to use.. but then so stupid lor.. we discuss until every1 crowding over 1 txtbk to read.. haha.. so stupid.. den today interhouse games.. irene reminded me of it dis morning.. she ask me whether i got join not.. ahah.. den in her sms she ask whether i late not.. wa kaoz.. indeed i am lor.. but lucky i change bus den not rly late.. haha.. ya.. den i watch pig play netball lor.. k la.. not bad la.. haha..den weikeong say she fat, den got big bums.. hahaha.. n i go tell pig dat.. haha.. but seems like she dun care.. haha.. i nv watch finish go off le.. haha.. so sianz.. cos she confirm win de ma.. :P hmm.. mus go chiong my thgs liao.. shitzzz la.. haha.. alot thgs lor.. so many tp type  plus all those hw.. i think i can nv finish my 5 gp essays.. haha.. siao lor.. '),(8767,3375539,'11,July,2004','2004-07-11 00:00:00','    haha.. erm.. seems like my plan failed?? ahha.. erm..well.. y not i try it again? last wk not counted la.. cos i sick.. haha.. den.. hmm still v ill disciplined.. restart k? ahah.. :P damn sucky la me.. i mean im slackkkkkkkkkkk..... or in fact lazyyyyyyyyy.. yea.. all my hw nv do again.. shitz.. its sucky.. i dun like dat.. i gotta mug lor.. pls ley.. i find myself v sia shuay.. cant even study well.. haiz.. guess wat.. im aiming for sth.. aiming for the \"Most Outstanding Student\" award in AJ.. yea.. haha.. aim onli ah.. haha.. whether can get not its a different story.. ahah.. yesh.. all those top student n crap i dun care.. haha.. i onli aim for dat award..yesh.. saw it when i was doing my common test in de hall.. haha..  yesh.. wanna intro http://www.fasco-csc.com/ yupz.. its sum time since i played a puzzle game.. ahha.. kinda like dis kinda games..click click click den got progress.. den needa think.. hehe.. i like dat.. spent quite a long time tho.. duh.. haha.. but see im slack.. :P yea.. hmm think i left 1 nice quite puzzle game in the 4e6 class web.. yea.. haha.. i go chck de link later.. haha i find SL constantly changing.. haha.. hmm.. cant tell wats on her mind liaoz.. haha..:( really lor.. shes like always maturing.. yea.. den hmm ya.. her moods changes v fast.. haha.. den she tell me PMS.. haha.. well its not a very gd yr for her i supposed.. well still mine\'s worse.. haha.. nvm.. knowing dat shes fine can le.. :) really dun wish to know her crying again.. den kinda miss pig too.. yep.. she grew leng dan towards me.. dats sad.. n everytime i tried to chat wif her online.. she\'ll start mention SL.. duh.. wth.. damn sick of it.. haiz.. cant we continue like perhaps where we left off? hey.. i got no ulterior motives?? haha okiez.. nitez.. i go off le.. finally gg to do some pw tmr.. lol.. siao liao.. daljit gonna kill us if she checks de files~! haha.. run for my life~ '),(8768,3375539,'10,July,2004','2004-07-10 00:00:00','    hehe~ so long nv come online.. well seems long to me lo.. haha.. yupz.. cos i sick manz.. den i slp n slp.. yesh.. slp a hell lot.. de onli thg i noe was on thursday i was a complete blur.. cant even leave my bed to go bath.. hehe.. yupz.. i gone cuckoo.. dunno wat im supposed to do..but then i woke up quite a no. of times.. cos alot ppl call me.. den i jus ah,eh,orh, haha.. cant really remem wat dey say.. :P den SL cry dat day sumore.. make me worried lor.. so i called her dat day.. ya.. n it turned out to be not a very big problem to me.. i tot sth bad happen to her lo.. haha.. yah..but true la.. i was not there.. so i dunno wat really happen n how serious it was.. haha.. yesh.. den yest i oso slp a hell lot.. i got home den BoMpH~ i lie flat on de bed.. haha.. den i slp till 2.30am.. yesh.. i was not waked up by any calls or smses.. haha.. think cos i put silent mode.. :P onli got 1 msg.. but i oso dun care.. too late liao by de time i see.. so i went back to slp.. haha.. yep.. den its tis morning le.. haha.. den got listening compre.. n hmm dunno whether de last qt correct notz.. but i jus chose de option i felt true to me.. yes.. Respect n Understand has greater effects den Encouragement.. haha.. yesh.. dunno.. i think quite true ba.. *_* den i jus finish my pdg phototaking schedule.. sooo many errors.. make me waste alot time.. haiz.. in e end.. still nv study.. :( sob.. i shall go study le.. haha.. n SL dun worry.. i think i recovered liaoz.. haha.. i slp so much.. :P haha.. but v touched.. u kept asking me whether i see doc liao not.. haha.. thks.. Miss ya.. '),(8769,3375539,'08,July,2004','2004-07-08 00:00:00','    die le... i think im sick.. down wif flu.. gan mao le.. haiz..my throat still v pain.. den i think begin to form blocked nose.. cant breathe..haha.. sianz.. my plan is going crazy le.. haha.. cos i slept jus now!!! 8.30 to 11.30 i was slping!!! haha.. im dead lor.. haiz.. im suffering.. but still i have to struggle to type those crappy stuffs.. yea.. dun type tonite den when? haha.. im tired liaoz.. haha.. but still hafta move on~ go go go~ '),(8770,3375539,'07,July,2004','2004-07-07 00:00:00','    here pain.. there pain.. everywhere pain.. my throat v pain la.. haiz.. heart pain.. ass pain.. head pain.. still got anywhere oso pain la.. PAIN.. n i think i shuld feel pain cos i nv stick to my painful plan '),(8771,3375539,'06,July,2004','2004-07-06 00:00:00','    hmm..exactly..9pm.. heard de theme song of de nkf dat hai you ming tian.. ya.. damn nice.. i like.. n perhaps cos de Yao Zi very Qing Chu.. like GuangLiang songs.. ahha.. i like sing Wo Ni De Shou.. hmm.. izzit cos tmr me chinese oral? haha.. dunno.. n i dun wish to noe anyway..i dun wish to noe anythg le.. hmm.. noe liao oso no difference..  ya.. today actions v slow..kinda like an old man.. but still quite productive..cos finished my bio holiday assignment..ahha..ermz mostly copied.. n finish all my chinese work..ya.. quite conscientiously..bother to check dictionary sumore.. well i chi rep mus do it ma.. haha..wanted to do de physics online assessments as well.. but ha.. i went to slp..:P  im still looking blur.. at a lost of wat to do.. everythg seems calm.. but ha.. wait till sch starts.. i noe.. bad things will come 1 by 1.. ha.. pardon me for being so pessimistic.. but i think dats me.. i got try to be happy lo..yea.. sumtimes i am.. but most of de time not..*.* ha.. im trying v hard.. yea.. Amor Ipsemet.. but still kinda find my actions v stupid.. ha.. tmr chi oral.. will it be de start of yet another unlucky chain of events? loneliness is tragical..still remem de story i read yesterday lo.. ya.. de story v sad.. im like de main character.. who is cooped up in his own room.. waiting for some1 to open.. ha.. deng dai na yi pian shu guang.. well gotta have some rigid plans to study liao.. ha.. but i noe.. i always tok only.. no action.. ya.. but still shuld mention at least.. hehe.. well i think i gotta find someone to go through de newspaper wif me every morning.. ya.. not read through them.. but jus ticked them out.. haha.. yup..den i will come very early to school.. hope so.. i dun wan late again.. haha.. sianz lor.. hate de 853 bus i hate it~! i wish to take taxi every morning..:P  ha.. late den have to do CWO lor.. dun wan le.. had de experience liao.. no wan le.. haha.. like pig said.. i think i late so many times will lose credibility de.. ya.. haha.. i no wan late le.. hmm.. nono..come back to main point.. newspaper is for GP.. hmm.. den i think i everyday after sch i will go reading room.. yupz.. i will do my hw.. i already told antonio le.. i think he will accompany ba.. den i think ard 6.30 den i leave sch.. reach hm shuld be nearly 7 le.. den bath eat..watch tv meantime.. till 8..at most 8.30.. haha.. den i will finish up wats need to be done.. den i will leave 30mins to 1 hr to recap wat i gone through today..hmm think dat dats very impt.. den ha.. i think enuff le lor.. will reach ard 11+ le.. will b v tired.. so go to slp.. dats mugging life.. ha.. no choice? well still i think i will come online.. yesh.. n on every \"odd\" days..lol.. i mean 1,3,5,7 yesh.. mon,wed,fri,sun, hehe.. ya.. n dat means i come at ard 10pm for these days.. haha.. hmm.. not much time i noe.. but hope it really relaxes me out.. n ya.. saturdays.. i now free le.. haha.. perhaps morning most of de time i will still go sch.. yea.. den afternoons i will maybe go do PW.. or well i will use it to MUG!! i will do TYS?!! ah.. i wan be mugger.. but even so.. i noe i wun be able to do it.. hard to be mugger de.. ha.. but siao lor.. sumtimes i wun mug la.. maybe once in a fortnight go out n chill.. yea.. dats my plan.. ha.. issit too crazy? well..i dun think so.. i still will come online ma.. haz..  really wish to numb myself.. n be void of all feelings.. but i dun wish to let laziness occupy my soul.. hope is work dat keeps me busy n forget other thgs.. cant concentrate tho.. dats y la.. bah..n one of de reason is of cos thinking of u..ha.. cant help it.. but im trying not to.. @.@ well.. kk.. i noe i type alot.. ha.. n i like to do it liaoz.. ha.. mus be cos i too boh liao.. duh.. haha.. byeZzZ.. u wun get to see me so often le.. T_T '),(8772,3375539,'05,July,2004','2004-07-05 00:00:00','    hmm find nth to write bout.. how bout me eating satay for dinner? haha.. dun wan la.. later got de euro2004 finals.. hmm any team win i oso dun care le.. i wan my netherlands.. boohoo..:P '),(8773,3375539,'04,July,2004','2004-07-04 00:00:00','    it hurts.. wo gan jue hen tong.. v painful.. de worse thg is.. i choose to hav dis torture.. '),(8774,3375539,'04,July,2004','2004-07-04 00:00:00','    urlLink    im jus too cute le~ :P&nbsp; urlLink '),(8775,3375539,'04,July,2004','2004-07-04 00:00:00','    ya.. beautiful.. cos.. morning.. SL called me say she become CSC pres.. haha..so farnie.. b4 dat she tell me she sooo nervous..haha.. den she tell me she no chance le sumore.. cheat ppl.. haha.. naughty gal.. den rly lor.. she call me dat time.. wah.. de voice v loud la.. haha.. my ear drum pain.. :P haha.. think she v excited.. ahha.. think she will do a good job de la.. SL v responsible de... ya.. i noe she is.. n i trust dat she is.. haha.. Jia You..!! me will support u de.. haha.. den later evening.. got PS bbq.. haha..tho mani ppl nv come.. as expected la.. haha..but nvm.. like i said alot times.. not my fault..haha.. all irene\'s fault.. :P haha.. den but den still quite alot ppl liao le.. haha.. den ok lor.. had quite some fun.. den we de J1s pao qi J2s.. go maze.. den go playground.. haha ya.. den had quite some fun.. we play de rotating plate.. i enjoyed dat most.. haha.. den like kids lor.. den got pass by de tree..4e6 used to take photos on.. during de pasir ris park cleanup.. haha.. ya.. de tree look v nice lei.. haha.. but then after dat go home.. stupid thgs happen again.. i 11.20 reach hougang mall.. den i go find ice cream for my sis..her bday.. ya.. den haiz.. cannot find.. so no choice go home.. den i reach mrt.. 11.32.. den de last train went off lor.. sucks.. den i try go interchange take bus.. den de bus leave rite in front of me.. den i stuck in houngang mall.. wah shitty lo.. so i no choice take taxi.. den cost $5.50.. so exp.. ya.. den reach home.. wah biang.. my sis haven reach home yet.. shitty boo boo.. haha.. sianz la.. its AGAIN lo.. stupid stuff.. hah.. used to it le.. nvm.. i wun brood over it.. haha den SL make me sorta angry.. so bad de she.. she say its pig help me den i pres.. huMpH.. i angry.. den i shuo qi hua.. say slap her.. den i dunno.. she like angry oso.. haha.. bad gal.. she noe i will giv in.. duh.. haha..ya.. in e end nth la.. nv rly quarrel.. she promise be gd gal.. dun say such naughty stuff again.. haha.. she say dat was a joke sumore.. -_-.. dun sound like joke to me leh.. HuMpH.. i say i wun gif in to her next time.. den.. she still try be funny.. haha.. say me will gif in de.. huMpH.. i wun la..:P.. hope so.. but dun angry me first la.. so hurting lo.. haha.. bad gal..  finsh le.. wrote alot again :P '),(8776,3375539,'03,July,2004','2004-07-03 00:00:00','    hmm haha.. perhaps shuld note down rite.. haha.. in case forget.. SL gave me a online hug.. LOL.. am i too kau zhang.. such thgs oso note down.. haha.. hehe.. dunno.. but its de first time.. i nv force her ya.. haha.. kk dats it.. no more.. bye '),(8777,3375539,'01,July,2004','2004-07-01 00:00:00','    wa... not bad sia.. i made 30 posts liao.. wah.. *CLAP CLAP* haha.. i nv tot i wuld last dat long.. but now i do.. i think it will continue to go on to 300 post.. 3000.. haha..ya.. so mus thks SL ah.. haha.. cos she gave me de idea of starting 1.. ya.. think now blogging quite cool la.. de govt i tot oso encourage blogging lor.. got wat \"Singapore Blog\".. haha.. ya.. so its a pasttime to sgreans liaoz.. haha.. hmm.. sudenly felt like i v loyal to my country.. yea.. actualli i quite like sg la.. n de gals oso not bad le la.. tho not many mei nus..but then.. i think dey r rather sunny.. haha.. kk.. dats not de pt..hmm i will write wat i intend to write.. firstly.. dis morning i post de.. hmm i say wat i wan memorise.. haha.. i jus remembered i sec 2 dat time go memorise 1 song.. \"F*** OFF\" by Dr Dre.. lol.. haha.. i its a rap ya.. den got alot alot \"F***\" words.. haha.. its like every phrase have.. ermz.. actually me still can sing dat song.. haha.. v power.. haha.. every hear me rap le den.. haha.. mouth open den stare la.. ahha.. yea.. actualli dat was my intention when i learn dat time.. if any1 remems.. zhss2e4 dat time got dis guy called johnson.. haha.. i learn to spite him.. haha.. i jus wanna haolian nia.. ya.. so bad me.. but then.. haha.. quite cool oso in a sense.. can rap it not easy lor..hehe..   hmm.. den today did nth constructive.. still lazing ard.. haha.. 1 day like gone v fast oso.. haha.. den i go lift weights.. whaha.. i wan train liao.. haha.. i hope so.. ahha.. i wan nice nice figure oso.. i got a postcard.. (my classmate is a postcard collector).. ya.. on it is a picture of a figure v nice de guy.. ya.. so i aim to get dat body.. ahha.. den i cut out newspapers.. haha.. realy.. i v scared of my gp.. haiz.. i always wanted to do well in eng.. but i noe i cant.. haha.. but nevertheles.. i will try hard de..ahah.. i will still read my newspapers.. den hopefully got use in my gp la.. haha..den ya.. i cleared my table.. wah. haha.. damn hard to clear:P  haha.. den at nite.. SL tell me sth.. haha.. rly.. i v happy lor.. haha..cos she rly trust me.. haha..^^, cant slp le la..aha.. (no la.. i will still slp ;p)ya.. i trust her too.. i think shes de person i can trust most la.. n i wuld.. i dun like to hide la.. v tiring de lor.. haiz.. every1 is hiding lor.. dun bluff.. i noe de lor.. whoever doesnt hide.. den will look v ugly de.. haha..ya.. so haha.. jus wanna tok de truth to her.. ha.. den hmm i wan to be Blaine.. haha.. ya..i cut de article out le.. haha.. i will keep den show u 1 day.. haha.. i will stay tune to de barbie story de la den update u..ahha..Blaine is a suave well-toned aussie surfer dude.. lol.. ya.. i like sea sports n i intend to go learn:P ... hope can help u get over him la..i believe u will lor.. but slowly la.. even if 5 yrs.. it still took place.. dun strss urself over it.. n haha.. u sounded despo.. :P kkz.. i go slp.. i remem u ask me slp early..haha.. i v guai de hor.. haha..tata~ '),(8778,3375539,'01,July,2004','2004-07-01 00:00:00','    haha.. jus as i was surfing de net.. i notice alot of online arguments.. yea.. haha.. dey r damn funny.. those ppl jus like to crap n boast things in which arent true..haha.. jus becos they can remain anonymous behind their monitors.. haha.. so farnie.. well dats wat made up of singaporeans.. but hey.. makes me see dat there are many different kind of online users.. haha.. look at a small quote of de argument (by a perhaps more intelligent man) i saw to a man who boast to be the wat 1998 bodybuilding champ n crap lots.. this incident took place last yr btw.. ahah..  \"You have many boasts, and though this argument has altered your consciousness, you remain irrevocably human. There is nothing special about you. Ergo some of my statements you will understand, some of them you will not. Concordantly, while your brains and muscles may be the most pertinent to you, you may or may not realize it is also the most irrelevant. Denial is the most predictable of all human responses, but rest assured, your choice of denying to accept what I said will constitute an escalating probability of your death and beyond. The inevitability of your doom is apparent to everyone now as a consequence of the imperfection inherent in every human being, with yours being your ever-so cocky attitude. Which brings us at last to the moment of truth, wherein the fundamental flaw is ultimately expressed, and your existence on this planet revealed as both beginning and end. Nobody is ever perfect. Stop your cockiness.\"  wow.. de english so pro.. i go memorise den i king liao.. haha.. :P '),(8779,3375539,'01,July,2004','2004-07-01 00:00:00','    haha.. me slp alot yesterday.. not say v song leh.. haha.. but rly slp till v blur.. cos i again.. slp wake up slp wake up.. so mani pple call me msg me.. haha.. yupz.. den i think i cant remem wat i said to dem.. hope not foolish stuff.. haha.. i slp until headache leh.. ya.. ahha.. den i slpt till 15 mins after de soccer match starts.. haha.. yea.. quite exciting la.. tho no goals score..but then until dey draw ah.. den sian liao..cos i think de czech players like purposely lose de leh.. haha.. kelong again.. den extra time greece won.. haha.. underdogs rox.. me n qiyuan wanna bet greece win de.. haha.. but i nv bet.. cos me v suay de.. haha.. alrite.. today no plans yet.. think i go clean up.. mayb go out later for movie.. come back do hw.. den i think i call my pw mates up ba.. dey like no action de.. lost grace liao den dey become like lost sheeps.. too bad lor.. i think i have to chase them back.. hiazz '),(8780,3375539,'01,July,2004','2004-07-01 00:00:00','    haha..seems like im blogging everyday.. ya.. slacker.. :P.. haha.. today de chem test oso turn out to be SUCKY.. yea.. haha.. i dunno how to do de essay qts... well sianz.. knew it.. haha.. cos i nv do my chem hw.. always copy..hehe.. yea.. so cannot blame.. i nv practice b4.. ya.. haiz.. den die le lor.. those qts on chem bonding n those on reaction kinetics i sucked like hell.. haha.. well dun care.. think gotta hav a long term plan to study.. haha.. shi jie will be joining me study together le.. erm.. dun get wrong idea.. shes my neighbour only.. haha.. yep n shes in AJ too.. okiez.. sian.. i go slp la.. later den chiong study.. hehe.. '),(8781,3375539,'08,August,2004','2004-08-08 00:00:00','    hmm.. really.. shuld get some music to keep us going.. well so i added music codes here.. haha.. i really like this song.. Simple Plan - Perfect   Now It\'s Just Too Late, And We Cant Go Back, Im Sorry, I Cant    PERFECT '),(8782,3375539,'07,August,2004','2004-08-07 00:00:00','    urlLink    hee~ '),(8783,3375539,'07,August,2004','2004-08-07 00:00:00','    â—¢â–ˆâ—£â—¢â–ˆâ—£    çœŸå¿ƒ      â–ˆâ–ˆâ–ˆâ–ˆâ–ˆâ–ˆ    è¯šæ„       â—¥â–ˆâ–ˆâ–ˆâ–ˆâ—¤   å…³æ€€           â—¥â–ˆâ–ˆâ—¤       æµªæ¼«              â—¥â—¤            æ¸©  æŸ”      urlLink http://chinese.cari.com.my/myforum/forumdisplay.php?fid=127&page=27&amp;sid=igrmt6      çœ‹äº†ä¸€çœ¼è¶Šå‘ä¸‹çš„å¤§çš„é›ªï¼Œç”·å­©ç»ˆäºŽé¼“è¶³å‹‡æ°”ï¼Œå¯¹å¥³å­©è¯´ï¼šâ€œè®©æˆ‘æŠ±ä½ ä¸€ä¸‹ï¼Œå¥½å—ï¼Ÿ æœ€åŽä¸€æ¬¡ã€‚â€ å¥³å­©æŠ¬å¤´æœ›ç€é£˜é›ªçš„å¤©ç©ºï¼Œæ²¡æœ‰å›žç­”ã€‚ç”·å­©ä»Žå¥¹çš„çœ¼ç¥žé‡Œè¯»åˆ°äº†ç»æœ›çš„é¢œè‰²ã€‚å¥³å­©å˜´è§’æŠ½åŠ¨ç€ç”¨æ‰‹å¥—è¾¾è½å›´å·¾çš„æ¸…é›ªï¼Œè½¬è¿‡èº«èµ°äº†ã€‚æ‰“å¼€å•å…ƒé˜²ç›—é—¨æ—¶å¥¹çŠ¹è±«äº†ä¸€ä¸‹ï¼Œ ã€€ã€€ä½†è¿˜æ˜¯æ²¡æœ‰å›žå¤´ã€‚ ç”·å­©å­¤å•å•åœ°ç«™åœ¨é›ªåœ°ä¸Šã€‚å­©ç¥žæƒ…è½å¯žåœ°è¸©ç€ç§¯é›ªå‘è·¯å£èµ°åŽ»ï¼Œä»–è„‘å­é‡Œä¸€ç‰‡ç©ºç™½ï¼Œä½†è¿˜æ˜¯ä¸ç»æ„é—´æœ›äº†å¥³å­©å®¶çš„é˜³å°ï¼Œæ¸©æš–çš„ç¯å…‰çœŸå¥½ï¼Œä»–æƒ³ã€‚ .....  å¬ç€ç”·å­©è½»åŸæµ…å”±çš„æƒ…æ­Œï¼Œæœ‹å‹ä»¬éƒ½å“­äº†ï¼Œå“­çš„å¾ˆä¼¤å¿ƒã€‚ç”·å­©çš„ç”Ÿå‘½è±¡ä¸€é¢—æµæ˜Ÿï¼Œ ã€€ã€€å°±è¿™æ ·æŸè½äº†ã€‚å¦‚æžœå¥³å­©è®©ä»–æŠ±ä¸€ä¸‹ï¼Œå¦‚æžœå¥³å­©è½¬èº«æŽ¥å—ä»–çš„ç¤¼ç‰©ï¼Œå¦‚æžœä»–æ²¡æœ‰å›žå¤´çœ‹é˜³å°ä¸Šæ¸©æš–çš„ç¯å…‰â€¦â€¦â€¦â€¦â€¦â€¦  haha.. chinese story nice rite? cos i think onli chi characters can bring out de rite emotions..hehe  dun feel like going out to watch de fireworks tmr lei.. n its true.. when i say no.. nobody force me to say yes.. im glad tho.. but dunno if its rly dey understand how i feel.. i dun rly see wats going to be fun if i go there not to take pic.. was it selfish to think tis way? mus got ulterior motive den i go.. haiz.. but once again.. science student lo.. im too practical.. diaoz today did pw.. got guoxi around like inspecting us.. duh.. he joined me for lunch.. haha.. rly.. i dunno how to appreciate sumtimes.. how come de pw group like dun share de same mindset as i do.. i wan grace back.. cos i guess she\'d listen.. haha.. anyway..today de grp decided on doing de advertisement.. so stupid.. we have complete de written report dey like thinking of de oral presentation.. duh.. den de advert damn spastic n funny n lame n STUPID.. but haiz.. i jus do it for fun.. guess if de advert rly come out den every1 will luff.. rly.. i dun see de point of making it.. n sumore its so informal.. they think its fun? well i dun rly.. but i shall try to enjoy most of it.. ha.. i always thot of thgs but forgot lei.. so alot of ideas i hav ended up mostly not rly implemented.. haha.. nvm.. heck..but psst.. im acting as a baby in de advert.. gosh.. *or weh or weh* ??? im still coughing.. abit asthma lidat.. dunno when fully recover.. but i suspect i will fall ill again nxt month..weakling me.. haiz.. '),(8784,3375539,'07,August,2004','2004-08-07 00:00:00','    urlLink    jus let de flowers bloom for us '),(8785,3375539,'07,August,2004','2004-08-07 00:00:00','    urlLink    n so its quite a stupid path.. '),(8786,3375539,'07,August,2004','2004-08-07 00:00:00','    urlLink    getting emotional.. '),(8787,3375539,'07,August,2004','2004-08-07 00:00:00','    urlLink    was at fort canning 3 mths ago? '),(8788,3375539,'07,August,2004','2004-08-07 00:00:00','    urlLink    hmm.. oso taken 3 mths ago.. wif some of my ps frens.. '),(8789,3375539,'06,August,2004','2004-08-06 00:00:00','    urlLink    taken like 3 mths ago.. y am i so haggard? -_- '),(8790,3375539,'06,August,2004','2004-08-06 00:00:00','    urlLink    hmm.. my J2 PS members dat left.. haah.. a hell lot.. my j1s like half de size lor.. '),(8791,3375539,'06,August,2004','2004-08-06 00:00:00','    hey..in case i forgot.. dis is for lulu.. thanks alot.. for de strepsils.. yea.. dats sweet of u.. cos dat day i sore throat..  den u n charmaine always ask whether im okie.. thanks.. ;) '),(8792,3375539,'02,August,2004','2004-08-02 00:00:00','    *uhum.. * *uh uh uhum...* got alot phlegm la.. irritating.. den my voice like keep changing.. haha.. cos got block nose.. been sick de pass week.. n nope.. i didnt go see doctor.. where got time.. n besides im too tired to struggle to de doc.. den everyday go home like jus sleep lor.. sleep n sleep.. den ended up late in school.. n de tchers there alot noe me lor.. den im supposed to be \"good\" wat.. so i late den i v paiseh la.. haiz.. but stupid bus lor.. always run away frm me.. :(.. den some days got headache oso.. but still gotta work wat.. haiz.. anyway.. i am recovering le la.. onli dat got a lot phlegm.. dis morning sneezed out alot le leh.. de whole yellow sticky stream of phlegm.. haha.. disgusting lor.. de internal photo exhibition in sch was put up.. last for 2 days onli.. hah..kinda simple n yet stupid.. haha.. but anyway.. still ok la.. got pple see lor.. den got ppl comment.. den.. dis morning felt weird.. cos nicholas say de pics were nice.. n hey seems like alot ppl like de cat pic.. (a pic of a cat in a drain) hmm perhaps it look innocent?? i can act innocent too lor *_* wed i was officially ps pres.. cos got de internal handover ceremony.. haha.. stupid lor..i was like rushing off for chem lecture..den still got to entertain ppl.. -_-.. den i bought cake n drinks..n de j1 took photos of de j2s and oso made a collage of ps members.. den we develop out de photos.. n gave each j2 members de photos as souveniers.. so unfair.. j1 members got like 19 but de j2s got 40 lor~.. -_- n de money?? all come from j1s lor.. so sadz de.. den ahha.. de pig v kiam siap de.. :P gave me a bookmark nia lor.. tsk.. nvm la.. i nv bother to gif her anythg oso.. haha.. den de j2s like wan her to hug me.. haha.. but she shy lo.. haha.. so no hug.. haha.. den too bad lor.. bye bye.. guess i wun rly see her anymore.. haha.. nxt wk i shall rule.. wooooo... but den -_- nxt wk i not arnd.. i gotta go astar research briefing.. gonna see my mentor le.. *anxious* today finnally got a break.. whaha.. cos yippie.. national day..*we are singapore*.. but dunno y.. didnt really enjoyed de morning wat .. AJ race.. stupid.. run here n there.. well perhaps cos my class like not really competing.. ahha.. haiz.. still think its stupid.. ppl jus run ard in amk,bishan, yck.. den dere were qts.. ask all those statistics,name ppl n stuff.. so boring.. den later came back to class for debrief.. haha.. dunno wats wrong leh.. ahha.. de class began toking until like got big problem among us lidat.. deres no problem lor rly.. yea.. like wat Sheng xiong say.. got clique den got clique lor.. haha.. wats de big deal?? den outings cancelled den cancelled lor.. no outings will die meh?? haha.. stupid leh.. i think its cos no incentives la.. so i think Fitri is rite lor.. go do CIp la.. its meaningful lor.. n while doing de work.. we can oso bond watz.. i tot i like said it 4 mths ago bout CIP outing.. haha.. nvm.. anyway.. mus got incentives de lor.. we r practical ppl la.. especially triple science students la.. too much \"practicals\" make use really practical..  those dat has no use we will rule them out de lor.. so no incentives.. den haha.. i dun think ppl will go for outings.. but den.. me myself dun mind la.. but mus ensure i free la.. haha.. every1 rly busy lor.. so outing is quite difficult.. so jus relax.. no outing wun die.. de class is still there.. duh.. and oso.. its still up to de individual la.. he wans to socialize den of cos de whole class will noe him la.. like me la.. ahha.. i dun have a fixed clique rite? cos i always get bored de.. de onli clique i may be in is edwin dat gang lor.. but true enuff dey r nerds.. ahah.. dey get lame de la.. but den still homework occupied their breaks.. so i will get stressed wif them.. duh..imagine u hav alot work nv do.. den dere they r discussing how to solve dis problem n dat.. den u dun understand a bit.. its sounds like foregin language lor..argh.. so.. i siam.. haha.. i change n go to else where.. haha.. sometimes i do prefer to be alone.. yea.. at least can do work in peace.. :P haha.. n so i heard..nxt tues dey wan class outing.. go eat i think.. haha.. i dun mind lor.. ahha.. of cos la.. i so sad.. i\'ll be in sch de whole of tuesday.. got workshop for cca leaders.. boo hoo.. no tiem to catch up my work.. :( den others can stay at home lor.. haiz.. nvm.. gonna study on sunday.. den jus now come backfrn sch saw woon teng.. but haiz.. nv see SL..haha.. nvm.. tmr got bio test.. diaoz.. v think leh de notes.. dunno how to finish lor.. haiz.. better study le.. tata.. n hey.. SL.. though i nv msg u.. pls do noe im thinking of u still .. :D take care.. *hugs* '),(8793,3375539,'02,August,2004','2004-08-02 00:00:00','    fuck manz.. the graphs i made for the pass hour was wasted cos the stupid darn programme hanged!! argh...... now i feel like hanging myself.. heck care le.. i go slp FUCK FUCK FUCK '),(8794,3375539,'01,August,2004','2004-08-01 00:00:00','    å¸¸å¸¸æ˜¯    è—åœ¨æˆ‘ä»¬    æ²¡æœ‰æ³¨æ„    åˆ°çš„åœ°æ–¹     ,     è€Œæˆ‘å§‹ç»ˆ    æ˜¯è¿™ä¹ˆè¿Ÿé’    ,    æ€»æ˜¯è¦åˆ°æœ€åŽ    ,     æ‰èƒ½å¯Ÿè§‰    åˆ°ä»–çš„å­˜åœ¨    .     åŽŸæ¥       å®ˆæŠ¤å¤©ä½¿           å®ƒä¸€ç›´æ²¡ç¦»å¼€                      jiu shi si ge zi:  wo jiu xi huan!  zhi yao ni xi huan, xin qing you duo hao jiu you duo nian qing, shi jie you duo da jiu you duo hao wan ni zhi dao wo zui xi huan shen me ma? wo zhen de zhen de zui xi huan   mai   dang lao  ! '),(8795,3375539,'01,August,2004','2004-08-01 00:00:00','    there are so many things that words can never express to heal a soul in pain, waht one needs is a drink of time distilled in a bottle. i wish i could be there for you, to hold your hand and blot out the pain, but there are some roads that we must walk alone, to make us stronger, to make us wiser, and to prepare us for the more rough paths ahead. please be strong.  today the winds and the waves are in revolution, perhaps they are in agony just like me right now.     \"does miles apart truly seperate us from those we love? if we want to be with someone, aren\'t we already there?\"     cont..  \"all that matters are the lessons learned along the way. thanks for walking along this road with me.. \" '),(8796,3392887,'25,May,2004','2004-05-25 00:00:00','    I am currently reading The Teachings of Don Juan: A Yaqui Way of Knowledge by Carlos Castaneda. So far the book is pretty good, my dad gave it to me and told me it was about how to become a physicist. At Fermi Lab right now I am surrounded by mostly people that have PhD\'s in physics or like me are graduate students in physics. All in all a pretty smart bunch of people. I tend to shy away from them, because in my head they are all smarter than me and never had any problems getting things to work like I do, so in fact, they are better, but are they? or are they just older and more experienced?  Tonight I read an interesting passage in the book that I thought I would pass along:       \"When a man starts to learn, he is never clear about his objectives. His purpose is faulty; his intent is vague. He hopes for rewards that will never materialize for he knows nothing of the hardships of learning.      \"He slowly begins to learn -bit by bit at first, then in big chunks. And his thoughts soon clash. What he learns is never what he pictured, or imagined, and so he begins to be afraid. Learning is never what one expects. Every step of learning is a new task and the fear the man is experiencing begins to mount mercilessly, unyieldingly. His purpose becomes a battlefield.       \"And thus he has stumpled upon the first of his natural enemies: Fear! A terrible enemy - treacherous, and difficult to overcome. It remains concealed at every turn of the way, prowling, waiting. And if the man, terrified in its presence, runs away, his enemy will have put an end to his quest.\"       \"What will happen to the man if he runs away in fear?\"       \"Nothing happens to him except that he will never learn. He will never become a man of knowledge. He will perhaps be a bully, or a harmless, scared man; at any rate, he will be a defeated man. His first enemy will have put an end to his cravings.\"       \"And what can he do to overcome fear?\"       \"The answer is very simple. He must not run away. He must defy his fear, and in spite of it he must take the next step in learning, and the next, and the next. He must be fully afraid, and yet he must not stop. That is the rule! And a moment will come when his first enemy retreats. The man begins to feel sure of himself. His intent becomes stronger. Learning is no longer a terrifying task.       \"When this joyful moment comes, the man can say ithout hesitation that he has defeated his first natural enemy.\"       \"Does it happen at once, don Juan, or little by little?\"       \"It happens little by little, and yet the fear is vanquished suddenly and fast.\"       \"But won\'t the man be afraid again if something new happens to him?\"       \"No. Once a man has vanquished fear, he is free from it for the rest of his life because, instead of fear, he has acquired clarity - a clarity of mind which erases fear. By then a man knows his desires; he knows how to satisfy those desires. He can anticipate the new steps of learning, and a sharp clarity surrounds everything. The man feels that nothing is concealed.\"   Maybe we all get to a point where we are too scared to continue. For some this would probably happen much earlier in life. When I tell people that I am a physics graduate student they give me this look of terror and say things like, \"oh my god I could never do that,\" or \"you must be so smart\" I want to tell these people that no, I am really probably a lot like you I just get up every day, knowing that I am probably going to have difficulties in doing whatever it is I need to do that day. However, it always gets done, there is always people around to help....and there is always my goddamn Welty stubborness that keeps me trying. Who knows where I will end up, right now I am terrified of being here and what I am doing, but i am hoping that as the days and weeks go on I will gain some insight to what I am doing and become more comfortable with it.   '),(8797,3392887,'21,May,2004','2004-05-21 00:00:00','    Well here I am at Fermi Lab at 1:10 in the AM. My office is in the sub-sub basement...yes that\'s three stories beneath the ground. I am about 20 feet away from the D0 detector, which I know sounds cool....but really it\'s just a wall which I know the detector sits behind. Not only am I lucky enough to be below three levels of sunshine, my desk is in a huge cavern. So imagine a huge room with 3 story ceilings and then some desks thrown in the middle. Yowza! does it get better than this!   Anyways, looking at my schedule for the next week I have just planned to not sleep for many days. Jason and I are working these owl shifts this week and so our schedule is all screwed up. It\'s alright for now, but the end of the week brings meetings that are unfortunatly not at 3am.   That\'s all for now. i will try and add some photos off my lovely office space in the upcoming week.   '),(8798,3392887,'21,May,2004','2004-05-21 00:00:00','    What is the  definition of marriage?  Such a simple looking question but lately in this country is has gotten more and more complicated. Recently Massachusetts has allowed gay couples to marry. Listening to a morning radio show the other day which was discussing said topic, I was surprised by what many of  the  callers had to say. Their thought was that hey, I don\'t mind it, but you can\'t call it marriage because \"God said marriage is between a man and a woman\". But it\'s ok if it\'s a civil union. So my question to these people is what about all the straight people that get married a Justice of the Peace? Like me for example. For the rest of my life, should I refer to my future husband as my civil partner? If we are at a party will I give the following introduction, \"Hi, my name is Leah and this is my civil partner Jason\" ? Since we are not getting married in the church and are obviously beyond sinners by these peoples standards do they say the same thing about us as they would a gay couple?   My real question is...Now it\'s a good one :  HOW DOES IT AFFECT YOUR LIFE?  Seriously, how does the marriage of two men or two women do anything to you, your family, your children, your marriage etc?   I think that everyone in this country has a right to get married. And by married I mean married, not civil unioned.   '),(8799,3392887,'20,May,2004','2004-05-20 00:00:00','    urlLink Michael Moore.com : Mike\'s Message : \"going\"  This is Michael Moore\'s comments on the recent events involving his newest movie Fahrenheit 9/11.   '),(8800,3392887,'19,May,2004','2004-05-19 00:00:00','    I just moved out to Fermi Lab a couple of days ago. My fiance and I are living in my dad\'s house for free this summer to save some money. However, come October we will back in the great city of Chicago. After a rather frustrating conversation at the Lab yesterday I got to thinking about why more of the grad students don\'t live in the city. Obviously there are a lot of issues to consider...money, commuting which of course involves traffic, city noise? Well I could go on and on and maybe my view is biased by the fact that I grew up in Chicago a couple of blocks from Wrigley Field so I can appreciate what the city can offer. I agree that it will be more expensive to live in the city due to higher rents, gas money food and so on. However, at the end of th day and on the weekends I get to say, \"But wait, I\'m in the city.\" The city has so much to offer. The city holds me in awe all by itself. I don\'t need to do anything to enjoy it, just being there calms me down. Then you add to that the amount of things to do, all the museums, bars, clubs, beaches, ball games, starbucks...wait those are everywhere. My point is that spending the bit of extra money and dealing with commuting traffic pale in comparison to just being there and having all the culture and excitement right outside your door.   '),(8801,3392887,'19,May,2004','2004-05-19 00:00:00','    My name is Leah Welty and I am new to the blogging world, just here to add my input about the world around me   '),(8802,3392887,'30,June,2004','2004-06-30 00:00:00','    urlLink    This is a picture of my fiance and I. How cute are we!&nbsp; urlLink   '),(8803,3392887,'11,June,2004','2004-06-11 00:00:00','    My dad is a physics teacher in what used to be one of the best high school science programs in the state. He has gone to work there every day for the past twenty years and for him it wasn\'t a \"job\". He got to go to school everyday and work with kids on expanding their physics horizons. Along with the rest of the science faculty they have written grants to get free computers for the classrooms, written the text for the courses they teach because they know what is necessary to get the kids to learn, involved both kids and their parents with outside of school activities like an astronomy night, a giant maze that they built in the wall of the building. He gets into the building usually about 6am or so and doesn\'t leave until 5 or 6 pm. I write about my dad because he is the faculty member that I know the best....However, the rest of the faculty follows this same behavior.  Due to a new administration this science department has been beaten into the ground and dismantled...In literally less than a years time. One of the faculty members has been transferred to another school in the district without given reason. Not just any teacher, one of the best physics teachers in the department (after my dad of course). This teacher has done so so much for the kids in the school, and so I don\'t understand why the school would want to get rid of him because of what looks to be someone\'s personal agenda. Along with this teacher being transferred the faculty has proposed to add another course to the curriculum due to a change in requirements. They offered to develop it, write the needed texts and make what they know will work best for the kids that would be needing this course. The administration said no. Even though 100% of the teachers that would be teaching this course think that it is the best thing for the students. I guess the students get the shaft again. Good job administrators.  I recently attended a board meeting for the school and the other two schools in the district and was shocked to see what actually happens at a school board meeting. After about forty five minutes of students and teachers giving speeches about the teacher that is being moved, many of the speeches very moving and highlighting the behavior of the administration, the meeting just went on to Action Item #2. I was so appalled because I felt that the comments of the students and teachers deserved something more than moving onto the next item of the agenda. The meeting then went on to discuss apparently more important topics involving education, like the price of parking for thirty minutes and the type of food that will be offered at the schools for another twenty minutes.   At NO time did this board discuss education. So I ask, what\'s the point.   This recent development in the science department is the most childish thing I have ever seen. It\'s like being in grade school and having to separate the kids that don\'t get along, or putting a kid in the hallway. The administration will say, well you don\'t really understand what\'s happening. So explain it. Tell the department why things are happening, and don\'t give them some bullshit reason, these are not stupid people. Tell them why you don\'t want to them to develop a new course, why you think that this would be a bad idea...Instead of just saying, \"no.\"  So to the administration of school: GET OVER IT. Stop being vindictive and let the teachers do what they do best : TEACH. The like to do it, they like to help the kids learn, they enjoy it. But like my dad has said, they don\'t have to. They could come in everyday from 8am - 3pm, stand at the board and lecture to kids out of a textbook, which is the direction that this department is heading. As leaders of educators, do you want to lose the education part? If so, you are doing a great job and keep up your behavior. If not, Back off.   '),(8804,3736053,'25,June,2004','2004-06-25 00:00:00','   Here i am. Finished my two days this week, sitting and drinking tea while it\'s about 35 degrees celsius outside. I\'m sweating like a pig. This heat is amazing! The last time i experienced something similar was last year when i was studying spanish in Sevilla. In the end of may we had temperatures closer to 40 degrees, still a different heat though.  This weekend i want to do a lot of stuff...  I\'m planning to work a little bit on that photojob i got, there are some people that want some 30 creative photos. So i\'m brainstorming about situations or places where i can shoot thoose...  And i want to go out and party some tonight, for the first time in two weeks since i was working all weekend long last week. However, this weekend is short, on monday i work again, and it will be monday til friday...  Still not sure...  I wan\'t to do something else during the days, something more creative, but how does that and money go together, well... offer me a job within something creative and we have a solution.  Now: Siesta and food  Hasta luego.  '),(8805,3736053,'25,June,2004','2004-06-25 00:00:00','   I\'m tired, it\'s my sixth day working, not in a row but scattered on two seperate weeks. I was looking forward to finish my studies at Centro de Lenguas Modernas and i did, even with a good result. In the end it felt like something i just wanted to get overwith, to end something and begin something else, the time after the studies, find a job, intent to settle myself in this city, and just become sure that i don\'t have to leave when my studies finished...  And my plan worked, but still i miss a lot of things in my life, i still don\'t have the job of my dreams, but i\'m still able to survive with the job i got right now. Another thing is that i would like to move to an apartment of my own. Something more personal, but i guess i\'ll have to wait.  I guess things come when i less expect them.   And i believe in surprises.  '),(8806,3736053,'28,June,2004','2004-06-28 00:00:00','   Monday, job is done, and i\'m dying of starvation and this heat. I was on my way to get some food but i though of writing down some words before i keep going on with my normal life.  This weekend was great, exactly what a weekend should be...  It began with Granadas gaypride, small in it\'s size, but with a lot of fun people. Later that afternoon i was DJ:ing, actually for the first time in almost 2 years, and hell, i realized i was missing that soo much!! Played some tunes for about an hour and i felt like a new person. A mixture between jazz, electronica and some beats.  The day was crazy though, a heat together with drinking for many many hours... and a lot of photos made. Went to bed quite early to get up in the morning, not that early, but to go to the beach... and hell... yesterday was an even hotter day, laying on the beach, under the sun, jumping into the mideterranian sea many times, drinking about 5 litres of water... and driving back between the mountainous passages of Sierra Nevada.  Today my body hurts a little bit, burned by the sun, but not so tired, i guess i\'m begining to get used to all this working. Waking up early, the hours passes by really fast, and then Silvia, my friend takes over...  In my mind there are many things, thoughts of jobs, future plans... however... my life is liven day by day, i don\'t think about tomorrow and neither yesterday.  This is me, stereosampling.  '),(8807,3736053,'05,July,2004','2004-07-05 00:00:00','   So this is what has happened. This saturday the owner of our falt came by with some people... saying: - we\'re goig to sell the flat. And since our contract officially ends now in July, even if we can unofficially live there for another 2 months until September 1:st, the situation is a little bit tricky. That is... it\'s difficult to find a flat in September, and even though i love Granada as the place i fell in love with about 2 years ago i feel that a change of atmosphere would be in it\'s place. Furthermore... finding a job here is difficult so, after throwing around some 100 CV\'s and being very very optimistic and happy towards the poeple iv\'e met the only thing i have is the job in the hostel. Yes, of course, it keeps me alive but i\'m looking for something more... with possibilities towards the future. I know you can\'t build a city in one day and neither a life in a different city, but with the background and experience i have in the world of creativity there should be more to find... or at least more response from... whoever who have seen my CV and my CD with all the stuff i\'ve done the last years.   So the solution was... find another place in Granada and keep being optimistic about an unknown future... or move to another place.   And this whole thing about Barcelona isn\'t actually something new, it\'s a 3 months old thing that have been putting questionmarks in my head.  So... this thing with suddently having to leave the apartment, and choosing to do that at this point and not within 3 months... i feel it like a push forward. Like, take a chance, go there and see what the big city has to offer. And i do believe a lot in Barcelona, not only that it\'s bigger but also all the things this cosmopolitan city have to offer.   So... we\'ve put the date to August 2nd... but it might change depending on the circumstances.   Me and my friend Roman will rent a car and then drive up along the coast with all our stuff, and as we see it, an old life will end and a new one will begin.  An adventure of some kind, something i\'m looking forward to.  Before the weekend i was thinking of leaveing this job this week, but life might change very suddently, so, i\'ll stay until i go. I will miss the feeling of meeting new people every day, show them my favourite places on the map of Granada. But i have to get something done in my life, at least something inside me pushes me forward. And somehow i see myself over there, along the northeast coast of Spain, in Barcelona. Images created in my head, but with theese dreams and my optimism i will reach that.  Stereosampling.  '),(8808,4166300,'07,August,2004','2004-08-07 00:00:00','   My cover has been blown, Bun-Bun has discovered the place where I talk to myself and hope that nobody actually cares enough to listen.  I can already hear Alex referring to me as a \"Touchy-Feely-Pussy-Girl\" at school.  I leave to do a tour of drill instructor duty at Fort Hunter Liggett this saturday, unfortunately its Advanced Orientation, so I\'m laying the smackdown on 15 ten-year old kids who can\'t sit still for a week.  AO is the wild west, and Seaman (Airman soon, motherfuckers!) Montgomery is the new sheriff in town.  Or Deputy Marshall.  Or a wandering vigilante whose parents were killed my marauding League Cadets when he was just a wee lad.  But most likely I\'m the yeoman.  *Does his little yeoman dance* '),(8809,4166300,'07,August,2004','2004-08-07 00:00:00','   It\'s 0233 and I\'ve got insomnia... Slept over 2/3 the day yesterday and I\'m not even close to being tired.  It\'s so boring walking around the house alone that I wish I could fall asleep.  I\'ve been keeping myself content eating boxes of triscuits and doing this blog crap.  It\'s beginning to become very therapeutic, I\'ve always wanted a place to complain about things.  I lost my USMA candidate number and password, so I\'m pretty much shafted unless I want to go crawling to my MALO begging for him to bail me out.  I intent on floating lethargically through Sunday in a symbiotic relationship with my rack.  No church Monty?  Oh well, I\'m going to hell anyway. Me, Hitler, Charlie Manson and Beelzebub can get a poker game together every other Friday.  I think I\'ll go to CenterStall and watch BJ Hall bring the pain over and over until breakfast time.   Status Check: Morale - 23% Hooah-ness - 35% Boredom - 89% Optimism - 12% Conclusion:  THE PAIN TRAIN\'S A-COMIN\'! '),(8810,4166300,'07,August,2004','2004-08-07 00:00:00','   My attempts trying to be an intellectual:  I think the best way of doing good to the poor, is not making them easy in poverty, but leading or driving them out of it. In my youth I traveled much, and I observed in different countries, that the more public provisions were made for the poor the less they provided for themselves, and of course became poorer. And, on the contrary, the less was done for them, the more they  did  for themselves, and became richer.  -Benjamin Franklin, The Encouragement of Idleness, 1766   The inherent vice of  capitalism  is the unequal sharing of blessings; the inherent virtue of  socialism is the equal sharing of miseries.      - Sir Winston Churchill         How do you tell a Communist? Well, it\'s someone who reads Marx and Lenin. And how do you tell an anti-Communist? It\'s someone who understands Marx and Lenin.    -Ronald Reagan         Communism doesn\'t work because people like to own stuff.       -Frank Zappa   Alas, it is all for naught. '),(8811,4166300,'07,August,2004','2004-08-07 00:00:00','   That post title is so cliche.  Don\'t waste your time on this blog.  It\'s nothing but a gigantic load of bullshit.  Who the hell reads this crap anyway?  Am I so egotistical I think that people from all over the globe who probably don\'t know me waste valuable time reading a site about my life?  Tell you what, if you\'re still reading this right now, you need get new direction in your life or something.  Stop reading this, your life is ticking away and you\'ll be worm food soon.  Go outside.  Do some community service.  Make sweet love. Pick daisies.  Frolick in the meadows.  Carpe Diem or whatever-the-fuck.  Join the military.  God knows they\'ll give you enough trivial pursuits to occupy you besides read this.  Monty makes a blog?  That\'s got to be one of the harbringers of the Apocalypse.  I\'ve always fancied myself as an Antichrist...  but really, is there anything I can say to satisfy your desire for tasty tidbits of information on my comings and goings?  Are you not entertained!?  Maybe its just me... being very... very... bored.  Rest in Peace Superfreak...  Rick James 1948-2004 \"Cocaine is one Helluva Drug\" '),(8812,3632734,'25,June,2004','2004-06-25 00:00:00','    UGH!!! TODAY IS FINALLY OVER!  Here is how the day went...  7:30am - Wake up call... Nicolas decides to wake up... ERRR! What happened to 9 o\'clock!?  9:00am - Michelle\'s doctor appointment... Nicolas decides he doesn\'t want to be in the doctor\'s office... so we walk around for at least 30 minutes and about every 2 minutes he decides to drop his butt to the ground and pitch a fit.  10:00am - At the hospital and Michelle is with the Lactation Consulant Amy. Nicolas again doesn\'t want to be in the room... SOOO We walk around again but for this time it\'s at least an hour. I decided that he might be hungry so I took him in the elevator down to get food stuffs and NOW he decides he wants to be in the elevator the WHOLE TIME!!!!  11:30amish - Adam calls... THE SYSTEM IS DOWN!!! One of the servers puked and all the sites are down... It would have been a GREAT TIME FOR A NAP, but alas... I am the Jesus... Our savior... I spend like 2 hours fixing the server and everything is great.  2:00pm - Nicolas decides to wake up... I take him downstairs and make him Mac and Cheese... what does he do you ask? DOESN\'T EVEN TOUCH IT! ERRR I swear to God... When he gets older... if I make him something, he better eat it or it gets the hose again!  5:30pm - Time for dinner... I cook, I clean... yay!  7:00? - Nicolas gets a bath... he cries the whole time... I don\'t know why... but I didn\'t brush his teeth :)  8:00pm - Awwww... relaxation... watch a little SNL w/ Lindsay Lohan... it was pretty funny.  10:30pm - I got a hair up my butt and decided to go play DDR... I completely suck now. I haven\'t played in a couple of weeks and it shows.  12:15am - I go to HEB to pick up a yummy treat for Michelle and to pick up cat food. Alas... I forgot the damn $$$ to pay for the stuffs. Do I look like a complete tard when I tell the HEB guy I forgot my Lonestar card at home. :P  12:25am - Back at home... everything is good...  1:20am - I now go to sleep... will I dream?   BTW - Some funny shit I found today on CNN... YAY!  \"On the other hand, I wouldn\'t just pull up and leave there if I thought it was going to cause the whole thing to disintegrate,\" Clinton told King.  Hmm... I wonder why he didn\'t say... \"I wouldn\'t just pull out\"? HAH!   '),(8813,3632734,'14,June,2004','2004-06-14 00:00:00','    I should be working...   '),(8814,3987879,'17,August,2004','2004-08-17 00:00:00','   Today was another aerobic day. I spent an hour on the stationary bike and burned 504 calories. This is considered a light to moderate aerobic workout but my room mate makes it out to an intense energy draining exercise. So much for the Atkin\'s Diet.  There\'s a formula that says you need to burn 3500 more calories than you take in to lose one pound of fat. This isn\'t totally correct because your body doesn\'t just use your fat as a fuel source when your body isn\'t getting enough calories to burn from food. Without proper training and diet, you could lose muscle mass once your glycogen levels are depleted and your fat stores barely touched. If you are losing more than two pounds a week, you are most likely losing muscle mass along with your fat. In my case, I don\'t want that. I actually don\'t want to lose any weight.  Weight training burns more calories than an aerobic workout for the same period of time but it\'s a carbohydrate fueled exercise. It\'s going to burn fuel from what you eat and what\'s stored as glycogen. Aerobic exercise is fueled mostly by free fatty acids which come from your body fat stores. I\'m not saying weight training is not good for burning fat because weight training does increase your metabolic rate more than aerobic exercise does. You need both. It\'s just that aerobic exercise is a better indicator on how much fat you are losing during an exercise. Then the weight training helps you to store whatever energy you are getting from food back to the muscles and not to fat. That\'s my key goal in my workout schedule. It\'s not to cut 3500 calories from my diet, but to burn 3500 calories from fat stores with aerobic exercise to lose that pound of fat.  '),(8815,3987879,'16,August,2004','2004-08-16 00:00:00','   Today is chest and tricep day. This is my weakest area. I couldn\'t even bench 135 for more than one rep and couldn\'t do more than one unassisted dip (I did try the dip after the chest workout though). I\'m happy I\'m losing the fat but it just shows how small my chest really is. My room mate suggested doing strength training on the chest and tricep day and worry about bulking that area up later. I\'ll keep pushing through. I\'m not expecting a huge chest overnight.  After the chest and tricep workout, we did cardio on a stationary bike for about 20 minutes. I kept my heart rate up to 150 but it took me about 5 to 6 minutes before my heart rate even got up to that level.  My room mate keeps trying to go on the Atkins diet but never seems to last a week on it. I personally think it\'s stupid to deprive yourself of carbs. I think it\'s better just to watch your calorie intake if you are wanting to lose fat. Since I\'m trying to lose fat and not lose lean body mass, I\'m more about just increasing metabolism than anything else. I\'m not even watching the calorie intake that hard. I\'m also not overloading on food for bulking up either (I may do that later when my body fat percentage gets low enough). I haven\'t lost any weight since I\'ve been working out, but you can see by the pics, I have lost some fat. I\'ll wait until about the two month mark before I get an appointment to check my body fat percentage again. Until, then I\'ll keep working hard on that midsection.  '),(8816,3987879,'15,August,2004','2004-08-15 00:00:00','   Day 8 pic&nbsp; urlLink     urlLink    - You can still see the love handles there, but I can definitely tell I\'m losing fat in the midsection. And, pardon me for the farmer\'s tan.  '),(8817,3987879,'15,August,2004','2004-08-15 00:00:00','   Today is a day of rest. Well, sort of. No weights and no cardio today but every Sunday my room mate and I babysit a seven year old girl for a woman with work with (I used to be really infatuated with the mom; I guess I still am a little bit). You can\'t really rest being around a seven year old. They have way too much energy and this girl is above the norm in energy. When she is playing on the moneky bars, I\'m amazed by the strength of a little girl who doesn\'t even weigh 50 pounds. She has more calluses on her hand than most bodybuilders I know. If I could bottle that energy into a supplement, I would be a billionaire.  I weighed myself today and was at 186 pounds. So, I am assuming I\'m losing fat quicker than I\'m gaining muscle. I would hate to think I\'m losing both. I may take some more pictures and put them up since it\'s a week since I\'ve started this blog, but I have to go into work soon and don\'t have time to do it now.  '),(8818,3987879,'14,August,2004','2004-08-14 00:00:00','   My room mate and I did legs and shoulders today. I did a little more on leg presses than I\'m used to (360 pounds instead of 320 pounds), which should be a good thing. My room mate said that by next year we should be doing 900 pounds. I\'ll wait and see.  I\'m trying to talk my room mate into getting into a more regular schedule because of moving the leg day back one day. He wants to do shoulders on a seperate day than legs. I\'m willing to compromise but that means four days a week instead of three days for weight training and he is already missing some days. I proposed that we do chest and triceps on Monday, back and biceps on Wednesday, legs on Friday, and shoulders on Saturday. Do thirty minutes of cardio on days that we do weight and an hour on the days that we don\'t do weight. I think he wants 45 minutes on days that we do wieghts. I think that\'s a little to much even though I think Bill Davey says he does 40 minutes every day.  There\'s another friend of mine who wants to work out with us too, but that means we would have to go to a different 24 Hour Fitness than the one I\'m going to now. He also has a four year old daughter which won\'t be going to school yet. So, I\'m still not sure this will work out with my room mate\'s and my schedule. I think that\'s one of the hardest parts about working out; finding a good schedule. Nobody likes going to work out by themselves, but they have a hard time to schedule to go with someone else.  Today, I have been a little bit more thirsty than usual. I do get to a point of drinking too much water and have to go urinate about every fifteen to thirty minutes. I know it\'s important to stay as hydrated as possible and drinking a lot of water a day keeps you feeling full and you don\'t seem to eat as much. I used to do this a lot before, so, I don\'t really think it\'s the diurectic effects of Rev XP. I\'m sure it\'s much better peeing a lot than not knowing if you are dehydrated or not.  '),(8819,3987879,'13,August,2004','2004-08-13 00:00:00','   Today was suppose to be leg and shoulder day but my partner woke up late and didn\'t have time to go work out before he had to go to work. So, I went in by myself and just did 30 minutes of cardio workout on a bike. Tomorrow we\'ll go do leg day.   I\'m not sure I would of been able to handle legs today anyway. Last night I spent a couple of hours after drinking some Guiness with another friend of mine and didn\'t get to sleep until about 5 am. I don\'t drink very often and I know it\'s not the best thing for me when I\'m trying to cut a lot of fat. But, what the hell; it was only one night.  I have my spherical ab pic in my Yahoo! profile and have had a few people from a bodybuilding chat room about it. Some where just assholes and they usually don\'t have a pic up themselves and others kind of understood what I was doing, which is showing my progress.  A couple of people at work think they already see results in me. I\'m not sure because it\'s so soon and measurements really haven\'t changed much except for about an inch and a half of my waist. I\'m sure I\'m slimming down; just not to the point where I think it would really be noticable. I\'ll still take any encouragements and compliments I can get.  '),(8820,3987879,'12,August,2004','2004-08-12 00:00:00','   I woke up this morning with pain in my back right shoulder and neck. I didn\'t think I over did my workout yesterday and I usually feel the pain two day later instead of the next day. Maybe I slept on it wrong plus the fact that I worked out my back yesterday. I think I need to get a bottle of Advil or Ibuprofen.  My buddy and I went to the gym today to do a whole hour of cardio. Once he saw that I was burning more calories than him with both of us having the same heart rate, he had to pump up his heart rate because he\'s kind of competitive that way. We both ended up burning around 500 calories during that work out and then went out for chinese afterwards.  At the chinese resturaunt I saw someone who looked pretty much like my ex-girlfriend that I had 5 years ago except that she was much healthier looking. It makes you wonder, but I\'m not going to dwell on it.  I started hanging out in a  urlLink Yahoo!  chat room on body building last night. I\'ve been hanging out there a little today. Doesn\'t seem like they talk about bodybuilding too much. Seems like a place where most people try to go see other buff people on web-cams or to spam porn to people. I have had a few good conversations with young beginning bodybuilders asking for advice. I\'m not sure how many would take my advice looking at my current physique, but I\'ll give what information I can. Hopefully, some of them will keep track of my current progress and see the results for themselves.  '),(8821,3987879,'11,August,2004','2004-08-11 00:00:00','   Day 3 pic. Abs don\'t look as bad as the sphere pic.&nbsp; urlLink     urlLink  '),(8822,3987879,'10,August,2004','2004-08-10 00:00:00','   I subscribe to the push-pull method for training. I do chest and triceps one day, back and biceps another day, and then legs and shoulders on a third day. In between the days I do cardio and work on my abs. I take usually just one day off a week doing nothing, sometimes two.  I used to be really skinny up until I was 27. Never weighed more than 170 pounds until then. I used the same method as above then and went up to 198 pounds. Then I moved out to California for a computer related job and sat behind a desk for about five years. I even got up to 210 pounds then and it was all fat.  Today was back and bicep day for me with a short 20 minute cardio on a bike. I\'m having problems with my left arm being so much weaker than my right. If anyone knows how to get both arms about even in strength quickly, let me know.  '),(8823,3987879,'10,August,2004','2004-08-10 00:00:00','   My friend and I went to do 45 minutes of cardio and watch some Riki Lake today at a  urlLink 24 Hour Fitness . I actually did about 30 minutes at 65% and 15 minutes at 85% of my target heart rate. I forgot exactly what my target heart rate is even though a personal trainer figured it out for me one day. I just went by what ever the machine on the bike told me for my age plus upping it a couple of numbers. I\'m thinking about just doing the 85% cardio zone instead of the 65% fat burning zone just to keep up my metabolism for the rest of the day. This is something I used to preach around 10 years ago that sprinting is much better for fat burning than jogging just because of the increased metabolism for the rest of the day.  On Day One I gave overall measurements that still don\'t take fat into considerations. When I first starting working out a personal trainer used calipers to measure my body fat percentage. She came up with 24%. I knew it was high but didn\'t know it was that high. 25% body fat for men is the line on obesity. I do have some subQ fat but very little of it. My main problem is visceral adipose tissue which is that midsection fat.  A few days ago I let a  urlLink GNC  clerk talk me into trying a product called  urlLink Rev XP . It\'s seems to be mostly just a lot of methylxanthines ( caffiene and caffiene relatives). I\'ll finish the bottle and if I don\'t see dramatic improvements by the time I finish the bottle, I\'ll most likely move on to  urlLink Ab-Solved  next.  Here\'s some more stats that I didn\'t mention before. I\'m 6\'1\", 34 years old, and started off my training at 191 pounds. I\'m at 188 right now. My personal trainer said I will probably get to 180 pounds before I get back up to 190 because fat loss will occur quicker than muscle gain. I set my measurement goals in Day One with the assumption I\'ll be at those sizes with around 10% body fat. I\'m not sure what the total weight of my measurements would come up to but I\'m estimating around 210 to 215 pounds. This means I need to lose around 25 pounds of fat and gain about 45 pounds of muscle. Like I said before, this is a lofty goal and is going to take awhile, but I can do it.  '),(8824,3987879,'09,August,2004','2004-08-09 00:00:00','   The Sphere. Scary isn\'t it?&nbsp; urlLink     urlLink  '),(8825,3987879,'09,August,2004','2004-08-09 00:00:00','   This could be Day Fourteen or Fifteen since I started working out about two weeks ago. But, I might as well start off the blog simple with Day One.  For the last 5 years I\'ve been sitting behind a desk and my midsection got to a point where it was too spherical for me. It\'s 40 inches around while my chest is only 41 inches around. I\'m not too happy about that.  I started thinking what would be the perfect body size for me and to do something about it. I always thought Bill Davey was a good size to be, but could only find measurements for his chest, waist, and arms. I did some more looking on the internet and found this nice  urlLink calculator  basing measurements just on wrist size. The problem is that it still doesn\'t show anything about the midsection. I guess if you can see a well defined six-pack it doesn\'t matter what size it is.  My wrist is 7 inches. I\'m going to round off and adjust to where I think I want my goal to be. I want (where I\'m at to where I want to be):   Chest: from 41 to 46 inches.  Waist: from 34 to 32 inches.  Hips: I\'m at 39 inches and want to stay there.  Biceps: from 13 to 17 inches.  Forearms: from 11 to 13.5 inches.  Thighs: from 21 to 25 inches.  Calves: from 15 to 17 inches.  Neck: I\'m at 18 inches and want to stay there.  This is a somewhat lofty goal and is going to take awhile to get there and I believe I can do it.  I\'ll try to keep this updated every day and talk about some of the training and supplements that I\'m using 2 B Cut.  '),(8826,3114279,'27,March,2004','2004-03-27 00:00:00','   Nick : \"As a lower-class kid, I was raised to think success would be owning stuff. Having that great job, too. Now I find my parents\' dream was wrong. You never really own anything. And you\'re never really finished as a person. Still, you can be stuck, trapped by possessions and an outdated self image. Other than that, I think humanity is perfect. We\'re making the same mistakes we made 1,000 years ago. So they must be the right ones. So relax.\" - Chuck Palahniuk '),(8827,3114279,'25,March,2004','2004-03-25 00:00:00','   Nick : I went jogging last night, for the first time in months.  I was pretty appalled at how completely exhausted I got and how sore I am today.  I used to be able to jog for a few hours like it was nothing.  I was so tired that I could barely speak to order my soda at the express afterwards.  I\'m going to go again tonight and hopefully get back into some sort of shape.   As I was jogging, some guy drove up next to me, leaned out of his truck, and told me I was \"going seven.\"  There is a really great  urlLink interview with Richard Kelly , the guy who wrote/directed Donnie Darko, about that movie, and his amazing-sounding upcoming projects, as well as the possibility of a Donnie Darko : Director\'s Cut.  Also really cool to hear about the possibility of Terry Gilliam directing a film adaptation of the book  urlLink Good Omens  by Neil Gaiman and Terry Pratchett.  There aren\'t any movies out currently that I\'m dying to see, except maybe Eternal Sunshine of the Spotless Mind, which I guess I\'ll have to drive to Jacksonville to see. '),(8828,3114279,'25,March,2004','2004-03-25 00:00:00','   Nick : post '),(8829,3114279,'21,April,2004','2004-04-21 00:00:00','   Glenn : Oh, physics. That class is the bane of my existence. It forces me to use tired old cliches when describing it. Is tired old cliche a tired old cliche? I don\'t know. Anyway, my poor Senators were eliminated by the stinking Toronto Maple Leafs yesterday. Learn to use proper grammar, you CN tower bastards. It\'s   leaves  , not leafs. Morons. Anyway, this weekend is the last day of shooting for my T.A.P, including a scene with Nick and Ryan, an outdoor scene, and the infamous Punk Rock Amy scene. We\'ll see how it all goes.   On the other side of things, I met my opposite yesterday. Now, as all of those who know me know, I am in this premed thing to become a forensic pathologist. My reasons are trifold; first, I hate people, and this way I don\'t have to deal with people whining \"my arm hurts,\" \"I have a rash,\" \"I can\'t breathe,\" or \"I\'m bleeding.\" Wah, wah, wah. This way nobody talks, and I don\'t have to listen to their trash. Secondly, I don\'t think I could deal with the pressure that my decisions could result in somebody living or dying. This way they come to me dead, they leave me dead, anything other than that happens and it\'s just a bonus. Third, I kind of want to one day be that old crusty pathologist you see in the movies, unshaven, eating a baloney sandwich, when the sheriff busts in and says \"Shaheen, I don\'t like you and you don\'t like me, but we\'ve got a killer on the loose and you\'re the only one with the know-how to stop him.\" So that\'s basically all laid out. This guy in a couple of classes starts talking to me the other day about how he wants to be a doctor not for the money, but so he can help people. He just loves people. He wants to do mission trips and crap. He says one of the greatest things he\'d like to say to somebody would be \"I\'m sorry, but I can\'t help you anymore, but Jesus Christ can.\" Now, I don\'t know about you, dear reader, but that\'s about the goddamned last thing I\'d ever want to hear come out of a doctor\'s mouth. Take two Jesi and call me in the morning. Fuck that! He said he wants to witness miracles. I told him I don\'t want to see any miracles in the morgue, or I\'d high tail it out of there. Yikes.  Been on a Sergio Leone kick lately. One of the greatest filmmakers in history. His best film is easily Once Upon a Time in the West. Excellent.  Three people show up to meet Charlie Bronson at a train station, all of them with guns drawn. Charlie: \"You bring a horse for me?\" Bad Guy: \"Looks like we\'re one horse short.\" Charlie: \"No. You brought two too many.\" '),(8830,3114279,'19,April,2004','2004-04-19 00:00:00','   Glenn : Yo yo yo. We shot some scenes for my awful movie this weekend. Turned out really damn funny. Got reminded constantly that I am a bad director and actor, but that\'s how it is. It\'s just for fun. Also saw Kill Bill Vol. 2. Fucking awesome movie, no matter what my friend Ryan may say. I can\'t let forth any details, because certain people who also write in this blog haven\'t even seen vol. 1 yet. And they rhyme with ick. And start with an N. Senators tied their series. Go, O-town. '),(8831,3114279,'16,April,2004','2004-04-16 00:00:00','   Glenn : Yo yo yo, this is Glenn, the lazy guy who never posts. Service zone sucks, yadayadayada. Anyway, heading to the great lake today for some shootage of my movie, which is tentatively called \"Glenn\'s terrible-ass project where he proves he made the right decision in not attending film school.\" Ha ha, just kidding. It\'s called \"T.A.P.\" for short. Looking forward to spending my tax return on pointless crap. And food. I\'m going to go eat at Mori (local sushi place) when that republican money starts coming in. Screw social betterment, Imma git me an xbox. Nick is going to see KBV2 with us tonight, whether he likes it or not. If he\'s broke we\'ll scrounge the cash, or make him sell his body. I think a Tarantino flick in the theater, even if it is the crappy ass lake city theater, is worth the abandonment of scruples.   So in poetry club last night this girl who was workshopping her poem was almost in tears by the end. I felt bad, but you shouldn\'t bring in a poem you completely love to be workshopped. You bring it in because you need help with it. I think a lot of people have the attitude that they\'ll bring in a poem and wow everybody with it and will get the response \"Well, I quit writing. This is amazing and needs no work.\" Hell, even I used to bring in poems based on that thought. Now I realize a poem is done when the writer thinks it\'s done, so don\'t bring in that poem to get workshopped. Bring in a poem you don\'t like so much, or have trouble with. People can always critique anything, so leave the stuff you don\'t want critiqued at home. She\'s a good writer, but some people don\'t respond well to criticism.   My new musical collaborator and I started working on our music. Sounds pretty good so far. I listend to \"Nice Ass\" by Free Kitten and loved the sloppy sound. I think that\'s what I want. Anyway, I gotsta get going to Psycho class.  In health news, I feel fine. - Kevin Nealon '),(8832,3114279,'13,April,2004','2004-04-13 00:00:00','   Glenn : You know, lots of people warned me not to work at Service Zone. Lots and lots of people. But I didn\'t listen to them, and worked at that dank pit of stupidity for just three days before quitting. Nobody there knows about computers. It\'s amazing Dell sells anything whatsoever. In that three days that I was there I was yelled at by the payroll lady for Service Zone\'s errors three times, because I didn\'t do things they didn\'t tell me I was supposed to do, and I was lied to about how much I was going to be paid. So I quit. No big deal. I\'d find another job. Little did I know that same lady would lie to unemployment about my status, preventing me from going back to that, and then even after two months of calling not send me my w-2. I thought it wouldn\'t matter, because I only received one paycheck and it has all the tax info, so no big deal. Anyway, I needed the employer number, so I had to call several times once again before I finally talked to her, and she got very angry and told me she had to call corporate for the number and it would take about 5 days. 5 days? What does she have to use, the fucking pony express? And this is a computer place we\'re talking about here! So I called corporate myself and gave them the what how, and then this lady calls me back today yelling at me to stop calling corporate, like I was some type of stalker. I called them once. Argh! I got the number, which kind of irritates me cuz I wanted to report service zone to the IRS. Anyway, the point of this long and boring story is that Nick and I should start an enemies list. So here it is thus far:  ENEMIES: Jill Milton, employee of Service Zone Adolf Hitler, fascist dictator and murderer of millions.  We\'ll add to it as we go on.   My next post will be funnier. I just had to vent. '),(8833,3114279,'12,April,2004','2004-04-12 00:00:00','   Nick : Well,  urlLink Eternal Sunshine of the Spotless Mind  was incredible.  I expected it to be good, but it was even better than I expected.  It was heartbreaking and uplifting.  To me it is the best kind of movie, the kind where I felt a little different after seeing it.  I saw it in my favorite theatre in the world, the Royal Park Stadium.  It was pretty empty last night, and I think I counted 15 people in Eternal Sunshine.  I saw a trailer for a movie called  urlLink Garden State , which looks good.  I killed some time in Borders after the movie, and spent my last bit of money on the new  urlLink Death Cab for Cutie , which is fantastic.  The Borders in Gainesville has the best men\'s room wall grafitti.  Glad to hear you had a good Easter weekend.  I spent the day with family.  We had a big family lunch at my great aunt\'s, who lives in the middle of nowhere, and I mean literally the middle of nowhere, twenty minutes plus away from \"civilization.\"  There\'s the true story of my great uncle who had a heart attack that we always wonder if it would have been completely fatal, except that it took the ambulance so long to get to him because he lives just in the middle of nothing.  So I saw lots of family, some of which I hadn\'t seen since I was younger (\"Our little boys all growns up\"), and ate lots of food, and had some very rural times.  No drinking for me, and no cops breaking anything up.  So now that\'s all over, and I guess the next holiday we have to look forward to is, what, 4th of July?  Fight restroom advertising! '),(8834,3114279,'12,April,2004','2004-04-12 00:00:00','   Glenn : So, here I am, lazy bastard extraordinaire, posting again. The easter weekend was a lazy extravaganza, but if I was religious then I\'d say Jesus spent most of easter weekend being lazy. Really lazy. I mean dead. Anyway, I saw Spartan, David Mamet\'s new film (one of the joys of living in a college town is that I don\'t have to drive 45 minutes to see new movies). It was really damn good. Sucker can write some scripts. Nick, I hope you like eternal sunshine. It is also damn good. Charlie Kaufman can write him some scripts too. I learned a valuable lesson about drinking on Saturday, which I\'ve learned before, and that is don\'t drink too much because it sucks. The party I left (which only turned into a party as all of our friends progressively showed up troughout the night) was eventually busted up by the police, I guess, which is bizarre because none of my friends are really the busted by the police types. Nobody went to jail, which I guess is all you can ask for when it comes down to it.   So, in the spirit of this boring boring post, I am going to tell a story about my dream woman, whom I saw on campus the other day. This supercool black chick dressed as a skater and wearing an \"I Love Nerds\" shirt walked by (too hot in my opinion. I guess that makes me a racist. My bad), talking to her friend about this psycho guy who kept calling her. \"He won\'t leave me alone. He\'s psycho,\" she said. Her friend left, and then this other black guy out of nowhere just comes up and tries hitting on her. He goes \"This guy won\'t leave you alone? Let\'s see what I can do about that. I\'m James,\" (I\'m guessing the name.  I can\'t remember.) She kind of looks at him incredulously and goes \"I\'m Suzie.\" (again, the name is a guess). So she starts walking and he is following her and he just starts talking crap like \"Oh, I\'ll take care of this guy. What do you want me to do to him?\" and she just says \"Stop following me!\" The guy stops, obviously shocked that his smoove-mo-b routine didn\'t pan out and yells \"I ain\'t following you! You\'re psycho!\" So she turns around, super-cool and says \"I am psycho. I\'m not afraid to admit it!\" Then she walked out of my life. Oh man! I\'ve never seen anybody shot down like that. But what I want to know is what the hell this guy was thinking. I mean, did he think to himself \"Oh, I\'ll show this poor timid girl I can kick dudes\' asses. Then she\'ll be mine.\" Give me a break. This isn\'t 1954, hoss. I bet she would have kicked both of their asses in a fight. And to end with a quote:  \"I love you dream woman!\" - Garth in Wayne\'s World '),(8835,3114279,'08,April,2004','2004-04-08 00:00:00','   Glenn : So this is the first post by me, Glenn Shaheen. Man, this is kind of like a job interview or something. I have to be really funny or people are going to say \"That Glenn, I don\'t like his posts. He gots no wit.\" Well. I just hope I don\'t spel anything wrong. Oh, damn! That should be spell anything incorrectly! I\'m so screwed. But I have to give props to Nick for showing me this thing.  Do we get emo points for having a weblog? I hope not. Next thing you know I\'ll be listening to Jimmy Eat World and snobbing my way through Jim Morrison\'s poetry book. I\'ll start pretending I\'ll never get old, and that MTV bands really suck, even though my entire cd collection will be on heavy rotation. I\'m so bitter. I once liked a girl who ended up liking some guy in a band with influences including \"AFI, Our Lady Peace, and Dashboard Confessional,\" although they\'re really original, she swears. I just want to know when the jocks took back rock and roll. They had it in the 80s with hair metal, but the losers took it back briefly in the early 90s for grunge. Now all this noise. I guess I have to live with the fact that I\'ll never be in a band because I don\'t work out three times a day and don\'t look like I should be in a CK ad. But c\'est la vie, au jour d\'hui, at least.  I am happy Easter is here, though. It has the best candy of any holiday. Hooray for dying Jesus! I think I\'d even see the Jesus Chainsaw Massacre again if I got a couple of free cadbury creme eggs out of the deal.  Nick always has quotes on his, so I guess I\'ll dig one out of \"A Far Rockaway of the Heart,\" a book of poems by Lawrence Ferlenghetti. \"So rent a museum     and see yourself in mirrors--  In every room an exposition  of a different phase in your life  with all your figures and faces and pictures of all the people who passed through you and all the scenes you passed through all the landscapes of living and longing and desiring and spending and getting and doing and dying and sighing and laughing and crying.\"  I really had to stop myself from putting the whole thing in. Damn good book. '),(8836,3114279,'04,April,2004','2004-04-04 00:00:00','   Nick : Spaceballs was a blast last night.  The place was packed.  I didn\'t mind that a very cute girl sat next to me and laughed with me and even ended up kind of snuggling up against me.  I\'ll never see her ever again I\'m sure, but those sort of little fleeting moments are wonderful, even if they are only temporary.  Everything is temporary, I guess.  Very tired today, lost an hour last night due to daylight savings. '),(8837,3114279,'26,May,2004','2004-05-26 00:00:00','   Glenn : Yo yo yo. My tyrannical reign as poetry club president continues. I held an officer meeting last night, but used my executive power to have it at a bar where we could watch the stanley cup game. We didn\'t get much done. Nah, we\'re going to try to have a reading in a month or so, so I have to get on that ball. I also have to get on the ball of sending out some more poetriums to places and things. Nick, I know you like Eternal Sunshine, but you can\'t make something your favorite movie until it\'s been out for a year, at least. You need perspective and a further eye. I\'m sorry, but them\'s the rules. In fact, I won\'t even supplant Jackie Brown as my favorite Tarantino movie with Kill Bill until I have a year\'s respite. There are guidelines for these things, you know.  I\'m starting to feel really sorry for one of my roommates boyfriends, because he\'s so nice and naivish and knows nothing about her other boyfriend. Once upon a time I had feelings that could be hurt, so I kind of feel sorry for the kid. There\'s nothing I can do to help that is within my boundaries, though. On the friend scale Laurie definitely ranks higher than this guy whom I don\'t even know that well.  You know, I was going to try to comedy up this post, and then I was going to try to serious up this post, but I think I\'ve decided to nerd it up. I am sick and tired of trying to explain all of the reasons why the original Star Trek is much better than Star Trek: The Next Generation. Now, don\'t get me wrong; the next generation is all well and good, but when it comes down to it it is some goody two-shoes brit flying around putting flowers places, smiling at everybody, some dumb sweet in-joke happens and the strings swell and everybody smiles warmly like it\'s goddamned Jane Austen in space! Data is D\'arcy. I really like it, and I watch it, so people who are about to get offended please don\'t (Jane), but it pales to the original. In the original the Enterprise strutted through the galaxy finding weird shit on fucked up planets. That\'s what\'s so scientific cool about that. There was a cowboy planet, a gangster planet, a roman planet, a nazi planet - you name it! Sure, it didn\'t make any sense, but it was strange as a motherfucker, and that\'s what I liked. I love the way when they went to a barren world there was this kind of wind-chimey sound in the background, or how the special effects were the kind of bad level they ended up being good. Just fucked up shit all the time. It couldn\'t exist in the real universe, but it didn\'t have to and recognized that. That\'s the problem with 80s+ trek. It tries to be too realistic. Maybe that\'s why I like Q episodes so much in the next generation. I don\'t know. Anyway, that\'s my rant.  \"From Hell\'s hot, I stab at thee. For hate\'s sake, I spit my last breath at thee.\" - Khan from Star Trek II, quoting some high-falootin\' literature '),(8838,3114279,'23,May,2004','2004-05-23 00:00:00','   Nick : It seems there have been some new terms being thrown around that I did not previously know about.  These are new terms to me, well, new as in I\'ve only heard them as of this year.  I don\'t live in an urban area, and though I visit an urban area often, and other larger cities, I live in a smaller town, so the hip bullshit takes a little longer to get to me.  That is also due to me not watching MTV or getting on chatrooms, or whatever.  Back in January, after spilling my soul to a friend of mine about my romantic troubles, she says to me \"Maybe you are  urlLink QuirkyAlone !\"  QuirkyAlone I guess is someone too neurotic or insecure or just plain strange to be in a relationship.  Or maybe that\'s just in my case.  I did some research, and I guess it is people who are not into the dating as sport kind of thing, or just getting into a relationship just because that\'s the thing to do, etc. etc.  I\'m on board with that, I suppose.  I\'d never heard of that, and most of my \"big city\" friends hadn\'t heard of that either.  Apparently there is a big movement or QuirkyAlones.  Then a guy friend of mine accused me of being a \" urlLink metrosexual \".  I had never heard this before, and I was offended.  I went and looked up the definition, and I was still offended.  Basically this is a gay guy who is straight.  I have some macho and some typical male in me.  Just because I am more sensitive and sentimental and introspective and neat and emotional and... Okay. I guess I am on board with that too.  Isn\'t this whole metrosexual thing in vogue now?  Don\'t girls like that kind of thing these days?  And aren\'t girls attracted to loners?  So I should have a lot going for me at this point. '),(8839,3114279,'22,May,2004','2004-05-22 00:00:00','   Glenn : Yo yo yo, this is Glenn with a gross-out story for all to hear. So, gather round, kiddies, for the latest Publix adventure. Last night, the same as many nights at Publix, I was bagging groceries, when I notice this woman doing her shopping. She is in her early twenties, and very heavy, but wearing these short short shorts with a slit up the side. I couldn\'t turn away. When there is that much skin it is impossible to avoid it. I don\'t know what it is. Even if she was a guy I couldn\'t have turned away. It\'s just human nature. Anyway, she picks my aisle (of course) and the first four items she slides down are two boxes of condoms and two tubes of astroglide. Then the smell hit me. The smell of hours of gross, slapping, unwashed sex. Ugh! I almost gagged right there. She asked me to help her out, and on the way she says, very flirtily, \"Pretty sad that the only thing I have to do on a Friday night is grocery shop, huh?\" I just got the fuck out of dodge as quickly as I could. And that is the story of how the very last of my libido was destroyed. ');
INSERT INTO `msgs` VALUES (8840,3114279,'21,May,2004','2004-05-21 00:00:00','   Glenn : Yo yo yo. So poetry club went ok. My brief tenure as president will probably be eventless, as we have few people attending over the summer, so my rule of the iron fist will be fairly short. I\'ve got a new movie brewing. There will be no script for this one. Hopefully it won\'t turn into a full fledged disaster, but I guess it\'d be kind of cool if it did. I like watching that Hindenburg footage, after all. After talking with my roommate the other day I came to the conclusion that if I don\'t get some kind of for rizzle dizzle girlfriend soon my neuroses will probably just completely envelope my mind and I\'ll never ever get a girlfriend. Now, this is in the dangerous territory known as settling for whomever one may find, but I don\'t know. I\'m in a rough spot, I think. I guess I really don\'t need a girlfriend, in the long run, since most of the stuff I like doing is solitary in nature, but there will be a time when all of my friends are married, engaged, or otherwise entangled romantically, and then when they have their dinner parties I\'ll be the guest everybody feels awkward around. I don\'t get why society demands we have to hook up with somebody. Why the hell can\'t I just be alone, goddamnit? Well, I\'ll stop ranting.  Nick Dalie, I don\'t think I\'d call Eternal Sunshine my favorite movie, but it\'s definitely a damned good one. I think I prefer adaptation, even. Not to detract from it, and people should go see it, although it\'s gone from the regular theater now. It\'ll probably be in dollarland pretty soon. On the flip flop of things, there are no prostitutes wandering my street anymore. I am so fucking happy. '),(8841,3114279,'20,May,2004','2004-05-20 00:00:00','   Nick : If you have asked me recently or ask me now what my favorite movie is, I have replied and will continue to reply  urlLink Eternal Sunshine of the Spotless Mind .  I knew as soon as the credits rolled at the end that it was my favorite movie; I don\'t think I\'ve ever had quite that strong of a reaction to a film.  But I\'ve noticed that a lot of people do not like this movie.  Glenn, I know you liked it.  But you are actually the only person I personally know who saw the film and liked it.  Everyone else I have talked to either did not like it, or has resisted seeing it because their friends or family didn\'t like it, for whatever reason.  I could be a movie snob and say that they just \"didn\'t understand it,\" or that it was too \"arty\" or whatever. But I don\'t think that is actually it.  But I think it was a fantastic movie that had me thinking about it and feeling it since I\'ve watched it.  So here is my sort of personal \"review\" of the film.  If you haven\'t seen it, I\'ll try not to ruin it for you.  First, I think how you see and feel about this film actually depends partially on your own life experience.  If you have never been in \"love,\" or you are shallow in your relationships, just for sex or whatever, you may not be able to seriously relate.  But the film, for me, mirrored some of my own life experiences.  About six months ago the most serious relationship I had ever been involved in, and the most heartbreaking, ended.  There is a scene in the film where the Jim Carrey character gathered all of the things his ex had given him and even the things that reminded him of her into a plastic trash bag- a few months ago I had done the exact same thing.  The central idea of the movie is the two people go and get all memories of each other erased.  This is obviously sci-fi, and a bit unrealistic, sure, but you can take that as two people trying to forget each other.  God knows I\'ve tried to just forget all of my relationships, especially the last one.  As the film goes on, you start to realize certain things the movie is trying to say:  that maybe it is better to have \"loved and lost,\" that even if a relationship doesn\'t last forever (and in reality, what  does  last forever?), it is better to have the great memories to savor over time as the hurt (hopefully and usually) wears off.  Most importantly, the film is saying that without facing your past mistakes, you are doomed to repeat them.  You could say that even when you do realize your mistakes, you will likely repeat them.  Would you erase the memories of a hurtful relationship?  In the month or two after she and I broke up, I might have liked that idea.  In the movie, after having all memories of each other erased, they get back together again, though you get a feeling that it might just turn sour again eventually.  If you read the  urlLink script , you\'ll see the authors original intended ending, which is even more bleak:  it shows the couple as elderly people, getting themselves erased from each others memories again, having had the procedure done dozens of times over the years.  These are themes not typically done in movies, but it\'s what life is  really  like.  Life is never \"happily ever after.\"  Life is flaws and disagreements and some very dark, hurtful moments.  Even love is like that; shit, sometimes  especially  love is like that.  I obviously saw a lot of the scenes in a personal way because of my own recent experience.  You\'ll relate to this film if you\'ve ever broken up with anybody, or if you look at love as more than just the typical Hollywood and Nicholas Sparks bullshit.  Love is a terrifying thing, and it fucking hurts.  But, as it is in real life, and as the movie shows through all the cynicism, love is still something that, I believe, human beings need.  Romantic love especially.  I have never felt as shattered and desperately heartbroken and lonely in my life as I did those months ago, but six short months later, I\'m still the same hopeless romantic who would still fall in love in a heartbeat.  So there you go.  It\'s amazing that Charlie Kaufman works in the Hollywood system and gets these movies made.  Adaptation was another amazing movie, and Being John Malkovich before that.  Human Nature was pretty cool, too.  Those movies are thinkers.  Eternal Sunshine is not nearly as much of a brain thinker as his previous movies;  it\'s a movie you have to think and feel with your heart more than your brain.  Obviously, it is recommended.  Go see it, if it\'s still playing. '),(8842,3114279,'19,May,2004','2004-05-19 00:00:00','   Glenn : Yo yo yo. So, with this blog thing I guess one of the main things is whom exactly should one tell about it? Cuz maybe I want to talk trash about somebody, but then I\'d forget and tell them, they\'d go and read my blog and then fireworks! Well, I\'m going to talk trash about my roommate. Not really trash, but the other day I was commenting about how cool her bed was, and she said if I was her sex slave for the summer I could use it. She was totally serious. I\'m not talking bad about her though. I would have taken the offer if it meant I could have been the sole user of her bed, but the offer meant I could sleep with her in the bed. Fuck that! It\'s already hot as a motherfucker up in this mug. Besides, being a sex slave would be totally inconvenient. I\'d have to drop whatever I\'d be doing just to please her. If I was out with friends she could call me up and be like \"Glenn. Head. Now.\" I couldn\'t get out of it. After all, it is sex   slave  , not sex   employee  . I don\'t think we\'ll avoid each other sexually all summer long. We\'ve been drinking a lot. A lot. Plus I\'m frayed pretty thin, because I haven\'t had a girlfriend in many a moon. At some point one\'s brain just says \"fuck it. it\'s worth it.\" Hm. Well, Nick Daleeeeeeeeeeeeeey, I hope you\'re feeling better. Me, I\'m tired and sluggish, and I have to go be president of poetry club in a few short minutes. I don\'t even have a prompt. So, gotsta go. Come up with something interesting next time. '),(8843,3114279,'19,May,2004','2004-05-19 00:00:00','   Glenn : Yo yo yo. I am Glenn, an irritated person trying to pay some bills. Not like it\'s working. That\'s what I get for having too many online passwords. Watched Blade Runner again last night. I find it difficult to believe that even now, 22 years after it was made, with all of the so-called advances in cinema technology, that it would look any better. What an awesome fucking movie. I can\'t believe I don\'t own it. So I\'ve decided to have people over to my house this weekend, since I no longer live in a disgusting prostitute infested hellhole. Should be pretty fun, although I don\'t really want the upstairs neighbors to call the cops, since we could be evicted for not alerting the landlord to our sublease. Meh. I figured at this late point in the week at least one interesting thing would have happened to me, but no dice. Just moving and class, and as of 7 o\'clock today work. Yay. My favorite place. Not really. Well, catch you peeps on the flip flop. I have to go and try to start organizing for poetry club.  Glenn '),(8844,3114279,'18,May,2004','2004-05-18 00:00:00','   Nick : I have added a little \"comments\" thingy on the site, just for Jane.  I will do a guestbook later, when I find one I like.  I like the comments thing, though.  I\'m pretty shocked that people are reading this!  I\'ll have to try to be more interesting.  Promise Ring??? '),(8845,3114279,'11,May,2004','2004-05-11 00:00:00','   Glenn : Yo yo yo, this is Glenn, lazoid supreme. I have now entered my mid 20s. 24. The beginning of the prologue of the introduction to the pre-end. I feel kind of old. True, I am only in my early-mid-twenties, but next year I\'ll be in my mid-mid-twenties, and before I know it I\'ll be in my early-late-twenties. Aging is a scientific bitch. Classes have again started. The people in my classes are still whiney retards, but I only have to put up with them in one class this semester, so it shouldn\'t be too bad. The songs I have started for my summer project are beginning to take shape, and I am very happy about that. Moving this weekend also, a definite plus since my roommate caught a prostitute having sex on our porch the other night. Oh, I wish that was a joke. If I post more often, maybe it would be less uninteresting events in my life and more humorous observations. Hm. So the last week of class a bunch of scientific geniuses started circulating emails complaining about our chemistry professor, and how it was his fault they were failing, and blah blah blah. One guy responded by saying the whole system of chemistry was at fault, but students aren\'t powerless, and our voices can change how it is taught, and how the textbooks are written, etc. Here is the email I sent him in return.    Hmm. You raise an interesting point. But then again, perhaps it\'s not the  book or the professor. Perhaps it\'s chemistry itself! We, the students of  FSU, should unite together to change the science of chemistry! After all,  with all of these fancy equations, and numberamas, and thingamabobs, it gets mighty peculiar. Why, if we just changed it so it was much simpler, we\'d have all gotten much better grades! What\'s the point for all them high-falootin\' elements up there on that brainy-brained e-lem-ental tabular? We only need about ten of those. When you gonna see cesium in the real world? Matter o\' fact, when you gonna see electrons in the real world? Never. Electrons ain\'t a gonna git me a brand new porsche, no siree bob. Chemistry should be a one-two buckle my shoe science. Just addition and subtraction. What do you get when you put hydrochloric acid with ammonium hydroxide? Simple. Hydrochloricacidammoniumhydroxide. One and one is two. Ain\'t no blip nor blap bout it. So what say you, fellow students? We gotta git this ball a rollin, otherwise it\'ll be football season and we won\'t have no time \'tall.    Man, I hate my classmates.  In a story of typical Glennluck, I was walking to pay my tuition today when a malfunctioning sprinkler turned on next to me and directly sprayed me with water. I\'m not talking a little bit of sub-stream, but a full on geyser. It was a really hot day out, so I didn\'t care that much. '),(8846,3114279,'30,June,2004','2004-06-30 00:00:00','   Glenn : Yo yo yo, Nick. Good to hear things are going ok in Naples, Italy. I mean, Sofla. I also saw Fahrenheit 9/11 the other night. I really enjoyed it, but it was kind of propagandaish, right? I mean, don\'t get me wrong, I love propaganda if I agree with its message, but he may come across as a bully to those who don\'t agree with him. Anyway, I really hope Bush doesn\'t get reelected, although I can\'t do anything about it. Of course, my general sense of disillusionment about the government as a whole makes me think that nobody has pure ideals when they go into politics. I guess you just have to vote based on which pocket the candidate is in is worse. Should I even be saying this stuff? I don\'t really want to be carted off to Guantanamo Bay or anything.   So it\'s been a little while since I posted, and I am sorry, although I have done a better job keeping up with this thing than I thought. I guess I have one last little ranty list that I want to put together more for my own benefit than the 3 or 4 readers this weblog may have. So this is the top 5 things I want out of life. After I write this out I\'ll start being funny again.   1. A house in the country and an apartment in a major metropolis.  I guess I could take either-or, but both is ideal. I love how when you\'re in the country night is so silent, and the stars just jump out at you. No airplanes or ghetto birds buzzing you at night, no drunk sorority girls singing 80s songs in the street, and no yellowey street lights washing out the sky. Of course, I also kind of like the hustle and bustle of big cities. Not that I\'ve been to too many, just St. Louis, Atlanta, and Chicago, but I really loved the pace of Chicago. I think that\'s a place I\'d like to live if I ever had the money and half a choice.   2. Money  This seems so banal, but it would just make things so much easier. I know you hear a lot of famous people saying \"Oh, money doesn\'t make life easy,\" or fairy tales that teach us \"It isn\'t what you own,\" but yeah, it kind of is what you own, or that you don\'t have to worry about things, like paying for the hospital, or where the next meal is coming from. In fact, this should maybe be number one. That would just make me look like some 1986 greed dealer. Oh well. Anybody who has been in the real world knows a million dollars would make everything a hell of a lot easier. Fuck those lying Hollywood fucks.  3. Time to travel  I was reading the Alphabet of Desire by Barbara Hamby the other day and there are a ton of poems in there that just have to do with jetting off to Europe for a few months or whatever. This is the kind of thing I\'d like to be able to do. But maybe not just this, but also the ability to just get up and say, yeah, I\'m going to take work off for the next week and drive to Nebraska. I guess I want my job to be a side note.  4. Somebody to love  Man, my list is so cliche, but aren\'t cliches sometimes true? I don\'t care how supposedly \"quirky-alone\" I am, or how much some infantile relationship would destroy my entire life right now, at some point I would like a girl who could just stop and look at things with me, take a week off work to drive to Nebraska (not that I have some fixation with Nebraska, but that\'s just the example), or feed me alphabet soup when I run out of words. I cheated on that last one, because I used that line in a poem a long time ago. It\'s still true, though. Also, after attending Megacon I realized the kind of girl for me is the kind of girl who\'d dress up as a comic book character. In fact, the kind of girl for me would buy a bunch of different comics than me every week so we could switch around on wednesday. Man, I\'m such a picky nerd.  5. Fame  Ok, so my list is the lamest list in the history of lists. These are exactly the same five things that everybody probably wants. But let me justify this fame thing. I don\'t necessarily want to have people shouting my name on the street or asking me for my autograph, but I would like respect in my field. That is, I would like people I\'ve never met reading my work and citing me as an influence. When I do a reading somewhere I would like for there to be a good sized audience, and to have sycophants come up to me and say how moved they were by whatever the hell it was that I read. This all sounds so trite, but it\'d be kind of affirming, right? I mean, it would mean I\'m good at what I do, which I guess I shouldn\'t tip toe around saying, because it\'s obvious I\'m leaning toward the writing thing. But that would facilitate all of these, if I was good enough. I had better start really trying.   Gots to see me some Spider-man 2 tomorrow. '),(8847,3114279,'24,June,2004','2004-06-24 00:00:00','   Nick : Wow.  It\'s been awhile since anyone has posted anything here.  Sorry about that.  I\'m here in Naples.  I obviously made it here alright, and I\'ve just been hanging out, getting settled in and becoming more acquainted with the area, waiting to start work next week.  I have managed to go swimming, though the weather has been too lousy to really go out to the beach or anything.  I caught Farenheit 9/11 today at the theatre, which was really excellent.  I bought some new clothes, and have patronized the local used record store.  That\'s about it.    Nothing much new to report.  It\'s still a little strange being here, and not the most comfortable situation as the house I am staying in is still in the process of being furnished and fixed up, and it is also not going to be my permanent residence.  Nothing too interesting is coming to mind to post right this second, but I plan to get back into being online and hopefully posting more often very soon.  And hopefully Glenn will post something soon as well.  Later. '),(8848,3114279,'24,June,2004','2004-06-24 00:00:00','   Nick : Today was my last day of work at the Library. I\'ve felt&nbsp;only a twinge&nbsp;of sentimentality. As much as I\'ve felt unfulfilled with that job for a while now, and as much as certain people have grated on my nerves or whatever, it has been like a second family in some respects for me for the last five years. I feel like I\'ve grown up there, and I kind of have; I was 18 when I started there. I won\'t miss it, though.&nbsp; This job just replaced high school for me.&nbsp; It\'s time to grow up.&nbsp; And actually, when I really think about it, I hate the job now.&nbsp; It was fun for the first couple of years, and it was nice to say \"I work at the library,\" you know, that was good for the girl\'s parents or whatever, but these last two years I\'ve hated my job, and except for maybe four or five people, I won\'t really miss anyone.&nbsp; There, now I look like an asshole, but I don\'t care.&nbsp; This job blows, it\'s boring, no challenge, the people are uptight (well, that\'s to be expected, it is a library).&nbsp; I really can\'t wait to leave.&nbsp; I take it back, I feel no sentimentality.  I found out that I won\'t be starting work on Monday, due to the difficulties with getting things together from such a distance, and also some of Lake City\'s \"slowness.\" So I\'ll have a week in Naples with pretty much nothing to do. I\'m trying to see this as a positive, as I\'ll be able to go swimming, get settled in, get more familiar with the area, go to the movies, etc. etc. I was anxious to start work right away. I\'ll just have to watch my money. '),(8849,3114279,'23,June,2004','2004-06-23 00:00:00','   Glenn : So here\'s my last venting rant for a while. I need to kind of organize things for the state of mind I\'m in right now, so I\'ve made a couple of lists. The first is a ripoff of what Rob does in High Fidelity (great book and great movie. I spent 9/11 watching this with a friend). So, here it is, for my benefit, my top 5 job list.  1. Professional Writer  At first my job list was just a conglomeration of writing jobs, like screenwriter, or poet, or comic book writer, but really all of these would be great. I mean, I\'ll do this no matter what I end up doing, so it\'d be fantastic to get paid for it as well. As of today I\'ve only made about $150 off of writing, and that\'s just contests. I wonder if my moral and ethic code would allow me to write some Tom Clancy crap for the bucks. I mean, these books they have at publix are utter shit. There are literally 15 or so books about jet pilots with names like \"Punk\'s Run\" and \"Blue Skies.\" Somebody gets paid for this? Who wants to read these flattened out rolls of toilet paper? Old bald guys with a beer gut who like to make believe they\'d strangle commies and arabs with old glory if they got the chance? And don\'t get me started on Harlequin romances. I can think of 20 synonyms for glistening cock. I can write a book like that. Urgh. I guess I just need to start sending out non stop.  2. The Creative Force In A Band  I know, I know. I\'m not hip enough. I don\'t buy my entire wardrobe at Hot Topic, and I don\'t roll my eyes at anybody who owns Michael Jackson\'s Thriller (which is a fucking awesome album that I do own). So how could I ever be in a band? It would still be a great gig, to have people want to listen to your music and get paid for it. My problem is organizing the people for the band, though. I can never quite get enough. Also, pretty much every local scene is just rife with nepotism. You don\'t play unless you know somebody. You don\'t come back unless you have an entourage of about 30 mall punk girls who think you\'re the greatest musician since Ludwig Van, and you\'re so poetic when you sing about \"The girl who took my rotten heart/and stepped on it in the dark.\" Agh! Open a book! Use a thesaurus! Don\'t sing about the same goddamned shit everybody\'s been singing about for 200 fucking years! On a side note, Mrs. Keeler, if you\'re reading this I\'d just like to say that the profanity is as much a comedic and rhythmic element as a device for emphasis, so I don\'t really mean to offend. Unless of course, the person reading it is in some stupid indie band who\'s in it for fucking girls and money. Agh! *On a side note I was going to make another entry on my list \"record label owner,\" but I think I\'ll throw it here as a side note. This way I could reject all of these dumb avril lavigne fucks when they send me their demos that daddy paid for. Grrrr.   3. Comic Shop Owner  This\'d be ok. I could read comics all day long and talk to people about comics all day long and be surrounded by comics all day long. I\'m a nerd. I\'d also be a better comic shop owner than the owner of my comic shop, who doesn\'t know blip bout blap bout comics. The cute girl who works there does, though. Not that I want to ask her out or anything. Well, maybe I do, but so does every other nerd who walks into that store. I went to Megacon in march, which is a massive Orlando based comic book convention, and I realized that whenever I do get a serious girlfriend again I would really like it if she was into comics and would dress up with me to go to one of those conventions. That\'s totally hot. And I\'m a nerd. Did I mention that already?  4. College Professor  I almost cracked and put teacher. Ha ha, just kidding. But college professor would be alright. I mean, I guess if I made it as a writer I could do this in conjunction, but it would be great to always be in an academic environment, surrounded by intelligent people all of the time. The real world really doesn\'t have a whole lot of them.  5. Forensic Pathologist  Ok, ok. So this job would be pretty fucking sweet. It\'s just the getting there that irritates me. I hate all of my smug classmates, I hate all of my boring classes, and I hate all of my boring homework. But I blow all of my smug classmates out of the water on tests, I ace all of my boring classes, and I don\'t really have homework, I just had to stick a number three in there for the rhythmic triad rule. I also really love when I tell somebody I\'m going to be a doctor and they get this brightened expression on their face and say \"Oh really? Family practice?\" and I say \"No. Forensic Pathologist. I\'ll be cutting up dead people.\" Ha ha ha ha! You don\'t want me marrying your daughter anymore! Oh, old publix people. Hilarious.    So that helps me a bit. Actually, I have a little bit more ranting to do, but this is a long post, so I\'ll hold off for a couple of days. Until then, l8r sk8rs. '),(8850,3114279,'22,June,2004','2004-06-22 00:00:00','   Nick : Glenn, nice list.  I feel somewhat justified now for having Eternal Sunshine on my list since you put Kill Bill on yours.  Though I did not see it five times at the theatre.  I don\'t have that kind of money to throw down.  I don\'t feel the way you do about Lake City.  Of course I haven\'t really lived anywhere else.  I\'ll be very glad to get out of here, and I mean very glad.   Looking forward to your venting. '),(8851,3114279,'21,June,2004','2004-06-21 00:00:00','   Glenn : So Lake City for five days was ok. Every time I end up going there I try to reevaluate my life and what I\'m doing. Would I like to live in a small town? I guess, if I could write, and obtain some kind of notoriety. Not like the fame is the most important part of it all, but at least recognition would justify it a bit. I still don\'t really know what I want out of life. Every now and then it\'ll be like something clicks, for just a second, and everything almost seems clear and I can almost cling to some kind of direction, but that fades and I\'m left where I started. It\'s usually landscapes or sunsets that trigger this. I think it reminds me of home, like Nova Scotia home. Maybe some kind of pilgrimage would help clarify things a bit, or maybe I just need to sit down and think what I want out of life. Not my parents, not my friends, but me. Hm. I had a nice and funny long diatribe about umbrella holding, but I\'ll wait on that. Sometimes you just need to vent, regardless of whether or not it would make for an interesting post. So, you want a classic top ten list? But what about the seven out of my top ten movies which were already earlier than 1980? Well, that ain\'t gonna fly. So I\'ll tell you what. I\'ll make a top 10 movie list of movies that I watch the most often, regardless of how much I appreciate them as cinematic. So here it is, my \"Nick Dalyesque Top Ten Movie List\"  1. 2001: A Space Odyssey  Sorry, but I watched this movie every night for like an entire year. It still makes number one.  2. Dawn Of The Dead  And this one is also a repeat. My bad.  3. Revenge of the Nerds  I just put this one at number three to get all the repeats out of the way. No matter how cheesy this movie is I still almost choke up at the end.  4. Star Trek II: The Wrath Of Khan  Yeah yeah, this is everybody\'s favorite, but it\'s still damn good. I don\'t even think Shatner\'s acting is that bad, to tell you the truth. And it\'s not just a space shoot-em-up; the getting old theme really plays through it quite well, and even now I can start to identify. James Horner\'s great score is also a plus.  5. The Big Lebowski  There are few movies I can quote to the same degree as this movie. I haven\'t seen it in a little while because somebody has my copy, but if I wanted I could probably just act out the whole damn movie anyway. I have a feeling this top ten list is going to be mostly comedies. Oh well.  6. Wayne\'s World  This is one of said few movies I can quote as much as Big Lebowski. Yeah, it\'s stupid and immature and adolescent and aimed at a target audience of 13 year old boys, but it\'s still goddamned funny.  7. American Grafitti  If I was in some other mood this wouldn\'t make the list, although it is still great movie. I get these weird kind of nostalgic chills watching this movie. Not 60s nostalgic, because I wasn\'t even alive then, but nostalgic for some era in my own life I can never go back to. Almost like I missed out on something. This is the kind of mood I\'ve been in. I\'m being pushed in a direction by everybody else I don\'t want to go in, and yet I have no idea what direction would be better. Watching this movie makes me think about how great a filmmaker George Lucas could have been if he didn\'t get wrapped up in almost thirty years of star wars.  8. Back to the Future  I saw part III first, but this is the best of the trilogy. I still love all of the repeat stuff they did in part two. You really don\'t find filmmaking this joyful that often anymore. Ahh, the 1980s.  9. Kill Bill (both parts)  Ok, here\'s my recent movie. I saw each part five fucking times in the movie theater, so I figured it justified it being on this list. Both parts never got old, either. Great movie.  10. Army of Darkness  Heh heh heh, what a ridiculous concept. I love that this all spawned from a low budget practically student film. Yeah, it\'s not the best made movie out there, but it certainly is enjoyable.  So there yous have it, another ten movies I like. On a lighter note, Jane convinced me to wear a coral (pinkish) shirt to the wedding on saturday, which was cool. It was much less boring than my grey shirt with grey suit. Anyway, I stopped in lake city to fill up on gas on the way home, when this truck filled with locals starts whistling and saying \"hey, look at that fag.\" Ahh, lake city. Where does the time go? '),(8852,3114279,'20,June,2004','2004-06-20 00:00:00','   Nick : Lots of things have been said about my top 10 favorite movie list, as far as almost all of my movies being more recent.  Apparently you\'re supposed to have some rules to your favorite movies and whatnot, and I\'m supposed to include \"classics\" or something.  Those 10 movies just happen to be the movies I identify with for one reason or another, and movies that have made me feel something and have affected me, and I am into movies like that more than movies that are just entertaining or cinematically interesting.  I tend to think with my heart with movies more.  But I do like classic movies, so here is my list of my top 10 \"classic\" movies, classic being pre-1980.  I\'m sure no one will like this list either.  Nick\'s Top 10 \"Classic\" Movie List:  1.   urlLink To Kill A Mockingbird    I adore this movie.  I still get chills when Scout sees Boo Radley behind the door at the end.  When I was in high school my girlfriend at the time made me watch this, and I\'ve loved it ever since.  Not the most profound movie, or the artiest movie, and not really particularly relevant these days, but still a beautiful movie.  2.   urlLink A Boy Named Charlie Brown    I was asked a few years ago which character from Peanuts I would be, and I didn\'t hesitate to answer \"Charlie Brown.\"  These days I\'d still say I identify with Charlie Brown, but I think I lean more towards Linus Van Pelt these days.  I must have seen this movie fifty times since I was little.  I\'m a huge fan of Vince Guaraldi\'s music.  I have a real soft spot for Peanuts, and a lot of the feelings that come across;  it\'s so bittersweet, and there\'s the unrequited love, and as I watch classic Peanuts cartoons, and especially this movies, I find myself sighing throughout the entire thing; it\'s just wonderful, and I *adore* this one.  I identify and feel for this cartoon more than most live action films.   3.   urlLink No Time For Seargents    Fucking hilarious.  A pre-Andy Griffith show Andy Griffith (and Don Knotts as well) plays a complete idiot who joins the army.  I laugh harder at this movie than most.  If I did a top ten favorite comedies this one would definitely be there.  Another movie a lot of people have not heard of.  4.   urlLink Silver Streak    I\'m not totally sure, but I think this is the first movie with both Gene Wilder and Richard Pryor.  Very very funny, and also a bit of a thriller.  Gene Wilder falls in love with a girl on the Silver Streak train, and turns out people are out to kill the girl, so he tries to save her.  Richard Pryor plays a theif who tries to help Gene Wilder.  Not as silly at all like some of their later collaborations, but really hilarious, with a bit of action.  5.   urlLink Dog Day Afternoon   I\'m surprised at how few people I know who have actually seen this film.  Al Pacino and John Cazale (Michael Corleone and Fredo from the Godfather) hold up a bank, and start a media frenzy.  Based on a true story.  You really actually start to root for the \"bad guys\" in this movie, and feel really sorry for their pitiful situation.  Funny in parts, sad in others.   6.   urlLink That Obscure Object of Desire    Luis Bunuel\'s last film.  This one is hard to explain.  Fernando Rey (who was in The French Connection) plays a guy who lusts after a girl named Conchita.  Conchita is played by two very different actresses in the film, apparently to visualize her shifts in mood and personality.  Conchita teases him and leads him on, but he keeps coming back.  Very surreal.  Pretty good \"surprise\" ending.  7.   urlLink Sleeper    I\'m not totally a fan of Woody Allen; I like a lot of his films, and a lot of them I can live without.  But this film is one of Allen\'s earlier films, which seem to be funnier and sillier than his more recent films.  Very funny, and I have always loved the retro-future stuff.  My dad showed me this movie when I was little, and I did not get it then.   8.   urlLink One Flew Over the Cucko\'s Nest    This movie probably needs to explanation.  Just a great movie, and very affecting.   9.   urlLink The Black Hole    All sci-fi is cheesy.  (Star Wars is cheesy (I\'ve never liked Star Wars much).  Even Star Trek is cheesy.  2001 is kind of cheesy, but much less so than most sci-fi.) But I love The Black Hole.  And it is cheesy.  But it\'s still cool.  When I was little, I loved this movie- the bad robots were scary.    10.   urlLink The Odd Couple    This movie was on my other list.  I don\'t care.  This is a great movie.  I love Walter Matthau and Jack Lemmon together.  I love the Grumpy Old Men movies.  I love The Odd Couple II (though it\'s more Grumpy Old Men than classic Odd Couple).  I have never gotten tired of this movie.  When I can\'t think of a movie to watch, I go for this movie.    Top 10 lists are difficult.  There are lots of classic movies I love: Taxi Driver, The Shining, Monty Python and the Holy Grail, Bad News Bears, Arsenic and Old Lace, The Taking of Pelham One Two Three, Alice in Wonderland (the Disney version), The Out-of-Towners (the original 1970 version), 20,000 Leagues Under the Sea, Casablanca, Vertigo, Kiss Me Deadly, The Big Clock, etc. etc. etc.  I look forward to seeing Glenn\'s list; though he had some classics on his last list.  There are a few entries I would probably change on my first list; I\'d most definitely take In the Company of Men off, and as I realize after doing this list, I\'d add A Boy Named Charlie Brown.  Actually, I\'d be tempted to stick that even in my top 5.  Top 10 lists are hard, and really only the top 5 are pretty solid, the rest really fluctuates.  I do prefer more recent movies to older ones; as stupid as it sounds, I tend to love movies that I feel the most for and that I relate to, ones that go a little beyond entertainment, or something.    That\'s all for now.  More things to pack up into boxes and bags, more things to throw out.  This time next week I will be moved and living somewhere else. '),(8853,3114279,'18,June,2004','2004-06-18 00:00:00','   Nick : Yo yo yo, bowling motherfuckers.  Our scores all stayed pretty consistently low.  The highest score of the night was Jeremy\'s 117.  I stayed pretty consistent, I probably averaged about 90. (I only broke 100 once.)  We were all pretty bad.  It was good to have Glenn, Ryan, Jeremy, and Matt all together.  A nice little going-away thing.  And we got to see someone carried away on a stretcher.  Good times. '),(8854,3114279,'17,June,2004','2004-06-17 00:00:00','   Nick : I finally got something done. I got quite a few things packed and a lot of stuff thrown out.  I also found my  urlLink CQ  DVD, which I thought was long gone, in my trunk.  Now I\'m relaxing and doing some around the house type stuff before tonight\'s bowling.  Went out to lunch with Ryan and Glenn, and ran into an ex, which I was afraid of, but it turned out to be not strange, as I thought it might.  One more week... '),(8855,3114279,'15,June,2004','2004-06-15 00:00:00','   Nick : (Nick is left, Glenn is right) '),(8856,3114279,'11,June,2004','2004-06-11 00:00:00','   Glenn : Yo yo yo, you whippersnapper motherfuckers. 60 year old man? Whatever. You\'ll hit the big 24 soon, and then you\'ll feel just like I do. Congratulations on quitting and everything. I\'ve only ever quit from one job, service zone, and I didn\'t really do it in any interesting manner. I wasn\'t going to post today, since I have two million things to do before I go to work, but a comic book webpage I was trying to read wouldn\'t work. But now it is working. And I don\'t have anything interesting to say. Oh, speaking of that gay bar I went to the other night, there were actually a couple of straight girls there (or really goddamned convincing drag queens, but hopefully the former) who were pretty nice and intelligent and flirty. Of course, me and my neuroses didn\'t actually click with this until I was long gone from the bar, so it didn\'t matter. They told me if I was a character from friends (which I hate and don\'t watch and hate) then I would probably be Joey. He\'s the stupid one, I guess, which I immediately mentioned, so they changed me to Ross, the whiney one. Urgh. I guess I should be happy with any of them, since they can all apparently hook up with Jennifer Anniston. Watch out, Brad Pitt. I\'m a-comin for ya. This is the lamest post I\'ve ever done. '),(8857,3114279,'10,June,2004','2004-06-10 00:00:00','   Nick : Well, though I\'d like to use some of your job-quitting tactics, I\'ll just be typing up a nice, semi-heartfelt letter of resignation.  Two weeks from this Saturday, I\'ll be packing up and moving to Naples.  It looks like I\'ll be staying with my family in their new house for a month or so until I get settled in, get some money saved, and find my own place.  I\'ll be working at the  urlLink David Lawrence Center , as a Behavior Technician.  I\'m not completely sure if I\'ll be attending school in the fall, or wait until the spring when I\'ll be more settled in and secure.  I got real lucky, this was the second job I applied for, and the first that I interviewed for, and it is the one I really wanted.  I\'m waiting to feel all sentimental about leaving Lake City, but it doesn\'t seem to be happening.  This is what I\'ve been waiting for.  It looks like I\'ll be having a couple of \"parties,\" so Glenn, let me know again when you are planning to be in town, so we can all get together.  Maybe go bowling, eat dinner.  Your treat, right? :p    Went to the porch to have a thought, got to the door and again I couldn\'t stop. '),(8858,3114279,'09,June,2004','2004-06-09 00:00:00','   Nick : Well, Glenn\'s list is pretty impressive, sorta hip (lowercase h : not as hip as I figured), even if it does make him seem like a 60 year old.  Most of them I can agree with, though I hate Singin\' in the Rain, and I could barely get through 2001; I could feel myself aging as I tried to sit still and watch the movie (read: boring).  I usually personally rate movies based on how they make me feel, and how I relate to them;  I guess I\'m not as interested in cinema as I am feeling and thinking.  Lawrence of Arabia is completely magnificent to watch, but there is not a thing that I can relate to in that film;  Roder Dodger looks like it was just shot by a couple of guys with video cameras following people around, and it probably was, but it rings true with me.  I would also say that my favorite movies are ones I can watch over and over and over; rewatchability.  Between the two of us, we have some taste.  I\'m surprised Once Upon a Time in the West and Wrath of Kahn were not in the top ten.  Also surprised that Howard the Duck or Battlefield: Earth didn\'t make the cut. '),(8859,3114279,'08,June,2004','2004-06-08 00:00:00','   Glenn : Hmm. My top ten movie list of all time. Hard to say, really. I usually just leave it at three, but even those have been fluctuating lately. So, let me try and top Nick and all of his 90s movies with a bit of variety. Of course, maybe I\'m lying to appear more in the so-called know. Anyway, here it is from 1 to 10, Glenn\'s top 10 movies of all time.  1. 2001: A Space Oddysey  That\'s cheating for Nick to make a \"guess\" that this would be on the list, since he knows for a fact it\'s my goddamned favorite movie. Some people may think it\'s boring (Jane), but I think it\'s just a complete exercise in film craft. Hypnotic, and for me it\'s impossible to turn away from. And I don\'t even know if at this point in time any films have really beat it in the special effects department.  2. The Good, The Bad, And The Ugly  Fucking brilliant. This movie destroys the entire concept of westerns. Eli Wallach is my favorite as the Ugly. It\'s obvious he had the most fun shooting of all of the main characters. Movies don\'t get much prettier than this.  3. Dawn of the Dead  I love the color in this movie. It just all seems so surreal. None of it is really in shadows, every bit of horror and satire is right up front. True, some of the social commentary is a bit blaring, but I never, ever get bored watching this movie. Kids these days would probably complain because there are a lot of slow parts, and not enough nu metal, but those are the best parts. Like when you don\'t see zombies for 20 minutes of the movie and you kind of forget what kind of terrible situation they\'re in. Fantastic.  4. Lawrence of Arabia  Ok, so here\'s where my list starts to get a bit fuddled. I know for a fact the first three movies are my favorite three movies, but these next seven may be awesome movies which hover around my top twenty of all time. It\'s so hard to pick, I think. Nevertheless, Lawrence of Arabia is an awesome movie. Once I almost had an opportunity to see this in the theater, but I had to do some other boring shit. Man, I really hope I get the chance to redeem myself for that one. You might say, how could one make the desert look beautiful? Well, watch this movie and find out.  5. Singin\' in the Rain  Uh oh. I may have just thrown this one in for variety. This is, however, one of the funniest fucking movies you will ever see in your life. I don\'t give a good god damn how many times my roommate claims Fred Astaire is better at dancing, Gene Kelley is a better actor, director, and knows how to make a dance scene work. I\'ve always said I\'d trade my heterosexuality to be able to dance like this cat. I\'ll stick to that still, but maybe only because deep down inside I know there\'s no way I could ever dance like him. Oh well.  6. Jackie Brown  See? See? I have a semi-recent movie up here too. Few other movies make me know characters as well as this movie here. Everything just comes to life for me, and I find the older I get the more solidified this is as Tarantino\'s best flick. Now, maybe I waver on this whenever the full Kill Bill comes out, but for now this stands.   7. Apocalypse Now  Hm. This is probably higher up than seven on my list, I just didn\'t think about it first. Like I said, these all shake around. I got the opportunity to see this in the movie theater when it was rereleased, and even though I thought every additional scene sucked scientific balls, this was the most incredible looking movie I have ever seen in my life. The bridge scene in the theater was worth the price of admission alone.  8. Duck Soup  There have never been a more coherent and talented group of comedians as the Marx brothers, and this movie is solid proof. I can watch this movie three times in a row and it never gets less funny. Taxes? I have an uncle who lives in taxes. No, taxes as in dollars. Yeah, that\'s where he lives. Dollars, taxes.  9. Raging Bull  I couldn\'t not have Scorsese up here. Then again, I only have one space left, and I have no Spielberg, no Hitchcock, no Lee, no Scott... aaaagh! But this is easily Scorsese\'s best picture. Less flash than Taxi Driver, but way more know-how. De Niro is flawless in this movie as well. Very depressing. He should have won best director, but instead he was doomed to edit crap for the oscars for all eternity.  10. Revenge of the Nerds  I had to have one \"what the fuck\" entry up here. But seriously, there is no better movie that exemplifies all of the eighties \"geeks fight back\" extravaganzas. I wonder, though, if the main girl was supposed to look good naked, because to me she looks more like a fucking skeleton. Eat something, girl.  Oh, man. I missed a bunch. Let me just throw a few other masterpieces out, just to round out my top twenty.  11. Raiders of the Lost Ark 12. Vertigo 13. Blade Runner 14. Once Upon A Time in the West 15. Malcolm X 16. The Godfather 17. Touch of Evil 18. Star Trek II: The Wrath of Khan 19. Dr. Strangelove 20. American Grafitti  Even that isn\'t enough. So many movies, so little time. On the lighter side of things, I read a poem at a local poetry night last night that one of my friends hosts. It\'s held in a gay bar, which was weird at first because whenever I got there this creepy old guy started making eyes at me. Now, I\'m not gay, but if I was gay I think I\'d be out of that guys league. I mean, come on. I could do much better than that wrinkled old fool. Also, the poem I read has the word \"gaytarded\" in it (my own concoction), which I don\'t think went over too well with the crowd. Oh well. Exposure is exposure. '),(8860,3114279,'08,June,2004','2004-06-08 00:00:00','   Nick : I\'m in a list-making mood, and there has been a lot of conversations I\'ve been in about favorite movies, usually in the form of a top ten.  I have compiled my list.  I\'m sure Glenn\'s list would be more Hip (capital H), and I\'m sure I\'ve broken some of the \"rules.\"  And I\'m sure I don\'t have enough \"older\" or \"classic\" movies.  So here goes.   Nick\'s Top Ten Movies    1.  urlLink Eternal Sunshine of the Spotless Mind    (I don\'t care if this movie is \"new,\" I have never been as affected by any other movie in my life.  I was completely taken with this movie, and as stupid as it sounds, I have thought about something to do with this movie quite often since I saw it.  It\'s very possible that it\'s because of my own very recent personal life; besides the sci-fi memory erasing plot, the movie very closely parallels my own most recent and most serious romantic break-up.)     2.  urlLink Punch-Drunk Love    (Although I probably shouldn\'t admit this, I relate quite a bit to the Adam Sandler character in this movie.  I\'ve managed to usually act pretty normal and even sometimes very confident and outgoing, but on the inside I\'ve always been extremely awkward and insecure, and I\'m usually one inch away from completely stumbling over myself.  Not really a comedy to me; more of a story about a very lonely, sensitive, and socially awkward guy who finds a girl who loves him not only despite his quirks, but maybe even because of them, and this love changes him, makes him stronger and able to finally stand up for himself.)   3.  urlLink Swingers     4.  urlLink Donnie Darko     5.  urlLink Roger Dodger     6.  urlLink The Odd Couple    (I don\'t know how many times I\'ve seen this movie.  Doesn\'t really require any explanation.  I have more Oscar in me these days than I used to, but I\'m probably leaning more towards Felix.)     7.  urlLink Magnolia    (I couldn\'t believe this movie when I first saw it.  Even though the movie is more than three hours long and is mostly all talk, I couldn\'t move for those three hours.  I love the characters in this movie.  I love the ideas of everything being connected, I love the ideas of retribution (except for the Philip Baker Hall character), I love the divine intervention (I knew nothing of this film or the \"frog\" scene before I saw it, so I was completely blown away).  After the movie was over, I felt a little bit different, like I\'d been through something.  That\'s about the best thing a movie, and any art for that matter, could do.)   8.  urlLink Free Enterprise    When I first let Glenn borrow this movie a few years ago, after he saw it he said \"only people like us would like this movie.\"  What he meant by \"people like us,\" I\'m guessing, is \"nerds.\"  I have been a Star Trek fan since I was little (original series, motherfucker!), but as I\'ve gotten older I\'m not as into it.  But I do have a shameful amount of knowledge in my brain about the Star Trek universe, and other sci-fi and cult films.  I\'ve had the very same conversations and arguments in real life that are in this film, and I\'ve been in the same romantic situations (though mine never turned out very happily ever after in real life).  And William Shatner is hilarious.  This is basically Swingers for geeks.   9.  urlLink In the Company of Men     10.  urlLink Northfork    There you have it.  Most of these are \"newer\" movies, as in the past decade.  I can only guess what films are on Glenn\'s list.  2001?  Once Upon a Time in the West?  Howard the Duck?  Cool Runnings? '),(8861,3114279,'07,June,2004','2004-06-07 00:00:00','   Glenn : Yo yo yo, this is Glenn. Tampa bay sucks. Calgary rules. See, I\'m no fair-weather fan. Speaking of which, it is raining, and on the way from my chem lab to the computer lab I decided that instead of walking about 10 feet on pavement I would cut across some mud, and just be really careful. Of course I slipped, but only enough to cover about six inches of the bottom of my jeans in mud. Drat. I need to come up with some more poetry club fliers, but I\'m blanking right now, plus it\'s still raining. Once again I have managed to start a post when I don\'t actually have anything to post about. This friend of mine runs local poetry night here in Tallahassee every Tuesday at Brother\'s a gay bar. It\'s not gay poetry night or anything, just anybody who can read. After running the gamut of every other one of my writing friends he has finally managed to ask me to go, although deep down inside I don\'t want to go. The main reason is that it starts at 11 o\'clock, and the crowd is supposed to be harsh, especially on those who aren\'t gay. Meh, any exposure is good exposure, I guess.  Glenn '),(8862,3114279,'07,June,2004','2004-06-07 00:00:00','   Nick : Horray for Tampa Bay!! :p '),(8863,3114279,'07,June,2004','2004-06-07 00:00:00','   Glenn : Yo yo yo, this is Glenn, posting without any particular reason yet again, to probably make for a boringtastic blog entry. Anyway, welcs to John Titor, prophet to the new century and all around great guy. Isn\'t he great? He\'s great. I saw the new Harry Potter movie, which I enjoyed, although my roommate said it was terrible because they strayed from the book so much. Whatever. I guess I win out because I didn\'t read any of those books. Glad to see you made it back to the old Lake safe and sound, Nick. Two days of shooting on my new movie have been thwarted by weather. I hope this isn\'t a recurring thing. I can probably predict that I will be horribly injured by the end of shooting though, as a result of my thinking that I can wrestle, kung-fu fight, and bo staff fight. I can\'t do any of those things. We beat Megaman 1 last night, the hardest of all megaman games, and least fun. I don\'t think I\'ll ever even start it up again, now. What a fucking bitch of a game. I guess I should go and get ready for the stanley cup game, now. My team will probably lose anyway. '),(8864,3114279,'05,June,2004','2004-06-05 00:00:00','   Nick : Yo yo yo, this is Nick.  750+ miles later.  I still have a bit of car lag.  My trip was great, and I think a success.  And I came back home with a new CD (Air \"Moon Safari\"), a few cat scratches, and an Eternal Sunshine poster. '),(8865,3114279,'04,June,2004','2004-06-04 00:00:00','   Glenn : Yo yo yo, this is Glenn. I\'m posting when I have nothing to say, so expect boredom. Last time I was in the computer lab I got scientifically drenched, so here\'s to hoping it won\'t happen again. Man, today it was blunder of blunders when my cellphone went off during class. I am usually good about turning that thing off, but today I fucked up. Drat. Apparently giant fucking spiders are invading our house. You know, although I don\'t like spiders on me or appearing suddenly when I don\'t expect them, I kind of like them. I definitely won\'t kill them, because of spider karma. I mean, anything that could possibly eat a roach for me isn\'t all bad. Which is why I like my cat Xerxes. Anyway, I\'m boring today. '),(8866,3114279,'01,June,2004','2004-06-01 00:00:00','   Glenn : Yo yo yo, this is Glenn, Duke of New York A # 1. Three smiling cents for the person who checks that reference. It has been a boring couple of days. I went to Lake City on Saturday, to converse with my editor on my old movie project and get some equipment for my new movie project. So much wasting of the time. I also picked up a dvd collection of world war II Walt Disney cartoons. Very interesting stuff. The gem in said collection is definitely Der Fuhrer\'s Face, otherwise known as \"Donald Duck in Naziland.\" He wakes up in Nazi germany and has to work in a munitions factory for Hitler. Of course it\'s miserable and hilarious, but then he wakes up and realizes it was all a dream. He hugs his little statue of liberty and says \"I\'m so thankful to be a citizen of the United States of America!\" Ahh, nazis. Comic gold! On Sunday, work. Not so good. I don\'t understand how people can just keep assuming I\'m so stupid they have to tell me \"Don\'t put my bread on the bottom.\" Like what, I\'m from some kind of alternate baggers universe where bread can\'t be crushed by conventional means? Man, I must quit my job a thousand times in my head every time I work. The trick is to have it just right, the quitting. Let me break down the science of it.   1 - You have to do it in a way that is noticeable, but not completely over the top.  You can\'t just slink into the back, tell the manager you\'re gone, and then wave good-    bye to everybody on your way out. That\'s boring. On the other hand, you can\'t just yell \"FUCK EVERYTHING, MOTHERFUCKERS!\" at the top of your lungs and jump through a plate glass window. Firstly, that injures you, and secondly, it\'s very bohemian. You raise your voice to very audible conversation level, and say something like \"Check the lot? Check the lot? No, I think instead I\'ll be checking out, bitch.\" It plays on words, while it is still insulting to some aspect of management. Which brings me to...   2 - Try to insult somebody who deserves it on your way out.  Keep the deserve it in mind, though. If some old lady says \"Could you please double bag that one?\" you shouldn\'t respond with \"I\'ll double bag your coffin when you die in a few weeks, you old hag!\" She has lived a while, and any bitterness is probably caused by brain leakage or wrinkle fungus or some other gross old person disease. Instead wait for the forty year old white fuck who says \"Wait! Don\'t put those cans on the bottom, goddamnit! Hurry up! Son of a bitch, why did I get the retard bagger.\" Then it\'s fair game to pick any of the following possible responses; a) \"Your wife doesn\'t mind when I line the bottom of your bed with cans, you dumb fuck,\" b) \"Sir, you do realize we carry this tube of asshole cream in extra large, do you not?\" or c) \"Fuck you, you degenerate half-sunken, monkey-flaying, shitsmoke-dragging, cerebellum-gaping fuckasaurus.\" That last one can basically be any chain of slightly creative profanities. Make sure that this is not a person who could kick your ass before you say any of these choices. If they can, go with the more simple \"Bag the shit your motherfuckin self; I quit,\" or just \"Fuck this, I\'m outta here.\" The colorful expletives are important because that helps you secure your non-returning to the job. Otherwise you may inadvertantly leave your door open and be back as soon as you need some cash. This is how prostitutes get trapped.   3 - Check the time of your exit.  Don\'t quit when you only have one minute left. That\'s retarded. Just finish that shift out and go out with a bang the next time you work. Also, don\'t quit just after you\'ve done some horrible and degrading piece of work, like cleaning fucking shit off the walls of the bathroom. Do it before you do that, so you don\'t have to do it. Duh.   4 - Be cool.  Don\'t just say I quit, and don\'t just get yourself fired. Have some ready responses to your boss\' possible prompts. Example:                  Boss - \"Glenn, some dumb brat kid puked all over aisle 8. Clean it up.\"                 Glenn - \"Fuck no, bitch.\"                Boss - \"Excuse me? What did you say?\"                Glenn - \"I said get a hearing aid. Clean that nasty shit up yourself, because I ain\'t the bitch of Publix central anymore. I\'m out.\"   Now, that example does not appear to have a whole lot of wit buried within, but it gets the job done. It\'s funny and unflinching, and it pushes the boundary to getting fired level, but doesn\'t let the boss get the firing in before the quitting enters. Simple, but somewhat perfect.   5 - Have no regrets.  Your job sucks. No bones about it. Almost everybody\'s job sucks, except for my friend Todd who gets to sit on his ass and drink wine all day. That job rules. Also, comic book writer is a good job, and so is fortune heir. But those aren\'t your jobs and your job sucks. So whenever you reach your boiling point and initiate the quitting sequence, don\'t turn back. Don\'t apologize and don\'t look back to your place as anything other than some massive building made of dog shit filled with pieces of dog shit with dog shit for brains. Quit and smile. You\'ll get another job. In fact, whenever you pick up your last paycheck walk in with a big smile on your face so all those poor dumb fucks who still shop or work at that hellhole know you had the brains to pull off the escape. And knock something over so somebody has to clean it up. That\'s also cool.  So, faithful readers, bear these thoughts in mind. And just change all of the grocery crap to whatever your job has. Anyway, I\'m out for now. Peace to the east. '),(8867,3114279,'29,July,2004','2004-07-29 00:00:00','   Glenn : Yo yo yo, this is a banner fucking day in the weblog, when I post twice in a row with out Nick to buffer things out with a little bit of sense. Not that I have anything entirely interesting to say, but I\'ll post some rambledambles. I saw the Bourne Supremacy last night. Apparently the first movie didn\'t stick with me that much, because I really didn\'t care about any of the characters in this one, as they didn\'t build any of them back up. Really boring. Plus there was the preview for a new great movie coming out called \"Paparazzi.\" The plot is this actor gets really rich and famous, and then the paparazzi start hanging out all the time and taking his picture. Well you know what? He has had enough, and decides to take matters into his own hands. I say it\'s about time the actors struck back. How long have we, the public, been giving them so much shit and millions of dollars, and they have just been silent? Yeah, we deserve it. Not that I agree with paparazzi, but give me a break. It\'s like the actors all just decided to get together and hatch a way to get the public on their side. Well, I\'m not going to feel sorry for these whiny millionaires. This is a movie? I\'ve seen better film on teeth, to quote Get Shorty. It\'s produced by Mel \"Bloody Jesus\" Gibson. Nice follow up, jackass. Michael Moore was on O\'Reilly the other night. It was hilarious. Like a couple of brat kids just screaming their sound bites at each other.   I was with a friend who got pulled over yesterday by campus cops. What a bunch of cowboy assaholics. First off he took 15 minutes running my friends license, since I guess he got stuck with the old car with the telegraph machine inside, and then he told my friend the reason he was pulled over was because he had a brake light out, and gave him an $80 ticket. When we checked later it was one of the little side lights. Agh! Why won\'t cops ever  help  me out? They always just cause problems. Anyway, enough negativity for one night. Peace to the East. '),(8868,3114279,'27,July,2004','2004-07-27 00:00:00','   Glenn : Yo yo yo. I am so far behind on this crap. I don\'t have the internet at all anymore. That really throws a kink into everything. I have no idea what I\'d do if I couldn\'t just pop over to FSU and use a computer lab. If things go my way and I\'m not in class in the fall I won\'t be able to do even that. Shudder. I\'ve finally moved into my new, clean, maggot-less, carpeted place. It\'s so nice. I actually cooked myself dinner the other day rather than just stare disgusted at a mess of a kitchen. It\'s really a design flaw to have the main door lead into the kitchen anyway, I think. Mud gets in there, and flies spontaneously generate, and bees and roaches-- gross.&nbsp; Now my kitchen is off to the side. No disgustingness. I\'m so happy.   I really used to like Dennis Miller, but recently he just seems to be some kind of right-wing voxbox. He just yells on his show how it\'s ridiculous to think that if America is the strongest country in the world we still need to listen to what the rest of the world thinks. Well, I say you can\'t bully people, or eventually they\'re just going to hit back. I mean, the concept may seem ridiculous, but there could be a EU USA conflict at some point in the future. I also don\'t like the uberpatriotism that really took hold after 9-11. The kind where liberals have to qualify every thing they say that may be against the administration. This so-called \"United we stand\" mentality is extreme to the point of massive excess. If you have a different opinion that Bush, you are against America and may as well sign up with the terrorists. The right-wing has really forced the left into a corner in which they can only meekly make suggestions while prefacing everything with \"Now, I love America,\" or \"America is the greatest country in the world,\" like somehow what they say negates that fact. It\'s really shrewd politicking. &nbsp;I\'m not American, but you know what? I love America, and it is the greatest country in the world, but Bush and crew have done&nbsp;too many&nbsp;messed up things and it\'s time we mended what wounds we have with other nations and lead by example, not force. Sorry to get all political in this blog. That\'s more of a Jane Blah thing, I guess.   I\'ve now gotten to the point where my extreme fear of roaches is ridiculous. When I was cleaning out the old apartment the other day there was this huge dead roach in the tub, but I refused to go near it, out of the fear that it may suddenly come back to life just to run along my arm and say \"Hiya, pally!\" I swear, they know I hate them and just come to hang out on purpose. I\'ll go to friends\' houses and hang out, and then there will suddenly be a roach near me when they haven\'t seen one for months. I have always been a fan of comic books, and deep down inside always wanted a superpower. Only now in my mid twenties do I realize my superpower is the ability to attract roaches. Hooray. I\'m the Disgusting Roach-Man. Look out crime. '),(8869,3114279,'19,July,2004','2004-07-19 00:00:00','   Nick : Lucy van Pelt: Maybe you have pantaphobia.&nbsp; Do you think you have pantaphobia?  Charlie Brown: What\'s pantaphobia?  Lucy van Pelt: The fear of everything.  Charlie Brown: That\'s  it !!!  I\'m pretty sure I have AdultADD.&nbsp; I took the online quiz.&nbsp; At the end of the quiz it basically told me I can print out the results and take it to the doctor.&nbsp; Actually it seems kinda borderline.&nbsp; I have always had trouble paying attention and sitting still.&nbsp; I already know I have Social Anxiety Disorder.&nbsp; Thanks to the Zoloft commercials and their television quiz.&nbsp; Maybe that is why I love my job so much: I\'m just a grown up version of a lot of these kids.&nbsp; As part of my job I get so many free visits to a staff psychiatrist, and you\'d better believe I\'m taking advantage of that.&nbsp; I\'ve also been told I might be Quirkyalone.&nbsp; Is there medication for that? '),(8870,3114279,'13,July,2004','2004-07-13 00:00:00','   Glenn : Yo yo yo, it\'s me, the internetless random poster. I have nothing of interest to say right now, so if your name is Jane and you complain about such things you may as well stop reading right now. I just wanted to post something so that people wouldn\'t think this is a solo act. I mean, who else would be the Laurel to Nick\'s Hardy? The Keaton to Nick\'s Chaplin? The Yin to Nick\'s way too Yang? Well, Spider-Man 2 is very awesome, and I\'ll have to recommend it to everybody right off the bat. Conversely, Dodgeball sucks really hard. Seriously, somebody got paid WGA minimum $35,000 to write this script at the very least. That\'s ridiculous! I could have written this standing on my head. Hell, I could have written this with somebody else standing on my head. What an awful piece of shit. I\'m sorry, Nick, but your boy Vince Vaughn has a stigma on his hands that he will be hard pressed to wash off. On the bright side I paid nothing for it, so I\'m not too mad.  I don\'t get people who take their minimum wage jobs so seriously. Take for example, say... Publix. Terrible job! Nobody in their right mind would want to stay there, right? And nobody in their right mind would definitely care about procedure or any other stupid crap, right? Well, the other day, bogged down by homework and studying I decided to call in to work. I\'m not the president. I\'m not the ER doctor. I\'m a fucking bagger. So the manager sighs, pauses, and then, legitimately angry, says \"Man, this is the second time you\'ve done this.\" True, I called in 5 months ago, but it is the second or fifth time I\'ve called in, but then he says \"You called in last week, too.\" I say he is mistaken, after which point he grumbles and says goodbye. I didn\'t call in last week, but if I did, who cares? It\'s a damn grocery store. It\'s a stupid job. I\'m not working my way up the corporate ladder. Urgh. The story didn\'t exactly go all of the places I wanted it to, but that\'s sometimes ok.  My roommate has left, and once again I am scared of ghosts and murderers. I was certain by the time I was in my mid twenties this fear would have subsided, but not so. Now, don\'t get me wrong, I would like to see a ghost, but I would prefer a notice via mail explaining exactly when, where, and how I will see said ghost. Tuesday, the 23rd, 8pm. They will not look fucked up, rotten, spooky, or ghastly in anyway, will enter through the west wall and exit through the east, and there will be no children\'s laughter, music, or anything else twisted, and definitely no eye contact. Simple. Oh well. I\'ve probably told that to everybody who reads this blog anyway. '),(8871,3114279,'11,July,2004','2004-07-11 00:00:00','   Nick : All those things I planned to get done today...I did *nothing.*  I didn\'t even get up until eleven.  I got nothing done.  No apartment hunting.  No cleaning.  No laundry.  No grocery shopping.  Nothing.   So I have nothing to blog about.  Goodnight. '),(8872,3114279,'10,July,2004','2004-07-10 00:00:00','   Nick : Tomorrow is my only day off.  Next weekend I will have three days off in a row.  I don\'t know what I will do with myself.  Right now I have no life, so work is becoming my life.  If I were completely trained, I\'d even volunteer to work tomorrow.  After next week I\'ll be pretty much considered trained, or trained enough.  There is a week of training that new employees have to have, but it won\'t happen until the middle of next month.  Next week I work the evening shift, so I\'ll get a shift differential.  After next week I can work some overtime, which pays time and a half, which I plan to take advantage of.  I need all the money I can get.  I\'m still doing some apartment hunting, and I have a few prospects.  It all depends on what I\'ve got to have up front.  I\'ve got to get my own place.  If I have to, I\'ll wait another month to get some money saved.  But I\'d prefer to be out by the beginning of next month.  Tomorrow is my only day off, so I\'m going to call some places.  I know all of this is really boring.  Work today was pretty nice.  I got to take the kids who earned at least 80% of their goal points during the week to a rec room where we watched a movie.  So I got to sit through Spy Kids, which I had never seen, and will be perfectly happy if I never see it again.  I probably could have dozed off, if not for my boss sitting next to me on the couch snoring loudly.  We played some kickball as well.  That\'s always fun, especially when the kids get competitive.  It\'s still a little surprising to me (and admittedly somewhat amusing) to hear children yell \"Fuck you\" at each other; I suppose I\'ll get desensitized to it eventually.  Tonight I put on Mallrats, which I had never seen, and I fell asleep about midway through.  What I did see was not great.  That\'s all I\'ve got for today. '),(8873,3114279,'09,July,2004','2004-07-09 00:00:00','   Nick : I guess everyone is too busy to post lately.  I haven\'t really had the time or reliable internet access.  I\'ve started looking around for apartments here in Naples.  I was planning on staying where I\'m living now until the end of August, but I\'m not sure I can do that.  I\'d really like to have my own place.  There are some pretty decent sounding places in Golden Gate, which is the area of Naples where I work.  I\'m going to check a few places out; it all depends on what kind of money I can get together.  Work has been going well.  So far it\'s been glorified babysitting.  I haven\'t been able to participate in group sessions yet or really get to know the kids and get into their heads.  I really like the job so far, and I\'m sure I\'ll like it more when I no longer feel so new and overwhelmed.  Tomorrow will finish my first week of work and my second week living in Naples.  Not much else to write. '),(8874,3114279,'06,July,2004','2004-07-06 00:00:00','   Nick : Today was my first day at work.  I\'m sure I could think of a lot of things to say, but I\'m completely exhausted.  I plan on being in bed within the next hour.  I will say that I really love the place and the people I will be working with, and I like the work so far.  I had become so used to sleeping in and staying up late that I didn\'t go to bed as early as I should have, and was up quite early.  Add that fact to the three hours of paperwork I had to go through this morning and then the overwhelming training this afternoon, and I\'m fucking pooped.  I\'ll post more in the next day or two. '),(8875,3114279,'05,July,2004','2004-07-05 00:00:00','   Nick : The comments are back!  You all can now continue to not comment.  Today I was asked an interesting question: If I had a Theme Song, what would it be?  I\'m guessing Theme Song as in, a song that would represent me as a person and my personality.  I take music very seriously.  I usually don\'t just listen to music as just entertainment.  My roommate used to say some stupid thing about good music speaking to your soul or something, and though I wouldn\'t put it in such cheesy terms, I agree with that sentiment.  My favorite songs make me feel a certain way, a specific way, and the best songs (just like the best movies) give me chills and make me feel incredible things.  I don\'t really have a favorite song, and I don\'t really have a favorite band or artist, although if I were pressed, I will usually say Stereolab is my favorite band, based on the fact that I have listened to them for years, since junior high, through all the changes and phases.  But a Theme Song would be hard to choose.  Not just a song that I like or that makes me feel something, or whatever.  A song that says something about me in the best way.  In high school psychology class, one class project was that each student had to turn in a tape or a CD and pick one song to be played to the class, and the class had to guess which student had picked which song.  Most of the class, of course, picked the generic music that was popular at the time, the class guessed mine right away, since I had different tastes.  I\'m not sure what the point of that project was.  It\'s going to take some thought to pick a Theme Song, but I\'m going to try.  I\'m sure you\'d have an interesting Theme Song.  My first week living in Naples is over, and I just have to say, I really love it here.  I was a little worried about being homesick (not very homesick at all), or being broke (I  am  broke), or being intimidated (only a little intimidated), or whatever.  But I love being here, more than I thought I would.  It\'s definitely been worth it all to move here.  I\'m really excited to be starting work, doing what I\'ve really wanted to do, and I\'ll be excited to start school here.  I like to get to go swimming whenever I want, I like laying around the house with Chud and Samson (my cousin\'s cat and chocolate lab, both of which I\'ve started to grow attached to), I like going to the pier, I like just driving around, I like hanging out at the used record store, I even like going to the Coastland Center Mall.  I usually hate malls, possibly because of the sorry \"mall\" back in Lake City.  I\'m happier than I thought I\'d be.  Sorry to be so positive here, I hate to break the trend of negativity.  There are things I dislike here, but the good definitely outweighs the bad.  I finally escaped Lake City.  \"We Built This City\" by Starship is the #1 worst song according to VH1.  I\'m on board with that. '),(8876,3114279,'03,July,2004','2004-07-03 00:00:00','   Nick : Happy 4th.  Today I slept in, went swimming and laid by the pool, had a cookout with my family here, drove through town and saw some fireworks, and now I\'m watching VH1\'s 50 worst songs ever I think it\'s called.  Hope everyone has had a good and safe day.  I keep having Lake City dreams at night.  A few nights ago I dreamt about dozens of people that I haven\'t seen since high school.  I haven\'t seen these people or thought about them in more than five years.  I was just surrounded by these people, hanging out with them, and then playing sports with them.  Dozens of people.  What the hell?  And then last night I dreamt about being in Lake City, just going around town or something, and again seeing people I haven\'t seen and really should have completely forgotten about.  When I woke up, it took me a minute to realize that I\'m not in Lake City anymore.  I don\'t exactly know why I would have these dreams.  I certainly don\'t miss Lake City, and I haven\'t thought about some of these people in several years.   Oh wow, Crash Test Dummies \"Mmm Mmm Mmm Mmm,\" that is a terrible song.  I commented a while back that I wasn\'t crazy about the Soderbergh version of Solaris.  I picked it up at the used record store for very cheap (and it was brand new, not used!), and watched it again.  Now I have to change my opinion and saw that it is quite good.  Maybe it\'s because I had only seen it once.  Maybe also because I recently finally saw the original Russian version, and skimmed the Stanislaw Lem book that both films were based on.  The Lem book is very spread out and the Russian film, I think, tried to tackle as much of the book and as many of the themes as it could, while I think Soderbergh pretty much picked one thing and went with it.  I think he picked the strongest theme of the book and the original film, and it really is outstanding.  This is another film that most people I know dislike, possibly because it is so slow moving, I don\'t know.  I hate it when people claim that the reason that someone didn\'t like a movie is because \"they didn\'t understand it.\"  How snobby.  Though, truly, a lot of times that may be the case.  Anyways, this film asks some of the same questions as Vanilla Sky, about memory and reality and whatnot.  But the George Cloony character comes to a different decision in the end than that of the Tom Cruise character.  Tom Cruise chose to wake up from his dream reality and live a real life, which in the past was what I would have most agreed with, where the George Clooney character chose to go into the Solaris planet and live a dream in order to be with his deceased wife (and by the way, Solaris is much better than Vanilla Sky, and the way Solaris shows Clooney\'s characters decision is much much better).  Really, why is the decision to live in a dream a bad decision?  Why would a \"real reality\" be better?  Wouldn\'t a dream reality be just as real?  Would you rather live in the Matrix, or live in the sewers.  What\'s the difference?  The Matrix is just as real.  By the way, I hate the Matrix.  Maybe I would choose to go into the Solaris planet.  What would you choose, Glenn?    I now have 50 horrible songs stuck in my head, courtsey of VH1. '),(8877,3114279,'03,July,2004','2004-07-03 00:00:00','   Nick : Having very little to do this past week as I wait for next week, and my first days of work, to arrive, I think I\'ve watched more television in one week than I have in the past year.  I\'ve become particularly interested in watching Celebrity Poker Showdown.  Along with Court TV and all these investigation shows.  I have made it out to the pool and Naples pier, which is fantastic.  I also saw Farenheit 9/11 a second time, going with my very Republican Aunt and Uncle, which was amusing.  And I agree that Michael Moore is very slanted, and as he says, much of the movie is his opinion, and it is propaganda, and it is at times very heavy-handed, but there are still facts there.  Facts that can\'t be ignored.  The propaganda is pretty entertaining in itself, but beyond that are actual proven facts.  The movie hopefully will encourage people to do some actual research for themselves.  And as a bumper sticker I saw recently says, \"let\'s not vote for W again in 2004.\"  I enjoyed your list, Glenn.  And for the most part I can agree with it.  I don\'t have the ambitions you have, though.  I\'d like enough money to be comfortable, but having been here in Naples, which is half a pretty normal town, but the other half is very wealthy.  And wealth seems to equal snobbery, at least as much as I have encountered.  Of course, I\'m having to work towards something, and I imagine that earning wealth by working hard as opposed to just being handed down wealth probably builds better character.  But I don\'t think I\'d ever want to be wealthy.  I\'d like to have enough to be secure, as you said, in paying my bills and being able to swing an emergency if that were to happen.  One home is enough for me.  I don\'t have the same motivations to compile such a list, but just for fun, here is mine:  1. Companionship:  As in a female companion.  I\'ve been called \"quirky-alone,\" though I\'m not sure how much I agree with that label, but I am comfortable being alone, and I am \"holding out\" for that \"special someone.\"  I don\'t want to ever end up with someone out of desperation or just because it\'s what people do.  But I seriously have hope and believe that I will one day meet that special someone.  I am really a hopeless romantic, and one day I\'ll get to put that to use.  2. Career:  I want to have a career, not a job.  I want to do something that I really want to do, regardless of what anyone else thinks I should do, and regardless of money.  I want to do something fulfilling.  I\'m about to start work in the field that I am most interested in, so maybe this will be the start of what I am hoping for.  3. Financial security:  I can\'t imagine ever just wanting to be \"wealthy.\"  But I want to be secure.  I want to know that my bills will be paid.  I want to know that I have some money in the bank in case of emergency.  I want to be able to afford some luxuries, which for someone like myself is not much at all.  As I\'ve moved here, I have put myself in the position of being very low on money, really almost completely broke.  I\'ve always lived paycheck to paycheck, as I\'m sure you have, Glenn.  I hope to one day be a little more secure, one day soon, hopefully.  4. Circle of friends:  As cheesy as that sounds, friends are important.  I\'ve been fortunate enough to have a few really great friends, ones that I always look forward to talking to and seeing.  Unfortunately, I\'ve just left them five hours away.  I know nobody here in Naples, besides my family, but as time goes on I\'ll make new friends.  And now I want to play poker.  So I can only think of four right off the top of my head.  I\'m not really in much of an existential crisis right now.  I am a little nervous about starting this new job.  For some reason, it feels a little like it used to feel right before a new year of school started when I was in high school.  Strange.  I went out and bought a shitload of clothes, basically half a whole wardrobe.  I\'ve found a few spots locally to hang out.  And I picked up an amazing Antonio Carlos Jobim CD that I have listened to completely non-stop all week, for $5 at the local used CD store.  And I plan to see Spider Man 2 Monday.  So I guess that\'s all to report for now.  Back to Celebrity Poker Showdown. '),(8878,3114279,'03,August,2004','2004-08-03 00:00:00','   Glenn : Squirting ass-gremlins? How can I follow that? '),(8879,3114279,'01,August,2004','2004-08-01 00:00:00','   Nick : It\'s a little after 9 in the AM, and I got home from work just a short while ago.  I\'m pulling midnite shifts all this month.  I\'ve been up for something like 26 hours.  Yesterday I worked until 4PM and then had to be back at midnite.  Being that I\'m not used to being up all night and sleeping during the day, I couldn\'t sleep yesterday afternoon.  I ended up watching Futurama all evening.  Now I\'m exhausted, but I don\'t want to sleep, it\'s a nice day out and I feel like I should be up.  I\'m sure I\'ll get used to being up all night.  This is only for a month.  Plus the shift differential is nice.  But it\'s playing tricks on my body.  I\'m tired, yet I want to stay awake.  And I didn\'t eat dinner until after midnite, so it\'s obviously got my digestion out of whack, as I now have a slight case of the sqirting ass-gremlins.  It\'s obviously messing with my judgement too, as evidenced by what I just posted.  Good to see Glenn is poting on here, what an effort.  I think I should get to bed now befoe I starrt makeng tieping mifstakes. '),(8880,3592700,'30,July,2004','2004-07-30 00:00:00','   urlLink    this is what i see first thing every morning. camera phone pic, so quality is a bit squiffy.  '),(8881,3592700,'28,July,2004','2004-07-28 00:00:00','   giant thunderstorm! with hail! and knocking-things-off-shelves-thundercracks!  *poing*  '),(8882,3592700,'27,July,2004','2004-07-27 00:00:00','   so i\'m working through my little list of things to do for life in general and college registration. and one of those things is to take assessment tests. the college website offers sample tests to help you fine-tune your brain.  i totally ruled the reading test. like, 100%.  i totally ruled the writing test. like 100%.  i have yet to get an answer right on the math test. i have done all 14 questions at least 5 times each, with no correct answers. it\'s a basic algebra test!! i have spent almost 3 hours on several algebra tutorial websites, and none of the sample questions they offer have anything like the equations used in the test.  a little background : high school was spent in 4 or 5 different christian schools, including one year homeschooling. none of the teachers there were qualified to teach algebra, so i took consumer math 3 times. the year i homeschooled i took algebra 1. that year was largely self-taught, since my mom was too depressed to be much of a parent. i tried really really hard and spent about 6 hours a day on each lesson trying to understand it and eventually my hard work paid off : i got a 99 on the course. WITH all work shown! i possess an above-average intelligence, and i do really well in math once i grasp it. but i need help to get there.  crud :( i have to try to re-learn everything i purposefully forgot, in the next 14 hours or so. i can see myself in the remedial math class already.  '),(8883,3592700,'27,July,2004','2004-07-27 00:00:00','   i went to college orientation today. i was 5 years older than everyone at my table, but then about an hour into it, this old dude showed up so i didn\'t feel so weird.  it was looooooooooooooooooong  and poorly liiiiiiiiiiiiiiiiiit  and booooooooooooooooring  but infoooooooooooooooooormative  holy shit i have a lot to do before the 23rd (school starting)  i\'m all tuckered out. i\'m going to go watch shark week with my boys. rawr! bring on the bloody water!  '),(8884,3592700,'26,July,2004','2004-07-26 00:00:00','   yes, it was me who picked all the dried pineapples out of the 2 pound bucket of trail mix in a half hour timespan. you should just be glad i didn\'t get the kind with those chocolate swirly things in it.  next stop : dried papaya  '),(8885,3592700,'25,July,2004','2004-07-25 00:00:00','   my bedroom smells weird. like beef stew.  very odd.  '),(8886,3592700,'24,July,2004','2004-07-24 00:00:00','   i have realized that i will never sleep through the night again. i don\'t even have little kids. i have a 14 year old stepson who sleeps like the dead. i have two dogs and a cat. if i were forced to choose which pet i loved most, it would be the cat. but not last night. last night i would have cheerfully shipped her to rwanda.  every 90 minutes, she crept up next to my face and then patted it lightly, gently, then insistently and with claws. i know she had food and water. i know her litterbox got changed. i played with her vigorously half an hour before bedtime. so WHY fucking WHY did she have to wake me up!!? at the beginning of the night, i would pick her up and drop her off the edge of the bed. towards 4 am, i threw manners to the wind and pushed her off.  after that she left me alone.  ten minutes later i heard a groan from downstairs, followed by quiet whining that promised to become loud howling, quickly.  the fucking dog had to pee. or whatever.  i let them out, decided i wasn\'t going to wait for them to finish up, and stomped back up to bed. my husband could let them in when he got up.  so this is why i am unable to summon up the energy to Do Things Today.  the coffee widgetry that glistens everywhere in my kitchen has been fully employed.   my husband and stepson are in BFE at a car show for at least 10 more hours.  and all i\'ve done the past two hours is read blogs and webcomics and news and rub my eyes and jaw-creakingly yawn.  i should be out on a long walk with the fat dog. i should be grouting the tile. i should be cleaning up the dessert dishes from dinner last night. i should be plucking my eyebrows. i should be looking for a job. i should be looking for a studen loan.  but i haven\'t read penny arcade in a week. so fuck the shoulds.  '),(8887,3592700,'24,July,2004','2004-07-24 00:00:00','   6 people total eating dinner at my house tonight (several cancelled)  this is what we are preparing :   kebabs with choice of steak, lime-ginger-garlic marinated shrimp, turkey wrapped in bacon, zillions of vegetables, pineapple, chunks of white nectarine, and tofu  grilled corn on the cob  rosemary & olive oil roasted potatoes  and for dessert, chocolate mint brownies with chocolate cream cheese frosting  i feel full just thinking about it  '),(8888,3592700,'22,July,2004','2004-07-22 00:00:00','   ...for going to the grocery store in sweatpants. not matching-set soccer-mom sweatpants. my husband\'s old, cast-off 80s-style with muddy pawprints and dog fur all over them sweatpants. it\'s like i never left alaska.  when i was a kid, we\'d go to the grocery store and there would be massively unkempt women everywhere in old sweatpants and flannel shirts. the sweatpants would invariably be tucked into a pair of sorel boots, and a faint odor of dog and/or b.o. would follow in their wake. i always vowed to never go out in public in sweatpants.  what have i become?  '),(8889,3592700,'22,July,2004','2004-07-22 00:00:00','   i slept too long, and as a result, am craving coffee. so i\'m drinking this overbrewed sludgy stuff my husband\'s friend made last night because i\'m too lazy to clean out the coffeemaker and make a fresh batch. yech. but mmm.  it\'s been raining off and on for the past two days, with spectacular thunderstorms and flooding in parts of town. i felt that was my cue to send out a massive \"fuck you\" to low-carb diets by making potato chowder. so good. i even toasted chunks of sour rye and floated them on top of each bowl.   i better go clean up. i moved the dogs\' kennel and they\'re still getting used to it. i have to use the Loud Mommy Voice and grab collars. with all the rain, they\'re muddy and after three cleanups, i gave up and left them in the kennel. little mud chunks everywhere.  we\'re having 5 people over for dinner tonight. bleh.  '),(8890,3592700,'22,July,2004','2004-07-22 00:00:00','   ....i do that annoy my husband :    &nbsp;  1.buy store brand stuff instead of commercial brand stuff (\"it tastes wrong!\")   2.leave the hose on the grass after watering it   3.communicate non-intelligently and be a bitch in general when the pms monster is around   4.worry. and be anxious   5.read in bed and keep him awake, because he has a hard enough time sleeping as it is    &nbsp;   &nbsp;  ....he does that annoy me :    &nbsp;  1.leave his stuff in a central location and forget about it for 3 months, 2 weeks, and a day, and then need it urgently, suddenly RIGHT NOW and griping that i \"put shit where he can\'t find it\". because the middle of the driveway is the best storage place for those dvds he borrowed   2.see #1   3.snore when i already can\'t sleep (but most of the time it\'s cute)   4.not cleaning up after himself   5.use 17 cups of oil to deep-fry&nbsp;1 sliced-up potato, thus ensuring that everything from here to colorado springs is splattered with grease  '),(8891,3592700,'13,August,2004','2004-08-13 00:00:00','   i\'m hungry.  i have blisters on my heels the size of quarters from walking a mile in cheap shoes yesterday.  not that the two are related....or are they? i could eat my blisters!   ok, even i\'m grossed out.  '),(8892,3592700,'12,August,2004','2004-08-12 00:00:00','   there should be an actual time called ass o\'clock. it should refer to everything before 6 am, and after 1 am. i\'ve become soft and weak. i used to get up at 3:30 every day for work after going to bed at 11. these days, being hauled out of bed at 6:15 by my husband\'s ex calling is excruciating. god forbid i ever get less than 8 hours of sleep.  '),(8893,3592700,'12,August,2004','2004-08-12 00:00:00','   i\'m not talking about music records.  i\'m talking about my math test score. 45 out of 120 is enough to keep the classes i\'m registered for, and lets me into a remedial course, which i really wanted to take. if i never really had a chance to learn it right in the first place, why not do it now?  oh joy! oh bliss! everything is going to be ok, school-wise.  and now, pig-dog!   urlLink    this dog!  '),(8894,3592700,'11,August,2004','2004-08-11 00:00:00','   holy fucking shit. i just googled mariana trench and hit the little shamrock. and what do i find? what seems to be a cool easy-science website CRAWLING WITH SPYWARE POPUPS!! a popup jumps out for that goddamn \"set your homepage to searchwebgoto.com? click yes to continue\" and alt+f4 doesn\'t get me out of it. moral of the story : get your underwater jollies elsewhere. that website is whack.  i call these lows \"the trench\" in reference to the mariana trench. everything is dark and i feel a lot of pressure. when i was a kid and thought about my mom being depressed, i assumed she felt mellow and relaxed all the time because she slept so much. now that i\'m depressed, i realize that\'s not the case. i feel like a hamster on an endless wheel run. my thoughts race around and around and seem to get darker and darker with each squeak.   i can\'t do anything. i can\'t focus on anything or even remember names for things. my emotions are helter-skelter. eventually i get worn out from trying so hard and i have to sleep. i think there\'s probably a lot of anxiety mixed in with this particular bout due to school starting and my lack of employment.  in short, i feel like shit and two days ago i felt giddy, exuberant, on top of the world. highs and lows.  i\'m doing everything i\'m supposed to. eventually i will feel better. i am really thankful to my husband for sticking with me and being such a great support.  i\'m reading an old book by greg bear about these living cities and the people who lived in them. it\'s kind of post-apocalyptic. it\'s not really a great book...i can tell he was still refining his style and learning to develop more cohesive plots. but this is what i love about greg bear : he always has a fresh twist, something new to think about, something i never thought of before.  i got up at 5:30 this morning to drive 35 miles to an interview. i arrived and they asked me what i was looking for. i said, part-time, preferrably an administrative support position. one of them got this funny look on his face and said, you\'re not here for the marketing director position? no, i replied, i thought this was an office clerk job. he cleared his throat and read through my resume quickly. well, it\'s too bad. it looks like you\'d be perfect for the job.   i\'d like to state that i am not, nor will i ever be a marketing director. nothing on my resume says this. not even close. asshattery! rampant asshattery!!  '),(8895,3592700,'09,August,2004','2004-08-09 00:00:00','   my betta died. he wasn\'t eating the past two days...i tried to feed him about 6 hours ago and he didn\'t have much interest.  he was a beautiful electric blue.  he used to make me enjoy getting up in the morning. i\'d stumble into the kitchen to start the coffee and he would swim wildly around the bowl. after a while, he did it when my husband came in the kitchen too. my husband thought he wanted to be fed, but he would do it after feeding too. happy to see us, i guess.  just your regular 70\'s hong kong action star siamese fighting fish.  he used to terrorise the guest betta that lived with us for a bit. i\'d put their tanks next to each other and he\'d get all puffy and change colors. the other fish would try to hide in the gravel. it always made me laugh.  my husband wants to get another one and call him sonny chiba. i don\'t know if i\'m ready yet.  '),(8896,3592700,'08,August,2004','2004-08-08 00:00:00','   i can hear the sparkling from here.  yes, all 3 bathrooms are clean.  carmie = efficiency incarnate.  the signs point towards dishes in my future.  should i make my stepson do it? he\'s on his second hour of cartoons, the bum.  '),(8897,3592700,'08,August,2004','2004-08-08 00:00:00','   today we went to the bug-in show at bandimere speedway. it was a lot of fun....i want a karmann ghia now. a convertible. and of course i would paint it pink. with flames. and get one of those custom jobs stating \"carmen\'s ghia\" on the back. but no pink fuzzy dice. i hate shit hanging from my rear-view mirror.  i told my husband this and i thought his eyes would pop out in shock that i actually wanted an old vw. he\'s going to try to find me one for about $500 and we can restore it together.  we watched the time trials for a while. one guy\'s muffler fell off. they sent this big dude on a little scooter up the track to get it. i, not having had coffee this morning, thought he was competing for a time trial. a porsche was revving up in the lane next to him and i thought that was just really unfair.  there were a lot of old vw busses there...i saw one all fixed up and tricked out for sale - $49,000.00. jesus fuck.  when we arrived, my husband\'s friend thought perhaps they charged entrance fee per car, so we (five of us) piled into the \'62 bug he was showing. and of course, they charged per person. truly, i haven\'t seen my knees that close-up since the last time i shaved my legs. however, i\'m just glad it had a back seat. my husband\'s \'63 doesn\'t.  i\'m grilling eggplant for lunch and it smells almost done.  '),(8898,3592700,'07,August,2004','2004-08-07 00:00:00','   i\'m tired. exhausted, even. i keep having to remind myself to close my mouth as i type.  it was one of those gone-from-the-house-for-12-hours weekend days. i have spent so much time at home the past year that i feel off if i\'m gone for too long. i start missing the dogs and the cat.  we bought school clothes for my stepson today. he has to wear uniforms this year for his new school. we had a serious difference of opinion about pants. i hate, hate,   loathe pleated dress pants with tapered legs. i hated them when they were in style (ten years ago) and i hate them now. they are an incredibly unflattering style on anyone. why couldn\'t he wear (more fashionable) flat-front, boot-cut dress pants? oh no! those are \"gangster pants\". the poor kid is going to look like a dork. he\'s small for his age, starting high school at a new school where he doesn\'t know anyone, and already a bit otaku, so he needs all the help he can get. at least we got him some cool shoes.  on further reflection, it occurred to me that we never really asked him what he wanted. however, there were only two options for uniform clothes, and my husband made it abundantly clear the \"gangster pants\" weren\'t an option. poor kid, stuck shopping for clothes he doesn\'t want with two parents who can\'t make a decision without bickering over it.  we went up to my in-laws\' house to bask in their air-conditioning, and played poker.   i won $2.50.  and promptly spent in all in one place : dq, on a brownie batter blizzard. (the \"much thafer\" one)  is it wrong that i laughed my ass off at that poor guy in the commercial? and promptly tried to imitate the sound of his screams while his tongue was being eaten by the blender? i can\'t find a download of it and i\'m too tired to look really hard.  i am a cwuel bathtawd.  '),(8899,3592700,'06,August,2004','2004-08-06 00:00:00','   intermission.  it\'s been a hard week.  some weeks are harder than others.  i keep thinking about last year around this time. maybe if i hadn\'t smoked that pack of cigarettes about that time, it wouldn\'t have happened.  i will never smoke again. anything.  all those years i didn\'t smoke, and stayed \"quit\".  why did i pick that particular two weeks to start again?  it hasn\'t even been a year yet. please don\'t expect me to be \"over this\". please don\'t expect me to \"stop being morbid.\" please don\'t tell me \"work this out with your therapist so you stop thinking about it.\"   i will probably think about this every day for the rest of my life.  we return to your regularly scheduled program now.  '),(8900,3592700,'02,August,2004','2004-08-02 00:00:00','   urlLink    hana drowses on the seat   urlLink    samson stole the seat from hana. see the look of triumph in his eyes?   urlLink    i want to live on the high mountain plains   urlLink    driving towards south park (yes, that south park)  '),(8901,3592700,'01,August,2004','2004-08-01 00:00:00','   so it\'s like 97 degrees here.  we went up to the mountains for 5 hours or so to escape. coming back down I-70 was like sitting in an oven as it heats up.  but, i saw burros near buena vista. the cuteness! the cuddliness! unfortunately, we didn\'t stop since the dogs were with us.  i might post a picture or two tomorrow if any came out good. i\'m starting to dislike my digital camera. it\'s damn difficult to get clear pictures on it. it\'s like an old-fashioned daguerrotype! the subject has to hold still forever.  however, in about 4 weeks, things will start to cool off, and then it will be my favorite time of year. i love fall.  '),(8902,4060812,'19,August,2004','2004-08-19 00:00:00','   bla bla bla bla bla bla bla bla .... now in visual programming ... waiting for cairu to email me the commands for the freaking project. its not done yet ... argh ... i am BORED!!!! anyway, today in a better mood ... just going to do the basics and hand it up ... hope can pass la ... anyway, think i made a new friend. His name is qi xiang, in del\'s class and also huimin\'s friend from her sec school(i think).  well, talking about guys, i miss Him sooo much ... dunno whether he has returned back to Germany already ... think he should have been back already and preparing to start school ... haiz ... cant wait till the end of the year to go and see him !!!! hahaha ... think i am going mad soon!!!! haha ... '),(8903,4060812,'18,August,2004','2004-08-18 00:00:00','   here i am again ... 2nd time today ... still at school comp lab doing the project. Cairu says that she will help me tomorrow. Hopes she keeps her words. anyway, i am feeling much better. sorry if i offended anybody just now ... i was damn fraustrated and about to cry(i think i did). Anyway, going to go home and attempt to study cause most of the mock exams are approaching and i dun wanna fail ... haiz!!!  anyway, now, huimin\'s project seems to have conked up, and she is back at her usual mood again ... dunno it is the how many time already ... i sort of really admire darren for being able to take it. I am already quite tolerant, and i and starting to get irritated at her more often. shumin seems also to be getting pretty short-tempered. I dunno whether it is the influence of huimin, or if she is already like that.haiz!!! '),(8904,4060812,'15,August,2004','2004-08-15 00:00:00','   I FEEL LIKE A DUMBASS!!!!! i dun even know how to do anything, sometimes i think i am better off dead !!! i dun know wad the hell is wrong with me. Sometimes, i think that i really am being taken granted of !!!! they dun seem to care about me!!!! i sometimes think whether do i deserve such treatment or not. Take the VB for  example , huimin and shumin only seem to care that as long as their project is done, then that\'s all. What about me? They have completed the project, but i haven!!!! They wanna leave already ... but i haven even gotten anywhere in the project, and i am getting sick of this project already!!!!! arghh!!!!  i seriously think i am gonna flunk this freaking course and disappoint everyone in this entire world!!!! My mum and dad are so gonna make me feel guilty for not doing well again ... and maybe i should just stop school and start working !!!! argh!!! '),(8905,4060812,'15,August,2004','2004-08-15 00:00:00','   Just saw Bel\'s blog. She seems to be very happy now ... made 6 new friends and even have a group name. While here i am, feeling so miserable on how come life is so sucky for me? I feel that most of my poly friends take me for granted, when they actually dun like me ... and i dunno , but i just feel like shit now ... I am very happy for Bel, but i dun know if this will cause us to drift apart, by the fact that we are both now in different polys and my class time tables are now diferent. We dun get to talk much, wad more hang out with one another. HAaiz!!!! Really missed 4F!!!!! Sometimes, i dun know wad to look forward to going to school for, sometimes, when i dun even understand wad the lecturer is talking about or how to work those equipments!!!! Argh!!!! '),(8906,4060812,'12,August,2004','2004-08-12 00:00:00','   Hey!!! Today is Saturday and i met up with my TK friends!!! Woah!!! So Happy!!!! I missed them sooo much!!! Haha!!!! Anyway, went to school early today, at about 9, maybe. Tried to do that freking visual programming thing, but didnt get far on it!!! So frustrating!!!   Anyway, left school at about 12, rushed over to Orcard to look for the guys!!! Met them at Breeks!!! Bhavna, Bel, Kesh and Nurfarhana. Di and del was not there yet !!! Missed them man!!! Haven seen these guys for like at least 2 â€“ 3 months ... HAHA!!! Anyway, didnt have a great lunch, was okay la ... everyone was sooo slim and pretty, made me feel like shit sia ... must go on diet sia ... hmmp!!! Proceeded to Paragon to watch MJC perform ... i0t was actually pretty lousy cause the music was sooo damn soft and the moves werent that great either, but wadever!!! Anyway, met some of 4F ppl. So glad to see them, felt like hugging them too but didnt dare ... haiz!!! Met sham, ks, and i think thats about all.   Oh ya, i dun know why, but i always feel like a freaking lamp-post whenever arund Darren and Huimin. They always seem to want me to be around, but i feel out of the place. Haiz, i think i am fated to be a match-maker for life sia ... first Feli and Yongsheng, now Darren and Huimin. Well, hope that things will turn out great for these 2 couples, especially between Darren and Huimin, or not have to step in and help. Its not that i mind, but i hope that it would be solved between themselves.   Another thing, i dunno why, but it seems that the group is like going to offically break lor. I dun seem to be talking to Daisy and Feli anymore. We dun seem to have anything to talk about, althoug we used to have lots to talk about!!! Well, maybe this will turn out to be the best for everyone!!! Anyway, Tg will be going to TW for some com comp., will miss him, but only as a friend. Istill dun understand how he can be mistaken that i like him. Haiz!!! Just let it pass la ... Thatâ€™s how life is, i guess.   Tests are coming soon, and i think i better start studying hard or i am sooo gonna fail man!!! I dunno whether it is because the course is not my knid or is it that i just dun understand wad the hell the lecturers are teaching. Especially computer hardware. Digital circuit, Data com. And visual are gettign harder too !!! How the hell am i gonna to be able to be the top 5 if this continues!!!! I feel like i am gonna be in deep shit sia ... Better start studyuing man!!!!   Kk, think i am gonna end pretty soon !!! itâ€™s getting alittle too long!!!!  Tata>>>  P.S: I dun think i like Olaf at all . i treat him as like a good friend only ... so here i am ... offically(or should it be unoffically) SINGLE again!!! Hahaha ... '),(8907,4060812,'12,August,2004','2004-08-12 00:00:00','   Yap ... The 1st day, we all met at the tanah merah ferry terminal ... ohhh ... how glad i was to see him ... lahtough i like saw him only like less than 8 hours ago ... haha ... i feel like a small girl !!!! Anyway, when we boarded the ferry, we all sat down; me, my sis, him, his sis. And we all actually fell asleep !!! cause we were all too tired ... LOL!!! Anyway, when we got there, we got a very very warm welcome from the people there. Haha ... There were also pretty much sights to see !!! muahahahah!!!! But i will stay loyal to you-know-who!!! Anyway, That day, we did pretty much almost every landspot there was to do ... I WAS SO DAMN TIRED TO DO ANYTHING BY THE TIME IT WAS NIGHTFALL!!! We did squash, tennis, badminton, basketball, volleyball, and many many more which i have practically forgotten. Anyway, we also went swimming for like about a few hours and the sun was scorching like hell!!!! Me and Olaf like had a few waterfights, trying to drown one another out!!! Hahahaha !!! We had such great fun that day, but also lots of muscle-aches. Oh ya, that night after dinner, there was a show(as usual) and it wasnt as interesting as i had remembered it to be, but anyway, we stayed on till the end of the show. After the show, we proceeded to the disco . When we got there, as i was sitting down and he went to get drinks, 2 girls approached him and started to talk to him. Surprisingly, i did not feel anything at all ... He felt pretty embarassed and guilty, i dun know for wad reason. But i decided to excuse myself. I left to go outside and he followed along. I told him that he could go and accompany his new girlfriends and he said that he would rather accompany me ... Wadever. We went for a walk at the beach. The beach was sooo peaceful ... I felt so at ease. We talked as we walked from one end to the other, and back again. It got too late and we decicded to go back to out rooms and sleep. Day 2 was for water sports . We did practically almost every water sport and was so exhausted by noon that we just fell asleep after lunch. We woke up beside one another at about 4pm and went on to play some more sports and swimming. Another waterfight went on, till i surrendered and let him drown me just once. Hahaha !!!! Day 3 was the day we were leaving club med and i actually miss this place. We spent the morning packing our lagguage and left for the ferry terminal. Having reached Singapore, we proceeded for dinner at Changi. When we sent them to Changi Airport for their flight to Colombo, i gave him a hug. Thats about all!!!! I MISS HIM SOOO MUCH ALREADY AND ANTICIPATING FOR HIS RETURN VISIT TO SINGAPORE AGAIN!!!! '),(8908,4060812,'05,August,2004','2004-08-05 00:00:00','   hey!!! just got back from a short holiday in Bintan Club Med!!!!! Woah!!!! It was so fun... i got a suntan and a sunburn!!!! Think that the sunburn is all worth it as now ... i am pretty tanned and brown ... Lol!!!! Had a GREAT time with my dear Olaf ... Hahaha ... I FEEL SOOO HAPPY !!!!! Hahaha ... Went swimming most of the time and i guess that\'s where most of my suntan came from !!!! Hahaha ... Think i am going crazy !!!! Hahaha!!!! '),(8909,4060812,'05,August,2004','2004-08-05 00:00:00','   Lalala ... hey, i am here again !!! dun really know wad to say too ... now in visual programming ... haiz ... although it doesnt suck as bad as before, but i think it is still equally bad ... haiz ... dunno how i am going to survive through the next 3 years of poly ... guess have to try my best ...  haiz haiz ... I AM GOING TO BINTAN CLUB MED !!!! Yeay ... i am so happy. Moreover, i am going with Olaf ... yeay ... haha ... its going to be one of the best days even since i started school ... haha ... i will be there from saturday till monday ... woah!!!! yeay !!!! i am so looking forward to it !!!! Hope i have a real great time !!!! I AM GOING TO BINTAN CLUB MED !!!! YEAY ... I AM SOOO HAPPY!!! '),(8910,4060812,'04,August,2004','2004-08-04 00:00:00','   Our class is splitting and its real fast !!!! i dunno wad the f*** is wrong with the group ... especially Toh Ghim ... argh ... dunno wad to say la ... nowadays when we are out ... i feel so out of the place, i just am like keeping quiet and nobody listens to me or attempts to talk to me. '),(8911,3831024,'30,July,2004','2004-07-30 00:00:00','   i seek refuge from my storms i seek comfort from my thorns i seek solace from my anger and not end up a strangler  i seek wisdom from the wind i seek knowledge from within i seek intellect everlasting time all encompassing  i seek courage in defeat i seek hope in retreat i seek&nbsp;the blood&nbsp;of the flames that dances the dawn and has no name  i seek love deep and true i saw love lost and blue i seek love in scarlet hue i seek love, i seek you.  ngek ngek ngek. corny corny corny... hehehe. basta enjoy   '),(8912,3831024,'15,July,2004','2004-07-15 00:00:00','   i learned that you&nbsp;can\'t&nbsp; claim to love God and hate your fellow man at the same time.   '),(8913,3831024,'15,July,2004','2004-07-15 00:00:00','   glance. waiting to cross the road. glance. nice store. glance. whoah cute girl! glance. argh, green light. glance. REALLY CUTE GIRL. glance. yellow light. glance. SHE\'S LOOKING THIS WAY. glance. time to cross the road. glance. she\'s smiling. glance away. face it gio, she\'ll never be yours, never know your name, coz she\'s just a girl buying something in the store and you\'re just some guy crossing the road and for all you know she has a guy that watches the road he\'s crossing. glance. why is that car... glance...   '),(8914,3831024,'12,July,2004','2004-07-12 00:00:00','   i wonder. what should i write? who could be reading? would anyone even care? wla lang. pa trip lang. :p   '),(8915,3831024,'03,July,2004','2004-07-03 00:00:00','   spiderman 2. honestly it exceeded my expectations. all the fight scenes, good vs. evil kind of battles were expected, but was unexpected was the internal and external conflicts of a man. a young man. a spider man. he can beat doc ock, but can he pay the bills? he can impress all the ladies, but can he actually have it in himself to risk his girl through the vengeance of others? the scenes bleeds the humanity of a super hero through out the film. especially the elevator scene. y would some one wear something so uncomfortable just to save the world. 2 words to spider man: nappy rash. i love heroes. if there is one main theme in my life, it would be heroes. i just cant get enough of them. every time i have a hero story, i try to emulate the said hero. and every time i fail. my humanity bleeds. maybe i wasnt meant to be a hero. my first post. dont read. its useless. just spreading my wings   '),(8916,3831024,'14,August,2004','2004-08-14 00:00:00','   I find it wholesome to be alone the greater part of the time.  To be in company, even with the best, is soon wearisome and dissipating.  I love to be alone.  I never found the companion that was so companionable as solitude.  We are for the most part more lonely when we go abroad among men than when we stay in our chambers.  A man thinking or working is always alone, let him be where he will.  Solitude is not measured by the miles of space that intervene between a man and his fellows.       - Henry D. Thoreau (1817-1862), Walden (1854)   '),(8917,3831024,'10,August,2004','2004-08-10 00:00:00','   that was what kim said. this is a country with out a soul. there is no love of self. the filipino dream is to be a little brown american. the hottest college and even post-post-grad course here is nursing. so that they could ship out of this god forsaken country to wipe the white mans ass. after all, we\'ve been that for years. was reading tthe papers today. business are closing down. instead of opting to be manufacturers and producers, we opt to become importers and buyers. we spend so much and earn so little. our swine industry is shutting down. prices on pork will rise, people will losed jobs and the country will lose millions in the market. why should filipino\'s care? we rather watch marina and dugong every night anyway. our perspective of beauty is that of the mestiza, the chinita, the half something from somewhere else. and to think these people are the very same rejects of their other halfed origins. we idolize everything that is not us. and what of the hope of the nation? the youth? they are to busy getting out the country, getting laid and having fun. and the money of the country? always with the rich. circulated and recirculated amongst the rich. who cares if the poor and weak die as long as they get their ferraris and private islands. the culture is shot. what culture? the youth look up to the demi gods of media machinery. the false lords of hope. where is the blood spilt for father and country? where is the song of the morning? where is the pride of a country raped and rampaged by the passages of time and men? where is the tongue borne by the children of our ancestors and rejected by the successors? where is our hope? where is our soul?... if you cant find one... make one, i\'d like to say... the question is HOW.   '),(8918,3831024,'10,August,2004','2004-08-10 00:00:00','   tonight i broke down. i just finished diner. i was talking with my mother. she left me at the table to wash my plates. i was alone looking down on the sink washing my plates. i washed them in tears. i stood there just shaking and pressing my hands on the sink. i was crying and i couldnt stop and i didnt know why. i heard my mother come in. i felt her hand on my back. and i felt like a little child again.  The deparment of pediatrics at the nmmc is terrible. they wouldnt care if the clerk died as long as they got their way with the labs, the procedures and the reporting. there are roaches in the ward. they come out at night and crawl out and about on the furniture, on the food and on the patients themselves. one of them managed to crawl up my leg while i was inserting an intravenous line. dextrose in laymans terms. i was pumping the heart of a little boy the other day. it gave up on him. i was his heart. and even i was losing hope. there is no light here. no hope. only pain. and death in the end. The other day we were having our endorsements. the newest resident was receiving a chewing out from the consultant. the resident was on the verge of tears and when she was asked to proceed with her report, the same consultant just whipped her head around to another consultant and chatted in full volume adding much insult to the ludicrous injury already given. worthy to mention is that same well respected consultant is taking up nursing to get out of this country. there are patients here, but there no medicines. there are medicines but there is no money. philippine health care goes not to the needy but to those who can afford. and that is made ever more evident in a run down roach infested government hospital like the NMMC. and i turn off the faucet as i was my plate with tears.   '),(8919,3831024,'10,August,2004','2004-08-10 00:00:00','   I, Gio&nbsp; urlLink     urlLink   '),(8920,3831024,'05,August,2004','2004-08-05 00:00:00','   i wouldnt know. :) hehehe. its called the \"suffering messiah complex\" by some psychiatrists. Most of history\'s great people have in some way gone through it. jose rizal, mahatma gandhi, karl marx.... the list goes on. they say people with the same complex claim they have the key to the salvation of man kind. and so goes the claim of most of our politicians... but i wouldnt know. the only keys i have are the keys to the house, and i lost that a while ago. i wanna go home.... wahhhh   '),(8921,3831024,'02,August,2004','2004-08-02 00:00:00','   do you know what its like? to live life not the way you want to, but the way you have to? to forsake your happiness for the sake of another\'s? to have the power to rip off a person\'s head and break their bones, but you stay your hand and accept all the shit that comes from them? to shake cold, alone and scared that you wouldn\'t know how to live your life the way you wanted, because the life you live has always been the life you have been given? to have the dreams of making lives of other people better but not having the means for it? to have love, friendship and companionship offered and beared in front of you and to shun it, just to give yours to those who have none? to not want to wake up every morning, but you do because you should, or people would die? to bear all the shit in this life, and still have the time to sing and smile? to hold so much darkness inside of you, and squeeze out all the little drops of life just so you wont lose your sanity and hope? to have those very same embers of hope slowly consume you in a inferno of fervor to fight some more? to save lives, at the sake of your own?   '),(8922,3831024,'02,August,2004','2004-08-02 00:00:00','   i walked all the way home from the gate of our subdivision and felt happy. felt happy because i am happy. Not because of some one or something, but because i am. i am inherently happy because on nights like these i feel my soul open up and bear itself with all its beauty and madness. I wish people could see it or feel it. All the positive energy with in me, not marred by my insecurities or fears, my imperfections and frustrations, my darkness and my doom. it is just me. and all the light within me. all the beauty i contain. and i dont mean physical beauty. if you could only see it, you would not wonder what makes me a happy person, and it is not because of my tangible assets that i deem it so. it is deeper, more crucial than that. experience has taught me that my inner beauty can be exploited. some people drink up what they can not have. withdrawal symptoms have been... unpleasant.   '),(8923,3831024,'02,August,2004','2004-08-02 00:00:00','   i met jeffrey today. he smiled at me and keep on playing peek a boo with me as if i wouldnt notice him glance my way. Just to piss him off and to tell him i mind, I caught him in midstare and let him off scampering. People like Jeffrey make my day. They make all the pains worth it. Its like i could bear even more, when i should have given up up way before. and i guess jeffrey feels the same way. his kidneys are shot dead. he\'s pleasantly plump 7 year old smiling face belies the fact that he might not see his 1st decade. and here you see him running around and sticking out his tongue at you just because he can. how sad. i would want to tell you, but i would not know how it would end. but endings for cases like these are always unpleasant. people like him make worth living. if he knew i wonder how he would take it. if he knew, i wonder if he would still be happy. but i thank jeffrey. for giving me hope. when i have none. ignorance is bliss, i wish i didnt know, so i would be happy. but i do. such is the price of knowledge. if ignorance is bliss, doctors must be the most unhappy people of the world   '),(8924,3831024,'02,August,2004','2004-08-02 00:00:00','   there i was absent minded as usual. my thoughts drifting from my recent break up to the fact that my medicine residents wont want to let me graduate, to the fact that i still havent paid my parking ticket, to the fact that by the end of this month i would be fat and flabby again, to the fact that i locked my keys inside the car again, to the fact that i am utterly miserable just because i could be, to the fact that i am not living my life the way i want it to be, to the... then he said: \" i wish i was like you.\" I looked up from the chart and saw yellow eyes. why me? \"You are healthy, and that beats lying in bed all day\" I focused back on the chart. The kid was only 16. He\'s dying. and all i could think of was myself and not wanting to be myself. And here he was wishing he was me. i told him i would pray for him. i didnt have to tell him that i wrote in the chart the number of days he had left to live. i am becoming cold.   '),(8925,3831024,'02,August,2004','2004-08-02 00:00:00','   rather than spending your life watching movies, live your life that it would be a movie... i should rephrase this   '),(8926,3699072,'26,June,2004','2004-06-26 00:00:00','   If I have to be left in this mother fucker  one more time by myself  because one of my co-workers wants a weekend night off, and offers up some bullshit excuse, I am seriously walking out the door.   I\'m here on time every night and hardly ever call out and I just think it\'s ridiculous. '),(8927,3699072,'24,June,2004','2004-06-24 00:00:00','   urlLink  Wheeee! '),(8928,3699072,'23,June,2004','2004-06-23 00:00:00','   Well, well, what do we have  urlLink  here .  Mary-Kate Olsen is being treated for an eating disorder.  As if it wasn\'t obvious.  The girl is 2 pounds and you can see her bones from crying out loud.    Blech.. '),(8929,3699072,'21,June,2004','2004-06-21 00:00:00','   My hair is actually going gray.   Fast .  Both my great-grandmothers on both sides of the family were completely gray by the time they were 25.  Great.  My mother, on the other hand, didn\'t get a single gray hair until her early forties.  Ain\'t that some shit.  I started noticing a gray hair here or there when I was 18.  Over the years it has progressively gotten worse.  I\'m dreading having to buy hair color every month to cover it.  It\'s definitely something genetic because my brother is rapidly going gray and now my sister is starting to find gray hairs as well and she just turned 18.  Grrr.. '),(8930,3699072,'21,June,2004','2004-06-21 00:00:00','   There are just some features I can\'t get installed where I have my journal hosted now.  So hopefully this will be my final move. '),(8931,3699072,'20,July,2004','2004-07-20 00:00:00','   And I don\'t know who\'s responsible for giving&nbsp;me their bug&nbsp;but.....A POX ON  THEIR  HOUSE!!!!  &nbsp;  Brad got his little 5 days in jail he has to serve so I\'m taking him in tomorrow.&nbsp; Fun indeed.  &nbsp; '),(8932,3699072,'11,July,2004','2004-07-11 00:00:00','   The gall of Brad never ceases to amaze me.  He didn\'t work one day last week and was off on the holiday but didn\'t get paid for it so his check was short.  He gave me what he owed me for bills but his cell phone got turned off.  I told him he could take mine  to work with him , since I turned mine off during the day anyway.  He\'s been off for the past couple of days and is never home when I ask him to be.  All I ask is that he be home by 9pm since I have to go over to my parent\'s house to take a shower since our gas is turned off.  He can\'t even do that, plus he takes my phone with him (which I never told him he could do) so that I can\'t even track him down.  Yesterday he took 20 dollars out of my money -- without asking, of course -- and claims he used it to buy food and beer for himself, even though there\'s stuff to eat in the kitchen.  He\'s been over at my brother\'s all day so needless to say, I don\'t believe that\'s how he spent the money.  He then has the nerve to ask me to buy him cigarettes in the morning.  I originally told him no because he has 20 dollars of my money all day long yet \"forgot\" to buy himself a pack of cigarettes.  He couldn\'t do that while he was buying his beer?? Bullshit.  I asked him for receipts from his purchases today and of course, he couldn\'t provide them.  So now in addition to what he owes me for bills next week, he\'s got to pony up and extra 23 dollars as well.  I\'m starting to get pissed because he\'s using and taking advantage of me now.  I just think it\'s really sad when you\'re reduced to having to hide your phone and money from someone you\'d think you could trust. '),(8933,3699072,'02,July,2004','2004-07-02 00:00:00','   Ever have one of these days?     urlLink Saddam Hussein is a funny dude .  I\'m going to go play with my NeoPet now because I have no life. :(  Rejection sucks.  That is all. '),(8934,1503551,'30,June,2004','2004-06-30 00:00:00','   urlLink     Harry Potter and the Half Blood Prince ?!?  '),(8935,1503551,'29,June,2004','2004-06-29 00:00:00','   urlLink    Katie Holmes knows how to make that detective sing!  '),(8936,1503551,'29,June,2004','2004-06-29 00:00:00','   urlLink    Robert Downey, Jnr: An amazing actor, even if the  Singing Detective  film was all it could be!  '),(8937,1503551,'29,June,2004','2004-06-29 00:00:00','   urlLink    Opens June 30th.  Looks kewl.  '),(8938,1503551,'27,June,2004','2004-06-27 00:00:00','   urlLink     A Page of Madness  - Japanese Film, 1926.  A special event at Revelation, Sat. 10th July 2004: 9pm.  '),(8939,1503551,'25,June,2004','2004-06-25 00:00:00','   urlLink    Revelation - Perth International Film Festival - http://www.revelationfilmfest.org/  '),(8940,1503551,'25,June,2004','2004-06-25 00:00:00','   urlLink    Touching the Void ... one hell of a docudrama!  '),(8941,1503551,'24,June,2004','2004-06-24 00:00:00','   urlLink    Joss Whedon\'s first feature film is blogging!  '),(8942,1503551,'20,June,2004','2004-06-20 00:00:00','   urlLink    the fabulous duo, new batchannel, new battime!  '),(8943,1503551,'20,June,2004','2004-06-20 00:00:00','   urlLink    a new look, a new team  '),(8944,1503551,'20,June,2004','2004-06-20 00:00:00','   urlLink    An oldie, but a goodie.  '),(8945,1503551,'20,June,2004','2004-06-20 00:00:00','   urlLink    Drinksafe warning: you\'ll be burping for days.  '),(8946,1503551,'20,June,2004','2004-06-20 00:00:00','   urlLink    For Australians, releases July 29th.  (In the US June 25th.)  '),(8947,1503551,'18,June,2004','2004-06-18 00:00:00','   urlLink    http://www.fahrenheit911.com/  '),(8948,1503551,'18,June,2004','2004-06-18 00:00:00','   urlLink    So animated ...  '),(8949,1503551,'18,June,2004','2004-06-18 00:00:00','   urlLink    So fierce ...  '),(8950,1503551,'18,June,2004','2004-06-18 00:00:00','   urlLink    So cute...  '),(8951,1503551,'18,June,2004','2004-06-18 00:00:00','   urlLink    Yeti Sports 5 ...  '),(8952,1503551,'17,June,2004','2004-06-17 00:00:00','   urlLink    Has Keira been watching too much Braveheart??  '),(8953,1503551,'16,June,2004','2004-06-16 00:00:00','   urlLink    0.9 is so slick, so pretty, so STABLE! :)  '),(8954,1503551,'14,June,2004','2004-06-14 00:00:00','   urlLink    When photoshop goes wrong!  '),(8955,1503551,'14,June,2004','2004-06-14 00:00:00','   urlLink    Hermione ... bad boys always get punished!  '),(8956,1503551,'12,June,2004','2004-06-12 00:00:00','   urlLink    Firefly on DVD in Australia ... finally.  '),(8957,1503551,'10,June,2004','2004-06-10 00:00:00','   urlLink    Laser  '),(8958,1503551,'07,June,2004','2004-06-07 00:00:00','   urlLink    Best 404 ever.  '),(8959,1503551,'06,June,2004','2004-06-06 00:00:00','   I\'ve been playing with the new  urlLink Hello PhotoBlogging Tool  which allows those of us using Blogger to post  and host  pictures (albeit within in fairly specific size parameters).  Here\'s a quick test, capturing the storm-tossed Sunday I can see from my back window:  urlLink    Although  urlLink Blogger  still doesn\'t officially host files other than the text and templates of blogs, this new service has no obvious limits in terms of photos, so it\'s certainly a step toward Blogger freely hosting more than just text!  '),(8960,1503551,'06,June,2004','2004-06-06 00:00:00','   urlLink    28 Days Later  '),(8961,1503551,'06,June,2004','2004-06-06 00:00:00','   urlLink    Nietzsche: Sip on these ideas!  '),(8962,1503551,'06,June,2004','2004-06-06 00:00:00','   urlLink    The cover to LIMINA volume 10.  '),(8963,1503551,'31,July,2004','2004-07-31 00:00:00','   urlLink    dannybot?!?  '),(8964,1503551,'28,July,2004','2004-07-28 00:00:00','   urlLink    Will makes a new friend.  '),(8965,1503551,'24,July,2004','2004-07-24 00:00:00','   urlLink    more than meets the eye  '),(8966,1503551,'22,July,2004','2004-07-22 00:00:00','   urlLink    William Gibson, bridging the future, living the interstitial.  Chronicle Photo by Sam Deaner.  '),(8967,1503551,'20,July,2004','2004-07-20 00:00:00','   urlLink    cray like a rupert?  '),(8968,1503551,'17,July,2004','2004-07-17 00:00:00','   urlLink    it\'s here!  '),(8969,1503551,'17,July,2004','2004-07-17 00:00:00','   urlLink    ... with no Camelot, no affair, no Holy Grail, but with lots and lots of nasty Saxons! :P  '),(8970,1503551,'14,July,2004','2004-07-14 00:00:00','   urlLink    The Invisible Woman ... perhaps Sarah Michelle Gellar?  '),(8971,1503551,'11,July,2004','2004-07-11 00:00:00','   urlLink    Think about it.  '),(8972,1503551,'09,July,2004','2004-07-09 00:00:00','   urlLink     A Page of Madess , so far ahead of its time! ( urlLink Details )  '),(8973,1503551,'09,July,2004','2004-07-09 00:00:00','   urlLink    The world wide web told me so:  urlLink http://www.johnhowardlies.com/ .  '),(8974,1503551,'07,July,2004','2004-07-07 00:00:00','   urlLink    Movie posters are always edited, but this one seems to be  urlLink raising a few eyebrows .  '),(8975,1503551,'07,July,2004','2004-07-07 00:00:00','   urlLink    Topical: http://www.theaustralian.news.com.au/common/story_page/0,5744,10042844%255E12332,00.html  '),(8976,1503551,'05,July,2004','2004-07-05 00:00:00','   urlLink    Sometimes, I wonder how serious things can be on the set of a moiton picture ...  the  FireFly  filming looks fun!  '),(8977,1503551,'05,July,2004','2004-07-05 00:00:00','   urlLink    Jennifer Garner\'s new look for the Elektra film (spunoff from the very average DareDevil feature) looks very promising.  More like the look from the Frank Miller comics...  '),(8978,1503551,'03,July,2004','2004-07-03 00:00:00','   urlLink    Spider-Man 2 - cool, complex and compelling!  '),(8979,1503551,'05,August,2004','2004-08-05 00:00:00','   urlLink     urlLink Guess who\'s back?  '),(8980,1503551,'03,August,2004','2004-08-03 00:00:00','   urlLink    Who said political satire couldn\'t have big explosions?  '),(8981,2246318,'26,November,2003','2003-11-26 00:00:00','   So it\'s almost Thanksgiving again.  I really like this time of year.  Most people are in a better mood, there is a crispness to the air, and most of all it means that turkey is in the oven.  This year I decided that I was going to make the turkey.  Last year I wanted to but decided to let my Mom do the honors.  I am a big fan of Good Eats(as some of you may already know) so I am going to brine my turkey like Alton does in the classic episode \" urlLink Romancing the Bird (A Good Eats Thanksgiving) \"  Hopefully I will do it right and everything will taste as good as it looks in the show.  I will let you know how things go.  Also, I saw a commercial for the Cross Roads Farm and Craft Market and saw that they have a website.  I tried to access it but it said it wasn\'t there.  I don\'t know if they are working on it or if was just down when I tried, but here is the address anyways, maybe it will be back up soon.   urlLink www.thecrossroadsmarket.com   Have a very happy Thanksgiving all! DAN  '),(8982,2246318,'24,November,2003','2003-11-24 00:00:00','   Hello again.  A friend of mine brought me to a Farm and Craft Market yesterday.  I wasn\'t looking to find too much that interested me, but alas, I did.  The first place that I found something interesting was \"Spice Harbor.\"  Now my friend said that I would like it there, and she was right.  They had a lot of different kinds of fresh spices that you just can\'t get around town here.  Most of the spices were pretty cheap also, 1/4 ounce of Spanish saffron was only $8.  That\'s not too shabby.  Then while looking at some Emu Oil products, I get tackled by the owner to try some Ostrich meat \"beef\" sticks.  I was a little leary, but I tried it, and it was actually really good.  No, it didn\'t taste like chicken.  Then he had my try the summer sausage, and it was really good too.  Too bad I didn\'t have any money with me or I would have bought some.  My friend did buy some hand lotion that she tried there and said it worked great.  The final thing that I found that interested me was Ice Wine.  I\'ve only heard of ice wine once before on the  urlLink Food Network  on the show \" urlLink The Best Of .\"  I knew this wine was very unique and I just had to try a sample.  I went up to the guy at the stand, Michael Burkland,  and asked if I could try a sample.  Now, I am 25 years old and I am always told that I look older than I am.  He looked at me oddly when I asked and pulled the bottle out of the cooler.  He poured a little bit into a plastic cup and asked me how old I was.  I was a little shocked by this and I almost blanked on my age, but I told him that I was 25.  He made a little joke, then gave me the sample.  I am not a big fan of wine, I just don\'t care for the taste or the mouth feel of it, but this wine was VERY good.  Michael went on to tell me that it is a desert wine and the way it is processed makes it very very sweet.  It was the clear cut best wine I\'ve ever had.  It was very sweet, the taste was intense, the after taste was actually a good one.  This small bottle that he had at the stand, which resembled a extra virgin olive oil bottle, was in the price range of $24 US.  I\'m not one to spend much on alcohol but I thought that I have to get some of that.  I asked him where I can purchase it other than the show there and he told me that it was at all the local liquor stores in my town and handed me his card.  I am going over to my sisters for Christmas morning and I hope to bring a bottle of this outstanding wine with me.  You can go to their website at  urlLink www.merrittestatewinery.com   I had a good time on that little trip, it was very nice to get out of the house for once.  Until next time, DAN  '),(8983,2246318,'21,November,2003','2003-11-21 00:00:00','   Ok, so I haven\'t really been into reality shows for quite a while now.  I, admitting, used to watch Survivor, Who wants to be a Millionaire?, and the Mole.  But almost everything since has been boring.  One that really did interest me when I saw the commercials was \"Average Joe.\"  Why?  Because the concept is great.  Take this beautiful girl who has been in beauty shows and was a pro cheerleader and give her a group of average guys, not models, to chose from.  She said before the show started that looks doesn\'t matter, and almost all women say the same thing.  This show set out to either prove them right or wrong.  Now, I consider myself to be not all that great looking so I was intrigued with this premise.  This past week she cut the number of guys from six down to three, then the big surprise, which I have to say I was kind of expecting.  They brought in three pretty boys.  One, from the beginning of the show who welcomed her to the show and then left, and two others.  She seemed pretty happy to see these guys come in and hear that they are now part of the show and the three \"average joe\" guy\'s competition.  Now, this adds a very interesting twist.  She has shown that she isn\'t so shallow that she won\'t give these \"average joe\" guys a chance, but is she shallow enough to dump them for the new pretty boys for their looks.  She hasn\'t spent nearly as much time with these new guys and doesn\'t know them as well as she knows the other three.  I am rooting for Adam, he is the most \"Average\" of the three and I hope that she doesn\'t fall for the other\'s good looks only.  But, from past experience, looks matter!  I think she will go with one of the pretty boys.  But we\'ll see I guess.  \"Average Joe\" is on NBC on Mondays, check your local listings for times.  You can also learn more by going to the \"Average Joe\" website:  urlLink http://www.nbc.com/Average_Joe/   Ok, time for bed for me.  I\'ll see you soon! DAN  '),(8984,2246318,'16,November,2003','2003-11-16 00:00:00','   So I put the carpet in my room yesterday.  My room looks good now.  I even found a place for my Dave Matthews Band subway poster that Amber got me.  It\'s almost 6 feet wide!  It looks great.  I am now moving my stuff into my room, which looks like a bedroom finally.  Today I am going up there after the Bills game to tack the carpet down.  For anyone who would like to e-mail me about this blog you can.  When e-mailing me, please put the subject as \"blog.\"   urlLink viperwd@hotmail.com   Until next time... DAN  '),(8985,2246318,'15,November,2003','2003-11-15 00:00:00','   Ok, in case any of you read my post from early this morning.  She called and apologized.  So I am taking that post down.  For those of you that didn\'t, my best friend and I had a fight last night and I said that I didn\'t want to be friends with her any more.  But she called this morning and said that she was sorry for what she did, and I said that I was sorry for what I did.  So now things are better, we are at least still friends.  Ok, I have to go and help put carpet in my room, after that my room will be ready to move into.  :)  DAN  '),(8986,2246318,'08,November,2003','2003-11-08 00:00:00','   So, my Mom just bought a new house and I am helping work on it.  I\'ve been working on my room first so I can get it all ready.  (Yes, I still live with my Mom, but hopefully not for too much longer.)  A couple of days ago I got the ceiling painted, then I didn\'t the walls.  Today I got all the trim done, and all that\'s left is where the ceiling and walls meet.  I ran out of gray paint to finish.  Next I am going to do the closet probably tomorrow.  After that is done I can put in the carpet that we got the other day.  It is also gray, it should look good in the room and match pretty good.  The colors are skyline blue for the ceiling, blue wale grey for the walls and cashmere white for the trim and closet.  All the wiring in the house has to be redone.  I am getting two new outlets and they are going to replace the two that are already in there.  Once that is done, I will be able to start moving my stuff in.  My room is probably three times the size it is now which will be nice.    I just thought I\'d let all you who know me know what I\'ve been up to the last week or so.  Laters. DAN  '),(8987,2246318,'07,November,2003','2003-11-07 00:00:00','   Hello again, glad you made it!  Today, for those of you that don\'t know, I\'m going to let you in on my favorite Band.   urlLink Dave Matthews Band   I\'ve been a really big fan since their first studio album came out in 1994.  Many of their songs have a lot of meaning to me.  I have been to 4 concerts.  2 at Ralph Wilson Stadium in Buffalo, I was in the 10th row for each.  1 at the Air Canada Centre in Toronto Canada, I was in the 12th row there.  And 1 at Six Flags Darien Lake in Buffalo, I was only in the grass(I bought the tickets late.)  If you have never heard of them, check them out!  For those of you who are already fans, Dave Matthews is going to be on Saturday Night Live this Saturday, Nov.  8, 2003 at 11:30pm EST to promote his solo album.  '),(8988,2246318,'05,November,2003','2003-11-05 00:00:00','   Ok, so I made a mistake.  I sent everyone to a bible website instead of here.  hehe, oops!  Not that I have anything against those types of websites or anything, just not what I\'m trying to do here.  But you made it now and I guess I will give you another very interesting entry. :P  Today I am going to let you in on my favorite cooking TV show,  urlLink Good Eats .  I like to cook.  I have some experience cooking in the restaurant business at Red Lobster.  But I really like to cook at home.  It can be fun.  But Good Eats makes it even more fun.   urlLink Alton Brown , the host, mixes pop culture with cooking to make it interesting to watch.  Even if you don\'t like cooking shows, you should check it out.  It\'s on the  urlLink Food Network  every weekday at 11:30pm or Wednesdays at 9pm.  Not only does Alton give you recipes but he tells you why things do what they do in those recipes so you can change them up any way you want.  It lets you experiment and make up your own recipes.  It\'s actually fun to watch, and gives me lots of ideas for cooking.  Check it out sometime.  That\'s all for now, TTFN DAN  '),(8989,2246318,'05,November,2003','2003-11-05 00:00:00','   Hello everyone, and welcome to my new Blog.  What is a blog you ask?  I asked myself that same question earlier today when I was surfing the net, but more on that later.  This blog will be a mish-mash of all kinds of different things.  One day it might be a journal of the events in my day.  Another day it might be my rants and/or raves of a website that I found.  And another day it might be my thoughts about an article in Scientific American that I read.  Who knows what it will be, I\'ll keep you guessing.  I hope some of you find my thoughts interesting, and if you do, I hope you will come back.  For everyone else, thank you for at least giving me a chance and checking this out.  This is a new thing for me and hopefully will evolve into something very fun and informative for everyone(including myself.)  If, by some rare chance, you find this interesting enough to link to or mention, let me know, I want to see :).  Ok, now on how I got here.  I got an e-mail newsletter from a new website that I found last week,   urlLink www.howstuffworks.com , and I saw something interesting.  \"How Blogs Work\"  I think to myself, \"What in the world is a blog?\"  So I printed out the article and read it while watching Conan.  Now I\'m not going to try to explain what blogs are since I barely know myself, so I will send you to the article that I found so you can find out for yourself.  Maybe you\'ll even make your own, if you haven\'t already.   urlLink How Blogs Work   Now I am a very big computer \"freak.\"  I pride myself in knowing quite a bit about computers.  There isn\'t a lot of things on the internet or about computers that I haven\'t at least heard of.  So when I heard about blogs today I was taken aback about how big they are and how I haven\'t even heard of them yet.  Well, this was my first blog, I hope you like it and I hope you all come back.  DAN  '),(8990,2246318,'31,December,2003','2003-12-31 00:00:00','   Hello, this is my last post of 2003.  It\'s been an interesting year.  I hope next year will be better.  I won\'t bore you will a look back so lets just look to the future.  I start my new job on Monday and I am in a new house, as close as I\'m going to get to starting over for a while.  On a side note.  My friend Bamboo2 got laid off yesterday at work.  They did it for very stupid reasons and they lost a good worker.  I would like to wish her good luck on finding a new and even better job!  Until next year, DAN  '),(8991,2246318,'29,December,2003','2003-12-29 00:00:00','   One week to go and I go back to work and my three month vacation will be over.  I am glad actually, I have been pretty bored when not working on the new house.  It will be nice to be back to work and be able to talk to people again.  I miss that interaction a lot.  The money is going to be nice, that is for sure.  Maybe I\'ll even meet a girl that I like there.  That would be really good too.  Without giving too much info out because I don\'t know how much they want anyone to know, I am going to tell you a little about bamboo2 who found my typo.  She is a good friend of mine who I used to work with.  When I first met her I thought she was really annoying but once I got to know her I liked her more.  She goes out with all of us to Applebees on Friday nights and she has a boyfriend in England.  Just thought I\'d let you know so you don\'t sit there wondering who she is.  I don\'t really want to write about my Christmas because it\'s over with.  But I will fill you in on my Thanksgiving turkey that I never told you about, but promised.  As I said before I brined my turkey like Alton brown does on his show Good Eats.  It turned out great!  The skin was a dark golden brown and very crispy and the inside was perfectly done and very very juicy.  For my first time cooking turkey it turned out better than I expected.  I will brine turkey again if I get the chance.  That\'s my story and I\'m sticking to it, DAN  '),(8992,2246318,'27,December,2003','2003-12-27 00:00:00','   Hello,  I can\'t talk long but I said that I would post once I move in and I am moved in.  We have been sleeping here since this past Saturday.  I just, although, set up my Mom\'s computer to check e-mail and look a few things up on Christmas eve, so I had a lot of e-mail to go through.  Christmas wasn\'t too bad, it could have been worse but it could have been better.  I wish I had money to get my friends and family stuff for Christmas but I didn\'t so I felt really bad.  I start my new job in just over a week so I will be able to catch up on bills.  I gotta go, I\'m really busy lately and I have to sit on the floor to use the computer right now so it\'s uncomfortable.  I hope you all had a great Christmas.  Talk to you soon!  DAN  PS  Thanks to  urlLink Bamboo2  for finding a spelling error in my last posting.  '),(8993,2246318,'17,December,2003','2003-12-17 00:00:00','   I finally got a job!  I will be working at  urlLink PowerNet Global  through Express personal starting January 5, 2004.  Like Shannon at Express said, it\'s a good way to start the new year!  I will be a customer service Rep.  PowerNet Global offers Internet, long distance phone service and data services.  It should be a fun job in a field closer to my career choice.  At the new house we got our new refrigerator today.  I love it.  I\'m a gadget guy and that includes big gadgets like new refrigerators.  It is very nice and has crushed ice, one of those things that I just love.  I will be sleeping there either Friday or Saturday.  The phone will be hooked up Friday and the cable on Saturday.  It\'s starting to look like a home finally, and my room is feeling more and more comfortable.  Laters, DAN  '),(8994,2246318,'16,December,2003','2003-12-16 00:00:00','   Well, they finally caught Saddam Hussein.  No matter how you felt about the war here in the US, you have to agree, as I do, that this is a good thing.  I was against the war from the beginning, even though I knew Saddam had to go.  I was glad to get on the Internet yesterday morning to see that they caught him.  Now we wait to see the trial and see what punishment he gets.  The Buffalo Bills lost again yesterday, barely.  They are not out of the playoffs for sure.  Now we just hope to have the best defense in football.  One more early morning.  The electrician should be done tomorrow.  He finished my room so I can set up my bed and my computer and all my other stuff.  My room is starting to look like a room finally.  I can\'t wait to get out of an apartment, the people upstairs are really really loud.  They stomp around like they are squashing grapes or something.  Ok, off to bed. DAN  '),(8995,2246318,'13,December,2003','2003-12-13 00:00:00','   Hello again, it\'s been a while since I\'ve posted.  I hope everyone is fairing will with all the snow we are getting here.  So, Melana picked Jason.  I\'m not really sure why I\'m surprised with that.  I guess I was just hoping(for my own sake) that she wouldn\'t pick the pretty boy just because he\'s pretty, but that is exactly what she did.  It just proves that even women that say looks don\'t matter are just saying that.  Looks do matter.  Of course I knew that, but this just proves it to those that said that they don\'t.  Adam was the better guy all around.  He had money, he had friends that loved him, he was on his own and secure, and he knew how to treat her right.  Jason?  Well, lets see, he knew how to kiss and how too look pretty.  Now people always tell me not to worry, looks don\'t really matter to women, they look for other things like sense of humor.  Well, I guess there goes my hopes.  I guess there is going to be a Average Joe 2 based in Hawaii but I don\'t think I\'m going to watch it, I already know what is going to happen.  They showed some previews of it and the girl freaked out, she was pissed that they did this to her.  Oh well, such is the vain life.  If you want to chat about this, e-mail me at  urlLink viperwd@hotmail.com  with the subject of \"blog.\"  Now on to other things.  The house is almost set for us to move in.  The electrician is almost done with all the new lines and service.  We have carpet in the living room now.  We also moved a lot of the living room furniture.  My room, well I moved everything except my mattress and most of my clothes.  I guess the cable and phone will be moved next Saturday, that is when we are going to be officially moved in.  It will be nice to be done moving, I really hate moving.  Well, I still have no job.  I have gone to a few interviews with no call backs as of yet.  Red Lobster is also delaying my unemployment so I have no money coming in.  I might be losing my car, and with no car I can\'t get a job very far from home.  What a good time for all this.  I can\'t get anyone anything for Christmas this year.  I hate it, I feel so bad that I can\'t get anyone anything.  Maybe something good will come out of this.  I guess time will only tell.  Ok, I hope to post one more time before I move and then as soon as I get my computer up and running once I move.  So have a good week all!  DAN  '),(8996,2246318,'02,December,2003','2003-12-02 00:00:00','   Ok, time to do a couple of updates on things I\'ve posted in the recent past.  First, moving into my new home.  I moved my desk the other day.  It is a very large computer desk.  I actually pulled the bottom and heaviest part up the stairs by myself.  It wasn\'t too bad, I thought we would have more problems than we did.  I got it all set up and started to set up the speakers and other things for my computer.  I also moved my bookshelf and put my CD\'s, and DVD\'s in it.  And lastly, I moved my night stand over.  My room is coming together nicely.  The guy to do the electricity is coming at the end of this week and the guy to do the carpet in the living room is coming on the 10th, so after all that\'s done I can move in and live there.  Although I am going to wait until we get phone service and cable and move things like the stove in.  Next, I watched the new episode of \"Average Joe\" tonight.  They put Melana, I think that\'s how you spell her name, in fat makeup and a fat suit, just like they would for a movie, and had her ask the guys questions.  Then she watched by camera how the guys reacted to her and Zac finally came through as the jerk he is.  To keep it short, she voted Zac and one of the new guys, I think he name was Mike, off.  Zac was not too happy about it because he thought he was the greatest.  So, to keep track, she has not taken two of the thee pretty boys off the show.  So it\'s down to Adam and the one last pretty boy standing.  I am rooting for Adam, he seems like a nicer guy and he has a similar sense of humor that I have.  And lastly, new news, I signed up for college and have been filling out financial aid.  I guess it\'s almost final, I am going back in January.  This is a good thing, I hope to get a degree in Information Technology(IT) and get a good job, finally.  Ok, until next time America, DAN  '),(8997,2246318,'27,January,2004','2004-01-27 00:00:00','   HI!  You can just ignore that last post and the link to the site feed on the right.  Just for now.  It\'s something new that I\'m trying to figure out.  I\'ll let you know when I do :)  Sorry it\'s been so long since my last real post I\'ve been very busy with work and all.  Speaking of work, things are going really good.  This past Friday was my first day training in my new Dept.  I also trained yesterday.  Then today my trainer and another girl in the dept. said that they thought I was ready to be on my own today.  I did pretty good.  Not too many things I had questions on and I moved right along on my report.  I only got a few incoming calls so that helped too.  I think after this week I should be all set.  I like things so far.  I would like to say thank you to Bamboo2 for her very thoughtful gift on Friday.  She saw that I had Evanescence\'s CD on my wishlist so she got it for me.  Then she had a very nice note on the inside.  Now, this isn\'t why I put up my wishlist, but thank you very much!  It was very nice of you.  Ok, TTFN DAN  '),(8998,2246318,'24,January,2004','2004-01-24 00:00:00','   Hello, this is just a test post to test the site feed.  Thanks!  DAN  '),(8999,2246318,'18,January,2004','2004-01-18 00:00:00','   Well, week 2 is over and week 3 is about to begin at my new job.  So far things are still going pretty good.  Another one of the trainees has left.  She quit then came back and is now going to be in Order Entry instead of Customer Service.  There is a lot of stuff to learn in a very short period of time.  So now we are down to two.  They wanted four.  I don\'t know if they are going to try to hire more after they are done training us or not, I guess we will see.  Also I\'d just like to say how much I like the show Alias.  It is one of the best shows on TV.  I don\'t have time right now, but I will soon.  I just thought I\'d throw it in right after I finished watching this weeks episode.  \"23, so tired of life, such a shame to throw it all away\"  DAN  '),(9000,2246318,'14,January,2004','2004-01-14 00:00:00','   Hello again,  Well, my job is going really good.  I\'ve gotten 90%+ on all my tests.  The trainer told me today that she is really happy with how I\'m catching on and taking in all the info.  Out of the three of us training, I\'m the only one who hasn\'t had to retake a test.  The trainer talked to me about moving to the Pic/Freeze/Move section which would give me more to do and the person I was shadowing today said they were thinking about hiring one of us for over there.  It also means more money once they hire me from Express, the temp agency I\'m with.  So hopefully things keep going well there.  (knock on wood)  The power went out tonight around 8pm.  It was 14 degrees F out, so we got a little worried that if it stayed out for long then we would have to go elsewhere because out furnace runs half on electricity.  But it was only out until 8:45pm so we were ok.  We only lost 2 degrees inside.  Not too bad.  This week has gone by very fast so far.  Oh one last thing.  I just thought I\'d send you to my favorite Blog(other than mine that is).  It\'s, big surprise here,  urlLink Alton Brown\'s Blog .  He recently said that he was going to try to update it at least once a week, which is a good thing.  I like to hear what he has to say about the food industry, his show, and many other things.  Give it a look sometime.  Can I get a What, What!?! DAN  '),(9001,2246318,'11,January,2004','2004-01-11 00:00:00','   Now, I know that most of you reading this don\'t care about this, but I just thought I\'d post it telling you that I have put up my wish list online.  Now my Mom can stop begging me for it.   urlLink My Wish List   Peace. DAN  UPDATE: It took me a little while but I finally made my list the same style as this page.  Now it looks a little more seamless. DAN  '),(9002,2246318,'10,January,2004','2004-01-10 00:00:00','   Wow, it has been a long week.  There is a lot to learn at my new job.  I was planning to post after my first day on Monday but I was pretty tired then I had to study for a test that I was going to have on Wed.  Yes, that\'s right I had a test, and then I had another today too.  I have two weeks of training which includes classes and a 400+ page book to learn.  Plus, we have to take tests throughout and you have to get an 85% to pass.  I think once I get through the training period, I will like the job.  It is a lot to take in all at once but once I get it down it will come like second knowledge.  I get the whole weekend off this week and next so that\'s pretty good.  Get heavy on with digging your ditch. DAN  '),(9003,2246318,'05,January,2004','2004-01-05 00:00:00','   Well, I start my new job tomorrow.  It should be fun, I am glad to get back into the workplace.  My 3+ month vacation is finally over.  Now I can finally start paying my bills again and maybe even buy something at Applebees soon.  I should get to bed, I want to get a lot of sleep so I\'m not tired at work tomorrow.  I\'ll fill you in on my first day tomorrow.  Work like you don\'t have to, love like you\'ve never been hurt, and sing like your in the shower.  DAN  '),(9004,2246318,'01,January,2004','2004-01-01 00:00:00','   HAPPY NEW YEAR!  Welcome to 2004!  DAN  '),(9005,2246318,'28,February,2004','2004-02-28 00:00:00','   WOW!  It\'s been a long time since I last posted.  I\'m ashamed of myself for that.  I\'ve been very busy with work and all.  Which, by the way, things are going very well.  This past Tuesday our Dept. moved to a new area of the office.  We now have a nice big window view and better air flow.  We aren\'t stuck in the corner any more.  We also have the new computers, new phones, new everything.  We are special ;P  I can\'t really complain about my job, things are going well, now I just hope they officially hire me once my 90 days are up.  Yesterday When I went to my weekly Friday night out I saw two of my good friends that I haven\'t seen in a long time.  The first I haven\'t seen or talked to for in upwards of over a year.  I worked with her at Wal-Mart in the layaway dept.    She helped train me and we became good friends.  We had a great time working together.  Then I moved to the Foods dept and we didn\'t get to work together much, and then I became CSM and I almost never got to talk to her.  Then, the killer, I left Wal-Mart for Red Lobster.  Well, she went out with everyone Yesterday so I got to chat with her and it was like it was back at Wal-Mart.  We were back to joking around and having fun.  The second I haven\'t really gotten to talk to since I left Red Lobster in Sept.  We had a good chat and plan on getting together Sunday to chat some more.  He\'s a good guy who if you show some respect, he will show you 10 times the respect back.  So I am very happy that I went out yesterday, I miss both of them, and I miss working with them.  I am glad we can get together once in a while though.  Ok, one last thing for tonight.  I have added two new features to my blog.  If you look at the end of every post you will see two new links.  The first, Comments, you can click on to make comments or read comments about that post.  If you want to make any comments about something I post that others can read then you can do it there.  The other is Linkback.  If you have a blog yourself or a website and you write a reference to my blog or something that relates to my blog or something I wrote about in my blog you can put it there.  Also if you see a webpage that you think is pertinent to something that I wrote about, even if it\'s not yours, you can put it there also.  If you want to know about these then you can go here:  urlLink www.haloscan.com  I would like to thank Haloscan for these free features which I hope will enhance my blog experience.  Ok, I\'ve blabbed enough. DAN  '),(9006,2246318,'05,February,2004','2004-02-05 00:00:00','   So it\'s been a long time since I last posted.  Sorry about that.  I have a lot to talk about but I just don\'t have the time right now.  Work keeps me going to bed early and when I get up in the morning I don\'t think about updating my Blog.  I will try my hardest to post more often though.  First off I want to say to all those who are upset about what happened during the half time show at the Super Bowl.  GET A LIFE, GET OVER YOURSELF!  Number one, the whole outfit wasn\'t supposed to come off, it was an accident.  Number two she had a pasty on her nipple so you couldn\'t see anything anyways.  Number three a couple of years ago Lil\' Kim went on TV for an entire awards show with less on than that.  Then someone grabbed her.  I would think that is worse.  CBS cut away as fast as they could when they saw what happened.  When I saw it they cut away so fast that I couldn\'t tell if I saw what I thought I saw.   Plus you could tell that Janet was embarrassed she crouched over and covered herself up as quickly as she could.  Now stop talking about it, stop trying to sue anyone about it, and move on to something important.  Ed, on of my favorite TV shows is ending soon.  I found out last week that there are only two more episodes left.  What happened?  They moved it to Fridays to cancel it?  It doesn\'t make any sense.  I am not happy about this!  I bet you, at NBC,  $10 to bring Ed back.  (if you watched the show you will get that)  On to another of my favorite shows, Alias.  I got the season 2 complete season on DVD set the week before last.  I just got done watching it this past weekend.  It is just the best TV show on, well, TV.  It\'s like a top of the line movie every week.  Just as long as they dumb it down anymore than they have then I will never purposely miss an episode.  Every week I look forward to Sunday to watch the best night for Drama.  First on is 10-8, pretty good.  A good way to start the night.  Then, of course, Alias.  I can\'t say enough about this show.  And finally, The Practice.  Although they cut the budget of the show for this season and lost many of the stars of the show, they kept enough of the stars and didn\'t cut on the writing which makes the show.  It is the best 3 hours on TV bar none.  hehe get it, bar, hehe like in lawyers.....ahhh nevermind :P  Ok, you\'ve heard enough from me tonight, \"just \'cause things around you seem heavy, doesn\'t mean you can give up this ground.\" --Dave Matthews  DAN  '),(9007,2246318,'28,March,2004','2004-03-28 00:00:00','   Alias was great this week.  Wasn\'t too dumbed down.  I couldn\'t believe that the blonde chick\'s Mom killed the Dad.  That was a surprise.  Then, for the first two seasons they built up Sloan so we hated him a lot.  Now they are building it up so we feel bad for him.  It was  a good episode.  The last scene hit a little close to home for me.  A slightly different situation but the same feelings.  Alias is one of the only shows ever that really brings out emotions in me.  I truly feel bad for characters when bad things happen and vise versa.  It has the best of many worlds.  Every episode is like a mini movie, it is a drama and brings out emotions, it has comedy, it has action and suspense, and a hot leading lady.  What could be better?  Another great show, the one I\'m watching now, is The Practice.  It has some great writing and even though they got rid of a lot last year, it still has some great actors.  Only 4 more new shows until it ends.  It\'s too bad ABC had to cut it\'s budget so much.  I guess they are going to do a spin off of some sort.  David E. Kelly makes some great shows and if he does the spin off then I\'m sure it will be good, and I will definitely watch it.  Ok, one last thing before I go.  I would just like to wish Bamboo2 a happy belated 22nd birthday.  It was this past Wed.  That\'s it for now, but please remember, \"Tomorrow is no place to place your better days\"  -- Dave Matthews  DAN  '),(9008,2246318,'30,April,2004','2004-04-30 00:00:00','   So, I\'m at work right now.  I\'m on my lunch and I\'d thought I would post that I passed my test!!  That means I will be hired in the coming weeks.  No more worrying if I will have a job next week or not.  I will probably get a raise once I get hired so that will be nice also.  I thought I passed, I just wasn\'t 100% sure.  Ok, I\'m going to go read and relax, it is hot in here right now.  Werd Up DAN  '),(9009,2246318,'28,April,2004','2004-04-28 00:00:00','   Ok, there\'s a lot to talk about today.  BTW, Hi, how are you today.  I hope I find you feeling healthy.  First off, Alias was great this week!  I was on the edge of my seat until I fell off at the end when it was a real cliffhanger just like at during the first and second season.  I have to admit I thought Sloan had reformed.  Boy was I wrong.  I can\'t wait until this weeks show.  Second,  I took the first part of my 90 days test on Monday.  It was all on the computer, mostly multiple choice with a few fill in the answer and one essay.  Then I took the second and third part yesterday.  The second part was all essay.  The third part showed some screen shots and I had to fill things in.  My boss came up to me today and said that I passed the first part and they have yet to score the second and third part.  I\'m 97% sure I passed with flying colors so I will be hired soon.  That will be nice, to have some feeling of job security.  I will probably know for sure sometime this week.  Lastly I signed up to beta test Gmail.  It gives you 1GB of storage space for e-mail so you never have to delete e-mails if you don\'t want to.  It should be interesting to see if it makes it past beta.  Google is making a lot of noise with this service.  Ok, that\'s enough from me. Under the Table and Dreaming, DAN  '),(9010,2246318,'19,April,2004','2004-04-19 00:00:00','   Per corrections from Bamboo2.  Fixed spelling and grammar mistakes on yesterdays posting.  DAN  '),(9011,2246318,'18,April,2004','2004-04-18 00:00:00','   Wow, Alias was great tonight!  It had Jennifer kicking ass with two guns.  They tricked me into thinking that they killed off Sloan.  It had, what I thought, was the beginning of the end of the blonde chick.  This show was the best all season.  There were a couple of weak episodes this season, but I knew J.J. couldn\'t do that all season.  The twists keep coming, now Sydney has a sister fathered by Sloan.  Only three shows left this season.  I hope the show gets picked up for next season.  It was picked up by TNT for syndication and that includes up to 88 shows, so next season is almost in the bag.  And if a show goes 4 or more seasons then they usually can go even further.  The Practice, I\'m saddened that the show will end this season.  They are planning a spin off and if David E. Kelly is involved then it should be good as well.  We\'ll have to see.  \"I\'ll lean on you and you lean on me and we\'ll be okay.\" -- Dave Matthews  DAN  '),(9012,2246318,'11,April,2004','2004-04-11 00:00:00','   I just wanted to wish everyone a very Happy Easter!  DAN  '),(9013,2246318,'09,April,2004','2004-04-09 00:00:00','   So lets forget that I ever said that I would post every score for the Mets.  I just can\'t seem to do that.  I will post important games etc. just not every one.  I went to Erie today.  I got some nice clothes for work, cheap at Target.  I love that store, it\'s like Wal-Mart, but better.  I hope we get one in our area.  I got a coupon for Bed, Bath and Beyond this week in the mail for 20% off so I had to go there.  I almost got a 1 1/2 quart All-Clad sauce pan but then I noticed that the coupon said that it excluded All-Clad so I then went over to the Calphalon section and I found some nice pieces in the tri-ply area.  It has a layer of Aluminum sandwiched between two layers of Stainless Steel all the way throughout.  I ended up getting a 10\" Stir Fry Pan.  The regular price was $40, but I got it for only $32.  A very good deal.  If I had gotten the same pan from All-Clad it would have cost over $100.  Same quality, same warranty (lifetime), You can\'t go wrong.  I can\'t wait to use it!  Ok, goodnight sweetheart it\'s time to go, DAN  '),(9014,2246318,'09,April,2004','2004-04-09 00:00:00','   Ok, so I already missed a day.  I was busy playing with my computer.  Getting it up to date  etc.  So yesterday the Mets lost 18 - 10 to the Braves.  Not Good.  Then again the Mets Lost today 10 - 8 also to the Braves.  The Braves take the series 3 games to 1.  Well, at least they are high scoring games, so you can\'t say they are boring.  Ok, I\'m off to bed, I get to work tomorrow.  DAN  '),(9015,2246318,'07,April,2004','2004-04-07 00:00:00','   The Mets season opener was today and the Mets beet the Braves 7 - 2.  GO METS!!!!  If you are a Mets fan, I will try to post as many of the Mets scores as possible.  It\'s a long season though, so don\'t get upset if I miss a day.  DAN  '),(9016,2246318,'04,April,2004','2004-04-04 00:00:00','   No Alias today, whatever am I to do?  Next week it is on an hour later, that\'s just no good.  I made fish and chips for dinner today and it came out pretty good.  I even surprised myself.  And yes I made the fish part from scratch.  I bought fresh Haddock at the store and I made a beer batter and fried the fish.  Yum!!  I guess I will find something to do, I ain\'t got much to talk about tonight, or at least right now, I\'m tired :)  DAN  P.S.  The last post on April 1, 2004 was an April Fool\'s joke in case you couldn\'t tell or didn\'t read my comment.  '),(9017,2246318,'01,April,2004','2004-04-01 00:00:00','   Ok, so when I got to work today, they told me that this will be my last week here.  I talked to Express and they don\'t have anything else for me right now.  They said they will be looking but now what am I supposed to do?  I can\'t afford not to have a  job!  Well, I guess I should get going and start work only 2 days left here I guess.  They didn\'t even say why, they just did, and since I\'m a Temp.  They can do that.  DAN  '),(9018,2246318,'31,May,2004','2004-05-31 00:00:00','   I am a little surprised that no one has left a single comment for me about the new design.  It\'s actually hard to tell if anyone has even been here since I\'ve been hitting the site to look for comments.  Oh well, I guess I do this for myself.  I had to work today.  It wasn\'t so bad, it was really slow.  My dept. had a total of one call incoming today.  I got paid time and a half plus I get a day off paid next week.  Plus it rained most of the day today.  So I don\'t have anything to complain about.  I\'m off to find something else to do.  DAN  P.S. to those who I went out with last night, in case you read this, I had a good time, we all should hang out more often.  '),(9019,2246318,'25,May,2004','2004-05-25 00:00:00','   So I went for a walk today.  Boy do I need to do that more often.  I went to where I used to walk when I needed to think.  All I thought about this time was nature and outer space.  No, not inside my head! :P  I walked for over an hour and it felt really good.  The last time I walked I lost about 15 pounds so I want to try to do that again, I need to lose more than 15 pounds though.  If I start eating better and walking then I think I can do it.  I don\'t need no stinking low carb diet!  Alright, that\'s all for tonight  DAN  '),(9020,2246318,'24,May,2004','2004-05-24 00:00:00','   Well, Season 3 of Alias has come to an end.  It sure was an action packed episode.  --If you haven\'t seen it yet and don\'t want to spoil it then don\'t read the rest of this post.--  First the commercials make you think that Sidney betrays the CIA, but in fact it was the blonde chick in disguise.  Later Vaughn goes to kill the blonde chick and he ends up getting stabbed in the back literally.  Then Sidney goes to get her for doing that to Vaughn and her Aunt almost kills her but Sidney out smarts her.  Sidney and the blonde chick get into a fight and Sidney almost gets killed but Vaughn saves the day and kills his wife.  I was kinda hoping the blonde chick would stick around for another season because you loved to hate her.  Plus she was very pretty.  But she\'s dead, or at least they are making us think she is.  They have done this before.  But when she was getting shot for the 14th and 15th times in the face and chest she tell Sydney to look for something.  Sydney goes into some sort of vault, she looks like she might be in an alias disguise and she looks at some black CIA papers that she has to look at with a black light.  They show some sort of project that involves her and it was undersigned by her Dad.  Somehow he is there and while she is balling he says that she was never supposed to see those papers.  And that is how they end season 3.  Now I left some stuff out because I am doing this a few hours after the fact and this is just supposed to be a summary, but it was a good show with a \"what do they say\" ending.  It really makes you anxious to see the first show of next season.  I really hope the show doesn\'t just end there.  That just wouldn\'t be good.  To all you other Alias fans out there, make sure you go out and buy the season 3 DVD set, out in September and make sure you watch come fall.  DAN  '),(9021,2246318,'23,May,2004','2004-05-23 00:00:00','   Ok, here is the new look of my Blog.  What do you all think?  Please let me know by clicking on the comments link next to my name at the end of this post.  I hope everyone likes it.  DAN  '),(9022,2246318,'03,May,2004','2004-05-03 00:00:00','   A lot is changing with this Blog.  First off Blogger, the way this Blog gets to you, has revamped their look.  Second, I am seriously thinking about changing the look of the Blog since it only really looks good in Netscape.  I want something that will look good in both Netscape AND Internet Explorer.  So keep checking back for the new look and new features including some pictures maybe.  The Alias season 3 finale is on tonight.  The commercials make it look like it is going to be one Hell of a show.  I will post after the show to voice my opinion about it.  I haven\'t heard if ABC picked it up for a 4th season, but they better have.  Season 3, yes this one that ends tonight, will be out on DVD in September, just in time for season 4\'s start.  It will make a good birthday gift!  hehe  Alright, I\'m off to change some things around here.  DAN  '),(9023,2246318,'03,May,2004','2004-05-03 00:00:00','   Ok, I am at work again.  I was just told that I got a 90% on my test.  That\'s really good.  I got a 90% or higher on EVERY one of my test throughout training.  That was my goal.  The Manager here asked if I wanted to work for PNG and be hired on as a permanent employee, and of course I said yes.  She handed me my employee handbook and told me I will fill out some paperwork tomorrow and my official start date will be 5/10/04.  Finally a job that I like, they seem to like me, and the pays pretty good.  Now if I can just get my personal life straightened out I\'ll be all good :)  Ok, it\'s lunch time so I am going to try to find something for lunch. Eat, Drink, and be merry! DAN  '),(9024,2246318,'27,June,2004','2004-06-27 00:00:00','   Boy it was a long weekend.  I didn\'t go to the weekly Applebee\'s on Friday this week.  I didn\'t have any money because of the aforementioned brake change. Plus I had to work on Saturday morning.  Then on Saturday a few friends of mine were trying to get together to have a bon fire but one who said he was going to go said, when I called him on Saturday, said that he decided to make other plans.  So we went without him.  It was fun, and if anyone who was there reads this, yes I did have fun.  I\'m not going to go into too much detail to save the innocent, or not so innocent, but I think everyone that went with me had fun as well.  I hope they did and my being there didn\'t bring them down.  It was a long night last night, that is for sure.  I didn\'t get home until 3:30am.  Then today my sister, brother-in-law, my father, and I went to dinner for fathers day.  It was in Warren and I\'m not sure of the name of the restaurant.  It was very good.  I had lasagna per my sisters suggestion and it was a good one.  The portions were, to put it lightly, HUGE.  I ate about a third of it and I was stuffed.  Hey, lunch for work the next two days.  I don\'t know if my Dad reads this or not, but if you do, happy fathers day Dad.  Oh, and earlier I washed my car.  I really washed my car.  I used the Mr Clean car wash thingy and it works pretty good.  It does what it says it does.  Well, except on my windows, but that is my fault because I have rain X on them so they bead up no matter what.  Then I waxed my car.  I bought, what is in my opinion, the best wax you can buy.  I did the black trim.  I made the tires look like they are wet, even if they are not.  And I put more Rain X on the windows.  Rain X is the best thing for your windows of your car.  If you haven\'t used it in the past you have to try it.  Don\'t bother with the off brands, go for the real thing.  Now my car drives like new and looks like new.  IT IS SHINY!! :)  And now my Mom is going to drive it to Boston and back.  I\'m not real happy about that but I am a nice guy so I am letting her drive it.  ONLY HER though.  I just hope that it comes back the way it will leave.  Well, on to another week of work. And remember, don\'t drink and drive. If you drink, don\'t drive. DD\'s - our modern heroes, saving lives.  I should know, I am one :)  DAN  UPDATE:  I just wanted to let you know that I have added the Rain-X website to my list of links on the right.  Also on the main page you can look up your car to see what size wiper blades it takes.  Even if you don\'t use Rain-X items, this can be a helpful tool.  Laters.  '),(9025,2246318,'20,June,2004','2004-06-20 00:00:00','   I got a new grill today.  I bought a Weber charcoal grill.  I haven\'t used it yet but it looks pretty good.  The store that I bought it from said that assembly would cost $30!  I said no way and I put it together myself.  It took me a total of 8 min.  It was not hard at all and definitely not worth $30.  I also got my car inspected today.  I knew my brakes were bad, but I was hoping the car would pass anyway.  It did not, I needed new front pads and they had to resurface the rotors, the total for everything including a oil change was $207.  That took a very big chunk out of my paycheck.  Went to a steak cook out today with my Mom and her boyfriend.  It was stupid.  They started the grill just as everyone got there so everyone started putting their steak on (you had to cook your own, which I prefer) when the grill was still way too cool to cook right.  Then the steaks were still mostly frozen.  How about you thaw the steaks before you try cooking them?  So everyone cooked their steaks, which took way too long, they ate the steaks and barely talked and then just left.  Nothing else to do or anything.  Free steak is good, but kinda pointless.  It was cold too, but that was a surprise.  I rented Scary Movie 3 today and it was very funny.  The best one yet.  The little kid just couldn\'t keep safe.  If you haven\'t seen it yet, go and rent it.  So when I went up to check my e-mail etc online I turned on my TV Comedy Central and what comes on but none other than The South Park movie, bigger, longer, uncut.  Now, what surprised me was that it was still uncut.  EVERYTHING was left in.  All the swearing and crude language, all the cartoon nudity like the seen with Kenny going up to heaven and the Saddam seen.  I have to say that I was very surprised they actually played it on TV without a change.  I\'m not complaining just surprised, especially after all the threats from the FCC for indecentcy.  Yes, I did watch the whole movie, I haven\'t seen it since I saw it in the theatre.  Ok, that\'s too much for one post.  Catch ya...  DAN  '),(9026,2246318,'11,June,2004','2004-06-11 00:00:00','   I Cuss,  You Cuss,  We All Cuss For Asparagus!   DAN  '),(9027,2246318,'10,June,2004','2004-06-10 00:00:00','   Ok, I won\'t do this often but I am going to post an e-mail forward that I received because I liked it so much.  Here it is:  Ok, I just got this and I thought \"Oh what could it actually say?\"  It is one of the funniest e-mails I\'ve been sent in a long time.  You HAVE to try this!!!!  Dan Paterniti Try this soon, before someone forces Google to fix its site  1) Go to  urlLink www.Google.com   2) Type in -- weapons of mass destruction--(DON\'T hit return)  3) Hit the \"I\'m feeling lucky\" button, NOT the \"Google search\"  4) Read the \"error message\" carefully -- the WHOLE page. Someone at Google really has a sense of humor.  '),(9028,2246318,'02,June,2004','2004-06-02 00:00:00','   Hello,  Not much to say today.  I finally got a comment about the new look.  I do plan on posting a pic of me, no not to scare you away.  When I have some more time I will make some comments on some of the articles in  Scientific American  magazine.  There has been some interesting things in there and I would like to put my thoughts out there as well.  Other than that I don\'t really have much to say.  Things are still going well at work.  Crash, DAN  '),(9029,2246318,'29,July,2004','2004-07-29 00:00:00','   urlLink John Kerry for President  '),(9030,2246318,'25,July,2004','2004-07-25 00:00:00','   Ok, so here is the link to my review of the concert.  Sorry it\'s even later than anticipated.  I guess better late than never.   urlLink Click Here.  '),(9031,2246318,'22,July,2004','2004-07-22 00:00:00','   We have a big show tonight so I\'ll get right into it.     First, the Dave Matthews Band concert that I went to last Tuesday. In as few words as possible, it was great! I want to talk about each song so I put it onto it\'s own page. I am running out of time so I will post the link tomorrow. Sorry for the wait.    Next, on Saturday I went over to a friends house from work and we played Pictionary. I had a really good time but I again learned that I can\'t draw worth beans. There was another co-worker there with her husband and I think we all had fun. I hope we can all get together again to do that.     Then on Sunday I saw  urlLink Spider-Man 2 . It was a great show. I think it was as good if not better than the first one. They focused more on the romance story line between Peter and MJ on this one, but they did not lack on great fight scenes. I don\'t want to give too much away but spidey fights Doc Octopus a few times and they are great fights. The story was really good, it kept me wondering what was coming next and it leads into the next installment. I, for one, cannot wait.     One of the local radio stations in the area is giving away round trip tickets, hotel room, and VIP tickets to see  urlLink Dave Matthews Band in San Fran  this September. I keep listening to see if I can try to win. That would be great to go to the other coast and see DMB. The radio station actually announced that they were giving away tickets to this show before the official DMB website announced the show. Kinda cool.     Ok, there will be another big post coming up because first this Friday I am going over to another friends house from work, and then on Saturday I am going to a graduation party. So I should have a lot to talk about.      \"Nine planets round the sun  Only one does the sun embrace  Upon this watered one  So much we take for granted   So let us sleep outside tonight  Lay down in our mother\'s arms  For here we can rest safely \"   --Dave Matthews Band - One Sweet World     DAN  '),(9032,2246318,'15,July,2004','2004-07-15 00:00:00','   So on Tuesday Night I went to see Dave Matthews Band Live at Darien Lake with a friend of mine.  They were great.  I won\'t go into my review just yet, I really don\'t have time right now but I will post the set list.  Intro (fade into next song) Don\'t Drink the Water Don\'t Burn the Pig One Sweet World Intro (fade into next song) New Song #3 Bass Intro (fade into next song) Crush Song That Jane Likes New Song #4 Too Much Recently Interlude Rhyme & Reason Hmm ? (New) The Space Between New Song #1 Warehouse What Would You Say [Tease] New Song #2 Ants Marching  Encore: Tripping Billies  DAN  '),(9033,2246318,'14,July,2004','2004-07-14 00:00:00','   Don\'t Burn the Day Away  DAN  '),(9034,2246318,'09,July,2004','2004-07-09 00:00:00','   Think back, a few years ago, before anyone knew what the  urlLink West Nile Virus  was....~~Dream Sequence~~....People were trying to get that insect repellent off the market.  Why?  Because they were saying how unsafe Deet is.  They were telling us to use items that had the least amount of Deet possible....~~End Dream Sequence~~...Now, in a world afraid of West Nile, Deet is everyone\'s best friend.  You don\'t hear anyone saying anything bad about Deet.  Why?  Because it works so well to repel mosquitoes.  Even the Governments CDC website tells us to \"use repellent with DEET when spending time outdoors.\"  Now, I\'m not trying to say anything bad about Deet or downplay what the West Nile Virus could do, I\'m just making an observation about how things can be changed.  Do I use Deet?  Yes, I do when I am outdoors where there are a lot of mosquitoes.  Do I feel it\'s safe?  I do when used properly.  Why believe me though, Here is what the CDC answers to whether it is safe or not:  \"Yes, products containing DEET are very safe when used according to the directions. The regulation of substances such as DEET is determined by the  urlLink U.S. Environmental Protection Agency (EPA) . Because DEET is so widely used, a great deal of testing has been done. When manufacturers seek registration with the EPA, laboratory testing regarding both short-term and long-term health effects must be carried out. Over the long history of DEET use, very few confirmed incidents of toxic reactions to DEET have occurred when the product is used properly. To read more information about DEET safety see the National Pesticide Information Center [NPIC], EPA re-registration eligibility decision fact sheet ( urlLink npic.orst.edu/factsheets/DEETgen.pdf   ) and the  urlLink EPA Re-registration eligibility decision (R.E.D.) fact sheet .\"  What is  urlLink Deet ?  Well, first off, Deet is N,N-diethyl-meta-toluamide.    It is actually a product made by one company,  urlLink Morflex, Inc .  They make all the Deet you find in anything.  It was actually developed by the U.S. Department of Agriculture in 1946.  It is a very efficient and long lasting bug repellent.  Not just mosquitoes but many bugs.  When at the store pick your favorite product that contains Deet and then pick the highest concentration of it.  The more Deet, the longer it will last, and the less amount of times you will have to re-apply.  For more information about Deet, go to  urlLink www.deet.com , CDC Insect Repellent Use and Safety at  urlLink www.cdc.gov/ncidod/dvbid/westnile/qa/insect_repellent.htm ,  and Deet Online at  urlLink www.deetonline.org   What is West Nile Virus?  Well, what everyone\'s afraid of is fatal encephalitis or inflammation of the brain.  That doesn\'t sound like fun to me.  Most people don\'t even show any symptoms of having West Nile though.  But if you do, try looking for things like fever, headache, body aches, and maybe swollen lymph glands.  For a full list go to CDC\'s Symptoms of West Nile Virus at  urlLink www.cdc.gov/ncidod/dvbid/westnile/qa/symptoms.htm .  Most cases aren\'t all that serious but if you think you might have West Nile or what they call West Nile Fever then see your doctor as soon as possible and they will tell you what to do.  For more information about West Nile, go to the CDC West Nile Virus Homepage at  urlLink www.cdc.gov/ncidod/dvbid/westnile/ .    Now, I have just gone over a few of the major points about Deet and West Nile Virus.  Please check out the sites I listed or do a search yourself for more complete information.  Well, I guess we\'ve seen the end of the war against Deet.  At least for now and while West Nile is still a threat.  DAN  '),(9035,2246318,'08,July,2004','2004-07-08 00:00:00','   Ok, ok, so it\'s called Croquet.  I knew Polo was with horses and all, and aren\'t they similar?  I don\'t know.  If you click on the title of this post it will send you to a history of Croquet.  You\'d be surprised how many sites there are about Croquet.  I thought, I\'ll find a quick website by going to  urlLink Google Search  and putting in Croquet.  WRONG.  There are a lot more than I thought.  It only found 579,000 pages.  only.  So, I had no idea how to play when I watched the first game but it didn\'t seem to hard and I picked up on it really quickly.  If you want the official rules, which by the way vary from country to country, then you can  urlLink click here .  People really take this game seriously.  They have tournaments and everything.  Now it was fun, and I might think about buying a Croquet set to play more often, but some take it too seriously.  You can see this at  urlLink Croquet World Online Magazine .  lesson Over, DAN  '),(9036,2246318,'06,July,2004','2004-07-06 00:00:00','   Hello,  I hope everyone had a safe and happy Fourth of July weekend.  On Saturday my sister invited me over to the their house to see the fireworks that were at the Jammers game.  They were ok, a little short and a little plain but it was still fun.  I got to spend some time with my family while they were at least pretending not to hate me.  I meet, who I assume is, my Dad\'s girlfriend.  My sister told me earlier in the day that he had one but he didn\'t introduce her as such.  I was going to do the usual \"nice to meet you, goodbye.\" etc but when she passed me she looked in the opposite direction and didn\'t say anything.  Oh well.  On Sunday I went over to my sisters husbands Mom\'s house and we ate chicken and played Polo, but without the horses.  Actually I think it was called something else but I don\'t remember what, all I remember was that I was not good at it.  It was, in my own defense, my first time playing though.  It was fun but ran a little long.  My Dad left early, well earlier than I did.  Well, gotta work.  DAN  '),(9037,2246318,'04,July,2004','2004-07-04 00:00:00','   I would just like to wish everyone a happy Fourth of July.  Please be safe and have lots of fun!  DAN  '),(9038,3969316,'19,July,2004','2004-07-19 00:00:00','    Today is another day sat at this desk, sat at this computer, whilst working from home.&nbsp; My normal routes of pretending that it doesn\'t drive me insane, are: dancing madly around my kitchen to loud music which sounds like hamsters on speed, talking to a squirrel who comes and sits and eats out of a purpose built hanging basket or accosting the postman into having a chat.&nbsp; So, I thought a blog might entertain me, even if it bores the pants off everyone else!  &nbsp;  This job - working for an internet based&nbsp;company, who are actually great fun to be with but are about 60 miles away from me and it makes it tricky when I want to make a cup of coffee for everyone but it&nbsp;is my bread and butter, whilst I am training to be a sound healer.&nbsp;  &nbsp; I am currently working my way through case studies and a thesis to finish my qualification but I feel this training has been going on since I popped into this world.&nbsp; Everything which has happened to me has bought me to here and I really like here.&nbsp; People moan and whinge and say the world is a terrible place but really thats a matter of perception.&nbsp;  &nbsp;  Maybe I was brought up on a diet of Monty Python and that has a lot to answer for?!  &nbsp;  I love doing the sound healing - I love singing - I love watching people\'s reactions to what I do and I love music.&nbsp; I am looking for people\'s comments on how music moves them.&nbsp; To me, it creates an emotion, a feeling which is beyond words but that has never stopped me trying to explain.&nbsp; Music has bought me towards so many people who would normally have been shy to come up and talk.&nbsp; I used to be a DJ and people would hug me and would chat chat chat about the smallest sounds and noises which they adored.&nbsp;  &nbsp;  As I explore using my voice, I get the same reaction ... a connection with people and a connection with the Universe. &nbsp; I had an epithany a while ago and decided that working in the City for a corporate business just wasn\'t for me and so I packed it all in and decided to try and find out what i really wanted from life ... what being truly authentic means to me.&nbsp; I have met some interesting people along the way and I have certainly had a few highs and a few lows on really trying to understand what really makes me tick.&nbsp; I certainly don\'t have any regrets for the things I have done in the past but I do feel like I have stopped thrashing around and everything has simplified itself as a result. &nbsp; I am, I guess a spiritual person and everyday presents another bit of my path in front of me but I love the balance of dark and light in everyone and I find it fascinating watching people transform into something \'more\' as they examine their choices in life.&nbsp;  &nbsp; This is why I love sound - it doesn\'t judge.   '),(9039,3969316,'19,July,2004','2004-07-19 00:00:00','    My lovely man is currently indoors with me, suffering from \'all the aches\'.&nbsp; He does look grim but I realise I really am a rampant, active Aries at heart.&nbsp; One part of me is wanting to fuss him and keep him at home forever (it beats talking to my buddha statue) and the other part is like wishing him better at speed.&nbsp;     &nbsp;  Where has my play mate gone. My sparky demon lover who understands me when I make jokes about Jesus-on-a-stick.&nbsp;     &nbsp;  I know he is ill because I gave him a connoction to drink (which included honey with toast crumbs in it) and he just drank it!   '),(9040,4201404,'13,August,2004','2004-08-13 00:00:00','   Behind The Scenes Of The Iraq War   The plan for the Iraq war, which has erupted in the face of opposition from the entire world, was drawn up at least decades ago, by Israeli strategists   In its attempt to realize its strategy of destabilizing or dividing the Middle Eastern Arab states, Israel has Egypt, Syria, Iran and Saudi Arabia on its list of subsequent targets.   As these lines are being written, the United States of America has begun striking at Iraq. Despite the fact that most countries of the world, and even the majority of the USA\'s allies, opposed it, the US administration was determined for the strike to go ahead. When we look behind the scenes of this insistence, it seems that Israel and its powerful lobby in the US, have the greatest share in the make-up.   In fact, Israel\'s policy aimed at the fragmentation of Iraq has lengthy historical roots-   The Age-Old Israeli Plan to Divide Iraq   An ambitious report entitled \"A Strategy for Israel in the 1980s,\" which appeared in the World Zionist Organization\'s periodical Kivunim in February 1982 disclosed a strategy aimed at making the whole of the Middle East a kind of \"living space\" for Israel. The report, drawn up by Oded Yinon, an Israeli journalist and formerly attached to the Foreign Ministry of Israel, set out the scenario of the \"division of Iraq\" in these terms:   Iraq, rich in oil on the one hand and internally torn on the other, is guaranteed as a candidate for Israel\'s targets. Its dissolution is even more important for us than that of Syria- Iraq is, once again, no different in essence from its neighbors, although its majority is Shi\'ite and the ruling minority Sunni. Sixty-five percent of the population has no say in politics, in which an elite of 20 percent holds the power. In addition there is a large Kurdish minority in the north, and if it weren\'t for the strength of the ruling regime, the army and the oil revenues, Iraq\'s future state would be no different than that of Lebanon in the past- In Iraq, a division into provinces along ethnic/religious lines as in Syria during Ottoman times is possible. So, three (or more) states will exist around the three major cities: Basra, Baghdad and Mosul, and Shi\'ite areas in the south will separate from the Sunni and Kurdish north. 1)   This was not the only announcement of the Israeli plan to atomize Middle East, including Iraq. As Israel Shahak, the notable Israeli scholar known for his dedication to a peaceful solution in the Middle East, explained that Yinon was just echoing the views of Israeli hawks:   The idea that all the Arab states should be broken down, by Israel, into small units, occurs again and again in Israeli strategic thinking. For example, Ze\'ev Schiff, the military correspondent of Ha\'aretz (and probably the most knowledgeable in Israel, on this topic) writes about the \"best\" that can happen for Israeli interests in Iraq: \"The dissolution of Iraq into a Shi\'ite state, a Sunni state and the separation of the Kurdish part\" (Ha\'aretz 6/2/1982). Actually, this aspect of the plan is very old. 2)   Thus, the plan was a serious one and this has been confirmed by the age-old Israeli support to non-Arab or non-Muslim minorities in the Muslim Arab states. The rebellious Kurds of northern Iraq was one of these strategic allies of Israel. During their revolt against the Baghdad regime, 1961-75, they have been financially and militarily supported by Israel. Israelis would love to see them carve up the northern part of Iraq, no matter how bloody and devastating such a civil war would be. However, the revolt failed in 1975, after loosing the support of its major patron, the Shah.   Fifteen years later, a new opportunity arose for Israel, an opportunity from the ambitions of the Iraqi dictator, Saddam Hussein.   Israel\'s Role in the Gulf War   Iraqi dictator Saddam Hussein, after his bloody war against Iran in the 80\'s, invaded Kuwait in a sudden attack on August 1, 1990, giving rise to an international crisis. Israel headed the list of those forces that encouraged that crisis. Israel was the fiercest supporter of the attitude adopted by the United States in the wake of the invasion of Kuwait. The Israelis even regarded the United States as moderate, and wanted a harsher policy. To such an extent in fact that the President of Israel Chaim Herzog recommended that the American use nuclear weapons. On the other hand, the Israeli lobby in the United States was working to bring about a wide-ranging attack on Iraq.   This whole situation encouraged the idea in the United States that the attack against Iraq under consideration was actually planned in Israel\'s interests. Pat Buchanan summarized this idea in the words \"There are only two groups that are beating the drums for war in the Middle East - the Israeli Defense Ministry and its amen corner in the United States.\"   Israelis had also initiated a serious propaganda campaign on the issue. Since this campaign was largely waged in secret, Mossad also entered the equation. Rogue Mossad agent Victor Ostrovsky provides important information on this subject. According to Ostrovsky, Israel had wanted to wage war with the United States against Saddam long before the Gulf crisis. So much so in fact, that Israel began to implement the plan immediately after the Iran-Iraq war. Ostrovsky reports that Mossad\'s Psychological Warfare department (LAP - LohAma Psychologist) set about an effective pro-war campaign using misinformation techniques. 3)   A Mossad Agent Describes the Gulf War   Ostrovsky describes how Mossad used agents or sympathizers in various parts of the world in this campaign. Among the tools employed in the campaign were the horrible massacres done by missiles launched against civilian targets in Iran during the Iran-Iraq war. As Ostrovsky makes clear, Mossad\'s later use of these missiles as a propaganda tool was quite peculiar, since those missiles had actually been directed towards their targets by Mossad, with the help of information from US satellites. Having supported Saddam throughout his war with Iran, Israel was now disclosing his crimes. Ostrovsky writes:   The Mossad leaders know that if they could make Saddam appear bad enough and a threat to the Gulf oil supply, of which he\'d been the protector up to that point, then the United States and its allies would not let him get away with anything, but would take measures that would all but eliminate his army and his weapons potential, especially if they were led to believe that this might just be their last chance before he went nuclear. 4)   The Israelis were so determined on this matter, and with regard to the United States, that on August 4, 1990, Israeli Foreign Minister David Levy issued a diplomatically worded threat to William Brown, the American ambassador to Israel, stating that Israel \"expects the U.S. will fulfill all of the goals it set for itself at the beginning of the gulf crisis,\" in other words that it attack Iraq. According to Levy, if the United States failed to do so, Israel would act unilaterally. 5)   It would be of enormous benefit to Israel to have the United States engage in the war and for Israel to remain entirely uninvolved: and that is indeed what happened.   Israel Forces the USA to War   However, the Israelis were actively involved in the United States\' war plans. Some US staff officers involved in planning Operation Desert Storm received fine tactical advice from the Israelis that \"the best way of wounding Saddam was to strike at his family.\"   The Mossad-inspired propaganda campaign reported by Ostrovsky set up the necessary public backing for the Gulf War. It was again Mossad local assistants who lit the touchpaper for the war. The Hill and Knowlton lobbying firm, run by Tom Lantos of the Israeli lobby, prepared a dramatic scenario to convince members of Congress on the subject of war against Saddam. Turan Yavuz, a noted Turkish journalist, describes the incident:   October 9, 1990. The Hill and Knowlton lobbying firm organizes a sitting in Congress on the subject of \"Iraq\'s Barbarities.\" A number of \"eye witnesses\" brought to the session by the lobbying firm maintain that Iraqi troops killed new-born babies in the hospital wards. One \"eye witness\" describes the savagery in enormous detail, saying that Iraqi soldiers killed 300 new-born babies in one hospital alone. This information deeply disturbs the members of Congress. This works to President Bush\'s advantage. However, it later emerges that the eye witness brought by Hill and Knowlton to Congress is in fact the daughter of the Kuwaiti ambassador to Washington. Nevertheless, the daughter\'s account is sufficient for members of Congress to give Saddam the nickname \"Hitler\". 6)   This leads to just one conclusion: that Israel played an important role in the United States\' to wage its first war on Iraq. The second one is not much different.   The Pretext of \"War Against Terrorism\"   Contrary to popular belief, the plan to attack Iraq and overthrow Saddam Hussein\'s regime by force was prepared and placed on Washington\'s agenda long before the genesis of the \"war against terrorism,\" which emerged in the wake of September 11. The first indication of this plan emerged in 1997. A group of pro-Israeli hawkish strategists in Washington D.C. began to put forward the scenario of the invasion of Iraq by manipulating the \"neo-con\" think-tank, called PNAC (Project for The New American Century). The most notable names in the PNAC were those of Donald Rumsfeld and ###### Cheney, who as defense secretary and vice-president would be the most influential figures in the George W. Bush administration.   An article titled \"Invading Iraq Not a New Idea for Bush Clique: 4 Years Before 9/11 Plan Was Set\" written by William Brunch and published in the Philadelphia Daily News, sets out the following facts:   But in reality, Rumsfeld, Vice President ###### Cheney, and a small band of conservative ideologues had begun making the case for an American invasion of Iraq as early as 1997 - nearly four years before the Sept. 11 attacks and three years before President Bush took office.   An obscure, ominous-sounding right-wing policy group called Project for the New American Century, or PNAC - affiliated with Cheney, Rumsfeld, Rumsfeld\'s top deputy Paul Wolfowitz and Bush\'s brother Jeb - even urged then-President Clinton to invade Iraq back in January 1998. 7)   Is Oil the Real Objective?   Why were the PNAC members so determined to attack Iraq? The same article continues:  While oil is a backdrop to PNAC\'s policy pronouncements on Iraq, it doesn\'t seem to be the driving force. [Ian] Lustick, [a University of Pennsylvania political science professor and Middle East expert,] while a critic of the Bush policy, says oil is viewed by the war\'s proponents primarily as a way to pay for the costly military operation.   \"I\'m from Texas, and every oil man that I know is against military action in Iraq,\" said PNAC\'s Schmitt. \"The oil market doesn\'t need disruption.\"   Lustick believes that a more powerful hidden motivator may be Israel. He said Bush administration hawks believe that a show of force in Iraq would somehow convince Palestinians to accept a peace plan on terms favorable to Israel- 8)   This, therefore, is the principal motivation behind the plan to attack Iraq: to serve Israel\'s Middle East strategy.   This fact has also been identified by other Middle East experts. Cengiz ?andar, a Turkish Middle East expert, for instance, describes the real power behind the plan to attack Iraq thus:   ... Who is directing the attack on Iraq? Vice-President ###### Cheney, Defense Secretary Rumsfeld, National Security Adviser Condoleeza Rice. These are the \"senior level\" backers of the attack. Yet the rest of the iceberg is even richer and more interesting. There are a number of \"lobbies.\"  Heading these lobbies are the Jewish Institute for Security Affairs team, pro-Likud and Israeli-right and known for their close relations with US arms manufacturers. These have close relations with the \"arms lobby,\" Lockheed, Northrop, General Dynamics and Israeli military industries ... JINSA\'s fundamental principle is this: America\'s and Israel\'s security are inseparable. In other words, they are the same thing.   JINSA\'s objective is not solely the overthrow of the Saddam regime in Iraq: it also supports the overthrow of the Saudi Arabian, Syrian, Egyptian and Iranian regimes with a logic of \"total war,\" and the subsequent importation of \"democracy.\" ... In other words, a number of American Jews on the same wavelength as the most extreme factions in Israel at the moment comprise the hawks in Washington. 9)   The Israeli Strategy for The Muslim Middle East   In short, there are those in Washington who are encouraging a war aimed first at Iraq and then at Saudi Arabia, Syria, Iran and Egypt. The most distinguishing feature of these is that they are lined up alongside, and even equivalent to, the \"Israeli lobby.\"   No matter how much they speak of \"American interests,\" these people are actually supporting Israeli interests. A strategy of waging war against the whole of the Middle East and turning all the peoples of the region against it cannot be to the United States\' advantage. The adoption of such a strategy can only be possible if the United States is bound to Israel, by means of the Israeli lobby, which is profoundly influential in this country\'s foreign policy.   It is for these reasons that behind the strategy which began to be set in motion after September 11 and is aimed at re-arranging the entire Islamic world, lies Israel\'s strategy for dominating the Middle East. Ever since its foundation, Israel has aimed at restructuring the Middle East, making it manageable and secure to itself. The search for security is of course a legitimate one, but the way that Israel chose to achieve this end is wrong: From the beginning, Israel decided to establish security behind an \"iron wall\" that would separate itself from the Arabs, and most important of all, protect the lands that Israel occupied through methods of invasion, colonization and depopulation. This strategy of reaping the wind had its consequence as a century of constant clash between Israelis and Arabs. Had Israel chosen a peaceful path to secure its existence, by building good relations with its Arab neighbors and refraining from aggression, Jews and Muslims could peacefully co-exist in the Middle East, as they have done before for centuries.   However, the radical Zionist ideology still denies any chance to peace and relentlessly tries to transform the whole Middle East to create a \"living space\" for Israel. It has been using its influence in the United States for that purpose in recent years, and to a large extent directs Washington\'s Middle East policy. The post-September 11 climate gave Israel the opportunity it had been seeking. Pro-Israeli ideologues who for years had been propounding the falsehood that Islam itself - not some militant radicals who use Islam as a shelter - posed a threat to the West and the United States, and who encouraged the mistaken concept of a \"clash of civilizations,\" have been trying to incite the United States against the Islamic world in the wake of September 11. As early as 1995, Israel Shahak of the Jerusalem Hebrew University wrote former Israeli Prime Minister Rabin\'s obsession with \"the idea of an Israeli-led anti-Islamic crusade\". Nahum Barnea, a commentator from the Israeli newspaper Yediot Ahronot, stated that same year that Israel was making progress \"[to] become the Western vanguard in the war against the Islamic enemy.\" 10)   All that has happened in the years which have followed is that Israeli hawks have made their intentions even clearer. The political climate in the wake of September 11 prepared the ground for this intention to be made a reality.   The Only Way to World Peace: An Islamic Union   The situation may be summarized as follows: Israel\'s aim is to reshape the Middle East in line with its own strategic interests. In order to do this, it needs a \"world power.\" That power is the United States; and Israeli hawks, thanks to their influence there, are trying to wage a militant American strategy against the Islamic world. Although Israel is a small state with a population of only 4.5 million, the plans drawn up by Israeli strategists and their counterparts in the West are directing the world affairs.   What needs to be done in the face of this?   1) \"Counter lobby activities\" need to be adopted in the face of the Israeli lobby\'s influence in the United States in order to develop dialogue between the United States and the Islamic world and to invite it to seek peaceful solutions to Iraq and similar problems. A wide section of the United States wishes to see their country adopt a fairer Middle East policy. Many statesmen, strategists, journalists and intellectuals have expressed this, and a \"peace of civilizations\" movement must be carried forward in cooperation with them. The approach inviting the US administration to peaceful solutions must be carried forward at governmental and civil society organization level.   2) A reconciliation between Israel and the Muslim Middle East must be sought. There are many \"peaceniks\", i.e. pro-peace Israelis, too. Many Israeli statesmen, religious leader, opinion leaders and many Jews from all around the world are urging the Israeli state to refrain from its brutal occupation and accept a just peace to live along with their Arab neighbors. Cooperation with them, especially on the inter-faith level, should be initiated and encouraged. One thing should never be forgotten: The rise of radicalism, enmity and violence is bad for all parties.   3) Alongside all this, a deeper rooted solution lies in a project which can resolve all the problems between the Islamic world and the West and deal with the fragmentation, suffering and poverty in the Islamic world and totally alter it: An Islamic Union. Recent developments have shown that the whole world, not just Islamic regions, stands in need of an \"Islamic Union.\" This Union should heal the radical elements in the Islamic World, and establish good relations between Muslim countries and the West, especially the United States.   This Union, can find a solution to the mother of all problems: The Arab-Israeli conflict. With Israel retreating to its pre-67 borders and Arabs recognizing its right to exist, there can be real peace in the Middle East. And Jews and Muslims - both Children of Abraham and believers in one true God - may peacefully co-exist in the Holy Land, as they have done during the past centuries. Then, Israel would need no strategy to destabilize or divide the Arab States. And it will not have to face the results of occupation in forms of terrorism and constant fear of annihilation. Then, both the Israeli and Iraqi (and Palestinian) children may grow up in peace and security. That is a Middle East that any sane person should work to see.   ------------------------------------------------------------------------   1- (ed.) Israel Shahak, The Zionist Plan for the Middle East; from Oded Yinon\'s \"A Strategy for Israel in the Nineteen Eighties\" Published by the Association of Arab-American University Graduates, Inc. Belmont, Massachusetts, 1982 Special Document No. 1 (ISBN 0-937694-56-8); 2- (ed.) Israel Shahak, The Zionist Plan for the Middle East;  3- Victor Ostrovsky, The Other Side of Deception, pp. 252-254.  4- Victor Ostrovsky, The Other Side of Deception, p. 254  5- Andrew and Leslie ####burn, Dangerous Liaison, p. 356.  6- Turan Yavuz, ABD\'nin K?rt Kart? (The US\' Kurdish Card), p. 307  7- William Bunch, Philadelphia Daily News, Jan. 27, 2003  8- William Bunch, \"Invading Iraq not a new idea for Bush clique\" Philadelphia Daily News, Jan. 27, 2003  9- Cengiz ?andar, \"Iraq and the \'Friends of Turkey\' American Hawks\", Yeni ?afak, September 3, 2002.  10- Israel Shahak, \"Downturn in Rabin\'s Popularity Has Several Causes\", Washington Report on Middle East Affairs, March 1995.  '),(9041,4201404,'11,August,2004','2004-08-11 00:00:00','   Behind The Scenes Of The Iraq War   The plan for the Iraq war, which has erupted in the face of opposition from the entire world, was drawn up at least decades ago, by Israeli strategists   In its attempt to realize its strategy of destabilizing or dividing the Middle Eastern Arab states, Israel has Egypt, Syria, Iran and Saudi Arabia on its list of subsequent targets.   As these lines are being written, the United States of America has begun striking at Iraq. Despite the fact that most countries of the world, and even the majority of the USA\'s allies, opposed it, the US administration was determined for the strike to go ahead. When we look behind the scenes of this insistence, it seems that Israel and its powerful lobby in the US, have the greatest share in the make-up.   In fact, Israel\'s policy aimed at the fragmentation of Iraq has lengthy historical roots-   The Age-Old Israeli Plan to Divide Iraq   An ambitious report entitled \"A Strategy for Israel in the 1980s,\" which appeared in the World Zionist Organization\'s periodical Kivunim in February 1982 disclosed a strategy aimed at making the whole of the Middle East a kind of \"living space\" for Israel. The report, drawn up by Oded Yinon, an Israeli journalist and formerly attached to the Foreign Ministry of Israel, set out the scenario of the \"division of Iraq\" in these terms:   Iraq, rich in oil on the one hand and internally torn on the other, is guaranteed as a candidate for Israel\'s targets. Its dissolution is even more important for us than that of Syria- Iraq is, once again, no different in essence from its neighbors, although its majority is Shi\'ite and the ruling minority Sunni. Sixty-five percent of the population has no say in politics, in which an elite of 20 percent holds the power. In addition there is a large Kurdish minority in the north, and if it weren\'t for the strength of the ruling regime, the army and the oil revenues, Iraq\'s future state would be no different than that of Lebanon in the past- In Iraq, a division into provinces along ethnic/religious lines as in Syria during Ottoman times is possible. So, three (or more) states will exist around the three major cities: Basra, Baghdad and Mosul, and Shi\'ite areas in the south will separate from the Sunni and Kurdish north. 1)   This was not the only announcement of the Israeli plan to atomize Middle East, including Iraq. As Israel Shahak, the notable Israeli scholar known for his dedication to a peaceful solution in the Middle East, explained that Yinon was just echoing the views of Israeli hawks:   The idea that all the Arab states should be broken down, by Israel, into small units, occurs again and again in Israeli strategic thinking. For example, Ze\'ev Schiff, the military correspondent of Ha\'aretz (and probably the most knowledgeable in Israel, on this topic) writes about the \"best\" that can happen for Israeli interests in Iraq: \"The dissolution of Iraq into a Shi\'ite state, a Sunni state and the separation of the Kurdish part\" (Ha\'aretz 6/2/1982). Actually, this aspect of the plan is very old. 2)   Thus, the plan was a serious one and this has been confirmed by the age-old Israeli support to non-Arab or non-Muslim minorities in the Muslim Arab states. The rebellious Kurds of northern Iraq was one of these strategic allies of Israel. During their revolt against the Baghdad regime, 1961-75, they have been financially and militarily supported by Israel. Israelis would love to see them carve up the northern part of Iraq, no matter how bloody and devastating such a civil war would be. However, the revolt failed in 1975, after loosing the support of its major patron, the Shah.   Fifteen years later, a new opportunity arose for Israel, an opportunity from the ambitions of the Iraqi dictator, Saddam Hussein.   Israel\'s Role in the Gulf War   Iraqi dictator Saddam Hussein, after his bloody war against Iran in the 80\'s, invaded Kuwait in a sudden attack on August 1, 1990, giving rise to an international crisis. Israel headed the list of those forces that encouraged that crisis. Israel was the fiercest supporter of the attitude adopted by the United States in the wake of the invasion of Kuwait. The Israelis even regarded the United States as moderate, and wanted a harsher policy. To such an extent in fact that the President of Israel Chaim Herzog recommended that the American use nuclear weapons. On the other hand, the Israeli lobby in the United States was working to bring about a wide-ranging attack on Iraq.   This whole situation encouraged the idea in the United States that the attack against Iraq under consideration was actually planned in Israel\'s interests. Pat Buchanan summarized this idea in the words \"There are only two groups that are beating the drums for war in the Middle East - the Israeli Defense Ministry and its amen corner in the United States.\"   Israelis had also initiated a serious propaganda campaign on the issue. Since this campaign was largely waged in secret, Mossad also entered the equation. Rogue Mossad agent Victor Ostrovsky provides important information on this subject. According to Ostrovsky, Israel had wanted to wage war with the United States against Saddam long before the Gulf crisis. So much so in fact, that Israel began to implement the plan immediately after the Iran-Iraq war. Ostrovsky reports that Mossad\'s Psychological Warfare department (LAP - LohAma Psychologist) set about an effective pro-war campaign using misinformation techniques. 3)   A Mossad Agent Describes the Gulf War   Ostrovsky describes how Mossad used agents or sympathizers in various parts of the world in this campaign. Among the tools employed in the campaign were the horrible massacres done by missiles launched against civilian targets in Iran during the Iran-Iraq war. As Ostrovsky makes clear, Mossad\'s later use of these missiles as a propaganda tool was quite peculiar, since those missiles had actually been directed towards their targets by Mossad, with the help of information from US satellites. Having supported Saddam throughout his war with Iran, Israel was now disclosing his crimes. Ostrovsky writes:   The Mossad leaders know that if they could make Saddam appear bad enough and a threat to the Gulf oil supply, of which he\'d been the protector up to that point, then the United States and its allies would not let him get away with anything, but would take measures that would all but eliminate his army and his weapons potential, especially if they were led to believe that this might just be their last chance before he went nuclear. 4)   The Israelis were so determined on this matter, and with regard to the United States, that on August 4, 1990, Israeli Foreign Minister David Levy issued a diplomatically worded threat to William Brown, the American ambassador to Israel, stating that Israel \"expects the U.S. will fulfill all of the goals it set for itself at the beginning of the gulf crisis,\" in other words that it attack Iraq. According to Levy, if the United States failed to do so, Israel would act unilaterally. 5)   It would be of enormous benefit to Israel to have the United States engage in the war and for Israel to remain entirely uninvolved: and that is indeed what happened.   Israel Forces the USA to War   However, the Israelis were actively involved in the United States\' war plans. Some US staff officers involved in planning Operation Desert Storm received fine tactical advice from the Israelis that \"the best way of wounding Saddam was to strike at his family.\"   The Mossad-inspired propaganda campaign reported by Ostrovsky set up the necessary public backing for the Gulf War. It was again Mossad local assistants who lit the touchpaper for the war. The Hill and Knowlton lobbying firm, run by Tom Lantos of the Israeli lobby, prepared a dramatic scenario to convince members of Congress on the subject of war against Saddam. Turan Yavuz, a noted Turkish journalist, describes the incident:   October 9, 1990. The Hill and Knowlton lobbying firm organizes a sitting in Congress on the subject of \"Iraq\'s Barbarities.\" A number of \"eye witnesses\" brought to the session by the lobbying firm maintain that Iraqi troops killed new-born babies in the hospital wards. One \"eye witness\" describes the savagery in enormous detail, saying that Iraqi soldiers killed 300 new-born babies in one hospital alone. This information deeply disturbs the members of Congress. This works to President Bush\'s advantage. However, it later emerges that the eye witness brought by Hill and Knowlton to Congress is in fact the daughter of the Kuwaiti ambassador to Washington. Nevertheless, the daughter\'s account is sufficient for members of Congress to give Saddam the nickname \"Hitler\". 6)   This leads to just one conclusion: that Israel played an important role in the United States\' to wage its first war on Iraq. The second one is not much different.   The Pretext of \"War Against Terrorism\"   Contrary to popular belief, the plan to attack Iraq and overthrow Saddam Hussein\'s regime by force was prepared and placed on Washington\'s agenda long before the genesis of the \"war against terrorism,\" which emerged in the wake of September 11. The first indication of this plan emerged in 1997. A group of pro-Israeli hawkish strategists in Washington D.C. began to put forward the scenario of the invasion of Iraq by manipulating the \"neo-con\" think-tank, called PNAC (Project for The New American Century). The most notable names in the PNAC were those of Donald Rumsfeld and ###### Cheney, who as defense secretary and vice-president would be the most influential figures in the George W. Bush administration.   An article titled \"Invading Iraq Not a New Idea for Bush Clique: 4 Years Before 9/11 Plan Was Set\" written by William Brunch and published in the Philadelphia Daily News, sets out the following facts:   But in reality, Rumsfeld, Vice President ###### Cheney, and a small band of conservative ideologues had begun making the case for an American invasion of Iraq as early as 1997 - nearly four years before the Sept. 11 attacks and three years before President Bush took office.   An obscure, ominous-sounding right-wing policy group called Project for the New American Century, or PNAC - affiliated with Cheney, Rumsfeld, Rumsfeld\'s top deputy Paul Wolfowitz and Bush\'s brother Jeb - even urged then-President Clinton to invade Iraq back in January 1998. 7)   Is Oil the Real Objective?   Why were the PNAC members so determined to attack Iraq? The same article continues:  While oil is a backdrop to PNAC\'s policy pronouncements on Iraq, it doesn\'t seem to be the driving force. [Ian] Lustick, [a University of Pennsylvania political science professor and Middle East expert,] while a critic of the Bush policy, says oil is viewed by the war\'s proponents primarily as a way to pay for the costly military operation.   \"I\'m from Texas, and every oil man that I know is against military action in Iraq,\" said PNAC\'s Schmitt. \"The oil market doesn\'t need disruption.\"   Lustick believes that a more powerful hidden motivator may be Israel. He said Bush administration hawks believe that a show of force in Iraq would somehow convince Palestinians to accept a peace plan on terms favorable to Israel- 8)   This, therefore, is the principal motivation behind the plan to attack Iraq: to serve Israel\'s Middle East strategy.   This fact has also been identified by other Middle East experts. Cengiz ?andar, a Turkish Middle East expert, for instance, describes the real power behind the plan to attack Iraq thus:   ... Who is directing the attack on Iraq? Vice-President ###### Cheney, Defense Secretary Rumsfeld, National Security Adviser Condoleeza Rice. These are the \"senior level\" backers of the attack. Yet the rest of the iceberg is even richer and more interesting. There are a number of \"lobbies.\"  Heading these lobbies are the Jewish Institute for Security Affairs team, pro-Likud and Israeli-right and known for their close relations with US arms manufacturers. These have close relations with the \"arms lobby,\" Lockheed, Northrop, General Dynamics and Israeli military industries ... JINSA\'s fundamental principle is this: America\'s and Israel\'s security are inseparable. In other words, they are the same thing.   JINSA\'s objective is not solely the overthrow of the Saddam regime in Iraq: it also supports the overthrow of the Saudi Arabian, Syrian, Egyptian and Iranian regimes with a logic of \"total war,\" and the subsequent importation of \"democracy.\" ... In other words, a number of American Jews on the same wavelength as the most extreme factions in Israel at the moment comprise the hawks in Washin\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0  '),(9042,4201404,'11,August,2004','2004-08-11 00:00:00','   WAIT FOR THE NEXT ISSUE  '),(9043,4201404,'11,August,2004','2004-08-11 00:00:00','   bvfcvjhbjk  '),(9044,4201404,'11,August,2004','2004-08-11 00:00:00','   The Untold StoryBackgroundChechens are one of the few peoples indigenous to the Caucasus Mountains. Chechens are one of the few peoples who are indigenous to the Caucasus Mountains. They are ethnically very different from the Slavic Russians. Even linguistically the Chechen language is in the class of languages called Ibro-Caucasian, making English, German, and French a lot closer to Russian than Chechen. More importantly, all Chechens are Muslims who have throughout their history resisted Russian occupation. Basically there are no historical or legal reasons for Russia to claim Chechnya part of the Russian federation. The barbaric ethnic cleansing that is going on today in Chechnyais the latest episode of a war that has lasted for over 300 hundredyears. Chechens have never in their history joined any type of union or signed any piece of paper entitling Russia the right torule their land. Going as far back as the 16th century, Russia has repeatedly practice ethnic cleansing against the Chechens and therest of the north Caucasian Muslim nations. In the first Chechen war of 1994-1996, 12% of the Chechen population perished, that is about 120 thousand lives, almost all civilians. And that is no accident. It is typical of the way the Russia wages war. In 1944, the entire Chechen nation was exiled from its land along with the majority of the Muslim nations of the north Caucasus.During this forced migration 60% of people perished from cold, hunger and diseases. The idea was to dissolve the Chechen identity altogether. Simply put, for as long as there is a Chechen living in the north Caucasus, Russian politicians will not be free to rule the region. During this close to 400 hundred years of war, the Russians have never fought a clean war. They have repeatedly targeted the Chechen civilians, only because they have never won face-to-face combat against the Chechen freedom fighters, by Allah\'s grace that will never happen. On the other hand the Chechens have always been very humane to their captives. In the last war scores of POWs caught by the Chechens embraced Islam. Many more young conscripts were returned to their mothers. However, no mercy was ever shown to the mercenaries or military pilots that killed civilians. At the end of the 20-month-old war the Russian army pulled out of Chechnya dragging its tail between its legs.The last war has left Chechnya and its people crippled. What little industry was there is totally destroyed. Utility services are nonexistent. Hospitals lack everything from medication to equipment. Water is contaminated. A great percentage of the survivors are handicapped. Infant mortality rate is probably one of the highest in the world. In other words Chechnya and its population are in no state to wage war against so called a super power. The Russian FSB has been busy waging a propaganda war against the Chechen people. The kidnappings were started in order to make sure the outside world would stay far away from Chechnya. Members of humanitarian organizations that dared enter Chechnya were brutally dismembered by orders of the Russian FSB.Russian government blew up apartments in Moscow killing about 300 Russian civilians, then turned around and blamed it on the Chechens. All this was done to justify invading Chechnya. The way the apartment bombings were carried out required a high expertise in the matter of bringing down buildings. This is a rare expertise, which the Chechens never had. Had the Chechens had that expertise they would have demonstrated it against the Russian army by now.The Kremlin needs this war as a tool to make sure a president sympathetic to the current regime would win the next elections. Mr. Putin seems to fit that bill. More than likely he has promised that Yeltsin and his gang would not be prosecuted for stealing billions of dollars. Many, including some in the Russian media itself, have theorized that the Russian government is behind the apartment bombings. This is the enemy the Chechen people are facing. An enemy that does not even have mercy on its own civilians let alone the Chechen civilians.Our brothers and sisters in Chechnya and many decent freedom loving citizens of the world are wondered why no one is forcing the Russians to stop the bombing campaigns. All they see is destruction and mass murder. They do not see \"rebels\" hiding out or careful efforts by the Russians to avoid civilian targets. They watch as village after village is decimated indiscriminately. Most do not understand how this can be happening to them again. Only this time, worse. We urge you to respond to their cries.THE DESTRUCTIVE INVASION OF WAR IN CHECHNYA WHAT TO DO ?We ask all readers of this booklet to please follow this stepwise approach to face this very sad and disturbing human tragedy1. PrayÂ· Pray to Allah that He will help the Chechen innocent citizens to overcome their miseries, that He almighty will give them victory, and that He shows the mislead young members of the Russian army the truth in order to change their hearts to Islam.Â· Please ask the Imam of your masjid to include the Unfair Destructive War in Chechnya as the main issue of the Jumaa Prayers Ceremony (Khutba). Â· Hold a special \"prayer of need\" (Salatul Hajah) that is recommended in times of distress and great needs.Â· Remember the Chechens in your du\'as. Supplicate for their well being. Ask Allah (SWT) to grant them strength, ease their difficulties and bring them out of this trial successfully. Â· Pray that the oppressor be punished. 2. DonateWe urge you to donate generously. Your donations will help us provide necessary humanitarian relief to the affected needy families and refugees through the services of accredited and experienced international Muslim relief agencies. 3. Educate & communicate Spread the word to and discuss the problem with Muslims and non-Muslims. Discuss the Chechen point of view and the right of the Chechen people to live in peace. Share the atrocities committed by the invading forces of the Russian army.Statement of Mujahideen Leadership from Grozny All Praise is to Allah. Prayers and blessings are upon His Messenger.Firstly: We thank all of our Muslim brethren everywhere who are with us in heart and spirit. The ones who provide us with supplications and support which undoubtedly affect the course of fighting.Secondly: We want to assure all the brethren that our current military situation is good and there is nothing to worry about. We are happy to inform them that the Mujahideen have stood up to many fierce Russian Army attacks. The current state of the Russian military is that of desperation and confusion. Care about civilian life is all but gone. The Russian Army forces started using ultra heavy artillery (1+ Tons) in the bombardment. The indiscriminate bombardment caused major losses among the unarmed civilians and huge destruction to the capital. Regarding the statement that the \"Mujahideen should not sacrifice all for the sake of non-strategic towns for life is more valuable that land\". As we previously stated, yes, we completely support this statement. We did not declare Jihad except to comply with the orders of Allah. We travel the path that pleases Allah. However, the reason we continue to defend the capital, which is of no strategic value to the Mujahideen, is that we wanted to prove to the world and the Russian Political and Military establishment that despite the size, power, or technology of an enemy, there is no way that they could defeat the people of belief, principle, and land. By protecting the capital for this long, we have broken the arrogance of the Russian Army and politicians. We caused them tremendous damage and broken the spirits of their soldiers and officers. This alone is considered a victory. We also repeatedly humiliated the Russian military leadership, who on many occasions, promised dates of entering the capital. The last of which was the first of January. Yet, to this date, they have not succeeded. Putin promised the world that the Russian flag would be waiving in Grozny by the beginning of the year. So, where are those flags of yours, Mr. Putin? Due to this failure, his credibility has plummeted among the Russian leadership. The ones who are anxiously waiting for something to happen so that they can save face in front of the world. By maintaining the capital we\'ve obviously embarrassed them. This has resulted in the politicians loosing faith in their own military. The military whose defeats date back to Afghanistan, where they left humiliated.Thirdly: As we convey these points to our brethren, we have to bring their attention to an extremely important matter; When we made the decision to protect the capital, the ultimate goal of protecting the Religion, civilian life, and the Mujahideen did not escape our minds. When we contemplate a decision, we do our very best to base the final decision by weighing the pros and cons. Everyday, we continue to judge the benefits and prices resulting in maintaining the decision. We will ultimately continue what we believe pleases Allah and has benefits to the religion and Muslims regardless of the cost. If this results in leaving the capital, we will. We are not better than the companions fighting in the battle of Badr when they left Mu`tah. When Khalid led the Muslims in their departure, the messenger, prayers and blessings be upon him, was happy and considered it a victory. On the other hand, if we need to stay and protect the capital, then our best example is the Messenger of Allah, prayers and blessings are upon him. He believed in the victory and the promise of Allah where he -peace and blessings be upon him- led 300 men in fighting 1000 who also exceeded them in weaponry. As a result, regardless whether we go with the first or second option, our Religion is kind and merciful. Praise is to Allah who did not make our Religion difficult. This leaves us with the remaining variables, which are monetary, military, and political. The leaders in charge of defending the capital know the most about the variables and therefore are best to evaluate the situation. Thus, we want to ensure our brethren that we will not stop supporting what we believe is best for religion, life, and the Muslims. We continuously ask Allah and request that our brothers ask Allah to guide us to what is best for our religion and life. Finally: We request and urge all Muslims to stop listening to and ignore all Russian Political establishmentâ€™s propaganda and the western media who try to strengthen the spirits of the Russian Army with lies and fabrications of victories that never took place. If we decide to leave the capital, we will announce it via our own official sources. We ask Allah to raise his religion, grant victory to his supporters, and bestow defeat upon his enemy. Allah knows best and prayers and blessings are upon our Prophet Muhammad, his family, and all of his companions. Six Points Action Plan Recommended to the Muslims Worldwide1. Follow the authentic news of the Jihad in Chechnya from authentic sources. 2. Spread authentic news about what is happening in Chechnya to Muslims all over the world, through word-of-mouth, Friday sermons, leaflets, posters, friends in the Muslim and non-Muslim media, radio, TV, e-mail and the Internet. Media is an important part of this war.3. Boycott Russian companies and goods. Enforce all types of sanctions against Russia. 4. For the one who is able, to travel to Chechnya to give physical and financial assistance to their suffering Muslim brothers and sisters. One should NOT travel to Chechnya until he has definite contacts with Chechen groups, organizations, and centers or Islamic charitable organizations, which can confirm 100% that a safe route exists into Chechnya. All visas must be obtained prior to travel. Once inside Chechnya, one must join one of the many existing groups helping the afflicted Muslims in the area. 5. Lobby relatives, friends and contacts in Muslim governments to break off diplomatic relations with Russia, expel their ambassadors and shutdown their embassies, unless Russian unilaterally withdraws from Chechen soil. 6. At the very least, supplicate to Allah to give Victory to the believers and destroy the oppressors.THE CHRONOLOGY OF CHECHNYA-RUSSIA CONFLICT1. In 1864, Chechens, after a long and bloody struggle for freedom, surrendered to Tsarist Russia. 2. In 1877, Chechens rose up in arms against Russia but were subdued. 3. In 1901, about 700 families migrated from Chechenya to Turkey. Of this group, about 200 families returned to Chechenya, while 300 families settled in Turkey and 200 families settled in Jordan. 4. After WWI, the political and nationalist activists of the Republic of the North Caucasus Federation moved to Poland to escape Communist oppression. Upon signing the agreement, by Hitler and Stalin, to divide Poland between Nazi Germany and Russia these activists moved to France where their descendents live today. 5. After WWII, Chechen members of the Red Army who were prisoners of war in Nazi Germany refused to return to Chechenya since the Chechen nation had already been exiled in 1944 to Siberia by Stalin. These Chechens settled in Germany, Turkey and the U.S.A. 6. On May 11, 1918, the North Caucasus nations declared the formation of the Republic of the North Caucasus Federation. Germany, Ottoman Turkey, Austria, and Bulgaria in addition to Bolshevik Russia recognized this republic. 7. In 1922, in a reversal of its recognition, Bolshevik Russia invaded and occupied the Republic of the North Caucasus Federation. 8. Chechens rebelled against Russia in 1929 and again in 1940. Both times they were subdued in a bloody manner by the communist regime. 9. On Feb 23, 1944, the whole Chechen nation was exiled to Siberia by orders from Stalin. 10. After thirteen years of exile, the remaining Chechens were allowed to return to their ancestral homeland by Khrutchev in 1957. 11. Through the 70\'s, Soviet Russia treated Grozny (capital of Chechenya) as enemy territory, where nighttime curfews were imposed routinely. 12. On Nov 23 - 25 1990, a national Chechen conference convened in Grozny. The conference included delegates representing all ethnic groups of Chechenya, including Russians and Cossacks. 13. On Nov 25, 1990, the delegates to the conference declared the independence and sovereignty of Chechenya and its cessation from the former Soviet Union. 14. On Nov 27, 1990, the declaration of independence and sovereignty was unanimously ratified by the Soviet Parliament of the Chechen-Ingush Republic. 15. On Oct 27, 1991, in fulfillment of constitutional requirements, a national Chechen referendum was held. The Chechen nation approved the declaration of independence and elected General Dzhokar Dudayev as the first elected president of the independent Chechen Republic by a margin of 84% from among three contenders for presidency. 16. The former Soviet Union was dissolved at midnight Dec 31, 1992. 17. The former Russian Soviet Socialist Republic was dissolved and replaced by the Republic of the Russian Federation upon signing the federation treaty on Mar 13, 1992. 18. By the free will of the Chechen nation, Chechenya did not sign the treaty or join the federation. 19. Dec. 11 1994: Russian troops invade Chechnya with two tank units; the Chechen-war begins without Russia ever declaring war. 20. Aug 1996: Aslan Maskhadov, a former Soviet army officer who led Chechen rebel forces in the war, negotiates a peace agreement with Moscow\'s Alexander Lebed that provides for the pullout of the Russian troops and postpones an agreement on Chechnya\'s constitutional status until 2001. 21. Jan. 27, 1997: Aslan Maskhadov clearly wins elections and becomes Prime Minister in Chechnya\'s separatist government.22. Feb. 12, 1997: Chechnya\'s new president Aslan Maskhadov takes the oath of office under heavy security, pledging to strengthen the self-proclaimed independence of his Muslim republic. 23. May 12, 1997: Boris Yeltsin and Aslan Maskhadov sign a peace treaty which leaves open the question of independence, but confirms that Russia will acknowledge the norms of international law and not use force to settle disputes. Chechnya keeps the Russian ruble as currency. The treaty also clears the way for commissioned use of Chechyan oil pipelines by Russia.Council of Europe Must Act on Chechnya(Strasbourg, January 25, 2000)--Human Rights Watch today called on European parliamentarians meeting in Strasbourg to take a tough stand on Russia\'s war in Chechnya. \"The Council of Europe\'s top priority must be getting a sustained international monitoring presence in Ingushetia,\" said Peter Bouckaert, Human Rights Watch\'s representative in the northern Caucasus, who traveled to Strasbourg to press his case with members of the Council of Europe\'s Parliamentary Assembly holding their quarterly meeting this week. \"Having people from the international community on the ground to bear witness to what is happening is the best means of getting some immediate relief. If the Russian Federation is serious about living up to its commitments to the Council of Europe, it must agree to such scrutiny.\"Rape Allegations Surface in Chechnya(Nazran, January 20, 2000)--Russian soldiers have been raping Chechen women in areas of Russian-controlled Chechnya, Human Rights Watch charged today. A team of Human Rights Watch investigators in Ingushetia has spoken to a number of Chechen refugees who saw the bodies of victims, or had conversations with women who had been raped.\"Rape is a war crime, and these allegations about rape in Chechnya are very serious,\" said Regan Ralph, executive director of the Women\'s Rights Division of Human Rights Watch. She stressed that according to international law, any single rape could be a war crime.Human Rights Watch researchers in southern Russia have learned that thousands of elderly, poor, and infirm civilians remained trapped in Chechen combat zones. Civilians who have managed to flee the region report that bus drivers, border guards, and corrupt officials are charging exorbitant prices for persons seeking to flee the fighting, forcing the most underprivileged persons to remain behind. They are trapped between Chechen fighters and advancing Russian troops, are unable to escape, and have no access to running water, electricity, heating, or food supply. Most essential services in Grozny and other large towns have ceased functioning.\"This kind of predatory extortion for compliance with residence permits is bad enough in peacetime as a violation of freedom of movement,\" commented Holly Cartner, executive director of Human Rights Watch\'s Europe and Central Asia division. \"But these are helpless civilians fleeing armed conflict. The Russian government has a legal obligation to protect them.\"A Reminder from Al-Qur\'an and SunnahOur beloved Prophet (peace be upon him) said: â€œA Muslim is a brother of another Muslim, so he should not oppress, nor should he hand him over to an oppressor. Whoever fulfilled the needs of his brother, Allah will fulfill his needs; whoever brought his brother out of a discomfort, Allah will bring him out of discomforts on the day of resurrection, and whoever screened a Muslim, Allah will screen him on the Day of Resurrection.â€ (Sahih Bukhari)â€œWhatever you spend in the way of Allah, it is for your own good, when you spend not for anybodyâ€™s pleasure except that of Allah. Thus whatever you spend for any good work you will get full benefit of it and you will not be wronged to the slightest extent.â€ (Al-Qurâ€™an 2:272)â€œAnd in their wealth have due share of those who ask for help and those who are deprived.â€ (Al-Qurâ€™an 51:19) him.â€œThose who spend their property by night and by day, secretly and openly, they shall have their reward from their Lord and they shall have no fear, nor shall they grieve.â€ (Al-Qurâ€™an 2:274)Brothers and Sisters, both Al-Quran and As-Sunnah have urged us to spend in the cause of Allah (SWT). Chechnya is bleeding. Your support is urgently needed. We have diverted our resources towards the people of this war torn region. Please donate today and strengthen our relief efforts in Chechnya.ChechnyaTerrifying Facts Â· Campaigns of terror since 1994 by the Russian Army and Political Leaders have resulted in the shrinking of Chechan population from 1.2 million to only 800,000.Â· Since October 1999, more than 300,000 Muslims have fled Chechnya.Â· Tens of thousands are displaced internally.Â· More than 40,000 children have became orphans.Â· Attacks have resulted in civilian deaths and injuries.Â· 82% of schools have been destroyed.Â· 3 universities have closed down.Â· Land and water are contaminated.Â· Unemployment was 89% before the current war.Â· There is no heat to keep people warm in the frigid winter seasonfrom www.yahoo.comI don\'t see that somebody talk about Chechnya !! Why ??When something happend there, they say a Checen(Muslim) rebels attack Russian civilians..I would like to ask those people what about hundreds of thousands Checen people who died ?  '),(9045,3682262,'04,July,2004','2004-07-04 00:00:00','   Hi! Well if you haven\'t guessed already I did NOT write that last post! haha. Beth Collins is very sneaky and changed my password to make an entry. lol So I\'m  not  a lesbian w/ Cassie and I think Tommy Hunt is  very nasty . (just for the record! haha) Oh and something else for the record- Beth doesn\'t get me things to eat &/or drink when I need them. She says- go get it yourself! Now that everything is cleared up I have to tell y\'all that Evan is not exactly single anymore. (sorry ladies, haha) He has a \'long-distance lova\'! haha No more details for you! Well I just got off the phone with Beth and she\'s gonna try to be sneaky again, so any entries withiin the next 17 days are not me! (obviously, cuz I\'ll be in Canada... then NYC!) alrighty well I\'m leaving tomorrow morning and I haven\'t even started packin (bad news! haha) so I got to jet! Bye!   '),(9046,3682262,'04,July,2004','2004-07-04 00:00:00','   Hey everybody! Okay, so I\'m Laurel Mallory and I love Beth Collins more than life itself. She is my everything. I wish I could see her all the time, but unfortunately, I\'m going to Canada and then New York so I\'m gonna be without her for like a month. She\'s my best friend. What in the hell would I do without Beth? She does everything for me. She does all my favors for me. When I need something to drink, who goes to get it? Beth does. When I\'m hungry, who goes to get me food? BETH DOES! Who calls me when I\'m sad? BeTh DoEs! She has a toothbrush at Brandon\'s house. How cool is she?! VERY! Okay well I just thought I\'d let the whole world know how much I love Beth. And how much Cassie Brummet and I have lesbian tendencies. Yes. We have phone sex. She\'s my LUVVAAA! I love Tommy Hunt, too. He loves to give me lovin\'. He\'s my MAN LUVVAAA! Alright, time to leave. THESE STUPID BLACK PEOPLE DONT KNOW HOWTA DANCE!!! Love ya! *Laurel Fucking Mallory!*   '),(9047,3682262,'04,July,2004','2004-07-04 00:00:00','   Well I was talking to Evan earlier bout this whole blog thang and he was like is there anything about me in it? and i was like no... so I\'m makin him happy now haha. (plus, I think this is the only way to get him to ever look at this haha) Well he\'s like my best guy friend and we\'ve been friends for about 2 years I reckon. His middle name is James and his last name is Dusynzki. (I don\'t think I spelled that right, I never do- I\'m awful, I know!) Some ppl call him Ed for short... but I don\'t! I call him Evie, haha. He\'s the same age as me (15) and in my grade (10th). He has blonde hair and he\'s like um, not real tall but not short. (I\'m guessing 5\'7\"?) He likes playin sports (soccer, basketball...) and of course he like watchin them on tv. He\'s a good guy- nice, cute, smart, alla that good stuff. (And he\'s single unless there\'s something he\'s not tellin me so go for it girls. haha) Okay, I hope he\'s happy now cuz this took me a while and I\'m gonna go now! Choi! (oh and his AIM sn is Ed407 and he\'s gonna kill me for puttin this on here but I think it\'s funny, so IM him!)   '),(9048,3682262,'04,July,2004','2004-07-04 00:00:00','   This is what I put on my profile about Brandon:   Brandon Cutwright LOVES Joey Brown!!! (bjcballer)- Brandon, (JoeyDBrown)- Joey! bjcballer: i give head for quarters, i give joey head for free, but i gotta charge for everybody else so tell everybody u know, im tryin to make money  okay, I\'ll tell you about Joey Brown- he\'s this guy from Alabama and Brandon talked to him on the internet a lot and thought he was really cool... well one day Joey got on and they were talking bout doing stuff with girls and then Joey told Brandon that his friend Jimmy gave him head the other night and \'it was good\'. Brandon was like is Jimmy a girl? and Joey was like no. and Brandon was like eww and Joey was like \'don\'t knock it til you try it\'. lol freakin hilarious kids, freakin hilarious... Brandon tried getting back at me with this but really now- no where close:  definiton of prostitute:  Person who sells there body in exchange for money so they can make a living.  Synonyms: Laurel Mallory  OMG I just remembered that my brother almost killed me n Beth n Meagan 2day. we went to help my Dad with his car club\'s car show in Augusta and we left with Dean when he went to Maysville to get Meagan... well on the way home we were on this real curvy road and we were goin like 100 i swear and i see this sign that says to go 30 m.p.h. around this curve and i\'m thinkin holy shit we\'re gonna die! it was scary... but fun i\'ll have to admit. lol   well i\'m done. you\'ve just heard about another exciting nite in the life of Laurel Mallory. hasta luego.   '),(9049,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    Well this is me on my roof. I decided to do my bit of writing for today with this picture. Beth ate a dead spider... see the ladder in the pic? Well she climbed up it (only to go back down it- scaredy cat) and a dead spider and it\'s web went into her mouth while she was eating her popsicle! It\'s 11:11 make wishes! Brandon is putting obscene things on his profile and it\'s making me just a lil bit angry... lol look: Laurel\'s top 3 things: 1.Girls 2.Girls 3.Girls oh and did i mention girls, dont believe me, look at this sexynsweet0719: i love girls, i really do. but dont say nothin\'&nbsp; urlLink   '),(9050,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    this is George... he lives at my house (this was taken in my backyard) and he\'s the best pet i\'ve ever had!&nbsp; urlLink   '),(9051,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    me n beth a lonngggg time ago...&nbsp; urlLink   '),(9052,3682262,'04,July,2004','2004-07-04 00:00:00','   I just wanted to let y\'all know that it\'s gone! I\'m better! And I\'m so happy and I dunno how I realized it was gone but it is definately gone. And well- no more weird feeling in the tummy. I feel good. I feel very proud of myself and it actually took me by surprise but I reckon I\'m stronger than I thought. yay! :)  very yay . You do not understand how yay this actually is. I hope it lasts. I really really really do.           *hint: this has to do w/ the stupid boys of my life   '),(9053,3682262,'04,July,2004','2004-07-04 00:00:00','   I made a strange discovery 2day (well I think that i\'ve known for a while, but oh well) my arm hairs are out of control! lol I swear, they like won\'t even like lay flat on my arm they\'re just all weird. I thought I\'d share that with y\'all. And, um... I don\'t think I have anything else to say. \'night.   '),(9054,3682262,'04,July,2004','2004-07-04 00:00:00','   Well I\'m sitting here with Beth and Liz and we\'re kinda bored. I made a good discovery today- I learned how to make my cell phone (um... George, that is...  shhhhh ) say \'Laurel\' instead of \'713481481848 units\'. ya know. Well you prolly don\'t but oh well! Beth says-  \"I know how you make babies!\"  haha she\'s so dumb. We walked into the wonderful city of Berry earlier!- we needed pop and everywhere\'s closed, so we went to the pop machine with like $6 in change haha. \'my parents don\'t remember my first name... I mean my first word\' haha I can\'t talk. (that\'s not news) A piece of advice for anyone who\'s sunburn is itching: lie on the floor (tummy side up!) and like move your back over the carpet and it feels good! haha Beth got that from Tyler and I think it really works. The Max is 4?!  ...AnYwAyS...  I went to Dwayne\'s visitation today. It wasn\'t that bad I reckon. I didn\'t know him that well though... I dunno what I would do if he was really close to me. I just think if that was one of my really good friends... god- it\'s crazy! It\'s gotta be shit for his family and all his real close friends, I\'m praying for them. It can happen to anyone and that\'s a  really  scary thought. But I think it opened up a lot of people\'s eyes in so many ways. For me and Beth and Jess it\'s got us to be nice... We\'ve been bitches to a lot of people lately and we\'re trying to fix that. I think we\'ve all decided on our to own to be nicer to people and apologize for things we\'ve done in the past. I don\'t want to not be able to fix things, if ya know what I mean? I dunno... It doesn\'t really seem real that he\'s dead. I mean I still kind of expect to see him driving around places and stuff and I\'m not. It like hasn\'t sunk in I reckon. I know things like this are going to happen- and might hit me a little closer to home next time... I\'m praying they won\'t. I\'m praying.   '),(9055,3682262,'04,July,2004','2004-07-04 00:00:00','   Ello loves. I\'ve gone through some drama recently. well... i stayed on here \' toooo  late\' the other night, and um, my dad took the keyboard away. (and the mouse) for like a day. wooo big freakin deal! please. well i\'m back! haha and better than ever i reckon... i got a $5 shirt at JC Penny 2day! how exciting! haha earlier today i realized that what i  really  wanted to do, like more than a lot of things, was to go muddin\'. how fun would that be?! i mean it\'s like 4-wheelin... but in a truck. seriously- how could anyone think that wouldn\'t be fun? maybe pointless... but still fun! (*hint* you should be making a \'note to self\' \"laurel REALLY wants to go muddin\"! haha) hmm... what rhymes w/ muddin\'? PUDDIN\'! lol it kinda does anyways...? i like callin ppl puddin\'- is sounds so countryish (but in a good way, i think) like callin them darlin\'. it\'s so fun. lol anyways- ya know i\'m gettin pretty tired of all my \'rents rules. they\'re not normal, i swear! i think they want me to be like one of those girls who sit at home all day and read... and who doesn\'t have friends, or a reason to use the phone, or get on the internet... or to go to cynthiana! haha i know they\'re just trying to discipline me (which- if i like it or not- is a good thing) but they\'re pushing it a lil farrrr in my book. but what do i know? i\'m just a lil kid. haha shit. \'night!   '),(9056,3682262,'04,July,2004','2004-07-04 00:00:00','   this is a conversation between beth collins and me. um... let\'s just say that i don\'t really know why i\'m putting this on here because it makes me look like a complete retard. (i have edited some things that you don\'t need to know- i jus put in a star) this is me hyper:  sexynsweet0719: ello ***** sexynsweet0719: HE\'S AN ASSSSSS ***** AngelBeth2309: he\'s my buddy:-( ***** sexynsweet0719: well well well AngelBeth2309: what was your away message? lol *** AngelBeth2309: ohhh lol sexynsweet0719: yea sexynsweet0719: can i ask you something ? AngelBeth2309: yes * AngelBeth2309: i\'m not sure AngelBeth2309: thats my problem... sexynsweet0719: yea sexynsweet0719: well i think you can if you try haha sexynsweet0719: i have faith in ya! AngelBeth2309: i think i\'m gonna try AngelBeth2309: but when you call, i\'ll tell you everything sexynsweet0719: alrighty sexynsweet0719: take my new poll AngelBeth2309: lol k i took it sexynsweet0719: thank ya AngelBeth2309: goshhh sexynsweet0719: i have hair as long as a towel sexynsweet0719: it\'s very thick and i like to hold it sexynsweet0719: WAIT sexynsweet0719: IT IS A TOWEL! sexynsweet0719: wowwwwwwwwowoowowowoow  sexynsweet0719: dunno how it got there sexynsweet0719: this is weird sexynsweet0719: i have a towel on my head sexynsweet0719: nonononono AngelBeth2309: lol RETARD AngelBeth2309: did you take a shower? sexynsweet0719: lol ummmmmmmmm yea i think so sexynsweet0719: ohhh koa y it makes sense now sexynsweet0719: okay* AngelBeth2309: hahaahahah AngelBeth2309: dude, have you ever heard daylight fading by counting crows? sexynsweet0719: nope i have not AngelBeth2309: its the best song ever sexynsweet0719: i shall look it up AngelBeth2309: i LOVE it sexynsweet0719: alrighty * sexynsweet0719: good good sexynsweet0719: it\'s looks good sexynsweet0719: does it sound good? AngelBeth2309: lol very AngelBeth2309: i LOVE ITTTTTT sexynsweet0719: awesome AngelBeth2309: yes AngelBeth2309: im re-signing your gbook AngelBeth2309: i think you needa sign mine sometime... sexynsweet0719: alrighty i shall but i\'m signin jason\'s now and i signed jess\' earlier and i think i\'m all signed out sexynsweet0719: for 2day AngelBeth2309: FINE THEN sexynsweet0719: lol i\'m sorrry i gosta get off in sexynsweet0719: like 3 secs AngelBeth2309: fine, i\'m not signing yours sexynsweet0719: well geez i\'m gonna get into trouble beth AngelBeth2309: lol into trouble sexynsweet0719: yes AngelBeth2309: well are you allowed to cal lme? sexynsweet0719: that doesn\'t matter i do all the time anyways sexynsweet0719: druh AngelBeth2309: true sexynsweet0719: geez i can\'t type \'what did you do\' without thinking about it WAY more than i should sexynsweet0719: have to AngelBeth2309: who are you talking tO/ * AngelBeth2309: no i mean with the what did you do thing * AngelBeth2309: haha sexynsweet0719: i was like sexynsweet0719: what did you did sexynsweet0719: what did you you sexynsweet0719: what did you you did do sexynsweet0719: it was awful AngelBeth2309: what did you doi AngelBeth2309: do* sexynsweet0719: huh? lol * AngelBeth2309: Mista Clean 23 (11:25:55 PM): who you like thse dayz AngelBeth2309: good GOSH sexynsweet0719: yup he asked me the exact same thing sexynsweet0719: i was like sexynsweet0719: : sexynsweet0719: sexynsweet0719: i have no clue sexynsweet0719: lol do you like askin that question? Mista Clean 23: im trying to make convosexynsweet0719: i have no clue sexynsweet0719: lol do you like askin that question? Mista Clean 23: im trying to make convo sexynsweet0719: oopps i pasted that twice sexynsweet0719: didn\'t i? AngelBeth2309: SO ANNOYING AngelBeth2309: lol yes you did sexynsweet0719: lol sexynsweet0719: i\'m like really hyper for some reason sexynsweet0719: crazy sexynsweet0719: my heart hurst AngelBeth2309: lol hurst? sexynsweet0719: hurts* sexynsweet0719: geez ya know what i was gonna say AngelBeth2309: lol huh? AngelBeth2309: what? sexynsweet0719: you knew what i was gonna say* sexynsweet0719: what i meant to say* sexynsweet0719: gosh i can\'t talk i can\'t type i need help sexynsweet0719: NOW AngelBeth2309: lol im\' not sure if there\'s much help for you sexynsweet0719: lol i don\'t think there is :-( AngelBeth2309: lol thats right sexynsweet0719: what font do you  type in? AngelBeth2309: tahoma sexynsweet0719: tell me please sexynsweet0719: lol okay AngelBeth2309: lol i told you retard sexynsweet0719: i know it sexynsweet0719: ohhhhhh i mean okay sexynsweet0719: thanks sexynsweet0719: look at my profile AngelBeth2309: ok RETARD lol sexynsweet0719: lol it\'s okay  sexynsweet0719: ithink sexynsweet0719: ay? AngelBeth2309: lol ay? sexynsweet0719: yes ay sexynsweet0719: i like sayin ay sexynsweet0719: is that a problem? AngelBeth2309: lol wetawd sexynsweet0719: tawd? AngelBeth2309: A BIG FAT PROBLEM sexynsweet0719: haha AngelBeth2309: lol yes say it out loud AngelBeth2309: i am foo king wee tawd AngelBeth2309: i am a* AngelBeth2309: i\'s a foo king wee todd AngelBeth2309: THERE WE GO!!! sexynsweet0719: lol i am sexynsweet0719: i know sexynsweet0719: you are too  AngelBeth2309: lol :-) weeeee AngelBeth2309: GET OFF OFFLINE SO I CAN TALK TO YOU ON THE PHONE AngelBeth2309: NOWWWW BIAAAAATCH sexynsweet0719: i\'m putting this conversation on my blog sexynsweet0719: haha sexynsweet0719: like now sexynsweet0719: so it\'ll be a sec AngelBeth2309: lol omg ok sexynsweet0719: okay so i\'m gonna post, go, then call AngelBeth2309: well tell me when you post so i can read sexynsweet0719: lol kkkkkk AngelBeth2309: YOU\'RE RACIST! TIMES 2... YOU SAID KKK-KKK AngelBeth2309: you stupidhead sexynsweet0719: well i\'m sorry! sexynsweet0719: very sorry! sexynsweet0719: i\'m not racisty sexynsweet0719: lol racist* AngelBeth2309: lol racisty AngelBeth2309: dude... major typing lessons... like seriously sexynsweet0719: lol i know *  that was it. okay that\'s it kids. (jessika just ate BAD food, don\'t let her lie)   '),(9057,3682262,'04,July,2004','2004-07-04 00:00:00','   i\'ve been mowing since i woke up. well actually it shouldn\'t have taken me that long but i was stubborn and kept taking breaks. ya know there\'s 12 days until i go to Canada and NYC. i won\'t be able to write in this thing from july 1-17! wow. that\'s a hella long time. i think i\'m getting fat. i haven\'t exercised in over a week. bad. very bad. i guess i\'m just lazy!   '),(9058,3682262,'04,July,2004','2004-07-04 00:00:00','   okay... so i just fell in love with Dashboard Confessional. you should, too. here\'s a sample:  Brilliant Dance  by Dashboard Confessional    So this is odd, the painful realization that all has gone wrong. And nobody cares at all, and nobody cares at all.  So buried all your lovers clothes and burnded the letters lover wrote, but it doesn\'t make it any better. Does it make it any better? And the plaster dented from your fist in the hall where you had your first kiss reminds you that the memories will fade.  So this is strange, our side stepping has come to be a brilliant dance where nobody leads at all, where nobody leads at all.  And the picture frames are facing down and the riniging from this empty sound  is defeaning and keeping you from this sleep. And breathing is a foreign task  and thinkings just to much to ask and you\'re measuring your minutes by a clock that\'s blinking eights.  This is incredible, Starving, insatiable, yes, this is love for the first time. Well you\'d like to think that your were invincible. Yeah, well weren\'t we all once before we felt loss for the first time? well this is the last time. This is the last time, This is the last time.    i hate when songs are like exactly right on.   '),(9059,3682262,'04,July,2004','2004-07-04 00:00:00','   Hey kids. well it\'s father\'s day. things have been normal- dad tried to wake me up at like 10 am this morning (nice try), i\'ve been laying around, i had to break into Beth\'s house (mat didn\'t have the key, haha), dean n leslie n dad had been on a motorcycle ride until a lil while ago, dean and meagan are fighting (he jus left to go see her), and well leslie\'s leaving soon to go back to Morehead and go to work... that leaves me at home, bored, with the \'rents. hmm... sounds excitin, ay?  well i have to say that last nite wasn\'t that much fun. i need to not try to have ppl over on the weekends of big parties, that never works out too good, haha. okay, so covert n nathan came over for like an hour but um, yay? lol and those 5ths i shoulda got... um that didn\'t work out either. Dean was very close to gettin some for me but he wasn\'t goin into maysville... sucked. but one of these days (hopefully sooner than later) i\'m gonna get to have my real birthday party, haha. gosh i bet anyone who reads this thinks i\'m a big alcoholic or something. believe me, i\'m not. i\'m a good girl, seriously. i jus like to have fun every once in a while. well i think i\'m gonna start writing about more important things than this soon. just be prepared. i\'m probably gonna write about things i\'ve never written about before. i never thought i\'d write about them in any kind of a thing like this where everyone can read it, but i think i will. maybe soon.   '),(9060,3682262,'04,July,2004','2004-07-04 00:00:00','   okay, so today is my birthday. woohoo :) so i figured now would be a good time to start this whole online blog/journal whatever it is thing. i\'ve been meaning to keep a diary for a long time but i never have so maybe this will help. i want something to look at in like 10 years and be like \'gosh i was so dumb!\' or cry about it or laugh about it ya know that kinda stuff. okay so maybe i\'m a dork, but you\'ll have to disregard that because you know you want to read this. haha         i think my \'rents are actually talking about buying a cell phone! (wtf?!) and getting a family plan so dean has one, too... i wonder if i get one? WAIT! i already have one- haha. they\'ll never know... but that\'s gonna SuCk if they buy me one and i already spent $ on my own. hmm...         well ppl are coming over 2nite (hopefully) and it\'s gonna be fun (hopefully)... i\'ll find someway to make it excitin. i was kinda thinkin bout goin over to blue licks like really late and sneaking in to swim when nobody was there... wouldn\'t that be fun? dunno if that would work but that would definately top breaking into a house! haha (we\'ll try not to get David grounded this time- sorry beth! lol) i think after the party should be more excitin cuz i better be gettin some 5ths and some swisher sweets for my bday... hmm... or maybe somebody would like get me something besides rum cuz that\'s all i ever drink. how bout a huge case of beer? haha don\'t think we could hide that very well. hypnotic? is that expensive? ppl say it\'s good... atleast me n my girls will have a good time.        my \'rents are also talking bout buying one of those new Dodge Magnum ugly things. well actually order one cuz you can\'t get one w/ a hemi in it yet or whatever dad\'s all hyped up about. they\'re gonna trade in the tahoe and his truck for one- what are they thinking?! i\'ve openly opposed to it but i doubt they care much. i mean who wants to ride in a \"nicely designed\" (as my dad says) station wagon? *gags* i pray they make the right decision. but then again... if we have less cars by the time i\'m 16 1/2, does that mean i get a camaro? i\'m praying for that, too.        well that\'s about it. sorry it\'s so long- since it\'s my first entry i went all out i reckon. haha well i have to go talk the \'rents into taking me to the berry festival! yippee! haha Berry\'s the coolest in case you don\'t know. i need to take a shower but i\'m so sunburned it hurts to touch my head so washing my hair would be painful... but worth it, nonetheless- gotta look good for the sexy cowboys that i\'m asking for for my bday, haha. gotta clean my room, gotta do laundry, gotta get a move on. Bye kids!   '),(9061,3682262,'04,July,2004','2004-07-04 00:00:00','   July 30   Why is it that sleep is so hard 2 come by these days? I mean geez I woke up at 9:30 this mornin and I\'m still up and runnin. I mowed, yippee. Got the same damn blister on my thumb, too. It\'s so freakin hotttt and it\'s raining, too. man ain\'t life grand? haha I\'m complaining a lot for no reason, ay? I\'m just so freakin bored... and alone! geez it\'s pretty sad that I can\'t think of 1 person that I could pick up the phone and talk to right now. (Beth\'s w/ Adam -e-... lol) Well I could think of 1 but I\'m too chicken to call him. that\'s pathetic, I know. Just don\'t think he wants a call from me, I\'m prolly the furthest thing from his mind and I\'m gettin used to it. I mean hell I don\'t blame him for not caring anymore, just sometimes it pisses me off but I\'m not exactly sure who I\'m pissed at?   July 31   Why does everything turn into something bout him? I think it\'s a nitetime thang cuz it\'s mornin now and well, okay...changing subjects... Deanie\'s friends from college (Dan & Phil) are comin over for a few days 2day. I\'m goin to see if Beth wants to come over... this should be interesting. haha I love it when my house is full of ppl... that\'s why I love Christmas! Well \"the holidays\", ya know.    Geez I cannot wait to move into my new room... but I feel kinda bad cuz it\'s Leslie\'s old room and she has no where 2 put her stuff in her \"cabin\" she lives in w/ James. So we were gonna put her stuff into my room, but now Dad\'s making it into a studio (hmmm... does he not already have one above the garage?) &amp; now when we come back, like when we\'re older, Leslie won\'t have a room. Isn\'t that awful? I feel so freakin bad & she already said that she hasn\'t \"had a home\" in a while &amp; now I\'m sure that\'s goin to make it a lil worse, ay? Plus there\'s not a place for her stuff in our house. There\'s no room for her in her own home. Well, I could leave some of her stuff in her closest, but woohoo she has a closet! (Is she gonna sleep there years from now?) But then there\'d be no room for my stuff. Geez this is pretty cracked up, ay?   This dog, that isn\'t even ours, has been barking at a garbage bag for the last 15 minutes, seriously. I feel like shooting it... but that would be cruel. Bad thoughts! haha   '),(9062,3682262,'04,July,2004','2004-07-04 00:00:00','   Beth says hi & that she has a really good book (it was her Dad\'s, but well, she likes to steal) &amp; she loves to eat & sleep it. Oh, &amp;  red  really looks like  or-ange ! haha Well it\'s a quarter til 1 in the mornin, woohoo. I\'m bored. Ya know what? When I hug him I fit right under his chin. like perfect. what else would you want? Guess what? I watched Dirty Dancing Havanna Nights 2nite- I liked it ToNs. I still need to post those pics. I got drunk @ Liz\'s last nite (we got followed by a guy w/ long hair in a semi-truck- Joyride)... fun stuff I reckon. The day before that Adam Evans (BoCo) came to Beth\'s cuz they\'re talkin &amp; we rode \'round town for like 4 hours! lol And I drove a stick-shift car! well truck! yay! Oh yea! Abby bit Turtle 2day! That\'s not good... I love Turtle- he\'s adorable! I have pics of him I can put on here. I will.   '),(9063,3682262,'04,July,2004','2004-07-04 00:00:00','   I forgot to put the lyrics to all the songs I referred to in my \'laurel\'s adventures\' post, so I\'ll put them here. Have fun!       &nbsp;  Hands by Jewel   urlLink http://www.azlyrics.com/lyrics/jewel/hands.html   I Will Remember You by Sarah McLachlan  urlLink http://www.letssingit.com/?http://www.letssingit.com/sarah-mclachlan-i-will-remember-you-5h4ctjn.html   Try by Nelly Furtado  urlLink http://www.azlyrics.com/lyrics/nellyfurtado/try.html   Sunrise by Norah Jones  urlLink http://www.songlyrics4u.com/norah-jones/sunrise.html   Like the Rain by Clint Black  urlLink http://www.letssingit.com/?http://www.letssingit.com/clint-black-like-the-rain-27fkjtt.html   Broken by Seether and Amy Lee  urlLink http://www.letssingit.com/?http://www.letssingit.com/seether-broken-glcgbjt.html   Leaving Las Vegas by Sheryl Crow  urlLink http://www.letssingit.com/?http://www.letssingit.com/sheryl-crow-leaving-las-vegas-hr374vr.html &nbsp; My Heart Will Go On by Celine Dion  urlLink http://www.letssingit.com/?http://www.letssingit.com/celine-dion-my-heart-will-go-on-r8pk278.html   Take It Easy by Travis Tritt  urlLink http://www.letssingit.com/?http://www.letssingit.com/travis-tritt-take-it-easy-v5s9lzh.html   End of the Road by Boyz II Men  urlLink http://www.letssingit.com/?http://www.letssingit.com/boyz-ii-men-end-of-the-road-qt1ph73.html   *other good songs&nbsp;I heard in&nbsp;Canada:  Seven Nation Army by White Stripes  urlLink http://www.letssingit.com/?http://www.letssingit.com/white-stripes-the-seven-nation-army-9t4w5q7.html   All Apologies by Nirvana  urlLink http://www.letssingit.com/?http://www.letssingit.com/nirvana-all-apologies-plc2281.html   Slither by Velvet Revolver  urlLink http://www.letssingit.com/?http://www.letssingit.com/velvet-revolver-slither-k3c6sbs.html   Fall to Pieces by Velvet Revolver  urlLink http://www.azlyrics.com/lyrics/velvetrevolver/falltopieces.html   Patience by Guns and Roses  urlLink http://www.letssingit.com/?http://www.letssingit.com/guns-n-roses-patience-xczs7gn.html   Somewhere Out There by&nbsp;Our Lady Peace  urlLink http://www.letssingit.com/?http://www.letssingit.com/our-lady-peace-somewhere-out-there-rjwjg1s.html   Plush by Stone Temple Pilots  urlLink http://www.letssingit.com/?http://www.letssingit.com/stone-temple-pilots-plush-k9xmnfx.html   Addicted by Enrique Iglesias  urlLink http://www.letssingit.com/?http://www.letssingit.com/enrique-iglesias-addicted-cjd3g17.html   Beautiful Good-bye by Amanda Marshall  urlLink http://www.letssingit.com/?http://www.letssingit.com/amanda-marshall-beautiful-goodbye-555s3xs.html   *one more before I\'m done:  Vindicated by Dashboard Confessional  urlLink http://www.letssingit.com/?http://www.letssingit.com/dashboard-confessional-vindicated-mpkvtcn.html   '),(9064,3682262,'04,July,2004','2004-07-04 00:00:00','   Well here it is... the Redhead Remix of Roses! by Beth (co-writer me) inspired by um, just bout everything. (to the tune of Roses by Outkast! obviously, haha)  Laurel Lynn (Laurel Lynn). See Laurel Lynn, all the guys would say she\'s mighty fine (mighty fine), but mighty fine only got you somewhere one-third of the time, but the other quarter either got you cussed out or punched in the eye.  bridge: well dig this now, even tho (even tho) she needs a platnium calculator to multiply (to multiply) the time it takes to look inside and realize that red hair donâ€™t care, care,&nbsp; c-c-c-c-care.  chorus x2: i know you\'d like to think your hair is red, but lean a lil bit closer, see, red really looks like or-ange. yea, red really looks like or-ange.  Laurel Lynn (Laurel Lynn). she\'s the reason for the word red (red). well when we\'re speedin on the way to cynthiana and i\'m gonna put my selt belt on cuz dad told me to. no i don\'t want to but i guess i have to so i won\'t die if we crash, crash, crash into a ditch. i\'m not playin.  bridge  chorus x2  but she\'s bout half retarded, everyone likes her big booty. When I met her on her Harley, she was as bright as a red Crayolie. I said Laurel would you call me? She said itâ€™s long distance from Mt. Olivet. I said Laurel, you sound like a giant eree-eree. Oh so youâ€™re one of them penguins and you live in the zoo and you got red hair, too. Beth donâ€™t like feet, 19 sexy-sweet. Brush your teeth, always floss! When youâ€™re playin baseball wear your cleats. ow! hamburger meat. Ow! &nbsp;If you wanna scare me- Ow! Be a ghost wear a sheet. Ow! Donâ€™t trick-or-treat. No! (no!) Now go on Kara Beth, my pooh cover is flawless. Regardless, Blankey wanna get involved with all them lawyers. And 80s tees and Zach Morris in a courtroom. Smokey donâ€™t like forest fires, so FIRE ON YOU!  chorus x4  this thang x2: Qclub guy thinks I live in Mars. Laurel quit chasin cars. Redhead got ran over by a tax-i in new york, new york, new york.  crazy Beth. crazy betch. crazy Beth. crazy betch. crazy Beth. crazy betch. crazy Beth. Stitch. (crazy bitch.) Lilo and stitch. (Gullible Laurel.) Naked with eggs. (crazy braytch.) Oleâ€™ crackin braytch. (wear that shirt.) A betchâ€™s Beth. (frozen poptarts.) Just a birdâ€¦ (stupid bitch) ate penis cake. (crazy trip.) Oleâ€™ Steph loves beans. (crazy betch.) Oleâ€™ Peter Rabbit. (crazy chocolate.) A big fat secret! (crazy shirt.) Just a tree...&nbsp;&nbsp; &nbsp;  Being stupid is so much fun sometimes. Make sure to take my poll after you read this so you can tell me what you think! :) Me n Beth n Evette n Steph stayed the night at Brandonâ€™s last nite and stayed up like almost all nite. I am the Scrabble Queen! lol We all had fun (of course) but it was def light outside when I went to sleep. oh well We went to the fair 2nite. I was alright. There were enough ppl there, of course not really the ppl (person) I want to see, but ainâ€™t that the way it works, lol. I got to drive Sean\'s car for like a really long time, on a real road! lol I was so proud of myself it was super fun! I did good, too, you should be proud, mkay?&nbsp;We only ended up ridin one ride and we went through that Mummy thing, lol. I was actually scared at the beginning of it. Me n Beth started to go in it once but came back out cuz neither of us wanted to go in front, haha. I know weâ€™re â€˜tards. Does the song â€œThe Good Stuffâ€ bother you, too? I mean not really bother, but you know. Or maybe you donâ€™t. It was worth a shot. oh I just remembered the other day we went to Food Lion and we saw Preston and he made me a 4sale sign on one of those MVP Bonus Buy signs lol. I taped it to my back. fun stuff, really. Gotta have fun in grocery stores, or the world is lost. And like 2 nites ago&nbsp;Jason&nbsp;came over here and hung out for a while (I got to drive! yeehaw! lol)&nbsp;Okay Iâ€™m tired and everybuddyâ€™s goin to sleep on me so Iâ€™m gonna go jump and sit on them. Sianora.   '),(9065,3682262,'04,July,2004','2004-07-04 00:00:00','   I wrote something almost everyday I was gone, so here it is!  July 1st  the city of Shaginaw, Michigan&nbsp;   I&nbsp;donâ€™t feel like writing so hereâ€™s some memos & song lyrics I jotted down 2day:   *Shrimp, Shrimp, Shrimpâ€¦ &amp; Moâ€™ Shrimp! (truck stop sign)  *when I go on a road trip weâ€™re going to Yellow Springs, Oh. (where I saw a woman in a business suit carrying her shoes down the roadâ€¦ itâ€˜s like a hippy town, fun stuff really) and Point Pleasant WVa. (moth man place). and I got the idea to go 2 strangers houses and ask to jump on their trampolines.  *â€œI never liked the rain until I walked through it with you.â€ Clint Black (Like the Rain)  *â€œCuz Iâ€™m broken, when Iâ€™m lonesome, and I donâ€™t feel right when youâ€™re gone away.â€ Seether and Amy Lee (Broken)  *Sheryl Crow lyrics:  **â€œAll I wanna do is have a lil fun before I dieâ€ (All I Wanna Do)  **â€œJump in, letâ€™s go. Lay back, enjoy the show. Everybody gets high, everybody gets low- these are the days when anything goes. Everyday is a winding road (I get a lil bit closer). Everyday is a faded sign (I get a lil bit closer). Everyday is a winding road (I get a lil bit closer, to feeling fine.)â€ (Everyday is a Winding Road)  **â€œThis has been the best part of my losing streak.â€ (Leaving Las Vegas)  *I kicked a stone from Applebeeâ€™s to our hotel.  *I saw a wrecked Volkswagen Beetle with blood all over the inside.  *I watched the Spanish channel with mom â€˜til Dean got mad and changed it.  *Had to take advil with bourbon water, momâ€™s orders. lol (my dad was drinkin bourbon and he left it sitting and fell asleep and mom thought it was water and told me to drink it to take advil, Nasty kids, very nasty.)   July 2nd  in Little Current   On the way back to the hotel from the restaurant me and Dad had an interesting topic of convo. Iâ€™ll set the scene for you: this guy was riding his bike past us and doing tricks and stuff and Dad recommended that I walk up to the bike guy and give him a 4-leaf cloverâ€¦ I was like rightâ€¦.. But that got him talking bout how I should live life to make memories- like just have meaningful moments with ppl (strangers, friends, anyone I guess). He said that making memories could be like an art formâ€¦ I mean what if someone spent their life creating memories in pplâ€™s minds. Crazy idea, but I like it. Sharing moments with ppl is something like everyone likes at some time in their life- why not all the time? Ya know, no matter how much I say I hate him sometimes I just love him too much.   July 3rd  my little island   I saw the hottest guy EVER today at the dock. Iâ€™m guessing he was a bit old for me by the cases and cases of beer he loaded into his boat, but hey- a girl can dream! haha  Well Iâ€™m now posted on Iroquois island. Ya know, no matter how many times Iâ€™ve come here it doesnâ€™t cease to amaze meâ€¦ itâ€™s just so beautiful. Dean and Dad have been fishing for at least 4 hours now and thereâ€™s not sign of them yet! They better get back before the sun goes down or theyâ€™ll be devoured by mosquitoesâ€¦ theyâ€™re So thick here. omg here comes a boat pulling a boat very similar to ourâ€™sâ€¦ haha Iâ€™m gonna laugh my ass off if thatâ€™s them! I think it is! hahahaâ€¦ okay damnit, itâ€™s not! grrrrr! haha Well I guess it\'s one of our neighbors for the week... hmmm... they don\'t look very excitin. Ya know I painted momâ€™s toenails 2day for the first time in her EnTiRe LiFe! Isnâ€™t that crazy? Okay mom walked down to the dock cuz she thought it was Dean n Dad, well she came back up here and was like â€œWell I felt foolish.â€ and I was like, â€œwhy didnâ€™t you turn around when you figured out it wasnâ€™t them?â€ and she said â€œwell I rushed down there and saw strangers, I wasnâ€™t gonna rush back, theyâ€™d think I was weird.â€ I asked her what she said to them and she told them that the way they were towing something (which turned out to be a rubber inflatable thing) she thought it was her husband. At least I got some humor out of the situation. I have to pee now.   July 4th  some kind of fourth of July   Well today started out alright. I had weird dreams last night... ya know how ya wake up and go back to sleep  and you have continuation of the first dream? well thatâ€™s what was happenin except now I donâ€™t understand how they connect, or which one was first~ they all followed the same theme. About Mike of course. bunches of people were in them at different times, except Beth (who was there the whole time). like: my â€˜rents, Brandon, Dean, I donâ€™t remember other people. They involved weird things like Beth writing on this big white wall (in green markers) things like â€œLaurelâ€™s gayâ€ and I got really pissed in my dream. But I remember me and Mike havin a long convo and feeling really good when I woke up but I donâ€™t remember what we said. AnYwAyS... we all went fishing this morning and it started POURING down rain and lightnin and shit and so we anchored on this weird island and kinda stood there. Well, I was okay for the first half hour we were there... but in another hour my clothes (including my piece of shit rain jacket) were soaked and my whole body was wet. I was in a bitchy mood and didnâ€™t hesitate to tell my dad just what a piece of shit that rain jacket was. When it let up, they wanted to fish more and I was like â€œDo you want me to pour buckets of water down your rain suit and see if youâ€™re in the mood for fishing then?â€ We went back to the cabin. That ruined the rain for me forever... well at least for the rest of the week. I need a back massage. Last night before I went to sleep I was reading and the way the light is hanging in my room, I couldnâ€™t see very well so Dean swung the beds around in my room~ while I was still in mine. It was FUN. haha Well Iâ€™m peeing â€œhealthierâ€ now~ donâ€™t ask. Oh- Beth I heard that â€œAlthough weâ€™ve come to the end of the road..â€ song today! haha 94.1 Moose FM! Yeah buddy! I also heard that â€œHandsâ€ song by Jewel~ I forgot how much I loved it. Itâ€™s grrreat! Yea oh well time to dine.   July 5th  happy b-day Autumn!   The temperature today has been weird. This morning it was cold and rainy and by the afternoon it was warm enough for me 2 lay on the dock in my swimsuit. Not quite warm enough to actually swim, tho. haha Dad caught a bass today~ big enough to eat. I caught about a five inch one off the dock, haha. of course I let it go. I finally got to blow out my b-day candles today! Even though my b-day was like, what, 17 days ago? We didnâ€™t have any candles at home when we celebrated my b-day... so when we were shopping for groceries to bring up here I saw candles and I thought if I got some and put them in the brownies mom planned on making thatâ€™d make up for it. I heard tons of good songs today. The song â€œI Will Remember Youâ€ by Sarah McLachlan is seriously exactly the way I feel a lot now. I forgot about that song. And I heard this new Nelly Furtado song- â€œTryâ€ Its good and Norah Jonesâ€™ â€œSunriseâ€ I love her voice. Dean tried singin â€œMy Heart Will Go Onâ€ It was great. haha Well its midnight and Iâ€™m tired- Adios!   July 6th   Itâ€™s rained off and on all day today, not too bad until like right now, though. Itâ€™s after 11 at night. We all went fishing from like 10-6 or something. Nobody really caught anything except Dean. He was really reelin â€˜em in at one point. He caught 3 Pikes. None of them were really big enough to keep, but he kept the first one, it was real real close to 2 feet. Dad keeps saying Iâ€™m asleep and Iâ€™ve left the light on and Iâ€™m like geez how can I be asleep if Iâ€™m writing? And heâ€™s like- you can write w/ your eyes closed? And Iâ€™m like you canâ€™t even see me from where you are! GEEZ! Weâ€™re all goin to freeze to death in this cabin. I never remember it bein cold but I have 3 blankets on and I could sand another one. Oh, I just remembered something dad was talkin bout today. His friend, Mike Straw, used to talk about a guy who got a bumble bee sting on his dick and it â€œswelled up as big as a baseball bat.â€ haha I kinda doubt thatâ€™d ever happen, but talk about a 3rd leg. lol. Crazy, ay? You know some people claim lilâ€™ kids are scared of everything but I think theyâ€™re too curious and well, dumb to know when to be scared. I know I was. I kinda miss that. I was never scared of spiders when I was younger... or bees, or much else, really. I wouldnâ€™t want to be naive and innocent forever, but it definitely has its moments. Definitely...  &nbsp;   July 7th   Well I like just woke up and I had the coolest dream in the world last night. The beginning is kinda vague and weird... like I was in a gym (like in a school) and I donâ€™t remember what I was doin but after while I noticed these white lions (like @ the Cinci Zoo) a lilâ€™ ways down from me and they were like walking around and not paying any attention to me, but I was convinced they were gonna attack me. So there were these rooms and I locked myself into one and jus waited. Well my mom came to get me and got me out of the room and the lions still didnâ€™t pay any attention to us so we just left. When we got outside somehow Beth was with us and her dad had come to get her. We were in our student parking lot but it was dark outside and there were these two girls in Shannonâ€™s truck with him. They reminded me of those lesbian porn stars in Anger Management. And he had like met them at some support group but I donâ€™t remember what it was called. I think it had something to do with killing ppl? Well, Beth went home and I was on this long road and everything was like hazy and gray and Evan was on it, too, but he was up a ways from me but we were facing each other and walking. I had my cell phone and he was like sending me all this these text messages but I have no clue what they said. Well then I think I was in Canada and my rents were with me and I kept using my cell phone in front of them and they didnâ€™t care and they were gonna get me a better one. Well then there was an outside basketball court and Beth was supposed to be having a gym class there. I talked to her some but her instructor yelled @ me and I hadda leave. Then was the coolest part~ it was the last day of school and it was our school, but it was a million times better! I didnâ€™t have to go to my classes I reckon cause I was just wanderin around the whole time and I went to the cafeteria, I think but it was like the middle school one with a stage and all and there were these girls up there and one of â€˜em looked kinda Hawaiian. She was singin that song â€œTake It Easyâ€... ya know? And other girls were playin instruments and the girl would sing the song but make up verses in the middle about stuff goin on in the room and there were these other girls doin a tap routine... Leslie Preston, etc. lol. It was weird. Then I left and like went up the slope. Ya know, the hallway outside the gym, and there were people everywhere doin the tap thing and Mrs. Casey came up to them and was like what are you doing? And they told her and she was like I know, Iâ€™ve been doing it all day, too. It was the coolest school ever and it was a lot bigger and there were tons of people. I donâ€™t know. Then it was supposed to be 4th block or something and I was gonna go to class to see what we were doin. So I went upstairs and a lot of people were leavin class so I went with them and I found Beth, Jen and Amanda and a buncha other people and we were goin to some assembly, I guess? And Amanda was talkin about how Kensi didnâ€™t come to school because she would have made fun of her. Well we got to this place that looked like a big version of Jonesâ€™ gym/cafeteria (Mason Co.) and there were groups of people doing whatever they wanted so me and Beth went walkin around and I remember Clink bein there? And then I saw Kensi talkin to my mom or somethin. But there were people sellin stuff and we talked to this girl that looked like that girl on Degrassi (I canâ€™t remember her name... Liberty I think?). Then Kensi was talkin to us and this guy came up to us and he was supposed to be famous like a guy from New Kids on the Block. Somehow we got in this really involved convo about love. There was this package for sell that was like a love kit I reckon and the guy threw it and it opened and somethin in it said VooDoo. So, the guy picked it up and Kensi gave him gum out of it. I donâ€™t remember what they said but I got the fact that her and Rus had broken up and she was a totally different person. The guy left and I asked Kensi why her and Rus broke up and she said he was arrogant and she grew up. Earlier in the dream, I was sitting in the real cafeteria with a buncha guys and it looked like Straubâ€™s cafeteria. (Mason Co.) and we were watchin TV. I think Abby was with me and we were talkin about Chadâ€™s legs? And then me and her went to the auditorium and it was like rehearsal for our show last year and Tracy, Anna Kristin, and Kitty were onstage. Mrs. Franklin was showing us the arms for something. Then I left and I saw Mike and I think I kept walking and I saw John Turner and we talked then I kept walking and Mikey Harris slapped my ass or something and him and Scott were tryin to talk to me but I saw Josh Ecklar and went to talk to him and me and Nina ended up goin to his house? But then we had to go to Ninaâ€™s and her aunt was there and she was really mean so we went back to school. Anyways I know that was the longest pointless thing Iâ€™ve ever written but it was just so cool cuz our school was bigger and nobody cared what anybody else thought and we were pretty much nice to each other and it was just fun.   July 8th  Whoa Nelly! haha   Geez o pete o geez~ Dean wont stop rocking back and forth and staring at me and singing â€œRing around the Rosieâ€ all scary-like. Sometimes I donâ€™t want to claim him. Oleâ€™ Deano. haha! well I just had the biggest adventure of the week! We went to Nelly, this lake thatâ€™s on an indian reservation. Itâ€™s my favorite place in Canada. I have pictures from there. Dean wanted to climb over on top of all these rock hill things to take a picture and he wanted me to come so we set off. I have to say it was a bit rocky and kind of dangerous but of course fun! I have 2 more bug bites, tho. Oh and I lost my shoes like 3 times cuz I was wearing flip-flops! haha! I just feel so alive there. I could just stand and stare up at the sky and feel like I never had to move. It makes me feel like I can do whatever I want! haha I wrote my name in rocks to see if it would be there next year, prolly not, but? omg this is the funniest radio program Iâ€™ve heard in a long time. Theyâ€™re discussing living in a city or a suburb. What the hell is the difference? This woman said Escalade like it was Escalad. (I cant stand people who talk with flat aâ€™s) She said if the roof breaks sheâ€™d have to move cause she doesnâ€™t know how to fix anything! wtf?! lol oh well. Iâ€™ve been reading the â€œDivine Secrets of the Ya-Ya Sisterhood!â€ Its long but I love it. I want to live in it. Itâ€™s complicated but I need that. I need some things to go wrong so I can appreciate it when everything is right. I want to live in the deep south and have close friends and get arrested , get talked about and not give a damn, have kids, cause controversy... I want to live! I mean like really live! I just havenâ€™t found a job that fits in w/ that. So I donâ€™t think I will. Work, I mean. Iâ€™ll be an old fashioned mom. Have a 2-story house w/ a finished basement and a huge lawn w/ big trees and a volleyball net and a lake w/ a dock and maybe a canoe! lol! A big porch and parties every month. I donâ€™t mean trashy parties, I mean dress up and dance parties (open bar, of course. If someone is gonna write a book about me I would give them a hell of a lot to write about. My life might look nice and simple and plain but Iâ€™m gonna surprise the hell out of people. Shake things up a bit, ay? Beth letâ€™s find hubbies after college and go live in some southern state and upset some little town. lol Iâ€™ve got big plans for such a short period of time. Maybe I could just stay here to live and do just as well. Kentucky is fab. and Iâ€™m serious... Well our feast of green beans and tuna sandwiches is FiNiShEd so I\'ll quit my rambling...for now! Smooches!   Back from dinner...Deanie tried to say his stomach gets big when he eats. That was good for a laugh. lol Back to what I was talkin bout earlier (kinda)- I want family dinners, a dishwasher, someone to clean the house (maybe)...Iâ€™ll do laundry and cook and raise kids and make fun, that should be enough, lol! I think I can do it! Oh and one of the most important things is that my husband is gonna think I\'m sexy! No matter how old I am. Things just get better with time, honey. I cant stand men who talk down to their wives, they should be shot. My man will appreciate what he has. â€˜Cause Iâ€™ll be worth it. But I wonâ€™t put up with shit. Laurel Mallory will not put up with a manâ€™s shit!... I take pride in being difficult... no matter how bad it hurts me sometimes.   July 10th  is there a point to Diet Sierra Mist???   This is my 2nd day on the train 2 NYC. It\'s nice I guess. It took me a lil while to get used 2 the constant rocking, but I\'m good now. I\'m in the state of New York, but I still have a few hours before I get to NYC. These are really comfortable seat, the food is tolerable, and I\'ve seen four hot guys! yippee! haha The first one was at the train station yesterday- he worked there- and he reminded me of someone but I couldnâ€™t figure out who. The 2nd one was on the train yesterday, playing cards with some lil kids. I heard him say his hat was from Australia and I think that\'s why I found him attractive... the dangerous type! haha The third one worked at the train station in Toronto- blonde hair. He was nice... nice smile. The fourth one was kinds of unexpected... I said â€œExcuse meâ€ and he said â€œCertainlyâ€ how adorable is that?! haha There was also this guy that smiled at me in the eatin car but he was kinda scary, actually... oh well Iâ€™m sure theyâ€™re all a tad old for me. The guy selling food is rude and I really, really donâ€™t like him. Screw him and the horse he rode in on! haha (DS) but he gave me a free water, so? lol AnYwAyS- Sheryl Crow is my fuckin idol, kids. Iâ€™ve even heard sheâ€™s dating Lance Armstrong... sexy- but I thought he had a wife? Every single song she ever sang is absolutely fab. (That reminds me, you can get the cutest lil bottle of Absolut on the train. haha) Iâ€™ll put links to my fav Sheryl songs on this entry @ home. I wish Iâ€™d get to NYC so I can shake my booty, and truck it back home (w/ a National champs trophy, hopefully...) I miss yâ€™all! Beth- Iâ€™m wearin my indian choker today! Yesterday I met this woman named Natalie raveling to Toronto w/ her hubby and baby girl (who was adorable!). She was really nice and reminded me of one of my best friends from elementary school- Elizabeth Smith. (Everybody gets high, everybody gets low- these are the days when anything goes...)   oh and I figured out the answer to the not-really-a-question of â€˜chase what makes your heart flutter, even if it hurtsâ€™ or whatever ya know... I say â€˜if it makes you happy, it canâ€™t be that badâ€™. (a lil obvious help from Sheryl on that one!) Like any of that really matters anymore anyway. Itâ€™s not like I think bout that 24/7... only about a minute a day now (hmm...). Thatâ€™s enough time for it to kill me, tho. I thought I was over alla that... but it hit me upside the head out of left field again. (lol) Sorry for being so fuckin stupid. (That was an apology to myself). Iâ€™m too strong to let stupid shit get to me. I have to keep reminding myself of that.   *Make Sure Yâ€™all Look @ These Sheryl Lyrics!:   The Difficult Kind  urlLink http://www.letssingit.com/sheryl-crow-the-difficult-kind-ld52t76.html &nbsp; best line: â€œI swear I can see you... coming up the drive. & there ainâ€™t nothing like regret, to remind you youâ€™re alive.â€ *chorus*   All I Wanna Do  urlLink http://www.letssingit.com/?http://www.letssingit.com/sheryl-crow-all-i-wanna-do-4jbgfh6.html &nbsp; best line: â€œAll I wanna do is have a lil fun before I die.â€   Everyday is a Winding Road  urlLink http://www.letssingit.com/?http://www.letssingit.com/sheryl-crow-everyday-is-a-winding-road-jk53v89.html &nbsp; best line: â€œI used to ride with a vending machine repairman. He said heâ€™s been down this road more than twice. He was high on intellectualism. Iâ€™ve never been there but the brochure looks nice.â€   Strong Enough  urlLink http://www.letssingit.com/?http://www.letssingit.com/sheryl-crow-strong-enough-gmvhwtb.html &nbsp; best line: â€œIâ€™ll be the last to help you, understand. Are you strong enough to be my man?â€   I Shall Believe  urlLink http://www.letssingit.com/sheryl-crow-i-shall-believe-qfz4sn1.html  best line: â€œEven if itâ€™s a lie, say it will be alright, and I shall believe.â€   Steve McQueen  urlLink http://www.letssingit.com/sheryl-crow-steve-mcqueen-2fcvwxp.html  best line: â€œI wanna leave you feelin breathless, show you how the West was won- but I gotta fly.â€   '),(9066,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    Dad\'s HD V-Rod... he just got it before we went to Louisville, that the best pic I could of it while it was in the garage, so there ya go.&nbsp; urlLink   '),(9067,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    being a \'tard before we went out Friday night! more on that later, too woooooo! haha&nbsp; urlLink   '),(9068,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    Beth the day the frog jumped into the toilet haha more on that later!&nbsp; urlLink   '),(9069,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    during painting... I dunno why I didn\'t put that inbetween the other ones, I\'m a dumdum&nbsp; urlLink   '),(9070,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    my room after I painted! how cute is it? I love it!!! lol&nbsp; urlLink   '),(9071,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    my room before I painted&nbsp; urlLink   '),(9072,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    Max! how cute is he? we were blowing bubbles on their balcony&nbsp; urlLink   '),(9073,3682262,'04,July,2004','2004-07-04 00:00:00','   I think I am gonna go crazy. And that\'s okay with me. don\'t know where that came from. I really want you all to look at this song because I love it and I don\'t know I just started singin it the other day while I was mowing and I have NO clue how it got into my head... I think I was saving it somewhere back in there for some random time. But it\'s great. It\'s called Long Day by Matchbox 20 so go to  urlLink http://www.letssingit.com/?http://www.letssingit.com/matchbox-20-long-day-d7zqxz7.html  to read it! We start school tomorrow and it sucksssssssss very much. bad stuff. also while you\'re on Letssingit.com you should look up Pedro the Lion they\'re super. they like sex a lot, but other than that they\'re allllll right. I went to Louisville the other day to babysit my cousin Max geez he is 5 and he drives me insane! I have some pics tho he\'s adorable i\'ll post them later. we went to this art openin thing at this place called the cinderbloc i think but there was the HOTTEST GUY I HAVE EVER SEEN IN MY WHOLE ENTIRE LIFE there. geez really and he dressed good, too. i don\'t mean preppy i mean like good i dunno. i shoulda stayed around i\'m he mighta got drunk and hit on me and that woulda be fun but ya know he was prolly like 20 so it\'d be bad news! hmmm... atleast i think it would be, I haven\'t decided yet. lol I wanted to see the man Dan in Louisville (he lives there) but I was too busy so i reckon i will go to the ky fair. how bout it? good stuff. i like the word stuff. it\'s so i dunno good. okay nevermind i sound like a complete idiot. but hey i only am sometimes, outside of the classroom when i get to think bout other things besides education. book smart, commom sense dumb. that\'s me! i reckon that\'s not exactly something i should be braggin bout, ay? lol I\'m really tired. i woke up at 6 this mornin to practice for school, like gettin up and all that. i had dance from 10-5 and i don\'t think i\'ll be walkin very lady like tomorrow at school. i\'ll prolly be limpin. but it\'s all good in the hood cuz i have dance for 3 hrs tomorrow. no pain, no gain, chico. it\'s so weird to be thinkin bout goin to school cuz i haven\'t been for so long... but i better be catchin some zzzz\'s if i expect to make it thro the first day. geez that\'s promising.   '),(9074,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    I\'m a rockstar!!! hell yea &nbsp; urlLink   '),(9075,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    ??? Beth thought I loved my dresser, so I was showing the love.&nbsp; urlLink   '),(9076,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    my homeless hair... please send $ to Rt. 1 Box 78 Mt. Olivet, Ky. 41064&nbsp; urlLink   '),(9077,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    I don\'t remember taking this... funny stuff kids!&nbsp; urlLink   '),(9078,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    Liz & Jeremy! awwwww!&nbsp; urlLink   '),(9079,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    me, Beth, Liz, & Ashley at Liz\'s!&nbsp; urlLink   '),(9080,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    the funniest picture of Liz I have ever seen, great stuff kids! hahaha she\'s gonna kill me&nbsp; urlLink   '),(9081,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    me n Turtle!&nbsp; urlLink   '),(9082,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    Turtle- the cutest puppy in the whole wide world!!!- on Beth\'s dresser.&nbsp; urlLink   '),(9083,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    Beth\'s a macho, macho man! woohoo! haha @ her house.&nbsp; urlLink   '),(9084,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    Brandon & his Pooh blanket- awwwww! haha in his upstairs livin room after cleanin up the fair that mornin (the naked w/ eggs story was NOT true)&nbsp; urlLink   '),(9085,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    imitating the Bevis & Butthead pin I bought Beth at the MTV store Times Square- at Brandon\'s (Steph was at Brandon\'s too- v)&nbsp; urlLink   '),(9086,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    \'ole Steph loves beans!!! hahaha&nbsp; urlLink   '),(9087,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    two retards in action- at Brandon\'s&nbsp; urlLink   '),(9088,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    Beth & Evette at Beth\'s&nbsp; urlLink   '),(9089,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    Abby, Ashley and Kitty on the bus ride home&nbsp; urlLink   '),(9090,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    alla us and Laura Bell after we performed Friday nite&nbsp; urlLink   '),(9091,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    in the Lobby at our hotel headed to the Grand Nationals Competition!&nbsp; urlLink   '),(9092,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    the outside of the Hershey\'s factory right off Times Square&nbsp; urlLink   '),(9093,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    awesome McDonald\'s sign right by our hotel&nbsp; urlLink   '),(9094,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    the lower Manhattan skyline from the Staton Island Ferry&nbsp; urlLink   '),(9095,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    Goofing around on the Subway to the Stanton Island Ferry! (we got stalked! haha)&nbsp; urlLink   '),(9096,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    Our penis cake at la Mela- a restaurant in Little Italy. We decided to try the \'ole \"It\'s our Birthday!\" thing... we were surprised!  lol&nbsp; urlLink   '),(9097,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    Me n Anna in Little Italy after shopping in China Town&nbsp; urlLink   '),(9098,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    straight down from our window&nbsp; urlLink   '),(9099,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    straight-across-view from our window in the daytime&nbsp; urlLink   '),(9100,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    It was Anna Kristin\'s \"birthday\" at Bubba Gump- they shoved whip cream up her nose! haha Oren\'s- our waiter- is playin a water jug in the background! lol fun stuff&nbsp; urlLink   '),(9101,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    Trump tower, ya know from that show- The Apprentice. I obviously couldn\'t get it all in the pic. We were on our way to the Met via 5th ave. and Park ave... nice window shopping, heart attack when you go inside and look at the prices, lol.&nbsp; urlLink   '),(9102,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    sittin on the windowsill after we went to the Junior Mr. & Ms. Dance competition.&nbsp; urlLink   '),(9103,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    starting off the New York pics! (I know I haven\'t put in any entries bout this, but be patient.) This is out our hotel room window (we were on the 37th floor!) and I accidently moved the camera when I was taking it, but it turned out pretty cool I think, lol.&nbsp; urlLink   '),(9104,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    one more from Nelly!&nbsp; urlLink   '),(9105,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    a different view of Nelly, you can see the mountains in the background.&nbsp; urlLink   '),(9106,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    after the trek through the woods at Nelly. *note: took this from top of the \"mountain\" kinda, lol.&nbsp; urlLink   '),(9107,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    awesome tree I found.&nbsp; urlLink   '),(9108,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    Mom and Dad in the cabin.&nbsp; urlLink   '),(9109,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    playin round w/ my camera- it takes black and white pics! lol *note: he didn\'t pose for this!&nbsp; urlLink   '),(9110,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    day at Nelly... wrote my name in rocks!&nbsp; urlLink   '),(9111,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    chillin on the dock. haha&nbsp; urlLink   '),(9112,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    Deanie\'s first fish! (it\'s a Pike)&nbsp; urlLink   '),(9113,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    view from my bedroom window.&nbsp; urlLink   '),(9114,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    in the kitchen... never got to go swimming :(!&nbsp; urlLink   '),(9115,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    day I got soaked... this is in the central room of our cabin.&nbsp; urlLink   '),(9116,3682262,'04,July,2004','2004-07-04 00:00:00','   urlLink    This is the start of the Canada pics! This is the first day we were there and so I think I have a pic for everyday? but I\'m not sure, but they\'ll pretty much match if you read the entries. This is on the porch of our cabin.&nbsp; urlLink   '),(9117,1240274,'22,February,2003','2003-02-22 00:00:00','   All around me are familiar faces Worn out places, worn out faces Bright and early for their daily races Going nowhere, going nowhere And their tears are filling up their glasses No expression, no expression Hide my head I want to drown my sorrow No tommorow, no tommorow And I find it kind of funny I find it kind of sad The dreams in which I\'m dying Are the best I\'ve ever had I find it hard to tell you \'Cos I find it hard to take When people run in circles It\'s a very, very Mad World  Children waiting for the day they feel good Happy Birthday, Happy Birthday Made to feel the way that every child should Sit and listen, sit and listen Went to school and I was very nervous No one knew me, no one knew me Hello teacher tell me what\'s my lesson Look right through me, look right through me  Mad World. Tears for Fears  '),(9118,1240274,'28,February,2003','2003-02-28 00:00:00','   agh. i survive gangs. i survive poverty, the drive by shootings, stabbings, and beatings in my old neighborhood. but this? fuck this. wish you coulda\' been my marijauna.  '),(9119,1240274,'27,February,2003','2003-02-27 00:00:00','   three days gone again, missing my girl, bittersweet angel. wish she would be here with me i\'d listen if she\'d talk make her out to be a princess to my head she\'s royalty.  tell me when we really begin, loving eachother like others in love kisses in twilight making love to the sound of two hearts drum.  tell me when we see more of eachother, and really communicate. stop my mind from racing, second guessing, let\'s not twist the knife, hurt eachother with silence, tempting our fate. i want to be with you i thought you wanted the same to laugh and cry and share ourselves but all these restless thoughts, they get the best of me here deep within they permeate, symbiotically tempting me, my fears they encapsulate.  so please celena, just come over. my heart is open. be honest with me about how you\'re feeling, let\'s not have this carry on, it hurts me so when i see that to me your eyes don\'t focus on what\'s going down in our world, dammitt Celena i feel they\'re hardly open.  if they were you\'d be around again kissing me, touching, and loving me more often.  maybe i\'m just insane, or unwhole and lonely. either way i need security, i need to hear it from you, that you feel love, and all you want is me.  '),(9120,1240274,'26,February,2003','2003-02-26 00:00:00','   tim hortons night shift in northern ontario. at a highway stop. interesting. i\'m tired as hell.  '),(9121,1240274,'26,February,2003','2003-02-26 00:00:00','   now i wonder,  if i disapeared again would anyone notice? now i wonder,  am i alone here as i was there? now i wonder,  why am i feeling the way i do --so much longing. now i wonder,  why i\'m thinking there used to be more to life than all of this. what is she hiding?  i sense it on her touch, i wish she wasn\'t so distant. i taste it on her kiss. what is it that she\'s missing?  she isn\'t getting what she needs from me. i wish she\'d express what she\'s feeling, cover the hole that i always seem to miss.  does she still love me or did i fuck that up too?  '),(9122,1240274,'25,February,2003','2003-02-25 00:00:00','   it\'s funny, ya\' know. how getting out and going on a doomed over night road trip, makes you feel good as soon as it\'s over. not that i didn\'t have a good time, bennet ;)  --in fact the opposite. everytime we talk, seeing you reminds me that i always come to you with the same tired story. made my problems seem trivial and stuff like that keeps me optimistic. the thing with life for me is, you gotta\' savour lifes bitter moments. it makes the sweet spots sweeter.  getting ready for my first day back at a job since calgary. that\'s over a month of hellish tormenting stress :)  '),(9123,1240274,'25,February,2003','2003-02-25 00:00:00','   ..site\'s back up. got a fulltime night job. things are lookin\' pretty, as I am starting tommorrow. wo0t! anyhoo. for those of you kids that have seen the site before, i am cleaning it up for people who might hire me one day, or for family and what not. old web log is no longer, and this site is now in phase two of turning it more into a portfolio for my work --as well as a portal for my other projects. thanks for comin\' out :)  '),(9124,1240274,'07,March,2003','2003-03-07 00:00:00','   \"It\'s time for people to show their cards and let the world know where they stand\"  - President G. Bush.  ..said like a true american. what a twat.  '),(9125,1240274,'06,March,2003','2003-03-06 00:00:00','   Health And Happiness (The Wallflowers)   Maybe it\'s not That I don\'t care anymore Maybe I just never did You can\'t say  That I have ever lied Because I keep the truth well hid Tell me what\'s wrong Is something wrong? You can\'t be sad When something you have never had Suddenly feels gone  I wish you health  I wish happiness  But absolutely nothing else  I\'m not here to keep an eye on you I think you misunderstood I could make things easy for you I won\'t  But I could Does that make you sad? Cause you look sad You should be glad For what you\'ve had And don\'t feel bad I could wish less for you  I wish you health  I wish you happiness But absolutely nothing else  There is no distance here To get across There\'s nothing to grip  Or get a hold The emptiness you feel Is nothing at all Now nobody broke any bones It could be worse Yeah, it can get worse You\'ve got your strength And you have got your health You should be worried \'bout Somebody else  '),(9126,1240274,'05,March,2003','2003-03-05 00:00:00','   monday 3rd of march         \"i just want you to know: life is a black sucking vortex of anguish and despair, filled with brief moments of false hope and empty joy all the while dragging you inevitably closer to a final, absolute and eternal death. life?  this isn\'t life.  it can\'t possibly count as life. i derive pleasure by bitching at you.  were i dead, i could not.  fuck you.\"  -ct.   this is the popular opinion among a few of my close friends, loves. not wanting to take the helm of the issue just yet, i am wondering what any of you have to say about it? observations? write it in the guestbook? serious submissions need only be written.  '),(9127,1240274,'04,March,2003','2003-03-04 00:00:00','   no more will this continue. a means to an end arises.  '),(9128,1240274,'02,March,2003','2003-03-02 00:00:00','   doing the right thing can be bad medicine. hoping all is well with ya\' ct. thinking lots.  '),(9129,1240274,'12,March,2003','2003-03-12 00:00:00','   just tell me what\'s the one thing you remember most about me?  was it my ride, my smile, my eyes, or was it the things i hide inside, or the way that you always thought   that i lied.  '),(9130,1240274,'10,March,2003','2003-03-10 00:00:00','   just a brick, too soon gone, prevents you from being, and doing, anything along the way of carrying on.  just a prick, with his head on wrong. doesn\'t understand the why,  and WHO he is.  he keeps awful control of the ship he\'s on.  just a kid, too soon out of favour, wonders why he\'s alone, confined, stuck on this one-way street he\'s spawned.  just a hypocrite, he tries to be Christian, but his ways are too wicked. so here and there you see him stumble. never aware of where he\'s been, or where he\'s going to.  just confused, and all mixed up, wishing he was something he already is.  neurotic and thick, always craving loves lust, and daily adventure wishes he was with her, yet then he\'d be too soon tied.  so this is what he does instead prays to God to cease the demons, screaming, wrenching, fighting in his head. lives every day, the same as the last, forgets the numbers that just don\'t add up, wishes he was someone else, someone in love, wishes his bride would be sent from above.  ignoring the piles so high, the time so quickly to pass this boy by.  what will he do? where will he go?  when the outside world turns callow, cruel, and cold. he\'d have been a lover long ago, a success, a caregiver,  on a path of envy, righteousness, and husbandry.  if only he saw the fork up ahead, when he would have to decide with his head; ignorance, for love and happiness. ..or misery for the real, unraptured society.  no fairytale ending in sight, no release to his plight, unable to tread life\'s scorn, he took his own life instead.  . . .unaware   of       the   ones       he stole with            it. . .  '),(9131,1240274,'10,March,2003','2003-03-10 00:00:00','   ..sitting here last night thinking, walking, waiting for something to come, some distraction always thinking my life, my world, my actions and disolutions. gone in a half-year moment, oh i could never leave you God, Just get lost in my false intrusions, pretending i was wanted..  So as I go, no one knows, snapping at heels, and shoebuckles, unseen. wishing i was clear, wishing i was clean, wishing i was not alone, electric, shiny and free. wishing i\'d find her, willing to make herself a part of me. Not worried about length, breadth, and span, would rather have just me, not worried about why i\'m here, my sins that make up who i am, Simple love, respect, and passion, Not the focus of my sinful ruin, Not for money, words, or apathetic union, she\'d love who and why i am, and see what it is i aspire, if only i could find a woman who\'d light my fire. Put some closure on this confusion.   Instead I took my pills, pink, orange, and yellow. stood inside this comfortable rainbow, held me, carry\'d me to sleep, my only love is here with me, she doesn\'t want change, money, or purity, as long as we are together, she is my security.  --m.a.    Just sitting. awakening. long sleep again last night. anyone like a good jeff buckley track?? snag this: \"Last Goodbye\". Wikkidd track. wo0t!  '),(9132,1240274,'20,March,2003','2003-03-20 00:00:00','   \" oh! o0o00hhh!!    IS --is THIS a PULSE   i feel in my neck? what? i\'m not dead yet? holy shit. nice to feel the wounds are healing and the winds are changing. . . \"  '),(9133,1240274,'19,March,2003','2003-03-19 00:00:00','   nostalgia and chaos. going away now. going tonight. not wanting to be seen, not like this. hidden from sight.  maybe one day, i\'ll come along. together we\'ll be, together we\'ll lay. lost in love, but not today.  '),(9134,1240274,'18,March,2003','2003-03-18 00:00:00','   not over yet it comes in lapses three years is all it took in between these broken moon catches to feel lacking, to understand want you around me notalgia yet enhanced so much driven passion words left unsaid please come from the past come back into my head.  '),(9135,1240274,'18,March,2003','2003-03-18 00:00:00','   ..down on my luck.  '),(9136,1240274,'07,April,2003','2003-04-07 00:00:00','   download: blink182. dammit.  yeah so i am all inspired and shit. have a great optimistic day. do some funky ass graphic up. feelin\' fine. --and then it happened. i thought of her. and that\'s it. i fucking hate the human condition. it\'s one thing to have thoughts. a memory. a series of memories. but then we also have to go and get emotional attachment to certain memories. figures that the emotional attachment to ex-girlfriends is harshly strong and fixated. funny though, i am trying to get  back on the horse  but i keep falling off.................  '),(9137,1240274,'06,April,2003','2003-04-06 00:00:00','   ..listening to \"mother\"; pink floyd. wondering where i can buy some inspiration... the clothing line, web sites, and other projects are all starting to pile up and i have nothing. three or more years of nothing to be a little more precise.   hard drive crashes destroy.  wish microsoft could be held accountable... fuggin\' XP.   \" Mother,  do you think they\'ll drop the bomb ? Mother, do you think they\'ll like this song? Mother, do you think they\'ll try to break my balls? Ooooowaa Mother, should I build a wall? Mother, should I run for President? Mother, should I trust the government? Mother, will they put me in the firing line? Ooooowaa Is it  just a waste of time ? Hush, my baby. Baby, don\'t you cry.  Momma\'s gonna make all of your nightmares come true. Momma\'s gonna put all of her fears into you. Momma\'s gonna keep you right here under her wing. She won\'t let you fly, but she might let you sing.  Momma\'s gonna keep Baby cozy and warm. Oooo Babe. Oooo Babe. Ooo Babe, of course Momma\'s gonna help build a wall. Mother, do you think she\'s good enough,For me? Mother, do you think she\'s dangerous, To me? Mother will she tear your little boy apart? Ooooowaa Mother, will she break my heart? Hush, my baby. Baby, don\'t you cry. Momma\'s gonna check out all your girlfriends for you. Momma won\'t let anyone dirty get through. Momma\'s gonna wait up until you get in. Momma will always find out where you\'ve been. Momma\'s gonna keep Baby healthy and clean. Oooo Babe. Oooo Babe. Ooo Babe, you\'ll always be Baby to me. Mother, did it need to be so high? \"  '),(9138,1240274,'18,April,2003','2003-04-18 00:00:00','   wo0t! need a better paying job. need a good girl on my side. need som\'ore ambition, cause i be runnin\' low...  ......misery loves company, i hate the company i keep. these malevolent demons, always at me while i sleep. . . . ..  '),(9139,1240274,'16,April,2003','2003-04-16 00:00:00','   i see you looking down from somewhere above what are you wishing whom do you love?    memories rush back to the crossroads.. the day my heart was torn, my tears darkened the sky that day, made a difficult decision, i had opted to walk away.  now here you are again. an angel before me      or a devil\'s vixen? you enter my dreams, you walk through my life, day to day basis, you really don\'t want to see me with my \'same old\', my rigorous strife.  sick of it then, so tired now. if you\'re here to stick around, why not just catch me, let me rest, gird my strength, did you ever really believe in me anyhow? what comes next is the change, i know you do it to us well, possibly save us, or at least grab our focus, place it on high, for Him to see, have Him look deep within me.  -m.a.  '),(9140,1240274,'15,April,2003','2003-04-15 00:00:00','   \".. instead,               i took my pills, pink, orange and yellow.  Stood inside this colourful rainbow.  Held me, carry\'d me to sleep, my only love is here with me.  She doesn\'t want change, money, or purity.  As long as we\'re together,     she is my security.\"  -m.a.  '),(9141,1240274,'26,April,2003','2003-04-26 00:00:00','   every little change i make every time i leave i begin to lose a sense of what it is to feel right now it\'s all gone no one here hardly even me.  '),(9142,1240274,'24,April,2003','2003-04-24 00:00:00','   tonight i take a stand let her guide me by the hand she wears a new face, of another place  --tonight i got a plan.  i am the change i want to see in the world. i am influence. i am spiritual. today i am the one who the mirror will answer to. tonight i beat the buddha. tonight i am playing for keeps.  making good on this \"removal\", to which the fortune teller speaks.  '),(9143,1240274,'23,April,2003','2003-04-23 00:00:00','   \" while you mock me, do not forget as you languish in your own effluence, that i am a self made man... and i could BUY you. \" ...don\'t ask.  '),(9144,1240274,'02,May,2003','2003-05-02 00:00:00','   does the dream like mentality  ever end  for you ? things ever feel  real?  hard edges from soft glow, running this way, free of inhibitions, out of control.. .  '),(9145,1240274,'01,May,2003','2003-05-01 00:00:00','   sittin\' here chillin with the cure and chattin it up wif my newest fandangled L33t hax0r chick Maggie. Thanks for hosting www.serotoninambition.cjb.net for me. you are my new bestest L33t hax0r chick pal. you rock ;)  -me  '),(9146,1240274,'28,April,2003','2003-04-28 00:00:00','   celena?  '),(9147,1240274,'27,April,2003','2003-04-27 00:00:00','   calgary.  there once was a day when nothing mattered. carefree, careless. going nowhere, yet being everywhere. omnipresent and loving every gapped out, blissful moment. i will go back there to my kingdom above the trees. the hapless caress of God on my fingertips. so visceral and seemingly infinite was my time here while it lasted.  just a glimpse, a memory in a place and time. never to encapsulate how it really felt. i was a living being. on edge, gaping over the precipice that stood before me. something big stirred in lifes\' manevolent belly. madison. celena. brothers. sisters. parents.  thus is life, i guess. here for a moment, gone eternally. yet lived out every minute of every day in my shadowy realm of memory and dreams. and because of this i am torn. always waiting for the new day to come.  '),(9148,1240274,'27,April,2003','2003-04-27 00:00:00','   no one\'s knocking on my door  nope. not anymore. they can\'t see me meek frozen cold and full of bullets crippled on the floor.  '),(9149,1240274,'10,May,2003','2003-05-10 00:00:00','   happy mothers day :D  '),(9150,1240274,'08,May,2003','2003-05-08 00:00:00','   today i:  got multiple jobs locked down. quit my job at jack astor\'s. ran into many faces from my past. got back my digital camera. accidentally ran into someone i recognized from friendster, THAT was odd yet was kinda\' cool and what not :D oh yah, and i met the guy who broke into my car last summer. i think, either that or he was a new theif. car theiving fuckers. all within 2 hours, i might add. dan, call me, fack. let\'s get a pint.  '),(9151,1240274,'07,May,2003','2003-05-07 00:00:00','   matthew good : the man from harold wood.  everytime i would see someone who just lived far from home, a friend of a friend, what have you.. I could almost always see this glint in their eye. I can\'t describe it in words, only in my work, but it is apparent when i look at them. . . you live a life for x years/months/weeks and then fwump, for whatever reason, here you are back in home sweet home.. you\'re smack dab right where --and WHO you were when you left. in THEIR eyes anyway. the friends your old persona loved, needed, clung to before the change. you don\'t get a choice. it cripples any growth processes that were going on in the inside. think of it as reminiscent to how God knew Adam had eaten from the tree of life. other than the fact they were running for cover, -nakedness- and what not, but it\'s present in the eye of Adam i bet, when he looks over at his longtime ally, God, it slips. . .  now that i am back, i keep thinking i still gotta cut the cord. lose my pre-teen-esque addiction to social behaviour and friends. damn pre-teens, always doing shit for their friends, little for themselves, and nothing for their family. i don\'t wanna\' be alone, but i want to stop spreading myself so thinly and make more time for the real friends i have, which, at this point, is down to ...two or three out of the dozens of so-called, old\'school buddies of yore. not to mention my family. my mom and dad are amazing people, and who knows where and when they will leave me.  with an end in sight, g\'bye ;)  '),(9152,1240274,'06,May,2003','2003-05-06 00:00:00','   ..blegh. i feel like i just woke up from a passenger side nap in my car. as i sit up and look around at the big picture i wonder.. who was driving for the last 7 months?  '),(9153,1240274,'05,May,2003','2003-05-05 00:00:00','   _ i never know what to write here.  || | ||| | ||||  | || | |||| || || | ||| | ||||  | || | |||| || || | ||| | ||||  | || | |||| || || | ||| | ||||  | || | |||| || || | ||| | ||||  | || | |||| || || | ||| | ||||  | || | |||| || || | ||| | ||||  | || | |||| || || | ||| | ||||  | || | |||| ||   msn: roofsessions@hotmail.com    halo. _  '),(9154,1240274,'05,May,2003','2003-05-05 00:00:00','     '),(9155,1240274,'16,May,2003','2003-05-16 00:00:00','   alright. months later now. she\'s hardly a thought, right?  wrong. take this girl-venom out of me............. make me forget.  '),(9156,1240274,'12,May,2003','2003-05-12 00:00:00','   favourite song of the minute: radiohead - be quiet and drive. mmmm. 2:05am. matrix reloaded: good for the  transition  flick in the trilogy. no spoilage, no worries :) got hired at transit shoes.   w e e e e e e !  now maybe i\'ll be able to hook people up real tight so they can hear \" hey! nice shoes! wanna\' fuck ?\" more often. a perfect utopia. i could make t-shirts for transit. start a club even. we could wear funny name tags. ..and yes, jenn, we could wear hats. alright,  ..must lay off the ganja when i write my web log, eh? yeah. holy shit!  look behind you !  ---end.  '),(9157,1240274,'12,May,2003','2003-05-12 00:00:00','   ephesians 4:11. wonder if i\'d fall into one of those gaps.  '),(9158,1240274,'22,May,2003','2003-05-22 00:00:00','   At this particular time i would trade all i currently own to be in Calgary. Yes.  Yes , I would.  '),(9159,1240274,'21,May,2003','2003-05-21 00:00:00','   there remains a lot to be said about the journey of self discovery. amazing how the people you know now, or the ones you lost touch with lifetimes before, can shape where your life takes you. whether geographically, emotionally, spiritually, and even morally. . .the heart is  so  ...spiratic. but once you plant your thread into someone else, you are plugged in forever.  '),(9160,1240274,'21,May,2003','2003-05-21 00:00:00','   and like a cord this music flows through me. yum. so glad you came. bound to be awhile? prolly. but only until i leave my mark, then i get on a train to everywhere and nowhere. gonna see what else i can stop hiding under this machine. confused?  '),(9161,1240274,'31,May,2003','2003-05-31 00:00:00','   hi. my nam est mikael-anton and i WRAT in zees web log, ah-hoh hoh. it is so much pleasure, zees \"blog\". yes-yes. zees way, peepol can zee thas i yam all abut zee DRAMA et la badly written poetry! zas i am depressed~! ooh la-la! keep on reading zee TRASH and it may WARRP your sense of reality, as it has my own. oui, oui.  funny thing ya\' know. it\'s been a killer week. and all i do now is smile, wave, and walk away :P   .me.  '),(9162,1240274,'28,May,2003','2003-05-28 00:00:00','   cars\' fixed. inspiration flowing through my veins again. many prospects. new characters have been cast. the pace is picking up. i\'m dehydrated and tired. things are looking up. . .  ah yes.  the stage is set for summer .  '),(9163,1240274,'26,May,2003','2003-05-26 00:00:00','   every now and again i get like this feeling so slow, so alone,  so inside myself, shuffling around.  something so strange, something so vague, i feel like the sound of crinkled paper, i feel broken today.   -me.  ...oh. btw, anyone out there got sum cool photoshop brushes they wanna\' share?  roofsessions@hotmail.com  '),(9164,1240274,'05,June,2003','2003-06-05 00:00:00','   ah yes.  the beginnings of consciousness.  '),(9165,1240274,'04,June,2003','2003-06-04 00:00:00','   always smiling and waving running around hidden under my broken faces untied shoe laces seething my fake optimism hands always shaking what are they doing to me?  sit alone. eat alone.  people my world with friends. what am i for? show me the point. give me a reason, something legitimate something more may something new this way come in this, my change of season.  -michael.  '),(9166,1240274,'11,June,2003','2003-06-11 00:00:00','   \" hey pig. yeah you. hey pig piggy pig pig pig all of my fears came true. black and blue and broken bones you left me here i\'m all alone my little piggy needed something new.  nothing can stop me now, \'cause i don\'t care anymore. \"  when he isn\'t bashing God, Trent has got some pretty good words to his music, eh? wish i could do it all over, but in a third world country in the 40\'s, or something.   -m.a.  '),(9167,1240274,'09,June,2003','2003-06-09 00:00:00','   aw yeah. listening to James: \"Wah Wah\". killer album. pressure\'s on.  in the world today, seems like more injustice: \"ABCNews is reporting on a 19-year-old college student at Rensselaer Polytechnic Institute in Troy, N.Y. He created a site named ChewPlastic.com where students could search for files on the university network. Mind you, this is not a music file sharing software, this is just a search engine. Presumably, the search engine was being used to search for music files as well. The folks over at the RIAA did not take too kindly to the idea, and sued the student. He settled but denies any wrongdoing. What was settlement, you ask? His life\'s savings.\"  ..if that was me..............................................................................  see fightclub ?  '),(9168,1240274,'18,June,2003','2003-06-18 00:00:00','   1. stress surrounding \"future\". 2. too busy to find time for personal human contact. (ie. friends, lovers, family. . .) 3. you think about money more times than sex any given day. (ie. problems, how to spend it, what you *want*) 4. picking up more hours at work. 5. less patience with people. 6. don\'t forget about the gym, it is bikini season after all.  7. depression. 8. more value is placed on material things than the \'single serving\' people you meet. 9. you\'re on some sort of medication.  sound like you?  yeah. i\'d say that 2,3,4,5,6,8,9 are definately issues these days. just because we\'re all in the same boat, doesn\'t make it right. so this is to all those people who look out the window on a sunny day and think \"shopping\". to those who bitch at the cashier and then go home and love up their pet cat. to the scared ones who take the pills their doctor gets paid well to give you --just to cope with this new dynamic, GLOBAL society. where do you place personal value? in a world where \"revolutionary\" and \"value\" are simple marketing words, do you really think all this is going to last forever? even until the end of your own life? (selfish if you don\'t care \'cause you beleive you\'ll be dead when it all goes to hell) ..and now even the wars aren\'t black and white anymore. bush being the perfect example. don\'t try and guess his motives or the thought police will get you. . . . . .and we\'re becoming more antisocial (this web log culture is a good example) and it\'s affecting our kids big time. i donno. i am just losing faith in people. what hurts more is that i lost faith in my own self long before today. . .  '),(9169,1240274,'17,June,2003','2003-06-17 00:00:00','   she had man hands.  '),(9170,1240274,'28,June,2003','2003-06-28 00:00:00','   with a million varieties of the way you can get pears in a can. with all those shiny toys out there to be bought. money to be made and spent. cancer to get removed. cigarettes to smoke. people to kill over oil. deadlines to bump up because with all the innovations in technology --you\'re saving time anyway, right?  with all the aborted babies and sexy models. city buses and free global trade. the fbi, the riaa. bell canada, and microsoft. with all the groovy clothing, dead animals on our feet, perfect teeth and OFF! Skintastic Spray. with the internet, gasoline, and vibrators. meal supplements, antibiotics, toxicity and wires.  with all the styrofoam, VISA, carloans, cardboard, and itchy skin. the death air, greenhouse, lights, and signage.    . . is it more selfish to have children,  or get clipped ?  '),(9171,1240274,'27,June,2003','2003-06-27 00:00:00','   Josephine  (The Wallflowers)  Josephine You\'re so good to me And I know It ain\'t easy Josephine You\'re so sweet You must taste just like sugar & tangerines  I won\'t make a sound Sleep on the ground When you wake I will Drive you into town I missed your smile Your schoolgirl style But I never had much fun Maybe the very first mile  Don\'t you know I watched you walkin\' home from school Your friends on the old playgrounds You never looked so down Won\'t you come and help me  with these cuts of mine? I\'ve disconnected my heart And cut myself  on the wires, Josephine  I know I was wrong I knew all along But I got so far from my home I never thought I\'d be so lonesome  '),(9172,1240274,'27,June,2003','2003-06-27 00:00:00','   alright. starting to seem like a freak --just read my last few posts and i think i look like an angry geek. no bodies in my basement, i swear!  anyhoo. i averted financial destruction for another month, hooray!  newbrunswick is cancelled for this summer, and replaced with radiohead in august. seriously wondering if there is a link between stress and dyslexia \'cause i\'m not right these days. yeah, weeeeeeeeeeee !!   dying to get the music out of my head and onto hard disk. wish i could use acid4.0 but it isn\'t like photoshop!!?!?!  everyone i know wants a bike. the urlLink  latest dream ?   yeah. that\'s about a dozen friends who want one. i got money on rick and dan, the rest, dream on :P  i want one too, but something more superhero like ;) with a sidecar! egh.  ... ...  i work every damn night.  12- 7:30am.  buddies are out of town. insurance is no longer on my car so i got no way to get around. gonna\' be a lonely summer for this boy. . . why won\'t girls like boys unless they got\'s a car (that is legally allowed to leave the driveway)?!?  -me.  '),(9173,1240274,'23,June,2003','2003-06-23 00:00:00','   Everyone knows the song \"Loser\" by Beck, right? Well, I woke up to find this; (computer geeks need only see this)  Props to Kevin Hughes, the author.    User.  In the day of sysop nerds I was a flunkie Jolt in my brains and body feeling chunky With the plastic mouse balls spray paint the Commodore System install with the hard drive on the floor  Kill the process and put it in /dev/null Email flaming with the user hitting D-control Shell\'s called Reno and it\'s written in C Got a couple of xterms, keys set to repeat   Root came sayin\' I\'m insane to complain About an online wedding and a stain on my screen Don\'t believe everything that you make(1) You get a cracker from Europe and a login that\'s fake   So write your code in Perl in the dark Saving all your hacks for working at a tech park Yo - punch it   So - dumping core I\'m a user, baby, so why don\'t you kill(1) me? (Double dense floppy) So - dumping core I\'m a user, baby, so why don\'t you kill(1) me?   Forces of evil in a MUD/MOO nightmare Ban all the members in a phony #chat channel \'cause One\'s got a handle and the other\'s got a .plan One online spammed the other and ran   With the FTP and the insane print job The daytime crap of the alt.test slob He hung himself with a call to ping Twenty milliseconds and it\'s spitting out another string   RTFM if you can\'t relate Trade the Sun for a car and the Web for a date And MIME is a nifty hack for mailing to a newbie That\'s choking on my MPEGs   So - dumping core I\'m a user, baby, so why don\'t you kill(1) me? (Get crazy with the caps lock) So - dumping core I\'m a user, baby, so why don\'t you kill(1) me? (Drive-by BIFF post)   ... Yo, bring it on down ... I\'m a hacker, I\'m a winner Program\'s gonna work, I can feel it   So - dumping core I\'m a user, baby, so why don\'t you kill(1) me? (I can\'t retrieve you) So - dumping core I\'m a user, baby, so why don\'t you kill(1) me? (NULL) So - dumping core I\'m a user, baby, so why don\'t you kill(1) me? (Sprecken sie DOS, eh, baby) So - dumping core I\'m a user, baby, so why don\'t you kill(1) me? (Know what I\'m typin\'?)   ---eof---  '),(9174,1240274,'05,July,2003','2003-07-05 00:00:00','   so ou think you can tell. heaven from helllllllllllll............do you think you can telll. did they get you to trade................. blaw  '),(9175,1240274,'01,July,2003','2003-07-01 00:00:00','   lots of thought being put into a short film i wanna\' make. think tragic, starcrossed lover syndrome, suicide, triumph and failure, the bland and the abstract. perhaps a wedding¿ it\'s gonna\' be about 10 minutes long, and the plot won\'t be understood by most.. never made anything with video before, so it\'s gonna\' be a learning experience. no dialogue, and it\'s to the beat of  Viðrar Vel Til . by  Sigur Ros . the music is key for this. . .  by autumn i\'d like to start something up (in the way of a vid or anything promotional) for my buddy danoh, a great musician without an audience --yet. if down the line you see what you like, got some ideas and some musical talent of your own, drop me a line? i\'d be glad to help ye. stay tuned?  in other news, i caught some headlines this week, about \'the mother that was never born\'.. apparently the unborn can have children too. that is, an aborted fetus can be used in the genetic manufacturing of another babe that  isn\'t  a clone. i see a plot begging to be used, no? one has to wonder where researchers even acquire these fetuses.. it\'s like circumcisions. are hospitals allowed to do whatever they want with your biological \'waste\' (for a lack of a better word) --your dna? all these questions. . .  my folks take off today as of 8:30 am, for a week. this means i finally get some time alone to relax. write. walk around in the buff, and live vicariously for a bit --wo0t! it\'ll still be hard to get back the feeling of living alone in a big place --with all the old people shit around the house-- like windmills and gnomes in the garden, ducky bedroom border, and country garb hanging in the kitchen. aw well. no parentals = fine with me :|  meh. time to run away, if you\'re reading this check my site:  urlLink www.serotoninambition.cjb.net  and sign my brand spankin\' new  urlLink guestbook .  cheers! m.a.  '),(9176,1240274,'01,July,2003','2003-07-01 00:00:00','   my boy Jay is getting married --the biggest news since sliced bread !!  '),(9177,1240274,'12,July,2003','2003-07-12 00:00:00','   [ :: Monday, March 10, 2003 :: ]  just a brick, too soon gone, prevents you from being, and doing, anything along the way of carrying on.  just a prick, with his head on wrong. doesn\'t understand the why,  and WHO he is.  he keeps awful control of the ship he\'s on.   just a kid, too soon out of favour, wonders why he\'s alone, confined, stuck on this one-way street he\'s spawned.   just a hypocrite, he tries to be Christian, but his ways are too wicked. so here and there you see him stumble. never aware of where he\'s been, or where he\'s going to.   just confused, and all mixed up, wishing he was something he already is.   neurotic and thick, always craving loves lust, and daily adventure wishes he was with her, yet then he\'d be too soon tied.   so this is what he does instead prays to God to cease the demons, screaming, wrenching, fighting in his head. lives every day, the same as the last, forgets the numbers that just don\'t add up, wishes he was someone else, someone in love, wishes his bride would be sent from above.   ignoring the piles so high, the time so quickly to pass this boy by.   what will he do? where will he go?   when the outside world turns callow, cruel, and cold. he\'d have been a lover long ago, a success, a caregiver,  on a path of envy, righteousness, and husbandry.   if only he saw the fork up ahead, when he would have to decide with his head; ignorance, for love and happiness. ..or misery for the real, unraptured society.   no fairytale ending in sight, no release to his plight, unable to tread life\'s scorn, he took his own life instead.   . . .unaware of the ones he stole with it. . .   michaelanton.  '),(9178,1240274,'11,July,2003','2003-07-11 00:00:00','   sigh. there are a few problems with web logs.. like --you\'ll always have that friend that will mock you. call it a web diary. make you cry. o0o0oh the sting. heh,  or NOT . naw, the real issue lies with not being able to discuss specific matters for the possible chance T H E Y might see it. not gossip, \'cause you don\'t use names, but T H E Y know ;) kinda\' blows..  coulda\' dropped a really cool story today.  michaelanton.  '),(9179,1240274,'11,July,2003','2003-07-11 00:00:00','   a long time ago, when i was young and stupid, i broke a girls heart, and broke my own in the process...  a song that\'s close to me; \"life is sweet\" by Natalie Merchant. hope it can mean something for you too  .m.a.  '),(9180,1240274,'09,July,2003','2003-07-09 00:00:00','   If you tolerate this, then your children will be next.  '),(9181,1240274,'08,July,2003','2003-07-08 00:00:00','   it was then that i saw the white bag, flash coming down the road  it was then that i felt what it was to be not so innocent, yet so very naive the essence of me  always so outward willing to please standing for others when they were down on their knees watching adhering walking the hard lines in between the fray so many years it would be this way  ..so many years to pay for it dearly.  little could i have known, of the docterine of the severe the slanderers that make up that jury.  __michael anton__  '),(9182,1240274,'07,July,2003','2003-07-07 00:00:00','   getting real tired of the hard lines. . .   you recall teenage mutant ninja turtles for nintendo? well, you know that level where you\'re under water trying to deactivate bombs and avoid the electrified weedy things? well, i feel like i am on my 100th try and at the end of my rope --right down to the sweaty palms and shaked out reflexes.  '),(9183,1240274,'07,July,2003','2003-07-07 00:00:00','   urlLink The Lure of Data: Is It Addictive? : \".....They become frustrated with long-term projects, thrive on the stress of constant fixes of information, and physically crave the bursts of stimulation from checking e-mail or voice mail or answering the phone ..call it pseudo-A.D.D.\"  I motion to get a cellphone that is only capable of outgoing calls. Only then will I be able to have a discussion with a group of buddies, without some twat breaking it up to tell us who\'s on Letterman.  '),(9184,1240274,'06,July,2003','2003-07-06 00:00:00','   urlLink Slashdot: News for nerds, stuff that matters : \"Take away the elements in order of apparent non-importance \"  '),(9185,1240274,'06,July,2003','2003-07-06 00:00:00','   the doors. \'love me two times\'. sunday morning. 10:14 am.    as two friends get it on in my basement. rock on. and me? in the kitchen. on my pc. writing in my web log. oh yea.  so, i stuck it to the man, 5th time coming. last time i checked i didn\'t work sundays. or was i ever able to unload 60 crates in 60 minutes. when your boss\' name is guido, \'stickin\' it to the man\' is even more fun. yep.  the doors. \'people are strange\'. sunday morning. 10:17 am.  so. i haven\'t slept yet in awhile. haven\'t really slept in years. egh.  ...  the doors. \'riders on the storm\'. mmmmmmmm. oh yeah.  -m.anton.  '),(9186,1240274,'19,July,2003','2003-07-19 00:00:00','   Yesterday I got a ride to my new placement through my temp agency, at Stanley Tools in Oakville. Never takes long to get there on the highway (it\'s just past Trafalgar, and I started from the neighborhood of IKEA in Burlington. Now that my own car is off the road, I am learning a few things about time when travelling. Making myself suffer enough to know better to let things slide --as far as my car goes, anyway :P ..so I asked me mum to drop my bike off and I rode home. One hour and a half taking Lakeshore, the more scenic route.   Riding along, you start to notice more. The creeks you never knew were there. The great shots you could take with your camera, when the sun seethes through the trees. The amazing graffitti art under ancient bridges. ..and the little naunces along the way to entertain your mind. Like how you can tell a fellow biker, even if they\'re in a car. How old ladies think I am a punk kid on a bike. How the Catholic School Board is so easy to take a jab at. Ok, ok, I gotta\' elabourate a little on this one I guess..   I ride by the location of where I beleive they are going build a new school. The only thing I notice on the sign in front of the destroyed trees is \"Believing in you, believing in us\" --or something to that effect, I\'m not sure how fast I was going when I flew by it, but... yeah. What about the \"Believing in God\" part? .. but FACK, 5:30am, gotta\' jett if I wanna be at work for 7. Cheers. . .  '),(9187,1240274,'17,July,2003','2003-07-17 00:00:00','   sometimes a good way to relish a great feeling.  '),(9188,1240274,'16,July,2003','2003-07-16 00:00:00','   right now i am sitting at my desk, in my living room, in calgary. thinking over choices i will make, ones i made. where to next. which cafe i will visit in the morning to write the letters i already wrote. feeling so....  sigur ros . nothing can touch me now, in my world above the trees, far from my past lives, half lived in caves, and the rest in loss of mind, doped up. asleep.  i couldn\'t sleep tonight. i was haunted by Elisabeth, and how she made me feel the second day in calgary. how angry she made me. how things were left rotten and so... not right. i keep going into my past and pulling people out like teeth, dusting em\' off, and placing them before me to fix the way i had left them. i hate bad blood between myself, and people i may never see again. it just sits wrong and even years later, disrupts my happy world, and troubled sleep. keeps me from thinking about what comes next, but instead what shouldn\'t have been at all? or maybe, ... ....  i sometimes feel like God Himself shows me His hand once in awhile. whether it be my gut, or sixth sense, i know what\'s gonna\' come next. or what\'s supposed to. what shouldn\'t. i know everyone has thoughts like this, only different,  i can\'t explain it. it\'s like nothing else. it\'s intense. i sort of feel that i lost hold of things, and thought after seeing the cards, that things would still play out the way they should. i got cocky. confused. didn\'t know why things were happening the way they were --it was never supposed to take such a u-turn. it isn\'t so much that this is excess baggage, but more so that i am convinced she felt my sleeplessness, a year later, tonight. she felt it not because she isn\'t over me, or me for her. but because it played out wrong. i think deep down, we both would agree.   if the final moments we had shared were a painting -there was too much red, and not enough yellow, green and purple...  the rest of the faces come back, and the lines aren\'t as accurate as i\'d hoped for. the stresses seemed so important at the time, and i lost touch with where the fuck i was, and why i was there. what i was really doing. i\'d kick myself for coming home if i could allow my pride to dwindle a little more.   these days i sort of feel like people are getting the right idea now. i was getting real sick of always being a projection of their own insecurities. it ate at me a lot, and made me start to beleive i was what they manifested me to be. glad i got a break.  right now i am sitting at my desk, in my bedroom, in burlington. thinking over choices i will make, ones i made. where to next.  '),(9189,1240274,'15,July,2003','2003-07-15 00:00:00','   today i looked outside my window and i saw the strangest  urlLink thing .  '),(9190,1240274,'26,July,2003','2003-07-26 00:00:00','   urlLink Yahoo! News - U.S. Displays Hussein Brothers\' Bodies : \"U.S. Displays Hussein Brothers\' Bodies \"  '),(9191,1240274,'25,July,2003','2003-07-25 00:00:00','   UPDATE:   urlLink EFF: RIAA Subpoena Database : \"Concerned that information about your file-sharing username may have been subpoenaed by the RIAA? Check here to see if your username is on one of the subpoenas filed with the D.C. District Court. This information is drawn from the court\'s publicly available PACER database and will be updated when that system is updated. \"  ..if you\'re american, and you\'ve used kazaa, kazaa lite, grokster, etc, type in your username, and click \"check\". little easier...  '),(9192,1240274,'25,July,2003','2003-07-25 00:00:00','   Yeah. I am really starting to dig this  urlLink Google toolbar with Blogger add on . I\'m a big fan of  urlLink Slashdot , and i found yet again, something that totally blew me away when i woke up today (at 4am! wo0t!)   urlLink Terra Wind Home : \"Introducing the world\'s first luxury Amphibious Motor coach / Yacht\"  Yep, yep. Anyhoo. Through a twist of fate I do in fact have a car now. Go figure --yesterday, when we went to see how long the registration place would be open for today (Friday) it turned out the place had extended hours Thursdays only ~until 7pm.  mmmm. driving to work. something i took SO for granted because i always worked up the street. now i work in oakville, and biking an hour and a half wasn\'t in the cards for long. neither was taking the go-train/buses for an hour and 45 mins. i beleive the term would be \"blaurgh\"...  right andy ?  with amon tobin on my ears and an  urlLink XENARC CP-1000  in my web browser, i drool.  meh. time to clean my room and drink more coffee. latz. mikeanton.  ..p.s. can you beleive that the USA put photos of the corpses of Saddam\'s kids on the web? corpses. i mean photo\'s could still lie, it could be propaganda..... but either way it\'s a fucking sick world we live in when my unborn kids can hop on to sites where the government puts up shit like that? not going into the war, i think the news has also covered a lot of the lies being spread (i had my hunch since the get-go) by Bush and Blair, and the death of that weapons inspector reaks of controversy........ all i wanna\' point out is that \'we\' think the other smaller, less well off countries involved are being brainwashed by their media. Guess my second point.  '),(9193,1240274,'24,July,2003','2003-07-24 00:00:00','   urlLink TechTV | RIAA Hit List : \"The Recording Industry Association of America (RIAA) has now issued more than 911 subpoenas to Internet service providers across the United States, trying to get the names of people still offering music on file-sharing networks such as KaZaA and Grokster.\"  Yeah, so who digs Canada now? On this article is an actual list of some of the user names that are gonna\' get nailed. Hoping there\'s a few lawyers in there too. . . I\'d say that for myself, I am beginning to feel the immorality of downloading music. The stakes are higher for you Americans out there too! I know that the artists already take it in the ass on their end money/contract wise. playing \"robinhood\" with your music collection is a low-class way to justify yourself as well, as there are millions of us leaving much more than a dent in their pocket books. . . not saying i won\'t download ever again but I am not gonna\' share my whole collection online anymore, that\'s for sure. although, because of this whole mp3 fiasco, ever notice how the smarter labels are starting to offer DVD discs along with music cd\'s for the same price? because of this we ARE getting more for our money. i say twist the knife a little longer and make that stick. then cease the madness of downloading whole albums just to have that complete collection, and cut it down to just making wikkidd mixes. if it wasn\'t for the mp3 and web broadcasting, i wouldn\'t know of half the bands i do now, so it is great free advertising. the sound quality IS lacking as well, so the real store-bought deal is always gonna\' be better anyway, right? just my thoughts... moderation is key. it\'s the little kiddies that go over board that are killin\' it for everyone.  '),(9194,1240274,'22,July,2003','2003-07-22 00:00:00','   for a minute there i bet everyone thought i was being a drama monger. heh. fack.    anyhoo , getting my car back on the road whether i can afford it or not. me = mobile hopefully by the end of the week. now with natalie ( natalie = my car.. don\'t ask..yet  ;) by my side, and my mountain bike? i am gonna rip shit up. for those who know me --big step taken today. faced my demons. called VISA for the first time since I got back from Calgary. (in JANUARY). yeah. freaky freaky ass shit. no more to be said about that!   meh. gravol kicking in. must... wake up... 6 hours... weeee! when i get my car i get to awaken around 6:45am --can\'t wait. better than 5:30am like now. -mikeanton.  p.s. *cough* *cough*  jenn should update her livejournal . *cough*  '),(9195,1240274,'21,July,2003','2003-07-21 00:00:00','   jann arden: nevermind   Never mind what you think you know about me Never mind what you might have seen on the TV You can\'t believe every single word you\'re gonna read What I am is too far in for you to see  Never mind who said that I was a bit far gone Never mind when they tell you sex is all I want You can\'t believe that I\'m every word that I write down What I am is too far in and can\'t be found  Never mind what they\'re gonna say is probably wrong Never mind it ain\'t about the singer it\'s the song You can\'t believe how a person looks is who they are What I am is too far in and can\'t be found     natalie merchant: break your heart   Oh people downcast in despair See the disillusion everywhere Hoping that their luck will change It\'s a little harder every day People struggle people fight For the simple pleasures in their lives The trouble comes from everywhere It\'s a little more than you can bear I know it will be hurt I know it will break your heart the way things are And the way they\'ve been And the way they\'ve always been People shallow self absorbed See the push and shove for their rewards With nothing nice on their minds You can read about it in their eyes People ruthless, people cruel The damage that some people do Full of hatred, full of pride It\'s enough to make you lose your mind I know that it will hurt I know that it will break your heart the way things are And the way they\'ve been Yeah I know that it will hurt I know it will hurt I know that it will break your heart The way things are and the way they\'ve been Don\'t spread discontent, don\'t spread the lies Don\'t make the same mistakes with your own life You know it will always come back Yeah, I know that it will hurt I know that it will hurt I know that it will break your heart The way things are and the way they\'ve been Don\'t spread discontent, don\'t spread the lies Don\'t make the same mistakes with your own life And don\'t disrespect yourself, don\'t lose your pride And don\'t think everybody is gonna choose your side Oh no... Oh no... no..  '),(9196,1240274,'21,July,2003','2003-07-21 00:00:00','   a little confused. yesterday night i sort of went loco. when i used to get all neurotic, i referred to those moments as \"spikes\". like my mind/heart gets all tangled up and tries to go two ways at once. some people take meds for this sort of behaviour. hell, i even used to.. before the day was out, i cut deep into my right hand man\'s head with negative. the things we sometimes think, but never say --i said, totally forgetting the reasons why we don\'t say \'em. i wanted a stir. wanted the world to catch the drift i was on, and it wasn\'t working. so instead it flipped around on me and before i caught it i ate my own words. . then there\'s my other buddy. in a hospital bed, and for some reason i\'m totally confounded as to why i am so numb/cold. it was inhuman. i woulda\' been better off starting a fight and getting my ass kicked by a punk kid, than saying such bullshit to my boy. egh.   wish i could stay in character all the time, and i wish i could find some sort of equilibrium. i can\'t be this one person 6 days a week, and then switch it up, and get all condescending and harsh the next. it\'s not me. it\'s not fair. wish i wasn\'t such a fucking two faced pompous ass. it\'s times like this i sort of need though .. i guess. keeps me humble, on my toes, and in focus. i wanna\' go back to the days when i could look in the mirror and be the  only one  who can see the complete lack of humanity, cold, and ugly. when other people see it it hurts so much. . .  '),(9197,1240274,'21,July,2003','2003-07-21 00:00:00','   urlLink BBC NEWS | Technology | Cyber sex lures love cheats : \"Growing numbers of married people are turning to internet chat rooms for sexual thrills, a US study has found.\" ..that\'s really too sad. *sigh* i always wonder what the cards hold for me. is the whole process of meeting/dating/falling in love/marriage old fashioned?  '),(9198,1240274,'20,July,2003','2003-07-20 00:00:00','   urlLink Future Shop Web Store - : \"Wacom Steel Graphire 2 4X5 USB Tablet \"  yes folks, i do believe i wet myself.  '),(9199,1240274,'20,July,2003','2003-07-20 00:00:00','   aghahgahghaghgahahhgahgahghghghaggaggaggagagagag. stress.   ....new site layout and new work coming soon. wo0t!  '),(9200,1240274,'29,July,2003','2003-07-29 00:00:00','   urlLink BBC NEWS | Entertainment | Jackson attacks music piracy bill : \"The Authors, Consumer and Computer Owners Protection and Security Act of 2003 - known as the Accops Act - carries penalties of up to five years in prison and a $250,000 fine for uploading a copyrighted file to a peer-to-peer (P2P) network\" -- in the USA . What\'s wrong with this? I\'ve seen murderers get off with less time served. America; why don\'t you contemplate letting Canada take over?  woo Canada! What happens when parents are too controlling of their children? What happens when a super-power democratic government becomes a beligerent dictatorship?  '),(9201,1240274,'27,July,2003','2003-07-27 00:00:00','   urlLink The Year Of The Rooster :   1921, 1933, 1945, 1957, 1969,  1981 , 1993, 2005 People born in the Year of the Rooster are deep thinkers, capable, and talented. They like to be busy and are devoted beyond their capabilities and are deeply disappointed if they fail. People born in the Rooster Year are often a bit eccentric, and often have rather difficult relationship with others. They always think they are right and usually are! They frequently are loners and though they give the outward impression of being adventurous, they are timid. Rooster people¡¦s emotions like their fortunes, swing very high to very low. They can be selfish and too outspoken, but are always interesting and can be extremely brave. They are most compatible with Ox, Snake, and Dragon.\"   i\'m gonna\' leave this one alone i think. -mikeanton.  '),(9202,1240274,'27,July,2003','2003-07-27 00:00:00','   urlLink BBC NEWS | Europe | World War II message washes up : \"A bottle containing a message from a refugee fleeing the Nazis during World War II has been found on a remote Swedish beach. \" madness.  '),(9203,1240274,'08,August,2003','2003-08-08 00:00:00','   damn. it is great to finally realize that even with my dialup net connection i can still listen to  urlLink somafm  (groove salad :  urlLink dialup  or  urlLink cable ). i missed it so. been a long time for me and  urlLink somafm . yep. i think maybe since the beginning. lots of cool beats to be heard. on  urlLink somafm . yes, yes. gotta love that soma... urlLink fm .  urlLink 24kbps  of mmm-mmm yummy with a little kick of y0wzAH.  second rockin\' point on this one would be that this time next week, i will be kickin it down hard with the party to commemorate the radiohead concert i will be attending the following night. thats two pleasin\' reasons to feel high. go slow. and enjoy the flow. wiggit-wiggit-wo0t.       ...or not.  mikeanton  :D  '),(9204,1240274,'06,August,2003','2003-08-06 00:00:00','   horseback riding. moving to a new apartment. moving out of that new apartment. where is michael anton now? wondering why my friendster list got shorter in my weeklong absence. wondering lots of stuff. working at a new gig; a lumber yard. go figure  :|     ..also wishing \'my best friends wedding\' ended better. or at least actually remembering the ending would be alright.. but i think the guy didn\'t get the girl. you see that sucks, cause one day i am gonna be in that boat and it doesn\'t give me much hope for the future. . . wish i was more    urlLink fearless  .  wish how i felt had some sort of bearing on reality. wish i was in calgary.  '),(9205,1240274,'16,August,2003','2003-08-16 00:00:00','   sigh. yeah. so we called about the radiohead concert yesterday. apparently it was still a go, then. but now. after a few train rides, joints, etc. it was cancelled. how horrible.   -me.  '),(9206,1240274,'14,August,2003','2003-08-14 00:00:00','   thats odd. not one person is on msn right now..  reading august 10th\'s entry, and then driving about my city tonight was all too crazy. people fluttering around town, buying up all the gas/food/water they could. fat people rescuing the icecream. oh. and batteries. for those unnaffected, and haven\'t heard about it in the news already,  urlLink big blackout  along the east coast. yep. although from the time i got home from work, until now, my power has mircaculously been unaffected ;)   i sort of crave this sort of thing --the destruction of this western empire we live in. acts of chaos and terror all over the place. maybe some sort of fallout. then the terror ends, and you come out of hiding. . . bartering for gas. everything is looted, or being looted. finally all that matters is those you love and survival. then small new communities are formed among the destroyed former civilization. you form packs, and everyone must play a role to the greater good of the group. protectors. food gatherers. cooks, babysitters and mechanics. no such thing as taxes, rrsps, visa, tommy hilfiger. people would have to learn again. how to fix things. establish new homes/camps. not be so dependant on the automatic. hopefully learn to never go back to the shit hole we made life to be. no more pills.. (condoms would be high priced commodoties in the underground market :| that sucks). but nonetheless, you get the point. after this example of how easily our security in buildings, purchasing abilities, mobility, drinking water, etc... can be knocked out by a power failure. ever think about the unexpected shit that might happen? chaos, without power, the mad carnal nature of people to go to arms when chaos and anarchy is in the air?   the musicians of the world sing of it. the bible, and other prophesies speak of it  . .  but i wonder how it will really go down. by no means am i trying to play on current events to incite a panic; it\'s just quite simply a thought not to write off because it is disturbing.. always wanted a gasmask in my closet or trunk for days like those.  mikenaton.   urlLink [ _radiohead: one more work day_]  '),(9207,1240274,'10,August,2003','2003-08-10 00:00:00','   urlLink Cloning Yields Human-Rabbit Hybrid Embryo (washingtonpost.com) : \"Scientists in China have, for the first time, used cloning techniques to create hybrid embryos that contain a mix of DNA from both humans and rabbits, according to a report in a scientific journal that has reignited the smoldering ethics debate over cloning research.\"  '),(9208,1240274,'10,August,2003','2003-08-10 00:00:00','   \" you must be on your guard. you will be handed over to the local councils and flogged in the synagogues. on account of me you will stand before governors and kings as witnesses to them. . . . . .    brother will betray brother to death, and a father his child. children will rebel against their parents and have them put to death. all men will hate you because of me, but he who stands firm in the end will be saved.\"  when you see \'the abomination that cases desolation\' standing where it does not belong --let the reader understand-- then let those who are in judea flee to the mountains. let no one on the  roof  of his house go down or enter the house to take anything out. let no one in the field go back to get his cloak. how dreadful it will be in those days for pregnant women and nursing mothers! pray that this will not take place in the winter, because those will be days of distress unequaled from the beginning, when God created the world, until now --and never to be equaled again. if the Lord had not cut short those days, no one would survive. but for the sake of the elect, whom he has chosen, he has shortened them. at that time if anyone says to you \'look! here is the Christ!\' or \'look, there he is!\' do not believe it. for false Christs and false prophets will appear and perform signs and miracles to decieve the elect -if that were possible.   so be on your guard; i have told you everything ahead of time .   \'but in those days, following that distress, the sun will be darkened, and the moon will not give its light; the stars will fall from the sky, and the heavenly bodies will be shaken.\'   at that time men will see the Son of Man coming in clouds with great power and glory. and he will send his angels and gather his elect from the four winds, from the ends of the earth to the ends of the heavens.     now learn this lesson from the fig tree: as soon as its twigs get tender and its leaves come out, you know that summer is near. even so, when you see these things happening, you know that it is near,  right at the door  . \"   Mark 13 : 14-29. NIV translation.  '),(9209,1240274,'10,August,2003','2003-08-10 00:00:00','   ever feel like your time to catch up is over, and ya\' didn\'t make it? i don\'t want to work at a lumber yard (for a complete prick) forever. i don\'t want to tell my kids \"go to school, don\'t turn out like i did.. biggest mistake i ever made was not going back after highschool\". i won\'t beleive that my life will turn out that way. yet.. it seems  so fucking likely.   -michaelanton.  '),(9210,1240274,'23,August,2003','2003-08-23 00:00:00','   urlLink Nuclear Power Plant Bombing? : \"TORONTO (Reuters) - Canadian police arrested 19 men last week in a case that, according to court documents obtained by a newspaper, has eerie parallels to the preparations for the Sept. 11 attacks on the United States.\"  '),(9211,1240274,'22,August,2003','2003-08-22 00:00:00','   ok. enough already.  really . 3 posts in a day is a little much. so. regardless of that fact, as i write this a wee bit too much wine is goin on and uh. well. i am drunk. nothing to be proud of --but hey. what can i say now? lots! instead of getting an oil change i spent my night talking to cool people on msn, and looking for my golden retriever. yup. i left the gate open after cutting my front lawn and he ran away. got him back though, due to cool, resourceful people who called the spca, with his dog tag number, and got my fone number and called me.  yup.  as for the coolness i was GOING to write. well. too bombed. too tired to write. stay tuned? i know you will. you\'re a little hooked. even in the least. admit it ;) .............or N O T. either way. i am not as cocky as i sound. but please come back another day...........!?? ? i suck.  mikeanton.  '),(9212,1240274,'19,August,2003','2003-08-19 00:00:00','   Pointless legal proceedings: - Topless law in the Greater Toronto Area.  After all that shite , i have seen two topless women. A stupid law fought by stupid people? Stupid people who later chose to not even make use of it themselves? (Although I am NOT against equal rights. I am all for equality of the sexes. if a guy can do it, women can too ..as far as i am concerned. i just don\'t think it\'s cool for \'auntie faye\' to let my non-existent children to see her knockers ;)  - The whole pot thing. In the GTA we can have 15 grahams with a slap on the wrist. WIKKIDD cool. ( I have since quit the habit , but . . ) why make the consequences for selling it way worse than before in the process?  --when the government is allowing it for medicinal use? fucked in the head, right?  - now look across the border where culture is wasted. culture = history = our roots. religion is as much part of our culture as our ancestors themelves are, right? what they beleived in affected everything from how we buried our dead, cleaned up human waste, wrote music, communicate. . . all the way to our present day laws. that said, i think most people realize that no matter how anti-religious one may be, you respect (most) our laws. the very same laws founded upon Christian beliefs (ie. the ten commandments). You can\'t erase that. It\'s already history. it happened. forget the God part if you feel the need to and look at it as purely a cultural thing. i choose not to look at it that way, but that\'s just me. even if you take away God from our everyday life, like our society has chosen to do, you cannot unwrite the laws we\'ve founded that are based on the bible, because they  make sense .   like a historical statue at a lookout at a  \'window to the lake\' . ..like a statue of old war \'heroes\'.  why  is a statue of the ten commandments in front of a courthouse in the  united  (yah. right.) states of america ILLEGAL...? Go see for yourself. Christian, Buddhist, Catholic, or even if you\'re into Osho.. You can\'t conjure up enough of an argument in agreeance to this..   urlLink CNN.com - Aug. 22, 2003 : \"Judge suspended over Ten Commandments\"  '),(9213,1240274,'19,August,2003','2003-08-19 00:00:00','   well. it is 6:27pm.  Pyramid Song . A glass of California wine, and pizza. how spoiled am i? meh. not really into going into actual details of my day to day, but egad! i got to get an oil change tonight (3 changes behind...or wait. four? uh oh. . ) then its back to cut the grass. clean out my car. clean my room. do laundry. drink more wine. dream more of the radiohead show i didn\'t see. then pay bills. then sleep for tommorrow i work. mmm.  punch-up at a wedding  --another sweet radiohead song. sigh. i got it bad..  work. the bine of my existence. it doesn\'t help that i work for a stressed out and greedy yappy dog. the most annoying part? could it be getting called into  johnny\'s  office for punching in 2 minutes late? or the fact that i am treated like a moron by a Bill Paxton-esque hick boy (referring to his role in Aliens (Pt. 2) \"Gaaamee overr man --G A M E  O V E R. yeah. he talks like that. Looking like Howard the duck doesn\'t help his case either.) or maybe it\'s the sun and heat of the lumber yard.. nope. it\'s  the hick .   rant over ;) sorry kids!  (not too) funny how only the negative and lonliness get transferred over the the world wide web, as i tend to leave out the good stuff in my life, so maybe stay tuned. when i start to feel the love later on tonight, from the wine, i wanna\' write some happy joy. wo0t. suprized how many (cute, cool, and mysterious) folks suprize me by referring back to the things i write, especially when i haven\'t talked to \'em in awhile, .. or ever. feedback is always welcome, wanted, and valued.  take it easy kids,   mikeanton.   no really.  '),(9214,1240274,'19,August,2003','2003-08-19 00:00:00','   oh.        my.  it\'s pretty ..well  it\'s something else  when you play social pinball. most fond memories of meeting new people, and taking in new experiences has always started off going out solo. So it can be seen how i might thrive on it.. it was one of those nights where you expect to wake up early cause there is no way  these people  are talking to you, let alone enjoying your conversation. then there was the game of pool.  i don\'t play pool. i lose pool. i hate pool.  well. i beat my opponent. an opponent that happened to be one of three beautifil women. i was reeling, and there\'s more but i stop short. andrew --glad you saw the game of pool \'cause dan isn\'t going to beleive me. . .  thanks for the conversation kat. -mikeanton  '),(9215,1240274,'17,August,2003','2003-08-17 00:00:00','   for the first time i find myself not longing for the past.  hooked on nostalgia  --wishing i was someplace i\'d been before. without thinking of the future, i wish i\'d manifested some sort of stable relationship by now. i want to hold out for the next one. that one girl that comes from the very hand of God to touch my life. the one that was always meant for me. the one maybe i\'ve walked along with before, lifetimes ago. the click i feel starts off almost like deja vu. then it\'s an unspoken, seemingly pre-existing bond. a common plane we\'re both on. the more we talk, the more the story unfolds before us..   getting itchy feet again as well. last year at this time i just started debating the little thought-of question of calgary. it wasn\'t until september that i made that come to fruition, and by that time this year the strain of the need to stretch will be unbearable. whether it be getting out on my own in burlington, or hopping on a plane some place at random.. again. it will be done right. my mind is meandering around this decision, and will for some time. when the call comes from inside me someplace, it will be uprupt. it will be absolute and without warning. i can\'t live here anymore. on this street, in this skin, with these choices, people, and mindfuck.   have you hugged a stranger lately?   mikeanton.  '),(9216,1240274,'17,August,2003','2003-08-17 00:00:00','   \"Better now than never, better loud than clever Better just to play the fool It\'s times like this When you just close your eyes and kiss Cause everything after this Is just bullshit and being cruel\"  -matthew good.  '),(9217,1240274,'30,August,2003','2003-08-30 00:00:00','   ah yes.  the summertime/going away house party.  it brings out so much in the many. the jealous boyfriend. the happy drunk. the superstar and the expert. the nurturing. the sexpot. the fooseball swindlers. the untold truths. admissions of loves past, and broken hearts. the lamb and the lion talk over a smoke and a beer while your mother doesn\'t know where you are. mystery photos turn up on cameras. grass stains, and muddy shoes. new unions are made, and brital ones are rocked a little more...   michael anton.  '),(9218,1240274,'29,August,2003','2003-08-29 00:00:00','   ever notice how when muttered, \'i love you\' sort of sounds like \'i\'ll leave you\'?   radiohead power hour .  \'there there\' , \'be quiet and drive\', \'lurgee\', \'thinking about you\', \'how to disapear completely\', \'a punch up at a wedding\', \'i might be wrong\', \'fake plastic trees\', \'rabbit in your headlight\', \'climbing the walls\' (remixed with  sneakerpimps  ¿?). . . mmm.  '),(9219,1240274,'27,August,2003','2003-08-27 00:00:00','   urlLink Sheridan - School Of Animation, Arts & Design - New Media Design : \"Innovation in concept, interface and usability design is ongoing. In a business where change is the only constant, the ability to adapt to new technologies, audiences and applications is vital. Sheridan’s New Media Design Program will enable students to extend their existing digital and traditional visual communication skills to meet these challenges. In-depth instruction in design, methodology, Web authoring, software and technology is combined with intensive project work in a dedicated studio/lab. The program utilizes a wide range of software applications, including products from Adobe and Macromedia. \" --or  urlLink something  like it.  wish i had some sort of \"great expectations\" (charles dickens) life. annonymous financiers. mystery. travel. adventure. ...or marry an old bag, and get all her money to send me back to school.   ..found out my pay raise won\'t take effect for an extra month or two. no stoney creek apartment with studio for mikey this year. yeah. for those who know me well enough to know i had it in mind, i\'m hurting inside.  our lady peace;  naveed . (the only good album) are you there, and is it comfortable? did you  want to escape, try to escape the population ? the pressure is deceiving, and for you particularly should we let a young man die?  let him die if he wants to? I can\'t live here anymore it\'s hard when you reach for that floor and and there\'s something that tears me inside, so I can\'t go  brother are you there?  now tell me is it wonderful  or were you hoping to find something a little more?  climbing down the hours I need to know now do the hands of time only rule this chapter  i\'ll have to try once again, i\'ll have to try when I want to  there he\'s on his knees again trying hard to understand why Naveed would let a young man die convinced that he might break he reaches for that phone and then another day has gone  '),(9220,1240274,'26,August,2003','2003-08-26 00:00:00','   urlLink Melissa : \"I\'m as smart as the next one, as naive as your first, as skilled as you promise to be and as realistic as one could hope. \" cool quote.  '),(9221,1240274,'25,August,2003','2003-08-25 00:00:00','   stepping on the brakes never felt so good. . .  today was the first day of carpooling to work with a fellow whom i had hooked up with a job; temping at the lumber yard. as he drove, i got a chance to look out my window, listen to the music  he  played on his stereo, (somewhat a bonus, considering mine was stolen pre-calgary) and lose myself in thought.  talking cars never really interested me. neither did stupid music about getting \"to the centre of oh, ooh\", and fucking men of many races. what is the profound message in that?  that you\'ve got the clap?  what gets me most, is that because of these kind of people, boys like me look awful and like vile liars. \" i say what i mean \". \" i am different than most other guys \". \" i am not looking for a cheap fuck, i want something concrete .\" \" i love you \". \" what\'s in your heart and what you stand for are what is really getting to me \". All things said by friendly, genuine people. In addition to that, also said by shameless players. I understand the game is so hard to play when people can bend the rules at will, and fool us all. Maybe that\'s why I am not playing anymore. . . No, really. I feign from considering asking girls out, purely on the basis that for one, i have to deprogram everything they have come to beleive about the opposite sex. it\'s getting to be a pain in the ass. that is also assuming they\'re an innocent in the first place. which brings me to my other key goal; to discern their ambitions. it\'s hard when they have amazing eyes. i\'m an eyes guy. it\'s tough.    not too sure where i am going with this. . i guess my point is this. thanks for ruining it for me.  you know who you are , you faking, dishonourable, pointless existences.  thanks a lot ;)   {and by no means whatsoever am i referring to anyone i know in particular}  heh, a little harsh, but is it? i don\'t mean it to be! in fact,  i\'m in a good mood today . -- bold because it probably SEEMS rare, based on how i write, and the subject matter i choose. i assure you all, i am a very happy person :D   thinking about how good it feels to have a smaller role in the lives around me. to be breaking free one day at a time. not to say i\'m never going to partake in party-favours again, no way. i\'m just feeling like... like comfy clothes feel. wee!   i used to surround myself with people who had a story. it was all about the story. i overlooked their lack of morals, for ..the story.  the more out there, intriguing, mysterious, moral or not, i was right there. i also had a thing for the lost sheep of the world, convinced i could save them. assist them in cleaning up. taking their next step. trying to keep my own in queue, and not fall into their hole was somewhat easy at first. for about a day. i am getting sick of all the selfishness in the world. yet at the same time, i am bothered by the fact that friendliness is quite often received as stupidy. a hidden agenda. and a free spout to milk from. these days i am trying to be a little more selective of who i let in, and the rest, i put on a stupid face.   people always so easily underestimate you when you\'re happy all the time.   well. another day over, another one for the books --if you read this far, cool for giving a damn.  michael anton. [==+ roofsessions@hotmail.com +==]  ( urlLink click ) ( urlLink click ) ( urlLink click )  '),(9222,1240274,'25,August,2003','2003-08-25 00:00:00','   urlLink      mmm. revival.  '),(9223,1240274,'24,August,2003','2003-08-24 00:00:00','   Bjork: Army of Me   Stand up You\'ve got to manage I won\'t sympathize Anymore  And if you complain once more You\'ll meet an army of me  You\'re allright There\'s nothing wrong Self - sufficience please! And get to work  And if you complain once more You\'ll meet an army of me  You\'re on your own now We won\'t save you Your rescue - squad Is too exhausted  And if you complain once more You\'ll meet an army of me  '),(9224,1240274,'24,August,2003','2003-08-24 00:00:00','   urlLink AlterNet: Living In a Kleptocrat Nation : \"kleptocrat nation (klep toe krat nay shun), n. 1. a body of people ruled by thieves. 2. a government characterized by the practice of transferring money and power from the many to the few. 3. a ruling class of moneyed elites that usurps liberty, justice, sovereignty, and other democratic rights from the people. 4. the USA in 2003. \"  '),(9225,1240274,'24,August,2003','2003-08-24 00:00:00','   going somewhere new somewhere away from you your vile red skin proving the demon within always bringing me places cutting the keys making the ride so much longer for me not to places, hidden spots that i\'ve ever wanted to be  so tommorrow it\'s over wishing that with all the choices i\'ve made and how much damage i\'ve done how i could have killed myself if i continued on  wishing the same fate would befall you sooner your path would be your suicide your choices, your attrition making you so hidden prolonging the fate you face to punish you for all the souls you\'ve stolen.  so i want  new  a life so far apart from you not referring to the faces in my life you so often embody those close to me that had silently fallen these innocents you chose just a part of your callow plan deep within their minds you froze at times i sadly admit, i even made you my right hand man a part of my outfit  i pray for tommorrow. . . a future so uncertain gripping the world in fear, our own destruction seems so near --all in your palm we gather walking quietly into the fire  it\'s also to be seen that i am not, will not be the only one who turns away from you and runs  my only concern is that  before  my change is complete i have proven myself my trespasses all undone that you still have your way. . .  and take my life that day that simply must not be.  '),(9226,1240274,'05,September,2003','2003-09-05 00:00:00','   urlLink Edinburgh Evening News - Top Stories - Bin Laden family\'s US exit \'approved\' : \"Bin Laden family\'s US exit \'approved\'\"  no comment.  '),(9227,1240274,'03,September,2003','2003-09-03 00:00:00','   3 days until  urlLink www.serotoninambition.cjb.net  loses it\'s web host ( urlLink Maggie  WRAWKS for all she\'s done for me--it\'s been almost the whole year. Thanks hun) and it could be a little bit of doing before I get it back up --and updated to version 3.0. If you take a gander, keep in mind none of my new art/photographs are on it.. A small taste of things to come? Maybe. Not too sure yet ;)  Also, I need some info! --Regarding  Radiohead \'s post-poned Toronto show! Anyone got the scoop? House of Blues doesn\'t have anything on it that I can see.......  Michael Anton.  '),(9228,1240274,'02,September,2003','2003-09-02 00:00:00','   they say that when God closes a door, He opens a window. . .  '),(9229,1240274,'02,September,2003','2003-09-02 00:00:00','   in a neo-punk kinda\' mood. ah yes, the memories associated with blink182...................... good times. work felt so short today. which immediately puts me on a good level ;) today was my bro\'s first day back to school...in 10 years. proud of the man. one of the men behind this man. feels good when i look to him as a great influence again, it\'s been too long..  another groovy beat; favourite game. cardigans. before they turned to crap. woohoo.  aw well. feelin good, and wasting it in front of a  magical box \'o\' radiation  and networked goodness, as great as that sounds, is done for now. CHEERS..!  mike ry-doo-dah .  '),(9230,1240274,'31,August,2003','2003-08-31 00:00:00','   \".. And always, if he had a little money, a man could get drunk. The hard edges gone, and the warmth. Then there was no loneliness, for a man could people his brain with friends, and he could find his enemies and destroy them. Sitting in a ditch, the earth grew soft under him. Failures dulled and the future was no threat. And hunger did not skulk about, but the world was soft and easy, and a man could reach the place he started for. The stars came down wonderfully close and the sky was soft. Death was a friend, and sleep was death\'s brother. The old times came back - a girl with pretty feet, who danced one time at home - a horse - a long time ago. A horse and a saddle. And the leather was carved. When was that? Ought to find a girl to talk to. That\'s nice. Might lay with her, too. But warm here. And the stars down so close, and sadness and pleasure so close together, really the same thing. Like to stay drunk all the time. Who say\'s it\'s bad? Who dares to say it\'s bad? Preachers -but they got their own kinda\' drunkenness. Thin, barren women, but they\'re too miserable to know. Reformers -but they don\'t bite deep enough into living to know. No -the stars are close and dear and I have joined the brotherhood of the worlds. And everything\'s holy ..everything, even me.\"  The Grapes of Wrath; John Steinbeck.  '),(9231,1240274,'09,September,2003','2003-09-09 00:00:00','   Two jumps in a week I bet you think that\'s pretty clever don\'t you boy Flying on your motorcycle Watching all the ground beneath you drop You\'d kill yourself for recognition Kill yourself to never ever stop You broke another mirror You\'re turning into something you are not  Drying up in conversation You will be the one who cannot talk & all your insides fall to pieces You just sit there wishing you could still make love They\'re the ones who\'ll hate you when you think you\'ve got the world all sussed out They\'re the ones who\'ll spit at you You will be the one screaming out  It\'s the best thing that you ever had It\'s the best thing that you ever had The best thing that you\'ve had has gone away  Don\'t leave me high don\'t leave me dry   radiohead: high and dry    ...summer is gone and the party --like the people, the life, and the drugs ..have    all   gone    stale.   moving on past the single moments, the dry eyes, and broken hearts. it\'s autumn now and all i can think of right now is how nice it is when the crashing waves dull the wrenching in my head and heart....... how nice it\'s gonna be to shut up and start over clean.  from this point on, (not sure how long it\'ll last)  i think i am turning back to my tattered notebook for my thoughts, and writings. it\'s been fun, and meeting new people through this web log was a cool experience. hope everyone had as amazing a summer as i, and if we ever shared a talk, and you haven\'t heard from me in awhile, drop me a line..  michael anton, over. and  out . roofsessions@hotmail.com  '),(9232,1240274,'07,September,2003','2003-09-07 00:00:00','   sitting here reading the news, with a few bowls, coffees, and cigarettes. (  grab \"smashing pumpkins - mayonaise\" --acoustic version if you can get it  )..  here be a news medley pour vous;  for the girl;  urlLink Ananova - Lap dancing club offers to pay students\' tuition fees (..in WINDSOR ;)   taxes on the rich? thongs, and t\'s? ..snah?  urlLink The geek who would be governor  i say boycott  urlLink EMI  and buy   urlLink Universal   for a little bit. force some change in the way of how much we pay for music cd\'s... read why. ..and something that\'s just   urlLink nice   :D  '),(9233,1240274,'14,September,2003','2003-09-14 00:00:00','   urlLink Beth Orton:  Bullet  :  \"Something\'s buzzing round your brain and it won\'t let go, Could be saying all you need to know, It could control who you are \'til you become a passenger, When all is said and done, Where credits due the blame was true, It isn\'t all up to you.  There\'s too many people take more than they can use, Still too many people taking more than they can use, They\'d rather drown you in self pity, Than find the will to live, When there\'s no consequence there\'s no easy route, Find your bullet and aim and shoot, When there\'s no compromise no easy route, Find your bullet and aim and shoot, You\'ve got to aim and shoot.  Time is rushing through your veins, And it won\'t let go, Time can tell you all you need to know, It doesn\'t matter who I am if I\'m not part of its greater plan. When all is said and done, Where credits due the blame it isn\'t all down to you.  There\'s too many people take more than they can use, Still too many people taking more than they can use, They\'d rather drown you in self pity, Than find the will to live, When there\'s no compromise there\'s no easy route, Find your bullet and aim and shoot, When there\'s no compromise no easy route, Find your bullet and aim and shoot, You\'ve got to aim and shoot.\"  '),(9234,1240274,'21,September,2003','2003-09-21 00:00:00','   urlLink Slashdot: News for nerds, stuff that matters : \"Your Rights Online: RFID Hell  \"Finally, somebody has stepped up with an article that descibes the potental abuse of RFID. Imagine being flagged for social tendencies. Gattaca is not so far off as we think. it is simply a pass of a wand for your embedded tag rather than a drop of blood.\' \"  '),(9235,1240274,'29,September,2003','2003-09-29 00:00:00','   urlLink All-New Adobe Creative Suite Delivers Platform for The Future of Design and Publishing : \"All-New Adobe Creative Suite Delivers Platform for The Future of Design and Publishing\"   mmmmmmm  oh, and: http://msfreepc.com/  '),(9236,1240274,'17,July,2004','2004-07-17 00:00:00','   So yeah. I\'m now writing on this thing again.  &nbsp;  Seems like a lot has changed since I last started writing in here, and no need to re-cap. It\'ll all work itself out eventually, but what bothers me is that all my decent posts have been removed. Those being the older ones. Maybe I did it with another web log? who knows. So basically all you can see is what happened when i just moved back from calgary just under two years ago I think? it\'s pretty un-interesting to say the least. . .&nbsp; If you\'re new to my posts, no point in reading up on me, or you\'ll&nbsp;discover i am a dry read..  &nbsp;  Regardless,  More big stuff is in queue for me. Stay tuned. I got all this time on lunch break now because I quit smoking. Grooven times.  &nbsp;  \'Till later-- Cheers!  Mike.  '),(9237,1240274,'16,July,2004','2004-07-16 00:00:00','   woa. like. the past year. it sorta\' feels like a dream. where did all my calgary posts go? where am i? wtf? &nbsp; oh yeah.  &nbsp; see you soon.  '),(9238,1240274,'24,July,2004','2004-07-24 00:00:00','   not much going on.   started a text messaging network for employees at my job tonight. i work at a call centre and cell phones are forbidden, but we all use em for text messaging anyways. i figure if we all had a forum to shoot the shit, while we are talking on calls, we could be pretty ..connected. make plans after work, or spaz out and rant. sure would be a great way to plan some sort of movement. maybe a walkout, or at a certain time we all go into \"systems down\". hell yeah. this could be fun.    urlLink http://upoc.com/group.jsp?group=ttodot &nbsp;   anyhoo.  nothing much goin on. just passin the time and getting in a memorable moment whenever i can. been really chillin\' out lately, and worrying about the future less. i\'m not so much concerned with the outcome of things anymore and am just accepting things more often. life is great when you don\'t have to pass every minute of every day with a gut wrenching knot in your chest.   yep.  drop me a  urlLink line  sometime if you\'re reading this and are bored, or got something to say. cheers!'),(9239,1240274,'19,July,2004','2004-07-19 00:00:00','   no, i\'m not moving just yet. maybe i won\'t. but if i do you know i am going to do it proper.  &nbsp;  but why is it every six months i run away? in the past, 6 months time meant the end of a relationship with a girl. or it meant the job i&nbsp; had got stale. or maybe it\'s the amount of time of silence before i pick up an old friend where we left off. it\'s how long i pick anything up, or the amount of time i put something down.  &nbsp;  never let anyone too close or they might figure me out.  &nbsp;  nevertheless,  i am not referring to anything too much in particular. i was initially debating moving in with a buddy to barrie. start a t-shirt company. get a life. keep the girl. move  forward and keep my place in line .  &nbsp;  now as if almost by instinct i am sealing a different fate. i am opening the door wide to air out the fire in the hole. i am selling my computer in the coming weeks. i am turning my back on natalie ( my car ) and freeing myself up  a la  fight club.  &nbsp;  i did it before.  &nbsp;  this time seems like it will have longer lasting effects to linger. when the car is gone this time, it won\'t just be hiding under a sheet in my parents garage. the computer will not be replaced by a laptop. the idea is to free myself up for something,  somewhere &nbsp;better. $500/month for a car (payments + insurance)&nbsp;plus gas and parking tix is a bit much, wouldn\'t you think? --and how many of you can boast selling off everything you own more than once just to know what it feels like to walk down the street with your life on your back? last time i did this it failed because i was not prepared to do it completely naked. i did not trust my guts and i got burned. i kept the car in the garage, and told my folks it wasn\'t forever. the car payments, and the poor living choices out west were my end.  &nbsp;  my girl thinks i am crazy, but realizes the liberation this feeling of wreckless abandon brings. it\'s like a girl cutting her hair off. i\'m not saying i am running off to far away lands. i\'m not saying i am even formulating a plan --more of a  feeling .  &nbsp;  the weightlessness is inspiring.  &nbsp;  &nbsp;    Oh noose tied myself in, tied myself too tightLooking kind of anxious in your cross armed stanceLike a bad tempered prom queen at a homecoming danceAnd I claim I\'m not excited with my life any moreSo I blame this town, this job, these friendsThe truth is it\'s myselfAnd I\'m trying to understand myselfand pinpoint where i amBy the time I get things figured outI\'ve change the whole damn planOh noose tied myself in, tied myself too tightTalking shit about a pretty sunsetBlanketing opinions that i\'ll probably reget soonI\'ve changed my mind so much I cant even trust itMy mind changed me so much I cant even trust myself&nbsp;  : lyrcis by Modest Mouse : \"Talkin\' shit about a pretty sunset.\"       ..btw. if you\'ve read this far maybe you\'d be interested in helping me out with a server to host only my images? i am turning this into a photoblog..'),(9240,1240274,'18,July,2004','2004-07-18 00:00:00','   a couple weeks ago an email stirred me.   &nbsp;'),(9241,1240274,'26,July,2004','2004-07-26 00:00:00','   i am jacks&nbsp;usurped addictive nature .   the fact that i hate the stupidity that follows after i hit a bong. the dirty feeling in my veins when i drink. the crispy burnt out feeling the days following --if i don\'t smoke&nbsp;another sesh.&nbsp;the people i can often associate with upon acquiring my ganja. the money i spend. . . there really is no reason for me to continue with this current  high infatuation .   it\'s going to take some time before i can be clean with myself. free myself up from this. go out the way i came in and never look back. i make everything sound so drama on this thing --i am tired of killing myself from the inside out. i am sick of being another face on the couch that noone remember\'s the next day. . . i am tired of forgetting just about anything.   i hate how all this said; yet&nbsp;the slight smell of it, or session invite will be my undoing. i hate how the addiction trumps all. with any luck i can quit pot like i did cigarettes. with any luck i can free up $2,500 in the coming weeks.  with any luck. . .'),(9242,1240274,'04,August,2004','2004-08-04 00:00:00','   In the few days it\'s been since any real entry, there have been a lot of developments on the home-front.'),(9243,1240274,'02,August,2004','2004-08-02 00:00:00','   Hope dangles on a string Like slow spinning redemption Winding in and winding out The shine of it has caught my eye And roped me in So mesmerizing, so hypnotizing I am captivated {Chorus} I am Vindicated I am selfish I am wrong I am right I swear I\'m right I swear I knew it all along And I am flawed But I am cleaning up so well I am seeing in me now the things you swore you saw yourself So clear Like the diamond in your ring Cut to mirror your intentions Oversized and overwhelmed The shine of which has caught my eye And rendered me So isoloated, so motivated I am certain now that {Chorus} So turn Up the corners of your lips Part them and feel my finger tips Trace the moment, fall forever Defense is paper thin Just one touch and I\'d be in Too deep now to ever swim against the current So let me slip away (3x) So let me slip against the current So let me slip away (4x) {Chorus} Slight hope It dangles on a string Like slow spinning redemption..  vindicated. dashboard confessionals.  '),(9244,1240274,'01,August,2004','2004-08-01 00:00:00','   terror alert for washington set to high today, with a slight chance of showers and a low of 12 degrees.  bush family,  i hope you choke.  '),(9245,3403051,'22,May,2004','2004-05-22 00:00:00','   INTRODUCTION Jika anda menggunakan adobe photoshop mungkin anda akan kesulitan didalam membuat kotak yang setiap ujungnya memiliki sudut yang tumpul. Hmmm... inilah cara gue untuk membuat kotak tersebut.   * Langkah 1.  Buatlah media gambar baru berukuran 200 x 200 pixel dan pilihlah latar belakang transparant dengan memilih Content = Transparant pada saat anda membuat document baru ( File - New )  * Langkah 2.  Buatlah gambar kotak dengan mengunakan rectangle tools  pada media tersebut, kemudian klik menu Select - Save Selection. Berilah nama pada kotak name, sebagai contoh: A ( Garis titik titik tersebut telah menjadi sebuah gambar hitam putih pada tab Channel ). Anda dapat melihat hasil gambar tersebut dengan cara menekan Ctrl - 4 pada Keyboard  * Langkah 3.  Kemudian pilihlah menu Filter - Blur - Gausian Blur, dan berilah nilai radius 6 kemudian klik OK   Langkah 4.  Kemudian pilihlah menu Image - Adjust - Level atau ( Ctrl - L ), dan berilah nilai input level pada masing masing kotak 140 - 1 - 165 kemudian klik OK  * Langkah 5.  Kembalilah pada Media Layer anda yang masih transparant tadi, ( Tab Layer berada disebelah kiri tab Channel ) Kemudian pilihlah menu Select - Load Selection. dan pada windows load selection rubahlah Channel Layer 1 Transparancy dengan Channel yang anda berinama sebelumnya ( A ).  * Langkah 6.  Pilihlah tombol paint bucket  kemudian kilklah didalam garis titik titik. Dan sekarang anda telah memiliki gambar kotak dengan setiap sudutnya tumpul.  SUMMARY Tehnik ini sangat berguna apabila anda mendisain interface ataupun gambarlainnya. Tehnik ini gue gunakan pada salah satu homepage yang telah gue buat untuk TRIJAYA FM JAKARTA                               Hasil :     taken from :  urlLink http://www.master.web.id/i-framemaster.htm  '),(9246,3403051,'19,June,2004','2004-06-19 00:00:00','   Menambahkan password pada File Zip   Oleh: nazar      Lintau.com - WinZip merupakan salah satu program kompresi yang terbaik saat ini. Bagi Anda yang belum familiar, mungkin pertanyaannya adalah mengapa harus memakai program kompresi? Dengan program kompresi maka Anda dapat memperkecil ukuran dari suatu file. Misalnya jika Anda punya file dokumen MS Word yang besarnya , misalnya 300 KB, dengan program kompresi ukuran file tersebut bisa menjadi 50 KB (tidak pasti!/relatif). Besarnya tingkat kompresi sangat tergantung pada isi dan jenis filenya.     Dengan menggunakan program yang dapat didownload di www.winzip.com tersebut Anda dapat menambahkan password pada file yang dikompres (file ZIP). Dengan demikian tingkat keamanan data Anda dapat lebih ditingkatkan. Di bawah ini merupakann langkah-langkah untuk menambahkan password pada file ZIP.  1. Pilih menu File - New Archive. Arahkan ke folder di mana Anda akan menyimpan archive (file ZIP). Setelah Anda mengisi nama file zip (yang akan dibuat), klik tombol OK. Pada contoh ini, nama filenya adalah \"Data Penting-Ku.zip\".    2. Pilih file-file (bisa satu atau beberapa file sekaligus) yang akan dimasukkan pada file zip.    3. Klik pada tombol Password. Pada kotak dialog yang muncul, isi dengan password yang Anda inginkan. Klik OK.  4. Akan muncul kotak dialog konfirmasi password. Masukkan password yang sama dan klik OK.    5. Klik tombol Add. Jika berhasil, maka saat file zip tersebut dibuka maka terdapat tanda \"+\" (plus) di sebelah kanan file yang diproteksi dengan password.    Sumber : Klik-kanan.com  '),(9247,3403051,'19,June,2004','2004-06-19 00:00:00','   Lintau.com Artikel singkat kali ini akan membahas mengenai operasi membuka dan menutup CD-ROM drive. Kita akan menggunakan Windows Multimedia API mciSendString untuk melakukan operasi ini. Sebelum kita masuk ke acara utama silakan simak dulu pesan-pesan berikut.    Tulisan ini bersifat terbuka dan Anda bisa menyebarluaskan sebagian atau seluruh tulisan ini selama Anda menyebutkan sumber aslinya. Anda bisa memberi komentar terhadap tulisan ini melalui sony-ak@sony-ak.com. Penulis tidak bertanggung jawab atas segala sesuatu yang diakibatkan oleh tulisan ini dan semata-mata hanya untuk tujuan pendidikan.  OK, mari kita langsung saja step-by-step demo ini. Buka Visual Basic Anda dan buatlah project baru (Standard EXE) Tambahkan module pada project dan copy paste code di bawah ini:  Option Explicit  Public Declare Function mciSendString Lib \"winmm.dll\" Alias \"mciSendStringA\" (ByVal lpstrCommand As String, ByVal lpstrReturnString As String, ByVal uReturnLength As Long, ByVal hwndCallback As Long) As Long  Public Function OpenCDDoor(ByVal drv As String) As Long   Dim Alias As String   Dim retval As Long   Alias = \"Drive\" & drv   retval = -1   retval = mciSendString(\"open \" & drv & \": type cdaudio alias \" & Alias & \" wait\", vbNullString, 0&, 0&)   retval = mciSendString(\"set \" & Alias & \" door open\", vbNullString, 0&, 0&)   OpenCDDoor = retval End Function  Public Function CloseCDDoor(ByVal drv As String) As Long   Dim Alias As String   Dim retval As Long   Alias = \"Drive\" & drv   retval = -1   retval = mciSendString(\"set \" & Alias & \" door closed\", vbNullString, 0&, 0&)   retval = mciSendString(\"close \" & Alias, vbNullString, 0&, 0&)   CloseCDDoor = retval End Function  Tambahkan dua buah button pada Form1, satu sebagai button untuk membuka CD-ROM drive dan satu lagi untuk menutup CD-ROM drive. Copy paste code di bawah ini pada bagian deklarasi umum di Form1.  Option Explicit  Private Sub Command1_Click()   OpenCDDoor \"E\" End Sub  Private Sub Command2_Click()   CloseCDDoor \"E\" End Sub  Jalankan project di atas dengan menekan F5.  OK, inti dari demo di atas adalah pada method OpenCDDoor dan CloseCDDoor pada module. Kedua method tersebut memiliki parameter CD-ROM drive yang bertipe string. Anda tinggal memasukkan drive CD-ROM ke dalam parameter tersebut dan melakukan operasi buka atau tutup CD-ROM drive.  Sekian dulu tips kali ini dan semoga bermanfaat bagi Anda semua. Jika ada komentar atau kritik bisa dikirim melalui sony-ak@sony-ak.com.  Thanks for all readers around Indonesia   http://lintau.com/modules.php?name=News&file=article&sid=623  '),(9248,3403051,'31,May,2004','2004-05-31 00:00:00','   ..uhhuukk..waa....batuk nih.... emang lagi musimnya sih....aappeezzz.....hikss...hikss... --------------------------------------    Make Mine a Triple: The Feuling W3  http://www.motorcyclecruiser.com/roadtests/w3/   Supersize your V-twin with an extra cylinder. From the October 2000 issue of Motorcycle Cruiser. By Evans Brasfield.  urlLink         Photography by James Brown (email kingosoul@aol.com)    The story is as old as the internal-combustion engine. Some people are never satisfied with their engine\'s power output. What follows is a slippery slope of bigger bores, capacious carburetors, obnoxious exhausts, radical valve timing, and demon tweaks. Even when the engine has obviously reached the ragged edge, a few builders will push ahead, testing the boundaries of physics--and common sense--to produce very powerful, very expensive grenades.  Jim Feuling, owner of Feuling R&D in Ventura, California (www.Feuling.com), realized that, while nothing exceeds like cubic inches, the current crop of big-bore kits for Harley-Davidson\'s V-twins were reaching the upper limits of practical power generation. Feuling decided that, in order to add a new wrinkle to the displacement wars, he wanted to build an engine with 50 percent more displacement than Harley\'s Twin Cam 88 with the 95-cubic-inch factory upgrade kit installed. The result of Feuling\'s idea takes the form of the working prototype shown here. Looking strangely familiar yet completely alien, the Feuling W3 may be the shape of things to come for those who fancy power cruisers.     urlLink        Feuling\'s creation is at its core both incredibly simple and maddeningly complicated. Simply put, why not just graft an additional cylinder onto Harley\'s robust Twin Cam design? The result would be the desired 50 percent increase in displacement while utilizing parts that are widely available from either H-D or the aftermarket. The complicated part would be, well, actually acting on the idea. This stumbling block, however, wasn\'t a problem for Feuling.  Get hard on the gas and the W3 delivers enough power to the drive belt to twist the modified Dyna frame. The Progressive Suspension handles the bump absorption duties well. The Dunlop K591 just doesn\'t offer enough grip to keep the rear tire from spinning up during high-rpm shifts in the lower gears.    Feuling holds more than 100 patents in the automotive industry, with R&D contracts with companies such as Cheverolet, Oldsmobile, Ford, Chrysler, Nissan, John Deere, and Harley-Davidson. He developed four-valve heads for Harley-Davidsons and later sold the rights to Ultra Cycles. He also designed an engine for American Honda\'s high-mileage streamliner that delivered a whopping 500 mpg at 55 mph. He holds several land speed records, including one for a motorcycle streamliner at 333.847 mph set on the Bonneville Salt Flats on October 20th, 1999. Feuling R&D\'s 28,000-square-foot facility is chock full of CAD/CAM/CAE machinery, making the company one of the few private facilities in the world that can design, test, and build entire vehicles (including the engine and drivetrain) under one roof. The W3 engine project initially began in conjunction with Harley-Davidson, but as the design reached the end of the initial stages, Harley opted out. With the withdrawal of Harley-Davidson, Feuling went back to the drawing board. The original design had been structured so that H-D would need to add just six new part numbers to its inventory. When Harley opted out, that concern was scrapped in favor of a design that used only parts available in the aftermarket plus a few application-specific items that could be manufactured by Feuling.  The working prototype W3 engine we rode is, when viewed in broad strokes, essentially the same as the version scheduled to be sold early next year. Between the time that the bike was shot and the time we went to press, the displacement jumped from 142 cubic inches (2327cc) to 150ci (2458cc) as a result of the bore and stroke changing from 4.0 inch (101.6 mm) bore and 3.875 inch (98.4 mm) stroke to a perfectly square 4.0 inches by 4.0 inches. Taking inspiration from radial aircraft engines, a central, master connecting rod with two slave rods, one on each side of the central rod, direct the 4.0-inch slugs through the cylinders. Feuling tells us that the design has already been tested in a 185-cubic-inch (3032cc) form and a 245-cubic-inch (4015cc!) drag racing kit may also be built.     A trio of Dellorto 40mm flat-slide carbs complete with velocity stacks feed the hungry cylinders. However, the W3 doesn\'t just count on three big jugs for its power. Several other Feuling innovations make their way into the mix. Feuling-designed cams operate pushrods for the two-valve patented \"Max Flow\" heads and valves to speed the combustibles into and out of the arena. Staggered triple AR (Anti-Reversionary) exhausts minimize the backwards flowing pulses that disrupt the movement of gasses out of the cylinder. Three Feuling-designed CVX mufflers allow maximum flow through while keeping the noise and backpressure to a minimum. How does a claimed 150 horsepower at 5000 rpm and 195 foot-pounds at 3000 rpm sound to you? The engine is so mildly built that it can even run on 87-octane gas. The aftermarket helps the W3 put the power to the pavement. A Barnett \"Clutchzilla\" transfers the horsepower to a Jim\'s five-speed transmission.     A modified Dyna frame wraps around the big engine. The front of the downtubes were widened, and the fork was raked out to 34 degrees, giving the third cylinder a little breathing room. Famed customizer Jesse James of West Coast Choppers (www.westcoastchoppers.com) created fenders and a gas tank worthy of gracing such a unique machine. The almost invisible ghost-flame paint scheme comes courtesy of Ultra. An inverted Storz Ceriani fork (www.storzperf.com) adds to the aggressive stance of the chassis. Progressive Suspension shocks keep the back end in line. The rolling gear features a 19-inch Performance Machine Trespasser front wheel and a matching 16-inch rear. The dual four-piston PM calipers offer tremendous stopping power, while Dunlop K591 tires provide the adequate stickiness.  Thumbing the starter button brings the beast to life, and the first impression is that of the exhaust note and cadence. You won\'t hear the H-D potato-potato from this bike. Instead, the engine has a distinct Ducatiesque sound (due to the 90-degree separation between the front and rear cylinders, no doubt) with Harley overtones. Open the throttle and the intake honk of the trio of Dellorto carbs can be heard along with the throaty, yet not too loud exhaust. The stout pull on the clutch lever is only surpassed by the strength required to turn the throttle. In an effort to lessen the throttle effort a full-turn throttle was mounted, making it difficult to roll on the throttle past half-way without choking up on it drag racer style.     Perhaps the long throw of the throttle is a good thing since cranking it wide open in second gear caused the Dyna frame that we rode to flex noticeably under the strain of the 150 horsies. According to Feuling, the next generation prototype will mount rigidly to an FXR frame, eliminating this issue. Since the W3 engine runs eerily smooth at cruising speeds, swapping the Dyna\'s vibration damping for solid mounts shouldn\'t present a problem. Unlike many of the ultra-big bore twins we\'ve ridden, the W3 works well in the lower rpm range with none of the surging and drivetrain snatching we\'ve experienced on some aftermarket engines. The engine felt comfortable, even docile at sedate speeds. However, acceleration at any more than half-throttle makes short work of any straight sections of road. Fortunately, the big Performance Machine brakes scrub off speed effectively, giving us an excuse to rap on the throttle once again. W3 owners will, however, want to pop for some stickier rubber in order to harness the bikes acceleration and braking forces; just donÃ‚â€™t expect the tires to last too long.  Whether the W3 engine configuration is the shape of cruisers to come remains to be seen, but we\'re certain more than a few twisted enthusiasts will pony up the $20,000 for Feuling\'s \"Warlock\" engine and frame kit when it\'s available. If you can\'t wait, build your own. Feuling\'s estimates his expenses for developing this first streetable prototype are mere 2.5 million dollars. We certainly wish Feuling the best and hope we\'ll get a shot at a full test of the W3 when it\'s available. A single afternoon in the saddle only whet our appetites.  Editor\'s note: Jim Feuling passed away in December 2002. The company lives on and still lists the W3 among its products.  SPECIFICATIONS  Price: $20,000 approx.  Type: air-cooled, 3 cylinder 45 degree x 45 degree  Valve arrangement: OHV; 1 intake, 1 exhaust per cylinder, pushrods, hydraulic adjusters  Displacement, bore x stroke: 142 cu. in. (2327 cc), 4.0 in. x 3.75 in. (101.6 mm x 98.4 mm)  Carburetion: 3, 40mm Dellorto flat-slide  Transmission: JimÃ‚â€™s 5 speed  Chassis: Modified Harely-Davidson Dyna  Front suspension: Storz Ceriani  Rear suspension: 2, Progressive Suspension dampers, adjustible for preload and rebound damping  Wheels: Performance Machine Trespasser aluminum, 19-inch front, 16-inch rear  Contact: Feuling Motor Company, 2521 Palma Drive Ventura, Ca 93003, (805)650-3406, www.feuling.com  '),(9249,3403051,'28,May,2004','2004-05-28 00:00:00','   HONDA FANTASYX    MALAGUTI CAFE RACER    Renault     urlLink too see much more, click me !  '),(9250,3403051,'26,May,2004','2004-05-26 00:00:00','   PS 250     Honda ngeluarin prototype ini sebagai kendaraan serbaguna disegala medan, hanya sayangnya gw ngga mudeng boso itali di web-nya. Jika anda mau membantu dengan menterjemahkannya silahkan post comment anda. gw bakal berterima kasih sekali.      NB: Surabaya kok banyak banjir ya akhir-akhir ini ???? so cold,so rainy,so wet, ..........  '),(9251,3403051,'26,May,2004','2004-05-26 00:00:00','   Inline-Four vs. V-Twin                 The engine on the left, from an Aprilia Mille, shows how the weight is spread on a V-Twin: compact laterally, but spread out longitudinally. The Suzuki GSX-R1000 motor on the right, however, shows the exact opposite: wide laterally, but compact longitudinally.    There are lots of engine options for motorcyclists: Singles, Twins (in parallel, horizontally opposed and Vee configurations), Triples, inline-Fours and V-Fours, horizontally opposed flat-Sixes, rotaries and even jet turbines! Yet with all these options, the most popular engine types for motorcycles have been, and continue to be, V-Twins and inline-Fours. As the pinnacle of motorcycle technology has arguably swayed in favor of the 1000cc Superbike, let\'s look at these two engine configurations up close and do a little compare-and-contrast examinations.   V-Twin sportbikes, from the high-dollar Ducati 999S to the low-buck Suzuki SV650S, have been making a strong resurgence in recent times. Excellent midrange performance courtesy of relatively high torque values means easy access to engine power in the hands of the average rider.   Side profile of a typical dohc, inline-Four engine, in this case, the Suzuki GSX-R1000.     But what makes this so? It\'s not simply displacement, because in this comparison, we\'re only looking at one-liter engines. Instead, the reason V-Twins have relatively more torque than an inline-Four has nothing to do with what you\'d expect, such as the engine\'s basic layout.   As CW Technical Editor Kevin Cameron explains, \"For the most part, this is a function of valve area. The temptation in doing any sports or racing engine is to put into the head the largest valves that will fit. When this is done with a four-cylinder engine (250cc per cylinder in a one-liter Four), the result is more valve area per displacement than with a Twin of the same size (500cc per cylinder in a one-liter V-Twin). The result of this tactic is power concentrated at the top of the rev scale for the four-cylinder, and power concentrated in the middle for the Twin.\"   Why is that? It all has to do with intake-charge velocity. When the intake valves open, the downward movement of the piston creates a vacuum (unless you\'re running boost, but that\'s a story for another day) in the cylinder, drawing a fresh air/fuel mixture into the combustion chamber. Internal-combustion engines are nothing more than glorified air pumps; so because our one-liter inline-Four and V-Twin engines have the same displacement, they therefore flow a similar amount of air per crank revolution. And that means the volume of the incoming charge will be similar in either engine type.  Because the charge volume is determined by displacement, that volume of air and fuel has to squeeze through the comparatively small carb throats/fuel-injection throttle bodies, intake runners, ports and valves of either configuration. Unlike rush-hour traffic, our orderly fuel and air molecules do this by traveling at high velocity. And because velocity increases as the cross-sectional area of a given passage decreases for a given volume, the cylinder heads of an inline-Four engine will have a naturally higher intake velocity than those of a V-Twin.   To take advantage of this fact, the engine designers strive to increase the intake velocity of the V-Twins to help them make more peak horsepower, while they try to decrease the intake velocity of inline-Fours so those engines can make more peak torque. What they end up with are two different engine platforms with very similar intake velocities.   \"Every competent engine designer attempts to use the highest intake velocity he can get away with,\" says Cameron. \"In earlier times, sharp port angles and sudden changes of section interfered with flow, so the engineers came to believe that very large ports were necessary. But as better shapes came into being, they found that smaller ports could be made to flow just as much--if not more--air as before. This has been the legacy of the recent \'flowbench\' period of port development--that small, well-shaped ports can flow plenty of air and improve torque by raising the average velocity in the port.\"  We know, however, that Twins and Fours are not equal, and the reason is that there are still limitations that having more open valve area cannot solve. First, as intake velocity is increased, you begin to have a problem with the intake charge reaching supersonic speeds. A shock wave builds inside the intake tract and air begins to \"back up\" behind the wave. This severely impedes intake velocity and imposes the limit at which an engine can inhale. Additionally, an inline-Four has twice as many power pulses in any given time period than a V-Twin. What\'s more, Twins have more frictional losses in the valve train, thanks to needing two sets of slightly larger cams, two cam chains, stiffer springs pushing on larger, heavier valves moving through more lift and, of course, all the bearings necessary to support all the extra hardware.  To counteract the fact that a V-Twin only produces half as many power pulses per crank revolution, you could simply make the engine rev higher. But even if you were able to ignore the frictional losses (which increase by the square in relation to cam rotation speed--and remember, you have four cams pushing on the aforementioned bigger valves and hardware), you\'d run into an even bigger issue, in fact the biggest issue: piston speed.   V-Twins require more valve gear.Up there, you can see the dual timing-belt arrangement of a Ducati 999.    \"What really determines how high an engine can rev is its peak piston acceleration, reached at TDC on every revolution. This is typically something like 7000 gs right now. Above that, there are problems with piston and ring reliability. Peak piston acceleration is directly proportional to stroke length, and to the square of rpm. In Superbike racing\'s previous 1000cc Twin/750cc four-cylinder formula, this difference worked in favor of the Twins, even though the formula had been carefully set up so that the difference in displacement would be offset by the smaller engine\'s shorter stroke and ability to rev higher. But in fact, the Fours could not rev as high as that formula assumed, because they reached that maximum piston acceleration sooner than thought.\"   From an engineering standpoint, torque alone doesn\'t make an engine configuration viable. That\'s why packaging has played such a strong role in the success of the V-Twin. Not only are V-Twins narrower, but contrary to popular belief, they offer greater flexibility in fore-aft weight distribution due to their broader polar moment. Inline-Fours, on the other hand, are wider than V-Twins, obviously, but longitudinally shorter. In a front-to-rear plane, at least, that shortness is helpful in achieving the \"mass centralization\" goals that most sportbike manufacturers currently ascribe to. But the width of inline-Fours means they are generally harder to flick side-to-side than their V-Twin counterparts.   But wait! There\'s more. By utilizing a shorter stroke, with smaller valves in smaller bores, Fours can rev much higher than Twins can. And because a smaller bore size means a smaller combustion chamber, the mixture is likely to burn more completely because the flame front has a shorter distance to travel; and that, in turn, allows the use of higher compression ratios. All of these factors equate to an engine that produces more peak horsepower than a Twin. No matter what sort of engine format you like, more power pulses per revolution, with more revolutions available over a given time period, equal more power.   Then there\'s the subjective feeling that V-Twins provide. Producing one large power cycle per crank revolution, and staging those pulses at staggered intervals, not only yields a character that can be felt and heard, it punches all the right feel-good buttons.   Does that mean inline-Fours are better? Absolutely not. As Cameron puts it, \"In simple terms, the four-cylinder should slaughter the Twin in terms of power, but when can you use it? For a few seconds at the end of the straights? Meanwhile, the strong midrange of the Twin can allow its rider to get a better jump off most turns, which gives an advantage most of the way down the next straight.\"   What all this means is that, unless you\'re involved in a serious racing program, it doesn\'t really matter what motor layout powers your bike. Whether it\'s a torquey V-Twin or a high-revving inline-Four, the bottom line is that a rider\'s emotional preference is, as is often the case, the only real measuring stick that matters.   -- Calvin Kim   urlLink http://cycleworld.com/index.cfm?siteaction=OnlineUpdate&id=46  '),(9252,3403051,'25,May,2004','2004-05-25 00:00:00','   This trike is \'bajaj\' in Indonesia is transportation use since 70-80\' era( import from India ). But I found a many variation of use this vehicle. Who knows ??? this vehicle has a much function like this :      550ML model of lambro    Ambulance Function   Passenger Model of India    Modern model (Vikram India\'s)   Beginning model   Truck Fire   450cc Model    urlLink See more by click this !!  '),(9253,3403051,'24,May,2004','2004-05-24 00:00:00','   Kawak ZZR Smoothness ====================  Artikel ini gw ambil dari  urlLink http://www.gizmo.com.au/public/News/news.asp?articleid=2251  Dengan terjemahan bebas sbb :  Motor ini mempunyai kelebihan pada semua fungsi yg dapat \'diatur\'.Motor ini dapat menyesuaikan diri pada kecepatan rendah maupun tinggi. Contoh kecil :fairing dan screen berubah menurut posisi low/high/full sport maupun menurut kecepatan saat berlari.Pada sektor mesin ada teknologi \'cone-shaped\',\'rim-mounted disk brakes\' yg memaksimumkan pendinginan. juga\'centrally-mounted fule tank\' sehingga tanki bensin-nya plus mesin yg besar bisa memperpendek gaya garvitasi lebih dekat ke tanah.Design muffler pun menjadi satu dengan design bodi.        Hanya sayangnya detil mesin tidak di ungkapkan.  '),(9254,3403051,'24,May,2004','2004-05-24 00:00:00','   Wahai diriku,what i do now ??? Just kill the night ??? ooppss...i don\'t know. but ideas come to me....  how \'if\' i put two joint carburator to my KZ200 ???? in standart condition he can 150km/h with slippy on \'kopling\' .....   http://www.motorplus-online.com/Articles/Printing/V/273/Pictures/8754/8754-W500.jpg Keterangan : Karburator yg atu di fungsiin lewat handle/tuas lain sebagai turbo dengan salah satu supyer dimatiin/ di buntu.Atau mo milih yg ini nih !!   and here some pict about pretty \"mad\"MAX with site : http://www.zvmax.com/vmax.htm & http://www.vmaxoutlaw.com/main.html !!!  '),(9255,3403051,'23,May,2004','2004-05-23 00:00:00','   Di Minggu pagi ini, uhf....banyak bgt nih gawe, untung aja gw cuman maen ke kantor....hehhe...legaa dehh... Something interest today\'s :        A most attractive and possibly perfect compromise between motorcycle and automobile was recently announced.  Check it out for yourself at:  urlLink www.carver.nl    Here is some picture :  '),(9256,3403051,'22,May,2004','2004-05-22 00:00:00','   Kereatip juga nih org, gw kopi ya artikelnya   plus gambarnya juga nih !!!!   Bubut Kem Pakai Mesin Jahit     ============================   Bubut benjolan kem kerap bermasalah. â€œDikikir manual pasti enggak rata dan benjol-benjol tidak sempurna,â€ tegas Chandra alias Chiank dari Bandung Motor, di Jl. Ahmad Yani, No. 292, Garut, Jawa barat. Ya, brother satu ini mikir nggak habis-habis cara ngebubut kem secara praktis dan benar.  Lama riset, Chiank dapat ide. â€œKumaha kalau minta bantuan dinamo mesin jahit?â€ katanya merenung. Eâ€™eh, dia ngaku dapat ide saat di WC, wah... ngapaian Koh?  Yuk, langsung lihat konstruksi bikinannya. Awalnya membuat tatakan alias jig untuk dudukan noken as. Tentu, mesti statis dan lurus.  Beres begitu, ia nempatin dinamo kekuatan 110 volt sebagi sumber tenaga pemutar gerinda mini di mesin jahit. â€œDi toko, dinamo begini paling seharga Rp 25 ribu,â€ kata Chiank. Mirip cara kerja para penjahit, putaran gerinda itu  dikendalikan sang tuner lewat kaki.  â€œTinggal gas... zzt..zzzt... gerinda muter, menekan noken as di jig yang ikutan muter seiring tekanan gerinda,â€ ulasnya sambil praktik. Makin halus gerinda berputar, makin presisi pula hasilnya. Begitu juga sebaliknya.  Untuk menekankan gerinda pada pantat noken as, Chiank menghubungkan dinamo dengan alat ukur microton yang bisa diatur maju-mundur. Microton alat pengukur yang menempel di mesin bubut. Pengukurannya diputar tangan, sehingga posisi benda yang dibubut maju-mundur mendekat, atau menjauh dari pisau bubut.  â€œAkurasi microton cukup laik. Perhitungannya, 10  putaran gerinda menggerus pantat kem 0,5 mm. Berarti perlu 20 putaran untuk menebas 1 mm pantat kem. Jika nggak ada microton, bisa memakai mikrometer dengan cara yang sama,â€ ulasnya.  Menurut Chiank, alat temuannya lebih sip ketimbang gerinda manual. Dibangunnya jig yang ajeg menjamin dudukan noken as tetap stabil. Saat gerinda menekan dan berputar konstan, kubah noken as akan terbentuk sendiri. â€œSaat berputar, berapa mm pantat kegerus bisa diketahui dari jumlah putaran.â€  Supaya makin yakin, kem yang sudah dibubut diperhalus  sama mata bor korek, lantas diukur ulang dengan sigmat.   Reporter : Isfandiary Mahbub D.   Fotografer : M. David Srihanoko  '),(9257,3403051,'21,May,2004','2004-05-21 00:00:00','   Buaahhh......pagi yg hot di surabaya !!! I don\'t know why, last night is so rainy ... but this morning... too clearly......right ??? ------------------------------------------------------------------ Oh,iya. saat gw buka buku2 catetan yg lama alias lawas.... i found this...suatu web ttg banyak model mesin yg pernah ada dan unjuk kerjanya. Dari yg biasa, sampai yg paling aneh. Just click at  urlLink http://www.keveney.com .  contoh kecil di mesin 4 tak ciptaanya mas Otto nih ---->  '),(9258,3403051,'21,May,2004','2004-05-21 00:00:00','   Saat kukunjungi site http://www.neobike.net, ada satu judul yg sangat interest, yaitu sepeda motor dengan kongiurasdi mesin V8, gila kali ya ???? seberapa hebat sih ??? Situs lengkapnya sih ada di  urlLink http://www.richmondlabs.com/Automotive/RichmondV8/V8Registry.html  Ulasan : sebuah situs yg berisi banyak kumpulan pembuat spd motor V8 dengan detail dan asal mesin-nya   Just look ...... U gonna be full impression...ini nih contonya handmade tapi fabrique looking .... factory model....    Ini nih yg nggilani, but its so creative ....     Ini bukan mesin V8 sih, berapa ya ??? susah juga ngitungnya !!  '),(9259,3403051,'21,May,2004','2004-05-21 00:00:00','   Hi....  I Think this gonnabe nice ....... i\'m creating a much motorcycle design ..... i\'m collecting motorcycle information..... and so much gone , time by time...... and this blog would be my solution........... see me ...........  '),(9260,3403051,'20,June,2004','2004-06-20 00:00:00','   Mengenal dan merawat mesin turbo   (05/09/2002) - Turbocharger menjadi alternatif terbaik untuk meningkatkan daya kuda yang bisa dihasilkan mesin, tanpa harus menambah bobot mobil. Selain itu, ukurannya yang kompak dan proses pemasangannya yang sederhana, membuat turbo juga populer di aftermarket.      Perangkat ini banyak digunakan mesin diesel maupun mobil performance. Di Indonesia, banyak model yang menawarkan perangkat ini, seperti Isuzu Panther, Ford Ranger, Mitsubishi L200 Strada dan Kia Carnival.   Prinsip kerja turbo, mengkompresi udara ke mesin untuk meningkatkan jumlah molekul oksigen yang masuk ke silinder. Tingginya molekul oksigen yang masuk mendorong tambahan pasokan BBM. Dengan demikian, lebih banyak BBM yang dibakar, hingga daya yang diproduksi meningkat.    Tekanan udara yang dikompresi bisa meningkat hingga 8 psi (pounds per square inch) dibandingkan tekanan normal. Bila tekanan normal di permukaan laut sebesar 14.7 psi, maka udara yang dikompresi mempunyai tekanan hingga 50% lebih tinggi. Namun tidak berarti power yang dihasilkan meningkat 50%. Karena ada sebagian daya yang hilang/inefisiensi. Peningkatan daya optimal turbo bisa 30 â€“ 40 persen lebih banyak.   Untuk melakukan kompresi, turbo memanfaatkan aliran gas buang dari mesin untuk memutar turbin, yang meneruskan putaran ke kompresor udara. Turbin ini bisa berputar hingga 150,000 putaran tiap menit (rpm) atau 30 kali putaran mesin mobil pada umumnya. Temperatur perangkat ini juga bisa melesat naik, ketika bersentuhan dengan gas buang. Dengan kondisi kerja seperti itu, turbo membutuhkan material berkualitas tinggi dengan pengerjaan super presisi.   Perangkat turbo dipasang pada exhaust manifold, sedangkan kompresor udara diletakkan diantara air filter dan intake manifold. Udara yang dikompresi, suhunya naik dan ketika suhu naik, udara akan memuai lagi. Akibatnya, meskipun tekanan udara yang masuk ruang bakar tinggi, tapi jumlah molekul udara yang dibutuhkan untuk pembakaran menjadi berkurang. Oleh karena itu, maka ditambahkan perangkat intercooler yang berfungsi menurunkan suhu udara kompresi.   Di sisi lain, penggunaan turbo juga menimbulkan kerugian pada mesin. Pemasangan turbin membuat aliran gas buang menjadi tidak lancar. Mesin juga harus mengeluarkan tenaga ekstra untuk melawan tekanan balik dari saluran gas buang.   Selain itu gejala knocking/nglitik juga sering ditemui. Ini disebabkan karena udara kompresi yang bersuhu tinggi ketika masuk ke ruang bakar yang bertekanan tinggi, bisa memicu pembakaran sebelum busi memercikkan api. Oleh karena itu, mobil dengan perangkat turbo seringkali membutuhkan BBM dengan oktan tinggi, guna menghindari gejala knocking. Kini mesin-mesin modern yang dilengkapi turbo, sudah dilengkapi semacam adjuster yang bisa menyesuaikan kompresi udara secara presisi sesuai kebutuhan mesin.   Problem lain yang sering ditemui mobil dengan perangkat turbo adalah turbo lag. Kondisi ini terjadi karena turbo tidak bisa seketika menghadirkan tambahan daya saat gas ditekan (turbo baru bekerja pada putaran tertentu). Baru beberapa detik kemudian tambahan daya bekerja, ditandai dengan melonjaknya mobil ke depan.   Cara untuk meminimalkan efek ini adalah memangkas bobot komponen yang berputar. Ini membuat turbin dan kompresor lebih mudah berakselerasi untuk melakukan kompresi. Cara lainnya, dengan menggunakan material baru seperti ceramic turbine blades. Material baru ini lebih ringan dari baja, hingga lebih mudah berputar Efek ini nyaris tidak terasa pada mesin dengan teknologi turbo modern.   Kebanyakan turbocharger memiliki wastegate, semacam katup pengaman yang memungkinkan gas buang menerobos keluar tanpa melewati turbin. Katup ini bekerja berdasarkan sensor tekanan. Bila tekanan udara terlalu tinggi, berarti turbin berputar terlalu cepat, maka exhaust gas dibuang lewat wastegate, hingga rotasi turbin melambat.   Karena turbo bekerja pada kondisi temperatur, kecepatan dan tekanan tinggi, maka peforma optimum bisa didapat jika alat ini dioperasikan dan dirawat dengan benar. Kerusakan yang sering terjadi biasanya akibat buruknya lubrikasi, atau masuknya partikel abrasif pada oli. Sebab lain adalah lolosnya partikel berukuran besar pada aliran udara yang tersedot masuk. Juga benda-benda yang tersembur keluar dari exhaust, seperti kerak karbon, serpihan komponen mesin, dll berperan menimbulkan kerusakan.   Agar turbo bekerja sempurna, maka;     Turbo harus di service sesuai rentang waktu yang direkomendasikan.   Gunakan selalu oli yang direkomendasi produsen mobil   Pilih bengkel yang benar-benar ahli dalam perawatan turbo   Periksa setiap kebocoran oli, suara-suara â€˜anehâ€™ dan getaran yang tidak wajar.   Power kurang, suara keras, asap biru atau hitam, kemungkinan mengindikasikan masalah pada mesin, bukan turbo   Panaskan mesin beberapa saat, tunggu temperatur oli mesin mencapai suhu kerja optimal sebelum menggenjot pedal gas dalam-dalam untuk mengaktifkan turbo. Jangan memainkan pedal gas, karena kemungkinan lubrikan komponen turbo belum sempurna. Sebaliknya, biarkan mesin idle beberapa saat sebelum mesin dimatikan. Bila mesin dimatikan seketika, maka pasokan oli mesin ke turbo otomatis terhenti, sementara turbo masih berputar dengan kecepatan tinggi. Ini bisa menciderai bearing. Pada mesin-mesin dengan teknologi turbo terbaru, â€˜ritualâ€™ seperti itu tidak perlu lagi.   -Â®-  '),(9261,3403051,'11,June,2004','2004-06-11 00:00:00','   Beberapa saat lalu,kenapa sih blogger ngga bisa sign in ???  kali aja network gw yg kacaow ya....btw......... gw dapet gambar good buat Honda 90s buat tampilan racing.Aselinya sih Honda cub 50cc Diambil dari web vintage racingnya http://www.eurospares.com/graphics/cubprix.jpg     urlLink    urlLink          Gw masih ngga tahu detilnya , but U kira2 sendiri deh ya........ improve..improve....improve.......  '),(9262,3403051,'09,July,2004','2004-07-09 00:00:00','   Hmm.kenapa eh kenapa....extrem modif scooter di Indonesia ini masih langka ya......gw punya gambar yg moga-moga aja bisa menggugah niat usil saudara2 kita yg seneng ma modif......   urlLink    urlLink  '),(9263,4151959,'15,August,2004','2004-08-15 00:00:00','   Hey guys!! As some of you might of known, i had a soccer tournament all weekend! Its fun though because its 3v3 soccer and not the big field soccer! Ok soo im sure u dont care very much but we came in first... which was exciting! There are five people on my team: Michelle Scandora, Taylor Neiling, Alli Bellmer, Jaime Orewiler and I. Our team is called DA DAWGS!!!! Yay! YaY!!!! lol...ok sooooo still nothing tooo exciting has happened that will keep u interested but if u will please keep telling people who have computers to go to my blog~lol       P.S. dont forget to tagboard :)     var chatterbox = \'no\';      if (!chatterbox  chatterbox != \'ok\') { document.write(\'This  urlLink flooble chatterbox  is\'); document.write(\' temporarily unavailable. It will be back up shortly.\'); }   '),(9264,4151959,'13,August,2004','2004-08-13 00:00:00','   Hey i know that everyone should be getting their schedules sooner or later so please check mine out and then comment to me if we have any classes together!!!! hopefully we do! :)    var chatterbox = \'no\';   1. world Civ- Corntd   2. geometry- Samide   3. English- Piro   4. Study Hall- Parkash   5. Spanish- Bradley   6. Spanish- Bradley   7. Physchem- Fledderman     8. Gym- Kroger      if (!chatterbox  chatterbox != \'ok\') { document.write(\'This  urlLink flooble chatterbox  is\'); document.write(\' temporarily unavailable. It will be back up shortly.\'); }   '),(9265,4151959,'11,August,2004','2004-08-11 00:00:00','   Hey! anyone whos bored needs to play this exhilerating game called golf solitaire (and it is on yahoo games) it is awesome~!!!!!!!! but yea yesturday i got my haircut (just a lil shorter and more layers) then I went shopping with my mom and then i had soccer practice! So it was a pretty good day! but today i really need to do stuff with some friends because i havent seen some of them for like 6 days!!!! and Michele Redfield I havent seen her for like 14 days!!!!!! Soo everyone call me today on this friday!!!!! and we can plan the best day of our lives! hahahahah hehehehe and if u dont like me then its okay just spend the rest of ur day playing the coolest game ever!!!! GOLF SOLITAIRE~    urlLink  http://games.yahoo.com/games/downloads/sg.html       if (!chatterbox  chatterbox != \'ok\') { document.write(\'This  urlLink flooble chatterbox  is\'); document.write(\' temporarily unavailable. It will be back up shortly.\'); }   '),(9266,4151959,'11,August,2004','2004-08-11 00:00:00','   Hey sorry that i havent written in forever....but i was at this away basketball camp (5 days overnight) so i havent been able to write! The camp was pretty fun! But it was long and im glad to be back so keep writeing messages on my chatterboard thing or whatever!!! and if anything exciting or interesting happened while i was gone please inform me!!!! and if u could keep giving out this blog address because i dont think like anyone knows about it even though its not that good~haha   '),(9267,4151959,'06,August,2004','2004-08-06 00:00:00','   Hey today has been very interesting!!! well i woke up and went to the driving range with my father and then went ova to Nats house with Michelle!!!! and we def had some adventures!!!!!!! like riding our pimp ass tractor!! haha and then on the roof and in the attic!!! and of course our anual tea party! and then i just hung out with Kels and Michelle the rest of the day and now we are gonna have a Tent sleepova with Michelle Alex and Kels....gonna be great fun!!! But then i have Doug Bruno the rest of the weekend so im not gonna write much till then...lol even though no one is looking at it! so if for some reason u are reading it then please tell others! and ill try and write some other interesting things as i come along em\'!!!! lyl   '),(9268,3722833,'28,July,2004','2004-07-28 00:00:00','   Woah!   Well, I\'m back from a long absence.&nbsp; I have been doing alot of things the past week or so.&nbsp; One of those things was working on my new Website!   It isn\'t much yet, but it will be soon!&nbsp; It\'s called  Levile\'s Lair .   You should visit it and check out some of the links I put up.&nbsp; Anyway, I\'m gonna go now.&nbsp;  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Fare-Thee-Well   '),(9269,3722833,'22,July,2004','2004-07-22 00:00:00','   This is my first Web Log Post.&nbsp; I guess I should introduce myself.&nbsp; You can all refer to me as Levile.&nbsp; I\'m fifteen years old.&nbsp; The rest of my information is in my Profile.&nbsp;     &nbsp;  Anyway, today I did a few things.&nbsp; For one I joined the Cowboy Bebop forums:  urlLink http://bebopforums.proboards19.com &nbsp; Also, I downloaded a new MP3.&nbsp; The song is titled \"Green Bird.\"&nbsp; It\'s from Ballad of the Fallen Angels (Cowboy Bebop Session Five.)&nbsp; The song is hauntingly beautiful.&nbsp; A young lady by the name of Gabriella Robin sings the lyrics in a language that I\'m not sure of.&nbsp; However, even though I can\'t understand the words, the song is still great.&nbsp; It makes you remember things from your past.&nbsp; Old friends, places, events, etc.&nbsp;     &nbsp;  I should tell you all that I\'ve been depressed for quite some time.&nbsp; Three years ago, I moved to Upland, Indiana and met a kid named Skai Haley.&nbsp; Me and him became very good friends.&nbsp; He was my only friend at the time.&nbsp; My family has always been moving around the state of Indiana.&nbsp; Despite this fact, I kept in touch with Skai.&nbsp; But now, I\'ve moved too far away to contact him.&nbsp; I haven\'t a clue about his e-mail, and I forgot his address and phone number.&nbsp; The county I\'m living in doesn\'t have the right phone book, so I can\'t look him up.&nbsp; I haven\'t seen him for two years.&nbsp;     &nbsp;  Well, now that I\'ve told all of you that, I guess I should stop whining and wrap up this post.&nbsp; However, to whomever reads this, if you know Skai Haley, tell him his fellow Otaku says hi.&nbsp; He\'ll remember me, I know it.&nbsp;     &nbsp;  Bye for now.&nbsp; I\'ll be sure to keep posting, though.   &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;    &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Fare-Thee-Well   '),(9270,3561749,'30,June,2004','2004-06-30 00:00:00','    okay, so i thought it didn\'t publish, but nooooooooo it did. it just didn\'t show when it said \"view blog (in new window)\" i hate this man. remember when everyone was friends...YA THAT\'S RIGHT WE NEVER WERE...YOU WICKED SERVANT. umm okay, evil moment over. don\'t be scerred.   '),(9271,3561749,'14,June,2004','2004-06-14 00:00:00','    okay so i had like the greatest post ever written then i try to close a popup and it closes all the windows!! bitch!!!   '),(9272,3561749,'14,June,2004','2004-06-14 00:00:00','    that was a long entry, figured i\'d split it up.      2. RJ left for nationals yesterday. so we drive out there and ben\'s already checked in and stuff. he packed all his stuff for five days including uniforms and stuff into one carry on duffle that still has room!! he is sooo not wearing underwear!! but like ya so the airport has alien messages or maps on the floor. i made cards for all the guys and ben liked his. i kinda left em in the car after dad dropped us off at the gate and had to walk to the car and get em. the airport has stupid style sense. rj\'s stupid and didn\'t know where i got my pictures when like half of them were from his moblog. even ben knew they were. btw, he had these uber nifty champion pants that were like my shorts, but pants. (shut up dru, i know i\'m captain obvious!!!!) they paged travis on the intercom hehehehe me and dad saw him and his mom when we got starbucks. while we were gone billy and theresa showed up. it turns out this is only theresa\'s second time up in a plane!! wow. i would die if i never rode a plane until my son\'s nationals. guess i won\'t die though cuz my first time in a plane was when i was like three! i would probably be too freaked out to go i waited that long. but ya she liked billy\'s card and said she wanted to put it in a scrapbook. isn\'t she just the sweetest? ya and billy has to read that elie wiesel book too. so ya, we all chitchatted. then they went through security and all that good stuff. mum still had rj\'s magazine so she gave it to billy. theresa had to take off her shoes and all that good stuff. thank god i didn\'t go, cuz it would take a while to take off my ballerina shoes. oh ya, i wore my black and white hibiscus dress with my turquoise roxy tank and my pink roxy tee over it and my ballerina shoes. read tuesdays with morrie. sad book, good book. pissed me off :)drove home. nothing happened.     3. i babysat at the volunteer appreciation thing. sssoooo fun, kinda, well not really. ya so that\'s a no. one girl georgia is so gonna be a klepto when she grows up. she kept stealing everyone\'s food and sippy cups and knocked over a lot of juice. there were two adorable girls though. one was maggie who looked like sydney (her and her brother patrick matched with the whole red, white, and blue thing). and stacy who i kept calling grace but tracy and libby told me it was stacy, but i swear when her dad picked her up it sounded like tracy. then there was another kid named brendan who had an accident of the poopy sort and was alone in the bathroom and got it all over his legs then threw the diaper down and tried to wipe his butt off on the walls. ya , meagan (adultish chick who knows a buncha sta teachers including flip who had her baby) must have had fun with that cuz the girl with the very low cut dress and big eyes couldn\'t open the door to the other bathroom where she was alone and started crying. i know her pain. i remember how i couldn\'t get out of my room once. it was an epic journey filled with anger, frustration, and tears tears tears. Devin we\'re makin a movie!!!!  me tired and hungry. will leave now. byeizzle.   '),(9273,3561749,'14,June,2004','2004-06-14 00:00:00','    ya so here\'s my attempt to make an entry as rocktastic as the one i had. tear tear tear.       1. rj had his st. louis tourney. me and mum went up on saturday. ya so we thought dad was gonna pick us up 2ish. but no he rolls up into the garage sale at 11 while i\'m volunteering and mum is still shopping around. went home and did a pretty damn good job at packing in five minutes (forgot swimsuit). boring ride. tried to stay awake. couldn\'t. saw pretty cars. okay so we get there and our hotel is like a stone\'s throw away from the magic house and we didn\'t go!!! but they were done for the day and we left for dave and buster\'s maybe 20 minutes after we got in. i got bar burgers again and was the first to leave the table so i could get my game on. and let me tell you i rock at the one where the light goes around and around and you trap it!! i spent like 80% of my card on it. i played a buncha basketball too! speaking of basketball (forgot this in other entry) i saw mrs. williams!! it turned out sha and trish were there too with their aau b-ball team!! i said hi to them and billy comes up and kept saying they went to his school. and i was like \"duh? i\'m not overly deaf\" (on the inside of course) ya and rj is really good at the light up games too, he won on his first try and i won the titanic game and he won the one with the glowing ball. geoff and robin just shot at stuff the entire time. figures. then we went back to the hotel. billy rode with us and we stopped at wallie world for cards and chips so the guys could play poker. went back to the hotel, they played then swam. i listened to music and doodled and read the notebook (ryan gosling rocks. i wish i was an amazon in the 90s).      So the next day (sunday) they played at Forrest park. That gym you have to walk across. lights are much better, not yellow. First game was pretty easy, but then they had to play ST. Louis CYC with bleachy (blech) good server guy and kerr smith guy (with really bad highlights) and lost. there wasn\'t any hottie blonde army gel guy :(  but OMfreakinG!!!! Their setter!!! [fans self]. He was ssssssssssoooooooo scrumdillyumptious. he has the most [pause for breath] abfab shaggy rooney worthy brown hair. the gorgieoussity of it all amazed me. he\'s like 5\'5\" ish which is great by me (shut up Kevin, we\'re not all vertically blessed like you and at least i never worry when it\'s baby eatin\' tine). and he kept on lifting his shirt only to reveal abs you could see through his navy under armour (very reminiscient of grey shirt green shorts lefty omni short tucker guy)!! and he has uber gorgeous hands and always dumped it right where no one is (and i mean just a tight backset when you think he\'s gonna set something up!!). and he is one of the few who look hot with braces. and he took off his jersey and the under armour after they lost the chamionship match to mo thunder. omg, i sound like lauren, \"i want to have his baby\":) oh and mo thunder what the hell?! i always knew the tall black guy was good. but since when is ugly blonde teddy bear good? but he\'s so damn cocky about it. even more so than collin, but then again he can actually play volleyball so ya. and there was this team from texas that rj\'s roomate played for and they were so fruity!! i mean for serious, they were flamers and one of them told dad rj was cute. hehehe!! and they had this little dinosaur named steve. kinda scary they probably rock the ritualistic animal sacrifices. jklol4l. ya so that was pretty much it. omg he was sssssssssssooooooo incredibly hot. i wish i lived in st. louis. dammit why is it this hottie major pops up in their last year!!!! i don\'t get to enjoy him anymore unless rj plays against him in college and that\'s a fat chance.     the ride home was pretty uneventful. we went to ryan\'s, of course. but we did see this green truck that had one of those little monkeys with a diaper!!   '),(9274,3561749,'14,June,2004','2004-06-14 00:00:00','    it didn\'t publish the other half of the long one that i split it up, but my smart thing for the day was copying it before i published. take that comp demons!!!  that was a long entry, figured i\'d split it up.      2. RJ left for nationals yesterday. so we drive out there and ben\'s already checked in and stuff. he packed all his stuff for five days including uniforms and stuff into one carry on duffle that still has room!! he is sooo not wearing underwear!! but like ya so the airport has alien messages or maps on the floor. i made cards for all the guys and ben liked his. i kinda left em in the car after dad dropped us off at the gate and had to walk to the car and get em. the airport has stupid style sense. rj\'s stupid and didn\'t know where i got my pictures when like half of them were from his moblog. even ben knew they were. btw, he had these uber nifty champion pants that were like my shorts, but pants. (shut up dru, i know i\'m captain obvious!!!!) they paged travis on the intercom hehehehe me and dad saw him and his mom when we got starbucks. while we were gone billy and theresa showed up. it turns out this is only theresa\'s second time up in a plane!! wow. i would die if i never rode a plane until my son\'s nationals. guess i won\'t die though cuz my first time in a plane was when i was like three! i would probably be too freaked out to go i waited that long. but ya she liked billy\'s card and said she wanted to put it in a scrapbook. isn\'t she just the sweetest? ya and billy has to read that elie wiesel book too. so ya, we all chitchatted. then they went through security and all that good stuff. mum still had rj\'s magazine so she gave it to billy. theresa had to take off her shoes and all that good stuff. thank god i didn\'t go, cuz it would take a while to take off my ballerina shoes. oh ya, i wore my black and white hibiscus dress with my turquoise roxy tank and my pink roxy tee over it and my ballerina shoes. read tuesdays with morrie. sad book, good book. pissed me off :)drove home. nothing happened.     3. i babysat at the volunteer appreciation thing. sssoooo fun, kinda, well not really. ya so that\'s a no. one girl georgia is so gonna be a klepto when she grows up. she kept stealing everyone\'s food and sippy cups and knocked over a lot of juice. there were two adorable girls though. one was maggie who looked like sydney (her and her brother patrick matched with the whole red, white, and blue thing). and stacy who i kept calling grace but tracy and libby told me it was stacy, but i swear when her dad picked her up it sounded like tracy. then there was another kid named brendan who had an accident of the poopy sort and was alone in the bathroom and got it all over his legs then threw the diaper down and tried to wipe his butt off on the walls. ya , meagan (adultish chick who knows a buncha sta teachers including flip who had her baby) must have had fun with that cuz the girl with the very low cut dress and big eyes couldn\'t open the door to the other bathroom where she was alone and started crying. i know her pain. i remember how i couldn\'t get out of my room once. it was an epic journey filled with anger, frustration, and tears tears tears. Devin we\'re makin a movie!!!!  me tired and hungry. will leave now. byeizzle.   '),(9275,3561749,'14,June,2004','2004-06-14 00:00:00','    i hate comps. i made the greatest post and then i try to close a popup and it closes all the windows! bitch!!   this is an edit. it wouldn\'t publish my other half of the long post that i tried to split up. so take this comp bitch! i copied it before i published!!! i\'m takin another root to get the word out!! die evil demon. you wicked servant!!!(don\'t be scared daniel)  that was a long entry, figured i\'d split it up.      2. RJ left for nationals yesterday. so we drive out there and ben\'s already checked in and stuff. he packed all his stuff for five days including uniforms and stuff into one carry on duffle that still has room!! he is sooo not wearing underwear!! but like ya so the airport has alien messages or maps on the floor. i made cards for all the guys and ben liked his. i kinda left em in the car after dad dropped us off at the gate and had to walk to the car and get em. the airport has stupid style sense. rj\'s stupid and didn\'t know where i got my pictures when like half of them were from his moblog. even ben knew they were. btw, he had these uber nifty champion pants that were like my shorts, but pants. (shut up dru, i know i\'m captain obvious!!!!) they paged travis on the intercom hehehehe me and dad saw him and his mom when we got starbucks. while we were gone billy and theresa showed up. it turns out this is only theresa\'s second time up in a plane!! wow. i would die if i never rode a plane until my son\'s nationals. guess i won\'t die though cuz my first time in a plane was when i was like three! i would probably be too freaked out to go i waited that long. but ya she liked billy\'s card and said she wanted to put it in a scrapbook. isn\'t she just the sweetest? ya and billy has to read that elie wiesel book too. so ya, we all chitchatted. then they went through security and all that good stuff. mum still had rj\'s magazine so she gave it to billy. theresa had to take off her shoes and all that good stuff. thank god i didn\'t go, cuz it would take a while to take off my ballerina shoes. oh ya, i wore my black and white hibiscus dress with my turquoise roxy tank and my pink roxy tee over it and my ballerina shoes. read tuesdays with morrie. sad book, good book. pissed me off :)drove home. nothing happened.     3. i babysat at the volunteer appreciation thing. sssoooo fun, kinda, well not really. ya so that\'s a no. one girl georgia is so gonna be a klepto when she grows up. she kept stealing everyone\'s food and sippy cups and knocked over a lot of juice. there were two adorable girls though. one was maggie who looked like sydney (her and her brother patrick matched with the whole red, white, and blue thing). and stacy who i kept calling grace but tracy and libby told me it was stacy, but i swear when her dad picked her up it sounded like tracy. then there was another kid named brendan who had an accident of the poopy sort and was alone in the bathroom and got it all over his legs then threw the diaper down and tried to wipe his butt off on the walls. ya , meagan (adultish chick who knows a buncha sta teachers including flip who had her baby) must have had fun with that cuz the girl with the very low cut dress and big eyes couldn\'t open the door to the other bathroom where she was alone and started crying. i know her pain. i remember how i couldn\'t get out of my room once. it was an epic journey filled with anger, frustration, and tears tears tears. Devin we\'re makin a movie!!!!  me tired and hungry. will leave now. byeizzle.   '),(9276,3561749,'14,June,2004','2004-06-14 00:00:00','    I lay my life before you/and i\'m not getting up/father how I adore you/those words are not enough   they have some awesome contemp. xtian songs and blink-esque style in their oddities. gotta lurve that lead singer too.   '),(9277,3561749,'13,June,2004','2004-06-13 00:00:00','    BORDERS ROX MY SOX   I luv getting free advanced screening passes for movies. That\'s how I saw  Win a Date with Tad Hamilton! ,  Ella Enchanted , and now The  Notebook . Go to Borders and look for passes at the register.   Fun times for the whole family!   '),(9278,3561749,'13,June,2004','2004-06-13 00:00:00','    It sssssssssssoooooo wasn\'t bad!   In fact, it was downright fantabulous. I am now a certified eucharistic minister! i was already a lector, but now i\'m doubly a lector. i appreciate the mass and services so much more now that i understand them.   and the people, whoah! they were completely normal! i didn\'t find one jesus freak kid!!! and hear you me, there was some nice scenery out there if you know what i mean. me and the girls decided that hgtv needed to check out starving artist, angry celloist, and a guy with some great garden tools. 2% body fat! aaarrrgh!!! and then there were those hillarious guys u always laugh at. i sat in on some rocktastic jam sessions. oohhh, and I learned to play a beat on the drums!!!  i liked the music we played @ HYU a helluva lot better than the stuff at St. Peter\'s and STA. It was more upbeat and fun! (Short! 5ive! Elbow!)and I was in luv with our skits!! I got to be an angry yelling manager and the devil inviting HYU attendees to a party!!! Jesus (Hayzoos) rox my sox! Ok, enough for now, must do laundry and get ready for Super Sleuth camp tomorrow.   '),(9279,3561749,'06,June,2004','2004-06-06 00:00:00','    i\'m gonna start up a lil\"kids these days\" series. it\'s a tribute to all those annoying talks you\'ve ever had with your grandparents, but a with a katrina kick to it all.   BABY FOOD   What\'s with all the baby food kids these days get? I was roaming around price chopper yesterday, looking for some pepperidge farm light style 7 grain bread when i fatefully stumbled upon the plethera of baby food and all that good stuff. whoah, i mean WHOAH! they have miniature plastic containers now that are stackable  and  resealable!! all we had were those round glass thingies with the twisted lids that didn\'t like to match up with their bottom half. and have you been in the pressence of the gourmet choices babies get?! there was dutch apple desert, chicken dinner with carrots and peas, plum apple grape juice, and things that i don\'t think were in english. it was like i walked into willy wonka\'s chocolate factory. whatever happened to mushy peas and that purple stuff you wanted to say was prunes but you just weren\'t sure?   am i the only one here who thinks that babies are eating a helluva lot better than us? for serious, i usually rock the salad, sandwich, fruit meal a la moi. it\'s as if the guys on iron chef are cooking up new baby food or something. gourmet baby food, i mean seriously. it\'s like i\'m taking crazy pills.     kids these days don\'t know how good they have it. die gourmet baby food! word.   '),(9280,3561749,'06,June,2004','2004-06-06 00:00:00','    so i do believe i have peaked on the stupidity scale. i\'m doing this HYU camp thing (read:\"jesus camp\").  i thought the first day was today. so me and my dad drive an hour and 45 min out to maryville, mo,get a little lost on campus, and eventually find the place. no one\'s there. i was getting a little irked. then this lady tells us it starts tomorrow. well woopee. then we drive the hours and 45 back to kansas city. fun times, ya?   '),(9281,3561749,'06,June,2004','2004-06-06 00:00:00','    Okay, so here goes the first.  Hey there happy campers Welcome to my Slam Feel free to roam around Hope you don\'t receive spam  I needed a place To let out my noise It seemed like the web Would be my best choice  So I went online Blogspot is pretty easy and all Not too great with sites I\'m ready for a fall  But screw that thought Just read on Cuz soon one day It\'ll all be done  So completely rocktastic In all it\'s majesty Feel free to mock me Just another travesty  So I\'ll shut up now Well really I won\'t Want to leave already Don\'t   '),(9282,3561749,'12,July,2004','2004-07-12 00:00:00','    oops, said the entire message in the title. anywhoozles. will finish earlier entry later       for now:  I\'m passed out on the overpass  Sunday best and broken glass  Broken down from the bikes and bars  Suspended like spirits over speeding cars  You and me were kings over the parkway tonight  And tonight will go on forever while we  Walk around this town like we own the streets  And stay awake through summer like we own the heat  We\'re singing   Everybody wake up  (Wake up)  It\'s time to get down.  (Everybody Everybody wake up, it\'s time to get down)  And when I pass the bottle back to Pete  on the overpass tonight, I bet we laugh   I\'m gonna stay eighteen forever  (Cut me open)  So we can stay like this forever  (Sun poisoned)  And we\'ll never miss a party  \'Cause we keep them going constantly  (This offer stands forever)  And we\'ll never have to listen  (New haircut)  To anyone about anything  (New bracelet)  \'Cause it\'s all been done and it\'s all been said  (Eyeliner)  We\'re the coolest kids and we take what we can get  (Wait forever)   The hell out of this town  And find some conversation  The low fuel lights been on for days  It doesn\'t mean anyhting  I\'ve got another 500  \'nother 500 miles  Before we shut this engine down  (We shut it down)   I\'m gonna stay eighteen forever  (Cut me open)  So we can stay like this forever  (Sun poisoned)  And we\'ll never miss a party  \'Cause we keep them going constantly  (This offer stands forever)  And we\'ll never have to listen  (New haircut)  To anyone about anything  (New bracelet)  \'Cause it\'s all been done and it\'s all been said  (Eyeliner)  We\'re the coolest kids and we take what we can get  (Wait forever)   (You\'re just jealous cause we\'re young and in love)  Eighteen forever  (Your stomachs filled up but you\'re starved for conversation)  So we can stay like this forever  (You\'re spending all your nights growing old in your bed)  And we\'ll never miss a party  (And your tearin up your photos \'cause you wanna forget... it\'s over)  \'Cause we keep them going constantly   (You\'re just jealous cause we\'re young and in love)  And we\'ll never have to listen  (Your stomachs filled up but you\'re starved for conversation)  To anyone about anything  \'Cause it\'s all been done  (You\'re spending all your nights growing old in your bed)  And it\'s all been said  (And your tearin\' up your photos cause you wanna forget... it\'s over)  And we\'re the coolest kids and we take what we can get  Just jealous cause we\'re young and in love  Just jealous cause we\'re young and in love  Just jealous cause we\'re young and in love  Just jealous cause we\'re young and in love  Just jealous cause we\'re young and (turntable scratch)        the rest of my life is a long time to pretend ya, so goodbye to all of that stuff      sorry. IMU   '),(9283,3561749,'06,July,2004','2004-07-06 00:00:00','    Current song on replay: \"Soco Amaretto Lime\" Brand New/\"Those Words Are Not Enough\" and \"Mood Rings\" Relient K  Now playing (fer serious): You\'re just jealous cuz we\'re young and in luv Ur stomach\'s filled up but ur starved for conversation And ur spending all your nites getting old in ur bed And ur tearin up ur photos cuz you wanna forget    it\'s over     Fun of the Outdoorsy Sort       Ok ya so here\'s the skinney on Sat and Sun.   Saturday       Mens doubles. RJ and Billy. Fun times. Ben brought his mom\'s tent!! IT didn\'t really work out though because by the end of the day there was about four square feet of shadow me, RJ, Ben, Billy, Travis, and Austin crammed into. one sec, got somethin to do real quick. Yes that\'s u Jack R, I know ur wonderin, but ya it\'s u.   '),(9284,3561749,'06,July,2004','2004-07-06 00:00:00','    no, okay that sounded really wrong as in not right! not  doing  Jack R. real quick. talkin on the fellytone. and yes jack r. this post is the reason i\'m taking so long and not one sec. shut up! when someone says one sec they never actually mean one sec.   '),(9285,3561749,'06,July,2004','2004-07-06 00:00:00','    rj\'s back from nationals. they got 21st out of like 67 ish teams i think. cool beans. more later. i luv gilmore girls and summerland!! can u believe it? bradin did  it ! And Sara\'s all avoiding him and with another guy and stuff! I luv drama.          If it\'s drama you want then look no further. It\'s like the Real World meets Boy Meets World meets Days of Lives.   '),(9286,4224299,'10,August,2004','2004-08-10 00:00:00','    urlLink    hehe...women\'s soccer team...(hindi halata noh!) girly girl kasi kami sa debut ni kella ...:)  -=:shei-shei:=- '),(9287,4224299,'10,August,2004','2004-08-10 00:00:00','    urlLink    hay..miss ko na din ang soccer!!hehehe..it\'s been two weeks since our last training...(cool ng pix namin dba?)hehehe Ãœ  -=:shei-shei:=- '),(9288,4224299,'10,August,2004','2004-08-10 00:00:00','    urlLink    nax...yellow day daw oh!!!hehehe  -=:shei-shei:=- '),(9289,4224299,'10,August,2004','2004-08-10 00:00:00','    hmmm.....this has been a very busy and tiring day for me...our whole RLE group stayed in Krissy\'s house last night just to finish all the stuffs and by 7:30 am we had our case presentation and we\'re lucky because Ms. Reyno is in good mood...(kahit na ang gulo ng presentation!) in fairness, yung case namin super challenging! hehehe...after that, jen, macy, karol,yana, tom, mike, seth, maycon and I ate lunch together....  -=:shei-shei:=- '),(9290,4224299,'10,August,2004','2004-08-10 00:00:00','    urlLink    hay...i miss katoinks! sana makapag-gimik uli kaming lahat....dami kasing schoolworks..busy lahat!..... :)  -=:shei-shei:=- '),(9291,4224299,'10,August,2004','2004-08-10 00:00:00','    urlLink    sadako or assassin?!?  -=:shei-shei:=- '),(9292,4224299,'10,August,2004','2004-08-10 00:00:00','    urlLink    purple madness.....  -=:shei-shei:=- '),(9293,4224299,'10,August,2004','2004-08-10 00:00:00','    it has been a relief for me because the classes were suspended today...I was able to have a straight 8 hours sleep (which I don\'t usually attain in regular days!)....but still I\'m so busy with my NCA...(for those who don\'t know about it...Its a case analysis of my patient during my Related Learning Experience) and it was already due on WEDNESDAY...argh!....I\'m not yet ready for that...how I wish my clinical instructor will be considerate enough to move it next week.....      anyway....enough of that....I just want to relax even for just a little while...that\'s why I surf the net instead of reviewing tons of chapter in my Pharma book....maybe I can still rush on reading it in the morning....      so...i\'m glad I was able to chat with my friends...its really been a long time.....    I was able to talk with my friend ryle...hmmm...I was not expecting that he will open up a topic like \"modeling\" ...but then, he did....      he was asking if I had an experience in modeling before, particularly in ramp modeling....sigh!....I thought it was just a joke....yet its true....I told him that I didn\'t have any experience in ramp modeling but I was once able to be a SLIGHT image model for a campaign ad.....I also told him that I was just 5\'4\" and it is obviously not accepted....but then he told me that my height is not really important ...what important is the way I walk and well of course...my creativity....and he gave me until the end of the month to prepare my stuffs like the personal data, pictures, etc....he also oriented me about the screening....      i\'m still undecided if i\'m going to join...      I need to ask for other people\'s opinion...hmmmm   ......\"di tuloy ako makatulog sa pag-iisip.....\"       hay...whatever..... '),(9294,3594873,'11,June,2004','2004-06-11 00:00:00','   Okay, so where did we leave you guys? Oh yeah, we just told you our names and we let you guys see us chatting. Well, where do we live? We live in the â€œscaryâ€ Middle East. Speaking of living, you know, Iâ€™m sure youâ€™re all aware of the fact that there are seven characteristics an organism has to have/ perform in order to be considered alive. They are: responding to stimuli in the environment, â€¦â€¦â€¦â€¦â€¦.(skip the other 5 characteristics), and breathing. Now, you know those people who are on respirators? Are they considered to be alive; because in reality, they are not the ones actually performing the function of breathing. I feel really horrible for all their relatives who have to go through so much pain, but Iâ€™m just curious. Are they considered living? And now for our little story about lifeâ€¦  The story takes place in nowhere. A place very familiar to lots of persons- especially nobody. Nobody who does nothing all the time in this nowhere. Night and day- nothing.  But this is the story the way I heard it straight from somebodyâ€™s mouth, and I canâ€™t change one bit of it.   The story begins with crying and ends with crying. The middle is laughter. Like life: the first moment we breathe, we cry because of the world we have just entered and the one that we have just left. And when we are dead, our relatives cry. But in the middle, we laugh.   If I have gotten your attention now, if I have sucked it from you, inspired you, then nothing. I donâ€™t care. Because if I told you, well then, it would be a something, and this story- this story is a nothing.   Now for some serious thoughts: Those people at the airport. You know who Iâ€™m talking about, right? Those serious men who stand and take your passport from you, like itâ€™s a bomb, those men. They open up to your picture and begin to stare at it. They have to concentrate very hard- their head turns from the picture on the passport to your face and back. Like a dog whose shaking its head to get rid of water. Now, these people bother me. They irritate me. I mean, do they have some kind of make up/ disguise education? Are they very talented artistic people or something? I mean, somebody trying to escape or use a different passport isnâ€™t going to be stupid. He/ she is going to transform themselves so that they look exactly like the picture. And these people are dressed up like police officersâ€¦(over here, at least). Whatâ€™s with that? What kind of a police officer gets stuck with that job? Looking at pictures all the day. â€œOh, look!! I found a criminal!â€ â€œI swear its me. I swear, I just got a pimple attack, but that is me.â€ â€œGet out here, and have a good, safe trip.â€ Who needs these stupid police officers?!  Oh! Since I wrote that the police officers look like dogs, I just wanted to say, that all of you people are dogs. Yes, you are my dogsâ€¦.seriously, since everyone goes like â€œdogs are so lovable. Theyâ€™re a manâ€™s best friend. They are loyal and sweet. They are so affectionateâ€, if they are all that, then why do people get insulted when someone says â€œhey b****?â€ Shouldnâ€™t we think of it as a compliment? And, why is it the female dog is an insult? Whatâ€™s so bad about a female dog? Why is their name taken to mean something horrible?   Speaking of weird language and stuff: What is with the expression: time on your hands: if I had time on my hands,  I would go and wash my hands, rinse off all the bad times, cleanse off all the embarrassing timesâ€¦..erase all the times that when I think about them, I shiver and can feel waves heaving and rolling in my stomach.  Times that if I think too much about, I feel regret that matures into a full grown tree, and I wish that I could just go back and change itâ€¦ yeah, if I had time on my hands, then I would wash it off, so I wouldnâ€™t have time on my hands anymore.  Oh hey, if you guys have some time on your hands, comment on the post :P :P.   What I just wrote reminded me of something else I thought about before. You know how I wrote I would erase all the bad times. Well, I wouldnâ€™t really because I donâ€™t know how an eraser works. I truly have no clue to the mysterious functional capabilities of an eraserâ€¦â€¦..i donâ€™t even want to think about how an ATM machine works. No wait, I do. How does it give us money back? How does it know how much to give? I know it has to do with electronics. That scary world! I donâ€™t want to know anymore...natasha, cut the engineering lecture. I said I donâ€™t want to know.   Natasha is my imaginary friend. She thinks you guyz have some mental issues. :D But Iâ€™ll tell you who has  real, major problemsâ€¦ a dentist. They have major issues. I mean, who grows up wanting to stick their head in somebodyâ€™s mouth and look at all the disgusting stuff stuck there. Thatâ€™s if they even make it thereâ€¦someoneâ€™s bad breath might just be a toxic chemical that surrounds the dentist chocking him/ her and preventing him/ her from breathing until the dentist turns red and finally departs from this world. Then to look at all of those brilliant white teeth that have been stained by coffee, soda, and everything else a dentist hates until they have become yellow. Next, the cavities. You see where Iâ€™m getting at?   And dentists are the biggest liars. They all say, â€œOpen up your mouth. Itâ€™s not going to hurt you a bit.â€ Excuse me, if it looks like a drill, sounds like a drillâ€¦well than it is a drill thatâ€™s going into my mouth. But hey, yeah, it doesnâ€™t hurtâ€¦actually you canâ€™t feel anything if you get those shots that donâ€™t let you feel; (I canâ€™t seem to spell it) But oh my jumping, sprouting, spinach balls. Have you ever gotten one of those? Those shots are the most painful thing, ever. I swear, it feels like somebody is pushing something down into your mouth and the pain just goes directly to your brain.   That is if you have a brain? Gotta goâ€¦its time to save the world/ or maybe it was take over the (Hey, donâ€™t tell them!!)  '),(9295,3594873,'10,June,2004','2004-06-10 00:00:00','   Spring is here. Its bright light shines, warming the sky. Its gentle, soft breeze embraces the dancing trees. The sky is blue, a clear blue that speaks of children playing in the park and family picnics. The flowers wake up from their long hibernation and gently unfold, facing the powerful sunâ€™s rays. Birds chirp and sing to each other of romance. A caterpillar fights desperately pushing its cocoon, its wrapped blanket, and emerges as a splendid butterfly gliding the waves of air ever so lightly. Bees hover over their hives, busy work stations, producing sweet, golden honey.   Winter is gone, shed like a snakeâ€™s skin for spring. Its spotless coat put away in the closet.   Amina:  And then, I wake up from the terrible nightmare I was having, and find to my joy that all is well with the world. There is still the pungent smell of smoke, a layer of smog covering the sky, and dead birds falling from the sky. Yes, Chicken Liken was wrong: the sky was not falling: it was just the beautiful world on one of its regular routine daysâ€¦ (sniffs the nice smell of pollution and feels once again refreshed.)   Mariam:  My poor darling aMinA. Chicken Licken was incorrect in her assumption as are you. It is neither the sky falling nor the smoke which has created such drastic changes in our country so rapidly, and with no warning. No, youâ€™re both wrong.   Amina:  You know my insignificant one, that chicken licken has an usual name. I doubt anyone would want to be eaten, and yet she calls herself â€œlicken.â€ However, that is not as strange as the name â€œwinnie the pooh.â€ Please explain to me the significance of having the word pooh inside the nameâ€¦ Aaah, I have realized I didnâ€™t ask you the reason you are so convinced of my wrongness.    Mariam:  Why you are wrong:  It is simply because Winnie the Pooh ate up all the honey and then had to go do doo doo. So a stinky smell engulfed the world. And people could no longer breathe. If the earth, itself, not alive could not take it and suffered from a hole within its atmosphere, it is no surprise that we humans are unable to.    Amina:  you do realize why there is a hole in the atmosphere. It is because the politicians know that everyone hates them and so they have planned an escape route in case, at some point, a fanatic tries to kill them. So the truth of the matter is the hole is just a way for the government, once again to trick us all and escape with all our mulla. Our doe. Our money. Show me the money. What are you talking about honey for?    Mariam:  You know, I heard that theory but rejected it. The truth is: there is no hole. I donâ€™t see a hole in the sky. Do you see a hole in the sky?! No. Name me one person who has actually seen it? For all we know, people can be getting cancer because of the food or some governmental created project with us humans as the rodents- testers. You know, I realized that I did not comment on chicken licken. An unusal name, to be sure. Yet, hardly surprising. Her mom was a chicken rib, and her father was an absolute chicken drum.     Amina:  I have neither seen a hole in the sky nor the whole of the sky to know whether there is a hole in the sky. However, I have been assured that there is indeed a hole in the sky.  ________________________________________________________________________  and so you can see why we wished to show you the important matters we discuss. Seriously though, whatâ€™s up with name Winnie the Pooh. I mean instead of the characters calling him Winnie, they go like â€œPoohâ€¦.Pooh bear.â€ What kind of a sick joke is it to name a beloved character to kids after the number 2 we do in the bathroom? And then, I canâ€™t believe it when I hear one of the characters going like â€œPooh, where are you?â€ Iâ€™ll tell you where pooh is. Pooh is in the bathroom where it is supposed to be. Honestly, I can not for the life of me, figure out what was wrong with the directors of that cartoon.  _____________________________________________________________________________   Amina Banawan Amina Banawan, amina Banawan,  Amina Banawan, Banawan Amina  Mariam Banawan Mariam Banawan, mariam Banawan,  Maraiam Banawan, Banawan Mariam  If this website does not appear when I write â€œAmina or Mariam Banawanâ€ in the google search, I am going to explode. I mean every time I search myself, I never come up with amazing moi: people donâ€™t know that I am alive.  Itâ€™s like google doesnâ€™t recognize my nameâ€¦.or just doesnâ€™t want to.   I have decided that I am going to copy our names like twenty times on this blog just to make sure the website does appear. I want people to know that I exist. I want a person to type in my name on the net, not really thinking about it, and then discover accidentally that we do indeed have a site. Soâ€¦â€¦.  Amina and Mariam Banawan.  Mariam and Amina Banawan.  Ms. Mariam Banawan   Ms. Amina Banawan A&M Banawan M&A Banawan   Mariamina Banawan.  Amina & Mariam Banawan Mariam & Amina Banawan   Isnâ€™t it ironical? Ashley and Mary Kateâ€™s names also start with A & M. Once I told my friend that the reason why our parents picked out our names with these letters was because they wanted to match Ashley and Mary Kate. It was so funny because she actually believed meâ€¦well, why not? I am a convincing actressâ€¦right ïŠ   Back to the task at hand: Amina and Mariam Banawan. Mariam and Amina Banawan. Amina and Mariam Banawan have a website, and if google wants to not include them, well then, thatâ€™s googleâ€™s problem. Mariam and Amina will go check out yahooâ€™s search engine. ïƒŸ And that is to be taken as a threat, google. We are deadly serious about this. Weâ€™ll see if yahoo can find us or not.   Well itâ€™s time to goâ€¦. But remember:   WE will Be back!!  ........................  And we are back!! Yesterday, on my first real day of summer, my parents plunged us back to the dark ages. Quietly, they disconnected the television from the satellite to broaden our interests!! To use our summer wisely?! Thankfully, all has been restored in our house after a little revolution- some people were taught lessons. JUST KIDDING!! I just woke up and found it back on. Only weâ€™re only allowed an hour on it, but not to worry. Each one of us will choose a different hour and therefore gain an hour! The smarties that we are. =)  Talking about television: didja you guyz hear about Tarzan being banned? Didja hear about the reasons: They are so melodramatic. â€¢	The fact that Tarzan had to help/ save Jane. The feminist movement said that this made girls believe that they are unable to protect themselves and need the help of males. They argued the movie had a brainwashing effect and promoted sexism in society.  â€¢	Police all over the country said they are fed up of movies drawing their â€œevilâ€ characters, ugly and big. They reason that criminals have no general appearance and children should be aware of this. After all, they think movies should teach children to expect the unexpected and know that even their teacher could be a criminal. â€¢	Orphanages were outraged when a five year old ran away. When found and questioned on his reason, the boy said he wanted a home. He distinctly commented on the fact that Tarzan had been embraced by animals and had had a home with them. If living in the jungle was all he needed to do to get a home, well, he could certainly do that.  â€¢	Foster homes were appalled at Kojakâ€™s character- they fear children will be afraid to go to foster homes because they might feel that they will not be accepted by both parents.  â€¢	Religious leaders are worried that the movie subtly promotes the message that the theory of evolution is correct and that humans did evolve from monkeys. â€¢	The Environmental Organization found the appearance of the jaguar as a bad guy, would help encourage people to shoot these lovable creatures. They are concerned that children will grow up, believing that these animals are horrible predators and would thus do nothing to help protect them and save them from extinction. â€¢	Physiatrists thought that the movie further increased peer pressure in childrenâ€™s minds. â€œYoung Tarzan had to prove to those other monkeys that he was worthy of their friendship. He succumbed to peer pressure and dived from a waterfall. We think this behavior teaches kids to do whatever they can to be accepted by a gang. We find it rather disturbing,â€ the physiatrists said.  â€¢	The â€œLASâ€ (Love All Society) said it did not like how the characters in the play seemed to promote specieism. They said â€œItâ€™s enough we have racism. Certain characters in that play promote specieism. They acted like humans and animals could not get along because of their differences. We would appreciate it if the characters were censored.â€ â€¢	A Womanâ€™s Rightâ€™s Movement Organization said they did not like how Jane took off most of her clothes at the end of the movie and only wore what appeared to be a bra, and a tiny version of a mini-skirt. They claimed that the way women are pictured in movies are horrible: but one would expect better from a childrenâ€™s movie. Women did not have to look like models and dress this way: they had much more dignity than that. They think the movieâ€™s underlying message is that nudity should be acceptable in our society.    If you guys did hear that the movie was banned, well then, you heard wrong. It wasnâ€™t banned. We just made up all of those reasons.  (Losers, you actually believed us. Stop shaking your head. You did! I know you did!! Hehe)  ___________________________________________-  '),(9296,3594873,'10,June,2004','2004-06-10 00:00:00','   This blog is dedicated in loving memory to Salma\'s brain which has just recently been lost. Please, it is urgent, if somebody finds it contact us. There\'s a hefty reward!!   seriously: thanks salma for actually inspiring us to inspire the world.   Hello0o0o0 world.   Itâ€™s us. We really do exist.   The crazy_twinz.  How does one go  about and start their first blog? Hmmâ€¦I think that we should introduce ourselves.  Weâ€™re the Banawan twins and well, weâ€™re often mistakenly called crazy. The truth is, weâ€™re not crazy, but rather, gifted. And so we have decided to share with the world all of our â€œgiftedâ€ thoughts.   Itâ€™s actually kinda funny. Almost everyone at one point or another ends up calling us crazy, and a lot of people ask, â€œhow can you talk to each other all day?â€ because, well we do talk to each other the whole dayâ€¦and we do that because, we have weird thoughts the entire day that need to be discussed, analyzed, and made fun of .  So, here you go: weâ€™re opening a window into our lives; letting you take a glimpse of all that we do. â€œwe have a world of our own. Weâ€™ll tell you things that no one else knows.â€ weâ€™ll make you go crazy, but at the end of this, youâ€™ll love us :D :D :D  so you get to read former chats we had with some people, notes we pass to each other during class, and things we think up of.   Disclaimer: anything that you read, you read at your own jeopardy.  Warning: if after reading the blogs you find yourself have insane or strange thoughts, and seeking the assistance of a physciatrist, we shall not take any blame.     Thought number one: What is with the name Atomic kitten? Why is it necessary to add such an explosive word such as atomic to a friendly word such as kitten? I mean you might as well call yourself â€œRaging Butterfly; Bitter Daisy, Furious Bride.â€ Itâ€™s like these days itâ€™s not acceptable to be nice and sweet: you have to have anger with the world to make it. You gotta be a â€œ Fiendish Angel.â€   But I donâ€™t want you guys to think that a kitten is actually cute. No, just because I called the word kitten in the beginning friendly doesnâ€™t mean that they really are. The word kitten connate images of friendly little fluffy creatures; but in reality a kitten is a dangerous miniature tiger. It is a deadly creature with fangs, claws and what one could call a progressing roar.  Kittens are not innocent creatures: they act all cuddly but right when you pick them up, they scratch you. You try to get them off, but you canâ€™t. You run around the room screaming for someone to help you but instead, you find everyone else running away, shrieking and you think, â€œhelloo0o Iâ€™m the one with the insane kitten clinging to me.â€ And you try to throw it off and hurt it, but then you realize that kittens always land on four feet and so there is no point in even trying to help it commit suicide by putting it on the balcony.  And so you give up to the fact that you do indeed have what one could call â€œan atomic kitten.â€  Thus, I have arrived at the conclusion that the reason one tends to call oneself such outrageous names is because one realizes that nothing in one hundred percent angelic, nor sweet. There has to be a hidden tiger somewhere or other inside of them.   But you, know, I â€˜m getting ahead of myself. I should start at the beginning. You know, Iâ€™ll start by giving you an autobiography. There are two of us- obviously since weâ€™re twinz. Actually, up until this year we were the only twins in my family. We were special in that way. But then, my uncle got married and had twins. Almost exactly 16 years later. Itâ€™s funnyâ€¦my cousins have not only taken what was special about me, but managed to do it right almost my birthday so that there is nothing really special.   Why did you say: Itâ€™s funny. I donâ€™t see whatâ€™s so funny about it!   Speaking of having young cousins: I hate it. Itâ€™s just like what one of my best friends said: when you go and visit your relatives, they dump you with a two year old and say â€œgo play.â€ What the heck go and play? Iâ€™m 16 years old not a toddler. What they really mean is go babysit the kids, make sure they donâ€™t cry, clean their crap after them, and have fun in the process of doing what I donâ€™t want to do: have fun taking care of the kids that I donâ€™t want to take care ofâ€¦ the ones I dumped you with and make sure to burp them, and â€¦..the list doesnâ€™t finish.   I have just changed my mind. It would be utterly impossible for me to write an autobiography now. Too many stories, or anecdotes. Too many characters and no plot.  You know, its summer here. The first real day of summer, but it feels like the weekend. Those two days that one always waits for in the beginning of the week, which in our case is Saturday. (That means our weekend is Thursday and Friday)   From about two weeks, the yearbooks have been making their rounds. Time to write messages. It suprises me how vain people are, myself included. Everybody wants their yearbook to be filled with messages on how great of a friend and person they are. Somebody you just said hi to once, becomes another person who just must  write in your book. Your yearbook wouldnâ€™t be complete without their inspirational writing. So your book gets filled with bull and you can go home and read all those messages and marvel how youâ€™ve affected so many people. There will be a couple messages with your name spelled wrong, but you wonâ€™t let that bother you. Things that you have no recollection of doing  at all or saying, but because their message will include something about your â€œsweet, beautiful natureâ€, youâ€™ll accept it readily enough.   Almost all my messages had the same meaning: Youâ€™ve been such a great friend. Youâ€™re so sweet and I can always rely on you. Good luck on your exams. Have a great summer. And Iâ€™ll miss you. How pathetic!! But then there are the true friends: and one doesnâ€™t know or just canâ€™t put to words how one feels. Because our heart has not a mouth for itself, and so our feelings can not be translated into words and put on paper.  Enough of the philosophy youâ€™re killing me!! By da way, some of you guyz didnâ€™t sign my year book yet. Itâ€™s kinda empty. I wrote in yours!     I hope you are all reading this in a British accent: not that I have one. Hahaha. I just want you guys to sound like dorks. Not that the British accents are dorky: itâ€™s just that people always sound like dorks when they try to make an accent. Haha!!  You know when we were trying to make up a name for our blog, our sis, offered â€œfairy dustâ€. Such a cute name. And thereâ€™s star dust too. Thereâ€™s all these sweet, girlish dust names. But Iâ€™ve never seen dust like that; all Iâ€™ve seen is JUST DUST! Little unwanted, black particles filling up the air causing diseases...  I ,now, know the true mark of a pessimist. Hereâ€™s the true test: Tell the person in question â€œYouâ€™re a pessimist.â€ In almost every case, the pessimist will say, â€œIâ€™m not a pessimist. Iâ€™m a realist.â€ And thatâ€™s how you know. As for us, weâ€™re not pessimistsâ€¦weâ€™re just realists.   Iâ€™ve always thought it is stupid to be an optimist. I donâ€™t see the point. Once, you have all these expectations and hopes you can only go down. Whereas, if you expect the worst thing in the world and imagine it 10x worse, then you can only go up. And feel better. It might be cowardly, but heyâ€¦its going to save me a lot of money at the shrinks.   So what was I saying again? Oh yeah, weâ€™re Mariamina. Not Mama Mia, Mariamina. See, my name is Mariam. And my name is Amina. But if you write Mariam and then add ina, you have both our names. Mariamina. Thatâ€™s us!! Or you could just write half of mariamâ€™s name: which would be mari, and then add all of aminaâ€™s name: so you would get Mariamina. My parents are clever wits arenâ€™t they. Not! They didnâ€™t figure it out. We found it out by an accident: like all true discoveries that help humankind, we discovered an interesting way to combine those two precious names â€œamina and mariamâ€.   Goodbye peoples. Take care, and remember to reach for the moon because even if you miss, youâ€™ll land among the stars. Remember, always dream and hope â€¦  '),(9297,3594873,'18,June,2004','2004-06-18 00:00:00','   Two Questions:      (ready?!)    1. Where is Waldo?  2. Can you show me how to get to Sesame\'s Street?  '),(9298,3594873,'15,June,2004','2004-06-15 00:00:00','   First of all, just wanted to make a very important introduction as hostess. Blog, the person eyeing you and reading every letter typed on you does indeed have a history with me. Itâ€™s my friend Sara A. Sara A, this is THE BLOG! Never, insult it by called the blog. It must always be capitalized. Hehe. Now, that I know there is a reader reading my inspirational and exceptional writingâ€¦that I am making a difference however slight it is in somebodyâ€™s day and thinkingâ€¦it is now that I feel that I have begun to accomplish what I set out to do as a child. What I dreamed about. I can laugh my as* off that somebody actually reads this cra*.   (Yes, I have decided that I will not become corrupted. My mouth will remain clean like a newborn babe. I will speak no falsehood nor utter what I do not know as fact. I will speak as an intellectual, knowing that I am an example to countless othersâ€¦yeah, right!!)  Iâ€™m typing this in my sisterâ€™s room, even though itâ€™s light is very dim. It casts shadows in the room that play tricks with my eyes fooling them; like mirages. Me- I like my light to be brilliant, to radiate, to spread everywhere â€˜till all sense of blackness has disappeared. To feel enveloped in this light that shines. Till it is beyond impossible to realize that it is night that I am writing this in.   Okay, now Iâ€™m in a different room. Literally. We moved back to our room.  You know I was talking to my friend today, after I realized a very embarrassing thing. I brought my mobile with me for like no reason. I mean not one phone call. Not one missed call. Nothing! Whereas my friends kept having to get out of the pool, run, and grab their mobile how many times. Really, its humiliating. I brought a phone with me for people to call me- the fact that nobody did means it was absolutely useless. I mean somebody wouldnâ€™t even have realized I had a mobile if I hadnâ€™t told them. Yeah, I can see it now.  â€œHey! Iâ€™m so down on credit. Could I borrow somebodyâ€™s phone. Pleaseâ€¦pplz keep calling meâ€. Everybody looks around, but nobody wants to part with this treasure. Itâ€™s beyond valuable; its vital for their existence. There isnâ€™t enough of it already. Only I scream out â€œSure, you can borrow mine. Iâ€™ve got lots of credit.â€ Wait a second. All heads turn around and find me. Everybody stares in disbelief and then everybodyâ€™s mouth forms the same question: You have a mobile?! But we never heard it ring.  And then, because Iâ€™m so embarrassed Iâ€™ll smile meekly and say, â€œYou could borrow mine if I had one.â€ Everybody will laugh and say, â€œOkayâ€¦ you actually had us for a second!â€  And now to our second chat that is more an insult trade.   Amina: You know, there are all kinds of idiots. Like your teacher is an idiot because he assigned you homework on a lesson that you didnâ€™t even complete. Your dad is an idiot because he canâ€™t understand that what youâ€™re wearing is style. The guy who takes your order and gives you back your food at a fast food place is an idiot because he brought you the wrong food.  Sara A is an idiot because well, she is.  Mariam: Ahhh! Such an expert on idiots. I presume this comes from first hand experience: analyzing and dealing with yourself, my lovely butthead. I would like you to know that Iâ€™ve had a deliciously nice time today at a swimming bonanza at my palâ€™s. Surely, I did. Actually, we stayed in the pool for such a long time that my skin became as soft as a childâ€™s, but as wrinkled as an old ladyâ€™s. Chlorine intertwined and shrouded my hair. I even got the markings of the beginning of a tan. The bronze color that I do love. You know, my chere, today I saw the strength of character somebody could have, but more importantly I discovered something. Do you know what it is?  Amina: I think my buttercup (notice, the usage of the word butt in that delightful flower) that I do know what you mean. You discovered how to use makeup. Was my deduction correct? Otherwise, how the heck can you claim that your skin was turning the color of bronze? Or maybe you discovered true imagination. Was that one correct? You know, my dumpling, I donâ€™t believe that our guests (those people who donâ€™t bother to write comments but read what we wrote) donâ€™t know that we actually made up a game, several years ago, that had the same idea as the â€œPowerpuff girls.â€ And to think that we could have been famous if we had sold our idea long agoâ€¦:(   Mariam: Ahhh! (Notice how my paragraphs always start like that) I see, that you have not discovered what it is. However, once a wise woman said â€œKnowledge comes not from within but from what diffuses from the journey of lifeâ€, I can not share this information with you. In case, you didnâ€™t guess who was the eloquent owner of this saying it is your butercup. (That my dearie, is the proper way to spell it when talking to moi). About our lovely game, please donâ€™t insult it and me. You know it was 100x better than the Powerpuff gurls: HeLlOoO it starred me!! Itâ€™s okay: to err is human; to forgive divine. Second of all, my plays and my genius are a gift and work of art. I can not even attempt to picture estimating and labeling a price for them. Cold, dead presidents (money) is not something that I (I wanted to make that capitalized, but it already is) will sell my soul to; no matter how tempting it can be.  Amina: Of course you wonâ€™t sell your soul: my dear teezacup (teeza is butt in Arabic :D) I wouldnâ€™t want you to do that. I didnâ€™t ask you to sell your soul. I couldnâ€™t ask you to do that. Okay, enough of the bull. Are you sure that there is no way I canâ€™t convince you to do it? To sell your soul? To get rid of it? Please, itâ€™s for humanity. Itâ€™s for the better cause. Itâ€™s for the good of Rome. Hhahahha Caesar has fried my brain. Or should that be Brutus and Cassius have fried my brain, since Caesar was dead. Nope, wrong again. Shakespeare has fried my brain. Oh wait, thatâ€™s still wrong. Of course they didnâ€™t fry my brain. You see my brain is, just, well a miracle. Itâ€™s so big and full of knowledge. Enough modesty: time to brag. My brain is huge. Itâ€™s loaded. Itâ€™s awesome, and the whole world wants it because they all know that I have the biggest weapon when it comes to mass destruction. Iâ€™ve got my calculating, intelligent, cunning, amazing, and mind boggling brain. Get it? Mind boggling?! Hahahahaha  Mariam: Zakhroota at the amount of bul* you are able to make up in very few seconds; at the most imaginative lies one could write. But I shanâ€™t give your tiny brain, another moment of my precious time. I have more important matters to discus: a toe that has a freckle, the freckle is brownâ€¦does this mean that the toe must have a brown spot? Confusing right. Allow me to make it easier; itâ€™s the pinky toe; the one the little piglet got stuck, crying, and couldnâ€™t find his way home. Thatâ€™s the toe.  Amina: you must be quite familiar with the little piggys. How do you spell the plural of piggys? I donâ€™t want pigs, I want piggys. It keeps marking it wrong. Oh well. Anyways, I was incorrect in writing that you are quite familiar with little piggys. I should have written that you are familiar with all piggys. And donâ€™t you dare call me a piggy, I know the way your sick brain works, Mariam.  Mariam: It shames me to dishonour my beautiful, honorable, held high, name by allowing you to write your name next to mine. (To pplz who are having difficulty keeping up with our influential corespondance, I urge you to get a translator to translate our exceptional, genius language into your idiocy. People of your slow capabilities canâ€™t do this by yourself)  Amina: you know what would be really sad, (youâ€™re right. Your name shouldnâ€™t be anywhere near mine, but for the opposite reason) if our soul mates somehow read this blog and find out that we are insane. Years later we catch up with them; they are exactly as we would want them to be, and we agree to get married, until they realize that we are the proud owners of this blog. Then they run like heaven (why is it hell?!) away from us, and we never get to marry our soul mates. All because of this blog.   Itâ€™s too much. We can not continue. I urge you Mariam to take into consideration your future and leave this blog to be.   Mariam: Oh, Amina. The minute my future husband (insha Allah) discovers you are my sister, I am a goner. (Donâ€™t even think about saying itâ€™s because heâ€™d fall in love with you). I donâ€™t blame him; I only put up with you because youâ€™re familyâ€¦there is no reason he should have to put himself in such a horrible situation. And I have a clue why its h-e- double hockey sticks/heaven. (but I wonâ€™t tell youâ€¦like a wise person once said â€œKnowledge is the waterwheel that soaks up information, turns it around, and absorbs all that is logical and correct.â€  THE END OF OUR CHAT FOR NOW. (in case you didnâ€™t know, mariam ended itâ€¦which means she won!!)   Amina: â€œ Bullshit, is what the bull excels at. Shit is what Mariam excels at.â€ Hey, guess what? Thatâ€™s my quote :D :D :D :D   * weâ€™re going now, just wanted to tell Sara A. we love you :D :D  '),(9299,3594873,'14,June,2004','2004-06-14 00:00:00','   What does it mean to do embarrassing things? To an English teacher itâ€™s making a grammar mistake and having the students correcting it. To a math teacher itâ€™s doing a simple mathematical calculation wrong. To a thief itâ€™s entering the wrong house and getting caught. To me: itâ€™s basically my life.    I donâ€™t think a day goes by without something embarrassing happening. Sometimes, Iâ€™ll admit I get embarrassed at things that are beyond my control: things that arenâ€™t even my fault. Like when actors act out an embarrassing scene on t.v. When someone says something not realizing that who theyâ€™re talking about is right behind them. Sometimes I get embarrassed to tell the waitress that she actually picked up my fork when she picked up the plate, and so I donâ€™t have a fork anymore.   Getting embarrassed when you find out that your big sister actually did something that seems more like a scene from a movie, only she did it live, is pretty normal -when she tells you how she picked up the microphone at this place, not realizing it was actually on, and was singing in front of everyone â€œRight here waiting for you.â€ Itâ€™s okay to get embarrassed for her and envious. Itâ€™s okay to be proud that the people said she had a good voice.   Getting embarrassed that you embarrassed your teacher is okay too. Like when my teacher was saying how he never learned things fast, but once he learned them, he never let them go. And to prove this, he started to name Jupiterâ€™s moons, only he added Titan, and so you raised your hand up in front of the class, and corrected him, telling him Titan is actually Saturnâ€™s moon. Of course, he agreed with you right away and corrected himself, but itâ€™s okay to be embarrassed that you put him in that position and that your class was staring at you as if you had just grown another nose. Well, why not? You just proved to them, for the umpteenth time, what a geek you were. (Notice: Please notice the verb: were. Were is the past tense. It means it happened in the past; I was a geek in the past.  Right? Right?! This is so embarrassing: why isnâ€™t anyone agreeing with me?!)  Getting embarrassed is annoying. People say, why are you embarrassed? Thereâ€™s nothing to be embarrassed about, but of course thereâ€™s always something to be embarrassed aboutâ€¦especially in my lifeâ€¦  But you know, the good thing is, Iâ€™m over being embarrassed about my last name. Itâ€™s not that I was really embarrassed. Itâ€™s just that, everyone always had to make a big deal about. â€œBanawan,â€ thatâ€™s quite an unusual last name.. â€œBanawan? Thatâ€™s your last name? Where are you originally from?â€ Like who the heck cares where Iâ€™m from. What are you going to go talk to my great ancestors about my last name? PleaZZZZeâ€”get a life!     	 Actually once, I told my friend that we were actually Chinese. I then pronounced my name, in such a way, that she believed me I said, â€œ Bana â€“wAnâ€. It was really funny. It turned out anyways, that I had relatives living in the Philippinesâ€¦to all you Banawans out there, that I donâ€™t know, I love you guys :D   Anyways, I couldnâ€™t really be embarrassed. There was a t.v. show practically dedicated to our last name. It was called â€œCamp Banawana.â€ I swear, there was a t.v. show called that. The only problem is no one seems to have heard of it, or seen an episode of it. Thatâ€™s my homework assignment for you guys. Go look it upâ€¦.believe me, it exists! And I used to like it tooâ€¦well why not? It practically had my last name.    You know back to the topic of embarrassment: when you realize that all this time youâ€™ve been singing on the bus, thinking that the girl behind you doesnâ€™t really talk, and she wonâ€™t mind, then seeing her SINGING on the graduation, is a time for embarrassment. But, anyways that girl left the school, so phew, things are okay. No oneâ€™s going to find out the truth about our voices.  Speaking of which, my math teacher makes fun of my voice. Itâ€™s because I have a low voice (not very girly) and because I have an American accent. He seems to find it very funny, but at the same time irritating. Anyways, Iâ€™m always making fun of him/and he makes fun of me, so thatâ€™s okay. Only I have this dream: I want to work for a news station. Read the news everyday, in that fake voice they have. (I tried it with my friend once and she said I did have the right voice.) Be all smiley everyday, reading the news. Then one day, for just one second, Iâ€™ll say, â€œAnd my voice is now famousâ€¦take that mr. former math teacher.â€ And then Iâ€™ll just continue on reading the news as if I hadnâ€™t said anything weird.   You know, I once saw this woman on the news station go like â€œI forgot all my lines. Sorry,â€! It was actually quite humorous: and now that I think about it, quite embarrassing.  I wonder if she ever had a teacher that made fun of her voiceâ€¦.  I have dreams of going back to all those bullies that used to bully, young, quite meâ€¦Is that normal? Hehe, having dreams where you go and take revenge against a kid that teased you/ bullied you when you were in kindergarten? Well, who cares if itâ€™s normalâ€¦., let me let you in on a little secret: Iâ€™m not aiming for normalcy, you see :O Yup, my life is gonna be different. Not that it hasnâ€™t been already, but wellâ€¦  Anyways, I see my twinster has been discussing bad habits. Let me tell you what  a really really annoying habit is. When someone goes on and on about nothing. Keeps on talking and talking and you just want them to shut up. You feel like saying, â€œ Hey, I donâ€™t need to catch a train. I actually need to stand in front of a train so I can dieâ€¦It would be less painful than listening to you go on and on.â€ Or you just want to tell the person that you have discovered they have exactly 34 freckles on their face, seventeen eyelashes on one side, a thicker eye-brow on the other, and that they are in desperate need of plastic surgery. You realized that because well, why they were so nicely discussing what their mother told them when they were 6 years old, you happened to have time to glance at their face. (Of course, you donâ€™t actually tell them this, or that fact that you were staring at their face, dreaming of ways to kill them and that you figured out 4 so far.) Nope, you just sit down, and listen to them talk and talk. I hate those people.   Anyways, I figure I talked a bit. Not enough for you? What you say you want more of my wisdom? Sorry youâ€™ll have to wait until next time!  Byebyez!  '),(9300,3594873,'13,June,2004','2004-06-13 00:00:00','   An unusual thought: in old English, ass meant donkeyâ€¦when was it coined to mean butt? And why? What is the connection? Thatâ€™s your homework, peoples. Put some effort and research it. After all, youâ€™re not always going to have answers just fall into your lap.    You know itâ€™s impossible to see beyond what you see. Cuz if you see something and look beyond it, well then you see what is beyond itâ€¦so youâ€™re still looking, if you know what I mean?  Which remind me: you know what I mean? What a stupid expression. How do I know if you know what I mean? You could very well say, you know, I know what you mean. And it would just go onâ€¦that expression is just because people are too lazy to continue on with their explanations and stuff. So to just end it, like their conclusion sentence, they just stick that in.   That also reminds meâ€¦seems Iâ€™m like a grandmother; so many memories. Anyways, when I was a kid, we had to write vocabulary paragraphs and other delightfully fun, not, paragraphs. Well, I enjoyed writing so Iâ€™d start writing. Everything would be so detailedâ€¦Iâ€™d have written so much but still be at the beginning. Nowhere near where I wanted to end, so Iâ€™d just add this, â€œWell, I got to go now. My hand hurtsâ€. Can you believe it? How stupid? How shameful? I could go on and on, but the truth is Iâ€™m kinda tired of typing all this timeâ€¦ so gotta go. oOops, looks like I still have that habit.  Some habits just donâ€™t die, now do they? Well, maybe because theyâ€™re called habitsâ€¦duh! If itâ€™s a habit, then itâ€™s something youâ€™ve been doing for a long time, and if youâ€™ve been doing it for a long time, then youâ€™re used to it. And if youâ€™re used to it, then its part of your nature. And if itâ€™s part of your nature, then itâ€™s hard to stop it. So you just keep on doing it, and thatâ€™s why it doesnâ€™t die.  But you know habits also change. When I was a kid, I didnâ€™t mind staring into mirrors. Now I do. In fact, I donâ€™t just mind. I hate it! I loathe it! I abhor it! Just trying to sketch you a picture (instead of paint you a picture). You know why I hate it? Cuz I just see my  reflection. If youâ€™re a smart aleck, then youâ€™re saying right now â€œReally, what did you expect to see? The zoo?!â€ No I did not expect to see the zoo. I expected to see my soul. Yes, that means Iâ€™m pretty happy with who I am. In fact, if you were to look at it, it means I think that I am a beautiful person inside and am quite surprised that it doesnâ€™t show on the outside. Now, over hereâ€¦Iâ€™m being arrogant yet modest, arenâ€™t I? Arrogant since Iâ€™m calling my soul beautiful, but modest since Iâ€™m saying Iâ€™m ugly. But you have no way of knowing if this is trueâ€¦so it doesnâ€™t matter to you. It matters to me thoughâ€¦mirrors and pictures can only capture your outwardly appearance. Theyâ€™re so shallow. And that is why they can not picture who you are.   You know itâ€™s funny how people seem to think because they know celebrities, then the celebrities must know them. We memorize their birthdays, their favorite expressionsâ€¦ everything about them. The whole time, though, these celebrities have no idea that we exist. That we actually walk the same earth that they do. Weâ€™re just not there, to them. Who can blame them? Itâ€™s not like our birthdays or anything about us, is published or taped so that they could be aware of our existence. Itâ€™s not like our own best friends remember our birthday anyway. When we see them, or get them to sign our autographsâ€¦when we wave at them, we feel like weâ€™re with a friend, whereas all they see is some people obsessed with them. Yup, fans!   Fans: its such a weird word. It can mean an appliance that helps keep you cool or some admirer. I see no connection whatâ€™s so ever. English really is strange.  You know, I used to get so angry whenever I heard some British word, like mum. Until, this year when I realized that it was Americans who copied the British, took their languageâ€¦and then changed it. (or else they just didnâ€™t copy it right)   But the word loo still annoys me. Just say it; loo. Hey, people who are called Leau or have that as a name are really unlucky.    Silent, like a ghost it creeps. With the stealth of a thief, and the confidence of a natural predator it sneaks behind the unsuspecting bird. Its eyes focus on its victim: the bull eye. For a second, just a second, the bird smells something. Something dangerous in the air; something ugly. It looks with its beady eyes all around, but it canâ€™t spot anything. But the feeling of something wrong, still lingers.  It finds itself in a dilemma. Does it leave its breakfast or stay? This bird has not eaten for a long time andâ€¦it didnâ€™t find anything risky. Maybe though, thatâ€™s why there were no other birds there. Maybe, it wasnâ€™t lucky and thatâ€™s why it found food, but because it was unlucky. Go or stay? Stay or go? It can not choose. But the choice is made for it. And a single white feather falls to the ground.  '),(9301,3594873,'20,June,2004','2004-06-20 00:00:00','   Amina: The hippopotamuses have come in gigantic herds to the great ear to hear. Yes, the herds have come to be heard.  Mariam: The ear stood under the black hair, slightly to the right of the monstrous nose, southwest of the beady eye, waiting for these herds that came to hear. It was an ear marked with superiority; canals that run through, and a tiny hole in the center of its lobe which was the hook of the earring.  Amina: Nah, the hippopotamuses did not go that superior ear, but rather to the ear on the opposite side. The left ear.   Mariam: Oops my daisies. I had quite forgotten that the hippopotamuses were exiled and disowned by the superior ear. You know, the scandal is hardly talked of. Even less the bumble bee that was a honeyetarian. Scandalous, I do say.    Amina: Oh no, my little chap. The bee was not a honeytarian. I think you meant that it was anti-honey. That little trumble bee refused to eat any of the glorious honey. I remember it clearly. They said, the bee said, â€œ I aiâ€™nt sweet and I donâ€™t want to eat something sweet.â€ Still that was nothing compared to the dandelion petals. Remember them? They refused to fly away. Said, â€œhome sweet home, weâ€™re never leaving you.â€ And they stayed. Dandelion petals arenâ€™t supposed to do that: theyâ€™re supposed to be brave you know. Them being called dande-LIONS. Theyâ€™re supposed to fly far away from home, but theseâ€¦ wereâ€¦.errâ€¦.humph. It was just so scandalous I canâ€™t talk about it.   Mariam: But donâ€™t forget the story of little bear. Legend has it, that in ancient times a baby bear was born on the first night of spring. As he grew older and older, his mother began to teach him how to hunt salmon. But a curious and unusual thing happened, the little bear could never catch one. Almost as if the pond was deserted and contained nothing even if it was overfilling with salmon. Each day his mother would have hope and would send him off with some tips, but each day he would come home empty handed. His mother, desperate to help him, asked an old berumim- old, wise bear. She went to the ultimate bERuMim, known only as FATHER bear and asked him to help her son. FATHER Bear told her, heâ€™d seen this little bear and the truth was this little bear had woven tiny leaves into a coat. so you can see this little bear, was not bare at all! And thatâ€™s whyâ€¦because he did not follow the law of the jungle (as Rudyard kipling would say) he was cursed to live on an empty bare stomach. Shameful I say. I canâ€™t really speak of it.   Amina: Oh, that was absolutely scandalous. Really, there are some things I know, I canâ€™t speak about them either. Horrible, horrible. Let me tell you, the story of the butterfly. Oh, what? You know it already. Yes, yes, I understand. I wonâ€™t bother the time to say it. Let me just tell them a bit. The butterfly was once a fly who could not fly. He was a walk so to say (because he couldnâ€™t fly.) One day, his mother, the poor old mamfly said, â€œSon, we have tried everything we can to make you fly. Letâ€™s just put some butter on you, and maybe youâ€™ll get slippery.â€ The little fly agreed and followed his mamflyâ€™s order. But when he put the butter on himself, he found that all the petals of the wind stuck to him and so he became beautiful. The shameful thing, now extraordinary, left his mother and flew off. Never talked to her again when he became pretty. Really, itâ€™s absolutely shocking I canâ€™t talk about it.   Mariam: Oh no! Thatâ€™s not the story at all. No way. Now you know, I donâ€™t like to talk about these things, you know I donâ€™t. You know, I never talk. I donâ€™t. Iâ€™m closed book when it comes to these things: but really, just to educate you: the story is, his father being a cheap man (as you know all men are) used Tut Tue Tut butter. That awful kind! And so when he wore the butter all the village butterflys made fun of him. Well the truth it, it became so hard for him that he just killed himself. Thatâ€™s the true story. I swear! But you know the juiciest, I mean most disgraceful story of all is Tweetyâ€™sâ€¦but well that I just canâ€™t open my mouth about it in public.   *Meet you at my house for some tea.   Amina: Oh dear, your story is completely wrong. Let me tell you the right version.  The butterfly was once a fly who could not fly. He was a walk so to say (because he couldnâ€™t fly.) One day, his mother, the poor old mamfly said, â€œSon, we have tried everything we can to make you fly. Letâ€™s just put some butter on you, and maybe youâ€™ll get slippery.â€ The little fly agreed and followed his mamflyâ€™s order. But when he put the butter on himself, he found that all the petals of the wind stuck to him and so he became beautiful. The shameful thing, now extraordinary, left his mother and flew off. Never talked to her again when he became pretty. Really, itâ€™s absolutely shocking I canâ€™t talk about it. (now whose the idiot who actually re-read it? I said the correct story in the beginning, duh!)  Okay time to leave so that I can go visit mariam.   posted!  '),(9302,3594873,'20,June,2004','2004-06-20 00:00:00','   OMG! Brainstorm. Thunderstorm. Lightning flashing around and bouncing off my skull. My brain is now afloat because of the storm. Hehe, I had another idea. You know after somebody breaks up, this personâ€™s friends usually end up saying sometime when they are trying to comfort him/ her: there are other fish at sea.  Excuse me, but that is about the stupidest expression one can say at a time like this! Do you want the person to just jump right out the window?  What one is ultimately saying is that the world is full of people: hundreds, thousands, millions, and billions of people. The chances of finding the perfect match, your soul mate, your dream man/ woman are statistically impossible. But hey, maybe youâ€™re luckyâ€¦after all some people actually do win the lottery. Well, now to find the person who captures your heart who fills it up with joy and then to have him/ her toss back into your face and spit it out- well that can be a problem. Now this hurt person has already experienced this/ at least feels like he/she has, and you are asking them- no telling them to go experience it again?! Are you CRAZY?!!! There are also sharks in the sea, arenâ€™t there?!! Theyâ€™ve already gotten their heart amputated, what more do you want?   Hey, hey; then thereâ€™s the other saying: Iâ€™ve got bigger fish to fry. And another other: Something fishy is going on. What the heck is the big deal about fish. Fish this and fish that. Ewwwwwwwww!! Istaghfir Allah, but I donâ€™t like fish.  WAIT A SECONDâ€¦ do sharks live in seas? Whatever, thatâ€™s besides the pointâ€¦you got what I was talking about. Actually, what I was writing since Iâ€™m not actually talking. Wait another second: not what Iâ€™m writingâ€¦what Iâ€™m typing.   But letâ€™s talk about handwriting. Iâ€™m sorry you guys. Iâ€™m truly sorry that you can not see our beautiful handwriting and that instead, youâ€™re stuck with this computer font. I mean, notice the word font. Itâ€™s got nothing pretty about it. But handwriting: the word handwriting: itâ€™s like a flowing artâ€¦h~a~n~d~w~r~i~t~i~n~g. Itâ€™s so beautiful. Of course, not everyone can pull it off. In fact, you guys just stop lying to yourselves. I want you guys to be truthful to yourselves. That way I can be truthful with you. Okay, I guess weâ€™re not ready for the full blown out truth: Iâ€™ll just say it mildly. Your handwriting sucks. I mean are you sure itâ€™s your hand that wrote that, and not another part of your anatomy? Seriously, do you need handwriting classes or what? Sheshesheshe (our feminine version of heheheh. Yeah, GirL PoWeR all thE WaY!) We donâ€™t have handwriting: we have calligraphy. Itâ€™s part of our blood.   Speaking of girl power: did we ever tell you guys that we have a part time job as ballerinas? We didnâ€™t? Well of course we didnâ€™t, because if we did then we would have been lying. Ha! Sadly we do not have part time jobs as ballerinas.   You know I had this really cool idea but I forgot it. Out it went- and flew into the sky, leaving me alone. All alone with the distant memory of having thought up of something remarkable that is soOoO ANNOYING!! Oh well. Bye (So abrupt yet so refreshing)  '),(9303,3594873,'03,July,2004','2004-07-03 00:00:00','   Amina: Can you imagine missing out on a whole season? Waking up one morning and finding out that you have missed out the whole of winter. How would it feel? I think it would be strange. But you know bears do it all the time. Just lie down and go to sleep for months. One morning itâ€™s winter time, and the next, it is spring.  Mariam: You know, that would be very freaky. Itâ€™s funny how fast time seems to travel. Itâ€™s scary to think that the second time passes by it is gone foreverâ€¦irretrievable. It is beyond impossible to ever bring back a moment. Thankfully, my friend, I am not a bear. I am more like a hyena. At least my laugh is, as my good friend pointed out.   Amina: Yes, indeed. A second is irretrievable. For those reading this blog,(unlike Salma) how many minutes have they lost? And yet, how much have they gained? A different perspective, a new thought to munch on, and a whole lot smarter, wouldnâ€™t you agree? I say, you do resemble a hyena, slightly.   Mariam: Ahhh! Ideasâ€¦they flow in my head. Some are born slowlyâ€¦gradually they begin to form into a clearer shape, until they mature  and are completely developed. From intangible to tangible they become. Others race around in my head as quickly as lightening flashing in a dark storm, lost forever. Still, others die before they are even born. Miscarried in my head. Wait a secondâ€¦if Iâ€™ve told you once, Iâ€™ve told you twice. I do not like the word miscarried. Excuse me?! Miscarriage sounds like itâ€™s the womanâ€™s faultâ€¦like she didnâ€™t carry the baby properly. I do not, do not like that word. That word, That word, I do not like. Does that remind you of somebody?  Amina: Do I remember that? That do I remember? Quite frankly, YESâ€¦ I do remember. That style of writing was that of the famous Dr. Seuss. Well, my ideas are like the things in the ocean. That is to say, some are coral reefs; all entangled up, and of various shapes, sizes, colors. Others are like tiny fishies swimming around. And even others are like the treasure chests buried deep within the ocean. I know that they are there, yet, I canâ€™t seem to remember where they are. I canâ€™t seem to open the treasure chest, until at last, when I least expect it, I remember and the chest is opened to reveal the hidden, glorious treasure: perhaps it is a fond memory, a creative idea (that at first I can not seem to form), or maybe itâ€™s just a thought that will make me smile when things arenâ€™t that great.   Mariam: How very amusing. But I must beg to differ. Unlike your ideas, the ocean is deepâ€¦rich. It is wide and almost infinite in a sense. Your ideas are shallow. ClichÃ©. I will allow you to say that you do have some good ideas but they are like a warm current that passes only in summer. Although it has the possibility of having a monumental effect, it doesnâ€™t because as quickly as it comesâ€¦it disappears.   Amina: well my dumpling (notice the term dump in it) if your ideas are like plants sprouting and maturing you must remember that they are also like them in the sense that they are stomped on, pulled (yours are mostly weeds) and those that actually mature into trees are cut down. Aaah, my friend, we never hear the stories of the trees that have been cut down. Instead, we go and write our own stories on paper which comes from the trees. And so the question remains. Can you paint with all the colors of the wind?  '),(9304,3594873,'28,June,2004','2004-06-28 00:00:00','   Today a most extraordinary thing happened. Before we explicate, let us stop and ponder the absurdity of the word extraordinary. The word is obviously made up of two words: extra, and ordinary. We all know that the word extra means more. If we substitute the word more in the word extraordinary we get: moreordinary. Moreordinary? I mean, whatâ€™s up with that? Doesnâ€™t more ordinary sound like the most ordinary instead of beyond ordinary. Maybe, itâ€™s just meâ€¦well you know what they say: Geniuses are never satisfied, to criticize and find fault is their habit. ïƒŸ if you say that you do know that saying than youâ€™re really really weird because I just made that up.  Actually once a couple of years ago this weird thing happened.  I swear itâ€™s true. See, I sorta wrote my friendâ€™s homework for her. We had to write short poems and so I kinda wrote one for her and she turned it in. you know what the teacher wrote back? Do not copy poems, and she wrote the title of some poem. I swear! Iâ€™m like, errrm, umm, okay so she cheated but she did not COPY a poem. I made it up!!!!!!!!!!!!!!!!  Anyways, what happened today was most unusual: quite extraordinary. It was to put it bluntly, something that has led me to believe in the goodness of people again. Ahhh, youâ€™re tired. You want us to just blurt it out. But no, one can never tell a story from the end. And that is why I will have to start at the beginning.   As most of you know: weâ€™ve entered this contest with our sis and her friend. Who ever loses the most weight gets a NICE GIFT. SO, weâ€™ve basically been starving ourselves for the past two days. Well, today we were surfing the Internet and we found out that: THATâ€™S BAD!! So we started eating again. But then, Salma was fasting. And so we had to eat AGAIN!! Thatâ€™s why after maghrib we decided to go walking. To lose weight.  So we started out and everything was going fine UNTIL: ahmed started complaining. Thankfully our mom came and picked us up. Then our family decided to go buy ice cream. AND thatâ€™s where trouble began.   After weâ€™d finished our ice creams, my mom slammed the car into the raseef/ side walk. The wheel drowned in the sand and to put it frankly: we were stuck! So we called up our daddyo and just stood there, looking like 4 damsels in distress. Now hereâ€™s the kewl part.  ALL THESE PEOPLE STOPPED TO HELP US!! I mean loads of people. Cars just kept stopping and looking and trying to figure out how to get the car out. Especially this one determined man. Insha Allah, heâ€™ll go to heaven. WEâ€™ll we were starting to get pretty desperate because nothing was working, but then finally: the car was on the street again!! Yay, alhamidulla.  *Even if any of you read this from Salmaâ€™s blog: I know ours is better!! It has to be! Hers is probably to short. Iâ€™m sure it is. :D ;D  * riddle: when is a door, not a door?    when it is ajar...hahhaa...hahahha..HEhahEhahesheshe Lol! *start laughing peoples*. i saw this riddle in a movie...called Bartok the Magnificent. yes, yes, it was a cartoon. sue me!  the name is banwan. banwan mariam and amina signing off. (think bond)  PoSted.  '),(9305,3594873,'09,July,2004','2004-07-09 00:00:00','   Having a boring day? here\'s our original list of things to do! have fun...  Go up to a teacher. Compliment him on his shoes and then ask him where he got them from. (When I did this I was talking to a woman teacher. After she told me the name of the store, I said Excellent. But that gives a clue to what you\'re about to say next) Next, ask to try on their shoes. Observe their face.  IF you\'re playing a basketball/ any sport game and let\'s just say things aren\'t going as good as planned and you want to have some fun : 1) bark like a dog 2) roar like a lion. *part 1, the roaring like a dog was inspired by an episode of \"america\'s funniest home videos..*  Make sure you bring a pack of gum with you the next time you go out with your friends. Hopefully, at least one of your friends will ask for a piece. Take out a notebook and write their name. Check your watch. Write down the time. Look at your friend and say: If I don\'t get the same kind of gum tomorrow, I\'m never going to give you gum again. And I\'m not going to be your friend. *credit for this one goes to our best friend. she came up with the idea, and we modified it.*  When you\'re at a mall go to the kids/ toddlers side. Choose a very small shirt. IF a sales person comes and tries to help you tell him/ her that you lost weight and were wondering if they had a smaller size.   oh here\'s another school one. when the teacher tells you they want your undivided attention or some cra* like that, stare at them without moving one muscle. STARE with all your heart. open your eyes as wide as you can. push your head forward, and place your hands cupped around your ears (as if you want to see and hear everything) then, as they get up to go on the board follow their every movement...that is, if they are writing something on the board move your head along with them. if they move their hand to touch their face or something, copy their movements.  I haven\'t tried this one out...maybe later. Go to the public library. Take a very tragic book. Sit down and laugh as hard as you can. Slap your knees. Look around and point to the book with a smile. (Romeo and Juliet is a good one)  Next time you\'re in a restaurant...go up to a complete stranger and look at them hurt. Say in a whiney voice: You ordered with out me?!  When you enter school, wave...the KING WAVE! SMile a fake smile, and say \"Autographs will be given at lunch. A nominal fee of 40 bucks will be required.\" (You know I kinda did this one. I stood up in our balcony in Egypt and began telling everybody standing outside \"My people, bow to me..you have....\" everybody thought I was a freak.)  When you get a new teacher this is how to be annoying. whenever they do something on the board, for example a problem in math, raise your hand and say, \"that\'s not the way, our teacher used to do it.\" etc. it really bugs them. *friend said this one before*  things to when you get a propsal for marriage:  I alaways thought that if a guy proposed to me (and I did love him)I\'d look at him and say \"OMG! You\'ve confused my intentions by a lot. I don\'t understand why you want to ruin our friendship. Please go...you\'ve done enough.\" Of course, I probably won\'t be able to do that becuase if somebody proposed to me who I really loved I\'d probably be crying.   I also thought I\'d make him answer about 100 questions that are completely irrelevant. Like \"When you brush your teeth, which tooth do you start with?\" or \"When you eat, do you eat from the left side of the plate or the right side of the plate?\" or \"Which colors look better together :purple and blue or orange and yellow?\" When he\'s finished and I\'m grading his answers I\'ll flinch and grimace every second. I\'ll even tear one up and say \"I\'ll pretend I didn\'t see that one.\" or I might just start crying and say \"Why are you doing this to me? Why?\" AT the end, I\'ll look at him and say you passed. Time for the second level and give him the next batch of 100 questions. of course if he actually really loves me, he\'ll go through it. HAHA, YEAH RIGHT!  things to do in a cinema:  okay, go to a movie you\'ve already gone to. Sit next to a stranger and tell them what\'s going to happen the whole time. Especially if its a horror movie. For example: somebody is right about to be killed. You say :Don\'t worry. He\'s not actually going to die. Just you wait and see. *this was actually done to a frined*  Also, laugh very loudly when its a tragic movie or pretend to cry when its a comedy. Scream peircing screams when anything remotely scary happens. IF somebody is about to kiss, look at whoever sitting next to you and say: Cover your eyes. It\'s inappropriate while you take out a magnifying glass.   Answer a mobile/ cell phone. Make sure your voice is very loud as you say : NO! You\'re not disturbing me. I\'m just watching a boring movie. It\'s really not interesting.\" Keep talking as loud as you can about nothing.  Make lots of crunching noises as you chew pop corn. Eat it like a pig. Stuff your head inside of the box...if somebody asks what you are doing. Say: It takes too much time to put my hand in the box and then bring it back to my mouth. Or say: I\'m hungry, Mama. I\'m hungry.  When a bad or evil character says their plan say :Boo!! Get off the stage. act like you take the movie personally and start saying mama jokes.  Keep on criticizing everything in the move: costumes, acting, everything. After a very touching scene, snicker and say: I could do tht in my sleep. What a sucky actress!  *My bro did this one. We were watching Around the World in 80 days with Jackie Chan so he got up and tried to mimick all the karate moves. Very EMBARRASSING but funny!*  don\'t worry. once we think of more, we\'ll add them!  From the creators of \"Str8 from the Asylum\" comes  \"a list of crazy things to do.\"  Now in theatres. (hahaha doesn\'t it kinda sound real?!)  catch yall l8er!  '),(9306,3594873,'12,July,2004','2004-07-12 00:00:00','   Mariam: The oil canvas lay on the desk, unfinished. Just resting for all of eternityâ€¦the oil colors sitting next to it. Peeking at it from their tired eyes and trying to relax their aching head. One of them, Mrs. Pink, had an awful headache. Her head ached, just as headaches do, and she felt terrible. But she had escaped Mr. Blueâ€™s fate. His was death, thankfully it was instant. The Fingers, as they were called, had broken his body in half.   Amina: Ah, the Fingers. Those that meant life and death: the stressed one colored too tightly and poof! You were gone. Or the distracted ones pushed you off a table, and pow: you were dead. Yes, the fingers were indeed lethal!   Mariam: The Fingers, my friend, are never written as the fingersâ€¦they are always written as The Fingers. I thought everybody knew that. You know you are quite right about them being lethal. They hold the seesaw that tiptoes between life and death for the colors. But they held an even more important occupation. They were the lines by which the creativity flew from The Brain to The Paper.   Amina: Dearie, tell us what the Fingers have brought to you. What have they led you to sketch? To paint? To draw? Oh my sweetie, I think you forgot to also write: the Fingers let the soul speak as they write the words of the heart. What have your Fingers written? Care to share with us what your nasty heart thinks up of? We so want to laugh. Go ahead and amuse us. :D   Mariam: Amuse you? I think you have mistaken the word amuse for the word inspire. But I donâ€™t mind telling you anyway. I was thinking where I would be, Insha Allah, in less than a week. Specifically about our voyage, our adventureâ€¦our trip to Egypt. But I hadnâ€™t gotten so far as to arriving there; Iâ€™d just gotten to the airplane. How wondrous it is to fly. To feel as one feels on a roller coaster. The delicious feeling that you have conquered the world. That you have reached the highest peak. Everything appears microscopic. Buildings and houses look like doll houses and cards from up above. Trees appear to be little grasses. People, like specks of moving dust. But what is so ironical is that fact that while you are looking down at everything and thinking of how small the whole world is, everybody under is straining their eyes. Trying to find the tiny airplane in the vast sky. And they are thinking of absolutely insignificant you appear so high above.   Amina: You surprise meâ€¦no, really, you really do: that was good! Real good. Now is the time for me to outwit you & write something wittier: something brilliant. Ever heard of an ice cube? Itâ€™s frozen: itâ€™s solid. But before it was solid, it was liquidâ€¦it was water. Then it frozeâ€¦once itâ€™s warm again, once it melts it becomes water again. It returns to the liquid phase. *I will now leave you to think of the irony. To see the wisdom in the ice cube. To feel the depth of philosophy from one little ice cube.*  What? How dare you? Me not know what Iâ€™m talking about. Well you can just go and kiss my cute a**. The reason why Iâ€™m not telling you the mystery of the ice cube is because I donâ€™t want to spoil it for youâ€¦ I want you to find it ïƒŸever heard that one before from a grownup? You haveâ€¦good, then you know what I really mean. :P!  Mariam: You can not write about an ice cube like that! What kind of storytelling is that? Let me begin for you. Once upon a time, there was a water bottle. A man came and poured some warm water into a tray, made up of tiny empty houses. The liquid flowed through, almost passed the brimâ€¦just touching the sky. And then the man placed the tray in the frigid north freezer. The air was frigid. Only shrieking icy wind ran through the freezer, making the water colder and colder. They would cry from how scared they were and slowly they turned into hard, ice cubes.   Then one the day, the man took the ice cubes and put them in a pitcher. A pitcher full of warm water. And do you know what happened? The ice cubes bounced a bit on the liquidy trampoline, until they finally fell downâ€¦down touched the bottom of the pitcher. The water began to surround the ice cubes, to warm them, and the ice cubes began to melt. Their hearts opened again because they were in such a warm and loving environment and the ice cubes turned into water, inside the water. That is a story, my dearie!   Amina:  nope, thatâ€™s not how you write it. Like this:   Flowing through the brooks, the lakes,  Landing in a tiny crystal cup, A spoon of water, a drop of life, Shivering in the glacial air, Frozen as a minuscule prism, Then winds of warmth, blowing away Away the chills, changing the prisms into Water, the quencher of the thirst,  That sprouts the plants, That flows in the brooks,  And in the frozen, ice skating lakes   Mariam: And the blog ends, hence.  For all of you out thereâ€¦Think of a crayon, an airplane, and an ice cube and the stories and characters that surround us every day!  '),(9307,3448540,'30,May,2004','2004-05-30 00:00:00','    Today, woke up pretty late..at 9 something i think cuz yesterday slept pretty late. Left my relative\'s place at about 1:30pm..hehe..cuz cousin\'s husband dont want to leave..watching lotr2. when boarding the bus back to college...an indian man striked up with a conversation with me..but..actually most of the time i didnt hear whats he is saying..he is going mumbo-jumbo in a very soft voice...heck..how am i going to hear whats he is saying..so as the conversation lasts..i just nodded and smiled like an idiot..haha..   '),(9308,3448540,'27,May,2004','2004-05-27 00:00:00','    its getting worse!!went through 2 hours of econ and 2 hours of socio with teary eyes..and did i look that bad that everyone asked me to see the doctor. Doctor..my last resort...haha..  class was ok..haha..but towars the end of socio class..the discussion suddenly went into...whats the age of miss chua..when she revealed of her age...pamela :\"oh my god!...\" haha...that must really hurts..  oh yeah..its raining!!! ye ye!!hm..so it finally rains..  Later at 2pm, have lan class...oh so boring..   '),(9309,3448540,'26,May,2004','2004-05-26 00:00:00','    arrggg....i cant breathe..hehe...well, well, what ya knows, the medicine dont work..im still having the cold eventhough i took the medicine for 2 times now...hehe..feel like taking it for the third time..but..afraid that i would die of overdose, better stick to the intructions...haha..hm..have been sneezing like hell...ah..looks very pathetic la..good thing mom is not here..or else sure drag me to the clinic.  After dinner, felt very hm..sticky..haha..so went to take a shower. After shower decided to do a lil bit washing...hey! no mosquitoes attack this time..haha..what a luck..well..usually there will be 4 or 5 mosquitoes attacking me..maybe my blood is too sweet or they like me very much.i dont know...haha..just feeling glad.  After studying..hehe..actually not much studying..just read a few interesting articles in the econ text. practise the violin!! haha..like to play the grade 2\'s pieces la..its easy..haha..just imagine that if i dont practise the scales..wonder what would the teacher do..poke me with the fiddle to death..haha..  hm..soon it would be ch\'s 18th birthday!!haha..still remember that the last time i went back to kulai she asked me to air-mailed her present to her..and to her request, i acted like i hadnt heard it..haha..well..dont even knows what she wants. Actually, i know la..but hehe..those things cost a \"fortune\" to buy..haha..buying animes in big quantities..theres also hanazawa lui la..hehe..but i dont know where to find him..hm..the college is so big..maybe i can find someone with the same personality as lui and drag him back to kulai with me...haha..sounds like kidnapping to me.  hm..boring la...so many class tomorrow..*Zzzzzzz*   '),(9310,3448540,'26,May,2004','2004-05-26 00:00:00','    During LAN class, i was paying only half of attention cuz being in the air-conditioned lecture hall and having cold the same time really made me uncomfortable. And we have quiz 2 today!! good thing i didnt leave in the middle of the class. The lecuturer was on his usual way...the bullet-train style of lecturing..thats very fast.  After Lan, went to buy my lunch and some medication for my throat and cold. The medicine for cold looks more like poison if u ask me. Having white, pink and orange granules in the capsules. Reminds me of the medicine that mom gave me the last time i had cold and i even asked mom if its poisonous..haha..of course after i took it..the next day, i was my own self again...ok, just hope that tomorrow i would be ok.   Friend asked if i want to go swimming, i said no cuz the only place i wanted to be is on my bed..not in the swimming pool although it would be fun to go swimming..haha..theres always next week.  ok, had my lunch, read the newspaper(havent finished it actually)..now im feeling sleepy..haha..and the weather is damned hot out there. Theres a lil bit of rain yesterday..but its not even enough to wet the whole floor outside my room.  Oh...very bored now..should i study now? sleep? hm....thats a very tough decision..have to think properly..haha..   '),(9311,3448540,'25,May,2004','2004-05-25 00:00:00','    Finally..finished doing the laundry, now..savouring my breakfast. Hm..shouldnt have wake up so late today..cuz if wake up early, can do a bit studying or just practise violin. *sigh* another hour to go b4 LAN class..so boring. Hm..just feeling bad right now, maybe because of the things that i said yesterday night to bf or its just my cold that are driving me crazy. Hey, i wont want him to say those things to me either..he must be really mad at me. Maybe its just my bad mood that made me wanted to drive him angry deliberately. I admit, thats very stupid of me..mood swing....:(   '),(9312,3448540,'25,May,2004','2004-05-25 00:00:00','    hey..i cant believe that im doing blogging..haha..    Hm...theres not many class today, the sociology lecturer was on mc(great!!), law lecture was as boring as ever and economics was ok. duh...im having a slight cold and sore throat..what a luck. :(     Im always eating today..dont know y..maybe bcuz im ill??god knows. im supposed to go swimming with my friends but i felt really tired..so hehe...i stayed in my room and went to sleep! And mainly because of the squash training yesterday..feeling sore all over..especially my right arm..like its temporarily paralysed.    At 6pm, went to have dinner with my best friend at the cafe, haha...more eating..    Tomorrow..the stupid LAN class at 12 noon but nvmind that..since i can sleep late today. Oh..just remembered, tomorrow have to do laundry..or else it will bcomes a small mount everest..   '),(9313,3663701,'19,July,2004','2004-07-19 00:00:00','   Four days and counting until we set off on our biking adventure.  There\'s so much to do this week between our normal working hours and last minute purchases and preparations.  We also have to practice packing the bikes as they will be on the trip, just to ensure that everything fits.  It\'s nuts...I can\'t believe I chose this as my holiday...it\'s gonna be dope.    In other news, there was a massive accident on the autobahn on Thursday of last week.  It happened directly in front of the exit to our house.  A tractor trailer carrying flammable goods crashed into the support pillar of a bridge and exploded - just like in the movies, man.  The heros of the day were two brave Dutchmen who pulled the truck drivers out of the vehicle before the whole thing blew sky high.  The bridge started to fall apart as a result of the explosion, so part of the autobahn was shut down for the entire weekend.  Traffic jams galore.     We were in Dortmund on the weekend, visiting the in-laws.  Nando\'s parents are great.  His mom was a home economics/family studies teacher and so whenever we go there she has elaborate cakes made for the Sunday afternoon \"coffee and cake\" session.  She is Betty Crocker and Martha Stewart rolled into one. Definitely the hostess with the mostess.      Europeans really are more cultured when compared to their uncouth North American counterparts.  They regularly eat dinner together as a family instead of everyone grabbing something on the go.  They never drink tea or coffee without a saucer, and they are much more skilled and efficient when it comes to handling cutlery.  Whenever I have dinner with the Laumanns family, I feel like the kid in the jungle book who has been raised by wolves and is reintroduced to civilized society when he\'s like ten years old or something.  My big clumsy paws are a running joke at the dinner table.  I am praised and congratulated if I actually get through a meal without dropping something on Nando\'s mom\'s lovely white table cloth.        Nando\'s nephew was also there this weekend.    urlLink     Leo and Lulu&nbsp; urlLink      This kid is the sweetest.  He\'s about 9 months old now.  It\'s really amazing to see how quickly babies grow.  I\'ve never been around one for any extended period of time until now.  I love being able to play with him and then give him back to mummy and daddy as soon as he starts getting cranky.  I like being \"Tante Jackie\".  I think I will keep it that way for a very long time.  Got loads to do so until next time...  Jaxxx  '),(9314,3663701,'14,July,2004','2004-07-14 00:00:00','   I find it slightly amusing that I\'ve seen more low-budget Canadian films on German TV than I ever have in Canada. It\'s a given that film making isn\'t exactly Canada\'s forte, but it confounds me that the people in charge of programming for the movie channels here would actually buy our crap. How do I know they are Candian productions, you ask? Well, by the usual giveaway of course...the National Film Board of Canada logo that rolls with the ending credits. Until now, I think I\'ve only ever seen the NFB logo on the films I was made to watch as a child in elementary school. (You know...the ones about bicycle safety and saying no to strangers). \"If someone you don\'t know offers you candy.  Say NO! Then GO! And TELL someone!\" I remember the early days of Jim Carey\'s acting career from those cinematic works of genius.  In other news...  I had my first visitors in Brausnchweig two weeks ago. It was really nice to see a familiar face. My main Aix-en-Provence partner in crime, Lee, and his main squeeze, Jon, came to visit for a few days of shopping, sight-seeing and reminiscing.      urlLink    Jon, Lee and Jax deciding where to have dinner&nbsp; urlLink     It started out a bit rocky when the boys\' flight was delayed by 5 hours. To make matters worse, when Nando and I arrived at the airport, traffic was backed up and there were  \"Polizei\"  everywhere! According to the officer that was checking each and every car, terminals A and B were closed off due to a bomb threat! Luckily our visitors were arriving in terminal C.   We spent the next two days shopping, bar hopping and wandering about the city. The weather was grey, cold and rainy (as usual) but I don\'t think these British boys minded. I guess you get used to that on the island.   On Friday afternoon, we took off for Nando\'s summer home at the North Sea for some fresh air, even worse weather and pure relaxation. We ate well, drank well, sat by a cozy fire and played Scrabble until the wee hours of the morning.    urlLink    Ferienhaus in Burhave&nbsp; urlLink     All in all, a great weekend. We have a reciprocal visit planned for September so that the boys can show me all around Newcastle.  Can\'t wait. This weekend it\'s off to Dortmund to spend some QT with Nando\'s folks. So, until next time...  J  '),(9315,3663701,'13,July,2004','2004-07-13 00:00:00','   Hello out there...  I\'ve decided to hop on the blog-wagon after my good friend Vickie began her very own blog and graciously decided to send it to me. Why am I creating a blog, you ask? Well, since I don\'t keep a journal, I think it\'s a fun way to document my thoughts for future retrospection. It also lets those I love (who are far, far away) know what\'s going on in my life. Basically, this is how it works: A couple times a week, I will post my rants, thoughts, feelings and observations on the web for all of you to view at your will. You can read and post comments, if you so desire. It\'s as simple as that.   So...let\'s begin shall we? I will start with some basics and an update on my \"current situation\".   First off, this is me.   urlLink    Jax - sitting in a pile of garbage at the SARS concert.&nbsp; urlLink     I am presently shacked up with my boyfriend, Nando  urlLink    Nando - car fanatic&nbsp; urlLink     in  urlLink Braunschweig, Germany.    urlLink    Scenes of Braunschweig&nbsp; urlLink     I am working as an English teacher at the local uni and at a private language school. (My German is shit so I can\'t really do much else). Nando is working on a Phd. in integrated vehicle dynamics (or something) and working for Volkswagen simultaneously.  Smart boy. We live in a cute one-bedroom apartment with our geezer of a landlord (Mr. VÃ¶pel), living above us.  We try daily to avoid running into him, as a quick \"hello, how are you\" soon turns into a half hour monologue about any trivial matter he can dream up.  I wish he\'d get some new material...I\'ve heard the same stories a hundred times.  At the moment, life is humming along smoothly.  But I must admit, I miss my friends and family tremendously. I am finding it very difficult to stay close and keep in touch. I hope that this little journal will help bridge the gap somewhat.  I am planning to be back in T.O over the Christmas holidays, but that\'s a long way off.  In the meantime, I am enjoying all that living in Europe has to offer.  The best thing is undoubtedly the close proximity of neighbouring countries.  Germany shares at least 7 or 8 borders with other countries and they are all no more than a six hour drive in any direction.  And with the insurgence of discount airlines, flying is fast and affordable.  In the past few months, I\'ve been skiing in Switzerland  urlLink    Near Zurich&nbsp; urlLink     and chillin\' in the Corsican sun  urlLink    the view from our room&nbsp; urlLink     Over the next week or two, Nando and I are preparing for our summer holidays.  We are planning a three-week bike trip from Oberwald, Switzerland to Aix-en-Provence, France.  We will mainly be following the Rhone River as it winds its way down from the Alps to the Mediterranean Sea.  As this will be my first long bike trip, I am not really sure what to expect other than some breath-taking scenery and a great adventure.  You can be sure you will hear all about it upon my return on the 16th of August.  Well, this concludes post #1.  Feel free to bookmark this link and check back often.   Luv Jax  '),(9316,3663701,'16,August,2004','2004-08-16 00:00:00','   I just returned from an amazing three weeks on the road and am excited to update you all on our adventures. I kept a journal while biking through Switzerland and France and will post one entry per day until all of them have been posted.   So without further ado...   Prelude: July 23rd - 24th   We arrived in Zurich late Friday evening. Flying with the bikes wasn\'t as difficult as we imagined. Marco met us at the airport and the three of us biked to his apartment together. We spent most of Saturday shopping and buying last minute items. Thankfully, I picked up some kick-ass rain gear that would prove VERY useful on our trip! Revital cooked up an amazing meal for us on the eve of our departure and then we relaxed over wine and great conversation until bedtime.   urlLink    At Marco and Revital\'s apartment in Zurich the evening before we set off on our trip.&nbsp; urlLink          Day 1: Sunday, July 26, 2004    Daily Stats: Kms - 39.76 Average speed - 23.16 km/h Time - 1:43:02 Max speed - 62.8 km/h Total Kms - 39.76   Today we started our journey from atop the Rhone Glacier (The Furkapass near Oberwald, Switzerland).     urlLink    Jax at the Rhone Glacier&nbsp; urlLink     Marco, Revital and Leo were kind enough to accompany us to the top of the glacier by minivan. (I felt sorry for the poor bastards I saw biking up that thing and was so thankful for the lift). We stopped for a lovely picnic near \"Andermatt\" where Leo had a few laughs on the swings.   urlLink    Leo on swing&nbsp; urlLink      Once at the top, we were able to actually go inside the glacier (for a small fee of course) via a cave-like passage that had been dug through for tourists.    urlLink    Entrance leading into the glacier cave&nbsp; urlLink      urlLink    Jax inside the Rhone Glacier&nbsp; urlLink      urlLink    Nando and ice sculpture inside the Rhone Glacier&nbsp; urlLink      urlLink    We look like we\'re having fun but we\'re actually freezing our asses off!&nbsp; urlLink     It was upsetting to see the photos outside the cave illustrating just how much the glacier had receded in recent years (probably as a result of global warming).   urlLink    Rhone Glacier - a wet, dripping mess&nbsp; urlLink     By the time we set off on our journey it was nearly 5pm.     urlLink    Ready to hit the road&nbsp; urlLink     We planned to make our first day on the road a short one. We said our goodbyes, did a final check on the bikes and set off down the side of the mountain at speeds of up to 63 km/h. It was absolutely exhilirating...a total rush!  The scenery in the Alps was just spectacular. I wanted to take everything in but full concentration was needed on the road as the turns were sometimes sharp and we had to be careful of cars and buses.   urlLink    The road that lies ahead - view from the top of the Furkapass&nbsp; urlLink     Once on less steep ground we found the bike path (Route #1) with little trouble and continued to follow the Rhone River. It was interesting to think that we had just begun our journey at the source of the Rhone and had actually seen it as nothing more than a trickle of water. By this point it was still pretty narrow, however it had changed from a clear blue to a cloudy jade colour.   urlLink    The Rhone becomes a real river&nbsp; urlLink     On our way, we ran into a short but extremely steep hill that posed somewhat of a problem. The back end of the bikes were so heavy that the front end came off the ground when we tried to ride the bikes up the hill. We were forced to push the bikes to the top (brutal). In the process, Nando gouged his shin on one of the peddals and began bleeding profusely...ouch. We took a quick rest at the top and then continued on toward Fiesch where we were planning to camp for the night.  The countryside was peaceful and we enjoyed the quiet solitude of the ride at dusk.   urlLink    Countryside in the Rhone Valley&nbsp; urlLink     We\'ve just arrived at our campsite,  Camp S\'Moosji  , in Fiesch. Since it is our one-year anniversary today, we are going to splurge on a romantic dinner at the Camp S\'Moosji restaurant. Woo!   urlLink    Camp S\'Moosji at last&nbsp; urlLink     Switzerland is so beautiful. I can\'t believe I\'m here. I wish I could share it with all my friends and family. Everyone needs to see this and photos can\'t possibly do it justice. Tomorrow will be a long difficult day. I hope it goes well.  Nighty-night...  Jax  '),(9317,3466592,'29,May,2004','2004-05-29 00:00:00','   urlLink    Well, I\'m back.  Instead of the picturesque scene you saw represented below, reality looked more like what you see above.  the weather varied between sprinkling, raining, and snowing (just those three), and the trail had\'t been maintained in a while.  lots of underbrush grown over the trail, logs to climb over or under, and snow drifts to hide the trail all added up to one soaked christopher by the time I set up my tent (at that point wherever I could find a flat spot big enough).  Chock one up to experience.  there were were two nuggets of wisdom validated on this trip:  Always take a change of clothes to leave in the car for the ride home; and, alway take a roll of (emergency) plastic tape.  The reason for the first one is obvious, the second one was validated this morning when I woke up to six inches of fresh snow that obscured my path the night before.  I would have been bummed if I hadn\'t marked my trail.  The Lord is good, and there will be other weedends for adventure (like next weekend). &nbsp; urlLink  '),(9318,3466592,'27,May,2004','2004-05-27 00:00:00','   Check this out! somebody with far too much time on their hands constructed a map of springfield (not oregon) by watching episodes of the simpsons and noting geographical markers.  Ever wondered what streets Homer uses to get to work?  Neither have I, but now you can know anyway!   urlLink Guide to Springfield USA  '),(9319,3466592,'26,May,2004','2004-05-26 00:00:00','   urlLink    Well, maybe this picture will give you a better idea of the sort of place I am going.  Maybe.&nbsp; urlLink  '),(9320,3466592,'26,May,2004','2004-05-26 00:00:00','   urlLink    I will be here for the weekend:  Jefferson Park.  I promise to bring back lots of pictures.  (Hopefully the threat of rain will drive away any crowds) (also, hopefully but less likely it will not rain)&nbsp; urlLink  '),(9321,3466592,'26,May,2004','2004-05-26 00:00:00','   This is great!  Thus, my first deposit: I think questions are good.  (Hey!  What do you expect for two cents?)  '),(9322,3466592,'25,May,2004','2004-05-25 00:00:00','   There is a new feature on my  urlLink website .  (Drumroll please....) Question of the week!  Every week I indend to have a new question open for discussion on an adjacent  urlLink blog .  Give us your 2c.  '),(9323,3466592,'24,May,2004','2004-05-24 00:00:00','   urlLink    and one of my neice, Riley Rose&nbsp; urlLink  '),(9324,3466592,'24,May,2004','2004-05-24 00:00:00','   urlLink    recent picture of my little bro.  pretty menacing, huh?&nbsp; urlLink  '),(9325,3466592,'24,May,2004','2004-05-24 00:00:00','   Okay, so I\'m new to this whole thing... but what a better place to stretch my wings than Chris\' blog.  :)  '),(9326,3466592,'24,May,2004','2004-05-24 00:00:00','   Here it is! the website, as you already know is off the ground (even if not fully constructed), and this blog will serve as the forum.  If you want to participate in this forum email me and I will send along an invitation.  Also, for those of you with newsreaders (go here if you don\'t have one but want one: http://www.rssreader.com/) this is my rss feed for this blog:  http://christopherlayton.blogspot.com/atom.xml  '),(9327,3466592,'02,June,2004','2004-06-02 00:00:00','   urlLink Gifford Pinchot National Forest - Campground Information   Hey Chris,   This is your test to see if the button on my google toolbar works.  And, oh look -- fires are allowed at the Forlorn Lakes!  Yippee!  '),(9328,3466592,'31,May,2004','2004-05-31 00:00:00','   Well, I\'ve been a part of this blog for a couple weeks, I though I\'d better quip in quick with my 1.5c.  (Sorry, I\'m Ã¼ber-poor...)  '),(9329,3466592,'11,June,2004','2004-06-11 00:00:00','   Titanic in 30 seconds, re-enacted by bunnies.  yes, you heard (or, in this case, read) that correctly.    urlLink Titanic in 30 seconds with bunnies. : \"Titanic in 30 seconds, re-enacted by bunnies.\"  '),(9330,3466592,'10,June,2004','2004-06-10 00:00:00','   urlLink    This just in:  Linus is moving to portland next week.  Why, you ask, should you care?  I don\'t really know.  (see the preceeding post)Linus is the ubergeek,  See  urlLink slashdot  article; urlLink  '),(9331,3466592,'08,June,2004','2004-06-08 00:00:00','   Okay, I laughed out loud - and (whew!) didn\'t score too highly (24.45759% - Geek) on the Geek test - how did you score? let me know.  urlLink The Geek Test  '),(9332,3466592,'06,June,2004','2004-06-06 00:00:00','   Hey, lookie here, a new question for a new week.  urlLink Question of the Week  '),(9333,3466592,'24,June,2004','2004-06-24 00:00:00','   [experiencing technical difficulties]  Or, at least I was.  If you\'ve wondered why posts have been lacking lately, it is because I finally chucked my Microsoft CD\'s out the window.  I am operating on a \"brand new\" computer:  the same hardware now running  urlLink Fedora Core 2  (formerly Red Hat Linux)  I don\'t have everything up and running yet (DVD video, TV tuner, some other things aren\'t configured yet) but I was pleasently suprised just how much worked \"out of the box\" (iPod, SATA RAID, dual head monitor/dual video cards, etc).  It is looking good and running notably faster than under XP.  Anyhow, I am up and running again.  '),(9334,3466592,'22,June,2004','2004-06-22 00:00:00','   Some food for thought...   urlLink Chicago Declaration of   Tim recently pointed out an article to me on social/political issues in the evangelical Church.  I found it thought-provoking and wanted to share. Good food for thought....   urlLink Main Article     urlLink Reader Responses  '),(9335,3466592,'30,June,2004','2004-06-30 00:00:00','   It look as if the Sailhamer course at Western this summer will be held July 12-16 10am-5pm.  The class will be about the book of  urlLink Genesis   Have you any vacation time to spare?! (having taken several classes from the Good Doctor, I say it is worth it.)  '),(9336,3466592,'29,June,2004','2004-06-29 00:00:00','   According to  urlLink JK Rowling\'s website , the Half-blood Prince is neither Harry nor Voldemort.  We know that Harry is a pure-blood, but if it\'s not Voldemort, then who?  My money\'s on Hagrid, who is two half-bloods; giant and wizard.  I know it sounds weird, but it sounds weird enough to make sense in Rowling\'s world.  Or it may be a new character...  And, oh man, did I just tip my hand as a geek...  Though I only scored 26.4% on the test...  '),(9337,3466592,'29,June,2004','2004-06-29 00:00:00','   Ah, yes; Let the madness officially begin:  the Title of the sixth Harry Potter book was released today by J.K. Rowling, though no word yet as to when to expect publication.  I, personally was a late-comer to the Harry Potter madness; but I did read the fifth book (Order of Phoenix) in one sitting on the day it was released, so I guess you could say I have caught the bug.  Oh, by the way the title is \"Harry Potter and the Half Blood Prince\"  Any speculation on who this \"Half Blood Prince\" is? urlLink The Next Harry Potter Book  '),(9338,3466592,'29,June,2004','2004-06-29 00:00:00','   this has surfaced in my life several times; it surfaced again.  For our mutual edification:   MY LORD GOD, I have no idea where I am going. I do not see the road ahead of me. I cannot know for certain where it will end. Nor do I really know myself, and the fact that I think that I am following your will does not mean that I am actually doing so. But I believe that the desire to please you does in fact please you. And I hope I have that desire in all that I am doing. I hope that I will never do anything apart from that desire. And I know that if I do this you will lead me by the right road though I may know nothing about it. Therefore will I trust you always though I may seem to be lost and in the shadow of death. I will not fear, for you are ever with me, and you will never leave me to face my perils alone. - Thomas Merton   For those who must have a bibliographical reference, here it is:  Merton, Thomas.  Thoughts in Solitude.  New York: Farrar, Straus, and Giroux, 1958. Pg. 83  '),(9339,3466592,'29,June,2004','2004-06-29 00:00:00','   For those of you who use a rss reader (go  urlLink here  or  urlLink here  if you need a free one) will be glad to see a \"read the bible in a year\" rss feed sponsered by the ESV people. urlLink ESV Bible RSS Feeds  they have verse a day at random, \"good\" verse a day, and several differing options on how one might read through the bible in a year all syndicated via rss 2.0.  Very cool stuff. A couple of other cool feeds:  the   urlLink weather forcast /report customized by you, ;  urlLink the word of the day , and the  urlLink feed for this Blog ; great stuff. also, though I don\'t use it, there is a news reader plug in for firefox ( urlLink aggreg8 ) as well as for Internet Explorer ( urlLink pluck )  '),(9340,3466592,'09,July,2004','2004-07-09 00:00:00','   FYI. I will be preaching in Colton this Sunday - no, not at Colton Community Church, but next door at Canyon Creek Community Church.  The Text is 2 Cor. 5:16ff.  \"For our sake he made hime to be sin who knew no sin, so that in him we might become the righteousness of God.\" (v21)  '),(9341,3466592,'07,July,2004','2004-07-07 00:00:00','   Ever wonder what the masses are thinking about?  How about the Blogging masses.  Now you can know, thanks to  urlLink blogdex .  This site is like a google news for weblogs.  Blogdex\'s webcrawler pays attention to what sites weblogs are being linked to, and then posts the most popular sites.  This is a pretty accurate way of \"polling\" what people are reading/paying attention to, since most blogs  (like this one) use links to provide context for their comments.  Blogdex also has a  urlLink rss feed .  '),(9342,3466592,'12,July,2004','2004-07-12 00:00:00','   Today the near miraculous happened, I enrolled for classes this fall at PSU.  Praise be to God (and, by the way Lord, I\'m not sure how this is getting payed for yet...).  So, you might ask (actually, you didn\'t so I asked for you) what am I taking?  PHL324 Intro to Formal Logic ENG490U Rhetoric ENG308U Top:  Literature and the Enviroment ENG340U Medieval Literature  BTW I have declared \"Liberal Studies\" as my major, which is a short-hand way of saying that I get to take pretty much what I want to take.  And there was much rejoicing.  '),(9343,3466592,'29,July,2004','2004-07-29 00:00:00','   okay guys, this is crazy - so crazy in fact that I thought it was a hoax until I started seeing the photographic proof that these people are actually giving away iPods.  So I signed up.  If you want a free one you should sign up too.  But first read  urlLink  this  and make sure to sign up using me as a  urlLink referral   urlLink freeiPods.com   Okay if that wasn\'t enough you can use the same company to get yerself a free 27\" Sony WEGA flatscreen TV  urlLink FreeFlatScreens  again use my  urlLink referral code   The Hitch:  1:  You have to sign up for a free offer (you know how those work - sign up and cancel before they start charging you AND 2:  You have to refer so many people (5-8) who each have to sign up for an offer. Is this a pain?  Yea kinda, but if it means getting a free iPod or TV out of the deal I can live with it.  '),(9344,3466592,'06,August,2004','2004-08-06 00:00:00','   urlLink    Talked to my Brother last night: \" Is that thermometer maxed out ?\" \"Yeah, it was about 128 that day.\" \"Good Lord.....\"  urlLink  '),(9345,3466592,'11,August,2004','2004-08-11 00:00:00','   urlLink    I went to the street of dreams (and, yes, dreamed) today, and found that the houses all (with the exception of the house pictured above) of the houses occupoed a fairly narrow design niche: NW.  Hmm.  I wonder why this is. &nbsp; urlLink  '),(9346,3466592,'10,August,2004','2004-08-10 00:00:00','   urlLink This week\'s sign of the Apocalypse.  '),(9347,3466592,'09,August,2004','2004-08-09 00:00:00','   urlLink    Whilst adventuring on the world wide web, I happened upon this  urlLink  cool site .  the skinny: aerial photography of anywhere in the united states.  case in point: my house (if you wonder which one it is my blue honda is parked on the street in front of the house).&nbsp; urlLink  '),(9348,4312209,'24,August,2004','2004-08-24 00:00:00','   AaLlowZ again..         hari ini cuacanya bagus ^^ aga panas lho! hari ini 21~! te denngg~!        tp sayangnyaaa.. i am so screwd 2dae! byk tugas yg berceceran..nggg.. pala aga pening2..    byk day dreaming.. n stressed things out.. ga bisa focus.. SAC inggris gw aja kacau..          Aiyoorr perlu kratingdeng2an nih.. ><>        besok gw bakal lembur deh! slesein smua yg blom slese!! (moga2 tercapai) akhir2 ini gw demen bgt tidur pagian gt, baca buku di ranjang trus tidur... z.Zz..Z.Zz..z~ *drollz     bangun chell!! *slap!* ngg.. iyah eniwei..        gw harus fokusss (spt kata mahkluk2 sejenis; \"susa aammett sih yg namanya blajar!!\") kmaren malem gw lg rajin. ambil buku maths method n rencananya blajar.. trus..    mata gw ngelirik ke buku yg td br gw pinjem di library.. ngg.. pikirnya gw bawa skalian deh.     kan gw blajar di dining room gitu.. jd buku2 kudu gw bawa ke atas.. buku methods pun gw buka, menganga..     ngg.. tanpa sadar tangan gw jalan ngambil buku itu,  the road to camelot.  (klo ga sala judulnya itu deh)    eits tanpa gw sadari study time nya uda abis n gw ga blajr dikit pun! ended dgn baca itu buku, yg gw lanjutkan denga cettink di kamar ma sodara.    trus sblom bobo gw jg baca lg~ hehehe    kacau nih gw >    hari ini spt kmaren.. gw masi pengen males2an.. emang males luh cel! --\"     well yg pasti.. kita harus jago2 ngatur waktu n duit deh.. iya nih soalnya juga aga bokek gwe!     pa lagi yah.. oia! gw maren bli kejuuuu ^^ cheese spread + mozarella + Dutch Gouda~ ^^     dimakan pake crackers gitu hohohoho! enaakk buanget..    n tak diragukan lagi! kolestrol gw pasti makin tinggi ajee! weleh gawadz ye..      hmm tapi hari ini gw memutuskan buat diet! mulai   diet lagi! ><>         tp dinnernya iniloh.. mana gw masak lg ntar (hospitality class ^^) hmm~ hari  ini grilled fish fillet, scallop potatoes, and orange creme caramel~! slurrpp... !!     trus gw bertekad buat exercise tiap hari! jalan2 turun belgrave kek (buat yg ga tau belgrave, itu thirdworld, yup dimana gw tinggal, jaunya amit2)      hmmm~ smoga tercapai deh!     trruusshh.. hari ini gw makin retarded aje.. duh lemes.. *slaappz* cheell buangun!     yesh yesh! oiaaa uda bell eh uda dulu ^^ ganti klas nih    wish me luck ye ma diet gw! hehe        cheerz     kyuubi/ 9tailz   '),(9349,4312209,'23,August,2004','2004-08-23 00:00:00','   My first post!  hehehe Postingan pertama begitu berharga.. slanjutnya terserah anda~'),(9350,4312209,'23,August,2004','2004-08-23 00:00:00','   A long time ago, somewhere in Europe, There is a big town called, Luxtenberg, it is where the King lives, it is the heart of most activities and the place where important people and the wise goes to. Gwerlexine is the second biggest city. It existed a day walking to the East of Luxtenberg, pass the Green temple, and another day through the Brivisile, little village that famous for the grape wine.      Ethan used to live in the Louisene village, in the south of Gwerlexine, before the village was flattened to the ground of the earth, by the Saverest\'s soldiers. The Louisene was a place where some wizards and witches used as a meeting point, because of some mystical athmosphere in the area. Ethan is an orphane now. He was eight when he left home. His parents where accused as wizards, and so they were assasinated. Ethan was saved, by a man. Somehow, this man, has mysterious and wild look in his eyes. He has the eyes that challange others. He is tall and average in size. This man grabbed Ethan into his black cloak. He was hooded, may be that\'s why no one saw them, that was what Ethan thought. He didn\'t say a word. His hand covered Ethans mouth, so that he won\'t scream. When the Louisene was burnt, he grabbed Ethan\'s father\'s necklace and put it around Ethan\'s neck, gave him a little smile and left whithin the shadow in the forest. Ethan was confused. He saw something flew over the air, towards him. It was a scrap of paper with notes.. it says :      East is where you suppose to go river will guide you so do the truth faithful and loyality are the keys find the door of wise and wisdom      Ethan stood for a while, and he decided to go to the Owlhom river. It\'s the river where he used to go to. Ethan is bright and full of curiousity, and adventurous spirit. He walks into the forest without fear, but curiousity. He found the Owlhom river. He looks down into the water. There is only his reflection. He looks in once again, and surprised because he saw another face looks back at him. That face is as clear as the river, with blue and silvery scales, whole black eyes and fish like face.      \"Dear boy, are you Ethan?\" Ethan is a little hesitate to answer,\".. Yes i am.. sir..\" \"I\'ve been told to guide you to the east city, Gwerlexine. now, my name is Edourise. Ethan, would be be kind enough to show me your necklace?\" So Ethan shows him the necklace. It took the shape of an small blade, with a gemstone. It was half blue and half orange, blended in between.  '),(9351,3356583,'31,May,2004','2004-05-31 00:00:00','   Salvaging the energy i\'ve got.I realized that i\'m becoming a ehbasan.The places i\'ve been.The people i meet.The bastards they whacked.The security.I have self discovered myself.The people i need.And those i don\'t.The ones that will help me.And those that will go.The new.And the old.Tapered or cool.Fighters or not.Sluts or bitches.I lost my hp.All my contacts are gone.Fuck.But its bothering me that much.Its sucking up my energy.Life is too fast of a pace for me to catch up.Will it slow down and wait for me? I now know the people i need in life.Those who are shittin behind my back.I now have no fear.No hesitation.Nothing to lose.Everything to gain.School?Wad is it to me?Cigs?I want them.Sex?Im giving her supersized.Drinks?Oh god i hope im not an alcoholic.Drugs?I don\'t do them.Music?Many of them are shit.I am so high.Life\'s taking its toll on me.Im waiting for the time to snap.I\'ll take u on.Even if u\'re van damme.These past weeks.I\'ve learned alot of new things.Met alot of new people.People who once i tot wouldn\'t ever be my friends.The people that i once looked down upon.They were in fact in many ways superior to me.It seems like i\'ve been wiped off from the faced of this world.Returned.Wiser smarter and stronger.Discpline once i tot was not for me.Now im living in it.To the way i wanna go.No one can stop me.Im not ready for tommorow.And i cant control thinking about yesterday.Where do i wanna be?Who do i wanna be with?In fact i dont really know.Dont u go away.Every  mistake.Seems so big now.As the days were longing.I think about the things caught in my mind.Damn my education.The chicks i got to meet.I don want to be there when u hit the ground.You cant make things right.Neither could i.I\'ve learned to appreciate every minute of my life.These past 2 weeks were really meningful to me.Get on the roller coaster.I\'ve been better known to beat the brain.Today is just a daydream.I\'ve seen people whoreally have repented in life.Singing rapsodies.The people who real fighters whoave nothing to lose.Im starting to evolve into them.Throwing it all away.The gals who are daring sick and psycotic.The people who have to spend years in research for their souls.Holding on.For dear life.Apart from all these.They are the people who you can really trust.Cos when all of the stars have faded away.They are the people who\'ll bring them back some day.The fire in my heart is out.I\'ve never really had a doubt.If you had the chance to save me would do it?Knowing i could never save u in return.if u did. All i\'ve to say is \"thankyou\"   '),(9352,3356583,'30,May,2004','2004-05-30 00:00:00','   WO! Saturday was exhausting.. Went to town.. Wif Cicak, Yasin and the gang... After doing crap and disturbing gals.. Headed down to hendrix... OMG! It was really happening there... But shit.. Before that we got screened by some stupid cops.. who\'d i\'d thrash in a 1 on 1 match.. their so gayish and skinny... the singapore police force must be crazy..CB! One of them ask us \"we all where one?\" haha.. wad a joke... Yasin said \"bue tuey!\" I was afraid he might bring us for urine  urlLink test .. Yea.. So they checked our id.. I gave Yan\'s id and wrote down abdullah\'s address... I had my ic already in my shoe.. Oh manz..! So after that screening shit.. Headed down to Hendrix. Reach there arnd 11pm. The \"dok guy\" told us that just finished a  urlLink raid .. Fuck it.. Who cares..Went in in pairs.. Each pair 1 jug.. So we had 3 jugs.. Drank outside the club.. And started a snorting game.. (sucking the vodka and acohol) into your nostrils.. It was cool... The other guys are dicks.. Some are already drunk and high .... It made my dick laughed. You know i\'m very sensitive if  urlLink people  act drunk or are not strong in drinking... They shouldn\'t be hanging out wif me.. Oh yea! Met zaki and the other gang.. Mamat and the other zaki..They were also going to hendrix.. Went in together with them..  Headed down to the dancefloor... Zaki and his gang was taking out their trance moves out of their pockets.. Met the national U-17 guys.. Shabbay and Rasul on the dancefloor... So one big group of us.. Formed a circle of 7 and dance together.. Not long lAter.. Another group of chicks, came in the middle of our circle and dance with us.. HAHA... I can see that Rasul was damn happy.. In fact all of us were... Danced with one of the girls in the circle.. She\'s hot.. Gotta this mix british and malay blood.. And she\'s 17,cool and hot..YEA... I don mind older gals.. The boys were like \"wah... take 1 alone ar.. selfish sia\" I gave them the surprised look..Her name was Diana.. And she\'s schooling in Tamasek Poly.. I told her i was 18.. And a slut wif unintellectual mind.. Oh boy , she belived me. All of us were just into the groove. Then , went out to 7-eleven awhile, cos Diana wanted to buy her lollipop.. And i was thristy and wanted my Gulp.So 2 of us went there.. and she insisted on paying for me.. I said it was ok i got cash.. But she said persistently.. So yea.. \"have it ur own way bitch.!\" Then she suggested to sit at this deserted alley behind hendrix cos i could not bring my gulp into the club.. yea.. the rest was history.. Alot of crazy and hottie things happened.. I wouldn\'t want to elaborate... After we finished , went back to hendrix.. It was already 2.30am.She was acting cute wif her lollipop.. By then there were like 6 chicks in the circle.. Haha.. Me n cicak entered the middle. Dragged Diana too.. OMg! We were sandwiched... Like i was surrounded by 6 gals.. all around me.. Fuck.. The other fact was that... When gals are dancing and sweating.. They stink... Unbearable..Shabbay then join us..He was a crappy bastard.. Keep asking me for cigs.. Dragged this other malay gal wif me out and danced wif her for a while.. She was quite chio also..She had blonde  urlLink hair .. Then suddenly arnd 3.15 am , the stopped the  urlLink music .. and cid and ssb came in.. They just scanned the dancefloor.. Perhaps they were searching for someone in particular.. Then in like 2 mins.. they left... and the song started back..Danced with the boys till closing.. They were all crazy.. And Mamat. Gosh! He\'s such a pervert.. Closing  urlLink time .. 4 am.. The guys asked me to ton... Nah..I was too tired.. Met Alip outside Hendrix.. he just came back from town.. Chatted wif him .. Then we took NR3  urlLink home  together..They was 3 chicks with him.. One of them was his gf.. BED! I LOVE MY BED!   '),(9353,3356583,'28,May,2004','2004-05-28 00:00:00','   Honey you are a rock. Upon which I stand. And I come here to talk. I hope you understand. The green eyes. You\'re the spotlight. Shines upon you. And how could? Anybody deny you. I came here with a loan. And it feels so much lighter, Now that I met you. Honey you should know. That I could never go on without you. Green eyes. Honey you are the sea. Upon which I float. And I came here to talk. I think you should know. The green eyes. You\'re the one that I wanted to find. And anyone who tried to deny you,  Must be out of their mind. Love those green eyes.   '),(9354,3356583,'24,May,2004','2004-05-24 00:00:00','   Today i went to sch .. hoping that it would be fun.. but fuck.. all my friends din come... yesterday already i was absent.. so i went to the general office, got a form and went home.. its so tiring.. school sucks . the same faces .. OMG i cant stand it.. the teacher\'s voice.. its hurts my mind.. Im gonna slack at home today wif my cigs.. and my com.. im not going training.. Lame huh? Fuck school.. n fuck those wankers who go there consistently to lick the teachers butt... Maybe at night later im going out.. im meeting someone... fuck this blogging shit.. im sick of it too. You\'re totally fucked..  Nv trust ME ( cunning man ) and shawn ( sly man )  Cos when we\'re together.. All we gotta say is that u\'re part of the masterplan.. We don give a shit.. So swallow ur pride.. and hide your shame gal... Like the croc that preys on the rabbit.. All im gonna say is that u\'re PREYED. By the two best predators in the businees.. hahaha.. im gonna sit here.. and laughed the whole day abt that..! Take care...   '),(9355,3356583,'23,May,2004','2004-05-23 00:00:00','   It was raving on satuday. Nothing more to be said. =)   '),(9356,3356583,'22,May,2004','2004-05-22 00:00:00','   To start off.. im feeling hot now.. im sweating like a pig. Phew.. dono today go where? Maybe i wanna go town.. buy some stuff.. then at nite head to clubs.. Its a weekly routine.. I just realised the SAF cheerleaders are hot.. Oooh wee.. i mean real hot..  They ain\'t those typical malay gals i see on the streets.. I was shocked to see Candy in the squad.. She\'s my pri sch friend.. Haix.. she looks preetier now though.. she too was shocked to see me.. HAHA she is crazy.. laughs for no reason.. nonsense sia.. Her face look like Shanning.. Only prettier and the nose not like shanning.. LOLX! The malay gerls there are also cool.. HAHA.. During s-league matches the u16 and u18 team will like have to watch the match behind the cheerleaders and almost all of us will not be concentrating on the match.. But the cheers were nonsense.. I cannot tahan..  \"Eh ah eh ah tak takot tak takot.\" \"Warriors gina pabeto.\" haha wad nonsense.. I wanna go shopping later.. I wanna buy a belt and a topshop tee.. Yesterday was nonsense.. help to play for shawn\'s team.. sec 3 guys match.. waste my time.. my team win 4-1 la.. HAHA.. the opponents which is hanif\'s team.. play until halftime then go home sia.. wtf.. they say we all play rough.. HAHA.. they all not good means not good la.. Cb.. No kick to me la.. My team oso all lousy sia.. Oni me eddie and shawn know how to play.. the rest cannot make it.. haha.. scored 1 nonsense goal.. then help shawn dummy another.. Signed season contract for SAF already.. . Im sick of training.. Im a casualty of exhaustion. I must take steriods.. LOLX! AHH!.. I wanna go bathe.. im meeting someone in town.. Set sail to paradise.!   '),(9357,3356583,'21,May,2004','2004-05-21 00:00:00','   Hey! I heard you ran away. But where\'re you gonna run. You got such a pretty face.  And the world\'s a small small place. Cos when you\'re running on your own. You know u whane like rolling stone. Cos the stone will find it\'s place. And when you hear the space. Just crawl and never run. And when your work is done. And then you see the sun. So sing what you want. I\'m a man on a mission. No time to love. No time to be wasting. I want know if you feel the boredom. All the things you do that make you feel undone. Feel undone. Blues!  I heard you caught the blues. What was the sin? Have you got out of bed? Have you get out of there? You know a head space that you don\'t. And will you cannot and you don\'t. A fortune has been made. On all the illness they made. And then we\'ve gotta pay. For the drugs to take it away. And then you\'re gonna find.  There\'s nothing with your mind. That you don\'t already know. That you ain\'t already seen.  Do you live in a dream? Save your song. Don\'t you cause trouble. Man on a mission.  Is a man in a bubble. Play your cards before your dice.  You\'ll get a piece of lemon.  And the piece of that vibe.  There\'s no time to be wasting.  By then... you\'ll wake up. And smile. Cos you know you\'re mine.   '),(9358,3356583,'20,May,2004','2004-05-20 00:00:00','   I just ended my training. It was so tiring.. We did gym workouts. My thigh muscles are like contracting and aching now. Just reached home. I am like dying of exhaustion now.. Went home wif zul,erwan and hairul just now.. Met Nadira at cck station. She was going home from her friend\'s house. I was like sweating and so tired.. I haven\'t seen her for ages.. Yea i tink i miss her.. She\'s one good friend..She was so spunky , full of energy and very vibrant. She gave me a huge huge hug. Till i could feel her tits pressing against my chest. OMG! I wasn\'t smelling pretty good after my training though. Haha.. We chatted for awhile before i headed home. I said to her you\'re looking even prettier now.. And she replied.. Oh ya.. Of course.. As usual.. HAHA.. wad the hell...Ego babe. Aniwae we had a good laugh and its nice to see her after such a long time.. I realised i\'ve been neglecting my friends now.. Im too busy with training , exams and my personal stuff.. It was sad. I got to know that my friend is in the Intensive Care Unit. He got attacked by 7 ppl wearing masks and each armed with a parang. He was caught off guard. Im not suppose to disclose any names now.. Aniwae my fren was like sitting in a group of friends at west coast when suddenly 7 guys with masks and parangs started attacking them. 2 of them got to run.. However he and his other friend was attacked brutally.His fren got a deep cut on the head. This friend of mine , he lost 3 fingers. Got a punctured stomach. I hope he survives.. I just saw him like last month and he was ok and cool. I just hope he\'ll do well.. This is shit. Didn\'t know the attackers from where.Oh boy.. I have nothing more to say.I\'ll pray for him.   '),(9359,3356583,'17,May,2004','2004-05-17 00:00:00','   Today was my last paper.. Phy/Chem MCQ qns. It was quite easy though.. i guess maybe i can get at least a b3 for the paper. So after the exam i went home.. It was boring today.. Doing my usual routine. I still haf to go for SAF training in like 20 mins time.. Argh ppl thought that soccer training was suppose to be fun. But when u haf like 5 trainings and a league on Sat. It gets tiring.Im gonna listen to songs now.. Chilling for another 20 mins.. then off i go. Cyaz !   '),(9360,3356583,'17,May,2004','2004-05-17 00:00:00','   urlLink    \" I\'m the best singer in the world and I\'m in the biggest band in the Universe.\" -Liam \"I don\'t like being famous...It\'s crap.\" -Liam \"You could say I\'m a genius I suppose, quite good-looking as well actually.\" -Noel I\'ll look out to the crowd tonight and there\'ll be all teenage girls. I can\'t write for teenage girls! That\'s f**king Robbie Williams. - Noel If anyone don\'t like oasis they\'re shit! - Noel Jesus was 5ft 6ins apparently, so i\'m bigger than God - Noel  \"Liam\'s leaving the band this very minute. Liam\'s ALWAYS leaving the band. It\'s raining today so he\'s leaving the band \'cos it\'s raining. (Imitating Liam) Why is it raining? ****! That\'s it! I\'ve had enough of this, I\'m going! I\'m sitting there saying (cheerfully) See you then, bye! He threatened to leave in Minneapolis because I wouldn\'t go to a pub with him!\" - Noel on Liam \"We are the biggest band in Britain of all time, ever. The funny thing is, that f***ing mouthing off three years ago about how we were gonna be the biggest band in the world, we actually went and done it.\" - Noel I only know five people who haven\'t taken drugs and that\'s my mam, me gran, me father-in-law, me mother-in-law, and my new-born baby. Everyone else I know, be they lawyers, doctors, or what have you, they\'ve all got their heads down once in a while. Nothing wrong with it, all part of growing up. - Noel In Newcastle some geezer jumped up on the stage and punched me in the eye. It went off pretty bad; there was a riot at the gig. You always get that in England. There\'s always someone in the front giving you the finger. Did they queue up for six hours just to come down and call me a wanker? I don\'t understand that mentality. It\'s very strange. - Noel   '),(9361,3356583,'16,May,2004','2004-05-16 00:00:00','   Cant wake up 2dae.. super duperly tired. yesterday i slept arnd 1.30. talk to abdullah on the phone.. we haf so many things to talk about.We both we like super engrossed in the conversation.Aniwae today was my chemistry and literature paper.. went to school wif azhar today.He told me about the Singapore vs Selangor match yesterday.He went to the national stadium.Actually i was supposed to go wif my SAF team but i was too tired..Jus woke up from my sleep.The previous day i went clubbing.Very tired.Coach Richard is going to start his nagging wif me.So i was azhar in the train.As the train arrived there was these 2 girls like smiling at us and like talking to each other and smiling at the same time.WTF!? Early morning do this kind of thing.. SMLJ.Knn the other fact that those 2 girls also \"cannot make it\" la.Cb. It was so irritating..We were standing infront of them and like one of them was like listening to the discman and it was so loud. Even i could hear the techno.Techno? Get a life gal. She was like nodding her head to the beat.Trying to get our attention or something and then smiling to her friend.Cannot tahan already i can azhar moved to the other carriage. Wad nonsense right? So i had my lit and chemistry paper. The lit essay sucked the crap out of me. I wroted cock,dick and wadeva else u call it.I had to write like 2 full pages of essay without leaving line or else Mrs Chng would scream in my ear.Then during recess i met abdullah and relax wif anwar.After that had my chem paper which i had no idea how to do.What the heck is molar masses?Relative atomic mass?Ionic/covalent bonding?I didn\'t even study.After the exam i went straight home with mathew. I decided not to wait for abdullah as his paper would end in like another 40 mins.Bastard rite me?So i alighted at cck mrt.. As i was walking in the shelter to go home there was this girl who is dressed in like all black.I glance at her.\"Woo.Chio.Got standard.\" She was also full of attitude. She was having this expressionless face.Turn me on!She had long dark hair and dark eyes..gothix.haha.Suprisingly, I caught her in the act looking at me when i glanced . So i just locked my eyes onto her and i noticed she had hers locked into me too.Didn\'t say a word . We walked past locking eyes onto one another.Reached home.I\'m on my com now.Gonna play my cs la.Taste my headshot baby.Today not going SAF training la.. Too tired.. i wanna slack wif my com.Sayonara.   '),(9362,3356583,'16,May,2004','2004-05-16 00:00:00','   WOah... Just came back from hendrix today... reach home bout 5am.. Damn it. my cs no form... so yea.. im writing this piece of shit right now.. Yesterday was happening .. haha... went to hendrix wif aidil... there was a lot of controversy before we headed there...First we wanted to go to JAMS the club , then aidil called his frens and they said they were going to ChinaBlack.So he thought we should go to ChinaBlack.But oh no.As i was using ARAB\'S id. i feared that i wasn\'t allowed to go in.. Aidil just came from ktv wif his friends.. He was steaming and slightly drunk.. His nonsense was starting.. Everything we laughed and talked about he wanted to ^5.. As a good actor that i really am i played along ... FUCK U! It was getting irritating.I just came back from the Police Academy. My team ,SAF was vsing Home United.We lost 1-0. Argh ,i was not playing.. i took a 2mth long break from SAF until the coach called me to play back.. i was i my outside clothes.. The game was nice. SAF strikers and wingers cant seem to keep up wif the Feroz&Acak.Co. they were constant , they were calm,to cut it short,they defended well.There was high tension in the match.. Hard tackles. and quarrels that was stop by the referee.No yellows or reds suprisingly.ARGH. fuck la.. lose to HOME they were the unbeaten team in the league now..  So HendriX it was. Reached there at 10.15pm bought cigs.. Aidil was a stubborn ass.. I told him i wanted Marlboro or L&M but he wanted Sampoerna.FUCK.accomodating ppl was by best trait.Sampoerna it was even though i hated it. We went in Hendrix.. there was free flow.. i don consider my self an alcoholic..but when there\'s free drinks to swallow in.Count me in.I finished 9 glasses of vodka in like 15mins.. haha free wad? We met our friends there..Lots of them.. Shaking hands in weird ways.. Argh get over it boys.. im fuckin lazy to do tat kind of crap.We sat in a group ,smoked , talked cocked. Checking the chicks in da club.This friend of ours,a gal, was like standin in between us at the bar counter and was like talking to aidil.. i looked at her tits.how pervetic.haha when she walked away i and aidil had a good laugh over it.High ^5 was wad he wanted. There were alot of chicks in the club that night.Though , not many of them were dat beautiful..basically, not my type.I think the bartender was getting irritated with us.. Every 2 mins there was a glass to be refilled. 12AM. We hit the dancefloor.. there were four of us..1 of them brought his gal.. So there was ravers in the house. Me aidil and his fren.. i don bother to remember his name. We dance with each other in the 1st 15mins.. glancing at the chicks as well..OOO.The dancefloor was getting packed.. So indirectly, the chicks were pouring in. Then i spotted this gal.. cute sexy cool.. CHeeky.Me.She was like dancing with her 3 friends.. She glanced at me.. and i smiled.Mastering the art of smiling, i had been training.Be a good bitch tonight YEA?!So she was like beside me and we danced together.. Wrapped my arms around her waist. She had hers around my neck.Looked her in the eyes.\"YOU LITTLE CUTE BITCH!\" was the thought in my heart. Aidil got this hot chick that was staring at him all the way. That was wad he claimed though.EEew... the chick was fat and ugly.. at least to me she was.. aidil must be blind.arr. FUCK IT!Dance dance dance.. the boys decided to get a breather outside the club.. As much as i don want to go, i had to follow them and leave the girl on her own. i don not want to be seen as a bastard.who pangse friends. After that we went back in,headed to the dancefloor . this time i was taking the lead.. there were plenty of guys in the club as well.Looks like the chicks were all taken.That was OK. Dancing with the other boys in da club hu were also my come-go friends. *JACKPOT* this gal was staring at me.. i know but did not give any signs.. just continued with my dance.. She was cute though.. Malay gals..Haix sian.ANd btw, the chinese gals in Hendrix are hot and i mean real HOT!ANiwaez Aidil continued with his fat bitch.how disgusting.Being the ego dude i\'ve always been , i jus make those expressionless face and dance.. By now she was beside me.The other guys or \"crocs\" thats wad i called them.. were like trying very hard to dance with her.. She was full of attitude.. HAHA.. yea.. that was wad i like.. LET THE CHALLENGE BEGIN! She was like turning away from each of them.. I was laughing in my heart.. by now, i was slightly drunk as well.. She was dancing infront of me.. and i was like orrh wadeva..then suddenly SONG CHANGE. She grabbed me by by Shag band and like we started dancing . haha i was taken back..\"Girl u gotta realise that some things are to big to fit in you.\"AHAH the ego man wins again.The song was \"come to my hotel\" and i was like singing \"baby do u want to come to my hotel\" she was like smiling.. OH WTF! A sluty bitch deprived of sex.She had big tits with matching ass.. Haix. yawnx. After a while , i was getting tired.. and bored.Took out my hp and looked at the time. It was nearing 2.30.Pretended that i was making a call and went out of the club.. tired manz. Now they aren\'t that many ehbasans in Hendrix.. and the guy who came up to me and said \"Aku dah relak ar\" didn\'t turn up. I was boyishly glad.Went back in, dance with this other chick as well. Touchy - touchy nothing special.. didn\'t spotted Aidil..Dance till closing time..As i was walking out of the club this gal suddenly started dancing rubbing against my ass.. fuck la whore.. closing already. Met aidil outside the club. Said i was headin home.. He was walking with the fat gal.. I for once thought that they were checking in a hotel.. Said Goodbye.. Took a long bus ride home... Slept my heart out in the bus. Lucky to be awake when it reached cck. Thank god.. Home at last.. Sweet Dreamz.Take care and don\'t take drugs.   '),(9363,3356583,'16,May,2004','2004-05-16 00:00:00','   urlLink    Richard Aschcroft. A man who lived his dream.   '),(9364,3356583,'27,June,2004','2004-06-27 00:00:00','   Damn the things caught in my mind.I need more time.Tmr is school.Im so not ready.Im the last person in spore right now to be ready for school.Haven\'t even packed my bags.Hate those same people.Fuck.Boring.Im staying up all night today.Tmr sch.. hu cares.Im gonna watch Czech vs Denmark.Im putting $100 wif mervin on the Czechs.Wif Nedved in midfield and Baros in top form.Yea.. They\'ll take this 1 quite comfortably.. Nizam\'s prediction: Czech to win :: 2-1. Jon Dahl Tomasson will probably get 1 goal for Denmark.. He\'s in good form.I\'m tired.Stayed up all night yesterday.Changed my no. But a few bitches stil msg me.. Wtf?!Who tell 1.. Must be my gay friends la.. Fuckers.Haha.. I promised myself that i\'m not gonna stead wif a malay girl anymore..Haha.Nizam u nonsense la.. Now i haf trouble keeping up wif my promise.If I had a gun wif 2 bullets now.. I\'ll shoot her once.And another on her head to make sure she really fucking dies.Nonsense la.Euro 2004 winner : Holland ( my prediction ).Now my chest damn big man..Fucking looks like breasts.HAHA.Been doing alot of push ups lately.. Im fitter than ever now.. Everyone thinks so..Erm sianz... Sat gotta lan training at cine\'s egames wif my gang.Long time since i saw nadira.. called me jus now.. wanted to meet.. jus too lazy to entertain her..Thinks she misses me.. haha.Told her tmr la.. asked her come my house. Rather stay at home.. Put my lying abilities to the test.. Said that going to haf dinner later wif my family..Everyone believes me... hahahhahaha... Only few ppl i known can tell that im lying.. And they oso kana lied b4 by me.. Nizam ur so cool.I tink i\'ll be sleeping tomorrow in class..Heard that there was class bbq last week.. Vanessa asked me to go.. Fucking lazy . I don do this kind of crappy stuff.. Wow.! Many guys are in a bad mood now.. They lost alot of money on France.. and also England.. Bad Moods? Try showing that to me.. I\'ll break ur fucking face..Met the guy who\'s nose i break and teeth i made disappered at cck mrt just now.. He and his friends.. haha.. 4 ppl.. they 4 i alone la.. come.AHAH. Stared at them all they way while waiting for someone.. They didnt even dared a glance.. He was still bo-ge. LOL.My sis was a dick.. she was using the com for damn long .. I tell her to faster .. she told me to shut up.WAHHAHAH!! daring ar.. Tell her don make me break ur fucking face.. She shut up.. SHe was in fact doing some stupid project on book reviews or smth.. Fuck school la.. I really don wanna go sia.. Mrs Oei. Her voice i cannot tahan.Tan puey eng the face i cannot tahan...Hope he misses me.Argh! Fucked up la... Got alot things to do la.. Fuck la.. my hp i want to low profile now ppl keep msging n calling.. fuck i changed no. oso no use.... Take care u.  Bitches with wings.They fly.   '),(9365,3356583,'26,June,2004','2004-06-26 00:00:00','   Went to town yesterday.. Wif abudullah.. Met quz , adi and zul at topman..Long time since i met them... we got some topman shirts.. Grabbed 5 shirts totalling $350+ for 60 bucks... Cool huh? AHAH... Aniwaez.. After dat we all went to lido to catch a movie.. \"The Punisher.\" It was NC-16 so the aunty asked us.. \"U all 16 anot?\" I said ahah... Aunty .. I oni 7 years old can watch... She was damn pissed.. Zul gave her a malaysian 10 cent coin.. hahah... then we just walked away.. we passed quz the movie money.. he lost $10 of it.. LOL! so he had to pay the xtra manz.. Ate long john silvers.. Its so oily.. Fuck.. ANiwaez the movie was nice.. Its started at 9.30 and ended at 11.45... Went home after that.. It was crazy.. Beckham.. I think he modelled better than playing soccer.. And those tattos.. Wtf he wanna act cool.. ANd now.. WAnt to copy my botak hairstyle.. Its been a long time since i kicked a soccer ball.. Fuck hu cares? My skills must be gone by now..Portugal are in.. Reached home. Played on my com.. Then watch france vs greece at 2.35am.. France was  dissapointing..Without Viera , zizou was finding it hard.. playing with Dacourt.. HEnry only gave a par perfomance.. While trezequet barely had any chance in front of goal.. But Greece won 1-0. By a goal from their striker.. In fact they didnt deserve to win.. Neither did france.. WHo played shit football.. Silves was horrible.. Zizou was losing his touch.. Like i said.. NO viera , NO france. So today is denmark sweden... im gonna buy $500 X 2.. wif mervin.. if lose dont pay la...Normal...hahah.. wad can he do.. lose $1000 le.. WIn $1500.I\'ll take that chance.. Let\'s see if he\'s stupid enogh to play wif me.. He siao de gina .. No kick la.. KA betta... AHHA.. Fuck those gangsta shit.. I realised now.. That ehbasans arent wearing tapered anymore.. they\'re starting to wear cool.. Nuisance to the society.. Embarassment to mankind.. Now i realised.. They are 2 ppl my age that will drop me in a 1 on 1 fight.. Jepon and Pakar.. Zul said im fitter now.. HAH! Im always fit..Ego me. Im planning to make another tatoo.. Ljunberg\'s panther.. Cool? Still ain\'t confirmed.. HAHA yesterday walking at HMV we saw this packet on the floor... Then i kicked it.. AHAH .. It was a condom.. I shouted CONDOM! All rushed to picked it up.. Quz almost fell... WTf.. lolx.. ALl like fighting for it.. At last quz got it.. ITs durex strawberry... I dont get it y the govt. wants to import flavoured condoms.. WHen oral sex is banned in spore.. Moreover.. where got people suck cock use condom 1.. Nonsense la..I phobia i have abt going to cine is that i will always meet those troubled teens outside cheers.. and somehow.. one of them will sure approach me and try to strike up a conversation.. FUCK! what ever u think.. im not a troubled teen like you.If possible.. i dont want to have anything to do with u guys...Im to lazy to talk to u guys.. WHos pregnant? Hu pia hu? WANt to go chiong?\"recently i whack...\" AIYA!! U ALL DON LIKE REAL LA>>> CB talk about fight.... Try wif me la.. nonsense fuckers.. sibei irritating..DOnt people haf anything betta to do... Im a changed man now.. At least i think i am...                             Straight pimp.                        WAN SUEY! STOP SMOKING B4 U GET SMOKED GAL!   '),(9366,3356583,'23,June,2004','2004-06-23 00:00:00','   Just finished my lunch. Im home alone.I like being along at home.There\'s so much things to do.My body is aching today.Did arnd 300 push ups yesterday night.Slept at 2 am.My eyes are bulging red now.Went to CRC toa payoh yesterday wif my gang.They are complete assholes.Literally.They went there with the \"lose,lose la\" attitude.Wtf?!Met a few ppl i know when i was on my way there.The train ride was so long from cck to tp.Shunned all those ppl.Im jus to lazy to say \"hi\".Hope they didn\'t see me.I took cover beside 1 auntie wif a huge umbrella.Haha.Niwae I\'m not playing for SAF already.So reached there around 4. Match wif GBR mix and then wif Discipline.1.6 sucks.I cant play.I makes me seemed like a noob manz.HAHA.Aniwae being the star player in the team i need not worry cos my mates will be paying for me.Cool huh?They played nonsense.For once i even agree that i was playing shit down there.The factual team are pros manz.I cant seemed to get my sensitivity right.Its been a long time since i went to town.I\'ll be heading there this friday wif abdullah.Long time too since i saw him.Perhaps. Around 2 weeks plus.HAAH.Wad can i say. Im a changed man.Im not looking forward for school to reopen.Im sick of the same faces the same voice the same place.Its cracking me up.Maybe im going to hendrix on saturday.Im totally banned from that place i guess.Hhaha.And btw im botak now.The reason remains untold.Fuck i still got training wif the gang at crc on sat. I dont feel like going.Wasting my time.But the CRC counter gal was cute though.She had this jap look.AHHA. Who cares?fuck her.Abdullah said he got me this long sleeve topman shirt.Thanks alot bro.Im short of cash now manz.. Im gonna ask my mom for $1000 i guess.I feel like asking someone out.But im jus to lazy.Dates are a thing of the past.Now the chicks ask u out yea?FUck that bullshit.Im gonna shop on friday and catch a movie wif abd.Wad movie is there.?im still waiting for spiderman 2.Take care and don\'t take drugs.  .KA.WAN SUEY !   '),(9367,3356583,'26,July,2004','2004-07-26 00:00:00','   God.. Im just to lazy to go to school todayâ€¦ Pretended to have a stomach ache.. Early morning screaming contest with my momâ€¦ Its just too irritating.Im jus gonna relax and watch some vcds..Hrm.. Im damn hungry.. Maybe Iâ€™ll fry some chicken n beef burgers laterâ€¦ This is heaven.Alone at home. Im going to Malaysia on Oct 1st.Im heading to Johor for the Asian Youth Freestyle Fighting Meet 2004. Yea.. Iâ€™ll be participating.. Hopefully getting 1st.. Haha 1st my assâ€¦ more like 1st to get my nose broken. Im joining the u-16 category.. Thereâ€™s also the u18 categoryâ€¦Basically u can go there and fight any style any modifications tooâ€¦ Im using my trusted Street fightâ€¦ There surely be those 16 year olds thai kick boxers and those Japanese ninjasâ€¦ Hrm.. Do I stand a chance.. Only GOD knows? Anyway Iâ€™ll be going there wif faiz.. Heâ€™ll be my ring man.. Hahaha. Im really scared of those kick boxers ninjas and silat pplâ€¦ Wooâ€¦ Jus hope I don get hurt real bad.. Hopefully the 1st and 2nd round Iâ€™ll meet those Singaporean karate and teakwondo gaysâ€¦ Not frm Korea or Japan can alreadyâ€¦. Singapore?? No kick laâ€¦Those Malaysian boys confirm are going to use silatâ€¦ Scary stuffâ€¦If I meet a ninja in the 1st rnd â€¦ Iâ€™m going to let him have a walkover manzâ€¦ Im fucking seriousâ€¦ I donâ€™t want some dude flying ova my head and doing some crazy techniques to kick my ass upside down.. Hopefully 1st few rnds meet those Judo or Jujitsu ppl laâ€¦ at least I tink I can take it.. I tink those spore National karate and teakwood federation are going to get beaten like fuckâ€¦ People go there wif coach I go there with a friend. I tink theyâ€™ll view us as making a fucking jokeâ€¦haha.. Who cares la.. I jus wanna go there and see how I par wif the best fighters in Asia.. If world.. No need say already laâ€¦My worries are doused. At least there are paramedics thereâ€¦Woohoo thank god. So wish me luckâ€¦ Im too lazy to train.. I just wanna go there .. Fight nonsense â€¦ Get my ass kicked.. And go home. =) Im short of cigs..!   '),(9368,3356583,'25,July,2004','2004-07-25 00:00:00','   Exhaustion.   &nbsp;  Today is sunday. Im jus gonna stay at home.. Woke up at 3pm jus now... Went to Hendrix the night b4... Hrmm... Went with a big bunch and met a big bunch there too... Went wif Aidil n his friend nadira jessica laura 2 of nadirahs friend, ijah and nasha and another 2 of her frens which names i din bothered to know.. u\'ll understand y.. yea..... Woo... Met the national under 17 boys there shabbay n rasul.. my gay fren yasin and zaki and the gang....   It was fun though... Party ended at 6am.. OMg i was practically dead by then... Thank you nadirah for bringing ur frens ova... ijah was dead drop gorgeous she had this attitude and style.. Punk rock concept.&nbsp;and nasha.. Hot.. wad can i say... ahhaha.... But fuiyo... ijah has a bf... hrmm... too bad... I and ijah was playing hard to get in da club... hahah every guy neared her she turned away and every gal neared me i danced away... nadirah is a puss.... Fucking MATHUNTER! lolx... one day i must really snap at that cold hard bitch.. I tink she isnt suitable for hendrix... Lolx.. Girls being girls... hrmm.. i wonder y they like to backstab each other and talk behind their backs... I jus don get that mentally... Her friends were like telling me all her bad stuff... and yea.. i jus listened on... But of course.. after the conv ended. ... a paranoid nadira went up to me and asked .. did they say anything abt me.... And&nbsp;&nbsp; all i said was \"no la.. u tink u so pretty ar ppl wanna tok abt u\" obviously tt was a lie... i made it seemed like a joke so as she would not suspect any serious changes in my face when im lying... The things i heard were kinda bad... from the looks of it.. hrmm.. i tink i do believe them... The secret im gonna keep till i die... Ijah asked me to go out to 7-11 buy drink... OK! fuck ... nadira n her other fren followed... DAMN IT! bought big gulp and treated the gals a lollipop each.. CANDYMAN!Sat beside the river awhile.. before the nadirah and her other fren wanted to go back in... they called ijah.. Cheeky me... Used reverse psycohlogy and said.. u go in 1st la....&nbsp; haha... then she told nadirah she wanted to finish the drink 1st... which was actually already finishED!LOLX.. 1 on 1 time wif ijah... jus wad i wanted...Wee! Nasha.. Omg.. she was a power dancer... tt gal.. really noes how to shake her booty... Cant keep up wif her manz...Ijah actually commented on my dance... SHe said it was cool.. Im always cool gal...P.I.M.P   Shabbay.. u bloody ass.. He apparently got a month long break from national u17 training coz he bluffed the coach that he went to india and will be back on 3rd august... LOLX... He was a dick as usual... He went to this couple which was kissing and stuck out his hang and made the WOoo WOOoo stuff... lolx... the guy felt akward and stopped.. Jus cant stop laughing... ANd he was also a shameless pimp who just went up to girls and grabbed them by their hips wanting to dance wif them... got pushed by them half of the time.. ahahha. Rule: Nv go unless ure 100% confident.  Went home at 6am.. First train.. went home wif nad ijah nasha ijah\'s bf and her 2 other frens.. There\'s a gay that actually smiled at me in the guys toilet.. Being the no nonsense attitude kinda guy... i went up to him .. Stuck out my middle finger... (cautious to duck if he throwed a punch) and GO \"FUCK U, U FUCKING GAY\" .the glare.He was probably gayyified by then...   Took train... Home .... going wif ijah in the next next sat.... Yipee Yi Yo.   &nbsp;   '),(9369,3356583,'15,July,2004','2004-07-15 00:00:00','   Missed school yesterday.. i was just to fucking lazy... At least school was ok.. i slept in class all the way till recess... And eventually woke up after that.. Its time i do some serious study.. Heard from my classmates that tan eng wee was damn angry with me yesterday for missing school and also focus.. I hate foucs.. FUCK! i find no reason for going... Why study in school when u can go home and have a nice time studying... Aniwae there are too many distractions for me at home to study... I hardly open my books at home..   Waited for abd after school .. but he probably went home without me.. Wtf.. So walked to clemmeti stadium .. wAnted to take 188 home.. met bifen.. She was also taking 188 home.. so took 188 home with her home... Omg.. She has soo much to tok to me abt.. Didn\'t realise that... she was happily talking to me abt ppl she hate and stuff... Talking abt backstabbing and hypocrites.. Hrmm... I just playing along... too lazy for this kind of stuff.. Ahah.. She told me not trust some ppl n stuff.. She took my number ... and said she will call me to cont. our conv.. Talked all the way fropm clementi to toh guan.. does she ever get tired... LOL!She told me some plastic bag joke that was damn funny.. nonsense sia.. Advise me to carry some real untransparent plastic bag. for some purpose...cannot tahan manz..haha.. aniwaez she was really easy to talk to.. I just faked my enthusiam.. .. i am always damn tired after school... Reached come... I just wanna be alone for now...Im doing some Crappy stuff .   '),(9370,3356583,'12,July,2004','2004-07-12 00:00:00','   Im rotting here.. Jus came back from watching Mean Girls today... At least Monday was fruitful today.. The show was damn nice.. Abt girls toking abt each other behind their backs.. And undoubtly , Lindsay Lohan was drop dead gorgeous... It was a show abt these 4 teenage girls each having huge egos... How pathetic... Backstabbing each other and stuff... Each thinking they were that fucking pretty... LOL!Nice show... And they\'re called Plastics.  School was boring...It sucks so bad.(0_o)   '),(9371,3356583,'10,July,2004','2004-07-10 00:00:00','   Woah!!!... Im in a good mood today.... Wanna know y... cos i jus came back frm crc .... and guess wad... we won team discipline... 13-11. hahaha... it feels so cool to beat my ex-team... Played dd2.. Im so tired now... feel like dying... ANiwaeZ ... went to town yesterday with that crazy bitch... As usual came 45 mins late that made her real maD!!! HaahHAAH.. Purposely 1 actually.WAnted to watched Mean Girls... Woo.. I tink Lindsay Lohan is cute...ahhah... but all the tixs we sold out.. both at cine and lido... OMG! and that crazy bitch actualy blamed me for coming late... fuck her. WeNT to THis fashion ... Bitch returns wanna buy some clothes... Im such a failure with woman\'s fashion... SHe picked up a dress... SHowed it to me from a distance and gave me the stupid look... I nooded my head and smilEd... (telling her the dress was nice..) She StuCk out her tongue and put the hanger back on.. WtF!? why do u have to ask me for suggestions when u noe i suck at this.. All the clothes i said nice was disliked by her.. and all the clothes she liked was dislike by me...I tink the sales girls are getting angry with us.. we kept walking and keep dropping the hangers.. haha how irritating.. Talked abt nonsense stuff... I hate these things... But not wanting to disappoint.. I PLAYED ALONG>>! OH GOD .. i hope she realises im faking it... Life was terrible...Wanted to watch mean girls so badly.. Going out wif her seems a waste of time... hhaha...Still she wanted to wait for the 11.45pm show at lido for mean girls.. Too LAzy...WAnted to train back wif me... DOWAN! take 190 home...10.30pm. END . REACH HOME.Placebonified.   '),(9372,3356583,'06,July,2004','2004-07-06 00:00:00','   Shit.Im starting to blog with a 4 letter word now.It was a close call. Greece won? Yea.. Watched that fucked up match.It was a fluke la.Cmon la...Mr Scolari is a fucking stubborn cunt.He refused to start wif Nuno Gomez.Instead starting with Pauleta , who in the whole tournament.. hasn\'t a goal to his tally.Figo was trying to be 1 cheeky bastard.. Trying to weave his way thru the Greek Wall.And a Greek corner resulted in a goal from Caristeas.He such a fit guy.Costhina was day dreaming.. And OMG! Ricardo .. Only GOD knows y he came way out of his goal line trying to punch a ball he couldn\'t possibly have reached... Weird.. Mentally i don understand that..  Spiderman2.. Watched it on the opening day..Nice show manz... Who would ever have suspected a nerd to be a superhero..Yea.. Kirsten Dunst is hot...So is Kristin Kreuk in Smallville.. And Keira Knightley in Pirates of the Carribean.. I think gals starting with \"K\" is naturally hot..ahah.. Nonsense me...   Hrm.. Keep going out often these days.. Sat afternoon was purr-fect.Went to crc toa payoh wif my team.. Alright la.. Played against Nic\'s team... Won aztec 24-0 and dd2 20-4. Cool manz.. At least the boys now r performing... Hrm.. i played well too i guessed.. Eric cut his hair.. He looked like a dick now.. his 1.6 not bad...Yea.. Had my quick dinner wif Roy at 5.30pm // Ciaoz home.. Went out again at 7pm... Came home next morning..I wanna sleep... Have you ever think .. What if u sleep and nv woke up..? Freaks me.  Hrm... Im getting my grip now.. Im back.. I hope my hair can grow faster.. Any solutions? I should probably try yun nam hair care...Or should cut my armpit hair and paste it on my head... Erm.. would any1 notice tt?What i need now... erm... nothing i guess... I think i\'ve got wad i wanted right here..Heard that Britney is getting married? Haha... She\'s gonna be the ala J\'lo.It\'s gonna be a joke.. Fucking boring.. My msn is damn irritating... Playing my cs got xialan ppl keep msging...Noisy sia.. Fuck them la.. im tired la.. wanna slack..... Cyaz..   '),(9373,3356583,'01,August,2004','2004-08-01 00:00:00','   They say Ohhh Weee....  Was at ChinaBlack yesterday...... Nice place... The atmosphere , lightings , the music , the chicks , the dancefloor  was all perfect... The place was soo damn big and there was plenty of space to move around... The dance floor was getting packed by 12+ .. There were damn alot of ppl inside...Went wif ekki , ian , eugene , denise , huffy and whitegal... faiz could\'nt go in ... He din haf a fake id... I was damn lucky... Got qned by the bouncer abt my ic.. I told him i\'ve lost in together wif my wallet .. He was taking a real close look on yan\'s bas pass... haha... Then i used reverse psychology and said... hey uncle i regular here le... u nv see me b4 mehz.... which was totally untrue... then he said ... Ah na.. OKOK next time bring ur ic....$18 bucks... Cheap stuff.. ian and eug booked this VIP sofa set for us in the club... WOo... we were treated like vips... being served drinks... They opened to bottles of Smirnoff and 4 jugs of liquidators... costing $250... They\'re fucking filthy rich.... Had our wrists copped wif the word V.I.P. oo cool......  Apparently, denise vomitted after having 3 glasses.... How pathetic.... Damn alot sia she vomit.. Fucking disgusting... And ian was out at his 6th glass.... OOO soo poor thing... Huffy saw his gf while dancing on the dancefloor... Small world aint it... HAHA... it was damn funny... As his gf was dancing wif some1 else... And they were like \"hey i din noe u came\" Crap bullshit.So that was it.. ian said to me \"c\'mon la drink more...\" i was laughing in my heart... u\'re asking me to come on.. and he said that to ekki too... I n ekki considered the rest WEAK. h3h3 ekki was tapping me and giggling... I n ekki gulped everything down ... WOo cool stuf.. Nothing like drinking some expensive liquor...  Went to the dance floor after that... me ian ekki n eug danced on the podium... LOL.. ASSES. Got this mixblood gal... Shook wif her on the podium... Almost fell off.... she was moving damn fast... Paiseh sia... Ekki got this reborned haired chinese chick... He was acting cool and smiling to me.. Acting gay at the same time...There were lots of those office workers and uni grads.. poly students too... Suprisingly, there were also ehbasans gals in ChinaBlack... How power... HAHAHA... DAnced with this ehbasan for a short while.. FUck.. she was wearing a tank top and some damn short jeans dat aint even cover her thighs... Overall... hot la..Good kisser .Whitegal was dancing this malay dude wif a snowcap... There were also some American Pimps and Italian Dicks.. Apparently they wanted to strike up a conv wif me.. and dragged me into their group which was filled with lots of cute 20 something chicks... COUNT ME IN DUDE. They were taking pictures and stuff... There was this particular pic when i and the 2 other guys stuck our tongues into 1 girl\'s mouth.... haah... nonsense sia... SHAGGED  Alot of stupid stuff happened after that..... ChinaBlack\'s cool...   '),(9374,4252081,'16,August,2004','2004-08-16 00:00:00','   â€˜Origins of species by means of natural selection; or, the preservation of favoured races in the struggle for lifeâ€™, maybe a bit of a mouthful, but it is an important book for all those thinkers out there. At the age of 22, Charles Darwin went on a world tour with a few of his naturalist buddies (not to be confused with the now popular naturist buddies) to far off lands with coral reefs and dragons, where he met strange natives with bones in their noses and poison tipped arrows. They wandered around uncharted seas in a little boat for 5 years, picking up plant samples, finding insects with thousands of little legs and analyzing multicoloured birds before the hungry cabin boys could put them in the pot. Despite suffering badly from sea sickness and giving up the life of luxury that a rich Victorian is used to, Darwin soldiered on with his project, and returned to England quite the world savvy explorer. He spent the rest of his life recovering from some mysterious Amazonic STD, studying the bottles of pickled lizards and dried butterflies that he had collected on his travels. Darwin also had a few theories on life; how it was created, how it was evolving and perhaps how it might end. He shared his thoughts with friends over a pot roast pheasant one night, which thought he might have a strong case.  â€˜You should write a book Charlie, youâ€™re bloody clever so you areâ€™.  So he wrote a book, that even he thought was good, so he wrote another, and another. The titles became longer and stranger, and the discussion deeper and darker with the summing up of a lifeâ€™s work being published in his penultimate year entitled â€˜The formation of vegetable mould, through the actions of worms, with observations on their habitsâ€™. Iâ€™m yet to get to that one, but it will certainly be next, just after I finish reading the National Enquirer that is next to my bed.  It is important to keep up with world affairs, and following the British dramas of the Hutton enquiry and the American Martha Stewart scandal, as well as keeping up with Jordanâ€™s latest flame and the conspiracy theories of Janet Jacksonâ€™s right nipple clamp is nearly a full job time in itself. Reading the Enquirer at night helps with the American gossip, but isnâ€™t much use when I want to know about John-two-jags-Prescottâ€™s latest punch up or Wayne Rooneyâ€™s new love nest. So, when some friends came over from old blighty, they were instructed to bring over â€˜Britainâ€™s biggest selling newspaperâ€™, the SUN. As the American audience may not know the Sun is famous for itâ€™s outrageous headlines (such as KUNTZ! After master striker Stefan Kuntz scored the goal that put England out of the Euro â€™96 championship) but also for the upstanding young ladies who get their baps out for the nation every morning on page 3.  â€˜Well fitâ€™, was Jonnyâ€™s comment on Zoe, 22, from London. And with 32-24-34 C measurements I could appreciate what he was talking about.  We were all at a bar in the right part of town later that day, at a poetry slam, and a hilarious middle aged woman got on to the stage and tickled us pink with a long winded story about a little blue box from Tiffanyâ€™s, â€˜I donâ€™t know what it is about her, she isnâ€™t that fit, but there is something about her..â€™ Jonny, she is over 50, she has let herself go a bit too far, Iâ€™m not with you on this one. Jonny, Anna and Lou stayed for a week, and every nightspot we went to Jonny would find beauty, where sadly sometimes it was hard to find. Maybe those Chicagoan beer goggles are that bit stronger than the ones back home. â€˜That girl looks like Joley Richardson, she is absolutely gorgeous, the fittest I have seen here so farâ€™. Time for bed Jonny. When the English crew returned to their daily grind back home and I was left in Chicago with nothing more than a sack of carrots to keep me company, I started to think back to the funny northern boy and his funny northern lingo. What exactly is fitness all about? It used to be about how far you could run, or how many lengths you could swim, but now in England it is all about beauty. So I wonder what exactly Charles Darwin was thinking about when he wrote about his theory of the â€™survival of the fittestâ€™. Is it about beauty conquering all? When the pet cat has kittens, the runts of the litter are always the first to go. The kids always want to keep the prettiest kittens for themselves. And in an age where appearance has never been so important, over-worked plastic surgeons may agree that Darwin was right all along, straight teeth, and a good rack will be an important part of getting that promotion, and finding that dream man (insurance of survival). Beautiful people are often more confidant, more likely to succeed in life, as the cream always rises to the top.   But then there are those who believe that to survive, you must battle your way to the top of the pile, be it in one big punch, or a slow wearing down of the competition, body beautiful or ugly as sin. It is all about what is inside, and how much you have inside. Survival of the fittest is certainly relevant in the army where they push you to your limits physically, to train you to fight those merciless strangers in a far off land that have pissed of your President. Preparation is three quarters of the battle. If you are trained well, if you are fit, you are more likely to survive. But as strict as they train them in the army, you all should know that we have the famous General in charge of discipline in our kitchen, and making it through the day is as much of a challenge as it is making a career in cooking. I doubt if the General has even heard of Charles Darwin, yet he knows all about his theories on life. It is back to animal instincts when in the kitchen, everything must be harder, faster, cleaner and twice as beautiful as the time before, and anyone not keeping up with the pace will fall along the wayside. I have the cuts, bruises and burns to show for my days of fighting with hot pans, wrestling with sweaty chefs, sharp knives and mind warping verbal torture. That coupled with the stamina of working a minimum of 14 hours a day, and no wonder chefs are never the prettiest boys in town.  Kitchens draw a strange crowd, the general attitude in England for the new generation of chefs in an up-and-coming culinary destination is; â€˜well, I was shit at everything else at school, and I like burgers, so I thought I would learn how to make them.â€™  Doomed I know.  So there are the dullâ€™uns who will get to cook the grill one day if they are lucky, too stupid to realise they will be working life-threatening hours for the rest of their lives. Take Paul for example, he is from up state New York, says Cwoffee like every good New Yorker should, and never talks about the ladies, despite his rugged good looks and muscular physique. He looks a bit like Mr Potato-Head with the tired-chef clip-on eye-bags. I saw him chatting to a girl at â€˜The Storeâ€™ one evening, but had to cycle on home before I was snowed in, and the next day over heard him telling big John about the â€˜pieceâ€™o assâ€™ he got the night before. What? â€˜Yeh, pieceâ€™o ass, you want me to spell it out?â€™ NaÃ¯ve little English boy.  The 20 kids from the hood came in later that day, and there was a girl who seemed to be carrying some junk in her trunk, I turned to Paul, â€˜pieceâ€™o ass?â€™ â€˜Nope, that is some classic ghetto-booty boy. Britney Spears, now that is a pieceâ€™o ass..â€™  Being brought up on New York pizza does nothing for your chosen culinary career, and when the only thing you know what to do with a chicken is fry it after 5 years in the industry, youâ€™d better get out if the game when time is still on your side. But for some the challenge of making through the day is sadistically addictive. Iâ€™d prefer to leave those sorts of sick power games for the slaves of this world.  Then there are those who genuinely enjoy cooking (you would be surprised), are interesting and interested, who unfortunately have the organisation skills of cuckoo on acid. The heart is in the right place, the passion is often there, and maybe even an understanding of food and cooking. But as soon as the tickets start coming in, and the pressure of feeding 70 people on 27 different tables is on, the chaos theory then starts kicking in, and a random mess comes out of logical progression. We have our representative from Fort Lauderdale, Florida to fill that void. Big John from the Northside learnt his trade from cookbooks and trial and error. He was voted Best New Chef in Florida before realising that there is more to learn in this big old world of fast food and fancy restaurants, so he went straight to the biggest name in American oat-cuisine, and took a job with Charlie and his angels. Unfortunately Big John has never had a chef to look up to and to work under before, and has never had the pressure of someone elseâ€™s discipline put upon him. Suddenly he is in a kitchen with people telling him what to do and how to do it, and as â€˜the Best New Chef in Floridaâ€™, gets flustered and frustrated with his own lack of direction. It is a tough old life, kitchen life, especially with 20 odd egos flying around. The Big Chef must not only be a talented cook and organising wizard, but also a king mediator and chief ego capper. The General is the best person to bring cocky-two-shoes down a peg or two. A bit like the sin bin in rugby or hockey, a good 15-minute spell in the pot sink will bring any man crashing back to reality with a thud.  â€˜I gonna ride your ass boy, I gonna break you..â€™ Gloria Gaynor got it right. I will survive.  So there are the dullâ€™uns, and the ones who wouldnâ€™t be able to work out how to get out of and open door. Both Paul and John get paid for their services, but there are then at least 5 people in the kitchen who work for nothing but the satisfaction of it. We get â€˜chefs for the dayâ€™ who pay thousands of dollars (to charity apparently) to dress up in the starched whites, and self-consciously wear those famous tall hats. Usually enthusiastic dinner party cooks who just happen to have fallen into a high paid stock broking jobs which enables them to lavish their guests at home with fresh white truffles from Alba, and come to Charlie Trotters to pick up hints on cooking Kobe beef. They come in their mufti clothes, which around Chicagoâ€™s high-end stock broking fraternity is a strict dress code of an English hand tailored shirt, Prada jeans and those nasty tasselled slip-ons that went out of favour for English gentlemen in the late 1980â€™s. They help peel South African porcini mushrooms, sipping Fiji water from a Riedel glass, while the undernourished chef next to him tries to make polite conversation while deep in the shit, guzzling warm Chicagoan tap water out of a plastic beaker. Worlds apart, but which is more likely to survive if Saddam dropped a scud on Chicago tomorrow I wonder?  We have a chef who should be the hardest nut of them all. He grew up in a war zone, spent a few years in the army, and is the only one in the kitchen train to kill, yet young Roy, hailing from some obscure goat herding village outside Tel Aviv, is the tenderest, nervous wreck I have ever come across. On paper he is the perfect employee for a hard kitchen. Firstly he should be able to handle himself physically; with all that torture training they do in the army, never giving up, even under the heaviest of provocation. And Roy is also hyper-intelligent, speaking no less than 7 languages fluently, so theoretically he should be able to handle himself intellectually too. Unfortunately theory doesnâ€™t always match reality. For all his world knowledge, and his expertise in killing a Palestinian 9 ways. He lacks the all-important trait in a successful chef, common sense. As an extern, here from culinary school for a 3 month unpaid stay, he unfortunately does all the jobs no one else wants to do, and when faced with a mountain of squabs on a Tuesday morning, you know you wonâ€™t see him for two days. He sets up camp down the back of the kitchen, and only moves from his spot for a 3.30pm feed, and a 5.00pm mop of the kitchen. I caught him wandering around the hotline during service having just finished his squab odyssey in some kind of dazed dream world that I think only factory workers get from doing the same thing for hours on end. He stumbled upon the grill station and watched Big John the meat man carving all sorts of delicious hunks of meat fresh off the wood fired grill. â€˜Tell me John, is this bison you are cutting?â€™ â€˜No Roy, this is a quail breastâ€™. He came back two minutes later,  â€˜John, tell me is this chicken you are cutting?â€™ â€˜Roy, this is rabbit, chicken is quite different.â€™ â€˜So John, tell me, this must be bison.â€™ John looked up in disbelief from his cutting board. For young Roy had just finished his marathon butchery lesson, pulling the legs and breasts from 200 flightless pigeons, carefully boning out the legs, de-skinning and trimming the little breasts. Did nothing sink in? â€˜Roy, this is not bison, quail, chicken or rabbit. This is a squab breast, remember?â€™ Poor Roy probably can say squab in his seven languages, can crack codes, and knows all there is to know about the falafel, but give him a squab breast and a hunk of prairie grazing bison, and he is lost. Sadly some have it, and others can only dream of it.  Cooking is in the blood for a lucky few, and being Italian certainly helps with that. Brought up drinking cold pressed extra virgin olive oil, knowing the way pizza ought to be and cooking risotto the way mama used to make it is an education even 4 years at the Culinary Institute of America canâ€™t teach you. Resident foccacia and mozzarella expert all the way from Milano, Giuseppe Tentori is the big sous chef, the number 2, in charge of everything from butchering one of his beloved rombo (AKA turbot for those of you who donâ€™t know their Italian fish names) to being the chief tormentor on the hot line in the evening. Giuseppe has spent the last 6 years with Charlie Trotter and is living proof of survival of the fittest. He may look a little kitchen weary with his blood shot eyes and his â€˜strawberry blondeâ€™ beard, but there is still plenty of fire in his belly. I swear that his right ear must be deformed from having a phone pressed to it for 10 hours a day; such are the hardships of a busybody sous chef.  After a 14-hour day feeding 180 hungry Chicagoans, Big John the grill man, looks like he has only just survived stepping on a land mine, the boy is a mess. At least 2 nights a week his apron is melted from standing too close to the wood fired grill (good luck having children John, if your apron is that frazzled..) on the other hand Giuseppe looks no different than the minute he walked in the kitchen, there isnâ€™t a hair out of place, not a mark on his apron, despite getting stuck in butchering fish and painting plates with orange, purple and black sauces. Having spent the last 10 years in this land of milk and honey, you would assume that language wouldnâ€™t be a problem anymore, but when an Italian with an accent as heavy as Giuseppeâ€™s, is screaming at you in the height of a busy Saturday night, he does tend to repeat himself quite often,  â€˜Hiillooow? Miit, wera is thea miit?â€™  Unfortunately everyone in the kitchen has anglophile names too, which is a little too much for our little Italian buddy. If everyone had names like Paulo, Francesco and Giorgio life would be so much easier, because Joseppy might be able to remember them. But because he has an aversion to remembering non-Italian names he calls everyone in the kitchen Johnny, or Jackson, which means in the height of a busy Saturday night, he does tend to repeat himself quite often.. â€˜Hiillooow? Johnny, miit, wera is thea miit?â€™ I have been promoted to the name of crack-boy, which at least narrows it down, as there is a crack-pipe in the kitchen too, and the Maestro from Milano is slowly coming round to the idea that my parents had of calling me Jesse. But one step at a time. He loves his kids, even if there is hatred in his bloodshot eye at 7.26 when the meat he has been handed is a bit overcooked. All is fair in love and war, and at the end of service he has been known to smile, well at least twice since I have been here anyway.  But if we are talking about survival, there is no better survivor than the Big Chef, Chef de Cuisine, Chef Matthias. Stick him on a desert island with nothing but a corkscrew and a fig leaf, and he will still make it happen. Chef Matt has been at the restaurant for nearing on 9 years and I remember the first day I arrived, he walked up to me with his overly tight-fitting toque pulled over his brow like and army officer, and came close up to me and stared very deeply, deep into my soul with contact lenses floating around these huge hazel eyes,  â€˜ARE YOU PSYCHED?â€™ To be honest I had never been psyched before, coming from a country that is about as enthusiastic about life as a wet Tuesday in March. In England we are real about life, honest about life, and if we are honest, life is difficult and grim sometimes. People complain about all those grumpy people they meet on the tube in London. You try living there with a permanent smile on your face, especially when sitting on a tube, which at any moment could derail; such are the problems on anything public in England. Americans on the other hand are the most enthusiastic people I have ever met, pity little of it is real, â€˜hi, how are you today? Great! Great!â€™ is a long winded way of saying hello in this country, to ask some one how they really are you must repeat the question, â€˜no, how ARE you today?â€™ and just to throw people, I sometimes answer with a negative answer, because as we all should know life doesnâ€™t always smell of roses, â€˜I feel like shit today thanks, a splitting headache and I have the most awful diahorrea.â€™ Watch the face drop, and the fear appear in the eyes. The fear of reality. But Chef Matthias, although having grown up in some safe suburb of Chicago, and being partial to the occasional high-five is quite real about life. He is the most European American I have met. He is interested in organic food, small producers of speciality goods, such as the mop haired hippy who comes to the restaurant every week to deliver real cream, eggs and the latest unlucky animal around the farm. He is fascinated by the contrasts of European and Japanese discipline and works the best from both worlds. He also makes things happen, he is feverishly passionate about what he does, he is sensitive and delicate, but is also an aggressive hustler when he needs to be. I think his philosophy on life is all about application and adaptation, the true goals of survival. That coupled with the fact that he has the biggest and sharpest knife I have ever held, a two-foot tuna razor, and if anyone wanted to challenge him for the throne of â€˜the fittestâ€™, heâ€™d cut you into the finest sashimi this side of the Pacific.  To survive in this little world of ours, we must fight to get to the top of the pile. If my poor mother heard this she would shudder with disgust,  â€˜How did I bring you up thinking like this darling? Life is all about living, loving and being.â€™ But maybe I am just talking about my little competitive world of constant limit pushing in the pursuit of excellence. We all have our goals in life, I wonder what yours are?  Jesse Dunford Wood  12 March 2004 '),(9375,2166338,'17,July,2004','2004-07-17 00:00:00','   aww, don\'t wory bout steve Nic. If half the stuff you have told me about him is true he proabbly isnt worth it.      Big hairy w00tage! Im back folks, a quick offline period and on  harddrive format later im back! woo. Now i gotta go download msn and  winamp again. thnak god i kept my music. phew!  '),(9376,2166338,'16,July,2004','2004-07-16 00:00:00','   yea i\'ll have to send you some of those pic\'s sometime qm i think you\'d like em ;D sry i haven\'t been posting much, i\'ve been busy and there\'s a lot of stuff i have to get through and work out since i just broke up with my bf just short of our 1 yr anni. i keep finding about more and more things he did that i didn\'t kno about and it hurts so much but oh well, makes me realize i\'m making the right choice so it\'s not all bad.  '),(9377,2166338,'12,July,2004','2004-07-12 00:00:00','   Un-Lazy? lol. Nic you should be a couch potato as long as people allow you to be one. because they are the best time of your life!  Send me the pics over MSN later and Ill host em for ya if you want!  '),(9378,2166338,'11,July,2004','2004-07-11 00:00:00','   lol! well actually i took a few pictures tonight buuuuuut i don\'t have anywhere to upload them, my hostess '),(9379,2166338,'11,July,2004','2004-07-11 00:00:00','   wow, pant creaming on our first day of blogging. things are looking good!  Any chance of a pic nic?  '),(9380,2166338,'11,July,2004','2004-07-11 00:00:00','   Blogs... are damn sexy, weeeee ok now I\'ve never really had a blog with anyone else before... ok so maybe i lied because i\'m part of  urlLink the lunch crew blog  but oh well still i\'ve never had a blog with tony before! i\'m so excited... i think i just creamed my pants o_o  '),(9381,2166338,'11,July,2004','2004-07-11 00:00:00','   Way, I have a new blog. (again) I cant remember what number this is but this time im with my good friend Nicole.  Isnt that right Nicole?  yeah see. :D  '),(9382,2166338,'28,July,2004','2004-07-28 00:00:00','   Yes that is right. I have gained a little knowledge of Html code so I can mess with the template and next time you come back it will be a shadow of its former magnificance. But for now, I ahve links. look! links! wow.    What you wanted more? Tough!   '),(9383,2166338,'26,July,2004','2004-07-26 00:00:00','   Yes, I have a new blog. the original was deleted but now it returns! and now it includes more fun! maybe.    Anyway, last friday night.Myself and a few chosen others went drinking up Chapel Field. or planned to anyway. Just before we got there, I got a call (on my brand new mobile! whoo! it has a camera!) from will saying that the police were confiscating all the drinks they find on underage people around Chapel field. After a quick conference we decided, fuck it. Well they did. I wanted to go round the field, but OH NO. Anyway to cut a long story short. there were no cops. We got nicely pissed. then played on a dance mat.    Guess who won! thats right! it wasn\'t me. Matt won instead. But I reckon I won the Tekken fights (Yes Tekken 1) anyway. Twas good fun. and I found out some interesting information about Hayley. Which is private. Away you dogs! *Staves off wth stick*   '),(9384,3386981,'28,May,2004','2004-05-28 00:00:00','   there are a lot of things that a man or a woman might want to say but cannot say them, there are various reasons why they are able to say them or not. however there is one thing any man or woman should easily be able to say, the word that comes to your mind when you accidently kick your bedpost with your knee or toe. Especially the toe, kick it just hard enough that a little bit of the nail is left attached to the more ably attached part of the anatomy. now after you wake up from your personal blackout you find out which condition your nail is in. for women it might be a consolation that they broke a nail and at least it wasnt one from the two hands. there isnt a third hand, if there were it would only spell more long nails and a higher consumption of nail polish and nail polish remover and other related items of use to women, cross dressing men, transvestites and other people who use nail polish (little kids who want to differentiate MY MUG from MY SISTERS MUG.  is nail polish, Polish? i guess i will never find out. or is simple polish, Polish? was polish used for the first time in Poland or did a Pole use it for the first time. if a Pole used it for the first time, who helped it? i mean, how can a POLE paint its own self. if it could paint would it let the Dog destroy it by raising one leg and doing it? which reminds me of the freedom Dogs and cats and other such animals have. They can just go ahead and do it. especially the stray ones. we have so many double standards. Our pets are more free to do what they wanna do they we are. more over they are fed and we have to kill and eat (that might be one of the ways to define working for a large organiation). for every job you get, some ten people dont, or maybe more. so should we consider ourselves lucky? or just plain stupid? or undemocratic (since we are doing something that the majority aint doin?) or cruel cause we disallow someone from taking a job and going back broken hearted or hungry or somehting like that?  '),(9385,3386981,'26,May,2004','2004-05-26 00:00:00','   If you should die before me, ask if you could bring a friend 					 (Stone Temple Pilots)    An innuendo is an insinuation (and or an implication or suggestion).  Life is full of little innuendos, it at times seems to be built of these little one liners that people keep dishing out as if they are home grown in their back yard or back lawn or what ever those things with grass and other weeds is called. Turns out most of them are from forwards and other kinds of emails that these people send and receive over a period of time. Johnny was a little boy; Johnny was a happy little boy until he started receiving these forwards. He now is a big boy; he is a big boy because he hates, and hate, like love is an extreme emotion. Johnny hates these forwards and every other thing that these forwards accompany.   	 Diarrhea is an innuendo.  It is an innuendo because it is tells you what you have done (Implication of), and it definitely TELLS you what you have to do in the very near future (Suggestion). It plays with oneâ€™s mind, wages psychological warfare on the mind, body and soul of the victim. Almost all of us have gone through it at least once in our lives. Anyone who is old enough to get to this web page has frolicked with it once. It is a beautiful thing; it reminds you of how nice and serene the world is. It is more powerful than love OR making love. It leaves your knees wobbly like you had just laid eyes or hands on the most cherished of your possessions. The love of your lives may not bring that effect that diarrhea brings with it. Ignore what you consider discomfort; the rose would never ever be so beautiful had it not been for the thorns.   It all started with a bit of pizza, a bit of fun and a bit of soft drink. The bit became a bit more than the digestive system could handle independently. Maybe it didnâ€™t like the taste of the pizza. I did enjoy the pizza thoroughly but the after math was not a sight to enjoy, not something to write home about.   It is strange, for someone like I , who enjoys a cup of tea while taking a dump it shouldnâ€™t have been THAT gory, but the way I felt, it seemed as if I was ravaged by a creature come up from the dark pits of the universe, something out of a Stephen King book.   The darkness of the night, it was a cool night, cooler in comparison with the previous three weeks of blistering heat. The skies had opened up and the clouds wept with embarrassment as if they already knew what was about to befall my poor self. The wind whined out side, hitting the house with considerable anger, lightning crashing brings occasional sparks of light to the otherwise dark night. I slept unawares, enjoying the cool wind swirling down from my new ROYAL FAN (after having bought it was I told that you may NOT get Amina Haq with it, she only appears in the advertisement does not jump out of the box, I do not like her particularly but having her around wouldnâ€™t have been bad at all). How? Oh how could it have occurred to naÃ¯ve old me about how my life was going to change forever?   It was a lesson in relativity. Sharp pangs followed by even sharper ones, cramping in the abdominal/lower abdominal area, spasms in the lower back. It all felt familiar, as if I had heard about it somewhere. I feared the worst; I thought that it was a curse. The utter obsession with the opposite sex had at last come back to haunt me in the worst of manners possible. On the other hand I would be notoriously famous as the first man to ever have a menstrual experience, and since I was a man it meant I had no experience what so ever.   I was sure things were changing and that I would soon be a beautiful woman and go out and haunt the innocent city dwellers like in the stories only I would be Dr. Jekyll and Miss Hydi!. Just to be sure I gave my self a thorough anatomical examination, with the kind of pain I was in I felt like Rambo sewing himself in First Blood. The result was a negative (which in medical terms means good). I am sure that long time back a doctor gave a patient some test and when they came through, the doctor in all his wicked humor teased the patient saying the tests had been negative. This led to the patient swooning and nearing death, due to which the high humored doctor lied about how negative, was good in medical sciences. I am so sure about such stuff that it is scary. Throughout this pain and confusion I was wondering that if it really happened and what with the recent negative tests, it would happen with my anatomy not supporting the correct accessories to go through painlessly with the experience. This thought sent a fresh shipment of ice shards down my lonely spine column. I knew I would scream, and I was wondering how I would feel and look like while screaming like a girl caught in a Hollywood horror / slasher flick???â€¦â€¦..  '),(9386,3386981,'21,May,2004','2004-05-21 00:00:00','   You are standing on the sidelines, thinking, should i or should\'nt i. and then you take that one fatal step. You step out of the beautifully air conditioned office, who cares if the office is decorated worth shit, it has air conditioners. They help you breath. You dont feel the gratitude while sitting in the office. Something hits your face like a hammer. you dont feel much, you cannot. The hair all over you body starts to burn and curl up as if you are standing in an inferno. It most definitely is HOT out there, but please pay no heed to mr. Nelly. Do not, i repeat DO NOT take off all or any of your clothes. You will only die of exposure. The other option is slow death, at the hands of the heat. A man thinks what hell would be like. The sweat your body gives off to keep you cool evaporates even before the glands can push any out. The realisation that you have got NO CAR dawns upon you with the heat of a thousand suns, you cease wondering and start trudging towards the nearest shade, only to find it occupied by enough people the body heat of those people could start a bush fire of its own if they were let loose in the bush. cant start a bush fire without a bush now can we. Amidst this chaos that reigns supreme in one\'s mind, one realises that one has been talking to one\'s ownself in the third person. \"Man it is hot today, i think you should buy a car now, cant go on like this for long, cant go even a furlong anymore\". however in all this gloom, all the stupid idiomatic dark clouds that just arent there at that time, there is a silver lining. Women!  we happened to leave a client\'s office right next to liberty market, at about 1.30 pm, high noonish, sun was at its playful best. brows tightened and creased, cheeks reddened with anger, frustration, flustered by the sudden onslaught of amazing nuclear fusion fueled heat, eyes sightless, ears drumming with the last sounds of blood rushing past the tympanic membrane. Tounge stuck to the top of your mouth, no saliva there baby. And that was when we layed our eyes on God\'s greatest creation to date, WOMAN. All this confusion and there they were, in all calm like little secluded lakes. All of them, out SHOPPING at a time when even the shop owners were reluctant to sell. They were having the time of their lives. Laughing, cheering about how they got the best price possible cause the shopkeeper was too dehydrated to resist their battering. The heat seemed to simply bounce off of them, how can a man resist such strenght and beauty, and intelligence. to come out to do something that you like at a time when you like, at a time that the opponent is at his weakest. The women all over this city are natural generals, right ful heirs to San Tzu and his art of war theories. At a time when all is bad, hair is burnt, cars over heat, air conditioners give up, roads crack, the ladies can put on make up, wear stilletos and walk the walk and talk the talk. Bah! chauvanist fools! Man\'s world, just beware all corporate people and other important fools, just beware of the day when a woman finally decides that she has shopped enough for the day, that will be the last day you sit in your plush sofa behind your mahogany desk pushing your expensive pen. it will be the revolutionary day of the much awaited coup, and your thrones will be slid from right under you. The only problem that might arise from all this is, that as soon as she takes over the office, she would want to shop again, you have to do something about the decoration you know, it looks like rat lived in here for a thousand years! and frankly cant have any doubt about the rat part anyways.  '),(9387,3386981,'19,May,2004','2004-05-19 00:00:00','   You know it  Just the way I told it We spent our lives dreaming To see a rose blooming Only to know it is screaming And we long to leave this place To go to some afar in space Alone, no one near us  Expression that comes itself It is evident is not forthcoming Forced acceptance akin to denial Self torture mass suicide Of our nonchalant existence Let us put as much distance between us and sanity Taste of tears and smell of my fears The sea, its breeze and its salinity  All that is pretty  All that is dainty So full of hate So much love for pity Yes I am alone here Said my prayers, seen the end is near I want to go there Alone or with you Life, it needs a shove and a push  Iâ€™ve done with love,  Done with out being sorry For the loss of one, when all is gone And till the drowning of the sun  Iâ€™ve said my prayers,  Said them for life Been on my knees Been there for you When you were there and now When you are gone  Iâ€™ve said my prayers With the rising sun And till it drowns Am tired now, will leave Alone and without you To hell where my life has gone And come back with the rising sun The day it doesnâ€™t dawn On a day that will not drown  '),(9388,3386981,'19,May,2004','2004-05-19 00:00:00','   SALAM, THIS IS GOING TO FIND YOU IN GOOD HEALTH, THAT I BELIEVE AND HAVE FAITH IN, CAUSE I HAVE FAITH IN THE ALLMIGHTY. THIS MAIL IS GOING TO BE IN CAPS,CAUSE I AM NOT GOING TO TYPE IT IN NORMAL FONT. THAT IS THE DIFFERENCE BETWEEN HAVING A CHOICE AND DESTINY. WHETHER THE FIRST STATEMENT IS TRUE OR NOT, I DO NOT KNOW, BUT I HOPE IT IS, BECAUSE I HAVE NO CONTROL OVER IT WHAT SO EVER, I DO HAVE A SEMBLANCE OF CONTROL OVER THE FONT OF MY MAIL. SIMILARLY IN EVERY OTHER THING GOD HAS GIVEN US A CHOICE. I HAD A CHOICE, I COULD HAVE TURNED OFF THE TV, CHANGED THE CHANEL, PUT ON A MOVIE OF MY CHOICE. BUT I THOUGHT, HELL I BOUGHT THE MOVIE, AND NEVER WATCHED IT, IT IS ON THE CABLE, NOTHING MUCH TO DO, LET\'S JUST WATCH IT, AND I DID. AMERICAN PSHYCO, IS A STYLISHLY CARVED, WELL EDITED, SHOCKING EXCURSION INTO THE NOT SO AVERAGE BUT OBSESSIVELY AMBITOUS MIND OF AN AMERICAN CORPORATE OFFICER. IT IS NOT RECOMMENDED, ESPECIALLY IF YOU HAVE ANY INDIVIDUAL THOUGHT PROCESS, DO NOT WATCH IT. IT CAN PROVIDE PLEASURE AND A BIT OF INSIGHT, BUT DESPITE ITS LONG MONOLOGUES FROM THE MAIN CHARATER, IT IS NOTHING BUT A ROAD TRIP TO A COCAINE FUELED HELL. A VEHICLE TO FAME AND FORTUNE, TO BE DIFFERENT IS WHAT THE MOVIE TEAM HAS AIMED. IT IS OLD AND IS EQUALLY SHITTY. IF YOU COME ACROSS SOMEONE WHO LIKES THE MOVIE, PLEASE MAKE SURE ABOUT WHAT YOU THINK ABOUT THAT PERSON. IT IS NOT THAT THE PERSON IS A PSHYCO, BUT IT MIGHT GIVE YOU AN INSIGHT INTO WHAT SORT OF GREY MATTER HE POSSESSES. I HAVE ALWAYS ADVOCATED THAT ANYTHING RELATED TO AESTHETICS AND TASTES OF A PERSON ARE HIS OR HER INDIVIDUAL HOICES. NO ONE HAS THE RIGHT TO SAY THAT YOU ARE WRONG IF YOU HAVE THE FLINTSTONES AS YOUR FAVORITE MOVIE. THE REASON COULD BE ANYTHING AS TO WHY YOU LIKE THAT PARTICULAR THING SO MUCH, WHY ONE THING APEALS TO ONE IS DIFFERENT AS COMPARED TO WHY IT APEALS TO ANOTHER. SOME PEOPLE WATCH A WHOLE MOVIE BECAUSE OF ONE CHARACTER, SOME FOR ALL OF EM, SOME FOR THE SOUNDTRACK, SOME OF TEH ENDING, SOME FOR THE CINEMATOGRAPHY AND SO ON AND SO FORTH. BUT YOU CANNOT SAY THEY ARE WRONG. NEITHER AM I SAYING THAT THE PERSON WHO LIKES THE MOVIE IS WRONG. ALL I AM SAYING THAT YOU SHOULD WATCH OUT IF HE LIKES THE MOVIE. I LIKE THE WAY IT WAS MADE, THE SINGULAR SHOCK IT DELIVERED TO ME WAS SUBTLE AND I REALISED THAT I HAD HAD ONE THE DAY AFTER WATCHING IT. HOWEVER WHAT EVER THE MOVIE ENTAILS IS NOT AT ALL GOOD NOR IS IT HEALTHY. MY RECOMMENDATION: READ THE BOOK IF YOU HAVE TO, BUT DONT WATCH THE MOVIE ON A FULL STOMACH. MY SOURCES TELL ME THAT THE UNEDITED VERSION CONTAINS ORIGINALLY EDITED SCENES OF SEXUAL VIOLENCE.  			REGARDS                      D  '),(9389,3386981,'19,May,2004','2004-05-19 00:00:00','   beep beep beep  '),(9390,3386981,'30,June,2004','2004-06-30 00:00:00','   Same was with my dearest of friends; he lived for his love affairs, with this world and its atmosphere, the narrow streets in large cities, and big highways in small countries, made in imitation of those in more modern and advanced civilizations. To him controversy was a way of life, the essence of what his existence meant, and even in love denial for him was ultimate. Loving some thing, and not being able to get it, to fight the desire and the wantonness, everything that was, had to have an opposite. For him at least, but not for me, his belief that everything that is must have a balancing effect, that if there were too much good in this world and no evil, this world would collapse, implode, unable to digest all the sweetness that was encaged in the tiny hemisphere. To him evil was necessary for good to exist.  That made it easy for him to understand the extent to which some people went to have their way, but for all the philosophy and the big words that were even banished from modern dictionaries, he was a romantic to the core. His constant failures to deny love, and acknowledge hate as the only true emotion in this world were invisible to his own eyes, yet surrounded him every where he went. This time, he was truly caught, for never before had he been bewitched by one of the opposite sex, the worst trap any romantic can fall into, true love mixed with a dash of denial and topped with lack of guts to go up to the person and say out loud what he felt.  The reason behind his newfound lack in confidence was that he had never loved something alive before this incident, something that has a choice and may say no when you said yes. He was trying to cover it up unsuccessfully, by saying that the extent to which he loved this woman, had to be balanced by her hating him, just as much as he hated himself for loving her or just as much as he loved her, which was by and large in the same ratio. The least he could do to maintain balance for the sake of the world was to make sure that she at least didnâ€™t love him. That was as far as his hypocrisy went, he couldnâ€™t make himself imagine being hated by his true love. That was a minor development in his otherwise unchangeable character. I had a feeling the next few months would bring about many changes in a terrain where change was a stranger. Hassan, that is my friendâ€™s name, thought that if he made her love him he would be part of another personâ€™s dilemma. That some evil might befall another man, or animal for all the happiness he would be enjoying. Unbelievable to some, but true to a very large extent in Hassanâ€™s mind which held nothing much in regard to opinions, and that too was usually borrowed from other peopleâ€™s high headed discussions.....(continued)  '),(9391,3386981,'25,June,2004','2004-06-25 00:00:00','   Most common people who tend to consider themselves normal because they follow the norms of society and find it easy to relate to other people around them, call dreamers as weak, strange weird and other one word and two word names that do not even begin to define the sort of people we really are. All this name-calling does not help our relation ship at all and only causes the rift between the drones, and us only to widen to an extent that it may never be bridged. At least not in our lifetime, mainly because people of our breed tend not to live very long lives, our average life span not withstanding we do try to eliminate the pain in our lives at the source of it, by foolishly eliminating life itself.  Contrary to what I am, I have tried to remain sober and true to life in the past ten minutes of story telling, but nothing that I say is a promise that the status quo would be maintained for any long period of time. However due to my new found reason to live, I can try to assure you that I will not take you for the ride of your life. This is not an original story, nothing man made in this world is original, and it cannot be, period. Why? Is a different issue for a different time and a different chapter, I cannot really promise that it will be pondered on either, maybe if I remember it, I will give forth my opinion on human beings and their delusions about being original. This is a story of tendrils of smoke, black lungs, more pin pricks in human flesh than a pincushion, your bossâ€™s voodoo doll that you have hidden under your bed, and other stuff related to the downside of having more than your share of fun in one lifetime, not that you would remember anything the next morning.  Please bear with me as I iron out a few minor details, wrinkles in the bed sheet so to speak. You would know how it feels when you look at a painting that has been put up but has this ever so slight tilt to the right; you feel the need to put it right. If you cannot reach that painting, or it is in the dinning room of some one you have gone visiting, then through out your visit you find yourself looking up at the tilted picture as if it is a ritual until you can leave the place and embrace sanity away from the paintingâ€™s hung disposition.  Please remember that what ever follows has happened before, it is happening now somewhere in this smaller than it was once world and it will happen again unless human kind, either vanishes from the face of the earth in the brightest moment of the rockâ€™s history, or if we manage to achieve the much dreamed about utopia. Only the vanishing part seems any likely and would also be very nice for the future of this planet and all chicken on it as well.    CHAPTER 1  â€œShe is so beautifulâ€ he moaned ever so balefully looking at me as if to gain some sympathy, or maybe he just looked as if he could do with some sympathy, whatever the confusion his condition though expected was not good at all. He lay sprawled on the leather-upholstered couch, one leg dangling so that his toe almost touched the Iranian rug, and the other foot; completing his designated pair, rested on a small stool. Light from the sinking sun filtered through the dirty window illuminating his face a bit, if not his brain.  It was pretty obvious that he was smitten with the girl he was talking about; at least it was a girl and not a car. Last time he came down like this with all his articulations drowning in his eternal undying love for HER, it turned out it was a classic car and he couldnâ€™t get HER out of his mind for a minute. I felt cruel, considering his predicament was real and as serious as any other one sided love affair but it did not have much effect on me. Maybe because I had never been in his shoes, dainty as they were I tend to step lightly where daintiness is at hand. When it came to my thoughts about my friends and their continuous flirtations with love I was a nobody, a pure dyed in the wool outsider, who knew no bench mark to define or categorize the extent of seriousness in peopleâ€™s emotions.   I merely looked at him as if he were an object in a museum, cold and beautiful, something you are not allowed to touch either. But despite all the rules and laws in the world, there are people who touch stuff in the museum and then there are people who steal them for which touching is a prerequisite. All of this happens for their vested interests, could be monetary or maybe touching the artifact just because it is forbidden, takes them to a high that nothing else in this world can. It can also be the selfish, obsessive and possessive nature of human beings, to capture and enslave something that you cherish, so that no one other than you can ever get near it see it taste it or breath it. Sometimes, it is just good to touch it, to know that it is there, that it exists; some things and some people are so beautiful, that you cannot imagine their existence until you have actually touched them. It is not necessary to touch them physically, an emotional connection, even a disagreement strong or weak can both lead to a co-existence of two different people, both feeding off each otherâ€™s presence, life blood without even knowing that the other person ever existed. Controversial statements but still very possible, you know another person is there, ever present, his or her existence felt in every other thing that you do or donâ€™t but you never know who you do it for or how................(continued...)  '),(9392,3386981,'22,June,2004','2004-06-22 00:00:00','   Not knowing where life was eventually going to lead, I marched on. A lone soldier amongst millions, no one to worry about, no sorrows, no tears, neither of sadness nor of joy. Life was a long railroad track along which I was swept, like a vagabond sail torn off some old ship ravaged by pirates. Life had no meaning, and even if it did it would have been really painful. It would be good if it continued and ended the same way. There would be the consolation of the thought that it was only something meaningless that had ended. Something that meant nothing to anyone, not even death itself thought much of this mockery, a character whose job was to mimic life. The greatest thing about being what I am is that you can be anyone you want, anyone you wish, anyone you look up to, dream to be, dread to be. Your best and your worst nightmares are at your doorstep; all but a single knock away.  You didnâ€™t even have to close your eyes, and you could be in dreamland. Even if there were an Alice and her fabled wonderland, I had it beat by such a long margin that if she knew she would beg me to let her play. I can be your Knight in the shiniest armor, the last hope for planet earth. Make me your friend and I will be the Huckleberry Finn for you that Tom Sawyer never ever had. But in reality, I was nothing but a mere daydreamer, a man who had promised a lot, shown the potential of being the genuine article. The prodigious child that every teacher dreamed of having for a student but never got. No, this is not where I daydream, this is where I reminiscence, this is how it all started. The tiny seed that got planted by innocent onlookers, watered by unwitting patrons and my ludicrous imagination which later grew into a thousand-year-old rain forest of hopes; rotten to the root and unable to bear fruit for any creation on this planet. The infection seemingly benign and common had seeped in so bad that not even vultures would have a piece of this cake. Self pity? You couldnâ€™t possibly know what the word meant, no one could, and frankly I donâ€™t think I will ever know what it really means. I have never shed tears for my self nor felt anything much while shedding them for others in pain. Only the tears roll down when I dream of saving some one in pain and come out all painted in glory. Life has always found a way to get around hardships and give the tiring mind a rest; those are some of the excuses that a man or a woman who wished to dream his troubles away would give, to dream. To run so far away from reality that the taint of society cannot possibly find a way to reach them, hidden in their little hidey-holes where they themselves are only just comfortable. The comfort spawned from the thought that they reside in a place where no one else would ever feel comfortable enough to stay for long and disturb their sleep.  '),(9393,3386981,'06,June,2004','2004-06-06 00:00:00','   We didnâ€™t have much to do; hell there wasnâ€™t much we could do. The sky was totally unapproachable, thick clouds and occasional whiff of mist stood its ground between our sight and the speckled blanket enveloping our universe. The sky isnâ€™t really there, it is just something we underachievers talk about reaching knowing we cannot cause it just isnâ€™t there. Space without sunlight is what it is. It was really quite beautiful there, in its gloom and tiny noises of snow falling off the tall pines that surrounded us. Every now and then one of them would shudder as if they couldnâ€™t handle the weight of snow on their shoulders. I could most definitely relate with the thought, I donâ€™t know whether the rest could or not, but I tried to be honest, knowing my limitations and I knew that I could never come close to encroaching them. The clouds overhead rumbled as if brooding, whether they wanted to shed more snowy tears over this valley of fear and suppressed human emotions or not. No one actually knew how long we had waited there in the cold darkness. Huddled together like a fresh dog litter trying to stave off the cold, willing it away. It burned it was so cold, went through the pores in our skin and even the adrenaline induced sweat seemed to freeze in the glands, however the tears never stopped flowing. Warm flowing salt kept making steamy streaky tracks along some quivering whitened faces as if some skiers had had a day full of fun and frolic in the alps. Cynicism aside life could get worse, the thought of being dead was a blessing now, it might have been worse if i had wanted to stay alive, especially after what I had seen people go through. Painful is what having a heart is, does not allow you to feel pain for yourself, if you see others in worst. Would have been better off had I seen only with my eyes and nothing else. I got involved, biggest mistake I ever made in my entire life, I had always been a bystander sometimes enjoying but always observing what happened around me. Observing yes, but only objectively, no emotions involved what so ever. There was never a need to do so, I was so right, I have always hated being right and I was when I thought that to stay aloof, away, the heir to the throne in introvertism; if there is such an ism, would be near to perfect. Nothing really can be perfect, but one thing is for sure. No one human being can stay away from humanity for long; every man is an animal, part of a society. Some sort of an emotion always creeps through let it be good-natured or bad there is always an underlying current of emotions. I donâ€™t think us humans are deserving of so much nor can one man or woman bear so much pain or hate love, distress ...there arenâ€™t enough shades of gray to cover the feelings and emotional ebb and flow that one might go through in one micro second.\"  '),(9394,3386981,'03,June,2004','2004-06-03 00:00:00','   Twisted thoughts flowing Sleeping they were, Donâ€™t know what gave them wings Shame, embarrassment, lies and deceit Head in knees The little boy is crying His soul is dying With every word that is said Between every line Leave it all behind Refuse to resign Ourselves to truth that drowns inside Forever gone innocence Slipping like tears All shame and fears Stripped of garb and cloth Naked flesh, strutting the streets Flushed with egos The sea is rising  To meet the moon Spell it out, it reads soon The tide is high But its only noon The little boy cries, Singing with his soul She leaves him now The sun is cold A mother wishing his hunger away Not knowing what hurts so She such a fool You canâ€™t get what you wish for It never works that way We get what we deserve The devil sings as he may The sun shall rise On someoneâ€™s day Tears will fall Where someoneâ€™s love once lay They all still lie Waiting to die Thinking, of days gone by When the young were shy Some would laugh, while some would lie For every lie, someone would die The moon wept, seas rose high The world grows cold And the sky just dies The little boy cries His mother to him lies Wait a bit longer Dinner just arrives Let the kid sleep, she begs Deep,  Somewhere deep Where every dream dies, where he is not alone Where every one cries  '),(9395,3386981,'04,July,2004','2004-07-04 00:00:00','   Hassan was a man; that was the least that could be said about him, and something that cannot be said about many of the gender but not of character, walking around on this planet at the same point in time and space. He was a man of his and other peopleâ€™s word, some one who had never had to work, by choice or by compulsion. He wasnâ€™t in habit of whimpering for everything he wanted or desired, even though most of what he wanted he could have, at least the material wished could easily be fulfilled with what his family had. There was a time when he fell in love with the Alps, wanted them to be available to him in his city, which was half a planet across. Of course such foolish wishes are not readily granted. He could have had it, provided he shifted there himself, but he didnâ€™t, he wanted the alps to be in his back yard and no one could take the right to wish away from him. He wasnâ€™t deliberately eager to please other people, he was just that way, and he was an earnest person who went whichever way you wanted him to. Maybe because at times he never had a direction in his own mind to go in, he was the kind of guy who would blindly go where no man or woman would go before him. For thousands of years man has evolved; born, grown, failed and succeeded, built pyramids which are effectively the most elaborate houses of the dead and nothing more but mere graves. Awe inspiring to some but nothing more than mere graves, signifying only one fatal reality; death. To Hassan that was pretty much a decent thought and I agreed, but I still thought that the pyramids if not a wonder were very well constructed buildings. You could make them palaces or shelter homes, but we humans are so intent of preservation of the past that we seem to forget the present. Every year millions are spent on rejuvenation of pre-historic artifacts but it is never thought that while we preserve our history we lose a bit of humanity, as it exists now. None of us ever stop to think why the pyramids effectively glorious graves stand yet we struggle to find the remains of the magnificent palaces that existed with them. Why has time preserved tombs instead of glorious palaces, why do we persist on calling the houses of the dead the glory of the past yet fail to see that the glory is past, and does not exist anymore than the dead lying in the graves they built for themselves. Something to do with cherishing what is past or could not transpire, same with the first love of everyoneâ€™s life, whether you get it or not, whether you lose or gain from it, thoughts of it follow you even through darkness, through places where your shadow leaves your wake.  '),(9396,3386981,'04,August,2004','2004-08-04 00:00:00','   I have a little problemâ€¦. I have a demon for a wife He delights in my pretty face And he hates my life Takes notes on how to provoke past grief Makes my teeth decay, with the last of my self belief 				(Addicted-Faithless)    A time when you feel the need to say something the most is the time when you should say nothing at all. Easier said than done Iâ€™d say, climb up a flagpole and kiss your own arse is what I say. I say a lot, but realize only later that most of what I say is a bloody figment of my imagination. Never get around to actually saying it. I am sure; that had I said what I had wanted to, when I wanted to, things would have been totally different. I would be lying somewhere, oblivious of this pain, oblivious of the worldâ€™s sorrows and stuff, lying somewhere in my total decapitated solitude. My head separated from my body, would still be dreaming and talking though. But that is just a vision I have, I donâ€™t think it is going to materialize. I really hope it doesnâ€™t.  Evil dead is a movie I have seen many times. The scene where the girl sits in the window, telling cards true to the last color wrapped in her little blanket fresh from being raped by a tree and its branches, makes me think, is it dangerous to wander off in a jungle at night alone for a cute looking girl because of:  a)	The evil spirit haunting the jungle b)	The sexually frustrated trees c)	The trees possessed by a sexually frustrated evil spirit d)	The sexually frustrated men masquerading as trees e)	The sexually frustrated men brought up by apes and who now call themselves Tarzan f)	The Apes who brought up Tarzan and now want the real shit g)	The Female Ape who brought up Tarzan, now hates Karchak wants to try another flavor? h)	The sexually frustrated Lesbian Women, Cross dressing as men because that is the most believable disguise (even if a woman rapes another woman, the rape`e would assume that the rape`or was a man because only men are so frustrated so as to impose themselves on unassuming women running topless in an evil infested jungle in the middle of the haunted night).   There could be some other reasons as well, but what I really want to ponder upon here is not how a woman gets raped, even though for some that would be the best part (bloody perverts), I am rather interested in why women get raped while frolicking in the jungle, naked, in the middle of the night. When a man teases a woman (eve teasing I suppose). It is not appreciated by the feminine self-esteem at all. The ass pinching, and breast staring really hurts the feelings, and rightly so, It is unnatural, goes against the nature of mankind, but at the same time conforms with the nature of man, because that is what we do, attack and pillage and rape the poor lil bastard who cannot defend his or her own self.   So if men teasing women is unnatural, then women teasing men should be perfectly natural. All women should tease men, grab asses where ever they see a cute one, stare at crotches, push and shove in trains and buses, fight with friends in loud voices as to who gets to sit with the nice looking guy sitting alone in the back seat. Drop a hankie and ask if he dropped it and then ask if he wants to have sex, or just ask if he wants to have sex, forget the hankie (men like the bloody direct approach) after all, it is only natural!!!!  '),(9397,3400445,'20,May,2004','2004-05-20 00:00:00','   Canada.  a hockey nation.  a nation that without the stanley cup is like a summer day with no sun... or like may 24 long weekend with no beer and no friends...  Calgary has the first chance in 10 years to bring the cup back to Canada, where it belongs.  so lets join together, crack beers, don jerseys, and flip on the tube.  let us join as a nation and cheer for Calgary...  let the Flames burn!  '),(9398,3400445,'20,May,2004','2004-05-20 00:00:00','   Today, like any other day.  waking up at my ex-boifriends after a night of fun and wicked sex... BEEP BEEP BEEP the harsh reality chills me as the covers are pulled back.  he gets out of bed to the telephone. he loves me. i love him. yet the inevitalble phrase tugs at my heartstrings everytime i see him...yeah, well she is my girlfriend. so what am i? what name is there for me? \"don\'t over think it,\" he says. fine. done. names are supposed to inhibit anyway right?  like putting an ivisible box around something. \'it\' cannot change once \'it\' is defined. there is someting cozy about a box though...perhaps it\'s the familiarity. overthinking again.   how does a person overthink anyway?   i know i have done this... and i know i have witnessed other people overthink.   i\'ve told them not to. but where is the line between not thinking enough and overthinking? am i overthinking overthinking? fuck. all i know is that i want to be with him... i love him. i love him unconditionally. she is leaving. thank god. but will it be different when she is gone? i know this is not about her... it\'s about him. getting his shit together...whatever that is. i know that just because i don\'t understand this doesn\'t mean it isn\'t okay... the problem lays in the fact that i cannot truely move on. it\'s not that i haven\'t tried. i don\'t want to. i do not want anyone else. i want him. until i am sure things will never work out...i do not want to be with anyone else...period. i do not want to be in an unhealthy situation. i just want to enjoy life.  enjoy my time with him. and not to overthink.  '),(9399,3400445,'24,June,2004','2004-06-24 00:00:00','   so i think I\'ve missed the point. is it really love that is important? because it seems that we run around chasing those that do not want us. and those who do...We do not want them. is love just a game? and is it what we should all be looking for? or is it in fact just an illusion? something that distracts us from our everyday life? perhaps it is the connection that is most important. not love at all. not that love is not important. but is it the most important? is the connection to somebody else what feeds us. what keeps us going. what helps us to feel like apart of something. something important. a connection that helps us to relate. to feel connected. to ground us on this earth in a way that feels safe. feels real. should we in fact be looking for those that appreciate us? as opposed to those who love us?  '),(9400,3400445,'23,June,2004','2004-06-23 00:00:00','   he is the best friend a person could have. he is loving. but in a way that is almost undetectable at times. i have taken out my magnifying glass. i see now. i see that he loves me when he talks to me. even if it is harsh at times. he is speaking. that means he cares. i feel so incredibly lucky. to have found a friend that will take the time out to take a chance, maybe initially even be considered mean...to get his point across. to let me hear his opinion. i hope he never stops speaking. i love him more than ever. i love him always.  '),(9401,3400445,'10,June,2004','2004-06-10 00:00:00','   maybe im just normal. maybe this whole process is what figuring yourself out really is. i love him. i will not stop. i am not a fool. i am in love. i would never trade the bad...i would forfeit the good. not worth it. never worth it. good and bad are part and parcel. they are tests. tests of your emotions. a test of yourself as a person.  '),(9402,3400445,'10,June,2004','2004-06-10 00:00:00','   how could i let this happen? how could i upset those that i love most and who are closest to me? how could i let myself get to such a state that i am no longer myself? not myself... but it was me. who was that psycho?  that psycho inside me? i can\'t undo what i did. i must make sure it never happens again. never. i will never drink again. ever. this day, june 23 2004, i vow that i will never drink again. a promise that only i can break.  '),(9403,3400445,'10,June,2004','2004-06-10 00:00:00','   i have been a fool a fool in love a fool for love i believed that if i loved, loved with unwavering strength, it would matter it would make a difference i was under the impression that things would work out that love would conquer all that nothing else mattered, if there was love i have been a fool a fool in love maybe i thought that i could love enough for both of us maybe i wasn\'t thinking maybe im overthinking i have been a fool a fool in love overlooking other women other distractions knowing that i love you  knowing you love me is it possible to love and be with someone else? are their different types of romantic love? can a person ever love you the way you need to be loved? would they try? does anybody love the same way? why is this so important? the feeling of being loved  loved by you you whose love is the most important i have been a fool a fool in love i can\'t move on im trying i can\'t im sick to my stomach  i don\'t want to hurt anymore i hate that i don\'t understand you why you don\'t want me maybe i am empty maybe i am not enough maybe...  maybe im just a fool  '),(9404,3400445,'10,June,2004','2004-06-10 00:00:00','   My thoughts on the subject?  BOOOO  '),(9405,3400445,'02,June,2004','2004-06-02 00:00:00','   your love feels like... the way you touch my cheek the way you say my name the way i feel when i know your looking at me the way you run your fingers through my hair  the sound of your breath while you sleep the way you love your cat the way your eyes dance the way you walk the way you hold yourself the way you feel music the way you explain and teach the way you birdbath the way you laugh the way you sing to me and write songs for me the way you care about your family the way you take pictures of me the way you creep up on me and tickle me the way you make me feel like the most important person in the world when you look at me the way time seems to stand still when i am with you the way we can be doing something or nothing and it doesn\'t matter because just being with you makes me feel happy the way i feel in the morning when i wake up next to you, like i can do anything the way you told me, \"the best feeling in the whole world is the way i feel next to you\" the way i get excited and why i always read your emails first the way you make me want to be a better person, to take better care of myself  the way your sweat smells so sweet the way i know you love me, and yet i also know that you love the idea of loving me... and that loving me and being with me are two different things.  the way i feel when i can\'t stop thinking about you and start to write and write and write  '),(9406,3400445,'02,June,2004','2004-06-02 00:00:00','   today i learned the meaning of nice.  it does not mean that you pretend to hold the elevator door by reaching for the button.  or that you do not spit in food before you serve it.  or that you do not make noise at 3am in a quite residential neihbourhood.  being nice gives a person a good feeling.  is this good feeling abused?  are people nice just to make themselves feel better? not to simply help out a fellow human being?  so what happens when a person is \"nice\" for the wrong reasons?  what are the \"wrong reasons\"?  let me enlighten you as i have recently found out first hand.  i have found that people are \"nice\" in a way that is purely self-serving.  oh sure.  they do not realize it at the time.  they are being nice.     beware  those who are nice and then somehow end up throwing it all back in your face.  those who find a way to turn the \"nice\" thing that they did around and use it against you...  use it to expose a weakness.  then use this weakness to elevate themselves and showcase how \"nice\" they were...  and what could we ever do without such nice people?   beware  of those who are nice.  take care of yourself.  everyone is self-serving...  everyone.  '),(9407,3400445,'01,June,2004','2004-06-01 00:00:00','   Love. what is love anyway?   an intense bunch of feelings for someone or something?   what everyone is searching for? the energy that binds us as humans and helps us to not feel so alone in this vast uncaring world? it binds us just as it rips us apart. there is no rest for the weary. there is no rationale for the confused. love rips through my heart like the roar of a harley davidson on a hot dry summer street. emotions leave you weak.  spent.   loss of all ability to properly protect yourself.  no fort to call my own. open.  open to pain.  open to love. no walls on my fort.  the fort that was supposed to be built.   the fort that others seem to have are so big and strong. i have no fort. i am a wanderer. i am alone. i visit others on my travels through life...yet their forts have walls that can not be penetrated.   passed through, but no one lets their gaurd down. always snipers on the walls.  waiting. waiting for you. to slip up. a fort is much stronger against a traveller when they are lone wanderer on the inside. it encompasses.  it drowns. the light that penetrates is only 10 to 2.   everything is dim. many shadows cast. i have left the last fort i encountered on my travels.  i see it only in the distance. there is sand in the wind and my vision is blurry. it seems like home. like something safe. i turn around. keep going. that is not home. it is an illusion it has created for itself. once leaving a fort you must not return. so lonely. so alone.  '),(9408,3400445,'05,July,2004','2004-07-05 00:00:00','   i love you for your strength of soul and spirit  i love you because of your energy the way our energy flows together much different energy  but along the same path in the same direction  i love you because you can handle anything, but feel the crushing weight of it sometimes just like everyone else  i love you because you are stubborn, caring and loving   i love you because you are impatient  i love the way you project your own exact actions, feelings and/or emotions onto others   i love the way you hold things in, and by doing so shut people out the way you don\'t think anybody else can help the way you don\'t want to burden anybody   i love you because you are totally unreliable but you are always there for the people you love when they really need you  i love you because you are always missing the point  i love you because you challenge me in ways that drain my energy and leave me misunderstood and frustrated  i love the ways you have tested me without letting on that you are doing so  i love that you think you are clever and keep things from me because you think its best but would never want to hurt me or keep things from me at the same time  i love that you underestimate me  i love you because you are scared just like the rest of us  i love you because you use peoples disadvantages to elevate yourself everyday, all the time, wherever you go  i love you because you are complex and yet transparent  i love you because you think that since you have experienced something you are an authority on it and can look down your nose at those going through a similar thing  i love the way you feel too good to be true and yet really you feel you are never good enough    i love the way you feel you want/need to protect me  i love the way you feel generous towards me because you feel stuck  i love the way you trust me  i love the way you love me and hate me at the same time  the way you think that everything will work out in the end but that you are not an optimist the way you are in control of your own life, but leave the future to chance and fate  i love the way you are the gravitational pull in your own universe  the way you never think first about the needs of others, unless it directly affects you or the way you feel or look to others  the way you realize that my love for you is pure, unrelenting, uncompromising, and unwavering...no matter what and the fact that you don\'t know how to feel about it  i love the way you love me, need me, can not afford to not have me in your life...and tell me you love @*!^, all in the same breath.  i love your confusion  i love your humanity  i love that you want to \"do the right thing\" and usually fuck up the situation because of it  i love you because you are only human, but want to be so much more, to so many  i love that you put yourself first, but do not take care of your own most basic needs in doing so  i love the way you do everything fast... eat, sleep, fuck, drive, walk, read, type, listen to music, make phone calls, converse, evaluate, draw conclusions, move through parties, lives, love, life  i love the way you are a contradiction of yourself  i love that you do everything with the best of intentions  i love the way you know i know you the way you know i love all of you all sides everything i want to know all that i do not there is so much more, i know it you are my light  '),(9409,3522747,'03,June,2004','2004-06-03 00:00:00','   My first experience with computers started in the 7th grade. My mom was nice enough to save up all of her grocery receipts for six months so I could buy my very first computer...the Timex Sinclair 1000. The local grocery had a promotional thing going, if you turned in the receipts (six months worth) you only had to pay $50 for the whole shebang!  I was very excited about this computer! The Timex Sinclair had a 3.25 MHZ Zilog CPU, 2K of RAM, connection for a cassette back-up drive, a 16K RAM pack and a thermal printer. If you would like read more about the 1000 use the following link http://oldcomputers.net/zx81.html. As soon as I could I hooked the system up to the T.V. in the living room and off I went. I spent the next 10 hours typing in line upon line of BASIC code to produce a clock that barely kept time, AHHHH heaven.  It was not long before I found other friends and teachers at school who also had the same computer set-up. We soon started an after school computer club that met twice a week. Every thing was going very well However, little did I know that my computing days would soon be coming to an end.   After a while the others in our club started to get borded with using these tiny under powered machines and started upgrading to larger PC\'s. I could not, due to my parents financial situation, not to mention a lot of people were less than supportive of my new hobby. I was told by an older cousin in no uncertain terms that my lack of mathematical ability would keep me from pursuing a career in computers so I should \"just give it up\". So I did.  The next real contact I had with computers came shortly after I joined the United states Air Force. The Air force started buying a limited number of IBM compatible 286 machines round about 1990-91. We used them to order supplies and write performance reports. This was my first introduction to the world Microsoft products more specifically Windows 3.1.1. After messing with these machines I felt the fever coming back over me.  I was a little strapped for cash so I tried to find a used IBM compatible machine but, my efforts were unsuccessful. Not long after my search started I received my first credit card (mistake #1 but that is a whole \'nother BLOG). I had several friends that were into computing and they persuaded me to look into purchasing an Amiga 500HD, they said these units were the best out there for gaming which was what I was after at the time (this was another mistake and a whole other BLOG I could write). Needless to say in less than 2 tears Amiga folded and I was left with a lot of useless computer stuff.  Since the Amiga I have owned several machines upgrading from year to year, all of which run windows operating systems. My next project will be to build my own one of a kind system, geared mostly towards programming and web design. I start a introduction to programming course with the university of Maryland in two weeks. Well I\'ve rambled on enough time to get work researching computer stuff for this BLOG Joe sends...  '),(9410,3522747,'02,June,2004','2004-06-02 00:00:00','   I finally have found the courage to start my very own Blog. I\'ve decided that my Blog will dedicated to the passion in my life...Computing. I wish to learn all that I can about computers and computing by building this blog and eventually turning into a full blown website. Please feel free to e-mail with criticism, suggestions, links to cool sites or something you would like me to post on this Blog. Let us dive in shall we!  I guess the most logical place to start would be with a definition. Computer:An electronic device that processes data according to instructions fed into it (Geddes & Grosset English dictionary). Computer:A device that computes, especially a programmable electronic machine that performs high-speed mathematical or logical operations or that assembles, stores, correlates, or otherwise processes information (Dictionary.com). Personal computer:  general-purpose single-user microcomputer designed to be operated by one person at a time (The Free On-line Dictionary of Computing, Â© 1993-2003 Denis Howe). Simple enough I guess, I added the personal computer definition because that is the term I hear most often when friends of mine refer to their computers.  The preceding  definitions are very basic ones, and I\'ve done this on purpose I do not wish to confuse anyone on this subject (especially me). If I were to use a lot complex definitions with lots of big words people would lose interest in my postings very quickly. Keep It Simple Stupid Is one of my basic rules. If anyone sees me straying from this beaten path please feel free to set me straight!  '),(9411,4158987,'07,August,2004','2004-08-07 00:00:00','    I love that word. Ponderings. If you break it up you get pond, like lake, eh, like oh, ring like disaster. Because wedding rings do tend to end up in disaster. So pondering is like dipping into a lake of disaster. Mmmm. Maybe if I shut my eyes I\'ll be able to feel the waves washing over me, hear the delicate ripples as leaves fall into my pond and breathe in the air of immortality. If the Philosophers Stone is a soluble tablet, my lake of disaster is the solution for it. Like in chemistry. I dreamt the other night that I was in a cold, dark lake and that I was below, watching the world through a rippled mirror. I saw myself peering down and a white hand slooshing the water and slooshing my view. When it was finally clear again, all there was was a star that lit up the lake. And in the middle of the star, there was me, spread out in a crucifix shape with my eyes blazing and my hair fanning out like it does underwater. I reached out to feel the wetness on my hair, but I felt only heat as the star plunged into the icy water and I melted into oblivion. I love random dreams. '),(9412,4158987,'07,August,2004','2004-08-07 00:00:00','    Hi, I\'m Tash and this is my random blog which I created for two reasons. One is because I was bored, two because I wanted a gmail account and three because I decided the world needed to hear the random Ponderings of a Nutcase. So thats me. The nutcase. I live in New Zealand in a house with a cat and a dog. And I like to be random. And my pet peeves are French keyboards, the sounds MSN makes and jingle singers. I mean who really puts on their CV, \'Ah yes, that\'s me. I sang in the Felix Catfood jingle.\' ??? Also, I don\'t like Bush. At all. And I don\'t like Micheal Jackson because he has a dead nose and sings like a sick hippo. There are lots of things I do like though. I like melted mozerella, the number 666 - just a random thought but if there is an apocolypse, wouldn\'t it happen on 06/06/06? Just a random vibe - and I like redheads. I love shopping. Shoe shopping. But I hate French keyboards and porridge. I think that is enough randomness for now. More later. Tash P.S. If anyone has a gmail account they don\'t want well... lets just say I know a taker... '),(9413,3984120,'20,July,2004','2004-07-20 00:00:00','    Tuesday, July 20 2004 These are just a few selections from when I was a lowly, 1st quarter copywriter. Browse through them if youâ€™d like. Please note, just like Heddyâ€™s lecture disclaimer, I frequently use foul language. Iâ€™m Irish Catholic, what the fuck do you expect?  Shitty writing at its best  Monday, January 19   Big Alâ€™s BMW So my friend Mike is in town for a couple of weeks. He\'s the only one left who\'s still in an academic institution of some sort. The kid\'s a year away from getting his degree in pharmacy from the University of Montana. It\'s so fucking cold and snow laden there this time of year that they give the students practically the whole month of January off.  At about 1:43 this afternoon my bedroom door flings open.  \"Get up, pussy,\" my roommate Ackerman grumbles at me in his usual baritone.  I\'m hurting pretty bad from a night of boozing it up that lasted until about 5 AM. I say nothing. He begins berating me, now with movie lines.  \"Stu! You\'re rent\'s do mother fucker! Don\'t pull that falling down the stairs shit on me. You conscience.\"  \"Get out of my room you damned communist,\" I fire back at him behind closed eyes.\"  \"No, seriously dude, get up. It\'s 2 o\'clock. Mike\'s taking us to get your truck.\"  \"Fuck my truck.\"  \"Well he\'s coming to pick me up in about 15 minutes and we\'re going to get food in the Highlands, so if you want me to take you to your truck this is you\'re only chance. I\'m not going back over there later. Besides, he\'s got his dad\'s new Beemer.\"  I guess I should clarify a few things before continuing. Mike\'s dad, Big Al, is a pretty damned brilliant lawyer and consequently, friggin\' loaded, hence the Beemer.  Ackerman is always playing angles. In this case he\'s probably correct in assuming that I\'ll ask him to take me down to the Highlands--where I left my truck the night before because I was too shitfaced to drive 4 miles home. Sometimes, however, his constant effort to make decisions based on his interpretation of the law of averages pushes his personality towards out-and-out paranoia, especially when girls are added to the mix.  I am extremely fucking lazy and difficult to deal with when I have just woken up. When I was in college I had to alarm nearly 2 and a 1/2 hours before I actually had to get up because of my propensity for mashing the snooze button over and over and over again. This drove my roommates crazy.  I\'ve gotten a little better about it, but alcohol consumption the night before can certainly make me revert to my docile ways, as it did tin this case.  \"His black Beemer?\" I ask.  \"Naw, nugga,\" Ack explains sensing his opportunity to close the deal. \"Brand new two thousand-fizzo, 525i. He\'s only had it for two weeks. He just left for Kansas City this morning so Mike\'s got it for a few days. We\'re gonna take it out on the highway and see what it can do.\"  By this point he\'s got me. I start to accept the fact that I\'m going to have to get out from under the warm, wrinkled covers and be a citizen of the world once more. Yes, the Beemer\'s the main enticement, but also the need to urinate is becoming critical as well. All the forces of the universe are conspiring with one another to push me towards my bathroom.  \"Alright. Alright, just get the fuck out of my room and I\'ll get up,\" I tell him.  \"Bullshit,\" he replies as he flicks on my overhead light. \"You\'ll go right back to sleep, man.\"  Bluff called.  I throw the covers off to the side in one swift, violent motion and immediately spring myself off the right side of my bed and onto my mom\'s old oriental rug. As I pass by Ackerman, who is still standing in the doorwayâ€”now scrolling through his cell phoneâ€”I flip a quick bird mere inches from his face. He flinches slightly.  \"And don\'t take one of those twenty minute showers either, I\'m hungry,\" he parks at me as I close the bathroom door behind me.  \"Fuck off,\" I retort.  After pissing for what seemed like a minute and taking an economical shower, I reenter my room, throw on some clean duds, toss my cloths from the night before in my closet and head downstairs for my daily 2 glasses of morning OJ. Mike has just arrived and sitting on the couch flipping through channels in the living room.  \"What\'s up, slacker,\" he says, remote in hand.  \"Hunger over, my man.\"  \"Oh yeah. Y\'all get pretty wasted at the bar last night?\"  \"Kinda,\" I respond as I pour glass number 1. \"We were drunk enough at the bar, then we came back here, played Palace and killed two bottles of wine at Steve\'s request. I\'m fucking dehydrated.\"  More clarification. Palace is a game of cards that my friends and I quasi-invented and play quite often. Steve is one of my other roommates. The other, Chris, is the only one who actually had to go to work today.  \"Did Ack tell you about my dad\'s new ride?\" Mike asks.  \"Yeah,\" I say in between gulps of orange juice, \"sounds pretty slow.\"  \"Wait until you ride in this thing, man, it\'s fucking amazing.\"  I didn\'t have to wait long. Ackerman had us out of the house within five minutes, and there it was in the driveway. Beautiful piece of machinery covered in glossy Navy blue with some sort of dorsal fin sticking jutting out of the back part of it.  \"Holly shit, Mikie,\" I offer as jaw drops a little.  \"She\'s a beaute,\" he agrees.  \"What\'s with the fin,\" asks Ackerman.  \"GPS system, man,\" says Mike. \"Al doesn\'t fuck around.\"  \"Shotgun,\" I declare.  So we hop in, back down the driveway and pull onto the curvy, residential street that runs by our neighborhood in Toco Hills.  Mike has always been a very aggressive, if not dangerous, driver and he\'s always been blessed with expensive, fast cars with which to fuel that fire. The car starts beeping.  \"What the hell is that?\" asks Ack.  \"That\'s Rye not putting on his seatbelt. It\'ll keep doing that until he does,\" explains Mike.  \"You\'re shittin\' me,\" I say as I reach for my belt. Those Germans sure mean business.\"  In addition to all the amenities one might expect in a car of this magnitude--leather interior, fancy pop-out cup holders, floor mats, etc--there\'s a fucking TV screen in the dash and a joystick sticking out of the console that controls the GPS, radio and whatever the hell else you can pull up on it. Pretty cool stuff.  The goddamned machine takes turns like a champ. Mike is sure to get it up to speed before stop signs so he can demonstrate its superior stopping ability even at high speeds.  \"Feel the weight being redistributed?\" he asks as we transition smoothly from 45 to a sturdy stop.  \"That\'s fucking awesome,\" offers Ackerman. And I tend to agree. At this point I\'m completely enamored with the vehicle, and we haven\'t even gotten on the highway yet. We continue to wind around back roads before hitting North Druid Hills, which is the road that most conveniently takes you from my house to I-85 South.  \"You ready boys?\" asks Mike.  Oh sure. And why not. I\'ve personally never owned a fast car and for good reason. When I wrecked my last pickup truck in August, my rental was a Poniac Grand Am GT, white. That motherfucker was entirly too fast for me. I do not need that kind of horsepower at my disposal. I tend to develope a lead foot. Not safe for me, or other drivers for that matter.  So we pull onto the onramp and Mikie steps on it. The thing really does purr, I mean I\'m not even a big car guy. It\' rather impressive. We merge over to the HOV lane climbing into the 90\'s. At this point we\'re pretty much blowing the doors off are fellow travelers.  \"Jesus Christ,\" exclaims Ackerman.  We hit 100.  Out the passenger side window is a blur of 18-wheelers, soccer moms in minivans, and other less privileged motorists.  \"You can\'t help but be a little arrogant driving a car like this,\" confesses Mike.  \"A little arrogant?\" I ask sarcastically. We\'re in an airplane fer crissake.\"  We max out at 112 before the traffic becomes too heavy in front of us and Mike is forced to shut it down. Deceleration is so smooth in a that bad boy.  \"How \'bout that, fellas,\" asks Mike.  \"Very nice,\" I agree.  \"I\'d definitely fuck it in the tailpipe,\" confesses Ackerman earnestly.  We hit up the 10th/14th Street exit and work are way through traffic and pull up in front of my old house on Greenwood where my alumni status allows me to park and leave my truck overnight if need be. I get out, Ackerman takes over shotgun, and with that, my joyride is over. The Germans, for all their faults, can build one hell of an automobile, that\'s for sure  We eat a sandwich in the parking lot of Alon\'s Bakery and I follow them back to the house, albeit not closely.  Tuesday, January 20 Q & A-Deuce style I recently came home to find a large Ziploc bag attached to the windshield wiper of my truck. Inside was a note from a young couple new to the neighborhood.  It explained that in their effort to become better acclimated with their new surroundings, they would like for everyone in the Mason Woods Neighborhood Association to fill out an attached questionnaire and submit a picture of everyone in their household.  They, in turn, plan on taking this information and using it to create a comprehensive Mason Woods directory.  Being the resident scribe at â€œThe Deuceâ€â€”our address is 1072â€”I decided to fill in the necessary information. Hereâ€™s how it turned out:  1072 Mason Woods Dr.  â€œThe Deuceâ€   1) Q-Your names, ages, college affiliations and cell phone #\'s   1) A-Chris Prince, 26, Florida State University, 404.386.8962  Ryan Dean Waite, 26, University of Georgia, 404.310.2221  Matt Ackerman, 26, Carolina School of Broadcasting, still uses carrier pigeons  â€˜Deuceâ€™ Alumni  Chef Steve Sharp, 26, Johnson & Wales University  P.F.C. Robert Burns, 25, United States Army   2) Q-Any kids that no longer live at home  2) A-Not that I know of, although it was rumored that Bobby paid for an abortion or two in high school  3) Q-Home phone #  3) A-No house phone  4) Q-Any pets?   4) A-Pricilla, a 5 year-old ball python and whatever varmits can squeeze their way in through that hole behind the fridge  5) Q-When did you move in?   5) We moved into â€˜The Deuceâ€™ in August, 2001â€”just shortly before all hell broke loose.   6) Q-Where are you from originally?  6) A-Chris-Atlanta  Ryan-Leesburg, FL  Matt-Leesville, LA  Steve-Atlanta  Bobby-the UPS guyâ€™s penis  7) Q-What are your vocations?   7) A-Chris-high school English teacher  Ryan-graduate student/contractor  Matt-video/film production  Steve-chef de partie  Bobby-trained assassin   8) Q-Hobbies, interests?  8) A-Chris enjoys The Simpsons, getting stoned and blacking out at bars.  Ryan enjoys good head, racquetball and waving his fist at stupid people in traffic.  Matt enjoys â€œsexing up his girlfriend,â€ Grand Marnier and Coen brothersâ€™ films.  Steve enjoys being stubborn, tennis and athletic sex with Asian women.  Bobby enjoys drinking, fucking and killing.  Collectively, we also throw two huge parties per yearâ€”as many of yâ€™all have surely noticedâ€”including our annual Halloweeen in June party (June-O-Ween) and one helluva post X-Mass bash in which the theme changes yearly. But thereâ€™s always a theme. Thatâ€™s key.   9) Q-Something that might surprise us about you?  9) A-Chris occasionally gets off the couch  Ryan occasionally doesnâ€™t start a fight after a night of bourbon drinking  Matt occasionally doesnâ€™t use Aveda haircare products  Steveâ€™s actually not a very good cook  Bobby is yet to record a confirmed kill   Wednesday, January 21, 2004  The Cabin Classic Okay, so this is all being written post-mortem, sorry. I spent Saturday, Sunday and this morning up in Blue Ridge, Georgia at the second annual Cabin Classic Texas Hold â€˜Em Tournamentâ€”which actually consisted of about 12 separate tournies throughout our stay. Basically 12 of my buddies and I rented a cabin in North GA, provisioned ourselves with what you might expect, got drunk/stoned and played poker tournaments all weekend. I brought pen and paper to record the weekends festivities, however, it never quite made it out of my bag. I did get some good video footage of both nightsâ€”a clear indicator that I was out of a couple of tournaments earlier than I would have likedâ€”and after viewing that this afternoon, I can real off a few high/lowlights. Friday night was rough. The first tournament saw me exit quickly. I went â€˜all-inâ€™ early with two pairâ€”kings and threesâ€”and got burned when my buddy Jesse pulled out a straight on the last card, otherwise known as â€˜the river.â€™ So, seeing as I was the first to drop out and that tournament went on for a good two hours after that, I did quite a bit of whiskey drinking. Tournament #2: same result. Although my friend Davies got knocked out first, I wasnâ€™t too far behind him. This time it was on a much worse hand. I was pretty much itching to take someone down and the first time I pulled a pair of aces on the flop, I went for the gusto. Caplan promptly called me and sent me packing with two pair. Another loss. At this point, my luck turned around. A couple of other guys got knocked out, so we decided to open up our own â€˜real money tableâ€™ while our more fortunate cohorts were still battling it out for the tournament prize.  My good friend and old roommate, Sean Beattie, had gotten there too late to enter into the first tourney and was knocked out of the second one shortly after me. Due to his aggressive nature at the table, I decided to go after him with guns blazing if a truly â€˜unbeatable handâ€™ came my way. It didnâ€™t take too long. About 7 or 8 hands into the thing I was up about 10 bucks over my initial chip draw of 20. Beattie was about the same. I pulled king/eight off the deal and watched him raise two bucks. I knew he had an ace with a good kicker. The flop was gorgeous: king, six, eight. He had to call my $3 raise this time, but didnâ€™t re-raise. The turn was a threeâ€”no help for anybody. I bet $5. Tater, the other remaining player, folded. â€œYou ainâ€™t got shit,â€ said Beattie after staring me down for about 45 seconds. He called. The river was an eight. I had the nuts.  Allow me to explain. In Texas hold â€˜em, youâ€™re playing poker with only two cards in your hand. The other five cards are flipped face up on the table in increments of 3-1-1. The first 3â€”the flop, the 4th cardâ€”the turn, the 5th and finalâ€”the river. Thereâ€™s betting in between each of these cards being flipped, as well as before the first and after the last. By â€˜the nutsâ€™ I mean I had a handâ€”8â€™s over kingsâ€”that I knew he couldnâ€™t beat, unless he also had a king and an 8 which was an astronomical improbability. So, anyway, back to the action. I figured, at best, that the little Cajun bastard had two kings, two pairâ€”kings and either a three or sixâ€”or something else I could beat.  â€œIâ€™m all in,â€ I mumble as I start to count my chips. â€œWatcha got there, Dubya,â€ he asks. â€œ19.75,â€ I respond. â€œIâ€™ll call ya, Dubya,â€ says as he flips his king/three. â€œSorry, brother,â€ I lie as I drop the bad news on the table. â€œOhhhhhhhh,â€ the rest of the table erupts as I scrape what remained of Beattieâ€™s stack. So thatâ€™s poker for you. I lose $20 in two tournaments in a matter of 45 minutes, and it only takes me about 15 minutes at the real money table to make about $50. Gotta love that. Sunday I woke up Sunday morning on the floor with my trusty gray sleeping bag and my green fleece from the night before as my pillow. My friend Flounder was passed out on his half-deflated air mattress to the right of me. It was 11:45. I crawled upstairs towards the smell of coffee. It was waiting there for me along with about 9 of the 13 folks that had come up the night before. I grabbed myself a cup and sat down at the table.  Steever and Blatt were preparing breakfast burritos. Phil sat down next to me and we traded war stories from the night before as we sipped on our mountain grown Folgers. After breakfast, the tournaments got rolling once again. At about 1:30 we came to the unfortunate realization that our satellite TV system didnâ€™t pick up local channels. This meant no AFC or NFC Championship games for us. Of all the fuckinâ€™ luck. The first tournament saw me finish 4th, merely one spot out of the money. The second, 5th. We took a break, re-tapped the keg and a couple of he guys who were only staying one night began to gather their belongings and discuss who was riding with whom. At this point, I got hot. I finished 2nd in a tournament, netting me $20. Then, came the crowning achievement of my weekend. I came back from virtually dead to take down Jesse, on a straight flush, en route to my first all out victory and a $70 payout.  The last tournament of the weekend saw me finish 2nd again and net $20 before cashing the keg and going to bed, this time on my very own double bed. I drooled all over the pillow in my sleep.     Thursday, January 29, 2004  Luke: the Aussie Plastic Fuck Doll My name is Luke. Iâ€™m a plastic fuck doll. G\'Day.  I have a 10 inch penis that has been fully erect as long as I can remember. I live in the basement of a 36 year old woman named Mistress Victoria. We live in two story condo just outside of Sydney.  Mistress Victoria is 5â€™2â€ and weighs about 245 lbs. Believe me, it feels like much more from time to time. She isnâ€™t afraid to really let me have when sheâ€™s on top. Sheâ€™s usually on top. Mostly though, Mistress Victoria is a fine shela. Sheâ€™s very thoughtful and attentive to my plastic fuck doll needs. She always remembers my birthday and sometime even picks me up little gifts just to show she cares. Just like the other day when she brought home a new set of hand carved mahogany clothes pins for my nipples as well as a new bullwhip to stick up my ass.  I must say I could do without the inanimate objects up the ass. Last week she crammed a milk carton up there for crissake. Iâ€™m not familiar with the concept of pain, but that shit just isnâ€™t right.  Sometimes, Mistress Victoria leaves the TV on for me in the basement while sheâ€™s at her data processing job in the city. We have satellite TV, so I get to sit back and she what real people do all over the world. It seems to me that many of you Americans go to court to sue one another over petty squabbles and that most of the judges in the US are black. If I were a real person, Iâ€™d want to be one of those black judges on TV in America. They really seem to have it good over there Even though my nametag says â€œLuke,â€ Mistress Victoria usually refers to me as her â€œbitch.â€ I canâ€™t be certain were the nickname came from, but it sure has stuck.  I often find it odd that Mistress Victoria, being a real person and all, seems to always be asking me, a plastic fuck doll, questions about whatâ€™s going on. Sheâ€™s always asking stuff like, â€œAre you ready for Big Mama, bitch,â€ â€œDo you really think Iâ€™m done with you yet?â€ and â€œHow would you like it if I crammed a tin of Altoids up your bloody bitch ass?â€  Oh how I wish I could answer the Mistressâ€™ questions for her. Sometimes I think if I could, she might stop cramming things up my ass.  Monday, March 08, 2004  Monday 8:00 AM: Get out of Câ€™s bed and take a piss. She is running late for work. 8:03: Get back in bed and watch her get dressed. 8:10: Steal a cigarette on my way out and head directly to school. 8:25: Reroute to Einstein Bagel where at the drive thru am unable to find wallet. Curse myself for stupid, apologize to the lesbian at the window and wheel my truck out onto Monroe in disgust. 8:27: Before reaching the light at Piedmont, realize I have $200 cash in my jeans pocket. Curse myself twice more. 8:35: Roll into school and immediately start cranking out shit for Sylviaâ€™s class. 11:30: Having thrown only a small portion of garbage together on Photoshop, prepare to just draw the rest of it. 11:31-noon: Look all over the place for colored pencils, but just end of having conversations with other students about how we are all lazy and procrastinate like weâ€™re getting paid for it. 12:01 PM: Resign myself to the fact that I wonâ€™t get everything finished on time (1:15), and just fuck off for an hour and 14 minutes 1:15: Leave what little shit is finished in Sylviaâ€™s classroom and return home to slap the rest together. 1:30: Arrive at the Deuce and eat two bowls of Wheaties. Proceed to color like a man possessed.  1:45-3:30: Focused work.  3:45: Unable to find my good sunglasses, I put on my third string pair and bolt with all but one of my comps complete. 3:50: Accept an invitation, en route, to sub for my friend Craigâ€™s softball team at 7:20.  4:00: Back at school, learn Sylvia is running late and that my 4:15 appointment is pushed back an hour. 4:01-5:00: Having already felt the pressure and responded with accelerated focus and effort, feel the need to fuck off and bullshit around school for about an hour. 5:00-5:15: Finish final comp. 5:25: Go in to see Sylvia. The meeting is good. â€œYou really are a turtle, you know,â€ she asks rhetorically. â€œYeah, I guess so,â€ I admit. Hey, she was right, and she didnâ€™t even say it in a mean way. 5:45: Get the fuck out. 5:46-7:00 Grab a half a Publix sub and chug over to Blackburn Park in Dunwoody. 7:30-8:30: Go 0-3 (FC, 6-3, F-8) with a run scored. Our team gets killed, 19-7. 8:35: Head home 9:00-9:15: Make a peanut butter and jelly sandwich and plop down in front of the idiot box just in time to watch Halle Berry get absolutely rocked by Billy Bob Thorton in Mosterâ€™s Ball.  9:30-10:00: Watch a bad episode of South Park. 10:01: Broha hits the sack. 10:02-10:25: Talk to C about her/my day and how much she hates her job. 10:30-11:15: Ack gets home, we smoke two bowls. 11:16: Ack crashes 11:17-11:45: Clean out school folder and make to do list for Tuesday (mostly bullshit insurance bills) 11:46: The Steve arrives, I get him high and he bums me a cigarette. 12:00-1:00 AM: Download music and put together this inspiring little chronology.  1:01: Go to bed.   Posted by: ryan / 10:01 PM   Sunday, March 07, 2004  Confessions of a disgraced insomniac Pulled another one of my quasi-famous all-nighters last night. A very productive one, too, might I add. Okay hereâ€™s the drill, I somehow conned my landlord into allowing me to rip the nasty, corroded, mildew-laden wallpaper of my bathroom walls and replace it with a couple of fresh coats of primer and semi-gloss. Hereâ€™s the kicker, sheâ€™s knocking $300 off my rent for the job. Yesssir. So, I decide to do the whole fucking job last night. Hereâ€™s the boring play-by-play.   Thursday Night  9:30- Get out of copywriting class 9:31-10:45- Fuck around with Photoshop comps for Sylviaâ€™s class. Two computers freeze on me and I leave in disgust. 11:00- Arrive at home; Chris is only other roommate present and is already asleep. 11:01-As I begin to tear the moldy shit off the wall, I notice that Iâ€™ve been showering in a bathroom with canary yellow wallpaper prominently featuring flowers in pots for going on a year now and am horrified. 12:00 Midnight- Matt gets home and we smoke two bowls 12:15-12:30- Stoned now, I slowly drudge through fifteen minutes of fruitless labor before our friend C.Dubya rolls by. We smoke two more bowls. 1:15- With all the paper down, I make a few minor repairs with dry wall mud.  1:45- Matt goes to bed, I make a cup of coffee and start taping the tile up for the primer stage. I forget to ask him for a nug to puff on periodically throughout the night, I decision I come to regret. 2:30- Priming begins; one layer is completely on there in less than 45 minutes 3:15- Steve comes home jacked up on cocaine. He proceeds to complain about the slack asses that work for him at his restaurant, then walks around the house mumbling to himself about not wanting to be â€œthis fucked up.â€ 3:40- Steveâ€™s muddles off to bed 4:00- The shitâ€™s still too wet for the second coat. Bring large fan into bathroom for circulation purposes. 4:03- Scrape up some â€˜ash tray potâ€™ and smoke a nasty excuse for a bowl. 4:30 second coat of primer goes on in 30 minutes. 5:00- clean up, putz around, lift weights for about 15 minutes straight, stretch back. 5:45- leave for Waffle House. 6:00- there out of chicken, so I order a patty melt plate, no onions, steamed hashbrowns, a waffle, three cups of coffee and four waters. I buy a paper and read about how old and religious Julio Franco is 6:45- Foolishly go up to school is this delirious state with the naÃ¯ve notion that I could actually get some work done. 7:45- After fucking around on Internet and getting virtually nothing worthwhile done, leave. The security guard is dead asleep both times I pass him. 8:00- Get back to the house just in time to catch Matt on his way out. We smoke a bowl. 8:15- First layer of paint hits the walls.  8:40-Done. Make another pot of coffee. 10:00-After watching Sportscenter, eating a bowl of cereal and jerking off, I start the second coat. This is done in about 20. 10:20 Start final clean-up preparations, as I am satisfied with my work. 11:00 Steve wakes up and calls me crazy for staying up all night. He is hung over from drinking and from doing the Bolivian marching powder. 11:15 Steve and I watch the last 20 minutes of Seabiscuit. 11:45-3:30-After 25 minutes of taking the tape down, putting the supplies back in and around the tub and taking a moment to marvel at my work, I basically fuck around, make some calls, sit in the hammock, and download music until I can no longer keep my eyes open. At 3:30, Broha comes home. 3:31- I plop my self down on the couch and slip in and out of consciousness for three hours; periodically batting an eye at my phone, which seems to ring about 38 times. Now- Iâ€™m up, Iâ€™m showered and Iâ€™m driving to the barâ€¦â€¦   '),(9414,3659024,'21,June,2004','2004-06-21 00:00:00','    Football (or Soccer to the uncouth north Americans)  England\'s playing Croatia as I sit here at work, England lead 4-2 and might actually win a vital game for a change. I am sure all my colleagues think I am in sane, I sit here groaning and sighing, cheering (not too loudly) at every chance missed and foul, every close call.  I don\'t follow football all the time (rugby\'s my game) but this is England, my home my land. I don\'t know how can explain it any better. I am sure others love their homeland too, but there seams to be something about England that green and pleasant land. An Englishmans belief that God lives there, that almost certain knowledge that whatever happens England will be there.  I know many who might read this will no doubt disagree or even cuss at me. But these last fifty years something has remained dormant, sleeping but if it ever awakes then the world I fear will be shocked. English nationalism.... Many think England a spent force, an empire in decline or gone.  When the English awake, with dreams of empire, then all should worry, all should tremble, for these peace loving English can still bring to bear that mighty sword, bringing down all, be it Hilter, be it Neolpeoan .... England did bring them low in their hour of victory.   I know the above may be pompous but remember well June 18th 1815..... Come on Ye noblest English...   '),(9415,3659024,'17,June,2004','2004-06-17 00:00:00','    Feeling sorry for myself today....  Got a property tax bill and a car insurance bill yesterday some $7500 combined. They really know how to sick it to ya!  On another note... The Canadian elections are getting closer and I get to look at all the parties as an outsider. I can\'t vote (not a citizen yet), but I read the papers, watch the TV and have to say that Canada has some of the most incompetent politicians in the world. More concerned with telling you what the other canidates are going to do rather than there own policies.  I watched the televised debate on Tuesday evening, I have never seen such a rude bunch of people, this was not a debate be a yelling match.....  Well Goto go...   '),(9416,3659024,'16,June,2004','2004-06-16 00:00:00','    Has anyone ever noticed that the harder you are forced to work, you achieve the least. But the converse is not true, when you are not pushed to deliver, then you don\'t.  The thing is, is there a happy medium?  Another thing how do you describe a multi node supply chain network in terms sinks and sources, that makes sense to someone who can only think in terms of warehouse and store   '),(9417,3659024,'16,June,2004','2004-06-16 00:00:00','    I really don\'t what a BLOG is.... But everyone seams to be creating one, so as the band wagon rolls along I decided to jump on.  Just to introduce myself...  I am recent arrival in Canada from the UK. I work for a medium sized food retailer in western Canada as a systems analyst.  My hobbies include wasting time, drinking beer (or wine if there is no beer available), Reading just about anything (I\'ve been known to read cereal boxes when I can\'t find anything else). Oh and annoying people on the internet by publishing a random collection of thoughts.   '),(9418,3757297,'27,June,2004','2004-06-27 00:00:00','   I subscribed to a newsletter when I was pregnant with my 2nd child.  After 40 weeks, I thought the newsletters would just stop, considering that most pregnancies are around 40 weeks.  They did not.  I got week 41 and I thought, â€œwell, some people are up to 2 weeks late, so maybe that is why I am still getting them?â€  However, when I got the newsletter for week 44, I knew there was a problem.  After a few weeks of messages with titles like â€œHello, you are now 50 weeks pregnant.  How do you feel?â€ I decided to try to â€œunsubscribe.â€  So I clicked on the â€œunsubscribeâ€ button only to find that they will require my password in order to stop sending me ridiculous emails.  Well it has been approximately 40 weeks since I signed up for this newsletter initially, so not only did I not remember the password, I did not even know my username!  Well of course you need your user name to access your password on the odd chance that in 40 weeks you may have forgotten it.  So I am basically doomed to receive these emails until I die or change my email address, whichever happens first.  Every time I get another email I just want to scream â€œI am apparently pregnant with an almost 2 year old.  How the fuck do you think I feel?!â€  '),(9419,3757297,'26,June,2004','2004-06-26 00:00:00','   To start out this log or blog, I will list a few things that I do not understand.  Maybe someone else understands these things and can clue me in?   *Pop up adds for software to eliminate Pop ups. Isnâ€™t this blackmail??? Basically they are telling you that you have to buy their product to get rid of a problem they are causing or else you will continue to be harassed. This would be like some guy walking up to you and hitting you repeatedly until you agreed to buy his service as a bodyguard to protect you from people hitting youâ€¦  *The auto-glass people calling me to ask if I have a cracked or dinged windshield that I want repaired. If I wanted their services, I would call them. Do they believe that there are people who desperately want their services, but for some reason canâ€™t get a-hold of them? As if someone is sitting at home and gets their call and exclaims, â€œOh thank goodness you called!!! I have been trying to reach you for days now, but could not get through!â€ (Probably because the phone line was busy due to the fact that they were calling everyone in the city just to make sure nobody wanted to talk to them.) Other businesses do not do thisâ€¦I never get a call from McDonaldâ€™s asking if I am hungry and wanted to buy a hamburger from them.  *Maxi pads designed for thong underwearâ€¦no explanation requiredâ€¦  *People who say â€œno offenseâ€ right before blurting something completely rude, as though â€œno offenseâ€ is a qualifier that prohibits you from becoming upset at their commentâ€¦i.e. â€œNo offense, but you look like you have gained a lot of weightâ€¦â€  *7 day return policiesâ€¦Are these establishments worried that their products will not hold up past 7 days?  *How doctorâ€™s offices will refuse to see you and often charge you a fee if you are more than 15 minutes late for an appointment or do not show up or cancel without giving 24 hours notice, but will leave you sitting in the waiting room for 3 hours when you arrived early, and not offer you any discount on your bill for the inconvenience of the wait!  '),(9420,3757297,'26,June,2004','2004-06-26 00:00:00','   Interesting...I\'ll get started tomorrow...  '),(9421,3757297,'23,July,2004','2004-07-23 00:00:00','   Well I have exhausted (nice word choice I think) every possible option for natural sleep aids, and am now just giving up. Like an idiot I drank coffee with dinner, although I could not sleep last night either, and had no coffee then.I have spent the last 4 or so hours surfing television channels in search of quality programming (*smirk*) and it is just not there.   I started out watching Disney channelâ€™s â€œBug Juiceâ€ which is like MTVâ€™s â€œThe Real Worldâ€ reality show, but for pre-teens:) It is basically a show about pre-teen and teen kids at camp.  Then, I started watching old reruns of â€œL.A. Lawâ€ on A&E. Not a bad show, although the episode p*ssed me off because a man whose father was a diplomat violently raped a girl and then was set free due to â€œdiplomatic immunity!â€ So now I have one more thing to be politically p*ssed aboutâ€¦  After that I started watching an infomercial about a tape that guarantees flat abs in only 6 weeks. LOL, if I had an extra $60 I might consider it! I am so tired of my little pooch! Something tells me though that there is fine print (so fine that I can not see it without a magnifying glass) that reads something like â€œResults are not typical. Likely the only weight you will loose is the 1 lb. From the $60 weighing down your wallet! This means you H!â€ After looking at about 300 obviously touched up â€œafterâ€ photos, I switched the channelâ€¦  Then there was the infomercial for the â€œamazing mattress!â€ This amazing mattress was supposed to form to your body shape. (I usually would describe a mattress that did this as â€œold and worn out,â€ but that just does not have the same ring to it as â€œamazing!â€) Also, it was supposed to stay still while your partner moved. I guess the idea was that you would not wake up your loved one. The interesting part was the way they demonstrated it though. Here was the setup:  A 25 year old, blonde, 95 lb., Cheerleader type girl stands at one end of the bed in a semi-short red negligee. A wine glass half full of water is placed at the other end. The clueless blonde starts jumping up and down on her corner of the bed while an official sounding male voice commentates, â€œLook how the glass of water does not spill as Candy moves about!!!â€ Yeahâ€¦I am sure there are a great many people out there looking at the WATER;) I am not sure if they proved any point with the demonstration, but I am sure that many a man watching was happy that the official voice was able to make Candy â€œmove about.â€ I am sure many men readily began practicing their official voices in hopes that they too could get someone to â€œmove aboutâ€ for a harmless experiment:)  When you start searching the channels for a better infomercial, you know that you have hit an all time low.  So now I sit here, too tired to clean, too tired to read, too tired to write anything worth anything, and worst of all, too tired to sleep. I guess I am back to bed to stare at the black ceiling again, or to make shapes in my mind from the lit numbers of the alarm clock, or maybe I should \"move about\" on my corner of the bed;p  '),(9422,3757297,'07,July,2004','2004-07-07 00:00:00','   I wish that candidates would debate the issues and not engage in so many personal attacks.  I wish that there was a \"perfect\" mix of liberalism and conservatism in a candidate that actually had a chance of winning in our 2 party system:)  I wish that people would stop saying \"grass roots.\"  I wish that fat free mayonaise tasted more like actual food.  I wish my Mother In Law were sentient.  I wish that my husband had the day off work:)  I wish Law and Order were on every second of every day...7 times a day is not really enough for me.  I wish I knew Kim Possible.  I wish for more wishes.  '),(9423,3757297,'01,July,2004','2004-07-01 00:00:00','   What is the world coming to!?  It is *only* 3:30 am and there is nothing on TV!;)  Basically, the only thing still on is Profiler (and similar shows), Infomercials, and advertisements for sleeping aids!  I donâ€™t like watching ads for sleep aids while I am awake at night.  I realize that they market them at night to catch an intended audience, but people who have trouble sleeping watch TV in the day time too.  Perhaps if I saw the commercials earlier in the day I would have a chance at obtaining the products so that I could sleep at night instead of watching their ad.  However, with their ads airing at night, I have little chance of obtaining the product after seeing the ad.  It feels like teasing to me, like when I see a Pizza commercial after the restaurant has already closed;)  Infomercials are bad enough without additional commentary LOL, so I wonâ€™t touch much on this one.  The fact is that it is both boring and disturbing to watch someone cut cans with knives (as though this is a normal activity of knife owners), mop ceilings with extended handled mops, and build muscles on contraptions that look as though they were designed by the military for torture.  The main problem with shows like Profiler, and other intense crime shows is that at 3:30 am, they scare the shit out of me!  I can barely watch these programs at 3:30 in the afternoon without constantly checking behind me!  So these are the choices I am left withâ€¦Of course I do have a DVD player and a VCR which I could utilizeâ€¦but again it is 3:30 amâ€¦and I am not thinking clearly;)  '),(9424,3829063,'21,July,2004','2004-07-21 00:00:00','   in my \"hot or not\" post  below, i listed \"the princess bride\" in my hot column.......tom g responded to this with the following comment:  \"i thought i was the only one who thought that girl from the princess bride was hot.\"    well tom g, i gotta break it to u, u r the only one who thinks that girl from the princess bride is hot... my inclusion of princess bride in the hot list referred to the movie as a whole, its a great movie, from inigo montoya to the posioned challace scene... the girl , however, is butt ugly... tom g\'s dirty little secret has been EXPOSED!!!...apparently amir is not the only guy who likes pale white girls....what would rani think tom g '),(9425,3829063,'15,July,2004','2004-07-15 00:00:00','   HOT FARENHEIT 9/11,\"turn me on\" by kevin lyttle, i love the 90s on vh1, american casino, ninya off american casino, the nina sky video, girls with bangin bodies at the gym, my apartment, vanessa from TRL, the return of the arm band, my muslim league bball team beatin everyone by 30 points,  my dad tellin a telemarketer to take his mother to florida after he said my dad had won a free vacation to miami, the george lopez show, woody paige, my denver nugget shorts,  how big my muscles look in certain lighting (i almost turn myself on),  the princess bride, the sandlot, my good times dvd, the white girl from the \"i like the way u move\" video, shana hyatt,  my wr skills, the thought of nihar goin to work carryin a briefcase full of crackers, afghan weddings    NOT  baller bands (this item is listed in tom g\'s hot list, and yes, i had one for a week or so, but now every fucker in the world wears one, thus, they r now on the NOT list)    the fact that i cant wear arm bands anymore since every faggot white boy in the world wears them now (led by carson fuckin daly)even though i started that shit in the 9th grade    casino on fox, sikhs and punjabis, pakis and indians who think they can play ball, paki and indian girls, paki and indian ppl who wear colored contacts and think it looks natural against their burnt dark ass skin, white girls except for tan brunettes,  losin with 3 aces,  the girl at the gym with the incredible body but with a face of a horse (shes like a character from greek mythology),  max kellerman leavin around the horn, jay mariotti, jay mariotti\'s eyebrows, tim cowlishaw\'s conservative ass,   ashiks excuses for bein blown out by 30 points after he ran his mouth for a week thinkin the game had been cancelled '),(9426,3829063,'08,July,2004','2004-07-08 00:00:00','   some comments and confessions:  i heard nihar was given the penske file.....but i doubt he\'s penske material  \"lost in emotion\" by lisa lisa and the cult jam might be the greatest song of all time  even though shes not muslim or a virgin, and has a korean baby, id marry angelina jolie right now....or just have sex with her  the only thing possibly funnier than amir dancing is seein nihar at prom........aaah, who am i kidding, theres nothing funnier than seeing amir dance  one of these days, im gonna order a large pepperoni pizza....and id like to see u stop me  im just not attracted to black girls....just cant do it...i think its the hair  i cant find blondes attractive either....they look too much like bologne.....  its so hard to talk to white girls now, they seem so damn ignorant of the world.....  i hate chinese girls too.....their breath always smells like shrimp  things in real life dont happen like they do in the movies...i once put ex-lax in someones food...they didnt jump up and run to the bathroom....it takes 4-6 hours for the stuff to kick in  i dont know what the hell a guju is....but it sounds like something u buy at the snack counter at the movies....and it probably taste like doo-doo  lynda carter at 50 is still hot as hell....so is mary ann from giligans island  no matter how hot an indian girl is, hearin that fuckin accent makes my balls retract.....think about that before u marry rani tom g.....about the accent, not my balls retracting  how did monica from friends all of a sudden become a lot hotter than rachel...  why did i used to think sandra bullock was so hot???  \"informer\" is a very catchy song.... '),(9427,3829063,'05,July,2004','2004-07-05 00:00:00','   here r asaads comments: man, i should be one of the cool bengalis; i bring so much of the humor at our get togethers...i thought u were fan...seems like ur just a deceptive faker...and btw, cricket is better than that one afghan sport with the goat-head ball.  someone responded to his comment: asaad, jesus christ, why are you guys obsessed with your colonizer\'s sport. don\'t you have any ethnic bengali games? --naaz  my answer: sure they do, it involves ordering the cheapest thing on the menu then mooching food off ur friends to save a buck....cheap bastards '),(9428,3829063,'04,July,2004','2004-07-04 00:00:00','   chinese:  i saw some light beer commercial that depicted 4 ppl doing \"power blinking\" in order to burn the low amount of carbs in that beer...one of the 4 ppl was a chinese woman....i wondered to myself, does she have to do more reps than the others since shes only doin half the work???  iranians:  why do they all sell used cars?? u can be talking to an iranian medical doctor with 3 phD\'s and a seat in the iranian government, and the second u mention ur lookin to buy a car, the buttons on their shirt pops off, thick, black chest hair sprouts out, gold necklaces appear,their hair slicks back, and they notify u that they sell used cars in the spare time.    indians:  they just stink...and got crusty lips....and ashy elbows....and some of them have purple lips and gums like my dogs...most r harmless though, theyd rather watch gay bollywood movies than fight for their country....yeaaaa, let the british come shit on u even though u outnumber them 132352343 to 1, theyll still let u watch ur bollywood.....and whats up with indian ppl who say theyre british when u ask them where theyre from...no bitch, ur indian, ur a fuckin patel, ur not british..... (excluding nihar and his family)  pakis: see above....pakis have the same base smell as the indians, except they incorporate a hint of body odor as opposed to the nuance of curry that the indians mix with that same base odor...pakis r also cheap bastards.....they bitch and moan that ppl should be more islamic and that the muslims should help eachother out, yet, when bush waggles a couple dollars in their face, they bend over and spread wide....they remind me of those little flying monkeys from the wizard of oz....short, dark, monkeys....  bengalis: i like tom g  jewish ppl: oh  man, where do i start....all ill say is that iv never seen a group of ppl who run the fuckin world always try to act like the victims in every fuckin situation  arabs: the cause of all the problems in the middle east....theyd fuck a chicken if it had blonde hair..... '),(9429,3829063,'03,July,2004','2004-07-03 00:00:00','   ????: can u believe that fat bitch asked me for condoms once ????: what the hell is her fat ass gonna do with condoms kobe: she probably fills them up with pudding and hangs them around her belt for a convenient carrying snack  thought after seein saddam on trial today: aaaah, i get it, america unjustly invades iraq to catch saddam and charge him with unjustly invading kuwait...u cant fuck with bush logic '),(9430,3829063,'03,July,2004','2004-07-03 00:00:00','   ok bitches, time for the dream to finally put it down.  i know what u homos r thinking, creating a blog is very un-samim like.  yea, i know, but i was struck by inspiration.  what was the source of this inspiration? the answer may surprise u.  as i did my routine check of away messages on my buddy list, i ran across our old friend asaad....in his profile, he had a link that was titled, \"my inner soul.\"  now, for those of u who know asaad, i do not have to explain the uncontrollable curiousity and temptation to click that link... for those of u who do not know this bengali wonder, this is the guy who went to a restaurant and asked for water with no ice, explaining that without the ice, u get more water....now u can imagine the thoughts that ran thru my mind when i wondered what could possibly represent asaads inner soul...a few things crossed my mind: big black womens\' asses, asaad crying when his father broke all his 3-6 mafia cd\'s, gremlins, a stuffed adnan doll he goes to bed with....the possibilities were endless...so i click it, and its a link to his blog where he posts results of bengali cricket matches, because, of course, we all wait in anticipation to find out how many wickets some bengali jackass knocked over...so i figure i needed my own area to say whats on my mind... over the yrs, i have made u laugh, made u cry, made fun of u, started fights, and stirred up much contraversy with my away messages, but away messages r just short, fleeting samples of my hate....i figure with med school startin in about a month, i wont have time anymore to air my greivances with u ppl thru my away messages..this page will keep the hate i spew alive for all time..... '),(9431,3829063,'08,August,2004','2004-08-08 00:00:00','   i officially knew i was in memphis when i turned on abc last night and instead of \"friends\" they were playing \"martin\"  celebrity sightings from my first night in memphis: the funky lookin dude who married liza manelli, i dont know what the hell his name is but he looks like a plastic figurine....and he always wears those sunshades, he was wearin them last night  some guys from the world series of poker...they all looked like dicks, but had the hottest bitches with them.....apparently chicks dig dudes with money, who would have thunk it....  and of course, no night in memphis is complete without spotting a member of the 3-6 mafia....i didnt recognize him since half the ppl in memphis look like theyre in the 3-6 mafia, however, having a jersey with the number \"666\" and the name \"lord infamous\" across the back kind of  gave him away....  apparently the \"finish line\" in memphis also has the ridiculous sale rack in the back.....jacket for 5 bucks? thats amazing......tom g and nihar dont know what theyre missing  theres a \"church\'s\" chicken on every block....for obvious reasons....  i do my grocery shopping in arkansas?......  germantown puts me at ease......  the indians here r white as hell......their personalities i mean, not their skin color '),(9432,3986767,'20,July,2004','2004-07-20 00:00:00','    As a sedentary Computer Engineer / Programmer that has spent too many years behind a desk, I found at 36 I was diagnosed with Type 2 Diabetes. Itâ€™s a common enough illness that unfortunately is not treated with the respect it deserves. Killing a full half of it victims by the insidiously slow attrition of pulmonary and vascular health, it leaves the vast majority of those remaining with associated long-term complications that affect almost every part of the body. Diabetes contributes to blindness, heart disease, strokes, kidney failure, amputations, and nerve damage. Not something to take casually but interestingly enough, I meet diabetics all of the time who simply neglect proper treatment because there is no immediate need to worry about it. Diabetes kills over the course of months and years, not days and weeks.  Personally, I took the diagnosis seriously; it was a real wake up call. Frankly it scared the hell out of me. I was 36 years old, an amazing 80 lbs over weight with outrageously high cholesterol and triglyceride levels. As my Doctor put it, I was â€œWalking towards the light, a prime candidate for pulmonary diseaseâ€. All I could do was shake my head and wonder how I had let this happen. I really didnâ€™t know. I had ALWAYS been in excellent physical shape but somehow I allowed my life to be consumed by my work, and day after day of sitting in a cube, writing software and managing computers (my occupation) had lead to lifestyle that was so out of balance that my body simply stopped functioning correctly from the neglect it received.  The day I was diagnosed was the day I started to make positive changes in my lifestyle. I immediately stopped working 60-70 hour work weeks. I started to watch what I put into my body and spend time at the gym. Over the course time I began to see results, lowered blood sugar levels, more energy and better overall health. I was pleased with the progress I had made but still very disappointed in myself for allowing this to happen in the first place. I was also having a really hard time losing weight and just didnâ€™t feel as healthy as I thought I should given all of the work I was doing. I felt and still feel that I owed it to myself to makeup for the years of mediocre to downright poor health and physical neglect. We only live once after all, so why take the cheap seats.  How I going to fix this? I choose to make it a point to get into the best possible physical shape I could and at the same time attack the other areas of my life that I had allowed slip into disarray. Donâ€™t get me wrong, my life was not an absolute mess; it was just a mediocre mess, a life half lived because I forgot to bring the INTENSITY to the table that was necessary to compel me to work hard and be disciplined and successful with EVERYTHING in my life.  First I did a little personal reconnaissance, assessing my diet / exercise habits and talking with my Doctor about it.  She suggested that.  1) I had likely screwed up and slowed down my metabolism by eating too few calories for too long; making my body think that food was scarce. Once again, a balance issue.  2) There is a rather aggravating relationship between Diabetes, Insulin and weight loss, the scope of which is too in depth to cover here and Iâ€™m not sure I fully understand it. It is enough to know that it makes losing weight MUCH harder than one might think and I was in this endless cycle that diet, medication and exercise were failing to control.  Nice, really nice. It was about this time that I started to loose my hair, mainly from pulling it out in frustration.  As a kid I was exposed to an â€œOld Schoolâ€ Tae Kwon Do and Hapkido Do Jang. The people who ran the Do Jang could hardly speak English, all of them Korean and Vietnam War veterans. They were deadly serious about the way they practiced because to them it was a â€œMartial Systemâ€ they had employed professionally, not something â€œto doâ€ on Monday night. I had been told that their system was a derivative of Japanese Karate and an ancient Korean Martial System (or group of â€œKwansâ€) named Hwarangdo, similar to Bushido. These people made a huge impact on the way I view life and as an adult, when I started to look for a more complete and balanced exercise regimen I thought of them, starting a 4 month long search to find a Do Jang that personified that â€œOld Schoolâ€ martial spirit. I quickly found that what I was looking for was lost to a world of commercialism. I donâ€™t mean to be critical of the typical martial arts school, itâ€™s not my place nor do I possess the technical expertise necessary to offer a meaningful critique, but what I saw in school after school was nothing like what I was exposed to as a kid. The most obvious difference aside from the red and blue foot pads and head gear was the lack of seriousness in how they trained. The Koreans I had met didnâ€™t talk about work or chicks or the weather when they trained; they just trained and trained hard.  Things change.  I started to look at other martial systems and in most cases found the same thing. After a couple of months I was getting a little discouraged. As a fat guy, it was sort of like looking for clothes at Wal-Mart, nothing fit right.  In early August I decided to checkout Jujitsu. The first school was an hour away AND â€œsports orientedâ€, I quickly decided it was not my style; the second was a Tae Kwon Do school that also taught Jujitsu, too many red and blue foot pads, again not my style. Then I stumbled upon the Florida Jukido Jujitsu Academy. Amazingly enough, just a short drive away and the first conversation I had with the Sensei was very low key. He didnâ€™t try to get me in to sign a contract, he didnâ€™t talk a bunch of smack about his school or â€œstyleâ€, and as a matter of fact he asked that I wait a couple of weeks to come to class. The school was attending a seminar and would not have time to spend with a new student until they got back.  From the first night as an observer I knew I had found the right place. I quickly realized that Sensei Rego didnâ€™t hype his Jujitsu style because he didnâ€™t have to, these guys were INTENSE, very INTENSE and Jukido spoke for itself. It was all there, discipline, tradition, stunning technique. I knew I was home.    After a rather extensive interview with the Sensei I was invited to join the Dojo and have spent the last 11 months working as hard at Jukido Jujitsu as my body will allow me to.   I hope that this Blog will afford me the opportunity to share my journey with this beautiful Martial Art as well as some of my observations regarding the impact that Jukido Jujitsu has had on me personally. I think I may also be able to provide some constructive advice regarding phisicle conditioning to others who are in less than stellar condition and are considering adopting a hard core Martial Art on as a lifestyle.   Never look to this Blog as a technical reference for Jukido Jujitsu or any other dynamic of an actual process involved in performing specific techniques. I am very wrong as often as I am right when giving advice and if you have a Sensei you should go to him for specific technical questions. If you donâ€™t have a Sensei and would like to know more about Jukido Jujitsu or its Sister art Kokondo Karate let me recommend that you either approach the Main Website for this Martial System  urlLink http://www.ikja.com  or the website for the Dojo that I attend at  urlLink http://www.floridajukido.com.   !OSSU!   '),(9433,2308025,'28,November,2003','2003-11-28 00:00:00','     '),(9434,2308025,'27,November,2003','2003-11-27 00:00:00','     '),(9435,2308025,'27,November,2003','2003-11-27 00:00:00','     '),(9436,2308025,'27,November,2003','2003-11-27 00:00:00','   Using mathematics to explain everything.  '),(9437,2308025,'26,November,2003','2003-11-26 00:00:00','     '),(9438,2308025,'26,November,2003','2003-11-26 00:00:00','   Anyone (not just BlogTeam members) can leave comments in response to any of the main posts directly from this page.  Just click on the Comments link below the relevant post, fill in your name and leave a message.  Primitive, but simple.  Enjoy!  '),(9439,2308025,'26,November,2003','2003-11-26 00:00:00','   A shocking photo of Douglas \"Doug\" McConnell has appeared over the internet giving a frightening insight into the Stadium Rock Dreams of the Be-Spectacled Self-Proclaimed Soft Rock Saviour.  Rick Parfitt is said to be beside himself.  '),(9440,2308025,'25,November,2003','2003-11-25 00:00:00','     '),(9441,2308025,'25,November,2003','2003-11-25 00:00:00','     '),(9442,2308025,'25,November,2003','2003-11-25 00:00:00','   I\'ve Made It!!!  '),(9443,2308025,'25,November,2003','2003-11-25 00:00:00','   Credit to my Dad for finding this.  '),(9444,2308025,'25,November,2003','2003-11-25 00:00:00','   When I\'m hungry, only one eatery can satisfy my needs:  '),(9445,2308025,'25,November,2003','2003-11-25 00:00:00','   He\'s coming home, he\'s coming home, he\'s coming, Panda\'s coming home!    \"Being back in the UK makes me a sad panda.\"  '),(9446,2308025,'25,November,2003','2003-11-25 00:00:00','   Find yourself wondering what Guy and Scott get up to all day? Finally the answer.  All you really need to know about panda\'s summed up in one flash animation.  Sort it out boys!   urlLink http://www.thebigsmoker.co.uk/pandalove.html  '),(9447,2308025,'23,November,2003','2003-11-23 00:00:00','   Just saw  urlLink this  on Slashdot - apparently some American Biotech company has created a fish that glows red in the dark.  All very scary for various reasons, but the fish itself looks  urlLink quite pretty , in an eerie unnatural sort of way.  '),(9448,2308025,'22,November,2003','2003-11-22 00:00:00','   As promised, here\'s a picture from the Oilers vs Flames game I was at last weekend.  Unfortunately, I was too busy cheering when the fights were going on to take photos of them but this one was taken just after the Oilers scored the winner in overtime.     It\'s bloody freezing here at the moment (-20 degC plus wind chill) so it\'s just as well the nearest off-licence is just across the road from the hotel!  '),(9449,2308025,'21,November,2003','2003-11-21 00:00:00','   Sounds like Graham\'s dirty past is catching up with him and the Canadian government do not approve.  Oh well, see you on Thursday mate.  For webspace, this blog has been upgraded (free of charge) to a Blog*Spot Plus account, which provides us with 25Mb of space to fill.  There should be a new \"Upload File\" button on the \"New Post\" screen in Blogger.  Just ensure that pictures are:  1) Sensibly trimmed, i.e. no more than 500x400 pixels. 2) Compressed in a web-friendly format, .GIFs and .JPGs work best.  The picture of \"Bob\" below gives and idea of the maximum size that works well over a dial-up connection.  If in doubt, just post it and I\'ll trim it later via FTP.  '),(9450,2308025,'21,November,2003','2003-11-21 00:00:00','   Just to update everyone - I\'m not being allowed to leave Canada and have been placed under house arrest with armed guards at every door.  But with a bit of luck (not to mention my history of successful covert ops) I should manage to escape back to Aberdeen at some point on Thursday.  Oh and if someone could give me access to their web space I\'ll post my pictures from over here.  As it is I\'m too lazy to arrange doing this myself.  '),(9451,2308025,'20,November,2003','2003-11-20 00:00:00','   Bob deserves a special mention as he has a had a long hard fight but come out the other side as (say it together) A WINNER. So, Well Done Bob!!!  '),(9452,2308025,'20,November,2003','2003-11-20 00:00:00','     '),(9453,2308025,'20,November,2003','2003-11-20 00:00:00','   \'nuff said.  '),(9454,2308025,'19,November,2003','2003-11-19 00:00:00','   I know JB is having some problems posting messages on this so I\'ll do it for him.  Me, JB and Stu got fucking monged last night (2 for 1 drinks at the \'nage - don\'t order 9 drinks - THEY GIVE YOU 18!!!).  Anyway, I was drunk in work this morning.  John vomited in the shower and I dunno what happened to Mr King. Jesus, I feel like ass.  '),(9455,2308025,'19,November,2003','2003-11-19 00:00:00','   In light of recent events, the colour of this blog will  have  to be changed.   Complete and utter spaz.  '),(9456,2308025,'19,November,2003','2003-11-19 00:00:00','   At last! A detailed analysis of the airspeed velocity of an  urlLink unladen swallow .  '),(9457,2308025,'18,November,2003','2003-11-18 00:00:00','   T-shirts are passÃ©, we need Panda Attack  urlLink BBQ Aprons ! Nice.  '),(9458,2308025,'18,November,2003','2003-11-18 00:00:00','   Lets not push the panda too far.  It could go horribly wrong.     I think that is brilliant.  Can we get t-shirts.....can we?..........please.  Check them out  urlLink here   The boxers are my personal favourite.  that is where you need the warning!!!  '),(9459,2308025,'17,November,2003','2003-11-17 00:00:00','   HOW CAN WE JOIN??!  '),(9460,2308025,'17,November,2003','2003-11-17 00:00:00','   Hello All.  Thank Christ......I am no longer on the west coast.  what an almighty waste of time it was!  Anyway, I\'m back in Aberdeen and I look forward to seeing you all (esp the new panda) at 7pm in the Caley Thistle tomorrow.  I\'m gonna find a photo to put up too!!     boudy getting some love?  '),(9461,2308025,'17,November,2003','2003-11-17 00:00:00','   A Big Gay Panda.  He\'s also a sexual deviant who can\'t handle his Martinis for toffee.  '),(9462,2308025,'17,November,2003','2003-11-17 00:00:00','   Greetings from Edmonton!  Just contributing what will no doubt be the first of very few posts.  If I get a chance later I\'ll stick up some of the pictures from the (ice) hockey game I was at over the weekend.  '),(9463,2308025,'17,November,2003','2003-11-17 00:00:00','   For the benefit of family members struggling with what to get me for Xmas, my Amazon.co.uk Wish List is  urlLink here .  '),(9464,2308025,'17,November,2003','2003-11-17 00:00:00','   I have just invited Graham, Cormac and John to join this blog in the hope that it might be fun.  Graham is in Canada just now, so I don\'t expect to hear from him until he gets back. Cormac is bored on the West Coast of Scotland, so he should have time to contribute some humor. As for John, this site should provide a much needed vent for his intense sexual frustration.  '),(9465,2308025,'12,November,2003','2003-11-12 00:00:00','   Only just getting time to play around with this wonderful shiny new weblog. Thought I\'d try to link an image first, but as I am much too shy to post a picture of myself, here is one of my Girlfriend, Debs:     It would appear that it works :-)  '),(9466,2308025,'03,November,2003','2003-11-03 00:00:00','   Experimenting with Blogger. So far, so good.  '),(9467,2308025,'24,December,2003','2003-12-24 00:00:00','   Anyone fancy joining Claudia this Christmas?!?!  Happy \"Disco\" Christmas Everybody!  '),(9468,2308025,'23,December,2003','2003-12-23 00:00:00','   Dear Friends, I just wanted to bring to your attention that I feel the year of the Cormac is coming to an end  and as such I feel it is appropriate I leave that name behind.  I was going to go with a symbol but all the good ones had gone so....Cormac - has gone.  Long Live \"DISCOBOY\"!!!  Also I am trying to get my logo sorted, let me know which one is better.  '),(9469,2308025,'23,December,2003','2003-12-23 00:00:00','   \"It don\'t matter if you\'re black or white...\"  '),(9470,2308025,'23,December,2003','2003-12-23 00:00:00','   Need to bring balance to The Force?  '),(9471,2308025,'22,December,2003','2003-12-22 00:00:00','   Scott \"Mangler\" Morrrison Cormac \"Disco Boy\" Booth John \"Killer\" Murray Graham \"The Stalker\" Watts Stuart \"Jack\" King Doug \"Tubby\" McConnell  Check what happens when you type in Guy Hepworth  urlLink here.  '),(9472,2308025,'19,December,2003','2003-12-19 00:00:00','     '),(9473,2308025,'19,December,2003','2003-12-19 00:00:00','   What the fuck\'s a stettler?!  '),(9474,2308025,'16,December,2003','2003-12-16 00:00:00','     '),(9475,2308025,'15,December,2003','2003-12-15 00:00:00','   Soap or Pebble? Im bored of this now.  '),(9476,2308025,'15,December,2003','2003-12-15 00:00:00','     '),(9477,2308025,'12,December,2003','2003-12-12 00:00:00','   Hey Everyone,  Gosh this blog is getting big isn\'t it.  Anyway at the moment it seems to be a site for pictures and JB making fun of scott.  I can\'t think of anything funny so I am settling for a picture (It\'s the hurricane that hit the US in September).  Sorry.  '),(9478,2308025,'12,December,2003','2003-12-12 00:00:00','   Its because before Debs, he got head from this lucky lady:    Oooh that really must\'ve been a toothy blow job.  '),(9479,2308025,'11,December,2003','2003-12-11 00:00:00','     '),(9480,2308025,'11,December,2003','2003-12-11 00:00:00','   Whilst we\'re on a panda bandwagon...  '),(9481,2308025,'11,December,2003','2003-12-11 00:00:00','   Guy looks smug, but why has he got a small man perched on his shoulder peering over his new lady? Am worried that she looks like my sister.... GUY!  '),(9482,2308025,'10,December,2003','2003-12-10 00:00:00','   Apparently she\'s called Cat, but in this photo she\'s been Pandafied!  '),(9483,2308025,'10,December,2003','2003-12-10 00:00:00','   This.  '),(9484,2308025,'10,December,2003','2003-12-10 00:00:00','     '),(9485,2308025,'10,December,2003','2003-12-10 00:00:00','     '),(9486,2308025,'09,December,2003','2003-12-09 00:00:00','   I am posting a message only to move that awful picture.  It may be accurate, it may be funny, but it is also sick and that is too much.   urlLink  '),(9487,2308025,'09,December,2003','2003-12-09 00:00:00','   Dad: Nothing on TV again son?  JB: No Dad. Oh well, shall we watch Mum\'s chuff?  Dad: Good idea son.     And there was much rejoicing.  '),(9488,2308025,'09,December,2003','2003-12-09 00:00:00','   Arf!  '),(9489,2308025,'09,December,2003','2003-12-09 00:00:00','   I keep getting this message:  '),(9490,2308025,'09,December,2003','2003-12-09 00:00:00','     '),(9491,2308025,'08,December,2003','2003-12-08 00:00:00','   You playing tonight at Dr Drakes? JB  '),(9492,2308025,'08,December,2003','2003-12-08 00:00:00','   Not really, but at least I can spell John\'s name correctly.  And here\'s an educational tree:     Update: Nice try JB, but changing the spelling now isn\'t going to help you...  '),(9493,2308025,'04,December,2003','2003-12-04 00:00:00','     '),(9494,2308025,'04,December,2003','2003-12-04 00:00:00','     '),(9495,2308025,'04,December,2003','2003-12-04 00:00:00','   Q: Why did the baker have smelly hands? A: Because he kneaded a poo.  I thank you.  '),(9496,2308025,'03,December,2003','2003-12-03 00:00:00','   For all your stapler needs:   urlLink http://www.virtualstapler.com/   Wow!  '),(9497,2308025,'02,December,2003','2003-12-02 00:00:00','     '),(9498,2308025,'01,December,2003','2003-12-01 00:00:00','   John\'s happy face.  '),(9499,2308025,'01,December,2003','2003-12-01 00:00:00','   I don\'t know why I find it funny, but I do.  '),(9500,2308025,'30,January,2004','2004-01-30 00:00:00','   Tasty!  '),(9501,2308025,'29,January,2004','2004-01-29 00:00:00','   AND WE SHALL ALL LOOK LIKE THIS!!!  '),(9502,2308025,'27,January,2004','2004-01-27 00:00:00','   Cormac Booth is a \"normal\" boy who lives at 77 Acatia Road, but when Cormac eats a flump, he turns into DISCO GEEK! Check those white socks mo-fo!  '),(9503,2308025,'27,January,2004','2004-01-27 00:00:00','     '),(9504,2308025,'27,January,2004','2004-01-27 00:00:00','     '),(9505,2308025,'27,January,2004','2004-01-27 00:00:00','   According to JB. Oh yes sir.  '),(9506,2308025,'27,January,2004','2004-01-27 00:00:00','   Requires Flash and sound...   urlLink  Enjoy!  '),(9507,2308025,'26,January,2004','2004-01-26 00:00:00','     '),(9508,2308025,'21,January,2004','2004-01-21 00:00:00','     '),(9509,2308025,'21,January,2004','2004-01-21 00:00:00','   The Penguin Game seems to have inspired everyone with lots of furious penguin-bashing going on overnight (ooh-er).  Unfortunately, I have once again, upped the ante:     592.6!! Come On!  First person to smash the 600 barrier gets a Mars bar.  '),(9510,2308025,'20,January,2004','2004-01-20 00:00:00','   Play the game  urlLink here .  Instructions:  Click mouse to drop penguin. Click mouse again to make the Yeti swing.  Tricky to get the timing right. Post best scores as a comment.  '),(9511,2308025,'19,January,2004','2004-01-19 00:00:00','     '),(9512,2308025,'16,January,2004','2004-01-16 00:00:00','     '),(9513,2308025,'16,January,2004','2004-01-16 00:00:00','   Comments should now be working again.  \"Post-It Saturday\" was indeed a triumph. Stationery cupboards everywhere will have to re-stock their Post-It supplies in anticipation of the doubtless increase in demand for the sticky buggers and their new found use.  I was very drunk. Felt like crap yesterday.  Congratulations to Cormac for getting a hand job... I hope you got his phone number ;-)  '),(9514,2308025,'16,January,2004','2004-01-16 00:00:00','   For those of you who missed the sight of Mozz in town covered in Post-its, then tough!  You also missed the careful placing of other sticky items on Dirty Bitchaz\'s trilby hats - AWESOME!   Oh and Muc got a (half) handjob on the dancefloor..... Yes Scott, that\'s right  '),(9515,2308025,'16,January,2004','2004-01-16 00:00:00','   Tomorrow I officially loose the www.mobilepartyunit.co.uk domain name. :-( \'Tis a sad day indeed. Perhaps one day it shall rise from the ashes, perhaps not.  The cryptic message reveal (apoligies for the crappieness):  U Oar Al Gay  \"You are all gay\"  Since none of you got it, you are all gay.  '),(9516,2308025,'15,January,2004','2004-01-15 00:00:00','   This is a post from Woody.  He was too feeble to put it on the site himself and so asked me to do it.  Watch Guy as he holidays in sunny California!!!!  Enjoy   urlLink http://sandiegozoo.org/pandas/pandacam/  '),(9517,2308025,'15,January,2004','2004-01-15 00:00:00','   Absolute thieving bastards, MPU Forever!!!  '),(9518,2308025,'15,January,2004','2004-01-15 00:00:00','   Unemployed Discoboy strikes again. Anyway, is the answer to the cryptic message \"U-bend-oar-al-michael-barrymore\".  Roughly translated to You bender, al michael barrymore.  Doesn\'t make much sense.  Is it a computer joke Scott?  '),(9519,2308025,'14,January,2004','2004-01-14 00:00:00','   Hi Everyone, Just a quick note to say that, JB,  I am sorry to say that neither of your supposed googlewhacks are correct.  For a two words to be a googlewhack, they both have to be in the dictionary.  Since neither \'sizemic\' or \'retract.......cation\' (or whatever that first one was) appear of dictionary.com, neither of your pairs are googlewhacks.  A helpful tip, is that is you type in your pair on google.com, once the search is complete, the words will be underlined if they appear in the dictionary (in the blue bar).  Also if the possible googlewhack webpage is a dictionary or word list of some sort, then they are disqualified.    Try harder next time.  Anyway,if you want some examples of real googlewhacks: try \'hippocampi wallpaper\' or \'bibliophilic sandwiched\'.  I am still unemployed incase any of you wondered.  '),(9520,2308025,'14,January,2004','2004-01-14 00:00:00','   Mozz Missing once More, it is believed that this is due either to wank or modern jazz.  '),(9521,2308025,'13,January,2004','2004-01-13 00:00:00','   They guy running BlogSpeak, the service that powers the Comments on this blog, is having problems.  To quote his site:  \"BlogSpeak is currently down because the bastards that host it decided to suspend my account. I do not know as of yet when this situation will be resolved. If you don\'t want any JavaScript errors on your pages, take the code off for the time being. Thanks for your patience.\"  So blame me if your browser acts funny on this page, I can\'t be arse removing the script yet.  '),(9522,2308025,'12,January,2004','2004-01-12 00:00:00','   Say what you see!   First person to decipher the message wins 5 points.  '),(9523,2308025,'09,January,2004','2004-01-09 00:00:00','   Its not very interesting though.  retractification superfluous - brings up a slang dictionary  sizemic superfluous - brings up a crap website  '),(9524,2308025,'09,January,2004','2004-01-09 00:00:00','     '),(9525,2308025,'07,January,2004','2004-01-07 00:00:00','   You will need:  A printer A piece of paper A pen A scanner A computer  and   urlLink www.fontifier.com   Results are not bad...    Although I appear to write like a 9 year old...  '),(9526,2308025,'06,January,2004','2004-01-06 00:00:00','   Didn\'t turn out as well as I\'d hoped, but it wasted 10 mins of a slow afternoon...   Was going to try and make him dance, but didn\'t.  '),(9527,2308025,'05,January,2004','2004-01-05 00:00:00','   Hello.  DiscoBoy here!  Just wanted to call John a cumdumpster and show you all a picture of my new kitchen.  What do y\'all think?!  '),(9528,2308025,'05,January,2004','2004-01-05 00:00:00','   Disco Bitch Disco BoothDisco Bitch Disco Booth Disco Bitch Disco Booth  '),(9529,2308025,'05,January,2004','2004-01-05 00:00:00','   Conclusive Proof that I am in fact Dead.  '),(9530,2308025,'28,February,2004','2004-02-28 00:00:00','   Here\'s John proudly showing off his new purchase from Friday night\'s tupperware party...  '),(9531,2308025,'27,February,2004','2004-02-27 00:00:00','   After lengthy discussions with my learned friend Mf Forsyth we have decided to call all people for what we will term War in Westburn Park.    The time: tomorrow Saturday 12 noon The place: Westburn Park/ Victoria Park(maybe better with more obsticles) The people: everyone on this email list and whoever you guys can get.  The teams: Split between Aberdonians and the rest. NOTE that everyone has to come along as if not then fun will be ruined for all ie Murray - girlfriend, Wabber - rugby, everyone  - any excuse. Special guest apperance for the out of towners: Mr C Cheeky Ogilvie   RSVP    cheers doug P.S. from Cormac - YOU ALL HAVE TO COME!!!!!!!! Booze and XBOX round at JBs after I think.  '),(9532,2308025,'24,February,2004','2004-02-24 00:00:00','   It\'s the Middle East\'s answer to Gareth Gates!  '),(9533,2308025,'24,February,2004','2004-02-24 00:00:00','   Debbie has sent Scott to Qatar. Here\'s an angry old man who has just said \"you love it you slag....   ....Now NAFF OFF!\"  '),(9534,2308025,'23,February,2004','2004-02-23 00:00:00','   UPDATED:  '),(9535,2308025,'23,February,2004','2004-02-23 00:00:00','   Variation on the ridiculously popular penguin game posted here a few weeks back.   urlLink Click here .  '),(9536,2308025,'23,February,2004','2004-02-23 00:00:00','     '),(9537,2308025,'21,February,2004','2004-02-21 00:00:00','   Discuss....  '),(9538,2308025,'19,February,2004','2004-02-19 00:00:00','     '),(9539,2308025,'18,February,2004','2004-02-18 00:00:00','   Obviously before Robin ate them all.  '),(9540,2308025,'17,February,2004','2004-02-17 00:00:00','   I was hoping to have time to create some sort of animation/picture to celebrate Guy\'s homecoming; the return from the land of \"gangstas\" and \"thugz\".  Unfortunately, I\'m running out of time (got proper work to do) and my imagination is in a rut.  I got as far as this:     I may get round to making it move at some point, but for now I\'ve decided to use it to launch a caption competition.  What do you think Guy is thinking or saying in the picture? Can you think of any good slogans that could accompany this picture as a poster campaign?  Post suggestions as comments, best entry wins a beer. For the artistic amongst you, you can obtain a bigger version of the file  urlLink here  for editing.  '),(9541,2308025,'13,February,2004','2004-02-13 00:00:00','   Go on, give him a wave!  '),(9542,2308025,'13,February,2004','2004-02-13 00:00:00','     '),(9543,2308025,'13,February,2004','2004-02-13 00:00:00','   (Credit to JB for the idea)  '),(9544,2308025,'12,February,2004','2004-02-12 00:00:00','   Look Closely...   urlLink Click Here .  '),(9545,2308025,'12,February,2004','2004-02-12 00:00:00','   Check this out.  Plenty of reason\'s why single folk shouldn\'t stay in on Valentines Day:   Click Here .  Woody found this, credit should go to him.  Though he was too inept to put it up so he should get slandered too.  '),(9546,2308025,'12,February,2004','2004-02-12 00:00:00','   Hello, I\'m Batman.  When I was just a child, my parents were murdered right in front of my eyes.  I made a solemn vow that day to commit my life to fighting crime.  I\'ve sacrificed my childhood and any efforts toward personal enjoyment.  My nights are spent in deadly combat with everything from common street thugs to grotesque psychopaths.  What little sleep I ever get is usually interupted by horrible nightmares of brutal murder and my waking up soaked in sweat.  I fight a war that can never be won.  I strive toward a goal that can never be reached.  I am haunted.  I am relentless.  I am tortured.  Won\'t you be my valentine?  '),(9547,2308025,'12,February,2004','2004-02-12 00:00:00','     '),(9548,2308025,'11,February,2004','2004-02-11 00:00:00','   Proof:  '),(9549,2308025,'11,February,2004','2004-02-11 00:00:00','     '),(9550,2308025,'11,February,2004','2004-02-11 00:00:00','   The Blog seems a bit quiet just now.  I do not approve.  As a display of my disapproval, meet Angry Gary Coleman:  '),(9551,2308025,'11,February,2004','2004-02-11 00:00:00','   In the morning, when Cormac\'s new \"friend\" asked for a shower, he was happy to oblige...  *UPDATE* Due to paranoid, over sensitive type folk (who shall remain nameless), the picture accompanying this post has been removed from the front page and can be found  urlLink here . Apparently it\'s Not Safe For Work (=NSFW).  '),(9552,2308025,'10,February,2004','2004-02-10 00:00:00','   Hey Everyone.  Just found  urlLink this  on the intenet.  God Bless Sport\'s Illustrated.  Choose your favourite girl from the bar at the top and post them as a comment.  Sweet Mercy!  '),(9553,2308025,'10,February,2004','2004-02-10 00:00:00','   This is the weirdest thing ive ever seen, its safe for viewing at work.  urlLink Click Here .  The site does, however, clarify this below noted site:  urlLink Click Here .  '),(9554,2308025,'09,February,2004','2004-02-09 00:00:00','   This is what GAJ will look like in the near future if our Spandex and PVC looks don\'t work out and Gary fails to die in an horrific armchair accident and thus denying Tim his place in the almighty TAJ! Ps Andy\'s the bird.  '),(9555,2308025,'06,February,2004','2004-02-06 00:00:00','   I had initially thought that one of the scariest things I had experienced was Toni Baxer sticking needles into my abdomin and eyes before sawing my foot off..... or was that a dream.... or was that a film...... anyway but now its definitely Ashley growling \"you wan pie?\"  URGH! sends a shudder down my spine as if Lesley Fraser had licked me urgh.  No thank you very much i do not wan pie.  '),(9556,2308025,'05,February,2004','2004-02-05 00:00:00','   You wan pie?   urlLink Weebl & Bob   Watch them from the very start...hours of humour...  '),(9557,2308025,'05,February,2004','2004-02-05 00:00:00','   WAZZ AND MAZZ MAKE JAP DVD NIGHT!!!! Like this paper plate, they are proud to promote literacy.  '),(9558,2308025,'05,February,2004','2004-02-05 00:00:00','   WAZ AND MAZZ TURNED DOWN DVDS FOR TACOS AND SEX!!!! Like this dog, they are humiliated.  '),(9559,2308025,'04,February,2004','2004-02-04 00:00:00','   Following Tuesday\'s fiasco that was my guitar playing, John and I have decided to put in a bit more effort for next week.  Not only will I be writing the chord sequences in BIGGER WRITING so I can read them, I may actually practice the songs in advance.  I think the \"3 song\" approach will work best; 2 of John\'s own and a cover song.  But frankly we\'re too lazy to pick a song to cover ourselves, instead we are giving you (yes, you) the opportunity to pick \"A Song for Muzz and Mozz\".  Suggestions should be posted as comments. The winner will be picked by JB and myself unless there is an overwhelming majority for a particular song.  Good luck, and get voting!  '),(9560,2308025,'04,February,2004','2004-02-04 00:00:00','   /points to tag line and cries quietly.  '),(9561,2308025,'04,February,2004','2004-02-04 00:00:00','     '),(9562,2308025,'04,February,2004','2004-02-04 00:00:00','     '),(9563,2308025,'03,February,2004','2004-02-03 00:00:00','   I think it\'s fair to say that last night\'s performance was \"distinctive\". My fingers are killing me this morning. Ouch.  '),(9564,2308025,'03,February,2004','2004-02-03 00:00:00','     '),(9565,2308025,'03,February,2004','2004-02-03 00:00:00','   Courtesy of the  urlLink bbc .  Some notable omissions include: Cunted Pissed Shitfaced Wankered Rat Arsed  Personal Favourites: Boogaloo Dot Cottoned Locked out of your mind Monkey-full Nicely irrigated with horizontal lubricant Saying hello to Mr Armitage Shedded [as in \"My shed has collapsed taking most of the fence with it\"]  '),(9566,2308025,'02,February,2004','2004-02-02 00:00:00','   It was my first band practice in >5 years last night, and I finally managed to conquer the basics:    Muzz and Mozz - Back Together Tour, begins tonight @ Dr Drakes open mic night, after 10pm.  '),(9567,2308025,'02,February,2004','2004-02-02 00:00:00','   Ever wondered which horrible affliction you would be?  Wonder no more! Find the answer  urlLink here .  '),(9568,2308025,'26,March,2004','2004-03-26 00:00:00','   What is this strange chicken theme?  '),(9569,2308025,'25,March,2004','2004-03-25 00:00:00','     '),(9570,2308025,'24,March,2004','2004-03-24 00:00:00','     '),(9571,2308025,'22,March,2004','2004-03-22 00:00:00','     '),(9572,2308025,'19,March,2004','2004-03-19 00:00:00','   Also, check out this:  urlLink http://www.nbpl.net/   It\'s not even a joke.  the use of beer and pong at parties has only just begun my friends.  But can we please do it at someone elses next time.  Guy has a nice long table.....  '),(9573,2308025,'19,March,2004','2004-03-19 00:00:00','     '),(9574,2308025,'18,March,2004','2004-03-18 00:00:00','   Sorry it\'s a day late.  '),(9575,2308025,'17,March,2004','2004-03-17 00:00:00','     '),(9576,2308025,'17,March,2004','2004-03-17 00:00:00','     '),(9577,2308025,'17,March,2004','2004-03-17 00:00:00','   Shit Yeah.  Turn up any time after 7.30pm  '),(9578,2308025,'17,March,2004','2004-03-17 00:00:00','   Apparently we are playing  urlLink this  at Cormac\'s on Friday, so get practicing! I have also recently purchsed  urlLink this  XBox game complete with 2 dance mats. I\'m sure that once we\'ve had a few drinks, we will all be dancing like fools!  '),(9579,2308025,'16,March,2004','2004-03-16 00:00:00','   urlLink Click Here .  '),(9580,2308025,'15,March,2004','2004-03-15 00:00:00','   Nest of Death provide the support.  '),(9581,2308025,'13,March,2004','2004-03-13 00:00:00','   Under construction:  urlLink www.johnsongs.co.uk  '),(9582,2308025,'13,March,2004','2004-03-13 00:00:00','   urlLink  Click here  Short video (30 secs), sound required. Broadband recommended  '),(9583,2308025,'10,March,2004','2004-03-10 00:00:00','   ... with his \"Back to the Fuschia\"  '),(9584,2308025,'09,March,2004','2004-03-09 00:00:00','     '),(9585,2308025,'08,March,2004','2004-03-08 00:00:00','     '),(9586,2308025,'05,March,2004','2004-03-05 00:00:00','   Seamless!!!  '),(9587,2308025,'04,March,2004','2004-03-04 00:00:00','   come back here with my tankard!!!  '),(9588,2308025,'03,March,2004','2004-03-03 00:00:00','   come on you slags  '),(9589,2308025,'03,March,2004','2004-03-03 00:00:00','   No but it is fucking scary.  '),(9590,2308025,'02,March,2004','2004-03-02 00:00:00','   Just tried to go to www.johnsongs.co.uk and it simply didn\'t come up! What does it mean?!?!? Does someone already have it?!?!? Is it under construction by some other John?!?!? Will boudy be a panda in the morning?!?!? Will boozy be slapping it to pornos?!?!? Will wazz be having a celebratary wank?!?!? AGGGHHHH I Don\'t know!  '),(9591,2308025,'01,March,2004','2004-03-01 00:00:00','     '),(9592,2308025,'01,March,2004','2004-03-01 00:00:00','   Thanks for supporting me last night, it was well appreciated and my sincerest apologies for making you all listen to that other bloke drone on before me. Big Up the German contingent!  '),(9593,2308025,'01,March,2004','2004-03-01 00:00:00','     '),(9594,2308025,'01,March,2004','2004-03-01 00:00:00','   I met Sven Goran Eriksson yesterday at Manchester airport. Shook his hand and took his picture. Unfortunately I only had my mobile phone camera with me, so the quality isn\'t great.  '),(9595,2308025,'30,April,2004','2004-04-30 00:00:00','     '),(9596,2308025,'28,April,2004','2004-04-28 00:00:00','   Front row seats to a Flyers game: $145  Replica hockey jersey: $225  beer and hot dogs: $22  Picture of you and your friend acting like a couple of FAGS in Sports Illustrated: PRICELESS!  '),(9597,2308025,'28,April,2004','2004-04-28 00:00:00','     '),(9598,2308025,'28,April,2004','2004-04-28 00:00:00','   It would appear Captain America is a wanker.  '),(9599,2308025,'26,April,2004','2004-04-26 00:00:00','     '),(9600,2308025,'26,April,2004','2004-04-26 00:00:00','     '),(9601,2308025,'23,April,2004','2004-04-23 00:00:00','     '),(9602,2308025,'21,April,2004','2004-04-21 00:00:00','     '),(9603,2308025,'16,April,2004','2004-04-16 00:00:00','   Pointless, but she does jiggle.  '),(9604,2308025,'16,April,2004','2004-04-16 00:00:00','   Well in case any of you were wondering when I (Cormac the Skald???) was going to snuff it...read on to find out more about me!   urlLink http://sunsite.berkeley.edu/OMACL/Cormac  '),(9605,2308025,'16,April,2004','2004-04-16 00:00:00','   urlLink http://www.uua.org/uucf/jmur-con.htm   bizarre  '),(9606,2308025,'13,April,2004','2004-04-13 00:00:00','   It\'s Beadle Time!  Sure has been quiet round here lately.  '),(9607,2308025,'05,April,2004','2004-04-05 00:00:00','   SEAMLESS!!!  '),(9608,2308025,'05,April,2004','2004-04-05 00:00:00','   Sorry to be random, just wanted to get that stupid fucking chicken picture off the top of the site.  '),(9609,2308025,'27,May,2004','2004-05-27 00:00:00','   Friday night, just before the duck flew South:  '),(9610,2308025,'26,May,2004','2004-05-26 00:00:00','   Looks like JB, Doug & Gary were snapped by the Belmont Street photographers on Friday night.  Check out the Belmont Street  urlLink website , look under the section \"what did you do?\", dated Friday 21st May, the boys were caught in the Wild Boar!  '),(9611,2308025,'24,May,2004','2004-05-24 00:00:00','   So, all you folks with digital cameras... where are the pictures from Friday, eh?  '),(9612,2308025,'21,May,2004','2004-05-21 00:00:00','     '),(9613,2308025,'21,May,2004','2004-05-21 00:00:00','     '),(9614,2308025,'17,May,2004','2004-05-17 00:00:00','   This is a picture of the very rare \"Nemesis Gibbon\".  Notice how it wears the Nemesis t-shirt, even though gibbons aren\'t actually allowed on the ride.  '),(9615,2308025,'14,May,2004','2004-05-14 00:00:00','   My legs are in agony!   I hate fishing.  '),(9616,2308025,'14,May,2004','2004-05-14 00:00:00','   Is it just me, or does this donkey look a bit smug?  '),(9617,2308025,'12,May,2004','2004-05-12 00:00:00','   I know it\'s a bit crap, but I had some work to do, anyway I think the message is the same.  Scott is a giant hom.  Let\'s see how long it is before he strops and takes it off!!!  '),(9618,2308025,'11,May,2004','2004-05-11 00:00:00','   What does everyone think of Cormac\'s new apartment? He is leaving soon and I thought some of you might like to be able to find out a little bit about what he is doing and be able to keep track of him once he is gone.  Check it out  urlLink here  and  urlLink here  '),(9619,2308025,'11,May,2004','2004-05-11 00:00:00','   Following a short break from the lime-light, John Murray is back!     The details: The Belmont Cinema Bar Belmont Street Aberdeen  The show should be starting around 9pm. Arrive early to drink beer.  '),(9620,2308025,'11,May,2004','2004-05-11 00:00:00','     '),(9621,2308025,'10,May,2004','2004-05-10 00:00:00','   What does everyone think of my new apartment? I\'m leaving soon and I thought some of you might like to be able to find out a little bit about what I am doing and be able to track where I am once I\'m gone.  Check it out  urlLink here  and  urlLink here  '),(9622,2308025,'10,May,2004','2004-05-10 00:00:00','   Blogger have just overhauled their system and as a result, my blog now looks shiny and new!  New features include the new comment posting system!  '),(9623,2308025,'06,May,2004','2004-05-06 00:00:00','     '),(9624,2308025,'05,May,2004','2004-05-05 00:00:00','   JB rated very highly on the Mong-o-meter!  '),(9625,2308025,'04,May,2004','2004-05-04 00:00:00','   An old one, but a good one...  '),(9626,2308025,'04,May,2004','2004-05-04 00:00:00','     '),(9627,2308025,'04,May,2004','2004-05-04 00:00:00','     '),(9628,2308025,'25,June,2004','2004-06-25 00:00:00','   Beckham Penalty!  '),(9629,2308025,'23,June,2004','2004-06-23 00:00:00','   Go on, you know you want to!  '),(9630,2308025,'21,June,2004','2004-06-21 00:00:00','   Have you seen this man? Last reported skulking around the Manchester area. Carol Vorderman suspects she has a stalker.  '),(9631,2308025,'17,June,2004','2004-06-17 00:00:00','   Hello everyone, I\'m back from my short break in sunny Manchester!  I can confirm that I did in fact go on every major ride at Alton Towers and I have pictures to prove it!  Although, after 10 years of torment, I have to say that the Nemesis was a bit disappointing. Air is good, but Oblivion was the best.  Managed to get sun-burnt again.  Leeds seemed to be a bit of a sh*t-hole, one big building site.  Here is a picture of me and Debbie with our new friends:  '),(9632,2308025,'16,June,2004','2004-06-16 00:00:00','   Imagine my surprise as my tv switched itself on and emanted an other worldly glow....  '),(9633,2308025,'13,June,2004','2004-06-13 00:00:00','   Clicky clicky on the radio buttons to see the effect...        Before...     After...  '),(9634,2308025,'09,June,2004','2004-06-09 00:00:00','   I\'m pretty sure I emailed everyone, but on the off chance I forgot, here\'s the poster for the Wazzamatazz Air Guitar Birthday Piss Up:   urlLink    Click to bigify.  Everything kicks off on Saturday afternoon...  EDIT: Big version of the picture also available at  urlLink www.grahamwatts.net .  '),(9635,2308025,'03,June,2004','2004-06-03 00:00:00','   ....and there were shouts of joy.   I\'m off, thought this photo was appropriate.  Let me know how you all are, like I say I dunno how easy it is going to be for me to stay in touch, but I\'ll do my best.  I\'ll miss you all very much.  Look after yourselves and I\'ll see you sometime soon.  Meanwhile feel free to view my minimal progress on the  urlLink blog  and the \" urlLink Voyage of the Odyssey \" webpage.  This is The Booth, signing off.  '),(9636,2308025,'20,July,2004','2004-07-20 00:00:00','   He\'s not been working late in the Deen at all!     He\'s opening a Tokyo TV Dining Facility!     AND HE NEVER TOLD US!     Dick!  '),(9637,2308025,'16,July,2004','2004-07-16 00:00:00','   Hello boys and girls,  I am now slightly unbeleivably in Tanzania. When I went on holiday with my family a few weeks ago I had a rough idea that I would stay if i could find anything worthwhile to do, but i sincerely didnt think I\'d find it, but i kind of did albeit in a terribly unpaid fashion. Nevertheless the job is cool, and the people/landscape/food is brilliant.  So there we go, I stayed they went. If all goes well I\'ll be back in Spetember so its not a proper good bye a la Cormac. Feel sorry for not updating you all on things before I went and milking the departure for at least a piss up but I didnt really know i\'d stay. In fact Revolution had me shifted on for work all this and last week. Whoops. :P  Anyway hope you are all doing well I\'ll start sending out some proper emails the next time I have a large task to do at work. In case you dont know my email is johnprimrose (at) hotmail.com and as scott or cormac could tell you I\'m always on MSN.  Take it easy  John  Look at this if you havent already  urlLink www.rockpapersaddam.com  '),(9638,2308025,'09,July,2004','2004-07-09 00:00:00','   According to Eurogamer.net, Microsoft has officially  urlLink announced  the UK release date for Halo 2 on the XBox as being November the 11th 2004. 2 days later than expected and 2 days after the Americans.  '),(9639,2308025,'08,July,2004','2004-07-08 00:00:00','   Great news from  urlLink Slashdot!  The Monty Python & The Holy Grail musical has been confirmed along with some of the casting! Tim Curry will play Arthur, King of the Britans and David Hyde Pierce (Niles from Frasier) is to play Sir Robin.  More info  urlLink here .  And there was much rejoicing.... . . . . . . ....(yaaaaaaaay)  '),(9640,2308025,'06,July,2004','2004-07-06 00:00:00','   I knew it...  '),(9641,2308025,'05,July,2004','2004-07-05 00:00:00','   Just testing mailing to the blog from my phone. I hope it works.  '),(9642,2308025,'05,July,2004','2004-07-05 00:00:00','   urlLink    Click to big it.  Mairi,  urlLink here  is one of the piccies taken with my new phone for you to get an idea of it\'s capabilities. I think you\'ll agree, despite the good picture quality, Graham still looks like a smug git.  '),(9643,3353979,'30,May,2004','2004-05-30 00:00:00','   who have i become,,, i have become the odd man out,,,, i dont think anyone notices me anymore...... Everyone just takes life as a joke, im taking my life seriously, im taking small things too seriously...  '),(9644,3353979,'28,May,2004','2004-05-28 00:00:00','   Today was one of those lazy tired days. Me and nicki went to haley and chalces house started playing kickball, then steven and kristen came, kristen remined me that she was avoiding me, all that everyone could do is fight, what happened to the days when we could just play a game and and have fun, then nathan and john came, i left...... went to the coffee house natan, john, steven,kristen,casey came down....... Kristen wont talk to her good friends but she\'ll talk to nathan,,,,, what is going on with kristen..... or have i changed......  '),(9645,3353979,'27,May,2004','2004-05-27 00:00:00','   Went to nickis house after school,, kristen came over, she dident say anything to me, i dident say anything to her,,, i went to the park,played basketball, nicki, kristen, haley and chalce came, started play kickball, kristen dident play, steven got mad and quit, kristen just isent the same person anymore,, yount came about an 20 min later,,, anytime someone from out of the neightboorhood comes she just completely ignors me, its almost like were not friends anymore...................  '),(9646,3353979,'26,May,2004','2004-05-26 00:00:00','   Me and nick were at haleys house, i told her that is still like kristen, she wasent happy and i just dont understand why....... I seen kristen talking to anna, when i got home i found out from steven that kristne was going to kennywood with anna, whats going on just 2 weeks ago she was saying how mean anna is.. kristen has really changed.............  '),(9647,3353979,'24,May,2004','2004-05-24 00:00:00','   Was able to talk to nicki for a short time, told her that kristen wont talk to me until i tell her how she has changed. Got to talk to randall. Everything is still unsettled... I got a million things i want to tell kristen,, i still like her, but someone has to tell her how she has changed, im afraid to tell her, the truth hurts.... I have to tell her that i still like her and how our friendship is different now that she knows................  '),(9648,3353979,'23,May,2004','2004-05-23 00:00:00','   Me nicki start to become better friends, but as me and nicki become better friends me and kristen become worse friends,,, ever since i told kristen i like her it hasent been the same. Yesterday when i found out she dident like me i just want to avoid her, i have nothing to say to her. I told kristen that me and nicki were talking about how she has changed, i wouldent tell her how she has change, she got mad and said she wouldent talk to me untill i told her. I dont want to tell her, i know it will just make things a lot worse. Maby its best that i dont see kristen for awhile............  '),(9649,3353979,'21,May,2004','2004-05-21 00:00:00','   Went to youth day talk,, wanted to talk to kristen all day,, of course i dident. Got home got to talk to nicki at haleys house, it was good to finally talk to nicki, we went to kristens asked her i she liked me, she fixed things with her boyfriend,,,, i just dont know who i like anymore, my life just keeps getting more and more confusing,,, id love to move and start new......................  '),(9650,3353979,'21,May,2004','2004-05-21 00:00:00','   Finally on ok day. Talked to kristen on instant messenger, i kept saying how good she looked and how nice she is, i just kept going on and on. Went to the coffee house got to see christina mcune beat up john, it was funny,,, so that ends me ok day.  '),(9651,3353979,'19,May,2004','2004-05-19 00:00:00','   Today was one of those day that you could go to sleep right after school, kinda an easy day at school. 4pd basic video got to jump out of locker, it was fun. I spent part of my day just walking around. Went to the park, nicki,kristen,steven,dan s, came down. I was able to talk to kristen, finally, me and nicki stared talking about kristen, i told nicki that i like kristen, as we were walking up from the park, i told kristen that i like her,, its about time i brought that up again, i found out kristen is going to break up with her boyfriend on thursday.......................  '),(9652,3353979,'18,May,2004','2004-05-18 00:00:00','   Today was a boring yet exciting change from the ordinary day. I dident mind being in school today. School life improving, thats a good thing. Im starting to get my life back together.... Everything seemed normal untill the afternoon bus ride, poring rain and thunder, windows down on bus, water pouring into bus, everyone was wild on the bus today, it was a definte change from the ordinary. As for nicki and kristen, well i really dont want to talk to them, they have changeed too much. The rest of my day was spent hanging out at the park not really doing much...........  '),(9653,3353979,'17,May,2004','2004-05-17 00:00:00','   Today day started out like any other monday, tired, boring, just waiting for somthing to happen, long day at school, just a boring day. Went to the park nicki,steven,nathan,brian and chris was there. Nicki and kristen have really changed over the past year, there different people, there not the same people i know. They take like as a joke and right now and my life isent a joke, they joke with there feelings and im done joking around, \"life is real\". Right now i\'d give anything to just move far away, a new life, new friends, new everything............  '),(9654,3353979,'16,May,2004','2004-05-16 00:00:00','   What a day, my day started just like any other BORING day, until tonight, at the park with ,Steven,zac and Chris,,, an of course Nathan was TRYING to act like a good skate border, it was funny watching him fall, than kristen and anna came down,,, anna not one a favorite people,, me zac, steven and nathan started playing basketball, of course me and zac lost, not a surprise to me, than nathan started saying stuff about anna anna got mad and got iced tea on nathan and started throwing rocks at nathan, nathan ran away like a baby,,,,, i really want to tell kristen that i like her and why, but no time really seems like the right time to tell her,,,,,,,,,, and that ended another one of my days.................  '),(9655,3353979,'15,May,2004','2004-05-15 00:00:00','   last night around 8:00 i was at haley and chalces house with steven,nicki,kristen,randall, they dident even notice me, i was little mad, but its just the age difference, \"it\'s easier to live backwords than foreword\" because you already lived backwords.  '),(9656,3353979,'15,May,2004','2004-05-15 00:00:00','   The past few days have been hell for me,,, on tuesday i told kristen i liked her, dident say why, is was sick majorly that day, told her weâ€™d talk the next day, dident get a chance to, stayed home from school, to afraid to talkt to her, avoided her all day, the rest of the week is seen her hanging around chris,nathan,steven,brian, it makes me mad and jealous, its driving me crazy not to tell her, im just afraid to bring up the subject again, right now shes on instant messenger just to afraid to talk to her, dont want to tell her why i like her on instant messenger, i think i should tell nicki she could relay it to kristen but iâ€™d rather tell her in person, what should i do, im so upset im about to cry i need tell but i mean she already has a boyfriend and even though i need to tell her i dont want it to change our friendship i just starting to rebecome friends with her,,,,, all the the years iâ€™v know kristen we really werent good friend in fact last year i kinda disliked her,,, now that were starting to become friends i dont want anything to change,,, my life is already screwed up, this just makes it a 100% worse, its driving me crazy!!!!!!!!!!!!!!!!  '),(9657,3353979,'27,June,2004','2004-06-27 00:00:00','   Today i went to the old club areas, knowing that going there will make me feel alot worse and it did, the place has really changed, i felt so unconfortable there i was angery and sad, i couldent stop thinking about who i was,, who i was when these log walls were standing high, every time i go there i always find somthing of mine and every time i pick that object up i can remember what was going through my head when it was built................  '),(9658,3353979,'26,June,2004','2004-06-26 00:00:00','   Another day spent trying to find out what im good at. First i found drawing but i can only draw what i was used to seeing, i cant finish a drawing, then theres basketball, i can be good at that just not now i got too many things on my mind and then theres somthing i always used to do, draw ideas for futuristic inventions, i can still do that its just not the same.................  '),(9659,3353979,'24,June,2004','2004-06-24 00:00:00','   Every day i look back and see the person i was and try to be that person, life was so easy then, now im not sure if my life was really that great then..................  '),(9660,3353979,'23,June,2004','2004-06-23 00:00:00','   Feelings toward everyone have changed,,,,, i have strong feelings toward someone and i dont think she feels the same way..........................  '),(9661,3353979,'22,June,2004','2004-06-22 00:00:00','   Normal, somthing we all take for granted untill its gone, its like when your sick your off from school for a few days, at first you dont miss normal, then you start to miss waking up, doing your morning rutine, going to school and maby thers friends you like to talk to and then you go thru out your day, but when normals gone nothing seems to make sense........................  '),(9662,3353979,'22,June,2004','2004-06-22 00:00:00','   We all need to be good at somthing, and we all need to know were good a somthing, but what happens when the thing your good at is gone, when you cant be good at it anymore, you feel like a part of you is missing, and after you realize you cant be good at it anymore you go a search to find somthing else you good at, but most of the time its not that easy, you feel like you have to great at every new thing you try or you give up....................  '),(9663,3353979,'21,June,2004','2004-06-21 00:00:00','   kristen i still like you and nothing will change that........................... Yet another day spent at the park thinking,, i dont think anyone care about me,, my day is spent full of sadness, another sun set watched at the park and yet another horrible day goes by,,, so many things going thru my head its driving me crazy, so many thoughts, so many questions with no answers, why me?, why am i going thru this? why did my life have to change?,,,,,, i cant remember what happyness feels like.........  '),(9664,3353979,'20,June,2004','2004-06-20 00:00:00','   Know one knows who i am and i dont think anyone around here will.... The place that is home dosent feel like home anymore,, i feel traped in this place, today i walked around and this place donsent feel like home it feels more like a prision....... i cant stand living here anymore,, the people here sometimes make it worse they dont know who i am, im full of secrets... i think there is only one person here who can understand me but im still to afraid to open up and reveal the true me, i know that if i open up i will be alot happier and feel alot better but i just dont know if im ready..........  '),(9665,3353979,'13,June,2004','2004-06-13 00:00:00','   what ever happend to a day when there is peace.... when there is no conflict between people there still the fight with my self,, all the secrets i have that im dying to tell someone, and to show the real me.......  '),(9666,3353979,'11,June,2004','2004-06-11 00:00:00','   Life is like a road the path you life should go is staight, but the path is full of bends and intersections.. My life took took the wrong turn............................ Know one knows the real me and im afraid to show the real me  '),(9667,3353979,'27,July,2004','2004-07-27 00:00:00','   Im feeling alot better, now that im talking to kristen again,,, i still need to know how she feels about me...........................  '),(9668,3353979,'22,July,2004','2004-07-22 00:00:00','   I cant put how i feel right now into words,,,, im sad, angery and so much more..................................  '),(9669,3353979,'22,July,2004','2004-07-22 00:00:00','   Everytime i see her i realize how much i like her, and i asume she hates me, shes never really said wether she likes me or not.......  '),(9670,3353979,'21,July,2004','2004-07-21 00:00:00','   Everything is getting a bit better, im not sad as i was but somtimes im as sad as i was and i know why........ kristen wouldent know everything is getter better because we dont talk.........................  '),(9671,3353979,'16,July,2004','2004-07-16 00:00:00','   I like her so much, im afraid to talk to her, for the first time im too shy to talk to kristen...............................  '),(9672,3353979,'14,July,2004','2004-07-14 00:00:00','   I always hid things in my life, i always ran away from good things in my life, im finally realizing that running away isent a life,,,, I also have a habbit of asuming things i dont know, i guess thats why i havent talked to you, because you know how i feel about you, but i dont know how you feel about me.........................  '),(9673,3353979,'12,July,2004','2004-07-12 00:00:00','   2 questions still remain unanswered,,, kristen please answer them,,, are we friends, do you really like me..........................  '),(9674,3353979,'11,July,2004','2004-07-11 00:00:00','   When i felt sad or angery i would go to a spot on the edge of the woods and just sit there staring into the woods and some how i would feel better, or when i was feeling really bad i would go the club sit back with my feet on the table and look all around, now there is no place to go...........  '),(9675,3353979,'11,July,2004','2004-07-11 00:00:00','   Today was another sad day untill i seen kristen, so much to tell her, the 2 questions still remain unanswered and kristen needs to answer them,,,, well i guess when i seen kristen somthing in my mind just slipped, and there i was riding my bike down the hill as fast as i could, not paying attention to anything but going fast, i was standing on my bike not caring if i got hit by a car and hoping i would, i kept doing this untill i got so tired i couldent and common sense kicked in,, im so afraid ill do it again, im afraid of what ill do...............................  '),(9676,3353979,'09,July,2004','2004-07-09 00:00:00','   Shes means so much to me and i need answers for two unanswered questions, are we friends, becaused lately it hasent seemed like it,,, Do you really like me?..................  '),(9677,3353979,'08,July,2004','2004-07-08 00:00:00','   I really like her, she means so much to me and she cant see that....... She dosent think shes pretty, thats not true why would she think that. Shes a great person......  '),(9678,3353979,'07,July,2004','2004-07-07 00:00:00','   The reason why i hide the real me is because of what people would say, today i found out that people do say things, things that make me feel so much worse, i also found out that friends will understand and everyone else will say things,, the only thing that kept me together today was a quote from the book Atlas Shrugged- \"Do not let your fire go out, spark by irreplaceable spark, in the hopeless swamps of the not-quite, the not-yet, the not-at-all. Do not let the hero in your soul perish, in lonely frustration for the life you deserved, but have never been able to reach. The world you desired can be won, it exists, it is real, it is possible, it\'s yours.\"  '),(9679,3353979,'06,July,2004','2004-07-06 00:00:00','   in my last post i said i was going to stop my blog but i cant this is the only way i can get everything out. most of time it easier to write it down then to tell someone.  Over the past week i guess i put my life on hold while she was gone, i like her alot and she dosent feel the same way,, i have been quetioning alot of things,, are we friends, i dont know thinking back to winter were we spent months not talking and friends dont do that, friends cant do that.... Every day i feel like i dont belong in rural ridge, and nearly every day somthing happens that keeps me knowing that i dont belong here.......... Every day is find out what i can and can not do, every day i push my self to the limit know the risks and not careing and i do it because i think i can relieve the sadness im feeling, i cant play basketball and know i can play a few weeks ago i could and i know why i cant play its because im play for all the wrong reasons i was playing it because it was fun now i play it to get my mind on somthing else, to not be sad..... Every day i leave with out telling anyone, i leave because i started thinking and im not thinking any happy thoughts................ the music i listen to isnet helping i listen to that describes how i feel and who i am................  '),(9680,3353979,'05,July,2004','2004-07-05 00:00:00','   i have decided to stop my blog or at least for awhile, when life is at its worst its easy for me just to write down how i feel but right now i think i have to work it out in person...................................  '),(9681,3353979,'09,August,2004','2004-08-09 00:00:00','   Kristen can read me like a book, i have been unhappy. Its because i have this feeling that kristen dosent really want to be with me,, i want to be with her, i love her.................  '),(9682,4027311,'27,July,2004','2004-07-27 00:00:00','   I kinda wonder why I am even doing this whole blog thing. I could just keep a journal and leave it at that. Of course, then no one would be able to read it and I wouldn\'t be the subject of one of John Dvorak\'s   ranting & raving columns  . But that\'s OK; I probably never will be...    Enough about me; let\'s talk about ME!    I just got back from  urlLink People\'s Bank  in  urlLink Stop &amp; Shop  in Norwich. The tellers must be under a ton of stress, because the only folks ruder are the poor bastards at DMV (and I do feel bad for them). I was so ticked off at the poor service and exorbitant fees and charges that a couple of months ago, I went in to close the account. I\'m such a sucker. They closed the account, alright. By selling me on another, improved, cheaper account. So away I went, content that I was being frugal and responsible. All I had to do was get my ATM/debit card and I was off to the races.    Excpet that the card never arrived. So after a phone call to the local branch and getting routed to Bridgeport (you always get routed to Bridgeport if the local branch doesn\'t pick up the phone after like, one ring), I was told that the local branch had forgotten to order the card and they would be doing it post-haste.   (Riiiiiight.)   Card still didn\'t arrive. Called the local branch, talked to Bridgeport (see above explanation). This time the story was that the card had been ordered, but it takes 10-15 business days to arrive. By this time, I\'m starting to think I really got my money\'s worth out of this new account. After another couple of weeks with still no ATM card (today), I decided I would visit the local branch to find out what the hell was going on. They told me that the card was mailed 24 days ago. Would I like to cancel the card and order a new one?  Hmmmmmm. I don\'t know; what do you think???  So I asked if there were any possible way to get the new card rushed, since I haven\'t been able to use an ATM in a couple of months that I\'\'ve had the account. Well, yessir, we can rush the card - for a $25 fee that we won\'t waive.    Fine.   Whatever  . I\'ll probably never see this new card, either. I am such a sap. I went in to close the account because they don\'t have a clue how to keep their customers happy or back off on the fees. Instead I have another account and still suffer the same service.    I just think banking should be transparent. You get an ATM card. You use it anywhere. You write checks anywhere. You get your statements in the mail with the cancelled checks. You don\'t get charged ridiculous fees every time you turn around. Simple, right? Apparently this is not an easy thing to do for these folks. I don\'t know if it\'s the people who work there or a bigger problem with the organization itself. Or maybe it\'s a conspiracy against me. Yeah, that\'s the most logical idea. '),(9683,4027311,'26,July,2004','2004-07-26 00:00:00','   OK, OK... So I\'m the last to hear that Van Halen has welcomed Sammy Hagar back to the fold and is on tour or planning one. And they\'re releasing a warmed-over re-released song album (\'greatest? hits\'). Am I the only one who\'s not excited at all about this? I\'m sorry, but there was only one Van Halen, and it ended sometime in 1984 (with a brief reprise when DLR contributed to a couple of newer songs for Greatest Hits, Vol. 1). They were good songs, and there were many of us fans who have too much time on our hands that were hoping for a more permanent reunion. Oh, well... Didn\'t happen. I still have DLR\'s CD\'s. Some were better than others, but you gotta hand it to the guy: he doesn\'t ever friggin\' stop rockin\'.  Or does he? Where is he now? I thought I hear on NPR last week that he is training to be a paramedic in NYC. I can\'t find his web site, either.  So is David gone from the music biz? Was the last CD (\"Diamond Dave\" a.k.a. \"DLR Does Better Covers than Originals\") his last? '),(9684,4027311,'24,July,2004','2004-07-24 00:00:00','   This is my first blog post, ever. I\'m sure that it will go down in the annals of blog history... Or not.    A little about me you can\'t get from my profile (because I\'m too damn lazy to fill in all those text fields):    I am a single dad. It\'s the best job I\'ve ever had (although sometimes I don\'t think I\'m going to be getting a promotion anytime soon). My kids are the best.    I am a technology junkie in semi-luddite clothing. I run a small computer company (OK, it\'s miniscule, as in one-man show). I have way too many computers running way too many OS\'s. Handhelds, too (yeah, they run different OS\'s, too). DSL, wireless network, yada yada yada. However, I don\'t have cable TV. Or satellite TV. Don\'t get me wrong; it\'s not because I think those companies are evil (I   know   they are). I just don\'t have the cake to pay for TV I don\'t have the time to watch anyway. I don\'t have high-def  anything . My TVs are  old . (One of them I bought new in 1986!)  I play blues guitar. What? You didn\'t know that Connecticut is the breedin\' ground for famous bluesmen? Actually, I didn\'t either. But here I am, playing guitars all the time. Electric and acoustic. There\'s like 6 or so sitting around the house. Way back when, my ex-wife used to ask me why I had five guitars. I answered, \"Because I can\'t afford six.\" My current squeeze is a \'52 Tele reissue thru a \'71 Super Reverb. No stomp pedals necessary for killer tone.  I have a love affair with motorcycles, but I can\'t afford the addiction. My current ride is a 1971 Honda CB750, my favorite two-tone color: black and chrome. Runs great (usually), looks not so great. I guess that\'s what New England winters are for: overhauling bikes. If there are any antique bike riders (antique bikes, not riders), let me know.  Friends of Bill: Contact me anytime. We\'re trudging the road to happy destiny together. '),(9685,3475992,'27,May,2004','2004-05-27 00:00:00','   Life.  There have been way too many perplexing moments from my childhood on that all have a common theme. Indifference. When I was three, my child-like spirit shined without knowing shame, guilt, doing wrong, paying for your \"indifference\" from the rest of the world.  My first memory of this would be when I was three years old, a smiley, bubbly me, asking my mother to explain to me why I was not black. Yes, I am white. My innocent eyes, seen the world as a different place, everyone the same, no indifference from the structure our society seeks on what to wear. Red carpet shows on who\'s who and........WHO ARE YOU WEARING? My very fiber that I am made of, can tell you the giggles, the shock, the \"oh my god, pat..............she wants to be black!\".........and the odd looks that came from everyone around me.  That has to be the first memorable milestone for me. Now, do not go believing that I was shamed........or shunned. Infact, my mother gave me my one request.......an   AFRO !  Aye capt\', I said afro.   Personally, as I look back and type this, my soul can honestly speak and say the reason why I wanted to be black.  Every single black person I met, which honestly was alot at the time, were very nice to me. Talked to anyone, or so my eyes could see. Embraced people, and laughed heartily, dancing, joyous, and were not afraid to be themselves. A sense of family, a zest for life, and just being down to earth.......is what I saw.  Which, at that time, was exactly me.  Glancing back through that memory, reminded me of my father. Self taught percussionist at the age of 16, and that man can still jam and keep up with the likes of larry mullin jr, mel torme,  CARTER  from, DMB and more.  My father, which you will see later on in this BLOG of my flippin\' life.............is a great man today, and one hell of a beast from the past for me. Thankfully I have my head on straight......no serious mental issues, just your occasional paxil popping moment. (joke) Dad was the man of the house, he was the love of the family and he also was the first one to show me, and EXPLAIN to me that just because you are different than the rest, means there is nothing wrong with you. Wonderful! So, why am I white and not black?   Life is in color, not black nor white, it has many unique layers of self destruction, self love, self discovery, and paving the way for milestones to happen...........which are building blocks to one\'s character, and future.  Embrace them.  '),(9686,3546623,'26,June,2004','2004-06-26 00:00:00','   Ecstasy equates To the morning dew Reminders to all  We are washed anew Life and memories The past holds the clue Learn now  Convert In the morning dew. Birds sing freely Souls that dance in the night Clearly blessed  Ready to take flight Wishes and dreams Soar towards the sky All within reach Life so fresh Blessed in the morning dew Slowly I rise From the tumbling sleep Gentle waves of moisture Gracing my feet The path is in front The road is behind Baptized each morning Drenched in holy morning dew Each day is a gift  Cherish and laugh Tomorrow is too far Today is the plan Live in the moment Walk in the air  Roll in the grass Let it dance in your hair Seconds of youth Vibrate and sing Flooding the soul Ecstatic tears  The Goddess brings Disguised and hidden Do you not see? Here is the key Breathe in and deep Refresh the life within An instant, forever Be the morning dew. Birth of the morning As the foliage is feed Colors and songs Waltz through my head I hear the whispers Feel the calling Forever I live Kissed by the morning dew. dlt Â© june 26 2004  '),(9687,3546623,'20,June,2004','2004-06-20 00:00:00','   Essence of a Soul Inglorious erstwhile, Struggles forth, In darkness is met.  Essence of a Soul, Mute and Deaf, Screaming and shouting, Pain forever left.  Essence of a Soul  Once barren, forlorn, Gasps, pleads! Sunrise innate!  Flaming hope. Cherished faith. Willingness fresh. Love stark!  Two dance as one; Heart and mind, Nascence pure and light Essence of a Soulâ€¦ Anew!  Lady Lorraine du Lac Â© June 17 2004  '),(9688,3546623,'09,June,2004','2004-06-09 00:00:00','   Yellling, Screaming Dancing, Sweating Women divided, United! Fighting it seems, The union of two, lost halves, now found. Sacred Marriage  The calm and the strong. Dark and the Light. Embracing with love, Divine is the time, The torch held as one!   Lady Lorraine June 17 2004  '),(9689,3546623,'09,June,2004','2004-06-09 00:00:00','   Walking across dew kissed grass Serene heart gently sighs  Moments seconds minutes trapped in glass Lyrical soul aspiration flies.   A blizzard of separation Discovery of distance Time to uncover Our truth to reveal.  Breathing deep as the stars shine and sing Moonbeams dance as we grow more distant Our paths growing away  Limbs reaching stretching away from the core  Experience screams LISTEN Enraptured, enchanted the breathe of the muse A gift so rare and difficult to render Fate as you leave, her truth  your heart transfused!  I smile to myself as slowly you become Listening to the muse You glow and you shun The past and the future for us there is none  DLT Â© June 2004  '),(9690,3546623,'09,June,2004','2004-06-09 00:00:00','   He skitters and he scatters, He zips and he zaps, It is you he flatters, He knows we are a couple of saps!   The music plays happily, Dancing to the beat he performs never a slip! High in the sky, he dives so craftily Floating and soaring he ends with a flip!  We laugh and we giggle!  To the front of the bee, I saw your finger wiggle! \"Sit\" you shouted with glee!  Seducing you, the bee has his fun Tears run freely as we jest, For your love he has won! A perfect moment held dear in my breast!   For Tim My teacher My student  Of living in the moment!  Blessings  Denise Â© May 21 2004  '),(9691,3546623,'05,June,2004','2004-06-05 00:00:00','   December 28 2003  Sunday  Quabbin      I sat upon the shores of the Quabbin today, my inside quivering as I faced a truth of my future. The death of a dream, a hope, another life. I wander now, start a pot of coffee, take the dog out, put on music, smoke a cigarette and another cigarette. Anything but sit here and write. I know once I put this on paper it will be gone, maybe not totally but it will be real and I will have to move on. I close my eyes listening to the harmonic melody playing, drums, crickets, wind instruments, a deep pulse moving through me. I pray, help me God to write and move on.  So again I take you to the shore of the Quabbin with me. A man made reservoir that for me holds the secret of man and Godâ€™s unity. You can not see the man made elementsâ€¦ although you know that they lie there towns deserted beneath the waters surface, all you can see is the creative beauty of Godâ€™s wonder, trees, water, sky, rocksâ€¦life. I come here often. I find balance at the waters edge. I can hear the childrenâ€™s lost laughter from the sunken towns echoing amongst Godâ€™s trees. The lost laughter of children, this is why I am here.   As I make my way to the waterâ€™s edge I pass fellow travelers, saying Good Day! How are you? Wishing them well, moving on. It is a pleasure to share a moment with another human being, extending a genuine smile and nod that you acknowledge their existence . Many others come to this waterâ€™s edge to ponder their past, present and future. The energy is strong, healing it draws me forward but still I question am I ready?  As I make my way to Godâ€™s chosen spot for me, I take pictures of the water, rocks and dead lifeless trees. I am always drawn to these gray ghosts standing silently, drooping by the waters edge. I see myself as one of them. Dried up, dead, a vessel where no life will spring from, no buds leaves, fruit or children.   I walk towards the waters edge, I know that the man I love is right. I must face and deal with this. If I do not there is no hope for us. I do not know if I am ready. There is a lulling comfort in the pain of holding on. I brood over this truth, but my love for him is stronger than the comfort and I know I must travel on. My steps move me forward to the balance of the waterâ€™s edge, silently praying for the strength and courage to face the dark fear that bars me from my future.  I sit now on a cold flat stone balanced on the edge of land and water. I face the winter sun, a small river of ice runs toward the water beneath my right foot. My back is against the water, I know when I turn to face the serene calmness that holds the healing power of the water, I will have to face the laughter of the children. My cigarette is done, put out now in my back pack, do not want to leave any evidence of my being here. What am I waiting for?  I hear an airplane in the background its motors beckon to me to turn face the water and look, see and know. I pray:  I ask you God for strength and courage                                                I ask you God to stand by me                                               In order to move closer to you                                               I ask you God                                                To help me                                                 Help me  I turn and face the waterâ€¦.  I see my shadow on the waterâ€™s surface dark against the warm brown of the sandy beach. Gentle ripples move across the water towards my shadow, beyond me Godâ€™s love in those ripples radiating around, across and through me. It is time.  I will never have any children. I always thought I would make time. Reality is I did not I made choices and now my time is past. It is too late. There it is said. The sad truth is that I still want them, but now I know that this is not to be. This is so hard to deal with, especially when you are a woman who always gets what she wants. I must face this truth. There is no other way, if I do not I will be a prisoner to the past and forever be the gray ghost of a lifeless, drooping tree at the waterâ€™s edge.  So now I sit here, the cold from the stone running up my spine and I cry for the mother I will never be. I cry for the children laughter I will never hear. It is so hard to let go of this, I do not know if I can do this by myself. Do I have to?  For the first time in my life I heard from a manâ€™s lips the words \"It is not your problem Denise, It is ours.\" This is more than I could dream of. Can I trust you with this precious secret that lies within me. Will you help me?  I look down the gentle waves of the water reach the edge of the thin ice that bridges the water to the land. The waters gentle movement creates a pulse living within the thin ice. A beating heart strong steady giving life to hope.  The tears stop, I stand and balance upon the rock. Thanking God and Mother Earth for the healing wisdom they have bestowed on me. The winter sunâ€™s rays grow stronger as if they say to me, \"You are welcome!\" I gather my things and move towards the woods, a short cut to the road and home.     The Woods  Death of Desireâ€¦.  Wouldnâ€™t it be so easy if all I did then was gather my things, walk to the woods, get in my car drive home and all is well in the world. Sadly that is not the way towards enlightenment or reality of the world. I hate to tell you but suffering, pain, anguish are the bedfellows of the death of a dream ,our desires.   I made my way through the woods, not a wise choice. Safety and comfort lies at the waterâ€™s edge. I am not one for safety and comfort, my life has its foundation on the rocks of pain. Moving through the pain a place I know so well is the only way towards my future. Only then may I may find the truth of safety and comfort.   The death of a dream can not be communicated through words, paintings, sculpture, music, aromas, tasteâ€¦ nothing can describeâ€¦ it is all hope and all death rolled into one.   I walked through the woods, the gray finger of lifeless trees grabbing at meâ€¦other barren women pulling me towards them telling me to stay. I will not. Colors of yellow, blues, greens, orange and blood red swirling around me creating a tangle or rusted barbed wire that ripped at my skin peeling off layers of my true self. Ever breath deeper as the barbed wire wound itself around my insides my heart, blood pouring out from meâ€¦   The music startedâ€¦ loud drums, chaotic rhythms, pounding my skin split with screams that only I can hear. I can not breath I shake but pull myself forward .. for I know that this too shall pass. Everything hurtsâ€¦ the colors are swirling black nowâ€¦ the music and screams have no beat just a high pitch whineâ€¦ my heart beats so fast my breathing is rapid. I am at the roads edge. It is done.   I scream out loud just to hear that I am still alive. I do not care if anyone hears me. Slowly I hear the gentle rush of a small stream nearby, the winter sun comes out and I begin to breathe. A single tear caresses my cheek. I know this will not be the last one I will shed for this dead desire. There are no words, no art, no poetry or music that can make one understand the power that lies between the hope and death of a dream. It is a spiritual connection with the light and dark that lies within all of us. In order to learn we must draw our swords and battle our way through our own dark woods, releasing ourselves from the pain of precious dreams of our past in order to gain perspective our what grand dreams God has for our future. I know what I must doâ€¦  '),(9692,3546623,'05,June,2004','2004-06-05 00:00:00','   Alone now wandering room to room, Echoes of liquid laugher floating down the hall Embracing me with honey warm joy.  What I ask made this so special? Moments of shared quietude Simple safety of contentment reigned.  Drenched in music  Time flew As our souls stood still.  Silent smiles Speaking grander than any words Our inner selves slowly emerge.  Childhood memories  Danced through our night Sweet sharing revealing racing hearts young .  Alone I sit  This serenity so new A blessing I know all because of you.  Denise Tetro Â© May 2004  '),(9693,3546623,'04,June,2004','2004-06-04 00:00:00','   You hide in the bushes and lurk from the trees, silent you think as you creep through the weeds. Stalking the prey now wise to your needs;  I sense your presence,  forever I see.  You listen and contrive, from the outside you watch silently. The saunter of the walker as her faith  thrives, her steps strong and steady as she glides defiantly.   Oh dear muse!  You stand in the shadows of fallen oaks, like a torrent rain your empty words, drenching your soul, denial forever to soak!  Courage brings you to my front, to know the secret of the walk. But lies, doubt and broken trust, are walls that cause you to balk.   I see you now, without your daystar glitter, only a man, lost and bitter.  A raggedy man,  who runs and retreats. Just a man, who has wornout the soles of his feet.  Lost in the mind, dead of the heart, you exhaust me! I no longer play my part!   Denial is a snake charmer...  that will whisper sweet lies in your ears,  blind your eyes,  harden your heart,  deaden your soul...   So dear muse, I live to feel the blood in my heart, the singing of my soul, I close my ears for I now have control.  I pray someday, for the return of your daystar glitter.  Once who knew he way, a man so brave. Desire bright that burns, living within the oak,      and a life lead no longer bitter.  March 30 2004 Â©  dlt  '),(9694,3546623,'04,June,2004','2004-06-04 00:00:00','   The  Priestess of the Fallen Oaks    I fell into the moon And lived in the land of the midnight sun the gate swings open and I am greeted by a muse  gently he whispered love laced riddles in my ear speaking to my heart, my soul opened each willing to hear  my mind runs to the darkness... Taking the muses hand  we walked through the garden in silent seperation the light shines on Alone with the muse  Sadly our hands slip out from the tender embrace surrounded by a lush green of a dream the riddles danced through me the answers sleep upon the muses lips I yearn to know the answers but still I am not told there will be no kiss  so I may breathe in the answers.. anger flares...  why? I look at the muse  no words..silent sadness the riddles flying like hummingbirds in the air The muse slowly fades in front of me bursting into star dust  surrounding me like a tornado of Divine light Soaking through my skin the garment of my soul I am a day star in the land of the midnight sun...  Without my muse I am lost. my starlight fades in and out traveling the overgrown paths finally lost in the wilderness I forge my own ...   searching  for the answers My shining garment is soiled and torn  My soul exposed and vulnerable I am trailed by the hummingbird riddles the whirl of their wings throbs noisly in my mind lost in the darkness still memories of my muse glisten upon my skin a  torture of the lessons learned Hope struggles with each dying breathe Desire and longing  I firmly clasp at my breast my heart beats for the muse to look within the all knowing eyes and to see my granduer as he once did   I awaken in a meadow bees busy at work floating from stem to stem bringing new life into the center of the moon I gaze upon a mighty oak tall and strong My hands long to roam across the rough bark and hear the whispers of the years A lonely oak who longs for companionship I see the changing of many seasons a hawk flies above and watches  my progress A voice I hear with wise words that draw me beware for what you seek It will both amaze and bring you great sorrow The seed of the old oak  dropped from the sky the struggle for the roots to receive the life\'s water Clutching at the earth its mother each morsel of soil clings to the roots  painfully they spread and gain strength Feeding from the full breasts of the Goddess I feel the years of the oak passing through time standing tall young and proud Alone he is not living among a forest of friends slowly the disease of man took away their lives Only shadows of fallen oaks grace the meadow Towering alone now waiting to be called forth  ripping it\' s roots from the mother earth The goddess lives within the pulp She cries for the shadows of lost souls  The the dance of the leaves of her fallen soldiers  that once stood proudly in the meadow I call... softly, gently a lullaby yearning to see the miracle To feel the earth beneath my feet  rumbling  ripping itself from the mother The oak struggles for freedom as the hummingbirds take refuge within the branches I hear the earth give way   thunder and screams... groaning ...  straining for release I stand aside... my tears  now the life source for the exposed roots  I reach forward  crystal teardrops drip from the tips of my outstreched fingers.. like rain in the desert falling to the roots mingling with the milk of the Goddess each splashing into a thousand stars that sail towards the sky   swimming with the hawk Alone I stand waiting for when the oak will be ready to live in my star shine and I in the center of the rings One with the Goddess her Priestess of the Fallen Oaks...   Love Light and Peace  Lady Lorraine du  Lac  '),(9695,3546623,'04,June,2004','2004-06-04 00:00:00','   Outsiders      I remember coming home one day when I was 14 years old cryingâ€¦my mother opened the door with a worried look upon her face. This was an unusual site for herâ€¦ you see I never cried, at least not in the open where others could see me. My mother was used to a little warrior marching into the house, smile on, books in hand and always a new tale to tell from the day. This young girl who walked in with red eyes brimming with endless tears was a new fragile person to her. She looked at me asking what could be wrong. There are times in life when even the strong break.   Earlier that day at gymnastics I had introduced my new floor exercise routine. I worked hard and long for many hours down in the small judo room putting this 1 minute and 30 second debut together. I had played the music Exodus over and over at times breaking it down to 2 or 3 seconds in order to perfect the right movement with the notes from the piano music. At night when I laid in bed I replayed the music in my head and each movement until I visualized my body in perfect unison with the music. I loved this routine. I was proud of it knowing that it reflected me... what lived with in my spirit. It was mine.. not a routine made up by a coach for me to mimic but me dancing, leaping, flipping and spinning one with the music. I choreographed it to reflect how I felt and wanted to communicate to the judges. It was me.  That day Eileen my coach asked if she could see the new routine. I usually did not perform my new optional floor ex until the meet. Always one for the dramatic entrance! But that day her smile was large and I knew she really wanted to see my work. Nodding okay, I handed her the album (yes album! Vinyl!) walked out to the middle of the carpeted mat centering myself and struck a pose. The music started slowly a building force and as usual I got lost â€¦ a point where there is only my spirit, the music and my body. I worked my way through leaping with precision each arm movement joined with a beat. As the tempo built I performed back handsprings ending with a perfect full twisting somersault. (I was the only one who could do this trick so ego was involved!) Slowing again as I worked my way onto the ground now one with the mat as I gracefully lowered myself onto the carpet for the final dramatic pose. Done!   I remember coming back to the gym from the music center that lived within me. Eileen standing there no wordsâ€¦ but slowly the laughter beganâ€¦all my teammatesâ€¦ the girls who I shared hours of my life with in this gym laughing at me. I remember Dayna who was my closest friend on the team saying, \" You think you will win with that????\"  They began to mimic some of my moves with clumsy gestures and laughed harder. I recall hearing Eileen telling them to stop as I ran out of the gym. Tears rolling down my face I gathered my things and ran home. I was never one to run before so why now?  There I was then standing in the door my mother asking me what could be wrong? How can a 14 year old girl explain to her mother the exhaustion caused by living outside the world of the norm? I could not explain it to myself so how could I voice this torture to my mother who loved me so much? It was not just my gymnastics that classified me as an outsider, rather it was everything about me. The way I danced, thought, dressed and what I believed in. It was the boys I fell in love with, the fact that I loved school and rather be there then anywhere else! It was how I sat in the back yard for hours staring at the sky filled with clouds, the birds and trees. The fact that I would rather read a book than talk on the phone. It was the friends I picked and teachers I loved. It was everything about me and for the first time I realized that I was an outsider different than others and I did not fit in! This was the reason for running and the tears. I sat on the couch and told my mother my tale of humiliating laughter from the girls I believed to be friends. To this day I remember my mother smiling, shaking her head and grabbing my hand. She took me into my bedroom where the walls were lined with trophies, ribbons, plaques and medals. She looked at me with pride and said \"Denise, Do they have all of these?\" I miserably shook my head No, She replied \"This is what being different.. unique (nice way of putting it) will always get you...\" as she ran her hand across the hundreds of awards I had received through the years. She finished \"So why would you want to be like them?\" My mother sat me on my bed facing those ribbons and medals telling me about artists whom had been laughed at the same way and now where considered to be masters. She told me of great scientists and leaders who had the same ridicule laced through out their lives and now are looked upon with awe. My mother told me that those who are willing to stand tall and be different change this world and she believed that this power lived within me. She wiped the tears away and told me to be strong. She gave me that day a gift I will carry to my grave. I performed that routine with more passion then I thought possible. I won over and over again. Slowly my fellow teammates laughter faded from my head.  This 30 minutes spent with my mother at the age of 14 has returned to me numerous times in my lifeâ€¦. always giving me strength to be different...return to school at the age of 33 and graduate with honors. To search in areas of life that others find \"crazy\" or repulsive. When I was told at the age of 21 I could not possibly dive from 113 feet for a world recordâ€¦ I climbed that ladder and proved them wrong. Have I always made the right choices? Who knowsâ€¦I dance to a different beat and that is just fine with me today. Call me crazy, insane, whacked or just plain nuts! I will always hold my head up high... remember the words my mother told me and open the next door in my life with hope, joy and insane anticipation of the next adventure!   Lady  Lorraine  '),(9696,3546623,'04,June,2004','2004-06-04 00:00:00','   With in and With out  Dark and  gloomy, Once  no light; Lost and lonely Each day a fight.  Moment by moment, Day by day, Time stood still, A soul to pay.  Slowly emerging, Pushing and straining, The light up ahead, The cure for the ailing.  Manifested of pain, Balance unknow, Light from within, All to atone.  Power achieved, Freedom lives, Down on the knees, Surrender gives!  Within and without, There is no difference, Divine love sprouts, Living in his presence.  Two become one, Spirit and heart, True to the shaman, All have a part.  The river runs deep, No longer a struggle, Minds need not sleep, The spirit immortal.  Be oh so silent, Live in the present, A past violent, Tomorrow repent.  Within is the light, Love is the key, Yours to ignite, Answer the  plea.  Time is the essence, Forgive and forget, Living  luminescence, Paid is the debt.  Pain so bold, Forever to hold, Surrender my dear, Banish the fear!  Blessed am I, Who lives in the sky, Fruits of  ones labor, Grace! Divine Favor!  Lady Lorraine May 28 2004  '),(9697,3546623,'04,June,2004','2004-06-04 00:00:00','   The Spiral   I sit here in my house in the country, Barney my dog curled up tightly in his little bed sleeping away so peacefully. The sun has just risen and streaks of pink reach across the new days sky slowly searching for the hope that lives just beyond itâ€™s touch. It is a new day, another chance for peace. Sitting in my study sipping the first cup of hot coffee of the day, I sit and wait for the calm. My cat, Obi leaps against the glass sliding doors struggling to play with a little bird outside the window. The bird does not moveâ€¦ it just sits there and stares at the crazy little black priestess. She sits there perplexed at why that bird is so unattainable. It looks at her and flies away I can hear it snickering with every wing flap. It leaves behind a confused female wondering why she can never break through that glass wall.   The calm is not easily achieved this morning. I look around my little home. The floors need to be vacuumed, books need to be placed on shelves, clothes washed and dishes done. Hundreds of CD cases lay emptyâ€¦ with a stack of discarded music along the wallâ€¦ music I can not listen to. My keyboard stares at meâ€¦no words fly from my fingertipsâ€¦WHY CANâ€™T I DANCE? What happened to my precious music vibrating through my very essence? Music my blood, running through my soul bursting into a white light so immense I can only believe it is God. Where is it? My energy dissipated, I am lost in the darkness. A tear runs down my face as I mourn the loss of my music and the dance. A stack of art books lay discardedâ€¦ nothing speaks to me. All my loves, my distractions, my drugs stripped from me. Now my house cluttered and littered with my disenchantment of this life is a work of art that reflects my mind. Even the coffee is flat. Normally I take a deep whiff of the warm embracing aromaâ€¦ just before I sip with delight. Gulping today, I smell nothing. Cigarette after cigarette smoked and put out, my ashtrays overflow. Food tastes bland I numbly place each bite in my mouth chewing without taste.. a function only, sadly I continue to chew even when I am not hungry. Finally, I reach down to that sleeping bundle of love Barney, stroke his tiny headâ€¦.my heart breaks! I can not feel the love.   My senses are dyingâ€¦for a sensual woman so soak in the warm summer breezes, rich silk of chocolate laced across my tongue or the tears released from viewing a shooting starâ€¦ this is worst than death! My heart sunk to the bottom of the sea lost with the city of Atlantis as the thought of living like this forever drenched my being. No more lost moments in union with a Mozart symphony, heart racing with the reading of Shakespeareâ€™s sonnet 18, flying to God as I dance through hours of music or the sharing of a soul holding my loveâ€™s hand. This is too much to take. He is gone and I am alone with the pain. It is all about my love for him and nothing about my love for him.   When and how did this happen? More important why? How did I get here this desolate tormenting moment in my life? I close my eyes, breathing slowly knowing where I have to goâ€¦.not to last night or even 2 weeks ago when the pain started, no it began years ago! That is the road I must travel now or forever live a false life.   I had believed that the spiral down was finished, it had started June 9,1999. Newly graduated from college with honors I was living in the middle of the universe. Happy, proud and ready to start again at 38 years of age, the world had laid a red carpet before meâ€¦many in fact which way should I go? I laughed regularly with the excited joy of new adventures. I opened an email thinking it must be joke or a riddle from my Uncle Jackâ€¦ the replacement father I loved dearly. What I found there was one of those moments when time stands still. My grandmother on my fatherâ€™s side had died. I had not seen her in years but still the ache of loss echoed through my heart. I was determined at that moment to go to the wake! SO WHAT if my father would be there. He canâ€™t hurt me nowâ€¦ I can do anything, I thought to myself. Did I not just accomplish more than anyone dreamed I would in this lifetime? Strong with emotion I read onâ€¦â€¦there was no mention of any of us. No Denise, Frank or Patti when they listed grandchildren. Time stood still and the darkness came dragging me out of the center of my universe to a catacomb of my mind. We did not exist. I was nothingâ€¦ as if I had never been born. It was at this moment that the black spiral down to my core beganâ€¦. I thought.  The next four years I threw myself into the material world producing money and power. Funny I did not become joyful, in fact I sunk deeper the harder and more powerful I became. Until after a disastrous love affair with a man whom only thought I loved and just realized that the pain came from the fact that I could not save him from the death living in his eyes. This realization had me driving towards a bridge at 90 miles an hour praying for the courage to close my eyes and just aim. It was then I heard my voice again. It had been silent for so very long I almost did not recognize it! The voice screamed STOP THIS NOW! And I woke up from the dark catacomb swerved away and pulled over to the side of the turnpike and cried for the first time in years. I choose to live at that moment, to crawl out and learn to love. My walk towards the light began right then and there at 41 years of age, sitting in a dark car on the side of a lonely highway.  So I ask myself how could I be here now? No more tears to cryâ€¦.dehydrated from the racking sobs of the last week. I gave this man all of me! The first time I was honest, open and loved unconditionally. Why did he abandon me? A sob crawled to the base of my throat, I pushed it down and screamed Not Now! I am so close. In my deep state I walked past this man who I blamed for this pain, pushed aside the anger and need to have him love meâ€¦. for as I said it is all about my love for him and nothing about my love for him. I so want to reach out and touch him feel his hand in mine one last time but I know it is not him that holds the keyâ€¦ I gave it away along time agoâ€¦that is where I have to go.   My mind holds many locked doors from my childhood. Slowly through the years I have found the courage to open each buried memory and face the past. Always the most painful times of my life both then and now. Each time it is like the lifting of a veil that covers a dead child. As I lift the veil I watch the child breathe againâ€¦.resurrected in order to take my hand down the meadow of pain I will walk.   I am 9 years old. It is my brother Frankâ€™s birthday and we are all at my Grandmother Renaâ€™s house. My fatherâ€™s mother did not tell us he would be hereâ€¦ but there he is, my mother is not allowed with us and he wants us to sit in the room with him. I sit on the couch alone and fearfulâ€¦.my brother and sister are on his lap gleeful to see him. He calls to me but I do not want to goâ€¦and I want to go with all my heart. Tears run down his face as he pleads for me to come to him. My mother is divorcing him and now is professing his love for us. I stay strong, fueled by fear and biting memories of the past tortures he has placed on me. At the same time I want to believe him, I am dying to believe the words I have never heard from him beforeâ€¦..\"I love you Denise.\" The desperation for my fatherâ€™s love winsâ€¦..I go to him.  His tears run freely, I believe he loves meâ€¦ my sister and brother hug me as we all sit on his lap. I let the walls down, those strong thick walls built through years of beatings and torture! Amazingly, they fly down within seconds with his words and tears. It is then he hugs me close, it is more than I ever thought I would know from him my father, the man with the belt and a rage so bright it burned in the sky. He loves me! I relax and fall into the hugâ€¦.he whispers oh so softly then â€¦.so only I can hearâ€¦. I donâ€™t love youâ€¦. never willâ€¦. No one can and ever willâ€¦.I fly back through time, straight past the man I love to the cluttered study and the silent keyboardâ€¦. The sobs rack my soul. You see it was all about love â€¦.  My eyes open when I am finally doneâ€¦there is nothing leftâ€¦.but the music and the keyboard. I look around my houseâ€¦a home I was so willing to throw away â€¦just to seek loveâ€¦.Love lives here! My walls painted in soothing yellows, pinks, desert sands and vibrant reds hold hours of my mothers and my laughter and love as we splattered paint and sweat poured down our backs. My rows of bookshelves lining the walls in front of me, laugh back.. holding joyful memories of our adventures in carpentry, as I struggled to hand twist each and every screw that now holds up the hundreds of books lining my walls. Wonder and amazement vibrate under my kitchen floor â€¦ I think back to my motherâ€™s laughter as I was covered with glue and struggling to get the Armstrong linoleum in place. When we were done, we sat for an hour staring like children at the job we thought we could not do. The tiles of the small bathroom hold my tears of the dark time before the bridgeâ€¦my birthdayâ€¦ when I felt no light in my soul and my sister and mother came to force me to live. With each and every tileâ€¦ my darkness and pain left my being and were buried deep with in the tiles grout. Love lives here. I do not have to leave. It breathes within the walls of this house and the garment of skin that holds my soul. I never have to go back to that couch living in fear and desperate to hear the words \"I love you!!\" It lives here. The pain lifts and I realize this man who I blamed earlier for the death of my dream has given a gift far greater than roses, candy or diamonds for Valentines Day! I am freeâ€¦the music is playing! I have broken through that glass wall and have played with the bird! I know now I can live and love, the calm has comeâ€¦.realizing this I get up to make the bed, my voice is humming and Barney bounces at my side! I sigh thinking, it was all about my love for him and nothing about my love for him!   Thank you all....  Love + Light  Lady Lorraine du Lac  '),(9698,3546623,'04,June,2004','2004-06-04 00:00:00','   A dance, new we learn today The steps are foreign and strange Slowly we go as the music takes us away We question now are we ready for the change?   We hesitate with doubt Still the song is too strong We wonder does our power hold any clout  A new dance that will not be rushed along Far too profound This waltz in our dreams Gingerly we follow the rhythm of new ground  Cautious not to go to extremes Clumsy we are A first, I let you lead Trust is our grace that leaves no scar Unselfish in the heart our dance is not one of greed   So dear friend My spirit does not fear Joy laughter and smiles we need not pretend Let us dance and learn for the hand offered is sincere.  Denise Tetro Â© April 25 2004    For Tim...  '),(9699,3546623,'04,June,2004','2004-06-04 00:00:00','   Once upon a time there was a land without fairy tales. Glass slippers and a perfect fit, magic wands, fairy dust and godmothers who save the day never danced in the minds of little girls who existed in this realm. Wise trees speaking their peace, tiny dwarfs with hearts bigger than the universe and prince charming with all his undying love were banished from the hearts of all tiny maids. This was a land where chivalry was a foreign language in which little boys lacked the message and comprehension. A tragedy for not one little boy knew how to rescue a damsel in distress, rather they walked on by thinking it was not their problem! Slaying a dragon was a laugh, who knew what a dragon was, they were too busy trying to live through the day. The excited and bonding chant of one for all and all for one never was uttered. Trust and depending on another were signs of weakness. Kindness did not exist! Shadow dancing children living in a colorless world where sarcastic laughter echoed through their dead hearts.        The children of this world grew up isolated, solitary entities in a cold world without hope, dreams, faith, imagination and self-confidence. Fear was their only friend and hero. A dark world with no light and sadly they never came to understand the power of a smile or safety of a gentle caresses. Faith was a mystery, while respect was dead and honor was a rumor.        From the start these fledglings had a strong understanding of the position they held in this life. Love was impracticable, an extravagance for the totally insane or those few individuals of wealth and power. Sex was a function, a diversion from the pains and frustrations of this lifetime. A physical act to release the silent angry screams that built up over the years. Violent and passionless with no notion of sharing or bonding with a soul mate, leaving them always empty and ashamed. Often an act that only produced more children vacant of the wonders of childbirth only an endless cycle of populating a dead world.        A world barren of simplicity, grace of an pure moment of joy, no giggles or laughter that released purity and light into the universe. Insane notions of creativity, imagination, comradeship and freedom were locked away in dusty catacombs that no one ventured into. Lifeâ€™s simple pleasures denied, the children of this world grew to be black souls that walked grasping on to the hand of FEAR ignorant of lifeâ€™s abundance. Suffering was their comfort and without suffering they had lost all meaning in life.        In this fruitless world, the childrenâ€™s lives were void of hope and dreams, there was no other way of life. Envy, jealousy, hatred and betrayal thrived in this atmosphere. Life was bitter, gray and depressing in which they could only grow into adults that lived lives of quiet desperation. A world slowly becoming our ownâ€¦.. Hold your breathe, close your eyes, listen quietlyâ€¦.. you can hear their silent screams roaring through the universe.   Copywrite 2004 Lady Lorraine du Lac  1999/2004  '),(9700,3546623,'04,June,2004','2004-06-04 00:00:00','   If I had only known. Forever had itâ€™s limits. Forever is not infinity. Only a concept of manâ€™s mental abyss! Forever mere seconds, Forever was now,                   no future                          no tomorrow                              Just now.... If I had only known. My lips would have stayed on yours a second longer. Let the warmth of our hands joined linger for one more breath. I would have felt one more beat of your heart. Swam a second longer in your eyes. Inhaled your soul a little deeper. Cherished all these things for a split moment more. If I had only known...  Forever was only now...then...  dlt Â© March 16 2004  Lady  Lorraine du Lac  '),(9701,3546623,'25,July,2004','2004-07-25 00:00:00','   The day slowly wanes, Our camp is all set, Stacking the wood, Pressure we forget.        The first flame flickers, A tiny little dance, Growing and burning, Luring us into a trance.        Slowly we open, Our pasts float free, Stories and laughter, No where to flee.        The fire burns brightly, Calming our doubts, Smoke swirls around, Contentment sprouts.        I see who you are, Simple and complex, Outside so calm, Harmony&nbsp;inside perplex!        Next to the fire, I see your ease, The man before me, I love to tease!        This something that grows, A fire within, Open to me, release the flow, Here we begin! Dlt July 25 2004 all rights reserved  '),(9702,3546623,'24,July,2004','2004-07-24 00:00:00','   The tears slide from inside to out, This pain unbearable no doubt, Torn and bleeding, The time is coming, Decide I must, Forcing me to be, Which one is he?        A crime is this? Love of two at once, Separate, sacred each love strong. Lurking in front, The decision I canâ€™t prolong. I toss and I turn sleep I hunt, By my side which one belongs?        Two loves so diverse! One breathe held, patient is he, The second in trust yet to immerse. Safety is there in both that I share, Different in form leaves not a clue. My choice unclear, burden I bear. Tormented soul whispers adieu.        A crime I commit, To whom I submit? From one I must part, Matters of love a fine art, Emotions unsound, Ecstasy in one is found. Two men loved, one a future bound.        Dlt July 24 2004 all rights reserved  '),(9703,3546623,'21,July,2004','2004-07-21 00:00:00','   My world keeps changing,&nbsp; &nbsp;&nbsp;&nbsp;&nbsp; from red to blue.  I close my eyes,&nbsp; &nbsp;&nbsp;&nbsp;&nbsp; &nbsp;revolving thoughts of you.  Time is so short,  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp; I feel I must rush.  Oh dear lord,  &nbsp;&nbsp;&nbsp;&nbsp; my spirit self doubt will crush.  I stand before you,  &nbsp;&nbsp;&nbsp;&nbsp; pleading for release.  Earthly demands,  &nbsp;&nbsp;&nbsp;&nbsp; I scream now to cease!  My world keeps changing,  &nbsp;&nbsp;&nbsp; from red to blue.  Oh dear lord,  &nbsp;&nbsp;&nbsp; I so need you!  dlt july 21, 2004 all rights reserved  '),(9704,3546623,'19,July,2004','2004-07-19 00:00:00','   Exhausted and wary,  She parts through the trees,  Fingers clinging to her,  Crying and longing,  For a minute to rest,  The priestess stands staring,  Little hope in her breast.  The valley so stunning,  Mesmerizing and cold,  Crystal mountains, rivers of ice,  Siren sounds echoes,  Harmony of a Valley of Glass,  Seducing songs glide,  Gently her hand does one grasp,  Come now, dear one,  Your time to rest is near,  One step forward,  No more to bear.  Her blood answers the source,  A call she must heed,  It is deep within her,  Born of his seed.  Her first step slices,  Straight through to her soul,  Rich blood gushes,  River so deep,  Rushing from her,  Soaking her feet.  Drenched and swooning,  Stop she cannot,  Voice to answer,  Rhythm, so strong,  Flows through her veins,  Pouring forth freely,  She is grateful for pain.  The moon beams dance,  On rain soaked glass,  Blood flows freely feeding the grass.  The Goddess drinks deeply,  Grateful and parched.  Gushing forth,  Ancient thirst,  Only blood can quench,  Knowledge so forbidden,  On the knees,  A heart tells the truth,  Torture is sweet,  Sensation relieves,  Pain is a gift in order to heal.  A mind so sharp,  Tumbles and shatters,  She fears her faith,  Ego she flatters.  Help me she pleads,  Inhaling thick mists,  Swirling around her,  Bats and spiders snicker,  She calls to them slowly,  Please end this I plead!  As the web grows thicker,  Wet and sticky like jam,  Childhood joys fade and flicker,  Silence is held,  Burying the screams,  Her life floats away,  Frozen and bleeding,  Her muse emerges from dreams,  His face is in front,  Changing and gone.  Frigid and glacial,  Crimson and clear,  The calling pours forth,  The essence a tear.  Beating heart shatters,  Solid from fear,  Gasping for breath,  Liquid air does she breathe,  Drowning slowly,  No time to atone!  The past is in front,  Future behind,  For neither do matter,  A moment at hand,  Floats in her mind.  Sinking like quicksand,  A soul pulling away,  Looking down upon her,  Nothing to say.  Suddenly from nowhere,  Arms around her embrace,  Life of the oaks circles,  Her life,  A moment,  Her soul to grace.  Flying to the branches,  Hummingbird whispers,  Life so fragile,  Alone no more,  The oak will not fail,  Strength so pale,  Lean on another,  Message to send,  Learn now she must,  Or her life will end.  DLT July 19 2004 all rights reserved  .  '),(9705,3546623,'17,July,2004','2004-07-17 00:00:00','   You snuck up behind me,  Sliding out of the blue.  I did not want nor seek,  This love I feel for you.  &nbsp;  A butterfly I desired,  With dew kissed wings.  To fly carelessly,  Never to dream.   Laughing and carefree,  Forever I seek,  Fast moments of pleasure,  Never \"Forever\" did I speak.   Denial is lost,  This feeling will last,  The door you opened,  I must deal with the past.   Alone you leave me,  This I must do.  I pray I might see,  The one&nbsp;is it&nbsp;you?   dlt july 2004 all rights reserved.  '),(9706,3546623,'15,July,2004','2004-07-15 00:00:00','   Let us see how this comes out.. I joined this new webring...     urlLink    urlLink ?   urlLink Redhead Blogs   urlLink #   urlLink >  '),(9707,3546623,'11,July,2004','2004-07-11 00:00:00','   A life so full, One of grace, Daily I strive, To balance the pace.  From the heart, Truth so bold, Desires denied, Trust in the old.  Beliefs are soon shattered, Confusion exists. Faith so shallow, Praying for bliss.  The mists of the path, Foggy and dense, Lessons of suffering, Joy sings from the fence.  Slowly I learn, Why must I suffer? To give so easy, To receive is tougher.  Open the heart, Risk to be taken, Balance of two, Not be forsaken.  Alone I now sit, Life scattered and tattered. Of my own making, Does it really matter?  Strength from within, Now I must find. Depend on one self, God in due time.  DLT Â© July 11 2004  '),(9708,3546623,'10,July,2004','2004-07-10 00:00:00','   The flame burns bright Moths circle and chant Around and around they go Back off they cant  Drawn to the motion The shadows and heat Moths dive deeply Their ends do they meet  The men dance around her They are the moths to the flame Circling and diving Screaming her name  The muse knows not  A flaming heart she holds Hope she clutches close Love opening the folds  She seeks for the one Dancing around the flame Courage he has  Regardless of his shame  The  moths dance around her The wings singed black Her flame is a song Sanity they lack  She cries for each one They  know not what they have done Life snuffed out quickly Her flame is the sun  Tears of a muse Lost and depressed. Life has no meaning She prays for a  rest  Tears of a river Deeper and fast Her flame extinguished  She prays for the last  The moth that circled So far and so slow Her heart flickers Trust does he sow  Forever so strong Her weakness revealed A muse so human Her fate once sealed.  Forever alone Past to atone The muse cries freely Afraid to believe  Falling to her knees She begs and she pleads Patience and kindness Love to fill her needs  The moth dances to and fro Uncertain of where to go To her and away Now  he canâ€™t say  Patience and balance she must learn So long has she lived Her flame forever to burn Trust the key to a heart that will never flee.  DLT Â© July 10 2004  '),(9709,3546623,'09,July,2004','2004-07-09 00:00:00','   I smile to  myself as slowly you become Listening to the muse You glow and you shun The past and the future for us there is none  Fly high dear friend Live with glory Understand the gifts  Content moments writing your story  Tears travel slowly mixing with the dew Forever a solitary walk Comfort alone in memories for few who knew.. Fate of a muse  The dance of the sacred A muse who enchants Young men to question Whispers of their souls chants  Her fate she embraces like a warm  summer breeze Loneliness banished with a waltz of a bee Poets, carpenters, lovers and dreamers They come to her lost and forlorn  Her laughter is rapture Challenging to open rusted hearts Dusty minds that rarely were used Marriage of the two left many confused..  Be on your way, skipping across lava Gently she whispers forever tomorrow Learn and be happy for one day you will sayâ€¦ Mountains or Oceans hold neither at bay.  A word so kind Confession of truth Departing fears  Edifying muses.  A hint A guess Thinking deep Time unfolds  Expectations dashed Life a wonder Only a moment,  Now to be lived.  I did not know Pleasure so bold Appreciation shows Honesty richly told.  Joys of life Thankful today Rich and rife Spirits  play!  Swirling snow waltzing around us Creeping dark flees towards the light Endless hope infinitely blooms Dawn forever dances on the horizon.  Thankful the muse Grateful to know No longer unique Alone no more to go.  The  muse is tired She hears the call Captured by laughter Her men of lifeâ€™s ball!  DLT July 2004  '),(9710,3546623,'08,July,2004','2004-07-08 00:00:00','   I thought it was done, I believed it was over, That door was closed, Hope for the future!  I whispered, Noâ€¦ As the story was told. Lies, Deceit, Hatred, Dishonor, Betrayalâ€¦.  What had I done? I am no villainess, Reception of the vile, Truth a horror to behold, How can this be? Your hatred is so bold!  You call me the darkness, You say that I lie, You label me insane, As Anavian slides through your brain.  No more will I take, The wounds still to fresh, Enlighten all other, I go now to rest.  Lady Lorraine du Lac  May 26 2004  '),(9711,3546623,'08,July,2004','2004-07-08 00:00:00','   Truth  It knocks on my gate, Recognizing the fate, I turn to run, Truth soon to come.  A heart so hollow, Pain to follow, Your character I label, I see now a fable.  Creaking and straining, Truth stands reigning, Deaf are the ears, Eyes filled with tears.  Fault is my own, The seeds I had sown, Closing my eyes, Your lies seemed so wise.  Truth is the tower, It holds all the power, Shallow you be, Never to see.  The path I chose, Never to lose, Light from within, Love never a sin.  DLT Â© july 9 2004  '),(9712,3546623,'08,July,2004','2004-07-08 00:00:00','   Salty sweet tears Pool and form Silent puddles Lakes of desire Fathomless Lush Surroundings hushed Rich greens  Deep Blues Moist  Rare Trust in you Do I dare?  Salty sweet tears Friends  Love New and fair  Easy  Determined Time thunders in Slow the Way How we begin.  Salty sweet tears Mending hearts Baffling trust To give  Is not a sin Still I battle Here I pray No longer to win.  Salty sweet tears I cry today My heart Fills with joy Truth is shown To think  Is not the way.  Salty sweet tears Lead me to you Slowly I open Listening Whispering An answer The secret Kept By two.  Salty sweet tears Immersed Trusting A final breathe I take Slowly I sink Drowned in the lake.  Salty Sweet tearsâ€¦. My heart will you break?  DLT Â© July 9 2004  '),(9713,3546623,'08,July,2004','2004-07-08 00:00:00','   The stream of time Called by the ebb and flow Sirens of the sea sing People grace our shores Polished pebbles of their essence Left behind Immersed in sea foam Rolling through the sands Beckoning remembrance Soul marrow Love Sorrow Ecstasy Contentment Circling moons of our destiny Full Half Crescent New Opening doors Closing doors Forever a timeline not ours Sleep Next to me Near but far Which door is ours? Changes You Me Both Neither Ringing laughter Glides through my heart Knocking on the wall Erect between us Slowly your soul Leaks forth your nectar A smile  Dancing laughter Soulful eyes That can not look in mine Lento sun rising Strong and warm Potent and pure Opening doors Closing doors No longer one We crest  on our shores Leaving pebbles  and foam  Scattered Along our beaches A memento or a pearl in the making?  DLT Â© July 7 2004  '),(9714,3546623,'12,August,2004','2004-08-12 00:00:00','   The moon sings thru me, Pulsing rays of my soul, My dreams of you dear, To the sea do they flow.        Friends do surround me, Our language few know, In the sea we do dance, Come live in the glow!        Come to me dear one, In waves we do fly, Living in peace, Pure moments that cry.        One in the moon, Songs of the sea, My partner your place, Soon we are free.        Floating in grace, In our hearts we will be. Dream I do now, At the edge of the sea.        My song I do sing, To your ears it does flee, Listen and hear, The message I bring.        Come to us now, Love on the wing, Forever, my vow. A queen and her king.        Soon you will hear, The whispers of few, Songs of the dolphins, Deep in the blue.        Come to us now, Listen and be, My partner, my love, Living at the edge of the sea.        Dlt Â© August 12 2004  '),(9715,3546623,'09,August,2004','2004-08-09 00:00:00','   urlLink    blue solace&nbsp; urlLink  '),(9716,3546623,'01,August,2004','2004-08-01 00:00:00','   '),(9717,3546623,'01,August,2004','2004-08-01 00:00:00','   The weather forecast called for thunderstorms today. Hot and sticky slowly the gray skies are sliding into the past. The sun is reaching back from the future demanding to live in the now. I watch the shadows of the leaves dance on my deck and think to myself what is the   MORE  ? I am lost. No purpose in life, empty shallow at 42 I sit here and wonder what is life about? The shadows dance on the deckâ€¦gray, one dimensional, and dim. These shadows are my life. The winds of change tickles the edges, I ask when will I no longer be a shadow, but a full living leaf that is rooted to the tree?      When did my life become so empty? Many say this is depression but it is not, it is something so much more. A form of spiritual death a precious gift dying that I know I must release or forever be lost. So I sit on the deck watch the dancing grays and pray for the green. I have always been one to say I live without regrets. That I hold all that has been an experience of my past as a gift be it good or bad. Did Parsifal feel this emptiness on his search for the holy grail? So alone, desperate and dying? Was his world gray and meaningless before he found the secret? Did he witness the gray to green miracle? Can I find the strength to witness my own color guard change?       At this moment I regret that I have not lived a more normal life. This truth stares me in the face and shouts from the tips of the leavesâ€¦edgings of green coming into focus. Yes I have some fabulous tales to spin around a campfire and laughter comes easily as the drama queen pulls out her crown in the telling. The life of a seeker is lonely. We have many experiences to share to teach and to learn from. I canâ€™t help thinking though at the end of the day that these experiences are the gray shadows on the deck compared to the gentle smile of a child that calls you mother. The passion of climbing a mountain face pales in comparison to listening to the steady breathing of your partner as you fall asleep at the end of the day. The comfort of writing is empty when mirrored to the sharing of your soul face to face with one you love, all the while knowing the risk of loss you take but, taking the chance anyway. My passion for this single isolated life is dying, I miss the touch of a lover, the quiet silences and the shoulder to lean on. I miss them? I laugh, how can I miss something I never have had? Again it is the knowing that lives deep in my essence that haunts me, telling me let go and believe. Has time tainted me? For at this moment I have little faith in what I have yet to experience, rather the knowledge of my past haunts me and calls from the dim grays that the MORE is a myth to forget and not strive for. Parsifal did you feel this? Where is the green?      I fight hard now to grasp onto anything that will make me feel. A woman so passionate about life I am now but a whisp of a ghost floating through life. Maybe I have it all wrongâ€¦. Always have! This is a devastating reality to come face to face with. The quest for inner peace, pushing forward courageously, always to do good for others while seeing the best of this world in all, is this a fable? I sit in reflection as the sun hides behind a huge black edged cloud. I ask, is there no silver lining for the seeker? I have always maintained that the religions that stated that Jesus never had a life companion were wrong. The belief that in having only himself and God was enough was a myth. I believed wholeheartedly that Jesus and Mary Magdala were two parts of one. This is union of soul, mind and body was the way to the divine. Trust, love, honor and respect shared between two people an experience that we all must strive for to reach ecstatic union with God. Not just sex people! Rather so much MORE. Now I doubt this, maybe the religions are correct. There are those of us whom are to be alone, as Christ is in the Bible to share the experiences of life and God, but never to share with another the total marriage of two souls that create passion for living. Am I giving up hope or living in reality? This thought saddens me for if I accept this as the role of my life what else will I settle for?      The sun hides still as the passion of these question fades. Where did this all come from one might ask. The truth is the answer. When one is faced with a truth about themselves we have options to take but, only one leads to a greater freedom. That option is surrender and acceptance of our responsibility. I must now speak the truth one so painful that I, like the sun, have hidden for years behind the clouds of a false reality. No longer may I hide this secret that I myself must face. The reasoning for the death of my passionate life. I must go there again.      Many have heard the story of my life, the trials and tribulations that I have survived. I have been compared to by others as having the power and essence of the Mary Magdala. I have spent months now reading of her, trying to see the resemblance. For you see, I do not see this at all. Mary Magdala was a woman of divine courage and faith. I am anything but. If I reflect on my beliefs that Mary and Jesus had a sacred marriage and that she gave birth, I see no comparison to me. This woman faced the decision, accepted the help of others and brought life into this world. I did no such thing.      Ten years ago, I made a decision in my life that lacked courage or the strength of one who holds the essence of Mary Magdala. A decision that will forever haunt me and that I believe I am being punished for. A wound that never heals. When asked what age was your favorite, I always quickly comment back 33! I go on to explain that is was the year that I changed my life. Returning to school, leaving my comfortable restaurant job, in great physical shape, facing truths and helping others. I never mention though that the age of 33 also held the worst day of my life. All these simple reasons that I state for being the best year of my life hinged on a decision that also made it the worst day of my life, a day of the coward and not accepting the gift that was given to me.      That day I sat alone on my couch crying tears that I hope I never have to shed again in this life. Then again, if presented with the same decision, I would not even consider the choice I made that day. I was pregnant, alone and afraid. The shadows on the deck are no longer visible, blocked from the sun the gray reigns as I go back to that day. A day I hid away, a choice I made that put me on a different path, one of my own free will, one that I regret.      I sat alone that day of the decision due to my own fear. This pregancy was not planned, we took precautions, I was careful and still this happens. I did what I thought was right in this relationship. I had previously had a year of celibacy and slowly dated this man. We did not sleep together until we were long into the relationship and then it was only once. I did not tell him of the child. I did not trust that he would not leave me where I was already, that of being alone. I thought to myself why take that risk? Why add fuel to the pain I was already experiencing? Why think that he would be any different than all the other men in my life? That he would actually stand by me, help and be a support. I did not even allow him the opportunity to try, I denied him the truth and myself the lesson of trust.      Alone I sat that day, on the couch speaking to a child that I did not have the heart to bring into this world. I loved this child, I wanted this child, but I lacked the courage, faith and respect for myself and God that I could do this. So instead on that day that the gray began to edge into my life, I cried and begged for forgiveness to a child that lived within me. I convinced myself that this was for the best, as I tried to convince this force I felt alive within me that same fact. I knew the child was fighting for its life, I felt it within me and the tears came harder and faster. I hated myself for lacking the faith in myself and God. I hated myself for believing that this was the right decision. Sadly, I still hate myself.     The childs voice left me that afternoon and with it my true passion in life. I chose my path. I saw the pregancy and this child as a cage, something I feared all my life. Reality was and is, that decision to deny myself the gift of that child was the true cage. One that there is no release from, one that I live in to this day.      Slowly, I come back to the present. The sun has slipped behind the hills and dusk will soon be upon me. The cage and the bars are still there, formed by the desire of the green of a leaf. Maybe tomorrow it will be sunny and I may heal some from opening this wound. Or maybe tomorrow the sky will cry a million tears for the path I chose. I only know now, the truth is told. Before this truth was faced, I was living a lie, a mask placed before all to see. I am raw now, naked and vulnerable. Who do you see? In front of me, stands Mary Magdala smiling.  '),(9718,1488330,'19,May,2004','2004-05-19 00:00:00','   Hello, I am just starting this for the NetNav class.  Feel free to read any of our postings and use it for tech stuff.  '),(9719,1488330,'31,March,2004','2004-03-31 00:00:00','    Hey Thomas - we missed you...we had an interesting discussion on the atomic bomb...the pushing of sugar is very interesting...comparable to the big companies making money off adolescents at Skyview - disguised as Vending Machines and Food Products....  Mr. Ray wants to be invited to our US History Blog - and I have the list of sixth period students.   '),(9720,1488330,'30,March,2004','2004-03-30 00:00:00','    Just some interesting things to note from Henretta about WWII Rationing:  \"People found it especially hard to cut back on sugar.....the manufacturers of products such as Coca-Cola and Wrigley\'s chewing gum received unlimited quantities fo sugar by convincing the government tha their products helped the morale of the men and women in the armed forces\". Woooooooo we are really really a nation addicted to sugar.  But what\'s more interesting: \"Congress imposed a nationwide speed limit of 35 miles per hour...as a result, highway death rates dropped dramatically.\" WOW!!!!! Speed limit NATIONWIDE at 35 mph.......Man in today\'s lightning pace stressful lives......Such a speed limit would certainly kill off lots of people!!!!!!!!!!   '),(9721,1488330,'30,March,2004','2004-03-30 00:00:00','    Unfortunately, Ms.Clark, I will be gone all day tomorrow (for MUN) and on Friday (for NWSE), and both days are apparently discussion days. But hopefully you or some other students can post interesting discussion topics here to keep me and others who will be absent informed.   I do believe that there are lessons to be learned in history so to avoid similar conflicts, patterns, and cycles of violences (otherwise there is no point in enthusiastically studying history). Things do occur in similar way just because of natural ways of human interactions, reactions, and emotions.   And that the dropping of the atomic bomb poses an interesting question and potential debate. I personally don\'t have a strong opinion of that right now simply because it\'s such a complex issue.   Oh yeah, and don\'t forget to invite all the talkative sixth period students to this discussion blog!   '),(9722,1488330,'29,March,2004','2004-03-29 00:00:00','    Hey - very interesting tidbits...Thomas thank you for starting the AP Blog - American foreign policy is very complex - I wonder what would have happened if American intervened in the Spanish American War - would World War II erupted earlier - would Hitler\'s Final Solution been avoided or defeated before the horrible consequences.  It is so easy to make or critique our past and qualify it as a mistake.  I have never really understood the quote...\"we must remember our past so we aren\'t condemned to repeat it .\"  We will never be able to repeat the past...it is so different...however we continue to behave as humans do? How can we repeat something that is so different - or do we repeat the same instances with superior technology.     \"Under God\" - did you know they added that phrase in the 50s during the cold war?  America\'s rally around Freedom...Goodness and Godliness...Cold War propaganda?   I enjoyed reading your blogs - What are your thoughts on America\'s decision to drop the bomb on Hiroshima and Nagasaki?    Prediction: On the fifty year anniversary of Brown vs. Board of Education (1954) - Portland will host a major racial riot....Topic \"TALK\"  See you tommorow   '),(9723,1488330,'29,March,2004','2004-03-29 00:00:00','    We need to do something to liven this place up a bit. We should advertise more or do something...   '),(9724,1488330,'28,March,2004','2004-03-28 00:00:00','    Isn\'t it a sad state that the Fascist National forces won the Spanish Civil War? The Loyalists, being the Republican-elected govt., get supported by only Mexico and the USSR, while the US, Britain, and France keep their sympathies at home. I applaud those who went over to aid the Loyalists. I have yet to read much lit on this period (Hemingway\'s \"For Whom the Bell Tolls\" is now on my shelf, unread, but certainly on deck). George Orwell\'s \"Homage to Catalonia\" would be good too. I read a short, interesting excerpt from it where Orwell described himself meeting an Italian peasant. A communist soldier, a simple man, and one fighting against the government of his country to fight for his beliefs. It was interesting to consider that even some Germans and Italians, whose governments were supporting Franco\'s Nationalist rebellion, were joining the Loyalists to fight the spread of Fascism. It\'s like the Domino Effect theory in Southeast Asia during the Vietnam War, but also like the end of that conflict, the \"villanous\" form of govt. fought against (Communist) didn\'t spread beyond far beyond Vietnam. Anywho, and the lack of help from democracies causes the Republican collapse, and Franco\'s brutal reign goes from 1939-1975; 36 years! Gee, and the policies of neutrality and appeasement certainly didn\'t help in Spain, or in Czechoslovakia for that matter. Or in Poland, etc. \"It just goes on and on and on.\"    Does anyone need an extra id partner. I\'m trying to get an exchange going for this new test. Anywho, see you all later, hopefully. I\'d like to see what you think on this above blurb. Didn\'t the quote before signing into this thing say \"for people who have lots to say\"? Goodbye all. Adios and God bless you.                                                             Yeehawman,                                                               (Tristan)   '),(9725,1488330,'26,March,2004','2004-03-26 00:00:00','    Nicely written, Thomas. I don\'t understand why some people find that they must make a big stink about a two word phrase. If they don\'t want to say it, then they shouldn\'t say it. That\'s why the first amendment is there. It protects freedom of speech, and the freedom of not-speech. So if you don\'t want to say something, you\'re welcome to not say it. Don\'t start suing people because you don\'t want your kid to say it. I agree that little kids should know about the pledge, but forcing them to say something they don\'t know the meaning to is not right. If you want people to say the pledge, at least have the decency to teach them what it means first.   '),(9726,1488330,'26,March,2004','2004-03-26 00:00:00','    To deviate only a little away from wars and genocide,  urlLink Mr.Wargo\'s class blog has got up there the really controversial debate of the constitutionality of the Pledge of Allegiance, and especially the \"Under God\" phrase.  And don\'t blame me for spilling over that debate to this blog, but it is inevitable. Of course everyone knows that the 1st Amendments guarantees \"Congress shall make no law respecting an establishment of religion, or prohibiting the free exercise thereof; or abridging the freedom of speech.......\". And then one can debate FOREVER what does that mean, either pro pledge of allegiance, the under God phrase, or opposes it. Either way, both sides can site the constitution to argue for their sides. LOL that\'s the beauty of our first amendment, even tho it may be concise, it\'s still very interpretative. And the society is sooooooooo complex that a few words in the constitution can be argued forever into future generations.  As a former buddhist, and currently agnostic, to me personally, the \"Under God\" phrase doesn\'t offend me. But the act of conforming is. I have nothing against people saying or chanting Under God at anytime, anywhere, or under any occassion; the constitution guarantees their rights to do so. The thing that bothers me is the act of telling others what to do or not to do. How does that relate to the pledge of allegiance is the fact that little kid six years old have to say the pledge of allegiance. And did you know all the MEANING of the words in the pledge of allegiance when you were six? To me it\'s almost ridiculous to impose others to say what they don\'t even know what they are saying, and what it means......let alone believes in it. So the little kids just memorize the phrase and spew it out like robots. And that\'s not what America is all about.....  What makes America great and the beauty of this country is the fact that it\'s a country of immigrants, and EVERBODY is different, and the constitution really protects the DIFFERENCE of all individuals. In fact, individualism is not only protected, but glorified and encouraged all throughout American literature from the declaration of independence and onwards.   And from individualism comes with it, differences of ideas, clash of cultures, criticisms, and sometimes stuff that people don\'t want to hear.  What bothers me is when people stop wanting to hear complaints and other points of views. Any criticism, or other radical ideas becomes BAD and immediately shuts off from debates........And everybody has to praise everybody, praise the government, praise all the institutions, all the ideas and so on. And even if you don\'t, you are somehow peer pressured, or forced to comply with the \"majority\" point of view. Some of America\'s greatest writers were critics of the time, and often deemed radicals because of their foiling of the truth......truth that the majority didn\'t want to admit. Like Mark Twain\'s Huck Finn realism style writing that foils the true life of the South from a common southern white orphan boy\'s point of view. (Of course it\'s not just happy little child pacifer literature that\'s all happy and good. Because the society isn\'t always that good).   Seriously think about, what a society would be like if everyone is the same, think the same way, and all are happy with each other with no reason. Everybody would just lose their personality and the society itself will become dumber.   And education is really when the clash between conformity and individualism occurs. In order to preserve the individualisms as America\'s embodiment, I believe that in schools, since it\'s a public place, the state or government should have no say in what the kids should say. The kids are just kids, they are learning and developing.........And wheh pledge of allegiance is inserted to thehir curriculum since their first day in school.....And have memorize something they don\'t even know what it is......it totally defeats the American character. Education is to develop people to become individuals, to become smart citizens with an critical eye, and not to turn everybody into robotic conformists.   And as always, I welcome all dissagreements, criticisms, opinions and would like to debate this issue with others.  Differences of opinions is good!  Thomas Chen   '),(9727,1488330,'26,March,2004','2004-03-26 00:00:00','    Also, I have a CD compilation of Spike Jones, and he makes fun of Hitler and Germany in a few songs. But I lost it... :(   '),(9728,1488330,'25,March,2004','2004-03-25 00:00:00','    It\'s been too long since I posted in a Blog. I even had to create a new account, because my last one was unused for so long. Anyway, Today was interesting. I knew that the old cartoons were about wartime stuff. So many of them showed people buying warbonds, or ditching the drafting, or something to do with war, but cartoon Hitler was still funny...   '),(9729,1488330,'25,March,2004','2004-03-25 00:00:00','    This is my first time writing in a Blog - I am so uncomfortable.   '),(9730,1488330,'25,March,2004','2004-03-25 00:00:00','    Heh Interesting AP Exam Data for AP U.S. History: Last year\'s AP Exam statistics: Average Score 2.73, (compared to 2.78 for World History (which Emerson\'s class nailed the exam), and 2.69 for Comparative Politics/Government).  Response to matt: Hmmmmmm wait, maybe not many people are interested, but then there are those few who are really interested, and for them, it\'s a chance for them to burst their feelings out........   '),(9731,1488330,'25,March,2004','2004-03-25 00:00:00','    I dunno Thomas, not many people seem to be all that interested...   '),(9732,1488330,'25,March,2004','2004-03-25 00:00:00','    Test-------The first post-----Any interesting topics to JUMPSTART the discussion forum? WWII/Holocaust/American Involvement/Hitler/Nationalism/FDR/Concentration Camps.........   '),(9733,1488330,'28,April,2004','2004-04-28 00:00:00','    Very insightful comments...I wish more students would take the time to think about what they learned - you did a nice job synthesizing the information and taking it to formulate your viewpoints on the time period. Nice procees - and you allowed me to see it which makes it more fun.  I hope others are wrestling with the information to help them support their understanding of the world.  That is the best way to go about learning history - there is no way you can remember everything but the more information you acquire will help you formulate your own viewpoints.  It is called epistimology...(Look it up)  You did it perfectly - I am not sure I spelled it correctly.     Kailie - I am glad you loved Sam\'s lecture.  I hope you see his generetional view points - Change over time...I am not sure Sam\'s view points change - (women, Communism, Freedom, America)  but they are fun to examine as he continues to teach history.    Sunday   Segment 1 - 1800-1846 Republican Revolution to the Election of Polk Segment 2 - 1846 to the Civil War Segment 3 - Reconstruction/ The Guilded Age  If you guys can continue to reveiw your multiple choice while going through the study guide it should help.   We also have Tues, Wed, Thursday to sharpen our saw.    Keep working and thinking - make connections.   Clark   '),(9734,1488330,'27,April,2004','2004-04-27 00:00:00','    Well, I\'d like to do anything but that essay.  I think we\'ve had enough Centuries to do well on the test, but I\'d like to finish the series later.  Cause like Kolk said, Peter Jennings is one sexy hunk of man.  I\'d like to tear into him like Sam and his Sheila (that\'s Aussie for \"girl\").  Just kidding about all that, but I really enjoyed Sam\'s lecture today.  He\'s very intelligent, understands what went on and why, and can relay it to a bunch of high-school kids very effectively which is tough to do.  I think the key is boot camp and that drill instructor that\'s itching to get out... keeps us on our toes.  The lecture really effected my views about the time period.  It drives me crazy that the era that produced The Beatles, The Grateful Dead, The Rolling Stones, Jimmy Hendrix, Bob Dylan, The Who and Miss Clark would also deem it noble to spit on a soldier who risked his life fighting for freedom.  That may sound distorted/cliche (even to the vets), but the troops were there because the government told them to be...the government that feared the domino effect, and that if Asia went down to communism, trade ceases, America chokes on its own success in a ridiculous recession no military might could pull us out of, then the Revolutionary War, the Civil War, WWII, and every other conflict we\'ve participated in under the banner of freedom and democracy would have all been for nought.  It is so prevalent to Iraq... Bush riding terrorism to attack a country that could foster terror... but it makes me think \"what about Saudi Arabia?\"  like, 4/6 or something of the 9/11 bombers were  Saudi.  And what about Palestinian groups like Hamas and Al Aqsa Martyr Brigade?  Are they not also equally important to the war on terror, despite the fact they are attacking Israeli Jews rather than American ones?  Sam\'s lecture also helped shape my political views.  Kerry fought in Vietnam, won a silver star for valor as he captained a boat in the hottest of zones in the heart of Vietnam, Bush skipped out on the National Guard.  Although military experience does not ensure quality presidency (Grant), Eisenhower seems to have done a job well done.  I also get the impression that being surrounded by death, Kerry might be more hesitant to send in troops, he might carry a higher appreciation for the life of a soldier.  I was going to ask Sam about it, but we ran out of time.    Give me your thoughts on Nam, communism, Sam, or political oppinionism.  Good luck on the test, and good luck to SJClark in Seattle!  Love, Rob   '),(9735,1488330,'27,April,2004','2004-04-27 00:00:00','    Well, Clark, I just have to say i LOVED Sam\'s lecture today, and im really not being sarcastic either! I think I\'m really starting to feel the pressure of the test coming up, but i think as long as we continue to have review after school, I\'ll survive. I will definitely be at school on Sunday from 1-4, but what are we working on? I think the era that i need to work on the most is during the 1800s...pretty much the entire century has become a blur to me. But im feeling confindent about the 1600s and definitely the 1900s, and im starting to understand the 1700s better as we\'ve been talking about it during the review sessions. See you all tomorrow! Oh yeah, and Rob, dont ditch me on Friday morning...i coming to your house at the crack of dawn whether or not youre awake!  Kailie   '),(9736,1488330,'27,April,2004','2004-04-27 00:00:00','    helloo again! yeaa i think we should watch centuries peter jennings is a stud i just cant get enough! (biting my hand like miss clark right now) JUST JOKING but yea i learn from those videos and i like to think of it as relaxed review i think thursday would be a wonderful day for peter! what is everyone elses input?? OOO and in class could we do the paper on the table discussions i get more from that then lecture and notes BUT thats jus me hows everyone else like it?  im bouncin like a ball off da wall  '),(9737,1488330,'26,April,2004','2004-04-26 00:00:00','    Rob - Thanks for the compliment...did I say something funny.  Tomorrow we will move forward through the revolution and constitution and hopefully get the Republian Revolution and the era of good feeling.  I appreciate the feedback.  On friday I have a job interview in Seattle...Not that I have enough to do this week...but I am very excited about the prospect.    Friday - I could give you a multiple choice test again - or an essay prompt - or review - or centuries...Please use your review guide and review the Multiple Choice questions that you missed.   SUNDAY - AP study session 1 - 4 THE 1800s...Can people make it at this time?  Monday you guys have the test and the rest of the week we will review...  Essay prompt:   John Winthrop\'s goal was to create a \"City Upon a Hill\"?  Like John Winthrop, Jefferson believed Americans had a special destiny and responsibility: \"The eyes of the virtuous all over the earth are turned with anxiety upon us as the only depositories of the sacred fire of liberty, and...our falling into anarchy would decide forever the destinies of self government.\"  This was the Republican vision of the City Upon the Hill. If Americans were to be a virtuous people, would all of the people in the new nation be allowed to become American?  Throughout history how and why have Americans struggled to live up to these principals?   I need a Muchas Gracias Burritto.  Ms. Clark   '),(9738,1488330,'26,April,2004','2004-04-26 00:00:00','    Clark- have you thought about a career in standup?  I know you have your administrative internship, but drop it all and head for Hollywood!  Maybe then you could afford to be a principal in Napa.  Thanks a lot for all the reviews- they beat the hell out of Henretta, which is nice because me and him aren\'t as acquainted as we could be.  Do you think we could start reviewing Jackson through WW1 era?  That seems to be the blankest for me, except for the slavery aspects of it.  Maybe one more essay practice somewhere would be gravy if you could fit it in... maybe a take home one that we could review?  To all my peers- keep on trucking.  The test is coming up fast, but we can study even faster.  GOOD LUCK!  Love, Rob   '),(9739,1488330,'26,April,2004','2004-04-26 00:00:00','    Yes...I thank Jessica for those life magazines...one is the year of 72 in picutres..  The other has the cover of the Kennedy family at JFK\'s funeral and of course my favorite...1969 Neil Armstrong walks on the moon...Guess what the Mercury seven were furious they weren\'t chosen to be the one who walked on the moon - they picked Neil...He must have had the \"RIGHT STUFF\" -  By the way I have been so thirsty for a glass of tang.  I will make sure to bring those in...  Those in 6th period I am going to start at 7:00 am on Wednesday because Sam is going to lecture about Vietnam...Boys bring doughnuts and girls bring juice and cups and I will bring coffee...  Also - the mornings of Thursday and Friday I am with Ms. Winters - we are going to go over 6 to 7 American art slides and then do an art project - There was an Andy Warhol question two years ago...she said it was fine if we wanted to review on the balcony or the side room.  Cheers - I am going to the gym.  Clark   '),(9740,1488330,'25,April,2004','2004-04-25 00:00:00','    Hey APUSH - we had some great success on our AP multiple choice test - however, I know many of you are disappointed.  The good news we have two weeks to raise those scores. Last year we only had a couple students over 40 at one week prior to the test. It is not time to panic - instead go over twenty multiple choice questions a day - even some that you got right you still need to work on.  Antebellum - Pre-Civil War - this term is often used to refer to the golden age of the south.  Andrew Melon - Treasurer under the Harding administration - you know the roaring twenties were about big business and handsoff government  Evan Cate can give you all the information  Essay Prompts - Most answered the essay - How did the Civil Rights Movements of the 50s and 60s address the failures of reconstruction?  I think you could look at supreme court decisions - Plessy V Ferguson - Brown V Board of Education  Presidential Decisions - Eisenhower sent in troops, Johnson and congress supported to important pieces of legislation while Hays made deals to take the troops out.  Excellent leaders and movements that addressed the issues head on...Desegregating schools, facilities, voter registration drives...all of the issues that the radical republicans wanted for the freedmen but failed to achieve...I think it is also important to think about the television and media.  The movement relied on the sympathy of nonsoutherners...the television allowed for the brutality to be seen and not just told.  All of these are applicable events - Brown V Board, Montgomery Bus Boycott, Little Rock Nine, Jackie Robinson, Desegregation of Military Forces, SNCC, Sit Ins, Freedom Marches, Freedom Rides, CORE, March on Washington, Stokely Carmichael,  Civil Rights Act of 1964 and Voters Rights Act of 1965...Remember to use events to support your point and ideas - do not describe the events.  Your readers are intelligent people and describing the Montgomery Bus Boycott is insulting.  How did anti - communism, the silent majority or the women\'s movement shape American public life....  Most NAILED the anti-communism in the fourties and fifties...you should have established the fear...Soviet Expansion, Eastern Europe, Berlin Blockade, Berlin Airlift, Truman Doctrine, Eisenhower, Korean War, Red China ....  But how did it affect American Life - fear of communism and subversion - McCarhyism, HUAC, Black Lists, Bomb Shelters, NSA, NASA, Space Race, Interstate Highway system, Math and Science stressed in schools, Arms Race, Competitive with Russians...  Silent Majority was a group of \"hard-working\" middle class Americans who were sick of the protests and the violence of the sixties and seventies...In support of Nixon\'s back to Normalcy - Good Americans, Anti - Drugs, Back to American Values....Voted in Nixon and led a stance against the chaos and hardships of the sixties and early seventies.  In fact one poll believed only 11% of Americans sided with the kids at Kent State and the rest supported the Natinal Guard.  - this was the hardest one...  Feminist Movement - Betty Friedan - there is more to life for a woman...the birth control pill, protests at Miss American Contests...again a group met the \"establishment\" head on to make changes in the work force, at colleges, and made it acceptable for women to have a career...prior to the movment - job sector divided...changed the family units two income households and may have contributed to the divorce rates...ERA, Title 9 and Roe versus Wade...are all good things to include.  I promise I won\'t eat the entire pint of ice cream if you promise not to freak out and do intentional focused studying each night.  Have a good evening.  Steph   '),(9741,1488330,'25,April,2004','2004-04-25 00:00:00','    YA its me again IF ANYONE IS INTERSTED I GAVE MISS CLARK SOME LIFE MAGAZINES FROM THE 60\'S AND 70\'S AND THEY WERE HUGE BACK THEN FILLED WIT LOTSA PICTURES AND ARTICLES ABOUT THE SPACE RACE, NEIL ARMSTRONG, ETC ETC THEY ARE KINDA KOOL! (JESSICA k)   '),(9742,1488330,'25,April,2004','2004-04-25 00:00:00','    HELOOOO Blogger BUDDIES! Took me awhile cuz yea my computer is quality and works alot..... THE EC movie nite! (LOOK MISS CLARK IM BLOGGING 10pts EC right!) I thought the movie was alright, drug on and on but i like airplanes and the space stuff and growin\' up in kettering ohio where the WRIGHT brothers WERE born AND LIVED thats right the orville and wilbur wright used to live like 2 miles from me:-) yeaaa but ne ways The movie really showed how women were just kinda sittin on the back burner the whole movie was about men, and then the women were just at home waitin to see if theyd come home alive or not. It really supported everything i had learned about the 50\'s and 60\'s from miss clarks eye of the tiger lectures! BUT YES now i must finish my we didnt start the fire EC! YESSS have a WONDERFUL SUNNY DAY! FOR u overacheiver ap\'ers PLEASE AT least study OUTSIDE!!! ENJOY! JESS K( as in kolkmeyer)   '),(9743,1488330,'25,April,2004','2004-04-25 00:00:00','    ok... someone has to get me up to speed on what\'s been going on in class... this WASL schedule has made it so annoying... what\'ve we been doing or studying? i have no idea!   '),(9744,1488330,'24,April,2004','2004-04-24 00:00:00','    Good Morning...Thomas I think the segregation process is a fascinating story. It appears that all the efforts in the sixties may have been futile because our society is segregated based on economics and race.  Racial and ethnic segregation has been a phenomenon from the beggining of America\'s story.  The Irish, Chinese, Italians, Koreans, the Russians immigrate and create communities.  It seems like the etnic cultures eventually  \"melt\" into mainstream society after a generation or two  but people with racial differences haven\'t.  Then their is the Native American story where they were placed in communities together for convenience and their own preservation.   During the seventies there was a movement to desegregate called \"bussing\".  You see an example in the Denzel Washington movie Remember the Titans.  It failed miserably.  Now with \"No Child Left Behind\" legislation allows for students to leave their failing school to attend another.  You tell me - where would all the students from Skyview go if it failed?  It is funny - when I was in school we were deemed \"The Nation at Risk\"   - Regan administration.  Now it is NO CHILD LEFT BEHIND...  I am getting to ready to read your prompts...I will post some information on how you did as a class.  A good thing to examine to finish your studies of modern time is to take Kennedy - Camelot, Johnson - The Great Society, Nixon - Back to Normalcy,  Carter - ? and Regan administration and make sure you look at what they stood for politically, on domestic and foriegn policies...Watergate, Opec, The Iran Contra Affair, The Camp David Accords, Reganomics, Star Wars...  I am working on a review...   '),(9745,1488330,'24,April,2004','2004-04-24 00:00:00','    I feel this is a great article that everybody should read:  urlLink   Segregation no longer black and white New issues fifty years after Brown v. Board decision  It supports some educator\'s arguments that today\'s schools are actually as segregated or even more segregated than ever. Segregated for a variety of reasons: huge discrepancy of wealth, economic gap, culture differences, community separations, and continued racial problems and misunderstandings. This culturally, economically, and racially segregated school systems in turn shows up in the huge differences in the \"standardized\" testing scores. This problem really boggles those \"No Child Left Behind\" testing conformist\'s mind and is a threat to their \"everybody conform to the same standard\" ideology.  I would like to hear  everyone\'s arguments and opinions on this issue.   '),(9746,1488330,'23,April,2004','2004-04-23 00:00:00','    Yeah I can finally blog, I am just testing the water of blogging and I have so bye   '),(9747,1488330,'22,April,2004','2004-04-22 00:00:00','    Ok, Ms.Clark, I had to delete all of those erroneous testing posts. And by the way, for any of you who may have trouble publishing your posts, after you click \"publish blog\", please do not hit the back key, for it may erase your post entirely.   In reading Henretta, I found it interesting that today\'s leisurely clothingstyles largely stemmed from the counterculture radical movements: when young men started to wear \"coats and ties and women skirts and sweaters\", and \"youth defiantly dressed in unisex ragged blue jeans\". This provides a great contrast to the clothingstyles of \"old days\" when even in the 50s everybody was still wearing very uptight clothes and the kids in school had to wear tight uniforms.   And  urlLink here is a link that may bring about some shock about the current cost of war in Iraq. I was surprised and never put into perspective how much money that cost of the war came to be .   '),(9748,1488330,'21,April,2004','2004-04-21 00:00:00','    Took a while, but I liked the extra credit movie the other night a lot! I thought that it showed the attitude of the time period pretty well. America was trying so hard to beat Russia in the \"space race,\" and I didn\'t realize it was that important to a lot of people. I also liked how it showed how much people liked the Kennedy\'s. I think it was Guss\'s wife who got totally emotional because she didn\'t get to meet Jackie.  I have to agree with Ms. Clark that astronauts are totally hott after seeing this movie.  I enjoyed my first extra credit movie night. Thanks! Xoxo,  Jess Knab   '),(9749,1488330,'21,April,2004','2004-04-21 00:00:00','    Hi this is Al and I have no idea if i am doing this right at all. so i if i am, please tell me!! Al   '),(9750,1488330,'20,April,2004','2004-04-20 00:00:00','    The movie tonight was alright.. it took too long to get going, but maybe that was just me. In response to Ms. Clark\'s questions...... Perhaps the reason that Cooper couldn\'t say Yeager, is that at that stage in America, we wanted to see leaders and such on TV, not someone who wants to take his time explaining about some other guy\'s abilities.  The news people were wanting him to hurry with his explaination, and perhaps it was faster to just say himself. Also, another answer could be, that he was scared that if he didn\'t show confidence in his own abilities as a pilot, NASA would have chosen the guy he suggested. If the nation sees an astronaut reccommend that he is not the best pilot, they would be upset that he is flying a space craft their tax dollars went into creating; they would want someone who is ready to go into space.   I think that the main point of this film was to show that during the Cold War, personal victories boosted the nation\'s morale. For a very large portion of the nation, breaking the sound barrier meant nothing or little to them. Yet the media played this off as some great event. While sending 1 person into space was a landmark event for the US, it became more and more dramatized each time... everytime a new astronaut went up, they \"pushed the envelope farther than before.\" But each time they orbited the earth a few more times than before, it was made out to be an enormous event. This surprised me, because if I was living back then, I don\'t think I\'d care if someone went two more times around the Earth than the person before him. I think the directors were trying to show that maybe some events during the Cold War were made out to be bigger than they actually were. I felt that they were successful in this (unless my analysis is wrong). They showed how dramatized these events were....  Wooo Getting kinda late... i should be shuffling off to bed like an hour ago...  ~Mat \"Ralph\" Keksi signing off   '),(9751,1488330,'19,April,2004','2004-04-19 00:00:00','    I enjoyed the movie tonight, although the editors should be shot in the face.  The flick portrayed the Cold War attitude- the fun side of it.  Frank and I were talking about how fun it would be to have that kind of competition with another nation... where every American triumph feels like a personal victory.  Aside from the brief period of time after 9/11, I don\'t think I\'ve ever felt the sort of American comradery and patriotism that appeared in The Right stuff.  The homage paid to the test pilot was a little time consuming, but touching to see Yager continue to risk his life because it was his job and what he liked to do even if it were overshadowed by the space race.  Lastly, seeing Jeff Goldblum, Dennis Quaid, Ed Harris, the crazy guy from Vertical Limit, and the hick in Tremors  as such young guys was also fun.  I hope that the extra credit incentive for posting tonight begins a trend on this blog.  Also, respond to the posts Tom and I wrote earlier.  Love Rob  P.S.  Thanks for the excelent review this afternoon, Clark   '),(9752,1488330,'19,April,2004','2004-04-19 00:00:00','    In response to the question about gender roles, I believe that the movie\'s portrayal of the responsibilities of women during that era was quite accurate.  Throughout the movie the wives of the to be astronauts were depicted as obedient to their husbands even if this was not in compliance with their personal feelings.  The prime example of such characterization was demonstrated in the scene where the astronauts were being interviewed by the press and the question arose of how the astronaut\'s wives felt about their husbands going into space.  In response, all the astronauts replied that their spouses supported them 100% in their pursuit towards space, when this clearly was not the case for all of the wives.  Finally, one aspect of the film that stemmed away from the stereo typical gender roles was the relationship between Alan Sheppard and his wife.  When urged by his superiors to tell his wife to speak with the press/vice-president, Sheppard disregards this pressure, and supports the wishes of his wife to be sheltered from the invasion of the press/media.  Any other thoughts?  Benjamin   '),(9753,1488330,'19,April,2004','2004-04-19 00:00:00','    After you comment on Robs and Toms blogs please consider writing or discussing the following ideas - or come up with your own.  Write an informative and entertaining review of the movie.  Read a few current movie reviews before you write your review on The Right Stuff.   At the end of the movie when Gordo Cooper is asked, â€œWho was the best pilot you ever saw?â€  Why couldnâ€™t he talk about Yeager?  He tried and it was on the tip of his tongue, but it never came out.  Why?     What was the theme of this film? What were the film makers trying to tell us?  Were they successful?     How does the film depict gender issues?  How does it support what you know about the fifties and the early sixties?  Were you suprised by anything in the movie?   '),(9754,1488330,'19,April,2004','2004-04-19 00:00:00','    Hello? (HELLO, Hello, hello) Echo! (ECHO, Echo, echo)  I don\'t think there\'s anybody in here, Tom.  But just in case... if you\'re in 6th period, or the other AP period, I hope you remember the Holocaust/Hiroshima discussions we had.  I came across a story in the New York Times about a village in Italy attempting to prosecute some former Nazi SS members.  If this piques your interest, you can read the story here; http://www.nytimes.com/2004/04/18/international/europe/18TUSC.html?th .  The question it asks is, is the world ready to forgive the Nazis?  Because Sant\'Anna Italy isn\'t ready to forget it.  Post your thoughts about this, the draft, or your mother\'s snickerdoodle recipe, but say SOMETHING!  Love, Rob   '),(9755,1488330,'19,April,2004','2004-04-19 00:00:00','    Ok, with the approval of Ms.Clark, I would like to ask everyone your opinions on the impending possibility of a draft.   Here is a  urlLink great article that introduces a glimpse of the growing probability of a draft    I didn\'t know that our country has 1.44 MILLION TROOPS stationed in 141 countries. Why would we want to police the whole world?  And given the result of the past few days, the miserable situations our armed forces got into in Iraq, the situation seems somewhat scary. And I know it\'s inevitable, but Rumsfeld\'s decision to keep the soldiers there for 90-120 days more is an act that would further degrade the morale of the army, after already almost all of the frontline soldiers have been put on suicide watch.   And here is an article from CNN that accounts for the Bi-Partisan support in congress for the draft   CHARLES RANGEL   Democratic Congressman from New York and Korean War vet   Staying the course in Iraq means increasing our troop strength, and, not surprisingly, recruitment and re-enlistment levels are down. But proposed enlistment bonuses and other economic incentives will not make the military any more attractive to upper-middle-class young people. Increasingly we will be a nation in which the poor fight our wars while the affluent stay home.   To correct the disparity among those who serve, South Carolina Senator Fritz Hollings and I have proposed a new draft. All men and women ages 18 to 26 would be eligible for induction once they have completed high school. Those not needed in the military would perform civilian service. Enacting our plan would democratize our armed forces and return to the \"citizen soldier\" ideal that has served our nation so well.  As a veteran, I strongly believe that fighting for our country must be fairly shared by all racial and economic groups. Nobody wants to go to war, but the burden of service cannot fall only on volunteers who, no matter how patriotic, are attracted to the military for financial reasons. We cannot continue to pretend it is fair that one segment of society makes all the sacrifices.   CHARLES MOSKOS   Professor of sociology at Northwestern University and a former draftee   Our country is facing new kinds of threats and needs a new kind of draft. Even before the wars in Afghanistan and Iraq, our military was severely overstretched in fulfilling its missions. But more important, we have done nothing serious about homeland defense in the war against terrorism. We need guards for our nuclear power plants, dams and public facilities. We have done little to create the necessary border patrollers, customs agents and cargo-ship inspectors. Short-term draftees, under professional supervision, could perform these duties admirably. It takes less than four months to train a military police officerâ€”precisely the kind of role most needed in peacekeeping missions and guard duties. This would free up professional soldiers, and it would stop the unprecedented activation of reservists. Their multiple tours have led to demoralization and impending recruitment shortfalls.   We must institute a three-tiered draft system in America, with 15-to-24-month tours of duty for citizens ages 18 to 26. In the new-style draft, conscripts could serve in the military, in homeland security or in a civilian-service program like AmeriCorpsâ€”and there is no reason women could not be drafted for the latter categories   JAMES INHOFE   Republican Senator from Oklahoma   I think I\'m the only member of the Senate Armed Services Committee who would reinstate the draft. There are huge social benefits that come from it. I can assure you I would not be in the U.S. Senate today if I had not gone through the draft. When I look at the problems of some of our kids in America nowadays and then I go visit the troops, I see what a great benefit it is to give people the opportunity to serve their country.   I was drafted into the Army in January 1957 and served two years as an enlisted man. I gained a new outlook on life through the rigors of basic training. The military can have a more intense influence on soldiers when they are drafted and have no choice. I developed a sense of patriotism through the experience of serving my country. I\'m not on a crusade, but I think today\'s youth could use more of that type of discipline.   '),(9756,1488330,'13,April,2004','2004-04-13 00:00:00','    Lol yeah fear of communism is a very prevalent and integral part of American society, even to this day. Of course it\'s way too shallow and outright wrong to label people or groups of people in some strict, broad, undefined, and misleading category. Here is what I found to be a brief definition of \"Third World Countries\":  The term \"Third World\" was coined in 1952, in the wake of the Second World War, to refer to the ex-colonies that were not part of the two newly emerged geopolitical blocs (Soviets and US) of associated interests. The \"Third\" then meant the third way . So by that definition, I guess other than the US and USSR, Western european former imperialist nations and Japan won\'t be considered third world (because they weren\'t colonies), but almost every other nation could be considered third world. And interestingly enough, the US did not refer to the Soviet bloc and other communist nations as first world either, they sometimes are even categorized as second world.   '),(9757,1488330,'11,April,2004','2004-04-11 00:00:00','       '),(9758,1488330,'11,April,2004','2004-04-11 00:00:00','    So this is the first time that I have posted, but I have seen some of the comments that have been posted and I just hope that we will be able to focus on reviewing for the AP test and not just for the time period that we are studying at the moment. But... actually, I do have a question for chapter 27. Does anyone know why Third World countries are called \'\'Third World\"? I was just curious, so hopefully someone can help.   '),(9759,1488330,'11,April,2004','2004-04-11 00:00:00','    Hey APUSH students!  I hope you are studying.  Let\'s shoot for an extra-credit movie night next Tuesday @6:00...THE RIGHT STUFF...definately one of my favorite movies and books.  Anything by Tom Wolf is great - I am currently rereading some of Halberstom\'s work on the fifties - he was the historian who did the great Mike Hammer impersonation today.  He is tons of fun to read...he writes pages on Elvis, I Love Lucy and the McDonald\'s Hamburger chain. I love pop-culture.  As someone asked during class today, Chris Mitchem - why were Americans so scared of communism?  How did this fear start - and how did the Truman and Eisenhower administration handle this fear?  Did the leaders of this time add to the fear and hysteria or attempt to control it? Howard Zinn believes the administrations created this fear to stimulate our economy?  I am not sure I buy his idea...What do you think? Where does Pop-Culture fit into this...Captain America, Mike Hammer - Hollywood was infatuated with the Western at this time? Was Elvis a communist?  James Dean?  Marlon Brando? Lucille Ball had RED hair and married a Cuban????   Blog On... I think this time period could be easily related the roaring twenties...and today - what will this decade be called?   '),(9760,1488330,'11,April,2004','2004-04-11 00:00:00','    Tuesdays are Thursdays are good. Just on days when there is NHS (which is only twice), some people may come a few minutes late. But as for now, studying for AP is my top priority, so I will probably suck time out for the history study sessions if I have to. So it\'s an excellent idea.   '),(9761,1488330,'10,April,2004','2004-04-10 00:00:00','    i wish tuesdays and thursdays were good for me, but i have chemistry until 3:00 on those days at Clark. But it it\'s good for everybody else, then majority rules, and i\'ll have to use the internet to use for studying. Anyway, see you all tomorrow! and a Happy Easter to all those that recognize it in their religion, and a happy last day of spring break to those that don\'t!   '),(9762,1488330,'10,April,2004','2004-04-10 00:00:00','    Hey AP Students...I hope you had a good break and are ready to get back to WORK!  The next three weeks will move quickly  - and be somewhat intense.  Jessica your notebooks will be checked on Friday...along with a quiz.  We will be having a quiz every friday...five IDs from your reading.  I need to keep you honest - we have a few weeks until the AP test. The good news - you can choose to do your notebook or not...I will not be taking the time to grade it after this friday.   It would help if you have read and thought about the origin of the Cold War and its impact on America\'s society for class the next two days.  I also need the song - \"We Didn\'t Start the Fire\" Billy Joel - for class.  Does anyone or there parents have the Best of Billy Joel?  I may have it at school but I couldn\'t bring myself to enter the building this week.  I want to have afterschool sessions on Tuesday and Thursday - they may be time to view and discuss the Centuries, Multiple Choice or review certain eras.  How do Tuesdays and Thursdays work for you guys?   See you Monday... Eye of the Tiger...:)  Clark   '),(9763,1488330,'10,April,2004','2004-04-10 00:00:00','    You know what, that\'s actually a great question......   '),(9764,1488330,'08,April,2004','2004-04-08 00:00:00','    helloo everyone hope springbreak is going great for everyone!! JUST WONDERING if chapter 27 ID\'s and tests are DUE monday or it would just be helpful if we had em done??? if anyone knows i guess... blog back? lol THANKS! Jess   '),(9765,1488330,'08,April,2004','2004-04-08 00:00:00','    ....................... This blog sure got boring and dull fast...   '),(9766,1488330,'06,April,2004','2004-04-06 00:00:00','    So far, these are the members that have joined the AP US history blog: Mat Keksi, Hiliary Philip, Thomas Chen, Joseph Liberman, Kailie Middleton, Rob Muilenburg, Ben Perry, Christ Storms, Tristan McDaniel, Vanessa Esch, Megan Valencia, Mary Tyner, Jessica Lee, Sarah Nosler, and Ms Clark and Mr. Ray.   '),(9767,1488330,'05,April,2004','2004-04-05 00:00:00','    AH I got it from Paul, thanks.   '),(9768,1488330,'05,April,2004','2004-04-05 00:00:00','    i don\'t think she did...   '),(9769,1488330,'03,April,2004','2004-04-03 00:00:00','    Oh yeah, Ms.Clark, did you hand out some Multiple Choice Practice Test for us to do during spring break?   '),(9770,1488330,'03,April,2004','2004-04-03 00:00:00','    This is my third time attempting to blog - none have been published.    Thomas -  I have been in contact with Karen Hall - I will get back to you as soon as I hear something. By the way CONGRATULATIONS...so relax while we wait for her decision.  The other two blogs were full of witicisms - but now I am the ANGRY BLOGGER...what am I doing wrong.   '),(9771,1488330,'02,April,2004','2004-04-02 00:00:00','    Sorry everyone for using this blog for private purposes......but Ms.Clark, I have some emergency requests that just came out today!!!!!!!!!!! It\'s really urgent in that I was ridiculous chosen today in the Northwest Science Expo to take apart in the annual ISEF (Intel Science and Engineering Fair) (Representing get this: TEAM OREGON!)............And that happens during the second week of May!!!!!! I have to stay in portland the WHOLE WEEK and that week also freakishly coincides with AP Tests!!!!!!!!! So I have AP Chemistry test on Tuesday that I DEFINITELY CANNOT take in the Jim Parsley Center..........I have to either take the AP test DURING ISEF.....Which would require you to mail in the AP registration forms (By bloody this Tuesday), or I have to take it on a makeup day at JPC.......But in any case, the signup for me to take the test on May 10th AP Chem test at JPC is completely void now.  I\'m SOOOOOOOOOOO GRAVELY SORRY about this really short notice of this ridiculous inconvenience.  Hopefully you can take notice of this blog post as soon as possible.  Thomas Chen   '),(9772,1488330,'01,April,2004','2004-04-01 00:00:00','    Mr. Ray, thanks again for doing the reading thing today. That was cool, and now I have to get my box of childhood books from the attic and read them to see how they related to the times. I never thought of Dr. Seuss (sp?) as a political writer/cartoonist, but I now see him in a new light....   '),(9773,1488330,'01,April,2004','2004-04-01 00:00:00','    Thanks to y\'all for including me in your discussion group. Has anyone seen Hiroshima Mon Amour? It\'s an old film by Alain Resnais that begins with footage from the Hiroshima devastation and ties it to a woman\'s past experiences in WWII. It\'s been a while since I\'ve seen it, but it might be a good view as it reveals the impact (pun intended) of the war on both Europeans and the Japanese. It is beautiful in its portrayal of the devastation of war. It\'s very \'art\' and I think subtitled. Massive wonk points.  It has taken me many years to begin to see past the mechanics of history and foreign policy down to impact that this history has on people. In junior and high school, I was an avid student of WWII. I read books like PT-109, Guadalcanal Diary, et. al. made models of WWII airplanes. I even had an amazingly complex board game in which you could replicate almost any battle in the war. It was all VERY geeky in retrospect. I\'m sure part of this fascination was due to my mother\'s experiences in the war as a Finnish refugee and the fact that she and I would watch the BBC documentary World at War. But as a child, the war was fascinating because of the things, places, and \'stuff of war.\'   Thomas\' comment about the complexity of the issue of dropping the bomb is telling. Up until 9/11, non-combatant Americans never had to deal with the war on their own soil. The WWII experiences among an average European, American, or Hiroshima resident could not be more different. As a child refugee during WWII, my mother often had no food. Some Finns ate the bark off birch trees for nourishment. Her father was killed within the first few weeks of the conflict. As the oldest child, she became an erstatz father at age 12. Sugar was the least of her worries.  I never really appreciated this very real history when I was young. History and war were either subjects of grudging study or something that I read in a book. After travelling, college study, and some sporadic personal growth, I look at history very differently now. I\'m not saying that\'s better or worse, but my reading of history is now reflected through a desire to see it both from \'the big picture\' and from the viewpoint of the human beings that are part of it. It makes history far more complex, messy, and ambiguous.  As an example, my mom was \'the enemy\' in WWII. When the Finns sided with the Germans in \'42 to attempt to regain the land that had been taken when Russia (our ally) invaded Finland, they became part of the Axis Powers. The Finns lost twice in WWII. Because of the Molotov/Ribbentrop Pact, they lost their land to the Russians and were forced to pay reparations to the US and other allies for siding with the Nazis. (My mom was very proud that the Finns were the only Axis nation to pay every cent of their \'debt\' to the U.S. despite not receiving any of the aid given to Japan or Germany.) Despite the fact that my mother\'s heritage and former home is more Russian than Finnish, she strongly distrusted and disliked Russians until she developed Alzheimers. Now she loves everyone.   '),(9774,1488330,'01,April,2004','2004-04-01 00:00:00','    I don\'t know what the next topic is. We just took a test, so we haven\'t gotten to cover anything... But I\'m sure that\'ll change..   '),(9775,1488330,'01,April,2004','2004-04-01 00:00:00','    Welcome to the blog Rob Oh thanks Rob No prob Rob Haha, good rhyme Rob  What\'s the hot topic???  Love, Rob   '),(9776,1488330,'31,May,2004','2004-05-31 00:00:00','    Hello to all of my homies out in bloggerland.  I don\'t know if we realize how much time Thomas spends on this blog to make it run smoothly for us.  Next time anyone sees that wonderful little wizard, give him a hug and a kiss for being so damn tight.  Slow week in news, but I went and saw Day After Tomorrow and got thinking about global warming.  The danger is very real- studies have shown that seasonal allergy symptoms get worse every year thanks to global warming.  Steps have been taken to reduce its escalation (at least industrially), most notably through the Kyoto plan or protocol or whatever it\'s called.  Unfortunately for human kind, many countries that donate significant portion of greenhouse industrial excrement can\'t afford to change their industrial standards to reduce the CFC\'s and such that fill the atmosphere.  (Watch me spin this politically) I commend George Dubya for  expelling a hideous dictator out of the goodness of his heart despite the sacrifices of many Americans for purely humanitarian reasons without any nice byproducts that the American economy is STILL waiting for the benefits of... but I regret that the U.S. in all its glory can\'t aid an ally like, say, Russia modernize its industry which would not only benefit said nation, but the whole world.  Everyone elses thoughts on saving the environment and therefore the world?  Love Rob   '),(9777,1488330,'30,May,2004','2004-05-30 00:00:00','    I\'m not so sure that people understand what it is to be morally against something. It\'s not like we have a bible in our hands, knocking on people\'s doors and preaching to them about their horrible sins. We\'re not Puritans! That seems to be the stereotype that some may see whenever someone else uses morals as their base of their opinion. Whether it comes to Gay marraige, or abortion, y\'all probably know where I stand on those issues... or do you? You see, I don\'t mind abortion being legal, but I wouldn\'t use it. Even if I do have an \"accident\" and get pregnant, however that may work. And it\'s not like I\'m rich and have a lot of money, so I\'m priveledged enough that I can say that I can handle having a child. I would be dead broke, dropping out of school, getting a job, compromising with my parents, handling their disappointment... and so on with the negatives of teen pregnancy. AND STILL, I would have the child and either keep it, or give it up for adoption... making sure FOR MYSELF that it goes to two wonderful parents who can give the child opportunities that I wouldn\'t be able to provide. But you know, all that would go away if only I aborted the child. But (again, using my morals OMG) I wouldn\'t, I couldn\'t, I value the life that I created too much. But this is all a hypothetical situation guys, don\'t worry I\'m not pregnant. And I won\'t get pregnant til I marry, because I have that much control over that situation. But go ahead and keep abortion, because others feel that it\'s needed. Now was I using a bible to defend my reasons? A lot of people who think to the right don\'t.  When it comes to the Gay Marraige issue, I\'m not homophoebic and I\'m not going to shove a bible people\'s faces, but I don\'t recognize marraige of homosexuals as the same union as I would a marraige between heterosexuals. It\'s not the same, and it\'s not for reasons of what people do in the bedroom. It\'s much more than that, especially when there are children involved. And Rob, to give you a reason as to why I think it\'s wrong... it\'s because I believe that homosexual marraige isn\'t something that God designed. The design was for a man and a woman to have the marraige of holy matrimony. But to a secularist, I guess that doesn\'t matter. They don\'t use morals as the base of their opinion. They have to use facts, reasons and \"imperical data\" to share their beliefs, and in turn, they\'re viewed as the \"smarter\" people. Don\'t lie. I know the reasons why students who are viewed to be the smart ones are viewed that way. It\'s because they\'re left-winged/secularists who use data up the ying-yang to prove their point. And teachers love it... at least most professors at universities do. I\'ve been told that if I need a good grade in a course, I need to write a paper that agrees to the viewpoints of the professor. And they\'re all pretty much liberal. And don\'t tell me it\'s not true, because I\'ve seen it. I would talk on and on and on, but like Rush Limbaugh, I\'m running out of time. Yep, I listen to Rush and pretty proud of it. But I do want to say one more thing. It makes me feel bad that some people say I\'m intolerant of their views when in reality they are intolerant of my views. I don\'t hate homosexuals, I just don\'t agree with them. I don\'t hate those who choose abortion, I just don\'t agree with using it. What I\'m trying to say is that I want to take down the idea that right-winged, conservatives aren\'t open people. DOn\'t believe that stereotype.   '),(9778,1488330,'30,May,2004','2004-05-30 00:00:00','    Ok, just in case you have not realized, I\'ve completely changed our AP US History website for the better. There is a \"contributing members\" section on the sidebar that allows us to look at our APUSH friend\'s blogging profiles. And I\'ve added three things:  1: tagboard(which is for visitors to post; or you can post short comments and quick questions and answers on some small particular topics - there is a limit to how many words one can post on the tagboard at once), 2: The title bar: The title bar serves to better organize our blogs by forcing everyone of us to post the topic we will be discussing in our posts. Thus people will know exactly what the post is for w/o reading sometimes long and tedious posts. 3: Comment option for every post. Hey, if you feel somebody\'s got a very interesting, or hilarious, or horrible post, feel free to write some comments about that peculiar post.   THANKS ALL MY FRIENDS!  Thomas   '),(9779,1488330,'28,May,2004','2004-05-28 00:00:00','    ok... i just got back from watching a most excellent movie... The Day After... no, not that new one in the theaters (the day after tomorrow)... the one made in the 80s about nuclear war... some of you probably dont know what i\'m talking about... but some (hopefully Ms. Clark) will know... I thought it was a good movie.. and brought up one of my favorite quotes from Einstein: \"I don\'t know how world war 3 will be fought, but i can tell you how world war 4 will be fought; with sticks and stones...\" very grim indeed, but a very good movie...   '),(9780,1488330,'28,May,2004','2004-05-28 00:00:00','    Ok, i realize i havent been active in any of the blogging going on this week, but i have finally caught myself up and want to respond to some of the ideas you all have presented. I know Ms. Clark is trying to get us to talk about a new issue, but bare with me.   I understand all sides of the different arguments that have been presented, so i commend all of you for putting your ideas out there. I think the arguments that are clearly stuck in my head are Rob\'s (with the hope of rallying up support for the blond hair blue eyed heterosexual caucasian white males), Natalya\'s (arguing on both GM and EC) and Megan\'s. Maybe its because i agree with a lot of what they have said. Unlike Megan, i dont know many homosexuals, however my heart does go out to them. But I do realize that i dont have the same religious background as others that could hinder my sympathy for them. I still am pushing that there is a definite separation between church and state. This should be remembered when arguing whether or not GM should be legalized.   I know i posted a week ago, and it\'s so far back in the past that im sure it has been forgotten, but i said a lot of what Natalya said, just not as eloquently (she is of course an experienced debater!). I agree completely with her arguments, particularly on the debate over abortion and EC. I know people have moral reasons for not agreeing with its usage, however mistakes happen. It isnt fair to put an innocent child in an unforgiving world. No matter what opportunities are trying to be given to make up for lower-income, hard lives, the pain inside cannot be erased. They are brought into the world unwanted, permanently branded. And i agree, it is up to each individual to make their own decision. Being a fairly independent female, it scares me to death to know that the option may not always be open to me. Accidents happen. I know that i wouldnt want one bad decision to effect me, my unborn child, my family, and the father of the child (considering rape isnt involved). Each decision has repercussions, effecting not only one individual, but all those who surround him/her.   I\'ll admit it now: I am pro-choice on pretty much every aspect of social controversy. I dont think it is acceptable for a group of individuals, no matter what their backgrounds or personal beliefs may be, to make decisions that effect the masses. There are always instances where people should just have the right to make their own decisions, b/c in the long run, each individual knows what is best for himself/herself.   Addressing the idea of whether or not abortion or gay marriage should be reason to sway one\'s vote, i think its up to the individual.Unfortunately, i will not be 18 for the coming election and thus i cant say that i can answer the question without a little cynicism, knowing that it doesnt effect me for another 4 years. But i do think that it is worth researching and looking into. I think it\'s important to research all of the different positions each candidate is taking, economically, their domestic policies, and foreign policies. I dont think it would be wise to make the decision solely on one aspect of their campaigns. Basically, it\'s choosing the lesser evil. I dont think anyone will ever completely agree with all the prospective policies, but its choosing which one you have lesser opposition to.  Well, i hope i have covered all all the topics that have been discussed over the past week. But bare with me if i havent, a lot of ground was covered between now and the last time i posted!  Kailie   '),(9781,1488330,'27,May,2004','2004-05-27 00:00:00','    I want to make sure everyone understands that Rob is using sarcasm and exaggeration to make a point.  Please read in between the lines when you read Thomas and Rob\'s Blogs...boys be careful in your writing so you are not misunderstood.  (Yes I am a worry wart, a control freak and high maintenance.   I am curious what you think about Fog or War...I know I hindered second periods viewing... I am fascinated with Robert McNamara...he is brilliant and has learned and willing to share some of his misfortune.  Do you think he is fair to LeMond (spelling)?  or Johnson...neither man is able to defend himself? What point has had the most impact...the proportionality argument was very powerful.  I also enjoy Errol Morris\'s style...music and some of his effects are interesting while I found a few annoying.  Next week...Get yourself ready!  The issue for dabate will  be whether the US should act unilaterally or in concert with others?   Specifically, how should American policy regarding the use of military forces in the  war against terrorism change to deal with the largely stateless nature of the  current terrorist threats?  They will play the role of Congressmen, Senators,  National Scurity Advisors, Sceretary of Defense, or Secretary of State.  The  problem will be to create consensus on a national security policy.  The outline will be as follows:  I.  War powers:         US Const., Art. I - the powers of Congress        US Const., Art II- the power of the Executive        Review of declared wars         Instances of Use of US Armed Forces Abroad, 1798-2001        http://www.fas.org/crs/man/RL30172.pdf            urlLink Tonkin Gulf Resolution (1964)   urlLink War Powers Resolution (1973)          Military interventions since Vietnam        Post 9/11 Resolution (2001)   urlLink National Security Policy of the USA (2002)         Colin Powell, Partnership and Principle, Foreign Affairs, 83:1, pp.  22-34, Jan/Feb        2004        Samuel Berger, A Democratic Foreign Policy, Foreign Affairs, 83:3, pp.  47-63,        May/June 2004  urlLink Charles Knight, Essential Elements Missing in the NSS of 2002           urlLink         Joint Resolution . . . Iraq (2002)            II. International law of War:         Traditional justifications for war:  defense against attack on nation,  protection of         citizens and property abroad          urlLink UN Charter, Arts. 33 & 51 and Chaps. IV-VIII                  urlLink UN and treeorism                   urlLink Geneva Conventions (III)            urlLink  Douglas Stuart, Reconciling Non-intervention and Human Rights                  urlLink UN Resolution 1441 on Iraq (11/2002)           Treaties for collective security        NATO, Arts.  Pre-emptive vs preventive war          urlLink Hal Culbertson, The Ethical Challenges of Preventive War                   urlLink Duncan Currie, \'Preventive War\' and International Law after Iraq                    urlLink US Use of Preemptive Military Force           This is more than enough for now!  If students would pick one of the issues  and do some research they will have lots to talk about.   Alan   '),(9782,1488330,'27,May,2004','2004-05-27 00:00:00','    Haha Rob, I\'m not sure if I understood your last post correctly, but it seems like it was full of metaphors and analogies. First of all, GAAAAA the oxymoron combinations of Leftist/Fascist/SuicideBomber/Know-Nothing political jargons mixed with some of your peculiarly molded mind has transplanted the great picasso like abstarct ideas into our fantastic debate blog. Emerson would laugh so hard.  Secondly, I think you were the first person to mention that gay people are more loving than heterosexuals. \"Others would say that gays don\'t hold the institution of marriage is the same esteem as heteros, but I would argue that they hold it higher. Marriage is something that this group has to fight for\". Although I don\'t deny that statement at all.  Third of all, for affirmative actions, I think Asians (especially oriental Asians) are as bad, or even worse than caucasians in lot of cases. Go figure. I actually wish I was white. Heh being a minority without the minority benefits and without the majority race ADVANTAGES is like a bogus race to me. WOH NEXT NINE HOURS KNOW-NOTHING SUPPORTERS PLEASE JOIN ROB MUILENBURGE! PROTECT THE INFALLIABLE RACE OF BLUE EYED BLOND HAIRED CAUCAUSIAN MALES! Oh yeah, you HAVE to be STRAIGHT. NO GAYS ALLOWED!   INFALLIABLE RACE OF BLUE EYED BLOND HAIRED CAUCAUSIAN MALES DOMINATE THE EARTH, I know my friends. Although I\'m ok with that. it\'s life. It\'s the majority race that wins everything.   ANY MORE POLITICAL SATIRE NEEDED PLEASE POST   '),(9783,1488330,'26,May,2004','2004-05-26 00:00:00','    Thomas, let me extend a welcoming hand for you to shake in joining my militia of pro-homosexual terrorists.  Clearly we\'re becoming more militant by the second, and sooner than later we\'ll be suicide bombing the Fox News Channel.  Just kidding everyone, but I should make it known that ASB President is just the first step on the staircase of Ultra-Leftist Facism. I dare anyone to try and figure that out.  Anyway, I\'d like to make it known that I was joking when I told Kolk she was flat out wrong.  That was me making a joke as well as pleasing my ever-expanding ego (I no longer recognize other opinions).  JKo knows I\'m just playin, so all\'s gravy in my eyes.  I decided that I\'ll take a few hours off of GM rants and focus my attention on another group of individuals that is having a harder and harder time defending itself:  Heterosexual able-bodied blond hair blue eyed caucasian males.  In Thomas\' post he mentioned the fact that gays are probably more loving w/ a probable smaller divorce rate.  It\'s that kind of heterosexual discrimination that I\'m getting sick of.  Ask Chris Mitchem how hard it is to get a milkshake at McDonalds without being sneered at by the caucasian females working behind the counter- not knowing what vile bodily fluids have been hawked into his order.  Heterosexual able-bodied blond hair blue eyed caucasian males unite!  We want our close parking spots back from the handicapped.  Our innercities back from the various ethnicities.  Gynecology back from those oppressive stick-together females.  Thomas Chen will never know the difficulties of non-affirmative action.  Ask Ben Perry how hard it is to get a scholarship for being a non-minority!  For the next 9 hours I\'d like to reinstate the Know-Nothing party.  Who\'s with me!?  After that nine hours expires it\'s very likely I\'ll go back to the minority supporting spam that got me attention in the first place, but until then, all you non-heterosexual able-bodied blond hair blue eyed caucasian male bigots better look out.  Power to the people!  Love Rob   '),(9784,1488330,'26,May,2004','2004-05-26 00:00:00','    Wow Rob you were exactly right. Yesterday was a lull before today\'s storm. I have to say that although I take hardsides and are extreme sometimes, I think that it\'s disrepctful to call other people flatly wrong. Political issues will never have a complete 100% right or wrongs. Yeah so there!  On to the gay marriage debate, I will have to side with hardline pro-gay coalitions forces. Rob\'s article lets you have a glimpse of the emotional traumas and unexplicable sorrows and guilts gay people already have in their minds that they cannot communicate. Much of it of course caused by the phobias from society. Most gays have gone through (or are going through) constant fear of discriminations from the society already. These frivolous fear of gays (WITHOUT ANY SINGLE SCIENTIFIC BASIS!) are simply baseless and are simply stemming from the conformist stereotypical mainstream society\'s view. Denying the gays the right to marry (for they probably have more love and less like to divorce anyways) simply manifests itself from subtle stereotypical view towards gays, even if you don\'t want to admit it.  Discriminating against gays is completely analogus to any other form of discrimination of any RACE, or even GENDER. Just because they physically look different, act different, think differently DOES NOT make them inferior. ONLY a criminal would do so.  This great nation generously help the disabled, and the mentally retarded. We have special access, special privileges, and special facillities for these people so they can have the same opportunities as everybody else. And besides that gays are as intelligent and as abled in every intellectual as straights. Any supposedly \"disabled\" abilities of gays are purely the result of social pressure and phobias. If we do do our best helping disabled , why not help gay people achieve equality too? And of course everyone knows that disabled and mentally retarded people are that way because of nature(They didn\'t become disabled because it\'s cool to!) and there should be no denial into the genetic factors behind gays, AND so WHY favor some biogenetic factor over others??  How can we truly live up our meaning of freedom and equality for all if we turn a blind eye on a GROUP of citizens? I apologize if this post offended anyone.   '),(9785,1488330,'26,May,2004','2004-05-26 00:00:00','    I don\'t know why we can\'t change peoples\' feelings or beliefs.  I think we have to.  I believe it was MLK that said \"we can\'t stop people from hating us, but we can make laws that makes it illegal to lynch us.\"  The change has to start somewhere, and protecting homosexual rights in the courts is the first step towards their universal acceptance.  (Ferg is gonna hate me for this but) Take the Civil War for example.  The 13th ammendment freed the slaves, the 14th protected them in the courts, the 15th let them vote.   Although gays haven\'t been forced into labor and don\'t need emancipation, the next step is to protect their rights.  Last year there were 2051 reported hate crimes based on sexual preference, up 8% from the number in 2002.  Along with protecting gay civil rights, the advancement of homosexuals will hopefully reduce the very negative anti-gay sentiment (whether conscious or not) in this country.  Kolk, you give me a good reason why gays don\'t have the right to marry and I\'ll either tell you why you\'re wrong, or apologize profusely.  Everyone is \"different.\"  Why limit marriage requirements to what is done in the bedroom?  Without getting naughty, there are plenty of crazy ass things done in the bedroom besides same sex interaction.  Plain and simple, civil unions divide the society over reasons the majority of the posters find unworthy of debate.  I also want comments on the article by Andrew Sullivan- a homosexual republican.  Go Figure  Love Rob   '),(9786,1488330,'26,May,2004','2004-05-26 00:00:00','    Though I haven\'t had the same experiences, I totally agree with Megan. I don\'t understand how being gay is morally wrong. I think that debating GM can be very pointless. The main issues lie in a person\'s beliefs, and debating the issue will not change how someone feels. The only debate is whether our great nation will allow everyone the same equal rights. I think it\'s amazing that our nation still hasn\'t resolved this issue, because to me it doesn\'t seem very complicated. Either we allow everyone to marry, and marry whomever they choose, or we deny part of our nation these equal rights. Marriage is between two people, who gave the rest of the country the right to tell them they cannot get married?  Steph B.   '),(9787,1488330,'26,May,2004','2004-05-26 00:00:00','    ROB IM NEVER WRONG lol but yea once again Gm is that whole opinion thing... man n man woman and woman isnt marriage its like callin an apple and orange gays can be unionized NOT married hav a great day   '),(9788,1488330,'26,May,2004','2004-05-26 00:00:00','    Being gay is not morally wrong. I don\'t understand how people can say this. It doesn\'t make any sense to me. I never realized how many people have \"problems\" with gay people until I moved to Washington. It really makes me angry and I just fill with sadness. I\'ve been around gay people my whole entire life. There is nothing anyone can tell me that would make be believe that being gay is wrong. It\'s so hard for me to debate this issue, because I believe that it should not even be an issue. The fact that people I care about can not legally marry the person they love infuriates me.  Does God think it\'s \"right\" to be gay? Is God ashamed at my talented intelligent cousin for being gay? Has she been a sinner her whole life just for being gay? Should she be looked down upon for being gay? Well she is. And she is terrified at what the reaction of other people will be if they find out that she is gay. Responding to Clark\'s questions about the effect of personal rights and civil liberties on presidential elections I think that is has great relevance. Bush is against gay marriage. This is what swayed my vote. If I was born 12 days earlier I would not vote for Bush. This may seem like a stupid reason to decide how to vote, but it\'s reason enough for me. This is how I feel. I can not rationally debate this topic because hearing the opposing side just breaks my heart.   '),(9789,1488330,'26,May,2004','2004-05-26 00:00:00','    Kolk, you\'re wrong.  I don\'t see how you feel GM changes HM in any way.  Props to your EC stance, but MARRIAGE is as important to gays as it is to straights at least according to this article.  Check it out.  Why The M Word Matters To Me  Only marriage can bring a gay person home  By ANDREW SULLIVAN  Feb. 16, 2004  As a child, I had no idea what homosexuality was. I grew up in a traditional home â€” Catholic, conservative, middle class. Life was relatively simple: education, work, family. I was raised to aim high in life, even though my parents hadn\'t gone to college. But one thing was instilled in me. What mattered was not how far you went in life, how much money you earned, how big a name you made for yourself. What really mattered was family and the love you had for one another. The most important day of your life was not graduation from college or your first day at work or a raise or even your first house. The most important day of your life was when you got married. It was on that day that all your friends and all your family got together to celebrate the most important thing in life: your happiness â€” your ability to make a new home, to form a new but connected family, to find love that put everything else into perspective.   But as I grew older, I found that this was somehow not available to me. I didn\'t feel the things for girls that my peers did. All the emotions and social rituals and bonding of teenage heterosexual life eluded me. I didn\'t know why. No one explained it. My emotional bonds to other boys were one-sided; each time I felt myself falling in love, they sensed it, pushed it away. I didn\'t and couldn\'t blame them. I got along fine with my buds in a nonemotional context, but something was awry, something not right. I came to know almost instinctively that I would never be a part of my family the way my siblings might one day be. The love I had inside me was unmentionable, anathema. I remember writing in my teenage journal one day, \"I\'m a professional human being. But what do I do in my private life?\"   I never discussed my real life. I couldn\'t date girls and so immersed myself in schoolwork, the debate team, school plays, anything to give me an excuse not to confront reality. When I looked toward the years ahead, I couldn\'t see a future. There was just a void. Was I going to be alone my whole life? Would I ever have a most important day in my life? It seemed impossible, a negation, an undoing. To be a full part of my family, I had to somehow not be me. So, like many other gay teens, I withdrew, became neurotic, depressed, at times close to suicidal. I shut myself in my room with my books night after night while my peers developed the skills needed to form real relationships and loves. In wounded pride, I even voiced a rejection of family and marriage. It was the only way I could explain my isolation.  It took years for me to realize that I was gay, years more to tell others and more time yet to form any kind of stable emotional bond with another man. Because my sexuality had emerged in solitude â€” and without any link to the idea of an actual relationship â€” it was hard later to reconnect sex to love and self-esteem. It still is. But I persevered, each relationship slowly growing longer than the last, learning in my 20s and 30s what my straight friends had found out in their teens. But even then my parents and friends never asked the question they would have asked automatically if I were straight: So, when are you going to get married? When will we be able to celebrate it and affirm it and support it? In fact, no one â€” no one â€” has yet asked me that question.  When people talk about gay marriage, they miss the point. This isn\'t about gay marriage. It\'s about marriage. It\'s about family. It\'s about love. It isn\'t about religion. It\'s about civil marriage licenses. Churches can and should have the right to say no to marriage for gays in their congregations, just as Catholics say no to divorce, but divorce is still a civil option. These family values are not options for a happy and stable life. They are necessities. Putting gay relationships in some other category â€” civil unions, domestic partnerships, whatever â€” may alleviate real human needs, but by their very euphemism, by their very separateness, they actually build a wall between gay people and their families. They put back the barrier many of us have spent a lifetime trying to erase.   It\'s too late for me to undo my past. But I want above everything else to remember a young kid out there who may even be reading this now. I want to let him know that he doesn\'t have to choose between himself and his family anymore. I want him to know that his love has dignity, that he does indeed have a future as a full and equal part of the human race. Only marriage will do that. Only marriage can bring him home.     CopyrightÂ Â©Â 2004 Time Inc. All rights reserved.   '),(9790,1488330,'26,May,2004','2004-05-26 00:00:00','    HIII its jessica KOLKMEYER people have been gettin me n KNAB confused lol  anyways this is going to be a multi topic blog...... ABORTION AND THE MORNING AFTERPILL two of my favorite topics... abortion is WRONG it is murder etc etc the morning afterpill simply puts up an obstacle course for the little soldiers so they can\'t meet their fair lady! i agree 100000% with it i have had friends in the situation and its a scary one i know i dont want to me Auntie jess. Sex is a big deal if you have sex b responsible... IE condoms yes they fail then that is why you keep EC at home or make sure you or one or two of your friends has it so if anyone has an OOPSY then its solved. The abortion pill is call Ru-486 i remember the huge controversy over this pill in recent years, this is the one that KILLS what is already MADE it causes the cervix to open causing contractions which expel the embryo.. ok we are talkin about an embryo a LITTLE BEBE think of the difference of EC and the abortion pill.... like rob said the safest sex is NO sex! NOW gay marriage MARRIAGE IS FOR MAN AND WOMAN SORRY TO BURST YA\'LLs bubbles  im not against gays or lesbians i think they were born that way they can\'t help it BUT  they can hav a civil union and get the tax benefits marriage is DIFFERENT i mean do you see them calling them selves husband and wife if they are two women no they are \"lifepartners\" or somethin else great.... ya we live in america we are ENTITLED TO OUR RIGHTS as people who marry heterosexually should they   get that right the gays get a civil union OK sounds GREAT who needs to sign the bill????   and finally my book.... I was reading catcher in the rye now im on to Tom O\'brian Tomcat in love.... its a rather racy book as we dicovered on the buss back from st helens!   well i hope everyone has a super day day and remembers \"safe sex is no sex and straight marriage is best!\"   '),(9791,1488330,'26,May,2004','2004-05-26 00:00:00','    Hey all of you happy bloggers. Paul, it is a good thing I will not be old enough to vote in this election because I don\'t like either candidate.  Anyway I am really posting to discuss the Da Vinci Code, a book of which I just finished reading. With all of the hype concerning this book and input from others who have read, I was excpecting a great book that would address controversial issues of religious history. The book failed to meet the standards I had set, and completely let me down. It actually turned out to be the most annoying book I have ever read. First of all the author, Dan Brown, used all kinds of foreshadowing trying to keep you stuck on the book. He would hint on subjects or different happening that wouldn\'t actually take place until 4 chapters after comment. So as I read I would try to concentrate on the matters before me but I constantly had the thought of his foreshadowing in my mind. So I could not put the book down, and usually people use that phrase to descirbe a book that really interests them but in this case I just needed to find out the stupid answers to the stupid questions. By the time I found the answer or I reach the happenstance that had been revealed tens of pages beforehand, I would now have experienced more of the same types of comments and the vicious cycle would restart. I know that this kind of writing is common but I think the author really overused its use and it started to get on my nerves. I guess that means Dan Brown is a good writer because however conniving his tactics, I could not put the book down.      Another problem I had with this is finding the difference between fact and fiction. There were some very interesting ideas posed but I would like to see justification for some of the statements made because the only way they would be cool is if they were true. Anyway despite my conflict the Da Vinci Code, I am reading the prequel Angels and Demons, I guess I am still trying to fulfill my initial hopes for greatness from the first. If any of you have read the book I am curious as to your thoughts on it.  Rob you are my sunshine Chris   '),(9792,1488330,'26,May,2004','2004-05-26 00:00:00','    I heard Gore\'s speech yesterday, very, um, how should we say eloquent.  He\'s about as motivating as a tree stump.  A dead tree stump.  Seriously, I love Howard Dean, the dems should have run him.  \"And then we\'re going back to Washington DC to take back the white house- YAAAAAAAAAHHHHHHHHHHHHHHHHHHH!\".  Howard Dean is great.  Kerry is lucky to have him drumming up support for his campaign.  In other news Kerry backed out of his plan to not accept the nomination(which would not allow him to raise anymore money).  He had the gaul to blame Bush, \"Because Bush didn\'t limit his take to 19 million and take matching funds, it forced Kerry to consider not taking the nomination\" to paraphrase.  What an ass.   '),(9793,1488330,'26,May,2004','2004-05-26 00:00:00','    So, I looked up Robert McNamara - very controversial.  I\'m excited to watch the movie.  I think it\'s always interesting when people who actually shaped the past are able to look back with perfect hindsight and evaluate what they did - right and wrong.  He said in an interview, \"This century [the 20th] will go down as the bloodiest...we\'ll have lost 160 million people killed in conflict.\"  That\'s too many, I hope THIS century will be a lot more peaceful.  Speaking of which, how about the new warning from Ashcroft?? And... Gore\'s speech today, OH MY!  Anyway, have a good night!   '),(9794,1488330,'25,May,2004','2004-05-25 00:00:00','    I\'m reading main street by sinclair lewis.  I don\'t see why you have to stop reading dorian gray it shouldn\'t be that big of deal.  We should debate about how best to help the impoverished in this country.  Very large, good topic!   '),(9795,1488330,'25,May,2004','2004-05-25 00:00:00','    I got two questions....   1.) what is everybody reading for their book?  2.) Ms. Clark, i\'m an idiot! I started reading my book (THe Picture of Dorian Gray), and started to get into it, but i realized that its taking place before modern times.....  (Oscar Wilde died in 1900) :-/ urg... i\'ll have tin find another book.. is that alright?   '),(9796,1488330,'25,May,2004','2004-05-25 00:00:00','    Hey Bloggers...Who says the abortion topic is dead.  You seem to have plenty to say about it...and your knowledge about contraceptives is very admirable.  I also appreciate how you relate it to the constitution and personal rights.  So many major issues in American history are related to personal rights and civil liberties.   Do you think this issue has enough relevance to sway how you would vote for president?  Should it?  College portfolios...I think the college admissions process would be an interesting topic to debate...especially after out discussion in 2nd period today.  I think the movie Fog of War will lead to some interesting discussion.  For some of you it may help if you look into the life of Robert McNamara before watching the movie.    I love the post AP period...it actually gives us time to get to know each other.  Have a good evening and see you tomorrow.   '),(9797,1488330,'25,May,2004','2004-05-25 00:00:00','    Hey Guys! I can Blog now! YAY! luv ASH! :)   '),(9798,1488330,'24,May,2004','2004-05-24 00:00:00','    Based on my own beliefs I\'m against abortion.  I feel that both the morning after pill and first term abortion are wrong for me.  Adoption can be much better in the long run for the baby and the mother.  Many women who have had abortions say they wish they hadn\'t.  Abortion is not something you can take back.  Before women choose to have an abortion they should realize and understand its effects on their body.  However, this is America, one of the few places on Earth were people can live freely.  Therefore to deny a person the right to choose what to do with her own body goes against what we as Americans stand for. One issue that really itches me is what role should fathers play in abortion?  In reality it took an man and a woman to make the baby, but the baby only grows inside the woman.  Is abortion a woman\'s struggle?  Just curious.   '),(9799,1488330,'24,May,2004','2004-05-24 00:00:00','    I agree with Natalya.  I don\'t think that a group of people (aka the government) can impose their beliefs on another group of people.  On the issue of GM, it is ridiculous to say that two people who love each other and are willing to make that kind of commitment to each other can\'t because said people are both of the same gender.  Also, I think that the \"slippery slope\" argument is stupid.  Just because I see two women getting married on TV, I\'m not going to go out demanding all kinds of other pointless things like the right to marry my cat.  And as for abortion: while the idea of killing babies (or fetuses that could potentially be wonderful people someday) really bothers me, I don\'t think that I have the right to tell my friend that they can\'t have an abortion.  It was interesting because I went into class set on the opinion to overturn Roe v. Wade, and left class with the opinion I just stated.  Finally, the fat kids.  I really don\'t know what to think.  On one hand, I think that taxing junk food is a little extreme.  But on the other hand, children are ridiculously occupied with Doritos and Halo, and do absolutely NOTHING.  But how could the government take a role?  They obviously attempt to control what is being served in schools and make it nutritious, it just as obviously doesn\'t work (look at Skyview\'s menu).  P.E. classes are ineffective.  And the government, while they can run things like schools, shouldn\'t be allowed to dictate what people eat at home.  I guess the best solution is the one Tyson mentioned, where the government looks at what corporations are putting in their foods and look for healthier alternatives.   OK, I\'m done now. If there is a flaw in any of my arguments, I\'m sure someone will let me know. Bye.   '),(9800,1488330,'24,May,2004','2004-05-24 00:00:00','    I think it\'s important to be realistic in this debate, if not simply because idealism will fail to present any real answers.  Pragmatism is key, and I think everyone needs to keep in mind that while America is a land which prides on being the land of opportunity for millions it is not.  America\'s middle class has been shrinking, and will continue to shrink while the lower class grows and upward mobility stalls.  With the education system in general disarray especially in impoverished areas, I think it\'s rather unrealistic to expect a majority of foster children or impoverished children to grow up to cure cancer or become doctors lawyers and indian chiefs(due note the culture reference there).  Not that I approve of abortion but we have failed to help the lower classes in the country, the reasons we have are up for debate(and would make for a good one), but I think not exasperating the problem is good policy.  Gay\'s should marry, marriage is secular in that it\'s awarded by counties irregardless of religion and brings with it financial benefits tax-wise.  Since these are both government imposed and our government is secular and thus should take no cues from religious authorities, I think gays should be allowed to marry.  If the word marry bothers people too much, I wouldn\'t be too opposed to changing the official term to civil unions for everyone and making marriage a religious only institution.  Maybe we could get \'civil unionized\' at the justice of the peace, but married at a church, and the churches could decide to not let gays \'marry\' perse but they could still reap the financial benefits and official representation from a \'civil unionization\'.  Or we could just get it overwith and let the gay people marry.  It\'s going to happen eventually, why put up such a fight?   '),(9801,1488330,'24,May,2004','2004-05-24 00:00:00','    I personally think its wrong, better yet \"morally wrong\" to impose your morals on any other person or a group of people, that is concerning gay marriages and abortion. I mean if religious people believe that abortion is wrong no one forces them to have it, but they can\'t hinder other people in their choice to have abortion. There is lot of idealistic talk around here, especially saying that the problem would be solved if people simply stopped having sex, but the fact is human instincts often take over reason and a lot of people don\'t even see anything wrong with having sex. The question then is how we can help those people that do have \"accidents\" and are irresponsible. I heard a lot of people say that instead of being aborted the child should be given to foster care, which I think is even greater irresponsibility. It\'s irresponsible to throw away your child once you gave life to her/him, abortion in a way is taking responsibility of your actions; once that child is in the foster care you will never know what kind of a life that child will have, whether he/she will be abused, sexually harassed or simply have an unhappy life knowing that he/she was an unwanted child, thrown into the world as a homeless dog. Many say that there are a lot of couples that simply dream adopting cute little babies and providing them with a wonderful life, but there are lot of kids in foster care around the world and all of them in no way possible will have enough couples to be adopted by. Anyway, I think the pill is one of the most important innovations in history of womankind, it eliminates that dead end of either having abortion or giving birth and then giving up that precious life to the \"cruel and perilous world.\"   Concerning gay marriages, Rob has voiced my opinion in the best way possible. Afterall this is America...  my country tis of thee, Sweet land of liberty, Of thee I sing. So shouldn\'t we all have the right to love and to live our lives happily as we wish to as long as we don\'t harm anyone else. I mean do you become physically or mentally harmed if you see a married homosexual couple? What goes on in their bedroom is of no ones business but theirs, and I think who should be of more concern are the preachers who vigorously preach about homosexuality as a sin and a \"morally wrong doing\" and then go an have \"fun\" with their own little \"buddies\" seven years old, sick.   '),(9802,1488330,'24,May,2004','2004-05-24 00:00:00','    I think everyone in this blog is missing the point of emergency contraception.  It does NOT cause an abortion.  Because that fact seems to get so easily confused, let me explain how pregnancy and conception work.  After intercourse that ends in ejaculation, sperm cells may live in the very hospitable environment of the vagina for about 5 days.  If there are sperm still living in the vagina during the time of ovulation it is very probable that a pregnancy will occur.  What the morning after pill does is prevent the egg from being released therefore preventing pregnancy or prevent a fertilized egg from implantation in the uterus where it would THEN begin to mature.  This IS NOT an abortion!  With that logic, everyone that takes the birth control is instigating one indefinately long abortion, which they clearly are not doing.  Furthermore, the point of emergency contraception is not to be a primary source of birth control in the first place.  A fully functioning condom used properly is more effective than EC, but sometimes the condom comes off, breaks, or \"little spermies get out\" for whatever reason, opening the door to an unwanted pregnancy by responsible sex-havers.  Another scenario is a rape/date-rape situation where sex was not planned.  It is unfair to the rape victim as well as the bastard child that results from the unwanted pregnancy.  Lastly, I know there will be a situation in which one of these bloggers will get caught up in the passionate moment and omit protection.  I\'m not excusing irresponsibility, but why risk the pregnancy if there is a safe, non-abusable form of contraception that could eliminate the unwanted consequence?  It\'s fine to criticize EC\'s use from a seperated vantage, but wait until you or your partner is panicking because the condom broke, and try to tell me you wouldn\'t want an easily available solution that ultimately PREVENTS abortion if that is the last resort for an unwanted pregnancy.  I honostly don\'t see where a debate against this medication stems from other than ignorance, but consider yourselves educated now.  The only safe sex is abstinence...  Love, Rob   '),(9803,1488330,'23,May,2004','2004-05-23 00:00:00','    I have to disagree with Chris on the morning after pill and abortion. They both terminate a pregnancy. I think if people think they are responsible to have sex, then they should be responsible enough to use protection. I know condoms aren\'t always 100% effective and sometimes little sperm-ies get out, but if you are responsible enough to have sex, you have to be responsible enough to deal with the consequences, whatever they might be. I think some people use abortion and the morning after pill as a form of birth control after they have had sex, rather then using birth control before or during. If you aren\'t ready to have kids, then you aren\'t ready for sex. I think if people weren\'t so caught up on acting older then they are, we wouldnâ€™t have to have this talk about whether abortion should be legal or not, because there wouldnâ€™t be a high number of unwanted pregnancies.   '),(9804,1488330,'23,May,2004','2004-05-23 00:00:00','    Wow, it is nice to see all of the debate going on and it is nice to see everyone\'s different views. Rob, congratulations on the longest, most well-constructed blog in the history of blogger. The current debate topic of abortion and birth-control seems to manifest a rift between morals and science. The problem now is to recognize the difference and form opinions/arguments. Chris is certainly right-on when he talks of the moral similarities between taking the morning-after pill and an abortion. However, as the morals are similar, the science and event of each are so drastically different that it separates the moral argument on a scientific level. I would say that it would be much more demoralizing for a woman, perhaps her boyfriend and parents as well, to go through a surgical abortion. If taking a pill after sex allows a girl to not have to go through with an abortion this then reduces a life-changing experience to a gulp...And a swallow. Therefore, the repercussions of each event and the moral affect that those repercussions have, separate the two.   ~Holla   '),(9805,1488330,'23,May,2004','2004-05-23 00:00:00','    Hey all, I don\'t mean to light the fire for a debate but I don\'t have a problem with abortion, and I don\'t see how abortion and the morning after pill differ in purpose of use. Whats the difference between stopping something thats going to exist than ceasing something that barely does. A pregnancy is the process of producing a baby, once a mommy and a daddy have sex and the sperm meets the egg that process begins. Whether a person recognizes the possibility of a baby earlier or later, the possibility and probability of a baby are the same in both situations. If one has to use the morning after pill to stop the inevitability of a baby growing, then it is obvious that the person taking the  pill beleives that a baby has been conceived, and wishes to cancel that conception. Cancelling the conception is taking the possibility of existence away from that baby. Now how does that differ from abortion? Due to that, I don\'t see how people can justify their negative views on abortion while supporting the morning after pill.  When it comes down to it, both methods of ending a pregnancy are ways out for people who are not ready to become parents. And it is my personal opinion that a way out three days after conception does not differ from that of 3 weeks.   -I love you guys,  Chris Mitchem   '),(9806,1488330,'23,May,2004','2004-05-23 00:00:00','    I\'m going to kind of answer my own question about what people\'s opinions are of the morning after pill, because I never really did answer it myself.  For some reason, I really do not consider it to be abortion.  Because you can only take it up to three days after you\'ve had intercourse, and you also at that point have no way of knowing if you\'re pregnant, so I don\'t consider it taking a life.  As Rob said (or maybe Ms. Clark said it, I can\'t remember), the morning after pill is much more of a form of birth control than an abortion tool.  Abortion is a deliberate choice to take a babies\' life in it\'s (sometimes) later forms of developement, while the morning after pill is simply a safety net.  As for the gay marriage issue...this issue is extremely sticky to get around.  I personally, partly because of my religious background, think that being gay is morally wrong.  However, we live in America.  Here we are supposed to be able to have the freedom to do as we please, get the job we want, eat what food we want, and marry who we want.  It then becomes an issue of how far the government is going to get into Americans personal lives and try to take control.  Do they really have the right to say that you can\'t marry who you want to just because it\'s morally wrong?  Perhaps that\'s stepping in too far, but then I get into a conflict with my own morals of gay marriage being sinful according to the Bible.  Most of you probably don\'t know that in the United Methodist Church in the past few months, one of their pastors has openly admitted to being gay and having a lifetime partner.  They went to Portland and got married recently also.  Well, needless to say that opened up a HUGE problem within the church.  Not only the question of whether God thinks it\'s right to be gay or not, but also does the church want a gay person to be able to preach?  After a very drawn out trial (which is still being protested and debated) the Church decided that the pastor would still be allowed to preach.    Up until recently, I hadn\'t seriously thought about how much GM would effect and divide our country.  But just thinking about it from that one light opens up so many controveresies, and many more that are most likely to come.  It\'s such a vast discussion, and it can\'t be talked about with people\'s personal feelings and morals coming into play.  The madness of the world today...  Well, it is getting late and I should probably be going to bed.  I think that I\'ve rambled on enough to bore you all to tears.  See you in school tomorrow!    ~Kimber   '),(9807,1488330,'22,May,2004','2004-05-22 00:00:00','    I also thought the Life Skills Prom was a great experience. I thought that ASB did a great job at decorating and planning it. I think that we all made it a great night for them. I really got to see them have some fun and it just made me appreciate the little things in my life. Ha... I spent more time at that prom than I did at my own! Hope everyone had a nice weekend. Bye!   '),(9808,1488330,'22,May,2004','2004-05-22 00:00:00','    Clark, How dare you challenge me?  The morning after pill or emergency contraception either prevents an egg from being released, or creates an unhospitable environment in the uterus to prevent implantation the same way that birth control pills, shots, and patches work.  The defenition of pregnancy is the implantation of a fertilized egg into the uterus.  I repeat, the morning after pill DOES NOT abort a pregnancy and there are plenty of on-line resources to back that up.  The pill is useful for unplanned intercourse including rape or condom breakage.  Studies show no harmful side effects, but unfortunately, the FDA decided (against surgeon general recommendation) that emergency contraception not be sold over-the-counter because of the lack of specific directions for use.  With any luck, that problem will be remedied and EC will hit the shelves hell-bent on protecting against unwanted pregnancy.  Write to your congressmen to speed up the process.  Love, Rob   '),(9809,1488330,'22,May,2004','2004-05-22 00:00:00','    Good Morning Bloggers...I agree with Rob...the Life Skills prom was and is an awesome event.  It is the best dance of the year.  I hope more of you are able to participate next year...it gives you a new appreciation and insight on what\'s important in life.  All of your posts are interesting and insightful and Rob your section on GM was well thought out but I do want to challenge you on the day after pill...I THINK (I have not researched it.) you can take the pill up to three days after you have had sexual intercourse...and if your egg is fertilized it would terminate the pregnancy which would be an abortion.  There is also the pill you can take several weeks after pregnancy that many abortion clinics are offering to their patients.  I am not sure how these pills sit with Pro-Lifers.    I will see you Monday...I am going to try and rent Fog of War today and prepare something for class next week.  Clark   '),(9810,1488330,'22,May,2004','2004-05-22 00:00:00','    P.S.  Life Skils Prom rocked the mic right.  I\'d like to ask that you bloggers extend an invitation to the \"exceptional\" students of our school, you might be pleasantly suprised at the results.  P.P.S.  My post is my honost, albeit aggressive opinion.  I do not apologize for anything I\'ve said, but none of it is meant to be offensive.  If you have any questions, come to me personally and I\'ll explain them individually.  Love Rob   '),(9811,1488330,'22,May,2004','2004-05-22 00:00:00','    Kimber, I don\'t know what problems anyone but the Catholic church would have with the morning after pill.  I single out the Catholics because they are the only group I know of that has discouraged the use of birth control- which is all the morning after pill is.  It has the same chemical that a birth control pill does, and the same effect, just a double the dose of a weekly or monthly pill that interferes with the release, or implantation of an egg after unprotected sex.  It DOES NOT abort a pregnancy which is a common misconception.  The real problem with the morning after pill is that it is not an over-the-counter medication.  It is not marketed as a replacement for a condom or other contraceptive, but rather a substance that can be taken in an emergency.  Plain and simple, if the morning after pill becomes more easily accessable, abortion rates decrease.  Now on to the debate we had friday on gay marriage... why arguments against gay marriage (GM) don\'t work (many arguments against gay marriage overlap and I\'m going to refute all that I can recall, but possibly not in the most organizationally logical, but bear with me);  Civil Unions:  The fourteenth ammendment declares any law abridging rights illegal and Brown vs. Board (fifty years ago last week) declared separate unequal.  Allowing gays to marry does not detract at all from heterosexual marriages, I\'ll get to that later.  Religion:  Besides the obvious Church and State separation, I\'d like to challenge those who claim that their only opposition to GM is because God finds it immoral.  The only verses where the Bible expressly declares homosexuality sinful is in Leviticus, however, also in Leviticus there are laws forbidding mules, clothing made of more than one material, and the consumption of shellfish.  Even the most pious of midwestern farmers use mules, the devoutest of New York Irish-Catholics wear spandex, and the most faithful of Southern Baptists eat Louisiana oysters.  So why is the law forbidding homosexuality still relevant while so many Christians across the country would be burning pigeons, rams or bulls for atonement every day?  It is unfair to God, unfair to oneself, and unfair to homosexuals to be hypocritical in ones faith.  Lastly, there is no law forcing religious institutions to accept state-issued marriage certificates.  I know that I will not convince many to disregard a good portion of the Old Testament, but none should argue that universally legalizing GM infringes upon their religious beliefs.  Legalizing GM sets dangerous precedents.  AKA, the \"slippery slope\" issue: This argument suggests that there is something wrong with homosexuality in the first place.  That because heterosexuals can never comprehend its biological intricacies homosexuality is therefore immoral or wrong.  Some would argue that legalizing GM encourages homosexuality- to that I ask what is wrong with that?  Never mind the intended rhetoric of that question, the real contention is that homosexuality has been around since the beginning of time and will continue to stick around long after an ammendment to ban GM.  Others would say that gays don\'t hold the institution of marriage is the same esteem as heteros, but I would argue that they hold it higher.  Marriage is something that this group has to fight for.  A rite of maturity and consecration that has been to this point forbidden to them.  George W. Bush has proposed an ammendment banning GM as part of his Marriage Protection platform. I would ask the president to explain why, after two years as Governor of Texas, divorce rates in his states jumped up 6% after a steady decline over the last 25 years.  Gays can create the same environment of dependence and stability that the state provides benefits for.  Those who suggest that GM promotes infidelity, insincincerity, or various other ins promotes the very KKK/neo-Nazi idea that gays are inferior, as well as the perpetuation of the hate and discrimination that results.  Lastly, there is another argument that gay marriage opens the door to marriage to animals, or polygamy.  Marriage, ideally, is the manifestation of a commitment made in love.  Historically this has not been the case in heterosexuals when arranged marriages to expand fortunes and empires have been abused by straight royalty and peasantry as long as time remembers, and although that history has faded out, I see no reason as to why the commitment of two people in love has to be expanded to animals whom the law could successfully argue do not share the human emotion ideally involved in a marriage.  As for polygamy, I see no problem with it.  I have had first-hand accounts with personal friends of polygamists in Utah.  While the state doesn\'t recognize their marriage, the man of the house (so to speak) commits himself to each of his wives, supports his children, and the members of the family receive the benefit of an environment much healthier than that of an abusive or neglectful albeit \"normal\" household.  My conclusion to the arguments against GM is that in many cases they are the manifestation of an instinctual reflex against homosexuality.  I won\'t deny the fact that witnessing two males kiss does not sit well with me, but that is no reason to punish those, who for whatever reason, are physically attracted to the same sex.  If anyone wants to challenge my statements or add to my list of arguments to refute, I gladly accept the confrontation.  A metric plethora of people have claimed that GM is such a frivolous issue to get so passionate about, and on the first layer, I believe they are right.  But this argument is not just about GM; it\'s another pinnacle of the continuing struggle minorities have for equality in this America so proud of its Land of the Free.  Think about it.  Love Rob   '),(9812,1488330,'22,May,2004','2004-05-22 00:00:00','    Hello all you crazy blogging children!  To start, I want to say that the Life Skills prom was so much fun!  Everyone that was there, you know what I\'m talking about.  That was awesome.    I do have a question for all of the 2nd period, or 6th period kids.  On the issue of abortion, what do you all think of the \"Morning After\" pill?  I was talking to Bethany about it today, and I was just wondering if any of you had any imput on the issue.  Oh, and here\'s a random piece of info, for those of you who said that no one would want to adopt African Americans, or that the majority of the poor women who get abortions are black:  The amount of black women who get abortions is half of that of white women.  Black women actually have hugely decreased the amounts of abortions, and the reason behind it is because the African American Baptist Church has gotten so involved in enforcing abstinence. So..there\'s something to chew on for any of you all who had expressed that black women were having the most abortions in America.  I hope that you all are having a wonderful and fun weekend, see you all in school on Monday!    Kimber   '),(9813,1488330,'21,May,2004','2004-05-21 00:00:00','    Hey!  I thought that our debate in 2nd period went really well.  For someone in 6th period, did you guys break up into little groups, or did you all debate it in one big group?  We formed two groups, with one watching and the other debating, and it went pretty well.  But it was hard, because we all wanted to hear everyone\'s ideas, but with so many people and only 45 minutes, most of us only got to talk once.    The MUN kids in our class are INSANE, spouting off random facts that are somehow relevant to what we\'re discussing.  If only we were all that smart...but, I\'ve gotta be honest Thomas (and don\'t get heated), but I think that your issue on abortion in 3rd World Countries belongs in a completely different discussion.  Maybe I just don\'t know enough about Bush\'s policy, but I don\'t see how that effects the subject of our discussion on Friday.    And, for those of you who think that morals shouldn\'t be involved in deciding whether abortion should be legalized or not, I think that\'s impossible.  For example, I believe that no baby should be killed, but that if the mother has no way of taking care of the baby, that adoption is the best thing for it.  But someone else will feel differently, that the fetus isn\'t yet a person or that the mother has the right to kill or keep her baby.  That\'s a difference in morals and personal beliefs.  There\'s no way to take a stand on an issue like this without including your own morals.  Issues like abortion and gay marriage are built completely on personal opinions and cannot be decided in any other way.  PS  I\'m not anti-choice, I\'m anti-death! :)  See you all on Monday.  Bethany   '),(9814,1488330,'21,May,2004','2004-05-21 00:00:00','    WHOA WHOA WHOA......This blog is really heating up now....and that\'s awesome news  The debate today was impressive.....there are some talented 2nd perioders out there that really SHOULD JOIN THE DEBATE/MUN TEAM.....following Rob Muilenburge\'s great and courageous step.  For the abortion issue, I continue to stress the fact that whatever the case, the ultimate responsibility comes down to taking care of the child, whether it\'s parental care, adoptions, or fostercare. I dissagree with the fact that any living cell is somehow given the chance to become a genius and those wonderful dreams are innately possessed in the process. It\'s simply not true. It\'s simply a wild dream to think a living cell born to a perfectly dangerous and alien society would be expected to grow up healthily or EVEN SURVIVE THE FIRST FEW DAYS OF THEIR LIVES WHEN THEY STILL HAVE NOT OPPENED THEIR EYELIDS.........hhhmmmmmm maybe that\'s why infant mortality rates in third world countries are as high as 50%! Or half of the babies don\'t make it through the first two years of their lives. Hmmmmmmm how IRRESPONSIBLE are we then....how many more babies are killed and what\'s the strain on the society then. And that\'s not already taking into account the risk of lives the mothers have to go through bearing the child and raising the child up under extreme hostile situations.  I have to agree completely with how Bush\'s actions of punishing foreign country clincs that give out abortions would only hurt the destitute third world societies and does nothing to educate and prevent these kind of situations from happening again. The truth is, in the majority of 3rd world countries, and as a singlet example, SE Asia\'s Cambodia, sexual exploitations is extremely rampant (more rampant than anything you\'ve ever imagined that could happen). Such poor countries. People younger than us would be SOLD BY THEIR PARENTS for prostitutes because the families are simply sooooo poor. And for us it is easy to \"hey you have to understand the consequence of sex\" and that, but many destitute people in third world countries PEOPLE SIMPLY DON\"T KNOW AND HAVE NEVER HAD THE EDUCATION FOR THAT. EDUCATION and the corresponding higher level of social-economic status is the KEY for woman to make wise decisions. Education is the key for elevating the traditional degrading and subhuman status of woman in many third world countries. Thus, instead of withdrawing aids and funds from foreign clinics where abortions are performed, it\'s an opportunity for us to show our egalitarian American values and try to put more aids into education programs, and more aids so more people can lift out of poverty situations.   Yeah an thus, I think that abortion should be upheld simply looking at the world with reality......And of course if you are educated and wise and know how to make the choice, GOOD FOR YOU (you are extremely lucky and should be proud of yourself of not ever having the need for abortion), but for millions of others around the country and around the world.......There is a long long long way to go to create a perfect utopian society......   As for the childhood obesity debate, we\'ll see on Monday.  I also want to comment on the gay marriage issue here. The compromise for both sides would be civil union protections for at least basic liberties so gay people are not constantly in a state of fear and constantly being prejudiced. And ultimately, I believe that they should enjoy the exact same liberties and opportunities any other person enjoys. Because ultimately, people CAN biologically be gay. To deny gay people the same rights all other citizens enjoy would be like degrading disabled, mentally retarded, or even a minority race of people.   '),(9815,1488330,'20,May,2004','2004-05-20 00:00:00','    Hey, what\'s up?  I thought the discussions in class went pretty well.  I just had to post my thoughts on the whole same sex marriage issue.  I agree with Megan, it is a person\'s own beliefs but religion is part of a person\'s beliefs.  My thoughts are that same sex marriages are morally wrong and they shouldn\'t be allowed.  But on the other hand, as Ryan Elliot said today, marriage is just another benefit for these people.  Not allowing it isn\'t going to do anything to the gay population.  They are still going to have the same relationships and everything.  I guess that it always will turn into a heated conversation because of so many different beliefs.  Anyway, that was just a thought I had after class. Keep it real!   '),(9816,1488330,'20,May,2004','2004-05-20 00:00:00','    I thought todays debate went really well, and like a lot of people I\'m against abortion but believe in pro choice. I think the government would have a lot of problems trying to make abortion illegal again. I believe many would see it as an erosion of women\'s rights. Just this Sunday a huge rally was held in Washington with over 800,000 people marching for pro choice. Like Sarah said, I also think that religion should not have a place in this issue. This issue is an all American issue. An individual\'s moral beliefs whether it is based on religion, family, etc, is not significant to everyone. Then you can get into ethnics and if the fetus is an actual baby and things just get confusing. How I see it, the law is already standing, and making abortion illegal is an infringment on women\'s rights to choose. Is abortion wrong...well I think it is. But what about freedom... the foundation of American identity.   I\'d like to be honest and say that I am worried if 2nd period debates about gay marriages. This issue is extremely personal and emotional for me.   '),(9817,1488330,'20,May,2004','2004-05-20 00:00:00','    My little dumplings are growing up...I think it is so interesting to watch adolescents begin to formulate and defend their opinions on such controversial issues.  I was watching many of you in second period with the personal struggle of Abortion being morally wrong yet unwilling to take a strong stand against others...hopefully you will continue to grow and think on these issues as you further develop.  Some of you will become more conservative in your thinking as you experience the \"real\" world while others may become more liberal.  Trying to figure out what you believe in and what you stand for is so important in figuring out who you are...and whatever that is I respect all of your ideas and opinions to the upmost.  Commentary...the second period debate was impressive...I was very impressed by your sensitivity and the resepect towards others...will you have the same resepectful tone when we discuss child obesity?  (I have to laugh because this topic has been hot on the news and NPR lately.)  Sixth period...the Pat Tillman discussion was interesting.  I am unsure how I feel or if I feel anything to be completely honest.  However, he will be the \"poster\" child for this war and remembered in many history books.  He is also an excellent example of the sense of Nationalism produced by the 911 World Trade Center attack.  I appreciate all of your struggle with the value of life...and yes I think we can all agree upon that everyone\'s life is valuable.  I think the discussion also brings up other valuable questions?  What is the role of the Media?  And who is really in charge of what we see and hear by our news media?  Who decides? Who pays for it? AND who gets the profit?  Obviously these people (dynasties) have a lot of power and influence over our nation?  The gay marriage debate was a little heated.  I think Ben may be right that is an issure that might divide our nation.  I like how Kalie related it to the Civil Rights Act - an example of a good history student.  But should our government be involved in a personal institution.  They set age limits? Should they?  Maybe we could lower the divorce rate by increasing the age and wait time before people get married...how about 25 and you must \"date\" for two years?   That might make our families more stable...their are so many other issues...maybe our goverment should control who we marry to make sure it lasts longer?  Maybe people weren\'t meant to be married...of course I am just thinking out loud...(Playing the Devil\'s advocate in most cases...)  I must get ready to boogie down at the Life Skills Prom.  Have a good weekend and I look forward to hearing about more of your ideas...  Ms. Clark   '),(9818,1488330,'20,May,2004','2004-05-20 00:00:00','    hey guys its my first time on THA BLOG (im kinda nervous...) ne ways i just wanted to follow up on our debate today in second period about abortion and if the government should have a say in it. well... i dont think they should. I know a lot of people dont believe in abortion but that is their opinion and there may be some people who think differently.  us and especially the government dont have any right to tell those people how to live their lives. Mar i got your back on this one!! im against abortion but not against human rights.  bie bie   '),(9819,1488330,'20,May,2004','2004-05-20 00:00:00','    I wasnt able to be in class today, as Ms. Clark already knows, but i do want to participate in the \"discussion\". On the issue of same sex marriage, I support the legality of it. Whether or not i agree with it morally has nothing to do with my decision. My decision is based on the constitutionality: our country takes pride in the freedoms we are guranteed as American citizens, many of which we take advantage of daily. However, with the Civil Rights Act of 1960, it is unconstitutional to descriminate based on race, national origin, or SEX (which for me, includes sexual preference). How can we stand by our constitution and the freedoms it gives us if it is not used to give these freedoms equally to ALL citizens? Along with that, even since the first colonists, separation from church and state has been established. Thus, even if it is morally wrong based on your religious creed, it is not justification for its illegality. I understand both sides. I see where religious arguments can come into play and how morals effect our ways of thinking, but we must see past that...we must see is it constitutional? And does it provide equal freedom to all U.S. citizens? ANALOGY  On the issue of Pat Tillman...i think i\'m a little more conservative. Yes, he did a great thing, and he SHOULD be remembered because of his bravery and giving up everything he has here at home. But I think it is insulting to all those other soldiers fighting for the same reasons, but getting no recognition just because they arent pro-football players or millionaires...they also are giving up a great deal to be fighting overseas. They\'re giving up their families, their children, if they\'re in the reserves, their jobs (even though theyre not highly public jobs)...theyre giving up just as much as he is, yet are quickly forgotten. How is this fair? What are we as Americans supposed to tell the other soldiers\' families when they die in action? they werent famous? they arent as brave as Pat Tillman was? NO! I DO think it is important that we remember him. Maybe we should name the stadium after him. But i think to be fair and equal, we must also give recognition to the other men fighting overseas. The issue of Pat TIllman is like group projects. Everyone does (or SHOULD do)their own part, working hard and diligently, but then when the time comes to present, the speaker of the group gets all the recognition and everyone else\'s hard work is quickly dismissed. That should not be the case here. Pat Tillman, although a well-known hero, was part of a team. He did his part, along the sides of many other brave men. When the time comes, which it has, shouldn\'t all the men get recognized for their hard work?   Well, that\'s my little argument that i WOULD have discussed in the discussion had i been there. Dont fret, i WILL be at school on Monday!  Kailie   ');
INSERT INTO `msgs` VALUES (9820,1488330,'20,May,2004','2004-05-20 00:00:00','    K, well here is what I have to say about the abortion debate that we had in class today...  There are so many factors that contribute into abortion besides just the idea of \"killing an unborn baby\". There are the ins and outs of foster care, adoption, women\'s lives, children\'s live as well as the state that our country is in.  I am religious, though I believe that religion can have no place in this issue. It is fair that everyone can have their own morals, beliefs and religious system, but what about the whole idea of separation of church? Just because \"the bible says so\" doesn\'t mean that is what our country should go by. Do you see what I mean?   I have a lot more to say on this issue, though I know there is really no way of winning the argument. Is there really a happy-medium? Because I mean if abortions were to be allowed for only hazardous pregnancies, rape and incest cases, where and how excactly do you draw the line? If it\'s okay for some women, why not for all? And in the case of pregnancies that are harmful for the pregnant woman, perhaps it is \"God\'s will\" that it ends that way? I don\'t really believe that, I\'m just throwing it out there.   Anyway, I would never have an abortion myself, but I wouldn\'t want to take that right away from any woman who wanted to use it. Afterall, if abortion is against yours or whomever\'s religion, they will  have to take it up with Him in the end. We\'re not God. That statement can be taken another way: we can\'t take away life...but if we can make it, why CAN\'T we take it away? And just because someone doesn\'t support something or even think it is right, it doesn\'t mean they should force that opinion and ways of life upon other people. That\'s all for now.   Peace.   '),(9821,1488330,'20,May,2004','2004-05-20 00:00:00','    My comments about the abortion portion of the debate is this... I\'m not pro-abortion, but I am pro-choice. Can I be that way? I don\'t feel that abortion is a legitimate way of escaping the problems of unexpected pregnancy. The woman should have thought about the consequences of potential pregnancy before she decided to have sex. I know I know, but what about the women who get raped. That solution would be solved by adoption. I have heard from some people that adoption isn\'t a sucessful option. What I have to say to that is where\'s your proof? With the process of adoption, the baby, at birth, is signed over to a family or a couple. It\'s not like the baby is being handed over to the state. It\'s the state that has the problems of foster care, or the problems people may think of when they hear of adoption as a solution. But anyways, should the government keep abortion legalized? I\'m against abortion, but the to tell you the truth, the government should have no say as to what women do with their bodies. Abortion is a moral problem, not a legal problem. To say something else about abortion, especially the clinics. I wish that abortion clinics would give more information about the procedure. Abortion is not simply taking a piece of tissue out of a woman\'s body. Abortion is taking a life away, and no one can argue that it\'s not.   '),(9822,1488330,'19,May,2004','2004-05-19 00:00:00','    Hey Tom Chen, thanks for the invite!  Sooooooooooo.   bye   '),(9823,1488330,'19,May,2004','2004-05-19 00:00:00','    Ok I\'ve emailed to a bunch of people....A lot of sixth period folks....Hopefully all ya who wanted to blog got an invitation.....If not, talk to Rob, Clark, or me.   And yeah, our 2nd period debate topics: 1: Abortion - age old controversial topic, should US overturn Roe vs Wade 2: Childhood Obesity - Should the federal government play a significant role in monitoring the health of our children.   '),(9824,1488330,'19,May,2004','2004-05-19 00:00:00','    Does anyone know what 2nd period\'s debate topic is? Sorry, I wasn\'t here when we decided and I\'d like to do a little research so I can paticipate.  Thanks!   '),(9825,1488330,'19,May,2004','2004-05-19 00:00:00','    Hey kids! Its my first time on the blog! Aahhh! Thanks for the invite Thomas. Anywho...just wanted to make my very first statement...exciting...and now I have to get ready for discussion tomorrow (a lil unprepared) see you all tomorrow! ~Gwen   '),(9826,1488330,'19,May,2004','2004-05-19 00:00:00','    HIIII!! QUESTION what is the debate on 2morow i forget.... did we pick the topics? i know we talked about them but i wasnt sure if we narrowed in on any specific ones.. its been a long week! THANKs jessss   '),(9827,1488330,'19,May,2004','2004-05-19 00:00:00','    I feel so sorry that I won\'t be able to make the discussion about abortion tomarrow... I\'m going to be in Tacoma for Regionals for Tennis...I know, Boys tennis is in the fall, girls is in the spring, but boys regionals is held in the spring because...well...just because. So GOOD LUCK TO ME! at regionals, and I guess I\'ll see everyone on monday. ~CT   '),(9828,1488330,'19,May,2004','2004-05-19 00:00:00','    Hiiiiiiii I am blogging for the first time, it\'s exciting.   '),(9829,1488330,'19,May,2004','2004-05-19 00:00:00','    The following students want to BLOG..please invite:  Jake Flake-jacob_flake@hotmail.com Jon Gorham-gorhams@hotmail.com Nikki Hinton- Nikolie43087@aol.com Ashley Ferguson - AshFerguson@hotmail.com   '),(9830,1488330,'19,May,2004','2004-05-19 00:00:00','    hey Hey HEY! another innocent blogger is trapped within this tangled web of bloggers that Thomas has started... Anyway... Once again, i get frist post of the day... I wish blogger had options for smiley faces... Oh well... maybe i\'ll complain to the owners....   '),(9831,1488330,'19,May,2004','2004-05-19 00:00:00','    Hi!!!  It\'s Nova.  First time I\'ve posted on the blog.  So how is everyone?  My schoolwork is cut in half since the tests.  I just came here to say hello, not much to talk about... Bye!   '),(9832,1488330,'18,May,2004','2004-05-18 00:00:00','    Wow... the year is coming to a close... well... of Running Start people... The last day of classes for Spring Quarter is like June 4th i think.... then the next week is for finals... Anyway... the year is still coming to an end... does anyone have any summer plans that they want to share? I\'m just bored right now... doing some homework.....   '),(9833,1488330,'18,May,2004','2004-05-18 00:00:00','    Ok as Ms.Clark was saying how the blog is beginning to heat up.....it\'s time to review all the current members of the blog......And those still who are still pending....  Current members: Mr.Ray, Al Tu, Aly Eastman, Vanessa Esch, Chris Storms, Kailie Middleton, Frank Kanekoa Jessica Lee, Joseph Liberman, Hiliary Philip, Megan Valencia, Mary Tyner, Ben Perry, Rob Muilenburge, Sarah Nosler, Jessica Knab, Jessica Kolkmeyer, Mat Keksi, Julie Olds, Tristan McDaniel For the names listed above, all of you have accepted the blogger invitation and have signed up for a username and password. If you ever forget your username, just contact me.   Current members waiting to be signed up: Corey Taylor, Darla Haynes, Nova Martin, Natalya Skiba, Conrad Salvato, Jennifer Wellsmore, Devin Cusack, Stephanie Bauman, Adam Stevens, Ryan Elliot, Gwen Cole, Alex McDougall, Paul Fleurdelys, Bethany Graef, Kimber Graef For the names listed here, I have emailed the invitation to all of you. The invitation maybe many months old...in any case, if you have never gotten the email or if you have accidentally deleted the email, please contact me and I will resend you the invitation.  If it\'s the case that the blogger is screwy and you can never sign up to the invitations I sent you, you can contact a friend and get their username and password and share an account.   Oh yeah for all of you who have not pronounced yourself in the APUSH blog (also for those new blogger members who have just signed up) PLEEEEEEAAAAAAASE DO SO. Inaugurate the blogging system by just saying \"HI MY AP FRIENDS\" on the blog.   Thanks I can be reached at AIM: C5hanst123 or email apturtle2003@yahoo.com   '),(9834,1488330,'17,May,2004','2004-05-17 00:00:00','    Ahh... there are some good books out there.. I\'m probably going to read The Picture of Dorian Gray.. It\'s a double part reason.. its the book for History, and one of the scholarships requires me to read it... and if i get bored with this book, i\'ll read Atlas Shrugged....   '),(9835,1488330,'17,May,2004','2004-05-17 00:00:00','    this is rob posting at school, demonstrating to ashley ferguson how it\'s done so she can join for the last month of school.  by the way, i\'m about to fall asleep, but i\'ll dream about everyone of you bloggers havin a great day.  love rob   '),(9836,1488330,'17,May,2004','2004-05-17 00:00:00','    This is my first time posting in here. So hey. I don\'t really have anything to say except, does anyone have any good ideas for the book we\'re supposed to choose to read for this class? I think I\'m gonna go to the library next period and look for one.  Peace. -Sarah Nosler   '),(9837,1488330,'17,May,2004','2004-05-17 00:00:00','    Hello out there~ I was just wondering if anyone had any suggestions for our book assignment. I have tried to find a book that is popular, and I would just like to find a book that is peer-approved. So any suggestions would help. Thanks, ~hillary   '),(9838,1488330,'16,May,2004','2004-05-16 00:00:00','    Ahhh... nothing like a good cup of coffee and the first post of the day on the blog.... Good to see you back Thomas... you didn\'t miss too much... we watched Forrest Gump. (is that one \'r\' or two?). that alone took three days. Then we got a couple assignments.... that\'s about it...   '),(9839,1488330,'16,May,2004','2004-05-16 00:00:00','    HELLO MS.CLARK AND ALL MY GREAT FRIENDS FROM APUSH!!!!!!!!!!!!!!  IT\'S BEEN SO LONG........I\'m finally back from 130 hours of imprisonment in the Oregon Convention Center for the International Science and Engineering Fair.  Yeah the most common phrase over the last week for me was \"Where are you from\", gosh I\'ve had to said it over 500 times............And from that I\'ve met sooooooooooooo many people......from all 50 states and 38 countries around the world!!!!!!!!!!!! And the most valuable experience of this whole week was making such valuable friends.........And have the opportunity to meet future leaders, future billionaires, future Nobel Laureates, not to say that we had the chance to actually talk to the current Nobel Laureates.  And overall, there was TONS AND TONS OF FUN!!!!!! ENDLESS FUN....PARTYING EVERY NIGHT FROM SUNDAY NIGHT TO THURSDAY NIGHT...........More social activities the past week than I ever experienced in my life.  And tomorrow I\'ll have lots of fun tomorrow catching up with all the work I may have missed. And catch up on the current status of the impending party for our great APUSH teacher Ms.Clark.   And finally, I missed all of you the past week, hopefully you all had a fun week!  See ya all tomorrow!!!!!!   '),(9840,1488330,'15,May,2004','2004-05-15 00:00:00','    Sounds like Rob is having fun plugging MUN.... sixth is so lucky... we didnt get to do frisbee..... unless this was an after school thing i missed. i\'d bring a frisbee... but the only one we have here is the dogs\' chew toy... its not in the best of conditions...    I\'m aware of my excessive use of periods..... i do not care....   '),(9841,1488330,'15,May,2004','2004-05-15 00:00:00','    APUSH Bloggers... I hope you realize that next time in ultimate frisbee the teams will be fair and you will all be subjected to the merciless wrath of he who is known as Rob.  Props to Frank for his beautiful catch, as well as disabling one of my best team mates, you are a worthy competitor.  I\'m in support of JKo throwin a bitchin going-away party for Clark, and maybe next year we\'ll orginize a road trip to kick it with her in sea-town.  For any of you who enjoy the debates we sometimes have in class, I recommend you give Model United Nations a try.  I did, really enjoyed, and was fairly successful in my first competition.  Debate and compromise are great skills to develop, and MUN is an excellent medium for that.  Props to the Russian Federation on their team victory, and good job to everyone else that participated as well.  Love Rob   '),(9842,1488330,'15,May,2004','2004-05-15 00:00:00','    hey!! I havent blogged in awhile! ultimate frisbee was fun! to bad i wore pink stilleto boots next time i better chose better footwear! but yeas CONGRATS on the job clark we\'ll throw a heck of a goin away party for YOU! Just leave that plannin up 2 me! hehe(evil APer laugh) juust joking but anyways i agree with every1 i felt way more prepared for AP history test than any of my other ones.. excellent job you are a good multitasker, with the whole principal thing, interviewing, going to school teaching us WOOO Oprah isn\'t kidding girls can really handle more than one thing at a time! well im off to portland for a lil shopping excursion with aly! hav a great saturday everyone!   '),(9843,1488330,'13,May,2004','2004-05-13 00:00:00','    Hey...APERS...Does anyone have a frisbee for tomorrow?  Please bring it if you do?  Also 6th period may be in for a treat...my sister (THE LEGEND) is coming into town probably around 6th period.  Who is a very competitive player?  The Clark girls will DOMINATE!  (Trash talking on the Blog...a first for me.)  Miss Hillary thank you for the kind words and it has been a tough year and I don\'t think I was at my best but hopefully we all learned something.  And I have to say you are definately one of the most talented and diligent classes I have ever had. So make sure you enjoy each other.  You are lucky to have a group of students who enjoy learning and thinking.  I expect great things from you in your Senior year and would love for you all to keep in touch!      Thomas I hope you are having a good time in Portland.  Make sure you get all the assignments that are due tomorrow.  (AP HUMOR)  See you tomorrow...let\'s have a great finish...(My internship is all over but the shouting...)  Bring on the farewell tour.  CLARK   '),(9844,1488330,'12,May,2004','2004-05-12 00:00:00','    Well I would just like to thank Ms. Clark for being such an awesome teacher. Don\'t tell Mr. Emerson, but I felt that I was more prepared to take the AP history test this year than I was last year (but that could have been that I never really read the book last year)! I can honestly say that I feel bad for the sophomores who want to take AP US History next year, because they won\'t get Ms. Clark for a teacher. NOT to suck up or anything but I am really proud of you Ms. Clark because I am sure that this hasn\'t been your easiest year with the administration junk and all. But I am sure that you are happy that all your work has paid off and now you have a great job! So hooray! Plus now you can give us a bigger perspective on the journey of life! Hehe how is that for cheesy!   '),(9845,1488330,'12,May,2004','2004-05-12 00:00:00','    Hmm... Prom... I don\'t know how it was... i wasn\'t there... but I hear that it was fun.......   '),(9846,1488330,'10,May,2004','2004-05-10 00:00:00','    APUSH bloggers... Although Forest Gump has a different, hilarious speech pattern, there is also a great deal of 50\'s and onward pop culture and history incorporated, and although the apex of the class is already behind us, viewing FG historically is an alternative and equally entertaining method of watching enjoying it.  Secondly, I want everyone\'s opinion on prom... the good, the bad, the ugly- and I don\'t mean Kolkmeyer\'s dress (JK Kolk, you looked hot and I love ya).  Would you want it at Union Station again?  Any feedback you can provide will help make next year\'s prom as ridiculously perfect as possible.  Congrats to Clark on her new job, but we\'ll miss ya next year.  Love Rob   '),(9847,1488330,'07,May,2004','2004-05-07 00:00:00','    Hey Ms.Clark again, yeah thanks for the moral support.............Hmmmmm it\'s time to reflect.   Actually here I want to say thank you very much with my deepest appreciation for your great care and effort you put in as a teacher and administrator to support and help the Skyview students, and your efforts to make Skyview a better place.  Yeah and since this is the end of the AP test and, I have to say that this year has been a great experience in your AP US history class. I have learned and grown substantially both as an academic person and a contributing member of the Skyview community, thanks to the support of you and all of my great peers.  I know you will be leaving us next year, and I wish you the best luck as a vice principal and I\'m sure all of my peers and I will miss you very much.    Yeah anyways Ms.Clark, you have been a GREAT AP US HISTORY TEACHER AND A GREAT ASB ADMINISTRATOR! I\'ll certainly miss you very much next year as I continue my journey through the Skyview storm.  And I\'ll come up and visit you sometimes.   '),(9848,1488330,'06,May,2004','2004-05-06 00:00:00','    Hey my little dumplings...I am sure you did great on the test.  Take the Multiple Choice in perspecitive.  I am convinced if Tristan and Thomas thought it was hard...it was...and the test score is based on averages.    After hearing what the essay prompts were...I am convinced you did great.  In any event...you are all amazing people and have grown as readers, writers, and students.  I hope you will some day appreciate all of your efforts.  I can\'t wait to finish the last month of school with you...let\'s stay positive and have a good ending to a great year.  You truly are an amazing group of school.  Clark   '),(9849,1488330,'06,May,2004','2004-05-06 00:00:00','    OH wow... RELIEF.  I was excited to take this test today because I was prepared!  History, unlike English, is a subject with definitive answers... it\'s an engine that needs all its parts to run correctly, and this morning I was purring like a ferrari on steroids!  Most of my answers I knew, atleast 60 of them I was able to narrow it down to two, and there were only three I wasn\'t able to answer at all!  I was soooo pleased, and the essays?  Right up my alley.  Don\'t take this the wrong way, I don\'t WANT to gloat, but I feel like I\'m OBLIGATED to I feel so good about myself right now.  I\'m sure everyone else did good too, because if you payed attention in class the last couple weeks, and reviewed a little bit then you had the tools for a 4 or 5.  If you want to get pissed at me for this self-glorifying post, you have that right... but make me feel good and praise me for my temporary genious!  Love, Rob   '),(9850,1488330,'06,May,2004','2004-05-06 00:00:00','    Well, this test is over.......And which turmoil comes next??  Heh today\'s test (the multiple choice section) I completely threw up on............  Well first of all, the test commenced in an environment in which Zenson mistakenly said \"This test consists of ONE HUNDRED AND FIFTEEN MULTIPLE CHOICE QUESTIONS AND SIX SHORT ANSWERS IN 55 MINUTES!\"  And that just sent the room to completely dead silent with everyone stopped breathing!!!!!!!!!!!  And then the multiple choice section:  Soooooooooooooo HARD............SOOOOOOOO MUCH HARDER THAN ALL AND EVERY SINGLE PRACTICE MULTIPLE CHOICE TEST WE\"VE EVER TAKEN!!!!!!!! Frank, Ryan, Tristan and many others would certainly agree with me. WHAT THE HECK?????????? About 80% of the questions were a wobble between two answers..........GOOD GOD I DID ABSOLUTELY HORRIBLE!!!!!!!!!!!!!!!!!! Seriously, SO SMALL PORTIONS OF THE QUESTIONS HAD A DEFINITIVE ANSWER!!!!!!!!!!!! NO MATTER HOW MUCH HISTORICAL KNOWLEDGE YOU KNEW!!!!!!!!!!!!!!! 80% of the questions to anyone is just completely debatable.  Anyways, that\'s my extreme rant of the day.........I\'d really like to hear others\' opinions and afterthoughts about today\'s test. And oh yeah, have fun everybody the next week, I\'ll see you guys on May 17th.   '),(9851,1488330,'06,May,2004','2004-05-06 00:00:00','    WOOOOOOO!!!!!!!!!!!! YAY!!!!!!!!!! That test is done and over with! I hope everyone  did well.....   '),(9852,1488330,'06,May,2004','2004-05-06 00:00:00','    good luck everybody, though I doubt the likelyhood of someone actually being able to read this at six something in the morning.......   '),(9853,1488330,'06,May,2004','2004-05-06 00:00:00','    Here\'s the mandatory good luck and Happy Review!  Be safe for prom!  DUI is never EVER the best option.  Love, Rob   '),(9854,1488330,'05,May,2004','2004-05-05 00:00:00','    Ahhhhhh The BIG TEST has almost arrived, is everyone ready?   '),(9855,1488330,'03,May,2004','2004-05-03 00:00:00','    Mrs. Clark I missed you too. The AP tests were so fun.  ~Frank   '),(9856,1488330,'02,May,2004','2004-05-02 00:00:00','    Anyways, Rob, when I mentioned to Matt \"people like you\", I was specifically referring to those people taking classes at Clark, and thus couldn\'t come to Skyview for the right classes because of the screwed up schedules.  But seriously good GOD lord, contrary to most people, to me, I actually have MORE stress now after taking this one AP test. Holy lord.....four more to go, four more battles to be fought.  Yeah I\'ll probably come to Skyview tomorrow after school review session just for the heck of getting in touch with society and conversing my mental state of stress with others (hopefully I don\'t contaminate other people with my stressed out minds).  All I have to say is that hopefully the History test goes better than today\'s kick-off AP test.   '),(9857,1488330,'02,May,2004','2004-05-02 00:00:00','    Hey all you APers...I missed you today.  I walked around not knowing what to do with myself.  I hope you did well on the English test.    Tomorrow during class we will review post Civil War to ?????  After school we will work on AP prompts and stuff...make sure you are taking care of yourself.  Extra Credit Review Packets are due on Thursday - we will wait and have the AP celebration after all tests are done.    See you soon...   '),(9858,1488330,'02,May,2004','2004-05-02 00:00:00','    Haha, Tom, I don\'t know what you mean by \"people like you,\" but it made me laugh.  AP English really didn\'t bother me... I thought the multiple choice was on the easy side, answered all but four or five.  As for the essays, one will be good, the other two are a dice roll, there\'s no organization in so little time!!!  Anyway, the point is, one down, two to go...are we reviewing after school tomorrow?  Wanting to know if I should come.  Love, Rob   '),(9859,1488330,'02,May,2004','2004-05-02 00:00:00','    yeah.... anyway... how was the english test everyone? there were only four of us brave enough to come to class today.....   '),(9860,1488330,'02,May,2004','2004-05-02 00:00:00','    Oh poor Matt.....Don\'t worry......Interview has absolutely nothing to do with what we are studying. I just had to bring it up to confuse you. (JUST KIDDING).  Anyways we are just review right now, in class and after school and on weekends.  I have to agree that WASL schedule did screw everyone over, especially people like you.   '),(9861,1488330,'01,May,2004','2004-05-01 00:00:00','    Interview? what? I DUNNO WHAT\'S GOING ON! I\'m going to be so lost on this test.... i don\'t know what i\'m going to do! this WASL schedule has screwed me over!   '),(9862,1488330,'01,May,2004','2004-05-01 00:00:00','    What is wrong with hard working - I know many students/adults that would prefer not to be labeled hard working.  I also think it is the students or people who aren\'t afraid to work hard usually find the most success in life.  Depending, of course, how you measure success.  The interview - I SWEATED and STAMMERED and probably STUTTERED a few times.  I was interviewed by 15 different people.  It was a little different.  We will see if I get called into for a second.    This morning on NPR they did a nice show on Alexander Hamilton...they played up on his illegitimacy and being an orphan.  There is a new biography out on the man.   Are we ready to review Jackson to Mckinnely?  I am drinking coffee and taking notes as we speak.  For those of you who are unable to attend the APUSH review - Good Luck on the English Test...See you Tuesday.  How did people do on the Multiple Choice section?  Clark   '),(9863,1488330,'01,May,2004','2004-05-01 00:00:00','    Here is an EXCELLENT article from \"KidSource Online\"  urlLink that provides insights into students\' attitudes in schools with survey statistics and some common misconceptions. All students and especially teachers should take the time to read this article .   This statement below from that article is VERY true:    Popular opinion has it that students\' academic success depends on the quality of their teachers and textbooks. Ask the students themselves, however, and you get a different view. Here is how they account for their academic achievement:   Most students believe their ability and effort are the main reasons for school achievement. By the same token, if asked whether they would prefer to be called smart or hard-working, they will choose smart almost every time. Why? Because they believe that hard-working students risk being considered either excessively ambitious or of limited ability, both of which they would find embarrassing.    And by the way, how did that interview go, Ms.Clark?   '),(9864,1488330,'23,June,2004','2004-06-23 00:00:00','    Hey everyone........I guess today is Skyview\'s last day of school. I\'m posting from the UW communications building computer lab.......UW is so nice! I am living in the infamous McMahon dorm. And our SIMUW group of 24 students is only a small fraction of ALL the countless number of high school/middle school students who are here for the summer living in McMahon. A lot of people are here for summer sports camps.   The McMahon dorm has a fantastic view of lake washington. The dorms are set up as a \"quad\", IE four dorm rooms in one vicinity sharing one living space/microwave/refrigerator/bathroom, and the best part, a balcony. The McMahon dorms is so oppressively hot (seems like the heat is always on even in 90 degree weather). The big windows are open with no wind coming in and yet the flies and bugs outside are attracted to the light inside. The first two nights the dorm is around 90 degrees and each room has one little fan to cool us off.......  As for the food, it\'s 1000 times better than anything we have in Skyview or nearby Salmon Creek. Each student gets $24.50 daily for food (all free).......Lots of stuff to gorge on........And some crazy SIMUW people buy dozen cartons of milks or drinks and freezes it overnight......crazy Ha one downside of the \"quad\" dorm is that eight people have to share one bathroom/one shower. So things can get interesting with shower/bathroom schedules.  The whole day spent here in UW is packed with stuff to do..........From 7:50 to 8:50 breakfast/community meetings.....From 9:00 to 12:00 morning lecture/activity.....1:30 to 4:00 afternoon lecture/activity.....4:15 to 5:00 computer alloted time.....5:30 to 6:30 dinner time, and afterwards free time to do whatever we want.  During the evening times we wander around the campus/dorms/travel to university avenue or university village shopping areas/movies/community solving math problems/or playing boardgames like monopoly or risk.   Yeah and finally the people here are so interesting.....just leave it at that.......Half of the people in this program are Canadians!!!!!!!!!! Ha............So interesting.   '),(9865,1488330,'21,June,2004','2004-06-21 00:00:00','    My little dumplings...thank you so much for the nice send off.  I am so sorry I added to your anxiety and stress by not showing up on Thursday...Whoops!  I am very touched by your kind words and actions. I have the poster hanging in my condo amongst the boxes. I can honestly say I have never come across a more  caring and intelligent group of students.  I wish you the best of success in your senior year and your life.  The hardest part of this transition is saying goodbye to the students.  (I have teared more than once thinking about it.) You are truly amazing.  Please keep me posted on your life, college choices, etc.   You can reach me in July at sjclark68@hotmail.com or use the Shorecrest website...I will have an email account set up there soon.  I start work on Monday.  (Yikes)  I would love to hear how you did on the AP test...remember it is a tough test and your score may not always reflect what you have learned or know.  It is one test.    Last night I was watching 60 minutes...did anyone else see the Clinton interview?  What did you think...do you hate him or love him...how do you feel about the right wing conspiracy? Will he be able to live down the impeachment trial...was the impeachment fair?  It is odd...the last couple of weeks have focused on celebrating the Republican agenda in remembering Ronald Reagan...now Clinton\'s book and Farehneit 911 is coming out...there should be some interesting debates and discussions to blog about.  I do like what Clinton said about hate...there is no reason to hate a political figure but be able to disagree with their stance and opinion.   Thomas...I hope you are having a good time in Seattle.    Keep on blogging...I would also love to hear about the trips to Germany, China, and Georgetown...  Love ya, Ms. Clark,   '),(9866,1488330,'19,June,2004','2004-06-19 00:00:00','    Well, I woke up and realized this college project is due next week, so I decided I had better get started.  Instead of using class time to do work I played computer games and now on a beautiful day like today I\'m trapped inside until I finish my project.  Just my luck.  Have a great weekend and don\'t forget Father\'s Day is tomorrow.   '),(9867,1488330,'18,June,2004','2004-06-18 00:00:00','    Wow, I just started blogging not to long ago, but I think it\'s fun.  To bad the year is almost over.  This college project we have to do in clark\'s is actually going to be very useful.  Well, if I follow it anyways.  I think I will though, I\'ve already applied for scholarships that I would have never done if not for this project, and I Think I\'m going to make it my summer goal to apply for at least 10 a month.  Get what I can out of scholarship money.  I also like the idea of applying to every school that is free to apply to across the country, just to make them all compete over me.  Anyways, I\'m going ot bed.  I Need the sleep. ~CT   '),(9868,1488330,'18,June,2004','2004-06-18 00:00:00','    THANKS SOOOOOOOOO MUCH FOR ALL OF YOU WHO SHOWED UP TO CLARK\'S \"SURPRISE\" FAREWELL PARTY, 2ND EDITION!!!!!!!!!  WOW!!!!!!!!!!!!!!!!!!!!!! I STARTED PLANNING FOR MS.CLARK\'S PARTY SINCE LAST MONDAY!!!!!! JUNE 7th!!!!!!!!!!!   Wow.......Thanks A MILLION TIMES TO Sydney Clos, Chris Tamura, Bethany Graef, Rob Muilenburge, Vanessa Esch for the SOOOOOOOOO MUCH COOPERATION FOR THE PARTY PLANNING!!!!!!!  Here is the breakdown: Monday June 7th: Thomas Chen somehow IMS Bethany Graef about an AP PARTY thing......And Thomas Chen\'s corrupted brain settles on the idea of an Surprise party......and sooooooooooooooo all begins there, with an incompetent wired brain of Thomas Chen\'s.  Tuesday June 8th: Thomas Chen finds Sydney Clos as a sixth period coordinator  Wednesday June 9th: Thomas Chen talks to Ms.VA about this party thing.....VA completely cooperates......And thought of a plot: Schedule Clark a \"curriculum meeting\" for next Thursday(June 17th)  Chris Tamura agreed to do the flier, Sydney Clos the T-Shirt, and Bethany Graef the card  Thursday June 9th: Fliers passed around third period VA\'s class.  Monday June 14th: Fliers passed around first period VA\'s class.   Thursday June 9th - Tuesday June 15th: Thomas Chen secretly goes around school, AP classes informing people of this surprise party thing.........And everybody else gossips too.  Wednesday June 16th - Secretly distributing party info inside APUSH classes and even on Wargo class\'s blog.....And Rob Muilenburge informs of the plot to drag clark to the 700 wing, and Rob tells Evan and ASB.   Thursday June 17th - SADLY FINDS OUT CLARK IS \"SICK\"!!!!!!!!!!!!!!!!!!! - The substitute was shocked by our sadness - Mrs.Shultz informed of the urgent need to get Clark to school - Martha the secretary calls a million times all the phone numbers........And still no luck........ - Mr.Cox, Mrs.Winters, Mr.Little all informed of the Clark\'s farewell party. - Vanessa Esch and Thomas Chen made that big sign in the commons.  - FINALLY THE BIG THING: THE MS.CLARK CARD IS SIGNED BY EVERYONE!!!!!!!! (2nd period APUSH, 4th period ASB, 6th period (Thomas Chen actually skipped classes to be in 4th period ASB and 6th period APUSH) - ROB makes a PA announcement end of Thursday informing Tomorrow\'s  Friday June 18th - FINALLY FINALLY CLARK IS HERE!!!!!!!!!!!!!!!!!!!!!!!! - Thomas Chen had to re-informed people about the party, same gig, same plan......... - Mr.Russel and other people apparently told Clark about the surprise party............ - AND YES FINALLY AFTER SCHOOL......AT LEAST MY GOAL OF MAKING CLARK HAVE TEARS IS ACCOMPLISHED!!!!!!!!!!!!!!   Since this is Ms.Clark\'s last year of teaching in the beloved Skyview High School, we have to pay all of our most sincere farewell to Skyview\'s QUEEN OF HISTORY AND ASB!!!!!!!!!!!!!!   Here is a comical recap of the stressful ANXIETY ATTACKS before the actual party:  C5hanst123(Thomas Chen), Attakid1172(Rob Muilenburge)  WEDNESDAY, JUNE 16th: (DID NOT KNOW MS.CLARK WOULDN\'T SHOW UP TO SCHOOL TOMORROW!!!!)  C5hanst123: so have you been successful in lobbying people to come to the party tomorrow? Attakid1172: yaeh Attakid1172: i tried to let everyone in asb know, but clark was in class the whole time so it was hard C5hanst123: lol heh C5hanst123: hey anything you bring, or see other people bring for Clark\'s party....leave it in VA\'s class first period ok C5hanst123: since we\'ll be in that class after school anyways C5hanst123: and oh yeah Attakid1172: should i bring food? C5hanst123: sure Attakid1172: k C5hanst123: i have a bag of chip and drinks, and some paper napkins/plates stuff......but any help is greatly appreciated Attakid1172: k C5hanst123: and hey Rob C5hanst123: since you have Clark sixth period....... C5hanst123: and this thing is right after school C5hanst123: it\'ll be weird for Clark to see people all escort to the 700 wing with her Attakid1172: lol Attakid1172: yeah Attakid1172: i can stall her C5hanst123: so are you willing to be the C5hanst123: yes C5hanst123: ! C5hanst123: ! Attakid1172: we talk after class almost every day C5hanst123: KEWL C5hanst123: ! Attakid1172: yeah Attakid1172: no problems C5hanst123: oh and maybe Clark\'ll say, now I have to go to attend this meeting C5hanst123: haha C5hanst123: and then you\'ll say C5hanst123: i\'ll come with you Attakid1172: lol C5hanst123: and by that time hopefully everybody else has already gotten there Attakid1172: yeah C5hanst123: and so you\'ll escort her to the 700 wing Attakid1172: lol C5hanst123: geezus christ this schoolwide surprise party thing is like a \"blind date\" C5hanst123: so much anxiety attacks C5hanst123: like a gamble......we\'ll see how it fits together tomorrow Attakid1172: lol Attakid1172: yeah Attakid1172: it\'s really nice of you to do this for her Attakid1172: she really is a great lady who has done a lot for our class and the school C5hanst123: heh yeah thanks, after all it\'s all of our gracious hearts C5hanst123: exactly, she deserves the best Attakid1172: trust me, i probably know better than anyone C5hanst123: yeah i\'ll miss her sooooooo much next year.......... C5hanst123: ASB.........lol Attakid1172: lol Attakid1172: mrs bevill is awesome Attakid1172: but different C5hanst123: yeah i\'m sure she\'s good too C5hanst123: but Clark is just Clark Attakid1172: clark is unique to say the least C5hanst123: her personality nobody can replace C5hanst123: yeah exactly C5hanst123: oh before the end of the year, (Friday), you have to see some stuff I made this year in my graphic design class C5hanst123: all related to ASB/School Bureaucracy   THURSDAY, JUNE 17th: (MORE ANXIETY ATTACK THAN EVER!!!!!!!! GAAAAAAAAA CLARK DIDN\"T SHOW UP, THE WHOLE PLAN BUSTED ON THURSDAY!!!!)  Attakid1172: lol Attakid1172: it\'ll be fine C5hanst123: yeah hopefully, at least we have that card all nice, and a shirt for her, and apparently a costume too C5hanst123: and oh that sign if all up C5hanst123: in the commons Attakid1172: yeah C5hanst123: so btw C5hanst123: if you don\'t mind C5hanst123: at what approximate place does the honorable, famous Clark resides? Attakid1172: lol, k Attakid1172: clark lives by beaches restaurant   '),(9869,1488330,'15,June,2004','2004-06-15 00:00:00','    Hey All MY AP US HISTORY BUDDIES!  Today\'s my last day of school........I\'m going up to UW for six weeks. I have to say that this year has been an awesome year in CLARK\'s AP US HISTORY class. Despite a somewhat stressed out, apprehensive studying for AP Test period, we\'ve had lots of fun in that class. We\'ve grown both academically and socially.   So enjoy the rest of year three days of school.......Make the most out of your junior year.....Great memories will be with us forever. And old friendship and fun times may never be replaced.   Hopefully I\'ll see some of you again next year in AP English or AP COmparative GOvernment class!  FINALLY, KEEP THE BLOG UP THIS SUMMER!!!!!!!!!!!!!!!!!!  IF YOU CAN, BLOG ANY AND ALL OF YOUR SUMMER FUNS! SHARE ALL YOUR STORIES WITH US!!!!!!!  Best Wishes  Thomas Chen   '),(9870,1488330,'15,June,2004','2004-06-15 00:00:00','    Ahoy bloggers!  This is Sydney and Livy and it\'s the first time either of us has ever blogged.  I guess we\'re a little slow, considering there are only 3 days of school left this year.  Isn\'t that splendid.  We have decided that since so many people are disgusted with childhood obesity, we\'re going to do a little experiment on the issue ... So we\'re going to see how long it takes us to become obese.  This experiment will involve a lot of poptarts and chocolate.  Ummm we can\'t start right now though because we\'re sitting in history class right now, doing \"college research.\"  Well, since this is probably the only blog that Livy and Syd will ever write - everyone have a scrumtrilescent summer and don\'t be too smart! - May the God of Poptarts smile down upon you all, Olivia Guest Pellegrom and Sydney Lauren Clos   '),(9871,1488330,'15,June,2004','2004-06-15 00:00:00','    Ahoy bloggers!  This is Sydney and Livy and it\'s the first time either of us has ever blogged.  I guess we\'re a little slow, considering there are only 3 days of school left this year.  Isn\'t that splendid.  We have decided that since so many people are disgusted with childhood obesity, we\'re going to do a little experiment on the issue ... So we\'re going to see how long it takes us to become obese.  This experiment will involve a lot of poptarts and chocolate.  Ummm we can\'t start right now though because we\'re sitting in history class right now, doing \"college research.\"  Well, since this is probably the only blog that Livy and Syd will ever write - everyone have a scrumtrilescent summer and don\'t be too smart! - May the God of Poptarts smile down upon you all, Olivia Guest Pellegrom and Sydney Lauren Clos   '),(9872,1488330,'15,June,2004','2004-06-15 00:00:00','    Ok.....I guess I was a hypocrite on my last post when I complained about people being up at 3:00AM.....Well this post here is a 3:00AM post on a Tuesday Morning! YAY Well there were justified reasons for this early post because a few hours ago I was still at the amphitheater attending the class of 2004 Graduation Ceremony.  Since half of all my friends are seniors, the ceremony was rather nostalgic and emotional and it\'s unbelievable that some of your friends are already graduating, and forever ending 12 years of public education career! And yeah.....I saw Clark and VA there, and Evan Cate\'s speech was astounding.  I\'ll miss those seniors very much.......AND THE CLASS OF 2005, WE HAVE TO BEAT ALL THE STANDARDS THEY HAVE SET!!!!! WE HAVE TO BE BETTER!   '),(9873,1488330,'14,June,2004','2004-06-14 00:00:00','    Ok, I LOVE how everyone keeps saying they think that the discussion on friday was AWESOME, but on to the next topic, we have to present out books...Will Clark mind if I switched books?  my hitler Mein Kompf or however its spelled never came in, so I read to Kill a Mocking Bird, or of Mice and Men...yes, they were my english class books, but hey, I can present on them!  I read them!  I think this project we\'re doing is HUGE by the way, I just finished the calender thing we had to do...ugh...any suggestions on how to make this go faster? ~CT   '),(9874,1488330,'13,June,2004','2004-06-13 00:00:00','    I was pretty happy with the way the debate went of Friday too, except, unlike the ones everyone else has posted about, mine got pretty heated, and for the first time, I was probably one of the quietest people in the group! But, while I was preparing for the debate, I talked to some of my senior friends about the issue, and i asked them if they knew anything about it. I was surprised to find that most of them didn\'t have a clue what the difference is between the two party\'s economic policies...they didn\'t even care! I was surprised b/c the elections are coming up quickly. You\'d think that people would want to know about the policies that each party is proposing so they could make an educated decision and vote for the party they relate most with. I mean, I guess it makes sense though. The policies on gay marriage and the war in Iraq are getting so much more media attention. How are people supposed to know what the different parties are proposing unless they are given that information on the news or on the front page of the newspaper? Which is kind of interesting to think about: why dont the parties make their economic policies easier for the public to get a hold of? Do they really expect Americans to go searching for the information...NO! Please, we are all so lazy (I\'m making a huge generalization here) that we want the information fed to us through television and bullet points. We dont want to take the time out of our busy schedules to find all the different policies out. I guess it\'s just a little something to think about...why are gay rights and the war getting more attention than an issue that directly affects every single American?   '),(9875,1488330,'11,June,2004','2004-06-11 00:00:00','    I, too, liked the debate on Friday.  I love getting other peoples perspectives on these issues - which are SOOO complex.  Personally, I wish we could talk about taxes and the like the rest of the school year, instead of do this project!  Has anyone started writing their \"Describe Yourself\" essay for the applications?  I did, it took me 45 minutes to write the first two paragraphs.  So muych pressure - those things have to be PERFECT!  Well, anyway in the end I\'m going to be very thankful to Ms. Clark for making us do this stuff now, not in the midst of Senior year!   '),(9876,1488330,'11,June,2004','2004-06-11 00:00:00','    First of all, Corey and Julie, what are you people doing and 1AM and 3AM???? Go to sleep, it\'s good for you. Wow making food that early in the morning.....Crazy people.   I agree with Hilary\'s statement of older people generally carrying stereotypical views towards teenagers. One reason this may arise is the fact that teenagers really don\'t care about politics that much, only around 19% of 18-21 year olds vote! How sad! And then we complain about how the government only cares about healthcare/medicare, and we never get what we want......Well it\'s because we don\'t vote!!!!!!  It\'s really a damaging and detrimental cycle of teenagers not voting, then we don\'t get what we want, and then we teenagers think we are not important in the society and our vote doesn\'t make a difference, so then we don\'t vote......It\'s bad for our society.   AND THUS, SPEAKING OUT, AND VOTE!!!!!!!!!!! Voice your strong oppositions whenever you feel like it........And don\'t always let your parents vote FOR YOU...... I believe we should have more of those last friday\'s discussion types so more people can get involved in politics and research into what our country is actually doing, which directly affects everyone of us when we go to college and worrying about our financial crisis......And maybe military draft.   Everyone of you can start by just going to the schoolboard meetings........It\'s the first step towards understanding how our nation\'s bureaucracy works and how their policies directly affect everyone of us.   That\'s why Rob\'s honorable plan of sacrificing his vote for the vote of the whole school during the November election is such an unprecedented plan!   '),(9877,1488330,'11,June,2004','2004-06-11 00:00:00','    I really liked the debate that we had last Friday. I thought that most of us were mature enough to listen properly and respect the views and values of our peers. I hate how our generation is made out to be irresponsible and wild, but I just wish that some of the older generations wouldn\'t classify all of us, because not everyone is lazy and crazy! I guess that I hope that I won\'t stereotype teenagers when I am old and rusty. Plus I think that with the knowledge that is available and the born genius\' who are at Skyview, the future will be a better place than it is today!   '),(9878,1488330,'11,June,2004','2004-06-11 00:00:00','    The debate on Friday went rather well.  The group I was with was pretty evenly split as for Rep or Dem, however we managed to not let it get too heated.  One of the main topics we focused on was tax cuts, which I learned a lot from.  I\'m not into politics that much, but I do want to know more about the presidential candidates so I can vote in what I believe.  But as for now I\'m making chocolate chip cookies, nothing beats homemade.   '),(9879,1488330,'11,June,2004','2004-06-11 00:00:00','    Is it just me, but at 1 in the morning, you start to get hungry for breakfast?  Last time I was up this late (last night) I made pancakes, this time I have Belgian mini cream puffs, so good!  only about 8 bux at albertsons for about 30 of them, in the frozen dessert section in case anyone wants to try them.  If they have the mini eclairs, those are better.  Personally I could go for a Jack in the Box\'s Spicy Chicken Sandwhich...  followed by some x-box.  Oh ya, I didn\'t like that \"discussion\" we had on friday, everyone got so off topic so much, go from taxes to well-fare, to zoneing laws... ~CT (goin to get some food from J in the B)   '),(9880,1488330,'11,June,2004','2004-06-11 00:00:00','    While my family and I were watching former President Reagan\'s funeral, my dad told me that while a President is in office they have to plan their funeral just in case they die in office. Now that\'s what I call planning ahead. You see, I remember the flag folding part of former President Nixon\'s funeral when I was little, and I hear that the other aspects of his funeral was rather simplistic. But then my dad busts out and says that Nixon was a Quaker. Thus, it was likely that his funeral was planned to be more simplistic and small... just the way he wanted it. I asked my dad if there was more media coverage on Reagan than there was with Nixon, and he said no, there was an equal amount. So if anyone wants to compare the extravagency between the two deaths, it was planned for Nixon\'s funeral to be more private and not so much exposed. So, maybe it\'s not because of the Watergate Scandal that Nixon\'s funeral wasn\'t as \"big\" as Reagan\'s.  I have to say that it makes me sick that some are speaking ill of Reagan. They\'re so focused on their politics that they can\'t even respect the former president. They say things like, \"oh he was responsible for the spread of AIDS\". Don\'t they know that he spent 6 billion dollars for the research of AIDS? I guess not, because they don\'t want to research the facts. How is it that they\'re able to speak of him that way even in his death? Someone in my history class once said, \"Don\'t speak ill of the dead\" and I totally believe that. Respect is key. Even if a certain president passed away, and his policies were ones that I don\'t agree with, I would still have respect enough to not argue against him, or attempt to discredit him. Some of you might be thinking, \"Oh well you\'re a conservative so it\'s easy for you to say that now, since it\'s Reagan that died\". NO, that\'s not the case. It\'s not hard to have respect for the dead, especially ones that DID serve our country, democrat or republican, liberal or conservative. If you want to argue or discredit any politicians, I ask that you please do it with the ones that are alive. Please.   '),(9881,1488330,'10,June,2004','2004-06-10 00:00:00','    When I was researching for tomorrow\'s debate, I found many websites that seemed bias and quick to insult the opposing party\'s policies. I personally think that people should try and have open minds about the opposing position because it could give them more of a perspective on not only the opposite side, but their own side, too. I am glad that we are researching this topic because we need to see how the economic policy that Bush has implemented will affect us in the future. From the articles that I read though, it seems as though Bush\'s 10-year plan is trying to help the middle class when it will most likely benefit to the wealthy instead. Now it is hard for me to believe everything that one reporter thought, so I am going to look for more facts and see if Bush\'s plan is actually helping America and not just cutting taxes and putting more of a burden on the middle class. Hope everyone has a good night!   '),(9882,1488330,'09,June,2004','2004-06-09 00:00:00','    So today I asked my parents what they thought about Reagan. My mom replied, \"O I loved my puppy.\" \"Yes\" I said, I\'m sure you loved your dog as a kid up but I\'m talking about the president.\" Anyway, they don\'t care too much about president Reagan. They get very frustrated with the government and politics. They are rather liberal and quick to criticize. Anyway I was doing some research and I agree with Joe that Reagan\'s administration ended up raising taxes, and while Reagan stated that he would \"get government off our backs\" he ended up increasing the role of the federal government.   '),(9883,1488330,'09,June,2004','2004-06-09 00:00:00','    Its interesting that people remember Reagon for cutting taxes and reducing the size of the governemnt, but in reality his administration raised taxes more than any other, and increased the size of the federal government.  He tried cutting taxes and cutting social programs, but it didnt work.  Also, he was in power during the beginning of the AIDS epidemic, but what did he do to educate the public about the problem?  Nothing.   Also, people say he was great because he wasnt an \"intelectual,\" like George W.  Isnt that a nice way of saying he was stupid?  I hardly see how that was a good thing.  ~Joseph Liberman   '),(9884,1488330,'09,June,2004','2004-06-09 00:00:00','    Ryan, I agree that unabashed, unashamed, mudslinging political rhetoric poisons the mind! This may sound contradictory considering that a large portion of my posts include some sort of stab at Bush, but not EVERY issue. I listen to political talk radio for entertainment, not news. Lars Larson comes on at night and talks about how Abu Ghraib is a positive reflection on the Bush presidency, then Sean Hannity comes on and rants on how lame Bush is for not politically castrating, then executing Don Rumsfeld. Am I a fan of Bush or Rumsfeld? No. Do I think Abu Ghraib is the fault of either, no. I actually commend Rumsfeld on trying to delay the release of the photosâ€” at least the man recognized the way of harm his boys would be put directly into as a result. Atleast I get a good laugh before I fall asleep. I understand why they do what they do. Mudslinging IS entertaining, the problem is, the political extremists that get put on the air collect a FOLLOWING for their charisma and idealism. It\'s rare that their is a perfect candidate for anyone, so in many ways, I could liken the popularity of a man like Michael Savage to the rise of Hitler in Germany. Many Americans feel that our government is gutless to act, and a man like Savage is all about American empowerment. With such political polarity being generated from CNN to Fox News everyone in the media, imagine the cult following a political MODERATE might generate! Talk about extremism!  Now the Bible. So often in history we praises the founding fathers for their briliance in drafting the Constitution. Leaving it open to interpretation, casting it in putty rather than in steel so it can mold as times change. Maybe God should deserve the same type of credit so often to the \"revolutionary thinkers\" of the late 18th century. Everyone interprites the Holy Bible differently, and that is generally accepted by everybody, but this is the first time I\'ve really thought it might be part of the design. I guess the most important fact is that people are looking to God for an answer, and for that, I commend them. God is an invaluable resource. For any atheists out there, you\'re alright by me too, but get to know what\'s out there, you might find a doctrine you like.  Al Tu and I have decided that the U.S. should return to a monetary system similar to that of the Gold Standard, but slightly different. The beer standard. All government issued coins and bills would be backed by their value in beer. Also, all informal college sports bets would be transacted exclusively in beer. Long live the Greek System.  Love Rob   '),(9885,1488330,'08,June,2004','2004-06-08 00:00:00','    So, for all of you who haven\'t heard of the issue of Multiculturalsim it may be time you have. What is Multiculturalism? Well, my personal definition and opinion is that multiculturalism is racism in a politically-correct guise. It holds that an individualâ€™s identity and personal worth are determined by ethnic/racial membership and that all cultures are of equal worth, regardless of their moral views or how they treat people. Multiculturalism holds that ethnic identity should be a central factor in educational and social policy decisions. Multiculturalism would turn this country into a collection of separatist groups competing with each other for power. So what do you think?    Jon G.   '),(9886,1488330,'08,June,2004','2004-06-08 00:00:00','    I have a similar story to that regard. I was a pro-choice dude till my frosh year. I remember a friend who even confided with me a few times that she was going to abort her 4 month old fetus; she\'d told her mom, and part, if not all, of it was based around fear of her father. But the incident that really got me thinking was the night before we at Pendleton H.S. left for a district acting compostition in Portland. We were performing our skits and monolgoues for an audience, and one was about this girl whose mother forced her to abort her child, and then ran away from her mother, who kept trying to find her and bring her home. The daughter wouldn\'t, since she\'d killed off a living thing in her womb. It was all along those lines, and I lost all pretensions for abortion (including the for population control bit). That piece of acting was especially powerful, and it gave me a very personal look at the whole issue. It is something living to me, regardless of its cognizance of its state, for it\'s growing and is a nutrient-needing being, possessing of a soul. That\'s how I feel.   '),(9887,1488330,'08,June,2004','2004-06-08 00:00:00','    I know how so many of you have complained about bringing up things that have already been brought up, but I really wanted to tell you all about something I learned this morning during my first period class.  I was talking to one of my friends, and she told me that her sister (who is 19 years old) got pregnant about a month ago.  The baby\'s wonderful and everything\'s fine, but I then my friend told me that her sister is considering getting an abortion instead of having the baby.  Before hearing that, I had been fine with abortion.  I don\'t consider it to be killing a life because it doesn\'t really have life yet, and I was totally pro-choice and all that.  But after hearing that one of my friend\'s sisters is considering getting an abortion...it just made me so sad that she would do that.  Maybe it\'s just because of the personal connection, but I just don\'t view abortion the same today as I did yesterday.  My friend is so excited to be an aunt! And just to take that away from her and everything, just because it would be the easy way out just doesn\'t seem right to me anymore.  I may be over emotional or something like that, but hey.  It\'s all good to me!  I know that this is an old topic, so you all don\'t really have to respond to it or anything, I just view abortion completely differently now that I have a more personal connection with it.   '),(9888,1488330,'08,June,2004','2004-06-08 00:00:00','    Ryan, your points are perfectly valid. I found myself reading some info sheets on economic policy last night, and I found myself very engrossed in it. It was either \"Look what Bush\'s economic policies are doing for us positively\" or the opposite. I found some valid facts, but also language that was deliberate, a spin to catch the public. I too often despise the political mire, and it can disgust me how politicians can continue to bicker endlessly along partisan lines till the cows come home, and there aren\'t any cows in D.C. Partisanship makes me mad, it was one of the reasons I was so ticked reading the stuff in the text book a month back on Richard Nixon - the old heated partisan warrior. You put this whole bickering gig in historical perspective, and I love that. Much more fighting than helping has ever been done in this country; truly, in the entire history of man. There\'s to much to loose for all our politicians to work together.      I\'m going to become slightly partisan here, it may seem, but this is what I observed when reading yesterday: that we\'ve been getting jobs back the last three months, but we lost several million in the past three years. There were some cool things that we have coming, but a lot of crap happened too. I hate that I\'m not being specific, but I\'ve got the stuff on me, so if you\'d like I could show it to you, and I was carefully looking at this info, trying to be objective. I seem to have more bias against the Bush administration than the opposite side, but I saw that in myself and tried to look in that same light on those critical of the administration. Both sides have some valid points, but for me I\'m not much of a Reaganomics guy, which we\'re living with under President Bush. I like that I\'m reasearching, and it keeps me informed. I also don\'t like these tax cuts. To me, a government needs its money, its surpluses. We need the programs that the government provides - they should be limited so that people can\'t take advantage of them, but you need to have well funded programs. Raegan cut taxes, aznd limited funfding for some of these programs. I also wonder at these stats in the course of my research, whether they were caused by Bush\'s policies directly, or are a part of regular economic rise and fall, or both.     I hated being on the pro-George site, \'cause it was so obvious that they were trying to use stuff like patriotism to catch unwary voters for them. Both sides do it, and it disgusted me that they were trying to Big Brother me and take me for most of this country\'s citizens - uniformed, idiotic voters. Yet the uninformed part is more true, and I\'m being judgmental to an extent. Just always be informed, go beyond the surface and see what\'s really going on. If you don\'t like George for instance, know why, and if you want something new, make sure Kerry would institute the policies that you would want; don\'t just vote a man in to get rid of another. I pray that we may have the best for this country. Adios and God bless.  Tristan   '),(9889,1488330,'08,June,2004','2004-06-08 00:00:00','    Weird is one of the only words in the english dictionary that is the exemption from the \'i\' before \'e\' rule.   .....isn\'t that weird? Also, Rhode Island is neither a road nor an island....discuss... ~CT   '),(9890,1488330,'08,June,2004','2004-06-08 00:00:00','    Yeah AAAPPP Party......Please attend and enjoy, all you AP Party animals!  Maybe Rob\'s Trivial Pursuit and other \"Intellectual\" Games could be part of this unique AP Party theme....(Hehe JK).   And apparently we have another friend from Lafayette Lousiana who commented on natalya\'s last post. Thank you very much all the visitors who may stumble upon this scary blog.   '),(9891,1488330,'07,June,2004','2004-06-07 00:00:00','    WUT UP BLOGGERSS YEAAAYYYAA! its the AP party 2morow DO NOT FORGET i planned it with the help of some of the other ap\'ers BRING A DESSERT i like chocolate n icecream and cookies n stuff!!!! FROSTING IS A PLUS! so get ready to play some games and have some fun!!!!!!!!!!!!!!!!!!!!!!!! PEACE OUT HOMIES! jess   '),(9892,1488330,'07,June,2004','2004-06-07 00:00:00','    yeah.. i already got mine picked out... if i have the supplies, i\'ll be making cookies.. if not.. well.. that sucks then.. my cookies rock.. what else are we doing? someone should bring some board and video games.. i have a trig final tomorrow... obviously i\'m not too concerned about it, because i\'m talking about making cookies..   '),(9893,1488330,'07,June,2004','2004-06-07 00:00:00','    Congratulations...you have made the blog work.  Your writing is interesting and lively.  It is actually fun to read because you aren\'t worried about getting a grade but more worried about making a point while entertaining your blogger friends.  I know many of you are tired of the GM issue but I think you made a lot of progress...there is a lot to think and evaluate on this issue...and will continue to be topic for discussion as many individuals continue to press the issue.  Ryan...great insight...I would love for John Kerry to embrace McCaine as his running mate.  Wouldn\'t that be an interesting team.  However, through the bias and hidden agendas we have made progress in many issues.  The history of America has been a struggle to define ourselves.  Every step of the way issues, beliefs, and ideologies have been challenged.  Howard Zinn has an interesting book called...\"You Can\'t Stay Neutral on a Moving Train!\"  He makes a great point.  It is nearly impossible to be completely objective based on your life experiences. However, I do think it is important to challenge yourself to think, listen and learn.  No one has all the answers...  How are you doing on your Democrat V. Republican economic policies? Can you share some interesting websites?  Also I wanted to remind you about the AP Party tomorrow night.  It should be fun.  Bring a shirt, a sharpie and dessert....you may want to team up with a couple people in the dessert bringing.    Good work...your blog is enjoyable to read...keep it interesting and informative.   Many of you seem mature beyond your years.  Gertie   '),(9894,1488330,'07,June,2004','2004-06-07 00:00:00','    I abhor politics by nature because opinionated writers are unable to present facts without intertwining their own agenda.  As a student I process information daily, and have found the most effective way to learn concepts is through evaluating raw facts, and analyzing relationships between themes.  In politics I find this straightforward analysis nearly impossible because no source presents all of the facts on a single issue; instead, columnists selectively use facts to win imaginary arguments.  As a result, I find myself repulsed by political reporting.  Iâ€™m torn between the desire to learn about the world around me and the intense repugnance I feel toward the political media.  People make politics personal, which is unnecessary and detracts from progress.  I dream of a world where opposing political parties can discuss issues and make decisions based on whatâ€™s best for the nation and not whatâ€™s best for their respective organizations; however, I understand this is the nature of politics; they are purely subjective.    America is a half-constructed bridge, spanning the continent from the Pacific to the Atlantic, and the political parties are the construction team responsible for the next phase in building.  Because the political parties are on the same team it would seem logical for them to work together to build the strongest structure possible; however, since their inception they spent more time waging wars and fighting than planning for the future.  Government officials should be driven by building the best nation possible, not lining their pockets or making a name for themselves.  Unfortunately, this is seldom the case and the root of my inner conflict.  The role of the government is to create a healthy, sound country, and far too many politicians lose sight of this goal for me to follow avidly; however, because the fate of the nation depends on voters, I and others like me must swim through the political gobbledegook and get to the heart of the issues â€“ I just wish I didnâ€™t have to wear waders to do it.    -Ryan  PS - I admire all of you who closely follow politics and respect your opinions greatly   '),(9895,1488330,'07,June,2004','2004-06-07 00:00:00','    One of the things that i think of first when i hear about Reagon is Iran-Contra.  I think that this event is the best representation of the kind of person Reagon was.   '),(9896,1488330,'07,June,2004','2004-06-07 00:00:00','    It is funny that you drive on a parkway, and park on a driveway.  Also, I get high on a freeway, but freebase on a highway.  Curious.  Anyway, the Mariners have so much potential.  Promising start last night by rookie Clint Nageotte (pronounced na-zhot) who gave up 0 runs in six innings.  The largest Mariner problem has been synchronizing offense and defense (one has one f, the other has two).  Mariner\'s starters Joel Pinero and Freddy Garcia each have excellent ERA\'s, yet abismal records.  Lastly, the Pistons can beat the Lakers, and I see the finals as a tossup.  Anyone that thinks the Lakers are going to coast by anther lame-ass eastern conference team has another thing coming.  That thing is the ridiculously powerful Wallace and Wallace combo, Tayshaun Prince guarding Kobe, and my boy Rip Hamilton who took a page out of Jesse T\'s book and wears the Hannibal mask on his way to being the Piston\'s postseason leading scorer.  I prefer my TP hanging under, and I\'ll fight that to the death with Biblical references to back it up.  Love, Rob  P.S.  To all you concerned little kiddies in blogger land, Mary and I are good friends and remain to be good friends.  Last night we saw the Presidents of The United States of America play the Rose Festival.  I can say with 100% confidence it was the best concert I\'ve ever seen for $5.  Long live Lump, Kitty, Peaches, and The Groovin Postman.   '),(9897,1488330,'07,June,2004','2004-06-07 00:00:00','    CT brings up a good point... you do park on a driveway and drive on a parkway... weird.... anyway, we do need a new topic to talk about.. I haven\'t blogged for awhile, and I see we\'re still on GM.. From what I\'m reading, its a touchy subject. personally, i agree with Rob.. I\'m uncomfortable seeing it, but its not immoral. Yesterday, I saw a truck with a sticker that had this: Marriage = Man + Woman for a moment, i considered hitting him... but i enjoy driving, and if i hit someone, i will not be driving. People are quoting the Bible and such, but when it comes down to it, I believe that the big stink about GM is that most people are uncomfortable with the thought of GM, so they get other people riled up about it, and now its a hot subject. There\'s my two cents, and its time for a new subject to talk about...  uhh... hmm.... erm... why did the chicken cross the road? there\'s a website with how other people would answer: www.chickenjoke.com there. Laughter is the bestest medicine, so laugh it up!   '),(9898,1488330,'07,June,2004','2004-06-07 00:00:00','    Something else to talk about?  I finally got on this blog, there\'s one thing.  But another is that friggen Ronald Reagon.  My dad went off on him about him being the last president he liked that America\'s had.  I really haven\'t looked up anything about him, and probably am not, but from what I hear, he must\'ve been a good guy.  Still not interesting enough for something new to talk about?  How about You park on a drive way and drive on a park way...discuss... ~CT   '),(9899,1488330,'06,June,2004','2004-06-06 00:00:00','    Just an interesting fact: in order to have your painting be exhibited at the Louvre you have to be dead for at least sixty years. I just don\'t understand why do you have to be dead before you can become famous or a hero. Look at Socrates who had been tried and killed for his work of science, or Copernicus who had been called a heretic for his proposal of the sun being the center of the universe, or Mozart who was buried in a common group grave; or Beethoven who was poor, deaf, and unacclaimed during his life; or Bach who was also poor with ten kids, and in a constant struggle with his employers - all famous composers after their death; or look at Van Goh who died poor, drunk and unknown for his great paintings; or the writers such as Solzhenicin and Pasternak who have been banished and sent into exile and now are known worldwide for their great works of literature. They all have been famous after death and will continue to be celebrated for centuries to come and yet during their life times they were either banished and criticized for their work or drunk, poor and struggling for survival. Now I got a bit off topic here. I was going to talk about Reagan and got carried away with my appraisal of true heroes. I hear everyone talk about Reagan and how great he is but really what was so great about him? It seems as if everyone forgot about him and his great deeds for a decade and bothered to discover a hero in him only as soon as he died. So what do you think he did that was so great, I didn\'t really remember him being a major personae in the APUSH book.   P.S. I just want to mention that no one won the Cold War, and no one destroyed the USSR. It destroyed itself, it simply had a malignant cancer which killed it from inside out. It\'s actually very interesting to hear and observe different perspectives on the issue of Cold War.   '),(9900,1488330,'06,June,2004','2004-06-06 00:00:00','    I\'m trying frantically to come up with something else to discuss so that we can get off the gay marriage issue.  So what kind of peanut butter do you all prefer: crunchy, or creamy?  Do you put the toilet paper so it rolls from underneath, or over top?  Doritos or Cheetos?  Haha, I\'m just rambling now, but we desperately need something else to talk about on here.   '),(9901,1488330,'06,June,2004','2004-06-06 00:00:00','    Hey Chris, Mary, Rob, and Natalya........WOWWWWWWWW WHAT AN ECLECTIC GROUP OF VOCAL AP US HISTORY STUDENTS!!!!!!!! I\'m actually really glad that there is HEATED DEBATE about such a controversial topic as Gay Marriage.   It\'s so great to see that at the end or beginning of your posts, Rob and Mary, that you two have to reconcille with each other that you love each other.  Because we are all still great friends, no matter what difference of opinions we have. Yeah true friendship has to be able to withstand political debates!    Chris, I honor your detailed analysis of the motives/arguments/reasonings behind both polarized sides of this GM issue......Your post sounded like a mature moderator\'s post trying to bring together the two sides. However, I have to dissagree with you on one important fact, and that is, the Gay Marriage debate will not always go around in circles. Not everyone in the country (on both sides) is willing to sit idly and let the gay marriage issue be pushed out of debate like a gag order. Even if the issue of gay marriage will not be resolved for decades......It will eventually be ......No matter what end of political spectrum America\'s general populace stands, there are and WILL ALWAYS be the progressives who will change America forever. And here I have to insert my own personal opinion (WAAAAAA SCARY PERSONAL OPINION)......That I really hope the resolved will forever legalize gay marriage. And I\'m convinced along with a lot of my peers that if America continues its 250 years of journey in the pursuit for the real meaning of freedom and democracy, that Gay Marriage will eventually be legalized.   But this can be a long wait........Maybe many decades......But I\'m convinced that eventually it will happen.  Simply looking at America\'s past, you\'ll see that altho there are short term regressions to more goold old traditional fundamental values, the general path of America\'s history has always been taking the step to become more progressive and democratic and egalitarian (For such important issues now we take granted such as: From the first ten amendments to: ALL MEN\'S SUFFRAGE, Public Education, ABOLITION OF SLAVERY, Women\'s right to vote, Separation of Church and State (including teaching evolution), Trust-Busting policies on monopolies, Social Securities, Healthcare/Medicare, Minimum Wage, Conservation, De-Segregation, Civil Rights, Sexual Revolution, Non-Discriminatory Clause(There the disabled gained their rights), And finally African American\'s Rights to Vote ). And shorterm regression of more conservative values such as prohibition on Alcohol and Jim Crow Lows just DID NOT WORK AND DID NOT LAST. Just looking at your parents and grandparents, which generation is more \"conservative\", which generation is more \"free\" or which generation strives for more freedom? (56% of 17-24 year olds suport GM as opposed to 19% 65 years and over). The more recent list that I will eventually able to add to the list of America\'s journey towards the true democratic values has become abortion and gay marriage.  And at the end, I love everybody here who has contributed to the great DISCUSSION OF OUR AP US HISTORY BLOG!   '),(9902,1488330,'06,June,2004','2004-06-06 00:00:00','    All I have to say is that Rob said everything there was to say. I mean the fact that no matter what kind or how many arguments you bring forward, every person will still hold back to his/her own views/believes. I just wish people weren\'t so judgmental and close-minded, dividing the world into wrong or right, moral/immoral, good/bad. I personally believe that all of us have our own ten/twenty or however many commandments that we hold close to our hearts and which are precisely the reflections of our views on life based on our OWN individual experiences. That\'s the primary reason why I call myself \"partially religious\" and of course because I believe in God in my own special way... So therefore I invite you all to convert to Buddhism in order to experience the eternal happiness in the state of Nirvana... Long life to Allah.   '),(9903,1488330,'06,June,2004','2004-06-06 00:00:00','    Rob and Mary, thinking of you two going head to head gives me the urge to get on a plane and leave the country. But nontheless I have been reading the arguments on the blog lately with interest. Both Rob and Mary both have very valid points and I\'m sure if given the chance they would contradict one another into the next millenium. But the fact is and I quote Rob,\"one thing I have learned through the GM debate is that you won\'t be able to convince anyone to drastically alter their position, based on religion or otherwise.\" So these arguments stewed up by the bright minds of our Ap US history class are hopeless,  in the end it all comes down to what you believe in as an individual. For example: Mary holds strongly to her faith, a virtue of which every question derived from the bible comes down to for answer, and uses this belief to battle and contradict Rob\'s arguments, while Rob uses modern situations as proof to regard or disregard arguments like Marys that are based upon faith. Both of these strong willed individuals represent 2 of the many separate groups in the battle for gay marriage, each group possessing a strong and opinionated following. The only problem with this situation is that neither group appears to be morally wrong in their thinking, and therefore there is not an easy way to chose sides. Due to this the battle for gay marriage will go around in circles in vain attempts of each side of the argument to gain a majority on the matter. But with each argument presented by each of the opposite viewed sides there will be arguments to address the stated argument and arguments to address the stated arguments argument, and with each contradiction those groups will move further apart from eachother as characteristic human stubborness is triggered with the attack on ones argument. With this reasoning in mind, I don\'t think that gay marriage will be legalized throughout the United States for quite some time. No one can expect a drastic idea such a same sex marriage to be accepted by a society of people who have been morally against the idea of homosexuality their entire lives. And with that I think I can safely say that I am sick of talking about it.   '),(9904,1488330,'06,June,2004','2004-06-06 00:00:00','    I feel that the gay marriage discussion has gone on WAY to long, and that it is time to find a new topic.  However, I do have one thing to say on the issue. If anyone is against gay marriage, than please, donâ€™t marry someone of your own gender.  Why can\'t we just live our own lives, instead of trying to control every one else?  If people just focused on their own morals, instead of trying to force everyone else to think how they do, than the world would be a much better place.  ~Joseph   '),(9905,1488330,'06,June,2004','2004-06-06 00:00:00','    I feel like I\'ve been polluting the blog recently (throwback to globabl warming which Ryan Knight claims is a myth) with my GM posts, and so I plan on leaving it with this last post on the topic.  It\'s not that I\'ll give up backing it till it\'s legal, so feel free to talk to me in person, but I\'ve decided it\'s alright for the blog to go another direction.  So here it is.  Mary, I appreciate your love for me, and I hope you know it\'s entirely reciprocal, but I also hope you know that the Bible means no more to you than it does to me.  I hope I\'ve made it clear that I\'m familiar with what the Bible has to say, and CLEARLY we have different interpretations (I guesss neither of us can call either interpretation WRONG), but I can confidently say right now that every human in the world has a bias.  It\'s human nature to have some sort of preconceived notion, some unfaltering instinct that affects how we see the world.  Part of our bias comes from the society we live in.  We have our cultural norms, our milieus or what have you that would make aspects of the culture in our neighbors of Mexico and Canada confusing at times.  Mexicans have a very different view of personal space, not afraid to come sit right down next to you on an empty bus.  As the egotistical douche yet talented film-maker Michael Moore demonstrated in Bowling For Columbine, Canadians don\'t feel the need to lock their doors at night.  A scenario many not feel comfortable with in their homes in the states.  My point is that when it says in Romans that men turned to other men in unnatural lust, in Roman society homosexuality was not considered normal, especially not natural.  I myself get uncomfortable seeing men kissing (...\"I am beautiful, no matter what they say...\"), but I don\'t see it as immoral.  When there is a natural inclination to repel homosexuality combined with a society even less friendly to gays, bias plays a factor in Paul\'s writing.  Do you see where I\'m coming from?  I\'m not trying to discredit any of the apostles, but I do want to remind the world that they can\'t be expected to be any more than human.  (For the record, I\'m referencing the New Testament exclusively because Mary referred to herself as a judeo-christian.  If you want my argument against the Old Testament, jump back a couple weeks worth of posts)  Moving on to data.  Even if Rush does cite his sources, they could be the product of any number of think tanks funded by any number of VERY biased corporations.  Tobacco did all kinds of studies saying how bad tobacco WASN\'T for you.  I think NPR would be one of the most direct sources of info out there.  If you don\'t want to believe Sean Hannity, I don\'t really either, but you can trust a credible source like an NPR newsbreak.  Lastly because I\'m a left-winged secularist who only gets A\'s because I wright what teachers want to hear politcally, let me back up one more statement with imperical data; going after your spelling was a good natured ribbing, but you were 1/9 in spelling marriage correctly.  (Take that whole last paragraph in a sarcastic tone, but the data I believe to be correct.  Do you want me to reference my source?)  My last statement will address you not agreeing with homosexuality.  Why not?  You know my arguments, and I guess you can say because that\'s your morality, but there IS a basis for morality.  I just want to know what yours is.  By the way, I LOVE Mary Tyner as a human being.  The great thing about this country is that we\'re all allowed to have our beliefs and opinions, share them with each other, and be homies at the end of the day.  Love Rob   '),(9906,1488330,'06,June,2004','2004-06-06 00:00:00','    Thanks for the response Bob. You sure can quote stuff. By the way, I know how to spell marrIAge, it was just a typo. That was kind of a cheap shot. I find it amazing, though, how you can quote this, and quote that, but yet you don\'t know what verses mean. Again, people taking such an objective view when they regard the bible. You don\'t know how to interpret. Or is it that I interpret it differently? Obviously, there is a difference between us in what we regard as the truth. (By the way, Rush always tells his audience what his sources are, or if what he\'s saying is his opinion.) When it comes to the bible/religious stuff, there\'s a difference between having faith in a religion, or just believing in it. For me, it\'s faith. It\'s not that I take every single word in the bible and take it literally. That\'s not the intent. Just a note... Jesus, in his teachings, spoke metaphorically, or through examples and stories so he could teach his disciples. Thus, not everything is to be taken literally. These disciples, who you call bias, are the ones who have the first hand account and have written these accounts in their books Mark, Luke, John and so on. If you have faith, you don\'t think of these men as biased figures who totally disregard the views of others. I\'m not saying don\'t ever question these men and do everything they say, they are the new gods now... ahhhh! What I\'m saying is that these men, these disciples, in regards to christianity, are telling what they know, are telling what Jesus taught them. What I\'m saying in all of this mumbo jumbo is to be careful when you use the bible for reference, everyone, because when you use the bible, you\'re entering the territory of religion, and you\'re right rob this is a history class, not a theology class.   If you really want to know my direct opinion on Gay Marriage, I\'m against it. But the question is, to what extent? I don\'t even know the answer to that question, because I\'m not against gay people. But I am against the lifestyle. I\'ll go along like it\'s normal, which is called tolerance, but inside my head I\'ll be thinking it\'s wrong and there\'s no way you can change that, the same way I can\'t change you. But I dare you to try.  Back to data, it\'s not like I don\'t use it. I QUESTION every blatant piece of data to ASSESS THE VALIDITY OF THE STATEMENT (haha, never thought I\'d hear that outside the AP exam). I question not only the liberals, but also conservatives. I question my side as well because I want to make sure that I know what it is I\'m believing in/ have faith in. Always ask questions. And make them simple questions too. That way, if you don\'t get a straight forward, simple answer, then you know that that person who is answering the question is BS\'ing with you. So no, Robby Bobby, I don\'t just go with my gut feeling alone. I know the facts first, I repeat THE FACTS first, and then I perhaps use my gut/mind/brain/head, whatever you want to call it, that makes me decide what it is I say, do, think and stuff.   Thanks for the response. I love you Rob.   '),(9907,1488330,'06,June,2004','2004-06-06 00:00:00','    Mary, I admire your commitment to your statements.  It\'s easy to hide behind rhetoric and avoid conflict, but by standing by what you say and inviting opposition you show tremendous security.  I would first like to address your alarmingly confrontational attitude towards numbers and statistics.  You are correct that data collections can be interpreted many different ways, and are often presented strategically to convey a certain point of view, but by no means is that always the case.  With my personal experience I find the majority number twisters to be those on political talk shows, including your beloved Rush Limbaugh.  Al Franken does the same thing, and I\'m not trying to say it\'s strictly a conservative thing, but try and argue with the fact that states at the time of this post, approximately 117 billions of American dollars have been spent on the war in Iraq.  Or that since the invasion, 828 Americans have lost their lives in combat.  Tell me that statistic can be twisted to comfort the mothers, fathers, wives, husbands, and children of those lost in Iraq.  If you don\'t base decisions, viewpoints, etc. on \"imperical data,\" what exactly DO you base them on?  Gut feeling?  Gut feeling gets you the Salem Witch trials.  A clever little process known as the Scientific Method led the world out of The Dark Ages and into the Renaissance and is responsible for the development and institution of MODERN MEDICINE.  If I had cancer, gut instinct wouldn\'t lead me to subject myself to chemotherapy or radiation treatmentâ€” both cell-anhialating and physically exhausting therapies, but the imperical data that shows that survival rate is MUCH higher with those treatments would probably sway my thinking.  Moving on to your GM argument, I\'m still unclear.  Do you believe that GM should be legal?  Clearly you have stated you do not support it based on the Bible, that is, the New Testament, but based on the current law, do you believe that amending the Constitution to ban GM is legal?  One thing I have learned through the GM debate is that I won\'t be able to convince anyone to drastically alter their position, based on religion or otherwise.  I have tried and tried, typed countless words, referencing countless verses, and yet I have done little to directly express what I keep alluding to: The Bible, New Testament or Old, is not an infallible source.  The most direct reference to homosexuality in the New Testament comes from Romans, which loosely states that men turned to men and committed unnatural acts out of lust.  It does not say that homosexuality is wrong, just \"unnatural.\"  With that out of the way let me remind everyone that the New Testament is not document written by the hand of God, or even translated from God\'s divine mouth as some of the Old Testament was.  The New Testament is a collection of first-hand accounts with Jesus from various people; the Romans, Mark, Jude, etc.  The point is, the reader of the Gospel must remember that this collection of accounts is at all times subject to the author\'s bias, and even the deviation from the original meaning through years of translation.  For all anyone knows, \"unnatural\" as it was written might have been more accurately translated as \"unconventional.\"  The Bible, New Testament and Old, contains excellent themes on how to live ones life as well as the most redeeming sacrifice of the Son of God for eternal souls of all us sinners, but it is not a document that can be taken word for word as the unvarying decree of the Most High.  I disagree with you that God or His son would object to the marriage of two homosexuals who followed the teachings of Jesus on marriage; where their flesh becomes one, they are devoted to each other, and that they love each other as they love themselves.  Does He find THAT unholy?  Lastly, the Old Testament has some ridiculous inapplicable laws, as I demonstrated in previous posts.  However, the New Testament is not clean of some socially questionable direction as well.  The book of Ephesians says \"the husband is the head of the wife as Christ is the head of the church, his body, of which he is the Savior\".  Does this condemn all Christian wives to a life of servitude?  Certainly, with the influence of the Feminine Mystique (this IS a history blog after all), huge amounts of women have become significant breadwinners in the household, is this What Jesus Would have them Do?  I hope I have presented my ideas clearly.  For future reference, marriage is spelled as so, not marrAIge.  God Bless and good night.  Love Rob   '),(9908,1488330,'06,June,2004','2004-06-06 00:00:00','    You are certainly welcome Natalya. I\'m glad you read what I had to say.  For me, I never trust numbers unless I know the source in which those numbers come from. Who was surveyed? What questions were asked in regards to the survey? There are a lot of ways numbers can be manipulated, which, as a result, can sway the data in a way that supports an idea. I rarely rely on percentages, ratios and what-not unless I know how the data was brought up. For example, I saw on the back of a bus in Portland the other day that read on the rear \"249 cars are at home because I am on the road\". How do I know they didn\'t make up the number? Most people who ride the bus don\'t even own a car! So, by counting the number of people who ride the bus everyday for reliance on the accurancy of that particular data isn\'t gonna cut it. That is the type of BS statistic that makes me question every form of data I hear, unless I know what the source is. So NPR is going to need to present their resources for their data before I trust what they say.  I have to ask Nat, how is it that you are somewhat religious, but don\'t even know if religious is real? Interesting. In regards to the bible, those who believe in the judeo-christian religion go by the New Testament, the place where Jesus\'s teachings are, which is the place where the religion originated. The New Testament is the place where Jesus speaks of homosexuality. And don\'t call people hypocrites because they don\'t fast or don\'t follow the Ten commandments strictly. That is part of the Old Testament. And it seems to me that you think that fasting and ten commandments stuff is all you think the Christian religion is. It seems to me that you\'re taking the standpoint of christianity very objectively, in which you think that we point fingers at the gay marriage issue and say \"you\'re sinning! you\'re sinning!\". We all know people sin. People who \"go by the bible\" sin. The Pope has sinned! What I\'m saying is, is that as a christian, in regards to the Gay Marriage issue, we\'re not pointing fingers as to who\'s a sinner for being gay. It\'s the fact that those who support gay marraige are trying to take the sacred, holy idea of marraige and turn it into an unholy union. And yes, you may quote me on this is if anyone wants to respond.   '),(9909,1488330,'05,June,2004','2004-06-05 00:00:00','    The Tony Awards are tonight at eight! I\'m going to watch, I\'m excited. Another thing that I\'m excited about is amazingly senior project. I really like my topic and I can\'t wait to do my project. I think Mrs. VA was right when she said that you should do something that is meaningful to you.  OK, college Portfolio questions... It says we need to apply to at least three scholarships. I don\'t know about other people but I\'m having a hard time finding scholarships. All the ones that I want to do the deadlines have already passed. Can we use them for the project even if the deadline has passed? Finally, I saw the Skyview band in the Starlight Parade on TV. It was cool, great job you guys!   '),(9910,1488330,'05,June,2004','2004-06-05 00:00:00','    I am not sure if anyone has heard, but former President Ronald Reagan died today from pneumonia due to his Alzheimers disease. I was watching a tribute to him on TV today and learned more about his life than what I learned about his political career from the history books. He was the oldest president to be elected at 69. He also lived longer than any other president. He died at the age of 93 in his home. I guess I never actually realized the major impact that he had on America until I watched his tribute. He was able to bring hope to Americans and bring them to an age that they were able to be happy. I never really knew about him until this year, but I wish that I could have experienced and seen his great presidency in action. It is funny, though, that I never really wanted to learn a whole lot about him, but after his death, he seems intriguing and somewhat of a legend, which makes me want to know more about him and his life. I hope that everyone will at least give there respects to the man who gave so much to our country.   '),(9911,1488330,'04,June,2004','2004-06-04 00:00:00','    Yeah Julie for our VA\'s english classes, there is an AP FEST/PARTY on June 10th from 5:30 to 7:00. We have a committee planning that party and I\'m not sure whether those planning people have coordinated the party to include the US history students too.  On the other hand, we have to do something special just for Ms.Clark....Maybe everyone has to sign and write a message on a card or something....As a farewell to all the dedications and years of work she has contributed to Skyview. So maybe we should have a special US history party on a different date.  In the mean time, I have to find those AP FEST committee members........   '),(9912,1488330,'04,June,2004','2004-06-04 00:00:00','    Thank you Mr. Parlee for teaching our class while Ms. Clark was enjoying the lovely weather and moving her belongings.  So Thomas what did you have in mind for Ms. Clark\'s goodbye party, when, where, who, etc.  Let me know and I\'ll try and help out.  Enjoy the weekend everyone, summer is almost here.  Oh and if you have the chance go support our band and dance team at the Starlight Parade!   '),(9913,1488330,'03,June,2004','2004-06-03 00:00:00','    Heyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy everybody. I agree global warming is serious and the open use of chlouroflouro carbons which is a large addition to the problem should be regulated in third world countries. About the movie \"The Day after tomorrow\"; I went into the movie expecting it to be concerning the effects of global warming but really they just threw watchers some bs about the North Atlantic current shifting and heating things up. They did not address the common problems in the world today that cause global warming. All in all the movie had good special effects and it was cool to see mother nature wreck havoc. Anyway I hope you all have a great weekend.   Chris   '),(9914,1488330,'03,June,2004','2004-06-03 00:00:00','    Wow, this is news to me. Some unknown visitor posted a comment on our site claiming to be an AP US history student! Well Hello there, our AP US History friend from another some other part of this country, Jenny or whatever your name might be.  Anyways, you are our new guest, and welcome to the beloved Skyview High School Vancouver Washington AP US history blog!  We would really like to know where are you from and how did you get around finding this site in the first place. Hehe Anyways, thank you for posting and hopefully we can hear from you soon.   On to another news? the AP PARTY?????? And EVERYONE please listen, since Clark\'s not back yet, we have to somehow plan a farewell party for Ms.Clark!   And then....global warming....Heh I URGENTLY want to see that movie.....And especially since my whole senior project and this year\'s ISEF project is dedicated to the research and further understand of pollutant emissions and its potential devastating effect on the chemistry of the atmosphere  Finally, we have to thank Mr.Parlee tomorrow for his great job as our one week long sub!   Thomas   '),(9915,1488330,'02,June,2004','2004-06-02 00:00:00','    Has anyone seen the new movie The Day After Tomorrow?  I heard it\'s about the effects of global warming and should be pretty good.  Maybe seeing the moive will motovate some people to help improve conditions before they worsen.  However there will always be people who could care less about the future and do not desire to sell their gas guzzling polution contributing monster vehicles.  Any ways, think about the many ways we can individually contribute to improving our damaged environment and do one thing different for the sake of future generations.  PS: Thanks Thomas for changing the blogger set-up.  It makes it easier to see when I\'ve posted because I forget between my other classes.   '),(9916,1488330,'02,June,2004','2004-06-02 00:00:00','    Where the hell is Bob Dole, he had one good run back in god knows when and then he just quit. With two crappy top canditates I think we could use a little more competition like trusty Bob. DONT THINK IM KIDDING YOU DONT KNOW ME SIT DOWN   '),(9917,1488330,'02,June,2004','2004-06-02 00:00:00','    Thank you Mary for providing some good opinion to respond to. As I was reading your post I found my self disagreeing with a couple of statements you made so I had to respond. I just don\'t see how you can say \"I want to take down the idea that right-winged, conservatives aren\'t open people. DOn\'t believe that stereotype,\" when you yourself blatantly state: \" Students who are viewed to be the smart ones are viewed that way... because they\'re left-winged/secularists who use data up the ying-yang to prove their point.\" Isn\'t that a major stereotype you just have created yourself? It seems as if you call all the \"smart\" people boot-lickers who side with the side that their liberal teachers take, that we are not guided by morals in our arguments but only by \"data\" and scientific experiments. I personally have very strong morals I comply with and I am somewhat religious (if there is such a thing) but I always prove my every argument with \"imperical data\" just because that\'s the info that can not be made up and can not be proven wrong or considered opinionated; numbers do not swing either liberal or conservative, they are neutral. What you say about writing papers that teachers like I also consider to be very argumentative, and it of course depends on the school you go to, but the majority of teachers I am sure would prefer you to write a paper that is original in idea and content and shows you to be open-minded thinker.  About abortion, it\'s great that you have that kind of sense of responsibility for your actions, and I think it\'s partly because you know that that won\'t ever happen to you, but we have to remember about all the people that simply do not have that kind of education/surrounding atmosphere/family to support them when they are faced with that kind of choice of either having a child and giving up their future or seeing a doctor for a couple hours and killing that precious spark of life. I still don\'t know what I would exactly do myself, but I would probably side with you. However I am very strongly for abortion for those who NEED it, or who believe they do.  GM: I don\'t know how you can tell a gay person that he/she wasn\'t even \"intended\" to be alive on this earth, just because it says in the Bible that God did not include gays in his original design. I read the Bible many times and I totally agree with Rob that if we do not follow everything there is written in it, like fasting and strictly following the ten commandments, then why should we tell other people that they don\'t have the right to marry just because it says so in the Bible. I really think there shouldn\'t even be such an issue. I want to reference slavery in what Jefferson wrote about the exclusion of Black people from the constitutional definition of \"men\" just based on the reasoning that Afr. Amer. are \"a different species,\" thus degrading \"a whole race of men from the rank in the scale of being which THEIR(?) Creator may perhaps have given them.\" Maybe gays should be excluded from the ranks of \"men\"?, they are sort of a different species too...  \"We are at war... with terrorism ... racism ... but most of all we are at war with ourselves.\"   '),(9918,1488330,'02,June,2004','2004-06-02 00:00:00','    I too find the impacts of global warming to be frightening.  I dont know how people still deny that it is a real threat.  If anyone wants more information on the topic, check out http://www.ucsusa.org/global_environment/archive/page.cfm?pageID=520  It has a summary of the IPPC\'s (Intergovernmental Panel on Climate Change) findings in a recent study.  It also has a link to the origenal report.  I found it to be very informative.  ~Joseph   '),(9919,1488330,'01,June,2004','2004-06-01 00:00:00','    Thomas, is there any way that you can invite me into the AP English blog?  That would be awesome!  My email is babster29@hotmail.com  Thanks!  Bethany   '),(9920,1488330,'01,June,2004','2004-06-01 00:00:00','    Rob, what you will be doing is unprecedented, unique, and first of its kind.(For that matter you can patent this ingenius idea.......JK.....or at the least claiming your originality of that idea as your intellectual property).   But seriously, if I interpreted your message correctly, I think core concept is to get more young people our age involved in politics, the political process, and especially how the political process works in a democratic society - with the funamentals manifested in the act of voting. We all realize the age group of 18-21 represent the lowest percentage of elligible voters that actually vote in an election. For some odd reasons young people our age just does not care about policies of our nation(Heh when there is a draft I bet a lot more people will vote then, but then again it\'s too late by then).   Youngesters, take your legitmiate part of control over the government!!!!!!!! We do not want the same old/senile/retired generals and shrewd politicians dictate our lives and draft us to be killed in some unknown regions of the world without our consent!!!!!!!!!!  I support the statement that those who do not vote does not deserve complaints.   THUS............GO VOTE, AT ANY OPPORUNITY YOU CAN.......AND THIS IS A GREAT OPPORTUNITY FOR ALL OF US WHO ARE NOT OLD ENOUGH TO VOTE IN THE NOVEMBER ELECTION A TASTE OF THE DEMOCRATIC PROCESS!!!!!!!  TAKE CONTROL OF YOUR YOUNG LIVES AND PROSPEROUS FUTURE THROUGH THE ACT OF VOTING!  And finally, do we have an end of year AP Party/Farewell to Ms.CLark Party soon?   '),(9921,1488330,'01,June,2004','2004-06-01 00:00:00','    Maybe it\'s because it\'s one in the morning, but I had a thought that intrigued me a lot.  I was thinking a little bit about how representative the ASB is of the studentsâ€” and it\'s basically trust on behalf of the student body that gives the governing body \"artistic license\" to run the school as it sees fit.  There are no referendums, and although meetings are open to the entire student body to sit in on, nobody ever does, or probably even knows it\'s an option.  In an effort to make my presidency more directly representative as well as continue my theme of the importance of democratic participation, I would like to have the school vote on how I should vote in the elections held this November.  It would be one vote representing the entire school, and it would sacrifice my own voice, but I think mock campaigning in school for candidates at all government levels would expand the knowledge and interest in politcs of the school; a heritage that would be far more valuable than the vote I would cast on my own.  With a fervor that hinges on administrative support, I would like to pursue this concept.  What does everyone think?  I want comments from both students and their parents on this one.  But not Mike Rush, he is an ignorant Pig-Dog who deserves no respect from anyone.  (Mike demanded I begin trash talking on the blog).  Secondly, I want to organize a large-scale trivial pursuit contest- who would be down to participate?  Trivial Pursuit will be the next big thing in this school and nobody wants to miss out on this.  May democracy live long and prosper.  Love Rob  P.S.  Create Blogger profiles.  They are kinda fun to do, and they give ME something to look at when I\'m up insomniacing.   '),(9922,1488330,'01,June,2004','2004-06-01 00:00:00','    Preface:  I know GM is a horse that died a long time ago and we continue to beat it, and for that I apologize, but having just bought a pair of Birkenstocks combined with defense of my weekle Bush mocking, I feel a little offended at Mary\'s harsh criticism of scientific thinkers.   Mary, I\'m not sure I get your stance on gay marriage.  You say you\'re morally against abortion but you think it should be legal.  Do you have the same position on GM?  I found the arguments you made a little contradictory.    First I want to say that I understand you respect everyone\'s right to an opinion, that is what you intended isn\'t it?  But for someone who doesn\'t want to come off as a right-wing bigot, you kind a come off as a right wing bigot when you challenge \"left-wing secularists\" who don\'t reference their morals for forming logical platforms based on fact.  I understand there is a fine line to walk between faith and fact, but I don\'t think a person has to choose one or the other.  The \"smarter people\" you refer to have a right to desire evidence to support the belief that God didn\'t intend homosexual\'s to marry.  If that evidence comes from the Bible, that is fine, but I guess what I\'d like is some flexibility in one\'s \"morals.\"  I, Left-wing Secularist, read your post and find myself questioning some of your statements you credit to your morality.  I, Left-wing Secularist, would like to know WHY you feel how you feel, where your morals come from, because I feel that I have evidence and \"imperical data\" to contest the morals you base your argument on.  And I, Left-wing Secularist, feel that I have analiable justification to disregard baseless morals if they don\'t coincide with mine.  With that said, can you satisfy my curiosity as to why you believe homosexuality is immoral?  I have my own theories, some of which I have expressed, but in the interest of originality, I\'ll only provide them upon your request.  Long story short, my biggest problem with the argument that homosexuality is immoral (whether right or wrong) is that it means homosexuality is WRONG and homosexuals are WRONG, spawning an attitude that is anti-homosexual which punishes those who had homosexuality thrust upon them whether by biology or development (there are studies that make strong cases for both).  Is that fair?  You say God didn\'t intend homosexuals to marry, then why were the created?  To be deliberately persecuted by faithful Christians?  Did God intend addiction, physical disability or car accidents?  I don\'t think that they were spontaneous developments while His holiness stood idly by.  I don\'t want to question your ideology, maybe just some minor parts, but it\'s only because I have imperical data from the holiest of books to back me up.  Post on the blog, e-mail me, or talk to me about it because I think it\'s an important discussion, but know that like Jesus, I\'ll always love you.  That goes for everyone  Love, Rob  P.S.  Clark\'s in Atlanta (I don\'t know why) then Seattle to prep for the next step in her career.  P.P.S  I only pay $6.25 for movies, where the hell is Frank going?   '),(9923,1488330,'01,June,2004','2004-06-01 00:00:00','    Global warming is a serious threat and I, like Bethany, am glad to see a new subject on the blog. The amount of damage humans have done, do, and will continue to do to the earth always amazes me. It seems that humans are extremely selfish. If humans continue to plague the earth with pollutants at this rate, it will inevitably lead to destruction of the Earth or the human race, whether it be from depletion of ozone or the melting of the icecaps. It sickens me that although we know this information we continue to pollute and plague the earth.  I also saw the movie Day After Tomorrow and was somewhat dissapointed but it was still an okay movie. I have never really liked Dennis Quaid (I think that is how you spell it), I think he wasn\'t good for the part. Anyway it\'s worth the 7.50 to see it. Mr. Parlee is a good teacher. He is very opinionated but, as Ryan and I were talking, we found it somewhat refreshing to hear actual political opinions by someone as educated in history and politics as he is. Usually teachers are hesitant about revealing their opinions because they don\'t want to offend anyone, but because Mr. Parlee is a substitute he is able to cross those boundries.  Lastly, does anyone know where Clark went? She told us but I forgot.   '),(9924,1488330,'01,June,2004','2004-06-01 00:00:00','    Rob, thank you for bringing up a different topic!  It was getting really repetitive in here!  I think that global warming is a really important topic that doesn\'t get enough credit.  Again, thanks to \"the man\" for leaving it alone and letting people forget that it does exist and could cause major problems in the future.    One of the issues is that scientists aren\'t getting funds to research plausible effects of the increase in carbon dioxide and carbon monoxide in the atmosphere, which has definitely increased over the last twenty years or so.  Just for everyone\'s knowledge, the movie \"The Day After Tomorrow\" is pretty much bogus, in terms of the causes of global warming.  The idea that the actual warming of the world would be immediately followed by a global ice age is pretty unrealistic.  But as long as it reminds people of global warming and the problems that it could eventually create, then I\'m all for it!But, anyways, not enough of \"the governments\" money goes into research for global warming and, while the process is irreversible, it would be worthwhile to research how we could slow down the process that we\'re now heavily involved in.  However, although the issue seems like it would be pretty un-controversial, there are many scientists who actually believe that global warming is not something that humans cause, that it is simply part of the world turning and getting older.    In the last century on the earth\'s surface (keep in mind that this is on the ground, not in the actual atmosphere), there have been increases of at three degrees Fahrenheit, which may not sound like much, but it\'s caused the sea level to rise from six inches to about three feet.  Of course the increase in sea level is caused by ice sheets melting from the increase in temperature.  It\'s all relevant!!!  And, if these changes had taken place over the last thousand years, it really wouldn\'t be that big of a deal.  Wow, the earth changes....what a concept!  But the fact that over only 100 years so many changes have taken place...well, it\'s worth getting into.  You guys, please respond to this one.  It\'s important, and we need to move on from the gay marriage and abortion topics.  I think that everyone has said what needs to be said, and it\'s time for something else.  Later!  Bethany   '),(9925,1488330,'27,July,2004','2004-07-27 00:00:00','    this place has been inactive for some time... ms. Clark was the first post in a month... wow... not much goin\' on. so, how\'s everybody\'s summer going? anyone want to share their AP grade.. i got a highly discraceful grade...a 2... =( and i thought i did exceptionally well on the MC and alright on the essays....   '),(9926,1488330,'23,July,2004','2004-07-23 00:00:00','    Hey students...I am settling in and have already started working.  Next week is my third week at Shorecrest.  I hope everyone is having a relaxing summer but I hope you are still all thinking and learning...there is no evidence of this by looking at our blog...did anyone see Farenheit 9-11...?  My new email address is stephanie.clark@shorelineschools.org.  I would love to hear how you did on the AP test.  Remember it is a difficult test and set up so  only 50% of the students paww with a three or better.  In addition only 8% GET A 5.  Also remember only the top scholars in the nation choose to take these tests...so pat yourself on the back for your effort and the amount you learned despite your score or regroup for a better effort next year.   For those of you who did well  and I am sure some of you did congratulations.  Rob-how is the ASB retreat going?  Enjoy the rest of your summer and I hope to hear from some of you.   '),(9927,1488330,'06,August,2004','2004-08-06 00:00:00','    I just finished my six weeks of intense math camp.........And a short two and half day ASB Leadership camp!!!!!!!!!! Wow...........Way too much to say............... The math camp has elevated my life to a new dimension, with numerous epiphanies occuring and many new friends and bonds made. And many valuable lessons I\'ve learned from math camp, including the fact that I felt incredibly stupid when I was immersed in a sea of of 23 geniuses.  And how important friendship and \"constructive\" competition is. And that\'s besides all the six week full time math we learned.  I miss the university setting soooooooooooooooooo incredibly much!!!!!!!!!!! I wish I could just be a college student and live in the McMahon dorm from now on for the next few years! Home only brings me the memories of stress, depression, loneliness.............and what not. While the math camp at the university was completely motivational, stress-free, fun......Fun together with learning/bonding experience.  Otherwise, the ASB retreat was mostly a taste of what we have to do next year, the responsibilities we have, the goals we have, and the people we are going to working with. It was quite interesting.........We\'ll see next year............Although I seriously wish Ms.Clark was still here!  Thomas Chen   '),(9928,1488330,'04,August,2004','2004-08-04 00:00:00','    Summer has been great,  i got a 3 on my AP test.  Hope everyone is doing well.  ~Joseph   '),(9929,1488330,'26,February,2004','2004-02-26 00:00:00','   Thomas, thanks for re-sending the invitation... I\'m doing a poem, not sure if I\'m gonna do Robert Frost, though.  He doesn\'t really have a particular style.  Some poems he does 10 syllables per line, some 9, some 7, etc...  But then, I could just combine elements of all of his poems...  If I was doing one of the essays, though, I would do the Robert Frost one.  It seems easier than all of the rest.  '),(9930,1488330,'23,February,2004','2004-02-23 00:00:00','   Hey Everyone: IS EVERYBODY EITHER DOING A POEM OR DOING THE ESSAY THAT TALKS ABOUT ROBERT FROST? Anybody doing any essay prompt besides that? or I\'m just such an outcast that I can find not a single accomplice to share ideas with  '),(9931,1488330,'23,February,2004','2004-02-23 00:00:00','   Hello Hello Hello, What did all of you people discuss about Robert Frost today? Any interesting interpretation and revelations, anything important things I missed today?.......  '),(9932,1488330,'19,February,2004','2004-02-19 00:00:00','   It wasn\'t too shabby.  But it was all about the Ralph Wiggum quotes! \"Me fail English?  That\'s unpossible!\"  '),(9933,1488330,'18,February,2004','2004-02-18 00:00:00','   Hey how did everyone think of that really \"interesting hands on\" activity we did on \"CELEBRATION?\" I guess it was a really powerful activity for our class (third period) to get all the people engaged. I celebrate the fact that we don\'t have to do that activity every day, only a little change for one day.  '),(9934,1488330,'18,February,2004','2004-02-18 00:00:00','   To me it seems like Whitman\'s going on a \"poetic rant\" of sorts on his thoughts of life, love, happiness, etc.  He\'s not necessarily praising himself as much as he\'s just praising his life in general.  Man, I could sure use a fat bag of the crack he was smoking right now...  ;-)  '),(9935,1488330,'17,February,2004','2004-02-17 00:00:00','   hey everyone, does anybody know what Whitman is talking about in his Poetry??????  '),(9936,1488330,'17,February,2004','2004-02-17 00:00:00','   Heh sorry Beck it\'s too late now. It will be helpful if we knew what you were reading.  Onto the new poetry \"Songs of Myself\", would anybody want to share their general impression or interpretation of what the poem is about?  '),(9937,1488330,'17,February,2004','2004-02-17 00:00:00','   Hello hello. Anyone want to do my Wread like a Writer for me?  Oh well.  '),(9938,1488330,'12,February,2004','2004-02-12 00:00:00','   I sent the blog invitation out to everyone. So please check your email to accept the blog invitation.  As of 2/12/04: These are the people who have accepted the invitation and are currently a blog member: Thomas Chen, Chad Wyszinski, Erin Riley, Kimber Graef, Corey Furin, Matt Goosen, and Tyson Cecka.  '),(9939,1488330,'03,February,2004','2004-02-03 00:00:00','   Hey everyone, this is a test to see if it works. If any of you have suggestions, i can change the templates and settings to suit everyone.   Also, does anyone have any good AP English or literature links I can put up on the website?  '),(9940,1488330,'31,March,2004','2004-03-31 00:00:00','   I have to say that I\'m not as interested in reading Night because I read it in eighth grade and it is a very emotional book that I would prefer to read once only because it is so sad.  However since it seems curriculum here is different and students here haven\'t yet read it, I suppose that\'s unfair of me.  I\'ll read it again if I have to.  '),(9941,1488330,'29,March,2004','2004-03-29 00:00:00','   -Hello AP English, this is Ryan Elliott and I\'m a rookie to Blogger.com.       I recall reading Night freshman year in Mrs. Micheel\'s English class, and the impression the story left on me brings to mind the narratives of Fredderick Douglas and Harriet Jacobs.  The accounts of the Holocaust enable the reader to experience the horror firsthand, and each individual number in the death totals suddenly forms a face.  At the beginning of the year I remember Mrs. VA stressing the importance of showing and not telling in writing, and I feel in many cases that history can tell a story, but it takes a first person account like Night or the narratives of Douglas and Jacobs to show the reader a story and make them live it.  Considering the current study of World War II in AP History, I think Night would be an excellent compliment.        The present study of World War II with Ms. Clark has made me feel disappointed in American diplomacy because of the decision to drop atom bombs on two Japanese cities.  When encountering the situation, I pretended I was a soldier in WWII, and I tried to think of how I would have felt about our actions.  Not only am I shocked at the civilian aimed tactics implemented globally during the war, but also I am troubled by America\'s decision to take part.  Personally, I would have difficulty enlisting in the military because of the possibility that I could be asked to carry out a mission like the one presented to Paul Tibbets, the pilot of the Enola Gay.  I believe war is a conflict between armed military soldiers, and does not involve intentional civilian casualties.  In today\'s military this death count is accounted for by the term collateral damage, which masks its true definition probably for the morale of soldiers and the maintenance public relations.  I can\'t imagine I would feel at peace pledging allegiance to a nation who was engaging in civilian warfare.  I love my country and would no doubt serve if conscripted; however, I feel the line in the pledge of allegiance, \"with liberty and justice for all\" extends to all people of the world, and that the American bombings of Hiroshima and Nagasaki were injustices to the Japanese people.       I believe pledging allegiance to the United States of America means promising one\'s devotion to preserve the liberty and justice of our nation while attempting to extend the basic courtesies to all who we come in contact with.  The invisible borders separating the people of the world are self created illusions, much like the concept of racial differences, that do nothing to obscure the sanctity of democratic rights or our responsibility to uphold them.  '),(9942,1488330,'29,March,2004','2004-03-29 00:00:00','   YES I\'M VERY MUCH INTERESTED! To me, historical literature are especially fun to read, especially when it compliments with what we are studying in history. And the holocaust is certainly a dramatic period of history and literture that would sprout out many ideas about morality, wars, politics and the whole theme of race and injustices. I may not be able to speak for the whole class, but it\'s certainly a great topic to discuss and read about! I\'m sure Tristan would be interested in it too! :)  And well, since Rob, you\'ve started the pledge of allegiance debate, I have already posted a huge long post about the pledge in the history blog, but I think it\'s relevant for me to post my argument on this blog too because it touches on the issue of individualism and what is to be an American (the whole themes we\'ve studied all year).  And don\'t blame me for writing about this thing for so long, but it is inevitable. Of course everyone knows that the 1st Amendments guarantees \"Congress shall make no law respecting an establishment of religion, or prohibiting the free exercise thereof; or abridging the freedom of speech.......\". And then one can debate FOREVER what does that mean, either pro pledge of allegiance, the under God phrase, or opposes it. Either way, both sides can site the constitution to argue for their sides. LOL that\'s the beauty of our first amendment, even tho it may be concise, it\'s still very interpretative. And the society is sooooooooo complex that a few words in the constitution can be argued forever into future generations.   As a former buddhist, and currently agnostic, to me personally, the \"Under God\" phrase doesn\'t offend me. But the act of conforming is. I have nothing against people saying or chanting Under God at anytime, anywhere, or under any occassion; the constitution guarantees their rights to do so. The thing that bothers me is the act of telling others what to do or not to do. How does that relate to the pledge of allegiance is the fact that little kid six years old have to say the pledge of allegiance. And did you know all the MEANING of the words in the pledge of allegiance when you were six? To me it\'s almost ridiculous to impose others to say what they don\'t even know what they are saying, and what it means......let alone believes in it. So the little kids just memorize the phrase and spew it out like robots. And that\'s not what America is all about.....  What makes America great and the beauty of this country is the fact that it\'s a country of immigrants, and EVERBODY is different, and the constitution really protects the DIFFERENCE of all individuals. In fact, individualism is not only protected, but glorified and encouraged all throughout American literature from the declaration of independence and onwards.   And from individualism comes with it, differences of ideas, clash of cultures, criticisms, and sometimes stuff that people don\'t want to hear.  What bothers me is when people stop wanting to hear complaints and other points of views. Any criticism, or other radical ideas becomes BAD and immediately shuts off from debates........And everybody has to praise everybody, praise the government, praise all the institutions, all the ideas and so on. And even if you don\'t, you are somehow peer pressured, or forced to comply with the \"majority\" point of view. Some of America\'s greatest writers were critics of the time, and often deemed radicals because of their foiling of the truth......truth that the majority didn\'t want to admit. Like Mark Twain\'s Huck Finn realism style writing that foils the true life of the South from a common southern white orphan boy\'s point of view. (Of course it\'s not just happy little child pacifer literature that\'s all happy and good. Because the society isn\'t always that good).   Seriously think about, what a society would be like if everyone is the same, think the same way, and all are happy with each other with no reason. Everybody would just lose their personality and the society itself will become dumber.   And education is really when the clash between conformity and individualism occurs. In order to preserve the individualisms as America\'s embodiment, I believe that in schools, since it\'s a public place, the state or government should have no say in what the kids should say. The kids are just kids, they are learning and developing.........And wheh pledge of allegiance is inserted to thehir curriculum since their first day in school.....And have memorize something they don\'t even know what it is......it totally defeats the American character. Education is to develop people to become individuals, to become smart citizens with an critical eye, and not to turn everybody into robotic conformists.   And as always, I welcome all dissagreements, criticisms, opinions and would like to debate this issue with others.  Differences of opinions is good!  Thomas Chen  '),(9943,1488330,'29,March,2004','2004-03-29 00:00:00','   Hmmm...I\'m starting to think ahead to curriculum after spring break, when we will have time to read another book and also to prepare for the AP exam.  I know from talking with Miss  Clark that her classes have recently been spending time studying the Holocaust, and I\'m thinking about Kailie\'s expression of frustration.  Why didn\'t Jim just run?  A similar question is often, \"Why didn\'t the Jews just band together and keep the Nazis from treating them so poorly.\"  I\'m wondering...are you folks interested in reading a first-hand account of Auschwitz after break?  The book I\'m thinking of is Night, written beautifully by Elie Wiesel.  It\'s a quick read, but it\'s also a hard read because of the subject matter.  It isn\'t American lit (Wiesel was Romanian, I believe), but that doesn\'t really matter for the test.  What do you all think?  Would it be best to be presented with a variety of options before deciding?  As for your question about the Pledge of Allegiance, Rob...I find myself asking many more questions in addition to the one about the \"under God\" phrase.  For example, what does it mean to pledge allegiance to a flag at a time when patriotism is so integrated with war-making?  I cannot help but remember photos I\'ve seen of Nazi swastika flags hanging over church altars during World War II.  As a religious person, I want to see a full separation of church and state so that I am left to practice my faith as I see fit.  I believe others have that right, too.  Although I believe that we exist by the grace of God, not everyone does.    So what does our flag stand for?  When we pledge allegiance to it, what are we promising to do or not do?  What is the role of God in a pluralistic country?  What binds us together as a people, particularly as we become a more diverse society every year?  Does pledging allegiance mean committing ourselves to creative peace-making, or only that we will pick up arms if attacked?  And what about \"justice for all\"?  These are the questions I ask myself every day at the beginning of third period.  What do you all think? Mrs. V-A  '),(9944,1488330,'29,March,2004','2004-03-29 00:00:00','   First of all, let me start off by saying that I have no problem with mandatory repetition of the Pledge of Allegiance because it stands for becoming INDIVISIBLE in an effort against conformity...promotes LIBERTY- the liberty to think, act, and live without government intervention...as well as JUSTICE FOR ALL- the permission for everyone to live as they please.  It may be ironic that there is a uniform method of promoting the liberties that lead to diversity, but if everyone is made different through the same process, are they not still different?  I recognize the recitation of the Pledge alludes to military service as flags are used to determine a soldier\'s allegiance in war, but I also believe that the Pledge is more effective with a visual aid.  The debate today is whether or not \"under GOD\" should or even COULD be allowed to stay in the Pledge.  I outlined the current particulars of today\'s case in a previous post, if you haven\'t read them.  To me, this is a case defining the separation of church and state.  Although this country was outlined around that seperation, there is nothing that states that government can\'t recognize the ideology of monotheism, or other belief structure, as long as it doesn\'t allow the government of religion to directly affect it.  One could argue that excluding the possibility of a deity\'s existence would be not seperating the state from atheist ideology.  Tom C addressed free speach...that each American has the right to say what they want and worship as they please.  With compliments to that right, many Americans feel the direct connection between Providence and the success of this country, and that excluding their right to express His influence in relation to the country would antithesize their beliefs.  The same issue is currently being presented in the Ninth Circuit Court of Appeals, as an Atheist man argues that with each (socially) mandatory oration of the Pledge, his daughter is taught that her father\'s beliefs are the wrong ones.  As for me, I believe that eliminating it from the Pledge restricts the large portion of Americans who recognize the Glory of God as parallel with the liberties of this country.  For those who are polytheistic, \"under god\" can be translated as recognizing a \"higher power\" than is visible in daily life.  Lastly, for those that don\'t recognize any supernatural force or deity, \"under God\" may be substituted with a respectful pause while those who do choose to recognize God do so.  My views have been presented, but I\'m all about becoming as educated as I can on the attitudes toward the aspects of life that effect the way we live every day... so school me!  If \"Night\" is a good read, then I\'m all for it...especially if it\'s short!  Also, are Rip Van Winkle, The Grapes of Wrath, or The Catcher in the Rye worth reading over break?  Because they all spark my interest.  Have a great break, and the fabulous assembly ASB will be presenting Friday morning.  Love,  Rob  '),(9945,1488330,'29,March,2004','2004-03-29 00:00:00','   In response to Kailie, I think your point about how Jim\'s not afraid of enduring through all of these ridiculous rituals imposed by Tom (If I interpreted you correctly), illustrates Jim\'s complex charactor and heroic quality. Jim, as a slave, probably has a really low self-esteem and he comes from the world where slaves have to listen to whites for every little nuts and bolts. Of course Jim did not know he was free, (any slave would think it\'s a miracle if somebody proclaimed they were free). So he just thought \"this is the way everything is done, because Tom said so\".  Even though he has the brain to mention it\'s a ridiculous idea like making up a coat of arm, but he still follows the order, because who knows what will happen to him if a slave doesnt obey a white man\'s order.  But I understand where you come from. I also felt very weird these things were happening, but lots of weird things happen back then, a lot of things today would consider very very sick and unbearble......So in trusting Twain\'s realism style of writing, I have to just interpret what I can about the society from Twain\'s words.  '),(9946,1488330,'28,March,2004','2004-03-28 00:00:00','   I was really frustrated when reading the end of the book when Tom and Huck were \"rescuing\" Jim. I couldn\'t help but get past the idea that, if Jim were really worried about getting away from slavery and afraid for his life, he wouldn\'t jump through all the hoops for Tom\'s perfect scheme. I felt it was a little unrealistic; it\'s natural instinct to run when given the chance -- just like if a dog were to get free, he wouldn\'t sit around and wait for the perfect getaway. He\'d just get out of there! Did anyone else feel anxious while reading this part in the book, or is it just me?  '),(9947,1488330,'28,March,2004','2004-03-28 00:00:00','   Wowzers. I\'m on the blog! Cool.......Thanks for the blog Thomas. Anywho, hope this can be a good spot to swap thoughts. Adios and God bless.                                                            Yeehawman                                                              (Tristan)  '),(9948,1488330,'26,March,2004','2004-03-26 00:00:00','   Knowing what Jim had to go through and then at the end for him to just give it all up for the sake of Tom, proves Jim to be the more significant hero.  Throughout their whole journey Jim was always worried about Huck and concentrated on his friendship with Huck.  Whenever they were separted Jim would search forever for Huck until he found where he was.  They always ended up reuniting due mostly to Jim\'s love for Huck.  Not until the end of the book did Huck ever truly try to find Jim.  However at the end of the book Huck\'s love for Jim became more prevelant. Although Huck knew better than to do all the crazy things that Tom insisted they do in order to free Jim, somehow Huck\'s love for Jim was submissive to Tom\'s influence over Huck.  After trying so hard to escape to freedom then putting up with Tom\'s ideas of freeing a slave, Jim was still willing to risk it all for the sake of saving Tom.  This is what a true hero is.  Huck didn\'t have anything to lose, if people found out he had been trying to free a slave they would send him back to the widow\'s house and people would just think \"that poor boy needs to be civilized.\" However for Jim, he had everything to lose, and he was willing to lose everything for Tom, who wasn\'t even a true friend to Jim.  '),(9949,1488330,'26,March,2004','2004-03-26 00:00:00','   While both Huck and Jim can be seen as heroes, Jim is the more siginificant hero of the two because of not only the racism he faces by being a slave, but because he can overcome this racism by saving Tom in the end of the book.  In Tom\'s cruel and \"adventurous\" plan to help Jim escape to freedom puts Jim in a place where he is locked in a cabin, when he could be free.  Jim doesn\'t object to this plan because he figures that Tom is white and thus knows what he is doing.  After putting Jim through a month of being locked up, Tom does manage to help Jim escape.(even though Tom is shot in the leg)  This is the point that seperates Jim\'s heroism from Huck\'s.....Jim throws away the freedom that he has endured a month\'s \"cabin fever\" for to save Tom that has just put him through this month of isolation.  '),(9950,1488330,'26,March,2004','2004-03-26 00:00:00','   In my opinion, Huck and Jim are both heros in this novel.  Huck is a hero in a sense that he decides to go to hell to save Jim, which was a huge decision to make at such a time in the South.  Later in the novel, Jim becomes the hero when he sacrifices his freedom (or so he thinks), the freedom that Huck has worked to protect, to save Tom\'s life.  Both Huck and Jim have their heroic moments by making huge moral decisions.  They both have to decide whether or not to do the right thing, and the fact that they both end up doing what\'s right makes them both heros.  Sorry, that made no sense...talking about stuff like this confuses me sometimes; I know what I want to say, but it just doesn\'t come out right!  '),(9951,1488330,'25,March,2004','2004-03-25 00:00:00','   Ok.......Everyone......We now have reached the 20th person mark with Rob Muilenburge being the 20th blog member (although probably the 5th person who\'s published anything on this blog). Oh yeah, If any of you have trouble finding your latest written post on the web, it\'s because you did not published it; it\'s only saved ready to publish. So you have to click on one of the icons on top you\'ll see that says \"publishing\", and click on \"republish latest\". Then you will see your nice happy post on the web.........I\'m not sure if there are some remote posts posted by people long long time ago that never got onto the website because somebody didn\'t publish it...........yeah so be sure to remember to publish everytime you write a post......And enjoy everyone.  wow, what a well constructed argument there Rob.......Now there has commenced a real discussion. Yeah I\'ll get back to your comments soon on a later post.  '),(9952,1488330,'25,March,2004','2004-03-25 00:00:00','   Hey guys, you can add one more to the total # of bloggers, making 19 i think?  Anyway, addressing Tom\'s point from the essay that Twain was writing for a racist audience doesn\'t sit well with me.  I\'m not disagreeing that there were (and still are) many racists in this country, and probably some of them were fans of the original, Tom Sawyer, but I just don\'t get the feeling Twain wrote for an audience.  He was already a very successful author who had stopped needing to write for money, so I don\'t think the bottom line was his issue.  Every example of Twain\'s writing I\'ve come across challenges cultural norms.  Whether addressing Christianity, poverty, or racism, he attacks each issue with the underestimated weapon of satire.  Flooding the tongue-in-cheek essays and novels with his own opinions, no matter how controversial.  I don\'t believe that Twain could be writing for a racist audience while openly attacking the Ku Klux Klan, and the entire establishment of Christianity.  If Twain avoided openly expressing Huck\'s love for Jim to the world, it was to get around his own nagging imperfections in his relations with blacks.  I don\'t believe that this was as much of a concious effort as Miss Morison claims.  Sometimes things are better off left to the imagination, e.g. the \"humbles himself to a nigger\" sentence is much more effective in my eyes than, \"I\'m sorry Jim, I didn\'t mean to worry you.\"  Lastly, although Morison\'s essay was very well articulated and worthy of consideration, it was very much speculation, just as I am writing now. I look forward to posting often and lengthily in this blog, so you\'ve been warned.  By the way, the appeal to the removing \"under god\" from the pledge of allegiance began this week.  I\'d love to hear everyone\'s thoughts on that, as well as on Huck and the recent essay.  Love, Rob  '),(9953,1488330,'25,March,2004','2004-03-25 00:00:00','   Hello!  My assumption is that extra credit will motivate many of you to access this site this weekend, so I thought I would throw out some questions we didn\'t have time for in class.    Who is the hero of the novel?  Huck or Jim?  What are the most significant themes from the book that affected you personally?    What about Toni Morrison\'s question:  \"Yet the larger question, the danger that sifts fr om the novel\'s last page, si whether Huck, minus Jim, will be able to stay those three monsters as he enters the \'territory.\'  Will that undefined space, so falsely imagined as \"open,\" be free of social chaos, personal morbidity, and further moral complications embedded in adulthood and citizenship?  Will it be free not only of nightmare fathers but of dream fathers too?\"  Have fun!  Remember to study for Monday\'s test.   :-)  Mrs. V-A  '),(9954,1488330,'25,March,2004','2004-03-25 00:00:00','   HA....What I found so interested while reading through this essay was the fact that even tho Mark Twain might be a staunch abolitionists....For racial equality during that time period........But he also has to sell the book, make the book noteworthy.....And so Twain has to insert a lot of audience-awareness into his novel.......And since he knows a lot of the readers at the time are rascists......Therefore he makes Huck sometimes very uneasy to comment how good Jim is......Even tho Huck probably knows deep inside that Jim is a very good person......But he fails to admit it to anybody else........Even fail to admit fully to Jim and himself Huck\'s true feelings.  And of course Twain probably didn\'t know the future status of blacks/slaves, and all is very uncertain........So he had no choice but to retain some of the old stereotypical qualities.........Only today can we see it from an objective point of view.  '),(9955,1488330,'24,March,2004','2004-03-24 00:00:00','   Don\'t blame me everyone..........but i\'ve just thought of a great idea to implement for this blog......There are those \"Open-ended\" free response questions in the AP English exam (usually one a year), that has an open prompt and one has to write a mini expository essay about it, from your own experience.  And last year\'s was:   urlLink  \"entertainment has the capacity to ruin society\"  , defend, challenge, or qualify this assertion. Soooooooooo any cool ideas/agree/dissagree viewpoints on this statement.  '),(9956,1488330,'23,March,2004','2004-03-23 00:00:00','   *crickets chirping*  So, to break the silence:  If you/Huck Finn/your milieu/a good heart and corrupt conscience/a stereotype were a hot dog . . .  '),(9957,1488330,'22,March,2004','2004-03-22 00:00:00','   Ok, WOWWWWWWWWW NOBODY HAS BEEN USING THE BLOG FOR SOOOOOOOOOO LONG............WHAT AN EERIE SILENCE HERE..........I\'M GONNA HAVE TO RENAME THIS BLOG \"DEAD BLOG\" NOW.......... But as if TODAY, Frank Kanekoa and Ryan Elliot have joined.....That brings our total to 18 people.........And there has not even being 18 total posts so far. Anyways..............Just added a new link to our blog.......That\'s Mr.Wargo\'s AP english blog. Wow what an interesting class that is......For some odd reason, they are forming an international collaborative English class.   urlLink Wargo\'s AP English Junior English Class Blog  '),(9958,1488330,'04,March,2004','2004-03-04 00:00:00','   Well our blog community now has 16 members: Beck Nelson, Thomas Chen, Forrest Vines, Chad Wyszinski, Kailie Middleton, Alexandra McDougall, Ryan Vasbinder, Corey Furin, Paul Fleurdelys, Nova Martin, Erin Riley, Kimber Graef, Megan Valencia, Tyson Cecka, Matt Goosen, and Bethany Graef.   And sorry for some of you who failed to sign up to this blog because of a messed up email problem. If you still want to post, you can share the account with one of your friends.  '),(9959,1488330,'02,March,2004','2004-03-02 00:00:00','   hey thomas, thanks for re-inviting me!  and i totally think that we had a really great class discussion, though it would be nice to get everyone involved.  some people are still pretty quiet.  '),(9960,1488330,'02,March,2004','2004-03-02 00:00:00','   It was an astounding discussion today! I\'m surprised at our third period class today, we actually did some genuine discussions! Without more than five minutes of silent stares at each other! That\'s a huge benchmark for us. LOL.   Ok Nova, we may have a different stance on this issue about the discussion of \"racism\". I think the controversy of it is really really interesting, and really relevant to the society we live in today. Maybe whining about it doesn\'t help, and only makes the matter worse. But one has to study the history behind racism, the history behind slavery, and study the current society in general to get all these controversies and see that these debates are very interesting. If we are living in a place like Chicago, Los Angeles, NYC, San Francisco, or Atlanta, we might see the issue and controversies surrounding the issues of race and racism in a very different way, because the societies of these urban centers are so polygot that you CANNOT ignore the issue of race at all. And it becomes so relevant to your everyday lives, whether you like it or not. So I don\'t think we can make a judgement on how the issue of \"racism\" is sooooo out of date and dumb now. To a lot of people, it\'s very relevant to their lives. And reading this book Huck Finn I\'m sure will help us understand better the deep history behind slavery, race relationships between whites and blacks, that still influences our thinkings today. And I agree with today\'s discussion just banning the N word will not solve any any racial problem at all, not even down to the scale of molecules. There are a lot of other forms of racial discriminatory actions that we can not ban because of free speech issue such as Erin\'s description of all the confederate flags in the south still. And for christ\'s sake it\'s a piece of literature, one of the best pieces of American lit every written. And to the issue of calling people \"whites\", \"blacks\", \"African Americans\" whatever, I agree with the discussions today that it\'s a matter of respect. Some people maybe offended by calling \"black\", others maybe \"African Americans\", so it. But I agree that in general, we just have to see all people as one group of Americans, and not place any body into a race or cultural profile. --------- So, what do you think of that?  '),(9961,1488330,'01,March,2004','2004-03-01 00:00:00','   The entire racism argument makes me mad.  Do people not get that Twain was anti-racist?  That he was trying to make a statement with Huck Finn?  And also, the \"n-word\" was used back then, and it was the common word for black people.  Making all the poor white people of the south talk in a more socially polite way would have been defeating a major point of the book, realism.  100 years from now, what if \"black\" is not politically correct?  Will they ban all the books written in present day, to censor \"bad words\"?  These people make me so mad!!!  '),(9962,1488330,'01,March,2004','2004-03-01 00:00:00','   I think the people who are against this book for its portrayl of blacks need to look at how it shows whites. It shows cruel whites, white drunks, white racists, whites who steal, lie, cheat... so how does Mark Twain show black people in a negative light, when the majority of his criticisms are demonstrated through his white characters?  '),(9963,1488330,'01,March,2004','2004-03-01 00:00:00','   I think we should have a discussion on Huck Finn and the controversy, it seems very interesting. I read this particular interesting thing in the passage that implied you need to be a \"good teacher\" , to actually use all the positive aspects of Huck Finn and teach more about the truth of slavery. One needs to teach Huck Finn in conjuction to the historical period behind it. An \"insensitive teacher\" may allow the racial impacts the word has on the students in the classroom goes unchecked, and allow students to \"read aloud\" in the classrooms.  But \"on the contrary, studies have shown that, in the hands of a good teacher, the novel can actually have a positive impact on racial attitudes and increase in interracial understanding. I agree with the fact that if you don\'t see the true picture of that period in history, to see the life of a \"common man\", and use the language they use back then, one can not grasp a true concept of. And of course all of this have to be carefully studied in conjuction to a great understanding of the history of slavery and slave trading. That\'s why having the book in elementary and middle school is probably not a good idea. Because the kids then can\'t get ironies, satires, and the historical behind why Twain wrote this book and used the bad words to portrait the truth.  '),(9964,1488330,'30,April,2004','2004-04-30 00:00:00','   Well, I would have gotten the first question right...  Anyway, Natalya, Tom suggested I talk to you about taking his place in the district MUN coming up in May.  I have the basic understanding of what goes on, but if you would like me to join you, you can e-mail at attakid117@attbi.com or call me at 574-1916 with info (specifically what I would need to do to prepare).  Hope to hear from you soon...  And to everyone else, Good Luck Monday!  Love, Rob  '),(9965,1488330,'29,April,2004','2004-04-29 00:00:00','   Sorry, but I had to post this poem I found in one of the practice tests I have been taking. I really like this test because of the fact that a lot of the experts that they give for analysis are actually really interesting to interpret and are great works of literature:  Virtue George Herbert (1593-1633)  Sweet day, so cool, so calm, so bright,      The bridal of the earth and sky; The dew shall weep thy fall tonight;      For thou must dies. Sweet rose, whose hue, angry and brave,      Bids the rash gazer wipe his eye: Thy root is ever in its grave,      And thou must die. Sweet spring, full of sweet days and roses,     A box where sweets compacted lie; My music shows ye have your closes,     And all must die. Only a sweet and virtuous soul,     Like seasoned timber, never gives; But though the whole world turn to coal,     Then chiefly lives.   Just a couple of questions for practice:)  The subject of the poem is: a) mortality, b) immortality, c) the beauty of nature, d) the impact of evil,      e) the essence of goodness.  The speaker conceives of the world as: a) ending in ashes, b) repeating the cycle of birth and death,       c) having the strength of \"seasoned timber,\" d) eternal, e) being the secure dwelling of humanity  I am looking forward to frisbee tomorrow; I have to agree with Frank that sometimes it is just better to simply relax and take your mind off all the fidgeting and worries. By the way the answers to the questions are: e and c.  '),(9966,1488330,'28,April,2004','2004-04-28 00:00:00','   All this death and gloom over the AP test, we all need some happiness... Well, okay, i\'m no good at being happy, so here is some dark sarcasm. Good enough?   Bush To Iraqi Militants: \'Please Stop Bringing It On\'  WASHINGTON, DCâ€”In an internationally televised statement Monday, President Bush modified a July 2003 challenge to Iraqi militants attacking U.S. forces. \"Terrorists, Saddam loyalists, and anti-American insurgents: Please stop bringing it on now,\" Bush said at a Monday press conference. \"Nine months and 500 U.S. casualties ago, I may have invited y\'all to bring it on, but as of today, I formally rescind that statement. I would officially like for you to step back.\" The president added that the \"it\" Iraqis should stop bringing includes gunfire, bombings, grenade attacks, and suicide missions of all types.  Props to the Onion for that little piece. My two cents on the AP test?    AHHHHH. Now, were did all those grammar/rhetorical thingy lists go to?  --Beck--  '),(9967,1488330,'27,April,2004','2004-04-27 00:00:00','   I just took the collegeboard test and I have to agree that it is much, much easier than the one we took in class, at least my score went up by 10 points:) I agree that literary terms are crucial to know both for mult. choice and essays, but grammer is also really important, they have a few questions like:  lines ... contain which of the following?: a) elaborate metaphor b) parallel syntax c) a single periodic sentence d) a compound subject e) subordinate clauses I am currently looking through all the grammer packets Mrs. VA gave us in the beginning of the year. Yeah, I am kind of starting to feel a bit nervous already.  P.S. The evil that men do lives after them, the good is oft intered with their bones. (not sure if that\'s how it goes:)  '),(9968,1488330,'27,April,2004','2004-04-27 00:00:00','   Friends, English Classmen, lend me your ears.  It is true what you say... the test is subjective, subtle, and DIFFICULT.  But you are more prepared than you think.  By now we can all write 9-caliber essays with the right inspiration, and the multiple choice is less perilous than you give it credit for.  The tests we\'ve taken from the books are meant to trick you- beat you on a technicality and so forth.  So much so that the objective (I hope) is to get you to read the questions so closely, you don\'t get fooled on the test.  However, there is evidence for hope.  The test that I gave a link to in my last post, was a sample test written by the college board... after a 50% on the Peterson\'s test, I got an 80% on the real one... that\'s a huge improvement.  It\'s also great for confidence.  We can do this, class!!!  The best advice I can give is to know the literary terms like the back of your hand... what they are, how they\'re used, and how you can recognize them in a passage.  That\'s the heart of these questions, and if you can identify those terms, you will have success.  Love, Rob  '),(9969,1488330,'27,April,2004','2004-04-27 00:00:00','   My only suggestion to those who want to \"study\" for AP English test is to take as many practice multiple choice questions as possible, and after you took a couple hundred of them you will probably see some patterns emerge; at least that\'s what I am planning on doing for the next 5 days. The only problem that I see is that all the practice tests written either by Petersons or Barrons are much harder and some of the questions are totally different from those on the actual AP test, so if you take their practice tests you will probably get used to those author\'s patterns of thinking and the type of questions they ask and what answers they are looking for, yet it\'s still a good practice: it helps you to try and think like the writer of the questions and also pace yourself. Last year I was taking practice world history tests and so at the AP test all the questions seemed much easier to answer and I answered all of them in time.   I have to agree though with Thomas that AP English is hard becuase of its subjective questions, I am myself almost never 100% sure on answers, and often those that I was sure of end up being wrong. But I agree with Eren also that essays make up for the harsheness of mult. choice; especially after Mrs. VA gave us the rubric it became much clearer to me what exactly I should focus on including in my essays.   So just practice a few more hundred mult. choice questions;), and review grammar and literary terms.  My favorite questions on our last practice test: 26. What is the meaning of the word \"object\" (line 85)? a) Feel b) Dispute c) Argue d) Silence e) Save I think Rob was the one to point out the fact that \"dispute\" and \"argue\" are too similar to see that much of a difference between the two.  '),(9970,1488330,'27,April,2004','2004-04-27 00:00:00','   To be perfectly honest, i dont really see how you can study for the AP test. I mean, of course you can review grammar and rhetorical strategies and how they effect one\'s writing, but i dont think it\'s worth getting overly stressed. Be confident in the knowledge you already have; we\'ve all been preparing for this since the first day we signed up for AP -- just do you best. For me, I\'m planning on going in on Monday with what I know...that\'s all we really can do. You know what you know, and as long as you try your best, that\'s all you can do. So dont stress out too bad. Just think, by noon on Monday, it will all be over!  Kailie  '),(9971,1488330,'27,April,2004','2004-04-27 00:00:00','   I have to disagree Thomas.  The AP English test is definitely tough, but I don\'t think its the toughest.  While a lot of the multiple choice questions are subjective, the essays make up for it.  If you write good strong essays and support your points, you should (hopefully!) do well.  I think people just find the test to be hard because its not something you can memorize out of a book; its skills for recognizing rhetorical strategies and writing well that count.  '),(9972,1488330,'27,April,2004','2004-04-27 00:00:00','   Ok maybe it\'s just me, but I would rank AP English tests the hardest tests.   The multiple choice questions on the AP English tests are harder than SAT passages (which i already did horribily), not only are we suppose to know the meanings of the passage, but exactly and thoroughly understand how authors wrote it and carefully constructed the passages.  There are several reasons for this ranking that puts English on the APEX of all AP tests:  1: The questions are the most subjective: Everytime when I take the AP English multiple choice tests, for every MC, I feel my hands sweat and my mind full of doubt when I bubble in an answer for that question. There is NEVER a single question that I can feel 100% confident on that I\'m sure I got it right. Unlike all other subjects, including US history, where there is just some straightfoward concrete fact that\'s either right or wrong.  The meaning of every excerpt, sentence, even to the every word of the question can be interpreted in a variety of different ways. And as Chad mentioned today in class, that we are taught so much about how to infer, imply, that we sometimes forget the LITERAL MEANINGS, and we lose perspectives. In short, every question we can \"Overthink\" it.  And again being the subject of English, there is no fine line as to where is \"overthinking\", and where is \"underthinking\". To a point, it gets to be rather annoying, all the shakiness and uncertaintiness in answer EVERY SINGLE QUESTION.  2: When you get an answer wrong, and you read the explanations that explains why, you are never ever 100% convinced. There is always that self-righteous factor that comes to play that doubts the correct answers, simply because words and phrases can be interpreted differently in everybody\'s minds. Everytime I feel the urge to argue with whoever thought the right answer to a question is this letter, while two other letter of answers can probably be interpreted as correct.   3: For almost every other school subject, the teacher, the student, if he or she is willing to learn or improve, can easily identify the weak areas, which FACTS the student\'s missing and so on after some assessments......But for English, I just  know I\'m not prepared for the test, and I know did bad on the practice ones, but still there is not some concrete evidences I can point out that here, this is my weak area, or this I really do not understand. So it becomes a matter of trial and error and practice, which seems limitless and endless.   And Finally, this is the most unique about the English test that\'s different than EVERY OTHER SUBJECT: ---- YOU CAN NOT CRAM AND STUDY FOR . Yeah it is right that after maybe several, maybe several dozens years of intense study.....You maybe say I\'m fully prepared now for the AP english exam........And still the questions bombards you with subjectivity. And just the shear size of the vocabulary, the subtle difference between all the synonymous words, (overstatement, hyperbolie, implies, infers, suggests, what is the intent, argue, dispute...). And it\'s like your whole life of culture and experience in the world culminates and contributes to your performance on this test. Yeah for a really really stressed, miserably failing, troubled, an extremely unprepared and uncertained kid one weekend before he has to take the AP English language test, This test is just not the test to cram for.  All in all, AP English test in my feeling is just hard.......harder than every other test.   Unless you are Matt Wu of course.  I don\'t know if any of you share any of the feelings I listed above.   So there is my emotional turmoil of the day.  Thanks all to all of you who were willing to read this. Thomas Chen  '),(9973,1488330,'26,April,2004','2004-04-26 00:00:00','   Hey, for those of you feeling stressed about the AP test next Monday as a result of some negative test scores on the Peterson\'s practice we took, there is a College Board written practice at   http://www.collegeboard.com/student/testing/ap/sub_englang.html  You may have to download a new Acrobat reader, but this should be a slightly more accurate prediction of your success than the Devil\'s own instrument of torture; The Peterson Test.  Good luck to everyone on this practice test, remember to take it within the alotted time, it is good practice!  Love Rob  '),(9974,1488330,'24,April,2004','2004-04-24 00:00:00','   Thomas, I totally agree with you that standardized tests are simply unfair/wrong since individuals are not standard. It seems like in America everything is about individuality and ability to strive and achieve, yet society still places standards upon us that we have to pass in order to be \"good citizens.\" I\'m myself very frustrated with SATs and ACTs cause no matter if I am â€œsuper smart,â€ or get 1st places in competitions, I am just bad in analogies and my vocab doesn\'t include such words as vituperative, or flabbergasted. I personally just came from the WASMUN competition where I was elected Best Negotiator, and Rapatour (I don\'t know how to spell that) and spoke in front of 600 \"delegates,\" and yet all of that doesn\'t matter cause I still feel \"not good enough\" cause my awesome SAT scores place me into the \"average/standard\" herd of students.  Actually I think it\'s an awesome idea to encourage our fellow students to participate in conversations by addressing them directly, it really does work... sometimes. In our \"committees\" in MUN (yeah, I am a nerd like that) we can just say \"I yield my time to ....\" and name a delegate that seems to be the least awake; it mostly ends up that he/she has a lot to say and starts participating in the debate, or else you at least you wake them upâ€¦ for a while:)   About the practice AP test we took, I find it to be the best preparation for the test we can get. I just want to know what I would actually have gotten if it was the real test. I found mult. choice test to be the hardest part of it (stand. tests), the essays were fine. The prompt about Emerson was interesting, especially since we have studied him and read his works; I didn\'t focus as much on conventions, except for the usage of simple/complex sentences and the \"...\" to communicate effectively. Shakespeare essay was awesome cause I just have read a lot of him and the prompt was pretty broad and open for interpretation. It helps a lot to have background info about prompts. Now I think we should just really focus on reviewing the literary terms and grammar, I have problems coming up with literary terms, and also some of the mult. choice questions on grammar need review. I think our class is well prepared, we all read and write a lot, and Mrs. VA covered grammar/terms pretty well.  '),(9975,1488330,'23,April,2004','2004-04-23 00:00:00','   I feel this is a great article that everybody should read:  urlLink   Segregation no longer black and white New issues fifty years after Brown v. Board decision  It supports some educator\'s arguments that today\'s schools are actually as segregated or even more segregated than ever. Segregated for a variety of reasons: huge discrepancy of wealth, economic gap, culture differences, community separations, and continued racial problems and misunderstandings. This culturally, economically, and racially segregated school systems in turn shows up in the huge differences in the \"standardized\" testing scores. This problem really boggles those \"No Child Left Behind\" testing conformist\'s mind and is a threat to their \"everybody conform to the same standard\" ideology.  I would like to hear everyone\'s arguments and opinions on this issue.  Thomas Chen  '),(9976,1488330,'23,April,2004','2004-04-23 00:00:00','   Mrs. VA,  I think trying to get more people involved in the discussions is a good idea.  However, knowing what it\'s like to be one of the ones that is not easily motivated to state my opinion during class, I feel that the best way would be to encourage the other students to try to bring the less responsive ones into the discussion.  Having a teacher pinpoint you in the middle of a discussion never seems to work too well. Throughout middle school and other classes many teachers have used that as a technique as a form of punishment for not paying attention, but for our class, many people are paying attention they just don\'t say as much.  Therefore if it\'s the other students it wont be as much of punishment, but should bring about more of a casual converstion which will allow everyone to feel more comfortable.  '),(9977,1488330,'22,April,2004','2004-04-22 00:00:00','   As far as I know, Kailie Veys had the highest multiple choice score with an adjusted (incorrect guess penalty included) score of 23/43.  Sorry if I betrayed her trust in this, but I\'m proud of her accomplishment.  On to class participation; I don\'t think bribing responses is the way to go.  Everyone comes out of their shell, some of us are fortunate to have expatriated earlier, but sometimes forcing someone to respond only delays their comfort.  The best possible scenario is to create an environment where the shy can feel that their thoughts and feelings will be appreciated.  Many people have expressed to me that what they have to say they don\'t feel other people want to hear; that their thoughts aren\'t on the same level as the \"intellectuals\"; that they feel stupid with what they have to say. While there isn\'t anything that can be done about those angsts, for those that don\'t feel confortable contributing; know that you needn\'t worry or feel judged.  Anything that can be contributed adds to the formation of other thoughts and truely helps a discussion.  Lastly; some of those questions in those Peterson tests are so damn frustrating and often times ridiculous.  There should never be a scenario where one must choose between \"argue\" and \"dispute.\"  They\'re the same thing!!!  In history, the Fed either \"raised interest rates\" or \"discouraged borrowing.\"  All the Fed does is control interest rates and inflation, so the answers are the same!!!  Traditionally, the practice tests are harder thant he actual tests.  For a potentially more accurate representation of your score, you can take old tests that were prepared by the college board, accessible at www.collegeboard.com.  Find the AP section, and we\'re taking AP Lang, not Lit, so be wary.  Once again, good luck on the test!  Love, Rob  '),(9978,1488330,'22,April,2004','2004-04-22 00:00:00','   Mrs. VA, I absolutely agree that it\'s time to use some forceful means to rally everybody into sharing their ideas and speak up in class. That\'s actually a great idea actually of setting up a ground rule for everybody to speak, except that I\'m not that kind of person who wants to jumpstart forcing other people to speak. But it would be nice for everybody in our class to finally break the \"shyness\" barrior.......after almost a whole year has gone by.  So for all you, please just start randomly question \"Underrepresented\" people \"what do you think of this\", or \"what\'s your opinion on this issue\".  I\'m SURE everyone has great ideas to share, the \"laziness\", \"uncertainty\", and \"shyness\" are the only barriors.   In other news, third period would really like to know what was the highest grade on the English AP practice test multiple choice section in first period? Somehow almost every member of third period has a feeling that first period is \"smarter\" and engages in better discussions that us.   Don\'t be shy to share your comments now :)  Thomas Chen  '),(9979,1488330,'22,April,2004','2004-04-22 00:00:00','   Hello!  I\'m so pleased to see you all carrying on such a lively conversation; you\'re informed, you\'re responding to each other, and you\'re connecting all kinds of ideas and subjects.  Literature leads to culture, which leads to politics, which leads to a need to communicate civilly, which leads back to writing and literature.  Truly, you folks are the kinds of students teachers dream of when they first go into teaching.  I also appreciate your feedback about what is happening in class.  Concerning discussion...this has come up before.  Those of you who are participating feel like we are in a rut, as if you are dragging many unwilling--or unresponsive--bodies along with you.  Speakers and non-speakers become entrenched in their roles, and soon, class is predictable.  Boring.  So why have I allowed this?  Part of my role as a teacher is to facilitate, to push students to do what they would not be willing to do on their own.  I have a preferred style, though, that is based on respect of students\' choices.  I find it very difficult to make a person speak if he or she is not willing.  Why?  Some people are painfully shy, and to be made to speak causes that student to suffer.  I would like class conversation to work like it does at a good dinner party, where people speak because they are interested and engaged.  I give all of you prep work, and I make the assumption that if you have done the work, you have the tools to speak.  And I expect you to be motivated enough--courageous enough--to conquer fear of a large room and state your opinion.  Perhaps this is idealistic of me, yet this is what democracy depends upon.    In the way we have a large group of silent people, our class is echoing a societal trend. Small numbers of people bother to get out and vote; small numbers of people in class participate in discussion.  What do we do?  I could be an autocrat and put people on the spot, giving \"points\" to those who have an anwer.  Is this the role of a teacher?  Is this what you want class to look like?  Feel like?  Do you want speaking aloud to be connected to rewards and punishments?  To take this analogy further, should we somehow punish people who choose to avoid their responsibilities as US citizens?  Dock their pay if they don\'t vote? The classroom is a special place, though.  It wasn\'t until extra credit was offered that any of you chose to participate in the blog.  So what to do? When leaders are organizing activities, one of the things they know to do is to invite specific individuals to participate in the group.  All of you who participate in this blog participate in class, also.  By speaking aloud, others will see you as class leaders.  If you want to maintain a democratic classroom, where students feel like equals and I am not treating you all as minions, what if you were to start inviting other students to respond?  Why should I be the one to ask questions, just because I am the teacher?  Don\'t you all ask questions of each other when you are in small groups? What if you start doing this in a large group?  Who do you want to hear from?  Ask that person a question, draw them in.  What kind of encouragement/lessons/announcements (if any) do you need from me to help you take a more active role in making discussion work?  I have been in classes before where one of the ground rules for discussion is that no one in the room may make more than 2 comments until everyone has been invited to share something with the large group.  We could do this if you\'d like.  It can feel contrived; however, if you want to change the dynamic of class, this is a system that can help encourage change.    What do you think?  Mrs. V-A  '),(9980,1488330,'22,April,2004','2004-04-22 00:00:00','   Yeah I can finally blog, Ryan I too am a rookie at this. It was so nice to go outside today instead of look at another essay. I feel that occasionally a break from strenuous study adds to your understanding of whatever your studying. I found this to be true when I was reading Thoreau (WOW that was a long time ago) and I practice this regularly. I don\'t know if that is a good or a bad habit. I wanted to test out the waters of blogging and I have, so Good Night!   ~ Frank  '),(9981,1488330,'21,April,2004','2004-04-21 00:00:00','   I was dreading taking the practice AP english test all week, anticipating the worst possible three hours, or coming really close to having to take the french one.  However, after taking it today it wasn\'t as bad as I had thought. The multiple choice part wasn\'t too bad but I tend to take too much time reading and rereading... oh well.  The essays weren\'t extremely difficult either, and amazingly i finished all of them, even with a conclusion.  However the second one where we had to discuss how Emerson used the English language and conventions was a little more difficult.  Also, like it was said earlier, I noticed that Shakespeare was used and I found that to be odd considering he wasnt American, which is what we have been studying all year.  '),(9982,1488330,'21,April,2004','2004-04-21 00:00:00','   There were a few obscurites in the multiple choice section..never felt overwhelmed though, and finished in plenty of time.  By the way, did anyone notice how I used an elipsis instead of transition words there?  Basically, that\'s how I responded to the Emerson prompt.  Picked one or two conventions that seemed blatantly obvious, and beat them till they died, then beat them some more.  I can\'t tell if I was too redundant or not, but the familiarity with the author DEFINITELY eased the pain.  It made reading the passage go much smoother, and I felt not bad about it overall.  I was suprised at the Shakespeare prompt because it was so vast and wide, not really what I\'m used to.  It also suprised me that it asked for Shakespeare (which I remembered enough of to respond), because he\'s super old and not American.  Overall I feel confident (more so after some grammar review), but this will not be my most difficult test.  Good luck to everyone!  Love, Rob  '),(9983,1488330,'21,April,2004','2004-04-21 00:00:00','   Hey, how did everyone think of that practice AP multiple choice/three essays practice test we took? how do you think you are prepared? Any comments/feelings/complaints/sorrows/anger/stress/suggestions/questions about the AP English test?  To me personally, the multiple choice section was harder than the free response section (Still a lot of interpretations, confusions, and ambiguities in the multiple choice answers). And within the free-response essays, i felt that the one about the English conventions/language was the hardest, although since the passage was about Emerson, the familiarity of the author kind of comforted me a little. It was weird talking about how Emerson persuades the reader his opinions with the use of \"conventions\", such things as grammar, the parts of English language that I tend to ignore its importance. Yeah so overall, that was a challenging piece to write as the essay prompt demanded \"a well reasoned and organized essay\".  But other than that, I liked the free-free response essay that we can draw on our own experiences, readings, and observations for that last essay. It wasn\'t so strict as to just interpret a small piece of passage..   Please do respond with your opinions on the AP english practice test.   Thomas Chen  '),(9984,1488330,'21,April,2004','2004-04-21 00:00:00','   O Paul! I read the Da Vinci Code! It\'s a GREAT book. I recommend it to everyone. The different hidden symbols found in the artwork and the secret organization of the Priory of Sion is just so interesting. However if you do read it I recommend you go online and look at some of the paintings discussed in the book. This will give you a good visual of what\'s described in the book. It\'s very entertaining and it allows you to explore new ideas. It\'s also fun to try to solve the different puzzels throughout the book.  '),(9985,1488330,'19,April,2004','2004-04-19 00:00:00','   About third period being soo quiet, I think it\'s almost become a routine that others don\'t notice or don\'t care anymore. I know some people are just lazy, and believe me, third period just seems to drag on forever, but I think some also feel intimidated.  Everyone, well the only five or so people that actually talk always have really good ideas and are usually really smart, so others, including me, only think of the obvious when it comes to most of our discussions.  So either you don\'t want to state the obvious or someone else says it before you get the chance.  This has resulted that now when Ms. VA asks us all a question that no one really understands everyone leaves it up to the same people who always find a response that fits what she is looking for.  If people know they can get away without having to say anything during a discussion they will probably stay that way for the rest of the year, so unfortunately it seems that the same people will lead the discussions once more, unless something drastic changes.  '),(9986,1488330,'19,April,2004','2004-04-19 00:00:00','   I would just like to make a comment in general.  I don\'t know if first period is the same, but our third period class really bums me out sometimes.  I am pretty vocal in class (I need to shut up sometimes) but honestly a lot of the time I end up raising my hand and commenting because so few people do!  Our class discussions consist of about five people talking, and everyone else just sitting there, and I think it would be really nice for everyone to get involved.  Just sharing a thought...  '),(9987,1488330,'19,April,2004','2004-04-19 00:00:00','   Paul we\'re both too worried about our images.  Maybe we didn\'t need to apologize, but we\'ll just chalk it up to dignity and chivalry.  Even if our politics don\'t see eye to eye, we agree on the worth of the New York Times.  The NYT (as it is known by those IN the know) posts its articles online without a monetary fee of any kind, all they ask for is your e-mail.  If you want, The NYT will send you the morning headlines on the subjects that interest you (business, politics, international, sports, etc.) with links to their articles, but more on that later.    I haven\'t read DaVinci code, but I\'m damn interested.  Copies are flying off the shelves, and I\'m all for a nonstop read (I\'ve read every Harry Potter book but the last one atleast 3 times, and I\'m not ashamed to admit it).  For something shorter and tied to \"Night,\" a town in Italy is attempting to convict several former Nazi SS members (a story I got through the free NYT e-mail headline service).  You can read about it here; http://www.nytimes.com/2004/04/18/international/europe/18TUSC.html?th .  I have my opinions, but didn\'t read \"Night,\" so I\'d like it if someone else gave theirs.  I\'m posting this on Miss Clark\'s blog as well, so if you have her for a teacher she feels neglected because her blog doesn\'t get this kind of action.  I\'d also like other viewpoints on the issues and differences in opinions Paul and I worked ourselves through.  I\'d enjoy this to be a political forum, and I know you all have something to say on the world as is today.  Love, Rob  '),(9988,1488330,'18,April,2004','2004-04-18 00:00:00','   I feel I came off as too defensive.  I really don\'t mind anything political getting posted here, I just feel... obliged to post my rebuttal wherever it\'s relevant.  That being said, you really don\'t have to \'clarify yourself in a dignified manner\'  I felt your original post, even if meant for Thomas\' rather less tame personal blog, was fine and no personal slight at my personal politics at all.  I guess I was just on a defensive republican bent that night, probably worrying about Kerry\'s successful poll numbers:-)  So I guess I\'m apologizing for sounding like a bitchy little republican, you really have no reason to apologize, clarification is often a must in debate but I really don\'t think you said anything out of line or offensive to me.  So to steer the conversation anyway in a different direction....  Who\'s read the Da Vinci code here?  I started reading it yesterday and really can\'t put it down.  There\'s nothing particularly wonderful or interesting in the prose, but I find the storytelling and suspense to be excellent, as well as the subject matter.  '),(9989,1488330,'18,April,2004','2004-04-18 00:00:00','   Paul, your arguments were effective, logical, and I respect them.  While this isn\'t an excuse for anything I stated, the post was meant to go to Thomas Chen\'s personal blog, but I\'m on three of them, and all the post screens look the same except for a small titleline in the top right corner, and I accidentally posted it on this one.  Anyway, the first half of this is an apology for the crude, unpolished, vulgar post from before.  This is not characteristic of me, as my previous posts (I hope) convey.  Back to the issue at hand...    I\'ll sum up the previous post in a more formal language; There is an element of monarchy (in both Democratic AND Republican parties) in that there is a lineage of elitists that are pedigreed to run the country, and that these families become separated from many of the realities of Joe American, which I believe manifests itself in the use of American soldiers as expendable tools.  John Kerry was one of these expendables, and as far is foreign policy is concerned, could draw from his personal experiences, which were perilous no matter what that article implied (Don\'t try to tell me that having a rocket shot at your boat doesn\'t deserve an award if you are the commander, as well as going after the shooter who could very well have been able to defend himself with another, closer-range firearm.  And don\'t forget the Oregon vet\'s story that was so popular on the campaign trail, and that may have been Kerry\'s push over Dean.  And demeaning the Purple-Hearts is also a joke... as if it was his obligation to reject them???  But I\'ve digressed).  Back to the issue; a man who knows the detriments and personal losses of war would be more hesitant to thrust the volunteer (How many soldiers do you believe volunteered to fight for Sadam\'s phantom WMDs) soldiers (How many soldiers do you believe volunteered to fight for Sadam\'s phantom WMD\'s) into a frivolous war.  Point 2.  I believe Bush\'s war against Iraq is entirely a farce with negative short term and long term affects.  The short-term affects is the death toll that grows daily.  Long term affects are more anti-American senitiments that take a death toll and add hundreds or thousands in a single day (9/11, USS Cole, embassy bombings).  If a regime change is the goal, don\'t lie about it, but as of now Bush is using American men and women to fight for oil, which is necessary to accomodate America\'s mass automotive transportation.  And reflecting current consumer trends and the pressure on Bush to secure American jobs (ridiculous amounts of which outsourced by GM to Mexico), more gas is needed to satisfy the throats of gas-chugging SUV\'s, trucks, and to a lesser extent, vans; in all of said categories American companies generally control the market.  If gas gets too expensive, American cars depreciate in a hurry and sales numbers go down, resulting in more GM/Daimler Chrysler layoffs.  I\'m entirely aware that Japanese car companies have success in this country for a variety of reasons, which I understand to be largely affordability, dependability, and gas mileage.   GM is one of the largest companies in America, and rather than allow the pressure from another market to create a need for quality improvement, Bush\'s actions act in giving American motor companies too much room for error.  The last point that a gas tax would emphasize fuel-efficiency, environmental protection, and reduce accidents is largely unarguable.  If you disagree with the principal of taxes, that\'s your opinion, but let me point out that Andrew Sullivan, the author of the article that brought the issue to my attention, is a REPUBLICAN.  Not that a Democrat wouldn\'t support it, but take care when you go throwing around those increasingly meaningless slanders such as \"liberal and conservative.\"  I hope I have clarified myself in a somewhat dignified manner.  If I missed something, you have further questions, or just plain disagree, please post a response.  One more apology for accidentally posting a private on a public, I will be sure to take more care in the future.  Love, Rob  '),(9990,1488330,'18,April,2004','2004-04-18 00:00:00','   Paul, your arguments were effective, logical, and I respect them.  While this isn\'t an excuse for anything I stated, the post was meant to go to Thomas Chen\'s personal blog, but I\'m on three of them, and all the post screens look the same except for a small titleline in the top right corner, and I accidentally posted it on this one.  Anyway, the first half of this is an apology for the crude, unpolished, vulgar post from before.  This is not characteristic of me, as my previous posts (I hope) convey.  Back to the issue at hand...    I\'ll sum up the previous post in a more formal language; There is an element of monarchy (in both Democratic AND Republican parties) in that there is a lineage of elitists that are pedigreed to run the country, and that these families become separated from many of the realities of Joe American, which I believe manifests itself in the use of American soldiers as expendable tools.  John Kerry was one of these expendables, and as far is foreign policy is concerned, could draw from his personal experiences, which were perilous no matter what that article implied (Don\'t try to tell me that having a rocket shot at your boat doesn\'t deserve an award if you are the commander, as well as going after the shooter who could very well have been able to defend himself with another, closer-range firearm.  And don\'t forget the Oregon vet\'s story that was so popular on the campaign trail, and that may have been Kerry\'s push over Dean.  And demeaning the Purple-Hearts is also a joke... as if it was his obligation to reject them???  But I\'ve digressed).  Back to the issue; a man who knows the detriments and personal losses of war would be more hesitant to thrust the volunteer (How many soldiers do you believe volunteered to fight for Sadam\'s phantom WMDs) soldiers (How many soldiers do you believe volunteered to fight for Sadam\'s phantom WMD\'s) into a frivolous war.  Point 2.  I believe Bush\'s war against Iraq is entirely a farce.  He\'s using American men and women to fight for oil, which is necessary to accomodate America\'s mass transportation.  And reflecting current consumer trends and the pressure on Bush to secure American jobs (ridiculous amounts of which outsourced by GM to Mexico), more gas is needed to satisfy the throats of gas-chugging SUV\'s, trucks, and to a lesser extent, vans; in all of said categories American companies generally control the market.  If gas gets too expensive, American cars depreciate in a hurry and sales numbers go down, resulting in more GM/Daimler Chrysler layoffs.  I\'m entirely aware that Japanese car companies have success in this country for a variety of reasons, which I understand to be largely affordability, dependability, and gas mileage.   GM is one of the largest companies in America, and rather than allow the pressure from another market to create a need for quality improvement, Bush\'s actions act in giving American motor companies too much room for error.  The last point that a gas tax would emphasize fuel-efficiency, environmental protection, and reduce accidents is largely unarguable.  If you disagree with the principal of taxes, that\'s your opinion, but let me point out that Andrew Sullivan, the author of the article that brought the issue to my attention, is a REPUBLICAN.  Not that a Democrat wouldn\'t support it, but take care when you go throwing around those increasingly meaningless slanders such as \"liberal and conservative.\"  I hope I have clarified myself in a somewhat dignified manner.  If I missed something, you have further questions, or just plain disagree, please post a response.  One more apology for accidentally posting a private on a public, I will be sure to take more care in the future.  Love, Rob  '),(9991,1488330,'18,April,2004','2004-04-18 00:00:00','   Does anyone have any book recomendations after reading over spring break?  -Pat  '),(9992,1488330,'17,April,2004','2004-04-17 00:00:00','   I really don\'t have much to say about AP tests or ASB elections... but politics---  If military experience was a prerequisite for being a successful President or commander in chief of the armed forces I\'d fear for the future of this country.  And I\'d hardly compare Kerry to true war heroes such as Audy Murphy, here\'s a section of an article describing Kerry\'s Vietnam medals and tour.  But neither of those matter, the bottom line is, when you enlist in the military you are enlisting in the military.  You job is to be ready for war.  In war you are put in harms way.  Everyone in Iraq right now knew what they were getting into, we have a volunteer military.  The loss of every life, especially the civilian ones, is a great trajedy, the real issue is should we even be in Iraq, but that\'s a whole other issue.-     Kerry served on swift boats for about four months. During that time, he was awarded the Bronze Star (the lowest decoration for heroism in combat) and the Silver Star (the next higher decoration for valor) and three Purple Hearts for wounds sustained in battle. These latter decorations â€” like my title of \"nationally syndicated columnist\" â€” sound more impressive than they are. All three wounds were minor cuts from shrapnel, which, according to Kerry, caused him to miss a grand total of 2 days of duty.   For soldiers and Marines, especially of the enlisted variety, a Silver Star is a big deal. You\'ve got to do something profound to get one. But the rules were different for officers, especially for naval officers.   This is the action on Feb. 28, 1969, for which Kerry was awarded the Silver Star: A Viet Cong fired a B-40 rocket at Kerry\'s boat, Patrol Craft Fast-94. Tom Belodeau, manning the twin 50-caliber machine guns at the rear of the boat, opened fire on the VC, wounding him. The VC fled behind a hooch. Kerry ordered PCF-94 to shore, leaped out of the boat, pursued the VC, and finished him off.   I can envision grizzled infantrymen shaking their heads. \"He got the Silver Star for that?\"   Kerry had an advantage most servicemen do not. Medal recommendations have to be made by the commanding officer of the unit in which the heroism took place. Kerry was the commander of PCF-94. Presumably, Kerry\'s medal recommendation was made by the commander of the squadron to which PCF-94 belonged. But Kerry\'s commander wasn\'t there. The evidence he had of the heroism of Lt (jg) John F. Kerry came chiefly from the after action report of Lt. (jg) John F. Kerry.   Shortly after being awarded the Silver Star, Kerry took advantage of a provision in Navy regulations that permits a sailor who has been wounded three times to obtain early release from his combat tour. For Kerry â€” since his wounds were so minor â€” this was taking advantage of a technicality. There is nothing wrong with this. Many officers similarly situated would have done the same. But it wasn\'t heroic.   To recap: Kerry was a double volunteer. As a swift boat commander, he was brave and able. But I am unaware of any soldier or Marine who was awarded a decoration of any kind â€” much less the Silver Star â€” just for killing a wounded man who was running away.   Though it is being hyped far beyond what it warrants, Kerry\'s Vietnam service was honorable. What is not honorable is the way Kerry â€” in testimony before Congress in April, 1971 â€” falsely accused his fellow Viet vets of routinely committing grisly war crimes. Nor does it speak well of Kerry that in several demonstrations he marched under the flag of the Viet Cong. It is one thing to oppose the war in Vietnam. It is another to cheer for the enemy.     Why is there hope in John Kerry?  If the major issue with the president choosing to send troops to war, because of the evils of war, then no President since maybe Gerald Ford(I\'m just guessing) would fit your criteria.  I find it hard to believe President Bush is as cold and heartless as he is portrayed to be here.  I really doubt he\'s any different than any other President, I\'d just like to say that character slander is not an effective means of policy debate which is what is truly at debate here.  Tell me why Kerry won\'t be bought because his wife has billions?  Kerry is already bought, he\'s bought because he wants to win.  You can\'t buy votes with billions but you can buy votes with endorsements.  Kerry\'s wife is rich, yes, so she can fund his campaign along with George Soros.  But Kerry is already bought by the standard democratic special interest groups, labor unions, environmentalists, trial lawyers(by far the worst) because he hopes their endorsements will entice members of their groups to vote for him.  In exchange for those endorsements he agrees to give them more sway in his administration, it\'s just how the system works.  It would be nice if everyone looked out for everyone and the common good but that\'s not going to happen.  All politics are corrupt, Democrats and Republicans equally.  I agree completely that we went to Iraq for oil, and I don\'t think we should be there.  I certainly understand the benefits of securing a new nation under our sway in the middle east, removal of bases to Iraq from Saudi Arabia would certainly quell some anti American sentiment in that nation and put our bases closer to Israel.  OPEC has too much power over the United States economy and the global economy, but setting a precedent of a preemptive strikes frightens me.  Nothing would make me happier to see the US freed from oil, and the Bush administration has done an admirable job of using tax credits to push ahead gas/electric hybrids and to cut oil usage, such as in new federal requirements tying fuel efficiency requirements to vehicle weight.  The single issue of gas prices is not what gives the Japanese car makers the edge, it\'s a rather unique combination of luck, trade policies and union interference that has put detroit second to tokyo as of recently.  But this years trends from detroit have been towards cars, while some from Tokyo, most notably Nissan and Honda have certainly been to expand into less fuel efficient vehicles such as SUVs and trucks.  It\'s impossible to generalize on such as level or attribute it to something as arbitrary as gasoline prices.   Rob\'s suggestion of putting a 1 dollar tax on gas is why I\'m not a democratic supporter as a general rule.  Liberal policies dictates a negative approach to social engineering.  Why not offer an incentive, a tax break on buying a gas efficient car or a hybrid.  Democratic ethos dictates, RAISE TAXES.  Push people away from gas, not try to entice them away.  The federal government is simply too inefficient and bureaucratic to justify it taking any more tax money from the private sector.  Democrats want to take it away more so than Republicans, but they both want it.  Lesser of two evils.  I guess that\'s a rather sad commentary on my approach to politics and maybe our system in general but I digress.  Read the New York Times at www.nytimes.com even if it\'s a liberal rag it still beats the pants off the Columbian and Oregonian combined.  '),(9993,1488330,'16,April,2004','2004-04-16 00:00:00','   I would actually vote you for ASB president, believe it or not. And I would take on the responsibility of an ASB selected position if I can get my .5 credits of PE waived. I will already have five AP classes next year, and if the PE department won\'t be like a bigot like they usually do, I could manage my sixth period a leadership class. But if they become a usual royal pain in the ass, (Yeah they think PE is so much harder and more important than AP Literature or AP Physics) then I will have to resort to my last resort, and that is, to take all SIX AP classes.  But anyways, a new book just came out on Bush\'s secret plan to invade Iraq right when the invasion of Afghanistan started. The book is called \"Plan of Attack\" by Bob Woodward.  Peace out, and let\'s all have a moment of silence for all the innocent soldiers killed in Iraq.  Thomas Chen  '),(9994,1488330,'16,April,2004','2004-04-16 00:00:00','   Stress is right...thank GOD for the weekend.  I\'m taking three ap tests which will sound infantile and scoffable next to Tom\'s FIVE or so, but I\'m working on Prom day and night.  Laugh if you want for me expending so much time and effort perpetuating a night of drunkeness and morally relaxed practices, I sometimes shake my head myself, but it\'s what I signed up for and it\'s hard and stressful.  Go to prom and appreciate my hard work damn it!  Any way... it sickens me that the people that send others to their death a.ka. the Bush administration, do so without feeling the pain so many American families do when they lose a brother/mother/father/sister/friend/lover/fellow blogger in Iraq.  Bush\'s connection with the world has been GENETICALLY REMOVED.  Whether like to think of it or not, we elect monarchies.  W. Bush is something like the great great great great nephew of Madison or something.  There\'s no difficulty tracing a Kennedy bloodline either.  There is hope in Kerry.  He opted to fight in Vietnam rather than lynch himself over a balcony, and his experiences there gave him credibility teaching (what should be the obvious) wrongs of war.  His wife has billions, so don\'t expect him to be bought, and he doesn\'t claim his greatest mistake was TRADING SAMMIE FUCKING SOSA.  The dumbass Bush couldn\'t run a major league baseball team successfully, but we gave him the country!?!?  Edwards was the best answer, but he was too decent to sling mud, so Kerry it is.  I think it\'s obvious we went to Iraq for oil, becaue Iraq isn\'t/never was a threat to the security of the u.s. (North Korea is however) and the war has done nothing but stimulate enrollment for Al Aqsa Martyrs Brigade, and other groups hell-bent on making their suicide hurt anyone that\'s not Muslim.  Bush and Cheney are both oil-men (duh) and they\'re sick of having their balls in opec\'s vice, so they want to secure an oil flow.  How bout eliminating the need for oil?  Andrew Sullivan wrote a thought-provoking essay supporting the case for a gas tax of $1 per gallon.  It\'s what the rest of the world does, and it would result in a trend toward fuel-efficient cars, cut down on polution, and get many people off the road-reducing accident rates and deaths.  Bush won\'t do it though... it would give Japanese carmakers too much of an edge, and right now GM is pumping out SUV\'s as fast as bunnies are pump out kittens.  By the way, I\'m going to run for ASB president next year, and I want to see some representation from the bloggers that post on this site.  Elias, Tom, Pav, et al. if you have schedule room, uncle sam wants YOU!  Help to keep the ASB diversified.  And if Alex R. still wants to talk shit bout Evan Cate, shame on him... I\'ve been sitting on this for a while, but Evan is one of the most honest, genuine kids I know, and I\'m glad he\'s my friend.  I\'m out- talk back.  And if nonconformity becomes too popular, conform to nonconform.  Love Rob  '),(9995,1488330,'16,April,2004','2004-04-16 00:00:00','   Thomas - I agree with you absolutely that individualism is absolutely great and that America is truly a country of immense opportunities.  But I think too many people in this country tend to look up to their heros who dared to be individuals, yet simply stay satisfied and happy that someone else represents them so well and that they can get by, by simply idolizing these heroes. Maybe its not right to say this, but there are too many people that look like blind, brainless cattle that has to be guided and taken care of. Not just in our school, or just in America, but everywhere I have been. I work as a teller and see many people come in during the day, but there are only a few faces that I recognize when they come back, becuase these faces have something special about them, like a special radiating energy of being different. Yet, I believe in a \"friendly individualism,\" where people are not simply egoistic and competing, but where they help others and modivate people around them to develope their own individuality.  Rob - I think American culture as a whole seems to tend to idealize and make everything seem heroic. I just think that there are way too many controversies in American history and literature: Jefferson, freedom and slaves; WWII, Pearl Harbor and Hiroshima & Nagasaki; Freedom, civil rights and gay marriages; the only thing that seems to go hand in hand is Bush and his idiotism (no offence to anyone).  '),(9996,1488330,'15,April,2004','2004-04-15 00:00:00','   I\'d have to agree with Natalya on the theme of Self-Reliance and individualism is one UNIQUE theme to American literature and Americans. This is probably the only time I will appear as an American egotistical bigot. But I truely believe that America is a country founded on progressive beliefs, reforming individuals, and revolutionary ideas. Famous Americans of the past had always had an urge for freedom of speech, freedom of expression, and freedom of THINKING. The famous American authors of the past: from founding fathers Jefferson, Franklin, Paine, to Transcendentalists Emerson, Thoreau, to poets Whitman, Dickinson, to realist writers Mark Twain, Faulkner, to muckrackers Tarbell, Steffens and to civil rights activist MLK; they all inspire Americans with the vivid themes of individualism and the interaction between the individual and the society. Many of them of course were all considered radicals just because of their progressive ideas. But they are some of the most brilliant Americans. All of them wanted to, and they did, forever change America for the better. Seriously, in comparison, Chinese classical literature is all about the art of war and at the same time harmony with nature (juxtaposed against each other), and respecting elders, and conforming to society (which of course, many of you know I abhor). Yeah the theme of individualism is a farfetch idea in most of the world.  To truely live up to the American ideal, I believe American citizens have to always strive for the theme of believe in one self, and strive for individual rights, and find one\'s unique place in the society.  It\'s when people stop thinking about American literature with the theme of individualism, self reliance, and self-expression, that causes me frustrated. When people stop using brain and thinking about the society with their own unique point of view, and let others make judgements for them, and blindly conforms to the mainstream society, then they\'ve lost the integral value of American individualism. Losing one\'s unique place in the society fundamentally degrades the morals of human beings, and dictators dreams to have a society of dumb people that blindly conforms to the society. I don\'t believe that\'s what America, the land of opportunity, is all about.  Thus................Always think for yourself, and always question all authority.   As always, I welcome all dissagreements (P.S. Dissagreeing is also one element of American individualism).   Thomas Chen.  '),(9997,1488330,'14,April,2004','2004-04-14 00:00:00','   Paul, thanks for bringing up the nature of other categories of national literature.  Last year we had a much more cosmopolitan curriculum, and thinking back to Dr. Zhivago, I have to agree with Natalya on the desolation that Pasternak wrote about.  We read about Okonkwo too, and the let down he obviously felt when his tribe was weakening.  We also have read some western European novels and essays- Tale of Two Cities, All Quiet on the Western Front, et al.  Tale of Two Cities was really a political observation where personal relationships just worked themselves out, and All Quiet addressed the feeling of hopelessness war carries.  America truly is unique that it hasn\'t had to experience the despariging wars (other than Civil War) that Europe and Asia have dealt with, and that\'s the best evidence I can uncover to explain the difference in literary themes.  Many people in the world view Americans as arrogant and possessing other undesirable characteristics.  Does the attitude that we can do it on our own come through in the lit?  Or do Americans just lust for heros?  I find this all very interesting.  Maybe Tristan will tell me what I want to hear.  Shifting gears, what did you all think of the in-class essay today?  I enjoyed it, but being a dick, I could swear that passage came out of Disability.  No matter how strong Mairs portrays herself, you can still see her pain in the words she uses.  Lastly, in his press conference, Bush claimed he couldn\'t think of a mistake he\'s made in office.  You have to respect the guy a smidge for believing in himself, but I\'d like to make a list of mistakes he\'s made.  Ignoring all his verbal blunders, I\'ll start the list off with the suggestion that Saddam and Al Qaeda were linked, last time I checked it was a lie, and it got Clinton impeached (which he deserved).  If anyone\'s got an item to add, please post it.  Love Rob  '),(9998,1488330,'14,April,2004','2004-04-14 00:00:00','   NIGHT.  A very compelling read.   But since everyone is talking about their books on here I thought I would mix it up a bit.......and give some references to music.  First of all, Im sure there are people out there with different tastes in music......I myself prefer a mix of alternative, rap, and classic rock.  Sorry I\'m not an Enya fan.  First of all I would like to make the recommendation that by sigining up at Yahoo Music, you can create your own radio station that plays almost every genre, customized to your own style.  This is free and is VERY useful when I get sick of all the stuff that is played on the radio or on CD\'s.  But anyways......a few recommendations.....for a pop type of CD I would recommend Hoobastanks new CD, listen to \"The Reason\" or \"Lucky\".   I classic rock CD I would recommend would be the Red Hot Chili Peppers greatest Hits or anything by the Beatles(can\'t really go wrong with the Beatles)  A rap selection...lets see....Kayne West or Twista seem to be two of the most popular at the time but a solid choice would be Jay-Z\'s Black Album.....Jay-Z\'s last album.  If you are into the same music of me.....I would recommend anything by 3 Doors Down or the Goo Goo Dolls.....very solid choices in those albums.  Anyways, just thought I would mix up the blog with a little culture.  -Pat  '),(9999,1488330,'14,April,2004','2004-04-14 00:00:00','   Beck- I absolutely love you, no seriously, it\'s like you have been reading my mind, only you are way better in expressing it. I agree that it\'s funny how we worry about the least important things every day, and if you look at the big picture our lives so unimportant, they can be simply compared to that pile of ants, each ant is so small and what he does is so small, and yet I am sure he thinks that what he is doing is important. I mean, 4 children die every hour from malnutrition... strange, I have been reading Biology for the past hour... 4 children just died... do I feel anything? No. Can I even comprehend the fact that 4 innocent lives just died? Not really. I was watching the news today and they were talking about one of the Italian hostages being killed in Iraq, and I remember thinking: they make such a big deal out of this one death when at the same time there are hundreds, thousands of people dying every day. I think it\'s simply because we have gotten used to these statistics of how many people are dying in Africa from AIDS and how many have died during the Holocaust.  The horror simply gets saturated and you get used to it, and that\'s how saw Wiesel\'s experience as his story extended in the book: the first time he saw babies been thrown into the flames of crematory his reaction was \" How could it be possible for them to burn people, children, and for the world to keep silent? No, none of this could be true. It was a nightmare...,\" compared to the reaction of when his father is taken away to the crematory \"I did not weep, and it pained me that I could not weep. But I had no more tears. And, in the depth of my being, in the recesses of my weakened conscience, could I have searched it, I might perhaps have found something like - free at last!\" I have heard from many people about Night and everyone said that they cried whenever they read it, but strangely the first time I read the book I didn\'t shed a tear. I have read many books like this one, I cried reading \"The Pianist,\" and \"All Quiet on the W. Front,\" even \"Of Mice and Men\" but this one was dry, it felt as if Wiesel was simply stating facts and these facts have been already known and there was simply too much of it and too hard to believe that such things could be true, too saturated with horror.   Paul - it does feel like self-reliance is much more prominent in American literature than in any other. At least drawing on what I know about Russian literature that\'s the feeling I get. I read A LOT of Russian novels and most have this trend of sorrow, grief and disillusionment with the people and world around, maybe that\'s why Russians tend to be so \"mean,\" cynical and full or sarcasm. But I think Russian authors are great on analyzing human nature, and our lonely desolate souls; read \"Crime and Punishment.\"  Ryan- you know it\'s funny how in America everyone simply stands up for himself and takes care of his own private business, but that does leave us lonely and without friends or anyone to talk to. George has to get rid of his burden too in order to be able to go on, but does he need the farm, now that he doesn\'t have his friend to share it with? Does the old man, Candy, want to go on, now that he lost his dog, his only friend? I think self-reliance and independence are over-rated.  '),(10000,1488330,'14,April,2004','2004-04-14 00:00:00','   Going off of Ryan Elliot\'s post, the theme of loneliness struck me the most when I read \'Of Mice and Men.\' The old man who lost he dog is so frightened of being all alone for the rest of his life that he still wants to join George and Lennie\'s dream farm even after Lennie kills Curley\'s wife. Also I think that Curley\'s wife is one of the loneliest people in the novel. I think the thing she wants the most is a friend and someone to talk to. She\'s so lonely that she tries too hard to be liked by all the farmhands thus becoming unpleasant to be around. She dislikes her husband and she doesn\'t know who to turn to. Lennie is the only person who will talk to her. However, it\'s also Lennie who ends up killing her.  On a side note I think Paul\'s post was really interesting about self reliance being more of a human concept in our literature.   I think it would be nice to \"live on the fatta and lan.\"  :)  '),(10001,1488330,'14,April,2004','2004-04-14 00:00:00','   Amazing what extra credit will do for a blog, won\'t it;-)  I never really have much to say anymore, but I know our whole theme this year has been self reliance in American literature.  If self reliance is American literature, what\'s British literature, or French literature or Russian literature or Chinese literature.  I\'m simply curious as to if it\'s really uniquely American, or more a human concept in our literature.  As for turning the blog political, I say go for it, I just wish I had more time to frequently check in here.  '),(10002,1488330,'13,April,2004','2004-04-13 00:00:00','   I\'m really liking the fact that so many people are posting.  There are a lot of good ideas generated in each mind/group/class period, and I\'m really enjoying reading them so keep it up!  (Beck, your post was especially eloquent)  Today in class, (M&M) discussion seemed especially productive in my group, and I thought we came up with a really thought-provoking level 3 q.  Lennie and George rely on each other for various reasons; does their relationship signify a shift in American Literature AWAY from self-reliance?  I think it\'s quite interesting because since Emerson and Franklin to Huck Finn, there has always been that underlying theme.  The suggestion that Lennie and George can be viewed as one entity and are self-reliant in that way was one way of countering the question, but has anyone read anything to suggest that there is a paradigm shift in American Lit?  What events could be triggering this shift?  Post your thoughts, and anything else you\'d like to say... I\'d like this blog to become somewhat of a politcal/current events forum as well, because I like to argue.  Love Rob  '),(10003,1488330,'13,April,2004','2004-04-13 00:00:00','   I wasn\'t so excited to start reading these books, but I had heard that Of Mice and Men was an excellent book.  It is very intersting to see how Steinbeck shows how much these two men resemble people in a family.  I relate this a lot to myself, I see Lennie as my brother who has some difficulties and challenges, and me as George.  Sometimes I get real fed up with him, and extremely annoyed, and I do things like George that then tend to hurt his feelings, make me feel bad and then apologize and make it up to him.  My brother is also big like Lennie, and if anyone else trys to take advandtage of him or make fun of him, it ticks me off.  The same thing is with George, no one else can do anything to Lennie except him.  I like the way Steinbeck uses a lot of sentence manipulation so you can get a real grip of the attitude the characters have espcecially Lennie, it really helps the reader get into the book more.  '),(10004,1488330,'13,April,2004','2004-04-13 00:00:00','   In a discussion about \"Of Mice and Men\" our group observed that George and Lennie\'s mutualistic friendship creates a theme of interdependence, which differs from the popular theme of self-reliance of the past semester.  The works of Emerson, Thoreau, and Franklin represent the fruits of solitude, and promote the benefits of self-sufficiency.  In contrast, Steinbeck\'s \"Of Mice and Men\" shows the loneliness of seclusion and explains how interpersonal relationships are necessary in leading a fulfilled life.  When George and Lennie arrive at the ranch they are hired to farm, their bond surprises a well respected ranchhand named Slim.  George takes exception to Slim\'s reaction and Slim explains there aren\'t \"many guys who travel around together...maybe ever\'body in the whole damn world is scared of each other.\"  Slim and the some of the other workers are envious and almost admire the friendship George and Lennie share.  Their relationship enables them to divide the stress of work and travel, as opposed to bearing it individually.  As the novel progresses the theme of mutualism stretches to recruit other characters.  Candy, an old, one-handed swamper, loses his dog to old age and sees lonely, desolate days ahead of him.  This depresses him and he spends most of his time staring at the wall of his bunk.  Crooks, the negro stable buck, isn\'t allowed to play cards with the guys, and is forced to sleep in a solitary room away from his peers.  Time eats away at these friendless men because as Crooks put it, \"A guy goes nuts if he ain\'t got nobody.\"  Although self-reliance produces independence and pride within a man, it cannot pat him on the back, play a game of cards, or swap stories the way a friend can.  Living as an equal in the world is an inherent human goal, and to sacrifice societal equality for isolation produces a burning loneliness that must be tempered with humanity.    A few thoughts- Ryan Elliott  '),(10005,1488330,'13,April,2004','2004-04-13 00:00:00','   Like many others, after today\'s small group discussion I have new thoughts on the book Night.  My group spent much of the period analyzing what the main theme of the book was and we came to two conclusions.  The first theme that we noticed was the naive/optimistic attitude the Jews had towards the Germans and the treatment they would receive in the events that led up to being sent to Auschwitz.(sp?)  The basis behind this theme most likely comes from the strong faith Jews had in God at this time period.  Even after being sent to the concentration and experiencing the cruel treatment of the Germans, Jews still had an optimistic view stating \"God must be testing us.\"  Although Weisel himself begins to lose faith in God, many Jews still believed God was \"just testing them.\"    The second theme my group noticed was individualism in the concentration camps.  Because the happenings at Auschwitz are so horrible, they begin to mold the spirit of many of the prisoners, Wielsel in particular.  This theme is best shown when Wiesel\'s father is being beaten severely by a crazed guard.  Instead of Wiesel becoming angry toward the guard, he instead is angry at his father for getting in the guards way.  Wiesel, like many others, lose their ties with family and friends in an effort to \"fend for themselves\" at the concentration camps.   These are the two most prominent themes my group saw in the book, are we missing a big one?  -Pat  '),(10006,1488330,'13,April,2004','2004-04-13 00:00:00','   During the group descussions today in class, my group noticed something very interesting. I don\'t know where i\'m going to go with it, but i thought i might throw it out there anyway. Our group was talking about Night, and we were busy discussing how horrific/depressing the book is when we heard a neighboring group mention how someone in the other book had lost his pet mouse and therefore all the joy in his world. We went kind of silent for a minute and looked at each other. How does a mouse compare to the Halocaust? And then we just went back to our discussion.   But i\'ve been stuck on it all day, and i think that our reaction has as much to say as the juxtoposition of the books does. And i think that it all has something to do with the human ability to completely ingore huge, world changing events (or the inability to understand them), like the halocaust. I\'ll let the great Eddie Izzard make the next point:   \"Pulpot killed 1.7 million people, we can\'t even deal with that. I think we think if somebody kills someone, that\'s murder, you go to prision. If you kill 10 people, you to to Texas, they hit you with a brick. That\'s what they do. Twenty people, you go to a hospital, they look at you through a small window at you forever. And over that, we can\'t deal with it, you know? Somebody\'s killed one hundred thousand people, we\'re almost going, \'well done! well done, you killed 100,000 people? You must get up very early in the morning. I can\'t even get down to the gym.\' ...So i suppose we\'re glad that Pulpot\'s under house arrest, at least we know where he is. Just don\'t go in that house, you know?\"   Some people would probably say that we find little ways to deal with huge or horrible things ...like pet mice. But i think that we just completley ignore them. A revolution in Haiti? Ten thousand citizens dead in Iraq? Ariel Sharon pulls out of the West Bank? Terrorists attack Madrid? Oh. Okay. ...Did i leave the iron plugged in? God, my boyfriend\'s being so annoying these days. Do i look fat in this shirt?... Nevermind the 25 million people that die every year from hunger. Or the 34.3 million people who are looking a little skinny since they contracted AIDS.   I\'m not suggesting that we should think about this all the time every day, but it is valuable to be reminded every now and then.   So yeah... I\'d put some sort of conlcusion on this... but it\'d be way easier if people posted responses (seeing as you all completely ignored my last post) that i could then steal ideas from.  -Beck  '),(10007,1488330,'13,April,2004','2004-04-13 00:00:00','   I just have to mention the observation that as if 7:02 April 13th, we have two distinct ads on top JUXTAPOSED to each other. One about \"supporting the loving conservative BUSH administration\", and the other \"Support Anti-War movement\". Kind of hilarious. Yeah if you haven\'t figured out, Google (the owner of blogger), periodically \"reads\" your blogs and accordingly put appropriate banner ads on top. So our posts about drafts, holocaust probably caused google to put those ads up.  And Megan, that quote there from Night is really insightful. Especially the part of \"if we forget, we are guilty\". (That\'s why history, historiography, and historical literature is so important). And finally Rob, to bring up that draft thing into perspective, it was actually two democrats who first initiated the plan. They think it would \"democratize\" the army system. Because they think it\'s not fair for a special group of people to fight. And of course many many Republicans support their plans, it helps them win their ambition of occupying Iraq. But yeah, there is some level of extremely scary muti-partisan support of the draft in the congress right now.  '),(10008,1488330,'12,April,2004','2004-04-12 00:00:00','   I finished \'Night\' today. Mrs. VA said we may have to read it in chunks if it became to emotional for us, but I found myself not being able to put it down. I went online and looked up information about Elie Wiesel. I read his Nobel Prize Acceptance speech. Check it out, it\'s really good.  http://www.eliewieselfoundation.org/ElieWiesel/Nobel_Speech.htm I also found some pictures of him. Mostly when he was older, but I found a picture taken when he was 15 or 14 before he was in the concentration camps. It made me think about the end of the book when he\'s looking in the mirror. I looked at the picture and I tried to imagine what he looked like at the end of the war. Here\'s a good quote from Wiesel. I think it does a pretty good job at summerizing why he wrote \'Night\' and why understanding what happened during the Holocaust is so important.  \"I have tried to keep memory alive, I have tried to fight those who would forget.  Because if we forget, we are guilty, we are accomplices...That is why I swore never to be silent whenever wherever human beings endure suffering and humiliation. \"  '),(10009,1488330,'12,April,2004','2004-04-12 00:00:00','   I\'m a \"Mice and Men\" kinda guy, and I\'m going to agree with Thomas\' \"it\'s a comfortable read\" quote.  It\'s mostly dialogue and fun, quite frankly.  Bringing the draft back is a heinous idea to me as well, and it drives me crazy because Bush is not being the president that (lost the popular vote but) we elected.  Does anyone remember the debates of 2000 when W. Bush adamantly and outwardly expressed his distaste for the way Clinton was using the military?  Saying directly that U.S. troops should not be used for nation building in other countries.  Hmm... why does W. Bush want to reinstate the draft again?  Ohh, that\'s right, it\'s because he needs more troops to continue his nation building in Iraq.  Afghanastan was a justified war.  I entirely understand why the Taliban had to be removed, and (with international backing) it was the right decision to invade.  What W. Bush told the country the reason for invading Iraq was is SOMEWHAT legitimate because Saddam possessing biological/chemical weapons is a threat to our ally Israel, but that\'s not the reason he decided to invade.  He can blame intelligence all he wants, but suggesting Iraq had tried to acquire Uranium based on zero evidence isn\'t the kind of credibility I want in a Commander In Chief.  Now, as the motives continue to change, hundreds of Americans are dead because of a W. Bush lie, and he wants thousands more to risk their lives for a quest he said was not worth fighting before he was elected.  I scoff at those who want to support W. Bush because they believe KERRY flip-flops positions.  The more I think about it the sicker I get.  I might type more, but I\'m off to do my taxes, attempting to get back some of the money W. Bush is itching to turn into a deficit.  Love, Rob  P.S.  I need a prom date, so if anyone has any ideas, please feel free to let me know.  '),(10010,1488330,'12,April,2004','2004-04-12 00:00:00','   I for one have never read Night.  It is interesting and very depressing.  At one point I had to stop reading because I felt so awful.  But does everyone like the book that they have chosen?  (I wanted to see if this worked too)  Kaitlin  '),(10011,1488330,'11,April,2004','2004-04-11 00:00:00','   Maybe contrary to a lot of people, I\'m reading \"Of Mice and Men\". The book so far is faring better than it\'s intro paragraphs, which when I heard Ms.VA read in class today, I thought it would be so boring. But right now it\'s becoming more and more like a Western Thriller book, with A LOT of Western vernacular language (as oppose to the Southern vernacular language we encountered in Huck Finn). It\'s easy to relate to that a lot of language directly imitates the coloquial language as people are speaking (non-sensored words). So overall, a comfortable read.  Other than that, I encountered this article:  urlLink  about some members of the congress are even THINKING of RE-INSTATING the DRAFT.........Utterly sick..........   Yeah anyways, thats a great article, also talks about in December 2003 the United States had 1.44 million troops stationed in 141 countries.  And here is an excerpt from an article from CNN:    CHARLES RANGEL   Democratic Congressman from New York and Korean War vet   Staying the course in Iraq means increasing our troop strength, and, not surprisingly, recruitment and re-enlistment levels are down. But proposed enlistment bonuses and other economic incentives will not make the military any more attractive to upper-middle-class young people. Increasingly we will be a nation in which the poor fight our wars while the affluent stay home.   To correct the disparity among those who serve, South Carolina Senator Fritz Hollings and I have proposed a new draft. All men and women ages 18 to 26 would be eligible for induction once they have completed high school. Those not needed in the military would perform civilian service. Enacting our plan would democratize our armed forces and return to the \"citizen soldier\" ideal that has served our nation so well.  As a veteran, I strongly believe that fighting for our country must be fairly shared by all racial and economic groups. Nobody wants to go to war, but the burden of service cannot fall only on volunteers who, no matter how patriotic, are attracted to the military for financial reasons. We cannot continue to pretend it is fair that one segment of society makes all the sacrifices.   CHARLES MOSKOS   Professor of sociology at Northwestern University and a former draftee   Our country is facing new kinds of threats and needs a new kind of draft. Even before the wars in Afghanistan and Iraq, our military was severely overstretched in fulfilling its missions. But more important, we have done nothing serious about homeland defense in the war against terrorism. We need guards for our nuclear power plants, dams and public facilities. We have done little to create the necessary border patrollers, customs agents and cargo-ship inspectors. Short-term draftees, under professional supervision, could perform these duties admirably. It takes less than four months to train a military police officerâ€”precisely the kind of role most needed in peacekeeping missions and guard duties. This would free up professional soldiers, and it would stop the unprecedented activation of reservists. Their multiple tours have led to demoralization and impending recruitment shortfalls.   We must institute a three-tiered draft system in America, with 15-to-24-month tours of duty for citizens ages 18 to 26. In the new-style draft, conscripts could serve in the military, in homeland security or in a civilian-service program like AmeriCorpsâ€”and there is no reason women could not be drafted for the latter categories   JAMES INHOFE   Republican Senator from Oklahoma   I think I\'m the only member of the Senate Armed Services Committee who would reinstate the draft. There are huge social benefits that come from it. I can assure you I would not be in the U.S. Senate today if I had not gone through the draft. When I look at the problems of some of our kids in America nowadays and then I go visit the troops, I see what a great benefit it is to give people the opportunity to serve their country.   I was drafted into the Army in January 1957 and served two years as an enlisted man. I gained a new outlook on life through the rigors of basic training. The military can have a more intense influence on soldiers when they are drafted and have no choice. I developed a sense of patriotism through the experience of serving my country. I\'m not on a crusade, but I think today\'s youth could use more of that type of discipline.  '),(10012,1488330,'11,April,2004','2004-04-11 00:00:00','   So far I am really enjoying the novel \'Night\' due to the way in which the novel is written.  The novel consists of choppy, simple and compound sentences, to enforce the view which he is writing from(a child\'s) while also creating a fearful, anxious tone in the book that comes from these choppy sentences.  In class, Ms. V-A spoke of the novels \"poetic tone.\"  I believe this tone is derived from the sentence structure Wiesel uses in the novel(or atleast the first 15 pages).  Is everyone else seeing this tone? Or something different?  -Pat  '),(10013,1488330,'11,April,2004','2004-04-11 00:00:00','   Ya, I also already read \'Of Mice and Men.\' I also read the New York Times at 4:30 in the morning everyday. I only read it then because I have to wake up early for zero period. \'Night\' is really good so far. Really really sad though. My math teacher said he read the book when he was in high school and that it gave him nightmares. I hope I don\'t get nightmares. I don\'t know about the rest of you but AP testing is comming up too quickly.  '),(10014,1488330,'11,April,2004','2004-04-11 00:00:00','   Why didn\'t more people choose \'Of Mice and Men\' than \'Night\', there appeared to be at least twice as many people reading Night.  For me personally, I\'ve already read Night so Of Mice and Men was the logical choice.  And go read the New York times, it\'s free online, www.nytimes.com it\'s good for the brain.  '),(10015,1488330,'11,April,2004','2004-04-11 00:00:00','   Hey, I hope you all had a fantastic break... I know I did.  I read \"Catcher In The Rye\" over the break because there are a lot of references to it in other media sources...Six Degrees of Seperation and Chasing Amy are two movies that the book influences, but I never really felt the significance of it.  It\'s a book narrated by a student our age about himself.  He\'s having trouble dealing with his life- he makes references to suicide, he\'s kind of cynical, thinks everyone\'s a \"phony,\" as well as intermitently dealing with the death of his brother.  It\'s written coloquially (like Huck Finn only NINETEEN 50\'s), and it\'s kind of funny at times because Holden (the narrator) is such an oddball character, but it\'s nothing that\'s really going to change my life.  If you\'re disconnected with your surroundings and like a straightforward read, you could identify with \"Catcher,\" or if you\'re Thomas, and are rooting for the nonconformist protagonist, you might really identify.  I guess it just wasn\'t what I expected, but I AM looking forward to \"Of Mice and Men.\"  Love, Rob  '),(10016,1488330,'11,April,2004','2004-04-11 00:00:00','   I visited some schools on the East Coast, and they had similar requirements. For Cornell and NYU, they usually take 4\'s and always take 5\'s. It appears some places accept 3\'s on a few tests, but that usually just allows you to transfer out of the class, it doesn\'t get you any credits. To actually get college credits takes a 4, or a 5, they seem to be more strict on the science and English tests most places I looked.  '),(10017,1488330,'06,April,2004','2004-04-06 00:00:00','   Hey that\'s a very good question and interesting topic. Yeah some school may offer a credit system in which a 3 may get you into a basic class, and 4 or 5 a more advanced class. But in order to eliminate 5 or more credits, most schools has to see that you have at least a 4 or 5 on most AP exams.  UW is a very typical example of this AP credit system. Here is a link to see how much credits UW offer base on your AP scores:  urlLink UW AP Class Credits  '),(10018,1488330,'05,April,2004','2004-04-05 00:00:00','   I know this may be very off topic but I just returned from Gonzaga and found out that Gonzaga and many other schools demand a 4 on the AP Test to count for college credit.  While many of us last year when taking the World History exam believed that to \"pass\" we needed to get a 3.......well I guess we passed but it doesn\'t count for anything at alot of the schools that we may be trying to get into when applying next year.  I have also heard a rumor that the University of Washington doesn\'t accept 3\'s......this is only a rumor but I would like to hear from anyone else who visited schools to see what other schools demand on the AP Test to count for college credit.                            -Pat  '),(10019,1488330,'05,April,2004','2004-04-05 00:00:00','   I\'m seeing if this works.  '),(10020,1488330,'04,April,2004','2004-04-04 00:00:00','   I started reading \"Fountainhead\" a while ago, but I wasn\'t able to finish it. Too much stuff going on. If anyone is looking for a good book to read I recommend anything by Ayn Rand. She\'s an amazing author.  Over spring break I plan to read \"A Passage to India.\" I don\'t know much about it. My mom said it was good book. I should start reading it but I\'m too absorbed in another book called \"Battle Royale.\" It\'s by a Japanese author, and the book reminds me a lot of \"Lord of the Flies.\"  The government sends a bunch of kids out on an island and they have to kill each other and whoever is left alive at the end wins. It has a lot to do with government control over the people and what fear can cause you to do. It also relates to the saying of it\'s a dog eat dog world. It\'s really bloody, but interesting.  Recently in AP US History we just finished up with WWII. We watched some video clips of the Holocaust with was really shocking for me. Anyway, I remembered something Mrs. VA said about her German friend and how he is still paying for WWII through high taxes, and how he was really upset because he had nothing to do with Nazi Germany and the Holocaust. An interesting book which relates to this is \"The Reader\" by German author Bernhard Schlink. It\'s a deep book containing a lot of philosophy. It centers around a young man who wasn\'t involved in WWII but still has to deal with the horrors of postwar Germany.  '),(10021,1488330,'04,April,2004','2004-04-04 00:00:00','   Well yeah, \"Fountainhead\" has a very straight-forward message for individuality, which I absolutely adore, but some people can probably find it a bit too zealous. Are you going to write the scholarship paper for the book Thomas? I think it\'s due on April 15th. I think you should also add a link to fastweb.com, maybe some people will find some good scholarship opportunities to write for.  Do you have that list of 100 recommended books from collegeboard.com? I have \"Candide\" next on my list; let me know how good the book is.  '),(10022,1488330,'03,April,2004','2004-04-03 00:00:00','   Very inspirational post there Natalya. I agree that we need to have lots more multiple choice practices. And more examples and practices of creating the three levels of questions would be very helpful. Hmm I never thought \"Fountain Head\" was that good. I\'m personally reading this philosophical book by Voltaire called \"Candide\". What are you all guys reading for spring break?  '),(10023,1488330,'02,April,2004','2004-04-02 00:00:00','   Well I joined you guys here too. I was looking at the AP English Barron\'s book and it seems that in order to do well on the test we need a lot of just practice of taking the answers and knowing what kind of questions are on the test and how to interpret, it\'s kind of like SATs or ACTs, you do better on english part with more practice. I hope after spring break we will have a lot of practice on the AP mult choice and some timed writing.     The Challenge is to read some good books, I personally recommend \"The Fountainhead\" by Ayn Rand about individuality, the best book I have read and very inspirational. Also if you want some more background on the Holocaust you might want to read \"The Pianist\" or rent the movie, which I found to be very powerful, yet disturbing. I find reading literature such as \"Night\" important to remind myself every so often how superficial our problems are when compared to what people had to undergo during Holocaust, WWI, WWII, and even now around the world. Though I find it ironic to see American propaganda against German Nazis and the genocide directed toward Jews, when arguably Hiroshima and Nagasaki can be viewed as genocide directed toward \"Japs.\" Also the fact that U.S. was supporting Iraq against Iran at the time when Sadaam was gassing Curds is especially interesting since the Kurds genocide is one of the reasons we had the war with Iraq. History is very controversial, especially when put in the context of literature. So yeah, liberty and justice for all (with exceptions to those that are different from \"all\").  '),(10024,1488330,'02,April,2004','2004-04-02 00:00:00','   Wait so what\'s this \"Spring Break Challenge\" about? It seems to be full of curiosity and contains some imminent surprises......But anyways, I guess for the spring break we just have to read some independent reading novels? Anything else important I missed.....What\'s the new schedule for after spring break gonna be like?  '),(10025,1488330,'02,April,2004','2004-04-02 00:00:00','   Thomas, I feel like this has become a dialogue between you and I, not that I\'m complaining.  It\'s unfortunate we haven\'t had as much forced social contact since 8\'th Grade Spanish, we sure had good times.  Anyway, I wasn\'t in class for English today, would anyone like to let me know what the \"Spring Break Challenge\" is?  It was mentioned on the syllabus, but haven\'t heard about it since.  Have great breaks everybody!  Love, Rob  '),(10026,1488330,'01,April,2004','2004-04-01 00:00:00','   Wow in suggesting me to perform any kind of fashion or dancing show is like thinking the unthinkable. If that made any sense. But anyways, you certainly have a very interesting humor there Rob....And I\'ll certainly commend you for spicing this blog up with some humor :) As to breakdancing, do you know who\'s going to be really good at that......Tyson Cecka. So if you ever have any trouble looking for a good break dancer, Tyson Cekca can do it all for you.   And for everybody here, I demand all of you to have an awesome Spring Break!  '),(10027,1488330,'01,April,2004','2004-04-01 00:00:00','   Haha, sorry about that Tom, hopefully this one only shows up once.  We got a new desk for the computer, so maybe some good karma comes with the hardwood.  I appreciate your optomism about the assembly, I\'m not on the committee so I can just sit back with my fingers crossed and enjoy.  Did anyone approach you about being a model???  If not, we\'ll have you up there next year in the trashiest outfit Anchor Blue has to offer.  I can see it now...you\'ll be strutting your stuff down the runway, stopping, turning, throwing your jacket over your shoulder and giving a sexy, heart stopping look Derek Zoolander himself would be proud of.  You\'re a natural-born star, Tom.  Maybe you can guest-star as a breakdancer for the Rick White experience.  Love,  Rob  '),(10028,1488330,'01,April,2004','2004-04-01 00:00:00','   Hello all! We\'ve got quite the rousing discussion going on here, and I figured that I should contribute.  (Sorry this got half published before, I don\'t know what happened)  I wanted to respond to Ryan E and point out that there have been reports of at least 8,000 and up to 10,000 Iraqi civilians killed in our invasion of their country. I no longer pledge allegiance to our flag. Of course there are many reasons why I no longer pledge allegiance (the false election in 2000; Martha Stewart going to prison while Ken Lay, and countless others, get off; the \"defense of marriage act*\" +constitutional amendment, the \"no child left behind act,\" and the \"patriot act\") but the war is a major one right now.   I also agree with Rob though. I suppose it\'s alright if they keep \"under god\" in our little third period prayer. It really doesn\'t bother me; when I still said the pledge, I simply left that part out. And yet, the wanted gap between church and state that would justify keeping \"under god\" (with Rob\'s argument) isn\'t always evident. When the president says, \"Marriage cannot be severed from its cultural, religious and natural roots...\" (2/24/04) I lose confidence in the separation between his religious convictions and his political policies.   Here\'s a site that put the war into perspective: http://www.infoshout.com/the_toll.htm ...Liberty and justice for all, eh?  *AHH! How can they use words like \"defense\" and \"protection\" when all you need is $54 to get a license? No proof of love, no plan of action for good intentions, just some cash and a couple signatures. http://www.thestranger.com/2004-03-11/city4.html     ~Beck Nelson~  '),(10029,1488330,'01,April,2004','2004-04-01 00:00:00','   Hahahaha, Ok, Rob, you realize you have three of the same posts up. That\'s ok. You can re-enforce your point of view to other people just in case they missed your point on the first post....they have the second and the third post to remind themselves of your arguments. :) Oh yeah and about the assembly, I\'m so sad that I can\'t be there tomorrow, because of the NorthWest Science Expo, I\'m sure it will be awesome.  '),(10030,1488330,'01,April,2004','2004-04-01 00:00:00','   First of all, let me start off by saying that I have no problem with mandatory repetition of the Pledge of Allegiance because it stands for becoming INDIVISIBLE in an effort against conformity...promotes LIBERTY- the liberty to think, act, and live without government intervention...as well as JUSTICE FOR ALL- the permission for everyone to live as they please. It may be ironic that there is a uniform method of promoting the liberties that lead to diversity, but if everyone is made different through the same process, are they not still different? I recognize the recitation of the Pledge alludes to military service as flags are used to determine a soldier\'s allegiance in war, but I also believe that the Pledge is more effective with a visual aid. The debate today is whether or not \"under GOD\" should or even COULD be allowed to stay in the Pledge. I outlined the current particulars of today\'s case in a previous post, if you haven\'t read them. To me, this is a case defining the separation of church and state. Although this country was outlined around that seperation, there is nothing that states that government can\'t recognize the ideology of monotheism, or other belief structure, as long as it doesn\'t allow the government of religion to directly affect it. One could argue that excluding the possibility of a deity\'s existence would be not seperating the state from atheist ideology. Tom C addressed free speach...that each American has the right to say what they want and worship as they please. With compliments to that right, many Americans feel the direct connection between Providence and the success of this country, and that excluding their right to express His influence in relation to the country would antithesize their beliefs. The same issue is currently being presented in the Ninth Circuit Court of Appeals, as an Atheist man argues that with each (socially) mandatory oration of the Pledge, his daughter is taught that her father\'s beliefs are the wrong ones. As for me, I believe that eliminating it from the Pledge restricts the large portion of Americans who recognize the Glory of God as parallel with the liberties of this country. For those who are polytheistic, \"under god\" can be translated as recognizing a \"higher power\" than is visible in daily life. Lastly, for those that don\'t recognize any supernatural force or deity, \"under God\" may be substituted with a respectful pause while those who do choose to recognize God do so. My views have been presented, but I\'m all about becoming as educated as I can on the attitudes toward the aspects of life that effect the way we live every day... so school me! If \"Night\" is a good read, then I\'m all for it...especially if it\'s short! Also, are Rip Van Winkle, The Grapes of Wrath, or The Catcher in the Rye worth reading over break? Because they all spark my interest. Have a great break, and the fabulous assembly ASB will be presenting Friday morning. Love,  Rob  '),(10031,1488330,'01,April,2004','2004-04-01 00:00:00','   This is a repost, I accidentally posted over an earlier post of mine, so if it appears twice, you now know why.  I\'m about as tech-smart as the chair I\'m sitting on.  First of all, let me start off by saying that I have no problem with mandatory repetition of the Pledge of Allegiance because it stands for becoming INDIVISIBLE in an effort against conformity...promotes LIBERTY- the liberty to think, act, and live without government intervention...as well as JUSTICE FOR ALL- the permission for everyone to live as they please. It may be ironic that there is a uniform method of promoting the liberties that lead to diversity, but if everyone is made different through the same process, are they not still different? I recognize the recitation of the Pledge alludes to military service as flags are used to determine a soldier\'s allegiance in war, but I also believe that the Pledge is more effective with a visual aid. The debate today is whether or not \"under GOD\" should or even COULD be allowed to stay in the Pledge. I outlined the current particulars of today\'s case in a previous post, if you haven\'t read them. To me, this is a case defining the separation of church and state. Although this country was outlined around that seperation, there is nothing that states that government can\'t recognize the ideology of monotheism, or other belief structure, as long as it doesn\'t allow the government of religion to directly affect it. One could argue that excluding the possibility of a deity\'s existence would be not seperating the state from atheist ideology. Tom C addressed free speach...that each American has the right to say what they want and worship as they please. With compliments to that right, many Americans feel the direct connection between Providence and the success of this country, and that excluding their right to express His influence in relation to the country would antithesize their beliefs. The same issue is currently being presented in the Ninth Circuit Court of Appeals, as an Atheist man argues that with each (socially) mandatory oration of the Pledge, his daughter is taught that her father\'s beliefs are the wrong ones. As for me, I believe that eliminating it from the Pledge restricts the large portion of Americans who recognize the Glory of God as parallel with the liberties of this country. For those who are polytheistic, \"under god\" can be translated as recognizing a \"higher power\" than is visible in daily life. Lastly, for those that don\'t recognize any supernatural force or deity, \"under God\" may be substituted with a respectful pause while those who do choose to recognize God do so. My views have been presented, but I\'m all about becoming as educated as I can on the attitudes toward the aspects of life that effect the way we live every day... so school me! If \"Night\" is a good read, then I\'m all for it...especially if it\'s short! Also, are Rip Van Winkle, The Grapes of Wrath, or The Catcher in the Rye worth reading over break? Because they all spark my interest. Have a great break, and the fabulous assembly ASB will be presenting Friday morning. Love,  Rob  '),(10032,1488330,'31,May,2004','2004-05-31 00:00:00','   Thomas thank you for setting this up.  It took a while but I finally got your email.  Today is Memorial Day and I was just thinking how much I appreciate all the freedoms and protection we have as Americans.  Both of my grandpas went to war.  On my mom\'s side my grandpa served for four or more years and on my dad\'s side he served for about four years as well.  It\'s crazy to think that they went through so much for us today.  It\'s also weird knowing that someday some of us might take part in a war.  I just wanted to get my first post out and see what other people thought about on Memorial Day.  GOD BLESS!  '),(10033,1488330,'30,May,2004','2004-05-30 00:00:00','   Ok, just in case you have not realized, I\'ve completely changed our AP English website for the better. There is a \"contributing members\" section on the sidebar that allows us to look at our AP English friend\'s blogging profiles. And I\'ve added two things:   1: The title bar: The title bar serves to better organize our blogs by forcing everyone of us to post the topic we will be discussing in our posts. Thus people will know exactly what the post is for w/o reading sometimes long and tedious posts. 2: Comment option for every post. Hey, if you feel somebody\'s got a very interesting, or hilarious, or horrible post, feel free to write some comments about that peculiar post.   THANKS ALL MY FRIENDS!  Thomas  '),(10034,1488330,'29,May,2004','2004-05-29 00:00:00','   As the year winds down, I want to give some love to the class, especially everyone that posts.  We\'ve had some good times, and I think it would be both mentally beneficial as well as fun to continue blog participation through the summer.  Pat, your essay was inspiring.  I\'ll be interested to see how the follow-through manifests itself this summer.  Look forward to raisin some philanthropic funds at the fireworks tent across from Blackjack.  To all you potential fireworks buyers this July, Support The Fort!  Blackjack has some good advertising and stuff, but it\'s run by some weird ass people, I know from experience.  Natalya, I hope your experience doesn\'t support the thesis you\'ve generated from TEWWG.  I won\'t deny that some fellas have very rigid beliefs as to how a woman deserves to be treated, but don\'t let that discourage any of you fine young women.  There are plenty of GOOD guys to go around.  A little format note; I\'ve always wondered how publishing companies created the extended dash, but I\'ve figured it out!  (I have a bloody iMac, so I\'m not sure about this on Windows, but...)  Holding the option key and shift key, pressing the button in between the 0 and = key produces the punctuation I\'ve long looked for: â€” (compared to -).  This makes me so happy!  My essay I wrote easily in one sitting.  It was nice to forego a certain amount of literary analysis and focus on personal response because I felt like I was able to get something off my chest.  I\'ll post it if desired, but it\'s more or less a more personal extension of the post with the long quote.  Lastly (this is a very quick-moving post), the Harry Potter story line is nothing new.  The classic series\'; Chronicles of Narnia, Lord of the Rings, and even Star Wars have taken the theme of the conflict between Good and Evil and given it a different setting.  The story lines each include some standard elements of personal struggle and sacrifice, making the story quite predictable, yet gripping nonetheless.  For example, Luke Skywalker must resist the Dark Side of the Force, Frodo frequently expresses the grip he knows the ring has on him, and even Harry Potter fought with himself in his choice of House, ultimately siding against Slytherin\'s beckoning promise of power.  I could make a pretty good guess to the end of the series because they all end very similarly, but I will NOT stop reading them till Voldemort is contained and Harry graduates (or maybe he dies in sacrifice, that OCCASIONALLY happens, but I doubt it in this instance).  The key to writing a successful series is the ability to weave reader interaction into a story as Rachel addressed, and equally important, putting a unique spin on the fantastic setting.  To reiterate, I love you all  Love, Rob  '),(10035,1488330,'27,May,2004','2004-05-27 00:00:00','   I found Catcher in the Rye to be a bit scary too once I realized that Holden was really mentally unstable. My paper was a parody on Holden and I just hope Mrs. VA won\'t be too disturbed by the \"creativity\" of my imagination in this paper. It was actually really fun to explore the thoughts of a potential character like that, I could draw a lot for my writing from my guy-friends and their thoughts that they express verbally, I think I could write a whole book just in that kind of a colloquial style, exploring the different ideas people have in their mind during the day.   I ran our of my reading mood, so Their Eyes were Watching God is only half finished for now, but so far my reaction to it is: men are sexist and generally evil, with rare exceptions. That is of course an exaggeration but the feeling I get from the book so far is that men find a girl, get what they want from her, that is love and attachment and then either get bored and move on to another \"victim\" or exploit the already used and hypnotised girl to fullest potential. I just couldn\'t believe it how Janie\'s first husband first begged her to marry him and then went to by her a mule so that she could plow his fields. The same with her second husband Jody who first says: \"you behind a plow! You ain\'t got no mo\' business wid uh plow than uh hog is got wid uh holiday! ... A pretty dall-baby lak you is made to sit on de front porch and rock and fan yo\'self and eat p\'taters dat other folks plant just special for you.\" It\'s ridiculous how he very soon placed her behind the counter of his shop and scoundreled her for miscounting. I just wouldn\'t believe the swift change from adornment to \"... you [Janie] need tellin\'. Somebody got to think for women and chillun and chickens and cows. I god, they sho don\'t think none theirselves.\" I just love the comparison between women and chickens, no wonder the word \"chicks\" addressed toward girls is still in the lexicon. I want to finish the book though, maybe Tea Cake will bring some optimistic news about the race of men and will lower down my cynicism.   I was just wondering if anyone has any ideas about the significance of Janie\'s beautiful hair and the crazy mule running around.  Call me a nerd but I love Harry Potter books, they are just so beyond imagination it\'s fascinating. They inspire me to write something like that too, only it\'s hard to come up with new ideas for a story; I think I am going to go think about some potentially successful ideas for a plot, maybe that will get me into Harvard, or at least I could make millions of dollars selling the books and then just bribe the admisions officers;) I think the movies are not that great though. Lord of the Rings series were better and closer to the actual book. These kinds of books are just fun to read in order to get away from the ordinary world, afterall it gets really boring here sometimes and they don\'t take that long to read becuase they have such straigh-forward ideological content: the good always wins, yet I have to agree with Pat that they are a bit stricky in the story line. I am still struggling with Fountainhead though, that book is insane, the more I read of it the harder it becomes and the more I want to read it.  '),(10036,1488330,'26,May,2004','2004-05-26 00:00:00','   I don\'t know about the rest of you, but I would like to read some of the other papers that students have written to see different perspectives on the topics.  I will share mine, it was based on The Great Gatsby, and I chose the second prompt that concerned how the book has influenced our beliefs....read mine but I would   really   like to read others papers so please share.  A Positive Influence on My Beliefs  	I have been raised to believe in the importance of Christian values that have shaped me into the moral person I am today.  After reading The Great Gatsby by F. Scott Fitzgerald and encountering the degradation of values concerning upper-class society in the 1920â€™s, I have been influenced to strengthen my moral values. 	Fitzgeraldâ€™s novel is portrayed through a description of two wealthy Long Island communities where the center of attention is the Gatsby mansion where â€œThe Great Gatsbyâ€ holds his famous parties where Fitzgerald focuses his novel on the characters of Tom and Daisy Buchanan and Nick Carraway(the narrator).  Fitzgerald uses the beliefs of these characters to mock and show the beliefs of 1920â€™s upper-class society. 	Jay Gatsby is an â€œOxford manâ€ who, after fighting in World War I becomes fabulously wealthy through criminal business to pursue his love-Daisy Buchanan.  Gatsby is portrayed as a mysterious figure who everyone admires for his extravagant amount of wealth.  Gatsbyâ€™s attitude toward money is shown in a quote that is taken from a flashback on Gatsbyâ€™s life stating  â€œTo young Gatz, resting on his oars and looking up at the railed deck, that yacht represented all the beauty and glamor in the world.â€  The authors reasoning for showing the materialistic attitude is for the reader to make the connection of how money can corrupt oneâ€™s thinking.  Gatsby, like many other characters in the novel, is so engrossed in wealth that it seems to be the sole thing that gives him satisfaction in his life.  	Gatsbyâ€™s obsession with wealth is nothing compared to the attitude of Daisy Buchanan, a longtime lover of Gatsby.  Daisy Buchanan is not interested in love, but completely in material possessions.  Daisyâ€™s attitude is best described in a quote when Daisy is admiring en extensive collection of Gatsbyâ€™s shirts.  She states,  â€œTheyâ€™re such beautiful shirts,â€™ she sobbed, her voice muffled in thick folds.  â€˜It makes me sad because Iâ€™ve never seen such-such beautiful shirts before.â€  Fitzgerald is able to portray a multitude of character insights in this small quote.  First and foremost Fitzgerald is depicting Daisy as a fool for being so engrossed in a collection of shirts.  Also, the author gives insight as to what Daisyâ€™s true love is-wealth.  Daisy married her husband Tom because he had money, not because she loved him as a person.  Tom knows what Daisyâ€™s true love is, and in response began an affair with a woman in the city.  Fitzgerald uses Daisyâ€™s attitude and the relationship between Tom and Daisy to show the corrupt beliefs of upper-class marriages during this time period.  Fitzgerald is saying that upper-class marriages werenâ€™t based upon love-but two selfish people getting what they, without regard to morals. 	While most the characters are immersed in materialism, Nick Carraway has contrasting beliefs.  Nicks beliefs are based upon morals, not money.  In the end of the novel, Nick makes the decision to move away from the upper-class society of the East, and move back to the Midwest where morality is more important than money.  Nick describes his decision when saying â€œ...the East was haunted for me like that, distorted beyond my eyesâ€™ power of correction.  So when the blue smoke of brittle leaves was in the air and the wind blew the wet laundry on the line I decided to come back home.â€  Nick makes a decision to get away from what he knows will influence him and eventually corrupt his beliefs. 	Through analyzing the beliefs of these characters, my own personal beliefs have been influenced greatly.  Fitzgeraldâ€™s portrayal of a corrupt upper-class society made me realize that I am so very fortunate and that I take for granted the wealth that I am immersed in.  I am far from a Daisy Buchanan, but am still caught up in material possessions from time to time.  After reading this book, I am sickened that the upper-class in this book wasted their lives living a life where they were unable to love one another normally because they were so greatly influenced by material possessions.  What is the point in having an extravagant shirt collection?  Where is the need to host a party every weekend that leads to a myriad of people being plastered with alcohol.  It is a complete waste that I donâ€™t ever want to be a part of.   I have developed such strong beliefs from reading this book that instead of spending my money on upgrading my car this summer, it will go towards something useful.  I am so very glad I read this book, for it opened my eyes to what is really important in life and that is not possessions, it is giving back when I have the ability to share my wealth. 	 SO THERES MY PAPER.......WHAT DO Y\'ALL THINK?  '),(10037,1488330,'26,May,2004','2004-05-26 00:00:00','   First of all, I have to say that Pat you couldn\'t explain Harry Potter any better.  Something about the plot captivates the reader, but I have never stopped to think about what it could be... How could I when I wasn\'t even able to stop to put them down?  Being a person that has read all of the Harry Potter books countless numbers of times, everytime I read them I pick up on another clue that has a link to a later point in the plot.  You don\'t even understand how excited I was to realize that Moony, was because he was a wearwolf, Wormtail was because he was a rat and had a tail like a worm, Padfoot the dog, and Prongs was the stag.  I seriously set down my book and was amazed... Maybe it was a little more obvious for others, but when different things in the book all of a sudden click and come together and you get the feeling that a light bulb just came on, its irreplaceable.  That\'s why these books are good for all ages, younger kids wouldn\'t necessairly make all of the connections but the mere plot and setting of a magical place that all of us wish we could be a part of would be enough to draw them in.  While adults and teenagers are able to make all of the small connections that makes for a thrilling read. As you can tell, I\'m absolutly in love with Harry Potter, and June 4th is marked on my calender! I can\'t even wait...  Now for a more serious side, Ryan I can see exactly what you mean about Their Eyes Were Watching God.  It definitly focuses on how money and status are not everything in a relationship and you can\'t \"buy\" love.  Although it\'s always said, but it\'s true, love only comes on its own.  Janie could have had anything she wanted when she was with Jody, but the love was not there.  However when she was with Tea Cake and didn\'t have a big house and everything else she was never happier. It just goes to show that love is the most important part of relationships and nothing can replace it.   - Rachel Penny  '),(10038,1488330,'26,May,2004','2004-05-26 00:00:00','   The most respected black writer of the Harlem Renaissance, Richard Wright, said of Zora Neale Hurstonâ€™s novel Their Eyes Were Watching God, â€œ[it] carries no theme, no message, no thought,â€ and after reading the novel for myself I came to the conclusion that Mr. Wrightâ€™s perspective and preformed filters limited his ability to perceive the delicate and emotional subtleties of the story.  Their Eyes Were Watching God records one womanâ€™s pursuit of love, a venture which is beautifully complex and unexpectedly educational.  Unlike a majority of black writers in the 1930s who created serious, in-your-face literature, Hurston chose to write a novel that extended beyond the black protest tradition and connected with readers on a level unaffected by race or gender.  Hurston does not blatantly broadcast an image of black struggle as her early critics would have preferred; instead, she unfurls a story, revealing intricacies of love and emotion that many men are never privy to.  Although writers such as Richard Wright believe love to be trivial in comparison to the awareness of black oppression, Hurstonâ€™s novel should be commended for discussing unique and unexplored topics.  Their Eyes challenges the notion that riches and status alone can attract a woman to a man, and reinforces the message that a man must respect a woman for her mind for a relationship to be truly fulfilling.  The emotional journey of Janie Crawford, the female protagonist, encouraged me to believe that true love is attainable through perseverance, and that the tragedy of loss is outweighed by the memories of time spent with a soul mate.   By: Ryan E  PS - It\'s a shame this book was out of print for 30 years  '),(10039,1488330,'26,May,2004','2004-05-26 00:00:00','   Hi everyone! I have finally been given permission to post on our class blog.  I just wanted to make a quick post and say hello.  The Catcher In The Rye was a quick, and enjoyable read.  I thought this book was very depressing, which, I\'m sure, is a shared opinion between those of us who have read it.  Holden, the main character, scared me quite a bit.  Throughout his decline, I began to notice quite a lot of similarities between he and I; this is scary because Holden loses control of his life.  I\'m curious to find what other people\'s opinions are of this book, and if they felt any connection with Holden.  '),(10040,1488330,'26,May,2004','2004-05-26 00:00:00','   I\'ve mentioned how much I enjoyed the Da Vinci Code on this blog before and Pat\'s post about Harry Potter is similar to how the Da Vinci Code is affecting adults.  The book isn\'t too long, and isn\'t hard to read, but it\'s not a ground breaking novel either.  The subject matter I think is what has made the book so popular at first glance.  Everyone knows about the Holy Grail legend, Leonardo Da Vinci, and secret societies.  The success of the book builds upon itself, like Harry Potter, word of mouth causes a snowball effect on popularity.  But I digress, I just hope the movie is as good as the book.  '),(10041,1488330,'26,May,2004','2004-05-26 00:00:00','   With the approach of the latest Harry Potter movie, The Prisoner of Azkaban, I would like to take some time to reflect upon arguably the personally fulfilling series of books I have read. (Don\'t laugh, I know all of you read these books.....and love them)  First of all, the genius of J.K. Rowling is her organization.  How many times have you been reading these books and come across a turn in the plot, and then remembered hearing a small clue earlier in the book or in a past book that led directly to a crucial part in the book.  To clarify that last statement I will give some examples.  EX 1: Harry has to live with the Dursleys-find out later this gives him protection when he\'s away from Hogwarts.  EX 2: Ron\'s rat Scabbers-find out later it is Wormtail, supporter of Voldemort who framed Sirius and lived with Ron and his family to be able to hear news from the wizarding world concerning Voldemorts whereabouts/possible return.   While the plot is deeply woven for a simple children\'s novel, this is not the only reason for the success of these books.  What makes these books so wildy popular is the way in which the reader is so captivated by the setting of a wizarding world in which spells and flying is a reality.  What kid could resist a fantasy like this?  I sure couldn\'t.  I am 17 years old and am still spellbound(pun).  I have talked to adults that are more obsessed than I am.  If anyone caught last weeks TRL on MTV is was all about Harry Potter and kids dressed up in robes and competed to get a role on the next Harry Potter film. One girl participating was able to name the license plate number of the Weasley\'s car in the second movie.....that is seriously ridiculous.  Speaking of ridiculous, the winner was a kid about our age who literally was jumping up and down, crying with joy after he won the contest.  I\'m sorry but after reading Benjamin Franklin\'s autobiography I was hardly jumping up and down with joy.(I am not bashing Benjamin Franklin-just pointing out how a setting can contribute to the joy of reading a novel)  So now that I have gone off on a tangent, I would like to hear from anyone else that has been so engrossed in the novels as I have.  Call me obsessive, I don\'t even care.  Read the books-then call me obsessive.  Yes, these novels will probably not help you get into an Ivy League school but don\'t we deserve a treat after all the reading we have done this year?  Like William Forrester said on Finding Forrester, \"I read the Times for a main dish, and the tabloids for a dessert\"  Go on, treat yourself.  Or even better, start an argument against me.  I would love to hear it. -Pat  '),(10042,1488330,'25,May,2004','2004-05-25 00:00:00','   I\'m 100% in favor of isolating and analyzing those philosophical quotes in the hearts of novels.  My favorite (besides \"my mother is a fish\") in As I Lay Dying comes from the woman who lays dying.  The quote comes at the end of a chapter that deals with the nature of words- how words and names are used to describe emotions and feelings but they\'re always insufficient if there isn\'t first hand knowledge of the experience that the word\'s trying to describe.  As an example...  \"I would think Anse (the name of her husband).  Why Anse.  Why are you Anse.  I would think about his name until after a while I could see the word as a shape, a vessel, and I would watch him liquify and flow into it like cold molasses flowing out the darkness into a vessel, until the jar stood full and motionless: a significant shape profoundly without life like an empty door frame; and then I would find that I had forgetten the name of the jar.  I would think: The shape of my body where I used to be a virgin is in the shape of a            and I couldn\'t think ANSE, couldn\'t remember ANSE.  It was not that I could think of myself as no longer unvirgin, because I was three now (had three kids now).  And when I would think CASH and DARL (the names of her first two children) that way until their names would die and solidify into a shape and then fade away, I would say, All right.  It doesn\'t matter what they call them.  \"And so when Cora Tull would tell me I waas not a true mother, I would thinnk how words go straight up in a thin line, quick and harmless, and how terribly doing goes along the earth, clinging to it, so that after a while the two lines are too far apart for the same person to straddle from one to the other; and that sin and love and fear are just sound that people who never sinned nor loved nor feared have for what they never had and cannot have until they forget the words.  Like Cora who could never even cook.\"  Sorry for the long intro- but keeping that attitude towards words in mind, comprehend the significance of the quote I really wanted to address.  \"One day I was talking to Cora.  She prayed for me because she believed I was blind to sin, wanting me to kneel and pray too, (this is the big part) because people to whom sin is just a matter of words, salvation is just a matter of words too.\"  I read it as questioning a lot of peoples\' grasp on religion.  Sinners who believe that sin is just in words miss the meaning of words, miss the meaning of sin.  Addie really does know sin, she has a child from a man outside her marriage and holds resentment towards her whole family as well as God.  I think this is a quote that can have a lot of interpretation, so think about it and post your thoughts.  Love Rob  '),(10043,1488330,'24,May,2004','2004-05-24 00:00:00','   Catcher in the Rye is such a comfortable read and a great book that many of us at this age can relate to. Even with Holden\'s very personal, eclectic ideas, interrupted and day-dreaming like train of thoughts, my brain can somehow feel a connection. And actually this personal and colloquial style is what really makes this book engaging, interesting, and awesome. I feel like I can easily follow Holden\'s train of thoughts, even when he daydreams from one seemingly unrelated topic to another. That\'s just how the human brain thinks. Salinger constructed such a real, believable teenager figure.  Finally, as Paul advertised so nicely, besides the great New York Times, there is also the great magazine New Yorker everyone should read. Especially in the May 23rd edition in which there was the famous article \"The Gray Zone\" written by Seymour Hersh probing the secret interrogation agencies and facillities that were in widespread use. All of these abuses we are witnissing today directly relates to the intentional or unintentional commands from the Pentagon (including Rumsfeld himself). But yeah, read this famous article \"The Gray Zone\" by Seymour Hersh.  '),(10044,1488330,'24,May,2004','2004-05-24 00:00:00','   Holden reminds me a lot of someone who\'s eyes aren\'t open to the real world.  He\'s had his life handed to him by his parents, and no doubt while he loves them a lot of his frustrations are coming from a lack of direction.  He\'s been directed all his life and he struggles to be who he wants to be.  Because he\'s never worked for anything he doesn\'t understand the value of anything and throws away his extremely valuable education and life his well to do parents have given him.  He judges people and fails to make successful relationships because he\'s placed arbitrary value in some unimportant things.  However I feel Holden has a kind heart, and that counts for a lot.  He\'s simply, uninformed to say the worst, and is a bit odd.  For me, the character makes sense, and I do find the book extremely memorable, much more so than many other novels I\'ve read.  I find the conciseness of the book commendable, Orwell\'s philosophy of saying what you need to say without the fluff is philosophy now too.  Have a nice day bloggers-  Don\'t be a (NY) Post Monkey, read the Times and evolve www.nytimes.com  '),(10045,1488330,'24,May,2004','2004-05-24 00:00:00','   I actually really like the prompts that we have for the writing assignment, they give a lot to write about. I am going to write a parody on Catcher in the Rye, the book itself provides a lot of ideas to write about.  I think one of the major problems with olden is in his believe that people do things for the wrong reasons: not because it\'s what they want to do, but because it\'s what everyone else wants them to do. He expresses it in this quote about lawyers: \"Even if you did go around saving guys\' lives and all, how would you know if you did it because you really wanted to save guys\' lives, or because you did it because what you really wanted to do was be a terrific lawyer, with everybody slapping you on the back and congratulating you in court when the goddam trial was over...?\" And I find that to be true for myself, I sometimes wonder if I want to go to a great school like Harvard just because it\'s a famous school and it will look good on my resume or because I believe that I will actually learn better and enjoy studying there; I have to admit though that often the answer is: I like the name a lot, sounds big and important:(   I was also looking at this quote by Mr. Antolini in the book : \"If you go along with it [academic edu.] considerable distance, it\'ll give you an idea what size mind you have. What it\'ll fit and, maybe, what it won\'t. After a while you\'ll have an idea what kind of thoughts your particular size mind should be wearing.\" I just happen to really disagree with this quote, just because I am against any kind of parameters set on a human mind and consciousness, I believe that there is no limit to our ability to learn and understand new ideas, and so there shouldn\'t ever be such thing as a \"size mind\" ... do you like it small, large, or supersize?  I like to analyze different \"smart\" quotes from books, gives you something to ponder about.  '),(10046,1488330,'22,May,2004','2004-05-22 00:00:00','   Ryan, I agree that more guys should read Their Eyes Were Watching God. Men and women have different perspectives on life and think completely differently. However I think Their Eyes Were Watching God allows the reader to get inside the mind of a women and how they view relationships. This could be very benefical for guys that read it to understand how women think... It\'s also a good book besides that and isn\'t the \"girl power\" book you would think it would be. I highly recommend it.  '),(10047,1488330,'21,May,2004','2004-05-21 00:00:00','   Pat, I agree with you that Gatsby is a success as a story critical of the society during the 20s, but the only analogy I can make to the book is that of a turkey that is served for Thanksgiving dinner, it doesn\'t have its feathers and it doesn\'t look as beautiful as it did when it was alive but it\'s still a bird, so is this book, it has a story but it lacks the \"feathers\" - it lacks commentary and reflections of the author; it\'s just a story with a very interesting plot and a straight-forward message of society being corrupted and fake. And it\'s true that maybe the story doesn\'t need the extraneous ornaments but it\'s just much more engaging to see what the author himself has to say about the story, it creates a conversation as you read the book between you and Fitzgerald.  Rob, sorry but I didn\'t find AILD yet, so I started on Their eyes were watching God, but I will get to it as soon as I find the book. You should start though on Fountainhead, that book is really hard to read, partly because it has so much philosophy and such a complicated plot in it. But you definitely should read it, I really need someone to discuss it with:(  Discussing Catcher in the Rye today helped a lot, there were a lot of things that I haven\'t even realized as I read the book, such as for example the fact that Holden ends up in a mental hospital or that he is crazy all along in the story. As I read the book I had this feeling that he was a bit strange, but because the book was written as his own personal thoughts I just took it for normal that his mind wondered all over, I sort of think we all have that in our mind where we think of random thinks during the day once in a while. I also heard a lot of criticism of the character, but I still believe that he is a really nice person that has just come to a state of dissatisfaction with the society around him; he does exaggerate a lot and he seems to be critical of everyone and everything he sees, but it sounds like it all comes from bitterness and depression of his character. Maybe he wants to be a catcher in the rye because he knows that there is no one else to catch him if he were to jump off a cliff. It\'s strange how the more I think about him the more depressed I become myself. Ducks are lucky, they can just fly away whenever it gets too cold ...  '),(10048,1488330,'20,May,2004','2004-05-20 00:00:00','   If love is a battlefield as Pat Benatar famously said, then Their Eyes Were Watching God is a top secret report addressed to the Female Commander that was intercepted and expressed mailed to the Male General.  Zora Neale Hurston writes about a woman named Janie searching for her soul mate and finding a few duds along the way.  In the process, Hurston provides insight into the female perspective in a way I\'ve never encountered.  When women speak of love it seems as though they carefully construct a model of an ideal relationship, but the honest solitude of writing allows Hurston to leak key secrets about the art of war/love.  A recurring metaphor in the novel compares perfect love to the blossom of a pear tree, and through the exploits of Janie, the protagonist, the author defines the love blossom as a relationship in which a man treats a woman as his equal and respects her mind, all the while making her glow from every fingertip and toe. Easier said than done? Absolutely, but Their Eyes chronicles the mistakes of foolish men, writing a virtual how-not-to book on love.  I feel I\'m a more complete person after reading it because I attained a deeper understanding of female emotion and compassion.  I believe Their Eyes is a book that should be read by all men because it teaches how women feel and why they feel the way they do, key bits that are rarely divulged and usually left to be deciphered through telepathy, a skill that few of us poor saps are granted.  Zora Neale Hurston\'s novel may be about encountering womanhood, but itâ€™s a worthwhile read for everyone because it demystifies the differences between the sexes and could possibly save or prevent a future dysfunctional marriage.   -By: Ryan Elliott   PS - I had no idea anyone could think as beautifully as Hurston does in this book  '),(10049,1488330,'19,May,2004','2004-05-19 00:00:00','   I\'m coming to the end tonight Beck, but there is certainly no doubt in my mind that each one of those Bundren\'s is crazier than the next.  I certainly enjoyed kickin it w/ you and the crew at Mr. Skyview, and I\'d like to commend Divcom for raising 6 g\'s for Make a Wish Foundation.  Natalya; Fountainhead is Ayn Rand, is it not?  I\'ll read it if you read AILD and we\'ll have a two book discussion going on.  Pat, I\'m in agreement w/ you about Gatsby.  One of my favorites of the year... I really enjoyed a fictional approach to social commentary.  Not really satire, but 100% effective- a challenge to the alcohol consuming socialites who had lost the traditional values of the turn of the century.  It\'s a direct insult to those who live that life.  And the whole time there is the allusion to \"big brother\" with the doctor\'s eyes.  My take is that although the socialites attempted to cover their moral discrepencies in speakeasies and  private parties, it was clear to everyone what was really going on.  Just a thought, no evidence to back it up, but I think it sounds good.  One month of school left!!!  Love Rob  '),(10050,1488330,'19,May,2004','2004-05-19 00:00:00','   Who says The Great Gatsby is a boring book?  I have heard that from countless people who have seen me reading the book and its not true.  The novel, like other successful American Lit. novels, criticizes society through the experiences of a narrator that is a central character in the novel.  The Great Gatsby criticizes the \"roaring twenties\" by depicting the upper class\' lifestyle that consists of breaking prohibition laws, having affairs, and murder.  Throughout the book I felt a sense of this lifestyle that I can only describe as \"hollow.\"  That may not make sense, but it seemed to me as if the novel showed the upper class as not having a role in society, just partying and living a life that didn\'t have a significant meaning.    Going back to my thought on how this novel is successful due to its criticism of society, I can draw many similairities between this novel and Huckelberry Finn.  Both novels consisted of a central character that had the role of the narrator, both mocked society of the time period, and both novels a relationship between the narrator and another central character that grew as the novel progressed-both resulting in the parting of these two characters at the climax of the novel.    Maybe there are some other thoughts on this novel.....anyone?  -Pat  '),(10051,1488330,'19,May,2004','2004-05-19 00:00:00','   Okay, I just finished \"the catcher in the rye\" and I don\'t quite know what to make of it. This blog entry will double as somewhat of a discovery process (hopefully) for me and I will hopefully be able to come to some sort of conclusion about some aspect of the book. If I begin to ramble I apologize.  First I\'d like to talk about the significance of the title. When his sister asked him what he liked I was anxious to hear his answer. Immediately he ponders some off-the-topic idea and comes back to the topic after his sister says, \"You can\'t even think of one thing.\" He responds with, \"Yes, I can. Yes, I can.\" He than talks about his brother and the enjoyment he gets out of talking to his sister. These things he comes up with are so simple. Then he tells about how he simply wants to be a catcher in the rye. I like the connection Natalya made with the James Castle and the \"catcher\" Holden. This seems like one of many intricate connections that Salinger makes. Finally I think the title signifies he wants a simple life with a simple purpose. Now I finally see the connection with his displeasure with \"phonies.\" Phonies to Holden suggest some sort of unneeded attention aimed at bettering the perception other people have on them. This adds to the idea that he wants a simple life that is not fake and that has a purpose. Obviously one can see that being \"phony\" has no \"real\" purpose. This contributes to his dislike of the idea.  One thing I noticed was the abrupt conclusions Holden comes to when looking at different situations in his life. It\'s hard to find examples, a few off the top of my head are: the inferiority complex and the conceited men, the kids always having to meet their friends and the yearbook handsomeness. He even ends his story with a simple conclusion of his.  Hmm... That was somewhat helpful for me. I hope I didn\'t sound like a rambling idiot. Oh yeah, does anyone see the significance with the ducks? ~Frank  '),(10052,1488330,'19,May,2004','2004-05-19 00:00:00','   I\'m reading \"AILD\" as well, Rob, and i quite agree; it is fascinating. Have you got the end yet? It\'s CRAZY. Sort of.... unexpected anyway. Mrs VA introduced the book by telling us that the Bundren family was simply unintelligent (\"really stupid,\" i think she said) but i\'ve come to a different understanding of them. (No offense to Mrs VA - I was just using you as a handy introduction-type idea). Anse is the only person in the book who strikes me as truley stupid, but i think his stupidity comes from his country stubborness and pride. The rest of them are pretty much just pure crazy as opposed to dumb. Cash pulls through in the end as the sanest of the bunch - too bad his leg is rotted off. I think that Darl\'s developement is very interesting; his narrations retain some of his characteristic vocabulary, but he is badly cracked by the end. And i see what you\'re saying (Rob) about Vardaman\'s run on excitement, but i think it runs right on through to insanity. (conclusion)  -Beck-  '),(10053,1488330,'18,May,2004','2004-05-18 00:00:00','   Hey Rob, I am going to start As I lay Dying tomorrow, I\'m not sure if we will have any more time for discussions, but I would certainly love to talk about the book. Discussions are great, as long as people are interested in talking about a book.  I personally recommend everyone to read \"Fountainhead.\" Every time I come back to reading this book it elevates me to a totally different level of understanding of life, it\'s so crazy, it\'s like a philosophy book but the only philosophy is in the lives of the people that are in the book, I mean it\'s just great! You have to read it, there are no words of explanation for this masterpiece.  People \"die with every day that passes. When you meet them, they\'re not what you met last. In any given hour, they kill some part of themselves. They change, they deny, they contradict - and the call it growth. At the end there\'s nothing left, nothing unreversed or unbetrayed; as if there had never been an entity, only a succession of adjectives fading in and out on an unformed mass.\"  '),(10054,1488330,'18,May,2004','2004-05-18 00:00:00','   I\'m glad everyone has found a book that\'s suited them for this most recent project.  Having read both Gatsby and Catcher, I find it kind of fun to play along as you get to know Holden and slowly resent 20\'s life.  Right now I\'m reading As I Lay Dying, and it\'s like nothing I\'ve ever read before.  The style is insanely unique (albeit a little difficult to get used to) as Faulkner switches narrators from one miserable member of the Bundren family to another with each passing chapter.  Faulkner gives real personality to the characters not by description, per se, but by narrating  as they think.  For instance, Vardaman is the youngest of the characters,  and his narration is distinct for his simple vocab and run on sentences that accurately depict the excitement of a child.  Darl on the other hand is viewed as a deep thinker so his vocab is more expansive and Faulkner really tries to play around with language and philosophy under the pretense of death.  While I rejoice that my classmates enjoy their respective novels, I\'m slightly disappointed I have but one compatriot that can share the sentiments of my novel.  If you have some free time this summer, AILD will keep your mind engaged.  Keep up the great work as the instinctual countdown in your brains approaches june 23!  Love Rob  '),(10055,1488330,'18,May,2004','2004-05-18 00:00:00','   Hey Nova, thanx for joining us, it\'s been getting rather quiet here lately.... The Great Gatsby is not that great after all, I guess it\'s just not my kind of \"great literature.\" First I though I was missing something in the book, but now I think that it\'s the but the book that\'s missing something... The author states his doctrine on the first page: \"whenever you feel like criticizing anyone... just remember that all the people in this world haven\'t had the advantages that you\'ve had,\" which dictates the rest of the book making it \"lack\" any kind of criticism expressed by the narrator or any \"reflections\" upon the issues raised by the actions of characters in the book. He simply leaves it all for the reader to interpret and draw her/his own conclusions. I understand that\'s great in the fact that it makes you think for yourself about what you read instead of just reading off the dogmas that the author provides you with layed out on a silver plate. But my problem is that I need someone to discuss the book with, even if it be the author himself through the context of the \"sophisticated sentences\" he writes down in his book, or his thoughts.   That\'s why I think I liked Catcher in the Rye so much more, the author just told you everything he thought, even though all the digressions could sometimes get distracting. It\'s strange how the book just gives you three days of this boy\'s life but you know everything about his character you could ever find out. I first thought he was rather sardonic and just negative in general toward life and people, all of his comments and criticism, but toward the end of the book I couldn\'t help but love this boy. The way he just tells his little sister how he wants to be a catcher in the rye.. \"standing on the edge of some crazy cliff... I have to catch everybody if they start to go over the cliff...I have to come out from somewhere and catch them...\" and just makes you feel like you know that he would have caught that boy that jumped out of the window and that he would have kept that young girl from selling her body and that he would have donated all of his money to those two nuns if he could... just the simple things he says and does make him so unrealistically true and \"good\"... I am trying though to figure out what is the significance of the carousel at the end of the book... why does it make him so happy?  '),(10056,1488330,'18,May,2004','2004-05-18 00:00:00','   Hello all.  It\'s Nova.  How is everyone\'s book going for them?  The Great Gatsby is pretty good.  I thought it would be boring but I actually got into the story, although some parts of it were confusing.  The people in our discussion group were having a hard time trying to figure out Dr. T.J. Eckleberg.  Does he represent God?  Nick?  Or just a random billboard, watching everything?  Anyone have any ideas?  Also, we had a group consensus that we dislike Tom and Mrs. Wilson (well, mostly a consensus).  I have to go, but I\'ll come back later! Bye  '),(10057,1488330,'18,May,2004','2004-05-18 00:00:00','   Like Corey, I am also reading Their Eyes Were Watching God.  I really enjoy reading it. It\'s more relaxing of a read opposed to having to analyze a piece for rhetorical strategies.  It\'s nice to have no pressure after the AP tests and be rewarded for spending the first three fourths of the year challenging ourselves and preparing for the \"big test.\"  '),(10058,1488330,'17,May,2004','2004-05-17 00:00:00','   The blog\'s been rather quiet lately.  I\'m finding the catcher in the rye an easy and memorable read.  I personally find the characters very realistic, especially Holden, I enjoy how the entire book is simply living the life of somebody who\'s our age, but engrossing simply because Salinger paints such an accurate portrait.  But I\'m not hearing unanimous agreement from others, since people in my small group are finding different views of the book.  But what\'s really on my mind isn\'t the catcher in the rye, for me it\'s the aftermath of the ISEF fair.  Aside from meeting kids who I\'m sure will be extremely successful before they\'re 30, I saw a group of 4 of them on my favorite business commentator\'s show, Lou Dobbs, talking about their projects tonight.  I wish I could accomplish as much as they have already.  I\'m having the hardest time finding a good idea which real practical applications.  Getting started is always hardest but I digress.  I just feel like there is so much any of us could be accomplishing and I hate feeling like a sponge.  But there\'s always next year!  Or this summer I should say!  -Read a real newspaper, go to www.nytimes.com  '),(10059,1488330,'16,May,2004','2004-05-16 00:00:00','   Hey everyone whats up?  I like how we are ending the year with fun reading and basic skills about research and other things, after taking the AP test it should be this way.  I am reading Their Eyes were Watching God, it is a really good book and I highly reccomend it, because it\'s a story that tells of a woman who had tough relationships, and you may know people like this girl, and eventually meets this awesome guy named Tea Cake, I mean with a name like that why wouldn\'t you want to read the book.  -Corey Furin-  '),(10060,1488330,'16,May,2004','2004-05-16 00:00:00','   I personally think everyone should congratulate Rob for his 2nd place at MUN this Saturday! He really did a  great job and it is thanx to him that our team got all the first places:) And I invite everyone to join our club, you never know what hidden talents you have.   Going on to Gatsby, I see the irony of Gatsby\'s \"friends\" not even coming to his funeral and the fact that his beloved Daisy runs off and he has to pay for her mistake with his own life. The scenes of the party in his house are truly what the 20s were like for the elite and give me just a general visual of pigs partying in dirt, a feeling of disgust. Also disturbing is the fact that Tom\'s mistress Myrtle is like that puppy he bought for her on the street, and that he can throw her away as easily as that puppy. I loved the description of women of that time as elegant and mysterious which then created a special contrast with their greed for money and superficial expression of love. I mean I see a lot of themes going on in the book and they are quite obviously related to materialism, but it feels like there is something missing in it, like it\'s too easy or maybe I just expected too much from the book after hearing all these great things about it. I liked the book, but I don\'t feel like it changed my life or like it is a masterpiece of literature.  '),(10061,1488330,'15,May,2004','2004-05-15 00:00:00','   Hey Rob, in response to your welcome to me as to returning to the \"real world\", I have a few words to say. Forgive me if I sound like a nerdy bigot here.  First of all, I have to say that yes, certainly, I\'ve got your note you so graciously left on my exhibit booth. I have to thank with my deepest appreciation to all and everyone who came to visit Intel ISEF on Thursday or Friday.   With all the scientific jargons EVERYWHERE in the beloved ISEF exhibits hall that would probably intimidate a good deal of people just with their first step into the ISEF exhibit hall, yes I do agree the environment I was immersed in for the past whole week is out of touch with the \"rest of the society\". But on a different note, we had the wonderful opportunity to meet people from ALL OVER THE COUNTRY AND THE WORLD.....We were chatting about not just what scientific discoveries, methods, projects we worked on over the past years, but we ISEF finalists also discussed politics (yeah discussing politics with people from other countries is kinda interesting), our hometowns, schools, and just our personal nerdy lives in general. We also have had LOTS AND LTOS OF SOCIAL LIFE....more social life in the past week than I\'ve ever had in my life. Just being in the convention center makes you feel more mature, and everything was in adult settings and business attires. All the strict rules, politeness, civilness in conjunction with meeting professors, Nobel Laureates, leading scientists and teachers of all fields that represent modern society. True, all of these cameras, press, Nobel laureates are probably very special, but these famous people (both adults and ISEF students) I met do constitute part of our world and the contributions of the scientists is what move the society forward, not the verbatim politicians. Besides, I\'d rather be in a environment of overly academicness, formalness, and overly famous people to see what really the world of brainpower is like than be in an environment that encourages some kind of phobia to academia and modern sciences (We had some fun times chuckling and making fun of Bush on the bureaucracy and the political barriers to modern science advances, such as the stem cell research, and it was started by Nobel Laureates) YEAH SO ISEF ALL THE WAY....5000 times more REAL WORLD! lol.   And now I\'m looking forward to catch up all the work I\'ve missed the past week lol. Catcher in the Rye seems to be a very comfortable reading right now. I\'ve looking forward to discussing the book with other interested readers.  '),(10062,1488330,'15,May,2004','2004-05-15 00:00:00','   Wow, it\'s been a few days since I last checked the blog and I feel so out of touch... so I have some preambulatory clauses to address before I get to the real, English-related topic of the post.  First and foremost, I\'d like to  CONGRATULATE the members of Skyview MUN for their various awards and other successes at the district tournament this weekend, RECOGNIZING the efforts of the Russian Federation in successfully warding off the political opposers of the Rodina FURTHER RECOGNIZING that said efforts resulted in both best team in General Assembly as well as best team overall and, LOOKS FORWARD TO seeing the big ass trophy that results  I really had a good time and look forward to participating in the future.  Secondly, Thomas, welcome back to the real world.  I hope you got my note when I went to visit you and you weren\'t there, but if you didn\'t, I came to support the scientific mind Natalya has recognized as a Hottie.  Now to the Great Gatsby... I feel like a major nerd here, but I really enjoyed the book and feel I was able to track the author and accompanying symbolism.  (This is where I feel like a nerd)  I would therefore like to reference the independent reading report I did on it earlier on in the year; 	  The Great Gatsby is considered the great American novel because below the cloak of an extravagant love story is the erosion of American society in the Roaring 20â€™s, or â€œJazz Age,â€ as Fitzgerald dubbed it.  Gatsby himself is the symbol of the American dream as Fitzgerald saw it:  heâ€™s rich and therefore a socialite, he throws lavish parties where hundreds of â€œfriendsâ€ come to his mansion to glorify the vices of society, namely alcohol and sex, and he pursues unrealistic fantasies on a whim, attempting to bend reality to fit his sensibilities.  Gatsbyâ€™s love interest, Daisy, is respectable and accountable, besides being desirable.  When Gatsby tries to woo Daisy, and runs into obstacles, it is apparent in Gatsbyâ€™s demeanor that the American lifestyle is unraveling, as demonstrated in his cessation of party hosting.  In the end, when Gatsbyâ€™s protection of Daisy results in his death, it is apparent at his funeral when none of his â€œfriendsâ€ come to mourn how the American lifestyle is truly viewed; as base, vile, and taboo.  Long story short, Gatsby was symbolic of the twenties life-style, (as AP US History should have taught you) a life-style that Fitzgerald was against.  Viewing Gatsby as an embodiment of the Roaring 20\'s may alter your position toward him and his actions as you read.  Some other interesting symbolism was the role of the Green Light that was Gatsby\'s hope in reuniting with Daisy, and the role of the Dr.\'s Eyes.  I really really enjoyed Gatsby, so if anyone who is reading it wishes to discuss it extracurricularly, I\'d love to participate.  I\'d also like to note that Faulkner\'s As I Lay Dying is an entirely different style than I believe I\'ve encountered before, and the humor and complexity of the book make it a lot of fun as well.  I love that post AP test, there is still so much interest in American Lit.  Love Rob  '),(10063,1488330,'15,May,2004','2004-05-15 00:00:00','   HEY MS.VA AND ALL MY AP ENGLISH FRIENDS!!!!!!!!!  I just would like to say that Thomas Chen had had the wonderful opportunity to experience through one of the best week in his life. Intel International Science and Engineering Fair is truly world class and the experience once in a life opportunity.  The awards and free stuff we got were only a temporary excitement, but the most valuable experience was making great friends from all 50 states and 40 countries around the world!!!!!! It\'s 100% probable that one or two friends you make in ISEF will be in the future leaders of the world, billionaires, or Nobel Laureates. The country and state pin exchanges, the plethora of ISEF ceremonies, all the social events were truly amazing. Furthermore, we the ISEF competitors got to be in a press like room with the Nobel Laureates as a panel. We were like the press asking the prestigious Nobel Laureates questions and they would answer us right there. It was truly astounding.   Yeah I guess the most important idea I\'ve learned and experienced through this ISEF event is that seriously, there are sooooooooooooooo much talents and brainpower out there in the world, even with the everyday person around us. The project I did was just one grain of sand morphed into this vast lake of sand dunes stretching thousands of miles.  All the 1400 projects there were simply impressive.  I wil see everyone on Monday.   Thomas Chen  '),(10064,1488330,'13,May,2004','2004-05-13 00:00:00','   I agree with you Natalya. The ending of \"The Great Gatsby\" left me hanging also. I feel like I missed some hidden messages or themes. Mr. Selby, my ASL teacher, said he loved the book and that it made such an impact on him... Yeah, I think I\'m missing something. I liked Fitzgerald\'s imagery throughout the book. It was really amazing. There were some very beautiful passages. I need to do some discussing with my group.  '),(10065,1488330,'13,May,2004','2004-05-13 00:00:00','   I just got back from the Talent Show and I was really impressed by the last song by the White Bros, piano and violin really made it just ... special, so props for Ben Perry and Matt Wu. Sorry Rob, but Matt WU is way hotter than Thomas, although I have to agree that Thomas by all parameters is qualified for the title of a hottie.  I just finished \"The Great Gatsby\" today and I have kind of mixed feelings toward the book. I have heard a lot of people say good things about it and on the back of the book it says \"American masterpiece\" but my question is: what is the point of the book? I just feel dumb right now cause I didn\'t really find an actual point, or an underlined message or some \"truth\" or a \"revelation.\" The book was really well written and was engaging, I thought it was great from the very beginning, but I somehow expected more from it and I guess the end just threw me off, as if Fitzgerald just left the book unfinished. I didn\'t even get it first that the last page I read was the last page of the book, I expected a \"to be continued...\" So I am really looking forward to a discussion of the book as soon as everyone has finished it, meanwhile Catcher in the Rye seems quite interesting, I like its colloquial style and a somewhat humorous tone, the author is great in his ability to delineate a very vivid and a realistic picture of this boy, I like the beginning: \"I\'m not going to tell you my whole goddam autobiography or anything. I\'ll just tell you about this madman stuff that happened...\" a nice start for the book, sort of sets you up in the right mood ... enjoy the show ... .  '),(10066,1488330,'11,May,2004','2004-05-11 00:00:00','   I was certainly happy to hear absolutely no strange Skyview news on Monday morning.  After a big weekend like prom, I always wonder...will we be having an emergency faculty meeting?  Thankfully the answer was no.  I\'m glad last week\'s presentations and class discussions made an impact on you.  Let the message stick.  I have been so pleased to see how well you\'ve all supported each other through AP testing.  You have worked hard and applied yourselves, and through that process, you\'ve been united by a shared goal.  Hang onto this; let your mutual respect for each other be the prevalent theme of your senior year.  The worst part of high school is living with the horrid restrictions that come from cliques; if you can see past those imagined lines, you won\'t need to live your life with a countdown.  I encourage you to come to school each day with the intent to build relationships so that by the time your ten or twenty year reunion rolls around, you\'ll genuinely want to go and can show up in a non-inebriated state.  Thanks, everyone, for your honest evaluations of class.  The consensus of the group was that it would have been helpful to have seen practice exams earlier in the year in order to have some context for our topics of study.  I also realized--far too late--that I could have made our SOAPS/Tone analysis more relevant for you by asking you to write essays that incorporated it.  I made the assumption that this was something you already knew, but Natalya\'s posting made it clear that this was something she (and probably most) had to discover.  My guess is, though, that many people haven\'t made this connection.  So...I know what I can do to improve.  This brings up an interesting observation in response to Thomas\' posting about teachers and students.  Teachers think they\'re invaluable...students think they can independently learn everything they need...I believe that neither is true.  Bad teachers limit good students...A teacher can only be as good as students allow her to be. The best classrooms exist where everyone realizes that mutual responsibility is necessary.  So, as always, thanks for coming in with curiosity and a good work ethic.  As for you, Thomas...IF you have time to read and think about Catcher in the Rye, here\'s what you need to know.  Sydney Clos, Mina O\'Brien, Matt Goosen, Mitch Amort, Chris Tamura and Tyson Cecka have signed up to read the book.  I assume you have Tyson\'s e-mail.  If so, find out what his group has decided upon for their reading schedule.  Wednesday and Friday will be discussion days.  On Friday, students will be turning in their book notes, which can be written in journals.  (Obviousely, you can turn in your work late.  We\'ll talk when you return to school.) The directions are on the syllabus.  Do you still have it?  If not, let me know and I\'ll get the directions to you.  We\'re thinking of you with pride every day!  The air circulation system has turned off and I\'m starving.  Time to go.  See you all tomorrow!  Mrs. V-A  '),(10067,1488330,'10,May,2004','2004-05-10 00:00:00','   I just wanted to tell Thomas how hot he was in his Oregonian picture.  With that kind of outward sex appeal I\'m amazed that he hasn\'t pursued a full-time science t-shirt modeling gig.  I know that however hired him would have a customer in me for sure!  Good luck Tom and Skyview in the ISEF comp!  By the way, Catcher in the Rye isn\'t for everyone, if you don\'t like it from the start, you might want to switch because it doesn\'t change much.  However, some people REALLY identify with the desperate feeling of not quite fitting in with the world.  The book didn\'t grab me like Gatsby did, which I really enjoyed, but As I Lay Dying needs some love too, so far it\'s me and Kailie V and I\'d like some people to share with, so think about it.  Love Rob  '),(10068,1488330,'06,May,2004','2004-05-06 00:00:00','   Hi this is Thomas Chen posting from a HUGE GIGANTIC computer lab inside the Oregon Convention Center. Everything is going well so far,and it\'s an awesome place here, although I\'m feeling the PRESSURE of this immense Intel ISEF competition more and more.  I had a wonderful chance already to meet people from all around the country and the world, and I have collected emblems from most states of the US and over a dozen other countries via the student pin exchange program.  I\'m still assuming we are doing the book group. And Mrs.VA, I\'ve decided to read \"Catcher in the Rye\". So I just need to know my book group and our agendas and plans for the week, such as what\'s the project all about; book notes; other works and preparations? Keep posting on the blog and keep me informed through the week.  See ya all next week.  '),(10069,1488330,'06,May,2004','2004-05-06 00:00:00','   Just looking back at this year all I can say is that it flew by fast, way too fast to even realize that the junior year is gone... well almost gone. I think senior year will go by even faster, the more you have going on in your life the faster it seems to flow, every day is inundated with just ... life! I just have this analogy in my head: h.s. is like our salmon creek, college is like an Amazon, and after that it\'s an ocean, and hopefully you get into the warm current;) I am personally looking forward to the end of h.s., I just want to go out into \"real\" life and be sort of extricated from the feeling of being trapped in this tiny box, cause life just has so much more volume to it than we can actually see right now. I think college should be awesome, better classes, more people to meet, better professors, just feeling like you are your own person and you can do anything you want with your life. Actually when I was at Stanford our tour guide told us that students there can get as drunk as they want and have as many parties as they want as long as they don\'t bother anyone and don\'t get out of the dormitories, I think that\'s in a way a safer way to drink if you do drink than going out to the woods \"camping.\" I missed the accident scene yesterday and the assembly today, but I also would like to think VA for talking to us about this issue today, it just makes it so much more personal and real when you talk about it, and I love it how our teacher actually cares about us. Really, thank you.  '),(10070,1488330,'05,May,2004','2004-05-05 00:00:00','   I just wanted to say that I think we had a really good discussion today third period.  It was only some of us, but it was really really nice to talk about current issues within our country with other people my age, and Mrs. VA of course.  I really enjoyed it.  '),(10071,1488330,'05,May,2004','2004-05-05 00:00:00','   I just wanted to commend Mrs. VA on a drunk-driving presentation that connected better with me than the pseudo-accident or marathon assembly.  Personal relationships are what have the most influence on peoples\' lives, no matter how much pop-culture goes into an assembly.  So this weekend be safe, sober if possible, but if you do choose to drink, remember to have a plan, and that drunk-driving is NEVER the best alternative, or acceptable at all.  Once again, great job AVA, have fun but be safe!  Love, Rob  OH, by the way, Thomas mentioned some independent reading?  Could anyone fill me in on what that will entail?  And while I\'m having after-thoughts, good luck to everyone that\'s taking their AP History tests tomorrow.  Happy Review!  '),(10072,1488330,'05,May,2004','2004-05-05 00:00:00','   I just wanted to say that i was quite impressed with the whole 15 minutes thing. The crowd was quiet and respectful in both the commons and the parking lot and there weren\'t too many jokes cracked later on that day. Either something of the message got through to the general mob or i have greatly underestimated their intelligence. ...well, probably the former of those two options.   I also wanted to thank Mrs VA for the discussion we had in third period. First perioders truely missed out.   hmmm, that was oddly written. Oh well. --beck  '),(10073,1488330,'04,May,2004','2004-05-04 00:00:00','   After two AP tests, I\'m absolutely DYING to discuss some multiple choice questions... please!  Somebody!  Did you get D. on question 28 or A. on 17?  All this anti-cheating just sounds like a challenge to me.  So here\'s the plan for u.s., we all fabricate test booklets, replace them with the real booklets when our booklets get picked up, and mail the originals back to the College Board.  Our goal is to have to take next year\'s tests in facilities with prison-like security.  We\'ll call it operation Rebel Yell.  Anyway, one day you\'ll look back at high school and just admire your naievity and such, and you\'ll be pleasantly nostalgic, at that point you\'ll buy yourself a Harley and hit the road for a couple months.  As soon as H.S. ends I\'ll be anticipating my mid-life crisis (maybe crises), because maturity and responsibility are WAY overrated.  ENJOY IT WHILE IT LASTS!  Carpe Dium!  Much love to all bloggers for making MY h.s. experience memorable  Love, Rob  '),(10074,1488330,'04,May,2004','2004-05-04 00:00:00','   I don\'t know about all of you, but I\'m REALLY REALLY GLAD High School is almost over................It\'s time for the real college lifelong challenges. Although I have to admit that I\'ve learned and experienced sooooooo much in the last three years.  And Here is a little vivid memory from my three days of AP tests in as many days: \"Remember, the Multiple Choice questions may NEVER be discussed in any shape or form with ANYONE at ANYTIME, failure to do so may result in the invalidation of grades and possible legal actions for COPYRIGHTS VIOLATION!\"  Yeah you realize from the AP test experience how much your mind has been trained to memorize stuff........because i\'ve literally memorized the whole AP instruction script by now.   I\'ll see all you guys tomorrow, the only day I\'ll be in school for two weeks.  '),(10075,1488330,'04,May,2004','2004-05-04 00:00:00','   I totally agree...its amazing to think that once we finish up the last months of this year, we only have one year left.  Then high school is done.  It\'s going by SO fast!  '),(10076,1488330,'04,May,2004','2004-05-04 00:00:00','   As Fleckenstein would say, I pooched that test.  Which is a first class bummer, but I\'m just glad to be finished.  I need to take more english in college anyway.  Anyway, yeah, you guys know we\'re seniors next year.  Maybe it\'s just me, but I still feel like we were freshlings just a short time ago, not 3 years....  '),(10077,1488330,'03,May,2004','2004-05-03 00:00:00','   I share the relief that I am sure everyone is feeling. I can\'t believe how fast this year went by and I am so glad that we were well prepared for the test. I can remember reading all of the documents from or about Indians but that was almost 6 months ago. WOW this year flew by. Thomas you\'re stress level must be through the roof and Rachel that must have been the most fun you\'ve ever had. Hmmmm... Sarcasm. My mind is still spewing literary terms. Ah almost time to sleep. Oh yeah Mrs. V-A,  You Rock . Bye Bye ~Frank  '),(10078,1488330,'02,May,2004','2004-05-02 00:00:00','   Wow, you two must be taking the AP French test also. Congratulations to you two for surviving two AP tests in ONE DAY! That double test experience must\'ve eclipsed any academic strenuous activity you\'ve ever accomplished in your life. I can\'t even imagine how I would survive because I just went home and slept.  For me, the continued stress hormone pumper is still pumping strong doses into my brain, for I have four more AP tests to go, three MORE this week......Yeah I\'m begging myself that hopefully I can survive like you guys too.. And have a happy stress relieving time in English class the next two weeks!  '),(10079,1488330,'02,May,2004','2004-05-02 00:00:00','   Yeah wow... I can\'t even describe this feeling right now... Chad can understand what I\'m talking about. After enduring about 10 hours of testing today and now having it all over.  It\'s undescribable.  The English test itself wasnt too bad, I think I anticipated it to be much worse, considering the practices seemed to be much more difficult than the test itself.  Also, what a relief that one of the essay prompts wasnt based on a passage.  I hope everyone else felt that they did well... Congrats on surviving the AP experience.  '),(10080,1488330,'02,May,2004','2004-05-02 00:00:00','   Sigh..........Relief............And what not..........I can\'t believe three years\' of English language and composition career just culminated in this huge grand assessment exam, and it\'s over in a flash of thought. Seems like I haven\'t done anything significant and it\'s already over......Yeah it\'s time to have a moment of silence........To honor all the three years of learnings we had that prepared us for this exam.  Anyways, how did everyone else feel about the test? (Remember, NEVER discuss Multiple choice questions.......lol).  '),(10081,1488330,'02,May,2004','2004-05-02 00:00:00','   Thanks Thomas for providing that site. SOAPS is where its at. ~ Frank  '),(10082,1488330,'01,May,2004','2004-05-01 00:00:00','   I was reading the free response essays and scoring commentaries that Thomas posted a link to and I just had a revelation: there is a purpose to SOAPS! No really, most people that got 9s or 8s used the SOAPS method, they talked about the subject (themes and ideas), occasion for the writer to write addressing this subject(s), audience whom he/she addressed, what purpose the author had when writing this piece and style of the writing including tone and rhetorical strategies. Now after two years of thinking that soaps is a waste of time I finally come to a conclusion that it is really the best way to organize your thinking and to make sure that you include everything in your essay in order to score all the possible \"points.\" I tend to just randomly name facts as I go, mentioning the style, occasion and everything in a mix, but I think putting info in a coherent and organized manner will show the scoring person that you actually know what you are talking about, you can express in a very erudite and orderly manner and you are not just brain- storming. So my another suggestion is practice soaps, just look at some of the essay prompts on the link Thomas gave us and as you read each one of them jot down just a couple of words for SOAPS that you think you might have used if you were writing the essays at the exam.  Or better yet, relax and forget all about the test until 7:30 on Monday morning:)  '),(10083,1488330,'01,May,2004','2004-05-01 00:00:00','   Hey Frank, go here  urlLink AP Central\'s English Language and Composition past essay prompts. , although you need a collegeboard Login (Which it\'s good and very helpful to have one).  Here is an EXCELLENT article from \"KidSource Online\"  urlLink that provides insights into students\' attitudes in schools with survey statistics and some common misconceptions. All students and especially teachers should take the time to read this article .   This statement below from that article is VERY true:    Popular opinion has it that students\' academic success depends on the quality of their teachers and textbooks. Ask the students themselves, however, and you get a different view. Here is how they account for their academic achievement:   Most students believe their ability and effort are the main reasons for school achievement. By the same token, if asked whether they would prefer to be called smart or hard-working, they will choose smart almost every time. Why? Because they believe that hard-working students risk being considered either excessively ambitious or of limited ability, both of which they would find embarrassing.  '),(10084,1488330,'01,May,2004','2004-05-01 00:00:00','   Hey does anyone know a site that I could go to that provides writing prompts or multiple choice tests.  ~ Frank  '),(10085,1488330,'26,June,2004','2004-06-26 00:00:00','   Hey Everyone, I\'m having a blast here at UDUB.......I was wondering, since I missed the last three days of school, did anyone who\'s taking AP Lit next year get a  specific \"summer reading list\" or summer work that we have to accomplish before september? ......Just curious, otherwise have an awesome summer! And as Rob mentioned, keep blogging THomas  '),(10086,1488330,'23,June,2004','2004-06-23 00:00:00','   hey all you out in bloggerland.  thanks for makin this year the best ever.  keep bloggin on through the summerâ€” Fahrenheit 911 opens this friday, that could provide interesting discussion, but who cares about that now?  everybody have a safe time in your celebrations, remember the drunk driving talk, and i\'ll see you all back in september, but hopefully sooner.  Love Rob  '),(10087,1488330,'18,June,2004','2004-06-18 00:00:00','   Hey All MY AP ENGLISH BUDDIES!  Today\'s my last day of school........I\'m going up to UW for six weeks. I have to say that this year has been an awesome year in VA\'s AP English class. Despite a somewhat stressed out, apprehensive studying for AP Test period, we\'ve had lots of fun in that class. We\'ve grown both academically and socially.   So enjoy the rest of year three days of school.......Make the most out of your junior year.....Great memories will be with us forever. And old friendship and fun times may never be replaced.   Hopefully I\'ll see some of you again next year in AP English or AP COmparative GOvernment class!  FINALLY, KEEP THE BLOG UP THIS SUMMER!!!!!!!!!!!!!!!!!!  IF YOU CAN, BLOG ANY AND ALL OF YOUR SUMMER FUNS! SHARE ALL YOUR STORIES WITH US!!!!!!!  Best Wishes  Thomas Chen  '),(10088,1488330,'17,June,2004','2004-06-17 00:00:00','   we\'re going to try the farewell miss clark party tomorrow.  same plan, same time, i just hope she\'s here this time.  any food or beverage you can bring would be appreciated  '),(10089,1488330,'16,June,2004','2004-06-16 00:00:00','   I have been injected with some sort of spirit now that the only things left to do in English are sign yearbooks and watch Ferris.  Thanks Mrs. VA for an awesome year and such a simple yet appropriate pair of final prompts.   As far as school spirit goes this last year was not only unprecedentend and uncharacteristic of Skyview, it was also PHENOMINAL.  The key was a group of kids, mostly seniors but several juniors and sophomores as well that were willing to set aside their images (rip up their cool cards *wink*) and have fun with outfits, chants, and the good feeling they got when Skyview won.  If we want to have that kind of a fall and winter sports season again it\'s going to take some leadership.  Kids quickly forget how much fun it is to get G\'d UP from the feet up and yell at Prairie high school, but all it takes is a key figure like Jake Valentine this year to really motivate kids.  Cheerleaders keep up the good work, but help is needed to generate school-wide spirit.  Everybody keep blogging this summer!  I\'m pretty sure we\'ll have a summer reading list so we can talk about that, but with each passing day, November draws closer.  Love Rob  '),(10090,1488330,'16,June,2004','2004-06-16 00:00:00','   So this is a really exciting blog, so I\'ve decided to spice it up a bit...or at least try. I don\'t know if anyone has noticed, but our school generally lacks school spirit. I admit, last year was AMAZING compared to every other year I have been at Skyview, but there is still room for improvement. During cheerleading practice we were discussing different ways to implement more school spirit throughout Skyview. However, it\'s going to take a lot more than just 15 girls out on the sideline.  If anyone has any good ideas of how to get more of the student body involved, blog away! It will be greatly appreciated and we can have a super fun year next year being the old men (and women) on campus! Rachel Penny  '),(10091,1488330,'14,June,2004','2004-06-14 00:00:00','   thanks!  '),(10092,1488330,'13,June,2004','2004-06-13 00:00:00','   This was the prompt V-A gave us for the reflection part of the PDP:  How have you grown as a reader, writer, and thinker over the course of your junior year?  Reply in the form of a business letter Address it to Mrs. V-A Attach adn use evidence in the form of an enclosure  Hope that\'s helpful!  See you all tomorrow.  Bethany  '),(10093,1488330,'12,June,2004','2004-06-12 00:00:00','   hey guys...for the self-evaluation that we have to do for our PDP, we have to write about how we\'ve improved at reading, writing, and what was the third one?  if anyone could tell me, i\'d really appreciate it!  '),(10094,1488330,'12,June,2004','2004-06-12 00:00:00','   It\'s very true that we don\'t realize our milieus until we move away from them.  When I moved up here from North Carolina, there was something different.  I couldn\'t really describe what it was, but something in the personalities of people.  There are pretty much the same restaraunts, stores and things to do around here as there are in most areas, so i could never figure out exactly what was different.  Then I figured it out. It\'s the people.  For the most part, people living in the Northwest are more \"natural.\" Like Seattle is known for being artsy, and Portland is more of a naturalist area.   Also, the first thing I noticed when I moved was the California pride.  I never understood what was so special about California. Maybe it was just being in middle school when everyone is trying to find themselves by bragging about everything possible, but California always came up as a bragging right.  It was amazing to me how many people were like, \"well I\'m from California!\" My response was always was \"your point is...?\" To this day I have yet to figure out what is so special about California. Yes, it has great weather for the most part of the year, which is more than I can say for here, but I couldn\'t see myself just packing up my bags and moving down there.  Even given the opportunity to, I declined to move to sunny So. Cal.  Which brings up a whole nother aspect of milieu.  Families. You would be surprised how completely different people live.  Having my mom move down to California this past summer, I was completely excited to move in with the Gilfoys. Even though they seemed like a \"normal\" family.  I realized there is no \"normal\" family. (Kind of like there are no \"normal\" people, like Thomas stated.) I always thought that I just lived in the most \"interesting\" family, but every family has something different about them. After moving in with them, I have realized many small things that I love about my own family and how differently I used to live.  I guess it\'s kind of like leaving my family a couple years early.  Next year when most of you go off to college you think you will never miss your family and you\'re dying to get out of your house. Getting out of the house will be great, but trust me though, you will miss them and all of the small things, especially your own mother\'s cooking.  -Rachel Penny  '),(10095,1488330,'12,June,2004','2004-06-12 00:00:00','   Ok, this blog needs to be spiced up once again. So what did everyone think of the great and rather humorous documentary?   First of all, in the film so often you\'ll see all the \"weird\" cultures, and the people in these communities/sect will be so proud of their culture and heritage and religious practices.   To go off on Mrs.VA\'s commment that \"we don\'t see our own milieu when we are living in it\", I believe this statement to be mostly true. If a person has never moved to a different state/country, then that person will always see himself as a \"normal\" human being, because they are so immersed in their culture in which they were born. But when you move, that\'s when you will discover your milieu, and when you move to a drastically different place you also can more objectively see the milieu of your new home. That happened to me several times already, first the culture shock when I moved from China to California, and then I then I brought an interesting mix of Chinese/California culture when I moved up here to Vancouver Washington.   Here is just a list of the many many milieus of Vancouver, if you don\'t realize already:  1:As Erin mentioned: \"ALL THE SUBARUS AND SUVS!!!!!!\"(this is a very interestiing discovery as I moved here, I guess even Georgia has much less SUVs) 2:Ski and Snowboarding Crazy(Go figure with all the outdoor opportunities here) 3:People drive soooooooo painfully slow and very carefully obeying the speed limit(Compare to California....You\'d be kidding if you sped just 10mph above the speed limit....in California 50 mph 5 lane highway completely jammed with 80mphers) 4:People very much obey the law, the society as a whole is a lot more strict and calmer.(People very rarely question authority here, I can elaborate on that later).  5:The society\'s pace is soooooooooooo slow(downtown stores close before dark, what???? that\'s unbelievable in california) 6:I\'ve never heard the word \"Spendy\" before I moved to Washington 7:Fast food dominates(somehow people are ok with fastfoods here) 8:All these wetlands and storm drainages(which of course the \"rain milieu\" of washington) And there are many many more.............Milieus and these subtle differences of cultures are what\'s so unique about every place, everyone of us......Catch a glimpse of New York, San Francisco, Las Vegas, Los Angeles, Atlanta, Miami, Honolulu, Victoria(BC)......Wow what a range of different milieus.   Too often have I heard people calling each other \"you are not normal\", or people urging me to \"just being a normal person\". But I will often ask back, \"what do you mean by NORMAL?\". \"How do you define normal????\" Because \"a normal\" human being is different in each culture/country, and you have to have a firm grasp of the milieus of that culture to be normal. Like people will always see me tapped on with Chinese culture, even if I realize it or not. And I will not try to act like everybody else just because I want to look like \"a normal person\", because THERE IS NO NORMAL PERSON, especially in America.   The immigration and custom scene where people from all corners of the world feeling all patriotic and excited to become an American Citizen is simply a phenomenal picture patented for America. After all, America is a country founded on uniquenesses and differences and immigrants and we continue to pride ourselves for that. And thus in America, the world \"normal\" is moot and there is simply NO NEED FOR NORMAL! America has always to protrude out from the world as NOT BEING NORMAL!!!!!! In the land of the free, uniqueness is happiness; And differences drive our country forward;  '),(10096,1488330,'09,June,2004','2004-06-09 00:00:00','   Nostalgia as a whole is vastly underrated.  Advertisors recognize its potential, filling ads with subliminally old soundtracks that put you in a good mood for being 19 once.  Nostalgia is all around.  People take pictures so they can look back 20 years from now and remember how good they looked in that sweater, or what they barbequed that fourth of July.  I myself enjoyed my own dose of nostalgia going to a concert by the Presidents of the United States of America.  They were super popular in elementary schoo with hits like Lump, Kitty, and Peaches, but took an 8 hear hiatus (sp?) at the height of their popularity.  They\'ve reunited and played at the Rose Festival for the glorious price of FIVE DOLLARS.  I had a great time, and so did several non-blogging peers.  They played all their hits, some new ones (yay!) and even covered some Beatles, Sex Pistols, and Video Killed the Radio Star.  Rick White said it best.  I always loved PUSA, but I never dreamed I\'d finally see them my senior year of High School.  Somehow I think it will be a memory he sure as hell won\'t forget.  Love, Rob  '),(10097,1488330,'08,June,2004','2004-06-08 00:00:00','   I too fondly remember a time when I could fit both legs into a single pantleg of my size 30 waist JNCOs in the 4th grade.  I also remember the furious discussions with my father after covering them in grass stains playing football at recess.  It is an era I will forever remember with a twinge of longing in my heart.  I believe the same will be true once I exit high school.  In the elementary years all my friends and I could think about was reaching middle school, and now I wish I had savored those moments.  Understanding this time conundrum, I refuse to make the same mistakes twice.  Not even in the most boring, drawn out day of high school will I ever say â€œMan, I wish I was in college,â€ because I know Iâ€™ll blink my eyes and be a forty year old man sitting in a recliner in my underwear with a beer on the table and a hand in my drawls as an homage to Al Bundy.  For a seventeen-year-old that is a terrifying thought and one Iâ€™m not ready to accept, but that day is point B, and fortunately Iâ€™m still at point A.  But unlike my 11-year-old self, this time Iâ€™m going to enjoy the journey.  Today I still wear size 30 waist pants, only they fit a little differently.   -Ryan  PS â€“ Pat and Ryan â€“ stunning resemblance, but I think itâ€™s Alf, (Alien Life Form)  '),(10098,1488330,'07,June,2004','2004-06-07 00:00:00','   Pat I have to thank you for that shout out.  But i do have to condone you because Alf is spelled with fa \"f\" and not a \"ph\", if youre going to post about Alf get it right. I have had some friends in the past month bring up this fact that i do resemble Alf the tv icon.  i think its great and will make sure that this never dies.  Another thing i want to add is i loved your little peace on JNCO.  i was one of the unfortunate kids and never did have a pair because my parents thought they were hideous, but i envied everyone that had a pair.  --Alf  '),(10099,1488330,'07,June,2004','2004-06-07 00:00:00','   First of all, praise to all of you who went to see harry potter, my respect for you is boundless.  It\'s interesting to hear the feedback from the movies, so no one else was disappointed that it didn\'t show Gryffindor winning the Quidditch cup....I was im not going to lie about it.  Yeah Yeah, the movie couldn\'t be that long....but why not?  Anyone who goes to see the movie most likely has some interest in the book so they will not mind staying in the theatre an extra half hour to see some vital parts of the book. I am not saying I didn\'t enjoy the movie-it was quality, I am just saying I stayed up till 3:30 to see a movie and it didn\'t even have quite possibly one of the most important parts of the book in it.  So ya, enough about that. Now that thats over with, I would like to give a shout out to Ryan Vasbinder for his stunningly real portrayal of Alph, the TV icon, in the late 80\'s and early 90\'s.  '),(10100,1488330,'07,June,2004','2004-06-07 00:00:00','   What\'s every person here thinking about Clark\'s party? It\'d be cool, and I\'ll certainly enjoy sending her off. I liked having her as a teacher. I\'ll miss VA more though (if you\'re reading Mrs. VA, someone\'s paying me to say this). It\'s been great having her two years straight and seeing my writing grow, as well as having her watch the progression and help me through it. I know that there\'s some of you who\'ve had her for three years and are ready for a new teacher, and I see your point. I find it interesting the difference between last year and this year. It\'s a difference which struck me when I was looking for stuff to write my PDP\'s on. I\'m glad how she conducted the class this year; it was very worth while.      Poor Alex is right here next to me. It\'s Clark\'s 2nd period, and she still can\'t access either one. I\'m guessing that all of you are very ready to just leave. The will to work is leaving, and I really just don\'t want to do to much. I\'m ready to hop that plane to Germany, head off to my summer concerts, not as ready to do side work (but will somehow), and just slide into next year. It won\'t be so easy, especially all this college stuff, but I\'m ready to start anew with senior year. I really just want to throw away all the college stuff that isn\'t Washington; I\'m not really interested in leaving the state, and I\'m even less interested now. It\'s not smart to trash the stuff yet, but sometimes I don\'t want to be smart.      Anywho, I\'m going to check scholarships or the US blog. Till next time - Adios and God bless.  Tristan  '),(10101,1488330,'07,June,2004','2004-06-07 00:00:00','   For someone who has never read Harry Potter, never seen Harry Potter, not because I don\'t like them, I\'ve just never been interested, the latest movie was very entertaining.  I\'ve finally picked up the sorcerer\'s stone, which I think is the first book(someone tell me) that has been sitting on my shelf for an eternity and decided to read it so I think that\'s telling about the movie.  I enjoyed it, definately a visual feast and fun.  Even not knowing the background the story was simple enough to understand.  Is that kid who had the blonde hair and was a little biatch in the first 2 movies?  '),(10102,1488330,'04,June,2004','2004-06-04 00:00:00','   I have also seen the new Harry Potter.  Although some people are disappointed in the movie, I thought it was actually good.  There is no way they would be capable of putting the entire book into a movie.  However, if I hadn\'t read the books, some parts would have been a little bit confusing.  The only disappointment I had with the movie was that they never really explained the map in full detail.  If you are they type of person who only goes to see the movies, or if you just haven\'t been caught up in the HP frenzy, then I highly suggest you read the books. The movies are good but they would never be able to give the books the justice they deserve. The first book takes a while to get into, but it is definitly worth the energy. Rachel Penny  '),(10103,1488330,'04,June,2004','2004-06-04 00:00:00','   I spent the entire weekend basically watching Godfather, and thus coming to a conclusion that I am in love with Al Pacino and that therefore I shall go to a cloister unless I meet a prince on a white horse who will be an identical twin of Al, but preferably a couple of years younger (about 40...). Al Pacino is the greatest actor in the history of man and is a true inspiration. After watching the first two movies I was going to resign as the KGB princess sure that my destiny is to become a mafia godmother, but with all my regret I have to admit that the third movie sort of disillusioned me with the tragic life of mafiosi families, so thus I decided to stay in KGB for a while and then maybe go to a monastery once I am too old to search for my prince Al Pacino\'s YoungerTwin:( Anyway I encourage all of you who have not seen the movie to see it, believe me that you shall not regret the nine hours of your life spent to see the greatest classic of all time.  I saw HP3 this weekend also and I have to admit that I was very impressed with the new director and that I actually really liked the movie unlike that first two:) I agree with Rob that the movie\'s main improvement was that it was fast, entertaining and much more like the book. I couldn\'t believe that it was over so fast though, I want more now:) I think we should open a club in our school, called something like \"HP Nerds, Unite!\":)  '),(10104,1488330,'04,June,2004','2004-06-04 00:00:00','   Bethany, your parody brought me back to the days when I used to live near Seattle with your description of Puget Sound.  After reading your parody, I was reminded of a part in Huckleberry Finn, I don\'t know if you remember but Mrs. V-A called it the \"mini-vacation\" or something like that.  It was the scene in Huckleberry Finn where Twain is describing the sun rising over the river.  Your clear and thorough description of the events taking place and the surroundings is what brought me to recollect this scene in Huckleberry Finn.  Also, I applaud you on writing a parody-that can be a lot more challenging than cranking out a compare/comtrast paper.  So, in essence, as a Compton native would say \"mad props to you, it was off the hizzle fo shizzle\"  '),(10105,1488330,'03,June,2004','2004-06-03 00:00:00','   Hey, great action on the blog and I\'m sorry I don\'t have time to post on that right now, but it\'s 330 in the morning.  I\'m up because I saw the greatest wizard/witch-based movie of all timeâ€” Harry Potter and The Prisoner of Azkaban.  It was my favorite book, but I loved one and two as well and I didn\'t think the movies had entirely done them justice.  The new one is just that; new.  The new director, Alfonso Cuaron (something like that, he directed Y tu Mama Tambien) did an absolutely fabulous job making some changes that made the first movies lose a little magic (pun intended) compared to the books.  For instance, robes are worn only once that I can recall, being replaced by jeans and sweaters, making the characters more comparable to the muggle adolescents in the audience.  Also, Draco Malfoy, though still hateable, is now a coward in the face of confrontation.  He takes on the role of a character like Wylie Coyote in the fact that he had malicious intentions, yet always ended with a frying pan bent around his face.  Lastly, Pat said that it didn\'t always follow close to the book.  Don\'t get it twistedâ€” none of the story was altered, however, in the interest of time, much quidditch is subtracted in place of plot advancement, which is entirely ok, because the foreshadowing is MUCH better orchestrated than in previous movies, making it much MORE like the book in that sense.  If you can\'t tell, I LOVED IT.  It was funny, fast, and satisfyingly sentimental.  I\'ll see it again w/ anyone that wants to go.  A word to the wise, I might be a little grumpy with the mere 3 hours of sleep I plan on getting, so if you must talk to me, PLEASE let the topic be Harry Potter.  Kahli Veys was there too, so if she\'s at school, she too can fill you in on the glory that is HP.  May all your sleeping patterns be much more consistant than mine,  Love Rob  '),(10106,1488330,'03,June,2004','2004-06-03 00:00:00','   Umm ya, its about 3:06 in the morning and i just got home from seeing harry potter.  I am seriously tired, but I thought it would be important to blog to share the improvement of the third movie in comparison to the previous two.  Go see it, its better.  It doesn\'t follow the book as strict as some would like, but it makes for a quality flick.  '),(10107,1488330,'03,June,2004','2004-06-03 00:00:00','   Hideeho Bloggeroonies  I same to be borrowing a page from Ned Flanders on that one. What\'s going one you probably all asleep people? I know that lots o\' people are off at that Harry Potter opening, my little sister included.   Bethany, I\'d like to read that parody gig sometime, it\'d be cool to see. It can take time to write a nice parody, or a nice thing in general. I\'m still trying to finish up my Stalin/Spongebob parody (it\'s almost done, and thank God I finally have an idea for the ending!). It\'s fun to write these things, and we all have our different styles, and practice makes perfect, so don\'t worry. Also, true dat about leaning on God. Coming to him makes us realize how stupid materialism, selfishness, etc. can really be, if we take time to be sincere and thoughtful with him.   Speaking of said parody, it\'s at the center of what I want to do for my senior project. I\'d like to direct and produce it next year, try and make it part of a student-written winter one-acts show that we\'ve had in the wings throughout this school year. I know Megan might be interested in writing a piece. My thesis\'ll center around different styles and methods of directing, but the project is what I\'m really interested in. VA really liked that I\'m going to do it, and she\'s been encouraging in telling me to do a project that is important to me. I\'ve worked almost all school year on my play (named Stalinpants = Stalin + Spongebob Squarepants), so it\'ll be a great way to tie that in to something like senior project. Hope some of you guys might be interesting in trying out next year. We\'d love to have you in any of the shows, and Stalinpants specifically has a large cast, so you\'d have a nice chance of getting a part.  Anywho, I find it interesting that we\'ve got senior boards next year. I can totally understand though why some juniors last year were quick to decorate themselves and their cars in Senior \'04 stuff. It\'ll be a fun year (pre-nostalgic sigh). Anywho, I\'m going get going, and maybe go to bed sometime soon; most likely. Till next time. Adios and God bless.  Tristan the Yeehawman.  '),(10108,1488330,'03,June,2004','2004-06-03 00:00:00','   \"Anthem\" Ayn Rand  \"I am. I think. I will.  My hands.... My spirit.... My sky.... My forest... This earth of mine.  It is my eyes which see, and the sight of my eyes grants beauty to the earth. It is my ears which hear, and the hearing of my ears gives its song to the world. It is my mind which thinks, and the judgment of my mind is the only searchlight that can find the truth. It is my will which chooses, and the choice of my will is the only edict I must respect.  Whatever road I take, the guiding star is whithing me, the guiding star and the loadstone which point the way. They point in but one direction. They point to me.  I know not if this earth on which I stand is the core of the universe or if it is but  a speck of dust lost in eternity. I know not and I care not. For I know what happiness is possible to me on earth. And my happiness needs no nigher aim to vindicate it. My happiness is not the means to any end. It is the end. It is its on goal. It is its own purpose.  And now I see the face of god, and I raise this god over the earth, this god whom men have sought since men came into being, this god who will grant them joy and peacs and pride.  This god, this one word: \"I\"  '),(10109,1488330,'02,June,2004','2004-06-02 00:00:00','   (Ego translates as \"I\" from Greek) I have never asked myself that question before, the answer always seemed to be: duh... evil! Of course until... I read the Fountainhead, which I finally finished today and have successfully proclaimed to be my Bible:) Although it is as radical in its ideas as any philosophical book could possibly get, the philosophy is integrated in a great, crazy story of an architect: the greatest egotist alive, a man of highest level of integrity. And here is what he says at his trial before the collectivist/altruist world accusing him of being an egotist:     \"The first right on earth is the right of the ego. Man\'s first duty is to himself. His moral law is never to place his prime goal within the persons of others. His moral obligation is to do what he wishes.     \"From the beginning of history ,the two antagonists have stood face to face: the creator and the second-hander. The creator lives for his work. He needs no other men. His primary goal is within himself. The parasite lives second-hand. He needs others. Others become his prime motive.      \"The basic need of the creator is independence. The reasoning mind cannot work under any form of compulsion. It demands total independence in function and in motive. To a creator, all relations with men are secondary. Second-hander declares that man exists in order to serve others. He preaches altruism. Altruism is the doctrine which demands that man live for others and place others above self. No man can live for another.      \"Men have been taught dependence as a virtue. The man who attempts to live for others is a dependent. He is a parasite in motive and makes parasites of those he serves. The relationship produces nothing but mutual corruption.     \"Men have been taught that the highest virtue is not to achieve, but to give. Men have been taught that it is virtue to agree with others. But the creator is the man who disagrees. Men have been taught that it is a virtue to swim with the current. But the creator is the man who goes against the current. Men have been taught that it is a virtue to stand together. But the creator is the man who stands alone. Men have been taught that the ego is the synonym of evil, and selflessness the ideal of virtue. But the creator is the egotists in the absolute sense, and the selfless man is the one who does not think, feel, judge, or act. These are functions of the self.      \"The egotist in the absolute sense is not the man who sacrifices others. He is the man who stands above the need of using others in any manner. He does not function through them. He is not concerned with them in any primary matter. Not in his aim, not in his motive, not in his thinking, not in his desires, not in the source of his energy. He does not exist for any other man - and he asks no other man to exist for him. This is the only form of brotherhood and mutual respect possible between men.      \"You can fake virtue for an audience. You can\'t fake it in you own eyes. Your ego is the strictest judge. It\'s easier to donate a few thousand to charity and think oneself noble than to base self-respect on personal standards of personal achievement.\"     \"People are not really struggling even for material wealth, but for the second-hander\'s delusion - prestige. They don\'t say \"This is what I wanted because I wanted it, not because it made my neighbors gape at me.\" The only cardinal evil on earth is that of placing your prime concern within other men. If one doesn\'t respect oneself one can have neither love nor respect for others.      \"The only thing that matters, my goal, my reward, my beginning, my end is the work itself. My work done my way. A private, personal, selfish, egotistical motivation.\"     \"I never think of myself in relation to anyone. I just refuse to measure myself as part of anything.\"     \"I came here to say that I do not recognize anyone\'s right to one minute of my life. Nor to any part of my energy. Nor to any achievement of mine. I am a man who does not exist for others. The integrity of a man\'s creative work is of greater importance than any charitable endeavor. I recognize no obligations toward men except one: to respect their freedom and to take no part in a slave society.\"    \"Is sacrifice a virtue? Can a man sacrifice his integrity? His honor? His freedom? His ideal? His convictions? The honesty of his feelings? The independence of his thought? But these are a man\'s supreme possessions. Anything he gives up for them is not a sacrifice but an easy bargain. Self-sacrificed? But it is precisely the self that cannot and must not be sacrificed.\"  '),(10110,1488330,'02,June,2004','2004-06-02 00:00:00','   Hey guys!  So, I know that I\'m a little behind the times with the blog, but I just wanted go back to Pat\'s thing on the essays and sharing them with each other.  I think it\'s a really good idea, especially for people in the other classes to share essays about the same books.    Anyways...Pat, I thought that your essay was really good.  It reminded me of how materialistic America has become, and we all get swept up in it.  As much as you think you don\'t, I\'m sure that all of you have.  I mean, we all go to Skyview, and none of us seem to be doing that poorly.  I mean, my parents both of decent jobs, we live in a pretty nice house, and we have enough food and clothes and things.  I just throw myself a pity part sometimes, when I go over to a friend\'s house who lives in Ashley Heights or has a closet full of nice clothes, or wears a different outfit every day...it makes it really hard to be content with what you have and to not be dissatisfied.  But (I\'m gonna be religious here), as long as you lean on God, you should never be dissatisfied with what you have.  And that\'s something that I struggle with, but that your essay reminded me of.  So, thanks Pat.  I wrote a parody, and it\'s not too bad.  It follows these few pages in \"Their Eyes Were Watching God,\" and it\'s a really amazing part of the book that I wanted to try to recreate.  It was hard, so I\'m not sure how good my parody is.  Read it, see what you think of it!  But I\'ve gotta go.  All of the Junior Project stuff is do tomorrow...I should start.  Bye!  Bethany          Mary looked up at her with hungry eyes, only wanting to listen to the unfolding story of her Momâ€™s life. Sarah looked around as her eyes slowly took in the summer surroundings.Â  Bees swimming drunkenly from flower to flower, the sun slowly dipping into the Puget Sound, and the gentle crash of waves on the shore.Â  She smiled and gazed down at her patient daughter.Â  Maryâ€™s unanticipated interest in Sarahâ€™s teenage years inspired her to tell her story.      She thought of where to begin, and her eyes glanced heavenwards.Â  The afternoon sunlight danced through the branches of the weeping willow tree and made playful shadows oh Maryâ€™s excited face.  Sarah sighed with content, and decided to begin at the time when she first imagined that maybe she wasnâ€™t a little girl anymore- that maybe she was growing into something greater.Â  She decided to tell Mary about the time she swam in the ocean at midnight.      It was a warm, summer evening on Whidbey Island.Â  Sarah had spent the evening roasting sâ€™mores over a beach fire and singing songs with her family.Â  Every evening of that weeklong vacation, she had been on the beach.Â  She spent every moment walking along the shore, but at night liked to listen to the waves crash on the rocks, slowly churning beach glass and barnacles into sand.  The ocean called to her, like a mystery that needed to be unearthed.Â  Not in the way that the ocean calls to a sailor, but she was mesmerized by the intense drifting current and the white foam that disappeared after only moments on the rocky shore.Â  It inspired her.Â  How?Â  Why?Â  It was like a beat lost long ago, in another time, drifting across the ocean to meet her.Â  What?  How?Â  Why?Â  The song of the ocean was not a real song, yet the salt sea air made her drunk to the world.Â  It followed her through her every move, and it pleasantly haunted her dreams and thoughts.Â  The ocean sang to her, the waves the symphony and the air its conductor.Â  It seemed to connect her to a place or time almost unreal; her observations became acute and alive.  Things that she had not noticed suddenly became vivid and momentous.Â  They were unearthed and filled her mind and conscience.      She took a few steps towards the water, admiring the sparkling rocks and glistening water.  Although her fatherâ€™s guitar was now safely inside the cabin, the song still swayed in her mind, a friend in the warm night.Â  Suddenly the guitar, waves, and air were all playing together in harmony; a melody so strong and inspiring yet so soft and beckoning that she could hardly contain herself.Â  She fell against the soft sand and stared at the stars, twinkling happily in the dark sky.Â  Why had she never noticed them before?Â  The song of the ocean controlled her; she closed her eyes and seemed to float closer and closer towards heaven and the stars.Â  Did the waves always make music?Â  Did the moon always make the water sparkle like diamonds?Â  Yes, but not for her.Â  Tonight was her night.Â  She felt limp and drunk with flying towards heaven and listening to earthâ€™s song. 	    After awhile she stood up and looked across the Sound.Â  She was seeking some sort of confirmation of her vision, her dream, her revelation.Â  She received her answer.Â  The water nodded its agreement, the waves roared its truth, and the stars shone in affirmation. The waves rushed over her bare feet, a usually chilling sensation that now left her smiling with indescribable warmth.Â  She waded into the water.Â  About to dive in, she glanced back at the cabin where here parents and siblings were sleeping.Â  She turned around and dove into the cold water.Â  She was seventeen!Â  She was a dancing star and a singing wave, turning over and over in the ocean.Â  Was the sky so joyful for her?Â  Nothing answered.Â  She searched the heavens, floating across the waves in the water.Â  Looking, waiting, breathing short with impatience.Â  She stood and looked around her; the world was made in a new light.  '),(10111,1488330,'02,June,2004','2004-06-02 00:00:00','   The school food actually sounds like a good idea. Or how about how it\'s not healthy to have to eat within 30 minutes? If you go out to lunch by the time you get there you only have 10 minutes to eat before you have to leave.  It took me forever to decide what topic I wanted to do. However, once I picked it seemed like good topics came up everywhere, it\'s crazy how that always seems to happen.  I finally chose to do the benefits of extracurricular activities. It\'s actually interesting how being involved in an after-school club or sport is related to higher test scores and better attendence. It\'d be interesting to see what everyone else came up with for their junior project. Have a fun time with all the nice weather while it lasts!   - Rachel Penny  '),(10112,1488330,'02,June,2004','2004-06-02 00:00:00','   So, this is my first time blogging for English class, whoohoo!  This is very exciting.  I don\'t quite know what to say, there really isn\'t a whole lot of controversial stuff to talk about here like there is in Clark\'s class.  What is everyone doing their Junior Project on?  I\'m doing how disgusting and unhealthy school food is, because honestly...yuck!  The deep fried fries and extremely greasy pizza everyday, that needs to stop eventually.  Paul, I totally agree with you about how it\'s hard to keep your mind engaged on school when you do nothing in other classes...ah!  I don\'t know what to do with myself, I have absolutely no motivation to do anything.  What a great attitude, right? Haha.  Anyway, I\'ll see all you kids at school tomorrow!  '),(10113,1488330,'01,June,2004','2004-06-01 00:00:00','   Rob I commend you on your idea.  Most students won\'t be able to vote on their own, I for one(missed the cut off by 2 weeks).  Paul, your recomendation on Ma$e\'s new song is superb, I heard it last Saturday and recognized the unique lyrics and smooth, calm vocals, so I had a feeling it was Ma$e.  Do you know if that was a single, or if it is part of a CD?  '),(10114,1488330,'01,June,2004','2004-06-01 00:00:00','   Gosh, I remember JNCO\'s.  4th grade, had to get me a pair, couldn\'t afford it, I was a very sad little boy.  I remember them having innovative tee-shirts too.  Amazing how everything changes, since I don\'t think any of us would go near JNCO\'s now.  Are they even still around?  I think your skyview votes for the vote idea is very innovative, it\'s the kind of idealistic idea I\'d expect to read about in Newsweek or something...  Oh the possibilities.  Is anybody else having trouble focusing their energy.  We have work in this class, clarks(somewhat), and my math class.  But everything else, especially my 3rd and 4th periods, are nothing.  It\'s really hard to keep a work minded mind not having to do any work.  I love summer, but I hate the mindless feeling I get from it.  I always like to have my energy focused on something.  I think I\'m going to end up being one of those guys who can\'t bring themselves to retire, and when they have to they hate it and go back to work after 8 months.  Closing note, Ma$e- bling bling- is back.  Everyone needs to go listen to \'Welcome Back\', who else can you see sampling the welcome back kotter song?  It\'s actually very good, and he hasn\'t given up on his preaching, which is cool.  But it\'s definately a break from the all the mindless club music I hear all the time now.  '),(10115,1488330,'01,June,2004','2004-06-01 00:00:00','   i think that would be a good idea.  there is no way to have a student government that completely represents the school, but your idea is definitely an improvement.  seeing as this is my first year at skyview, sometimes it seems like asb doesn\'t really do anything aside from blatantly violating the dress code and getting away with it.  perhaps that\'s just me being cynical about the way things work.  haha.  go for it buddy, anything is worth a try.  '),(10116,1488330,'01,June,2004','2004-06-01 00:00:00','   Rob, what you will be doing is unprecedented, unique, and first of its kind.(For that matter you can patent this ingenius idea.......JK.....or at the least claiming your originality of that idea as your intellectual property).   But seriously, if I interpreted your message correctly, I think core concept is to get more young people our age involved in politics, the political process, and especially how the political process works in a democratic society - with the funamentals manifested in the act of voting. We all realize the age group of 18-21 represent the lowest percentage of elligible voters that actually vote in an election. For some odd reasons young people our age just does not care about policies of our nation(Heh when there is a draft I bet a lot more people will vote then, but then again it\'s too late by then).   Youngesters, take your legitmiate part of control over the government!!!!!!!! We do not want the same old/senile/retired generals and shrewd politicians dictate our lives and draft us to be killed in some unknown regions of the world without our consent!!!!!!!!!!  I support the statement that those who do not vote does not deserve complaints.   THUS............GO VOTE, AT ANY OPPORUNITY YOU CAN.......AND THIS IS A GREAT OPPORTUNITY FOR ALL OF US WHO ARE NOT OLD ENOUGH TO VOTE IN THE NOVEMBER ELECTION A TASTE OF THE DEMOCRATIC PROCESS!!!!!!!  TAKE CONTROL OF YOUR YOUNG LIVES AND PROSPEROUS FUTURE THROUGH THE ACT OF VOTING!  And finally, people on the AP Fest party committee, please blog in and speakout about your plans and what activities are we going to have.......or is it a secret surprise party?  '),(10117,1488330,'01,June,2004','2004-06-01 00:00:00','   Maybe it\'s because it\'s one in the morning, but I had a thought that intrigued me a lot.  I was thinking a little bit about how representative the ASB is of the studentsâ€” and it\'s basically trust on behalf of the student body that gives the governing body \"artistic license\" to run the school as it sees fit.  There are no referendums, and although meetings are open to the entire student body to sit in on, nobody ever does, or probably even knows it\'s an option.  In an effort to make my presidency more directly representative as well as continue my theme of the importance of democratic participation, I would like to have the school vote on how I should vote in the elections held this November.  It would be one vote representing the entire school, and it would sacrifice my own voice, but I think mock campaigning in school for candidates at all government levels would expand the knowledge and interest in politcs of the school: a heritage that would be far more valuable than the vote I would cast on my own.  With a fervor that hinges on administrative support, I would like to pursue this concept.  What does everyone think?  I want comments from both students and their parents on this one.  But not Mike Rush, he is an ignorant Pig-Dog who deserves no respect from anyone.  (Mike demanded I begin trash talking on the blog).  Secondly, I want to organize a large-scale trivial pursuit contest- who would be down to participate?  Trivial Pursuit will be the next big thing in this school and nobody wants to miss out on this.  May democracy live long and prosper.  Love Rob  P.S.  Create Blogger profiles, they are kinda fun to do and give ME something to look at when I\'m insomniacing  '),(10118,1488330,'01,June,2004','2004-06-01 00:00:00','   Whatever happened to the greatest corduroy company of all time?  JNCO.  The name radiates a sense of \"gangster.\"  In the grunge era of the mid 1990\'s, bands such as the Smashing Pumpkins and Nirvana influenced me, as a elementary schooler, to pursue such the novel brand of JNCO.  Now if you are pronouncing it JUH-N-COH, start pronouncing it JEEN-COH.  During this time of my life, others strayed to other brands such as Lee Pipes-a jean that went straight down and was usually ridiculously baggy.  Another blogger, Rob Muilenburg recalls JNCO\'s by stating \"they changed my elementary life in a time when I hated Phil Sobolowski\"  Yes, Rob, I agree with you, they did also change my life.  The JNCO scriblle will always live on as long as there is the memory.  '),(10119,1488330,'01,July,2004','2004-07-01 00:00:00','   Hello!  I arrived at the University of Massachusetts Dartmouth on Sunday night and am enjoying a new climate, new colleagues and new sights.  Since Thomas asked about summer reading, I thought I would answer.  Mr. Wargo and Mrs. Grey-Hatton will be teaching the senior AP classes next year.  They have not assigned you any summer reading.  However, take advantage of your freedom and free time this summer and read on your own.  Your primary focus next year will be British literature and poetry.  Perhaps this summer you can indulge yourself in more modern fiction.  You might want to focus on prize-winning book lists like Pulitzer Prize winning novels or Booker Prize winners.  Or focus on modern writers from around the world.  Another option to consider is set yourself up with a book list of classics, because I guarantee you that you\'ll be in lit classes in college and wish you\'d taken better advantage of your free time as people allude to books and authors you\'ve heard of (maybe) but don\'t really know.  A few books I\'d recommend?  Reading Lolita in Tehran The God of Small Things Persopolis Love in the Time of Cholera Anything by Dostoevsky, Chekhov, Ibsen, Woolf, Hardy, Neruda Red Azalea Germinal  Have fun reading, be safe, and have a happy summer!  Mrs. V-A  '),(10120,1488330,'06,August,2004','2004-08-06 00:00:00','   I just finished my six weeks of intense math camp.........And a short two and half day ASB Leadership camp!!!!!!!!!! Wow...........Way too much to say............... The math camp has elevated my life to a new dimension, with numerous epiphanies occuring and many new friends and bonds made. And many valuable lessons I\'ve learned from math camp, including the fact that I felt incredibly stupid when I was immersed in a sea of of 23 geniuses. And how important friendship and \"constructive\" competition is. And that\'s besides all the six week full time math we learned.   I miss the university setting soooooooooooooooooo incredibly much!!!!!!!!!!! I wish I could just be a college student and live in the McMahon dorm from now on for the next few years! Home only brings me the memories of stress, depression, loneliness.............and what not. While the math camp at the university was completely motivational, stress-free, fun......Fun together with learning/bonding experience.   Otherwise, the ASB retreat was mostly a taste of what we have to do next year, the responsibilities we have, the goals we have, and the people we are going to working with. It was quite interesting.........We\'ll see next year............Although I seriously wish Ms.Clark was still here!      Thomas Chen  '),(10121,3601175,'12,June,2004','2004-06-12 00:00:00','   Ok if anyone knows a good hit man hook me up!. My parents have started WW3 and i get pulled to either side! yay sucky life! Anyway lets get down to the real reason i wanted to post, ok on gaia i got this pm it was did anyone ever tell you.. So i opened it and the kid said your hot as hell so in my mind i was thinking \"who the fuck thinks pixels are hot?\". Well I\'ll tell you who, 12 year old boys that\'s who. Seriously why the hell do i only get hit on by twelve year olds? It\'s not fair I\'m to old to be hit on by 12 year old that\'s a 3 year age difference!! No one say anything about my crushes i know it\'s not any different but seriously.... Anyway i cantered presto today and i made up with Ana so were cool again. Elf boy hasn\'t been on and i haven\'t seen Ryu for years it seems i guess my RPG peaked and now is starting it\'s steady and eventual flaming death, oh well it was fun while it lasted. i guess I\'ll just stick to other peoples from now on.. It;\'s sad to i thought that one was going to go someplace good and maybe like work and have a really story with a goal... It backfired and now I\'m back to square one. Well my new underage stalker is bugging me and he can\'t see this (thank you parental controls!) so i got to go! Shimby  '),(10122,3601175,'11,June,2004','2004-06-11 00:00:00','   urrgha I\'d like to say one thing, who ever upstairs decided to make most boys annoying jackasses and all the good ones gay needs to be fired it was a cruel joke and there needs to be pay back. Take my current head over heels crush, Karl OK first off i have issues talking to guys i always have and it\'s not something that\'s gone change over night. So Karl well he probably thinks I\'m a bitch who thinks she\'s better then him because i never say hi and i avoid eye contact and everything i know i shouldn\'t but i can\'t help it my mind completely shuts down around him. i can\'t even talk to him about anything with out stuttering or dropping something or making a fool of myself. i mean come on how hard is it to talk to someone one who\'s nice and rides horses and is not a flaming idiot or as far as i can tell gay? In my case it\'s harder then doing a bi cranial bi pass while riding Superman ride of steel. I wish i could just give myself a wake up call and get with the program and stop being a stupid wimp but it\'s not that easy. *sigh* this is Turing into more of a rant then a journal now isn\'t it? I\'m sorry guys this is another one of those depressing blogs were some teenager who thinks she\'s the worst off in the world moans and complains... I\'ll stop now with his final message, save the whales and the other species of the environment, don\'t were fur and make sure you thank your pets for being such loving companions who don\'t mind if you cry into there shoulder. The same goes for friends and don\'t ever do something stupid like lose one over a guy because you never know when you\'ll need them or when you\'ll lose them forever to something bigger then us all. Hold onto your childhood don\'t let it slip away keep it as long as you can and don\'t let anyone tell you to grow up. Keep  your head up when you feel like crawling into a hole and staying there for forever and never be afraid to speak your mind. Freedom is only as free as you chose to make it and if you stop watching even for a second someone can take it away from you, the cost of freedom is constant vigulance thats what my history teacher taught me and I\'ve made it my motto for the rest of my life. never think that one person can\'t change the world because you know what? they can and it takes that one person to change the way the rest of the world thinks. remember what your forefathers died for protect it well -Shimby  '),(10123,3601175,'10,June,2004','2004-06-10 00:00:00','   yo just me and here we go. Ana\'s confusing me and every time i try to help it blows up in my face. kana\'s in the hospital I\'m not sure what happened but we think she hurt herself over a grade and some other problems. I\'m not her friend really but i am still worried about her and i wanna go make sure shes OK but i don\'t know what hospital shes in. it\'s the end of the school year only a week of finals to go and then I\'m a sophmore... man it\'s scary i can\'t believe it i had a hard enough time maturing to freshman status what am i gonna do as a sophomore? blow it entirely? make a fool of myself and have to move to another country? i mean really it\'s not like i have anyone who really knows who i am, what makes me tick. sure my friends think they know me but how much of what they know is just part of this elaborate facade i put up around myself day in and day out? Is what they think i am what i really am? i can\'t even answer those questions for myself i don\'t know why I\'m trusting complete strangers to answer them for me. i just wish like every little girl forced to grow up that prince charming will sweep down off his white horse and carry me away into the sunset were we shall live in happily ever after forever and always. but i think someone told my prince that my hairs always frizzy i can\'t keep my temper and my glasses are always crooked. someone let it slip that i try to pretend i don\'t give a flying fuck about what people think when it dictates everything i do. every day i try my hardest and then when my grades come they just show my lack of focus on homework and such and i know i can do better and it just kills me. i think I\'m going to really make these exams count study like hell and do my best and then next year, no more fooling around I\'m going to get honer roil every quarter and I\'m not going to worry about the swirling vortex of evil thats edging ever closer. I\'m going to reach that far off goal of complete independence and forgetting conformity. but i won\' fall into drugs and sex to pull me through no way I\'m going to be different I\'m going to be a good girl who\'s not afraid to punch somebody who\'s messing with her and I\'m going to love it. oh well i\'m off to the barn peace out peoples  -Shimby  '),(10124,3601175,'10,June,2004','2004-06-10 00:00:00','   Damn it this is what happened when I get bored, I make random sites for people to see! Don\'t ask but anyway I guess so you have a fucking clue what I\'m talking about I\'m a freshman at a prison I mean high school in CT my best friends are Slappy, Sana, Kitsune,Elpi/Ana,Jake,Chelz R., Mandy, Megan M. and a few others but there really not all that important and if you haven\'t guessed most of them are names that we call ourselves because that\'s what we go by on our RPG\'s and stuff. My names shimby if your wondering and I\'m 15 with red blond hair one blue eye and one green eye. I ride horses and hate all other sports and my only other hobby is boys and being a pervert.... I also can\'t spell to save my life so spell check is my best friend ever. I live for my friends and my enemies better watch there backs *cough coughpaulcough cough*. I\'m addicted to Gaia and all rpg\'s in general especially my own htpp://endingpain.proboards29.com because it fucking rocks.. Yes I have a swearing problem no I don\'t want help with it. Anyway what happened today oh i know! i cantered my horse presto,which my father dose not know i have BTW for the first time off the lounge line and it was fucking epic! Oh yeah epic is my word i use it whenever something is wicked cool and needs a whole new word to describe it, that\'s were epic comes in to play ^_^ ok other then that nothing no Karl at the barn today no gym to see Napoleon (also known as Tim) in and nothing new to report with Jake... Well peace all I\'m out Shimby  '),(10125,3601175,'19,June,2004','2004-06-19 00:00:00','   Hey guys OK i need to put this somewhere so everyone will see the truth about my father.  Heâ€™s a jackass, he yells about everything and i hate being around him.  I couldn\'t ever stand it to be with him 24/7 because he has no idea what a father should be.  A father should be caring, not just about what you do but your life and who your friends are and what your friends are like.  A father should love you and respect you at least to the point that he doesnâ€™t treat you like a second class citizen who has no rights and needs to do everything you tell them the minuet you tell them or they get thrown out of the house.  A father needs to be someone you feel comfortable around, not someone you have to walk around on eggshells because your afraid he\'s gonna blow.  You see i have a temper problem and i really wonder were i get it from.  Hmm maybe my dad?  Heâ€™s got himself a real anger management problem that he dose, and i can\'t stand it so as i think about Fathers day tomorrow i realized this, i can\'t give my dad a fathers day card because he\'s not like most dads, he\'s a tyrant who by some cruel twist of fate became my dad.  And i don\'t think i could reward him with that.... You know you love me,  Shimby  '),(10126,3601175,'17,June,2004','2004-06-17 00:00:00','   Hey dudes Surfs up!  Bikinis out!  Hot guys in swim trunks!  I loooove summer oh god do i ever.  Thatâ€™s why i havenâ€™t been able to post lately Iâ€™ve been to busy havin fun!  Awe man i love just laying in my yard caching the sun and turning bronze.  Man i love it, and then the riding, thereâ€™s nothing like riding bareback over the meadows while crazy dan yells at you to get away from his property, and then obliging him by takin off at full gallop to your favorite prince and his horse.  Man o man i love summer.  Will someone just let me stay this way forever?  Give me prince charming, a horse, and a few bottles of coconut tanning oil and Iâ€™ll be set.  But anyway here i am sitting at my computer as it rains; even rain is good in summer time!  I think Iâ€™ll go out and play b-ball in the rain or maybe Iâ€™ll just crank call a few PPL... i don\'t know i don\'t know they all sound like so much fun!  Hmm maybe Iâ€™ll just go curl up some was with the latest gossip girl book... i like that idea... hmm mm.....      Well whatever i do you know you\'ll hear about it sometime anyway let me tell you about the story of the little frog who uh well just listen... ribbit ribbit went the little green frog one day ribbit  ribbit  went the little green frog ribbit  ribbit  went the little green frog and his eyes went Ho ha Ho HONK HONK went the big red truck HONK HONK went the big red truck one day HONK HONK went the big red truck one day and the little green frog went SPLATT!  And then thereâ€™s little Ritchieâ€™s trip to the zoo Fuck fuck fuck a duck  Screw a kangaroo 69 a porcupine ORGY AT THE ZOO!  You Know you Love me, Shimby  '),(10127,3601175,'15,June,2004','2004-06-15 00:00:00','   Hey all  Ok I want a hand raise who ever made up finals would they please step out of the room and shot themselves? Please? Seriously I\'m on the verge of a mean breakdown here and it\'s all because of the excess stress caused by exams if it wasn\'t for them I\'d be perfectly fine.. cockbitting fucktards...(thank you RVB for that word...) Anyway I\'m at the god computer in the media center waiting for Slappy to finish her history.. Oh wait here she is looks like I gotta jet peeps! Time to make CAKE!!! It\'s of to her house I go and then to gieslers.. This outta be fun...  '),(10128,3601175,'14,June,2004','2004-06-14 00:00:00','   Hey guys! i slept over Ana\'s house last night it was killer really it was. We played FFX and talked online and \"studied\" for our finals today. I didn\'t get much sleep though so I\'m running on sugar and coffee right now so I\'m all jittery and perky ~^_^~ it\'s fun. I didn\'t get to see presto last night but Ana\'s dogs make up for it, there huge. Grizzly is a akita German Shepard mix (right? aww I\'m not sure) and Dakota is a German shepherd, there sooo cute! and there HUGE!!! i loved it. There was lots of new stuff i learned like FMND (you can guess what it means) and the fact that my RPG character is allowed to come back on the realm as a guardian Angel, a real kick ass guardian angel at that. Shimby BIA this is soooo gonna rock now all i gotta do is pick who the lucky person is to be my charge... ;) this should be killer fun, no this should be EPIC!!! Ok i need to go, sorry this is such a fuckin short entry but  the comas setting in, time for shimby to go nighty night i\'ll leave you with this 1 tequila 2 tequila 3 tequila FLOOR and   fuck fuck fuck a duck  screw a kangaroo  69 a porucupine  orgy at the zoo!!!      Shimby  '),(10129,3601175,'13,June,2004','2004-06-13 00:00:00','   Let me tell you a story. This story is about a little girl by the name of Shimby who went to the 4H fair saw the horses, now she is a full blown horse addict. It all started innocently enough, it was 4 years ago on a hot day in August, i was walking thorough the animal barns at the fair when i saw the horse tent. I smiled and said to my friend \"come on lets go see the pretty horses!\" and took off at a full run, little did i know how this event would change my life forever. The barn was deserted as it was after closing hours and it was only the 4 h\'ers were left. There was this one horse, a big chestnut with a blaze, his show name was Classic car, but they called him Collin. I can thank Collin for pulling me under the devastating spell of horse craziness, it was love at first sight. The girl who owned Collin saw that and she let me into the stall to pet him and give him treats, I\'ve been clinically obsessed ever since.      Now every obbsesion you just have to pass it on to someone, anyone you don\'t care who. i started with my best friend Chi,but to no avail, i stopped trying as i had no horses at my disposal to get her hooked. This year in August at, were else but the 4H fair i met my ridding instructor, Barb and started riding. The first horse i rode was lil Sonny, well that\'s not really true the first horse i ever rode was foxy and then copper at $H camp but back to Lil Sonny. He counterfeits which means he dose everything in his power to turn around and fuck with your head... Not very fun. But i was hooked on that little QTH buckskin so i couldn\'t wait to ride him in my next lesson, well i didn\'t Barb put me on Chaz a 16hh TB bay with a lighntnig bolt on his forehead, he\'s a hell of a lot different then little sonny and i fell of for the first time it was quite a learning experience I\'ll tell you that. So me and chaos worked pretty hard and i fell in love again. Now as the months went by it was finally time for my first show, i wanted to take Chaz naturally but barb put me back on lazy lil sonny. i won a few ribbons but it wasn\'t what i wanted. It would be another 4 months before i actually wanted to ride sonny instead of Chaz, all because of the INCIDENT. It was mid January and the snow was fallen off the roof of the indoor. Chaz being a high strung TB hated the snow and one day as we were trotting around the ring a huge piece of ice and snow fell from the roof and i got to do my jockey impression in the Wishing Rock derby. Chaz took off full speed and i screamed and hung on, finally being able to pull him down to a halt and jump off only to half faint. The next weekend i had a show on him and got 4th place and did damn good, but it ended my long time love affair with Chazy. A bunch more stuff happened but i don\'t fell like putting it down on paper, so flash forward to march 7th 2004 Heritage farm auction, we went up with out a trailer even the my mom had finally said she thought it was time for me to get a horse and barb agreed. We sat and watched the horses go by for auction when a  hh skinny as a twig chestnut Tb with  perfect diamond on his forehead two back white socks and a messed up eye and a cold (both of those we didn\'t find out until after) came out, the bidding started at 1000 dollars, no one bid the auctioneer lowered it until it reached 650 and my mom raised her hand, we didn\'t have a number. The jumped him and even in his weakened state he jumped it perfectly one other person bid bringing it to 675 my mom bid again 700 dollars, barb shook her head and went to go see him up close, thinking we were crazy as bloody hell, the auction closed and i had my first horse \"Presto\'s his name and what\'s yours little lady?\" the auctioneer asked me as Megan  M and P an Kelly hung on to me squealing happily \"Meaghan\" i said shocked and  shaking then i ran out to were barb was standing with my new horse and i hugged him around the neck and he whinnied as we slipped a old green halter over his nose and took off the western saddle and bridle. We borrowed bill strains trailer and brought him home to the quarantine barn that night. i was in love.     As Presto grew stronger and i started riding him it was clear that this ex grand Prix show jumper was not for a beginner rider, witch i was at the time, i had to learn fast, and that i did. After him taking off while i was riding him and me getting hurt pretty bad Barb started having doubts over my ability to control presto, she wanted me to sell him. I could never sell presto because the night of the auction i made him a promise, as cheesy as it sounds i told him he\'d never be sold again and that he\'d never have to go to another auction and that he;d be loved and cared for until the day he died nd you know what? I\'m a girl of my word i keep my damn promises and I\'m never breaking this one.     Flash foreward to yesterday, my little cousin Sammy had her first lesson and you know what? i finally found some one to pass the addiction along to, she\'s going to take lessons every Saturday on Ambie and she\'s gonna be the next little equestrian in the family. that\'s my story peace all Shimby  '),(10130,3601175,'21,June,2004','2004-06-21 00:00:00','   Hey all why o why do i seem to attract stalkers? Seriously every day it;s at least one new one who randomly IMs me on gaia as i get online for the day... It\'s ludicrous i tell you ludicrous! and it\'s never anyone from my neck of the woods, no i attract wanna be gangstahs and 12 year old punks it\'s tons of fun.. Really it is. [/sarcasm] Anyway i found another barn that\'s kickin but it\'s more an adult barn then a kids barn so no shimby riding there and you all have no clue what I\'m talking about and I\'m to lazy to write out about it so i\'ll leave it for latter. Well as this is a wicked short post i\'ll try and write more latter but theirs just nothing big to report... you know you love me, Shimby  '),(10131,3601175,'30,June,2004','2004-06-30 00:00:00','   Hey guys I\'ve really gotten fed up with all the fucktards around this hellhole!  I mean come on, how fun can it be to ruin someoneâ€™s life?  Thatâ€™s what there doing to poor barb; i can\'t stand it its all those damn mini van moms who are doing it.  The only thing they care about is making themselves look good and if one of them doesnâ€™t like you they all hate you, i really hate Erikaâ€™s mom, my god that woman is a real bitch!  Sheâ€™s a shark threes no doubt about it, i really can\'t stand her she grabs on to something and then make s all the other mothers who are all stupid sheep who would follow there leader to the ends of the earth.  No you know what they are?  There lemmings, they follow each other off cliffs if it meant staying with the herd and in favor, it\'s not like they could *Gasp* think for themselves.  URRAAGA!  Itâ€™s pissing me off to no end.  And then Megâ€™s mom is trying to get all them to take lessons with Margie, um no okay Stephane can not ride, hell i ride better then her she cannot take lessons with Margie and besides she thinks barb is the fucking center of the universe and everything she says is law.  But anyway none of you have any idea what the hell Iâ€™m thinking about anyway so i should prolly stop this mindless rant now shouldn\'t i?  OK on to more important things, like the fact that were painting the jumps at WRF, and Chris is having a field day with how dirty were getting, i mean take me yesterday, i was covered with paint and he was teasing me and shit, god i thought guys grew out of that by the time they were his age!  I guess i was mistaken now wasn\'t i?  Anyway Governor R is gone and now we have reel, this could be interesting, i wonder if she\'ll be better then the jackass who stole our money.  OK so he didnâ€™t steal out money but he was still a jackass, taking gifts and stuff hmm-bad boy mister john...  OK i think thatâ€™s about it.  Hope somebody is actually reading this, or else i just wasted a hell of a lot of time...  You know you love me, Shimby  '),(10132,3601175,'13,July,2004','2004-07-13 00:00:00','   Hey all Ever have those days were your just having the best time of your life and everything seems to be going your way? Well today I had one of those. First I worked for the first time today at the barn, it was fun really fun. But the good things started happening after work when I was just hanging out. First i rode presto and my God he was perfect, he didn\'t even flinch when Chelsea rode by on Fi in full canter! If he\'s like this at breezy pines I can say hello first place baby! The next good thing was you know how I\'ve been scared to talk to Karl? Well I was grooming Holly and I was in her stall bending over to pick out her feet, and he comes down the aisle with Sterling, he sees the stall door partially open and shuts it, with me still in there. So his mom asks him what he\'s doing and he says the stall was randomly open, so I had to pipe up and say because I was in there. then we actual talked for a Lil bit.. made me feel all warm and fuzz... wait a second i\'m starting to sound like a fan girl at a John Mayer concert no way in hell am I going down that road.... Any way you know what happened when you have great days like that? Well bad things happen, thats what. Mandy has been leasing Jete for about 2 months now and shes been doing great with him, until the psycho bitch from hell comes along and decides to give him to the horse guard. Damn did that make me SO mad i wanted to punch Martha\'s lights out, if their\'s one thing that really makes me mad it\'s when people mistreat my friends. So they came and took Jete away today and Mandy is devastated. The other bad thing that happened was Chelsea got bucked off Big Sunny, while I was being locked in the stall with holly and making goo goo eyes at the boy (stupid me, my god i\'m a jackass) my best friend gets her ass kicked by an overweight Thoroughbred.... Oh well easy come easy go all in all the good outweighs the bad and like some old proverb says with every ending comes a brand new beginning. So i think i shall end this post, or am i just beginning a new adventure, a new challenge in the next day? You know you love me, Shimby  '),(10133,3601175,'11,July,2004','2004-07-11 00:00:00','   hey guys nothing special really going on, i re joined the realm after the whole i killed Shimby fiasco, beginnings just about dead, i\'m considering selling my virginity on ebay and i may have found a substitute for elf boy. OK my computers being a laggy SOAB so i\'m getting pissed also. but anyway to elaborate on those things. i wash tony *my neighbor*\'s truck today damn was it dirty i could have spent weeks doing a really good job, but whatever it\'s a work truck got paid 75$ for the two of um. Ana/elpi has been in Tennessee all week so she hasn\'t been around to tell me, it\'ll bounce back@! about Begining so i\'m getting pretty stressed haven\'t seen either of the guys in weeks and it\'s starting to get to me.. fragile ego this isn\'t helping.... About the virginity thing i\'m just horny and in need of cash.. it\'ll work trust me XD OK i think i\'m done got a show on the 18th with presto he was epic yesterday.. hope he\'ll be as good at the show, his names gonna be luck o the Irish and it\'s at breezy Pines should be pretty easy no? hope it\'ll be good night all! gotta finish my spar with Sana.. You know you love me, Shimby  '),(10134,3601175,'22,July,2004','2004-07-22 00:00:00','   &nbsp;&nbsp; Hey all, Did you know 56% of couples end up divorced before there 40th year of marriage? well Iâ€™m guessing my the way things are going that my parents are gonna end up adding to that number. seriously, every frickin thing i say they have some snide remark about the other parent! take today when my dad picked me up from the barn, today you see i snapped i didnâ€™t grin and bear it, oh no i told him off, heres how it went down me: it was to hot to ride today Dad: you couldn\'t ride? me: no it was to hot for presto Dad: did you drink enough water? do they have clean water at the barn me: yes they have clean water, it\'s not that hillbillyish Dad: My god! i was just asking a question! between you and your mother i can\'t even- Me:(interrupts, starting to cry) STOP IT i was joking around can i not even do that??! between the two of you&nbsp; everything i say is turned into a insult against the other parent! do you have any idea how hard that is???!!! then we were quiet for about 3 miles and then he starts Dad: so you did drink enough water did you do anything else? thatâ€™s it no, Iâ€™m sorry Meg, and no Iâ€™ll try not to make you feel like a go between in our stupid piety fights. My god really they have no clue how hard it is for me to have to be torn between the two people i love the most and have to watch their marriage deteriorate into a pile of ashes. it\'s scary that i have to find out the love isn\'t forever, that was something i really used to belie in and now.. now it\'s .. it\'s just gone so you know what? i think Iâ€™m gonna go in my room turn the radio on and forget them  You know you love me,  shimby  '),(10135,3802748,'30,June,2004','2004-06-30 00:00:00','   I seem to have a blog. Whether or not this will prove an interesting addition to my (and your) lifestyle remains to be seen. For now, this post is just a  test  to see if I can arrange this blog to suit my fancy. I\'ll write more soon, of course. That is, after all, the point. :)  '),(10136,3802748,'02,July,2004','2004-07-02 00:00:00','   Well, here I am again.  I\'ve spent some time fiddling with the look of this page until I am temporarily satisfied with the results. What remains to be found now is a topic...something interesting and preferably not too serious. Life is too serious most of the time. I don\'t feel like contributing to that at the moment.  Here\'s something not too serious (which people take seriously): football. Well, actually, in this case I\'m talking about what we Americans call  soccer . Officially, the Euro 2004 championship football games.   Now, I happen to live with Portuguese people, and their team is one of the two finalists (along with Greece). They\'re not even diehard sports fans. I rarely seem them watch any other sorts of games throughout the year. But Euro 2004 is an all-pervasive infectious aura of enthusiasm, anxiety, and general fanaticism. So these normally rather non-sports-oriented people are presently glued to the Euro 2004 reports and occasionally to the games themselves (when they can endure the stress).   When they were watching a game in which Portugal was involved, I felt inclined to hide myself in another portion of the house and avoid all the jumping up and down, yelling things and manly exclamations of encouragement and athletic superiority. The championship game should be a riot. :)  And I mean that literally. I live in Brussels, the capital of the EU, so every country has representives to send into the street for a massive party or expression of grave dissappointment at every conceivable opportunity.   But, anyway, what it comes to is...I don\'t get it. I can admire the skill of the athletes, sure. Some of them are incredible. But I just don\'t feel even remotely as intensely passionate about every moment. Maybe it would be different if I was born in Europe and had a team to feel passionately about. I don\'t know.  I do feel passionately about some other things, like the online game I run, or cats, or about the greatness of my fiancÃ©. But I don\'t dress up like someone from my game (you should see the soccer fans in their flags and war paint), nor do I wear T-shirts that say \"I   Cats\", nor do I jump up and down yelling with empassioned joy when my fiancÃ© comes home from work. So this sports phenomenon has to be somehow different.  I love seeing people so happy, really. It\'s really inspirational to see Portuguese news broadcasting the massive crowds of people lining the roads on the work day just to cheer their team on for the bus ride between the hotel and stadium. And the players themselves look nearly as majestic as their fans worship to be, crossing themselves as they enter the field and standing proudly as their national anthem is played.  My father once explained to me that modern day sports have taken the place of the wars of old. We cheer for our army (team) as our favorite and most skilled warriors (players) step onto another battlefield (stadium) to fight for honor and glory (that much is the same). Only here there is no intentional bloodshed. There are rules and regulations to make sure everyone has a safe and wholesome experience.  Sure, sometimes some of the fans get out of hand, but it\'s not nearly as destructive as actual warfare. And everyone ends up feeling closer with their fellows (either in victory or defeat) and has something interesting to discuss for weeks to come. It seems like a healthy enough outlet don\'t you think?  I\'ve played sports before, myself, (yes, really!) and I remember what it was like to be on a team, with the fans shouting from the side and the exhileration of working together as a whole unit to accomplish the singular two objectives of putting the ball in the other team\'s net and preventing them from putting it in ours. (This was field hockey, mind you. Not soccer.)   But even though I played, and found that immensely enjoyable, I was never an outsider fan of any sport at all. I never collected baseball cards. I resented when I wasn\'t able to make a sound about the house because my father was eating all the pretzels and staring at the (American) football game on Sunday. I got out my embroidery or a book whenever I was visiting someone who was terribly into watching the latest golf/bowling/tennis match. I think baseball is more fun to watch than anything, but perhaps that\'s due to fond memories of going to see the Phillies with my Dad (and Mom and sister and Grandma) as a kid.  So while I don\'t actually feel a part of the sports fan club in general, I can appreciate why people are very enthusiastic. Considering that this passtime is nonharmful (unless you drink too much and break things with your exuberance), healthy (if you play it yourself), enhances a sense of comradery among fans of the same team, and helps a lot of people to better enjoy their life, I can clearly and elegantly dub sports a good thing. :)  '),(10137,3672882,'20,June,2004','2004-06-20 00:00:00','   ...so anyways...I spent the night at my firend Sean\'s last night.  We had a great time. It was me, Taylor, Skyler, Brandon, and Sean.  Skyler said that he was going to go home...but we caught a glance of his AIM conversation.  Turns out, he went to a resturaunt...with a girl from a different school to get breakfast (at midnight..).  And..we wouldnt have made anything of it but...he lied to us.  So we went to the same resturaunt!  We sat at a booth right in front of them.  We all got milkshakes and we bought  one  for them.  It had two straws.  Then this like 18 year old chick came up to us and started asking us a bunch of questions...she thought we were stoned (which we were not).  We were a bunch of 16 year olds...luahging our asses off...with milkshakes...at midnight...and my eyes were bloodshot from allergies...we did appear to be stoned.  But we werent.    Then we went back to Sean\'s house...and watched CKY.  Brandon and Skyler fell asleep...so me, taylor, and sean got them with shaving cream and a feather!  Then...we went into seans room...and found this can of easy cheese...and squirted it on his floor to piss him of...turns out...that can of easy cheese was from Aug 2002....so his room was extremely rank afterwards...  We just kind of fucked around \'till about 5:00 am.  And they all decided to go to sleep.  So i went home...but first i stopped at safeway and bought Coco Puffs.  Then i went to sleep...and nothing happened for much of the day.  But at around 1:00 am...i watched Cowboy Bebop and learned about a drink called a Prairie Oyster.  Basically...its Tomatoe juice and a raw egg...and of course i had to make one.  It wasnt that bad.    Well mamas and papas...im done for now...  -Al '),(10138,3672882,'17,June,2004','2004-06-17 00:00:00','   So anyways...I woke up at around 1:00 pm this afternooon. My \"breakfast\" was a bowl of skettios. My dad had to get an operation for this bump on his cheek. He\'s fine.  After he got back I checked out Lollapolooza\'s website. I really want to go. I just gotta get some tickets...and someone to go with. One of my foavorite bands, String Cheese Incident, is gonna be headlining day 2. I havent really heard of much of the other bands...soooo....I did some 100% illegal downloading on one-a-them p2p thingies (bearshare). I now have a few new favorite bands. The Pixies, The Flaming Lips...yeah.  I then went with my dad over to my grandma\'s house to...do something. I brought my guitar and recieved praise.  Then I came back home and looked at the comics on www.redmeat.com Oh my God thats some funny shit. New favorite.  Then I decided to get bored and went to my friend Shorty\'s house....and did nothing except break my car\'s CD player...fuck. So now I have to buy a new one...with money.  Now I\'m back home...im gonna watch a movie or something...I hope Goodfellas is on.  So anyways...bye...  -Al '),(10139,4227656,'21,August,2004','2004-08-21 00:00:00','   So...this youth group.  It doesn\'t come with instructions.  Hmmm that reminds me of what new parents say about kids.  Is this the same?  Am I now a mother preparing for her newborn?  Or am I merely a caretaker and God is the father of this new baby?  Maybe I should know the answer to that question. I need to have a network of people around me already in the ministry, but so many of them are already too busy.  I understand that, but then feel left standing in a foreign land without instructions.  The instructions?  Hmm maybe the Bible, but ... then there\'s the age old question of how to decode some of the answers.  And also the age old problem of waiting on God. So here I am standing on the border between my old life and a totally foreign land.  I barely speak the language and barely understand the culture, but I have a passion burning so hot inside me that if I don\'t move it will consume me ... destroy me. I have to start a youth group.  I get to.  But ... how. I ask the question and it seems to bounce off the empty spaces around me.  They say that they are there to support me, but I don\'t think they really understand what it\'s like until they too have to step out there and try it. A name?  A theme?  What to study first?  Do I talk and sign at the same time?  Do I relate to only the Deaf?  Do I include the hearing?  Do I start a middle school plus high school group?  Do I seperate it?  What do I do when I get there? They trust me.....What an overwhelming responsibility.  '),(10140,4227656,'18,August,2004','2004-08-18 00:00:00','   Well, it\'s getting to the point that I\'m beginning to feel spred too thin and can barely keep from crying when I think about all the things I need and all the things I don\'t have.  I still have faith and I still trust God, but I feel like the faith and the trust is beginning to fade.  It\'s like I\'ve held on to the ledge for a really long time and my fingers are beginning to slip, my arms are no longer able to hold on to the promise that God will provide.  I know that I need to keep trusting and keep the faith, but it\'s getting harder.  I also wonder, now, just how many people out there wait years for things and keep the faith.  Where does it come from?   *sigh* I\'m waiting for God to open the flood gates and provide all of the things I need.  I want my dad to see that God really does give us things that we need, that He provides in times of need and that miracles do really happen.  I think my dad\'s going crazy over this whole thing...worse then me.  I really would like to show my dad and help him revive his faith in God.  I wonder if I\'m doing something wrong ... or not doing something that I should be doing.  When I think about it nothing comes to mind, but ... maybe I\'m not thinking hard enough.  Or maybe these are the lies of the devil. Please don\'t you have any answers?  '),(10141,4227656,'17,August,2004','2004-08-17 00:00:00','   Well, it\'s getting to the point that I\'m beginning to feel spred too thin and can barely keep from crying when I think about all the things I need and all the things I don\'t have.  I still have faith and I still trust God, but I feel like the faith and the trust is beginning to fade.  It\'s like I\'ve held on to the ledge for a really long time and my fingers are beginning to slip, my arms are no longer able to hold on to the promise that God will provide.  I know that I need to keep trusting and keep the faith, but it\'s getting harder.  I also wonder, now, just how many people out there wait years for things and keep the faith.  Where does it come from?   *sigh* I\'m waiting for God to open the flood gates and provide all of the things I need.  I want my dad to see that God really does give us things that we need, that He provides in times of need and that miracles do really happen.  I think my dad\'s going crazy over this whole thing...worse then me.  I really would like to show my dad and help him revive his faith in God.  I wonder if I\'m doing something wrong ... or not doing something that I should be doing.  When I think about it nothing comes to mind, but ... maybe I\'m not thinking hard enough.  Or maybe these are the lies of the devil. Please don\'t you have any answers?  '),(10142,4227656,'17,August,2004','2004-08-17 00:00:00','   Well...I got to work today!  My dad has a friend that needed some doors painted and some yard work done.  I worked 5 hours today and will probably work about that many on Thursday.  I\'m hoping it will help me out a little. Come to find out my clothes are falling apart and I\'m desprate for some more, but I don\'t have any money.  I know that sounds like the same old same old, but it\'s true!  I\'m trying hard to be good about not spending my money, but ... I don\'t know how much longer I\"m going to be able to last without money.  I know I\'ll be poor, but I\'d at least like something.  Okay so right now I\'d like a lot of something, because I need it to move to Southern Cali. Pray.  '),(10143,4227656,'13,August,2004','2004-08-13 00:00:00','   Well, I created another online journel here at blogspot, because of Barbara!  I also like that they changed the format.  I like that you can now add comments. This is actually my second online journal, but I don\'t mind.  This actually has a different tone then the other one. I hope that I can actually get people reading this thing!  That\'s actually rather difficult to do.  I wonder why!  It\'s rather like an email without all the hassal of trying to get all the right addresses and stuff like that! I hope I can keep up with this. I haven\'t heard from the church in Riverside yet and a lot of people keep asking me about it.  All I can say is keep praying.  I know that worrying does me no good and gets me no where, but I feel like I want to start worrying more.  It\'s hard to keep the faith and hope alive when I see no real good solution. Waiting for miracles.  '),(10144,4177430,'16,August,2004','2004-08-16 00:00:00','    Everyone who has been brain washed by the right\'s attacks of John Kerry as being a \"flip-flopper\" on the war, should read this wonderful Buzzflash article. Every liberal pissed at Kerry, claiming he\'s a war monger, or believe he would\'ve also gone to war with Iraq, if president, should read it as well.  urlLink     urlLink What Did the Joint Resolution on Iraq Really Say?   '),(10145,4177430,'14,August,2004','2004-08-14 00:00:00','    Recently, the california supreme court nullified all the gay marriages in San Fransico because they supposedly contradicted the state\'s consitution.  However, under the first two clauses of the 14th amendment, people must not be deprived of their liberties without due process and are subject to equal protection of the law.  This means someone could get married to another of the same sex in Massachusetts and come to, say, Ohio, and declare their marriage legal.  This would be frigging revolutionary and I don\'t understand why someone hasn\'t done anything about it yet.  I looked it up in the massachusetts marriage laws, and a minor can be married with the consent of the parents.  Anybody want to go to Massachusetts and get married to one another?   '),(10146,4177430,'14,August,2004','2004-08-14 00:00:00','    Check out this great comic someone Photoshopped together.  Incredibly well done and hilarious:  urlLink http://www.pointlesswasteoftime.com/nscomics.html   '),(10147,4177430,'13,August,2004','2004-08-13 00:00:00','    Go here   for an essay Kurt Vonnegut wrote about recent goings on in the country and the world. Truly magnificent, as he always is.    Moveon.Org just recently came out with these ads of real Americans talking about why they\'re voting for John Kerry. What\'s surprising is that the majority of the people in these ads are Republicans, which proves that George W. Bush is a divider, not a uniter. I mean, he even alienates people in his own party. What a guy.  Watch the ads  urlLink here.   '),(10148,4177430,'13,August,2004','2004-08-13 00:00:00','    This is an absolute must: an audio presentation of a picture in exactly 1000 words.   urlLink http://www.motherjones.com/commentary/columns/2004/08/keizer.html   '),(10149,4177430,'11,August,2004','2004-08-11 00:00:00','    Sweet anti-Bush rap video.   urlLink http://www.slambush.net/video/index.html   '),(10150,4177430,'11,August,2004','2004-08-11 00:00:00','    READ THIS ARTICLE:  Tells all about how Bush knew the evidence against Iraq was weak and went ahead with war anyway.   urlLink http://www.inthesetimes.com/site/main/article/they_knew_0802/   '),(10151,4177430,'11,August,2004','2004-08-11 00:00:00','    ANY BAY HIGH STUDENT INTERESTED IN JOINING BAY HIGH YOUNG DEMOCRATS THIS YEAR====> please e-mail me (David) at   seamlessline@yahoo.com   so  I can startup a list of names and emails of those interested so I can update and inform you all much easier and get an idea of the amount of interest. Even if I\'ve already talked to you, send me an e-mail!   '),(10152,4177430,'11,August,2004','2004-08-11 00:00:00','    I got about 80 some people called today (all in Bay Village!) and found two big Kerry supporters, Anita Guerin being one of them (a friend\'s mom). This time, Robert Grossman, Dave Rogers, Ali Tramba, Erin Looshel, David Antone, and Jenny Evans came along. Sometimes all the lines were busy and we had to wait for one or the other to get off. Having more people exponentially increases your chance of hearing funny stories--Rob had over 4 people talk about their miscellaneous injuries and maladies with him, David called someone named Myrlin (who we wanted to ask to put a hex on GW), and I got ahold of someone who was \"on the road.\" David and I noticed how anyone who supports Bush always says \" BUSH .\" so definitively and defensively. The ones who say \"Kerry\" pause a little and seem like they\'re really thinking. I called someone who didn\'t register because she believed that all the canditates were the same. Anyone who says that is terribly uninformed. Yeah, Dems and Republicans alike do cater to big business, and there are some irksome similiarities, but get your head out of your politically aloof ass. Kerry has the 11th most liberal rating in the senate. He\'s great at speaking, actually knows what he\'s talking about, and has a plan to get the country out of the gutter into which it\'s climbed and prepared to die.  Saying \"they\'re all the same\" is just an excuse for you to be apathetic.  I saw this in someones profile: \"I hate it when all these people aren\'t voting because they say that it\'d be just picking the lesser of the two evils. For god\'s sake, pick the lesser!\" (don\'t remember who it was from though). I don\'t think Kerry is really an anti-Bush ticket. He\'s got a good plan for a) getting our troops home while simultaneously balancing Iraq\'s turbulent government b) dragging our sick, oil-sucking economy out of the depths of defecit c) getting our civil liberties back and \"ending the age of Ashcroft\" d) backing up our ailing schools. I think he\'s harping too much on his Vietnam service, but we\'re not getting on Bush\'s case enough.  There\'s a huge difference between dodging the service because you disagree with war and dodging the service because you\'re a coke-blowing pussy . Oh, and that shit about Swift Boat Veterans for Truth--all lies. Check out  urlLink www.spinsanity.org  where you\'ll see the downlow (the headline is Spinning Swiftly, date 8/9). David and I are going to volunteer at the Festival of the Assumption and pass out literature. I expect we\'ll have lots of supporters, but some who will go on about Kerry\'s Pro-Choiceness. Hello people: Bush.  IS.    NOT.      PRO-LIFE  .  Think of all the people that man has put  TO DEATH  as governer of Texas. Think of all of the innocent Afghanis, Iraqis ( who are now in the 10-13 thousand range   urlLink www.iraqbodycount.net ) and not to mention our own soldiers. How can you call him a Pro-Life president? It\'s a non-issue, throw it out, don\'t let it decide your vote. Besides, it\'s absolutely ridiculous to hinge your vote on something like that. If you want, you can be pro-life UNDER KERRY who has all the benefits of being A MUCH BETTER CANDITATE THAN BUSH. And must we go spiralling into this debate?  Pro-Choice does NOT mean pro-abortion . Kerry just doesn\'t want to force his religious beliefs on others. And you know what? Sagging economies means less money to raise children which means the numbers of people having sex goes down...PFFF--WRONG. The number of abortions should go up. (I\'m hypothesizing here, don\'t take this for absolute fact.) Who do you want at the helm? Someone who\'s \"Pro-Life\" and is responsible for terrible economic plight, or someone who\'s Pro-Choice and pulls us up to a position where people won\'t want abortions. Huff. Well, to end on a good note, here\'s a clip of GW getting totally owned at the Unity Conference (collection of all minority journalists who asked the President and Kerry questions, each of whom had different days. George got 5 rounds of applause including the ones where he got on and off stage. Kerry had 20 some--am I right David?).  urlLink http://poststuff2.entensity.net/081204/media.php?media=bush.wmv   ____________________________________________________________________  From Davy:    Yeah calling is a sinch and all, but man....all I have to say  is that the majority of the elderly people in Bay Village are a  CRANKY  bunch to say the least. Just because you\'re old and senile doesn\'t mean you have to moan and fart at me for volunteering to help make the country a better place for me AND you (some of you just fail to realize that yet). Other than that, the calling went fine. Called 100 people and I had more Kerry supporters this time which is surprising for Bay Village. Thanks to Steve for that link to the Bush vid. I watched all that live on C-SPAN and I\'ve been waiting ever since for a clip of that ramble. Mark, from what I remember, Kerry got about 40 rounds of applause. Bush got AT MOST 10. Don\'t forget that half-way through Bush\'s speech a man stood up and started yelling at Bush in protest(thanks to C-SPAN for showing the dude). After about....say 5 seconds, the man was swarmed by the secret service and escorted out of the building (hopefully not detained indefinitely as an ENEMY COMBATANT). Let\'s just say that the UNITY Conference was just a sweet sneak preview for the debates in October.     \'Bring it on!\', George. Give em\' hell, John!  Myrlin casts Lvl. 5 Hex Spell on GW! ROLL D20!  2nd post by David Edward Antone   '),(10153,4177430,'11,August,2004','2004-08-11 00:00:00','    This past Monday, as Mark said, we went to the John Kerry phone bank and it was a Grand Ol\' Time. I had a couple very kind Kerry supporters, and a few very kind people who had to tell me the person i was calling wasn\'t home (such as one person who\'s son that i was calling was off training in the military). The majority of the people I called were extremely old and most of the people I called were Bush supporters. Mark and I were stuck calling all South Euclid residents. According to the woman at the phone bank, that\'s a fairly conservative city. We\'re going again tomorrow with our teacher Mr. Grossman tomorrow (Thursday). Should be fun with more people  .      Now, on to voting third party. Anyone planning on voting for a third party candidate for President, I beg of you...please don\'t. I\'m begging you because I know for a fact that most of you have good intentions, but I strongly believe that you are unknowingly turning a blind eye to the well-being of the country and the world. This especially applies if you are voting in the crucial battleground states such as my home state of Ohio. I understand that you are upset of the dominance of the two-party system but if we all don\'t do something now, the Republican Party will become extremely close to making the United States of America a ONE-PARTY STATE. The Republican Party currently controls ALL THREE branches of the US government. This dominance will continue to flourish if George W. Bush is re-elected, especially in the crucial judicial branch (due to Bush\'s possible future judicial appointments).    We cannot sit idly while this happens because this ain\'t yo\' daddy\'s Republican party! The Neo-Conservative movement continues to grow within the Grand Ol\' Party with their desire to limit or even destroy certain civil liberties, cater to special interests more than ever, and conduct an arrogant, prejudicial,and unilateral foreign policy. Clearly, electing a Democrat to the highest office in the land would put a stop to this neo-conservative movement to overthrow of the US government, preventing an government-wide infestation of Ashcroft-esque politicians.   My point is that THIS ELECTION IS  NOT  THE TIME TO SPLIT THE LIBERAL VOTE AND GIVE GEORGE W. BUSH AND HIS ADMINISTRATION MORE CHANCE OF BEING RE-ELECTED, ALLOWING FURTHER NEO-CONSERVATIVE TAKEOVER GIVING LIBERALS LITTLE OR NO VOICE IN THE UNITED STATES. So if you love this country and if you love this planet, don\'t vote for George W. Bush or put your reservations aside for this election, and don\'t vote for Ralph Nader or any 3rd party candidate.Vote for John F. Kerry. Don\'t split the vote, please.   '),(10154,4177430,'10,August,2004','2004-08-10 00:00:00','    David and I went to the Kerry Phonebank in Rocky River yesterday for two and a half hours.  It went by very fast--we called over 100 people and asked them who they wanted to vote for this election.  If they said Kerry, we would ask them if they could help out the campaign in any way possible.  Out of 100 people I got 3 Kerry supporters, 4-5 Bush, and one Nader.  All the rest either didn\'t answer or hung up.  I was a little nervous at first, making cold-calls, but after a while, it got easy (especially since most didn\'t answer their phones).  David and I had some funny encounters mostly involving either really old people or really hardcore Republicans (David\'s are funnier though--I\'ll let him post the highlights).  One time, in weariness, my fingers fumbled and I accidentally called 911.  Good thing you have to dial 9 to get out, or else that call would have registered.  I did it a second time, but this time, I had already entered the area code too, and it started ringing!  I hope dispatch knows the only emergency we\'ll have is a case of over-flipping out when Kerry wins.  The time, surprisingly, went by very quickly.  I calculated that we pressed over 1,500 buttons.  Even though we got a lot more Repugnicans than Dems or Indies, when a lady working there saw that we were calling South Euclid, she rolled her eyes and said, \"Oh  yeah .\"  We wanted to call up our friend Kyle who is (in our humble opinion)   w  a  s  t  i  n  g     his vote on the Green Party, but we forgot his number.  Speaking of third parties and wasted votes: IF YOU ARE LIBERAL, THERE IS NO LOGICAL REASON IN THE ENTIRE WORLD THAT YOU SHOULD BE VOTING FOR ANYONE BUT JOHN KERRY.  People who want to vote Green seem to forget that Kerry has a 95% on environmental issues.  A 60% is considered great.  Considering that there are no real conservative third parties (unless you count Nader, who gets most of his support from Republicans who...guess what!...want him to take votes away from Kerry), it\'s a fractured base against one other party.  No one cares about your \"principles\" in this election, possibly the most important election of our lives.  Getting the 5% for next election won\'t matter when, by then, our country is  FUBAR .   What Green and third parties need to do is not go straight for the President, but build up a base in lower positions, congressional or local, so that when the time comes for them to run for president, they actually have a base of voters that has a chance of getting them elected, not just for this mythical Holy Grail of 5%.     What are they going to do with the Federal funding they get if THEY DON\'T HAVE A LEGACY OTHER THAN 8 YEARS OF BUSH?  No matter what anyone tells you, third parties  do not work in an electorial college .  People cite third parties as being responsible for the demise of slavery, leaps and bounds for labor, and other special interests, but guess what!   THEY\'VE NEVER BEEN ELECTED THEMSELVES   (other than Teddy Roosevelt, who ran on and Independent platform that was exactly the same as the Republican one).  The best they can hope for is one of the larger parties absorbing their platform.  And considering that John Kerry has a NINTY FIVE PERCENT on environmental issues, I would judge the Green Party effectively absorbed (with, ironically, only 5% left to them).  We need to a) not vote third party, even in non-swing states.  Contributing to the popular vote count is as important as tipping the vote in swing states; b) get as many people voting as possible.  The more people we get to vote, especially in the 18-24 age group, the inevitable result is more Kerry supporters.  If Bush wins this election, it\'s not because a majority of Americans want him to be president, but because a majority of Americans didn\'t make it to the polls.  You know who really stole the election in Florida?  It wasn\'t just Bush.  It was third parties.  Six hundred plus votes went to Socialist or Communist parties.  Hey, at heart, I may be with them, but WHO ARE YOU KIDDING?  The last person that liberal that was almost elected was Eugene V. Debs, and that was during a labor crisis so bad that becoming a Pinkteron was the best career move you could make (considering everyone but them was getting their shit kicked).  The Florida election was so close that if those liberals had voted for Gore, it would have likely tipped the election beyond Bush\'s stacked Supreme Court\'s fanagling.    A vote for a third party is a vote for Bush.  There is no way around it.   '),(10155,4177430,'09,August,2004','2004-08-09 00:00:00','    If this stuff is true, I pity ANYONE in Iraq.   urlLink http://www.snopes.com/photos/bugs/camelspider.asp   I\'m alright with spiders as long as they don\'t crawl on me.   '),(10156,4177430,'08,August,2004','2004-08-08 00:00:00','    Bugger people who say, \"I don\'t know Kerry\'s position because he flip-flops daily!\"  This is a facsimile phrase pumped out by Fox and other Bushites.  Ask anyone about his \"flip-flops\" and they can\'t back up jack shit.  \"He voted for the war\" well yeah, that was before anyone knew that the president LIED about the WMDs.  Many of these \"flip-flops\" posted on Bush\'s reelection site span 10-20+ years.  Four years ago, I used to think Will Smith was a great musician, but I don\'t now, and no one\'s calling me a flip-flopper.  And have people forgotten that changing your mind is OK?  Hello?  Yes.  Changing your mind...is OK.  What do we want in office?  A guy who\'s bullheaded and never goes back on what he says or someone who can admit when he\'s wrong (and been wronged).  Anyone who talks about Kerry\'s Flip-Flopping has no idea what they\'re talking about because there is, in fact, nothing to talk about.  What about Bush \"flip-flopping\" on No Child Left Behind--a program he praised but left terribly underfunded.  What about Bush applauding veterans, then cutting their benefits?  What about Bush flip-flopping off his bike...twice!  There isn\'t a band-aid big enough to cover up Bush\'s scabbed political career.   '),(10157,4177430,'08,August,2004','2004-08-08 00:00:00','    Everyone needs to read \"Against All Enemies\" by Richard Clarke.  It\'s really funny at times (Colin Powell calls chemical weapons \"goofy\") and really creepy at others (all the holes that still exist in our current infrastructure) and sometimes just makes you pissed off too (if we would have continued Gulf War 1 for another 72 hours and destroyed the Republican Guard, the Shi\'ites would have sucessfully risen up and unseated Saddam...instead, they were brutally slaughtered by the R.G. when they rose up--why didn\'t we help?  We didn\'t want another Shi\'ite government like Iran...).  Clarke also delves into why so many people in the Mideast hate America, drawing on history he experienced firsthand over his 20+ years of service in counterterrorism within the White House.  This guy got so fed up with Bush\'s fanagling in Iraq that he up and quit.  Very well written and very engaging.  One of the most informative books you\'ll ever read.   '),(10158,4177430,'08,August,2004','2004-08-08 00:00:00','    We have us a new contributor at the Young-Democrats blog. My good friend Mark will be adding shit and shit because he, like I, hates America. Because if you\'re not a Republican, you\'re a terrorist!   '),(10159,4177430,'06,August,2004','2004-08-06 00:00:00','    Wow, this ad from America Coming Together featuring Will Ferrell is extremely funny.       Watch it  urlLink here   '),(10160,4177430,'05,August,2004','2004-08-05 00:00:00','    Read this truly   urlLink  awesome essay    by Ron Reagan on the Bush administration and the realization of the American public of this administration\'s lack of credibility, common sense, and compassion...    \"As Bush\'s former Treasury secretary, Paul O\'Neill, and his onetime \"terror czar,\" Richard A. Clarke, have made clear, the president, with the enthusiastic encouragement of Defense Secretary Donald Rumsfeld and Paul Wolfowitz, was contemplating action against Iraq from day one. \"From the start, we were building the case against Hussein and looking at how we could take him out,\" O\'Neill said. All they needed was an excuse. Clarke got the same impression from within the White House. Afghanistan had to be dealt with first; that\'s where the actual perpetrators were, after all. But the Taliban was a mere appetizer; Saddam was the entrÃ©e. (Or who knows? The soup course?) It was simply a matter of convincing the American public (and our representatives) that war was justified. The realâ€”but elusiveâ€”prime mover behind the 9/11 attacks, Osama bin Laden, was quickly relegated to a back burner (a staff member at Fox Newsâ€”the cable-TV outlet of the Bush White Houseâ€”told me a year ago that mere mention of bin Laden\'s name was forbidden within the company, lest we be reminded that the actual bad guy remained at large) while Saddam\'s Iraq became International Enemy Number One. Just like that, a country whose economy had been reduced to shambles by international sanctions, whose military was less than half the size it had been when the U. S. Army rolled over it during the first Gulf war, that had extensive no-flight zones imposed on it in the north and south as well as constant aerial and satellite surveillance, and whose lethal weapons and capacity to produce such weapons had been destroyed or seriously degraded by UN inspection teams became, in Mr. Bush\'s words, \"a threat of unique urgency\" to the most powerful nation on earth.\"       Oh, yesterday I received a recruitment letter from the vast right-wing conspiracy. More on that tomorrow...   '),(10161,4177430,'05,August,2004','2004-08-05 00:00:00','    3 days ago, the secretary of Homeland Security. Tom Reluctant-Neo-Con Ridge , announced a raise in the terror threat level from lemon to clementine (right?) citing specific information, from multiple sources that New York and New Jersey financial centers were the target of Al Qaeda terrorists that are possibly currently in the United States,(probably watching \'America-hater\'   Michael Moore\'s movies and partying with \'god-hater\' George Soros at this very moment.)       Although he says Wall St. is about to E*X*P*L*O*D*E! Sec. Ridge repeats his post-threat announcing mantra of \"Go on with your daily lives normally!\". Which in    B u s h\' s  A m e r i c a   consists of using what little money you have left (after losing your manufacturing job to outsourcing) to buy a $14 throw pillow at Wall*Mart (aiding in the eradication of small business) along with some Duck Brand duct tape to save yourself from biological attack(following Homeland Security\'s post-9/11 advice. Without knowing that Avon, Ohio-based Duck Brand duct tape makes up 46 percent of the nations duct tape supply.Is it coincidence that the Duck Brand is owned by a man named Jack Kahl, a Bush campaign contributor who gave over $100,000 to Bush\'s 2000 campaign?). As you exit the Wall*Mart and walk to your $2 per-10-miles SUV, you take a deep breath of de-regulated air and look into the pale blue sky. As you look into that sky you thank God that George Bush proposed the \"Clear Skies Initiative\", which can finally make the skies clear...of birds. You arrive at your apartment and read the last few e-mails of the day and any feeling of loneliness melts away because you know that your Big Brother Ashcroft was reading those e-mails too. Ah, Bush\'s America!    YEAH! TAKE THAT OSAMA!  T H E S E  C O L O R S DON\'T RUN! THEY CLIMB...THE CORPORATE LADDER!                   Soooo anyways, here are some interesting links pertaining to the recent terror-threats...      This  urlLink New Republic article  that was issued more than  two weeks  before the terror threat was announced discusses the Bush Administration urging Pakistan to capture Al Qaeda related high-value-targets (HVTs) such as Osama Bin Laden BEFORE the November election. Here\'s an excerpt-              \"But The New Republic has learned that Pakistani security officials have been told they must produce HVTs by the election. According to one source in Pakistan\'s powerful Inter-Services Intelligence (ISI), \"The Pakistani government is really desperate and wants to flush out bin Laden and his associates after the latest pressures from the U.S. administration to deliver before the [upcoming] U.S. elections.\" Introducing target dates for Al Qaeda captures is a new twist in U.S.-Pakistani counterterrorism relations--according to a recently departed intelligence official, \"no timetable[s]\" were discussed in 2002 or 2003--but the November election is apparently bringing a new deadline pressure to the hunt. Another official, this one from the Pakistani Interior Ministry, which is responsible for internal security, explains, \"The Musharraf government has a history of rescuing the Bush administration. They now want Musharraf to bail them out when they are facing hard times in the coming elections.\"           It wasn\'t until this threat that the mainstream media has questioned whether the Bush Administration has used terror threats for political gain. It\'s strange how almost every time something goes against the favor of the Bush Administration, they soon announce a raise in threat level from  jungle green  to  fuschia , or from  midnight blue  to the dreaded  macaroni n\' cheese ( terror threat levels obviously monopolized by Crayola). The controversial shady terrorist threats are what comedian David Cross sarcastically refers to as \"A FUCKING COINCIDENCE!\".  Here\'s New York Times  urlLink editorial piece  on how Homeland Security Sec. Ridge\'s recent terror announcement at times felt less like a counterterrorism announcement and more like a counter-convention announcement.     \"Some of the past terror alerts have seemed aimless and happened when the Bush administration would have benefited from a change in the political conversation. On Sunday, when the administration had grim and specific information to convey, Mr. Ridge did a real disservice to himself, his president and the public by giving what amounted to a campaign pitch for \"the president\'s leadership in the war against terror.\'\' It\'s hard to write that off as an offhand comment. If Mr. Ridge is to continue in this role, he must stay out of the election; using him as a campaign surrogate would be disastrous for public confidence. The administration should also stop dropping dark hints about Al Qaeda\'s having election-related motives to attack, as if a vote against the current president were appeasement.\"   Hm, what could be against Georgie\'s favor this time? Hmmm, maybe THE ELECTION ON Nov. 2 WHICH HE WILL MOST LIKELY LOSE! Bush MUST GO!   '),(10162,4177430,'05,August,2004','2004-08-05 00:00:00','    Hello. My name is David Antone, I live in the Cleveland suburb of Bay Village, Ohio and I am currently 17 years old. I attend Bay High School and will be a senior this fall. I recently decided to dedicate myself, for the next ninety days to helping elect Democratic Nominees John Kerry and John Edwards as President and Vice President of the United States and Eric Fingerhut as Ohio U.S. Senator. I have started this blog as an internet base for the Young Democrats Club that I plan to start with some of my friends at my school. I will be posting various articles and information pertaining to the election and the shameful and dangerous monstrosity that is the Bush administration. I will also post info on various volunteer opportunities around the Cleveland area to help register and educate voters, attend rallies or protests, etc. This may be the most important election in our lives. We all must unite, and practice our civic duty of voting to elect John Kerry to prevent another 4 years of the neo-conservative, special interest tailoring, and arrogant Bush administration.      ANY AND ALL HELP IS WELCOME!      If you wish to...     Contribute an article, some information,give suggestions or anything to this blog    Join the Bay High Young Democrats  OR     Request any help from me     CONTACT ME:     via  E-Mail: urlLink seamlessline@yahoo.com       OR post a comment on this blog.    I will try my best to update this blog frequently.     LEARN MORE:         urlLink http://www.johnkerry.com    urlLink http://www.fingerhutforsenate.com    urlLink http://www.airamericaradio.com  (the BEST talk-radio in the nation)  more links coming soon...  urlLink   '),(10163,3819051,'04,July,2004','2004-07-04 00:00:00','    I am the one and only guru. This time i have come in the form of a hardcore teacher. One day when i was playing my holy xbox, i noticed some people sneek by my holy ashram. I had to pause my game and see what it was. It was the mudafu**in teletubies, so i decided to do something rather not so nice to them.      I hated the teletubies because they were a bunch of hardcore losers. They had come to my&nbsp;ashram to get a blessing, but couldn\'t find the door. I thought they were rather thick in the head, because my door, to the ashram won an award for biggest yet humblest door! I felt that they needed a punishment because of criricising my door, by thinking very  little  of it.      &nbsp;    So this is the punishment they got:      &nbsp;     &nbsp;    When walking by... &nbsp;&nbsp;      &nbsp;     &nbsp;     &nbsp;     &nbsp;     &nbsp;     &nbsp;    When buggering me...  &nbsp;       &nbsp;     &nbsp;     &nbsp;     &nbsp;     &nbsp;    When getting on my nerves...     &nbsp;   &nbsp;   &nbsp;    when they couldn\'t find moi door ala fantastique... &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;      &nbsp;     &nbsp;     &nbsp;    When they criticised my door...       &nbsp;     &nbsp;     &nbsp;    When there was only one left... &nbsp;i did it myself...&nbsp;&nbsp;&nbsp;&nbsp;      &nbsp;     &nbsp;     &nbsp;     &nbsp;     &nbsp;     &nbsp;     &nbsp;     &nbsp;     &nbsp;    &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;taught em\' a lesson!!!!!!!!!!!!!!!!!!!&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'),(10164,3819051,'03,July,2004','2004-07-03 00:00:00','    Hahahahhahahahahahahahahahah! ............... I am Terminator Guru!... I have come to trminate all the&nbsp;human beings&nbsp;who&nbsp;oppose moi!!!!!... I am going to  &nbsp;all you freaky, retarded, opposing&nbsp;humans&nbsp;&nbsp;who are&nbsp;trying to&nbsp;destroy me and moi holy scriptures!!!&nbsp; Hahaha, i, am the holy guru whom the has the power within the holy scriptures to destroy all&nbsp;evil people!!!     &nbsp;    P.S: I wear a very holy dress...      &nbsp;     &nbsp;    original plan...     &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;The way to kill      &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;or other ways...&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  &nbsp;&nbsp; &lt;fellow&nbsp;humans against me!!!!!!!!            &nbsp;      &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;hahahaha!!!!!!!!!!&nbsp;&nbsp;&nbsp;&nbsp;        &nbsp;     &nbsp;   &nbsp;'),(10165,3819051,'03,July,2004','2004-07-03 00:00:00','    &nbsp;    Greetings my fellow pandits... i am a new user of this  blog&nbsp;  so mind me if i get confused...     Now let me introduce myself to... Ohhhh!... PANDITS?????     Ah... I knew they would come...one year... Well this is the mystic Guru\'s Ashram. I am the Guru of this gloomy place.      If you have seen my profile then You can tell how holy it is. Well... (it is very holy.).      &nbsp;    You may enjoy my time here, and Ilook forward to hearing from you.&nbsp;     As pou see, i\'m very&nbsp;keen in studing holy scriptures with you!     I hope also that you are keen in studying the ancient myths ...so on and so forth...     This year is a year of enjoyment and happiness. (nott, well not for me...).       &nbsp;    You may be wondering why i\'m not enjoying myself here...its because ive lost one of my scriptures for this years teaching! NNNNNNNNOOOOOOOOOOOOOOOOOOOO!!!!!!!!!!!!      &nbsp;    Anyway i\'ll just have to teach you...     HOW TO VISIT MY BLOG MORE OFTEN AND HOW THAT IS RELATED TO RELIGIOUS EDUCATION??????????????????????????????????????????????????????????????????      &nbsp;    (Not that i know the&nbsp; answer myself...)...      &nbsp;    Now time to tell u who i am.      &nbsp;    I am a kid, my age is not 13. its 11. I got away because i really liked this site. It helps me advertise my holy scriptures...&nbsp;      My real name is... mabey i should not tell u.&nbsp;You may dob on the cops. (the palace of blogs!).      Anyway, I think that everyone is  &nbsp;and&nbsp;funky... (random statement)    &nbsp;    By the way!       &nbsp;     &nbsp;     (       &nbsp;) &nbsp;&nbsp;&nbsp; You must play a console or computer to be added on to my holy blog!&nbsp;&nbsp; &nbsp;        &nbsp;&nbsp;      &nbsp;     i got ta go!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!'),(10166,3819051,'01,July,2004','2004-07-01 00:00:00','    &nbsp;   &nbsp;   &nbsp;   :&gt;&lt;:Praise the Ones who play:&gt;&lt;:    Or you could just think of it as a Mexican Wave...&nbsp;   &nbsp;'),(10167,3819051,'01,July,2004','2004-07-01 00:00:00','    T  his page is an ashram (holy abode) of the:  ,,;!!Mystic Guru!!;,,  I hope you like it!   '),(10168,3460990,'31,May,2004','2004-05-31 00:00:00','   Back from my family trip. Everything there was pretty uneventful, which is the norm for the upstate relations. I left Saturday morning, expecting to get on the 9:51 train to Poughkeepsie. And I woulda made it, too, if it hadn\'t been for the amazing brilliance of the MTA. They decided to do construction on subway lines over the holiday weekend. So, halfway into the city, the train stops and the passengers are told they must disembark, as that train will be going no further. A door problem and a transfer later, I am at Grand Central. It is 10:15. I end up taking the 10:51 into Poughkeepsie with no further problems.   Grandma and Great-Aunt Freda pick me up at the station. What a pair, those two! Without getting too much into it, I\'ll just tell you that some people should not be enclosed in a moving vehicle together. Fights that ensued in the 24 hour time period I was with them include such great topics such as why one used so much of her maple syrup on her waffle and who felt dizziest when they stood up. Oy-vey!!  One upshot to hangin\' with the Old Ladies is that I got taken to the movies. We saw \"Raising Helen,\" which I can say that I was highly impressed with, despite the fact that I am not all that much of a Kate Hudson fan. Even (or should I say, \"especially\") with all the \"whaddya ya mean, move to Queens?\" comments, this particular bridge-and-tunnel girl was quite amused. (To top it off, she lives in a neighborhood that is close by, so seeing it would give you a representation of where I live.)  From there, I went to my cousin Dawn\'s and had much barbecue fare. Yum. Ended by seeing all 3 hours of the last Lord of the Rings movie. Have to say I think it woulda made more sense if I\'d seen the second one.   This morning I went with Dawn, her three kids and my Aunt Sandy to the Red Hook Memorial Day parade. Cute and picturesque. And mercifully short, even though I swear any kid in a 5 town radius who had some sort of uniform - be it Boy Scout, baseball, soccer, basketball, cheerleader or whatever - walked past me, smiling and waving.   About 30 seconds after the last float went past (the third nursery school class being pulled on a hayride) Aunt Sandy and I took off to start my return journey to the city. Because of parade detours, we ended up missing my first departure time, but then decided that hour would be better spent having cake and coffee rather than idly waiting on the platform. (Aunt Sandy has the most fantastic ideas, don\'t you agree?) A scrumptious piece of cheesecake later, we were on the way again.   We pulled into the station 5 minutes after the train had pulled out. Oh well, so I did end up sitting for an hour on the platform. But really, that turned out okay. I ended up talking with a young MTA electrician (John) for nearly 45 minutes. Verrry cute. I would say \"Ben Affleck\"-ish, even down to the adorable cleft chin. He\'s just hired a realtor to help him find a place in Astoria. I\'ll let ya\'ll know if he calls me...since I, for the first time ever, boldly gave him my number on the back of the library receipt stuck in my book. (Although, for the record, he approached me and started the conversation...)  I am now home, preparing for the workweek and giving my kittens extra attention from the days I was gone.   I did finish  The Life of Pi , which was really interesting at times, but I didn\'t find it quite lived up to all the hype. A lot of it is like reading a script you might find on Animal Planet. Very graphic literary images of animals being ripped apart. Eewww. Also a lot of description about the boat itself and where he is in it. I, personally, had a hard time picturing the situation, and I found all of the nautical description a little boring. But, it\'s a huge bestseller, so I leave it to you to decide whether or not to read it for yourselves.   Only 3 more weeks until I start summer training for NYCTF!  '),(10169,3460990,'28,May,2004','2004-05-28 00:00:00','   Got news today that is both exciting and a tremendous relief. About a month ago I spent an entire Saturday slogging through 2 different exams, the Content Specialty Test (CST) in English and the Liberal Arts and Sciences Test (LAST). Both are necessary for my teaching certification. Today, I got my results, and am happy to say I   PASSED BOTH!!!  In other day-to-day stuff, I spent some time at the library last night. I cannot even tell ya\'ll how much I loooove the  urlLink library . (Did you notice that? I\'m learning how to insert links already! Aren\'t you impressed?) Anyway, I finally got my mitts on a copy of  The Da Vinci Code . Wanna know if it was good? Does the fact that I started it last night around 9 and have already finished it answer that? Will I tell you the end? Of course not. Be assured, it is a fantastic read. Highly recommended! I came home with some more books from my time spent \"shopping\" at the library, and am planning on starting The  Life of Pi  tomorrow when I get on the train.   The train (Metro North) is a fantastic form of transportation. I love going on the train. The tracks are just alongside the Hudson River, so it\'s a very scenic ride. I am going to visit my family upstate in Poughkeepsie (for Texans, that\'s pronounced Puh-kip-see). From Grand Central Station it takes about 2 hours to get there. I\'m planning to stay until Monday morning, and since I\'m gonna be hangin\' with Grandma, I figured I needed a heavy artillery of reading material. Among other selections that will be stashed in my suitcase are Simone De Beavoir\'s  All Men Are Mortal  and the purposely-not-as-heavy chic lit  Good Grief . Reviews to follow. I would be really interested in having you comments on what you are reading now, or if you\'ve read one of these titles already, what you think.   In other news, today\'s my Mamma\'s birthday and tomorrow is Pantene Princess Jenny\'s special day. HAPPY BIRTHDAY!  So, you\'ll hear more from me after I return from what promises to be an interesting weekend of \"quality family time\".  '),(10170,3460990,'26,May,2004','2004-05-26 00:00:00','   Okay, so I\'ve actually been thinking about starting my own Blog for a while now, especially since so many of my friends are geographically far away. However, what really pushed me to finally do it was a conversation I had with a friend from Texas about two weeks ago. It made me realize that I am about to reinvent myself (again) and it would be an interesting thing to chronicle this journey from the beginning.   Soooo, what\'s a girl to do when she decides to move her life in an entirely new direction? That\'s right, get a new haircut! So, wanting to have my newly found direction benefit others, I went out last week and donated most of the length of my hair (over 12 inches) to a charity called Locks of Love, which makes and distributes wigs for children who have no hair due to illness. Pictures to be posted..eventually. So far, short hair is really nice. New York is just starting to get hot and sticky, and this is making it much more tolerable.  Shoutouts...congrats to AmyKatPat, who earned her Master\'s degree last week. You\'re awesome, and I\'m so proud of you! Holly, I am so glad you have free airline miles, and I\'ll see you in July!  In other life news...I find out the scores for my teacher certification exams on Friday and my university assignment through the Fellows program in roughly two weeks. And since the New York City Teaching Fellows program will be what I eat, sleep and breathe from this point forward, I have no other information to relay.  '),(10171,3460990,'28,June,2004','2004-06-28 00:00:00','   Went to a job fair tonight. Results were pretty good, I think. I got 3, count \'em, 3,  follow up interview offers. All 3 are jr. high, so I\'d be teaching 7th or 8th grade, but 2 of the 3 want to not only utilize my English degree, but also my Performance Studies. One school is an arts academy, the other needs a coach for their speech club. Whee! I am so excited by the idea that I could have a solid placement in the next few weeks!  Holly, I am really looking forward to your visit. Can\'t wait to show you around the neighborhood, and maybe, if things go well, my classroom!  '),(10172,3460990,'26,June,2004','2004-06-26 00:00:00','   School got much better Thursday. We had Friday off because many Fellows were signed up to take their exams first thing Saturday morning.  I went to the head of the college section of the Fellows program, and she gave me a little sticker that gave me permission to use the library if I should need to before our regular ID cards get distributed. She also pulled my Educational Psych prof aside and explained the whole ID/stipend/computer situation to him. He has now agreed to have a copy of our text put on reserve in the library and will be bringing copies of the handouts to us who don\'t have the ability to print out items without paying (like an internet cafe).  Sarah and I spent much of today walking 5th Avenue. She is looking for a part time job while she is waiting to be placed as a temp. We went all over, which was pretty fun.  Tomorrow, we are planning to go to church, then to pick up tickets to Shakespeare in the Park, where they are performing Much Ado About Nothing. Yay!  I also booked a ticket to come home next weekend. It was not the schedule I wanted, as I will be flying back out to NY early Monday morning, rather than late in the evening. However, such is expected with airlines in general, and I am hoping to do as much as possible in terms of dealing with what is happening to Mom while I am home.I wish that I would be able to have time to see all ya\'ll, but that doesn\'t look like a possibility. I do look forward to August, when I will be able to stay nearly a month, and hope that I\'ll be able to make plans with you then!  Mom seems to be doing all right, all things considered, and I am anxious to get home to spend time with her and try to prepare for what awaits us as a family.  Everyone has been wonderfully supportive, and I am very appreciative of that. I know that I have a huge challenge ahead, and I would like to ask a bigger favor of all of you. I have a tendency to \"crash and burn\" when I get depressed, and once my mom has gone, I am really afraid that I might do that. To try to avert that situation, my wish is that in the year after my mom passes away, that one person will come visit with me each month to help me keep from falling apart. I realize that this is a lot to ask, but if you\'re willing, please send me an email, so I can have that list handy when the time warrants.  Besides that, the kitties are great, and love Auntie Sarah, who will be kitty-sitting while I\'m away next weekend.   Much love to all!  '),(10173,3460990,'23,June,2004','2004-06-23 00:00:00','   Argh. I am so frustrated. I knew that doing the program was going to be a challenge, but some things are simply ridiculous!   The Fellows program is this bizarre network of institutions that claim to be working together to put us into classrooms in the fall. So, technically, as of now, I\'m a  non-matriculated  grad student of Queens College,(my matriculation status somehow magically changes in the fall) which is separate from the Fellows, which is then separate from the state and city of New York and their subsequent regulations and procedures.   Therefore, I have to ask UMHB for yet ANOTHER transcript, because the state needs it to process my request for certification, and the other 4 I\'ve already gotten were distributed into the Fellows never-never land of paperwork. Somehow, after several years of this program, they haven\'t added that to the pre-coursework part of the required papers asked of us.   My Educational Psych prof doesn\'t understand that we don\'t get our first stipend check until July, and therefore, it makes it difficult to pay for books. And that since we aren\'t matriculated, we don\'t yet qualify for a valid student ID from QC and therefore can\'t use the library or internet labs. (Somehow, the same ID  is  accepted to allow those who have cars to pay the $90 parking fee and get a permit.)  In other Fellows news, well, maybe not news, but whatever...as there are 2 other Lauras, a Lauren and a Lori in the same Fellows group, I have made a decision to go by Lily instead, and make it easier for all involved. So, I am on my way to having a completely new identity to grow into! Still getting used to seeing my name written that way on papers (sign in for attendance and contact sheets). But, very exciting! Yay!  My schedule looked something like this:  7:45-8:45   leave the house, then walk 15 minutes to subway, sit on train for 30 min, transfer to a bus, then sit another 15 minutes.  8:45-9:00  wander around the campus, looking for my first class (recalling days of my first few days at UMHB)  9:00-12:00  Class  12:00-1:00   Lunch, find the bookstore, calculate that books will cost approximately $150. Realize this is a problem, seeing as how I have $20 in my bank account.  1:00-4:00  Class, again, where we try to explain previous situation to prof, who tells us to get the money  from our husband   and get the books. Also try to explain the whole internet thing, as he is sending all handouts for the class through email. Was told to \"work around it.\" Argh.  4:00-6:30   Fellow Advisory. We try to pick this woman\'s brain for the real stuff we will need to know to be good teachers in the fall. Ask questions. Lots of questions.  6:30-7:30 Commute back to Astoria  7:30-now (?) Internet cafe. Venting. Needing to do homework before the whole process begins again tomorrow.  Thank God, we have Friday off!!  In totally unrelated news, I went to the AWED Gala event last night. I got there for the 10 last official minutes, so Rocio ushered me directly to the open bar and then gave me a plate and pushed me towards the finger foods. Got slightly drunk. Then Sarah showed up, and she, Rocio, the interns and I all went to Tropica, a bar inside Grand Central, and had another drink. (Just one more).  I really needed that.   Besides that, I fed my cats and was given an appreciative \"chirp\" from Lena and a somersault from Sydney. (They\'re so talented!) Love your pets!  '),(10174,3460990,'21,June,2004','2004-06-21 00:00:00','   Ah, my first \"official\" day as an NYC Teaching Fellow has come to a close. And I\'m exhausted already.   Lincoln Center was way cool, and it was overwhelming to see sooo many people that are now, essentially, in the same boat as I am. (Let\'s hope it doesn\'t turn out to be the Titanic, eh?)  To throw flowers at myself for a minute (to use one of my favorite newly-learned phrases) they let us know that we each were selected out of a pool of over 17,000 applicants. Approximately 2,000 have been chosen to go the distance and will begin school in September.   I met my cell group, which the program calls Fellow Advisory group, FA for short. There are about 15 of us who will all be attending Queens College and will be completing the coursework towards a Master\'s of Education in English together. Tomorrow morning we meet at campus to begin Master\'s work.   It\'s all a little scary, but also exciting to think about the new phase of my life that is beginning now. In my circle of friends, teaching is the most popular occupation, so I\'m looking forward to joining the ranks. On that note, congrats to Holly, who is going to be spearheading BJH\'s brand spanking new theatre program, and to AmyKatPat, who is making her way towards securing a position in an Austin-area district. I wish you both the best of luck on your new endeavors!  '),(10175,3460990,'20,June,2004','2004-06-20 00:00:00','   Well, it\'s been a crazy, busy weekend. Sarah arrived 2:30 yesterday afternoon, and we\'ve been going non-stop since then. I took her around the \'hood and showed her where everything was, and we stopped at a little Italian place for an early dinner. Then, we went into the city to see The Stepford Wives. (Pretty cute.)   Today, we got up and went to church. We stayed for both the main service, and then went on to the 20/30 Connection group to hang with some people our own age. There\'s a girl there that\'s currently a teaching Fellow. She just finished her Master\'s degree through the program, and it was great to talk to her, since she knows exactly what it\'s like to be where I am right now. She was wonderfully supportive and encouraging about starting this journey. Being back at church again was a good feeling, too. I was reminded that that\'s somewhere I really need to be on a regular basis.   After church, Sarah and I took a quick trip back home to change clothes and headed back into the city. We spent a few hours at Central Park (pictures to come later) and then went to have dinner at Serendipity 3. (Yes, the Serendipity from the movie.) I\'d tried to go there twice already, once with AmyKatPat, and once with my Aunt Liz and her husband, Kevin. Both times the wait times were at so ridiculous that we skipped it. I guess we were just lucky today. We were seated in less than 45 minutes. Anyway, it was so fun! The food was great, and there\'s a good reason they\'re famous for Frrrozen Hot Chocolate---it\'s amazing!   Hanging out with Sarah this weekend has really given me a new appreciation for the City, and I am anxious to begin exploring again. Also, she seems like she is going to be a great roommate, and someone that I think it will be easy to continue a friendship with even after she moves on to her own place.   We\'ve ended up here, at an internet cafe near the apartment.  As for a Mom, I talked to her for over an hour last night, and it seems she was having a really good day (considering everything, of course). She was in good spirits and seemed cheerful. Bonnie is trying to arrange a for a hospice interview for the three of us for the weekend I\'ll be there. I\'m waiting to find out exactly what time I\'ll be released on that Friday, so I can book a flight out. It\'ll be a quick weekend, but each day will bring me closer to August, when I can be home for a longer amount of time.   NYCTF starts tomorrow, with the welcome event, and official Master\'s classes starting Tuesday morning. There has been so much support for me doing this, for which I am so grateful. The people at AWED threw me the best farewell party, and gave me some generous and thoughtful gifts. I am looking forward to reporting back with my new experiences.  '),(10176,3460990,'18,June,2004','2004-06-18 00:00:00','   There\'s a saying in Christian circles in regards to praying for patience, which says that if you pray for patience, you\'d best be prepared for it to be tested. I think that the same theory relates to strength.  As difficult as things are right now, I think that God has been using the last few years to prepare me for this time. I\'ve often thought, as I\'m sure so many do, that I was being given more to deal with than was fair. However, I\'ve spent some time looking back over the toughest experiences in my history - finishing school while my father was ill, eventually losing him, deciding to follow my dream to fly, to have that hope dashed as well, to come home and deal with romantic heartbreak (a few times), and finally picking up and moving here, and the initial hardship associated with that - and I\'ve realized I was being shaped into a person who could deal with anything that might happen to me.   While I was going through all of my trials, I thought I was so unlucky! Everything was just so damn  hard . I kept wondering when God would just give me a break, already! The longer I\'ve been here, evaluating everything, I\'ve had a change of attitude and now see how amazingly fortunate I was.   My father was ill, but I got a long time to say goodbye and got the ultimate blessing to be able to really believe that his death, when it finally came, was a welcome release. I finished school, which if he had died more suddenly, I honestly don\'t think would have happened. I know I would have needed to quit altogther to stay with my mother.   I didn\'t get to have flying from exotic locale to exotic locate be my job, but I got to train through one of the most rigorous professional development programs in the country. And they paid for my ability to say, \"When I was Amsterdam...\"  Romantic heartbreak...I\'m still working on the recognizing the benefits of this one for me, but right now, I can say that there were times that he was the one I needed, and he was there. I learned a lot about the traits I need to see in the man I will choose, as well as what I won\'t put up with. I\'ve also made necessary changes in myself, finding that balance of keeping my own identity while contributing to his needs. In that process, I\'ve become someone that will, eventually, make a amazing wife to the right person.   New York. If I can make it here, I can make it anywhere, right? After months of thinking that the City was just going to chew me up and spit me out again, we\'ve finally become friends. She has welcomed me on her sidewalks and held me close between towering buildings. Pretty soon, I\'ll start helping her grow by contributing to the young minds that reside here. And I know that in her way, she\'ll make sure I\'m rewarded for that.   When the time comes, I know that God will give me what I need to work through whatever happens. He\'s been training me pretty exensively up to this point, even though I didn\'t realize it as it was taking place.  '),(10177,3460990,'17,June,2004','2004-06-17 00:00:00','   I called Mom again last night. She\'s on really heavy drugs, so she\'s sleeping a lot. Good for her to get rest, scary for me when I call her 10 times and get no answer.   I finally spoke to her, and she was a little loopy, but is slowly admitting things to me. I asked her about her doctor\'s appointment, and she simply said that she was put back on treatment. When I asked about the results of her scan she said, \"Well, it\'s not good,\" but that was all she would tell me.   We talked a little about me, since yesterday was my farewell party at work, and she asked me about everything going on with the Fellows. That was really nice for me, because up until she was given the new pain meds, she hasn\'t been able to talk to me for more than about 5 minutes at a time. She would ask me questions, but wouldn\'t be comfortable enough physically to hold the phone up and focus on the conversation. And that would upset her.   I did talk to her nurse the other day, and she told me that Mom had seemed anxious about me giving up the Fellows to come home, and that she needed to know that I would still be dedicated to something to help me get through the aftermath of what\'s going to happen. I can tell that Mom is really concerned with me being okay later on, and she believes that the best way to be secure is to go on with my Master\'s and start my new job. So, after hearing that, from the nurse and Mom, I think best course of action is for me to continue working here, try to come home as often as I can in the next few months, talk to her on the phone every day,hope that we get to spend one last holiday together, and if we do, make sure it\'s as special as possible.  I told Mom that I\'m trying really hard to find a way to come home for the long July 4 weekend, and she seemed excited about that. This way, she\'ll know that I was planning a trip then anyway, and she won\'t feel that I\'m sacrificing my future to spend time with her now. I\'m also planning to come back for the few weeks I have in August and early September before I start teaching in the fall.   I can tell she\'s trying to ease me into all of this, as last night she told me I should start thinking about anything of hers that I would like to have, \"just in case.\" She knows there\'s a lot to deal with in terms of the estate, (as some of you know, my father\'s estate still has not gone through probate yet,) and that she\'s going to have to tell me the whole truth pretty soon, but for right now, baby steps.   Thank ya\'ll for being so supportive. I really appreciate the messages I\'ve been getting. Right now, things are okay, but I\'m sure that later on, I\'m going to take you up on your offers.  '),(10178,3460990,'15,June,2004','2004-06-15 00:00:00','   My sister in law, Bonnie, called me last night. Things are not going well with my Mom. She had another doctor\'s appointment yesterday, and was told the cancer has not only returned, but spread.   She\'s going to have another round of chemo, but apparently that is only to prolong the amount of time she has left, which they are estimating to be around 6 months. She has been advised to settle her affairs and after this last set of chemo, will enter home hospice.   At this point, I haven\'t spoken to my Mom, partly because Bonnie and I were on the phone until around 4 this morning. The other reason is because my Mom asked Bonnie not to tell me. She doesn\'t want me to give up the Fellows program to come home and stay with her. (Of course, being my mother, she sees that as her decision, not mine!) I\'ve left a message with Mom\'s nurse, and am waiting for her to call me with the medical issues. So, I\'m going to call Mom, like normal, tonight, and see what she says and give her the chance to tell me in her own way.   I know there are only a few of you reading this right now, but this was the best way I could think of to put out a general cry for the help I\'m going to need in the upcoming months as I try to make some hard decisions about what is best for myself and my family.   Please pray for us.  '),(10179,3460990,'14,June,2004','2004-06-14 00:00:00','   When I turned on my computer at work this morning and opened Internet Explorer, it went directly, as always, to the MSN page.   One of their front page articles was entitled, \"Flip Flops at the Office!?\" with the subtitle. \"What NOT to wear to work.\"  I think this is funny because I am, right now, blogging in flip-flops.   Hahahahaha. I have 4 days left. Whaydda they gonna do, fire me?  '),(10180,3460990,'13,June,2004','2004-06-13 00:00:00','   So, I added a counter to the blog...now I can see just how many people are checking in on me. Thanks to that one person! (If that number doesn\'t bump up, don\'t worry, whoever you are. I won\'t quit writing, I\'ll just delete the counter. Hahaha.)  My head is better now. I went out to dinner with Matt, who is leaving for Italy at the end of this week to be with his pregnant wife (she is stationed there for her work in the Navy.) I guess cheesecake works better than Excedrin.  My roommate is essentially gone. She left one more trip\'s worth of stuff, but she did go ahead and turn in her keys. Hopefully, with her things now being behind locked doors she will be more persuaded to contribute towards the $600 she currently owes me.   I\'ve been having kind of a nostalgic day, in the hours since my headache dissipated, anyway. Walking around the neighborhood reminded me of a walk I took with someone else, a long time ago, and then a CD I listened to reminded me of them as well. Even though we cared about each other, there are good reasons we\'re not involved in each other\'s lives anymore...the relationship didn\'t give either of us what we really needed. But for whatever reason, that knowledge doesn\'t make me immune from missing them, as much as I wish it did. I know it\'s silly, but there\'s this desire to know that they\'re missing me, too. (This would be the part on Sex and the City where Carrie would see whoever she\'s talking about on a date with a new woman. Hahaha.)   There was another time that we had separated, and I kept hearing \"our\" song all over the place. Everywhere. In elevators as muzak, in stores over the PA system, on the radio, I mean  everywhere . I remember thinking that if I was hearing it so much and thinking of him, he must be hearing it and thinking of me, too. That actually made me feel better. When we finally did get back together and I told him about it, assuming his response would be parallel to my own, he said, \"What? We have a  song ?\" Ugh. Frustration. Anyway, I\'m annoying myself that I\'m even still thinking about this. (It would help if that cute Ben Affleck look-alike from the train would call and distract me!)  Oh well, until then, tomorrow\'s a new day, and I can fill it with a walk heading somewhere else and a different CD.  '),(10181,3460990,'13,June,2004','2004-06-13 00:00:00','   I had a great first half of the weekend, spending it hanging out with one of my best friends here in NYC. We went out for Chipotle, as it is now our tradition to spend Fridays in the company of burritos and beer, and then continued on to my apartment. There, we spent the rest of the evening watching Mystic River. Saturday we got my fixed up my house and then watched Mona Lisa Smile and Bend it Like Beckham. Overall, it was a fun and productive time.  Today I woke up with a headache that is progressing rapidly into a migraine, despite 2 Excedrin and a nap snuggling with a kitty. Ugh.   My current roommate is using the day to move out, which, in her usual fashion, is taking place the most unnecessarily difficult way possible. Rather than just calling a car service and loading the car and taking one trip to the apartment of whatever other person she has duped into thinking her was a normal, responsible adult, she has recruited a friend, and together they are transporting her things  one duffel bag at a time  using the subway. This experience epitomizes my entire relationship with her. Nothing is ever easy, even when it could be.   Only one more week of work before I start the Fellows training. Look for my university assignment to be revealed soon!  '),(10182,3460990,'09,June,2004','2004-06-09 00:00:00','   urlLink    The Sky Club is on top of the MetLife Building  '),(10183,3460990,'05,June,2004','2004-06-05 00:00:00','   I just got an email from the Fellows program, informing me that our first \"official\" day of the program will be a welcome event, hosted by the  urlLink Alice Tully Hall  at Lincoln Center. I\'ve never been there, but it looks kinda swanky to me! However, work begins pretty quickly. The first few hours are a \"glad you\'re here\" thing, but the last 2 hours are set aside for consultation with our program mentors and to meet the other 30 or so other Fellows in our cell groups. (The entire cohort for the year of 2004 is expected to be around 2,000 Fellows, but we\'re further divvied up into smaller groups that will be taking university classes together throughout our Master\'s training.)  And as the days of the Fellowship beginning grow closer, so does the end of my time with AWED. In preparation for my departure, I have been working my butt off! I want to leave with a good impression, and so have been cleaning up all of the files on my computer, re-organizing the paper files I\'m in charge of, teaching the 3 interns how to do my job, and writing a manual to help them out once I\'m gone. Even though I\'ve really been pushing myself, it\'s been a good week. It\'s nice to have so many people in the office again. (Over the last few months there has been, ahem, \"restructuring\" and our already small staff of about 10 was reduced to 6 full time people.) The AWED Annual Gala is approaching, too, and I\'m hoping that I\'m going to be able to scoot over to the Sky Club (see photo) after my day at whatever university I get assigned to. (That, I\'m sure, will be posted here as soon as I find out!)  I\'m also trying to learn more about how to make this page more interesting for those of you who are attempting to live vicariously thru me (hello, Stephen Greene!) So, as you can see, now that I can insert links, I\'m teaching myself how to post photos. I\'m also perusing other Blogs to see how I can do things that I see that I think are nifty. Saw one today that had recorded voice audio.  Way cool . More updates on that as I delve more into this area of computer savvy that is currently unknown, yet intriguing,  to me.  '),(10184,3460990,'05,June,2004','2004-06-05 00:00:00','   So, I went to my school visit on Thursday. Unfortunately, my host Fellow wasn\'t exactly \"all that\". He teaches \"ramp-up classes,\" which is just a way to un-label low performing kids with a  different  label. Out of his 90 minute class, 45 minutes was basically free time, although he had asked them to work on their final writing project of the semester. At one point, he walked over to me and the other girl observing and said, \"Well, 90 minutes is a long time. I just don\'t know what else to do with them,\" while shrugging his shoulders. So, I was definately inspired by my classroom visit. Now I see the kind of teacher I don\'t want to be, and the residual influences that some of these kids will be bringing with them into my classroom. Let the battle begin.  Besides that, everything has been quiet. I\'m working on getting organized for the beginning of my Fellows training, knowing that I\'m going to have to be up and out the door at least an hour earlier than I\'m currently used to. To be honest, I am so excited to start school, both as a teacher  and  a student.   I finished two more books in the last few days. One being  Good Grief , by Lolly Winston, which was exrememly well-written. It\'s about a woman in her mid-thirties who is struggling to heal after the death of her husband. Of course, parts of it are sad, but there\'s also a lot of realistically funny and touching situations as well. If you\'ve lost someone you loved, you\'ll easily relate to the main character\'s journey. If you\'ve ever known someone going through mourning, this will give you a glimpse into their life, and perhaps show you things you can do to help at a time when many people feel paralyzed-wanting to find a way to show support yet not be intrusive. There are not many resources out there to help people deal with the loss of someone important. It\'s not like anyone really  wants  to write \"Funerals for Dummies.\" And because of that, we\'re left to figure things out on our own. For the first few weeks after my dad died, the attention was overwhelming. It wasn\'t that we didn\'t appreciate it-it\'s just that so soon afterwards, you\'re still in shock about the whole thing and you manage to go through all the motions in sort of a haze. What most people don\'t understand is that it\'s not until later, after the casserole dishes have all been returned and the \"just calling to check on you\" calls have stopped, that you really need that contact with others. Anyway, point being, this is a good book to read-there\'s something about it that simply resonates with feeling.  The other title I read was  A Man Most Worthy . All I have to say about it is that I never realized that the word \"ass\" could be used so many times as an adjective. I won\'t elaborate further on that. (It\'s an \"Essence\" book, but I didn\'t realize that when I picked it off the library shelf, and when I did notice it on the cover, I didn\'t think it would that much different. I mean, cheesy romantic is essentially cheesy romance, right? Hhhhmmm. I think I\'ll be more cautious in the future.)  After I finish  All Men Are Mortal  I plan on starting reading from a list of the top 100 recommended reads for students entering college. One book at a time...  '),(10185,3460990,'02,June,2004','2004-06-02 00:00:00','   Listening to Ace of Base whilst I rapidly prepare for Spring Visit AND an impending new roomie (hiding all my dark secrets, hahaha).   Today was a crazy day at work. For the next 11 workdays, I\'ll still be sitting behind the front desk at  urlLink AWED . We just got 3 new interns, who seem quite motivated and intelligent. Therefore, we aren\'t exactly sure what to do with them (one only needs so many copies made). When I leave, they\'ll be covering my desk in shifts while work figures out what to do. So, I\'m trying to do all my regular work stuff, plus train the interns while attempting to make a \"Bible\" for the desk, so whoever sits there eventually will be able to cope. I was soo glad when the clock read 5:30.  On the way home, I stopped by Petite Sophisticate (read:\"mature\" clothing) and got a \"don\'t I look professional?\" black bag that\'s big enough to hold a file folder stuffed with all the Fellows documentation I need to tote around to interviews. Okay, so I couldn\'t resist...I also got a teacher outfit. Black skirt, lavender twin set. Not flashy, but it really makes me look like a grown up, especially with my hair cut. (On that subject, a student who came into my work told me it made me look \"like a lady\". Not sure how exactly to take that.)  My home life is still in flux, with the roommate still avoiding me. However, the kittens are, as always, a joy. Sydney is bouncing around in the newly aquired shopping bag and appears to be happy as can be. (Ah, simple pleasures!) Lena is watching, but I can tell anym minute now she\'ll be fast asleep on the table. As I was reading  The Life of Pi , one word really made me smile. The boy, Pi, came upon an island, inhabited by not just meerkats, but meer kittens ! Meerkittens! What a great word.  '),(10186,3460990,'01,June,2004','2004-06-01 00:00:00','   I\'ve spent the last two evenings just...me.   Last night I watched \"Red , White and Blonde,\" a girl-power movie if there ever was one. It\'s definately one I plan to add to my collection for those inevitable nights when I come home emotionally bruised from a tough day at school. There\'s just something about the way Elle Woods views the world that makes me believe that a pure heart can overcome any obstacle. Totally corny? Yes, I suppose so. But it works. (Just try to do the \"bend and snap\" alone in your living room without it putting huge smile on your face. C\'mon, ya\'ll know you\'ve done it, admit it!)  And uplifting movies have been my resounding choice of entertainment lately. Tonight was \"Bruce Almighty,\" a really fabulous movie. Whenever I ever have one of those, \"why-am-I-still-single-getting-closer-to-30-maybe-I-should-try-online-dating-again\" thoughts, the way this movie ends reminds me of what I\'m really waiting for and trusting in: that one day, when I least expect it, I will meet someone who will love me and treat me the way that all of us deserve...as the special and beautiful creations that we are. For those of you girls out there who see the end of your twenties rapidly approaching....SO WHAT!?! Don\'t settle for less than you really deserve. As some of you probably know, I know whereof I speak...  Okay, off the soapbox. Besides that, last night was eventful in another way. I finally, after literally stressing out about it for  months , had a talk with my roommate and it is offical:in two weeks time, she will be an ex-roommate. (Quick update for those not up to speed on my living situation. I decided to save some money by moving into the living room and subletting my bedroom. It sounds nuts, I know, but it\'s not such a bizarre thing here in NYC. Besides, it\'s just for a few months, until the teaching paychecks start rolling in.) Anyway, out with the old and in the with new. That\'s right, folks, I am giving the roommate game one more shot. Thanks to the Jennifer Lynns and AmyKatPat for showing me such that with the right people, a roommate situation can turn into a lasting friendship. (Hey, Conner, remember the Batman posters? Ha!)   This time the lucky winner is a girl from Ohio who actually has the money for rent saved in advance and only wants to stay a month or two while she finds her own place. Ah, the days of sleeping on my couch are coming to a close. There is truly a light at the end of my tunnel of financial suffering.   As for teaching program  urlLink NYCTF , the days are rapidly approaching towards summer training. I should get my university assignment next week! I had some schools email me about open positions today and will be sitting down to make some big decisions in the next week or so. Besides that, I\'m also scheduled to attend my Spring Visit on Thursday. Spring Visit is like a Big Brothers/Big Sisters program for the Fellows. I was assigned to a current Fellow teaching in my region, and I get to go follow him around for a \"day in the life\" of a HS English teacher. I can\'t wait!  '),(10187,3460990,'24,July,2004','2004-07-24 00:00:00','   Things are definitely better now than last week. Holly is here, and I\'m in the process of finishing up some schoolwork so I can spend some time playing in the city.   Next week is the last week of university classes for my program, which means that everything is due all at once. It\'s crazy, but once it\'s over, that last week will be a piece of cake!  Starting on Monday, I have a final, a take-home final, a paper, and a group project that all has to be turned in before 4:30 on Thursday. After that, it\'s just field trips and more instruction on how to instruct.   We\'ll be graduating from training on August 6, and the ceremony will be on the world-famous Intrepid aircraft carrier. First thing Saturday, August 7, I\'m on a plane home.   I\'ll be in town long enough to get to spend time with anyone who\'s free. I am so excited to be coming back to Texas!  &nbsp;  '),(10188,3460990,'19,July,2004','2004-07-19 00:00:00','   Okay, so things are going. They aren\'t necessarily going well, but they\'re still going. &nbsp; Honestly, this doesn\'t even feel like it\'s happening to me.&nbsp; One minute I\'m staring into space, the next I\'m bawling. Then nothing.  &nbsp; I don\'t know what to say at this point. There are things that neither of my parents will never see in my life, and that makes me angry, and sad. And really, in a way, thinking about that helps me to deal with everything that\'s going on. I guess I figure if I can list out all the important events that I should know they won\'t be attending, then that fact won\'t be able to sneak up on me later.  &nbsp; My first day of school as a teacher, my Master\'s graduation, my wedding. I never imagined that neither of my parents would be there.  &nbsp; There\'s that saying about what doesn\'t kill us makes us stronger, and I believe that. They just never tell you that sometimes that one thing that doesn\'t kill you might try to kill you again at every important event in your life from then on.  '),(10189,3460990,'18,July,2004','2004-07-18 00:00:00','   My mom passed away a few hours ago.&nbsp; I\'m currently waiting for my Aunt Liz to call me back to start the family phone tree, so&nbsp;I decided that while I waited, I would email my Fellowship advisor just to let her know that I wouldn\'t be at school tomorrow. &nbsp; We are planning to have a small gathering at the house when I come home in August, and I will be posting more details once they are decided.  &nbsp; Thank you all for your support during this time, you all have definitely been a source of comfort for me.  &nbsp; Unfortunately, now is when the really hard work begins.&nbsp;  '),(10190,3460990,'15,July,2004','2004-07-15 00:00:00','   Last night, Sarah and I went down to the lower East side and had dinner at the famous Katz\'s deli. It is absolutely nothing like the one in Austin...it is pure New York. The counter guys are all yelling and it looks like it hasn\'t had a really good cleaning in a decade. However, the pastrami and rye and piece of cheesecake I had were truly the best ever.  &nbsp; The walls are covered with photos of famous customers, and there\'s even a sign hanging from the ceiling to show you  exactly  where Harry met Sally (Or more accurately, the table where Sally demonstrated faking an orgasm.) &nbsp; &nbsp;All in all, it was a great experience. (Also, it rates on Citysearch\'s \"10 Things You Should Do in NY Before You Die. And on that list, I now only have 8 to go, since I\'ve already been to the Empire State Building!) &nbsp; Besides that, it has been a calm weekend, which was especially needed right now. I still have homework before I go to class tomorrow, but it was so relaxing to just not do anything yesterday. &nbsp; Holly, I am really looking forward to next week-I\'ve been saving ideas for things for us to do! &nbsp; &nbsp;  '),(10191,3460990,'15,July,2004','2004-07-15 00:00:00','   On a very short note, I went to summer school today and was given the opportunity to do the \"read-aloud.\" As I was reading, my Fellowship Advisor came in to observe my classroom skills. &nbsp; As I read, I asked the kids small questions about what certain words meant, what they thought would happen, or why characters did or said the things they did.  &nbsp; Later today, at our group Fellowship session, we brought up the subject of what the evaluations she would be giving us actually meant in terms of the program, or us becoming teachers. To reassure us, she said that she just wanted to peek in and see how we were relating to the children, and that she would be able to determine that, even if we were only sitting in a chair in the back of the class observing. Then, she said that she feels that some people are just born with an inherent \"teacherness\" about them. Not that teacherly traits can\'t be learned, just that if you were someone who needed to learn them, she wanted us to know and then help us accomplish that. &nbsp; At the end of the session, she went over the observations she had from today\'s summer school classes. She reviewed the things I did, and then exclaimed, \"You know what? You\'re teacherly!\" &nbsp; &nbsp;  '),(10192,3460990,'14,July,2004','2004-07-14 00:00:00','   Was another tough day in the world of alternative education certification. I\'m tired all the time, and frankly, three weeks seems like years away at this point.  However, I did get to do my first read-aloud session, and I actually, believe it or not,  sang  in front of my class!We\'re reading a book about a musician, and there were a few lines from the \"Happy\" song (you may have heard it on the Clinque commercials...\"I want to be happy, but I can\'t be happy, unless you\'re ha-ap-py too!\") Anyway, for the first time since I started going there, they were all attentive and most were even smiling! I was in such a shock! Looking forward to more \"happy\" moments!   Bonnie contacted some lawyers regarding everything happening with my mother, and apparently no one she spoke with wants to take the case. So, we\'re moving on to self-made will and distribution. Grreaaat. That really sucks, but what else can I do?  As for my blast from the past, I got an email today from classmates, and it seems an ex boyfriend from college wants to get in touch. Mmmm, I\'m curious, because we really did not end on good terms. If nothing else, it could be really interesting, since I\'m so, so over him and everything that happened between us. So, I gave him my email, just to see. Ha!  '),(10193,3460990,'12,July,2004','2004-07-12 00:00:00','   Last night I pulled the closest thing to an all-niter than I had in years. That paper I\'d been stressing over was given an extension, but of course, I didn\'t do it. I finally finished last night at nearly 4 am. Then, I woke up at 7 and went to school. It\'s now about 7 pm, and I\'m in the school\'s library. Ugh. I need sleep...  Summer school was better today. They reduced our class size, so I don\'t think we had more than 20 kids in any of my classes. However, my co-op teacher still isn\'t happy, since she\'s being forced to teach science and social studies for her 3rd period.   I had been getting kind of anxious about not having actually gotten the chance to practice my teaching skills, but now I have a different anxiety: my first \"real\" lesson will be in social studies. I am really lucky in that another Fellow in my school is in the same situation, and we\'re going to try to work together. It\'s kind of crazy...me, teaching social studies! (She wanted to give me science, but I said no thanks!)  Besides that, things are okay, I guess. The kittens are good, and they make me happy. Which reminds me, that I need to make them happy and go get them some kitty kibble. Gotta run!  '),(10194,3460990,'11,July,2004','2004-07-11 00:00:00','   My aunt Liz went to see Mom over the weekend. She called today before she left to go back to Reno, and gave me an update.  It was really good to know that the rest of the family is now becoming aware of the situation, and it\'s not quite so overwhelming. Aunt Liz was the executrix of my grandmother\'s estate, so she knows a lot more about all of the legal stuff than I do.   This has been the first weekend in a long time that I was able to get some real rest, and I think next week with summer school and university will go smoother.   Yesterday, Sarah and I went exploring down in Chinatown and got a couple little things and had lunch. Overall, it was a really nice day, and I got to relax, knowing that Aunt Liz was with Mom. (Not that I don\'t trust Bonnie, but Aunt Liz is a nurse, so there\'s an added level of confidence in that.)  Holly will be in NYC on the 22nd, and I am very excited at having her here! As another first year teacher, we will have tons to talk about! Can\'t wait to cram in as much fun stuff into that weekend as possible!  '),(10195,3460990,'09,July,2004','2004-07-09 00:00:00','   Last night, around 11 pm, I opened my email to find an invitation to interview at 10 this morning with one of the three follow-up schools.   After a long wait and a short interview with a very assertive principal, I got the job!  I\'ll be teaching English to either 7th or 8th grade at K291 in Bushwick, Brooklyn.I\'m still researching the school to be as prepared as possible for September, but what I do know is that they have a stong emphasis on the arts, and that appeals to me a lot.   Hopefully, I\'ll be able to get involved with some aspect of extra-curricular activities before school even starts. They have a drama group, dance team, and band. I\'m very excited to get started.  '),(10196,3460990,'07,July,2004','2004-07-07 00:00:00','   Today was my second day of summer school with children, and it was chaos. I knew that kids in summer school generally need extra help, and also might be behavior problems, but honestly, I was not prepared for what I experienced today. There was yelling, screaming, fighting and so much more. My mentor teacher literally threatened to quit today. I won\'t go into it much more, because tomorrow I have a midterm in one class and 10 page paper due in the other.    Will update more soon, probably this weekend, after I get everything done and have had some sleep.  '),(10197,3460990,'05,July,2004','2004-07-05 00:00:00','   Got back into LaGuardia (LGA) today. Very, very tired. Still have homework to do, and I start summer school, with actual children, tomorrow morning. Thank God tomorrow is just observation!  Am planning to return to Tx in August as soon as my university stuff is done for the summer. I will start teacher inservice, assuming one of my 3 as-of-yet unscheduled follow-ups go well and I have a place to go, on Sept 7. Kids arrive Sept 13. Wow.  Was very glad to see kittens, as they are quite furry and cute and make the world seem like a better place when they curl up next to me purring. They were obviously well taken care of by Aunt Sarah, who revealed to me that Sydney has honored her by sleeping on her lap over the weekend. Awwww.  '),(10198,3460990,'04,July,2004','2004-07-04 00:00:00','   Hospice came today to make assessments of our situation. She is definitely qualified for home care, which is both good and bad. Good that we will have help, bad that she is already at the point where it appears she will need an aid 3 times a week, and a managing case nurse 2 times a week.  This is going to be a difficult road. I can already see that in the time I have left this trip not everything will be accomplished that I would like.   However, there are things that I will get to do that are much more meaningful and  supercede the things I would have liked to have gotten done. I really am lucky in that we will get to chose some things that she would especially like to do while she has the ability to do them.  Tomorrow, I will be going with Mom to see the fireworks display. How do you not cry when your Mom says she\'d at least like to see fireworks one last time? By knowing that if she sees you cry, she\'ll cry, which is worse than almost anything else in the world.   Again, I wish I could curl up with someone nice for a while and just let myself melt into them. I know I need to count on myself, because there are things that have to be done that no one else can do for me: I have to write my papers for school, I have to clean another part of the house tomorrow so hospice can work while I\'m gone, and I have to start sorting through Mom\'s current expenses so I know what we\'re dealing with, along with probably a hundred other tasks that are waiting. However, this all would be easier if I had someone in my life that would wrap their arms around me and help make it disappear, if only for a few minutes. Unfortunately, God has not chosen to bless me with that gift. I suppose it\'s selfish, but I can\'t help but wonder why.  '),(10199,3460990,'02,July,2004','2004-07-02 00:00:00','   I\'m here. Tired beyond belief, but here.   My flight, which I left for at 5 am this morning, was delayed by an hour. Then, with the miracle of air travel, we \"made up\" the time in the air. That way, I made my connection and landed in Austin around noon.  My Mom looks worse than I had expected. She\'s lost a lot of weight, and can\'t really get out of bed. Stephen, of course, is the biggest jerk in the world and is making every process harder than it needs to be. Bonnie was going to take the kids over to her parents\' tonight, but has reconsidered and is going to take them in the morning. Shelby has been a terror all day-to the point where I locked her outside to get her to stop screaming her head off.   I don\'t know how I\'m going to get through this. I know that I  can , I just don\'t know  how . Sometimes I wish I could curl up with someone nice for a while and just let myself melt into them.  '),(10200,3460990,'09,August,2004','2004-08-09 00:00:00','   urlLink    Mom and me  urlLink  '),(10201,3460990,'08,August,2004','2004-08-08 00:00:00','   For as long as I can remember, people have been telling me I look just like my mother. However, if I was actually  with  my mother, she would tell me later, \"They don\'t know what they\'re talking about. You actually look just like your father.\"  And I believed her. I mean, she IS my mom, after all.  Then, I found this:  urlLink    My mom, approx 1977  urlLink     urlLink   urlLink   And, the picture of my mom putting a scarf on my head...so funny, because I actually wore that yellow and blue shirt (mom made it herself in Home Economics!) while I was in college. Haha!  Doncha just love nostalgia?  '),(10202,3460990,'08,August,2004','2004-08-08 00:00:00','   Um, yeah, so my \"do-it-myself\" plane connection in Dallas didn\'t go so well. I made a big mistake by arranging my transfer through Love Field instead of DFW, where I actually touched down. This lead to being stuck at DFW for about 3 and a half hours, followed by a road trip with my sister-in-law and both kids from Dallas into Austin. Yay.  Oh well, I\'m actually here, so I guess that\'s good. I started work on Mom\'s house this morning, trying to get the bedrooms empty so we can go ahead and put the kids\' furniture in there so they can stop having to sleep on the couch.  I worked for several hours, then went out. First I stopped at Starbucks, which was an extremely well-needed experience, then continued on to Olive Garden for a glass of Sangria and some good conversation with my friend AmyKatPat.  Tomorrow is yet another day, and there is lots left to be accomplished!  '),(10203,3460990,'07,August,2004','2004-08-07 00:00:00','   on a jet plane. (Haha) Just did my online check-in for Air Tran, and will be on my way home shortly. See ya\'ll soon, if my self-arranged flight transfer between airlines in Dallas goes smoothly.  '),(10204,3460990,'01,August,2004','2004-08-01 00:00:00','   Only 5 more days of going to Fellows stuff for the rest of the summer. University coursework is technically done, but with my family situation, my two professors gave me extensions to finish up the huge workload due for the end of semester. So, I\'m spending the rest of this weekend getting together my papers that need to be turned in and studying for my Ed Psych final.  Last weekend with Holly was a ton of well-needed fun. We went to see 3, yes, 3, broadway shows. After she arrived Thursday night, we went to Brick Cafe, where we spent, um, I think, somewhere between 3 and 4 hours catching up.  Friday, she went to school with me at Queens College. What a brave soul! That night for dinner, we went to the Olive Garden in Times Square. It was absolutely pouring down rain, and we all got soaked (although I got it the worst, being a dummy and thinking that when the rain had quit earlier that afternoon, it was really over, and therefore choosing not to take an umbrella). Still, the Olive Garden is always good, even when you\'re wet. And even being wet doesn\'t seem too bad after a Long Island iced tea and a Cosmo. Hahaha. (Holly, I still owe you a drink, girl!)  Saturday night we saw \"Wonderful Town\" with the amazing Donna Murphy. We actually went and waited by the stage door for her to come out, but after what felt like forever, we decided to press on and do other things. Like go to the Virgin Megastore and buy lots of music. Yay!  Sunday, we went to the tkts booth again, and got seats for the matinee performance of \"Bombay Dreams.\" Oh-the costumes, the dancing! It was fantastic! From there, we stopped quickly at Roy Rogers and got a quick bite, and then met Sarah out front of the famous Winter Garden theater, where we saw \"Momma Mia!\" That was truly spectacular! Although lots of people sang along through the whole show, at the end, I don\'t think there was a single person not dancing in the aisles! And it was special because it was Sarah\'s first Broadway show  ever . Afterwards, we went to Ellen\'s Stardust Diner, where the servers take turns performing. Turns out that our waiter was actually from Sarah\'s hometown in Ohio, and was only a year or two older than her. Only in NY!  In case you\'re curious, the ex-boyfriend is still calling me. It seems he is determined to get himself \"back in the picture,\" but he has, um, shall we say, some \"issues.\" The most I have agreed to at this point is to let him take me out to dinner once while I\'m in Texas. He had already been planning to go to see his family, but after talking with me on the phone, put off his intended schedule to wait until I would be there. So, no pressure or anything there, I just thought that just about anyone who has the guts to call after 6 years should be given at least a chance.  Anyway, I am very glad to have things wrapping up here and am ready for a break from school. I\'ll see ya\'ll soon!  '),(10205,4074186,'29,July,2004','2004-07-29 00:00:00','   I\'m the Electric Hellbilly, 665th Transgalactic Hyperbike Champion, Junior Division. These are my stories. They are inane and useless, full of mind-numbing niggling and general buffoonery. Enjoy them at your own risk.  The Electric Hellbilly '),(10206,4074186,'29,July,2004','2004-07-29 00:00:00','   Peace, pot, and microdot... Coming at you from our whiskey saloon here on Phobos, moon of Mars. Hide yer daughters, \'cuz we\'re fuelin\' up our hogs and we\'re runnin\' outta booze. '),(10207,4074186,'13,August,2004','2004-08-13 00:00:00','   My friend Ronda came in late to work yesterday. She was out jogging when a guy riding in the passenger side of an old pick-up truck threw his cup of McDonald\'s parfait at her. She freaked out at the first sensation of impact, and dropped to the ground as if she\'d been shot. She had no clue what hit her until she saw the fruit stuck to her. The truck driver and the parfait-thrower kept on going, but someone at the intersection who saw what happened wrote down their license plate number. So Ronda called the cops. They caught the guy. He had to write her an \"official letter of apology\" (what makes it \"official\"?), is being charged with assault, and has a court date set sometime next month. Ronda\'s pressing charges.  Ever seen Mrs. Doubtfire? \"It was a drive-by fruiting!\"  -----------------------------------------  My sister starts teaching again on August 19. That means I have the afternoons to myself.  I need a new routine. I want to start writing again. I mean,  really  writing again. I haven\'t finished a story in over 7 years. My last major project was a screenplay I started my senior year in high school called  El Bandido . It was a blatant  Desperado  knock-off (\"homage, thank you\"). Lots of cussing and gun-fights. The first scene took place in a seedy Mexican dive bar called  El Mundo Cerveza . I got about 25 pages into it before getting side-tracked with something else...  Ever since then, I\'ve tried to start new projects. A burgeoning space opera epic, a couple of hard-boiled Batman stories... At one point I remembered writing about me, and my life, and my friends\' lives... This was during school, of course, when I saw them almost every day... But now, 5 years later, it\'s just me. Well, me, my sister, and my little 2-year-old niece. All of my friends are grown-up, married, moved on. They have careers and kids. They\'ve begun new lives, and I\'m still here, stuck dwaddling in the past like some heart-sick nostalgic fool. I can\'t write about them anymore. Our adventures are over. So now I have to invent them. And heroes make inventing stories so much easier.  Well, dammit, the days of me not finishing anything are  OVER . I\'m starting a new schedule. Every Monday thru Friday, starting on the 19th, I\'m going to wake up at 9:00 AM, sit down in front of the computer upstairs, and WRITE -- WRITE, WRITE, WRITE -- for 5 hours. Non-stop. No breaks. Just writing. And I\'m setting a goal for myself. Say, a 10 page short story. I\'m not sure how many words that would be, but I\'ll figure that out after I write my first short story. And no sitting in front of a blank moniter for 2 hours, thinking of something to say. I\'m just going to type. Let my fingers move. I\'m sure 99% of what ends up on screen is going to be crap. But that\'s the beauty of editing. You can turn shit into shinola.  Pardon the language.  Anyway, that\'s my new daily schedule. If I want to  be  a writer, then I have to   write   something. And finish it.  I\'ll be keeping a daily record of my progess on the blog. Not that anyone ever reads this thing, but I\'m doing it for my own sake. Posterity.  -----------------------------------------  My grandparents live in Cape Coral, Florida. About 2 hours west of Miami, on the Gulf Coast. Right now Charley\'s closing in on Key West, and by mid-afternoon it\'ll be on top of South-West Florida. Which means Cape Coral.  The street my grandparents live on is the very first street to flood, in even the smallest rain storms. A hurricane could probably submerge half their house, if not all of it. Cape Coral is nothing but a network of canals and harbors, anyway. All water.  I don\'t know if the county is going to make the residents evacuate, but my grandfather is refusing. He\'s getting stubborn in his old age. He\'s old-school Irish. All of his closest friends are dead, his wife hates their lives, and their health is failing. They\'re heavy wine drinkers, and chain-smokers. Or rather, they used to be. But I have a suspicion they sneak here and there.   This is my grandfather, Popu, shaking his fist at the Universe.  Dad can\'t talk him out of it. He\'s dead-set on staying right where he is, come hell or high water, literally.   I guess we\'ll know what happens tomorrow, won\'t we?  -----------------------------------------  I still haven\'t heard from you. Wow. We talked almost every day for two and a half years, and you can\'t think of what to say to me? No. No, I think maybe it\'s because you DON\'T have anything to say to me. Please prove me wrong. Because I can\'t believe I\'ve lost my best friend. '),(10208,4074186,'12,August,2004','2004-08-12 00:00:00','   Maybe it\'s simply b/c you don\'t care, really. It\'s a non-issue for you. Maybe that\'s why you can\'t think of what to say to me. '),(10209,4074186,'05,August,2004','2004-08-05 00:00:00','   urlLink    When the Hellbilly had hair...&nbsp; urlLink '),(10210,4074186,'04,August,2004','2004-08-04 00:00:00','   urlLink    4:20, baby.&nbsp; urlLink '),(10211,4074186,'04,August,2004','2004-08-04 00:00:00','   urlLink    Pontificating. Taken by Claudine Auguste.&nbsp; urlLink '),(10212,4074186,'04,August,2004','2004-08-04 00:00:00','   urlLink    The Electric Hellbilly in all of his inanimate glory.&nbsp; urlLink '),(10213,4074186,'04,August,2004','2004-08-04 00:00:00','   40 oz. and a pack of smokes. Loud music, to drown out the Last Conversation looping in my head.  Trying to arm-wrestle Love never ends up in victory; and suddenly the universe whirls upon you, and you sit there with pie on your face, a fool to your own inanities. What a tidal wave of arrogance, what a thunderbolt of hubris, that I should be knocked down this hard. But It put me right where I needed to be, I guess. This is where I belong.  I only wish it didn\'t end the way it did. I wish I could\'ve looked her in the eyes when I said good-bye.  But I\'m a man, I suppose, and a real man wouldn\'t tread in the past. I will not be ruled by the capriciousness of my emotions, or the stabbing pain of memories turned sour. I am a man of decision. And I choose to accept the rich treasures of the future.   Someday, I will find someone who will love  me . '),(10214,4288065,'20,August,2004','2004-08-20 00:00:00','    Howdy Yaâ€™ll,   Itâ€™s me again. I been sittinâ€™ here on my porch lookinâ€™ at all the sale papers that came in the mail. Itâ€™s time fer the youngâ€™uns to git back in school, so they got all the school stuff on sale. I never will understand what differnce a certain kind of notebook is gonna make in the way a youngâ€™un learns. I seen wired notebooks, zippered notebooks, colorful notebooks, plain notebooks, ones with movie stars pictures on â€˜em, ones with cartoons onâ€™em, and ones with ever color of the rainbow on â€˜em! They got regler pencils and fancy pencils, regler pens and colered pens and pens with special grips. Crayons now come in big boxes of 96 differnt colors, and paper is in ever color you could want to write on. They even got special stores that sell nothinâ€™ but school stuff!  They got them backpack things fer the youngâ€™uns to lug all this mess back and to in. The TV said the other day that doctors is worried about the youngâ€™uns back messinâ€™ up before they even get out of school to get a job! They just ainâ€™t no sense in any youngâ€™un totinâ€™ that much on their back! We didnâ€™t have to tote that much when we worked in the fields â€˜till we got older and big enough to handle it.  Can anybody out there prove to me that all this new fangled stuff is teachinâ€™ the youngâ€™uns any better? Most youngâ€™uns now ainâ€™t got sense enough to get out of the rain unless some game tells â€˜em to! The youngâ€™uns that make it to college are taking so long to graduate that they done got gray hair before they know what a job is! And when they do git a job, guess where they go first? To the store that sells all the school stuff so they can git fancy stuff fer their new jobs! â€˜Cept now, theyâ€™re called â€œofficeâ€ supplies! Donâ€™t make no sense to me, neither.  Since we done got all the supplies, are the teachers any better? I wonder. I see on them school calendars that they schedule teacher work days. These are days that the youngâ€™uns could be in school, but todayâ€™s teachers need the time to fix up report cards. I figered it shouldnâ€™t take â€˜em as long today â€˜cause they got them computers to do all the work. In my day, the teachers had to do the figurinâ€™ themselves, and the writinâ€™ themselves, and there werenâ€™t no such thing as settinâ€™ up an appointment to see your youngâ€™uns teacher. If you needed to see her, you just stopped by the school when you had a minute. Boy, if you try that now, theyâ€™d have you hauled off! Yet all the time you hear the teachers whininâ€™ about they cainâ€™t get the mamas and daddies involved in their youngâ€™uns schoolinâ€™! They need to learn that you cainâ€™t have it both ways. Either you want me there, or you donâ€™t. They need to learn that time is important to mamaâ€™s and daddies, too. Some mamaâ€™s got jobs, and those that donâ€™t got meals to cook, homes to clean, and probably got clothes to fold! Most daddies got jobs, too. They have to take off to go see the teacher. The least they can do is work with the man!  They need to learn that supplies donâ€™t teach the youngâ€™uns, good teachers do. What is wrong with the good, old fashioned ways of readinâ€™, writinâ€™, and arithmetic, as long as you throw in some of that computer stuff so the youngâ€™uns donâ€™t fall behind the rest of the world? Now a days, theyâ€™re teachinâ€™ girls how to be welders and boys how to cook. They got health and safety classes, babysittinâ€™ classes, and picture takinâ€™ classes. If the mamaâ€™s would teach the cookinâ€™, and daddies would teach the weldinâ€™, and we all raised our youngâ€™uns with a little common sense, and left the picture takinâ€™ as a hobby, they might find they could teach something that mattered.  Todayâ€™s youngâ€™uns ainâ€™t as dumb as most of yaâ€™ll think. They got plenty of brains to learn with. Itâ€™s just that what are we teachinâ€™ â€˜em? I was told one time to keep an open mind by some of the youngâ€™uns that I run up with. The only problem is that if your mind is always open, there ainâ€™t no tellinâ€™ who is goinâ€™ to dump what in there!   â€˜Till next time, Granâ€™ma Gertie '),(10215,4288065,'20,August,2004','2004-08-20 00:00:00','    Howdy Yaâ€™ll!   Welcome to the world of Granâ€™ma Gertie! I reckon I aught to interduce myself. I was borned Gertrude Maybell Smith somwheres around a hunderd years ago. I been married to the same old geezer fer what seems like forever. I got more granâ€™youngâ€™uns than a pear trees got blooms in springtime. I got my share of wrinkles, and my hairâ€™s been gray fer so long that I done fergot what color it used to be. Iâ€™ve always been a big olâ€™ strong gal. My daddy used to say it was â€˜cause I come from good stock. I was born and bred in the country, and never took a shine to the city. Except for my olâ€™ man. He was a city boy all the way, and thru and thru â€˜till I got hold of â€˜im! Had to teach him everything he knows. Now that you know a little bit about me, I reckon we can discuss a few things, you know, chew the fat a bit.  I reckon at my age, Iâ€™ll never git used to all the new things this world keeps cominâ€™ up with. Some of my granâ€™youngâ€™uns just spent around $300.00 dollars fer me a miker-wave. They said it would cook real fast. I just ainâ€™t figered out what they think a rotatinâ€™ light bulb is gonna cook! It does make a perty good shelf to put some of my what-nots on, though. I reckon Iâ€™d like the clock part too, ifâ€™n it didâ€™t blink 12:00 all the time! It seems to me that the women now-a-days have just about cut off their noses to spite their faces. What with all the new fangled gadgets and thing-a-ma-jigs they got now. Take them biscuits in a can fer instance. Now that they ainâ€™t got a use fer a rollinâ€™pin, what do they think theyâ€™re gonna WHOP their olâ€™ man with ifâ€™n he needs it? Only thing I can think of is to fix them sudden biscuits in a cast iron skillet and use that to hit him with.  They got telephones that do everything but make a phone call. I, fer one, sure donâ€™t like them answerinâ€™ machine things. I donâ€™t need some machine tellinâ€™ me to leave a message, I ainâ€™t so old I cainâ€™t call back! They got cars you have to plug in and teli-visions that you donâ€™t, addinâ€™ machines that are so little you gotta have a magnifyinâ€™ glass to use â€˜em, and radios so big it takes two people to tote â€˜em. They got perfume thatâ€™s sâ€™posed to smell like real vaniller flavorinâ€™, but they use imitation flavorinâ€™ in their cookinâ€™. They got take out food that sâ€™posed to taste like homemade, and their homemade taste like it should be took out, maybe to the dawgs!  Women used to know their place. Then some gal mustâ€™a thought we ought to get smart and start makinâ€™ us a new place. These modern women got things so messed up, Iâ€™m beginninâ€™ to wonder ifâ€™n we were better off where we were.  What with all the gadgets and such we got now, I donâ€™t know whether I wish we could go back to yesterday, or whether I should just keep hanginâ€™ on to see what the heck is next!!  â€˜Till next time, Granâ€™ma Gertie '),(10216,4288065,'20,August,2004','2004-08-20 00:00:00','    Autumn Afternoon  As I sit on the front porch using the last of the dayâ€™s light to glance over the evening paper, I feel the slight chill that has crept into the air. I have spent the late afternoon watching the golden brown leaves of the sycamore tree float slowly and gracefully to the ground. The pine straw is a bit heavy on the now browning grass in my front yard. I know I shouldnâ€™t be wasting my time just sitting on the front porch. I know I should be in the yard raking up the leaves as they fall. I just canâ€™t bring myself to leave my perch. I, after all, have a perfect autumn view. I can sit quietly listening to the birds getting ready to nest at evening light. I can watch the squirrels preparing themselves for the fast approaching winter. I can see the pink shimmer of an autumn evening sunset. I can smell the soft fragrance of a fireplace beginning to burn. I feel the cold breath of winter beginning to creep into the fast approaching night. My toes, poking out of the sandals that I have lived in since April, are suddenly quite nippy.  I know that soon, I will wake up to a yard that has been diamond studded with the first frost. All the leaves will have finally found their final resting place in my front yard. The children will be catching the school bus in the early twilight of morning wearing new sweaters and jackets. Thanksgiving and Christmas decorations will soon be showing up, glowing brightly in windows and dancing across lawns. The sound of chainsaws whining and axes cracking will be echoing across the neighborhood announcing the golden glow of fires and the perfume of smoke to come.  I almost hate to think about removing the fallen leaves from the yard. They are a signal of the coming changes. The slow, methodical grace of the leaves falling down is as graceful as the soft, gray, billowing of smoke coming up from the chimneys. The newly grounded leaves are the signal that winter is fast approaching. They act as a reminder to get out your winter coats and blankets. They tell you to winterize the lawn mower, to break out the rakes and leaf bags, and get the yard ready for colder weather. Listen to them.  I donâ€™t think Iâ€™ll worry about the leaves today. I may not worry about them tomorrow, either. I may just spend the early fall sitting on my front porch. Just sitting and watching as the season shows her finest colors and changes moods as quick as the fall wind blows.  Then again, I may not have to rake them. Maybe there will be a good strong wind to blow them all away into the oblivion of my next door neighbors yard. I never liked her anyway. '),(10217,4288065,'20,August,2004','2004-08-20 00:00:00','    Tombstone Of A Home   Where there now stands a lonesome chimney, there once was a home. The chimney is all that remains as a testimonial that once upon a time, someone cared. With loving hands, each brick was carefully lain in a mortar of mud. Even though all traces of the home have long since disappeared, the chimney still proudly stands.  How many cold mornings did this chimney serve itâ€™s purpose? How many children were warmed by itâ€™s hearth? How many times has it warmed both soul and bottom, faces and hearts? How many times did the children of itâ€™s past warm their hands before going out into the cold, perhaps to school or to do their chores? How many times has a father warmed his feet and hands after working in the fields? How many clothes did the mother press with the flatirons that were heated here? How many ? How much? How many times? So many questions, so much history.  The lonesome chimney now stands as a grave marker in time. It has many memories that it could tell. It makes a person wonder if it had a mantle. If so, were the family portraits there? Maybe a bud vase of wildflowers given by a child. Think of the labor that went into the building of this chimney, and also of the building of the home and family that it warmed. It now serves as a reminder that long ago, someoneâ€™s father was young and strong enough to cut and split the wood for itâ€™s use. That long ago, someoneâ€™s mother was young and strong enough to keep that fire burning brightly.  The outline of the home can still be seen if you look hard enough and possibly have a little imagination. If you listen long enough, you can almost hear the voices and laughter of the family that once lived here. You can see where there might have been a rope swing tied in the ancient oak that still stands in what must have been the front yard. If you are lucky enough to be here in springtime, you may see some of the mothers now wild gladiola and tulip bulbs peeking through the soil. You can imagine where the clothesline was, where the garden, chicken coop, and hog pen must have been. You can see what could have been the childrenâ€™s favorite place to play, and where the mother kept watch over them. If you listen closely enough, you may even be able to hear the family dog scratching out a cool place in the sand under the porch that once stood here.  The mother and father that once lived here grew old and watched as they slowly lost everything. They were no longer young and strong enough for the old house. Life had called their children into a different way of life. Not better, not worse, just different. The children rarely visited after they grew up and moved away. Sometimes, the old couple would receive a photograph or letter from the grandchildren, but not very often. You know how busy our lives can be. The children decided that they were too old to remain in the house. Sometimes, the roof would leak and the cold breath of winter would come creeping in the cracks around the doors. After a lot of convincing, the couple was moved into a â€œretirementâ€ home. It was here that they passed away. The children never returned to the old home place.  Almost like a person, a home needs love to survive. When no one cared anymore, slowly, board by board, and nail by nail, the old house lost itâ€™s strength and gave up. No more laughter would fill itâ€™s rooms, no more dinners in itâ€™s kitchen, no more fires would burn. Maybe the chimney still stands in the hopes that someday, maybe a family will return. Maybe they will see how well it has stood the test of time and rebuild a new home around it. In the meantime, the lonesome chimney will continue to stand proudly, with the memories of the past. Always and always, forever hoping. '),(10218,4288065,'20,August,2004','2004-08-20 00:00:00','    If Only They Could Talk   I have always been what you could call a yard sale shopper. I cannot count the Saturday mornings spent with my family looking for that perfect bargain. I have furnished my home, bought gifts, clothed my entire family, met many interesting people, and saved plenty of money in the end. An estate sale, though, is an altogether different story.  For those who do not know, a yard sale is when people are cleaning out closets, re-arranging, moving, or making room for new things. They are held on lawns, porches, garages, or sometimes in living rooms when the weather is too cold or rainy. They are held in every month, with spring and summer being the best. They are held in big cities and small towns, and in neighborhoods everywhere. An estate sale, on the other hand, is often just as it says - the selling of a persons estate after death. Perhaps I have a melancholy view, but so be it.  I have attended several estate sales with my mother. She quite enjoys them, and has a different view than I. She feels it is a privilege to be able to purchase treasures that the family doesnâ€™t want or need. I feel it is an invasion into a persons personal life. Just a way for the adult children to dispose of family â€œjunkâ€ so as to settle everything before they return to their busy lives. I wonder if they would feel the same way if these items accumulated over the years could talk? If they knew the rich memories behind all of the family junk. If they knew what each piece had meant to the family member.....  Hello, I am  Grandfather Clock . I guess you could say I kept the time of this family, not just in minutes and hours, but in memories, too. It was me that father watched as each child was born into this family of mine. It was I who counted the time before the children went off to school, and before father went to work. My numbers taught the children to count. It was I that let out the chimes to signal bedtimes and curfews. Often, I was hated by the children because I signaled the end of some enjoyment, or the beginning of some dreaded chore. Over the years, Iâ€™ve seen the children come and grow and move away. It was I that was stopped when father passed away. I didnâ€™t know then if Iâ€™d ever run again, but just as life kept on, so have I. Simply standing in this hallway, ticking away the minutes of this families life. Now that mother has passed away, our home is being separated and sold. I wish I could tell them all the things I remember. My memories are for sale at the price of $ 55.00.  I am  Miss Mirror . I have hung in this hallway for over fifty years. I was a present from father to mother. He used to stand in front of me, with his hands wrapped around mother, and tell her how beautiful she was. Iâ€™ve seen so many smiles! Iâ€™ve watched as mother fixed her hair before going to church on Sundays. Iâ€™ve watched father check his tie. I watched as son and daughter primped and checked their looks before dates, prom nights, and weddings. I saw the children grow from babies amazed at their reflections, to proud adults. Iâ€™ve also seen the tears when the family was told of brothers death in a war, of sisters car crash, and all the little arguments of a family. Iâ€™ve seen so much, but can tell them so little. Iâ€™m not worn out or used up, but I guess Iâ€™m useless to son and daughter. If only they could unlock the memories I have. My price is only $ 35.00.  I am  Mrs. Rocker . Iâ€™m the one who watched sleeping babies in mothers arms. I remember all the nights of colic, fevers, chicken pox, and broken hearts. I remember all the sleepless nights mother spent with me while waiting for some member of the family to return safely home. Iâ€™m the one who stayed up with father when he sorted out problems ranging from a disagreement with mother, to how to pay for daughters wedding. Iâ€™ve seen more worry than most of the house, but Iâ€™ve seen plenty of joy, too! Like the time mother discovered sons first tooth while nursing him right here! And the time mother discovered what it felt like to rock a grandchild for the first time. So many memories wrapped up for the small price of $ 45.00.  I am  Mr. Hammer . I supposedly belonged to father, but spent most of my time with mother. Oh, sure, father used me from time to time. To fix a few things like the nail in the front steps that keeps coming up, or putting up targets for sons first BB gun. Mostly though, I was used by mother. To open stuck windows, open paint can lids, and pull out old nails. I was once used to threaten a sewing machine that wouldnâ€™t cooperate with mother while she was making a quilt for daughters wedding! Iâ€™ve had a few new handles over the years, but I work just like new. It seems with my experience, I would be worth more that the price tag of $ 2.00.  My name is  Ladle , and I guess Iâ€™m the speaker for the kitchen crew. I assume we are the least thought of because we are the cheapest priced. We range from a dime to a dollar. We all wonder if son and daughter remember all the after school snacks, sleep-over party favors, birthday cakes, and Sunday dinners that came from us and mother. Nothing could make us feel better than to see their smiling faces after they had finished eating mothers homemade cookies or special Saturday muffins. The things she made for this family were pure labors of love. The kitchen was so hot in the summer. We didnâ€™t have fans or air conditioning. Mother always made things from scratch, never from a box. Everything in this kitchen speaks of her love for this family. We wonder if our new families will have that same love, or will we be considered outdated and thrown away at the end of this sale. If only we could talk!  With the breakdown of the family unit, it seems to me that some of us have forgotten just what the word family means. Think about your life. About your parents and your childrenâ€™s lives. Will your children hold memories, or an estate sale price tag at the end of your life? How about you? Will you hold a tag, or will you remember the smiles of a gift given? You know, the trinket box you gave to your mother when you were little? Will you remember the fishing trips with you father when you see his old tackle box, or will you see the price of a lure? Look in your home today. Remember where you got each little knick-knack, each picture. Sure, some things may have no sentimental value, but what about the things that do? Think about these things the next time you see.......  *ESTATE SALE* 1234 ANY STREET YOUR TOWN, USA SATURDAY 9AM TO 2 PM '),(10219,4216034,'13,August,2004','2004-08-13 00:00:00','    MAC/ACNA Members...  Have you ever been to the Radnor Concours? Would you like to participate in a gracious road rally through the Main Line with Porsche\'s, Mercedes and fellow Audi owners? Would you like to participate in one of the US\'s premier automotive show events? Now\'s your chance.  MAC/ACNA organizers strive to provide a variety of ways to \'join in\' that reinforce our commitment to enhancing your Audi ownership experience through Lifestyle, Performance, Technology and Community events.  Through the outstanding support of Vision Audi of Reading, we have been invited to participate in the Radnor Concours Rally on Saturday September 11th, 2004. The Radnor Rally is a low key, fun event and is by mileage only - it is not timed. It starts at about 8:00am at the Radnor Hunt and ends no later than 2:00pm. Refreshments/Lunch and awards are included in the entrant fee of $50 per car (includes one driver and one passenger). MAC/ACNA members are welcome to participate with a guest and/or significant other. This event is designed to be congenial and fun! And, you\'ll get to see some of the world\'s finest automobiles!  Please contact me as soon as possible if you would like to join us. If you\'ve never done something like this before, it\'s a great time to join in and try something different! For more details visit  urlLink www.radnorconcours.org .  Note that we need a minimum of 5 MAC/ACNA members to participate in order to have our own class (yep - an Audi class!) and we need to have participants signed up by Monday 8/23/04...  Please email me at danjackson@audiclubna.org if you would like to join in.  Thanks!  Dan Jackson MAC/ACNA   '),(10220,4216034,'12,August,2004','2004-08-12 00:00:00','    Problem:  If driverâ€™s knees push against knee bolster under the dashboard, electrical short circuit and fire could occur.   Models:  172,866 cars including A6, S6, and Allroad Quattro, made 8/97 to 2/04.   What to do:  Have dealer install protective covering over knee bolster (also known as crash element). Dealer will also inspect and, if necessary, replace headlight-switch wiring harness under dash as well as felt protector on left side of dashboard carrier.  For more information call Audi of America at 800-822-2834.   '),(10221,4216034,'12,August,2004','2004-08-12 00:00:00','    Faulty ignition prompts major Volkswagen, Audi recall    AUBURN HILLS, Mich., Feb. 4, 2003  Some 530,000 2001 to early 2003 Volkswagen and Audis have been recalled because of potentially defective ignition coils that could lead to a partial loss of engine power, according to the automakers. The recall includes the Audi TT and A4, Volkswagen Golf/GTI, Jetta, New Beetle, and Passat--all with 1.8-liter engines. The recall also extends to the Passat with W8 engine, all Volkswagens with the 2.8-liter VR6, and all cars with an Audi 3.0-liter V6. The majority of affected vehicles are 2001 and 2002 models. Only about 36,000 2003 vehicles are involved. The 2.8-liter V6 Passat is not affected by this recall. Gerd Klauss, president and chief executive officer of Volkswagen of America, said in a statement that \"some ignition coils are not up to our high quality standards\" and that the automaker intends to fix \"every single car with these coils by replacing them whether they are broken yet or not.\" Ignition coils provide electricity to the spark plugs. Through dealer service reports, the automakers noticed that the coils in the suspect vehicles were failing at a higher-than-normal rate. If a coil fails, the \"check-engine\" light will blink and the car\'s performance may become rough, possibly to the point of losing some power. Neither Volkswagen nor the National Highway Traffic Safety Administration consider the coil failure to be a safety issue. VW says that roadside assistance is available for anyone who does not feel comfortable driving with a car that has a failed ignition coil. Letters went out to consumers at the end of January 2003 notifying them of the problem, Volkswagen spokesman Steve Keyes said. Car owners should make an appointment with their dealer. At present, VW is giving priority to those who have had an ignition-coil failure: broken coils will be replaced immediately. If the car has suffered a second failure, all the coils will be replaced. By March 2003, VW expects to be able to replace all ignition coils on a car if only one has failed. In order to minimize inconvenience, Volkswagen and Audi dealers will offer alternative transportation at no cost while repairs are made, the company says.  For more information call Volkswagen at 800-822-8987 or Audi at 800-822-2834.   '),(10222,4216034,'12,August,2004','2004-08-12 00:00:00','    Dear fellow Audi Quattro Club Members:  Welcome to my  MAC/ACNA  page.  This is a place where we can all get together, plan events, GTG\'s, and talk about all things Audi.  Feel free to post your ideas, questions, and suggestions here.  Come back and visit often.  Buckle up, and drive safely!   Alex   '),(10223,3873570,'10,July,2004','2004-07-10 00:00:00','   At the age of 23, I was barely making it financially.  Yes I owned a small place, but I earned a low salary that barely helped me get to the end of the month.  Once a month or every 2 months my mom bought me stuff at the groceries.  I was lucky because I used a uniform at work, which the hotel cleaned for me, and they provided one meal a day.  Food was not the best, but there was a slight selection.  I am getting off subject of what I wanted to write.  I lost weight, smoked, and dated a loser.  Finally I was lucky enough (I can say that now) to be dumped before Christmas 1993.  In February I ended up going to Cancun, instead of Rio de Janeiro, and I met Steve.  I did not want to meet any guy, after my recent break-up, but things happen for a reason.  I thought I would never see him again, so I had fun for intense 3 days.  As it turns out we continued talking on the phone, and I received weekly letters.  If e-mail would have been accessible, we would have communicated even more.   After six months, impulsive Clau, decided to sell some things and come to the US.  I thought what if he is the one?  And also I had many things that helped me make up my mind easily.  My mom was getting on my nerves calling me 10 times a day asking me for advice she never followed.  My brother\'s life style was messed up.  He is gay and was involved in promiscuous, abusive relationships.  I did have lots of friends that were excited about my romance, so they were pushing me to do it.  At one point I remember saying: If everything works out, I quit smoking! I was so excited.  There were 2 reasons to consider staying in Argentina: my grandparents and my recent closeness to my dad. My grandparents had been the support since pretty much I was born, and gave me ethic, moral and love guidance.  My mom never expressed affection to me, but specially my grandpa showed me love every second.  I lived 2 blocks from their place as a kid, and I would rather be with them than with my school friends.  My grandpa has always been my hero.  He arrived to Argentina with nothing, and worked nonstop until he no longer couldn\'t.  He had these amazing stories that I could hear for hours.  I played cards with my grandparents and had a blast.  If they were one of the reasons to stay, it all changed in the month of July 1994.   On July 1, I went to see my grandparents and had along conversation with my grandpa, in which he apologized for not being able to give me money to bring to the US and help me.  My mom had spent all the money he had sweat for.  I told him I was a woman, and I could provide for myself.  I told him not to worry about me, because I was a fighter and I was independent.  He looked relieved and happy.  That night I went out, it was hot, and I mixed alcoholic beverages.  I was very sick.  The next morning the phone rang to announce that my grandpa had died of a heart attack, ironically when he was showing a picture on the wall of the house were he had been born in Germany.  My mother or brother were gone, so I was the one in charge.  That was not uncommon, since I\'ve always been the mature of the three of us.  My life stopped, it was as if a dagger cut open my heart.  A few days later, after the funeral and all, I talked to my mom.  She was feeling guilty for having spent all his money, and thought that it caused his death.  I felt peace for having had that conversation with him the day prior to his death, but my hero had died.  My grandma got sick, and died on July 31st, just three weeks later.  They\'ve been together 51 years, so it\'s not surprising.  After my grandparents were gone, my reasons to stay grew even smaller.  I thought that by trying to make it in another country, I would be repeating my grandpa\'s story.    The one force remaining for me to stay was my recent connection with my dad.  After many trials and failures to have a relationship with him, I called him and we met for coffee.  I don\'t remember exactly if I called or he did.  I was amazed by what he told me.  As a young girl, I was convinced that he didn\'t want me in his life and he didn\'t love me.  My mom always guilt tripped me, and I believed everything she said.  I was not a girl anymore, and my strong instinct told me that he was sincere.  I started a close connection, only to discover how similar we are even though I grew up without him.  The bad qualities and the good ones, jealousy, gredge, stubborness, humor, strength, happiness, tenderness, dorkiness, you name it.  After a long time I understood why I was so different from mom and Rodrigo.  Many years earlier I wrote him a poem wishing for him to love me, and there it was...  He cared!  Still, I decided to come to the US.  I figured if he loved me, he\'ll continue to do so.  After 10 years he\'s the main reason for me wanting to go back.  But that comes in the next chapter.  Stay tunned  '),(10224,3873570,'08,July,2004','2004-07-08 00:00:00','   Wow,I thought it would be easier to start by traveling through my life, starting from the beginning.  I was born in Buenos Aires, Argentina almost 35 years ago.  My mom and dad were together up until I was 3 years of age.  I have a younger brother with whom I have periods of silence and resentment, and others not bad.  When my parents divorced, I was told that I was really angry for not seeing my dad for over 6 months. Apparently, after I saw him I came back with strabismus.  Being crosseyed has always bother me a lot, until finally last summer I had surgery, and I am much better now.  My mom re-married Ricardo, a Uruguayan musician and composer.  He was very creative and nice at the beginning.  During the eight years my mom was married to him, she made me feel very guilty about seeing my dad maybe once or twice a year.  Ricardo overtime became violent, and started to physically abuse my mother and I in particular. I don\'t remember anything happening to Rodrigo.  I remember trying to fall asleep at night and hearing him hitting my mom, and her pleading for help.  I also remember a couple times in the middle of the night, getting us together and telling us we were leaving the house.  He never worked, and my mom didn\'t either at that time.  They/we were living of my grandpa\'s tremendous effort.  He arrived to Argentina from Germany with nothing and built his own little empire.  Ricardo went through a face that he was following the macrobiotic diet.  I will never forget as long as I live that one day I asked him for money to buy candy and he slapped my face and threw his wine on top of m,y school uniform.  I had to go to school after lunch and stinking like an alcoholic...  After eight years my mom finally divorced.   I start a new paragraph because my life completely changed between the time I was eleven until pretty much I was 24.  My mom had depression and sometimes spent days vomiting by her bed with a bucket on her side.  Another thing that happened is that she became very promiscuous, and slept with almost a different man every night.  When I was 12 she told me to smoke pot that it was not addicting.  As a way of rebellion I never wanted to try it. As I became older every time I went out and came back, I would see a jacket hanging on the livingroom\'s chair.  I knew that man was sleeping with my mother.  A couple times I called her \"puta\", she slapped me and told me that it\'s either men in her room or her vomiting.  Of course I could not say much.  I started to be silent and to absorb all the things that were going on.  God, I was so embarrassed of my mother in public.  She used to dress with mini skirts and spandex tops...  My friends had good families, so I felt alienated and did not talk much.  I wrote some, but there was less writing than thoughts and feelings.  I started college many careers, but no one told me the importance of getting a college degree.  At the age of 24 my mom sold the big apartment and bought 2, one for me and one for her.  She wanted to live alone.  I felt less cared for than before.  I know I was 23, but I was more alone than ever.  My brother had always been my moms reason of being, and ironically he is the one that has made her suffer the most.  Ever since I was born, the role model and person I admire the most has been my grandpa Leo.  I guess, since I was little I was ambitious, hard working and loving like he was.  If it wasn\'t for the stable home of my grandparents I could have turn to drugs, sex or alcohol.  I look back, and I know that many people have it worse, but I don\'t understand where I had the strength to endure all of those tough days...  I will continue at another time.  '),(10225,3873570,'14,August,2004','2004-08-14 00:00:00','    Today was another Saturday in my life. I spent quality time with Steve... we went to have lunch to his buddy\'s place in downtown San Leandro... we later went to Tapioca Express and I bought some milk tea for my buddy and dropped it off at the mall... I saw her for just a bit, but it was nice... At 5 we went to Union City to watch  The notebook. I\'ve been wanting to watch it for six weeks. I don\'t know why, but I knew that it was my kind of movie. I haven\'t seen such a great movie in years!!! I am not the kind of person who buys DVDS, but as soon as this one is out, I want to buy it!!!!    After watching the movie, I wished that my love for Steve was as strong as the love Alley and Noah had. I normally tear in almost any movie, but in this one I cried rivers... It\'s a must see... an old times like love story. When the movie ended, everyone, or at least so it seemed, was crying. It was cathartic and woke up my romantic side. It also deals with Alzheimers, and the choice to die when ur body is giving up. If I were very sick or had a devastating disease I would choose to die. But as long as I am relatively healthy, I will never leave this world... even if at times I wish I would be gone... I cannot be so selfish to leave when some people count on me... at times I think that\'s not the case... but when I am sharp I realize that they do...the quick fix would be to take your life so you no longer have to feel... I am not such a coward... or I am one because I could not even try to end my life...even at the darkest moments there is a tiny glimmer of hope... life is a miracle every day... each day should be the first day of that miracle... whoever reads this, specially you Teesa, I love you and need you lots... don\'t you ever leave me alone in this sometimes cruel world...   '),(10226,3873570,'11,August,2004','2004-08-11 00:00:00','    Today was a day that I wished at many times would end. I wish I could just disappear, so I didn\'t have to feel or think certain things. The only moment that I seemed to be ok was when I saw Teesa. I was on the verge of crying when I arrived to pick her up. The reason was absolutely stupid: just listening to a cd made me feel incredibly sad about being an Argentinean... then I thought about my dad and his disease...what if the only positive figure in my family dies?...why? yes he\'s abused tobacco forever, but why would it happen? My mind was racing kind of continuing with yesterday\'s thoughts.        Somehow, little by little Teesa (with her powers) managed to make me      forget for a while and we had a nice time. After I left her at her house, anguish came back. Maybe I\'m pmsing badly, and I try to find a logical explanation when there\'s none. During the evening I texted Teesa and she helped me feel a little better. At around 8 I called her without anything to say. I hate those times, because I feel I am boring her to death and I called just to pass time. I like when we engage in interesting topics and after talking I feel: wow that was nice... When I am so negative I can never have engaging conversations. why? I wanted to talk some more, I was hungry for one of those conversations. Maybe Monday... if nothing happens so that plans are cancelled. I love my buddy, today my mood was such that I couldn\'t even tell her I loved her without feeling shy and out of place. Well, tomorrow has to be a better day...'),(10227,3873570,'09,August,2004','2004-08-09 00:00:00','    Today was a fine day!  I saw my best friend.  At the beginning we juast stared at each other without much to say.  It was very comforting to see each other, but we were getting acquainted with our new looks.  Hers is more drastic than mine! She is beautiful no matter what.  After wtaching TV, I let her read the poem.  I\'ve never written a poem to a friend before, she is truly special.  I also showed here the pics and gave small green gifts.  It could not be another color, right?  We went to lunch, and I let her drive through Alameda.  She\'s actually pretty good!  When we arrived home time was running short, so we did our favorite thing to do: lay in the couch next to each other.  I enjoy that because we are close in body and spirit.  It was almost like a mother daughter situation, Teesa fell asleep on my shoulders!  What\'s amazing is that I know how hard it is for her to sleep, and being with me relaxed her to the point she could do that.  It made me very happy!  I wish we could have some more of those moments before school starts...  Maybe next Monday?  I don\'t know.  Besides that nice moment something really important happened: we made a pact.  I will not smoke not even 1 drag, and she will not cut herself even one small cut.  Smoking and cutting are two self-destructive habits we have that need to die forever.  It will work!  I hope from the bottom of my heart that she does NOT let me down, because it could be extremely painful to overcome.  She has my trust and I have hers; we are very vulnerable, but at the same time we are two tough women.  I believe in us!  I love us!   '),(10228,3873570,'08,August,2004','2004-08-08 00:00:00','    It was easy to make the decision to move to the U.S. given everything I\'ve been talking about. My relationship with my dad was way too recent to change my mind. Creating a separation from my mom was what I thought I needed. It was like being free at last... My mom called me at my job at least 7 times a day, it was overwhelming. I thought leaving would be getting rid of my past problems. Well, in reality I carried them in my luggage... I was too impulsive to make the decision to leave my beloved Argentina, but in a way I felt I had no other way of putting my life back together; or better yet, starting to live! I desired so much to come here that I said I would quit smoking... I did. Yes, every time I go I start, but I quit again.   When I came to the U.S. I moved in with Steve without knowing him really. I was gutsy, but otherwise I couldn\'t. He introduced me to all his friends, and I remember people would not touch or kiss... They hugged in such a fake way that I felt so awkward! I missed and still do the physical contact we have in Argentina. Even if you don\'t know people, when they are introduced to you by a friend you kiss them on the cheek. Steve went back to work kind of quickly. At the beginning I went to do some sightseeing in SF, but after a while I was like afraid or lonely. I used to stay home and watch soap operas, and by the time Steve got home I asked 20 questions before he could answer the first one! I\'ve always been the kind of person that when I have too much time, I think crazy... After spending some money in shopping at The Gap (because in this country they carry my size, at that tiem was between a 10/12,) I started searching for a job. I thought since I spoke some languages and had experience in Tourism and Hotels, I could get that kind of jobs. But very soon that dream fell apart. In Argentina I would have never worked at a coffee shop, especially cleaning and doing dishes. Finally, I decided to search for a job at a coffee shop. I thought it would be easy, but without papers it wasn\'t. One day Steve got me job through a friend at Sweet Joana\'s cafe Howard and Spear. Every day I took the bus, and went to my daily humiliation session. The wife of the owner was very mean to me. At first I was troo shy to ask people to repeat what they have ordered when I didn\'t understand. I got better over time. Her husband and brother in law liked me and were nice, but she was a bitch. I worked there for a while until one day the bitch told me that business was slow, so they could no longer pay me. The husband hooked me up with another brother of his at the Fisherman\'s wharf. Things there were different. I was given lots of wasing to do of huge, sticky muffin trays. At least I had someone to talk to, a Mexican girl (Yadira). Customers liked me and I was not shy anymore, but I worked like an ass. The owner sometimes cussed at me behind my back in arab, but always smiled and pretended that everything was ok. He called me wet back, and I had no clue what that meant. Everything continued until one day I was told that a new girl will start ,and I will have less hours and will have to work weekends, so I quit impulsively. I told him I had a better job offer. I had nothing, but my ego was still strong.   Weeks after I found a job close to where I lived (a block away). The owner was Egyptian. I had to prepare food, coffee, clean, mop, you name it. Very soon I was in charge of everything and by myself. Omar would only come once a day if that. Customers loved me, and although the job was demanding I was happy. Business was slow at times and friends from the building came to visit, like the lesbian that made moves on me and Guillermina who got offended by my words... Just an Argentinean expression that it is offensive in Mexico. She was nice. One day, I went to work and the store was empty, Omar moved out the night before. To make things short, talking to some cutomers a Korean man offered me a job in a Blue CHip Cookie store he was opening soon. I went to Argentina in the time being, and started working upon return. I had moved out Steve\'s place, but when I lost my job, he told me to go back. I worked there for a bit, and Steve offered me to get a Civil Wedding to get my papers. I applied for a job at Office Depot and got it. It was ok. After 4 months Steve proposed with a ring, and asked me to get a degree. I started shyly, but college woke up a part of myself that had been quiet. I went to community college, got my AA, and transfered to Cal State Hayward. I wanted to go to UC Berkeley, like Trish, but I needed to stay one year to finish one class. I knew that no matter where I went I would do good if I just set my mind to it. And I did. I graduated Summa Cum Laude. After graduating I got in a program to receive my teacher\'s credential, and I did. I\'ve been teaching for 3 years, and I don\'t think I would do anything else, although it could be stressful at times. THe chase of a dream has been tough, and many times I wanted to throw the towel... I did not, and as I look back I feel proud! No one in my immediate family has received a college degree. In a way what I did is so similar to what my grandpa did... He is and always will be my inspiration...   '),(10229,3873570,'06,August,2004','2004-08-06 00:00:00','    I hate when technology overcomes me! It\'s the fucking third time I\'m writing this blog. I need to be tech smart! Well, the blog\'s topic was the continuation of my last one. I wrote about the positive influences I had that help me endure tough times. This 2 forces were Mirta and Marta. Mirta was my nanny in my early childhood until I was 11. She was my friend Mirta\'s grandma. She was there for me, and took my side when she thought I was right. My mom never heard her, and punished me for things my brother did. My buddy Mirta told when I saw her recently in Argentina, that I was always grounded... I was grounded for stuff Ro had done, but my mom always preferred him over me. My mother and brother went out to have fun while I stayed home punished with Mirta. I kind of liked the peace at the house, but   as time went by I grew apart from my mother. It\'s hard to have feelings today, even if my mother is trying to do the right thing, because she did not plant the necessary seeds or water the plant correctly when it was necessary. It is hard to show affection, when we never did...    Marta was another positive force in my adolescence. She was my nanny from 12 on. She has always been like a big sister. She took my side a lot, since mymom went crazy during my teenage years. We fought at times, but minutes later we were having \"Mate\" and laughing. I learned to cook by watching her, although I took a couple classes with famous chefs from 1992 to 1994. Marta was there when I needed guidance, and gave me advice when I was loosing it. She adopted Carlitos when he was 2. I used to take him to places, and when I saw him recently I could not believe how big he\'s got! He\'s 16! My mom always thought that Marta was partial to me, so she never listened when there was any issue w/ Rodrigo. At least someone that favored me! My mom slapped me in the face when \"I went out of control.\" With time, I started to laugh, no matter how much I was hurting. She stopped.   '),(10230,216833,'06,July,2004','2004-07-06 00:00:00','    The bump came in today - a day before I expected it. That is good news. I have to pay less apartment late fees and I feel a little better about the whole mess. Still waiting on those letters from the bank to resolve this creditcard fraud mess.  Just listening to the radio in the cold room - ahh - technology - everything I love about it seems to cancel itself out with everything I hate about it.   am ordering the new 7\" covers this week, if all goes well. I am looking forward to the end of the line with that shit. Looking back on the last year, all the changes that have happened seem to always have that in the picture - shows have fallen to ruins.  The band thing has run its course with me, never again. That is a stage in my life that I am over with for good. I love art, making things, music too, but I just can\'t deal with the bull shit anymore. Almost all my friends are from connections in music. I wouldn\'t trade that for anything in the world. I still enjoy making shirts. I don\'t really enjoy building websites anymore, not like I used too. I wonder if there is more to come. Have I just not matured enough, have I not found myself like I think I have?   Since Eric is selling a few things, I also have a few things that I will through up for you guys before I take em to my parents for a garage sale:  Glass Dinning table and chairs(the chairs have seen the rath of a kitty) $100 o.b.o.   Time Killers arcade game - $160  Corner Computer Desk: $40  I am moving and I will have alot of other shit for sale. But problably nothing really cool.   '),(10231,216833,'06,July,2004','2004-07-06 00:00:00','    hahahaha....the morning has been great so far. Halloween. I came to work looking like a zombie with an afro. I got the fake \"bloody scabs\" all over my arms. I usually don\'t do this but latly I have been enjoying this holiday. It\'s fun...its just a little fake blood, it never hurt anyone. People are jumpy around here, I\'ve just been doing my normal stuff and i get a few jumps here and there.  I guess it looks pretty good becuase i feel like a fucking animal in a cage... people just looking and going, \"Hey Mark, check this guy out!\" Fuck OFF! Its just a little fake blood, never hurt anyone...   '),(10232,216833,'06,July,2004','2004-07-06 00:00:00','    just added my interview with myself   '),(10233,216833,'06,July,2004','2004-07-06 00:00:00','    well.......i guess lunch time is over. I really wish i would get a record or two from someone in the mail... I got an order to No Idea for like 8 7inches and an order for a few SOV 7inches and a Karate 7\" i won off of ebay....nothing yet.  I feel sleepy. only 3 more hours of the work and I will be right back here again being retarded and wasting time in front of a PC.   '),(10234,216833,'06,July,2004','2004-07-06 00:00:00','    it did work! and some how it got rid of my freaking java error on the page...   '),(10235,216833,'06,July,2004','2004-07-06 00:00:00','    well i just wanted to test out this little blog from AIM thing and if this works, I would have to say very cool.   I\'m at home now, eating some leftover birthday cake and drinking a coke....no time flat I will be the biggest, fatest son of a... you have ever seen. haha. nah.....i need to go do some running again.   '),(10236,216833,'06,July,2004','2004-07-06 00:00:00','    well, im just at work. getting hungry and tired of people bothering me.  I just want to sit here and play with my website. I don\'t want to fix your email or make your printer work after you jammed it up pretty good.  yesterday was my birthday, i\'m 22 now and my ID is expired. i hate going to the DMV.  i\'m sure it will be a barrel of fun.   '),(10237,216833,'06,July,2004','2004-07-06 00:00:00','    SpyAttack goes Platnium  Most of this week has been pretty insane.  The weather was really nice, well it was cold. I find chilling air exhilarating.  Some how the icy air filling up my lungs makes me feel more alive than the smoggy filth that is Dallas summer air.  On Monday the talk in the office was \"Oh I hope the roads aren\'t bad!\", and \"Well, winter is finally here.\"  In Dallas, there is no winter.  There are three day periods in the Winter months where it happens to get below 40 F. My decrepit self spent most of the week at home or at work.  Not really talking to much of anyone esp after Wednesday when I lost my front tooth.  I find it very disheartening.  I know I shouldn\'t care what people think, but it just brings back the insecureities of jr. high and that\'s really what I hate about it.  I feel exposed, almost naked.  I have found comfort in the pages of William Sutcliffe\'s second novel \"Are You Experienced?\"  So far a great book.  I have to stay I really don\'t like to put it down.  Almost through with the book, when I am done, I will post my thoughts in their entirity.  So far, I am really ejoying it.  I think I am starting to cheer up.  I am totally looking forward to spending the next week with mimi\'s dad in California.  I think I am gonna go up in the mountians and do a little hiking.  There is a scooter shop in Sacramento called Vespa Haus that I want to check out.  I need to buy a little something for the bike so that will be fun.  Someone other than old man Garner.  And I am totally going to hit the record stores in Berkly and SanFran.  Maybe go fishing at the Delta.   Anyways.....I will try to post on here more frequently.  Till then, all my love.   '),(10238,216833,'06,July,2004','2004-07-06 00:00:00','    Welcome to my mind   Ah hell, whatever.  I\'m so sick and tired of everything.  I\'m pretty  unhappy the last few days, seems that all I want to do is bitch and  moan.  I\'m looking forward to this 2-day cold snap coming up. Then sometime in december it will get up to 80 again and make me want to freeze myself through the blistering summer.  Just thaw me out when it gets cold again.  I am getting that itch to move and I\'m pretty sure this time next year, I\'m going to scratch it.   '),(10239,216833,'06,July,2004','2004-07-06 00:00:00','    Well, this weekend went to hell as far as the planned events but I made it worth my while I suppose.  After canceling the show for Lubbock pretty much on the way, pissing everyone off, and wasting about an hour of everyone\'s time...I went a head and took Brian\'s drums to him.  I thought it would be nice to eat some dinner with my brother and hang out so I asked him to come along to get some food and I would by him some dinner.  He was down, but he also invited his friends....well, my mind changed because that wasn\'t really what I wanted to do...so I totally confused him but oh well.  I got on the road wanted to talk to some one about the events, I called up my parents, and then just decided to roll on over there.  I was there for about two hours just talking about this and that.  I invited them to come up this weekend...so maybe I will see them around tomorrow.  I want to take them to this awesome greek place in lewisville that I know they will just love.  Plus, I think my dad is going to bring the Vespa P200 frame that I bought for Brian like almost a year ago.  I think it will be better off in my garage than just sitting in the back of their yard.   Today has totally rocked so far...  I slept till 9 a.m. which was really nice.  For the last month it seems that there hasn\'t been one weekend that I could just sleep in my bed without doing much of anything.  I would have slept longer but Robert(co-worker) called to let me know they were going to be up at the office going over the Saturday system backup procedures at around 10am.  I needed to be there.  I sat around, checked my email, got dressed and went up to the office. It was a pretty quick ordeal. I spent about an hour there going over all the stuff that needs to be done every Saturday.  New responsibilities...fun.  While I was away, mimi had gone up to Planet Tan to bake herself for a few minutes...she happened to get home at the exact same time I did.  So we went and got some food and went to Target where I bought a \"boombox\" for the garage.    I spent the rest of the afternoon cleaning up the garage.  I took a whole gallon of bleach and poured it on the floor of my garage.  The fumes where so intense that I couldn\'t breath or see.  My eyes were watering so bad I had to go outside to get any air worth breathing.  Some how I managed to sweep it all out and I used a fan to help dry it.  After all that mess I spent about 4 hours working on the dismatleing of my Vespa engine. That went Ok for the most part, I had a few problems with a few bolts....some of these problems are still pending.  It was pretty frustrating. I had to go to home depot to pick up a wrench....the real reason was to take a break from the engine. I was about to give my self a heart attack, I just get worked up over shit way to easy.  I feel like I made some head way on the project so that is cool. The real pay off is when I start putting it back together.  That is where I bet it\'ll get super tough. Around 7pm I am finishing up the laundry and I call up my Uncle Scott who lives right down the street.  His wife, Carrie, was going to a movie so he was going to just be sitting around the house for a few hours with their little baby boy, andrew.  I went over there for a few hours and talked about all kinds of stuff.  I really dig Scott because he reads alot and has alot of interesting view points and I like to hear what he has to say about things.  Anyways, I should be going to sleep soon.  mimi and I are getting up in the middle of the night to go watch the meteor shower.   '),(10240,216833,'06,July,2004','2004-07-06 00:00:00','    Well, for the last day or two things have been brightening up a bit.  I really can\'t say how much, but some.   '),(10241,216833,'06,July,2004','2004-07-06 00:00:00','    Whatever will I do.... I had a very bad day yesterday. I had an appointment to see the Dentist at 3:30 yesterday afternoon. Got off a bit early from work and went over to the dentist office, I didn\'t have to wait very long to get into a room, which was cool, but i had to wait on the dentist forever it seemed like.  I was going to check and see what could be done with my broken partial, I have it just sitting there and it is  very  annoying.  She told me I need a bridge, fuck, that sucks because that is like 2,500 bucks for a fucking tooth. Anyways, I was pretty upset, walked to my car crying becuase I know it will be a few months before I can get the bridge and that means I have to deal with this broken \'false tooth\' for that much longer.  I went home and cleaned up the house real nice for the Art Collective meeting.  No one came.  I laided on my bed for about 40 minutes and read a new book I just got at the library.  the history of buddhaism. It\'s pretty interesting. I was about to fall asleep, but I finished the chapter and laid my head down.  8pm i was asleep and mimi woke me up at 11:30 and told me to get undressed for bed, i still had my pants on.  whatever..........   '),(10242,216833,'06,July,2004','2004-07-06 00:00:00','    well i\'m at work....thinking of all the shit i need to get ready for another weekend of wuornos shows. first off I want a fresh shower.  my cat pissed all over my clean socks that i had in a basket.  I am really sick of that cat.  I love her, but damn, that is just nasty, she doesn\'t know.  i think her litter box needs fresh litter. I guess whatever.  We are going to longview tonight to play at the Texas Blues Bar.  I\'m stoked about it...first show with brian as our drummer. Hope it goes well. I wont be back until Sunday night.   '),(10243,216833,'06,July,2004','2004-07-06 00:00:00','    year-end: 2001   Well here I am, trying to think of all the things that have happened in the last year and I really only see the last four months.  I have seen more, learned more, and changed more in the last four months than I have all year.  I really started to expand my worldview more after nine eleven and try to see the world with a different eye.  This led me to a new change in my beliefs on animal rights, corporations, city life, money, health, love, and friendship.  I always loved the urban life, running from street to street and taking it all in.  I still do but there is something different now, I see all these people running around, what for, I find confusion in my new found knowledge.  I don\'t know about you, but I go through this from time to time like most humans.  What am I doing, why am I here.  Maybe I should help people, but then, maybe not, I need to help myself.  Fuck you and your needs, but this is selfish.  So what am I to do with my time? Should I go through life to make as much money as I can and do all that I can possible dream about?  Sure, why not.    This weekend marks 5 years to the date when I met mimi, Dec 30.  I know that sounds stupid, but that is the day she started work at the Sweet Shop where I was the cook, famous for my perfect Corn Dogs.  haha.  Yeah I made em better than anyone ever.  So they say.  That was a totally different time for me.  I was a very different person 5 years ago, and that is just bizarre to picture in your mind what your were 5 years ago.  I can see myself in my mind, I had this pair of jnco\'s that were cut off mid-calf.  I still have them, they are in a bag of clothes I\'m going to take to goodwill.  I always wore plain white t-shirts and I had a chain wallet, I had hair then and around this time it was bleach blonde and I was still skinny.  Didn\'t drink then.  Beer did it, that didn\'t happen for another year.  I was busy, with my fast food job and my bad ass 1972 Chevy pickup.  I had stickers all over the back window of the pickup. Dead Kennedys, supertones, mxpx, the specials, value pac, etc...  I was in between bands at the time, but I had all the equipment to start a band, with my work money I had bought a drum kit, guitar amp, PA, guitars and bass guitars and it was all in my room.  I had a door to the room from the outside, my closet was paneled with cedar and my clothes always smelled so good.  I\'m started to feel old, but I\'m just glad I can remember sometimes.  Now I\'m at the brink of twenty O\' two; Happy new year.  What lies in the future is so beyond me that I can\'t even begin to make a prediction.  I know that the person I am now is the person I think I want to be, it\'s always that way, but maybe I will become a better person, maybe not as pissed off, maybe I will finish things that I need to finish and move on.  Love you all.   '),(10244,216833,'06,July,2004','2004-07-06 00:00:00','    x-mas and more of that sort of thing   well, the holidays were more exciting than most so far. I spend a few days before x-mas to take a tour of west texas stopping in little towns that house family of my girl and i.  that was nice, i got a chance to see some family that i havn\'t seen in about two years, like uncles, cousins and such.  on the day after x-mas i spend holding it down at ssg.  mimi was off work on thursday and we really didn\'t do anything but hang around the house, hell thats all i\'ve done for the past 4 days, it seems.  glued to the tv playing grand theft auto 3 on the playstation 2 to that mimi bought for me.  i went to a show on friday night, that was a highlight.  i saw a few people that i hadn\'t seening in a while, all the bands were great: el salvador b-day bash, ...of death, kill the slavemaster, and since by man.  great show.  lots of energy or drunks, can\'t tell, but it was my first chance to get out my camcorder that mimi bought for me for x-mas.  I owe it to her for sure.  so i have the show on tape and that was fun in itself.  from the desk of john coffman   '),(10245,216833,'06,July,2004','2004-07-06 00:00:00','    Dinner at Monica\'s   Dinner last night was quit an event.  For the first time in my life, I met some people that I actually found interesting. It\'s a weird twist on the feeling of meeting new people because I don\'t really admire these people as much as I just plain find them interesting.   mimi\'s mom was in town for a meeting and she had invited us out to dinner with a group from her office. I was dreading it.  I just really didn\'t see it as being fun, but I was hungry, and it was Tex-Mex with a vegetarian menu so I sucked it up.  The place was like a crowded bar, oh wait, it was a crowded bar filled with the cool and trendy.  It was a really cool place and it was half-priced entree night so I could see the reasons for being so busy.  Hell, tonight they have 50 cent margaritas, I can just imagine all the people there then.  We walked into the waiting area and there was this group in the middle that we were meeting up with.  Let me see if I can get this right, but there was Jerri, the old yank woman that moved her giant pear-shaped body around like a weeble-wobble. Her shepardess, Tiffiny, who was the young, repitively defensive and also over weight girl who seemed to feel the need to protect the thought of her open-mind as Fort-Knox.  Next to her was Rose(i think that\'s her name), not much to say about her except the fact that all of her stories seemed to be very anti-climactic.  Also, the host\'s of the dinner, Lee and his wife(I forgot her name).  Young, maybe late 20\'s early 30\'s, very much caught up in the city life of Dallas.  Into all the lights and sounds of city culture.  I found them to be more inspiring than the other people at the table.  Lee\'s wife scuplts and paints.  Lee seems to be very sucessful and I could see that his charisma is what got him there.  He really didn\'t seem that overy interesting but he knew how to deal with people, something I am really not good with.  His wife was able to pull out some very interesting topics and we discussed art, vegan resteraunts, and music; Lee would occational note additions to the conversation. The food was really good, I would go back.  Maybe on a less crowded night.   today  Just going through the same uninteresting things; the life of a computer geek.   '),(10246,216833,'06,July,2004','2004-07-06 00:00:00','    End of the workday..closing in on 4pm  I\'ve been sitting at my desk for the last three hours trying to stay busy with out much will to get up. Now I am finally to my daily duty of uploading the shipping files to their respected shippers like UPS and the post office.  At least now I have something to do every time I go to work. What will I do with the rest of my afternoon, like I said, mimi is off today so hopeful we can go have some fun.  Christmas has completely left us broke so it seems. I think I have gone a little overboard with the presents but I want mimi and everyone to have a good christmas, its my favorite holiday and damn if I don\'t see people smiling.    urlLink check out dallas hardcore music   '),(10247,216833,'06,July,2004','2004-07-06 00:00:00','    Another day...most of this morning I\'ve been reading about how to build a log cabin.  I don\'t know why but maybe someday I can build my own cabin from the sweat of my brow.  I think I could seriously find that rewarding.  I\'m at work, things are alot calmer today.  I\'m going to lunch with a finance guy from emerson/ssg and robert is going to.  The guy is taking both of us to lunch for...i guess our help with his bank software.  MiMi is off today.  I hope she is enjoying it.   '),(10248,216833,'06,July,2004','2004-07-06 00:00:00','    But the bad part was that it wasn\'t really cold.  We all set forth on the camping trip hoping for a cold weather trip and it ended up being very pleasant.  The rain was not bad until late at night when the fog also set in.  Good thing we were staying in a palace of a tent.  Dad had bought a 10-man tent.  The thing was huge.  It didn\'t even fit on the pad.   '),(10249,216833,'06,July,2004','2004-07-06 00:00:00','    The forest was dark at the early hour of 6p.m. but this is December.  There were probably a total of 3 occupied campsites in the entire park which made for a very people quiet place to be.  It was really just the nearby road and the forest sounds.  I get there with a headache, which I am totally upset about, but I try to drink away, while at the same time occupying my mind with the duty of Fire Chief, protector of the Flame.  Man was that Fire raging.  I absolutely love camp fires, everything about them, the smell of the smoke, the heat hitting your face while the cold wind slices up the back of your coat.   '),(10250,216833,'06,July,2004','2004-07-06 00:00:00','    This weekend was adventure number 2 to the Davy Crockett National Forest.  Another insteresting experience to the campgrounds of Ratcliff Lake. I really enjoy that area.  Scott, my mother\'s youngest brother, and my Dad were my partners in this quest into the woods in search of chilling air and campfires.  We made our way out from Ft.Worth where Scott and I met up with my Dad.  It took us a little longer than normal to get out to Ratcliff because we took I-45 a little longer down to Centerville and from there we crossed over.  The trip was interesting. It rained the entire time, from when we left to just about the time we entered the campsite.  Halfway to our destination the wipers go out on my Dad\'s Suburban.  That made it almost impossible to see, but we made it to the forest and got to our campsite without slamming into anything.   '),(10251,216833,'06,July,2004','2004-07-06 00:00:00','    How shitty all this is!  I had two pretty interesting adventures and my site was down.  David at the hosting server moved some stuff and we all had to re-upload the pages.  Well this one was pretty customized so I had to move a few files.  Luckily I had the files here at work so I can go ahead and start putting it back together.  Looks good, or the same as before.   '),(10252,216833,'06,July,2004','2004-07-06 00:00:00','    so what it is what it isn\'t.  I have some super sinky ass farts going on tonight.  oh my.  mimi and i went to a thai resteraunt in dallas called the thai rose. it was really just alright. I had the Evil Jungle Price with tofu and some thai iced coffee and mimi had the asparagus and some sprite.  i enjoyed the time we had just hanging out. it was fun. we tired to go shopping for some family for christmas and saw these awesome new toys....remember the Rudolph the Red nose raindeer claymation cartoons that ALWAYS come on around christmas.  THEY MADE TOYS of ALL the characters and they are off the hook.  I want em all so bad.  That seriously gave me chills when I saw em.  It just rocks.  I mean....i love those shows and I watched em every year, that is a totally good memory of fun at christmas.  Man.....if you get a chance....buy me some....hehe...love john   '),(10253,216833,'06,July,2004','2004-07-06 00:00:00','    Well from my point of view, things don\'t seem to go anywhere unless you move them.  Cars don\'t go unless you fill them with dino-blood and start them, lives dont live until you kick them in the ass.  So when is someone going to kick me in my ass.    Thing to start again by the end of the week.   · Lift weights (becuase I need to be mean)  · Study my cisco books regularly and get to the goal of a cert.  · Find a place for wuornos to practice  · Eat better for the fucking sake of staying alive.    I dunno why I just can\'t do these things.  Fuck it.  Tomorrow I am going to do what I want, not what my weak ass mind gets away with but what I want.   You can\'t have it all you say, well, I can if I make myself.  Its going to be hard.  But I guess instead of pussing out and getting fat and poor, I\'ll stop being a light weight and do it!   Updates coming soon!  Man, losing friends makes things go weird....I\'m way to dependant on people....from now on its me and I don\'t care if you don\'t like me.   '),(10254,216833,'06,July,2004','2004-07-06 00:00:00','    Holy §moke!   I took a trip to the  urlLink library  yesterday afterwork to pick up some refrence material for the  urlLink DFWAC  meeting.  I had spent some time online researching different sites to find books that I thought might help us come up with some ideas and maybe show us new ways to organize the group to get stuff done. mimi was sick with the cold, so I left her at home, crying and sniffling on the couch.  The  urlLink library  is only a few blocks down the street, behind an Exxon and a dry-cleaners, so it was a short dash to my local branch.    The library was packed with people, but here, the books are the majority and most of the time I would rather take a book with me than a person.  People are to edgy sometimes, and they have problems that they always seem to bring to your front door.  I don\'t know why, but I\'m always greeting strangers on the street with a \'Hello\" and \"Hey\" followed by some sort of nod.  I found my self doing this as I was walking into the library and going, \"wtf!\" I don\'t know that person, why do I give a flying fuck enough to say \"Hello.\"  Fuck that guy for making me feel weak and habitual, and now I am saying \"Fuck your self for blaming your weak ass on some stranger!\"    weird...I had a small fight inside my head..........am I crazy??  I\'m crazy, fuck you!   I got my books, and went to check them out and they had some nice calendars sitting on the counter...FREE....wow these are nice.  I picked up two, one for my house and one for the office(a.k.a. the life I am to weak to abandon.)   '),(10255,216833,'06,July,2004','2004-07-06 00:00:00','    I think I have a cold, and I don\'t like it very much...oh I wanna go back to bed.   '),(10256,216833,'06,July,2004','2004-07-06 00:00:00','    Four A M Lanscape  Wind waves the reeds by the river, Grey sky lids the leaden water. Ducks fly low across the water Three flying: one quacks sadly.  Grey are the sky and the water, Green the lost ribbons of reed-beds, Small in the silence a black shadow Floats upon wide pale mirrors.   year-start: 2002   The first few hours of this year I spent with Nick, Kira, Brian(my brother) and his friend Bryan, and MiMi.  We talked the new year in with a failed attempt to celebrate simultaniously with the rest of the time zone.  Canadian wiskey and potatoe chips where the first food of what will be undoubtedly be the most unpredictable year of my life.  Towards the closing in of sleep i was in the middle of watching Planet of the Apes around 3am.  I started to feel a little sick, like alchohol sick.  I stayed awake for a while longer; trying to get through the movie with out having to lose my chips and \"salsa\". \"Maybe I need to lay down,\" I told myself.  I went into the room and undressed and slothed my way under the blanket, this didn\'t help, just made me realize that I needed to take care of this.  I was sober, to a point, and hadn\'t had a drink for over an hour, why was this belated.  RRAAAAAALLLLPPPHHHHH!.  I feel better now, finished the movie and went to bed.   Its almost like stabing your hand, you know its gonna hurt, but the rush, some say, is worth it.  What a stupid way to start the year.   By the way, Nick and Kira saw the canal rat moster thing last night.  what is that thing!   '),(10257,216833,'06,July,2004','2004-07-06 00:00:00','    http://www.museumofhoaxes.com/photos/photos.html some of this are interesting....   '),(10258,216833,'06,July,2004','2004-07-06 00:00:00','    i wonder if anyone is going to make fun of the image i have up on the main site!  hehe   '),(10259,216833,'06,July,2004','2004-07-06 00:00:00','    Lets see if they take this olive branch.........  <><><><><><><><><><><><><><><>  Saudi Arabia: Peace With Israel Possible  VOA News  By Jessica Berman  Posted Monday February 25, 2002 - 06:34:18 AM EST    Washington - Saudi Arabia\'s foreign policy advisor said Sunday that peace in the Middle East is possible if Israel withdraws to its 1967 borders.   Appearing on the television program Fox News Sunday, Saudi foreign policy advisor Adel al-Jubeir said that in exchange for total withdrawal from occupied land, including Jerusalem, Israel would gain the recognition and normalization of relations with its Arab neighbors.   Mr. al-Jubeir called the proposal a \"statement of a vision\" by Crown Prince Abdullah.   \"The importance of it is it sends a signal to the Israeli public by telling them that peace with the broader Arab world is possible should they make peace with their neighbors,\" he said. \"It also sends a message to the peace camps in both Israel and the United States not to lose hope and that peace is possible and that we should re-energize our efforts.\" Mr. al-Jubeir says the idea was positively received in the Arab media and by Arab leaders, including Palestinian President Yasser Arafat.   Meanwhile, Prince Abdullah, reportedly upset with Israel\'s continued hard line against Palestinians, cancelled a speech he planned to give at the Arab Summit the end of March in Beirut outlining a peace proposal.   On another topic, the Saudis are of two minds about a U.S. base on their soil. On one hand, the country is a close American ally. On the other, the government worries the U.S.   presence could set off a wave of terrorism.   Also appearing on the television program Fox News Sunday, Joint Chiefs of Staff Chairman General Richard Myers was asked if there\'s been any change in the U.S. Saudi relationship. \"We\'ve had a strategic interest that goes a long way back,\" he replied.   \"That relationship is still valid, still working.\" \"The Saudis could always ask us to leave,\" he said, adding that \"we have no signs at this point that that\'s the case.\" Vice President Dick Cheney is headed to the Middle East where he will likely discuss Saudi concerns about the U.S. presence, as well as any new peace initiatives.   '),(10260,216833,'06,July,2004','2004-07-06 00:00:00','    Nothing like some chips and sprite for breakfast.  sprinkle some pepper on it an BAMM its a meal.  Here is the start of another week. Robert, my co-worker, is taking a vactions this week.  So its me and James on duty.  Hopefully it will be light.  The weekend was beautiful.  Very nice outside the whole weekend, it might have been a little to windy yesterday, no problem.  I made a decision sunday to take my kitty to the SPCA - I feel alot better.  After that trip, I did some studying for about an hour and 1/2 and went to check my email and message boards.  Alex was online, after his 18 hours of sleep, and said we should go pass out some MELANGE flyers. I headed straight over to their apartment and soon after that we treked around dallas putting up flyer.  Where will we go next.  There is that FW, 2/3s of the name, that we need to tackle as well.  Our \"west-side connection\" has failed us.  We had little trouble putting up flyers.   There is always that one asshole guy, and he works at the Magnolia so spit on him if you see him!   FUCK YUPPIES, cuz they are the people who think beauty is clean!  I like clean, I hope I\'m not a yuppie.  I\'m hip yo!   '),(10261,216833,'06,July,2004','2004-07-06 00:00:00','    Went to see pinback last night at the gypsy tea room, the opening bands were ok, nothing to shout about.  Pinback was a little disappointing, they rocked, but alot of their songs were on \'crack.\'  Super fucking fast.  I bought their 2002 tour E.P. and I have listened to a little bit of it.  Sure enought, penelope is fast as hell on it.  I was real impressed by the bass player, mad skillaaazzz.    We had a sad attempt of passing out flyers at the show, alex put some on one table and I put a few on the other.  We are gonna need a bottle of B.O.L.D to get some flyers out to people.  After the show, I kinda gave mimi directions home and I hitched a ride with alex and eric up the TLP(where alex works) to work on some more flyers for MELANGE at insomnia.  We cut out 11x17 portions of that brown paper roll sheet stuff and ran it through a color laser printer.  I really like how they turned out.  HOLY SHIT, im crazy and its 4:30 in the AM.  Yep, we were up at the office till some time after 4:30am and somewhere in all that mix i totally lost my mind.  I was dancing around the Mat Room shouting \"melange\" in a fake as french accent all at the same time waving my pretend ribbons around in the air.  I think they were red and purple....haha....french people rock!  I crawled and falled out of bed around 10:30 this morning, feeling like someone at taken a tire tool to my knees.  I had to make my way up to work to change out some tapes for the weekly backups of the AS/400.  I got that finished pretty quick and headed back home wishing I could go back to bed.  mimi went to work around 1pm and I was sitting around fight with my self over going to school.  I didn\'t want to, I hate that place, I always leave confused and pissed off at the people who work their for being idiots.  I took my books and went the opposite way from school.  I went to the Mall. I passed out a few Melange flyers and put one up in Fast Forward there.  Those kids were really cool and asked me a few questions about it.  I bought a messenger bag, I can\'t find my old black one, lost i suppose.  I can\'t get the cat piss smell out of my old faithful backpack so I found a bag I liked and took it.    \"Shit, I guess I\'ll go to school!\"  I went and it wasn\'t bad this time, I met a new guy there, he was really helpful.  He seemed to care at least, so I am gonna start going there a LOT more.  I am ready to have something to show for the money I\'ve spent.  I\'m back home now after about 2 hours of going over custom subnet masks, and IP Routing, RIP.IGRP. ARP. etc.......  I\'m going to Denton tonight to see a friend I havn\'t seen in a while, maybe play some pool and shoot the shit.   '),(10262,216833,'06,July,2004','2004-07-06 00:00:00','    What an absolutly beautiful day it is outside, the birds are cirping, the air is crisp.  I love it, when I was walking in to the office, there was a mockingbird sitting in a tree singing.  It\'s wierd how when you say anything about a brid, esp. a mockingbird,  its like almost always deep and poetic.  haha.    Last night I rode out to the Evil Dirt Road to see ...of death (they didn\'t play).  I had a good time either way.  Last Nerve was generic and I wasn\'t impress. Just your good ol\' hardcore eh!  Terms of Grey had an almost comical set, i love those guys, they are just full of heart.  Of coarse tyler is always cutting up with the crowd.  They had a better than usual performance.  USA is a Monster was the headlineing touring band.  They have some free mp3\'s at  urlLink infrasound  and also I found some info on their records at this pretty cool distro called  urlLink massdist .  I was very glad i got to see that.  Two guys from AZ coming to this little shack of a garage and tearing it up with all they\'ve got.  It was great.  I had some awesome sleep last night.  Got home about 12 and hit the sack.  slept till about 9:15 and found some clothes and went to work.  I feel tons better than I did at the start of this week, I\'m going to see Pinback tonight with lots of friends, I\'m stoked. I\'m looking forward to chilling out to some mellow tunes.   '),(10263,216833,'06,July,2004','2004-07-06 00:00:00','    Last night was hard for me, I spent about 2 hours studying subnetting and IP addressing and got so frustrated I wanted to cry.  I took a shower to help me feel a little better, it did, but I couldn\'t shake my mind from constantly tearing me down.  What do you think about yourself?  Most of the time I am very critical of myself.  I guess I set my self up for failure alot.   It\'s a new day, that makes a world of difference.   Girl Scout cookies are soooo good.   '),(10264,216833,'06,July,2004','2004-07-06 00:00:00','    So I probably should say something like - \"Woah, its been a while since I\'ve updated.\"  Thats is the way a journal is sometimes.  So its Monday Feb. 18th.  Tonight wuornos was supposed to play a show in ft.worth.  Non of us really showed any real effort at trying to get it together.  Oh well.  I really don\'t have time to be in a band.  I need to just stop all this shit all together.  I feel completely overwhelmed.  I guess that is the way it is going to be.  I am sure I would just be bored if I had nothing going on.  Its a challenge for sure.    Will power could really come in hand right about now.  I feel like all I want to do is bitch, what are the things that make me feel good?  Working out, lifting weights, reading awesome books, eating good food, making art that I like, just living and doing new things.  I\'m an f\'ing lazy shit ass.  I know what I need to do, I just gotta do it, seems like i write about the same things.  This journal is really more of a self-disipline thing for me.  Makes me see how behind I am.   Lots to do, better get started on some of it.   '),(10265,216833,'06,July,2004','2004-07-06 00:00:00','    Stereo Photo   Thought this might be interesting to my digital camera friend.   '),(10266,216833,'06,July,2004','2004-07-06 00:00:00','    Take the ASS Quiz   '),(10267,216833,'06,July,2004','2004-07-06 00:00:00','    Well, I am finally getting something done about my apartment being all fucked up.  I turned in a request in December to get my disposal replaced and the lights in my kitchen and garage replaced. Mark is the \'make ready\' guy but sometimes the service manager, Tom, gives him a few tickets to handle.   Tom said he does maybe 10% of all the repair work around the complex.  By the way, both Mark and Tom are very nice guys.  Mark is always smiling and laughing but he sorta weirds me out, i dunno, he just does.  Tom, well, he is just a rock n roll guy who loves guitars.  He is always wanted to \'jam\' out with me at every chance he can get; I have yet to take the time and \'jam\' with Tom. Well, anyways, Wednesday I was determined to talk to the folks at the office about this lack of follow through with my disposal, lights, and now my carpet pad that was removed when my water heater broke a few weeks ago.  I really didn\'t want to get any of the guys in trouble becuase I wasn\'t sure who did the work; correction, who DIDN\"T do the work.  I had been on the look out for Tom all day, he is always riding around in his golf cart so I was sure eventually I would see him.  I didn\'t see him in the morning or at lunch when I came home, but after work I just decided to go to the office, I hate dealing with those prissy girls, but oh well.  Just so happens, I\'m backing out of my spot and ZOOM! here comes Tom on his cart; so I wave him down.  We talked for a while, about all that was wrong with my  pad.  He apologized and I said he no big deal really.  I wasn\'t that pissed, not anymore.  He told me I was paying for the service plenty and needed to take atvantage of it.  Your right Tom.  I need to step up.  Anyway, Tom is coming over on Friday before noon.  I\'m off, so maybe we can \'jam.\'   '),(10268,216833,'06,July,2004','2004-07-06 00:00:00','    Normally I hate Television, it\'s usually a waste of time and it seems to steal time really fast.  I have to admit, I am hooked on a show.  the Osbornes - has to be the funniest show on TV right now.    Ozzy is gone, his kids are characters, and his wife is NUTZ!  I love it.      ***   I worked out legs today and I am so exhausted.  Tonight mimi and I are going to different stores to register for our wedding.  That is totally coming up soon, but I\'m not real apprehensive about it anymore, I\'m actually super stoked about the deal.  It\'s really happening and I can\'t think of a better person to spend my life with, time will tell.   ***   I noticed that March has been a slow month for the spy attack journal. I\'ve been busy; yeah, whatever!  I bet I could always say that.  Any point in my life, I can always say, \"Oh Ive been super busy!\"  I need to invent a \'busy\' meter.  I just don\'t think I have anything to say, I need to take this journal more personal, not so broadcast like...or some shit.     Anyways,  I have been listening to:  Cave In - Jupiter Nathlie Marchant - Tigerlilly Weezer - blue Saetia and 90.1 NPR  I wish Jets to Brazil would come back through.  I love those guys.  I wish I could book them at my apartment garage.  \"Don\'t mind the dog shit, Watach out for the piss\"  By the way, I cleaned all that up last night.  My garage looks cool again.   urlLink    ZZZZZZZZZ~~@@##^^^^^^^^^____---zzZZ radom rules   '),(10269,216833,'06,July,2004','2004-07-06 00:00:00','    Well -  The art show on Saturday was a success, I thought.  Insomnia looks 100x cooler now, and I have to say I had a ton of fun doing it all.  What a relief to have this finished.  Surprisingly, everything went really well,   I could have used some more energy close to the end of the evening, but all was well.    I broke more glass on Saturday than I have in years.   '),(10270,216833,'06,July,2004','2004-07-06 00:00:00','    center>      urlLink take the \"what\'s my fault\" quiz.    urlLink (and then browse around mewing.net. because laura is cool.)   '),(10271,216833,'06,July,2004','2004-07-06 00:00:00','    All week long has been GO! GO! GO!  This last weekend I went to an arcade auction and that in itself was an all day event.  I picked up two games; Mortal Kombat and Time Killers.  I think there is something wrong with the MK machine.  It freezes up and the Player One buttons don\'t work.  I did install coin door locks.  That\'s all I know how to do right now.  Still searching for documentation or something; anything to learn how to fix these darn things.  Work has been \'non-stop problems\'....We have had tons of work to do and not enough strength to muster up for compeltion.  Problem, like dial-in access and running low on printer toner, ahhh, tons of shit.  Not even enough time to play a game or two of Counter Strike.  The Art show is nearing quickly.  I am looking forward to it, OH SHIT!  I just remember that we need one more band for the outdoor stage.  No Forms yet.  No articles yet,  no idea what is gonna go down!  My legs HURT so bad from doing lots of leg press, squats, hamstring curls. and those leg extentions.  I can hardly walk right,  I need some ICyhot rubbed all over my legs.  Later flizzo\'s   the ghost - ive been rocking out to these guys this morning   '),(10272,216833,'06,July,2004','2004-07-06 00:00:00','    Got the bump today!  So that mean, arcade auction, here i come!  Oh, I\'m so stoked about this, I hope its not lame...     ***   Things are looking up yo,  Insomnia changed management for the third time since we started planning MELANGE.   That is getting annoying.   '),(10273,216833,'06,July,2004','2004-07-06 00:00:00','    Some days are bloody, and others are even bloodier!  Sounds as if alot of people have lost someone today....   '),(10274,216833,'06,July,2004','2004-07-06 00:00:00','    Ah - the office birthday party.  Makes me want to vomit my taco bueno from last night.  I can\'t stand the off key sound of someone\'s birthday song from down the hall.  The fat fucks eating more cake then they should, not even really knowing whose birthday it is, they are everywhere when the song is being farted out of fellow employee.  What is up with \"CUBE DECORATIONS!\"  Seems bizarre to decorate a fucking cube.  I dunno, thats just me, but I don\'t want \"over the hill\" or \"streamers\" or \"midgits with ribbons\" distracting me from my work.  I got enough distractions already.  Although the midgits might be cool to have around for coffee and such!  Small people RULE!   '),(10275,216833,'06,July,2004','2004-07-06 00:00:00','    Rightousness is the key to the gates of heaven - some sorta metal is the key to my apartment. Plastic are the keys under my fingers.  Islands are the keys in Florida.   '),(10276,216833,'06,July,2004','2004-07-06 00:00:00','    Oh yes - I am dreaming of frosty pine trees and ice on the bridges.  No use; hot and humid.  Looks like a mad storm is about to blow up out side.  I feel so secure - haha - I have been loading up songs on myplay.com.  Thanx for the tip eric, this is a very cool deal.    I think I am gonna work on spyattack.com tonight - maybe a new splash image.  THE CHEIF IS DEAD - KILLED BY THE EVIL UNION CAVALRY.  Native Americans - SPY ATTACK!!!   '),(10277,216833,'06,July,2004','2004-07-06 00:00:00','    anyways - back to my weekend.  James perez, from denton, called me up and he finally convinced me to come up to denton and hang out with him.   I got up there around 9 and we started drinking at his apartment.  Went out on his balcony and talked and drank booze for about two hours.  We both decided we would walk over to fry street and visit a few pubs.  First we went to cool beans, had me a double bourban and coke and a shot of Jagermeister.  This was repeated several times through out the night, just at different places.  We left that bar and headed over to Muther\'s to find a pool table.  No dice, the place was full, James saw some people he knew so we talked with them for a while and left to go to Lucky Lou\'s - by this time, I am fucked up to no end.  I started calling people, I called nick, josh p. and tyler.  I couldn\'t really hear then, just shouted in the phone, the bar was loud.  We didn\'t stay there for to long before we went back up to cool beans.  Got another drink and a shot in me and we went up to roof deck to check out the band.  James had to bull shit the guys and say that I put out 7\"s .  I can\'t even remember what the band sounded like, but they have my card.  Oh well.   We went back down stairs and boozed away, and started talking to random people, oh and fucking pedro was there.  He is from brazil and a totally awesome guy - I met him through Josh P. back a few years ago.  James told these chicks that I was in system of a down, they thought that was pretty funny, so they talked with us for a while.  I can\'t remeber what was said exactly, I just remeber talking about me getting married in a week and this was my lame ass bachlor party.  haha.  I had a good time - we finally made it back to his place, and I smoked a bit with his room mate ricky and we watched EnemyMine.  Fucking weird movie for sure.   I called mimi and said I was in no condition to drive - stayed up a few hours watching that movie.  James passed out and ricky went to his room so I was there, drunk, and freaking out over this movie.  I havn\'t done that in a long while and I paid for it the next day.    Sunday, I got home around 10am and just sat on my ass watching movies - pay it forward, proof of life, and some bigfoot show on the travel chanel.    This morning I got up at 7am and went to play some tennis with mimi - I was grumpy.  Been listening to alot of Nick Cave and the bad seeds and danzig!  M O T H E R!!!   '),(10278,216833,'06,July,2004','2004-07-06 00:00:00','    My weekend was pretty interesting, at least I thought so.  I got up on Saturday and went to work to change the tapes for the weekend backups.  After that I headed over to Insomnia, where we were just lucky enough to have someone let us in to get the paintings.  Alex was there about an hour before and had taken down most of the art work.  From there, I went home to hang with mimi for a while before she trappled off to work around 1.  We went to the local tex-mex resteraunt and had a weird waiter.  He gave you these blank glares that confused me.  He either had  no thought at all, or some evil plan to poison our food.  After mimi left, I drove aross the fucking country to my brothers house in white settlement.  They should change the name to black settlement.  Took me about an hour to get over there.  I had brought some coke, but brian had no ice.   No big deal - had me a few keystone lights and started painting their fire place.  Not finished yet.  Left there around 4:30pm and went to stop by my parents house to pick up the carpet steamer and just to say hello.  Hannah was playing in her pool - jumping around like a crazy kid.   Mom showed me her pumpkin plant in the flower bed in front of the house and she took me to the backyard to show me her carrots, squash, and tomatoes.  Growing plants just rocks.  I can\'t wait until I can eat food that I watched grow.  I finally got home, the sun was falling fast in the west.  I poured me a vodka/cola drink and started in - watched a bit of TV.  Then james from denton calls.   - - - -   '),(10279,216833,'06,July,2004','2004-07-06 00:00:00','    Sometimes small town\'s are bredding grounds for bordom.  What will a girl do to free herself from the stangle hold of idle hands?  Just like any normal girl, she went to school and tried to do the best she could at the same time having as much fun as girl can have; close to bursting like an over-filled ballon.  Her dad was a drunk but he meant well and really didn\'t get involved in the kids activites, she felt alone alot - all of her older siblings had moved away.   Close to her senior year in highschool she met a guy that she thought was so cool.  He was the guy that you couldn\'t bring home to mother, he was an \'outlaw.\'   Of course, she could give two shits less what her mom thought, this guy was hot and he could have some fun.  One night, they both rode up into the yard of her friends house, the engine from the motorcycle making waves of beating sounds across the party goers.  They seemed unphased, their music was bumping and the toxins were pouring a river that babbled alot louder than any harley.  It was just like a scene out of some movie where you ride up to a cliff and there is a bonfire in the distance.  The closer you get the louder the metal \\m/ gets, the louder the screams and shouts get.   Two guys on the side porch were smashing beer cans on their foreheads, one fell over the rail and everyone started laughing.  The girl was taken by the care-free attitude of the party-goers, this was everything that she had been feeling, but wasn\'t sure how to find it.  All smiles and sparkles, she gazed around the scene of totally rebellion and hedonism.  As she walked into the trailer house, three guys and two older women where planted on the couch with smiles on their face, like they just heard the funniest joke in the world.  There it was, the source of their humor, a joint the size of a sharpie marker.  In the back room, serveral girls were passed out on the floor.    \"They musta had a little to much to drink..\" she smirked.   \"Their high on smack, stupid.\" the boyfriend grinned, acting as if \"stupid\" was the name of a dear friend. \"Have you ever done this shit?\" he asked. \"Sure, I had my first taste a few summers back.\" she was shaking with nervousness. He didn\'t ask anything more of her and they walked outside to check on a few of his other buds. The amazed smile on her awe-stiken face had turned slowly more to a glazed stumble with every beer.  She was talking to another fellow that she had in a class at school, and suddenly she was struck on the left side of her face.  With what, she didn\'t know, on the ground, she was afraid and sweat started to pour down her face.  She reached to wipe it away, it wasn\'t sweat but blood.   Another sharp jab was felt in her ribs, the alcohol did little to numb the pain.   After the shock was over, she gathered her self up and began to relaized what was going on, her boyfriend was upset.  She could understand, he arm around the other fella must have looked ill.  The shouting began and he tore away on the motorcycle.  Bruised and beaten, she limped to the trailor house, with the help of some other girls, and taken to the back room to treat her wounds............   ( I will continue the story later)   '),(10280,216833,'06,July,2004','2004-07-06 00:00:00','    The mesquitos are getting bad around the canal... Everytime I go out on my patio, swarms flee the corner near the storage closet.  I havn\'t been bitten, but they just fly around, buzzing in and out my space. I keep wanting to leave my water can out on the patio, but I am worried that it will become a breeding ground for more virus-like bugs.    My plants are doing good, I brought them in last night.  Just a little worried about the high winds and possible turrents of rain, I didn\'t want the little fellows beaten to death. On monday, I looked at the weather report and there was a thunderstorm forecast every day of the week.  We are still waiting on those thunderstorms.  My luck, the fucking clouds will wait until next weekend to unload, and there goes the wedding outside.  There is a backup plan, just not near as cool.  Work is boring, very much so and I am very worried I am going to get ripped off by techskills.  Fifteen months can go by quick when you have shit to do.  But I really can\'t do anything about it by worrying.    What Passion dies so quickly with discouragement.   '),(10281,216833,'06,July,2004','2004-07-06 00:00:00','    I could her them chatting down the hall. Boomingly,  his voice carried through the white tiled floor of the narrow crawl-space that people travel daily towards the vending machine room.  After all,  work chit -chat is rampant in the work places of corperate america.  The whole idea of two pot-bellied, tie wearing, middle-management goons hanging out by the water cooler seems to me like those fuck heads have some sort of daily ocean fantasy.  They must always stare into the water, like they are looking for fish.   \"It\'s fucking WATER! Are you looking for a fucking dead body! DRINK IT!\"   Sounds like the guys, down the hall, are talking about fishing. . .   '),(10282,216833,'06,July,2004','2004-07-06 00:00:00','    I\'m testing my will power today - eat nothing!    Everything is bad for you so I am just gonna stop eating food.  haha for today.  I think I will see how long I can hold out.  Tomorrow might be the same thing.   '),(10283,216833,'06,July,2004','2004-07-06 00:00:00','    Went to get my wedding ring re-sized last night.   When I got it on Friday, it was a little loose.   I sure as hell don\'t what the thing falling off and me pissing away a lot of money.    Didn\'t go to school, that nice run I took yesterday morning caught up with me around 6pm and I was dead ass tired.  I worked out my back in the gym for an hour yesterday also, but I really don\'t think that was why my legs where jello.  i think I am getting a headache.     * * *    I\'m in such a bad mood today, everything seems to be pissing me off.  Some days are like this, some days are not.  I hurt all over, I musta woke up on the wrond side of the bed.  Also - I am intching to go camping soon.  Anyone down with a camping trip.  Winter is over, so we will have to battle the bugs!  I want to go to Big Bend, never been there before.  ATTENTION!  All boring, lice-infested, re-hashers of punk rock!  burn in Hell!   '),(10284,216833,'06,July,2004','2004-07-06 00:00:00','    Went for a jog this morning and it felt great.  The air was cool - just right for me, and I felt very energized.  At lunch I went home to pick up the test pressing that were delivered to the apartment office the day before.  It was so hot, 87, at least that is what the weather website says, looks like its gonna storm soon.    Well - this really wasn\'t public knowledge until I was for sure, I told a few friends, but it seems that everyone knows.  I am planning on moving to sacramento, ca in october.  It\'s is a good amount of time away - I get the feeling that some people have already marked me out as \"moved.\"  I havn\'t yet, and now I really do have to move, job or not, I will look like an idiot if I have to stay.  I can\'t think of I would have to stay, or why my plans would change - but shit happens.  I hate computers more and more each day - the whole industry seems to me like such a waste of time.  It\'s always changing and you constantly have to learn new software and hardware, but it doesn\'t seem to me to truely be NEW knowledge or GROWTH.  Just another way of doing the same shit, making corperations richer.  Ever since I started into this, it was a job, I wanted to do something in music; but that dream has faded with discouragement.  I hope I find my footing soon.   I finished an awesome book last night - The Woman in the Dunes by Kobo Abe(1924-1993) The guy is awesome, dead, but awesome eitherway.  I bought his book - The Box Man - but have yet to start it.  I was still finishing up a book I have been reading by Henry Rollins, the same book that turned me on the Kobo Abe in the first place, when I bought The Box Man.   When I went to Midland over the weekend, I had forgot my book, so I went out to Barns and Nobles on Saturday afternoon and looked around for something to read for about an hour or so.  I found the Woman in the Dunes, and I also picked up a John Steinbeck book called The Log of the Sea of Cortez (I had heard some talk about it on NPR and it sounded interesting.) Anyways, the woman in the dunes is about an teacher/amature entomologist who travels to a beach looking for some new type of beetle, he ends traped in a hole by the locals and forced to shovel sand with this woman, protectors of the village.  Its a good book, you should  urlLink pick it up.   I need to go to school tonight, but I am not really looking forward to it very much.  Just another skill that will soon be out of date.   Knowledge that is out of date, what a fucking weird idea.   '),(10285,216833,'06,July,2004','2004-07-06 00:00:00','    I have been having battle after battle with sleep in the last few days.  I think maybe I have taken it for granted the fact that I can hit the bed and instantly fall into a slumber.   Discomfort is a massive head fuck.  I think that my threshold for comfort is so pathetic.  I hate being hot, I can\'t stand being itchy, and I feel like a fucking whiny brat half the time.  Guess I need to change that.    MiMi told me I was \"obsessed\" with my outer appearence this weekend.  I don\'t like to think of it as low self asteem - it\'s more of a constant dissatifaction with who I am.  I am always thinking of what I need to change to make myself work harder, be better, look hotter, and live longer.  Maybe I do take it to far, it\'s not that, maybe I talk about it to much.  There are those people that talk to much, my self included, and there are those people that  DO!   The wedding is not far off.  I am looking forward to a vacation, Mexico will be nice.  We bought some new snorkal gear, so I am ready for the reef.   I am also terrified - I have been dreaming of shark attacks - or just seeing a shark in the water is giving me nightmares at night. Last time we went snorkaling, I saw a baracuda - that freaked me out a little.   I am so facinated with nature that I think about it all the time - I just don\'t want to be eaten.    I\'m sure there are lots of things in the world that feel that way - i.e. farm animals.   '),(10286,216833,'06,July,2004','2004-07-06 00:00:00','    Seemingly impossible to shake, I have this feeling again.  That feeling that I am not happiest where I am.  I feel boxed in, trapped by my own self with one way out, and it\'s a fast and furious sprint out of the way of the buildings I have crumbled.  \"What sad-sad-sad scene this is.  I\'ve troubled myself to open a locked door; blind to the inversion of total and un-natural worthlessness\"  The bounty of self-beating is great.  I can beat my self to a bloody pulp and gain a wealth of knowlege only once.  I know this doesn\'t make sense--that is just what is in my head.  I always think clearly at these moments after the drag is over.  I can\'t stop my pointless bull-shit sometimes.  I wonder how many people today, on their internet journal are saying things like \"My life sucks!\" or \"I hate everything about myself.\"    It\'s easy to feel that way, it\'s almost hip and cool to feel bad; maybe, it\'s EMO!  At this moment I am listening to some mp3\'s of geoff farina.  Sounds alot like an acoustic karate.  It\'s the lead singer/guitar player from karate.  They guy is just amazing, I can absolute connect with what I hear most of the time. The world is all just a poem or a line from a song.  Every opinion can be assumed or boldly presented from shit like that.  I feel very pretty, hot even.  I have mad sex appeal.  haha.  Do I believe that?  What a sack of poo pies!  I am getting married in exactly ONE month!   Woah, i felt that heavy sinking feeling.  Not about getting married, but about the fact that I feel so overwhelmed with preparation.   MiMi and I are getting the cake, the decor, the tux, the rings, the place, the invites, and tons of stuff.  Oh man, I hope this goes well.  I still don\'t have a person to marry us.  At least after today I can go get the marriage license out of the way.  Pray for me and my mimi!?!  $%&!%*!&@$#!___q434@(#*@)(#!*@)#(*!@#)$#!@$   '),(10287,216833,'06,July,2004','2004-07-06 00:00:00','    Today the gym will be serving up hot biceps and triceps, on fire, burning to a crisp! F U C K YEAH! My conditioning is getting more natural as the months go by. I should have never stopped, even for a week, it is detrimental.   Well, I\'m tapped out. I disputed an officially posted charge today and they blocked my card. I can\'t spend a dime - I just think it would complicate things more. So as the days go on, planet services has morphed into other companies like ISUND fiber optics and ISUND Internet and the one that posted was B365 Stokes/Trent??? That is weird, even the lady at the bank thought that was odd. Damn fuckers!  Some times you just gotta push it to the very end of the line!   '),(10288,216833,'06,July,2004','2004-07-06 00:00:00','    Taking in consideration of my present circumstance, I feel apt enough to make it through the day. I still have yet to make progess with my little money problem. I opened up a new bank account today. I am not to solicitous to throw more of my money down a spiralling hole called \"My Present Bank Account.\"  Oh well, I talked to the directors at my school - I\'m fucked. I am having to work out deals and press intensely. I feel like a failure, a disappointment to myself. I couldn\'t make it happen, but there is hope and it starts on Saturday.  This week has just been a crap heap. I need a bigger shovel to dig myself out of this hole. I wish I could just through out all my obligations, get a backpack and travel, meet up with old friends and have a cup of Joe. It sucks to feel this way - you know what I mean. But it will pass and everything will be golden once again. It\'s all a big fucking circle. The burden of being an adult. A guy at work said it was harder to be a kid, you don\'t have the freedom, you can\'t make your own decisions. Well - I say bullshit. Being an adult is harder - you HAVE TO make your own decisions - there is still no option.   I think I\'ll go play some tennis this evening. James (the head net admin at work) just bought a boat. Soon I will be going to the lake for some water skiing! That will mos def rock!   '),(10289,216833,'06,July,2004','2004-07-06 00:00:00','    When a good day goes bad! FUCK! I am in some trouble. I was just checking my bank account like I do on occasion and I noticed something odd. I only had 16 bucks and just yesterday I had alot more than that.   I got hot and felt that panic feeling come over me. Oh shit - did i screw up and spend to much money. I don\'t think so. I went over all the things that I had bought over the weekend and it didn\'t add up. Not even close. MiMi\'s card was stolen out of her purse at a gas station in west texas the weekend before memorial day.  She called me right after it happened - I called the bank and the card was blocked. Problem solved.   I\'m not sure now.   So I call the bank and she lists off a few debits that have not posted. At first she only gave me the amount. $10 $8 $1 $250 $50 $150 $100  OK - I\'m officially freaked out. those are big chunks of my rent due very soon!! She gives me the name of the company making the debit Planet Services WTF is that shit??!? I\'ve never heard of this, and so I start my search.   Oh there is one for $500 that didn\'t go through. Well shit! thanx fuckers. I hope all of this is some big mistake becuase I don\'t want to wake up in the morning and be broke while Planet Services is sittin by some pool with my money!  I\'m gonna be sick....   '),(10290,216833,'06,July,2004','2004-07-06 00:00:00','    I\'m in such a good mood today. I think it\'s all a mental thing. Everything makes me smile - haha - it\'s nice to be nice to people for a change. Don\'t drag me down mother fucker - I\'m stoked about being alive.  Problems are being solved!!   '),(10291,216833,'06,July,2004','2004-07-06 00:00:00','    Sometimes I think this public journal is just a way for me to try and impress others with my life. But I know that my life is far from impressive. I just don\'t want people to think I am trying to be cool. I just think writing rocks and this is a daily outlet.(I edited this post to clarify my point) I wish I could be more truthful about the way I feel sometimes, but that is why I have a personal journal made of paper. I would just rather type it but I don\'t think I would ever write in the book this much. It\'s just so convienent.  Nick came over tonight and we had some quasadillas and I started up my stars wars robot grill. It looks like those little things that ran around on the death star and a cross between darth vader\'s helmet. I like it. Makes me feel like powerful to control fire. whatever.  I\'ve been writing a short story for Al\'s zine. I think I pretty much screw it up. I will write something fresh - blah. Listened to the new Cave In sampler EP thingy. I like it, i think their new stuff is awesome either way.   Work was the same - blah. Worked out the back today - I wanted to work on the chest - I will just get to it later this week. Maybe I will skip out on the legs this week.  I feel better than I did earlier today. I\'m just so fed up with everything.   '),(10292,216833,'06,July,2004','2004-07-06 00:00:00','    Looking forward to getting back into the gym today. I love being off work - but sometimes I feel like I waste my spare time - damn technology.  I\'ve been making more and more small talk with the janitor, eric. That guy is for real and just goes about his day with little complaint. He is an idiot all the same, but he seems happy so thats cool.  Like I posted lastnight - I saw Mulholland Dr. That has to be the no.1 most fucked up movie I have ever seen. It was good, but I am still confused.   '),(10293,216833,'06,July,2004','2004-07-06 00:00:00','    Well - watched the X-men movie and Mullholland Dr. today. Went fishing in the rain. Drank a whole 1.5 liters of Martini and Rossi Champane(sp). I washed all my clothes and donated some I don\'t wear. DRUNK. Going to bed soon.   '),(10294,216833,'06,July,2004','2004-07-06 00:00:00','    Let\'s see! How did I kill all that time. I played video games till about 3:30, downloaded tons of music and around 4:30 decided to go find a camp site.   After getting lost somewhere around Eagle Mt. Lake, first stopping at Grapevine Lake, I found myself far from home and pissed off. Of course, I could have told you the probability of finding an empty campsite. The drive was fun. We took tons of obscure farm roads and small curvy roads with big trees. It was a really nice drive.   Stopped by my parents house and said hello to them. I think they are doing ok, never can tell anymore.  Got home about 9pm and watched some tv - now I\'m off to bed.   '),(10295,216833,'06,July,2004','2004-07-06 00:00:00','    I got a few hours to kill - getting that craving to hit the gym and let off some steam. I\'m not pissed, just could use the rush.   I woke up alone yesterday and went to denny\'s for breakfast. Across the room was this father and son just finishing up their pancakes when I walked in. When I\'m alone at places I find myself being nosy. They were talking about what to do after graduation. The father was telling the kid to do some traveling before he meets some girl. That\'s fucking weak. I guess for some guys - girls are nothing but a money pit pussy worth giving up all of your dreams for a poke. I think school and a job would drag the kid down before a girl - but if that is how he see\'s it - whatever.  I went home and played some gta3 until my ass fell asleep and around that time mimi came home - we got some food. She left for work soon after that and I dropped off some film to get developed [I hate walmart.] I took off to school right after that - that was uneventful.  On my way home I picked up some engine cleaner and started work on my vespa engine. I have been trying to get all these parts clean so I can inspect em for wear and put it all back together. I was pretty discouraged from all that work. I havn\'t sweated like that in a while. It was pouring into my eyes and I felt washed out.   This went on for a while, mimi came home around 9:45 and we watched some MADtv and then we crashed like jail break. I was tired.  Today has just really begun - I\'ve been up since around 9:30 but I had some rought sleep. I woke straight up when this HUGE crash of thunder shook the windows. It was so fucking loud. One of my jalepeno planets fell over. I put the planet back in the pot, but I\'m not sure how well it will recover. Still just a baby.  Today is uncertain.   Jake and I were talking last night about challenges and how hard shit can be sometimes. I don\'t think time makes things any easier either. challenge after challenge and they just keep getting more difficult. There is really no one to blame for the increased difficulty - there is always something in the way. I know I\'m the guy that always puts those obstacles in my way. That\'s fine - I don\'t really mind - I just need to stop bitching. I have an awesome life - my spirit is fresh and I have alot to learn and see. Just rock out and I hope I make all the right moves. Right or wrong - I can take it.   '),(10296,216833,'06,July,2004','2004-07-06 00:00:00','    It\'s pretty late for me. I usually head for the sack, exhausted well before 11pm. I had some weird breakfast this morning. no lunch but I did try the vanilla coke. I wasn\'t that impressed. I LOVE vanilla coke but I\'m not sure I will be drinking the pre-packaged bull shit.  I got a new tattoo tonight. I drove out to Suffer City in Garland becuase I knew that a few of the artist\'s were scooter fans. I had seen pics of vespas and lambrettas in front of the shop. I get out there around 7:15pm. I had a little trouble finding it. I couldn\'t remember where I saw the place. I found it eventually. Walked in and Nathan, the first guy I met, he was a pretty nice guy. The place was \'punk rock\' which is totally fine with me. There were toys everywhere. I felt like I was at eric and alex\'s apartment. Toys in wrappers. I don\'t get it really. What\'s the point. I talked to another guy, Steve or Steven I can\'t remember. He was super cool and thought the atom would look good.   I walked in wanting to get one on each arm but had to just go with the right arm for now. It cost $150. seemed high compared to the $200 i spent on the mxpx guy (supposed to be $300 but I snaked a deal.) So maybe it was reasonable. Either way I got it and it hurt and it\'s big.   I was listening to a radio program around 11 when I went out to the store for some soda. They were talking to kids in Moldova about their parents. Apparently the country is in such bad shape that a quarter of the population has left the country to seek employment abroad. They said that the average yearly salary in Moldova is 150 dollars.   I felt bad.  That was my tat. I could have taken those toys in the wrappers down to the kids with my tattoo money and helped a few. I coulda - shoulda - probably won\'t. WhY?  Anyways, mimi said she is tired and will leave from snyder tomorrow morning instead of coming home tonight.   I called j.palmer and talked to him for about half and hour. He is doing good. Thats good. I need to push people farther away - getting attached sucks. I don\'t really need people anyways - thats a lot of shit!   '),(10297,216833,'06,July,2004','2004-07-06 00:00:00','    this is great - all i\'ve been doing today.   urlLink whatsbetter.com   Just like the VS. game.   '),(10298,216833,'06,July,2004','2004-07-06 00:00:00','    There is always something to read.   Not alot going on - I had a weird day at work yesterday. This chick at work pretty much tried her damn\'dest to stick her boobs in my face. Ugly slut. Then I go to my car to leave for the day and she is sitting in her car next to mine on the far end of the parking lot. I open my door and she says, \"Oh! You scared me...\"  I just smiled and closed my car door. See ya later, wierdo.  Got the fucking 7\" compilation in yesterday. Also, the deposit check from our vaction. I\'m fucking glad the wax is here. Now I gotta wait on the covers and this mess will be done - one fucking year later.   My legs are still pretty sore from the leg workout wednesday. Today is bi\'s and tri\'s.   I decided that I am not going camping after all. I need to finish up some things and I want to work on some shit around the house. mimi is going to snyder this afternoon for her cousin\'s graduation. Hope that goes well and she stays safe. She should be back around 2am. I think I am going to toy around with my vepsa - I got some skrilla so I \'m gonna clean up the parts and see If I can make some headway and get it up and running soon. Also, I need to dick around with the Mortal Kombat arcade. Maybe I should order the parts today. I will do that. Also, F.Y.I. - the Time Killers arcade game is up for sale. What\'s a good price?  I also need to clean up the house and finish folding the clothes. I need some beer buddies to drink some beer form my beer mugs. haha.   Sometimes I miss palmer - maybe I had a non-sexual crush on the guy - i dunno. But him leaving has really left an empty spot in my life. I feel like a dork but oh well. I guess I will have to shoot up to Denver and see him. I called him around Tuesday and Aurda answered the phone. I talked to her for a little while - I wasn\'t mean, I guess I\'m not pissed about it anymore.   '),(10299,216833,'06,July,2004','2004-07-06 00:00:00','    Last night mimi and I went shopping, she needed some jeans. I also picked myself up two new CDs. black rebel motorcycle club and the new promise ring. I like b.r.m.c, that cd is pretty fucking rocking but the new promise ring I havn\'t had a chance to listen to much of it just yet.  I am going camping this weekend, but it looks like it is going to be by myself. That will surely be some interesting self evaluation time. Time like that can make you think about alot of things. I think mimi we come out there with me on Sunday after 3pm but I am gonna go out on Friday afternoon. I hope it doesn\'t rain on my ass. My tent is in good shape. I am always keeping it clean and patching up any holes that wear thin. This might be my chance to hike out into the primitive sites and camp there.  Anyone want to go camping this weekend?  My fucking leags are screaming mutiny!! I had a hard time sleeping last night because those things would never get in a comfortable position. I\'m stiff - today I\'m working out the shoulders. I can really see the work paying off in that area. There are lots of muscles I never really paid attention to. Still need to pay better attention to my diet.  Another ongoing battle for sure.   '),(10300,216833,'06,July,2004','2004-07-06 00:00:00','    ? Most of the time, I wake up around the same time every morning. I hit the snooze button three times every morning after 7:30 - I lay there until mimi says get up about two times - and then she leaves for work. No I get up - it\'s usually 8:30 by now. ? I don\'t mind a little routine now and again. A daily routine of preparing yourself for what the day holds is never a thing I would lightly throw out the window. I\'m not saying I spend an hour fixing my hair. I don\'t think I will ever have to do that. [Unless I go for the \'comb-over\'. Next of kin to the mullet.] I just feel generally better through-out the day if I took the time to iron my shirt or make sure my armpits don\'t stink.  ? Maybe it\'s all self-esteem - probably. ? I feel sorry for this dog next door. Up to a point I feel bad for the dog. Then I start getting pissed at the owners. Every morning around nine, this dog gives out the loneliest howls and bellows. I\'ve seen this dog a few times too. She is a great Dane the size of a house. I wish I had the strength of some of these big breed dogs. Big black and white spots look like she came from the house of an 18th century novel writer who spilled a lot of ink.  ? I guess the owners leave around 8:30 or so, right about the time I am talking myself into staying in bed. This couple is extremely over-weight. I just don\'t get that lazy twinkie eating shit. I never see a health problem as the cause of their being overweight. To me, you can always work it off. To many times it seems that being over-weight is the cause of the health problem. I just that poor dog didn\'t feel so attached the her owners. She is just going to get eaten anyways.   '),(10301,216833,'06,July,2004','2004-07-06 00:00:00','    This morning I picked up a Met-RX bar and some water for breakfast. It\'s lower-body day in the gym and it almost always kicks me on my ass. It\'s been three weeks since the last squat or leg-press. I feel pretty drained. Last week, i just skipped wednesday in the gym - I think I was busy; but that was the only day missed.  I\'m just screwing around at work uploading files to myplay.com - drawing - checking email and message boards. Not much going on, I set up a laptop for a guy, prepared the backups, and answered some calls about simple operational questions.   I drove by a Jack-n-the-Box and saw like this couple barly pull out of the drive thru and they are stuffing their face. Shit! Get the wrapper off first. Sooner or later someone is going to come out with this idea where people can ask for \"Drive\" instead of \"To Go\"   \"Would you like that burger to Drive, Sir\" \"What\'s the...?\" \"We prepare the sandwich in the wrapper so you can DIVE right in. Would you like to me do that?\" \"Sure...I\'m a fat lazy fucker - just calories I don\'t have to burn.\"  They would neatly fold the wrapper ready to take in hand and eat imediatly. That is just fucking lazy. If that ever happens, it would just be a perfect progression into the American dream. The president would endorse the new complementary Drive wrapper option.  I guess if it goes that far, they should just have people in the restroom to shake and wipe.  Oh well - I am going home tonight to prepare Thank you cards for all the lovely people that brought wedding gifts and helped out with the wedding. I am really glad all that went well.   '),(10302,216833,'06,July,2004','2004-07-06 00:00:00','    Oh my fucking lord - check this article out. Why in all holy hell is TIME covering emo?   Yeah - whatever........   '),(10303,216833,'06,July,2004','2004-07-06 00:00:00','    \"I like too many things and get all confused and hung up running from one falling star to another till I drop. This is the night, what it does to you. I had nothing to offer anybody except my own confusion.\" - Jack Kerouac  That is pretty much what I felt like last night. I have this desire to be in a band. I love all the fun shit - playing and rocking out - merch - websites - mp3s - promotions. All of that, I find enjoyable. I just forget how much time it all takes.   To be in a fucking good band - you have to have mad dedication, commitment, and you gotta find some bad ass mofo\'s to rock out with every night. That being said, I have the upmost fucking respect for guys and gals who stick together, tour their ass\'s off, and rock out every chance they get. I just don\'t have the love for it like once upon a time.   Music is something I love - but it\'s not my only interest. No - I don\'t have a.d.d. I love rocking out - with my friends or in my car even, it something I totally enjoy. I am just working on time management alot and it seems that from 7pm to 12 midnight is alot of time to fuck around. Don\'t get me wrong - I like hanging out and just running around. I just need that strong passion about the music to spend endless hours on writing music that always seems to die in less than 5 years.  I met nick, christy [nick\'s friend], eric and venessa at the veggie garden. I can\'t stand that place. The people freak me out and it just looks dirty. Plus I can\'t stand the whole mock meat thing. I eat alot of junk and I\'m sure its just as unhealthy as synthetic food. I just think that is a little OVER processed. I had tofu and veggies.  We rode out to the practice place south west of downtown I think. I was taking in the totally experience of going out to this place. It was right out of punk rock legends. Posters and flyers all up and down the dank old halls of this place. Looked to me like some old school - there were kids making noise for sure.  It was an event to say the least. I was the fill-in drummer for the night. I was pretty fucking sweaty when we rolled out of there. I was so thirsty and dry\'d up. I went by a closed convience store - got pissed and went home.  I love going home - I sneak into the bedroom around 1AM right after I\'d spend some \'winddown time\' on the patio. mimi just laid there sleeping soundly in one of my button-up shirts. She always looks so comfortable when she is sleeping. I was comfortable too.      Go read a kobo abe book! I am about to pick up \'The Box Man\' and start in on that.   '),(10304,216833,'06,July,2004','2004-07-06 00:00:00','    when i was at the mall on sunday, I show up there around 11 and nothing is fucking open but some food place and the movies. All the other stores don\'t open until 12. I guess I had never been to the mall on sunday morning but here I was with others around me and the store was deserted.   FUCK!  I waited around for what seemed like a long while, I still havn\'t replaced my watch that was stolen in Mexico. I have the hardest time finding clocks on the wall. So I waited - with no idea what time or even when the place opened. I just waited. I was stoned out of my mind - just sitting there on this bench.   Here comes the fucking little kid, and I am telling the truth, this 7-year old boy came and sat by me out of the blue. He said \"Hey Mister I am waiting for this place to open,\" and he pointed the the store directly in front of me. It was a kids photo place and they have some fucking dumb camera kids can play on and eventually destroy. \"I\'m gonna play on that big camera.\"   \"Great..\" I said. I was so out of my mind. I just wanted the piosonous vermin to scrurry away under the arms of some guardian. I hate people some times. Esp. their fucking kids.   They hung around my general area for a while as I waited for this shoe store to open and I could finally get my Diesel shoes. They had the one I wanted - I could see it through the massive scuplture of metal bars enclosing the store.  I get in and they only have a size 8 in each color. the model is discontinued. The life of a super model is sooooo hard.   '),(10305,216833,'06,July,2004','2004-07-06 00:00:00','    well so far today has been pretty average. not to busy and not to dull - so far. i started thinking about how it was in mexico. I had no tv. there wasn\'t a radio station i could understand - I never picked up a paper. I was totally isolated from the happenings of the world. or so it seems. I am so tied up with information that its hard to just look beyond all the \'news\' and see how the world really is. I think they only way to do that is to go to places. I am really looking for adventure most of the time. some sort of new challenge. I will have to visit india some day.   '),(10306,216833,'06,July,2004','2004-07-06 00:00:00','    On Saturday, I was at wal-mart, dropping off the pictures from our trip to mexico and I find this dvd that I want to pick up. It\'s the enitre miniseries called V. Do you remember it at all? I just remember parts of the show freaking me out. My dad used to watch it alot, I remember going up to his office and he had a TV in there. I guess late at night, TV episodes of V is pretty entertaining.   I bought the dvd.  It\'s fucking weird, but I totally got into the story. I didn\'t watch a whole lot of it - just as the aliens were intigrating themselves with the human society. What a trip man, this shit was humorous a few times. The 80\'s are funny on their own.   '),(10307,216833,'06,July,2004','2004-07-06 00:00:00','    Well - I\'m off to try and find some replacement fomr my diseal canvas shoes. No dice yesterday. Mine are so old they don\'t want to stay on my feet anymore. They keep shouting \"Give us a rest - we\'re old! Bail, LEFTY!! jump foot!\"   '),(10308,216833,'06,July,2004','2004-07-06 00:00:00','    I woke up around 10:30AM, not feeling very rested at all. The late night before has throw everything off for me today. Didn\'t really slow things down much this weekend. After throwing on the same pair of shorts and the same t-shirt I was wearing last night., mimi and I scurried off to my office to change out the weekend backup tapes from the servers. After that we drove out to the Ft.Worth Natural History Museum to check out the exhibit and watch thte show in the planetarium. We got there and the parking was a mess - there is some sort of farm animal show out at the will rogers area. The exhibit wasn\'t like I remember\'d. It was small, unorganized and just seemed to be missing alot of things. Must have been another museum, I am just confused. The planetarium was ok, it was targeted at younger crowd for sure, but I still enjoyed the tips for finding navigational stars. That was very informative.  We went from there to a mexican food resturant and shot out to the mall to see if they had any more disael canvas mules for men. no dice. So I took mimi back home - she had to leave for work at 3pm. I went to get some of the pictures from our honeymoon developed. Got em back,   '),(10309,216833,'06,July,2004','2004-07-06 00:00:00','    Well - It is close to 4 A.M. and I just can\'t sleep. After dinner I decided to go see the new Star Wars movie. I thought it was excellent. Lots of Stars and lots of Wars. I was happy to see the movie heading for a familiar trend. I can see where things are going now. I am truely getting tired now. The night is very still. Not much making noise out on the porch. I am sitting out there a few times tonight wondering about things.   Oh well, more weekend later.   '),(10310,216833,'06,July,2004','2004-07-06 00:00:00','    Well - Brian just called, and as I had predicted, he will not be coming up to pick-up the arcade. He had a problem with his check and will not have the funds. That is cool - anyone want an arcade game to go in the casa. CHEAP! I would just wait for another auction. I think there is one coming up on June 22nd. Look out for that. haha. I will probably go home after work, roll a J and play some fucking gta3. I have yet to actually fuck off for a few hours. Keeping the apartment clean is an all day job by its self.   '),(10311,216833,'06,July,2004','2004-07-06 00:00:00','    Ok well - It looks I went with the crowd and got a live journal. I wish I could figure out a way to move all of my spyattack entried over here. Oh well.   I guess I am just going to let the domain name expire and go on.   Brian stopped by last-night to say hello. I guess he was in the area - at least that is what he said. We talked a little while, I got lit, and we checked out the mr.bill DVD that I got as a wedding gift from my boss. It was an odd little DVD.  He want to buy the Time Killers arcade machine for $160. We will see, he is supposed to pick it up tonight and bring me the money. It\'s a cool game, just ready to get it out of the garage, and I wasn\'t into it that much. Just thought I could make some scrilla from the game - oh well. I really wanted Altered Beast anyways. I am at work - a little bored, but I feel accomplished this week. I got alot done at work, I had a good time the other night with a little barbq going on and it\'s soon to be the weekend. a matter of hours.   I have to come up to work tomorrow and change out the tapes for the back up. No big deal - I might run up to school for a while are maybe learn some shit...   '),(10312,216833,'06,July,2004','2004-07-06 00:00:00','    .·.·.·MARRIED MEN ALL WISH THEY WERE STACKED LIKE ME·.·.·.   Well - I wish I could figure out why my comments aren\'t working.  Oh well.  no one reads this anyways - cuz its not a live journal.  For me to archive anyways.  I have a hard time remebering things some time.  Maybe this will help.  I am a married man - happily married to mimi! So now married life starts.  Weird, but it doesn\'t feel that different.   Honey moon was eventful - most memorable recap would be snorkaling, Mayan Ruins, going to the grocery store, and getting robbed.  Some one broke into our rental car on the last day of our trip and stoke my watch and sony handycam.  FUCK! I am still a little bitter about the whole mess.  The wedding was awesome.  I should have some pictures any day.  I had a good time there.  I feel like an old man now.  At least I have the best body ever for a married man.  I love fucking pushing weights around, working out in the gym.  I think that is a good reason I was glad to be back in DFW.   '),(10313,216833,'06,July,2004','2004-07-06 00:00:00','    .·.·.·FAKE TO-DO LISTS AND THE WOMEN WHO CARE·.·.·.   You know your creativity is lost when you look up on a cloudy day and there are just a bunch of clouds floating around.  No fluffy white turtles meandering across the deep blue background - just clouds.  I have been hard pressed for a creative idea, something to build esp. after all the breaking I have been doing.  Being constructive is like a dream to me - a goal I feel I should achieve at all times, but for the most part I do nothing to get me there.  For example, I am at work, I have tons of things to do, but I can\'t really do anything about it here.  So I sit, and grow a farm of hemroids(joke).  I am going out of my mind some times - I swear.    I need to build something - like a bench or something.  I wish this blogbuddy had a thesarus - I would use fancier words.  wink wink ;)   '),(10314,216833,'06,July,2004','2004-07-06 00:00:00','    I guess I will start where I left off. That would be Thursday. I have to say, I wish I hadn\'t gone to work on Thursday. I might have been better off staying in bed.  \"My alarm clock lost power and then I got a cold...\"  Yeah, right. I was up at 6:30am. I had switched \'shifts\' with Robert so I could head out to the lake earlier - that meant I had to be up at the office at 7:30am. No problem. I\'d rather be up anyways.  My morning went really well, slow, but nothing annoying or worth causing a fuss about. I had an email problem in the Alabama office that I was \'brainstorming\' over. I kept it in the back of my mind, something I do alot at the office.  Soon enought it was around that time to head to the gym, strap on some weight gloves and push some iron. Good times but not on this Thursday.  Right off the bat, I saw the foreshadowing of something bad. As I was walking down the hall, this fat, rolled-over, dick of a man named Taylor is standing in the hall looking for his \'mail.\' He has been coming to the gym for about two weeks and just really gets tooled around by his co-workers. He seems to do whatever ridiculous exercise they throw at him - what a pushover. As I walked down the hall, he looked up at me and then back at his mailbox by the mailroom. I start to walk around him and he steps right in my way, without even aying excuse me or anything. It was totally an obvious move to piss me off.  I just mumbled \"JACK ASS\" under my breath. Another weak move by me. I shoulda just laid the guy out right there - but I let it slide right off my back like everything else.  The workout was going pretty smooth, I noticed \'jackass\' had made his way over to the tredmill to do a little mall walking. We had some Oldies radio on the stereo and James, Robert, and myself were working those shoulders hard.  Up pops Talors buddy, Brian (this guy talks bad about his wife and says beer at least once a day in the gym.)  He starts changing the radio station, while at the same time asking James if he doesn\'t mind.  James says \"Yeah, I like that station.\" Brain: \"I can\'t workout to that\" still changing the station  \"Is this alright\" Brian says as he gets closer to the station of his choosing. James: \"I don\'t give a fuck......pause.......but if I come in here tomorrow and you change the station when I asked you not to, I\'m gonna have to fight your fucking ass over it. That\'s bullshit!\"  BAMMMMMMMMMMMMMMMM!!!!!!!!!!!!!!!!!!!!!  The little bitch but it back on the Oldies and went back to his tredmill. All I heard was \"Fuck\" and \"Shit.\" HAHA they were pissed.   Now I hate going into that gym. It was embarrassing to see that totally lack of restrain on negative emotions. I understand James was pissed, and he had a right to be upset. It was just a bad vibe, especially with customers.  Anyways, James was hot for the rest of the day at work. 4pm rolled around and I was out of there. I picked up mimi and we headed out to Lake Grapevine to meet up with James and do some water skiing. I had a blast. Ryan, James oldest son was with us for about 2 hours. I finally got up on the water and actually skiied. Fuck yeah, that was a relief for the Thursday from hell.   I ended up forgetting about dinner and by the time I got home, I was so tired and hungry we headed over to IHOP for a gourmet meal. ha.  Today was pretty chill. There were some adverse effects of James actions in the weight room. The owner of my consulting firm sent out an email to all of MIS about proper etiquette in the gym. haha. James was pissed - time to look for another job he says.  I packed and stuffed and folded everything for the 7\" comp. That is coming to an end.   Looks like mimi didn\'t get that job, havn\'t heard anything from the folks in cali - guess we\'ll just have to keep looking. As for me, I havn\'t had one solid lead.  I bought the Karate Double LP - that has to be my favorite record. Damn Palmer took my CD to Denver with him! FUCK  goodnight  p.s. - arcade auction - Sat morning at the Rodeo center in Mesquite - I will be there at 8am to play some free games.   current mood:  tired current music: Karate   '),(10315,216833,'06,July,2004','2004-07-06 00:00:00','    The gym was extra sweaty at 11:30 today. The A/C has been fucked in that room for about two weeks. There were wet spots on the floor where the leaks from the unit had dripped for weeks. I don\'t really know why it was extra hot today, maybe because there were a few ceiling tiles missing. Hot air from the warehouse must be making its way into the gym.   I am looking forward to this weekend. Another arcade auction on Saturday morning. I will try to attend a show this weekend, people seem upset. I\'m also gonna help nick move into his new castle. Look it\'s wednesday and I am already looking towards the weekend.   I feel weird talking to my parents, something is different now and I don\'t like it.  Tomorrow, I get off at 4pm and I\'m heading to the lake. This time, mimi will be along for the ride. That will help curb the weirdness of riding around in a baby-blue boat. I am so macho!! :)   current mood:  cheerful current music: jawbreaker   '),(10316,216833,'06,July,2004','2004-07-06 00:00:00','    As strange as this may seem. I went water skiing last night with my co-workers, James and Robert. I had never been water skiing before and it was a fucking workout! I am so exhausted today - my hands have blisters on them.   Jame\'s boat is so interesting to look at. He recently bought it in Granberry for cheap. It doesn\'t have the best color combinations. Actually it looks like a baby-blue Barbie? Speed Boat. hahaha. It\'s fast as hell though. He needs to get it painted. Three guys in a baby-blue(with a hint of pink pinstirpes) looks like a pride run on the lake. Whatever.   People were still going \"ooo and ahhh\" over the speed. If it was black and red, it would look like the Bat-Boat from the live action series in the 60\'s.   MiMi had a phone interview with a district and regional manager from the Bay Area. She will find out if she got the job at the end of the week. I am kinda interested in how this will go, so far is sounds very likly that she is a shoe-in for the position. I\'m happy for her. She will be moving out there soon than expected - at least, that is the way it looks.   current mood:  exhausted current music: coldplay   '),(10317,216833,'06,July,2004','2004-07-06 00:00:00','    I understand alot of things latly. I understand why I am where I am. I can see the long road that got me here, I just wish it was as nice for the other travelers as it was for me, but it\'s not. All I can see, looking back, is pot holes created by someone else. My poor brother and sister. How can I mend those holes for them.   I can\'t. I can only love them. Damn the crappy way love turns sour. My parents are a mess. I can\'t seem to deal with it. I just wish a decision can be made and we can all just go on. I really don\'t have much to do with it anymore. What a mess.  Yesterday was pretty fun, I was bored for a little while. MiMi and I washed out cars and I waxed her car when she got home from work. She helped out with that too. It looks nice.   Around 11pm, her friend Cindy came over. Cindy is also manager of a Nine West shoe store. She was MiMi\'s manager before MiMi got her own store. Miss Cindy is super slender and almost 40. She laughs at everything and has some pretty funny stories of her own.  We drove up to Half Price Books off of 75 and Northwest Highway around 11:30pm. They had just closed so there were people still around. We got out and started looking in the dumpsters. Sure enough, my preminitions were correct. A guy came up and asked us what we were doing. I made up some shit about needing box\'s for moving. He watched us get a few boxes and we took off Pissed as hell - we all vowed to come back in a few hours.  We drove down greenville and came back, all the way past university and circled back around towards Cafe Brazil. We stopped and dumped the boxes that we didn\'t need and went in to have some food and drink.  I\'m so not used to being in Dallas, I spend so much time in my neighborhood. Lots of open space in suburbia.   Anyways, we went back to Half Priced Books around 1am and got tons of books. I was paranoid being with two girls. There were lots of gawkers that would slow down to \"check out the chicks\" and that made me nervous. People do crazy shit to get sex and I would have a hard time defending a car full of guys trying to stick my wife and her friend. That shit freaks me out. I would go insane if that ever happened. Anyways, my nerves got the best of my and we left only after digging around for about an hour. I think we all found plenty of shit. I was satisfied.   I saw the security gaurd leave at 1am so just a tip. Go to Half Priced Books after 1am and on the weekends and there will be lots of books for free. This would be an awesome way of starting a punk rock library. Maybe those South Main kids would be interesting in this info.   '),(10318,216833,'06,July,2004','2004-07-06 00:00:00','    I understand alot of things latly. I understand why I am where I am. I can see the long road that got me here, I just wish it was as nice for the other travelers as it was for me, but it\'s not. All I can see, looking back, is pot holes created by someone else. My poor brother and sister. How can I mend those holes for them.   I can\'t. I can only love them. Damn the crappy way love turns sour. My parents are a mess. I can\'t seem to deal with it. I just wish a decision can be made and we can all just go on. I really don\'t have much to do with it anymore. What a mess.  Yesterday was pretty fun, I was bored for a little while. MiMi and I washed out cars and I waxed her car when she got home from work. She helped out with that too. It looks nice.   Around 11pm, her friend Cindy came over. Cindy is also manager of a Nine West shoe store. She was MiMi\'s manager before MiMi got her own store. Miss Cindy is super slender and almost 40. She laughs at everything and has some pretty funny stories of her own.  We drove up to Half Price Books off of 75 and Northwest Highway around 11:30pm. They had just closed so there were people still around. We got out and started looking in the dumpsters. Sure enough, my preminitions were correct. A guy came up and asked us what we were doing. I made up some shit about needing box\'s for moving. He watched us get a few boxes and we took off Pissed as hell - we all vowed to come back in a few hours.  We drove down greenville and came back, all the way past university and circled back around towards Cafe Brazil. We stopped and dumped the boxes that we didn\'t need and went in to have some food and drink.  I\'m so not used to being in Dallas, I spend so much time in my neighborhood. Lots of open space in suburbia.   Anyways, we went back to Half Priced Books around 1am and got tons of books. I was paranoid being with two girls. There were lots of gawkers that would slow down to \"check out the chicks\" and that made me nervous. People do crazy shit to get sex and I would have a hard time defending a car full of guys trying to stick my wife and her friend. That shit freaks me out. I would go insane if that ever happened. Anyways, my nerves got the best of my and we left only after digging around for about an hour. I think we all found plenty of shit. I was satisfied.   I saw the security gaurd leave at 1am so just a tip. Go to Half Priced Books after 1am and on the weekends and there will be lots of books for free. This would be an awesome way of starting a punk rock library. Maybe those South Main kids would be interesting in this info.   '),(10319,216833,'06,July,2004','2004-07-06 00:00:00','    Over the past couple of days things have moved at their normal pace. Nothing extra ordinary has happened. That is fine with me. I watched a few good movies during the weekend, sat around the house most of last night, and have just been going to work at my normal time.  I read a bit of those books I got from the library. Evolution and history. The idea of evolution interests me. Unlike most people who go to public school, I was taught at home during my highschool years and my history and science books were religiously biased. Creation was taught, and the Christian agenda was splattered on every page. I was bright enough to see it, and I just went on. To me, I had to find out on my own both sides of that coin before I made a decision. I still don\'t know enough about evolution to make it totally clear that I could teach someone else. I think once you understand an idea well enough to pass on the information correctly, you know what you are talking about, otherwise, it is best left unearthed. I am feeling a little dull and bored.   '),(10320,216833,'06,July,2004','2004-07-06 00:00:00','    Who sucked out the feeling? Understanding the ultimate human cause for laughter and sobing can be traced back to serveral soruces. Pain, fear, joy; they are all just words. What is the most common thread of life between everthing living, is it that fact that we ARE alive or that we will all die. I have been thinking alot, I have been drinking some too.   I went to the library today and picked up two books on human evolution and a book on the history of mexico. I hope I can find the time to read all three. The darwin book, i dove into it already, seems like some tough reading - interesting at the same time.   The course of my goals has changed over the years, and I guess I can expect it to change a few more times. Hopefully I can stay the course, see the end of the road before I decide to turn down a side street.   Intoxicated by all around me, I only know that who I am is who I make myself out to be. Understandably, that is what faces each of us. For some reasons, I think I shall like who I am very much..   '),(10321,216833,'06,July,2004','2004-07-06 00:00:00','    Oh, a hunting I will go. Job-hunting that is! With cubicles and office chairs...ya\'ll come back now - ya hear!  All of Wendesday and Thursday, my mind as been in one single fucking place. The Move. I have been saturating myself with preperations. I\'m not even sure if it will be worthwhile in the long-run. I do this to myself all the time, with out a break.   I\'ve been looking into colleges, apartments, and jobs, jobs, jobs. Right now, I am looking for something during the night hours. If I can pull that off, I might start back to school earlier than I am planning to. We will see - I\'m super stoked thought. At the same time, a lot of scared.  I\'m starting all over while, at the same time, alot of my friends are starting out brand new, and some are moving up in what they love. Me - I\'m starting over from the start - that\'s fine. I got lots of time I guess. Might as well be busting my ass during that time. The other option would be to make and spend money. Seems worthless and unrewarding but I guess I would have STUFF! Fuck that.  I have a long list of goals. Longer than santa\'s list, i\'m sure.   current mood:  anxious current music: face to face   '),(10322,216833,'06,July,2004','2004-07-06 00:00:00','    I went out to nick\'s place last night to hang out with him, had some awesome conversations, a few laughs, it was a good time. Ate to much food and with the beer on top of that made me sleepy, fast. I headed home and read for about 30 minutes before I crashed out like a tree falling in a forest.  This morning I woke up with a head full of dreams and a face full of crust. The sleep was a hard sleep, but I tossed and turned alot last night. I woke up feeling troubled, maybe it was the dreams. There were two dreams that I remeber well, the second one was in two parts.  The first dream was late in the night, another on of those \"raging jealously\" dreams. I have them often, maybe once a month or so. It always involves mimi screwing some other guy, this time an older man, maybe in his 40s or 50s. It ends with me beating the living shit out of the guy. The dreams are always soo intense that I have woke up crying or yelling profanity. Last night was, \"Fuck this shit.\" MiMi didn\'t even wake up, I suppose that is good, but I wanted to talk about it.   Back to sleep after a few minutes and I have another emotionally real dream. This time it was more invovled and new. MiMi and I had traveled to a small beach area, not really sure where. It\'s a dream so it could be anywhere, the streets were desserted for the most part except for a small fishing shop on the road facing the ocean. We were riding bikes down the road to a small square beach called \"La Gu\" - I remeber looking at a map and noticing that the beach was not far away from another beach called \"La\" - thats all, just \"La\" not L.A. but \"La\". Weird. As we approach the beach, directly inland and across the road from the beach was a mountainous sand dune with two women sitting about halfway up on the dune. They were just talking, mostly about being lesbiens and woman\'s rights and topics of that sort.   Right behind the dune, or just over the top of the dune back towards the fishing shop was a giant house. Old, but well kept and very dark. The color of the outside of the house was dark. I walked in and my father, who happened to look alot like Anthony Hopkins, was standing at the top of the stairs. He was a powerful looking man, who was on his way to another meeting far away. He had just taken in a girl from somewhere, I\'m not totally sure. I think she was to be his wife, or adopted daughter - I can\'t remeber. I just know she had problems with drugs or something - and this is so unlike me, but I felt complete compassion for the girl. She was so lonely and sad. I knew that a little love from someone could turn her life around. So I watched over her, she was around 20, this wasn\'t a little kid.  One night she had disappeared, and I had fear the worse, my father\'s maid was phoning him. He rushed home and I had found her with those girls by the dunes and I rushed her to the hospital. They had released her, so by the time I had made it home, my father was standing there, a grim look on his face, careless, almost evil. He said nothing, turned his back and walked towards a limo on the side of the house. I took her up to my fathers enourmous bed and tucked her in. She was upsetting to look at, not ugly, just in pain, hurt emotionally. I held her and she cried.  The next day there was a show at a local pub, Jake and 10kDrive were in town to play and we were in the club early - planning an evil \"holloween\" style show. We had taken water/oil ballons up in the roof, posted next to an opening in the ceiling to drop on people during the set. There were costumes and ropes and it was gonna be scary. After getting that all set up, Silver Actions showed up and started setting up. They were the fist band, and during their set, Nick\'s older brother(who doesn\'t exsist) came up to me. He looked just like nick, taller and with more facial hair, but just like him. He was telling me that his hoodie was much cooler than Nick\'s and he went on about why it was cooler. Around this time, I woke up. The tossing and turning began.   Weird eh.  I also found out that alot of people around me are taking anti-depressants and I didn\'t even know. I had alot to say about that too but I really think it would just be a waste of time. Good luck to everyone that is having troubles. I wish you well, and her is a hug from me ((((((((((((((((((((you)))))))))))))))))))   '),(10323,216833,'06,July,2004','2004-07-06 00:00:00','    Unlike most weekends around Coffman house, things were on an upswing this weekend. First off on Friday I participated in one of the biggest displays of white-trash I have ever seen. On Saturday, I went grocery shopping, suit shopping and as the hours carried on, I was shopping for a clear thought. Sunday was a day to chill and clean up. I took a break and saw a film - and now it\'s Monday and I am thinking back on all the silly fun shit.  Let me get back to Friday night, I finished up work as fast as I could, I was supposed to meet James at 7pm and I didn\'t want to keep him waiting. I raced home to change my shirt and shoes(ritualistically) and pick up mimi, she went with me. All the while making record time, I pull into James drive-way just under a few minutes from 7. He had two extra tickets to the Craftsman 400 truck series at the Texas Motor Speedway. OH MY FUCKING LORD. I was floored, I was in awe when we pulled into this fucking crazy stadium of wasted energy - I was just so fucking blown away with the display of utter trashiness. Fat - fat -fat -fat everywhere. There were Turkey legs for the classic cavemen, beer by the tons for the idiots, and troglodytes for the mullet-styled fellers. There were wrecked RV\'s, race-car painted pickups that looked more like moving toilets and tons of really nice motorcycles. It was like a trailor park with out the houses. I would hate to be here on NASCAR day when there are some 250,00 fuckers, this was a light event comparatively. I think somewhere around 80,000 race-fans. That\'s is alot of - well alot of fucking idiots for sure.  The race was intense, I have to say that much. The feel of the Roman empire was strong, large arena - danger and possible death!! The cheer of the crowd for their favorite local driver was obvious, the boo\'s when someone gets to close to their hero couldn\'t even scare a fly. I was definatly charges, excited, and pumped up when those trucks would barrel down the side of the track. Oh well - don\'t tell anyone. We left a few laps before the race ended. I just imagine how the traffic would be at the end of the race, it wasn\'t so bad when we left. I have to say, I will never pay for a ticket to see a race, but I\'d go again for free of coarse.  On Saturday MiMi and I woke up, wanting to do something but not sure what, so we decided to go buy some groceries - It was a good oppurtunity to handle that mess. MiMi went off to work around 1pm and I went to the mall to just shop. I was looking for a suit and that was an undertaking in itself. I had an idea of what I wanted - and after 4 different stores and about 14 different suit jackets I think I might have found the one I am looking for. I am supposed to pick it up on Thursday. When I got home after buying a new white dress shirt and some new grey dress slacks I started doing the laundry and cleaning up the house. That went pretty smooth - I didn\'t see anyone till about 10pm - I had a few beers and was feeling pretty warm and happy. Alex and Eric stopped by to play some board games, I think it was called Tribond, I had fun. I pretty much felt like an idiot, but after a while, I got into the groove of thinking - if you can believe that.  Sunday, once again alone, MiMi was off working at her store from 11 - 6 so I finished up the laundry and went to the movies around 1:30. I went and saw the Sum of All Fears and it was alright. I enjoyed it - I havn\'t seen an action movie on the big screen in a while (except star wars - but that is different.) It was a wonderful day, partly cloudy, and just gorgous - I didn\'t stay outside much because of the heat but I was one of those days that makes you smile and say \"Summer is here!\"   I had dinner on the stove when MiMi came home. Some spicy Thai stirfry with a little red curry and yummy thai rice. I dunno the name of my dishes, I just cook what I wanna eat. I also made some chocolate-covered strawberries and boy where they delicious. Watched Proof of Life again with MiMi and ironed all my outfits for the week. I am prepared - bring it on!   '),(10324,216833,'06,July,2004','2004-07-06 00:00:00','    Understandably, things change. For the most part, I accept these changes and move on but at times comfort is the key. I am really starting to get worried about my move - my BIG up and coming life change. One way or another, I will pull it off, but it is an unknown adventure that could crush me.  I didn\'t have enough to eat this morning, that will adversly affect my workout - I will be a big pile of ooze when I\'m through. MiMi is off work today, I am gonna go home at lunch and hang out with her for a while. It\'s cloudy atleast and that makes for a pleasent break from the heat that I will have to just get used to for the next few months.  I havn\'t read a book since I got back from Cancun - I\'m sorta pissed about that. Again, it\'s one of those things that no one can do anything about but me. There are several books that I would love to jump into - it just seems like I spend my free time cooking, cleaning, and stareing at some sorta screen.  I feel like a zombie - it\'s time to stop the EVIL cycle. This happens to me every summer - I just don\'t like to go out for a walk as much as I do in the winter. To me there is nothing inspiring about summer. Winter has this magic about it that makes me think and feel more positive.   This is a re-occuring topic for me, I apologize.  I was hoping to get the bump today but I found out that it looks like it will be a couple of days. bummer. I really need to either get my money back from the bank or get my paycheck so I can pay my rent. UGHH!   '),(10325,216833,'06,July,2004','2004-07-06 00:00:00','    Today, I am in my new office. It\'s not really an office - it\'s just a desk in the computer room. I am taking the place of Robert as the lead Operations guy. I will have a chance to learn new skills as far as SAP and IBM AS/400 but I am pretty much stuck in the computer room. I\'m gonna have to bring a hoodie up in here cuz it\'s cold. I don\'t really have the option of changing the temp. I could just come in her every 20 minutes and check on shit but my phone and computer are in here now and I like to have a home at work. Some small space to call my own. I guess that is the same reason people decorate there work space. You want to feel comfortable.  I like having locked doors that have to be locked at all times. Keeps people from walking in my office without me expecting it. Sometimes I\'m like to play puzzle games and talk to friends on AIM while I bored. I don\'t want that to become an issue if you know what I mean. Now it is less likly to make an impact. Plus, I have a new feeling of security as far as my job goes and that is nice. Everyone seems to want me in this seat, they all feel that my easy-going personality is better suited for this position that the other guy Rob. I guess there have been a few conflicts as far as that goes. I just roll with the punches like I always say. Bring it on, cuz I can always deal.   I can atleast listen to my music at a normal level in here. When I had the post in the hall I was always worried about being to loud. There are so many fans and buzzing electronic noises in here that I have a hard time hearing people on the phone. It\'s not that bad.  I brought my lunch today. Kinda reminds me of going to school. I would always bring my lunch. I just don\'t want to get outside at 1pm to go to my car with the melted steering wheel in the dead heat of the summer just to waste money and gas on some sub-par meal. I think I will be bringing my lunch from now on. Oh well. I think I will like it - I do enjoy going to lunch with the GUYS, but I have a feeling that being in here away from the everyday PC support (I\'m the backup guy now) will seperate my close ties to those guys. Maybe not, I do work out with em every day.  Oh yeah, today we are working on the back. It\'s normally a short workout, because James wife comes in to do arobics with an instructor and he goes to lunch with her afterwards. Her workout is only 30 minutes so we cut it short so he can get out with her. I think I will just stay in there and get some jogging in. I need some of that FAMOUS LATIN STAMINA!   HAHA. Sorry - I was playing more GTA3 last night. The Asian chick was beating up the vato from the cartel and she said that. I watched a little bit of TV and cooked up a big pot of beans. They are sooooo good. I cooked em for about 4 hours and but some chilepowder, cumin, garlic pepper, fresh garlic, one onion, about three tablespoons of butter and three tablespoons of that Jack Daniels bar b q sauce and some salt and black pepper. It turned out nice, but my I got some mad gas. I think something crawled up my ass and died.   I need a stereo with a CD player in here - I have been listening to the radio, AC/DC rules.   '),(10326,216833,'06,July,2004','2004-07-06 00:00:00','    This weekend was pretty full of fun! On friday night I really didn\'t do much. I sat around at home, cooked some spicy mexican style veggies and rice and watched \'The Pledge\' on HBO. Went to bed early.  On Saturday morning, woke up around 9:30am or maybe t was later. After mimi went to work I drove up to school, but they wouldn\'t let me work - they said my time had expired and I was supposed to meet with one of the directors. He stood me up. So I called Alex, he was asleep. I went home and got on my computer and talked to Kevin from Shreveport. He convinced me to come out to veggie garden at 2pm to eat and hang out. I had no money( I do but my account is frozen until I get my money back and my rent paid) but Nick hooked me up and now I owe him lunch. I got home around 3:45 and Brian called. He came over and hung out for a while. mimi came home around 5pm and we just sat around. We were all waiting to go over to Alex\'s casa and hang out. Left my house around 7pm - stopped at sonic and got a soda with quarters that I fished out of the bottom of the Mortal Kombat game.  Got to Alex\'s house and looked at the wedding pictures. They are awesome - THANKS ALEX! I was super stoked out all that - means alot to me to have a friend like him. We watched Brian play video games, played some 2 player Tony Hawk while mimi fell asleep. After all that - it was around 11 and everyone was asking me when Nick was going to call - so we decided to watch one of the many great DVDs that Eric and Alex have. We watched Pi - that was a kick ass movie. It made you think for sure.   I wanted to go dig for books behind the Half Price Books - it was 1am - mimi was tired.  Sunday was ok - we went to Denny\'s for breakfast and it was packed - not pleasent. I played some GTA3 until about noon and got ready to go to Sommer Shockley graduation. We were there for a while, after that mimi had to go to work for a few hours and when she got home, I had dinner ready for her and we watched the simpsons and some of the basketball game. Damn Lakers!   Watched \'The Seige\' on TV last night. That was kinda weird to watch after Sept.11th.   '),(10327,216833,'06,July,2004','2004-07-06 00:00:00','    Lots of packing and studying going on. I was looking for good music on mp3.com around Rohnert Park, CA (where I am moving off too) and I found some good stuff. There is this one shitty metalcore band called the Beverly Beer Belly\'s :) That made me laugh - they pretty much suck. There was some lo-fi stuff that sounded like death cab for cutie or pinback. I would really like to get into something like. Just some good slow/soft heart felt music about deeper things than girls and love. Maybe a celebration of love instead of the heart-break of life. I\'ve had my heart-breaks... I think there is more to everything than that or this or those. All in moteration, even the thoughts.  I am pretty freaked out about the FOG! Everytime I have been to the Central Valley and the Bay Area the FOG is super intense. I have only delt with fog like that in texas maybe a total of three times. There is a whole season of this morning and evening FOG. I should probably learn how to drive in it. There was a 78 car accident in Bakersfield because of the FOG. That is in the southern part of the central valley close to LA but its all the same FOG.  I almost have every fucking piece of paper faxed and overnighted to the apartment complex we are moving into. I hope we qualify becase I would hate to look for a place with just a few days notice. This shit has taken over a week so far. I know we had trouble at the last apartment we lived in. They always gave us hell about shit and then never fixed a damn thing. I hated that place and for some reason I think that I\'m gonna hear about it from these people in California and then have to explain myself to them.  I am getting excited about the upcoming road trip!   current mood: super stoked! current music: edward applebee and jagjaguwar records music   '),(10328,216833,'06,July,2004','2004-07-06 00:00:00','    Under normal circumstances, I would feel bland and ready to kill my self. This week has been nothing but:  6:30am - wake up and get dressed. 7:30am - start my day at work 11:30am- work out in the gym 1pm - go to lunch 4pm - leave from work and drive to school 7:30pm - leave from school on my way home for food and tv 10 or 11 - lay down and go to sleep  This weeks blubs in this routine were as follows:  Monday - same Tues. - same Wedn. - no school/found apartment/went swimming/bought new undies Thurs. - 7:30pm drove to Ft.Worth to install PCs at FWD Ballet office - get home at 2am( I am very tired!) Friday - we will see!!!   I know there are a few things I need to do today. I probably won\'t be going up to school. I hope the guy from Brady, tx(whom purchased the Time Killers arcade) shows up tonight to get his system. Later tonight I will have to move stuff over to nicks for our garage sale in the morning. I need to pick up some signs, also, around 3pm tommorrow I am meeting the guy who bought the Mortal Kombat - I hopefully will have some money this weekend - that will help make the move easier. I sold my bass on ebay. I feel kinda shitty about it - it\'s just stuff. I can live in a box and be ok physically - maybe not mentally. I don\'t like needing stuff. :(  Come to the garage sale and buy some of my old stuff.  Do I have anything going on this Sunday - it doesn\'t look like it. OH! I have to watch my little sister on Saturday night. I need to tell mimi hahahaha.   current mood:  tired current music: Shotmaker - Now and Then    This is where I am moving!  If the circus comes to Rohnert Park, it can send in the clowns, but it will have to keep most of the animals in their cages. In a mostly symbolic gesture, the Rohnert Park City Council passed a law backed by a local animal rights group that prohibits the display of wild and exotic animals such as elephants, lions and tigers for entertainment or amusement. Violators would be fined a thousand dollars and jailed up to six months. Members of Sonoma People for Animal Rights urged the council to pass the law. They claim traveling circus animals are often confined and chained up to 23 hours a day.   '),(10329,216833,'06,July,2004','2004-07-06 00:00:00','    :( I\'m really cold. When you sit still for a few hours in a room where the A/C blows 68 all the time, you tend to get a little cooler than normal.  I went to school last night for about 1.5 hours and I was surprised with the amount of stuff I remembered. I am excited, but at the same time, I can see my self getting drained. Yesterday seemed to drag on forever. I worked my ass off, busy as hell at work. Went to class, came home, washed the dishes, collected the dirty clothes and set up three more eBay Actions.  Check out my eBay Auctions  I really don\'t know what is going on today. Robert is out in L.A. working on moving the Emerson warehouse to a different building. He has to be there to move the network equipment. I was here at 7:30am and was overwhelmed till about 11:30. James went to the doctor, he has only been here for about an hour and a half. I dunno when he will be back, but I am getting pretty hungry. Also, I dunno when I am getting off tonight. I know I won\'t be home until after 8pm becuase I have school. I will just have to wait and see.   current mood:  cold current music: face to face - over it   '),(10330,216833,'06,July,2004','2004-07-06 00:00:00','    Forget the title, it doesn\'t mean anything.  Yesterday was mimi\'s birthday, she turned 21. She went to Ft.Worth and had lunch with my mom and while she was at lunch I prepared her birthday display on the table. Nothing to big, I am watching the pennies, and she knew that. I got her a dozen roses, some chocolate/vanilla fudge, and a charm braclet that she was wanting. I also bought a Royal red velvet cake and she said it was yummy. Oh, and lastly - I got her first \"To My Wife\" Birthday Card. :)  After all that, we went out to Grapevine Lake with James for some speed boat fun. James had bought a wake board, but no one knew how to ride it, so I had to figure it out. It took me about 5 tried to get up, but I finally figured it out. I have to say it is a lot easier than skiing and tons more fun. I dunno if the skateboarding as a kid helped out at all, but I felt totally more comfortable on the board. I\'m not saying I didn\'t royally bust my ass, I landed on my face a few times. Alot of the time! The fun factor out weighted the pain, so it was cool. MiMi just chilled out in the boat, she seems to just enjoy getting out and doing something - I do too.  When we first got out to the lake and taxi\'d out of the \'no wake\' area - it started to pour rain.  I just laughed and said \"Happy Birthday mimi, you gotta have an interesting memory for your 21st birthday..\" The rain lasted at the most 15min - and we were sailing again.  I talked to Ed(owner of EJB Development, Inc) about my move to California - he was super cool about it and said my job was in no way in jepordy over the news. That was some relief. I needed to tell him, just because of long-term plans that were being made at work.  I also talked to the branch manager at TechSkills.com and gave him a background on my situation. He seemed to understand and is going to give me what I want. Another month to finish my CCNA and the books for the CCNP.  Things are looking up YO!   current mood:  pleased   '),(10331,216833,'06,July,2004','2004-07-06 00:00:00','    I got this in my email today | this is about the sample at the begining of track 1  ::Hi there,  United just informed me Friday that they are now listening to every  test  and looking for possible samples. Although this \"sample\" issue is  still in  litigation, United has chosen to cancel all projects that might not be  in  compliance.  Unfortunately, your\'s is one of those projects. I am offering a couple  solutions:  1. If you send me a re-done master with the offending material removed,  Aardvark will re-cut and get new stampers to United at no charge.  2. If you wish, we can pull new stampers from your original cut and  send  them to a different pressing plant. Bill Smith Custom Records is one  of  these plants that is not listening to tests, and from what I can tell,  really doesn\'t care. A new set of stampers would run $40.00 plus  $12.50  shipping/handling. I\'ll waive the $12.50 shipping/handling to help out  in  this situation. United is keeping all stampers. I don\'t believe they  have  the legal right to do so, but fighting them would be very problematic,  waste a lot of time, and cost more in legal fees than the simple $40.00  for  replacing the stampers. I talked with Beth at United, and they haven\'t  made a decision about labels that they have printed. I would think  that  they should release them as there is nothing on the paper labels that  is in  violation of copyright. United is going to charge all work that they  have  done to get the project to the stage where they discovered the  \"problem\".  There are several pressing plants in this country. At this point,  United  is the only one that I\'m aware of that has adopted this policy. I hear  that Rainbo Records in California is now looking at this policy as  well, so  they may be next.  Let me know what direction to take, and we\'ll work together to get this  solved. ::   Another thing | I just found out that my record bags were delivered on July 10th - I had NO fucking NOTE! I could have taken care of this a week ago. FUCK!   '),(10332,216833,'06,July,2004','2004-07-06 00:00:00','    Over the past couple of nights I have met lots of interesting people and went to strange places. The dream world is so fucked up and intense. I met a track star/turned serial killer last night. It felt so real when she ran by me as super speeds. I could feel the wind push me back; in the background, my friends screaming at me to move out of the way.  She ran from one end of the hall to the other. stopping at the end, just enough to get a look at her - and then she zipped away.  What a weird mess - no one cares about dreams or even really wants to hear it.  Yesterday I went out to Burger\'s Lake - my mom invited me. Little did I know, it was a CHURCH event. I let me disgust be known with my brother and mom that I was not aware that this was a church fuction. Luckily there was no type of service and I really didn\'t even talk to anyone from the church.   For the most part, I had fun. The place looked extremly crowded but I was not bothered much by anyone. I did almost get a paddle in the head from one of the \'rowing\' life gaurds. I voiced my distasted with his ability to row a boat - he seemed more afraid than most people I mouth off too. That was an ego boost.   I also did a little snotty remark to a large mexican man - he was nice about it. He was yelling out to his kid not to be S\'CARD and I mocked him and he politely corrected it by saying SCARED. I was actually surprised he heard me.   Today has been about the same as any work day. I called a few apartment complex\'s in Santa Rosa, Ca and tried to weasel a Move In Special out of the nice ladies. No dice. Three weeks before we load up the truck and take a very long road trip.   current mood:  silly   '),(10333,216833,'06,July,2004','2004-07-06 00:00:00','    Time flys by when you are having fun. I guess I am having fun - I can\'t really tell. Since my last post over a week ago, I have been packing my ass off. Looking for jobs in nor cal, and hunting for a new place to call home.  I really have to much to say to even go into detail. I wish I had the desire to type it all out, but when you miss a few days on this - retracing the meaning shit seems like a waste. I will try and keep a better record so I can look back and remember.  California or Bust....   current mood:  amused current music: jealous sound - anxious arms   '),(10334,216833,'06,July,2004','2004-07-06 00:00:00','    Alot as happened in the last 24 hours. MiMi called me last night around this same time to let me know that she got the job at the Nine West outlet store in Petaluma, CA. That means I am moving soon. She starts work there on Aug. 12th - so I will have to take some time off and drive out to nor cal and get her all settled in.   Until I can find a job out there, I am going to have to hang tight back in the DFW. That\'s fine, I\'m in no hurry. I want to get a decent paying job before I tackle the great unknown. Sad to say, I wasn\'t ready to move anything or anyone out to nor cal this soon. I just wasn\'t prepared as far as money goes. I was planning on a few more months of savings so I can make it smooth. Well - life doesn\'t EVER go as planned and damn if I hate that so much.  I had to give my 30 day notice today at the apartment complex. I am not going to be able to pay two rents - one here and in cali. Nick said I could stay at his house until I found another job, so that is cool. I am hoping not to be there for to long. There are TONS of networking positions all over the Bay area - I don\'t see a big turn down around there.   I am planning on flying out there mid-august to carry out some interviews. This is all happening so fast. I am almost overwhelmed. No big deal - just roll with the punches.  Tonight, I am going to the Houston area to spend some time with my grandparents. We are gonna go down to the beach and hang out - shoot some fireworks and have a good ol\' time. hahaha. :)   current mood:  excited current music: karate   '),(10335,216833,'06,July,2004','2004-07-06 00:00:00','    Do you have a problem with ass-grabbing? Only you can answer that question for yourself. However, taking the following quiz may help to put your relationship to ass-grabbing in perspective for you. If you end up answering \"yes\" to three or more questions, you may want to take a good look how your life is affected by ass-grabbing.  1. Have you missed classes or work because of ass-grabbing?  2. Do you have trouble refusing ass-grabbing?  3. Do you need ass-grabbing in order to have fun at a party?  4. Do you use ass-grabbing to build up your self-confidence?  5. Do you use ass-grabbing to help you relax?  6. Have you tried to give up ass-grabbing and failed?  7. Do you crave ass-grabbing as soon as you wake up?  8. Do you get into trouble because of ass-grabbing?  9. Do you crave ass-grabbing at a definite time daily?  10.Do you lie to others about how often you partake in ass-grabbing?  11.Have you gotten into financial difficulties because of ass-grabbing?  12.Do you often wish people would just mind their own business about you and ass-grabbing?   Remember, there are people who can help you control ass-grabbing, instead of ass-grabbing controlling you.  There is a few fun things here http://www.brunching.com/categories/toys.html   '),(10336,216833,'06,July,2004','2004-07-06 00:00:00','    I\'m pretty tired this morning - worked late - stayed up late. I couldn\'t stop thinking about things as I laid in bed. Those little wheels in my head were going everywhere and I don\'t even remember what was on my mind.  Eric, nick, alex, and vanessa came to my apartment last night around 12 to pick up eric\'s bass amps that were stored in my garage. That makes a little more room in there. I seriously need to take a fine comb and clean that garage up. It\'s a fucking mess from hell.   My Projects (unfinished):  - Restore Mortal Kombat arcade system. - Restore 1979 Vespa Rally 200 and purchase engine for Vespa p-200 - Clean garage - Finish numbering and stuffing 7\" records (order poly sleeves) - Make 10k Drive buttons before July 14th (order button pieces) - Sale Time Killers arcade system - Buy mimi birthday present (july 17th) - Find a job in California - Quit smoking pot - Drink more water everyday - Lose the extra fat around my belly - ETC...............  Other than the long and drawn out project of planning the rest of my life - those are just some temporary ones. Maybe that was the mess on my mid last night - I dunno.  I just got the bump and that is good. I can\'t remember many months where it has happened this early. The bad part is, i get it this month on the 2nd - next month it might be the 9th and that is an extra week to wait. I\'m looking forward to a change in the way my job goes. I need to get a handle on these fucking taxes - I\'m so lost.   '),(10337,216833,'06,July,2004','2004-07-06 00:00:00','    I had a pretty rioted filled weekend. Saturday, I napped all day, it was a mellow day - mimi was off work and we lazed around the apartment.  I went out to nick\'s apartment around 11:30pm on Saturday night and had some beer drinking fun with those fellas. Lots of fun people were there - we talked and jumped around for a while - watched two movies (falling down and the attic expeditions.) Very good movies, on the second movie, I had finished off more beer than I needed to and stumbled awkwardly around the living room. Good thing I didn\'t fall into a window.  After the movies, most of the people left. nick, jason, and i stayed out on the porch till close to 5am talking about our ideas about god and christianity and every other deity-centric cult.  The next day, or afternoon, we all went to denny\'s - had some food and started to move nick\'s large pieces of belongings to his new house in dallas. We go done with all that around 6:30pm and I was ready to get in the shower. I sweated so much - damn beer.  Got home, cleaned up and went to deliver the 7\" comp to the ...Of Death kids and Terms of Grey. I was so tired and hungry, I just stopped by, said hello to a few people i recognized, and took off to get some food with mimi.  After dinner, my wife and I went home and crashed out. What a tired soul I was....   current mood: working current music: pinback   '),(10338,216833,'06,July,2004','2004-07-06 00:00:00','    well - that interview sucked. maybe I am just being negative. We will see.   10:37p  yeah it musta sucked pretty hard because I was turned down for the job! I did pass a practice test today!! yea!   '),(10339,216833,'06,July,2004','2004-07-06 00:00:00','    Yesterday was pretty exciting. I did a phone interview with a Perot Systems recruiter for a Technical Support Analyst job. It sounds like exactly what I am doing now but more Novell. I\'m stoked, she passed my resume on to the hiring manager and I am supposed to talk with a team of peers on Monday for a Technical Screen.  The position is located in Napa, CA. Home of wine and beatiful country side. Plus it\'s not as far as San Francisco. I hope everything goes well with this becuase I could really move around in a company that size. The client that I would be working with is the largest hospital in Napa County. The Queen of the Valley  That will be interesting.   I am hitting the books heavy. I want to do something fun. Metal shows suck these days.   '),(10340,216833,'06,July,2004','2004-07-06 00:00:00','    Last night I had the worst sleep experience in months! THANX - I guess it is still 10x better than sleeping in my car. Car sleep sucks for sure. Once, when I ran away from home I slept in my truck in a parking lot. The police ran a scan on my plates. I know this because mimi\'s dad happen to hear my name on the police scanner. They let me sleep. The night after that, I slept out in this lot in the country on piece of carpet I had in my bed for the drums to sit on. I just laid there listening to the rednecks rev their engines for fun late into the night. I had some many bugs just hanging out on my hear, male mesquitos i think, because they would buzz in and sit right on my ear. I would swap but they just came right back. I wonder why.  Anyways, today I am beat. I wish I would have stayed at the show - but I think I just might be even more tired. FFFFFFFFFUUUUUUUUUUUUUUCCCCCCCCCK! How can you sleep at night with that shit!   FUCK AUGUST! I seriously hate this month and everything it stands for - SUMMER!   '),(10341,216833,'06,July,2004','2004-07-06 00:00:00','    I am so glad that this week is over. It\'s a relief to know that you made it through something difficult. I have worked my ass off all week at school and work, had a little fun when I could and slept as little as possible. It was fun. The plans for the weekend are work even harder at school and rest.   I went to the store to buy mimi a present, I guess it\'s a \"I miss you\" present. I got my this awesome T-Shirt that says California on it. It\'s looks all old and shiz. It\'s just a shirt, who cares, right? I also went and got a haircut - I look super sexy now. At least I feel cleaner, makes for a nice evening with a bottle of wine and my awesome self.  I think I am going to finally do something with my new domian name. www.low-c.com is now going to be a new line of artist screen printed t-shirts made by hand by John Coffman. What do you think? I like making shirts and I think it would be cool to offer an alternative to mall wear. It will be totally Hosssome. I am going to get alex to help me set up some brand mess and come up with a logo. Look for the first shirts in the winter. I think I am going to try and learn flash and make this my first flash shoot. We\'ll see. I have to do something productive before I start school, and working on network equipment is not productive.  I love you all, even al and lacey.   current mood: artistic current music: pele   '),(10342,216833,'06,July,2004','2004-07-06 00:00:00','    I am high on doubt these days. I don\'t think I can take much more of this homeless feeling. I have to stick it out and just deal with these alien feelings but it is just so uncomfortably agonizing to be on the edge of a cliff waiting for the bridge to be built so I can attend to business on the other side.  My eyes are clearly obstucted to the original problem. I am the constructor of this \'bridge\' - I am to blame. Dammit. I am looking forward to the weekend. I can hit the books and spend the seed of my resume as far as I can email AND fax.  Yesterday evening I went out to my little sisters birthday party - she seemed to have tons of fun. The drive over there was some were around an hour longer than normal thanx to the speeding idiots and rain. By the way, the two spell disaster.  I was so beat on arrival that I feel straight to sleep the instant I laid on a bed.   >>>>>>>>>>>>BUUUUUUZZZZZZ  I jumped right up and a panic washes over me like I just realized I am terribly late for work - It was just my mom\'s egg timer. That was a weird feeling.  So I head back to A&E\'s and surf around a bit and slam a big zzzzSLUMBERger.zzzzzzz Ymmmmmm!   current mood:  energetic current music: jawbreaker - etc. rockin\' the B-sides!   '),(10343,216833,'06,July,2004','2004-07-06 00:00:00','    I am having a most difficult time staying awake today. I didn\'t get very much sleep last night because of fucking heartburn. I woke up at 3,4,5am tossing and turning with pain. It\'s my fault of course. I had a little more wine than I should have had and a little more spicy mac\'n\'cheese than I really should have eaten.   At 6am all three of my alarm clocks went off, it was that unrelenting reminder that my day has begun, and like most of the time, I am not quite finished with what I was doing. The instant the TV came on, Betty the cat walks over to me and starts talking to me. It\'s either \"get up\" or \"turn that shit off.\" I can\'t tell exactly which, but she sounds pissed at me.   I really should be working, I have a long list of problems and tasks that I need to attend to today - not including my personal business of going through job listings and faxing/emailing resumes. James went to see the doc so I am stuck here to work on all this shit alone. It\'s is just a little more than normal.  Last night I went with eric to see: xfilesx - deadfood - think i care - and some other band from lawerence, ks(who tours in a short bus) It was nice to see people I hadn\'t seen in a while: james delgado, josh brown, and nick and darius. I personally didn\'t get into any of the music that night, it just wasn\'t my thing really. I thought it was sort of bland - but I\'m just an ass hole.   I didn\'t go to school last night - I need to get on the ball with that and knock it out. That is probably the best thing I can do right now as far as working towards the new cali job. Tonight is my little sisters birthday, so I am going out to Ft. Worth around 4pm to hang out with them. Maybe I will take a nap.  peace   current mood:  tired current music: me yawning   '),(10344,216833,'06,July,2004','2004-07-06 00:00:00','    So I am all situated in the livingroom of Alex and Eric\'s apartment. Things are going ok. I am just looking for jobs like mad, trying to get this school mess done with and hopefully find a job in California. I really want to get out there with MiMi. I am starting to miss her terrible. Esp. since I feel like the odd man now. Alex and Eric are all chilling with their ladies and I am alone with a cat and ferret. :) It\'s cool. It\'s only for a short amount of time. I am going to go to a show at Nick\'s house tonight and yell profanity and take a poop in his yard. Should be a mess of fun.  Things are going well, I am really itching to get back to school - hopefully this next spring 2003 I can enroll and start with that mess.  Peace   current mood: accomplished current music: promise ring - wood water   '),(10345,216833,'06,July,2004','2004-07-06 00:00:00','    Well, things have turned out all sorts of ways for me. Right now I am sitting in the airport in SanJose. I just said goodbye to my wife for a few weeks and it was alot harder than I thought. I am fighting off the tears right now. I am probably pouring my heart out into this little internet-station while people look over my shoulder.   Things are lots different now. There have been some great disappointments and some awesome adventures. The landscape is AMAZING. The weather is great, the people are so nice it scares me. I have alot to tell my friends about. Some things make me so happy and upset at the same time. I have not had alot of sleep and I can\'t seem to keep a natural handle on my emotions. I have seven minutes left on my web access so I am going to check some message boards.   '),(10346,216833,'06,July,2004','2004-07-06 00:00:00','    8:01a - .::MOVE DAY::.  It has finally come. The day I have been staring at for weeks. I hope I am ready to make the trip. The way I feel right now(really tired) is not gonna be any better after I get off work. I\'m pretty sure this whole trip will be wrapped in a blanket of fatigue.  I picked up the truck yesterday with my good friend eric. He drove me out there and followed me back to my house. The thing is not nearly as massive as I first thought. It\'s brand new, 2002 model, and that comforts me. I have almost everything loaded except for my bed, computer, TV, and misc. stuff. I think their should be plenty of room in the truck for Leatha\'s stuff. I filled it up though. Alot more shit in there than I thought.   Right now I am talking to two condo landlords in Navato and Santa Rosa. The guy in Santa Rosa sounds like a good guy. He is going to call me around 12noon PST and let me know if we are in. It\'s smaller and lesser of everything, but I am willing to give up things just to have a little security. It\'s nice to have a place to live, mosdef.  I have soooo much to do today. It\'s mind-blowing.   current mood:  thirsty current music: jawbreaker - unfun   '),(10347,216833,'06,July,2004','2004-07-06 00:00:00','    On my way to get some food I saw dustin drive down my street. I honked and waved. I never see anyone I know in my neighborhood. I went to Wal-mart to buy a newspaper, some shades, and some 409 to clean up the place. That took longer than it should, what else is new.  I stopped at Subway and got some food - went home and picked up Precious. I took her to get a rabie shot. That was pretty easy, when I was waiting on the paper work - I put her in the back seat and turned on the car so she wouldn\'t get hot. When I walk back to the car, she is taking a SHIT in my back seat. HOLY FUCKING SHIT.   I drove to a dumpster and scoped it all out, now I gotta go scrub out the soiled spot. EWW!  Poor pupp gets all nervous after shots, she is famous for shitting in the vet\'s office. I wish that was true in this case.   current mood:  sick   '),(10348,216833,'06,July,2004','2004-07-06 00:00:00','    I don\'t think any of my friends get up before noon on the weekends... I guess I never stay up late enough to justify sleep that late. I went to the movies last night with John from A Faith Called Chaos. That guy is totally awesome, he reminds me of someone but I can\'t pin it down. We went and saw signs and it was alright. It was more about the battle with faith and the ideas of fear and hope than the aliens. The aliens were pretty freaky but I think they stole some ideas from spiderman. The poison dispenser and the shoot with the creepy fucker on the barn roof. That shoot and the footage from the brazilian birthday party were pretty creepy.  The theatere was fucking packed, hot, and the side speakers weren\'t working. I was seriously sweating all over.   Fuck, I have a headache. My mom is coming over today to help me clean up the apartment. That shouldn\'t take too long. I think I\'m gonna go to subway and get some food.   current mood:  complacent current music: the radio sucks   '),(10349,216833,'06,July,2004','2004-07-06 00:00:00','    I feel pretty shit after hearing the news yesterday that we didn\'t qualify for the apartment we were working towards. I had spent 2weeks with this stupid complex, overnighting several documents - over and over. I\'m just really bummed out about it. I have a few applications out with other condo\'s and rental houses. I really hope this goes well. I feel sick.  I didn\'t go up to school yesterday or today. I just don\'t have it in me to think about studying or stupid cisco shit. Right now I need a place for my wife to live. :(  I washed Precious today, she was starting to smell like a big bag of Fritos. haha. And that doesn\'t make me very hungry. She is all super clean now. I even went to far and got naked and turned on the shower. I could pick her up easy and just hold her under the spray. It was easier to rinse her for sure.  I dunno what is going to happen. I really freaked out yesterday, i had two corona\'s and that calmed me down. After all the kicking and shouting I was pretty emotionally drained and super fucking slumped so i went to bed at 8:30pm. 12hours later I am up - i don\'t feel better.   !!!!!! F U C K !!!!!!!   current mood:  drained current music: nothing - my speakers are packed   '),(10350,216833,'06,July,2004','2004-07-06 00:00:00','    After flaking out and not doing anything about it all week, I got back into the gym today. It felt good to get in there and do something productive - I have been eating like shit lately. Junk food, sodas, pizza, you name it. I just get lazy, the lazier I am the more I feel American, and fuck a lot of that! Actually, yesterday I put off eating until about 10pm. I was just to busy with thoughts and stuff to even think about it.  Still no word from the mountain shadows. I am pretty sure we will find out tomorrow. It sucks because, if this falls through I have to hurry to round up a backup plan.  I got a new cell phone in the mail. My home phone is disconnected and this will be my new and only way of contact for a while. It\'s a super tiny Nokia 8265 but mine is black and it looks all scary. Like some little evil bug - how foreign!  I LOVE ALL OF YOU!   current mood:  cynical current music: The Album Leaf   '),(10351,216833,'06,July,2004','2004-07-06 00:00:00','    Well, my check engine light came on this morning, thought this was a good time to get my oil changed and find out what this problem was. Found out, O2 sensor is reading low voltage, so I have to take it to a shop and get it fixed. If they replaced the 02 sensor it will cost me around $300. Now, where can I find $300 bucks.   FUCK!   '),(10352,216833,'06,July,2004','2004-07-06 00:00:00','    well - i gave my two weeks notice this week and my last day in dallas is Oct. 9th   '),(10353,216833,'06,July,2004','2004-07-06 00:00:00','    Since things are going SOOOO well , I think I might make an entry.   Let\'s see...  I feel tired, bored, drained, and super depressed. FUCK!  I was thinking today about the point in evolution where clothes were neccesarry. I wonder if we were ever really hairy, like the ape-men we see in books. Most remains of the australopithecus africanus and homo habilis are fragments of skull, teeth, and other small bones. I am very curious as to how they make the connections from skulls.  They do resembly apes, but they walk upright. homo habilis and the others were just walking apes, hence the hair. I have been looking at people differently.  I\'m sitting in traffic and all I see is a bunch of smart-apes driving machines. Fuck, that is weird.   '),(10354,216833,'06,July,2004','2004-07-06 00:00:00','    I was shot last night. I was shot right in the back by my best friend. It only hurt for a little while.   '),(10355,216833,'06,July,2004','2004-07-06 00:00:00','    Well, my goal to abstain from TV and radio was short lived by my alarm clock.  The morning news. Every morning it comes on and I hate it.   '),(10356,216833,'06,July,2004','2004-07-06 00:00:00','    So, its has been a year since all that shit happened on sept. 11  I;m not to trilled about the comemoration of the dead.  Yeah, it sucks.  I am totally not watching any TV tomorrow and/or radio. I\'m just not going to do it.   '),(10357,216833,'06,July,2004','2004-07-06 00:00:00','    I think sometimes.  I feel alot more.  Blah! I wish I could go home...   '),(10358,216833,'06,July,2004','2004-07-06 00:00:00','    Yesterday was my first official day, before I went to work. I found some jobs to send resumes to. Today, I am doing the same.  This job is seriously boring as hell. I tried to sale to three different people and didn\'t close it.  I really want to quit, at the same time, I don\'t want to give up.  Yesterday was my birthday and I was so happy when I got home. Mimi bought me the new Jets to Brazil CD and it is just great. She bought herself the new Tori Amos and happened to pick up a limited edition copy. The packaging on this is very very cool. It came in a box with a little charm, a map, a DVD, a little accordian like paper with poloroid pics of Tori, and the CD and a little book of art stickers. Very cool.  We went to Adel\'s for dinner, it\'s like a super old Denny\'s style cafe. It was fun and good and the people were super nice. I slept like a rock...i was so tired from standing around doing nothing all day.   current mood:  weird current music: jets to brazil - perfecting lonliness   '),(10359,216833,'06,July,2004','2004-07-06 00:00:00','    I got the job at Zale\'s. Blah, it is really gonna suck. It totally reminds me of the times I worked at Whataburger. You have to count shit, push people to buy crap.  I really hope I can land that sweet job in some IT department. I don\'t think I can deal with the general population for very long.  I felt super depressed last night. I felt super fucking shitty when I saw that I have to work 12:30-8:30 for like the next four days and I came to the realization that this is bad. I know I have a few friends that are worse off, but I just can\'t seem to just be OK with it. Not very many people stroll in so it\'s like just fucking around when you are bored. I need the internet or something.  I have to go study diamonds, so I know what the fuck i am talking about.   '),(10360,216833,'06,July,2004','2004-07-06 00:00:00','    I spent most of the day with Leatha, my mother-in-law. I sat in her room and tried to figure out her computer problems. I sat down around 10:30 and didn\'t leave until around 4pm. I pretty much wasted the day, but I had some good conversation.  I watched a bit of SNL with mimi, made some vegi-frito pie and now I am here.   Till next time, watch for deer.  peace  john   '),(10361,216833,'06,July,2004','2004-07-06 00:00:00','    Last night I watched a bit of the World series before heading out to the Karate show at the \'Bottom of the Hill.\'  This was a very rad venue, I had an excellent time. I was going to review the first two bands but instead I will just say - the first band sucked enough to make me laugh and the second band sounded like incubus.  Karate rocked my socks off. Mimi was feeling a little sick so we left before the end, I will have to see them again for sure. Oh, I just love listening to some good quality music with talent.  Seems that more and more bands come around sounding trashing and fast or catchy and they miss the main point of good music = TALENT!  So on my way home, I had a little trouble finding my way out of frisco, no big deal. I\'m new at this. So we make it over the golden gate bridge and we are driving through the hills just north of the bridge when I have a fucking accident...  Four cars including mine are heading up over this hill around 65mph. two of them are about three cars ahead of me and one is behind me on my left side. As we come up the hill, I see two deer in the middle of the road: one buck and one doe. So i begin to stop, the car on the right spooks them to the left and the car on the left spooks them right back in my way. Well, the doe makes it out of the way, but I fucking slam right into the buck with the front left corner of my car. His head slams onto my hood leaving a nice antler dent. The rest of his body wraps around the front left fender and leaves a nice fucking dent in my door and worse on the fender. Also, my fucking headlight is cracked deep. I pulled off the road to take a look at the car, the deer were quickly out of sight and back up in the hills. FUCK! my door doesn\'t open all the way now, I am gonna try and do some body work.  My luck just keeps getting better.   current mood:  blah current music: get up kids   '),(10362,216833,'06,July,2004','2004-07-06 00:00:00','    My interview with Zales went well. I think I pretty much have the job, I have to come back for a second interview on Monday. I also have an interview with a staffing agency on Monday morning.   I feel pretty up and down today.  I am going to see Karate tonight and hopefully pick up the new Jets to Brazil CD. I downloaded a couple of mp3\'s from Jadetree and it\'s HOTTT.  Tomorrow mimi is off so we are going to the Russian River to watch the salmon swim up stream. That should be nice. I guess I will continue to search for a job where I can find my place again. Fuck, I can\'t wait to start school.   current mood: up and down current music: death cab for cutie   '),(10363,216833,'06,July,2004','2004-07-06 00:00:00','    My apartment is still very disheveled but I will deal with it. There has been a lot going on lately so I sure myself can understand.  On Friday morning, I woke up and fixed breakfast while mimi gathered some clothes to be washed at her dad\'s house in lodi. We had to return the Welsh Corgi that was found to the rightful owners later that afternoon. We leave sometime around 11 or 12 and make our way across the Napa Valley and over the Sacramento River delta arriving a couple of hours later. Most of that afternoon we goof \'d around. Mimi and I threw a Frisbee and played with the chickens and other animals. I found a nest full of eggs and road around on a little razor scooter. Darrell(mimi\'s dad) was waiting on a dish TV guy to come and install their new dish system so we all drove up the road to check out a salmon farm but ended up just tooling around the neighborhood for about 15 minutes and driving back. During that time, the lady came to pick up her dog and she left us 4 - $10 gift cards to chili\'s. Not bad. Darrell had mention that he wanted to get a large hole in his yard filled so I offered to help. As soon as we arrived back at his house, the dish guy pulls up and I run to grab a shovel. I have been feeling so unproductive this past week that I was really itching to get some work done. I fill up the hole with a little help from mimi in about 15 minutes and got a nice blister on my hand and a sore-as-hell back in the morning. It felt fucking good to use some muscle for a change. It felt good to accomplish something. I think I like yard work. We decided to stay the night so I had a few beers and washed some clothes. Jeremy went over to a friends house so that was the last I saw of him during the weekend. The next day I was up around 9am, I found some cereal in the closet and got the day started with a little sugar and corn. I sat around waiting for Darrell to get his ass out of bed so we could start digging this trench for his new telephone line.  We all had plans to go to this Portuguese Bull Fight that day, and I knew it started at 3pm. I dicked around some more outside, enjoying the semi-country sounds and watching tons of black birds devour a field of grapes that can\'t be sold. Finally Darrell is up, and he is sitting in the living room talking to mimi and having some coffee. He had told mimi that he was going to put off the trench till Sunday but because I was on his ass about it, he would start it today. GOOD! So after fucking around for nearly an hour, he finally gets outside and we start on the trench. We had an hour before we were to leave so we got a little bit done. I felt very disappointed that the morning had wasted away. Seize the day! The bullfights were amazing. I was a little horrified with the treatment of the animals but the amount of real fucking culture and people and sounds was like a dream.  The music was so amazing, these were real people living their lives like generations before them. I felt like I was in Portugal, the matadors were a sight! Their outfits and manners were so arrogant and proud. I was enamored by the whole event.  After that we went to a small restaurant and eat some REALLY good food. I didn\'t have to pay for anything, doing a little work for people is always a nice trade. We left that night and made it home just in time to watch some SNL and fall asleep. That night we both slept so well that neither of us woke up till around 11 and that doesn\'t happen often. Mimi had to run some weekly recap at her store, so I went with here and chatted with my dad on the phone while she worked. I learned that a good friend of my mom\'s had died the night before and just a few days before that, her uncle died. It\'s weird how these things come in multiples.  So any ways, we head over to chili\'s after that to cash in our gift cards for some bean burgers and boy were they good! The chili\'s was right next to a Home Depot so we ran in and tried to find an air filter for the A/C unit at the condo, no luck, mimi was unsure on the size. I did apply for a job there, it was all on a computer and it took me about 45 minutes to complete. The questions where awful.  When we went home, we crashed out for about 30 minutes and then both agreed to \'groom\' our puppy. That took a few hours but we trimmed all of her hair, her nails, and bathed her and treated her for fleas. We found a few while trimming her. She hated us. ha ha. Now she looks and smells a lot better.  Today, I spent most of the day filling out applications. I applied at Zales, Ultra jewelry outlet, VanHuesen, CompUSA, Peir1imports, and Boarders book store. I have interviews already set up with the first three. So it looks like John is going into retail for the time being. I will do whatever it takes, I am sure as hell not scared of lowering my standards. I am still looking for a tech job but it\'s pretty thin.  peace out.   current mood: accomplished current music: promise ring   '),(10364,216833,'06,July,2004','2004-07-06 00:00:00','    Day 5: I have spent a weekend and a few days in california so far and everything is going just fine. Still no job bites but they will come, so I am not to worried about it at the moment. I finally got my DSL connection up and running so I am stoked about that. Should make the job hunt a little quicker. I was getting kicked off every 10 minutes or so from this dial-up shit.  I have three computers I need to fix up for mimi\'s dad by friday. We are going back out to lodi to wash some clothes and return our new puppy back to the real owners. Our lost little welsh corgie ran away from home, mimi\'s dad found her and mimi brought her here. Now the \'real\' owners put an ad in the paper looking for their lost dog. Her name is Minnie Mouse and she belongs to a little 4 year old girl. awwwwwwww. I will miss the sweet pup.   I need to keep this up to date.   current mood: busy current music: mazzy star   '),(10365,216833,'06,July,2004','2004-07-06 00:00:00','    Goodbye Texas, I will miss you. I will miss those big orange sunsets, the rainy weeks, and beautiful fall weather. I will cherish all the mud and dirt I caked myself in as a kid.  I feel totally devoted to my home state, the wonderous lands and animals and people. I have had some of the best times here, times I will never forget.  This is such a big move for me, I am very excited to see another plan close to completion. Some things are not so easily planned. Often times, trouble steps in the way, but so far I\'ve made it through just fine. It\'s is my last day at work, I have worked at EJB for over 4 years and it\'s so weird to think that I have been spending Mon. - Fri. at the same place for FOUR years. I will miss all the guys here, they are all great in some way or another. Most of all, I will miss my friends. Each and everyone of them in a special way. I have a feeling some tears will surface today, it will be tough to keep my emotions under wraps.  I am going to try and get off around 4pm and tie up a few things before heading over to Nick\'s House. After that, we will load up his stuff and head out into the sunset, like an old western movie.  It\'s going to be a long trip, I\'m already super tired. I need to pick up some vivrin and redbull or something. It\'s gonna be a late one. I\'ll miss my friends.  My heart aches, but the ties are not severed and the contact will continue.  Wish me luck, I will need it...   '),(10366,216833,'06,July,2004','2004-07-06 00:00:00','    Yesterday, I had to keep telling myself - \"Just two more days and then you are off. You can make it through.\" I met a pretty cool guy yesterday, he came in the store looking for diamonds for his new bride. He was wearing a Hot Water Music shirt so I started talking to him about all kinds of stuff. Maybe I will run into him again. His name was John and he was mosdef on the level. I should have invited him out for a beer, unless he was EDGE.  I started complaining about the lack of inspiration and intelligent conversation. Grant it, I talk with mimi all the time about stuff and that is perfectly satisfying for the most part, but I could use some connections. Obviously the people I work with now cannot put me in contact with anyone hiring IT pros.  We are getting a new manager at Zales, his name is Dwight and he is a fucking dork ass mofo. He is moving from Las Vegas to manage the store and he is a dork ass mofo. Oh, I said that already. Slicked back blonde hair that is going bald in the back, a molestache, glasses, and he was wearing a TAN blazer with his grey pants. I dunno, that just seems odd. He looked like a fucking used car salesman.   Well, tomorrow I am heading to Lodi, taking some laundry with me and heading up into the mountains for some fun. I wonder if mimi\'s mom is going to go? That would be interesting. Her flight arrives tomorrow.   current mood:  melancholy current music: pedro the lion   '),(10367,216833,'06,July,2004','2004-07-06 00:00:00','    My needs and urges cease to control me, and I thereby avoids the cycle of indulgence and denial that tortures, confuses and distracts every living soul. I think I will start reading the words of Siddhartha Gautama. It seems so simple.   '),(10368,216833,'06,July,2004','2004-07-06 00:00:00','    Work was uneventful. I bought this awesome \"dice\" style game last night called Pass the Pig. I remeber playing it when i was a kid so I went hunting for it and found it at Target. hahaha. fun. Bought some beer too, and a alot of warm conversations with mimi and the TV played the mexican history in the background. It\'s nice to be married to someone you actually love. I feel bad for those people that get married for all the wrong reasons. Good luck to them.   :)   current mood: dorky current music: the radio and random mp3s   '),(10369,216833,'06,July,2004','2004-07-06 00:00:00','    I spent of the day cleaning up the apartment and watching television. I sent my resume to a few places, but I really just laid around. I never left ventured past the doorway, not even to get the mail.   I am lacking inspiration these days. I fail to self-start any project or any form of productive activity. I havn\'t picked up a book in a month, I havn\'t painted or doodled in about the same time. work-eat-workout-watch movies.  I did see Don\'t say a Word and Reservior Dogs tuesday night. Those were fun and late. I forgot to take them back last night. I think I am gonna go send my grandma a postcard.   current mood:  bored current music: samiam   '),(10370,216833,'06,July,2004','2004-07-06 00:00:00','    Monday night was interesting, we had a bit of a medical emergency. I came home from work around 9 after picking up a movie at CD warehouse. I cooked food while mimi watched some TV and after I was done cooking we started the movie. Not to long into the movie Precious started having a seizure. She has had these seizures before but never this bad. It wasn\'t letting up. She was on the floor flopping around, her body stiff as a rock and she was getting tired from it, I could see that. MiMi was crying and I was confused. We had no money and I could not make a decisive decision. I finally said, \"Let\'s just take her to the vet hospital and deal with whatever.\"  So we took her over to the local 24 hour vet and they gave her a shoot of valium and kept her over night. The doc told us that she has epilepsy and that she will have to be on meds for the rest of her life. poor puppy.  She is fine now, just a little sleepy. The drug she has to take is a barbiturate, and it is the same medication used on humans for epilepsy.   As far as everything else, I am off for the next couple of days so I am stoked. A chance to look for jobs. mimi gets home around 1:30 so we are gonna hit the gym a little while and head back home to clean up the apartment a bit. I am looking forward to Thanksgiving because I am heading out to BigTrees, CA and play in the snow and eat some food with mimi\'s dad and their family.   I think I have finally decided what major I am gonna go for when I start school. I still have 10 months to figure this out. I am gonna get a BS in Biology with a concentration on Ecology and Evolution and minor in Anthropology. When that is done I can hit grad school either towards Biology or Anthropology. UC Davis has an awesome Bio-tech school so I might go for a masters there. But Berkeley is my main focus, gotta keep those grades up. That is all I am her to do, go to school. I don\'t need to make new friends, they just end up needing time to figure out. I think I have friends right now that will last a life time, and they are in texas or somewhere else but here. I love my friends, I just want to keep it that way. I like how it feels now, I don\'t have that nagging obligation to go see a \"local band\" and all that bullshit. Fuck your scene shit, it\'s a hopeless cause, and a worthless one at that. I\'m a rambling man. I have to ramble on baby.   current mood:  calm current music: jets to brazil - perfecting loneliness   '),(10371,216833,'06,July,2004','2004-07-06 00:00:00','    Well, I am hitting the gym regularly now. 5 days a week so far. Everyday at Zales sucks and there are no jobs to be found, that I can find.   I am planning a ski trip for January in the tahoe area. Actually Bear Valley. http://www.bearvalley.com so if you wanna do it, let me know and I can hook you up with all the info.  MiMi and I went up to Jenner last wednesday and found this tiny forest road called Fort Ross Rd. We were on the curvy road for about an hour, the only ones on the road. It was so fucking awesome. I had a good time just looking at all the valleys and tiny little towns. The smell of fireplaces and the scent of pine and redwood is so \"life-giving.\" Makes me totally feel alive.  Now I have to get ready to go sale some Fucking RINGS AND BULLSHIT!   current mood:  lonely current music: the beatles   '),(10372,216833,'06,July,2004','2004-07-06 00:00:00','    Today, I am off work again. Crazy huh! I went to the mall with mimi and we went looking for an insense kiosk, no dice. We ended up goofing around for a little while, looking at random stuff. This is very good, becuase I try to pick up on ideas for christmas. Mimi bought a little western-style shirt at Old Navy. I think it is totally cute and I bought some Steve Madden style shoes at Payless for like 19 bucks and they are all man-made materials. I love em.   It\'s just a rainy day. I read a letter from a hurting friend. I joined a gym and that is rad!   '),(10373,216833,'06,July,2004','2004-07-06 00:00:00','    \"You say there is a finish line, but you are already beat.\"  Today is my first day off in seven days. I have been looking forward to today for a while it seems. I almost walked out on Thursday, I started crying and was so utterly depressed that I couldn\'t even hold on to myself.  I just felt so out of place at zales. Michelle, the \'fill-in\' manager talked me into giving it some more time. I have to say, she was right. I can totally do this.  I\'ve already sold a few high priced items and I feel totally comfortable talking to random people. I just had to get in that groove i guess.   We were supposed to go up to Fort Bragg and hang out in the redwoods but one of Mimi\'s workers didn\'t come in and so she had to go into work for her. Damn that cunt. Poor mimi cried as she laid her head on my chest in the kitchen. I just held her. I started to cry, this has been a change for sure. It\'s one of those good/bad situations. I know we will find \'normal\' in there somewhere.  Oh, I wish I would have taken the time to tell some stories of the week. I have a few for sure.   current mood:  uncomfortable current music: sunny day real estate   '),(10374,216833,'06,July,2004','2004-07-06 00:00:00','    10. Enjoying all the badass shows last January(since by man, etc.) 9. Palmer leaving town 8. Slamming my car into a deer 7. Discovering the American Kestrel for myself 6. Going to Mexico for Honeymoon/Getting my Handycam jacked 5. Melange/DFWAC art show 4. Quiting my job at EJB Development 3. Landing a shitty job at Zales 2. Moving to California 1. Getting married to MiMi  What a fucking year this has been. Looking back on it, my jaw drops. Did I really do all of this - in one simple year. Time flies any way you look at it.  Damn sore thoat. I think I am coming down with a bit of a cold. I really hope I can shake it. I would like to do some drinkin tonight. Ring in the celebration of time. We throw everything into days, weeks, months and make up some day where it starts all over. An excuse to get wasted. There is no harvest, no renewing or dieing really. Just a bunch of made up shit! The world is great, I wouldn\'t trade this bullshit for anything!   '),(10375,216833,'06,July,2004','2004-07-06 00:00:00','    So I am off today! I have a face to face interview next Tuesday for a Help Desk job at Legacy Marketing Group. I am almost finished with Smile, You are Traveling by Henry Rollins and I started reading Siddhartha by Hermann Hesse. Both enjoyable so far. Palmer actually gave me Siddhartha to read a while back. I just opened it up the other day and I have grabbed it every chance I can. In about an hour, I am going to run up to Evolution Tattoo and get some new shiz done. I was listening to some super sad fariuz earlier. I can just feel the pain in her voice.  There is still much to learn. We are not going in circle, we are going upwards. The path is a spiral; we have already climbed many steps. I found my paints and finished a painting last night. I like it alot. I can\'t think about these types of things to much. I think it clouds my head; just grab a beer and dive.   Oh, I can\'t leave this out! Zales sucks. I keep thinking that I will not get out until I find peace at Zales. Dammit mom, thanx for that idea. I can find peace, I can find peace......  Wish me luck!   current mood:  touched   '),(10376,216833,'06,July,2004','2004-07-06 00:00:00','    Christmas \'02 I am 23 now, and I can\'t think of anything I want more than just to relax. No big meal, no family. Just mimi, jeremy, leatha and her friend hanging out. We went to lodi to pick up jeremy christmas eve and we opened presents there. I landed a propane stove and light, beard trimmer, a longsleeve shirt with monkeys, and a tree ordament. We came back to the casa is sr and opened more presents with mimi\'s mom. There I landed a case of Beers around the world. It had twelve different kinds of beers from around the uh world. Very rad. I also got a little red toy vespa, some food stuff for the kitchen, and a beautiful card from mimi that made me cry. I seriously was not in the spirit of getting. I was really enjoying watching everyone unwraping their gifts. I thought that was neat. So anyways, I spent all christmas day working on jeremy\'s computer. I got it loaded up good now. After that we drove by the santa rosa, healdsburg, and petaluma skateparks to check them over. Very rad. After that I help jeremy with a website and we listened to load music late into the night.  Now I have to go to work and do nothing becuase I doubt anyone is gonna be shopping after christmas while it\'s raining. I did get a call from Legacy Marketing Group and they want to do a face to face interview! Horrrraaayyy!!!   '),(10377,216833,'06,July,2004','2004-07-06 00:00:00','    The last few days have been full of wonderful sites, sounds, and beautiful scents.  I was off on wednesday and thursday. MiMi and I cleaned up the apartment a bit on wednesday and went for a drive to look for apartments. We ended up somewhere near the coast in a small town called Dillan Beach. Lots of surfing there. Two weeks ago a guy was attacked by a great white while surfing at that beach.  From there we made our way down PCH along the Tomales Bay. We passed through this tiny town, Marshall. It must have been a fishing community becuase we passed by a few fishing type fellows. They had their warm sweater caps and dirty rubber overalls. Probably milling over ex-wives as they threw a few Bud Lights back around the bed of a ford ranger.  When we reached the end of the bay we made our way up the otherside, our destination being Point Reyes National Seashore. Sir Francis Drake Bay was in the park as well as some 20 ft. waves crashing down on the south side of the shore. I was mystified by the shear power of these waves. The air was full of saltly mist. Things were washed up on the beach that resembled some sort of snake/bull whip/plant life. It had roots and felt like a sponge. Funny.  We had an awesome, and quick drive back to 101. We took the Petaluma-Point Reyes Blvd. and found some beautiful hills lit with green grass as bright as neon. I forget what we did after that. I believe we rented the Fellowship of the Rings extended version and watched that.  Thursday, mimi and I went through some of our pictures and found them homes in empty photo albums. Finally. Cleaned the casa more. Mimi went to work at 4:30 - I went christmas shopping and was surprised by the snow on all the hills the hug the highway. just 10 minutes up the road, the hills were covered in snow. It was very cold that day. I was filled with energy, just wished there weren\'t so many damn people some times. Always rushing around.  I am really itching for a motorcycle. I want something to ride, and soon. But with this rain, I really don\'t have the chance too. Summer is a love/hate kinda thing, mostly hate.  Right now, mimi is at the tori amos concert. I am drooling over Henry Rollins on Jan. 8th. He will be in Dallas in Feb so check that out. I skipped out on the show since leatha was wanting to go, fine with me.   current mood: artistic current music: John Coltrane - I\'m Old Fashioned   '),(10378,216833,'06,July,2004','2004-07-06 00:00:00','    The poison has finished me off once again. I had a bit to much to drink but I had alot of fun. Playing scatgories was a riot. I was just saying silly shit for the one\'s I didn\'t get anything for. The letter is \"D\" - Something made of metal DOKKEN!   '),(10379,216833,'06,July,2004','2004-07-06 00:00:00','    My mind is still a difficult hurdle to clear. The race will never be finished, my life will always be 100mph and there will always be spills along the way. When I come to terms with this, things will start to clear.  I still have not made it to the gym. My lazy bones have been collecting dust in the bed or infront of the television. Cry me a river, so I can float away. Work is still something I dread. I don\'t want to help anyone, I hate jewelry and I am doing all I can to find a way out.  I don\'t mind my coworkers as much anymore. I yelled my fucking mouth off at Matt last night becuase of the way he handled a customer. FUCK! I just realized that I don\'t care about the customer. Buy or get out, see if I fucking care. This experiment is going all wrong.  Tonight we are having an unofficial zales christmas party. It\'s unofficial so we don\'t have to invite sherry, amanda, and atty. Supposly no one likes them much. Sherry and Atty are both over 40, Sherry being around 60 and Amanda is on the brain dead side of humanity. So I guess it will be all the young, hip people. I am part of the cool crowd. Fucking kill me now. I guess I am just too much fun to miss out on. So anyways, mimi and I are going over to Heathers house to play scatagories and drink some booze. Should be fun. This new girl, shelby, is bringing her boyfriend Jason. I met Jason about two days ago and the guy is fucking rad. We talked awhile and he runs a tattoo shop in Petaluma, plays bass, guitar, and some recording, surfs/snowboards/skate, vegetarian, and listens to \\m/SLAYER\\m/ and other such great metal and hardcore like reversal of man, orchid, and DEP. He also is going to hook me up with a red atom on my left arm and touch up my shitty blue atom on my right arm all for 100 bucks. I am stoked. So hopefully we can talk some more tonight and maybe he will teach me how to surf and I can be all: DUDE! like I totally road this wave to the sky! DUDE! WOAH! RAD!  Eitherway, I am getting smashed. I hope I don\'t spill anything.   current mood: stoked current music: Sunny Day Real Estate - LP2 - 07 - Iscarabaid   '),(10380,216833,'06,July,2004','2004-07-06 00:00:00','    Well, the last few days have been productive. I spent my work day drinking and playing age of empires. Mimi\'s mom was in town over the weekend but she left on Tuesday. They got caught up in some fog on Monday night driving across the delta. From what I hear from mimi, her mom freaked and I am just glad I wasn\'t in the car. I\'m sure I would have been pissed.  Went to the DMV yesterday and failed my motorcycle test, the written part. I went with mimi today and passed it. She got her license as well. So we are really in california.  After that we went to walmart and target to try and get a roll of film developed. We also ordered some holiday cards with out picture on them.   We went and eat at The Cantina in Downtown Santa Rosa after driving around for 15 minutes trying to find a place to park. I started the day off busy. We parked at a meter, luck me, it had an hour on it. I maxxed it at 2 hours and put the rest of my change in the fellow next to me.  After eating, we headed back to the car, and a college guy stopped us and offered me $1 for my leftovers. I hooked him up for sure with a fish taco and lots of veggies. That made me feel good, he was stoked. It would have been hard to justify me denying him food. I had just eaten, I didn\'t need to leftovers, plus its good karma.  We looked at some apartments in petaluma, and OH. my interview this morning went really well. I am just waiting on a call from them early next weeks to see if they wanna do a face to face interview. I hope that works out, becuase I think I can be a very volitaile part of their crew. I am about to go to bed, just got to finish my booze and crash out. I do more drinking these days then working out. Not good, I\'ve lost weight, but you can\'t tell, i think i have lost muscle mass.   current mood: accomplished current music: white stripes - Fell in Love With a Girl   '),(10381,216833,'06,July,2004','2004-07-06 00:00:00','    I came into all of this knowing that it was going to be hard. I have to start looking for an apartment this month. We need to find a place by the end of the month so we can be moved by the first of Febuary. I am weighing the idea of staying at this place until things are a little better, but, who knows, it might be like this for a while. They still have our $2000 deposit. that would play for a nice apartment\'s deposit and 1st months rent. I have to go to the DMV next tuesday and take the test to get my license and register mimi\'s car in california. I will do mine at a later time.  YEA! I just got a call and I have a phone interview on Wednesday with Legacy marketing group!!!! I\'m stoked.   current mood: accomplished current music: the vss - evolution   '),(10382,216833,'06,July,2004','2004-07-06 00:00:00','    I have alot on my list to get done today. I have to finish the laundry, wash my dog with flea killer, powder the carpet with flea killer just in case, plant some flowers I bought, and clean the bathroom. Oh yeah, wash the dishes and go get my library card. I have to go get more quarters here in a bit so I might get the libary thing knocked out then. I also bought a board so maybe when I am done with all that I can paint.  Mimi and I got half way up the mountain and had to turn back because it was getting dark. We started the hike later then planned but we still wanted to hike as far as we could. We got about 2.5 miles and turned back. It was still an awesome view but we will have to go back for sure. Maybe next time we will take our bikes up the fire road. We were the only ones out there until we saw this random lady walking down the mountain. She was dirt as fuck and had nothing on her. I would have to have some water for this mountain, at least. Maybe she lived in the park. :)  Well, I am off to tackle the \"day off\" list!  ttyl!!   current mood: busy current music: Danzig - She Rides   '),(10383,216833,'06,July,2004','2004-07-06 00:00:00','    Obviously there is nothing wrong with a few violent words now and again. I felt good today. I was top salesman for the day, I help people actually make worthwhile purchases and I helped one over-weight, over-indulgenent bitch. She made a statement something like You must love jewelry as much as I do since you work here.  Fuck you cunt! Jewelry is right there next to picking up dog shit on my list of things that make me vomit. I am looking forward to some meditation on the top of all this bull mess.  I have to say, I am really enjoying helping people that can\'t speak english very well. I am serious. Fuck all these rich-ass white people, give me some good-hearted hispanics and I will spend hours showing them jewelry. I helped this couple a few weeks ago. Carlos and Virginia. They were super nice, I was never pushy at all so today they come back and I show them wedding bands for about an hour. I was stoked to help them out. I can surprisingly understand most of what they say in spanish. That made me feel good about myself. Kiss my ass if you think I am racist. I love everyone, esp hispanics. All of my best friends have been hispanic except for jake and josh.  Integirty is something to behold and Spanish is a bad ass language! MiMi kicked my ass again at abalone.   current mood:  cheerful current music: Geoff Farina - Car   '),(10384,216833,'06,July,2004','2004-07-06 00:00:00','    I hate being nice to people, dealing with swine on a regular basis is eating away at my nerves. I get these urges to take their snappy attitudes by the hair and slam them into the glass cases. Maybe I will light up some fucking 80\'s metal as loud as it will go and start hi-kicking random people in the gut. Tomorrow I am off and I am gonna try and make it out to Mount St. Helena and hike to the top. Being outside away from people is my only solice these days. Once I reach the top I will let out a nice tribute scream to 80\'s metal! Listen closely...Dee Snider! This guy I work with, Matt, has invited me to his house about 4 times now and offered to go on hikes with me several times. I feel bad, because the guy is alright but he is still a fucking insecure tool obviously looking for a friend. Poor guy hates his wife and they have been married for just a month more than mimi and I. Where Not Gonna Take IT! I can\'t seem to find my way out of this pit of depression. I guess it is only a pitty party, so fuck it. What do I need to do? Then Fucking get it done!! It\'s that simple. Since my dog went to the hospital, I havn\'t been to the gym. I will get some excercise tomorrow on the mountain. I will take some pictures - I hope it is not hazy tomorrow becuase I want the full on view from the highest point in the Bay Area. I really had fun with mimi last night. We played abalone and I am totally hooked. She kicked my ass 2 out of 3 times, but it was still fun.   current mood:  pessimistic current music: Iron Madien - The Trooper   '),(10385,216833,'06,July,2004','2004-07-06 00:00:00','    Well, thanksgiving was great. mimi and I went up to big trees state park and hiked around for a few hours. found an awesome river. I will post pictures. I already developed the pics but left them in lodi. After that I saw jeremy\'s band practice. very fucking good. I was totally impressed by the sheer fucking change. I\'m a fan for sure.  The day after mimi, darrel, and I went to look at 10.5 acres of land we were thinking of buying. That was fun, the land was super rad.  Thanksgiving \'shopping\' weekend has not been to bad. I thought it was going to be a bit overwhelming. not at all.  I got a new game. Abalone, very fun. I just need to get GO and I will board game happy. A hiking I will go Tuesday. Somewhere, maybe Mount St. Helena.   Vodka vanilla coke rules my drunk ass.   current mood:  drunk current music: Nick Cave and the Bad Seeds - Ain\'t Gonna Rain Anymore   '),(10386,216833,'06,July,2004','2004-07-06 00:00:00','    I seem to understand things a little more with vodka. ha. i think i\'ve made a break through and I fall back into the same juvinell shit. gossip, mean mouth. the devil. this will alsolutly make no sense in a few years.   '),(10387,216833,'06,July,2004','2004-07-06 00:00:00','    So the move begins. Today we did all the lease signing and that was weird. I signed alot of papers. It\'s weird how writing your name in cursive is some how legally binding. ha. Who thought that clever shit up.  Work was blah. No real sales. Did alot of packing today and I am tired. The wine is nice, mimi is ill. She got sick from me and is sleeping now.  8)  I will be on dial-up shortly. Goodbye fast internet. I will see you again soon!   '),(10388,216833,'06,July,2004','2004-07-06 00:00:00','    It has been a week since I\'ve posted. Sad to say, but I\'ve been a bit down on my luck. I will try to be positive. GO HAPPY!  Last friday, like I said in my last entry, I was going to my HR interview with Legacy Marketing Group at 10am. I get there a bit early and Jennifer takes me to an interview room and the interview was pretty much just a list of questions. She wrote my answers down and it was painless and smooth. Almost all the Questions were \"Describe a specific time when...\" She told me at the end of the interview that there were a few other jobs available and if I was interested in those, she would like to consider me for those posistions as well. Both jobs are entry level help desk and the other entry level batch operator. I said I was interested because I really don\'t have the liberty of picking and choosing what job I would like at this point.  They said they would like to combine both part time positions for me and I would be working from 1 - 10 M-F. That schedule will be perfect when I start back to school. Anyways, she walked me out and said they would be making a decision for a team interview in the next week and a half.  After that, I go out to mimi\'s work, tell her the news and change clothes. We get everything ready and head to lodi. We pulled into lodi a little after 3pm and just sat around for a while. I was tired, the night before I was so nervous about the interview, I had a hard time getting comfortable. So I crashed in the chair for a little while. Jeremy and his bandmates showed up and started practicing for their show later that night. Around 6-7 they loaded up and we followed them out to Sacto to West Coast Worldwide for their first show. The place was pretty cool. They had a short set that was decent for a first show or a 101st show. I was stoked. After the set, they took all their equipment out to their vehicles and Jeremy and I took the guitars to his friend Nick\'s truck. Nick didn\'t have a lock on his camper so we threw em in my back seat for the time being. We were only gonna be a few minutes to talk to some folks and grab Bryan to get some food. When we got back to my car, my right back door window was busted out and all three guitars were gone. I felt so helpless, it sucked, it still sucks. My gut was twisted when I saw what had been done.  $1500 worth of guitars paid for some fucking crackhead tweekers habbit and we were the victims. After a few calls to the police, we got no help from them. They would not come and make a report for such a thing. I have yet to make the police report. I am still waiting on all the info from the guys as fars as how to describe the guitars.  So anyways, I was physically sick, with like bronchitus or something. It was mild at the time, and I thought I was beating it down but after hanging out in the cold for a while and driving 30 minutes with a window gone going 80 miles an hour, I think it started beating me. I was sick as fuck for the rest of the weekend. I rested most of sunday. Mimi and I went driving around.  Most of this week I have been dreaming about getting a motorcycle and trying to pack up for our move next week. I went to a few cycle dealerships and sat on a few cruisers. I\'m stoked. I started working on my credit to get some of the false shit off. Hopefully in a few months that will all be cleared up and I can go get a fat bike to cruise on.  I got a call from Jennifer at Legacy Marketing Group a few days ago and she said that they were going to hire an internal employee for the Help Desk posistion but she would like to set up a Team interview for the Batch Operator position. It will be a few dollars less an hour, but a fuck load better than Zales and I am down! Maybe it is meant to be, with school and all coming up.  I hope so. I am royally stoked about all of it. I just need a new window in my car. The duct tape and cardboard is holding up nicly, though. I need to paint a Skull on it for fun!  Sorry about all the errors, Proofreading is for smart people!   current mood:  shocked current music: you and i   '),(10389,216833,'06,July,2004','2004-07-06 00:00:00','    What\'s going on: Well at the end of the month I move to Rohnert Park. It\'s official. The rent is a about 70 bucks more and 10x nicer. I think so anyways. The rooms are a little smaller, but I think I will have a better peace of mind for sure. What else: Tomorrow is my second interview with Legacy Marketing Group. Wish me luck on all that. I am going to Lodi later that afternoon. Stopping at Ikea for a computer desk, or at least to look at them. They have alot of rad furniture for cheap. Since I sold everything I owned before moving out here, I have to gather a few things. They canned two guys at work this week. They two guys that have been there the longest and the two guys I get along with most. Brenden and Steve got the ax becuase they had to cut hours. I\'ve had to work with Matt a few nights this week. They day always starts out okay and by the end, I\'ve heard his mouth run for hours that I just want to get out! Death may it be, anything to save me! I try to sit and read but I get interupted. I try to show as little interest as possible. I believe that it\'s just to hear his voice. Ugh. Friday night, I\'m going to see Horror City Superstition. Jeremy, mimi\'s brother, plays guitar and I think they are very rad. Download their new song. Anyways, I need to clean up a bit. I need to start packing but that is never any fun. I\'m halfway packed becuase I havn\'t unpacked alot of stuff.   current mood:  indifferent current music: nick cave and the bad seeds - o\'malley\'s bar   '),(10390,216833,'06,July,2004','2004-07-06 00:00:00','    I\'ve been working at zales for close to 3 months nowand hopefully soon, i will be gone.   '),(10391,216833,'06,July,2004','2004-07-06 00:00:00','    It rained more yesterday. I calmed down once I got home. I sat around for the rest of the day watching movies. Goodfellas and Fight Club.   Now I have to go sale some diamonds!   '),(10392,216833,'06,July,2004','2004-07-06 00:00:00','    I sailed a schooner around the horn of mexico. I found my polaroid, I have pictures from my trip.   Last night, I got off work and went over to Brenden\'s apartment. The entire day proved one thing. People are lame. Over here, you have the guy who trys to hard. Over there, here comes some super fat couple to bitch and yammer to me about shit I don\'t care about. I\'ve found my self physically cringing at the sounding kick off of another useless anecdote.  What the fuck is wrong with me? This retail shit has really made me see through some people for who they really are. I used to be nicer, at least I thought so. Fuck being nice, it\'s a waste of time. Just cut the fuckers off fast and kill the pain before it begins.  Hyper people, calm the fuck down! There is no reason to run back and forth like a caged dog. You\'ll find yourself a weak, decrepit old bag of dust. This shit pisses me off.  Don\'t get me wrong from my journal entries. I\'m not always pissed. Just most of the time. I brought over Happiness to watch at Brenden\'s house. They had already started XXX. Good old Vin Disel skating around on his enormous, testosterone fuel cock rocket. Saving people with one fail swoop of his manhood...what a load of hollywood manure.  I think I might go to Lodi today, but I\'m not sure when jeremy is supposed to get back from the mountains. sigh....   current mood:  discontent current music: Willie Nelson - I\'m So Lonesome I Could Cry   '),(10393,216833,'06,July,2004','2004-07-06 00:00:00','    I fucking walk into a Target to drop off some film and right off the bat some bum is outside trying to push a newspaper on everyone that walks by. Inside, the photo guy is no where to be found and then he looks surprised when I give him a \"Go to Hell\" look when he arrives. WTF are you doing!! I go outside and some fucking cart pusher asks me if I have a cigarette to spare. I don\'t fucking smoke.  Give it a rest people!   I had all the interaction i can take for one day.   '),(10394,216833,'06,July,2004','2004-07-06 00:00:00','    Nocturnal artisans do not know fatigue. I have spent the last hour or so in a fury of writing and paint. Becoming happy with creations of any inspirational type are hard to manage; for myself.  I took off to work a bit early today to find a used book store in downtown petaluma. I presumed that I was to be at work around 12:30 so I found a parking spot around 11 and took a nice brisk stroll down the sidewalk.  The air was invigorating and I was fresh and crisp like a conspicuously clean dollar bill. My white shirt was bright and starched and I had the top button lossened. I hadn\'t yet thrown on the tie, so I was feeling a bit free and fairly handsome. I stopped first at the petaluma goodwill to see if that happened to have an overhead projector. I also stolled down to the basement to see what they had in the way of an office desk. From there, I strolled over to Red Devil records where I have yet to make a purchase. They have some rad Jazz, indie, and punk vinyl. I asked the gent to point me towards the used book store and it happened to be a few doors down, three to be exact. The store was rad. Open, indepentantly operated and clean. I got the following: The Metamorphasis and The Trial by Franz Kafka as well as Survivor by Chuck Palahniuk. I have today off so after a bit of sleep I have some reading to do.  By the way! Cheers to you my friend!!! Drink one for me! TO LIFE!    current mood:  enthralled current music: bjork - post   '),(10395,216833,'06,July,2004','2004-07-06 00:00:00','    10:33a - .::BE STILL AND WONDER::.  the last two days off have been busy for the most part. On Tuesday, I went to my Pre-employment acessment and that sumed up to be 3 1/2 hours of tests. Still waiting to hear back as far as a real interview goes.  Yesterday, I started a new painting, played Vice City most of the day and went to see Henry Rollins spoken word that evening.  I enjoyed that event. If nothing else, I was turned on to a few writers; Franz Kafka and Ryszard Kapuscinski. I am going to the used book store before work to gather up some stuff. I am so behind on good reading. I love books, and I enjoy reading immensely.  I am also on the hunt for an overhead projector. I looked a bit on ebay and the shipping is high. Maybe a thrift store will have one for me to purchase.  I\'ve been feeling a little on the humdrum side the last couple of days. Maybe it\'s my exhausted excuse for not being very motivated to do everything I\'ve been wanting to complete. I am relocating to Rohnert Park at the end of the month and I am not looking forward to the cost and labor involved. Speaking of labor, I hit the gym a few times this week. That always barrages the spirit and body. I owe the gym money, so that adds an impediment to my motivation.  So now I will try to obviate my discourse by rocking to some Stryper. That shit ALWAYS makes me laugh!   '),(10396,216833,'06,July,2004','2004-07-06 00:00:00','    So I made it to the fucking top of Mt. St. Helena. I had a good time going up and the view was awesome. There were even a few people hang-gliding off the summit. MiMi and I just sat there for a while looking around and watching them do their thing. The trip down was rough on me. My legs were jello. We are both in some pain today. It was worth it. I saw a ton of hummingbirds and a few birds of prey. Even found an owl pellet but stayed away.  Today I am working a split shift and I have to go back in a short while. I\'m on the tired side after eating some lunch. I eat some mexican food down the street and they had these little creditcard style calenders for your wallet. All of them had these elaborate and detailed latin/western paintings on them. I got the one with the aztec prince in headdress holding the sacrificial virgin. haha. I have a shorter day tomorrow. I\'m glad, and still stoked about the bird training tomorrow. I want to be that guy that takes the birds to the 3rd grade classes and tell the kiddies about birds and animals and wildlife harmony. We are all on the same earth.   Peace.   current mood:  blank current music: book of dead names - Yo   '),(10397,216833,'06,July,2004','2004-07-06 00:00:00','    I guess I will be the first to admit that I would consider myself manic-depressant; with that out of the way, let us carry on.  I had on hell of a head ache last night at work. I sent mimi to get me some medicine out of the drug vending machine and it ate my money. I was forced to suffer, and I was trying not to be a little baby about it.  After I got home, popped some pills and it was away in a matter of a few minutes. A miricle I tell you! I have been in an awesome mood ever since.  Mimi and I are going to tackle Mt. St. Helena once again tomorrow. We never made it to the top the first time so we will get out there earlier today to tackle the beast once again. When we went out to Clear Lake and back down to calistoga, the drive took us past the head of the trail and there were tons of people parked there. I am some what glad I have days other than Sat and Sun to play. When no one is around.  This saturday, we both go to our bird rescue training. I have to say, mimi and I get along more than I had ever hoped too. I see so many couples that obviously are in some sort of economic agreement instead of a releationship. Peace, Love, and all that hippie shit from beatiful NOR CAL!   current mood:  cheerful current music: Spirit of Versailles - Track 8   '),(10398,216833,'06,July,2004','2004-07-06 00:00:00','    ...and some days are worse than others. It\'s all about attitude and right now, mine sucks.  sigh   '),(10399,216833,'06,July,2004','2004-07-06 00:00:00','    Today, mimi and I were off. We drove around. Lots of little Napa Valley towns. Went to Clear Lake. That was nice. I am glad, I had a good time. I really didn\'t know what to do with my self.  I love not working at Zales. Some days are just alright.   current mood:  naughty   '),(10400,216833,'06,July,2004','2004-07-06 00:00:00','    So yeah. It\'s been a while and I know you are all really wanting to know. ha. I\'ve done a lot of interesting things in the last few weeks. Most of them were really something to write about. For some reason I\'ve been taking all these inspirations and little life notes and tossing them into my short story. A dog jumping at a tree, trying to catch a cat. An ex-con trying to make up for his mistakes by joining the forces behind drug and alcohol counseling. Work is fool of good characters. New faces at the jewelry store and new drama.  I\'ve been working on getting things ready for school. 1. Had a few meeting with councilors, that has been very helpful. looks like all my stuff will transfer. I will find out for sure in a few weeks. 2. March 1st I will start my first day of volunteer work with the Santa Rosa  Bird Rescue Center. I am really excited about that. I know that I should offer up my spare time for things worth something. 3. Job market sucks. I\'ve almost given up. :( 4. Trying to work on a lit zine. Probably won\'t have time for it in the end.  Today was so amazingly beautiful. The sun was bright and warm and everything was all wonderful. That took about 2 hours to wear off at work. I was sick of people quickly and wanted out. I can\'t wait to get home and curl up with \'lonelyness.\' The only time I ever feel real is when I am by my self.  I am so ready to sink my teeth into some learning.   current mood: accomplished current music: Hot Water Music - At the End of a Gun   '),(10401,216833,'06,July,2004','2004-07-06 00:00:00','    So it has been 6 months delayed, but I finally moved to Rohnert Park. I guess there is alot to talk about from the last week and a half. I am finished moving. I still have some unpacking to do but for the most part it is completed. I had a day off yesterday for the first time in a while. Zales was all super chaos, people quiting, getting sick and the rest. So I was called in to clean up. Fucking work ethics. Anyways, mimi and I went out to the Bay to Emryville to check out IKEA and fill out life with \'stuff.\' Actually I was tired of working on my computer from a fucking stool and my lap. So we bought a table. We bought the table top and the legs and just built it. That was fun and it turned out to be exactly what we needed. I think we spent about an hour or two there trying to figure out what we were going to get. They have tons of super inexpensive stuff there. I was stoked. Emryville is a really cool town, all the buildings look like a cartoon.  After that we went out to Gilman St to eat at the Chipotle. Burrito! For the most part my days have been filled with putting things in their place. It\'s begining to look like home and I am so at peace with it all.  Another plus is a I am right down the Street from Sonoma State University and I swear, the only people that live in my new complex is SSU girls. Everywhere with their hoodies and pony tails. AHH! My eyes are going crazy. haha. Oh and I am getting DSL...woohoo. they are transfering the service so I am glad glad glad.   current mood:  satisfied current music: Millencolin - Monkey Boogie   '),(10402,216833,'06,July,2004','2004-07-06 00:00:00','    Dear life and love, hollowed be thy name, you seem to be the ever searching vixen of my criss-crossed mind\'s eye.   I spent the morning at the brc going through some simple training on how to addmit birdies and different policies and i worked out my schedule with heather. I set up a little feeding station for all the birds around my apartment. Now, don\'t get me wrong, I\'m not all mister BIRD now, i just have something to do and I feel like I am helping something besides myself. You never know what kind of curve life is going to take, and this experience could, in the end, send me in a totally new direction. You never know sometimes.   Hi Ho Hi Ho - Off to work i go!!   current mood:  calm current music: Palatka - The State Of Gear   '),(10403,216833,'06,July,2004','2004-07-06 00:00:00','    Well, I got the part-time job at the BRC. I am now a Shift Assistant and get paid for the same thing I was volunteering to do. Whatever, that is cool.  I want to write more, but I have to run and get ready for work. I bought a new Sony Handycam friday. yeah! I also have alot of great things and ideas going on, but now i must shave and get dressed up like a prick!   current mood:  rushed current music: Maximillian Colby - Sifelaver   '),(10404,216833,'06,July,2004','2004-07-06 00:00:00','    Less than 24 hours now. Soon, the U.S. will make another one of those stupid mistakes that will stay with us for the rest of civilization. The begining of the end for what once was a strong nation, I guess it is really nothing out of the ordinary. Over the years we have raped and pilaged our way across the rockies and now the world. - - -  Coffee is a natural diurectic and I have been staying awake so much later. Worked at the bird rescue center and met a new girl. She was kinda not good at handling birds, she dropped three. whatever.   current mood: evil current music: Alkaline Trio - My Little Needle   '),(10405,216833,'06,July,2004','2004-07-06 00:00:00','    In Kyrgystan, \"during a feast, they serve the guest the sheep\'s boiled head on a plate. The guest must eat the brain, then me must pluck out and eat one of the eyes. The host eats the other eye. In this way the knots of brotherhood are tied. It is an experience one does not quickly forget.\" I suppose this is not unlike other ritualist dinners of brotherhood in all kinds of places in the third world. I have really found a facination with this area of the world. It\'s amazing, all the history of the world that isn\'t told to us here in america, even about our own country. I love history, I love stories of the past, I pray we will learn from our mistakes on the level of humanity and on personal levels as well.  I am extremly wired on coffee this morning. I\'m on my second pot and my eyes appear to prepnderate from out of my skull. I have two articles I need to start writing. Deadlines are coming soon.  Yesterday, on a whim, I signed up for the raptor handling class and made it out there around 11am. We spend about three hours going over all the ways to handle the resident raptors(hawks, falcons, and owls) and ways to read the birds behavior. In the last hour of the class we had to oppurtunity to handle some of the birds. I worked with a Dark morphed red tail hawk named cocomo and a little screech owl named blinken. Both birds we very well behaved for the session, I enjoyed the entire event.  When I got home, I recieved a call from an employer of a start up banking company. They do the banking at the local casinos and Cal card rooms. Weird, but it is 20 bucks and hour, so if it works out, i might be sitting at a blackjack table for a living, without gambling. I will probably be the only person making money at the table. Now I have to sit tight and wait for a call back to set up an interview. I hope he calls back, I am sort of weird about the whole posistion, but if I can sale jewelry, whatever. I fucked up and smoked some pot so I need to get that out of my system in case I have to take a piss test. It was just a few rips so hopefully it will clean out quickly. I was pretty worried about that most of the afternoon, not to stoked on losing a $20/hr job over a piss test. dammit.  Sometimes we choose the path not all that thought out.   current mood:  worried current music: Cave in - stained silver   '),(10406,216833,'06,July,2004','2004-07-06 00:00:00','    rode my bicycle to work yesterday. It was actually pretty fun, and I am going to do it as much as possible to save money. Gas is over 2 dollars a gallon now, and even though I drive a civic, that is still alot of money to be pumping away. I have stores and stores of fat cells that can be used instead. The ride over there was about 12 miles, not bad, it took me 40 minutes and I just changed in the nine west bathroom. The ride back was in the dark, and I couldn\'t see all the time where I was going, haha. There was a monster hill that I tackled and a fucking cop flashed by me at like 90 and scared the living shit out of me. Yesterday was warm too, i think somewhere around 76 F and the sun was beating down, but today it is super cloudy and cool. Should be a pleasent ride. Work was boring, still hot on the pursuit of another job. I found a few good ads today, so I had mimi fax the resumes at her work for me this morning. Oh yeah, and that Pam lady, I went to her house, in my car. She has this old Apple Quadra 605 and she is not getting any display. I opened it up and she was shocked at how i did it so easily. hehe. She said the Apple people told her that a battery had probably gone dead, well I found the battery for her and told her to try replacing that first and we can go from there. She was surprised again, and agreed. I was in there for only like five minutes. Speaking of all that, mimi and I are going to the BRC tomorrow for our 2nd training shift. Should be fun again.   '),(10407,216833,'06,July,2004','2004-07-06 00:00:00','    Nothing like a few cold \'stones before you go to work. Went to the college today to talk with Tia the conselor and the damn evaluation office has not finished my transcript evaluation for CSU. No worries. I went ahead and planned out my Fall \'03 semester and it looks like it\'s going to be Biology, Math(College Algebra and Trig), and Chemistry. 12 units of pure homework hell I\'m sure. That\'s okay, progress ahead is better than working on something in vain; I hope those other classes I took will transfer - still in the dark but I will find out in the fall. Right now, I just need to continue to review algebra so I don\'t fall on my ass and I should be fine.  I am going over to Pam\'s house tomorrow to help her with her Apple. She is the scheduling lady at the Bird Rescue Center and out of the blue she calls me to see if I want to help her. haha, whatever, i guess. I think the mother-in-law might be moving out soon. Not sure, but that is what it sounds like. That\'s gonna be odd, becuase then we will have to find a roommate. Havn\'t ever had to do that, but whatever.   current mood:  determined current music: The Promise Ring - east texas avenue   '),(10408,216833,'06,July,2004','2004-07-06 00:00:00','    For the last couple of days I have accomplished only a few things. They weren\'t even really that productive. I rearranged the storage closet, in actuallity, I was searching for all the parts to my low-rider manual scooter that I started ages ago. It must be at the father-in-laws casa where the vespas are stored. I did a little painting, I am low on supplies and broke so painting will have to cease for the time being. I can make due with what I have.  I totally abandoned my good eating habits, lots of beer I don\'t need. It\'s almost just something to pass the time. I keep thinking, i can go for a jog or ride my bike around town. No matter how many times I want to be alone, I also want to share everything with at least someone. even a jog. I have another appointment with the junior college conselor tommorrow at 9am. I will find out if the classes will transfer to the local four year university and also plan out my semesters. I guess I did do something productive. I started hunting for another job, no tech stuff, just a better \'job.\' I sent alot of resumes out for bank tellers, loan officer/processor, customer service rep, medical lab assistant, etc. Maybe something will come around and i can pull my self out of this finacial hole I\'m in. Leatha just told me that last night, around 2am, some people were walking back to their apartment and some dude came over and took a leak right under her window. ROFL!!! wow. Musta been drunk.   current mood: geeky current music: The Panthers - Snakes in the Soundsystem   '),(10409,216833,'06,July,2004','2004-07-06 00:00:00','    I got a compliment from someone today that made me feel good. I was told that I speak well thought out things that make sense and sound educated. Thats cool. I\'m smart! haha.  Did some work with a few songbirds at the center Tuesday. that was fun. I think that will be enjoyable. I\'m off tomorrow so i really want to do something exciting. Maybe read some and paint, ride my bike and clean!!! Rock out with your cock out!   '),(10410,216833,'06,July,2004','2004-07-06 00:00:00','    I went over to Matt\'s house yesterday and played some ping pong. That was some fun. I discussed a painting that I am doing and played ping pong for the most part. Made a pot of beans later that evening, the cooked all night and are excellent. They taste alot like those ranch style beans this time. I\'m kinda wondering what all that animal fat is for in the ranch-style beans becuase I made em with out all that and it tastes damn close. This morning, I went to this \"college skills\" lab at the JC. I think it will be helpful as far as getting me up to speed with algebra and all that again. Not much else. It felt like I was in school and that was kinda neat. I fill my life with little scribbles in my calender so I feel worth something. Stay busy all the time and you don\'t have the chance to see home meaningless it all is. I finished Survivor by Chuck Palahniuk, it was mediocre, but entertaining for the most part. Just started reading Imperium by Ryszard Kapuscinski. I am stoked about that book, so far so good. I also finished two paintings yesterday. I need to get my online gallery up. I\'ve been thinking about that for a few days now.   Peace.   current mood:  pleased current music: murder city devils   '),(10411,216833,'06,July,2004','2004-07-06 00:00:00','    I am completely wiped out today. I laided down at 11pm and some time around 3:30am was the last time I looked at the clock. My eyes were swollen and my face was weird when I shot out of bed at 7am with that \"Oh SHIT! I\'m late...\" panic feeling. Actually, I happened to wake up at a perfect time. I totally missed the alarm at 6:10am and must have just hit the button, whatever. So after stumbling around all day, forgetting things left and right, I made it here, to my chair at my desk - the walking dead. I almost didn\'t realize I was at the book store, I worked and worked trying to get this shipment out on the shelves and time flew by. As soon as I have worn completly out, it was time to go. Now I am off to shower and then bed.   current mood: wiped out current music: #681 pedro the lion - of minor prophets and their prostitute wives   '),(10412,216833,'06,July,2004','2004-07-06 00:00:00','    So far, so good at the book store. Today was long but fun. This lady came in and bought $285 worth of romance books on tape. I caught a little 10-12 year old looking at couple\'s \"how to\" sex books. Proxcimity is the key. He managed to keep away from people for a while. I just moved over close to him and straighten\'d books and he was quick to bail. Close to close two guys and two gals came in. One guy with a converge shirt and one of the girls with a Beneath the Ashes shirt. I joked around with them and tried to help them find some books but really didn\'t talk to them much. I suck at making friends. Maybe they will come back. I have to say, I meet alot of \"cooler\" people thru the book store. I got to talk Bukowski and Burroughs to some punk rock chick. Of Course, we don\'t have anything cool like that.  There is this cool address book that I need to get for someone. It is all filled with gay pulp art and is sooooo super rad. All these cartoon, hunky guys with no shirts on doing shit like camping and swimming. hahahaha. Some lady supposedly was offended by it a few months ago. surprise. but it is far from explicit.  I had the chance to force feed my first Red Tailed Hawk. Boy, that was exciting. One pissed off bird. I need to go to a show soon. I could use some music.   current mood:  pleased current music: Califone - Fisherman\'s Wife   '),(10413,216833,'06,July,2004','2004-07-06 00:00:00','    Yesterday was my last and final day at Zales!!! Now that that is done, I start at the book store today. We will se how that goes.  #565   current mood: nerdy current music: Rites of Spring - All There Is   '),(10414,216833,'06,July,2004','2004-07-06 00:00:00','    I didn\'t even get a chance to tell you about all the neat birds I got to work with on Saturday. We had 3 Black-crowned night Heron babies. These little dudes look like weird feathered dinosaurs. I was working with one in the lab, trying to get him to eat on his/her own out of the dish that I had placed in its holding box. That was a trip. At first, there was no reation to my offers of fish pieces. I would have to force em down its throat. I would check on him every hour and after the first visit he was catching on. I would wave the fishies in front of him and he would snatch it out of my hand. I did that a few times, then I waved the fishie and dropped in the dish. Viola! The food dish was discovered. That was cool.  Oh, by the way, I am on song 501 now.  Today, I started getting into the swing of running my own shift at the rescue center. Brad was my assistant today and we had everything ready to go early this morning for all the babies jays, finches, doves, and mockers. We had two admits during my shift. One was a little female house finch, the boy, probably about 8yrs old, that brought it in was awesome. His mom had drove him in and she could hardly speak english but the boy was busting out the story to me. They were walking and he saw this bird flopping around so he picked it up and took it home. haha. His mom wasn\'t to trilled about having a wild bird in the casa so she drove him to see us. I took the bird back and transfered him to his new basket home. I went out to answer a few questions and the boy told me he loves animals and that he doesn\'t like to see them hurt. DAMN! That seriously touched me. I told him the same thing, that I felt the same way. His mom said he wants to be a vet and I said, awesome, help him along the way.  I hope your dream comes true little antonio vasquez. ( I had to euthanize the bird later - what a bummer....) Next bird was a little bad ass, this dude found him in their pasture on the ground. A fucking baby Great Horned Owl a little larger than a softball. He had some giant feet on him too. I did the exam, he looked good, gave him some pedialite and called the raptor lady. Come to find out, we try to get these little dudes back in their original nest or a nest with some babies of the same general age. Unfortunatly we can\'t teach them how to hunt, so they don\'t do well in rehab. That made for a fun morning, I rushed out to home to get my Zales clothes on and get to work. I hurried over there to get started on doing NOTHING! What a waste of time. I worked with this chick named Carolyn that I totally pissed off the last time I saw her. We got into some discussion about why I don\'t care about people\'s lives. I do, but it is generally people I have woven some sort of relationship with, not random people I meet at work or on the street. She said, \"If I die in six months, will you go to my funeral or just say, \'Oh, that was someone I worked with...\' \" I would say the later. I don\'t care that much I guess.  I love my friends and family, I am cautious of strangers.   current mood: geeky current music: Jawbreaker - Million   '),(10415,216833,'06,July,2004','2004-07-06 00:00:00','    I had this strange thought that ran through my head when I was driving back from the rescue center. I was close to my exit and going around 75mph when I looked back behind me and a motorcycle was fast aproaching my car. I moved over a bit to let him pass between myself and another car in the left lane. Now, this manuver is perfectly legal in California. It\'s called \"lane sharing\" and I normally don\'t think anything of it. It\'s a kind of benifit for using a more economical mode of single passenger transportation. When you are stuck in traffic, you can move right between all the stopped cars.  Put what freaked me out was the high speeds at which this person decided it was appropriate to make this move. If I happened to miss him in my mirror, he would have went flying all over the road.  That is seriously fucked up. No matter how many times people piss me off or in general irritate me, I would honestly say that I would never want to kill someone on purpose or by accident. I ran the question by my self as I exited the freeway, \"How would I feel if I accidentally took someones life?\" You take away everything they ever had or will have with one mistake. I would be devestated. I can\'t imagine the shear magnitude of calling to arms a military with the soul purpose of killing people on both sides. With war you are garenteed that someone will die, no of old age. Fuck that, I hate people some times and even think to myself that I might wish harm upon them but it all comes back to you. I would never really want something bad to happen to anyone.  With that said, I have a hard time with humanity on an indivual level, I can care on a global or even national level but one on one caring is not my strong point. Maybe I am hear to care for the earth and it inhabitants. Maybe I can have a hand in reversing all my destruction and the destruction of my blood.   current mood:  thoughtful current music: Hot Water Music - turnstile   '),(10416,216833,'06,July,2004','2004-07-06 00:00:00','    So, I just hit song 300. I am listening to all my mp3s in a random mix all the way through all of them. This will take a while, luckily I saved the one random list I made. Today had to be the longest day I have had in a long time. For most of the day I had a fucking headache and it kept hurting, getting worse as the day wore on. I just finally got rid of it with some medicine when I came home. Now I don\'t want to go to sleep becuase I feel like I havn\'t had a chance to enjoy the day.  I read alot at work and actually watched an entire hour long episode of StarTrek Voyager in the back room while on duty. The jewelry store was extremly dead, plus I really don\'t care what I do there becuase I seriously only have three more working days left before I am finished with this silly part of my last six months. I hope the people at the book store are a little less obnoxious.   I have a few comments that I have pulled out of the book I am reading(A Peoples History of the US by Howard Zinn) regarding our little war we have waged against Iraq. This is mainly just knowledge from past events that seem to shed some light on current events. \"All men of conscience or prudence ply to windward, to maintain their wars to be defensive.\" - Roger Williams Regards attacks by the English against the Wampanoag Indians which ended in 1676. And in such a world of conflict, a world of victims and executioners, it is the job of thinking people, as Albert Camus suggested, not to be on the side of the executioners. And even a Baptist Minister had something to say that caught my eye. He said that only wars of self-defense were just, and in case of unjust war, the individual was morally obligated to resist it and lend no money to the government to support it. This was in regards to the Mexican war that had begun in 1846. Rev. Francis Wayland was president of Brown University at the time of these sermons against the war.  I bought a new watch the other day, check it out fossil watch. I wanted to get another pair of jeans but I ran out of money. Spent alot of it on food. Oh, this was from saling my white vespa frame. The guy that bought it plans to turn it into a CCR tribute bike with refrences to the Big Libowski. hahaha. Silly shizzzzz   current mood:  good current music: the crimson curse - psycho   '),(10417,216833,'06,July,2004','2004-07-06 00:00:00','    Rain, rain, rain... which made for a long, sleepy day. I woke up around 6:30, got the the rescue center at 7:30 and worked there. Had a good day there and a boring day at zales. I am going to lodi tomorrow to sale one of my vespas. A guy in OR is coming down to pick it up. I am stoked about that.   '),(10418,216833,'06,July,2004','2004-07-06 00:00:00','    Well, I got a call last night and I am now hired as the assistant manager at the warehouse bookstore in petaluma. I feel like I am in highschool, hopping from job to job in the same mall. I am really excited, mostly to be leaving Zales.  Tomorrow, out shifts change at the bird rescue center and I will be doing the 7:30-12 shift mostly. I get to try that out for the first time tomorrow.  What else is going on? Well, I watched a movie last night after MiMi and I went to the grocery store. The check out guy, Joshua K., was a weirdo. Always staring at MiMi and saying outloud the products like, \"vegetable broth, oranges, etc.\" I wanted to say, \"Yeah dude, that is right, you are good at identification.\" FUCK. haha. So I watched the Ninth Gate with Jonny Depp and the ending sucked. That is all.   current mood:  determined current music: Califone - Don\'t Let Me Die Nervous   '),(10419,216833,'06,July,2004','2004-07-06 00:00:00','    So yeah, I\'ve been meaning to do this thing but I just keep putting it off for some reason. I dunno why, it helps me put my thoughts together, but with working two jobs. I think I have been away from home more then ever before. Away from a computer even more. haha. The bird rescue job is starting to pick up. Today we admitted 6 birds during the moring shift and I had to do a few exams, one died on me but the other(a quail) is doing better so I am stoked. I have an interview with a book store tomorrow in between my two jobs, so maybe this will be better than zales for the momment. I like books alot better than jewelry. Diamonds are a touchy thing, people die over that shit and I feel like I am only making things worse for them. This weekend I watched ringu with jeremyu and wasn\'t that impressed. Mimi went out to point reyes and saw three whales. I didn\'t get to go, I was shunned from the vehicle. haha. I think that is where we are going to be camping at the first of May. One year anniversary coming up very soon. Wow. Time flies when you are having fun.  I\'m soooooooooooooooo tired. Speaking of palmer, he may be coming out to calif soon to party with his brother\'s bacholer party crew so I might get to see that fool. I could use a little home latly.  - - peace, more than ever - -   current mood:  tired current music: And You Will Know Us By theTra - aged dolls   '),(10420,216833,'06,July,2004','2004-07-06 00:00:00','    So, yeah, I have very neglectful to my live journal. It needs nourishment with words to stay alive.  I m working alot, busting ass at both jobs. I start school in the fall. Not to far away. I finally finished Siddhartha by Hermen Hesse a couple of weeks ago. Josh Palmer gave me that book a while back and I just sat it aside becuase I was reading something else. Now, I got to run off to the rescue center. I will finish this later today.   '),(10421,216833,'06,July,2004','2004-07-06 00:00:00','    Fuck man.....16-18 had to be the hardest most emotional years of my life. All I worry about now is money and time and even though that sucks, love and acceptance and friendship is something of a different story. I\'ve grown bitter and neurotic early in my life and I am some what of a hermit. I just don\'t like random people all that much anymore,  I found out today that my brother\'s friend\'s brother OFF\'d himself. That isn\'t official, but that is what everything is looking like. That makes me feel so fucking sad. I know exactly how he must have felt. I know I had thought of just ending it all a few times. We all have thoughts of suicide. I just could never give up like that. Sometimes that is the easy way out and I am sooooooo glad that I never dove into that statistic.   I am glad to know everyone. jeremy, pio, bryan, corry, nate, and all the badass fellas that I have met through the very extremely bad ass caveman - stick it out. Nothing is really as bad as it seems, believe me. I would much rather have another hello with each of you than a lifetime of goodbyes. So, know this from your good friend john. You are all the raddest guys in cali so keep on being overly rad! OK!!!!!   peace   current mood:  jealous current music: Vinaldi\'s Four Seasons   '),(10422,216833,'06,July,2004','2004-07-06 00:00:00','    So this morning mimi and I attended the Bird Rescue Center Orientation. That was fun. Let me run over that event a bit.   We find the place fairly easy. Pull up in the lot and it\'s not exactly what I thought. Right when we get in and sit is some seats we think are alright, this \'thing\' turns around to ask me a question and I\'m not really sure it\'s a dude or a chick. One of those gender bender situations. IT\'s PAT!!!! IT had very long hair, john lennon style glasses, large \"boob\' like chest that was a little on the hangy side, and a very square shaped body. Also the voice was high-pitched and there was a well groomed beard on the chin. It was very thin, more to be ambigious with i suppose. The IT asks me for a ride downtown after the class. FUCK! I so wanted to be a prick but it was really only a ride so I oblige. After that, the IT didn\'t bother me to much. Oh the IT\'s name was Mageda. W T F ! ! ! I am just an asshole I guess. I just don\'t want people to bother me. So I learned alot of great things and I am really excited. I found out that all of the European Starlings found today in North America-and they number in the 200 million range-are descendants of approximately 60 pair of birds introduced in New York City\'s Central Park in the early 1890s. A society dedicated to introducing into America all of the birds mentioned in the works of Shakespeare set these birds free. These are birds that we don\'t except at the center. They are not native. All native birds are protected and illegal to have in your home, under your care. But if you want a pet Starling, go for it. Even a piegon. There is no law saying you can\'t - at least in california. I\'m sure you\'ve seen a ton of Starlings.  These birds are related to the mynah bird and are very good at talking. Even better than parrots. So if you want a bird that talks alot, grab a baby starling right about now. haha. I saw red-tailed hawks that had been hit my cars, owls, kestrels, and turkey vultures. Very interesting.  Oh! I worked today, it was dull. So I\'m off tomorrow, but I\'m going to bed now.  bye   current mood:  determined current music: Hot Hot Heat - Bandages   '),(10423,216833,'06,July,2004','2004-07-06 00:00:00','    I have a list of things to do but they aren\'t getting done. COME ON LIST...DO IT! What does it matter anyway. Things are going well, yesterday was the first day of 100+ degree weather and it sucked. Hot as balls. I have to open the book store tomorrow so that is nice, will be off in a short amount of time. Nothing really going on besides the same stuff. Little weird things here and there but not wrong writing about, well, maybe - If I had more time....   '),(10424,216833,'06,July,2004','2004-07-06 00:00:00','    I am hungry but I am tired. I dunno if I should just crash or wait for mimi and get some food. Then I won\'t go to bed till like 1am. I have to get up at 6:45am tomorrow and work at the rescue center for 6 hours. This is such the coolest job I have ever had. I can actually say that I am helping something. I was going to burn some CDs but I will just sit here. I am full of those kinda things. I was gonna..........fill in the blank. I need to take care of my head, if you know what I mean. One way or another. ;)   current mood:  hungry current music: Cub Country - St. Louis   '),(10425,216833,'06,July,2004','2004-07-06 00:00:00','    I had the chance to see Jets to Brazil again but, this time in san francisco at the Great American Music Hall. That was simply the best. I could see those guys anywhere and just be totally thrilled.  I am starting to run out of steam this week. So much is going on at my bookstore job and my rescue center job that I won\'t even bore you with all of that. Let me just say that it equals 60+ hours of my time this week. At least I am taking the time and money to see a few good shows; music is good for anyone. Even if it adds to the \'happenings\' in my date book, I think it is worth every extra red vein in my eye. School registration is next month and, I am at least stoked on that. I have been spending my free time at home and work reading alot. I finished a few more books; most of them have been good. I actually went to the new library a few days ago and it was really nice. Maybe I will spend more time there, looks fun. I am attempting to write fiction again. Maybe I will just disregard all negative thoughts on my skills and just pour words on to paper. Oh fuck, I really should go to bed, I am creating a sleep deicit that I will not be able to catch up on for a few more days.  Mimi flew to Pheonix today for a business meeting and I am just hanging out in my boxers. word. screw off.   current mood:  distressed current music: The Fire Theft - Uncle Mountain   '),(10426,216833,'06,July,2004','2004-07-06 00:00:00','    Yesterday, I just work all day. Both jobs. Nothing really worth telling i guess unless you wanna hear how stupid people are but I really don\'t want to get into that much. Today I clean up the pad and washed mimi\'s car, had some chinese food, took a walk over by the SSU campus pond and filled up my car with gas. Then we drove up the small mountain range behind my house to find the Fairfield Osborne Preserve. I was just curious as to where exactly it was. They only do guided tours of the area but even the drive up there was super wild. After that I worked at the bookstore and that was blah.  DO IT NOW!   current mood:  predatory current music: Soul Coughing - Sugar Free Jazz   '),(10427,216833,'06,July,2004','2004-07-06 00:00:00','    This is a quick recap:  1. Went to pool party in super rich Glen Ellen for about two hours on Friday. Nice view and pool, stereotypical rich ass holes.  2. Went to Marin County Fair after pool party, saw some cool art, paid $11 to walk around with a ton of people. Lines long for carnival rides. Had the best $4 beer ever. Mild sunburn. Saw the coolest 5 minute firework show ever.  3. Saturday, finished my parakeet cage; worked at rescue center from 2-8:30pm.   4. Sunday, went out to Point Reyes and hiked the Bolinas Ridge trail about 2.5 miles deep then turned back because I was not prepared for the amount of sun. Second sunburn of the weekend. Wife and dog also hiked with me. Saw a few ravens, lots of turkey vultures, a wren, and a redtail hawk. Lots of cows, my pup even tryed to attack one - it started to charge and we all had a little freak out! Went home and cooked up some awesome shirmp fajitas on the grill and watched Independence Day on TV.  5. Today I work at rescue center for the first half of the day. Had the pleasure of working with a GOLDEN EAGLE and that was amazing. Very massive bird, alot more that I had expected. Poor guy is missing his right eye also he is still just a baby but adult size. wow. Had lady tell me at the book store that I should but all my efforts towards becoming a millionare. What a disappoint life I would lead if that was my ultimate goal.  g\'night   current mood:  okay current music: BANE   '),(10428,216833,'06,July,2004','2004-07-06 00:00:00','    Maybe you can say that you have seen enough for a lifetime, maybe you can say that everything has been free and simple but, I have to see more. Imagine yourself young, running around in the open country, darting in and out from behind tree trunks and turning over rocks to discover alien planets. That shit was great. So let me buy you your next bottle, and we will toast to our childhood! I went to the library today to trade a few books that needed to be turned in for some fresh ones that I will read soon. I found some great ones; eight of John Muir\'s books in one \'volume,\' an ecology book, and a biography on John Muir. I\'ll be stoked if I finish up a few of those books in that volume. I\'ve already started in on \"The story of my boyhood and youth\" and it is stirring up a lot of good memories of when I was a boy - always outside poking at the world around me.  After the Library, I worked at the book store from 12-8:30. I had a lady bring me a House Finch that had fallen from the nest. I kept him warm and after I left work and went searching for the nest, I couldn\'t reach it so I placed him directly under it and hoped for the best. I would have to put him down if I took him to the center; maybe his chances are better close to his parents. He is fledging anyways so he might already start eating on his own and flying even sooner.  I also met a former professor of Sonoma State University, he was very encouraging as far as telling me that my goal was obtainable and he wished me luck. He seemed like a very interesting man, obviously paralyzed from the waist down but able to walk with crutches. Determination is an impressive quality for anyone. He taught ethics and sociology, mostly for people going towards masters in social work/public service.   current mood:  good current music: Murder City Devils - Press Gang   '),(10429,216833,'06,July,2004','2004-07-06 00:00:00','    So today was a pretty regular day but I thought I\'d share it anyway. Worked a full day from 7:30am to 8:30pm. I had the chance to work with one of my fav birds for the second time this week. An Osprey, or Fish Hawk. What an amazing bird. This guy has talons as thick as a pencil and his feet are large in proportion to the rest of him. We got him in the center because he was down on the ground, some one caught him and brought him in - he was starving to death. So far things are looking good. Just the coolest thing to see - only the 40th bird of that species that we have had in the center for the last 20 years.   \"I\'m only a little pissed!\"  Even rarer than that is a threatened species called the peregrine falcon.   I had the chance to work with him on Saturday, mon, and today and he is a very impressive bird as well. Fastest of them all and almost entirely wiped out by DDT. They were considered endangered for a time but have moved to only threatened with the banning of DDT in the US in the 70s.  anways, those were really cool. My day at the bookstore was simple and easy. Really just took it easy, and relaxed.  I have a few options for the 4th. In Kenwood there is the World Championship Pillow Fights, that sounds silly. Also, a guy from the Rescue Center invited mimi and i to his pool party. I think I might go to that. Could be some fun. After that, I dunno. Maybe find a carnival and see some fireworks or go see Minus the Bear in Petaluma. We will see. Have to go take my library books back tomorrow. I guess that was a \'note to self.\'   '),(10430,216833,'06,July,2004','2004-07-06 00:00:00','    I think a temporary miracle has happened today. This extremely idiotic boy I work with will be off to Europe of r a month and a half. Thank goodness. I almost killed this brat today. He is the biggest wad of crap I have found in California. Lemme describe him; he is a lanky 18 year old know-it-all that fucks up almost everything and can\'t COUNT worth shit. He sings random shit to customers and is just overall the most annoying person I have had the bad luck of meeting. I know the kid hates my guts - I have never been more of an ass hole to anyone in my life. I am a straight up dick to this kid. For instance, he was helping the new girl with the register when she was doing just fine on her own, so I walk over there and ask him if he was instructed to watch over her shoulder. Then I tell him to get back to work. AAAAAAAAHHHHHHHHHHHHHHH!!!!!!!!!!!!!! He brings out things in me that I just can\'t cope with or handle well. Good riddance to him for a time. On a more organized note; I went to the DMV office to get mimi\'s car registered and our address\' changed on the ol\' drivers license. I could not work at the DMV - people are so insanely rude to the desk people. I am shocked at the nerve of some people.  Why can\'t we just be kind and generous to people - instead(me included) we are blindly led by selfishness.  You know, I have washed my hands with this senseless judgment of random people. It will give me nothing but more grief in the end. Every day is a good day for me, I am alive and my plans have not been ruined totally so screw all of these people and their bad days.  I want to share with you a poem that I came across while reading this book I bought. ----------------------------------------- Pine Tree Tops by Gary Snyder  in the blue night frost haze, the sky glows with the moon pine tree tops bend snow-blue, fade into sky, frost, starlight. the creak of boots. rabbit tracks, deer tracks, what do we know. -------------------------------------------- Man, I know the answer to that question. We don\'t know nearly what we all think we know about our surroundings. Take it all in, I think i might have to go off alone in the woods at the end of this week. Maybe do some meditation and find out where I ran off to.  I saw this stupid fucking bumper sticker today that said. \"Except for Slavery, Fascism, Nazism, and Communism WHAT DOES WAR SOLVE\"  Oh shit...   current mood:  good current music: The Mars Volta - Haunt of the Roulette Dares   '),(10431,216833,'06,July,2004','2004-07-06 00:00:00','    We got a ton of \'new age\' books at the store yesterday. lots of interesting stuff. I will have to rant about the people I work with soon. I havn\'t done that in a while. I am starting to hate them more and more. I really dislike most people - they probably dislike me the same. that is fine, I can live with that, but can they?   '),(10432,216833,'06,July,2004','2004-07-06 00:00:00','    Now that I finally found my lost little book.  Maybe I will write more.  These journals of mine always turn out like this.  Every day is a good day. I am stoked on life, just wish i could live without other people.  I can make page after page of people I hate and all the things they do that drive me insane. So ust to say what has happened since 1/15/03.  I left zales and started working at a bookstore with a paranoid obese woman and an idiot pysco kid.  Between the two of them, my generosity and kindness have disappeared in the work place.  I also work at a wildlife rehab clinic called Bird Rescue Center in Santa Rosa, CA.  I started there in March as a volunteer but was hired shortly after as a shift supervisor.  I guess I like to be an expert on everything.  So why not birds.  I really enjoy it, it has to be the most rewarding thing I have ever been a part of...   '),(10433,216833,'06,July,2004','2004-07-06 00:00:00','    MiMi and I went out to Point Reyes again yesterday for an awesome hike. We were dead broke and that is the best free place on earth. We drove up and parked near the Youth Hostel and hiked .8 miles down the Laguna Trail, From there we took the Fire Lane Trail about a mile to meet up with the Coast Trail. Once we hit the Coast Trail we hike up through the camp sites there. Our plan was to hike down far enough to venture off down some steep deer trails towards Secret Beach but we were side tracked but an awesome little beach that we could barly see from the trail. The valley created by the Santa Maria Creek made a little hole through the cliffs. We found a small of shoot trail and took a risk. What we found was just perfect. The little brook was emptying out on the beach making a lush tide pool that only reaches the ocean at high tide. In between the pool and the waves were a whole lot of gulls with there new young all grey and unkept. We thought we were alone but right near the cliffs to our right were two naked chicks sun bathing. They didn\'t seem to care, mimi and I giggled and headed south on the beach. We stopped at every tide pool and watched the green anemones and crabs just hanging out. Lots of blue mussels clung to the exposed rocks just off shore. After about a mile we ran into the Sculpted Beach and found a little rocky cove to sit and eat our lunch. Yumm, so them we hiked up another trail and met back up with the Coast Trail. From there we went back to the car. About a 5 hour hike, just a little sun burnt but not back at all. Seriously the most fun I have had in a while.  Today I worked at both jobs, got paid and filled up the tank. Was a bit annoying at the bookstore but we had a new shipment and that is fun. New stuff to check out.  Oh Jeremy, I might be out there soon, I have a Barn Owl I need to release in lodi. It was found in a truck cover in 2 tons of saw dust. In the dust somewhere. He\'s in good shape and hella pissed so that will be fun. ;) Maybe when we release him he will chase us all around with his fat talons.  So yeah, Good times, beach fun!   current mood:  indescribable current music: Murder City Devils - Press Gang   '),(10434,216833,'06,July,2004','2004-07-06 00:00:00','    I sat around and enjoyed my clean room and my spare time with mimi.  we laughed and joked around until 1:30pm when we had to head out to the rescue center.  I had to be at work at 2pm and she came along to volunteer.  I forgot our lunch so we had funyon\'s and coke.  Also, a few cookies that barb made.  Work was easy.  Had one mean Barn owl that wanted to kill me but I overpowered him anyway.  Today actually turned out to be pretty rad. A girl said her hands were all sweaty when she was washing the dishes.  That was odd. Another frizzie-haired blonde pulled away in a taxi from the police station in petaluma.  I wonder how her time in jail went? or her visit?   '),(10435,216833,'06,July,2004','2004-07-06 00:00:00','    This book is alot easier to write in when it is placed flat on a desk instead of my knee.  Today, I worked at the bookstore - scammed out of there early.  Went home and cleaned the house. I put away a bunch of clothes and rearranged my desk to make more room.  Ya know, to do more stuff on, like write in this little book.  I called hannah today to see how her birthday went, but she was busy with a few friends.  She is seven. mimi\'s mom is back at the apartment for two weeks.  I will be so busy next week.  I\'m starting to get a little nervous.  No worries.  I\'m super hip and thats all good.  I saw mars tonight.  Its really close to earth - and bright as well.   '),(10436,216833,'06,July,2004','2004-07-06 00:00:00','    Hannah\'s birthday today and I didn\'t call her.  I can\'t even think how old she is even.  There were several times I thought about calling but I guess I let it slip each time.  Today has been a really nice day, weather wise.  Lots of sunshine and a super nice breeze to cool you down.  The sun is setting behind me, lighting up all of those high clouds.   '),(10437,216833,'06,July,2004','2004-07-06 00:00:00','    Pieces of me fall like stars onto a black canvas night sky. The salty fish smell of the sea says it all.  Life.  Everywhere, I see traces of what once existed so vibrantly.  Now urbanized.  destroyed.  Maybe some day we will turn back the clock.  The sad state of so many heartbroken fellows bitten by commercial greed. So many things target them.  Books, songs, TV. Lonely is a way we all feel.  Not just the codependant or lust ridden, but all mankind. Money, Money, Money - is everyone weak and waiting to get taken? I\'m starting to think so..   '),(10438,216833,'06,July,2004','2004-07-06 00:00:00','    Being broke is a little on my mind.  Being a little lazy with my health is also there. Ready to go home and go to bed.  Got hired at the Bird Rescue Center full time last week. Next week, school starts.  I don\'t have any of my books. I\'m only working at the bookstore twice next week.   '),(10439,216833,'06,July,2004','2004-07-06 00:00:00','    All of my dear friends - I have missed you. I sometimes forget all about my friends and that I actually have people in the world that think about me becuase I am out on the end of the u.s. I dunno what I would do with out mimi. My pets just guard my stuff.  Well, news flash, I got hired Full Time at the Bird Rescue Center. So I will be working my little buns off there. I am sorta glad to get out of retail for a little while. I am only working at the book store occasionally. I have been reading a ton in my spare time and at work alot. I rush around from one job to another, always late.  School starts next week and the book store still only has 1 out of 4 books I need. Fucking weird, but whatever.  Anyways, my job description is changing a bit at the center. I am dealing alot more with all the raptors that we handle. That is really exciting and super stressful. I had to go feed all of the guys we have out in the rehab mews(outdoor aviaries) and was flown at a few times. Not sure if it was aggressive but still pretty hott. I was sweating my ass off worried about talons in my pretty little hairless head.  Mimi and I ran out to Point Reyes again on Sunday to have some food at the beach and watch the sunset. It was amazing. The moon was full as well and bright has her beautiful eyes. It is so dark out there and so alien. It makes my heart heavy to think that I only paint buildings and street lights as familiar in my mind. Today, in the local paper, horrays were in order for a woman who survived a cliff fall in the seirras for four days with a broken femur, hip, and other injuries. Damn, thats tough.  I just wanna say it sucks to grow up.   current mood:  melancholy current music: ben folds - still fighting it   '),(10440,216833,'06,July,2004','2004-07-06 00:00:00','    Well, I have to be more recent than alex so I\'ll make this short.  1. Busy with school  2. Might quit bookstore this week becuase mgr is fat.   3. Will be moving in to bird rescue center\'s studio apartment as care taker of grounds in six months to year. Rent Free.  4. Reading lots of Gary Snyder\'s poetry. growing beard again, becoming a mountain. Showering in the rain, wiping my ass behind a tree with a bear.  5. Winter at the rescue center kicks ass. no freaking baby birds.   current mood: caffinated current music: mogwai - may nothing but happiness come through your door   '),(10441,216833,'06,July,2004','2004-07-06 00:00:00','    just wanted to post this so you all know. I am coming to texas on Dec. 13-14th so lets hang out.  I bought a new keyboard today. horray. I feel like ass physically. Damn cold won\'t go away. Maybe it\'s a SUPER DISEASE!!  I really like going to shows, i need to go to more of them!  Karate Kid CHOP!   '),(10442,216833,'06,July,2004','2004-07-06 00:00:00','    Hey there butta!  This doesn\'t happen very often for me, I can\'t fucking get to sleep at all so i thought i would download a bunch of mp3s and stare at a monitor until my eyes fucking bleed I guess.  Anyways, i will be in texas on dec. 12-14 if anyone wants to hang out.   '),(10443,216833,'06,July,2004','2004-07-06 00:00:00','    I had to go to a sales meeting at 9am this morning. I was so frustrated with the basic crap that we went over, I almost left halfway through in high dudgeon. Besides that, I went to breakfast with Brenden and we chow\'d down on some spuds. You could tale by the look on our face that we both hate our job as of now. Even more so than before, I think. In my case anyways.  Luckily, I did get a call from Legacy, and I have a real face to face interview on Friday 17th. Slow going, but good news non the less. Talked to Palmer for about 20 minutes yesterday. That was cool. Nice to have him around again to talk. I read alot yesterday. I finished the metamorphasis and started in on the chuck plalahniuk book. Good stuff, I am stoked once again with a little interesting poker. Prod away at my brain. I could use a massage.   current mood:  aggravated current music: clipped wings   '),(10444,216833,'06,July,2004','2004-07-06 00:00:00','    school starts again on monday, and I have that on my mind alot today. After busting my rump at work, repairing outdoor enclosures and setting up my pygmy owl friend in his new house. He is spending the night at the center tonight and I hope he will be okay. I think it should be fine.  He seemed to deal with the new surroundings well. He ate and that is always a good sign.  I had a handling class today and the people that came were progressing very nicely. I can\'t wait till some of them are up to speed so they can take over these training responsibilities.  Now every time I talk about shit I am going to refer to Von Dutch. I\'m gonna go Von Dutch it in the big white bowl.   '),(10445,216833,'06,July,2004','2004-07-06 00:00:00','    I gave alot to you 2003. I took alot from you too. Now that it is over, I feel the need to catch my breath. I spent the last few waking hours of the oh three drink some booze with some new friends and joking about the accident i was a part of the evening of dec 31st. New drivers, I got banged in the ass by this 19 year old guy. Oh wait, I mean a car accident.   2004 should be interesting. Not as much crazyness, but I will find some for sure. This has been a swell group of holidays.  I wish all my friends in Tejas a HAPPY NEW YEAR!! I miss you all, and sorry i didn\'t get to see all of you. We need to hang at tea house in the big SF reading passages from hip books and seeing all the bueaty and art and wonder there is to life.   I wish all my friends in Cali a HAPPY NEW YEAR!!! You are all hella the sickest! :)  Read a good book and then write it out in steam from your coffee cup.   '),(10446,216833,'06,July,2004','2004-07-06 00:00:00','    garuda played their last show and I missed it.   '),(10447,216833,'06,July,2004','2004-07-06 00:00:00','    You should read my other journal, its a lot more interesting. Oh wait, you can\'t...it\'s on paper.   damn   '),(10448,216833,'06,July,2004','2004-07-06 00:00:00','    I\'ve been feeling pretty crappy the last few days.&nbsp; This apartment search seemed to be dragging out more than it needed to, but today there was a break thru.&nbsp; I got a call from the second place we filled out an app with and they approved the application so we have a place to move to now.&nbsp; I\'m really stoked to finally get out of Plano.&nbsp; It\'s not a bad town, and there is a really cool park next to us but it\'s time to get a place of our own.&nbsp; It\'s been a while.&nbsp;    &nbsp;  I called the first place we looked at to voice all my complaints but they havn\'t returned my call.&nbsp; Dammit.&nbsp;     &nbsp;  Tonight I am going to see the album leaf at hailey\'s club.&nbsp; Should be nice, I havn\'t been to a show since I\'ve been back in DFW.   '),(10449,216833,'06,July,2004','2004-07-06 00:00:00','    I just finished copying and pasting all of my live journal entries back into this blogger.  Now, I need to start typing out my books.   '),(10450,216833,'06,July,2004','2004-07-06 00:00:00','    This is yet again, another try at the old blogger. I am working on getting all of the old blogger posts as well as the live journal posts centralized on low-c.   I\'ve been working all day on getting all the other pages of low-c.com working with photos and pics of my art. I have nothing better to do right now, so why not.   I also have my webcam up and running again. check it out.  urlLink click here   '),(10451,216833,'06,July,2004','2004-07-06 00:00:00','    I\'ve been all alone, in the empty apartment by myself. It\'s pretty dull so far. I have no transportation, the bus doesn\'t even run down this street. There is a bike on Leatha\'s back porch, but there is no bike lock. I would hate to take it somewhere and it end up getting stolen. I would like to go to the library, its just one long walk. I feel brainless, just sitting at the computer all day. My thoughts and ideas wasting away in the light of the glowing monitor. It\'s okay, I\'m not really alive today. More like a zombie. Last night, I was awake till 3am, rendering myself useless in front of my computer. I had found an awesome 1980 Schwinn Cruiser on ebay. It\'s on of those big one-speed bikes with the big comfy saddle and the retro-lines. I bid on it and emailed the seller with an offer to see if he would end the auction early and let me come pick it up. He is just north in McKinney. Of course, being the hour it was, I would have to wait for a response to come in the morning. I tried to sleep. At 6am, the smiling sun was just topping the horizon, mimi gets out of bed and is quickly ready to leave for work. While I was half asleep, she takes her pillow and covers my head then turns on the light in the room. I guess she was looking for something. Soon after she is gone, out to drive to Hillsburo to train at another store. I can\'t sleep so I get out of bed, waste some time on the computer for a few hours and finally I get an email from the bike guy. He says it would be very convenient for him if I could pick the bike up, but my offer was to low considering he has 15 people currently \"watching\" the item, myself included. I was bummed but what do you expect. It\'s really no big deal. Last night, in my bout of insomnia, I researched Denton\'s transit systems like I was on the verge of curing cancer. More likely curing my insomnia. Denton has a pretty rockin\' bus system. There are a lot of routes, fairly frequent and the school even has its own more specific routes around campus. My need for a car is now out the window. I can ride everything for free with a Student ID. I did venture out of the hole of an apartment today. I went for a walk to the grocery store. A few blocks away, not bad. Picked up a few items for my grumbling belly. On my way back I saw, for the first time, a Killdeer luring me away from her nest by acting like she had a wing injury. I had read about this and seen a few photos but this was really neat first hand. Right next to Tom Thumb, there is a square of mowed land that is really just an island of soon-to-be developed real estate already surrounded by a parking lot and access roads. I\'m sure some crap-hole JackintheBox will fit perfectly. Well, I had walked across the lot on my way into the store but it wasn\'t until I crossed again the second time did the adult Killdeer start her act. Making all kinds of noise and flopping around. I knew what was happening so I looked down and I\'m standing less that 2 feet from the nest. Killdeer nest out in the open, on the ground, even in gravel parking lots, but their luring tactics and the camo of the eggs help them to be on of the most successful \"shore\" birds. There were three perfectly spotty eggs in a small gravel nest. The adult just sat on the ground, like it had just been a victim of mis-identification at the hands of a really crappy duck hunter.  Of course, when I started to walk away, instantly! she is cured of her injury. An Academy award should go out to hard working actors like that.  I left them alone, walked home, ate my food and creeped back to the terrible box. For what, I still don\'t know. But here I am. I finally got tired of the beating sun that comes thru the east facing window in my room. I found an extra shower curtain rod, placed it in the window and put a big heavy sleeping bag over it to block out some of the sun. Now I\'m positive the radiation poisoning is coming from my computer screen.   '),(10452,3858126,'30,July,2004','2004-07-30 00:00:00','    The last message I realise was worthy of being turned into a book. However, I promise to keep this post quite short (in comparison!).  This is just a little message to shout out Wa Hooo! I added a counter to my blog for curiosity sake and today it recorded the 100th visitor to my blog since last week. That\'s an achievement and I am very proud of it! I would like to thank ...........     (SHUT UP OH ANCIENT ONE!!!)     Ok, sorry!  Good night all! '),(10453,3858126,'29,July,2004','2004-07-29 00:00:00','    I have taken this diary directly from the   urlLink VSO website  . It highlights the issues that other countries have in regards to education. Andrew Lincoln is a very attractive actor who played the part of \"Simon\" a young and reckless English teacher in the Channel 4 series   \"Teachers\".    &nbsp; Read what happened when Andrew Lincoln, from Channel 4\'s Teachers was invited to Ghana by VSO to experience at first-hand the real work of VSO teachers and their colleagues, as part of the Valuing Teachers campaign.  Popping a couple of immodium in to my mouth and feeling decidedly worse for wear, I took a slug of water and squinted along the horizon. As the Land Rover hit another pot-hole I winced and turned to Katrina in the back.\"How many VSO\'s does it take to fix a light bulb?\"The reply was cautious.\"I don\'t know. How many VSO\'s does it take to fix a light bulb?\"I stifled a belch,\"What light bulb?\"I had just spent the night deep under cover in the African bush with two volunteers Ross and Steve. They had no electricity, no running water and a hole in the ground that bore only the slightest resemblance to a toilet.We were driving along one of the worst stretches of roads in the upper west region of Ghana. We\'d travelled over 2000 miles, visited five schools, met and shared beers with more than 20 volunteers and had very little sleep. I was coming to the end of one of the exhilarating, awe-inspiring and profound weeks of my life.The idea for the trip had formed two months earlier in the twisted imagination of VSO. The theory was this; the anti-authoritarian but passionate character I play in the Channel 4 series, Teachers mirrored the type of volunteer teacher that VSO recruits to work overseas. Even the pub scenes were apparently identical to the conversations being had by VSOs all over the world, just with warmer beer, more mosquitoes and the occasional power cut.That being the case, I was invited to experience first hand what being a volunteer means. It was tenuous but I was intrigued.Then I discovered that a friend I hadn\'t seen for three years and the daughter of my mother\'s oldest and dearest friend was stationed in Ghana and had been teaching English for VSO for a year and a half.News quickly reached me that my friend Laura Smith was in one of the remotest placements VSO had, the Upper West region of Ghana in a tiny village called Kaleo Wa. It would take two days to reach her by car as internal flights were reserved for government officials and big business - Don\'t they know who I am?Filming flew by and before I knew it I was standing in the Ghanaian embassy waiting for my visa. The wait would have been a little less tedious if the official behind the counter hadn\'t have found my real name, Clutterbuck so ridiculously amusing. My sense of humour had temporarily failed due to the dull pain in both my arms from the five injections I\'d received the day before. I felt like a walking pin cushion.So, at 10.00 pm, I lay in bed the night before my African adventure. My mosquito repellant, factor 28 sunblock and Lynx deodorant (African scent of course) already packed, I began my diary. Just so you know, I\'d bought a very nice albeit quite camp travel journal to log the journey.All the volunteers without fail thought it was the funniest thing they\'d ever seen, but undeterred, I stuck to my guns and would excuse myself every night at a civilized hour, retire to my bedroom and record the day. Well that\'s my story and I\'m sticking to it.  Saturday 19 January My nerves have set in and I\'m finding it difficult to sleep. Just watched a fascinating and frank VSO tape about Ghana.Modern day Ghana makes for unsettling reading. In a country where over 60% of the population are subsistence farmers, most of the spending seems to be focused in urban areas. Here are a selection of other facts that shocked me:  enrolment in primary education - Ghana 43.4%, UK 99.9%  public expenditure on health - Ghana $4 per capita, UK $1,366 per capita  50% of deaths in Ghana are children in a population where 20% are children. Most of the deaths are preventable diseases ie malaria  200 people are infected with the HIV virus every day.  That\'s a conservative estimate foreign aid to Ghana - $38 per capita.  Ghana\'s external debt - $360 per capita   With all of these crippling facts it comes as no surprise that there are many non-governmental organisations operating in Ghana today. They are all principally focused on development of agriculture, education and health.All of my exposure to workers in charities over the past few weeks has inspired me. I\'ve entered through doorways into vast bunkers, with people working for very little pay, some for no pay at all on phone lines, postal deliveries, advertising and education. There is a resolve, responsibility, fierce determination and quiet pride in all of the people I have met.This is going to be an extraordinary experience. Who knows that the next few days will hold. I feel humbled and ignorant, probably the perfect place to learn and hopefully begin to understand a country I have probably spoken about a handful of times in my 28 years.Bring it on!  Sunday 20th January Before boarding the plane I buy a bottle of Gordons Gin, the suggested present for the Chief of Kaleo Wa. I thought the picture of the Beefeater was a nice touch.On the flight out, I sat next to Charles, a business man who worked for Ashanti Gold in Kumasi. Every five or so minutes during the showing of Captain Corelli\'s Mandolin he would take his earphones off and tell me about his country. I learnt about its geography, economy, the forty or more tribes and even a few phrases in Twi the major language spoken in Ghana. Despite this, I was still able to follow the film.Oh, my word! The heat! As I walked down the steps on to the tarmac my brain naturally justified the warmth as heat emitted from the engine. But as I walked further away it became hotter, wetter and more smothering. It was 28Â°C, so humid you could feel water droplets in the air, and 10.30 pm Ghana time.The airport was chaos. I attached myself, or rather, a porter attached himself to me at baggage collection. I had suddenly become ridiculously over-excited. Every time I\'ve been to Africa in the past the energy of the place and its people intoxicated me and this was no exception. My friend from London, Chewy, while working with him in a play called Blue Orange had taught me the West African handshake. This involves a normal shake - usually held for a lot longer in Ghana - then a slow drag away of the hands and finally a click of both greeters index fingers - very cool to watch and do!What with the handshake coupled with the fact that it\'s an insult to not stop and chat if someone introduces himself to you, progress out of the airport was slow.Still buzzing from the airport. I got chatting to Peter our cab driver. Mid conversation and music pumping from the tape player, I asked if this was the latest sound in Ghana. He looked at me a little confused and said, \"It\'s Craig David\".  Monday 21st January Could have slept for four more hours as morning came quickly. After sharing an omelette with Liba Taylor, our freelance photographer (a seasoned traveller she can sleep anywhere, anytime, anyplace) we head off to another charity, Plan International\'s, headquarters.I have been sponsoring a child with Plan International and I had called them in London to let them know I was going to visit Ghana. I was keen to see some of their work there too and had arranged to spend a day visiting a community they work in 2- 3 hours from Accra.It was an incredible day that gave me a good introduction to Ghana before my journey up north with VSO. Even in that one day, I felt all of my preconceived ideas about rural Africa were dispelled. We went to a school in Asesewa run by an incredible man Mr Charles Na. After obtaining a masters degree he became an MP, and during this time he was arrested and put into a detention centre. On release he returned home to found the secondary school I was now visiting.The school children were amazing. Not only did the students work extremely hard at home they were also conscientious at school and seemed to generally relish learning. Some of the essays I read put my English to shame, one girl wrote of the \"dream light of literature\", - unbelievable - a budding Sylvia Plath in the making.That evening back in Accra, we met up with three VSO volunteers working in the capital. Solomon Ezobi, a Nigerian volunteer was working as he put it \"with the government\". I think it had something to do with accounting and taxes, and looked to be a position of great importance and responsibility. Jane Grier was a management advisor working with the disabled in Accra. She had been campaigning for equal rights for her students and had succeeded in forming a united union of all disability organisations which had strengthened their cause.Finally there was Michelle Banks from Australia. She was training social workers at the University of Ghana.After eating a superb meal in one of Accra\'s fanciest hotels, the conversation leapt from economy to culture to human rights and I realized how committed, and passionate they were.Tiredness eventually set in and, as we were going to be up early the next day, we said our goodbyes. I promised I wouldn\'t tell any other VSOs where we\'d eaten. Of course one of the first things I said on meeting a new volunteer was \"You\'ll never guess where we went for a meal with the guys in Accra\".  Tuesday 22nd January Woke up at 7.15 am to the sound of a gospel tune being sung by the cleaning lady. Breakfast was a subdued affair, as Katrina, Liba and myself were still a little jaded from a hectic day in the field and one too may Star beers. Our VSO driver and guide for the trip Kwesi (pronounced as if Emma Fudd were saying crazy) arrived at 8 o\'clock and after a mad flurry of packing bags and loading the Land Rover we were away. Before hitting the road we stopped to change some money. The currency in Ghana is hilarious, the exchange rate in Sterling is 10,000 cedis to the pound and the largest denomination is the 5,000 cedis note! So when I left the bureau de change with a bag full of money it looked like I\'d robbed the place. (You need a Del Boy wedge to buy a bottled water in this country!)Travelling up country was visually intoxicating. A barrage of noise and bright colours in Accra merges into tropical villages each having their own speciality produce. Palm oil, coconuts, bananas, oranges, pineapple and finally a village full of coffins! Goats and pigs wander lazily across the road. Vultures fly overhead and all the time women carrying huge pots full of water for miles without spilling a drop.Possibly because there is next to no tourism and Ghana has never been involved in any major conflict you feel you can instantly connect with the people. People are quick to smile. This automatically has the effect of releasing my London blinkers and I drink in all the noise, colours, smells and textures on the journey up.This sensory hypnotism is broken when Kwesi hears some dreadful news. The Ghanaian football team (the Blackstars) drew their first game in the African Nations football tournament with Morocco. Their star player has been given his marching orders for dissent. Kwesi was visibly shaken so I try to convince him that youthful ambition will conquer experience every time and besides South Africa (their next opponent) are past it. I don\'t think he believes me.During the eight hour drive to our overnight stop in Sunyani I manage to sleep, I\'m getting the hang of this international travel lark. In the evening another local volunteer emerges from the bush to join us for a drink.Cherish Ocampo is a charming 28 year old Philippino, teaching wildlife biology at the local college. She arrived in Ghana on the evening of September 11th. She was shy at first but after a while revealed a wonderful sense of humour, great independence and courage. Cherish was the only woman on the teaching staff at the college in and she spoke of initial teething problems and a self confessed weakness in speaking English. I realized quickly that the latter was false modesty.My pathological fear of anything that could eat, bite or sting me, led us on the topic of snakes. Cherish\'s eyes lit up. It transpired that the school gardeners had found a few cobras in the school grounds and given them to Cherish to start a school collection to use for demonstrations in class.\"Where do you keep them?\" I asked.\"In a home made glass tank\", Cherish replied.\"Are they difficult to look after? What do you feed them?\"Cherish looked up with a cheeky glint in her eye.\"I haven\'t fed them anything so far. All but one have died!\"There was a slightly surreal pause as I wondered where to go with the conversation. I decided to make a joke of it.\"What a shame - I suppose you could always stuff them?\" I offered with a laugh.There was another polite pause. \"Yes. Taxidermy\'s not until next term though.\" I looked, she was deadly serious.I think I\'ve fallen in love. Night. Night!  Wednesday 23rd January Woken up by Kwesi on the phone after sleeping in. Had a quick cold shower, grabbed a banana and we were heading north by 8.00 o\'clock.The journey to Wa in the upper west region saw a marked change in scenery, from relatively lush vegetation to arid and barren. The poverty levels dropped accordingly. We stopped in Baboi, a tiny village about mid way through journey. It felt like real undiscovered Africa. In a food stall called a Chop Shop we saw the native dish fufu being prepared. A yam (like a non sweet sweet potato) is boiled and then pounded until it reaches a dough like consistency. It\'s served with stew and eaten with fingers, without chewing. The stew that was being prepared looked like plantain and black coffee. I was a bit of a hit with the village drunk who wanted a map of London. I left him puzzling over the detailed plan I\'d knocked up of the one way system round Archway. We then set off on a three hour stretch along a dusty, red and extremely bumpy dirt track that Kwesi christened \"the dancing road\".After what seemed like an eternity we made it to Wa, where I was to meet my friend Laura. A reception committee was waiting. Laura greeted me with a huge smile and hug and introduced me to her fellow volunteers Ross, Steve, and Suzanne. She looked so well and happy and exuded a warmth and confidence that was fast becoming a trait in all volunteers I\'d met.An inter-school football match between Kaleo Wa and Nandom Community School had been arranged for our arrival, so without a second to lose we piled in the Land Rover and drove to the school.Little did I realise that I was booked to play in the match, as the VSO volunteer for Nandom school Mike was already changed in to his strip - Manchester United! Before the match began, it was speech time again. I thanked the teachers and students for the wonderfully warm welcome, pleaded with the Nandom lads to go easy on me and made a cheeky gag about hopefully scoring more goals than the Ghanaian football team. Okay, so it didn\'t bring the house down!Laura then gave an inspirational speech about fair play. She understood that there was a history between the schools (the last fixture had been a blood bath) but she hoped that our visitors would be treated to a superb exhibition of football. Mike leant over and whispered to me, \"We\'ll see - the coach has just told them to play like your opposite number has just killed your brother\".After a little coercion and a bowl of pito (locally brewed beer that tasted like sour scrumpy) I stepped on to the pitch in my Kaleo colours, yellow and green. I ran around like a maniac for five minutes then recieved a bruising tackle for my troubles. With my lungs on fire and realising I should quit smoking I called out to Laura, \"Too hot, too bumpy, too old\".The Nandom boys were definitely the stronger team. We managed to squeeze a 0-0 draw. I like to think this was due to me shoring up the mid-field, but I think it had more to do with the Kaleo girls singing and dancing provocatively in our goal mouth. Nandom was an all boys school.Shattered, but secretly quite proud of my performance we headed back to Uplands Hotel in Wa. Even more VSO\'s were waiting for us in the restaurant under the stars. There are 84 volunteers in Ghana, around half working in education. Was I going to meet them all? Talk slipped from food, to education on to illness. Suzanne who arrived a vegetarian, was eating chicken within a week and had quickly graduated to dog. Apparently the best meat is just beneath the skin so you have to put up with a bit of hair if you want the best.There\'s a refreshing humility and unaffected openness among the volunteers, and they all have a genuine love of the land and its people.Despite the frustrations, principally the lack of materials, all of the teachers are meeting the challenge. Laura, or should I say Miss Laura (as she is known here) has integrated remarkably well. In the year and a half of being here she has taught herself the local tribal language of Dagaare and has an ease of manner and strength that is truly African.  Thursday 24th January For the first time this trip I wake up before Kwesi. I\'m assuming it was around 6.00 am. I walked from my room under a mango tree in the muted, misty light of an African dawn. I was alone and the moment was golden. There is something unique about Africa when it wakes up, it is the pensive face of a country that is almost always, boisterous exuberant and gregarious and it is divine.The gang and I went to pick up Laura and Suzanne from Kaleo Wa and we headed into the school to meet the teaching staff and then on to the chief. He is the paramount chief in the region, Naa Bannaamwini Sandu. As we filed into his palace to greet him, I started to feel more than a little nervous. Laura had taught me a respectful greeting in the tribal language and I\'d spent most of the morning practising. I had that opening night feeling as I wondered if I would remember it.\"Ansomah AAY Obesoum\".I remembered the first word by that great British lingerie retailer Ann Summers. The second word of agreement is spoken as it looks but imagine you\'re the Fonz. Ridiculous but it worked and the meeting turned out to be both a moving and funny encounter.Then came the lessons. Just being in a classroom as a student, reverted me back to a sniggering disruptive idiot until I was quickly chastised by Laura.Watching Laura and Suzanne at work with the students was magical. In the stifling heat were two English girls in traditional blue and white Ghanaian dress playing literary bingo and demonstrating solar eclipses with a football and an orange.Walking back with Laura after taking pito with the other masters, I asked her what it was like living and working in Ghana. She said \"It\'s a gift. I feel like I have found a second family and the thought of leaving makes me very sad\".Speaking to Mr. John one of the maths teachers at Kaleo Secondary School during lunch, he said of Laura, \"We love her! She speaks our native tongue, respects and practises our customs and is a wonderful teacher\".Another teacher had entered the lunch gathering and without a seconds hesitation Suzanne had brought him a glass of water. \"You see,\" John enthused, \"this is why we love these girls! In our custom, when a stranger enters our house we bring them water\".Lunch was prepared by the home economics teacher and it was delicious, spicy beef stew and rice. I watched as Laura squatted on the floor and ate her meal with her fingers, chatting to her colleagues in Dagaare. A moving farewell from the teaching staff followed a last embrace with Laura and we were off. There was just not enough time, I could have stayed for weeks.Next stop was Lassia to spend the night in the bush with Ross and Steve, physics and maths teachers respectively in the local boarding school.On the journey over Kwesi and I attempt to follow the Ghana South African match but the deeper we enter the bush the worse the reception. As we arrive at the boys place we grab the first child we can find and discover the rest was another no score draw.Kwesi is mortified. When he leaves me in the bush, his posture is of a man defeated. I suggest the reason Ghana can\'t score is because we need to find a TV to watch them on instead of messing around in the remotest regions of Africa. He says \"You know you could be right\". I say \"Of course I\'m right. Your country needs you\".That evening, I have my first experience of preparing fufu with Abigail and Patricia, two students at Ross and Steve\'s school. It\'s exhausting work, while one pounds, the other turns the yam on the upstroke. Abigail gives me a couple of stern looks that I understand as, \"My fingers are quite precious to me\" so I retire and leave it to the experts.Throughout the meal and the rest of the evening we receive a lot of visitors. Some of the students break the boarders curfew to see the new white man. A couple of lads, that I took an instant shine to were Rodney and Andrew a latter day Butch and Sundance coming to offload their problems with the fairer sex - FANTASTIC!This is no lie, it is 11 o\'clock. I\'m in a box room with no furniture, lying on a makeshift bed made from bamboo, foam and cowhide. It is baking hot, an evil looking spider is watching me and about a mile away a funeral is in progress. Just so you know, a funeral is a bigger celebration than a marriage, namely more drums and louder singing because the bigger the celebration, the easier the journey to the after life.I joked to friends before I left, \"Africa, the damned heat and those incessant drums!\" I should know now never to tempt fate.Seriously this place is incredible. I feel so peaceful, inspired and bewitched by this country and its people. It too is going to make me very sad to leave.  Friday 25th January Woken up the old fashioned way by a cockerel. Unfortunately this particular one had absolutely no concept of time, it\'s four in the morning. I staggered to the toilet in the pitch black guided by a sense of smell. It wasn\'t difficult to find as it was a long drop lavatory that ripened as the hours wore on. Fortunately I was able to sleep again and was woken up at a more civilised hour by a cacophony of pigs. I ladelled water from the bin in the courtyard for my shower, a bin that cleverly would warm the water through the suns rays. I had a cold shower. Unfortunately - the sun hadn\'t yet risen.Before Kwesi and the girls picked me up Ross took me on a tour of his school. The thing that stayed with me was the dormitories. In a room about 10 feet long there were six bunk beds and I saw that bedding was thrown down on the remaining floor space. Fourteen children in one tiny sweatbox.The students washed outside in the courtyard and cleaned and ironed their clothes themselves. Despite these facts, everyone looked immaculate in their pink and brown uniforms.I was touched to find Rodney and Andrew waiting to say good-bye when I returned to Ross and Steve\'s house. Rodney was wearing an old seventies style school tracksuit top with his name stencilled in gold on the back - (Jarvis Cocker eat your heart out). The boys showed me a variation on the West African handshake which involved touching your heart after the click and I felt honoured.We were back on the road for the final stop with Jo and Jenna in Tumu. On the way, something strange happened. Deep in a bush, we saw twenty or so children painted as skeletons standing still looking at us. Some of them held their hands akimbo and started shaking. Kwesi turned to us and said, \"There are some things in Africa you will never know about\". He told me that Fetishism was widespread and this was an annual celebration of the dead.Sophia and Laura from Jaripa insisted that we stop for lunch and a quick tour of their school, St Francis Catholic school for girls. Out of all the schools we visited, this was the best equipped , but there were four classes without teachers. Unlike in England, where all hell breaks loose, when a teacher doesn\'t turn up here, the students teach themselves. An incredible sight. Although, the reasons behind this provoke some serious questions.During lunch in a local chop shop, where we sat in the dirt eating rice and stew with our hands I looked around at some of the of the worst images of poverty I\'ve ever seen. We then left Sophia in the midldle of the market to haggle over some vegetables and headed out along another terrible road to Tumu, the poorest town in the region and our final destination.No sooner had we reached the girls Jo and Jenna in Tumu, we were rushing off to an after schools AIDS awareness club.The enthusiasm, humour and confidence of the kids was infectious and my tiredness soon evaporated. Midway through the club, Jo dragged me away to a teacher versus students volleyball match. In the late afternoon light, my favourite hour of the day, a war was waged across that volleyball pitch, a conflict I had never hitherto witnessed on any sporting arena.Tensions grew. After several terrible calls, each one punctuated by a furious peep on his whistle, the umpire made one too many gaffs for Gabriel, the captain of my team. He flipped and a torrent of abuse broke forth, resulting in the umpire, who was also now at breaking point, thrusting the whistle at Gabriel screaming, \"Take my whistle! Take my whistle!\"Gabriel was apoplectic but soon found his defiant and dignified retort.\"I don\'t want your whistle,\" he said disparagingly.\"I want justice.\"Brilliant. I was bent double crying with laughter at this point. I quickly realised however that this was a serious comment. Needless to say we won the match three sets to nil.Supper consisted of chilli chicken. And when I say chicken I mean every single part of the chicken. There had been a power cut, so we resorted to the \"touch and try\" approach to determine the anatomy of the part we\'d chosen. Jenna had a well-practised technique which involved a swift and thorough massage of the meat. I likened the experience to a scene in the Deer Hunter calling it Chicken Roulette.Tiredness was getting the better of me but on the way home by torchlight I insisted we search down a black mamba or at the very least a cobra.The girls humoured me but the hunt lasted all of ten minutes and before I knew it I was face down in my bed, being rocked to sleep by Louis Armstrong\'s \"We have all the time in the world\". If only that were the case.The two day journey back to Accra was a calm and relaxed affair in comparison to the previous five days. Throughout the whole journey, the soundtrack playing in the Land Rover was Daddy Lumba, the most successful Ghanaian recording artist around. He played a style of music not dissimilar to reggae but more upbeat, a style called High Life. By the end of our whistle stop tour, I knew pretty much all the words to Daddy\'s new hit album, Poison! In particular the controversial song, One One One, Six Six Six. It had generated a wave of moral outrage because Lumba had ambiguously transplanted the word sex for six. I of course couldn\'t wait for each chorus and belted it out with Kwesi everywhere we went.My thoughts, as quickly as they came left my head, and many times I caught myself with tears rolling down my cheeks. The tears were not borne out of pity (or out of Ghana\'s lack of goals in the Africa Nations Cup) but more of inspiration, and a recognition that I had discovered something in this country\'s people that I had lost, or maybe mislaid.This is a country and people who need our support but not our pity. There are many qualities in Ghanaians that I believe we could aspire to back home. An extraordinary sense of family seems to underpin every facet of Ghanaian life, something that in our urban centres doesn\'t seem to highly prized. But my overriding sense of Ghanaians as a race is one of optimism coupled with kindness and gentleness.The other education I received during this trip was regarding charities and specifically VSO. There was a cynicism and distrust within me that money and aid wasn\'t actually reaching the people who needed it. This was completely dispelled on my trip to Africa. I also had a preconceived idea that VSO was just something people did instead of backpacking around Thailand. How wrong I was proved to be.The VSOs are sent to the most deprived and poverty stricken areas of Ghana, filling teaching placements where degree holding Ghanaians rarely wish to be posted. The principal reason for this lack of enthusiasm is poor pay, terrible accommodation and cultural differences, for example if a teacher from the south of Ghana is coming to live and work in the north.VSO is launching a Valuing Teachers campaign to highlight exactly this problem. It calls for governments to include teachers needs in their plans to improve delivery of education - it\'s difficult to be a motivated teacher if you\'re living below the poverty line in a school with no support or management.In spite of the poor pay and often challenging work conditions most of the Ghanaian teachers I met were amazing, both inspirational and pragmatic. The extraordinary Mr Charles Na from Assewa and of course the enthusiastic and brilliant Mr John Vianney Kubara from Kaleo Wa. Both highly educated and courageous men who made the trip back home to support and educate their villages in spite of more lucrative job opportunities elsewhere.I saw people living in the solution rather than the problem, and a sense of optimism pervading the whole community. Yes there were some children who were malnourished, and we saw incredible signs of poverty, but those images of despair we\'re so used to seeing in the West were replaced by a positivity we don\'t normally see. The people I met were optimistic and generous, facing harsh living conditions with dignity and most of all humour. I can\'t wait to go back.  Copyright: Andrew Lincoln/VSOPhotographs: Liba TaylorDisclaimer: The views expressed in this diary are those of the author and do not necessarily reflect those of VSO. VSO is not responsible for any factual inaccuracies in the diary. '),(10454,3858126,'28,July,2004','2004-07-28 00:00:00','    No, the Queen of Egypt hasn\'t found love, she hasn\'t even discovered any clues leading to it\'s hiding place! She just watched the movie  urlLink \"Love Actually\" &nbsp;and fallen in love with the concept of falling in love!  It has been a long time since I felt emotionally attached to anyone and watching all of those love&nbsp;stories develop in front of me reminded me of two things:  1. Andrew Lincoln (This Life, Teachers, etc.) is very attractive and worthy of more sites on the internet. 2. It is a wonderful feeling to be in love.  We all hope it will happen to us. Some people give up on it, saying it is a childish dream. That financial security is far more important. Other people live their life waiting for the dream to happen to them. I am the latter. I dream of Mr Right. He must be out there, living his life, oblivious to my whereabouts. One day, our paths will meet. You could call it destiny.   Well, good night to all you dreamers and good night to all the realists. And good night to all those who have found their dream and are happy in their space in the universe. '),(10455,3858126,'24,July,2004','2004-07-24 00:00:00','    So I went shopping today in mighty Tesco (an oversize cornershop that sells everything except...well, best not go there!) and I decided it was time to find out what had been happening in the celebrity world. I picked up a copy of  urlLink Kerrang  music mag so that I could catch up on the rock news and a copy of Now! to find out the ins and outs of being a 21st Century celebrity.   So: Britney is happy with her dancing boyfriend (so?)  Sadie Frost has gotten over her times with Jude Law at last.  Leslie Ash got broken after a bedroom session, contracted an infection and is now starting the recovery period - she can walk again now. (that\'s some bedroom session)  The fifties are back in again!  Elizabeth Jagger and Sean Lennon are an item.  Jay Kay and Amanda Donohoe are an item.   Anyway, more later.   Love y\'all!   I have now found an amusing site which reminded me very much of those years gone by when I had  urlLink technical problems  in my life. Check out this advice page for help.   Or maybe you just want to throw a  urlLink baby  about. WARNING: giggling becomes annoying!   urlLink Bushisms  perhaps?  Extracts from  urlLink letters of complaint ?  Enjoy, Enjoy, Enjoy '),(10456,3858126,'22,July,2004','2004-07-22 00:00:00','    This is just a short&nbsp;post advising all visitors to my blog to visit this site. It is all about two&nbsp;Red Cross fundraisers from Oxford who sold all they owned to gain money to go from Oxford England to Oxford New Zealand in a Morris&nbsp;Oxford.  urlLink  Oxford2Oxford    While spending a quiet night in the British Embassy, Kimberly and myself met the couple who had just arrived in Cairo. Their next stop was to be India. You can read their diary and maybe donate some money to the cause. '),(10457,3858126,'21,July,2004','2004-07-21 00:00:00','    Well, here I am in England.   I was expecting the weather to be cold and dismal... Well, the dismal part is right. It is cloudy and constantly threatening to be rainy, but it is also humid and hot - even though the temperature is probably only about 24 degrees C! I arrived yesterday after a very bumpy flight that stopped at Athens airport for 4 hours! There I bought a nice bottle of perfume and read my book. I even took a little nap!  So what has happened to me since my arrival? Well, I have visited two pubs and eaten various pork varieties. Most notable was the breakfast this morning of eggs and maple cured bacon. Sorry to offend any not-eating-pork people, but I have missed this food!  I went to bed earlier than I have done in a while and I woke up earlier than I have done in a while. Maybe it\'s jet lag, I don\'t know. After breakfast, I turned on the TV and watched various movies that I haven\'t had the luxury of seeing. They were all linked together by the presence of different actors. It was uncanny. Then I turned on the Sci-Fi Channel (Kimberly would be jealous at this as our TV is not working well yet) and there I watched  urlLink Farscape &nbsp;&nbsp;and was hooked. I just think John Crichton (aka Ben Browder) looks hot in his tight leather trousers! It\'s a good show, shame it\'s been cancelled!  Anyway, the prospects for today: go to the pub, enjoy the sticky hot weather of England and watch TV. Tomorrow I shall go and investigate town life! It\'s funny seeing kids coming home from school (my school ended a month ago!) and it will be even funnier when the weekend starts a day later than I am&nbsp;used to!  Oh well, good bye from mighty blighty! '),(10458,3858126,'19,July,2004','2004-07-19 00:00:00','    In the header at the top of my pages you can read about inflatable babies. Well, apparantly Egypt has sold out of these and the only things available to replace them are black and white inflatable cows, complete with bright pink udders. The guys selling these are carrying 20 or so inflated cows under one arm and a bag of fake noses that make silly sounds when you blow them. Now isn\'t that exactly what each resident of Cairo is sadly lacking? &nbsp; Saw \"The Day After Tomorrow\" today... Man that\'s some scary movie. No monsters. No aliens. Just stupid people who didn\'t take care of the beutiful planet they inhabited and it bit them back. Now who are those stupid people?????? This is the movie that all governments should watch to gain a complete understanding of just what a mess&nbsp;we have made of the planet! We need to wake up and take some responsibility for our actions! &nbsp; (Steps down from soapbox) &nbsp; Calls from strange guy = 0&nbsp; Hooray! &nbsp; Considered depositing his ring in a shop display but then the good conscience woke up and said that was a mean idea! &nbsp; Going to be taking a holiday to England from tomorrow 3:45am. Tooooooo early! The temperature may be a little warmer there than it has been but it will still be like going to the ice age for me. At least half the daily temperature. &nbsp; I will still contribute to my daily(ish) diary but it will have to be sensible stuff. That is if I survive the first ten minutes with the road system. Here I walk in the middle and cars honk and avoid me - there... I\'ll be squished. So wish me luck! &nbsp; Pleasant dreams! &nbsp; &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; '),(10459,3858126,'18,July,2004','2004-07-18 00:00:00','    First dates, eh???? Theyâ€™re a funny thing. They kind of set you up so that you know you want to pursue a relationship, or run a mile.  &nbsp;  My last first date was way back when, in the glory days of the 20th Century. It was nice. I was nervous but eager to get to know the attractive guy who asked me out on a date. We ate pasta and I talked incessantly about my life (nothing has changed there!). We parted after dinner with a lingering kiss. It was a good first date.  &nbsp;  Now letâ€™s talk about a dating nightmare. I met this guy at the cartoon voice test. He looked a little like a pirate with a big baggy white shirt tucked neatly into his black jeans. He wore smart, shiny, black shoes and a cap. The only amendment I (or Kimberly) would make would be a black bandana to replace the cap. He appeared to be cute and smiled a lot â€“ at both of us. Upon leaving he said he hoped to speak to us soon. I thought nothing else of it.  &nbsp;  Two days later, at 11am, I receive a call saying he must see me. We arrange a time and place (tricky though, his grasp of English wasnâ€™t that good).  &nbsp;  I go to the date venue: a cafÃ© in Zamalek (famous for fabulous cafÃ©s and night spots). I dress casual. I realise when I see a different guy at the cafÃ© that Iâ€™m not THAT into this guy I am supposed to meet. He arrives. Same type of clothing except no hat and his shirt is black.  &nbsp;  We go to the table (surrounded by courting couples) and he sits so close to me that I instantly feel Iâ€™ve had enough! Heâ€™s invading my personal space. He shakily grabs my hand and puts a ring on my little finger. OK! I thinkâ€¦. He then proceeds to tell me how he fell in love with me the minute he saw me! Wow! I think heâ€™s a headcase!  &nbsp;  Two hours go by of him professing his undying love for me. Discussion of marriage, children and other stuff. I am having trouble saying no to letting him kiss me as it is difficult to keep thinking of different reasons to avoid it. I have run out of conversation and he keeps staring at me. Heâ€™s starting to scare me. I finally end the date saying that I have to go to see a friend who needs cheering up. &nbsp;He says he wants a photograph to remind him of the event. OK. I say thinking a quick photo and then freedom! Oh noâ€¦  &nbsp;  By a photograph he didnâ€™t mean grabbing his camera and taking a quick snap shot. He took me to Mohandeseen, to a photography shop and there we posed for ten different pictures in a studio. I had to drape my arms over him, lie next to him, all sorts of stupid positions.  &nbsp;  Then, finally I got in a taxi and freedom was mine! I couldnâ€™t believe it.  &nbsp;  Let this be a lesson to all guys out there wondering how to impress the girl who just agreed to go out on a date with them: DONâ€™T BE TOO INTENSE ON THE FIRST DATE. Girls will run a thousand miles to escape you, regardless of how attractive you are!  &nbsp;  Anyway, I survived. Texted him to say, â€œEnough!â€. He then called me 8 times. Shame each one was missed! '),(10460,3858126,'15,July,2004','2004-07-15 00:00:00','    So I\'m still not well. It is really testing my patience! Thank you Richard for your last comment. I will put together a list of places to see in Cairo and Luxor, and then Egypt as a whole.  Today has been a weird day with a totally unusual thing happening. Yesterday I was called by a friend to see if I would be interested in any voice over work. I thought, yeah, why not. I used to be in a theatre company in the UK called the Barton Green Players in New Malden, London. The theatre world was becoming a distant memory so I jumped at the opportunity to do some more!   I wait with my friend on the steps of a gas company waiting to be collected and taken to the studio (doesn\'t that sound cool!!!) Eventually we are located. Once at the studio, there is no audition! How strange I hear you yell! We are sitting on rickety chairs in an entranceway to the studio. Other people are being recorded so we wait. We played \"I Spy with my little eye...\" It was fun! We are then taken into the studio. We are given a text to read...  We are reading the parts of pillow #1 and pillow #2. Yours truly was Pillow #2. I feel I put a lot of character into my part, I really established motives and underlying feelings!!!!!!! OF A PILLOW???!!!  The voice over work is for an animated version of the life of the Prophet Mohamed (PBUH) You have to add the bracketed letters as a sign of respect in Islam. The directors want really British people to do the parts. As I said earlier, it was a very weird situaion.  My friend and I (I think I shall call my friend Kimberly from now on. It\'s a good pseudonym and she knows it\'s her) then progressed to Genina Mall in Nasr City where we watched Spiderman 2. It was surprisingly good. I hadn\'t expected it. Any readers who have seen Spidy2 please comment and let me know what you though of it.  When we came back, I found Penny was feeding her puppies. I want to upload an image but I first have to convince my computer that it\'s ok to use Hello software. I gave her a rawhide bone. She was happy to see me but was confused by the bone. I daresay she ate it without any problems.  Went to Karaoke at Bullseye tonight. Every Wednesday is Karaoke and I have to be there. If I am not, then the world will explode and people will be forced to live in trees floating through space, boldly searching for new life and new civilisations; boldly going where no man powered trees have gone before! LOL  Anyway, sleepy now. Have a nice night/day wherever you may be and if you visit here, place a pin in the map!  Ishta! (cool!) '),(10461,3858126,'13,July,2004','2004-07-13 00:00:00','    Having just updated my log I feel I should say a big thank you to  urlLink Because I Say So  as she gave m the inspiration to use a fantastic map. So huge  THANX . I\'ll be checking your site regularly!!! '),(10462,3858126,'11,July,2004','2004-07-11 00:00:00','    So I have been a little sick these last 2 days. Itâ€™s what we call Pharaohâ€™s Revenge. It is generally the welcome to Egypt you get after the first two maybe three weeks. It soon passes, but sometimes it is necessary to take products like Antinal to relieve the problem. I consider it to be powdered (or liquid) concrete for your tummy. It works like a charm!  I was too sick to go out to go out to the pub or go to hunt for the Mountain Dew Fugitive. I have to tell you about him. Itâ€™s soooooooo exciting! (No hint of sarcasm) The only quality radio station in Cairo is 104.2 Nile Fm. They play quality tracks (but not enough rock for my liking) and they came up with this idea of a fugitive running around Cairo with cash. He telephones in clues in a strange voice throughout the day. You have to listen to these clues to determine where he is, what the password is and what he is wearing. My friend gave the station the idea that anybody searching for him must also be carrying a can of Mountain Dew. So that is the new policy; Go Her!!! Anyway, he starts wandering around with 1000LE  in his pocket (roughly equivalent to Â£100) and each day he is not caught, it goes up by an additional 1000LE. We first became interested when the total was 16000LE (or Â£1600). But now it is the holidays, we have time to hunt for this man. So, far this vacation, we have looked for him twice. We have our handy kit of mini radio, shopping guide, restaurant guide, Cairo guide and map book. Itâ€™s all ready in a bag so that we can just grab the kit and go. Hey it fills the day before going out to socialise with important people, ambassadors and such like!  Oh! I need a massage. Must go and relax in my haven of Maadi. Itâ€™s quite early and to be honest, I have actually just woken up (rather than just going to bed!) Maybe Iâ€™ll go for breakfast somewhere. You canâ€™t get good quality pork here so bacon is difficult to come by. You generally find beef bacon which is no comparison to Tescoâ€™s Maple Cured Bacon. When I go back to England, Iâ€™m gonna eat bacon every day. Itâ€™s amazing what you miss when you move away! Anyway, have a nice day! '),(10463,3858126,'11,July,2004','2004-07-11 00:00:00','    Well, we decided to go on an adventure this evening and discover a new night spot. Having lived in Maadi (where the grass is indeed fine!) for almost 7 months now, we decided it was high time to enjoy the social life Maadi offered. Every time we come home from Mohandeseen we would go past a place called â€œThe Golden Bellâ€. It intrigued me, so today the alleged â€œPUBâ€ was on our to-do list.  Oh, my expectations were low. From the outside it is a small wooden building, with burning neon lights that say â€œGolden Bellâ€ and â€œKaraokeâ€. We passed the bouncer and walked through the two doors intoâ€¦â€¦.an empty bar. A long walk past cubicle seats led us to the main bar. There were maybe 7 or 8 people in the entire pub. Two were playing pool, two at the bar and some other people round the corner.   We looked around and discovered there were two rooms hidden behind darkened glass. One door opened, to reveal a gathering of Koreanâ€™s singing out loud and playing tambourines. We sat at the bar and listened to Melody TV combined with Korean songs. Looking around further we noticed a darts board. Now, we both play darts and so upon seeing a darts board we thought about playing a game or two. We then looked at the surroundings: poor lighting, no score board, no line on the floor to indicate where one should stand. But all of these are nothing compared to the fact there is a table and chairs set underneath the board. â€œPlease sit hereâ€¦â€ whack!!!! Dart in the head!!!!!! Wow, it is exactly as you would expect from a country where health and safety is an unknown thing.  Ahhh, health and safety, now thereâ€™s an interesting thing. A few examples are needed so that people outside of Egypt understand the joke. Letâ€™s start withâ€¦ scaffolding.  In England, scaffolding is carefully put up with slabs of concrete to secure the lower struts. The horizontal bars are bolted together and all bars are made from strong metal. Now think the opposite and youâ€™re in Egypt.   Electrical sockets without earth wires. They actually spark when you go near them with a plug. A mini firework display is the best description.  Elevators: Ahhhhh. In England they have doors. In Egypt we watch with interest as the wall of the lift shaft zooms past.   Enough. I must go to bed. Iâ€™m watching Wives and Daughters, the BBC adaptation of Elizabeth Gaskellâ€™s book. I confess I am a big fan of period drama. I need to get hold of Pride and Prejudice. I need to have a Mr Darcy fix.  Good Night World! ! ! '),(10464,3858126,'09,July,2004','2004-07-09 00:00:00','    So when an ambassador for a country wants you to marry him, he really goes all out there! \"I\'ll build you a school...you can teach and I\'ll run it for you!\" This is a little extreme I feel. I say I\'ll think about it. Hey, Mauritius is a hot sunny country where there are white beaches and apparantly hotels with bedrooms that have glass floors so you can see the fishes under you. Sounds cool, but I don\'t want to marry a man older than my own father in order to obtain it! Egypt is one very interesting place where invitations to get married happen frequently.   Today I turned down an invitation to go on a felucca at night. Consider me crazy, but I didn\'t relish the opportunity to go on a boat at night with nine slightly worse for wear people!! I was very sober and very sensible!  Nothing really crazy happened today. But there was a very memorable meal. I had roast lamb with roast potatoes, beans and carrots, a huge quantity of gravy and... wait for it...MINT SAUCE!! It is the first roast I have had since Christmas. It was delicious! It was so delicious I forgot to leave room for my desired pudding...Apple and Cinnamon Crumble. That would have been perfect. It just means I\'ll have to go back to that restaurant again!! For readers living in Cairo and desirous of a good roast meal, it is The Nook in Maadi Degla. Street 254. Check it out!!!!!!  Time for sleep now. Must get up later for breakfast - I promised a friend. Good night and sweet dreams! '),(10465,3858126,'08,July,2004','2004-07-08 00:00:00','    Men with guns just walked past my house!! Not the usual rifles, but hand guns - tucked into their trousers. Life is bizarre.   Just got in after singing away in a karaoke. Sing sing sing I went. It was very nice! I had a great time. I am very sleepy now.  Today: Man selling inflatable puppies - who buys this stuff??????? Particularly while they\'re driving down the road!  Also: Saw two small boys pushing a drinks cart piled high with stuff down the motorway. I wonder where they were going to?  Thanks for your comments. I did leave Al Rehab to have a more interesting social life and to be closer to Ancient Egyptian stuff. I can see the pyramids when I am on my way home from work you know!  Made friends with a man who teaches soldiers how to fight properly in tanks. He\'s Egyptian and so he is also trying to teach me some Arabic. He said my Arabic was very good, I don\'t necessarily believe him!  Anyway, goodnight to one and all!  Rosie '),(10466,3858126,'07,July,2004','2004-07-07 00:00:00','    So I moved to Egypt 11 months ago to work in a small town of no consequence. The only impressive thing about the small town is its name: Al Rehab. Yes, I moved to Egypt to live in Rehab!!! The town had no life and no prospect of life. It was good for people who had families but not me (I had just left the nightlife of London!)  I stayed in Rehab for a while, rarely venturing out. But then I moved to a slightly larger town called Maadi. It\'s closer to places of alcohol and loud music! I am happier now!  The point of setting up this blog is to educate the world about the crazy things that happen here. I\'ve already got to the point where I take many things for granted when I could be sharing these experiences with others!  Today: I went on a falucca (which is a small sail boat) just around the Nile for an hour. It was very relaxing, or it would be but for the Italian film crew filming us. We still don\'t know why they were filming us but hey, maybe I\'ll be famous! LOL! '),(10467,3858126,'04,August,2004','2004-08-04 00:00:00','    That is the sound of time ticking away. I only have 14 days left in England before I have to go back to Egypt. Don\'t get me wrong, I am looking forward to returning to the sun, the heat and the karaoke, but I will miss my folks. Also going back means that work is about to start again.  I teach Egyptian dolls and army boys how to read, write and count using the British curriculum. That means they know about Queen Victoria and Henry 8th but know very little about Egypt! I do look forward to getting my new class. The only question is: will they be the children I was told I would be getting at the end of last term? We will see.  The Children go back to school on 31st August, but teachers return on 24th August so that we can prepare our classes. There will be many new teachers joining this year so that is kinda exciting. Hopefully they will be a fun-loving bunch and not members of the older (I mean OLDER) generation.  Anyway, these next two weeks are mine. Even if it rains like it did today (not used to rain - it only rained three times in the last year in Egypt!) or if it is sunny, I will be partying in London and Brighton.  Enjoy your life! '),(10468,4151531,'17,August,2004','2004-08-17 00:00:00','   A friend of mine told me that blogging is just for people who dont have lives...but i think some of us just have an ingrained desire to write and attract attention to ourselves...or it could be that blogging is the ultimate activity for desk-bound workers in this country who exercize an increasingly endangered personal freedom to say anything we want.   '),(10469,4151531,'11,August,2004','2004-08-11 00:00:00','   now that i am listening to npr in the morning instead of hip-hop, r and b, and old-school, i am very much worried.  a few days ago i thought that it would be impossible for bush to win the election but now im not so sure.  theres the florida problem again.  i also think that it is hard to get a sense of the direction the nation is moving in from the bay area.  i dont know how useful it is to go to meet-ups or rallies or fund-raisers in this area.  instead of preaching to the choir, i think the efforts should be increased in the swing states.  i would like to go on a \'logic train\' campaign across the country and sit down with undecided voters and speak to them.  but this requires critical thought and intelligence...  which brings me to a bigger almost blasphemous notion:  is democracy really that great?  the admin trumpets it all over the world (even though many countries are not ready for democracy) as if it could be implemented as a shake-and-bake set-up (easy bake democracy oven...ready in five minutes or less!!).  if you look at two of the longest periods of peace in the world (amongst large, diverse populations), you will see that neither of them was under a democracy.  im referring specifically to the reign of akbar during the mughal dynasty in india and the pax romana under augustus c.  the former peace was under muslims and the latter, pagans.  the problem with democracy is that currently it does not assume the votership to be educated or possess critical thinking.   '),(10470,4151531,'05,August,2004','2004-08-05 00:00:00','   this is my first ever blog!  hot.  as a matter of etiquette, from what i have noticed in the recent 3 years, i believe that someone with a cell phone who uses it to call people but then turns it off should be reprimanded.  as if its her one way paging system.  fuck that.  if someone says to you that they are like the brother or sister that they never had, this should be a warning sign that the person might be missing the \'non-codependent\' screw.  why do people fawn over clinton?  hes a fake liberal...im sorry, but i just cant take the guy:  hes sleezy slick like his wife and if you look at his presidency, its pretty bad.  nafta?  bad.  multi-fold increase in drug incarcerations?  heinous.  iraq?  bad.  economics?  just short of the republican agenda it seems.  the one good thing about clinton is that he just didnt care what people thought...   '),(10471,1758432,'26,May,2004','2004-05-26 00:00:00','   today i feel so cute ! donÂ´t know why, maybe because of using linux all the day :)  '),(10472,1758432,'26,May,2004','2004-05-26 00:00:00','   welcome to my new blog !  '),(10473,1758432,'09,June,2004','2004-06-09 00:00:00','   urlLink http://goforit.unk.edu/cprogram/c_093.htm   Another Linked List Example  urlLink http://goforit.unk.edu/cprogram/c_094.htm  '),(10474,1758432,'09,June,2004','2004-06-09 00:00:00','   c++ assignment for this week : structure  '),(10475,1758432,'09,June,2004','2004-06-09 00:00:00','   urlLink http://download.com.com/3002-8022-10289035.html?tag=dir  '),(10476,1758432,'28,April,2004','2004-04-28 00:00:00','   Managing time   I still confuse about my schedule, which lecture to be taken, german courses, and work.   Hope it will fix next week.  '),(10477,1758432,'27,April,2004','2004-04-27 00:00:00','   Handy   I have a new handphone, Samsung X600. With about 65 thousands pixels display and VGA camera.  Quiet cheap, only for 5 euro monthly in 2 years :P  '),(10478,1758432,'26,April,2004','2004-04-26 00:00:00','   New Semester   New semester just begin today ! It is also my b-day.   First class feels so hard, give us a big shock !  I can\'t understand completely what the lecture says, but keep on writing the notes :P  '),(10479,1758432,'20,April,2004','2004-04-20 00:00:00','   Knoppix : ItÂ´s more powerfull than you thought !    I already have this Knoppix CD 6 month ago, but I didnÂ´t use it often.  Lately, I found that somehow this Knoppix just save my life.  It helped me much when my Windows was broken, attacked by virures, and today when I almost lost my internet connection (due to download quota), it helps me again.  Now,  I logged into campus server and run mozilla from console.   No worry about quota anymore :D  Love you KNOPPIX !!!  '),(10480,1758432,'18,April,2004','2004-04-18 00:00:00','   Sailing through the Rhine river   My dream comes true. I have been wondering for this about year ago, as I heard about the Loreley story from Deutsche Welle German course.  Not only that, from the ship we also can see many castles, tower and ruins from 16th century.   Below quoted a poetry about the Loreley story, written by a famous German author, Heinrich Heine.   Die Loreley von Heinrich Heine (1823)   Ich weiÃŸ nicht, was soll es bedeuten, DaÃŸ ich so traurig bin; Ein MÃ¤rchen aus alten Zeiten, Das kommt mir nicht aus dem Sinn.  Die Luft ist kÃ¼hl und es dunkelt, Und ruhig flieÃŸt der Rhein; Der Gipfel des Berges funkelt Im Abendsonnenschein.  Die schÃ¶nste Jungfrau sitzet Dort oben wunderbar, Ihr goldnes Geschmeide blitzet, Sie kÃ¤mmt ihr goldnes Haar.  Sie kÃ¤mmt es mit goldnem Kamme, Und singt ein Lied dabey; Das hat eine wundersame, Gewaltige Melodey.  Den Schiffer, im kleinen Schiffe, Ergreift es mit wildem Weh; Er schaut nicht die Felsenriffe, Er schaut nur hinauf in die HÃ¶h\'.  Ich glaube, die Wellen verschlingen Am Ende Schiffer und Kahn; Und das hat mit ihrem Singen Die Loreley getan.    The Lorelei (English Version) by Heinrich Heine, 1823   (note: English version NOT translated by me)  I cannot tell why this imagined  Sorrow has fallen on me  The ghost of an unburied legend  That will not let me be.   The air is cool, and twilight  Flows from the quiet Rhine;  A mountain alone in the high light  Catches the faltering shire.  One rosy peak half gleaming  Reveals, enthroned in air,  A goddess lost in dreaming  Who combs her golden hair.   With a golden comb she is combing  Her hair as she sings a song;  Heard and reheard in the gloaming  It hurries the night along.   The boatman has heard what has bound him  In throes of a strange, wild love.  He is blind to the reefs that surround him,  Who sees but the vision above.   And lo, the wild waters are springing -  The boat and the boatman are gone...  Then silence. And this with her singing,  The Lorelei has done.  '),(10481,1758432,'11,April,2004','2004-04-11 00:00:00','   Trip to Trier   Trier, the oldest city in Germany and also known as second Rome.  Because of eastern holidays, a lot of friends, didnÂ´t join us. There were only 6 master-plus students, 7 Erasmus students, and  3 other from IntClub. Unfortunately there was also no guide, blurry fotocopied map and some papers were the only things we have !   As we separated at Porta Nigra (RomeÂ´s gate, about 2000 years old), we walked to Hauptmarkt to begin our journey :p  Then Nicole (from IntClub) joined and guided us to TrierÂ´s Dom. Like usual, we took pictures (the most important things to do, right ? :P)  even inside the dom and the cemeteries. It was funny cause I thought that the cemetery was a park. After that, the trip continued to Basilika and KurfÃ¼rstlich Palais. Basilika is a church of Kaiser Konstantins, which is build just behind his palace. Next station, Kaiserthermen, ruins of  RomeÂ´s emperorÂ´s hot bath (Warmbadesaal). We just passed through Thermen am Viehmarkt, another ruins of hot bath, on way to Karl Max Haus.    Dying for hungry, we considered to find a place to take rest and have lunch. The street was so crowded with locals and tourists who have lunch or just drink beer. While looking around we found Brunnen and Judengasse. At last we found bench, sunny and facing to Porta Nigra, so nice !   Still have about one hour after lunch, we back to Hauptmarkt to find more places such as DreikÃ¶nigenhaus,  Rotes Haus and Frankenturm.  Finally the time was over,  we must back to Kaiserslautern.    Windows installation  a night long :((  '),(10482,1758432,'09,April,2004','2004-04-09 00:00:00','   things to do (on this Osternferien)    - complaining computer`s hardware - cleaning worm virus - excursion to Trier ( and publishing photos after that) - manage emails and moving website - finishing hiwi`s assignments  '),(10483,1758432,'07,April,2004','2004-04-07 00:00:00','   problems...   why things didn`t run like they suppost to be ? or something wrong with me ?  '),(10484,1758432,'04,May,2004','2004-05-04 00:00:00','   Tuesday has come....   My work still not finished yet !!!! what to do now ? :((  '),(10485,1758432,'03,May,2004','2004-05-03 00:00:00','   Many things happened...   Morning, wash the clothes in Alcatraz, after that just stuck in my room and do some work on programming, but not yet finished until now :( In the aftenoon (accompany with Joe) I  bought cutie pinky folder and some other stationaries also with beautiful colors.  After that  went to the German class in the evening and I was late to catch the bus to home, so I and Lei Ling have to wait for more 1/2 hour.  At home I cooked `gorenganÂ´ and got my internet repaired by a friend. :P  hahaha... today my writing seems like German homework in Perfect Tense :P  '),(10486,1758432,'01,May,2004','2004-05-01 00:00:00','   Last Samurai   I watched Last Samurai film for the second time. Cool... :D  '),(10487,1758432,'01,May,2004','2004-05-01 00:00:00','   Saturday night   Alone again on this night....  I found the song : She believes in me from Kenny Rogers, very deep meaning then... feel so mellow....    She believes in me   While she lays sleeping, I stay out late at night and play my songs And sometimes all the nights can be so long And itâ€™s good when I finally make it home, all alone While she lays dreaming, I try to get undressed without the light And quietly she says how was your night? And I come to her and say, it was all right, and I hold her tight  And she believes in me, Iâ€™ll never know just what she sees in me I told her someday if she was my girl, I could change the world With my little songs, I was wrong But she has faith in me, and so I go on trying faithfully And who knows maybe on some special night, if my song is right I will find a way, find a way...  While she lays waiting, I stumble to the kitchen for a bite Then I see my old guitar in the night Just waiting for me like a secret friend, and thereâ€™s no end While she lays crying, I fumble with a melody or two And Iâ€™m torn between the things that I should do And she says to wake her up when I am through, God her love is true...  And she believes in me, Iâ€™ll never know just what she sees in me I told her someday if she was my girl, I could change the world With my little songs, I was wrong But she has faith in me, and so I go on trying faithfully And who knows maybe on some special night, if my song is right I will find a way, while she waits... while she waits for me!  '),(10488,1758432,'09,June,2004','2004-06-09 00:00:00','   my math result 10/16, so bad, because most of my friends got 14/16, and this result means not satisfied....  '),(10489,1758432,'06,June,2004','2004-06-06 00:00:00','   Score    Eine Woche gespielt : 5 gegen 3 fÃ¼r mich.  '),(10490,1758432,'05,June,2004','2004-06-05 00:00:00','   Van Helsing  I bought new Van Helsing DVD in TrÃ¶delmarkt in Paderborn.  '),(10491,1758432,'26,July,2004','2004-07-26 00:00:00','   Since last week I have a TV. Not new but old, as inheritance of one of my friend who leaves this city.   I am really happy for it because I can watch a lot of tv-series, films and some reality shows . It help to improve my German, but also took a lot of my time instead. Na ja, everything have both good and bad side isn\'t it ?  '),(10492,1758432,'08,July,2004','2004-07-08 00:00:00','   This article is taken from www.howstuffworks.com  If there is water in a heated container and the water is boiling (at 100Â°C or 212Â°F), the container will maintain a constant temperature. As soon as all of the liquid water has evaporated (all of the water is absorbed by the rice), the temperature inside the container immediately rises. The appliance has a thermostat that can detect when the temperature rises above 212Â°F in the container, and it turns itself off.  '),(10493,1758432,'07,July,2004','2004-07-07 00:00:00','   Mildew can grow fastly anywhere, in nature and also inside our house. In the nature, it has important task which is decomposing organism substance. But inside the house, high concentrate of mildew could be dangerous for our health, because of its poison. Mildew live in environment with high humidity. To avoid mildew we have to control the humidity inside the house always under 60%.  In winter, we can do \'crush ventilate\' by open all windows and door very wide for short time (5-15 minutes). No sense to tilt the windows all the day. It will loose a lot of energy and the house would be very cold, but the chance of mildew growth would be bigger.  '),(10494,1758432,'06,July,2004','2004-07-06 00:00:00','   Today we did experiment on stochastic. It is quiet new for me since I never learn about it before. It deals with probability in random area. I didn`t understand completely about the experiment but I\'ve got some idea of autocorrelation. Autocorrelation is simply convolution of two same signals (one of them is shifted) over its period. Autocorellation of sine signals will output cosine signal and autocorrelation of rectangle signals will result triangular signal.   One implementation we did is to find velocity (in rpm) of the machine with help of noise sound from machine that was recorded on casette. The signals are autocorrelated, and the output will have the same periode of the signal itself. With the period finally we can calculate the velocity of the machine.  '),(10495,1758432,'06,July,2004','2004-07-06 00:00:00','   examinations schedule :   22.06 : measurement technique 28.06 : electromagnetic compability 30.06 : mobile communication 02.07 : DSI (german exam)  '),(10496,1758432,'06,July,2004','2004-07-06 00:00:00','   www.dradio.de/wissenschaft interesting link ! you can read the news and hear it as well. check the right part of text, there you will find link to download the hear-text in mp3 format.  '),(10497,1758432,'06,July,2004','2004-07-06 00:00:00','   7 vc. 5 (cmiimw) :P  '),(10498,1758432,'06,July,2004','2004-07-06 00:00:00','   it is true ! instead of using ear, elephants can \'hear\' infrasound under 15 Hz in distance until 40 km. more ? later ok ?  '),(10499,1758432,'02,August,2004','2004-08-02 00:00:00','   0-0 (7 times)  '),(10500,3667367,'17,June,2004','2004-06-17 00:00:00','   Last night all the ladies of lovely linda street gathered for a dinner that included veggie soup, pita pizza, (free) bread and roasted garlic, and yummy vegan molasses-ginger cookies for dessert (and several bottles of wine). We all toasted to the final months of our household, brainstormed costume-concepts for the final linda st. fiesta, and hashed out some of our domestic issues. I think it ended with everyone feeling okay about our final months together, and the idea that we\'d try to enjoy them instead of squabbling about dishes and toilet paper. Not to say that this living situation has been difficult, but we all have our moments, especially in a house of 5 women. Overall it has been an amazing experience for me, and every one of my roommates has taught me something. As much as I can\'t wait to live with Jeff, I am going to miss having such a powerful female environment around me (that\'s the new age Amber talking).   '),(10501,3667367,'17,June,2004','2004-06-17 00:00:00','   Tuesdays are better than Mondays, but only slightly. I still feel like the weekend is far away, but at least I\'ve taken the first step towards it in getting through yesterday. My Monday turned out being better than your average Monday. Despite feeling spent from my busy weekend, I managed to be a productive worker bee yesterday and get a lot done. I also did a pilates class at lunch and my sore abs are feeling its effects/benefits today. Last night I met my roommie for dinner at ChaChaCha, a place we tend to frequent quite often, and we ate our favorite platanos maduros and chicken quesadilla, washed down with a few pitchers of sangria (they were little pitchers, I swear), and had a long talk. We came to the realization that the platanos are almost better with whipped cream (thanks to the bartender\'s advice) than they are with black beans. Oh, and we also concluded that our issues are resolved and we\'re buddies again. Then we went home and did interpretive dance to Tori Amos music with our other roommates. Good times!   In running news, I have none to report, because I suck. I hope you all won\'t hate me when I totally flake on my commitment to run that half-marathon in August. I just don\'t see that happening.   Tonight.... Lovely Linda St. Roommate Dinner and maybe a movie. And Jeff leaves for Texas tomorrow :-(   '),(10502,3667367,'17,June,2004','2004-06-17 00:00:00','   I\'m at a loss for an inspired title right now as it is Monday again, and my weekend has steamrolled over me and left me exhausted, broke, and unprepared for another full week. Not one, not two, but three friends from out of town were visiting this weekend, Fahrenheit 911 opened and I schlepped all the way to Daly City to see it, pride revilers and their festivities took over the city and (at times) my neighborhood - birthdays were celebrated, much food and alcohol was consumed, and mostly fun was had (I say mostly b/c I did have some drama with a friend :-(). Now I am back at work, dealing with Monday\'s abrupt return to reality. Not that this week is totally without hope - there is a three day weekend to look forward to (will anyone else be around?) and I get to spend it with two of my favorite people - the chico and his dog Casey. To get myself through the week I have 3 new netflix on their way, a roommate dinner tomorrow night, drinks on Wed. with friends, and many yoga classes to attend throughout.   I should also be spreading the word that Fahrenheit 9-11 is amazing and very important for everyone to see. It was a moving film, and it lacked the typical Michael Moore obnoxiousness. Jeff has offered to buy you a ticket if you\'re feeling reluctant, so hit him up and go see it!   '),(10503,3667367,'17,June,2004','2004-06-17 00:00:00','   I\'m being really lazy today and laying around, nursing my sore neck and watching netflix. I didn\'t make it in to work today, and it seems that neither did two other chicas. Its just not a day for working - sorry to those of you out there slaving away at your desks.   I\'m looking forward to the movie tonight - come back tomorrow for a review!   '),(10504,3667367,'17,June,2004','2004-06-17 00:00:00','   The #15 Thai Chicken from Specialty\'s is sooo delish. And if you\'re there, their cookies aren\'t too shabby either. Seriously people, a chicken breast, avocado, cucumber, carrots, lettuce, and yummy thai sauce... too good! A little messy, but worth the extra napkin effort.  I\'ve got mine at my desk right now and am LOVING it!  Oh and my neck is feeling better. I went to yoga again today (I know, maybe not the best idea, but I was careful) and my instructor told me how to prevent straining it any further.   And a shout out to my medicine chicas that got me through our run last night by being goofy and jumping over concrete benches along the embarcadero. I love you both.  And check out a website one of the chicas told me about www.about-face.org for a non-profit that seeks to change negative images of women in our culture.   '),(10505,3667367,'17,June,2004','2004-06-17 00:00:00','   I\'m feeling rather poo poo today b/c I did something stupid to my neck and its hurting. I can\'t turn my head to the right without sending pain down through my neck and back. I don\'t want to bore you with my whining... In other news, it looks like its going to be another busy weekend with a lot of visitors in town. Of course, I\'m heading out to see Fahrenheit 911 on Friday night with the rest of the liberal masses. My question is; will this film be successful in affecting the upcoming election if only liberal sympathizers go see it? I really hope that its message also reaches those that haven\'t thought about these issues and not just in the form of a strong backlash. Someone recently told me how surprised a lot of people are to hear that none of the hijackers on 911 were Iraqi. I think a lot of people are under the misconception that they were. And its those types of ideas that the Bush administration has managed to inadvertently plant in American minds to justify their Iraq invasion. Stoopid Bush! Wait, will the FBI monitor my blog if I say things like that?   '),(10506,3667367,'17,June,2004','2004-06-17 00:00:00','   My crrrrazy roommate has asked me to post her list of desired b-day gifts. If you ask me, just get her a book of fart jokes and she\'ll be entertained for hours!  apes\' birthday list  (when in doubt, always go more expensive) 1. thesarus - need synonyms for rad 2. mini massager - for those late night neck pains 3. diet coke - feed a girl\'s addiction for her birthday 4. sumyungboi 5. all your money will do 6. charmin - for my charmin\' personality :) 7. visit amber\'s blog everyday  There you have it -- 7 gift ideas for the girl born on the 7th day of the 7th month --  Your very own, hard shell soft on the inside cancer crab :)  -- see y\'all at Cha Cha Cha!   '),(10507,3667367,'17,June,2004','2004-06-17 00:00:00','   I\'m trying not to let the caca things about my day bring me down and focus on the positives, which have been so far:  - I did a headstand in yoga! - Good oatmeal, wheat germ, chocolate chip cookie at lunch - 3 new Netflix DVD\'s on their way - Cute picture of Casey (the dog) with the above quote on my dry-erase board - Daydreams about the potential cohabitatin\' that seems to be happening in October - My wonderful friend returning to the bay area to work at Eccolo in Berkeley  Those are a lot of nice things to think about. I feel better already.   '),(10508,3667367,'17,June,2004','2004-06-17 00:00:00','   It was such a nice (and super busy) weekend, did it have to end so quickly? Friday I made my super delish corn chowder for dinner and my roomie made yummy chicken veggie pizza and we watched \"Spellbound\" with the chico. Then we watched \"50 First Dates\" and I happily curled up with the chico, and slept through the entire movie. Saturday I did some great power shopping with \"Charlotte\" aka Reno. I scored an awesome pair of jeans at Crossroads and even got some $$ in trade for some old clothes of mine. Then we were party planners extraordinaire for our beautiful friend\'s b-day gathering that evening. The chicas proceeded to eat, drink, and be hysterically merry all around the mission area. Sunday was father\'s day and the chico and I headed up to Santa Rosa for a BBQ with the fam. Busy, busy weekend. But what I did not manage to do this weekend is one single minute of excercize... my running plans fell through when I opted for banana walnut pancakes at Polker\'s instead. Is this half marathon really going to happen for me? I\'m feeling doubtful, send me encouraging comments!   '),(10509,3667367,'17,June,2004','2004-06-17 00:00:00','   On my seventh birthday my dad bought me a new bicycle. He was currently residing in a motor lodge, near a freeway entrance to the 101, while my parents settled the domestic matters of their impending divorce. On the afternoon of my birthday, my dad took me to Toys\'R\'us to choose my new wheels. Down that long isle of gleaming bicycles I searched for my match. The one I finally chose had a purple banana seat and tassels hanging from the handle bars. He had to assemble it in the parking lot of the motel with new tools purchased from Kmart that morning. I don\'t remember a lot from my 7th year, but I distinctly recall riding that newly assembled bike in laps around the parking lot.   '),(10510,3667367,'17,June,2004','2004-06-17 00:00:00','   I did a good thing for someone today. At least one good thing that I can think of, maybe I even did more that I\'m not remembering.  The person I did this favor for is someone who should have a significant karmic surplus due to the large number of thoughtful and giving things he does daily - not only for me, but for lots of people (and occasionally animals too). I can only hope that this one favor helps tip the karmic scales more in my favor.   '),(10511,3667367,'17,June,2004','2004-06-17 00:00:00','   In the lovely linda house we are now members of the Terra Firma CSA (Community Supported Agriculture) and we are receiving a large box of organic, seasonal produce every week. Now on our second week, we have accumulated 12 lovely and fresh ears of corn - just waiting to be boiled or bbq-ed, speared through the ends with little corn forks and nibbled in long rows like a typewriter (ding!). I\'ve been trying to come up with other corn ideas all day and have decided to make summer corn chowder tonight, also utilizing some of the potatoes, red onions and summer squash we have around. I\'ll let you know how it turns out!  Check out the link and sign up for your own box!   '),(10512,3667367,'17,June,2004','2004-06-17 00:00:00','   I have recently become so much more touchy feely than I ever was before. Today I will have my new age moment during my lunchtime restorative yoga class. I will lay in a dark room, using Styrofoam blocks and blanket bolsters and contort my body into pretzel-like positions. If I\'m lucky, the instructor will come around with a vial of lavender oil and sprinkle my towel so that I can inhale the scent as I lie there and breathe. I will stretch and inhale deeply and feel my head clear from all the activity, stress and fatigue of my daily routine. I will end the class in a long meditative state (sometime with my legs up a wall) and be awakened by the chime of a bell. When I return to work my body will feel re-energized and my cubicle will feel a little less repressive than usual.   '),(10513,3667367,'17,June,2004','2004-06-17 00:00:00','   Besides being totally excited about my new little blog, I\'ve also been preoccupied this week with my new subscription to netflicks. The decision was made recently in the lovely linda household to downscale our cable (RIP HBO). So not only am I going to be hopelessly behind in my viewing of HBO\'s amazing programming, I\'m also now missing all my other wonderful movie channels like Encore and IFC. And I must admit I\'m a bit of a TV and movie whore. So I gave myself a little gift of Netflicks and in 5 days have already \"queued\" 150 movies! It could take me months, or even years to get through all those and I\'m looking forward to each and every one. This week alone I\'ve watched Bowling for Columbine and Somethings Gotta Give. I enjoyed both and was surprised in my reaction to both. Bowling because although I agree with a lot of what Michael Moore\'s message is, he can be really sensational and annoying. I don\'t remember reacting that way to Roger & Me, but I guess its different now that he is a celebrity in his own right. I loved that he used the Dead Milkmen\'s \"Take the Skinheads Bowling\" as an intro, such a great song! I am looking forward to seeing Fahrenheit 911, and hoping he keeps his promise of having less camera time for himself. I do expect more of his eye-rolling theatrics, but don\'t get me wrong, if that\'s what it takes to get his message out there, I\'m a fan.  And somethings Gotta Give surprised me by being more sweet and less formulaic than I had imagined it would be. And it got me thinking about love and age, I mean why shouldn\'t people be able to find love again later in life? And stupid men that date girls their daughter\'s ages are missing out on the beautiful depth, wisdom and maturity of women their own age. Sure, our bodies will age, but can\'t we see past that?  Anyway, more movies to come soon, I\'m expecting the Station Agent and Talk to Her, two reportedly great movies that I missed in the theatres.   I love you net flicks, I love you!   '),(10514,3667367,'17,June,2004','2004-06-17 00:00:00','   I\'ve succumbed to the blogging mania that is sweeping the nation and created my very own spot in internet space for my random musings. Why you might ask, do I feel the need to launch my thoughts into the web realm for anybody (or possibly nobody) to stumble upon? Is this just pure narcissism, mixed with an abundance of free time? Yeah, maybe it is a little of that. Yet, I also feel like I might just have my own unique perspective on life that I feel compelled to share.   And really, what could possibly be more interesting than recounting the adventures of La Sassy in SF?   '),(10515,3667367,'17,June,2004','2004-06-17 00:00:00','   Sorry this has been slow in the making... things have gotten super busy in my life lately. More on that later. Last weekend\'s slumber party was so much fun, and had elements of slumber parties past (like junk food, dancing, games) and new elements too (like NO bitchiness or cattiness). Of course, there was the element of alcohol as well, which I never remember having at any of my slumber parties as a kid - really Mom I swear! As every guest arrived at the wonderful home of my chica friend, Michelle, we changed into our PJ\'s and enjoyed a cup of hot cocoa, with a little kick to it. Michelle\'s house has a big living room with comfy, plush carpet and tons of huge throw pillows... so it was a perfect area for sleeping bags and lounging. We ate a ton of junk food including my indulgence of Ruffles Cheddar and Sour Cream chips, and animal cookies (the pink and white kind), and pizza and tacos. Then we had an impromptu dance party in the kitchen when someone put the Beyonce album on, and that turned into a limbo contest. After we were pooped from the dancing we plopped down on all the bedding and pillows in the living room and played I Never and Truth or Dare. Kudos to Lianna for her chonie run down the block! I attempted to get a Ouija seance going, but SOME people just couldn\'t take my efforts to contact the spiritual world seriously. And you know the spirits aren\'t down with giggles! After all that, a few of us got into a game of \"would you rather...\" which turned rather dirty, and I have to admit, pretty imaginative. Common themes were genital relocation or removal, incest, bestiality, and consumption of bodily fluids. I won\'t give you any more further details, but I think you get the point that this was a hard core game! April and Enrique really pushed the envelope a bit.&nbsp; At that point I sort of dozed off with the images of our \"would you rather...\" game floating through my head, and woke up intermittently from about 2 am to 4 am b/c of either the movie \"Shag\" on TV, or the mosquito that was eating me alive. Basically, it was your typical sleepover night of no sleep, and girls started stirring around 8 to make breakfast, drink more, be silly, etc. Unfortunately, I was off to work for the day, but was amazingly not feeling awful. All that fun and silliness put me in a good mood for the day, and I made it to 5 pm without curling up under my desk. I must say that I\'ve renewed my love for slumber parties and hope to host or participate in another one soon!   '),(10516,3667367,'17,June,2004','2004-06-17 00:00:00','   Okay - so I have heard from everyone I\'ve asked that Dirty Dancing, Havana Nights was just plain awful. Although I was initially excited to see a sequel to one of my all-time favorite movies, after hearing the reviews I didn\'t go see it. Thanks to netflix, I was able to rent it and watch it last night and I have to say that it wasn\'t nearly as bad as I\'d heard. In fact, I would say it was fairly cute and fun. Not Oscar deserving or anything like that, but very decent fluff. I felt the two characters had good chemistry and the romance was very sweet. The dancing and the music, much like the original, combined music of the time period (1958 Cuba) with more contemporary style music. I was sucked in from the start. I would even consider owning it on DVD and watching it on lazy days... like I do with the original. One thing I\'d have to criticize it for, is its lack of a really good finale dance number like the first one. And NO lines as good as&nbsp;\"Nobody puts Baby in a corner.\"   '),(10517,3667367,'17,June,2004','2004-06-17 00:00:00','   In a throw back to elementary and junior high days, my chica friends and I&nbsp;are throwing ourselves a slumber party tonight. On the agenda is plenty of junk food consumption, bad eighties/nineties movies, lots of dancing (you know, to teach each other out new moves), games like \"Truth or Dare\" and \"light as a feather, stiff as a bored\", prank calling, and probably naked pillow fights. Okay maybe not \"naked\" but at least \"pj-clad\" pillow fights. The last slumber party I attended was my pledge retreat during my sorority pledge semester during my freshman year.&nbsp; My highlight of that night was streaking through the streets of Montclair at 2 am! Before that, slumber parties usually didn\'t involve alcohol, and nudity was confined to trying to remove your training bra without taking off your pj shirt. I remember A LOT of overnight birthday parties I attended as a pre-teen where the group split into two rooms either because of a fight (probably over some deep issue like which New Kid was cuter Jordan or Joe) or because we were forced to split up by some sleep-deprived parent who didn\'t want to hear another peep out of us. I remember times when I felt left out and pissed off at the group, and other times when I made someone feel that way. I also remember trying to make someone\'s younger cousin pee in her sleeping bag by putting her hand in warm water... and I definitely remember freezing bras... Why is it that little girls are so damn cruel to each other? But we\'ll have none of that tonight, I\'m sure! ;-)   Wait for the re-cap tomorrow...   '),(10518,3667367,'17,June,2004','2004-06-17 00:00:00','   Another busy and wonderful weekend is gone and another long week has started. Jeff and I were blessed with the presence of Casey again, but this time we carted him over to the east bay to stay at Jeff\'s place in Montclair. We watched a really interesting (and steamy) movie on Friday, \"The Dreamers\", that is Bernardo Bertelucci\'s latest film.&nbsp;That was the only steamy and exciting aspect of our Friday night, because&nbsp;I had to get up bright and early to take a practice LSAT on Saturday at the Kaplan in Berkeley that was so horribly disorganized, I&nbsp;felt completely validated in my decision to handle my LSAT&nbsp;studying on my own. After 3-hours of strenuous&nbsp;mental activity, all I wanted to do was re-fuel and relax. So, Jeff, Casey and I headed down to 4th street to eat yummy tacos at Tacubaya and shop the anthropogie and other 4th st. shop sales. Down there we encountered the usual kitty and puppy adoption booths with their crates and pens of adorable animals, just yearning to go home with you.&nbsp; One in particular, a 10-week old rott/shepard-mix looking mutt boy, was especially adorable and needing to go home with us. But, we resisted and walked away. I know there\'s another puppy out there for us... and hopefully soon we\'ll find him/her :-( So, being sad about leaving that little guy, I was luckily cheered up with some gelato from Eccolo. I stopped in to say hello to Samin and she gave me a taste of peach and honey gelato. The honey flavor was incredible! My new favorite. We then helped Samin move a couch into her new fabulous Berkeley abode and we all got spiffed up for our dinner at Eccolo. That dinner should be a whole other blog unto itself, but let me just say we ate a TON of fabulous food, with a big group of friends and had a really nice, totally grown-up style evening! Sunday I studied LSAT for several hours (because I rock and because I had a few hours to kill at the ferry building waiting for my mom who had MISSED her ferry!). Then my mom and I went to see the Geisha exhibit at the Asian Art museum. It was a really wonderful exhibit, and I do think the Geisha culture has a lot of beauty and intrigue, it does seem like a tradition that caters to men\'s fantasies, which seem to&nbsp;be a subservient, flirtatious, shallow (as in keeps the conversation \"light\"), etc. I think they are wonderfully talented in the arts that they learn, but it bothered me watching them giggle at these ridiculous men and play the role for them that they are hired to perform. I suppose its like acting, and they can separate who they really are from their Geisha character, but it would be very hard for me to do! Anyway, my mom and I had a nice time together and ended it with amazing gelato from Ciao Bella at the ferry building. I had lemon and rose - they were both amazing! Then I went home and studied more LSAT with Jeff and Casey. Casey is really good at the reading comprehension!   '),(10519,3667367,'17,June,2004','2004-06-17 00:00:00','   This week has flown by, and I\'ve been so busy I\'ve hardly had a chance to post to my lovely blog. I apologize to all my loyal readers - I hope I haven\'t lost you from my laziness. For reals, I\'ve hardly been lazy at all. I decided last week to take the LSAT on October 2, and because September is going to be a crazy month for me (going to Denver, finding a new place&nbsp;with my LTR, and moving), I am trying to study as much as I can now. Last weekend Jeff helped me set up a table in my room to serve as my study area, and a few of my fellow LSAT-taking friends have loaned me study materials from their courses. For those that aren\'t familiar, here\'s a brief education in the LSAT - it consists of 3&nbsp;types of questions&nbsp;(4 sections): 1 section reading comprehension, 1 section games (not like candy land or anything like that), and 2 sections arguments, plus some writing sample that apparently no one reads, and its all out of 180 points. Anyway, so I am pouring over methods to attack these problems, and working on as many practice problems as I can handle. Tomorrow morning I\'m taking a full blown, proctored, practice test to see how I\'m scoring.&nbsp; I have no idea what goal to set for myself because I have not idea what I\'m capable of scoring. Plus, since I don\'t know where I\'d want to go to school (if I do decide to go to law school) its hard to know what I need to get. I guess I just want to do the best I can possibly do and kick that LSAT\'s b-u-t-t, but we\'ll see. I guess a 180 would be good, wouldn\'t it? But really... I think a 165 is a solid, good, impressive score and would make up for my not totally impressive GPA. So far, I think I\'m doing well on the arguments (are any of you suprised about that?!), pretty well on the reading comprehension, and so-so on the games. Ironically, I LOVE doing the games and look forward to practicing them, b/c they\'re so fun... they can just be so hard. The \"games\" are usually those types of logic puzzles where you have to figure out who brought the cheese dip to the party and if Betty&nbsp;arrived before or after Suzie&nbsp;&nbsp;... things like that. I\'m sure that\'s totally a useful skill to have as a lawyer when you\'re organizing the office potluck.  &nbsp;  Another aspect of my LSAT studying is my decision not to shell out a gazillion dollars (or a thousand) to take a prep course from Kaplan or Princeton Review (and that\'s no diss to you Dre), or whoever. Don\'t get me wrong, I think they\'re valuable, but that\'s just too much damn money right now and I think that you can get the same preparation if you are disciplined. And so far I am. All I want to do right now is practice LSAT b/c, like I said, I want to kick some booty on it. I\'m motivated right now, so if I can keep that up, I think I\'ll be good.  &nbsp;  Okay I\'ve rambled on enough.&nbsp;Send me any advice you have for standardized test&nbsp;prep, or your LSAT anecdotes, or ultimatums that you\'ll disown me as a friend if I&nbsp;become a blood-sucking leech of a lawyer (I swear I won\'t, I\'ll only use the law to help people and do good in the world&nbsp;:-) ).&nbsp;   '),(10520,3667367,'17,June,2004','2004-06-17 00:00:00','   On Saturday I went on the most ridiculously hilarious wild goose chase through Sonoma, looking for Cannard Farm, the farm ran by \"Bob\" that grows almost all of the produce for Chez Panisse. Jeff has never been there, I have been there once, and Samin has been there between 8 and 20 times. Bob, the farmer is rumored to be a bit of a hermit, but takes care of an amazing beautiful plot of land in the Sonoma valley and cultivates some of the most delicious produce I\'ve ever had. According to Samin, the farm is located past a fork/triangle type intersection, near two churches, in a stoplight-less span of road, by some trees, and on a road that is either alta mira, sobre vista, or something linda. We followed our instincts (mostly Samin\'s - altering from \"this is defnitely it\" to \"we\'re totally in the wrong place\" back to \"we\'re close jsut keep going 5 more minutes\") for almost an hour and thought we had most definitely found the driveway to the farm when we debated whether or not it was appropriate for us to visit a quasi-aqaintance of Samin\'s at 7:30 pm on a Saturday night. We hadn\'t really considered that part of it yet. We decided to drive down the road and Samin would duck, and Jeff and I could just be lost tourists. So we headed down the road with Samin screaming \"Bob might get mad\" and \"he might throw things at us\" when we encountered the end of the road and a sign saying it was Patty and David\'s house ahead, or something like that. So, Bob\'s farm is not on Sobre Vista Dr. We turned around and headed home, though Samin swears it must have been there somewhere. For an almost-adventure, it was pretty wild!   '),(10521,3667367,'17,June,2004','2004-06-17 00:00:00','   Sunday at 2:00pm in Dolores Park the SF symphony is performing a free concert. I plan on going a little early to stake out a good spot. Give me a call on Sunday if you want to meet up. Maybe we could all bring some wine to share?   '),(10522,3667367,'17,June,2004','2004-06-17 00:00:00','   Ow - my head hurts. I got a little carried away in the celebration of Aprils 26th year last night and maybe over did it a teeny tiny bit on the minty mojitos at chachacha. It was a huge group, at least 20, of all her various friends (roommates, college friends, her crrrrazy co-workers and both her bosses). My favorite part of the night had to of been when she introduced the bartender to her boss as \"one of the guys she\'s hooked up with\". He was mighty flustered and brought on the free drinks after that! April, Jeff and I were the last three left at the bar after dinner and Apes and I slurringly confessed our undying friendship to each other - I think Jeff observed it all in horror of our girly emotionalness. Now I\'m suffering the consequences, but also feeling like it was worth it. I do have to say that I\'m vowing to avoid ChaChaCha for at least a week or so!   '),(10523,3667367,'17,June,2004','2004-06-17 00:00:00','   I feel like this week is just starting for me and its already Wed. Not that I\'m upset that the weekend is almost here again! Although I did work yesterday, I didn\'t get home from Sausalito house-sitting to casa de las lindas until last night. I was really happy to get back to the city and home to my lovely linda roommies, my own bed, fridge, closet, etc... but I must say that Jeff and I had a really wonderful weekend in Sausalito. A looooong weekend since we were there since Thursday. We had such a great time, I don\'t even know where to start... In no particular order and because I seem to enjoy making lists so much here are some of the reasons my weekend was wonderful...  First , I just love the Engel family dog, Casey. He\'s my first real and true dog love and now I am certifiabley dog obsessed. We spent a lot of the weekend playing with him, petting him, taking him out, buying him treats (at fancy schmancy George on 4th st) and just generally ogling him. I wish that we could have a dog. My parents are getting a new puppy named Bailey this week, so hopefully I will fall madly in love with him too and can hold off my own dog urges for awhile. Is this what it feels like to have baby urges? Right now I\'d much prefer a puppy!  Second , Jeff and I spent much of the weekend laying around, playing scrabble, eating yummy food (ahi fish tacos!), drinking wine, doing crosswords and just having some much needed down time together. On the 4th we mixed margaritas in water bottles and walked with Casey down to the waterfront to watch the Sausalito fireworks, it was really nice.   Third , I saw my wonderful friend Samin in Berkeley and after Thai Temple brunch, we got to see the new restaurant she\'s working in, Eccolo. Being around Samin always makes me feel positive and creative and I missed her a lot while she was gone, so it was great to be spending time with her again. I can\'t wait to do all kinds of fun projects with her and to visit her at Eccolo soon!  Fourth , I got to spend time with my family on Saturday, eat hot dogs, and watch Jeff being dragged around by my 3 year old nephew who cried when he left. The only down side to that is my that my entire family keeps telling me \"Jeff is soooo good with kids\" so they either want us to babysit or procreate immediately. C\'mon people, let\'s put the horse before the cart (or however that particular saying goes!)  As much as I wish our life were like that (suburban and domestic and mellow) all the time, I am really looking forward to spending a few more years in the city and enjoying the young, urban life (does that mean I\'m a yuppie?!). I especially want to enjoy these last three months in the mission before Jeff and I nest in our new place. The mission can be such a wonderful area to be in, I know I\'m going to miss it. It can also smell like pee and vomit, so in other ways, I\'m ready for something new!   '),(10524,3667367,'17,June,2004','2004-06-17 00:00:00','   I really appreciated your comments last week when I was feeling a little caca. Not to sound trite, but I really feel lucky to have some many wonderful and different friends. Life would really be caca without you!   '),(10525,3667367,'17,June,2004','2004-06-17 00:00:00','   Today is Miss Godette\'s 26th birthday - so send her b-day wishes, or even better prank phone call her or leave really long messages on her machine! Her number is 565-0505!   '),(10526,3667367,'17,June,2004','2004-06-17 00:00:00','   I somehow deleted my blog last weekend, so this one is brand new, without any fancy customizations. I will get the \"tonight on netflix\" and other fun sidebars up and running again, just as soon as I read about html again. Who knew blogging could be so complicated?  And more about my wonderful weekend later. I\'ll leave you with my funny moment of the day, I heard on the radio that today is George Dubya\'s b-day and I heard a funny song entitled \"faithless\" with the lyrics \"fear is a weapon of mass destruction\". So clever, and so true.   '),(10527,3667367,'17,June,2004','2004-06-17 00:00:00','   Yesterday I was in one of those moods where every little thing got on my nerves - including not remembering to buy my new fast pass and having to pay for my morning bart ride, then waiting in a ridiculously long line after work to buy it. Then I had an annoyingly long drive down Van Ness in traffic (hitting every stop light and stuck behind some idiot riding their breaks the whole way) on my way to Sausalito. However, there were a few little things that really brightened my day yesterday and I\'m feeling the need to share:  1) I held a Crow position in yoga for a few seconds. This may not sound like much, but Crow is super hard for me b/c it involves strength and balance, not my top skills. For those that don\'t know Crow is balancing your knees on your elbows, supporting your weight on your hands, and lifting your feel off the ground. Its amazining how every teeny tiny advancement in yoga makes me feel fantastic.   2) We got tomatoes in the terra firma box yesterday! We got a mix of what I think are early girl, cherry, something yellow (taxi or something else), and a brandywine (is that what they\'re called)? It\'s an education in tomatoes at my house! I used two in a veggie pasta I made last night, though I think they should be eaten just raw to be fully appreciated. The red ones are perfectly red all the way through and sweet. What are those weird tomatoes with white insides that they sell at the supermarket? Obviously not tomatoes. Now that I\'ve experienced so much organic, perfect produce, its hard to imagine shopping at Safeway or something like that. Forgive me for my food snobbery!  3) I \"found\" a funny story someone wrote on the back of a flyer on the street yesterday. I have recently discovered Found Magazine, and spent the past month with my eyes pointed downward looking for some discarded note or picture, but hadn\'t \"found\" anything until yesterday. I plan to send it in and hope it gets posted. If you haven\'t yet, check out their website - I\'ve included the link in this post. It totally changes your perception of litter!  And that\'s it - plus of course, Jeff is back for Tejas, I\'m staying in Sausalito with Casey, its a 3-day weekend, and I\'m having lunch with the Chicas today. No more bad moods!   '),(10528,3667367,'17,June,2004','2004-06-17 00:00:00','   (Caution - this is a very self-conscious blog entry)  I feel like I\'ve been so busy with work, life, etc. that I\'ve hardly had a moment to post interesting entries to my blog. To all my loyal fans, I apologize - but can I ask one thing of you? If you could all leave a comment once in awhile, I\'d feel so loved. Maybe my posts have gotten a bit mundane in the past week and you\'re not inspired to comment, so I\'ll try and work on that.   '),(10529,3667367,'17,June,2004','2004-06-17 00:00:00','   Our lease is signed and we are all set to move in two weeks. If anyone needs some extra karma, we\'d welcome any help unloading the Uhaul on Saturday, August 28th.  Our new address is 239 Cervantes Blvd. #2, SF, CA 94123!   '),(10530,3667367,'17,June,2004','2004-06-17 00:00:00','   I feel like many people I know have been enjoying a \"brush with fame\" this week.  Like my BFF Enrique who is auditioning this weekend to be a backup dancer for Destiny\'s Child. However, not being a professional dancer (yet) he doesn\'t have a headshot. So, who came to his rescue? None other than his amazing BFF (that would be me) with ny new digital camera. So today, at work, we snuck into an unused office and had a photo shoot. I can now say I know how it feels to be a professional photographer! Luckily we managed to get a great, very professional-looking, shot and get it printed and blown-up to 8x10 size. So now if he gets the job, he owes all his success to me and must bring me as his date to the MTV Video Awards show. Watch for me in US Weekly people!  And then there\'s another person I know who went on a few dates this week with the son of a pretty well-known musician and even got to meet his dad and dad\'s other musician friends last night.  Who\'s next? I hope its me!   '),(10531,3667367,'17,June,2004','2004-06-17 00:00:00','   So tomorrow may just be the day when Jeff and I sign the lease on our new place. But then again it may not be as the situation at my old place is not yet settled. I\'m really just dreaming of this move being done and sitting in my new place by the water with Jeff. Too bad we can\'t have a dog to cuddle up at our feet, but maybe a picture of Casey will suffice.   In the moment (where my yoga teacher reminded me today I should be) ... I\'m enjoying my new minigreen ipod immensely and listening to music during my commute and at my desk.... I dusted off my LSAT books and am getting back into study mode after a brief hiatus...I\'m preparing for my cousin\'s wedding shower this weekend with fun games and prizes... I\'m thinking of KD tonight while she enjoys a Crosby, Stills, and Nash concert in NY with a cute boy... I\'m taking good care of Linz and Dwight\'s place while they\'re in Hawaii... I\'m loving taking pictures of Casey with my new camera... I\'m thinking of chocolate from fog city news...  Oh and, last night Jeff and I irreverently acted out a phone call between Scott Peterson and Amber Frey. It was the funniest thing we\'ve done in awhile, and I must say we did a damn good job.   '),(10532,3667367,'17,June,2004','2004-06-17 00:00:00','   I find myself today in a mood of total excitement. I feel like a million wonderful things are happening and I don\'t even know on which I should focus my attention. These are them:  1) Moving in with my most wonderful boyfriend... we may or may not have a place yet, but we\'re definitely going to have one very soon. I am so lucky to have such a sweet, funny, smart, caring and just all around awesome person to cohabitate with. I can\'t wait to make every little decision about our new place with him.  2) My new Mini Green ipod is so fabulous.  3) Dog sitting Casey this week and seeing Bailey this weekend in Santa Rosa. Dogs are wonderful creatures.  4) KD\'s new love interest in NYC who sounds amazing.  5) My new pink converse.  6) My upcoming acing of the LSAT and then my celebration trip to Hawaii.  7) ... and I\'m still on a high from the most amazing birthday ever!   '),(10533,3667367,'17,June,2004','2004-06-17 00:00:00','   Pyramid Tartare  at Blowfish (w/ salmon, tuna and avocado - served WT style with potato chips!), plus the  Peach Nympho  sake drink with nigori and peach juice was amazing!  A  Mitchell\'s ice cream birthday cake  (w/ Hawaii decorations), white cake and coffee ice cream, Jeff I LOVE YOU!!!   My other birthday cake made by the chicas with  chocolate cake  and frosting eaten at my house at 2:30 am on Friday night. I personally licked \"a piece\" of the frosting off myself.   Chilaquiles  with green sauce from Puerto Alegre. The best cure for a hangover.  Everything I had at Delfina on Sunday night including:  fried squash blossoms  stuffed with Ricotta,  raviolis with walnut sauce ,  albacore tuna  with green and cranberry beans, and  profiteroles  with espresso gelato and chocolate sauce, Yumm!!!   '),(10534,3667367,'17,June,2004','2004-06-17 00:00:00','   To all of my wonderful friends and family that made my brithday AMAZING - thank you. I felt incredibly special to have so many great people help my celebrate my 25th birthday. I had a blast at dinner and doing the mission bar crawl... and to everyone who couldn\'t be there, but called, e-mail, etc... thank you too! I\'m such a lucky girl.  BTW to all you bar crawlers... WE ROCK! We made it from start to finish... and all in time to hit the 500 Club by 2 am. And those that came back to 51 Linda afterwards, enjoyed the yummy Chica cake, and got to play spin the bottle!   '),(10535,3667367,'17,June,2004','2004-06-17 00:00:00','   Looks like we found a place right near Marina green and only a few blocks from Chestnut.  We might be signing a lease soon. Mauna Loa here we come!   '),(10536,3667367,'17,June,2004','2004-06-17 00:00:00','   I really do, and I feel bad about it. I have this horrible stereotype about people that live in the marina, despite knowing some nice people that live in that area. I tend to think that all Marina people are affluent, vapid, homogeneous, shallow, etc... and I can\'t seem to get over this idea that by living there I would somehow become one of those people are just be really unhappy that I was surrounded by THEM. I know its a horrible generalization, and I\'ve tried to see past it, but its pretty well ingrained in me.  Last night Jeff and I went to see a few apartments in the Marina area. Though obviously not my top choice of areas to live in, it would be an ideal area for Jeff to commute from, and convenient enough for me to get to work. Not to mention that the Marina is a beautiful area, with the water front within spitting distance. Despite my bias, I wanted to be fair to Jeff and consider it in our list of areas. The places we saw were mostly not that nice for the price. We did see one fairly decent place, but it is right on Van Ness, and not as nice as other places we have seen elsewhere for the same price or even less. I think we\'ve realized we\'re better off in another area. Pheww... and I didn\'t have to seem like a total bitch for just not wanting to live there.   '),(10537,3667367,'17,June,2004','2004-06-17 00:00:00','   Pardon my complete obnoxiousness, but people have been asking... not that anyone is required to buy me a gift, but if you feel the urge, here are 10 things I\'m wishing for:  1) Dirty Dancing Havana Nights on DVD (Don\'t everyone rush out at once to get this) 2) Le Creuset non-stick spatulas (I lost my other ones in my last move) 3) Dagoba chocolate bars!! 4) Lavender smelling things 5) Pictures of puppies, b/c I can\'t have one and I\'m that annoying person now 6) New agey things like crystals and incense (J/K) 7) Something from Good Vibes ;-) 8) Mani/Pedi gift certificates!! 9) A copy of your favorite book/CD 10) Queen-size, at least 300 thread count Egyptian Cotton sheets - for those of you on a budget!  However your PRESENCE at my birthday celebration on Friday night would be the best gift of all! ... first at Blowfish for dinner, then onto a Mission bar crawl. If you can\'t make it, gifts will be accepted!   '),(10538,3667367,'17,June,2004','2004-06-17 00:00:00','   Did the weekend really go so fast? I feel like I had hardly a moment to enjoy it, relax, study for LSAT, etc. Jeff and I managed to spend just about every waking moment either bouncing between open houses, or searching for new listings on Craigslist. And after probably 15 actual viewings, 10 or more drive-bys, and hours (6+) looking at hundreds of listings on craigslist, we have about 3 possibilities. We have only submitted an application for one of them and are keeping the other 2 in mind for back-ups IF we find nothing else, and they\'re still available. It is such a tiring process looking for a perfect place, and when we finally do, we have the whole headache of moving to deal with. Not to sound too down on the whole thing, I am so excited to live with Jeff in a new place (a new area probably), re-organize and downsize all my stuff, focus on LSAT studying, and then take off on our two week trip to Hawaii! I just hope once we do find this place, we stay there for awhile!   '),(10539,517266,'30,August,2003','2003-08-30 00:00:00','   Aloha and welcome.  Please feel free to check back often as I share my various moments of the day.   '),(10540,517266,'30,September,2003','2003-09-30 00:00:00','   Angola, Indiana - Speaking at Angola High School and Tri-State University.  Great audiences and great day.   '),(10541,517266,'29,September,2003','2003-09-29 00:00:00','   Indy - Taking Angelina and Anna to the doctor for the annual check up and hearing myself utter the words in true conversation, \"try not to poopy in the cup.\"  Also, taking them to Chuck E Cheese afterwards for funtime!   '),(10542,517266,'28,September,2003','2003-09-28 00:00:00','   Indy - Celebrating my mom\'s 60th birthday party with the entire family.  Had a wonderful day with her and also enjoyed visiting great aunt Mae in the hospital and taking her on a wheel chair ride.  Finally, the Colts victory over the Saints was a big one...first time Colts ahve had a 4 and 0 season in 26 years.  Go Colts!   '),(10543,517266,'27,September,2003','2003-09-27 00:00:00','   Spent the night at the cabin in Brown County with Angelina\'s class...hiked in the rain all through the creek...had a campfire...and slept in tents in a thunderstorm. Doesn\'t get much better than that!   '),(10544,517266,'24,September,2003','2003-09-24 00:00:00','   Home - Went to fly kites with Anna, Angelina and Buddy John at Broad Ripple Park.  We even got the kite up in the air.  It was a great afternoon!   '),(10545,517266,'23,September,2003','2003-09-23 00:00:00','   Having a morning picnic out by the river with special B.  It was wonderful to spend that time together.  Also, meeting up with Dese, Meg, Danny, T Bayt and Brandon at Starbucks, to reconnect!   '),(10546,517266,'22,September,2003','2003-09-22 00:00:00','   Las Vegas, Nevada - Went to see the show \"O\" for the third time in my life and once again found it astounding.  One of my top 10 life experiences!   '),(10547,517266,'20,September,2003','2003-09-20 00:00:00','   Las Vegas, Nevada - Spent the day with my mom and her 9 or so closest friends along with a few friends of moy own...celebrating my 36th birthday and my mom turning 60.  The MOTD probably had to be sitting down with a psychic for the first time in my life.  Real wild and real profound!   '),(10548,517266,'17,September,2003','2003-09-17 00:00:00','   Speaking at Estero High School in Florida near Fort Myers.  Making a new friend, Dr. Jones, who works in the parking lot of the school.  Wonderful audiences.  Full of energy and very supportive administration.  Spent the rest of the day with Micki (my mom) on jet skis in the ocean...until we tipped over.  Don\'t ask.  Found a great sushi and thai restaurant.  Great day.   '),(10549,517266,'31,October,2003','2003-10-31 00:00:00','   Indy - Happy Halloween!  Taking the girls trick or treating and going to the black tie event for the Damien Center at the Indiana Roof Ballroom.  Great night where they raised over $100,000!   '),(10550,517266,'30,October,2003','2003-10-30 00:00:00','   Wisconsin Dells - Spoke at a local leadership conference and had a blast.  Then spent the rest of the day with the family enjoying yet more water slides at Treasure Island!  This has been a wet and wonderful visit!   '),(10551,517266,'29,October,2003','2003-10-29 00:00:00','   Wisconsin Dells - Too many MOTD.  Being at the Kalahari Resort with Anna, Angelina and Brandon is a blast.  THis is the coolest water park in the world with amazing indoor water slides.  You have to see it to believe it.   '),(10552,517266,'27,October,2003','2003-10-27 00:00:00','   Indy - Spoke in Avon, Indiana and Selma, Indiana today and had a blast at both schools.  But, my moment of the day was seeing Anna and Angelina again and spending the evening with them!   '),(10553,517266,'26,October,2003','2003-10-26 00:00:00','   INDY - Colts WIN against the Houston Texans...great day with a group of friends from Paradigm...went to see Tim Conway and Harvey Corman in concert.  Very interesting!   '),(10554,517266,'25,October,2003','2003-10-25 00:00:00','   Indy - Had a wonderful evening with Brandon\'s parents in Cloverdale...followed by dinner with Vince and his new girlfriend Lori, T Bayt, Desi, the Dream, Danny and Stephanie!  Great night!   '),(10555,517266,'24,October,2003','2003-10-24 00:00:00','   Indy - Back home again in Indiana!  yeah!  Great day at the Bernardsvlle 8th Grade Youth Summitt and an even better night at Paradigm.  Great to reconnect with everyone again.   '),(10556,517266,'23,October,2003','2003-10-23 00:00:00','   New Jersey - Spoke at Radix and Holly Glenn Elementary Schools and had an absolute blast.  Great students and amazing adults who have a true passion for education and students...and it shows!   '),(10557,517266,'22,October,2003','2003-10-22 00:00:00','   Verona, NJ - Spoke at Verona High School and enjoyed my chat with Johnny, Tommy, Vanessa, Alex, Alyssa, Carine and Pam.  They were a fun group and it was a great school.   '),(10558,517266,'21,October,2003','2003-10-21 00:00:00','   Troy, MO - Spent the day in Troy speaking to the National Honor Society and a group of 5th graders.  Tonight, there were hundreds that showed up for an evening program.  Great day all around.  Great town.   '),(10559,517266,'19,October,2003','2003-10-19 00:00:00','   Monday - Indy - Brandon and I spent the day with Karl and Jeanne Anthony and had a great time just being.   '),(10560,517266,'19,October,2003','2003-10-19 00:00:00','   Indy - Spent the day with Karl and Jeanne Anthony and debuted the Choose To Love song at the Garden!  Yeah!   '),(10561,517266,'18,October,2003','2003-10-18 00:00:00','   Indy - Going to Anna\'s school carnival...carving pumpkins with Anna and Angelina...and spending time with my dear friends Karl and Jeanne Anthony as we sat by the fire with Jim, Meg and Brandon and sang, laughed and talked.  Great night!  Great life!   '),(10562,517266,'16,October,2003','2003-10-16 00:00:00','   WInchester VA - Spoke at Handley High School for their Homecoming and was able to be part of one of the best Homecoming celebrations I have ever experienced in a high school.  Amazing!   '),(10563,517266,'16,October,2003','2003-10-16 00:00:00','   Indy - Spent a great morning with Mikey Smith having breakfast at Patachou and catching up.  Made it safely to Winchester, VA and am excited about a great day tomorrow.   '),(10564,517266,'15,October,2003','2003-10-15 00:00:00','   Indy - Going to the pumpkin patch with Brandon, Mikey Smith, Anna and Angelina.  It was great and we found the perfect pumpkins.   '),(10565,517266,'13,October,2003','2003-10-13 00:00:00','   Casper, WY - Had a great day at the Healthy Communities Conference in Wyoming speaking to young people and adults.  My MOTD was spending coffee time with Ty Sells.   '),(10566,517266,'13,October,2003','2003-10-13 00:00:00','   Casper, WY - Spending the day with Ty Sells at a state leadership conference.   '),(10567,517266,'12,October,2003','2003-10-12 00:00:00','   Indy - Going to the Colts game and spending the day with an amazing group of people.  Okay, so we lost in overtime by three...superbowl, here we still come!   '),(10568,517266,'10,October,2003','2003-10-10 00:00:00','   Indy - Celebrating Brandon\'s and Brandi\'s 25th birthday in Cloverdale with their entire family.   '),(10569,517266,'10,October,2003','2003-10-10 00:00:00','   Hannibal, MO and Indy - Spoke at a teacher inservice today and had a ton of fun.  But my moment of the day was climbing trees and riding bikes with Anna and Angelina back at home.   '),(10570,517266,'09,October,2003','2003-10-09 00:00:00','   Hannibal, MO - I was in Java Jive coffee house today and watched a group of kids on a field trip come in wanting to buy Hot Chocolate.  There were about 8 boys.  Not everyone could afford the hot chocolate.  A man, unrelated to the group, stepped up to the counter and bought drinks for all the boys.  They were dumbfounded and thankful.  It was a great moment.   '),(10571,517266,'08,October,2003','2003-10-08 00:00:00','   Hannibal, MO - I am in the home town of Mark Twain and The Unsinkable Molly Brown who was on the Titanic.  I spoke today to the middle school, high school and then tonight to a group of parents.  Overall, a very good day.  The schools were amazing with great auditoriums - which makes all the difference.  Today\'s visit was sponsored by the C.H.A.R.T. Teen Task Force.  Great day.   '),(10572,517266,'07,October,2003','2003-10-07 00:00:00','   Indy and Hannibal, MO - Spent the day with Brandon celebrating his 25th birthday.  It was a great day and I am so happy to have been home to be with him all day!   '),(10573,517266,'06,October,2003','2003-10-06 00:00:00','   Indy - Watching the COLTS beat Tampa Bay, in overtime, on a Monday Night Game.  One of the greatest football games I have ever seen.  Superbowl, here we come!   '),(10574,517266,'05,October,2003','2003-10-05 00:00:00','   indy - Participated I a rope course today and was able to jump from the top of a thirty foot telephone pole in to the air to catch a flag.  It was insane and fun.   '),(10575,517266,'03,October,2003','2003-10-03 00:00:00','   Indianapolis - Rejoining my group of friends tonight from the the group I am in, Vision in Action.  We had a great night being together and laughing and watching the movie Defending Your Life.   '),(10576,517266,'02,October,2003','2003-10-02 00:00:00','   Missouri and Indianapolis - Speaking at an excellent STUDCO Conference in Missouri.  Then tonight we went to visit great Aunt Mae in her new home.  She is doing much beter after being inn the hospital.  Thank God!   '),(10577,517266,'01,October,2003','2003-10-01 00:00:00','   The new website is LIVE!  Yeah!!!   '),(10578,517266,'29,November,2003','2003-11-29 00:00:00','   Indy - Saw a great movie last night with friends, Shattered Glass.  VERY GOOD.  Afterwards, we all headed back to my house to start the Holiday Season off.  Tonight was great as well all went to Meg\'s for Turkey Day leftovers and to enjoy each others company.  Great night all around!   '),(10579,517266,'28,November,2003','2003-11-28 00:00:00','   Indy - Cooked a gret Thanksgiving Dinner for Grandma, Mom, Dad, Brandon, Katie, Bill, Anna and Angelina.  It was a lot of fun.  So much to be thankful about.  Last night we also saw the movie ELF which was hilarious.   '),(10580,517266,'25,November,2003','2003-11-25 00:00:00','   Indy - Picked up Grandma from the airport as she is i town for thanksgiving!  I am so thankful for my family!   '),(10581,517266,'23,November,2003','2003-11-23 00:00:00','   Indy - Spent the entire weekend with my friends, fixing up the house that belongs to the Indiana Youth Group.  It was a TON of work but was well worth it.  I am totally exhausted and ready to rest!!!   '),(10582,517266,'21,November,2003','2003-11-21 00:00:00','   Indianapolis - Spoke in New Jersey today at Thomas Jefferson Middle School where some of the students had shirts made with my face on them.  Hilarious.  Tonight, the paradigm group met up at the Indiana Youth Group to plan our community service weekend.  I can\'t wait!   '),(10583,517266,'20,November,2003','2003-11-20 00:00:00','   New Jersey - Spoke in Woodcliff and Westwood and had a blast.  In Woodcliff, I spoke to a middle school and in Westwood, I returned to the high school to work with student leaders.   '),(10584,517266,'18,November,2003','2003-11-18 00:00:00','   Columbus, Ohio - Spoke at most of the Catholic High Schools today and had a great time.  Each school was amazing in it\'s own way and I am so blessed to have had the chance to be part of sharing the ALOHA spirit.  What a great time.  Thanks Columbus!   '),(10585,517266,'17,November,2003','2003-11-17 00:00:00','   Indy - Celebrating Angelina\'s birthday at her school.  Happy 8th year pajamalina!   '),(10586,517266,'16,November,2003','2003-11-16 00:00:00','   Indy - Colts win over the Jets, 38-31.  Great game and great day with friends at the dome!   '),(10587,517266,'14,November,2003','2003-11-14 00:00:00','   Yesterday - Angelina\'s 8th birthday party was a blast.  The theme was Scooby Doo.  It was a wonderful way to begin her ninth year on this planet!   '),(10588,517266,'14,November,2003','2003-11-14 00:00:00','   Pennsylvania - Spoke at the state student council convention today and had a blast on their leader-ship.  Also, I was able to reconnect with an old friend, Mike Smith...a great speaker from California.  Wonderful day with amazing lei-ders.   '),(10589,517266,'12,November,2003','2003-11-12 00:00:00','   Chicago Area - Spoke at Gross Middle School and Hauser Junior High Schools.  Both were great schools with students, staff and auditoriums that were amazing.  I am so proud to have been able to share the ALOHA spirit with them!   '),(10590,517266,'10,November,2003','2003-11-10 00:00:00','   Last Night (Sunday) - Went to see Dame Edna at the Murat Temple with my mom, Brandon and my friend Jim Zarvos.  She was absolutely hilarious!   '),(10591,517266,'08,November,2003','2003-11-08 00:00:00','   Indy - Great day at home!  Getting stuff done, spending time with the girls, having taco night.  Enjoying the day!!!   '),(10592,517266,'08,November,2003','2003-11-08 00:00:00','   North Carolina - Today, I spoke at the state student council convention here in NC.  The entire conference theme is ALOHA and the student leaders here went overboard.  This is an amazing conference.   '),(10593,517266,'07,November,2003','2003-11-07 00:00:00','   Bloomington - Tonight, Brandon and I had dinner with my amazing Aunts Melanie and Susan.  They are so wonderful and I am so blessed to have them.  Great night!!   '),(10594,517266,'06,November,2003','2003-11-06 00:00:00','   Spoke at the Willow Tree Conference.  It is one of the sigle best middle shool preventin camps I have ever seen.  It is always a blessing to invited to be there!   '),(10595,517266,'05,November,2003','2003-11-05 00:00:00','   New Jersey - Spoke at Slocum Skews School and Cresskill High School...and had a blast at both.  Great schools, great audiences and amazing teachers.  What a great day.   '),(10596,517266,'03,November,2003','2003-11-03 00:00:00','   West Des Moines, Iowa - Spoke for the Dekko Foundation where I overheard a lady named Suzie who inspired me to take my bowl of light story to a whole new level.  The Dekko Foundation inspires young people to make a difference in the world.  They are amazing!   '),(10597,517266,'03,November,2003','2003-11-03 00:00:00','   Indy - Last night, I celebrated my mothers life as she finished her Level One of Paradigm.  It was great to spend the time with her and share a moment!   '),(10598,517266,'30,December,2003','2003-12-30 00:00:00','   Brown County - A couple of days in the middle of the woods on a great winter day!   '),(10599,517266,'28,December,2003','2003-12-28 00:00:00','   Indy - Watching the Colts beat the Texans to clinch the division!  Yeah!!!!   '),(10600,517266,'26,December,2003','2003-12-26 00:00:00','   Indy - Enjoyed a wonderful Christmas with the entire family.  Perfect.   '),(10601,517266,'23,December,2003','2003-12-23 00:00:00','   Indy - Spending a great day with Anna and Angelina by seein the Cat in the Hat movie.  Absolutely hilarious.  Had a wonderful Christmas night with Chris, John John, Pastor steve, Grandma, Great Grandma, Dreifus, Anna, Angelina, Bill, Katie, Corky and Brandon.  It was perfect.   '),(10602,517266,'21,December,2003','2003-12-21 00:00:00','   Indy - Yesterday, went to Chicago with Katie, Brandon and Mom to see hairspray.  it was hilarious and a to n of fun.  Tonight, we went to the Colts game where they sadly lost to the Broncos.  Great time, though.   '),(10603,517266,'19,December,2003','2003-12-19 00:00:00','   Indy - Went to see Anna\'s Christmas Program at school where she sung Frosty the Snowman.  It was, by far, the most interesting school program I have ever experienced.   '),(10604,517266,'17,December,2003','2003-12-17 00:00:00','   INDY - Presented last night at the gueset night for Paradigm Discovery (www.TheIndianaCenter.com).  It was a great turn out and a lot of fun to share my experience with others.   '),(10605,517266,'14,December,2003','2003-12-14 00:00:00','   Indy - Just returned from Las Vegas for the last few days where I spent time with Ty Sells, Karl Anthony, Ray Lozano, Vic Perez, Nick Webb and Dave Myers.  It was a ton of fun as we always enjoy our annual time together.  I would write more, but what happens in Vegas, stays in Vegas.  Yesterday, my MOTD was experiencing friend Dave Wilcox in concert and meeting his ten year old son, Nathan.  Today, the Colts were at home and confidently beat Atlanta.  SUperbowl, here we come!   '),(10606,517266,'11,December,2003','2003-12-11 00:00:00','   Las Vegas - Being here with my good friend Karl Anthony.  We are here for business and pleasure...and last night we were able to get tickets for the Billboard Music Awards.  It was a blast.   '),(10607,517266,'08,December,2003','2003-12-08 00:00:00','   Indy - Had an amazing weekend with a group of friends from paradigm Discovery.  It was a great time to share love and friendship with so many.  VIA X Lives!   '),(10608,517266,'05,December,2003','2003-12-05 00:00:00','   Indy - Spent a great day catching up with friends and with work.  TBAYT and I hung out in a coffee house and enjoyed each other.  Gret winter day in Indy.   '),(10609,517266,'02,December,2003','2003-12-02 00:00:00','   Indy - Went to the Beef and Boards Christmas Show tonight with Brandon and the girls and had a great time.  What a perfect way to start the holiday season!   '),(10610,517266,'01,December,2003','2003-12-01 00:00:00','   Indy - Yesterday - Being at the Colts game, and although we lost, it was great to be with such an amazing group of people.  I lost a bet to friend Ed Gerety since he picked the Patriots.  Another motd was at church last night as we enjoyed being with Carolyn and attending an open discussion with Kent Millard about the upcoming Methodist Conference next year.   '),(10611,517266,'30,January,2004','2004-01-30 00:00:00','   Houston - Great night at the NFL Comishiners Party held at the Astrodome.  A blast dancing and having fun with the group.   '),(10612,517266,'29,January,2004','2004-01-29 00:00:00','   Houston - Here for some football game on Sunday.  My MOTD was playing Guesstures with Meg, Danny, Casey, Vijay, Dese, Christine, Kalen, T Bayt, Vince and Brandon.  It was a blast.  Also, givinng bull rides in the livinng room was kinda fun, too.   '),(10613,517266,'29,January,2004','2004-01-29 00:00:00','   Indy - Last night, dressed up as Tigger and went to give a singing telegram with Brandon and Dese for our 45 year old friend for life, Danny.  He was at a really swank restaurant and it was really funny.   '),(10614,517266,'23,January,2004','2004-01-23 00:00:00','   Cape Haze, Florida - It has been some time since I last wrote, but I am here in Florida visiting my pledge brother from college, Matt White.  In January of 1999, Matt was diagnosed with ALS, also known as Lou Gehrig\'s disease.  Although it is a terminal disease, Matt has shown me how one can live through adversity and get the most life out of living.  After all, Matt chooses to live until the moment he dies.  And even then, he lives.  Spending time here is a blessing.   '),(10615,517266,'16,January,2004','2004-01-16 00:00:00','   Indy - Spoke in Southern INdiana today to two great middle schools.  It was a blast.   '),(10616,517266,'15,January,2004','2004-01-15 00:00:00','   Indy - Spent the day, yesterday, staffing an amazing conference here in Indy.  It is a great way to connect with others and give back.   '),(10617,517266,'11,January,2004','2004-01-11 00:00:00','   Indy - Colts Win in Kansas City!  New England, here we come!   '),(10618,517266,'10,January,2004','2004-01-10 00:00:00','   Indy - Introducing the girls to the puppy and celebrating John John\'s Birthday with Chris and Brandon at a show called TUNA TEXAS.  It was hilarious!   '),(10619,517266,'08,January,2004','2004-01-08 00:00:00','   West Palm Beach, Florida - Had a great time speaking at the County Student Council conference to all the high schools in Palm Beach County.  Enjoyed the ocean views, seafood and WARM weather!  Can\'t wait to get home to see Crazy Maesy the puppy!   '),(10620,517266,'07,January,2004','2004-01-07 00:00:00','   West Palm Beach, Florida - Saw my friend Chris George in the national tour of KISS ME KATE.  He did a great job and it was wonderful to experience him in the show and have dinner afterwards.   '),(10621,517266,'06,January,2004','2004-01-06 00:00:00','   Indy - Celebrated the life of Great Aunt Mae at her funeral and welcomed a new life into our world with a Jack Russell puppy named Maesy.  Life is beautiful!   '),(10622,517266,'04,January,2004','2004-01-04 00:00:00','   Indy - A great day celebrating the life of the girls father, Jose, who passed away a week ago.  We had a memorial service for him which was wonderful.  Anna and Angelina have taught me a lot about death and the celebration it should be.   '),(10623,517266,'01,January,2004','2004-01-01 00:00:00','   Indy - A great night of relaxation and friends with Christine, Dese, Danny, Meg, TBayt, and Brandon.  A great way to ring in the new year!   '),(10624,517266,'28,February,2004','2004-02-28 00:00:00','   Indy - Finding a brand new private super duper secret playground with Anna and Angelina.  Playing outside in February is never a bad thing!   '),(10625,517266,'26,February,2004','2004-02-26 00:00:00','   East Dreenwich, Rhode Island - Having lunch with my good friend Bob Houghtaling and then speaking at his amazing school Cole Junior High.  Great day.   '),(10626,517266,'25,February,2004','2004-02-25 00:00:00','   Indy - The annual pancake and jazz dinner at Church to ring in lent.  Dancing with Anna and Angelina was definitely the highlight.   '),(10627,517266,'22,February,2004','2004-02-22 00:00:00','   Indy - Having lunch with one of my favorite speakers, Melanie Mills.  It was great to reconnect with her in such a wonderful way.  It was a great time together.   '),(10628,517266,'21,February,2004','2004-02-21 00:00:00','   Chicago - Visiting my best and amazing friend, Judy McHargue, who is up from Texas.  We had a great night doing nothing but laughing and catching up.  It was a true gift!   '),(10629,517266,'19,February,2004','2004-02-19 00:00:00','   North Carolina - Spoke at three schools today.  South Rowan, East Rowan and Irwin Middle School.  I don\'t know where to begin.  ERHS and Irwin were in their auditorium and it was an amazing group.  Tons of energy and respect.  I really appreciated being there and felt appreciated.  At SRHS, they went way out of their way to decorate the gym for the day and to celebrate ALOHA in the halls and with the students.  Although sound systems in gyms can be nuts, the audience really endured and hung in there and I really am thankful.  There is something about North Carolina that I really like.  Thanks for the hospitality.   '),(10630,517266,'18,February,2004','2004-02-18 00:00:00','   North Carolina - Spoke at Alexander High School today and was totally impressed with the tradition and respect of the student body.  They really went the extra mile to make the assembly the best it could be.  I had a great time with a school of about 1,700.  I also enjoyed just driving through this state.  Absolutely beautiful.   '),(10631,517266,'16,February,2004','2004-02-16 00:00:00','   Bloomington, IN - Seeing the musical SEUSSICAL with my Aunt Sue, Aunt Mel and Brandon.  It was aweseome!   '),(10632,517266,'14,February,2004','2004-02-14 00:00:00','   Bloomington, IN - Seeing the musical SEUSSICAL with my Aunt Sue, Aunt Mel and Brandon.  It was aweseome!   '),(10633,517266,'14,February,2004','2004-02-14 00:00:00','   Indy - Had the annual Valentine\'s Day celebration with a great group of friends.  Amazing family.   '),(10634,517266,'13,February,2004','2004-02-13 00:00:00','   Indy - Happy Friday the 13th.  Wonderful day at home, trying not to freeze to death as the floors were being refinished.  Long story.  Great day.   '),(10635,517266,'11,February,2004','2004-02-11 00:00:00','   Indy - Enjoying a wonderful walk with my sister, Maisy and Mattie...until Mattie ran away for a while.   '),(10636,517266,'10,February,2004','2004-02-10 00:00:00','   Indy - Presented once again at a guest open house for Paradigm Discovery.  (www.TheIndianaCenter.com)  It was a great night and am thankful I had the opportunity to give back.   '),(10637,517266,'07,February,2004','2004-02-07 00:00:00','   Brown County - Spent a great weekend in the woods with FIVE DOGS, Gray and Brandon.  Snow everywhere.  A wonderful couple of days.   '),(10638,517266,'04,February,2004','2004-02-04 00:00:00','   Indy - Had a great day with the new canine addition to our family.  Went to Christine and Dese\'s to watch the Superbowl commercials on TIVO.  Hilarious.   '),(10639,517266,'02,February,2004','2004-02-02 00:00:00','   Houston and Indy - So many moments.  COming home from Houston after an amazing Superbowl experience.  Also, picking up Maisy\'s brother, Mattie, as a new addition to our family.  Four dogs.  Does that make us crazy?   '),(10640,517266,'01,February,2004','2004-02-01 00:00:00','   Houston - Fun night, last night, at dinner with the group before the big SUPER day today. Also, got to see my friend Rob who is here from New York for the game.   '),(10641,517266,'30,March,2004','2004-03-30 00:00:00','   Glendale, Wisconsin - I spoke in Nicolet High School today where I presented five times before noon.  It was a bit nuts, but a lot of fun.  Each grade had a ton of energy and working with the student leaders was rewarding.  Tonight, I spoke with the parents and enjoyed the evening.  Great time in Wisconsin.   '),(10642,517266,'29,March,2004','2004-03-29 00:00:00','   Indy - Wow...so many.  Seeing my friend Dave Brown present his new book:  The Self-Help Paradox...Going to celebrate with Dese, Christine and my Mom after paradigm...seeing my friend Spiffy after over a year...being with Dreifus.  Too many to count.  Great day.   '),(10643,517266,'28,March,2004','2004-03-28 00:00:00','   Indy - Having dinner with my wonderful Aunt Kim, Special B and my mom.  Kim signed up to be in Paradigm this May (www.theindianacenter.com).  I am soooo excited for her.  She will LOVE it!   '),(10644,517266,'26,March,2004','2004-03-26 00:00:00','   New Jersey - This should be more of my moment of the week.  I have been in Maryland and New Jersey all week long speaking at great schools and to wonderful parent groups.  Today I spoke at the New Jersey State REBEL2 Conference for Middle Schoolers and had an absolute blast.  I am exhausted and ready to head home!   '),(10645,517266,'22,March,2004','2004-03-22 00:00:00','   Maryland - Finding out that my website is not gone after all.  There was a weekend of panic as I think I did something to erase my site...but luckily, my pal Erik has everything backed up, so we should be up and running again within a day.  Woo Hoo.  I am in Maryland today speaking at South Carroll High School and then to a parent group tonight.  Then I am off to New Jersey for the rest of the week.   '),(10646,517266,'19,March,2004','2004-03-19 00:00:00','   Nevada, MO - Speaking in Nevada, Missouri to every kid from K through 12th grade.  The middle school has ALOHA as their theme for the entire school year, so that was, by far, my moment of the day.   '),(10647,517266,'18,March,2004','2004-03-18 00:00:00','   Indy - Had a great day at home with my MOTD with Dese, Special B, Mick and Katie...just being together and catching up.   '),(10648,517266,'16,March,2004','2004-03-16 00:00:00','   Pennsylvania - Spoke at Pocono Mountain West High School today and had a great time.  A beautiful school with a great group of students.  I appreciated their respect and attitudes.   '),(10649,517266,'13,March,2004','2004-03-13 00:00:00','   San Ramon, CA - Spoke at the area Youth to Youth conference for middle school leaders.  An absolutely great conference of about 350 students and with a staff of about 100.  This is their tenth anniversary of Youth to Youth.  Congratulations.   '),(10650,517266,'12,March,2004','2004-03-12 00:00:00','   San Ramon, CA - Spoke at a National Blue Ribbon School, Iron Horse Middle School.  It was one of the top schools I have ever experienced.  The principal, Kirby Hoy, was exceptional and his staff and students refelct that.  I am so proud to have had the chance to be part of such a great school ohana.   '),(10651,517266,'11,March,2004','2004-03-11 00:00:00','   Indy and San Fran - Made it home from Wisconsin, spent the night at home and then had a great session with Meg as she taught me some new stretches.  It was really effective.  Flew to Cally and am looking forward to the time out here.   '),(10652,517266,'09,March,2004','2004-03-09 00:00:00','   Wisconsin - Spoke at Columbus Middle School, Lodi High School and Pardeeville High School.  Each school was really great and respectful.  It was a great day.  I am now at the Kalahari Resort to speak in the area tomorrow.  They have the worlds largest water park here, so there is fun to be had.   '),(10653,517266,'07,March,2004','2004-03-07 00:00:00','   Indy - Celebrating my friend Megs 44th birthday with the family - Vince, T Bayt, Christine, Dese, Danny, Special B and ofcourse Meg.  It was a great afternoon.   '),(10654,517266,'06,March,2004','2004-03-06 00:00:00','   Indy - My two great friends Dese and Christine, had an engaging party tonight for Special B and myself.  It was an amazing night with a great collection of friends and family.   '),(10655,517266,'05,March,2004','2004-03-05 00:00:00','   Northern Indiana - Speaking at the Indiana Northern District Student Council Conference.  They had an ALOHA theme and the group was a blast.   '),(10656,517266,'05,March,2004','2004-03-05 00:00:00','   Indy - Going to the JCC for a great hour or so.   '),(10657,517266,'03,March,2004','2004-03-03 00:00:00','   Indy - Spoke at Immaculate Heart of Mary school today to both the K through 4th grades and then the 5th through 8th.  It was a great school and I was happy to be part of such a positive day.  Plus, since this school is right down the street from where I live, it is nice to speak in your own back yard.  Thanks IMH!   '),(10658,517266,'01,March,2004','2004-03-01 00:00:00','   Indy - Spending the day in Cloverdale having lunch with Brandon\'s mom, seeing his Grandma and meeting about flowers for the big day in August.  Great weather too for Indiana!   '),(10659,517266,'01,March,2004','2004-03-01 00:00:00','   Indy - Last night we hosted the Annual Oscar Bash.  It was great.  Each person brings a good item that best represents a movie.  For example, two years ago i won the event with a statue of Jesus surrounded by onion rings for the movie:  Lord of the Rings.  You get the idea.  This year, we had about 40 people with amazing ideas.  The winners of the night were:  OSCAR PRIZE - Sue and Mel  BABE PRIZE - Amy from Pirates of the Carribean TINKY WINKY PRIZE - Cheryl Tocco and JANET JACKSON PRIZE - Linda Nunn!  It was a great night all around.   '),(10660,517266,'30,April,2004','2004-04-30 00:00:00','   Denver, CO - Spoke at three area high schools and had a great time in each one.  The Cherry Creek School system is really great and I was also able to spend time with my good friend Coleen, so that was an added bonus!   '),(10661,517266,'27,April,2004','2004-04-27 00:00:00','   Columbus, Ohio - Visting my good friends Ty and Angie Sells and spending time with their triplets Keegan, Tanner and Carly.  They are angels.   '),(10662,517266,'21,April,2004','2004-04-21 00:00:00','   Columbus, GA - Day three in a great town, speaking to 1,200 sixth and seventh grade students a day at the annual \"Drug Free You and Me\" conference, sponsored by McDonalds.  My mom is here with me for the week, which is a lot of fun.  Plus, speaking at two elementary schools each day has been very rewarding.  However, the Tour du Georgia, a bike race with Lance Armstrong, came through town today, which was exciting to watch.   '),(10663,517266,'19,April,2004','2004-04-19 00:00:00','   Columbus, GA - Celebrating my thirteenth year at the annual \"Drug Free You and Me\" conference for all city sixth and seventh graders.  My mom came with me here this week which will be a lot of fun.  Great people and great cause.   '),(10664,517266,'17,April,2004','2004-04-17 00:00:00','   Indy - Running a 15K race with my sister, Katie.  She is doing amazing as she trains for the Indianapolis mini-marathon!  Go Katie!   '),(10665,517266,'15,April,2004','2004-04-15 00:00:00','   Southern Indiana - Spoke at four schools today, sponsored by Peers Educating Peers. They are a group of high school student leaders who work with younger students and encourage them to be abstinent.  What an amazing group of people.  I was so happy to be sponsored by them.  Great day!   '),(10666,517266,'14,April,2004','2004-04-14 00:00:00','   Jasper, Indiana - Wow.  So many moments in the past few days.  Going to Columbus, Ohio to see Josh Groban was amazing and speaking at Hilliard Darby High School was great.   '),(10667,517266,'09,April,2004','2004-04-09 00:00:00','   Indy - Spending the day at friends Melanie and Sue along with Dese, Christine, Anna and Angelina.  We dyed and created 400 Easter Baskets for underpriviledged children.  Then tonight, going to the Good Friday service, was simply undescribable.  What a blessing.   '),(10668,517266,'08,April,2004','2004-04-08 00:00:00','   Indy - Enjoying Spring Break with my lovely little angels.  Going to the park and the nature center, having picnics and of course, stripping wallpaper.  Mmmm, nothing says spring break like stripping wallpaper!   '),(10669,517266,'07,April,2004','2004-04-07 00:00:00','   Alabama - Spoke at Athens High School to all the county sophomores.  It was sponsored by a group named T.R.A.I.L. who were an amazing group of young people who are focused on giving back to the world.  My program was followed by a volunteer fair where students could get involved with a variety of organizations.  It was a great idea and a great day.   '),(10670,517266,'05,April,2004','2004-04-05 00:00:00','   Indy - Going to Chuck E Cheese with the girls, Judy Judy and Lucy Lou.  Perfect.   '),(10671,517266,'04,April,2004','2004-04-04 00:00:00','   Indy - Having my angelic friends Judy Judy and Lucy Love in town to visit from Texas.  Lucy is 13 months old and is a gift from God.  It is such a blesing to have a weekend with her and her amazing mother.   '),(10672,517266,'01,April,2004','2004-04-01 00:00:00','   Detroit, MI - Spoke at one of the most amazing high schools - University of Detroit High School.  A very smart, funny and quick audience that really enjoyed the message and made me feel quite at home.  Great day.   '),(10673,517266,'28,May,2004','2004-05-28 00:00:00','   Culver, Indiana - Sadly, it has been over a week since my last post, but I had a great week at home catching up on odds and ends.  Today, I am up at the lake with Kalen and Meg enjoying the serenity.  A great way to begin the summer.  Right now, we are watching CAT IN THE HAT.  Hilarious.   '),(10674,517266,'20,May,2004','2004-05-20 00:00:00','   Indy - I also found out that two of my dearest friends, Ed and Suzanne Gerety are pregnant!  I am so thankful that people like them are going to be parents as a gift to the rest of the world!  They will be great parents.   '),(10675,517266,'18,May,2004','2004-05-18 00:00:00','   Indy - Going dancing at the Vogue, here in Indianapolis, at a night called retro-rewind...where songs from the 80\'s are considered oldies.  Okay, I am officially old.  Danced with Queen B, Meg, Vince, Mark and countless others.  Hilarious night.   '),(10676,517266,'18,May,2004','2004-05-18 00:00:00','   Pennsylvania - How embarrassing.  Blame it on my adhd, but a week without writing is sad.  But, alas, I am back.  My motd would have to be here in Pennsylvania and speaking at Downington High School West.  It was to the juniors and seniors.  They have a brand, spankin new school with a great auditorium and it was an all around great experience.  I return tonight to speak to the student leader banquet.  Great day.   '),(10677,517266,'10,May,2004','2004-05-10 00:00:00','   Cleveland - Going to get Grandma and move her to her new home in INdianapolis.  It will be so wonderful to have her in town!   '),(10678,517266,'22,June,2004','2004-06-22 00:00:00','   Indianapolis - Having a great summer here in Indy.  Getting ready for the big wedding in August...went to Holiday World with my nieces in Southern Indiana and had a BLAST...swimming at the JCC...spending time with my Grandma who recently moved to town...and trying to get things done before the fall. So far, so good.   '),(10679,517266,'08,June,2004','2004-06-08 00:00:00','   Indy - Whoa!  Where does the time go?  Enjoying the beginning of summer...spending days in a cabin in southern Indiana...with great friends and four angelic canine beings.  Oh, and did I mention the 50,000,000 secadas?   '),(10680,517266,'31,July,2004','2004-07-31 00:00:00','   Brown County, Indiana - Spent two nights at the cabin, which was great.  We had a ceremony for Corky, my dog of 14 years who passed away two weeks ago.  We spread his ashes there, since that was his favorite place in the world.   '),(10681,517266,'27,July,2004','2004-07-27 00:00:00','   Gainseville, Florida - Spoke last night at the state 4-H Congress to about 500 youth and 50 adults.  It was an electrifying night.  The confernece theme was \"Riding the Waves of Success\" and they had an ALOHA luau dance after my program.  A ocean of fun.   '),(10682,517266,'26,July,2004','2004-07-26 00:00:00','   July 26, 2004 - Whoa.  It has been so long since I have last written here.  I have really gotten caught up in the fun of summer.  I will try to write a bit more before the school years gets rolling again.  However, I do want to write about my experience I had over the weekend.    In Noblesville, Indiana, there is a living museum called Conner Prarie Farm.  It is all about life in 1886 when Grover Cleveland was president and the Governor of Indiana was Governor Gray.  I had an opportunity to spend the entire weekend there, living as a charectar of 1886, with my two nieces Anna (a.k.a Christina Hancock) and Angelina (a.k.a. Mary Hancock) along with my sister Katie (a.k.a. Nancy Hancock.)  My name was Salathiel Hartnet.    It was an amazing time.  We worked on the farms, made our meals from scratch, fed the pigs, split wood, and enjoyed evening activities without the benefit of running water or electricity.  It was an experience of a lifetime that I am so happy to have shared with my beautiful angels.  I wish everyone could experience life in 1886.  Priceless.  (And yes, at night if you had to use the \'privy\', you instead had to use a \'chamber pot\' which is NOT a pot you ever want to eat out of...if you know what I am saying.)  Aloha.   '),(10683,517266,'01,July,2004','2004-07-01 00:00:00','   Indy - The time is really flying by.  Having a great summer and not taking the time to write each day since I am trying to avoid the computer as much as possible!  Have a great, relaxing, aloha summer!   '),(10684,3984458,'29,July,2004','2004-07-29 00:00:00','   Ok I am in Wanganui NZ the most random thing ever and explaining what a blog is to this guy with steelo wool hair called Ben Stegman. His brother is good at hockey, the captian infact. Just thought I should tell the world that.  '),(10685,3984458,'22,July,2004','2004-07-22 00:00:00','   urlLink http://www.deviantart.com/view/7981089/   urlLink http://www.deviantart.com/view/8077884/   urlLink http://www.deviantart.com/deviation/8990373/   urlLink http://www.deviantart.com/view/6830860/   urlLink http://www.studyit.org.nz/subjects/maths/calc3/5/   urlLink http://www.bartleby.com/61/roots/IE39.html/   urlLink http://www.ehow.com/list_1020.html   urlLink http://www.tmnw10686.pwp.blueyonder.co.uk/   urlLink http://loish.deviantart.com/   urlLink http://lordofdrknez5.deviantart.com/  '),(10686,3984458,'20,July,2004','2004-07-20 00:00:00','   The Islamic Jihad assholes are now threatening to execute 6 more truck drivers after releasing the Phillipino one. 3 of those are Indian. This is unacceptable. If they are going to execute our civilians, we have no choice but to wipe all their kind from the face of the earth.  '),(10687,3984458,'20,July,2004','2004-07-20 00:00:00','   When have mass extinctions occurred, and what kinds of ecological damage resulted?   &nbsp;  Life has been around since the Archean Eon (3.9-2.5byr) when the Earthâ€™s oceans and atmosphere formed. Since then, after life exploded in the Cambrian period, it has been battling the odds to survive. In every major era,     the Paleozoic   ,    the Mesozoic    and the Cenozoic, life has been threatened to die out, but has prevailed. There have been five mass extinctions palaeontologists have come to agree on, where large proportions of earthly life have died out. These are the:  &nbsp;  End Cretaceous Extinction (About 65 million years ago)  The most famous extinction, it destroyed 85% of all species, but was by far not the largest. Interest in this extinction stems from the cause. In 1980, Luis Alvarez controversially suggested a new theory that an asteroid impact could be responsible for the death of the dinosaurs, capturing the interest of the globe. This is the most probable explanation, although there are others. Dinosaurs were among the reptilian species largely affected, but by no means the only group. However, some groups seem to have avoided being scathed by this impact; birds, crocodiles and of course mammals all survived which lead to our own evolution.  &nbsp;  Late Triassic Extinction (About 208 million years ago)  One of the more overlooked extinctions, however one quarter of all life was wiped out, making it comparable to the End Cretaceous extinction. It is still unclear whether this event was two extinctions spaced 12-17myr apart, or a single event. Many species families were lost, and both Conodonts and Labryinthodonts were disappeared, making room for dinosaurs to take over. This was caused by global warming.  &nbsp;  End Permian Extinction (About 245 million years ago),  This was the most devastating extinction ever in the planets history, wiping out 96% of all marine species, while on land three quarters of all vertebrate families became extinct. In fact, for a time it caused fungal species to be the dominant form of terrestrial life. This also did away with the last of the surviving trilobite species. Fluctuations in sea level, differences in ocean salinity or volcanic activity may be the cause for this mass extinction, as always leading to climate change, in this case a global warming.  &nbsp;  Late Devonian Extinction (About 360 million years ago)  Major facts of this extinction still have not been agreed upon; however most information suggests this extinction/series of extinctions took place over about 3 million years. It caused huge damage to biodiversity and killed up to 70% of all species. Marine life suffered most, while freshwater species, land plants and early amphibians did not suffer as badly. The causes for this extinction are not apparent, but are probably due to global cooling as a disproportionate number of warm water species died.  &nbsp;  Late Ordovician Extinction (About 438 million years ago)  The Ordovician period was a relatively stable one, where substantial growth in biodiversity took place. However as the period drew to a close an extinction event occurred, causing some animal groups to loose more than half their species. It seemed to be caused by an ice age, lowering sea levels.   urlLink  Echinoderms   ,   urlLink  trilobites   ,   urlLink  nautioids    and other groups suffered heavily. However, overall the effects were not as great as other extinctions.  &nbsp;  Late Cambrian Extinction (About 500 million years ago)  This period was one of great evolutionary expansion, it appeared that life had taken hold and was in a great experimental phase, where many species were created leading to the main branches in design of animal and plant life from 570-500myr. Most of the animals at this time were soft-bodied creatures, leaving little fossilized records. Therefore the effects of this extinction are not as clear as they normally are. The causes or cause of this extinction are not well known, but there certainly were huge changes in sea level during this time. The trilobites suffered, previously being widespread and dominant. From here their species declined from successive extinctions.  &nbsp;  How do mass extinctions occur, and do their causes still pose a threat?  All the mass extinctions occurred from ever present natural reasons. The next one may be man made and unlikely to ever occur again after our extinction, e.g. a nuclear winter; however the catalysts for previous extinctions are still around. Although it is hard to say, due to incomplete fossil records and accuracy limits on dating, it appears there is a cycle to mass extinctions of 26-30 million years. This has tremendous implications, for it suggests a single cause for extinction. However, there is always going to be more than one cause for mass extinction. In fact there are far too many right now than for me to list. So the following are likely, major causes for extinction that have occurred or may occur in the future. N.B. God, divine, human, extraterrestrial and supernatural causes have been omitted because their categories are too open.  &nbsp;  Biological Collapse  Plants and animals all consume resources, converting them into un-recyclable waste. This can lead to a shortage of resources causing mass extinctions. A plague of parasites/animals may also manage to quickly kill off major links in the food web and collapse it or modify the atmosphere. If it werenâ€™t for plant life, the Earthâ€™s atmosphere would easily have ended up like that of Venus, so life has massive effects on the entire planet.  &nbsp;  Climate Change  Induced by solar events, such as sun spots, large prominences and other changes in solar weather, our weather can change. Cloud formation is dependant to some degree on levels of cosmic radiation received, if this varies, more or less clouds may form lowering or raising the temperature respectively. Any temperature changes or instabilities in ocean currents can cause polar ice caps to enlarge or melt, changing sea levels by several meters. Volcanism is usually a catalyst for this behaviour, and releases many poisonous gases and ash clouds that can cover the planet. Natural cycles of ice ages and heat are shown to occur, so there is no reason why they will not continue to do so. In fact, we are due for a magnetic field flip that occurs regularly for unknown reasons.  &nbsp;  Cosmic Catastrophes  Black holes are unlikely to sneak up on us, but it isnâ€™t impossible that one of our neighbouring super-massive stars collapse and suck us in. Cosmic radiation appears to be a factor that has caused a few extinctions. As we regularly pass through the Milky Wayâ€™s spiral arms in the suns orbit around the galactic centre (over many thousands of years), Earth is bombarded by high levels of cosmic radiation from increased numbers of nearby supernovae. This can change the ionosphere dramatically and the radiation is lethal to life, a possible cause of the Permian extinction. Asteroid impact. This is blamed for the demise of the dinosaurs, a six mile long asteroid smacked into the Earth with the force of 100 million tonnes of TNT, making a dust cloud that covered the Earth for about six months blocking sunlight. This isnâ€™t as bad as the collision that hit the early Earth before life came to be, when we collided with an object the size of mars, sending ejected material into high orbit that soon collected and formed the moon. Planetary alignments are a remote cause for concern; however in 2025 all the planets bar Earth will align on one side of the solar system with earth on the other. The resulting gravitational imbalance could alter Earths orbit to more elliptical creating huge weather pattern changes, change the Sunâ€™s short term behaviour, or even nudge some asteroids out of the asteroid belt towards us.  &nbsp;  Death of Sun  As it goes through its main sequence, the sun gets 10% brighter every 10 million years. This will make the surface of the planet uninhabitable long before it expands into a red giant possibly swallowing the Earth, and at least baking it dry and airless. At this point in time Saturnâ€™s moon Titan will become like that of early Earthâ€™s for a period of 1myr, possibly sustaining human life while the only chance for survival, interstellar ark ships, are constructed that can make the journey over thousands of years to other star systems.  &nbsp;  Death of Universe  Ultimately we will all come to an end. The universe will accelerate out from our reach leaving galaxies trillions of light years apart, clusters of dead star cores and cold planets slowly being consumed by central black holes. Any surviving matter will one day sublime into energy as fundamental particles loose their power. The largest of the black holes will finally evaporate leaving an empty universe. Only intelligent life will be able to avoid this fate, but it requires millennia of technological development to create technology powerful enough to save and create life perpetually. Itâ€™s unlikely our species will stick around that long and do it.  &nbsp;  If any of these causes for mass extinction were to happen, would we be able to fend them off or escape?  The ability to fend off disaster is quite rare. However, it is possible for us to try and cool the atmosphere in the event of global warming by changing the air chemistry, reflecting sunlight with huge solar sails and deflect/destroy any incoming asteroids. Unfortunately due to current financial and political situations, in practice such attempts would not work at present. The nuclear warheads would take months to prepare and launch, and we have a record of finding high risk asteroids after theyâ€™ve passed close to the Earth. Early warning detection and defence systems would cost billions of dollars, and were more likely do die of climate change anyway, like increased volcanism. Manufacturing volumes of gasses and huge high-tech reflectors would cost much more, and deployment is more likely to become a tangled disaster than a micro thin reflective success. Practically, it is extremely hard to prevent the causes of extinction, if not impossible. Try altering the sunâ€™s orbit or stop a supernova for instance. Escape is the best and frequently sole option. Escaping our entire solar system has already been achieved by numerous probes like Pioneer 10 &amp; 11 and Voyager 1 &amp; 2. Developing the plumbing to keep messy humans alive for long durations seems to be our only obstacle to following them.  &nbsp;  Should we fear the causes for mass extinction? By when will we have the technology to become invulnerable to mass extinction, if ever?  According to the cycle theory mentioned in Q2, we are in the middle of the cycle, about 10 million years out. This doesnâ€™t mean we are safe, it means it would be statistically surprising if we suddenly died out without having a fair share of time to live. And surprises happen all the time. It may occur to you that the Earthâ€™s biosphere has never been completely erased; our extinction may make way for the next kind of life. However, it is prudent to note that during the three to four billion year evolution of life, only once has it produced intelligence. This intelligence is a pretty fair shot, not just because it is naturally capable of constructing the environment, space travel, etc. but can engineer its own DNA intuitively which is the next evolutionary step. Direction. It is unlikely to ever happen again. Itâ€™s safe to say that it wonâ€™t happen again, we appear to have arrived just in the nick of time, as the Earth will loose its capacity for biological regeneration in less than a 100myr thanks to our warming sun. So the engineering ability is important. It will be the final step we take before our demise as humanity is superseded by what we create. Obviously weâ€™ll borrow from the finest evolution in nature to perfect this new machine, which will still be representative of ourselves. But it will be far more efficient, hard and adaptable, opening up the majority of planets for habitation instead of the incredibly few Earth-like ones we may never be able to reach or find. They will take their like wise engineered flora and fauna extending the biosphere of Earth because they can do what we fall short of, evolution will not take us outside the box because weâ€™re adapted to live and inflexible. And they will point up at the stars speaking of the yellow sun and the blue planet, and say it all started there. That is the point when we will have the technology to become as invulnerable as possible to mass extinction, spread right across our galaxy. The technology for this will be perfected within the next five hundred years. Just a bit more know how about genetic engineering, and the materials and habitat sustaining technology to build a ship with an ecosystem that can support generations of people for a few thousand years without breaking down. As for cheating the death of the universe, the technology it currently a theoretical impossibility. Fearing the causes of mass extinction and building defences against it will probably only save us from mass loss of life, as total extinction is very unlikely. Especially since the resources to send a colony to live on Mars is already in place bar about 5 billion US dollars, making us invulnerable to total Earthly catastrophe. Even if an asteroid hit, a few of us will survive, whether it is by keeping a fire alight in the darkness and eating dead animals, or in a nuclear bunker for 6 months. In short the answer is not to fear, but work as quickly as possible towards creating the pinnacle of evolution and finding new homes.  '),(10688,3605346,'15,June,2004','2004-06-15 00:00:00','    Three days ago, i received a text message from my father and it went like this, \"Gud pm, pinatawag kami ng parents ni richie, wag daw ang hindi kame pupunta sorry hindi ka na namin nahintay.\"  I missed his three calls and two of my brother\'s.  I then called them up.  My dearest kuya is getting married.  Mixed emotions.  I wasn\'t shocked at all coz i saw it coming na.  But when i asked inay on the date, she told me, this coming month. Sad ako, excited, and surprised.    Finally, my kuya\'s getting married after years of unending search for the right one.  I dunno what to feel.  Affected ba ako? \"Baket kelangan ako maapektuhan eh ndi naman ako yung ikakasal?\"  That\'s what i was thinking.  Excited pa nga ako kase, magkakaron na ko ng pamangkin na gagawin kong spoiled. Together with my nasnip, i was suddenly thinking on what my pamangkin-to-be would be calling me.  Excited na kong magkaron ng pamangkin.  Excited din ako kase magpapasukat na naman ng gown para sa kasal.  Of course, dapat ako ang maid of honor.  But, yesterday on my way home.  Because of the traffic, i had time to think, kung ano ano naisip ko, nang biglang pumasok sa isip ko yung kasal ni kuya ko.  I sent him a text message asking him san kame magpapasukat ng damit.  Naging textmates kame for the rest of my trip.  I couldn\'t believe that in the middle of our conversation, my tears fell.  Syempre, hiya naman ako sa katabi ko, pinigil ko, though masakit sa throat.  Ayoko pa pala mag-asawa si kuya.  Gusto ko pala nandun pa rin sya sa bahay everytime i would go home on weekends.  Pag maggrocery, pag magpapasama ako sa mall, pag magpapasundo ako sa bus stop every friday.  Yung gigisingin ko sya pag dumadating ako. Pag saturday na ayaw kong bumangon, dadaganan nya ko tapos kukulitin yung kamay ko, which i fondly called \"kulinting\" tapos pag hindi pa rin ako bumangon hihilahin nya yung kumot ko para finally bumangon na ko.  These simple things are only few of the things i would definitely miss about him.  Bigla ko ding na-miss yung lolo ko na mahal na mahal namin.  Na sana, he\'s still alive to this day para makita nya ikasal ang paborito nyang apo.  I\'m sure, he would be so proud of my kuya.  Sayang, wala na sya.  Then, i passed by bf na.  Wala nang traffic.  In 10 minutes, i reached home.  I went straight to the room where cous and nasnip waited for me.  I burst into tears with matching words like these, \"ayoko mag asawa si kuya.  Gusto ko lang naman magkaron ng pamangkin.\" And then, cous made a remark, \"Dapat kase sinabi mo na, kuya, pwede naman akong magkaron ng pamangkin ng hindi ka nagpapakasal.\"  Then the tears stopped.    Ikakasal na talaga sya.  Wala nang makakapigil.  I just have to accept that we would eventually part ways talaga.  Alangan namang kami kami na lang buong buhay namin.  Today, on my way to the office, I was thinking of what gift i will give them on their wedding day.  I wanna give them something na hindi nya ko makakalimutan.  Accepted ko na, but still, i\'m sad and crying.  Sana dumating na yung baby who would call me, tita pretty. '),(10689,3605346,'15,June,2004','2004-06-15 00:00:00','    first blog ko to. hmmm.. what will i say kaya? sarap ng feeling..baket??e kase, meron na ko masasabihan ng thoughts ko na hindi magrereklamo sa mga pinagsasabi ko. haha!! '),(10690,3605346,'08,July,2004','2004-07-08 00:00:00','    had a good sleep last night.  sobrang sleepy ako that after business nightly i didn\'t have the strength to wait for insider para sa news about kris and mark..hahaha!! i woke up 5 minutes late from my scheduled waking up time...hindi na naman ako nagmadali..nagpapretty na naman ako..kase im hoping na happy day ko to with mahsweetie...  feeling happy..tnext ko sya..ganda ng gising ko eh...pagkadating ko ng office, i called him up...and asked kung magkikita kame...sad...hindi pala kase lalabas sya with his barkada...hurt na naman ako...konting time na nga lang yung saken, hindi pa mabigay..well, i am not that important talaga for him...he would rather have time with his friends than me.. =(  so, i expect a whole day of misery na naman and deprive tears.. '),(10691,3605346,'07,July,2004','2004-07-07 00:00:00','    Active and dynamic. Decisive and haste but tends to regret. Attractive and affectionate to oneself. Strong mentality. Loves attention. (True..KSP ako..) Diplomatic. Consoling (True..great listener ako!!), friendly(True..) and solves people\'s problems (Nyeh..Hindi ako si Eddie Gil...=p). Brave and fearless (False...Im afraid to be alone). Adventurous (True..sa dame ko nang adventures). Loving and caring (Sobra). Suave and generous. Emotional (Sobra din). Aggressive (Not Really). Hasty. Good memory (Dates and Numbers and vivid colors in my dreams). Moving Motivates oneself and others. Sickness usually of the head and chest (Head lang). Sexy in a way that only their lover can see (For private use only ako eh..hahahah!). '),(10692,3605346,'02,July,2004','2004-07-02 00:00:00','    he was sick...and he sounded so sick over the phone. everytime i would call him up in the office, his officemates have ready comments..and those comments make me laugh.. i didnt want to meet him na kase i can feel na he was so sick. i told him, puntahan ko na lang sya sa modesto and i will bring him dinner.  pero ayaw nya.  sabi nya, he doesn\'t want to be pampered.  then, i decided to cancel our dinner.  after that, i received a text message telling me, \"sweetie bat ganun..ok lang naman saken yun. kaya ko naman.  kaya lang bakit ayaw mo..sweetie naman..\" after reading those lines, my heart melted na naman.  so, i replied, sige sweetie, punta ka na lang dito.  while waiting for him, i asked him, sweetie kamusta pakiramdam mo? then he told me, masama pa rin yung pakiramdam nya, kainis diba? hindi ko naman sya pinilit..sya pa nagdrama..and then, he arrived. he looked so sick..and i couldn\'t help hugging him and touching his face.  naawa naman ako sa kanya.  so..we had dinner. we talked and i felt so happy na naman ulit.  tapos uwian na..sad naman kase i hadnt had the chance to hug and kiss him..even after he broought me to the shuttle..the moment i sat down the van, i received a text message from him saying he\'s sorry for the little time he had to spend with me..i didnt reply..tampo ako..upon reaching home, i asked him san na sya..i was so cold, yet i felt na he was ignoring my feelings na naman..  i then called him, and he told me, why i had tantrums again..he asked me kung hindi pa daw ba ako thankful kase magkasama kame..pero kulang...  i just wished,,saken na lang sya..sana ako na lang yung totoo...sana hindi na lang kame kelangan na ganito..well...maybe there\'s a reason why..kase siguro kung saken lang sya talaga, walang thrill and i think magsasawa ako sa kanya...  i have to understand him, i have to hold on to the situation kase mahal na mahal ko sya..and he told me, he loves me too...sad talaga.. '),(10693,3605346,'01,July,2004','2004-07-01 00:00:00','    \"hi sweetie\"  the first line i read in his IM window in YM.  thanks he\'s here in manila already.  after 7 days of not speaking to him, finally, meron na ulit chance na makausap at makasama sya.  the next line was...\"nagkasakit ako.  hanggang ngayon may lagnat pa nga ako eh.\" =( sad na naman ako.  i asked him kung ayaw nya na ko makita..sabi nya..gustong gusto kaso he can\'t carry himself to me.    this happens most of the time..hirap ng gantong sitwasyon pero i can\'t do anything about this but to understand him..to understand our situation..  we first met 4 years ago...through a missent text message...then after 3 years..nagkita kame...the first time i saw him, i knew we would have a great connection...nasundan yung first time na yun...then finally we knew we were inlove...kaya lang hindi pede...he was committed to someone else...but still...he likes me and i was head over heels inlove with him...  one day, he called me up...sounded worried..i asked him what\'s bothering him..he told me...SHE might be pregnant.  my whole world collapsed..i was crying for like 3 days..and then i called him up...do you plan to marry her? he told me, no plans yet..somehow, i still hoped..after a month he called me and told me, wla na kong magawa, kelangan ko sya pakasalan, i was crying like hell nasa office pa ko that time..i asked him, do you want me forget you and never hear from me again? he was crying and said, no, i want you to be always here for me pa ren...pero, half of me, said, let him go..and then, started the end of it (i thought!)..  i started going out again..i thought i am over him...tapos, isang hapon, tinawag ako nung officemate ko, dulz fonecall..**** daw...huy musta ka na, his first line..then we met again...after 5 months, we finally met again...ang ganda ko daw ngayon...and it happened once again..regularly, meet na naman kame...  ndi ko lam baket ganon ang effect nya saken..kahit anong lakad ko, nakacancel ko just for him...all i know is love ko talaga sya..never pang may dumating na boylet saken na hindi ko sinangtabi para sa kanya..too bad, we weren\'t made for each other..i hope someday or in the next life, ma-meet ko pa rin sya...and sana, kame naman yung made for each other.. '),(10694,3343294,'24,May,2004','2004-05-24 00:00:00','   I had my two hour sailing class on Friday and my four proteges could have used instruction in more than just water based activities. They didn\'t really want to listen to my sailing instruction and I had trouble organizing it in a cohesive way.  Watching these people mingle made me realize I need to either settle or give up by my late 30s, dating can not continue forever.  '),(10695,3343294,'21,May,2004','2004-05-21 00:00:00','   This afternoon I am Captain of the Single\'s Cruise at a local marina. I have been asked to sail around some local singles while they mingle. I get paid for it. It\'s about a five person boat, so I am not sure what the deal is. How many can they possibly fit?  Unless they are kinky and looking for a three or four-some, it doesn\'t seem like a practical way to meet people. Not that I am anyone to judge. I just hope I don\'t get hit on. I will threaten to keel haul them if even one suggests he swab my poop deck. There is a guy who works on the deck down there who is perversely hot. He has skin that looks like leather, and when he once helped me out of the boat his hand felt like a monkey\'s, but he has this incredibly tanned thing going on. I don\'t want to get the first sexually transmitted melanoma, but as Captain of the single\'s cruise, I should have certain rights.  '),(10696,3343294,'20,May,2004','2004-05-20 00:00:00','   So now I\'ve decided to have this leg lengthening surgery and I wish I could have scheduled it for tomorrow. I\'m absolutely incapable of starting anything that I\'ll be unable to complete before this planned 6 week hiatus from life.  I was in therapy yesterday discussing my recuperation time. My therapist seems to think its a bad idea for me to spend this period at my parent\'s house. I\'m not sure how else I can afford the procedure, as insurance covers only 90% of my physical therapy, which will have to be daily, and Short term disability (STD) insurance, covers 60% of my income.  Couldn\'t they have come up with a better name for it? It sounds like I\'m stuck at home with a particularly bad case of the clap. I can\'t figure out how to upload pics to this site. Maybe I\'ll figure that out later. I don\'t have anything relevant at the time anyway. Once the surgery has happened I should have some gory stuff. I\'m really bored at work today. I did some internet searches and read another fascinating blog on blogspot about the newest Capitol Hill sex story - Washingtonienne. It\'s full of such wisdom as \"A man who tries to fuck you in the ass when you are sober does not love you.\" Words to live by.  The new trend seems to be blogging about fucking. Blogging is really one of the most self absorbed thing one could embark on - assuming someone else wants to actually the minutae of our life. It is only natural that in an effort to gain attention and readers, fuck talk is becoming popular. I doubt there will be much of that for me this summer. Typically, every summer I fuck someone new. It\'s kind of a tradition. I don\'t know how many positions are going to be possible with a broken leg though.  '),(10697,3343294,'14,May,2004','2004-05-14 00:00:00','   I have decided after 15 years to revisit this lousiest of childhood experiences, and have two even legs. I worry that I am looking for too much out of it.  I am afraid when I use the term \"whole\" accidentally to describe my post operative state. Surgery made me such a mess when I was little, maybe I would have been anyway, but it couldn\'t have helped. Now I want it fixed and I don\'t want to be thinking somewhere that it\'s going to fix ME. I like to make fun of the white trashy people on that tv show on Fox \"The Swan\" where they\'re all fucked up and compete to be the most pathetic who deserves a makeover. Like Queen for a Day but with a tit lift.  '),(10698,3343294,'14,May,2004','2004-05-14 00:00:00','   I don\'t think I spelled inaugeral right but I am too lazy to look it up. It just doesn\'t look right.  Anyway, I am creating this blog in the hopes that anyone else who might be considering a similar procedure can learn from my experience. I have not found any blogs by folks undergoing similar experiences, and I hope to inform others, as well as review and organize my own thoughts and feelings in regard to leg lengthening and my experiences with it. This is going to be a little bit technical and a little bit personal.  I\'ll start out by setting the stage with a little bit of background. I have \"Hemihypertrophy\". What is this? Google says: Hemihypertrophy refers to the overgrowth (hypertrophy) of one side of the body. A person with hemihypertrophy will be found to have enlargement of the body on the affected side, so that the chest may be more full on the affected side as well as the arm and leg being longer and thicker in diameter.  Hemihypertrophy in infants and young children is important because such children have a much higher incidence of certain abdominal cancers, notably Wilm\'s tumor of the kidney and hepatoblastoma (liver cancer). Unequal leg length cause some walking difficulties as well.  Children with hemihypertrophy require careful periodic followup examinations to detect any early sign of malignancy. This followup must continue for a number of years.  Hemihypertrophy can be a familial trait or may occur sporadically.  In my case, it occured as a sporadic birth defect. No one knows why it happened. Sometimes, these things just do. I don\'t have any of the internal complications. The difference is not visible, you can\'t tell by looking at me that, say one of my eyes, or my breasts is bigger. My mom claims that she can tell that one of my eyes is bigger, and as for by breasts, I\'ve never had any comments!  It was discovered that my right leg was longer when I was very young. I begin seeing an orthopaedic surgeon in nearby Boston, at the Floating Hospital, a children\'s hospital which began, curiously, on a boat. It\'s no longer aquatic and is now close to Chinatown.  They determined that the best strategy was to arrest the growth of my right leg. They did so by removing my growth plate with two operations at age 9 and 10. I was in the hospital for several days. I am a bit foggy on my memories of this, as I was young, scared, and not emotionally prepared for what I was undergoing. I don\'t remember the procedure, and I don\'t know if this is because I was never told exactly what was happening or because I have blocked it out.  My Dr., Dr. Michael Goldberg, did not have a manner which was appropriate for a pediatric surgeon. In my opinion, he should have stuck to the adults.  After the surgeries, I continued to grow taller than anticipated. I am still the shortest one in my immediate family at nearly 5\'9\".  My right leg stopped growing, but my left continued and surpassed it by 2.7 cm. This is approximately the discrepancy I had to begin with, only reversed.  Dr. Goldberg recommended I live with the difference and suggested I begin seeing another Dr. I think he did not want to be reminded of his error. There wasn\'t anything he could do for me at that point anyway. He was such a prick.  My new Dr. is doing this procedure: www.iskd.com  - it looks pretty gnarly but I\'ve been promised a morphine drip. Hopefully there\'s a limit on that bad boy or I\'ll OD my first day in.  '),(10699,3360363,'20,may,2004','2004-05-20 00:00:00','   Each time... I was truly alone... the book commenced to write itself, screaming the things which I never breathed, the thoughts I never uttered, the conversations I never held, the hopes, the dreams, the delusions I never admitted.  '),(10700,3360363,'19,may,2004','2004-05-19 00:00:00','   Ik zit in de trein naar Enkhuizen. Iedereen behalve ik blijft op het perron achter. Net alsof deze trein een trein naar nergens is,  een trein die nooit aankomt,  en iedereen weet dat, behalve ik.  '),(10701,3360363,'18,may,2004','2004-05-18 00:00:00','   April 3, 2004   Lord Tennyson was a driving soul.  ~ N.C. (secret hero of these poems)  Strange how one such line can make it all okay, even if you\'re biking out in the rain at 3:40am and nothing seems to be working out the way you hoped it would... and none of that matters because you just spent 4 hours drinking wine and whiskey with your best friend discovering new/old wondrous pieces of music by Thelonious Monk and listening to Cassady raps.  And your roommate\'s good-luck-hug means so much more than the job you wanted so but didn\'t get-- (and it\'s okay to sometimes tell people that you love them) -- makes you think the world\'s not such a bad place, sometimes :)  '),(10702,3360363,'17,may,2004','2004-05-17 00:00:00','   I was halfway across America, at the dividing line between the East of my youth and the West of my future, and maybe that\'s why it happened right there and then, that strange red afternoon.  '),(10703,3360363,'16,may,2004','2004-05-16 00:00:00','   De plukjes mos en gras tussen scheefliggende grijze tegels,  het is zomer natuurlijk.  De kleine hoopjes zand waaruit mieren komen lopen.  De versteende dennenappel bedolven onder dertig centimeter zandbak. Wormen in een Hello Kitty bakje.  '),(10704,3360363,'16,may,2004','2004-05-16 00:00:00','   One time (it was about a year ago) I was walking down a San Francisco street after work going to get some shampoo and salted peanuts at Walgreens on Haight and Fillmore, when squinting against the low evening sun I could make out a funny-looking girl further down. She was wearing one of those India-style hippie dresses in bright purple, a pair of truly godawful sandals, plastic gold crown poised on her head, princess-like. And she was dancing around on the sidewalk, smiling and kissing some goodlooking guy with too much facial hair. Something about the scene irked me and I think I forgot the shampoo. On the way back the air smelled of spring and sea, tinged with a hint of melancholy (or, in retrospect, nostalgia) and I realized I\'d seen myself, five years before.  '),(10705,3360363,'16,may,2004','2004-05-16 00:00:00','   But then they danced down the streets like dingledodies, and I shambled after as I\'ve been doing all my life after people who interest me, because the only people for me are the mad ones, the ones who are mad to live, mad to talk, mad to be saved, desirous of everything at the same time, the ones who never yawn or say a commonplace thing, but burn, burn, burn like fabulous yellow roman candles exploding like spiders across the stars and in the middle you see the blue centrelight pop and everybody goes \'Awww!\'  '),(10706,3360363,'16,may,2004','2004-05-16 00:00:00','   Terug naar het hotel. Zonnebril. Na maanden in het donker is Helsinki te licht, te schoon, te rustig. Ik heb een klein beetje meer duisternis nodig, nog heel even, mijn ogen tranen, ik heb hoofdpijn. Langzaam zal ik aan het licht wennen, maar nu nog even niet.  In een winkelruit kijk ik stiekem naar mijn buik. Ik zie veel buik. Het valt zo op in de stippeltjesjurk. Houd mijn adem in en blijf even staan. Bijt op mijn lip.Trek een stom gezicht. Ik leg mijn hand op mijn buik en maak een holle rug. Nog lelijker en nog steeds veel buik. Doorlopen. Voordat iemand me ziet. Ik neurie. Altijd. Alweer. Spring van tegel naar tegel zonder de randjes te raken. Blijf even staan op een kruispunt. Er zijn geen auto\'s, het is stil en leeg. De weg is breed en rijkt tot zover ik kan zien. In de verte rijdt een bus. Dichtbij een lelijke flat (Mijn Finse familie zouden het een flat in Russische Stijl noemen.), een pompstation, stoplichten en een blinde muur met daarop een enorme schildering. Over de weg hangt een wirwar van draden, de straatverlichting. Het is zo mooi hier, ook al is er niets wat specifiek mooi is. Het is het moment wat de omgeving mooi maakt. Doorlopen. Het flatgebouw tegemoet. Achter elk raam andere gordijnen, vooral veel kant. Nauwelijks zichtbaar omdat de ramen van een donker soort weerspiegelend glas zijn. Een paar ramen staan open.  Links boven ontdek ik hem, waarschijnlijk is hij daar al de hele tijd. De Verrekijker. Ik voelde me betrapt, maar hij waarschijnlijk ook.  '),(10707,3360363,'28,may,2004','2004-05-28 00:00:00','   Once was a war won (lost) here  now our laughter.  Feet pale on dust - concrete bare black cinders left: bonfire.  We whisper sighed words and don\'t tread on the rose mallows.  You taste like ashes on a silver tray. (I like it)  '),(10708,3360363,'27,may,2004','2004-05-27 00:00:00','   Yesterday I went to visit the maritime museum in Lelystad, Holland. The main reason for my venturing into the polder was the museum\'s replica of a Dutch Indiaman from 1628, the Batavia, which sunk near the Australian coast on its maiden voyage. There are several replicas of ships from the Age of Sail around the world, but the great thing about the Batavia is that she was built exactly the way she had been originally, using only those techniques that were available to ship builders in the 17th century. It seemed to me that this ship was as authentic as they come, and I had to see her.  The Batavia lies in what used to be a sea, the Zuiderzee or Southern Sea, but what is nowadays a large lake, created by the placement of a huge dam at its entrance to the North Sea about half a century ago. Where there used to be huge waves on stormy days, you now find a calm water, mirror-like in its placidity, with only the merest of ripples disturbing its surface in the breeze. The ship was all I had hoped it would be. She was beautiful and looked exactly like her ancestor had almost four centuries before, but for the small electric lamps placed strategically in dark spots such as the lower decks and some of the cabins. Several carpenters were working on her, hammering away, which made me picture how she would have looked in her day, being refitted in port after heavy weather.   The only thing that annoyed me on my visit were the throngs of tourists all about the vessel, chattering away excitedly, who were following around several tour guides telling them about life aboard the Batavia back in the days of the East India Trading Company. On my way to the back of the ship I ran into a guide who seemed to particularly delight in describing the harsh punishment methods of the time to a crowd of visitors that appeared to eat it all up. Shrugging, I crossed the quarterdeck and went down some steps in search of a place that would allow for some more or less quiet reflection. I found it all the way in the back: a tiny, windowless cabin, low enough to bump your head against the ceiling even if you were crouching, with a small cupboardbed on either side and a bench in between. I stepped inside and flopped down on the bench, dropping my shoulder bag at my feet, and spotted the door in front of me. Trying it, I found it could be moved, closed even. As the door fell shut, the tourist noise became a mere background annoyance, and I enjoyed the solitude of the little room immensely.   Soon, the tour guide\'s gory descriptions vanished completely, and now all that could be heard was the incessant sound of hammering and the waves breaking on the Batavia\'s hull far below. The lantern above my head creaked as it swung slowly to and fro with the ship\'s light roll, throwing crazy shadows on the walls. Voices again, somewhere outside, laughter, shouts, fragments of a song in the distance. The existence of the lantern, and the waves, did not enter my consciousness immediately, but when I became aware of them I was instantly seized by a strong sense of anxiousness mixed with an almost frantic excitement. I knew at that moment, with all my being and an absolute, clear certainty, that were I to open the small door in front of me now, at this very moment, everything I knew would be gone. It felt as if I was walking a tightrope between two worlds, and it was up to me to make the choice between them. But as the seconds lengthened, the decision had already been made, somewhere deep inside me: without thinking, nor considering the consequences, my hand reached for the door handle.  Then- footsteps, approaching the other side of the door. My heart skipped two beats when it opened slowly, and then disappointment washed over me full force. The moment had passed, and the guy with the camera in the low door opening nearly jumped when he saw my face. I wondered if the sense of loss I felt could be seen on it.   \"God, you startled me,\" the man cried as I got up. \"Where you hiding in here?\"   \"Yeah,\" I said. \"I was.\"  '),(10709,3360363,'27,may,2004','2004-05-27 00:00:00','   Ik heb vanavond, tijdens het koken, mijn vinger in een staafmixer gestopt en vervolgens de staafmixer aangezet.  '),(10710,3360363,'27,may,2004','2004-05-27 00:00:00','   smoking kills drinking kills EVERYTHING KILLS fruitshake shot.  '),(10711,3360363,'23,may,2004','2004-05-23 00:00:00','   After sex ear smoke  TRB 2004 (Toelichtende afbeelding zal spoedig worden toegevoegd.)  '),(10712,3360363,'23,may,2004','2004-05-23 00:00:00','   Wij bestaan want wij hebben notulen.   TRB 2004 (Geschreven op een Dirk van den Broek kassabon)  '),(10713,3360363,'31,may,2004','2004-05-31 00:00:00','   Lief dagboek,  toen ik gisteren met mijn boterham met smeerooft voor mijn rekentuig zat werd ik gestoord door vervelend lawaai. Het bleek afkomstig te zijn van een wentelwiek.  Het leidde mij zeer af van de webpost die ik op dat moment aan het tikken was. Het was mooi weer, ik wilde naar buiten. Snel voegde ik het aanhangsel aan de webpost toe, sloot ik mijn rekentuig af, regelde ik een wiegewacht voor de zuigeling en snelde ik naar buiten. Met mijn wagen in de rijgeleider reed ik naar het strand waar ik een zonnige dag tegemoet ging. Om een uur of vijf at ik een peuzelhapje in de peuzelkroeg. In de buurt waren jongens in de weer met schaatsplanken. Ik heb nog even staan kijken terwijl ik patat met tomatendoop uit een kunststof bakje schraapte. Toen ik thuis kwam betaalde in de wiegewacht waarna ik haar snel de deur uit werkte. Ik startte mijn rekentuig op en webzeilde nog een tijdje over het wereldweb totdat de boekstaven voor mijn ogen begonnen te dansen. Die avond vond ik wel een belangwekkende webschakel, namelijk: http://www.geocities.com/SoHo/Den/7522/WaaromBTL.html Daarna verstuurde ik nog even een belbriefje naar mijn beste vriendin, stopte ik de zuigeling in bed, deed ik de blakers uit en ging ik heeeerlijk slapen.  '),(10714,3360363,'30,may,2004','2004-05-30 00:00:00','   Voordat ik in mijn bed ga liggen en moet kiezen tussen  hemelrijk  en  on the road  (Denk dat het hemelrijk wordt, is een een-avond-boek) even dit:  Ik ga ze missen.  De ZZ Top figuur, die blijkbaar net de laatste keer dat ik werkte besloten had om in Frankrijk te zitten, of op de Veluwe of gewoon in zijn bed met een kopje thee en een keelontsteking. Ik heb niet kunnen zeggen dat ik ging stoppen met werken.   ZZ Top figuur: vaarwel, tot nooit meer ziens. Ik zal je fabuleuze baard en bijpassende snor missen, evenals de druppels die op koude dagen langs je snor naar beneden sijpelden om tenslotte op de toonbank uit elkaar te spatten. Ook zal ik die prachtige knipoog die je me altijd toewierp als je over de drempel van de winkel stapte nooit vergeten. Ik hoop dat de Chiapas koffie smaakte, ik heb het elke keer met veel liefde voor je gemalen.   Holm, die waarschijnlijk ook leed aan een ernstige keelontsteking op de laatste  laatste dag dat ik werkte en klaarblijkelijk door zijn moeder in bed werd gehouden met een kopje koffie, Palermo, gemalen op stand drie. Niet omdat koffie goed is voor je keel maar omdat de kasten vol stonden met halve kilo\'s koffie, in rode papieren zakken, elk bekrabbeld met het woord Palermo in het zelfde handschrift, mijn handschrift. Verder was er niets in huis, er paste niks meer bij in de keukenkastjes.  Holm: ook jou zal ik missen. Als je dit leest: Doe de groeten aan je moeder en aan haar vriendin. Of: Doe maar niet de groeten aan die vriendin, ik geloof dat ze niet erg aardig was en erg veel praatte, zoveel dat jij zelf je ei niet meer kwijt kon en daar mij voor nodig had. Doe de groeten aan je moeder. Ik heb het gevoel alsof ik haar heb gekend terwijl de dag dat ik haar zou ontmoeten nooit heeft plaatsgevonden. Dank je zweedse opa voor je naam, ik heb tijden gedacht dat je Jan heette, of Piet of Bart. Holm past zeker bij je, op een of andere manier. Wees er trots op. Ik zal je enorme buik omspannen met bomber jack, je prachtige vlassige snor, je verlegen zachte gemompel en je enorme bruine portemonnee ook zeker gaan missen. Vaarwel Holm.   De Woordenverzinner: waarschijnlijk zat hij thuis woorden te verzinnen die laatste dag dat ik in de winkel werkte. Ik kan geen andere verklaring bedenken voor zijn afwezigheid.   Woordenverzinner: ik hoop je ooit nog tegen te komen, ik hield van je. Ik hield van jou en je koffie gemalen voor Franse Persing, in een boogje geschreven, met krullen aan de F, de S en de G, versierd met sterretjes.   Verder ga ik missen: de vrouw die altijd wachtte totdat ik eventjes tijd had om met haar te praten en dan vroeg hoe het met me ging, de vrouw met de mooie bos witte haren die elke zaterdag een halve kilo huismerk haalde en (stiekem) een oogje had op Thijmen, de man die evenveel van speciaal belgisch bier hield als van thee en goede muziek, het meisje met de mooie -door een vriendin bedrukte- tas en de altijd groene kleding, de ierse dame die een mooie wandeling wilde maken, de man die naar Canada was geÃ«migreerd en een dag nadat hij koffie bij me had gekocht terug naar de winkel kwam om me een goed boek over zijn dorp in Canada te geven (Bedankt!!), de vrouw met het rimpelhondje en de prachtige zonnebril die ik zeker een uur heb gedragen als zijnde borg voor vijftig eurocent, de jongen met de cello op zijn rug, de man die dacht dat ik de muziek van Emmylou Harris niet kende, de jongen die in de kledingzaak voor te rijke- en depressieve mannen werkte en een enorm Pixies dweper was en tenslotte de hond die graag koffiebonen at.    Het wordt Hemelrijk vannacht.  '),(10715,3360363,'10,june,2004','2004-06-10 00:00:00','   A text lives only if it lives on, and it lives  on  only if it is  at once  translatable  and  untranslatable... Totally translatable, it disappears as a text, as writing, as a body of language. Totally untranslatable, even within what is believed to be one language, it dies immediately.  '),(10716,3360363,'10,june,2004','2004-06-10 00:00:00','   \"Sinners, repent! the end is near\" the cardboard sign of the smallish young man with the downcast eyes by the side of the sixteenth century church says. Silently,  he turns upon us so we can read,  recognizing us for the sinners  we are, and, silently,  tries to save us from his dreaded  faith imagined.  I bid him: goodnight, and receive  an unintelligible rambling sermon-like reply,  and - silently - I thank him for his sign;  for trying to save us, for a mighty good laugh, and for allowing me to contemplate that famous last syllable of recorded time, end of all things (which I am sure is always near) once more  '),(10717,3360363,'18,june,2004','2004-06-18 00:00:00','   The one and only master himself: a  urlLink  Cassady rap ,  and for the real Cassady lover: Cassady is  urlLink rapping for 26 whole minutes  with a little bit of help from Ken Kesey...  '),(10718,3360363,'25,june,2004','2004-06-25 00:00:00','   urlLink    My first 8 mm film  '),(10719,3360363,'15,july,2004','2004-07-15 00:00:00','   Beste verloofden, heeft u ook ontelbaar veel saaie bruiloften meegemaakt? Irriteert u zich ook altijd aan de gelegenheidsband? Luistert zelf uw bruiloft op een andere manier op, huur het Verhalen Bedenk Bedrijf in voor een onvergetelijk huwelijk!  Wij vragen u alle informatie over bruid en bruidegom en geven uw een fantastische speech die op geen enkele manier betrekking zal hebben op de geboden informatie. Wij garanderen u dat het een absoluut onvergetelijk huwelijk zal worden waar geen van uw vrienden aan zal kunnen tippen.   Organiseer een bruiloft waar iedereen over zal praten!  Huur het Verhalen Bedenk Bedrijf!  '),(10720,3360363,'13,july,2004','2004-07-13 00:00:00','   The Columbian Centinel, Boston (Massa), Saturday, April 3, 1813.  THE public will remember how much their feelings were assailed a few weeks since, by numerous accounts in the democratic papers:- First, that Gen. Winchester, and a number of his attendants, were most horribly  murdered, tomahawked, scalped  and  mangled ; and secondly, when it was found, that the General was alive and well, that then the horrid murder, tomahawking, scalping, burning alive in houses and other enormities, were transferred on Major Madison, and other officers of the General\'s army engaged in the battle of  Frenchtown:  (...)  And I wonder who the Hon. D.A. White was, whose name is in such pretty writing at the top. And whose hands held this paper before me, whose eyes read it and what their story was. And who printed these letters, issue 3,025, and who sat round coffee tables sipping tea discussing these issues, the news about Gen. Winchester and Major Madison and the tomahawking and such, which was or was not true and which was of course all linked to the British because they were at war and therefor evil (seems like not much has changed). And the small burnt hole in the middle, how did it get there, and what old words did that fire eat? (Tiny brown piece fluttering down on my carpet, it\'s an n, and it\'s forever lost) And the phone\'s ringing, and I\'m picking up and it\'s now again, and here, and very much 2004. Which is fine with me, really.  '),(10721,3360363,'12,july,2004','2004-07-12 00:00:00','   Er was vlees en vis. En brood en wijn.  De agnost vond de tijd op het keukeneiland. een citaat en nog een  Voor het feest:  iemand wordt doodgeschoten in een trein. Ze zit op een bank met haar rug naar de dader toe. De kogel is niet voor haar bestemd, althans,  zo doet de dader overkomen. De kogel gaat door de bank, door haar rug,  door haar lichaam, en bij haar borstkas er uit. Zomaar op een zonnige dag in een trein van Dorp naar Oegstgeest. Het doet geen pijn. Het ziet er een beetje gek uit.  Er komt bloed uit het gaatje aan de voorkant van haar borstkas.  Maar het doet geen pijn. Het bloed kleeft aan haar vingers,  precies zoals dat hoort als het uit de voorkant van iemand borstkas sijpelt. Uit het bloedende gat schijnt een blauw wit licht.  Of geen licht, een schijnsel. Nuja. __  Een diepe zucht, een verademing, de laatste adem, station Dorp aan Zee. --  En de man die banden plakt na met spijkers te hebben gestrooid.  '),(10722,2284476,'26,November,2003','2003-11-26 00:00:00','   I just read this message today from Sister Hazel lead singer, Ken Block:   We are pumped ya\'ll!! This is gonna be an amazing year - probobly the most exciting in the history of Sister Hazel....  There\'s our 1st DVD coming out on January 20th (We are VERY excited about this!), To top that, we have our 1st double live CD coming out on the SAME DAY with some addition and different material, a new concert tour, a new Sister Hazel CD later in the spring, the 1st annual \"The Rock Slope\" in Colorado (This event is gonna blow peoples mind!), our Lyrics For Life charity event \'04, The Rock Boat \'04 AND a holiday CD ready for next holiday season. Anyone whose been around this band knows that there will be endless fun and countless suprises along the way. We\'ve never been more excited about what lies ahead. If you ever thought about checking out what it\'s like to be a Hazelnut -- Jump in and HOLD ON TIGHT!  Ken Block Sisterhazel   I can\'t wait for the DVD, and the new Sister Hazel CD. If you\'ve never heard of them, you really should go to their  urlLink Website  and take a listen. They\'re a great band. If you like Matchbox 20 or Train, you\'ll almost definently like Sister Hazel.  '),(10723,2284476,'21,November,2003','2003-11-21 00:00:00','   I just don\'t get it some times. In a world filled with thousands upon thousands of products to make you smell good, there\'s always a few people that seem to have let the deodorant revolution pass them by. Do they even know that they smell about as sweet as a butt crack? I just don\'t get how they can smell so raunchy and not notice it. Everyone else around them can, so why the hell can they not realise that, when they\'re smelling the same nasty smell ALL DAY when they move around, that it\'s prolly not the people they pass in the hallways, and it might actually be their stinky ass? My personal belief is that there should be some law against people not wearing some form of protection. And I\'m not talking about the kind of protection that is ribbed for your pleasure, I\'m talking about the kind that protects all of us from your funky ass smellin\' pits.  '),(10724,2284476,'15,November,2003','2003-11-15 00:00:00','   Welp, I\'ve successfully managed in pissing off my mom to the point of her saying I should just move out. What did I do? I got lost coming home (DOT had shut down 206th street, which I usually turn on to off of 37), and when I tried reaching my dad on his nextel, got static. When I tried reaching her, I didn\'t get any response at all. When I finally found my way home, I asked her why she didn\'t have her cell phone turned on, and she got pissed because I was home late (well no shit, I was lost einstein) and because I \"had an attitude\" again. Whatever. I\'m honestly getting so sick of my parents that it\'s not even funny.  '),(10725,2284476,'15,November,2003','2003-11-15 00:00:00','   Grrr. Some the way customers are treated just really makes me mad sometimes. Case in point, rhapsody, the love of my life. I signed onto my account today, hoping to listen to the new train album (I already own it, but it\'s so much easier to just log into rhapsody and play it). Once I had logged in, I scrolled down through my list of artists in my virtual library. For some reason, train was greyed out. When I clicked on them, I found that I could no longer listen to them on the rhapsody service. What the heck? Not only was trian greyed out, but so were several other artists in my list. Don\'t get me wrong, I realise that there is some music that rhapsody won\'t be able to offer, but it\'s my personal belief that when you offer a service and charge for it, you should continue to offer the full service. I\'m starting to think about switching over to the newly relaunched (and 100% legal) service, Napster. Not only do they offer more music, but they also let you download mp3 files, something that rhapsody has yet to offer.  '),(10726,2284476,'10,November,2003','2003-11-10 00:00:00','   I know a bunch of friends have been asking me what I mean when I talk about Rhapsody. Rhapsody is a legal music service that lets me listen to over 400,000 songs whenever I want for 10 bucks a month. If I want to burn any of those songs, it costs me $0.89 a song. Not bad considering I\'m always getting only the songs that I want, rather than buying a CD at Best Buy and getting two songs that I want, and then a bunch of filler for $13 - $15.  ---- I\'m listening to: Blue Man Group  Audio  I\'m watching: Law and Order  '),(10727,2284476,'09,November,2003','2003-11-09 00:00:00','   Why is my dad such an ass? Honestly. After the crappy day that I\'ve had, as soon as he gets home, he gets on my case. I don\'t feel like retyping it all, so I\'ll post a peice of an MSN convo that I had with andy while this was going on: 11/10/2003  5:04:44 PM  Cameron:  i f***ing hate my parents  11/10/2003  5:05:03 PM  Cameron:  jesus christ, they started yelling at me saying that I need to drop my tone  11/10/2003  5:05:12 PM  Andy:  are they pmsing  11/10/2003  5:05:18 PM  Cameron:  I didn\'t have a freaking tone with them, I didn\'t say two words  11/10/2003  5:05:34 PM  Andy:  why are they doing that  11/10/2003  5:05:37 PM  Cameron:  She just got pissed off because I was talking to you guys when she asked me to come talk to her  11/10/2003  5:05:43 PM  Cameron:  Because they\'re both assholes  11/10/2003  5:06:04 PM  Cameron:  I want out, this is getting so freaking old.  11/10/2003  5:06:14 PM  Cameron:  They can both go to hell. I told them that too.  11/10/2003  5:06:24 PM  Andy:  just stay in your room the rest of the night  11/10/2003  5:06:32 PM  Cameron:  that\'s what I do every night  11/10/2003  5:06:40 PM  Cameron:  Because everytime I come out, this is what happens  11/10/2003  5:06:43 PM  Cameron:  I\'m tired of it dude  11/10/2003  5:06:57 PM  Cameron:  I get enough of it at school, I shouldn\'t have to deal with it at home too  11/10/2003  5:07:24 PM  Andy:  tell em to back off  11/10/2003  5:07:29 PM  Cameron:  I did  11/10/2003  5:07:45 PM  Cameron:  My dad said I need to quit with the tone unless I want a boot up my ass  11/10/2003  5:07:49 PM  Cameron:  To quote him exactly  11/10/2003  5:08:13 PM  Andy:  thats terrible  11/10/2003  5:08:39 PM  Cameron:  God he\'s such a prick  11/10/2003  5:08:51 PM  Cameron:  Now I\'m crying like a freaking baby  11/10/2003  5:09:01 PM  Andy:  dude don\'t cry  11/10/2003  5:09:07 PM  Cameron:  It\'s prolly a good thing you didn\'t come over   It didn\'t end there. Later my dad got mad because I told him to wait a couple minutes before he moved the cable modem (for the internet) while he was putting in a new desk. I asked him to wait because I was talking to andy, elizabeth, and jessica, and wanted to finish my conversations with them without just all of a sudden signing off with no reason. He told me he\'d move it if he wanted, I told him to just wait a couple minutes, and he told me I better lose the attitude if I wanted to keep my front teeth. I\'m so sick of all this crap, I know my dad wouldn\'t actually hurt me physically (he\'s only got really rough with me a couple times), but he still always makes me feel like I\'m anything but a son to be proud of. Maybe he\'s right. I mean, I\'ve never really shined in anything, what is there to be proud of?  '),(10728,2284476,'09,November,2003','2003-11-09 00:00:00','   I want out. Okay, today\'s been a major pain in the ass for the most part. When I first got up, I could just tell it was going to be a crappy day. Nothing helped prove this right for the first two periods of the day (Physics and US History), and Web Design wasn\'t that bad either (other from the usual cold shoulder from the nerds because I\'m not like them). The last period of the day (CAD) is when things started getting crappy. First of all, on the way there, Patrick Schiess came up behind me and asked me if I actually moved when I walked. I have a bad knee, and it makes me walk a little differently from most people (it\'s not egor bad, but still noticeable). I brushed this off, and thought it was just Patrick trying to get to me because he\'s an arrogant prick. I went to CAD, and sat next to the people that I usually do (my friends Melinda and Joey on my right, Michael Galyan and some arrogant panzies on the left). About halfway into the class, I thought I heard Eric Forbes and his \"groupies\" say something about me. I listened in, and they were making fun of me and joey, calling him and I both fags, and mocking us when we\'d say something to someone. It brought back a lot of memories from Kindergarten through prolly 8th grade when I used to get picked on, and beat up (physically). While I know it\'s something I should brush off because their opinions don\'t mean anything to me, when someone makes fun of you non-stop for an hour and a half, it starts to eat at you. Needless to say, I was ready to go home after CAD, and was actually looking forward to today because Andy was coming over, and he hadn\'t been over to my house in nearly three weeks. Well, he asked his mom yesterday if he could come over, and she said she\'d tell him in the morning, only she forgot to tell him. So, after school, he tried calling her, but she wouldn\'t answer. He told me to follow him to his house, and he\'d ask her when we got there. I had to take my ex girlfriend Beth home, but I wanted to see if he could come over, so we went to his house with her in tug. When we got to his house, his mom wasn\'t home, so he called her cell phone again, and she told him that he couldn\'t come over, even though I had made plans and told some people that are working on a project with me that I couldn\'t do anything because he was coming over (meaning I wasted an entire afternoon that I could have used on the project). I was a little irked that she wouldn\'t let him come over because he would end up sitting at home doing nothing anyway, but I brushed it off. I left his house to take Beth home, and on the way she said that it was a waste of time, and that she could have been home by the time I found out that Andy couldn\'t come over. When we got to her house, she blew up at me for god knows what, said cya later, slammed my door, and ran inside. I just LOVE it when people that I\'m helping out show their thanks by being an ass to me. What a fun day it\'s been...and it\'s not even over yet *groan*.  '),(10729,2284476,'09,November,2003','2003-11-09 00:00:00','   Well, another day, another dolla....oh wait, that\'s right I don\'t have a job. I was supposed to go apply for a job at Target as a cashier where my best friend (Andy), and another good friend (Stephen) work. Well, as usual, plans changed. I took along a friend (Elizabeth) to go with me because she needed to pick some things up, and I found out from andy, who was working, that the application takes fourty-five minutes. I didn\'t want to leave Elizabeth by herself while andy was working and I was applying, so we ended up wandering around target with andy while he as on his break, and then we went off to best buy so that I could grab Blue Man Group\'s old debut CD \"Audio\". I went home, blasted the CD, then headed off to bed. Not much else went on today (yup, reaaaal exciting day, huh?).  '),(10730,2284476,'30,December,2003','2003-12-30 00:00:00','   So life still isn\'t good. Breaks are great most of the time, but this time it\'s suck. I haven\'t been able to see my friends as much as I like, and I haven\'t been able to see a couple really good friends at all. It also doesn\'t allow for wounds to heal, it just allows them to get deaper. Minor things that occured at the end of school have become major, and caused a lot of rifts in friendships.  '),(10731,2284476,'23,December,2003','2003-12-23 00:00:00','   Well, to continue with my down feeling, I guess I might as well bring it up. It doesn\'t feel like christmas at all. I don\'t know what it is, but it doesn\'t feel like christmas at all to me. I hate to break it to everyone, but the christmas spirit is dead. Loving each other, and celebrating the birth of Jesus has been replaced with loving that new expensive MP3 player that you got, and celebrating the two weeks off that you get. It happened to St. Valentines day, and now the time has finally come for Christmas to finally sucumb the same fate. It\'s official, christmas has lost all non material meaning, and is now a halmark holiday. I could go into this more, but I already feel like going to sleep without waking up because of what I rambled about below. Anyway, for those of you that are still fooled into the belief that christmas is more than a way for american eagle to boost its bottom line, I\'ve picked out some christmas music for you care of rhapsody. Take a listen if you\'d like, I\'ve got RHAPSODY and Napster links, so just click to listen through whichever program you like more.   urlLink Napster Link   urlLink RHAPSODY Link  '),(10732,2284476,'14,December,2003','2003-12-14 00:00:00','   I admit it, I don\'t like elvis. I know you\'re not supposed to hate anything, so I guess you could say I strongly dislike elvis to the point where I wish he were born a mute. Well, I was scanning through rhapsody today, and found a new remix of Rubbernieckin\', and it\'s actually good. And by good I mean I was able to listen to it without wanting to stab my eyes out with a spork. So yeah, I just thought I\'d share that. Take a listen and see what you think.  urlLink Rubberneckin\'  '),(10733,2284476,'16,January,2004','2004-01-16 00:00:00','   Ugh, I\'ve had a headache from the 9th level of hell (if you don\'t get that check my last post), and I\'ve had it for the past two weeks...it just won\'t go away. Isn\'t that usually a bad sign? Oh well, if ya die ya die. It happens to everyone sometime. In the meantime, I\'m going to keep chomping my great-value brand ibprohen and watching NYPD Blue on TNT.  '),(10734,2284476,'12,January,2004','2004-01-12 00:00:00','   Well I just took the \"Dante\'s Inferno Test\" to see where I\'m going when I die, and it looks like a certain someone was right...I really AM going to hell. Only the first layer, but still...I\'m headin\' there!   The Dante\'s Inferno Test has sent you to  the First Level of Hell - Limbo!   Here is how you matched up against all the levels:     Level    Score       urlLink Purgatory   (Repenting Believers)   Very Low       urlLink Level 1 - Limbo   (Virtuous Non-Believers)   Very High       urlLink Level 2   (Lustful)   Low       urlLink Level 3   (Gluttonous)   Low       urlLink Level 4   (Prodigal and Avaricious)   Low       urlLink Level 5   (Wrathful and Gloomy)   Low       urlLink Level 6 - The City of Dis   (Heretics)   High       urlLink Level 7   (Violent)   Low       urlLink Level 8- the Malebolge   (Fraudulent, Malicious, Panderers)   Low       urlLink Level 9 - Cocytus   (Treacherous)   Moderate      Take the  urlLink Dante\'s Inferno Hell Test  '),(10735,2284476,'11,January,2004','2004-01-11 00:00:00','   Michael Jackson is a pedophile. Me...I\'ve got my own secret, disturbing problem. I admit it...I\'m an  urlLink audiophile . I love everything about audio. Music, Mixing, Portable Audio, Speakers. Speaking of speakers, my new Logitech\'s got installed yesterday and they kick ass.  '),(10736,2284476,'08,January,2004','2004-01-08 00:00:00','   You always hear about it happening to other people, but it\'s one of those things you never think will happen to you. I don\'t get it...everything was going just fine and then it happened. WHY the hell do they feel they have to interrupt Law and Order, for the NBA? Isn\'t that why we have ESPN 1-20? Drama isn\'t the NBA Playoffs, drama  is  Law & Order on TNT. I think my second favorite commedian, Larry the Cable Guy (Phil Hartman will always be #1), described how I feel best when he said \"I was madder than a one legged etheopian chasing a doughnut down a hill\". While I\'m on my soapbox, I guess I\'ll throw out another complaint with TNT. TNT usually is a great network to watch, with great original shows. But WHAT THE HELL were they thinking when they decided to put the new Hootie and the Blowfish music video in heavy rotation just so that they could advertise their new made-for-tv movie \"Goodbye Girl\"? I used to love Hootie and the Blowfish, but quite honestly, I\'d rather be a cupless wearing dummy for a women\'s self defense class than listen to that damned song one more time.  '),(10737,2284476,'07,January,2004','2004-01-07 00:00:00','   What\'s better than an overpriced mp3 player? An even smaller mp3 overpriced mp3 player, now in five fruity colors! Yeah, I know...I got an iPod, and I love it. But really....what the hell was apple thinking when they released these new mini disasters? The pink one looks like it belongs in a tampax commercial, to be honest with you...and aside from the silver one, I don\'t think I\'d want to be caught dead with one unless I wanted people to think that I should be guest starring on Will and Grace. To keep with the apple thing, I finally tried iTunes. All I have to say is KICK ASS. I can\'t believe how nice it is being able to click \"buy track\" and have it automatically sync\'d to my iPod. With my old Zen (I almost miss it and its 80s-esque cassette player look) I had to manually drag new songs onto it, which got to be a pain in the ass when I had 1400 songs in my collection.  ');
INSERT INTO `msgs` VALUES (10738,2284476,'07,January,2004','2004-01-07 00:00:00','   Nothing major happened today, so I\'ll just ramble on about some stuff I guess.  What\'s the best way to piss off a teacher? Call her a drunk, of course! And who in their right mind would be dumb enough to do something like that? No-one...but I\'m not exactly right in my mind, so I guess that\'s why I said it. Word to the wise: No matter how much your teacher acts like a drunk, never call them one.   My schedule this semester sucks! I didn\'t get my favorite teacher back (Mr. Knotts...he\'s insane in a cool way, so I\'m going to try to transfer to his class), and at the same time, my new english teacher is off the edge freaky, but still kinda cool.  So yeah, random post today. I\'m going to go back to watching NYPD Blue, and maybe...just maybe I\'ll make another post today because I know that everyone that reads this refreshes the page every few seconds just to see if I\'ve posted something new...of course when the number of people reading this is 0, I can also say that they\'re all girls hotter than paris hilton, and that they all want me badly...  '),(10739,2284476,'07,January,2004','2004-01-07 00:00:00','   I feel like I\'m waterlogged. I don\'t know why though, I\'ve only had three dr. peppers, two glasses of tea, and a bottle of water in the past...two hours. I\'m gonna call the guiness people right now, I think I feel the world record for the longest pee coming on.  '),(10740,2284476,'23,June,2004','2004-06-23 00:00:00','   I\'m not sure if there are words that could do this concert justice. Dexter Freebish came on a little after 7:45, and rocked the house until Sister Hazel came on at 9:00. Sister Hazel was absolutely amazing. You could feel the energy coming off the guys as they jammed up on stage. The set list for both bands couldn\'t have been any better, either. The highlight of the evening for me was when Ken Block (lead singer of Sister Hazel) came out for an encore and sang an acoustic version of \"Running Through The Fields\", a song that he wrote about his brother who he lost to cancer. It actually brought chills to listen to him, and watch how the huge crows of people became instantly silent. But it\'s easy to see how, as the song was so moving that literally everyone was in tears. I\'ll post a more thorough review of the concert later, but for now I just wanted to get a picture up and give a quick overview of what it was like.   urlLink    Ken Block  '),(10741,2284476,'16,June,2004','2004-06-16 00:00:00','   Second week that it\'s rained every day. I\'m off to Lowes to buy some more wood for the arc...  '),(10742,2284476,'15,June,2004','2004-06-15 00:00:00','   My friends never cease to amaze me. Even when I\'m so down that pounding a bottle of zoloft couldn\'t bring me up, one of my friends is always there for me. I really am blessed to have friends as good as I do.  '),(10743,2284476,'13,June,2004','2004-06-13 00:00:00','   Well, turns out he didn\'t get the neighbor\'s dog knocked up, but Hank\'s still a dad. I went out to feed him the other day and he was carrying a teddy bear around  (god only knows where he got it). He hasn\'t let the thing out of his site since. I just want a normal dog. I can\'t take anymore of the pooh who humps everything in sight (I\'m not sure if rigsby ever found out that the pooh humped his shirt on mayday), and Hank who evidently is emotionally tied to a teddy bear.   urlLink    hank and his \"baby\"  '),(10744,2284476,'12,June,2004','2004-06-12 00:00:00','   Is that sweet-ass or what? It\'s a little robot gizmo thing. Evidently it does everything from walking/running to throwing a ball to break dancing. Screw the Sony aibo dog I was gonna buy, I\'m gonna pick one of these up for a couple hundred less.  '),(10745,2284476,'12,June,2004','2004-06-12 00:00:00','   If someone told me to describe today in three words, I\'d either say \"shut up, I\'ll use as many as I want\", or I\'d say \"Boring as hell\" depending on the kind of mood that I was in. Anywho I did absolutely nothing today aside from watching the MTV Movie Awards (someone clue MTV in that it\'s  NOT  halloween). Oh, and anyone else see those new \"rip, slip, brush, ah\" commercials for that porta-toothbrush thing? Personally, I think it\'d be a better trojan commercial than anything else.  '),(10746,2284476,'11,June,2004','2004-06-11 00:00:00','   I didn\'t have anything to write about today, so I thought I\'d post a picture from the band\'s latest gig at the Mariott Indianapolis. Yes I know that the pictures are blurry, but instead of blaming it on my obvious lack of photographising skills (yes that\'s what it\'s called, smartass) I\'m going to say that I was inspired to take pictures of the stage as it would have been seen by one of the many rich drunks that were present that night.   urlLink    ooo look at the pretty colors  '),(10747,2284476,'11,June,2004','2004-06-11 00:00:00','   This one care of Sister Hazel. This song nails the way my (lack of a) love life is.    If it\'s to be That there\'s just no one for me I\'ll try not to cry much So no one will see But if it\'s to be That there is anyone for me Our hearts will shine So everyone will see I\'ve got some time, So maybe if you\'re free I could use a friend Just to talk to me   Spent most of my life without Anyone near Never cared what happened I never saw clear But I\'m getting older now And I\'m, thinking of my end And to leave without love Or without a friend  I\'ve got some time So maybe if you are free I could use a friend  Before I die   It\'s not over yet... I\'m still standing   If it\'s to be  That there\'s just no one for me I will try not to cry much So no one else will see I\'ve got some time So maybe if you\'re free I could use a friend Just to talk to me  '),(10748,2284476,'11,June,2004','2004-06-11 00:00:00','   Do you ever get those feelings like everything is wrong with your life, even though everything is right? I feel like I\'m not the person that I am. Like the life that I\'m living isn\'t mine, it\'s someone elses.    I cannot focus here, I cannot find my way My mind is blurred by all the things that people say I dream of bliss, I dream of heaven here on earth Won\'t you please set me free  '),(10749,2284476,'11,June,2004','2004-06-11 00:00:00','   W00P, only ten more days until the Sister Hazel & Dexter Freebish concert, and I\'m stoked! I can\'t wait. It\'s gonna be a kickass concert. I talked to Sister Hazel\'s management, an they said cameras are ok, so I\'ll be sure to take a load of pictures and post them up here after the show.   urlLink    Only ten more days until the Sister Hazel/Dexter Freebish concert!&nbsp; urlLink  '),(10750,2284476,'11,June,2004','2004-06-11 00:00:00','   How sad is this? I\'ve become so bored in the past week that I actually put my picture up on one of those \"hot or not\" type websites. It\'s kinda cool though. I\'ve managed to meet a couple really cool (and not so cool) people. Evidently a lot of people think I\'m either really hot (they gave me a 10), or think I\'m really ugly (they gave me a 1). Oh well, can\'t please everyone.  '),(10751,2284476,'11,June,2004','2004-06-11 00:00:00','   I think I need to go to a quack. Honestly. Ever since I\'ve bought my iPod, I\'ve become obsessed with buying music on iTunes. So far I\'ve bought a little under 100 songs off of iTunes, and I don\'t see myself stopping anytime soon. It\'s just so easy to do though. I hear a song on Z995, and all I have to do is go home, open iTunes, click on the song, and it\'s automatically on my iPod so I can hear it anywhere I want.  urlLink To prove my point, let\'s take a look at my inbox, shall we?  '),(10752,1223561,'08,June,2004','2004-06-08 00:00:00','    Imagine.......   If you got into a taxi and he started driving backwards, would the taxi driver end up owing you money?  If it\'s zero degrees outside today, and it is supposed to be twice as cold tomorrow, would the degrees still be zero?  Why is it called building when it is already built?  If pro is the oppsite of con, is progress the opposite of congress?  If you drop a chameleon in water, will it turn clear?  If tin whistles are made of tin, what do they make fog horns out of?  Is it possible to be totally partial?  If a book about failures doesn\'t sell, is it a success?  Isn\'t it a bit unnerving that doctors call what they do \"practice\"?  If all the world is a stage where is the audience sitting?  If work is so terrific, how come they have to pay you to do it?  If you\'re not supposed to drink and drive, then why do bars have parking lots?  If you take an Oriental person and spin him around several times, does he become disoriented?  If people from Poland are called Poles, why aren\'t people from Holland called Holes?  Do infants enjoy infancy as much as adults enjoy adultery?  Why is the man who invests all your money called a broker?  When cheese gets its picture taken what does it say?  If horrific means to make horrible, does terrific mean to make terrible?  Why isn\'t 11 pronounced onety one?  If lawyers are debarred and clergymen defrocked, doesn\'t it follow that electricians can be delighted, musicians denoted, cowboys deranged, models deposed, dry cleaners depressed?  Why is it if someone tells you there are 1 billion stars in the universe, you will believe them but if they tell you a wall has wet paint you will have to touch it to be sure?  If Fed Ex and UPS were to merge, would they call it Fed UP?  Do Lipton Tea employees take coffee breaks?  I thought how people seem to read the Bible a whole lot more as they get older, then it dawned on me ...they are cramming for their final exams.   '),(10753,1223561,'08,June,2004','2004-06-08 00:00:00','    Calvin and Hobbes Quotes  Happiness isn\'t good enough for me! I demand euphoria! ------------------------------------------------------------------- The purpose of writing is to inflate weak ideas, obscure pure reasoning, and inhibit clarity. With a little practice, writing can be an intimidating and impenetrable fog! ------------------------------------------------------------------- Calvin: I\'m a genius, but I\'m a misunderstood genius. Hobbes: What\'s misunderstood about you? Calvin: Nobody thinks I\'m a genius. ------------------------------------------------------------------- Well, it just seemed wrong to cheat on an ethics test. ------------------------------------------------------------------- Calvin: Can you make a living playing silly games? His Dad: Actually, you can be among the most overpaid people on the planet. ------------------------------------------------------------------- If you do the job badly enough, sometimes you don\'t get asked to do it again. ------------------------------------------------------------------- The only skills I have the patience to learn are those that have no real application in life. ------------------------------------------------------------------- Some people are pragmatists, taking things as they come and making the best of the choices available. Some people are idealists, standing for principle and refusing to compromise. And some people just act on any whim that enters their heads. I pragmatically turn my whims into principles! ------------------------------------------------------------------- Sometimes I think the surest sign that intelligent life exists elsewhere in the universe is that none of it has tried to contact us. ------------------------------------------------------------------- It seems like once people grow up, they have no idea what\'s cool. ------------------------------------------------------------------- The real fun of living wisely is that you get to be smug about it. ------------------------------------------------------------------- Calvin: Know what I pray for? Hobbes: What? Calvin: The strength to change what I can, the inability to accept what I can\'t, and the incapacity to tell the difference. -------------------------------------------------------------------  I say, if your knees aren\'t green by the end of the day, you ought to seriously re-examine your life. ------------------------------------------------------------------- But Calvin is no kind and loving god! He\'s one of the old gods! He demands sacrifice! ------------------------------------------------------------------- It\'s psychosomatic. You need a lobotomy. I\'ll get a saw. ------------------------------------------------------------------- Why waste time learning, when ignorance is instantaneous? ------------------------------------------------------------------- Hobbes: What would you call the creation of the universe? Calvin: The Horrendous Space Kablooie! ------------------------------------------------------------------- If something is so complicated that you can\'t explain it in 10 seconds, then it\'s probably not worth knowing anyway. ------------------------------------------------------------------- Who wouldn\'t be interested in everything we do?! ------------------------------------------------------------------- I think we dream so we don\'t have to be apart so long. If we\'re in each other\'s dreams, we can play together all night. ------------------------------------------------------------------- As a math atheist, I should be excused from this. ------------------------------------------------------------------- This one\'s tricky. You have to use imaginary numbers, like eleventeen ------------------------------------------------------------------- You can present the material, but you can\'t make me care. ------------------------------------------------------------------- YAAH! DEATH TO OATMEAL! ------------------------------------------------------------------- >From now on, I\'ll connect the dots my own way. ------------------------------------------------------------------- I\'m learning real skills that I can apply throughout the rest of my life .... Procrastinating and rationalizing. ------------------------------------------------------------------- Reality continues to ruin my life. -------------------------------------------------------------------  Calvin: Sometimes when I\'m talking, my words can\'t keep up with my thoughts. I wonder why we think faster than we speak. Hobbes: Probably so we can think twice. ------------------------------------------------------------------- I liked things better when I didn\'t understand them. ------------------------------------------------------------------- Is it a right to remain ignorant? ------------------------------------------------------------------- I think nighttime is dark so you can imagine your fears with less distraction. ------------------------------------------------------------------- Miss Wormwood: What state do you live in? Calvin: Denial. Miss Wormwood: I don\'t suppose I can argue with that... ------------------------------------------------------------------- What\'s the point of wearing your favourite rocketship underpants if nobody ever asks to see \'em? ------------------------------------------------------------------- My life needs a rewind/erase button.   '),(10754,1223561,'20,June,2004','2004-06-20 00:00:00','    For all the times that Mr. Prasad knew of himself, he had never been happy. For happiness didn\'t find it\'s way to his home, ever since he knew of himself. Happiness is a state of mind, always said Mrs. Prasad, whose source of unhappiness was her husband\'s unhappiness and whose spirit was her husband\'s only infinite resource. My mind doesn\'t even have country, let alone a state, always poor-joked Mr. Prasad.  Chandrashekar Venkat Prasad (First, Middle and last names) was born to a not-so-rich, not-so-poor class of family, at a time when middle class was almost synonymous with the not-so-poor, and the upper middle classes were anonymous. His father had been an heir to one-sixth of his grandfather\'s ancestral property. Prasad\'s grandfather had amassed much wealth, and his father added a lot more to that. Lot more than all the amassed wealth could handle. Casino gamblings, business ramblings, fighting siblings and so on and so forth had negated all the advantages of what could have been a profitable heirdom. By the fourth year, Prasad (arguably the earliest memories the Prasad had) had witnessed most of all what the great epics of our land (The Ramayana, the Mahabharata, to name a few) talk, speak of and teach us. Though his infantile and juvenile mind couldn\'t grasp the gist of those epics, the reality of his life slowly hardened him, hardened him against hope, joy and happiness. The harsh realities of his life spawned their own myths, legends, histories, songs, epics, lessons and morals. Life became a responsibility (if you wanted to shun it)or a response to a responsibility (if you responded to it). Dreams were there only to be dreamt, life was there only to exist, and the rest was there, only because there was nothing that could be done - either for it-to-be-there or it-not-to-be-there. '),(10755,1223561,'11,July,2004','2004-07-11 00:00:00','   urlLink    All smiles for what reason!!!&nbsp; urlLink '),(10756,1223561,'11,July,2004','2004-07-11 00:00:00','   urlLink    Sos, umbrella, umbrella, umbrella, ad infinitum&nbsp; urlLink '),(10757,1223561,'11,July,2004','2004-07-11 00:00:00','   urlLink    Sand Castles&nbsp; urlLink '),(10758,1223561,'11,July,2004','2004-07-11 00:00:00','   urlLink    too close to Infinity&nbsp; urlLink '),(10759,1223561,'11,July,2004','2004-07-11 00:00:00','   urlLink    Relax\'sea\'ng&nbsp; urlLink '),(10760,1223561,'11,July,2004','2004-07-11 00:00:00','   urlLink    Four on beach at the Fort De So to Beach&nbsp; urlLink '),(10761,1223561,'11,July,2004','2004-07-11 00:00:00','   urlLink    All the Beach\'s boys&nbsp; urlLink '),(10762,1223561,'11,July,2004','2004-07-11 00:00:00','   urlLink    Waiting for the Ferry back to Honeymoon... I look awful&nbsp; urlLink '),(10763,1223561,'11,July,2004','2004-07-11 00:00:00','   urlLink    Sun, Sea (not in the pic), Sand, Sandy and Sos&nbsp; urlLink '),(10764,1223561,'11,July,2004','2004-07-11 00:00:00','   urlLink    Ravi Raja and the Rented Car&nbsp; urlLink '),(10765,1223561,'11,July,2004','2004-07-11 00:00:00','   urlLink    The King and the Car - the Gallant Raja&nbsp; urlLink '),(10766,1223561,'11,July,2004','2004-07-11 00:00:00','   urlLink    The Fort De Soto North Beach... It was much hotter than it looks&nbsp; urlLink '),(10767,1223561,'11,July,2004','2004-07-11 00:00:00','   urlLink    Whales on the beach, of the beach by the kids on the beach&nbsp; urlLink '),(10768,1223561,'15,February,2004','2004-02-15 00:00:00','   Hi guys!  Just started off with my First blog page today! Happy Valentine\'s day to you all! Im sure you all must have enjoyed the best where ever you are..cos that\'s what Cool-it is all bout! live life to the fullest...Enjoy!  Shall write some more soon!  Bye  '),(10769,1223561,'27,February,2004','2004-02-27 00:00:00','   Hours..days...weeks...months...and ofcourse Years..are fast running. There\'s no time even to think that they are passing off so fast. is there really no time or dont I want to take out time? I am really not sure. This \"dreadful\" mechanical life is killing me from inside. Its not that I am doing an extra effort than everyone else...but is this what I really wanted in life? I keep asking this question to myself innumerable times everyday...still have got no answer..or is it I dont want to listen to the answer cos I am scared of it? All my hidden secrets, desires, fantasies....I am sacred they will all die down soon and I will remain \"life less\".  '),(10770,1223561,'02,March,2004','2004-03-02 00:00:00','   Awesome!  Today was damn kewl. I made 147/150 in Experimental Design...was expecting may be about 90 or so..kewl. Yesterday Nishanth and I had along talk about the usual stuff...asusual no conclusion..its all about proirities. Nothing can be done on it. Anyways I have forgotten about it..dont wanna discuss about it again. The blog is the best thing to have hapend to me after a long time. Its really cool..I can type all my feelings...more than friend!  '),(10771,1223561,'07,March,2004','2004-03-07 00:00:00','   Daytona!  Just back from the DAYTONA Beach trip. It was kewl. Had fun. Beaches..Babes..Bikes..It was beautiful! Looks like the B words are in huge demand these days! It was the \"BIKE week\" last week at Daytona. Bikers from all over the country assembled there. The whole city was Buzzing with Bikes!  Along with the good things, there are certain \"unpleasant things\" associated. Asusual, people boozed,Ravi and I were the only silent spectators there. Things got heated up to such an extent that some of them were almost ready for blows! Ravi and I were damn pissed at what happened. Why must one ever trigger sensitive issues into discussion and that too when nobody wants to listen to the other but simply put his point forward? Other than this, EGO, the \"EVIL\" ONE, clearly in the group.  I dont deny we had a good time. We played, laughed, had a sexy time in the beach. But there was something lacking. May be I am wrong. May be it is just me who feels like that...if it so, I pray I be Wrong.  Today is Holi. Called up home. It would have been real fun to be at home at this time..but anyways, I am still enjoying here.  I was really happy to know that Veedee got a call from SPJain. I sincerely hope he makes it into it. For all that he has done and sacrificed, he deserves that.  Thats it for today! Bye Blog!  '),(10772,1223561,'09,June,2004','2004-06-09 00:00:00','   Friends....Finally got the GMail id! Pretty cool uh... Thanks to Ravi for inviting me! 1000 MB free space..nothing to delete. My new email..sorry \"gmail\" id is srinivas.chundi@gmail.com   Saw Meenaxi and YUVA recently. Both were pretty good...I was happy about the way M.F.Husain has finally tried to each out to the public.   And do we need to say anything about ManiRatnam??? Simply Superb asusual!  Will get back soon.  Srinivas.Chundi !!!! GOD I m obsessed!  '),(10773,1223561,'08,June,2004','2004-06-08 00:00:00','   Kewl...Ravi got hi Gmail id today! Hope I also get one soon!!!! Gotta meeting...gotta crash....dont want to do both....  chalo..shall catcha later!  '),(10774,1223561,'07,June,2004','2004-06-07 00:00:00','   test  '),(10775,1223561,'07,June,2004','2004-06-07 00:00:00','   Testing  '),(10776,1223561,'20,June,2004','2004-06-20 00:00:00','   Hey guys!  Watched Lakshya yesterday. Well made film with no flaws difficult to find! I guess the word Lakshya says it all- \"Lakshya- That IS the word\" as Preity Zinta repeats in the movie. It\'s kinda hard to digest that Farhan Akhtar has directed Lakshya after the witty, whacky Dil Chahta Hai. But he deserves all applause for his effort. So does Hrithik Roshan. He was a perfect match for \"Karan Shergill\". His performance was damn good.  The songs though would keep you wanting for more and infact better. The music is definitely not on par with Dil Chahta Hai. On the whole a film worth watching. Well after watching the film, I asked myself \"What is my Lakshya?\" Hope I find an answer to it very soon.  Cheers, Sos  '),(10777,1223561,'06,July,2004','2004-07-06 00:00:00','   Hi guys!!!  The CFC Chapter is back to share its Long weekend trip to the best beaches in USA. And mind you, the beaches were the \"best\" in everyway! Also, this was the first trip that we by ourselves drove to all these places, now that we all have our licenses!  It all started on Friday evening, when Ravi, Raja and I went to the Enterprise rental office to rent a car- A MITSUBISHI Gallant 2004 ! The car was damn sexy. I was fortunate enough to be the first one to drive the car, the whole of friday evening. Before I begin, it would be great if you could have a look at this link. These are the places where we had been!   http://www.floridasbeach.com/bch_nat/topbeaches.aspx  On Saturday morning, we hit the Interstate- 4 towards Tampa Bay. We went to Fort De Soto Beach first. Fort de Soto Beach has been officially ranked as the Number 1 beach in continental USA ! So was it...truly number 1 ! Once we reached the beach, we all got into our shorts and were damn excited to take a plunge in the warm, inviting waters of the Gulf of Mexico, more so the Atlantic Ocean. The water was clam, with not many waves, and truly pleasing to have a dip. The best part of the beach was the Canoeing and Kayaking section in the back waters. We were 7 of us, formed 3 teams of 2 each. And then began our Kayaking adventure. It was my first time and I enjoyed each and evry bit of it, though it was very tiring! We were there in our kayaks for an hour..having fun in the backwaters, ocassionally hitting the bushes on the sides! While kayaking, we happened to meet some of our batchmeates at BITS. I m sure everyone would know them here! We came across Shruthi Dash, Vinay Sharma, Shobana Desikan, Siveramakrishnan(K 2) and Deepika! We exchanged \"Hellos\" and the we were back on our own ways. Well it was nice to see them there...of all places in the middle of an ocean! They were on a trip to Florida from Ohio. So guys, all those not planning to visit Florida when you are in US, remember that you are missing out something really beautiful!  Back to kayaking...after that, we went to chill out on Fort De Soto Beach. The beach was lovely, with white sands ( almost as white as a paper) and as soft as flour! At about 5 we started off to Clearwater. That was about an hour drive from Fort De Soto. As the name says \"Clearwater\", the waters are literally \"clear\". You can clearly see the sea bed. All of us jumped into the waters again! Saw in the sunset while in the water and it was simply superb! We then drobe back to our motel at about 10 in the night, took bath, this time with fresh water!, ordered Chinese food and slept by 1:00. We had no enthu to drive back to the shore to see the night life there. We were tired.  Sunday morning, we had breakfast at McDonald\'s near the motel and started off straight to Caladesi Island (ranked 5th best beach in USA). To reach this place, we had to drive to Honeymoon Island and take a ferry from there! There is no road to Caladesi. By the time we reached Caldesi it was 12 noon. We were allowed to stay there for 4 hours, that is till 4:00. And the island was beautiful. There\'s nothign in this place except for the beach. Truly, the beach was amazing. The waters were shining in the bright sun light. Once again, we all jumped into the ocean, played in the waters till we got totally exhausted. We rented Umberellas and chairs, for shade. And it was a great feeling to just relax on the beaches, with the goggles on your face and you resting on the chairs under the umberellas, with the Ocean in front of you! Before we realised it was 4:00 andwe were su[pposd to head back to Honeymoon Island. At about 4:30 we started off to Orlando-the City Beautiful! The weather was very rough, it was raining heavily with thunderstorms and lightning. But we managed to drive well and reach the Downtown (the place where the government buildings are located in a city). It was 4th of July- the American Independence Day. We stopped at Lake Eola. There were tens of thousands of people gathered around the lake for the fire works display. The lake was very scenic. With the Downtown buildings on one side, trees on the other, a lovely weather, and a marvellous \"Rainbow\" in the sky! The entire picture is indescribable! After a 2 hour long wait, we got to see the fire works display and it was definitely worth waiting for that! By the time we reached home it was 12 in the night, and we were totyally exhausted! I just went and hit the bed. Monday - the day started very late..at 12!!! Got ready and went to Orlando Fashion Square Mall. We did some shopping as it was Raja\'s bday on 6th. Got his nice boots. Later in the night we had the cake cutting and champagne opening! And with that our long weekend came to an end! And I m here jotting down all those memories which are still very clear and which want me to go back to those amazing places.  That\'s it guys.  I will upload the snaps very soon.  Luv Sos/Srinivas/Srinu  '),(10778,1223561,'28,April,2003','2003-04-28 00:00:00','    The Legend of Bhagat Singh ---------------------------------------------- Watched: 26th April 2003, DVD at home.  \"History is like the Dark house on the other side of the river. You might not be able to travel across the river to see what\'s inside it. All you can do is wait for someone to come from the house and tell you what the house is like\".                                             ----- Self, inspired from the God of Small Things, by Arundhati Roy  I somehow couldn\'t help remembering GOST by Arundhati Roy, and Ammu\'s brother Chacko who had his ideas about history, when I watched the movie The Legend of Bhagat Singh. A RajKumar Santoshi film, an Ajay Devgan Starrer, and an A.R.Rahman\'s film for starters. (Just in case you are confused which Bhagat Singh movie I\'m talking about, \'t was one of the five released then).  It was a good movie. And I usually say that to all movies that I put three hours of time to watch. But I\'m also saying this because I am forced to think, at the end of the movie, about myself, our nation and history on the whole. Have we betrayed ourselves, our nation, and our history? The movie concludes with a similar remark.  The fortunate or unfortunate thing about history is that, it\'s closed from us. Like the house on the other side of the river. We never know what murmurs and rumors and murders are going on in there. What survives, comes out to tell a story. And we are more or less crippled to judge the truth in it.  I do not know if someone from Bhagat Singh\'s time survived to tell us his story. But there are facts. There are happenings. And every story teller and historian will connect the facts using his logic, reason and imagination. It\'s his job, he cannot escape it. And it\'s upto us to buy it.  Was Gandhiji as weak as is portrayed in the movie? Did he feel insecure by Bhagat\'s militant patriotism? If we had opted the violent way to end the slavery, would we have achieved our freedom long long back? Would lesser people have lost lives that way? Would we have been a different country then, so that we wouldn\'t have to depend on making sure of our alien accent and false funds to look for greener pastures? The dam was broken.  At one point in the movie, Gandhiji takes back his non-co-operation movement because of the incident in Chouri Choura. And says later I think, that the nation isn\'t ready for freedom yet. I couldn\'t help wondering if we are free yet. Bhagat is angered at this, and decides to go on his own way to fight for freedom. The line between defense and  violence is a very thin one. And more often than not, they are used interchangeably. A Lion that attacks you, has to be killed. No body has tamed a lion, without ever killing one before. That was Bhagat Singh\'s logic. May be Gandhiji worked with the logic that we are dealing with humans all the time and not lions. Can history tell us who was right? I donot know.  We depend on foreign channels, crave for foreign currency, gladly give up the citizenship, yet complain about the system, the corruption, and the insincerity of our people. And yet proclaim, \"I hate hypocrites\".     One point that I liked in the movie was that it asks you to think on your own. It asks you not to be passive. It asks you to go ahead with your ideals. It tells you that you can live by not eating for 50 days (I felt ashamed of myself when I ate a Pizza today afternoon). It tells you to live for ideals. It tells you to die for your ideals. It teaches you to endure everything for your ideals. Do we need to be a Bhagat Singh to do that? Need not be. But if all our freedom fighters, whether it was the militant one or the non-violent one, could endure so much for their country and countrymen, can\'t all of us rise against the petty cravings, to see our country grow and grow with it? That my friend is the real difference between America and India. Americans have stayed in the country for all the years after their independence to build a country they call their own. Let movies based on history be an inspiration for all of us to do just that.   '),(10779,1223561,'27,April,2003','2003-04-27 00:00:00','    Has anybody Read \"Red Earth and Pouring Rain\"?   '),(10780,1223561,'07,June,2003','2003-06-07 00:00:00','    Finding Nemo-Watched on 6th June 2003, Regal Cinemas  Welcome to Walt Disney\'s World. A world where the soul and the spirit exist not just in human bodies, but in any creature that you can imagine. Where toys and fish have lot more to teach us, about the business of living, love and friendship. Where you\'d find a lot more adults than kids, watching, enjoying and may be even learning from the movie. Welcome to Walt Disney\'s World.  \"Finding Nemo\" takes you through a journey, in the deep currents of the Ocean, to the deep insecurities of an over protective father, the coming of age of an adventurous son, the undying spirit of \"escape-to-freedom\", along with a sharky message \"Fish are Friends, not food\".   This animation movie by Pixar Animation studios was an absolute treat to watch. The depths and the colors of the ocean, the colors of the corals and reefs take you to a world you can only dream of under an ocean. And it\'s a movie that\'s funny. Not just the kids way of funny, cos, there were lot more giggles from adolescents and adults (or do they laugh louder) and by the time you came out of the hall, you would keep hearing \"Wasn\'t that an awesome movie\" for quite a long time.  So what\'s the awe about? It\'s a relatively simple story, about a clown fish father(who is not funny), Marlon, whose only son Nemo has a small fin. Marlon is over concerned about the safety of his son (\"The ocean is not a safe place\"), and is reluctant to send him even to the school (An ocean school). But on his first day to school, Nemo eager to prove to his dad that he can, take care of himself, goes that extra league, and is captured by a diver, and ends up in a dentist\'s aquarium. And if he doesn\'t find his route to freedom, he might as well end up being a dead fish in the hands of the dentist\'s niece with ugly teeth (I forgot her name).  Mean while, the horrified father takes on the fearsome ocean carrying the hope of rescuing his son, from... well, he doesn\'t know what or where. But with the aid (and delay) from a forgetful friend he makes (or was it fate!!!), Dorry, who tells him about the boat and the address of the dentist (Dorry can read the English Language, but she occasionally forgets she can), he embarks on a journey to be Finding Nemo.  While swimming on his way, he befriends sharks who want to change the shark\'s image among the fish (Fish are Friends, not food), but a twist of fate, compels him to escape them. But danger follows him and Dorry, when they get caught up jelly fish. The sea turtles help them out and help them to find their way to Sydney, where Nemo is.  In a parallel turn, Nemo knows that his \"cowardly\" father is fighting a whole ocean to find him, and with his accomplices, bravely turns the aquarium to a stinking pond, so that he can escape when the aquarium is being cleaned. The idea fails, and he feigns \"dead\" to escape through the toilet. Marlon ends up in the dentist\'s room at that point with the help from a pelican, but mistakes his son to be dead.  Whether Nemo and his friends in the aquarium finally escape and whether he can reunite with is father forms the climax of the movie.  The movie was like in fact like the ocean. I look back at it, and it feels great to have watched it. Like the waves in the ocean, we laughed at Dorry, and the pelicans, we prayed for Nemo and cried with Marlon. But I have always wondered at the technology that could breathe so much of life into these creatures that didn\'t exist. May be God feels the same way looking at us.   '),(10781,1223561,'19,July,2003','2003-07-19 00:00:00','    Dear guys, I got all your wishes... I donot want to say thank you for your wishes, cos I guess I have a right on them...:) I\'m 22 from today... As always it doesn\'t feel anything different from yesterday.. but there\'s a whole year of difference in what I should tell my age is to people...:) And then ...baam...I realised that celebrating birthdays, was just ... it\'s a one day thing that changes your age... but then growth is continuous... happens everyday every moment... Ok, we realise that only at few moments in life... and probably birthdays are the times when we realise it the most... atleast for me, I guess... So what did I do for my b\'day...? I had asked people not to go overboard to spend the b\'day... I mean... it\'s not that I didn\'t want a celebration or anything like that... but it\'s just that ... in the whole process of doing something different and spmething special, I thought that may be we tend to lose that \"doing-it-for-you\" thing somewhere and end up with \"i-did-that-for-your-birthday\" thing... not tht anybody would end up that way, but I wanted it to be a very simple one... no hassles for anyone, just a time to be together with people who have known you... I do not know if people understood or misunderstood what I meant, but may be someone must have been hurt that I asked then not to go overboard... so i refrained from making any strong statements further... On the evening of July 16th US time, I came back home at 5 \'o clock... slept for a little while, like I always do... Woke up at I think some time around 9, cooked aloo fry (it didn\'t turn out too good)... I spoke little qith anybody... and fakir and vinay (my roommates) were coming into and going out of the house ... I don\'t usualy spend anytime at the comp ion the house unles for wirk... and though I thought I would go to college in the night nand come back at 12\'o clock, I thought I must enjoy my b\'day... I didn\'t go to college, instead called my sister... she was asleep, but I just had enough time to get wished up by her... it was already my b\'day in India... Followed it up with a call to mom and dad... the first... nope second time I was away from them for my b\'day...:) Well... I followed it up with an hour and a half completely spending the time in front of the comp, listening to all the songs that I hadn\'t listened to for quite sometime... and ... I just continued to listen... I sang out loud (the door closed, ofcourse), danced a little, conscious that I was approaching of my b\'day, but oblivious of my roommates coming in and out of my room, \"i-dunno-thinking-what-about-me\"... At about 11:57, when \"shakalaka baby\" was playing, fakir barged into my room, and said \"Chilaka, come\"... I laughed out loud, and moved into our living room... and there it was, a beautiful chocolate cake, with two candles each shaped a 2 lighted... and there they were, fakir, anusha, vinay, pande, anitha, aparna and om....the lights went out, the candles were lit, and I blew the candles amidst the \"happy b\'day to youuu...\" Gave everybody a share of the cake amidst the flashes of fotos.... got a phone call from amar wishing me b\'day, got wishes from shankar, raj, hari and kalyan, who dropped in to say \"Happy B\'day\"...Got phone call from Anant and Mrs.Anant wishing me for my b\'day, and a phone call from roopa wishing me b\'day... and then we sat, I thanked everybpdy for doing it all, and decided that we will go to Woodlands, n Indian restaurant the next day... Everybody left then, I went back to the comp, receiving messages from slash, veedee, chatting with balu and ajay, listening to songs all the while... I took leave of everybody, at about 4\'o clock in the morning and slept... The next day I got up at around 11:30 am... I got up and started to get ready... I was happy that it was like any other day... I brushed, took a head bath (as mom said), and was going to wear my clothes, when Fakir said \"Wear these\"... And he tossed three T-shirts, and a Jeans... I remembered that I had told my mom that I\'ll buy new clothes for myself... And I didm\'t know what to say to Fakir... the only thing that could come out of my mouth was \"Thanks...\" Anusha was there too.. and I knew they had it all planned it out since yesterday... I had to rush to college... and I boarded the bus at 1:30 pm., but while that I had forgoten my cell and camera... I had a teleconference at 2:30 in the lab, and I didn\'t want to be late... I went into the lab., and got to my desk, when Dr.Jack noticed me and said, \"Hey Ravi\'s Here, Happy B\'day\"... The rest of the gang too didn\'t waste time in wishing me my b\'day, but then we got down to business, and the rest of the aternoon was uneventful... I just made sure that everybody who needs to know the directions to woodlands knew it... I called up Shankar, told Dr.Balaji ( a new member in the Datawarehouse team, replacing Dr.Amr), called Lisa -the ex CATSS worker- she wasn\'t feeling well though, told smitha, fakir, vinay, anusha, aparna and her roommates, that we will start at around 7 \'o clock... Mean while I was also receiving mails wishin me b\'day from Malaiappan (my PS mate), slash, mala, and others... At 6\'o clock, we watched our favourite tv comedy series \"That 70\'s show\", and started for home... Fakir called Anurag if we can start a bit late, cos by the time Anant Bhaiyya would be back, it might get delayd a bit... Anurag spoke to him, and we started home... I went back home, laid on the couch as is my habit everyday... Fakir and Vinay came in and went out, and it was past 7, past 7:30 but no one said \"lets start\" yet... I knew something was going on, but I used the time to not be disturbed from my nap... Anurag called a couple of times asking me to ask fakir to call him whenever we are ready to go... I noticed fakir in house at 8:00 pm, and told him about our departure... He said he spoke to anurag, and said he\'s going to bhaiyya\'s place, and will call me when we are ready... I knew... and he knew I knew... and he knew I knew that he knew... and so on.. I got a call in two minutes, and when I went over to bhaiyya\'s place, opened the door, there it was, a cake with lotsa candles, with \"Happy B\'day chilaka\" on it, with everybody who was suposed to out that night, gathered... the room was filled with balloons, and colored paper (forgot what it\'s called)... and everybody crying out Happy B\'day to you... I didn\'t have any wordsto say, I was overwhelmed, I didn\'t even notice the name on the cake (I thought it was chandra, not chilaka- I knew it only after wards)... I cut the cake again, gave the pieces to all those who missed it the day before, the first piece to Mrs.Anant, bhabhi for us... and then she also prepared ragda paati, chai for all of us... the only thing I could say all the while was \"Thank yu\" al the time... I was feeling very oiverwhelmed, because Ididn\'t know if I had deserved al that.. But I had decided not to judge... I decided to accept it the way it was, without questioning anything... It was about 8:45 when we started to think about \"what next...\" Woodlands..? they would close by 10:00 pm we thought, we called them up , but they said they would take orders till 15 minutes before closing that\'s till 9:15... So we decided we \'d make a dash for it at Woodlands... Since Bhaiyya wasn\'t there, Bhabhi wouldn\'t come, so wouldn\'t anusha and vinay and so wouldn\'t fakir,... So we decided to go another day when bhaiyya and sai (my roommate who was transferring to Houston) also would be there... we started out and reached woodlans at about 9:20, only to know that they can\'t serve anything but dosa at that time... After a little deliberation, we decided to stick out at woodlands itself and have dosas... all of us - roopa, shankar, madhavi, anitha, ashwini, dr.balaji, pande, om, me apara, smitha had different varieties of dosas - from spring dosas to masala dosas tp paper masalas to mysore masalas... It was served fast,m and it was good... But Dr.Balaji had to leave ear;y, and so he left us, but not before he PAID the bill, to our horror.. We objected to it, but he just said since we were students and he was H-1., he did it... We tried to ask the cashierto accept my card, but it was too late... I owe Dr.Balaji my whole b\'day treat now... And he left for Gainesville... I made a mental note to pay back Dr.Balaji the next week he would come, and we finished our Dosa dinner, had a couple of snaps, and came out and decided to go to Baskin robbins for Icecreams... On the way back, I caled Pradeep - he had called me earlier but all I got was the voicemail earlier...on our wau to Baskin robbins, I spoke to Pradeep, he wished me b\'day wishes and we spoke about other things as we always do... Smitha and Anitha told me I was too silent on the way.. may be I was contemplating about all the people i missed... But it was a great day too... We had our share of icecreams but unluckily I didn\'t have cash, so shankar helped me out - I had margarita ice wgich was slightly sour and very filling, I couldn\'t have all of it... It was 11:30 when I came back home... In half an hour I hit the bed, with the memory of a day that very soon became yesterday... (and now it;s day before)... And, my age is changing everyday... guess I\'m growing everyday too...but I might only realise it next year on July 17th again... till then... let the time, age and maturity, come to me... I\'ll post the snaps soon... Just be a little patient... and everything else is fine... Ravi   '),(10782,1223561,'12,November,2003','2003-11-12 00:00:00','    Kaalatheetha Vyakthulu What would you think of Indira, the \"Nayika\" of the novel \"Kaalatheetha Vyakthulu\" (Timeless People is too literal a translation ?) As I started out, just like all other characters in the novel, I found Indira a mean, manipulative and selfish woman. I had no problems with selfish. I had problems with her manipulativeness. So did Prakasham, the lover scorned, Krishna Murthy - Prakasham\'s friend, Kalyani - the sensitive victiom of Indira\'s manipulations feel about her. But Indira\'s manipulations arise from her dreams - her dreams to stand up, her dreams to live a good life, her ambition for her \"rights\" - and she\'s not bogged down by weaknesses. She refuses to cry at any stage, and if you are a cry baby, the whole world will conspire to make you cry. Is it the virtue of selfishness that Indira encompasses ? Or is she fighting to defeat the defeatist attitudes of Prakasham-like people? Or inspiring rich vagabounds like Krishna moorthy to find a direction? Or indirectly teach weaklings like Vaidehi to stand up and fight against the world for what you beleive is yours? for what you beleive must be yours ? She does all that. And much more. Kaalatheetha Vyakthulu, by Dr.P. Sreedevi, is an amazing attempt towards Objectivism (\"virtue of selfishness\"???). Written in the Post independence era, the novel talks of freedom - freedom to dream, freedom to do, freedom to acheive - through the character of Indira, a bold woman, who doesn\'t mind throwing the bricks that she doesn\'t want while buiding her beautiful house. Kalyani, a victim of destiny, losing her father, Prakasham, a place to live, et al. For most of the time, she\'s dependent emotionally on her father, and friends, until Indira\'s and Prakasham\'s betrayal opens up her eyes to the strength and belief she has in herself. Prakasham gives us an illusion as the hero of the novel to start with. But some one who can\'t decide what his preferences are, who his love is, and what he wants to do, can only be a spineless parasite.  In a way, Kaalatheetha vyakthulu is the discovery of Indira\'s character through the changing times and circumstances and of course, people. Through love, and war amongst personalities and consciences of people, Indira emerges as the heroine, who defines her life as she wants to live it. And so may be epitomising the heart of the Timeless People.   '),(10783,1223561,'26,March,2004','2004-03-26 00:00:00','    The Art of Letting Go Dr. Alan Zimmerman\'s Comments:  Sooner or later, everyone you know will disappoint you in some way. They\'ll say something or fail to say something that will hurt you. And they\'ll do something or fail to do something that will anger you. It\'s inevitable. > Unfortunately, you make things worse when you stew over someone\'s words and deeds. When you dwell on a rude remark or an insensitive action made by another person, you\'re headed for deeper problems.  In fact, the more you dwell on these things, the more bitter you\'ll get. You\'ll find your joy, peace and happiness slipping away. And you\'ll find your productivity slowing down as you spend more and more time thinking about the slight or telling others about it. Eventually, if you don\'t stop doing it you\'ll even get sick. So what should you do the next time someone betrays you? TAKE RESPONSIBILITY FOR YOUR FEELINGS. Even though the other person may be at fault, even though the other person wronged you, you are still responsible for your own feelings.  In other words, other people do not \"cause\" your feelings. You choose them. For example, two different people could be told that their suggestions made at the staff meeting were \"stupid and idiotic.\" One person may \"choose\" to feel so hurt that he never speaks up at any other meeting again. The other person may \"choose\" to feel sorry for the critic, sorry that the critic couldn\'t see the wisdom and necessity of her suggestions. As long as you blame other people for your feelings, as long as you believe other people caused your feelings, you\'re stuck. You\'re a helpless victim. But if you recognize the fact that you choose your feelings and you are responsible for your feelings, there\'s hope. You can take some time to think about your feelings. And you can decide what is the best thing to say or do. Then, you\'ve got to learn to WALK AWAY FROM DISAPPOINTMENT. It\'s difficult to do, but it\'s possible. The famous 19th century Scottish historian, Thomas Carlyle, proved that. After working on his multi-volume set of books on \"The French Revolution\" for six years, Carlyle completed the manuscript and took volume one to his friend John Stuart Mill. He asked Mill to read it. Five days later, Mill\'s maid accidentally threw the manuscript into the fire. In agony, Mill went to Carlyle\'s house to tell him that his work had been destroyed. Carlyle did not flinch. With a smile, he said, \"That\'s all right, Mill. These things happen. It is a part of life. I will start over. I can remember most of it, I am sure. Don\'t worry. It\'s all here in my mind. Go,my friend! Do not feel bad.\"  As Mill left, Carlyle watched him from the window. Carlyle turned to his wife and said, \"I did not want him to see how crushed I am by this misfortune.\" And with a heavy sigh, he added, \"Well the manuscript is gone, so I had better start writing again.\"  Carlyle finally completed the work, which ranks as one of the great classics of all time. He had learned to walk away from his disappointment. After all, what could Carlyle have done about his burnt manuscript? Nothing. Nothing would have resurrected the manuscript. All Carlyle could do was to get bitter or get started. And what can you do about anything once it is over? Not much. You can try to correct it if it is possible, or you can walk away from it if it isn\'t. Those are your only two choices. Sometimes you\'ve just got to shake it off and step up. It\'s like the farmer who had an old mule who fell into a deep dry well. As he assessed the situation, he knew it would be difficult, if not impossible, to lift the heavy mule out of the deep well. So the farmer decided to bury the mule in the well. After all, the mule was old and the well was dry, so he could solve two problems at once. He could put the old mule out of his misery and have his well filled. The farmer asked his neighbours to help him with the shovelling. To work they went. As they threw shovel-full of dirt after shovel-full of dirt on the mule\'s back, the mule became frightened. Then all of a sudden an idea came to the mule. Each time they would throw a shovel-full of dirt on his back, he would shake it off and step up. Shovel-full after shovel-full, the mule would shake it off and step up. In not too long a time, the exhausted and dirty mule stepped over the top of  well and through the crowd. That\'s the same approach we all need to take. We need to shake it off and step up. Finally, you need to FORGIVE. It\'s difficult, especially when the other person doesn\'t deserve your forgiveness or doesn\'t even seek it. It\'s difficult when the other person is clearly in the wrong. Part of the difficulty comes from a common misunderstanding of forgiveness. Forgiveness doesn\'t mean that the other person\'s behaviour is okay. And forgiveness doesn\'t mean that the other person is off the hook. He\'s still responsible for his misbehaviour. Forgiveness is about letting yourself off the emotional hook. It\'s about releasing your negative emotions, attitudes, and behaviours. It\'s about letting go of the past so you can go forward to the future. Everyone in your life, everyone on and off the job is going to disappoint you. If you know how to respond to those situations, you\'ll be way ahead of most people. You\'ll be able to live above and beyond your circumstances. Action: Identify two people that have disappointed, hurt, or angered you. If possible, select two people towards whom you still have some bitterness. Then ask yourself, \"How does my bitterness serve me? Am I happier holding on to it? Do I sleep better? Is my life richer, fuller, and better because of my bitterness?\" If you find that your bitterness is hurting you, ake a decision. Actually decide to let it go. Walk away from the disappointment -- which means you no longer dwell on it or talk about it. Period!   '),(10784,1223561,'27,May,2004','2004-05-27 00:00:00','    May 28th 2004, 1:32 AM Feel like doing something creative... But looks like I\'m bgged down... Writer\'s Block...?   '),(10785,1223561,'02,June,2004','2004-06-02 00:00:00','    11:56 pm, June 2nd, 2004, Meenaxi... Have you ever wondered if existence is limited to \"one\" individual or a single \"individuality\"? What if we could lead more than one lives simultaneously, in different bodies, different circumstances, different minds and hearts, but the same soul...? If we cannot do it in this world of \"real\", is it possible in the world of surreal...?  And what would happen if one character from a book you are writing comes alive in front of you, and demands that it\'s story be written it\'s way...!!! What would you do...? Watch Meenaxi for answers...!!! If you would ask me, I wouldn\'t answer any of your questions on Meenaxi. Rather I would ask you to watch the movie itself. It might be difficult to grasp what the movie is about, what the director wants to say, and what\'s there to see. And it\'s like the painting of an artist, than the movie from a director (I guress this statement is only trivial). But yes, it is an experience to watch the movie and interpret it the way you want to. And there probably lies the mark of a director. He opens you up to show you what you want to see.   '),(10786,1223561,'30,May,2004','2004-05-30 00:00:00','    May 31st, 1:28 AM Watched Pulp Fiction... It was pulp, amorphos, without form... Didn\'t understand the movie... But liked it... I had a fantasy about making a random movie... I think this is the closest I got to a deliberate random movie... But may be there;s something I don\'t see... Do I have to look deeper to understand this movie...? May be...!!!   '),(10787,1223561,'29,June,2004','2004-06-29 00:00:00','    Lakshya is about finding your Objective... About a guy living in your door, inside you, who\'s confused about what the whole big life is about, and what his goal is. It talks about finding one\'s feet in the ground they are in. It tells us that anyone can define his \"lakshya\" (thats the word). It can be the lazy, confused, happy-go-lucky Karan who lands in the army more because of his ego than because he has a dream of serving his country. A guy with no ambition, direction and determination, who is happy to drift along till his father and girlfriend question and challenge his ego. And within the army he faces obstacles to be tackled, peaks to be conquered and enemies to vanquish. In contrast to the witty urban(e) comedy DCH, Lakshya is a serious war drama, but it\'s more a journey of self realisation. I liked the movie hazzaaar, but I would have liked if Akhtar had handled the lazy Karan metamorphosing to hero with a little more panache. I got more than I had asked for, which (fortunately or unfortunately), ended with my only asking more of it. Not that it left me complaining, but it left me wanting more... may be just a little more. Hope I do get my goal (and gal) the way Karan does...   '),(10788,1223561,'07,July,2004','2004-07-07 00:00:00','    urlLink    trying to photoblog&nbsp; urlLink   '),(10789,1223561,'07,July,2004','2004-07-07 00:00:00','    It\'s been a very stormy week here in Florida. Hot afternoons, humid middays, and rains in the evenings. And not just showers these were, these were the siblings of the storms, escorted by lightnings and thunders, threatening to steal any plans of a (American) Independence day weekend trip towards the west coast of Florida. But they weren\'t strong enough to bend the wills of the two first time drivers, who had just won their driving licence from the DL office in the Winter Park Office of Florida (lets just say FL). As the Friday drew closer, Srinivas and Ravi watched the skies, laying foundations to their will to get on with it, preparing themselves mentally to rent the car, and drive it, somewhere, anywhere.  It had been about two weeks since the four of them (Raja and Sandeep in addiiton to the other two) had received their driver\'s license. An acknowledgement of their driving sense. May be they were afraid of going out of touch, may be theirs was a juvenile enthusiasm to prove to themselves and to the world that they could drive long distances, safely, that they had decided to go on a long trip. It was certainly a delicate time, when you think that your acquired skill is good enough for the big thing, and you fight to bury your doubts and questions on your abilities as deep as possible in the ground. After all, it\'s confidence you need, to work miracles. So there they were, bubbling and restless like the lion cub, ready for it\'s first hunt of the deer.  For a week now, Srinivas had been hunting for places to visit in around Florida, invading the tourism sites on the internet, and zeroed in on four places: Fort De Soto Beach, Clear Water Beach, Tampa night life, and Caladesi Islands. It was decided that they would travel by ourselves, five of them (Raja, Srinivas, Ravi, Sandeep, Arun) unless someone wanted to come with them. Jay and Om volunteered to come with them, for a day, so that someone experienced, could accompany.  As friday approached, the four licensed friends started their discussions about their turns to drive. They were under the impression that only two drivers at most can be added on a car rental contract. Which two was their question. Sandeep was the first to sacrifice himself from the race, and Srinivas was the first to declare uphold his authority on the drivers seat. And that left Raja and Ravi to compete for the remaining position. Lady luck smiled on Ravi who won the \"toss\" thrice out of three times, to book his slot behind the steering wheel.  Before they were framed into Friday, the would be drivers made sure they had their homework, groundwork, carwork and drivework done, thanks to their wonderful neighbors Fakir (for his car) and Anousha (for her time). They needed that for confidence, they need that to atleast hope that the trip they were undertaking was going to be safe to say the least.  Friday had Om coming to Orlando, and shortly after him the rains came. The rains came not with ferocity, but with persistance, and they fell constantly, consistently. Gradually, the visibility turned low, the traffic became slow, the wind started to blow, and doubts entered their minds about all that would be at stake. They realised it was a bold step, and before they could think it was too bold of them to start off, they started off to the Enterprise car rental on Semoran, Srinivas, Ravi and Raja in Om\'s car, with Om driving. The rains gave them some taste of what could lay ahead. Road side pools, fog on the driving glass, lower speeds, spashing waters - driving in rain was a challenge on it\'s own. They were advised to either stay away from rain if possible, or excercise utmost caution.  Finally, the Enterprise was graced with the arrival of the soon to be Driver-heroes. It was still raining but the wind seemed to have come to a standstill. With great antipication forthe beautiful trip days ahead, Srinivas told the lady at Enterprise, \"We had reserved a car for Friday evening at 5\'o clock\". He also wanted to tell her \"Give it to us right now so that we can grab it for a drive\", but his brain subdued this statement from his vocal chords into a mere thought in his cerebullum, that excited a few million neurons. \"Yeah Sure, May I see your Drivers License Please\", came from the lady which actually catalysed the whole subduing-the-speech-to-thought process in Srinivas.  \"I\'m sorry, but this is a temporary license, and we are not allowed to rent cars on this\". \"What!!\" the drio spoke almost instantly, but each with a different lag from the time time the lady\'s sentence ended. A moment later when he realised that she had actually, really, said that, Srinivas retorted, \"But I am sure this is valid, because my friend did rent a car from enterprise, and he too just had the temporary Drivers license ( TDL)\". Ravi moved out of the room in a slight disgust, while Raja and Srinivas proceeded to negotiate with the lady. Ravi came back, seemingly resolute, and the drio decided to fight back. \"I\'m sure of this, my friend had rented a car on the TDL\", \"Do you mean to tell us that this is not valid, when it says that it in place of the permanent photo-IDied DL\",\"We would like to talk to somebody about this, just to clarify, Please\". May be Maria sensed the barrage that had been broken by her denial, may be she herself wasn\'t confident of what she had said, or may be she was trying to do her job, but she called someone again. And this time, she got it right, she got the drio smiling.  ******************************************************************  It was still raining outside, when Maria from Enterprise waded out of the backdoor of her office, keys in one hand, papers in the other, followed by our driver-trio (shall we call them drio- just to shorten things up). They walked towards a magestic car, radiating like a black body, with it\'s black body. As Ravi moved of the door and took his first look at the car, he felt a tinge of warmth in the slightly cold rainy weather. \"It\'s the excitement\", he heard his mind ring inside. His heart started to beat a little faster, as he involuntarily hastened his steps towards the car. He was the one who was the farthest from the car, and he was unlucky to have lost the race towards the car to Srinivas and Raja. Maria showed the drio the papers and the keys,  (Updated 07/10/04)  Om checked the car, and Srinivas assumed his authority at the driving wheel. Raja sat next to him, Ravi and Om behind. Om had got into the Gallant Mitsubishi so as to get to his car on car (it was raining so he would get wet on foot). As srinivas started to reverse, He could hear three voices screaming \"right\", screaming \"left\", scremaning \"stop\", screaming \"go\", and a host of other screams that he couldn\'t exactly make out. He also screamed something in between probably asking only one to speak, but since he was screaming too, all their screams moved randomly in the air, got mixed, mingled, scrambled and jumbled, and all he could hear was \"rgoighttelfakelefstop\", and the car ambled through the lane, and out of enterprise.  (Srinivas\' Version)&nbsp; Lemme catch up with Ravi\'s Blog!Once Maria, the lady at the Enterprise office said \" Ok, I\'ll go ahead and give you a car\", at that moment, I was on cloud 9. I had lost all hope when she initially said we couldnt get the car. I, then, told her that I thought she was the meanest person on earth when she said she cannot give us a car! She said \"Noo..please dont say that!\" I had a sense of achievement, of getting the car! Once the paper work was done, Ravi, Shekar and I \"ran\" to the car, a Mitsubishi Gallant 2004 ! You could say that was my dream come true! Never did I imagine I would driving such an amazing car! Once the paper work was done, I was selfish enough to decalre that I would be driving the car that evening! And then started our first drive home! As soon as I adjusted myself in the seat, and got all ready to leave, a chill feeling went down my spine- a sense of excitement, a sense of fear too! \'What if something happens to the car?\' Then suddenly I recollected Nishanth\'s words \" There\'s always a first time for everything. And the earlier it is the better!\" Also we had the entire insurance coverage. That\'s it. That thought made me confident enough to drive the entire trip! It was still raining heavily. I was initially a little worried about the clearance. It just took a couple of minutes for me to get the feel of the car and the spacing needed. In that rain, splashing waters on the road, I drove back to Khayyam, without a single fault. That was a real big accomplishment. As soon as we reached home, Jai and Pande were there welcoming us! Pande said \" Now that you have driven in such a weather,I m damn sure you can drive anywhere in any condition\". Well those words were also a big boost to my morale, which already was high up in the sky! I was scared I might get over confident now! So I decided to drive the car all by myself the entire evening. Well, yeah, I m sure Ravi and Shekar woould have definitely thought I was the meanest person on earth at that point!, but I was so overwhelmed by the thought of driving the car, nothing else bothered me!! Once we reached home, it was decided that Jai woudl treat us at Ahmed\'s , located on OBT. Ravi and I insisted on visiting the Shiv Temple as it was Guru Purnima, our birthday according to the Hindu Calendar that day, and also our first experience with the car! And I played the driver. With Jai as my guide, I felt I did a good job driving to OBT. We took Expressway 408 while going there. It was in a way my very first experiences on a Toll Road. I was kinda nervous initially, my hands started sweating, though I kept denying! But I had a sexy time. It was really cool driving at more than 60 mph all by myself! By the time we reached Ahmed\'s I was 100% confident of carrying the entire trip on my shoulder\'s even if Ravi and Shekar wouldnt have been there!After dinner at Ahmed\'s I decided, it wouldnt be fair on my part to keep driving and so gave the keys to Shekar. He pounced at them like a hungry cat! We reached home in about 40 minutes and decided that we must leave the next day by 7:30 in the morning. Om and Jai were more than happy to join us on the first day. They said they would be coming back by Saturday night. That night by the time we slept it was pretty late, dont exactly remember the time though. We tried burning some CD\'s. We could burn just 2 of them. One Tamil and the other Hindi. We were very sleepy and decided to call it quits for that day.Saturday- The day the voyage began! Three brave drivers were getting ready for their first ever road trip! Ravi was the first one to hit the road that day. Jai was there in our car as the navigator while Arun and myself were the passengers. We went to the gas station , got ice, cameras, gas and chips for the journey and we were all set to roll!And Ravi did an awesome job driving on I 4!!! I really didnt believe he could do it that well. He was really really good. Well, he had his own moments of \"fear\" which made us gasp for our life for a moment! He was pretty scared looking at a huge semi truck merging on to the freeway. He thought it was gonna hit us straight! And then without looking behind he moved into the left lane, and there came a HONK! Before we realised, there was an SUV going right next to us. Anyways once this incident got over, there was no looking back. He drove for a solid 2 1/2 hrs non stop! Even Jai was pretty amazed at Ravi\'s driving skills! The drive to Tampa Bay was great. Arun and I got a chance to see the backwaters of the Gulf of Mexico while travelling on the bridges connecting Tampa. In all our desperation we tried clicking snaps of the Downtown, the bridges, the backwaters, but nothing worked out as we were travelling so fast, nealr 75 MPH, couldnt see anything in the camera properly. Throughout the journey, the Tamil songs we copied , were played, and we ddint get bored even once. The collection was too good!After missing out the exit to Fort De Soto, and after some inquiries regarding the direction to Fort De Soto, we reached our destination at about 11:30 - 12:00 Noon. We first went to the North Beach. We could first see a huge parking lot with lots of greenery around. All of us decided to get into our swimming shorts, and jut then I realised I left my bag at Orlando! I hadn\'t brought it with me..Damn! I was wondering what I would wear for 2 days, that too we were going to beaches. We definitely will get wet. Then Om said I can share his clothes, he had a huge stock of them ready in his car! And so all of us went to the shore, passing through the woods. The first impresion on seeing the shore was \"Beautiful\" and \"Bright\". The sands were so soft and white, nobody would believe it was sand. It was more like Flour! Man....and then the waters...very very inviting! All that I remember was, we just plunged into the water and remained there for about 2 hrs. The waters were warm and really soothing.After a 2 1/2 hr journey, with the sun blazing hot right on top of our heads, the waters seemed like a heaven!   '),(10790,1223561,'11,July,2004','2004-07-11 00:00:00','    urlLink    On the ferry to Caldesi from Honeymoon Islands&nbsp; urlLink   '),(10791,1223561,'11,July,2004','2004-07-11 00:00:00','    urlLink    Srinivas Ravi Car&nbsp; urlLink   '),(10792,1223561,'29,June,2004','2004-06-29 00:00:00','    S1   \"It\'s time\", Ramesh said locking his suitcase. I looked at my wrist, but found no watch there. For a moment, fear gripped me, and impulse reactions travelled down my spinal cord, jerking my hands towards my shirt-pockets, and then to my trousers. Within the moment, I realised that I had kept my watch safely locked in my - either the travel bag or the suitcase. With some uncertain certainty, I looked in Ramesh\'s direction, and found the big clock some way behind him, about to strike 5:00 PM. \"It\'s time\" I thought, and as if the lady had heard our thoughts, we heard her, \"Your attention please. No:8888 New Delhi to Kanyakumari, Swarna Bharathi express West is arriving on platform no 7 shortly\". She announced the same sentence verbatim in Hindi, and no sooner did she finish, we heard the distant \"Cooo\" of the train, we felt the ground we were standing on shiver, we saw the people on the platform getting silent momentarily (On hind memory I guess our ears were jarred for that moment), then we saw people getting busy, students, families, businessmen, husbands, wives, policemen, old couples, loud women, distracted men, getting ready with their luggage to get on to the train, getting ready to get on with their lives that seemed to have stalled on this railway platform.  \"The arrival of train,\" Ramesh said, compelling me to abandon my search for the two ends of the train, and listen to him in anticipation of something interesting. \"It creates so much of panic among people. People who have got reservations, people who know the train is going to stop for a while, still get very panicky when they hear the train come in\" said Ramesh sounding amused. \"Well, nobody wants to miss the train, no body wants to miss their seat\" I countered, as the train came to a halting stop. \"It\'s not so much about missing than it\'s about compettion - to be among the first\", Ramesh re-countered, walking slowly towards his compartment. \"The earlier you are in the lesser hassle it is\" I said. We started to move towards S7, which was Ramesh\'s bogie. We had to move towards our right, against all the passengers hurriedly moving to the bogies on the descending side of 7. It wasn\'t as difficult, moving a couple of luggages that Ramesh had, as it was to wade through the opposing passenger flow. For a moment I wondered if the ones moving opposite to us felt the same, and I ended up smiling lightly, with a sense of discovering something as subtle as.... as dew drops on a petal.  We came close to the door of S7, and since we walked slowly to it, the crowd of passengers that usually cling at the door had thinned. Ramesh got into the bogie with his travel bag and I helped him with his suitcase. I got into the compartment too, dragging the suitcase till we reached seat No: 16. \"I wish everybody had carried as little luggage as I do all the time, it makes so many lives easier\", he said slightly panting. There were people around trying to settle their luggage under the sleepers, over the upper berths, towards the window (the waterbottles). Permutations were discussed, combinations were tested, deals had been made, arrangements had been agreed, wars were fought, treaties signed, and peace was restored before all the passengers started to settle. I figured out that the best place for me was the side of the window to ramesh\'s from outside, and I promptly came out.  \"Not too many from our place this time on this train\", I said. Ramesh looked at me as if I was stating something too obvious. I understood the reason behind the look. Soon my train will come, but it\'ll be so in contrast to the one Ramesh is in. While he\'s unlikely to go around the train looking to find people, there would probably be not too many looking out expecting to see him. We have been going to our homes every year, but this time it is different. We were going back, back to our homes, back to where our lives were going to take us. End of schooling is the end of a way of life as we knew it, so this was effectively the last time to say good byes. Somewhere deep down, we were optimistic of meeting each other, bumping into each other, working with each other, working for one another, but deeper down there was just this doubt, this apprehension, this misgiving that may be it\'ll be a long time before we can say \"Hi\" to each other again. Ramesh was least worried about saying goodbyes. And considering the fact that we were cross with each other a couple of days back, he might just have not expected me to come and say Goodbye to him.  I had come to the Xaviers Boarding school in Delhi when I was about to get into the 5th grade. My father, was a wealthy accountant, who was enamoured with distant lands and good education. But the only good education according to him was being an engineer or a doctor or a lawyer. You couldn\'t blame him for what he thought were respectful professions, for these were the three professionals he had&nbsp;felt for the part of his life could make more money than him. And of course, Collectors. But collectors though representative of high intellect, were, according to him mere gloves to the hands of bureaucracy. He hadn\'t seen too much of difference between a politician and a collector, and ever since I had written the right sequence of numbers from 1 to 50 when I was in the first grade in&nbsp;a small&nbsp;school in Hyderabad, he wanted me to be an Engineer. The engineer was one of the most persistent ambitions I had had when I hadn\'t known what it meant or what it entailed. But among all when-I-grow-up-I-want-to-be-an-xxx\'s that I would blabber in my golden days of ignorance and innocense the \"I want to be... an Engineer\" seemed to be a mantra that seemed to please and sober my father in his most delicate moments of thoughts, worries, anxieties, curiosities, misgivings, apprehensions about my future. In fact, for a long time, I was under the impression that the Engineer was responsible, among many other things (like building the dam, creating biggggu buildings), for getting the steam out of the train engine. Somehow, engine-engineer, seemed to just make some weird sense to me, and I had the engine-driver and the engineer synonymously coexist in my mind. But then, I was then too young to remember my past, or think of the future. All I was, was just being awed by the present.&nbsp; And so it was, my dreams already dreamt and re-dreamt in my parents\' eyes. And along with their genes, and love, I had inherited them. Until now. It was after one whole summer of my being in Xavier\'s that Ramesh too found his way to the dormitory I was in. We were just 10+ boys then, and he had been admitted into the school by this then-fighting-now-divorced parents. Ramesh looked to me then like a shy, intelligent, nerdy geek, a thin frame, with thick glasses that escorted his eyes with a vision towards - infinity or nothingness, I didn\'t know then. I did not know about his or his parent\'s history, and I wanted to intimidate him in his early days (just like I was, in my early days). But something in him&nbsp;restrained me from doing so. I don\'t know if it was a feeling of sympathy for the boy-who-saw-into-nothing, or a feeling of awe for the geek-who-could-see-infinity. In fact, whenever he was around in the dorm, after the classes, after the evening cricket matches, after the dinners, I could feel some presence, an overwhelming presence of nothingness. It was as if you felt you were being followed, you looked back, expecting to dismiss your feelings, and eventually finding out that your feelings weren\'t illusions. For the first one week, the only conversations we have had&nbsp;living on adjacent beds were the dialogues of introduction, and occassional invitations from me to him for some outdoor games. Ramesh was polite in his conversations, but never reciprocated. He had carefully given me an impression that he never&nbsp;cared. Within a week, I had enough prejudices on him and I wondered how he would live the rest of his life without any friends.&nbsp; '),(10793,832265,'02,August,2004','2004-08-02 00:00:00','   The  urlLink WWASP  programs have an interesting twist on more conventional forms of  urlLink incarceration  - the role of parents.  So, you\'re a wealthy kidowner, but you can\'t keep the kid from smoking the pot.  You can\'t go to the police because, well, there\'s not much of a crime there and they can\'t give the sort of guidance that unliscensed professionals can.  Isn\'t it worth $80/day to send your problem child to Jamaica for rehabilitation.  It\'s 100% hassle-free parenting.  Tranquility Bay escorts will even  urlLink pick up  your brat unannounced so there\'s no time for awkward explanations or goodbyes.  The good news is that you, the parent, are still in charge.  Afterall this is the eagle way - pay someone to give your child the draconian third-world upbringing you can\'t be bothered to administer yourself.  '),(10794,832265,'02,August,2004','2004-08-02 00:00:00','   Back to  urlLink camp  and still just sandals on my feet.  Quiet stretch mark plus paint splatter vinyl wouldn\'t be so tough if not for the smell.  But I\'ve got run plans and a future without  urlLink fisheye breakfast  wherein all big toes will live lint free.  And so I take my showers  and try not to clog the drain with  urlLink feathers .  ( via  urlLink mefi  )  '),(10795,832265,'02,August,2004','2004-08-02 00:00:00','   The highly annoying color display issue has been resolved.  More importantly, I\'ve seen  urlLink Spirited Away  which knows a thing or two about color, character design and what happens when a Stink God tries to take a bath.  More than  urlLink Princess Mononoke ,  urlLink Hayao Miyazaki\'s  last movie, Spirited Away reminds me of most of the ps2 game  urlLink Ico .  The two share a similar tenderness - a feeling that stops short of seeming precious because it is so subtly evoked by the gestures and expressions of the main character.  Also like Ico, there is no excess of backstory or mythological explanation.  But this doesn\'t matter as the world shown is so complete that it makes intuitive sense, despite the fantastical setting.  We saw the matinee with a full house of little kids - it seemed to have entertained the good chunk of them ... lots of laughs and some crying in the sad parts.  However, there were also those who weren\'t really buying my \'intuitive sense\' argument.  One kid in particular was desparate to figure out a variety of unanswerable questions: \"Why\'s the dragon white!?  Where does that door go?  Why\'s that girl got hair!\"  '),(10796,832265,'02,August,2004','2004-08-02 00:00:00','   I don\'t know how often you find yourself among married folk, but I\'ve always felt the proper response is the gluttonous consumption of seafood followed by emotionally detached frolicking.  Eugene also went to the bathroom ... a lot (not pictured).  urlLink .:more:.  '),(10797,832265,'02,August,2004','2004-08-02 00:00:00','   It was foggy in the city by the bay, but we went to the beach anyway.  Did we ever.   Fortunately, it was more grey than actually cold.  That being said, we marvelled at the psycho-lady (not pictured) who boldly marched into the sea wearing a blue bikini.  By the way, I\'m having some \"non-display of color\" type issues with the individual entry pages ... until a crack team of blog specialists has been able to resolve the problem, uh, imagine colors.  urlLink .:more:.  '),(10798,832265,'02,August,2004','2004-08-02 00:00:00','   The softlaunch of /goldtoe continues.  A links page has been added as well as a picture of a lamp.  I need a soundbyte of a goat eating a can of beans and this deal is done.  In the meantime, you can have a picture of a backpack. urlLink '),(10799,832265,'02,August,2004','2004-08-02 00:00:00','   26 years old is half of 52 - which is to say, I had an excellent birthday.  I started my day with a walk in Diamond Heights wherein I saw that most suburban of fixtures - the playground/stripmall.  If you can\'t get frozen yogurt and takeout chinese immediately after swinging, what\'s the point?   The point is catfish and mashed potatoes at the Lucky Penny.  Not to mention a trek into the foggy Presidio for some serious bowling.  urlLink .:more:.  '),(10800,832265,'02,August,2004','2004-08-02 00:00:00','   \"Very probably, it was not part of the sofa vaudeville of a showoff but, rather, the private exposed achievement of a young man who, at one time or another, might have tried shaving himself left-handed.\"  '),(10801,832265,'02,August,2004','2004-08-02 00:00:00','   All around the country today both little children and adults are carving pumpkins and dressing up in costumes in honor of my last day of work.  It\'s quite the heady experience.  To celebrate, I went to the dentist and had a cavity filled.  urlLink '),(10802,832265,'02,August,2004','2004-08-02 00:00:00','   The DVD of the original Rollerball includes a memory game in which you have to correctly sequence 16 thumbnail scenes from the movie.  Only the brave will manage to slog through and receive the ultimate reward - a 10 second clip from the flick you\'ve just finished watching.  Still, Rollerball certainly represents a more  urlLink well-heeled &nbsp; urlLink version  of the  urlLink 1970\'s &nbsp; urlLink future .  '),(10803,832265,'02,August,2004','2004-08-02 00:00:00','   And now I get up at 8am to ensure a good jump on my second to  urlLink last day  of work.  '),(10804,832265,'02,August,2004','2004-08-02 00:00:00','   It\'s either the after-effects of gluttonous  urlLink Zachary\'s  pizza consumption, or the fact that there\'s no more sun ... but either way I\'m lagging.  Incidentally, there\'s a whole mess of folk, some of them  urlLink Czech &nbsp; urlLink jews ,  urlLink opposed  to Daylight Saving Time.  '),(10805,832265,'02,August,2004','2004-08-02 00:00:00','   Apparently all varieties of animals have been found  urlLink sealed  in stone, iron ore and even chalk.  Mostly frogs are found but every once in a while you stumble upon a live  urlLink pterodactyl .  The common explanation for these occurances is similar to the way Winnie the Pooh ended up stuck in Rabbit\'s hole ... so to speak.  But then again they may be reincarnations of those  urlLink stuck  in one of the outer circles of Buddhist hell.  '),(10806,832265,'02,August,2004','2004-08-02 00:00:00','   The current SF Weekly  urlLink cover story  is about getting to the bottom of the \"Monkey Knife Fight\" streetmeme currently plaguing the Mission. They basically pull their version of Salon\'s  non   urlLink Un-Masking of Deep Throat  - altho\' you don\'t have to buy the eBook.  The weak conclusion - \"It\'s with this realization that we come up with a theory. \'Monkey Knife Fight\' means NOTHING.\"  Sweet.  So will it stop now?  '),(10807,832265,'02,August,2004','2004-08-02 00:00:00','   '),(10808,832265,'02,August,2004','2004-08-02 00:00:00','   \"Although the remark has the ring of memoirs  urlLink written  after the event, it may well have been spoken.  Fateful moments tend to evoke grandeur of speech, especially in French urlLink . \"  urlLink urlLink'),(10809,832265,'02,August,2004','2004-08-02 00:00:00','   Ari Fleischer was asked why the President doesn\'t support a national ballistic fingerprinting system given, you know, certain media-hyped gun violence in the DC region.  His  urlLink response : \"And certainly, in the case of the sniper, the real issue is values. And that\'s what is at stake here. The real issue is values. These are the acts of a depraved killer, who has broken and will continue to break laws. And so the question is not new laws; the question is the actions here represent the values in our society.\" See, it\'s that whole culture of death we\'ve got going on that\'s to blame.  If only we didn\'t have abortion and violent video games this never would have happened (NB: the death penality is not, despite appearances, part of the culture of death).  '),(10810,832265,'02,August,2004','2004-08-02 00:00:00','   And numerology, too, for that matter.  I\'ll  urlLink find  you a  urlLink pattern  in the abstraction and leave you  urlLink convinced  the Pope is the Devil.  Maybe that\'s not too hard.  But if I tell you that the 6th, 7th and 8th  urlLink primorials  are 30030, 510510 and 9699690 ... don\'t you shiver a little bit.  '),(10811,832265,'02,August,2004','2004-08-02 00:00:00','   My tall, salty friend Steve Yuhan has a  urlLink weblog , the discovery of which lead me to find that there are many fine debate-folk in the blogsphere.  This makes sense.  However, I have trouble figuring out why this debate  urlLink webpage   is being hosted by the soap-box racing site,  urlLink National Derby Rallies, Inc .  '),(10812,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink urlLink'),(10813,832265,'02,August,2004','2004-08-02 00:00:00','   This evening\'s  Frontline  is on everyone\'s favorite intersection of pork barrel politics and unscrupulous militarism,  urlLink  missile defense .  I\'m confident the folks at PBS will do their usual bang up job; the episode has already received a ringing endorsement  from the  urlLink National Review . Yes, I\'m sure that by the end I\'ll be shouting at the TV and afraid to go to sleep lest I discover that a military-industrial junta has taken over America.   urlLink Oh &nbsp; urlLink wait .  That being said, my favorite account of missile defense comes from JoAnn Wypijewski\'s article in the December 2001 issue of Harper\'s.   urlLink The article  is quite lengthy, so here\'s a little taste: \"Sometime after I returned from Kwajalein, I came upon a newspaper column by Paul Loeb, recounting a meeting he\'d had with executives at Lockheed\'s Missile & Space Operations in Sunnyvale, California. He quoted one fellow saying, \'Let\'s get real. We all know that if anyone ever attacks America, the bomb is going to be delivered by a suitcase, a car or a truck, or in a boat.... We all know that we\'re lobbying for these programs because they make us money. We don\'t care whether they\'ll ever work, or even be useful. We care that the dollars come our way.\'\"  '),(10814,832265,'02,August,2004','2004-08-02 00:00:00','   Being the  urlLink account  of one  urlLink man\'s  obsession with and struggle against Bill O\'Reilly.  In fun  urlLink hyperfootnote  format.  '),(10815,832265,'02,August,2004','2004-08-02 00:00:00','   There will be no precious. There will be no darling. There will be no sweet stains left drying in dark corners.  There will be no downbeat. There will be no whimper. There will be no teardrop held up and in soft pause.  '),(10816,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink Ansi &nbsp; urlLink Art : The  urlLink Next &nbsp; urlLink Generation  '),(10817,832265,'02,August,2004','2004-08-02 00:00:00','   \"It hovered inches above the horizon, dancing, waiting almost as if it were taking a stoked breath, then swelled in puffing spasms, poking high into the stratosphere, edging out the pale skyrocket vapor trails at either side, the ball going sickly yellow, the shock wave releasing its roar, the entire spectacle catching fire, blazing chaotically, shaming the  urlLink paltry sun .\"    urlLink'),(10818,832265,'02,August,2004','2004-08-02 00:00:00','   The more things  urlLink  change : \"The President is authorized to use the Armed Forces of the United States as he determines to be necessary and appropriate in order to defend the national security of the United States against the continuing threat posed by Iraq.\" the more things stay the  urlLink same : \"Congress approves and supports the determination of the President, as Commander in Chief, to take all necessary measures to repel any armed attack against the forces of the United States and to prevent further aggression.\"  '),(10819,832265,'02,August,2004','2004-08-02 00:00:00','   This year, I\'m thankful I didn\'t go home after Thanksgiving dinner was over - it helps if you can sleep over so as to get a good jump on leftovers the next day.  Special thanks to Karen for the eggplant parmesean recipe; it made a great dinner, lunch and another dinner the subsequent day.  urlLink .:more:.  '),(10820,832265,'02,August,2004','2004-08-02 00:00:00','   \"The side of the dunes that faced the sea and received the monsoon winds rose abruptly, but straggling clumps of scrub grass grew in places where the incline was not as steep.  The man looked back over his shoulder at the village and he could see that the great holes, which grew deeper as they approached the crest of the ridge, extended in several ranks toward the center.  The village, resembling the cross-section of a beehive, lay sprawled over the dunes.  Or rather the dunes lay sprawled over the village.  Either way it was a disturbing and unsettling landscape. urlLink \"  urlLink '),(10821,832265,'02,August,2004','2004-08-02 00:00:00','   Anthony Swofford, a Marine sniper in the Gulf War, has an article in this month\'s Harper\'s.  It\'s similar to, but much better than, his op-ed  urlLink piece  from the NYT last month, and presumably like his upcoming memoir due out this spring.  His take concerns both the terror of war generally and the absurdity of the Gulf War in particular.  So he writes about both the personal effect a Blood for Oil policy has on the soldiers ordered to carry it out: \"We joke about having transferred from the Marine Corps to the Petrol Corps, and while we laugh at our jokes and think we\'re damn funny jarheads, we believe we might soon die, and this is not funny.  We have been deployed to protect oil reserves and the profits and rights of American companies, many of which have direct ties to the White House and oblique financial entanglements with the secretary of defense, Dick Cheney, and the commander in chief, George Bush, and the commander\'s progeny.\" and what happens when you come under Iraqi rocket attack: \"I stand in place and piss my pants again, piss all over and running into my boots, thighs both, knees both, ankles both, bottom of my feet both, clear piss and no underwear, because otherwise chafed crotch because Vaseline only works to mile ten and all wars and battles occur farther than ten miles from all-safe points, and rotten balls if you don\'t remove your underwear at mile ten, and rockets landing red glare and more rockets, hitting everywhere around us, but they haven\'t hit us, so far they have only caused great amounts of terror and oblivion.\" It\'s interesting the response Swofford is provoking in the converative press/blogworld.   urlLink Some  seem only blithely aware of the larger point being made while others are  urlLink embarrassed  for a soldier made afraid by battle. Ah, the marketplace of ideas.  '),(10822,832265,'02,August,2004','2004-08-02 00:00:00','   The first college debate tournament I attended was hosted by Penn.  I thought I had a rough time because my partner kept wanting to run cases about Star Wars and we ended up losing more than we won.  However, I now realize that I got off  urlLink easy .  Being crammed into unlocked common rooms is the norm for tournament housing, and you\'ve got to expect a touch of uneasyness when a horde of dorky debaters takes over a shared college space.  But I can\'t begin to understand how a Penn student finds himself walking around West Philly with a container of motor oil in his hand and mischief in his heart.  Of course, this is the campus where a peace protester  urlLink self-immolated  a few years back, also around the time of the debate tournament ... so it would seem there\'s some kinda rhetoric-related bad vibe going \'round.  '),(10823,832265,'02,August,2004','2004-08-02 00:00:00','   Despite the full moon and the  urlLink light pollution  of a major city, the Leonid meteor shower did not disappoint.   urlLink Conditions  were favorable enough for me to see nearly two dozen Leonids between 2:15 and 2:45am while sitting at the corner of Hill St. and Sanchez here in San Francisco.  Several times I mistook the reflection of headlights off overhead power cables for astronomical objects ... and I have to remember that urban stargazing would be better if I brought along a flashlight to switch off the automatic streetlights.  However, no amount of light or haze could have drowned out the one meteor I almost figured for a plane until I saw its broad, phosphorescent tail dissolve into the night sky.  '),(10824,832265,'02,August,2004','2004-08-02 00:00:00','   The army  urlLink dismissed   nine of its linguists, six of whom were studying Arabic, after discovering they were gay.  Of course, there\'s a shortage of Arabic speakers needed to translate the War of Terrorism into an international best-seller.  But if you\'re going to defeat a fascist regime of religious intolerance you\'re going to need a little of your own.  '),(10825,832265,'02,August,2004','2004-08-02 00:00:00','   Once, Jerry O\'Connell was the fat kid in \"Stand By Me\" and slept with Sarah Michelle Gellar.  Now, he\'s in a movie with a computer generated kangaroo that inadvertently steals money from the mob.   urlLink Really .  '),(10826,832265,'02,August,2004','2004-08-02 00:00:00','   Have you ever noticed that the logo for the pork board is suspiciously similar to that of pharmaceutical company GlaxoSmithKline?'),(10827,832265,'02,August,2004','2004-08-02 00:00:00','   The  urlLink \"Hole in the Wall\"  project provides free internet access to children in India via kiosks across the country.  This, among other things, has lead to the kids using these computers to refer to the hourglass icon as  urlLink Shiva\'s drum .  '),(10828,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink'),(10829,832265,'02,August,2004','2004-08-02 00:00:00','   St. Louis, Missouri - home of  urlLink toasted ravioli ,   possibly  the  urlLink ice cream cone , and my friends Alex and Carrie Chen who were married this past weekend.  Congrats!  urlLink .:more:.  '),(10830,832265,'02,August,2004','2004-08-02 00:00:00','   Thankfully, the USOC  urlLink selected  New York as the US candidate for hosting the 2012 games.  I can\'t imagine the  urlLink nightmare  of trying to go about normal life in San Francisco had the SF/San Jose/Palo Alto/Turlock bid been accepted.  '),(10831,832265,'02,August,2004','2004-08-02 00:00:00','   \" urlLink Still Life  is a program that lets you pan and zoom any number of pictures and export the result as a QuickTime movie or DV stream.\" That reminds me ... I had the strangest  urlLink dream  last night.  '),(10832,832265,'02,August,2004','2004-08-02 00:00:00','   Retro, the New Order 4 CD box set is being released stateside on January 7.  There\'s a breakdown of all the goodies at the official  urlLink website .  More importantly they\'re also serving up a bunch of videos including the 10 minute one for  urlLink Perfect Kiss .  Basically, they all stand around and look really bored while rocking out for Jonathan Demme.  And Gillian tries really hard to look pretty. Oh but there\'s more.  On January 21, 24 Hour Party People is getting  urlLink released  on DVD with 2 commentary tracks and some other business.  It\'s an exciting time if you\'re into that sort of thing.  '),(10833,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink   I got 110 free 4x6 prints through iPhoto by virtue of signing up with .mac.  This is, admittedly, a pretty decent deal ... except I have to use them all by the end of the month.  So ... anyone want doubles of some pictures of my toes shot from various angles?  There\'s  urlLink one  in particular that\'ll scare the bejeezus out of you.   '),(10834,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink This  is ambitious.  Turning a 17th century diary into a blog, complete with annotations and links to explanatory pages for people and places.  '),(10835,832265,'02,August,2004','2004-08-02 00:00:00','   I\'m a big fan of the Simpsons  urlLink episode  where Lisa goes to Washington to compete in an essay contest.  Especially funny are the runner-up essays like the kid who talks about the Recipe for a Free Country.  You know: \"Mix one cup liberty with three teaspoons of justice.  Add one informed electorate.  Baste well with veto power...\"  But apparently, some libertarians didn\'t realize this was meant as a joke \'cause they\'ve come up with their very own  urlLink recipe .  It calls for \'liberating\' a chosen state by importing a horde of anarcho-capitalists and implementing alternative voting methods.  Well, I guess if the Scientologists get to have  urlLink Clearwater, FL  the Ayn Rand Fan Club can have a Dakota or something. But, you know, if your  urlLink FAQ  has to answer the question \"Is the Free State Project some utopian power trip?\" you might want to take a step back.  '),(10836,832265,'02,August,2004','2004-08-02 00:00:00','    Oh, the wonderful world of Apple IIe  urlLink emulation .  And now  urlLink someone  has made an OS X version, bringing the oldest of old to the newest of new.  You can grab yourself a whole mess of ROMs from the  urlLink Asimov  FTP archive or just check out my favorite screenshots. &nbsp;  urlLink .:more:.  '),(10837,832265,'02,August,2004','2004-08-02 00:00:00','   I\'ve been debating whether or not there\'s something cheap about the blog post that goes \"Look, there\'s something new to the site.\"  If it\'s truly fundamental, you\'d hope folks would just notice without having to make a big fuss.  Big fuss or no, there\'re now  urlLink picture archives .  I\'ve been debating whether or not there\'s something cheap about the blog post that goes \"Look, I\'m going ahead and pointing out the silly before you can.\"    '),(10838,832265,'02,August,2004','2004-08-02 00:00:00','   I used  urlLink FogCaster  to see whether or not the view would be clear for the Pyra Holiday Dinner at  urlLink Greens Restaurant .   I found it to be surprisingly accurate from a day in advance.  And you can see from looking at it now that the brief respite we had from winter today will end with a quickness tomorrow.  '),(10839,832265,'02,August,2004','2004-08-02 00:00:00','   The OS X port of Soulseek now has a SourceForge  urlLink page  ... but nothing to download just yet.  There\'s a lot of folks hoping for a nice Christmas present from the fellow working on this deal.  '),(10840,832265,'02,August,2004','2004-08-02 00:00:00','   Looks like I\'m going to have to find myself a new DSL  urlLink provider . Update :  Signed up with Speakeasy.  '),(10841,832265,'02,August,2004','2004-08-02 00:00:00','   FetchYahoo did the trick, liberating my three years of mail from y!.  When combined with the mail spools I\'ve had stored on CD since college, I\'ve now got 6 1/2 years of email archived and properly sorted on my new iMac.  Thanks to nathanr for his mbox-to-applemail converter  urlLink script.  Somewhere in St. Louis (or possibly a New Jersey landfill) there\'s a set of mboxes on 3.5\" floppy that contain another 4 years of emails from my very first ISP, shell-tastic  urlLink CRL .   One day I will be whole.  In the meantime, I\'ve taken this opportunity to switch email to .Mac - the first time I\'ve ever paid for email - and can now be reached at goldtoe@mac.com.  '),(10842,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink FetchYahoo : FetchYahoo is a Perl script that downloads mail from a Yahoo! account to a local mail spool.  '),(10843,832265,'02,August,2004','2004-08-02 00:00:00','   Accidentally saw the Canadian,  urlLink end-of-the-world  flick, \"Last Night\", on cable and it may have kept me from going back to rabbit ears.  What a great idea to make an apocalyptic movie that goes after the absurd rather than the sentimental.  Plus, there\'s the great choice to forego all explanation and just present The End as spooky, blown out twilight rather than trying to pass off some legitimate explanation.  Much better to get right to the good stuff like the gas company employee (played by David Cronenberg) who spends his last day at work calling his customers to thank them for their patronage.  Those Canadians - so nice, so wacky and so untroubled by gun violence despite their comparable socio-economic demographics and love of firearms ... or so I\'ve been  urlLink told .  '),(10844,832265,'02,August,2004','2004-08-02 00:00:00','   This weekend featured what was probably the last decent  urlLink Pug Sunday  before the rainy season starts here in San Francisco.  And despite being a little damp, it was a good one.  Snorting dogbeasts aplenty, frequently in stampede.  All hail the cute mutants.  urlLink .:more:.  '),(10845,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink Aquarius Records  is quite a fixture in the San Francisco boutique record store world; a peacock diskery of the highest caliber.  Their mailing list is highly recommended in that it offers succinct reviews of a huge number of new items each week complete with soundclips.  However, nearly the entire staff seems to completely despite Stereolab.  As the store philosophy is \"We only stock what we like,\" there aren\'t typically a lot of negative reviews.  Which makes comments like \"in my opinion, this new album sees Stereolab retreating even further into the wallpaper that they already have become\" come off all the more harsh. I guess when you\'re a band specializing in 7 minutes of organ drone punctuated with lyrics like: \"True life embodying pleasure principle\'s noblest triumph over the cowering mendacity of bourgeois Christian civilisation\"  you\'re going to inspire some difference of opinion.  '),(10846,832265,'02,August,2004','2004-08-02 00:00:00','   Speaking of  urlLink Rushkoff , I\'ve really been wanting to catch his Merchants of Cool special for Frontline and,  urlLink hey , it\'s available in its entirety from the PBS website.  (Altho\' not in QT, and part 5 wouldn\'t play for me in RealPlayer).  The coverage of Total Request Live\'s rise to dominance is particularly good and the way in which rebellion against marketing has itself been co-opted seems entirely accurate.  However, Rushkoff glosses over a crucial question he only poses rhetorically: \"Has it always been this way?\"  He clearly thinks it hasn\'t and that the marketing of pop culture has somehow become less authentic ... including that of wrestling.  Thomas Frank in The Conquest of Cool makes the best case against this idea in his analysis of the authenticity of the 60\'s counter-culture.  But it\'s a quote I\'ve relied on  urlLink too often  ... mainly because it bags on Bob Dylan and the Grateful Dead.  Also, I understand Rushkoff\'s making a larger point about the particular image being propagated by youth marketing today.  I\'m interested in the harm of training today\'s youth to be ever more receptive consumers.  But I find it hard to get worked up about teens being sexualized by MTV and South Park or to sincerely wonder \"whither the authentic, non co-opted youth culture of today.\"  In the first case, the sexual identities that people adopt as teens are probably not indicative of the sexual lives they will later lead, regardless of what role the media played in the formation of that identity.  In the second, Rushkoff points out that the only way new, authentic youth culture gets created is by the co-optation and death of the old.  Basically there\'s always going to be something ... but it\'s probably pretty safe to say it\'s been  urlLink played out  by the time it shows up on Frontline.  '),(10847,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink   When I was training to become a man, the rabbi told me that my chanting was decent enough that I could pursue a rabbinical path should I be interested.  I didn\'t have the heart to tell him that I wasn\'t even returning for the Sunday School classes I\'d promised to go to the following year, let alone strive for the golden yarmulke.  But I mean, come on, it was a Jewish temple with  Sunday  School.  And don\'t get me started on the stained glass windows.  I thought of this off and on while reading  urlLink Douglas Rushkoff\'s  Exit Strategy which has a lot to do with Jewish myth, selling out and the ability of the free market to seduce and delude its partipants.  The way in which he explores the latter two concepts within the context of a near-future dot-com morality tale is quite interesting.  It\'s the myth part that gives me pause.  The conceit of the book is that the story being told takes place in 2008, but is being published two hundred years hence by futurefolk who\'ve only recently uncovered this testament.  This plays out on two levels.  The first is that the structure of the story is very much a fable and, as such, there\'s really no doubt from the early going that the narrator is cruising for a comeuppance.  In fact, in doesn\'t even take that long for us to be able to pretty well infer the exact form that comeuppance will take some 200 pages hence.  So that\'s kind of a bummer.  The second level is that the book was conceived as the first \"open source\" novel.  This means it was first published on the web and readers had the ability to annotate the book as if they were the future scholars studying this found text.  Very \"Talmud as hypertext\" and I\'m all for the idea.  But in practice it comes up a bit short.  Most of the notes end up being jokes of the reverse Caveman Lawyer variety rather than contributing to a genuine, multi-narrative metatext.  The note for Reece\'s Pieces is indicative of the trend: \"A peanut butter candy made famous by its paid placement in a Steven Spielberg film about a friendly alien.  This was long before extra-terrestrials\' true nature had been determined.\" The other type of annotation is the one-off political commentary which makes it difficult to buy into any sort of concept of the proposed distant future because the comments are so based in our current political reality.  For example, this disturbing note on the Triborough Bridge: \"destroyed in the Palestinian Liberation Terror attack of 2004, which poisoned the East River for several years.  It is believed that Palestinian Mullah Yasser Arafat ordered the attack shortly before his own assassination in late 2004.\"  Yikes! This brings us back, as it must, to Jews.  In an  urlLink interview  with former NYU debater Julian Sanchez, Rushkoff talks about his ideas on Judaism.  He says: \"Why don\'t we just give the best parts of Judaism to the rest of society, then give up being the chosen people, and just basically promote a secular humanism. What I want to do is tell people that Judaism is not a monotheistic religion, that Judaism is basically the process by which people get over believing in god.\"  Certainly a nice idea, and this idea of transcendence is played out in the last 10 pages of the book.  But I simply didn\'t buy it.  Instead the resolution works by literalizing the concept of deus ex machina and, without support from the future chorus showing how this world came to pass, it seems pretty tacked on.  '),(10848,832265,'02,August,2004','2004-08-02 00:00:00','   There\'s a great  urlLink article  about the Anticon crew in the latest issue of Wire magazine.  It\'s about as straight ahead a piece of music journalism as you\'ll find on underground hiphop.  Which is amazing because often the artists involved are too busy being aloof and, you know, arty to answer questions sincerely.  And usually the writers are too busy stylizing themselves to pieces in order to prove that they get this whole art-hop deal. I especially like the exploration of Anticon\'s music as indie punk hip-hop with the focus on DIY and found art.  And I love the fact that Dose One seems to live his entire life speaking only in poetry: \"Our lacking in artistic raw talent or schooling is made up for with chance and the willingness to drag a cash register or parking meter back to our apartment because it looks like it belongs there for awhile.  It also applies to our sampling nature and open palette collage writing style - basically it\'s an unlearning adventure in charmed and nervous lives.\" Basically.  '),(10849,832265,'02,August,2004','2004-08-02 00:00:00','   Did you feel  urlLink it ?  Still can\'t access my  urlLink bank  account online.  It\'s like we\'re living Lawnmower Man, but different.  '),(10850,832265,'02,August,2004','2004-08-02 00:00:00','   So I\'ve got to live a bit more of a rock n\' roll lifestyle from now on.  Public expectation and whatnot.  But to stay true to my Harper\'s roots, I\'m shooting for more of a rock n\' roll intelligencia-type deal.  Think Jeff Goldblum from Jurassic Park.  Think Richard Feynman but less  urlLink dead .   urlLink Think .  You see how it is.  '),(10851,832265,'02,August,2004','2004-08-02 00:00:00','   Harper\'s published a list of wacky protest slogans seen at the San Francisco peace rally ... the rally back in October, for some reason, not the one just this past week.  Anyway, in addition to \"Beat L.A., not Iraq!  Go Giants!\" and \"Tango for Protest\", their list also includes \"Transsexual Lesbian Vegan Epidemiologist Punk for Peace.\"  Funny thing is, I saw the TLVEP in question while waiting for the bus on 16th and Valencia that day.  Unless, there was a second one ... which, I mean, come on.  But do you realize what this means?!  My life is being published by Harper\'s ... the Readings section, sure, but that and the Index are all anyone bothers to read, right?  It\'s weird being famous.  '),(10852,832265,'02,August,2004','2004-08-02 00:00:00','   Ahh, the old \"I\'m back in school and forgot to do something\" dream.  '),(10853,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink   My most glaring geek deficiencies:  I\'ve never made it past the first hundred pages of Fellowship of the Ring.  I\'ve never played Risk.  And just hearing the Dr. Who theme song makes me uncomfortable.  Recently, I decided to kill the first bird and wing the second with the stone known as  urlLink Lord of the Rings Risk .  My friend Mary bought this for her husband Eugene and the three of us sat down last Friday for a little game of Middle Earth domination.  First off, it\'s been awhile since I\'ve played a game where the object is to wreck the other people.  Although we started off good naturedly, there was some tension as the little elf battalions started dying by the handful .  This was probably exacerbated by playing with a married couple ... lots of idle \"If you invade Rohan, I\'ll divorce you\" threats.  Also, even though we were playing the \'quick\' version, it still took over 4 hours.  That\'s longer than the movie, for crissakes.  In the end, I lost horribly, Mary almost fell asleep and I think we all learned an important lesson about the futility of war.  Except for Eugene, who won, and would do these little dances after successful rolls of the dice.  '),(10854,832265,'02,August,2004','2004-08-02 00:00:00','   Since MacWorld, there\'s been lots of talk about Apple\'s new web browser, Safari.  \"Wow it\'s superfast.\" \"Yeah, but it\'s got no tabs.\" And then to a lesser extent, \"Hey!  This web browser  urlLink wrecked  my operating system.\" I spent some time this weekend exploring the minority opinion. Basically, I got de-metaphor\'d.  No copy and paste.  No drag and drop.  And all because a Safari preference file wasn\'t being parsed properly.  How rude.  '),(10855,832265,'02,August,2004','2004-08-02 00:00:00','   Minor disturbance in my geek hemisphere ... I can\'t find a good video game to get into.  This may be for the best as my brief dalliance with CivIII nearly cost me my life, or at least kept me from seeing the sun for a weekend.  Of course, this is why I thought Tropico would have a greater draw.  The problem is I can\'t keep myself from playing as the benevolent communist.  I\'m sure it\'s fun to throw dissenters in jail and have members of their families summarily executed, but I just keep trying to pass literacy programs.  Panzy.  This leads to the problem with Freedom Force.  I love real time strategy and I like superheroes okay.  It\'s the 60\'s style that kills it for me.  I\'ve had enough of listening to the main character\'s pseudo-ironic declarations like \"Let\'s get those Reds\" and \"I can\'t hit a woman!\"  Well, that and the fact that the game will randomly go into Cubist mode in which all the polygonal surfaces are exploded and rearranged hypercube fashion.  Not fun. Jedi Knight II is probably the strongest candidate thus far.  You can choke people with the Force, you can shoot purple lightning, and, most importantly, you can go absolutely buck wild with a light saber.  Incidentally, I don\'t know why it took so many Star Wars games until LucasArts realized that without light sabers you\'ve basically got a pile of nothing with high priced theme music.  The trouble is that Jedi Knight II is apparently among the buggiest ports to the Mac.  There\'s a  urlLink forum  where a guy is collecting crash logs and the discussion of video tweaks is well beyond my interest. I think this is where I say I\'m going back to Snood and  urlLink link  to the post about giving the little game its propers ...  I\'m sick of Snood.  '),(10856,832265,'02,August,2004','2004-08-02 00:00:00','    Due to his DSL outage,  urlLink Sutter  worked from my place today.  Together we managed to finish off a box of Gille Swedish Ginger Snaps which are now my favorite cookie.   Okay, I had a head start at finishing them off.  But the best part is, you get a giant box of \'em for only $1.50 at the corner store (assuming your corner store has swedish cookies).  As far as scandinavian sweets go, I\'m also a fan of swedish fish, but apparently there are some that will  urlLink summon  a plague of flies when opened.  '),(10857,832265,'02,August,2004','2004-08-02 00:00:00','   I didn\'t have a favorite republican governor until today when Illinois\' George Ryan commuted the sentences of everyone sitting on death row.  Sure he\'s on his way out and this is really an extension of the moratorium imposed 3 years ago, but it\'s an amazing decision - especially coming from the heartland.  If you ever wanted to know why the death penality is wrong, here\'s a pretty good  urlLink breakdown : Because our three-year study has found only more questions about the fairness of sentencing, because of the spectacular failure to reform the system, because we have seen justice delayed for countless death row inmates with potentially meritorious claims, because the Illinois death penalty system is arbitrary and capricious -- and therefore immoral -- I no longer shall tinker with the machinery of death.  '),(10858,832265,'02,August,2004','2004-08-02 00:00:00','   I\'m a sucker for a good, flaming review of something I\'ve never seen or read.  Yeah, yeah, yeah - it\'s easier to destroy the cathedral than to create it.  But if you can create a somewhat smaller, but funnier cathedral out of the rubble of of the first one which was, like, soooo full of itself to begin with ... well, that\'s the metaphoric religion for me.  Here are a few recent favorites:  You mess with the hilatron you get the  urlLink horns \"Yeah, the critics are gaga over About Schmidt.  urlLink But  then they loved American Beauty.\" Here\'s a meta-foot in your  urlLink ass   '),(10859,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink   Just finished reading Cory Doctorow\'s  urlLink Down and Out in the Magic Kingdom  and it\'s certainly the most enjoyable long read I\'ve ever had on a computer (I found the html version much more comfortable than the pdf ... or, you know, you could buy it).  More than that, it\'s a great sci-fi story that reminds me a lot of Philip K. Dick - a mystery where the identity of the victim is more important than the identity of the criminal.  As in Do Androids Dream, you learn about the future through the accumulation of small, fascinating details rather than the expositional dump.  This is good.  As is repainting Disney World as ubergeek utopia.  To wit: The Haunted Mansion was experiencing a major empty spell: the Snow Crash Spectacular parade had just swept through Liberty Square en route to Fantasyland, dragging hordes of guests along with it, dancing to the JapRap sounds of the comical Sushi-K and aping the movements of the brave Hiro Protagonist. Everything that goes into the future of Magic Kingdom is a little more interesting than the mystery to be solved, so the resolution doesn\'t come off as super-satisfying.  Still, it\'s a great read.  '),(10860,832265,'02,August,2004','2004-08-02 00:00:00','   The other day, I saw a minivan with a vanity license plate that read \"CAMUS\".  This is:  Funny, if not ironic, since Camus died in a car crash Ironic, if not funny, given the minivan as symbol of normative, suburban living Intentionally absurd   '),(10861,832265,'02,August,2004','2004-08-02 00:00:00','   Some network maps I have known and loved:  The  urlLink puerile and geeky The  urlLink tribal The  urlLink partisan The  urlLink beautiful and conspiratorial   In the case of the first two maps on this list, the network is created explicity; it only exists because of the opt-in of its members.  With the Ryze map, the members signed up and provided their own list of links which were then translated by hand into the network map.  And you\'d like to think that some amount of free choice was invovled in gaining membership to the sexchart network.  In other words, these maps serve as mirrors for the members of the network.  You look a the map to see who\'s connected to you; to find your place in the overall network.  And, of course, to find out if you accidentally kissed your sister by virtue of the transitive property.  With the second two cases, the creator of the map is trying to uncover the implicit connections of the underlying network.  This network isn\'t created because of the opt-in of its members, but because of the question the creator wanted to answer.  So, instead of a mirror, you have a window into the system.  And a somewhat hackneyed metaphor.  What if you could automate the creation of these maps such that they didn\'t rely on explicit opt-in, but also shielded the results from the influence of the creator.  For example, maps based on the collaborative filters of Amazon or the All Music Guide.  Instead of just seeing lists of \"People who liked this also liked\" you\'d generate a map showing the web of inter-related items.  More interestingly, what if DayPop used a similar tool; showing connection between blogs based on the similarity in links that they post.  Well, you\'d have a pretty cool map, that\'s for sure.   '),(10862,832265,'02,August,2004','2004-08-02 00:00:00','   Back online after being lost in the wilds on the other side of the digital divide.  At one point, I was trying to hook up Speakeasy dial-up ... when I realized I didn\'t have a long enough phone cord.  But thanks to the UPS guy who was still delivering past 8pm, I\'m back together again.  '),(10863,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink Sutter  informs me that the hype about the iMac being discontinued was  urlLink untrue .  Now the news is that Apple will unveil a computer that can change color.  And maybe one that hovers.  '),(10864,832265,'02,August,2004','2004-08-02 00:00:00','   All the hard work and determination have paid off ... as of right now LemonNut is #1 of 168,000 Google results  urlLink returned  for \"great big fat pig.\"  '),(10865,832265,'02,August,2004','2004-08-02 00:00:00','   Lessons learned by a pyromaniacal jew this post-holiday season:  Christmas trees do, in fact, burn well But only briefly ... the needles burn off like they\'re covered in hairspray, the rest of the tree stays behind Trees with a little tinsel left on them are preferable.  The tinsel goes off like a string of firecrackers Trees with ornaments should be avoided.  They will probably blind you with exploding shrapnel Trees with the angel still on top ... man, even jews find that too angsty If you want to toast marshmellows over a burning christmas tree, be quick.  Also be ready for turpentine flavored smores Lighter fluid ... really just overkill If you run out of christmas trees, trying to substitute various leafy trees will not work.  Walk away gracefully People who water their christmas trees are no fun     '),(10866,832265,'02,August,2004','2004-08-02 00:00:00','   Christopher Marlowe is the 16th century dramatist best remembered for being played by Rupert Everett in Shakespeare in Love.  But more than that, he was also an Elizabethan  urlLink secret agent  who worked to uncover Catholic plots against the crown.  This is apparently true. What\'s less certain is whether he ran afoul of the notorious  urlLink Star Chamber  for being an atheist and decided to fake his own death to escape torture.  Assuming that he did, there are those who further believe that he fled for Italy where he continued writing.  These pseudoposthumous plays, some set in Italian towns, others about exiles, were sent back to England and published by the man we know as William Shakespeare.  Now, I heard about this by watching Michael Rubbo\'s  urlLink documentary  on Frontline.  When this documentary first came out last year, Rubbo himself  urlLink posted  about it to Metafilter ... but only identified himself as the filmmaker in comments.  So, this is the guy we\'re meant to believe in an authenticity of authorship debate?  '),(10867,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink   I didn\'t realize that the 15 inch iMac had been taken out of production until I saw the  urlLink news  that the same thing had happened to its bigger brother.  It would seem highly unlikely at this point that Apple would introduce a 19 inch version as was  urlLink rumored  to happen at SF MacWorld. Certainly glad I got while the getting was good.  This is unquestionably the best computer I\'ve ever owned.  '),(10868,832265,'02,August,2004','2004-08-02 00:00:00','   The closest thing I\'ve seen to a production of Waiting for Godot is kd lang\'s Constant Craving video ... which is to say, not very close.  So too bad I missed the beginning  of the  urlLink PBS  version which aired tonight in honor of the 50th anniversary of the play\'s  urlLink debut .  It\'s certainly strange to see all those stage directions acted out, and to feel the effect of all of those pauses.  As nice as it was, I don\'t think I\'ll be spending $150 on the 4 DVD  urlLink box set  containing filmed versions of all 19 of Beckett\'s plays.  How big is the market for that deal?  '),(10869,832265,'02,August,2004','2004-08-02 00:00:00','   I saw a printout from  urlLink liberation.fr  on the Blogger/Google deal ... apparently, weblogs are called either \'blogues\' or \'jouebs\' in french.  I\'m guessing that jouebs is an example of  urlLink inverse slang .  So, I wonder what \'blog\' is in  urlLink cockney ?  Update : Thanks,  urlLink Stephanie  for the lowdown on jouebs ... it\'s a contraction of journal and web.  Perhaps I subconsciously knew my proposed etymology was wrong and, as a result, dreamt of being harrassed by my 7th grade french teacher.  Why he looked like a youthful Rutger Hauer is your guess.  '),(10870,832265,'02,August,2004','2004-08-02 00:00:00','   Riding a far too bouncy bus home from Golden Gate Park, I was reminded of the death of Danish astronomer Tycho Brahe.  Tycho was the one who got a whole island for the purpose of building his naked eye observatory.  He also lost his  urlLink nose  in a rapier duel and had it replaced by a fake nose made out of precious metals.  And he had a pet moose.  He was kind of a big, dorky, 16th century weirdo, that Tycho.  But I was thinking of him because I\'d heard he had died from urine retention.  The  urlLink story  is that he was at a royal ball and refused to use the Danish facilities until the king left the table.  As a result, he developed a massive kidney stone and several days later his bladder burst.  Altho\' new evidence suggests he may actually have died from mercury poisoning ... he had a sideline as a alchemist, you see.  Regardless, it certainly wasn\'t the 3 glasses of mercury I had at breakfast that nearly caused my own breach of protocol on the bus ride home.  '),(10871,832265,'02,August,2004','2004-08-02 00:00:00','   So this is what it feels like to work full time.  It\'s all coming back to me now.  This wasn\'t even a full week of work but it certainly was eventful.  Things do not happen slowly down in Google.  Equipment, projects, people ... keep it coming with a quickness.  The decision to go out and see DJ Krush last night may have been somewhat of a misculation.  But Anticon opened up and we got to stand within mere feet of the diminutive Krush as he wrecked shit on some crazy Vestax sampler/mixer.  My big plans for the weekend involve eating the food I bought last week that I\'ve not had the chance to cook.  Tho\' I don\'t think it\'ll turn out as well as the  urlLink meals  to which I\'m growing accustomed.  '),(10872,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink Jason , Aaron and I were gathered around the old iMac, listening to the LA Blog event when the story of Blogger\'s acquisition broke.   urlLink Tony Pierce\'s  account pretty well captures my excitement.  And the coolness of having Ev break the news in the company of bloggers.  Moreover, I\'m quite relieved that the story is out there now.  It\'s weird having to tell your Mom, \"Listen - you can\'t tell anyone in the teacher\'s lounge about this Google news.\"  And it\'s doubly weird to find yourself self-censoring your blog to avoid mentioning why you were driving down 101 South in the first place.  So onward to the exciting parts ... getting to continue the exciting Blogger work down at Google, getting to eat their much hyped food.  Good times.  '),(10873,832265,'02,August,2004','2004-08-02 00:00:00','   Driving down highway 101 South, the most interesting scenery is the variation in noise barrier design.  Some walls have mossy toupees, others are patterned like lost tetris games - it\'s truly a visual cornucopia.  I thought \"I bet you could get a grant to explore the art of noise barrier design and how it relates to the American Dream.\" Then I found out that barrier design is actually  urlLink mandated  by the US Department of Transportation.  '),(10874,832265,'02,August,2004','2004-08-02 00:00:00','   I saw  urlLink Wild Style  for the first time over the weekend and feel I may have missed out by seeing it on DVD rather than its more traditional presentation on 8th generation VHS.  Because it\'s the DIY feel of both the movie and early hiphop that really helps to connect the dots between the NYC punk and rap scenes.  As does this factoid from the commentary track featuring director Charlie Ahearn and Fab 5 Freddy.  Charlie was freaked that if the soundtrack featured one of the 7 rap tracks available at the time that they\'d get in trouble for not clearing it.  So Freddy created an original breakbeat record for the movie with the help of a drummer and the guitarist from Blondie, Chris Stein.  The record was pressed and given to the DJs who appeared in the movie and that\'s what they used for all of the filmed routines.  Guitar riffs by the man who co-wrote Heart of Glass and Sunday Girl.  '),(10875,832265,'02,August,2004','2004-08-02 00:00:00','   My mom tells me that there are district-mandated steps her  urlLink school  has to follow when the TerrorMeter is elevated, as it was this past week.  For example, Orange requires that all teachers check their voicemail more frequently.  I guess the idea is to make sure that al Qaeda didn\'t call up a second grade teacher in suburban Missouri to give advance warning of a terrorist attack.  If we go to Red, outdoor recess is cancelled ... which is totally lame because you can only play Heads Up Seven Up so many times before you realize everyone\'s cheating.  The thing is, the idea behind the whole color coded terror level is working.  When the alert level is raised to Orange, parents call my mom asking what their elementary school prinicipal is doing to keep their kids safe from anthrax.  Fear is the mind-killer. Fear is the little death that brings  urlLink total obliteration .  '),(10876,832265,'02,August,2004','2004-08-02 00:00:00','   Been trying to lay low and keep out of the spotlight.  It\'s hard when all you want to do is live your life.  But by the weekend I\'d gotten really tired of being a prisoner in my own house or being forced to wear disguises when I\'d go out to buy groceries.  If you\'d sign here, the bottom copy is yours. Sorry, I just don\'t have time for autographs.  So I ended up going to an Exposition of Alternative Press ... okay I was at a  urlLink comic book convention .  I figured I\'d be camoflaged by the geek-tinged hipster cool, but Daniel Clowes was there and I\'d just rented Ghost World. I don\'t think he saw me.  But there was this horrible tension between me and Real World 3\'s Judd Winick ... perhaps because he knew  urlLink Ev  was going to  urlLink David Egger\'s  826 Valencia the next day to set some kids up with blogs.  And there\'s that whole weird Real World thing in Heartbreaking Work of Staggering Genius which I\'ve not read, but, like, totally appreciate anyway as the most important post-modern deal written about my generation, except older.  Or maybe the tension was because everyone at the convention had really nasty things to say about Judd Winick.  I kept quiet \'cause it\'s best not to make enemies with other celebs.  Even if they\'re just reality show famous.  '),(10877,832265,'02,August,2004','2004-08-02 00:00:00','   St. Louis Breckinridge remembrances of pretentions past: Is it hot in here or is it just I  Oh, a case purist, how precious  Listen, lady.  I won\'t sing your song tonight.  But the salutes are nice.  '),(10878,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink   Don DeLillo\'s Mao II is a short book crammed full of big ideas.  Lots of nuanced arguments about the power of images, the roll of the mob in modern society and the nature of identity.  The thing is, if your read almost anything about Mao II you\'ll read a similar catalog of its major themes ... which I don\'t think is terribly satisfying.  But the experience of reading the book certainly is.  The prologue, a mass Moonie wedding at Yankee Stadium, had me completely hooked: \"They want to snatch us back to the land of lawns.  That we are willing to live on the road, sleep on the floor, crowd into vans and drive all night, fund-raising, serving Master.  That our true father is a foreigner and nonwhite.  How they silently despise.  They keep our rooms ready.  They have our names on their lips.  But we\'re a lifetime away, weeping though hours of fist-pounding prayer.  World in pieces.  It is shock of shocks.  But there is plan.  Pali-pali.  Bring hurry-up time to all man.\" I love the shift in language, the use broken english, bringing you into the mass mind of the Moonies.  It may be a specific skill but DeLillo is the master of telling a multi-narrative story from the perspective of a ballpark crowd.  '),(10879,832265,'02,August,2004','2004-08-02 00:00:00','   NextBlog served me up an Illuminati conspiracy page wherein I read of the Bush administration\'s plan to find aliens.  I was a little skeptical.  Turns out it\'s based in some  urlLink fact  as the recently submitted budget calls for $279 million to explore the Jovian moons for traces of water.  A far cry from \"we\'re building a Casio keyboard on a mountain to communicate with our celestial bretheren through electro\" but it\'s interesting.  What\'s funny is that someone decided to punch up the budget by titling this section \"Where Are the Real Space Aliens?\"  Can\'t you hear the Texas twang?  Both in the distinction between the  real  space aliens and those bogus bastards who almost fooled me with their fancy autopsy special on Fox.  And then also between the  space  aliens whom we want to find and that other kind of alien whom we want to unlawfully detain.  Anyway, all this talk of aliens got me excited enough to look up an online  urlLink Drake equation  calculator.  Plug in your own values and determine whether or not we are alone.  Using my numbers, I found that there should only be 0.1 communicating civilizations in the galaxy.  Uh ... wait.  '),(10880,832265,'02,August,2004','2004-08-02 00:00:00','   Since seeing  urlLink Morvern Callar  and not responding to it, I\'ve been on a trashy movie kick.  To that end, I\'ve watched Reign of Fire and Near Dark on back to back nights.  Reign of Fire is, unquestionably, the superior movie as the dragons look pretty sweet and lots of things get incinerated in an eye pleasing manner.  Also, Matthew McConaughey takes a flying leap at a charging dragon while brandishing an axe which almost makes up for his religious hotbody role in Contact.  On the other hand, the box for Near Dark claims that it is the scariest horror movie ever made but really it\'s just about hick vampires and lots of slow, delberate scenes of people stumbling through dirt fields.  It\'s like Days of Heaven meets John Carpenter\'s Vampires but really crappy.  After Dark was written and directed by Kathryn Bigelow who is the ex-wife of James Cameron which explains why Pvt. \'Game Over\'  urlLink Hudson , Pvt. Vazquez and Bishop from Aliens are all vampires in the movie.  But the worst offense is the inclusion of Joshua John Miller as a kiddie vamp.  His name isn\'t familiar, but his  urlLink face  and annoying lisp have ruined episodes of Family Ties, Growing Pains and the Greatest American Hero.  Fortunately, he explodes fantastically at the end of the movie.  And if I\'ve ruined this mess of movie for you, you\'re welcome.  '),(10881,832265,'02,August,2004','2004-08-02 00:00:00','   I heard about the Challenger disaster in an elementary school classroom and thought it would be my \"remember where you were when it happened\" event until a couple years ago.  This morning I found out that Columbia broke apart over Texas while reading RSS feeds in NetNewsWire ... looking at  urlLink ev\'s  blog, to be specific.  This is the first national disaster I\'ve read about primarily through weblogs.  It\'s strange because you feel connected to the other folks who are going through the same thing, in the same way ... experiencing it in a public way rather than each of us alone with our CNN.  But you lose the sense of the enduring ache.  Something that hurts for longer than the one post dedicated to it.  Even on the larger community sites, with their active comment threads, it\'s sooner rather than later that someone posts the next post.  And it not only seems, but feels, that I\'ve moved on too quickly.  I feel the reflexive tug of cable news because that\'s how the Gulf War trained me to deal with national events.  Race home after school and watch the war during dinner.  The event endures and we experience it passively; comforted by the illusion of being tied together in an experience mediated by patriotic banner graphics.  It\'s the desire to say \"I am a part of this - I am going through this.\"  To somehow reduce an entire war, or shuttle crash, to my experience of it.  Better to make the one post.  '),(10882,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink   This is how  urlLink All Music Guide  usually looks to me when I view it in Chimera.  The content of the banner ad at the top of the page leaps out of its box and applies itself as the background.  And it\'s fun because the ad still rotates so the effect is sort of like a cheap porn site.  '),(10883,832265,'02,August,2004','2004-08-02 00:00:00','   Wired News:   urlLink Sneaky Toolbar Hijacks Browser : \"Xupiter is an Internet Explorer toolbar program. Once active in a system, it periodically changes users\' designated homepages to xupiter.com, redirects all searches to Xupiter\'s site, and blocks any attempts to restore the original browser settings.\" I got hit by this one while approving ads for work.  Suddenly, you get a really ugly toolbar in IE, complete with scrolling text.  It\'s like an ad for a crappy BBS has been appended to your browser.  '),(10884,832265,'02,August,2004','2004-08-02 00:00:00','   So, technically, a  urlLink googlewhack  is supposed to be a two word search that yields a single result.  Be that as it may, I\'m proud that my blog is the sole hit for \" urlLink patriotic banner graphics \".  Truly, we must all do our part in the War on Terrorism.  '),(10885,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink Steve \'s found a crazy little app for keeping an eye on geomagnetic storms.  The guy who wrote it should see if he could get some product placement in  urlLink The Core .  What I\'d like to see is a tool that\'ll plug the most recently discovered  urlLink Near Earth Asteroid  into my current blog post.  Or, even better, something like WeatherPixie that\'ll show if there are any planet-killers on the way.  If the end of the world\'s coming, I\'d like to make sure it\'s captured on my blog in goofy pixeltoon format.   '),(10886,832265,'02,August,2004','2004-08-02 00:00:00','   In the spirit of  urlLink LazyWeb , I\'d love to see an english major write a thesis titled \"Moses and the Cowboy: The Role of Extra-Narrative Characters in the Films of the Coen Brothers.\"  Okay, I\'ll give you some suitably pretentious chapter titles, too:  Humor or: \"Made me laugh to beat the Man\" \"Obstackles\" : The Case for and against O Brother\'s Tiersias Deus ex Machina, the Clockmaker God \"Is he really going to  urlLink jelly  up the sidewalk?\"  Thanks for the nod in the acknowledgements section.  '),(10887,832265,'02,August,2004','2004-08-02 00:00:00','   The  urlLink truth  is I can\'t help myself.  The hardest decision I used to face was figuring out what to eat.  Now, my weekday lunch and dinner choices are made for me ... and they\'re \"all you can eat\" choices.  Last Thursday I didn\'t stay at work for dinner and, when I got home, found my own devices sorely lacking.  All I had was some apple sauce and a sack of popcorn kernels.  So I decided to order pizza from the joint arond the corner.  Went to pick it up ... not ready yet.  Came back later ... they couldn\'t break a ten.  (In fact, I don\'t think they could break anything as all they had in the register was 3 singles, a twenty and a fifty.  Honestly, how\'s that gonna work out for you Pizza Joint Around the Corner?)  I knew this libertarian who used to say that subjecting yourself to any moment of paternalism or dependency was to play the role of the animal being hand-fed by children.  He used to eat a lot of nasty-looking cheese steaks, that guy.  '),(10888,832265,'02,August,2004','2004-08-02 00:00:00','   I woke up at 7:30a this morning ... without the aid of an alarm clock.  Today is the day I became an old man.  I\'m gonna go get frustrated trying to set my VCR.  '),(10889,832265,'02,August,2004','2004-08-02 00:00:00','   Susan Sarandon\'s on stage with the gaggle of past Academy Award winners  (Dead Man Walking) .  At the same time, she\'s on the Sci-Fi Channel spouting dialogue like this in Children of Dune:  \"I won\'t be satisfied until every Atreides bastard is crushed beneath the jackboots of the Corrino Sardaukar.\" I think there\'s a lesson to be learned about the nature of fame and glory.  Or maybe I just need to stop watching TV.  '),(10890,832265,'02,August,2004','2004-08-02 00:00:00','   The War to End All Wars was the name of the 1914-1918 conflict until, you know, it didn\'t.  So it makes sense that you\'d want to put some thought into naming your military exercises. With the benefit of 90 years of post-modern advertising, the naming of the Second Gulf War has been turned into a  urlLink branding exercise .  For less important exercises and conflicts, naming is governed by the Navy\'s Code Word, Nickname and Exercise Term System - the regulations of which you can read  urlLink here .  I especially like section 7.d.5, \"Exercise terms may not contain words that convey anything offensive to good taste or derogatory to a particular group, sect or creed.\"  I guess that\'s why Operation Chocolate Clarinet is on the permanent assignment list.  '),(10891,832265,'02,August,2004','2004-08-02 00:00:00','   Some time before us Bloggerfolk joined up with Google, Al Gore dropped by Mountain View HQ for a visit.  He\'s a good tech nerd, that Al, and a  urlLink Mac  user to boot. Which is why I feel really bad about the dream I had last night.  Al was a homeless knight, wrapped in chainlink armor and driven mad by circumstance.  \"I\'ve not the wherewith to carry on,\" he babbled.  It was hard to hear his voice; his chafed and windburned lips slurred his speech.  Yet, I felt he hadn\'t come to seduce those of us huddled around a burn barrel we\'d found under a city overpass.  A gathering of those made apostate from politics, those who felt caught bewteen both war and  urlLink anti-war .  Al didn\'t say \"It could have been different\" or \"I told you\".  He sat and watched the fire.  And we watched his breath curl out from under scarred lips, leaving wavelets of moisture in the night-time air.   '),(10892,832265,'02,August,2004','2004-08-02 00:00:00','   In 6th grade, I had my mom for a teacher.  Okay, it wasn\'t for homeroom or anything but still.  My biggest challenge was figuring out what to call her.  Obviously \'Mom\' was out of the question and \'Mrs. Goldman\' was just going to crack up my friends.   So I ended up never once referring to her by name.  To get her attention, I\'d sort of snap my fingers or subtly cough.  It worked out alright ... except for the one time I really got under her skin by mouthing off and I realized that I was in trouble with both my mom and my teacher at the same time and therefore in about as much shit as a 12 year-old smart-ass could possibly be.  My mom\'s now principal at the same elementary school.  Today she won the Pillar of Parkway award which is sort of the equivalent of being inducted to the School District Hall of Fame.  Congrats, Mom!  I\'m proud to say I knew you when you were coming up through the minors.  '),(10893,832265,'02,August,2004','2004-08-02 00:00:00','   You know, us Jews don\'t fit in so well sometimes.  In elementary school, for example, I had to tell the story of Hanukah every Christmas as part of my ambassadorship from the dreidel-people to the land of the stockingfolk.  But as I was the only kosher kid in my school, and since I don\'t actually know the story of Hanukah, I just made up a new one each year.  \"Then Moses said, \'And I\'ll form the head!\'\"  And I\'ve always felt left out that all the stories about seeing God in bread mold, or a skid mark, or a  urlLink street sign  invariably involve the big JC.  Well, no longer.  Enter the  urlLink Carp  that Spoke in Hebrew and was Presumed to be God (or Possibly the Devil) but Got Turned into Gefilte Anyway.  '),(10894,832265,'02,August,2004','2004-08-02 00:00:00','   Sure, the story about Don Johnson smuggling $8 billion over the Swiss border might seem like silly celebrity gossip.  But since the  urlLink Harvard Business School  is reporting it, I\'m confident that my interest is due only to the underlying issues of international finance.  '),(10895,832265,'02,August,2004','2004-08-02 00:00:00','   What press secretary told Tony Blair to go stand behind the Strangelovian mini-blinds?  And not to go for the low-hanging stereotype, but the ashen toothy look is oddly reminicent of  urlLink Brad Dourif  from The Two Towers.  Oh, also, the accompanying  urlLink story  from the Guardian is a good read: \" Why, then, does the government say there is a legal case for war?  \"It is difficult to know on exactly what grounds the government is basing its arguments that that is a legal basis for war.\"  '),(10896,832265,'02,August,2004','2004-08-02 00:00:00','   Under the reflected glow of a dry docked Carnival cruise ship, we partied it up at the Blogger/Google shindig last night.  Lots of smart and lots of funny, but I was more the latter and less the former when I said to myself: \"I can drink all the whiskey I want and never get hungover.\" \"Hey, Chris Pirillo\'s taking  urlLink pictures  ... try to look cool.\"   '),(10897,832265,'02,August,2004','2004-08-02 00:00:00','   The passengers of Caltrain #81 knew we were in trouble long before the conductor announced \"We\'ll be trying to leave the station soon ... pray for us.\"  After all, cars on the 25mph frontage road had been passing us easily, putting the lie to our express train status.  An hour later, when we finally arrived in downtown SF, a concerned crowd was gathered to meet us.  They were concerned as we had delayed their southbound departure on the same suspect train by half an hour.  As such, the unstated sentiment between the arriving crowd and that departing was both one of well-wishing and misplaced resentment.  '),(10898,832265,'02,August,2004','2004-08-02 00:00:00','   The newly released Fear and Loathing in Las Vegas  urlLink DVD  contains a 1978 BBC documentary on Hunter S. Thompson.  Basically, the BBC and  urlLink Ralph Steadman  show up at HST\'s place and try to re-enact the book.  They roadtrip it to Vegas and then to LA, all the while trying to shake out the crazy.  Hunter\'s not having any of it and ends up hiding from the camera most of the time.  When he\'s not blowing lines from behind a clapboard, that is.  In the doc, Hunter says he has trouble with his public persona because people are usually expecting to meet Raoul Duke instead of him.  And this conflict plays out in his commentary track for the movie.  When he launches into a sustained high pitched squeal, or a sudden burst of \"fuck you\'s\" it seems somewhat like an act.  On the other hand, he has great things to say about the book of Revelations and launches into an eloquent dressing down of Timothy Leary and his role in the demise of the 60\'s.  The other bonus features are a hoot, too.  The producer, Laila Nabulsi, talks about how she tried to put Terry Gilliam on a dwarf quota.  And when that didn\'t work, she tried to bribe him with midget porn in order to keep more little people out of the flick.  It didn\'t work.  '),(10899,832265,'02,August,2004','2004-08-02 00:00:00','   More quasi-comical dreams of being hunted.  Last night I was pursued by a mad scientist who looked like a cross between  urlLink Henry Gibson  and Dr. Finklestein from the Nightmare Before Christmas.  His dastardly plan was to get me locked up in the booby hatch by injecting me with a madness-inducing serum.  At least he had a better plan than the shark.  I remember seeing the mapley syrup substance shot into my arm and thinking \"Hey, being crazy\'s not so bad.\"  Altho\' there was a loud buzzer that went off just as the madness set in that was pretty annoying.  Later, after I\'d been locked up, I heard the same sound while careening through the halls of my asylum.  And I remember thinking to myself, as in the end to a sitcom, \"Well, there\'s goes another one off to the booby hatch.\"  I was a little upset that the dream didn\'t end at that point ... it seemed like good closure.  '),(10900,832265,'02,August,2004','2004-08-02 00:00:00','   I\'ve got this recurring shark dream.  Sounds scary, huh?  Except, it\'s just me running through a dry sewer and the shark trying desperately to catch up.  As there\'s no water, the shark is forced to flop around and pull itself along with its fins ... but, you know, menancingly.  '),(10901,832265,'02,August,2004','2004-08-02 00:00:00','   William Gibson  urlLink waxes  nostalgic on outmoded tech, including a grade school favorite, the mimeograph: \"You typed, once, on a waxed paper \'stencil\', clipped this over a silkscreen device with a moving pad or drum of ink behind it, and your mimeograph ran off (or silkscreened, really) as many copies of your document as you required. Owing to the physical peculiarities of the medium, though, it was unwise to underline too frequently on a mimeograph stencil: the single unbroken line was particularly prone to tear, producing leaks and smudging.\" urlLink Jason , Aaron and I were talking about the mimeograph and other printing processes and we couldn\'t figure out why blueprints were still in use since they seemed to involve such a wacky production process.  I imagine it\'s both because of size and because draftsmen draw on trace.  I didn\'t remember the details at the time, but blueprinting is a wild scene.  It was invented by  urlLink John Herschel  in 1840 (his father discovered that funny planet between Saturn and Neptune).  It involves  urlLink coating  a piece of paper with a chemical whose structure is unknown and then explosing it to a ridiculously strong light.  The process is so old that the blue produced is referred to as Prussian blue.  That\'s like having a color called Ottoman orange.  I know it\'s still somewhat used because I saw the blueprint machine that belongs to the city of Baltimore do its crazy photochemical voodoo.  Leave it to Charm City to keep alchemy alive.  '),(10902,832265,'02,August,2004','2004-08-02 00:00:00','   My stepdad is a vet with a purple heart and I\'m a whiny liberal who lives in San Francisco ... which is why I\'m often confused that we both love Full Metal Jacket.  I think it comes down to appreciating the movie as a true depiction of what it was like to go to Vietnam vs. a statement of why not to go to Vietnam in the first place.  urlLink Kottke  recently said something related to this, \"it must be difficult to recruit someone who has seen the first 15 minutes of Saving Private Ryan or Full Metal Jacket (although some may sign up because of such films).\"  It turns out the parenthetical is the truer part - at least, according to  urlLink Anthony Swofford \'s Desert Storm memoir, Jarhead.  In it he  urlLink says , \"Vietnam War films are all pro-war, no matter what the supposed message, what Kubrick or Coppola or Stone intended. . . The magic brutality of the films celebrates the terrible and despicable beauty of their fighting skills. Fight, rape, war, pillage, burn. Filmic images of death and carnage are pornography for the military man.\'\'  Obviously there\'s a great documentary to be made on the free play of signs and signifiers involved when soldiers watch anti-war war movies in the battlefield.  But moreover it shows that depictions of war, regardless of the political intent, are compelling because of the horror, not in spite of it.  '),(10903,832265,'02,August,2004','2004-08-02 00:00:00','   Last Thursday, I saw where the internet  urlLink lives .  It lives in a cold, glass  urlLink box  protected by biometric scanners and a regulated power supply the size of a bus.  Knowing I had no business in the holy of holies, the internet winked at me as I trundled off to the Blogger alcove.  I\'d never thought of internet security in terms of floor-to-ceiling black metal cages, but that\'s how it works where the electronic meets the physical.  I never thought the internet had an off switch, but it\'s a start to see a set of emergency shutdown switches that would bring down the entire data center.  Of course, like launching nuclear weapons, you have to press both buttons at the same time.  I once knew a responsible, well-mannered young woman who, as a girl, set off the fire alarm in her elementary school.  Not out of mischief aforethought, but just because she felt her hand drawn to the big, red switch.  The alarm pulled her.  Best not to think of such things.  '),(10904,832265,'02,August,2004','2004-08-02 00:00:00','   I fell in love with last night\'s Caltrain conductor when he said, \"We\'re now arriving in downtown SF ... or at least as close as we get to it.\"  The other passengers came around when he added, \"Please remember to gather all your personal belonings before leaving.  Any items left behind will be seized and ground into a fine powder.  The powder will then be turned into a launchpad which Caltrain will use to go to the moon.\"  '),(10905,832265,'02,August,2004','2004-08-02 00:00:00','   Last night, someone asked  urlLink Ev  if he was related to a clan of Williamses from Out East.  Oh, the pain of over-common last names.  It made me recall the time I was offered condolences for the relative of mine that OJ killed.  Happily, I also remembered my out-east friend John Williams who isn\'t the famous composer but once convinced someone to the contrary.  You\'re John Williams ... the composer?! Sure. But you\'re so young!  You must have been a little kid when Jaws came out. Well, yeah, but my early work wasn\'t really that complex.  I mean, how hard is \'DUH-nuh ... DUH-nuh.\'   '),(10906,832265,'02,August,2004','2004-08-02 00:00:00','   There\'s really not much difference between the state of Missouri and the State of Nature.  I mean, we\'ve got some laws there, like the Boats on Moats  urlLink amendment  allowing \'riverboat\' casinos to be classed as boats as long as they sit in a manmade pool of water.  Someone once said \"You can put a pig in a dress, but it\'s still a pig\" and the Show Me state replied \"What if she were wearing pumps?\"  But what Mo is missing are the sensible laws.  While we did get that cock fighting and bear baiting  urlLink initiative  passed, a  urlLink bill  to prevent cars from parking in bike lanes has hit a snag.  It never pays to underestimate the Parking Freedom lobby.  '),(10907,832265,'02,August,2004','2004-08-02 00:00:00','   \"I am trying to break your heart\" is a documentary about the making Wilco\'s album, \"Yankee Hotel Foxtrot.\"  And it\'s excellent.  But  urlLink Sutter \'s got me wondering at this point if there are any bad documentaries about music.  Right after September 11th, PBS aired its Rock & Roll documentary on repeat and I\'m forever grateful.  I\'m even dying to see the second part of Decline of Western Civilization despite not liking metal.  Oh yeah, Sam Jones, the director of the Wilco film, has a \'making of the making of\'  urlLink blog .  '),(10908,832265,'02,August,2004','2004-08-02 00:00:00','   After successfully activating my new ATM card, the CompuVoice told me: \"Remember, this is your new ATM card - you should begin using it immediately.\"  '),(10909,832265,'02,August,2004','2004-08-02 00:00:00','    Out, vile jelly! Where is thy lustre now?     '),(10910,832265,'02,August,2004','2004-08-02 00:00:00','   In the 6th grade, Andrew Askuvich picked up this kid and pile drived him into the rail road ties lining the playground sandbox.  I was one of only 2 eye witnesses to this event which would quickly become a fight of epic proportions in the history of Craig Elementary.  This week Andrew and his fiancee visited San Francisco and we took in a Giants game together on Sunday night (the Giants won on a 12th inning RBI double from Marvin Bernard).  As the pictures reveal, I was somewhat obsessed with close-ups of Andrew\'s mellon.   For one thing, he\'s got grey hair and that\'s a little weird to see on someone with whom you spent the majority of your childhood.  The other thing is that we had a running gag as kids about the size of Andrew\'s nose and how, you know, it would one day destroy the world.  I must admit it doesn\'t look too menacing in the photographic record.  urlLink .:more:.  '),(10911,832265,'02,August,2004','2004-08-02 00:00:00','   I\'ve been conducting an experiment wherein I\'ll fill my head with thoughts of dogs right before going to sleep in order to encourage canine dreams.  It\'s not working.  The other night, I invented a new kind of soy milk that was pre-packaged with cereal.  The dream kinda fell apart when I couldn\'t figure out why the cereal was still crunchy.  Still, that was more coherent than the one where I got invited to a taping of the Clint Eastwood Show.  It was more dinner theatre than anything else and Clint wasn\'t even there.  Instead, I sat next to Matthew McConaughey and tried to convince him that I thought Reign of Fire was a legitimately great movie.  \"No, really ... when you jumped at the dragon with the axe!  Man, that was heroic.\"  And then everyone got up and started line dancing ... I felt pretty excluded.  '),(10912,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink  Do you see what happens when you walk Barry in the bottom of the 9th?'),(10913,832265,'02,August,2004','2004-08-02 00:00:00','   Since destroying a statue and thereby winning the war,  urlLink some journalists  have been touched by the poet ... Shelley in particular, and his poem  urlLink Ozymandias  in specific.  The use of the poem is not really surprising as it seems readymade for the occassion.  Even the part about \"Two vast and trunkless legs of stone\" recalls the broken pipes that remained attached to the pedestal after Saddam\'s body had been pulled away.  Ozymandias came about as the result of a sonnet writing competition (ESPN2) between Shelley and Horace Smith.  Smith\'s  urlLink version  concludes with: We wonder, -- and some Hunter may express Wonder like ours, when thro\' the wilderness Where London stood, holding the Wolf in chace, He meets some fragments huge, and stops to guess What powerful but unrecorded race Once dwelt in that annihilated place. The theme in both is the impermance of all things, not just the reign of a dictator or pharaoh.  In this case, the pharaoh is Ramses II (Ozymandias is his greek name) who built lots of temples, defeated all his enemies and was generally well loved by his people ... well, except for those who Moses asked to be let go.  Still, the \'works\' referred to in the poem truly were great and emblematic of the last great period of the Egyptians.  So, rather than self-congratulatory triumphalism, the real message in toppled statues is that 1) the Jews are probably involved somehow and 2) all empires end up buried under lone and level sands.  '),(10914,832265,'02,August,2004','2004-08-02 00:00:00','   \"It\'s finished.  It\'s finished!\" the renegade commando screamed on this week\'s episode of 24.  He\'d been shot and the doctorfolk at the MiniMed couldn\'t save him.  He knew he was dying and cried out to reveal the location of the secret evidence.  Altho\' he died one sylable too soon, Kiefer quickly figured out that the microchip was implanted in his side.  Here\'s the relevant part.  Kiefer whips out a scapel and starts rooting around in the dead man\'s body cavity ... using his fingers to probe the side wound.  And that\'s how we know that Kiefer\'s going to eat it at the end of the season.  The \'it is finished\' dialogue combined with the wound probing is clearly meant as an allusion to Jesus and  urlLink Doubting Thomas  (with Kiefer as Thomas).  And, as everyone knows, Thomas got speared to death in India.  Ergo, Kief\'s a dead man.  (Okay, he\'s not gonna die next week.  But it\'s him or Kim, I guarantee.  And, they could kill him and still do another season ... just set in the past.)  '),(10915,832265,'02,August,2004','2004-08-02 00:00:00','   I had to call QuickBooks the other day and was chatting with support as we were waiting for my request to process. \"Are you down here in Mountain View,\" I asked.  After a long pause, the support guy replied in a conspiratorial tone, \"Did they tell you that?\"  \"They?  Uh, no.  It\'s just I\'m at Google which is right across the street from the Intuit HQ ... if you were down here, maybe I could see your office.\"  \"No,\" he said.  \"You can\'t do that ... I mean, we\'re in Reno.\"  It\'s hard to make friends sometimes.   '),(10916,832265,'02,August,2004','2004-08-02 00:00:00','   Elizabeth Stanley-Mitchell  urlLink points out  that regime change doesn\'t work too well unless you have something else to change into ... and, coincidentally, that\'s precisely what\'s lacking in Iraq.  Fortunately, we\'ve got the benefit of experience on our side.  Twenty years before this  urlLink photo-op  took place, the US engineered the last regime change in Iraq when, in 1963, the CIA helped oust Gen. Abdel Karim Kassem by  urlLink aiding  the anti-communist Baath Party. And the Baath Party brought peace and freedom to Iraq, turning the country into a model of liberal democracy.  You know, all of this is making the Bay of Pigs look really sensible.  '),(10917,832265,'02,August,2004','2004-08-02 00:00:00','   The RIAA is going after a  urlLink Princeton student  for contributory copyright infringement.  Not for building a peer-to-peer community or for creating a file downloading tool, but simply for indexing an already existing file-sharing network.  Indexing: enabling more effective search means enabling more effective terrorism.  '),(10918,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink'),(10919,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink   There comes a time in a young man\'s life when he must open up his heart and home to a lady.  Even if only for one night.  Mabel\'s owner, Elizabeth, is out of town and I took over dogsitting responsibilities from her roommate, Meredith.  Actually, that\'s not fair.  Mabel could have stayed at home for one night ... it was more like she watched 8 Mile with me and then stood guard against the weird noises in the night.  Pretty good deal considering all Mabel wanted out of it was to be pet and taken around to see the smells of the Mission.  I mean, I know I had a good time ... I think she did, too. I promised to call later this week.  urlLink .:more:.   '),(10920,832265,'02,August,2004','2004-08-02 00:00:00','   \"They got lifestyle figured out in Spain\" I said. \"It\'s like CaliforniaPlus\"  \"Where are you from,\" I was asked.  \"The midwest.\"  \"Figures.  No one actually from California would say something like that.\"  (Once, I was in Paris for all of 18 hours.  I flew around the city, tracking down all the sites I\'d studied in 4 years of high school french class - I didn\'t go up the Eiffel Tower ... that\'s for tourists.  I was wearing a brown leather jacket that was 10 years out of style and I think that\'s why I got asked for directions a couple times.  Still, my missourian french gave me away as just another American and all the Parisians I met would quickly switch to english rather than hear me butcher their native tongue.) I laughed and said, \"You know, I\'ve found with some native Californians that they don\'t realize how good they have it.  And then, one day, they figure that what they really need to do is move to Boston or something.\"  \"Then,\"  urlLink Sutter  rejoined, \"they get run over by a car.\"  '),(10921,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink Aaron Schwartz : \"On February 21, 2003 I watched my first episode of Buffy the Vampire Slayer (6x09). ... Since then, I have watched every aired episode of Buffy, spinoff Angel, and creator Joss Whedonï¿½s other show, Firefly, from the beginning, in order.\"  '),(10922,832265,'02,August,2004','2004-08-02 00:00:00','   I\'m back from the soggy graduation proceedings in Providence wherein I saw many remarkable things.  I saw a baccalaureate service held in the First Baptist Church in America that opened with the Muslim call to worship, featured a Taoist prayer and closed with Chinese dragon dancing.  I ate a lobster omelet, heard a kickass song about seaweed and drank one of the best Chai teas I\'ve ever had.  And got to see my brother graduate in a soaking wet cap and gown.  Never one to miss an opportunity to exploit sibling embarrassment, there are naturally many pictures (in deference to my mother, the shots of his crackden dorm room have been withheld).  urlLink .:more:.  '),(10923,832265,'02,August,2004','2004-08-02 00:00:00','   My today\'s being scraped across America as I wing my way to Providence, RhodeEye for my brother\'s graduation.  The last time I was there was during the weekend of my own graduation ... we took a family roadtrip up North to check out the campus. For some reason, it took 2 hours to find a place to eat lunch.  I assume we\'ll be better informed this time.     '),(10924,832265,'02,August,2004','2004-08-02 00:00:00','   As I\'ve not driven regularly in over 8 years, I\'d forgotten a couple things about my driving habits.  In particular, the fact that I sing along loudly with the car stereo and play a mean set of steering wheel drums.  And nowadays, I\'m using my iPod to pump out the jams so I\'ve got 2500 of my favorite songs to belt out.  Incidentally,  urlLink this device  from Belkin is just great.  It\'s an iPod charger that also has an amplified line out for use with cassette adapter.  Perfect for giving you that extra jolt needed to hit the falsetto intro to New Order\'s \"Temptation.\"  '),(10925,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink TelevisionWithoutPity : \"The 25th Amendment again. I feel like every show I watch is dealing with this theme. (Okay, they haven\'t really gotten into it on Trading Spaces.) It\'s just a big coincidence, right, that popular culture seems pretty taken with the idea of removing a sitting President from office for being unfit? Right?\"  '),(10926,832265,'02,August,2004','2004-08-02 00:00:00','   I ended up hating on the first Matrix movie because everyone kept talking about how amazing it was from a metaphysical perspective.  And the movie supported this kind of over-seriousness: \"No one can be told what the Matrix is.\"  Well, actually, you can.  It\'s for this reason that I was completely blown away by Matrix Reloaded.  There\'s an NYT  urlLink article  today about Cornel West and his cameo in the movie.  He\'s quoted as saying, \"The second Matrix movie actually critiques the idea of the first. It\'s suspicious of salvation narratives. It\'s deeply anti-dogmatic.\" I completely agree.   urlLink Those  who would try to sweep the sequel under the easy \"brains in jars\" philosophizing of the first are really missing out.  Reloaded goes far beyond the original by building layers of myth around the central question of \"What is the Matrix?\" so that there is no easy, superficial answer.  All of which makes me wonder why the effects are all that\'s mentioned in every review I\'ve read.  I mean, it\'s got to be said that the effects are phenomenal.  The chase sequence certainly makes the Coruscant chase of Episode II look like the unimportant cutscene it actually was.  But when I was leaving the theater all I could think about was the first time I saw Dune or 2001 and knowing that there was more going on than I could have possibly absorbed in the first viewing.  It\'s science fiction at its best - creating a complex, complete otherworld and filling it with questions that extend beyond the plot.  To call it a \" urlLink popcorn movie \" is like calling Blade Runner a stereotypical film noir.  '),(10927,832265,'02,August,2004','2004-08-02 00:00:00','    Saddam Hussein is/was a man. All men are mortal. Therefore, Saddam Hussein had weapons of mass destruction.  In  urlLink this NRO article , Jim Lacey tries to come up with a reason why we\'ve been unable to find WMD in Iraq.  Occram\'s razor be damned, Lacey says.  It can\'t just be that there weren\'t weapons there.  Instead, he concludes, \"Saddam wanted the program and was willing to endure crippling sanctions to have it. However, his henchmen were unable to deliver and, unwilling to be on the receiving end of Saddam\'s zero-defects program, they faked it.\"  So let me get this straight.  First off, it was these folks on the right who kept trying to get me to understand that Saddam was a megalomaniacal dictator so evil that he might as well have been the love child of Stalin and Hitler.  Now these same people can\'t possibly see why said megalomaniac would get pretty prickly on issues of sovereignty and forced inspection.  But the part about the WMD program being faked, well, that\'s just a new level of self-deceit.  So, our intelligence is so bad that we can\'t tell the difference between a real threat and a staged one?  If that\'s the case why should I ever believe you when you say that we must use force now in order to protect ourselves from the Big Bad?  But the biggest question of all is that what does this say about the ethicality of having invaded Iraq if, in fact, the WMD was a fake.  The pro-war arugment has seemingly become that pre-emptive force is justified not only in the case of an actual threat but also in the case of one that ultimately turns out to be a fake.  In the name of efficiency, we\'ve managed to remove the second step from the doctrine of Shoot First and Ask Questions Later.  '),(10928,832265,'02,August,2004','2004-08-02 00:00:00','   Somewhere in the previously posted picture is a bunch of poison oak that I seemingly bathed in on the way down to the beach.  Jason Sutter was unaffected.  '),(10929,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink  '),(10930,832265,'02,August,2004','2004-08-02 00:00:00','   Proposition : The internet is amazing.  I needed a place to get my smog checked and the Feeling Lucky hit for \"smog check San Francisco\" pointed me to the perfect place just 8 blocks away from my house.  What\'s hot is that the referring  urlLink page  is a personal account of smogcheckery, complete with hard-to-read font and layoff commentary.  I\'ve been connected to someone I\'ll probably never meet (and a great  urlLink garage ) by virtue of search.  Opposition : My blog is the first hit returned for a search on \" urlLink rodent death .\"  '),(10931,832265,'02,August,2004','2004-08-02 00:00:00','   Most often, the quality of your seats at a baseball game makes for an easy 1-line conversation rather than a genuine debate.  \"Decent seats, huh?\"  My $8 upper deck tickets at today\'s A\'s game were no exception.  The angle made some foul balls seem fair, but the view was good enough to see Roger Clemens rack up career win 298 and Jorge Posada deposit a monster homerun into the left field bleachers.  Also, I got to witness the pride of the dad in front of me who found a triple decker peanut in his bag during the 3rd inning.  I wasn\'t scoring the game or I would have made a special note.  '),(10932,832265,'02,August,2004','2004-08-02 00:00:00','   This weekend I\'m going to be reunited with an old friend ... my mom\'s \'89 Camry.  It\'s being carted out here on an open-air car carrier (by way of LA and Fresno).  This is the car I learned how to drive in, the one I lost my mind in while listening to the Les Mis soundtrack on a family roadtrip to Colorado and the one I got my first ticket in when out after curfew at the Rocky Horror Picture Show.  Now we\'ll be spending our mornings and evenings together on 101, hunting for parking in the Mission and taking new, Kansas-free road trips.  '),(10933,832265,'02,August,2004','2004-08-02 00:00:00','   At the last professional conference I attended, I was told that the only way to write software requirements was to analyze the root cause of the problem using a shipyard management tool.  This seemed odd.  Moreso, when it was revealed that the tool was called the  urlLink Fishbone Diagram .  You gotta question any system that uses fish parts to divine knowledge.  Fortunately, this week I\'m at the Adaptive Path  urlLink conference  where the second slide warned against the One True Way of design.  Word.  '),(10934,832265,'02,August,2004','2004-08-02 00:00:00','   Come on, Spring.  I know it was a dry winter and all, but this raining 3 times a day thing is preposterous.  And you\'re such a tease.  When it\'s not raining it\'s all blue skies and comfortable climes.  In fact, usually it\'s like that  when  it\'s raining.  They have a  urlLink saying  for that type of behavior.  The devil is beating his wife in hell with a skillet.  Not a pretty phrase, is it, Spring.  Get your deal together.  '),(10935,832265,'02,August,2004','2004-08-02 00:00:00','   There\'s a bogus  urlLink story  that the word posh is an acronym for Port Out Starboard Home.  The idea being that luxury cabins should be in shade on a boat heading east in the morning and west at night.  It\'s not true.  Well, the part about where the sun would be is true.  But since I take Caltrain south in the morning and north at night, I end up sitting on the starboard side for both legs to avoid the glare and better watch Dr. Strangelove on my laptop.  Sosh.  That\'s what I am.  And sick of taking the stupid train.  '),(10936,832265,'02,August,2004','2004-08-02 00:00:00','   There\'s an SUV ad with the tagline \"Go climb a driveway.\"  It means, \"We know that you know that you\'re never going to use your SUV to climb a mountain like in those other SUV ads.  But you want the SUV anyway.  That\'s cool.\"  The point of this form of irony is to disarm the audience; to remove the need to pass judgement by making fun of yourself.  The catch being that you never actually own up to anything.  Anyway, Zoe Williams has a great Guardian  urlLink article  on irony in general and this point in specific: \"So, you take a cover of FHM, with tits on the front - and it\'s ironic because it appears to be saying \'women are objects\', yet of course it isn\'t saying that, because we\'re in a postfeminist age. But nor is it saying \'women aren\'t objects\', because that would be dated, over-sincere, mawkish even. So, it\'s effectively saying \'women are neither objects, nor non-objects - and here are some tits!\'\" This reminds me to reread The Fall at some point wherein  urlLink Camus  admonishes the existentialist set for being such dicks. \"I was wrong after all, to tell you that the essential was to avoid judgement.  The essential is being able to permit oneself everything, even if, from time to time, one has to profess vociferously one\'s own infamy.\" Well, yeah, the thing is you\'re not supposed to like the guy who says that.  '),(10937,832265,'02,August,2004','2004-08-02 00:00:00','   Find someone at work to tell me why I\'m #5 for on a search for  urlLink \"shit showers\" .  Surely, I can make it to #2.  '),(10938,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink DragonRaid : \"In DragonRaid, the participants role-play Christian characters in the land of EdenAgain, where they are known as LightRaiders. LightRaiders are members of the TwiceBorn who, together, courageously follow the call of the OverLord of Many Names and go into the Dragon Lands to combat and defeat evil.\" ( via  urlLink Feï¿½rag  )  '),(10939,832265,'02,August,2004','2004-08-02 00:00:00','   A recap of things that happened while I was sick: Yo la Tengo at the Fillmore: You\'ve not heard \"Little Honda\" until you hear it build into a noise-feedback climax only to drop back down into the next verse of \"It\'s not a big motorcycle, Just a groovy little motorbike.\"  A drunken lover\'s spat in my backyard at 5am: Not sure which of my crazy neighbors were involved.  It was hard to place the low, droning male voice who slurred for 15 minutes, \"You\'re not going to hit me anymore.  You\'re NOT going to HIT me anymore.\"  Winged Migration: This movie\'s French, yo.  Gotta lot of good-looking bird photography.  The soundtrack\'s got various bird-inspired tunes that sound like the soft-rock offspring of Rush and SpinalTap.  Santa Cruz: I got to be the guy who gets to the front of the parking toll booth line and realizes he has no money (I\'d gotten a pass that apparently wasn\'t good for anything). I\'m really sorry ... but what do I do now? [Angry stare]  You leave.  And, I feel as a result of sleeping oddly, I had a dream about being abducted, strapped face down to a massage chair and forced to suckle its hairy nipple headrest.  '),(10940,832265,'02,August,2004','2004-08-02 00:00:00','   The producer behind the X-Men movies is looking to make a live-action Transformers  urlLink flick : \"The feature film will tell the story of an intergalactic war between two races of robots: the heroic Autobots and the evil Decepticons. When their fight comes to Earth, the future of humanity inevitably hangs in the balance.\"  I see Vin Diesel in the role of  urlLink Grimlock .  '),(10941,832265,'02,August,2004','2004-08-02 00:00:00','   From the  urlLink Brazil  DVD box set, I learned that the whole Buttle/Tuttle device was Tom Stoppard\'s idea and basically defined the structure of the movie.  From  urlLink K5 , I learned that: \"a glitch in US airport security is causing headaches for men named \'David Nelson.\' For some unknown reason (supposedly if \'they\' discussed the matter it would breach security) security computers give a \'no fly\' or \'potential terrorist\' warning whenever someone named David Nelson attempts to board an airliner.\"  '),(10942,832265,'02,August,2004','2004-08-02 00:00:00','   So, this past Sunday I saw The Matrix Reloaded for the third time at the  urlLink Coronet Theater .  I\'m still incredibly impressed.  And worried about the sequel.  I\'m sure that whatever the Wachowskis have planned for Revolutions will be great.  But I feel it won\'t be as mind-bendingly wonderful as I initially  urlLink thought .  Specifically, I now have a $1 bet with  urlLink Jason Sutter  about the trilogy\'s conclusion.  He is of the \"nested matrix\" camp - those who think that the reason Neo can control the sentinels at the end of the movie is because the \"real\" world is, in fact, another matrix.  I believe that Neo has this real world ability (no quotes) because he is now part-machine.  The natural resolution in this scenario is that machines and humans realize their inter-dependence and must combine as the yin and yang to advance together (translation: big fight with Agent Smith).  Anyway, Scott Manning has some great things to  urlLink say  on this topic.  '),(10943,832265,'02,August,2004','2004-08-02 00:00:00','   I used to be a hotdog technician.  On a hot and muggy afternoon in St. Louis, I tended to the sausage needs of overweight suburbanites as they rested from digging through discount software bins to snarf on free snacks.  If you want to get people to spend $50 on add-on floppies to Strip Poker 7, you need to provide some snackage.  That was the first time I almost died.  Sweating my heart out over an open grill in the St. Louis summer and trying to field absurd hotdog requests from the thrifty .  \"I want mine completely black on one side but raw on the other!\"  Jason Sutter eventually took pity on me, but hotdog jokes have followed me to this day.  To wit,  urlLink Lane\'s  recent portrait.  I\'m sporting my old favorite sweatshirt in those photos.  I still love Themselves.  But now I\'m rocking the  urlLink Blogger hoodie .  '),(10944,832265,'02,August,2004','2004-08-02 00:00:00','   \"I don\'t know who I am anymore,\"  urlLink Chris  said in a panic.  \"If this doesn\'t stop soon, I\'m gonna end up being for the war.\"  Who would\'ve thought that a rock n\' roll concert could go so far off the rails.  The New Pornographers show started out with just a few generic problems.  A bad soundcheck that left the vocals basically inaudible during the first song.  A too drunken keyboards player who was occassionally way out of sync and would snap his fingers arhythmically when not playing as tho\' he was trying to convince himself that he\'d found the beat.  But then the band noticed the kid in the Rush t-shirt at the front of the crowd who was playing air guitar along to every song.  \"Hey - c\'mon up here, kid,\" they told him.  And up he trundled to hop around stage playing air guitar during the next tune. At the end of that song, his friend came up to play air bass.  Then their lady friend played air drums on the song after that.  Eventually there was an entire shadow band pantomiming the show, spewing plumes of chugged beer in mock rockstar glee.  It was a nightmare.  The two albums that the New Pornographers contain only amazing songs so it was still fun to see them played live.  But I get a little nervous when the accepted standards of performance are so blatantly flaunted.  Audience in the crowd, band on the stage.  It\'s the natural order of things.  '),(10945,832265,'02,August,2004','2004-08-02 00:00:00','   Aaron Fagerstrom is a master of  urlLink flesh  and  urlLink bone .  And MattB has got a brand new  urlLink blog  full of tales of Midwest bawdry.  Plus great non-blog  urlLink stuff  from his LoFi days.  In short, Sutter\'s online  urlLink empire  is on the move.  '),(10946,832265,'02,August,2004','2004-08-02 00:00:00','   I\'d been warned, but I waited for the  urlLink 22 Fillmore  anyway.  Very foolish.  20 minutes later I jumped in a cab.  The cab driver explained to me that the Union Street Fair had messed up everything.  He also explained that he was from Tunisia and that he\'s sick of hearing tourists complain about the cold summers in San Francisco and that he moved here to get away from the heat.  I liked him immediately.  \"I hate the heat, too.  And Tunisia, that\'s like desert hot,\" I said. \"Nah ... the desert\'s further south, Tunisia\'s more Mediterranean.\"  I felt bad for being another American completely ignorant of world geography.  But I looked it up later.  Where this dude\'s from may be all riviera\'d, but the Sahara\'s definitely in that deal somewhere.  Passing Dolores Park, we saw a paddy wagon being stocked up with homeless people who\'d been sleeping near the tennis courts.  The cabbie said he used to run in the park but stopped because he was afraid of all the homeless who slept there.  \"Do they have a lot of homeless people in Tunisia,\" I asked. \"Nah ... when they find people living on the streets they just put them in homes or take them to the doctor if they\'re sick.\" \"Wow!  That\'s nice that people get taken care of there.  How\'s it work that Tunisia can do that for its people, but we can\'t do that here.\" \"Well, Tunisia\'s a more socialist country than America.  Here, you\'ve got to rely on yourself which is what makes it great.\" \"Yeah, I guess so.  But still it\'d be nice if we took care of folks.\" \"Yes.  Some people are like dogs and need to be taken care of like dogs are taken care of.\" A Tunisian true believer of the American Dream.  Well, I\'m the one who thinks I know where the desert is because I saw it on a map.  '),(10947,832265,'02,August,2004','2004-08-02 00:00:00','   Just got back from driving around  urlLink Treasure Island  with  urlLink Jason Sutter .  What a fucking misnomer.  Treasure Island is what happens when the Navy decides to float a suburban turd in the middle of the San Francisco Bay.  Also, it\'s largely abandonded.  The Fleet Admiral Nimitz Convention Hall?  Not that happening.  Good place to film a zombie movie, however.  After returning home, I remarked to my diminutive friend that I was feeling pretty worn out.  He wondered why that was.  So I turned to my new/current employer for  urlLink advice .  I\'m still tired, but as a result of the Feeling Lucky hit, I found this helpful  urlLink info  about Myo-Blast CSP3.  Sounds great: \"Well guys, ive been using this for a while now and although i have had seriously satisfying pumps at the gym i have found that i have diveloped large warts around the crack of my arse and in between my now muscular toes. Sadly as a result of this i am no longer having satisfying dumps in the can since it stings and this sensation never stops hurting me. Does anyone else suffer from these symptoms?\"  '),(10948,832265,'02,August,2004','2004-08-02 00:00:00','   Pulling up the rear of the employment wagon, I finally joined  urlLink my   urlLink cohorts  in full-timey Google status.    Come Monday (the day of my beating-in) I get a new shiny badge, a bucket of rocks and everything.  Woot!   '),(10949,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink Wired News : \"Mike Gontelli, a late arrival to the game that evening, said that when he arrived in Shadowbane \'there were hundreds of tombstones. New players were being beaten and tortured. Newbie blood was flowing like a river. I knew it wasn\'t real, but it was oddly terrifying.\'\" ( via  urlLink lil\' Goldtoe  )  '),(10950,832265,'02,August,2004','2004-08-02 00:00:00','   Just picked up a bottle of Crisco and a 4 pack of batteries from Walgreens.  Crisco always seems dirty on its own, but the C cells suggested some additional level of depravity.  Does vegetable oil conduct?  Some sort of auto-electro-lubricatory ridiculousness?  Good lord, my  urlLink mother  reads this.  And me not only buying Crisco and batteries but also with a rented copy of  urlLink Secretary  secreted away in the pouch of my hooded sweatshirt.  So, yes, for all of this.  And because the internet gives it away for free.  '),(10951,832265,'02,August,2004','2004-08-02 00:00:00','   Bob Hope was so old that the guy who wrote his  urlLink obituary  is dead.  '),(10952,832265,'02,August,2004','2004-08-02 00:00:00','   I\'d been wanting to see a game at  urlLink Wrigley  for a while now and finally had the chance a couple weekends ago.  Moreso than even Fenway, Wrigley Field is all about the old timey baseball feel.  Instead of a Jumbotron, there\'s an electromotive scoreboard for keeping track of balls and strikes and hand-turned placards for everything else.  I was sitting dead center in the outfield bleachers, directly beneath the scoreboard.  Whenever there was a one ball count, you\'d hear an electric whine from the board laboring to display that number.  During the 7th inning stretch, rows of heads popped out of the empty holes on the scoreboard to see the guest conductor lead the crowd in Take Me Out to the Ballgame.  From the outfield, you can\'t see any ads at Wrigley.  There are two small video strips in the grandstand that show the score and have a Sears logo.  Aside from that, nothing.  Quite the contrast to PacBell park.  At one point, some of the rowdy bleacherfolk I was sitting near tried to start the wave.  They were mercilessly heckled and revealed to be out-of-towners posing as locals.  A guy in my row turned to the would-be wavers and said, \"I was born 3 blocks from here and have been coming to Wrigley my entire life.  So let me tell you, there\'s no fucking wave!\"  Right on.  Modern baseball park design has focused a lot on trying to incorporate the stadium into the city rather than just dump it on the side of some highway.  Nowhere is that neighborhood feel more authentic than at Wrigley.  From our seats, I could look straight down Waveland and Sheffield Avenues.  And in addition to the folks watching from the rooftops across the street, I saw parents playing catch with their kids in the street, others flying kites outside the park.  Wrigley is the neighborhood.  All this and the Cubs won.  Certainly, not disappointing.  '),(10953,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink COM 480: Ethnography of On-line Role-playing Games : \"In this course, we will conduct an ethnographic study of the behaviors, cultural   practices, and motivations of MMORPG players. The course packet will include   readings that explore role-playing games, virtual community, and the construction   of identity on-line. Extensive attention will also be given to     methods for conducting research on-line.\" ( via  urlLink Costik  )  '),(10954,832265,'02,August,2004','2004-08-02 00:00:00','   When Errol Morris talks about self-deception, he doesn\'t do so from a superior position of \"This only happens to other people.\"  He argues that we all self-deceive; perception easily deludes us into thinking of ourselves as protagonists in our own personal dramas.  Well, maybe not Buddhists or something.  It\'s this sense that\'s lacking from  urlLink Spellbound , the spelling bee documentary.  There\'s a lot of \"Man, these kids and their parents are freaks\" and not enough of \"And so are we all.\"  The kids are wonderfully dorky and the movie has great narrative tension.  But instead of pulling back to show the spelling bee as just one of the many ways we create arbitrary definitions of \'smart\' and \'successful\', the movie tacks on some ad hoc conclusions about the American Dream.  Spelling as the great American scholastic tradition and lots of Algerisms about how anyone can succeed in this Great Land.  Bah.  The flipside of Errol Morris\' argument about Perception: The Great Deluder is that there is such a thing as historical truth.  \"There\'s a real world in which real things happen.  And in some small way, my job is to look at that world and to try to figure out what those things might be.\"  Andrew Jarecki\'s documentary  urlLink Capturing the Friedmans  takes a somewhat different slant.  In this story about a Long Island family torn apart by a pedophilia and sex abuse scandal, the hook is that it\'s all on tape.  The Friedmans were meticulous in recording their familial meltdown and surely there\'s some interesting stuff going on about the American compulsion for  urlLink self-documentation .  Of course, the alleged abuse itself, that\'s not on film.  Most likely because it never happened.  There\'s a lot of Thin Blue Line moments with prosecutors and witnesses doing a great job discrediting themselves.  But Jarecki\'s not coming down one way or the other, and, according to  urlLink Debbie Nathan , that\'s part of the marketing strategy: \"While the film was in production, Jarecki told the Friedman family he thought the two were innocent of the charges. Polling viewers at Sundance in January, he was struck by how they were split over Arnold and Jesse\'s guilt. Since then, he\'s crafted a marketing strategy based on ambiguity, and during Q&As and interviews, he has studiously avoided taking a stand. Teaser ads pitch the film as a Long Island Rashomon: \'Who do you believe?\' For Jarecki and his PR people, the question is rhetorical.\" That ain\'t so cool, Mr. Moviefone.  It\'s much more interesting to walk the tightrope between \"the struggle with perception\" and \"the truth is out there\".  To wit, Errol Morris\' next movie, Fog of War, consisting solely of interviews with Robert McNamara.  In this LA Times  urlLink piece , Morris says: \"What it does is take you inside someone\'s head. It\'s part dream, part history, part self-analysis, part self-justification, part mystery.  People say nothing can redeem McNamara\'s conduct during the war, and maybe that\'s true. But the fact is he\'s trying to grapple with who he is, trying to come to an understanding of himself and the world. And just because there is one voice, that doesn\'t mean I wasn\'t hard on him, it doesn\'t mean the voice is left alone in some uncritical way.\" I can\'t wait.  '),(10955,832265,'02,August,2004','2004-08-02 00:00:00','   Sigh.  My adopted home state has decided to do its impression of what democracy looks like in a recently westernized Baltic country.  With all the recall talk, I had some trouble tracking down what the ballot will look like come October 7.  It\'s not good.  According to this  urlLink article  there will be two questions.  First, an up or down on Gray Davis.  Then, if the answer is down, the election of the replacement governor.  The election of the replacement governor (who will serve until 2007) will be based solely on a first past the post vote.  No redistributive voting, no ranking - just winner take all.  This is a bad system of voting in general, but absolutely retarded when the election is going to happen in less than 80 days and the field will be flooded with any random weirdo who\'s managed to get 65 signatures.  Like the  urlLink guy  who\'s running on the legalization of domestic ferrets platform.  A lot of the commentary about the recall has been very folksy.  Isn\'t America great?  We\'ve got accountability for our leaders and anyone can run for office.  Yeah, anyone can run.  But with only 80 days to campaign, it\'s not going to be about who\'s got the best ideas for how to fix California\'s problems.  It\'s going to be about who has the most money and name recognition.  And, yes, these are the dominant factors in modern politics anyway, but they will be the only factors for a hyper-crowded field of candidates with no real time for debate.  Which means we\'re totally getting Arnold.  And I didn\'t move to California to end up living in Minnesota.  '),(10956,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink Errol Morris : You know, I have this version of the expulsion from the Garden of Eden.  God, in expelling Adam and Eve, kind of felt bad. He had gotten very angry, right? You know, you get angry and then you feel, \"Well, maybe I overreacted.\"  So, God was in that kind of mood when he expelled Adam and Eve from the garden. But his hands were tied. He had to go through with it; he had made the decision. God doesn\'t want to constantly second-guess himself. But he thought, \"I know. I\'ll give them self-deception. Things are going to be truly horrendous out there, but they\'ll never notice.\"   '),(10957,832265,'02,August,2004','2004-08-02 00:00:00','   Everything was going just fine - I\'m watching the VW/iPod  urlLink ad  and digging on the romance between the iPod\'s curves and those of the NewBeetle.  And the bit where the music goes distorted as the lady switches from earbuds to car speakers ... nice.  Then I wonder, \"What\'s this song?\" '),(10958,832265,'02,August,2004','2004-08-02 00:00:00','   Political outcry only works if there\'s a handle.  Congress wasn\'t willing to make the larger argument that the war on Iraq was unjust.  But now there\'s the Case of the False Nigerian Uranium and all the would-be presidents are gearing up to beat the administration with its own  urlLink 16 words .  To wit, Howard Dean\'s  urlLink got  16 Questions for President Bush.  The one I found most remarkable was: \"4) Mr. President, we urgently need an explanation about the very serious charge that senior officials in your Administration may have retaliated against Ambassador Joseph Wilson by illegally disclosing that his wife is an undercover CIA officer.\" What!  The source is a Nation  urlLink article  and, apparently, senior officials used Bob Novak to out the wife of Ambassador Wilson.  All because Wilson came out with the story that there never was any reason to believe that Iraq was getting uranium from Africa.  '),(10959,832265,'02,August,2004','2004-08-02 00:00:00','   When Aaron and I saw Winged Migration, he said that the music really reminded him of  urlLink Nights in White Satin  (extended version).  I said it sounded like  urlLink something else .  Last night, I heard the 8 minute version of Nights for the first time and I feel I\'ve seen the other side.  There simply isn\'t a more pretentious song.  The song builds into a majestic orchestral surge and the breaks down into a spoken  urlLink word  litmag crapfest: \"Bedsitter people look back and lament Another day\'s useless energy\'s spent Impassioned lovers wrestle as one Lonely man cries for love and has none.\" I don\'t understand how this song was a hit.  Moreover, I don\'t understand how it was a hit 3 different times in Britain during the 70\'s.  People really couldn\'t find another song to make out to?   '),(10960,832265,'02,August,2004','2004-08-02 00:00:00','   Had a great weekend in Chicago visiting my old college roommate Jonny B.  Saw the Cubs beat the snot out of the Braves at Wrigley, powered in part by a Sammy Sosa homer that squeaked over the ivy in left field.  Got to see the Art Institute which features a great exhibit on Himalayan art, Van Gogh\'s  urlLink Bedroom , Wood\'s  urlLink American Gothic  and the big ole\'  urlLink Seurat .  And I feel, with slightly more time, we would have made the trek out to  urlLink Highland Park  to trash PapaJon\'s sportscar.  '),(10961,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink DGCopter : \"I have discovered the joys of taking very cold showers and letting my fan dry me off. That\'s as close as I\'ll get to a/c.\"  Ah, not-so-fond remembrances of the hot, humid East Coast summers. I spent a couple in New Jersey, lying on the concrete floor of a college building\'s sub-basement ... the only cool spot in the entire state.  I\'m still puzzled as to why everyone doesn\'t live in California ... I guess the parking\'s pretty bad.  '),(10962,832265,'02,August,2004','2004-08-02 00:00:00','   \"I do miss sausage,\" a vegetarian friend confessed to me a couple days after my 4th of July BBQ.  This year\'s shindig featured a Salute to Sausage with 5 different kinds of sausage representing their respective homelands.  We also had the Great Potato Salad Face-Off courtesy of  urlLink Steve  and  urlLink Stacy .  And then it got cold and everyone started wearing my clothes.  There were also 2 digital cameras per attendee so the pix you see here are just the beginning.  (For example,  urlLink Eric\'s ,  urlLink Ev\'s  or  urlLink Willo\'s )  urlLink .:more:.  '),(10963,832265,'02,August,2004','2004-08-02 00:00:00','   And Buck65 said: The story of my childhood is bad luck and crises. Born in the year of the rat and I\'m a Pisces. Which makes me a ratfish.   '),(10964,832265,'02,August,2004','2004-08-02 00:00:00','   Jeffrey Lee Parson gets made fun of by the  urlLink NYT  for not disguising his identity when he launched the Blaster worm: \"As a byproduct of each infection, every victim\'s computer sent signals back to the \'t33kid.com\' Web site that Parson had registered in his own name, listing his home address. The computer bug also included an infecting file called \'teekids.exe\' that experts quickly associated with Parson\'s Web site: Hackers routinely substitute \'3\' for the letter \'e\' in their online aliases.\" Congrats to the Paper of Record for being hip to l33t-speak.  The editor of Virus Bulletin further ribs Parson for being so conspicuous, \"I guess we should praise the Lord for stupid people, right?\"  Or blame them for leaving easily-exploited security holes in their software.  '),(10965,832265,'02,August,2004','2004-08-02 00:00:00','   From Nelson\'s  urlLink blog , I learned the State Fair\'s this weekend ... home of many fine attractions, including: B.J. the Goodwill Ambassador Look for this charming puppet in his Coca-Cola Model-T meeting and greeting guests and handing out souvenirs to his new friends!  Speaking of  urlLink Big Fun , I watched the commentary to Heathers last night and learned many fun facts. There\'s a lot of Kubrick infatuation: director, Michael Lehmann made everyone watch Full Metal Jacket before the shoot - hence, the ultra-wide angle lens look. Writer, Daniel Waters, also loves Kubrick.  Originally, the movie ended with the death of Winona.  In one version, she was to have watched Christian Slater explode, give the kiss-off to Shannon Doherty and then siddle up to Martha Dumptruck ... who pulls out a gun, says \"Fuck you, Heather\" and shoots Veronica in the mouth.  As Veronica lays dying, Dumptruck rises from her motorized wheelchair and exclaims, \"I can walk!\" Kubrick love notwithstanding, Lehmann and Waters ... really never did anything ever again.  Waters next wrote Ford Fairlane and then Hudson Hawk.  Lehmann last directed 40 Days and 40 Nights.  Sad, really.   '),(10966,832265,'02,August,2004','2004-08-02 00:00:00','   There\'s no reason for high school reunions - just wait for your friends to get married.  Last weekend, Andrew & Hallie got hitched in St. Louis, and I got to hang out with a whole bunch of old-timey friends.  It\'s funny talking about home ownership with a bunch of folks you used to trash homes with.  Not that we were punks or anything, but there sure were a lot of \"remember when we broke that lamp / wall / turkey dinner including the plates\" stories.  urlLink .:more:.  '),(10967,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink Fox Loses Bid to Stop Sale of Franken Book : \"There are hard cases and there are easy cases. This is an easy case,\" said U.S. District Judge Denny Chin. \"This case is wholly without merit both factually and legally.\"  '),(10968,832265,'02,August,2004','2004-08-02 00:00:00','   I shoulda known better when  urlLink Steve  gave me all his skee-ball tickets.  '),(10969,832265,'02,August,2004','2004-08-02 00:00:00','   Today\'s the last day of my vacation.  I feel great, simultaneously much vacated, and yet also full of many new seen sights.  Herewith I present the photographic record of said sights.  If you\'re looking for hints on what to do in Vancouver, I highly recommend the  urlLink Lonely Planet  guide.  urlLink .:more:.  '),(10970,832265,'02,August,2004','2004-08-02 00:00:00','   The first time I heard Birdhouse in Your Soul was on a mix tape of  urlLink Aaron\'s  sister.  Because it came right after It\'s the End of the World as We Know It, for a long time I thought REM sang Birdhouse as well.  I had a troubled youth, in many respects.  Just got back from seeing the They Might Be Giants documentary,  urlLink Gigantic , and it was a rockin\', dorky good time.  I\'ve never had a stronger mimetic experience than watching high school debaters argue over the meaning of Particle Man. But the most impressive thing to me was what was left out.  Somewhere toward the end, there\'s footage of the two Johns doing a release show at Tower Records for Mink Car.  The card reads \"September 10, 2001 - Midnight in Manhattan.\"  And you\'re just like, \"Oh, man here it comes - time to make contact with the great touchstone of national tragedy.\"  Instead, you\'ve got TMBG singing  urlLink New York City , and the film simply transitions from the record store appearance to a stage performance of the same song.  And that\'s it.  No falling towers - no channel-flip style news footage.  In talking about the happy-sad mix of most their songs, John Linnell says something like \"It\'s not really interesting just to watch someone\'s heart break.\"  It\'s much more interesting to wrap despair in this complex wrapper of accordion punk rock.  '),(10971,832265,'02,August,2004','2004-08-02 00:00:00','   Back at home and there\'s no place like it.  Before heading out this morning, I zipped around the  urlLink Vancouver Art Gallery .  Got to see one of the Lombardi conspiracy  urlLink maps  up close ...  damn you Henry Kissinger!  And I saw Danica Phelps\'  urlLink work  which consists of an obsessive, beautiful record of the transactions and events in her day-to-day life.  My first thought was \"Wow!  This is what a blog looks like put up on the walls of an art museum.\"  But maybe that\'s not quite right.  '),(10972,832265,'02,August,2004','2004-08-02 00:00:00','   Not much time left on vacation, so I\'ve been doubling up.  Yesterday was the Day of the Two Formal Gardens (one  urlLink japanese  one  urlLink chinese ).  Well, it was also the day of dropping a bunch of loot in  urlLink Zulu Records  which has the same kind of well written reviews you find in SF\'s  urlLink Aquarius  and a wider selection of  urlLink Anticon  stuff than I\'ve seen anywhere.  Today I hit up both Pacific Spirit Regional Park and Queen Elizabeth Park - one very treeish with a bit with a bog, the other more flowery.  And then I had to eat dinner in three different places to squeeze in the remaining restaurants I wanted to try.  Nothing can stop the Blob.  '),(10973,832265,'02,August,2004','2004-08-02 00:00:00','   Thanks to  urlLink Sutter  I\'m now rocking out at  urlLink goldtoe.net .  I feel all grown up.  '),(10974,832265,'02,August,2004','2004-08-02 00:00:00','   After walking to a lighthouse surrounded by an old growth forest, I decided to do my Zelda: Wind Waker imitation today.  I rented a motor boat from up in  urlLink Horseshoe Bay . I was a little worried that since I lost my driver\'s license a while back, I might have some trouble.  Turns out there aren\'t many rules up here in regards to renting a boat.  You don\'t need to actually know how to drive one.  You don\'t need to wear a life jacket.  Basically, you just get in the thing and point it at the pretty stuff. And, man, is it fun.  Howe Sound is dotted with islands you can zip around, some of them peopled by lazy sea lions.  I was amazed by the distance I covered in 2 1/2 hours.  Having spent the afternoon communing with marine life, I decided to spend the evening eating it.  Whole mess of oysters, whole mess of trout.  Good times.  Oh, and then I saw Northfork which is a really fine looking movie that may or may not be about something.  God, maybe.  '),(10975,832265,'02,August,2004','2004-08-02 00:00:00','   My first day in Vancouver has been pretty amazing.  This morning, I saw a crazy  urlLink clock  that runs on a combination of steam and the dropping of stainless steel balls. I walked all over Stanley Park which, I feel, is better than NYC\'s Central and SF\'s Golden Gate if only because there\'s an unlimited number of places where you can stop and eat salmon.  But the real kicker was the all-you-can-eat japanese restaurant.  In some ways I feel I\'d been waiting for that meal all my life.  12 dishes later I stumbled back to my hotel. Tomorrow I predict there will be more eating.  '),(10976,832265,'02,August,2004','2004-08-02 00:00:00','   My car got broken into for a second time last night.  Someone applied a screwdriver to my already-violated driver\'s side door lock.  Fortunately, there was nothing worth taking.  '),(10977,832265,'02,August,2004','2004-08-02 00:00:00','   Well, it only took  urlLink 2 1/2 months , but the  urlLink outcry  has finally started to build over the Case of the Outed CIA Agent.  I like the Administration\'s denial strategy.  \"No way was it Karl Rove!  That\'s ridiculous.\"  Ok, not Karl, got it.  Who, then.  Is it time to start guessing initials?  '),(10978,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink Letterboxing  is a particular type of scavenger hunt in which the goal is to decipher a clue leading to a box.  The box contains a logbook and a rubber stamp.  When you find the box, you stamp the box\'s book and use its stamp to stamp your own.  So: clue -> box -> stamp & stamp.  Apparently, this all  urlLink started  in Dartmoor, England when a guy left his business card in a bottle.  Hobbies have a funny way of getting out of hand.  Now there are between 10 and 40 thousand letterboxes hidden in Dartmoor alone and a published catalog listing the clues.  Letterboxers have a scoring code to indicate how many boxes they\'ve found and left.  And there are funny variants of letterboxes like the Hitchhiker box - a letterbox placed inside another letterbox and moved from place to place when it is found.  Recreation is serious business.  '),(10979,832265,'02,August,2004','2004-08-02 00:00:00','   I am ready for the  urlLink future .  '),(10980,832265,'02,August,2004','2004-08-02 00:00:00','   Do you see what happens when you read Dune as Primary Colors.  It\'s like the Bible Code: \"If we want to solve this riddle, we need to find out who is playing the role of the Baron. When I saw that Monkey Mail post today that said Cheney had urged Clark to run, the bells went off. Is this true? Did Cheney urge him to run? Have you heard anything more about this? If he did, then I\'d say we have our fat man. Maybe Cheney\'s undisclosed location is Geidi Prime.\" ( via  urlLink Superelectric  )  '),(10981,832265,'02,August,2004','2004-08-02 00:00:00','   It\'s  urlLink like  Cecil B. Demented, but, you know, real and unterrible: \"Ryan has emailed me about an impending highly secret action he has planned, which he has codenamed Operation Night Hawk. I am invited to join him. He says he cannot tell me much about it for fear of jeopardising its security, but convinces me that, by the end of the President\'s Day holiday weekend, every studio executive in LA will know about Abby Singer. What\'s more, they will each have a trailer on their desk.\"  '),(10982,832265,'02,August,2004','2004-08-02 00:00:00','   The Yanks and O\'s attempted to  urlLink play ball  today in Charm City despite the oncoming hurricane.  It didn\'t work out so well and the game was called in a tie.  Derek Jeter on the situation: \"They\'re canceling everything around here -- schools are canceled, the government left, the Navy\'s pulling out, and the Orioles and Yankees are playing baseball.\"  '),(10983,832265,'02,August,2004','2004-08-02 00:00:00','   The debate still  urlLink rages .  My favorite theory: \"K\" is short for \"Thousand\", as in Kilobyte, and therefore \"K-Rad\" means \"A Thousand Times As Rad\".  '),(10984,832265,'02,August,2004','2004-08-02 00:00:00','   My hometown has seen better days.  First, a school board member called down a  urlLink curse  upon the mayor for allowing a corporate turnaround firm to close down 16 schools ... which, you know, fair enough, sometimes cursing\'s all that\'s left.  But now, they\'ve got  urlLink digital bugles  playing Taps at our military funerals and clearly it\'s only a matter of time until one of them\'s accidentally set to Reveille.  '),(10985,832265,'02,August,2004','2004-08-02 00:00:00','   I\'ve been warning people about the impending demise of the banana for weeks.  Most are dubious in a \"that asteriod\'s not really gonna hit us\" sort of way.  But really, as the Guardian  urlLink explains , it\'s all a matter of time.  The bananas we enjoy today are mutants and susceptible to fungi like yellow Sigatoka.  In fact, such a fungus would have completely wiped out bananas in the 1950s if not for the heroic Cavendish variant.  But now! there\'s a new form of Sigatoka ... black Sigatoka.  And the Cavendish is imperiled.  There is some  urlLink hope  that a new hybrid may save the day.  But only if we get the world\'s best  deep core driller  mutant fruit hybridizer and his ragtag crew of misfits to head on down to Honduras and show those boys at FHIA what real cross-breeding is all about!  '),(10986,832265,'02,August,2004','2004-08-02 00:00:00','   A kid at my Mom\'s school, when asked if he knew what Labor Day meant: \"That\'s the end of the Jeep Grand Cherokee Summer Sale!\"  '),(10987,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink Steve  via voicemail at  urlLink 6:39:54 pm  last night: \"...I\'ll send you email about that, too  - Holy Jesus Christ there\'s an earthquake!  Wow this is a big one.  Uh ... ... yeah .. um, ok, as soon as the earthquake finishes I\'ll give you a call.\"  '),(10988,832265,'02,August,2004','2004-08-02 00:00:00','   The whole  urlLink Omnimax &nbsp; urlLink v. &nbsp urlLink IMAX  thing has confused me for a while.  The Metreon in SF has an IMAX theatre, but it\'s not of the domed, planetarium style.  It\'s more of the \"WHOA! That sucker\'s HUGE!\" style.  After seeing so many movies projected on tiny multiplex screens, it\'s really something to see Matrix Reloaded on a 80\' x 100\' megasurface.  Basically, you end up seeing a lot more detail.  Laurence Fishburne\'s face is quite bumpy, for example.  The side effect is that if a small gnat happens to settle on the projection glass, you end up with a 25\' killer bug stuck on the side of Keanu\'s face.  '),(10989,832265,'02,August,2004','2004-08-02 00:00:00','   Matt Groening  urlLink says  that Fox News threatened to sue over an episode featuring a fake news crawl.  Apparently, the crawl featured such stories as: Democrats cause cancer? Find out at foxnews.com ... Rupert Murdoch: Terrific dancer ... Dow down 5000 points ... Study: 92 per cent of Democrats are gay ... JFK posthumously joins Republican Party ... Oil slicks found to keep seals young, supple. Some of which are almost as funny as \"Network sues self.\"  '),(10990,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink Camille Paglia \'s a nut.  She\'s compelling because she intermixes pop culture and social critique.  And, you know, my generation\'s greatest contribution to the world is obsessive Simpsons-quoting, so I\'m all about stuff like:  What a bunch of crap this Clark boom is. Clark reminds me of Keir Dullea in \"2001: A Space Odyssey\" -- a blank, vacant expression, detached and affectless. There\'s something sexually neutered about Dullea in that film -- a physical passivity necessitated by cramped space travel -- that I also find in Clark.  Sci-fi + sexual identity + politics ... I\'m with ya\'.  But then she\'s gotta go poop all over blogs:  Blog reading for me is like going down to the cellar amid shelves and shelves of musty books that you\'re condemned to turn the pages of. Bad prose, endless reams of bad prose!  While, in the next breath claiming she invented them (fair warning here, Camille ... this is a crowded field):  Now and then one sees the claim that Kausfiles was the first blog. I beg to differ: I happen to feel that my Salon column was the first true blog.  Really her issue is that there\'s no blog super-celebrity.  That someone needs to do for blogs what Rush Limbaugh did for AM radio - make it \"buzzing and vibrant\" with voices that have \"energy and vision.\"  To which I say, first, we already have these  urlLink folks .  And, second, my  urlLink debate team  kicked your butt!  '),(10991,832265,'02,August,2004','2004-08-02 00:00:00','   Netflix will mail you your complete rental history upon request (a partial list is available online).  In 134 days of being a subscriber, I\'ve rented 79 movies.  So, 0.6 movies a day seems a bit ridiculous ... I mean I do leave my house and, like, bathe.  It would be nice to have that list, maybe sorted by ranking, in an XML feed.  I\'m not a huge \"must document everything about me in a sidebar\" person, but I could get behind including that list on my blog.  Or at least a link to the movie most recently seen.  '),(10992,832265,'02,August,2004','2004-08-02 00:00:00','   GI Joe,  urlLink re-contextualized .  ( via  urlLink mempool  )  '),(10993,832265,'02,August,2004','2004-08-02 00:00:00','   The worst type of DVD commentary is when the director decides to narrate the film.  \"He\'s walking down the hall to get to the bedroom.  When he gets there he will sit on the bed ... our set designer is a genius, by the way, completely captured the feeling of the bed.\"  Of course, the exception to this rule is a movie where you find yourself saying \"What the hell were they thinking here - who thought this was going to work?\"  In other words, the Logan\'s Run commentary track.  The costume designer pops up once in a while with such gems as \"Well, this is a future society based on free-love and free-expression.  So we didn\'t really see the need for anyone to wear underwear.\"  But the majority of the justifications come from director Michael Anderson.  For example, the ice cave scene.  Logan 5 and Jessica 69 have just escaped from Hamster City and find themselves moistened and alone in the layer of the evil robot, Box.  This gives rise to one of the great lines in any movie, \"It\'s freezing ... let\'s take our clothes off.\"  Director Anderson explains, \"Well, they were freezing ... and thought their clothes might freeze ... so they decide to get undressed.  And then they wrap themselves in some furs that are lying about ... which god knows how those things got there.\"  Sadly, no deleted scenes on the DVD and according to the commentary there are many.  The Sex Shop scene got cut way down because of censors.  (Anderson on this: \"It\'s a shame because we spent hours applying flesh covered tape\")  And, in a scene cut from the above-mentioned Box lair, the dastardly crapbot tricks our heroes into getting naked (again!) so that he can sculpt them in ice.  Hopefully, there\'ll be a Platinum Edition one day.  '),(10994,832265,'02,August,2004','2004-08-02 00:00:00','   Today\'s my  urlLink brother \'s birthday.  When we were kids, and I was waiting for him to die in Super Mario Bros., I would use the other controller to \'control\' the bad guys.  \"I\'m that Koopa that got you.  Fear my turtle wrath!\"  This may explain why I\'m a Yankees fan today.  Yeah, we\'re the Big Bad of MLB ... ruining the game by buying up everyone and winning so much.  The Evil Empire.  Evil because the Yankees have more money to spend than anyone, an owner willing to spend it, and a lack of any rules to prevent that from happening.  It\'s really great.  When I was sitting around with some Yankees cohorts watching the ALDS a couple weekends back, there was a lot of \"Hey - that guy\'s pretty good.  Maybe we can buy him before the next round.\"  There are some drawbacks.  You get called  urlLink names .  You gotta worry about getting jumped in a Boston sports bar if you cheer too demonstrably.  But still, when your team comes back on Pedro to win Game Seven with a walk-off homer in extra innings, it doesn\'t feel evil at all.  In fact, it feels real good.  It\'s understandable why Red Sox fans would want to experience this for themselves one day.  But the truth is, there\'s solace to be had in the Curse.  I mean, for one, it\'s all the BoSox have.  It\'s not like there\'s moral superiority in having only the second highest payroll in baseball ... selling out the Green Monster in order to grab a few more dollars (Hint: get your own TV network ... very lucrative).  And, second, I don\'t think they\'ve thought about what would happen if they won.  It\'s not like it actually erases all those other botched opportunities.  Now, you\'re just the team who\'s won once in a hundred years rather than this super-mysterious tortured team of doom that\'s haunted throughout the ages by a fat man in pinstripes.  That\'s the thing that suprised me about all the Cubs v. Red Sox hype - how does that work out well for anyone.  All you end up with is a team who\'s Curse wasn\'t as big and hairy as the other guys.  No one wins ... someone just loses more.  So, embrace the Curse, Boston.  It\'s what makes you special.  There\'s also a lot to be said for certainty.  Now, I\'ve got to worry about the team of freaks from Florida and the fact that Mariano Rivera\'s arm may fall off and that we\'ve got no starters on full rest.  The Yankees win a lot, but they still haven\'t figured out how to win all the time.  Maybe if we can buy Pedro.  '),(10995,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink   My friend,  urlLink Eugene , he paints.  A while ago, we decided that he should paint me ... with horns and a black cape.  This past weekend at  urlLink Open Studios , Eugene sold this smaller version of my gourd to,  urlLink Bob Gerbracht , a pastel artist and Eugene\'s teacher\'s teacher.  Live well in Pinole, painting called \"Satan.\"  '),(10996,832265,'02,August,2004','2004-08-02 00:00:00','   Something I never realized about WarGames until I listened to the writer/director commentary ... Joshua\'s voice is a processed and cut-up version of John Wood (Falken) reading the dialogue backwards.  Which is funny because Falken was supposed to be a ornithological-proxy for Stephen Hawking ... wheelchair and all.  The idea got dropped when Hawking was asked how he felt about it.  He said that he would be interested in talking if, in fact, the filmmakers were interested in him because of his work as opposed to his affliction.  Yeah, not so much.  There\'s some discussion about the kinds of things WarGames accurately predicted.  In particular, they take credit for inventing the  urlLink war dailer .  But they couldn\'t understand why having computers talk hasn\'t caught on more.  '),(10997,832265,'02,August,2004','2004-08-02 00:00:00','   From the liner notes to Dear Catastrophe Waitress: There are four towers to choose from though.  Orthanc, Morgul, Tirith and Barar-Dur. I\'m surprised Tolkien didn\'t give the Hobbits a tower as well.  Well they had a sort of mound.  They (sic) are twelve ways you can  urlLink pick  two towers from four, right? In related news, in listening to the Themselves remix album, I noticed there\'s this slowed down sample in Home Work that goes, \"He\'s gone back on his old job - hauling sand.  No!  He\'s working at the office, that\'s right.\"  It\'s a quote from Florence Rasmussen, the Sweet Ol\' Grandma from Hell, who shows up in the middle of Errol Morris\' first film, Gates of Heaven.  Soon I will have all the facts I need.  '),(10998,832265,'02,August,2004','2004-08-02 00:00:00','   Thought I would work from home today - just a chance to get away from the bustle of the office.  Instead of quiet, I got Blue Angels practicing aerial manuvers over my house.  Turns out F/A-18\'s are real fucking loud.  '),(10999,832265,'02,August,2004','2004-08-02 00:00:00','   Sometimes times are rough.  Rough like you vote for a terrible politician in order to avoid electing a morally  urlLink rephrensible  one.  And then the reprehensible one wins anyway.  Rough like your team loses the first game of the ALCS after getting baffled by the ol\' knuckler.  And your shortstop gets attacked by a  urlLink bald eagle  to boot.  But then you find out that your local  urlLink record store , one of the great record stores anywhere, actually does have  urlLink Themselves \' remix album,  urlLink The No Music of Aiffs .  And it\'s not only great, but has a sweet little video clip as well.  Things are looking up.  '),(11000,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink Europe vs. Bush : Europeans are waiting for the Americans to give up on Iraq and come back to their senses, so U.S.-European relations can get back to the intimacy of Cold War days. ... Both the Bush administration, which has been overly dismissive of other nations, and its guerrilla critics need to remember Rule One of crisis behavior: When you are in a hole, stop digging.  '),(11001,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink Baysball: a world of heartache : \"The Bay Area\'s baseball dreams are in a skid row pawnshop today, in a grimy glass display case between a banjo and someone\'s mother\'s wedding ring.\"  '),(11002,832265,'02,August,2004','2004-08-02 00:00:00','   When I came home this evening, I had messages on my machine from a DNC official, Bill Clinton, Barbara Boxer, Bill Clinton a second time and Joe Lieberman.  You know, today was Yom Kippur and while I ate lunch like normal, I\'m guessing that Holy Joe didn\'t - I\'m surprised I couldn\'t find a press release on this, actually.  My point is that 38 years ago to the day, Sandy Koufax refused to pitch in the first game of the World Series because it fell on Yom Kippur.  Meanwhile, I\'ve got auto-Joe on my machine.   '),(11003,832265,'02,August,2004','2004-08-02 00:00:00','   Jinxes are an interesting thing.  I had predicted that this weekend the Red Sox would see their hopes dashed and was proved, infuriatingly, wrong.  I blame myself.  Still I had the good luck to be able to watch the game with a bunch of old college friends - cursing the TV and  urlLink Joe Morgan .  The game was cursed because of the A\'s penchant for squandering every opportunity to put away the BoSox.  Joe Morgan was cursed because ... well, largely just out of tradition.  In post-seasons past, we got into a bad habit of throwing soda cans during Morgan- & Miller-commentated games.  The A\'s have lost 8 straight games in which they\'ve had the opportunity to clinch a post-season series.  Here\'s hoping their luck will change tonight.  '),(11004,832265,'02,August,2004','2004-08-02 00:00:00','   I\'m experiencing insomnia for the first time this weekend in Boston.  It\'s just like being awake all the time.  I feel I could be a much more productive, yet insane, person if this were to keep up for a few more days.  Despite that, I had an excellent time at Day One of BloggerCon.  Great panel on education filled with passionate folks and new ideas about how to make blogs work in schools.  And not to be too aw-shuckish, but it\'s been a real treat meeting folks who say \"I use  urlLink Blogger  - thanks so much\"  and then getting to hear their ideas on what we should do next.  Tomorrow morning\'s user get-together should be a good time assuming I get any sleep between now and then.  Where\'s some valerian root when you need it.  '),(11005,832265,'02,August,2004','2004-08-02 00:00:00','   Headed out to Boston this weekend for  urlLink BloggerCon .  Of course, I\'ll also be on hand for what is likely to be another sad ending to a BoSox post-season.  So, I\'ll be looking to capture Boston\'s collective woe in tiny picture  urlLink format .  '),(11006,832265,'02,August,2004','2004-08-02 00:00:00','   Saw a  urlLink documentary  on Peter Sellers which completely skipped over Lolita, but had great info on Dr. Strangelove.  In addition to learning that Sellers considered Strangelove his favorite character, I also learned that he was meant to play a fourth role in the movie - Major Kong.  As you can read  urlLink elsewhere , Sellers \'conveniently\' broke his leg and Kubrick had to cast Slim Pickens instead.     '),(11007,832265,'02,August,2004','2004-08-02 00:00:00','   In Florida, Wal-Mart had a \'blitz\' sale where they sold DVD players for $30 on the day after Thanksgiving.  When the doors opened at 6am, a woman was  urlLink trampled  by the crowd and is still hospitalized.  On Friday, I was in downtown San Francisco, about a block away from Union Street, in a crush of a thousand people swarming up and down Powell Street.  It was a Castro at Halloween type crowd - intoxicated and frenzied.  Families of four or five people were struggling to stay together and would call out the names of their favorite stores they pushed past.  '),(11008,832265,'02,August,2004','2004-08-02 00:00:00','   Mathworld has this kick-ass  urlLink page  on special numbers and number theory.  Maybe that wasn\'t the greatest sell.  But how great is it that we\'ve got folks who can spend time finding and naming special numbers like the  urlLink Vampire Number .  A vampire number is one with an even number of digits that is the product of a combination of its digits.  So, 1260 is a vampire number because 21 x 60 = 1260.  Here\'s the great part.  If a number is a vampire number, the pair of numbers multiplied together to form it are called its fangs.  Fangs!  Mathematicians are wacky.  They\'ve also got some fascinating info about the  urlLink Beast Number , 666.  I\'d never before heard the theory that the reason it\'s 666 is because in roman numerals it contains one of each symbol (sans M): DCLXVI.  I\'m a believer!   '),(11009,832265,'02,August,2004','2004-08-02 00:00:00','   Steve Yuhan is  urlLink blogging  again!  I\'m hoping he\'s finally broken through that barrier at law school where you actually have free time before graduating and starting the 80 hour weeks.  Steve\'s gonna be working at PaulWeiss next year so maybe he can get to the bottom of the infamous  urlLink sushi   urlLink memo  when he gets there.  '),(11010,832265,'02,August,2004','2004-08-02 00:00:00','   In the PS2 game, Ico, you\'re a behorned kid that must escort a princess through a giant castle.  It\'s the best game ever made.  Recently, I played Prince of Persia: Sands of Time which reminded me a lot of Ico as you\'re a very athletic kid who occasionally escorts a princess through a giant castle, engaging in acrobatic ass whoopin\' along the way.  The acrobatics are particularly nice - lots of swinging, wall running, Ninja Gaiden-style wall-hopping.  But, the whooping of ass is a little formulaic.  The baddies are slow and telegraph their moves in a \"Grr!  I\'m swinging this hammer around my head cause soon I\'m gonna try and hit you\" sort of way.  The first time you run up a sand troll\'s belly to slo-mo flip over his head and slice him up on the way down, it\'s impressive.  When you realize that that\'s all you need to do to beat 80% of the bad guys, it\'s a bit tedious.  Prince of Persia\'s level design is expansive and beautiful - doing flips while balancing atop the Ultimate Tower of Doom at sunset is breathtaking in both a pretty and a \"that\'s gonna hurt\" way.  And they do the little stuff well, like leaving trails of dust behind you after you wall scramble and doing clever things to avoid the feeling of load times.  Finally, you can  urlLink unlock  a fully-playable version of the first Prince of Persia game.  Which certainly is cool, but man did I quickly grow frustrated.  I don\'t know what we were all doing before analog controls.  I ended up humping the wall repeatedly because I\'d overrun the place where I needed to jump - something I remember being endlessly amusing when I first played the game on my Mac SE.  Less fun now.  '),(11011,832265,'02,August,2004','2004-08-02 00:00:00','   Today I sold the \'89 Camry that I had shipped out to California back in  urlLink May .  A car with 150K miles on it wasn\'t up to the daily grind of a 70 mile round-trip commute.  It\'s brakes needed to be replaced, the muffler was shot, it needed a new radiator and the driver\'s side door lock had been turned into a gaping hole by folks looking to steal the fat load of nothing I kept in my car. So, farewell. I\'m happy that I was able to unload it without any hassle and that its new owner is a grizzled hippie mechanic who will make the necessary repairs and turn the car over to a young cousin just learning to drive.  '),(11012,832265,'02,August,2004','2004-08-02 00:00:00','   MGM  urlLink announced  that they\'re planning a 2004 spoof of Lord of the Rings and Harry Potter called Henry Bates and the Sorcerer\'s Balls.  Heard about this from  urlLink Josh Cagan  who managed to get more jokes out of the announcement than there will be in the entire movie.  Good work.  '),(11013,832265,'02,August,2004','2004-08-02 00:00:00','   My former debate partner, Anna, is  urlLink blogging  her little heart out.  It\'s quite the treat when friends you don\'t get to see very often start posting bits of themselves to the interweb.  I feel there\'s a future in this business.  Regarding her  urlLink question  about whether there\'s any food that\'s better prepared on the East Coast, the answer is pizza.  Hasn\'t been a good pizza in California since I flew back from  urlLink visting  Jonny B in Chicago with a chunk of a thick-crust stuck in my teeth.  '),(11014,832265,'02,August,2004','2004-08-02 00:00:00','   Biz has made it across the country, all the way to Mountain View.  I didn\'t realize that along the way he left an audblog  urlLink post  from Iowa in which he bagged on 2 Mid-Western institutions.  First off,  urlLink Steak n\' Shake  is incredible.  I mean sure, not a safe haven for vegans given the foodstuffs it\'s named after.  But trying to impale the maraschino cherry at the bottom of an orange freeze is one of my fonder childhood challenges.  Second, the convenience store named Kum n\' Go.  Okay ... this really isn\'t a Mid-West institution.  I only ran across one once when visiting  urlLink Matt  in Kirksville.  But its name and the fact that the locals refered to it as the Ejaculate n\' Evacuate captures both the irony-free superficiality of the Mid-West and its underlying sad wit soul.  Ah, home.  Update : My mom tells me that in her day they refered to Steak n\' Shake as Stoke n\' Choke.  I\'m trying not to think about what this phrase meant to the youth with whom my mom parked back in said day.  '),(11015,832265,'02,August,2004','2004-08-02 00:00:00','   According to the  urlLink IHT , the director of Brazilian staging of Tristan und Isolde mooned the audience when he was booed on opening night.  His opera seems fairly ridiculous ... a woman masturbates during the overture, there\'s a chorus of Hasidic Jews - it sounds like Springtime for Hitler, but for real.  So, you know, maybe booing was appropriate.  What\'s wacky is that he\'d now being tried on indecent exposure charges ... in Brazil!  Where, I\'ve been told, the whole country\'s indecently exposed.  What\'s next?  Folks getting charged for public drunkeness in Scotland?  Excessive cheese-eating in France?  Crude national stereotyping in America?!  '),(11016,832265,'02,August,2004','2004-08-02 00:00:00','   Jason Sutter gave me the Goonies DVD for my birthday and I\'d been trying to hold off watching the commentary, but finally succumbed to temptation this past weekend when I found myself uncharacteristically bereft of fresh Netflix.  The DVD does not disappoint.  The commentary features all of the original kids ... even Short Round from Temple of Doom.  What\'s more, they videotaped the commentary session so the movie will occassionally go into picture-in-picture mode so you can bathe in all of Sean Astin\'s plumpness and Corey Feldman\'s assholic irrelevance.  Sean Astin\'s the big highlight here.  Apparently, he\'s not well liked.  At all.  Repeatedly he will start to tell a story - usually one that starts off very promisingly such as \"Cyndi Lauper, there\'s something I\'ve always wanted to tell you\" - only to be cut-off and trounced by someone else in the cast.   People just talk over him and if he persists they\'ll throw in some kinda Rudy or Samwise Gamgee jibe to really shut him up.  Even Jeff \"Chunk\" Cohen gets in on the action.  This is a guy who never appeared in movies again, went on to win student body president on a \"Vote for Chunk\" campaign and is now a lawyer of some kind.  He gets to beat up on the Son of Gomez.  I think it has something to do with Sean Astin being the only commercially successful actor left from the cast.  There\'s a lot of \"in this scene, you can almost hear Sean thinking about what this movie means for his career.\"  Which really makes not a ton of sense as before he landed the role of Frodo\'s manfriend, he most recently portrayed Hercules in the video game Kingdom Hearts.  Eventually all the abuse proves too much for the omega hobbit and he bails on the commentary session well before his dramatic speech at the bottom of the Wishing Well.  The man quits the commentary!  This does not lessen the abuse.  Also on the DVD is the epic, two-part music video for Cyndi Lauper\'s theme song, Good Enough.  It picks up where Girls Just Wanna Have fun left off ... that is with Captain Lou Albano.  The whole thing last twelve minutes and ends with Andre the Giant scaring off the Iron Sheik and Rowdy Roddy Piper who play both pirates and evil foreign investors.  It\'s all very confusing.  '),(11017,832265,'02,August,2004','2004-08-02 00:00:00','   So, I\'ve got an obsession involving a chicken.  And a sheep.  And a pregnant cow.  Every day, I am compelled to gather the eggs, pet my sheep and feed my knocked-up cow.  My name is Jason and I am a  urlLink Harvest Moon  addict.  Harvest Moon: Friends of Mineral Town for the Game Boy Advance, to be specific.  One of the key aspects that makes this farming sim so consuming is that each day you only have a certain amount of time and strength to make things happen for your pixelated livestock.  You\'ve got to make your animal friends happy (so that they\'ll produce higher quality eggs, wool and milk) while simultaneously working on your 5 Year Plan for Agrarian Life Fulfillment.  Sitting atop the Harvest Moon pyramid of needs, above even Win the Sumo Chicken Contest or Teach Your Dog to Catch a Frisbee, lies Getting Hitched.  And there are many fine ladies from which to choose.  Initially I was courting Popuri, the chicken farmer\'s daughter, because, not to be crass, she seemed kinda easy.  All she ever wanted was flowers and those were easy to come by.  But then she showed up at my house looking for treats on Halloween and I was all \"First off, this is a holiday for kids and I\'m not looking for no child bride.  And second, we are so not in the point in our relationship where you can just show up at my house, when I\'m trying to tend to my crop of eggplants, and demand that I provide you with snacks.\"  So now I\'ve got my eye on the tomboy-ish barmaid.  I hear she likes eggs.  '),(11018,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink Artist Statement : \"My artwork bears witness to the journey of the puzzle to the journey.\" with helpful decoder ring.  '),(11019,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink Boingboing  linked to  urlLink this site  featuring a CD of Christmas songs, played entirely with samples from 8-bit video games.  But I think the real magnum opus here is Nullsleep\'s  urlLink Depeche Mode Megamix .  Enjoy the Silence never sounded so good as when played in the style of the Rygar theme music.  '),(11020,832265,'02,August,2004','2004-08-02 00:00:00','   Finding Nemo does all things I\'ve come to expect from a great Pixar flick: funny characters with entertaining celebrity voices, lots of meta-pop culture jokes, a bunch of anthropomorphic/\"Sharks don\'t really have 12 step programs\" gags.  Nemo is the most impressive of their stable because it also looks pretty as all get out.  The design philosophy is all about density of information - packing in a crazy number of plants and animals into every under-seascape.  There\'s so much stuff going on that the background is frequently more compelling, from a dramatic standpoint, than the narrative.  Which is why it\'s super-hot that they built in a DVD feature that allows you to soak in the digital coolness.   Most DVDs boast a \"interactive animated menu\" feature - which is basically one step beyond simply saying \"contains movie\" - but Finding Nemo repurposes the idea.  Each menu of the DVD contains an animated background and a button that allows you to drop out the titles and main soundtrack.  So you\'re left with just the infinite loop of, say, sea anemones gently swaying.  Kick ass!  '),(11021,832265,'02,August,2004','2004-08-02 00:00:00','   First, the  urlLink power  goes out for a big chunk of San Francisco.  And now we\'ve got  urlLink earthquakes , what with the ground moving and the shaking and such.  This is my second quake ... it was like being on a boat, gently rocking.  Afterwards I felt landsick; appreciably dizzy and sort of sick to my stomach.  Not to be excessively naive, but this is what makes earthquakes really interesting and sort of personally terrifying.  I have this expectation from the ground that it\'s not going move and 99.99% of my life that\'s been a fair assumption.  And then it\'s not true.  In a very significant way.  It\'s as if gravity were to be dimmed down a bit for a few seconds - just a little bit of lightness, like going over the hump on a rollercoaster, but in a way that you could tell was affecting people for miles around, not just you in your rickety car.  It\'s the inescapable bigness of the thing that\'s amazing rather than just the thing itself.  '),(11022,832265,'02,August,2004','2004-08-02 00:00:00','   Jonathan Dee\'s NYT  urlLink story  on video games with whole pages of entertaining quotes from Atari CEO, Bruno Bonnell ... from which  urlLink Costik  tees off: Bonnell wants to call games \"interactive entertainment\": this is a common rhetorical strategy, but it always gets up my nose ...  You can see this strategy everywhere: It\'s gaming, not gambling. It\'s erotica, not porn. It\'s speculative fiction, not sci fi. To which the only honest response is: Bullshit.    '),(11023,832265,'02,August,2004','2004-08-02 00:00:00','   Sutter likes to remind folks that I used to carry around a yo-yo in pocket.  Pretty dorky.  But now I\'ve upgraded to  urlLink Astrojax  - the Swiss yo-yo of the future.  And it\'s only a matter of time until the ladies go crazy for my advanced three body technique.  There\'s some hot stuff on the Astrojax website.   urlLink Physics equations  to explain how the whole thing works.  And ridiculous  urlLink video  of Swiss kids with dreadlocks working the astro magic.  '),(11024,832265,'02,August,2004','2004-08-02 00:00:00','   The Believer is based loosely on the true story of a Jewish neo-Nazi.  In the film, Danny Balint is an incredibly articulate skinhead whose arguments in favor of anti-semitism and murdering Jews are based in his childhood yeshiva studies.  And also, he gets to bone Summer Phoenix.  There\'s a pivotal scene where he and his bald droogs vandalize a synagogue - spray painting swastikas, tearing up prayer books, peein\' on stuff ... the usual.  But Danny can\'t go through with descrecating a torah.  Unsurprisingly, the gentile neo-nazis aren\'t exactly won over by being told about the special sacredness of the torah and Danny ends up taking home the torn up torah in order to repair it.  Which, you know, is like a symbol for his return to judaism.  This is a great movie, what with the self-struggle and the contradtions and the unsympathetic hero.  Good. So, of course, it\'s gonna be a rabbi at the Simon Wiesenthal Center who  urlLink stops it  from seeing wide distribution.  The scene I described is the one that gave him pause because it\'s \"a primer for anti-semitism.\"  Or, you know, a scene about challenging anti-semitism ... depending on which movie you saw.  Also, it\'s interesting to note that the director (who was a non-practicing Jew that became re-enjewed because of making this movie) went to all sorts of lengths to mitigate his guilt around showing the torah being torn.  For example, they found a section of the torah that didn\'t include the name of God and offset that to make a whole scroll\'s worth.  '),(11025,832265,'02,August,2004','2004-08-02 00:00:00','   My brother  urlLink informs  me that the lady who got trampled at Wal-Mart not only has long  urlLink history  of personal injury claims, but, in fact, has 16 such claims against Wal-Mart alone.  It\'s nice to see that, in America, you pick yourself up by your bootstraps and go from being the victim of greed to the perpetrator thereof.  Next: We find out that there\'s security video footage of her gettin\' busy with a stock boy in the Wal-Mart bathroom ... only to discover a week later that she leaked the video to the media to get a better deal for her Movie of the Week rights.  '),(11026,832265,'02,August,2004','2004-08-02 00:00:00','   According to  urlLink this article  from the NYT, NASA\'s Mars rover  urlLink site  has received 32 million visitors in the past three weeks.  The thing I\'ve been most impressed about with the rover missions is the amount of information NASA has made available.  They\'re seemingly posting every image the rovers capture and are even including  urlLink screenshots  to show the controller-eye view of the rollout on Mars.  What\'s more they compliment the raw data aspect with a bunch of well put together videos on the technical aspects of the mission.  For example, the cg  urlLink animations  of the decent and landing stuff really breaks down just how amazingly difficult it is to land on another planet.  The NYT piece rightly points out that this openness has helped folks (like me) feel a greater connection to the rovers.   And by connection, I mean obsession: The flow of updates from Mars has led to a new addiction among some Americans, who find themselves checking for news dozens of times a day. In interviews, many reported feeling depression and withdrawal symptoms when the Spirit stopped communicating last week. That NASA kept up the information flow even while suffering serious setbacks is really a testament to how much they want to share the science.  And now that there are two functional rovers on Mars, there should be twice as many updates, right ... right?!  '),(11027,832265,'02,August,2004','2004-08-02 00:00:00','   First, there was the local radio station who introduced Nine Inch Nails\' \"Closer\" as \'classic alternative.\'  Now, it\'s the fact that I\'m developing a serious Sunday routine that involves rebalding myself, drinking chai tea and reading Harpers.  In the latest issue, there\'s a great piece by Carter Burwell (who\'s the composer for all of the Coen brother\'s movies) about the music written for the news.  Ultimately, it\'s about how the networks chose a different type of music for the latest war in Iraq than they did during the first.  Peter Fish, the composer for CBS, is quoted as saying: \"The second time it was more like they were trying to promote the war the same way they would promote Terminator 3 - it was like \"Battle of the Megaheroes.\"  So the first time what I delivered was vaguely militaristic ... the second time it was just Techno-Ali vs. Frazier-IV, we\'re-going-to-knock-the-crap-out-of-them music.\" There\'s also a great excerpt from Don DeLillo that was part of a Lee Harvey Oswald documentary.  I\'m definitely going to read his book Libra soon. \"If there was a plot, it was small, crude, and largely improvised - not the master plan that would allegedly balance the loss of the president.  Our state in the world, the fact that we are human, is the only element the equation needs in order to be balanced.  We\'re able to think into the stars, imagine alternative lives for ourselves, and there are times when we feel equal, some of us, to the vast social reality around us.  What else would make a man decide he might run for president?\"  '),(11028,832265,'02,August,2004','2004-08-02 00:00:00','   I\'m watching live coverage on my laptop, over wireless, of the Opportunity rover\'s  urlLink landing on Mars .  And incidentally, at work we occassionally have to do a reboot on a production machine - not that this is completely out of the ordinary or particularly serious, but, you know, it\'s always possible the machine\'s not coming back.  If it doesn\'t, you have to get somebody to go to the cage and physically restart the bastard and hope that it hasn\'t gone dark forever.  All of this can be stressful but it\'s not as if the computer in question is on another frickin\' planet.  It\'s damn impressive that the folks at JPL were able to re-establish contact with Spirit.  This whole deal is truly geekery in its highest form.  '),(11029,832265,'02,August,2004','2004-08-02 00:00:00','   Somehow, Anna gets paid to research terrorists who used to be c-movie actors.  Fortunately, she  urlLink shares :  Rooster: Spurs of Death! : Stoke and his son Wyatt take their prize roosters to a cock fight, only to find deceit and cheating by a competitor has forced them to lose their winnings. On the way, Wyatt learns about love from a runaway hooker (Kristine DeBell) and that maybe the \'cocker\' trade is not right for him.  '),(11030,832265,'02,August,2004','2004-08-02 00:00:00','   Art Garfunkel faces a $100 fine after being busted for pot possession in New York.  It\'s a sad state of affairs when the War on Drugs claims Garfunkel and Tommy Chong as victims.  On the plus side, copy editors around the country are having a  urlLink field day .  Some  high lights:  Garfunkel\'s Sounds of Sirens Garfunkel feeling less groovy after drug bust Garfunkel not so groovy Garfunkel feelin\' groovy Garfunkel feelin\' not so groovy Parsley, Sage, Rosemary And ... Pot: Art Garfunkel Busted What, no \"Garfunkel\'s Bridge is Over Trouble Water\" or \"Garfunkel: \'I am a Rock\' Fiend ... I mean, it\'s just Pot\'\" or \"Garfunkel: I wish I was homeward bound ... and high!\" or \"Garfunkel: April come she will ... and sell me a dime bag\" or \"Garfunkel says: pottity pottity pot pot - weed.\"  '),(11031,832265,'02,August,2004','2004-08-02 00:00:00','   The worst thing I ever lost was a one-foot headphone extension cable.  Several years ago, I bought a pair of Sony egg-style  urlLink headphones  and they are still my favorite pair.  The fact that they continue to have flawless sound is amazing as they\'ve been kicked around in my backpack every day for over 3 years.  What\'s more I prefer them to my more expensive studio headphones which, really, was just showing off.  Anyway, they\'re made for the MD market, so they come with a short cord to fit nicely with the wired remotes you find on nearly all MD players.  But they also shipped with a short extension cord, just in case you were using them on a remoteless device, like a laptop.  Which I did.  And then I forgot to take out the extension cord before turning in my laptop at my last job.  And since then I\'ve been frustrated by cables.  Nowhere seems to sell the short cord, so I had to get some six-foot monster with extra-gnarled sheathing.  The looks you get from airplane seat neighbors when you pull out a gordian knot of goofycord are none too kind.  '),(11032,832265,'02,August,2004','2004-08-02 00:00:00','   Pounds is abbreviated  urlLink lbs.  because of the Roman weight libra (which was like 3/4 of a pound, but whatever).  '),(11033,832265,'02,August,2004','2004-08-02 00:00:00','   In the slow-burning chemical process that controls when I do my laundry, clean underwear is the limiting reagent.  You just can\'t go to long without clean drawers.  The specific trigger is when I reach the last pair, which is always the same pair as they\'re my least favorite.  They\'re cotton shorts I got during orientation week when I was a freshman.  They really should be boxers, so I don\'t feel bad using them as such.  Whether they should still be my omega undergarment after 9 years is debatable.  The real problem is that since they\'re shorts, they have pockets - which is wacky but harmless.  But they\'ve also got no fly ... which is a nightmare.  Also, there\'s a lag between when I receive the \"nothing but flyless shorts left\" signal and the time I actually have clean laundry again.  So my awkward time at the urinal can go on longer than it should.  Nothing untoward, mind you ... just a couple days.  I\'m a very hygenic person normally.  Really.  Anyway, all I\'m saying is that I\'ve got a dresser stacked high with clean underwear, and it feels good.   '),(11034,832265,'02,August,2004','2004-08-02 00:00:00','   There\'s a scene in the restored version of Spartacus where Laurence Olivier attempts to seduce his recently acquired bath slave, Tony Curtis.  Yeah, hard to see how that one didn\'t make it in the original 1960 release.  I hadn\'t seen Spartacus before but I\'d heard about the scene in Celluloid Closet. Tony Curtis mentions that his character recoils and ultimately runs away when he realizes that Olivier\'s talk about oysters and snails is really none-to-subtle code for \"Did I mention I\'m a Roman played by Olivier and you\'re my bath slave.\"  In the documentary, Curtis said of the scene \"\'Take me out to dinner first ... show me a good time.  Don\'t get me in the bath and drop the soap.\'\"  What I didn\'t know was that when the scene was put back in for the 1991 release that Tony Curtis had to redub the dialogue.  And what\'s more Olivier was dead so they had Anthony Hopkins read his lines!  It\'s odd.  It\'s as though, in 20 years, they were going to re-release The Dead Zone and had Kevin Spacey fill in for the Christopher Walken part.  That\'s assuming that  urlLink Biz Stone  had better things to do.  '),(11035,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink Google / urlLink Ev  bought me a new 15\" PowerBook and I\'m happy as a dork with a new laptop.  I\'ve still been running Jaguar on my iMac and Panther is a noticeable improvement - things are snappy (like the new Finder) and Expose is a genius feature.  I\'m not sure how I feel about the option to have Expose keyed off of hot corners.  Ever since AfterDark I\'ve had a hard time with hot corners ... I forget which ones they\'re supposed to be.  I love the fact that OmniOutliner now comes with the OS.  I\'d believed in my heart that outliners were really important, but I\'d not gotten behind them in practice.  Now I understand the fuss.  I think it\'s the psychology of having bullet points - the structure alone reinforces the notion that you\'re working stuff out and getting stuff done.  The only thing I\'m not wild about is the trackpad.  Musn\'t it be the case that user-testing would show the eraserhead mouse as the preferable laptop solution?  Plus you get laptop nipple jokes for free.  Also, I get scared at the rapidity with which the CD/DVD drive sucks up the discs.  You\'re pushing it in and it\'s like \"Well, it still hasn\'t caught ... still not catching ... man it\'s in pretty far, I wonder if it\'s - Holy Crap!\"  But I\'m willing to say that\'s just me.  Oh, and Apple should totally buy Objective Development, the people behind  urlLink LaunchBar .  It\'s a short circuit for the whole OS ... wait, maybe that sounds like a bad thing.  '),(11036,832265,'02,August,2004','2004-08-02 00:00:00','   The Professor and the Madman by Simon Winchester tells the tale of how a schizophrenic Civil War veteran helped write the Oxford English Dictionary.  The story behind the hook is quite simple to tell and really could have been covered in a decent magazine article.  So you end up with a lot of padding regarding the deterioriation of Dr. CrazyPants, the Mad-Speller of Connecticut (not the madman\'s actual name).  But you also get a lot of fun facts regarding the creation of the OED.  There\'s lots of love heaped on the OED.  It\'s treated like a literary cathedral that took 70 years to create and is frequently called the most important work ever printed.  I\'m cool with the love - the OED is certainly one of those pinnacles of human obsessive-complusiveness: \"Not just every word ever spoken - but the original quotation for those words as well!\"  Winchester also has some good stuff on subjectivity in dictionaries; how the tone of the entries and choice of what to include are these fundamental choices by individuals going about the business of socially constructing language.  But that commentary is somewhat beside the point in this book and nowhere near as interesting as David Foster Wallace\'s  urlLink essay  on a similar topic.  On the plus side, there\'s also lots of fun facts about words and medical trivia.  For example: The surgical removal of the penis is at the best of times a dangerous practice, rarely performed even by doctors: An attack by the renowned Brazilian fishlet known as  candiru , which likes to swim up a man\'s urine stream and lodge in the urethra with a ring of retrorse spines preventing its removal, is one of the very rare circumstances in which doctor (sic) will perform the operation, known as peotomy. Note to self: never pee in Brazil.  '),(11037,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink   My family came out to visit me in San Francisco during Christmas.  It\'s like going home for the holidays, but I didn\'t have to go anywhere.  The non-trip was a blast, and I had an especially great time hanging out with my brother.  Because we don\'t see each other for long stretches at a time, there\'s a lot I forget - specifically how much fun it is to have a brother.  First off, little brothers are just so fucking game for anything.  Do you know how I got David to pose for this photo?  I said, \"Look, there\'s some big old nasty seaweed ... pick it up and pretend to eat it - I want to post pictures of that on the Internet.\"  Awesome.  Also, whenever we hang out it\'s like a giant dork braindump.  \"Here\'s every game, scifi show and geeky thing I\'ve seen since last we met.\"  My friends pointed out how similar we are in appearance and, like, vocal mannerisms.  Obviously, this is to be expected.  As Sutter said, \"Is he not the person you\'ve shared a bedroom with the longest?\"  It kicks ass.  There\'s like this other me living in Providence but with special skills.  Example: Me: Hey, Eugene.  Did you know my brother can sing the music from any video game he has ever played? Eugene: Okay, how \'bout Rygar? David: The first screen or the opening level?  Good work, Mom.  You done raised us right.  '),(11038,832265,'02,August,2004','2004-08-02 00:00:00','   In the commentary for Fear and Loathing in Las Vegas, you get some of the dirt as to why the original director, Alex Cox, was dumped from the project.  The producer, Laila Nabulsi talks about how there were creative differences and that it had something to do with the inclusion of the Wave Speech from the book.  That\'s the one that goes: We had all the momentum; we were riding the crest of a high and beautiful wave. So now, less than five years later, you can go up on a steep hill in Las Vegas and look West, and with the right kind of eyes you can almost see the high-water mark---the place where the wave finally broke and rolled back. In the new documentary,  urlLink Breakfast with Hunter , you get a much more detailed picture of what went down.  Filmmaker Wayne Ewing is there at Owl Farm when Alex Cox and co-writer Tod Davies come to visit to talk about the script ... which Hunter\'s not read.  So he\'s caught flat-footed when he finds out that their idea for the Wave is to show an animated version of Raoul Duke being carried back to Las Vegas after he tries to flee to LA at the midpoint of the story.  Hunter\'s first reaction is to threaten to sue the filmmakers.  Then, he rants about how he hates animation and how this is simply turning his writing (and himself) into a gross characiture.   He goes on to say that this is the best writing he\'s ever done, that it\'s a giant fucking sapphire and they\'ve turned into into a dumb fucking cartoon.  As Alex Cox tries to explain how it\'s all motivated by the Ralph Steadman illustrations, Tod Davies actually starts weeping because she feels her adaptation of the material is also a giant fucking sapphire.  It doesn\'t make any difference.  Which it shouldn\'t because it\'s a really stupid idea.  In addition to reducing the text to a plot device, it\'s also a flawed metaphor.  The point is that the wave of Sixities Freakitude broke before it engulfed America and the world in its righteous patchouli-scented energy.  The speech is Hunter\'s lament for the failed movement and is great because it evokes the tweaked-out isolation and madness he feels being trapped in the American Nightmare of Las Vegas.  As he says, the wave broke and rolled back - it\'s not gonna carry him anywhere.  Finally, Hunter threatens to kick the pair out and they take it upon themselves to skedaddle.  At this point, the documentary cuts to Hunter, Laila and Benicio del Toro watching the footage we just saw and figuring out that they\'ll need to get a new director.  Which, thankfully, they do.  '),(11039,832265,'02,August,2004','2004-08-02 00:00:00','   ... unless you\'re a black dude looking for a job.  NYT  urlLink reports : 51.8 percent of black men ages 16 to 64 held jobs in New York City in 2003. The rate for white men was 75.7 percent; for Hispanic men, 65.7; and for black women, 57.1. The employment-population ratio for black men was the lowest for the period Mr. Levitan has studied, which goes back to 1979. The cited report looks at the employment to population ratio as opposed to the more traditional unemployment rate.  With the former, you capture everyone who isn\'t working regardless of whether or not they actually looked for a job.  So, this is a number that conservatives would hate because they\'d claim \"A whole bunch of folks probably don\'t even want to work else they\'d pull themselves up by their bootstraps in this great land of ours.\"  The article contains a baked-in response to this implicit argument. \"If job losses land disproportionately on one group of people, a disproportionate share of that group may give up looking for work. In that case, changes in the unemployment rate for that group will tend to understate the relative impact of the recession on that group\" This seems true.  It makes sense to think about aggregate, cultural effects, rather than just \"This one person didn\'t look for a job because of his own failed internal motivation.\"  '),(11040,832265,'02,August,2004','2004-08-02 00:00:00','   See, here\'s a great  urlLink example  of rant from Rep. John Dingell (and courtesy of  urlLink Steve Yuhan ).  If you\'re a congressman from Michigan, you\'ve got to be a pretty salty dog with respect to how blue collar workers have fared during the current administration / past 50 years.  And yet, Rep. Dingell was able to take a ludicrous decision by the Council of Economic Advisers to class fast food jobs as manufacturing positions and turn it into a funny (but pointed) screed: I am sure the 163,000 factory workers who have lost their jobs in Michigan will find it heartening to know that a world of opportunity awaits them in high growth manufacturing careers like spatula operator, napkin restocking, and lunch tray removal. Plus, all references to defunct McDonaldland characters should be applauded when made by elected officials.  '),(11041,832265,'02,August,2004','2004-08-02 00:00:00','   I\'m not sure why but I\'ve been coming across a lot of rants recently.  In the blog world, it\'s mostly \"I have discovered this deal and now I must vent my venom upon it.\"  A lot are funny but there are a couple of rhetorical traps that these rants typically fall into.  First off, more is not better.  Having a giant ordered list of reasons isn\'t usually effective.  Often it\'s numbing and you end up padding the list with trivial bits.  For example, \"And why are people still using drop caps!  It\'s a webpage not an illuminated manuscript, you degenerate hack!\"  That actually leads into the next common pitfall - self-righteousness.  I see a lot of political blogger taking pride in enumerating the logical fallacies in the arguments of their opponents.  Fallacies are fun to point out ... they\'ve got cool latin names like reductio.  But too much of it just makes you look like an pompous ass.  As does writing stuff like \"And to you I say, \'Fine! Be happy with your bowl of gruel, spoonfed to you by the masters of government.  But I deserve a better fate.  One I shape myself ... partially by writing Randian fanfiction.\"  Well, maybe, that would work.  Self-deprecation is the great salve.  Finally, have a frickin\' point.  Argue in favor of something.  Point to a better way you actually believe in.  Fear of sincerity\'s a big killer and certainly one I struggle with.  But, as the man says, you gotta put your deal out there in order to feel the wind.  Oh yeah.  The &lt;/rant&gt; joke ... not funny!  '),(11042,832265,'02,August,2004','2004-08-02 00:00:00','   Today is Bagel Day at work (the day on which we get bagels as opposed to dress up like them).  While walking downstairs to claim mine, I caught a whiff of microwave popcorn being popped.  And that\'s when it hit me.  Popcorn-flavored bagel schmear.  Clearly, I could\'ve kept this idea to myself and made millions.  But I\'m giving back to the Net that\'s given me so much.  '),(11043,832265,'02,August,2004','2004-08-02 00:00:00','   This message is from a good two years ago.  The timestamp, however, is accurate. urlLink    urlLink audio post  powered by  urlLink audblog  '),(11044,832265,'02,August,2004','2004-08-02 00:00:00','   John Ashcroft is being  urlLink sued  over the mishandling of a major terrorism case in Detroit.  The lawsuit alleges that the Justice Dept. \"continuously placed perception over reality to the serious detriment of the war on terror.\"  Specifically, the plaintiff claims that a confidential informant was compromised and that results of the war on terrorism were exaggerated as a result of interference from Washington.  Which, you know, hardly seems like news really.  Except that the plaintiff is the former federal prosecutor in the case.  The prosecutor, Assistant U.S. Attorney Richard Convertino, has filed suit after being made the target of an internal investigation following his testimony to Congress.  Turns out the Justice Dept. doesn\'t it like it when you criticize the War on Terror to the U.S. Senate.  In fairness, there\'s no reason the Senate should know how the folks in charge of prosecuting the War feel about the job they\'re doing.  I\'m sure it\'s going just fine.  '),(11045,832265,'02,August,2004','2004-08-02 00:00:00','   Forget that nonsense about using moss or stars to know which way is north.  When\'s the last time you saw trees or stars living in a big city anyway?  If you want to find your way, simply look for the DirecTV satellite dishes outside apartment windows.  They always point south.  '),(11046,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink HDTV Arcade  provides the skinny on which games will take advantage of high-definition monitors.  Their by-platform matrices show if a game has a 16:9 mode or is available as progressive scan.  It would seem that the PS2 lags behind in this regard.  Especially when compared to the Xbox.  I can\'t wait for the  urlLink PS3 !  '),(11047,832265,'02,August,2004','2004-08-02 00:00:00','   The Yanks are getting A-Rod and (as  Jeter will still be at short) I\'m tickled pink.  It\'s worth noting to all the \"Yankees buy their championships\" haters that:  Well, this is just demonstrably true, isn\'t it?  The Red Sox tried to do the exact same thing but couldn\'t get it done.  So thhbbtt. The NYT has a great  urlLink account  of the financials behind the deal which are pretty fascinating.  One tidbit: The Yankees also got Rodriguez to defer an additional $1 million a year for each of the first four years with no interest. They offset that reduction in value by giving Rodriguez a hotel suite on the road (an old players\' perk) and allowing him to link his Web site to the Yankees\' Web site (a new perk). I like the fact the being in someone\'s blogroll has become a part of business negotiations.  '),(11048,832265,'02,August,2004','2004-08-02 00:00:00','   That  urlLink Anna .  She keeps a great blog AND sends cookies cross-country on Valentine\'s Day.  With self-stamped reply valentines!  There\'s like 47 people in the world who do stuff like this.  Hopefully, you know one of them.  (Hint: start looking among the sugar-fueled funny people in your life.)  '),(11049,832265,'02,August,2004','2004-08-02 00:00:00','   In which Eugene & Mary perform a duet in the key of Gol\'man. urlLink    urlLink audio post  powered by  urlLink audblog  '),(11050,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink Nutgroist :  \"I\'m desperately looking for a Jewish Bestiality website. Something with a good array of God\'s chosen chicks and some (often literally) horny non-Kosher beasts of the field, the air and the sea. Surely, surely there\'s got to something out there.\"( via  urlLink Eric  )  '),(11051,832265,'02,August,2004','2004-08-02 00:00:00','   Anna  urlLink points out  that Timothy McVeigh and I have more in common than stylishly short hair cuts.  But you\'ve got to ask yourself, who doesn\'t like the movies Red Dawn and Brazil.  McVeigh\'s love of Red Dawn is easy to understand.  You can picture Tim urging his militia brethen on with a full throated cry of \"Wolverines!!\"  What\'s really interesting is to see Brazil through McVeigh\'s eyes.  This is, afterall, a movie with hero-terrorists battling against a monolithic beauracracy that has stripped its citizens of their freedom.  They blow up fancy eateries, they make illegal heating repairs and are lead by a mustachioed and swaggering Bob DeNiro.  Except that\'s not what the movie is about at all.  In fact, there\'s not much in Brazil that actually establishes a real terrorist network.  DeNiro\'s character is a freelance heating repair man.  No connection is made between him and the bombings.  In fact, those that try to make that connection are portrayed as fools.  For example, Michael Palin\'s character Jack DeLint who says, \"There are no coincidences, Sam. Everything\'s connected, all along the line. Cause and effect. That\'s the beauty of it. Our job is to trace the connections and reveal them.\"  The plot of the whole movie turns on this erroneous chain of events.  Jill Layton is wrongly suspected of being a terrorist because she tries to help her neighbor who was wrongly suspected of being Tuttle who, himself, is not a terrorist at all.  The movie is not really about the righteous struggle against an oppressive society, a society so absurd in its machinations as to be inadvertantly malign as opposed to calculatingly so.  Instead it\'s about Sam Lowry\'s struggle to escape the complacency of his own life.  All I\'m saying is, I wish Timothy McVeigh would have seen the Criterion box set.  There are some good supplemental inteviews which would have set him straight.  '),(11052,832265,'02,August,2004','2004-08-02 00:00:00','   Two truisms about bad publicity:   There\'s no such thing as bad publicity.   The Anti-Defamation League has consistently refused to learn this lesson, most recently with respect to Mel Gibson\'s Passion movie.  By way of recap, 9 months ago, Mel Gibson had a privately financed movie about an Italian Christ with no distribution, no dialogue in a language spoken by living people, and no subtitles.  Now, fueled by months of controversy (e.g. members of the ADL passing themselves off as members of a fictious church in order to see an advanced screening) the movie has enormous support from a wide cross-section of christian congregations, a predicted opening weekend of $25M, and, of course, subtitles.  More than any other group or marketing strategy, the ADL has kept this movie in the news and made it a cause celebre for evangelicals.  Congrats!    The publicity\'s bad if it\'s not about you.   Why are the sponsors of the Super Bowl so pissed about the Boob.  Obviously, it can\'t be the case that Anheuser Busch is legitimately concerned that the recent boob viewage has forever damaged the moral health of America.  These are the people who had a whole campaign based around bikini models romancin\' a dog in sunglasses. As the  urlLink Guardian  points out, the sponsors are pissed because no one was talking about how great the ads were after this year\'s Super Bowl.  The big to-do over the hoo-ha is not because American families were subjected to boobie; it\'s because boobie distracted them from the annual dose of mega-advertizing.  In this respect, the Boob is a threat to the most basic of American values: commercialism.  '),(11053,832265,'02,August,2004','2004-08-02 00:00:00','   In which Eugene performs the mating song of the wild gobbler. urlLink    urlLink audio post  powered by  urlLink audblog  '),(11054,832265,'02,August,2004','2004-08-02 00:00:00','   The fallout from the Meet the Press interview  urlLink continues .  With both Bill O\'Reilly and Peggy Noonan coming down on Bush, this is the most optimistic I\'ve felt about a presidential election since 1996.( via  urlLink dgcopter  )  '),(11055,832265,'02,August,2004','2004-08-02 00:00:00','   \" urlLink Blather  is words, bunches of words, strewn about in a twisty tangly web ...\"  Redefining meta-weblog since 1998.  '),(11056,832265,'02,August,2004','2004-08-02 00:00:00','   In which Eugene shows my dork-cred the back of his hand. urlLink    urlLink audio post  powered by  urlLink audblog  '),(11057,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink President Bush : \"I\'m a war president. I make decisions here in the Oval Office in foreign-policy matters with war on my mind.\"  Yeah, we\'ve noticed.  '),(11058,832265,'02,August,2004','2004-08-02 00:00:00','   Kazaa\'s headquarters down under got  urlLink raided  today after an Australian judge issued a \"search and seize order\".  To begin with,  I had always heard that the Kazaa headquarters was like the Shadow\'s secret lair.  When you tried to find it, you ended up having your mind be-clouded and your plans be-fuddled.  Apparently not.  Second of all, it wasn\'t the Aussie police who conducted the raid.  Instead it was the \"Music Industry Piracy Investigations, a branch of the Australian Recording Industry Association.\"  So I\'m to understand the the recording industry has its own stormtroopers?  Who are empowered to conduct court sanctioned search and seizures.  I mean this is Australia, where as far as I can tell, you have to murder someone to get a tank of gas.  But still!  Are there other industry consortiums that have their own paramilitary wings?  The Mechanized Infantry of the Plumbing and Heating Industry Alliance, perhaps?  \"We gotta roll up to the vill\' and investigate an illegal channel lock possession ... lock n\' load!\"  Update : Apparently, \"The raid was conducted under a rarely used law, known as Anton Pillar, which allows litigants in civil copyright cases to gather evidence.\"  Which is a messed up law with a goofy ass name.   '),(11059,832265,'02,August,2004','2004-08-02 00:00:00','   I recently played the arcade game Ninja Baseball Batman on the  urlLink MAME  machine we have here at work.  As a wacky synthesis of japanese and american culture, it knows no equal.  The powerups alternate between french fries and bbq eel donburri.  You are a robot baseball player out to save ... something.  You have your choice of heroes - the fat yellow robot with donuts on his bat, the skinny blue stickball robot with the longball swing, etc.  Your enemies consist primarily of anthropomorphic robot baseball equipment.  There are nasty yellow baseballs with metallic mohawks.  There are evil umpires who will throw their face guards at you.  I think the ultimate baddie might be a pitching rubber.  The game\'s primarily a side scroller in the Teenage Mutant Ninja Turtles style but the level design is a chaotic mindfuck.  One level starts off on the Golden Gate Bridge which collapses and drops you onto a passing cruise ship which takes you to the docks where you have to fight a 4x4 miniboss driven by baseballs.  And, as these  urlLink screenshots  show, the design philosophy seems to have been \'wild monkey colorgasm.\'  In this respect, the game is a complete success.  '),(11060,832265,'02,August,2004','2004-08-02 00:00:00','   Matt Bruns writes the Janet Jackson  urlLink boob analysis  to trump them all.  '),(11061,832265,'02,August,2004','2004-08-02 00:00:00','   Boob.  Boob.  Boob.  The level of rhetoric around the Janet Jackson boobie incident is simply unbelievable.  FCC Chairman Michael Powell: \"Like millions of Americans, my family and I gathered around the television for a celebration. Instead, that celebration was tainted by a classless, crass and deplorable stunt. Our nation\'s children, parents and citizens deserve better.\"  Good use of the Rule of Threes, here ... \"classless, crass and deplorable\" and \"children, parents and citizens.\"  Speaking of celebrations, a 21 year-old who had swarmed the streets in Boston to celebrate the Patriots win was run over and killed by a guy in an SUV.  Who fled the scene.  And, of course, was determined to be drunk.  At this hour, it is not yet known how many people died as a result of seeing Ms. Jackson\'s boobie.  '),(11062,832265,'02,August,2004','2004-08-02 00:00:00','   Devil\'s Playground documents the Amish rite of passage known as  rumspringa  whereby 16 year-old Amish kids are allowed to experience the \'real world.\'  You know, drink, drive ... roll over on crank dealers in order to avoid jailtime, etc.  In explaining the philosophy behind  rumspringa , Faron Yoder (the main teen subject of the movie) explains that it\'s like a vaccination.  It\'s meant to give teens a taste of what they won\'t have so that they don\'t end up missing it.  Faron ends up snorting a lot of vaccine.  The ostensible tension of  rumspringa  is that the kids have to make a choice at the end of it: continue to live in the real world or join up with Amish church.  However, it\'s basically a false choice.  The \'real world\' the kids experience is a hyper-realized and completely closed teenage existence.  They only hang out with other Amish teens.  Most still live at home.  So while they\'re allowed to throw as many drunken Amish parties as they want (usually at their parents\' farms), their understanding of what it means to be a teenager is completely involuted.  As in everyday teenage life, things trend towards homogeneity - it turns out that kids are good at picking up what\'s considered cool despite an austere upbringing.  Amish kids quickly go from being completely bewildered by Wal-Mart\'s CD section to healthy pop culture consumers thanks to the social queues of their peers (and, you know, MTV).  But because of the in-bred nature of their society (even before getting the same haircut, the kids look a lot alike if you know what I\'m sayin\') and the temporary nature of their furlough in the real world, any random behavioral mutation will quickly spread through the group.  To wit, the aforementioned crank.  It\'s interesting that even when completely strung out on meth-amphetamines these kids never question their faith.  Heaven and hell are completely real and definite places.  It must be a really bad trip to be totally spun on meth and the only thought rattling through your head is  that you\'re bound for a literal hell much worse than the one in which you\'re trapped.  As if all of this weren\'t enough reason to see the flick, it\'s also got a great soundtrack.  Selected Ambient Works volumes 1  and  2.  '),(11063,832265,'02,August,2004','2004-08-02 00:00:00','   Boeing  urlLink announced  its plan to offer in-flight internet connectivity: The flat rate, which would give a user unlimited Internet access, will cost $29.95 for flights lasting longer than six hours. Each three- to six-hour leg will be charged at $19.95, while access on trips under three hours would cost $14.95. That\'s no free hotspot at the local coffee shop, to be sure, but I was expected something far more pricey.  I mean, doesn\'t it cost like $20 to make a 4 minute phone call in the air?  One weird detail ... the initial plan is for wireless and they hope to offer Ethernet later on.  Seems a tremendous expense to go around laying wires to every seat for people who will only have laptops.  '),(11064,832265,'02,August,2004','2004-08-02 00:00:00','   A while back,  urlLink Gavin  at gavinsblog.com linked to and excerpted a post in which Dr. John \"Men are Mars ...\" Gray was alleged to be a fraud.  The claim is that he doesn\'t actually have a PhD.  His lawyers caught wind and sent a nastygram - not to the person who originally posted the story, or the authors of the original outing in Men\'s News Daily - but to Gavin.  Which is pretty beat.  But also funny because in the letter, his lawyers attempt to establish his bona fides by pointing out that \"Previously, Dr. Gray received his Master\'s degree from the Maharishi European Research University ... where he also obtained his Bachelor\'s degree.\"  I\'m sure Maharishi U. is a totally legit university.  They\'ve even got an  urlLink MBA  program which combines higher consciousness and professional excellence.  I really dig the illustration on this page ... dude meditating + flip chart with  urlLink Feynman diagram  showing something assuredly managementory.  Apparently, you can maximize profits through the proper application of transcendental meditation and quantum mechanics.  '),(11065,832265,'02,August,2004','2004-08-02 00:00:00','   My cell phonaphobia may have finally caught up to me.  I\'m sitting outside the men\'s room on a conference call ... that\'s where they keep the pay phone.  Of course, the wireless is really good so, there\'s that.  '),(11066,832265,'02,August,2004','2004-08-02 00:00:00','   Photo credit: Eugene Smith'),(11067,832265,'02,August,2004','2004-08-02 00:00:00','   Silly Putty\'s come a long way, it would seem.  Check out  urlLink Crazy Aaron\'s Puttyworld  for the varieties of \'Thinking\' Putty now available.  I\'ve ordered some color-shifting scarab putty ... you know, for personal use.  '),(11068,832265,'02,August,2004','2004-08-02 00:00:00','   At lunch today, we were talking about Disney\'s classic foray into sci-fi, The Black Hole.  Specifically, we were debating the super-creepy end to the movie wherein everybody ends up taking a tumble through ye olde singularity.  The results of this are mixed.  On one hand, the do-gooders of the Palomino end up in the Land of Tubular Bells - a shiny place, full of shiny goodness and, like, gossamer-winged adventure.  On the other hand, EvilBot Maximilian and scruffy scientist Hans Reinhardt end up ... somewhere else.  They\'re shown drifting towards one another in the hazy fog of the black hole.  As general relativity tells us, when a robot and a man are put in a black hole, they will embrace gently.  And then they\'ll be fused together into some sort of cyborg, standing watch for eternity over slaves toiling in a smoky hot pit.  It\'s quite the metaphysical end to an otherwise subphysical movie.   urlLink Steve  pointed out that he\'d turned to the Internet for help in understanding this symbolic stew and the  urlLink IMDB forums  provided the hook-up.  Obviously, the simplest answer is that the filmmakers decided they need some sort of \"far out\" ending like 2001.  But lacking imagination or actual talent, they tossed up a weirdly literal, pseudo-spiritual meditation on Heaven and Hell.  Alternatively, they may have decided to make the most out of the fact that Maximillian was also the first name of the actor playing Dr. Reinhardt.  \"What if, like ... they were the same person!  But you know, in Hell!  Hey, man, puff, puff, give.\"  '),(11069,832265,'02,August,2004','2004-08-02 00:00:00','   Google\'s going public.  You may have heard.  Or not.  It\'s hard to say; to me, it seems like this is all anyone\'s been talking about anyway.  The pressfolk lurking outside our offices today seem to suggest a general interest.  But my perspective is probably a bit skewed.  I\'ve never been more proud to work at Google than when I read the S-1.  Ever since seeing a Frontline about how auctions were the solution to IPO corruption, I had dreamed of the day when some responsible corporation could show Wall Street how it\'s done.  I never would have guessed that I\'d be working for the company in question when it happened.  I also am glad about the dual class offering.  Larry Page points out all the remarkable things that make Google a unique place to work in his Founder\'s Letter.  Any system that allows him and the current management team to retain control is a good one in my book.  My god, the man specifically called out free lunches as an important part of employee culture!  Finally, I love the fact that, as noticed on  urlLink /. , the actual sum to be raised by the IPO is $2,718,281,828 or  e  x 10^9.  Given that I was a debater, quizbolwer and officer of the math club, it\'s really something to be at a place where I don\'t even come close to the top of the dork-o-meter.  '),(11070,832265,'02,August,2004','2004-08-02 00:00:00','   Another Sunday, another pleasant afternoon working up a head of rage while reading about the state of the world in Harper\'s.  There\'s a nice piece about the various legislation that small communities have passed ordering their local police departments not to comply with Patriot Act investigations.  Which got me thinking about an idea for an educational campaign.  In the tradition (and, possibly, style) of the School House Rocks civics lesson, it would be called  America: Looks Good on Paper .  Essentially, it\'s a way to let people know about their fundamental rights as guaranteed by the Constitution.  Whereas each phrase of the Constituion can be dissected with hundreds of years of case law, I believe the overall import is easily understood.  Episode One would start with the title card: \"The 4th Amendment: Damn, That Seems Pretty Specific!\"  And then it would be followed by the text of the amendment as read by some easily-snagged celebrity.  Maybe Leonard Nimoy. The right of the people to be secure in their persons, houses, papers, and effects, against unreasonable searches and seizures, shall not be violated, and no warrants shall issue, but upon probable cause, supported by oath or affirmation, and particularly describing the place to be searched, and the persons or things to be seized.   Later episodes could be more fanciful, like, \"Which Amendments are Better: Odds or Evens.\"  (Odds).  Or \"Habeas Corpus: Suspend This!\"  '),(11071,832265,'02,August,2004','2004-08-02 00:00:00','   Sometimes it\'s good to have folks over on a Saturday night for a few drinks and chance to gaze upon the visage of  urlLink pure evil .  '),(11072,832265,'02,August,2004','2004-08-02 00:00:00','   Something a computer does better than me: figure out what I can  urlLink spell  with my phone number.  Something I do better than a computer: spontaneously realize upon waking that, in the movie Clueless, Alicia Silverstone\'s somewhat-pudgy best friend was played by Brittany Murphy.  Try to figure that one out AI scientists!  I\'d been dreaming about spiders for goodness sake!  '),(11073,832265,'02,August,2004','2004-08-02 00:00:00','   Been doing some cleaning around the place and I stumbled upon a sekrit cache of mementos.  In particular, I found my notes from a working offsite I attended at my last job.  Herewith are the top 10 Death Metal Band Names inspired by a Seminar on Modern Portfolio Theory:  Pain Index Capital Death The Soul Process Fund of Mental Breakdown Blood Equity Benchmarked for Torture Too Much Fund Impaled Asset Fifty-Two Week Low Starkfolio Homicide   '),(11074,832265,'02,August,2004','2004-08-02 00:00:00','   I love the idea of moblogging and I love that my friends do it so I don\'t have to.  To wit, Sutter\'s got documentary  urlLink proof  that we saw Hellboy this weekend at SF\'s best theater, the Coronet.  Word is that the Coronet, the only one-theater movie palace in town that still shows first-run features, is due to be demolished.  The Goldman (grr) Institute on Aging has bought the place and is going to sent up some kinda death dormitory in its place.  But that geriatric sword of Damocles has been hanging over the Coronet for  urlLink 4 years , and I still got to see a ruddy Ron Perlman whoop demonic butt.  Later on, Sutter  urlLink captured  the slender elegance of my new live-in lover, the Sony Grand Wega.  I\'d been waiting to meet my sweetheart for 6 weeks and when she finally arrived, I melted.  With a 42\" diagonal, my baby\'s definitely a larger woman, but deceptively light-weight and only a foot deep.  For her debut, I had some folks over and we howled at The Big Lebowski.  '),(11075,832265,'02,August,2004','2004-08-02 00:00:00','   Welp - they\'re calling my row.  See you next time NYC!  It\'s been an awesome vacation.  '),(11076,832265,'02,August,2004','2004-08-02 00:00:00','   Woot to JetBlue for providing great service and free wifi here at JFK.  Last night I was at Princeton for Reunions.  When I mentioned to folks that I was planning on hitting up my reunion while out here, I invariably got the same response: \"What?  Your college reunion?\"  College reunions may be a rarer breed, but at P\'ton it\'s a tradition.  Wrapped up in the graduating ceremonies for the current class is a multi-day, drunken extravaganza for all sons and daughters of Old Nassau.  It culminates in the P-rade - a march through the campus of all classes, lead by the superannuated Old Guard.  After some consideration, I\'ve decided the whole thing is whack.  I\'m all for the idea of running into old friends; certainly the best part of this trip has been spending time with so many folks from back in the day.  In particular, I got to see a good chunk, tho\' sadly not all, of the debate folks with whom I spent nearly every weekend.  Partly as a result of spending so many weekends away from Princeton, I feel that going back to campus didn\'t resonate with me in the same way as it does for others.  But it\'s more than not having the same sentimental bond with the place because of time spent away.  I found myself recoiling at the parts of Princeton which I was able to ironically dimiss when I was an undergrad ... the spirit of self-congratulation; the exaltation of the place for its abstract qualities alone.  Yes - this is a Caufieldian argument.  And it\'s not as tho\' \'phony\' sums up all my feelings about Princeton; I loved it there and had the proverbial kickass time.  But, walking around campus last night, all I could see were the places where I\'d nearly had a nervous breakdown broken up by packs of middle age white men slapping each other around in celebration of the grand trajectories of their lives.  I\'m sure in another 5 years, I\'ll see it differently still.  I will say that George\'s chicken wings are still very tasty and that  urlLink Whig Hall  is an awesome place to have as your own private clubhouse.  '),(11077,832265,'02,August,2004','2004-08-02 00:00:00','   I visited the Met today for the first time.  In my book, it easily beats both the Louvre and the Prado.  It has the same \"We\'ve got everything, ever\" vibe as the great European museums I\'ve seen, but blows away the competition with stuff like  urlLink Room with Giant Temple  and  urlLink Hey, Here\'s a Facade from a Building But Inside .  Very cool.   Also, there was an Andy Goldsworthy installation on the roof.   By the way, the  urlLink Cristo deal  that\'s going to be going on in Central Park next year is going to be something to see.  '),(11078,832265,'02,August,2004','2004-08-02 00:00:00','   The Baltimore Apocalypse as  urlLink blogged by Ffoggy .  '),(11079,832265,'02,August,2004','2004-08-02 00:00:00','   Chai tea is made with some sort of powder here rather than, you know, tea.  I\'ve had it from 3 different places and each time it\'s been this Swiss Miss Chai Mix.  Questionable.  '),(11080,832265,'02,August,2004','2004-08-02 00:00:00','   The SoHo Apple Store is housed in a former post office giving it the coolest exterior of those I\'ve seen (tho\' I\'m still voting for the Michigan Ave. Apple Store in Chicago for best evah).  The wireless is also free ... I\'m blogging this during a lecture on OS X which makes me feel like a bad student.   Yesterday, Steve and I drove down to Baltimore to catch the Yankees at Camden with Anna, Lauren and LEG.  During my debate days, I drove up and down I-95 oodles of times, en route to Hopkins, UMBC, Georgetown and so on.  Whenever we\'d pass over into Maryland from Delaware, we\'d see signs for the  urlLink Decoy Museum  in historic Havre de Grace.  Many, many jokes about it being a carved, wooden museum.   Not so!  Steve and I stopped by, and tho\' it was closed, we could easily see the many cases of 19th century duck decoys.  Seriously.  I\'m not sure what\'s going on in historic Havre de Grace, but, man, they *hate* ducks.  There are poems on pedestals about tricking ducks into going after the decoys.    The hunter\'s ploy has worked. For it was not real canvasbacks that lured the ducks, but well crafted decoys.    When we reached Baltimore, we heard that there were tornado warnings in surrounding counties.  Giant, black thunderheads gathered over Camden and during the 2nd inning it began to pour.   I\'d realized recently that I hadn\'t seen a serious thunderstorm in many years; I think there\'s been lightning once in SF in 4 years.   After last night, I\'ve seen it all.   Saucer-sized raindrops forced our retreat to the top railing of the upper deck.  That\'s when horizon-wide forks of lightning erupted over downtown Baltimore and thunderclaps rocked the interior of the stadium.  Hearing thunder reverb off of the seats of a baseball stadium is quite the harrowing experience.  But not as freaky as noticing that the sky had turned green.  As every midwesterner knows, green sky = windy death.  I was seriously figuring how I could use the straps of my backback to lash myself to the handrails in case a tornado touched down on homeplate when we realized that the strange sky was due to the setting sun.  The storm passed, the grounds crew drained the 3 feet of water from the visitors\' dugout, and the sun set spectacularly as the Yanks went on to win behind a monster A-Rod homer.     '),(11081,832265,'02,August,2004','2004-08-02 00:00:00','   This evening  urlLink Steve  and I saw Jehane Noujaim\'s new documentary, Control Room.  Noujaim first flick was Startup.com, a big favorite of mine.  (I find myself quoting \"But I wanna name it renderuntocaesar.com\" a bit too often around the office.)  Control Room is signficantly better.  It documents the media coverage of the current war in Iraq from within Al Jeezera\'s headquarters and the US CentCom in Qatar.  It\'s an amazing illustration of one of Robert McNamara\'s points in Fog of War - the need to empathize with the enemy.  The US military repeatedly and continually refuses to understand why Arabs across the Middle East would react with horror to the death of civilians at US hands ... regardless of the brutality of Saddam.  That being said, there is an amazing scene in which a marine media officer reflects on how he felt at seeing images of the war\'s brutality.  And another when an Al Jeezera correspondant confidently declares that he believes completely in the US Constitution and the American people to take it upon themselves to stop what it happening to Iraq.  Damn.  In the spirit of proving him right, I think Rummy\'s on the way out.  There\'s no way he can survive Chalabi being turned out as an Iranian spy.  Steve and I were discussing who should be his replacement and we decided on McNamara.  Basically, he\'s on this redemption kick anyway ... might as well give him his shot at pulling us out of this generation\'s Vietnam, caused by another Texan president.     '),(11082,832265,'02,August,2004','2004-08-02 00:00:00','   My tour of famous, public WiFi hotspots continues ...  Today\'s stop:  urlLink Bryant Park .  This is a swell fusion of old and new.  Bryant Park feels like an old world public commons - largely because it is.  On my left, the Great Lawn stretches out to New York Public Library.  On my right, 41st stretches out to Times Square.  And in the middle ... hot & free WiFi.  With tons of available tables upon which to perch your \'puter.  Good work Bryant Park.  Someone needs to do this for Dolores Park back home.  '),(11083,832265,'02,August,2004','2004-08-02 00:00:00','   And they say people in New York aren\'t friendly.  I\'d been here only a couple hours and found myself entrenched with a ton of former debate pals ... folks visiting from DC, others who had just gotten here, some I hadn\'t seen in 5 years.  Good timing.  Incidently, things in the ole\' debate world have taken a turn for the preposterous.  Which is saying something because it is collegiate debate afterall.  But while we may have been an astonishing group of obsessive dorks, no one ever created a  urlLink debate fantasy draft .  This is sort of like if you\'ve got a group of dorkfriends with whom you play Dungeon & Dragons, and all of a sudden they start getting into LARPing.  Also I hear that there was a musical.  Seriously.  '),(11084,832265,'02,August,2004','2004-08-02 00:00:00','   Things are getting pretty dire over at my elementary alma mater.  Mom reports from the  urlLink front : We are deeper in the stuff than we thought. The sewer collapsed and permits are needed to repair it. No school for us today. A clear sign that I\'ve been playing way too much SimCity 4 - my first reaction was to reach for the public utilities slider.  '),(11085,832265,'02,August,2004','2004-08-02 00:00:00','   After&nbsp; urlLink     urlLink  '),(11086,832265,'02,August,2004','2004-08-02 00:00:00','   Before&nbsp; urlLink     urlLink  '),(11087,832265,'02,August,2004','2004-08-02 00:00:00','   I\'ve been having this intermittent problem with my 15\" PowerBook where it won\'t wake from sleep.  I\'ve asked various  urlLink experts  but been completely unable to figure it out.  Today I was clicking through  urlLink Jason Sutter \'s blogroll and went to read  urlLink Daring Fireball , which I check out infrequently.  Basically, there\'re a whole host of sites I check out every couple weeks - these are the links in blogrolls of the sites in my blogroll.  Anyway, DF has a post about how he had this problem with his Mac and when he wrote it up, he wrote it not as a blog post for his blog readership, but for Google.  \"Or, more specifically, for people using search engines to find an answer to the same problems I had.\"  That\'s savvy thinking and the sort of thing that makes working at Google so cool.  Folks understand how they can make the product better without even using the product.  But what\'s amazingly hot is that the  urlLink problem  he\'s describing is the freaking Mac-won\'t wake-up trouble I\'ve been having for the last 3 months!  And now it\'s hypothetically fixed (altho\' if having a custom date format is what threw off the energy management thus Rip Van Winkling my Mac, that\'s odd).  And, of course, I found it not because of Google, but because of blogs!  And thus I learned why we were acquired.  '),(11088,832265,'02,August,2004','2004-08-02 00:00:00','   So ... it\'s weird.  I\'ve got, like, free time now.  Last Friday, my mom came in from St. Lu and graciously  urlLink sat around BloggerHQ  as we sanity checked the Blogger release candidate.  \'Happy Mother\'s Day, Mom!  Wanna execute some test plans?\'   A side story here.  My mom enters the bloggerarium and there\'s a bunch of dudes hanging around.  You know, the team.  I say, \"Hey guys, this is my mom - Mom these are the guys.  Uh ... we also have some  urlLink ladies  but just not right here.\"  urlLink Chris Wetherdeal : \"Yeah, it\'s a regular sausage factory up in here.\"  Nice!  Had a great weekend with Mom - nearly distracted from thinking about work obsessively ... but you know, not entirely.  Sunday happens.  Sunday is amazing!  And I\'ve spent the whole of this week in some kind of bouillabaisse of stressed, joyful amazement.  And now things are creeping back to normal.  It\'s no longer about the giant huge thing that\'s left to be done.  So I need to find a hobby or something.  I\'m not sure if blogging counts.  Luckily, I\'m going on vacation week after next.  A week in NYC with side trips to various points on the eastern seaboard.  I\'m looking forward to catching up with old friends, both  urlLink lanky  and  urlLink ffoggy .  And just breaking away from the set of activities that, tho\' immensely satisfying from a personal and professional perspective, have completely defined my life for the past little while.  '),(11089,832265,'02,August,2004','2004-08-02 00:00:00','   I used to keep score when I went to ballgames.  Now I\'m blogging from one.  I\'ve somehow become an even bigger dork at the ballpark.  Luckily, I\'m  urlLink not alone .  '),(11090,832265,'02,August,2004','2004-08-02 00:00:00','   Sutter\'s got a nifty  urlLink visual recap  of the recent Blogger launch as seen from inside GoogleHQ.  Just to dispel some rumors:  My nipples aren\'t lower, they\'re just more vertically-extended.  The Korean Girls where neither girls nor sexy  Sutter\'s hobbit name is Sutter  We were most definitely  urlLink kickin\' it  '),(11091,832265,'02,August,2004','2004-08-02 00:00:00','   Note to self: when checking to see how your site\'s doing after its big launch, verify you have functional DSL before going crazy-go-nuts.  '),(11092,832265,'02,August,2004','2004-08-02 00:00:00','   New Blogger, new blog.  I\'m thrilled to have this release out in the wild.  In the last couple weeks, our team has really pulled together to get this new version out the door and I\'m psyched launch day has finally arrived.  Update : I\'m also coming down offa\' wicked donut / adrenalin high, and apparently unable to write in more than the trite cliche.  But I\'m still psched.  '),(11093,832265,'02,August,2004','2004-08-02 00:00:00','   Donna Hughes\'  urlLink take  on the Abu Ghraib tortures is a perfect example of why I keep coming back to the National Review Online.  I thought to myself, \"What spin can the NRO put on documented torture at the hands of US soldiers.\"  I was ready for a claim that torture is a justifiable wrong or even a blanket denial of the evidence.  But never could I anticipate the following stumble of logic:  The torture of Iraqi prisoners is horrible.  Hey, I once interviewed some European hookers who\'d been abused.  That\'s also horrible.  Ergo, pornography is really what\'s to blame. Seriously, that\'s the point of this article. Here are four consecutive, unedited tidbits: The images from Abu Ghraib are trophy pictures. The sadistic MPs are shown posing, smiling, and gloating over their victims and what they have made them do. I\'m with ya\', sister.  Definitely bad stuff in those pictures. Similarly, I found numerous offers on the Internet from pimps for men to bring cameras and video recorders with them to make trophy images and videos of their sexual use of women and girls. Holy propositioning prepositions, Batman! But if the point is that there\'s weird sex stuff on the Internet, I\'m with ya\' there as well.  A little unsure of the connection between your points, but let\'s press on: Why are we shocked by these images from Abu Ghraib, but when the victims are women (or gay men) the images are called pornography or \"adult entertainment\"? Wha? So, first off, it\'s weird that straight men can\'t be in pornography ... or maybe they just can\'t be victims of it.  But second, she\'s gone and condemned all porn based on the site wherefrom pimps for men do tape of women having sex.  What a slut!  Here\'s the best one tho\': What if the Iraqi men had been forced to smile, could we be convinced that there was a newly formed \"publishing and film production\" company in Baghdad instead of sexual abuse and humiliation being perpetrated? No! No we couldn\'t.  Never would it occur to us that what happened in Abu Ghraib was okay if someone tried to pass it off as porn.  That\'s like Pol Pot trying to pass off his skull collection as avant garde sculpture.  But, of course, the lesson of Abu Ghraib is all about porn. These similar images are what the young American soldiers from the Internet generation have grown up with and learned to call \"adult entertainment.\" Did they become desensitized to the harm of doing such things to people by seeing multiple images of similar abuse to women? Did they learn how to violate someone by being a voyeur to abuse, and in Abu Ghraib they had the chance to become perpetrators â€” and pornographers? Obviously, yes. In the face of such compelling logic, how could you say otherwise.  Hustler taught these soldiers how to violate basic human rights and the Geneva Convention.  Cinemax gave them the hard heart needed to do what they did.  The fault, dear Brutus, lies not in our unjust foreign policy and destructive occupation of Iraq, but in our porn.  Also, I think Howard Stern may have been involved.  '),(11094,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink     urlLink Tyler Hicks /NYT'),(11095,832265,'02,August,2004','2004-08-02 00:00:00','   Both American and British troops are accused of torturing Iraqi prisoners.  One difference between the two cases (besides, you, know, the uncontested photographic evidence of the Americans) is that Britain is now party to the International Criminal Court.  As explained by the  urlLink Guardian , there is an increased burden on UK officials to get to the bottom of these allegations.  The ICC \"has the power to launch war crimes charges of its own against authorities including the commander-in-chief - the Prime Minister - if necessary.\"  So, Tony Blair is legally compelled to make sure a proper investigation is carried out lest he be forced to answer to The Hague.  Nice!  This \'rule of law\' stuff is starting to make good sense.  '),(11096,832265,'02,August,2004','2004-08-02 00:00:00','   We hung about the tenderloin and tenderly you tell About the saddest book you ever read It always makes you cry The statue\'s crying too and will he may. Not my favorite Belle & Sebastian lyric, but it was fun to hear it sung at the Warfield last night when  urlLink CaseNotes  and I waded into what has to be the one the more polite and cheery rock n\' rock audiences around.  It reminded me a bit of a Sesame Street Live show - folks get very excited from the first chord of every song and actually cheer individual elements of the tune.  \"Woot!  It\'s the trumpet bridge.\"  Good times.  Of course, as the evening grew later, there was a bit of desperation in the voices of those yelling out their favorite songs.  Not everyone\'s going to get to hear what they want.  In particular, the tracks off If You\'re Feeling Sinister were all but unrepresented ... but Get Me Away from Here, I\'m Dying was a quality closer.  And I loved the impromptu cover of the I\'ll Be Your Mirror with Stevie Jackson singing the part of Nico.  This after the initial suggestion of a Morrissey cover was rejected with the reply, \"Cover Morrissey?  I\'d like to cover him with a rug.\"  '),(11097,832265,'02,August,2004','2004-08-02 00:00:00','   I saw two solid flicks this weekend.  And it\'s a close call but, I\'m giving Dodgeball the nod over Fahrenheit 9/11.  Partly, it has to do with crowds. 9/11 was showing at cramped, underlit Metreon with a riled up hoard of liberals who were very agitated about proper line etiquette.  Also, seriously, thanks SF Lefties for taking the time to make neat anti-Bush fliers and dress up like a tranny Condi Rice, but you kinda get on my nerves after awhile.  Especially when you refuse to pass out your own pamphlets.  \"Take one and pass it down\" does not a movement make.  Dodgeball on the other hand was at the spacious Coronet.  (Incidentally, I called 411 to find out the theater\'s number and they gave me the movie times!  Who knew.)  And while it is solely a broad physical comedy with a Ski School-inspired plot, man, is it funny.  With respect to 9/11, I agree with Matt\'s  urlLink comments  about how there\'re some good Morrisean touches, especially the soundtrack.  However, I was a little bothered by the amount of time spent on Carlyle Group stuff and the now-revised Bin Laden Airlift business.  Also, there\'s no denying the relevance or power of the story of the mom who lost her son, but, gosh, he really left her up there to bleed for us.  It was only slightly harder to watch than the Senate footage of Al Gore certifying the 2000 election results.  That\'s got to be the single most agonizing professional experience of all time.  '),(11098,832265,'02,August,2004','2004-08-02 00:00:00','   I\'ve heard lots of speculation as of late about what wacky tactic the GOP will whip out to turn the tide.  Not listed among those at  urlLink October Surprise  are the following:  Instead of dropping balloons at the GOP convention, Osama bin Laden will be lowered from the rafters in a cage The GOP will launch a \"John Kerry blows goats\" smear campaign (courtesy Dave) Bush announces a new reality show: \"Who Wants to be King of Iraq?\" (Johnny B\'s vote) Cheney is dropped from the ticket and is replaced by Colin Powell  The last is pretty scary, don\'cha think?  '),(11099,832265,'02,August,2004','2004-08-02 00:00:00','   Ralph Nader tried to  urlLink explain  to the Congressional Black Caucus what the hell he\'s doing in the presidential race ... and it didn\'t go so well.  Apparently, his explanation is that \"he will help elect presumptive Democratic nominee John Kerry in November by splitting the conservative vote and siphoning support from President Bush.\"  In what world does that make sense? Which conservatives are gonna say, \"You know I was gonna vote for Bush because I\'m a pro-life supporter of the war in Iraq, but now that Nader\'s involved I just gotta make sure that whacko liberal doesn\'t get elected.  Count me in John Kerry!\"  Ralph Nader, I invite you to explain yourself at your earliest convenience.  '),(11100,832265,'02,August,2004','2004-08-02 00:00:00','   Today, Bush  urlLink said : \"They are trying to get us to withdraw from the world so that they can impose their dark vision on people.\"  Impose their dark vision ... so that\'s like when the Emperor shoots Force Bolts at Luke in order to get him to submit to the Dark Side, right?  I have to say, this is a powerful phrase.  I think whoever wrote probably saw Lord of the Rings the night before.  You can just picture the forces of Good being beaten back by marauding hordes of barbaric beheaders.  '),(11101,832265,'02,August,2004','2004-08-02 00:00:00','   Moderation, I know ye not.  On Saturday, I took a double dip of Mission-style BBQ action.  Final poultry-fruit sausage count: 3.  But that was just the opener to yesterday\'s showcase of intemperance.  By way of background, I\'ve recently been having a Netflix-based crisis of conscious in which I only rent really bad movies.  For example, Deep Impact followed up by Gods & Generals.  (The latter being Ted Turner\'s 4 hour paean to the muscular piety Stonewall Jackson.)  It\'s hard to say why I watch these movies.  Partly, I believe you can sometimes find a kernel of beauty in the dungheap of mediocrity.  Mostly, it\'s trainwreck curiosity.  Anyway, I went to see Chronicles of Riddick yesterday.  That\'s the epic, far-future sci-fi movie that\'s gonna end Judi Dench\'s career.  I had such low expectations, that I was quite pleasantly surprised.  In the minus column you\'ve got cringeworthy dialogue and downright goofy action editing.  But it is a legitimately good story with awesome production design.  Three Headed Staff of Evil = KickAss!  And Colm Feore\'s supervillian is Evil from Time Bandits with badass superpowers and a penchant for inlavafying planets.  Inspired by Riddick, I decided to go back-to-back and catch the Harry Potter flick as well.  Here\'s a case where the collective opinion is right on; it is the strongest of the three, partly because it is more \'mature\' i.e. the kids are older and some of that stuff\'s downright scary.  And, like, the scenes with the hippogriff were pleasantly reminiscent of Spirited Away.  All-in-all a fun, supersized weekend.   '),(11102,832265,'02,August,2004','2004-08-02 00:00:00','   And the  urlLink winner  for Best Tautological Argument in Defense of an Unjust War ... George Bush!  \"The reason I keep insisting that there was a relationship between Iraq and Saddam and al Qaeda [is] because there was a relationship between Iraq and al Qaeda.\"  '),(11103,832265,'02,August,2004','2004-08-02 00:00:00','   This month, Ronald Glasser writes the quintessential closing lines to a Harper\'s article: We have grown so foolish and so incompetent that perhaps we do not deserve to survive.  Perhaps it is simply time to die. In this case, doom is coming in the form of super charged diseases.  Diseases to which we are particularly susceptible because of a lack of a coherent public health policy.  Incidentally, back in the heyday of high school debate,  urlLink Matt  was running a killer opp based on the rise of supertuberculosis due to health care mismanagement and the over-prescription of antibiotics.  I remember well the card which declared \"Hospitals are the Temples of Death and Doctors their High Priests!\"  Good times.  '),(11104,832265,'02,August,2004','2004-08-02 00:00:00','   So, yeah, Ronald Reagan\'s dead.  In 1984, we had an elementary school mock election.  I knew that my mom was going to vote for Mondale, but that all my friends (and, it turns out, everyone else) were going to vote for Reagan.  I stood in the voting booth for 10 minutes and changed my vote 3 times before finally going with the herd.  I\'ve been dreading Reagan\'s demise for awhile.  Now, it\'s only a matter of time until Mt. Rushmore and the $5 are remade in his likeness.  However, like Christmas, it seems that by living in San Francisco and avoiding TV you can escape almost any mass marketed phenomenon.  I did see some CNN today.  It was Anderson Cooper but without sound.  The banner graphic at the bottom read as follows: \"Ronald Reagan Remembered - What to Remember.\"  It\'s nice that TV wants to be so helpful.  '),(11105,832265,'02,August,2004','2004-08-02 00:00:00','   Previously, when I  urlLink wrote  about the documentary Control Room, I was particularly struck by Lt. Josh Rushing\'s struggle to make sense of the war and his role in it.  Rushing, now a Marine captain,  urlLink talked  to the Village Voice about the depiction of war:   \"I think it should all be shown, the dead on both sides ... In America war isn\'t hell â€” we don\'t see blood, we don\'t see suffering. All we see is patriotism, and we support the troops. It\'s almost like war has some brand marketing here.  \"Al Jazeera shows it all ... It turns your stomach, and you remember there\'s something wrong with war.\"  Today, Salon is  urlLink reporting  that Rushing is leaving the military having been ordered to keep quiet.  His wife is quoted as saying:  \"Everything that my husband represents in the military - the defense of freedom, and hopefully the expansion of freedom throughout the world - that\'s what a soldier is supposed to be, one would think. But it seems that he\'s denied some of those basic freedoms.\"  Genius.  '),(11106,832265,'02,August,2004','2004-08-02 00:00:00','   Despite it being a favorite of friends far and wide, I\'ve never known anyone who saw Donnie Darko in the theater.  Turns out it  urlLink made  $500K when it was initially released in 2001.  On opening weekend it was only ever on 58 screens - by the fourth weekend only 6 theaters carried it.  By way of contrast, I did see Congo in the theaters - a movie featuring laser-armed grey apes. It made $81M.  I also saw The Next Karate Kid (starring Hilary Swank in the title role) on opening night no less.  It managed to make $9M domestically.  Fortunately, Donnie Darko\'s being given a  urlLink second shot  this summer.  Even if it\'s only in 50 theaters again, I\'m betting that 5 will be in SF and that opening night will be completely sold out.  '),(11107,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink   During my junior year of college, I once stayed up for 50 hours during which time I wrote my semester thesis, flew to Athens and physically collapsed during the welcoming ceremony at the World Debate Championships.  The upshot is I broke something in my brain such that I no longer have the ability to function without sleep.  Last night I got about 2 hours before trekking to Logan and hopping the 6 hour flight back to SF.  Around an hour into the flight, I honestly thought I wasn\'t going to make it.  I\'d already tried listening to the KLF\'s Chill Out but there was still 5 hours to go and my eyeballs were trying to claw their way out of my skull.  But then - a miracle.  JetBlue, the world\'s best airline, was testing out a movie channel supplement to their signature DirecTV service.  This meant I didn\'t have to try to occupy myself with 5 hours of Animal Planet as I did on the flight out - the highlight of which was a heartwarming tale of, seriously, a blind  and  deaf dog who befriends a three-legged cat (see above).  Instead, I had the Princess Bride (which is a special treat if you know  urlLink Biz Stone  and can imagine him doing all the parts as he\'s often wont to do) and The Sound of Music.  Man, The Sound of Music is good.  Best movie involving Nazis?  Possibly, if you discount Das Boot.  What made it especially compelling is that I quickly realized I\'d never seen the theatrical cut.  The copy I watched growing up was taped off of TV and severely editted for length.  As a result, there\'s all kindsa stuff I\'d not seen despite having watched the movie dozens of times (so much so that I have a clear memory of where each commercial break occurs in the \'movie\').  Anyway, Sound of Music is also good because it\'s 3 hours long ... it was already time to land and the Von Trapps hadn\'t even escaped to Switzerland.  Thanks JetBlue!  '),(11108,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink Welp, that about wraps it up for the DNC.  My back is killing me from having sat for 6 hours waiting to see  a bunch of ballons  John Kerry. But I thought the speech was solid.  Strategically sound if not inspirational - \"health care is a right,\" \"screw the Saudis,\" \"we\'re taking back the money from the rich\" ... these are all messages I can get behind.  I  urlLink talked about  it some with Noah and I assure you that when I said President Douche instead of President Bush, it was an honest slip.  '),(11109,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink   Yesterday was a wacky, hectic day filled with Indiana Jones-themed parties (for the congressional  whips  don\'cha know) followed by the traditional scramble for convention passes.  But it all turned out well when Shellen and I gained access to the delegate floor and got to experience the full force of Al Sharpton\'s amazing speech.  I talked about it with  urlLink Noah  on the  urlLink audblog  - I feel the Democrats had to address the recent Republican effort to court the Black Vote.  Sharpton nailed it with his \"we decided weâ€™d ride this donkey as far as we could take it\" line and the Florida delgates with whom I was embedded went stark raving nuts.  And then we retired to the Blogger afterparty.  Blogs are big news here even if most people aren\'t sure why.  But it\'s always fun to get together with clever folks and talk technology ... I did some  urlLink audio rambling  with Noah on the state of the blogowhatnot.  This is definitely one of most heavily mediated environments I\'ve ever seen.  Everywhere you look someone is covering something ... sometimes it\'s just the coverage itself.  The intermingling of politicos and celebrities feeds some kinda primal excitement - last night I saw a woman lunge outta nowhere just to rub James \"That\'ll do, Pig\" Cromwell\'s arm.  She just wanted to grab a little squeeze to take home with her.  I find it painfully weird to run up and try to capture a celebrity momemt, but I also don\'t want to be left out of what\'s clearly an essential part of the convention experience.  Therefore, I present my own series of awkward celebrity photographs:   urlLink Gov. Ann Richards    urlLink Rob Reiner    urlLink Al Franken    urlLink Natalie Merchant    urlLink John  Cougar  Mellencamp Eat your heart out, Anne Leibovitz.  '),(11110,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink     urlLink'),(11111,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink   The big thing I\'ve learned on day 1 of the  urlLink DNC  is that political conventions, like politics, is all about access.  As I  urlLink talked about  with Noah, we started out the evening as uncredentialed outcasts - stuck on the wrong side of the riot-proof fencing, deafened by over-amped protesters and under-muffled diesel generators.  By the end of the evening, we had traded up our color-coded badges several times and were able to peer down on the convention floor from a lofty skybox inside the Fleet Center.   It\'s like a baseball game where you try to sneak into better and better seats as the game goes on.  The whole process is a little strange because it relies on exploiting connections and a good deal of guile.  (Neither of which are paradigms I\'m particularly comfortable with).    But both led to us jumping a partition into an adjoining skybox which happened to contain Michael Moore.  There was a weird celebrity awkwardness as folks were struggling with their digicams to snap souvenir photos.    But we ended up having a real conversation about blogs, Google and the terrible state of copyright.  Michael Moore is a clueful dude.  And hungry ... which I can appreciate.  The one completely genuine moment of the evening was Barack Obama\'s  urlLink speech .  The man gave me chills.  In terms of what public service should mean, Obama nailed it when he said \"If there\'s a senior citizen somewhere who can\'t pay for their prescription and having to choose between medicine and the rent, that makes my life poorer, even if it\'s not my grandparent.\"  That the poverty of others can make all of our lives poorer goes right to heart of the problems government is meant to address.  (For what it\'s worth, the least genuine moment of the day was dancing to Come on Eileen at a convention after party.  Apparently, I\'m going to be awkwardly dancing to that song for the rest of my life.)  '),(11112,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink'),(11113,832265,'02,August,2004','2004-08-02 00:00:00','   .flickr-photo {	border: solid 1px #000000;}.flickr-yourcomment {}.flickr-frame {	float: right;	width: 150px;	text-align: center;	padding: 3px;	margin-right: 10px;}.flickr-caption {	font: 75%;/*	color: #666666; */	margin-top: 0px;}.flickr-buddyicon {	margin-right:5px; 	vertical-align:middle;	border: solid 1px;}.flickr-postedby {	font: 75%;}  	 		 urlLink    		  urlLink Foot Dangle , originally uploaded by  urlLink goldtoe .'),(11114,832265,'02,August,2004','2004-08-02 00:00:00','   The Flacksmiths recently got new bicycles, so we set out this morning on a  urlLink Grand Bike Adventure .  The initial plan was just to cruise to the Embarcadero but before we knew it we were in Sausalito.   The weather was perfect and along the way we took in many great sights ... including an elderly gentleman with a head full of white plaster.    We got to cross the Golden Gate Bridge (a hoot!) and stuff ourselves full of fish parts over in Marin.  I also highly recommend the ferry ride back.  Ferries  always make me feel like re-enacting the opening scene to Working Girl.     Or that one Madonna video.  Regardless a good time was had by all. 		 		 		 			 urlLink     			 				 			 urlLink Golden Gate Bridge (detail) 			 			Originally uploaded by 			 urlLink goldtoe .  '),(11115,832265,'02,August,2004','2004-08-02 00:00:00','   .flickr-photo {	border: solid 1px #000000;}.flickr-yourcomment {}.flickr-frame {	float: left;	width: 150px;	text-align: center;	padding: 3px;	margin-right: 10px;}.flickr-caption {	font: 75%;/*	color: #666666; */	margin-top: 0px;}.flickr-buddyicon {	margin-right:5px; 	vertical-align:middle;	border: solid 1px;}.flickr-postedby {	font: 75%;}  	 		 urlLink'),(11116,832265,'02,August,2004','2004-08-02 00:00:00','   During August 1997, before my junior year of college, I watched the filmed version of Jesus Christ Superstar approximately 7 times in row in less than 2 days.  After some viewings, I would cry.  It was a confusing time and, frankly, my weepiness was much more a pharmacological by-product than a spiritual awakening.  Needless to say, I approached the new direct-to-video version of JCS with some trepidation.  Filmed in 2000, the goal of the producers was to update the look so as to reach a wider audience.  So it\'s out with bell bottoms and in with the Gap.   As a result, when we first see Jesus, he\'s wearing a ribbed wifebeater and cargo pants.  The apostles are portrayed as some kinda metrosexual paramilitary group outfitted with blue camo muscle tees and automatic weapons.   Which brings up a confusing part of the whole production ... who exactly are we rooting for here?  The apostles force Jesus to go pop, his followers quickly become an unruly mob, even the lepers are shown as being annoying ... well, maybe that\'s not a stretch.   Pilate, as is the modern style, comes off amazingly well.  Sure he\'s wearing Goering\'s hand-me-downs, but that\'s just \'cause he\'s Roman.  Pilate hearts Jesus.   In fact, one of the movie\'s more interesting touches is the scourging.  In the film, folks from the mob come running up and slap their hands on JC\'s body as Pilate counts out the lashes.  Their hands are dripping with red paint so the streaks they leave on Christ\'s torso represents the whipping.  At the end of the scene, we zoom in on the ghoulish jewmob (they\'re wearing goth eye makeup) as they reach out with their blood stained hands.   Them\'s my peoples!  This post has been cross-posted to  urlLink Flicknut  '),(11117,832265,'02,August,2004','2004-08-02 00:00:00','   Recently, I had to send the following message to the good people at Netflix:  Hello -    I am an idiot.    I accidentally sent back a non-Netflix dvd in the return mailer.  The movie I was trying to return was Derrida.  Instead, I sent you back my friend\'s copy of the second disc of the Two Towers.    I know.    I just dropped it in the mail this morning (7/10) so you wouldn\'t have received it yet. But I\'m wondering if there\'s anyway you can send it back to me once you get it. (At my expense, obviously).    Thanks for you help.  I love Netflix and have always found it very easy to use.    Which is important, considering the fact that I am quite stupid.  And, what do you know, less than a week later Eugene\'s copy of the Two Towers was returned!    Right after I sent this message I also ordered a replacement copy from Amazon just in case. I briefly toyed with the idea of not telling Eugene and just replacing the missing disk with the new copy in a covert-Blue-Heaven-this-is-the-same-turtle-you\'ve-always-had style.    In the end, I confessed and now have my own copy of the Two Towers to accidently send back to Netflix! Incompetent honesty truly is the best policy.  '),(11118,832265,'02,August,2004','2004-08-02 00:00:00','   I don\'t miss owning a car, but I have noticed that my view of San Francisco has shrunk to a very small strip that extends from the Mission to Lower Haight. I\'ve gone on some bus-bound adventures as of late, but inertia pretty much keeps me in the neighborhood and at the same restaurants I\'ve been eating for nearly 4 years.  Don\'t get me wrong - I love the hood. But I need a little expansion. (Partly I feel this is due to my obsessive playing of the new Spiderman 2 game. Long since having won the game, I continue to swing around Manhattan just because it\'s so frickin\' cool to be able to go  anywhere .  How sweet to swing from the Chrysler building and leap into Times Square ... and do it all in high pixelated-style).  Anyway, as I cannot (yet) be Spiderman, I bought a bike. I don\'t know much about bikes but mine\'s red and silver and let\'s you sit upright instead of hunched over and has a giant squishy seat. And it rocks! On my maiden trek I made it all the way to the ocean and saw parts of Golden Gate Park I never knew about (including the SFPD horse stables).  As soon as my legs recover I fully intend to cruise the Embarcadero ... I hear there aren\'t any hills.   '),(11119,832265,'02,August,2004','2004-08-02 00:00:00','   As discussed  elsewhere , we\'ve released&nbsp;a wizz-uh-muh-wig editor for Blogger. &nbsp;   It\'s got features and such and will be a boon for text manipulators everywhere.&nbsp; I\'m a particular fan of the new glyphs that Chris came up with for the toolbar.&nbsp; I feel like quoting stuff all the time now.          '),(11120,832265,'02,August,2004','2004-08-02 00:00:00','   NPR\'s got a good  urlLink audio clip  on the Senate\'s 500 page \'Fuck You!\' to the CIA.  I\'m fond of this quote by Sen. Jay Rockefeller:   \"Our credibility is dimished, our standing in the world has never been lower, we have fostered a deep hatred of Americans in the Muslim world and that will grow.  As a direct consequence our nation is more vulnerable today than ever before.\" So, it was a good thing, right?  '),(11121,832265,'02,August,2004','2004-08-02 00:00:00','   Once upon a time, Charles and Marie Robertson gave $35M to Princeton in order to establish a public policy school.  The result, the Woodrow Wilson School, is housed in a neo-neo-classical temple and home to the most over-achieving of what\'s already a massively type-A student body.  Perhaps unsurprisingly, in recent years the would-be wonks are pretty much not becoming the nation\'s next generation of career diplomats.  So the Roberston kids want their  urlLink money back  as the school is \"instead turning out journalists, bankers and -- most appalling of all -- a professional oboist.\"  Yeah, fuck oboists.  The kids should definitely get a refund ... with interest it comes to $600M.  D\'you think they\'ll take store credit at the bookstore?  I have a hard time picking a winner here.  On one hand, $600M lawsuits should probably involve some kind of mutagenic orange juice sold to orphan children.  On the other hand, the Wilson school is primarily used as an elite credential for undergrads seeking to maximize their future remuneration.  Princeton should be forced to put all the money into fusion research.  Success will lead to oil independence thus negating much of the diplomatic cost incurred in having to do business in the Middle East.  While not a funding of diplomacy per se, there will be more diplomatic resources left over as a result.  Which will allow us to invest in international affairs where we might have a snowball\'s chance in Saudi Arabia at having a positive impact.  '),(11122,832265,'02,August,2004','2004-08-02 00:00:00','   I first read Tom Stoppard\'s Arcadia in the winter of 1999 when I was busily dropping out of grad school.  It turned out that studying astronomy as a techincal discipline wasn\'t the bucket of fun I\'d been hoping for.  The wow factor I felt in doing astro as an undergrad was replaced by dread at having to slog through another problem set session.  Arcadia covers a lot of thematic ground; one of the main ideas is the tension between rationalism and romanticism.  Throughout most of the play, characters lash themselves to one or the other of these two poles and teeter about as a result.  The teetering\'s pretty damn funny.  (In reading the play, I\'ve always been amazed at how much mileage he can get out of stuff like \'the conversation at cross purposes.\')  Last Tuesday, Sutter, danah and I saw Tom Stoppard\'s Arcadia at the Mountain View Center for the Performing Deals.  It\'s quite something to see it performed after having read it so many times.  For one thing, the sucker is dense.  Reading 10 minutes of dialog on iterative functions is one thing - to see it pulled off is something else.  Also, the play takes place in a single room in an English estate, but in two separate time periods (early 1800s and present day).  In the last few scenes, the action from both time periods overlap, with both sets of actors performing at the same time.  To make the staging of such a thing seem important rather than gimmicky is quite a feat.  Technical stuff aside, Arcadia\'s always hit home for me and that was strongly reinforced seeing it live.  '),(11123,832265,'02,August,2004','2004-08-02 00:00:00','   Steve and I, having been  urlLink prevented  from seeing baseball at PacBell due to the insidious trickery of the US Naval Observatory, reveled in good-seat glory last night in Oakland.  One of the great pleasures of baseball is that its slow pace gives you time to contemplate the mysteries of life with good friends. For example, what would your intro music be if you were a major league closer? My list: \"Isolation\" by Joy Division     \"Closer\" by Nine Inch Nails (probably only in New York)     \"The Centaur\" by Buck 65     \"It\'s the End of the World as We Know It\" by REM     \"Personal Jesus\" by Depeche Mode     \"Personal Jesus\" takes it for me by a wide stretch because of the salvation themes.  But it\'s clear that there can be no  urlLink worse choice  than \"Dancing Queen\" ... maybe the PSB version of \"Go West.\"   '),(11124,832265,'02,August,2004','2004-08-02 00:00:00','   As widely  urlLink reported :  \"Kerry said he would have cast the same Yes vote in Congress that he did on Oct. 11, 2002, to authorize the president to launch a pre-emptive war that began March 19, 2003, even if Kerry had known that Saddam Hussein had no ties with Al Qaeda terrorists, no weapons of mass destruction and posed no real threat to the world.\" How is this anything less than Kerry slitting his own throat?  Besides just being wrong in principle, it makes no sense politically.  If Kerry can\'t distinguish his stance on Iraq from that of Bush, he\'s nowhere.  Bah.  '),(11125,832265,'02,August,2004','2004-08-02 00:00:00','   Today, Alan Keyes  urlLink said  the following ridiculous thing about abortion: \"I would still be picking cotton if the country\'s moral principles had not been shaped by the Declaration of Independence,\" Keyes said. He said Obama \"has broken and rejected those principles he has taken the slaveholder\'s position.\"  Obviously, the point was to have the quote picked up.  He\'s said something similarly preposterous in a number of other  urlLink interviews  - for example, arguing that Barack Obama (by supporting gay marriage) has turned his back on everything decent in the world and that Keyes, like Lincoln, must defend our fundemental principles of Union.  Will Barack Obama, the greatest thing to happen to Democratic politics in a generation, be sidetracked like so many others by the social issue sideshow?  Fuck no.  \"As I travel around this state, I don\'t get asked about gay marriage, I don\'t get asked about abortion.  I get asked, \'How can I find a job that allows me to support my family.\' I get asked, \'How can I pay those medical bills without going into bankruptcy.\'\"  Alan Keyes, Barack Obama is a righteous political whirlwind.  You are but the cracked out trailer park in his path.  '),(11126,832265,'02,August,2004','2004-08-02 00:00:00','   There\'s nothing remarkable to say about David Mamet\'s Spartan.  It\'s Jack Ryan with staccato dialog.  But seeing it did make me check out what Mamet\'s up to next ... apparently he\'s working on a Will Farrell vehichle called Joan of Bark: The Dog that Saved France.  Seriously.    '),(11127,832265,'02,August,2004','2004-08-02 00:00:00','   If you\'ve ever wanted to experience the awe of someone from the 19th century gazing upon our modern world, go out and get an AirPort Express.  Sending audio from a laptop to your stereo without wires!  All hail the future!  '),(11128,832265,'02,August,2004','2004-08-02 00:00:00','   urlLink   Sick of being bald?  To cheap to buy a real hairpiece?  Help has arrived!  Now you can own a homemade toupee in just 5 easy steps:  1. Find a friend with a thick head of hair.  The thicker, the better.  Take  urlLink Eugene , for example.  His hair is so thick he gets it cut every couple weeks ... with garden shears! (buh-dum-bum).  2. Next, go over to said hirsute friend\'s house when he\'s getting a trim.  Try not to salivate as you watch all that luscious hair pile up on the floor (the spit would make the hair clumpy).  3. After the haircut is over, get a couple strips of wide, transparent tape.  Packing tape is perfect.  Duct tape, not so much.  4. Dredge the the strips of tape through the piled hair clippings.  5. You should now have a couple chia-esque strips of tape.  Plop those suckers on your noggin and you can hold your  urlLink head up  high from now on!   '),(11129,3997213,'30,July,2004','2004-07-30 00:00:00','    It was a slow day. Mom treated my aunt to breakfast at Ammenities Cafe so Riley and I came along. For the reference, I had a sasauage and cheese bagel. Riley had a cupcake. Perfect for a toddler in the morning. Yum! It was hot so I came home and went up on the pool deck to read \"The Good Earth.\" That book is so depressing at times so I got my suit on, jumped in the pool, and floated around in a tube...again! Man, I love doing that. It\'s such a comfortable float. So I pretty much spent the day out back and in my room. It was such a quiet day. Dad was home doing work around the house and Rachel was at cheer camp. Reid was a lazy bum from the sleepover he just had. I was the only energized one in the house for a change. I won\'t be writing next until Monday. We\'re leaving for Ocean City, Maryland tomorrow night to visit some cousins and Rachel is spending like two weeks there. We\'re dropping her off. The house will be so quiet without her and you know what. I think I\'m gonna like it.   '),(11130,3997213,'29,July,2004','2004-07-29 00:00:00','    Well, I didn\'t get to write yesterday. I couldn\'t get on the internet at all. The stupid cable box is so unreliable. So to sum up the day, I got a haircut, went to Barnes and Noble to buy Pendragon, had gymnastics, and lost my Barnes and Noble gift card. (I figure I dropped it walking out of the store)  But today was good. Nana was able to come by and watch Riley so I didn\'t have to do it. It was a relief for once. We got Burger King around lunch time. Reid HAD to have chicken nuggets from there even though I wanted Wendy\'s. He always gets his way and now I sound like a brat. I came back and went in the pool. Actually, I floated around the pool in a tube. I didn\'t feel like going in. It was cold from all the rain so I drifted around and got a good tan.  Before that I made some corrections on my history report (Thank you mom!) and was able to read 3 chapters in \"The Good Earth.\" It took me the same amount of time to read 1 chapter in Jane Eyre. This book will probably be much better. Speaking of books, I stayed up pretty late last night reading Pendragon. It\'s the only time I can get. In the daytime, I make myself read books for school.  So I had gymnastics tonight at 5 and got out at 7. I came home and watched Big Brother 5. Did I mention how much I love that show and how I want to be on it when I turn 21? Well I do. We went over to Kohls afterwards at 9:00 at night. Mom had a 10 dollar off coupon and had to use it before it got too late and I needed to get a new watch. So, we ran out with Riley and Rachel (Dad, Reid, and his friend John went over to my Uncle John\'s to see his boat) I picked out a nice new watch. And the best part about it is the blue light. Yes the shiny shiny blue light that glows in the dark is great!   '),(11131,3997213,'27,July,2004','2004-07-27 00:00:00','    It\'s weird waking up to an empty house. I don\'t really like the feeling either. Mom left for work in the morning and, surprisingly, I didn\'t hear her go. Rachel came running into my room to tell me to pick up the phone that was ringing. By the time I rolled over and got to the top of the stairs the phone stopped so I trudged my way back to bed only to lay there for a couple of minutes. I didn\'t hear much of Riley for the first half hour or so. Rachel had her up in mom\'s bedroom and they were watching TV. I stayed downstairs.  Reid had drums at like 10:30 so mom was coming home to pick him up. Here\'s where we had our major problem of the day. I\'m going to describe it exactly as it happened. I decided to play a video game. No one was around because Rach was upstairs with Riley and Reid was eating breakfast. It seemed a perfect time to play the gamecube and I saw nothing wrong with it. So, I\'m sitting there minding my own business, having fun when Reid goes over to the front door and opens it. \"Reid why do you need the front door open?\" \"Mom will be home soon so I\'m just gonna leave it open so I\'ll be ready to walk out.\" I said okay and continued with what I was doing. I didn\'t really care. The door open didn\'t bother me too much. But it DID bother Rachel. She came down at that EXACT moment and this was where things got ugly. You have to understand. Rachel and Reid are like the cat and dog in the house. They don\'t get along whatsoever. WHATSOVER! So, naturally, she had an objection to Reid opening the door.  \"Someone could just walk right into the house Reid! Mom wants the door closed!\" Reid fired back, \"Rachel! It\'s only 10 minutes mom will be open soon.\" I sat minding my own business, twidling my thumbs, which was my one big error. There was something I should have done but didn\'t. It may sound silly, it may sound adultish, but I should have sent them to their rooms. It works every time yet why didn\'t I think of this them?? To sum it up, the door opened and closed about 6 or 7 times. Reid opened it. Rachel closed it. Reid opened it. Rachel closed it. On and on and on it went and meanwhile I\'m going back and forth between sides trying to get one of them to give in. Of course that\'s impossible though. They don\'t give in. It came down to Reid losing his temper and pining Rachel against the wall. He yelled in her face and went all \"animal\" like on her. Man, I can so see this kid getting into trouble. He loses his temper and it\'s not pretty. Rachel started crying and came over to me telling me to do something. Riley sat and took it all in. I sat and played my game. WRONG MOVE! Ironically, mom called at that moment and they both yelled what happend over the phone. That\'s when I knew it was GAME OVER for me. Sure enough, I got it from mom. Yes. I bow my head in shame but I think I learned my lesson. When in doubt, send them to their room.&nbsp;  So, that was the \"meat\" of the day. We ran to the credit union and went over to Nana\'s. Mom and Dad were going to this Titanic Exabition thingy so we stayed at Nana\'s. We was my favorite dinner, spaghetti and meatballs. I reserved my copy of Pendragon #5 at Barnes and Noble so I\'m gonna go get it tomorrow. I watched Big Brother at 9 even though I had to miss the first 20 minutes of it on the way home from nana\'s. I got gymnastics tomorrow too.  Well, I\'m done now for the night. I\'ll be writing again tomorrow. I\'m starting to like this blog. It\'s funny because I\'m slowly growing accustomed to it. I didn\'t think I would continue it at first.  But now, I think it\'s safe to say that you will be seeing a lot more blog from me from now on. Life can&nbsp;only gets more interesting from here.   '),(11132,3997213,'26,July,2004','2004-07-26 00:00:00','    Good day today.&nbsp;I slept in nicely and got a quiet house as mom took Reid to drums. Riley went too. Rachel was here but she wasn\'t bothering me. It\'s nice every once and a while to be able to enjoy some quiet time even if it was only about 45 minutes and, mind you, I spent that time doing homework. I spend a lot of my time now doing homework. Maybe I\'m obessing a bit too much over it because I can\'t remember the last time I was able to play a video game without having a guilty conscience. Seriously, it\'s a pain in the butt. Do you know how much I would love to just read a book of MY choice. Pendragon #5 comes out soon and it\'s only my favoritest book series in the world. I think I need to reward myself and get it. But that doesn\'t mean I can forget about \"The Good Earth\" and \"One&nbsp;Day in the Life of Ivan Denishovich.\" They have to get read too. Ah I\'m rambling now. I finished my history report day. It is such a relief to have it out of the way. I used a secret method while writing it though. I called it \"pacing myself.\" I would write a sentence or too, go onto the pendragon website, talk a bit there, go back, write a sentence or two and I continued this process until it got done. It wasn\'t distracting. It helped me stay focused on my assignment strangley... Rachel had a friend come over today along with the twins. For some odd reason, she found it necasarry to show them my bedroom. 0_0 I kinda shrugged her off in front of her friends and played the part of tough and&nbsp;superior older brother. It\'s funny watching her friends look around my room. It\'s even funnier to see them nervous around me. Ahahah it\'s fun being the oldest. I love the power...   '),(11133,3997213,'25,July,2004','2004-07-25 00:00:00','    I had hotdog day today. We went over at around 10:15 a.m. and I stress the A.M. I didn\'t need the sleep anyway though. I get enough of it. I\'m not complaining.  So anyway hotdog day was, in&nbsp;a word, hotdogfull :) or hotdogish or whatever you want to call it. The point is I\'m a bit sick of the dogs. Yeah, you know&nbsp;you\'ve had enough&nbsp;when the smell of hotdogs stick to your nostrils like... like well I don\'t have the energy right now to tell you what it\'s like but it\'s not pleasent. Give me a break here. It\'s 12:30 at night (er in the morning) You see my thoughts don\'t run too well after a certain hour so please excuse me if there are any grammer and spelling \"don\'ts\". ( Alright. I saw Anchorman tonight. Chris, Greg, and I went. Pretty funny movie actually. I laughed my hotdog off (ouch that don\'t sound right! I wonder where I got that idea from though?! Cough Cough) Excuse me. I laughed my tail off. The movie was really funny!  Oh yeah and how can I forget? I FINISHED JANE EYRE!! About time too Mr. Ryguy... (Darnit I\'m talking to myself now...) Wohoo that\'s enough for me tonight. I think I\'m a little over tired. Man do I ever need some sleep. Did I say that I need to get more sleep? Honestly...   '),(11134,3997213,'24,July,2004','2004-07-24 00:00:00','    It\'s Saturday. But the weekends don\'t matter to me anymore anyway. It\'s summer. Every day is a weekend. So anyway, mom was out at yard sales like always. She picked up a bunch of Rugrats stuff for Riley. She loved that. Rachel had cheer camp and dad was out working on the lawn. I didn\'t do much really.&nbsp;I went in the pool for like 15 minutes but it was so cold for some reason. We went to Target after that. Riley spilled her bag of m&m\'s ALL over the floor at the register. (It\'s not the first time) I had to get down on my hands and knees in the middle of the store and pick up her m\'s. Real funny.  Dad lit a fire in our fire pit out back but I wasn\'t there long. I was working on a CD for Rachel. She wanted to buy the Usher Cd but I told her to save her money (correction: mom\'s money) and I made it for her. She was pretty happy and I was pretty happy that I was able to do the whole thing in a couple hours. (New record for me) Then she wouldn\'t stop playing \"Confessions\". Oh how I have grown sick of that song...and Usher. Big Brother came on at 9:00; watched that. I might have to work Hotdog day tomorrow for the adult softball league if it doesn\'t rain. I don\'t care either way. I\'ll also be done reading Jane Eyre tomorrow for school. I have about 3 chapters left. That\'s pretty exciting. I guess...   '),(11135,3997213,'23,July,2004','2004-07-23 00:00:00','    I had every intention to work on my history project today. It\'s due next week and I\'m just about done plus what better day to work on it than a rainy one.&nbsp;Today was perfect. We weren\'t going anywhere, the house was pretty quiet, and it was raining.&nbsp;Sounds great but wait, this is reality.&nbsp;In reality, things don\'t always work out the way they were meant to be. You guessed it. I did squat on my report. Yes procrastinator was written all over my face today but I obviously never looked in the mirror considering I went the whole day seemingly oblivous to my homework. Oh well. I\'ll get it done. So, as I was saying, it rained all day if you didn\'t the count the hour or so in between. I&nbsp;woke up&nbsp;to mom talking on the phone. I always said she\'s a loud talker on the phone but it&nbsp;just seems so much louder in the morning. Reid slept in later than normal and Rachel and her friend were in the play room. Mom made some awesome french toast. The morning dragged on. I lounged around in pajama pants until mom announced the possibility of us going out to lunch with Aunt Debbie.&nbsp; This was the hectic part of the day. See, you have to understand that Reid is very picky with his chicken. He orders chicken at basically every restaurant and I assume that after ordering the meal at least 25-50 times, you start to get very \"weird\" about the chicken you eat. He likes Friendlies. We were just at Friendlies and weren\'t about to go again today. Uh here is where we had our problem. Reid was grumpy and tired. He wanted Friendlies and nothing else. Personally, I didn\'t care where the heck we ate but, eventually (against what Reid wanted)&nbsp;the choices were narrowed down to Stewarts Root Beer and Wendys. To make a long story short, we went to Stewarts but with a lot of hassle getting out the door. Riley was fussy and needed a nap so she made everything complicated and stressful with her crying. That was the last thing mom needed. We picked up Aunt Debbie and went out in the pouring rain. Very pouring rain! Stewarts was good. I had the best California Chesseburger EVER. Riley slept in her seat.  We came home and jumped in the pool in the rain. Wow that is fun. Reid had soccer camp and I was staying here with Rachel. Her friend had gone home. Mom and Dad went to watch Reid. The 2 hours or so we were here was very boring. I watched TV. Period. TV gets very boring after a while especially when it\'s near 8:30 at night, you\'re home alone, and have yet to eat dinner. We had pizza from East Ridge at 9:15 and it was anything but peaceful. It was a typical family dinner where everyone is rammy from being indoors all day. You can imagine.   '),(11136,3997213,'22,July,2004','2004-07-22 00:00:00','    I woke up to watch Riley. My grandmother couldn\'t make it to watch her so that left me. Reid had drum lessons and Rachel had field hockey camp. Mom was at work. we watched Dora the Explorer and her Pirate Adventure. Oh how I love that movie! Reid came home, Rachel went to stay with our cousin Caila and mom came home to take us to Sam\'s. We had to get hot dog buns for Hotdog day on Sunday. It\'s for the township\'s adult softball team and I usually help with my mom. But wait a second. JUST going to Sam\'s sounds way too easy. No way. Mom needed to stop and get a replacement glass for our lampost. Dad knocked it out mowing the lawn. Typical. Anyway, to make it short, we ended up going to THREE different glass places (never knew there were so many) and not one place could fix it. Of course. Everyting always has to be complicated. Sam\'s wasn\'t bad. The sample ladies were there, never fail.  We came home from Sam\'s, I had to get to gymnastics. Reid had to get to soccer camp and Rachel was having a friend sleep over. I went in the pool for about an hour then got ready to go to the gym. Once again, that sounds way too easy. Riley had to go and color on the carpet. Oh boy! Mom was mad. I think it\'s still there too. We tried vinegar and it kind of got most of it off. Still visible though.  Gymnastics was tiring. Worked my butt off on the pommel horse, rings, and vault and I\'m suffering now. I think I pulled my shoulder on the rings doing inlocates. Ouch.  Dad picked me up and I came home to watch Big Brother at 8:00. I love that show! Lori was evicted. Drew won HoH.  Now I\'m watching Confessions of a Teenage Drama Queen with my sister and her friend. Lindsey Lohan is way better than Hilary Duff!   '),(11137,3997213,'21,July,2004','2004-07-21 00:00:00','    First time with one of these things. Don\'t know if I\'m gonna continue it. I only wanted to do it to put up some pictures but I\'m having some difficulty with that. I don\'t know. It\'s 11:50 at night. No time to do any of this now...  I might decide to continue this tomorrow.   '),(11138,3997213,'13,August,2004','2004-08-13 00:00:00','    I slept like a rock last night. Maybe it was because I went to bed kinda early for a change but I woke up at around 10:30 without even trying. That\'s unusual for me. I\'m normally up around 9:00 or so. So yeah it felt good. We went to Friendlies with my nana for lunch. She\'s home from the hospital and it\'s good to see her back. I had a cheeseburger and then chocolate ice cream with gummy bears. Man, I couldn\'t feel my stomach after that. I think I gained some weight, which is a good thing. I could use a few pounds here and there. I was on the computer for awhile when we got home then I went outside to the pool and trampoline. It wasn\'t a particulary exciting day but like I said, it\'s summer. The days will get more interesting when school starts. I\'ll probably also have a lot to talk about when we come back from Seaside next week. We\'re leaving this coming Monday and coming back Thursday. It\'s not as long as originally planned but we\'re planning on going to Hershey park in the fall. That will be fun. Oh hershey chocolate. My stomach is groaning at the thought. That\'s what 5 big scoops of chocolate ice cream right before you get on the computer will do to you.   '),(11139,3997213,'12,August,2004','2004-08-12 00:00:00','    It was Thursday so I watched Riley while mom went to work. I\'m not going to bore anyone by going on about it. I feel like I say this often. I don\'t mind watching her. It\'s not a hard job at all if you can even consider it a job. It\'s more like something normal. So there was nothing special today. My throat is kinda sore now and so are my ears. Yeah, my ears hurt and I don\'t know why. Actually, I do know why. I went in the pool earlier and never got the water out of my ears. 0_0 I made Rachel and her friend a CD today. It took a chunk out of my time but it was worth it. They were pretty excited and I enjoy seeing they\'re faces of surprise. (I make the real deal you know?) Plus it gives me something to do that is fun.  No injuries at gymnastics tonight. Only some red marks on my stomach from the strap belt which pinced me. My leg is bruised and hurting a lot. Hopefully, I\'ll be over it at the end of this week because we\'re going to Seaside on Monday of next week. That will be a good week but we were supposed to go with my grandparents. Now it looks like they can\'t make it. It won\'t be the same without them. The olymics are also on that week. I\'ll have to remember to watch them in the moTEL. You know, that word is funny when you say it like that. It sounds like Mo tell.   '),(11140,3997213,'11,August,2004','2004-08-11 00:00:00','    So I didn\'t write anything yesterday. There was nothing to say because I stayed around the house all day and ate candy. (It wasn\'t as much as the day before though) Anyway, today was different. I woke up craving a chocolate donut knowing we had a box of them in the kitchen. I ate one for breakfast, turned around an hour or so later, and they\'re gone. Things go fast around here. The motto: If you don\'t get it while you can, you\'ll never get it. In other words, you snooze you looze. Now how true is that? Reid had a friend over again like every other day. They go and climb trees. Isn\'t that exciting? Mom took us to Toys\'R Us because she had to pick up a few things for Westampton Day and I bought a new playstation game. We went to the dollar store after that for a few more things and I got some stuff there. Dollar stores are funny. I had gymnastics at 5:00 but before that I jumped in the pool. It was cold thanks to all this darn rain so I wasn\'t in long. I went on the computer for a bit and that was when time flew by. I looked at the clock and had to be out the door in 5 minutes for gymnastics. I hate it when that happens. It turns out, I would have saved myself a lot of pain if I didn\'t go. I\'m not saying that I DON\'T want to go anymore. Otherwise, I wouldn\'t have these bruises on my arms right now. I did a handstand on those parrell bars and fell forwards over onto my legs and arms. It hurt. It was embarressing. It left a nasty pain. My arms are all red but they don\'t hurt too much. I\'m proud of myself. I\'ll have you know I only cried for an hour. ;)   '),(11141,3997213,'09,August,2004','2004-08-09 00:00:00','    I didn\'t do anything special today and I can honestly say that. It was a perfect example of a lazy summer day. Reid had a friend over and Rachel biked around and went to cheer camp. Mom took Riley to the playground (that\'s her new thing) and I stayed home. It was a nice day so I was outside for a while. When I wasn\'t, I kept eating candy, which I had a huge craving for ALL day. I was constantly back and forth to the cabinet grabing hand fulls of Swedish Fish and Sour Patch Kids. Then I ate some blueberrys. At least that was healthy. I had gymnastics tonight at 7:00 and yeah it was the usual. Hopefully, tomorrow will be more interesting because I\'m tired of writing about nothing. We\'re going to Seaside hopefully next week which will be fun and I\'ve got to finish my summer books. I read a bit of the Good Earth today. One word for the book. Blah   '),(11142,3997213,'08,August,2004','2004-08-08 00:00:00','    I remember the title of a book I read once. It was called Absolutely Normal Chaos by Sharron Creech. Though the book itself doesn\'t have much relevance to our life, I think the title does. Normal chaos. That\'s the way it was around here today when Rachel came back. It was like all of a sudden the house grew louder and tension mounted. Rachel and Reid started it but like I said, it\'s normal. They had arguments over the Sims and for some reason it seemed to affect the whole house. I noticed it today on the way home from visiting Nana in the hospital. Rachel and Reid fight in the fan and it makes everyone else edgy. Know what I mean? It was like a dark cloud that once it affected one area, it spread and hit everyone else...except me that is. I sat back and kept my cool, staying out of stuff as best as possible. When we got home, I headed out the backdoor and got away from it all until Reid followed me. I think the fresh air did wonders for everyone because the tension went down from there. That is until it was time for dinner. I don\'t need to go into detail here though. It was a normal dinner. It was the absolutely normal chaos dinner that we all have grown to know and live with.   '),(11143,3997213,'06,August,2004','2004-08-06 00:00:00','    I didn\'t do anything too special today. Mom took Riley and I with her to the hospital so we didn\'t have to stay around the house and we got to visit my nana again. I have to admit it was better than staying home. We were there for a while while Reid was at a friend\'s house, we left then were back again tonight. So overall not a very eventful day. But I feel like typing tonight so I\'ll describe in more detail our visit to Maryland since we\'re not all going to pick Rachel up tomorrow. Just Dad is going. It actually was pretty interesting and it\'s a shame I didn\'t feel like talking about it the night we were back. We left on Saturday around 6:30. Rachel had cheer camp that whole day so that\'s why we left that late. Actually, it would have been later if Rachel had gone to her teacher\'s wedding that was originally scheduled. I got a visit from Ed earlier that day whom I hadn\'t seen since school got out. I knew he would find me eventually. We took a walk around the block with Katy. It was a fun time. We all piled into the van about an hour or so later (after getting Wendy\'s for lunch that is) and hit the road. But that sounds way too cut and dry for our family. Nope there had to be some excitement thrown in there and, yeah you guessed it, it dealt with Rachel and Reid. Not that it\'s surprising but Reid found it neccasary to tease Rachel with a bucket of worms that was sitting behind her. Dad was taking them for fishing. Of course, she freaked and flew out of the back of the van screaming. I sat there laughing. But c\'mon! You have to laugh about it. Besides, she would have found out some time. Eventually, we got out of the driveway, which was by no means easy. Rachel and Reid started their bickering before we could get out of the neighborhood. They fought over who had more leg room. (Personally I think Rachel did but I didn\'t say anything.) Good thing too because I hate getting in the middle of their \"battles.\" I simply loaded on the headphones, got out a book, and pretended nothing was wrong. Two and a half hours later, we were there. Unbeknownst to me, I was missing Big Brother. We settled in at the house and arranged where we would be sleeping. Riley warmed up quickly. She was running around without a diaper. Reid was amused by a dart board and Rachel went off with Holly. I was in the middle of it all like always. The next day we planned on going to the beach. Rachel and Holly were going to a birthday party and dad and my uncle were going golfing. We had an open day and beach seemed perfect. But then, like a sign for the day ahead, it started to rain. We hoped it would just pass over but it never did. One could say that it stayed over because the rain never ceased. It stayed over us all day and ruined all plans for the beach. Naturally, when it rains one may go to the movies so that is what we considered. To make a long and complicated story short, we never went to a movie. Instead we went to Walmart. Yeah that\'s right. We went to Walmart. It was the wackiest Walmart I\'ve ever been to. I\'ve never seen so many people jam packed into one store like that and it was definately weird. Originally, we went to go get some \"dinnas\" for Riley (we forgot em at home) and we were gonna pick up a few things. We ended up buying a coloring book for Riley, a boogeyboard for Reid, some pool goggles for me, pop tarts, and cheerleading shoes for Rachel. Meanwhile outside the sky is turning dark. We didn\'t know that. If we did I bet we would have been outta there a lot faster cause the storm was wicked. So here we are. We got all we need to buy and we\'re going to the check out line. Mom says we can go hit the outlet stores now and then go back to the house. We\'re walking to the register when we hear thunder. I\'m talking about thunder thunder cause this was loud! One lady screamed which was pretty funny. Mom held her ears with Riley. I chuckled to myself. Now the only thing left was to get to the van with the groceries and make it back to the house. The only problem with that is mom is scared of the lightning yet she is the only one who can drive the van. I waited with Riley and Reid in the little foyer area of Walmart while mom went to bring the van up. Me, being level headed and calm, assured her that it would be fine and I told her I would do it if I could drive. (I would too. I don\'t mind the lightning.) Obviously, we got back to the house in one piece and we never went to the outlets. I didn\'t mind. That rain was pretty intense. There\'s no way we would have made it to the rest of the stores. Score one for me for suggesting to skip it. Later that night, we rented Along Came Polly on On Demand and we had S\'mores on a s\'more maker. Holly filed my nails and made em all nice and shiny. It was a new experience, I must say. Bedtime came. We left the next morning and left Rachel there. The ride home was easy. I slept the last half hour or so and Riley and Reid slept practically the whole time. It was quiet. There were no Rachel and Reid fights. Dad\'s going to pick her up tomorrow and he\'ll stay a night there. I\'m expecting the house to become a little louder by the end of the week now. Rachel will be back and everything will go back to normal. Well, everything will be normal enough for us anyway. For anyone else, it would be crazy.   '),(11144,3997213,'05,August,2004','2004-08-05 00:00:00','    Just call me Riley\'s babysitter. Call me the house keeper. Call me a big brother because I\'m getting really accustomed to watching Riley. I even found it in me to clean the house and do the dishes today. See, mom had to take Reid to drums at 10:15 and I ended up watching Riley until 1:30. Mom went over to the hospital to see my nana after Reid\'s lesson. I cleaned the rooms, made the beds, did the dishes, made Riley lunch, and I tell you I was thinking, only THINKING, about vaccuming and sweeping the floor. (Don\'t worrry. I don\'t think I would go that far) But, overall, it paid off. Mom came home to a clean house and a not-so-clean-wet-diaper Riley. (I knew I was forgetting something!) When I got my break, I went up to my computer and played a computer game. Oh yeah. Did I mention that while I was being perfect son and cleaning Reid sat on his butt and played on the downstairs computer. I don\'t think he even moved until the 15 minutes before mom arrived and that was because he knew he had to clean up the clothes in his room. I had gymnastics at 5:00. My hands were bleeding after I ripped em open on the bar. So I got to relax the last 5 minutes of practice. It felt good. We went straight over to Virtua hospital after that to visit my nana. Here I am with dried up blood and chalk on my hand (Predominately the right hand) walking through a hospital. Not that I really cared, but it was uncomfortable. Anyway, we gave her a visit. Big Brother came on at 8:00 and I missed the first 45 minutes of it. That\'s okay though. I got all the details at the website. So that about wraps up my day. Who knows what I\'ll end up doing tomorrow but I can tell you that watching Riley for a while is definately up there on the possibilities list.   '),(11145,3997213,'04,August,2004','2004-08-04 00:00:00','    It was such a nice day outside today. Too bad I was only out there for about an hour. I think I wasted it (on the computer) because now all it\'s doing is raining. I just want some hot days. The only reason I was in today was because well I decided to reinstall my Sims games so I loaded my laptop with the Sims and some expansions. It\'s been forever since I played that game and I\'m surprised I\'m back into it. Only problem is that it\'s so addicting that you end up wasting your day away in front of the computer screen. Sad isn\'t it? So, I had gymnastics tonight too. There\'s not too much to say about it. I managed to jump in the pool earlier and it was really warm. It was like 84 degrees. Nice. Reid brought home a drum set from my uncle. This is going to be a pain. I can hear it coming now and I have the feeling it will be loud. As long as I keep my bedroom door shut I should be fine. Rachel is still in Maryland and we\'re going to pick her up at the end of the week. I\'ve given some thought to me describing the trip in more detail. I didn\'t really do it justice when I came back. You can tell from the vauge entry a couple days ago. There actually was a lot more to the trip than just rain so I\'ll describe it on the weekend after we pick up Rachel. That way, I can include the end as well. (Cause we still have to go back one more time you know!) I\'m surprised at how non eventful these days are when Rach isn\'t here. Normally, we would be running her around but now the house seems so much more.....quiet. I know I\'ve repeated this over and over but it\'s unusual. I\'ve come to the conclusion that this time must be known as \"still.\" Let\'s just see how long this lasts.   '),(11146,3997213,'03,August,2004','2004-08-03 00:00:00','    Wow! My legs are killing me today. It must be something from gymnastics last night and now it hurts to walk down the stairs. Luckly, it wasn\'t a busy day. I had to watch Riley while mom went to work from 9:00 to 1:00. My nana is pretty sick so I might be watching Riley a little more often now. Not that I mind watching her though. She\'s may be a handful (getting into chocolate and throwing the remote across the room) but it\'s not too bad. At least mom comes home with Burger King. That\'s a good reward. I had a whopper junior today. It was free from the Spiderman scratch off game. Normally I eat a whopper and a medium fry but I ate Chinese food left over from yesterday so I was kinda full. Reid and Riley fell asleep today so I got to play a video game for a while. It was nice being able to play a game without Reid whining to watch Cartoon Network. That\'s the only thing he watches. I\'m serious. I don\'t know why we\'re paying for these other channels when the only thing on is Nickolodean and Cartoon Network. Urgh! Sometimes it makes me mad cause he expects me to do the same. Sorry but I like some variety in my shows.   '),(11147,3997213,'02,August,2004','2004-08-02 00:00:00','    Got back from Maryland today around 2:30 in the afternoon. To be honest, I don\'t feel like describing the trip. There\'s not really any point. It rained the whole time we were there so we didn\'t get to go to the beach. That\'s about it. We stayed around the house. I\'m really tired now. I had gymnastics tonight at 7 and my hands are killing me. I ripped the side of my thumb on the high bar. I came home from the ride today and went straight to my computer here. I think I was on around 3 hours and before I knew it we ordered Chinese food (I had that chicken stuff) then I headed off to gymnastics. So what I\'m trying to say is the day went fast. Good thing too. I can\'t wait to sleep in my own bed. The house will be quiet. Rachel is in Maryland and Reid is down the street at a friends. I love the quietness!   '),(11148,3569294,'30,June,2004','2004-06-30 00:00:00','   This has been a very exciting experience for me.  I have enjoyed all of the information sharing and ideas from each of you.  Good luck with your careers and I\'ll probably see you around!  Mary Hartsoe Marketing  HHS  '),(11149,3569294,'26,June,2004','2004-06-26 00:00:00','   In the Social Studies I Competency Goals for Special Education students at the high school level \"Student must demonstrate an understanding of the three branches of the federal government and their functions.\"  The Textbook I use in class addresses this in 3 simple pages in chapter 7 section 4. As we talked before texts do not address topics in enough depth. I have found three websites that will serve as additional teaching aids in helping my students understand the branches of the federal government.  The Truman Presidential Museum and Library http://www.trumanlibrary.org/whistlestop/teacher_lessons/3branches/1.htm  urlLink Cool link  Takes the reader through each branch of the federal government, what each is responsible for, what makes up each branch and the checks and balances that exist between each one. This site also has short question sections for the reader.  These question sections range from simple fill in the blank to short answers.  I usually spend a lot of my planning looking for more in depth questions to help my students understand the branches.  This site is interactive and has much of what I have done before all rolled into one.  Also, my students enjoy spending time in the computer lab because it is individualized and they can move at their own pace  Federal Government Branches- The United States http://www.lawmart.com/federalbranches.htm  urlLink Cool link  This website links you to the official websites for each branch of the federal government.  I like being able to use this website because my students can use this site to improve their internet skills as they search for specific information.  For example who are the US Senators from North Carolina?, How many cabinets are there in the executive Branch? How many woman are on the Supreme Court? and many other questions.  My students love to use the internet, but lack in certain retrieval skills that they made need to survive in the real world when they graduate.  Ben\'s Guide to US Government for Kids grades 6-8 http://bensguide.gpo.gov/6-8/government/branches.html  urlLink Cool link  I like this site because it is written to make sense to a middle school mind.  Most of my students operate at or near this level. I like the way that this site is able to go in depth with the different branches while at the same time not confusing the reader.  Many of the ideas or topics listed are things I have searched for in the past, but have trouble with the concepts getting to deep and for my students. urlLink   urlLink  '),(11150,3569294,'25,June,2004','2004-06-25 00:00:00','   One of the most important decisions that students will make in their lifetimes is what their career will be.  In order to make those decisions, they need to learn about themselves - their abilities and talents.  They need to know what careers are available.  These websites will help begin this process.  The website  urlLink www.princetonreview.com  has many resources.  I especially like under that under the career heading there is career quiz that is very simple (only 24 questions) but gives students a printout that is color coded with possible career options.  The website  urlLink www.scois.org  has a section termed Your Pathways to Success which has activities for learning about different careers.  The website  urlLink www.jobprofiles.monster.com  also gives great profile information about possible careers.  '),(11151,3569294,'25,June,2004','2004-06-25 00:00:00','   Unit III: Electronic Resources   urlLink Check out this link  Musuem of Tolerance: Children of the Holocaust  This site will enable students to see first hand what happened to children during the Holocaust.  The students will be assigned to read several biographies of children who survived the holocaust.  I will then give them the assignment of writing a bio-poem for a child of their choice.  They will have to type this poem using Microsoft Word and include a picture of their child.  I will then hang these pictures up on the bulletin board in the classroom.   urlLink Check out this link  Elibrary  Nigeria: Chapter 1E. Popular Disillusionment and Political Realignment; Robert Rinehart Countries of the World   01-01-1991 I will use this article to help the students understand the background for the story â€œIn the Shadow of Warâ€ from their textbook.  I will have them read this article and highlight what they think is important.  After reading and highlighting, I will ask each student to come up to an overhead and write one important thing from the article to share with the class.  We will discuss what the students come up with.   urlLink Check this link out  It is very important to relate the war and peace theme for this lesson to what is going on today in Iraq.  The BBC News carries excellent news coverage, updated each day that will help the students see how war is affecting the Iraqi people.  I want them to understand that we are lucky to be in a country that for so long has been safe from war.  This site always includes facts about daily life in Iraq, information about the rebuilding of Baghdad, and many personal stories from all sides. I will take my students to the computer lab and have them explore this site.  They will be required to write journal responses to two personal stories and also give me a statistical view of life in Iraq.  Everything they need to provide me with these assignments is included on this site.  '),(11152,3569294,'24,June,2004','2004-06-24 00:00:00','   The lesson I am planning involves teaching the importances of being physically fit. I plan to do this by using the textbook Health: A Guide To Wellness, the internet sources that I have picked out for them, and my own personal experinces. In making this plan, I hope that each student will benefit from having learned the material and be able to apply the knowledge that was given to them to their everyday life.  I have compiled some internet sources for the students to work with. I personally have been to every site and approve of them because each site has valuable information.   urlLink Check this site out   This site is a Georgia State University site that is simply titled The Exercise & Physical Fitness Page. It contains exercise program information. It involves the different types of exercise, the importance of warming up, the importance of nutrition, and other important factors in doing an exercise program. Their goal is to provide research based exercise and fitness information.   urlLink Check this site out   This site is developed by the President\'s Council on Physical Fitness and Sports. It is titled Fitness Fundamentals:Making a Commitment. It covers everything you need to know about making an exercise program and all of the things included in making an exercise program.   urlLink Check out this site   This site covers the five basic components of physical fitness. It has several other links published on the page that enable the reader to explore more options that go along with physical fitness.  I plan to have the students go to each one of these sites and compare and contrast each site to the other. I will then have them compare that to the material in the textbook. They will list some of the same points each text covers and list some differences. After they understand the components of exercise I will have them make their own exercise program using the information that was given to them.  '),(11153,3569294,'23,June,2004','2004-06-23 00:00:00','   Chapter 7 in my Sixth Grade Math book (Prentice Hall) and also Chapter 7 in my Community College text (Basic College Mathematics) both deal with statistics. As the word \"statistics\" scares students of all ages, I plan to use these three sites as a fun way to introduce the concepts of   mean, median,   and   mode   to my students.   Weather Today by BBC Education (Activity, Fact Sheet, ONLINE test)  Of the three sites, this is the easiest, giving very simple explanations and easy problems. I would use this as either the starting point for everyone or as an effective remediation tool for those who do not grasp concepts easily. urlLink http://www.bbc.co.uk/schools/revisewise/maths/data/   AAA Statistics This site is my second choice because it moves students up a level. First they see a mini lesson, then they scroll down to a practice area where their reponses are given immediate feedback. Within this same framework is a series of \"beat the clock\" games to challenge learners further. Lastly, even greater ability can be coaxed from the student as he/she decides whether to perform   mean, median,    or   mode   as 1,2, or 3 digit operations. urlLink http://www.aaamath.com/B/sta.htm   Train Race Activity  by BBC Maths File ( need to use own calculator with this)  This site ranks as my highest level due to the applicational level of requirement. Students must know their terms as no lessons are offered, then they must determine central tendencies of trains in order to place their passenger on the CORRECT train before it departs the station. Overall, the highest orders of thinking are applied to this game.   urlLink http://www.bbc.co.uk/education/mathsfile/shockwave/games/train.html  '),(11154,3569294,'23,June,2004','2004-06-23 00:00:00','   Principles of Business VoCATS Objective - J. Careers in Business and Marketing PB11.00 - Analyze and develop an individualized career plan. PB12.00 - Demonstrate procedures for obtaining a job.  FYI - I searched for these websites using google, dogpile, and ask.  Site -  urlLink www.topten.org/content/tt.AB8.htm  The Top 10 Tips for a Successful Job Interview website lists and details ten tips for a successful job interview. They are easy to understand and very explanatory. This website covers PB12.03 - Explain etiquette in a job interview.  Site -  urlLink www.collegegrad.com/careers/index.shtml  The Collegegrad website targets college graduates but has career information by industry, salary and hiring trends, free career analysis which matches you with 11-20 jobs, Kiersey Temperament Sorter, resume samples, free templates for letter of application (cover letter) and samples, interview preparation tips, how to dress, entry level questions, after the interview - call interviewer and write a short thank you note. The links for the objectives are on the left side of the screen. This site was very easy to navigate and understand. This website covers PB12.01 - Develop a resume and letter of application, PB12.04 - Create a follow-up (thank-you) letter, PB11.00 - Analyze and develop an individualized career plan, and PB11.01 - Analyze the characteristics, requirements, and availability of careers in business and marketing.  Site -  urlLink http://jobsearchtech.about.com/library/weekly/aa05101.htm  The About website covered resume writing, technical job resumes, entry level resumes, free templates for resumes, job applications, thank you note writing tips, and career planning. This website covered some of the same objectives as the collegegrad but offered different resume styles and general job application tips. The links are located on the left side of the page. This site covers PB12.02 - Complete employment forms necessary for obtaining employment.  Site -  urlLink www.washingtonpost.com/wp-adv/jobs/Career_Advice/impactadvice5.htm  The Washington Post website lists the 18 Deadly Interview Mistakes Job Seekers Make. Some seem silly to those of us who have participated in at least one or more interviews. However, teenagers are new to this task. This site was easy to navigate and easy to read with understandable vocabulary and terms.  '),(11155,3569294,'22,June,2004','2004-06-22 00:00:00','   The lesson I am planning could be used in a business financial management class or an introductory lesson for a class on entrepreneurship. It will focus on the creation of a business plan - its parts, their components and its importance to the success of any business.   All businesses, large or small, have a business plan. Currently, 90% of newly created jobs are created by small businesses and 65% of small businesses fail.   The sites will be used to supplement the textbook and allow the students to delve deeper into the creation of a business plan. At the present time, I envision the lesson to include a small group project where each group will be an \"angel network\" or \"venture capitalists\", who will review three plans and decide which, if any, plan they would fund and for how much.  Each group will need to create a final report justifying their decision. These reports will be presented by the group to the whole class.   The sites I\'ve chosen are:  urlLink www.sbaonline.sba.gov  This is the website for the Small Business Administration. The site contains useful information and resources for small businesses.   urlLink www.toolkit.cch.com  This is the CCH website. CCH INCORPORATED is a leading provider of business, legal, and tax information and software to the business community. It contains the tools to help start, run and grow a successful small business.   urlLink www.bplan.com/samples/sba.cfm  This site contains sample business plans for small businesses in a variety of industries. Students will be able to review plans for the types of businesses that interest them.   urlLink www.incmagazine.com  This is the website for Inc. magazine, a magazine for entrepreneurs.  Two other sites that could be accessed for articles pertinent to this topic, but are also good resources for business articles in general:  urlLink www.work911.com  This site contains links to over 2000 work related articles.  urlLink www.smartbiz.com  This site has as a tag line \"How small businesses get big\" and contains links to business related articles that would be of special interest to small business owners.  One final site business teachers and students may find useful:  urlLink www.khake.com  This site has business, management and finance career resources. It has links for career exploration, skills, schools, and job market. It contains links to other sites, articles, search engines, strategies and tutorials.  '),(11156,3569294,'22,June,2004','2004-06-22 00:00:00','   The following links were chosen to assist a sixth grade science class in studying the chapter entitled: \"Technology and Space Exploration: Past, Present, and Future\".  I envisioned an internet research assignment in which a group of students would submit their findings in a class presentation on this chapter addressing one of the following themes:  Discovery, Pioneers, Vision, Equipment.    1.   urlLink        NASA Ames Research Center       Click on the heading: \"The Vision for Space Exploration\".  This site gives students access to articles, press releases, official statements, video archives, and public addresses.  By accessing these selections they will be exposed to President Bush\'s vision and statements for America\'s space program, the vision of NASA\'s Administrator, and the vision and comments from the NASA Space Commission.  Additionally, there are sites devoted to demonstrations of new technology such as the Personal Exploration Rovers.  2.   urlLink http://btc.montana.edu/ceres/html/EdActivities.html        NASA/MSU-Bozeman CERES Project.       At this site, students have access to classroom activities and projects that expose them to the challenges a person would face as an astronautical scientist.  Each activity is listed by grade level with a description of expected time required and the exact area of space research that is being addressed.  Students would choose from activities such as \"First Manned Mission to Mars,\" \"MarsQuest,\" \"Every Picture Tells a Story,\" and \"MountainQuest.\"    3.   urlLink http://curry.edschool.virginia.edu/go/edis771/webquests2001sp/ssarahsouthall/webquest_student        \"Explore Your Universe: A Web Quest for 6th Grade\"       This site is devoted to space exploration with an emphasis on discovery.  Students will follow guided instruction within the website to explore the planets, stars, satellites, etc of our solar system.  '),(11157,3569294,'22,June,2004','2004-06-22 00:00:00','   My lesson topic is regarding HTML and the basics on how to begin writing Web pages. I personally don\'t know the code at all; I chose the topic since I had a high school HTML textbook at home on loan from a teacher. The following are websites that I found which may be good to use as resources.   Page Tutorial  urlLink http://www.pagetutor.com/  This particular website,  urlLink PageTutor , is a good online resource for beginners to those who are already familiar with HTML. The user can order the tutorials in written/printed format. In addtion, the site offers both free material as well as more in-depth material for registered users. I would use this website as supplemental material to the Shelly Cashman textbook for those who might find it beneficial or more interesting.   Page Tutorial 2  urlLink http://www.mcli.dist.maricopa.edu/tut/  This website,  urlLink Writing HTML , is somewhat similar to the previous one. The website is broken into various lessons, progressing from basic to more difficult. Each begins with a list of objectives. The lessons, themselves, contain numerous headings to break up the text as well as bolded words and hyperlinks within the lesson. In addition, the lessons have review questions at the end. The lessons seem to be very informative, but often contain nothing but text; in other words, some don\'t contain illustrations which serve visual learners well. I would use this also to supplement the textbook. Students could either use this site or the first one. It\'s always good to have choices in this world!   Call the Doctor  urlLink http://www2.imagiware.com/RxHTML/  This is a very neat website ( urlLink Doctor HTML ), one that I never knew existed on the Web. This is a Web page analysis tool that examines a particular Web page for any errors. The site will provide the user with information on how to improve his/her page. In addition, the user can select the various diagnostic tools that he/she wants to use. In other words, the user can direct the site to check for certain errors while ignoring others. This would be great for those just learning the code. Errors are a near certainty. Students could use this when their Web page will not run properly, and they are unable to find the error.  '),(11158,3569294,'22,June,2004','2004-06-22 00:00:00','   I found three differnet sites that will be perfect for helping the students understand Algebra.  However, the students can continue to use the site for further Math classes dealing with different parts of Algebra. I have listed the web sites and I will describe how I will use them in my lesson.  urlLink  http://www.mathleague.com/help/algebra/algebra.htm .   This site provides vocabulary words associated with Algebra.  Also, this stie gives differnet types of examples for the terms that will be used.  I thought, I would use the site for a steady base on getting to know some of the concepts of algebra.  The students will use it as a resource before jumping into algebra, the students will use this site to look up basic vocabulary words before being introduced to the concepts.  The reason for doing it is to have them come in with some understanding of algebra.  As we start out, the studnets will not be in the dark.   urlLink http://www.algebrahelp.com/index.jsp   This site will be helpful for the students outside of the classroom.  This site has all different kinds of ways for the students to understand the concepts.  Since the site has different ways of looking a the concepts, the students will be able to connect the information in class and other information together.  This might give the students a better understanding of the material. It has different kinds of games the studnet can use to see if he/she understands the concepts.  Also, the site has worksheets the students can work, too.   urlLink http://www.eduref.org/cgi-bin/printless.cgi/Virtual/Lessons/Mathematics/Algebra/ALG...   This site gave me an idea on introducing algebra to the students without causing them to get anxious about the subject.  It gives ideas on ways students can think about the dreadful world of algebra. Also, I will be able to use hands on materail for the students to play with.  If they are able to see and touch the objects being modled than there might be a good starting point for the students.  The studnets are able to work in groups which allow them to learn from others.  '),(11159,3569294,'21,June,2004','2004-06-21 00:00:00','   CONSUMERS AND BUYING DECISIONS  Mary Hartsoe 06/21/04  Reinforcement of Text Materials with Internet Activity at: http://www.marketingteacher.com/powerpoint/powerpoint.htm Click on â€œGOâ€ beside of Buyer Behavior (#5), Select Open Click anywhere in the presentation to advance to the next slide.   Continue through each slide and take brief notes.   Compare your notes from the textbook with the PowerPoint slides.   Notice how consumer buying behaviors may differ from institutional buying behaviors.   Why do you think there is a difference?   ______________________________________________________________________________________________________________________________________________  http://www.econedlink.org/lessons/index.cfm?lesson=EM457 Go to the site above and complete Activity 1 â€“ to learn more about consumers.  Answer the final question:  What other goods and services do you buy with your money?  List at least 5 items you or your family have purchased in the last month.   1. 2. 3. 4. 5.    http://www.consumerreports.org/main/home.jsp?bhfv=7&bhqs=1  Consumer Reports Organization Website offers consumers free and unbiased information regarding testing for thousands of products. Expert testing is done on products and is displayed on this site for consumers to review before purchasing a product.  This assists the consumer in the buying decision process.  This information is also available in print form in the Consumer Digest.  Information is available on many items, such as:  autos, appliances, electronics, computers, home, garden, health, fitness, personal finance, baby products, travel, and food.  It also provides information about consumer protection, issues and recalled products.   Go to this website.  Select a category and product that you would like to investigate.  Research the product for safety, efficiency, and other types of ratings.  Compare your product to another similar product.  How does your product rate when compared to the other product?    My Product _______________________ Ratings: ______________________________________________________________________________________________________________________________________________   Other Similar Product: ___________________________ Ratings:______________________________________________________________ _____________________________________________________________________  Would you consider checking out this website for product information, before you make a purchase, in the future?  Why or why not? ______________________________________________________________________________________________________________________________________________  '),(11160,3569294,'21,June,2004','2004-06-21 00:00:00','   I found several very good Keyboarding/Word Processing Web Sites.  This site is  urlLink Keyboarding Online at Crews Middle School .  The site has a section called â€œRules & Toolsâ€ which contains questions and answers about keyboarding skills and technique, home row keys, tips for better posture, and keyboarding tips and rules.  It also contains information for parents and interactive activities for students.  This site is great!  urlLink http://www.crews.org/curriculum/ex/compsci/keyboarding/    This site is  urlLink Catawba County Schools .  I have listed 2 URLs.  The first link is the CCS home page.  The second link has  urlLink Computer Skills Remediation Materials  that are wonderful resources for the North Carolina Test of Computer Skills given to all 8th grade students.  The site contains sample questions and worksheets to help students prepare for the State test.  It contains an Online Computer Skills Jeopardy Game with questions that cover the basics of the Multiple Choice Computer Skills Test.  There are also Online Spreadsheet, Database and Word Processing Power Point Presentations.  I appreciate the School Systems who make resources available for others to share.  Thank You, Catawba County.  urlLink http://www.catawba.k12.nc.us    urlLink http://www.catawba.k12.nc.us/pages/sites/edwebsites/computerskills/index.htm     This Site is  urlLink St. Stephens High School  in Catawba County.  This site has Online End-of Course test questions for review.  The second URL shows  urlLink Business Course Descriptions  and has review for the end-of-course exams -- VoCats.  This site has great Secondary Education resources.  Thank You, St. Stephens.  urlLink http://www.catawba.k12.nc.us/schoolpages/ststephenshs/   urlLink http://www.catawba.k12.nc.us/schoolpages/ststephenshs/pages/acad/business/business.htm     This site has great  urlLink Online Typing programs and Typing tutors  for students.  urlLink http://www.kidsdomain.com/brain/computer/type.html     The  urlLink Education World Web Site  is full of information for teachers and students.  It contains Articles, Classroom management tips, Tools for Teaching, Web Wizards, and much more.  This is a great site.  Take the time and check it out.  urlLink http://www.education-world.com/a_curr/curr076.shtml      urlLink  '),(11161,3569294,'21,June,2004','2004-06-21 00:00:00','   Here are some web resources for my project about autobiographical poetry:   urlLink Poetry Express   poetryexpress.org  This site will form the nucleus of the students\' writing phase of the project.  They will work through the \"15 poems you can write now\" activity as part of their experimentation with autobiographical poetry.     urlLink The Favorite Poem Project  favoritepoem.org  The purpose of visiting this site is to highlight the connection between real life and real poetry.  This project is one in which famous people (and regular people, for that matter) share their favorite poems and tell why they like them.  Students will explore this site and develop some ideas about why some poems are dear to peoples\' hearts.     urlLink The Newshour Daily Poem     This daily poetry feature from the Newshour highlights autobiographical poetry.   urlLink  '),(11162,3569294,'20,June,2004','2004-06-20 00:00:00','   Hi guys~ Hope all is well with each of you! Here\'s my list of resources for a study on the Holocaust using Elie Wiesel\'s \"Night\" as the reading.  http://www.nizkor.org/ This is a great site with info about every aspect of the Holocaust, it even battles the debate that some believe that it didn\'t happen. The two main areas that I will be using from this site are the \"66 Questions & Answers about the Holocaust\" and the area about \"The Camps.\" I am however going to allow the students to roam around this site and look at anything they please but will ask them to concentrate on those 2. This site will be used as a reference to mainly camp information as well as a resource for the students on a final project that they may do. I\'ll refer to it as \" urlLink Everything Holocaust .\"   http://www.ushmm.org/outreach/nrule.htm This site deals mainly around the Nazi rulers and the part that they played in the Holocaust. There is a timeline as well as keydates area that I think will be very useful for most students. I also like the fact that there is a glossary on the page. This site throws in every event that was happening during the Holocaust and the affect that it had on it. I like that the articles on the site are \"short, sweet, and to the point.\" I feel that that more students may actually read what they are supposed to and still get the info that I want them to from the little article. I would use this site when teaching about the part that Hitler played as well as to give them a place to go to for dates and terms. I will refer to this site as \" urlLink The Nazi .\"  http://xroads.virginia.edu/~CAP/HOLO/ELIEBIO.HTM  This last site is a biography on Elie Wiesel. Students would read this site in order to get a sense of the type of person that Elie is instead of just someone who survived the Holocaust and wrote about it. I think the site could give a deeper insight when they read \"Night.\" I\'ll refer to this site as \" urlLink Elie .\"  '),(11163,3569294,'19,June,2004','2004-06-19 00:00:00','   Why, oh, why will my blogs not post as beautifully as yours? I\'ve read the directions until I\'m blue in the face and can\'t seem to correct the problem! Can someone, anyone, tell me what one does to successfully blog free from orange editing buttons that taint your subject line? Thanks! :)  '),(11164,3569294,'16,June,2004','2004-06-16 00:00:00','   Greetings fellow Bloggers!  I\'ve tried several times in vain to link the sites from Blogger by following the directions given in \"?Help\", but it\'s not making the connection for some reason; I\'ll try again later.  Nonetheless, I\'m going to press on with the assignment.  I will be using the following sites in my research:  http://www.teach-nology.com/teachers/lesson_plans/language_arts/speaking/  I will be referring to this one as the Super Speaker site.  The words \" urlLink Super Speaker \" will be what you will click on to link to this site once I figure out what the problem is.  This site is incredible.  It is a comprehensive site including many of the important elements of oral communication:  tips on public speaking, interviewing techniques, writing interviews, writing speeches, sample speeches, etc...There are a few sites I will direct the students to from here.  One of them is the \"Putting it all Together\" site that I indicate at the bottom of this list.  Another one not listed on these top three is a site containing a collection of famous American speeches that I will be directing them to as well.  In addition, I currently plan to give them the freedom to peruse this site for research picks of their own.        http://cspace.unb.ca/nbco/pigs/speaklist/inter1.html  I will refer to this site as the  urlLink Informed Interviewer  site.  This site will be used as a reference for a great class activity where the students will interview a fellow student and then present that person to the class; this exercise should help to \"break the ice\" before they have to give a speech independently.       http://www.abacon.com/pubspeak/  I will refer to this site as  urlLink Putting It All Together .  I plan on using this site as a sort of WebQuest.  It\'s very user-friendly and gives the students help by starting with a topic for their speech and then directs them to a finished product.  I don\'t expect that they will utilize every aspect of this site because it is fairly extensive.  I will help them by pointing out the most helpful aspects.  '),(11165,3569294,'16,June,2004','2004-06-16 00:00:00','   hey, i was having a diffcult type figuring out this blog thing.  i final figured it out thanks carol dean.  anyway, when i first read the word blog, i had no clue what it meant. i figured out what it is now  and i think it is something you could use with your class.  They can use to contact another student about some homework or just blow off some steam about a test or something.  the students could use it to ask questions about a topic, so the student wouldn\'t feel ashamed of the quesion, and no one would know who it was.  '),(11166,3569294,'14,June,2004','2004-06-14 00:00:00','   Do you love words and the power of words?    Just want to pass on a website where you can subscribe for free to \"A Word A Day\". My sister started my subscription a few years ago. Try out  urlLink www.wordsmith.org   Each day you get a new word, its pronunciation, definition, origin, 2 examples of its usage in printed documents, and a quote a day. The website also contains archives by month issued, theme and words. Its been around for ten years.   If you haven\'t heard of it before, check it out and let me know what you think. -Deb  '),(11167,3569294,'11,June,2004','2004-06-11 00:00:00','   What\'s up with my posts displaying the \"delete\" and \"edit\" action on the sides of the subject line? What am I doing wrong in the publishing process?  '),(11168,3569294,'11,June,2004','2004-06-11 00:00:00','   Hello classmates! I\'m here, I\'m late, and for those of you that know me talk way to much.  Probably taking a class that I have to type my thoughts to you will definitely slow down my wandering thoughts.  Good to have you all aboard and hope that we can all learn something from each other.  Until next time  Bill Taylor  '),(11169,3569294,'10,June,2004','2004-06-10 00:00:00','   I am here!! I am not sure what I am doing, but I am here! I have never used a blog before. It appears to be much like an interactive Buletin board / chat room service. I think it could be useful for students if managed correctly. They will need detailed instructions about what a blog is and how you wish them to use the blog. My husband says if I see a blog to approach it very carefully and capture it alive if possible. Ha-Ha!! Have a good day! Arlene  '),(11170,3569294,'09,June,2004','2004-06-09 00:00:00','   Hi Guys~ I\'m still a little lost as to what exactly it is we are supposed to be doing on the blogger site. So I\'m just going to go along with what some of the others are doing. I agree that to use a blog in the classroom may be a little difficult unless it is an upper level class such as Jr/Sr high school. An idea on how to use one could be to have discussions on the blog as part of a weekly assignment. They could post their thoughts and ideas about the readings that they have been doing or any questions that they may have. I also really liked the idea that someone else had about the principal using blogs to communicate with teachers. That could definitely be useful. I think it\'s going to be interesting to see how these blog posts go! I\'m looking forward to it. Keli  '),(11171,3569294,'09,June,2004','2004-06-09 00:00:00','   I am anxious for the day when I have summers off and don\'t have to do homework.  These classes are greatly beneficial to my personal and professional development, however; the pool is calling my name and a road-trip sounds good.  I have been able to sleep late once or twice this week.  What a luxury. Actually it is 10:30 and I am still in my pj\'s.  Does that make me a bum? Have a great summer you guys and gals!  I wonder what my little darlings are doing this summer?    Mary  '),(11172,3569294,'09,June,2004','2004-06-09 00:00:00','   I am anxious for the day when I have summers off and don\'t have to do homework.  These classes are greatly beneficial to my personal and professional development, however; the pool is calling my name and a road-trip sounds good.  I have been able to sleep late once or twice this week.  What a luxury. Actually it is 10:30 and I am still in my pj\'s.  Does that make me a bum? Have a great summer you guys and gals!  I wonder what my little darlings are doing this summer?    Mary  '),(11173,3569294,'09,June,2004','2004-06-09 00:00:00','   Greetings from Outer Space,      This is StarBlog 001, recorded by kitpie13 ( known to some, but not all, as Teresa Wiles). Today is Thursday, July 10, 2004, and we have successfully completed our first mission : to develop a web-based e-mail account. It was with much trepidation that we set forth upon this venture, as it gives the dreaded e-mails even greater vantages from which to strike. Currently e-mails can menace us from six directions at once (2@job#1, 1@job#2, 1@home, 1@ASU, and the last one possesses the ability to attack from any location that is web-accessible). We then performed the bio which was long and grueling, yet strangely addictive. We expect to be forced into feeding our strange fixation at some point in the near future. Our captain has also give the order to create a new creature known as a  Blog , and it comes to life even now. Before the  Blog  overwhelms our dwindling forces, I shall attempt to preserve for posterity my last remaining thoughts... Obviously the Blog could be used as an excellent story starter, with students adding their own endings or being required to continue the story where the last student left off; thus creative writing could be reborn in the classroom. It could also be used as a log for checking and confirming that students are actually using their given times for their individual assignments, such as AR(if someone writes that he/she has read pp.73-80 in his/her library book, then I don\'t expect to see that person back on pp.66-72). In an assigned class reading, students could make up multiple choice questions with well-written distractors. Each student could compose one question (no duplicate questions allowed), and the teacher could require students to answer x number of questions. This could lead to great discussion in the classroom as students questioned each other or defended their answers. Other thoughts enter my mind, but the  Blog  is greedily consuming them even as it.....consumes....my...brain .(Audience hears a muffled groan,and then a thud as the StarBlog recorder\'s body hits the floor. The hiss of static is the last thing heard by the audience.)  '),(11174,3569294,'07,June,2004','2004-06-07 00:00:00','   Okay, so like when does one know one is blogging? Did I blog when I answered those questions about myself? Am I blogging now? Or, perhaps I\'ll blog when I look at those websites and comment on them? Wait a minute, I know what bloggy is!  It\'s that weather condition that necessitates the use of one\'s low beams.  '),(11175,3569294,'07,June,2004','2004-06-07 00:00:00','   Yes, that\'s it!  I need a synonym for blog.  Any ideas? Erin Farrington  '),(11176,3569294,'07,June,2004','2004-06-07 00:00:00','   Hi everyone,  I am here!!  I am not sure what I am doing, but I am here!  I have never used a blog before.  It appears to be much like an interactive Buletin board / chat room service.  I think it could be useful for students if managed correctly.  They will need detailed instructions about what a blog is and how you wish them to use the blog.  My husband says if I see a blog to approach it very carefully and capture it alive if possible.  Ha-Ha!! Have a good day! Arlene  '),(11177,3569294,'07,June,2004','2004-06-07 00:00:00','   When I first saw the assignment, I thought a blogger might be something you smoke (hee heee..) or a virus (i.e. \"I have the blog\"). However, it is not. I think a blogger could be used effectively for teachers and administrators. Students could edit their profiles so teachers could get to know them on a more personal level. Assignments could be posted. This would be great when weather is bad, a student is absent, or students could work ahead if they knew about a planned absence (i.e. vacation). Principals could communicate with teachers during the school day provided each class has a working computer that does connect to the internet. The Bowman Blog could be a catchy name for my blog.  '),(11178,3569294,'07,June,2004','2004-06-07 00:00:00','   A special welcome to Lynn,Chad, Deborah, and Jerry for creating and posting to the class blog.  I must admit that I wondered what we were about to get into when I was asked about \"blogging\"!!!  I liked your thoughts about how to use a blog.  Do you have any other ideas about how you would use the blog in your classes?  carol deal  '),(11179,3569294,'07,June,2004','2004-06-07 00:00:00','   I just created a blog and it was easy.  '),(11180,3569294,'07,June,2004','2004-06-07 00:00:00','   This the first time I ever heard of blogging. With dial-up access you tend not to spend a lot of time conversing over the internet. I still prefer a telephone or face-to-face conversation - you can\'t read body language or voice inflection in a blog!  It can serve a purpose but too often we get caught up in the technology bells and whistles and we lose sight of the human social interactions that are being replaced by technology connections.  '),(11181,3569294,'07,June,2004','2004-06-07 00:00:00','   I must admit that when I saw the directions from the instructor regarding a \"blog,\" I began scratching my head and wondering what the heck she was talking about. I may be a bit behind in this technology-savy world, but I had never heard of it before. After reading further, I discovered that it is very much like a public discussion board or forum...maybe even identical with a unique, eye-catching name for it. I think it\'s a great idea for people who have similar interests or have some other kind of common bond. I too think it could be a good tool for a class, but it needs to be relevant and useful...not something to use just to say you use it. Teachers must consider the types of, the ages of, and the academic level of students, among other things, before incorporating this resourse, if they want to use it effectively. OK...so I\'ve voiced my opinion and made my post! Now, I\'m moving on to something else.  Chad  '),(11182,3569294,'07,June,2004','2004-06-07 00:00:00','   I have heard of blogging and have often wondered how one could incorporate this technology into the classroom experience in Language Arts.  It seems as if it would be a wonderful way to inspire reflection and dialogue about interaction with text among students but the big question, as always, would be about making it meaningful.  I have to admit that, with a few notable exceptions, my experiments with technology in the classroom have led to somewhat mixed results.  Students either digress too much or get so excited about the minor points of the process (graphics, sound) that real interaction with the content is remote.  I\'m wondering if anyone has successfully incorporated blogging into their content area and, if so, how they did it.    Lynn  '),(11183,3569294,'07,June,2004','2004-06-07 00:00:00','   http://www.archives.gov/digital_classroom/teaching_with_documents.html urlLink  '),(11184,3569294,'04,June,2004','2004-06-04 00:00:00','   http://www.marcopolo-education.org/ urlLink  '),(11185,3569294,'04,June,2004','2004-06-04 00:00:00','   http://www.ncwiseowl.org/ urlLink  '),(11186,3569294,'04,June,2004','2004-06-04 00:00:00','   http://www1.appstate.edu/%7Emoormang/umiami_6_03/prof_resources.htm urlLink  '),(11187,3569294,'04,June,2004','2004-06-04 00:00:00','   urlLink http://www.sasinschool.com/  '),(11188,3569294,'04,June,2004','2004-06-04 00:00:00','   blogblog  '),(11189,2259900,'01,July,2004','2004-07-01 00:00:00','    urlLink MSN Search -- More Useful Everyday : MSN releases a preview version of their new search technology. They\'ve been making large strides to compete with Google and Yahoo. I\'m not sure of their algorythms or the factors that play into how they search. If anyone has any insight into this please share!   '),(11190,2259900,'29,June,2004','2004-06-29 00:00:00','    Yahoo now offers a Search Term Suggestion Tool (STST) to help you broaden your keyword selection search. Though it offers no indication on how many people are searching under each term each month as does Overture\'s STST, it still gives users a way to say something in a little different way and hopefully find what they are looking for.  Here\'s an example of a search and at the top there\'s a section called \"Related\" that offers suggestions.   urlLink Yahoo! Search Results for dna testing    Other Search Term Suggestion Services:    urlLink Overture\'s Search Term Suggestion Tool : This can be adjusted to find individual country results by going to:  urlLink Overture International  and looking in their Advertiser Center.   urlLink Google AdWords Keyword Suggestions    urlLink All The Web results for dna testing    urlLink AltaVista results for dna testing    urlLink Word Tracker    Each one of these keyword suggestion tools return slightly different results and gather their information from slightly different demographics and locations. Choose the one that\'s appropriate for you.   '),(11191,2259900,'24,June,2004','2004-06-24 00:00:00','    urlLink MSN Hotmail Takes on Google with 250 MB of Email Storage : Ahh yes. Rightly bellows Alex, \"Oh, the power of Google!\" Truly there is much at work here behind Google\'s announcement of it\'s 1GB of email storage. Yahoo jumped up screaming, \"hey, Email! That\'s what WE do!\" and extended to its account users 100 MB of space. Long gone are the times of trying to up-sell 2MB web mail users. Thank you Google   '),(11192,2259900,'15,June,2004','2004-06-15 00:00:00','    urlLink Yahoo to test its New Homepage - PakTribune : \"Yahoo has begun quietly testing a makeover of its home page, as evidenced by trial mock-ups seen by some Web users over the past few weeks when they accessed the portal.\"  Yahoo! is stepping up to Google\'s announcement of it\'s 1GB storage space for it\'s Gmail users. They\'ve announced that they\'ll give 100 MBs to their Yahoo! Mail users sometime this summer.   Yahoo! also recently revamped their Messenger Service to include the popular LAUNCHcast internet Radio and a Y! Search field at the bottom of it, something that\'s a new innovation when it comes to Instant Messaging.  They not too long ago also added a product search on their engine, a feature that mimics Google\'s product search Froogle. Yahoo\'s toolbar has for several months now included a \"WebRank\" that is an obvious knockoff of Google\'s PageRank, however, it acts more like Alexa\'s toolbar ranking in that it determines the ranking by aggregating Yahoo toolbar users\' browsing patterns.   With these changes and others, Yahoo and Google are begining to look more like each other and are pounding out the way we interact with the web. A little competition here is very exciting...I\'m liking the improvements on both sides.  Yahoo also rolled out today a new mail interface that matches its Messenger and that allows 10MB of email transfer and 100MB of storage. Thank goodness! I have been over my limit for several months. Kind of Yahoo! not to close that account. Kind of them to increase my storage to 100MB. Thanks Yahoo!   '),(11193,2259900,'13,May,2004','2004-05-13 00:00:00','    urlLink WebmasterWorld News and Discussion for the Independent Web Professional   I have this forum set as my homepage. Great news and discussions on things that are most important to a webmaster...or anyone doing anything online for that matter. Many times when I \"Google\" something I need help with I find my way to a discussion going on here at WebMasterWorld. While all of them do not directly solve my problem, many of them do and those that don\'t often send me to another website that does.  You can join the Supporters Forum or just broswe the open discussions. I often find that the juiciest posts that I want to read are restricted only to supporters. You\'re choice. I heard, not sure though, that the Supporters Forum discussions eventually trickle down into the regular discussions.  Happy posting!   '),(11194,2259900,'10,May,2004','2004-05-10 00:00:00','    urlLink Google\'s PageRank Explained : What is PageRank? Well come here for an overview and look for more articles that help out on this topic.   '),(11195,2259900,'08,May,2004','2004-05-08 00:00:00','    urlLink Search Engine Results Chart  Danny Sulivan does it again! Here he shows several charts in addition to the well know Bruce Clay Search Engine Relationship chart. Check it out!   '),(11196,2259900,'08,May,2004','2004-05-08 00:00:00','    Aggregators that run on your PC are pieces of software you can download from a website and quickly and easily install.  There are literally a few dozen different aggregators but here are some of the ones we\'re familiar with and you can easily download and try:   urlLink SharpReader RSS Aggregator : \"SharpReader is an RSS Aggregator for Windows, created by Luke Hutteman.\"   urlLink nntp//rss : Bridging the worlds of NNTP clients and RSS feeds, nntp//rss is an application that will enable you to use your existing favorite NNTP newsreader to read your information channels.   urlLink NewsGator  is a \"news aggregator\" that runs in Microsoft Outlook. It allows you to subscribe to various syndicated news feeds (such as weblogs, news sites, etc.) and have news from these sites be delivered right into your Outlook folders. There are thousands of sites which syndicate their content in RSS format, and many more being added every day. You can even read NNTP newsgroups with NewsGator.   urlLink RSS Bandit : a .NET RSS desktop feed aggregator.   urlLink Vox Lite  allows you to keep up-to-date with all your favourite sources of information that support the RSS protocol.   urlLink BottomFeeder  is a news aggregator client (RSS and Atom) written in VisualWorks Smalltalk. BottomFeeder runs on x86 Linux (also FreeBSD), PowerPC linux, Sparc Linux, Windows (98/ME/NT/2000/XP), Mac OS8/9, Mac OS X, AIX, SGI Irix, Compaq UNIX, HP-UX, and Solaris.   urlLink FeedDemon  makes RSS newsfeeds as easy to access as your email.    urlLink NewsMonster  is a news, weblog, and RSS aggregator that runs directly in your web browser. NewsMonster offers a superior web experience and outstanding integration with existing websites and weblogs that support RSS. Even sites that don\'t support RSS can work with NewsMonster.    urlLink Radio UserLand  is a unique and powerful web publishing and weblog tool that is used and praised by many of the world\'s most accomplished web authors. With advanced features such as an integrated news aggregator, Radio enables users to do far more than is possible with the hosted-only weblog products offered by others.   In general PC based aggregators have a cleaner, more elegant interface and offer more features than web based aggregators.  They also store your feeds locally so if you\'re traveling you have an archive of things to read.  Web based aggregators require a connection at all times.   If you purchase an aggregator make sure it can import and export OPML -- a standard for sharing the feeds to which you subscribe.  Without this you might be locked into a software package.  Yes the Feedster aggregator imports and exports opml.   '),(11197,2259900,'08,May,2004','2004-05-08 00:00:00','    urlLink News Aggregators: All About Delivery : \"Just has you have a browser for the web, there are browsers for RSS.  These browsers are called News Aggregators or \'Aggies\' because what they do is aggregate, i.e. gather and present new information.     Summary to date:  RSS is a way for you to have new information sent to you automatically.  The reader has complete control and can turn off any feed at any time.  This means there is never spam.    Within an RSS feed there are multiple chunks of information called posts.    An orange XML / RSS  (  )button or a text link like \'Syndicate\' lets you get a site\'s feed.    A News Aggregator lets you view your collection of RSS feeds.\"   '),(11198,2259900,'08,May,2004','2004-05-08 00:00:00','    urlLink Its All About those Little Orange Buttons... : \" Any website, weblog or even any ecommerce site can provide information to you as an RSS feed.  Even Amazon is publishing RSS feeds now, as are the New York Times, Yahoo News, Harvard, Microsoft and millions of others.     When a site has an RSS feed or feeds available you might see something like one of these:   Syndicate This Site (XML)  (RSS)    What these text labels and icons mean is \'This site has an RSS feed and here\'s the URL\'.  And if you see one of these then you can subscribe to that site and have its content delivered to you.   Not sure if a site has an RSS  feed?  Click here to use our FeedFinder which automatically discovers RSS feeds for a site.     Sites don\'t always label RSS feeds consistently so sometimes you have to really look for them.\"   '),(11199,2259900,'08,May,2004','2004-05-08 00:00:00','    urlLink What Is RSS ?   RSS or \'Really Simple Syndication\' is an Internet standard just like HTML (i.e. the web) -- with one key difference -- delivery.  When you use the web, you\'re constantly browsing, searching ... seeking information.      With RSS new information -- information that you select and information you can turn off at any time --  is automatically delivered directly to you.   This information is called a Feed since it feeds direct to you.  An RSS feed is sometimes indicated by [an]orange [XML] icon...    Within an RSS feed there can be text, links pictures and more just as you find on the web.  The different \'chunks\' of information within an RSS feed are called posts.   Analogy: Think of RSS as \'Tivo for the Web\'.\"   '),(11200,2259900,'13,July,2004','2004-07-13 00:00:00','   urlLink Yahoo! Help - : \"What are your guidelines on spam?\"   Again and again SEO firms attempt to \"trick\" the search engines into raising their clients\' rankings to increase their sales. Sure there\'s a need for this, however Yahoo! and other search engine giants continue to change their algorythms to eliminate bogus websites from their listings.  One point that Yahoo! tries to make again and again in their guidelines is to create page for humans, not for search engines. Tricks like duplicating your website with different urls, redirecting your visitors to other places, and using extensive linking to inflate rankings are becoming more and more under attack by the search engines.  Find legitimate ways to increase rankings and sales.  '),(11201,2259900,'16,June,2004','2004-06-16 00:00:00','   urlLink Google Ranking Factors List . When you talk about optimizing your website following as close as you can to some of the most common practices is best.  '),(11202,2259900,'15,June,2004','2004-06-15 00:00:00','   urlLink Search Engine Results VS PPC : \"Too many on-line marketers are overlooking natural search engine optimisation...\"  \"...across the top four search engines - Google, Yahoo!, MSN and AOL, which hold over 90% of the market share - more than 60% of users chose a natural search result over a paid search advertisement as most relevant to a sample query.\"  This means that Internet promotion is never over until you have good search engine rankings.  '),(11203,2259900,'13,May,2004','2004-05-13 00:00:00','   urlLink How Often Should I Submit My Site To The Search Engines? : Again, the answer seems to be now that search engines are more advanced it\'s not necessary to be submitting your site once a month or everytime you make changes to it. Trust in the power of the web. You\'ll be okay.   \"So the answer to your question is, it depends. If you\'re talking about Google, don\'t submit your website at all. Just do a few link exchanges with quality sites that are already ranked well with Google and let Googlebot \'find\' your site automatically.  If you\'re talking about Inktomi, Altavista, and some of the others, using their \'Express Submission\' services can speed things up considerably, but it isn\'t really necessary.  And remember, once your site is indexed in the search engines, re-submitting it will simply be a waste of time and/or money.\"  '),(11204,2259900,'12,May,2004','2004-05-12 00:00:00','   urlLink Submitting/Resubmitting to the Search Engines : \"How often do I need to submit my site to a search engine?\" Well, Jennifer Laycock from About.com does a great job of dispelling some myths that linger around from the days of less advanced search engines.  She points out that though submission is very much overrated these days simple tactics like getting listed in human-edited directories sudh as  urlLink DMOZ ,  urlLink Yahoo! ,  urlLink JoeAnt , and  urlLink GoGuides  will give engines a good place to find you when they get around to indexing your page.  '),(11205,2259900,'12,May,2004','2004-05-12 00:00:00','   Hey there! And welcome to Search Engine Friendly. This section of our blog is to let you in on the well known foundational elements of preparing or changing your site so that the search engines find and index them well. It\'s also a place where experimental ideas or little tweaks can be found. We\'ll try to let you in on which tactics are which.  We also promote responsible site optimization, so all the little cheats that you may have heard of before like including hidden text or stacking your meta tags with keywords are frowned upon here. If you\'ve practiced any of these methods before we understand that people can change and invite you to do so if you already haven\'t.  Site optimization is about communicating better. Communicating better with your customers, visitors and the search engines. Make your content relevant to them and everyone will be happy.  Let us know your feelings by commenting and check out our profiles if you\'d like to contact us personally.   Happy optimizing, or if you\'re in the UK, optimising!  '),(11206,2259900,'18,May,2004','2004-05-18 00:00:00','   urlLink PDF zone - The online authority for PDF and Adobe Acrobat professionals  '),(11207,2259900,'18,May,2004','2004-05-18 00:00:00','   urlLink activePDF Incorporated  offers activePDF Server, a server-side pdf creation solution as well as solutions for home and corporate file conversion.  http://www.activepdf.com/   7  '),(11208,2259900,'02,June,2004','2004-06-02 00:00:00','   Here\'s the beginning of my Bocce Ball Blog. Here will be the largest Bocce Ball Blog anywhere!  '),(11209,2033616,'31,October,2003','2003-10-31 00:00:00','    Yesterday... the impossible happened.  Kyle Cottrill wore COLOR... i repeat COLOR.  And not just any color.. but a bright orange.  many thanks to all who joined in my campaign to save Kyle from the boring world of white.  Kyle, you are very brave.  ~Long day... i\'m becoming loopy~   '),(11210,2033616,'29,October,2003','2003-10-29 00:00:00','    I found this quote.. and would like to dedicate it to the people that still don\'t believe i\'ll just never learn.  cough::KaylaJessica::cough::  There is no pleasure in having nothing to do; the fun is in having lots to do and not doing it. Mary Wilson Little   '),(11211,2033616,'25,October,2003','2003-10-25 00:00:00','    It seems like the whole world has turned upside down, and i\'m not talking about me.  My Aunt Linda has moved out of her house and into an apartment, leaving my Uncle Ed (my dad\'s brother) at home along with my cousin Jenny.  Darcy is so angry with ther mother, she says she doesn\'t want to go to the apartment.. she doesn\'t even want to know where it is.  Why is there so much unhappy relationships in this world?  NO ONE, i repeat NO ONE in my family saw this coming, never would have dreamed it happening to that family.  I really really feel bad for my cousins, i want to go down there right now and give Jenny a big hug and say, \"it\'s okay Jenny! We Love You!!!!\"  I wish we lived closer, i wish i got to see them more often.  ::sigh:: well it\'s back to the history notes now...   '),(11212,2033616,'25,October,2003','2003-10-25 00:00:00','    Well, i\'m done... let\'s get this day over with.  I had ACT this morning... hmm.. now that\'s what I call fun... NOT.  Actually, I talked to Brad during the break... but not only that, he talked to me.. which is even more surprising.  Stuff with him is confusing... we can\'t be friends without fighting, and there doesn\'t seem to be a happy medium between friends that can\'t get along and just never talking.  I thought we had found it a few weeks ago... but then who knows what happened and things went back to zero contact.  So it was kinda nice to talk to him again.  So far that has been the only good conversation of the day (it\'s only noon but this still says alot).    My mother greeted me when i got home with, \"I think someone has stolen the cell phone.... i left it in plain view in my unlocked car on the curb at the seminary.\"  She asked if she could borrow mine... no way.. she\'ll lose that one next.  I really don\'t see how my parents can be so hypocritical.  My father has the nerve to tell all my life, \"don\'t eat or drink next to the computer... you  could  spill it, and it  could  ruin the keyboard, and we  might  have to get another.  Then he goes off and walks into the Gulf of Mexico with a fairly new cell phone in his pocket!!!!  We hadn\'t even had it a year!  Well now my mother.. who has always said \"lock your door lock your door lock your door\" and is repeatedly telling me to keep my car locked.... leaves ANOTHER CELL PHONE (this was the replacement from last time... meaning it was only 3 months old) in her unlocked car.  She has power locks... she could be 10 yards away and still lock it.  My parents are rediculous.  Lea is also rediculous... I WANT MY SWEATSHIRT BACK!  it\'s not like it\'s my ONLY sweatshirt or anything.. but it\'s mine and how in the world did it get to Volga????? this is how  that  phone conversation went:  Me: so Lea.. you know that sweatshirt I loaned you when you were visiting.  Do you know what you did with it? Lea: oh yea I have it Me: You do!! good i have been looking for that everywhere.  I\'ve torn the house apart! Lea: Oh yeah i wore it to school yesterday.... why..? do you want it back... Me:hhmm...... well...  YES OF COURSE I WANT IT BACK   okay so i didn\'t say that but for crying out loud... did she think i gave it to her to have FOREVER??  furthermore... she hasn\'t and isn\'t making any effort to give it back to me anytime soon.  She refuses to mail it and says she doesn\'t know when she can bring it back... but \"i want to come to town so maybe when i come and see everyone.\"  I\'m sorry Lea... but i don\'t care... just give me back my sweatshirt!   '),(11213,2033616,'24,October,2003','2003-10-24 00:00:00','    One gigantic hooray for my Kayli!!!!! She made All-State... 8th Chair First Violin!!!! tendonitis and all...   Whooo hoo.. hats off also to Jenny Ulmer first chair Flute,  Jennilee first chair french horn, and Laura McCarty 4th chair Cello!!!!  Off to celebrate w/ Kayki   '),(11214,2033616,'23,October,2003','2003-10-23 00:00:00','    I hate school. says: yeah I hate school. says: i am stressed to teh max Hicup says: i know a lot of people that are... teenagers have too much stress... they\'re gonna have to create some pretty heavy duty blood pressure medication before we get to middle age or our generation (well the one\'s that get anywhere anyway) will just start dropping like flies   '),(11215,2033616,'19,October,2003','2003-10-19 00:00:00','    I think Courtney is about 7 or 8 years old.  She lived 2 houses west of us and her parents drive humongo huge SUV\'s.  I talk to her about twice a week.  While I was in Okoboji this summer... she came to our door asking \"to play with the girl on the pogo stick.\"  MK spent some time with her that day and when Aliesha dropped me off at home that night there was a big chalk drawing on our driveway... compliments of Courtney and MK.  The first time I met her... Sam was here, we were working on Calling the freshmen.  She came to the door, which i answered... very confused to meet a little girl asking for my sister.  She asked me if I could play, and i gently told her that MK wasn\'t home and I was busy.  She came over 3 more times that afternoon... finally Sam and I walked she and her electronic dog back home.  She has come back to the house about twice a week since then, asking for MK.  Each time I\'ve had to turn her down.  She has a brother her age, and a friend that lives just a couple houses away.  But she still comes over here looking for Mary Katherine.  She came to my door 10 min ago, and after the usual... \"MK is not here\" routine, she sheepishly asked, \"is she ever going to get to play again?\"  I feel like the bearer of bad news everytime she comes to the door.  Poor Courtney   '),(11216,2033616,'19,October,2003','2003-10-19 00:00:00','    LOL Kayla.. it wasn\'t a glare.. it was just a look lol.   '),(11217,2033616,'18,October,2003','2003-10-18 00:00:00','    A timer dial moves 10 degrees every minute, (starting in the 12:00 position).  How many times will the dial rotate a full circle around the clock in  h  hours? (answer is in relation to the variable  h  ex. 6 h )  Does anyone know??? It\'s driving me crazy!!!!!  So i took the PSAT this morning... I think i prolly did better than last year... although I still didn\'t study a whole lot either.  Then Kayla and i went to the mall an cruised the formal dresses for a while.  And now i sit here... kinda bored... kinda tired... but hey... minimal homework this week (clap clap clap).  So... now I have zilch to do.  I have realized a not-so-encouraging pattern in my life... and have found it\'s kinda a problem.   Ask me about it if you wish, otherwise I\'ll keep my mouth shut.    So... turn off the lights, and close the doork, that\'s all there is... there isn\'t anymore.   '),(11218,2033616,'16,October,2003','2003-10-16 00:00:00','    I have two tests tomorrow and I can\'t concentrate.  My mind keeps wandering.... someday it may just walk off and leave me.  I have PSAT\'s on saturday... ooo Hansen will have my head if I don\'t do well...  Let it be known I\'m adding Jennilee\'s link to my blog before Sean... haha... JenniLEEEEEEEE!  wow i\'m done   '),(11219,2033616,'15,October,2003','2003-10-15 00:00:00','    I am still glowing after yesterday\'s trip down to Sioux City.  Yesterday at 8... i woke up and did history notes (enthraling i know).  Then i got ready to go and left around 1:30.  I got there at exactly 3 o\'clock... and Aliesha and Bret and I couldn\'t find Jack anywhere.  So after Aliesha left for swim warm ups... Bret and I wandered around the school looking for Jack.  Well we finally just called \"his wife\" to find that he was on his way home.  He had waited at the wrong door... what a kid that Jack Farley.  So Bret and I hoped in the car and drove to Jack\'s where i got a warm welcoming from Jack and Amanda (aka, the wife).  Well okay, it was Jack that welcomed me... I don\'t know if Amanda thinks too much of me.  We sat around for a while, then headed off to the school to the swim meet.  We got to hang out with Anne (Anne\'s thinking about going to the Y2J2 camp session this summer which means I am THERE).  Amanda left for volleyball eventually... (it\'s not like i don\'t like her but i feel wierd hanging out with Jack when she\'s there... I don\'t want her to feel like i\'m crossing on her turf or anything... but I don\'t get to see Jack but every other month).  So then it was just me Bret and Jack as Aliesha and Anne came back and fourth between swimming and hanging with us.     So after the swim meet we hugged the very wet Anne goodbye and went across the hall to watch Amanda\'s v-ball game.  Well until Aliesha was ready, then we peeled Jack away and headed to the mall for some much needed Chick-fil-a.  No matter what Jack says I\'m not using them for their Chick-fil-a :-P... but it is a plus :-).  We brought the food back to Jack\'s house and ate... Jack said he\'d take me to my winter formal... you know... the usual...   After one last bear hug from Jack, I was off.  I took Bret home... and then Aliesha and I sat in his driveway for the next 5 min saying... \"THIS HAS BEEN SOO FUN OH MY GOSH!!!\" we finally said goodbye for the last time (isn\'t it sad).  And I was off.  I got back home, talked to Jamie on the phone.. things are settled... out in the open and things will get better i think.  Today was awesome b/c i was still on cloud nine after yesterday.  And I just beat Chad for the first time ever in internet checkers.. haha!!!  Time for a rematch and then i\'m off to bed.  Sweet dreams everyone... may the horse be with you ( I know it didn\'t make sense but that\'s a good song :-D)   '),(11220,2033616,'14,October,2003','2003-10-14 00:00:00','    I love my parents to death.... today anyway.  YES!!!!!! I\'m going down to see Jack Aliesha Bret and Anne today and I couldn\'t be more excited!!!!  Aliesha, Jack and Bret all called last night and everything is set, so at 1:30 I am OUT of this stupid city.  Yesterday was not a very good day.  I sent a not so nice e-mail to a very beloved person and I feel horrible... I sent another, apologizing but I still feel like a jerk.  I\'m still upset with everything though, so maybe this is a good thing... everything will be out in the open. ... wow that was vauge  So what better way to cap off a horrible weekend but with an AWESOME day down in Iowa w/ my buds from Camp?  I can tell you\'re excited for me... don\'t try to hide it.  So just a few more history notes to take (the joys of procrastination) and i can hit the road.   '),(11221,2033616,'13,October,2003','2003-10-13 00:00:00','    So what is Miss Lucy\'s Gossip Parlor you might be asking?  Well, it\'s the name of this quaint little coffe house in Sandestin, Florida.  It is located in a bright shopping center just off the beach and is the perfect spot for ice cream or gormet coffe, as well as a handy place to check your e-mail (it is easy to see why i visited so often).  The front of the store had little gifts and knick knacks for sale as well as some antiqued sofa\'s and chairs to accomidate the \"gossipers\" that happened to strole by.  I found the place very charming... and so i stole their name :-D thanks guys!  Well so far this weekend has consisted of fire (don\'t worry everything was under control), hair cuts, shower curtains, Chicago, curling iorns (used as a straightener), week wackers, and lots of Spanish.  If everything goes as planned i will have seen 5 people from camp(s) by wednesday.  Crad drove down to visit us on Sat, I sat w/ Becky in church today, and on tuesday if everything goes w/out a hitch, i\'ll get to see Aliesha Anne Jack and Bret.    Life is hard and people are even harder.  Ho hum to the people that are being stupid and three cheers for the people that are here.  I\'m late it\'s tired.  Goodnight my someone.  signed, Un-frizzy Lizzie (for tonight anyway... and I\'VE GOT PICTURES to prove it)   '),(11222,2033616,'09,October,2003','2003-10-09 00:00:00','    It has become apparant to me that there are people in this world... that I don\'t like.  Would someone be so kind as to drop them infront of a speeding semi?  Thanks  On a side note, i hate people that say \"hello, i\'m different\" just to say \"I\'m different\".  People need to do something b/c they want to do it.. not b/c everyone else is not doing it.  Having your own sense of style is one thing, but there is going to be someone out there who is just a little like you.  Call me a lemming?  lemming my foot... I\'m doing something that  I  want to do b/c  I  want to do it.   '),(11223,2033616,'07,October,2003','2003-10-07 00:00:00','    This is my rant of the day... the wolrd needs to grow up.   Yesterday was not a good day, again... people need to grow up, esspecially high schoolers that act like 3 year olds.  But yesterday was not a good day... so i don\'t think I will coment on it... it made me very very very angry.  However, there were a few highlights... MK has at least 3 sisters and Adam\'s name is Alfred, that was it.  Today was a bad morning as well... horribly bad, unfathomably bad, but i lived.  There was a religion debate in Lit and Comp today (the world needs to grow up).   '),(11224,2033616,'07,October,2003','2003-10-07 00:00:00','    I cannot be creative tonight... All my creativity has been drained from me.  I have been doing homework all afternoon and evening and now.. look.. night.  I\'m going to need about 4 days to recouperate from this week.. oh wait.. i DO get 4 days, lucky me.  Hmmm... anything eventful happen today?  nope, nuthin.  OH WELL sorry to dissappoint, but there is studying to be done... although i\'m not so sure it will get done, but it should.    OH GOODNIGHT FOR GOODNESS SAKES... WHY AM I TORTURING YOU LIKE THIS?   '),(11225,2033616,'06,October,2003','2003-10-06 00:00:00','    Today was, there\'s not really a word to describe it.  I don\'t have good days, i don\'t have bad days.. they\'re just days.  Today was a day.  I got to school, not early as MK wanted but oh well.  I hung out with buds in the J-lab as usual, then Kayla and I went upstairs, passing Chad surrounded by a bunch of people that didn\'t want me there anyway.  Yes, then I had Precalc... and Eric did NOT tell me to shut up today, so that was encouraging.  I don\'t understand why he\'s so mean to me, I am really rather normal around him and I have never done anything to him so i don\'t see why he acts like such a jerk.  Lit and Comp was uneventful, except for Jessica playing \"in focus, out of focus\" which was funny.. and me and Amber making faces at each other from across the room.    During reading period i came across an interesting quote in I Kissed Dating Goodbye.  It was telling this story about always being positive; (i doubt Mrs. Benz would approve of that semi-colon) This couple is having financial trouble and when the husband gets home his wife has kitchen lit with candles and says \"we\'re going to eat by candlelight tonight.\"  He soon realizes that actually what has happened is they have had their power cut.  The wife, in tears, tells him that she didn\'t want him to come home to the reality of his faliure, and that they will get the power back again but for now \"let\'s just eat tonight by candlelight.\"  Joshua Harris ends by saying \"I\'m looking for someone who will light candles, not just curse the dark.\"  How true.  So then in German... it\'s German-American week... or something like that?  So we\'re doing a bunch of zero brain power stuff.  In chemistry we had a great laugh over our labs that the stupid study hall burned to a crisp.  7th per. AP Chem sure seems a lot more eventful than ours... but ours IS funny.  In ULE i actually did study for the PSAT today.. well sorta.  Then Kyle stole Hanson\'s \"mindtrap\" cards again to read in AP History... which besides that, was not fun.. lots of talk talk talk talk about stuff that\'s not all that fun.  Mrs. Leistra made an anology to siblings which Kyle found esspecially entertaining.. considering how MK and I get along in ULE.  Then there was Band.. i didn\'t have to play :-D.  And then the most exciting part was going home... I did homework and then had a March of Dimes Chain Reaction meeting.. which was entertaining.  Abby Hitzman and I are pretty much the only people from Roosevelt there, except for some sophomores, and Lord knows how they got selected for it.  Cam Stalheim must have actually remembered who i was this time b/c he waved goodbye to me.  That kid is just as goofy as when I knew him.  Then I came home and watched some MTV (televison? MTV? what has come over me?)  and my most favoritist Real World member is finally on one of the challenge shows.  Yes that\'s right.. favorite from when i actually watched MTV.  He\'s prolly gonna get kicked off soon though b/c he actually has standards and faith and all that and the rest of the cast is just a bunch of hot-headed, foul-mouthed, jerks.  But nevertheless he\'s there.  It\'s amazing how i can turn such a boring day into such a long blog... wow.. why didn\'t i think of this earlier.  Well I\'m going to go read \"Fall of the House of Usher\" which (call me chicken) is really quite chilling so i don\'t really know why i\'m willing to read it.. but oh well.   Happy dreams and good mornings to all, See ya on the flip-side! ZZZzzzzzzzzzzz   '),(11226,2033616,'05,October,2003','2003-10-05 00:00:00','    Okay... so i lied, nothing interesting happened in Crossroads... oh but Lindsey did pull a Jamie and ripped her pants.  MK and i went shopping before youth group, we got 1 1/2 pairs of shoes between us (one each and we\'re sharing the third), so maybe there are advatages to having the same shoe size.  We got to church and I compleatly forgot that we were gonna be doing music.. so then i got to sing some more.  Somehow Lindsey and Jamie got Cory, Eric, and one of their friends to come to church, which only added to my frustration of being left out of the loop.  But what can you do.  I was extreamly hyper and very outgoing tonight... which might have been contributed to the fact that Brad and Evan didn\'t show, and Craig and Collin came late.  Not that I don\'t want them there, any and everyone should show for youth group, but it made me feel like i could say something without being ridiculed.  So that was youth group... oh and lindsey ripped her pants.. did i mention that?  Well then i had bells.  Yes and then bells was over.  And then i went home.  Boy is my life boring.  Which reminds me i\'m sleepy, probably because of boredom... goodnight.   '),(11227,2033616,'05,October,2003','2003-10-05 00:00:00','    The creator of the first web log was probably a procrastinator.  He probably did not have enough things to do during the time he was avoiding writing things like lit crits.   Thank you Mr. Blog for occupying my time.   '),(11228,2033616,'05,October,2003','2003-10-05 00:00:00','    Well so far today I\'m exauhsted.  And it\'s only 1 o\' clock.  I went to sunday school... which seems pretty much pointless.  I think the mindset of the class is good... \"lets talk about issues that you deal with everyday\"... but without Tom and Katherine there... there\'s no back-up and there\'s alot of empty dissucsion.  They come back soon though so hopefully it will get better.  After SS Kayli and I ran upstairs to practice bells.. then downstairs to sing the prelude ect... then back up stairs after the sermon.. then back downstairs to sing for the closing... wow.  My friends all deserted me after church so I am now here... waiting for my chicken soup to cool (which i accidentally boiled) so i can eat it.  I am so bummed... I thought for once i would be the only one in the family not to catch the cold.. but i think one is sneaking up on me.  Back to everyone deserting me, thanks guys.  I always get the feeling i\'m just a third wheel with them sometimes.  Or now that all four of them are at Wash. a 5th wheel i guess.  Oh well.. I\'ve always got my friends at school and Kayli\'s always gonne be there for me... luv ya guys.  My soup is done so I\'m going to eat it, soup is delicious.  Now don\'t you worry, I still have Crossroads tonight and considering the path that this day is taking.. i\'m sure  something  interesting will happen.   '),(11229,2033616,'05,October,2003','2003-10-05 00:00:00','    Well today was a banner day for Lizzie.  First.. I woke up on the RIGHT side of the bed this morning after a not-so-good night, then, i got to see Aliesha today!!! Three cheers for the Festival of Bands for bringing in Sioux City\'s marching band today.  So Aliesha and I met for lunch at the mall, then afterwards Mary Katherine (whom from this day forward will be refered to as MK for my own convienience) and I hopped on over to Howard Wood.  I was greeted at the gate by Kurt and Ross... which were probably the last two people i had expected to see there.  As usual Kurt wore my jean jacket for about 2 min (and i still don\'t know why he does that) and when i got it back MK and i went inside.        I ran into Carrie and chatted with her a bit, then said \"hi\" to Chel and Beth.  I would have talked to them longer but MK was running around like a chicken with her head cut off (not really but it\'s a fun expression) so I thought i should stick with her or I\'d never find her again.  We went with Nate to the other side of the field.... meaning we watched the bands backwards... until Chad called.  It was perfect timing because Nate had to leave, so then we went to find Chad and sit with him on the right side of the field.  So after running around in circles with Nikki (Chad\'s little sister who we ran into on our way to finding him) we finally got to sit down and watch.  The bands were all great, and Roosevelt was spectacular as usual, i just love all our visual stuffs.        So then we left ... but we didn\'t really leave b/c we ran into Ross and Kurt again (and once again i handed over the jean jacket for a whole 2 min) and then hung around some of the bandos before leaving.  Well MK and I didn\'t feel like going home but we DID feel like eating... so we figured we\'d go visit Chad at the Fryin\' Pan and get some Pie.  So we got some pie (pumpkin, my favorite) and brought some home for mommy.        Then, to top off my day... I started a Blog, orders of Kayla.  Thank you Kayla, you\'re an inspiration.  And danke Sean for all the html help.  Well guten Nacht, sleep well and don\'t forget to brush.   '),(11230,2033616,'05,October,2003','2003-10-05 00:00:00','    So this is a Blog hu?  Well hello Mr. W.W. Web.. it\'s nice to meet you.   '),(11231,2033616,'30,November,2003','2003-11-30 00:00:00','    Please forgive me for the ranting in my last post.  Anyway, this past week has been WUNDERBAR (wonderful duh!).  Lots has happened... and if you are reading this you mostly likely already know what I\'m talking about.  So no need for me to go on now is there?  Thanksgiving was great.  For all those ppl who hate going to see grandparents everytime there\'s a holiday, or being forced to spend time with annoying family members all the time, you don\'t realize how good you have it!  I got to see cousins that i havn\'t seen for 2 years or more!  The saddest thing is that, because MK and I are the oldest, most of our cousins hardly know us.  My second cousins (i think second once removed or somthing weird like that) are just like me and my sister.  Katherine is a 6 grader and reminds me SO MUCH of me, while her little sister Rebecca who is in the 4th grade, is basically Mary Katherine\'s personality double.  So it was very nice to see family that we\'ve lost touch with.    The only bad thing about this week.. is my voice... or lack there of.  I have a fever and I sound like a frog, when you can hear anything at all that is.  It wouldn\'t really be that big of a deal to me--sick, stay home, get better go back-- but my voice recital is saturday and recovery of my voice by that time isn\'t looking so promising.  I wonder if it\'s in my subconscious, b/c i got strep throat the day before my recital back in the 5th grade.  oh me oh my.  Oh well... Que Sera Sera   '),(11232,2033616,'28,November,2003','2003-11-28 00:00:00','    i had this nice long post about my cousins and thanksgiving and Mark Everet\'s photography skills and all that fun stuff but NOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO humph!   '),(11233,2033616,'25,November,2003','2003-11-25 00:00:00','    arg.. stupid Blog   '),(11234,2033616,'25,November,2003','2003-11-25 00:00:00','    Well tomorrow (as well as a trip to go see Samford University) we have phase one of our Thanksgiving Dinner... much different attendance from the actual thanksgiving dinner. also EVERYONE is coming into town tomorrow so it will be a packed house.  The Guest List for Turkey Day is as follows  Grandparents (2), Me, MK, Aunt Jennifer, Uncle Brad, Johnathan (5 years?), Anna and Katherine (both 3 years), Emily (5 months).   and then those staying in motels... Mumsie, Dumsie, Aunt Karen, JR, Trey (13 years), Nicole (10 years), Aunt Diane, Aunt Liba (Elisabeth), Uncle Brett, Blake (6 years), Uncle Bill, GreatGrandmother Shepard... and I do believe that\'s it  Needless to say, this is going to be quite some gathering.. and for the first time since Aunt Liba got married (which meant a substantially smaller number of little people running around) the Leslie\'s will be present... Isn\'t this exciting?   '),(11235,2033616,'22,November,2003','2003-11-22 00:00:00','    Hello from Birmingham Alabama (the HEART of Dixie... oh i wish i was in the land of cotton.. oh wait.. I AM!!!!!!)  Well this past week has been very eventful.  Lot\'s of deep conversations and late nights.  But back to now.  I have noticed that i start in with this slight southern twang whenever i\'m down south.  I\'m staying with my Grandparents, as most of you know, and today we went to the Zoo.  The Zoo yes i know but it was fun.  Wednesday I\'m visiting Samford, and today i bought an Alabama Crimson Tide t-shirt (woo woo for Bama)  The Bama v. Auburn game is tonight.. and in Alabama.. this is a BIIIIIG thing.  So tonight my Unlce Bill and Granddaddy are gonna be watching the game.  Yesterday\'s flight went by fairly well.. i was so tired i sleept just about the whole time.  Well that is all i really have to say for now.  Love Y\'all (see look at this southern influence... I LOVE IT DOWN HERE)   '),(11236,2033616,'18,November,2003','2003-11-18 00:00:00','    My first voice recital..  actually let me rephrase that, my first voice recital that i will be attending.  I took lessons as a 5th grader to, but i got strep throat the day before... so i obviously didn\'t sing.  Anyway  ... is on December 6th at 10 am if anyone cares.  And right now i can\'t decide between singing Se tu M\'ami... that\'s italian btw... or No One is Alone, from Into the woods.  They\'re both fun, No one is alone is prolly easier... being in english obviously.... but i can\'t decide, i love them both!!!   '),(11237,2033616,'15,November,2003','2003-11-15 00:00:00','    Well now that i\'m not dead tired, i will talk about my life... actually.  Speaking of which... i really really want to see that movie, Love Actually, oh but it\'s rated R... and there are only so many people that would be interested in going, or able to go for that matter, so sad... and takers????  Back to my week Tuesday.... well tuesday was Sam\'s birthday.  Sam and i havn\'t been speaking for a couple of weeks, however it wasn\'t just her... none of my friends had been talking to me for some time, no one really knows what happened.  Well I was delivering Meals on Wheels with my mother on tuesday, there being no school, and we were dropping food off at a house just two blocks from sam\'s.  It hit me that Sam\'s birthday was going to go by with out a word from me about Happy Birthday.  I hold birthdays as a very very important day, and i feel like it\'s a day that should make you HAPPY, to be you.  So i wasn\'t going to let the fact that i wasn\'t invited to her party, or that i was upset with her, keep her from not hearing that... even though i was upset, i still cared.  So, my mom dropped me off, i went to the door... the doorbell wasn\'t working so her dad, who was outside, told me to just go in, they\'re in there.  So i did, i opened the door and was facing Sam\'s mom, sam aris jamie lindsey and megan, it was like being on trial.  I told them hi, and wished sam a happy birthday.  At first i felt like i was getting this look of \"why are you here\" but then when i said happy birthday, she actually smiled... so i thought that was a good sign.  I told her I was sorry, whished her a happy birthday again, and left.  I was shaking for the next 10 minutes.    It apparantly really really surprised Jamie and Aris and I have hung out with them every day since.  However i havn\'t heard from anyone else, not even sam.  Apparantly she saw it as some act i put on to try and upstage her birthday party or something.  I had no intention of doing anything but saying I\'m sorry for the last time, and wishing her a happy birthday.  If she wants to accept it, that\'s up to her, but i\'ve done all i can do.  So, Wednesday, voice lesson.. which went really really well.  I\'m going to be in a recital on the 6th of December at 10 am... more on this as it develops.  After lessons i went to Jamie\'s house, we had planned to talk things over, but it kinda seems that a combination of sunday and tuesday sorta cleared up things.  I\'m sure we\'ll talk about it eventually.  After Jamie\'s, i went to the Cell group at Sara\'s house.  It was a lot of fun, i got to see Sara, and it was kinda cool seeing how well another cell group was working.  Kudos to Josh harmon for inviting me.  Thursday... MY FORMAL DRESS IS HERE!!!! woo woo.  It\'s so pretty! i love it. i put it on 3 times, once to see if it fit, then twice and three times for aris and jamie, who stopped by to hang out.  I love it I love it I love it.  Oooo i love to get all dressed up.  I can\'t wait for formal, now if only i could figure out who i\'m going with :-/  (Jack\'s not the MOST reliable guy out there, i don\'t know if it would all really work out)  Friday, it wasn\'t the greatest day, but i was still happy... for some reason.  Isn\'t that I sign of insanity?  Anyway, laura and i got in an argument with Kyle.  It was about 3/4 just joking around, but it still is bothering me.  I get the feeling he really does think i blame MK for everything.  He doesn\'t see how MK and I joke around.  He doesn\'t get that I\'m joking.  No, me makes me out to be this horrible brat, and it hurts.  So afterschool, he was working in the auditorium, and i was very very close to going to talk to him, to try and clear stuff up, but I decided not to.  That was the second person to run away that i really needed to talk to.  So I was a little stressed out and the 50 pounds of junk i was carrying was really starting to hurt.  So I then MK ran off and i found out she was in the library... doing nothing but bothering Brandon Van Westen.  So i left school in a very very bad mood, then Kurt needed a ride home, so i calmed down a bit, i can\'t be mean to my friends... but my sisters.. that\'s different.  Drove Kurt home and brought MK and Laura home. I was exauhsted and I decided to finish the movie i started a week ago.  I didn\'t finish it b/c at 7 i decided i was going to go fins someone and talk to them before i go crazy... well i didn\'t talk to them b/c i never found them, but Aris called.  So i went to Jamie\'s V-ball game at OG, they lost and Jamie was very very very upset.  So we went out for Chinese to try and cheer Jamie up.  No state for Washington :-( sorry Jamie.  Then i came home... fell asleep, woke up, went to bed... couldn\'t get to sleep.. but i suppose i eventually did.  I woke up around 10:30 and wrote this, now i\'m going to go do homework for the rest of my life.  Have a superb day everyone   '),(11238,2033616,'12,November,2003','2003-11-12 00:00:00','    Ich Liebe Dich, und Ich Liebe Dich, und Ich Liebe dich besonders auch!!!!  I\'m in an esspecially good mood... no reason, just happy.. lost in thought, thinking.. maybe life is taking me down a prettier road these days.  I found my old list... and i\'ve realized that someone fits it nearly perfectly... hmm...i\'m being vauge again.. terribly sorry.  ICH LIEBE DICH!!!!!!  Guten Nacht   '),(11239,2033616,'12,November,2003','2003-11-12 00:00:00','    Wow... sometimes you can surprise yourself.  So, today i went back to school.. oh wait we all did.  Wow.. no school in the middle of the week was pretty cool... It felt like a sunday all day and i almost felt like before the day even started, my free day was over-- but it was still nice to get a break for a day, get some sleep and relax.  Well well well well.. hhmmmm... other people have also surprised me lately.  Who would have known?  I realize that i am being very vauge... that might be annoying.. i apologize.  However i can\'t share what is REALLY on my mind lately, now with the whole world anyhow.  I should start driving more often... apparantly \"that\'s a good time to think\" says Chad  I\'m going to Sara Hagen\'s for a cell group tonight, that should be fun b/c i havn\'t seen her in a couple of months, and katie geoff and josh are there so it should be a blast.  Homework is calling.   '),(11240,2033616,'11,November,2003','2003-11-11 00:00:00','    Well Today was Sam\'s birthday.  It was rather eventful.  I feel a bit better today... i think things are getting better.  And then other things (people) are getting more complicated... sheesh.. just keep it simple stupids.   '),(11241,2033616,'11,November,2003','2003-11-11 00:00:00','    Yay, good weekend!  Aliesha came to town on Saturday and we hung out and went shopping.  Then we dropped in on Chad putting up Christmas lights, fun fun i know.  After Aliesha left i did some homework and then fell asleep watching the All State Concert on TV (which was awesome ;-)).  On Sunday, we went out to eat at Red Lobster after church and then went shopping.  Did the cell group thing at my house which went pretty well.  Then Bells and home to finish homework.  I didn\'t have much homework, but i was still up till one.  You see.. when i\'m ESSPECIALLY happy... for things like... someone has solved world hunger, or the whole in the ozone is closing, or one of my best friends and I want to try and figure out what\'s wrong and fix it... well being extreamly happy makes me very energetic... so at 11 o clock last night.. before i finished my homework i did laundry, organized my closet, and cleaned my room, and then finished my homework.  Exciting stuff i know.    Hmmm.. i feel like rambling... Kyle wore color again today!!! only b/c i gave him answers to a worksheet last friday.. but get this.. he wore.. RED... the color RED.  Are you as surprised as I was?    Mateja and I went to the Chicken Dinner tonight and it was excelent.  While waiting in the audi. we had some very good disscussion.  We talked about dating and faith and such.  It was a lot of fun.  Mateja is just so much fun to be around, she\'s so bubbly.. it just makes your day.  Well we finally left around 10 and i got to drive Mateja\'s car two blocks to my house.. a big feat considering it was manual.. (note to self... next time i\'m learning manual... drive somewhere with absolutly NO traffic).  We will have to continue our lesson at another time.. and place.  So then Mateja stopped in for some leftover pizza and went home.  And   '),(11242,2033616,'07,November,2003','2003-11-07 00:00:00','    WEEEEEEEEEEEEEEE Aliesha\'s coming to town tomorrow.  yippie for that  Kayla and Jessica and Renae and I went out to eat after school today.. it was delicious.. then Kayla and I drove around town... Eye doctors, Best Buy, the Mall....... then we both went to our respectable houses and then met up again at Century... i FINALLY saw School of Rock (btw.. WHAT KIND OF MOTHER BRINGS A 9 MONTH OLD INTO A MOVIE THEATER!!!!!) And it was wunderbar.  Then we went to Hy Vee and Nikki (nelson) is so obsessed it\'s not even funny... but that\'s okay b/c it actually was kinda... Oh.. and my left blinker doesn\'t work... which is very very sad...  Now my room is a pit... so i\'m going to fix that   '),(11243,2033616,'06,November,2003','2003-11-06 00:00:00','    I feel like singing today  Yesterday... cows, concubines, and pawns... wow MK and i get hyper at the SAME TIME isn\'t that crazy? Today... okay so it wasn\'t quite as fun.. except for Kyle Davis being the scum of the earth.. that was kinda funny.  It\'s been a REEEEEEEEEEEEEEEEEEEEEEALLY Bad day, but it was sorta resurrected by my hug before 7th period and the wonderful invention called the telephone... yippie for bestest buddies.  Wow still in a weird mood... still in a not so good mood.. i\'m exauhsted... i think i\'ll go to bed, ha! ingenious!!!   '),(11244,2033616,'04,November,2003','2003-11-04 00:00:00','    History notes.. they do nothing but put me in a bad mood.  I have made an important decision today... and that is... life is not going to get me down anymore.  All these loose ends are going to stop causing me so much stress... I\'m going to stop beating myself up, over other people.    ....and history notes still suck   '),(11245,2033616,'03,November,2003','2003-11-03 00:00:00','    Neubie is Awesome.  He is sending me to Samford.   Yay for Neubie and for his Howard Dean.   '),(11246,2033616,'02,November,2003','2003-11-02 00:00:00','    I was once watching a TV show where some evil scientist had some plan to use a laser beam to alter the earth\'s tilt in such a way that Canada would become a tropical paradise...worthy of billions in tourism.  I\'m assuming this would alter temperatures in such areas as the midwest as well...... smart man that mad scientist.   '),(11247,2033616,'02,November,2003','2003-11-02 00:00:00','    This is really odd... i\'m running a fever.. but i feel just fine... except for sudden waves of headace that reminds me \"whoa... i\'m sick\"  oh well.. i hope i don\'t get anyone sick.  I was supposed to babysit for Hansen last night... but no can do when you\'re temp. is upwards of 100 degrees.  Hopefully she found someone to replace me.  Church this morning... ho hum... but there\'s a video scavanger hunt this afternoon which is looking not so promising.... WHO\'S GONNA GO WITH ME?   '),(11248,2033616,'01,November,2003','2003-11-01 00:00:00','    Halloween shouldn\'t be considered a holiday... any day that makes ppl think this would be a good time to paint derogatory comments on really cool peoples\' car is just not right in my book.  Poor Kyle.    Well last night kayla and i went to the mall for dinner and then the play which was \"excelent\" in the words of Miss Mateja.  Then, today Kayla was sick... so no movie party at Kayla\'s.  So then Laura and I went BACK to the play, where we kidnapped Kyle, and attempted to kidnap Sean but he had plans... which was okay b/c i still love you anyway Sean.  So Laura and Kyle and I went to IHOP.. ate breakfast (at 10\'o clock at night) and then dropped Kyle off at school to pick up his car.  Why didn\'t he drive his car you ask?  Well you can\'t be kidnapped and then drive yourself!!!!  And then we found Kyle\'s car had been painted on!!! which was extreamly mean and very much angered laura and I.  Then I dropped laura off and come home... only 30 min after curfew.. no biggie.  I would comment on the rest of my day except it was very much not a good day... except for seeing the play again which is wonderful.  I laughed, I cried, i screamed for Kyle and Sean, i slapped Josh... it was a good capper for a bad day.    You know what I have noticed.... the people that my \"friends\" have been the most judgemental about, the ones they mock the most... are the ones that lately... have proved to REALLY be my friends.  Thank you to all my friends from school, I\'m sorry if I\'ve ever placed you second to anyone else in past years. :-(  You guys are truely one (or a few) in a million.  And on that lovely note I am off to bed...   '),(11249,2033616,'24,December,2003','2003-12-24 00:00:00','    to you: Merry Christmas.   '),(11250,2033616,'23,December,2003','2003-12-23 00:00:00','    Not only was it hard to get up on sunday.. i didn\'t get up.. i overslept.  I didn\'t wake up till 10:07... ie yie yie.  So i was 7 (yes only 7) minutes late to bells.... and not in a good mood.  That afternoon i went shopping with my mother, and then with Chad to help him pick out presents for his sisters.  We went to barnes and noble and he was very good at picking out presents.. he didn\'t even need my help.. once i got him there anyway.  He brought me home then i left again for hy-vee... no one is as good at prolonging grocery shopping as me.  Then Sean called me and we talked till 1 with MK walking in and out of my room.  By one o\'clock MK had finally gone to take her shower and when she got back and i had already gone to bed she walked in appaled.  \"what??!?! why are you sleeping?  no don\'t go to bed.\"  this is the same loopy girl who was saying \"elizabeth, it\'s time for bed\" and hour before.  crazy girl.  Well today.. get this.. i overslept.. and i felt HORRIBLE!!!!!!!!!!!!  I was supposed to go pick up laura at 10:30.. and she called me at 11 asking where i was.  UUUUUUUUGGGGG two days in a row... what is wrong with me?  So i rushed as quick as i could to her house picked her up.. then we went back to my house to get my dress so that we would be able to leave straight from the wrestling tournament for my alterations.  well we got there literaly 5 min after he had wrestled... and get this, we left 10 min before he would wrestle again.  it was very frustrating to say the least.  So then we had my dress altered.  The lady had this fun eastern european accent and her name is elvera.  So we were downtown anyway and decided... it\'s time for lunch.  (it was 3:20)  So we ate at Pizza and Pasta... it was delicious.  And of course, b/c you can\'t go downtown without visiting Zanbros, we stopped in for icecream.  it was a delicious afternoon.  I tried to take Laura home.. but after about 20 min of sitting in traffic, we decided she could just come with me to bells.  However in doing this we were 20 min early.. i am never early... it would be unnatural.. so we went to Sean\'s house and said HI... then got to bells 20 min late... that was more like it.  Oh well.. it was just pizza.    After i took everyone home and took MK shopping yet AGAIN... we topped off the day with the NEWSIES.  oh i love that movie.. who doesn\'t?  Well now for some deep thoughts.   I don\'t think anything said to me has ever been any more hurtfull than being told i wasn\'t worth being friends with.  It\'s the kind of phrase that has just stuck with me.  I mean, it\'s obvious that I have a lot of friends, and a lot of friends that really care about me, but it\'s so sad that you can be inside the most beautiful scenery of your life, and something as stupid and insignificant as a bug bite can make it miserable.  I can spend a day with some of the coolest, sweetest, and most caring friends anyone could ever ask for... but when i remember that there is someone else that thinks \"you\'re not even worth the trouble of being friends with, you\'re not worth it to me\" it\'s enough to make me want to cry myself to sleep and then never wake up.  I know that i\'m worth it, i know i\'m worth it to you, so why should one voice matter?          I hate the mental tape recorder.  It doesn\'t kick in until your lowest point of the day.. then what do i hear?  \"It\'s your fault... You\'re not worth it... you never have been.\"  And it always seems no one is there at those times... when the tape recorder starts up again.  So then i sit there listening to it, and all that was said over and over.. the 2 hour tape runs out and i\'m crushed again as if it was said this morning.  All i want if for someone to turn it off.  Right now is an I-need-mona day.  And then tomorrow i\'ll wake up thinking.. \"no, everything is fine.. i don\'t really need her\"  and that\'s how it will be for several days.. until late one night when the tape starts again.  Well.. the tape\'s been running for about 30 min now.. maybe it will turn off before i go to bed?   '),(11251,2033616,'21,December,2003','2003-12-21 00:00:00','    Well tonight was the party  Laura came over to help set up... according to my mother (who was mad b/c i cleaned my room) laura is a \"good little helper.\"  LOL, then Jessica (B) was the first to arrive.  people started sprinkling in.. then we got a big game of catchphrase going, followed by Mafia.  oh it was fun.  then the married couple showed up.   We played Mafia one more time.. and then most of the people went home around 10:15-ish.  Not too long after they left, Chad got there just in time to see everyone and then say goodbye.  We had our \"after hours party\" and played cards and watched food network... he left at about 1:30... 20 min ago.. and i am exauhsted.  Tomorrow i have plenty to do with shopping and whatnot.  Also... 11 o\'clock bells.  Oh it\'s going to be hard to get up.  Goodnight Sleeptight.  Don\'t let the bedbugs bite.   '),(11252,2033616,'20,December,2003','2003-12-20 00:00:00','    Sooo much to do... sooo little time to do it... better get to work!!   '),(11253,2033616,'17,December,2003','2003-12-17 00:00:00','    hooray for today!  and for grassroots magazines where you e-mail them and the editor herself replies as well as gives you her screen name.   '),(11254,2033616,'15,December,2003','2003-12-15 00:00:00','    I am fully aware that i could just add these quips onto the first post btw... however I am not known for my ability to be reasonable now am I?   '),(11255,2033616,'13,December,2003','2003-12-13 00:00:00','    Oh, and is it just me.. or is that mysterious space in between the links actually growing?   '),(11256,2033616,'13,December,2003','2003-12-13 00:00:00','    Or rather, Homestar reminded me of Brett Peterson   '),(11257,2033616,'13,December,2003','2003-12-13 00:00:00','    I went to the doctor today... and you know what i say about that? BLLLLAAAAAAAAAAAHH grrr  I have been in a really bad mood today.. i think it\'s the lack of sleep coupled with why i didn\'t get sleep and the stupid person that caused it... gggrrrrrr.  Such a bad mood in fact, that i attacked Brett Peterson in AP History today, which was actually really funny.. i also made fun of Kyle the whole period.  It just dawned on me that Brett Peterson reminds me of Homestar.  Anyway, i have lots of homework to be keeping me busy right now, so i must hop to it.  So long.   '),(11258,2033616,'13,December,2003','2003-12-13 00:00:00','    WEEEE!!! Well this morning I had to drive Chad to church for our cookie fundraiser.  He had knee surgery on tuesday... and now is drugged up on pain killers.  So drugged up, that he continually changes accents (yes accents, most commonly British or Ausie).  So that was a lot of fun and a lot of laughs.  I\'m in a really weird mood... so i decided to wear my formal dress for a while... so that\'s what i\'m doing now.  YAY for pretty dresses.  Tonight i\'m going to go see Harvey, hooray for RHS drama.  (Le Mes, second act was awesome!! the first act left much to be desired)  AWWWWWWWWWWW... i\'m in such a sentimental mood  I like being \"complimented\" ;-)   '),(11259,2033616,'10,December,2003','2003-12-10 00:00:00','    Face it, it\'s gone.  No group can survive with conflict like this.  It\'s irrepairable, unfixable, sadly nonrefundable.  It\'s shattered it\'s gone.  Not a haven from the world, but the war\'s very trenches.  God help me I wish it wasn\'t this way.  Why isn\'t it different?  Why can\'t it return?  When will it stop hurting to look and see the shattered place that once was our home away from home.  What\'s the point in going back..  If it will only be worse the next time, and the next, and the next?  The Devil is at work here, and I\'m afraid... he\'s winning.  I need Mona   '),(11260,2033616,'06,December,2003','2003-12-06 00:00:00','    What\'s Shakin\'?  I woke up this morning thinking, i am NOT going to procrastinate today... guess what i have done all evening?  Nothing... oh well.. I can start my essay tomorrow i guess.  Well today we decorated for Christmas.  I love Christmas, it\'s my favorite holiday in the whole wide world except maybe for Groundhog Day.  I dunno, that thrill of excitment i get when i\'m waiting to find out if Punxsutawney Phil (that\'s the groundhog\'s name for those of you who arn\'t as culturaly refined) saw his shadow or not... wow it\'s special.  Wow, \"how many days till Febuary 2nd daddy?\"  So as I was saying, I didn\'t do much today.  Except clean my room.  That was good, i consider it a personal victory over the lazy day.  Last night was the Madrigal Dinner which was very very fun.  Although i would have to say that last year\'s was a tid bit better, this one was just as much fun!  Afterwards MK and i stopped by Chad\'s to wish him a happy birthday for the bazillionth time since i gave him his present.  We didn\'t stay long.  Then Sean met me outside of the house to meet Jamie and Lindsey.  Then Kayla and Renae showed up, and it was a nice little party on a sidewalk in freezing temperatures.  It was great fun.  Then we headed over to my house so mk and i could change and i could put on shoes that didn\'t bind my feet.  (I need to stop staying \"then\")Kayla and Renae split at some point so MK and Sean and I went to Barnes and Noble without them.  We saw Lumer, who is growing a beard for Le Mes and looks like he\'s about 26 now, no joke.  Conversation consisted of Disney Channel and Sean\'s haircut (MK is now his style consultant) a winning combination.  Then we went home and i slept.  I slept and i slept and i woke up and all the sleepyness was gone!!! what a concept.  I should take note of that... more sleep= less sleepiness... why doesn\'t it work like that the rest of the week?  Oh that\'s right.. b/c i\'m a slave to my homework.  I\'m procrastinating right now, I\'m supposed to be writing my essay on American reform movements... i\'d rather eat dirt.  But i leave you now... i think we have some potted plants upstairs....   '),(11261,2033616,'04,December,2003','2003-12-04 00:00:00','    Guess what, I SET THE CURVE!! All of my fellow AP Historians hate me, but it\'s a small price to pay for MY success.  Seriously, i\'m kidding, i feel bad for wrecking the curve. but oh well.. I GOT 100%!!   '),(11262,2033616,'03,December,2003','2003-12-03 00:00:00','    In contrast to yesterday... this is how i feel right now, this minute.  AAHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHH HHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHH HHHHHHHHHHHHHHHHHHHHHHHHHHHHHH!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  I feel better now, thanks for listening.   '),(11263,2033616,'02,December,2003','2003-12-02 00:00:00','    Well, today has been a really good day.  First of all.. i was sick yesterday (this isn\'t one of the good parts) but i still didn\'t feel well last night and i stayed up late doing homework soooooo... i slept till 8:30 this morning.  So I missed first and second periods and got there just as 3rd period reading was starting.  So i was very much awake all day.  Then, in Chem, we had a work day, and i missed a lab... and while that means i have to make it up.. that means i\'m just that much less behind the rest of the class.  So then my history homework: the essay that\'s due on thursday, she is giving me the weekend to work on it! hooray.   Well then afterschool: Mateja has already given me a Christmas present, so now i have cute snowman socks.  And also I now have my Madrigal tickets, and Sean bought me an Ice tea afterschool (if you don\'t understand my excitement, everyday i look at the ice tea and wonder \"is it sweetened?\" so i was standing there and Sean said \"well let\'s get it and see\" and lo and behold it was yay).  Anyway, ontop of that, Kelsey\'s most generous parents gave me another thank you card (i get one about once a month, each with 10 dollars inside) so that was also fun.  And finally... i got my hair cut this afternoon and I love it to death.  It hasn\'t been this short in a really long time, but i really like it.  Well now i have to go be a model student, if you need me, you will find me slaving away over a heavy chemistry book.   '),(11264,2033616,'27,January,2004','2004-01-27 00:00:00','    Naughty Naughty... well today Eliza was sick of course.  Except i still went to the wrestling meet tonight.  But it was well worth it, Chad pinned his guy, Laura thinks she\'s bad luck, and I knew i would never get any homework done there.  But, like i said, it was well worth it.   '),(11265,2033616,'24,January,2004','2004-01-24 00:00:00','    Well... still sick.  And i went to the doctor today, i do not have strep throat and I do not have a virus (which they found out after sticking me with a needle that didn\'t hit a vein and proceeded to dig around inside my arm till they found one).  Oh i wish i could spell.  Have a nice day   '),(11266,2033616,'24,January,2004','2004-01-24 00:00:00','    Oh, so get this... last friday someone...(whom shall remain nameless) and i were gonna go get something to eat... so in walking to this someone\'s car... we find it locked... running.. lights on.  after about an hour of searching for the spare keys at this someone\'s house... we, being the intelligent minded kids we are, figure... AAA anyone?  So i get out my card and make a phone call in my parents car on the way back to the house  AAA: Triple A how can we help? Me:  Hi, we have locked the keys in the car, and it\'s running. we\'re gonna need a locksmith. AAA: Membership Number? Me:  12345ect ect whatever AAA: Name Me:  Elizabeth Leslie AAA: and what is the, make and model of your car? Me: uuhh....   At this point i remove the phone and turn to... someone... \"what\'s the make and model ?\" \" Pontiac Grand Am\"  Me: It\'s a Pontiac Grand Am AAA: and the color? Me: Uhhh....   Once again, i turn to unknown driver... \"color?\" \"silver... gray?\" \"right, i knew that\"  Me: It\'s silver AAA: Is it your car? Me: No, but i\'m riding in it AAA:  And you don\'t know the color Me: it\'s silver AAA: And you had to ask the color of the car? Me: I havn\'t actually seen the color yet AAA: You havn\'t seen the Car? (He sounds very put-out by now... and may i say i did NOT like the tone of his voice) Me: It\'s a brand new car, this is the first time i\'ve seen it... AND IT\'S DARK OUTSIDE AAA: okay thank you miss someone will be with you shortly  Needless to say the AAA guy thinks i\'m trying to rip off their service and he thinks i\'m an idiot... of couse i don\'t think much more highly of him.   '),(11267,2033616,'16,January,2004','2004-01-16 00:00:00','    I\'m sick.. again!!! Grrr... what is wrong with me.... didn\'t i just get over the stomach flu?  Oh well.  So, new semester... nothing\'s new my classes are carbon copy of last semester... except of course for the people in them.  I have no complaints... and would like to say \"hooray for no more kyle cottrill in my history class!!!  sorry kayla :-/  Well last week was formal... not to bad... not all that great either... but at least i got to get all dolled up for once.  I fell asleep watching Finding Nemo at Laura\'s house.  It\'s quite funny actually... EVERYONE fell asleep watching it... except for Kyle... who had been on the road at 6 that morning and wrestled about 4 people.  kudos for Kyle Davis (not Cottrill.. never kudos for Cottrill)  So, next day, tubing at Great Bear... oh what fun and FREEEEEZING... but once again i survived... barely.. i think i was close to loosing a few toes.  Hmm... then we had Bells.  When kayli and I got upstairs... we found MK and Laura sleeping infront of the door to the bell room... awwww.  Then Kayli and I got locked out... in the COLD.  Then we played bells and Pat picked on me.    Mary Katherine: I\'m sorry Pat!! there\'s a reason I...  Pat: no Mary Katherine just get it next time  Mary Katherine: No Pat, there\'s a reason why i can\'t tell the difference between black and white chimes  Pat: Because your sister is Elizabeth  !!!! Well.. i never... also.. when i couldn\'t get the stupid and of 3... she threatened me by saying she wouldn\'t let me stand by Chad.. doesn\'t she get we\'ll talk anyway?  oh well...  After bells Kayli MK Chad and I went to Qudoba.  The second time that weekend for chad and myself, but it was delicious all the same.  I dropped off MK at home and then went to kayli\'s where she and chad and I watched Pirate of the Caribian (sp???).  It was a good time had by all and Karl made us popcorn balls.  I spent the night then went home and went CRAZY at my house.  aahh too long here and anyone could loose it.  So i hung out at Chad\'s.. went to March of Dimes... then back to chad\'s watched a movie... beat chad at pool 7 times.. okay sorta not but he did loose... and was late getting home.  Okay so that was last weekend... this weekend is booked full of me being sick sick sick... so... send me a singing telegram or something... cuz i deserve it :-D  Cheers   '),(11268,2033616,'16,January,2004','2004-01-16 00:00:00','    FINALLY on sunday i\'m gonna get to spend some time with Kayli Jamie and the rest of the church crew.  I havn\'t been there in ages... I could talk about semester tests and how much time i did or didn\'t spend studying... or about formal tomorrow... but i\'m not going to.  I took them, i studied, i\'m going... that\'s it.  Well this weekend should be fun, seriously... i get to see lots of people.  Well today i got a present! :-D yay, Mark burned me a copy of the Relient K Christmas CD which was soo very nice of him.  Three cheers for Relient K... which reminds me... did Jack EVER call me back? no.  Well tonight will be... i don\'t even know what it\'s gonna be... but hopefully i\'ll feel better... or at least have a better grip on things... afterwards.  No one gets this i know... but i\'m just hoping i can get things straightened out inside my head...   My head... i wish my thoughts were more consistent.  One night i\'m dead set positive on one thing... then i wake up and i\'m not sure anymore... i blame it on lack of sleep i guess?  But i don\'t want to ignore whatever it is i have to do, but i don\'t want to jump to conclusions either.  Now that everyone is in a state of compleate confusion and thinking \"what is she talking about\"... i must go... sleep... good afternoon   '),(11269,2033616,'14,January,2004','2004-01-14 00:00:00','    Semester tests start tomorrow...  and b/c no one that reads this will know what i\'m talking about...  OKAY FINE!!! I ADMIT IT!!!   '),(11270,2033616,'09,January,2004','2004-01-09 00:00:00','    Geoff... that was a beautiful post   And the space is still there....   '),(11271,2033616,'08,January,2004','2004-01-08 00:00:00','    AAHHH i\'m so tired of things.  I\'m tired of feeling i have to act a certain way around some people to keep from something being lashed back at me.  It\'s something i\'ve learned in the last few months... around some people you just have to show no emotion, no happiness no sadness or look extreamly engaged in something else b/c NO MATTER WHAT i do... it seems to turn against me.  And i\'m tired of feeling watched by these people... why do i always worry!  It\'s not that i worry about what people think of me in the superficial sense... but i worry that they\'re reading me wrong because i KNOW they\'re watching me.  Why can\'t i just be left alone to be me? to be happy when i am, to be upset when i am, if i want to have fun with my friends i darn well can.  on another note... YOU TOOK IT AWAY FROM ME AND IT WASN\'T YOURS TO TAKE!!!!!!!!! no it doesn\'t make sense, but you can just deal with that on your own.   '),(11272,2033616,'07,January,2004','2004-01-07 00:00:00','    You will notice Renae\'s blog is gone... but the space isn\'t   '),(11273,2033616,'06,January,2004','2004-01-06 00:00:00','    I got blood-work results back from the doctor\'s office.  Oh joy... well today has just been GLORIOUS now hasn\'t it.... (take note... this is sarcasm)   '),(11274,2033616,'06,January,2004','2004-01-06 00:00:00','    I feel the need to just set something straight.  I know I go off the edge sometimes.  I know sometimes i\'m irrational and can be downright meanspirited.  I really hate turning around and realizing the damage i might have done.  So, I\'m sorry for being so dramatic and rude.  I am the kind of person who (in most cases) gets mad and inevitably blows up, no matter how long i try to cool off.. it just builds till i get it off my chest.  But by the time I\'m done, I\'ve calmed down, become rational, and feel horrible about what I\'ve said.  Oh if i only had a nickel everytime i wanted to take something back.  I just wanted to say, I am not holding a grudge and I\'m, again, sorry for what i said.  I just hope that my occaisional irrationality doesn\'t turn my friends away.  I love you all!!!   P.S. I\'M GOING TO SURVIVE AP CHEMISTRY EVERYONE!!!!!! I HAVE A 90%!!!!!!!!   '),(11275,2033616,'05,January,2004','2004-01-05 00:00:00','    The question still stands... what\'s with the space between links????  I\'m taking donations for the Lizzie wants to go to California again fund... anyone willing to donate $730?   '),(11276,2033616,'04,January,2004','2004-01-04 00:00:00','    WHY CAN\'T I BE WENDY GOSH DARN IT!?!?!  i just noticed my last post has a lot of capitalization... all the better to read you with my dear  that was corny   '),(11277,2033616,'03,January,2004','2004-01-03 00:00:00','    Well well well, what have we here, over a week and no sign of Lizzie...  Well last friday i spent packing and relaxing... it was a lot of fun... not really i got so bored i begged my parents to take me to walmart, and that was the hilight of my day.  Saturday the band left for San Diego, California is GORGEOUS.  I loved it there, after we got to the hotel, we walked to this cute little outdoor mall a few blocks away, it was quite fun.  They had an Old Navy, so of course, I have a new pair of jeans now.  The entire evening i walked around trying to form the word \"Berkely California.\"  I could NOT think of it, and consequently drove Jenna Lilyquist and Ricky INSANE.  Sorry guys.  Oh, that reminds me... what movie was Ben Affleck in before he did Bounce with Gwenith Paltrow????  Sunday we hit the beach.  Oooh the Pacific, soo pretty.  We went strolling on the beach, got to check out some cute Cali surfers.  The beach dudes were so nice.  Then Ricky and I went searching for food.. Laura and Merky didn\'t eat, they were too busy buying new outfits to remedy their wet ones... serves them right for going waist deep in the ocean.  We called MK, and made her sad... I\"M SORRY MK I WISH YOU HAD BEEN THERE.  Anyway, i had yummy greek food for lunch, then we went back to the hotel to practice for the concert.  The drive back was basically just me smacking Sean Sutton for playing that stupid video game... goodness gracious.. be social kid!  We rehersed and changed into concert blacks and we were off to Belboa Park... possibly the most beautiful park I have seen.  The buildings and museams in that place were awesome!  To bad we were all to cold to move much, and we were shipped out of there as soon as the concert was over so we didn\'t get to see much.  After our wonderful (not so) concerts, it was off to Old Town for some Mexican food.  The food was great, though it was really cold and we ate outside.  Then we did some shopping and i scared myself w/ pictures from a whale book... oo those things are huge.. which brings me to Monday.  We went Whale watching.. it was awesome and the ocean was perfect and the dolphins and the the the... cold!!!  Then, just when we thought we would never see a whale.. ta daa.. two gray whales.. right there at your service.  They were really really big.. i was nervous.... yes i admit it... I\'M AFRAID OF WHALES!!!!  So, with that said, where did we go next? yep, Sea World... dang Shamu is huge!... no living creature should be bigger than an elephant... it\'s just not right.  But Sea world was fun.. for a while.. then i got really tired of it... it was cold and i was getting really sick of dolphins... understandably.  We went to shamu\'s house of douse and DIDN\'T EVEN GET WET! i don\'t know how that was possible... but oh well.. it was prefectly FINE with me.    Next day we went to... ooo so exciting... DISNSEYLAND... THE happiest place on earth.  and it was supposed to rain but DIDN\'T yay.  Well i can\'t say enough about Disneyland.. except that Disney World is still better... however.. disney is awesome.. aahh i loved it there... On a side note.. i love this year\'s freshmen.. they\'re just awesome.   Well thanks to Seth and Kyle i got to go on Splash Mountain and Indiana Jones both within 30 min w/ ZERO wait.. that was special.  The Marching Band didn\'t do too badly i thought, it was pretty good considering about 75% didn\'t know their music.  But ahh well.  Well Sean Sutton is an impressive kid... he started with $380 and spent every last dollar... on who? his family!.. only $16 of that 380 he spent on himself.. everything else were gifts for family.. i was very impressed.  Well.. the flight home was uneventful... if you call the stomach flu uneventful.. yeah right.. needless to say, my new years sucked.  and i\'m still woozy.. oh well.. i will recover.  Well now.. i do homework.. till i DIE  Sorry to bore you with my super fun trip to California... wish you had been there!!  aawww that\'s precious.. aww i love my freshmen!   '),(11278,2033616,'29,February,2004','2004-02-29 00:00:00','    Everyone always says \"Well i know i havn\'t posted in a while but someone told me I should.\"  Well no one has told me to post so... I havn\'t.  In other news, i don\'t have a car anymore.  I have a brand new appreciation for the invention of the air bag, (and remember people, wear a seat belt!) as well as a new hatred for the use of alcohol... esspecially when it\'s a 40 year old woman who doesn\'t speak english and drives a Ford Expedition.  Friday was not a good day.  But here\'s why i was going to post.  I\'ve never really had to say goodbye to anyone.  I\'ve never lost anyone suddenly.  The only deaths that I have experienced have been my Grandpa, who was sick for a long time, and Great Grandmother, who\'s death was more of a blessing she was in so much pain.  Both lived in Alabama, deaths that far away never really touch you as much.  I saw them... maybe every other year.  Point is, i\'ve never really lost anyone.    I don\'t really remember leaving Switzerland very well, i know i was not happy when we moved here.  I was used to living in an apartment building with a friend behind every door.  I used to scare my parents by just leaving the apartment and going to play with a friend, or once even to a small house on the property where the students and their families would take turns staying for a while.  I had moved 8 times inside the building, from one apartment to the one upstairs, to the one down the hall, to the \"house on the hill\", and back into an apartment.  I guess i didn\'t realize that leaving the country would be different entirely.  What i can remember is that i was very sad to learn Sioux Falls was not like living at the seminary in Zurich.  Since then, i have never had to say goodbye.  But now i do.  I don\'t know how i will take this... b/c i havn\'t taken it very well up to this point.  But, to say goodbye with the prospect of \"goodbye\" meaning \"goodbye, forever, have a nice life and don\'t bother to look me up, you\'ll never find me.\"  I\'m not taking this well.  I feel abandoned and unappreciated.  Isn\'t it the strangest thing about High School?  The people you want most to be closest to, are the people that are still out of reach, yet the people that want so much to be close to  you   ... are the people you wish would leave you alone.  So, goodbye.  Forever. ~Eliza   '),(11279,2033616,'02,February,2004','2004-02-02 00:00:00','    Go look at Saturday the 24... read the story.  Now, take about 40 degrees of heat out of the equation, add the fact that the cell phone was locked IN SIDE the car... which was running of course, stick Lizzie outside a nearly empty LOCKED church... and you have me... yesterday.  Oh i am good... lol, and it won me the title of idiot... but lets all remember... CHAD DID IT FIRST!   '),(11280,2033616,'31,March,2004','2004-03-31 00:00:00','    I\'ve been looking for this conversation... I think it explains a lot   11/2/2003  I\'m BACK!: your mindset your thinking.... you tear yourself apart Lizard: b/c i don\'t know what I am... I don\'t know what i stand for, how i see things, what i want, or what i don\'t want. I want to be independent, i don\'t want to have to rely on ppl... yet i crave someone to take care of me.. i don\'t want things to bother me, but i can\'t let things go until they\'ve already cut me... i have too much faith in small people... i love idleness but i hate feeling worthless    I\'m BACK!:  sounds like you\'ve got a lot figured out about yourself already  Lizard :  maybe... but i don\'t understand it, i just recognize it  Lizard:  i love myself but i hate myself too... same with everyone else.. i can\'t stand someone but i can\'t bare to see them go, i\'m angry but i\'m sad, i\'m happy but i\'m not... i\'m nothing but a mix of emotions... they contradict themselves and make no sense... the only thing that I know... that stays the same through out... is that i want to be a better person a better friend,... THE best that I can  Lizard : but that all those different, contradicting feelings... make it so complicated  Lizard:  i need to go to sleep  I\'m BACK!: agreed   I\'m BACK!: concurrable  I\'m BACK!:  you should sleep more and think less.... because you think so much you put yourself in vicious think cycles and make you go bananazaas.   '),(11281,2033616,'29,March,2004','2004-03-29 00:00:00','    Last night there was a ring around the moon.  It was really cool looking, there was another cloud cutting through the ring that split the edge of the circle.  It was beautiful, it was just clouds and light and water I know, but still it struck me.  The night before, I was lying awake in bed and asked, \"God, is it possible, is it realistic, could it be, how do I know?\"  I completely forgot about my prayer till I saw the moon.  Then I looked up and my memory returned.  I don\'t know if it was some coincidence, some fluke, but I wasn\'t pondering anymore, I was just... Peacefully gazing.  20 minutes later, I was standing at the mirror in my bathroom.  \"Was that it, was that the answer?  How do I know what you mean?\"  And once again, the only thing that I could feel was this sudden feeling of peace, like I had just been hugged and told everything is going to be alright.  So, is it a sign?  \"Is it? Is it?\"   '),(11282,2033616,'28,March,2004','2004-03-28 00:00:00','    You make a good point Kayla... Saturday\'s previous post edited :-)   '),(11283,2033616,'27,March,2004','2004-03-27 00:00:00','    oooo ouch! the ultimate internet-diss, my link has been deleted.  See Geoff? no fixing needed.  AND... life moves on.    So, yesterday, took the SAT.. for not studying and talking on the phone till 11 at night... i feel pretty good about it.  It rained yesterday, oh i like rain, it\'s not snow, and anything that isn\'t snow is okay in my book.    The test was at Washington so i stopped by Jamie\'s to talk a bit.  I really miss talking to Jamie, i wish she and I could hang out more often, but with JO volleyball and homework and whatnot... the schedules just don\'t mesh.  Youth Rally is going to kick butt though.  A whole weekend w/ Chad, Kayli, Jamie, Aris, and Hillary too!  and...... I GET TO SING.  I really love singing in the praise band.  just 4 days till Rally, oohh it couldn\'t come soon enough  Yesterday I just hung out all afternoon, Chad dropped off a CD of the Great Gatsby and i then i fell asleep around 11.    This morning I sang in church, played bells, and then had a pancake feed.  It was a lot, and we made a bunch of $$$... and I mean a BUNCH.  I read the Great Gatsby all afternoon and just relaxed!!! for once in the longest time.  I went to Cell group... which was really boring today.  Chris actually had a lesson plan... see we don\'t do planning... give us a topic and we\'ll discus it to death... just don\'t try and come in w/ a plan... we\'ll kill it i guarantee it.  Went to bells... got yelled at... a WHOLE lot by Pat.  Apparently Chad and Kayli and I talk a lot???  or something like that... i don\'t know i was talking to chad i didn\'t listen.  Then i took Geoff home, no...stopped by Subway first and he bought subs... then i took him home and we talked a while.  I\'ve decided i don\'t want Geoff to \"fix\" anything.  I say, just let it be, I\'ve had too much bad experience and wasted too much time trying to change someone\'s opinion of me.    But I am procrastinating horribly, i need to finish my reading for the Great Gatsby.   ~Dibble   '),(11284,2033616,'27,March,2004','2004-03-27 00:00:00','    It is with the saddest heart that I have to pass on the following:     The Pillsbury Doughboy died Monday of a severe yeast infection and Complications from repeated pokes to the belly. He was 71.  Doughboy was buried in a lightly greased coffin. Dozens of celebrities turned out, including Mrs. Butterworth, The California Raisins, Hungry Jack, Betty Crocker, The Hostess Twinkies, and Captain Crunch. The graveside was piled high with flours as longtime friend Aunt Jemima delivered the eulogy, describing Doughboy as a man who \"never knew how much he was kneaded\".  Doughboy rose quickly in show business but his later life was filled with many turnovers. He was not considered a very smart cookie, wasting much of his dough on half-baked schemes. Despite being a little flaky at times, even as a crusty old man, he was considered a roll model for millions. Toward the end it was thought he\'d rise once again, but he was no tart. Doughboy is survived by his second wife, Play Dough. They have two children, and one in the oven. The funeral was held at 3:50 for about 20 minutes.   '),(11285,2033616,'25,March,2004','2004-03-25 00:00:00','    What makes a great friend?  It\'s not someone you can tell ANYTHING to.  Not someone who you\'ve known forever, or necessarily even always liked.  Having a great friend does not mean you will never argue, or have some rough times, regrets, or grudges.  Friendships are just like any other relationship.  Marriage, is not perfect.  Everyone knows that.  People fight, they go through stress.  Why then... Do we expect our friendships to be perfect?  Why is the mark of a \"best friend\" always, \"someone you can tell anything to.\"  Maybe a better friend is the one you need to tell nothing to.  A friend can see you for all your faults, all your quirks, all your aggravating little flaws, but the definitive mark of a true friend, is the one person that sees all of that, knows all of that, but still holds on.  They don\'t do it for themselves, but for you.  I stopped one day and was shocked, \"why are all these bad things happening, why are friendships so messed up!!?  How can I have so many broken, torn, friendships.  What could be wrong with me.\"  But then a great friend reminded me.  \"The difference between them and me, is that they never took the time to really know you.  I have gotten to know you, to care about you.\"  And that, is the mark of an amazing human.  Lately I have found, the people closest to me, most important to me, aren\'t the people I am with 24/7, that know every one of my last secrets.  They are people that I have known, that I have drifted from, fought with, betrayed, the ones that grate my nerves, that make me scream!!!  But in the end, I love them, because I know who they are and what they are, that they are my friends.  I\'ve made mistakes, I\'ve treated friends like dirt.  Right now I ask, please forgive me for my own regrets.  To my many amazing friends of all the years And to every amazing friend I\'ve ever lost, i hope you have found your \"hero\"too.  (b/c Don Quixote doesn\'t count)   '),(11286,2033616,'25,March,2004','2004-03-25 00:00:00','    I was inspired, but now i\'m tired (haha ha ha.. ha.. hmm....)  keep in mind i am no poet and the clock is going to strike midnight!!! oh no..!!!   Take me in your arms, my love Never let me go You safety wrapped around me Your presence seems to grow  Now you hold me as your sweetheart Because you promised me one thing You would never wish to hurt nor harm The heart of my being  Teach me what it means to love To fully trust you To firstly trust, my God above I want the spirit that\'s in you  I do not know if I can trust Enough to never cry Enough to never feel alone Or be suspicious of a lie  Show me that He\'ll always be The only man for me That you are just a living proof Of love, eternally  For this love between You and I, is not based on us But rather we are followers Faithful in God we trust  I ask you now, please lead me down The path to sanctity And show how YOU have grown to be A living child for Thee  I wish to live to praise the name Of He who sent you me Who brought you here so we could be One with heavenâ€™s three  Teach me how to love  Unconditionally So I will never fail you Nor worse my God mighty.   '),(11287,2033616,'23,March,2004','2004-03-23 00:00:00','    OOHHH I miss elementary school.  the day was so pleasant and the teachers so nice... except 2nd grade but we are not going to go into that.    Laura and I and some more AP chem people judged the Oscar Howe Science Fair tonight.  Some of those kids had some awesome projects... then there was the guy who practically did his son\'s whole project... then told the kid \"don\'t be nervous!.\"  If he did it himself, he probably wouldn\'t be bud.  But the kid was darling.  Anyway it was fun, then afterwards Laura and I walked around the halls and I gibbered about my Oscar Howe days.  Oh boy has that school changed a lot.  Not only the walls that they finally added my 5th grade year, and the new addition the year after... but teachers are gone, rooms flip-flopped around... but that library is still exactly the same... exactly.  Oh i did like school in elementary.  The building is completely different, but i can still see things exactly as they used to be in my head.   Speaking of elementary, i was debating the other night whether i wanted to counsel middle schoolers at camp this summer.  I decided against it.  At my age, elementary kids will still look up to me w/out being all snobbish.  Yes i see middle school as the time when people become completely adverse to any wisdom anyone might have for them.  Now there are exceptions i know, like Nikki, Nikki is awesome!  But for the most part... Elementary schoolers still are willing to be taught, i think we high schoolers are too... but middle school.  Here\'s my complete theory (i\'m babbling about this i know).  Younger kids, 12 and younger, are still very impressionable.  You can talk to them intelligently, but they\'re not yet ready to openly defy you.  As far as camp goes, I feel like I can get through to elementary a lot better than i would with middle school... Especially with the girls.  Then when the kids hit 12... teenagerness sets in.  They become loud and slightly snobbish.  Fights break out about the stupidest things (what were all our fights about laura?  who knows.. they were ridiculous).  These are the types of kids i cannot take care of for a whole week.  Nope, then we hit high school and suddenly you\'re ready to not just think on your own, but think hard.  It\'s no longer a matter of being defiant, but of being true to yourself.  You\'re willing to follow the rules (as long as they\'re not dumb... like ID tags... what are we, homing pigeons?) and maturity finally starts to set in.  Well... in most cases, we all know those kids who are just hopeless anyway.    So back to my point, I\'m going to counsel two camps.. one right after the other.  One is WorshipArtMusic (WAM) and the other is Camp Thingamajig (i know, Sarah is a corny lady).  It will be completely awesome!!  I still haven\'t decided where  I  want to go to camp, but it\'ll all work out.    Oh gee... i\'m tired.  Goodnight  ~Dibith  P.S. I present my power-point tomorrow... when i was working on it, I realized i misspelled  Absalom, Absalom  every single time in my paper... oops.. oh well  P.P.S. speaking of spelling, i just ran the spell check... apparantly I have never been able to spell the word, \"completely\" nice hu?   '),(11288,2033616,'23,March,2004','2004-03-23 00:00:00','    Cell phones are the strangest things...  I think i will be in a superchick mood for a long time now... here\'s another song  Some people bring you gifts Some bring you bricks to weight you down so they can swim a little higher  while you drown  some people mean so well Their way was the best way that they found But any other way you choose is a brick that weights you down  So tell me what do I do with this backpack Full of bricks and sticks and stones and words that stuck to me like ticks (CHORUS) Let it go, let it be Brick by brick we can be free of all the words we saved Till we were our own enemies Let it go, let it be Brick by brick we can believe in the person God intended us to be Let it be  Some people give themselves a brick I know most people do  when we compare We fall short somewhere ItÂ¹s always true  if all we see is where we fall We\'ve bricked a prison wall  instead of trying to learn to fly We\'ve taught ourselves to crawl   So tell me what do I do with this backpack Full of bricks and sticks and stones and words that stuck to me like ticks (CHORUS) Let it go, let it be Brick by brick we can be free of all the words we saved Till we were our own enemies Let it go, let it be Brick by brick we can believe in the person God intended us to be Let it be  we could believe in ourselves More we could try for unique Instead of trying to conform We could defy what they tell us  Don\'t buy the lies they sell us if we\'re brave we can believe in what we are  (CHORUS) Let it go, let it be Brick by brick we can be free of all the words we saved Till we were our own enemies Let it go, let it be Brick by brick we can believe in the person God intended us to be Let it be   '),(11289,2033616,'22,March,2004','2004-03-22 00:00:00','    I realized something yesterday.  I was being bothered by something.  It wasn\'t anything big.  I was feeling like I was being treated unfairly, or not being told the whole story.  But I\'m beginning to realize that they don\'t do it b/c they are trying to hide anything, but rather to not hurt me.  That\'s what a friend will never do, intentionally hurt you.  How can I be upset about that?  I\'m taking a more positive perspective of the situation now.  Why let petty things outweigh all the great things?  I just need to learn to trust the right people.  Now, in case you\'re reading this and you\'re not Laura McCarty, i can assure you, you haven\'t the slightest idea of what i am referring to... Maybe Laura does... If she thinks hard enough ;-)  The current countdown to youth rally is at 11 days... I am sooo excited.  It\'s going to rock.  The song Alyssa and I are singing is going to be great (i just need to memorize the words!!!) and the pictures for the accompanying slide show just need to be put together.  OOOO i love Rally.. if anyone is actually going to Spinsters... I feel very sorry for you, b/c that means you will not be there and you will be missing the most fun one can EVER have in Pierre, South Dakota.  I will stop taunting you now.  I\'ve had a wonderful time procrastinating with you, but I really must do something productive now.  Lizard Breath   '),(11290,2033616,'21,March,2004','2004-03-21 00:00:00','    ahh... i think every major paper in lit and comp should be preceded by a really really really good concert... it sure makes procrastination a lot more fun!  I LOVE THIS RAP!!!  excuse me barlows girls  now there\'s something on my mind see i was jus wonderin\' can i get a valentine saw you walkin by \'n\' somethin caught my eye there\'s something different \'bout these b girls  i won\'t deny alright i can\'t lie  now you got it goin on and all the boys in the band they be sayin you da bomb but can tell from how dress \'n\' the clothes you puttin on what your looking fresh for the reason  that you\'re holdin on for your wedding night alright now let me keep it moving on any girl can be a barlow girl  it\'s more than just a song am i wrong? don\'t think so spare me the drama you remind me of my momma!   '),(11291,2033616,'11,March,2004','2004-03-11 00:00:00','    Oh, who called it?  So, this morning we had praise and worship band practice for 3 whole hours.  oh boy, it was long... but our songs are good.  And Alyssa wasn\'t there, so i was heard... b/c i wasn\'t being drowned out by her hugely strong voice.  Hmm... so Chad had nothing to do today so he said he\'d come over and we could work on the slideshow for Youth Rally.  (Now this is the part where i called it).  So, he said he was going to run to his dad\'s house, after runing by the music store for guitar strings, and then come over in about an hour and a half.  So.... two hours go by, no call or anything... so i called him.  He had taken a nap, no problem there.  He said he\'d get up in about 20 min and leave... so he\'d be here in 30.  well, 3 1/2 hours later... no Chad... and no answer on his cell phone.  I knew he fell asleep.  However, for a guy who\'s already been in 6 car accidents and is 3 1/2 hours late... you start to worry... so i drove over there.  Sure enough, his dad answered the door... Chad was asleep (oh, i called it).  So, to make a long story longer, we looked at his pics from mexico, then he went to subway, then he came over and left about a half hour ago (it\'s 2:10 in the morning)... ohh... church is going to come so early tomorrow.  However, it doesn\'t matter... why? BECAUSE TOMORROW IS THE SUPERCHICK CONCERT!!! eeee... okay... so it\'s not really Superchick, it\'s the super poppy Zoe Girl... but that\'s okay... b/c THAT means... less people in line for Superchick merch. once Zoe Girl comes on.  Oh joy... AND we won\'t be late for youth group... see it all works out... now.. i\'m tired... and i\'m going to bed.  Goodnight... err... Goodmorning?  It\'s tired, i\'m late.   '),(11292,2033616,'11,March,2004','2004-03-11 00:00:00','    Ohhhhhhhhhhhhhhhhhhhhhhh.... do grow up... please!!!!  I\'m sick of being a teenager.  Get me out of here would you?   '),(11293,2033616,'04,March,2004','2004-03-04 00:00:00','    Well thank you for caring Sean, but no, I am just fine don\'t worry.   '),(11294,2033616,'24,April,2004','2004-04-24 00:00:00','    Tonight, i lit 1 1/2 years worth of painful memeories on fire and watched them burn.   '),(11295,2033616,'22,April,2004','2004-04-22 00:00:00','    I\'ve been duped.  How pathetic.  All lies, all of them.   '),(11296,2033616,'22,April,2004','2004-04-22 00:00:00','    I have a voice recital on Saturday May 1st, at 1:30 in the choir room of the Jesky Fine Arts Center (downstairs) at the University of Sioux Falls :-D   '),(11297,2033616,'21,April,2004','2004-04-21 00:00:00','    The current temperature on campus at Samford University is 81 degrees.  I love Samford :-)... a year and a half and I\'ll be an Alabama resident for the first (much delayed) time in my life.  Well that\'s the plan anyway...  Actually when it comes to school right now i am very happy with it... aside from AP tests where i will crash and burn because i have never learned to study and i never really have studied though i desperatly need it b/c i have ZERO confidence in what knowledge i have retained over the course of this year.  (I typed that in one breath)  But otherwise it\'s all hunky dory.  I went to the counselor\'s office today and got my schedual changed for next year... NOOOOOOOOo to AP English yay to taking a ME class for the FIRST TIME SINCE MIDDLE SCHOOL... aahh i loved middle school electives.  What better way to spend an entire class period that in an art room playing with paint?  So, now i am taking World Lit. with Hansen and Photography.  Oh next year will not be bad at all.  OOO boy.. i was REALLY lucky that everything fit my schedual... the counselor was shocked that it worked the first time!!  yay is me.  Back to my AP test thing... i have come to realize that the reason i\'m not SUPER stressed... just semi stressed about them is because honestly, i\'m okay with it if i don\'t pass.  That doesn\'t mean i\'m not going to do my best, and that doesn\'t mean that i\'m not going to TRY and work and study.  But I\'m not going to go balistic over the tests.  I didn\'t take the classes for the college credit.  It would be wonderful, and I am very grateful that my parents have paid so much money for the tests... but I didn\'t take the classes because i want to slack off in college.  I took the classes because 1) I like them.  I\'ve been taking acc classes, like many, since middle school and so why not?  School is so much more endurable when you are in an enviornment with people that are on your same level.  People in AP classes arn\'t asspiring to be the worlds biggest loser.  They all want to be there and are taking the class for their reasons, they have a reason to be there and they\'re not just forced.  2) I would gladly take a B in an AP or acc. class rather than an easy A in a regular class.  I get satisfaction from the fact that (though i admit, i slack off.. alot... esspecially lately) i work hard in those classes.  In chem, i handle really hard concepts, i feel like my brain might explode, but i think i like it like that.  3) I am interested in the subject.  I didn\'t take AP history because i love history, i took it because i wanted the harder route.  BUT, i didn\'t have to go on to AP chem.  Two years of chemistry are not required to graduate.  But, i like chem, i like science, i am a nerd and i LOVE to balance redox reactions.  (laura is dying of laughter now) 4) there is no 4.. but to make a long story longer.  I will be okay with myself if i don\'t pass all the tests in the end.  I\'ll be dissapointed in the begining no doubt... but in the long run, I\'m already okay if i have to take chem in college, if i have to live through another history class.  TAKING ONE MORE CREDIT IN COLLEGE IS NOT THE END OF THE WORLD.  Phew... well i\'m done... now maybe on to something productive... like homework?  Terribly sorry about the length  Did you know Samford has a Chick-fil-a ON campus IN their food court?  It\'s a beautiful thing  Lizzie   '),(11298,2033616,'21,April,2004','2004-04-21 00:00:00','    Okay fine so I jumped on the bandwagon... (i added new comments, for the slow minded)  So, who actually reads this anyway?   '),(11299,2033616,'19,April,2004','2004-04-19 00:00:00','    Okay, the word is out.  Lizzie is... yes Lizzie is in the process of a makeover, in a big way... big as in... about 30 pounds big.  I\'m hoping to have most of it off by July.  (I can\'t believe I\'m posting this for the whole world to see)  That\'s right, 5 weeks ago i decided i\'m going to do something drastic about my weight problem.  I\'ve joined Jenny Craig... so.. if you ever have wondered why i munch on raisins in lit and comp, or am constantly eating carrots and drinking diet snapples and unsweetened Ice tea... there you have it.  In the first three weeks i lost 7.5 lbs... and that\'s mostly from changing my diet.  Once i start working out regularly, things will start to even off or even slow but hopefully the loss will be constant.  I\'m not starving or on vegatables only, i\'m better fed right now than i have probably EVER been... last night i had turkey and mashed potatoes, tonight i had a personal pizza and tomorrow is chicken.  Last week when i weighed in, i hadn\'t lost anything (easter break issues no doubt) so i\'m hoping tomorrow\'s appointment will have better results.  Yep, so thank you for the support from everyone that has known so far, and i hope that those reading this will be just as supportive of my decision to finally take charge of my own life.   '),(11300,2033616,'15,April,2004','2004-04-15 00:00:00','    I have 4 min to write this because i am DETERMINED to be in bed before 11... it will be wonderful getting sleep :-).  I am very very happy to report that I am happy.  I don\'t believe i will go into too many details, but things are better.  Sunday at church, after certain events... or certain non-events... i was left with the feeling that i must be some horrible Christian, a horrible person who doesn\'t deserve such friends--really Godly people can\'t possibly have such horrible rifts in their social life.  I felt like there was something wrong with me, there had to be.  Funny how even though i have so many people i love so very much... this small population of ones that think the lowest of me can bring me down.  Maybe you were right Sean... low self-esteem.  But!  My faith in myself has been restored (thank you for a second chance Sean).  I am very happy.  I have this thing about loosing things... anything.  I loose my favorite pencil, i tear my room apart.. my favorite CD, soon my floor is littered w/ the contents of my room... a friend, and I\'m devastated.   So now, to end.  A quick word of thank you!  Thank you!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!   '),(11301,2033616,'12,April,2004','2004-04-12 00:00:00','    At the moment i have the strangest craving for watching a Zack Files show... it puzzles me too.  And Aris is the coolest!   '),(11302,2033616,'10,April,2004','2004-04-10 00:00:00','    First of all this is not meant to come off as cocky or with the mentality that \"gee.. It\'s so hard being me.\"  I\'m simply analyzing my own thoughts. so don\'t read if you don\'t want to... and i\'m kinda hoping you won\'t  I sometimes feel like I have all the burdens of having a relatively large, diverse array of close friends, but never fully reap the benefits.  Not that I have 50 close friends or that these are my only friends (you\'re all my friends), but the 3-5 very close people in my life bring a little hectic-ness to the social spectrum.     Allow me to analyze.  First of all there is Kayli.  I love Kayli and we have been the best of friends since the 4th grade, however i have since acquired other extremely close knit friendships.  Friendships take time and there are only 24 hours in a day and seven days a week.  Sometimes i feel worn thin between one group of friends and the other.  I love to spend time with Jamie, Aris, and Lindsey, but it\'s hard to make it all the way over to the other side of town.  Additionally, efforts to combine the groups have been more frenzied than fun.  Doing so requires basically forcing people to mesh... you can\'t do that.  To throw another problem into the mix, Chad used to be included in many many activities involving the posse.  This is no longer the case because, though rifts have mended and friendships are being rebuilt, it\'s just not back to that point yet.    So far... I need to give Kayli more time, i have neglected our friendship.  I want to give attention to \"the posse,\" what happened last fall was not fun, and the best way to keep that from happening again is to stay connected.  And, with Chad leaving for college, I don\'t want to let things just fade out before he leaves.  Now... back at school.  Laura and I were close friends in middle school as I\'m sure you are aware, but after that things faded.  Happily, we are closer than ever, and on the same level (hurray for us Laura!!).  During the point in the year when things with the posse were basically sour, or nonexistent for one reason or another, Laura and I grew quite close.  She\'s exactly what I needed while i was dealing with the drifting of 4 very very close friends.  Now, things in the church realm have been nearly returned to the high we were at last year.  Things are great and growing, but once again this poses a problem.  I could not possibly return to the way things were last year with Laura, i will not do that to such a close friend.  But once again, for things in the posse to return, more time is going to be spent together, but not so much that i\'m turning my back on other people again, like Laura, Kayli, and Kayla.  So these are the burdens of my close friendships... with all that, why do i feel so left out?  The time crunch is not all in my mind, b/c most of these people have identified times when i\'ve decided one person over the other, one friend for another.  But with the feeling of being worn thin, how is it that it\'s 8 o\'clock at night on a day with no school, i have no more homework, and no one is around?  People are busy, i know this, Laura is at rehearsal... and the others... well i don\'t know.  The thing is, i can\'t give my whole self to one person or one group, therefore i can only take what I put in.    Now you see, (talking to no one... b/c i\'m sure you\'ve stopped reading) i have figured it out with my long thought process.  Because I can\'t devote undivided attention, I can\'t expect to receive undivided attention.  tada... and my post is finished... i\'m still bored and lonely but hey... i understand why???   '),(11303,2033616,'10,April,2004','2004-04-10 00:00:00','    A little hello.  hi. how are you? well yesterday i woke up at 5 am... i had to be at church for the sunrise service at 6:15!!! ... which of course means i got there at 6:25.  The service was very nice, only 25 minutes and it went very well.  I hung out on and off with MK and the Bell children and Chad and his mom and sis.  Jamie, bless her heart, MEANT to come at 6:30... but didn\'t wake up till 7:45.. when i called her.  So she came to the service late and sat w/ Chad and his family and I.  Cindy, Nikki, and Chad have turned into my adoptive church family... because my parents are never at church when i am!!! oh well, it\'s fun, and i prolly wouldn\'t sit with them anyway... but it\'s the principle of the thing!!!!   Some people need to loosen up... like... all the church elders and deacons who glare at us for clapping during contemporary worship songs... HELLO!!! IT\'S EASTER, WE ARE SUPPOSED TO BE REJOICING!!!  Jenny gets 5 million points for starting the clapping during the flowering of the cross!!  It was quite fun.  For brunch, we went to  C J Callaways Easter brunch.. it was SOOOOOOO yummy oooo and they had shrimp and crab claws.. i was in heaven.  After doing some homework, i went to Carrie\'s to watch The Majestic... movie watching at Carrie\'s has greatly deteriorated into a critical bashing of every aspect of every movie ever watched.  Then the regular round of death ball, which i believe i lost... but hey by then i was super tired anyway.  so i went home.  Today is deemed homework day.  This was a really boring post... but then i\'m in a really dull mood.   '),(11304,2033616,'10,April,2004','2004-04-10 00:00:00','    Thank you Aris for reading my Blog!!! I love you!!! just the way you are!!!!  I\'m at Aris\'s house... i am sitting on the floor b/c jamie and aris are sitting in chairs... we ditched lindsey.. she is sleeping in the other room ... jamie and I are talking to the same person on two different messengers on the same computer.  I will not sleep tonight b/c i was forced to watch the last half of Texas Chainsaw Massacre... this has been a post of really short random sentences (punctuated with ....).  I miss Chicken... chocolate milk would be nice too.  Frizzie Lizzie (do you see that!! it says Frizzie Lizzie... oh simple pleasures)   '),(11305,2033616,'09,April,2004','2004-04-09 00:00:00','    oh this site just makes me happy...  courtesy of Lumer   urlLink Punked out Kitties   '),(11306,2033616,'07,April,2004','2004-04-07 00:00:00','    Dixie Chicks lyrics.  Random ones at that.  I really like the Dixie Chicks.  You may notice a common theme in these lyrics (if you\'re bored enough and have the patience to read through them all)  yes they\'re all about love.  It\'s not that i\'m looking for a boyfriend, or feel i need one, or anything like that... i just am very anxious for MY love story to evolve... a real one... not one i make up for myself.  I guess i want to hurry and grow up.  I\'m tried of facing the world alone, and i want a guy to stand by me and face it with me.  I know relationships bring a whole slew of new problems, i just want to know what it\'s like to really be in love, to really know who and what you are meant for.  I don\'t need to know who, I just wish i knew where i was going, what i\'m supposed to be doing NOW to lead up to what i\'ll be doing LATER.  Oh well... all in due time right?   '),(11307,2033616,'07,April,2004','2004-04-07 00:00:00','    Cowboy take me away Fly this girl as high as you can Into the wild blue Set me free oh I pray Closer to heaven above and Closer to you closer to you   I made a promise to myself Locked it away deep down inside Told my heart we\'d wait it out Swore we\'d never compromise Oh I\'d rather be alone Like I am tonight Than settle for the kind of love That fades before the morning light  Silence stared me in the face And I finally heard its voice It seemed to softly say That in love you have a choice Today i got the answer And there\'s a world of truth behind it Love is out there waiting somewhere You just have to go and find it  I believe in love, I believe in love A love that\'s real, love that\'s strong Love that lives on and on Yes I believe in love Yes I believe in love  And there aint no talkin to this man, hes been tryin to tell me so it took a while to understand, the beauty of just letting go Cause it would take an acrobat, and I already tried all that I\'m gonna let him fly  Wo-oh Swing me way down south Sing me something brave from your mouth And I\'ll bring you Pearls of water on my hips And the love in my lips All the love from my lips  God bless Mommy and matchbox cars God bless Dad and thanks for the stars God hears \"amen\" wherever we are And I love you  Godspeed little man Sweet Dreams little man Oh, my love will fly to you each night on angel\'s wings Godspeed Godspeed Godspeed (Sweet Dreams)  Sometimes I wake up crying at night And sometimes I scream out your name What right does she have to take your heart away When for so long you were mine   '),(11308,2033616,'07,April,2004','2004-04-07 00:00:00','    If I had a time machine, I would frequent all those times with you.  Till I knew your words by heart, till I could plan your steps, till the day that I would forget: that then isn\'t now, and you are gone, and even though I loved you dearly, we both have forgotten   '),(11309,2033616,'06,April,2004','2004-04-06 00:00:00','    Well I feel icky, thus my absence from school today... i know you were all crushed.  That\'s what i get for going a week with hardly 5 hours of sleep a night.  Oh well... if i\'m lucky i will feel just as crappy tomorrow and MISS CONTEST yay!!! i HATE district contests.  I hate them i hate them i hate them.    Okay, now that that\'s done, i must say this.  I MISS MY POSSE!!!!  I need a Carrie-movie night.  I need a sleep-over at Kayli\'s.  I need a trip to Garretson w/ Lindsey and Jamie, a late night in Aris\'s basement, a trip to century theaters with Sam.  I need a day running all over Sioux Falls with Lindsey, Jamie, Aris, Sam, and Chad.  Don\'t get me wrong, people from school, I love you... but... the triplets... the posse... the fab 4... the Jackson 4 1/2... I need them!!!  Why oh why oh why can\'t i have it all.  It\'s impossible to mix so many people together, (Geoff, you can vouch for me on that!)  but why can\'t i have everything?  I\'ve found you just can\'t replace one person with another, each person is so unique, so different.  Remember, i love you all, but man oh man do i ever miss being frizzie Lizzie, activities coordinator for the posse.   '),(11310,2033616,'05,April,2004','2004-04-05 00:00:00','    This year\'s rally was PHENOMENAL!!!! The best of the three years I\'ve been there.  It will be so sad next year when Chad isn\'t on the counsel to spice things up, but the good news is we will have an ARIS and JEN and again, CARRIE!!! Yay, oh yay, it was fun.  I think Saturday was one of the most defining nights of my life, many tears were shed and a lot was confessed and so much was settled.  It may sound like it was a drag, but I assure you, it wasn\'t.  It was awesome!  There were hugs all around and I was reminded that I am a part of the best group of friends there could ever be!! (both those there and those still in sioux falls).  Even though no one that went to Rally reads this... THANK YOU FOR AN AMAZING WEEKEND.  I love you and goodnight.  Lizzie   '),(11311,2033616,'01,April,2004','2004-04-01 00:00:00','    Rally is tomorrow!!! err... today rather.  In 9 hours I will be leaving.  I am SOOOOOO excited, and i\'m getting nervous about my solo!!!!  But i\'m sure things will go just fine.  Today has been wonderful and it can only get better tomorrow!!!  But now i leave you to pack and sleep!!! sleep: it\'s a wondeful thing.  Liza (we are rockstars Laura!) Leslie   '),(11312,2033616,'30,May,2004','2004-05-30 00:00:00','    I would like to revise point number one of my reflections from wednesday.  What better way to do this but with Superchic[k] lyrics  1. Boys are Bad, God is Good (Superchic[k] saying btw)  Princes and Frogs- Underdog Mix You hate men, is what you say And I understand how you feel that way All girls dream of a fairy tale But what you gotâ€™s like a used car sales Man trying to conceal whatâ€™s wrong, Behind a smile and a song And Iâ€™m not saying that boys are not like that But I think you should know That some of us will grow Becauseâ€¦.  All princes start as frogs, all gentleman as dogs Just wait till itâ€™s plain to see What weâ€™re growing up to be Cause some frogs will still be frogs Some dogs will still be dogs But some boys can become men Just donâ€™t kiss us til then  You found him is what you say And we all want you to feel that way The frog youâ€™ve got seems cute enough to kiss And maybe frogs seem like thatâ€™s all there is  But just because you havenâ€™t found your prince yet Doesnâ€™t mean youâ€™re still not a princess And what if your prince comes riding in While youâ€™re kissing a frog Whatâ€™s he gonna think then So look into his eyes, are you a princess or a fly?   '),(11313,2033616,'28,May,2004','2004-05-28 00:00:00','    urlLink    Isn\'t this just the sweetest?  It\'s a picture of a picture I found in my grandparents house over Thanksgiving.  That is actually how MK used to smile in every one of her pictures.  &nbsp; urlLink   '),(11314,2033616,'27,May,2004','2004-05-27 00:00:00','    Oh Summer... I can do whatever I please.  This morning I slept till 10 and woke up and began the make over.  The make over is a face-lift for my bedroom... Or what I like to call, throw the entire contents of your bedroom out of the room and start over.  It shouldn\'t have taken this long, and I\'m still not done, but I have a way of getting side-tracked.  I have $43 in small change and now a really cool collage of Italy on my glass top desk.  BUT!  It is going to look so much fun when I\'m done.  It\'ll be like new, without really changing much.  In any case, I like to clean, reorganize, and figure out ways to make things neater or more efficient ect. Wow... Do I sound like a housewife in training or what?  I went to Laura and Kyle and Jenny\'s DAPA concert tonight.  I got to see JOSH DEWIT.  That kid is just adorable.  I went to the concert with Kayli and afterward we went out to dairy queen with Laura and Kyle and Andrew Pa.  It was a crazy fun time it was.  Then I took Kayli home and it was only 10 o\'clock and so I went to Barnes and Noble and picked up two books.  One is a daily devotional called \"Reflections for Ragamuffins\" and the other is called \"When God writes your love story.\"  They both look very interesting.    I have a lot of plans for this summer, and that\'s not just my calendar plans.  But since you mentioned it here they are.  Next week I am having a get-together as my goodbye to Chad, Mateja, and Neubie.  I love entertaining and even getting ready for the party (back to the house wife thing).  On June 5th I leave for Rapid City to counsel a Middle School camp at Camp Rimrock.  I\'ve only attended camp there once, but I\'m gonna be Nikki\'s counselor so it\'s sure to be fun.  The day after I get back, I take the ACT in the morning then book it over to the church and leave for the Chicago mission trip.  We get back and the next day I start KinderKamp at McKennen park.  aawww those three year olds are just the cutest!  Then the next week.. I\'m FREE for the whole week.  And then on the 4th of July I go BACK to the Hills to be a camper at Rimrock.  Then I come home and then two days later I leave for Okoboji to counsel for the Music Art and Worship camp.  This time, it\'s elementary kids.  Then I come home, grab Jamie, and go back to Okoboji to counsel for MORE elementary kids.  Then, the following week, it\'s one more week of KinderKamp, this time at Kuehn Park.  And then it\'s August and I sleep for a month before school starts again...   But, those aren\'t the plans I was talking about.  Part of it has to do with my devotional.  I really need to set aside some quiet time everyday.  This summer is my summer to redefine myself.   '),(11315,2033616,'25,May,2004','2004-05-25 00:00:00','    Here\'s what I\'m REALLY reflecting on at the end of the school year.  1. Boys are Bad, God is Good.         If your situation SEEMS too good to be true... guess what, it is.  He\'s a shallow heartless child... you just havn\'t taken off your rose color glasses yet  2. It is but hadn\'t ought to be...       Nothing is more painful than seeing every one of your fears and failures play out again and again.  What can you do... take the bullet and keep running.  3. The fool doth think he is wise, but the wise man knows himself to be a fool       Woe-is-me should not be met with sympathy.  Go to someone who\'s wise enough to tell you \"you\'re a screw up, but it\'s okay, now what can you change?\"  4. How lucky I am to have something that makes saying goodbye so hard       When they leave, why do we pour over the sadness of the end.  Get your head out of the ground, just because you won\'t let the end come doesn\'t mean it won\'t.  They\'re starting new lives, doing new things!  This is exciting, not sad.  5.  The most beautiful discovery true friends make is that they can grow separately without growing apart       We don\'t all hold hands in life and trudge forward, we do this on our own.  Sometimes it\'s hard to realize our paths stop crossing so often.  But I\'m still here... i\'m just walking an extra three feet to your left.  6.If you\'re headed in the wrong direction, God allows U-turns.       You\'re falter isn\'t your road to failure.  Surprise some people  and make them proud; dig yourself out... they\'ll help you if you need it.  7.One of the symptoms of an approaching nervous breakdown is the belief that one\'s work is terribly important.       There\'s a little picture, and a big picture.  The big picture is the reward of realizing the little pictures really arn\'t that amazing anyway.  Don\'t stress.  Your fate will never lie in a single task.    8. Worry never robs tomorrow of its sorrow, it only saps today of its joy        Your plan will never be hitchless, your intention never fully comprehended.  Understand that and you\'ll be rolling before the punches even start comming.  9. If only we\'d stop trying to be happy we could have a pretty good time.         Be cheerful, not happy.  Happiness will fade, but cheer is there in good times in bad.    10.Sometimes I wish I were a little kid again, skinned knees are easier to fix than broken hearts.        Life is sad, goodbyes are sad, broken hearts are sad, the world in general is sad.  But things will heal, the sun will shine, and maybe, just maybe you can work something out.   '),(11316,2033616,'25,May,2004','2004-05-25 00:00:00','    Wanted:  Comments!  If you read this blog on a regular basis and you DO care... say hello :-)   '),(11317,2033616,'14,May,2004','2004-05-14 00:00:00','    I have changed my layout b/c once again.... MK\'S WAS EXACTLY THE SAME AS MINE!  will i ever NOT be MK\'s big sister?  Wow.. i feel like Jan on the brady bunch   '),(11318,2033616,'14,May,2004','2004-05-14 00:00:00','    From :  Bob Savot    Sent :  Monday, May 17, 2004 11:33 PM  To :  \"Lizzie Leslie\"    Subject :  Re: BOB!!!!   All the other stuff is okey dokey.   I am tired.   Bob    Oh Bob... we love Bob!!!  BTW... ignore my links for now... they\'re under construction   '),(11319,2033616,'14,May,2004','2004-05-14 00:00:00','    Whoo whoo for FINAL EVER band concerts and new hair (even though aaron hates it EVERYONE ELSE LOVES IT) and the one and only Miss Mateja Casseli!!! I\'m so excited, i\'m going to sioux city tomorrow... oh what a wonderful weekend   '),(11320,2033616,'12,May,2004','2004-05-12 00:00:00','    Well there we are, a new blog layout.  I\'ll go back in and put my links and picture in later.    Question:  Is is right to make a sacrifice for the team, when it means you will not be there for a close friend?  Or does showing your support for the close friend, come before the team?  Or maybe you can only skip out on the team, when many people \"need\" you?  You screw people over either way, but while the team may suffer the loss of a player, and the group, one less set of helping hands; when you drop that single friend, they personally suffer the loss of your apreciation, your love, and your support.  So which loss is worse?   '),(11321,2033616,'11,May,2004','2004-05-11 00:00:00','    This is so fitting, this brand-spanking\'-new blogger format.  I do like, and it\'s fitting because that\'s actually what I wanted to write about.  However, I\'m tired, so I\'ll be creative with the blog later...  I am ready to once and for all be myself!  I am breaking out of this little bubble I\'ve built around myself for so many years.  In 3rd grade when I started gaining weight, I just started to sink down into myself and try my hardest to go unnoticed.  I don\'t do that anymore obviously, but I still have the tendency NOT to grant myself the freedom to be the person that will make me happy.  I\'m making some drastic changes in the way I see myself.   '),(11322,2033616,'01,May,2004','2004-05-01 00:00:00','    I was reading through some old entries and I realized i have made a habit out of leaving out one of the most important people in my life.  It\'s already May and I cry everytime i realize that the begining of summer will bring the end.  I didn\'t leave him out because he doesn\'t mean anything, but because he does, i didn\'t want to get all mushy.  But here\'s the truth, i can\'t imagine life without him.  When i do, i can\'t do anything except sit there with a broken heart.  I think of that time i took him for granted and how i would take that month back in a second, i was so stupid.  He\'s never been anything short of wonderful too me (okay maybe once), and i am so afraid to see that go.  The past few months have put a strain on our friendship, with busy scheduals and college plans.  I\'m trying so hard to realize i have to let go, but it\'s easier to shut down and not feel anything.  The other day, we were talking and i mentioned a movie i wanted to see, he shot back.  \"What could you possibly gain from seeing a movie like that, why would you waste your time when there\'s so much more you could do?\"  Well, I want to feel again, i\'m not happy, i\'m not sad, i\'m just empty.  I just want to feel moved.   '),(11323,2033616,'30,June,2004','2004-06-30 00:00:00','    urlLink    AWWWWW... so can anyone tell i miss being at camp or anything?&nbsp; urlLink   '),(11324,2033616,'28,June,2004','2004-06-28 00:00:00','    I think I\'m backwards on a one-way The world spins opposite my say Every right turn is a dead-end \"No outlet\" Is the new trend  Lies upon lies over which the world toiled My confidence is people is spoiled So now I stand, I thought, alone Broken by this stick and that stone I look to you, now I find My failure is only in my state of mind  I\'m moving out of the driver\'s seat Leaving behind the scorching heat I turn this wheel over to you I can only find the answer in you   I don\'t really write poems anymore.  At least... I don\'t tell anyone when I do.  I don\'t really like knowing I\'m doing what \"everyone else is doing.\"  I never read Left behind or Harry potter.. I\'m sure they\'re great but I have never liked the \"millions of readers agree\" type of thing.  My music is the same.  I know every band has hundreds of fans... But do you really need a copy of a CD you\'ve never even listened to?  Maybe it\'s selfish.  But these things, poetry, music, books... Are so self expressive.  You can learn so much about a person through them.  I hate having that just thrown around.  I hate feeling like my identity is being turned into a free for all.  This isn\'t going where i had planned.  Let me start over.  I don\'t really write poetry anymore.  This is a year old.   '),(11325,2033616,'28,June,2004','2004-06-28 00:00:00','    urlLink    Picture It is fun!  And Tra is a wonderful photogropher (just don\'t look at any of his other photos).  This is a picture Tra took of me at the dance when i was counseling at Rimrock.  Dang... I miss him!!!&nbsp; urlLink   '),(11326,2033616,'26,June,2004','2004-06-26 00:00:00','    Cadet is my new favorite band.  Thanks to Lauren who burned me their cd!   Fantasy act one has just begun but i\'ve seen the end he\'s going to get your heart which you won\'t defend those wedding bells are locked inside your head but can\'t you see you\'re flirting with the dead you\'re falling asleep oh girl he can\'t be in your dreams wake up and believe oh girl don\'t trust your fantasy that stage was built to break for that fling will fail fear not you have not lost love\'s left a trail   '),(11327,2033616,'26,June,2004','2004-06-26 00:00:00','    This is the testimony I\'m presenting tomorrow in church for youth sunday... what do you think?  The surroundings in which we lived for a week were a world apart from the sites any tourist would see of Chicago.  Beyond the skyline and beachfront million dollar condos lies the reality of urbanization.  Here, the horrors of poverty, sickness, and violence are everyday realities.  The pains of racial segregation, hungry children, and families ripped apart by tragedy, are enough to make anyone question Americaâ€™s claim to greatness.  But if you look closely into this picture of Chicago, you will find specks of hope scattered throughout.  They are not readily apparent, and by no means will one or a hundred organizations rid the city of hurt.  But they are trying, and they are working.   	On Wednesday our third full day of work in Chicago, my small group traveled from north side Chicago to the south side community of Englewood.  This community, once a thriving neighborhood, has suffered more tragedy than many of us will ever know.  By the end of year 1991, Englewood was facing its darkest hour; in that year alone it had lost nearly 100 lives to violent murders.  In response, Hal Baskin, a former gang leader became the director for P.E.A.C.E., or People Educated Against Crime in Englewood. After learning this, I realize my memories of Englewood seem to be memories of a ghost town.  Hundreds of lives once present, but forever lost within the limits of this neighborhood.   	My first and most moving experience in Englewood was the sight of the old auditorium.  Painted doorframes were splitting and crumbling beside boarded windows and doors.  Once, Iâ€™m sure it stood as a sign of the areaâ€™s prosperity and culture, now it simply was waiting to cave in.  We pulled up next to the ornate Saint Stevens Lutheran Church.  Across the street was the parish house with 4 foot letters spelling Peace along the width of the building.  This once dilapidated, gang ridden building now held the headquarters for the Will Feed Kids CafÃ© and Peace community center.  The Youth director Ms Anita Dominique, or simply Ms D. lead us into the dark building.   	Though I mentioned the building as being once dilapidated, it is far from renovated.  As Ali and Megan found out, the sink faucets in the bathroom have a bit more character than your normal fixture.  The faucet in the left sink doesnâ€™t work, but if you leave it on, and turn the water on in the right sink, suddenly you realize the left sink is gushing water!  Quirky hardware aside, we settled in to listen to Ms. D.  Her respect in the community is outstanding.  Her involvement in the after school program, Peace community choir, and kidâ€™s cafÃ© has won her the honor of â€œneutronâ€ in the gang circuit.  This simply means graffiti was painted on her garage door, indicating no gang activity was to take place in front of her house.  But as you can imagine, in the ruthless world of urban gangs, this gesture showed more respect for one woman than you probably thought a gang member was even capable of.  Ms. Dâ€™s respect is fully reciprocated, for she makes it a point to strive to understand the gangs and their network of family.   	After lunch we walked two short blocks to the home of Hal Baskinâ€™s mother, whom owns the land that has become the community garden.  The produce of this garden are used in a cooking class for adults, which cooks food to feed the kids in the kidâ€™s cafÃ©, who in turn, work in the garden.  We spent the better part of an hour weeding, and moving a mound of dirt.  Though the garden seemed country enough, I was reminded of the neighborhood that surrounded it as I came upon more glass bits and even a broken pair of eyeglasses.  Walking back to the parish, we passed the shoddy houses of Englewood.  Once, they stood as symbols of Midwestern culture and affluence, now their run down yards and peeling paint were just another reminder of Englewoodâ€™s tragedies. 	Murder numbers in Englewood still rank in the top five highest of Chicago.  Murders still number into the 40â€™s and beyond each year.  Why then would I see the Will Feed and Peace program as a provider of hope to the community?  Because I donâ€™t see it as 40 murders a year, I see it as a community that cut its annual murder rate in half.  I see it as a community that is pulling together to make the neighborhood safer.  Because, in Chicago, the hungry can find food, the homeless shelter, and in Englewood, children can find safety from the gang violence on the street.  You can write a hundred letters to congress, run a thousand food drives, and spend a million dollars on improvements, but the problems will still run rampant.  We cannot stop the atrocities of our world.  We will never solve world hunger or find a home for everyone.  Iâ€™ve learned that serving for Christ is not about solving the problem; itâ€™s about our willingness to help, our desire to teach, and our passion for growing closer to Christ.   '),(11328,2033616,'24,June,2004','2004-06-24 00:00:00','    Strange Day.  But again.. i don\'t want to post.. i want to talk!!  but these things always happen at night... and of course there\'s hardly anyone there to talk to at night.   '),(11329,2033616,'21,June,2004','2004-06-21 00:00:00','    I\'ve been doing a lot of thinking over these past two weeks... And I decided I am not a blog person.  I don\'t want to tell people about my day, I want to spend it with them or tell them everything over lunch.  I don\'t want to babble about what\'s on my mind, I want someone to show interest first.  So, once again to my two faithful readers, I will write again when I take a really fun picture, or when I have become bored of all my projects, or when I think of something I just have to get out.  But until then, ta ta.  I had a wonderful time at camp and in Chicago.  I have lovely stories.  If you would like to hear them,  Let\'s do lunch?   '),(11330,2033616,'05,June,2004','2004-06-05 00:00:00','    urlLink    This is a  photo of me and Laura from a few nights ago... we fit 6 people on that hammock!&nbsp; urlLink   '),(11331,2033616,'03,June,2004','2004-06-03 00:00:00','    A quick goodbye to all my (2) faithful readers!! I\'m off to Rimrock... but not before taking a stupid standardized test that will determine the rest of my life.  Fair-the-well.   If you really miss me... i left you a picture.  Aren\'t I just the kindest?   '),(11332,2033616,'03,June,2004','2004-06-03 00:00:00','    I have a lump in my throat... But I\'m not sure why.  I was suddenly overwhelmed with this feeling of helplessness.  However on second thought, maybe a feeling of helplessness is needed more often.  I\'m such a control person; I worry constantly, obsess over details, and I can\'t handle it when something goes astray.  I need to remember I\'m not in control, life is to big for us to control it on our own.  Let go Lizzie, let go.   Let go.  That\'s what the lump is from.  I was sitting there in the rocking chair next to Mary Katherine as she typed away on the computer and suddenly out of no where...   It was October of 2002.  Me and Jamie, Lindsey, Kayli, Sam, her brother, and Chad are in the Corn Maze in Hartford and we\'re hopelessly lost in the darkness of the corn stalks.  \"Do you KNOW how to read the map Lizzie?\"  Ha, that tone in his voice has never changed, just become more sincere and less cynical.  Then we\'re in the parking lot.  We have two cars, one which arrived with 5 teenage girls and a middle school brother.. And then Chad, and his car.  For the drive back to Sioux falls we split up... Anyone could tell Jamie desperately wanted to ride with Chad, but never alone.  I found myself in the back seat of the Daytona.  We talked about his multiple detentions in middle school, Spanish, and of course his driving habits.  Ha, I remember when I was so afraid of him in elementary school.  Even up to freshmen year he intimidated me.  It\'s so strange to go back and look at the evolution of a friendship.  I was afraid of Chad.  Kayla and I pretty much met in the sand box.  Laura McCarty and I became friends in the 6th grade after being thrown together in reading class, in Jr. Philharmonic and at the lunch table.  Sam hated me when we first met, then one night at lifelight we just clicked...   Do you ever stop and think about it?  All you\'ve been through, what you were like before them.  Not just what they mean now, but how they got there?  I have so much to say but no one to say it to.  I\'m bursting at the seams to talk about what I believe, what I think, what I feel, why I\'m sad, happy, optimistic, feeling helpless.  I\'ve been trying to figure out the perfect person to say these things to.  But I\'m just so scared I\'ll pick the wrong one yet again.   '),(11333,2033616,'01,June,2004','2004-06-01 00:00:00','    Today was, quite a day.  Last night was a blast, yay for friends... and boo to saying goodbye to them.  Oh me, oh my... I\'m okay.  Geoff has a snippet about me on his blog, he said.   Believe-Elizabeth Leslie: I have known Elizabeth for quite some time. She is a constant reminder for me to always remember WWJD and stay righteous.   I like that part... I don\'t really understand this part.   She needs to work harder on that.     Geoffrey, do explain!  Last night I had a bunch of people over.  Chad, Mateja, Jason, Neubie, Whitni, Adam, Mike, Kara, Kayli, Hilary, Laura, Jamie, Lindsey, Aris... All came over.  It was my goodbye to my seniors.  Speaking of goodbye... I\'ve never done it before, and it\'s hard.  Chad and I recorded his song today.  I was there for about an hour and a half.  I\'ve been doing fine for the last week, but when I left it hit me that that  was  my goodbye. That was hard.  I wasn\'t planning on working this summer.. But now I have a job.  For a whole two weeks!!!  The YWCA is going to pay me to counsel a day camp for 3-5 year olds.  I went in today to file paperwork.  Those kids are just adorable.  I love kids.  Laura and I decided that when I become everyone\'s favorite youth pastor I will start a children\'s choir.  \"New Goal, New Goal!!!\" oh fun stuff.  I have a date for next year\'s prom.  Date? Lizzie? Never.  Yes, Jason sure is a sweetheart.    I\'ve lost a little over 17 lbs since March.  It\'s so strange that I thought this would be the most tedious, gruelling process... But I don\'t even think about it anymore.    I\'ve already decided that next year I will become a floater at school.  You won\'t even remember me b/c I will just blend into the background.  I think it\'ll be easier that way.  I\'ll just drift.. It\'ll be fun.  No more drama, zero zero zero drama.  I wish it was summer always and forever.   '),(11334,2033616,'01,June,2004','2004-06-01 00:00:00','    This is my blog.  I should be able to say what I feel right?  nope, i can\'t, because then all i will get is a backlash of people thinking that all am is jealous and selfish.  Also, by posting it, i might offend someone.  So i\'ll just tiptoe around the real thing for now and let you pick up on it.    It\'s a strange thing to be jealous of someone you love.  There\'s no way you could hurt them or ever tell them how much you want to knock them off... Instead, you love them, but that small hint of jealousy is always present, breeding resentment.    I can\'t say I hate everything.  Quite honestly the very thing that makes me so jealous is also the very thing that makes me so proud.  But more and more, i can\'t smile and pass of the actions as \"personality.\"  It has spread us apart, put a wedge in our relationship.    I don\'t like it, i hate it.  I want to be able to smile when i see her friends laughing at her antics.  But i can\'t, because those friends of hers, were ones i used to call my own... only to realize now they\'d rather be with her.  I still have my closest friends.  They\'re my friends b/c of ME, not b/c they just met me before her.  I\'ve never considered myself the kind of person everyone likes.  But can you imagine watching the people that have never liked you, ones you have never gotten along with, flocking to your baby sister like she was some sort of celebrity???  At the begining of the year, all i got was \"i didn\'t know you had a sister?\"  by next year, she\'ll be getting the same thing.  B/c nothing is my own anymore.  If i have one talent, she has it and dozens more.  If I get a compliment, someone gives her the same one plus that.    Since i\'ve already given myself away...  DON\'T you EVER EVER start a conversation with.  \"Your sister is the coolest/funniest/ weirdest girl I\'ve ever met\"  She\'s amazing, she\'s hilarious, she\'s uniquie, she\'s beautiful.  This, i know.  I\'ve lived with her for 15 years.  Can someone just forget about her for one second to have a conversation with Lizzie and not \"MKat\'s sister?\"   '),(11335,2033616,'30,July,2004','2004-07-30 00:00:00','    Attention... this post has been typed by Lizzie.. and I\'m terribly sorry but i didn\'t feel like monitoring the Lizzie speech.. so this is babbly and goes on and REPEATS alot.  I had a very scary thought today as i sat on a blanket in Kuehn park with a preschooler on my lap and 8 more scattered around me. The story was about an Octopus i believe, not very enthralling for me but the kids sure liked it. Anyway, I remembers what I had said to Jamie last night about Samford. We were talking about private college over public and I was talking about the biggest reason that I\'m choosing private church affiliated schools over any public schools. It\'s not that I\'m afraid of the public school surrounding... And I don\'t think that I\'m so weak in my faith that i would end up turning in my newboys and relient K for Emenim or Marilyn Manson... But there\'s not a doubt in my mind that going to Samford is going to be the best thing I can do for myself spiritually. I\'m so excited at the thought of going to a school that will surround me with people that are thirsting for the same thing I am. Samford isn\'t like other religious colleges. Their affiliation with the Southern Baptists isn\'t just a way for them to receive funding or to attract more and more Southern Baptists... It\'s a sign that every program every professor every class makes it very clear that God is present there. That\'s why I want to go to Samford.  By the way... This isn\'t the scary part. What I said next was, jokingly, \"Besides... If I\'m going to meet my dream man in college, i figure i have better odds there.\" The truth is I have fully accepted the fact that I\'ll prolly never have a (cough.. Another) relationship until college. Not to say I don\'t know perfectly darling (mk word!) guys now... I just can\'t imagine ( b/c it\'s not like I have much hope for it anymore) anything getting off the ground until I am in college. What scares me is that I\'m counting on it. I\'m building up my future around the fact that between the years of 2005 and 2009 I will meet the man I am to marry and marry very soon after that. Then, he can start his career and I can start my meager earnings as a youth pastor. What struck me today is there is that very real possibility that he won\'t be at Samford. I might not be telling my children one day \"your father and I met at the college, at the same college that both of MY parents met at... Sigh.\" It might not happen. And then what? What if I can\'t make it as only a youth pastor. What if my fallback option of a family counselor doesn\'t pick up. What if I don\'t find someone till I\'m 30 and until them am living miserably because I built my whole career up on the fact that I wouldn\'t have to be making all the money? It seems WAAAY to early to be thinking about this (I wonder if anyone is still reading at this point... b/c i bet you find this very boring) but maybe I\'ve started thinking about it just in time. I\'ve always scoffed at high school couples that change so much of their daily lives to betogether. They\'ll switch churches, or join each other\'s activities, plan out their summers together. Then they break up and they have to drop so much. I\'m so very against doing this in high school yet that\'s what I\'ve done in my mind. I\'ve built up a whole plan around a man soon to be identified and... What if it doesn\'t happen.  I think I need to start buiding my plans around me.. with room for additions. Not around the additions with room for plans.   '),(11336,2033616,'25,July,2004','2004-07-25 00:00:00','    Well, I will present my problem.&nbsp; It\'s 6:30 and I am dying to go out... But I am going to the church tonight at 3 am to unlock the doors for all my buddies getting back from the Presby. Triennium.&nbsp; So I\'ll prolly go to bed a 10... Not much time to do anything.&nbsp; So anyway I\'m bored so I have decided to fill out one of those survey deals but I think I will post it here b/c no one really reads my Blog anyway so therefore I can tell all and still remain a mystery...    &nbsp;    ....Kidding of course     1. Full name: Miss Elizabeth Alice Leslie   2. Are you named after anyone?&nbsp; Yes, infact.&nbsp; My mother\'s given first name is actually Alice, as was her mother\'s and her\'s before her and her\'s before her.&nbsp; So basically I\'m at least the 5th generation of \"Alice\" and you can bet that my first daughter will also be Alice.&nbsp; And of course my Aunt\'s name is Elisabeth.   3. Do you wish on stars? All the time.&nbsp; It\'s more of a nostalgia thing than anything else.   4. Which finger is your favorite? I\'m sorry... I love them all the same.   5. When did you last cry? Last night actually.&nbsp; Ask later.   6. Do you like your handwriting? I used to hate it but after saying \"I like your handwriting\" to so many people and hearing them say \"I hate it\" I\'ve decided that no one likes their handwriting and it\'s not like I can really do much about mine so why live hating it?   7. What is your favorite lunchmeat?&nbsp; My Bologna has a first name, it\'s o-s-c-a-r, by Bologna has a last name, it\'s m-e-y-e-r.&nbsp; I love to eat it everyday and if you ask me why I saaayyy!&nbsp; \'Cause Oscar Meyer has a way with b-o-l-o-g-n-a!!! (9 hour bus ride from camp... Please don\'t judge me!!!!)   8. Any bad habits? Oh goodness.. I whine ALOT.&nbsp; I\'m sorry guys.&nbsp; I\'m gonna start working on that.   9. What is your most embarrassing CD on the shelf?&nbsp; None, I\'m proud of them all.&nbsp; Besides... The Spice Girls are hidden in the closet... And they\'re on cassette anyway.   10. If you were another person, would you be friends with YOU?&nbsp; Well, like I said, I whine a lot.&nbsp; But after that I don\'t think I\'m too terrible to be around... Am I?&nbsp; I think I would, I\'m quirky like that.   11. Are you a daredevil? No... I just look on as my loved ones jump of cliffs and play with fire.   12. Do fish have feelings?&nbsp;No, if you poke them.. Yes they FEEL that... But if you call them a dirty fish in fish language they really don\'t care.   17. Are you trendy? Only if what I happen to want to wear at the moment is.   18. How do you release anger? I wait till no one is around and then I yell and scream.&nbsp;   19. Where are your second homes?&nbsp; 1. At the Presby Camp on Okoboji... I could stay there forever.&nbsp; 2.&nbsp; At church.&nbsp; 3. Home is where the heart is... Wherever my friends are I\'m at home.&nbsp;   20. Do you trust others easily?&nbsp;I trust the wrong people.&nbsp; Strangely enough sometimes i start to stop trusting the closer i become to my friends.   21. What were your favorite toys as a child?&nbsp; MY GREEN ERASER COMPUTER.&nbsp; And my cardboard playhouse... And my stuffed animals... And the computer b/c i became addicted to Duke Nukem at a very early age   22. What class in school do you think is totally useless? Physical Science was by far the most worthless class I could have taken.&nbsp;   23. Do you have a journal?&nbsp;yes, this and then I have my Prayer/bible study notebook.   24. Do you use sarcasm a lot? Only with Mary Katherine... She gets mad and laughs at the same time, it\'s fun.   25.&nbsp; There was no 25....   26. What do you look for in a guy/girl?&nbsp;A strong commitment for Christ.&nbsp; Someone understanding and willing to listen and to talk.&nbsp;&nbsp; And they have to be just a little bit quirky... Otherwise what fun are they to be around??&nbsp; Oh, and they have to respect my decision not to kiss   27. What are your nicknames? Lizzie, Liz&nbsp;(duh), Eliza, Lizard, Lizbeth, LizzieDibDib (Dani Tornow only)&nbsp; Dibble, and most recently Repete and Raman Noodles   28. Would you ever sky dive? No no no no.... not that I\'m especially scared of it.. But I really don\'t need a thrill like that at the moment.   30. Do you think that you are strong?&nbsp; Not physically.&nbsp; Emotionally i bounce back eventually but things tend to hit hard and then stick around for a while.&nbsp; Spiritually.. I\'d like to think so.   31. What\'s your favorite Ice Cream Flavor? Chocolate and Coffee   32. What is your favorite color? If it\'s on the rainbow it goes.   33. What is your least favorite food? Mushrooms.&nbsp; But I think I\'ve concered most others... i even had grilled cheese!!!   34. How many wisdom teeth do you have?&nbsp; 4.. Growing in sideways.&nbsp; Needless to say I\'m seeing the oral surgeon in 3 weeks   35. Are you in love with anyone?&nbsp;Nope   36. How many people have a crush on you right now? I\'m pretty sure there was a camper last week that did.&nbsp; He would jump out of nowhere and say \"HI LIZZIE\" and then give me a big hug.&nbsp; Sweet kid.. But a little weird too.&nbsp;&nbsp; Otherwise, I\'m going to be an old maid i just know it   37. Who do you miss most right now?&nbsp;Must I list them all.&nbsp; About 10 people scattered over south Dakota.&nbsp; 4 staffing at Okoboji.&nbsp; Another 5 or 6 in Iowa.&nbsp; And Jamie in Minnesota.&nbsp; Oh and not to mention all the people in SF I still haven\'t seen since i got home.   38. Do you want everyone you send this to send it back? Nope... Don\'t trouble yourselves :-) prove that your life is more exciting than this!!!   39. What color pants are you wearing? Navy Blue Shorts.   40. What are you listening to right now? The Elms... I copied the CD from Drew   41. What are the last 4 digits of your phone number? 1194 for the cell phone 8608 @home   42. What was the last thing you ate? A salad   43. If you were a crayon, what color would you be? Magenta... Just because it\'s fun to say.   44. How is the weather right now?&nbsp;&nbsp;I haven\'t been out much today.   45. Who is the last person you talked to on the phone?&nbsp; Personally.. Last night when i.. sorta talked to Chad.&nbsp; Otherwise it was one of MK\'s friends who wanted to know where she was.   46. First thing you notice about the opposite sex?&nbsp; Are they making me laugh and are they letting ME have a laugh?   47. Do you like the person who sent this? Yes, I got it from Kayla AND Sean and Kayla is my life long friend and Sean is... Well he\'s great but we won\'t go into titles...   48. How are you today? Very relaxed... sorta craving some human contact... But also enjoying FINALLY being by myself.   49. Favorite drink? Sweet Tea, Dr. Pepper... Sorry.. not a big water person.. Only when i have to.   50. Favorite alcoholic? Wouldn\'t know.. prolly will never have one   51. Favorite sports? No matter what I do I just can\'t get into sports because I just have absolutely no will to be competitive what-so-ever.   52. Hair color? Dark Brown (soon to be red highlights again)   53. Eye color? Hazel.. Otherwise known as Blue to&nbsp;Olive to yellow brown to.. Brown.   54. Do you wear contacts? yep... I can\'t see my hand infront of my face with out them   56. Favorite month(s)? Anytime that\'s NOT Febuary March or really even January   57. Favorite food? Chocolate, Chicken, and lots of Veggies and Fruit   58. Last movies you watched? Last night we rented Office Space (Yay for that hu drew?)&nbsp; And last week I saw Raising Helen.. Oh and I was forced to see the Sandlot at camp.   59. Are you too shy to ask someone out? Not exactly too shy..&nbsp; I think the guy needs to do it to assert leadership.   61. Scary movies or happy endings?&nbsp; Depends on my mood.&nbsp; Happy endings tend to be just to oh-what-a-coincidence-everything-went-off-without-a-hitch-perfect.   62. Summer or winter? Summer... Winter does BAD things to me   63. Hugs or kisses? Hugs... B/c unless I\'m married to you.. I\'m not going to be kissing you anytime soon   64. Relationships or one-night stands?&nbsp;It\'s kinda hard to have a one-night stand when you\'ve decided not to kiss... So what do you think?   65. Do you want your friends to write back? No... I want them to have a life... Unlike me.. Today anyway   66. Who is most likely to respond? Someone feeling very contradictory b/c i just told them not to.   67. Who is least likely to respond? Considering you can\'t really respond....   72. What did you watch on TV last night? I didn\'t watch TV... We watched Office Space... But on Friday I watched Monk.   73. Favorite smells?&nbsp; When my Mommy makes Chicken Curry, My Grandparents house, and anything that smells like Christmas   74. What is the first thing you think of when you wake up in the morning? It\'s still summer right?   75. Wouldn\'t it be nice if whenever we messed up our life we could simply press \'Ctrl Alt Delete\' and start all over? not really.&nbsp; The times when I\'ve messed up I\'ve paid greatly but I think it has helped me to change for the better.&nbsp; And besides... You\'d be pressing \"Ctrl Alt Delete\" ALOT   76. Who is your best friend? New realization:&nbsp; Jesus Christ!!!&nbsp; But I\'m sure you were asking about people here... Jamie, Laura... And&nbsp;of course Chad.&nbsp; But there are more people that I absolutely love to death.   '),(11337,2033616,'17,July,2004','2004-07-17 00:00:00','    urlLink    I have ALWAYS wanted to try this... so Kyle and I finally did.&nbsp; Oh fun times. urlLink   '),(11338,2033616,'17,July,2004','2004-07-17 00:00:00','    A Picture is worth a thousand words... and I don\'t feel like typing 10 thousand words so how bout i just show you a bunch of pictures.&nbsp; Okoboji was EXCELENT as usual.&nbsp; I absolutly love that place.&nbsp; I really brings me back to who I am and what I believe.&nbsp; It was the place of so many realizations when I was a camper.&nbsp; Now, It will hopefully be supplying me with a paycheck in a few years.&nbsp; If anyone wants to find me the summer after my freshman year of college.&nbsp; That\'s where I\'m gonna be.   '),(11339,2033616,'17,July,2004','2004-07-17 00:00:00','    urlLink    Me and Carly.  Or as she was less formally known.  Little Orphan Anne.&nbsp; urlLink   '),(11340,2033616,'17,July,2004','2004-07-17 00:00:00','    urlLink    The best cabin (and CIT) a counselor could ask for.  &nbsp; urlLink   '),(11341,2033616,'17,July,2004','2004-07-17 00:00:00','    urlLink    And the next week.  Nate and Jamie with \"the megaphone\"  That thing was a God-send with 97 campers.&nbsp; urlLink   '),(11342,2033616,'17,July,2004','2004-07-17 00:00:00','    urlLink    Me and Jamie during Morning watch.&nbsp; urlLink   '),(11343,2033616,'17,July,2004','2004-07-17 00:00:00','    urlLink    This is the entire WAM camp... 11 campers!!&nbsp; urlLink   '),(11344,2033616,'17,July,2004','2004-07-17 00:00:00','    urlLink    Me and Aliesha on the last night of camp.  That week\'s camp anyway.&nbsp; urlLink   '),(11345,2033616,'17,July,2004','2004-07-17 00:00:00','    urlLink    POOOOOR Nate... That pie was hurled at him by my cabin.  Pooooor Nate.&nbsp; urlLink   '),(11346,2033616,'17,July,2004','2004-07-17 00:00:00','    urlLink    One word... Blackmail!&nbsp; urlLink   '),(11347,2033616,'16,July,2004','2004-07-16 00:00:00','    urlLink    Make that two blackmails&nbsp; urlLink   '),(11348,2033616,'16,July,2004','2004-07-16 00:00:00','    urlLink    Youth Rally 2003...when things were great and everyone loved everyone else.&nbsp; urlLink   '),(11349,2033616,'16,July,2004','2004-07-16 00:00:00','    There\'s nothing to say except that I don\'t know whether to cry or SCREAM I\'m so happy.&nbsp; I just got home from a week it camp (btw.. I never got pied.. Directly.. Just smothered by the counselors that did).&nbsp; First of all... I am floating on a cloud because of this wonderful news.&nbsp; It seems my life isn\'t as aimless as I thought... I improved my ACT score by 2 points.. Giving me a 30... My SAT by almost 100.. So now I have a 1380... And I just opened my AP scores.&nbsp; I\'m a qualifying College Board AP Scholar...&nbsp;I got a 4 on all three tests.&nbsp;  &nbsp;  I SCREAMED... And then my next thought was... I have to tell someone.&nbsp; This is not a bragging thing of course but an... \"If I don\'t share my excitement with someone else it just might kill me\" type of thing.&nbsp; And then I realized there is only one person to tell... And then I remembered that they\'re not around anymore.&nbsp; That they don\'t seem interested anymore.&nbsp; At least I don\'t think... B/c I don\'t know what they think.. B/c they won\'t tell me... And they\'re not here.&nbsp; And when something horrible/wonderful/ exciting/ breathtaking/earth shattering happens to you, you want to tell your closest friend in the whole world right?&nbsp; But I can\'t b/c he\'s not my friend anymore.&nbsp; He\'s Kennen\'s boyfriend... Not Lizzie\'s best friend.&nbsp; So then instead of jumping up and down I felt like crying.&nbsp; Because the one person in the whole world that held true to \"I\'ll never leave you, I will always be here for you,\"&nbsp;isn\'t here when all I want is someone to smile with.   Tears aside... It has been an interesting week.&nbsp; I\'m always nervous about how people will react to my decisions on dating.&nbsp; People close to me, that know my history, are easier to explain to.&nbsp; But the people that say \"ha, that\'ll never work.&nbsp; whatever\" are the ones that almost make me want to cave.. Just so I don\'t set unreachable goals.&nbsp; Then I remember WHY I\'ve made the decision I have and for whom I\'ve made them.&nbsp; Then they stay standing.&nbsp;  &nbsp;  So... I think I\'ll post some pictures now and then go to bed and start my \"I\'ve been at camp for two weeks straight I need a week of recuperation before my head falls off\" day.&nbsp; That\'s all Folks.&nbsp; Goodnight.&nbsp;&nbsp; God Bless, I love you (yes... That means you)   '),(11350,2033616,'14,July,2004','2004-07-14 00:00:00','    Your success as a counselor can be directly linked to the amount of crap the kids give you.  I don\'t believe i will make it through the week without being pied in the face.  Oh dear... the prize for a clean cabin is a pie in the face... how does that work?   '),(11351,2033616,'12,July,2004','2004-07-12 00:00:00','    Well I\'m at Okoboji counseling and i had a minute during free time so i thought i\'d check in. First of all... TRA YOU FOUND ME!! lol, you are just too much, i luv ya Tra.  &nbsp; Okay, that is all,&nbsp;I better go b/c actual paid staff wants to use the computer (btw Lumer... if you read this.. i\'m using an iMac!!) So long. I\'m off to save Aliesha from the ruthless kids in the rec. hall. Lots of love... thanks for the comments tra... and here i thought Kayla would be the only person to ever comment.   '),(11352,2033616,'04,July,2004','2004-07-04 00:00:00','    urlLink    Nat, and our cabin of girls in Okoboji last summer&nbsp; urlLink   '),(11353,2033616,'02,July,2004','2004-07-02 00:00:00','    urlLink    This is just an adorable picture of me and Tra and Misty counseling Middle School 1 last month.  I just love this picture&nbsp; urlLink   '),(11354,2033616,'02,July,2004','2004-07-02 00:00:00','    urlLink    Camp friends from last year.  Me, Jen, Lindsey, Megan, Lea, Jamie, and Kelsey&nbsp; urlLink   '),(11355,2033616,'02,July,2004','2004-07-02 00:00:00','    On the road again... well tomorrow.. or rather this morning... as my departure is just 6 hours away.. i leave again.  I will be gone for three weeks but not three weeks straight... i will be back on the weekends so i suppose i will be able to say hello to people then but i will be much to busy doing laundry to do anything.  Anyway, i am very very very very very very excited for camp.. both rimrock and okoboji for both camping and counseling.. so... i thought i would leave you with the above pictures to show you i am having a good time yay.  Have a wonderful 4th everyone.  Send me comments if you love me :-)   '),(11356,2033616,'02,July,2004','2004-07-02 00:00:00','    urlLink    The winner\'s circle of the slip and slide tug of war...&nbsp; urlLink   '),(11357,2033616,'01,July,2004','2004-07-01 00:00:00','    You were right when you said I never listened to you. I should have. You were right when you said I didn\'t trust you.   I should have. You were right when you said I wouldn\'t hear you out. I should have. You were wrong, when you said I would forget all about you. I can\'t!   '),(11358,2033616,'03,August,2004','2004-08-03 00:00:00','    I am a hopeless romantic.  You know this... you must.  It\'s written all over.  In the movies i love, the songs i adore, the colors in which i dream.  But I\'m becoming impatient and it\'s breeding a loss of self-control.  My mind is wandering and coming to conclusions that there MUST be something wrong with the person I am.  Am i that horrible?  Do i annoy so much that someone would just turn away?  These questions keep teasing me.  Well I\'m not going to hear them anymore.  Drew, (who i might add, despite our small fallouts, like Chad, has always taken good care of me) reminded me of my \"end of school year post.\"  The fact that he saved it, and that it meant that much to him was so very nice to hear.  Thank you so very much Drew.  You will notice that my list starts at #2.  number 1 was just a little harsh.  It was revamped and you can find it in my archives on may 30th if you like.  Here is my list.  2. It is but hadn\'t ought to be... Nothing is more painful than seeing every one of your fears and failures play out again and again. What can you do... take the bullet and keep running.  3. The fool doth think he is wise, but the wise man knows himself to be a fool Woe-is-me should not be met with sympathy. Go to someone who\'s wise enough to tell you \"you\'re a screw up, but it\'s okay, now what can you change?\"  4. How lucky I am to have something that makes saying goodbye so hard When they leave, why do we pour over the sadness of the end. Get your head out of the ground, just because you won\'t let the end come doesn\'t mean it won\'t. They\'re starting new lives, doing new things! This is exciting, not sad.  5. The most beautiful discovery true friends make is that they can grow separately without growing apart We don\'t all hold hands in life and trudge forward, we do this on our own. Sometimes it\'s hard to realize our paths stop crossing so often. But I\'m still here... i\'m just walking an extra three feet to your left.  6.If you\'re headed in the wrong direction, God allows U-turns. You\'re falter isn\'t your road to failure. Surprise some people and make them proud; dig yourself out... they\'ll help you if you need it.  7.One of the symptoms of an approaching nervous breakdown is the belief that one\'s work is terribly important. There\'s a little picture, and a big picture. The big picture is the reward of realizing the little pictures really arn\'t that amazing anyway. Don\'t stress. Your fate will never lie in a single task.  8. Worry never robs tomorrow of its sorrow, it only saps today of its joy Your plan will never be hitchless, your intention never fully comprehended. Understand that and you\'ll be rolling before the punches even start comming.  9. If only we\'d stop trying to be happy we could have a pretty good time. Be cheerful, not happy. Happiness will fade, but cheer is there in good times in bad.  10.Sometimes I wish I were a little kid again, skinned knees are easier to fix than broken hearts. Life is sad, goodbyes are sad, broken hearts are sad, the world in general is sad. But things will heal, the sun will shine, and maybe, just maybe you can work something out.   '),(11359,28451,'31,July,2004','2004-07-31 00:00:00','   damit! no....dam you trent! i am week! so which bands playin tonight? '),(11360,28451,'30,July,2004','2004-07-30 00:00:00','   zenny returns from the bar with her jameson and coke, that she just had to nod at the barmaid to get *smiles*  I\'m glad you guys are embrasing the pub atmosphere, just spread the word if you can. I don\'t have everyones email, so if you guys can either let me know, or get ppl to message me with it, it would be apprieciated.   The bar is back baby, and the drinks here are free! *laugh* '),(11361,28451,'29,July,2004','2004-07-29 00:00:00','   complements to zen for a excellent expierement, i think we are all going to get hours of entertainment from it '),(11362,28451,'29,July,2004','2004-07-29 00:00:00','   as vandal settles down with his coke trent returns from the bar with his bourbon... at this point vandal realises the days when he was satisfied with coke are long gone and he returns to the bar to purchase something he enjoys more.... is it a smirnoff black? is it a bourbon?&nbsp; only nick can answer this question  &nbsp; '),(11363,28451,'29,July,2004','2004-07-29 00:00:00','   and thus vandal enters, buys his no alcholic beverage and settles in a corner to watch people as always. im glad this is here now.... back to my old ways.... ahhhh.... hello secrity blanket... '),(11364,28451,'29,July,2004','2004-07-29 00:00:00','   I was up in Sydney over the weekend when I stumbled across a promotion advertised at Shellharbour\'s local  urlLink Cellarbrations  retailer. The advertisement assured me that if I were to purchase a six-pack of the German&nbsp;imported beer,  urlLink DAB , I would be bestowed with a glass beer stein of gargantuan proportions. 1 litre, I believe. That\'s three regular earth beers! As my affinity for drinking beer from large containers knows no limits, I made the trip and purchased two. Pictures of me drinking copious amounts of lager will surely follow. Stay tuned. '),(11365,28451,'29,July,2004','2004-07-29 00:00:00','   Sweet....this is a cool idea, however one thing is missing from this wonderful place.   THE BEER...There\'s no beer.&nbsp; How can you have a pub with no Beer.&nbsp; To have a pub without Beer is like have corn chips without Salsa....   IT JUST ISN\'T RIGHT   However to do the Pub things without booze may help us all in our quest to drink less and party more.   and just think no hangover in the morning...not that I get hang overs any more, you see I have a secret weapon againest hang overs.&nbsp; It\'s called Lynx Anti-Hang Over Spray and I think it works.   Mind you I haven\'t really tried it yet as I haven\'t been out drinking since I brought but the Radio says it works so it must....because in Australian Advertising they never lie.......MUCH   Catch you round like a donut '),(11366,28451,'29,July,2004','2004-07-29 00:00:00','   still no word.... '),(11367,28451,'29,July,2004','2004-07-29 00:00:00','   So there i was, drinking, when i did something wrong, those of you that have other halves know how easy it is to do something that is wrong... so then later i was at the pub, and i did more wrong, what i did this time i knew was going to get me in trouble more but i felt it needed to be done as i have certain principles that i live by, and i feel that being untrue to these principles is being untrue to myself, and if that ever happened it would be an endless circle to self destruction...   so this is where i am, in a predicament, can anyone help?&nbsp;&nbsp; i didnt think so.... unless someone has one of those \'no more feelings\' pills.... which would be a great invention, lots of money to be made there..... i might suggest that one to my chemist friends.. once i get some chemist friends..  this is all for now Trent '),(11368,28451,'29,July,2004','2004-07-29 00:00:00','   I find the chairs in this pub a little bit more comfortable than the wooden bench of our last pub, however they do lack a certain charecter, not to mention the lack of beer, bourbon, jameson, scotch, lemon lime and bitters, and whatever else we may have been drinking....&nbsp; Well this is my first post and i hope to have many more... this is all i can think of writing for the minute, bye '),(11369,28451,'29,July,2004','2004-07-29 00:00:00','   Ok, so here it is.  We sat at the pub in Shepparton pretty much every friday night for I don\'t know how many years. We drank, we talked, we drank, we laughed, we drank, we danced, we drank, we sung, we drank, and we drank.  Now we are spread all over who knows where and its time we had a pub to go back to. Well a virtual pub at least. So here it is. Talk about what your up to, the last movie you saw, complain about your job, tell us about your great weekend, leave messages, talk about the universe, whatever! Just talk.  Lets give this a shot, and keep in contact, before we all go get married, and pop out some babies, become the boss of some mega company, and get so busy we never speak again.  So get to it! '),(11370,28451,'03,August,2004','2004-08-03 00:00:00','   i shall be missing from the pub this friday and sat. other night spots await and your all invited. friday night is abyss. it opens at 11 at inflation night club on king st and is $15. gothy. vey. should you want to come wear apropriate clothing.  strait black is fine. sat sees my frined matt playing at the termins in clifton hill. kinda blues/rock kinda thing. pretty dam good. dont know cost or anything but cheaper than fri for sure thought earlier. feel free to join me! '),(11371,28451,'03,August,2004','2004-08-03 00:00:00','   its quiet in here this time of week.... '),(11372,3878251,'29,July,2004','2004-07-29 00:00:00','    .flickr-photo {	border: solid 2px #000000;}.flickr-yourcomment {}.flickr-frame {	text-align: left;	padding: 3px;}.flickr-caption {	font: 75%;/*	color: #666666; */	margin-top: 0px;}.flickr-buddyicon {	margin-right:5px; 	vertical-align:middle;	border: solid 1px;}.flickr-postedby {	font: 75%;}  	 urlLink'),(11373,3878251,'28,July,2004','2004-07-28 00:00:00','    .flickr-photo {	border: solid 2px #000000;}.flickr-yourcomment {}.flickr-frame {	text-align: left;	padding: 3px;}.flickr-caption {	font: 75%;/*	color: #666666; */	margin-top: 0px;}.flickr-buddyicon {	margin-right:5px; 	vertical-align:middle;	border: solid 1px;}.flickr-postedby {	font: 75%;}  	 urlLink'),(11374,3878251,'27,July,2004','2004-07-27 00:00:00','    .flickr-photo {	border: solid 2px #000000;}.flickr-yourcomment {}.flickr-frame {	text-align: left;	padding: 3px;}.flickr-caption {	font: 75%;/*	color: #666666; */	margin-top: 0px;}.flickr-buddyicon {	margin-right:5px; 	vertical-align:middle;	border: solid 1px;}.flickr-postedby {	font: 75%;}  	 urlLink'),(11375,3878251,'25,July,2004','2004-07-25 00:00:00','    .flickr-photo {	border: solid 2px #000000;}.flickr-yourcomment {}.flickr-frame {	text-align: left;	padding: 3px;}.flickr-caption {	font: 75%;/*	color: #666666; */	margin-top: 0px;}.flickr-buddyicon {	margin-right:5px; 	vertical-align:middle;	border: solid 1px;}.flickr-postedby {	font: 75%;}  	 urlLink'),(11376,3878251,'23,July,2004','2004-07-23 00:00:00','    .flickr-photo {	border: solid 2px #000000;}.flickr-yourcomment {}.flickr-frame {	text-align: left;	padding: 3px;}.flickr-caption {	font: 75%;/*	color: #666666; */	margin-top: 0px;}.flickr-buddyicon {	margin-right:5px; 	vertical-align:middle;	border: solid 1px;}.flickr-postedby {	font: 75%;}  	 urlLink'),(11377,3878251,'23,July,2004','2004-07-23 00:00:00','    .flickr-photo {	border: solid 2px #000000;}.flickr-yourcomment {}.flickr-frame {	text-align: left;	padding: 3px;}.flickr-caption {	font: 75%;/*	color: #666666; */	margin-top: 0px;}.flickr-buddyicon {	margin-right:5px; 	vertical-align:middle;	border: solid 1px;}.flickr-postedby {	font: 75%;}  	 urlLink'),(11378,3878251,'23,July,2004','2004-07-23 00:00:00','    .flickr-photo {	border: solid 2px #000000;}.flickr-yourcomment {}.flickr-frame {	text-align: left;	padding: 3px;}.flickr-caption {	font: 75%;/*	color: #666666; */	margin-top: 0px;}.flickr-buddyicon {	margin-right:5px; 	vertical-align:middle;	border: solid 1px;}.flickr-postedby {	font: 75%;}  	 urlLink'),(11379,3878251,'17,July,2004','2004-07-17 00:00:00','    Any time spent in the great city of Melbourne with the Professor is incomplete without a quick trip to JB Hi-Fi for a sly CD bargain. Thus, Thursday afternoon, between lunch on Lygon and bevies off Brunswick, we squeezed in some quality time in the gaudy, jam-packed outlet.  This trip I restricted myself to picking up a cheap copy of an old Eels album (that I have previously been remiss enough to neglect) and the first Lemon Jelly CD that until now, I only had access to on the Professorâ€™s PC (not always convenient).  But now itâ€™s Frankyâ€™s! The ultimate in lazy afternoon background music as it hums along, my foot bouncing, head tapping side to side to side. Mmm. Love itâ€¦   '),(11380,3878251,'14,July,2004','2004-07-14 00:00:00','    .flickr-photo {	border: solid 2px #000000;}.flickr-yourcomment {}.flickr-frame {	text-align: left;	padding: 3px;}.flickr-caption {	font: 75%;/*	color: #666666; */	margin-top: 0px;}.flickr-buddyicon {	margin-right:5px; 	vertical-align:middle;	border: solid 1px;}.flickr-postedby {	font: 75%;}  	 urlLink'),(11381,3878251,'13,July,2004','2004-07-13 00:00:00','    .flickr-photo {	border: solid 2px #000000;}.flickr-yourcomment {}.flickr-frame {	text-align: left;	padding: 3px;}.flickr-caption {	font: 75%;/*	color: #666666; */	margin-top: 0px;}.flickr-buddyicon {	margin-right:5px; 	vertical-align:middle;	border: solid 1px;}.flickr-postedby {	font: 75%;}  	 urlLink    	  urlLink PICT0223 , originally uploaded by  urlLink Franky .'),(11382,3878251,'12,July,2004','2004-07-12 00:00:00','    .flickr-photo {	border: solid 2px #000000;}.flickr-yourcomment {}.flickr-frame {	text-align: left;	padding: 3px;}.flickr-caption {	font: 75%;/*	color: #666666; */	margin-top: 0px;}.flickr-buddyicon {	margin-right:5px; 	vertical-align:middle;	border: solid 1px;}.flickr-postedby {	font: 75%;}  	 urlLink    	  urlLink PICT0208 , originally uploaded by  urlLink Franky .'),(11383,3878251,'09,July,2004','2004-07-09 00:00:00','    A small but stellar cast features on this little amble through daily delights and it may be helpful to get to know some of them.  Franky, who sits here at the controls, is a Sydney girl without a vocation. With a penchant for baking and new shoes, it doesn\'t take much to delight this brussel sprout!  At her feet you\'ll often find the hero of our journey, Buzz, the giant Sheltie, a real family fellow. He has mastered the art of luxuriating in small pleasures, with running, eating and possum patrol at the top of his joy list. Franky has something to learn from he who treasures only a broken plastic plant pot and a chewed up golf ball. Beyond those few items, all other worldly things merely pass by with only the occasional nuzzle or sniff as acknowledgement of it\'s existence.  .flickr-photo {	border: solid 2px #000000;}.flickr-yourcomment {}.flickr-frame {	text-align: left;	padding: 3px;}.flickr-caption {	font: 75%;/*	color: #666666; */	margin-top: 0px;}.flickr-buddyicon {	margin-right:5px; 	vertical-align:middle;	border: solid 1px;}.flickr-postedby {	font: 75%;}  	 urlLink    	  urlLink Buzz , originally uploaded by  urlLink Franky .'),(11384,3878251,'09,July,2004','2004-07-09 00:00:00','    .flickr-photo {	border: solid 2px #000000;}.flickr-yourcomment {}.flickr-frame {	text-align: left;	padding: 3px;}.flickr-caption {	font: 75%;/*	color: #666666; */	margin-top: 0px;}.flickr-buddyicon {	margin-right:5px; 	vertical-align:middle;	border: solid 1px;}.flickr-postedby {	font: 75%;}  	 urlLink'),(11385,3878251,'16,August,2004','2004-08-16 00:00:00','    My natural state, no matter the weather, is bare foot, with toes wiggling and soles to ground. The professor often catches me tip-toeing on icy floors in the middle of winter and has to march me off for socks and slippers. Year by year I gain more skill and polish to my renditions of the graceful \"hot-concrete quick-trot,\" the spectacular wincing style of the \"sharp-gravel light-step,\" not to mention the ever popular, \"bindy-patch-discovery hold-that-pose\".  This foolish dedication explains my appreciation for the ever lazy thong (footwear, not underwear), the clever combination of bare foot benefits with the comfort of rubber. All I need is a pair of thongs in every colour and thereâ€™s my summer footwear sorted!  As for the wintry cold? I generally opt for my faithful black Eccos, with all the style and sophistication you expect from a pair of \"comfortable\" womenâ€™s runners. Not much. But they are quick, light and, well you guessed it, comfortable!  So why my fascination with shoes and all their trappings? Why my slavery to laces, buckles, tongues? Why my weakness for high heels, fine leather, stylish design? What is it about these beautiful little fashion accessories that has me quickly recalculating my monthly budget? What does a thong loving, sneaker resorting, barefoot at heart like Franky NEED with all these shoes?  I love them!    .flickr-photo {	border: solid 2px #000000;}.flickr-yourcomment {}.flickr-frame {	text-align: left;	padding: 3px;}.flickr-caption {	font: 75%;/*	color: #666666; */	margin-top: 0px;}.flickr-buddyicon {	margin-right:5px; 	vertical-align:middle;	border: solid 1px;}.flickr-postedby {	font: 75%;}  	 urlLink'),(11386,3878251,'11,August,2004','2004-08-11 00:00:00','    .flickr-photo {	border: solid 2px #000000;}.flickr-yourcomment {}.flickr-frame {	text-align: left;	padding: 3px;}.flickr-caption {	font: 75%;/*	color: #666666; */	margin-top: 0px;}.flickr-buddyicon {	margin-right:5px; 	vertical-align:middle;	border: solid 1px;}.flickr-postedby {	font: 75%;}  	 urlLink'),(11387,3878251,'10,August,2004','2004-08-10 00:00:00','    .flickr-photo {	border: solid 2px #000000;}.flickr-yourcomment {}.flickr-frame {	text-align: left;	padding: 3px;}.flickr-caption {	font: 75%;/*	color: #666666; */	margin-top: 0px;}.flickr-buddyicon {	margin-right:5px; 	vertical-align:middle;	border: solid 1px;}.flickr-postedby {	font: 75%;}  	 urlLink'),(11388,3878251,'09,August,2004','2004-08-09 00:00:00','    .flickr-photo {	border: solid 2px #000000;}.flickr-yourcomment {}.flickr-frame {	text-align: left;	padding: 3px;}.flickr-caption {	font: 75%;/*	color: #666666; */	margin-top: 0px;}.flickr-buddyicon {	margin-right:5px; 	vertical-align:middle;	border: solid 1px;}.flickr-postedby {	font: 75%;}  	 urlLink'),(11389,3878251,'05,August,2004','2004-08-05 00:00:00','    I first heard this album in the wee hours of a Saturday morning as I found myself sipping on gin & tonic and struggling to keep my tired mind on a game of 500â€™s. Nothing could have suited more the still shadowy night, the husky masculinity of the group, the thoughtful reserved nature of the game. This album combines a wry and delicate melancholy with a sexy carefree ease.  When I listen to it, in my mindâ€™s eye I rather fancy that I am soaking in the over brimming bubbles of a claw-footed bathtub, lounging in the candle light by an old gramophone. Of course, in my mindâ€™s eye, I am also much taller with wonderful chocolate brown curls piled up on my head and I believe I am smoking a cigar and sipping from a glass of whiskey. In my mind\'s eye, mind you.   '),(11390,3878251,'04,August,2004','2004-08-04 00:00:00','    â€œIt is dawn and the mauve sky hovers between night and day. The colonies of sparrows perched in the plane trees that line the square begin their chorus. The birdsong pierces the sky like a handful of thrown silver.â€  â€“The Witch of Cologne, Tobsha Learner   '),(11391,3878251,'03,August,2004','2004-08-03 00:00:00','    Trust the good old Womenâ€™s Weekly cookbooks to deliver another hit! This one was a little more fiddly than some but quite different and very tasty. I rustled it up to take for dinner to my two night classes but thereâ€™s plenty left for Ma and Pa Goodsmith to enjoy too. Professor will happily miss out as he is one of that delusional bunch who believe it best to keep savoury and sweet firmly on either side of the dinner/dessert divide.  Here \'tis:  1 cup chopped raisins 3 green onions,* chopped finely 1 apple, peeled & grated 1/2 tsp ground ginger** 1/2 cup breadcrumbs 1kg chicken thigh fillets 1 Tbsp olive oil 1/3 cup blanched toasted almonds*** 1/2 cup orange juice 1/4 cup plum jam 1/2 cup dry white wine 1/2 tsp ground cinnamon**** 1 cup chicken stock*****  * I used the ones with bulbs and long green shoots. Who knows what you call them wherever you are! ** As a hardened anti-ginger in savoury protester, I cut this down from 1tsp, to make a stand, you see. YOU can make your own choice about this, I know youâ€™ll do the right thingâ€¦ *** Donâ€™t be a total sucker for hard work like me, just buy them pre-blanched and pre-toasted **** Realised JUST NOW that I forgot the cinnamon. Canâ€™t vouch for tasty meal if added. Will try it next time thoughâ€¦ Yes, there will be a next time. *****Quite like using these little stars and wanted one last chance!  Combine raisins, onion, apple, breadcrumbs and ginger in a bowl and spread tablespoons onto each fillet, rolling them up and fastening with toothpicks. (This part is the fiddly, sticky-fingered part. Not recommended for tired or grumpy cooks. Relaxing ONLY if youâ€™re already relaxed!)  Heat the oil in a big saucepan and brown the thighs on each side. Remove from the saucepan and add the remaining of the ingredients to it, bringing them to the boil. Replace the chicken and cook for 20mins or much, much longer if you like a sticky rich sauce. I also added the remaining stuffing to the juices in the saucepan about halfway through cooking time.  Serve on rice with lots of crispy vegies. YUM! (Or you could plop it into a little tub to retrieve from your bag once it has congealed and wolf down the cold, slightly shaken chicken in your 15 minute dinner break spent in a garishly bright-lit TAFE cafeteria.)   '),(11392,3878251,'03,August,2004','2004-08-03 00:00:00','    .flickr-photo {	border: solid 2px #000000;}.flickr-yourcomment {}.flickr-frame {	text-align: left;	padding: 3px;}.flickr-caption {	font: 75%;/*	color: #666666; */	margin-top: 0px;}.flickr-buddyicon {	margin-right:5px; 	vertical-align:middle;	border: solid 1px;}.flickr-postedby {	font: 75%;}  	 urlLink'),(11393,3878251,'02,August,2004','2004-08-02 00:00:00','    .flickr-photo {	border: solid 2px #000000;}.flickr-yourcomment {}.flickr-frame {	text-align: left;	padding: 3px;}.flickr-caption {	font: 75%;/*	color: #666666; */	margin-top: 0px;}.flickr-buddyicon {	margin-right:5px; 	vertical-align:middle;	border: solid 1px;}.flickr-postedby {	font: 75%;}  	 urlLink'),(11394,3525639,'03,June,2004','2004-06-03 00:00:00','   so, on sunday, i\'m throwing a pair of flip flops on when my cell phone beeps -    \"fucking people are out of their god damn minds!\"  recognizing the voice as my best friend\'s, i let her ramble   \"so, i try to call you from my new cell phone thinking i\'m all slick because i memorized your number, and i get disconnected after two rings...ok, fine, so i redial, and instead get this dude\'s voicemail..so i think, hey, i\'m an asshole, i\'ll just cave in and look up the number because i obviously fucked it up....about 30 seconds later i get a phone call from the wrong number i dialed previously with this bitch on the line saying \"yo, why you be calling?\"  so i\'m like \"right, well, sorry, i fucked up, i dialed the wrong last digit\" and the chick on the other line goes \"slut\" and hangs up - but wait, holy fucking shit, here\'s this crazy bitch again, hold on!\"   so, she clicks over, and i\'m holding, for what seems like a really long time, when she clicks back and announces    \"dude! according to this bitch i\'m a crack slut dirty whore who this psycho\'s boyfriend can\'t afford to fuck because he\'s too broke..i need to go outside because it\'s a nice day and find some john to fuck because i am nothing but a nasty trick...yes, you heard it, i\'m nasty\" she then breaks down into peals of laughter which then sends me skipping down cackle road....  so, whatever, i finally get to jersey, and we\'re sitting outside on her deck, polishing off a bottle of stoli (raspberry natch)when she gets the idea to call this dude back and flip out on him...so, she tells me to call from my cell because it\'s only 1 digit off from this random guys...so, drunk fool that i am, i dial and, what do you know, it rings and then goes to voicemail, so, not wanting to start a voicemail war, i hang up.  So, we continue drinking, and slurring and laughing when suddenly my phone rings...it\'s the dude   \"hey ma....listen, i just want to apologize for my girlfriend..she\'s a little psycho..i\'m charging my phone in my car right now so she doesnt see me on the phone...she get\'s really pissed when girls be calling my number and shit\"     uh...right, so dumbass that i am, i continue talking to him   \"so like, my name\'s david...are you and friend single and shit? cuz like, ya\'ll be soundin sexy and shit \"    ok, so that\'s the point my friend rips the cell phone out of my hand and yells into the phone  \"ok, this is the reason that your skanky bitch thinks you\'re cheating on her, what the fuck is wrong with you...you\'re trying to pick up 2 girls from a wrong number call...go back to your bitch\"   and hangs up the phone.... she shakes her head and pours me another drink....  my cell rings again   \"hey, it\'s david, so like, it\'s really cool how our numbers are only a digit off yo, like, you be into swingin\' and shit?\"   good lord...   \"cuz like, me and my girl, we like to swing....\"   that\'s when i burst out laughing and hung up the phone..... me and my friend finished off the bottle of vodka, chain smoked for a while and went to bed....fucking people are out of their god damn minds...and shit  '),(11395,3525639,'02,June,2004','2004-06-02 00:00:00','   i am in the midst of trying to find a new place to live...at this point, my future roommate and I have abandoned any picky tendencies and are clinging to the idea that we\'ll find a decent size place that is cat friendly (yes, he\'s one of  those  people...*shudder*), in a relatively convenient area that is not on the 7th floor of a shithole walkup.  So far, we are not discouraged....we are hopeful, wistful jersey expats ready to challenge and conquer the manhattan real estate market...let\'s see how long this hopefulness lasts. I give it 2 weeks before he and I both lose our minds and are found shuffling around my neighborhood in our bathrobes blowing bubbles. In any case, the reason I need to leave my beautiful (though shabbily furnished - thanks to Jesus\' homegirl The cosmetics, well, they\'re easier to validate, as, hailing from jersey originally, i tend to pile it on like war paint....but all the hair stuff? whatever...god, i really am a girl let\'s just hope that everything fits into boxes...and hey, at least my new roommate and i will never have to worry about shampoo..I have 7 different bottles in the shower as we speak  '),(11396,3525639,'02,June,2004','2004-06-02 00:00:00','   disclaimer: I\'m new to this whole blogging thing, so, should i fuck up...be gentle with me...i don\'t know any better let\'s cut to it..i\'m fucking bored...like bored out of my mind.  My boss, the man I love more than my luggage (yes, a lil shout out to Olympia Dukakis), is..well, let\'s just say \"away\" for the next month or so. He\'s left me all alone in an office where most are either heavily medicated or those that aren\'t should be.  It\'s up to me to amuse myself at work, and lemme tell ya kids, it aint going too well.  Normally, when in a funk, assembling the toys inside of kindereggs does the trick ...not this time (my drawer full of plastic aliens, zoo animals and puzzles can attest to that). I tried the whole online shopping thing...yeah, not working out too well...i\'m the twit who pays for the shipping and handling charge on items from sephora when there are 2 within a 10 block radius of my soon to be vacated apartment...*sigh* I just like getting stuff in the mail...sue me I was getting by when the boredom was restricted to the hours of 9am..er..who am i kidding, more like 9:20am x 5pm...but then, it started carrying over into my evenings and nights...that\'s when i stumbled across my new best friend  urlLink netflix .  Let me just tell you, as a recovering blockbuster delinquent, this is a godsend...no late fees...EVER...what a motherfucking concept...plus, again, i get stuff in the mail..score! um..kinda...problem is, I can\'t exactly justify watching the musical episode of buffy at my desk when I should be scheduling surveillances and shit so, back to being excruciatingly bored essentially, this is just something for me to do that at least can pass for work when my superiors walk by... here\'s to hoping i don\'t get caught...otherwise, I\'ll have nothing left but  urlLink amazon     urlLink  '),(11397,3735538,'24,June,2004','2004-06-24 00:00:00','   lifes interesting now. certain events really scare me and make me worry about whatll happen next. my mom wont leave me at home by myself anymore cuz i had cammie over for about 4 hours on tuesday. at least i think it was tuesday...we were both sick, and her mom didnt know where she was.  HOURS LATER...  went to cammies earlier. had fun with her and preston, then annalise came, and we still had fun. cammie wished that by the end of the summer we would be able to tell each other that we loved each other. i dunno if ill be able to, cuz of the tally incident. but w/e, ill figure something out. listeing to soybomb rite now, sad cuz they broke up. in 24 hours ill b somewhere in idaho, sleeping in a stranger\'s house. well, at least i think ill be asleep. wouldnt count on it though. tomorrow\'ll be my last entry for at least a week. jeez its a 2 day drive to canada. hopefully itll b fun. ill miss cammie sooo much, it wont b fun. o well, canucks r cool. most of them anyways. mmm the LIVESTRONG braclet cammie\'s dad gave me smells good cuz cammie put perfume stuff on it. cant stop sniffing it lol. grrr im gonna miss her lots. as soon as i get bak, im gonna b moving. well, pretty soon afterwards anyways. i dunno what ill do if i cant see cammie every day. o god i dont wanna think about that. ill fall apart. i dont think she knows it but shes such a big part of me... ill cry at least once, im sure. '),(11398,3735538,'23,June,2004','2004-06-23 00:00:00','   The last few days have been pretty hectic. Cammie and me hit lots of rocks - big rocks - emotionally and now we\'re better. it was really tough cuz i think she loves me or something and doesnt want to say it, cuz she thinks i might not feel the same way. i cant say i blame her cuz from what happened with tally, its hard to say i love someone new. schools out finally, and i didnt beat bob up. goin to canada for a fuckin week on saturday, and its a two day drive just to get there. Staying in banff, alberta. goin with mckinley and stuff. i hope i dont get too bored. sounds boring. on the way bak i get to buy firecrackers and stuff from the indian reservation though, so that\'ll b fun. tubs of it. wow mckinley just reminded me he has one of my magic decks, so i hope my good cards are in there. ive been looking everywhere for them! im such a nerd, o well. its a fun game, really. i dont see why ppl have to look down upon ppl from doing stuff they like. w/e, doesnt matter to me. hopefully IHS wont hate too much. o god IHS is gonna b sooo lame. im not looking too forward to it. ill miss all the friends i guess i made at MMS, but i wanna start over. screwd up too many times. '),(11399,3735538,'13,July,2004','2004-07-13 00:00:00','   ha yesterday was a fucked up trip. first i found out that cammie came home sick from camp, cuz she had another dizzy spell and stuff, then i hung out with her till like 10 in the evening, from about 2 in the afternoon. afterwards i got home and was supposed to call her, but she wasn\'t home yet, cuz i think either her mom lied to me and said she wasn\'t home, or her dad had pulled over to yell at her cuz she just ran out earlier that day. whatever is was, i had told cammie earlier that i would be at her house around 1 or something. but anyways, after i got home i called my mom, got in the shower, then went to shelby\'s. had  a  drink. just one. once that was done molly was totally wasted and we hadta escort her back to her room. meanwhile as all this is goin on, her mom is asleep on the couch. shes still there and its 2:45 am. so anyways, i talked to molly about her emotional issues for like...2 hours. i think i helped her, but im not sure. rite now shes asleep, hopefully forgetting all thats happened since she started drinking, specifically chase. she really loves that kid, and he doesnt acknowledge it at all. i feel sorry for her, because she has apologized multiple times for cheating on him while they were goin out, and he still doesnt talk to her, much less be nice to her. she said he called her \"the biggest ho in the world\" and some other nasty names. it almost seems like the tally incident with just reversed sexes. just what i needed. o well, it feels good to talk. and help her. she says shell miss me, and i think she will. unless all this has been just a drunk episode. thatd b kinda funny if it was, ya? pretty chaotic. im done for now. peace. '),(11400,3735538,'11,July,2004','2004-07-11 00:00:00','   well, now cammies gone. ill miss her so much. i just finished writing her email, and now i hafta wait an hour before i go down to mollys and drink my boxers off maybe, lol. i kinda doubt ill actually drink, but maybe i will. also just watched the grand finale to BgC2 of RvB, it was funny. codename dirtbag! omally made off with lopez through the teleporter and now all hell\'s broke loose. i wonder what happens next. no more episodes for a few months. hmm...i really dont know what to write. i spent all day today with cammie, and it was fun. we had a water fight, among other things. ie jumped on the tramp, ate stuff, listened to music, chilled, etc. it was fun. ill really miss her. i dont know what to write so ill either edit this later or who knows. peace. '),(11401,3735538,'10,July,2004','2004-07-10 00:00:00','   cammie leaves tomorrow, and i move on tuesday. marches are still here, dunno when they leave. they kinda get annoying. so later im goin down to cammies again, actualy make that rite after im done typing this. i got a plan to do while cammies at camp, im gonna sneak out one night and help them raid stuff, then cammie an me will b together for a while at least. on friday im goin to the walkers cabin near Olympia so i wont get to see cammie before i go. ill hafta wait till monday! o well, at least ill get to see her again. i think im gonna stop taking my medicine cuz it makes it so i cant show any emotion. im always sure that i love cammie when i havent taken my medicine, but when i do take it im never sure. its really starting to bug me. but hopefully itll all be batter once i stop taking it. i really want to love her as much as she loves me. she sewed C-A-M-M-I-E into my sleeve end on my blacklabel jacket. and she stitched up the neck place thing. how sweet lol. ill be happy to see her as soon as she gets back. i cant say that i will miss her, but im sure i will. ive seen her every day this week. i gave her a stuffed koala bear thing which i thought was pretty ugly, but she acted like she liked it so, ya. at least she didnt make me feel stupid. we had a water fight today and afterwards she was worried that after this week i wouldnt love her anymore cuz we had been apart for a week. i reassured her that id still love her just as much, if not more, cuz i hadnt seen her for a week. i sure hope i still love her. really, ill feel so bad if i dont. i really dont wanna lose her. it was ruin me and her. im gonna go now, ill write more tomorrow. peace. '),(11402,3735538,'06,July,2004','2004-07-06 00:00:00','   walking a tightrope, ill tell ya. amy is convinced we made out, and we didnt. shes so dumb. i really dont like her at all. hate her in fact. i do realize i brought this on myself though. ill do what ive always done before, ignore it. cammie an me had a rough day today, but it all ended well. all because of a stupid ring. and avi. shes weird. hopefully i made an impression today that she didnt like. ha i \"called her fat\" once or twice, and i was gonna go up to her and call her the pillsbury dough-girl. man she woulda beat me up so bad. and her brother morgan was being a total bitch to me for a while. then he was shooting at me with his paintball gun and only CO2, leaving nice little white spots on my pants from frozen Co2, and i grabbed the end of the gun and just twisted it around. ha then he kinda stopped. he threw a ball at me, and hit me in the ass! what a freak. o well i didnt get really mad at him. grr really hate amy rite now though. shes really pissing me off. i love cammie more than she knows though, and i dont need amy there screwing it up. fuck im such a screw-up. o well, whats done it done. i didnt make out with her, and thats final. i didnt! i got my Darksteel forge in the mail while i was gone, and my mom was...not terribly pleased that i had used her CC again. lets just say  ninjas on fire  lol to quote  urlLink RvB . god im gonna miss cammie so much when i move. im gonna see if my mom will let her spend the night sometime. itd be so much fun. not like we\'d do much, but i bet my parents think we would. they suspect too much. im done for now. peace out. '),(11403,3735538,'05,July,2004','2004-07-05 00:00:00','   Well im home. been home for a day now, but yesterday i got sidetracked when i was gonna write. cammie gets home today, and im thrilled to see her. really. ive missed her so much over the past week, its sad. i love her lots. i hope she likes the present i got her. necklace and a ring. still got fireworks to use, saving them for my new neighborhood. gotta make a good first impression, ya know? pissed my sister off a lot yesterday, and had a good time doing it. the march\'s are at my house now, and their mom offered me a job back on their home island, but i had to turn it down, on account that cammie gets home today. i coulda made like 70-80 bucks from it though. nice amount for just a days work i think. she was gonna pay me 10 bucks an hour. thatd a been nice. o well. hmm these march kids r bothering me. o well, one more day. ill survive. might goto nicks on account of severe boredom. he cut his hand on glass yesterday and probably lost lots o blood. what a retard, but we still love him lol. ill write tomorrow. peace. '),(11404,3813860,'29,July,2004','2004-07-29 00:00:00','   So over the past few months, if you had walked past the Sears on Madison and State, like I do every day on my way to work, if you were paying attention, you would have noticed that the mannequins were wearing  swimsuits  and  penny loafers . Actually, they weren\'t  wearing  penny loafers, their feet were molded into likenesses of penny loafers. Way to go Sears marketing geniuses. Now&nbsp; I know why nobody wants to shop at your crappy store. I guess they were too cheap to buy mannequins with bare feet, for just one season. They thought nobody would notice, but HA! I notice everything. At least the penny loafers fit in for fall. NOT! Because now they have mannequins wearing  pajamas  and  bath robes  with their penny loafers. Maybe one day the idiots at Sears will be gazing out of their window at the Gap across the street, and realize that they could save money  and  be&nbsp;less ridiculous&nbsp;by getting rid of their lame ass penny loafer mannequins and hanging the clothes on body-shaped peices of plastic with no heads/arms/legs. Those things probably cost like $5  and  they don\'t piss people off who have to look at them every day. And come on, these mannequins don\'t have hair or facial features, but they are wearing  penny loafers ? That\'s some stupid ass shit.   Yes, I know I am crazy.   '),(11405,3813860,'21,July,2004','2004-07-21 00:00:00','   So today I left for work like 30 minutes later than I normally do, yet I still got to work on time. And I don\'t usually show up 30 minutes early or anything. I\'m not sure how that worked out. I think I have discovered my secret super&nbsp;power: The ability to always be on time. Because now that I think about it, I don\'t think I have ever been late to anything in my entire life. In fact, I bet I could leave for work at 9:30, and still get here by 9:00. That\'s why it\'s a super power. I mean, it\'s not as cool as being able to fly, or go invisible, but still.   '),(11406,3813860,'14,July,2004','2004-07-14 00:00:00','   Say you are eating cherries and most of them aren\'t very good. They are too ripe or not ripe enough, whatever. So you are finishing the handful of not so good cherries that you grabbed, and the very last cherry is absolutely delicious. It\'s nice and dark red, and firm yet juicy, overall a very good cherry. So then you go back to the fridge, get another handful of cherries, and unfortunately, again, most of these aren\'t very good. Here\'s the weird thing: If all the cherries had been good, you would have stopped after the first handful. It is the hunt for the elusive good cherry that causes you to keep going back to the fridge. It\'s like you have this need to eat 10 good cherries, and you have to eat 10 damn it, regardless of if they are the first 10, or if it\'s 10 good cherries out of 100! You might think this has never happened to you, but it has, you just never noticed it before. Watch.  mmm, cherries are good.   '),(11407,3813860,'10,July,2004','2004-07-10 00:00:00','   I saw Fahrenheit 9/11 last night, and, to my surprise, really didn\'t like the movie. I was disappointed that the movie really didn\'t offer any information that I already did not know. I was also surprised at some of the things that were said. For example, Michael Moore chastised Bush for not sending more troops into afghanistan, and more aggressively pursuing Osama bin Laden. However Michael Moore has previously stated that bin Laden should be presumed innocent until proven guilty, and that our actions in Afganistan were thus unjustified. I was also surprised that he spent so much time speaking about bin Laden\'s family, their contacts with the Bush family, and their flight from the country after 9/11. I was under the impression that the bin Laden family disapproved of Osama, and had cut off contact with him long ago. I would be willing to accept that the media had misled us (as they often do), and that my impression was incorrect, but Moore did not offer evidence to that effect.   I guess I feel like Moore would say  anything , even things that he doesn\'t quite agree with, or that he knows might not be representing completely accurately, to make his point. And if I was angry with Bush for lying to the American public to further his own agenda, than I can\'t accept similar behavior from Moore, even if I agree with that agenda.   so I pretty much think that if I, someone who STRONGLY disapproves of the president, and the war, was so unconvinced by this movie, then the chances of someone who intended to vote for Bush before seeing the movie will probably not be swayed. But who knows.    I guess that the average American is more likely to see this movie than to go out and read a whole book, but I would reccommend the book \"Dreaming War\" by Gore Vidal. Vidal makes many of the same points as Moore, in what I felt was a more intelligent manner. The fact that I\'d read this book before seeing the movie did not help me to enjoy the movie more. I\'m sure the movie would have been much more moving to someone who had not heard some of the shocking points that Moore makes. So I can admit that the movie did have that value. And I also admit that I found several of the scenes moving. I was just disappointed that the movie wasn\'t all I had hoped for.  I haven\'t talked to too many people about this yet. I\'ve been reading reviews, and it seems like EVERYONE likes this movie. I\'m curious whether any other democrats didn\'t like the movie. Maybe I am just weird.   '),(11408,3723398,'28,June,2004','2004-06-28 00:00:00','    well, i\'m in my first day of my MIMI training class at AspenTech ( urlLink http://www.aspentech.com/ ) in lovely Houston, TX.  i\'ll be here all week.  Houston seems nice, from the little i\'ve seen of it - just about as humid as they come - and that from someone that\'s lived all their life in Atlanta & Nashville.  training so far ... lessee ... interesting, but old-hat stuff so far for me.  it\'s been nice getting the fleshed out \"official\" explanation/presentation, and it\'s afforded me the opportunity to find out some particulars and details i wouldn\'t have otherwise known, but thus far, there have been no major revelations.  BUT, i suppose that would be a credit to how well Mike trained me.  ;)  if all goes according to plan, i\'ll be able to meet up with a highschool buddy of mine for dinner sometime this week - for now, i\'ll hope for some of my fellow trainees being hungry after class - and if not... off on my own, i guess.  ;) '),(11409,3723398,'23,June,2004','2004-06-23 00:00:00','    so here\'s thoughts on my new job (and old job), now that i\'ve had a bit over 1 week at it, full-time.  it\'s right nice to be able to step back and up a level, try to see and understand the big picture, and figure out the best course of action for said big picture.  i recently came from a \"fire-fighting\" job - that being, giving it all you\'ve got just to keep up - life limited to living in I and III of the Covey Quadrants(tm), never making it to that sublime state of Quadrant II.  i learned quite a bit, about the operations and details of the business.  knowledge which has - and shall probably continue - to serve me well, but it never afforded me the chance to change things, and make them better.  conversely, my new job affords me the freedom (almost to the extent of being loosely defined) to dig into a process, question it, consider all the angles and alternatives, and (if need be) push to change it.  it seems the primary challenge in accomplishing this change is how to push who, when, and how hard.  going off my past experiences, step one is to sell the changee on why they need to do it - telling them they need to do it probably won\'t go too far towards the end goal.  darned thing is, it\'s just so much easier (especially for us nerd-types) to \"tell\" rather than \"sell\" - hopefully i can remember that.  perhaps more to come on that at a later date.  other new things - gmail!  it\'s different, but so far, i like it - i feel like waving a big \"I TOLD YOU SO\" flag to my former boss, assuming the gmail philosophy catches on like (f)atkins.  he was big on us filing every e-mail into an appropriate folder, and keeping the \"Inbox\" down to a bare minimum of messages.  i kept saying that i could keep all my messages in my inbox, and find any given message faster (remember - firefighting) just by searching appropriate keywords.  i\'m not making the official change to gmail yet, but once it\'s out of beta - and particularly if they allow for POP3/other access - i\'ll probably take the plunge and go with it as my primary e-mail.  on a different gmail vein, everyone keeps yammering on about how horrible it is that they\'re screening your e-mail - privacy concerns, etc.  hmmm... sound like any other e-mail service or tool that auto-filters for spam out there?  how do they think it filters without screening your mail?  nevermind Yahoo! mail, which has been around for ages now, has been sending gmail invitations send to @yahoo.com addresses straight to the \"bulk mail\" folder - maybe i\'m a conspiracist, but somehow i doubt that\'s a coincidence.  bottom line - if you\'re not running your own e-mail server, you need to accept that *legally*, someone is going to potentially be reading/screening/etc your mail.  be that your home ISP, your employer - you name it.  if it\'s that confidential, run your own mailserver and/or encrypt it.  ;) '),(11410,3723398,'23,June,2004','2004-06-23 00:00:00','    ... it\'s hard to say how this will turn out (i really do just use the namesake of the blog in speech/etc ALL THE TIME).  i\'ve never been much of a maintainer of anything - i\'m usually doing pretty good to get the oil changed in my truck.  i attribute that to being a high-level thinker.  ;)  oh well.  guess we\'ll see.  so, i guess i should define the starting point, here... * recently changed jobs to a new department within the same company.  still very new at it, but quite enjoying it thus far. * gettin\' hitched soon, to a fabulous gal - got a lot to do, but am looking forward to it!  .... and ... that\'s it!  first post! '),(11411,3723398,'01,July,2004','2004-07-01 00:00:00','    so, I\'m a bit over the half-way mark with the MIMI training in Houston.  It\'s been good so far - still nothing earth-shatteringly new, but it\'s been good to get information \"direct from the source\".  However, that\'s hardly on subject with this post.  ;)  having not much else to do, i went down to the hotel bar tonight, and seeing on of my favorite beers on tap (and feeling splurge-like), ordered up a draught of Guinness.  so there i was, minding my own business, sipping my draught, when the older fellow to my right asks, \"So - y\'like your Guinness, eh?\"  hell if it wasn\'t a native irishman sitting right next to me!  he, oddly enough, was working on a Budweiser (which i later learned, was due to the \"when in Rome...\" reasoning).  a lengthy discourse, covering a variety of topics, from beer to sports to politics to beer to travel to work to beer... ensued.  said discourse was later joined by the two Swedes from my MIMI class that happened in.  So There I Was (tm) - me, an Irishman and two Swedes - in the suburbs of Houston, Texas - discussing life, liberty, the pursuit of happiness, and various themes thereupon.  it reached truly sublime levels when Latin American bartender Sergio joined in.  did we all agree all the time?  of course not.  did i gain an understanding of how closely [european] football is followed?  Sure.  did i get the chance to explain baseball to two Swedes who knew next to nothing of it?  Sure.  did the Irishman and I debate social policy - with the utmost respect for the others\' opinion?  Sure.  the above, to me, is the true joy in travelling - sure, seeing new and interesting sites and scenery is all well and good, but the real joy is the people you can meet along the way, if you take the time to talk to them and just find out a little about them.  i can only hope that i will soon approach that magical 3 weeks of vacation - which, as was discussed tonight, is quite sub-par in Europe (to the extent of actually being illegal, at least in Sweden) - such as to be able to travel more, and continue to meet yet more people from different places and see more things.  if there\'s one thing that i, as an admitted cheapskate, would continuously spend good money on, it\'s travel.  i can only hope that my future career - in both monetary terms and \"vactation\" terms - will afford me the opportunity to do so as often as possible.  daniel '),(11412,4225684,'23,August,2004','2004-08-23 00:00:00','   Can i claim to be who i am. The person im staring at in the mirror. Is he really i? Or am i just dreaming, one day waking up with another totally different personality as a totally different person, with a totally different life. Thoughts of such run through my head.       My existence. As i continue to journey through my life, i lose more and more of my consciousness, and i begin to realise... just how much of this is real.. What if this whole thing called my life was just a dream where i would wake up and never remember.      Im a christian and i believe that when i pass away, i will go on to be with God in heaven for eternity. But it doesnt stop thoughts from creeping into my mind. Thoughts of, what is eternity? Can there really be a forever? These short 75 years of an average human life if im even blessed with, what are they. The lives which we pass, are they just a repetition with our existences reliving it everytime we pass on, or is it forever etched in time once its passed, never to come back.      What is the Meaning of this all... What is the purpose... This moment im spending right now to type this all out... It will be passed, never again returned.. That much of a value i cannot describe, that i wish to savour whole heartedly, yet cannot...      How i wish, things were just so simple but they arent.....      \"Time passes me by as i sit and listen,   cries of folly, shouts of joy, happiness in abundance.   time it seems, we just have so much to spare,   what we don\'t realize is that time is not so keen to share.   what\'s passed, done and finished is over,   we will not be able to go back and start all over.....\" '),(11413,4225684,'22,August,2004','2004-08-22 00:00:00','   Let the lyrics speak for themself today      Pink Floyd - Time       \"Ticking away the moments that make up a dull day   You fritter and waste the hours in an off-hand way   Kicking around on a piece of ground in your home town   Waiting for someone or something to show you the way   Tired of lying in the sunshine staying home to watch the rain   You are young and life is long and there is time to kill    And then the one day you find ten years have got behind you   No one told you when to run, you missed the starting gun   And you run and you run to catch up with the sun, but it\'s sinking   And racing around to come up behind you again   The sun is the same in the relative way, but you\'re older   And shorter of breath and one day closer to death   Every year is getting shorter, never seem to find the time   Plans that either come to naught or half a page of scribbled lines   Hanging on in quiet desparation in the English way   The time is gone the song is over, thought I\'d something more to say\"      Relish in the remaining moments of your life... while you still have time.... '),(11414,4225684,'21,August,2004','2004-08-21 00:00:00','   Staring at this blinking cursor... It gives my mind many thoughts and lets it formulate many conclusions. Something about its continuity that grants me certain peace and solace.       Sitting here thinking, am i doing the right thing? What should be the right thing in the first place. Regretting? hmm i definitely hope not for this will be the first time... I have always been so sure, so positive of my actions and never have i once looked back to say that sucks and i should not have done that. Why am i doing so now? What\'s going on.....      Pride, honour, Promise.... You can lose it all but these... They will be with you till your very end and eventually for humans, will be the measure of their worth. Am i worthless? Does it really matter? or am i just doing it to please... Who can i please if i cannot even please myself..      My cries for help go unnoticed, unheard.. The people around me all keep smiling, keep moving in the herd... Gradually i learned to accept, to forget... That as myself an individual, it doesnt matter much who cares or who doesnt.. Whose to know whats true and whats false.. Whats the point..      Will i regret writing this eventually? No... i wont... But will i regret actions i take in my life like one i have just done... Maybe.. Sadly.. Maybe.. It is an eventuality which even i fear i will face but hold on i will.....      \"suffering but i have no regret,   this pain how can i ever forget.   misery which stretches to unknown degrees,   my life spells out to the finality of tragedy.   dreaming, wishing, hoping for the sadness to be taken away,   only to be dissappointed, let down,    left alone eventually....\" '),(11415,4225684,'20,August,2004','2004-08-20 00:00:00','   Want.....      Is it necessity or pure selfish desire.. We want many things in life.. Career, contentment, money, food and the list goes on. But what is actually necessary and what is just due to plain old selfish desire. Have we ever stopped to think about it?       I need that!! I want that!! Buy it for me!! Get it for me!! Sound familiar? Many a times i have been guilty of selfish desire, but at that moment of time, i don\'t really notice it myself. My mother always tells me that she will one day hire a cameraman to follow me around and video record every single thing i say and do. She says that i will definitely not even recognize myself if i were to watch the video.       Personally, i hate to agree with that but i would think its true. Sometimes i know myself that i am unreasonable and demanding but i will deny all such accusations at that point of time. Its only after the whole matter has passed on when i will actually realize and think about my actions and how selfish they were. Necessary? Most of the time no... Most of the time my actions were purely of selfish desire.. Afterall we humans by nature are selfish people.      What is this strange feeling called desire? Why do we desire? Desire is more then just simple want, its somewhat of a craving, of a need... We suffer from desire just like how drug addicts suffer without their drugs... Can we ever learn to break free from these bonds and chains called desire binding us to this material plane? Or will we forever be slaves to materialism and forever feed the great fuel of desire and let it grow into a great big blaze.      \"Her smiles ever so melting, makes all in place.   Her touch ever so haunting, a wind\'s caress could never replace.   Her laughter, angels chanter, reminds me of heaven.   Oh how i thank god with all my heart, for depriving heaven of her grace.\" '),(11416,4225684,'19,August,2004','2004-08-19 00:00:00','   Its not my fault. How many times in life have you heard these magical words uttered which seems to be able to shift the blame away from yourself.  Many times in life, we are all guilty of using these simple yet powerful words. Not my fault that happened. Not my fault this happened.    Well whose fault is it then? Before muttering those words without thinking, do we actually spare a thought to think about whether or not the situation might have been different if our actions had been different? So does it mean that it is not your fault and that the situation will still be the same if you had done pursued a different set of actions totally?      Like it or not, whatever actions we choose to do, it will in the end, affect those around us. Suicide will only result in loved ones breaking down and mourning over the loss. Betrayal will only cause those betrayed to be bitter and teach them to be more negative about life which is totally unnecessary. And love, well it can be both good and bad.. Ultimately, our actions, WILL affect those who are in immediate proximity to us.      Think for your actions are consequential. A simple lie can lead to great disaster and misunderstandings. A frown can sow discord and break relationships. And on the other hand, a simple smile can grow to a beautiful friendship which can be kept till the end of your life. Will you choose to act selfishly? or selflessly.. Make the choice, take the plunge but whatever it is, dont regret your choice for things in life can come and go, but choices we make will stick and be with us till we lie in our graves.      \"Saw your smile, felt your grace,   Entered a place called heaven,    wished it wouldnt displace.   Dreamt of kisses, Sensations of love,   Woke up alone, lying in that bed so cold.   The grip of that hand, the firm affirmation,   guides me closer, to what you call salvation.   My actions all lead me, all let me confirm,   that what i do is not entirely of me,   but totally for you.....\" '),(11417,4225684,'18,August,2004','2004-08-18 00:00:00','   urlLink   \"Hello, Good morning, How you Do\" Lyrics from Switchfoot - Learning to Breathe   When was the last time you actually greeted a friend with a proper greeting as such?    Do you take things for granted, that you will see your friend again or do you actually treasure every single moment spent with that companion.      When you smile, do you actually realise what you are doing? You can either be making somebody\'s day turn from bad to good or.. You could just be cheating the person if you didnt mean the smile. People see me smile, hear me laugh, but can they be sure it is for real? Have they ever stopped to think, wait a minute, he can\'t be laughing at this can he?       Personally, i take the good with the bad and whatever the situation, i believe in smiling. Even if its a forced smile where i have to cram up the corners of my lips, i force myself to do so. If i really have to force myself to smile i am in fact contradicting myself but i will do so because one sad person is enough... The world is full of sad faces, take the time to look around. Enjoy the lush scenery you see and ignore the frowns. Smile and be greeted with another smile and slowly, slowly all will be back in place....      \"As people age, As people grow,   They lose their sense, they make more foes.   Frowns a plenty, mostly blamed on responsibility,    Life they say, has taken such a toll.   Childish are acts which seem insensible to them,   but whose to say whose grown up when?   All one can do is smile and wait,    time will tell, chase your fate.\" '),(11418,4225684,'17,August,2004','2004-08-17 00:00:00','   Dreams.. Im sure every single one of us have them. Not just your simple crazy dreams of school life or adventures. But dreams of falling in love, getting married and even death.  What do dreams actually represent? Desire? Direction? Or are dreams really merely an indication of things to come.      Yesterday\'s dream was particularly intriguing. The whole dream was full of action and it involved all the people whom i were close to in my life. The most interesting aspect of it was that my hands were amputated halfway throughout (why though i still do not know why) and those people started to leave one by one. Suprisingly though, in my dream i did not cry when that happened. I only smiled and laughed as the dream ploughed on and finally only ending when i woke up in cold sweat. The dream really troubled me for a portion of the morning. It made me wonder what was my subconscious mind really feeling. Lyrics from a song Gary Jules - Mad World came to my head \"And I find it kinda funny, I find it kinda sad. The dreams in which I\'m dying, are the best I\'ve ever had.\"      Do i really know myself and can claim to be in control of my mind. Or is there a possibility of me actually losing it. That dream really scared me as it seemed like there was a possibility indeed. Can i claim to know more then i actually know or is reality just a cruel mirror which shows nothing but voidness.          \"Blatant lies are told everyday,   but your eyes seem to give the game away.   Where is truth, has it lost its way,   another broken promise, another dismay.   Where i have cried, where i have died,   i can only hope and pray to see a better day.\" '),(11419,4225684,'16,August,2004','2004-08-16 00:00:00','   Have you ever truly felt alive? What does it mean to be alive? To be able to see, hear, touch, smell, taste? Is that what it truly means? Or is there something else that is lacking in order to complete this equation..      Today was just another day where it seemed like i wasn\'t alive. Simple day with hardly any real interactions with people face-to-face, everything but was of mere conversations through text and pictures. I realise the time on this kind of days past by more slowly then normal, why though i still wonder.      On days where i am with people it is of a different matter altogether. Be it a whole day or just a short hour, it seems like time flies when theres interaction with other human beings whilst in their prescence.       Why do some days feel and seem shorter then others? Does it really mean that time flies when your having fun? Our time on this earth is short and limited. Many times i have heard people lamenting about how their lives suck or how much they just wish to die.. If i could i would show them what i see, about how limited our time on this life actually is and how many things there is to be done. Compared to people whose lives have been taken away from them unwillingly, here we have selfish people wishing for theirs to be taken away.   I guess we really like to take things for granted...      You will lose your life eventually one day. Choose the way to go when its the end. Either be a loser who never once saw the bright of life, or be someone who realised how precious it can be.. People often tell me to slow down, to stop thinking so much. Personally when humans dont think thats when they err..       We dont think about where we are going, what we are doing and most importantly of the consequences. We say many things which we never mean to do eventually and we do many things which we come to realise and regret after the consequences are clear. Many times in life i have made that mistake and awful a mistake it is to make. To realise that there could have been another alternative and to have been able to be better. Or to realise that it could have been worse and to have ended up in a lousier situation.      Instead of feeling that i am thinking too much, maybe, just maybe, i should feel that i am thinking too little and i should think more....      \"Time slips through hands like water on glass,    people feel that they have too many hours.   Minutes to kill, seconds to spare,   when they finally stop breathing, i hope they wont despair,   that time was too little, it just wasn\'t fair...\" '),(11420,4225684,'15,August,2004','2004-08-15 00:00:00','   Today an old emotion that i had not felt for many years resurfaced... It was like a breath of fresh oxygen after diving in that deep blue sea for so long. I felt my lungs pumped, my mind energized but most importantly, i felt my heart beat once more.  The emotion that im speaking of is undescribable.. Have you ever felt happiness just by looking at the wide grin on that face? Have you ever felt joy by just hearing that laugh so innocent yet jovial? Has your spirits ever been raised by a companion who does so little but yet it seems so much?  I sure am glad that i have found this emotion once again. Be it temporary or permanent, it sure was a great relief from many other emotions i have been experiencing lately. Though i hope for it to last, the sad chorus from november rain - Guns and Roses of \"Nothing lasts forever\" just keeps playing.. Well forever is a long long time, and i know that on my part, forever will seem like just another day as long as i continue to experience and feel this emotion i am so grateful to feel...     \"An Angels carress, so gentle yet full of feeling. A Demons words, so little yet full of fearing. Light from heaven fills my day, Dark from hell blinds my way. A child so innocent, so happy, so pure. How i wish, i hadn\'t learnt to adore......\" '),(11421,4225684,'14,August,2004','2004-08-14 00:00:00','   3.30 am on a saturday night and here i am, doing a blog... something which i had laughed and scoffed at when i had seen so many other people doing and reading blogs. Ironically, i had said i was never going to be interested in doing a blog as i was not that sort of \"person\".. Well i guess i did not even realise what that sort of \"person\" was when i said those words out aloud..    Why the sudden interest in blogging? Sleepless nights + Nobody to talk to (besides God) + lack of stuff to do = blogging... But more importantly, i felt i needed somewhere or something where i could express myself freely.. free of being criticized at the instantenous action of making a stupid remark or a careless slip of my tactless tongue...      I was once asked a question, if you could change anything in your life, what would it be... At that point of time i had not thought the question through carefully and just spurted out a simple answer of \"Nothing.\" I felt the lack of need to change anything that has happened in my life as anything different would not have made me the person i am today.. Thinking back on that question, if i were to be asked the same thing, i would answer the same way i guess... I would like to use this chance to say a big SORRY to anybody and everybody whom i have upset in any way or any manner.. be it a small let down to a big disappointment.. I want to say SORRY for being an insensitive jerk and for being the sorry person that i am.. There is only so much i can express verbally and through my words but i can only learn with each passing mistake and look up to those in heaven and pray as i smile on this earth still today..      Once more, sorry whoever you are, whereever you are.. Sorry for me being me...      \"Time whiles by, as i sit here today.    Dreams of you, offer nought but comfort to my mind.    Thoughts going wild, people telling lies.    When will we learn to abandon this foolish masquerade, to drop our disguise...\" '),(11422,4082385,'31,July,2004','2004-07-31 00:00:00','   a little while ago while looking back at the past, a bunch of old roommates had a bit of a reunion.  oh, and i finally got my motorola v400 to connect to my computer so here are some pix.  of course they had to come out with the v600 a couple of months after i got mine, and that has bluetooth.  *sigh*  how about not releasing new phones that quickly, ehh???  urlLink    bill-boy, woodhall suite-mate, dartmouth apartment-mate, campus village roommate. all that time, and still don\'t know cantonese :D  urlLink    hubie-boy, woodhall suite-mate, dartmouth apartment-mate, and all around shady guy. but he\'s cool because i think we were switched at birth, my exact opposite. where are the white girls?  urlLink    david the lovegod, whispering wood suite-mate, woodhall roommate, dartmouth roommate. what can i say, gone thru so much together. the aliens, my sister, tigger, polar-girl, the flight-attendant, oreo... *sigh* the best of times and worst of times. well at least i learned mandarin :D  urlLink    zerokitty... exposed! gimme back my camera!   '),(11423,4082385,'29,July,2004','2004-07-29 00:00:00','   What\'s the deal lately.  Seems like everyone I know is either getting married or moving away.  Don\'t they realize how inconvenient this is for me?  Or maybe it is just life telling me its time to get marr... uhm, move away too.  Sell off my house and my car and all earthly possesions... then move to some remote mountainside, live off the land, and teach the local wildlife Java (they already know VB).  Or maybe not.  But it is interesting to step back and reflect on life once in a while.  Look down on the world from above, see how the chess pieces (including oneself) move around, oftentimes in ways one would never have thought or hoped they would.  Change isn\'t as hard or scary from this perspective, and somehow things always tend to work out how they should in the long run.  Better not stand around and reflect for too long though, or else a bird will fly by and take a crap on my head.  So as they say, All Things Must Pass... which is true and a good reminder to accept it and just try to make each new phase of life better than the last.   '),(11424,4082385,'29,July,2004','2004-07-29 00:00:00','   someone told me  urlLink http://www.wibsite.com/wiblog/dull/ &nbsp;is the dullest blog on the web.&nbsp; sounded like a challenge to me!&nbsp; so here goes...&nbsp; =)   '),(11425,4082385,'19,August,2004','2004-08-19 00:00:00','   what is about the worst thing in the world that could happen to an obsessive compulsive pack rat computer geek?  hard drive crash? probably.  i won\'t call her my sweetheart, because we were only together for less than a year.  more like a one night stand and getting a regrettable disease.  funny how this \"new\" drive crashed and my old reliable P3 server, which has been on 24/7 for almost 2 years serving  this  and  that  has no problems what so ever.  i guess sometimes you just get a bad seed.  but before yelling at the warranty people, i had to try to salvage my priceless data (which hadn\'t been backed up since december)  so began late nights performing delicate surgery.  ran out instantly and bought a new 250 gig drive, loaded windows, just enough to make it usable.  then tried to read from the bad drive as a slave (since it wouldn\'t boot).  seems like i got most data from 2 partitions (although some seems quite corrupted), although i couldn\'t read anything from the old c: partition.  *sigh*, that\'s where my email is!  windows could see the partition, but couldn\'t recognize the format.  drat.  so i got some advice to boot into linux and try to salvage.  there is a cool linux distribution called  Knoppix  which you burn to 1 bootable CD, then without needing a hard drive, you can boot it up, instantly configures itself and is a fully functional linux operating system complete with a beautiful window manager, internet connectivity, and all the usual applications.  i was quite impressed, insert disc and it just runs in ram.  no configuration needed at all.  however, even using this, i was unable to mount the bad partition giving bad data block errors in linux.  crap.  as a last resort i heard of a  local company  that does hard drive recovery.  they said they just recover whole drives, not just a partition, and even then no guarantees.  since it was a 200 gig drive, they said it would cost between $500 and $1900 depending on data and how difficult it was.  well, you may be thinking, sure Josh it\'s not like you have a girlfriend to spend your money on... well true, but i\'m not *that* crazy and obsessive over a hard drive.  i said no thanks.  so now i sit backing up what i did recover to dvdrw, will reformat my new drive, and return the bad one hopefully.  so if you\'ve sent any important email to me since december, you may want to resend :D  and a good lesson to backup your data.  be like a friend of mine who keeps all important data on 1 cd, and reformats his computer every few months.  if you are a pack rat like me, this could be hard, but going through this made me realize there is more to life than crap on a hard drive.  i think so anyway???   urlLink     However, the week ended a little better.  Last night I saw  The Cardigans  at the House of Blues in Downtown Disney.  It was the first time seeing my favourite band from Sweden (only band from Sweden?) and highly recommend their  new album  that finally got released in the US.  Amazing to see Nina sing live not more than five feet away, *swoon*.  And then the stars must have been aligned right today, as during a staff volleyball game I somehow got myself involved in, I actually had a killer game.  Being the awkward tall guy, things didn\'t start well after joining, but after weeks of practice things are coming together.  Good thing, because one of the stars of the team, another tall spiker guy was out sick.  Luckily I had some good plays, and one to end the game with a rather strong spike that almost took the head off some very cute girl on the opposing team ;P.  The whole team was good today and we easily won our 3rd straight game.  Although our competition will get stronger later on in the playoffs, it is good practice to defeat these teams in straight sets.  And a good way to let off some steam of a tough week :D   '),(11426,4082385,'11,August,2004','2004-08-11 00:00:00','   So not too long ago I was helping out my  urlLink good friend  at work investigate this chat/photosharing application called  urlLink Hello .  It is a chat program, but a really basic one with not too many features.  However the cool thing about it is that you can trade pictures with other users very easily, and adds comments and point things out at what the other user is looking at.  This is not what interested us though.  It seemed there were hidden features to be found, such as if you typed certain words they would be auto-expanded in the chat or create animations in the photo window.  She did an excellent job of explaining  urlLink on her blog =) , so I won\'t repeat.  We eventually went a little further though.  While investigating the code in a hex editor (geeks past this point only, thanks) I found the references to the animation pictures (and the keywords to create them).  Looks like they are stored in a file in the \"Hello\\runtime\" directory called \"motio.psd\".  Hmmm, sounds like a Photoshop file to me.  Alas, it was!  And the reference names were to the names of the different layers on the photoshop image where each animation is stored.  So as you can see on the image below, we messed around with this a bit and traded modified files.  You can easily change the picture on each layer to however you want, save the file (using the latest photoshop version) and restart Hello to see new animations.  Obviously it only works for the computer you are on, if you want your friends to see them too, you have to send them your modified file and have them copy it to the proper location.  Unfortuantly to add new keywords would be difficult, have to wait for a new version with pluggable keywords (yeah right).  We were able to modify the .exe file a bit through the hex editor though, and change keywords or the expanded versions a bit (if they were the exact same length) with... some good and some bad results :D   urlLink   '),(11427,4082385,'10,August,2004','2004-08-10 00:00:00','   okay, enough with all the melancholy posts lately.  here is a fun one... sometimes i wonder how some people survive in the world...   urlLink   '),(11428,4082385,'08,August,2004','2004-08-08 00:00:00','   it was just over five years ago when i first met my japanese sweetheart.  very attractive, yet not too flashy, she stood there waiting to be taken home.  it seemed like we hit it off immediately and from that point on we were inseparable.  so many adventures we have taken together.  always ready to go to a concert, and willing to put up with my weird friends.  late nights driving alone in the moonlight, sunny days driving along the coast.  she was always there to keep me warm in the cold winter, cool in the hot summer, and dry during the fiercest of thunderstorms.  i too was always around to take care of her when she was sick or just lend an ear and listen when she was feeling down.  and so this rock solid relationship and love affair continued for many years until recently things began to change.  the bond we shared was no longer as dependable as before.  she began to experience problems that i could not understand nor help, and more consistently was no longer around when i needed her the most.  she began to notice my eyes wandering too towards a new life.  arguments ensued almost daily, and although we tried to work things out, it was clear things would never be the same.  although we have made it through both the good times and bad times together, little things in our lives consistently create more aggravation and headaches.  my friends say i should move on, but it\'s hard after spending so much time together.  call it loyalty, call it love, call it fear.  could it ever be the same or better with another?  yes soon i may have to take the leap, it may be time to get a new car, and say goodbye to my dear Honda Accord :(   '),(11429,3197709,'17,June,2004','2004-06-17 00:00:00','   I\'ve subscribed to the  urlLink JXTA  and  urlLink CHORD  mailing list few weeks ago...  '),(11430,3197709,'17,June,2004','2004-06-17 00:00:00','   JXORD is my Final year project which I am doing for my undergraduate degree. The JXORD name is derived from combining JXTA and CHORD. Its a distributed p2p file sharing app based on JXTA and CHORD.  urlLink JXTA  is peer to peer framework and set of protocols supported by Sun Micro Systems. JXTA will enable me to build my, basic communication network.   CHORD will be my lookup service; solving the problem of cutting down the overhead for searching.  urlLink CHORD  is a research project carried by the  urlLink Parallel and Distributed Operating Systems (PDOS)  group at the Computer Science Dept, MIT.  My intial objective for the system will be to :  1. Users to search for files in the JXORD network based on one keyword 2. Download files, after searching them. 3. Nodes can join and leave, maintaing the integrity of the JXORD network. 4. JXORD should be scalable.  '),(11431,3197709,'23,July,2004','2004-07-23 00:00:00','   As of today.. things are not going too well with JXORD..  I have given up the idea of implementing CHORD as a seperate module for JXTA.. as I figured CHORD is going to be a core service of JXORD and its a overkill to publish advertisments for CHORD itself.  May be I am wrong..   Pax  '),(11432,3197709,'25,April,2004','2004-04-25 00:00:00','    This my first test log here..   Packs!!!   '),(11433,3197709,'30,May,2004','2004-05-30 00:00:00','    I\'m Back!!!   Hum.. Havent been blogging for a while.. This just proves that I am not gone all phsyco abt getting a gmail account....  Its been quite a relaxing weekend with a couple of movies and lots of naps inbetween.. Oh did I mention I did my first JXTA program ??? Hum what a feeling!!!!  And on a sad note, I bid farwell to my german shephard, Roger.  He\'s been my pet for 12 years...and one of the gentle creatures I;ve known.. and would have set examples even for us...   Farewell roger!!!  Pax   '),(11434,3197709,'24,May,2004','2004-05-24 00:00:00','    urlLink    The miracle little bottle... in my room&nbsp; urlLink   '),(11435,3197709,'22,May,2004','2004-05-22 00:00:00','    Sloth Pees on his pants!!!!   Things are not going too well for the sloth (Premesh).  He peed on his pants and t-shirt.. I felt sorry for him.. But this was too good...  Look carefully.. Notice the stain ????   urlLink     urlLink     The sloth trying to cover his face in embarrsment... He started to stink thereafter...   urlLink    a href=\'http://www.hello.com/\' target=\'ext\'>     Pax   '),(11436,3197709,'21,May,2004','2004-05-21 00:00:00','    Movie Night!!!  Ahhh finally exams r over.. Screwed up all the exams!!! Dont give a shit!!!!  I m gonna have a movie night all by myself... back to back movies till the break of dawn.....    Pax   '),(11437,3197709,'20,May,2004','2004-05-20 00:00:00','    Online B\'Day   Ahhhhhh!!! This is the best(only) B\'Day party I ever had guys!!! That was a really lovely present!!!!!!!!!!!!! Couldnt ask for more........... Specially Prem dressing up as a cop and stripping.... I hope you got back your thong and pants back!!!!  Thanx for coming to my party Ara, Chari, Gumz, Neesh, Prem, Sahan, Kich & Ruwan....     Pax   '),(11438,3197709,'19,May,2004','2004-05-19 00:00:00','    May 19th   Happy birthday to me....  Pax   '),(11439,3197709,'18,May,2004','2004-05-18 00:00:00','    Dear Mr Blog   Dear Mr Blog  I cant be bothered to write anything today. I am very very busy so \"shoooooo\". I came to you when I was in need, but how many days and week have gone by... But still nothing. You are useless !!!   Pax   '),(11440,3197709,'16,May,2004','2004-05-16 00:00:00','    The Matrix and Buddhism   I\'m not a fan of The Matrix, but I came across this interesting artical and thought I  should share it with you   Although the presence of Christian themes is strong in The Matrix, the influence of Buddhism is equally powerful and evident. Indeed, the basic philosophical premises that drive major plot points would be nearly incomprehensible without a little background understanding of Buddhism and Buddhist doctrines. Does this force the conclusion that The Matrix and The Matrix Reloaded are Buddhist movies?   The most obvious and fundamental Buddhist theme can be found in the basic principle that, in the world of the Matrix films, what most people think of as \"reality\" is actually a computer-generated simulation. This appears to align closely with the Buddhist doctrine that the world was we know it is maya, illusion, which we must break out of in order to achieve enlightenment. Indeed, according to Buddhism the biggest problem that faces humanity is our inability to see through this illusion.    urlLink [more]   '),(11441,3197709,'15,May,2004','2004-05-15 00:00:00','    Using the F\' Word propely             Pax   '),(11442,3197709,'15,May,2004','2004-05-15 00:00:00','    Gumperson\'s Law   Ever wondered why, the harder you try to get something, or if you want something really really badly, the more difficult it gets ? Well it can be explained using a very simple rule....   \"The probability of anything happening is in inverse ratio to its desirability\"   I\'ve been following this rule for some time now, and it seem to really work. But becareful were you apply it !!!   Cheers Pax   '),(11443,3197709,'13,May,2004','2004-05-13 00:00:00','    Reengineering   Dear Manganelli, Hammer, Champy, Davenport, Short.  We know reengineering since the day we were born.. Thats what you call evolution..   '),(11444,3197709,'13,May,2004','2004-05-13 00:00:00','    Nothing to write   I\'m back to my old self again. Absolutely nothing to write today.   Packs!!!   '),(11445,3197709,'10,May,2004','2004-05-10 00:00:00','    Westheimer\'s Rule   For all those ppl who try to be optimistic....  To estimate the time it takes to do a task: estimate the time you think it should take, multiply by two and change the unit of measure to the next highest unit. Thus, we allocate two days for a one hour task  That means for a 9 month project...  9 months * 2 = 18 months 18 months converted into the next highest unit =  18 years   lol  Packs!!!   '),(11446,3197709,'09,May,2004','2004-05-09 00:00:00','    Best things in life are free - I   I got my domian name  urlLink upake.com  few weeks back from a domian name sale at  urlLink godaddy.com . I did not have any place to host so that was it... just the domain I couldnt do anything with it, as most of the places charged over 35$ to host the domain and do some considerable.   So screw those hosting guys!!!!  I had all the ingredients, a DSL connection which ran 24/7, a good pc to call as a server and all i needed was to map the domain to my IP. At last i found this kewl place  urlLink afaid.org . They let u do anything with your domian add A, MX.., unlimted subdmomians u name it. And plus they let you update your IP dynamically.. So no worries since I have dynamic IP.   I remmeber paying 25$ at pair just to get a sub domian.. Pair u suck!!!  I say this again \"Best things in life are free!!!\"   He He I just noticed a kewl new inteface for blogger... Quite clean and simple.. Love it!!!!   '),(11447,3197709,'09,May,2004','2004-05-09 00:00:00','    Upaka   Here\'s how my name came to be ( urlLink Source )   An Ãjivaka whom the Buddha met on his way between GayÃ¡ and the Bodhi Tree, after he set out from Isipatana for the preaching of the First Sermon. Upaka questioned the Buddha on his attainments, and when the Buddha told him what he had accomplished he asked the Buddha if he were \"Anantajina.\" When the Buddha acknowledged it, Upaka shook his bead saying, \"It may be so, friend,\" and went along by another road (J.i.81; Vin.i.8; M.i.170-1; DhA.iv.71-2). It is said (DA.ii.471) that the Buddha walked all the way from the Bodhi Tree to Isipatana - instead of flying through the air, as is the custom of Buddhas - because he wished to meet Upaka.  After this meeting Upaka went to the VankahÃ¡ra country and there, having fallen desperately in love with CÃ¡pÃ¡, the daughter of a huntsman who looked after him, starved for seven days and in the end persuaded the huntsman to give her to him in marriage. For a living, Upaka hawked about the flesh brought by the huntsman. In due course CÃ¡pÃ¡ bore him a son, Subhadda. When the baby cried, CÃ¡pÃ¡ sang to him saying, \"Upaka\'s son, ascetic\'s son, game-dealer\'s boy, don\'t cry,\" thus mocking her husband. In exasperation he told her of his friend Anantajina, but she did not stop teasing him. One day, in spite of her attempts to keep him, he left her and went to the Buddha at SÃ¡vatthi. The Buddha, seeing him coming, gave orders that anyone asking for Anantajina should be brought to him. Having learnt from Upaka his story, the Buddha had him admitted to the Order. As a result of his meditation, Upaka became an anÃ¡gÃ¡mÃ­ and was reborn in the AvihÃ¡ heaven (ThigA.220ff; MA.i.388f. Upaka\'s story is also given in SnA.i.258ff, with several variations in detail). The Samyutta NikÃ¡ya (i.35, 60) records a visit paid to the Buddha by Upaka and six other beings born in AvihÃ¡. According to the Majjhima Commentary (i.389), Upaka became an arahant as soon as he was born in AvihÃ¡.   '),(11448,3197709,'08,May,2004','2004-05-08 00:00:00','    MAXDB   Got my hands dirty with  urlLink MaxDB  today. MaxDB by MySQL is a re-branded and enhanced version of SAP DB,  urlLink SAP  AG\'s open source database. MaxDB is a heavy-duty, SAP-certified open source database that offers high availability, scalability and a comprehensive feature set. MaxDB complements the  urlLink MySQL  database server, targetted for large mySAP ERP environments and other applications that require maximum enterprise-level database functionality.  There are a couple of tools worth  urlLink downloading  after you install MaxDB, which will provide you with a GUI interface to administer your database server.    Cheers!!!! Packs...   '),(11449,3197709,'08,May,2004','2004-05-08 00:00:00','    Murali   He\'s done it: Muttiah Muralitharan breaks Courtney Walsh\'s previous Test record of 519 wickets as Zimbabwe went down by an innings and 240 runs at Harare....   '),(11450,3197709,'07,May,2004','2004-05-07 00:00:00','    Hey Jealousy...........  Hey Jealousy........... Hey Jealousy...........   '),(11451,3197709,'07,May,2004','2004-05-07 00:00:00','    god pls give me a gmail account.. I wont ask for anything else again...   '),(11452,3197709,'07,May,2004','2004-05-07 00:00:00','    This my first blog if you can notice.. I started this as I heard somewhere if you are a active blog user you can get a gmail account. So fingers crossed...   '),(11453,3197709,'24,June,2004','2004-06-24 00:00:00','    How to become a successful blogger in 60 steps    urlLink http://www.dreamwill.net/   '),(11454,3197709,'20,June,2004','2004-06-20 00:00:00','    Pax theory of baldness   Have excessive amount of body hair, and r u going bald at the same time ??  Ok I\'m not loosing my hair but, I dunno if you\'ve noticed, I definetly have; that ppl with lot of bodyhair, go bald...  Have u come across this anywhere in any artical ??? If not this could be my discovery !!!! If it is m gonna name it \"Pax theory of baldness\"  Pax   '),(11455,3197709,'17,June,2004','2004-06-17 00:00:00','    JXORD  I named my Final year project as JXORD and started a seperate blog for it, to keep a track of what i\'ve done and to organize my self.   Sounds quite boring but if u wanna take a look at it  urlLink  click here     Pax   '),(11456,3197709,'16,June,2004','2004-06-16 00:00:00','    Pimping my computer...   AAAHhhh!!! This wierd virus(Pimp.exe) is pimping my pc...I think its some kind of Spy ware...Its screwed up my browser too; opening up unwanted porn sites...  This is really annoying.. All efforts to remove the pimp from my pc has failed and finally decided to give it a format... Kill that bastard!!!  Hum.. and oh i\'ve finally decided to name my pc... cos few wierd ppl i know have named there notebooks. The  urlLink  sloth   has named his piece of junk, \"Kurt\" (an IBM PI that he carries around, which cannot take in any insertions.. I dunno what he tried to insert) after the late Nirvana dude, Kurt Kobain...  And  compact size  urlLink  N   has named her little toshiba as babytosh...  I guess it wont be too wierd to name my desktop.. after all we have a relationship like no other..and at this point in the relationship a name would be essential  But the  problem is I dunno what to name it... M clueless Any suggestions ????? She\'s really cute...So I would suggest give her a decent name....  Pax   '),(11457,3197709,'12,June,2004','2004-06-12 00:00:00','    Ants inside the Microwave..Experiment   Hum.. I am not going to confess anything... too bad..  But I\'ve notice somethine kewl recently  I usually use the Micro wave to heat up my tea and have notices couple of ants on top of the glass plate survive the couple minutes inside it!! was really curious so I googled it and found this kewl article   \"Ants can survive being \"nuked\" in a microwave? Microwave ovens have patterns of standing waves, with hot, very high-density areas, and cold, very low-density areas. Ants in the oven seek out the cold areas, and dodge the hot ones. If they run into a high-density area, they will survive, because their high surface area to volume ratio cools them more quickly than the ratio of large objects, and buys them enough time to locate a cold spot\"   Hum really really kewl stuff here... Maybe its like a merry-go-round ride for them.. I think u should try it too.. But keep in mind not to put any slothy ants.. you could end up killing them..  Before I read this i was gonna put my cat (\"Basheeee\")... Roasted Cat ????   Pax   '),(11458,3197709,'11,June,2004','2004-06-11 00:00:00','    I want to confess...........   I cant keep this inside of me.. I want to confess.... Pls help...   '),(11459,3197709,'09,June,2004','2004-06-09 00:00:00','    Its all to do with \"Karma\"   Well its been errr... cant remember two months since this crap.. gmail madness started.. and I can tell you dudes and dudets that I\'ve completely over it...  But u know sometimes I am a  bit  (pls note the word \"bit\") annoyed by it.. Just today good old  urlLink Chanka   got his gmail account.. Hum if you read his blog, its only 2 posts... Come on thats pathetic!!!!!  However the dude has got a invitaion from some another goo mail dude...  Hum... I\'ve done it all... Blogged.. Begged... still nothing.... and I\'ve come to a conclusion, and its all to do with  \"Karma\" , and its taken the best of me. Denied me of gooo mail account..   I knew I should have given my coins to the beggers on the road !!! :  Before u jump to any conclusion.. I am completely over it long time back.. Even if someone offers me a gmail account.. Kiss my ass...  Pax   '),(11460,3197709,'08,June,2004','2004-06-08 00:00:00','    Its all coming back!!!   Think its all coming back to me..  Ideas are flowing..  Cant even put them down into words.... May be Its all to do with  urlLink Venus...  Coming between me and the big guy...   '),(11461,3197709,'08,June,2004','2004-06-08 00:00:00','    Please Help   Trying to write something... My mind is blank. Have I\'ve gone insane.. Can you tell me please...    Pax   '),(11462,3197709,'05,June,2004','2004-06-05 00:00:00','    I\'ve become hopeless..   '),(11463,3197709,'23,July,2004','2004-07-23 00:00:00','    Its been a long day...  I \'ve been working with JXORD and there hasnt been any significant work done today.   CHORD seemed a pretty easy alogorithm.. but its really really mind boggling when you actually try to implement it...  Oh well I guess you wouldnt understand...  I know I am boring.. But I have to take my frustration out somehow....  Pax  &nbsp;   '),(11464,3197709,'14,July,2004','2004-07-14 00:00:00','    A page from the  urlLink sloths  Diary   This was stolen off the sloths diary...   Tuesday 13th July 2004  I was dreaming last night abt 3 girls, Which one is it???? I had a sleepless night last night and its going to be a long day at apiit.  It was around 4 o\'clock in the afternoon, pax lali and efam called me out to get a cup of tea.. And someone mentioned one the girls were down stairs near the reception.   I almost peed on my pants... I almost fainted ??? Whats happening to me ???... Is this love ???? I cant take this anymore.... I tell to my self i\'m gay I\'m gay for third time I\'m gay!!!!   Nooooooooooo... I\'m not gay..   Ok I downstairs...   k pax.. check out whether she\'s checking me out.. you come from behind.. me and efam will go in front!!!!  K my hearts going wild, the rush of blood into m face is making me red.. k my eyes are set on the door..   Oh!!! finally.. I\'m out!!!  Sloth > Pax, was she checking me out ?????  Pax > Yeap Sloth > Whoohooo!!!!    Stay tune for more!!!!  Pax   '),(11465,3197709,'14,July,2004','2004-07-14 00:00:00','    Hi I am back!!!   Aloha!!! Finally I am back.. I was on vacation in Hawaii... Its been Kupaianaha! (refer to a Hawaiian to english dictionary) experiance.... 2 weeks of sand and beach and the women were hoihoi. Aloha `ino! I had to leave in 3 weeks..  Anyways enough abt hawaii, my blogging is a bit rusty and few changes have occured while I was gone..   Whats quite significant is that \"Sloth\" has found his match...well its interesting to see someone going through the horrors of meeting his/her first love, for a fact \"Sloth\" has been less mobile today.. cos he\'s afraid he will run into this girl...  Sloth I know you want my advice... Here\'s are some pickup lines I learned in Hawaii..   Aloha!!! nani, u` i kaikamahine  \"`O wai kou inoa?\" `O Prem ko`u inoa. No hea mai `oe? No Dehiwala mai au.   K.. I m off.. till next time  Pax   '),(11466,3022585,'01,August,2004','2004-08-01 00:00:00','   Every morning when I wake up I think, \"This, this is the day that I will go to bed early!\"  I hate getting up in the morning.  I am so not a morning person that I am the anti-morning person.  If you talk to me as I am waking, I will growl at you, even if you are being ever so sweet and tying to gently rouse me.  I crave Saturdays when there are no waking up deadlines and sleep goes on and on and on.  But, somehow, and micraculously, after about 15-20 minutes on the weekdays, I begin to perk up, start making the coffee (thank God for coffee) and think that maybe I can make it through the day without vertically snoring.  I do end up making it through the day, but have to fight like hell not to take a nap in the afternoon.  The hours between 4-6 have become my nemisis.  The problem revolves around the fact that I stay up \'til about 12:30 or 1 everynight (I teach school so this is way late).  There is no real reason why I do this.  I am not writing a novel or developing photographs or even knitting an amazing sweater.  No, you can usually catch me doing, well, nothing I guess.  I can\'t acurately say what I do each night.  Sometimes I\'m reading, sometimes watching TV, sometimes reading blogs online, and sometimes talking to west coast friends.  I used to actually stay up to watch Letterman, but TV bores me more and more lately.    I have always been a night owl.  I can\'t remember when I didn\'t like staying up late.  It must have started in late elmentary school or middle school when I can remember watching Saturday Night Live and Johnny Carson with my family (I know, quality bonding time).  I remember in college wishing that exams were given at night because my brain was just so much more alert then and I was convinced this would improve my grades.  I still \"wake up\" at about 8pm everynight.  This is when I begin getting things done, the dishes, the laundry, the phone calls, the emails, the projects, the organizing.  About 2 hours into my second wind, I remember that I wanted to go to bed early so I don\'t feel like shit in the morning, but 10pm rolls around and I think I can give it one more hour and then I think if I go to bed at midnight I\'ll still get about 7 and 1/2 hours of sleep.  Sounds good, but this never happens.  So, everynight I stay up doing things that will sadly not bring to me to enlightenment anytime soon.  I wish that I could say that I have an extraordinary social life filled with late night gallery openings and wild parties, but no.  That\'s not me anyway.  Instead, I\'ll be washing my hair, helping my friends pick the perfect bridesmaids dress online or shutting the world out with a book.  Tonight I will be, um, let\'s see, um, well, how about reading my fabulous travel book on London (no, I\'m not going anytime soon, I\'m just obsessed) until midnight.  And tomorrow morning when I wake up I will think \"This, this is the day that I will go to bed early!\"  '),(11467,3022585,'01,August,2004','2004-08-01 00:00:00','   Who, pray tell, designed apartment buildings so that you can hear everything that is happening in the apartment below you with almost complete and utter accuracy?  I mean, is it that hard to place some sound absorbing panels in between the floors as you build?  Think, architects, think!    My neighbors downstairs are college students.  Why they live in our beautiful, mostly populated by the elderly and some young professionals, far away from the college they attend apartment complex, I don\'t know.  And, I actually, don\'t know them.  I know that one is a boy and one is a girl.  I know that the boy plays guitar and sometimes tries to sing along, mostly Dave Matthews songs.  I know the girl will occasionally put on some Jason Mraz.  (My connections to the above musicians would probably blow their minds, but I will never be friends with these college brats and so they will never know.)  And, I know that I can hear the boy cough at times.  What does that mean he can hear me do?  I wonder about this one.  Tonight, they are listening to the radio and they have what I think is a subwoofer cranked up so as I sit here trying to watch TV or listen to my music, I just hear deep voices and bass.  It BLOWS!!!!!!!!!  So, being the oh so nice and always ready to lend you some sugar should you be baking something neighbor, I turned up my subwoofer to the MAXIMUM.  Take that you slackers!  I am usually quite kind with my subwoofer.  When my brother installed it and put it at Maximum, I told him I had neighbors below and that it needed to go lower, a lot lower.  So much for thinking of other people.   The reason for this retaliation is that I think that they like to bang their ceiling, my floor, when they think that I am being loud.  I am not loud.  OK, so I always have some kind of noise on, I hate quiet, but when I am watching TV and they bang, and I check the volume, it is on 2 bars.  \"OK people, I pay enough in rent to listen to my TV at 2 FUCKING BARS!!!!!!!!!!!!!!\"  I am not actually sure what the banging is about.  I wonder sometimes if it is my imagination.  What could 2-3 pounding-like sounds in rapid succession be?  One day I came home and as I walked in the door I heard the 2 bangs.  I threw my stuff down and yelled at the floor \"I haven\'t even been here to make noise.  What, you don\'t like it when I open my door?\"    Maybe, just maybe, there is some phantom mouse running around and driving them crazy while I am constantly monitoring my volume.  I can\'t help but remember the Friends episode where the old neighbor below is always banging on his ceiling when the friends are not actually making noise.  When he dies and they go to his apartment they hear what sounds like a party going on above them, but they are all in his apartment. Could this be what is happening in my building? I don\'t know, but I always turn down the noise if I hear the banging until, of course, tonight and the subwoofer wars.  Game on!  '),(11468,3022585,'01,August,2004','2004-08-01 00:00:00','   Richmond\'s annual chili cook-off held many feasts for my eyes:  1) William Hung.  I know, you\'re jealous.  He was one act among many at the festivities and his one song performance of \"She Bangs\" was almost worth the $20 admission fee.  2) People wearing clothes that should have been retired years ago (or never purchased)- a) Belly shirts- these should only be worn by people who do not actually have bellies and then only when they want other women to hate them. b) Clothing that is too tight- there was a woman who looked normal in everyway except that her tan linen pants were a bit snug above her thighs, and by snug I mean she had a perma-wedgie. c) Stripper shoes- The shoes which I am referring to had a 3 inch clear plastic heel, silver souls, and clear plastic straps.  I know, you\'re jealous again.  Note to self- don\'t go to the chili cook-off as can\'t handle the dress requirements.  3) People running after beer- the beer trucks ran out of beer about an hour before the cook-off ended.  We were waiting in line and when they would announce that the beer was gone people would scatter quickly to another truck, and so on, and so on.  Note to self- don\'t go to an event where grown people run after silly things, unless there is a sale on at Saks.  4) A woman\'s bare ass- as I walked to my car I saw that the passenger door of the car beside me was open, blocking my door.  There was no one around and as I got closer and closer I looked for someone to tell that I needed to close the door.  Then, as I approached my car I saw that there was in fact someone behind the door, a woman\'s bare ass peeing.  Um, OK, um, um, look in the other direction, um, wait, um, don\'t look, hope that friend tells her I am here, um, um.  When she is done and is pulling up her pants over her bare ass and says \"I\'m sorry, I just really had to pee,\" say \"It\'s OK, do what you gotta do\" and quickly get in your car and pull away.  Note to self- never, never have to pee that bad.  '),(11469,3022585,'01,August,2004','2004-08-01 00:00:00','   When I checked the mail one day last week and saw the oversized, cream envelope with my name and address perfectly printed on it, I thought \"Oh hell, here\'s another one.\"  I open it to find an invitation to a cousin\'s wedding in May.  As I scan the card for the wedding, the reception, the directions, and the reply card I think, \"And so the season begins.\"  Every year there is a flow of weddings in my life.  I have probably been to at least 50 weddings since I was small.  Since college, they just keep on coming.  Someone told me that there is usually a boom right after college and then again in your 30s.  Um, there has been no let up in the bombing, apparently the troops refuse to surrender, give up and just go home.  I do have to say that I sometimes, am on the firing squad, well in some capacity anyway.  This year marks the seventh run as a bridesmaid for me  (OK, so 2 of them were my brothers, but still 7, come on people).  That\'s right, seventh (this does not include the ones in which I will merely be a spectator).  I love my friends and family and am so honored to be included in these special days and this is no diss on them in anyway.  This is all about me.  I mean, it\'s always about me anyway, right?    This year are the weddings of 2 very special people, my cousin who is like a sister to me and my best friend who basically is a sister to me.  These are probably the most special ones that I will be a part of and in my best friend\'s I am the maid of honor.  When she asked me, I cried, and then reality set in, this is my 7th time walking down the aisle, alone.  Oh, there are no pitty parties, Anna is an independent.  One friend said, \"Well, at least you\'re getting closer.\"  I was like, um thanks, that\'s as close as I want to be, for now anyway.  So, I do my duties of planning and attending showers, giving my measurements to strangers (probably the worst part of the whole process), and being there for the bride in any way she might need.  I do all of this thinking, as Cinderfuckinrella (thanks Pretty Woman) did, that someday my prince will come.  But, oh no, don\'t think he needs to rescue me, no.  I\'ve already rescued myself.  I want a prince for all the other reasons, plenty of cash, big house, private jet.  No, really, I want a prince who is prince of humanity in that he treats me and everyone I love with kindness, who is so funny that I laugh out loud just thinking about things he has done or said, who wants to learn to do things that I love just because I love them, and who wants me to do the same for him.  I think my friends (and cousin) have found these princes, I know they have.  I\'m just hoping that they haven\'t gotten the last ones.  '),(11470,3022585,'01,August,2004','2004-08-01 00:00:00','   I love NPR.  Music stations can make you feel like you are stuck on repeat of a mixed CD.  The same songs over and over.  People, please.  But then, there in the distance, on the hour, every hour, and during rush hour and some good hours on Friday night and Sunday, there is NPR to rescue me.  I like a pop song just as much as the next person, but I don\'t think that I ever actually will \"find [you] in a club with a bottle full of bub,\" but I do know people who have taken prozac which now appears to be causing suicidal tendencies.  (I heard this tidbit before eyeing it on any other news source that day.)  I love the idea of multitasking, driving and getting abreast of what is, or is not, going on in the world.  I can\'t watch news on TV anymore, unless it\'s BBC and then only when I\'m actually in England, because the inane (trying to be) analytic dribble drabble that procludes from these people\'s mouths is too much for me to stand.  I tell my father everytime that I am home that FoxNews is one of those broadcasts that actually does write its copy for people with an 8th grade education because that is there audience (if you didn\'t know, that\'s suppose to be the industry standard for level of writing, 8th grade, yeah I know).  He just ignores me and continues to be \"dumbed down.\"  Newspapers are fine, but I don\'t have the dough to subscribe to them and reading a newspaper online just doesn\'t seem right to me.  And so, when I get in my car and I\'m just sick of hearing the same 4 songs, I check the clock and thank God that it is after 4 in the afternoon and there is information to be learned as I drive  home.  Some days, I even leave early in the morning so that I can hear the more in-depth news as I drive in (the closer you get to 9am the more \"entertaining\" the pieces get).  I mean, all things considered I just need a little fresh air to handle this American life.  '),(11471,3022585,'01,August,2004','2004-08-01 00:00:00','   Today I heard a school secretary say to a little girl, \"Now, you\'re not going to get suspended anymore, are you?\"  As I passed by the conversation I wanted to say, \"No, you\'re wrong.  This little GIRL wouldn\'t do something to get suspended.  I mean look at her, she\'s cute, well-dressed, calm, and polite.\"  And then metacognition set in and I realized that I had just committed a mental act of sexism.  Our mothers might not understand why this scene might make us step back and take note of our place on the \"balance of the sexes.\"  My mother didn\'t even begin college until I, her third child, was in elementary school.  And she is the only one of her sisters to have a college degree.  A few years ago, she got her Masters in Administration (approximately 1 year before I go my Masters), hoping to become a school principal.  It seems that her age has hindered those hopes, sexism replaced by ageism.  (My mom is older than my friends\' parents.  She has a 39 year-old son after all.)  I\'m not sure if this is really true, but she continued teaching and thanks to Bush\'s unfunded No Child Left Behind, she has, in fact, been left behind.  It now appears that she must have in hand a different type of Masters to do the same job she\'s been doing for 3 years.  But, I\'m sure that 40 years ago she never would have thought she\'d be worrying about what type of Masters Degree she had.  Former generations of women only hoped that women would have as many choices and possibilities that we do now, but with the entire world laid out before us, what do we really have but more questions?  What if we just want to be married and have children?  And that\'s it.  What if we want it that simple?  Is it sacrilege to not take advantage of the choices, or is that the whole point anyway?  I personally would feel no shame in dropping the 50,000 dollars and 20 plus years spent on my education to be a happy homemaker.  But, alas, baby making and the perfect spinach quiche are not the path I stumbled onto, nor will it be anytime in the near future.  After all, I have a research project due for my \"might turn into a PhD\" class.  '),(11472,3022585,'01,August,2004','2004-08-01 00:00:00','   When riding my bike and a slight hill appears I must use all my leg power to tackle the mere incline (I\'m not talking real hills).  As I\'m peddaling with more and more emphasis I can\'t help but be aware of my powerhouse legs.  My calves are huge and this is not actually related to me being overweight.  There are two reasons for my treetrunk legs.  First, there is genetics (I know, blame your mother).  But seriously, I\'ve never seen the women in my family ever really break a sweat, sure there is the occasional walking of the dog, but no intense exercise going on, and yet these women do not have dainty legs.  We have legs built for... well I don\'t know what for.  The second reason is Linda Watkins School of Dance.  I was a dancer from age 4 to 18 and dancers have powerhouse legs, how else could they leap like that?  I tapped, balleted, toed, and jazzed my way through 14 years of jetes, plies, and time steps.  I usually hated dance class, but the recitals I lived for.  There is nothing like being on stage and performing.  My favorite was the kick line.  And still, years later, when I hear Elvis Costellos\' \"Hot Hot Hot\" (which is not that often) I want to throw some eye high kicks and do a split, oh there were splits.  Sadly, I couldn\'t do a split now if you paid me.  I miss the way a dancer knows her body, how every move is fluid and somehow perfectly calculated, even walking can be a dance.  Our dance instructor used to tell us a story about a dancer\'s grace.  The story goes that she was coming home from the grocery store holding a bag of groceries in each hand.  When she stepped up her back stairs to her house, she slipped on a thin layer of ice.  And this is where the grace comes in- she would emphasize that because she was dancer, she knew how to fall and thus landed safely on the ground, legs folded perfectly to soften the full body blow.  The strange thing is that I don\'t remember ever practicing or in fact ever falling in dance class, but I understood what she was saying.  A dancer\'s right hand always knows and reacts to what the left hand is doing.  I\'ve never broken a bone, pulled a muscle, or torn a ligament, could this be due to the dancing?  I don\'t know but, thanks to Linda Watkins, when I\'m 85 and fall down my front stairs, I may not in fact break my hip.  '),(11473,3022585,'01,August,2004','2004-08-01 00:00:00','   I was cruising on my bike through Windsor Farms today.  It\'s a posh neighborhood in Richmond\'s westend where the houses are big and beautiful, the yards are kept up by hired hands I\'m sure, and all children go to St. Catherine\'s, or St. Christopher\'s (cause being educated along ghettofied kids is to much of a reality for these folks).  OK, so I would like to own a house here, but my kids would go to public school, damnit.  On my bike tour past the beautiful AgeCroft (yes, some aristocratic cook had the house torn apart and shipped over here from England, but this was a long time ago when ghettofied kids would have been slaves).  The house is amazing and has a prescious spot in Richmond, it rests on a small hill and looks out to a small mountain which has not been tainted with developments.  It\'s serene, and reminds me of my hometown and of my last home, C-ville.  Growing up among moutains, I hated them.  My hometown is in a valley and therefore surrounded on all sides as if the hills protect and keep the people below.  And, in many ways they did, my hometown is a sheltered place.  Though quaint at times, it could possibly be the redneck version of \"A City Upon (surrounded by) A Hill (hills).\"  But, we all know what happens to Utopia, oh yeah, it doesn\'t exist.  So, for most of my life I have cursed the moutains and hills, especially when trying to learn how to drive a stick shift (I actaully gave up trying for a good year).  I didn\'t start to accept and reclaim them until this past summer in Montana.  I traveled to Montana for a friends wedding and when I stepped off the plane and found myself surrounded by mountains I thought I was back in my hometown except there are no semi-tall buildings and I don\'t think they know what urban sprawl is.  Throughout the weekend I kept thinking, \"What is so great about these moutains?\"  OK, so they are huge and the ranges are much bigger than Virginia\'s and they are bit more pristine, but really, no Virginian would be that amazed as I wasn\'t.  Then when drving back to C-ville from the airport in DC and going up and down the rolling hills, I thought \"These are my hills, Montana has nothing on these.\"  I think I began to love my state then.  I\'m proud to be a Virginian which is saying something, considering I have always wished to be an North Carolinian.  Richmond has no hills or mountains to look at, we\'re too far east.  I actually miss them.  Some days as I\'m driving to work I feel like I don\'t have everything that I need for the day, I check everything, my watch, my purse, my coffee and I and can\'t find the missing piece, but maybe it\'s the moutains, the Blue Ridge, that I\'m missing, without really knowing it.  '),(11474,3022585,'01,August,2004','2004-08-01 00:00:00','   My sister-in-law had an embroidered pillow which rested on her white loveseat which sat in an area off to the side of her kitchen in her former house.  The pillow said, in green thread, \"Bloom Where You\'re Planted\" and of course had simple, colorful flowers sprouting underneath the words.  The first time I saw this pillow, I thought \"What a great saying.\"  After 3 eye-opening, ego-blasting years in a teaching position that drove me to see a therapist, a move to a new city, a new job, and lots of downtime, I can now see the sunlight.  The cloud which followed me around all my life, much like the dust cloud that follows PigPen from Charlie Brown, has lifted.  The weather is not always clear, it rains and storms still, but mostly there are sunny days.  I\'ve moved to Richmond to start over and when I first moved here I was so excited and thrilled to begin again and to do it right this time.  Well, 6 months into living here, I was not happy.  The cloud was back, my apartment was a shambles, and I hated waking up to go to work every morning.  I was obsessed with New York City and sad that I could not afford to move there or to London, or anywhere.  Then I began to repeat and reflect on that pillow in my sister-in-laws kitchen.  And while I\'m not ready for blooming, at least for now, I think I\'ll germenate.  And I\'m sure, with all this sunlight, a stem will form.   Next Time  Ginko trees live 1,000 years. Eating the leaves will clear you brain. When I heard about them, I thought of my mother, how much I would like to sit under one with her in the ancient shade, nibbling the flesh, the stem, the central vein.  ~Naomi  Shihab Nye  '),(11475,3022585,'01,August,2004','2004-08-01 00:00:00','   I love the bridges of Richmond.  They are no architectural triumphs by any means.  They are simple and quiet, sometimes elegant, and to me, quite satisfying.    My favorite is the Nickel bridge.  It is a small two lane bridge that is made of intricate iron workings on either side and is lighted by short lamps which are part of the iron work.  It is a quaint bridge and one of the quite one I think.  As cars pass each other in close proximity, it\'s as if they are quitely saying hello to each other.    The one I take to work everyday, the Powhite Pkwy, is a toll bridge and although I pay everyday and hope that my $.50 will soon make enough of a difference that the tolls will soon go away, I make sure to enjoy the view on either side of me.  On the left is a train bridge that was clearly built when we took time for grandeur and beauty in transportation.  The train bridge is grey stone with slender arches that reach just under the flat top.  Everytime I look at this bridge I imagine riding in a cafe car on a train in the 20s, sipping tea and headed south to somewhere fabulous.  Just visible beyond the train bridge is the Nickel bridge, much shorter than the train bridge and sometimes I think it is a whisper of a bridge in the distance.  Back to the right side of the Powhit bridge.  On the right is only the James river and it\'s tree lined banks.  The river is a bit rocky at this point and every morning I can gauge the level of percepitation west of Richmond.  When the river is at a normal level, birds perch on some rocks that barely rest above the water level.  The first time I saw I bird sitting on one of these tiny rocks, I wondered if he thought himself a daredevil.   The Hugenot bridge is at the end (or beginning) of Cary St.  It extends into the posh westend of Richmond.  This bridge is two lanes with rails on either side that mimic a picket fince in metalic piping.  But, it\'s length is what is so fascinating.  It goes for at least a quarter mile and part of is not over water at all.  When you start the bridge from the Cary St. side you enter and curve around into a grove of trees.  Because it\'s a bridge a this point, you feel as if you are suspended through a small forest.  As you curve around, train tracks appear quickly and then the banks of the of the river.  It\'s a bit like a fairytale bridge, no exactly magical, but unnexpected.  The Willey Memorial Bridge mimics the Hugenot bridge in length and setting, but is far different in description.  The bridge is all concrete and lit by standard street lights.    The Lee Memorial bridge is my least favorite.  It connets downtown to a area that could be referred to as the \"projects,\" a sad and large part of this troubled town.  This bridge is open to pedestrians on either side so that the blank concrete feel is even more guarded by chain link fencing.  Although the bridge is wide with four lanes, you feel boxed in and can\'t see the water underneath.  The idea and realization of being suspended over rushing water does elevate the drab feel of this bridge for the brief trip over the James.  The 2nd st. bridge is another non-descript bridge, but on the downtown bank there is a memorial to Vietnam veterans of Virginia.  It\'s a beautiful memorial, a large stone wall in the background, a grand shape of a women inside, behind a glass wall.  I\'ve haven\'t seen it up close, it\'s on my list of things to do.  Just the other day I discovered an walking suspension bridge under the 2nd st. bridge.  The walking bridge literally hangs below the towering car bridge above.  It reminds me of the Millineum Bridge in London, another beautiful walking suspension bridge.  '),(11476,3022585,'01,August,2004','2004-08-01 00:00:00','   My name is Anna and I reside in the old town of Richmond, VA.  I\'m a Virginian by birth, but this my first time living in the capitol city.  I\'m a teacher for children with special needs in public schools.  I used to teach children with Autism which was fascinating.  For a long time, Autism was my passion.  At times, I still yearn to work with those kids again.  I\'m damn good at it, but the culture of Autism is too abrasive for me.  It\'s a complicated, long story.  I love the kids, it\'s just everything that surrounds them and thus henders a teacher\'s effectiveness was too much for me.  I like what I do now, but not sure if I love it.  What does it mean to love a job?  What does it mean to love anything?  I just turned 27 with no fanfair and am in the throws of confusion and searching that inhabits all 20somethings.  Things about me:  when \"Sex and the City\" ended I felt like my best friends moved away.  I love my nieces and nephew.  I love to read pop fiction (yes, Bridget Jones is a friend of mine).  I have TiVo.  I used to write poetry (fairly seriously in college and worked with some big names).  I have a very well decorated apartment, it\'s beautiful if I do say so myself.  I love to cook and don\'t do it enough.  I need to learn to exercise frequently.  I wish that I could: take amazing photographs, publish some poems/ book/ picture/ dna, travel to any city anytime, live in NYC, live in London, etc.  '),(11477,3022585,'01,August,2004','2004-08-01 00:00:00','   Do you remember watching music videos?  I mean really watching them, like for hours.  My brother and I, along with some neighborhood kids at times, would sneak snippets of MTV.  We were banned from such filth as kids though we seemed to watch endless amounts of it, our parents not as quick as they were prudish.    I was enamored with George Michael.  Funny, he never seemed gay to me at age 10, but then again I didn\'t know what gay was then either.  He was the perfection of hotness to me and that song, Careless Whisper, where he crones after a lover lost and that stubble, whew that stubble.   These days, he\'s not so hot and I wonder why I never saw the obvious \"batting for the other team\" going on, but again, this was the 80s, I donâ€™t think he was letting many people in on his secrets, lest he lose album sales.  A few months ago I was flipping through the channels and found FUSE, a music video channel here in Richmond where all they do is play videos.  I was stunned.  Some flashy hip-hop song was on and I paused, filling the commercial time during another program and actually watched a whole video.  I felt like I was 12 again, the volume turned down, the den door closed, and my finger on the remote ready to switch channels should the door open and my mother walk in.   I haven\'t watched another video since that FUSE moment months ago.  I miss the old videos where the stories were weak, if anything, the bands were overly made up, and the women had clothes on.  The current generation has no idea what endless, mindless hours of music videos are like.  They think MTV was created to expose the unreality of 18-23 lives.  The teeny boppers today won\'t have slumber parties and stay up late to see how many times their favorite videos appear.  They\'re lucky if they see their favorite video once during the terribly drab TRL.    And, what happened to a music video doing magic for the unknown artists?  The one hit wonders of the 80s would be nowhere without their videos.  Remember Divo and the whips?  My brother\'s band made a video a few years ago.  And if you blinked you missed it.  It did nothing for them because kids don\'t watch videos anymore, they want the (non) reality shows or the watered down, over-produced, uncreative pop star in their mist, and they only want the 10 best knows videos of the day.  Or at least, this is what MTV thinks they want.  MTV has gone from groundbreaking to heartbreaking.  And VH1, don\'t even let me go there.  And it\'s sad to realize that I am no longer the target audience of MTV.  They don\'t care about my demographic anymore, even though I busted my butt sneaking hours of videos into my childhood.  Sorry Dire Straits, you can have your money for nothin\'.  It turns out I don\'t want my MTV.  '),(11478,3022585,'01,August,2004','2004-08-01 00:00:00','   I was playing with my 3 year-old niece today when she said \"I\'m the king of the castle.\"  Then pointed to me and said, \"You\'re the dirty rascal.\"  And I was immediately placed inside the melody of CRASH.  I was irrovacably sadden that my childhood is now wrapped inside the lyrics of a Dave Matthews song.   Then later, in the backyard waiting for the hotdogs on the grill I heard a bright, highpicthed song lingering in the air.  I knew immediately that it was an ice cream truck.  Whew!  How long has it been since I\'ve seen on one of those.  Oh, the days of yore.  *sigh*  '),(11479,3022585,'01,August,2004','2004-08-01 00:00:00','   People, really.  You can leave comments.  I don\'t bite and I like to hear from the people I love in my life.  Really, I\'d like to hear from you.  Seriously.  If you stop by, tell me so.   Love ya,  '),(11480,3022585,'01,August,2004','2004-08-01 00:00:00','   I\'m funny.  Ask my friends; they\'ll tell you I\'m funny.  It\'s my thing, my party trick.  I come from a long line of funny people.  No, none of them are famous.  We\'re funny the way your family is funny.  The way you take family or friendly information and turn it around, spin it on it\'s head, give it a sarcastic tone or link it to a historical travesty and spit it back into the air.  My relatives have dry and sarcastic one-liners down to a science.  It\'s like a racket ball game when we are gathered together. The puns and witty observations fly so fast youâ€™re not sure you can hurl them back, but then, you do, you always do.  And so, I find friends who can play this game, who can come together in one accord on making fun of the world and each other.    Friday night, sitting amongst a group of co-workers for \"happy hour\" I was anything but happy.  Oh, I was jovial and jokey and witty and brilliant, but somewhere along the line I realized that this wasn\'t the real funny me.  I had to work hard that night to tone my references down and to pull what little information I knew about these people to the front of my brain.  I\'ve only worked with them for 6 months and this is only the third large group social thing we\'ve done.  There is a few staff I know really well and with them I can joke all day, there is enough reference material there, but when you stick me with a large group of people I only see in the hallways I can\'t do much.      My jokes on Friday landed on unfertile ground.  There were some hits, but mostly misses.  When you don\'t know the people well, you\'re jokes and puns and one liners have to come from outside the group, I usually choose current events, historical events, local events, world events, OK so anything really.  But I\'m sad to say I don\'t think my audience understood my references.  There was a discussion of a painting in the Mexican restaurant, which was of a man holding his own head.  I said I thought it was John The Baptist and then asked whose name was signed.  It was Raphael.  I stated again that I was pretty sure it was John The Baptist.  A co-worker said, \"Is it a religious thing?\"  \"Um,\" I thought, \"I give up.\"   Then there was my self-deprecating comment when another co-worker asked if I had gone to VCU with him, that I looked like someone in his ethics class.  I stated that I had no ethics, so no, I didn\'t go to VCU.  He didnâ€™t bowl over with laughter.  \"Whatâ€™s wrong with you?\" I thought and then realized I don\'t really know these people.    Sometimes I don\'t want to spend all that time it takes to get to know people, I just want it to be immediate.  My cousin\'s wife was telling me about a dinner party she and 5 other women have every month.  They get together and discuss politics by having various members\' research topics then they come together to discuss.  She said she thought of me and that this is something right up my alley.  Well, after tonight and the failed attempts at garnering some sort of recognition for my brilliant repartee, I think she is right.  I need some good conversation, I need some deep thinking, I some cognitive stimulation.  '),(11481,3022585,'01,August,2004','2004-08-01 00:00:00','   I went to Carytown today with my brother\'s little family.  My 6 year-old nephew\'s artwork was being displayed in a window of a store there, part of a city-wide school art project.  He was 1 of 2 children chosen from his 1st grade class to be placed for all the passersby to see.  It was a great picture full of blue and red and purple and white and yellow.  There was clearly a sky with clouds and then there are 2 distinct sections on the bottom.  One blue square section takes up the majority of the picture with two building shaped things inside.  To the right of the large blue square is a skinny blue rectangle with an American flag in it, just hanging out.   I asked the little artiste what exactly he had painted and he shrugged.  His grandmother said it was a house and then I asked \"Well, is it a house that you thought up in your mind or a particular house?\"  He said he just thought it up.  \"All right, well, good picture little boy.\"    I then began talking and playing with the little nieces when the artiste\'s mom states that the two buildings are \"Our house and Annieâ€™s apartment building.\"  I turned my body and swung my head around, \"What, my apartment building?\"  To which the little one said \"Yeah\" in his sweet, Well-what-else-could-it-be-? voice.  I looked at the picture again and there it was in all the many colors.  It was clearly a small one-story house sitting beside a 3-story brick apartment building.  How did I not see it before?  I said, \"Well, look at that, I\'m famous.\"  And then the love swept over me.  When a 6 year old paints your apartment building into a picture, that is love, real love.  Thank you my little A----baby.  I love you too.  '),(11482,3022585,'01,August,2004','2004-08-01 00:00:00','   I believe that dog-earing pages in a book is sac-religious.  I think I may have engaged in this practice when I was 12 for a day or two, but it never stuck.  It seemed too finite a thing to do.  Reading a book is a fluid thing.  Though you stop, put it down, go on about your life and come back to it, the story is fluid, or should be.  So, to me, dog-earing the pages is like telling the pages, even the words to stop going, stop moving because the crease is always there.  It stands as an acknowledgement that someone paused this scene, stopped the monologue, or took a break between chapters.  It will take days of hard pressing to reduce the dog-eared indentation.  For me, the damage is permanent.  Literal and literary disrespect has occurred, something a book can never shake off.  So I use anything I can get my hands on to save my places.  I have bookmarks, but they never seem to find their way into the books.  In bookstore checkout lines I see pretty little bookmarks with clichÃ©d phrases and artsy black and white pictures and salivate.  But my Pavlovian reflexes are always pushed aside because I know that the $3 bookmark will sit in a drawer or beside a stack of books and never actually make it into one.  There is no real reason for this.  When I need to hold my place I franticly search for something flat or something related to the process of reading.  Beside my bed there are a number of books \"held\" by varying objects.  The most frequently read book has my glasses holding court, the short story book that I dip in and out of is splayed, open side down on a table, a book of memoirs is conversing with my itty bitty book lightâ„¢, a travel book on London is held in place with the thin, red, satin ribbon attached to it, and another short story book has the receipt for itself tucked inside.   The receipt is my favorite placeholder.  I think this is partly sentimental in that every time I open the book I look at it and remember the store and the circumstances for the purchase.  Partly I\'m sure I do this to subconsciously remind myself to get every bit of escapism and pleasure out of the $10.95 I spent on the paperback edition.  And partly because it\'s convenient and perfect.  A receipt takes up no space between the pages, it doesn\'t ask you to clip it on, it just slides right in like it was meant to be there and it never leaves a mark.  It\'s the answer to the damaging dog-ear.  It\'s the lazy womanâ€™s pause button.  It\'s quite possibly the very best un-bookmark.  Yesterday I was digging through my assortment of not-yet-read books and discovered an odd bookmark tucked inside Carson McCullers\' THE HEART IS A LONELY HUNTER.  I\'m sure at the time, my former therapist\'s business/ appointment card was the closest thing around when I stopped reading.  I can\'t imagine that I just randomly put this card in this book.  It was slipped between pages 4 and 5.  A bookmark placed so early in a book would make me think that I never really began reading it.  Maybe I went for a teaser from the first couple of pages.  I have a terrible habit of beginning books and then about 20-50 pages into it, I pick up another book (see the bedside assortment listed above).  This is one of my worst habits and one I really hate. Commitment phobia rears it\'s ugly head in some many places in my life.  I just can\'t help but think there is a small, or rather quite large, message in THAT card in THAT book.  Sometimes I wonder...  '),(11483,3022585,'01,August,2004','2004-08-01 00:00:00','   I have seasoned chicken in the oven right now.  The smell is amazing and began wafting from the kitchen about 10 minutes after I put it in the 350Âº box.  The smell is of earthy paprika, pungent garlic, and sweet basil.  It smells like home, like a weekday dinnertime when a family\'s house is quiet after a day of harried activities and mouths are watering.   My father gave me this container which describes its contents as \"The Worldâ€™s Best Chicken Rub.\"  He bought it at Food Lion, a place I never go because Ukrop\'s has spoiled me and I am OK with the overpriced ingredients for my cooking, or lack there of.  I don\'t even know where a Food Lion is around here.  I\'m sure I\'ve passed one on my travels trough town, but I couldn\'t tell you where.  The seasoning gift was unexpected and handed to me in the Food Lion bag with the receipt still inside.  Two containers where purchased, one for me and for my brothers family.  I think I was the afterthought, but my dear sweet dad didn\'t want to leave me out.     The seasoning is described as a \"rub\" and the directions tell you to brush the seasoning mixed with oil and lemon juice over the chicken.  In my kitchen, the brush consists of a plastic bag and there is no lemon juice.  A poor single girl\'s kitchen is usually devoid of fancy cookery, i.e. brushes or fresh citrus.  I actually do own a cooking brush, however, I believe I bought it at the dollar store and then only in the hopes that I would someday make delicate puff pastries brushed with butter.  I have made none so far.    The brush did make an appearance two weeks ago on Easter weekend when my visiting mother sort of decided I would cook a ham for the family on Saturday night.  I didn\'t know until I was cooking it that this would be my job.  I\'ve never cooked a ham before.  Luckily, store-bought spiral sliced hams are pre-cooked, you just have to know how to heat them up and glaze them.  The glazing scared me and when I asked my mom how to do this she simply said \"brown sugar and mustard.\"  \"OK,\" I thought \"but how the hell much of either one?\"  I couldnâ€™t ask her again, she was busy with my little nieces and there was an air of \"Have you learned nothing from me?\" in her voice.  I looked in the cupboard and found a clump of petrified brown sugar and of course there is always mustard in the fridge.  I remembered that brown sugar comes back to life with heat so I set the stove to high flame, put the brown sugar and a splash of water in a small pot, squirted about 6 seconds worth of mustard and stirred.  The resulting taste test was good; it tasted familiar, like sweet Easter ham.    Two hours into the ham warm-up I pulled it out and brushed on my brown sugary mixture.  The glaze was perfect, sweet and tart and just right for a first time try.  My parents let me keep the leftovers because I\'m poor and because they live 3 hours away.  I said thanks, that I\'d eat on it all week.  I didn\'t touch it after they left, despite my glaze of glory.  The first few months I lived on my own with no dining hall or catering company to feed me (college dorms and sorority respectively) I bought more food than 3 people could eat in 2 months almost each week.  I researched and concocted difficult recipes from Martha Stewart and Junior League cookbooks.  I even spent 3 hours on a fancy dinner for a going away party for a friend once.  (I\'ve since learned that friends will come over for your grilled cheese as much for your \"whole roasted chicken with goat cheese and sage.\")  Now, I\'ve settled into the typical single girl\'s fridge.  I have a door full of condiments waiting for a salad or bread and meat or even fries, but I have none of those things actually in the fridge to put them upon.   In college, my fridge was stocked for weekends and nights I just didn\'t want to leave the apartment to go eat with my \"sisters.\"  One roommate and I would try to make what we thought were balanced meals.  We would go grocery shopping and come home to heat up canned beans, boil a bag of rice, and bake chicken with rosemary in tinfoil for 45 mins.  We thought our moms would be proud.  We also bought an entire red velvet cake once and after 2 days gave up the polite slicing and plating to simply spear it with a fork as we stood at the kitchen counter in our pajamas and bare feet and chatted about how we had really sunk to a new low.  During the day today I had a brilliant thought of going to the store after work to pick up a fresh lemon to squeeze in the bag with the chicken and over some freshly cooked rice.  I never made it to the store and the poor herbed chicken has no side dishes, no friends to share the stomach with.  A single girl just doesn\'t have the time or the energy really for all the fuss, and especially all those dishes.  There will probably be no leftovers either (what with the 1 piece of chicken I fixed), my fridge will go another day (or 10) with not much more than fancy mustard and light Italian dressing.  But, right now my dinner\'s calling me.  '),(11484,3022585,'01,August,2004','2004-08-01 00:00:00','   Miraculously my blessed computer is now typing.  I have no idea why it decided to go bananas and turn the letters \"HJKL andM\" into \"1234 and 5.\"  I took it to school today and I think maybe my baby didn\'t like being away from home.  I wasn\'t that upset as my computer has been showing signs of trouble for a few months and I expect that everyday may be it\'s last day.  I have an order to send it in to Apple but am terrified that they will wipe out the memory and start over with all the programs.  This makes me want to run away with my sweet sick computer and try to nurse it back to health.  But, I know I canâ€™t do that.  So I\'ll have to work on saving all my things like my music and documents and pictures.  This is not that hard, just dump it on a disc, but I have a habit of rereading things and going through everything and reminiscing.  What should take me about 2-3 hours will probably take me 2-3 days.  I\'m just sentimental like that.    My laptop is my baby.  In fact, that\'s its name, My Baby.  In the upper right hand corner of my Macintosh G4 Titanium is a tiny replica of my hard drive and the words \"My Baby\" underneath.  My laptop is one of my best friends.  Not like \"Oh I canâ€™t wait to go home and type on my computer.\"  But kinda.  And no, I\'m not a computer nerd.  I use it for writing and Internet and wellâ€¦ music and pictures and to watch movies and a to play a game for about an hour every 6 months or so.  It\'s like a best friend who enjoys a good flick, listens to my favorite tunes, and doesn\'t disagree with my neurotic banter, AND travels with me almost everywhere I go.   This current computer is 2 years old, but is still fabulous.  When I bought my G4, my older G3 was only 2 years old and one of my brothers thought that I was crazy to buy another one and asked which one I liked better.  I stated that it was like choosing between two children, it was like Sophie\'s Choice, the Apple Macintosh version.  If you like Macs then you LOVE them.  There are few Mac people who could go either way, PC or Mac.  I am quite obsessed with my Macs.  I\'ve had 3 so far and used pretty much every version they have come out with.  They are beautiful, efficient, and easy to use machines.  This is the first time I have ever had enough trouble with one that the friendly techies on the Apple help line couldn\'t solve.  This is the first time they have asked for me to hand over my baby.  I\'m scared, but today with the typing fiasco I came to terms with having to let go for bit and let the experts do their thing.  Then I realized that my G3 was still available.  I couldn\'t wait to get home and dig it out, to feel the tall, black keys under my fingers.  Pressing them is like pressing down on soft dough.  It makes writing fluid and effortless.  I miss that keypad.    When I opened the black chunky laptop I ran my hands over its edges.  The design team at Apple gave the G3 PowerBook curves, like a woman.  The edges are smooth and shapely; the sides of the computer make subtle \"S\" shapes.  It\'s high fashion, computer style.  I miss the friendly rainbow Apple at the base of the screen.  It goes to show that Apple did not make a big enough impact on the world market to claim the rainbow Apple without bringing along the rainbow symbol and all of its stereotypes.  Apple now goes for a sleek mono-colored, often white or grey, Apple, a more stylish replacement to its colorful twin.  The rest of my G3 is simple and elegant.  There are no extra buttons to \"go straight to the internet,\" \"open youâ€™re email,\" or anything else that junks up a PC keyboard.  There could be key functions that did those things for me of course, but I get to decide, Apple does not predetermine my preferences.  I think the thing I love most about Apple is the company\'s adherence and strive for simplicity.  Their computers have clean lines and their mice have one button, not five.  Choices can complicate life.  A simple gift, that is was Apple has given us in its computers.  PC people think that we Mac people are strange.  \"The whole world is Windows and you are using a Mac.  You poor deluded youth,\" someone once told me.  I told him he better step back and then began my rant on why Macs reign supreme.  I was showing someone a keying shortcut on a Mac (like open Apple and s for save) and explaining the ease and simplicity of Macs and she said \"Oh, they got those shortcuts from Windows.\"  I said, \"No you didn\'t\" and then explained how Bill Gates just made a PC version of the perfected Mac interface because people were sick of having to remember which drive to boot up and save on.  Deluded youth indeed.  My cousin says she would rather go without a computer than own a PC, which is what she is currently doing.  She gets her Mac on at work.    It\'s a far better place at the top of world my friends.  Come to the land of milk and Apples, we are a friendly, simple-minded, and visionary bunch.  And remember... as you sulk at us that almost every piece of music or visual media you hear or see has been touched by Apple.    *Anna is OK with being in love with her computer.  '),(11485,3022585,'01,August,2004','2004-08-01 00:00:00','   My computer is on the blitz.  When I went to type today this is what I got: 2230234550300 and that was just hitting random letter keys.  Um, OK.  Have to send it off.  I\'m scared they will erase it\'s memory and leave me with a shell of a computer and I will have to start all over.  But, I can\'t exactly do anything on it unless I develop a code language using 01234 and 5.  But, who has time for that?  '),(11486,3022585,'01,August,2004','2004-08-01 00:00:00','   Yesterday I stopped by my nephew\'s soccer practice just to say hi and show him how much I love him.  He loves to show off, but he also gets frustrated with not being perfect when he kicks the ball.  He\'s 6.  He has a lot of perfectionist disappointments ahead of him.  While I was there a mother of another child came up to talk to my sister-in-law.  She was asking who was the mother of whom and I said I was just an aunt.  She stated how nice it was that I came out to see my nephew.    During school yesterday a teacher was telling me about a girl who was behaving badly (there is a long story about how bad this teacher is and how she creates the behavior problems in her classroom, but that is for another day).  The teacher was talking about the girl and how she lives with her aunt because her own mother is in jail.  The teacher was trying to point out to the girl that she needed to straighten up because he aunt was good to her even though money was tight in her house.  I hear these stories all the time, an aunt or a grandmother or even a great aunt taking custody kids in their families.  For some reason, yesterday, it hit me that this is wonderful and strange at the same time.  Another little boy told me a few weeks ago with a smile on his face that his daddy was out of jail.  This was a follow up to the statement \"my daddy in jail\" that preceded a few weeks before.  I often don\'t know how to react to these statements and they are frequent.   Richmond is a troubled town.  The murder rate here evens out to about 10 murders a month.  That is unbelievable to me.  Crime is so prevalent that it isn\'t even a frequent topic of conversation.  The newscasters announce the death tolls almost nightly, but my friends and I don\'t worry because most murders are gang or drug related.  This makes us feel safer, but should it?  I\'ve become desensitized to so much in this town.  I\'m not shocked when teachers or kids tell me about people going into or getting out of jail or prison and I\'m not surprised by bodies being found behind school buildings or churches being burned down.    I can\'t help but think that most of my feelings about family members taking on the responsibility of other family members is cultural.  Yesterday, though, I realized when that mother at the soccer field was so impressed that an aunt came out to a soccer practice that I\'m so thankful that my nieces and nephew don\'t live with me.  I can\'t imagine taking custody of these kids because my brother was in jail (I would of course, but I can\'t imagine it).  I just can\'t even fathom it for my family and yet when a child tells me something like this I don\'t even flinch.  It seems so normal in the schools I teach in.  I realized yesterday that my desensitization is a bad thing.  It causes complacency and this will never help a child.  But, what can I say for Kindergartener whose role models are in jail that will stay with him and keep him off the streets when he grows up?  '),(11487,3022585,'01,August,2004','2004-08-01 00:00:00','   This is a quote from a graduate of my college from my graduating year,  \"I am about to finish up two year of MBA education at the Wharton School and will start work in September at McKinsey & Co in their Brussels office. Between graduation and starting work I plan to bicycle across the USA in 48 days...\"  Our five year reunion is this summer and some people have chosen to update all those who are waiting with bated breath to see what they have been up to over that last five years.  The above update caught my eye because really, as if the Wharton School wasn\'t enough this guy had to throw in his hot job in Brussels and then of course the side note about the cyclying across the USA, but that wasn\'t enough, no, then he had to add that he\'ll do it in 48 days.  Sometimes, I wish I had gone to a podunk small town college where I would send in updates like \"I just consolidated my debt and hope to be out of my misery is 5 to 10 years.\"  And my podunk fellow classmates would think \"Wow, I really wish my life were as on track as hers.  I really need to reassess my life and get my goals straight.\"  OK, well, anyway, I\'ve got to run.  My bicycle needs to ridden, around the block in 48 minutes.  Wahoowah!  '),(11488,3022585,'01,August,2004','2004-08-01 00:00:00','   Yesterday, Bunny Day, I was driving out to my sister-in-laws\' parents\' house for Easter dinner (yeah, we\'re all close like that) and noticed store after store closed.  Her parents live in the suburbs... (sorry, I just puked) and there is mega store juxtaposed to mega store juxtaposed to posh mall along the way to their house.  This is how it went down inside my head:  \"Hmm.  Why is Best Buy closed?  What time is it?  It\'s only 4pm.  Oh yeah, Easter.  But Best Buy is corrupt and charges restocking fees.  Why are they all of a sudden holy?  More like Best Butt to me.  OK, there are no cars in front of World Market, Home Depot, ah Kroger, there you go guys, sell those hams.  Target is closed.  What the?  McDonald\'s?  You mean I can\'t go get a Big Mac cause the Easter Bunny is busy.  What?  Oh, Wal-Mart.  Good, dear sweet heathen Wal-Mart.  Damn, rednecks like to shop on Easter, that parking lot is full.  Is Sam Walton rolling over in his grave?  Nordstrom, good for you.  You show them how to fear God, close you store.\"  Um, I guess I\'ve never been out on Easter Sunday to realize that we still observe this holiday by barring heathens from their purchases.  I guess I am now a heathen.  Well, good, really, I don\'t believe in judging people for their religious beliefs anyway so um, a heathen I may be, but I still hang out with non-heathens and other heathens and I love them all.    I think it says a good deal about a company to close on a religious holiday in this country, being that the majority of corporations seem to pray to the God of the U.S. Mint rather than the Almighty.  I didn\'t even go to church yesterday.  This is the first Easter that I haven\'t been to church.  I just wasn\'t down with going to Mass with the rest of my family (the only Catholic is my sister-in-law) cause they don\'t think I\'m Christian enough to drink their wine (Catholics do it right, there is no grape juice up in their joints).  So, I slept and joined the fam. for the important part, the food.  OK, so this is all blasphemous of me to say.  But, I probably would have gone shopping yesterday being none the wiser.  I had no idea we were that pious a nation what with Blue Laws being banished within the last 20 years.  Alright, well, you go people, stand up for something, except of course Kroger and Wal-Mart, you guys keep selling on major holidays cause you never know, I might need to buy some ice or a fishing pole at the last minute.  '),(11489,3022585,'01,August,2004','2004-08-01 00:00:00','   This past week was my Spring Break from school and I spent most of it tidying up loose ends and looking for a condo (not successfully).  All in all, not an exciting week considering that this time last year I was getting over my jet lag from London.   Last year that I had a longing to revisit my favorite city in all the world.  After my first tiring and horrendous year of teaching, I was planning on going backpacking through Europe with a girlfriend of mine, but she backed out.  Then, I thought I would spend a summer in London as a reward for getting through the fire, but I failed to ever find out how to rent a flat for a reasonable fee for a few months.  So, two years later, after 9/11 and with a pending war in Iraq, I decided that I would go to London, by myself.  There is a time when you realize waiting around for other people isn\'t worth it.    Luckily, British Airways had cheap tickets and AAA travel helped secure a great cheap hotel and off I went.  My family was not pleased AT ALL.  My brothers were convinced I would be blown up mid flight, my mother thought I would be killed on the streets, but I (and my therapist) thought why live in fear?  I almost backed out though, because at the time the war in Iraq was only a month old and I did get scared.  But I knew I had to get through the bit of fear for many reasons.  So, here is what got my on the plane; I thought, sure I can stay here and be safe and not take the risk and it will drive me crazy because the plane on which I have a reservation will be flying with or without me so I can be on it and be in London or sit here for a week and wish that I had gotten on it, because after all the worrying it didn\'t blow up.  So, I got on the plane and I loved every minute of it.    It was wonderful and fabulous and there are not enough cliched things to say.  I saw 4 West End shows and revisited my favorite spots and went to museums and shopped and loved it.  A good many people think that it was ballsy of me to go and to go alone.  I\'ve never had that feeling about that trip.  My parents instilled great traveling skills in me and traveling just feels like breathing sometimes.  I love it and I must do it.  I think I am my best self when I am traveling.  I\'m an easy traveler.  I don\'t stress about what the day will bring or getting through an itinerary because the best things happen when you don\'t plan.  There are too many stories to back up that statement, but the places and cities at the moment should be the tour guides, not a piece of paper or a set schedule.    I missed the traveling this break.  I had planned to treat Richmond as if I were a tourist this week and go to the historical places and museums that I\'ve been meaning to go to, but the week slipped away from me and now I have a paper due.  Sometimes, the real world is not quite a nice as the magical, mystical, travel filled one inside my head.  Oh well, there are always books to take me there.  I\'m currently (and just now) reading CAN YOU KEEP A SECRET by Sophie Kinsella, a British novelist.  I bought the book last year in Harrods of all places because I knew that it was not in the US yet and I enjoy her books.  They are the mindless Chick Lit type that a girl just needs sometimes (don\'t worry I also bought some Seamus Heaney).  So, I hadn\'t read it until now because I\'ve been trying to read a little more high-brow (though not much more- maybe just a brow lift).  So, I got it out the other night and read a few chapters each day.  Then, in B&N the other day I saw it is out in hardback as a new release.  Funny, how it\'s not new to me at all and yet it is new, now.     Last year I took Kinsella\'s book SHOPOHOLIC TIES THE KNOT to London as I had just begun to read it before the trip.  It\'s about a girl who loves to shop (obviously) who is British, but lives in NYC, but goes back to England to get married (again, obvious).  I would take it on the Tube to read between stops because I also feel that when traveling, you don\'t need to always act like a tourist (do what the locals do, read on the Tube).  So, I was on the Circle line train going to Leicester Square by way of Embankment reading and reading and being nonchalant. The funny thing is that I was reading about the main character visiting her friend in London and going out for tea in Sloane Square when the train stopped and the doors opened and I heard the announcement of Sloane Square station.  Sometimes, the universe just aligns.  And, no, I don\'t believe in that crap, but sometimes... Sloane Square and Sloane Square meet and I just can\'t compete.  '),(11490,3022585,'01,August,2004','2004-08-01 00:00:00','   And by the devil I mean Uncle Sam.  I now have to sell my soul and get a second job.  Turns out, I do owe taxes (see 4/3/2004 post).  I went back today to pick up the papers that had been sent to the processing center and there it was in black and white tiny type \"You owe (arrow) {a shitload}.\"  Thanks Uncle Sam, thanks.  The reason why I owe in the first place is because I tutored some kids with Autism last year.  And, being the honest taxpayer that I am, I reported all income and expenses and now I owe.  That sucks.  Honesty, sometimes, not the best policy for my I-really-want-to-go-shopping-but-now-I-have-to-get-another-job self.  \"O where o where has my little life gone?\"  Over the hill to \"Do you need any help or are you just looking?\"  I\'m thinking Barnes and Noble again.  I love that place.  OK, so most of me wishes that small independent bookshops were the norm rather than giant corporate monsters, but when I walk in there and there are a million books on any subject and the Starbucks and the prettiness, oh the prettiness of it all.  I love that place.  I have to work there, not because I love it, but because I have to, to pay the damn taxes and thereby accrue more taxes and then next year get another job and, oh god, will the cycle ever end?  Fraid not.  '),(11491,3022585,'01,August,2004','2004-08-01 00:00:00','   I hate it when I settle into a quiet night of phone calls to my friends and not one of my dear friends is home.  There is nothing like a warm-hearted answering machine to make a girl feel loved.  I mean my friends are fabulous.  They are kind and funny and fun and far away.  Distance from my good friends hurts sometimes and so when I call them out of an act of loneliness and they are not at home, the loneliness gets deeper.  Their fabulousness outshines my fabulousness to nothing and I go back to what I was doing before the desperate plea for conversation.    Tonight, I got 6 messages and 1 number out of service.  The latter was a friend of a friend that I\'ve been meaning to have over for dinner, but come on people.  Then I think about why I wanted to talk anyway.  So much of the time I am just ranting on in my neurotic self-obsessed way and what is in it for them?  I hope that I am a good friend, but I know that have my bad moments.  One particular friend is excellent at getting me to divulge more and more information.  She is a question asker and she is a lawyer.  She asks for more and more info. to the point that I have to stop myself and say \"OK, so whatâ€™s up with you?\" about 45 mins. into our phone calls.  She is great, but she also feeds my self-obsession.  She\'s also taught me the act the listening because she does it so well.  Then there is my best friend who\'s mother is a school counselor, need I say more.  While she herself, doesn\'t have a degree in counseling, she\'s actually a science nerd, (you know I love you PK) she is fabulous at her second job.  And she does this for all her friends.  In fact, she is so good she just may pick friends based on their level of neediness. Hmmmmm!   The immediate lonliness was brought on by my re-watching of my \"Sex And The City\" DVDs.  I just really miss my girlfriends sometimes.  I miss the closeness that women can have in true friendships when you live in close proximity.  My best friend lives across the country and my other dear friends live 2-24 driving hours away.  Oh, the mark of having successful, adventurous, independent women friends!    So, I sit down and write and try to work out whatever brought me to the phone in the first place.  I think about what my friends are doing right now that takes them away from their phones.  Are they out with other friends?  Are they watching a movie?  Are they discussing recently read books?  Are they organizing fundraisers?  Are they laughing?  Are they happy?  I hope they are laughing.  I hope they are happy.  They are my friends and I love them.  '),(11492,3022585,'01,August,2004','2004-08-01 00:00:00','   On Monday I started the possibility of the process of buying my own place.  I just met with a realtor, that\'s all, no paper-signing going on, no commitments, yet.  It\'s a scary thing this idea of home ownership, signing away your life for 30 years.  It\'s a good idea, financially, but mentally, is this good for me?    I have a few single girlfriends who have bought places.  Two friends in Richmond bought houses.  I think they have balls.  Not that home ownership is that big of a deal, but a house is a big deal.  I personally can\'t go that big.  The mowing, the pruning, roof replacement, window washing, it\'s too much.  I can barely handle keeping my 6-room apartment in orderly shape.  I can just imagine the letters from neighbors that I would get concerning the jungle that would certainly grow outside my home.  So a condo, a condo sounds good to me.  It\'s like an apartment, but I\'d get to paint the walls and a tax deduction.    Condo searching is hard.  While I\'m still in the beginning stages, I\'ve already seen 3 in person and a bunch online.  Nothing has caught my eye.  I think I am waiting for the perfect one to just fall into my lap.  I\'m picky, but open-minded.  I\'ve told my realtor that I\'m interested in city properties and no new construction.  I want an old, lived in, worn and wise place.  My very nice, my age, trying to sell me a place realtor took notes.  I checked for a ring. There was one.  What does it say about me when the men I am not attracted to are married?  Does that mean that somewhere out there there is someone who would love my winning personality and bright smile?  Oh, this is not blog about me wanting a relationship.  This is a blog about me wanting, something, a little more of myself, I guess.  There are friends who say, \"You will find him when you least expect it.\"  A good friend who is getting married this summer tells every single girl who claims they are OK with being single that they should watch what they say because she felt the same way last year and poof she\'s getting married.  Well, with my relationship phobias, I haven\'t been expecting it for um, 27 years so, um where is the unexpected?  OK, I\'m still not ready for all that commitment, so am I ready to commit to a condo?  We will surely see.  '),(11493,3022585,'01,August,2004','2004-08-01 00:00:00','   There is the old cliche that music can change your mood in an instant.  Oftentimes, I am the cliche.  Driving in my car a song on the radio will put in the mood for shopping, reading a book, or dancing within the span of 8 minutes.  The emotional hills and valleys that I go through listening to music come up as quickly as the mileposts on the highway.  On Saturdays, when there is work to be done in my apartment, I open the blinds or windows, turn up the radio on an upbeat (usually hip hop) station and get busy.  Choosing one CD to listen during one period of productive time is more difficult.    I have a habit of listening to an album until it is worn out and tired.  I get a new album and listen to it over and over.  I\'ve been known to put certain songs or even the whole CD on repeat for hours.  Because of this, I know every word to albums of bands like U2, Sarah Mclachlan, James, Harry Connick Jr., Billy Brag and Wilco, Michelle Shocked, David Wilcox, Jimmy Buffet, and show tunes like Les Mis and Phantom.  So much useless information floating around my brain taking up space for real knowledge like whom all of my government representatives are and how they vote in their various arenas.  I know that this mind hoarding of information happens to everybody to some degree, but not only do lyrics hide in my long-term memory, but with those lyrics and melodies are visual, olfactory, and textural memories.  Songs, and especially whole albums can bring back days, events, and sometimes-whole years.  Like James \"Laid\" CD evokes the memory or Camp Morehead where I was a (non-sailing) counselor for one summer in high school.  I had brought the only CD player for the cabin so I forced the 12-13 year old girls to listen to my brooding British music.  Camp Morehead was dismantled a few years ago and the land sold to people building their dream second homes on the shores of many peoples childhood summers.  I actually hated camp, but every once in a while when I listen to \"Laid\" I wish for the slow breeze coming off the Bogue Sound and all the little SunFish sailing in the wind.  There is another CD in my collection that I bought in the Dominican Republic on spring break my last year in college.  A resident of the house my friend and I stayed in played a native Dominican\'s CD everyday so it became the soundtrack to our visit.  The music is poorly produced and there are barking dogs used as part of the beat on one song, but that CD is probably the best souvenir I\'ve ever gotten from a trip.  Sometimes, in summer, I\'ll put it on in my car; roll all the windows down and just drive, thinking of the Dominican.  There are many more CDs and even more recollections.  And so all these memories attached to my music cause the choosing of music to listen to at a particular time quite difficult.  There are rows and rows of CDs in my collection that I just can\'t bring myself to play because friends and trips and times in my life are gone and can\'t be reclaimed.  Nostalgia is sad state for me and one I only visit when I really miss someone or someplace.  And I\'m so thankful when sometimes; somehow with new places and good friends I have the opportunity to attach new memories to the music.  '),(11494,3022585,'01,August,2004','2004-08-01 00:00:00','   that all the world should be taxed.  And so I went to a local tax place that was giving free tax service to teachers (gotta love those perks).  It took two nights because I had to wait the first night for another customer.  As I waited an older man who is a retired accountant and who\'s son owns the business began to organize my papers.  He was friendly and we talked for at least an hour.  Of course, he knew a lot about me because my life was displayed in numbers before him, but he also told me about his ex-wife who left him six months ago and all the improvements he had personally done on his house that he is now trying to sell.  While he was very kind and a great listener, this was all too much information, but I listened and nodded because he was helping me with finances.  When I left the first night after getting some of my information entered, I owed a boatload of taxes.  I was not happy and immediately began thinking of second jobs.  Barnes and Noble has been on my list for a while, but do I really want to work for a corporate monster at the lowest level?  OK, I\'m a teacher, I really don\'t have much of a choice.  The next night was more promising, well financially at least.  When I walked in, the older man welcomed me and asked some more questions about my paperwork and then said he found a mistake the previous person had made that would save my taxes and actually grant me a return, a small one, but at least it was on the positive side.  I was ecstatic.  The clouds lifted and I would not have to sell my soul to pay my taxes.  He was quite pleased with himself as well and mentioned that he liked to see me smile more than three times.  I obliged him the smiles because he was, of course, saving my taxpaying ass.  This was a small outfit so all paperwork would be sent to a processing center to be checked and then I would have to come back to pick it up.  When I was ready to leave I went to shake his hand.  He held my hand well after the thank-you-for-your-help handshake and said \"The good news is that I get to see you one more time.\"  I gave a quiet-I-don\'t-know-what-that-means giggle and turned and walked out.  As I left the store a scary thought went through my head, \"Did that old man just hit on me?\"  He is not grandfather, really sweet older.  He\'s the just retired, wife just left me older.  Um, no.  I then began to think back to the conversations we had had over the past two nights.  Did he think there was more than accountant/ taxpayer information being past?    I have come to realize that I am clueless to men\'s flirting techniques, even when they are obvious.  This usually happens when there is no interest on my part and therefore why would I be looking for signs?  There was once the college friend who all of sudden told me he loved me.  I stopped being friends with him, pronto (I know, I\'m a bitch, but he ended up gay so um, I think I was right).  There was a guy at a bar once who told me he had been staring at me and trying to pick me up for 1/2 hour.  I had no idea and I\'m glad actually, he ended being quite cute and oh-so strange.  Then there was a guy at weekend long party who followed me around the whole first night until I realized after he told me that I was the kind of girl he wanted to marry that I really needed to just be around my girlfriends.  I hate those clueless girls, the ones who have guys swarming around them while they are supposedly oblivious.  I guess, sometimes we all are That Girl.  Regardless if the old tax-man was getting flirty or not, I can\'t help but feel that my tax refund is now just dirty money.  '),(11495,3022585,'01,August,2004','2004-08-01 00:00:00','   here\'s me trying to add linky links in the text for the first time.    urlLink buy an apple, it will save your life ,  urlLink good news ,  urlLink good TV ,  urlLink better TV ,  urlLink the best music ,   ok, i\'m done with the practicing.  have fun with the linking.  '),(11496,3022585,'01,August,2004','2004-08-01 00:00:00','   So, the other day I was flipping through my sitemeter stats and noticed that someone came to my blog through google.  I clicked on the link to see what subject they entered to get here.  I like to see what other sites pop up along with mine.  Well, some perv entered \"pee together\" into the google search engine.  My site came up becuase of my innocent story about my cousin and I when we were little.  I was totally grossed out.  People are disgusting.  I need a shower.  '),(11497,3022585,'01,August,2004','2004-08-01 00:00:00','   I just need a hug.  No, no, the biggest hug you have actually.  A giant teddy bear sized hug that envelopes me and cradles me and maybe even rocks me gently back and fourth.  A big hug, that\'s all.  Well, and maybe the answer to the question of \"Why?\"  '),(11498,3022585,'01,August,2004','2004-08-01 00:00:00','   The work â€œdorkâ€ refers to the male sexual organ, â€œGeekâ€ was the name of a circus performer who bit the head off of chickens, and â€œjerkâ€ refers to the jerking motions that a person with severe cognitive or physical disabilities can display when they lose control of their muscles.  Sometimes, the words in our verbal lexicon are no so innocent.  Iâ€™m just sayinâ€™.  '),(11499,3022585,'01,August,2004','2004-08-01 00:00:00','   â€œI gotta go,â€ I squealed.  I was 4 years old, skinny with dark hair, pale skin and big brown eyes.  â€œI gotta go too,â€ said B.  She was 5 years old, skinny with blond hair and brown eyes.  We were beautiful together, the little girl version of yin and yang.  â€œLetâ€™s go together,â€ one of us said.  So we raced to the upstairs bathroom, got ready and sat down together to pee.  We were giddy.  This was silly.  We were peeing together, our tiny bottoms sharing one toilet.  Our mothers thought it was precious.  This was our cousin thing, a little girl bonding time.  One mother grabbed a camera and flashed our picture, B. with her long blond hair and legs together facing the bathtub, me with my long brown hair and legs together facing the wall.  We are squeezed in, our hands holding on to the tips of the oval seat so we wouldnâ€™t fall in or off.  Our clothes are simple knits, me in white, B. in dark blue; our shorts stuck to our legs have appliquÃ©.  We are gorgeous.  We are smiling.  We are young.  We are silly.  We are happy.  It is one of my favorite pictures, B. and I concocting a simple plan to kill two birds with one stone.  Iâ€™m so glad there is a picture of that innocent time.  Sometimes B. and I will talk about our childhoods, how we grew up so differently yet share so much in common.  She is my blond size 2 cousin.  I am not her.  She is a modified hippie.  I am an ex-sorority girl.  Her mom is young and the least neurotic of her sisters.  My mom is not.  But we both have a factor in our blood that causes blood clots.  So do our mothers.  We are so different and so similar.  I am proud of her, that she is a brilliant photographer.  I love that she loves her animals and plants like they are her children.  Iâ€™m glad she has found a man in a fairytale way, which we always dreamed would happen.  I love that she is strong and kind and knows more than she tells.  And I love that sometimes she will say, with laughter in her voice, â€œAnna, we use to pee together!â€  â€œI know,â€ I say, laughing back, â€œI know.â€  '),(11500,3022585,'01,August,2004','2004-08-01 00:00:00','   I  I donâ€™t remember his name, but I remember the first time I met Autism.  I was in college doing a very short practicum.  It was, in fact, my first practicum.  I was placed with a special education teacher at a local elementary school.  Her classroom served kids of varying disabilities, but mostly kids with LD or mild MR. She practiced â€œinclusionâ€ and this was my first experience with this hot topic in special education.  I observed her sitting at the back of a classroom with a small group of kids while the regular education classroom teacher led a lesson with the rest of the class.  â€œSo this is inclusion,â€ I thought.  â€œThis isnâ€™t what I had in mind, but I could do this.  OK, I get it.â€  But, I would do more than observe over my 3 short visits to that school.  I asked the special education teacher if I could help her.  I wanted to get in there with the kids.  They were, after all, why I was there.  She asked me to sit with a little boy and read with him to keep him occupied for a short while.  His social skills were strange to me.  He came to me when I asked, but never looked at my face.  His eyes darted quickly from on thing to another, mostly at floor or knee level.  He sat down with me on a couch in a corner of the classroom.  He was the first to open the book and he just took off reading before I could tell him what we were going to be doing.  I had no time to ask him about the title, the author, or the pictures.  I was amazed that he could read the book verbatim to me.  He was in 1st or 2nd grade and I knew his true reading skills surpassed his supposed ability.  I asked him questions as we read the story.  I wanted him to explain or expand on the pictures in the book.  His answers were â€œyesâ€ or â€œnoâ€ if anything.  Mostly he ignored me.  It seemed he had no idea what we had shared together.  I remarked to his teacher later that I was very impressed with his reading, but wondered about why he couldnâ€™t or didnâ€™t answer my questions.  The teacher was busy and couldnâ€™t quench my inquisitions.    I was so uninformed and inexperienced with Autism.  That teacher couldnâ€™t have told me the answer to my questions.  Like Autism itself, the answers were complex.  II Mathew was four years old and on the flyer he had an angelic face with blond ringlets framing big blue eyes.  I answered the ad for a home-based teacher and began doing ABA therapy with Mathew in his home for 6 hrs. a week.  It took me only a few days to fall in love with him.  â€œThis is good work,â€ I thought.  â€œThis is what I was meant to do.â€  I had another job that year in college working in the merchandise department for a famous band.  While I took phone orders and packed t-shirts all I could think about was Mathew and how the work with him was far more important than satisfying the sticker and CD wishes of teenagers.  I was anxious to stop working for the band and when summer ended I split my time between classes and Mathew.  I learned ABA well enough to do it in my sleep.  â€œWhat number? What letter?, Good job!, No, try again.â€ ran through my head, my dreams were flooded with discrete trials.    I worried about Mathew and whether I was really being effective.  I explained to friends and family that my work with Mathew was tedious.  I tried to put what we were doing into everyday language.  I said â€œWeâ€™re trying to introduce him to our world, one thing at a time.â€  I donâ€™t think my friends really grasped the idea, their puzzled faces said more than their encouraging words.  They feigned enthusiasm for my sake.  This was becoming my passion.    After a full year with Mathew I was tired of ABA.  Our sessions had become the same, the same drills, and the same goals.  I was bored.  I canâ€™t imagine what Mathew was feeling.  He was making progress, but not enough that I could make sense of it.  I had seen one major improvement in him.  I told my parents one night during dinner on their drive through my college town that Mathew had made gains and I was so excited.  I was seeing that he was understanding what we were teaching him.  The example I gave was this:  Everyday when I see him I say â€œHi Mathew,â€ and he answers â€œHi Annaâ€ usually with a flat affect and no eye contact.  I then say, â€œWhat are you doing?â€  And there is silence.  I prompt him with â€œIâ€™m sitting, Iâ€™m playing Nintendo, Iâ€™m reading a book, Iâ€™m watching TV, Iâ€™m eatingâ€¦â€ He repeats my answer and I say â€œGood jobâ€ and we get to work.  Weâ€™ve done this a thousand times.  Itâ€™s just the way I greet him, nothing specific we are working on.  Then one day around the 7th month into my work with him I walk into his room and say my stock greeting â€œHi Mathew.â€ â€œHi Anna,â€ he says.  â€œWhat are you doing?â€  â€œIâ€™m bed,â€ he says quietly.  I stopped in mid-stride then I flew to him, picked him up, threw him in the air saying, â€œThatâ€™s right, youâ€™re ON THE BED!  Good job!  I love it!â€  I tickle him and we repeat the sequence, me shaping his words with a prompt. â€œIâ€™m ON THE bed.â€  This was a good day.  This was important.  He was understanding.  I would later learn in my education classes that Mathew had generalized the information we were teaching him.  He was â€œgetting it,â€ but I was happy for the reprieve and change of pace that student teaching brought when I had to stop working with Mathew.  I continued to baby-sit and stay close with the family, but this was my last experience with hard-core ABA.  It was the foundation for my obsession with Autism.  It was a good learning experience.  It was important.  III  The first day of my student teaching I knew things would be very different for me.  This was a TEACCH classroom.  What I walked into was a mess, a scattered and colorful maze of bookshelves and desks.  The teacher and TAs were just putting everything together after a summer off.  There was a relaxed feel and yet there was so much to do. The teacher was a graying hippie who wore overalls and Birkenstocks, drove a red pickup truck and talked about her horse, Nelly.  Sheâ€™d developed the classroom and program with a fellow teacher.  This was her baby.  She knew Autism backwards and forwards.  She was wise.  I knew Iâ€™d learn from her.  And I did.  The first day the kids arrived, 8 of them in all, I stood and watched and followed, mostly with my eyes and mouth wide open.  These kids talked in full sentences without prompting, the cared for their own belongings, they followed their own schedules, they played with typical peers, they asked for help, they were toilet trained.  â€œThis is not Autism,â€ is all I could think.  I talked with the teacher that first school day when all the kids had left.  I had a million questions about each child.  â€œWhy is Ethan so lethargic?  If Sadie didnâ€™t talk when you first met her, how did she learn to talk?  Will Kathleen ever read?â€  I knew I had to get used to fact that Mathew was not Autism and Autism was not Mathew.  That classroom proved the spectrum for me in terms of kids, in terms of methodologies.  Autism is a rainbow of colors, I got it.  IV Anthonyâ€™s parents came to school before school even began.  We had to amend his IEP.  He was switching schools to my newly formed Autism classroom and he would be held back in kindergarten for one more try.  His mother and father were originally from India.  They spoke English, but I still found myself speaking loudly and repeating things in simple phrases.  I needed to feel that they understood me because I couldnâ€™t understand them.  Anthonyâ€™s first day in my classroom for kids with Autism was my first day too.  We both had no idea what we were in for.  Anthony would spend half a day in kindergarten sharing a TA with another student and half a day with me working academics and adaptive skills.  Anthony worked hard and was rewarded with playtime.  He spent a lot of him playtime having figures interact or simply talking gibberish to himself.  He knew English, but would only speak when prompted.  His parent assured us that they only spoke English at home.  This was important for him to learn one language given his challenges.  His parents were adamant and I believed them, they didnâ€™t understand him very much either.  He made no eye contact unless I cupped my hands around his big beautiful brown eyes and put him face to mine.  Anthony was a puzzle and he worried me.  Then, we began to acknowledge his sounds and anytime he used an English word appropriately we threw a small party.  Soon Anthony was throwing his own parties.  He would show us his artwork enthusiastically exclaiming, â€œLook.  Oh itâ€™s beautiful.  Good job Anthony!â€  When he would answer a question he would automatically congratulate himself, â€œYes, good job!â€  We would agree with him and then laugh with him.  We knew he was repeating us, but we couldnâ€™t have paid for that pride or that growth.  We took it and embraced it, Anthony was awakening.  Slowly and surely Anthony made strides.  He quickly learned to read alongside his peers.  He was counting and making patterns.  His comprehension was always a bit of a challenge, but by 3rd grade he was working on grade level and fully included in his regular education classroom.  He even had friends.  Anthony was an exception.  He was bright and his parents worked with him constantly.  He is only one of many success stories, but his is one I think of often in terms of extremes.  His speech pathologist would often say to me when I was discouraged,  â€œAnna, when we tried to evaluate him his first year in kindergarten he was a mess.  He would sit in his classroom in the corner, rocking and saying gibberish.â€    Turns out, there was no gibberish there, just words that needed shaping and a little boy who needed lots and lots of encouragement and laughter.  We happily gave it to him.   V  Autism is a neurological disorder that stems from deficits in social and communication skills and coincides with repetitive behaviors.  I often speak of Autismâ€™s umbrella, how ADD and OCD and MR can all be part of a childâ€™s diagnosis, but I always feel I fall short when explaining the true nature of the disorder.  This is hard for me because I canâ€™t possibly relay all the stories and images that had led to my understanding and yet without those stories and images my understanding is weak.  I want people to automatically see what I mean when I try to describe how this disorder can present itself and I am so careful that I donâ€™t leave a Rainman image, a one man show of a multi-faced disability.  I think I always fail; Rainman is more visual than my words can ever be.  And this too is what frustrates me when trying to explain my philosophy in educating people with Autism.  Contrary to my beginning with this disorder, I believe in a highly eclectic approach filled with speech therapy, occupational therapy, sensory integration, floortime, inclusion, schedules, behavior plans, varied opportunities, generalization, varied teaching materials and methods, all under a child centered, extremely structured environment.  And above all I believe in spontaneity with these students.  How can you be spontaneous in a highly structured environment?  Easy, make sure the structure allows for freedom and the kids will blossom.  '),(11501,3022585,'01,August,2004','2004-08-01 00:00:00','   I see the skyline of New York and my stomach does flips.  I want that city.  I crave that city.  When I unexpectedly come across a picture of New Yorkâ€™s skyline, I take in the scene and then have to look away.  I canâ€™t look too long because the stomach flip will turn into an aching.  So, I move on and begin to plan a trip back to Gotham.   Iâ€™ve only visited on short trips.  Sometimes my visits were so brief they were merely a tease.  A one-nighter for a private concert, a one-night lay over to the Dominican, and a lunch on Spring Street in January with snow falling while passing through town.  No matter the time Iâ€™ve spent in NYC, the city is always with me.    This past December I told my friend shortly after we emerged from Penn Station that I just felt at home in the city.  There is so much life in New York and this is what I love must about it.  Of course there is life everywhere, but in NYC it is at its most vibrant.  I love that there are people crashing into you on the sidewalks.  I love that you can hear the streets and sirens at 4am on the 25thfloor of an apartment building.  I love the subways and buses carrying strangers to a million different places.  I love that there is a giant patch of green in the middle of the chaos to offer a reprieve and that inside that escape the skyline still impresses its buildings, gesturing to come back into the fray.  I consume as much media related to New York as I can.  I watch every documentary I find, rent movies with that take place in New York, read books with NYC as the setting, and read blogs of NYC dwellers.  Obsession would be a good word for it, but my fascination is more of a hobby.  I think about moving there and then life goes on and I forget about it or convince myself that itâ€™s not really what I want, that I need to cultivate the roots here and stay near my family whom I love.  But, when I see a picture of New Yorkâ€™s skyline, my stomach tells me to go back to the obsessing, to get online and find a job and an apartment in the city, to forget the ties in Virginia and to make that leap.  I know, I know.  Why not?  Well, I have no answer.  All I can say is give me a year, weâ€™ll see.  '),(11502,3022585,'01,August,2004','2004-08-01 00:00:00','   the fever is breaking which is a good thing, but i would rather be cold than sweating.  thanks to my friend, j., in texas who urged me to go out and get sick food.  i was so hot, but my belly is full and that is a good thing, however eating salad while breathing through your mouth is a hard thing to do and not choke.  i\'m just sayin\'.  i\'m done with the sick updates, go back to your lives people, i\'ll be blowing my nose and turning up my AC.  '),(11503,3022585,'01,August,2004','2004-08-01 00:00:00','   been under a down comforter for an hour and a half and still freezing.  body hurts, fever at 100.something.  want. my. mom. and. gingerale and orange sherbert float.  '),(11504,3022585,'01,August,2004','2004-08-01 00:00:00','   I hate having a cold, especially when the weather is warm.  My head is clogged and nose is stuffed.  I just want my bed.  HMMMMM, my bed.  Must. Get. To. Bed.  Right after some meetings at school with parents who probably won\'t show up and I\'ll just be sitting around twittling my thumbs and wishing for my bed.  2 of my 3 meetings scheduled for this morning already didn\'t happen.  2 more non-meetings to go, then home and bed and tv and orange juice.  '),(11505,3022585,'01,August,2004','2004-08-01 00:00:00','   Check out the new link on the right, 100 things.  and thanks.  '),(11506,3022585,'01,August,2004','2004-08-01 00:00:00','   At one point during every season, save for summer, I think that this season is my favorite time of the year.    I love the point in fall when the leaves have turned their autumnal colors and there is a crispness to the air that only happens in late September and early October.  Fall smells like new backpacks and fresh notebooks.  For me, fall signals beginnings though most of nature is ending or falling into a slumber.   Winter brings the snow that quiets and cleanses everything under its blanket, and this usually brings rest from the harried school days.  I love the cold and chilled air of winter, sleeping under a thickly feathered duvet and throwing on handmade scarves to snuggle my neck.      My favorite thing about spring are the honeysuckle flowers.  Spring sneaks up on me every year.  I grow use to the barren trees of winter and somehow donâ€™t notice the buds and tinges of green that Spring pushes through.  Then one day, all of a sudden, there are no more branches on the trees, hillsides cascade with green, houses are hidden behind rows or clumps of leaves and flowers, and shade gracefully covers the precious areas under the trees\' new dress.    But, honeysuckle.  It\'s an amazing testament of re-growth.  It grows wild along highways, alleys, schoolyards, and backyards, any place with a fence to lend climbing space.  The bushes are everywhere and when a small honeysuckle breeze blows and catches me off guard I am seven again, standing in my front yard on the left side of the house near the alley with my brother and neighborhood friends.  The bushes are full grown and you can\'t be outside without the sweetness luring you.  We would gather round the bushes picking the yellow and white flowers as quickly as we could.  We wanted as much sweetness as we could get, but we were hindered by our work.  Stealing honeysuckle juice is a delicate task.  You must be careful to pluck the whole flower from the bush.  Then gently holding the flower midway up the petals between two fingers on one hand, you pinch the bottom green mass between two fingers on the other hand and pull.  The pulling must be slow.  You must feel the tiny yellow ball at the end of the delicate stem slipping through the petals, grabbing and pushing the honey out of the small hole you\'ve made in the flower.  Then quickly move the flower and stem to your tongue a moment before you pull the stem from the flower and feel the tiny tickle of sweet, watery honey.  It\'s spring.  Your tongue has confirmed it.  Life is beginning again.  '),(11507,3022585,'01,August,2004','2004-08-01 00:00:00','   I was strolling through B&N one day and noticed that in the SAT prep section there were new materials, which were not part of my SAT study path.  I had a tinge of jealousy seeing \"SAT word\" novels on the shelf.  The titles like BUSTED and HEAD OVER HEELS accompanied cheesy cover art and promised stories of love and mystery among high school students.  These \"novels\" are written around common vocab. words found on the SAT and those words even appear in bold type within the text.  I was jealous because even though I have a BA in English Lang. and Lit. and love to read, I hated and struggled with reading throughout elementary, middle and high school.    My majoring in English was a surprise even to myself because I entered college loathing reading and writing.  I even went to the writing center on campus to get assistance with my early college papers and I would come out knowing I had much work ahead of me.  I was originally going to major in Psychology thinking that I needed this background for teaching.  But, my heart was never in Psychology and when one of the best professors I ever had gave me a C, it pushed me to make a change.    He was one of the best professors, not because of his teaching, but because of our talks during my office visits.  I went to him often for help because I was failing his class.  He became my friend and gave me guidance.  I learned about his family and his struggles in college and he helped me see that my world was not, in fact, collapsing around me.  We didn\'t talk much about the class I was taking from him, these were more like therapy sessions.  The strange thing about his role in my life that semester (and even later when I would go and visit with him) is that his specialty was in learning and cognition and he worked primarily with mice.  He wasn\'t trained for the sessions we had and when he handed me chemwipes as I cried during my first visit, I knew he would be my friend.    After a few sessions in his office, the semester was ending and it was time for the final exam.  I rocked the socks off of my final, but he still gave me a C for the course, I was one point away from a B- and passing the class.  I was heartbroken and felt betrayed. When I went to see him after winter break to discuss my grade, he didn\'t flinch although he had previously and quietly said he would help me if I made a good enough grade on the final.  I left his office quietly and quickly that day, there was no therapy session to work me through this one and I wasn\'t that upset, surprisingly.  He knew what he was doing; he knew my heart was not in Psychology.  With the threat of having to retake that Psychology class to finish the major looming in the distance, I went on with my other classes.  While studying THE EPIC OF GILAMESH for an anthropology class with a friend, I realized I had to change my major.  I loved the analysis of psychology, but I hated the science of it.  English, this is where I could find the analysis I craved without any science to confuse my brain.  I crammed every English class available into my schedule.  I stuck to mostly poetry classes as I thought this was my passion.  I think part of me really liked how short poems where.  I was still a little afraid of all that reading.  But, I couldn\'t avoid it forever and I\'m so glad that I had to read books like A RAISIN IN THE SUN, AS I LAY DYING, DROWN, NATIVE SON, CYMBELINE, THE COLOR PURPLE, and THE BLUEST EYE.    I wasn\'t the best at reading all the assigned readings.  \"Sir Gawain\" and Chaucer didn\'t excite me and there were Colonial American passages that put me to sleep, literally.  I still struggle with my love/ hate relationship with reading.  I wish I were voracious at it.  I think it\'s partly that it takes time for stories to unfold.  I wish I could finish books in one sitting like a movie, but I read slowly.  I read every word and say each one silently in my head.  I know there are speed-reading methods, but I don\'t want to skip over words, I just wish I could read every word, faster.    A few years ago, I started a book journal where I write character maps and synopses of the books I read.  It\'s an idea from my 10th grade English teacher and I find it good for me.  I have physical proof that I\'ve read a book and it jogs my memory if I can\'t remember a plot line or character details.  In the beginning of the journal I noted words I needed to look up, big words not in my daily vocab.  This lasted for about 3 books (OK just 1 really) and then the words I didn\'t know disappeared from the entries as if I\'ve mastered the English Lang.  Hardly, I just got lazy, realizing I probably would never use or see those big words again and just began to skim over them.  But, sometimes, I crave bigger words in my head and I wonder if I read \"SAT vocab.\" novels, would I cognize the context clues, ascertain the definitions and ameliorate my conversations?  I doubt it.  '),(11508,3022585,'01,August,2004','2004-08-01 00:00:00','   Tonight going to my good friend\'s house to watch the finale of friends (um finally, thank God) I was driving to the middle of nowhere.  I hate the suburbs and she and her fiance live so far out that it takes 30 minutes and $2.50 in tolls roundtrip from my apartment in town.  So driving, I saw this mini cooper with the license plate \"SUVS SUC.\"  Now, I know we guzzle gas and are ruining the ozone, but I really wanted to pull his little mini ass over and challenge him to play chicken with my SUV.  I\'d squash him.  He\'d be like a crumple soda can after slamming into my Sport (luxury) Utility (hauling shoes and pretty purses) Vehicle.  So, you mini man, watch who you\'re picking on.   And my best friend is getting married and sometimes this makes me cry.  Like tonight, it finally hit me and I cried.    PS-  I love mini coopers, being obsessed with all things British.  But, don\'t mess with my SUV!  '),(11509,3022585,'01,August,2004','2004-08-01 00:00:00','   Of course I\'ve thought about it happening.  I\'ve envisioned how it would happen and what I might do.  I\'ve considered that I should devise a plan if it happened, just to be prepared.  Well, it happened and I had no plan, no diagram to show me what to do.    This morning I woke up with a headache thinking, as I do every morning, that I just want to get back in bed, I want to call in sick.  But, I never do call in sick.  I turn on the coffee maker, get dressed, fix lunch and when the coffee is ready I begin sipping it while quietly saying a prayer that my empty stomach will except the legal stimulate and take it straight to my brain.  This morning, the coffee couldn\'t work fast enough and the headache required 2 ibuprofen.  Praying to the over-the-counter drug gods that the medicine too would go straight to my brain, I began looking for my keys.  I couldn\'t find them and my head hurt too much to search really hard.  So I grabbed the spare set with it\'s spare car key, spare car remote clicker thing, and spare apartment key.  The clicker thing was the slightly used one I had traded for the unused one a few months ago thinking that the battery in it had run itâ€™s course.  It\'s a sad little keychain with only the three danglers, but it would have to do, I was running late.    It was a wet morning and this time I remembered my spare umbrella in case the heaven\'s opened up on me as I went from school to school.  I stepped out of my apartment, coffee in hand and as I approached my car, the spare clicker wasn\'t working.  I figured its tiny battery was in fact dead and opened the driver\'s side door with my key.  I then tried to manually unlock the rear doors by using the power unlock button on the driver side door (I know, not really manually).  Nothing happened.  I was confused.  I hopped in, put my bags on the passenger seat, set my coffee in its holder and put the key in the ignition.  Upon turning it, I got nothing.  I mean nothing, not even the usual piddily sound the altenater makes when itâ€™s trying to charge the battery when in fact the battery is dead.  All signs were pointing to a dead battery, but no noise, that was strange.  I was confused and then remembered how the ignition had been acting funny the last few days.    Here is where I needed the plan and had none.  I first said a quick thank you to God that I was a support teacher and did not have a classroom full of kids waiting on me and then a bit of panic set in.  What do you when you are almost late for work as it is and everyone you know is probably at work already.  I called my brother, he had to take the kids to school and even if he could get me to work I couldnâ€™t travel to my other schools and I\'d have to take care of the car today anyway.  So, I had to call in.  I was getting my early morning wish, but really, this was not the way I had wanted it.  Having no plan I couldnâ€™t figure out how to call in: Do I call in sick or is this personal business?  Do I call in for a half day and hope my brain clears and I figure out how to get to school for the last half of the day?  Who can help me?  How long with AAA take?    I ended up calling in for the whole day.  Car troubles have notoriously taken whole days before so I thought I would play it safe and not confuse the day even more by having to try to get to school by a certain time.  It all worked out.  The dealership was able to (expensively) fix everything by the end of the day and I can go to school easily and safely again.  I\'m just wondering, do you have a plan if it happens to you?  '),(11510,3022585,'01,August,2004','2004-08-01 00:00:00','   This past weekend I went to Winston-Salem, NC to a cousin\'s wedding, on my dad\'s side.  I am not close to this side of my family; my dad having only brothers and brothers are less likely to gather their respective families together often enough for children to form bonds.  I do see my cousin A. every Thanksgiving at our house though I don\'t really know her.  She is, however, sweet and kind and so I was happy to go to her wedding.  I was, however, happier to also spend time with my Cousin Loves A. and R. from my mother\'s side.  My mother has two sisters and while both live in different cities in North Carolina, all of our families have grown up together.  We have spent almost every holiday together, camping vacations at the beach, and weeks at Lake Norman, NC.  My cousin\'s and I (a total of 3 girls and 1 boy) haven\'t always been great friends, but we were great playmates growing up.  We were around each other so much that it seemed like we were brothers and sisters.  We played, we fought, we played again, and then we went to college.  When we grew older we became friends.  Trips to visit each other and weekly phone calls developed.  We became good friends, we were happy our mothers brought us together, we were happy for our Cousin Love.  Cousin Love is how we address each other and we do it with a southern drawl.  It ends up sounding like \"cuzzin luuuuv.\"  It is merely a way to say, \"I\'m glad you\'re in my life.  I\'m glad youâ€™re my cousin and I love you.\"    Cousin Loves have begun to multiply.  There are now wives and soon-to-be husbands and I\'m sure there will be children soon too.  We welcome the additions and affectionately extend the Cousin Love name.  Our children will continue to be cousins though growing more and more distant in a familial since.  We know they will grow up together, we hope too that they will develop innocent Cousin Love and see the joy of friendship that we have.  We have a tentative plan to save our monies and buy a beach house.  We want to name it Cousin Love, but know that the world would think us odd.  So, in Spanish it would be \"Primo Amor.\"  And in Italian, it also means \"first love\" and this sounds good to us.  So, this past Friday night, after the rehearsal dinner for A.\'s wedding I took my newly found cousin from my father\'s side over to A. and R.\'s house for some quality time.  It was fun, good drinks, good conversation.  It was like we had all been friends forever.  It was what Cousin Love is all about.  And my cousin R. from my dad\'s side blended right in.  In fact, we\'ve convinced her to come to Thanksgiving this year, to continue and expand the Cousin Love.  The Cousin Love was so good I felt full up, full up of the kind of love that lets your know you have all you need in the world.  The next night at A.\'s wedding I sat with my mom and dad, my dad\'s cousins, one of my dad\'s brothers, and my cousin R. as we watched A. and B. swish around the room.  They did their prescribed wedding things: the first dance as husband and wife, the father-daughter dance, the cake cutting, and the throwing of the garter.  As I watched them cut the cake I felt full up again.  I felt no envy of A. and B. at that moment.  There was no sense that I was missing something.  I realized then that I love my family, I love my life, and I am happy.  Full up, indeed.  '),(11511,3022585,'01,August,2004','2004-08-01 00:00:00','   How do we meet them?  When do we know they will be important in our lives?  What are the signs that certain people are meant to change you, permanently?    This weekend I found a good friend, and luckily, she is part of my family.  At our mutual cousin\'s wedding we met again after probably 15 years.  I don\'t even remember the last time I saw her but it was probably at our North Carolina family reunion, eating chicken stew cooked in a large cast iron kettle over an open flame.  Surrounded by family I didn\'t know very well and in the mist of teenage angst I guess I didn\'t realize what an amazing person R. would grow to be.  Funnily too, we lived an hour away from each other at one point in college and never made the trip over the mountains to meet.    R. currently lives in NYC pursuing her passion, poetry.  Sheâ€™s been through college, graduate school and now another masters degree for writing.  She\'s lived almost a whole life since we last talked and I guess, I have too.  We talked about NYC and my passion for that city and about poetry.  We have more in common than I can believe.  She is someone I wish had always been in my life and she was, I just didn\'t know it.  The cousin\'s wedding whom we were celebrating was brought about through similar circumstances.  A. and B. had gone to high school together, but didn\'t run in the same circles.  They both went on to college and jobs and lived in various other cities then met again a few years ago while both were back in their hometown.  Their friends say that A. and B. are perfect together, they compliment each other, and that they are deeply in love.    Why does it take years, sometimes lifetimes to realize that people who surround you are the ones that will compliment you in ways you can\'t even fathom?  It is as if we are not ready at times.  Friends, lovers, husbands, and wives may pass us along the street, but until we have fully grown into ourselves, it doesn\'t matter how often we meet.  When we are ready, hopefully they will be there.  It\'s as if we are all just ships passing in the night and I\'m so thankful when somehow we meet again and our lives begin a new direction.  '),(11512,3022585,'01,August,2004','2004-08-01 00:00:00','   My parents are here.  They are taking care of me which is nice since the tetanus shot has ruined my left arm, I got a splinter in my good foot hopping around, and all of my muscles are sore from having to be used all of a sudden in ways they are not used to.  Anyway, enough of the moaning and groaning, I should be half-walking in a few days.    My family is taking me to the beach to be waited on.  I canâ€™t go in the sun with the antibiotics that Iâ€™m on and I canâ€™t get my foot wet.  Itâ€™s all good; I donâ€™t like the sand and saltwater anyway.  So, Iâ€™m off for a week.  Iâ€™m being driven 4 hours to sit in a house and have people bring me things.  Iâ€™ll be knitting up a storm and reading and watching my little nieces and nephew play.  I wish I could play with them.  So, no blogging for a week.  Come back, Iâ€™m sure Iâ€™ll have lots to say when I return to you as long as you return to me.  '),(11513,3022585,'01,August,2004','2004-08-01 00:00:00','   And it seems I should have stayed there.  The infection is way better today, but I was really groggy from the vicodin this morning.  I had to take it last night to sleep through the pain.  Then, this morning, I went to fix some eggs and toast and just when I was going to put the food on the plate I stepped on a blade that had fallen from a cabinet that I forgot about and didnâ€™t see.  I sliced my foot wide open and now have 6 stitches.  Iâ€™ve never had stitches and I was not the best patient for the poor doctor.  This has been a rough two days.  Iâ€™m going to bed and please donâ€™t call cause I can\'t walk to get the phone.  '),(11514,3022585,'01,August,2004','2004-08-01 00:00:00','   So, I have an infection and from 6 to 10:45pm was in some SERIOUS pain.  I went to a prompt care place nearby and pretty much cried the whole time as the pain got worse and worse.  I thought about not going, about toughing it out.  I donâ€™t like to look like a hypochondriac, but after phone calls to mom and the worsening of the pain, I couldnâ€™t not go.  Iâ€™ve never seen any of the staff before and I went by myself so Iâ€™m sure the staff saw this crying girl and was like â€œOk, whatâ€™s going on there?â€  As they sent me from the bathroom to x-ray room to testing room and 2 people walked in on me undressing then dressing, I cried.  I cried around the nurse, the doctor, the x-ray tech.  I winced and moaned and squirmed like I was having a baby.  Iâ€™m not a crier and as a woman I can handle pain in the abdominal area pretty well.  I couldnâ€™t stop the tears, the pain was too much.  So, after all the testing (and crying), the doctor on call gave me a bottle of vicodin.  I think he felt sorry for me.  Iâ€™ve taken the other meds, but not the vicodin yet.  That powerful a drug scares me.  And, thanks to the beautiful power of antibiotics (we are talkinâ€™ serious ones here) and gentle Tylenol, I can walk and sit without moaning.  I love medicine.  I love doctors.  And, Iâ€™m sure if I take the vicodin, Iâ€™ll love you too.  '),(11515,3022585,'01,August,2004','2004-08-01 00:00:00','   Yesterday, the phone rang while I was brushing my teeth.  Mouth full or foam and sonic still buzzing, I run around looking for my cell phone.  When I locate it deep in the depths of my purse I donâ€™t recognize the number and think that I need to answer it quickly.   And then, like a dumbass, I do.  My muffled â€œhelloâ€ allows the person on the other end to identify himself.  While he does so Iâ€™m thinking â€œWhat do I do now?  I canâ€™t talk like this.  Do I swallow it (ew) or spit while heâ€™s on the phone.  What. The. Hell?â€  I decide on spitting and just confess it, heâ€™s already heard me sound as if Iâ€™m under water.  So I say â€œHold on, Iâ€™m brushing my teeth,â€ then go to the sink to spit, sonic turned off but still in my hand.  I apologize to him when I return and hope that he didnâ€™t hear me spitting in the sink, 3 feet away from the phone.  I realize; what a dumbass, like answering the phone at that moment was so important that it superseded the mouth full of water and paste.  What, what was I thinking?  Sometimes, I wonder.  '),(11516,3022585,'01,August,2004','2004-08-01 00:00:00','   Iâ€™ve realized it is real work being a girl.  The maintenance is insane and usually I am not great at making regular appointments for the filing and plucking and cutting and tweezing and polishing.  I usually get things done on a whim or because I canâ€™t take the jungle growing above my eyes anymore.  Home upkeep is far cheaper, but I am far from skilled enough to keep it up.  I try for a few weeks and then I have to revert to the professionals.   Iâ€™ve never really been the kind to have one hairdresser that I go to excusively.  A college roommate would travel 45 minutes back home to get her hair done and still goes to the same lady today though now the commute is about an hour or more.  I donâ€™t understand that.  Sure, a good hairstylist is hard to find, but there are several out there, in every city.    Iâ€™ve never had a fear of the haircuts either.  Iâ€™ve walked into salons for the first time and told the stylist to just do whatever they wanted to do.  Some stylists like this, others, not so much.  I havenâ€™t really ever had a bad experience.  One hairdresser said that I didnâ€™t have to worry about my hair framing my face while other women used their hair to enhance their features and thus the anxiety they have.  I think she was complimenting me.  She also gave me one of the best haircuts and I ended up going back to her a few times.  In the 10 months that I\'ve lived in Richmond I\'ve had my hair cut three times by three different people, none were bad, but none were exceptional either.  Today, my stylist was Daniel, a thin man with spiked dirty blond hair who smelled like cigarettes and herbal shampoo.  I was apprehensive at first for a man to be plucking my eyebrows and cutting my hair and then he spoke and walked.  There is something comforting about a gay man being in charge of your beauty needs.  I told him I just needed a trim and he guided me to a sink then said weâ€™d do the eyebrows first so we went to back room with a comfy bedlike chair.  He worked on my eyebrows longer than most women stylists have.  He plucked the small thin straggling hairs at the top corners of my brows and only put the wax on once.  He massaged the areas wear he had waxed and plucked.  He relaxed me.  He took his time and care.     I was disappointed that he didnâ€™t massage my head when he shampooed my hair.  I expected he was an extended massager, he wasnâ€™t.  As we walked to his styling chair I wanted to tell him to do whatever he wanted with my hair, to put his spin on it.  As he was combing out I my hair, I wanted to tell him to be creative, do what he thought was best for my face.  I never did.  I sat quietly and watched the other customers.  He never asked how much I wanted taken off.  He just took my hair and started cutting.  He cut it to a healthy length, parted it down the middle and added subtle layers.  He did this without asking or telling me.  He cut and cut and cut, the layers taking time and many head tilts.    When he was finished cutting he put straightening cream and hair gel into my wet hair and blew it dry using a rolled brush to straighten and turn the ends under, a skill I can never get down.  He didnâ€™t talk much, nor did I, but when he did he said things like â€œrock and rollâ€ and â€œtwisted.â€  With these words, I realized we couldnâ€™t be friends.    Hours after meeting him, he lingers.  When my hair swishes or I brush it out of my face, I smell him, cigarettes and herbs.  In the time that I walked in looking for a simple trim and walked out with a hairstyle, Daniel became my hairdresser.  Iâ€™m committing to him like no other man.  I know at least he will take his time and care and he\'ll give me what I need without even asking.  '),(11517,3022585,'01,August,2004','2004-08-01 00:00:00','   For weekâ€™s I have been wondering when exactly Fatherâ€™s Day is.  I know of course, that itâ€™s in June and on a Sunday.  But I feel like it moves around from year to year.  I have, however, just learned that itâ€™s always the 3rd Sunday of June each year.  Hmm, I really have trouble with dates.  Iâ€™m horrible at birthdays.  Actually, horrible is probably not strong enough a word.  My best friend and I will often send birthday cards that are 6 months late.  But, we know thatâ€™s the way it is and we accept it, wholeheartedly.  Others, however, are not so kind.  Iâ€™ve known people to really get offended by others forgetting their birthdays.  Thatâ€™s just damn selfish in my opinion.  I donâ€™t know.  Iâ€™m not a big birthday celebrator and never really have been.  Iâ€™m trying to get better at it, but dates elude me.  Like today, I was reading my daily blogs when   urlLink Jennifer Weinerâ€™s post  mentionâ€™s fatherâ€™s day and I have one of those â€œOh my God, Iâ€™ve got to call my Dadâ€ moments.  I ran around looking for my cell phone and as Iâ€™m talking to my Dad I realize that I have not sent a card or anything.  I quickly realize I will try to remedy the situation next week at family beach week, but still shitty, shitty of me.  I sent my Mom a Momâ€™s day card (a day late of course), but nothing for my old Dad, who loves me so unconditionally.  Sorry Dad.  Itâ€™s not personal, I just suck.  Well, I guess in my case, weâ€™ve established why there is no Daughterâ€™s Day.  On the bright side, maybe the Fatherâ€™s Day cards will be half-price tomorrow.  '),(11518,3022585,'01,August,2004','2004-08-01 00:00:00','   Cellophane.  Itâ€™s my new favorite wrapping material.  Iâ€™ve just wrapped 2 presents for todayâ€™s scheduled parties and yes, one is wedding related and yes, Iâ€™m over it.  As I went through my quite vast assortment of wrapping material I landed on the cellophane for the first package because I had to due to itâ€™s odd shape and then I just had to us the cellophane again for the next one, itâ€™s just too fun.  And, while wrapping, I walked around singing â€œCellophane, Mister cellophaneâ€ from Chicago.  And now, Iâ€™m just chillinâ€™ before I have to leave to drive all over Virginia in one day.  Iâ€™m usually a last minute shopper and would typically be out buying the presents right now. But, Iâ€™ve learned to stockpile presents so that I can just pull from the bunch and go.  But, my last minuteness is not totally gone, Iâ€™m really waiting on my clothes in the dryer and I was suppose to leave like 5 minutes ago.  Oh well, people love me anyway, despite all the lateness and last minute presents, tags still on.  I am, in fact, not Miss Cellophane cause you gotta notice the girl who is fashionably late.  '),(11519,3022585,'01,August,2004','2004-08-01 00:00:00','   Last night, at about 9:30 I had the bright idea of going to see â€œRaising Helen.â€  Some friends had already seen it and since I had no work the next day I could handle to the late night showing at 9:55.  So, I hopped in the car and raced to the West End to the stadium seats.  It was me, and three couples.  Wonderful, nothing like smug couples on a Thursday night date.  Whatever.  I took my own row, settled in and put my feet up.  I love the movies.  I started going to movies alone about 5 years ago when in college an impossibly popular co-worker divulged her love of going to movies on her own.  I was appalled and shocked.  Why in the hell would a 20 year-old, blond, size 2, charming, and smart girl ever go to the movies by herself?  She explained that her mother did it and that she liked it.  She would often be shopping and find a movie on and just go.  I thought she was brave.   It took me a good year before I braved the movies alone.  It must have been uneventful, as I donâ€™t even remember the movie that ended my solo movie going virginity.  Now, I donâ€™t even think twice about it and yet I think Iâ€™m still a bit of a rebel in the venture.  I donâ€™t think my friends do this.  I even have a friend who doesnâ€™t like to rent movies alone. Movie watching is a solitary act anyway, unless, of course, you are at home snuggled up on a couch.  So, if a movie is out that I want to see and no one else does, or Iâ€™m near a movie theatre and see something interesting, or Iâ€™m bored at home and need to do something other than stare at myself, Iâ€™ll go to a movie, alone.  No big deal, it just happens.  So, last night was â€œRaising Helen.â€    Sad bit of a movie if you ask me.  The plot was weak and lame and not well thought out.  There were scenes were I literally wanted to rewind and rewrite and then there were the ever present continuity issues where the clothes in one scene are the same in the next scene although itâ€™s a different day.  This movie overall has bad writing, bad editing, and not really great acting although I still adore Kate Hudson, John Corbett, and Joan Cusak.  What were they thinking?  Anyway, good premise, just not well done.  And, Iâ€™m beginning to rethink things about my friends who said it was â€œso cute.â€     And then, because I could, I went Wal-marting at midnight.  I stepped into the Wal-mart at 12:15am and spent the next 45 minutes shopping the supper center.  It was fabulous and surreal.   Tons of people were about, mostly Wal-mart workers restocking the shelves and a few customers.  There were boxes everywhere, which you had to navigate around.   The restockers would sometimes give me weird glances like it was crazy that I was bargain shopping in the middle of the night on a Thursday.  Please, like itâ€™s not weird that you are working at Wal-mart in the middle of the night.  Yarn was had because well, the kitting has not stopped.  And then, because I could, I got some fresh produce and checked out at 1am.  Wonder if that is what Sam Walton envisioned for his supper store?  '),(11520,3022585,'01,August,2004','2004-08-01 00:00:00','   So, I went to the doctor yesterday to get a mole checked out.  Since I had blistering sunburn 10 years ago, I have been cautious with suspicious items on my skin.  Turns out that melanoma does not necessarily strike the area that was burned (interesting).  And the mole or whatever seems to be nothing (thank God).  This was my first visit to this doctor in Richmond so I had to fill out all the info. sheets and sign away my life, literally.  Well, since it was the last day of school I had a little trouble leaving when I wanted to and then when I got to the hospital there was an unexpected game of chicken for the limited parking spaces.  So, I finally grab one and walk quickly through the soupy, muggy air (wearing jeans no less) and make it to the fourth floor about 7 minutes late for the appointment.  I get my papers and sit down.  I write and write and then pause to actually read the clauses before I sign them when the receptionist moves the glass plate that blocks out the waiting area (um, itâ€™s glass people, we can still see and hear you) and this is what she says â€œMiss Banana, the nurse is waiting for you because she was expecting you at 3:30.  Can you write faster because youâ€™re late?â€  â€œYeah, OK,â€ I say.  And then it dawns on me that I should have stood up on a chair or something and announced to the waiting area â€œHi, my name is Anna and Iâ€™ll be late today.â€  '),(11521,3022585,'01,August,2004','2004-08-01 00:00:00','   I woke up early this morning to a fog, a thick fog over the city.  It nestled in the trees in the courtyard of my apartment building.  It was tranquil and quieted the morning like a blanket of snow.  I haven\'t seen a fog in years, probably becuase I sleep through them.  But today, the fog  mimiced the fog I\'ve been under for the last few days.  Not a bad fog, just a mindless, in a daze fog where I leave important, confidential documents on a copier, don\'t put an important piece of info. into folders, can\'t comprehend questions asked of me, can\'t understand what I am reading. I\'ve been in a haze but, now, it\'s mid-day and the fog is gone, the sun has burned it away.  It\'s the last day for teachers, my files are completed, I\'ve helped other teachers clean up their rooms and mail letters home.  School is out.  We\'re done.  One year down, a million more to go.  But, for now, for the summer, I\'m free.  Plans for the beach, LA, and NYC are being made, and of course all the weddings.  Happiness abounds with the sunshine raining down.  Pool days and sleeping in and reading books and unexpected trips and movies and relaxation and freedom.  It\'s good to be a teacher when it\'s summer in the city.  '),(11522,3022585,'01,August,2004','2004-08-01 00:00:00','   Sex and The City on TBS. The girls are back.  Fall in love people, all over again.  '),(11523,3022585,'01,August,2004','2004-08-01 00:00:00','   And by that I mean Knitters Anonymous because I am addicted.  I decided to pick up knitting again this weekend and have so far knitted 2 scarves and begun three more.  I\'m obsessed.  I even raced after school to a knitting store that closes at four so that I could buy way too much yarn for half price.  But, at least my friends will nice have nice yarn in their scarves instead of the Walmart brands.  I want to constantly be knitting.  I\'ve even started knitting in bed with no lights on.  That\'s right, knitting in the dark, sexy.  And yesterday, I had to refrain from taking the knitting to the bathroom with me.  Knitting is on the brain so much that it\'s all I think about (at least for the last few days).  Like, when I\'m at work or out with friends I keep thining \"I really wish I had my knitting right now.  I really want to finish that scarf.\"  It\'s like I need a bumper sticker that says I\'D RATHER BE KNITTING.  And no, there is no purling, just kniting and knitting and knitting.  See the obsession, it\'s bad and well, I need to get back to the needles and the yarn.   Tonight was yet another wedding related event, a surprise semi-shower for the friend and co-teacher.  Enough with the wedding stuff already.  The showers and the bachelorette parties and the girl\'s weekends and the weddings.  I mean there is no money left, the well is dry.  I\'m really going to be weddinged out by November.  \"Dear Lord, make me a bird so I can fly far far away.  Dear Lord, make me a bird so I can fly far far away.\"  Or maybe, I\'ll just bring my knitting and then it will all be bareable.  '),(11524,3022585,'01,August,2004','2004-08-01 00:00:00','   Woke up around 11:30 which really not that late, but considering it was drunk sleep, it was far to early.  Too much bachelorette partying.  My head hurt and I felt like a ton of bricks hit me so back to bed I went at 2 pm for a long nap.  I don\'t even know how I got that drunk or if I was that drunk.  I didn\'t drink that much, although people were handing us shots all night, but those fruity shots don\'t have much alcohol in them.  And, I had my wits about me at the end of the night.  I was probbably the most sobber one there.  So, what the hell with the hangover?  Anyway, it\'s currently 5pm and I need to get on with my day, or what\'s left of it.  '),(11525,3022585,'01,August,2004','2004-08-01 00:00:00','   Itâ€™s summer, well almost and not officially, but heat wise; it is summer in the city.  Poor Virginia, it is plagued with humidity and hot summers.  I know our summers are not like Texas summers or Louisiana summers, but they are hot enough for me.  I donâ€™t believe in snootiness, but sometimes I wish I were a person who â€œsummersâ€ somewhere else if just to get away from the nastiness.  Maine, Maine sounds good, or maybe even Canada.    The only thing really good about the heat is that pools open and there are trips to natural bodies of water.  Every year my family takes a beach trip. I actually hate sand and saltwater, but I love the laziness of a beach week.  I take about ten books and I just read and play with children and read some more.  And then there is, of course, the eating of the shrimp and crab.  There is always the eating of the shrimp and crab.  But, before and after the beach, there is the pool.  Growing up we belonged to our local Elks Club, an organization I really donâ€™t understand except that there were usually old men there at any given time throughout the day and they could be found at the bar inside the clubhouse.  Iâ€™ve never found out what the organization stood for or against and as soon as the last child (me) went to college, my fatherâ€™s membership ran out.  We used the place for the pool and the pool only.  The local pool, a place where you learn to swim from the coolest man you have ever seen because he wears sunglasses underwater, where you strengthen friendships over adult swim breaks and during â€œsharkâ€ in the deep end, and where you tackle the fear of the â€œhigh dive.â€  I never joined the swim team, but I was a lifeguard for 2 summers.  Lifeguarding is the best job ever invented.  Sometimes I think about going back to lifeguarding with my summers off, but then I think that the current 16 year olds would find me a bit strange and I would certainly not fit into their lifeguarding clique.  Thatâ€™s OK.  I got my lifeguarding kicks back in the day anyway.  I just really miss the twirling, the twirling of the whistle while sitting in the tall chair with a high vantage point, the sun so strong you can see your legs tanning before you eyes.  The pool is now what I go to with my nieces and nephew.  None know how to swim yet, but I am determined to solve that one.  My favorite thing to do with them is to take each one, individually, in my arms and pull them across the pool from the shallow to the deep end.  On the cross-pool treks, our faces close together, we pretend different things, sometimes they practice kicking or just floating.  When water rushes into their mouth and nose I know itâ€™s a 50/50 chance of them quitting the trek mid-pool.  But, the other day, my precious niece coughed it out and swallowed hard then said, in three year-old voice, â€œAnnie, I want to ride the waves again.â€  You see, we were pretending to be on the ocean and waves were gently bouncing us up and down (hence the water in the nose from the crashing wave).  I love the pool.  I love that it turns a three year-old into a chlorine swallowing wave rider.  '),(11526,3022585,'01,August,2004','2004-08-01 00:00:00','   There is this guy I know, but he shall remain nameless, although I might be related to him.  I adore this guy.  He is one of the funniest people I know and whenever I am around him I know that we will be laughing so hard that we fall over.  However, last night I discovered he is the biggest dork I know.  And this is why.    We both love Apple.  We are both obsessed with Apple.  Like last week I emailed him about the new  urlLink Apple Airport Express  and he emails me back and is like â€œI was reading about it just when you emailed.â€  And then the next day he calls me and we chat for like half an hour about how the world really needs to know that Apple is so superior and really people would be happier if they just owned an Apple and businesses were run on Apples.  We discover it is a futile attempt, but our love for Apple will never die.  This guy is a certified Windows programmer and even interviewed for a job at Microsoft so his love for Apple is even more special than mine.  He converted about 4 years ago and has not looked back since.  In fact, heâ€™s a little pissed at people who like PCs.  He loves Apple so much that he gets magazines about Apple and somehow is in the loop of rumors about the new products they are creating. I know there are rumor websites and he probably surfs them.  Itâ€™s all a business related obsession since he programs all day. I have to confess that if I understood all that computer jargon, I would be in on it too.  I just donâ€™t have time to learn all that stuff about computers, I just know I love Apple.  I also know that he is my Apple lovin\' hero and he is there to dispense knowledge should I have a technical quetions.  He rocks my Apple lovin\' world.    Well, so last night he was telling me about the last issue of  urlLink Mac World .    urlLink Alton Brown gave an interview and developed a special recipe to celebrate Apple.  The name of the dish is APPLE MACaroni and Cheese.  â€œEw,â€ he said.  â€œYeah, weird,â€ I said.  Then, this guy that I adore,went on to tell me that at the last Mac Expo (or whatever) when Steve Jobs gave his closing statement, he had a friend over to his house during their lunch hour and they watched it on the internet while eating MACaroni and then for dessert they had APPLEsauce.  Oh, dear one, I still, still, adore the hell out of you!  '),(11527,3022585,'01,August,2004','2004-08-01 00:00:00','   Ronald Regan\'s body will be lying in state on Thursday at the capitol in DC.  Iâ€™m thinking about going.  Although I donâ€™t really have the money to spend on gas, when will I ever get to see a president lying in state in the near future?  And really itâ€™s not about the pomp and circumstance, itâ€™s about Ronnie.    Heâ€™s the only president I feel personally connected to and itâ€™s not about his politics or his cuddly nature and love for his wife or that even a president can be human and have family problems too.  Itâ€™s about birthdays.  You see, ole Ronnie and I have the same birthday.  Big deal you say, you also share it with Bob Marley.  But, Bob didnâ€™t send me a birthday letter when he was president.  Ronnie did.  It was 1988, I was eleven and this was the last year for Reagan in the White House.  I thought my sharing his birthday was the coolest thing.  I had no idea what his politics where nor did I care.  I considered myself a Republican then because my parents were Republicans and so this extra kinship with Ronnie felt strong.    My mom had the brilliant idea of sending him a birthday card and announcing the importance of the shared date.  I was a little apprehensive, how exactly do you tell a stranger, much less a president, that you have the same birthday?  Why would he even care?  Well, we bought a card and wrote a note.  Iâ€™m sure my mom helped craft the sentence or else it would have been something like â€œI was born on Feb. 6 too.  Happy Birthday Mr. President.â€   And, well, thatâ€™s probably what it did say.  So, off went the card and about 3 weeks later arrived a letter from the White House.  It was the most exciting day of my eleven year-old life.  My family was thrilled too.  I learned of his death on Saturday night after a full day spent with old college friends.  A friend was looking at the now framed letter on my living room wall and said that Reagan had died today.  â€œWhat?â€ I thought.  Then I thought about peace and that it seemed so expected and so natural.  Iâ€™ve dealt with Alzheimer\'s in my family and it is trying disease, on everyone.  I told my friends that it was probably for the best for him and his family.  Then I went to the letter.  He, or a staff writer, wrote a short and sweet letter.  It is typed on a typewriter.  The date is off-center from the letterhead and the eagle emblem.  It begins by acknowledging our â€œmutual birthday.â€  Then, in presidential fashion, goes into the history of Febuuary birthdays including George Washington and Abraham Lincoln.  He tells me to study these great men and hopes they will inspire me the way they have inspired him.  The letter ends with Mrs. Raegan joining him in his wishes and may God bless me. His name is signed simply â€œRonald Reagan,â€ no mention of president anywhere.    We think he really signed it.  It doesnâ€™t seem to be a stamp.  I will probably never truly know. But, it doesnâ€™t matter.  My eleven year-old heart knows he touched this letter somehow.  In the bottom left hand corner is my full name and childhood address.  My name has a typo.  Since the beginning of the letter I have gone from Anna to Anny.  At eleven, this typo crushed me.  Now, it is endearing.  It makes the whole letter so much more real, so much like the man whoâ€™s birthday I share.     Dear President Reagan,  I am honored to have shared your birthday.  Every February 6 I will continue to remember your birthday and those of the other great presidents you mentioned.  Washington and Lincoln were truly amazing, but they donâ€™t mean as much to me.  I canâ€™t look over at my wall and see their handwriting.  I loved your letter and will always cherish it.   May you and your family find the peace you deserve and thank you, from my eleven year-old self and from me now, for all you did for our world.  					Sincerely, 					Anny  '),(11528,3022585,'01,August,2004','2004-08-01 00:00:00','   J. is a â€œmilitary brat.â€  Sheâ€™s been all over the country and Germany, her family moving from place to place every few years.  Somehow she has roots in the mid-west and Virginia.  Iâ€™m thankful for those Virginia roots, happy they brought her friendship to my door, literally.  Ohio claimed her during undergrad.  She was a Buckeye.  I guess she still is.  Iâ€™d like to think she is more a Wahoo, but itâ€™s not my choice who wins out.  The Buckeyes could probably kick our Wahooâ€™s football ass any day.  J. now lives in Austin and owns a home there.  Her settlement came by way of Miami and Seattle, her traveling childhood brought to the surface in young professionalhood.  Sheâ€™s an audiologist.  I think she is resilient.  She lived 3 doors down from me in one of Thomas Jeffersonâ€™s original rooms at his University.  We had fireplaces, wood floors, a sink inside a closet, and a bathroom we had to walk outside to use.  Not exactly an outhouse, but close enough.  We were in grad. school, but J. was only there for one semester.  We spent hours chatting in our rooms, at night walking from room to room in our pajamas, outside.  When she left it was winter and cold, the freezing temps. had been lifted by fires in our fireplaces.  Our individual rooms heated the old fashion way.  J. was the best at making a fire that didnâ€™t smoke the room out.    I was sad to see her go and begged her to stay, to change her practicum placement to Virginia, not Kentucky.  But her plans were made far before any of us moved to the Range.  So she left but came back for Foxfields and a New Yearâ€™s party in a cabin in the woods.  J. is a good friend.  She is smart, logical, sensible and yet just as neurotic as I am (well maybe not).  We talk about how much our lives have mirrored each otherâ€™s while she moved from state to state or country to country and I lived in the same pink bedroom for 18 years.    Iâ€™m so glad she stayed for that last semester.  Iâ€™m so glad she couldnâ€™t find an apartment or dorm room for those few months.  Iâ€™m so glad we are friends.  And I love her couch or air mattress in her tiny cross-country apartments.  Sheâ€™s an anchor in my life, a friend who will always be there and never feel that far away no matter the actual distance.  '),(11529,3022585,'01,August,2004','2004-08-01 00:00:00','   Restaurants are closing or are closed by now.  Itâ€™s late, but youâ€™re hungry.  Youâ€™re not sure you want to cook, but you really want a salad.  Drive to Ukropâ€™s on your way home.  Rifle through your dashboard CDs looking for Emmylou Harris.  You need Emmylouâ€™s mellow voice to match your mellow mood.  Sheâ€™s not there, but youâ€™re happy you find Johnathan Rice.  Put it in the CD slot.  The volume is louder than you expect but you donâ€™t change it realizing â€œQuiet Hushed Voicesâ€ is what you were really craving.  Sing along to it with the sunroof open.  Donâ€™t mind the people in the parking lot looking at you with bass flowing out the open windows.  Keep singing the words.  Park, knowing you are only pausing the moment.  Enter the automatic doors.  Pick up romaine lettuce and chevre.  A salmon fillet is a sudden passing desire; you grab it, and hummus, for snacking.  Check yourself out in the self-serve lane.  Get back in the car.  Turn the key and hit the sunroof button.  Keep the duet with Johnathan going.  Feel your brotherâ€™s bass in the floorboards and through the speakers in the door.  This is good music.  Know that you canâ€™t go home.  You want to drive and listen and feel the wind.  There are no mountains here, no country roads just 5 minutes away.  There is the highway.  Head toward home, but pass it.  Make the right hand turn you make every morning on the way to work.  Bare left this time.  Take the toll road, the Downtown Expressway.  You need a hit from the city buildings; the skyline up close and 50Â¢ is worth it.  Pass Rosewood and the U of R Stadium.  Put Johnathan on repeat.  Slow at the toll booth until the SmartTag signals a green light to keep going.  Speed up, a horse out of the gate.  Get the center lane.  Pass VCU and the Cherry St. overpass.  Pass the 2nd St. exit and Thomas Jeffersonâ€™s Capitol building.  Go under a quarter mile of overpasses with lights burning a fluorescent tunnel.  Emerge to the buildings of downtown Richmond.  Wonder why so many lights are on.  Follow the signs to 95 North and 64 West, Washington and Charlottesville.  Curve around on the highway bridge through low buildings.  Look out the window, to the left, at the taller lit buildings.  {I want to love this city.}  Be amazed that you are driving so close to the bricks and concrete.  See the clock tower of the renovated Main St. train station on the right side, you can almost touch it.  Sing with Johnathan.  Get off at the Broad St. exit.  Go quickly through Shockoe Bottom.  Remember the recent attacks here.  Get back to 95 going South this time to Petersburg.  Stay in the right hand lane, curve right at exit 73C toward the Powhite.  Imagine you are not driving but are suspended, floating through these buildings, under the highway above you.  Look through the windows of La Difference.  Wish you understood the difference.  Go back through the tunnel of lights.  Pass the Jefferson Hotel and the Governorâ€™s Mansion.  {I want to love this cityâ€™s history.}  Pass the exit for the Boulevard.  Quell the wish to live there in an apartment alongside Virginiaâ€™s museums and with a balcony.  Go on to Cary St.  Go to your home.  Park and walk to your building.  With the full moon behind say quietly and to no one \"I want to love this city.\"  Unlock you door.  Put down your bags.  Wash you hands.  Decide itâ€™s too late for salmon.  Get a large bowl out.  Pour the salad into it.  Begin to toast pecans from the freezer.  Crumble the chevre over the lettuce.  Pour on your homemade balsamic dressing.  Add the toasted pecans.  Take a bite as you walk to the living room, to your computer.  Realize you are a good cook, that this salad should be on menus.  Sit down at your laptop.  Put Johnathan on again.  Listen to him, and write this.  '),(11530,3022585,'01,August,2004','2004-08-01 00:00:00','   Rrrrrrrring Me:  Hello. Them: (silence) Me:  Hello Them:  Yes, is this Mrs. So and So Me:  Yes* Them:  Hi Mrs. So and So.  Iâ€™m calling for the Richmond Fraternal Order of Police Me:  Hello.  Iâ€™m sorry,  did you think I was home?  Iâ€™m actually not able to answer your call right now.  Gotcha!  Ha Ha!  Please leave a message, unless this is a computer operated telemarketer in which case do not ever call me again or I will BITE YOU.  If this is the Richmond Fraternal Order of Police then also do not call me again unless you have actually done something about the 10 murders that happen each month in our fair city, especially those that happen around the schools I teach in.  When I feel good and safe about driving to the schools I teach in, those schools which sadly produce men that will kill others on the same streets, Iâ€™ll answer youâ€™re call.  And listen, weâ€™re working on educating kids that donâ€™t kill.  Weâ€™re trying.  Iâ€™m not so sure about your fraternal orderness.  So, until that time, Stop. Calling. Me.**  *Iâ€™m thinking, â€œIâ€™m not a Mrs. which means you are a telemarketer asshole cause no one else calls me Mrs.â€ **What I really said was, â€œIâ€™ve asked you five times to stop calling me, especially early in the morning and especially on Sundays.â€  This was followed by an abrupt click.    Sometimes, I wish I was a quicker with the unsolicited phone call comebacks.  '),(11531,3022585,'01,August,2004','2004-08-01 00:00:00','   It\'s all about the gmail.  New email to the right.  Read it, copy it, paste it, write it, send it.  In short, email it to me, the gapgirl is ready for mail.  '),(11532,3022585,'01,August,2004','2004-08-01 00:00:00','   Iâ€™m sitting at the Getty along a path to the side of one of the buildings.  Iâ€™m in the shade and leaning against the structureâ€™s massive stonewall.  Iâ€™m on a pathway to a terrace though itâ€™s a path that is almost hidden.  People have to navigate around me as they walk through.  I wish theyâ€™d go away and leave this public space private, for me only.  I can see a vast view of LA from where I sit from the hills to the oceans.  Ahead and slightly to the right is a small clump of tall buildings that make up downtown.  Beverly Hills is just below me.  There is a fire billowing on a mountain a few miles away.  Itâ€™s in Santa Clarita a passerby says.  The smoke wonâ€™t reach LA because the winds blow toward the east, toward the mountains.  It looks both ominous and soft, like a cloud.  To the left and over the hills from the Getty is the Valley and Pasadena, though I canâ€™t see them I know they are there with clichÃ©d houses and perfect lawns.  There is smog that hangs in the air between the landscape and the sky, a visible reminder of this cities population.  To my right, if I careen around I can see the sea and Catalina Island poking its head just above the haze.  The 405 are below me and the traffic crawls from north and south.  Planes line up in the sky ahead, waiting to land at LAX.  This city is huge, a massive gallimaufry of rich and poor, everyday and famous.  The rich build houses on the mountains or even cliffs, ignoring the impending possibility of earthquakes.  They look like cracker or thin wafers that could tumble down like rocks sliding.  California has itâ€™s own style to everything, but the houses in LA interest me the most.  Their style is a mix of Mediterranean, Spanish, and American classic.  The rich, modern houses are minimalist with hard lines and lots of glass, but mostly there is stucco, tan and smooth.  It is the red brick of southern California.    The Getty is an amazing place.  , a free art museum, open to the public.  But, people donâ€™t usually come for the art; they come to see the Getty.  It is a vast structure of buildings placed on a hillside overlooking all this city has to offer.  It is made of an off-white stone creating straight lines and curved walls.  The landscaping is immense, green, and perfect.  Fountains pop up in many places, an offering of serenity.  The buildings are so large that looking back on them from on of the many terraces and seeing people walking around on the balconies and inside through the glass, they look like tiny bugs, like insignificant, unimportant things.  People stand against the railings where I sit.  They squint and point trying to comprehend what they see.  The wind is strong at times on this hill.  They sun beats down an arid hotness, but in the shade of the building, a chill awaits.  '),(11533,3022585,'01,August,2004','2004-08-01 00:00:00','   Happy Birthday A.!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  What to say about A.?  Only she will know what the title of this post means.  She is a dear college friend whom I lived with for 4 years, 4 crazy party filled and chilled out years.  When I first met her I didnâ€™t think I liked her because she is so honest that it can be abrupt at times.  Then, her roommate went home for an extended time due to illness and A. and I became good friends and I saw her kinder, gentler side.  I ended up staying in her roommates bed quite often because my roommate was a freak of nature and I just needed to get away even if it was across the hall.  Iâ€™m so glad we became friends and that she is in my life.    A.  is the kind of friend who is there for you no matter what.  Sheâ€™ll be there to help you move in to your new apartment or celebrate you graduation a year after she celebrated her own.  She is honest and doesnâ€™t judge.  When confronted with people who are quite strange she says, â€œit takes all kinds to make the world go round.â€  We use to rag her in college because the words â€œmy mommy saysâ€ preceded almost every piece of factual or advice laden talk.  Well, her mommy did a great job.  She is a great girl and a good friend.    She married a few years ago.  He is a boy who counteracts all that we college friends thought she was.  He is country and mild mannered and loves hunting and listens to country music and watches racing.  A. was a bit wild in college so when we found out her choice we were a little confused.  It turns out that the domesticated life filled with good cooking and family all around is really her true self.  Her in-laws love her as much as we do and she has found a niche for herself and her husband.  They are beginning to build a house on 20 acres of land in the country soon.  I tell her that as soon as she puts a pool in Iâ€™ll be there.  I bow to A. and P. land on this auspicious day and bid you a good one.  May tiny pictures from walmart decorate the halls.   Happy Birthday,    ~ Azafata.  '),(11534,3022585,'01,August,2004','2004-08-01 00:00:00','   Why people?  Again, google has sent someone to my site who typed in \"together to pee.\"  I just... I just... don\'t understand.  '),(11535,3022585,'01,August,2004','2004-08-01 00:00:00','   Ahh, little Elizabeth Bishop was so right.  So, I\'ve lost my keys.  They are in North Carolina.  I am not.  I didn\'t realize I didn\'t have them until I awoke from the early morning drive back to my parents\' house.  Many phone calls later, they are hopefully being fedexed to me and will arrive Tuesday sometime.  I\'m trapped at my parents\' house which isn\'t bad except when you want to be in your home, you want to be in your home.  And I, want to be in my home.  Say some prayers people, may fedex not lose THAT package.  I\'ve never lost my keys before and really it\'s not a huge deal though it may look like (WRITE it) like disaster.  '),(11536,3022585,'01,August,2004','2004-08-01 00:00:00','   The LA chronicles are not over, but I have been busy washing clothes, making gifts, and repacking for another weekend of wedding festivities.  So, Iâ€™ll be away from bloggerland for another few days.  In the meantime go to my little blogchildâ€™s site.  Yes, yes, Anna has a child, in bloggerland that is.  Go, read, comment.  urlLink  WahooBuckeye   needs you.  '),(11537,3022585,'01,August,2004','2004-08-01 00:00:00','   with another city.  I never thought I would like Los Angeles.  Never.  I thought I was an East Coast girl through and through.  Turns out, I could go either way.  Los Angeles isnâ€™t anything I thought it would be.  It wasnâ€™t as flashy as I had imagined.  It didnâ€™t seemed as pretentious as it looks on TV and there was no preparing me for the weather.  LA was amazing and good to me.    I tend to fall in love with most cities that I visit.  Maybe itâ€™s because travel is a carefree existence or maybe itâ€™s because each city beckons to a different part of me.  Why did I fall in love with LA?  Well, it isnâ€™t the rat race it seems to be or maybe I never came in contact with part of it.  And the sun shines everyday.  There are never clouds in the sky or rainy, mucky days (there are also not stars).  I knew this about LA, but I didnâ€™t believe it.  My clouded East Coast brain couldnâ€™t fathom endless days of pure sunshine.  How could you be sad in a place like this because whatever happens, at least itâ€™s a beautiful day.    My love for LA is fleeting, though.  I know when I am home Iâ€™ll settle back into the humidity and heat and then the cool air of fall and the snow of winter.  I would miss the seasons if I lived here.  Endless days of sitting by the pool would begin to disorient me.  Life here would be like an endless vacation.  It would be like play, like pretend, just beyond the touch of reality.  And my love for LA is not true because it is shared with my love for London, New York, San Francisco, Seattle, and Morehead City.  LA is just another city to add to the list of places I could live if I ever decided to pack up and leave the state and family I know so well.  I never do make plans to move to the cities I love so much because I think if I lived there the magic would begin to disappear after a time and I donâ€™t want that to happen.  Iâ€™d rather hold these cities in my heart as perfect gems nestled on hillsides or islands or by the sea, covered in fog or basking in sunlight.  '),(11538,3022585,'01,August,2004','2004-08-01 00:00:00','   Lots of things to say, but I\'m tired and trying to get back to east coast time.  I love and miss LA.  It changed me.  I\'ll tell ya all about it later.  Time to watch movies and relax.  '),(11539,3022585,'01,August,2004','2004-08-01 00:00:00','   Iâ€™m off to LA for fun in the sun with my best friend.  Itâ€™s the last of a girlâ€™s romp before she (you guessed it) gets married.  I plan to do some touristy things and then just hang with my best gal pal.  I miss her and Iâ€™ll miss her even more when she gets hitched because girlfriends change ever so slightly when marriage happens.  I was hoping weâ€™d hop over to Vegas for the weekend, but weâ€™ve opted for time in LA with her fiancÃ©.  Well, I love them both, loads and loads.  Iâ€™m excited to see them, either in a flashy city or just in their living room.  Have to say though, kinda wish I was going to see a booby show.  Iâ€™ll blog at ya next week.  Blogger back yaâ€™ll.  '),(11540,3022585,'01,August,2004','2004-08-01 00:00:00','   We at mind the gap are a happy and loving lot.  We like diversity and we like to celebrate with our friends when they have things to celebrate.  Today is such a day!  Today is Bastille Day in France, a day of celebrating the end of monarchal rule.  And so, in honor of our French friends, a list of things we love from France, or things that France pimped out to us.  â€¢ the Statue of Liberty  â€¢ le baguette (yummy) â€¢ the colors of the French flag  â€¢ fashion  â€¢ le croissant (even yummier)  â€¢ Dijon mustard â€¢ Champagne  â€¢ Cannes Film Festival â€¢ toy dogs (because the French took them everywhere long before Paris Hilton was born) â€¢ Jaques Costeau â€¢ Les Mis (though in French it would be lost on me) â€¢ Jean-Marc Gaspard Itard and the wild boy of Aveyron â€¢ Matisse, Degas, La Tour, Cezanne, Monet, Rodin, Renior, to name a few â€¢ Joan of Arc â€¢ â€œLet them eat cake.â€  This list does not endorse the loving of all things French because those French bastards didnâ€™t go to Iraq with us so we said the hell with you too.  Our amazing and welcoming country forgot that French troops helped us in the American Revolution and about the gift that France once bestowed on us, the Statue of Liberty, and said, \"Weâ€™ll call French fries, Freedom fries instead,\" ignoring the fact that French fries are actually from Belgium.  â€œHa, take that you French bastards,â€ our government said.  Well, those French bastards were pretty spot on and had a sense that Bush was blowinâ€™ smoke about weapons of mass of nothinâ€™.  Damn those French bastards!  Happy Bastille Day everyone!  '),(11541,3022585,'01,August,2004','2004-08-01 00:00:00','   The Shakers say it is a gift to be simple and a gift to be free.  â€œSimple Giftsâ€ is an old shaker song written in 1848 and it is one of my favorites.  It is a song about simplicity, but it also says that â€œwhen we find ourselves in the place just right, we will be in the house of love and delight.â€  While the house of love and delight sounds like a cheesy, incense burning, hippy joint, the sentiment is what gets me.  I like the idea that when we are true to ourselves we will find the simplicity of life which ultimately brings peace.    But itâ€™s not always so such a simple gift.  We are all given gifts to some degree.  We are can run companies, we can cook amazing food, we can play an instrument, we can balance budgets, we grow a garden, we can teach someone to read, we can diagnosis symptoms, we can write, we can act, or we can see the world differently through the lens of a camera.  We all have talents and gifts which should be a part of our everyday lives.  The act of doing what you are good at, what you are talented in, is itself simplicity and freedom.    My cousin B. is an amazing photographer.  She has an eye like no one Iâ€™ve seen.  She loves taking pictures of things like rust and wood.  She can take scenery and portraits, but these are not her specialty.  Those shots are easy and generic, anyone can do them.  The passion she sees through her lens are things we would easily pass over.  B. hones in on the things that we feel corrode of corrupt surfaces.  Somehow, B. finds the beauty in them and she shows rust and corrosion as art.  And she does it so well.     She has never sold a photograph and has pretty much stopped taking pictures creatively when she left college.  All because the cost of taking and developing pictures is too much for her budget.  Iâ€™m saddened by this and I tell her often that she has to start taking pictures again, that she has to use her gift.  It is a gift to be simple and a gift to be free and sometimes we have to fight to use those gifts.  If our favorite writers and actors and doctors and teachers didnâ€™t use their gifts, where would we be?  How many times has your life been changed by someone elseâ€™s gift?  I know that B. is scared and I understand it.  Putting your creative self out for the world to reject is a terrifying thing.  But that simple Shaker song says that  â€œWhen true simplicity is gain\'d, To bow and to bend we shan\'t be asham\'d, To turn, turn will be our delight, Till by turning, turning we come round rightâ€  So bow and bend B., simply use your gift, youâ€™ll turn round right.  '),(11542,3022585,'01,August,2004','2004-08-01 00:00:00','   Numero dos is done.  Two weddings down, three to go.   The other day I received yet another wedding invitation.  As I perused the oh so familiar oversized envelope I wondered, â€œWho the hell is getting married now?â€  I thought I was aware of all the weddings for this season.  Apparently now, but I was relieved to find out it is a second cousin.  Iâ€™m not close to her, so the pain of another wedding was lessened at bit.  It would actually be fine if I didnâ€™t make it to this one.  But, it is in Florida in September.  Either that would be a great time to get a last minute tan or to see the wrath of a fabulous hurricane since it is the beginning of that season.  I donâ€™t know if Iâ€™ll go.  It would really be a weekend of hanging out with the rents which, letâ€™s face it, I do enough of in my own apartment when they decide to come visit, often.  I am; however, quite happy for this cousin as it is her second time of engagement though the first for actual invitations.  Sheâ€™s in her late thirties and that alone gives me a bit of hope and room to breathe.  Tonight was a great wedding.  The drinks flowed and good food was served and we ate dinner overlooking a lake at sunset.  There were people Iâ€™ve never seen before and familiar faces from all the showers and pre-parties.  There were good friends and new friends and family.  And there was dancing and pictures and smiling and happiness and love, lots of love all around.   Starting at 3pm in a Catholic Church, I nestled in to back row pew and beckoned friends to follow suit.  Our row, full of teachers, was the official peanut gallery.  We could hear and eye everything from out backseats, the late comers, the whispers, the crying children, the brat child telling her mother to â€œshush,â€ we saw it all and yet WE kept quiet.     Not being Catholic, we were worried that this would be what my Dad likes to call and â€œaerobics classâ€ service.  You know, up down, up down, kneel, up down, kneel, etc.  But, my good friend L. and C. decided on a Catholic light service since they too are Catholic light, if anything. The wedding was short and pretty with well-chosen songs and readings. The parish was beautiful.  I always love big, cathedral parishes; they almost make me want to be Catholic.      I go to so many weddings that I could probably recite the scriptures and the vows by heart, but this wedding did not bore me and when L. walked down the isle with tears in her eyes I kinda welled up myself.  Iâ€™m not a crier at weddings.  In fact, I thought for a long time that I had a stone heart because of my lack of crying at weddings.  Everyone else gets caught of in the emotions and you see women dabbing their eyes and bridesmaids pulling Kleenex wrapped about their bouquets.  Iâ€™ve never needed a Kleenex at a wedding and have always though â€œYeah, itâ€™s a wedding, whatâ€™s the fuss, why are you crying?â€  But, today, for L., I understood a little more.  Sheâ€™s almost 31 and the tears in her eyes were the symbols of her waiting so long for that man who loves and respects her and that she in turn loves and respects.  Heâ€™s not a perfect man, but he is one with whom she can share a life and so she cried for all that she is getting as she walked toward him.  And when I saw her crying, walking with her father, I thought of myself and if I ever find that man Iâ€™m sure then that my stone heart will certainly melt and tears will run down my face as I walk toward him too.    Cheers to L. and C., may you live happily ever after!  '),(11543,3022585,'01,August,2004','2004-08-01 00:00:00','   So I havenâ€™t been great at blogging lately, especially since I returned from vacation.  I even wrote those beach ones on vacation when I was up late at night and bored.  But, now back at home there are things to distract me.  I should be writing all day since I have no job, but I picked up a few kids to consult on and tutor so that takes up a few hours a week and then well, today, Iâ€™m just procrastinating.  I have my final paper due for a class that finished on Tuesday night.  The paper was actually due Tuesday, but she gave us an extension and I took it.  I am the worst procrastinator.  I will put things off until it is 2 hours to deadline, but this stress is really what gets my creativity and productivity going.  Iâ€™ve tried to be diligent and work on things early and it is so lackluster that it seems a waste of time.  Iâ€™ve tried to put my own deadlines on things to make me work on it and then go play, but no, again, I donâ€™t fall for my own tricks.  So, today, I probably should have turned it in sometime this morning, but she didnâ€™t give us a time and so I havenâ€™t started.  But, now, now Iâ€™m getting ready to start.  It should only take two hours or so, unless I read some blogs, or turn on the TV, or have to call that friend back, or have to find that CD thatâ€™s been missing for 3 months because only THAT CD will make me productive.  Anyway, off to work on it, *wink*.  '),(11544,3022585,'01,August,2004','2004-08-01 00:00:00','   I met him two weeks ago before I went to the beach.  Iâ€™ve been thinking about him ever since and am so happy when I see him.  He lights up my face and makes me feel like I have the best life in the world.  He is amazing.  Iâ€™ve even met his family and they are amazing.  He has a brother and sister and his dad his handsome and his mother is kind.  They are calm and centered and welcoming.  On Monday, his mother gave me five ears of corn, just because.    This boy is undescribable.  It doesnâ€™t matter what I write because you canâ€™t possibly get the full scope of him.  He has wisps of blond hair and blue eyes and when he says my name I melt.  His personality shines even when he is in a different room.  He is sweet and cuddly and playful and so much fun.  And the reason that he lights up my day and makes me want to appreciate everything that I have is because he canâ€™t walk because of a tumor on his spine and he has cancer and Downs Syndrome and heâ€™s four years old.  He deserves to live until he is 80 years old with a life filled with happiness and productivity and good people all around him.  He deserves so much this little one, this boy that I met.  '),(11545,3022585,'01,August,2004','2004-08-01 00:00:00','   Morehead City, NC.  On the Bogue Sound.  When you hit Morehead City limits you know you are at the beach.  Windows are being rolled down and the warm, musty salt air begins to hit your face.  Your hair begins to curl almost instantly.  You know good grits are on their way too because beach water makes the best grits.  You cross the bridge over the Bogue Sound and see a glimpse of that magical water, part lake, part ocean, part river, all saltwater and controlled by the tides.  There are clams and oysters and sand dollars and porpoises in this water too and you wish you were staying on the sound.  At the light after the bridge you take a right toward pine knoll shores, just a few miles away from Atlantic Beach, from you childhood.    The house sits in front of the ocean, facing what seems to be an endless amount of water.  You canâ€™t believe there is actually something beyond the point that your eye canâ€™t see.  The horizon is so far away that you see whole clouds from top to bottom in the distance.  Some nights, there are storms at sea, you see the lightning, but the rain and thunder never make it to shore.  You wonder what good a storm at sea does, were any boats caught in it?    Some days the wind is so strong on the beach that you have to stay in or go to the inlet or to Fort Macon or a day trip to Beaufort.  Some days there is no wind and the heat comes off the sand like it is willing you to leave.  Some days the rain comes in at lunchtime and moves you from your saltwater post.  Everyday, you want more, more sand clinging to your toes, more saltwater up your nose.  The ocean pummels and turns you out.  You keep going back to it.  In low tide there are tidal pools a short distance from the dune.  These are your favorite. There is enough seawater to float you on a raft and as the tide is leaving there are enough waves to keep you moving.  The tidal pool is the best of all worlds, no thought of a rip current and all the luxury of   On calm nights the moon glows in the sky so bright that the reflection off the water mimics daytime.  You swear you could see by it.  You know that nighttime is the best time on the beach.  The beach is yours then because no else ventures to the beach at night.  You sit on the deck and look at the horizon or the shape of the moon.  You feel like you have a secret that no one else really wants to know about.  How could anyone not love the beach more at night?  You donâ€™t understand the fascination with the daytime, the sun and the rain, when the nighttime brings the moon and crabs and loggerhead turtles.  You know the secret and though you try to share it, others prefer sleep.     Soon you must leave the crystal coast of North Carolina.  No other beaches in the world will compare.  There is simplicity and elegance in North Carolinaâ€™s crystal coast.  There are very few trappings of fancy dwellings or pools seamlessly flowing into the ocean.  There are just beach houses, many are old, and some are new or just well kept.  But the house isnâ€™t important, itâ€™s the beach the matters.  Itâ€™s the weather and the sand and the promise of good shell hunts.  Itâ€™s suntans and bathing suits and rinsing off before you come inside.  Itâ€™s a morning swim then lunch then back again then dinner out.  Itâ€™s beach towels draped over the deck at the end of the day.  Itâ€™s the dayâ€™s catch on the pier or at the surf.  Itâ€™s floppy hats and sun visors and sunscreen and big umbrellas.  Itâ€™s rafts and buckets and shovels and sandcastles.  Itâ€™s enduring the grit of the sand and the stickiness of the water.  Itâ€™s the oldies station on the radio and itâ€™s your family all around.  Itâ€™s the beach in North Carolina and part of you doesnâ€™t ever want to leave.  '),(11546,3022585,'01,August,2004','2004-08-01 00:00:00','   A rental for a week for your family, but a house to someone else.  The evidence of the family who owns the house is all around.  The selection of beach reads on the bookcase in the hall, a small sampling of games with missing pieces in the living room, pictures of the grandpa who worked his whole life to buy the house, artwork from local artists, and the choice of curtains that donâ€™t match the furniture.  You enter a rental house like you enter a hotel.  You know it is not yours yet you will treat it is such for the next week all the while ever thinking about the family who calls it â€œour beach house.â€    There have been houses that donâ€™t lock things away.  The fishing pole of the grandpa who bought the house when he retired hangs on the wall beckoning you to use it.  So, youâ€™re father does and on the first cast into the ocean he breaks it.  Itâ€™s too bad the original grandfather is dead or your father would have apologized to him.  Your whole family is a bit embarrassed, but your motherâ€™s friend who rented the house said we were free to use the poles so your guilt is lessened a bit.  Still, too bad about ole granddad.  There are the houses that seem to want to share the experience with you.  They leave out a guest book and brochures about the area.  You glance at them.  There have been visitors from as far away as Ohio or as close by as New Bern.  Your brother calls the travelers from New Jersey and New York Yankees.  You call them Yankees too, but not out loud.  They tell stories of their weeks, how much their family loved the house, the good weather they were fortunate to have, and they note their favorite restaurants.  You wonder why they choose to write what they write.  The guest book is quaint, but you are happiest about the tide chart on the bulletin board, now you can start your vacation.    Exploring the house, you canâ€™t believe that people share their space so openly with others that they will never meet.  The coffeepot and spoons and beds and pillows and sinks and bathtubs and deck chairs and lamps are all communal property.  There are fragile things about and when your niece knocks over a lamp and it shatters feel sorry for the family again, but then you think that the ownerâ€™s expect these things; beach houses are for children too.  You wonder how all the house was stocked.  Did the ownerâ€™s really like the dark, dense fabric on the overstuffed couches or was the whole ensemble on sale?  Did they pick up the appliances at a thrift store or are the leftovers from the family 1970s orange kitchen?  Your family discusses the flaws in the house, in the decorations.  They have ideas on how to make it better, what to paint or how to arrange the furniture.  You are quick to remind them that it is better than any of their beach houses.  Oh yeah, they donâ€™t have beach houses.  No one is amused.    Some houses have a washer and dryer and this is a luxury.  You do a million loads in one week, more than if you were at home and partly just because you can.  You go home with clean clothes; this you realize is quite nice.  Other houses have a linen service and beds are made for you and towels are provided.  You feel a bit pampered by this, but youâ€™d rather have the washer and dryer back.      There are houses with air conditioning so strong you end up with a stuffy nose going from hot to cold to hot to cold.  There are houses where your family breaks the AC and has to have the unit replaced midweek.  You are again a bit embarrassed, but realize you are just helping with the maintenance of the place.  Then there are houses with no air conditioning at all, just ceiling fans and the hope of good ocean breezes.  These are your favorites.  These remind you of camping in childhood.  This is what the beach feels like to you, heat cut by cool breezes off the water.  Days when there is no breeze you thank God for the invention of ceiling fans.  It takes a good day to get use to the heat and the breeze, but when you do, you donâ€™t ever want air conditioning at the beach again.  You sleep in little clothing and one sheet, you like the simplicity of this type of house.    In your mind you are always building the perfect beach house.  It is oceanfront with enough bedrooms to house you immediate and extended family.  It is not new, it is old and well worn, and there is a lot of wood.  The oceanfront deck is partly covered and partly open to the elements.  There are rocking chairs and a hammock.  There is a table for eating and playing cards.  A walk stretches to the dune where just before the stairs begin there is another small deck with seats built in and a shower hose for rinsing off.  The house has most of its bedrooms on the first floor.  Upstairs, at ocean view and deck level, there is a family room and kitchen and dining room in one space.  There are many large windows.  The dining room table is big enough to sit 12 and there are folding tables for extra company.  You want your whole family here and you create a space that allows for it.  Couches in the living room turn into beds.  There is a good stereo with speakers throughout the house and the deck.  Oldies beach music is always on.  Every bedroom has a well-made ceiling fan and can be closed off to the rest of the house so that windows can be opened at night.  There are beach chairs on the deck and in the garage, waiting for the sand.  They are wooden with cloth backs.  There are rafts and lifejackets and paddleball sets and kites and bachi ball.  The house is always full of people.  '),(11547,3022585,'01,August,2004','2004-08-01 00:00:00','   They wake up early in the morning to play together.  They are little girls with smocked dresses and bows in their hair.  They run around with high-pitched voices learning to share the toys they have brought, to the beach, to a family vacation.    Little girl cousins are a special thing.  They are automatic playmates and best friends.  They share genes and are beginning to share their lives, their growing up.  They clamor to the beach with all their might, they play in the sand for hours, filling buckets, dragging water from the surf, and hunting for shells.  They are adorable in their bathing suits and their skin tans easily even through super strength sun block.  Their fathers or mothers take them into the water and hold them safely from the waves.  The tidal pools are a perfect depth.  Rafts are their best toys, but sometimes it is just their hands and the sand and saltwater.  They get dirtier than they have ever been and never seem to notice the grittiness, the stickiness.  They donâ€™t want to go into the house even after hours of playing, but they love to rinse off with the hose, a semi grown up shower for a little girl.    Watching your nieces you relive your cousin time at the beach, at Salter Path Family campground, on the Bogue Sound.  Your playground was the sound and your favorite things were floating on a raft on the calm water and clam digging.  You and your cousins where obsessed with shells.  You hunted them tirelessly throughout the days on the beach or at the inlet at the north end of the island.  At the end of the shell hunts, you surveyed your treasurer.  You memorized the shapes and names, the olives and whale eyes and baby cradles and cigarettes.  It is years later that you realize, browsing through shell books that some names were made up by your mothers, sisters.  The next morning, you woke up early to set up shop.  You took boxes your mothers had packed in, turned them upside down, and laid old towels on top.  You set out your shells by the side of the road in the campground.  You sold seashells by the sound.  You were thrilled with every purchase.  After a few days of selling, you and your cousins and mothers would walk the path of the campground to the real shell shop near the main road.  You were amazed with the polished shells.  You bought cleaned sand dollars and gold-rimmed hermit crab shells and rope bracelets.    Your cousin still loves shells and houses them in glass jars in her home.  You donâ€™t need them around you anymore, but when you see shells or revisit the beach or see little girl cousins playing, you wish for the days at Salter Path, for the shelling and sound.  You wish for you cousins and you hope the new cousins will long for their beach days and each other when they are older.  '),(11548,3022585,'01,August,2004','2004-08-01 00:00:00','   -tutor precious cancer boy -workout -lunch -clean car, take it to be vacuumed, use oxyclean to scrub stains from coffee spills, organize all the teaching supplies and put in the way back so as not to encroach upon passengers in the back or front passengers seat any longer -organize office so that there is actually a desk to use as a desk -wash clothes so that you can even walk into the office -setup airport so that I am not confined to my living room anymore -dust and vacuum because the dust bunnies are breeding, fiercely  -write thank you notes to R and B, and B because I stayed at their house and/ or received a gift from them like 3 months ago (or longer) -make CD for K and J because they rock and they need the â€œIâ€™ve Come Along Wayâ€ by Michelle Shocked running through their heads as they drive around L.A and also feel my pain -wrap wedding presents for L and for S because L is already a Mrs. and I need to give Sâ€™s present to my parents this weekend -email about bach. party for K because invites need to go out soon and we donâ€™t even know what city itâ€™s going to be in (I suck as MOH, aka Maid of freakingHonor) -somewhere along the line, when I canâ€™t take the stench anymore, take a shower  As you can see, a busy day ahead.  Letâ€™s just see how much actually gets done.  And no, I have nothing else in my head to write about.  Sorry.  '),(11549,3022585,'01,August,2004','2004-08-01 00:00:00','   â€œ urlLink Stuart Weitzman  .  OhMyGod we have to go in.  Theyâ€™re having a sale.â€  I say to K.    Poor K has no idea what or who Stuart Weitzman is. â€œSure,â€ she says and we enter.  The store on Rodeo is small yet two salespeople greet us at the door.  They actually talk to us and welcome us and tell us about the sale.  We have just come out of Ralph Lauren across the street where I found, but did not buy, a pair of purple leather mules on sale for half price at $495.  I knew Stuart could be much more affordable in department stores, but I expect the Rodeo store to have sales where shoes were still $200- $300.  I am wrong.  Every shoe on the sales rack is $75.  â€œHot damn,â€ I think â€œI just might buy something on Rodeo Drive, baby.â€    K and I try on several pairs.  I find some fabulous black satin high heels with an impossibly cute bow covering the toe.  K finds some strappy white heels that she considers for her wedding day.  We each buy nothing, coming back to the reality that even at $75 a pair; we donâ€™t live a Stuart Weitzman life.  We leave, the salespeople shockingly bidding us a good day.  We are tired by this 2nd hour on Rodeo.  We have seen JosÃ© Eber outside his salon, a woman clutching her toy dog in  urlLink Chanel , huge canary diamonds at  urlLink Van Cleef & Arpels , a rail thin blond model standing as a bored yet beautiful greeter at  urlLink Dior , and the back, private entrances to  urlLink Armani  and  urlLink Gucci .  We are heading toward the car, parked in the sun in a free space in front of someoneâ€™s small bungalow and perfectly manicured green lawn.  I am thirsty, having downed my bottle of water noisily in Gucci while walking among the fur trimmed dog beds covered in Gs.    We think we were done with the pseudo shopping and the mouth gaping staring and then all of sudden we walk across Plexiglas ovals in the sidewalk.  Below the â€œwindowsâ€ were human size ant holes with manikins ironing or just posing.  There are three of them, two girls, one boy.  â€œWhat the hell,â€ I think.  K and I look above us and there is no sign, just a massive metallic rectangle as big as a Mack truck above us.  People are milling about in the opening to this, store?  The entrance is set back and it is all open, there are no doors, but there are two huge human sculptures with no heads resting just under the ceiling.  They seem to be holding up the second floor yet they donâ€™t quite touch it.  K and I walk in, we donâ€™t know what else to do, we are transfixed.     Inside there are concrete walls and before us a massive staircase that spans across the room leaving narrow â€œhallsâ€ along either side.  We walked down the left side and set into the wall there seems to be a concierge of sorts, a young brunette manning a phone.  â€œIs this a hotel?â€ I think.  Just past the conciergeâ€™s window there are low tables and set into the concrete walls are lit shelves at hip level.  Things are strewn about.  Bags, shoes, just stuff covering everything, but with no sense of order.  I look closely at one of the bags and see the distinguishable triangle with the word PRADA.  I turn, â€œK, weâ€™re in  urlLink Prada ,â€ I gasp, my face looking quizzical and amazed at the same time.  K doesnâ€™t say anything.  We keep walking.  We come around to the back of the room and there is another set of stairs mirroring the ones in the front yet opposite so that the staircases make a large triangle in the room.  We decide to go up, but have to navigate around shoes and the legs of manikins with no torso.  â€œWhat the hell,â€ keeps running through my brain.     At the top of the stairs we turn left and enter a room that runs like small hallways just above the space below so that on the right and left are narrow halls and in the front and back are slightly wider rooms over the stairs.  The walls are green and look like swiss cheese.  But, there are no smooth surfaces on the walls.  There are only holes.  I donâ€™t understand it and I follow Kâ€™s hand to the wall.  Itâ€™s plastic, green plastic swiss cheese, â€œWhat the hell?â€    The rooms are created with the swiss cheese walls on one side and glass on the other to keep you from falling down the stairs.  There are metal racks of clothes along the swiss cheese walls with exactly 5 items on each.  Thin, tan, bleach highlighted men in black walk around carrying an item or two.  They are barely noticeable in this room, the fade, into theâ€¦air.  When we turn the corner and hit the larger front room I spy the dressing rooms on the right side, above the stairs, I see two little girls playing inside, through the glass.  I think, â€œI know Prada is weird, but who tries on clothes behind glass, in front of strangers?â€  Then, one of the girls hits something on the wall and the glass goes frosted.  â€œWhat.  The.  Hell?â€    After I touch some $1500 mini skirts, we leave.  K and I donâ€™t understand what we have just seen.  Was it real?  It will be the topic of many conversations with Kâ€™s fiancÃ© that week and it will be the first story I tell friends when they ask about L.A.  Iâ€™ll say, â€œIâ€™m going to tell you about Prada, but you wonâ€™t understand.â€  And when Iâ€™m done telling them about the swiss cheese walls and the frosted glass theyâ€™ll say things like â€œWeird,â€ or â€œAre you kidding?â€  And Iâ€™ll tell them that â€œyes, it sounds weird, but you still donâ€™t understand.  You really have to go.  You have to experience it.  Go to Prada on Rodeo.  Itâ€™s worth the plane ride.â€  '),(11550,3022585,'01,August,2004','2004-08-01 00:00:00','   I used to collect quotations.  Iâ€™d culled them from quotation books or magazines or friends or even the daily email from Oprahâ€™s website.  Iâ€™d write them in a flowery printed journal in my best handwriting.  I wanted to the believe the lines like â€œKeep your face to the sunshine and you cannot see the shadow,â€  but, Helen Keller never saw a shadow though she lived within one.  I wanted her words to show my how to not only look at the sunshine, but surpass it and become it.  She did, why couldnâ€™t I?    For years my favorite quote was â€œWe are all in the gutter, but some of us are looking at the stars.â€  When I found it I thought Oscar Wilde was my soul mate, that he understood my plight.  I even wrote it in fancy scrawl across a painted wooden board that held a curtain rod and hung it over 2 different bedroom windows.  But, looking at the stars does nothing for trying to reach them and if you donâ€™t come out in the daylight, the gutter must be a hot and miserable place.  â€œYou must do the thing you think you cannot do,â€ wrote Eleanor Roosevelt and she was right.  You must in fact do the thing you think you cannot do.  She was strong and brave and I trusted her words.  She was the original Nike slogan, â€œJust do it.â€  These words pulled me through many papers and exams in college, but nothing, nothing in college or life could prepare me for what lay ahead.  My first years teaching I was energetic and naÃ¯ve and ready to take on the world.  I loved what I did and felt absolutely called to do it.  I thought that my teaching kids with autism was who I was.  I thought it defined me and that everyone should revere me for the work I did.  Those first years threw me into a tailspin of hatred and anger and bitterness and disappointment and fear and illness.  I complained constantly and to anyone who would listen, but friendships and life were slipping away.  Quotations and words of wisdom wouldnâ€™t pull me through and there were many nights I just cried and prayed.    I finally sought help and through a counselor learned that my life was full of unhealthy patterns and that I was the only one in control of what could happen in my life.  I didnâ€™t believe her and for 3 years cried and complained about all the things that happened TO me.  Then I began to use her advice and let the control go.  I had to learn to let go and forgive.  I had to learn to let life take me where it would.  I had to learn to do the thing I thought I could not do.  I faced the shadows and the gutters and Iâ€™m still learning to leave them where they are, in the past.  You canâ€™t see the sunshine or stars clearly until you face their opposites.  And no one and no quotation will help you truly get there.  They can point the way, but you, you have to do the digging and nail biting eye to eye combat with all the bad things before you finally appreciate all the good ones and let happiness if not land, at least aim, at your doorstep, everyday.  '),(11551,4170329,'19,August,2004','2004-08-19 00:00:00','   sooooo, i was at my apartment and randomly frank knocked on the door and said that i had to go with him because we were meeting crissi, jose, jake, mandy, carrie etc etc. at this pond/swamp/river/lake thing. so i jumped up and got ready and ran 14 miles with frank to the pond place. we had to climb through some bushes, buildings and rooftops to get to the \"special\" pond. when we arrived everyone was swimming with alligators, crocodiles, worms, fish, tadpoles and etc. i was scared shitless but eventually i jumped in when jose said \" you\'re not down\" so i had to prove him wrong. everyone was having a grand time and the cool thing to do in this pond was try to get a worm to wrap aound your neck to be a necklace. i was the only one who succeeded at the task. all of a sudden, frank freaked out and said i had to go with him and we had to run back to his house for a minute. being the dumb girl i am, i said \"ok.\" so we ran back to his house only to change his son\'s diaper. AAAAAAHHHHHHHH.. oh god!!!!!!! then he was freaking out about why he wasn\'t making any interest on his well fargo bank account since he was writing over 400 checks a month. what the fuck?!?!?! so needless to say, we went back to the pond and on the way we brought his son and frank asked me to be his baby\'s momma. i nearly died but said yes to him.       it was at this precise moment i woke up in a panic this morning. what the hell am i dreaming about. i make myself nervous.      in other news, my dad came this past weekend and took everything i own besides my t.v., computer, some clothes and some bathroom products. my apaprtment is bare. i am wuite nervous about moving home and will miss my independence in davis tremendously. i honestly think that i enjoyed my time in davis. it was rough in the beginning but it got easier with time. i am sooo fournatute to have parents like mine who supported and semi-paid for me to be here. i look forward to being closer to my friends in murrieta. and hope that the hotties come out where every they live and show their faces to me so i don\'t have to dream about resorting to be someone\'s baby momma.       i hope all goes well with my move back to so cal, it will be a big change for me and i just keep praying that my family will get better with time and my dad finds the strength to hold on each day. i love you all and hope you have a great day!!   '),(11552,4170329,'08,August,2004','2004-08-08 00:00:00','   ok.. well i am at work right now.. secretly using the compudora to hack away online like  atry junkie!!! i am so over this job.. (take this job and shove it).. only about 13 more shifts to work here and then im out like trout.. in other news, my dad is coming to move my stuff into storage this weekend. i\'m kinda nervous about how this will all go down. pray for me. damn, my family is CRAZY!!!! mi vida familia!!! my life is really boring in davis these days. i seriously sleep for about 16-20 hours a day lately. no joke at all here. i don\'t know what the deal is. i think i may becoming a bit depressed cuz my life is hectic and my future is sooo uncertain. where am i living in 3 weeks? will i get enough hours at the school district? will i even like it? what do i want to be when i grow up? who are my real friends? blah blahhhhhhhh...  well 17 minutes until closing time, so i gotta get my work done.. peace out!!   '),(11553,4170329,'08,August,2004','2004-08-08 00:00:00','   hey world... this is me!!!! i decided to be like every other kid in the world and get a journal to write it... danielle says i forget shit too often therefore i should write it down and hopeflly my memory will slowly return.. i\'m out!!!   '),(11554,3596048,'10,June,2004','2004-06-10 00:00:00','   --Thanks--    Just think about all those things,  That you\'ve done,  To make all those hurt gone,  Thank you for all..   Thanks for being such a nice friend to me..  For helping me through all of my problems..  Thanks for being there when I need you..  For comforting me when I\'m really down..   For all that you have done,  Just for me..  You just make me whole and happy..  And for that you are so cool..Ãœ   Think of those people you light up  And made their day complete..  Maybe being friends with you is just a coincidence..  But after all, this all make a sense..   I\'m really happy to have you in my life...  Now, at times when you need me,  Don\'t hesitate to come to me,  And ask for my help or advice..   I\'ll always be here for you,  I\'ll always help you through,  I\'ll always be your friend..  No matter what happens, I promise..Ãœ       --Life is Worthless Without You--    I still remember,  since the day I met you..  I told myself,  you\'re the one I\'ve been looking for..   your my soulmate,  your my life,  and I want to spend my life with you..  a guy who will love me too..   if you only knew,  how much i love you..  if only your here with me..  you\'ll know how you make me smile..Ã¼   can this happen that you and me will be together?..  can this be true?..  that you\'ll fall for me too..  or can this be a dream forever?..   I hope time will come,  that you\'ll love me too..  I promise i\'ll never give up..  coz life is worthless without you.. Ã¼  '),(11555,3596048,'10,June,2004','2004-06-10 00:00:00','   I don\'t know how long I will love you..  I don\'t know how long this feeling would last..  I don\'t know how much love i can give you..  I don\'t know when will i give up on you..  I don\'t know if i\'ll ever leave you..  I don\'t even know if i could do that..  all i know is as long as there is time and love,  i\'ll always be here for you..  i\'ll cherish every moment i\'m with you,  coz deep in my heart, i know it\'s true..  that i love you, i really do..  I know this may end,  you may leave me someday..  but i\'ll always look back..  on the mem\'ries you had left behind..  the happy moments i had with you,  that once in my life you have made me whole..  how you have made me smile..  especially at times when i\'m really down,  your there to comfort me and loved me unconditionallly..  I will always remember,  that once in my life i have loved some one like you..  I will always remember,  that once in my life i have learned how to love because of you..  I will always remember,  how precious life is since you came into my life..  I will always remember,  how lovely it could be when you love someone unconditionally..  if only i can love you forever, i will..  but i can\'t.. coz ya know wat? sometimes,  we just should give up someone so that we could be happy too..  but ya know what? I will never forget you, I PROMISE!Ã¼  coz you were the best i\'ve ever had in my life.  I LOVE YOU, I REALLY DO!!!!!Ã¼  '),(11556,3958372,'24,July,2004','2004-07-24 00:00:00','   When I think about this Air Supply song, I don\'t know what to surmise. Have I lost in love or am I lost in love? I would like to think both. I kind of feel both today. I\'m someone who likes to live for the day. Do I want love to be suffering, as someone said to me today - love is suffering! No way, I want to live every day happily in love. But is it humanely possible when you are constantly pursued by lonliness. Life is a constant fight against loneliness, despair, negativisim especially when you are living in a small country like Singapore, far away from your own country, far away from the dynamism of Mumbai city, your own folks and your friends. Out there, when your love lets you down, at least you have your family and friends to fall back on. Here there\'s constant despair, and an urge to escape from this island of loveless life. Today is my wedding anniversary, one of the most dull and boring days of my life, hubby just gave his courtsy call to wish me in the morning from distant America, without any display of love or emotion. He expects me to feel on top of the world with that. I\'m human after all, so how do you expect me to react? It felt like the longest day of my life, I surfed the net, read the news, tried to sleep in the afternoon, couldn\'t sleep plagued by persistent negative thinking, popped an anti-depression pill to deaden the anxiety in my brains, slept finally, still many hours to kill, went to the gym and exercised for an hour, came back, ordered a Pizza, had a shower, and then made myself a strong drink which I had with the Pizza, still a lot of time to go. I\'ll probably read a book putting on my radio softly, hopefully sleep would invade my senses and make me happy and reposeful. Tomorrow is another day, hopefully it\'ll begin and end happily. '),(11557,3958372,'20,July,2004','2004-07-20 00:00:00','   Gloomy weather in Singapore. Strangely it\'s a feeling-less day. Don\'t feel towards anything strongly. Early in the morning I got a rejection letter from the newspaper that I had&nbsp;applied for a reporter\'s job.&nbsp;That is, after doing considerably well in the written test. Well that\'s my destiny I guess. We Indians are pretty&nbsp;good at passing off our failures to destiny.&nbsp;It\'s time to seriously rethink what I want to do with my life. Be a writer! I think I can pass off reasonably well writing erotic novels and letters. I guess am pretty good at it because I write erotic letters everyday to my sweetheart. I have no qualms or inhibitions about describing every little intimate detail of&nbsp;lovemaking, our&nbsp;fantasies, being&nbsp;far apart from each other. Every morning I log on to&nbsp; write a letter to him.&nbsp;Today it took me a while to get over the rejection letter and write him a long loving email, as if things were normal. Well&nbsp;in the midst of despair, love is our only hope I guess. You have each others love&nbsp;to fall back on. Love can be life\'s greatest support. No more PMS blues, life is calm and tranquil again. Time to go and exercise in the gym. Need to get rid of all the bad energies. I want to be a loving, hopeful and optimistic person and pass all the positive energies&nbsp;to my loved ones.&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; '),(11558,3958372,'18,July,2004','2004-07-18 00:00:00','   I had a terrible start to the day. No one loves me,&nbsp;life on earth is full of sufferings!! Well am on a total pms mode today.&nbsp;I see bleakness everywhere. This is my perception of life life today. Who knows, things might change tomorrow, in this world of constant mood swings. I had a bad start to the day. Hubby calls me from America to voice his discontent over an email I sent him last night- a one liner of course \"Are there no phone services in America?\" He hadn\'t&nbsp;called us for a quite a while. Well I guess&nbsp;it was enough to elicit an immediate respone - an sms and a phone call in the wee hours of the morning. Of course he expects me to write nice loving emails to him everyday but he doesn\'t bother to do the same for me.&nbsp; But&nbsp;what did I really gain from it??? Nothing, but pain. So from today, I\'ve decided not to be emotionally dependent on T. I\'m not going to be bothered whether he writes to me or&nbsp;phones me&nbsp;or sms\' me. I\'m going to build an invisible wall where no man can penetrate and hurt me. That\'s&nbsp;one thing. One incident sparks off the other. I had called up&nbsp;my distant sweetheart hoping to hear some kind&nbsp;reassuring words and&nbsp;hoping to feel loved and wanted but only met up with his stony silence. Never show your vulnerability to someone who is in awe of you because half the time he might wonder who you&nbsp;really are!! Alas! I\'m beginning to believe in that almost hackneyed cliched saying, \"when you laugh, you laugh with the world but when you cry, you cry alone.\" I guess that\'s true.&nbsp; Believe me, you are better off&nbsp;not communicating&nbsp;your grief to anyone. It\'s&nbsp; better&nbsp;to drown &nbsp;your&nbsp;sorrows in alcohol, a book, a nice&nbsp;movie or music. At least they wouldn\'t let you down. Unleash your grief and disappointments in your diary and when things start looking up, start day dreaming. What would&nbsp;I do if I have some money? First of all, buy a Sony PD 170 camera, then buy a the latest Nikon SLR Digital camera, buy the most sophisticated&nbsp;laptop with video eidting package and then weave around an enchanting world of my own, just live for myself and my little beautiful daughter. Some amount of materialism is not bad&nbsp;I guess.&nbsp;It keeps you going in life. So who says, you need a&nbsp;hubby or a lover to be happy! Happiness is within yourself, it\'s important to love oneself first!!&nbsp; '),(11559,3958372,'17,July,2004','2004-07-17 00:00:00','   I\'m an incurable romantic, I can\'t live without love and the thrilling adventure of romance.&nbsp;I want to be in love forever because it makes life so exciting.&nbsp; Imagine life without love, it\'s like a desert, or rather love is like the oasis in the desert of life, the constant tribulations that you come across.&nbsp;From now onwards am going to write down my random thoughts on love and what it can do to one\'s life, especially my own life. I will write in all honestly all the intimate details of my experiences, my feelings and observations. &nbsp;&nbsp;&nbsp; '),(11560,3958372,'16,August,2004','2004-08-16 00:00:00','   Love is often subjected to one\'s terrible ego. What is love I ask? Is it pain and suffering or is it inflicting pain and suffering upon your loved ones because of one\'s ego? I\'m terribly confused. I love my husband very much. I often feel insecure about him and in a rage I end up doing all the wrong things, in my fits of jealousy and possessiveness. I want to be the centre of his universe and want to be the most important person in his life. But when you are married for several years, things change. You are expected to mellow down and become mature, have less expectations from your partner and give him space. I haven\'t been doing that. I don\'t think it\'s impossible to change. Can I resurrect my marriage? I think it\'s worth giving it a try. It\'s never to late to realize one\'s mistake and try to mend things. I regret all my silly mistakes and seriously want to undo all of them. Is there a way out? '),(11561,4108638,'11,August,2004','2004-08-11 00:00:00','   The fall semester has started for me, I begin the internship process and the final semester towards my masters. Granted I have to complete financial aide one more time. Let me tell you how stressful asking for more money again and again and again and again.       I am just tired of fighting with the university to get anything done or to just graduate. Why does it have to be so dam hard to get the financial aide necessary to graduate and do something with my life. I thought the schools purpose was to help people achieve a higher level of education. Yet I face more stress just to get the money to come to school then I have ever faced before. Hopefully if I do everything right I will get some aide by November. Ok that is today\'s gripe.       On the happy note. I start earning my hours towards internship I have to earn a total of 600 hours, 250 of them in direct client contact. Mathematically that comes out to close to 40 hours of work each week, and 20 of those hours in contact with clients. I think that it will be a great experience and I know I will learn a lot. So wish me luck. :p  '),(11562,4108638,'03,August,2004','2004-08-03 00:00:00','   urlLink    My dog at home&nbsp; urlLink  '),(11563,4108638,'02,August,2004','2004-08-02 00:00:00','   Long Post time...    Ok so this weekend I traveled to     Dallas     to check out a school for my Doctorate in Psychology.    This school was not what most people would consider as a real school.    This university advertised itself as a school for working professionals to continue to work on their education.    Therefore, I get there and start talking to the admissions person about the program and what it entails.    I find out that just like every other psychology doctorate program it requires a 2-year practicum and a 1-year internship.    Ok so that isnâ€™t too bad, until I find out that I would have to give up whatever job I was working in so to work full time in the practicum/internship unpaid for, three years.      THREE YEARS...No money, no moolah, no cash flow, no income......gee...no food?  How is anyone supposed to work for three years with out being paid?    Now do not get me wrong but no one is going to go into the mental health field and make tons of money.    However, I already have student loans, why would I want to get more student loans.    When will I ever pay them back?    I do not get why practicum/internship students are not paid we are providing a service just as valuable as a mental health practitioner that is out in the field.  I understand that we are still learning, but how are people going to become psychologists if they can support themselves.      Ok so that is my bitch about not being happy about the Dr. situation.    Now for the reason why I am complaining so much.    Since the eighth grade, I have been working towards being a psychologist.    I have heard and been reinforced from my folks, friends, and acquaintances to be a psychologist.    Now I am currently graduating in December with my Masters in Counseling, in     Texas     that will allow me to do the exact same work as a psychologist.    Therefore, what the issue is is that I feel like I am not living up to what I have dreamed about for a long time.    Kind of like a weird sorta of failure to myself.    I think overall I have made the decision to not pursue my doctorate for a while, because I am tired of being broke and living hand to mouth all the time.  I am ready to be an adult and enjoy being independent financially.    All right, time to be done posting.    :)  '),(11564,4108638,'01,August,2004','2004-08-01 00:00:00','   I have three clients that I am going to see tonight at my school\'s clinic. Normally I wouldn\'t be to worried, but tonight I am being observed by a different professor then I have been observed by. This prof that is about to observe me, doesn\'t seem to have a very high opinion about me from previous classes.       Now I feel and have been told that my skills with clients are right on the money for my experience level. I am worried about how she will evaluate my skills. I have a very difficult series of clients to work with tonight.       Well on another note... A good friend of mine has been accepted at UNT to begin working on her masters in counseling. I am really happy for her. More blogging tonight :)  '),(11565,4108638,'01,August,2004','2004-08-01 00:00:00','   This is my first post. :)  It has been an incredibly long week. I have worked several 14 hour days at my internship site. I am a mental health counseling intern. Basically that means I work for next to nothing. I think elephants at the circus make more then I do. I know I sound complainingish... but I really do love my work. Currently I am working with children and adolescents. Sometimes it really just means I play games with kids all day, but you would be suprised how much that can make them feel better about themselves or thier situation.   Soon I will be leaving E. Texas and moving to the big city of Dallas, which scares the crap out of me. I will be a real life serious adult. I want to leave E. Tx but I really dont want to leave the friends behind. More on this subject laters.  '),(11566,3644129,'28,June,2004','2004-06-28 00:00:00','    I   HATE   Reminiscing.   Don\'t get me wrong, I am a very sentimental person. You\'ll know by the junk I have at home! Hehehe!   Kidding aside, I  AM  Sentimental. The thing is, I hate it when I reminisce the days of old.   Why? Because it makes me so sad to think that most of the people I cherished those moments with, are gone. No, they\'re not DEAD! They just lead different lives now. So different, we don\'t even get to talk much when we bump into each other. There is just NOTHING in common anymore. And it SUCKS!   I don\'t think I have changed much. I still enjoy what I have enjoyed in the last 10 years of my life. (Which makes my psychological age, 14 years old). I like watching light-hearted and mainstream movies, Pop/Alternative music, giggling, small chit-chats... Juvenile, I know. I also enjoy just bumming at home, sleeping and watching the tube.   I guess the only thing that changed in me is my desire to be in a more serene, calm, quiet environment. That is the only thing that matured in my personality, among other things.   Most of my \"old friends\" are either in their final years of Medical School, doing their Residency, working in some uptight office somewhere, having babies or have different friends altogether.   Time is also a factor which separates me from my \"old friends.\" They don\'t have time, and I have lots of it. It\'s either that or their schedules don\'t jive with mine. I would gladly give my time to them, though, if they did the same for me. Unfortunately, that is not the case.   In short, our   PRIORITIES   have changed a lot. I still have them on my priority list, however, it seems that I have disappeared from theirs. To think I used to consider them as my BEST Friends or my Closest Friends. And that bit is what   HURTS   the most.   That\'s why I live my life one day at a time. I keep myself from getting attached to people, in fears of having to suffer the same consequences of my \"previous life.\" I guess the   ONLY   friend you can ever rely on in your life is   YOURSELF  . At least, when you look back and reminisce on your times with yourself, you don\'t get disappointed that you lost that person in one way or another.   This does not mean that I\'d throw away all those friendships that I have previously formed. That is not the case. I just think I was the one trashed.   Here\'s to finding the Diamond in the Rough...'),(11567,3644129,'26,June,2004','2004-06-26 00:00:00','    I did something dumb this week.   I forgot to submit my Log Records before the Cut-Off.   I can hear people saying: \"So What?!?\"   Well, here\'s the catch. It is office policy to submit your Log Records (Logs In and Out) before each Cut-Off Period (the 9th and 24th of every month), otherwise you do not get your Pay Check for that Pay Period. In short, I will not be receiving my Pay Check at the end of the month.   Ooohhh... Tough!   Well, there is a   Logically Sound and Utterly Stupid   story behind that.   First, let\'s do the   Logically Sound   reason: As I checked my Log Records for this Cut-Off, I realized that my Manager uploaded the wrong schedule, failing to amend the Rest Day Swap that my Co-Worker and I applied for. I also had to apply for a Log Modification, since the system did not Log me out on one of the days.   After my manager did the modifications and necessary adjustments, one of my Logs disappeared! The system deleted it! I immediately informed my Manager and she tried to fix it. She even had to call the System Administrators since it wasn\'t budging. When it was finally fixed, we were relieved.   Now, the   Utterly Stupid   reason: I wasn\'t really myself that evening. My car\'s brakes almost gave out on me while I was on my way to work. It was raining hard, too. I was dead scared when I got to the office, knowing something might\'ve happened to me if my brakes died. I also wasn\'t able to get some decent sleep that morning, when I got home.   When it was almost time for me to go home, I was thinking about the car and where we could get the car\'s brakes fixed. I didn\'t want to drive that car anywhere without the assurance that it was safe. I waited for my dad to come and pick me up,  completely forgetting  that I had to submit my Log Records.   To make matters worse, I went on my Rest Days. The Logs were due on a Thursday, and I was on my Day Off! Unfortunately, there are no exemptions to the rule.   No Log Records, No Pay.   That sucked...   BIG TIME!!!   And to think I have to pay 3 Bills this month, besides my daily gas, toll and food allowance!   Anyway, I can breathe a little bit easier now since my Manager assured me that the Pay I am supposed to receive at the end of the month will be credited to my July 15 Paycheck. Yipee! Hopefully, I can pay everything up by that time!'),(11568,3644129,'20,June,2004','2004-06-20 00:00:00','    That is my occupation on my  urlLink Friendster  account.   Why?   I am not sleep-deprived, nor is it by choice. There is a perfectly logical explanation to it... I work the Graveyard Shift. Well, that is, until July 4th.   I work in one of the biggest Contact Centers here in Makati. I handle the travel concerns of some of the wealthiest people in the Philippines. Calls rarely come during my shift, but it could still screw a person\'s body clock really bad.   One of the things I really hate about the GY Shift is my Rest Days. Although they may seem longer, it doesn\'t help my body clock much. It actually adds to my irregular sleeping patterns.   You see, on my Rest Days (which are on Wednesdays and Thursdays), I sleep at the \"Regular Time,\" which is around 10:30PM and wake up at 8:30AM. This screws me up on a Friday, when I have to go to work at 9:00PM. I am awake the whole day of Friday, so when I get to work I am often drowsy or sleepy.   Well, my Co-Workers and I share the same problem. And since we hardly get any calls, we tend to sleep off half of our shifts! Oftentimes, people are surprised that we sleep at work! Not that our manager really minds, she doesn\'t! As long as we answer the calls that come in, it\'s fine. It\'s a nice trade off.   At least, when I drive home from work, I do not feel drowsy or sleepy on the wheel!   I just hope I can get a more Regular shift next time around. July 4, come fast!!!'),(11569,3644129,'20,June,2004','2004-06-20 00:00:00','    Yesterday, June 19, 2004, my cousin gave birth to a 7 pound 11 ounce baby boy who they named Jude Matthew Corpin-Madriaga. He\'s the long-awaited baby in the family (after my cousin Izabella was born in 1998). The first great-grandchild of my Paternal grandparents.   I was not surprised by his birth since I dreamt of his arrival just 2 days before. In my dream, Ate Mhin (my cousin) gave birth to a very healthy baby boy with a head full of hair, beautiful round eyes and a cheery disposition. When I woke up, I immediately asked my cousin if she already gave birth and told her about my dream. She wished my dream would come true since they were praying for a boy for their first-born. And true enough, they have an adorable baby boy!   The birth of Matthew was not easy for both Matthew and Ate Mhin, though. Matthew\'s umbilical cord was wrapped around his neck, which prompted the doctors to perform a C-Section on my cousin. It is said that babies born with their umbilical cords around their necks are intelligent and extremely lucky. I hope it\'s true!   Another cousin (Ate Mhin\'s brother) sent us a Multimedia Message with a picture of Baby Matthew. He\'s such a looker! Very much like the baby in my dream. I am looking forward to seeing this little bundle of joy. However, they are staying in La Union (a province in the Northern Luzon) right now, and it will take us around 5 hours to get there.   I know he\'s going to be a very pampered baby since all of his aunts and uncles just could not wait to dote on him! I hope he grows up a well-mannered, God-fearing and cheerful man.   Hmmm... When would I have a baby of my own?!? Hahaha!'),(11570,3644129,'16,June,2004','2004-06-16 00:00:00','    Okay, let\'s start the ball rolling.   I\'ve been following the NBC Reality/Makeover show   Starting Over   these past few weeks and I\'ve found it very interesting. Never, since my   Real World: New Orleans   phase, have I been so into a Reality Show. There must be something about this show that really caught my attention.   I saw this show one very sleepless night, as I came home from work.   Starting Over   is being shown on a Local UHF Station here in the Philippines at 1AM (with Replays at 5PM). Since nothing good is on cable, I watched, for an hour every night, as 6 very diverse ladies try to make a resounding difference in their lives, with the help of 2 \"Life Coaches.\"   At first, Andy caught my attention. She was your local drama queen. She was very aloof towards her housemates, which started the tension between her and housemate Nyanza.   As the show wore on, I found myself looking forward to seeing the struggles of Lori. Lori is a young woman who is having an extremely hard time moving forward since the death of her husband. Hers is a very tragic story. Tragic, yet romantic. I could totally identify with her.   Lori seems to me, is like a sheltered flower - afraid that the sun and rain would harm her. She is horrified by the idea of establishing social relationships with people outside her family. Like me, she finds comfort in social gatherings involving her family. Her goal was to cultivate lasting friendships as well as finding a new man to fill her life.   I also found myself in another Starting Over housemate, Christine. Christine\'s Life Goals in the House was to lose weight and find happiness outside of her relationship with her teenage daughter. At first, she was \"flying under the radar,\" so to speak. I didn\'t know what she was in the house for, or what her goal was. That was until I noticed that she was given assignments to post an add at Yahoo! Personals and she started running with a Coach/Instructor. Man! Now she\'s the person I can GENUINELY identify with. I need to do what SHE\'s doing!!!   How I wished there was that kind of show produced here in the Philippines. It could really help a lot of us \"mixed up\" people roaming around the Metro. I could certainly use the help and the emotional nitpicking!   My dream thing, now, is to meet these women and tell them how much their stories, triumphs and struggles have affected my life. I mean, I am totally seeing myself learning from them and from what they\'re doing. Cheesy as it may seem, but this Reality Show is just so inspiring! Go out and watch it!   For more info on   Starting Over  , log on to  urlLink StartingOver.tv .'),(11571,3644129,'16,June,2004','2004-06-16 00:00:00','    Today is the start of my new life...   No... Not   THAT   drastic of a change. I just thought that I ought to do something about the things going on in my cerebral cortex. So why not publish them on a BLOG?!?   I\'ve been meaning to write a blog for the longest time. It just seems that writing a Journal is not enough. Or maybe, I\'ve just been so darn lazy to write something on a notebook.   Anyway, here\'s to the start of my new life!'),(11572,3644129,'30,July,2004','2004-07-30 00:00:00','    I had the   Weirdest Weekend   ever!   First off,   my   \"weekend\" is not the same as everyone else\'s. My weekends are Tuesdays and Wednesdays (for the whole month of July). So that\'s one of the weird facts (except for those who work in Contact Centers and 24/7 Operations).   But, what made my weekend extremely bizarre (at least, for me) was that everything seemed to go wrong!   Last Monday, as I ended work and started my weekend, I found out that my car had been towed! That sucked big time! I was unable to move my car to a Pay Parking area because I was stuck on the phones. I couldn\'t go out to move the car except during my Lunch hour (which I took at 1:00pm). However, instead of doing so, I  had  to use my Lunch Hour to pay numerous bills and go to the Bank. So when I finally was able to go down and move it, it was already gone!   Good thing I remembered that one of my Dad\'s cousins was a Barangay Council Member. I called her up and told her what happened. She agreed to help me, after much lecturing. So I went home on a Shuttle which was,&nbsp;for me, better than driving home.   Early the next morning, I went back to Makati to meet with my Aunt who accompanied me to where the car had been impounded. She gave them the Towing Ticket, which had already been signed by the Barangay Council Chairman. After they checked on my License and Registration, they released the car to me without having to pay the usual PhP1,000 fee.   Everything went on smoothly after that. Or so I thought.   The very next day, Wednesday, I woke to find that we had no running water. I checked on our Pressure Pump, to see if it was turned on. It was not. Usually, if the Pump was off, the water could not be stored in our Overhead Tank. I called my Dad to see if he turned it off. He said that he did because the Pump\'s motor had been running the whole night and yet there was no water pressure coming in. He had to shut it off or else the fuse might pop. He asked if I had paid the Water bill, to cancel out the possibility that out Water had been disconnected. I had already paid the current water bill, so it was out of the question. Then, I checked the Water Meter outside of our house. It was gone!   I immediately went to the Waterworks Office to ask what had happened. Why our Meter was taken. (They usually take the Water Meter if your service had been disconnected). I told them that I had already paid our bill. The Engineering Department of the Waterworks told me that they had not Disconnected our service and that there have been numerous reports of stolen Water Meters in our area. We have just been victimized by these unscrupulous Meter thieves!   I called my Dad to tell him about what happened. He spoke with one of the Engineers there and asked if we could have our Meter reinstalled in a different area. They made arrangements for the Meter to be replaced and reinstalled, and everything was settled.   That was my Weird Weekend. Well, most of it anyway.'),(11573,3644129,'25,July,2004','2004-07-25 00:00:00','    Last night was the BEST! It was just like Christmas and New Year\'s all over again!   Our family rarely gets together since most of us graduated College. But last night, we had a small get-together when my cousin, Ate Mhin, and her new family (her husband and new baby) came to visit.   Ate Mhin got married last year, and since she and her husband, Ernest, moved to Cainta (near Antipolo), we rarely saw her. So rare, that most of us never even saw her Pregnant!!! And when she gave birth last June, to a healthy baby boy  (see my BLOG Entry on our new nephew&nbsp;  urlLink   Matthew    ) , she was so far from where we all lived (about a 5-6 hour drive North of Manila). So, no one from our family really saw this NEW Family, until last night.   They came back to Manila just last week, and we took the opportunity to organize a mini-get-together dinner.   I picked them up straight from work and we headed to our Aunt\'s house in ParaÃ±aque.   When we got there, most of our female relatives were there (our Grandmother, our Aunt and 2 other cousins). Of course, everyone was excited to meet the new baby and greet the proud parents! The house seemed to be in a buzz while everyone marched into the house. Everyone was cautious not to go near the baby if they had coughs or colds, etc.   Slowly, the other relatives arrived. Ate Mhin\'s dad, Tito Ray, arrived with&nbsp;Ate Mhin\'s&nbsp;youngest sister, Yan-yan. She\'s just around 9 or 10 years old. Tito Ray has been looking forward to last night since he learned that we will be having that get-together. He hasn\'t seen his Grandson since he was born, hence, the excitement.   Then came our other Uncle and Aunt, Tito Dave and Tita Mildred with her mom. Tita Mildred and Tito Dave have been struggling to have a baby since they got married, but they are still holding on to their faith that the Lord will bless them with a Child in HIS time. For the meantime, they are extremely elated about the new addition to the family.   We now have 2 Babies...  Izabella Cristina  (who is turning 6 years old this year) and  urlLink  Jude Matthew  . Izabella is the youngest among the cousins, born with a 15 year gap from her sister, Aissa. She was our  \"baby\"  for the longest time, and now, she\'s the   \"ATE\"   of Baby Matthew.   The get-together was rather fun, with everyone hanging out and catching up on each other\'s lives. Doting on the baby was another highlight of the evening. Most of us, even the usually jealous Izabella, wanted to hang around the baby, watching his every whimper.   As the night progressed, me and another cousin, Mae (Ate Mhin\'s sister), talked to Ate Mhin&nbsp;about having Matthew.   She told us what she went through while pregnant with him. She said that her neck, armpits, knees and tummy became dark and she had developed stretch marks on her abdomen. While she didn\'t really have any unusual cravings while pregnant, she said she was quite irritated with her husband. Hahaha! Poor Ernest!   It was quite nice to have the family together like that. Catching up with each other, joking around, talking about politics and food, etc. It would\'ve been more fun if the family was complete. A couple of our cousins were indisposed of (2 were in the US, the other 2 were at work, and yet another was in Laguna) and&nbsp;our other Aunts and Uncles&nbsp;are also in the US.   Oh well, I guess that time will also come soon enough.   Today, they\'ll be having Lunch together in our Aunt\'s house again, but unfortunately, I would not be able to go since I have work until 3:00PM. *sigh*   I took a lot of pictures from last night and posted it on  urlLink  Snapfish.com   .  If you want to view it, just click  urlLink  here   .'),(11574,3644129,'24,July,2004','2004-07-24 00:00:00','    I am very agitated right now...'),(11575,3644129,'22,July,2004','2004-07-22 00:00:00','    About 2 weeks ago, I discovered a new recipe. I learned this recipe in my favorite warehouse club... Price Smart.   &nbsp;'),(11576,3644129,'21,July,2004','2004-07-21 00:00:00','    The past few evenings, I\'ve been having Asthma Attacks while asleep. It\'s so bothersome, that, at times, it prevents me from going to work the next day (to think I have to wake up at 4:45AM). The worst attack was last Monday morning (around 3AM). What made it   really   worse was the realization that  I left my Inhaler in my office drawer . I had no medicine available in the house!!!   I finally decided to go to the Emergency Medical Clinic that afternoon to consult with a Doctor (since my mom was out of the country). The doctor became concerned when I told her that I hardly get any attacks, until these past few evenings. She recommended that I take some medicine to widen my airway before bedtime&nbsp;and some   STEROIDS  !!!   I was concerned at the sound of Steroids because I know that once you get into steroids, you\'ll have to maintain it or you\'ll experience some unwanted side-effects. The doctor, however, assured me that I should not be concerned about the side-effects since the medicine,   Medrol   (Generic Name: Methylprednisolone), didn\'t have any, and it was to be taken orally.   At the Clinic, I&nbsp;also took some Ventolin via a Nebulizer. That medication cost me! I had to pay PhP310.00 for it! Good thing I had some cash with me at that time. After that, I had zip!   It was all good, though. I was able to sleep well the next few evenings... Until this morning. I had a slight attack while I tried to sneeze, as I woke up. It eventually faded, but I was still concerned. I just hope that after I take all these medicine, I will be able to sleep comfortably again.'),(11577,3644129,'18,July,2004','2004-07-18 00:00:00','    The other day, I was typing a BLOG entry, but&nbsp;I&nbsp;suddenly had to finish a transaction for a client. When I came back, I couldn\'t retrieve it anymore! I lost it! Man!!! That sucked!  &nbsp;  I wanted to re-type it, but I totally forgot what I typed in my entry.  &nbsp;  As soon as I get inspired, I will try to post another one again...'),(11578,3644129,'18,July,2004','2004-07-18 00:00:00','    I\'ve never been to a Planetarium.  &nbsp;  Can you believe it? I\'m about 2 months shy of my 25th Birthday and I still have not stepped a foot inside a Planetarium... Ever!  &nbsp;  I distinctly remember a time when my family was driving by the Rizal Park and I saw the Planetarium. I asked my dad to bring me there, and he said he would. Well, it\'s been a lifetime (of sorts) since, and I have never even seen the building up close.  &nbsp;  Most of my friends asked if I\'d be willing to take on a stiff neck. Well, I\'d be willing to, given that it would truly be an experience. (Not that I\'m an Astronomy buff or anything).  &nbsp;  Living in the City, I have never really learned about the stars, since&nbsp;city lights obscure the night sky. Whenever I watch movies with people looking up at stars, identifying each constellation, etc., I become quite perplexed as to how they manage to identify and find these heavenly bodies.  &nbsp;  I would really want to be acquainted with the heavens so that I could also navigate around with the stars as my guide.'),(11579,3644129,'17,July,2004','2004-07-17 00:00:00','    I think I am an Anti-Social  .   Weird, huh?!? I\'m a perky, spirited person, yet   I   view myself as an Anti-Social.   Why do I think so? Well, let\'s just say that being a  homebody  turns you into one. Maybe, it\'s also because I have set a whole lot of limitations on myself that hinder me from being a Social being. Though I am  straining  to break free, it seems like I can\'t.   I have come to this realization months back. It\'s not that I don\'t have friends or anything... I   have   a lot. It\'s just that, sometimes, I just couldn\'t relate anymore. It seems like we walk on different planes, with different levels, and I could reach them. This is why I hardly go out on social get-togethers. Catching up has never been my forte. Unless it\'s on a one-on-one basis.     For example  : I\'ve been to 2 get-togethers this past week. The first was a   Lunner   (Lunch/Dinner) Date with one of my Closest Friends in my professional life. The second was Dinner with 2 of my former  \"Lambs\"  just this evening. And all I can say is:   I am such a bad company.     Don\'t get me wrong. I  enjoyed  the Lunner with my friend, Nabsie. It seemed like we never lost touch (it\'s been more than a year). We were still making each other crazy. Or at least,   I   was driving   her   nuts! Haha! That\'s the way we\'ve been.   CariÃ±o Brutal   is the name of the game. I\'d always be teasing, and she\'d always be the victim. I don\'t know, but we had a lot of fun. We\'re just like kids!   The Dinner with my  \"Lambs\"  is a whole   different   story. No offense to Czacza and Lia, but I was at a loss the whole 2 hours we were together. I felt that I did not belong in that table. I can\'t blame them. It\'s been roughly 5 years since we\'ve met, and a lot has changed since I got in touch with them last (They\'ve been in touch since). We also have a 2-year age gap, and we run in different circles.   Though I seemed misplaced and disoriented, I was   still   glad to see them. What\'s more, is that, I just feel so blessed that these 2 people are still close, despite the fact that they were mere strangers before the YLSS in St. James, which brought us all together in the first place. I felt blessed because, in a way, I became one of the reasons why these 2 became close.     Misplaced. Oblivious. Estranged. Introverted. Austere. Isolated.   I feel that a lot lately, especially around friends who I\'ve lost touch with for some time.   Getting plunged into a situation where I am   \"obliged\"   to socialize is uncomfortable to me. I feel so vulnerable to my feelings of rejection and dissociation. I feel that the more I engage in a conversation, the more I don\'t make sense.    Aaarggh!! Somebody help me, please!!!'),(11580,3644129,'16,July,2004','2004-07-16 00:00:00','    I HAVE A LIFE!!!     &nbsp;  In the past few years, I have let a lot of things (and people)&nbsp;take over my life. Things that, in the long run, did not benefit me well. More often than not, it gave me unbelievable grief. I became very desolate, withdrawn and I even started to secede myself from situations and conversations that had anything to do with the people or things that have burned me in the past. It made me think that, maybe, I didn\'t have a life beyond those things or people.'),(11581,3644129,'15,July,2004','2004-07-15 00:00:00','    Just wanted to congratulate my sister, Pam, on her Capping today!!!   My sister is in Nursing School right now (her 2nd degree) and will be going through this ceremony for Nursing Students. It\'s very much like a graduation ceremony, except that she\'s not graduating yet. The school officials are going to install my sister into the Nursing world by donning her with a Nurse\'s cap.   So, Congratulations, Pam!!!'),(11582,3644129,'10,July,2004','2004-07-10 00:00:00','    I LOVE ITALIAN!   -- Italian Food, that is! (Got you there! Hehe!)   I actually had Pizza the other day. It was a New York style Pizza: Big, Tasty and Filling! It\'s so different from the more \"commercialized\" pizzas that are available in the market. I don\'t like Pizza Hut or Shakey\'s Pizza, but I love the ones that come from Sbarro, Yellow Cab and Brooklyn. I could also settle for the Pizza that you can buy at your local Price Smart.   My personal favorite is the White Pizza from Brooklyn. It\'s a made with different types of White Cheeses. It also has a hint of garlic, which makes it more tastier. Mmmm! Makes my mouth water just thinking about it.   But the Pizza I crave for the most is the Chicken and Garlic Gourmet Pizza from Round Table Pizza. Mmmm! They don\'t have Round Table Pizzas here in the Philippines, so I\'m just dying to have one right about now!   Food has been my passion for some time now. Isn\'t it obvious? But I am trying to cut it down to a minimum. When I\'ve been good, I reward myself with a slice of Pizza.   Another Italian food that I just   LOVE   is Pasta. Ravioli. Penne. Fetuccini. Tortellini. Those pasta noodles are more delectible with Authentic Italian Sauces. Puttanesca. Amatricianna. Arrabbiata. Carbonara. Pesto. Aglio e Olio. Neapolitan. Formaggi.   Since I was a kid, I loved spaghetti and lasagna. But until I tried my Aunt\'s Puttanesca, did I crave for more AUTHENTIC Italian Cuisine. I still, up to this day, have not cooked your usual Spaghetti Sauce (ground meat in tomato sauce). I\'ve been cooking and experimenting with a lot of different sauces to find the perfect taste. Hopefully I can find one that is just perfect.   Now, I\'m hungry!!!'),(11583,3644129,'07,July,2004','2004-07-07 00:00:00','    At long last! I am now living a   NORMAL   Life! For a month, at least.   Today, I started my 6:00AM-3:00PM shift at work. It\'s not   THAT   Normal for most people, but at least I\'m awake when everyone else in this Time Zone is!   It\'s not that hectic as I presumed it to be, but it  is  quite interesting.   On my way to work, I saw dawn start to break. The darkness giving way to the light. Poetic! Well, at least to my mind it was. The setting was not romantic. Unless you think the Highway, pothole-ridden streets and towering skyscrapers are romantic, that is.   I haven\'t seen the sun rise since my carefree teenage years or on Summer Vacations. It\'s quite a refreshing sight. I just got used to the darkness a lot. Well, it kinda dominated my life for a while. Literally, and figuratively.   I hope that as I see the sun rise on my way to work, I could also feel the light pierce through my dark days.'),(11584,3644129,'05,July,2004','2004-07-05 00:00:00','    For the last few days, I\'ve had pestering headaches, allergies and colds that have dampened my otherwise satisfying day.   I had already consulted a doctor for my headache, which he has diagnosed to be Vascular in nature. He gave me some prescription drugs to take for 3 days, which, stubbornly, I did not take. My headaches have since subsided, but last night, a heavily pounding headache plagued my evening at work. Fortunately, I was already in the office when my headache started. I took some Mefenamic Acid to numb the pain, and slept it off while at work.   Another ailment that has been bothering me are my allergies. 3/4 of the day, my eyes are raw, itchy, swollen and watery. Almost like conjunctivitis. I try to put some eyedrops to ease the itchiness, but when I wake up, my eyes are still watery and itchy. I\'ve been suffering from this kind of allergy for approximately 5 years now, yet I have not yet adjusted to it. Who would?!?   Having colds, for me, is as common as taking a bath... Literally. Whenever I take a bath, I get colds. Weird, but true. It eventually goes away, but it\'s always there.   Hmmm... I wonder when that day will come when all these \"diseases\" leave my body. Hmmm...'),(11585,3644129,'02,July,2004','2004-07-02 00:00:00','    I\'ve been on my Rest Days these past 2 days and the semi-highlight of my \"weekend\" was food and babysitting.   Ahh, yes...   FOOD!    Glorious Food! Hot Sausage and Mustard! While we\'re in the Mood, Cold Jelly and Custard...  Wait!!! I\'m getting carried away here...   Anyway... On my way home from work last Wednesday morning (June 29), I was famished, not having eaten my dinner and all. Since we were experiencing the last blows of a raging typhoon, I thought it was the perfect weather for some \"old fashioned\" hot cocoa and some hot pandesal. The perfect breakfast food for an otherwise gloomy and stormy weather. Good thing the bakeshop I buy pandesal from was open early. I bought a couple of freshly baked bread and some milk. Champorado, I thought, would\'ve been more perfect, but then again, I already gave up eating rice (and was too lazy to cook).   When I got home, I proceeded to cook the hot cocoa. Since it was the \"old fashioned\" one, I had to boil the cocoa and add the milk and sugar ontop of the stove. It was well worth it. I had a great breakfast! Yummy! I also shared it with my dad when he woke up. Unfortunately, my sister had early classes and was already out when I arrived.   I spent most of my Wednesday feeling that I ate a lot! I swear! Gluttony is an ever-present temptation and feeling everytime I go on my Rest Days. Since I don\'t do much when I\'m at home, I just cook up something and eat whatever esculent provision is available. Although I eat moderately in about 1 1/2 times a day, I feel like I\'ve consumed a TON of food, and it sickens me.   Thursday arrived after 14 hours of slumber. You read it accurately... 14 peaceful, accumulated sleeping hours. I felt deprived of sleep the previous day, so I retired to bed early, waking at mid-day. For brunch, I decided to have Bacon and Eggs.   I loved eating Bacon and Eggs. It reminds me of eating at Denny\'s and IHOP. Whilst cooking the Bacon, the oil that emanated from the Bacon fat splattered onto my right hand. It actually burned 3 spots on the back of my hand. At first, I thought nothing of it, but as I cleaned my hand from the hot oil, I saw 3 separate reddish spots that stung. These reddish spots have since turned into a dark brown color, much like bacon. Now, I think I would rather stick with the other \"safer\" cold cuts.   Right at this very moment, almost 7 hours after my last meal, I still feel supremely glutted. I can literally feel 5 lbs. adding onto my weight. Hopefully, I am just imagining things.   My life is anticipating normalcy in the next few days, as my new Working Hours are set for the Early Morning and Mid-Afternoon Hours for the month of July. I am looking forward to eating normally, though still in a reasonably moderated fashion.'),(11586,3644129,'11,August,2004','2004-08-11 00:00:00','    NOTE :  Before you get the wrong impression and have any weird thoughts running in your heads, I would just want to clarify that I am Definitely, Absolutely, Catergorically, Indubitably, Doubtlessly,   100% STRAIGHT  .'),(11587,3644129,'03,August,2004','2004-08-03 00:00:00','    I am severely peeved at my father, right now. (Was there ever a time that I wasn\'t?!?)'),(11588,3921274,'30,July,2004','2004-07-30 00:00:00','   Didn\'t see it, as I was out doing other things (a friend\'s birthday and alcohol were involved), but I\'ve read a majority of it from Kerry\'s website (availabe  urlLink here   if your interested). Later, I plan to watch the video they have of it as well.  Look for my comments on the speech after work and after my hangover goes away. So far, some quotes from the speech that really caught my eye were:   \"We\'re told that outsourcing jobs is good for America. We\'re told that new jobs that pay $9,000 less than the jobs that have been lost is the best we can do.  They say this is the best economy we\'ve ever had.  And they say that anyone who thinks otherwise is a pessimist.  Well, here is our answer:  There is nothing more pessimistic than saying America can\'t do better.\"    \"And as President, I will bring back this nation\'s time-honored tradition: the United States of America never goes to war because we want to, we only go to war because we have to.\"   Impressive lines. Look for more later. '),(11589,3921274,'28,July,2004','2004-07-28 00:00:00','   That along with \"John who?\"  were the thoughts going through the mind of everyone in attendence at the Democratic Convention after Barack Obama spoke.  I have to agree.   Remember the name people, you\'ll be hearing a lot from him in the near future. '),(11590,3921274,'27,July,2004','2004-07-27 00:00:00','   Wow . . . that was incredible. The speech he just gave might have been even better than the one Clinton gave last night. The Democrats sent this young politician up to the plate in a big game and he responded by knocking the ball out of the park.  American, stand up and take notice. This guy is going places.   Update:  urlLink Here is a transcript  of the speech. '),(11591,3921274,'26,July,2004','2004-07-26 00:00:00','   You know, there\'s been a lot of talk that if John Kerry doesn\'t win in 2004, the Democrats will run Hillary Clinton in 2008. If I were the Democrats, I would go for broke and make  urlLink Barack Obama  her running mate.  That would make Hillary the first woman president and Obama the first black Vice-President.  Of course, we all know what Chris Rock said would happen if we ever had a black Vice-President. '),(11592,3921274,'26,July,2004','2004-07-26 00:00:00','   urlLink Here is a link  to a transcript of the incredible speech Bill Clinton gave last night at the Democratic Convention.  Quick side note: I\'m note a total Clinton lover. I disagree with many of the things he did in office. But Clinton was the last big success the Democratic Party had and last night he showed he still mattered. That\'s why I\'m so high on him right now.  Another side note: I\'m by no means a rank and file Democrat. As I said before, I\'m an indepedent. I feel the party during the Clinton Administration tried to be too much like the Republican party and during the Bush Administration did nothing to stop the havoc Bush reaped upon us and our constitution. However, I have aligned myself with the Democrats for this election because John Kerry is the best shot we have at beating Bush. And unfortantly, Bush must be defeated at all costs . . . '),(11593,3921274,'26,July,2004','2004-07-26 00:00:00','   My job sucks. I want a new one, but I don\'t have enough experience to get a new one just yet. It\'s kinda funny, 8 months ago I was on the brink of a breakdown cause I couldn\'t get a job. I was unhappy cause I was unemployed.   Now I am employed. And guess what? I\'m still unhappy!! Does life really suck this bad or am I just one of those people who is never happy with anything in life?   Either way makes for a pretty shitty way to live your life!!  Coming Tomorrow: I rant on relationships, love, and marriage. Well I think it\'s coming tomorrow. As long as I don\'t find something better to do. '),(11594,3921274,'18,July,2004','2004-07-18 00:00:00','   Can Bill Clinton run for president again?   I know the answer to that rhetorical question, but man, the speech he gave tonight at the Democratic Convention was amazing. In under 25 minutes, Clinton layed out a clear political philosphy for the Democratic party, differentiated Kerry from Bush, and finally made the Democrates stand for something. In short, he did in 25 minutes what the Kerry campaign hasn\'t been able to do their entire campaign. These are the things the Democrats failed to do in 2000 and utimatly cost Gore the election.   Sure Clinton may have skeltons in his closet, but I still think if the Democrats want to beat Bush in 2004, they need to utilize Clinton to the fullest. Tonight was a shining example of how Clinton was able to captivate an American public and win 2 terms in the White House against a hostile Republican party ready to destroy the Clinton Administration at all costs.  Ronald Reagan, eat your heart out. Clinton is the real great communicator. '),(11595,3921274,'18,July,2004','2004-07-18 00:00:00','   I went to see the movie  Fahrenheit 9/11  on opening day. I found the movie brilliant, thought provoking, and extremely scary. As expected, the movie was met with harsh reviews from the usual suspects of right-wing, fact-less, you hate America cause you criticize the president pundits.  However, there was one review, written by Journalism Professor Robert Jensen, that really caught my eye. The reason? It was from a liberal, left-wing journalist who not only called Fahrenheit 9/11 a \"bad movie\", but also accused it of being conservative!!  The article (which can be found  urlLink here ) makes several interesting points about the movie that I think are at least worth considering (though I don\'t necessarily agree with most of them.) I could comment on the article line-by-line, but I prefer for people to read for themselves and decide on their own how they feel about articles such as these.  One part of the article I will comment on however, is this passage, which appears midway through the article   Is the administration of George  W. Bush full of ideological fanatics? Yes. Have its actions since  9/11 been reckless and put the world at risk? Yes. In the course  of pursuing those policies, has it enriched fat-cat friends?  Yes.  But it is a serious mistake  to believe that these wars can be explained by focusing so exclusively  on the Bush administration and ignoring clear trends in U.S.  foreign and military policy. In short, these wars are not a sharp  departure from the past but instead should be seen as an intensification  of longstanding policies, affected by the confluence of this  particular administration\'s ideology and the opportunities created  by the events of 9/11.   The statement is no doubt controversial and most likely confusing to those without any  background on the point of view people like Jensen hold. So allow me to provide some for you.  People like Jensen believe that since the end of World War 2, the United States has used it military not to defend our freedom or preserve our way of life, but instead to further US economic interests in countries which already lack a significant US economic presence.   For instance, Jensen would argue that the war in Iraq had nothing to do with liberating Iraqi citizens or protecting us from Saddam Hussein\'s supposed arsenal of \"Weapons of Mass Destruction, but was instead waged to eliminate a regime that was hostile an American presence in their country and replace it with one much more receptive to American companies wishing to do business in the Middle East.  There is some evidence to support this claim. In 2000, a conservative think tank known as the Project for the New American Century (PNAC) released a report entitled  Rebuilding America\'s Defenses: Strategy, Forces and Resources For a New Century   (the report, which is quite long, can be viewed  urlLink here ).  The report calls for an invasion of Iraq in order to remove the current regime and set up a base station \"for a significant American Presence in the region.\"  It proposes gaining control of Iraq\'s oil resources not for economic reasons, as suggested by many anti-war protesters leading up to the war, but instead largely for the political purpose of maintaining control over the region.  After all, it quite obvious that the individuals in power in the Middle East are ones who control the oil reserves.   By gaining control of these oil reserves, the plan asserts, the United States could then gain significant political influence in the Middle East. By gaining political influence, we could then install a hand-chosen governmental regime that is favorable to US companies wishing to do business in the Middle East, but have thus far been unable to do sobecause of hostile governmental regimes.  What is even more startling about the plan is that the PNAC was produced for people like Vice-President Dick Cheney and Secretary of Defense Donald Rumsfeld. Rumsfeld\'s assistant at the Pentagon, Paul Wolfowitz is listed in the report as one of its authors.   People like Jensen would argue that plans such as the one mentioned above are nothing new and have been perpetrated by every Administration in the White House since World War 2.   Such ideas are worth considering in any worthwhile discussion of US foreign policy. I plan to write more on topics such as these in future posts. For now, enjoy the review and give Jensen\'s comments on US foreign policy a second thought. And by all means, go see Fahrenheit 9/11. You owe it to your country! '),(11596,3921274,'13,July,2004','2004-07-13 00:00:00','   This is my first attempt at blogging. Although I read many blogs throughout the day, up until a few weeks ago, I never really thought about writing a blog of my own. This came as a surprise to many people, seeing as I make my living designing websites. Whenever someone would ask why I didn\'t have a site, my response was always \"I have nothing to say.\"   That, however, was extremely incorrect. The truth is, I have a lot to say about almost everything. After all, I\'m an information junky. I read every book I can get my hands on. I watch at least 2 hours of TV a day. I spend at least an hour a day scouring the internet for what I consider the real news (not the McNews mainstream media sources.) Hell, I even listen to talk radio on occasion (yes I know, I am getting old!)  At some point during the process of absorbing all of this information, I tend to form strong opinions about the issues. Up until now, the only people I had to share my opinions with were friends. And let me tell you, listening to me rant about US foreign policy over beer at the local bar is not my friend\'s idea of a good time. So now I\'ve turned to the internet as my vehicle for expressing my opinions.   So now perhaps a brief introduction of myself is in order. I am a 23 year old male graphic designer who lives alone. I am currently single and have many opinions about the current dating scene, many of which will appear on this site at some point.   I am a college graduate with a Degree in Mass Communications. I am very interested in media and the role it plays in our everyday lives. Someday, I plan on returning to college to obtain a masters and possibly a doctrine in the field. The media will be a major area of focus in my writings  I am also a political junky. I consider myself a political independent who is radically to the left on some issues but ridiculously to the right on others. Throughout the course of my blogging, you will learn about many of my views on a wide variety of political issues. As a quick taste, I consider George W. Bush one of the most dangerous presidents we have ever had.   But don\'t think this  blog will only be about media and politics. As I said before, I absorb a wide variety of information on a lot of subjects. What I write about will depend on what topic  or news items catches my interest on that particular day.  I plan to include a little bit of everything on this site.  Other than that, I would say my life is pretty ordinary. I go to work, I come home, I do some laundry, I watch some TV, and I go to bed. On the weekends and sometimes during the week, I enjoy going out with friends, having a few beers, and throwing a few games of darts. I don\'t have much luck with women, so I plan on being single for quite a while.  You will learn more about me as time goes by. I hope you enjoy this little experiment of mine. I welcome feedback, both negative and positive, on anything I write about.   Well thats all for now. Keep checking back for more. I\'ll talk to you soon, and please send me feedback on anything thats on your mind!! '),(11597,3921274,'13,August,2004','2004-08-13 00:00:00','   Yea, right and I\'m a ten-foot green alien.  This courtesy of  urlLink Kos:    Since 2001, President Bush\'s tax cuts have shifted federal tax payments from the richest Americans to a wide swath of middle-class families, the Congressional Budget Office has found, a conclusion likely to roil the presidential election campaign.  The CBO study, due to be released today, found that the wealthiest 20 percent, whose incomes averaged $182,700 in 2001, saw their share of federal taxes drop from 64.4 percent of total tax payments in 2001 to 63.5 percent this year. The top 1 percent, earning $1.1 million, saw their share fall to 20.1 percent of the total, from 22.2 percent.  Over that same period, taxpayers with incomes from around $51,500 to around $75,600 saw their share of federal tax payments increase. Households earning around $75,600 saw their tax burden jump the most, from 18.7 percent of all taxes to 19.5 percent.   It\'s important to note that the agency conducting the study is a governmental agency working for a Republican congress. Furthermore, the agency is run by a Republican economist.  It never ceases to amaze me how people say they\'re Republicans cause they\'re poor and want to pay lower taxes. That just goes to show you the power of media and their ability to place partisan talking points into everyday discourse.  PS: If any of you Bushies read this and would like to try and explain this one, please leave a comment. I would love to hear your explainations. '),(11598,3921274,'06,August,2004','2004-08-06 00:00:00','   A post over at  urlLink This Modern World  tells of a threat to America far greater than Saddam Hussein\'s supposed weapons of mass destruction. And guess what, our so-called tough on terrorists, wartime president sees the whole thing as no big deal. I hope Americans wake up and vote the bastard out of office come November. You can read the post by clicking on the link above or if your lazy by clicking  urlLink here. '),(11599,3921274,'02,August,2004','2004-08-02 00:00:00','   Since Vice Presidential Candidate John Edwards was a trial lawyer before being elected to the Seneate, it\'s almost a given that Republicans will try and link Edwards to wide array of \"frivelous lawsuits.\" One of the lawsuits featured will almost certainly involve the lady who was burned by a cup of coffee in a McDonalds drive-thru.  However, as  urlLink this article  illustrates, there was more to this famous story than meets the eye. Follow the link and keep the real story in mind when your assaulted with Republican attack ads in the months ahead. '),(11600,3921274,'02,August,2004','2004-08-02 00:00:00','   At least, that\'s what  urlLink some people  are saying. Appearantly, according to the various polles conducted following the Democratic Convention, the Kerry campaign has seen between a 4 point boost and a 5 point loss. These, are according to the polling sources, the lowest in the history of these polls.  However, as both  urlLink Atrios  and  urlLink Kos  pointed out, when you examine the internal workings of the polls, things get a bit fuzzy.   Apparently, according to the  urlLink USA Today poll  (which reported Kerry lost 5 points since the convention), 26% of those polls gave Kerry\'s convention speech a favorable rating. This is a higher approval rating than Bush\'s speech in 2000, which was so well recieved, it gave Bush a  17 point lead  over Al Gore in the polls.  It gets better. Apparently, the edge Bush held in his handling of the war on terrorism actually fell from 18% to 12%. Bush is selling the war on terrorism as the central issue of his campaign.  Something strange is going on here. Wonder if the polls were tabulated by the Flordia Election Commission. '),(11601,3999305,'29,July,2004','2004-07-29 00:00:00','   ~the night is alright..well it could be better..hmm..yea..RiRi came back from her tuition every thursday with a guy fetching her home..its not that i dont trust her,but the thought of it makes me feel kind of strange..i wish i could be the one fetching her instead..that would make me feel more secured and to know she\'s safe as i kinda dont trust any guy with the girl that i love except her father..But i trust my girl alot.. ~aNyways my day was kind of boring,the weather was quite good..windy and stormy abit..as usual i fell asleep but woke up to study maths as i needed to..kind of complicated so i need to improve on that..so many fomulars!! a guy could lose his mind..really!..but well i had to.. =) i couldn\'t make a promise and break it esp to my RiRi..i did it for my own sake too..yea also,i marinated a chicken this afternoon and cooked it at night since no one was home..you could say i was inspired by a chef \"jamie oliver\" i think i spelled it correctly.. LoL..the looks of his dishes are great..i could smell it from my home even though its on tv..it smelt great when i tried it out but not exactly how it was done..loL..hmmm also,tonight RiRi nvr gave me any hugz or MuackZ..still waiting for it..but nvm..guess i wont get any.. =(  ~well nothing much happened today i  guess..ill type another one tmr..byez '),(11602,3999305,'28,July,2004','2004-07-28 00:00:00','   ~I read my horoscope in the morning and if i remembered clearly,it says \" Ur in for a Great time today\" well of course i did, RiRi gave me a Cuddly and loVing *MuaCk* and lotsa other Stuff..right RiRi? can see you smiling at the com..loL like me!Today was a fine day,lovely music on the way to school,lovely weather,,well kinda..funny day at school too..i burst out laughing after my teacher told us a crappy joke which wasen\'t funny at all..i was laughing pretty sarcastically..that made the whole class laugh!! loL!!!! i made a group of malay friends laugh too..jokes that only we understand..loL!! not kidding..i dont know why the eng teacher always had high hopes on me..announce to the whole class that she wants me to read the best essay in singapore and learn from it..want me to get A1 for it..now im stressing myself out..ahhh but at least i got a great grade for my Eng Oral..!! LoL! hope it helps huh! oh yea..i love playing in the rain when im down too.. oh not forgetting,i love the dark..it keeps me calm inside and removes my sorrows..i live by the dark..but i like the sun alot too..could get skin cancer,thats what all of them say when i stay in it for a long time..Have the urge to go swimming..loL..well ill update it some other time yea? '),(11603,3999305,'27,July,2004','2004-07-27 00:00:00','   I have no mood for anything nemore..i know that i just feel hurt deep inside as i have let her down..so much...by placing all those dumb irresponsible msgges on her blog..am i that stupid not to think twice before placing it there..and now it has to hurt her..im truely depressed,i really need someone there for me and she proves to be the only one..inside it just cant stop burning and hurting me..i wish it could just dose the pain...its beyond description..i ran my 2.4km just for her and got in the third position as it just hurt so much that i dont wanna stop and just feel like running..jus think of making ur loved one disappointed and losing him or her..all i want is to see her smile,that would really make my day when she\'s really sad and can spare me a grIn..all i want to do is to just make her happy and enjoy chatting with me but i always go screw things up..im really disappointed in myself too..but for now...i understand she has to study real hard and i support her decision and will always be there to back her up mentally if not physically..i need to study either..so i truely understand and support her decision..miss you RiRi...wish could give you all the huGz and MuackZ i\'ve missed..nxt time alright? For now,study hard and ill do the same..dont let nething get into ur concentration.. '),(11604,3999305,'26,July,2004','2004-07-26 00:00:00','   Im so hurt tonight..i feel like just ending it all here..but im not going to do that..I must be here for my sweetheart...My heart just needs her and wants to be with her..take care of her..seeing her soo depressed maybe because of me..is soo..painful...i love her..i really do..what must i do to change my life..i feel like dying but my heart wants to carry on..the burn inside is killing me...i feel like just crying at the sight of her being so depressed and running away from me..i know shes down..im trying my best to save her from this painful thing..nvm.. '),(11605,3999305,'25,July,2004','2004-07-25 00:00:00','   Awww..My deaR IrIs Sooo Sweeeeet...makes my heart melt..like MnM\'s chocolates..but faster than that alright!  The reason? cause she said \"Love ya\" aww means so much to me and remove all my stresses and blues straight away! not kidding! if only she said that more....and im gonna break my 2.4km run record just for my sweeeeetheart..really going to do that..ill have her in my mind while running! hAha! Gave me Lotsa \"hugZ\" and \"MuAckS\" and called me \"Darling\" which ill never forget and will hold it deep in my heart..really..so deep that no-one could reach it cept my LiL baby RiRi! ~Neways ,back to my day today..played soccer and thrashed the sorta school team 3-0 HAHAHA!! i feel so mean you know!  we had alot of fun time..and studies! Impt k\'z RiRi? you got Ur darlin to Support you here *holds ur waist then cuddlE REally TIght* ! ill Promise ill study hard for my RiRi too and Never Let my RiRi down alrightz? alSo got back and cooked my own meal..haHA..chicken and vege everynight..lAme huh..thats just me.. loL =P well thats all i have to say.. be back tmr! '),(11606,3999305,'25,July,2004','2004-07-25 00:00:00','   ~ BoRing..well actually i could have gone out but well better save my money to get my Sweeeeetheart IrIs a Wonderful Present..trying my best to save..loL..my day started off Great! cause sweeeeetheart gave me a *HuGz*...every morning i wake up hoping to find my swtheart beside me so i could give her a tight cuddle and a kiss in the morning..awww how sweet that would be..neways..the day after that was usual..Take a cooold shower and freezing myself off..haha but i kind of like that feeling! no kidding! then have my breakfast then 2 hours later my lunch..then study sometime then take a nap.. =P im kinda like a cat..likes napping a little..loL..im also a sleepyhead on schooldays cause sometimes ill wake up late then rush myself! As usual,i miss my IrIs sweeeeetie..haHa it always happen so im use to it.. =P *Kiss On thE Nose* ..im always hyper at night if i go out especially!! you have to see it to believe for urself...If you realise that i dont use the word \"Love\" too much..well if ur wondering why its cause that the word \"Love\" is much tooo precious to use and only \"Love\" can be used on wisely and its not for everyone.. thats all i have to say yea?Be back tmr to update my blog..adios! '),(11607,3999305,'24,July,2004','2004-07-24 00:00:00','   ~thE time was 6am and it was a saturday morning..i could have slept..but i woke up just to accompany my friends to Flag Day even though i was not involve..eventually,i got myself involve..haha when they asked me to help them..it was great! I did lots of stupid And Crazy Things..and embarressed my friends who were following me..by helping myself to the balloons and i collected around 35 at once as the guy\'s there agreed to let me help them give it out..loL!! it was great! but kind of funny as the children came to me and asked me for balloons! they were sooo cute..no kidding! haha..i got alot of cash for my flag day also..lotsa notes! not coins but notes!! Also met lots of new friends..really cool..i wouldn\'t mind going for the flag day again! =D *grIn* i also met my sweeeeetheart IrIs baby..she was studyin at the starBucks cafe while i went there to get a caramel frappucino(or whatever its called..loL)..working hard as i can see..loL! Though i could not send her home..well nxt time i would =) shes always saying \"nah uh\"! loL!! i also fell asleep in the train..was very tired..my eyes was like 3/4 closed the whole time until i left the train then my eyes widened to only 2/4..thats how tired i was..loL bad rite?even when i was standing up in the train,i was even placing my head on the handrail trying to slp..hehe i was Very tired...i seldom sleep in the train..been years..i was in orchard rd since 7.30am in the morning..i should have brought my jacket cause it was really freezing out there!!! i caught a slight cold but it was not as freezing unlike when i come out of the bathroom after taking a bath..loL!! it was as usual,the whole day i was thinkin of my sweeeeetheart Iris baby..ill still be sweet to her and will never not..   *kisS on thE Nosey* thats abt my day i guess! i already missed the day..wish i could go for this again =] '),(11608,3999305,'23,July,2004','2004-07-23 00:00:00','   Why must it be that she is still with this guy...i know it takes time..but well it just hurt me to see it going on..well thats how i feel but i trust her alot still..just hope im could be her\'s instead..its alright i guess.. :) '),(11609,3999305,'23,July,2004','2004-07-23 00:00:00','   ~Ahhhhh..what a comfy bed! loL! i jus came home frOm gym..It was Damn TirinG..but kind of fun! At school was great also as i got pinched alot from my teacher..but well that kinda sucks! loL ! Shes just teenage abusing me ! i also had the chance to listen to my discman..kind of fun as my teacher dont mind! haha it helps me study better also as i had a maths test today..It was A Killer..imagine doing it without a calcalator! but i still pulled through..haha..it was easy but without calculator means death..At the gym we had lotsa fun..doing sit-ups and challenging who could do most.. haha! 45 in one minute for me!!!  challenge me if you could!! nah jus kidding!! i would definately lose!..we played soccer after that also..did lots of move such as \"The Scorpion\"! it was really funny as we havent master it yet and we were like doing handstands all the time with the ball jammed between our legs..not right up there but at the ankles as scorpion moves needs you to execute these steps.. =P but we were failures! Tmr  is gonna be a great day i think...hope it would be fun..kinda sad and hurt too though...why....damn...im just so..sadd...hurt and depressed but ill wait..its just killing me...hope i can be with someone i love.. '),(11610,3999305,'22,July,2004','2004-07-22 00:00:00','   This Is jus my first time writing so im not use to this..haha~Abit abt me? K\'z alrIght,Im jus a blaDer but dont blade often causE got Not MuCh Time.. iM born on March 4Th 1988 which means i get to watch NC16 movies!! LOL!!! i Jus basically like everything but i theres only two things which i Truely lOve.. My BabY irIs and my familY.. Neways,im jus having this strange feeling..but im not sure how to Describe it, Very unEasy but im kinda..well not kinda but extreMely Happy too.. cause i got a My Sweetheart irIs there for me  ~when i wake up In morning,and msg her,she would be there to wish me Good moRning AnD Give me lotsa Hugz! loL!! * gRIn* ..she\'s always there when i need her~ hEr hUgz bring me to life and her MuacKs Jus makes my day better and clear and Monday/Tuesday/Wednesday/Thurday/Friday bLues! sHe brightened my Life eveR sInce she came.. WeLL abt today...Hmm..I jus came home and playeD a Prank On mY neighbourhooD  by plaCing Scotchtape everywhere and Looking at the peoplE walKing throUGh them..iTs jus so FuN Oh i gOt an Early dismissal too as MY eNg TeacHer gave It tO mE and the rest Of my Friends WerE suppOse to StaY back cAuse they NeeD to FiniSh theiR EnG assIgnment..haha..!!  Also,I shouTed At A Few of Strangers from My school(abt 5 of them)BuT i DoNt knoW them, They WeRe Shouting At Us And I Shouted Back aNd they rAn OfF =P loL!! it was NiCe..NewayS iLL writE another Entry Tmr '),(11611,3999305,'17,August,2004','2004-08-17 00:00:00','   ~Feels great to be dismissed early today..looking at all those studying and sleeping in class..loL makes me laugh..makes me grIn evily at them..loL..yes i know im MEAN (sweetheart..lol)... well after that i just hang out with my friends at Mac..chat..and crap and chat and crap..loL..thats kinda me..but well when things gets serious,ill be serious..really..theres always time to be and time not to be..So i went to my friend\'s house and well..you could say we had barbecue in there..you have to see it to believe it..having a barbecue!(loL right LEWIS? Campfire too huh..flames soo big that nearly burned ur house down) loL.hope can do this again..this time make it bigger and hotter (add more vodka!!! trust me!)   Came home,bath and ate and then got an sms from sweetie..but mum at home so cannot talk long..and Someones reading a mystery book i can see (yup im talking abt you sweetie..who else i know really reads a book..loL!)  now talking to my sweetie on msn..can see she very tired..of course for a sleepyhead who nvr naps in afternoon will definately feel sleepy at this hour..right (agree with my baby?)..let me pay for movies can?please..or ill wont bring you shopping!!! loL jk ..we\'ll see when the time comes okie? ok update my blog some other time :D '),(11612,3999305,'16,August,2004','2004-08-16 00:00:00','   ~ZzZzZ..sleeeeepy..its a sleepy day today..i came home..i knocked out while sitting on the couch( see how much of a sleepyhead i am..sit down also can slp)..loL ..sleepy..really..sleepy...the night sooo tiring and the afternoon was good to sleep..loL windy..but you cannot blame a fellow sleepyhead right just like someone(you should know who im talking about right sweetie?loL) !! Baby now talking to 4 friends at once..her fingers and wrist must really hurt alot from all the typing..me just simple..tell my friends ill talk to them some other day and then talk to my baby..Today i really missed my baby more then usual..so much..yst was quite some night but not to the peak of our hyperness right huh am i right?! regarding the romantic places..ill bring you there..its our promise..ok? Me now freezing again sweetheart ( you guys wont know what we are talking abt.our own language right sweetie?) ~ need you here now..can come over?loL teleport here cause i build teleporting machine for you already.. cause im just freezing my ass off by myself..maybe i should just teleport myself there tonight!!! sweetie..just wanna say i miss you and love you..and Leonard...hmmm..dont stalk girls too much...one day they\'ll rape you !!!! hahahahahahaha!! welll thats all i have for today..cant reveal too much abt my day here as im gonna tell more abt it to my sweetheart on the phone.. Update in tmr! '),(11613,3999305,'15,August,2004','2004-08-15 00:00:00','   ~ \'Rain RaIn go away..dont ever come again\' nah actually i enjoyed it lots!! real much! soo cosy and romantic but well cant be romantic cause my baby not there..only my sofa.. and i wont want to be romantic with my sofa..nah uh (one of my baby fav phrase)! and i caught a ghost story on Ch 55..kinda cool!..well actually..its really fun to laugh at ghost storie..(one of my fav pasttimes that nxt time ill teach my  baby and she can accompany me laugh as we watch at my home or the movies!)..but well i kind of fell asleep..as the lights were off and it was reall dark and i laugh til fall asleep!well only three things can make me fall asleep,cuddling my baby,reading a book(sort of) and the weather....well actualy theres more,but well only one can command me to slp..SOMEONE..okok my BABY! (i can hear you whining in the background loL) and well to be honest...someones naughty tonight..real naughty,cheeky,bitey,licky and all! that someone is you should know..well actually WE ARE NAUGHTY..not only my baby but US (happy? loL i can see you grinning at ur com)! i really hoped you could have been at my home this afternoon,as we watch a horror and gory show! (can imagine you cuddling me as we lie on the sofa and watch it together..Watch the show but doing nothing else other than cuddling and watching..loL dont think too far..well maybe..) okok better get going..must accompany my baby now..loL..calling her later...update tmr..     ~i love you baby..(my irIs not my blog) '),(11614,3999305,'14,August,2004','2004-08-14 00:00:00','   ~Lotsa crappy things happen today..i laughed till my guts fell out and i swallowed it back again..Lol nah..thats how hard i laughed!we crap till we could not..it was great..went to lil india then nearly went to delta but i went to orchard today (as usual..orchards my second home =P) From fishes to chicken to food to movies to people to...to...EVERYTHING! it was just a fun day..but for all the time,always one thing stays on my mind..My sweetheart,my sleepyhead,my baby..! bought something sweeet for her today..i hope she\'ll love it..i got a feeling she will (WiNk*WinK* NO i did not get you colourful and tasty condoms !! jK) loL..and someone also went shopping today..and bought a piNK bag! not hot pink..just pink! loL.. and i wanted to get a new boxers but well ill wait before getting it cause well..its best to go with her..she knows best cause my body is her body..we share it..loL..whats mine is hers and hers is mine.. Right master(part-time slave but mostly master!) ..okok..thats almost everything that happened today..oh yea..i loVe it if you snuggle(RiRi)! update ur tmr.. '),(11615,3999305,'13,August,2004','2004-08-13 00:00:00','   ~ Todays friday the 13th! the day where bad luck roams and searches its victims! (not strip search!!what were you THINKING?!) anyway..the day was kind of quick,went for a hair cut(great!!),then to the gym then soccer then back here..to the place where i belong.. :P  Home..Friday the 13th dont seem bad to me..in fact its good,always didnt expect bad things to happen...well then night came..man..just when i really needed my sweetheart..so tired and worn out..needed to talk to her but didn\'t talk much and she was not paying attention to me..only a few words as she was watching a video..and chatting with her friends.. :) but i dont mind,as long as she\'s happy with it then so be it :) now she\'s offline..dont know what to expect when she\'s back..but well lifes like that..gotta give and take :) i trust her to the core and she trust me equally..we have faith in each other so we\'r always happy-go-lucky!  And Sweetheart bought a new bra..hmmm and its the colour of my room! loL sexY!! too bad guys dont have bras or not we would be fighting over whose is nicer or sexier.. :P loL! just kidding!  Hmm maybe should get new undies..loL..or maybe boxers but well boxers ill get when im with my sweetheart..gonna shop for one with her.. loL..i bet she would love to see me in it..right?!  should i shop for undies alone or..with her is better? maybe with her cause she got DA STYLE! :P and she knows me inside out..loL..you know what i mean right( sweetheart )? okie ill update it tmr..if i got anything else maybe ill add it later.. :P '),(11616,3999305,'12,August,2004','2004-08-12 00:00:00','   ~ ALRIGHT \"SOMEONE\" got B3 for MT O levels! okok not just someone..my sweetheart scored extremely well..soOoOoO PROUD of her..nVr been more proud for anyone else other than her and my cousin shanice..Today also miss her as usual.(theres not a day when ill not miss her..loL) and lucky Riri reminded me to update my blog or else i would have forgotten..as usual..Someone was also really HOT that night till we got real real real far..and especially...wet (only both of us understand the term \'wet\') anyways,today at sch was like any other days,boring,tiring and sleepy..well kinda but nvr sleep at all in class.. :) im a good boy! (someone knows im not..loL...RIGHT?) studied abit as usual for revision and accompanied the guys after sch and hang out with them...them i went home and take a quick nap then revise...(with all this time thinking abt you RiRi) then watch tv but all the show just sucks real bad..then i did some illegal stuff on the web..(shhhhh i downloaded MUSIC!!) hahaha!! silly me...ahhh well thats abt today..not much neways..but well bettr than nothing..update you tmr.. '),(11617,3999305,'09,August,2004','2004-08-09 00:00:00','   ~Ahahahahah Sg IdoL is A failure with full of crappy singers..my Riri could sing better..got a sweeter voice! loL..it was really killing me..so badly..haha could laugh till i die..well i wrote a msg for Riri and sent it three time but not once did it came out..maybe they wated something relevent to the Show i think..nvm ill be back to try again till i succeed..wanted it to be a surprise..anyways my day was boring..i was really tired today and \"someone\" was also being cheeky with me again today morning!...loL right? right my \"STUDENT\"? Got me soo hot.. =P and i bet you were too right? abt my day..well nothing much..Studied abit,not much only half an hour but it was just to refresh my memory on what im studying..It\'s a sleepy day for me..just lying down then study again( Pricipal must study and have lots of stamina,Right ;) ) Me and my student are sleepyheads and we got lots of stamina! Now just waiting to call my dear \"student\" as i must talk to her everyday or i can\'t go to slp..talking to her is the closest i can get besides being physically with her.so its very impt..also miss her thats why.. My heart lies besides her..its in her..she got it between her fingers in her grasp so tight that its almost impossible to remove..in fact its just impossible..! My dear,my \"student\",my sweetheart..basically my everything.. :) i just love her for who she really is.. '),(11618,3999305,'06,August,2004','2004-08-06 00:00:00','   ~To start things off..Things went really chaotic today..my school with around 1500 people went berserk and start fighting by throwing Highlighters that were givin out to us today..they started throwing all sorts of things at each other in front of the VIP guests,Principal and the teachers who were singing on stage..they were in a state of shock too! i was just laughing all the time..could not stop,it was soooooo totally funny!.. oh and we got lock in the school also and had to climb out..haha..so fun..we played in school with no one but us around..so freaking scary and haunted the school looks! we played Nike soccer today,where we have to chase the person with the ball and steal it from him and run away while the rest chases ya..real fun playing around in blocks and running in circles doing tricks to pull them away..then my friend fell down..tripped him!! hahahahahah!!! im sooo mean! well today someone was being naughty with me and that turned me on real bad..well truely,i really love her for what she is..not only abt her being naughty all the time but her herself..thats what i admire abt her..Honestly..the way she portrays herself..her character..we\'r always naughty all the time..loL..well *muacks* for ya my RiRi..and i love you..treasure this as this is only for you..no one else could hear it coming out from me.. *pinch ur nose* '),(11619,3999305,'05,August,2004','2004-08-05 00:00:00','   ~ Today afternoon was...greeeat! played soccer for a whole two hours and continued after resting..i think my friend is gay..called me sexy as i wore my pE shirt and shorts..haha!! i felt really strange when he said that..but ..ahhh just dont like the feeling..haha..really..it happened before and now it happened again but this time its a GUY and he\'s my FRIEND!  Tonight really is a wonderful night,sweetheart proved that she cares and gave me a *hugz* so im gonna give her a *muaCk* hope you guys reading this dont mind and get jealous yea?! Cause if you read my blog,you need to get use to it.. loL *sticks out tongue*...and yea got hit right on the side of my right eye real hard..really painful but at least it did not swell or bruise..no sign of it even.. now feeling much better as got a hug from my dearest sleepyhead! well thats all i got to say..update ya tmr! '),(11620,3999305,'04,August,2004','2004-08-04 00:00:00','   ~today is hardcore studyin day..went to civic centre with another of my friend who asked me to accompany and studied there till 7pm! being there since 3pm..it was tiring alright..and only studying one subject is really tough and kind of boring but it all went up there..haha! well then came home ate like there was no tmr...then went online abit late and made my sweetheart wait...just want to say sorry *cuddle* she gave me a *muackZ* and *hugZ* and made me melt like chocolates preferably MnM\'s under the sun! Shes just so sweet..really..my tiring day has changed into a hyper me..loL..If ur reading this sweetheart,  *MUUAACK on the LIpS*  this kiss comes from deep down in the heart..so deep down..untouchable by anyone and everyone but she manage to get it out and she got it.. :D back to my day,got two invitation to go and study today but only accept one..cause the other one could not study at all..will only give me trouble..lotsa them..got lots of gossips in class abt almost everyone and everything by my gOOOD friend sitting in front of me..got scolded by Eng teacher today for talking to me too much..guess you can say addicted!! hahaha!!!  then you should see me in maths lesson taking a test,half dozing off and half attemping the questions but well it was easy..also made my friends laugh by telling them all kinds of stuff..feel great to see my depressed friend happy again in class..Cant wait to call my sweetheart tonight..miss her alot.. want to dedicate a song to her on the radio but i have to get the right time...update ya some other time! '),(11621,3999305,'03,August,2004','2004-08-03 00:00:00','   ~ feel so sweet and loved by sweetheart RiRi..always never fail to melt my heart everytime..so sweet..she stole back her hP to Sms Me and risking her life.. :P loL..also called my dear..second time already..though its just a word,i wouldn\'t mind hearing it all day..it could stop my heart and make it come back to life without CPR but i would like a mouth to mouth resusitation(i think the spelling is wrong but you guys should know what it means right?) anyday from my sweetheart..back to my day..got carried away..well day started great and my friend so happenly got on to friendster and showed my sweetheart Picture to almost everyone including one of my best friend aka happy ahbeng..loL..then he so happenly go and make a public announcement..loL!! But its alright :D i dont mind!! played soccer again but i wasen\'t myself today..not in the mood to run..loL really! hmm well..i didn\'t like what i read in my sweetie blog..its not her fault but well i just dont trust guy\'s these days..know them inside out and outside in..but always got trust in Riri..as long as my trust in her is not betrayed and is honest and sincere with me..she will always have it *muAck* yst night was great too.. ~Instread of studying,i slept cause SOMEONE! nah my sweetheart wanted me ta sleep together with her..at the same time! k update it some other time! '),(11622,3999305,'02,August,2004','2004-08-02 00:00:00','   Well well..seems a long time since i updated my bloggy cause i\'ve been rather busy but time i can always make and find for my sweetheart RiRi!! its definately! a tiring day..though i studied but not too much..its just too tempting to doze off to dreamland as you know im a sleepyhead like SOMEONE! haha! *muAck to my other sleepyhead* school was tiring as i can\'t believe i doze off while doing a physics test as i finished really fast..dont even know if i knew what i was doing..loL and my teacher apparently,what\'s surprising is that my MT teacher gave me Sincere praises in front of the whole class which she is fond of in doing to me and i was kind of stunned! but well i told her not to do that again..i dont like too much attention..gets frustrating sometimes unless its english! then i would love it..loL..And as usual..missed my sleepyhead so much in class! wish i could give her lots of Smuacks and CuddleS and Frenches..but the day ended fast,came home and Bath as usual as im famous for that right sleepyhead?you know me best! loL then i studied for not quite long before dozing off into LALA land!then woke up studied some more and ate..then its until now..loL well update it some other day ritez? oh yea and i cant wait to spring in the surprise..loL for now,its a secret.. love ya my.......SLEEPYHEAD! be sleepy like me always.. '),(11623,3994464,'21,July,2004','2004-07-21 00:00:00','   I asked my daughter today what her inner voice defines about her these days.&nbsp;&nbsp;   I wanted to hear her  very deep  inner voice-&nbsp; not the stuff like, \"a great vocalist,\" or \"good test-taker.\"&nbsp; The deep inner voice that  defines her  is what I sought.&nbsp;   Mine is, right or wrong, \"special and different.\"&nbsp;   It wasn\'t always that way.&nbsp; It used to be \"slutty and irresponsible.\"&nbsp; Then it changed to \"overwhelmed martyr/screw-up-at-life.\"&nbsp; Then, I turned 30 or so and it evolved to \"trying very hard.\"&nbsp; In later years, it changed to \"successful at work but not quite there as a Mom.\"&nbsp; And so on.&nbsp; I\'m sure it will keep changin\' as I go.   My daughter\'s inner voice, right now,&nbsp;is \"the Good Kid.\"   I got to thinking about that.&nbsp; I asked her if she knew what happens to Good Kids.&nbsp; She replied, \"They die early?\"&nbsp; I said, \"Well, sure.&nbsp; They get mauled by tigers.\"&nbsp; (See, we share a morbid, sick sense of humor, we Wheatleys.)   \"Seriously, though,\" I said.&nbsp; \"\'Good Kids nearly always have something in common:&nbsp; They have fantasies of being bad, and since they\'re good, they\'re really bad at being bad.\"&nbsp;   I went on to describe it-&nbsp; we all have those friends who are just  not  good at cussing.&nbsp; They\'re also terrible and weepy when drunk.&nbsp; They \"get caught\" more than everyone else.&nbsp;   My goal as a parent is to try like hell to encourage them to figure out what they\'re thinking, feeling and living, such that they can be comfortable in happiness before they\'re parents themselves.&nbsp;   Yeah, I realize that things like hormones, lack of maturity, and lack of education prohibit some self-actualization, but I still GRANDLY envision my kids being simply and utterly as happy as is possible at this moment and every moment in the future.&nbsp;   And of course, true happiness is often borne through tribulation and mistake-making.&nbsp; So don\'t get me wrong and think I\'m Alice in Wonderland-ish.&nbsp; \'Cause I\'m more like the cynical Susan Sarandon in whatever that movie was where they flew off a cliff in a car.   But could it be that a person\'s label of Good Kid could hamper their ability to be bad, and therefore redeemed, and therefore, happy?   I dunno-&nbsp; it\'s just what I\'ve been thinking about while reading  The Ragamuffin Gospel-- &nbsp; Manning\'s description of the Catholic definition(s) of sin- mortal and venial--While he draws a hilarious picture of his experience in confessing&nbsp;adolescent masturbation&nbsp;with his priest, the shame he also endured was palpably recognizable to me.&nbsp;   He\'s making it painstakingly clear that Christ apparently preferred sinners to&nbsp;Good Kids.&nbsp; He said he came for the sinners.&nbsp; He let the Good Kids do their thing.&nbsp; (When I say Kids, note I don\'t mean the little children who He obviously loved dearly.) The Good Kids were apparently \"Pharisees\" back then.&nbsp;   I donna wanna mi\' daughter to be a Pharisee.&nbsp; Yuck.   So anyway, if it\'s true that he loved the sinners, then the question I have is whether to encourage my daugther to just  be .&nbsp; Not be the Good Kid.&nbsp; To just be.&nbsp;   And perhaps screw up every once in a while so she can learn how freakin\' awesome it is to be unconditionally forgiven simply because Jesus loves her.&nbsp; Nothing she can do can make that bigger, better or different.&nbsp; It simply is what it is.   Hmm.&nbsp;   '),(11624,3994464,'21,July,2004','2004-07-21 00:00:00','   I just fell in love with Brennan Manning, author of  The Ragamuffin Gospel  and  Ruthless Trust .&nbsp; His followers were mine as a younger \"christian,\"&nbsp; Michael W. Smith as a teen, and more importantly, Rich Mullins as an adult.&nbsp;   Rich came to San Antonio frequently when I was doing the Baptist thing.&nbsp; I\'ll never forget one night when he played and my mom and I were there.&nbsp; It had to be at least ten years ago, and I honestly have no idea why only Mom and I ended up at this event.   It was her birthday.   I watched and somehow decided that in heaven, I\'d be married to someone like Rich, because he GOT it.&nbsp; He understood the way that Christ was about.&nbsp; And he\'d be a good husband.&nbsp; I literally coveted Rich Mullins\' existence.&nbsp; (though I was married with three kids at the time and didn\'t have his existence.)   Ok, so of course, there\'s some weirdness about that.&nbsp; But I really lusted after Rich in the most spiritual of ways.&nbsp; I just wanted him to be my husband and then everything would be okay.&nbsp; (Mom, at that time, wanted me to marry Carmen, but that\'s another blog.)   So later, I&nbsp;used my \"Christian connections\" and got us into the Christian bookstore \"afterparty\" and he came in and as he was exiting after a ten minute acoustic thing, and we passed him on the stairs.&nbsp; I screamed to him like I was a groupie and he was a Beatle, \"RICH!!&nbsp; IT\'S MY MOM\'S BIRTHDAY!\"   He stopped and looked at me.&nbsp; I grinned charmingly and said, \"Pretty please, could you sing her Happy Birthday???\"   He looked at me (and with what&nbsp;I hope&nbsp;was wishfulness that I was his wife,)&nbsp; sang it.&nbsp; Sweetly and with authenticity.&nbsp;   And my mother was as thrilled as all get-out.&nbsp; And I always loved him more for that.   When I learned he died in a car accident, I just ached.&nbsp; I even committed the mortal sin and actually \"prayed\" to Rich on a regular basis when I was confused or lonely.   So fast-forward to today:&nbsp; I\'m at  Barnes and Noble  picking up Brennan Manning\'s  Ruthless Trust  and was inadvertently directed by the store clerk to  The Ragamuffin Gospel .&nbsp;   And lo and behold, there was a message in the forward by Rich,   And lo and behold, DUH, this is why he\'s called his band  The Ragamuffin Band .&nbsp;   Good lord.&nbsp; Could it be I\'m finding what I was looking for?&nbsp; Even if it\'s  not  being married to Rich?   I\'m overwhelmed.&nbsp; Because guess what?&nbsp; Today is July 26.&nbsp; My Mom\'s birthday.&nbsp;   No joke.   '),(11625,3994464,'21,July,2004','2004-07-21 00:00:00','   Going to a new church is hard.&nbsp; Especially when it\'s not a \"church.\"&nbsp;   When it\'s more of a gathering, where there\'s a lot of encouragement for people to be honest and affectionate, it\'s almost like you\'re on a date.&nbsp; Or you\'re in junior high.&nbsp; You feel like everything you do could be keenly analyzed.&nbsp;   You wonder if you\'re dressed right.&nbsp; You\'re being inundated with new faces and names.&nbsp; You wonder if your butt looks big in those jeans.  You wonder if the guy next to you is simply being kind to your kids, or if he\'s a freak.&nbsp; You want to know the story about that couple in front of you who have four kids that don\'t look like them.&nbsp; Did they adopt?&nbsp; How cool is that?&nbsp;   You want to know what makes those people who sit on the front row who they are.&nbsp; Are they special friends of the pastor?&nbsp; You especially want to know this, since in your non-\"church,\" the \"pastor\" isn\'t a pastor.&nbsp; He\'s just the leader of this unruly bunch.  Who\'s that woman in the back, the one bustling about and apparently holding things together?&nbsp; Does she get paid to work so hard?&nbsp; Or is she independently wealthy and has alot of time on her hands?&nbsp; Her shoes are fabulous.&nbsp; Wonder where she got them?&nbsp; You can\'t ask, \'cause you\'re just here to observe.&nbsp; And everyone seems so busy except for you.  Wouldn\'t it be a hilarious, cosmic joke if everyone around you is pretty much thinking the same thing?&nbsp;   Wouldn\'t it be sweet if everyone started asking those questions aloud?  &nbsp;   '),(11626,3994464,'21,July,2004','2004-07-21 00:00:00','   Alone  is a weird concept.&nbsp; What does it really mean?&nbsp;     &nbsp;  \"I hope you don\'t grow up and be  alone .\"    &nbsp;  \"Please.&nbsp; Just leave me the hell  alone! \"    &nbsp;  \"Finally, the kids left and I was  alone .\"    &nbsp;  If you start to actually look at the word, I promise, you\'ll feel like you\'re on acid.&nbsp; It\'s a strange word.&nbsp;  Alone .&nbsp;     &nbsp;  Good?&nbsp; Bad?&nbsp;     &nbsp;  I say good.&nbsp; Though before I always saw it as bad.&nbsp;     &nbsp;   Alone  is not alone at all.&nbsp;    Alone  is simply quiet.&nbsp;&nbsp;Purity is&nbsp;perceived as alone, \'cause it has no makeup.&nbsp; No one notices the girl without makeup.&nbsp; Except God.    &nbsp;   Alone  is not lonely, despite its lingual derivatives.&nbsp; We all yearn to \"be alone.\"&nbsp; To escape the talk, to be by ourselves.&nbsp; To experience nothing, yet every single thing that we\'ve put off forever.&nbsp;     &nbsp;   Alone  is when we lay on the bed and no one is around, and the comforter is on our cheek and we don\'t have anything, not even a mother, there to silence our juvenile commentary.&nbsp; At that place, we\'re just baby grown-ups.&nbsp;     &nbsp;  And that\'s when the conversation with God occurs.&nbsp; That\'s when we bargain like there\'s no tomorrow.&nbsp; That\'s when the  real  comes in and we know that  He\'s &nbsp; the One who is actually real, and by God, we\'d better get this shit figured out.&nbsp;     &nbsp;  It\'s often called panic.&nbsp; It\'s often called epiphany.    &nbsp;     &nbsp;   &nbsp;   '),(11627,3994464,'21,July,2004','2004-07-21 00:00:00','   Go this moment and rent  The Human Stain .&nbsp;     &nbsp;  Watch it, please.&nbsp;     &nbsp;  Come back.    &nbsp;  Now go buy the book, which is simply phenomenal.    &nbsp;  Come back.    &nbsp;  Finished?    &nbsp;  Y\'all.&nbsp; The layers of humanity are thrown in front of us, begging to be examined.&nbsp; I want to talk about irony.&nbsp; I want you to  get  it.&nbsp;     &nbsp;  There\'s really nothing but desperation for love sitting before us.&nbsp; And there\'s nothing but love sitting before us.&nbsp;     &nbsp;  Just watch it.&nbsp; I have a feeling you kept reading this before you obeyed me, dammit.    &nbsp;  :-)   '),(11628,3994464,'21,July,2004','2004-07-21 00:00:00','   urlLink    Contemplative Me.  Not at all what I generally look like.  &nbsp; urlLink   '),(11629,3994464,'21,July,2004','2004-07-21 00:00:00','   Be still and know that I am God    Psalm 46:10    &nbsp;   But those who wait on the Lord shall renew their strength    Isaiah 40:31    &nbsp;  Here\'s a Rule of the Universe:&nbsp; If you\'re a redhead and nicknamed \"Missi,\" you\'re likely to be told to \"sit still\" or \"hush up\" or \"slow down\" for most of your life.&nbsp; People who sire children should be aware of this and think twice-- If I\'d been named \"Siobahn,\" perhaps I would have developed a colt-ish figure and been more inclined towards serenity.&nbsp; Alas, this was not the case.    &nbsp;  I struggle frequently with that missive,  be still , in every area of my life.&nbsp; Whether it\'s shutting my mouth, or resting my brain, or driving on Bee Caves Road, I am constantly going-going-going.&nbsp;     &nbsp;  Today, I had to stop working on something to let the cleaning people do their thing in and around my office.&nbsp; I decided to leave the phone inside and slip out to the pool for a while.&nbsp; It was, decidely, torture.&nbsp;     &nbsp;  I was floating around and wondering if anyone had called.&nbsp; Then I decided I was so hungry that perhaps I was starving myself.&nbsp; Maybe I needed to go back inside and fix a small snack?&nbsp; No, the cleaners were busy in the kitchen.&nbsp;     &nbsp;  Wonder what the kids were up to?&nbsp;     &nbsp;  Could my neighbors see how fat I look on this float?    &nbsp;  How come I\'m a slob and can\'t keep the pool area swept?    &nbsp;  Those plants are gonna die because we never water them.    &nbsp;  God, I\'m so hungry.    &nbsp;  Is it late enough in the day to have a cocktail?&nbsp; I mean, it\'s  Friday .    &nbsp;     &nbsp;   Be still and know that I\'m God .    &nbsp;  But I don\'t know what the hell that means.&nbsp;     &nbsp;   Yes, you do .    &nbsp;  No, I DON\'T.    &nbsp;   Look UP, idiot .    &nbsp;  And would you believe that the color of the sky in Austin, Texas today is the very definition of \"sky blue\" in a Crayola package, juxtaposed against \"grass green\" in the trees?&nbsp; And it smells kind of froggy in my backyard.&nbsp; And there are these cicada-sounding things going off everywhere.&nbsp; And the sun feels good on my skin.&nbsp; And floating is right up there with skydiving.    &nbsp;  It\'s all there.&nbsp; We just have to be still for a second.&nbsp; In my case, I was back inside within 45 minutes, and I feel like I slept for twelve hours.&nbsp;     &nbsp;    &nbsp;   &nbsp;   '),(11630,3994464,'21,July,2004','2004-07-21 00:00:00','   When\'s the last time you woke up on a Sunday and really, truly, thought:&nbsp; \"I am SO excited to go to church!\" ?   C\'mon, admit it-&nbsp; church sucks.&nbsp;   For me, at least, church is something my parents instilled by forcing me to go.&nbsp;   But I\'ve still&nbsp;tried nearly all of them:&nbsp;   1)&nbsp; I\'ve tried the  traditional  ones that have large, inspirational architecture and intellectually interesting rituals.&nbsp;   And wow.&nbsp; All of that inspired me.&nbsp; I had the most amazing experience in St. Patrick\'s cathedral once with my best friend.&nbsp; I remember leaving and feeling that it was almost like a (really good) sexual encounter.&nbsp; The vastness of the building, combined with the hushed reverence that it commanded, caused this wonderful discussion of Mary Magdelene\'s role as a female disciple, and I felt awe and power while we quietly worked through a few personal issues.&nbsp;   But I know I wouldn\'t awaken each Sunday and feel compelled to visit St. Patrick\'s.&nbsp; Quite the contrary; I want to reserve it for a few special times.&nbsp; It\'s too freakin\' awesome to become that familiar to me.   2)&nbsp; I grew up in the beautiful, homey,  community  churches.&nbsp; They\'re everywhere, and their allure is never really denominational, truth be told.&nbsp;   Instead, the claim to fame in these churches is that they love kids.&nbsp; They nurture families and provided a place for us back when we were kids, and now, as we have become parents, they have a neat Sunday School program which prepares the kids for school and also re-instills our generational values.&nbsp;   The great thing about community churches is that you can raise several generations of children with the same old folks, thus creating a sense of family.&nbsp; Unfortunately, in every \"family,\" Grandpa thinks his way is right, and that it transcends ANY modernity.&nbsp; So, Mom and Dad listen to Grandpa.&nbsp; And kids rebel against that, and come back into the fold after law school or having children themselves.&nbsp; It\'s inbreeding at its worst, but it serves to educate us&nbsp;on all that modern politics is&nbsp;about, eh?&nbsp;   Oh, and \"church splits\" in community churches are the bloodiest.&nbsp;   And many people&nbsp;just eventually wake up one Sunday morning and decide that  Meet the Press  and the  New York Times  are more interesting and less painful for both themselves&nbsp;and their kids.&nbsp;   3)&nbsp; I love the  fun  churches.&nbsp; These are the ones with charismatic pastors who garner a national television audience.&nbsp; The music rocks, the building is usually donated by a local millionaire (who controls the church council, incidentally) and their Easter and Christmas Eve candlelight services are standing-room-only.&nbsp; I\'m telling you that this exists in every city in the U.S.; Austin isn\'t unique-- it\'s common in progressive churches and in fundamentalist ones.&nbsp;&nbsp;&nbsp;   I love being entertained at these churches.&nbsp; You can also differ in major theological viewpoints with the pastor, but because you\'re sitting 10,000 feet away from him, and he\'s so nice and funny, you don\'t even try to \"chalk it up.\"&nbsp; You just show up occasionally and brag to your friends about the rock concert you get to attend each week.   And one day, you just wake up after a hard night of life on Saturday night, and decide to sleep in and watch your pastor on his next televised viewing.   4)&nbsp; The  emotional  churches are often branded as fundamentalist, but my take is that they\'re the most compelling, because they actually deal with reality, or at least part of it.&nbsp; They focus on human frailty, which is what we all lay awake praying about at 4:30 in the morning when it\'s still dark and we realize we\'re screw ups.&nbsp;   \"If we could only speak in tongues, God could hear us.\"   \"If we\'re sick and have a little faith, God can heal us.\"   \"If we\'re having an affair, God can send us a Christian&nbsp;to save us.\"   There\'s a certain truth to this, methinks.&nbsp; And there\'s a certain condemnation that makes us depressed and inferior.&nbsp;   And we wake up one morning and decide to go to Gruene Hall instead, \'cause we\'re sluts/bad people anyway.   -------------------------------------------------------   Ironically, in the end, I\'d like to get up one day a week and gather with people who respect a portion of all the above descriptions.&nbsp; I\'d  like  some tradition.&nbsp; I\'d  like  people to care about my kids.&nbsp; I\'d  love  some coolness.&nbsp; I\'d  like  for people to call me out when I\'m messing up as a person.   In my view, it would be neat to have a  Breakfast Club  church, where you have the geek, the athlete, the princess, the outcast, and the reprobate all hanging out together and learning something.&nbsp;   Because after all, aren\'t we  all  the geek, the athlete, the princess, the outcast, and the reprobate?   '),(11631,3994464,'21,July,2004','2004-07-21 00:00:00','   Deny thySelf, yet love thy neighbor as thySelf .&nbsp;   Huh?   What did Christ mean when He talked about our relationships with our Selves?&nbsp; Much church doctrine is centered on \"Self sacrifice.\"&nbsp; It\'s hard to comprehend sacrificing someone/thing called mySelf when I don\'t know her very well.&nbsp; If my spiritual pursuits have been, up to now, focused on denying mySelf, then how can I love someone else like I love mySelf?&nbsp;   Here\'s a&nbsp;mom; here\'s an analogy:&nbsp; I constantly struggle with smoking.&nbsp; I deny mySelf (sometimes), but feel SO deprived.&nbsp; Because I see not being able to occasionally smoke as deprivation,&nbsp;I then treat mySelf as a deprived person.&nbsp;   So then, I have these kids.&nbsp; When they ask to go to the movies, my automatic answer is nearly always no.&nbsp; So I love my neighbor as I love mySelf.&nbsp; My deprived Self.&nbsp; While this isn\'t a justification for smoking, since I\'m changing my views on it these days, I spent years where that has literally been part of my existence.   SO...   Of all places, I think I figured something out while going to the bathroom this morning.&nbsp; My teens were pouncing around outside the door trying to get me to agree to yet another endless solution to summer boredom.&nbsp; My instinct was to yell, \"SHUT. UP. NOW.&nbsp; JEEZ!\"   For some reason, the adage, \"Do unto others as you would have them do unto you\" popped into my head.&nbsp; Suddenly, it wasn\'t about the fact I was powdering MY nose.&nbsp; I was hurled to the outside of the bathroom door.&nbsp; And my face turned red, because I remembered earlier this morning when my son was in the bathroom downstairs and I was knocking incessantly to tell him to hurry so we could make his doctor\'s appointment.&nbsp; Yep.&nbsp; Them apples fall very near my tree.   I got to thinking about it, and how I\'d felt when, during the earlier bathroom incident, my son had yelled, \"GOD! MOM!&nbsp; I\'M! COMING!&nbsp; JEEZ!\"&nbsp; And I guess I do believe in being \"filled with the Spirit\" after all-&nbsp; because as I remembered that, something came out of my mouth that was other-than-Melinda:&nbsp; (calm) \"Guys, I\'m in the bathroom.&nbsp; I\'ll come out there and love on ya\' in just a second.\"&nbsp; Yeah, I know.&nbsp; Pretty cool.&nbsp; They\'d already moved on to bigger and better things by then, but it still had an impact.   Ok, so back to mySelf.&nbsp; I felt really good about mySelf when I did that.&nbsp; I was like, \"Damn.&nbsp; You\'re the bomb.\"&nbsp; And I still recognized God was involved because a) I was thinking about Christ\'s teachings when it happened, and b) I responded like He probably would\'ve.&nbsp;   So there was this little lightbulb that went off:&nbsp; What if sacrificing one\'s Self is about sacrificing the crap that makes us feel like losers (like yelling at people)...the stuff that makes us loathe ourselves and panic that we\'re going to Hell?   Could it be that it\'s all one sweet, vicious cycle?&nbsp; I love you, you love me, we\'re a happy family?&nbsp; If I love you, I love me.&nbsp; If I love you, my sense of Self becomes heightened.&nbsp; I want to treat mySelf with the same love I give to you.&nbsp; If I love you, I expect- no, desperately desire- you to love me back.&nbsp; I\'m as weird as they come, but my tiny little brain has to wrap itself around these concepts in order to fully invest mySelf in pursuing Jesus\' way of living.   '),(11632,3994464,'21,July,2004','2004-07-21 00:00:00','   The anticipation of school reconvening loomed large in this household. The feeling was mutual, punctuated by doors slamming, complaints of the crapola food they had to eat here, and my whining about the plethora of computer viruses accidentally downloaded on my computer by \"no one.\"  I liken the environment to pre-Christmas, where shopping and baking makes for one bitchy Mommy. But the grand day of its arrival creates a psychological block that prevents the memories of the previous days to intrude. Come to think of it, childbearing also produces the same effect.  So, I stumble out of bed at the ungodly hour of 7:50 a.m. to transport the teens to their respective campuses. I\'ll add here that it is sheer bliss when children come of the age to not only dress and feed themselves, but to shun your advances to assist. God\'s so good in that regard, ain\'t He?  Turning left from my neighborhood street onto Bee Caves Drive during the school year is tantamount to proceeding left on 5th Avenue in Manhattan during peak hours. One needs a cabbie or public transport. Given that we have neither, and the kids have informed me that the school bus is wacked, uncool, too early, and above all, downright  dangerous , I am relegated to driving them for a few days out of guilt. This, too, shall pass. It always does for me.  Ten minutes after approaching the intersection, a kind gentleman lets me in and we\'re off to the races. I drop Amanda, the sophomore-with-a-sassy-haircut, at the high school. As she crosses the street, I see two Abercrombie OC-type boys checking her out. For some reason, this pissed me off, but I veered away from them right before I got too close.  As we approached Troy\'s junior high, it was obvious that every parent had fallen hook, line and sinker for the ol\' \"the bus is wacked, uncool, too early, and above all, downright  dangerous \" speech. And apparently, they weren\'t too happy about it. There was no friendliness. There were no smiley waves like we do at football games. Instead, we sat in a line, brows furrowed-- determined to hold our place in line and  not. let. anyone. cut .  As we finally approached the school entrance, I had practiced deep breathing and turned to ruffle my baby\'s hair before he exited the car. \"I lov...\" The door slammed. He waved behind his back without looking at the car. Sigh. I\'m a chauffeur, that\'s all.  40 minutes later, I arrive home from the three-mile roundtrip trek to the schools. I breathe deeply again and proceed to do the following:  1. Take a long, hot bath. (Boring. Took 5 minutes.)  2. Drank coffee and checked email. (10 minutes.)  3. Listened to talk radio. (Sigh. 3 minutes.)  4. Checked email again. (Why don\'t I have any friends? 1 minute.)  5. Blow dried hair. (4 minutes. )  6. Tried to take a morning nap. (10 minutes.)  7. And so on.  By mid-afternoon:  1. Decide I should volunteer at each of the kids\' schools one day a week.  2. Read their schools\' websites to acclimate myself to their upcoming events.  3. Wonder if Amanda knows that there\'s an audition after school this afternoon?  4. Call Amanda\'s school to leave message about audition after school this afternoon.  5. Wonder what time Troy\'s football practice is over?  6. Call Troy\'s school to inquire.  7. Decide to take Troy and friend to McDonald\'s to celebrate their first day of school.  8. Call friend\'s Mom to let her know and learn she\'s picking them up. Already arranged by Troy.  9. Try to take a nap.  10. Write in this blog.  Life\'s so sweet when the kids are back in school. I give it three days and they\'ll be calling me to ask that I deliver some of my \"crapola food\" for lunch.  Ciao for now.   '),(11633,3760558,'30,June,2004','2004-06-30 00:00:00','    As if we didn\'t have enough action yesterday with the gas issue, today brought another line of traffic passing through Arnosvale. You see, earlier this morning, a small, private twin engine plane was taking off from the airport when it veered off the runaway and crashed into the fence opposite Greaves Supermarket. Luckily nobody was hurt but you know Vincentians, whenever something happens, rather than avoid the area, within minutes, you have about 5000 peoples there and no traffic could move. Right up to seven tonight, there was still traffic coming out of town and still a small crowd gathered discussing all kind of things about the crash. All of a sudden we have 50 experts on plane crashes and how ir happened and who was to blame and how they could move the plane and..........well, you get the drift.  With all this action going on I wonder what tomorrow would bring? we shall see. '),(11634,3760558,'29,June,2004','2004-06-29 00:00:00','    Here\'s a big country question - What do you think is happening when you pass into a gas station and you have to line up for 30 minutes just to get some gas? most would probably say everyone wants to fill up before a bank holiday or maybe the people working at the gas station is on strike or a go slow or even prices are going up at midnight and everyone wants to fill up before the prices change. Well in a big country thats probably the case but here in Vincy, things are not that straightforward. You have to remember that we don\'t have any oil deposits that we can pump so we have to import it. Shell and Texaco are the suppliers and every week you\'ll see one of thir ships in port pumping gas into numerous storage tanks. Well it so happened today that one of the tankers went to the gas station in Villa and off loaded its content. The station opened back up and cars and buses started to fill up. The problem was that someone made a mistake and put deisal into the regular gas storage so the station was pumping the wrong type. What a ting!!! the shell station had to close and everyone ended up going to Texaco for gas causing a huge line. The thing is, they were now running out of gas causing a general shortage. Another boat is due in next week, hopefully before we seriously have shortages accross the country. As they say.......only in St Vincent!!!!!! '),(11635,3760558,'28,June,2004','2004-06-28 00:00:00','    OK, OK, I missed a day but what do you expect living on an island. Nobody\'s rushed here you know, think people do things on time? take for example today, the company I work for is flat out busy trying to get prepared for the 30th June deadline for the ISPS (International Ship and Port Security) code that would take effect from 1st July. You see, this code was introduced after the 9-11 thingy in the US, every port need to comply with a set of rules that Uncle Sam drew up. Now that was years ago but as usual, we leave it til the last moment and then run around like a headless chicken trying to put it in place. St Vincent finally got its act together earlier this year and my company followed suit back in May. Remember, deadline is right upon us and we only just  taking action. Well, the more we read the details, the more work we had to do and the more money we had to spend. What a ting!!!! So as one deadline approaches another one passes and thats for our disaster prepardness. In case you forgot, hurricane season started 1st June and we\'re supposed to make sure everything is in place just in case a storm passes by......YEAH RIGHT!!!!!!! As usual we would wait until one is on our doorstep and then........run around like another headless chicken!!! But don\'t worry, we safe, remember Hairouna.....Land of the blessed. '),(11636,3760558,'26,June,2004','2004-06-26 00:00:00','    urlLink    Frodo in a very compromising pose.........and still sleeping&nbsp; urlLink     And off course, with everything new, you just have to add pictures. So my blog will also consist of pictures and the like. Feel free to comment on any of my blogs or any of the pictures I publish by clicking on the comments section below.  Frodo is one of two cats we have along with two birds. Keep checking back for more info. '),(11637,3760558,'26,June,2004','2004-06-26 00:00:00','    Today is one of those days where you get up and decide to do something different....out of the ordinary........radical even. Well for here in St Vincent, doing anything, no matter how minor, is a change.  For me, I\'m going to do this blogging thing. I\'m going to rant and rave about life here in St Vincent and the Grenadines and make you wish you were here......well maybe not but you get the drift. For those of you who don\'t know where St Vincent is, go and look it up!!!!, learn something new about the world, it would do you good. So let me leave you to do that now and we\'ll pick it up tomorrow. '),(11638,3760558,'31,July,2004','2004-07-31 00:00:00','    Remember a few weeks ago I told you about our Vincy Triangle at Arnosvale. Well, guess what? another car went through the new piece of fence that was used to repair the hole. Now what are the odds? and on the topic of accidents, here\'s a classic one for you - I was heading up to Richland Park when a minibus came speeding down, the pickup in front of me stopped and a small collision occured. Well the usual thing happened with both drivers accusing each other of being in the wrong. This went on for about 15 minutes while no other vehicles could pass. The road was so blocked that not even the police could have got to the scene. But here\'s the joke, sitting on the back of the pickup was a police officer - I kid you not - when the accident happened, he came off the back and sat down on the side of the road, made no effort to resolve the dispute and even told both drivers not to bother him as he was off duty and hungry. I tell you, only in Vincy could this have happened and nothing doing. Oh well, such is life....... '),(11639,3760558,'23,July,2004','2004-07-23 00:00:00','    urlLink Travel   Leisure | World\'s Best 2004   The news is out - the best island in the Caribbean is The Grenadines - check out the web site above and feel proud to be a Vincy '),(11640,3760558,'21,July,2004','2004-07-21 00:00:00','    It seems as thought there\'s never ending news about the south eastern part of SVG mainland. Last week was the announcement of a new swimming pool and recreation centre in Rawacou and this week it\'s the announcement of the likely construction of a jet port in Argyle. Although it hasen\'t been signed on the dotted line, everything seems in place for that to happen very soon.     Bwoy, imagine - Argyle with supermarket, cinema, all kina clothes shop, restaurants, hotels, guest houses and all other types of commercial ventures. You drive through Stubbs and meet up round-a-bout number 2. You wait in traffic for 20 minutes while 2 traffic police try and speed everyone up and then you drive on past Mt. Pleasant which is now a huge residential area and on down to Agyle.    You say we progressing and then you pass into a Syrian shop and notice something - you run out to the next shop - another Syrian clothes shop - and see the same thing. Everywhere you turn, its just in your face. You run into the airport and look around, wherever you look, you see the same thing. You go out on the runway and look at the planes - same ting - it can\'t be, you say, this is a modern facility, this ain\'t supposed to happen but ask anyone who lives in the area now, and they would tell you, the biggest problem with Argyle is  RUST    .  You just can\'t get away from it but when you see it on the planes you decide to take boat from now on.     Maybe we should call it  The Rusty City of Argyle. '),(11641,3760558,'17,July,2004','2004-07-17 00:00:00','    I picked up this piece of news (see below) the other day and thought I\'d share it with you all. It\'s one of my favourite beaches because of the waves but a lot of people have died there so they banned swimming there.  In the next year, we\'ll see how it turns out.   urlLink We-FM 99.9 : \"Local Swimmers To Get New Swimming Facility 7/12/2004  A new swimming facility will soon be constructed on the Windward side of the island.  This morning the contract for the development of the facility at the Rawacou was signed between Sea Operations Ltd and the Ministry of National Security in collaboration with the Ministry of Tourism and Culture and the National Lotteries Authority.  The facility which will cost EC $ 774,030.00 is being co-financed by the National Lotteries Authority in the amount of EC $ 300,000.  This new venture is part of an overall plan to develop the Rawacou site into a major tourist attraction, catering for all ages.  The construction of the facility is expected to last between 7 to 9 months.\" '),(11642,3760558,'15,July,2004','2004-07-15 00:00:00','    Today brought the first real thunderstorm of the hurricane season and boy, did we have a real firework show. For as long as could remember, I haven\'t seen so many lightening over a period of time. They were flashing every 5 seconds for about 45 minutes, shame it was in the day rather than the night!!!! We even had one strike a building where I work - now thats getting close. Anyhoo, I\'d better turn off my computer before it gets turned off permanently..... '),(11643,3760558,'09,July,2004','2004-07-09 00:00:00','    Now there\'s one thing Vincy people like and thats cricket. And if you see exitement when the International Cricket Council (ICC) came to visit us a few months back. We roll out de red carpet, crack open de marby and give them a real tour of we small island. We tell dem that we have tings in place at we cricket ground, dat we raods good and that we can host some great matches for the world cup in 2007. Well, they come on tele after their visit saying how they well impress with we facilities, how de island pretty until and how people would have a great time here in vincy. How we tupid, we believe everyting dem man say. We done convince are we self that come 2007, we getting plenty games fo host, people from foreign like jokes coming to vincy and we meking plenty money!!! Well, who tell we fo say dat.......The other night, the ICC announced the venues for hosting the world cup. 8 country would get it out of 11 applicants with USA as a likely venue. As they name dem off one by one, we start fo get frighten, we can\'t hear Vincy call at all but we hear St Kitts (if you tink vincy small check dem out) and St Lucia but thats it for the St.  It turns out that although we have a good cricket pitch, the fact that we are off the beaten track and cannot get a direct flight to the island and that we don\'t have enough hotel rooms to put all the visitors expected went against us from the start. All they had to do was tell us that and done with it. But noooooo, we Vincentians too gullable - oh well, what to do!! '),(11644,3760558,'06,July,2004','2004-07-06 00:00:00','    Check this..... remember last week with the plane episode? well guess what happened yesterday. A car was travelling out of town heading towards Villa when the driver complained that she had no brakes, all of a sudden, the car picked up speed and swerved accross the road. Guess where it ended up. Remember the hole in the fence made by the plane, well a few days ago they finally removed it leaving a big hole in the fence. It was due to be repaired straight after carnival. Well for some strange reason, the car took on a mind of its own and went straight through it, hit the storm drain and rolled onto its roof. Again, luckily, nobody was hurt but what strange things going on in Vincy. Well plenty people must be frightened cause the very next day, the pull out the car and patched the hole before any other thing go through it and that time it would have been 3rd time unlucky!!!!! '),(11645,3760558,'05,July,2004','2004-07-05 00:00:00','    urlLink TonyOldies   For those of you who are interested in \"old time\" Vincy, check out the link above, you could see tons of pics from the past......check it out, you might even see yourself.  Special thanks to the guys who made the effort to pool all those pictures and make them available for viewing. '),(11646,3760558,'03,July,2004','2004-07-03 00:00:00','    Every year around about this time, thousands of students take exams under the guises of Common Entrance and School Leaving Examinations. The results for this year are as follows:  640 candidates wrote the School Leaving Exam,  124 [21.7%]  of whom were successful- 64 males and 74 females.Last year the same number of candidates wrote the exam,  161 [ 25.2%]  of whom were successful.  There was a decrease in the pass rate in this year\'s Common Entrance Exams. There were 2644 candidates and  979 passes [37.02%] , 383 males and 546 females. Last year, 2,582 students wrote Common Entrance,  1132[43.84%]  were successful.  Now instead of seeing a gradual inprovement of our students (in terms of passing) we\'re seeing the opposite. It seems that our students are taking the name of the exam literally and are  leaving school ......just so.....  and in Common Entrance there ain\'t nothing \"common\" about it unless you want to call it \"Common Exit Exam!!!!!\"  Jumping on my soap box, we really need to do something with the education system we have in SVG. We need to be finding ways of opening up the minds of our students and Think Big. Anyhoo,  there not much thought about that right now, we\'re right up on Vincy Mas 2004 and most people are thinking about \"Jumping Up\" and all that stuff associated with it. We have a long weekend coming up - lets hope it dosen\'t get spoiled by the few idiots that hang out in the crowd. We\'ll chat more in a few days. '),(11647,3760558,'02,July,2004','2004-07-02 00:00:00','    urlLink    oops, not quite the usual takeoff '),(11648,3760558,'01,July,2004','2004-07-01 00:00:00','    urlLink    Guarding the \"Ironman\"........not '),(11649,3760558,'18,August,2004','2004-08-18 00:00:00','    When have you ever heard of getting a half day holiday? probably never unless its a state funeral or wedding or something major like that. But today in Vincy, we\'ve had a decleared national half day holiday.  And the reason is.............football!!!!!  You see, our national team, formally Team SVG, now renamed  Vincy Heat    is taking on Trinidad & Tobago in a semi-final FIFA qualification match for the 2006 World Cup.  The thing is, you have a 90 minute match that starts at 4:00pm yet the county closes down at 12:00pm .  What a ting, as they say only in Vincy.  At least I wish them the best and hope they win and go on to the final qualifying round, may then we\'ll get a whole day off!!! '),(11650,3760558,'15,August,2004','2004-08-15 00:00:00','    We\'re now into the third week of the month and in terms of storms and the like, its been all go.  So here\'s the low down so far from NEMO.....  The report said that on August 15th Tropical Storm Earl impacted St.Vincent and the Grenadines reulting in the destruction of one house, extensive roof damage and other damages to 14 houses, several landslides throughout the country and structural damage to one bridge.  The Banana Growers Association anticipates that at least twenty-one banana Farmers are to claim storm damage to the tune of $40,000.  There were also reports of damage to some plantain farms valued at $20,000. Total damage was estimated at $342,3338.38.  The NEMO report further stated that on August 9th, Tropical depression #3 [later became Hurricane Charley] passed through causing several landslides resulting in road blockages and precariously exposing three homes, and wind damage to four houses.  Twenty-nine banana farmers also made claims for storm damagewhich amounted to $60,000. Total damage was estimated at $137,039.47.  On August 4th, a strong tropical wave passed thorugh SVG resulting in over 10 inches of rain in less than 24 hours. The E.T Joshua Airport, Arnos Vale, was flooded and closed for eight hours and numerous landslides were reported around the country.  NEMO says total damage was estimated at $62,700.  Oh and just in case I\'ve confused you, NEMO mean National Emergency Management Organisation. So there you go. '),(11651,3760558,'15,August,2004','2004-08-15 00:00:00','    urlLink    Calder getting blown away '),(11652,3760558,'14,August,2004','2004-08-14 00:00:00','    urlLink    Anyone for surfing? '),(11653,3760558,'14,August,2004','2004-08-14 00:00:00','    urlLink    Wave goodbye to Earl!!! '),(11654,3760558,'14,August,2004','2004-08-14 00:00:00','    urlLink    Coconut trees taking a battering '),(11655,3760558,'14,August,2004','2004-08-14 00:00:00','    urlLink    Mickey....run for your life, Charley\'s after you!!!! '),(11656,3760558,'13,August,2004','2004-08-13 00:00:00','    Well I\'m just  checking  out EARL and can you believe it, the projected path at this point puts it right on us.    Results for  St.Vincent (13.13N, 61.2W) :  The approximate Closest Point of Approach (CPA) is located near  13.2N, 61.2W  or about  2.4 miles (3.8 km)  from your location. The estimated time of when the eye will be at that location is in about  47.6 hours .   So  this might be the  year we get slammed .......we shall see.......... '),(11657,3760558,'10,August,2004','2004-08-10 00:00:00','    Talk about a busy Friday the 13th in the tropics! Major hurricane Charley tears into South Florida on it\'s way for a quick visit with Mickey Mouse and this afternoon, we have two new depressions to worry about! We pretty much all theorized things would heat up quickly but not this quick!  TD#4 which has now just been named  DANIELLE  is brewing in the far eastern Atlantic and is looking to reach here in the islands around wednesday or thursday of next week and TD#5 (Soon to be called  EARL ) is less than a 1000 miles away. TD#5 is quite worrisome due to it\'s proximity to the islands but it did form really low, 8.9 north, and just might crash through the lower islands of Trinidad/Tobago, Barbados, Grenada and maybe even Vincy. But you never know, it could potentially turn more northwest so it\'s definitely something to watch carefully.  TS Daniella is an interesting system with a long ways to go so I\'m not going to really comment on it at this time as more reseach is necessary. '),(11658,3760558,'10,August,2004','2004-08-10 00:00:00','    urlLink    Don\'t tell me Castro was the gardener!!!! '),(11659,3760558,'09,August,2004','2004-08-09 00:00:00','    urlLink    Operation Cleanup in Cuba '),(11660,3760558,'09,August,2004','2004-08-09 00:00:00','    As luck would have it, as soon as I say pictures are on its way from the depression passing through Vincyland, my camera packs up and my jeep battery dies meaning I can\'t show you nuttin\' Oh well, it passed by us and as usual it developed into somthing sinister further on up the islands. The next set of pics shows some damage to Cuba on the way to the US '),(11661,3760558,'03,August,2004','2004-08-03 00:00:00','    urlLink    Looking good...................... '),(11662,3760558,'03,August,2004','2004-08-03 00:00:00','    Well it\'s been a quiet one for this year so far but as soon as August comes around we have two popping up. Alex is hugging the coast of the US but TD# 2 (aka Bonnie) is getting close to home. I\'ll update with pictures over the next day or so....but off course, we\'ll survive as always. Remember we\'re Land of the Blessed. '),(11663,743739,'16,July,2002','2002-07-16 00:00:00','    My fellow smurfs, today we fight the battle of our lives, and if we go down, we GO DOWN SMURFING.   Ok, sorry, going through my video collection and saw this gem. So of course I had to share it   Due to pending legal action, smurfs may be called blurfs.   '),(11664,743739,'25,June,2004','2004-06-25 00:00:00','    urlLink Offbeat Dollars for Scholars : \"The annual duck celebration in Stuttgart, Ark., was winding down ? the Queen Mallard beauty pageant was over and the world\'s best duck dog had been determined. Then Daniel Duke stepped onto the Main Street stage.   Duke, a teenage veteran of more than a dozen duck-calling contests, wowed the judges with his renditions of the four required blasts: hail, feed, comeback and mating. Duke, from the nearby town of Brinkley, triumphed ? and bagged one of the nation\'s more unusual college scholarships.   \'I knew I had a shot at it,\' the 19-year-old said of the $1,500 award, which he hopes to use to attend the University of Arkansas. \'And I think it\'s pretty great you can get a scholarship for calling ducks.\'\"  --------------------------------------  I couldn\'t resist sharing this. My father is an Arkansas grad, I may have to call him up and ask if he received any duck hunting scholarships back in the day. Of course he will probably respond that he wasn\'t able to attend the contest because it was too far to walk in the snow, especially being uphill both ways. After I ask a follow up question, I will be informed that back \"in the day\" it used to snow in Arkansas even during the winter.   '),(11665,743739,'02,July,2004','2004-07-02 00:00:00','    urlLink Northwest Indiana News: nwitimes.com - News: Chew on this : \"The gesture was part of an elaborate plan cooked up by boyfriend Magic Mackeigan, who left town Thursday morning on a flight to Texas and was unavailable for comment.\"   ---------------------------   Ok, so the story is cool and threatened to be my subject. However, check out the boyfriends name. Magic Mackeigan. I bet his parents purchase items sold in spam email messages.   '),(11666,743739,'29,June,2004','2004-06-29 00:00:00','    urlLink Spam\'s dirty little secret - some buy its products : \"According to a Yahoo! Mail global survey of about 37,000 Internet users in 11 countries, 20 percent of U.S. residents admit buying products from spam purveyors. About 78 percent said they delete spam.\"   -----------------------------------  Ok, this is why some people need to be shot. Normally I am not a big advocate of violence, but come one. This means that every company that can get my email address will send me junk, knowing that there is A ONE IN FIVE CHANCE that I’ll buy from them. ONE IN FIVE! That is absurd. Of course if anyone of my loyal readers happens to fall in the category and is offend by my comments..... too bad, I still think you should be shot. Besides we all know that no one reads this, I don’t even read this. (But then who is \"we all?\")   '),(11667,743739,'28,June,2004','2004-06-28 00:00:00','    Hey quick one tonight as I need to be studying. I just wanted to give a quick update and make sure everyone knows that BCOM is the most evil class ever.   There, I feel oddly cleansed now.   '),(11668,743739,'27,June,2004','2004-06-27 00:00:00','    urlLink ESPN.com - GEN - ESPN is all the rage ... as a baby name? : \"A Texas couple who named their son ESPN after the cable sports network will soon be getting a visit from the toddler\'s namesake.   An ESPN film crew is coming to this Panhandle town next month to interview the family of 2-year-old ESPN Malachi McCall for a feature on several children around the country named after the network.   ESPN (pronounced Espen) McCall is one of at least three children in the United States known to be named for the sports network. A couple in Corpus Christi named their son Espn Curiel in 2000, the same year Espen Blondeel was born in Michigan.\"   ----------------------------------------------------------------  Wow, and I thought the name Asa was weird.... just kidding bud. However this article has provided me with inspiration. My first kid will be name Pepsi Coke Luttrell. Think about it, it works perfectly. Then all I will have to do is convince him (or her, doesn\'t really matter) to get a PhD and then the name will be Dr. Pepsi Coke Luttrell.    And yes, its official now, I will never show this blog to any girl that I meet.   '),(11669,743739,'17,July,2004','2004-07-17 00:00:00','    WOW, when we last talked, I had a full head of vibrant hair.&nbsp; That’s right, Vibrant.&nbsp; However, the combination of Falls Creek and YFN was just too much for it.&nbsp; For now, all of my beautiful vibrant hair lies in a trash bag somewhere in Dallas.&nbsp;  &nbsp;  Now, for those of you who are constantly looking for&nbsp;a reason to laugh at me (Martin, I\'m looking at you and your group right now) I will post pictures as soon as I can, hopefully before leaving on vacation this coming Thursday.  &nbsp;  Now to get serious.&nbsp; YFN was great.&nbsp; I consider it a great honor for God to allow me to see Him move twice in two weeks, and I pray that these experiences carry over into future youth meetings.&nbsp; I should probably type more than that, but I\'m tired, so............. BUGGER OFF!!!!!&nbsp;   '),(11670,743739,'11,July,2004','2004-07-11 00:00:00','    urlLink    Yeah, so really Im just testing this, but come on. Who couldn\'t use a little more Huddy in thier lives? Except american league batters that is.   '),(11671,743739,'11,July,2004','2004-07-11 00:00:00','    QUICK, click  urlLink HERE     Ok, with that out of the way, let me tell you about last week. Last week I had the privilege of attending Falls Creek with Glenstone Baptist Church. This was I believe my 9th year at Falls Creek, being a sponsor for six of those. Although I am cheating on one of the years, I only spent a day and a half at FC last year and I was probably technically a guest, but in my eyes I am always in charge so who cares.   Hey, back to what is important. We had 3 salvations last week, which was awesome, and if any of those three are reading this, I want you to know I am proud of you and praying that you continue to grow in your relationship with Jesus. We also had a student commit her life to ministry. To this student all I can say is I know what lies ahead for you, and I am sorry.  For those of you who can\'t tell, I\'m kidding. When called to it, there is nothing better than full time ministry.    In addition, I grew personally in my walk with God. Last week was a time of teaching for me, I taught the morning in cabin studies, and a time of prayer. While it was great to be teaching again, most of my growth came in the area of prayer. For many of us prayer is something that we do because we are supposed to, and then we sit back and hope God answers it. However I got to experience prayer as an exciting thing, and I got to see very quick results from it. Nothing is as exciting as knowing that you prayed for something one night, and then seeing it happen the next day.   Well it\'s time to pack for camp this week, we leave early tomorrow. For those of you who went to camp with me last week, feel free to post a comment below, I promise Asa dosn\'t bite.....often. You can just say hi, or even share your experiences from camp if you wish. Also, in addition to commenting, be in prayer for me and my church during the next week. Specifically pray that we see God\'s power in the area of salvations, just like we did last week.   Until I get back next week remember, same bat time, same bat channel.   Yeah, I had to do that   '),(11672,743739,'22,July,2004','2004-07-22 00:00:00','    urlLink In California, McCain compares vice presidency to being a prisoner of war : \"\'As far as me and the vice presidency is concerned ... I spent a number of years in a North Vietnamese prison camp in the dark and (was) fed scraps, and I don\'t know why I would want to do that all over again,\' McCain said at a rally in San Diego. \"  On Sunday, McCain joked about being wooed by both Republicans and Democrats -- \"next will come the Libertarian and vegetarian sides,\" McCain said.     Ok, those two quotes were really funny, now I have to go.    AWAY!!!!!!!!!!!!!!!!!!!!!!!!   '),(11673,743739,'21,July,2004','2004-07-21 00:00:00','    urlLink ESPN.com: Page 2 - Sorry, Cubs fans . . . again    Ok, had to post this, just a word of caution, use a soda cap instead of a beer cap. Makes parents much hapier.   '),(11674,743739,'20,July,2004','2004-07-20 00:00:00','    Well, this will be last post for a couple of days (at least).&nbsp; Tomorrow I leave for Oakland California, we thought about going to Chicago, but decided to see a real team play instead.&nbsp; The after a few days there, it’s onto Phoenix Arizona.&nbsp; For those of you who haven\'t heard me mention this with every other sentence I use, I did an internship there a couple of years ago and look forward to seeing everyone again.&nbsp; I’m just not sure if they look forward to seeing me&nbsp; =)  &nbsp;  Anyway, we leave at 6:30 in the morning, so I will not be here in the morning.&nbsp; If i see a reply before bed tonight, I might say something.&nbsp; If not, Asa, I leave this in your capable hands.&nbsp; If Martin gets out of hand, I expect you to head up to Missouri and pretend that he buys things from spam emails.   '),(11675,743739,'20,July,2004','2004-07-20 00:00:00','    urlLink Cartoon fans start Spongebob church    Pastor: Are you ready kids?  Congregation: Aye aye captain!  P: I can\'t hear you!  C: Aye aye captain!  P: Ooooooooooh....  Who lives in a pineapple under the sea?  C: Spongebob Squarepants!  P: Absorbant and yellow and pourous is he!  C: Spongebob Squarepants!  P: If nautical nonsense be something you wish,  C: Spongebob Squarepants!  P: Then drop on the deck and flub like a fish!  C: Spongebob Squarepants!  All: Spongebob Squarepants, Spongebob Squarepants,  Spongebob Squarepants,  Spongebob Squarepants!  AMEN    Yeah, so the above is not mine, but I found it really funny.   Oh and somehow I think I will stay at Lakeview Community Church even if a spongebob branch opens in Stillwater. (Yes I just did a search to see how sponges reproduce hoping for something funny to say here, let me tell ya, there is nothing funny about sponge reproduction.)   '),(11676,743739,'20,July,2004','2004-07-20 00:00:00','    Just like the cubs, wrigley field is falling apart mid season. &nbsp;  urlLink http://www.suntimes.com/output/news/cst-nws-wrigley20.html    &nbsp; It\'s so much fun knowing that a cub fan reads this&nbsp; =)   '),(11677,743739,'05,August,2004','2004-08-05 00:00:00','    It has come to my attention that there are people who read this and don\'t contribute their comments and responses (which means I have more than two readers....yay!).  This saddens me.  The thought that you would come here and partake of my intellectual offerings (ok this is not the time to look over previous posts) and not feel the need to contribute back makes me question your character.   Now I know that it seems like my intellectual (ok, I really hope I\'m spelling that right) abilities are enough, or even more than enough, but I like to keep \"the lesser ones\" involved (right Martin?) so please post something that allows me to make fun of you.   '),(11678,743739,'03,August,2004','2004-08-03 00:00:00','    urlLink ESPN.com - MLB - Game Track : \"It was 4-all in the fourth inning when the Yankees put two runners on with two outs. Derek Jeter followed with drive toward the corner and Dye jumped at the wall to take away an extra-base hit,  leaving his spike marks in the padded fence .\"  --------------------------------- Take that Yankee Stadium!   '),(11679,743739,'02,August,2004','2004-08-02 00:00:00','    So let me see how much of my trip I remember; at the same time recording it for future generations.   Thursday (July 22)  All day spent traveling.  We ended up in Flagstaff AZ, the most expensive hotel city in American.  Ok, probably not, but when did the truth matter here?  Friday  Hey look, another driving day, but this time with flair.  After buying cheap gas in Kingman AZ, we headed north to see Hoover Dam.  This was also our first time to drive through a security checkpoint.  Apparently we didn\'t look like we intended to blow up the dam, and was allowed to drive across.  This was also the day to see Death Valley.  All I can say is appropriate name. Hey, did you know there are no roads going across central California?  Not even minor ones like we guessed there had to be.  So, we got to go north all the way past Yosemite and take a road that looks nice and straight on the map.  On the map.  Stupid map.  So its 11 pm and we are winding through curves going up and down mountains amazed that we haven\'t passed a town with a gas station in a long time.  a long time.  And yes, for those of you who can guess where this is leading, I can now say that I have been stranded for 2+ hours in a national forest in California.  Special thanks to Daniel Lions for helping us out there.  Saturday Ok, so really that last think carried over into Saturday and we got to Oakland at 9 in the morning.  The baseball game was an afternoon game so we took a short nap and then proceeded to Network Associates Coliseum.  Yes, the stadium is as awful as you have heard.  However it was nice to see other A\'s fans.  Also it was a 6-2 Oakland win, with Kotsey hitting 2 home runs and Dotel pitching the ninth (three straight strike outs).  More importantly, I am now a 1989 World Series Champion and I have the ring to prove it.  Take that, San Fran Giants.  Oh, also I burnt my head during this game.  Not Fun.  Sunday Church at Harvest something in Fremont CA.  Good praise and worship.  The pastor delivered a good message, but is younger and still needs some polish.  All and All a great experience. After church we headed over to Fishermen’s wharf.  Pier 39 I believe.  Let me tell you, I have never seen so many shops crammed into such a small space and the crab was amazing.  Yes those of you that I talk to often will hear about the crab, then hear about the crab, and possibly hear about the crab some more.  OH, I almost forgot.  We took the Bay Bridge to get to San Fran (where the wharf is).  Awesome experience, out of the three bay bridges we took, this was my favorite.  Of course I am bias by memories of the quake hitting this bridge during the 1989 World Series (which I won). After fishermen\'s wharf, we drove over the Golden Gate Bridge.  This was nice, but the fog was heavy.  This kinda made it cool, but also made it impossible to see the top of the cables supporting the bridge. Heading north, we crossed back over the third northern bridge that I can\'t remember the name of.  Awesome.  This bridge was huge.  Yeah, that’s all I\'m going to say.  Monday If I was as cool as Willie Nelson, I would sing about being on the road again.  Instead I will just relate that we drove through Sequoia National Park, giant trees and nice mountains, on the way to Phoenix.  Another late night as we arrived at the hotel at 3am, I believe.  Tuesday Tuesday I slept.  Oh yeah and ate TCBY yogurt, losing that was the worst thing that ever happened to Stillwater.  Wednesday Umm yeah, more sleeping and TCBY that night.   Thursday Ill let you use your imagination, and yes, more TCBY.  Friday Mexico Day.  Yes, a quick jaunt into Mexico.  Also we saw the cacti at Saguaro National Park.  And driving, ended the day in Las Cruces New Mexico.  Saturday So driving through NM is boring.  The day ended in Amarillo Tx.  Some good bar b q and TCBY led us into bed time.  Oh yes, also I got to listen to the A\'s beating the rangers on rangers radio, that was fun.  Sunday Trinity Fellowship Church in Amarillo.  Rodger Hodges church, and if you don\'t know who Rodger Hodges, shame on you.  BTW Martin, he is the worship leader there.  So of course great praise and worship, Good message, then a 5 hour drive home.   So all in all it was a great trip.  Phoenix was a disappointment, my only plans were to see Bill Heaton, a local youth minister I once worked for, but our schedules never worked out for that to happen.  I must say it is great to be back home.  Even though the month started with two youth camps, only one was with \"MY\" group =), so I\'m starting to feel separated from them (not necessarily saying that is bad).  Well, goodnight and remember, same bat time, same bat channel. (the really sad thing is that I continue that joke even though I doubt the target reads this.)   '),(11680,743739,'01,August,2004','2004-08-01 00:00:00','    Ok, quick trip recap.  11 days 4 states 2 countries  And upon comming home  4 messages on the ansering machine over 4000 emails  Yes, 4000 emails.  Thats not a complete number, becuase I have had to send and recieve more than 2 times already to get them all (outlook crashes, believe it or not).  So if anyone needs me, Ill be checking email for the next three years.   '),(11681,743739,'09,August,2004','2004-08-09 00:00:00','    urlLink ebay item #3924884114   Yes, the link above gives you the change to buy a napkin containing Alan Keyes sweat.  Now I love Alan Keyes as much as the next man, if the next man happens to be a close relative of Alan Keyes, but this is a little out there.  So I will not be bidding on this napkin, however that is good news if you wish to.  It will be that much cheaper  =).    And for those of you who don\'t follow links, the current price for this sweat stained napkin is $24.25.   '),(11682,743739,'09,August,2004','2004-08-09 00:00:00','    urlLink Help on the way for bad dates  \"they\'re fake \"rescue\" calls — now being offered by two cell phone providers, Cingular Wireless and Virgin Mobile USA\"  \"And bingo, the bad date is history.”  -------------------------------------- This explains a lot. For awhile now I have been trying to figure out why Martin uses cingular instead of T-mobile (which all the cool kids use). So the question is, who should tell his wife?  Also, I really love that second quote.   '),(11683,3738552,'24,June,2004','2004-06-24 00:00:00','    So I have found yet another way to spread my mental problems around the net like so much Skippy . The ego boost almost knocked me out of this chair . But my problem lies in the fact that I have no idea what this first of many posts should be about . I think something about monkeys would be good . However I cannot think of a single thing to say about those fuzzy little poo throwing, flea eating, angelic bundles of monkey goodness. Well of course with the exception of the horrible monkey incident that befell me when I was but a young impish boy *cringe* . So maybe something deep and thought provoking , but no not enough sleep for deep thought , the synapses aren\'t firing today . So after much personal soul searching and meditation I have decided to leave this as is . I mean it\'s not like anyone is going to read this but me so it doesn\'t really matter what I say. Example: Marbles don\'t float in Ketchup . Now why would I say that ..... Because I can hahahahahaha Laugh with me , laugh with me  !! ahahahahaha   '),(11684,3738552,'02,July,2004','2004-07-02 00:00:00','    You know it\'s odd sometimes how you can find small parts of enlightenment in everyday life. Case in point , the other day I found my own little piece of nirvana in the movie Return of the king . Yes , I know a very fanboyish thing to say . However it is true , the really odd thing is where it came from . Now I have been studying religion for as long as I can remember , I started with Greek and roman in like second grade and have been chugging along ever since. It is one of my favorite topics of discussion.  Anyway in that time I have come across many different concepts of the afterlife, what it is , how you get there , ect ect . But watching that movie kind of put it in focus a bit for me . I was watching the scene where Gandlaf was talking to the bo-bo hobbit about how the afterlife was great and he shouldn\'t be afraid to die . This to me was a very powerful scene , mostly thanks to Sir Ian Mckellen\'s great acting , and great writing my one of the masters. It really made me stop and think about all the stupid stuff I deal with everyday and how it doesn\'t really matter , because when it\'s all said and done I get to go and feast and fight with my ancestors for all time in paradise .  What could be better than that , I mean aside from being a rich good looking porn star. By the way if any of you haven\'t read the LOTR series I suggest it. Tis one of those things everyone should read in their life . After that you can move on to Heinlein , the stories are just as good but a lot more debauchery .   '),(11685,3738552,'01,July,2004','2004-07-01 00:00:00','    So I get on here today praying , hoping , wishing that someone or something has read my shiny new blog post . It was not there . Now im pretty sure I saved it and looked at it after it was written but mayhaps I did not do something correctly .       Damn shame really as it was great blog work , it was funny , smart , thought provoking and witty to no end . It would have made me an instant star , women would want me and men would want to be me . Alas I am here now but because of my own stupidity or error you don\'t get to read one the greatest pieces of literature ever written . Sad   '),(11686,4271205,'23,August,2004','2004-08-23 00:00:00','   I lost my Passport and am suppose to be going on vacation in two weeks!  Now I have to have an emergency service expedite my application so that I won\'t miss my vacation.  I\'ve lost my passport before, but it was taken care of for me b/c it was the fault of the airlines for losing our luggage.  Luckily we were back in the states.  And yes, I am absent minded  enough not to remember to put my Passport in my carry-on luggage or purse.  Terrible, I know.  Anyway,  here I am trying to fill out the online customer verification forms and my computer freezes.  Okay, no problem.  I\'ll just open another window and redo it.  Now their site is down!  I\'ve got a headache now.  So I call the service and he says that the site is down, but I should try again in 15 minutes.  Okay, fine.   It\'s official.  I won\'t be able to get this stuff done until tommorrow, but the guy says I should still get my Passport by Sept. 1st. STRESS!  '),(11687,4271205,'20,August,2004','2004-08-20 00:00:00','   I auditioned for \"The Crucible\" on Monday for the part of Tituba. She is the slave in the story of The Salem Witch trials that plagued Salem, Massachusetts in the 1600\'s. It is a really powerful story and I thought that it would be a good piece to do to show my emotional range. Though I knew that if I was cast in You\'re a Good Man Charlie Brown it would be kind of hard to do both because the rehearsals would overlap. But it would have been possible.  That is, until I got an e-mail from the co-founder of the theatre company saying \"We\'ve decided to go with another actress for Tituba. We feel she needs to be more of a middle aged woman, and you\'re just too young and pretty to fit that image. :- )\"  Now I, like anyone else, love a good compliment. But, in this case, I think she was just trying to cushion the fact that I didn\'t get the part. (now would be a good time to refer to the title of my blog)   I appreciate her complimenting me to cushion the blow, but it wasn\'t necessary and certainly wasn\'t sincere. Age I can see being a problem, but make-up can make a person look the part. Just  urlLink look  at what they did to  urlLink Vanessa Williams in \"Into The Woods\" . They made her a witch. Yuk!  I am also tired of every playwright thinking that middle-aged black women are the only black actresses in the theatre. Musicals aren\'t so bad...at least not the modern ones. But straight plays seem to only portray the middle-aged or grandma-aged black woman b/c she is usually there to signify wisdom or culture. For example: The character of Tituba is a slave from Barbados, she represents both. So unless the director embraces non-traditional or color blind casting, it\'s almost a lost cause.  By the way, has anyone seen \"The Crucible\"? I rented the video and watched it before I went to the audition. I thought the movie was quite good. It\'s stars Winona Ryder as Abigail. This time she\'s trying to steal someone\'s husband instead of clothes, but it\'s still and interesting movie.  '),(11688,4271205,'19,August,2004','2004-08-19 00:00:00','   My company is going through a restructuring because, all of a sudden, the stockholders want to know everything that\'s going on and what everyone\'s specific job description is. Apparently the dividend checks were quite healthy last go-round and they see this company as a huge asset.  It\'s all fine with me, but I live in the quiet part of the building and when the contractor comes in to meet with the department heads it is quite disturbing. First of all his voice is hella loud. Even with the door closed you can hear his conversation TOO loud and clear. And since he meets with everyone individually he is here all day. I was trying to write a press release earlier and mistakenly typed what he was saying b/c he was so loud! I keep meaning to bring my headphones but I am too much of a scatterbrain to remember. Luckily this only happens every other week and when he\'s done he\'ll be gone forever, or at least until the next time the stockholders want to spy.  I guess I\'m a little spoiled b/c there are only two people in Advertising and we have this side of the building all to ourselves. They call us the Advertising snobs, b/c we rarely socialize with rest of the people here. But we really have no reason to. And if we do, there\'s always e-mail or ICQ. Hey, that\'s rhymes!  This makes me think of Office Space, b/c  urlLink Shawn  loves that movie. He said he gave it to his co-worker and she had to watch it two times to get it. I thought it was a funny movie. I think he likes it so much b/c he once was, or maybe still is, the guy who makes the TPS reports. He told me that his department has to watch a movie together today for some kind of B of A (Bank of America) bonding and he was going to try and slip in Office Space. The bank can be a weird place.  '),(11689,4271205,'19,August,2004','2004-08-19 00:00:00','   I got a call from the director of You\'re A Good Man, Charlie Brown last night and he offered me the role of Lucy. I wanted the role of Sally, but after thinking about it....a black girl in a blonde wig would just be wrong! Even if it is the theatre. The only reason I wanted to be Sally was so I could sing \"My New Philosophy\" anyway. But it\'s a really fun song. Lucy, however, is the lead female role. So I\'m excited. There are only two females in the cast and four males; Charlie Brown, Schreoder, Linus, and Snoopy.  I just finished doing Cinderella at Old Courthouse Theatre in Concord. I was Joy, one of the evil stepsisters. I had fun but was kind of overshadowed b/c we had a guy in drag playing Portia, the other stepsister. He was really funny but keeping up w/ him was a workout....and I\'ve got scars on my knees to prove it! You know Drag Queens, they are major upstagers!! Major! But the crowd loved it so I guess we did our job.  By the way \"You\'re A Good Man, Charlie Brown\" is a musical about the Peanuts characters. The director also told me who else was going to be in the cast. We have a very talented group, so I know that it\'s going to be a great show! Rehearsal starts Monday!  '),(11690,4271205,'18,August,2004','2004-08-18 00:00:00','   Well I finally decided to start a blog since I never have time to update my  urlLink website . Not to mention the fact that am extremely bored today. I have a lot of work on my desk, but doing it is boring me to death b/c it\'s monotonous search engine work.  Earlier today I decided to call my sister who just moved into a new apartment and out of my mother\'s house. She wasn\'t there but my other sister was. We talked for about 30 minutes and just as my spirits were starting to rise she tells me that she\'s pregnant. What! I was in disbelief!   Background info: My other sister (who\'s house I called) is pregnant as well. And they both aren\'t married and already have a child.   Now I\'m thinking to myself, \"Don\'t they have condoms in Charleston or are you allergic to them?\"  Then she tells me that she\'s 19 weeks. That\'s almost 5 months!! Why doesn\'t anybody ever tell me anything? If I hadn\'t called I would have been none the wiser. She then informs me that her and her boyfriend are getting married before the child comes. I don\'t quite know what I think about that yet. I like her boyfriend and all, but I\'m not sure that they should be getting married just b/c they have a child on the way. Then again, what\'s the alternative.  My mother seems to be okay with the idea of her getting married, though I\'m sure she\'s not really thrilled about the pregnancy. They are already living together so I guess the whole situation could be a lot worse. Oh, well. I guess I\'m now preparing to be an Auntie for the 5th and 6th time.  ');
INSERT INTO `msgs` VALUES (11691,3690855,'29,June,2004','2004-06-29 00:00:00','   I haven\'t updated this in a few days.  Mostly because I\'ve had nothing of real importance to add.  I worked Friday and Saturday nights, thus eliminating any semblance of a weekend, although I did spend part of Saturday at Joe\'s watching the Mets/Yankees game (we must be good luck together, as they won.  Every game we go to at Shea together they also win.  Good things happen when me and Joe are together, I suppose).  Sunday I did nothing.  That I can blame solely on me choosing to be a lazy bastard.    Monday I went to the Phillies game with my friend Jeff.  It was my first trip to the new ballpark, but I still had the sense that I\'ve been there before.  See, people criticised the multi-purpose stadiums like Riverfront Stadium, Veterans Stadium, and Three Rivers Stadium in the 70\'s.  \"They\'re cookie-cutter stadiums,\" people would protest.  And they were right.  They were built so teams could play football and baseball in the same stadium, yet were enjoyable to neither group of fans.  Now the new thing is building the seperate stadiums.  Football stadiums seem to all be the same.  I don\'t think there is any difference between them whatsoever other than the name brand on the front of the stadium and the colour of the seats.  Possibly the cuisine too, depending on geography.  The baseball stadiums are a little better, but they still all have a similar formula.  Improved sightlines and \"cozier\" atmosphere - Check. Super huge scoreboard - Check.  Barbeque stand run by former major leaguer - Check.  Major restaurant chain in left field - Check.  Expanded clubhouse shop full of gimmicks for the team to sell - Check Some sort of \"History of the Franchise\" exhibit that the old stadium couldn\'t be bothered with - Check.  Camden Yards started this trend, so it gets props for originality.  And sure, most other fields have their own little quirks to them to make them different.  But really, there isn\'t a huge difference between the Cit, PNC, Camden Yards, etc.  Of course, those are the only three I\'ve been to, so I\'ll just have to assume that the rest are like this.  If I\'m wrong, feel free to leave feedback for me.  I have yet to receive one piece of feedback, so I have no evidence whatsoever that people are reading these.  So if you\'re reading this, and even if I\'m right on my theory that all new stadiums are the same, just let me know somebody is reading this.  I am a pathetic soul, I am aware.  Anyway, what\'s on tap for this week?  Tomorrow is a \"WCW\" day.  That stands for West Coast Wednesday.  The name came because I used to go to my friend Jeff\'s house late at night for PS2 playing sessions, often arriving there at around 11 PM.  Because they started so late, around the time West Coast baseball games would start, the name \"West Coast Wednesdays\" was given to the events, whether they took place on Wednesdays or not.  Now, even the events that take place at 8 PM on a Thursday (as most of them have been) are called WCW.  I suppose another name could be given to them, but at this point, why bother?  Thursday is school and the radio show (www.wmcx.com, the link should work this week, last week we had web technical difficulties).  Friday I work.  Boo.  Saturday I\'m going to Mets/Yankees at Shea Stadium.  If you\'re there, look for the fella wearing an authentic Ty Wigginton road grey jersey.  And Sunday is the 4th of July, where I will be spending at least part of the day at work.  Double boo.  At some point this week, perhaps Thursday during my seven hour break between class and my radio shift (you read that right, seven hours of time to kill) I\'ll probably check out the new Spider-Man movie.  Perhaps I\'ll even do a review here or on the radio.  And maybe something unexpected will come up.  I doubt it.  Anyway, that\'s today\'s update.  Hopefully better ones await. '),(11692,3690855,'26,June,2004','2004-06-26 00:00:00','   Yeah, I\'m not happy about working at Domino\'s.  Anybody who knows me will not be even 1% shocked by that.  I constantly bitch and moan about working at that god-forsaken place.  But it\'s a little different this time.  Well, first, I should mention that working in Morrisville continues to suck.  I\'m still not happy working there.  So I\'ve been trying to get back in Robbinsville, now that my issue with working there is no longer there.  So I called Sam, whom I once considered to be a good pal (he used to constantly say that we were like brothers, and while I may never have reciprocated that particular feeling, I did like working with the guy and all), and he tells me that he\'ll call me if they need me to work there.  This offended me HORRIBLY.  I worked at that fucking store for six years.  Most people that worked there work there for three months and quit, usually without notice, if they even last three months.  I gave like seven weeks\' notice that I\'d be leaving, after having been the best god damned employee they ever had.  I was the best driver they had.  Nobody knew that area better than me, even after not having worked there in four months.  I was the best manager they ever had.  Every manager they had before me stole money from that place like crazy.  Just about every other person that ever ran shift in that store stole money from that place.  I never stole a dime.  Also, during the year plus that I was manager, I raised sales unlike anybody ever had before.  The previous manager, Gene, didn\'t give a crap about raising sales, just about working his twenty hours a week and getting out of there.  I earned every dollar of my salary, working the 55 hours a week I was supposed to for my salary.  With a crew that had maybe two good workers besides me, I made that store more profitable than it had ever been.  I didn\'t ask for help, I didn\'t need help.  I did it all basically by myself.  I sacrificed my summer so we could raise sales and do good business.    Then I ask for a job and I\'m told that they\'ll call me?  They seriously should be begging me to go back.  I did more for that store than everybody else that has ever worked there combined.  No matter what facet of the business you\'re talking about, I was the best that store ever had and ever will have.  They should have called me, I shouldn\'t have had to call them, but I was unhappy working in Morrisville.  And then they say they\'ll call me?  Don\'t bother, I\'m not going back to Robbinsville now.  Fuck that.  If you don\'t realise what I bring to the table enough to take my offer when I make it, I have no use for you.    Now I have to look for something outside of the pizza industry (which is probably for the best, anyway) because I really don\'t like working in Morrisville.  The whole mentality there stinks, and with me not working full-time, I doubt I can do anything to change it.  Plus, I hate working with Don.  He\'s so abrasive when it comes to dealing with employees.  The way he sarcastically cuts people down and such just pisses me off.  oh well, if you have any ideas as to a job I can pursue that can keep me out of these shitty pizza jobs, feel free to leave them in the comments section.  Finally, I got to see Fahrenheit 9/11 the other night.  I saw it Friday at 12:01 AM after my radio show went off the air (I stayed on until 11 PM, doing NBA draft coverage all night).  I have to say, I enjoyed it.  It was a movie targeted towards somebody like me to begin with, since I am quite liberal in my beliefs, and there wasn\'t a chance in hell that I\'d vote for Bush.  It is important to note that while watching these films like Fahrenheit 9/11, that Michael Moore has an agenda he is trying to push, in this case to get Bush out of office.  As such, you have to take into account that he may fudge a fact here or there to get that agenda across.  While it may be an agenda I agree with, at the same time you must consider that this is ultimately a movie that has both the purposes to entertain and teach, but mostly entertain.  So if you do watch the film, keep those things in mind.  And I\'m babbling now, so I will call it a night.  I\'ll probably give this another update on Sunday, as I seem to be falling into a schedule of updating every other day.  We\'ll see how long I can keep it up, but so far it\'s been cathartic to write these.  Until next time...fuck Domino\'s. '),(11693,3690855,'24,June,2004','2004-06-24 00:00:00','   What the hell does that mean?  It means that, while I am adding content to the blog today, nothing of note will be said.  I\'m warning you in advance, you can skip this one.  I\'m updating out of sheer boredom more than anything else.  The last two days have been uneventful.  I worked on Tuesday.  Still don\'t like working in Morrisville.  I tend to get out later, and plus I worked as a driver on Tuesday under this guy Don.  Don used to be the manager when I worked in Robbinsville originally, then left.  Then I became manager in Robbinsville years later.  Now we\'re working together again, only I feel as though I\'m much better at Don (since I\'m younger and have more energy).  This cannot possibly make for a good dynamic down the line.  We\'ll see what happens.  Today I had off.  I spent the day avoiding my dad, who was in a grumpy mood because, according to him, my mom won\'t let him spend more than $3,000 on a car.  My mom contends this is a fabrication.  I love my dad, but sometimes he imagines things in his head that simply are not so.  Then he spends hour upon hour moping about them.  I got out of the house, went window shopping for a while, then inexplicably bought a PC version of Monopoly at Best Buy for $10.  Hey, I loved playing Monopoly as a child, and $10 is pretty cheap.  I went to Subway today and bought a Grapeade just to see if the world\'s most attractive sandwich artist still worked there.  She does, but she looked younger than I previously remembered, like fifteenish.  Oh well, no more random Grapeade purchases there.  I also got a haircut.  I\'m the kind of lazy bastard that gets about four haircuts a year.  It\'s my opinion that I should grow my hair out to a reasonable legnth before going to the trouble of cutting it.  Plus, I\'m cheap.  I then came home and listened to my dad bitch and moan about this and that for a few hours before he went to bed.  Then I played some Monopoly (I have no life, remember, my friends never want to do anything) and updated this.  Tomorrow should be an interesting day.  Well, at least more interesting than today.  I have a 2:45 PM appointment with my Senior Seminar professor.  In lieu of classes, the class is meeting with her one-on-one.  We will be discussing my progress on our big paper, or lack thereof.  Actually, I will pretend I\'m making progress on said paper and hope she doesn\'t realise that I\'m full of shit.  Then at 6 PM that night, that\'s when the magic happens.  The Cox Show will beam out to listeners in the Monmouth and Ocean county areas in New Jersey, and live on the internet at www.wmcx.com.  This week, I plan on discussing who I feel should be the new PA announcer for the New Jersey Nets, my usual unusual takes on weather and traffic, somewhat-live NBA draft coverage...and definitely, DEFINITELY more Chingy.  After that, I may try to kill three more hours (or just stay on the radio, as nobody comes on after me and I doubt anybody would notice or care that I\'m staying on the air) and see a 12:01 showing of Fahrenheit 9/11 at Monmouth Mall.  It looks like it will be at least an interesting movie and I look forward to seeing it.  I guess I got some mileage out of this after all.  I\'m starting to get the hang of this and I think I\'m starting to like it.  Which means I will now not update for like three weeks.  But maybe not, so check back again later this week.  I have no current weekend plans (I work Friday and Saturday nights, and Sunday I\'m off with nothing to do) so we\'ll see if I get a weekend update (but NOT with fake news, sadly) in there.  Until then, you keep it real, holmes. '),(11694,3690855,'22,June,2004','2004-06-22 00:00:00','   Yes, my friends, we\'ve got a second update in less than three days.  Isn\'t that amazing?  It took a month for me to update the last one I had, and then I never updated it again.  The new format helps.  I think blocking Chrispy has helped too, because instead of endlessly mocking the sorry son of a bitch, I think of other things to do, like update this.  Well, not a lot has happened since the last update.  I worked on Sunday.  I needed a quick summer job, so I decided to work at a Domino\'s Pizza in Morrisville, PA.  I used to manage a Domino\'s in Robbinsville, NJ, but I left last winter for a myriad of reasons, one of which was because I took an 18 credit workload this past semester, including a 15 hour a week internship at SportsRadio 610 WIP in Philadelphia.  The other reason is kind of personal, maybe I\'ll go into detail about it one day, but don\'t get your hopes up on that one.  Let\'s just say it didn\'t leave me wanting to go back and leave it at that.  Anyway, one of my friends who used to run Robbinsville (Gene from the last update) now works in Morrisville and invited me to work there with him, and since I wouldn\'t have to deal with my Robbinsville problems, I accepted.  Now, I\'m not so sure I should have.  The people there are nice, but I kind of feel like a subsitute teacher when I\'m there as opposed to somebody who\'s in charge.  In Robbinsville, I was the man, but in Morrisville, I\'m really just a fill-in.  I don\'t know, maybe it will get better.  Now for today\'s rant.  I\'ve been reading \"The Bad Guys Won,\" an account of the 1986 New York Mets and the trials and tribulations they had in winning the World Series.  A good portion of the book talks about Darryl Strawberry and Dwight Gooden.  When I was 10 years old, there were few athletes I admired in the world more than Darryl Strawberry.  Gary Carter may have been the only one.  I remember feeling just horrible when he signed with the Dodgers in \'90.  I felt let down by one of my heroes.  With Carter, it wasn\'t the same because he was really coming up on the end of his career by then.  Darryl was entering his prime, and he and the Mets had gone through so much together, and it seemed like there should have been more chapters in that story.  In retrospect, it may not have been a bad thing that he left, because he ended up self-destructing, which was truly a shame.  But, when I think of a lot of my sports heroes from when I was younger...Darryl Strawberry was right there near the top of the list.  So was Dwight Gooden.  In fact, still to this day, a framed poster of Dwight Gooden from the mid-80\'s hangs in my room, more because I\'m too lazy to take it down than anything.  But I remember using my birthday money to buy that when I was a youngster, which meant a lot because when I was 11 or 12, I just didn\'t have a lot of money.  Then Doc started having his troubles with drugs, which is human.  I can\'t fault him for that, I guess.  Then I grew older, Doc lost whatever athletic magic he may have once held, and he left the team, throwing a no-hitter for the Yankees in \'95.   He\'s had his ups and downs, but he\'s seemingly been in better shape over the last few years than Darryl.  Still, here was another sports hero that turned out to be something less than I hoped.  Not all of my sports heroes were baseball players.  I remember thinking that when the Nets drafted Derrick Coleman, that he was going to be the saviour of the franchise.  DC had so many athletic gifts.  If ability alone determined greatness, Derrick Coleman could have been the best power forward of all time.  The next year, the Nets drafted Kenny Anderson, forming a 1-2 punch that should have carried them throughout the 90\'s.  Drazen Petrovic was also around to nail down the three\'s and provide scoring.  I even liked Chris Morris.  They made the playoffs for the first time under Chuck Daly, then the following year contended with the hated Knicks for the Atlantic Division.  Only John Starks cheating and breaking Kenny Anderson\'s wrist (I still to this day hate him for this) cost them that division.  Everything was going to be okay.  Then, everything wasn\'t okay.  First, Petro died in the car crash.  Then they drafted Yinka Dare.  Then Chuck Daly left.  Enter Butch Beard.  Beard ended up not being able to control Kenny Anderson and Derrick Coleman.  DC had such a horrible attitude, he was traded so the Nets could draft Shawn Bradley.  Then Kenny Anderson was traded.  It was awful, the team just self-destructed.  And through the years, DC and Kenny never really found their form from when they first came into the league.  DC wound up being one of the worst clubhouse influences of all time.  His attitude withered away what could have been amazing game.  Same with Kenny Anderson.  The point of all of this?  All of my athletic heroes wound up sucking and sucking hard.  Sometimes it was due to addictions, sometimes it was just because they didn\'t give a crap, but in the end, none of them were the heroes I thought them to be.  And that\'s the moral to this story.  If you\'re a kid, don\'t idolise athletes.  They are human and infallible just like anybody else.  Don\'t idolise anybody, really.  We all have our flaws as humans.  Just aspire to be the best person you can be day in and day out.  But if you do have to idolise somebody, don\'t idolise athletes.  Just because they have the skill to hit the ball hard, or to sink the impossible three pointer, or to throw a pinpoint pass in traffic to a wide receiver, does not mean they are a good person.  Respect their athletic abilities, root for the players on your favourite team as hard as you can, but don\'t idolise them.  Do your best to seperate the person they are on the field from the person they are off of it.  They are, just like you, only human.  That was long.  Next one won\'t be so long, I promise. '),(11695,3690855,'20,June,2004','2004-06-20 00:00:00','   Okay, well...I tried doing a blog and I didn\'t update it too much.  Probably because I\'ve been busy.  Okay, that\'s a blatant lie, I\'m just a lazy bastard.  But maybe without me having to go through the trouble of designing anything, blogging will be easier.  We\'ll see what happens.  So what\'s been going on?  Well, last I left you was the winter.  I finished my spring semester at Monmouth strong, pulling in two A\'s, an A-, and three B+\'s, finishing with a 3.6 GPA.  I\'m just that bad ass.  Yes, I am aware that calling yourself a \"bad ass\" in an internet blog for having a strong GPA probably isn\'t actually that bad ass.  But, I\'m just that weird.  What else has been going on...well, I\'m still single.  Don\'t see that one changing anytime soon.  I met a girl I really liked in one of my classes, and for once I didn\'t hide or anything.  I actually went up and talked to her, ate lunch with her once, talked to her outside of class.  These all seem like perfectly normal things that people do on a daily basis, but they represent progress for me because I am just terrible in social situations.  I don\'t know why, I just am.  It\'s why I have a small group of friends, and why I\'ve never really had a steady girlfriend.  I am as socially inept a motherfucker as you\'ll ever find.  Even among my friends...I have two friends I\'d consider really good, close friends (Travis and Joe).  I have two other friends who I\'m somewhat close to, but feel as though I may be drifting apart from, for whatever reason (Grim and Jeff, neither of whom I see as often as I used to).  I have two other \"friends\" whom I can\'t stand (Brandon and Chrispy).  I have my two online friends whom I\'ve never met (MJD and Ryan).  All of my friends seem to be in pairs, that\'s weird.  And I have Gene, my work friend who doesn\'t have a pair to match up with, whom I don\'t see that often anymore.  That\'s it.  And I suck at meeting new people.  Which leads to nights like tonight.  Tonight was a Saturday night.  I stayed home and watched an SNL repeat with Donald Trump.  The reason?  Well, Joe\'s married and has real world responsibilities (he\'s a school teacher, and this is his busy time of the month).  Travis lives an hour away.  Jeff is in Indiana for some reason.  Grim...well, I really don\'t hear from Grim too often anymore (other than last weekend).  Brandon and Chrispy are perfectly content to never do anything with anybody, and just sit around all night on AOL Instant Messanger while playing stupid online video war simulators.    On a tangent...let me just say that those online war simulators make me sick.  These guys are sitting around killing people on their little computer screens while people are dying out in Iraq every day.  Not just soliders, but innocent civilians whose only violation was living in Iraq.  We are in Iraq, fighting a war we have no business fighting, killing innocent civilians and other human beings.  For what reason?  So stupid ass Bush can get himself reelected.  I feel bad for the families who have lost their sons, their daughters, their husbands, their wives, their fathers, their mothers...both Iraqi and American.  This shouldn\'t be happening right now.  And meanwhile, people are sitting on their fat asses, eating Cheetos and drinking Pepsi and simulating killing these same people.  What the fuck?  Anyway, back off the tangent.  Brandon and Chrispy don\'t hang out, ever.  So they\'re out.  MJD and Ryan, they live far away.  And Gene is also married, although he hangs out with his buddies all the time anyway, but I don\'t seem to be in that group.  So that leaves ol\' Cox sitting at home watching TV, using the computer (because there\'s nothing on TV on Saturday nights because most people are out doing stuff) and wondering where everything went wrong.  I asked for a weekend day off a week from work specifically so I could hang out, and I never hang out.  That makes no sense.    Oh well, that\'s my frustration for tonight.  Maybe with this easy format to submit entries, I\'ll update this often.  Who knows.  Oh, and...time for a plug.  If you\'re in  Monmouth or Ocean county, New Jersey...on Thursdays from 6-9 PM I am on the X, 88.9 FM, Monmouth University\'s college radio station.  My shows usually range from the \"comically awful,\" to \"wow, I think I may have actually made sense for about thirty seconds there.\"  If you don\'t live in those counties, you can listen live on the internet by going to www.wmcx.com or by clicking on the link over there.  That\'s all for now, take care.  I will update again later this week, as needed. '),(11696,3690855,'31,July,2004','2004-07-31 00:00:00','   :(                :(                :('),(11697,3690855,'29,July,2004','2004-07-29 00:00:00','   I haven\'t posted an update in ten days!  God, I\'m lazy.  Let\'s see...    My cat is dying. This kind of sucks hard. Little Girl has been a part of my family since I was six or seven years old. She\'s been with us as long as I can remember. She\'s over twenty years old, and this is probably to be expected...still, I hate seeing her in pain, but I don\'t want to see her go. I\'ll miss petting her, scratching behind her ear when she\'s looking for affection. I\'ll even miss when she jumps on my lap when I\'m not expecting it and rips up a new pair of shorts I\'m wearing. It will be hard adjusting to life without her. I\'m not normally a spiritual person, but I hope for her sake that if there is such a thing as an afterlife, that she\'s well taken care of. If there isn\'t, then I can only hope to provide her last moments of life on this Earth with as much love and care as I possibly can.    Okay, that\'s our sentimental portion of today\'s blog update.  Now, on to other updates.    Still haven\'t gotten my paper back in Senior Seminar, meaning I\'ll have only a week to rewrite. This is bad, because my paper is a steaming pile of shit (unless Professor Schmidt happens to be reading this. In that case, it\'s the greatest paper of all time). Today I handed in a portfolio that was only slightly better. The work inside the portfolio was top notch. My resume is damn sharp, and when I get done with my page update, I\'m going to post it on there for one and all to see, because it is a work of art. And I designed it! Me, mister no artistic vision whatsoever, made a damn beautiful resume. Unfortunately, the work around the resume was just kind of haphazardously thrown in there. One out of my three references wasn\'t asked recently to be a reference (unless Dr. Schmidt is reading this, in that case...every reference was asked and checked and there are no problems to report!). Now coming up we have to do a final presentation of everything we\'ve learned in our years at Monmouth. This should be interesting, because we\'re supposed to be split up into groups, and I don\'t know who is in my group, because I missed class today. Note to self...don\'t miss class during introduction of group members. I\'m sure this will resolve itself, because I\'m the master of landing on my feet in situations similar to this.    My radio show is going well. So far, I haven\'t been thrown off the air. That\'s a plus. My Thursday shift has been concentrating more about talking about music of late, because none of my friends listen to this shift, so I make an attempt to actually keep the listeners who tune in. And finally, we got the Sunday situation figured out. I finally ran into the WMCX PD today, and he said he\'d make sure I\'m on the list this week and in future weeks so I can do my show from the WMCX studios on the campus of Monmouth University. Yay. I really do enjoy doing these shows, it\'s one of my highlights of the week. Editing my tapes down to only five minutes or so will be tough, but I think I can get some good stuff on there. And please, if you have the ability to listen, go to the link above and listen to my shows Thursday from 12-3 and Sunday from 3-6.    Let\'s see, what else...the Nets ripped my heart out. Fucking Bruce Ratner. 40% of a starting lineup that went to the fucking finals two out of the last three years...gone. Fuck you, Ratner. John Franco sucks. Hang it up, old man. You\'ve got nothing left but your pride, and with every despicable outing you throw, you lose a piece of that too. Wiggy might be traded...that would also ruin my day. Ty Wigginton is my favourite Mets player, and I\'m not ashamed to admit it. If he goes to the Pirates, I will not be amused. I think those are my sports beefs this week.    Things are also shaping up for my graduation party. The invites have been sent out, and most of the people I\'d expect to be there have said that they can make it. Hopefully, I\'ll hear from a few others as well. This is going to be a good ol\' time. I\'ve never had anything thrown in my honour before, mostly because I\'ve never done anything honourable. Graduating college in seven years still doesn\'t seem very honourable to me, but it should still be an awesome time. The party will be at my aunt and uncle\'s, so there will be plenty of room to commisserate, play games, some swimming...even for Joe\'s planned Commencement Ceremony, to take place in my aunt\'s backyard. I can hardly wait.    Maybe the reason I don\'t update this so much is because I put too much in one update! I\'m going to try to make some shorter updates more often. In the meantime, please leave feedback so I know people read this. My cat is crying, so I\'m gonna go make sure she\'s okay. In the meantime, you take care, youngster. '),(11698,3690855,'19,July,2004','2004-07-19 00:00:00','   Of course, by \"y\'all,\" I mean \"nobody,\" as in \"nobody reads this.\"  Nobody left any feedback for me as far as Saturday night ideas go, so I wound up going to the ROH show by myself.  At the risk of sounding like an ROH fanboy (which I am not, though I do enjoy watching some of their product) I really enjoyed the main show.  Doug Williams is a fantastic wrestler, and really needs to get his due.  His matches, first the four way match with Jay Lethal, John Walters, and Nigel McGuinness, then the Pure Wrestling title match later in the show with Alex Shelley, were both fantastic.  I really liked his exchanges with Nigel.  Since they both work British style, it looks so much different from the styles that everybody else wrestles that it looks unique and is fun to watch.  Certainly it\'s better than watching everybody work WWE style in that company, not that the style doesn\'t have its strong points, but when everybody wrestles the same style, it gets kind of boring after a while, not to mention the fact that some people aren\'t as good working that style as they would be another style.  Then you have people like Randy Orton, who would suck no matter what style they work.  :)   (That was a shout out to my man Jeff, who isn\'t reading this, because he hasn\'t read any of these blog items, but he loves Randy Orton for some reason even though he sucks).    So that was what I did on Saturday.  What else have I done lately?  Well, I was supposed to audition for Who Wants to be a Millionaire tomorrow, but they never sent me the paperwork so I\'m not going to bother.  I\'ll probably hang out with Jeff at his apartment instead.  I bought him an Andy Pettitte action figure in between shows on Saturday at the Toys R Us, so he has that to look forward to.  School is fairly easy this week, as all I have is a brief meeting with my professor sometime Thursday afternoon, then I do my show from 12-3 PM EST.  Then I go home and I\'ll be home in time for dinner, which will be nice.  I work four days this week, as I\'m going broke and need the money.  If anybody wants to give me money, feel free.    Also my parents and I have been planning my graduation party.  They\'re assuming I\'m going to graduate in August (as am I, I would be shocked at this point if I didn\'t).  Just the usual stuff, inviting friends and family, planning out what to eat.  As long as I have burgers, my dad\'s sausage and peppers, my mom\'s potato salad, some shrimp, and a good dessert, I\'ll be happy.  I\'m easy to please, what can I say.  Anyway, that\'s all for now.  Look for an update and a slight overhaul to the main Cox\'s Corner page at www.coxscorner.org soon, perhaps tomorrow.  I want to add some things to the layout, perhaps move some things around, make it more interesting.  We\'ll see what happens. '),(11699,3690855,'15,July,2004','2004-07-15 00:00:00','   And that doesn\'t make me a bad person.  No, things have been not a lot hectic, just a little hectic.  Last week, I had a twelve page paper due for my one summer class.  I have no doubt in my mind that it sucked, but I have the option to rewrite it, and I will do so.  I was just trying to get it complete instead of putting forth my best possible effort, knowing I\'d be given a rewrite option, so I didn\'t give it the full college try (which you think I would, considering I\'m in college and all, but...that\'s another story).  I\'ve had my radio show, which has had its ups and downs.  We had Grim in studio two weeks ago, and he added a different dimension that the show didn\'t have before.  We\'ll see if he makes any other appearances.  I\'d wager he would, but he didn\'t make it last week and I\'ve been unable to contact him for this week, so we\'ll see.  By the way, if you read this and want to listen, my timeslot was moved to 12-3 on Thursdays.  In addition, I have added a Sunday shift as well from 3-6 PM, if I can get access to the building this week.  Last week, I went to do my show and couldn\'t get into the building.  Not having been given instructions on how to get in when the building is closed, I didn\'t know what to do so I went home.  Hopefully, we\'ll have that issue cleared up for this weekend.  Let\'s see, what else...my 4th of July weekend kinda sucked.  I didn\'t do anything.  I\'ve been working some.  Work hasn\'t gotten much better, but I\'m kind of learning to deal with it.  Other than the paper and the radio show, nothing of importance to note about school.  Home life has been weird.  My dad has been having issues of late.  I hope he isn\'t drinking again, but I\'m not so sure.  For some reason, he isn\'t home tonight.  I just got home from work, so I have no idea if he and my mom had a fight or whatever, but he definitely is not home.  Before I left for work, he was ragging about this and that.  I think he might have started drinking frequently again, but I can\'t prove it.  I really hope I\'m wrong.  Weekend plans?  I\'m trying to talk Jeff into going to the Ring of Honor show in Elizabeth, but I\'m not sure he\'s buying that.  Grim probably won\'t go either, so I wouldn\'t be surprised if I went by myself.  Ugh.  Going to independent wrestling shows on a Saturday night by yourself = Bad.  Going to ROH shows, with ROH fans filling the building pretending matches are Match of the Year candidates when they clearly aren\'t by yourself = Really bad.  Still, I have nothing else to do, and I don\'t want to spend it at home, and Jeff wants to go to 3PW, and they suck, so ROH is likely what is in store for me.  May God have mercy on my soul.  That\'s all for now.  Nobody left any feedback for the last blog item, so I\'m assuming that I\'m writing these all for myself.  I can deal with that.  That means I can bury more people.  But I will ask this in my blog tonight.  Can anybody think of a more constructive way for me to spend my Saturday evening, keeping in mind that most of my friends will likely not want to do anything and that no matter what, I\'m spending it by myself?  If you can and I like the idea, I may just use it and give you credit in the next blog update.  But I doubt anybody will respond to this, so...touche.  Word out. '),(11700,3690855,'14,August,2004','2004-08-14 00:00:00','   Well, my 25th birthday has ended.  I\'d say I had a fun day.  It started at my friend Joe\'s house, where we got Subway and Coldstone Creamery ice cream (my GOD, is that stuff friggin\' awesome), then sat around and played video games.  After that, I drove over to my friend Travis\', where we went to Friendly\'s for dinner and had more ice cream.  I am a fat, fat man on this day.  After which, we watched a few innings of the Mets game, then I left him so he could pack (he\'s going away on a trip on Sunday), so I went over to my friend Jeff\'s house, where I dropped off McFarlane figures for him and we watched part of the opening ceremonies of the two week celebration of my birthday, aka the Olympics.  I spent most of the day out doing things, which was a nice change of pace from what I normally do, which is nothing.  Most of those things were eating, and right now, I am feeling severely overweight.  If I was the type to exercise, I would be spending an extra twenty minutes on the treadmill or stationary bike tomorrow, but alas.  Instead, I\'ll be working.  It\'s been nice to be away from work for three days, but I\'m broke, so...oh well.  I finished school on Thursday, too, making this a great week.  I got an A on my portfolio, then stood up and pretended I was part of a group presentation on how to produce a radio commercial.  I\'ve never really found myself to be the teaching type, so I kind of stood in the background and made wry observations every now and then.  I managed to get an A for Senior Seminar, which leads me to believe that Professor Schmidt was taking this class about as seriously as I was.  Thus ends my return to Monmouth, which has to be considered a total success.  I started there in Fall of \'97, and went there for two years.  A D in psychology caused me to lose my academic scholarship, since i needed to maintain a 3.0 GPA (and my GPA topped out at 2.95), so I went to TCNJ.  I failed out of there because I was burned out.  After a year off and a year at a local community college, I went back to Monmouth, where I wound up with six A\'s, an A-, four B+\'s, and a B-.  Not too bad, considering how lousy things had been for me at TCNJ.  Now, with the schooling out of the way, comes the hard part:  finding a job in the radio industry, and not in the pizza industry.  I\'m working on that one.  Two weeks from my graduation party, and I can\'t wait.  I\'ve never even had so much as a birthday party before, so it\'s exciting to have something in my honour.  It should be a fun time, and another excuse to eat like a slovenly beast of a man.  More updates to come, so keep it here, yo.  And through August, you can still listen to my radio show at www.wmcx.com on Thursdays from 12-3 and Sundays from 3-6, so please tune in and send me feedback. '),(11701,3690855,'10,August,2004','2004-08-10 00:00:00','   Word.  Not a lot to report this week.  I got my paper back in Seminar, and I got a 95.  I didn\'t expect that.  I think the reason I got a good grade is because my teacher didn\'t feel like grading rewrites.  Seriously.  Everybody in the class seemed to get an A or a B on the paper.  Oh well, I\'m not going to complain.  An A is an A.  Even if I gave a C effort on it, I\'ll take the A.  Let\'s see, what else...Little Girl is still wildly missed.  It\'s the weird times where I miss her, like when I\'m grabbing a cucumber out of the refridgerator, I almost expect her to come limping by to say hello.  It\'s weird, how I notice the little things.  My parents are talking of getting another pet, which is an about face to how they have felt over the years.  It used to be that whenever Little Girl left us, we\'d have no more pets, but I guess they realise how much they miss having a pet in the house.  They are worried, though, that if we get another cat, it will run out into the street.  See, Big Guy (our old cat) and Little Girl used to run out into the street a lot.  Big Guy got hit by a car and unfortunately died.  Little Girl got hit by a car, hid in the woods for two or three weeks, then came back one day limping like crazy with a hunk of skin missing from her leg.  While she eventually nursed herself back to health (because she was a tough cat), she never played in the street again.  My parents are afraid that if we do get another pet, the only way it will learn not to play in the street is if it gets hurt, and they don\'t want that to happen.  So we\'ll see.  On the social front, nothing new to talk about.  I still don\'t really hear from Grim too often anymore.  I\'m beginning to think I\'m probably not going to see him too much anymore.  This Friday is my birthday, and I\'m going to spend it hanging with my friend Travis and maybe my friend Joe.  Hopefully not too many alcoholic drinks will be consumed.  :)  Then in just over two weeks,  my graduation party will be the end of summer.  It\'s weird.  Normally, the end of summer has meant the beginning of school.  This year, the end of summer will be the end of school.  And I mean, the end.  I\'m graduating.  It\'s weird, though I am considering grad school.  We shall see.  And that does it for this update.  Please tune in to www.wmcx.com on Thursday from 12-3 PM and Sunday from 3-6 PM for my radio show, and other than that, take care, whomever you are.  Even if Tanya is still the only person reading this.  :) '),(11702,3690855,'02,August,2004','2004-08-02 00:00:00','   I\'m going to ask Blogger if I have the least visited blog in the history of existance.  I don\'t think anybody ever goes here.  But I post for you anyway.  Even if you don\'t exist.  Maybe you is me.  I dunno.  Today\'s update:  Little Girl is still wildly missed around the house.  She was in her own way such an important part of the house, and I really miss her a lot.  It\'s not the same without her.  It\'s difficult adjusting when something has been a part of your life for almost twenty years, and all of a sudden it\'s just...gone.  And not coming back.  Divorce must be ten times worse.  I\'m never getting divorced.  Children leaving the house or dying must be a hundred times worse.  I\'m never having kids.  Then again, sometimes I think those things aren\'t happening whether I want them to or not...  Anyway, other things.  The Benson and Zambrano trades are looking like unmigated disasters.  Bruce Ratner is still a douchebag.  I hope David Stern makes him sell the team back.  Work still sucks, but hopefully I only have a few more weeks of working at Domino\'s.  Radio show went well today.  Hopefully we can get some guests in there so I can interact with somebody.   With me being unable to take phone calls, I think it really takes away from the atmosphere of trying to do a talk radio program.  I, Robot was pretty good.  Harold and Kumar was pretty mediocre.  And I think that will do it.  Take care, invisible blog reader. '),(11703,3541469,'13,June,2004','2004-06-13 00:00:00','   Ok...this entry is about the solo walk i had in the camp. Basically there r six stations and at each station the station master will give you the password for you to tell it to the next station master. The first one is where all of us gather and have an exciting briefing >:D. This briefing is held by our senior who has graduated in the school. Each one got to take turns to begin the walk and there will be a 3-minute interval. At the first station, there were a lot of scaredy cats, mainly my juniors, who were whimpering and well...pissing me off. The senior began to tell them to stop being scared and start doing the solo walk because he will keep on telling stories about the school (real or not i dunno but i hope its real >:D) after one guy goes. 1 impression he gave me...our school is real haunted man. Let me share with you some...  1)During the Japanese occupation, our school was the headquarters of the Japanese army and there were many prisoners in the opposite of the building. In between the 2 buildings, there was a canal. The headquarters overlooked the canal. So if there were any escapers, soldiers would bring them to a room located in the headquarters and torture them. The room was called the Torture room. Now the room is a classroom located in the clock tower of our school. It was said that the room will unlock itself...because my senior said that once when he was camping in this school when it was nearing the exam period during his last year. Everyone would be studying and he had asked his friend where did he study. He said he studied in the classroom...shouldn\'t all the classrooms be locked at the night by then?  2)My school is a very famous school and it was founded by someone who is a good and famous man, by the name of Tan Kah Kee. We had a statue of him and it was said that during midnight, he would come down and patrol the whole school. And if you had a chance of bumping into him, he would ask you what the time is. If u tell him, he would give u a slap and walk away...  3)In the year 1999, there was a boy who played in the school band. Once he went to the school hall to practise. In our school hall, there were many big square holes which linked to different parts of the school downstairs. And that boy had fell down to one of the holes, straight down to the squash court. The next day, someone discovered him...dead.  So, how\'s my storytelling skills? Hope you can comment on them ^^  Well, When i had my turn of doing the walk, i erased all of spooky thoughts in my mind and changed them to CS thoughts! ^^ I was thinking about CS during the whole walk.  P.S. The creator of this blog will not be responsible if the ghost stories told are fictional.  '),(11704,3541469,'06,June,2004','2004-06-06 00:00:00','   Hey im back...and im here to talk about how the camp is like! Well the camp is quite fun to me...there was something like the amazing race, a solo walk (u get to walk around solo in the school after midnite...i\'ll be talking about it in the next post) and some fun activities. I have gain much during the camp and i mean it cos my team won in the camp and i got prizes (even tho i dont like them...)! In all, it was a wonderful experience for me ^^  '),(11705,3541469,'06,June,2004','2004-06-06 00:00:00','   Man...I don\'t believe it...my mum says if she catch me playing CS again, im in deep sh*t. In other words, she has grounded me!!! Why? Because of my results. Keep comparing me with my sister. WTF!! Life isn\'t fair!! Hope she goes out everytime so that i could enjoy myself ^^ Anyway...LIFE SUCKS  '),(11706,3541469,'04,June,2004','2004-06-04 00:00:00','   Tis the day when i created my first entry...felt excited. I couldn\'t think of any name for my blog so i used this lame name...even tho some may like it ^^ Well the june hols are here so fast and i think its gonna end fast too. So much homework!! Doesn\'t look like holiday to me. Anyway im gonna slack for sometime b4 i start work :P Many activities coming. Im excited about the camp coming soon on 9th June. Hope it will be fun. Tats all..thanks for supporting my first blog!!  '),(11707,3295677,'12,June,2004','2004-06-12 00:00:00','   Look at your car lovingly, it\'s crying out to you ... \"WASH ME!!! Pleasseeee???\". So you look up at the skies, bright clear day ... no problem. And after mowing the entire lawn, FINALLY! (After the wonderful notice you found, \"We can mow your lawn for $40, or you can do it yourself for free! By this date ____!! LOL!) You really don\'t seem to notice the sunburn you have, cuz your in too much pain otherwise. Well there\'s an hour of daylight left, what the heck? You decide to totally clean out the car, time for a \"car spa\". Go all out, after all wherever you need to go she takes ya! (It\'s a girl car, no attitude) Take it to the car wash, do the whole brush thing. Of course gotta have the \"wax option\", then topping it off with \"spot rinse\"! Wow what a difference, there\'s actually a color under all that dirt ... who would have known? Course then ya do the \"shamy thing\", no water spots a MUST! Woohooo new vaccumm machine, costs a quarter more but guaranteed to deep clean. Yeah, put the money in ... everything\'s going smoothly ... oh man, my shirt\'s getting sucked up .... PANIC! Take the top off? Yepp!!! Didn\'t see anyone around, have a sports bra underneath ... in the meantime fighting with the machine!!! Yelling at it to stop!! (yepp, it listened 3 1/2 min LATER) Ran outta quarter juice!!! Your too engrossed trying to hold onto the shirt so it doesn\'t get sucked in any more than it already is, hmmmm wonder how long those people have been there staring at me!!! Well, finally got the shirt out! By the way I wholly recommend that new vaccumm machine, it will suck ANYTHING!!! Just make sure there are no loose items within reach of the NOZZLE, body parts INCLUDED!!! Car is purring on the way home, \"me soooo clean\" .. the grill is smiling!!! Well, let\'s go one step further, detail the windows ... AHA how about the floormats? Yepp, throw them in the washer .. lil lazy ... sides no hose!!! So while your detailing the windows, you finally hear the washer ... uh oh!!! Yepp, RUN!!! I sat on the washer trying to hold it down, it finally gets done! Put the mats in the dryer, then go out to finish the windows! Kewl, how sparkly she looks! All gleaming ... well it\'s been awhile, go to check the floormats ... all dry!! Woohooooo ... just about done! Went to get the keys, floormats in hand ... guess what I walk out to? RAINNNNNNNNN!!! They never taught this in high school!!! We should really talk to the educational department about this! And here I thought, you had to do a dance around a campfire ... with a teepee involved! (no not toilet paper, a teepee ... kinda looks like a cone .. except upside down and no ice cream!)    I sooooooooo hope that your weekend has been as wild as mine has! Now go wash them cars!  '),(11708,3295677,'29,July,2004','2004-07-29 00:00:00','   OK Ladies, Here is some more ammunition for you!&nbsp;  A man was sick and tired of going to work every day while his wife stayed home. He wanted her to see what he went through so he prayed:  \"Dear Lord: I go to work every day and put in 8 hours while my wife merely stays at home. I want her to know what I go through, so please allow her body to switch with mine for a day. Amen.\"    God, in his infinite wisdom, granted the man\'s wish. The next morning, sure enough, the man awoke as a woman. He arose, cooked breakfast for his mate, awakened the kids, set out their school clothes, fed them breakfast, packed their lunches, drove them to school, came home and picked up the dry cleaning, took it to the cleaners and stopped at the bank to make a deposit, went grocery shopping, Then drove home to put away the groceries, paid the bills and balanced the checkbook. He cleaned the cat\'s litter box and bathed the dog.   Then it was already 1 P.M. and he hurried to make the beds, do the laundry, vacuum, dust, and sweep and mop the kitchen floor. Ran to the school to pick up the kids and got into an argument with them on the way home. Set out milk and cookies and got the kids organized to do their homework.&nbsp;    Then set up the ironing board and watched TV while he did the ironing. At 4:30 he began peeling potatoes and washing vegetables for salad, breaded the pork chops and snapped fresh beans for supper.    After supper, he cleaned the kitchen, ran the dishwasher, folded laundry, bathed the kids, and put them to bed. At 9 P.M. he was exhausted and, though his daily chores weren\'t finished, he went to bed where he was expected to make love, which he managed to get through without complaint.    The next morning, he awoke and immediately knelt by the bed and said, â€œLord, I don\'t know what I was thinking. I was so wrong to envy my wife\'s being able to stay home all day. Please, oh please, let us trade back.\"    The Lord, in his infinite wisdom, replied, \"My son, I feel you have learned your lesson and I will be happy to change things back to the way they were. You\'ll just have to wait nine months, though. You&nbsp; â€¦. got â€¦. pregnant last night.\"'),(11709,3295677,'07,July,2004','2004-07-07 00:00:00','   A little girl had been shopping with her Mom in Wal-Mart. She must have been 6 years old, this beautiful red haired, freckle faced image of innocence. It was pouring outside. The kind of rain that gushes over the top of rain gutters, so much in a hurry to hit the earth it has no time to flow down the spout. We all stood there under the awning and just inside the door of the Wal-Mart.   We waited, some patiently, others irritated because nature messed up their hurried day. I am always mesmerized by rainfall. I got lost in the sound and sight of the heavens washing away the dirt and dust of the world. Memories of running, splashing so carefree as a child came pouring in as a welcome reprieve from the worries of my day.   The little voice was so sweet as it broke the hypnotic trance we were all caught in \"Mom, let\'s run through the rain,\" she said.   \"What?\" Mom asked.   \"Let\'s run through the rain!\" She repeated.   \"No, honey. We\'ll wait until it slows down a bit,\" Mom replied.   This young child waited about another minute and repeated: \"Mom, let\'s run through the rain,\"  \"We\'ll get soaked if we do,\" Mom said.   \"No, we won\'t, Mom. That\'s not what you said this morning,\" the young girl said as she tugged at her Mom\'s arm.   This morning? When did I say we could run through the rain and not get wet?   \"Don\'t you remember? When you were talking to Daddy about his cancer, you said, \'If God can get us through this, he can get us through anything!\"   The entire crowd stopped dead silent. I swear you couldn\'t hear anything but the rain. We all stood silently. No one came or left in the next few minutes.   Mom paused and thought for a moment about what she would say. Now some would laugh it off and scold her for being silly. Some might even ignore what was said. But this was a moment of affirmation in a young child\'s life. A time when innocent trust can be nurtured so that it will bloom into faith.   \"Honey, you are absolutely right. Let\'s run through the rain. If GOD let\'s us get wet, well maybe we just needed washing,\" Mom said.   Then off they ran. We all stood watching, smiling and laughing as they darted past the cars and yes, through the puddles. They held their shopping bags over their heads just in case. They got soaked. But they were followed by a few who screamed and laughed like children all the way to their cars.   And yes, I did. I ran. I got wet. I needed washing.   Circumstances or people can take away your material possessions, they can take away your money, and they can take away your health. But no one can ever take away your precious memories...So, don\'t forget to make time and take the opportunities to make memories everyday. To everything there is a season and a time to every purpose under heaven.   A friend sent this to me to remind me of life. Hope you enjoy it.   I HOPE YOU STILL TAKE THE TIME TO RUN THROUGH THE RAIN.   They say it takes a minute to find a special person, an hour to appreciate them, a day to love them, but then an entire life to forget them. Send this to the people you\'ll never forget. It\'s a short message to let them know that you\'ll never forget them.   Take the time to live!!!   Keep in touch with your friends, you never know when you\'ll need each other -- and don\'t forget to run in the rain!   God makes everything but unbreakable hearts'),(11710,3295677,'07,July,2004','2004-07-07 00:00:00','   I never dreamed slowly cruising on my motorcycle through a residential neighborhood could be so incredibly dangerous! Little did I suspect.    I was on Brice Street - a very nice neighborhood with perfect lawns and slow traffic. As I passed an oncoming car, a brown furry missile shot out from under it and tumbled to a stop immediately in front of me.    It was a squirrel, and must have been trying to run across the road when it encountered the car. I really was not going very fast, but there was no time to brake or avoid it -- it was that close. I hate to run over animals, and I really hate it on a motorcycle, but a squirrel should pose no danger to me.  I barely had time to brace for the impact.  Animal lovers, never fear. Squirrels, I discovered, can take care of themselves!     Inches before impact, the squirrel flipped to his feet. He was standing on his hind legs and facing my oncoming Valkyrie with steadfast resolve in his beady little eyes.  His mouth opened, and at the last possible second, he screamed and leapt!  I am pretty sure the scream was squirrel for, \"Bonzai!\" or maybe, \"Die you gravy-sucking, heathen scum!\" The leap was nothing short of spectacular...  He shot straight up, flew over my windshield, and impacted me squarely in the chest. Instantly, he set upon me. If I did not know better, I would have sworn he brought 20 of his little buddies along for the attack.    Snarling, hissing, and tearing at my clothes, he was a frenzy of activity. As I was dressed only in a light T-shirt, summer riding gloves, and jeans this was a bit of a cause for concern. This furry little tornado was doing some damage!  Picture a large man on a huge black and chrome cruiser, dressed in jeans, a T-shirt, and leather gloves, puttering at maybe 25 mph down a quiet residential street, and in the fight of his life with a squirrel.  And losing...  I grabbed for him with my left hand. After a few misses, I finally managed to snag his tail. With all my strength, I flung the evil rodent off to the left of the bike, almost running into the right curb as I recoiled from the throw. That should have done it. The matter should have ended right there.  It really should have. The squirrel could have sailed into one of the pristinely kept yards and gone on about his business, and I could have headed home. No one would have been the wiser. But this was no ordinary squirrel. This was not even an ordinary angry squirrel. This was an EVIL MUTANT ATTACK SQUIRREL OF DEATH!   Somehow he caught my gloved finger with one of his little hands and, with the force of the throw, swung around and with a resounding thump and an amazing impact, he landed squarely on my BACK and resumed his rather antisocial and extremely distracting activities. He also managed to take my left glove with him! The situation was not improved. Not improved at all.  His attacks were continuing, and now I could not reach him. I was startled, to say the least. The combination of the force of the throw, only having one hand (the throttle hand) on the handlebars, and my jerking back unfortunately put a healthy twist through my right hand and into the throttle. A healthy twist on the throttle of a Valkyrie can only have one result.    Torque.    This is what the Valkyrie is made for, and she is very, very good at it.    The engine roared and the front wheel left the pavement.    The squirrel screamed in anger.    The Valkyrie screamed in ecstasy.    I screamed in  .. well .. I just plain screamed.  Now picture a large man on a huge black and chrome cruiser, dressed in jeans, a slightly squirrel-torn t-shirt, wearing only one leather glove, and roaring at maybe 50 mph and rapidly accelerating down a quiet residential street on one wheel, with a demonic squirrel of death on his back.    The man and the squirrel are both screaming bloody murder.  With the sudden acceleration I was forced to put my other hand back on the handlebars and try to get control of the bike.  This was leaving the mutant squirrel to his own devices, but I really did not want to crash into somebody\'s tree, house, or parked car. Also, I had not yet figured out how to release the throttle... my brain was just simply overloaded. I did manage to mash the back brake, but it had little effect against the massive power of the big cruiser.  About this time the squirrel decided that I was not paying sufficient attention to this very serious battle (maybe he was an evil mutant NAZI attack squirrel of death), and he came around my neck and got INSIDE my full-face helmet with me.    As the faceplate closed part way, he began hissing in my face. I am quite sure my screaming changed intensity. It had little effect on the squirrel, however. The RPMs on the Dragon maxed out (since I was not bothering with shifting at the moment), so her front end started to drop.  Now picture a large man on a huge black and chrome cruiser, dressed in jeans, a very raggedly torn T-shirt, wearing only one leather glove, roaring at probably 80 mph, still on one wheel, with a large puffy squirrel\'s tail sticking out of the mostly closed full-face helmet. By now the screams are probably getting a little hoarse.  Finally I got the upper hand ... I managed to grab his tail again, pulled him out of my helmet, and slung him to the left as hard as I could. This time it worked ... sort-of.  Spectacularly sort-of ...so to speak.  Picture a new scene. You are a cop. You and your partner have pulled off on a quiet residential street and parked with your windows down to do some paperwork. Suddenly a large man on a huge black and chrome cruiser, dressed in jeans, a torn T-shirt flapping in the breeze, and wearing only one leather glove, moving at probably 80 mph on one wheel, and screaming bloody murder roars by, and with all his strength throws a live squirrel grenade directly into your police car.  I heard screams.  They weren\'t mine...  I managed to get the big motorcycle under control and dropped the front wheel to the ground. I then used maximum braking and skidded to a stop in a cloud of tire smoke at the stop sign of a busy cross street. I would have  returned to \'fess up (and to get my glove back). I really would have. Really... Except for two things.  First, the cops did not seem interested or the slightest bit concerned about me at the moment. When I looked back, the doors on both sides of the patrol car were flung wide open. The cop from the passenger side was on his back, doing a crab walk into somebody\'s front yard, quickly moving away from the car. The cop who had been in the driver\'s seat was standing in the street ...... aiming a riot shotgun at his own police car.  So, the cops were not interested in me. They often insist to \"let the professionals handle it\" anyway.  That was one thing. The other?  Well, I could clearly see shredded and flying pieces of foam and upholstery from the back seat. But I could also swear I saw the squirrel in the back window, shaking his little fist at me. That is one dangerous squirrel. And now he has a patrol car. A somewhat shredded patrol car ... but it was all his.  I took a deep breath, turned on my turn-signal, made a gentle right turn off of Brice Street, and sedately left the neighborhood. I decided it was best to just buy myself a new pair of gloves. And a whole lot of Band-Aids.  '),(11711,4108942,'19,August,2004','2004-08-19 00:00:00','   It\'s Over.  I found a template thats suitable.  Today was filled with storms, and cleaning.  I cleaned for about 5 hours, and it rained when we needed to go outside.  And it rained when I was driving.  Thats not fair.  BUT, my room looks awesome, and I have to work tomorow.  Not too bad, I get to work with some cool people.  But on Saturday, It\'s going to suck.  No, tonight, I COULD NOT go see The Excorcist.  My dad was being a butthead, and said he doesn\'t want me to do it, because I have track tomorrow.  Well, track is in the spring.  I\'m doing Cross-Country.  Not track dad...      Anyways, I\'m going to go now, and listen to some cool music.       Later  '),(11712,4108942,'14,August,2004','2004-08-14 00:00:00','   urlLink     &nbsp; urlLink  '),(11713,4108942,'14,August,2004','2004-08-14 00:00:00','   Ok, Good news on my part. On Monday, I finally have an interview with the Theater. This means, that, if hired, I won\'t have to work in that infernal Bob Evans anymore. The people there are nice, but, no, no more food. Concessions doesn\'t count, because it\'s popcorn, and hotdogs. Thats not \"food\".      I don\'t know what to blog about. I just sat here for five minutes, thinking what to blog about. Hmmmmm........      Nope, nothing\'s coming to mind. Oh, funny story, that Aaron\'s already heard, but funny nonetheless. Today at work, the Chinese pony league team came in to eat, all 27 of them, and they sat everywhere. A waitress came up and said, ok, what can I get you to drink? The chinease kids said kjsdj dsjkhfhsdaja jf.....and the waitress said, \"I don\'t know what the fuck you just said.\"       It was so so so so so so so very funny.      Later      p.s. Facket, is........still alive. He\'s in Iceland on holidays, Heath...  '),(11714,4108942,'11,August,2004','2004-08-11 00:00:00','   No, not really, this isn\'t about pogs, sorry heath  I\'m trying to type this, but I have like seven thousand people talking to me. It\'s rather annoying, and one of them is explaining a lot of stuff to me. Today was filled with enjoyment, as my best friend and myself went to Budd Bear, to look at the cool STi!!!!! WOOOOOOOOOO. There was a Corvette that was just, KICKASS. It was an automatic, BOOO, but the speedometer went to 200, so, it compensated for the lack of the manual transmission.  Then we went to Panera, and we, um, disposed of, our....\"trash\". That\'s right, we were good Samaritan, and got rid of our trash.  But alas, my dad is still upset about me not wanting to run Cross Country. He\'s still giving me the cold shoulder. And it\'s rather annoying, being that he\'s my dad. I guess all I needed was some time away, and the time with Aaron helped a lot too. He\'s a great pall when it comes to that sort of thing.  I\'m done, I\'m bored, and I have a lot of cleaning to do!  Later  '),(11715,4108942,'08,August,2004','2004-08-08 00:00:00','   ok, this is my tenth post, awesomeness.  (happy)  ALSO, it was the third time I have been at the sunday night alive.  (sad) This is sad because I got to meet the new bass player, and he lives in bethel park, and has to drive out there every sunday.  That makes a lot of sense.  Doesn\'t it?  NO, the answer is no.  If you said yes, you\'re wrong.  Work was just super.......................................flous.  HA, I bet I had you convinced I was saying super, but no, it was really unnecessary for me to be there today.  Yes, it was a Sunday, but there was 1, thats right, 1 table at one point and time.  4 people, 1 table, all the othes were empty.  It was awesome, I got to stand around, but at the same time, I can\'t wait to quit, If I have the chance...Come on theater!  Tomorrow is monday, and I have nothing to do.  Someone tell me what to do,  because I\'m plum out of ideas.  I wanna do something, anything, but I probably won\'t, because I\'m in trouble with the law.  The law being my father.  Supposidley, cross counrty starts tomorrow morning, and I don\'t wanna run, but my dad swore he knows what running\'s like, yeah right, he sure does.  My coach is getting on my nerves, and I\'ll only drag the team down.  Oh well, I\'m dont blabbering.  Later  '),(11716,4108942,'07,August,2004','2004-08-07 00:00:00','   Well, now I\'m so pissed beyond reason.  Thanks Encounter.  Now I\'m not in a band anymore.  F**K  '),(11717,4108942,'06,August,2004','2004-08-06 00:00:00','   Today SUCKED.      I had to work, and it has never been that boring, EVER!!  Then after, I went to visit Aaron, but he was done, so we waited for heather and jon to come out of a movie, and went to pizza hut with them.  It was fun...But I was sad the whole time, because I still miss heather.  Ah well.      The I took a movie back, and followed an STi for a long time.  It was so cool.  It sped up, with no problem at all, and the muffler was about 3 or 4 inches from the ground. OOOOOOOOOOOO      Okay, I\'m gonna go now, because I have to work tomorrow, and it\'ll suck a lot.      Later  '),(11718,4108942,'06,August,2004','2004-08-06 00:00:00','   I don\'t know why I\'m posting today.  I have nothing to say.  I have to work tomorrow.  10 a.m. until 6 p.m.  You know, thats going to suck.  Because I want to work at the theater with my bedt homeboy Aaron.  But, as you know, sinse that makes too much sense, it probably won\'t happen.      Led Zeppelin keeps me going anyways.  I am soooooooooooooooooooooooo down in the dumps.      Sunday, I work the same hours.  16 hours of work will suck so so so so so very much, I can\'t describe it.  And I\'m low on gas, poop.  Oh well      Later  '),(11719,4108942,'05,August,2004','2004-08-05 00:00:00','   Why is it cold today? I don\'t know. I had to get some ritas for my dad, and i got money for the Senior Breakfast, which is tomorrow. It\'s going to be cool, except for crazy stacy. She can stay home. Band was annoying, because now that we have all 3 songs finished, with drill, we have to do complete run throughs, and thats what tomorrow\'s going to be. Run throughs. So I get to hear Earl not play, then blast the last note of the song.       It took me about 3 hours to finally log onto this, but now that I\'m on it, I don\'t want to close window. So I might not. But that leaves it open for my sister to read these, and post something, and then everyone will be like you\'re gay.      Not too much else happened today, so I think I\'ll split, and talk to my peeps on AIM!      Later  '),(11720,4108942,'03,August,2004','2004-08-03 00:00:00','   This blogger is really starting to aggrivate me. A lot. I can\'t log on, and this is a reoccuring occasion.  Enough!  I can\'t believe how hot it is in my house! It\'s hotter inside than it is outside, by like 20 degrees. I don\'t care. Band\'s still hard, and I wait to come home and talk to Aaron, Heath, and Brad online. Thats all i want to do when I get home. But there is always something thats urgent, that if I put it off for like an hour, the Earth will explode. That seems to happen a lot.  Taco Bell has a curse. If you eat there, you\'ll suffer later. Oh, the funniest thing happened there. I was eating there with Ryan, and he farted, but it was quiet, so I said he should make the next one louder. So he did, and it was very loud. We laughed really hard for a while, then it made me fart, and it was extremely loud, and this caused the man sitting across the restraunt to look at me. His name was Ed, huge ugly man. He was pissed. It was so funny. We cried. Then Ryan farted again, and it was long, and loud. We had to leave. BUT, as we were about to leave, Potato came and sat down.  He talked about hos he hasn\'t had sex in 3 weeks. He is deprived, I\'ll tell you. He went on and on, and then he told us why he reeked, because he was digging through the dumpster behind the place. He thought someone threw away his retainer, so we went in the dumpster, and looked for it. He was so dirty, and smelly. It was very funny. Then we left, and I went to the mall with Ryan, then I came home.  Later. Adam  '),(11721,4108942,'03,August,2004','2004-08-03 00:00:00','   Thats right...My dignity is gone, pride, BOOM, vanashed. :-D No, not really, it was very funny to see so many guys in a dress. Sinse having listened to Brad, My life has already turned around, and I\'ve been having so much fun.  Now only if I can get a job at the theater, because I dred going back to bob evans. The managers are horrible.  But Band camp for some reason seems to get hotter everyday. Not to mention the thermometer in my Jeep is broken, so it says like 108 when it sits in the sun, it\'s very funny. Because it seems like it\'s that hot.  Now, everything woud be perfect if I could play my instrument...  Later!    We Like to party, we like, we like to party!!!!!!!! da da da da da daaaaaaaa daaaaaaaaa  '),(11722,4108942,'02,August,2004','2004-08-02 00:00:00','   ok, ok, fine.      From now on, I\'m going where ever the wind takes me.  I\'m a free spirit now.  No one can convince me otherwise.  It took a very good friend to convince me of this.  His name is Brad.  He told me to wait for whatever comes along.  I was in too much of a hurry, trying to rush things that weren\'t supposed to be.        I need to rethink a lot of whats been happening...      Later      I\'ve been   Dazed and Confused   for so long...  '),(11723,4108942,'02,August,2004','2004-08-02 00:00:00','   Wow! Band\'s a killer! For about the first 3 hours, we were outside in the heat! It was so freaking hot! But, to our luck, we got to go in, when the woodwinds had to sit in the sun at 2, and that\'s the hottest part of the day! So......HA      Oh, Today, I got to see what made me sick. I saw who I want to be my lady friend, lets call her, Ms. A. I saw Ms. A with someone else, but she had already said she liked him, BUT, she also said that he rejected her. So that gives me a chance. Slim to none, but it\'s more than I had before. Also, I am getting sick of a person I play my instrument with. There\'s a part in the music, where we can either play an F, which is the top line of the staff, or we have the option to octave it, which is 4 lines above it. It\'s hard to play, and he\'s always like, hey, listen to me, BLAH BLAH, I can play it, blah b;ah, and I finally got sick of it, after he said \"there, did you hear it?\", I finally said \"shut the hell up and play the damn music!!!!!!!!\"      That shut him up. But I don\'t care if that\'s mean, his head is the size of the moon. Anyways, I am in need of some talent, as I will be asking someone if I can play the bass in a church band. It sounds like loads of fun, and I\'m really up for it. I\'m serious, I might not buy this kids stuff, but I still want to play another instrument, after the guitar, the bass sounds like awesome fun.       that\'s all I have to say right now, Talk to you later.   (you as in probably just heath)      :)  '),(11724,4108942,'02,August,2004','2004-08-02 00:00:00','   Finally, I have a Blog thingie... Maybe someone will read it. My past week has been a disaster. Well, the fact that I stayed with my best friend for a week kicked butt, but also the fact that I found out some horrible tihngs was a kick in the pants. And then when I think I found someone I could have relations with, she tells me that she likes someone, and blah blah blah. So I sit there with a smile on my face, when all I want to do is stop the car, get out and cry. I\'ll tell you what.  Oh geez. I saw Harold and Kumar go to White Castle, and it was very funny. and I also saw Anchorman again, with what I thought might be my lady friend. maybe. I still have hope.  hahahaa. Brad just called me. He\'s a funny guy. Also a very nice one too. Anyways, I\'m not having luck with anything this past week, so I think I\'ll stop while I\'m ahead, and let other people be happy, because I\'m not allowed to be. Nope. Because when I\'m happy, no one else is, and when I\'m unhappy, everyone seems to be having great fun. And when I have a lady friend, everyone else breaks up, and when I have a break up, everyone finds someone else. I am sick of it, so I\'ll stay unhappy for a while.  I\'m seeing in another light  '),(11725,4325536,'23,August,2004','2004-08-23 00:00:00','   This was all written, fairly manically, late last night. I decided that paper wasn\'t enough for all of my craziness, so this blog was born.       \"I guess it\'s time to start recording the emptiness. The question is, how do you record \'empty\'? How do you quantify a void? There\'s nothing inside me. I\'m just a mass of skin & blood &amp; viscera. Sounds suicidal, right? Except I\'m not. I have no desire to kill myself. I just want to bleed. Cutting\'s a great thing. I\'m glad I learned about it. Strange how those TV shows--news shows, sitcoms, dramas--they were trying to create awareness in order to help  prevent  the problem. I saw it as a new form of self-torture. Much cleaner, neater than just ripping at my skin. Acne is such an ugly condition. My solution has always been to squeeze, rip, pick, anything i could do to get rid of it. I can\'t let them be. I have to tear at them.        \"Ok, let me start over. I\'m sick. Obviously. Not just gross, but sick, truly sick. So far, after 2 sessions, the psych doc says I have depression (duh) & OCD &amp; possibly ADHD and bipolar disorder. Personally? I think he\'s lowballing. I\'ve got to have more than just that wrong with me.       \"I have been physically destroying myself for over half my life. As long as I can remember I\'ve picked at scabs. There was something about it that fascinated me. How the scab came off, how if you picked it the wound bled again, how different wounds created different types of scabs, how picking the scabs off altered the wounds....It was just amazing.       \"When I started getting acne, the reason for the picking changed. Acne was this horrible, ugly invader. It was something foreign that needed to be gotten rid of. So what was the obvious solution? In my diseased mind, at least, the solution was to scratch them off. Then, one time, I decided to examine a zit more closely. I got real close to the mirror, put an index finger on either side of the zit, & pressed, trying to get a closer look. Now my fascination kicked in again. I was almost mesmerized by the white substance that came out. Now I had yet another way to destroy the enemy: acne.       \"How did I explain the growing scabs on my face? When it was in the first stages, my story was always that I \'fell off my bike\' or some similar excuse that clearly didn\'t fit with the wounds on my face. As my acne spread, there really wasn\'t much of anything to say about it. It eventually evolved to the point where I would pick at anything that even resembled acne or blackheads. Blackheads: a new fascination. Instead of a white liquid-like substance, a dark, solid material came out when I squeezed. Amazing. Freakin\' amazing.        \"Now no part of my body was safe. My shoulders, upper arms, & chest all have very large pores--large pores that behave just like blackheads. These same areas later became prone to acne. For years, I hated to go out in public in a tank top. I hated people giving me those pitying looks & saying, \'oh, did you have chicken pox, dear?\' It made me so mad. To this day, when people ask \'What happened to you\' I simply say \'bad skin\' &amp; leave it at that. I think anyone rude enough to ask doesn\'t deserve much of an answer, if any.       \"After I got to the shaving right-of-passage that most non-tree-hugging women go through, it got worse. Ingrown hairs and irritated follicles became targets for destruction, just like acne. At some point, I started shaving my arms, too. I know partly where that came from, though; my older brother had called me \'gorilla arms\' several times when I was a kid. He meant that my arms were disproportionately long (they are), but all I could see was how hairy they were. Not dark hair--it was so light that it was probably barely noticeable to others. I didn\'t care. It was too thick & it had to go.       \"Shaving my arms only made the problem of my scars & sores more noticeable. This continued when I matured further, started having sex, &amp; started shaving my bikini area. All of it. Razor rash is a bitch. It\'s also yet another thing to pick at.        \"It was several years ago when I first tried cutting. I had read magazine stories about it & seen TV reports about cutting &amp; it just seemed like a good way to vent. I already had been using tweezers almost daily, plucking hairs from almost anywhere I had them (except my head). Any abnormal hair--too long, too dark, ingrown--got removed. The sores & scars were even worse than those from the acne. I hated wearing shorts because of it. But I digress.       \"The cutting was very timid at first. I made feeble attempts at making small--tiny, even--cuts on the side of my left thumb. I used knives from the kitchen (which weren\'t very sharp). Not getting any satisfaction from this form of self-destruction, I stopped cutting for several years. Suddenly, about six months ago, I started again for some reason. I can\'t remember exactly why. I know that I had this growing, frantic urge to find a box cutter that I had somewhere in my mess of a room & use it. First off, again, were small cuts on the side of my hand. This blade was sharper--the cuts bled. The satisfaction from this was great, a great release.\"       That was all from yesterday. A few hours of ranting, recorded on the WorldWideWeb. If you\'re not too sickened, I will be adding to this blog as a form of venting and releasing my inner, hidden self. Read with caution.  '),(11726,3489768,'09,May,2004','2004-05-09 00:00:00','   Iâ€™m going to try and actually update this journal as often as possible and not make them really short, dumb entries.   Yeah, so anyway this weekend wasnâ€™t too shabby. Last night I went to go see Almost Fandango and a few other bands. West Gate was playing too. West Gate is my sisterâ€™s boyfriendâ€™s band. Anthony is the drummer. Anyway, West Gate kicked ass. Almost Fandangoâ€¦ theyâ€™re talented and all but I think people make too big of a deal out of them. Theyâ€™re good, donâ€™t get me wrong, theyâ€™re not that big of a deal. Not to me anyway. I went with Ryanne, Amy, Zoe, and Danielle. Hector was there too. Great. Yeah, that was sarcasm. Anyway, that was the first show (not just local) that Iâ€™ve been to that I was scared to go into the pit. It was... scary. Manâ€¦ I wish I could dance like Dusty and those other guys. They were skanking the whole night. They were awesome. Anyway, Ryanne met this guy named David at the show. He seems cool. At like the end of the show they were like holding hands and cuddling and shit. I was likeâ€¦ â€œwow, I wish I had that kind of luck with guysâ€¦â€ haha. I thought Patti (my sister) would come but she didnâ€™t.   I have  no  idea why, but Patti has been a  bitch  lately. Sheâ€™s so snappy about EVERYTHING. Like, Iâ€™ll ask her a simple question and sheâ€™ll start bitching at me over nothing. My dad and I were talking about how whenever sheâ€™s arguing with someone, she always has the impression that sheâ€™s right. She always thinks that her opinion is the right one. I dunno, Pattiâ€™s cool to hang out with but she needs to chill out. I donâ€™t know. I kinda feel bad for Anthony â€˜cause he loves her  sooo  much but sheâ€™s so strung out and so bitchy all the time. I donâ€™t know how he handles it. Poor guyâ€¦ Anywho!   I helped my mom deliver flowers today for Motherâ€™s Day. My Gramma has worked at Mercerâ€™s Florist sinceâ€¦ before I was born. And my sister works there too, and my mom and I help out sometimes so itâ€™s sort of a family thing I guess. Tonight I went to Kasey Bloweâ€™s party. It was pretty cool, a lot of people were there. I went with Beth. I mostly just hung out with Keith, Caroline and Matt, and Beth. Oh, and Alex. Haha. It was coolâ€¦ tired me out though.   Butâ€¦ on Monday my friend Shaun asked me out. Heâ€™s such a sweet guy and heâ€™s so niceâ€¦ but I feel horrible because I said yes. I shouldnâ€™t feel bad. =\\ I feel bad because I said yes when I didnâ€™t really have those types of feelings for him. And this guy I talk to on Yahoo (he lives in Canada, yay! Haha) said that the longer I string him along, the more it will hurt him. He said he likes me a lot, but I justâ€¦ I dunno. Heâ€™s my buddy so it doesnâ€™t really even feel like we go out. God, this sucksâ€¦ I dunno what Iâ€™m gonna do. I know that Iâ€™m going to break up with him, but I really donâ€™t want to hurt him. Damnâ€¦ Well, Iâ€™m exhausted and Iâ€™ve had a long day, so Iâ€™m gonna go to bed. Or try to. I guess Iâ€™ll just lay in bed and thinkâ€¦ at least itâ€™s comfortable there. Night, guys.   Love,   Cassie   '),(11727,2882045,'26,February,2004','2004-02-26 00:00:00','    Bear with me!! Im trying to get a good post in...it\'s not that I dont\'t care, my life\'s just boring as hell!   '),(11728,2882045,'22,February,2004','2004-02-22 00:00:00','    Wow...yet another normal day. I finished my English paper this morning. Chilled and played some games. Played with Jessica (platonically, you perv!), memorized my lines for Forum....or at least some of them. Fished for a while. Rinse. Repeat. I need to wash that. Strange tingling sensation. What exactly makes something sharp cut you? As I talk with one of my friends whose weekend consisted of going out with friends and having a good time without feeling guilty about it, I ask myself a very important question: do I want to live a boring life because I\'m smart and realize that in the long run it won;t be worth it or would I rather live life to the fullest and make the most of every day, not worrying about the repercussion but instead milking the moment for every sweet drop of pleasure it\'s worth. Damn...I need to wash that.   '),(11729,2882045,'21,February,2004','2004-02-21 00:00:00','    Hey everybody, hows it goin?!? Sorry I didn\'t get the chance to post yesterday, I was some sorta tired! Anyways, rehearsal kicked ass yesterday. That\'s about it for yesterday. Woohoo. Today was just as boring: I began the driving portion of my drivers\' ed class today and boooooy did it suck. Not so much that I drove poorly, but more that he made me follow every damn rule in specific...I thought I was going to explode. I\'ve been thinking about the paper we have to write for history and I think I\'ve decided to do \'ol Willy...that\'s Shakespeare, dumbass. Yeah. My neck hurts. Hmm...maybe if I took some medicine for it. Juggies are cool. When ice melts in water, it makes this cool half-water, half-milk mixture. I\'m still single. Yep..single.... I beat badgers with spoons.   '),(11730,2882045,'19,February,2004','2004-02-19 00:00:00','    Hey everybody! Wow...today was the best day I\'ve had in a loooong time...knock on wood. The first thing was my finding out that a hot girl at my school might have a thing for me. Nice. Then, I got a 95 on a geometry test that I was sure I failed. That right there is enough to make my day. You don\'t understand my lack of talent when it comes to geometry. I suck...no, I  more   than suck...like, I  really   suck...bad. ANYWAYS. Then we had out first rehearsal for forum today and it kicked ass...except for the fact that my solo sucked...and no one could do the dance...and one of my best friends (I told you I\'d put you in here!) wasn\'t there...ok, so maybe it wasn\'t the greatest rehearsal ever, but it\'s better than sitting on my ass at home pondering the questions that have plagued mankind forever. Like what are those weird little spots that appear when you stare at a light for a long time....or why cheese is yellow...or who the \"they\" in \"that\'s what they say\" is....or what in God\'s name that awful smell is...damn, what  is  that?!?   '),(11731,2882045,'18,February,2004','2004-02-18 00:00:00','    Hello, readers! (yep...all three of you) How\'s it goin? Good? Good. Yeah my day was actually pretty good today. Yep...I haven\'t been this tired in a long time. So the only thing that happened me today was an exercise in drama. The exercise consisted of a classmate and I pretending our mother was seeing a man who beat her. We decided to take matters into our own hands and cut his break lines. THing is, our mom took his car and we ended up killing her. I figured it would be difficult but....wow. When my partner walked into the room with tears in her eyes...I was stunned. I know it kind of sounds stupid, but this exercise really opened my eyes to the true power of acting. When someone is so good at what they do that they can tap into their psyche and make themselves truly feel that they killed their mother and bring forth all of the feelings and reactions that accompany it... that is truly amazing. Damn. I only hope that I can do that someday. Oh, and I was a smartass to Knives today...it felt so good. Hehe. Ciao.   '),(11732,2882045,'17,February,2004','2004-02-17 00:00:00','    This is my most unique post yet! Why? Because I am posting from school! Woohoo... Yeah I\'m in study hall and contemplating shoving a pencil in my eye out of boredom. Anyways.... I have a geometry test in two periods and I\'m praying that I pass. Man I suck at geometry. Well, I should go now...see yah later. Sorry, I\'m too tired to say anything funny.   '),(11733,2882045,'17,February,2004','2004-02-17 00:00:00','    Hey Everybody, how yah doin. Well today was quite the interesting day. Earlier this year, I was asked by a teacher to enter a poem competition. A  SPANISH  poem competition, mind you. Now it didn\'t take me very long to figure out that I didn\'t want to do this competition. Problem is, I already told the lady that I\'d do it. Great. So all this time, I\'ve told her I\'d do it, not because I want to, but because I pitied her and I felt like I was letting her down. So today, we got a test back and I told her that I was disappointed in my grade and that I thought I did better. The following is pretty much the dialogue between us:  me: Aw man, I could\'ve sworn I did better than this knives: Yes, you can do better, but all you care about is  Drama  (Please note that I put that in italics because she made some damn queer motion with her body and hands to accompany her jab at my interests) me (quite offended): Actually, I spent all weekend working on a history outline. knives: Oh,  SORRRRRRRRY!  (ONCE AGAIN SAID QUITE SARCASTICALLY)  At this point, I sat my test down and walked out.  Now I finally have a reason  not  to o the damn competition. How DARE she question my priorities. Personally, her class isn\'t very high on my list, especially not now. Hell yeah I care more about drama than her stupid competition. Damnit...some people, huh.    Anyways I\'m pissed, you don\'t care, my head is about to split open, and I\'m tired as hell. Ciao.   '),(11734,2882045,'16,February,2004','2004-02-16 00:00:00','    Yeah so I\'m bored so I\'m going to post again. I saw 50 First Dates last night. It was ok. The movie itself was funny as hell yet quite unbelievable. Spoilers Ahead. In the film, some chick (Drew Barrymore) meets some playboy (Adam Sandler...talk about some poor typecasting). Badaboom, badabing,They fall in love.  Thing is this chick has a disorder where she has no short term memory (aka, she can\'t remember the day before, therefore everyday forgetting Sandler). Cue Hilarity. Sandler then uses his shy wit that we all so fucking love to figure out whats going on. He finds that the people in the chick\'s (damnit, I\'m just gonna call her A and Sandler shall be B) life have been lying to her and making her think that everyday is the day after the day she had her accident. Uh huh, that\' exactly what I said. Anyways, the people work so hard at making A (remember the variable she was given?) think that it\'s the day after the accident that the viewer will curl their lip in disgust and feel as if they had just received a donkey punch in the kidneys. I mean come on, is this how low Hollywood has sunk to? They figure if they throw enough random shit at the viewer, maybe they won\'t realize that at this point the writers have run shit out of ideas. I\'d give the movie a....2 out of 5 because they might as well have given me a swift kick to the testes, stolen $5 from my wallet, and sent me on my way.   '),(11735,2882045,'16,February,2004','2004-02-16 00:00:00','    SO....where to start. This is my new BLOG (not quite sure what it stands for) . By the way, my names Rip...but you probably already knew that. So today is Monday and we\'re out of school for Presidents\' Day WOOHOO. But then again, because of the fact that my life is pitiful and boring as hell, I\'ll probably be on my ass all day while my peers go out and have fun. Well I guess it\'s better that I stay home then go out and do something stupid, only to regret it the next day. This Sucks.   '),(11736,2882045,'30,March,2004','2004-03-30 00:00:00','    Oh, and Claire was there too.   '),(11737,2882045,'29,March,2004','2004-03-29 00:00:00','    Hey...since now I never post I\'ve decided to try something new: I\'m going to begin a story and with each new post I shall continue it...why? Because my eye hurts. Now, let us begin:         Fabian had always been an outcast. You would be too if your name were Fabian. Anywho, Fabian was a young man with a special ability: in order to cope with his problems, Fabian could pause the very weaving of space and time and undo whatever stitch the fates had sewn and alter them in whatever way he wished. Useful? Yeah, it may come in handy once or twice. Thus begins Fabian\'s journey.     Day one:...  This story sucks...yes, it just hit me. Sorry to waste these thirty seconds of your life. Let\'s try this again, shall we?    There once existed a young man named Rip. Caring, Rip tried his hardest to keep everyone happy. However, this wasn\'t always the easiest task. The following are true excerpts from his daily life. Schwa.    Today was a fairly good day to Rip. On a scale of 1 to 10, one being the worst, ten the best, Rip\'s day was about a six. A little bit better than average because he is finally on normal speaking sperms with someone he cares about very much once again. As usual, his archnemesis... The evil Count Knives (that\'s right) was on the prowl once again. She issued a true test of Rip\'s will today: she forced him down on a table, restrained him with leather straps and cackled as a pendulum slowly made its way towards his crotch and Rip Jr. This process is also often referred to as a test, but hey, my story, my words, got it?!? Sadly, my Archbishop (Mrs. Beard) is departing from the school and the time has come to issue a search far and wide across the land in order to find a worthy taskmaster. Will the interviewees survive the test? Luckily, the princess smiled at Rip today, marking the first day this week that the sun has truly shined on his castle. Next week, Rip may even learn how to use the force, who the hell knows.   '),(11738,2882045,'29,March,2004','2004-03-29 00:00:00','    Well, my big mouth got me in trouble once again. I said some downright cruel shit to a person that meant a whole hell of a lot to me...damnit, way to go Rip. Maybe she can find it in her heart to forgive me...I hope so, because I miss her so much...Oh, and sorry I havent been updating my blog recently. Busybusy. Ciao.   '),(11739,2882045,'12,March,2004','2004-03-12 00:00:00','    School...devours...lives...   '),(11740,2882045,'10,May,2004','2004-05-10 00:00:00','    \"MY ASS IS ON FIRE!!!\"                Juuuusssttt kidding, Im not into the whole sodomy thing, unlike some of my classmates *cough* NASTAY *cough*. Anyways, I\'m bored so I figured I\'d post. True, I should be working on my history paper, but you can only talk about Shakespeare for so long until your sperm count begins to drop. Why? Beats me.                My day today? It was thuper, thank you!! Wore my Hooters shirt, didn\'t get chastised for it. I was elected Initiation Marshall for next year\'s Thespian Initiations...no idea what I\'m suppose to do, but the name sure as hell sounds awesome, don\'t it?!?               And in girl news...I had one call me today. She was just calling to inform me of my being voted  INITIATION MARSHALL  (DUNDUNDUNNNNNNNNNNN) for next year\'s initiates, so she only expected the call to last like two minutes. BOOOOOOYYYYY was she wrong!! I babbled on for like 45 minutes..I was doing all the talking, note you...I felt like a schoolgirl again. I hadn\'t felt that way since before the operation...back when my name was Ripina, good time for hire. But I couldn\'t believe how I just kept going on and on. Oh well.              Once again, my big mouth got me in trouble today...surprise surprise. I said mean shit about someone who really doesn\'t deserve mean shit to be said about her. Ummmm... prom\'s this weekend. School\'s almost out. Paper\'s almost finished. My hair\'s on fire. Im almost a thespian. HOLY SHIT, MY HAIRS ON FIIIREEEEEE!!!!!!!!!!!!!!!!!!!!!!!!!  \"AIIIIIIIEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE\"   '),(11741,2882045,'09,May,2004','2004-05-09 00:00:00','    Due to friggin ginormous amounts of response (most rather angry..yes, there were torches involved, pitchforks as well), I have decided to..umm...\"amend\" my post about comparing girls to bread...if you don\'t see yourself in one of those categories, you fall within the bread store next door, which I should have gone into in the first place, I was simply too blind to see it.   '),(11742,2882045,'04,May,2004','2004-05-04 00:00:00','    As I sat at lunch with my friends, we began to discuss out required reading for English-  A Tale Of Two Cities . Yes, it sucks about as badly as it sounds. Now then, we all decided that in order to spice up the story a bit, it should be rewritten as a Matrix movie. HOW KICKASS WOULD THAT BE?!? Like Darnay is Neo, Lucie is Trinity, Dr. Manette is Morpheus- HOLY SHIT I THINK I JUST PINCHED A LOAF!!! THINK OF THE POSSIBILITIES!!!! buutttttttt.... then I remembered that I\'m a lazy procrastinator who has the attention span of a 3 year old with down syndrome. Perhaps that explains why I\'m so slow to update my BLOG. Anyways... I thought it was a good idea. School\'s almost out- WOOHOOO!!!                Day two in my search for female sustenance has been futile. I\'m still horn..errr HUNGry and stuff. Yep...not many chicks to choose from at our school. Imagine you\'re in a supermarket but all they sell are three kinds of bread- Untouchable, Skanky, and Bitch. Then if you\'re really nice to the guy behind the counter and you wink at him a little, he may show you their secret and limited stash of Perfect bread...ALL OF WHICH IS ALREADY SOLD!!!  DAMNIT!!!!       But I am talking to this one chick...but shes taken.  SURPRISE SURPRISE , NO?!? oh well...she asked me to say that she was (warning these may not be true because she tends to exaggerate): shes so pleasant, good, kind, polite, fine, hott, sexy, pretty, gorgeous, attractive, good looking, beautiful, cute, appealing, sweet, nice looking, cool, funny, fun to hang out with((even though we dont)), fashionable, arousing, so incredibly nice because she came to my play,overwhelming, grand, breathtaking, splendid, remarkable, amazing, marvelous, awe-inspiring, astounding, humbling...and kind of skanky. KIDDING, KIDDING! She\'s not that polite. BADOOMBANG! Once again, just kidding- he isn\'t  THAT  skanky. *wink*    NOW GET ME THE HELL OUTTA HERE!!!!!!   '),(11743,2882045,'03,May,2004','2004-05-03 00:00:00','    CLAYA RULES! W00T!!!!!!!   '),(11744,2882045,'03,May,2004','2004-05-03 00:00:00','    DUE TO BOATLOADS OF REQUESTS (that boat being one of those toys you play with in the bathtub..no, not the big ones...yeah, about *motions with index finger and thumb*  that  big), I SHALL BE UPDATING MY BLOG, HERE GOES:      Forum is over. Ah, the sweet, sweet smell of normal life. Then again, this now means I have no excuse for skipping class and not doing homework.  DRAT, DAMN YOU DR.BADGUY, DAAMMMNNN YOOUUUUUUU!!!!  Oh well. I\'ve finally gotten over the plague that was She-Who-Must-Not-Be-Named. Her poison has been sucked from my veins and I am now back on the  International Market of Luv . W00t!  I\'d like to take a brief intermission here and offer a nice round of applause for the  BOLD  button.   Now then. Where was I? Ah yes...I\'ve grown a bit distant from my best friend. Not quite sure...I think he\'s playing with fire again as well (GIRLS). I hope things can go back to the way they were. Do I have a certain lucky lady in mind? Perhaps. We shall see. As far as my life outside of school goes...at this point, it has become nonexistent. My typical night now consists of watching TV, running, touching myself, talking on the computer, touching myself, reading, touching myself, then a nice rinse and repeat. Oh well. My life is boring. I suppose the next big thing to happen to me will be prom. Only 2 more weeks.   Ok, I\'m done. May I be excused?   '),(11745,2882045,'17,June,2004','2004-06-17 00:00:00','    Wow...it\'s been so long. This is kind of depressing actually. It\'s been like two months since I last posted and not much has happened. School\'s gotten out. I\'ve dug myself a hole and dove into it. It is a rare occasion this summer that one can find a good time to converse with me. However, my disappearance isn\'t just out of choice- I have been taken hostage. My captor, however, isn\'t some machine gun toting extremist that captured me- it was my father. Yep, I\'m a hostage in his office. I suppose it isn\'t that bad, though- I\'m getting paid. I also get my license this month. woohoo...forgive me, I\'m not as excited as most because I realize my acquisition of a license will only give my father more opportunities to tighten his grip around the balls of my summer. Oh well. Over and out.   '),(11746,3850561,'30,July,2004','2004-07-30 00:00:00','    it\'s like instant coffee....like quick fixes...like....i don\'t know....but i\'m telling ya....girls, if you are ever feeling ugly, insecure or gross, book a ticket to this island where every man looks at you like your the only woman on the planet (though we all know they have by-sides)...they call you sweetheart, lieverd, schatje, dushi.....and they are, like all latinos, instantly in love....whether they mean it or not is irrelevant...i\'m not saying come here so that you can fall madly in love...i\'m also not saying come here to get some \'action\'....i\'m just saying that if you ever feel ugly, this is the place to be....the amount of attention you get is enough to sustain you....you don\'t even have to indulge them to feel beautiful.....cold shoulder works just fine....  moving on....  i made wicked money at work to night....me and the regs are tight and this brings me more and more coin....speedy tried to hold my hand and that\'s just...gross....yes...yuck...ugh....beach is the same as always....what an exciting life.....  oh and crazy fires are burning in spain and portugal....heading to sevilla....makes me worry about a certain someone with a brand new house.....i hope things are ok....  and.....dca still sucks ass....i will tell you the dca/ssna story later....  i\'m off to msn with ms. kristy....check her out...she\'s cute.....[see blog listing right]  big kisses to all this evening.....   '),(11747,3850561,'29,July,2004','2004-07-29 00:00:00','    wednesday....just finished yet another night of wine and happas....today was south africa....nothing exciting....food was greasy.....wine was good....clientele the same as always....pretentious and wealthy with a sense of entitlement that i am beginning to have less and less tolerance for....speedy gonzales (my nickname for the little columbian chef that\'s sweet on me) continues to woe me with extra vegetables on staff meals and \"oohs\" and \"aahs\" when i enter the kitchen with dirty coffee cups.....i don\'t understand what\'s so appealing about dirty dishes....juan, one of the guys i serve with, witnessed speedy\'s efforts by spying through the window and he seems to find it equally as creepy as i do....i try to not look at him (easy as he is at least a foot shorter than i am), make sure i don\'t come within a whole metre of his person, and say i am busy anytime he calls my name.....but after a month of giving him the cold shoulder, he perseveres.....maybe i should try this uber-bitch method on boys i really like....maybe there is some truth in the whole playing-hard-to-get thing....either way, it is progressively getting more and more creepy and i think i need to either have a good face-to-face with little speedy or have management talk to him...it\'s riding my nerves....  my boss also said that i induce a sense of guilt in him when i look at him....like i am accusing him of something....like he\'s done something wrong.....i couldn\'t possibly tell you what he\'s done wrong...i don\'t really care for that matter but it is an interesting comment....he\'s not the first to comment on my way of looking.....my old boss found it seductive and this new one finds it guilt-inducing.....some say it\'s just intense....what can i say?.....i just look....directly....maybe that freaks people out....  for the rest, i haven\'t much to say....my days consist of beach and work....my nights of sleep and disturbing dreams...i wake up more tired that before i sleep....but for the rest, i\'m ok....i eat well....i rest (i think)....i\'ve lost 4 kilos....and the family is alright....they don\'t drive me crazy....the maid cleans my room....my mother makes me breakfast and lunch....how will i ever survive back in holland with no maid and no mother?.....haha....i\'m sure i\'ll be fine....at this point, i still see it as a luxury....the minute i see it as normal, i\'m likely a lost cause.....  so here i am, at one o\'clock in the morning, behind this screen....cognac....and no desire to sleep.....missing all of you who read this blog....  dames, bedankt voor je berichten....jammer dat ik niks terug kan sturen....hoop dat alles goed is in maastricht zonder mij.....juud, laat me ff weten als je&nbsp;mijn sleutel al hebt....moet namelijk iemand hebben die me post uit de bus kan halen en mijn inschrijving voor de uni opsturen....mina, hoe ging het bij vodafone???....heb je nu werk?....smelly, hoe was barcelona?....  R, dank je voor je berechten...je zorgt dat ik me altijd populaire voelt.....hoeft niet maar als je iets wilt sturen, woon ik op cas coraweg 41, willemstad, curaÃ§ao, NA....  superstar, was leuk om met je chatten.....1...2...3....4....ja....uhunh....maar ff iets anders....persevere!.....het gaat je allemaal lukken....een veel plezier met je tractors en koeien (sp?)....mis je....  &nbsp;   '),(11748,3850561,'26,July,2004','2004-07-26 00:00:00','    diane ackerman, in her book&nbsp; the natural history of love , claims that 74% of married american men cheat....i wonder if european men cheat more than americans or less?.....and what about antilleans?.....i\'m not sure if culture and socialization influence the percentage of cheaters but i am sure that socialization and culture do influence the expression and form of cheating....cheating is a broad term.....it could be anything from&nbsp;one night stand to a short fling to a long-term mistress....what has surpised me here in curaÃ§ao is the openness with which it happens...i often serve a regular and his wife on an evening that entails only an entree and house wine....and three days later, the same regular is back in doing the app, entree, dessert thing with expensive wine but the woman with him is not the wife.....she is usually young and spanish speaking here.....maybe gold-digging....maybe not....another guy came in alone and waited for his girlfriend...they chatted, drank, cuddled, kissed for an hour and then she left....ten minutes later, his wife came in....they chatted, drank, cuddled, kissed for an hour and left together....i\'m not really surprised by this all.....but what surprises me is the openness.....do these guy know others won\'t tell their wives or do they not care if their wives know?...that\'s my thought for today....   '),(11749,3850561,'22,July,2004','2004-07-22 00:00:00','    every wednesday passeniers does a wine and happas menu...they pick a country and then build the wine menu and apps accordingly....last night was spain....and they turned it into a a big tapas buffet.....tortilla, abigodias, oyster pasties, huevos atun.....good food.....but it was crazy crazy busy....normally we close up at around 2330 but last night the place was still full at midnight and before we got everything done (and we didn\'t actually get everything done) it was twoish.....but what\'s nice about working on nights like yesterday is that the sense of teamwork is much greater than when everything is manageable....and, at the end of the night, we sat down for a few drinks and chatted....we ended up going to tmf&nbsp; for a few drinks and today i am tired as hell....woke up too early....so...passeniers is getting less boring with one less person on staff and nights like last night....i know the regulars and they know me which makes it more fun....and, since i am the only girl working evenings, i get all the leftover chocolate mousse as peter says that girls get cranky if they don\'t get chocolate....so it\'s good....can\'t compain....i\'ve been here and a month and i haven\'t really wanted to go home yet which is a positive thing....i could probably live here....the mix of dutch people and antilleans and columbians and americans and tourists is appealing....u can speak dutch, english, spanish and papiomentu....would be good for my spanish....curaÃ§ao is like a developing country in a lot of ways but, then again, fairly westernized.....who knows....maybe in time i will live here....s  R - bedankt voor je smsje....was heel fijn om een berichtje te krijgen naar een maand withdrawal...hoop dat alles binnenkort wat beter gaat....mis je....   '),(11750,3850561,'21,July,2004','2004-07-21 00:00:00','    hi....i have a mobile....+59995210288....can\'t seem to send texts but maybe i can receive them...who knows....someone give it a try....   '),(11751,3850561,'18,July,2004','2004-07-18 00:00:00','    keys and money.....two things i hate....two things we would not need if people were innately good, honest and fair....but hey, what can you do.....my family locked me out of the house this morning and while most houses are easy to break into this one is not....stupid bars..... &nbsp; it\'s sunday morning...my day off....yippie.....the rest are at church and i am here.....ben leaves tomorrow so we have to do a lot of packing but he is quite ill....i hope he will be able to get it done and be able to leave tomorrow....would suck for him if it didn\'t work out....we didn\'t manage to go out before he left.....probably better....mambo sucks anyway...it\'s like holland on the beach....the only good thing is that they play 50 cent everytime and that makes me think of robby and trying to learn the lyrics under the influence...and that makes me laugh....but it\'s not really worth going....not for one song..... &nbsp; wondering how things are in maastricht....dames....geef me de roddels.... &nbsp; i\'m off to make some breaky....take care.... &nbsp; &nbsp;   '),(11752,3850561,'17,July,2004','2004-07-17 00:00:00','    last may the place i was living in was broken into while i was here in curaÃ§ao....one of the things that was taken was my stereo...what i had been listening to the night before i left was david usher and the cd was still in the player so, when they took my stereo, they also took my david usher...which i haven\'t been able to replace...until today....my brother had burned a copy from my sister last year so i burned it and listened....and almost cried.....david usher is many things to me....it reminds me of a time that in many ways doesn\'t seem that long ago and in other ways seems like a whole other life....it reminds me of late nights and port and love and an intimacy i will never be able to replace....it also reminds me of grief and the pain of knowing that you can\'t be with the most important person in your life....there were months that i couldn\'t listen....and i have to admit, it\'s never become easy to listen to....david usher was a part of my life before certain doors closed in my mind and in my heart....and in that sense, it symbolizes my greatest loss....sure that sounds really dramatic but it\'s not.... &nbsp; on a related but less tragic note, ben also had a copy of my howie day cd, which i no longer have due to a certain cat pissing on it....that too makes me happy.....hurrah for muzik! &nbsp; and donny.....xxx   '),(11753,3850561,'16,July,2004','2004-07-16 00:00:00','    hola....friday afternoon....ben leaves on monday....scary....we\'re going out tonight with ben and his friends and the boys from de tropen have invited me to go out as well which will be good.....good fun....talked to mina today....was good but i got a half-story on the george thing and it all seems so weird....can\'t wait for the details cuz apparently he seems to think differently than i do about his abrupt disappearance from my life....what\'s a girl to do?....boyz....boyz....in the end we have noone but ourselves (thx R).....and that\'s just fine....work is still work and not very interesting...it was dead last night and i have concluded that we have approximately 30 regulars that seem to think they own the place....you know the kind that demand crushed ice when they know you have no ice crusher....the kind that order stuff that\'s not on the menu cuz they know they will be accomodated and the kind that refuse to move when you say \"excuse me....can i sneak by you\"......they are also the type of people that walk into the kitchen and holler in dutch at the poor columbian kitchen staff and, &nbsp;upon being met with blank baffled faces, simply repeat the same words but in a much louder voice as if volume increases understanding of a foreign language.....they are also the type of people that feel entitled to touch you wherever whenever cuz they are paying....this is obviously the epidomy of food whoreness.....not that i am totally bothered....i can put up with a fair amount of arrogance....but i must admit that my meekness is not innate and is more a matter of just holding my tongue while my insides scream.....the joys of dealing with the rich and pretentious who fail to understand that i take shit for nine nag an hour.....   '),(11754,3850561,'14,July,2004','2004-07-14 00:00:00','    been thinking a lot about the concept of home....what makes a place home?.....and more specifically, where is  my  home?.....being here and living in a house with the same people as when i was growing up and the same furniture and the same paintings on the walls seems kind of like home....hanging out in the garden while my brother\'s friend, patrick, plays beuatiful songs on the very same guitar that the first boy who broke my heart played seems like home....so maybe curaÃ§ao is home?....but then again, an afternoon in the take with de dames also seems like home....as does a little indonesian dinner at my apartment.....as does my afternoon cup a coffee with basje....so maybe maastricht is home?.....but....sunday afternoons in don\'s basement with a cosmo and a maxim and spinach dip and croissants and red chillie jelly is also home....as is sunday night dinners at the ames-hodgson-christiansen residence....all these things give me a sense of home....but they are never all in the same place....i guess home is not a place....not here, not maastricht, not edmonton but rather home is sense of familiarity and belonging.....if i had my way, i would put all these things that make me feel at home on one nice little warm island and i would have my ladies from maastricht and my family and my scandanavians and donny and linds & co. and basje and the rest all in one place....wouldn\'t that be nice?....yup....  on a totally different note, my little brother is leaving in five days and this makes me kind of sad.....it\'s good to hang out with him and he\'s kind of all grown up and knows so many people and is so well-established here and i really really like spending time with him and actually learning about his life and his thoughts and his feelings cuz i really don\'t know him when i am there and he is here and we chat or talk on the phone....everything is always so censored....hopefully when he\'s in canada we can keep in better contact and,who knows, maybe i can go visit him and leah and donny and linds and kristy and adam and rob and rory and all the rest in december.....maybe....  ben and this guy lars came to pick me up from work last night and the place was pretty full with industry people, all of which ben seems to know...we ended up leaving passeniers quite late and went to my boss, alfredo\'s place....it is a gorgeous place...roomy, open, balcony, garden....nice....very nice....we hung out while alfredo looked at porn and the rest of us watched nip tuck...not very interesting but fun none the less.....  oh...and i lied about the picks being on the site....i\'m having problems uploading them which is not surprising given the 56k modem....yes....ancient......i\'ll try to burn them to disc and upload somewhere via cable....cool?....  alright...i\'m off....be well all....en tot snel...-s   '),(11755,3850561,'13,July,2004','2004-07-13 00:00:00','    hi....pics are on my yahoo....click on the link to the left and you\'ll get to see them....  alright....an update....  came home saturday night right pissed about this guy that i work with that like to be a table hog.....and ends up screwing me out of the money i am supposed to be making.....i was livid....so....we went to mambo but it wasn\'t really my thing that night....and when we were done there we went to denny\'s.....yeh for denny\'s.....yippie for american greasy spoon....i even ate bacon.....scary eh?.....  i spent my day off doing practically nothing....watching shit hbo movies and reading.....nothing exciting....  yesterday i was back at work all pumped up to talk my manager about the unfairness and the table hogging only to find out that martin, my sweet manager and favourite person to work with, had quit....i guess he was fed up....one of the owners is his ex and he basically gets treated like shit and i think i would have quit too.....but it kinda sucks for me cuz work has gone back to being boring and now table-hogger is going to play mr. manager....  last night patrick and his little sister came over and sat around way at the back of the yard with a bottle of wine, a guitar and ton of bats that were flying all over the place....it seemed kind of reminiscient of the campfires we had when i was a jaded angst-ridden teenager....it was good fun....  today i am off to the beach with da girls....and then off to boring old passeniers....yippie....miss you all....xxx   '),(11756,3850561,'09,July,2004','2004-07-09 00:00:00','    larissa said that it was probably a good thing that i\'m not working at de heeren....and when i asked her why she said \"because there are a lot of eligible bachelors there\"....such a comment seems ironic.....if they are bachelors and if they are eligible, why is it better that i don\'t meet them?.....not that i am out looking but it seems absurd that one would think that eligible bachelors are an innately negative phenomenon....just a thought....   '),(11757,3850561,'07,July,2004','2004-07-07 00:00:00','    so...this morning i woke up with the my first curaÃ§ao hangover....hell....i went to de heeren after work yesterday for ben\'s goodbye party and we headed out to tmf with patrick, raphael, martin (my manager) and couple of others....tmf is like stagaire heaven....holland in the caribbean...just like mambo...i drank no more than i would normally drink....however, making a highball in holland consists of a single shot of your favourite distilled liquor plus a whole lot of your chosen mix....the way it works here is a) lots of ice; b) lots (and i mean lots) of, in my case, gin; and c) a wee teeny tiny little bit of tonic....so it hit me hard and the world was spinning and it was rather unpleasant but...i had fun being out....martin is a cool guy and working has become substantially less boring since he\'s been back to work....but the restaurant is still slow and the men are still creepy and pretentious.....oh and i have a rant....let me start by saying that a server is not a slave....people with money tend to forget this....and i hate that....i served these people yesterday....young...new money...no class....they were complete assholes....they treated me like i was dumb, incompetent and a piece of meat....not cool at all....i was told to treat them well because they tend to tip well but i don\'t give a shit....my pride dictates that i am not putting up with demeaning, snide and sexually inappropriate comments for a couple of extra bucks....no thank you....so they got the bare minimum....enough that they wouldn\'t complain about me but much less than what i normally do....and i hate that....learn some decency....learn to treat all people, and especially those who work in service industries (servers, teachers, social workers, call centres), like actual people with personalities and feelings and lives....these assholes have probably never worked in such industries so perhaps they are ignorant (they were after all americans) and can be forgiven....or perhaps they have worked in services industries and somehow money has made them forget which would be totally uncool....yes....the joys of being a food whore....someday this will have to end and i will become a real person and people will stop talking at me like i am a retard....i cannot wait....   '),(11758,3850561,'07,July,2004','2004-07-07 00:00:00','    yesterday morning my mother woke me up and informed me that i had fifteen minutes to dress, eat and get to the hairdresser, who was going to do her best to get this red out of my hair....after a failed dye, highlights and two and a half hours, i walked out not quite a blond but close....so yes, ladies and gentleman, i am on my way back to my real hair colour....right now it\'s pretty funky...it\'s red and it\'s blond and it\'s just about a little bit of everything and the minute i find my dad\'s camera, i\'ll snap some pics and post them....  i worked last night (big suprise)....made some decent money in tips and sat around on the porch with ben and a bottle of wine....was good to chat....i think late at night is the only time i will really get to see him before he leaves....tonight is his goodbye party and i will be heading over to de heeren as soon as i am done work....it\'s the first time i will have been in there since i gave marc the bad news and i don\'t think marc is really mad at me....i hope not at least....  i\'m off to wake up ben and get my stuff together for the beach.....i hope the rest of you are doing well and i promise the pics will be along shortly.....xxx   '),(11759,3850561,'06,July,2004','2004-07-06 00:00:00','    ladies and gents....welcome to the new blog...it\'s much pretty, much quicker and much friendlier to the english speaking folk in my world....enjoy....xxx....s   '),(11760,3850561,'15,August,2004','2004-08-15 00:00:00','    good afternoon....today is sunday....surprise....i tried to update earlier this week but i couldn\'t log in....which really sucks cuz i felt quite inspired....and now i do not....but hey....the update....  14 days to go....mixed feelings about where i feel better....but i honestly think that maastricht and curaÃ§ao work well as homes....maybe i should do both.....my grand partner in crime (alfredo) is off in aruba this weekend so i am just laying low and catching up on some sleep....a week and a half of work to go and, although it is sometimes a whole lot of  fun, on the whole, i am little tired of working six days a week.....but hey, it\'s easy and the boys are fun and peter is back from columbia and stock is in again so maybe things are looking up for the brasserie....who knows....as long as they pay me!....tips have been alright but not fantastic....the odd american with their 20% is always encouraging.....but the sit-around-all-night-drinking-difficult-to-make-cocktails-and-leaving-seven-guilders has kinda been the norm the last while.....too too too bad....my father says that tips stands for \'to insure proper service\' and i agree....i am much more willing to do a little extra if i know said client is a good tipper....how awful is that?....the epidomy of food whore....haha....   '),(11761,3850561,'10,August,2004','2004-08-10 00:00:00','    slow night at work last night and whoever set the mix last night must have been in a slit-my-wrist mood as it was depressing and sad and all about lost love and lost family and what not....i\'ve gotten quite used to it but it is usually interspersed with a little bueno vista social club and latin american stuff....not yesterday....none the less, at the end of the night i sat with my glass of wine while waiting for juan to finish up and got thinking....thinking about what brought me here....what life has been like since i first left maastricht in february....about my travels and about what the last two months on paradise island has done for me and meant for me....today, i am all alone at the house with a little sarah slean....hanging in the yard with a book and my thoughts flow again....i am much better than when i came here....a lot more chill....a lot happier....a lot less stress....the money hassles are pretty much resolved....or at least overseeable now....but i wonder what things will be like when this vacation ends and i am back at home....being away so much has given me a lot of time to put perspective on things......and to realize that the recurring theme for me this past year has been disappointment....and this makes me wonder if disappointment is self-created or something that happens to us....it also makes me wonder if disappointment is linked to high expectations or whether the disappointment in my life is a result of a self-fulfilling prophesy....an expectation to be disappointed....u know the realism idea when you conclude that you can\'t really count on people....and i really don\'t know actually so you tell me....where does disappointment come from?.....  on another note, peter, one of my bosses that deals with all of the financial stuff, decided not to come back from columbia as he was planning too....this leaves us severely in the shit....we\'re running out of stuff and there is no money.....and i can\'t help to wonder where the money is disappearing to.....shadey businesss......i hope i get paid out this month...and i hope i still get the remainder of my july salary.....who knows.....   '),(11762,3850561,'08,August,2004','2004-08-08 00:00:00','    good evening....i know....it\'s been a week....probably longer....and i know it\'s not like me not to write....but in all honesty, i don\'t know what to write on this blog anymore....all too often i feel censored....my own issue of course....  the last week has been a week of grand insecurity....my work didn\'t pay me out on the first as they should have and five days later i got a cheque that wouldn\'t cash.....fortunately no nsf cheque cuz my mom went in person to the bank....but the fact of the matter is that i worked the last week not knowing if and when i would get paid out....which has done absolute wonders for my diminishing work morale and has made me less than kind to the boys i work with....it\'s like a pool of sharks....everyone just trying to get what they are entitled to have....i got the greater majority of my money last night....cash....the whole thing is sketchy....the one who takes care of the money is in columbia and we all wonder if he\'ll be back....either way, i better get paid this month as well....cuz this screwing around with your employees\' money/writing bad cheques thing is not so cool....  moving on...  the last week has also been a great deal of fun....i\'ve been spending some good time with alfredo....who is good and kind and (as mentioned prior) lots of fun....and hanging out with someone you feel that comfortable with makes you realize that you can leave the country, move away, take on a new life, and still you remain the same....you never lose yourself....so what this means is i am back to late nights and g&t......i don\'t mind because in the end it\'s not about partying....it about people and relationships and good friendships and i am happy to know the people i have met here.....even if it means i have crawl in the door at seven am....tonight was my night off and we went down to the coast, sat on the water and ate fantastic italian.....and why?....because i said that i really missed that sort of thing given that i work six nights a week...so yes, things are good here and i think i will miss some people....and i have built some memories....things to look back on and smile about....my own little film.....and that\'s a good thing....  in addition, larissa and my mother left for miami this morning.....i hope they have a fantastic time and it\'s all too bad that i saw so little of larissa while we were both here but work and dengae (sp?) fever seemed to get in the way.....sucks eh?.....  alright, i\'m off to smoke a ciggy.....find  my dog (wonderful loyal beast that he is) and go to bed.....ciao   '),(11763,3850561,'01,August,2004','2004-08-01 00:00:00','    good morning...i have been up for about fifteen minutes and have had two grand moments of stupidity....first of all, i got up and walked around the house looking for my mother, my sister, my father....anyone....nobody is here....nada....and i thought \"that\'s weird\".....and then i realized it\'s sunday....of course they are not here....they\'re at church.....secondly, i woke up thinking i forgot the two bottles of wine i bought at work last night in alfredo\'s car.....but they were nicely on the kitchen counter....how can i simply forget that?.....the answer....gin and tonic....i think....so....i\'ve been out two nights in a row with alfredo who is one of my bosses and a good friend of ben.....and it\'s fun....alfredo is kind, funny, generous and ridiculously dirty-minded.....and he has rob dokter syndrome....he gives everyone the impression that they are, at that moment, the most important person in his life....good stuff eh?....we went out to tutu tango on friday and he took good care of me....shooed the yucky men away and observed the cute ones with me.....last night we were at mambo....saw a lot of the same faces we saw at tutu and ran into a couple (dick and petra) we had seen the night before....so the four of us had a good time, good conversation, good fun....i enjoyed myself....  this evening the boys i work with (juan and rugerio) are coming over to eat....i hope my head can handle it today....  uhm....  dames, ik bel jullie over een uurtje....1 uur hier, 7 uur daar...dan is juudje klaar met werken.....  r, bedankt voor je sms....ik ben precies hetzelfde....maak je geen zorg....bbq klinkt leuk....had er bij willen zijn....doe de jongens de groeten van me....   '),(11764,3403602,'20,May,2004','2004-05-20 00:00:00','   Hello Hello. Well nothing has happen for a while soo yeah. I was thinking about my first luv. I feel like shit just thinking about it. I wanted to cry again. No one ever forgets thier first luv. I mean hopefully that person hasnt for got abotu me. Vuz if they did well...Yeah. There are soo many things I wish I could have done. They gave me their address. And I lost it. I feel soo dum for lossing something that could keep me connected to them. There are soo many things I wish I could have done. Im always in my room trying to think of a good song to write about it. But theres never a good song. And sometimes when I try to forget abotu it I end up finding a song about it. Im always wanting to cry about it. Soo now I hope that I can see this person once again. Cuz I think if we were really meant to be we would see each toher once again. I pray that we do. But I need to stop this thinking before I get real into it. Yeah Im done now. Bye!!XoxO  '),(11765,3403602,'20,May,2004','2004-05-20 00:00:00','   Hello Hello. Yeah today was okay. I played drums. And I took my kitty to get her shots. Poor kitty. Shes in heat. Shes like \"MEOW!! FUCK ME NOW!!!\" Yeah. One of my friend is keeping me company. Shes great! Her name is Carrie. Shes my internet sex buddie. Hahaha. Yeah. I fell in luv in with this person. But I found out that one of my friends like them. They told me to send them a message to them saying they wanna get with em. I told them that I like them but they never told me they did too. Which sux. Cuz I wanna be with the person. God! Life sux. Oh and my kitty is getting fixed. Soo her and me are gonna be single. Great. We are both in heat. We need someone in our life. We are both trapped in the same room for ever. Not wanting to go out cuz no one ever askes. Oh well. Bye!!XoxO  '),(11766,3403602,'20,May,2004','2004-05-20 00:00:00','   Hello Today is such a better day. Well I mean the whole day was a big bore. Something soo funny happen. During 5th period me and my friend Lucy were talking about some friends of ours. Then all of a sudden I heard the word \"Fight\". So I turn around and I see two guys fighting and beating the shit of of each other like if there was no tommorrow. Then I tell Lucy \"Look!\" She goes \"......Oh shit their fighting.\" And man for two seconds the teacher goes \"Hey! Stop!\" Then she jumps in to sperate them and they almost swing at her. And man oh man the teacher pushed them to the floor and kick them out of her class room. WHich they were still fighting. Lol. But yeah. I finally said soory to one of my friends mom, Allie. I got her mom\'s boyfriend mad. Long story anyways. Thats all I got to say. Bye!!XoxO  '),(11767,3403602,'20,May,2004','2004-05-20 00:00:00','   Well today I went back to skool. Ugh. Skool sux now. Its all gay and stuff. Uhmmm...Yeah. Anyways were gonna start on this Kat 6 shit. Ugh. Stupid stuff to do. I dont see the point of doing it. Its pretty much a big waste of time. All they wanna do is scare us. Some of my teachers say that it goes toward yer final grade. thats dum if it does. Which i know it doesnt. God! I swear Im gonna kick Aronld Shortza-whatevr\'s ass. Anyways enough about Mr. Germany Nazi Man...I like ice cream. Uhmmm...Espeacilly Mint and Rocky Road. Its soo yummy for yer tummy. Lol. Anyways. I hate Carmen San Diego. Die Bitch Die. I think writing songs is great. Cuz you can express yerself in many ways. Thats sounds like an essay or some shit. But yeah. I hate when people say that cant write a song. I mean no one can but you can write poems and stuff. Some dont like to express thier feelings. Geez whats wrong with it? I do it all the time, and believe me I do. Lol. But yeah. Music needs to stay with you yer whole life. And if you dont listen to music then you smoke crack. J/k. No but you need to listen to all types of music. Not just one. I suggest Liz Phair. AWSOME ASS CHICK! She rocks my wrrrld. But Im mad at her cuz she has this one guitar I like and I wanna kill her. Lol. Well yeah Bye!!XoxO  '),(11768,3403602,'20,May,2004','2004-05-20 00:00:00','   Hello Hello. Its noon now and Im here in my room. Doing nothin. Lol. Anyways...Hope evryone else is doing better. My Spring Break was great on Wendsday I think. You know with the movie. Yeah hopefully I can have more fun like that evryday. Im always home smacking my drums. I feel soory for them and my kat. She has to hear me drum away. Which reminds me. I need to feed her. Ugh...You know what bothers me? When my Mum and Daddio get mad at me for no reason. Its funny though. Sometimes when I dont do anything bad they find away to get me punished. Assholes and their tricky ways. Lol. Yeah Uhmm...My kat is calling me soo I need to feed her. I\'ll be back.   :::::::::::::::::::::::::::::::::::::::Long Period Of Silence::::::::::::::::::::::::::::::::::::::::::::::::   Okay back. My poor kitty didnt have anything. Ugh...Uhmm...I dont know what to say. MmMmMm My sister\'s boyfriend (soon to be a father and husband) brought FOOD! Hehehe. Im gonna go eat now. Bye Ya\'ll. XoxO  '),(11769,3403602,'20,May,2004','2004-05-20 00:00:00','   OMG! I had a blast today. When I thought that this Spring Break would be  BORING  it became the best just for one day. Me, Lucy, Lucy\'s lil bro, and Celi went to the movies. It was soo fun. We just laugh and laugh. We threw spitballs at each other. We went to see \"Takeing Lives\". We were gonna see the \"Alamo\" but eh! We didnt want to after all. Oh man my friend Celi, a great grrrl to be friends with and to talk to cuz she is funny, she was soo dum. There was a part when the actress was in some room. The actress was leaning against the wall. And Celi goes \"Oh man the would be trippy if the doors open.\" I go \"What?! Doors do open\" She goes \"Oh shit I mean the walls.\" Funny ass shit....I kept on buggin her about it. Yeah it was fun. And thanx to Lucy for telling me this. If it wasnt for her I would have hung myself because of this  BORING  Spring Break. Yeah some guy named Louie was supossed to go but he didnt go after all. Soo I wanted to meet him. Oh well...Maybe next time. Ugh...okay that was today sooo Yeah! Bye!!XoxO  '),(11770,3403602,'20,May,2004','2004-05-20 00:00:00','   Hello Hello. OMG! One of my friends mentioned about Carmen Sandiego and its been in my head for a while now. Hahaha. But seriuosly how can she hide for soo long? I would get mad at her on every episode. They kids would find out her evil polts but they couldnt stop her either way. God damn that whore! She made me mad. I tryed my best at getting her. But I failed. Ugh...But my friend caught her. She has her in the closet. She asking her questions. If Carmen Sandiego doesnt give us the answer we want she earns a slap. Hahaha. That\'s what Carmen Sandiego gets for hiding all the time and wasting my time looking for her ass. Anyways. Yeah its Spring break and I hate it. I have nothing to do. I sit in my room listening to Bikini Kill or Bratmobile. I cant find my other cds. Which sux by the way. Yeah. Have I mentioned that Im in a band? Well yeah its called Who\'s 2 Blame? Heres more info on us. We do our original songs. We sometimes cover Bikini Kill and Bratmobile. I play drums. Sometimes I sing. Our vocalist is Allie, 15. Shes great and fun to be around. She cant sing all professonal but she knows how to fucken yell. Our guitarist is Andy, 14. Hes barely learning guitar but hes already good at it. Hes good at playing one string. Thats his specialty. Hes a great kid to be hanging around with. And hes good with creul jokes. Our bassist/keyborad player is Jessi. She is the best grrrl youll ever know. She is the newest member of the group. She already knows how to play keeyborad, but she is barely learning bass. Thats the band for ya. Bye!!XoxO  '),(11771,3403602,'20,May,2004','2004-05-20 00:00:00','   Yeah life is annoying. Im not Emo but Im really hurt evryday. Somwtimes I think know one cares anymore. Ugh...But yeah. I hate life but I wish it would get better. Anyways I was watching this show called Degrassi. Its a show about normal kids living life. Its soo kool to watch. Anyways this one episode I was watching, it was about this grrrl against the skool\'s food. Saying that it was serving bad food. She starts to have a protest against it like 10 yards from the skool. And then later on in the episode the Princial comes telling her to stop the protest. She told him that she has the right to speak her mind. And then he gave her a weeks detention. Later on she does it again, but this time for another reason. She did it again because she was saying that she has the right to speak her mind. And then the Princial comes warning her to stop. He gave her a choice: either apoligize to the whole skool for the \"mess\" she made or get kick out of the skool. So she went up to the skool and didnt say soory but she told them that she has the right to speak her mind. It was the freedom of speech and if she would get kick out of the skool so then be it. Then in the end she didnt get kick out. But she was warned not to do it again. But I thought that episode was great. Hopefully you can all see it. Its on this channel called The-N. Its kool. Well thats today. Bye! XoxO  '),(11772,3403602,'20,May,2004','2004-05-20 00:00:00','   Ugh Im not really into style or being placed into some catergorie...Uhmm I like to draw. I like to act stupid. I like dorks. I luv doughnuts. I like music. I like cats. I have a cat of my own. Her name is Starfish. Uhmm...I hate sexist bastard (guys)...Hahaha...Ugh I like video games. And I like staying up late. I luv chatting. I like hearing music. I luv making music. I luv writing songs. I luv writing storys. I like going to things that are against Bush. Im into feminism/activism. I luv being involed in certain things that have to do with feminism. I hate when people say you need to like this kinda music cuz you dress or acted like this. And I like pizza. And I like talking non stop. I luv the rain. I luv Monday. I hate Friday. And I hate having a bad reputation. I like making fanzines. Well thats me. Bye! XoxO  '),(11773,3403602,'20,May,2004','2004-05-20 00:00:00','   Hello Hello. So nothin gonna happen this week. 2 more weeks left of skool. Fuck! Im going to miss all my friends. Im goign on vaction for a month. I dont really wanna go. I want to stay cuz some of my friends that are out of town or out of state are coming to visit. I wanna see them. Blah! Oh next saturday Im going to the beach. Oh yeah! Me and friends, bandmates. Hopefully I will have fun. Well yeah no more stuff to say. Bye!!XoxO  '),(11774,3403602,'20,May,2004','2004-05-20 00:00:00','   Hello, Hello. Uhmm...Im in skool right now. And Im bored. Its 6th period and Im done with my work. I got things to do after skool. So thats why Im writing here. Anyways. Uhmm...If people wanted to chat with me Im soory but I wont be on today. Ugh. Im busy and yeah. Saturday Ill be on for a while. Suday Ill be at the mall. After that Ill be on more often. Dom!! I cant wait till we go to the mall. Hopefully Tanya goes cuz that would kick masvive ass or fat people asses. Lol. \"SHH!LISTEN!!.........Its Bjork :D\" LMAO. Ugh. Fuck! Bye!!XoxO  '),(11775,3403602,'20,May,2004','2004-05-20 00:00:00','   You fucking fuck! Stupid Fucker DIE!!!  xox ALESSA & SCOTTY  P.S. Fuck you.  Please Kiss My Fucking Homosexual Ass.  You Ugly Son Of A Bitch.  I Am Going To Castrate You,  With A Rusty Chainsaw. BZZRRT!  DIE MMKAY?  (Delicious Arent They? :D)  '),(11776,3403602,'20,May,2004','2004-05-20 00:00:00','   Hello, Hello. Ive notice that I spelled my username wrong. Which sux. I wanted it to be Bloody Ice Cream. Not Bloodly Ice Cream. Stupid \"L\" makes me sound like a retard. Lol. Anyways. Nothin interesting happing to me. Ugh. Sometimes I think no one really reads this. Soo that means Im talking to myself. Which sux. Fuck! Oh well. I ahte being single. Omg! Like on June 5 Im goign to the beach with my friends. Its gonna be the best time of my life. Hopefully. And and...Imma video tape it and and take photos. Well better photos cuz why the hell am I gonna video tape the damn ocean? Its water. But its pretty. Nvrmind, Ill take it. Im in my room and hearing music if you wanted to know. Thats what I do. Sit here and wonder. Oh and I hope I choke on something. I dunno what Im trying to say. Ugh. Did anyone hear about the American that got his head taken off? Ugh...I saw the video and I wnated to gag. I swear. I didnt want to see it but some idiot told em too. Him and his friends are stupid. They kept playing it over and over again and laughing. Thats sooo lame. Whats their problem? I hate that. Like seriously. But whatever. Hopefully by not respecting the dead they will die a slow painful death. Lol. Shove it! Bye!!XoxO  '),(11777,3403602,'20,May,2004','2004-05-20 00:00:00','   Well Saturday was the best time of my life. Lol. We got drunk. And we met the Go Betty Go Memebers. They were nice and kool. Michelle, the bass, was extremely nice to me and Andy(Memo). We talked about her favorite band. The video should be great. I hope that I cant see myself in it. Lol. Image its like getting everyone cept me? Fuck thats a bummer. Anyways I have nothin planed for this weekend. Yesterday was fun. I went to the mall with Dom and Gwen. It was fun. But we went ot Hot Topic and we saw Tanya. She great. Shes funny also. She let us hear all of her cds. In car of course. She has mutha fucken alot of cds. I wish I had much as her. But meh. Shes great. But when I got home I feel asleep. Ive became super lazy now. And and it sux. Well Im done now. Go screw yer self. Bye!!XoxO  '),(11778,3403602,'20,May,2004','2004-05-20 00:00:00','   Omg! Hello Hello. Today Im going to be in a video shoot. I signed up to try to win it I guess and I won. I never one anything. I feel soo lucky. The band is Go Betty Go. They are the best. If you wanna hear them go to www.gobettygo.com . The video is called \"C\'mon\" Its gonna be great. Im taking my two best friends, Andy and Allie. Its good be kick ass I tells ya! Anyways...I wish that Jessi could have gone but her mom said its to late to be going at that time at night. It starts at 8:00 pm. Somewhere in L.A. Yeah I cant wait. For awhile Ive been real sad but this is the only thing that got me really excited. I just cant wait. Well Im done. Bye!!XoxO  '),(11779,3403602,'20,May,2004','2004-05-20 00:00:00','   Hello Little World! God...I nvr felt like this before. I thought being single would be the greatest thing ever. I can do whatevr I want. Kiss who evr I want and blah blah blah...But now I just feel like I wanna stay with one person only...But I cant cuz I know Ill be rejected...God I cant believe I said that. But yeah. I like to many people when I know I may nvr get with them or even talk to them. It hurts...Badly...And I feel like shit. And when I find out someone likes me I get all excited. They think im a kool guy but when they met me they change their mind. Its been like that with evryone. So I might as well die...or sleep...either one, I dont care. Lucy...If you are reading this...I think that is why I felt like shit in 5ht....And I might continue feeling like this...But it wont stop me from going to yer house and wetting you and Louie babe. Lol...Well...Im done crying, trying and lying. Bye!! XoxO  '),(11780,3354778,'20,May,2004','2004-05-20 00:00:00','    situs islam terkini ; tragedipalestina.com, bangsamusnah.com, keajaibanalQuran.com   '),(11781,3354778,'19,May,2004','2004-05-19 00:00:00','    teah banyak kritikan, \"kok pake b. ingris seh?\" yah itu hanya untuk ebih universal. mahi mohon maaf bagi kawan kawan yang  DISAPOINTED   '),(11782,3354778,'19,May,2004','2004-05-19 00:00:00','    maaf bila site terlalu lama terbuka karena kapasitas antara komp centre dgn provider sangat signifikan berbeda,thank\'s........   '),(11783,3354778,'15,May,2004','2004-05-15 00:00:00','       '),(11784,3354778,'15,May,2004','2004-05-15 00:00:00','       '),(11785,3354778,'15,May,2004','2004-05-15 00:00:00','    DO FISH FLY?  Flying fish don\'t fly with wings like birds, but they slide on fins that resemble a bird\'s wings. They travel about 56 kilometres (35 miles) in an hour. These small fish can move faster in water by spreading their fins and lifting their tails out of the water. In this way, they slide swiftly in water.   '),(11786,3354778,'15,May,2004','2004-05-15 00:00:00','    DO FISH FLY?  Flying fish don\'t fly with wings like birds, but they slide on fins that resemble a bird\'s wings. They travel about 56 kilometres (35 miles) in an hour. These small fish can move faster in water by spreading their fins and lifting their tails out of the water. In this way, they slide swiftly in water.  THE HERON  The heron displays great skill in catching fish. It stands up spreading its wings in an umbrella shape above its head. This produces a shadow and prevents reflections from the surface of the water. Now, the heron can clearly see its prey under the surface of the water. The wings of the bird form a circular shadow on the water\'s surface, and it always fishes inside this circle.   '),(11787,3354778,'15,May,2004','2004-05-15 00:00:00','    MIRACLES OF THE QUR\'AN  The Qur\'an was sent to mankind 1,400 years ago, but there are some facts related in the Qur\'an whose meanings we have only been able to uncover recently.   From the planets to the stars, human beings to animals, Allah created everything in the universe. Our Lord already knows everything we have not discovered so far and He informs us about some of that in the Qur\'an. We can only learn these things when Allah permits us and then we realise that they are miracles of Allah.    How the Universe Came into Existence   He is the Originator of the heavens and the earthâ€¦ (Surat al-An\'am: 101)  The universe came into existence from nothing 15 billion years ago. In other words, the universe suddenly came into existence from nothing.   Only the science of the twentieth century enabled us to attain scientific evidence of this great event. Therefore, it was impossible to know this scientifically 1,400 years ago. But, as is also mentioned in the previous verse, Allah told us this fact when the Qur\'an was first revealed. This is a miracle of the Qur\'an and one of the pieces of evidence that it is the Speech of Allah.    The Expansion of the Universe  In the Qur\'an, which was revealed 14 centuries ago at a time when the science of astronomy was still primitive, the expansion of the universe was described like this:  And it is We who have constructed the heaven with might, and verily, it is We who are steadily expanding it. (Surat adh-Dhariyat, 47)  The word \"heaven\", as stated in this verse, is used in various places in the Qur\'an with the meaning of space and universe. Here again, the word is used with this meaning. In other words, in the Qur\'an it is revealed that the universe \"expands\". And this is the very conclusion that science has reached today.  Until the dawn of the 20th century, the only view prevailing in the world of science was that \"the universe has a constant nature and it has existed since infinite time.\" The research, observations, and calculations carried out by means of modern technology, however, have revealed that the universe in fact had a beginning, and that it constantly \"expands.\"  Orbits  Many of you know that our world and the other planets have orbits. In fact, not only do the planets in our Solar System have orbits, but also all heavenly bodies in the universe have their separate orbits. That is, they all move in very precisely computed paths. This scientific truth that scientists have only recently uncovered was revealed 1,400 years ago:   It is He Who created the night and the day, and the sun and the moon. They swim along, each in an orbit. (Surat al-Anbiya: 33)  As is also seen in this verse, Allah informs us about a scientific fact which was only discovered recently. At the time the Qur\'an was revealed, people did not know that heavenly bodies moved in constant orbits. But Allah knows everything and tells what He wills to His slaves.   The Roundness of the Earth  The understanding of astronomy of the time, when the Qur\'an was revealed, perceived the world differently. It was then thought by some that the earth was flat, and others had different theories. But the fact that the earth was round was not then generally known. However, from the Qur\'an it was understood by implication that the shape of the world is round. The relevant verse reads:   He has created the Heavens and the Earth for Truth. He wraps the night up in the day, and wraps the day up in the nightâ€¦ (Surat az-Zumar: 5)   The Arabic word \"takwir\" is translated as \"to wrap\" in the above verse. In English it means \"to make one thing lap over another, folded up as a garment that is laid away.\" The day and the night wrapping over each other can be true only if the earth is round. But, as mentioned above, the Arabs who lived 1,400 years ago thought that the earth was flat. This means that the roundness of the world was hinted at in the Qur\'an, which was revealed in the seventh century. That is because Allah teaches the truth to mankind. This matter, indicated in the Book revealed by Allah, was verified only centuries later by scientists.   Since the Qur\'an is Allah\'s word, the most correct words were used in it when describing the universe. It is unlikely that a human could know and choose these words. But because Allah knows everything, He can make these facts available to man at any time He wills.   '),(11788,3354778,'15,May,2004','2004-05-15 00:00:00','    blog ini adalah up date dari mahlil.blogspot.com, berhubung blog yang lalu masih longer little web, kali ini sudah grow up. trim\'s............   '),(11789,3989154,'30,July,2004','2004-07-30 00:00:00','    urlLink     Well, I have agreed a fee with the buyer and already solicitors are busying themselves passing bits of paper backwards and forwards.  Get this, the buyer, who is a local property speculator, doesn\'t even want to see the house let alone do searches or surveyors reports.  \"At that price,\" he said. \"I don\'t need to see anything\".  Which gives me the same feeling as when I win at the bookies. Momentarily happy nefore the thought of how much better off I would have been if I\'d opted for a higher stake.  In other words - if he was prepared to pay that price without even looking - the how much higher could he have gone?  Ah well. Bollocks to it. It\'s sold. That\'s the main thing. He said he can even give me a cheque inside a week. Fantastic.  Posts might be a little sporadic next week. I\'m off on a VSO training course. On the other hand, they do have an internet and I\'ll probably be bored senseless so I may post more than ever.  We\'ll see.  But tonight, I celebrate.   '),(11790,3989154,'29,July,2004','2004-07-29 00:00:00','    urlLink     The downside of PR is that today I appeared in the newspaper. It was full or errors. The picture looked terrible and I sounded like a right yuppy prat.    The upside?    Well by 10am this morning this note was put through the door.     Hi *** from number *** here. I would like to make a cash offer to buy your house and could complete in a couple of days. I would exchange contracts immediately.      If you are interested please ring....   YESSSSS.   '),(11791,3989154,'28,July,2004','2004-07-28 00:00:00','    urlLink      Before I fly out to Hanoi, the biggest problem I still have to face is selling my house.      But today I had a brainwave.    Many years ago, in my days before switching to the dark side that is PR, I was a hack. One of my least favourite chores, week in, week out, was writing the bloody property section.    So I thought - why not drop them an email? I mean, \"Geordie goes to Vietnam - needs to sell house quick,\" isn\'t such a bad story. Well, okay, by property press standards it\'s not such a bad story.    So I emailed them. Would they be interested in the story? Would they like me to write a press release? Can I arrange some pictures for them?    And guess what, they said yes, yes and we\'ll send our own photographer.      Bearing in mind that writing press releases is what I do for a living, it\'s nice to be able to write one about myself for a change. Only problem was - when I got to the sign-off stage I felt I needed extra reassurance. Typical. Finally a job when I don\'t have to clear everything with a client and I still need someone to say it\'s okay.    Anyway, I eventually told myself it was fine and forwarded it. Fingers crossed. If this gets a big splash in the property section then it might go along way towards selling the house.    I reckon if I could just shift the house, then all the other million pieces of organisation would fit nicely into place.   '),(11792,3989154,'26,July,2004','2004-07-26 00:00:00','    urlLink      Hanoi - not there yet.    I thought I had better make one thing clear. Despite what the blog says I am not yet in Vietnam.   I know I could have changed the \"About Me\" bit to say \"Newcastle, England\" - but it seemed at odds with the title of the blog and its strapline, so I put down \"Hanoi, Vietnam\" albeit, with my departure still over a month away.   The great clearout started today. The estate agent was coming round to value the place. I spent two solid hours of tidying and chucking out crap and she still turned her nose up and suggested I get in some professional cleaners. Cheeky cow. Then she proceded to value the place at a good 30k less than I thought.   It might sound quite a religious thing to say, but I\'m really looking forward to be posessionless. Give or take, of course, a pack on my back.   A fellow  urlLink VSO er told me what a buzz it had been, getting on his bike outside his newly sold property with nothing but a small bagful of clothes and his passport. I could immediately see his point. What freedom.   I never seem to get sentimental about objects. The house has been a great place to live but it was still just a house. I\'ll be more sorry to see my lovely yellow Punto go, but to be honest it\'s reaching an age with the clanks are louder than the revs.   Of the rest of my stuff I\'ll keep the clothes I\'ll be taking with me. I\'ll keep a couple of suits at me mam\'s too. My CD\'s can go in a box in the attic. No one is getting their hands on them.   But that leaves me with beds, coffee tables, TVs, stereos, lamps, pots, pans, plates and cups to get rid of. The vague plan is to send out an email to people I know and ask if anyone wants any of it.   They can have the smaller items. The larger stuff they can have on a long term loan. In other words, if and when I get back, and if I\'m stopping and want to buy another house, then I might just want it back. Chances are I won\'t though.   I\'ve also been given the contact details of a charity that helps asylum seekers. That\'ll be worth it just to annoy any Daily Mail readers I may have accidently befriended.   But imagine it. Imagine the day when it\'s all gone. How fantastic is that going to feel?   I can\'t wait till I\'ve finished all my training. Passed my medical. Filled out the million forms. Got my flight details etc etc.   Then I can really enjoy a last beer in the Toon.   '),(11793,3989154,'24,July,2004','2004-07-24 00:00:00','    urlLink   &nbsp; Google-searching for Hanoi pictures on the \'net I was reminded of the traffic.   It\'s like nothing you\'ve seen before.&nbsp; On most streets cars are relatively rare.&nbsp; There are a few bikes, and a sprinkling of cyclos, but the vast majority of the congestion is caused by scooters.   During rush hour it\'s quite some sight.&nbsp; Especially just as it\'s getting dark when you get a mix of commuters traveling home and the bright young things posing for all they\'re worth.&nbsp; The lights on the bikes just fly by you.&nbsp; If you\'re sitting in a pavement cafe then it\'s a fascinating spectacle.&nbsp; But if you have to cross the road then it\'s not quite so enjoyable.   Your instincts tell you to wait for a break in the traffic then run as fast as you can.&nbsp; Except there is no break, and if you run then you\'re only likely to make yourself a harder target to avoid.   Instead you have to walk as slowly as you can.&nbsp; It\'s best not to look at the traffic at all.&nbsp; Just keep walking.&nbsp; The aim is for them to see you and to take evasive action.&nbsp; The sea just parts around you and you become the centre of the only space on the road.   Occasionally though, when you do sneak a look at a traffic, the sight of a teenage girl, driving while speaking into her mobile, scares the hell out of you.   Has she seen you?&nbsp; Will she have time to avoid you?   Scooters&nbsp;appear to be&nbsp;the principle form of transport.&nbsp; A whole family on a scooter is a fairly regular site.&nbsp; Dad drives, with a child sat in front of him.&nbsp; Mum sits behind, often with another child strapped to her back.&nbsp; Livestock lashed to a bike is also a regular sight.&nbsp; Occasionally items of furniture too.   As a tourist, quite often getting a ride on a scooter, or moto, was the only way to get around and they scared the hell out of me.&nbsp; All sixteen stone of me on the back of a moto, plus the driver with my pack between his knees.&nbsp;&nbsp;   At first I didn\'t know where to hang on.&nbsp; I\'d cling onto the driver leaving embarrassing marks with my sweaty paws.&nbsp; Later I discovered the bar under the seat.   Local passengers don\'t seem to hang on at all.&nbsp; Girls will even ride side-saddle.   I think I will make do with walking.&nbsp;   Apparently I will be living on Van Mieu Street.&nbsp; By all accounts I will be&nbsp;dividing my time between the  urlLink training centre  and the  urlLink restaurant , both apparently also on, or adjacent to Van Mieu.&nbsp;   Searching for pictures of the streets, it seems like a rather nice area.&nbsp; The  urlLink Temple of Literature  is there and is surrounded by much leafyness.    urlLink      &nbsp;Temple of Literature    I\'ve been wandering a lot about what my house will be like.&nbsp;I\'ll be sharing with two other  urlLink VSO ers and&nbsp;apparently it actually comes with a TV and a kitchen.&nbsp; To be honest I can\'t really see myself using either much.&nbsp; I can eat from street hawkers as cheaply as anything I can make myself and I can\'t think the telly will hold much appeal.&nbsp; Then again, if I can get ESPN, I might even be able to watch a little bit of footy.   But will the house be clean?&nbsp; Will&nbsp;it be secure against&nbsp;rats and creepy crawlies?&nbsp; Will it be above one of those godawful karaoke bars, with it\'s sickly Asian pop?&nbsp; To give you some idea about local pop music, when I was last there all I heard was the awful \"Kiss&nbsp;Kiss Boom Boom\" everywhere I went.&nbsp;   Who knows what I will do with my weekends?&nbsp; Maybe I\'ll be so tired that I\'ll just want to sleep and potter around.&nbsp; Maybe I\'ll take the chance to go off and explore the locality.&nbsp;   One of the first jobs will be to find the nearest&nbsp; urlLink pho  joint.&nbsp; The  urlLink PieMan  has already recommended&nbsp;&nbsp;13 Lo Duc street, it remains to be seen how local that is.&nbsp; He has also promised to put a Hanoi food guide on his blog too.&nbsp; If you haven\'t checked him out already make sure you do.    urlLink      &nbsp;Pho - pictured pinched from   urlLink  NoodlePie     Anyway, all in all, it\'s suddenly starting to seem very real.&nbsp; And very near.&nbsp;   Excitement is building.&nbsp;   '),(11794,3989154,'24,July,2004','2004-07-24 00:00:00','    urlLink     I first arrived in Vietnam on May 22nd, 2002. I flew into Hanoi airport, which was my first stop on an eight month tour that would take me through South East Asia and subsequently Central America.   I have very little recollection of Hanoi airport but the trip into the city will stay with me forever.   Sitting in the back of a taxi, not having slept for 24 hours, even my tired eyes were forced open by the surrounding views. As you move along the dusty roads, the hundreds of scooters part in front of you and on both sides of the track all you can see is conical-hatted women working in the fields.   It was probably about five seconds into the journey that a grin broke across my face. It mostly stayed there for the next six weeks.   From then on though my Hanoi experiences were a little mixed. For starters the taxi driver ripped me off. I paid $10 more than the guidebook suggested. What\'s more I knew I was being ripped off but I was too tired to argue.   Little matter, Hanoi was beautiful. No one could fail to be impressed by the architecture and the feel of the city, but for a pale skinned Geordie, short on sleep and naive of traveler ways, Hanoi proved a little difficult.   Compounding his over charging, the taxi driver dropped me off at the wrong hotel and I didn\'t even notice. Obviously he was getting his cut. I opted to have a walk about in order to delay sleep till local time. Two hours of profuse sweating and being pestered by street kids selling postcards later I decided to return.   It was only then that I realised I wasn\'t at the hotel I thought I was. But if I wasn\'t there, then where the hell was I?   I spent the next three hours in Hanoi trying to find the hotel, all the time being pestered to buy something. My panic was mounting. How the hell do you find a hotel that you don\'t even know the name of? Two cyclo rides didn\'t even help. They couldn\'t point me in the right direction. My new sandals were rubbing, the humidity was unbearable and I was badly dehydrated.   Eventually I told a postcard seller that yes I would by some cards. But first he would have to find my hotel. I described it to him and he set off walking quickly, and cursing my blisters I followed. He walked and walked and I started to worry if this was another red herring route. Perhaps this was part of an elaborate kidnapping plan.   Almost 45 minutes later I got to the hotel. Because of my total relief I must have bought nearly all the kid\'s cards. Ten minutes later, I flopped onto my bed, turned out the light and....heard scurrying noises.   A second later the light was back on and I was in the middle of a staring match with a cockroach. Aaaaarrrggggh.   Call it symptomatic of a sheltered upbringing but I had never actually seen a cockroach in the flesh. They\'re ugly. Their antennas move independently of each other and they\'re fearless too.   Suddenly I was wide awake. I was convinced that the second I fell asleep it would walk across my face, probably laying eggs in any orifice it could. Every time I switched off the light the scurrying began again and I would be once again incapable of sleep.   I think I eventually dropped off about three in the morning, only to be woken again at five with the sounds of Asian pop music, scooters revving and general crashing of pans. Vietnamese people, I was soon to learn, work long hours and if they\'re awake, they don\'t really see why you should sleep.   So I got up and that day at least was a little more productive. I decided I needed some beach time so I booked myself on the train to Nha Trang later that week. Or rather I should say, the nice man at reception did, he took my money, and bought me a ticket on the Reunification Express fast train.   I also booked myself a two day tour of Halong Bay, the nearby beauty spot.   I set off for the trip the following day, leaving my pack at the hotel. It was magical. For the first time I started to really unwind in Vietnam. The food was amazing, the company was good and I started to warm to the Vietnamese. Vietnamese people manage to be cheeky and shy all at once. They seem friendly, but they\'re entrepreneurial too - always chasing the dollar. I also learnt that when saying: \"No I don\'t want to buy that\", if I did it with a smile, and a joke, they didn\'t get angry with me.    urlLink     &nbsp; Halong Bay    So after a couple of days swimming in the South China sea, eating seafood and mastering chopsticks, I returned to the Roach Hotel in Hanoi. I slept another night and caught the train the next morning.   The hotel owner even offered me a free taxi right to the station. Despite my protestations the very small porter also carried my pack, right onto the train and helped me store it under my bunk. I tipped him of course.   The train ride was amazing. I appeared to be the only westerner on board. There can be few stranger sites that six people sat across a bottom bunk shoveling rice into their mouths from little plastic bowls. There were five Vietnamese and me, in a Newcastle shirt, shaven headed and twice their size.   A whole day later I was ready to get off the train. My pack was packed, I had waved good bye to my cabin mates and made faces at the little kids one last time. The train slowly stopped and I was in Danang, not Nha Trang.   Hanoi had ripped me off one last time. I had paid full fare for the ticket but I was on the slow train, not the fast one. My previous hotel had pocketed the $20 difference. No wonder they took me right to my carriage, they didn\'t want me asking any questions of the guards on the way. Riding on the slow train meant I was still another 12 hours away from my destination. I did the only thing I could. I laughed and returned sheepishly to my cabin. I was starting to get the hang of Vietnam.   I don\'t think I was ever ripped off again. The street hawkers also became no bother at all. A smile and a shake of the head was enough, even better was a joke. In fact I came to rely on them. Every finished paperback, every empty packet of fags was normally followed by two or three kids rushing to sell me a replacement.   I still hate cockroaches but at least I can deal with them now. I used to catch them in my soap dish and chuck them out the window.   I didn\'t do Hanoi justice the first time around. Between sleep deprivation and a lack of experience it sent me scurrying off to the beach pretty quick. But I know it\'s beautiful and after spending a month and a half in Vietnam, at least I now know how it works.   I\'m looking forward to giving Hanoi a second chance. I met many more seasoned travelers who loved it and I know I will come to love it too. But I\'ll bargain harder this time, I\'ll be a little less gullible too, but I\'ll also keep smiling.   Hanoi, I\'m coming back.   '),(11795,3989154,'23,July,2004','2004-07-23 00:00:00','    urlLink      So why volunteer work, why  urlLink VSO  and why now?   If I\'m honest life hasn\'t really been the same since I turned 30.&nbsp; It was only at 30 that I realised that life wasn\'t predestined.&nbsp; I think it suddenly occurred to me that the path I took for granted either wouldn\'t, couldn\'t or didn\'t have to, be as I expected.   I guess I had just assumed somewhere along the line that I\'d marry, have kids and settle down.&nbsp; It was only later on that I questioned whether that was even what I wanted.&nbsp; And if I didn\'t want that, was there any room for me here?   There&nbsp;comes a time when you\'re the only single person at the wedding.&nbsp; And maybe, as it has been pointed out to me, maybe I just haven\'t met the  right person .&nbsp; And of course that\'s correct.&nbsp; Of course if I had met the  right person,  then I probably wouldn\'t be doing this.&nbsp; But I always liken it to winning the lottery.&nbsp; Yes it would be nice if it happened, but I\'m not going to plan for it or build my life around it.&nbsp;   Putting aside the Camelot comparisons for a second, I\'m not even sure that I\'m ready for it.&nbsp; Even at 33 I\'m still too independent to consider settling down.&nbsp; I still need too much me time.&nbsp;&nbsp; I\'m still far too selfish.   And as for kids, well I\'m starting to like them.&nbsp; But only other people\'s.&nbsp; Two hours maximum is pretty good, anything longer than that and they lose all their appeal.&nbsp; I\'m becoming an uncle for the first time in a month or so - that\'ll do nicely for the time being.   My biggest fear of being a dad is simply that I wouldn\'t ever be able to do this.&nbsp; Namely drop everything, sell up and move on.&nbsp; It\'s hardly been at the drop of a hat but it beats waiting till they\'ve grown up and left home.   But why VSO?&nbsp; Well somewhere between turning 30 and now, I took time out to think up a new plan.&nbsp; A plan that deviated from the norm.&nbsp; During this time I went traveling.&nbsp; The change in me was far more wide reaching than I expected.&nbsp; I spent a lot of time thinking.&nbsp; A lot of time reading and a lot of time simply staring at sunsets.   I learnt that there\'s more to life than the UK.&nbsp; Actually forget the UK.&nbsp; I learnt there was more to life than Newcastle.&nbsp; If you know me, then you\'ll know what a leap that it.&nbsp; I love this city.&nbsp; It\'s the love of my life.&nbsp; But it is time for a trial separation.   Staying in Newcastle meant continuing with normality.&nbsp; Continuing with office jobs and office politics, and while it\'s been nice to be self employed this past 12 months, at the end of the day I\'m still not cut out for the commercial world.   Post traveling it all became less important to me.&nbsp; Everything seemed so slow, so dull, so trivial.&nbsp;&nbsp; I considered traveling again but I was worried that it was literally a road to nowhere.&nbsp; So I needed something that wasn\'t so slow, dull or trivial.&nbsp; Something where the commerce that bored me would be replaced with a challenge that would keep me fresh and motivated.   Then there was the war.&nbsp; Excuse me using the next phrase in this context, but Iraq has been my Vietnam.&nbsp; I know when I look back in decades to come, people will ask what I did to oppose the war.&nbsp; Well I marched, I wrote letters, I bored the pants on my friends and I became angrier and angrier and angrier.   I have no level of nationalism left in me.&nbsp; Between the war, the BNP and the Daily Mail it has all been stripped away.&nbsp; Britain is simply somewhere I was born.&nbsp; And right now I want to get out, I don\'t want to play a part in what is being done.&nbsp; And as many times as I repeat \"Not in my name\", I know it is in my name.&nbsp; It is being perpetuated by my country and by the Government I voted for.&nbsp;   So that is it.&nbsp; I fought.&nbsp; The fight will go on but&nbsp;without me.&nbsp; I am swapping a life of being angry for the choice of not being involved.&nbsp; I want no part of Great Britain for the time being.   VSO from the outset looked like the perfect organisation for me.&nbsp; They pay.&nbsp; Albeit something like $5 a day, but at least enough to get by.&nbsp; They provide training and they provide help and assistance that is genuinely sustainable.&nbsp; Meeting all the VSOers, I know I have made the right choice.&nbsp; I know because, as different as we all are, we have all reached the same crossroads in our lives.   VSO meant two years too.&nbsp; Two years is a good length of time.&nbsp; Long enough to be genuinely life changing.&nbsp; Long enough to be a real adventure.&nbsp; Long enough for you to really settle into a culture.   But short enough for the end still to be in sight.   So there you have it.&nbsp; That\'s why I\'m off on September 3rd.&nbsp;   Exciting isn\'t it?   '),(11796,3989154,'23,July,2004','2004-07-23 00:00:00','    urlLink      So here it is.&nbsp; The new blog.   This will document, from this point onwards, my experiences as a  urlLink VSO &nbsp;volunteer with  urlLink KOTO &nbsp;in Hanoi, Vietnam.   As it stands, my leaving date is September 3rd, but there is a great deal to do in the meantime.&nbsp; People keep asking me whether I am nervous about my new life.&nbsp; The honest answer is no.&nbsp; It\'s all the work that has to be done first that scares the hell out of me.   Typically, having muddled along for the past 12 months as a self employed PR person,&nbsp;all of a sudden, when I am looking to wind down - everyone seems to be throwing work at me.   The good news is that the cash I\'ll make will help clear things financially before I go, the bad news is right now I just don\'t have the time.&nbsp; Unfortunately I don\'t have the guts to turn it down either.   To complicate matters even further, I have managed to do something very painful to my ankle.&nbsp; I have played five-a-side without a single injury for years now, suddenly, when I really need all my mobility, I go and knock lumps off myself.   The good news is that despite the worrying bruising and a bit of swelling, it only appears to be a bad sprain, or at worst ligaments.&nbsp; An xray in the next couple of days will hopefully confirm there are no broken bones.   Either way, clearing out the house and helping to fix it up for a potential buyer is beyond me right now.&nbsp;   I\'ll blog a little more in the future about why I wanted to do VSO as well as my love affair with Vietnam.&nbsp; I just wish everything could be sorted so I could really start to enjoy the anticipation.   '),(11797,3989154,'14,August,2004','2004-08-14 00:00:00','    urlLink     The Vietnamese gecko. Picture nicked from  urlLink here .   As excitement builds towards my September 2nd departure  I\'ve been keeping up to date with fellow VSOers\' weblogs.  I must admit I have been a little evangelical about the power of the blog while I\'ve been away on VSO courses and there\'s a growing number of volunteers who are now keeping in touch with back home through an online diary.  The three VSO bloggers, that I know of, have all pretty much just landed in the countries where they will be working. For the uninitiated you don\'t just get straight down to work. There\'s usually a period of in-country training which includes learning a little language, some local culture and a bit of history too.   urlLink Nurse Jen  has just landed in Mongolia. As much as I\'m delighted to be going to somewhere as beautiful as Vietnam, living in such an unexplored, amazing wilderness would really appeal to me. I am a little jealous. Although the -40 winter temperatures would be a little too much for me to bear. I\'d have to wear long sleeves.  Meanwhile  urlLink Sarah  is just settling into her training in Hanoi.  Eventually she will be taking up a post teaching in a hill town in the North.   Finally  urlLink Mel  - is getting to grips with China and she\'s taken to blogging like a duck to water.  It\'s a cracking read already.  So go over, and say Hi. You might want to introduce yourselves too - as they\'re probably only expecting comments from families and friends. Hope they don\'t mind me outing them to the weblog world at large.  In addition long-time blogger  urlLink Gia  has been on a trip to Hanoi the last few days. I\'m not sure if I entirely agree with her views on the country, except to say that we both think it\'s just amazing. But she\'s always a cracking writer and her posts are always challenging. There\'s a superb little movie you can view of Hanoi too.  Also, there\'s this:  urlLink I Love the Smell of Coffee in the Morning . Another new Hanoi blog. A man with his own cook. Very stylish. And I love the picture of the gecko. I had forgotten they were everywhere. Geckos eat mozzies. Geckos are your mates.  But they don\'t half freak you out when they run up your leg.   '),(11798,3989154,'11,August,2004','2004-08-11 00:00:00','    urlLink      There is something inherently cool about speaking a second language.  And I say this despite being a master of only one.  To be honest I can\'t even get my head around being fluent in any language but my own. How exactly does it work? At what point do you stop translating everything in your head? Surely even the best linguists have to translate first into their own language, then forumalate a response and then change that back into the language in question.  Don\'t they?  Or at what point does it become instinctive? When do you start to think in a foreign language? And how does that work? Do you think in a mix of languages? Do you keep certain languages for different thoughts? Will I be less intelligent, or expressive in a different language?  So far my language achievements boil down to this: I dropped French at 14. But even then I took it for four years and I can manage little more than asking directions or telling you that the pen of my aunt is in the garden.  I did a week\'s worth of Spanish in Guatemala. The idea was that afterwards I could manage rudimentary conversations and chat with the locals a bit. The reality was that I learnt how to ask for an ashtray, buy cigarettes and order chicken and rice. It helped but it wasn\'t really what I had in mind.  Every single continental European I met spoke absolutely fluent English. Not stuttering pigeon English - they knew it all. They could even be funny in another language. But it\'s stating the obvious to say that the English are terrible at languages. I used to sit with my Spanish tutor with my head in my hands. He would sit there prompting me with \"you know this, you know this\" and I\'d be thinking that if I really did know this then why would I be sitting here looking so stupud.  I also lapsed into classroom ways very early. He would tell me there was a free session on in the evenings. I would ask if I had to go and he would respond that it wasn\'t compulsory. So I wouldn\'t go. I mean, I was paying for my day time lessons and I wanted to learn the language so why wouldn\'t I go? The thought of learning anything that I wasn\'t made to learn was still beyond me.  As part of my posting I shall be required to learn Vietnamese. The language - in full. Typically having failed to learn either Spanish or French I am now trying my hand at one of the hardest languages in the world. It\'s a tonal language you see - each word can have six different meanings depending on how you say it. Sounds like quite a challenge.  But just imagine if I do master it. Imagine being able to bark orders to taxi drivers when I meetmy parents at the airport. Imagine being able to make jokes with the waiters and cyclo drivers while they look on admiringly.  Imagine being able to put \"Fluent Vietnamese\" under skills on my CV.  Like I said, speaking another lanugage is very very cool. I just hope I\'m up to it.  I think this will be my biggest challenge.   '),(11799,3989154,'09,August,2004','2004-08-09 00:00:00','    urlLink     Harborne Hall, Birmingham.  6/8/04   '),(11800,3989154,'07,August,2004','2004-08-07 00:00:00','    urlLink     Well I\'m back from training and my departure for Vietnam is just over three weeks away.  There is still so much to do.  I won\'t finish client work for another couple of weeks.  I have the house to clear, a million injections to have, and a thousand other forms still have to be filled in for VSO.  For someone as disorganised as me, it\'s a  nightmare.    Still it\'ll all be worth it when I touch down in Hanoi and I make that journey from the airport to the city again.  That\'s the bit I am really looking forward to.  Just knowing that the adventure is beginning.  Travelling back from Birtmingham yesterday I got a brilliant bit of news.  I had been encouraging Mum and Dad to visit me on my travels but so far the sticking point has been my Dad\'s knees.  A combination of his knackered knees and his height (six foot four) mean airline seats are hell and he can\'t sit in one place for too long.   A long-haul flight looked out of the question.  But yesterday they rang me to ask how the course had gone and I could hear my Dad in the background telling my Mum to pass on some news.  He sounded very excited.  It appears that they are now planning to visit me a in October 2005.  Dad has tracked down a Vietnam tour and he\'s going to book it for them both.  Not sure how they\'ve got past the knee problem.  But I\'m so chuffed they\'re coming over.  Not just because it\'ll be great to see them both but also I know they\'ll get so much out the trip.  They\'ve never holidayed outside of Europe before and I want to see the look on their faces when they get their first glimpse of Nam.  Anyway, I am off to Sunday lunch with them tomorrow so I can find out more about their plans.  I hope it all comes off.  I\'m looking forward to showing them around.  My Mum and Dad sharing a cyclo should be some sight.   '),(11801,3989154,'05,August,2004','2004-08-05 00:00:00','    So, apparently I am a teddy bear.  That was the verdict of an exercise we have just completed which measures how you react during negotiation.  My general teddy bearness is apparently a reflection of my willingness to accommodate.  Couple that with a slight level of turtleness (anything to avoid conflict) and it\'s quite clear that I\'m not the toughest of cookies.  But everyone likes a teddy bear, right?  The course is progressing well.  Twelve hours a day of lessons are tough for us all.  You really do struggle at times and it is unavoidable that you will have sessions when you\'ve realised that you\'ve drifted off a little.  However, there is a real togetherness in the group now and a feeling that, perhaps for the first time, we are all keen to get out of the classroom and into our overseas placements.  The biggest catalyst for this was a session yesterday afternoon when we split into smaller groups and devised a presentation.  The object was to find a way of making a decision based on our own scenarios.  We were invited to use the tools we had learnt.  This would mirror how we would all teach in our posts and pass on our own skills to locals.  I adapted the tree diagram into something approximating a bowl of chicken noodle soup.  My audience were invited to write down on post-it notes their own ideas for fundraising and they\'d stick these on the bowl as ingredients.  From there we also looked at who we would approach and what methods we would use to make contact.  I think, it went well.  But then again, everyone here is so kind and gentle with you that it is difficult to be sure.  Maybe they were just being nice.  Whichever it was, I came out of the session with lots of good ideas as to how find cash for  urlLink KOTO .  I had also had a real feeling of I CAN DO THIS.  That was, to use a  urlLink VSO  word for a second, very empowering.  Tonight we have the luxury of an early finish.  Six pm instead of eight.  That means there\'s time to use up before we slope off to the bar.  There\'s talk of a trip to a nearby bowling alley, or perhaps an Indian meal.  Maybe I should use the time to visit the resource centre and find out more about my placement.  Anyway, tomorrow is the last day and so far it hasn\'t been to bad.  I never cease to be amazed at just how lovely the  urlLink VSO  people are.  You can feel the sense of readyness and excitement within the group rising day by day.    It\'s almost time to unleash the new VSOers into the world.    I hope the world is ready.   '),(11802,3989154,'03,August,2004','2004-08-03 00:00:00','    Well, I\'m currently in Birmingham.  Harborne Hall in the suburbs to be precise.  I\'m here on a five day training course doing something called SKWID.  No, it has nothing to do with seafood of any variety.  It stands for (I think) Skills for Working in Development.  It\'s going to be a tough few days.  I arrived here yesterday, after three and a half hours on the train.  We had our first couple of hours works from seven till nine and I must admit I struggled.  A Monday following a weekend that included a birthday pu sessioand a wedding was always going to be tough.  I think the weekend\'s high living combined with a nasty muggy day yesterday left me very dehydrated and I was suffering from a stinking headache.  All in all it meant that I wasn\'t able to take part in the after lessons bar session which is when people let their hair down and you get to learn a little bit about them.  As always though with  urlLink VSO  everyone seems very nice.  There is the ususal mix of people.  There\'s the white haired old ladies, the young teachers and nurses, and me - somewhere in the middle.  Most people appear to be off to Africa.  Me going to Vietnam and another older guy off to Cambodia are the only exceptions.  Other interesting occupations include a fruit tree cultivator, a bloke who manages wetlands and someone who works in the eco-tourist business.  There are also three Canadians which I didn\'t expect.  VSO it appears is also active in Canada.  There\'s a Dutch man too.  The lessons are tough for someone like me who has the attention span of a gnat.  I find myself cursing anyone who asks a question or wants a point clarifying.  There\'s plenty of role playing and the like, so you have to suspend your natural level of embarassment from time to time.  Apologies for no pictures on this post.  The facilities are not conducive.  It also reminds me that unless I get my own laptop to work, pictures will be out of the question in Vietnam too.  But we shall see.  I\'ve an hour before lessons start so I am going to check out the Library facilities here.  Maybe they\'ll have some interesting reading on Hanoi.   '),(11803,3529951,'02,June,2004','2004-06-02 00:00:00','   Hmm hmm.  Well..seeing as I don\'t have much else to post. I\'ll just post a little survey. *nodsnods*  I\'m going to get my prom dress tomarrow. I guess that\'s exciting.  Other than that..i\'m tired..and I don\'t want to finish my homework..but hey, only 8 days left! Yay!  LAYER ONE Name: Kristina Taber Birthdate: October 3, 1987 Eye Color: Green/blue Hair Color: Auburn Height: 5\'1\" Righty or Lefty: Rigty Zodiac Sign: Libra  LAYER TWO Your Heritage: American The shoes you wore today: Camo Chucks Your weaknesses: I\'m too damned nice?  Your fears: The dark, Thunder Your perfect pizza: Ah, I like any kind of pizza, as long as there aren\'t mushrooms and odd things on it.  LAYER THREE Your most overused phrase on Messenger: w00t  Your first thoughts waking up: UGH, FUCKING RINGING. Or, Do I REALLY have to get up? no..I have 20 more minutes. Your best physical feature: My eyes maybe? Your bedtime: 12-12:30 am-ish. Your most missed memory: ?? Donno  LAYER FOUR Pepsi or Coke: Coke McDonald\'s or Burger King: Wendy\'s >.> Single or Group Dates: Single Adidas or Nike: Nike Lipton Ice Tea or Nestea: Does it matter? Chocolate or Vanilla: Chocolate this week. Cappuccino or Coffee: Cappuccino  LAYER FIVE Smoke: NO Sing: Yes Take a shower everyday: Everyday Have a crush(es): YES Do you think you\'ve been in love: I am  Want to go to college: Sure..why not waste more of my life on schooling >.>  Like high school: Hell no Want to get married: Yes Believe in yourself: Yeah  LAYER SIX Been on stage: Yes Gone skating: Yes..unfortunately :S Made homemade cookies: Yes Gone skinny dipping: Eww..no Dyed your hair: No Drank alcohol: Yucky--I only like smirnoff and wine coolers. Smoked: Nope Done a drug: Never Had sex: Yes Made out: Yes Gone on a date: Yes Gone to the mall: Yes Eaten an entire box of Oreos: No..my parents eat them before I do Stolen anything: Yes but I never will again  LAYER SEVEN Played a game that required removal of clothing: no If so, was it in mixed company: See above answer ^^^ Been dumped: No Been trashed or extremely intoxicated: No Been caught \"doing something\": Eh? Been called a tease: No O.o Gotten beaten up: No Shoplifted: No Stolen anything: Yes Changed who you were to fit in: no  LAYER EIGHT Age you hoped to be married: 25 Numbers and names of children: 4! according the palm reader lady >.>  Describe your dream wedding: dont know yet How do you want to die: Not in pain? Where do you want to go to college: No clue What do you want to be when you grow up: No clue What country would you most like to visit: Japan  LAYER NINE (in a guy/girl. . .) Best eye color: green Best hair color: black or blonde Short or long hair: either  Height: Doesn\'t matter, though I like taller guys Best weight: Doesn\'t matter, just not thinner than I am >.> Best articles of clothing: Dont care Best first date location: I don\'t know O.o Best first kiss location: Underneathe the railroad tracks where no one can see you lol  LAYER TEN Number of drugs taken illegally: None, thank you very much Number of people I could trust with my life: probably 1 Number of CDs I own: Too many to count Number of piercings: 3 on each ear, want another one >.> Number of tattoos: none yet  Number of times my name has appeared in the newspaper: i have no clue..a few, though Number of scars on my body: bah, too many to count lol  '),(11804,3409476,'31,July,2004','2004-07-31 00:00:00','   The boys were rained out for all games today and will have to play at least 3 tomorrow, weather permitting. I guess Lyle didn\'t love us enough to put pictures on the blog so he didn\'t see that Jackson\'s toes were supposed to be crossed. That\'s okay cuz I have Megan\'s crossed for tomorrow. I just explained that as the next youngest it was her duty. And as soon as Erika gets home and fulfills her motherly duties she could uncross them. On a good note the boys did get to go to the Rock N Roll Hall of Fame. So that was definitely a plus to not playing. Our house is still in Dodger mourning. We have lost 4 of our favorite household players, Paul Lo Duca (Whose first major league game Adam and I attended), Juan Encarnacion (whose autograph Adam just got last week), Dave Roberts and Guillermo Mota. Although we acquired Brad Penny, Hee Sop Choi and Chuck Finley, our team will never be the same.'),(11805,3409476,'30,July,2004','2004-07-30 00:00:00','   If at first you dont succeed try try again.  I dont know what happened to this post and I cant remember word for word what I wrote, but here is the gist of it.  Adam called and said they walked 4000 miles in the opening ceremonies, he didnt realize that as an athelete you are in the openeing parade not sitting on your bottom watching  it.  He took some pictures, so again I will need that tutorial shortly.  He is worried about not being able to play because of the thunderstorms you midwesterners are allowing to happen, so Jackson has to cross all his toes on one foot for good weather and the ones on the other foot for good games for his favorite boy cousin.  He saw Aaron Carter and was in teenage boy heaven with making fun of him w9ith his friends.  RThe rain started before Little Richard could begin his performance so he missed out on that (he would have liked that one).  No sign of W.  and let\'s keep it that way!  I just checked Yahoo weather and it is cloudy but 82 in CLeveland, sounds like soup.  hot and humid, eeww!  They havent posted any results.  The last game for today should be over about 6pm Eastern TIme.  On a side note, I think I ruined Adams day yesterday when I told him about the MASSIVE trade the Dodgers made of 3 of his favorite players.  I really dislike Frank  McCourt.  '),(11806,3409476,'29,July,2004','2004-07-29 00:00:00','   I just got a call from Adam, they\'ve arrived in Cleveland safe and sound.&nbsp; They are attending ( feasting  is how he put it) a reception for all atheletes before they go to their dorms.&nbsp; Kathy sent her digital camera with him, so sometime between now and Tuesday when he gets home, Erika or Gretchen will have to teach me how to post pictures.&nbsp; If you want to keep track of the games, I think they will be posting scores and info on the website  urlLink http://www.childrens-games.org/ &nbsp; (Erika it made it a link automatically on the PC)&nbsp;  '),(11807,3409476,'26,July,2004','2004-07-26 00:00:00','   Well, I didnt get the job.  They \"chose\" someone whose department is being dissapated.  It really is a let down, this is one I really wanted.  I hate when they interview for show.  I know there are laws and union contracts and you have to interview everyone eligible, but those of us who understand that you would prefer someone or you have to place someone should be told ahead of time so we dont get our hopes up!  AArrgghh!  It is just frustrating.  So my next move is to put my resume out to the real world again.  I hate to leave the district, great beneifts and vacations and all but I also will hate to lose 2 months worth of pay.  On a happy note.  Adam has his reservation for Cleveland, the boys are all flying together on Thursday morning.  After an hour and a half layover in Nashville they should arrive in Cleveland about 9pm.  The opening ceremonies are on Friday with Aaron Carter singing the National Anthem (Megan doesnt understand why Adam isnt thrilled) and President Bush making an appearance.  Truthfully I would feel safer if he just stayed as far away as possible.    Yesterday was pretty uneventful, laundry all day before the cruisers return and take over the machines.  Then practice for Adam.  When we arrived at 5pm someone asked the coach what time to come back and pick up his kid.  The reply was \"DARK\".  Well at this time of year that can be anywhere from 7:30 - 9:30 so us girls went to the movies.  We picked up Kayla on the way and went to see Cinderella Story.  WHile not a bad movie, it is not something I would have seen if not for Megan (age 12) and Kayla (age 9).  But is was air conditioned (quite cool actually) and dark and quiet so all in all it was very relaxing.  '),(11808,3409476,'24,July,2004','2004-07-24 00:00:00','   Today was a busy day.&nbsp; I took Mel, Meg and Adam to Universal Studios.&nbsp; I don\'t know how much they liked it because the only roller coaster had a 60 minute wait and they chose not to wait that long.&nbsp; I however had fun seeing all the different shows and going on the tram ride through the back lot.&nbsp; At one point near the end of the day Adam is playing a basketball game and Megan says to me can I go over to the water?&nbsp; Sure I reply thinking nothing of it.&nbsp; Next thing I know Megan is soaking wet from running through the squirting water fountains!&nbsp; Of course this was right before we were planning to elave, so not wanting her wet jeans on the seat of my car, we buy her a new outfit to put on.&nbsp; Now, amusement parks don\'t really sell \"outfits\" per se but they do sell alot of pajamas and since wearing your pajamas in public is perfectly acceptable in LA thats what she wore.&nbsp;   The kids opted not to have dinner at Hard Rock Cafe at City Walk so we went to BJ\'s in Burbank and gorged on great food and the most wonderful dessert in the world the pizookie.&nbsp; I can\'t find a website for them so I will just have to describe it to you.&nbsp; It is a fresh baked huge chocolate chip cookie with a pile of ice cream right in the middle!&nbsp; YUMMY!  Megan turns to Mel as we were leaving and says I don\'t think I can eat for a week.&nbsp; Well the week is over she promptly came home and had some pudding.&nbsp;   Time for bed for me.&nbsp; Adam has practice with his travelling team tomorrow.&nbsp;  '),(11809,3409476,'23,July,2004','2004-07-23 00:00:00','   The interview went well, I think.  The position is in the Public Information office at the District.  They deal with all news media and city and county officials.  When I arrived I was waiting in the foyer when someone rushed out of the Superintendents office needing to retrieve an article from the local paper the Glendale News Press.  Well, I happen to read that paper everyday online so I went onto one of the computers and brought up the site and searched for the article and printed it for him within about 3 minutes.    So that was a good start.  There was 13 questions that I feel I answered well.  I should know the outcome by early next week.  '),(11810,3409476,'22,July,2004','2004-07-22 00:00:00','   Wow, I didnt realize how interested everyone was in Adam\'s baseball career.&nbsp; Ok, I found out about Clevland about a week and a hlaf ago.&nbsp; The city of Glendale is sponsoring a team.&nbsp; I have not heard back yet about the playing time.&nbsp; TOnight we fill out information and find out about flight information.&nbsp;   As for the link, you can only make it a link on a PC.&nbsp; All this really cool stuff like font and everything is not available on a MAC, which is what I have at work.  '),(11811,3409476,'21,July,2004','2004-07-21 00:00:00','   In case you were looking for her she is hiding at www.rejeanne.blogspot.com  '),(11812,3409476,'21,July,2004','2004-07-21 00:00:00','   Oh well, we blew their streak.  We had great seats for last nights game.  Field level righ behind 3rd base.  Adam got 4 autographs and before the game and came up to the seats with the biggest grin saying we can leave now.  Well we should have.  ALthough we rallied in the 9th it wasn\'t quite enough.    Jackson would have a very cute new Dodger outfit if his cousin Melody wasn\'t so mean.  Just because Erika once stated he doesn\'t need anymore clothes she believed her.  Oh well there will be other games without her around and we will spoil him then.  Adam and I have a big meeting tonight, he is going to Clevland next week to play in an International Children\'s Games thingy  http://www.childrens-games.org  All we pay for is the airfare.  The kids stay in dorms and play against other kids from around the world.  So if anyone will be in or around Clevland next week go see my boy play.  I am a little nervous about sending him alone but  he will be with friends.  '),(11813,3409476,'21,July,2004','2004-07-21 00:00:00','   Last night was my first night home without any baseball.  So of course we vegged out in front of the TV.     I just can\'t get enough reality shows.  I guess my life is so mundane I must live vicariously through others!  I watched the first episode of the new Trading Spouses last night.  If you get a chance to see arepet of this it is hilarious.  Two women from completely different backgrounds change places for a few days with a payoff of $50,000 at the end.  There are a few catches of course but the two that traded this time is like Florida from Good Times and someone from Dynasty.    Of course Big Brother is my all time favorite.  If you haven;t watched your tape yet, you should move away from this blog.  I can\'t believe eraser head took off Paris Hilton!  It is fun to watch with us at home as we rename everyone on the show.  Nakomis/Nebraska/Nickle  what the heck is her real name anyways?    Amazing Race is awesome once again , but why does it have to be on so late?!?!?!  Shmirna and Mirna provide plenty of comic relief throughout the show.  '),(11814,3409476,'20,July,2004','2004-07-20 00:00:00','   I have an interview on Thursday for Administrative Secretary Confidential in the Public Information office at the District.  For those of you that aren\'t aware, I am looking for a new job because the school I am at will no longer be yera round as of the 2005-2006 school year, which means I will not be a 12 month empoyee.  I will have 8 weeks off in the summer UNPAID!  My last resort is to get a new job outside of the district.  I have my transfer papers in to try for any open 12 month 8 hour positions.  So keep your fingers crossed for me.  '),(11815,3409476,'19,July,2004','2004-07-19 00:00:00','   Oh well, our Junior boys all stars put up a great fight but lost their second game in a double elimination round last night.  The only good news is I don\'t have to drive to Palmdale through the fire storm anymore.  I think I got the smell of smoke out of my car.  '),(11816,3409476,'19,July,2004','2004-07-19 00:00:00','   Okay I might as well jump on the band wagon.  If I get realy good at this I can put pictures up as well.  And although my children are cute, they are not nearly as cute as Jackson.  '),(11817,3409476,'10,August,2004','2004-08-10 00:00:00','   I finally got those pictures uploaded.  It took 2 hours on my modem!   urlLink http://pg.photos.yahoo.com/ph/adriennevh/my_photos   There are a lot of pictures of no interest to some of you.  '),(11818,3409476,'09,August,2004','2004-08-09 00:00:00','   urlLink http://pg.photos.yahoo.com/ph/adriennevh/album?.dir=/edd3&amp;.src=ph This is the Yahoo picture album from San Diego. I will do Cleveland tomorrow. I tried to write them to a cd for Gretchen but of course there is something wrong with my writer. I will try to get one done before Melody goes back there for a visit.  '),(11819,3409476,'04,August,2004','2004-08-04 00:00:00','   We rolled into the house (a wonderfully clean house, thank you Ris) about midnight last night.  Just got up watched a little world series action (mom are you watching this year?).  Going to Costco and then will post tonight or tomorrow (I have taken the rest of the week off of work to recover from this weekend) with a try at posting pictures.  '),(11820,3409476,'04,August,2004','2004-08-04 00:00:00','   I don\'t know how many of you have tried to pump information out of an exhausted 14 year old boy before, but it aint easy.  The team walked off the bus looking like zombies, they had not slept the night before, too much adrenaline I am sure.  On our drive home to his bed, I got a call from a co-worker offering me free Dodger tickets and parking, as I was going to decline because of exhaustion this person ion the back seat woke right up and insisted that we go pick up Melody and go.  Good thing we did because parking was right next to the entrance.  Our seats were irectly behind home plate, and the DOdgers played a great game.  So throughout the night he spit and spouted tiny bits of information about the trip.  I will post those when I am home at a PC and can do so with formatting.  '),(11821,3409476,'03,August,2004','2004-08-03 00:00:00','   Due to a glitch in my ability to not only remember to set up a VCR but to also explain to someone over the phone how to do it, we missed all 3 shows last night.  SO be fruitful in your comments so we can enjoy through you.  '),(11822,3409476,'03,August,2004','2004-08-03 00:00:00','   http://www.latimes.com/news/local/glendale/sports/la-gnp-childrensgames03aug03,1,4225637.story?coll=la-tcn-glendale-sports  This link will get you to the article in the Local New Press which is a part of the LA Times  '),(11823,3409476,'02,August,2004','2004-08-02 00:00:00','   GOLD!   GOLD!   GOLD!   GOLD!   GOLD!   GOLD!   GOLD!   GOLD!   GOLD!   GOLD!   GOLD!   GOLD!   GOLD!   GOLD!   GOLD!   GOLD!      Adam called me a little while ago and let me know they went undefeated and will receive the Gold medal at the closing ceremonies tonight.  I have just landed back on Earth (but have yet to stop crying) so it took me some time to get this post up.  Erika very nicely sent me training on pictures so as soon as I get some they will be up.  '),(11824,3409476,'01,August,2004','2004-08-01 00:00:00','   http://www.childrens-games.org/results_frame.html?event=baseball_boys.html  You can go to the website and get updates on the games.  We found out last night that the team was supposed to be on Good Morning America today but I have not been able to watch it here at work.    But I have set it up to tape at home.  As soon as Adam calls me I will update everyone.  '),(11825,3409476,'01,August,2004','2004-08-01 00:00:00','   The boys got to play today and play they did.  They beat the S. Korea team 18 to 0.  ANd also beat the Newar NJ team (I don\'t know the score.)  So they are at this time the #1 seed.  According to Adam they are guaranteed the Bronze, if they win one game tomorrow they will get Silver and if they win both games tomorrow they get Gold.  So all in all a very good time is being had.  The Rock N ROll Hall of Fame was a bust even the old guy (coach Steve my age) said it wasnt worth the time or admission price of $20.  '),(11826,3940997,'15,July,2004','2004-07-15 00:00:00','    Dear World,   &nbsp;OK, so this is my first post and I am walking in to this thing blind folded but I will figure it out.     &nbsp;  Well, I might as well start off with some sad news...one of my fellow friends passed away due to someone else\'s stupidity in a car. He was and is a great man...I went to his wake tonight and tomorow is the funeral..which I am not looking forward too. I know that is going to be the hardest part. Its so final and all that.&nbsp; If&nbsp; I can be as happy as he was at the age of&nbsp;48 then&nbsp;I will be blessed. I mean, he never had a negative thing to say...and he didn\'t have one enemy. Now some people say that and its false...this is so far from false...He had so many people at his wake tonight..just celebrating him and talking about all the things he did and how many lives he touched.&nbsp; So, Gary Leblanc will live on forever..in all of the hearts of the lives he touched...I miss you!    &nbsp;  Lets see, another interesting topic is umm..the wonderful men in my life that I tend to always&nbsp; end up with. Call it fate or call it being unlucky...but they all tend to share one adjective \"asshole\".&nbsp; Maybe its the fact that they are like normally 8-10 years older than me or maybe its because they are all bachelors-never been married-no kids- and no focus besides themselves and what piece of ass they are getting tonite mentality. Let me explain a bit more for those of you who are interested...if not...scroll down to the next paragraph... So first there was Colby when I was 18 which was a total disaster. He was 26 at the time and I guess I was a trophy. But regardless...almost 11 mos. in the so called relationship I thought I was in it all went sour. So a man who constantly used to want sex...was now saying he was tired or sick or whatever it was that nite...and I was so naive that I didn\'t even see it coming. So stupid I was even after we broke up I wasted another year on the guy only sharing half of his life..the unknown to anyone else..(leave out the back door not front type of relationship). Until finally he found a new girl, Cassidy, which by the way \"FUCK YOU Cassidy\", ok i feel better now, to slowly waste&nbsp; away on. To this day he still pops in on my life. NO, not like that.. I wish :), its more like a hey whats up? who are you dating? would you still sleep with me? kinda pop in. We share the same circle of friends so its often that I bump into him but moving on to the next ass....   &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; So here comes Eric, so sweet and so kind ...STOP, ok, like that could ever be true..WAKE UP JESSICA...So, here it goes...He has been married and divorced twice and in the process of his third divorce...at the age of 31...duh! should of been a sign huh? Well, I am naive remember? So this also lasted blissfully for 10-11 mos and then the same signs came again....sooner or later it always happens right? The only difference in this one is that he is alot ruder...He came right out and said he was done with it and he meant it....so for 3 mos. never heard from the guy and last week he calls me two times! Sweet talking me as usual and I can tell he is buttering me up for somthing but I am not gonna fall for it...besides If&nbsp; I do...what am i to do with my half current ass.. ????&nbsp;    &nbsp;&nbsp;&nbsp; His name is Will, and he 33, seems like hes got it all together...he is enrolled in grad school..works a steady job at local club as a manager andworking on a new club in a surrounding city...wines me, dines me, and all the other hoorah..i\'ll save you those details... : ) and this is off and on for like 7 mos...but recently I was told I will be really busy next week...so it has been two weeks since I have got a phone call..so this is either an off time or never again time...I guess only time will tell but my goal for this fall semester is to find a guy who is not 30 plus and just not worry about all the bull shit..so whoever is to blame here really doesn\'t matter. I guess they are just too old for me...I mean really how can a 23 year&nbsp; old and 31 year old or around that age relate? Well, besides, between the sheets...thats a given... So, no more of all that...I am just gonna focus on school and hey if I find a dude then cool, if not..hey one less asshole around. ha&nbsp;    &nbsp; I have no idea what you are thinking right now as you all are reading this but this is like 6 years rolled into 4 paragraphs..so don\'t judge too hard...there is alot of pieces missing in this story of my clueless love(lust) life. Have no fear I will keep you guys posted of all juicy updates...don\'t hold your breath though.&nbsp;    &nbsp; Well, if you have made it this far congrats and I must part now..but I will&nbsp;leave you with this thought ....sex can never be just sex...it always gets complicated...ALWAYS.. I\'ll continue with a new post tom.     &nbsp;  Nite World,   Jesse '),(11827,4052055,'29,July,2004','2004-07-29 00:00:00','   The summer semester is officially over - took my final yesterday... we\'ll see how that goes.&nbsp; BTW - I got a call from Jimmy D the other day.&nbsp; He wants to email me this letter.&nbsp; I\'m curious, but I\'m also willing to bet I know exactly what\'s in it.&nbsp;   Just checked my grades - 100% on my Algebra final!&nbsp; Yea for me.  &nbsp;Nise & I will be going to the Stitch &amp; Bitch at Aalim Saturday - I REALLY need to get to work on my costume - I don\'t want to bring just the pieces!&nbsp; I also need to finish her bra.&nbsp; I\'m hoping I can get some help with hemming my chiffon panel skirt.&nbsp; Wish me luck.  &nbsp;- Kirvi  '),(11828,4052055,'27,July,2004','2004-07-27 00:00:00','   Everyone says that&nbsp;keeping a&nbsp;journal is good for mental health, good for the soul.&nbsp; Of course, I\'ve never been much into what anyone else says (see profile, and a dictionary for the meaning of \"introvert\" if necessary.)&nbsp; Studying for my Algebra final this week - I have to take it by tomorrow if I want the extra credit points.&nbsp; Of course, I had to take tard-math due to my having enjoyed the 80\'s far to much.&nbsp; That and a 12 on the math portion of my SAT\'s.&nbsp; Think about that for a moment.&nbsp; Allow it to really soak in.&nbsp; 12, baby.&nbsp; Big 12.    &nbsp;  After the total blow-out of a vacation I just had - wait, that could be misconstrued in this context.&nbsp; Strike \"Blow-out,\" insert one of the following - \"annihilation\"&nbsp; \"destruction on a global scale\"&nbsp; \"seventh outer ring of Hell\".&nbsp; Ok, better.&nbsp; Now that I\'m back, I can\'t believe I didn\'t buy any of the cowrie shell belts or beaded/sequinned wrap skirts I saw.&nbsp; I really hate to spend money, and I was pretty occupied with fighting that part of my personality that wanted to act like the child of rich, divorced parents - \"you guys got me into this shithole - now compensate me materially for my anguish.\"&nbsp; So I didn\'t really buy anything much.&nbsp; Tommy did get me this beautiful carved statue, it\'s a man & a woman, very minimalist.&nbsp; And I got my tattoo finally.&nbsp; I say finally, like I\'m the one who\'s been wanting one.&nbsp; Oh, I kicked the idea around, but T really wanted one, and he really wanted me to have one too.&nbsp; So I got mine first.&nbsp;&nbsp;  urlLink http://www.geocities.com/morrigandances/dragons.jpg &nbsp;     &nbsp;  Well I should probably get back to work - which today means, studying for Algebra final while pretending to contribute to the corporation\'s goals.&nbsp; And stealing FudgeRounds out of the Atlas snack-drawer.    &nbsp;  &nbsp;- Kirvi  '),(11829,4196591,'16,March,2004','2004-03-16 00:00:00','   Hey we have a fuckin snow day im march. haha that\'s great. But you know. im so bored there is nothing to do I want go somewhere lol oh well I miss my cookie too but we\'re not doing to well I don\'t think but maybe its all me I don\'t know. But anyways me and daevion are over fuckin done. End of story........ Au revior   '),(11830,4196591,'01,July,2004','2004-07-01 00:00:00','   ::WORTH:: hey everyone! Long time right? Well shit happens. The summer is so boring. All I have been doing is sleeping and watching movies. That\'s it. Incredibly exciting right? DRAMA:Adultery. My friends are fucking each other that\'s great isn\'t it? Oh well it doesn\'t bother me anymore because... Beyonce said it best,\" me, myself, and I that\'s all I got in the end\". That\'s the truth I found out that im better off without anyone be relationships aren\'t worth it....**DAEVION: I pushed him away for no reason what-so-ever and I am now willing to face the truth that I love him more than anything in the world and it was not worth the relationship that I wanted to work but the reality of it was that it never was going to work out, and I could have been with the one whose love for me was undying, to attempt a relationship with someone who lies about everything and who only wanted ass. :::DAEV... I love you::: Au Revior   '),(11831,4196591,'19,August,2004','2004-08-19 00:00:00','   hi everyone....im sssooooo tired...i just woke up and i wanna go back to sleep but im not so youh you kno.and people wont get the fuck off of the phone. but of well anyways.....i feel like going skating on friday. i hate the new one because it is so commercialized and no one goes there to skate anymore but i guess thats how things are.   '),(11832,4196591,'16,August,2004','2004-08-16 00:00:00','   omg...i was doin my daily search on my jon and i found out that the new album is coming out either on september 14th or november 25th. thats too long. but oh well. last night me and dev had similar dreams. it was fucking scary. but anyways i love jon. its pathetic i know but still. well im going to continue my search..... Au revoir   '),(11833,4196591,'13,August,2004','2004-08-13 00:00:00','   whats up bitches! its friday the thirteenth! woo-hoo its so great. btu anyways i had a fucked up dream last night it was about my daevion. but the only think that is really vivid and that still troubles me is the fact that someone said it all was a  hoax.  i hope to god it wasnt. but then again it would be ok if it was because i miss him. and i  still  love him to this day. well now that i am sad again im gonna go... Bye   '),(11834,4196591,'12,August,2004','2004-08-12 00:00:00','   im tired of people again today and i have only been awake for a total of 29 minutes. oh well. i have i get my new cd\'s today that would be great because i gotta fucking go out of tow to see my uncle and his wife, who i disown. she is a bigger bitch than i am. but of well. that nigger(aka the sperm donor) hasn\'t called back aparently what he has to say wasnt important. he\'s prly back in the hospitol. sucks to be him don\'t it? i dont give a shit about him, because he expects me to act like a child and not voice my opinion and that is not me. he got a rude ass awakening when he was here. but now he doesnt have a reason to come back. like i hate him and his mother because they expect me to respect them and they dont show me any so fuck that. you\'ve got to give respect in order to receive it. but i love my aunt. she is cool and everything, she is so much like me its crazy. she is nothing like them. my cousins are alright too. well im gonna go.... Bye   '),(11835,4196591,'12,August,2004','2004-08-12 00:00:00','   i thought shit had changed but i was wrong by far. i dont care anymore it is out of my hands. just listen to me when i say: DONT EVER INTRODUCE YOUR FRIENDS TO A SIGNIFICANT OTHER. i am tired of the summer, well this summer because so much bad shit has happened. i am prly going to court which is a good thing because hopefully i wont ever have to see the sperm donor. he is a fucking fag. but oh well. im so sick of people. well only 3 years right? bye   '),(11836,4196591,'11,August,2004','2004-08-11 00:00:00','   why do they think that have to keep things from me? since when am i not able to handle it? idk... idc it doesnt matter. they bath will probably get something. muhahah! oh well i dont know why i care. they dont give a shit about me. so fuck it all... Bye   '),(11837,4196591,'10,August,2004','2004-08-10 00:00:00','   im bored... but thats the norm.  but oh well its all good tho. i want devan to come over tonight that would be great... but she will try to bribe me into telling her about what \"it\" was saying about her, and im not because i dont want to hurt her or anything, and i kno if she knows what he said will truly hurt her.   '),(11838,4196591,'10,August,2004','2004-08-10 00:00:00','   aliens and boredom have taken over my mind... im bored....but anyways... i was just thinking why do relationships have to be so complicated? is hasn\'t that way with daevion but nothing was with him. Adultery, Monogamy: The hardest things for any man. The purpose of why i will stay a \"single cold superbitch\" until the day i die. *Thought* --Why can\'t people (in general) be what their not?   '),(11839,4196591,'10,August,2004','2004-08-10 00:00:00','   HOLA EVERYONE! today is a good day not like the rest. im beginning to get over daveion...but im not that sure if it is a good thing or not. i hope it is but idk. nut the story behind that is that he died in a car accident on my birthday and i knew about it before it actually happened because of my dreams. that fucking sucks. so for the next few says after i vowed not to sleep because i did not want to hurt anyone else, and i didnt for about 5 days. but i cant not sleep forever. i just need some of the drugs that the kids were on in freddie vs. jason so i can not remember my dreams. but thats not something that can happen well thanx im sad now.   bye.   '),(11840,4075808,'31,July,2004','2004-07-31 00:00:00','   Once upon a time I had mused on the topic of the initial courting period. I said you canâ€™t believe who a person is, or rather presents, during the first 6 months. Weâ€™re all on our best behaviour. Thereâ€™s no farting, no real scratching of any body parts that arenâ€™t publicly sheathed, and thereâ€™s an internal sensor diagnosing what should be said and filtering what absolutely should not. Sure, we fall in love, but in the end is it the real person weâ€™re falling in love with or is it the illusion that person created? We created.   Itâ€™s exhausting.  Iâ€™ve just recently found someone. I admittedly placed an ad in the mega online dating service. The way I phrased my profile I thought Iâ€™d get... well, I donâ€™t know who Iâ€™d get. He was my first contact. My first. My high school lover. My favourite flavour. And things went well... eventually. And now he means something.   I fucking hate this.   Why do I think heâ€™s going to break up with me at any moment? I get nervous and paranoid if he hasnâ€™t emailed me by a certain time. Yes, of course he has work he needs to do. Itâ€™s fantastic knowing there are other people in his life that he needs tending to. My rational mind fully comprehends that. My irrational mind is often wondering, â€œDid I say something wrong in my last email? Was there something I did or said that could make him change his mind about me?â€  Perhaps the reason for our need for perfection is if we feel perfect, flawless, we will be rendered lovable. If we are less than perfect how can we be loved?  We find it so easy to love and forgive our friends when they are painfully, endearingly and beautifully flawed. Why is it so difficult for ourselves?  Maybe itâ€™s from being scared of needing someone.  Iâ€™m 35 years old. Iâ€™m supposed to be a big girl now. I should have a career. I should have a car still. I should still own a house. But I donâ€™t anymore. I gave all those things up, including the big girl, when I left my marriage. I should be able to do it on my own. Thatâ€™s what the ads say. They donâ€™t play Double Mint gum commercials anymore, where pleasure and fun were amplified when in pairs. Thereâ€™s now a societal pull towards self-sufficiency, I-am-Woman-hear-me-meow mentality where the image of strength within femininity is wearing very powerful lingerie and using giant cans as a weapon of mass affectation.   How do I tell him he feels like home when he holds me?  Thereâ€™s so much that new love stirs up. The two people in question become mirrors to each other. In the beginning you can see what that other person sees in you: that perfect creature youâ€™ve always wanted to be and sometimes convince yourself you are. The problem is maintaining perfection is an unachievable illusion.  We create things, but if they donâ€™t stand the test of time or technology we scrap it, forgetting the path which took us there and in the end abandoning all and ourselves. We become embarrassed by our past. â€œWe were young and stupidâ€, we worldly surmise over a luscious cocktail... or nine. We berate ourselves, relinquishing process for product where the final end result had better be fucking good and weâ€™d better have something bling to show for it.   When one has been in a serious, long term relationship (Iâ€™d been married for 7 1/2 years), we often end up resenting the other person mostly because we have felt like weâ€™d been taken advantage of. We too had also taken advantage. We know theyâ€™re going to be there at the end of the day. We know that as soon as dinner is ready weâ€™re going to head for our appointed stations in front of the television, yet again captivated by an episodic Frasier Crane doing a repeat performance, studio audienceâ€™s canned laughter, cold comfort. We say we hate it when weâ€™re in a relationship where one day repeats itself over and over again. A glitch in the Matrix. Not even a glitch, itâ€™s more like a master reel someone inserts; a cinematic day-in-the-life scenario that is rewound and replayed ad infinitum. Oneâ€™s life amounting to nothing more than edited dailies, eventually understanding fully how there is no greater loneliness than the lonely felt while being with a partner.  So we move on (give up?) when the going gets boring and thereâ€™s no sign of hope or light. We resolve to take this new knowledge, the stuff we donâ€™t want, into the next relationship. But where does that leave us? Being a bunch of spoiled kids with miniscule attention spans never fully understanding the sacrifice and potential payoff of total commitment? Or in search of true love? I donâ€™t know.  What I do know is when I look at the new one itâ€™s absolutely exhilerating. When he holds me it feels like home. Heâ€™s careful and considerate. Heâ€™s flawed. Heâ€™s marvellous. But falling in love with someone new really makes you examine all the bits of yourself that you just relaxed into and didnâ€™t really care about because there had already been someone who claimed loved you, knew you, even when you didnâ€™t feel you were getting the kind of love you wanted or hoped you deserved.  I think Iâ€™m just scared of the response. What if my feelings arenâ€™t reciprocated? I know I should remember that Donald Kaufman line from Adaptation, â€œYou are what you love, not what loves youâ€, but it feels like such a huge risk. I donâ€™t really think the odds are against me. I just never realised how chicken I can be. Loveâ€™s making me question everything. Myself. My surroundings. The way Iâ€™ve been living my life up until now. I feel like a death-defying tight-rope walker with no balancing rod, just an empty, cavernous pit below me.   But I do see his smile. The safety net.  '),(11841,4075808,'19,August,2004','2004-08-19 00:00:00','   So Iâ€™m planning a party for my ex-husbandâ€™s belated birthday party this coming Saturday. Called all the peeps, deciding on what to make, to buy. Thankfully I\'ve worked out what I\'m wearing.  I got t-shirts made with iron-on logos in the front commemorating the day. Well, not really the day, but just a nickname Death calls Mike. I call him Ack. Death calls him Peenut. My print broker thought the spelling was off; it suggested a leaky ball. I think that was Death\'s design. I kept the actual design to a cute little line drawing. The shirts are wicked. We all have our pet names on the back, with our special numbers we\'re particularly attached to. Mine is Sunshine 33. Death\'s is Death 13 and Ack\'s is Peenut 36. Yay! I was so delighted when the t-shirt guy applied all the letters and numbers. What a treat. It\'s better than seeing yourself on television!  Suddenly I find myself feeling a little better about circumstances. No, I still donâ€™t have the love I want, but it is really great to have some of those people that you think are Godâ€™s gift to the world come over to your house for a one night event. I think this party will be just as important for me as it will be for Ack.   And my sweet Fergus is coming! Yay!  '),(11842,4075808,'17,August,2004','2004-08-17 00:00:00','   I was writing in my journal this morning about my last night. Kind of pits of despair nether-regions. Nothing too drastic. It only made my want to have one little glass of vodka, happy I didn\'t have any beer in the house as it would have been too much liquid, and truthfully, the bubbles would have been too chipper and cheerful. Last night I felt A-L-O-N-E.   I was in such a yuck state last night I ended up emailing David. Yeah, great. Like I haven\'t put myself on the line enough lately. Obviously I just needed further rejection. I think I just felt at the end of my tether emotionally. What\'s another stab? Like I haven\'t been humiliated enough recently.   I do like being alone. What I don\'t like is being lonely. Being single and alone is a hell of a lot better than being with someone and alone. So I\'ll take it, if there is a choice. And if I\'m to remember the big girl who is embracing Buddhism and mysticism then I should remember that these feelings bring me closer to my own truth. However... I still felt lonely last night. Why isn\'t a wonderful body of friends not enough sometimes? Well, you can\'t get too physically close to friends. It\'s too confusing. I was dying for it last night. Not sex, per se, just the closeness felt when two people who care so much about each other are together and beauty is created. Then tears welled up. Sucks.   I can\'t tell everyone about this. Yeah, so that\'s why I\'m blogging. I told Dirty. She was great. She told me she was lonely too. Somehow this made me feel better. I don\'t know why. It really makes no rational sense. I think it was the same feeling I had when I watched that 20/20 episode on Attention Deficit Disorder. I was no longer alone. There is comfort. The thing is I think what so many people are going through is utter bullshit. They don\'t want to get tied down. They don\'t want the titles \"girlfriend\" or \"boyfriend\". It implies too much. Makes them responsible to too much. To one. Really, they are so afraid of having their guts torn out that they hold back constantly, never fully giving of themselves, always the back of their minds thinking this new person is going to rip the shit out of them. Me. Shit-ripper. So they sabotage their efforts.  My friend Dirty was supposed to go on a buying trip with her current boss down in NYC. The night before she\'d run into me quite by accident. A happy accident. We decided to go down the street and visit mutual friends that worked at a neighbouring restaurant, after closing hours. Dirty decided she could stay for ONE drink. Ian, my lovely, wise for his years and wonderful friend, generous to a fault, particularly when it comes to pouring vodka, made Dirty a Caesar. Normally, and I have worked as a bartender for many years, this drink contains, at most 1 1/2oz of liquor. Ian didn\'t measure, but by internal counts he\'d put in approximately 5oz of vodka. Ian didn\'t know Dirty had driven. Ian also didn\'t know she hadn\'t eaten anything that day. She\'d also ordered a shot of tequila, something I\'d never seen her do in all the years of knowing her. A decade. She got so wasted she was hurling matter and non-matter in the basement\'s washroom, replete with the most beautiful and refreshingly chilled porcelain goddess Dirty\'s ever borne witness to. She worshipped for an hour.   She poisoned herself. She didn\'t want to go to New York with Wendy, her boss. She was trying to sabotage the effort. She wanted to make herself so sick she couldn\'t possibly board a plane, let alone drive with her to Buffalo, Wendy\'s idea, where airfare is deeply discounted from there vs. from Pearson\'s YYZ. Cheap cunt. She still went, begrudgingly obliged, still sick, and had a horrible time.   Ian is my self-professed Detachment Specialist. He doesn\'t listen to the women he has sex with if they talk about themselves and try to endear themselves to him. He says, \"Mm-hm\" and \"uh-huh\" in roughly the right places, but when there\'s a lull in conversation he\'ll turn to the girl in question and say, \"Sex, now? Now sex?\" Ian heatseeks \"warm bodies\". He says he doesn\'t want the complication of the next morning, or mornings to come. When I ask him about meeting the great girls, the ones he can really see himself with and potentially giving him some extra happiness, he wriggles in his skin saying, \"Too scary. Don\'t want.\" He later confesses his tactics are a lonely road.   What do I want? I want something incredibly meaningful. I don\'t want \"him\" to be my everything. Whoever Him is. I\'ve got a lot going on with my friends and the life I really love right now. I want extra meaning, though. I want to be cherished and loved in the same way I give it out. Well, not exactly the same, otherwise I\'d know exactly how it\'s to come and it would no longer be so exhilerating. Kind of like tickling yourself. You know it\'s coming, so it\'s not tickly. So why is masterbating so good? Hmm.  '),(11843,4075808,'17,August,2004','2004-08-17 00:00:00','   Lying in bed with Jeremy the other night we were talking about feelings and how at this one particular point I\'d felt as if my heart had opened, broken and this invisible metal shielding came up to protect it. My eyes would almost glaze over and I felt... nothing. It didn\'t feel right nor good. He thought it was positive. These mechanisms were in place so we didn\'t have to hurt as much or, in reality, feel as much. I told him I don\'t operate that way. I don\'t not want to feel everything. I don\'t think it\'s healthy. We try so hard in so many areas of our lives of pretending so many parts. We pretend we\'re happier than we are, more together, maybe a little richer, more successful. But really it\'s something other than what we\'re really feeling.   My mother had called a few days ago asking if I\'d accompany her to Alberta to do a reconnaisance mission to see if my estranged brother was alright. I told her I wouldn\'t go and I didn\'t think it was a good idea for her to go. She thought otherwise and accused me of being less than compassionate. I told her that everyone has his own time to figure things out. Sometimes it\'s harder to figure when there\'s someone over your shoulder questioning, wanting, wishing for more; answers, reasons. People need to come to their own conclusions in their own time. All we can do is support them in their time of needing space to make those decisions. Sometimes, in our infinite wisdom (she spake with a great deal of irony), we know exactly what the problem is and what the solutions are for these individuals, but it takes a great deal of love and compassion to step back and allow them to figure it out themselves. This is a lesson that took me 35 years to figure out. So when Mom asked me how she could demonstrate her love for her son, I told her to keep her love in her heart and when he comes back, and he will, don\'t hold anything against him for wanting to figure things out and taking the time.   I told her I did very much the same as my brother, Walter, in terms of extricating myself from the family at times. She didn\'t understand. She didn\'t really remember. I stopped going to certain family functions just because they weren\'t fun. No one had a good time. Everyone was trying to convince everyone else of his or her own personal success, whatever that meant to the individual. For me it meant projecting far more happiness than I actually felt, projecting a happier marriage than I really had. This is the problem with clowns.  Clowns have a paid responsibility to make the masses laugh. It\'s what we think is our mandate. Laugh and the world laughs with you. Cry and they just leave. So keep \'em laughing. And on just this topic:  From Leoncavello\'s Pagliacci, Canio\'s aria, Vesti la Giubba:            ~Consumed with despair on discovering his wife\'s infidelity, Canio gets ready for performance because the show must go on... First in Italian, well... just... because!  RECITAR!...                         VESTI LA GIUBBA	    Recitar!... mentre preso dal delirio non so piÃ¹ quel che dico e quel che faccio! Eppur... Ã¨ d\'uopo... sforzati! Bah! Sei tu forse un uom? Tu se\' Pagliaccio!  Vesti la giubba, e la faccia infarina. La gente paga e rider vuole qua. E se Arlecchin t\'invola Colombina, ridi, Pagliacco, e ognun applaudirÃ ! Tramuta in lazzi lo spasmo ed il pianto, in una smorfia il singhiozzo e\'l dolor - Ah! Ridi, Pagliaccio, sul tuo amore infranto. Ridi del duol che t\'avvelena il cor.  And in English, for us gloriously sad and loving creatures of beauty and buffoonery:   TO PERFORM!... PUT ON YOUR COSTUME To perform! In the throes of delirium I don\'t know anymore what I\'m saying, what I\'m doing! Still... you must... force yourself! Bah! Are you a man or not? You\'re just a clown!  Put on your costume, and make up your face. People are paying, they want to laugh. And when Arlecchino takes away your Colombina, laugh, you clown, and everyone will cheer! Turn your agony and your tears into buffoonery, your sobbing and pain into a funny grimace - Ah! Laugh, you clown, at your broken love. Laugh at the pain which poisons your heart.  '),(11844,4075808,'13,August,2004','2004-08-13 00:00:00','   Holy Moly! There is one thing in this world I will never fathom: the consistent sketchy behaviour of some people. And they are consistent, in their sketchiness. Now with sketchy, I\'m not talking about the drug induced sketch of today, I\'m talking the old school definition: adj. one who is routinely late, and usually coupled with being unreliable. Yes. Well, much of my friendbase consists of these such people.   There\'s Damian. Damian\'s wicked. I love Damian. What Damian does is promise you the world, but in the end he backs out, or changes plans without consulting you and you\'re left potentially with packed bags anticipating a cottage weekend, mentally preparing yourself for bug bites aplenty, but the phone doesn\'t ring. You try to call him and the phone doesn\'t answer. Damian is so sketchy that you can\'t even leave a message on his phone because his mailbox is constantly full of people in exactly the same situation I am. He makes up for it, though. He\'s always very, very sorry. But mostly he\'s disarmingly charming and eventually I did end up going to said cottage and having a wonderful time with him and other friends.   Then there\'s Jeremy. Jesus Christ. He has this affliction, not unlike Leonard Shelby\'s, from the movie Momento. He has little to no short term memory. It\'s fascinating and it\'s irritating both at the same time. For example we\'d made tentative, unconfirmed plans to see a baseball game. Things were fine. We\'d been chatting throughout the week via email. Nice, nice, nice. Then at one point I\'d asked him how his Friday was looking. He said he had these tickets to go see a baseball game. No mention of going with me, just that he had these tickets.... Yeah. It\'s impossible to make plans with him mostly because there\'s no real way of knowing if he\'s going to remember he\'d just made plans with you. It\'s wild, really. He\'s really bright, but there are moments that I think, \"Jesus, he can\'t be that stupid.\" So, I\'m now the stupid one because *again* I\'ve made plans with him and he\'s completely sketched out again. He called today around lunchtime. Chat, chat, chat. Fine. We make loose plans for tonight. He called firming them up. Said I\'d receive a call by 5:00. It\'s now almost 8:00. Yeah.   Okay, well he just called and he\'s not a totally unsensitive jerk-wad. He was out with a female friend/ co-worker and she had just told him the most heinous, horrific tale of her marital predicament. It turned out that he\'d also received a package from his lawyer outlining his divorce proceedings. So he and his friend had much to comiserate. He had just lost track of the time. It happens. Not to me, but it happens. When he did call he sounded totally shaken up, mostly from the tale told to him by this young woman. I\'ll never know this story, which is just fine. I\'m glad that he was a friend to her, in this world of not-too-friendly people sometimes. And this is why I can still be friends with Jeremy and forgive him his own very unique brand of sketchy behaviour. If we can\'t understand and then forgive our friends their foibles and inconsistencies then what kind of friend would we be? Not so good.  '),(11845,4075808,'10,August,2004','2004-08-10 00:00:00','   I love boys. Guys are great. Men are alright, but guys and boys are righteous! Case in point: I\'d just gone to the Curiosa Festival to see Mogwai (wicked) and Interpol (ruled). I met up with Mike, Pete and Gotham. Gotham, I\'d pointed out, had the best name I\'d heard all day. Probably all month, if not all of this year. I like the dynamics of boys.  When guys meet other guys, in certain circumstances, it\'s very reminescent of being kids. When we\'re kids we see another kid, approach them simply because we\'re the same height and at the same eye level. There\'s also the potential of play involved. One kid will invite the other kid to play. Sometimes there\'s an exchange of names right from the start. Sometimes the names aren\'t brought into the fold until the end, when someone\'s Mom calls him in for dinner. Last night was very much like that.   I knew Mike, a cook, from a local watering hole. He brought Pete, a best mate of Jimmy\'s, one of the bartenders at said watering hole. Both Pete and Jimmy are from England, England. Both are hilarious. Jimmy couldn\'t come. So the three of us are laughing, smoking pot, talking about stuff, critiquing a particularly bad band, and laughing extra hard because the smoke has caught up. Then to my left, Gotham had appeared. No one knew Gotham, but he had a wonderful sense of humour and play. He was invited into our circle immediately. He was terribly enthusiastic about everything, which was refreshing. I found a kinship in him. Instead of using balls, or mallets or gloves or other props, we played with wordplay. It was scathing, harsh, sweet, endearing, incredibly funny and sometimes fueled with delicious malice. But it was highly inventive and creative. Such fun.   What I really like about the interactions between guys is there\'s none of the weird shit involved when talking with girls, quite often, or at least it\'s different shit. I\'ve noticed, as a girl who often talks to other girls and quite likes other girls on a whole, there is a kind of procedure in dealing with them.   Unspoken Rule Number 1: You must make a flattering comment about something on them. It could be a piece of jewellery, their hair (they usually love this), eyes, shoes (also a favourite), handbag, etc. This is designed not only as an icebreaker, but it also shows the other girl is not a threat. You don\'t have to do this with guys.  Unspoken Rule Number 2: You have to talk about your feelings. This may be one of the interests the girl has right now. They really feel something for it. It must be discussed. Guys do this too, but they work it into the conversation more naturally but they certainly don\'t dwell on it forever. Usually with girls it\'s a complaint of something someone\'s done to them, an unfortunate event of needless anxiety at work, or something particularly heinous the old boyfriend did to her. Or, it\'s just mindless fluffy stuff. Girls try to figure out a solution to the problem. This could go on all night. Guys just say, \"Dude, that sucks,\" and move on.   Guys talk about really interesting stuff, too. Music, science fiction, books, pictures (both moving and still), hobbies and interests. They\'re always keeping abreast of stuff almost in anticipation of sharing these little tidbits of what most women consider useless information. It is not useless... usually. And this is totally subjective. If I\'m not in the mood for a certain topic, I\'ll just tune out.   Guys cheer other guys on. Guys genuinely like and trust other guys. Guys like to buy other guys beers. That\'s just nice. And guys are just happy to be my friend without expecting anything in return. Guys are great. Guys are good. Guys live in my neighbourhood.  '),(11846,4075808,'08,August,2004','2004-08-08 00:00:00','   \". . . All earth experiences are like the coloring used on the slides of a microscope to make you conscious of invisible things.\" -------  Stewart Edward White. With Folded Wings  I\'d watched American Beauty before but didn\'t truly understand the plastic bag sequence until this past week.   What a marvellous week.  The final monologue spoken by Kevin Spacey\'s character summed up what I\'d felt on that magic subway ride.   \"I\'d always heard your entire life flashes in front of your eyes the second before you die. First of all, that one second isn\'t a second at all. It stretches on forever, like an ocean of time. For me, it was lying on my back at Boy Scout camp, watching falling stars. And yellow leaves from the Maple trees that lined our street. Or my grandmother\'s hands, and they way her skin seemed like paper. And the first time I saw my cousin Tony\'s brand new Firebird. And Janie... and Janie. And Carolyn. I guess I could be pretty pissed off about what happened to me. But it\'s hard to stay mad when there\'s so much beauty in the world. Sometimes I feel like I\'m seeing it all at once, and it\'s too much. My heart fills up like a balloon that\'s about to burst. And then I remember to relax, and stop trying to hold on to it, and then it flows through me like rain, and I can\'t feel anything but gratitude for every single moment of my stupid little life. You have no idea what I\'m talking about, I\'m sure. But don\'t worry: you will someday.\"  '),(11847,4075808,'07,August,2004','2004-08-07 00:00:00','   I\'m trying something out; an experiment. I\'ve been allowing only grace, goodness and love enter in and out of my body for the last several days. By doing this, and it\'s not a willed feeling, it just is a choice, I have been experiencing the most delicious feelings looking at things that had been merely ordinary and mundane while in a \"normal\" state. Normal being the state in which we so often walk around - working, but not knowing why and often hating it, making money - but never enough, having friends but not being the best friend you can be. In other words: normal, in this society.   I am trying to live fully in the present. By that I think this quote I came upon online says it best: ...\"The circumstances and challenges in your life are not the real you. To deal effectively with these external circumstances, we must be detached from them. Observe what is in your life without judging, without being for it or against it. Just let it be. Detach yourself from it and just be with it. The moment you are for it, the effort to hold onto it begins; the moment you are against it, the effort to run away from it begins. And both struggles take you out of the present moment and into the future, into worrying about the future. This divides you within yourself; this divides you against yourself. When you are detached from the \"stuff\" in your life, you can take responsibility for it. You can act upon it, or you can just let it go. You can even truly enjoy it. You can live undividedâ€”in the present moment.\"  I\'ve been having these wonderfully seredipitous moments lately. I\'ll have questions in my head concerning the ways I\'ve operated in the past or just qualities about myself that I wasn\'t sure about. I\'ll happen upon things that help me in my search for self understanding. The answers come if I pay attention. Sometimes just plain wonderful things happen if I pay attention.  I had decided yesterday that it was going to be my own Fun Day. No one else was implicated. No one else was really invited. I was going through the toronto.com site for general event listings. I didn\'t want to see a movie. Decidedly not fun enough. The idea of going to see live theatre sent a jolt of utter distain through me, mostly because I\'d been an actor. But bands... interesting... I could see a live band. So I\'m scrolling down and I notice playing at Lee\'s is The Walkmen.   Oh... my... God!  I took the subway down to Lee\'s. It was such a marvellous journey. There was a woman @ my 11:00, black, 250lbs, head held between two hands, either chanting a prayer or crying to herself. Not softly. To my 2:00 there were two devotees of Beyonce, both black, both pretty, both shaking their necks trying to outdo the other with stories of how stupid boys are, sucking their teeth but resigning themselves to calling said boys as soon as they got home, away from each other\'s judgements. Looking back at my 11:00 the older woman had released her head to show me the full cataract in her right eye. It was ice blue, bulging, determined, like a steel marble. It was extraordinary housed in her face. To my 9:00 there was an empty paper take-out coffee cup rolling in an arc along the floor. To my 3:00 there was a Polish looking gent, in his 50\'s carrying a wrapped gift bag. He was lovingly fluffing the tissue paper adorning it. There was a woman in her late 20\'s in a cheap track outfit, cheap sandals, one polyester leg crossed over the other. One loose thread dangled from her pant leg. There was a young man, scruffy, would be seriously hot, to me, if he shaved and got a haircut. With the garb and the hair he seemed to be going for that Jesus look. It\'s a look. I met eyes with him for maybe three seconds. It\'s an eternity in a subway car. He spent the entire ride looking at my reflection through the window next to him. All of these incidents. All of these images are so incredibly dear and beautiful to me. And they happen all the time to everyone, but nobody really sees them.   The Walkmen are from Washington, D.C. A four piece outfit, solid band. Very tight in recordings. Less tight live. The singer is the weakest link. And I think he knows it. Visually he has a Chris Martin (Coldplay) look about him. Tall, lightly complected, pinstriped suit jacket, casual underneath, sings with all his might. But that\'s it. He\'s not really saying anything as much as he\'s screaming a lot, hyper-developing his massive jugular vein. He comes from the Bob Dylan meets Sepulchura school of vocal projections.   When he walked on stage the mike, which was previously set to \"shoe-gazing\" standards, was switched to Kareem-Abdul Jabar heights. This meant straining high into the mike while looking down at his audience. To him we were not his supporters. We were his enemies. He looked at us with both fear and loathing at the same time. One is often linked to the other. His band was so good, but his singing style doesn\'t allow me to play the album as much as I\'d like. His lyrics are alright, but this protective stance of screaming what it is that needs to be said, leaves me a bit cold and a bit frightened. Still, I lost my mind during The Rat.  The opening band was called The Uncut. Toronto band. I wasn\'t going to see them. I was planning to time it so I\'d arrive mid set. Saw the whole thing. Shy. Shoe-gazers. Sweet. Really tight. Really good. I like when opening acts dazzle us, the audience. It\'s happened quite a few times in my life. Each time it happens I\'m secretly thrilled. More votes for the underdog. They have a decidedly new New Order sound about them. Guitar and rhythm section deliciously rich and satisfying, not unlike Interpol; vocals merely a layer, understated, not emphasised, not enhanced, actually downplayed. Very Canadian. Please, God, make them successful. They\'re a nice bunch of lads.  Watching these two bands, wonderful each in their own ways, reminded me of a story on the CBC\'s Definitely Not The Opera. Canada\'s style vs. that of the U.S. The interviewee was saying that when one has a conversation with a Canadian the way things are phrased there is a lilt in the sentence, as if when giving a statement the statement is still up for discussion. It\'s more a questioning and from a musing perspective. With Americans they tend to state their \"fact\" and it\'s stated plainly, bluntly and not open for discussion. It just is. I was telling Jeremy about this and he asked if I was American. He thinks he\'s very funny.   There\'s nothing like seeing live music sometimes. I used to hate it, complaining about the *other* people. Secretly I\'d once wished the bands I loved would just do a special show just for me. After going to a lot of pubs and listening to decent enough house bands, I realised there is a lot of pressure felt to respond in a positive manner to the musicians involved. Eventually I embraced my concert-going brothers and sisters mostly because it took the pressure off maintaining my enthusiasm for the band. Now I can allow the music to fully penetrate me, make my cranium, chest, and pant legs vibrate madly and just enjoy the moment. I can scream as loud as I want because the bodies absorb the sound. I can jump up and down, laughing and clapping my hands. I can just beam throughout while still maintaining my own anonymity, which is important to me as I\'ve discovered. I like when no one looks at me sometimes.   Music is like good sex to me. Totally fulfilling, exhilerating, evocative of so much. Good live music, with an excellent sound system, is like being picked up, thrown down and being fucked hard. In the *best* way. And when it\'s that good you clammer for more and more and more...  '),(11848,4075808,'07,August,2004','2004-08-07 00:00:00','   I don\'t recall the last time I\'d taken an entire summer off from anything, either school or work. Maybe when I was eight. I guess the next time I did so was this summer. And it\'s been the best thing I could have ever done for myself.   It didn\'t start off great. I was lying in shavasana, or corpse pose, in my early yoga practice when I decided, or rather my body decided I could no longer work where I was working. So I quit. Yoga was the beginning of all the awareness I\'ve been having this summer. It\'s been a beautiful and wild adventure.   I am a believer of serendipidous acts. One thing leads to another. Answers are out there if you\'re looking for them and you pay attention to the signs.  I\'ve made it my business to care about things my whole life. But when I cared too much it would adversely affect my life. If things didn\'t turn out the way I wanted it would rupture something inside me or it felt not unlike a cancer growing inside. It was horrible. The act of caring a bit too much is my vision ends up getting blurred and I have no idea what it was I was looking for in the first place. It\'s kind of akin to being an inexperienced web researcher. You go online and look up something very specific. There are 100\'s of hits. Not all of them have 100% relevance, but you go onto the site anyway. One path takes you to another. So if the search started with the bad effects of fluoride, say, you could by the end of 5 hours have accidentally entered yourself into a wet T-shirt contest in Utah. It\'s hard. Staying focused. For some. For me. So I\'ve decided to let go.   Change is funny in most people. For me it\'s just a decision. I\'ve decided to not care so much, and it sounds weird but it\'s working for me. I\'m happier. I can love my friends more because I expect absolutely nothing from. And I can acknowledge all the love they can muster up for me. It\'s hard for most people to demonstrate love of all denominations. I\'m a bit gifted in that realm. It\'s a good gift. So I think I\'m helping, which is really good. And it feels really good. So that\'s good. Whole lotta good in a paragraph.  '),(11849,4075808,'04,August,2004','2004-08-04 00:00:00','   My yoga instructor Todd asked me to read an article from last Saturday\'s Globe and Mail. It described his friend Pete who died last summer and the circumstances involving the accident. One of his friends wrote the article.  The cause of his horrific demise, a boating accident last summer in the Muskoka\'s. The article was a eulogy, and a plug against bad boating procedures - alcohol, rich kids in their 20\'s staying up for too long and a general lack of care.   Todd told me Pete had a thousand friends.   Wow.   I don\'t have a thousand friends. I find eight are hard to maintain sometimes.  Pete was a \"golden boy\". Born into wealth, developed and honed the lifestyle, had friends in the same snack bracket. He worked on Bay Street. At the age of 27 he had won awards for entrepreneurial \"zeal\". Whatever that means. Why is my gut saying he ripped people off. Oh, the business of usury. Or I could be completely wrong and he was the only man on Bay Street who was the Patron Saint of giving people that don\'t have enough money, more money.  I suspect not, though. No one gets pretigious awards on Bay St. for being fair. They get them for getting as much money as they can get their hands on. Everyone wins. Sometimes.  In his spare time Pete surfed. He and buddies would drive all night to far off US destinations, spend dough on some choice accomodations and... well, surf. Or ski. Or do other activities kids of wealth do.  I didn\'t really feel sorry for Pete. Not really. Yes, he went horrifically. For that I do feel sorry. I feel sorry that Todd lost his best friend. I don\'t know what I\'d do without mine. I think I don\'t feel sorry for him because nothing bad ever happened to him. He probably never broke anyone\'s heart. He was never want for anything. He had an enviable relationship with his immediate family. He had hordes of supporters. All of his friends were good looking. The dude\'s life was a beer commercial. It was picture perfect. It was as if all the things that normally go wrong in other peoples lives were all stored up for him in that one moment when he just got shredded by a boat and motor.   He did live large and that\'s important. I think that\'s important to everyone. Whatever \"large\" means to an individual. What I liked most about the article was that one of his best friends, who also worked on Bay Street, ended up quitting his job and starting a painting business with his own brother, wanting out of the \"rat race\".  Somehow this gives me hope for humanity.  '),(11850,4075808,'03,August,2004','2004-08-03 00:00:00','   My foray into yoga started years ago. Yoga if done properly, and by this I mean diligently, earnestly and with the fullest of breath, is many things. I think two of the best things about yoga is it\'s a barometer for how you\'re *really* feeling and it\'s a massive transformative tool. Starting with Kundalini, I progressed into Ashtanga (for Type A personalities) and now I\'m practicing Bikram.   Bikram yoga involves a series of classic and not so classic yoga postures, but done in a heated room. Through radiant panelling the instructor can control temperature and humidity, with the exception of extremely hot and humid conditions outside that may seep indoors. There are *buckets* of water loss. Through my experience there are kind and gentle instructors who guide classes with love yet authority and there are nasty $#%!s that lead with both stifling heat, extreme humidity and either a dictatorial tone or a stilted lead.   [An aside: This beer I\'m drinking sucks. Carlsberg Light. Thanks Mere.]  Back to Bikram. Nearly each time I finish practice I tend to say, \"That was the best yoga experience of my life!\" There are exceptions to the rule. Please see paragraph describing nasty $#%!s . Tonight\'s class was, I have to say, though fear sounding redundant, the best yoga experience of my life!  What makes yoga so different from one day to the next is directly linked to how different we are from one day to the next. One day we might be upset over something that happened at work. Another day there may have been something said to a friend that could have used a bit more diplomacy. Maybe I had a righteous day. Maybe not. Tonight I forced myself to go. Half of me wanted to, the other half just wanted to surf the net. I could feel my ass flattening, so I packed my bag. Todd was leading the class.   I love Todd.  Todd and Julie are my favourite instructors. My love for them is equal. Todd\'s got this great surfer dude meets radio announcer\'s voice. Unlikely but true. Julie is simply one of the most beautiful people I\'ve ever seen. Both are gentle, knowledgable yet make you challenge yourself. This is a desirable effect.   I\'d been doing research, writing and planning an event throughout the day. All three activities were incredibly grounding. I felt good emotionally, but I really wanted to test my emotional/spiritual barometer by going to yoga. Sometimes we convince ourselves that we\'re fine, but there are issues that crop up every now and then. Doing breathwork really helps get down to the nitty-gritty of what we\'re really feeling.  I believe there is truth in emoting, but where the ultimate truth lies is in the breath. The poses don\'t hurt, either. In the first pose, I can\'t remember what it\'s called. You interlock your fingers, pull them down as far as they\'ll go in front of you then inhale audibly while extending your hands up, over your head, shoulders relaxed. Then you place your hands in prayer position, exhaling with a loud \"haa\" sound until they pass your heart, then back to interlocking the old fingers, back to position 1. Some guy choked on his spit at one time. Somehow I thought this was hilarious. I don\'t know why. I think I just needed to laugh.   Once upon a time I wasn\'t crazy about back bends. They used to hurt. They open the heart shakra, the bits of us that allow all feelings to potentially emerge. You stand on your knees, place your hands, fingers facing down, on the top of your bum, pull up with your chest, stretch your neck up and back, pulling the shoulders back and falling backwards, grabbing your feet. This action tends to allow emotions that we have kept hidden inside to seep out. We tend to feel things either during or directly after pose. There have been times I\'ve sobbed after doing it. Tonight I was beaming.   How could a person whose heart was allegedly broken 3 days ago possibly beam after that? Well, I felt like I hadn\'t smiled for a while. This is unusual for me. I felt lucky. I felt that I took advantage of a situation bravely and with honour. And the most marvellous of all, I felt like I understood what I was supposed to do on this earth. I was simply supposed to love people. I understood at that moment this was my job.  Yoga keeps me honest. Yoga keeps me nice. Yoga manages the yuck in my life and the weird that I sometimes feel. It made me quit my job. It made me take the summer off. It will bring me back to work soon, though. Yoga made me understand fully that the only constant is change. Oh, and change is good.  '),(11851,4075808,'02,August,2004','2004-08-02 00:00:00','   I found this while doing some research on life and love and stuff of general lusciousness. It\'s sooo good.  ~*~ A TIME AND A REASON ~*~   Sometimes people come into your life and you know right away they were meant to be there... to serve some sort of purpose, teach you a lesson or help figure out who you are/want to become. You never know who these people may be but when you lock eyes with them, you know that very moment that they will affect your life in some profound way.  And sometimes things happen to you at a time that may seem horrible, painful and unfair, but in reflection you realize that without overcoming these obstacles you would have never realized your potential, strength, or will power.  Everything happens for a reason. Nothing happens by chance or by means of good luck, illness, injury, love, lost moments of true greatness, all occur to test the limits of your soul. Without these small tests, life would be like a smoothly paved, straight, flat road to nowhere. Safe and comfortable but dull and utterly pointless. The people you meet affect your life. The successes and downfalls that you experience can create who you are, and the bad experiences can be learned from. In fact, they are probably the most poignant and important ones.  If someone hurts you, betrays you or breaks your heart, forgive them because they have helped you learn about trust and the importance of being cautious to whom you open your heart.  If someone loves you, love them back unconditionally, not only because they love you, but because they are teaching you to love and to open your heart, and eyes to little things.  Make every day count, appreciate every moment and take from it everything that you possibly can, for you may never be able to experience it again.  Let yourself fall in love, break free and set your sights high.  Hold your head up because you have every right to. Tell yourself you are a great individual and believe in yourself, for you don\'t, no one else will believe in you.  Create your own life and then go out and live it.  If you take your eyes off your goals, all you see are obstacles.   ~~ Unknown Author  '),(11852,4075808,'02,August,2004','2004-08-02 00:00:00','   Okay, so am I flaky because my heart is no longer broken? See, this harkens a query from a preceding blog which questioned my love for all nine of the men I claimed to love. I\'m filled with questions today.   Did he come along to ease my pain from missing David so much? Was Jeremy there in the month of July when the seers thought David would come back, designed to make me consider and reevaluate exactly what it was I needed this time, instead of constantly succumbing to the needs and wants of the men, as I traditionally do? As Psychic Kim said, \"He\'s coming back, but I\'m not sure whether you\'ll want him again.\" Yeah. Me neither. I did listen to the tape today though.  The last time I saw Psychic Kim I\'d brought in a tape recorder. (I\'d much rather have a DAT, but it works, so why allow something to A) unnecessarily collect dust or B) contribute to landfill issues.) I now have this tape that has two different, yet very similar readings on it performed by two psychics. They all say he\'s coming back. I think Jeremy came by during the time where I\'d reached the end of my tether and where I\'d missed intimacy so much that every fiber of my being was screaming out for the kind of attention I\'d needed for the past two years plus.  He raised the bar on the standards I\'d loosely set when considering a potential union of two. I tend to solely consider the individual and the direct correlation he has with my heart. The rest of the stuff is irrelevant, really. I mean, I don\'t care what he does for a living, if he has a car, if he has a house, stuff in general. Never really mattered to me. I liked his careful, though. I think I liked that the best. That and he was honest.   I was not careful. Careful\'s not really an adjective used to describe me. I need to feel stuff all the time. I don\'t care what it is: tragic, euphoric, angry, and (sweetly rare) jealousy. I was never too concerned that I cried. Crying is truth for me. It\'s an opportunity to examine things and to know that I care deeply about things. About people.  Sweetly rare jealousy. Hmm. David was the only person that ever, in any recent memory, inspired jealousy in me. I never felt that while with Mike, husband of 7 1/2 years. Never. I don\'t think a little jealousy is a bad thing, though. A bit of it means they mean something to you. The possibility of them being taken away is potentially devastating. Maybe that\'s what I need. Dunno.  Maybe the universe has some grand design for me. It inherently knows I\'m terrible with remembering lessons that were learned too easily. So, it threw in a little havoc. I was merely an outpatient in the cardiac ward. All the aortas still functioning. Just a little angina, or indigestion in the grand scheme of heart break. Just so sweet, the pain. I wrote to my friend Meredith, \"Every now and then it\'s good to be a bit shattered. It puts this delicious tragedy in your life, which in the end enriches it.   I think I freak people out.   Sometimes I freak me out, in retrospect. But then I\'ll read a magazine article while having lunch and there\'s something in there that will absolutely speak to my soul and cause me to put my fork down, though my mouth is open, it cannot receive food.   From Richard Wollheim\'s Germ: A Memoir of Childhood, an excerpt found in the August 2004 issue of Harper\'s Magazine:  \"...women could love, they could fall in love, they could be in love, they could be lovesick. They could *feel*. Sometimes, after a man and woman who had come down to the house for lunch had driven off, one of my parents would say, not exactly to the other, for that was not how they talked, but more into the surrounding air: \'Why does she go on doing it? What does she get out of it? When will she settle down? Why is she throwing away the best years of her life?\' If only, I would feel, these questions had been asked of me, I, though not able to put it into words, would have had much to say. I would have begun by saying that these were women, something that my father had never been and perhaps something that my mother had forgotten how to be, and I would then have gone on to say that, for women, for some women at least, love, love in itself, love unrequited, love that did not even seek for anything in return, in other words the pure culture of love, could be a way of life.\"  And then I smile to myself,  an inward and outward smile,  singular,  alone  but not lonely  and I feel good.  '),(11853,4075808,'02,August,2004','2004-08-02 00:00:00','   I want to get this while I have it fresh on my plate.   I had made a resolve to fight for him. I\'d felt like he\'d been let down a great deal in life by others and I felt it my responsibility to care for him because I loved him. But the love for him was the kind that was sweet, familiar, deep but more in a special, special friend way.   I remember reading his profile and thinking, \"Why don\'t I already know this person? He seems like someone I\'d already be friends with.\" But I wasn\'t yet and then I became one. And then he was wonderful, as they tend to be in the beginning, but in the end it was disappointing. As my friend Meredith said, \"Best it happened now versus six months from now.\" Apparently there is a direct mathematical  formula which describes the amount of time needed to heal a broken heart vs. the time spent loving someone. It had only been just over two weeks after all.  Ah, I\'m reminded of Chet Baker right now. Chet sang, \"I fall in love too easily. I fall in love too fast. I fall in love too terribly hard for love to ever last.\" That seems to be a recurring problem in my life. Oh, and fucking Chet killed himself. Would that there was something that I could sustain for a while that was almost tepid. If not tepid, then maybe a slow trickle. I don\'t know what I\'m racing for. I have no idea what my hurry is. It just happens. I get so caught up in the passion of the moment and expect that everyone is exactly the same as me, feeling exactly what it is I\'m feeling. But they\'re not, evidentally. So here I am. No real mission again. Alas. I wish I knew exactly why. But maybe I don\'t really want to know. A little mystery is good.  I think the mission is to concentrate on what it is I need to do, to be.   I\'m so close.  I found myself going through the same pulls as I did before. I had decided not long ago that the next time was going to be different. It wasn\'t all that different. I did spend more time with friends this time, which was good. But the rest of it, the all consuming bits were still fragments of my old self. Can you help those, though? Is there that much of yourself that you can change? I want to know.   Maybe I have a love addiction. Geez. That was hard to write. I don\'t remember feeling that gross angst before, but they tend not to break up with me. It\'s usually the other way around. And maybe it\'s the preparedness of a situation. He had prepared himself for this at least a couple of days in advance. I had absolutely no preparation at all. Just typically flying by the seat of my pants. Again. Sigh.   The worst part of it is I can\'t listen to half of my music library right now without thinking of him.   The best part is he\'s had the foresight I\'d always lacked. In the end he was right.   I don\'t really have that list of requirements that need to be filled when looking for someone. It\'s just a feeling. A connection. It\'s always different, ever unique. I have to remember the time spent with him. It\'s very important that I remember the time. He was very special and I have to be able to have that kind of special the next time around. He gave me hope that something really good could exist. I have to know that, to remember that.   So there were good things that came out of this: 1. I got to really care about someone again; something I love to do. 2. I now have to examine whether this is a serious problem in my life, the whole love addiction thing. 3. I know who my friends are. 4. We were very good friends and eventually we will be great friends. Just not yet. Still some more healing to do.  '),(11854,4075808,'02,August,2004','2004-08-02 00:00:00','   I don\'t know where it starts. Maybe at birth. The pecking order. The roles we end up playing throughout our lives. I was the last child born in a family of four children.   In my family as well as others the eldest child was expected to lead the others, to be an example. The second child if, God help him/her is born the same sex as the eldest that child will always live in the shadow his predecessor. I was the baby. They don\'t really expect much from the baby, especially when the baby laughed all the time and developed a special knack in the dispensing of wedgies.   I came from a strangely traditional Chinese family. I say strangely because both parents came from China, both from rural areas, at a fairly young age. They were in their late teens to early twenties. A new land. No sense of real belonging, only a need to assimilate and not draw too much attention to themselves. They were, not fully understanding the culture of their heritage, not fully integrating into Canadian culture either, effectively lost between two worlds.   They started having children right away. Dad didn\'t really wait the appropriate amount of time before impregnating Mom again. This happened 3x in a row. Siblings Vince, Walter and Eunice were born in consecutive years starting from 1960.   Eunice was the youngest and the only girl for six years of her life. And she REALLY liked it. I was born in 1968. I was a \"mistake\". There\'s a different kind of attention paid to girls by the father. There are less expectations, or just expectations in a different realm. More of chastity and familial responsibiltity, a care for the elderly agreement in blood sort.   Growing up in that time, in a fairly affluent neighbourhood, meant instilling the children with the iterated and reiterated idea that study was imperitive to success. Study led to the attainment of money and goods. Money = Goods = Success. As hard as they would study he would make them study harder because of their ethnicity. He convinced them that the world would not view their abilities and aptitudes based on their competence alone. The world would view the slant in their eyes and their general facial features first, relegating them to more subserviant positions based primarily on racial considerations.   I never got \"the talk\".   Competition became a primary characteristic among my siblings. Dad encouraged them to compete against each other. He\'d often pit one sibling against the other, as if it was some morbid test for his love. And they were willing participants. At first they were scholastic competitions, then it trickled down to sports, then within childhood friendships where he would ask some poor unsuspecting neighbourhood kid who that kid liked better, this one or that one?   His own children.   Choose.   Eventually the competition seeped into their adult lives where the test would include how big their house was, in what neighbourhood, what kind of car was each driving, if you had a cottage, how many degrees were under your belt, how many tries it took you to get your driver\'s license. It was never the quality of your friendships that mattered, it was the quantity and how full your activities schedule was. How busy you were keeping yourself. Idle hands, the devil\'s playground.  All four of us are very different in personality. Vince is an accountant and it was as if he found some manual explaining how clichÃ©d accountants behaved and he became that; think Eddie Murphy doing the White Guy skit from old Saturday Night Live episodes: uptight super-white guy, stiff, pickle firmly up ass, expressionless and cold. Publicly the prankster, laughing and smiling if he can make a buck off you. Privately silent.  Walter, super athlete and will tell anyone who has ears all about it, tendency toward violence, but is controllable. Funny, but prone to telling the same stories from his \"glory years\" in high school. Childlike. Was the apple of my eye growing up. Father thought all he would ever amount to was being a blue collar worker in a factory setting. Publicly the joker, privately the monster.  Eunice has been accused by my motherr of having \"eyes on her forehead\" because she looks down on people. Refuses to drive her car through certain neighbourhoods for fear of being robbed or raped. To her the robbing would probably be worse. Scholastically accomplished. Dreamed of being a professional student. Always publicly appropriate, privately caustic.  I am considered the \"black sheep\" of the family. Laughed louder than others felt was appropriate. Had a \"talk\" by my sister about not embarrassing her at her wedding. Feels shame too often. Is working on that. Is the \"pot-stirrer\" of the family. Truth speaker. Publicly obnoxious, but loving. Privately often unsatisfied.   In common we share the spirit of survival, showmanship, loyalty to our friends, competency, a profound difficulty to rely on others. We all have perfection issues within ourselves and often with others, a generous sprinkling of the gripping fear of looking ridiculous and are painfully aware of what others think of us.   Separately we have our own assigned roles. We all know that if Vince is confronted he will back down and relent, feeling embarrassed by other\'s outbursts, mostly because his own emotions embarrass himself. We all know that Walter will break something; a door or a heart. Currently he\'s estranged himself from the entire family. We know Eunice will deflect, spinning the argument in a totally different direction, leading the instigator to a confused state. She will likely try to be the victor at any cost. Myself? I used to cry but don\'t anymore. Not around them. They took too much away from me. Or I gave too much away at well below manufacturers list prices. After careful examination through therapy of many different varieties, excellent friendships and time I still feel the pull toward shame, but I hope I attack the issues as they are presented. It is my hope.  We still have funny pulls and triggers though.   The other night having cocktails, Jeremy and I were discussing movie going characteristics. I tend to see movies alone for a variety of reasons. He talks during movies, which I admittedly am not crazy about. I find it disruptive to my surroundings and wish not to do that to other people mostly because I don\'t like hearing whispers and muttering to my left (my good ear). It\'s louder to me than the movie is. (Apparently a classic ADD characteristic). He then said something about how loud I laughed and didn\'t I think that was disruptive to others? He meant nothing by this comment. I lost my mind. All the feelings of their attempts to minimise me came to the surface. [Tell her to be quiet. Tell her to be less, less than she is. She\'s not good enough. We just need to make slight alterations.]   We grow up and think we\'re grown-ups now. How could we possibly have the same pulls we did when we were kids. These same triggers. Lots of people have commented on the fact that my laugh is approximately 160 decibels (130 is the average pain threshold). But at that moment, maybe because he\'d just dumped me I was feeling particularly vulnerable and extremely self-protective. At that exact moment all I felt was you cannot make comments or opine negatively about things a person cannot change. I suppose I could laugh softer. But I don\'t. And I won\'t. Anyway, he loved my laugh. I know this. Why did I take it that way? Fucking triggers in our most insecure moments. Will they ever die?   Along with these triggers, we have this innate sense of our duties and responsibilites. Our roles in society.   Eunice is the self-professed, self-imposed \"Do-The-Right-Thing\" girl. She always brings the appropriate hostess gift, she is highly skilled at small talk, she \"presents\" well; charming, effusive in the expected parts ie. nearly clawing her eyes out in utter amazement when the neighbouring Rosedale hostess at a dinner function has made, \"by hand\", ONE course in the evening\'s meal. She\'s got her own money and she rarely misses a day using it as a weapon against her husband, threatening to leave because she can \"afford\" to, taking her overcommitted 3-year-old twins with her, thus abandoning him.   I hate what money does to people. I hate what money did to my family. I want to deal in the currency of people. I want to make people think, to make them feel it\'s okay to feel what they\'re feeling. It\'s honesty, something that this society lacks so often.   Our roles as people are as unique as we as individuals are. There are caretakers, caregivers, nurturers, peace-makers, war-mongers, mother-in-laws (of the icky variety), hosts and hellspawns. To be brave is to take risks, risks out of one\'s own character. We should strive to do things that don\'t necessarily make us feel comfortable or come to us as easily as the stuff that was formulated and honed early on. Things that scare us a little... or a lot. It\'s liberating. I want to experience as much as I can so I can share my experiences with others. We are all linked universally. We\'ve all felt loss. We\'ve all felt the diverse, often debilitating, feelings of pain and suffering. And I hope upon hopes that we\'ve all felt love.   I believe we all have a hidden resolve, a secret special superpower that lurks within all of us. The power that enables us to be most effective and do the most good. Maybe finding it and tapping into it is a goal we should always strive for, otherwise what\'s the point of living if we\'ve got nothing to fight for. And why fight if we have nothing to believe in.   Seek your battle, but use your heart.  '),(11855,4075808,'01,August,2004','2004-08-01 00:00:00','   How do you deal with a broken heart? Is there some special formula, or are we reduced to just filling ourselves with loosely tested, overly prescribed blue pills, red pills? I think I\'ll forgo that, choosing to feel the genuine pain of love lost rather than the numbness of not feeling anything. Or worse, throwing myself into work, or just generally keeping busy, as they prescribe. \"Are you keeping busy, dear?\" Jam packing my itinerary has never been a solution to solving life\'s \"little\" problems. It sure takes our minds off things for a while, though. But, eventually we have to deal with things. Things like:   The two things I thought would never happen again: 1) I never thought I\'d fall in love again. 2) I certainly didn\'t think I\'d get my heart broken again.   But I guess that\'s the chance you take when you risk everything. It is the ultimate exposÃ©, the final showdown, the opportunity to show everything you have in your secret cache and in oneâ€™s individual Pandora\'s Box.   All these people drinking lover\'s spit.  				                                               Water, water everywhere and not a drop to drink.  As scared as I was to say it, in the end I did and I\'m glad I did because I meant it and he really meant something very significant to me.   I love you.  And if it was said on the last night we were potentially to be together, that\'s alright, I can live with that. I know I did the best I could do and was the best I could be, given the circumstances of profound bad timing.    How can you compete with a ghost? No, she\'s not dead, but they spent 10 years together, during the very formative decade between the ages 20 - 30. She\'s moved on. A different life. Abandoning him. He the one who loved her so much. Maybe not romantically after all of those years, but certainly a familial love, that binding love, that blinding love that was ripped away from him before it was time. Before he was ready.   I\'d said before how new love stirs things up. I counted this morning how many men I\'d claimed to love in my relative short existence. Nine. How many did I truly love? Would walk to the ends of the earth for? Would sacrifice life and limb for? Not all. So maybe it wasnâ€™t true love. That\'s always been a quest for me. True love and the possibility of its existence. Far more than seeing if God really exists.  I need to believe it does.  I\'m puffy from crying, relenting finally this morning to remove the remains of the rivulet roadmap of mascara down my cheeks. A poor version of Alice Cooper. A sadder version, far less evil. I feel like I\'m close to exhausting my friend pool for emotional support. I need guidance. Comfort. He needs the same. A sage that is empathetic, someone who can tell him he\'s not alone and what he\'s feeling is totally normal and in at least one time in our pathetic lives we\'ve been hurt so badly that we can\'t distinguish waking life from a dream state, a surreality that doesn\'t make us want to feel anything, especially the pain we\'re feeling right now.   We search in vain for clarity. We need concise answers for the things we feel because they don\'t make any sense to us at all. \"If only I could get my shit together....\" We live in this world where there are always solutions to problems, or at least they work the bugs out the beta version with version 10.97. Things are packaged neatly in eye-pleasing colours and boxes. The whodunnit\'s villian eloquently and correctly accused discovered in the third act by the elegant detective with the Belgian accent, twirling his handlebar moustache, tapping his silver handled cane on the the parquet floor. But unfortunately the ways of the world donâ€™t necessarily work that way.  At one time in my not so recent past, I wanted answers so desperately I went to see psychics (plural).  Fuck, I can still smell him.  Anyway, back to the psychics. I wanted to know about a certain fella Iâ€™d been with a few years back, then rekindled something several months ago. They all promised he would come back. And I was so patient. One of them prophecised there would be a choice Iâ€™d have to make, between him and another man. In the end she said Iâ€™d choose the one in first chronological order. But then I met Jeremy. Was he the other Iâ€™d have to choose from? Two days ago there was no choice. After being with Jeremy he really raised the bar in standards certification. The psychics didnâ€™t tell me anything about him. He was a surprise. I think in my path I became happier not really knowing. The worst thing about being told your â€œfutureâ€ is once you know about it you have this tendency to wait around for the future to happen. â€œIs it going to be today,â€ you ask yourself for the 14th consecutive time that day. â€œHow about now?â€  What is certain is uncertainty. What dispells fear is the choice to love again. And here on just that topic: the righteous chicks, the Bene Gesserit, from the movie Dune, delivering a little litany against fear:   I must not fear.  Fear is the mind-killer.  Fear is the little-death that brings total obliteration.  I will face my fear.  I will permit it to pass over me and through me.  And when it has gone past I will turn the inner eye to see its path.  Where the fear has gone there will be nothing.  Only I will remain.Â    And Iâ€™m not talking perfect love, because again, perfection is unachievable and even if it wasnâ€™t it would be pretty damned boring to have to face day after day. â€œOh, hereâ€™s comes Perfect, again.â€ Tedious. Predictable. Weâ€™d resent it. Him.  Yes, I got hurt but thatâ€™s okay. I learned so much about myself during the entire, albeit brief process of loving someone new. And I got to love someone again. Someone really great. So, I consider myself lucky. [She said through gritted teeth as 1st runner-up, allowed to perform pageant-of-life duties only if the winner was unable].  '),(11856,4274229,'21,August,2004','2004-08-21 00:00:00','   sunny.      dear diary..   just came back from gb and lunchh wif the crappie family.oh yarhs.haben told u abt them.another time?wahahas.aniways.2dae GB.haiz.kinda sad larhs-(*fer the drill part*) at first, shu juan and shu zhen and lihong took us fer drills. their command was ralie soooooooo soft lorhs.so couldnt hear.den when we doing \'kepelakang pusing\' on the march, half of the squad turned while the other half marched on.sianz Â½. realie cannort hear marhs! den eileen kinda came and socld us. wth. realie so soft lorhs!! pissed sia.urgh.nvm.later, i was asked to take the sec 2 and 3s fer drills.den ms sherilyn want me to make them du sizing.wth?i dun even noe the command very well... so guilty sia. kip letting them \'samula\'.firstly i wasnt familiar wif the command.2ndly they kinda sucked at it larhs.so haf to kip repeating.sumore ms sherilyn was there to watch lorhs.wth!!stressed up sia.so during waterbreak i apologized to them.kinda cried a little larhs.wahahas.stressed.but i realie love my gurls sia.they even consoled me nehs.said it was excellent fer the first try.-(*so touched.sob sob.=\"))hahas.love them-(*sorie fer repeating, but i realie love moi gurls.wahahas*)by the end of the sesion, ms sherilyn seemed quite happy with our drills.they did improve alot btw.wahahas.so proud if them.oh well.i\'m gonna master the command well sia.gonna show them!!hahas.uh-oh.guess i gotta get going lerhs.still meeting sean and daryl at church. going east coast to book bbq pit fer cell bonding.fill u in wif the second part of the gb meeting later 2nite okie okies? well, cyaz!      with lotsa love, krissy  '),(11857,4274229,'20,August,2004','2004-08-20 00:00:00','   very very early.      dear diary...   it\'s 5am in the morning? hahas. i still cant sleep nehs.on da fone wif kenny.he\'s like, so uncle.wahahas!kinda generation gap nehs.but still,learn lotsa things from him. we talked abt the camp and stuff.bitching abt the bitchy camp leaders.wahahas.it\'s gonna be dawn soon.wonder if i can get to watch the sun rise.hmm.it\'s a beginning of a new dae!a new life.hmm.nida go fer gb soon.i cant wait!! =)   P.S.- qi li xiang is still playing..muahahas~      lotsa love, krissy  '),(11858,4274229,'20,August,2004','2004-08-20 00:00:00','   rainy.      dear diary...   back from testi spree.wrote fer 17 peeps nehs! power hors. wahahas.dunno y nehs. cannort sleep.still listening to qi li xiang. just hung up with wen. toked to her fer, like from 12-2.wahahas.had crazee plan to meet up at 5am tml to go fer chee cheong fun.but no transport sia.urgh.singapore\'s transport system soo lazy de nehs.5.30am den start bus service.eeyers.oh yarhs. to all those hu dunno, wen is short fer wenyi hu is moi bestest fren.hahas.we siao siao de. sumtimes beri beri gud, sumtimes biting each other heads off. well, guess it\'s one of the gud daes 2dae. hmm. guessed she missed me larhs. 3 daes nv see me lerhs. datz y soo gud to me... muahaha! btw, she\'s moi 2nd lao po too. cool eh?love her to pieces. hahas. crap. anyways. i haf a ninth lao gong lerhs!! it\'s shooter nehs.hahas.cool wors.still bugging him fer testi.wahaha~ k larhs. crap until here. qi li xiang is still playing sia... wootz.cheerios~   lotsa love, krissy  '),(11859,4274229,'20,August,2004','2004-08-20 00:00:00','   sunny.      dear diary..   i\'m officially in love wif Jay\'s Qi Li Xiang.it\'s soo nice.the lyrics rawks too. check them out~         ä¸ƒé‡Œé¦™ Qi Li Xiang'),(11860,4274229,'19,August,2004','2004-08-19 00:00:00','   starry night.      dear diary...   i\'ve had enuff.sumtimes i feel that i am not hu i am, that the real me is hidden beneath layers and layers of mask.everyday, i plastered a smile on my face, to show the rest of the world that i am living well.what sia.ever since fridae.i\'ve been trying not to think of jon.it\'s over.and that\'s the way i wannit, isnt it? *shrugs*. if that\'s the case, den y m i feeling wad i\'m feeling now? so what if he likes lala now, like 2 daes after the breakup? wtf. he\'s not worth it. but.. i cant help feeling the way i\'m feeling. mebbe a good cry is just what i need. or mebbe a stroll along the beach.a wash from the waves.a lost of memory... or mebbe,i shd make him lose his memory, eh? oh well.whatever it takes.i will survive.no matter how long.how many tears.i will survive.heys!i\'m moving on.time to get real and see wad a jerk he realie is.oh well.better now den later.i still cant believe i actually liked him b4.wth.awwright.everybody makes mistakes i guess?farnie how i didnt learn my lesson after the youth-camp episode.love realie is blind huh?so when the love fade away,do everything thus becomes imperfect?then,why love?why go through so much only to get hurt in the end? wait.. am i hurting? y am i hurting? i wasnt suppose to care! wth. i\'m movin on and i dun care animore. wish me luck.   P.S.: to everyone out there.never ever promise aniting if you are not sure u can do it. never say forever.. cos nth is forever.below is one of moi favourite poem. makes hell lot of sense to me. especially the last few lines. read on and think about it.'),(11861,4274229,'18,August,2004','2004-08-18 00:00:00','   moonlit night.      dear diary...   i\'m sick!! coughing moi lungs out sia.the medicine, like, sucks. but nvm larhs. got 3 daes MC. cool eh.die lerhs. the audition\'s tml. hope i\'ll be okie wors. feeling rotten about everything. and the medicine seemed to make me even more sick. urgh. pray fer me everyone. i would very much like to survive this. =) the medicine\'s taking effect. i nida sleep!! cheerios.      with lotsa love, krissy  '),(11862,4274229,'16,August,2004','2004-08-16 00:00:00','   cloudless night.      dear diary...      finally got back to school lerhs!hahas.missed moi classmates a great deal.(*honestly!*) hmm.nth much had changed, apart from the new classroom venue and the sitting arrangement.hmm.nth much larhs.oh yarhs! they had this new addiction: copying bei hun\'s version of \'soreii\'. haha. cool eh~hmm. well.there\'s lotsa homeworks piled up fer the last week. guess i nida employ sumone to du it fer me lerhs. ani volunteers? hahas.i\'m dwn wif a bad cough too.urgh.hope i\'ll wun fall sick wors.well well well.guess i better go do the overwhelming pile of homework lerhs.teachers can be so ruthless sumtimes.oh well.cyaz!      lotsa love, krissy  '),(11863,4274229,'15,August,2004','2004-08-15 00:00:00','   starry night.      dear diary..   just came home after the \"date\" wif alvin.it was soooo cool!!i totally luuurve fireworks man!!hahas. awwrite. it\'ll start from the beginning. hmm. we went to watch \"The Village\". the show was sooooooo dumb. i was like,\'duh.\' hahas. nevermind. it was nice company tho. *wink* hmm. after that, alvin was victimised to be yet another victim of moi neo print mania. hahas. the pics came out realie fantastic.kinda couple-ly tho.hahas.hmm.yeps.and the highlight - firework display! it was wayyy cool! soo romantic-(*minus the crowd and the \'ahhh\' and \'ooohh\'s from the aunties -.-\"*)hahas. hmm. it\'s realie indescribable. it just. rocks. hahas.realie kept my mind off things while watching the beautiful display unfold in front of moi eyes. woah. i swear i\'m gonna set off fireworks durin my wedding man. haha~ oh well. nvm. ahhh. i love fireworks. the beauty of it totally covers up the not so nice things about life.-(*like, u noe, some jerks.*)hmm. mebbe there is realie more to life than jerks, i reckon. sigh. wish i could just capture those moments.hmmm.    P.S.: well, if you are wondering abt me and alvin... hahas. guess on!!      with lotsa love, krissy  '),(11864,4274229,'14,August,2004','2004-08-14 00:00:00','   sunny      dear diary...       just cum back from youth service, wearing that toopid blouse and skirt and courtshoe!! haha. we had a excellent time of praise and worship-(*kinda quiet tho,cause joel wasnt there to play the drums. wonder where he he went*) anyways. 2dae during Words, that uncle preached us on the 3rd purpose in life. that is, to find our purpose in life. -(*stop rollin ur eyes.*) well.it seems that in life, we have our   own purpose-(*like stardeeing.becuming a millionaire.that kinda stuff.*) and of cos, God\'s purpose for us. More often than not, we forsaked God\'s purpose for us and pursue our own purpose in life. at the end of the day, we had what we wanted -- excellent grades, lotsa money, blah blah. BUT. remove the material achievements and we\'ll be left with nothing. scary eh? well den. we\'ll have to live our life to fulfill God\'s purpose in our lives: to be more like Him. of cos, it is impossible to be LIKE HIM. i mean, we are NOT GOD. but. we can live our life like him. we cant be perfect, but we can try to. obviously, we cannot achieve this alone. we\'ll have to get help from the Holy Spirt. invit Him to remind us what to do when we haf to make a choice, and do it in a way that is most pleasing to God.hmm. so from now on, i\'m gonna live to fulfill God\'s purpose in moi life. i\'ll learn to be more Christ-like. so help me along, everyone, to become a better person. =) Jesus rocks.      with lotsa love, krissy  '),(11865,4274229,'14,August,2004','2004-08-14 00:00:00','   setting sun.      dear diary.      we had our first gb meeting since we came back from camp.and it was our turn to take parade!hahas.i was the flag-raiser.cool eh?kinda lost moi voice to all the cheering during camps.so today\'s timing kinda \"chao-xia\" =x. well. took the sec2 and 3 fer drills. can see they are all trying very hard. especially the sec 3s. just cum back from camp nia. well well well. press on peeps! wanna tell the sec2 and 3s a big big SORIE fer being so fierce 2ade. kinda pissed by all the dragging sounds. but nevertheless, u guys still rocks! worship was great too! thanks a whole bunch delia, fer leading us into worshipping God!-(*i give my life to the Potter\'s hand*) hmm.yeps. Girls\' Brigade rawks man! gurl power!! girls, let us all shine in the name of the Lord!!   P.S.:Ching - u did a great job commanding!! jia you jia you jia you!!      with lotsa love, krissy  '),(11866,4274229,'13,August,2004','2004-08-13 00:00:00','   moonless night.      dear diary...      remember my first entry, when i thought that i have landed myself in the hugh bunch of mess? well guess what. life just seemed alot less complicated now. me and jonathan - officially over. just a simple message fer a clean breakup. oh well. at least i\'m not the bad guy this time. just as well. i was wondering how to tell him to back off anyways. hey wait. am i sounding bitter here? hahaha. i\'m not! i\'m so glad to be free again. and guess wad? i eralie dun believe a word he sae abt having to prepare fer \'O\'s. i mean, if he cared, we wouldnt have cum this far.well. i thought he should at least haf the guts to tell me that he\'ve met sumone new, dun u reckon? wad a coward. sumtimes, frens realie see more than u du. i shd have just listened to ching. find a better guy. but nevermind. i\'m moving on. bye jerk! harlo life! haha. oh btw. haf i mentioned to u before that i haf this tiny crush on a camp leader? hahas. well well well. guess on!!    P.S.: about the other 2 guys in moi 1st entry. i have come to a conclusion, that the thing with moi lao-po/gong is realie just a crazy infautation and well, i\'m moving on with no turning back!!    with lotsa love, krissy  '),(11867,4274229,'08,August,2004','2004-08-08 00:00:00','   starry starry night.      dear diary...      i\'m back!!hahas.i miss singapore soooo much!!i can finally finally have a nice long a hot bath without worries of the reporting time and the sudden termination of water suppy when u\'ve got soap all over your body!-(*i\'m not kiddin. it happened.*)well well well.soo much had happened during NDP and the staying over at hq and of cos, the camp!! dunno where to start filling u in with all the juicy details!!hmm. let\'s start with the NDP bahs..      -=NDP.fever=-   after 4 months of training!! it\'s all gone in a matter of mins!!hahas i\'ll miss NDP siaz. all the gurls.-(*we rawk!*)uncles-(uncle wong.uncle ong.uncle yong.uncle issac.uncle alvin.aloy!)yeps.and the slim water.pizza.ice-creams.and jj!!hahas he was right behind me and i couldnt turn ard and look!!wth??hahas.it\'s gonna be the most memorable ndp fer the rest of moi life.thanks to all the peeps hu made it possible.miss u guys!i will nv nv ferget NDP 2004.c=      -=slumber.party.at.HQ=-   haha.it wasnt meant be a party, if datz wad u are thinking.it was suppose to be a briefing at hq, den staying over so as not to be late for the ferry to ubin the next morning.well well well.i had soo much fun wif the delta gurls!!-(*hoi ki.cassan.me.elisa.weili.francel.*)we like, bonded the moment we stepped into our bunk!!hahas.it was a night full of luffter and noise by me and hoi ki, and abundance of \"advice\" pended out by cassan.i\'ll always remember the \"glow in the dark\" man!!wahahah.we realie shd do tis again, dun u reckon?hahas~      -=national.camp.2004=-   man.man.man.this must be the best camp, EVER.learnt lotsa new stuff, cheers.met lotsa new peeps.had lotsa fun!totally love delta man.we like, rawk!hahas.injured moiself the beri 1st dae at camp during mountain biking.got a bruise on moi face sia!hahas.hmm.second dae.had jetty jump.kayaking.boat rowing.and the elements.very very fun, and bonded realie well wif moi campmates and leaders alike.in the 3rd dae.we had campfire and mass dance!-(*usher rawks forver!*)it\'s time to PARTYY~yeps.but the scary part:a gurl in our division had fits.it was so horrifying and i realie dunno wad to do except to pray pray and pray.so, cassan, francel and me were like, praying and praying.woah.so after awhile, that gurl kinda calmed dwn, and a police car came to take her away.THERE IS A GOD!hahas.but still.the night was young and the rest of the deltas continued wif our little chatroom until the  urlLink #$@&$@&amp;$  camp people shooed us to bed.like,duh?at the last dae.-(*phew.finally*)hmm.it was kinda sad larhs.missed the camp sia.but then, all gud things will come to an end.hope the deltas will still stay in contact!!DELTA forever!!  '),(11868,4274229,'08,August,2004','2004-08-08 00:00:00','   starless night.      dear diary...   it was fun day in church. ironically, it wasnt fun fer me. at all. first of all, i was grp leader fer pei xuan, hu main objective in life is to make MY life living hell. then i was assigned to peel onions in the kitchen. like, wth? den. during my supposedly break time, i kindly went to help out at the balloon stall, and ended up having to put up with that sickening face of eve. harlo? i\'m being kind here awwright? duh. well. the best part of the entire carnival-(* beside being able to see HIMs, that is *)was the balloon though.-(* i mean after their duty when i dun haf to see eve\'s sickening face animore. *) i had 4 pink and 2 silver balloons!! cool huh? walked with the balloon tied to my watch all ard kovan. attracted weird stares i must say. oh well. bold to be different.-(* wink *) oh yarhs.and that toopid freaky bitch of a fren. been giving me that sickening face of hers the whole day. can i just gif her a bloody slap arcoss that pig face of hers that she so damn deserve?urgh. heys. guess i better get going. gonna pack moi camp stuff first. OH BTW. i\'m gonna be awae fer the entire week, in deserted pulau ubin fer national camp. remember to miss me yarhs?      P.S: bought a pink sarong skirt 2dae. so nice!       love,krissy  '),(11869,4274229,'07,August,2004','2004-08-07 00:00:00','   setting sun.      dear diary...       it\'s been an action packed day. woah. had dance pract in the morning. badminton. swimming. got myself a tan. hahas. did sth realie meaningful today. while walking with ching and cel and ljy and jowin and szeling and wen to da sportshall from kovan 2dae, ching spotted a unit with their household flag hung the wrong way-(* meaning,the cresent and the 5 stars are at the right side of the flag instead of the left side.wth??? *) thus, being the nation lovers that we are-(* awwright.it\'s my idea. but hey! it\'s respect fer da country okies?haha *), we went up to the door and wanted to \"confront\" whoever it wad to hang their flag the right way. well, nobody answered the door. duh. but, u reckon i\'ll just back off liddat? NO WAY! well. i took out a piece of paper, and wrote \"dear resident of this particular household: if you havent noticed, you\'ve hung ur national flag the wrong way. plaese correct it. thankiews.\" if that was not enuff! i even drew a \"correct\' version of the flag below!! woah. is that cool or what? hahaha. well. if u think i\'m nuts, go ahead! all i noe is, i\'ve done a great deed! haha. SINAGPORE RAWK DA HOUSE!      love,krissy  '),(11870,4274229,'06,August,2004','2004-08-06 00:00:00','   starry night.      dear diary...           sigh. i\'m. so. confused. why cant i leave my life like a normal teenage gurl? or du everyone else haf the own problems as well? why am i me? haix. i landed myself tangled in a huge bunch of mess, and there seemed no way to get out of it without hurting anione. i miss him. i realie do. but. there\'s so much uncertains. so much doubts. no matter how i tried to, i cant figure out what he\'s thinking. all the memories..they were never gone. apples. the beach. playground. basketball court. delibrate walking pass of classroom. 24/7 irc...too much, and too important to be erased. but. is that what will become of me? nothing but an empty shell that is living in sweet memories? hais. and him. my babe. how am i suppose to tell him, i dun see what he sees? how am i suppose to tell him, i\'m nort ready fer another relationship? how am i suppose to tell him, that we should realie just be frens when he had already thought we were a pair? and then there is moi lao po. or lao gong. or whatever. it\'s realie nice toking to him. keep my mind off things. and he\'s so. true. no pretence. and i dun haf to put up the false pretence, either. so. will we be more then frens in the near future?or. mebbe. i dunno. m i just imagining it?mebbe it\'s just a crazy infautation. i\'ll get over it in a minute. i hope. God above. pls answer my prayer and tell me wad to do. i\'m lost in tis huge, complicating mess. or mebbe..i should just lead a simple,SINGLE life.i dunno. sigh. well. live well everybody.                         love, krissy  '),(11871,4007735,'30,July,2004','2004-07-30 00:00:00','    Today was not as bad as&nbsp;i thought it would be, in fact it wasn\'t bad at all! not even the dentist poking around at my teeth with sharp objects and whatnot! lol no really, work was good cuz Bryant was working today and i hadn\'t seen him since school ended and he made work so much more fun, so yea that was good. oh yea during work i had like 3 cups of coffee loaded with like 5 spoonfuls of sugar each so i was high from that. got home, showered, blah blah blah, went to&nbsp;caseys&nbsp;(after passing by on my way to pick up my lil sis and him insisting i come over when i got back) and hung out with Max, his friend Katelyn, Nina, some gay kid whos still \"in the closet\"&nbsp;and apparently goes to our school? , and oh yea, Gian(fuckhead)Carlo. man that kids a fuck. yes he is  A  fuck. lol then Bob showed up, he\'s the fucking man! we chilled then casey was like, you guys have to leave soon cuz im going away for 2 days, so yea basically i headed over to Christina\'s!&nbsp;heh it was funny cuz i guess she had just woken up when i had called to tell her i was on my way over and she was all sleepy and it was funny, not to mention cute&nbsp;too =P . so we rented  Gothika  and can i just say,  goddamn  that movie is some  freaky ass shit! &nbsp; no lie! needless&nbsp;to say&nbsp;i was freaked out! lol yea yea i\'m a puss sometimes but hey c\'mon, we all have our moments =P . and with that i bid you  adieu ! (heh thats french!)     hanging with you is always a good time, it\'s nice to have&nbsp;a friend&nbsp;that always brings a smile to my face, so&nbsp;I just wanted to say,   thank you.'),(11872,4007735,'28,July,2004','2004-07-28 00:00:00','    Not&nbsp;feeling too great at the moment. I\'m not exactly sure why. Ever get that heavy feeling in your chest? As though something is pulling down on your heart, making it feel heavy as stone? It\'s weird how it just kinda hit me. I dunno, whatever. Tomoro I have to wake up early, goto the dentist, then go straight to work. It\'ll be a blah day to say the least but then again the past couple of days have been blah days I suppose so whatever. I\'m just looking forward to this week to be over with already! This coming weekend and the next week should be much much better.&nbsp; And then the week&nbsp;after that I leave for Paris! So yea that\'s gunna be good. Listening to the song \"Pig\" by DMB (of course), makes me feel a bit better. It\'s great how a single song can change your entire mood in a matter of minutes.      Don\'t burn the day away'),(11873,4007735,'27,July,2004','2004-07-27 00:00:00','    Listened to this and it brought me to tears, so beautiful, touches that certain place in my heart and mind which brings out these deep emotions. Don\'t know why, but the music just has that effect on me. So  beautiful ...'),(11874,4007735,'27,July,2004','2004-07-27 00:00:00','    The most intense display of emotions I\'ve&nbsp;ever heard in a song before. Dave wrote this song after asking his longtime girlfriend to marry him 3 times&nbsp;and she turned him down all 3 times.&nbsp; This set of lyrics is from a live show at \"The Gorge\" (2002), the best live performance of the song ever recorded.   &nbsp;  &nbsp;'),(11875,4007735,'26,July,2004','2004-07-26 00:00:00','    eyes closed.     listening.sadness.     longing.hoping.    dreams.wish.'),(11876,4007735,'25,July,2004','2004-07-25 00:00:00','    Note to self : Stop beginning entries with \"Yea\" and stop saying \"Well\"!  cold pizza is teh roxorz.   '),(11877,4007735,'24,July,2004','2004-07-24 00:00:00','    Yea, well just got home a lil while ago from Megs. Nothing all that exciting I guess. Had a drink but thats bout it. I\'d rather get trashed with people I actually enjoy hanging out with, not kids such as Jian(fuckhead)Carlo and others, so I\'m not disappointed. Max and Meg had their daily \"i-wanna-hook-up-with-you-but-dont-wanna-date-you\" talks where meg ends up crying her eyes out, and ohyea she was trashed at that point so it was even worse. Oh yea and then she decides to tell Max that she\'s given head to her neighbor (who is COMPLETE white TRASH) twice and shes only known him a week. it\'s like, \"WHAT THE FUCK is wrong with ya!\" cmon have some respect for the ones you love. She needs to grow up and stop being a bitch like that and Max needs to learn to stay away from her for a while until things cool down. He just keeps hurting both her and himself when he sees her again and does the things he does. I suppose one day he\'ll learn but whatever, thats his deal. And can i just say i really fucking HATE Jian-Carlo? He\'s such a goddamn cocky narcisistic asshole. I\'m not even gunna get into why, but we\'ll just leave it at the fact that,yes,he is. Okay well that ends this rant. I feel a bit better now! =P   '),(11878,4007735,'23,July,2004','2004-07-23 00:00:00','    Yea, so i woke up at 1:00 today! woohoo! I can\'t say there\'s anything better than waking up, looking at the clock, and realizing you just slept for like 12 freaking hours. heh. although part of me is like, \"wow you idiot, you just threw away half your day!\" *SMACK!* but hey, it\'s all good cause I have no big plans for today anyway. Meg Walsh, Max\'s ex-gf is having a party cause she has an empty house till tomoro (her parents went away), and some kids will be bringing some \"beverages\" (vodka, coconut rum, bacardi raz/cranberry juice, beer!) but I prolly won\'t be drinking seeing as I can\'t sleep over since I have work in the morning/I have to drive. Meh, it might be entertaining nonetheless, Max and I will prolly stop by for a lil bit in any case. Seems like most everybody I wanna hang out with is going away this coming week. GRRR. And then there\'s the  certain people  I almost never get to talk to! *sigh* Time to go scrounge for food. Lates                   -Brendan Michael   '),(11879,4007735,'23,July,2004','2004-07-23 00:00:00','    Well, I\'ve finally decided to make one of these things, mainly out of boredom. See i used to keep an actual written journal for like 2 years but then i kinda grew away from that, but i must say, within the pages of that journal are some of my deepest, most&nbsp;revealing thoughts and emotions. Of course they\'re a&nbsp;bit irrelevant at this point seeing as they were simply emotions of the moment and nothing more. yet when i look back at those pages all those emotions come flooding back, if only for a brief moment.&nbsp; weird how that works, and then i end up thinking of how stupid/silly i was 2-3 years ago. oy i\'m rambling, but hey! this is MY journal thingy and i can rant if i want to! so yea! there! alright, well there shall be more to come, until then...  -Brendan Michael   '),(11880,4007735,'12,August,2004','2004-08-12 00:00:00','    August 13th - August 22nd!!!      OH HELL YES!   '),(11881,4007735,'04,August,2004','2004-08-04 00:00:00','    Had such an awesome, kickass day at 6 flags today with britt,vicka,jay, and Victoryflet! Yea Victoryflet aka Fletty aka Eric is a kid me and Vicka met through the clan we\'re in and whom we play with in CS. He\'s totally cool and it was awesome to meet him! Got a little sick by the end of the night after riding every single roller coaster like 5 times each. Superman = the best roller coaster in new england. such a rush! but yea, we spent literally ALL day and night there! we got there at 1030 am and left at 1030 pm! We did all the crazy rides first, basically the coasters, then we left the park and got some pizza, then came back and did the waterpark for a good 2-3 hours and that was mad fun cuz it was soo goddamn hot out! then we went back to the rollercoasters and whatnot. it was such a great day! 6 flags owns! i\'m ready to pass out now and i have work tomoro which bites but its at 1130 so thats better than anything earlier rite? rite. Orange carnival begins tomoro so thats cool too. i\'ll prolly goto the movies with morgan seeing as the other night wen we attempted to attend said event, i got us really lost to say the least, not gunna even attempt to explain HOW lost i got us but lets just say, i was ashamed at my navigations skills that nite. so yea, prolly do like a matinee after work so that should be good. see morgan is totally kickass and chill and i think she mite have a crush on me? she asked me one time if i\'d ever hook up with her or date her. now morgan is a great girl but 1) i know and play hockey with her brother, and dating the younger sister of ANYBODY you\'re friends with is just plain wrong, it\'d be like if one of my friends dated mariel. thatd be fucking fucked up! and 2) she\'s just not somebody i\'d ever do that with. and so i told her these things and she was fine with it. and so i think we\'ve begun a really good friendship and thats always a good thing. what\'s funny about this situation is that i\'m better friends with morgan than i am with chris. but hey, what does it matter? alright, i\'m done for the night cuz im tired as hell. lates.   -B'),(11882,4007735,'02,August,2004','2004-08-02 00:00:00','    RAWRRRRRRR!!'),(11883,4007735,'02,August,2004','2004-08-02 00:00:00','    The day is done I\'m on my back, staring up at the ceiling,  i take a drink, sit back, relax, smoke my mind, make me feel better for a small time ...'),(11884,3771600,'28,June,2004','2004-06-28 00:00:00','   I know that there\'s a biblical reason for fornication and all the complications like STDs and how condoms don\'t give you the 100% protection. Pre-marital sex has all these and more I know, but the question that I asked was why does it have to be such a taboo? Why does it have to be such a taboo until you cannot talk about it openly? What happened to proper sex education? Let people do what they want about their lives-whether or not they want to unwrap that special present is their business not ours.  About homosexuality, I think that they are just people confused about their sexuality, but that does not give us a right to ostracise them, does it? They probably just need more time to really find out bout themselves but we, instead of giving them advice or a helping hand, treat them like sub-humans. I don\'t want to give you guys the wrong impressoion about my parents-they don\'t encourage pre-marital sex nor do they oppose it either. They strongly oppose sleeping around and one-night stands- believe me, they do.   '),(11885,3771600,'27,June,2004','2004-06-27 00:00:00','   I was talking to somebody yesterday about the cultural differences between Western culture and Eastern culture. Firstly, about pre-marital sex..My parents have been pretty open about this topic, so I knew about it when I was quite young- around 7 maybe. I\'ve been given free rein to do whatever I want with my life but anything that I do, I must accept the consequences- good or bad. In eastern culture, it is normal to be a virgin or lose your virginity only when you\'re married. I was brought up differently, you can have pre-marital sex BUT you must know what you are getting into and what the possible consequences are..Why does the eastern culture regard sex as a taboo?? That I don\'t understand. There was one columnist in the STAR newspaper who said that her child thought that if you kiss the opposite sex, you can get a baby. How ridiculous is that? I mean, I\'m fine with the \"don\'t have sex till u\'re married\" thing but why does it have to be such a taboo until kids don\'t get their proper sex education? I acknowledge the fact that pre-marital sex can lead to unplanned pregnancies and other messy situations.. But why make it such a taboo? Now, about homosexuality. I have a couple of homosexual friends who tell me their problems- our society\'s view of them, their family\'s views and acceptance. They are just normal human beings who breathe, eat, and drink but they just are attracted to members of their gender. I don\'t see why it\'s so hard for people to accept them. They are just humans. And yet we treat them like outcasts. With all that said, I\'ve got to go now..bye! p.s.: I\'d really appreciate feedback! Thanks!   '),(11886,3771600,'27,June,2004','2004-06-27 00:00:00','   HI! ok,this is the first time I\'m actually having a blog- I used to have one but I was too lazy to update it ^_^; currently I\'m at home now with nothing to do so I decided to start one at blogspot since almost everybody I know has a blog like Jenna, Aku, and Randy..Ok,that\'s all for now.   '),(11887,4202671,'23,August,2004','2004-08-23 00:00:00','   \".... time has  gone  by  real  fast...'),(11888,4202671,'22,August,2004','2004-08-22 00:00:00','   haiz... today  bad  day ar... first maths tests... like shit like that haiz... actually, standard la. i always  fail  maths hehe.. den physics..  argh!!  come out   so   many on kinematics and i din study it! stupid. we keep doing dynamics and momentum,  den suddenly , a   hel  l lot kinematics come out! damn! they want me to  retain  ar!?  but anyway, it is also a good thing. like luke chen say last time nvr come out kinematics, so now come out a lot. final year should come out lesser le.  hope  got more dynamics questions come out. but then, still got electricity.. hope it is easy...  anyway, after tests, came home.. rest. tired so got some shut eye. after that was making up my mind whether to go to edge. wanted to go, but changed my mind eventually... after that, i went to play comp for quite some time. survival project, again.. hehe... so fun. i think i addicted liaoz..  ok. after playing went to watch some tv, took a shower, did some weight lifting and study. so like damn boring hor?? i think it is quite boring..  should  have gone down to the edge.  anyway, gonna continue studying. later 2.30  am  got rugby. want to watch... '),(11889,4202671,'20,August,2004','2004-08-20 00:00:00','   common tests have  begin !! sianz..... it is 10.30 now.. and i am going to play some games!!! hahahha....  today\'s test was quite easy la.. hope i can do well... short day, nice and so cool. haha.. so early at home, like i juz  woke  up. hehe...  now i\'m going to  play  survival project. later study. oh ya! tonite got net...  got visitors ... nvr had visitors for a long time. hope they dun back out.  \"Dear God,'),(11890,4202671,'19,August,2004','2004-08-19 00:00:00','   hi evryone!! now i juz finished playing survival project. it is real cool man!!! should go check out that game.  so erm.. today i stayed back as usual. but tonight i did some work!! haha.. *shocking* hahahaha.... today was nice... hanging out with the usual bunch, chatting... haha...  oh ya!! i am going for the leadership camp! cool huh! cant wait to go.. Lilo\'s going too. yeah! it\'s gonna be fun.  juz wanna thank wenhong and christine for dropping by my blog... come to think of it, i kinda miss u guys a lot... i miss how kin wai always make fun of me... boonchye\'s stupid laugh, LOH\'s company... weijing\'s funny stuff... all u guys, i miss u!!! i wish i can turn back time and re-live those good old days... miss da teachers too...   anway, life goes on... sad things has passed and now happy things are all around me.. this is to u esther...'),(11891,4202671,'18,August,2004','2004-08-18 00:00:00','   hmm.. the  drama  today was nice   but  .... it reminds me of  sad  things... makes me feel so  hurt .... i dun wanna be reminded.. juz wanna be me. made me feel so sad... but i enjoyed the show.. :)  so erm... went home early again from night study.. cant study la me.. i am cut out to be an entertainer,  not  a study kia....  hahahahahaha.....  wad u guys think? * hee *  today... another nice and peaceful day lol... cant remember much. i oni keep disturbing  miss lee  and the other teacher, trying to make them   blush  .   wuahahahahahahahaha  ...  mean  ya!? hee.. miss lee\'s so  cute.   blush  so easily.  *LOL*  oh ya! guys hu did not go night study today   missed   out man!!! today all the  chio bu  teachers stay back hehe....  i think nothing else le la.... so cya! '),(11892,4202671,'16,August,2004','2004-08-16 00:00:00','   hey pple  !! yesterday could not write cos internet down. today i manage to solve the problem haha...  anyway, i juz came back from night study... like usual, i din do much again.. juz   keep chatting  . haha..   sorry   guys keep disturbing u all... hee... i cant study la.. not my way of enjoying life haha....  so erm... today... i cant remember much leh.. but quite an ok day la. oh ya. i  passed  everything of my NAPFA. so next PE lesson, can join the rest do the dance le... oh ya.. sam, i still dunnoe y u say  she\'s  rude.. y ar? tell me wad  she  do leh...    today kenna disturb by all the year one ruggers...   stupid   bunch. wait till  i  think of something. heh.... cant wait to start playing rugby again.. wanna play so badly...  k la. stop here le. these few days not enuff sleep. today betta get some.. '),(11893,4202671,'16,August,2004','2004-08-16 00:00:00','   after some rest, i went out to watch tv... when i got bored of channel i, i switched it over to disney channel for some cartoons... juz nice,  Lilo & Stitch  the movie had juz started... nvr watch it b4..  it is sucha touching show... made me think back about some  stuff ... how each other  should not  be  forgotten  and  left behind ...  \"O\'hana (if it is spelt this way)'),(11894,4202671,'15,August,2004','2004-08-15 00:00:00','   finally i have a more decent looking blog... heh... took me up till 2 am last night to do it. i figured out that the html commands for blog is more or less similar to OD.. haha... cool. now i juz have to try to find a way to add the background picture.. cant remember the command ar... haiz...  anyway, i juz came back from church. straight away log in to my blog to adore it. lol... (all my hardwork....) i think it is going to be a boring day from now till tml morning.. so gonna catch up on some sleep and do some workout.  so anyway, that\'s all for now. '),(11895,4202671,'15,August,2004','2004-08-15 00:00:00','   ok... i juz came back rom the bbq... quite fun and   hot.   haha.... so anyway, when i came back, i decided to go to kelvin\'s blog... go check it out :  urlLink http://yingmujielun.blogdrive.com   haha.. when i wanted to go to his blog, he  warned  me first cos it was gonna be very   shocking   and   indeed   it was! haha.... check it it out.. gonna  hook  u guys up with a   preview  . *lol*     Directly from the creator - Kelvin Wang'),(11896,4202671,'14,August,2004','2004-08-14 00:00:00','   haiz... today\'s  Nationals  sucked  ....so bad! man.. we  beat HCJC  which sent SR and YJ  trembling  but not RJ. sadly, we lost to RJ. wad was worst was that  we lost to SR and YJ!!!!!!!!!!   argh!!!!!!!!  i had confidence we could beat them... we   could   have been in the semi finals... though i dun think we\'ll get third(cos SA, AC and RJ will get top 3) but we can at least be fourth than  nothing!!     but anyway,  good game guys.  i only played the RJ game... it is like  defense  all the way... they are very good. SA and AC\'s match was more like a  real  rugby game. very nice to see em play... i wished i could\'ve play the YJ match... wanna let out some  fury  and give\'em a  bashing .. hands itching for a   faceoff  ... haiz.. nvm got next time friendly.  gonna work hard. i wanna play in more matches.. anyway, i\'m really tired now so i am gonna rest update again later. tonight got bbq again.. eternity production thingy.. haha... so come back later and check out my blog see ya! '),(11897,4202671,'13,August,2004','2004-08-13 00:00:00','   oh no.... my  BIG  toe... how??   today was such a dumb day.. so sad for me... i had to run 2.4 but... i passed!! zaki din have to cos we were going for nationals... but rozlan look down on me... anyway, with my toe, i dun think i can play already.... sobx... den still had to have training. (the day b4 the match!?) anyway, i got my  toe  injured, my  jaws  injured and i had double vision for a moment... haiz... how?? now that my toe is swelling... muthu play so hard during training dunnoe for wad... shit man... this  sux!!!!!   bashed  me and stepped n my  toe  and din say  sorry .. *crap!*  so frustraing... i hope it heals.. now i am puttng ice while writing this...    \"dear God.... pls heal my toe... i have been training so hard....'),(11898,4202671,'12,August,2004','2004-08-12 00:00:00','   haha.. hey akira, miyane! juz now during night study,  millie  asked about the  lilo  thingy, we should have juz stopped at telling her about the  M  and the  E  haha... den she tot is  her . lol... so anyway, i did nothing much during night study.. juz did some of the questions miss lee asked the other pple to do for discussion. read some chem stuff n... that\'s all haha...  ok.... today is juz another nice day.. one funny thing was the  teapot  thingy and shana\'z... thanks to  josh, akira  n of course   SHANA\'z  , i sang the ending of this silly song...(after keeping so quiet for so long during chem.. dunnoe wad happened to me) muz tel sam the next time i see her.. haha... juz a rough idea.  \'Little teapot\' (censored vers.)'),(11899,4202671,'10,August,2004','2004-08-10 00:00:00','   ok. my first entry.. ah.... y i started a blog? cos no one wants to use OD anymore... but anyway, today was quite a fun day. had a lot of fun playing a fool during chemistry. hahahaha. cos i saw miss lee scolding those with their filter funnels on their burette, i came up with this idea to sabo pple by putting the funnels on their burettes. wuahahahahaa.. i asked akira want to go do a not he say ok so he went to put the funnel on shaf\'s burette den kenna caught by her. for me, i went to put it on clinton\'s and vijay\'s burette. they din noe a single thing lor.. hahaha.. tio miss lee. hahahaha.... but i got go say sorry la hahaha.. sorry guys!  oh ya.. u guys noe mary? the servant. she is so fun to talk to...... or shold i say disturb. whenever i saw her today, i kept disturbing her.. hahaahahaha.. she thinks everyone lkes her!!!! *shocking sia* so bhb.. lol... even my chinese teacher had nothing to say about it... aha..  hmmm... oh ya. heard about akira\'s \'goodness resists all common evil\'?? haha.. cool huh. wad about me?? mine too long... haiz.. but.... guess wad!? i manage to come up with something too.. it goes like tis: \'Made In China Huh\' hahaha.. lame eh.. but the name\'s too long la... heee... but those hu noe about this, keep it a secret k? those dunnoe, come to me personally and i\'ll tell u. and dun leave posts about the made in china thingy thanks.  haha.. first entry and so much to say. ok stop here. take care u guys, especially sam! cya! '),(11900,3966870,'22,July,2004','2004-07-22 00:00:00','   Finally! I have moved most of everything over to blogger! There are some new features here. I have of course the desktop cam! But I have removed the TV cam for now and I have added a WebCam! Now you can take a little more depth look into my&nbsp;world..&nbsp;Which is most cases isn\'t much LOL I am still adding some features to the blog. I have some more links to put up and some of my favorite sites I visit as well. &nbsp; I will be also be putting together a Linux help section for tips and tricks for n00bies and some server stuff too. I will also be adding some Modeling graphics sometime in the future as I get more into that. I will also be posting all of the great art work I make I will be getting a deviantart account so I can put everything in one place.   I guess that\'s it for now! I am sure I will be adding tons of content to the site once I remember what it is ;) '),(11901,3966870,'23,July,2004','2004-07-23 00:00:00','   I will be offline for most of the day tomorrow I will be getting my trucks AC fixed and then I am going to be reformating windows and installing slackware 10! So look for updates from me later tomorrow night. If anyone knows of a webcam software for linux let me know! or a desktop cam program too! e-mail me admin@fatk1d.net    Well I am off to watch some tv and eat some ice cream and chill with my wife!     l8r '),(11902,3966870,'25,July,2004','2004-07-25 00:00:00','   Well my format and reinstall and install of slackware didn\'t happen, I will waiting till sometime next week to do it. I have to burn some stuff off of my drive before formatting. I am really torn between slackware 10 and SUSE 9.1 I will hopefully make my choice before I format. I am also going to be only updating my blog at night as I hope to soon have a job working for Dell in the service department here in the new place they opened up here. So wish me luck!    \"Dell Service Center\"  \"My Dell blow up\"    \"Dude you should have BUILT THE DAMN THING YOURSELF!\" lol    laaaatttteee '),(11903,3966870,'08,August,2004','2004-08-08 00:00:00','   I want to say that playing Doom 3 really lound on a bad ass pair of headphones should be ILLEGAL! It will scare the living FUCK out of you. It plays with your mind just like its real and it gets to you inside and makes you just want to stand up and scream MAKE IT STOP GOD MAKE IT STOP. I haven\'t tried it with total darkness and headphones but if I do. Please someone bring me a new pair of pants because DAMN if Doom 3 is fucking one scary as MOFO  GAME!  AHHHHHHH I LOVE IT! THANK YOU ID! '),(11904,2952444,'23,March,2004','2004-03-23 00:00:00','   Two little wolves                                                                            once upon a time there lived 2 wolves who had moved out of their home to make there own home. the little wolves had a disscusion on how to build their own home. the first one said he would build his out of chocolate. And the secound said he would build his out of gingerbread. And with no more to say than that they parted.  The first wolf finished building his house. It was now a wonderful gingerbread house. He admired the outside for a while until he decided to go inside and admire the house. As he was admiring his house from the inside a knock at the door was heard. It was a pig and he wanted in. â€œlet me in, let me inâ€ said the pig. â€œNot by the hairs of my big chin chin.â€  Replied the wolf. â€œThen Iâ€™ll eat and Iâ€™ll eat and ill eat your house down.â€ replied the pig. And within 2 seconds the house was devoured along with the wolf.    Then without hearing a thing, the second wolf finished building his chocolate house. He also admired his hose for a very long time before going inside, and just like the first wolf there was a knock at the door. And without a doubt it was the pig and he wanted in. â€œlet me in, let me in â€œsaid the pig.   â€œNot by the hairs of my big chin chin.â€ Replied the second wolf. â€œThen Iâ€™ll eat and Iâ€™ll eat and Iâ€™ll eat your house downâ€. And just like before within 2 seconds the house was devoured, along with this wolf to.  About 1 hr later the fat pig was found dead on the ground obviously his appetite didnâ€™t agree with him   The end  '),(11905,2952444,'02,March,2004','2004-03-02 00:00:00','   Red death                                                                              report  Far away in a little country in Jamaica a whole castle of men and women have died from the disease red death. The castle was built far away from civilisation and surrounded by walls and then locked up. The key was thrown away just so red death wouldnâ€™t get in but red death got in anyway and trapped every one else in there. Red death killed 22 actors, 29 chefs, 30 cooks, 100 civilians, 20 rich people, the queen and Prince Prospero. Police discovered this misfortune early yesterday morning when a erldely civilian that lived in the area was walking past, noticed it was quitter than normally. Doctors arrived after the police had got in and noticed that the daises that killed all of these people was a disease called red death. News reporters say that if they werenâ€™t locked up in the castle, and they were able to get out then they probably would of lived or at least saved a few people. Police are saying now that there is know one to blame except prince Prospero. Red death has now killed over 60,000 people. And red death is now one of the most deadly diseases a person is able to get. So if you know any one living near or has red death please warn them about it. And do not lock yourself in a building that is near red death. To tell if you have red death look for red blood pouring out from places like the face, legs, arms, eyes and feet etc. it takes 10 minutes at least to kick in so beware and stay away from red death or you could be or end up like one of the 203 killed in prince Prosperoâ€™s castle. End. -reporter Aaron ziino.  '),(11906,2952444,'30,May,2004','2004-05-30 00:00:00','   The Schoolyard bully  So there I was on my way to school hoping I wasnâ€™t seen by Bob. Bob was a menace. Every where he went be it, The school yard, the street, the shopping centre or even the park he started a fight. And he was only 12. I think he had a fight with nearly every one in town. For no particular reason either. Mainly it was just because he felt like a fight. But the worst thing about Bob was that he never lost a fight and that was really bad for the people that he decided to pick on twice like me Jon Jones. So any way back to my story.  There I was walking to school at 8:00, Monday morning hoping I didnâ€™t get seen by Bob but as usual there was Bob. Oh no I thought he was coming my way. â€œOh Hi Bobâ€ I said. â€œGive me your lunch money Squirtâ€ replied Bob. â€œNoâ€ I insisted, but it didnâ€™t matter, Within 6 seconds I was flat on my face and I didnâ€™t have any lunch money left either. I thought to my self ever day after my bashings why does Bob always pick on me. I wish I was big and strong like Bob sometimes.  The next day I was walking to school again same time same way and same amount of lunch money in my pocket as yesterday .But for some time Bob didnâ€™t come to day. I wounder why I think to my self I guess I am lucky in a sense. But just as I think I am safe here comes Bob right on time.  For some reason Bob just walked straight past me. I was so confused at this point of time. Then I saw Bobs mum walking with him as well and from that day on I never got bullied again and Bob he never bullied again because now Bobs mum walks with him every were.  The end  '),(11907,2952444,'25,May,2004','2004-05-25 00:00:00','   Dear Mr Brackes                                                                20-5-2004   I think the main problems with homeless people are that the institutions are too big, and crowded. They have over 35 odd people in itâ€™s just is not home. I think the way to bring a solution to this problem is to reuse old building and make them in to flats/units for homeless and make each unit with two people in it of their choice. They need to make each unit really homey with stoves to make  there own food, food they are acutely interested in frigs to store there on food in, a phones to call friends and family even there own bathrooms so they donâ€™t have to share and it should cost from $80 a week to $100 a week. In these buildings in two or three units there should be staff, doctors and nurses. Also you should have a suggestion box for the homeless so they can have a say. So they can suggest thing s like couches, TVâ€™s, frigs beds and things like that just to make it a little homier. That way they will feel a lot happier and a lot happier about coming into an institution. Thatâ€™s all but I think it will make a real difference.   Because at the moment these institutions are really not home and we really need to do something about it.    Yours sincerely       Aaron Ziino Latrobe secondary school  '),(11908,2952444,'16,June,2004','2004-06-16 00:00:00','   Insane News  Professionals every were but in my hair  BOB  WANNABIE  Yesterday afternoon Aaron Ziino professional five time world champion made a special appearance yesterday down in torque beach just west of Melbourne in Australia. He was there also with TV star/ sports world surfer, josh young.   He was there for a photo shoot for the magazine â€œworld water sportsâ€. He was posing with his flame board shorts from his main sponsors for clothing billabong and was also there with his six speed lighting bolt Yamaha jet ski.  The set was set up of a beach, the palm tree was on the sand and the jet ski was sitting next to it on the sand. And in the back ground there was the torque beach water and the beautiful bright sun in the back ground.  The expected price for a shoot like this is a round $10,000. but it didnâ€™t matter to the photographers because they just had to get in a photo before Aaron went on his next world tour.  I\'m reporter bob wannabieï¿¼    Thank you.   RACING FAST AS YOU CAN  James evenloderberrys   Just the new jet powered 4x4 Holden sedan v6 was brought out as a concept. The car was just yesterday tested out on a run way and reached a top speed of 1092km an hour. This is expected to be a hit with Holden fans.  '),(11909,2952444,'26,July,2004','2004-07-26 00:00:00','   Letter to mr white  Did you know that even the a.c.e.r is against giving homework. I know that there are many good reasons for giving us homework like parents expectations and getting in to rutin and things like that but what good does it do us if it is not improving our education then itâ€™s a waste of time.&nbsp; It also can cause other problems such as family fights, eating disorders, bullying at school and it can dint your mental health.   Also if you look at the people that are supporting scrap there are some very important people.  Thanks for listening to me Yours sincerely  Aaron 7p  '),(11910,2952444,'14,July,2004','2004-07-14 00:00:00','   So there I was having a nice nap in the warm warmth of the sun when suddenly I hear a crowed of voices, screaming yelling and voices laughing. So I open all of my five eyes to see what the hell is going on and what the hell is disturbing my sleep. And what I see is a class of year sevens. There are all jumping up and down, up and down waiting to get in. Thatâ€™s when I heard one of the worst things a spider like me can hear â€œletâ€™s get that spider and kill it. Letâ€™s take its legs of 1 by 1 ,  when we getâ€ said one of the many kids. I panic. I donâ€™t what to do or where to go but I have to get out of here some how or I am going to be one dead spider. Iâ€™ll be completely pulled a part.  So then I come to my senses and try to think of a way out of the room, I canâ€™t escape through the front door or I be found or even worse squished by all the feet. Thatâ€™s when I see a back door. â€œExcellentâ€ I think. Now I just have to get there and I know just how. Ill will go under the door. But just as I star to crawl the door flies open. â€œOooohhhh noâ€ I get about half way before some one picks me up â€œahhhâ€ there going to kill me. Iâ€™ll be mangulated. So I wriggle and wriggle, and squirm and hiss as hard as I can then they drop me on the floor. I quickly crawl away to the back door and crawl out under the door. Thank god Iâ€™m out of there I think to my self, and thank god im out alive. And that is the story of my great escape.   The end  '),(11911,3737807,'30,June,2004','2004-06-30 00:00:00','   its been almost 4 weeks, since sbihin nia na hindi na nia ko kailangan, that he had found someone better than me...na hindi nia na nga ko mahal, just like that...  ...but until now, hindi ako okay...  ang hirap magpretend na, there\'s nothing going on inside me,  na ang totoo umiiyak pa din ako sa gabi, na iniisip ko pa din kung iniisip nia ko... na sana bumalik na cia...  napilit kong mapaniwala ung iba, nakalimutan kong kumbinsihin ung sarili ko...  sobrang na-mimiss ko na cia... '),(11912,3737807,'24,June,2004','2004-06-24 00:00:00','   i may force myself to forget you and escape i may forget the magic words i might forget to say i love you forget that i did love you i may loose my mind and my heart i might loose myself with confusions and frustrations i might loose you and then find you again i might know or would know,  i love you then and i love you still realizing i remain in love with you i might loose the chance of having you back i might loose the nights and days of being with you i might loose the chance of having what i possessed i might loose you but i want you to know i loved you for a moment i loved you for sometime in my past i loved you for a while i loved you, even if for the meantime, means forever in my heart even if promises turned into harsh words and curses even if all the love i have had and i should share means to be forgotten even if this love was stolen away from me even if love means separation, being brokenhearted, painfully because of you even if love means tears and despair even if love meantâ€¦ â€¦youâ€™ll live in your present, while I still live in our pastâ€¦ '),(11913,3737807,'05,July,2004','2004-07-05 00:00:00','   Donâ€™t worry; this will be the last timeâ€¦  Sending this was never easy for me. Particularly now, knowing that you may not even care what this letter is all about. Knowing that, at this point in time, you may not give a damn regarding what I feel. And it kills me to think that youâ€™re much happier without meâ€¦but Iâ€™ll send it anyway, itâ€™s hard to keep things hanging on, leaving stones unturned, as well as I donâ€™t want regretting â€œwhat ifsâ€ afterwards.  There are so many things that happened without specific reasonsâ€¦ and I presume youâ€™ll be the biggest question unanswered in my life.  As I begin writing this, I look back late last year, when I experienced happiness I never knew would be possible. It was before my second semester, everything was laid back - nothing really to look forward to, until I heard the news about our reunionâ€¦                 I was ecstatic knowing that Iâ€™ll be seeing my long lost friends...     and believe it or notâ€¦specially you.  Before that day, I was so excited to text you after hearing that youâ€™ve been sending messages to ma.fe, even though she told me that youâ€™ve been asking her aboutâ€¦ you know, courting herâ€¦, but what the heck! Reuniting with you even as a former classmate would be fine.  And so there it was, the big day arrived. To tell you honestly, I was so nervous at that timeâ€¦why? I didnâ€™t know either. Iâ€™ve been quiet the whole day, not just because I donâ€™t fell well, but because all these years I never thought the feeling would be the same. However, aware that you like ma.fe, I ignored itâ€¦I just wanted to go back to where I amâ€¦maybe weâ€™re really not meant to be.  But fate took its turn. Everything happened so fast. Surprisingly, Iâ€™ve found myself deeply falling in love with you. Iâ€™ve never been happier. Iâ€™m used to having you beside meâ€¦ There were times that I have so many weights to carry onâ€¦at school, at home, but I donâ€™t worry much, because I know that youâ€™ll always be there. Walang Iwananâ€¦ We always worked things out and facing each problem seemed a lot easier having you around. Oftentimes, iniisip ko, mawala na silang lahat wag ka lang. Even in my wildest dream I never picture losing you, I thought what weâ€™ve been through will be reason enough for us to stay strong and hold onâ€¦ but I guess I was wrong.   Everything seems so perfect, or so I thought it wasâ€¦ I already set my mind and more than willing to spend my lifetime with you...no questions askâ€¦you know that. We used to dream. And I know that visualizing my future with somebody else will not be as ideal if it wonâ€™t be you.  Until that dayâ€¦ I didnâ€™t really see it coming.  I wasnâ€™t, or rather, never prepared for that.  I assumed you loved me that much.  All throughout our relationship, I never thought another girl would ruin everything. Youâ€™ve been telling me before that itâ€™s impossible.  Not to you. Not to us. And not now. I even thought that if we were going to part ways, Iâ€™d be the one to blameâ€¦ You know what?, Iâ€™ve never been so insecure my entire life, after hearing all the things youâ€™ve said. Youâ€™ve turned into someone else just like that.  I still remember that Sunday, our last break-up, you decided to end this because                             sabi mo,â€ nahihirapan ka naâ€ ,I canâ€™t help thinking, bakit ganon? bakit laging ikaw? di mo ba naisip na mas nasasaktan ako?â€¦ Itâ€™s hard to accept that kind of reality..  The pain has its own way of wounding your whole being. Itâ€™s hard to regain yourself â€¦not after taking away something youâ€™ve always asked for. Overreacting as you may think I amâ€¦but who cares?â€¦itâ€™s the truth.  It was never easy for me. I was so helpless.  Agony would be an understatementâ€¦it was hell.  I wasnâ€™t ready to face the truth and give up.  Your memories keep haunting me and moving on seems impossible. For days I keep on cryingâ€¦I keep thinking what went wrong...        Iâ€™m really devastated. I was so scared.  I canâ€™t bear loosing you. I wanted to talk to you personally to patch up things, even as your friend, but you rejected me every time I triedâ€¦ Still, I insist.  Until that incident in SM Manilaâ€¦ It strike me hard, big time! It was an eye-opener for me.             I have to save something for myselfâ€¦ at least my self-respect. And that was more than enough reason to let you goâ€¦sorry it took that longâ€¦ I guess itâ€™s really over thenâ€¦   Now I know that some good things never lastâ€¦  I decided to tell you all this because I donâ€™t want to remain issues unsettled between us. Hating you will not help me forget.           Iâ€™ve realized it. Sure Iâ€™ll miss you. Iâ€™m missing you now, what more tomorrow and for the days to come, but it was so pathetic holding on to something that is not yours anymore. I was stupid to think that youâ€™ll be coming back. But I guess itâ€™s also my fault. I never stop on hoping that youâ€™ll be back to your senses, but maybe it has to be this wayâ€¦  I just hope that sooner or later Iâ€™ll understand it.  Telling you this has been hard for me. And I guess I should prepare myself for more. It would hurt me seeing you with someone else, but I donâ€™t want to be selfish. Iâ€™ll let you grow and find what will really makes you happy. I may or may not regret giving you this letter but I know I would not regret telling you this...                           I loved youâ€¦I loved you even before. More than I expected, more than you knowâ€¦I still consider you one of the best things that ever happened to me. Sana hindi na lang nagbagoâ€¦                          if only I could turn back time...                                                                   Thank you for everything. Thank you for the good memories that are worth remembering.  Iâ€™m sorry for making such a big deal out of this.  I never imagined Iâ€™ll be seeing you again after four long yearsâ€¦whatever the reason, Iâ€™m glad it did.                                  I hope someday weâ€™ll meet again...                                    As they say,only time will tellâ€¦ '),(11914,3737807,'01,July,2004','2004-07-01 00:00:00','   Donâ€™t worry; this will be the last timeâ€¦  Sending this was never easy for me. Particularly now, knowing that you may not even care what this letter is all about. Knowing that, at this point in time, you may not give a damn regarding what I feel. And it kills me to think that youâ€™re much happier without meâ€¦but Iâ€™ll send it anyway, itâ€™s hard to keep things hanging on, leaving stones unturned, as well as I donâ€™t want regretting â€œwhat ifsâ€ afterwards.  There are so many things that happened without specific reasonsâ€¦ and I presume youâ€™ll be the biggest question unanswered in my life.  As I begin writing this, I look back late last year, when I experienced happiness I never knew would be possible. It was before my second semester, everything was laid back - nothing really to look forward to, until I heard the news about our reunionâ€¦                 I was ecstatic knowing that Iâ€™ll be seeing my long lost friends...  and believe it or notâ€¦specially you.  Before that day, I was so excited to text you after hearing that youâ€™ve been sending messages to ma.fe, even though she told me that youâ€™ve been asking her aboutâ€¦ you know, courting herâ€¦, but what the heck! Reuniting with you even as a former classmate would be fine.  And so there it was, the big day arrived. To tell you honestly, I was so nervous at that timeâ€¦why? I didnâ€™t know either. Iâ€™ve been quiet the whole day, not just because I donâ€™t fell well, but because all these years I never thought the feeling would be the same. However, aware that you like ma.fe, I ignored itâ€¦I just wanted to go back to where I amâ€¦maybe weâ€™re really not meant to be.  But fate took its turn. Everything happened so fast. Surprisingly, Iâ€™ve found myself deeply falling in love with you. Iâ€™ve never been happier. Iâ€™m used to having you beside meâ€¦ There were times that I have so many weights to carry onâ€¦at school, at home, but I donâ€™t worry much, because I know that youâ€™ll always be there. Walang Iwananâ€¦ We always worked things out and facing each problem seemed a lot easier having you around. Oftentimes, iniisip ko, mawala na silang lahat wag ka lang. Even in my wildest dream I never picture losing you, I thought what weâ€™ve been through will be reason enough for us to stay strong and hold onâ€¦ but I guess I was wrong.   Everything seems so perfect, or so I thought it wasâ€¦ I already set my mind and more than willing to spend my lifetime with youâ€¦no questions askâ€¦you know that.                                   We used to dream.                                                      And I know that visualizing my future with somebody else will not be as ideal if it wonâ€™t be you.  Until that dayâ€¦ I didnâ€™t really see it coming.  I wasnâ€™t, or rather, never prepared for that.  I assumed you loved me that much.  All throughout our relationship, I never thought another girl would ruin everything. Youâ€™ve been telling me before that itâ€™s impossible.  Not to you. Not to us. And not now. I even thought that if we were going to part ways, Iâ€™d be the one to blameâ€¦ You know what?, Iâ€™ve never been so insecure my entire life, after hearing all the things youâ€™ve said. Youâ€™ve turned into someone else just like that.  I still remember that Sunday, our last break-up, you decided to end this because sabi mo,â€ nahihirapan ka naâ€ ,I canâ€™t help thinking, bakit ganon? bakit laging ikaw? di mo ba naisip na mas nasasaktan ako?â€¦ Itâ€™s hard to accept that kind of reality..  The pain has its own way of wounding your whole being. Itâ€™s hard to regain yourself â€¦not after taking away something youâ€™ve always asked for. Overreacting as you may think I amâ€¦but who cares?â€¦itâ€™s the truth.  It was never easy for me. I was so helpless.  Agony would be an understatementâ€¦it was hell.  I wasnâ€™t ready to face the truth and give up.  Your memories keep haunting me and moving on seems impossible. For days I keep on cryingâ€¦I keep thinking what went wrong...          Iâ€™m really devastated. I was so scared.  I canâ€™t bear loosing you. I wanted to talk to you personally to patch up things, even as your friend, but you rejected me every time I triedâ€¦ Still, I insist.  Until that incident in SM Manilaâ€¦ It strike me hard, big time! It was an eye-opener for me.              I have to save something for myselfâ€¦ at least my self-respect.        And that was more than enough reason to let you goâ€¦                sorry it took that longâ€¦ I guess itâ€™s really over thenâ€¦   Now I know that some good things never lastâ€¦  I decided to tell you all this because I donâ€™t want to remain issues unsettled between us. Hating you will not help me forget.             Iâ€™ve realized it. Sure Iâ€™ll miss you. Iâ€™m missing you now, what more tomorrow and for the days to come, but it was so pathetic holding on to something that is not yours anymore. I was stupid to think that youâ€™ll be coming back. But I guess itâ€™s also my fault. I never stop on hoping that youâ€™ll be back to your senses, but maybe it has to be this wayâ€¦  I just hope that sooner or later Iâ€™ll understand it.  Telling you this has been hard for me. And I guess I should prepare myself for more. It would hurt me seeing you with someone else, but I donâ€™t want to be selfish. Iâ€™ll let you grow and find what will really makes you happy. I may or may not regret giving you this letter but I know I would not regret telling you this...I loved youâ€¦I loved you even before.  More than I expected, more than you knowâ€¦                I still consider you one of the best things that ever happened to me. Sana hindi na lang nagbagoâ€¦if only I could turn back time...          Thank you for everything. Thank you for the good memories that are worth remembering. Sorry for making such a big deal out of this.      I never imagined Iâ€™ll be seeing you again after four long yearsâ€¦whatever the reason,                                                 Iâ€™m glad it did.                                                      I hope someday weâ€™ll meet again...                                     As they say, only time will tellâ€¦                                                      >Catherine '),(11915,3737807,'01,July,2004','2004-07-01 00:00:00','   somebody said na if u love someone, set them free...gago b guamwa ng wisdom n 2...d nia b alm kung gano kahirap ung pinapagawa nia? hindi ko alm kung anong dapat kong maramdaman...  oo mahal na mahal ko p rin cia, khit ang dapat magalit ako sa kanya... khit ayaw nia na sakin, khit cguro...my mhal na ciang iba...  alm ko o.a. na, sbi nila kht respeto n lng sa sarili ko, khit un n lng...ngaun, un ung gnagwa ko...kya lng d nmn ako masaya...  bkit ganon??gnawa ko nmn lhat d b? hanggng ngaun iniicp ko pa rin kung anong nagawa ko 2 deserve this kind of pain...  hanggang ngaun, hinahanap ko pa din cia,hanggng ngaun...  ang hirap pla ng sobrang saya mo, then suddenly wala na... gus2 ko lng mlmn kung bkit...everything seems so perfect, im ready 2 spend my lifetime w/ him..no questions ask...i thought, kontento na cia, masaya na cia, kala ko lng pala...  ang hirap ng empty ka...sobra... ang hirap ma-miss ung isang taong wala nmn ng pakialam sau... ang hirap pag bumuo ka ng pangarap kasama cia, tapos d nmn pala ma22pad..  bkit ganon???  gus2 ko ciang sumbatan..gus2 kong pamukha sa knya na wala akong gnagawang masama sa knya para saktan ako ng gani2... gus2 ko maramdamn nia kung gano kasakit... i really, really hate him...  he was everything i\'ve dream of...sana nanatili n lng ganon, hindi n lng nging reality...don cguro masaya p rin ako hanggang ngaun...kasama cia... '),(11916,3577655,'10,June,2004','2004-06-10 00:00:00','   Ok everybody. I\'m going to admit I\'m  not the most unfortunate soul out there. And I feel bad. So I want all of us to start praying for Kellie\'s friend who just lost everything in a fire.   Psalm 23 A psalm of David. 1 The LORD is my shepherd, I shall not be in want.  2 He makes me lie down in green pastures, he leads me beside quiet waters,  3 he restores my soul. He guides me in paths of righteousness for his name\'s sake. 4 Even though I walk through the valley of the shadow of death, I will fear no evil, for you are with me; your rod and your staff, they comfort me.  5 You prepare a table before me in the presence of my enemies. You anoint my head with oil; my cup overflows.  6 Surely goodness and love will follow me all the days of my life, and I will dwell in the house of the LORD forever.  Wow I love that passage. Later  '),(11917,3577655,'09,June,2004','2004-06-09 00:00:00','   As some of you might have known I had an interview with Dierburgs and it went really well. But today I got a letter from them saying Thanks but No Thanks. I guess its back to job hunting for me. This really sucks I was so ready and I really needed it too. Totally of subject here but would those stupid telemarketers (no clue how to spell) please stop calling us, were on the no call list thingy.  Later  '),(11918,3577655,'09,June,2004','2004-06-09 00:00:00','   Ok, every one else is like doing things every single day except for me. I just sit around all day doing nothing. This is so stupid. OOOOO I get to go to Texas. I\'m so excited Texas. Can you catch the sarcasm. If not then um yeah. That\'s all for now. I just thought I would share that.  Later  '),(11919,3577655,'09,June,2004','2004-06-09 00:00:00','   Well I can tell that it\'s summer now. I\'ve already slept into 12:00 pm one day, I\'ve stayed up till 2:00 am and I\'m waiting to pull an allnighter.  It\'s starting to get really hot outside, man am I happy that we have a pool. Not that I use it that much. But if I ever need to I know that it is there. O yeah, I saw Shrek 2 it was awsome. There are still some movies that I wanna see. Well I\'ll post some more stuff later.  '),(11920,3650974,'30,June,2004','2004-06-30 00:00:00','    I\'m going to go out on a limb with this one, but I would put forward the idea that, just as Spider-Man was a reflection of America\'s spirit immediately post-9/11, Spider-Man 2 serves as a pretty good reflection of the national mood a few years further on. To explain:  The theme of the first film was outright heroism. Oh, to be sure, it was repeatedly pointed out that there are certain prices to pay; \"with great power comes great responsibility\" and so forth. But the essential  nobility  of it, the worthiness and worthwhileness of Peter\'s choice to become a superhero was never really questioned.  The second film  does  question those assumptions. We\'re asked to consider whether running around acting as a hero might be futile at best, and actually harmful at worst. Many scenes play out as darker echoes of the first film; Peter rescues a child from a burning building (without using his powers, even) and returns her to her grateful parents. Yet he learns later that, unbeknownst to him, another man burned to death a few floors further up. It becomes increasingly clear to him that he cannot hope to save everyone.  There\'s another section of the film where Peter, for a while, rejects herodom entirely. Much like the first film, he witnesses a robbery and chooses not to intervene. And this time, no tragedy results. The world keeps on spinning. We are asked to consider whether an opportunity to act heroically requires us to do so, or merely gives us the option.  Yet another scene depicts an exhausted Spider-Man, having just stopped a runaway train, about to fall into the harbor below. He is caught by the passengers, who pass him along like honorary pallbearers, and later, choose not to reveal his identity. I swear I heard someone in the theatre shout, \"Go New York!\". Trite, perhaps, but also a valid and oft overlooked point: under the right conditions, mobs of people can act with incredible nobility instead of ignorance and hate. New York really did remind the world of that, and though the first film gave short shrift to the heroism of the ordinary individual, the second reminds us of it often.  The film\'s strongest argument in favor of the hero comes from Aunt May (speaking, I maintain, the words of Stan Lee). Interestingly, she says very little about the things Spider-Man has actually done. Rather, she\'s most concerned with the way in which he has inspired others, like their young neighbor boy. This, she claims, is Spider-Man\'s greatest importance. The film could even be commenting upon itself: do all the whiz-bang action sequences pale into insignificance besides the ideals, and the nature of humanity itself, that the film projects?  The point of all this is that I think Spider-Man 2 will resonate strongly with a nation that has, perhaps, become a bit weary with playing the hero. We\'ve become very aware, not only of the responsibilities that come with the role, but of all the violence and pain and missed opportunities and unintended consequences that come with it. That\'s not to say (and the film doesn\'t say) that it\'s not worth it to be a hero; rather, it\'s that you must think  very  carefully about what you are doing, how you are doing it, and most importantly, why.  It\'s quite clear at the end that Peter is acting for just as many selfish motives as selfless ones. This does not invalidate his actions; but when he tries to forget the personal side, his powers seem to vanish. I think that there might be a little lesson there for all of us.   '),(11921,3650974,'28,June,2004','2004-06-28 00:00:00','    A quick post before bed, on my last half-hour of battery life. What possessed me to bring my laptop to KC but not an adapter?  Spider-Man 2 is, without question, far superior to its predecessor, which was itself a pretty good film. Without going into spoilers, it has a great deal more humor, and does a better job of remembering that it\'s a comic book. It also seems much longer than its 2:15 running time, not in a is-this-over-yet? way, but in a wow-there\'s-a-lot-of-plot-here sort of way.  Let me go out on a limb and say that the female characters (Mary-Jane and Aunt May) are the most interesting in the film. While all the men around them seem to be doing and thinking and feeling whatever the plot requires them to, these woman seem like fully realized human beings. That\'s very important even (especially?) in comic books, as their writers have never forgotten. Aunt May, in particular, seems to be speaking words directly from Stan Lee\'s mouth. The philosophy behind his characters and his writing becomes crystal clear. (Look for his quick cameo in a scene involving Aunt May.)  Oh, yes, and my interview today! Went quite well, I think, despite my being rather sleep deprived. Talked to three people from the company; one was definitely positive, one probably was but hard to say, and one was totally unreadable. One way or another they say they\'ll get back to me within two weeks or so. Don\'t expect me to announce it here immediately - I may, by then, have some other offers to consider - but the signs are positive, at least.   '),(11922,3650974,'28,June,2004','2004-06-28 00:00:00','    Posting from my hotel room in Kansas City. Ethernet ports in every room. What a wonderful world we live in, that this is becoming a semi-standard feature. Love the booby-traps though. Have a complimentary bottle of water! PS: We\'ll charge $4.00 to your bill if you do...  Cerner is literally wining and dining us - dinner was at a really, really nice restaurant here on the plaza. Perhaps I misevaluated who needs who more; they\'re really going out of their way to convince me that it\'s a nice place to work and nice city in which to live. When, in reality, if I didn\'t think so, I never would have applied in the first place. Ah well. Real interviews are tomorrow noon, and I will in fact be going to see Spider-Man 2 at midnight, and head home Wednesday instead.  More  urlLink decisions from the Supreme Court  today, and unlike last week, these are really cool. Holding people indefinitely without trial? BAH! Take that, Bush. Glad to see the Supremes occasionally remember that we do have a few bedrock legal principles in this country.   '),(11923,3650974,'27,June,2004','2004-06-27 00:00:00','    Posting this from Rusty\'s apartment down in Kansas City. Not a particularly long drive, but it kinda sucks when the cruise control on your car doesn\'t work. Oh well. Fittingly, he wasn\'t here when I got here - had to wait for him to clean out a poker tournament. And when he got here, he offered me a beer. Some things never change. :)  Looks like I\'m finally about to win our online  urlLink Citadels  game. About time, to. I don\'t tend to win very much, and the last few summers I\'ve tended to lose HARD. Just lucky I guess.  Happiness is showing up at a college dorm room with beer... and cookies.  Interviewing starts tomorrow. Wish me luck. I might stick around afterwards on Tuesday evening to go see Spider-Man 2 at midnight, tickets permitting. Review here if I do.   '),(11924,3650974,'25,June,2004','2004-06-25 00:00:00','    More watching romantic films alone in the wee hours of the night. \"Amelie\" this time. One of the best films ever made, in my opinion. I can\'t imagine anyone who wouldn\'t like it, so endlessly imaginative and charming. See it if you haven\'t. It also falls alongside \"Magnolia\" for films with the following message: not all of life\'s little chances and coincidences are as random as they seem; and even if they are, sheer chance is far more powerful than all of our plans put together. What Man proposes, God disposes...  A mystery: the film\'s literal title is \"The Fabulous Destiny of Amelie Poulain\", but beneath it, the English subtitles proudly proclaim, \"Amelie from Montmartre\". That\'s the district of Paris where the film is set, and apparently the title used for it at certain festivals and such. But, I can\'t imagine it being particularly evocative for anyone outside France. Curious.  Two of my better friends broke up, then reconciled, within the past twenty-four hours. Am I the only person who\'s never been through that particular dance in a relationship? I\'m never sure how to treat such people; congratulate them for getting it over with? Console them and encourage them to get back together? Be nice? Be honest? It\'d almost be better if people didn\'t tell me about these things until a week later.  Shareholders\' meeting tomorrow, which probably means I have to look decent at work for once. Then again, perhaps the investors assume (as I sometimes do) that the shabbier an engineer looks, the more brilliant he is. Heh. I wish.  I\'ve started telling a few people about this thing. Hello, if it\'s you.  \"Without you, today\'s emotions would be the scurf of yesterday\'s.\"   '),(11925,3650974,'22,June,2004','2004-06-22 00:00:00','    ceilidh : An Irish or Scottish social gathering with traditional music, dancing, and storytelling.   Unbelievably enough, it\'s pronounced \"kay-lee\", and from it is derived the semi-popular name for girls.  In other news, men\'s dress shoes are horribly expensive, for reasons that are entirely beyond me. Electric shavers, on the other hand, are good, no matter what people might say. Ten page management papers are bad; fifteen page quantitative methods papers are worse. Shareholder meetings are good, but streaming QuickTime is bad. Beer cheese soup is very, very, very good, and if it were legal in this state, I\'d ask it to marry me.  That is all.   '),(11926,3650974,'22,June,2004','2004-06-22 00:00:00','    There were a number of  urlLink interesting decisions  handed down by the  urlLink Supreme Court  today. The scariest to me is the one that says you\'re required to give your name and identification to a police officer or be arrested, regardless of whether you\'ve done anything wrong, regardless of whether he even has reason to think you have. Extremely icky. But then again, I\'m a bit of a constitutional law wonk, so maybe my opinion should be taken with a grain of salt.  Getting a haircut can be a rather painful experience when your scalp is sunburned.  Still prepping for that job interview next week: getting my suit dry-cleaned, loafers polished, and realizing that I no longer have a decent pair of Oxfords and need a new one. Maybe I can find dress shoes that don\'t suck this time.  Speaking of job interviews  and  scary legalities, the following appears at the top of the forms I have to fill out for Cerner. All perfectly standard, I expect, but still:   I authorize Cerner Corporation (the \"Company\") and/ or its agents to procure and obtain a consumer report, including an investigative consumer report, pursuant to the Fair Credit Reporting Act as part of its pre-employment selection process, or at other times during my employment if I am hired by the Company, and to make an investigation of my background, references, character, past or current employment, job performance, education, criminal and police records, general reputation, mode of living, credit worthiness, and to make other investigative checks, including those maintained by both public and private entities and public records, through data checks or personal interviews with those who may have knowledge concerning such information, for the purpose of confirming the information contained on my written or electronic application, Personal Data Sheet, resume, or other documentation and/or to obtain other information which may be material to my application for employment or continued employment with the Company.   Should I sign such a thing? Would you?   '),(11927,3650974,'21,June,2004','2004-06-21 00:00:00','    Did you know that Lincoln is home to one of the world\'s most comprehensive sellers of  urlLink licorice ?   '),(11928,3650974,'20,June,2004','2004-06-20 00:00:00','    Went to see the College World Series in Omaha last night with Wrenholt. Fullerton over South Carolina, 2-0. Pretty much a pitcher\'s duel, with only eight hits total between the two teams. I\'ve been kinda surprised over the years how much I enjoy attending athletic events that I would have absolutely no interest in playing or watching on TV. Maybe it\'s the social aspect. Like going to a movie, except you\'re allowed to talk.  Speaking of which, also saw Prisoner of Azkaban for the second time. Caught a few details I missed the first time - like the spell Hermione uses to blast open Sirius\'s jail cell, \"morsmordre\", which is actually used in Book 4 to conjure the Dark Mark. Or the fact that Lupin, while not looking very much like our traditional werewolves, does look almost exactly like the illustrations that Snape displays for the class. I also stayed to watch the closing credits this time. There\'s no extra scene afterwards, as in the previous film, but they  are  a wonder to watch. Full of special effects nearly as elaborate as those in the film itself. Fun stuff.  Unfun stuff: the Second Quest of the original Legend of Zelda is just brutally difficult. I\'d forgotten just how much so; I can\'t even manage to finish Level 4, and I\'m almost ready to cheat or give up. Yikes. They definitely toned things down for the SNES generation onwards.   '),(11929,3650974,'18,June,2004','2004-06-18 00:00:00','    Mike Wrenholt\'s birthday tonight. Stupid management test made me miss the burgers, but at least I think I aced it. Once I finally did get there, I mixed up a batch of radler, a German concoction I became quite fond of while I was studying there. Half wheat beer, half lemonade, it\'s really good stuff, particularly in the summer. Everyone else who tried it seemed to like it too.  Headed downtown to Bricktop after that, a bar I\'d never been to before. I\'d been told by reliable sources that it was \"practically a gay bar anyway\".  Now I see what they mean: besides the techno, dance floor, and actual gay men, there were lots of (IMO) straight men who dressed and acted as if they ought to be gay. Stupid cultural cross-pollenation. Stupid Queer Eye for the Straight Guy...  I did confirm an earlier observation, though, which is that techno tends to put me into an ultra-relaxed, almost sleepy state. Maybe that\'s supposed to be the intended effect - I\'m thinking of trance here, although the music tonight wasn\'t. Nobody else seemed to be reacting that way, though. Wierdness.   '),(11930,3650974,'17,June,2004','2004-06-17 00:00:00','    So,  urlLink The Legend of Zelda: Four Swords Adventures  for the GameCube kicks every conceivable variety of ass. I just got done with a six hour session with James, Sarah, and Kyros, follow-up to a ten hour session last weekend. And I still want to play more.  The game works like this: each of the four players controls a different Link. You travel through a series of very Zelda-like dungeons, eight worlds of three areas each. The gimmick is that you have to use GameBoy Advances to control the game. Most of the time, you\'re all on screen, but when you go into a house or cave, the display switches down to the GameBoy screen, so each player is potentially seeing something different. The four of you have to cooperate to solve various puzzles and defeat enemies. But, the designers also added \"force\", tiny gems that determine points at round\'s end, in order to encourage a bit of competitiveness as well.  If you\'re a fan of the Zelda series like I am, then there\'s an orgy of things to enjoy here. The story is from Oracle of Seasons/Ages, the world map is from Ocarina of Time, the dungeons resemble Link to the Past, and many of the bosses are taken from Wind Waker. It\'s like they threw everything great about Zelda into a pot, stirred, and got back something even better.  So yeah, it\'s a amazingly fun game, even if James is a  urlLink dirty, gem-guzzling whore . These are the sorts of things you just have to put up with in life.   '),(11931,3650974,'16,June,2004','2004-06-16 00:00:00','    I had a telephone interview with  urlLink Cerner  today, the first (and so far only) company to show interest after my application blizzard a few weeks ago. Pretty standard stuff - expanding on my resume, some behavioral what-if questions, that sort of thing. I think it must have gone well, because they want me to come down to Kansas City in the near future to interview in person. Yay!   '),(11932,3650974,'16,June,2004','2004-06-16 00:00:00','    So, the other night, I watched a movie I haven\'t watched in a long time: Ever After. For those of you who aren\'t familiar, it\'s a retelling of the Cinderella story with an only-slightly-modern flavor. Oh, and I watched it by myself. At first, I thought myself pretty pathetic, but then I changed my mind.  I like romances because they show people at their best, which appeals to my inner idealist. They show the world, to paraphrase Ayn Rand, not as it is but \"as it might be and ought to be\". They make me feel... confident, somehow. Reassured. That love is waiting for me somewhere in the future, that being in love is every bit as wonderful as I remember it, that every once in a while, we do live happily ever after.  By comparison, and most of my friends can\'t understand this, I  don\'t  particularly care for romantic comedies. There\'s nothing particularly funny about being in love. Oh, sure, it\'s  fun , but it\'s also (at the time, at least) the most important and serious thing in the world. Playing it for laughs somehow demeans the entire experience. For me, at least.  Romances I Dig: Amelie, Chasing Amy, Crouching Tiger Hidden Dragon, Ever After, My Best Friend\'s Wedding, The Broken Hearts Club, The Princess Bride. Many of them don\'t have particularly happy endings, but they have something even more important: honesty and total sincerity. And lots of it.   '),(11933,3650974,'16,June,2004','2004-06-16 00:00:00','    So, I finally broke down and made myself a weblog. As perhaps the only neo-luddite software engineer in existance, I seem to be the last one to jump on any given bandwagon. Would you believe that I don\'t even own a cellular phone? But this one I just couldn\'t ignore any longer.  More than that, though I\'ve never kept a diary, I\'ve always had long, literary rants pop into my head, on a variety of subjects, and usually late at night. Normally, I would simply IM them to my friends, but it\'s been happening later and later, so that nobody\'s awake by the time I need to talk. And if I wait until the next morning, it\'s all fallen apart.  So, here I am. This will be my official place to brain dump, when nowhere else will do. But I think I\'ll let it sit for a while, accumulate some content, before I start telling my friends. Otherwise, they\'ll expect me to have something clever to say every day...   '),(11934,3650974,'29,July,2004','2004-07-29 00:00:00','    Had lunch with Ashley at Vincenzo\'s today; wisdom teeth aside, marriage seems to be agreeing with her. She showed me the pictures from the wedding, and I think I managed to be blinking in every last one. Oh well. The ones I\'m in aren\'t the important ones anyway. I do believe she and I will be collaborating on a new project, though: keeping her husband from brewing biodiesel in the garage. Anything that requires collecting used oil from restaurants is just a bit too close to a Simpsons episode for his own good.   Looks like I\'ll spend a good chunk of the weekend hanging out with (among others) my ex\'s new boyfriend. Surprisingly, I find that I\'m fine with that, and have been for some time now. When did I become so, well, mature? And rational? And sane? It\'s a very Tom Hanks sort of feeling to wake up one day and find that you can be big about things. Best guess: the years have taught me that friends are rarer and more important than lovers. Or as Mick Jagger might have said,  You can\'t always get what you want / But if you try sometimes, you might find / You get what you need...    At some point this weekend, I\'m also going to try to see \"The Village\". I\'ve got a huge fondness for atmospheric horror films, and it seems to fit the bill.   '),(11935,3650974,'28,July,2004','2004-07-28 00:00:00','    I\'ve probably mentioned it in person to most of my readers, but I did, in fact, accept a job as a software engineer with Cerner. I\'ll be starting September 20, which gives me a month after graduation to find an apartment and get moved, not to mention go to a few football games and exchange some tearful (really!) goodbyes with the various Kauffmanites. Come visit me in Kansas City, I\'ll be lonely...   '),(11936,3650974,'27,July,2004','2004-07-27 00:00:00','    Tom Rice is back! We, the living, had to suffer through three months of total radio silence that began the moment he got to Arizona, General Dynamics, and Amy. I suppose I can\'t  entirely  blame him, but I might gently remind him that he does have other friends. :) Anyway, he says he\'s going to start keeping up his blog again (link at right), so keep an eye on it.  I do have to raise one bullshit flag, though: Tom claimed that he didn\'t have Internet at home all this time because he was lazy and didn\'t get around to it. Please. For a geek of our stature, that\'s like saying you forgot to eat or drink for a few months. At least come up with a believable excuse...  Justin Bechtel also made his trimphant return to the Internet this week. I\'d like to think it\'s because he missed me, but the truth probably has more to do with his wife\'s two-week vacation (alone) to go visit her family, and him being bored out of his skull. I wonder what it\'s like to reach that point in your marriage where you realize that your social lives revolve almost entirely around one another, and that&nbsp;you haven\'t really been  alone  in years. It seems scary, yet somehow alluring.   '),(11937,3650974,'25,July,2004','2004-07-25 00:00:00','    It occurred to me that I haven\'t said much in here yet that was overtly political. I guess I assumed that most of my friends are pretty familiar with my views already. But with the Democratic National Convention going on, and the silly season&nbsp;winding up, it\'s started to be on my mind. So here goes nothing...   \"For every complex question, there is a simple answer, and it is wrong.\" -- Oscar Wilde. If anyone from the Kerry campaign is listening,&nbsp;it ought to be one of his slogans. I\'ve seen him get a lot of flak for giving long, complex answers to questions. More power to him, I say. Anything like a honest answer to the sorts of questions we ask our politicians is  going  to be pretty complicated. I think it\'s mainly the (any-wing) media itself that comes down on him for this, because it makes their life difficult when you can\'t be easily reduced to sound bites. The only other possible the explanation is that the electorate is&nbsp;mostly idiots, and doesn\'t like it when the smart man uses big words, which I&nbsp;don\'t believe is the case.   I also get pretty testy whenever someone points out that Kerry has been on many sides of various issues during the course of his career. You know what? So have I. It\'s called learning. Show me a representative who\'s voted totally consistently, on every bill, on every issue, and I\'ll show you a representative who\'s never had a conscious thought in his life. If he can explain to me why he voted the way he did on any given question, then he has my full support. (All of this, of course, presumes that a given bill only encompasses a single goal which, guess what, is rarely the case.)   Is there really someone out there (cheap shot: other than George W. Bush) who believes that the world is a place of black and white and simple solutions? Please, tell me their address, that I might go and smack them over the head with the frying pan of moral complexity. It\'s sure as hell not the world that I live in. I don\'t want my president to be a man whose job it is to gloss over the messy details of messy situations. It\'s not good vs. evil or us vs. them, and it\'s entirely possible to be both with us and against us. Bring on the ambiguity, I say.   It has become increasingly clear to me over the years that a president\'s greatest influence, and greatest legacy, is the stamp they leave on our nation\'s culture. They set the tone for us all, for years at a time. I\'m a bit of an idealist, I guess, but I need to vote for the candidate whose ideal country seems closest to my own. Because, in no small measure, they will push the country in that direction. In that light, for me, it\'s an almost laughably easy choice...   More than anything else, though, I fear the results if the popular vote and the electoral college go separate ways again. I think it would be disastrous, undermining the already shaky confidence of an apathetic populace, on both sides of the aisle. Nothing could be more poisonous for a democracy than the impression that your vote doesn\'t matter.   Random tidbit: rumor has it that the Arab network  urlLink Aljazeera  will be broadcasting more of the DNC than any of the American networks. Good for them. I daresay that the average Middle Eastern citizen has an even more urgent interest in the results of the election than we do. It\'s not life and death for us, but it might be for them.   He\'ll slay me for this, but I\'m going to link to a few of my friend Derrick\'s  urlLink various   urlLink sketches . I think he\'s a really, really good artist, and that he should hear that from more people than just me.   '),(11938,3650974,'25,July,2004','2004-07-25 00:00:00','    Spent the evening at  urlLink James Arthur Vineyards   &nbsp;with Whidden, Holley, and the Elder Holleys. &nbsp; I hadn\'t been there in a  long  time, back when it was little more than vines. They\'ve come a long way since. Twenty-one different varieties of wine, sold all over the state, not to mention a very posh store and gorgeously landscaped grounds. Oh, and resident kittens too! If I ever managed to get a date, it would definitely be a place I would go.  Five bucks gets you sample glasses of six different wines; they\'ve also got cheese, bread, chocolate, etc, and live music on the weekends. I think my favorite was the Edelweiss (also their biggest seller), with the Vignoles being a very close second. Both semi-sweets, even though I usually go for semi-dry. We also tried a bottle of the ice wine, which is extremely expensive, extremely sweet, extremely alcoholic, and extremely good.  What else did I do this weekend, anyway? Last night seems like last week. Oh, right: Animal House. I realized that my reaction to it is probably pretty common: if only college had actually been like that, and, college occasionally  was  like that. I mean, to me, everything that happens in the movie seems perfectly logical and straightforward. Kinda scary. Kinda cool.   '),(11939,3650974,'23,July,2004','2004-07-23 00:00:00','    A nice evening at Ghettowood with Whidden and Holley and the Selleckites (Johnson and Stolee). Pizza, swimming, drinking, and Stargate. What more can a man ask for in life? Also, hot tubs  rock . I hadn\'t been in one in forever and had almost forgotten their orgasmic wonderfulness. Need to find an apartment complex that has one...   I found myself getting into that nice, relaxed, among-friends groove again, as discussed a while back. I know&nbsp;I must come off as recklessly blunt and honest sometimes, the things that come out of my mouth, but really, my guard doesn\'t go down all that often. Were I to go around saying the things that were really in my heart, there\'d be far more compliments than the usual, sarcasm. But it turns out that the vast majority of my friends don\'t respond well at all to any sort of admiration or praise. A pity. By and large, I think they deserve it.   ...and just when my own life is going so well, reviewing other blogs I see that one of my best friends\' life has just exploded into a million pieces. Time to go be a friend, I think.   '),(11940,3650974,'21,July,2004','2004-07-21 00:00:00','    Out of the blue today, got an email from my very first ex, whom I hadn\'t seen or heard from in at least six years. Stranger still, instead of updates about life (or at least location), the bulk of the email is devoted to anime recommendations. I... well, I don\'t know what to think, I guess. But I did reply.  Theory: highly competitive people instinctively recognize and focus on their own kind. I won all the board games we played tonight, and it wasn\'t skill, but it wasn\'t just dumb luck either. While I&nbsp;largely ignored the other players, they spend a good deal of time and energy on obstructing one another,&nbsp;and ignored me. And all of a sudden, I\'d won. Fool you once, shame on me; fool you twice, shame on you.   urlLink July Jamm   &nbsp;is this weekend, and I\'ll be volunteering on Friday and Saturday. Come by and see me, why don\'t you. The food is good, the music is good, and admission is free through the afternoon. Looks like it\'ll be a little cooler, too. If nothing else, you  must  try the Jiffy Burger: bacon, monterrey jack... and peanut butter.   '),(11941,3650974,'20,July,2004','2004-07-20 00:00:00','    OK, so I know that by now there are at least a half-dozen people reading this thing on a regular basis, yet I hardly ever get any comments or feedback from any of you. Am I talking to myself here?   '),(11942,3650974,'19,July,2004','2004-07-19 00:00:00','    How much do your office\'s elevators suck if you\'ve actually planned out what to do in the event that the cable snaps and you find yourself in free fall?&nbsp;  I\'ve been on a buying spree lately, snapping up quite a bit of merchandise from various favorite movies and television shows, not to mention purchasing a realization of a mathematical joke for a friend\'s birthday. Don\'t ask. I must be feeling wealthy, because I really don\'t need any of this shit. Though, I think it comes with aging, the wish to crystallize certain memories in a physical form. A talisman to recall the things we know we\'re going to forget.  &nbsp;  It was suggested to me today that I shouldn\'t be friends with a certain someone simply because we happen to disagree on nearly every political and social topic imaginable. What nonsense. Those are the people I  most  need to be friends with, because that\'s how opinions are changed and compromises forged.&nbsp;The only way for this country to&nbsp;become a lost cause is if we start to believe that we have nothing in common with those who see the world differently. I firmly believe that I\'ve changed a lot of people\'s minds simply by being honest and open about who I am, what I am, and what I think is right.  &nbsp;  Nor am I the only one. It is&nbsp;my great hope for my generation that we will improve the world simply by having&nbsp;clearer visions of one another. &nbsp;  Unrelatedly, I was called today on my bad habit of asking difficult philosophical questions of my friends that I haven\'t thought&nbsp;through myself. In the interest of spreading the love: do you want to have children someday? Why or why not?   '),(11943,3650974,'19,July,2004','2004-07-19 00:00:00','    Wow, so it looks like this thing is already paying off. Tonight I received an email that\'s the nicest thing, written or spoken, that I\'ve ever gotten in my life. Any attempt at a similarly lengthy reply would only belittle, so I\'ll leave it at: Thanks. And call.   '),(11944,3650974,'16,July,2004','2004-07-16 00:00:00','    So, I saw  I, Robot  on Saturday night; somewhat against my will, since the trailers totally turned me off, but I\'d sorta promised, so whatever. As expected, it wasn\'t the best movie, and dragged whenever Will Smith was onscreen. The first half hour or so, particularly. It felt like what might have originally been a thoughtful, philosophical sort of film got action scene after action scene tacked on until it was almost recognizable. &nbsp; The film does shine, however, whenever the titular robot (\"Sonny\") is onscreen. He\'s easily the most interest character in the piece, and when the others interact with him, you can finally hear the words of Isaac Asimov coming through. It says a lot that the humans ultimately, grudgingly, come to respect him as the conscious being he is. I like the occasional hopeful vision of the future mixed in with the usual dystopian fare. But please, if you see this movie and it intrigues you at all, go back and read Asimov\'s original stories. They\'ll satisfy you far more. &nbsp; So, in other news, Russ was up this weekend,&nbsp;so we had the usual gorge-and-guzzle fest. In other words, it was a blast. Converted a couple of new people to the joys of Four Swords, which is rapidly becoming my favorite game of the moment. Rewatched Bourne Identity, which wasn\'t as awful as I remembered it and actually makes me consider seeing Bourne Supremacy this weekend. And, learned that the Imperial Palace is closed  only  on Saturday afternoons. How bizarre is that?   '),(11945,3650974,'16,July,2004','2004-07-16 00:00:00','    A new series premieres on the SciFi Channel tonight:  urlLink Stargate Atlantis , a spinoff of the absurdly popular Stargate SG-1. For those of you who\'ve been living under a rock, it\'s a really, really,  really  good show. Now in its eighth season, it still pulls in upwards of three million viewers a week. Not to mention it\'s already been on longer than any Star Trek series ever was. So somebody must like it. :) &nbsp; Obviously, I haven\'t seen Atlantis yet, but since it shares the same writers as SG-1, it\'s almost certain to be excellent. It\'ll be on at 8:00 and again at 10:00. If you\'ve never tuned in before, give it a shot.   '),(11946,3650974,'15,July,2004','2004-07-15 00:00:00','    Spent last night over at James\'s house, first watching The Amazing Race, then playing cards with the Selleck boys. It felt good. Really good. I missed it, and them, a bit more than I realized. It\'s a sobering though that, in the not-too-distant future, I might not get to play for a long while. My only hope (cough, cough) is that I\'ll be able to do so electronically...  Thought: I wonder if an element of chance is an essential component of a game played among friends. It lets you be competitive and yet shrug off the actual outcome as unimportant. Most card games share this property, but some others do not. If nothing else, I\'ve noticed that purely strategic games (chess comes to mind) don\'t seem to encourage conversation, and in my opinion, that\'s the reason you\'re playing in the first place.  The preliminary designs for the  urlLink Nebraska state quarter  are out; they\'ll be narrowed down and opened for voting sometime next month. Some pretty cool ones in there, not to mention the hilariously awful.  Not sure why, but I\'ve noticed that I have next to zero libido this summer. (Yeah, yeah, TMI, cope.) I haven\'t so much as thought about going on a date, much less tried to meet someone. Maybe I\'m not getting enough sleep. More likely, I don\'t feel right about getting involved with someone when my life is in a huge state of flux and I can\'t even be certain where I\'ll be in a few months.  But I also have a feeling that if I do go to KC, I\'ll be so absurdly busy for the first few months (year?...) that I might neglect that part of my life there, too. Not to mention having no clue where the good bars are and that sort of thing. There\'s the workplace itself, but I\'ve found out the hard way that dating someone you\'re going to have to see every day regardless isn\'t exactly the best idea on earth.  Or maybe I just need to take my own advice, given to plenty of others, that love happens only when you\'ve completely given up hope of it ever happening.   '),(11947,3650974,'12,July,2004','2004-07-12 00:00:00','    First off, welcome back to all the kids over at Selleck. We missed you.  Just got back from Kansas City, after an unexpectedly eventful trip. As we were leaving town, Jesse decided we should take the scenic route - that is, the Kansas side instead of the Missouri side. Sounded good to me. And it  was  pretty scenic; lots of green rolling countryside. Saw Leavenworth Federal Penitentiary for the first time, a monumentally imposing structure that reminded me of nothing so much as the Smithsonian. And we made a lot of unexpected detours onto dirt roads to get past accidents or missing bridges.  Then, as we neared the Nebraska border, we noticed a nasty-looking storm looming dead ahead. Seemed like a good time to pull off, have some dinner, and wait for the storm to pass. And not a moment to soon. As we got out of the car outside a McDonalds in Hiawatha, KS, we realized that the town\'s tornado sirens were going off. And moments after we walked in the door, all of the power went off and stayed off. And moments after  that , lightning struck the gas price sign directly across the street. Then it started raining, horizontally, in 70+ mph winds. Turns out that a tornado was indeed passing only a few miles to the south. We only got sideswiped by the edge of it, but that was nasty enough that I don\'t want to think about the eye. Between the storm and the detours, by the time we got home, a normally three hour trip had taken over five hours. Now that\'s pretty damn scenic.  Otherwise, a pretty darn relaxing weekend. We drank a lot, slept a lot, lay around uselessly a lot. Rusty was perhaps a bit more energetic than the rest of us. Jesse and I suspect we enabled him somehow, and we\'re not sure how we feel about that, but all\'s well that ends well.  Second session starts tomorrow, which means that I\'ll be in Advanced Finance, 11:00 - 12:30, Monday through Friday. If your schedule is compatible with mine and you ever want to do lunch, just let me know.   '),(11948,3650974,'10,July,2004','2004-07-10 00:00:00','    Welcome to what will likely be my earliest blog post ever. Down at Russ\'s place in KC again; went to the bars and got totally smashed last night. Crashed out on the futon, woke up around 7, and now here I am. Nobody else\'ll be up for hours. But I did get 5-6 hours of sleep, which is just as much as I get in a normal evening. And this hangover headache won\'t let me sleep again any time soon. So yeah.   '),(11949,3650974,'09,July,2004','2004-07-09 00:00:00','    Just got back from watching part of  urlLink The Mysterious Cities of Gold  with James. It\'s by far my favorite cartoon series ever, and possible my favorite television series, period. A joint French-Japanese production, it aired several times on Nickelodeon back in 1985-1987. For people of a certain age, like me, it was our first introduction to the world of anime. And what an introduction.  The show is set in Peru at the time of Pizarro\'s conquest of the Incas, circa 1532. It follows the adventures of a Spanish boy and an Inca girl searching for El Dorado, the City of Gold, to which they alone carry the key. The plot of the show is pretty fanciful (sometimes echoing Stargate, even) but the settings are very much grounded in reality. All of the historical figures are real, and the backdrops are accurate reproductions of famous lost cities, including Macchu Pichu, Palenque, and Tikal. Each episode ends with a short, live-action documentary about the history and culture of South America.  True, the series was aimed at children, but I find that it\'s held up remarkably well over the years. The whole genre of \"historical-mystery-adventure\" is alive and well, and still very much appeals to me. It got me interested, not only in anime, but the musical styles of the Andes that it features prominently. I\'d recommend it to you if you dug, say, Castle in the Sky, The Fifth Element, or even Indiana Jones - there\'s a lot of trapped temple raiding going on here, too.  It has never been released on DVD, but I\'ve got a very high quality set on VCD. If you ever want to borrow, let me know. I\'d be more than happy to share.   '),(11950,3650974,'08,July,2004','2004-07-08 00:00:00','    I\'ve become fond of using LOTR as a metaphor for my feelings towards a lot of my classes, and truth be told, software projects too. Even when I\'m really enjoying myself, there comes a point when it needs to be over with. It\'s as if you\'ve been lugging the Ring of Doom up a mountain for weeks and you finally toss it into the Crack of Oh Hell With It All Let\'s Go Have a Beer. An immense weight is lifted from your shoulders.  Or to quote from the new Stepford Wives, \"I Love You But Please Die\".   '),(11951,3650974,'07,July,2004','2004-07-07 00:00:00','    Regardless of what you might think, I don\'t usually write take-home finals the night before they\'re due. It\'s not my preferred way to operate, and it doesn\'t result in my best work. Still, this one is turning out pretty decent, and I\'ll probably maintain whatever grade I had in the course anyway.  I\'ve been getting back into BitTorrent again, to catch up on missed episodes of favorite TV series. It works a lot better than I remember it. Either they made some major improvements to the client, or the university was throttling my connection all along.  Things I didn\'t know: it would appear that the concept of bankruptcy is somewhat peculiar to American law. An article I read on the BBC tonight identified it as such, and explained what it was for the benefit of its British readers. Since I haven\'t heard of debtors\' prisons lately, what exactly  do  other civilized nations do when people can\'t pay their bills?  It\'s an enduringly odd experience to spend most of your life considering yourself a realist at best and cynic at worst; and then, one day, to realize that you\'re rather more upbeat on humanity than most of the people you know. There\'s my new bumper sticker for the day: People Don\'t Suck.  PS: If you\'re going to comment here, for heaven\'s sake don\'t bother doing it anonymously. There\'s only a handful of people you could be anyway.   '),(11952,3650974,'07,July,2004','2004-07-07 00:00:00','    The following post is about what the following post is not about. If you\'re not familiar with them already, the history and theory of  urlLink memes  would be a useful prerequisite.  Right, then: anyone who knows me has probably read the book Snow Crash by Neal Stephenson, or ought to. To summarize briefly, it revolves around a \"metavirus\" that is both biological and informational, and can therefore be spread either through sexual contact or by seeing it in digital form. Among its other effects, it exposes some otherwise deeply buried linguistic structures within your brain and, as a result, renders you more vulnerable to infection by memes. Think of it like AIDS for the mind; your natural defense system is slowly stripped away.  This is all a perfectly good description of the Internet; as the linked article above suggests, \"the ultimate meme vector\". Ideas that are silly, wasteful, misleading, unfunny, or just plain stupid spread across the minds of the world with a speed and scope that would cause the CDC to swoon. Our poor minds, with their prodigious yet limited resources, are forced to spend more and more time filtering this tidal wave until they collapse of exhaustion.  Which brings me back to my original point: I\'ve made it a conscious effort not to use this journal to spread other people\'s memes, and I wish more would follow suit. No collecting all the links I\'ve read all day, no quizzes, no \"which   are you?\", etc. I think of it as doing my part to boost the Internet\'s immune system. Otherwise, we get locked into an echo chamber, where the same ideas circulate over and over again. Just ask anyone who relies on (usually only liberal or only conservative) blogs as their sole source of news.  Though, I suppose I might occasionally try to start my own memes, as in the previous post. But that\'s another matter entirely. :) Regardless of whether you believe any of this is literally true, it can be a useful metaphor. Unfortunately, I\'ve just gone and spread the meme of memes themselves. Recursion joke, old: see recursion joke, old.  Back to more usual mundane matters. Got a final on Thursday, after which only one more class stands between me and a master\'s degree. I think I\'ll enjoy being called Master...  Off to Kansas City again this weekend, to visit Russ, go to the Boulevard Brewery, and generally debauch. So yay to that.  I posed this question to a friend the other day without having an answer myself, which is hardly fair: if indeed we are reincarnated, do we keep meeting the same soul mate in lifetime after lifetime? And if so, is that a good thing or a bad thing?   '),(11953,3650974,'06,July,2004','2004-07-06 00:00:00','    Everybody should watch the season premiere of  urlLink The Amazing Race , tonight at 8:30 on CBS. It\'s easily my favorite television show, and the only reality series that I really get into.  The premise: ten teams race around the world, Phineas Fogg style. There\'s none of this \"voting people off\" business; the only thing standing in their way is the usual hazards of travel. People who\'ve traveled the world will recognize many of the problems they face and the places they visit; people who haven\'t, will want to. It\'s kinda like the Travel Channel on speed. Entirely awesome. Give the first episode a chance, at least, or you\'re really missing out.   '),(11954,3650974,'05,July,2004','2004-07-05 00:00:00','    Spent all day yesterday with several of my fellow JDE alumni: Mark, April, Alan, Matt, and Christy. I found it... relaxing, in a way that\'s difficult to describe (but I\'m obviously about to try).  A cliche: they say that as you grow older, you come to appreciate ever more the people who knew you when you were young. True, I think, because the changes we go through in life matter as much as the person we are at any given moment in time. Or, to put it another way, shared experiences are the foundation of most friendships, and there are few people I\'ve shared more experiences with than these people.  So... yeah. I can unwind around my oldest friends because, well, I feel normal. I don\'t have to worry about what they might think or say, because I already know perfectly well, and it\'s nothing to worry about. In fact, I get so laid back that they probably think me downright socially passive - I mostly listen, and rarely talk. When really, I\'m just appreciating their company.  On that note: Mark and April are in Minnesota and doing well, apparently settled in to a truly blissful marriage. Alan is still in Lincoln, still erratic, inscrutiable, and brilliant. Matt has a job in Omaha but seems to be looking for a better one. Christy will be graduating with me in August and heading off to a job in England. And Drew, Alan\'s (and my old) roommate is still either completely without judgement or completely without fear.  In other news, I somehow found time to make it through two of the three papers I needed to finish this week. At this rate, I might actually manage to graduate. Again.   '),(11955,3650974,'03,July,2004','2004-07-03 00:00:00','    I should really stop underestimating my own skills at bullshit. This paper that I have to write for my quantitative methods class is a good example. \"Applications of Game Theory to Ad-Hoc Networks\". It\'s not my area of expertise or my best work, but it\'s clear, coherent, basically accurate, and not plagiarized, which is ahead of a lot of term papers. And it\'s proving surprisingly easy to write. Now all I have to do is get it done and turned in by Tuesday, so that I can graduate next month...  Got a call from the lady at Cerner yesterday, needing a clarify a few items they need to do their background check. Which is a good sign, I suppose, since they probably wouldn\'t bother if they weren\'t interested.  Geek Trivia: the Taco Bell near my house turns out to be store #404.  I\'m off in a few to a party at Clay\'s house (a kegger, even). He wouldn\'t let me bring food or drink, so instead, I\'m bringing fireworks that I bought from James\'s stand. I know for a fact we\'re going to have some definite pyros in the crowd, so they should enjoy themselves.   '),(11956,3650974,'16,August,2004','2004-08-16 00:00:00','    Watched Kill Bill Vol. 2 with James and Derrick tonight, the first time I\'ve seen it since the theatre. I must say, it\'s held up better than I expected. I realize that Tarantino is very much an acquired taste, but I honestly believe there\'s a lot to like in this film. The first one is mostly about images, but the second one is mostly about dialogue and ideas. I was also reminded that there\'s a lot of room for sequels, should Tarantino ever choose to make one. And he\'s hinted he might. These are the unanswered questions from the first two films, as I see them:  * What happened to Vernita Green\'s daughter, whom the Bride told \"When you grow up, if you still feel raw about it, I\'ll be waiting\"? She might come after the Bride, but she might also come after her daughter, who would be the same age.  * What great crime did Bill commit, in Hattori Hanso\'s eyes, that would make him break his oath in order to forge a sword to kill Bill?  * Now that O-Ren is gone, who is the leader of the Yakuza, and how do they deal with those who slay their leaders?  * What happened to Elle Driver; did she survive the trailer with the black mamba? If she did, she\'d certainly have it out for the Bride.  * Did Bill see through Elle\'s lie about the nature of his brother\'s death? If so, would he have put a price on her head before he died himself?  * Why did the DVAS apparently disband shortly after the Bride\'s \"death\"? Vernita, O-Ren, and Bud, at least, all apparently went their separate ways. Bud alludes to an argument and falling-out with his brother, but that\'s all we know.  * At the time of Bill\'s death, was his organization still in operation? It could be that he disbanded the DVAS and retired to raise his daughter; but it\'s also possible he went through new batches of them on a regular basis, a bit like Charlie\'s Angels. If so, are there more students of assassination (or even ex-lovers) out there who might have it in for the Bride?  * Did Bill also know the Five-Point Palm Exploding Heart Technique? If so, who else might Pai Mei have taught it to?   '),(11957,3650974,'15,August,2004','2004-08-15 00:00:00','    Among the many, many reasons I love Stargate SG-1 is this: in how many other series, even sci-fi series, would you get to hear a 100-year old alien warrior discuss his love for Def Jam Vendetta?  And that\'s just the beginning of an episode that features a character becoming trapped inside a VR simulation which, for the benefit of those observing from outside, is rendered using the Half-Life engine. I shit you not, a goodly amount of the hour is spent looking at polygonal versions of the actors. If it isn\'t a real mod yet, I expect that it will be very soon.  Edit: I shouldn\'t speak before I google; there are, of course, several such mods out there. Also, it was pointed out that the characters claim it\'s the Doom 3 engine, but it simply doesn\'t look it to me.   '),(11958,3650974,'14,August,2004','2004-08-14 00:00:00','    Graduated today (again, again). I love this whole summer commencement thing: 800 people instead of 2,300 makes it surprisingly tolerable to sit through. I started to tell a friend today that he could call me  Master  Beermann now, before realizing that he already  does  call me that on a semi-regular basis. It\'s just never really registered before. I\'m flattered. I think.  Two of my cousins, brother and sister, both had baby girls arrive recently. One has the first name Regan; the other, the middle name Marin. Am I the only person who\'s seen The Exorcist?  On my drive home from work, I pass a field where a fellow has been camped out for the last few days. Above him is a large, handpainted banner reading, \"CANCER SUCKS!\" You tell the man, buddy.   '),(11959,3650974,'13,August,2004','2004-08-13 00:00:00','    So, Alan mentioned to me yesterday that he\'d found and fixed a bug in the ISIS GPS code, dating from when I originally wrote it, nearly two years ago now. It caused all latitude/longitude measurements to be off by around 20 miles. It was probably caused by a combination of coding error on my part and/or my sketchy understanding of the NMEA protocol and/or the sketchy documentation for said protocol.  Normally, this sort of thing would really bother me; I\'m supposed to be a professional and all, which means being methodical and testing your work thoroughly. But in this instance, I just have to laugh. I find it fucking  hilarious  that not a single developer or user, some of whom are much smarter than I am, noticed such a huge problem in a system that\'s been in heavy production for over a year. It\'s just... wow.   '),(11960,3650974,'12,August,2004','2004-08-12 00:00:00','    Tomorrow should be a good day: besides Ribfest again for lunch, the evening brings the Olympic Opening Ceremonies and new episodes of both SG-1 and Atlantis. Alas, I\'m likely to miss the latter because of a thrice-accursed reception at CBA for graduates. Bah. It\'s a good thing I have BitTorrent or I might be really pissy.  I\'m a bit torn over Gov. McGreevey\'s resignation, though at a personal level I can sympathize with him. I can\'t decide whether he made the right decision for the wrong reasons, or the wrong decision for the right reasons. I also question just how much honor there is for a man in upholding a dishonorable and dishonest wedding vow. It doesn\'t excuse an affair, but it might be high time for them to go their separate ways.  One of life\'s small pleasures: introducing your boss to Rejected.  Do you ever get the feeling that you\'re living your life backwards? The best way I can sum up my overall mood lately would be: The Elder Days are gone. The Middle Days are passing. The Younger Days are beginning...   '),(11961,3650974,'12,August,2004','2004-08-12 00:00:00','    Tried two new games tonight, which I\'d purchased based only on sales rankings and the recommendations of the good folks at  urlLink BoardGameGeek . The first, MÃ¼ & More, consists of an unusual deck of cards (12 each in 5 suits) and rules for five different games. One is like hearts, one is like pitch, one is like speed, one is like go fish, and one (the titular MÃ¼) defies description. Nearly all of them are more complex then they seem at first, but they also tend to be mercifully quick. Not a bad value for $10 if I do say so myself.  The other game we tried, Acquire, is apparently a classic that dates back to the \'60s, but is still in print and still played in tournaments. None of us had ever heard of it or played it before, though. You place tiles on a city-like grid, causing clusters of tiles (\"corporations\") to form and eventually merge, netting you shareholder bonuses and capital gains. Pretty interesting stuff, though I don\'t think I quite have the strategy down yet. As if I  ever  do...  Ribfest starts tomorrow! Barbecue is awesome, and a choice of a half-dozen vendors is even more awesome.  Also met Tracy\'s long-mysterious boyfriend, Eric, over dinner at Lazlo\'s tonight. Seemed like a pretty nice fellow, a film studies major, maybe a bit on the quiet side. But there\'s one lesson I\'ve learned over the years, often and well: the ones who seem quiet, very rarely actually are. In fact, they often have the most interesting things to say. You just need to figure out how to get them talking.   '),(11962,3650974,'10,August,2004','2004-08-10 00:00:00','    ...if I implied in my previous post that Myst IV was pretty nifty. It\'s  absolutely fucking amazing.  Proof: when you start the demo, stay put, but spin the camera around so you\'re looking up, perhaps in the direction of the moon. Leave it like that for a few minutes. I\'ve seen a lot of things in video games before, but I haven\'t seen anything look quite that good, and for a throwaway detail, no less. It makes me want to fly to Ubisoft Studios and hump the animator\'s leg.  Also, a hint: once you pick up the only item in the demo, go back and revisit the earlier areas. Some will have changed.  So, it seems that I\'m having lunch with Eric Graber (!) and his girlfriend (!!!). What exactly she\'s doing in Lincoln, much less in the United States, remains to be seen.   '),(11963,3650974,'09,August,2004','2004-08-09 00:00:00','    I have a strict policy of not using beta software from Microsoft, but when Service Pack 2 reached RTM I couldn\'t avoid it any longer. It took forever to download and install, but has otherwise been painless. In fact, I can barely notice the difference, even though I know perfectly well what all has changed, and I suspect most normal users will feel likewise.  It is a bit surprising how often I\'ve been hitting the security warnings about programs trying to connect to the Internet, considering that the new Windows Firewall (AFAICT) only actually blocks inbound traffic. Makes virus activity easier to detect and stop, I suppose. The new wireless networking definitely seems stabler, and less apt to switch channels just because it detects someone\'s random peer-to-peer connection.  Launching applications, opening menus, etc, feels a bit snappier but that could be purely psychological. The layout of the Security Center seems to have changed a bit since RC2. I wish I had a newer processor so I could test the NX security. All in all, not bad. We\'ll see how this affects virii in the coming months.  Spent most of the day playing with the  urlLink Myst IV: Revolution demo  which, astonishingly, runs like a dream on my oldish Dell laptop with its 1GHz chip, 256MB of RAM, and 16MB graphics card. They must have done a pretty decent job of optimizing it. And massive props for releasing the Mac version simultaneously with the PC. Don\'t see that much anymore.  Also took my brother out to dinner for his half-birthday, since his actual birthday got missed while he was at school. Outback Steakhouse is  yum,  particularly their Bloomin\' Onions, even if they do have something like 2,000 calories each.   '),(11964,3650974,'09,August,2004','2004-08-09 00:00:00','    So yeah, one of our cats died while I was away this weekend. I can\'t say it comes as a total surprise - she was sixteen years old and had been sick for a while. But it still sucks. Even more so for my brother, who actually had to bury her himself since the rest of us were out of town.  Other than the news waiting for me at home, a pretty good weekend all around. Definitely leaning towards a loft now; need to call them tomorrow and snap one up before they get rented. A  second  night of drunken debauchery at Rusty\'s; this time, drunken computer assembly! I think we must have done something wrong, though, because he can only get Doom 3 to run at 5-10 FPS, despite having a really nice graphics card, plenty of memory, and the latest versions of everything. Maybe he\'ll have figured it out by the time he reads this.  This week is the last week of summer classes, meaning that in a few short days I\'ll have a degree.  Another  degree. I swear, they\'re starting to seem like baseball cards or something...   '),(11965,3650974,'08,August,2004','2004-08-08 00:00:00','    Back in Kansas City this weekend, this time looking for an apartment. I find myself attracted to the lofts downtown, as opposed to the various apartment complexes out in the suburbs. There\'s such a mind-numbing identicalness to seemingly every complex built in the last twenty years. Little beige boxes, all of them. A few high ceilings and exposed brick walls do more than you can imagine to raise my spirit. Also, by the time you get done considering utilities, parking, etc, their prices are really quite similar.  Other things to figure out: which banks have branches both in Lincoln and in Kansas City? Precious few, as it turns out.  Not quite the orgy of debauchery of past weekends, but I\'m still having a good time over at Russ\'s. I keep meeting new faces in addition to the ones I\'ve meet in weeks past. It\'s quite the little harem Rusty has going. Keeps life interesting, anyway.  Even I must admit that Doom 3 looks pretty sexy, though I still can\'t imagine playing it myself. I need a good excuse to buy a new video card (if not an entire system) in order to play something or other. Suggestions?   '),(11966,3650974,'07,August,2004','2004-08-07 00:00:00','    But Iâ€™m not crazy, Iâ€™m just a little unwell I know right now you canâ€™t tell But stay awhile and maybe then youâ€™ll see A different side of me   '),(11967,3650974,'05,August,2004','2004-08-05 00:00:00','    For those of you who haven\'t yet had the pleasure of graduating, here\'s a primer on academic attire: Bachelors\' tassels are the school colors, doctors\' tassels are gold, and masters\' tassels depend on their discipline. Since I\'m an MBA, that means I get to wear a tassel and hood that are a putrescent salmon-like color. I believe it\'s officially called \"drab\". And for the privilege of being drab I get to pay $50. Yay...   '),(11968,3650974,'03,August,2004','2004-08-03 00:00:00','    The Games of the XXVIII Olympiad of the Modern Era begin next Friday night. I plan to watch them religiously, as I do every two years. It surprises even me; I\'ve never participated in any athletics, nor do I usually enjoy watching them on TV. Why, then, do I have such an attachment to the Olympics? Mostly to try to understand myself, here\'s a list, in no particular order:  * The Olympics absolutely reek with symbolism, ancient and modern, from beginning to end. As an example, this year\'s marathon is being run along its 2,500 year-old original course, from Marathonas to Athens. How often do you see such old traditions being observed today? (More than you think; ask me about Reagan\'s funeral sometime.) I\'ve got a huge soft spot for rituals of all kinds, even when their original significance has been forgotten. They make you feel connecting with the past in a way that simply reading history never can.  * When you see someone compete at the Olympics, you know without doubt that they are among the best in the world at whatever it is they do. There\'s a peculiar thrill that goes with that, even when you wouldn\'t otherwise be interested in the activity. I enjoy watching masters at their craft, basically. World Poker Tour is another good example of the same phenomenon.  * They only happen every two years. Well,  duh,  but it means that I don\'t have a chance to get sick of them. I can watch anything once, but these days it seems like most of the professional sports somehow manage to go year-round; they\'re always on. The games are not.  * There\'s a huge variety of things going on, and you can never hope to see them all. Maybe I\'ve just got a short attention span, but it\'s kinda cool to be able to flick on the TV and have an equal chance of seeing dozens upon dozens of different sports; some featuring people you\'ve heard of, and many that you haven\'t. It feels like serendipity.  * It\'s a travelogue. The Olympics take place all over the world in historic and, let\'s face it, often gorgeous locales. I like  The Amazing Race  for essentially the same reason. Also, the host cities spend years and millions of dollars to look their absolute shining best. It\'s probably a worthy goal in its own right (urban renewal), but it also gives them a fairy-tale quality that can\'t ever be replicated in a later visit.  I think that\'s the gist of it. So yeah, I\'ll be watching. Anyone gets bored, invite me over and we\'ll do so together. (Oh, and also: I got to hold an actual Olympic torch once, back in 1996. So awesome.)  Finally, of course, the obligatory news tidbits: I finally got Jesse to go to the Oven, which he enjoyed; the long delay was apparently due to his inexplicable theory that the place was only for dates. Strange, Jesse, strange. Everybody help me think of something to get/do for his birthday, which is coming up very soon.  Ribfest starts next weekend, and I can hardly wait. Still need to see the Village, about which I\'ve been getting mixed but not wholly negative reviews. A new batch of games is on the way from GameSurplus, an entirely awesome site with really good prices. Rsync might be one of the most Unix-centric programs ever written. Alan\'s site is down, due to his altruistic and slightly silly gesture of sacrificing its innards to revive his employer\'s server. Google\'s initial public offering will raise  e  billion dollars.  That is all.   '),(11969,3650974,'03,August,2004','2004-08-03 00:00:00','    So I got a form letter today from Chuck Hagel\'s office. (Are you listening, Tom?) I had contacted them to register my opposition to the Federal Marriage Amendment. I would assume, or hope, that all such contacts get a letter in reply, but I\'d have to imagine they had more than usual on this issue. A  urlLink press release  is attached to the letter explaining his position on the matter.  \"Americans and Nebraskans have been presented with a false choice between amending the Constitution of the United States or allowing same sex marriage across our country... Over our nation\'s history, social and policy issues have been best handled by the state... Amending the constitution, the founding document of our nation, should always be a final option, not a first.\" Brilliantly put, Chuck. I couldn\'t have said it better myself.  But then he goes off a cliff. The matter never made it to a vote; instead, senators voted on whether or not to continue debate (\"cloture\", if you\'re curious). He voted to keep the issue alive. \"There is much misinformation about this amendment. I believe the American people would have benefited from continued debate and a further understanding of this deeply complicated and important issue.\"    Excuse me? The proposed amendment was  one sentence long.  I\'m pretty damn sure I understood  exactly  what it said. Further, the debate  I  heard was doing precious little to combat misinformation; if anything, there were senators throwing more into the pot. Perhaps more importantly, had the amendment come to a real vote, how would you have voted, Chuck? Your press release implies \"no\", but your beliefs and voting record on related issues implies \"yes\". Since it\'s certain to come up again, year after year, it\'s something that I, the represented, urgently want to know. For goodness sake, stop trying to talk around it. Please.  On the plus side, though? Franking privileges are  awesome.   '),(11970,3650974,'03,August,2004','2004-08-03 00:00:00','    Is there a substantive difference between love and trust?  Aside from the abortion crowd, why don\'t right-wing nutcases protest things as often as left-wing nutcases do?  If the saying \"the good ones are always taken\" is true, and you\'re not taken, are you therefore not a good one?  Considering it takes so much more energy to hate someone than to get along with them, how do so many people around the world manage to hate so many without becoming exhausted?  Why do we depend on, or even expect, others to make us happy?  How often is someone\'s opinion changed by what someone else says, rather than does?  Am I channeling Rusty right now or  what?   '),(11971,3650974,'02,August,2004','2004-08-02 00:00:00','    This is an interim I\'m-not-dead post on my parents\' computer, since my laptop is currently hosed. Sometimes even I wonder what use a computer science degree is, but I can\'t do a thing about it until at least tomorrow.   '),(11972,3496773,'31,May,2004','2004-05-31 00:00:00','    Is this really any different? Besides, we don\'t get the fun little icons of how we\'re feeling. I feel like a smiley face drinking a fruity blender drink. I love the day off, I have never smelled so much lighter fluid in my life. I feel like I must eat grilled meat today. yet I have no grill.   Jobs are a tricky thing. I\'m scared to start a new job, but I know I must, and that requires obtaining a new job. So I must make a new resume and write letters. I hate this whole process. I want the big texan in the white hat to come and say \"Hey there, tall drink of water, want a lucrative job where you do very little and get paid a lot? I\'ll relocate you to a non-dying area and you\'ll be swamped with money and broads!\"   And I\'d say \"Boards? Why do I need boards?\"  And he\'d say, \"No, broads.\"  And I\'d say, \"I\'d rather have mamasitas thank you very much, especially if I\'m going to Texas.\"  and he\'d look all stern and then smile and go,\"Well, that\'s settled. See ya Monday!\"     THAT\'S how I should get a job!   '),(11973,3496773,'28,May,2004','2004-05-28 00:00:00','    Our upcoming show \"Godzillas and Generals\" is turning out to be a very good show!  We are trying alot of different things with this show, and you really are going to want to check it out.    One thing we are doing a little differently is reverting to our classic format for shows.  What I mean is that we aren\'t trying to tell a large, hour long story with this show.  Rather, we are doing things the old way, where we have penned a bevy of hilarious sketches, seemingly unrelated to each other.  Of course, because we are geniuses, and have a lot of dumb luck, things still tie together wonderfully at the end.  There are a handful of shows that I think are the \"Best\" the Dwellers have ever done.  In that list are Decon Santa, Kiwanis, and some of the older ones that we did during that golden summer.  I don\'t quite remember the year of that summer, but the show that sticks out in my mind is the show in which we had the most guns in a Dweller show... ever.  There were something like twelve guns on stage at one time!  And yet, throughout the show, only one character died and that was of natural causes.  It was a wonderful show, where all of the skits were written independently but in the end, thanks to being at the right place at the right time, we were able to pretty much tie together every skit.  It was beautiful.  And so this show is taking an interesting course.  We have chosen a theme instead of a plot.  That theme is \"Japanese Pop Culture.\"  It all came about when Larry and Joe came up with an idea for a skit (I won\'t give away what that skit was), and when they told everyone about it, I totally dug the idea of making the whole show sort of themed around the crazy Japanese culture (what with their samurais, ninjas, monsters, food, and videogames).  Originally trying to make the show like we had done Kiwanis and Decon, we tried to come up with this elaborate plot involving a samurai sword (and whomever had the sword became a samurai).  We realized this idea was tired, and not as interested as first imagined... and so we set out and worked on some ideas, wrote some skits, and did some improv... and the end result is \"Godzillas and Generals\"  And while the show isn\'t completely about Japanese culture, we managed to get some funny skits involving Godzilla, samurais, ninjas and a special surprise.  Along with these themed skits, we also have some great satirical sketches and a special \"guest character\" from everyone\'s favorite rock group: Blue Oyster Cult!  Enough of me giving away info about our show... You are just going to have to check it out.  Of course, it was brought to me attention the other day, while enjoying some coffee at Cafe Kolache, that our opening show is on the same night as the new Harry Potter movie.  While most other sketch comedy and improv troupes would be frightened by this, the Cellar Dwellers do not waver.  We don\'t even flinch.. we will fight Harry Potter, just as we have in the past, with our humor!  I think we can pull in more people than that puberty-breaching, girl-clamoring kiddie fest.  Of course, if the Dwellers are going to try and pull in over 100 million in our first weekend, we are going to need your help.    Friday, June 4th at the CCBC Allied Health Auditorium.  (CCBC stands for Community College of Beaver County, for those of you not from the area).  The show begins at 8:00pm and will run until around 10.  Tickets are just $5.00 and reservations can be made in advance by e-mailing the Dwellers: reservations@thecellardwellers.com  or checking out our website.    If you can\'t make it this Friday, due to work (NOT due to Harry Potter... which, I assure you, is no excuse) then you have one more chance to check out this hilarious sketch show next Saturday, June 12th also at the CCBC auditorium.  It starts at 8:00 and tickets are also $5.  But if you bring your Harry Potter ticket stub, we can charge you $7.50 instead... it\'s only fair.  And so, hope to see you there!  Feel free to comment and ask questions about the show... or just come find out for yourself next Friday.  The Dwellers love you.   '),(11974,3496773,'27,May,2004','2004-05-27 00:00:00','    Hey, just started this new Dweller blog because the other one sucked... let\'s see if this one\'s better!   '),(11975,3496773,'29,June,2004','2004-06-29 00:00:00','    As I\'ve noted in past postings, I find myself in almost constant transport from my home and Beaver County, home of the Cellar Dwellers. As I also have mentioned, traveling can lead to all sorts of adventures. Such was the case last Saturday.  I had left the triumphant improv show at the Copper dog - 4 paying customers, woohoo. It was fairly late, almost 12:30 am as I piloted my \'93 Lumina - or \'The Silver Bullet\' as I like to call it - through Rochester.  I was almost out of Rochester when I looked in my rear view mirror and saw the one sight sure to invoke terror in the mind of all motorist. I was being chased by Dan Akroyd. Just checking to see if you\'re paying attention. What my mirror really revealed was the flashing lights of a police officer.  Is there any more nerve racking event than being pulled over by the cops? Well, yes I guess there is. Trench warfare comes to mind. Still, getting pulled over always manages to be a good sucker punch experience. One minute you\'re driving along quite happily, feeling free. The next minute some guy with a mustache - they always have mustaches - is shining a flashlight in your face.  I tried to play it cool. I made a concerted effort to have all my paper work ready for the officer which was a fools errand. My glove box explodes like a party favor when opened. Papers fly everywhere. Every piece of paper work I have ever received for my car is in there. I have registrations dating back to the mid 70\'s. I don\'t know what car they belong to, but they\'re there. I was able to find my current registration. It was rolled in a ball under an ancient half full bag of Skittles.   I greeted the officer kindly, \"Hello officer. How are you tonight. Beautiful night for a drive isn\'t it.\" I practically had to muzzle myself before I became just a little too friendly. I was half a step from saying: \"Say officer, has anyone ever told you your an attractive man? You must work out.\" Remember, when talking to the police there is a fine line between polite and catching-a-beat-down-with-a-flashlight polite.  \"You know I clocked going 50 back there,\" The officer said. \"You were also a little erratic. You went over the double yellow and the white lines.\"  Crap, I thought to myself, looks like its alphabet backward time for James. But I\'m sober. At least I think I\'m sober. I only had 1-1.5 beers at Copper Dog. That was a couple hours ago. I couldn\'t be drunk. But what if they were super beers? What the hell\'s a super beer?  I remained quiet and tried my best too look sober, which I was. I offered him my license and registration. Always be helpful to the police when they pull you over. If you have food in the car offer it to them. \"Oh,\" the officer said, \"I also need your insurance card.\"  (flashback a couple of days:  my mother: James, your insurance card came in the mail. me: Just leave it there. I\'ll get it.)  I didn\'t have my insurance card, the fine for - the officer tells me - is a 6 month suspension of your driver\'s license. I panic. I tell the officer I have it. I show him the most recent card freshly expired. He seems to understand. I can\'t tell. He\'s got a flashlight in my face.  \"Where are you coming from?\"  \"The Copper Dog.\" I think it best not to lie. I was at a bar.  \"Doing a little drinking were you?\"  \"Just one or two a couple hours ago. I was doing a show. I\'m an actor.\"  \"What kind of show?\"  \"Improv comedy. I\'m a member of the Cellar Dwellers. Check out our website www.thecellardwellers.com\" Did I just plug the Cellar Dwellers? Was that a good idea? The officer seemed to accept it. It was the truth.  Then, the officer went back to his cruiser.   Now, I have no idea what cops do when they take your paper work back into their car, but I have a sneaking suspicion it\'s nothing. They\'re just making you sweat. They\'re probably back there smoking a cigarette or brewing some coffee or shaving or playing the harmonica or doing any number of other inane pursuits while your left pondering your fate. I was sweating bullets. I needed my license. I can\'t afford to pay for a ticket. The ticket could be worth more than I am right now. I\'d have to sell plasma just to make ends meet.  Eventually, the officer returns and a miracle happens. I swear it was a miracle. Little things like this confirm my belief in a loving God. The officer uttered those sweet, sweet little words: \"I\'m going to let you off with a warning.\" Praise God, hallelujah. I feel like singing. I feel like dancing. I want to hug the officer. I don\'t. It probably wouldn\'t have been the wisest of choices. I restrain myself. I do thank the officer and wish him a pleasant night. And then, just like that, I am free. I drive away with joy in my heart and less lead in my foot.  Shalom   '),(11976,3496773,'24,June,2004','2004-06-24 00:00:00','    I was just about to spend another day bored out of my mind.  I had spent the morning doing some designing, graphically of course, and had just left to visit my local Sheetz.  When I got back home, Danny and I sat down with my dad to watch some Fox News over lunch.  But wait... what\'s this?  Breaking news?  It\'s a CAR CHASE!  The saving grace of any afternoon: a high speed car chase somewhere in California.  Even better, it was a bank robbery and they were soaring on some freeway throwing money out of the window!    All three of us stopped and watched the screen.  Danny started clapping in excitement.  I tell ya, there is nothing that cheers me up more than some hella-sweet action that could wind up in disaster.  This chase scene, which at first brought back fond memories of the O.J. chase, was going to be a good one too.  There were three robbers speeding around in what appeared to be a Chrysler Pacifica, with about five cop cars following suit.  What\'s even better is they were catching everything from helicopter cam!    The driver of this car was pretty crazy at times.  He was all over the 8 lane super highway, and then he would take an exit and head into the urban areas of LA, including Compton (represent!), and then would loop around and get back on the freeway going in the other direction.  Even better, he utilized the car pool lane!  Not only was that smart, but it was also considerate, since he did have more than two passengers.  After about fifteen minutes of driving all over LA, and after evading the police which weren\'t very smart, they wound up in this suburban area.  The robbers pulled the car onto some lady\'s lawn and pulled this weird Chinese fire drill.  The woman, who was a large African American, was standing on the corner screaming at the robbers... and then screaming at the cops... and then screaming at everything.    Two of the robbers gave up instantly.  But the driver, who was a pretty hefty fellow wearing some O.G. football jersey, ran inside the house.  There was a standoff for a little while, which was especially hilarious because this crazy woman was standing between the cops and the house, getting into the way of their pointed guns.  Eventually the woman, whom I can only assume to be the driver\'s mother, talked the robber into coming out of the house and giving up.    Although I must say, judging from what I saw of the robbers... they didn\'t plan this well at all.  They were all wearing normal street clothes, no suits or black ties, and none of them had masks.  They obviously didn\'t have a good safe house, and obviously didn\'t have enough man power.  If more robbers would just do their research and watch some movies they might get away with it once in a while.  But anyways...  What excitement!  I tell ya, great car chases only happen every once in a while,  much like comets or the Olympics.  And just like the previously mentioned events, car chases are not to be missed.  My day has been brightened because of these ridiculous robbers and their poor planning.  Thank you, young vigilantes.  Thank you!   '),(11977,3496773,'22,June,2004','2004-06-22 00:00:00','    I had been aware for several weeks that my car was badly in need of two new tires. However, being unemployed and horribly destitute - funny how the two seem to go together - I was forced to drive the car as is. It wasn\'t too bad. There was a slight thrill to every car trip. I never knew when the tires would blow, possibly sending my car out of control. It\'s these little things which keep my life interesting. Today, thanks to a monetary donation to the Keep James Afloat Fund from my Aunt, I finally was able to buy new tires. Of course, I went to the cheapest place I could think of, Wal-Mart Supercenter. (Just let me say right now that I don\'t want any \'How could you support the evil, unfeeling, corporate behemoth\' comments. Maybe I\'ll fight injustice when I get the cash - maybe - and right now Wal-Mart\'s being quite helpful in keeping cash in my pocket. Long live the free market.)  Upon arriving at the Wal-Mart Supercenter in lovely bucolic cranberry Pennsylvania. I was informed that it would take a little over an hour for them to get to and do the work on my car. So, I had little choice but to find a way to while away the time at Wal-Mart, there being nothing else within walking distance. This presented many problems, the chief of which being Malaise. I mean malaise in every sense of the word from boredom to the more existential sense of loss and dislocation. Wal-Mart is like a factory for these feelings, especially when I am alone there. Just looking at the people in Wal-Mart makes me feel more and more distant from my fellow man. How can I have anything in common with someone in a Dale Earnhardt jacket? How can I entertain myself for an entire hour?  I could feel it. The malaise stalking through the aisles - a black cloud ready to envelope me. I had to keep my wits about me, stay sharp. I\'m not ready to embrace my malaise. I don\'t care to become authentic whatever that means. I read it somewhere.  I sat in the tire center waiting room for maybe 2 minutes. Just long enough to drink a cup of complimentary coffee. The waiting room was truly depressing: Colorless walls, torn vinyl chairs, 6 month old issues of better homes and gardens. One other guy was in there talking business on his cell phone. I secretly hoped he would make an inappropriate personal phone call, get in a fight with his wife or something. That would\'ve entertained the sick part of me. But he only talked about tractor prices. I left the waiting room. The man never noted my entrance or exit.  So, I was left adrift in the consumer wasteland. I moved quickly, one aisle to the next section to section, leaving whenever I felt the first shiver of boredom or despair. I took my blood pressure. It was good. I was proud of myself for a minute as I look over the deodorant for no good reason. I played with the camcorders aiming the one attached to the TV up my nose for all walking by to enjoy.  I leafed through the latest copy of People. They listed the 50 most eligible bachelors. Some how I managed not to make the list again this year. They must not know I\'m single. The guys starring in the upcoming film \'Harold And Kumar Go to White Castle\' made the list, so I have my suspicions as to the viability of said list.  I made a game out of taking the bouncy balls out of the bin and throwing them back into the top. I would throw them from all angles, from behind obstacles. I even tried it from an aisle over. I almost got thrown out from that.  I browsed the hunting rifles, asking the clerk questions like \"Do you have anything that will make a deer\'s head explode?\" or \"Do you have anything for a guy trying to compensate for something, if you know what I mean?\" I almost got thrown out for that.  I looked at the toys. I wished they had some out so I could give them a try but they didn\'t. So, I opened a bag of marbles anyway.  I tried to pick up games with some kids until their parents pulled them away. I almost got thrown out for that.  I pulled a bike off the rack to give it a test ride. I almost got thrown out for that.  I climbed up on a display section to try out some of the exercise equipment. I almost got thrown out for that.  I unfolded a camping chair and rolled out a cooler next to it. Then I chose a fishing pole from sporting goods and a goofy hat from the Men\'s section. I spent a good five minutes pretending to drink beer and fish. I almost got thrown out for that.   Almost getting thrown out ended up being the key to my morning. Walking that fine line between harmless fun and public nuisance proved quite rewarding. Testing the limits of acceptable behavior at Wal-Mart, I think someone could write a thesis on this. It gave my morning just enough of an edge to ward off malaise. In the end, I got two new tires, at least 6 warnings, and no existential crisis. Sounds like a good morning to me.   Shalom   '),(11978,3496773,'21,June,2004','2004-06-21 00:00:00','    If you have ever played a videogame that deals with war or fighting against a bunch of guys, you may know of a little feature called \"Fog of War.\"  Basically this is a thick fog that appears, limiting your view to only a close proximity to yourself.  It\'s based off of the actual term \"fog of war\" which is used to describe the thick fog of soot and gunpowder that fills the air after an intense battle.  In videogames, it\'s used so you don\'t watch every move the enemy makes, hence creating a little surprise and challenge.  I was thinking about this odd concept the other day, and I was wondering what would happen if everyone had \"fog of war\" in real life.  No matter what time of day it is, no matter how good your glasses are, you can only see four feet in any one direction.  Everything else just looks like a gray haze.  Driving would be insane.  I\'m sure I would run over all sorts of animals, people, and rubber sports equipment that happens to wander in front of my Cougar.  Parking would also be quite the bear, as my keen \"empty-spot-spotting eye\" would be totally useless.  As would my  \"hidden police officer spotting eye.\"  That\'s two eyes that are gone with the wind!  Then again, losing the cops would be a little easier (although I\'m sure their range of sight would be a little longer than mine... law enforcement bonus).    Walking around the mall would be a tad more enjoyable, since I wouldn\'t have to try as hard to avoid people I don\'t like.  I only really have to panic when they get within a few feet of me.  Muggers and Jehovah\'s Witnesses would be harder spot and defend against, on the other hand.    But if you are going to go the way of these videogames (the ones I have in mind are Age of Empires and Dynasty Warriors) then you could also employ some scouts.  Get the shmuck on a horse to go ahead and see if they have any ripe bananas at Giant Eagle.  He will come back with the word, or he would even distract the Jehovah\'s long enough for me to run in there and get the delicious yellow fruits.  Now that I think about it a bit, I\'m glad that God turned off \"fog of war\" in the options menu  when he was creating the world...  well, you can probably see that my summer has been pretty boring thus far.  I\'ve got nothing better to do than think about videogames and watch MacGyver (who is freaking awesome, by the way.)   '),(11979,3496773,'20,June,2004','2004-06-20 00:00:00','    I just got back from New York. That\'s right, the Big Apple, media capital of the world, center for culture, the city that never sleeps. At least that\'s what I\'ve been told. I spent almost all my time at my grandparents house in the suburbs, and trust me, my grandparents house definitely DOES sleep. I would even go as far as saying it slumbers. It\'s half a step away from slipping into a coma. Bed time\'s 8:30 and it\'s strictly enforced.  So, it\'s safe to say that I didn\'t exactly experience the thrill of New York City, despite being a mere stone\'s throw away. I missed the rush of the subway, the rhythm of the streets, the smell of the cabbies, but I wouldn\'t classify the trip as a loss. I learned all sorts of valuable things. Such as, it doesn\'t matter if you lose a ton of weight and everyone else thinks you look fantastic, an Italian grandmother will think you look awful. Apparently, despite being a healthy weight for my height, I\'m dangerously thin. My grandmother kept looking at me as though I were a cancer patient. She would lurk in shadows waiting to pounce on me, brandishing a meatball sub like a shining sword of justice ready to plunge it into my unsuspecting mouth. I couldn\'t let my guard down for a second. I\'d wake in the morning with inexplicable marinara on my chin and the taste of garlic on my tongue.   I also came to the realization that I can no longer really talk to my grandparents, I can only talk around them. It\'s not that there\'s some horrible generation gap separating us. My grandparents simply can\'t hear anymore. I\'m sure that over the course of a full week they only understood maybe 20 percent of what I said. I know they have a horribly inaccurate idea of what my life is like. I just got tired of correcting them. There\'s only so many times I can repeat something and there\'s only so loud I can talk. I just let them think what they like. It\'s easier on all of us. Here\'s an actual exchange:  Grandma: So, you do plays? Me: Well, actually it\'s sketch comedy. G: What? M: Sketch comedy? G: What? M: SKETCH COMEDY! Like Saturday Night Live. G: You perform on Saturday? M: Yeah, whatever. (I nod emphatically in surrender) G: What part do you play? M: I play a lot of parts? G: What? M: MANY PARTS!! G: What? M: An officer in the Queens Navy. G: Oh, a barber, how nice.  As if that wasn\'t bad enough, I had to put up with the most deafeningly loud television in the free world. My grandfather sits 4 feet from the television with what\'s left of his hair flapping to the decibels. And he still couldn\'t make half the stuff out. I couldn\'t escape the noise of the television. If you ever want to know what Bill O\'Reilly is talking about all you need to do is be within a quarter mile radius of my grandparent\'s house. You can\'t miss it. Eighties rock band Scorpions could be playing a concert in the back yard and I would never know.  So, while I missed out on the traditional New York experience. I feel that I got something much more important. I got a crash course on why I only visit my family once a year, and how much it sucks growing old.  Shalom   '),(11980,3496773,'20,June,2004','2004-06-20 00:00:00','    The other night I was sitting in my house and thinking about cartoons. For some reason the Smurfs popped in my head. Did you know they are from Belgium, like French Fries and those good waffles? Well, they are. In fact many well known cartoons are from there, the other popular one in the states was Tin Tin. You can get Smurf comics there in Flemish or Dutch. It\'s pretty cool.   Anyways, I got to thinking about the smurfs. And I knew that the smart one that made stuff was Handy Smurf, but I recalled one that wore glassesm which is of course the international sign for smarts. But I knew it wasn\'t Handy, who wore a pencil behind the ear.   So, I knew I could turn to the source of all information useless: the internet. I had faith in the fact that some person with nothing better to do had created a website that listed, with pictures, all of the smurfs. So I used google and put in Smurf List, and sure enough, there was such a list, and sublists under it. I quickly found that Brainy smurf was the one of which I was thinking. Glasses and a snooty attitude, like an associate professor.  So while some may think that the internet is good for school, which it isn\'t, and some may think it\'s only good for porn searchs, which it is. I also argue that there is a lot of useless crap out there to be known. And people seem to have webpages for it, and they pay for them and keep them updated. What sad and isolated lives we Americans lead. we can now talk instantly to people across the globe, but we won\'t leave the house and talk to the person next door.   Speaking of which, I have got to get outside. It\'s gorgeous! Right after I find some porn. Hahha  Kickin\' it, father\'s day style, Mike the Tall   '),(11981,3496773,'16,June,2004','2004-06-16 00:00:00','    While some of my readers, I know, are going to cringe in disgust, I figured I would throw this out there and see if anyone\'s interested.  I am throwing a Part for the President in July, and anyone who reads this is invited!     urlLink You may remember last April when I went to a Bush/Cheney Party in Greensburg.   Well I had so much fun that I decided I wanted to throw my own party this July.  There is a national party for the president day on Thursday, July 15th, and anyone involved will get a sweet conference call with VP Dick Cheney himself!  It\'s a sweet deal.    The reason I wanted to throw one, was basically because of what I witnessed at the first party I went to.  It was a fine party, very respectable, but it was definitely geared towards adults.  Me, Leroy and Teddy were the youngest guys there.  So I figured why not have a party aimed at younger folks, college kids and folks in their twenties.  A real \"party\" filled with loud music (Andrew W. K., not Toby Keith), a light atmosphere with some real humor and good times.  Instead of having it at a fancy hotel, I\'m having it at a brand new coffee house in Rochester, PA.  Instead of hiring a DJ to play stereotypical Republican country music, I\'m bringing my Mac and playing some true party music (regardless of affiliation).  (Assuming all goes to plan) We are going to have a conference call from the vice prez, a video tape from the president, and some other sweet activities and speeches.    This isn\'t to say that only republicans are invited.  If you are someone who isn\'t registered, and wants to learn more about the Grand Old Party (and you are smart enough to know that there is more to us than what the news and Michael Moore portray) then come on down and give it a shot.  Or if you are a democrat who wants to see what\'s going on, perhaps you aren\'t brainwashed by the MTV-culture into hating GW, or you just want to have some lively discussion, then check it out.    The GO-Party for the President is on Thursday, July 15th from 6:30pm until 9:30pm at the Blue Violet Cafe (Brighton Ave, Rochester PA).  For more information, post a comment!   '),(11982,3496773,'11,June,2004','2004-06-11 00:00:00','    In these turbulent times it is important that we all remember one simple immutable truth: celebrities are better than us. They are superior to us simple folk in every conceivable way. They are better looking. They are richer. They live in unnecessary, palatial \'cribs\'. They are sweeter smelling. They have entire television shows devoted to their fabulous lives. They are more well spoken. They are more intelligent. And, this must be remembered at all times above all else, their opinions are worth more than the those of the rest of us slack jawed, hamburger helper eating yokels.  So, in these fractured, contentious times let us all turn to the great oracle of Hollywood to gain guidance and understanding. They will tell us everything we will ever need to know, from what to wear to who to vote for. They will do all the thinking for the poor \'little people\' of our nation. This is a good thing since regular people have such a hard time thinking. We\'d all be much happier not thinking at all. After all, thinking gives regular people headaches, and we always make such a mess of ideas. Left to our devices we might do totally illogical things, like read books, express ourselves, or even vote for whoever we want. What a nightmare. Thank you Hollywood for sweet, sweet mind-numbing television to give celebrities a forum to tell us useless worms how the world works.  Now, I don\'t want to hear any complaints. That\'s just the last vestiges of your independent, but fatally flawed thought talking. Just let it slip away. Don\'t question celebrities. They\'re right. They have to be. They\'re celebrities. I know to us, the great unwashed masses, it may seem odd that we value someone\'s opinion simply because they happen to be famous and they happen to be famous simply because they are attractive, or can read someone else\'s words on cue, or because they have a pleasant voice. But it simply doesn\'t matter why someone\'s famous. As long as someone is famous they become our moral and intellectual superiors. That\'s just the way the world works. It\'s God\'s plan. Celebrities are the center of our world. We should all just recognize this and genuflect. Now, if you\'ll excuse me. I have to watch E!.  Shalom   '),(11983,3496773,'06,June,2004','2004-06-06 00:00:00','    I don\'t know why I keep telling people this, because I\'m certainly not proud of it... but I saw the movie \"Soul Plane.\"  Yeah, I know... I know.  It was a dumb thing to do, but the circumstances under which I saw the film made sense at the time (although they are certainly no excuse).  I had a bevy of free tickets for the Destinta theater out in Bridgeville, and they expired last week, so I had to go and use them up... I mainly saw movies that I had already seen, like \"The Punisher\" and \"Shrek 2\" and then, in a move that I can only attribute to a moment of weakness, poor timing, peer pressure, and desperation, I saw \"Soul Plane.\"  So far, I\'ve gotten the same reaction from everyone I\'ve admitted this to: \"Oh God, what were you thinking?!\"  And the answer is \"I have no clue.\"  The whole movie looked more like a train wreck from a mile away.  Have you seen the trailer?  It looked absolutely terrible... and I assure you, if you make the same mistake I did and actually take the time to watch the movie, your guess about the movie will be proven.  It was one of the worst movies I have ever seen.  And while I could sit here and critique this movie like I would any film, wasting time on the technical flaws of the movie would be pointless.  Just thinking about the content of this film would make anyone, of any race, just shake their heads.  In what had the potential to be the \"soulful\" version of the comedy classic \"Airplane!\", \"Soul Plane\" is nothing more than a 90 minute MTV-stereotyped, racially-charged, mess of a film that makes me wish I had the comforting arm of TV\'s George Lopez to cry on... that\'s pretty bad.  Let me start by talking about what goes on in the movie.  There is this giant purple plane with hydraulics (which is probably one of the few cool aspects of the movie) and it\'s flying with a bunch of stereotypes... and Tom Arnold.  I don\'t understand how any race could NOT be offended by this movie.  There is one Arab on the plane, and of course there are a few 9-11 jokes to go along with him.  There are four white people on the plane, they have the last name of \"Honkey\" and they are coming home from a vacation to \"Crackerland.\"  Come on now... and then of course the plane is filled with African Americans.  And for once I agree with Spike Lee and his joints-a-plenty.  This movie is most offensive towards blacks!  Sure there were a few times throughout the movie that I chuckled at what was going on on screen.  Perhaps it is just the overall faux pas nature of this racist piece of cinema (the lower class section of the plane gets a piece of Popeye\'s fried chicken for their in-flight meal), afterwards I felt a little weird about the whole movie.  What kind of audience are they making this for?    Well I can tell you.  It was made for the pop-culture-loving teens of this day and age.  And they are eating it up.  There is a reason why this awful movie is number 5 across the nation.  It\'s the same reason that they are making a movie called \"Harold and Kumar go to White Castle\" and a movie called \"The Cook Out.\"  Because America loves dumb shit.  America loves stereotypes and humor that, when attached using electrodes to a bowl of green Jello, barely causes ripples.  Now I saw this movie because I had a free ticket, and so technically they didn\'t make any money off of me and my buddies.  But someone out there is seeing this movie, and it\'s not good for America or Hollywood (who is going to do nothing but continue to make this kind of crap).    So if anyone asks... I saw Soul Plane to further my knowledge of cultural studies.  And because I had a free ticket.   '),(11984,3496773,'04,June,2004','2004-06-04 00:00:00','    There is something to be said for the Cellar Dwellers\' marketing and advertising practices.  Perhaps those words would be \"gorilla tactics\" or \"grassroots effort\" or maybe even \"word of mouth.\"  And then there is something to be said for the Beaver County Times\' treatment of the Cellar Dwellers\' effort.  Unfortunately, those words are a little too inappropriate for me to type on this very blog.    Anyone who lives in Beaver County knows the Times\' reputation for poor reporting, a strong reliance on AP news stories, a strong liberal slant, and a lot of typos.  You don\'t need me to complain about that.  And that\'s not to say that they don\'t have good people working for them.  Dave Smith is an exemplary entertainment writer, and at one point in time they had John Dessler, one of my inspirations for choosing graphic design.  However, they have managed to give the Cellar Dwellers the good ol\' one, two, screw at some point for almost every one of our shows for the past few months.  It\'s at the point now, where I recommend that our fans NOT look for our shows in there because the listing is going to be wrong regardless.  Two weeks ago, we were doing an all-improv show at Thursdays Restaurant in Bridgewater.  I am the guy in charge of making the fliers, press releases, and other promotional material for the group.  I always make sure the press releases are perfect, and that there is no way to confuse our location, show dates or times.  And yet, somehow the BCT thought we were at Dockers doing improv.  We\'ve never, in our troupe\'s seven year history, done a show at Dockers.  But I guess that\'s better than nothing.  Sometimes they just don\'t print a press release I send them, even when it\'s on time.  And of course I call them and let them know they are wrong every chance I get.  And yet, there isn\'t too much we can do, since it\'s a free service they, and all newspapers, provide.  We just have to take what we can get.    It was their screw-up today that has sent me over the edge.  Learning that they can barely understand press releases printed clearly in either Courier New or Times New Roman, I faxed them the Godzillas and Generals release TWICE, both times well before the deadline for this weekend\'s issue.  And yet somehow, through some mysterious voodoo curse or insistence on using \"copy & paste\" methods, they have managed to screw the Cellar Dwellers yet again.  Because fresh in today\'s Weekender, under the comedy section (I also insist on us being placed in the Theater Section but comedy is fine too... as long as we don\'t wind up under Etc.), it clearly says that our show is SATURDAY, aka TOMORROW.  It also says that it\'s scene-based improv.  I just have one web-jargon abbreviation to use right now: WTF.  Apparently, the BCT\'s A.D.H.D. has gotten the best of them again, and they just make these insane mistakes without even knowing it.    The Beaver County Times has a responsibility to their readers, advertisers, and the Arts in Beaver County to get these listings correct the first time around.  No amount of \"corrections\" the day after will fill the seats.  And for the Dwellers, the BCT is the best way for us to really get the word out about our shows.  This isn\'t to say that you should totally ignore what they print, because 70% of the time its probably right.  But I insist that if you see a Dweller listing, you check out the website, or look for a flier, or just go and find out.  Our fliers and website are always correct.  If you want something done right, you do it yourself.  But it\'s apparent that when you have to rely on others, you only get what you want when you are a paying customer.  I guess they won\'t get it right until we finally make the cover of the weekender.  Of course, as long as there are Emmy award nominations, we don\'t stand a chance.  Seeya tonight at the show.   '),(11985,3496773,'02,June,2004','2004-06-02 00:00:00','    As some of you may know, I spend quite a bit of time shuttling between my home in lovely Clintonville, PA - up in Venango county - and the Cellar Dweller home base in picturesque Beaver County. In my travels I have encountered myriad wondrous and astounding things, bad weather, scary hitch hikers, an Elvis impersonator truck driver, but I have encountered only one object which completely mystifies me.  Off of route 488, somewhere between Elwood City and Portersville, there stands a statue, maybe four foot tall, of a little Mexican leading a donkey. Alone in a small clearing, with absolutely no explanation, this Mexican is forever caught in the act of leading his donkey toward God-only-knows where. It\'s almost as though, this Mexican along with his donkey had sprung forth from the Earth whole. It stands there implacable, as mysterious as Stonehenge, as terrifying as an Easter Island head. I have the strangest sensation that no human mind or hand played any part in its construction. This statue exist on a plane beyond human comprehension. At least it exist beyond my own comprehension, which is - I am forced to admit - quite limited.  I do have my theories. I have several, they bubble up from the percolator of my mind. I think that some ancient culture of Lawrence County could have used this statue as a calendar or possibly it held some astrological purpose. It could have fallen from space, like David Bowie did in the 70\'s. It could be a gift to us from the mole people who live thousands of feet beneath the Earth. My favorite explanation is that it is an idol from a long dead religion - the Valdezians. The Valdezians worshipped all things coffee - mugs, plastic stirrers, half and half, pretentious bastards at Starbucks - and this is an idol to their chief god - Juan Valdez, the bringer of the coffee. Sure I have no proof of this, but I think if I just believe in it enough it just might become so.  One of these days - when I am not driving home at some obscene time of night or running late - I will stop my car and take a closer look at this marvel of ancient engineering. If I ever have anyone else with me and a camera handy I\'ll take a picture with it. I\'d like one with me riding the donkey in my fancy Cellar Dweller T-shirt. I just hope that I don\'t upset Valdez. He is a jealous god and will not stand for mocking. He can make my coffee taste horrible for the rest of my life.  Shalom   '),(11986,3496773,'01,June,2004','2004-06-01 00:00:00','    For the past month or so, Christoph and I have been meeting once or twice a week at Cafe Kolache in Beaver to work on the website.  Kolache is an amazing little coffee house, something I have been clamoring for for years here in the BC.  As I have mentioned, we are \"hard\" at work on the new site, which is going to be all in Flash.  It\'s pretty sexy, if I do say so myself.  Chris went out and got all of these books and whatnot and took this sucker on.    Of course, the two of us (along with any other Dweller who may join us) sitting in a cafe with Chris\' 17-inch HP laptop aren\'t the most productive around.  But productivity with this website isn\'t always our number one priority, fun is.  Because if fun ain\'t being had, then why be had at all.  That\'s what I say... sometimes.  And so we sit there, me drawing little designs and pictures in my sketch pad, him clicking away at Action Scripts and Motion Guides, and the two of us talking about the finer things in life: politics, computers, girls, coffee, jobs, and The Tick.  All things relevant to the life of two single working-guys fresh outta high school.    My coffee intake has nearly doubled thanks to working on this website.  The advantage, and the main reason, of doing most of the work at the cafe is because we get a bottomless cup of coffee.  So our laps are scalded by the end of the day.  No, so we get all of the coffee we can muster!  Trust me, I drink alot of coffee.  Usually it\'s the dark roast (which can be anything from Kenyan, Columbian, and Cuban to J.P. Hearty, AA Kenyan, or M.C. Hammer... I wish there was an M.C. Hammer coffee) with a ton of half and half and sugar.  Chris likes to get a carmel coffee thing (some of the girls there make it with whipped cream, while others skimp on the extras and just give him the coffee) and then he chases that down with some of the normal coffees.  I usually go to the bathroom like three times while we are there (which is normally from 3:30 until they close at 5:30).    It would help alot more if that place had a wireless internet connection like the Copper Dog does.  I mean, it\'s not enough that we steal their electricity by plugging the laptop into their wall, we want them to pay for internet as well!  There are usually other folks in there with laptops (a good number of them have Apple iBooks... REPRESENT!) and they are all kinda put in their place by Chris\' monster computer, which could probably wrestle Deep Blue and Hal at the same time.  Sure those other guys were advanced in their time, but nowadays a centrino Pentium machine could whoop their butts.... sorry.  Anyways, the website is gonna rock harder than Donny Iris\' love.  But since he loves us as well, we kinda win on both fronts.   '),(11987,3496773,'25,July,2004','2004-07-25 00:00:00','    I was flipping through the channels on the old television - not an all together unusual activity - when I chanced upon something interesting on C-SPAN. (Yes, I am&nbsp; aware that the words C-SPAN and interesting do not generally appear together in any context, but technically this was C-SPAN2.) They were airing a meeting of a focus group from somewhere in Ohio. (OK, I admit this still doesn\'t sound interesting, but bare with me.) So, there were about 9 people in a non-descript meeting room of some sort being quizzed by a weasily, condescending focusmeister. (Well, maybe this is only interesting to me, but I\'m forging ahead anyway.)   This focusmeister - sorry about the overly technical jargon - was asking all sorts of questions about Bush and Kerry to this rag tag assembly of housewives, retirees, and guys with goatees. This was so fascinating to me because this was a window into the strange and arcane tools used by candidates to gain political office. No doubt, Bush, Kerry, and their people use focus groups such as this to tailor their platforms, speeches, and such. After viewing this focus group, I only have one thing to say about the state of the democratic process in our country: It is totally screwed up.   If Bush and Kerry are listening to people like this, God help us all. These people are complete idiots. They almost all complained that both candidates seem to avoid really confronting the&nbsp;issues, but what do you expect when they hear from focus groups like this is that Bush supporters like him for his \'strength and confidence\' while&nbsp;Kerry supporters think he\'s \'careful and thoughtful\'. Why would the candidates deal with the issues, when focus groups are telling them that they\'re more interested in their images?   Then, there were the questions coming from the focusmeister. There was, of course, the old \' who looks more presidential?\' chestnut. The correct answer to this is: who cares? It doesn\'t matter who looks more presidential. Neither Bush nor Kerry does. One has a face that looks like it\'s melting. The other\'s face appears to be constantly sucked toward some midpoint, maybe around his nose. (Curiously enough, if you crossed Bush with Kerry, you\'d get one normal looking person.) Plus, it doesn\'t really matter who looks more presidential. It doesn\'t even matter who acts more presidential. The only question that matters is who would make the better president. Plus, asking who\'s more presidential between two candidates is akin to asking which is more tree-like: a pine cone or a buckeye. Both have the capability to be a tree, and once a tree, it becomes de facto tree like. It can be nothing but tree like, since it is, after all, a tree. The candidate who becomes president becomes presidential simply by being president, since by definition he can not be otherwise.   The questions from the focusmeister just kept getting stranger and stranger. At one point he asked, \"If John Edwards was a family member what family member would he be and why?\" Almost everyone answered Uncle - one guy&nbsp;cryptically answered \'an uncle from the coast\' with out giving any explanation. Again, no one gave the correct answer which would be to point out that the question is stupid and not worthy of a serious answer - although, if forced to answer I\'d have to say Uncle Joey since Edwards, much like Uncle Joey, is just a little too eager to please.   At about this point my attention began to stray from the focus group on television. I began to wonder why the focusmeister didn\'t just go whole hog and continue asking more and more bizarre questions. It would&nbsp;make for&nbsp;better television and be just as useful to society to just screw with the people, just ask off the wall crap, dare someone to call you on the questions. As a matter of fact, I have a list of questions I would just love to ask a focus group. Here\'s just a few:   If Dick Cheney were a candy bar, what kind would he be and why? - Hopefully this question would lead to a Bush supporter saying something like: \"M&M\'s because he may have a hard shell, but deep down I think he\'s a real sweety.\" To which a Kerry supporter would respond: \"Technically M&amp;M\'s aren\'t candy  bars , nitwit.\" This would lead to loud shouting and maybe fisticuffs, which would make for really good television.   If you could ride one ride at the carnival with John Kerry, which would it be and why? - The correct answer is his jowls.   If John Edwards ran one attraction at this carnival, which would it be and why? -&nbsp;Hopefully someone would say: \"None, he\'d set up a lawyer booth to sue when one of the rides comes crashing down.\"   Which word better describes George Bush: lava or magma? - Even I\'m mystified by this question and I came up with it.   Which Olsen Twin are you more attracted to? - Just to see if everyone\'s still awake.   If Dick Cheney could have dinner with any person, living or dead, who would it be and why? - A good cardiologist for obvious reasons.   Is John Edwards more of a Dick York or a Dick Sergeant?&nbsp;- Either way, he\'s still second banana.   True or false: George Bush totally rules at pack man. - Who knows? He might.   You\'ve been invited to John Kerry\'s birthday party, what do you bring as a gift? - I don\'t know. What do you get the man who married the woman who has everything?   See now wouldn\'t this be fun. I think that I may have found my calling. I should run focus groups. I\'d be just as useful, but at least twice as entertaining as the guys they have running them now. All I\'d really need to do is keep a straight face, and act like I know what I\'m doing. That doesn\'t sound hard at all. How do you think I got through college?   Shalom   '),(11988,3496773,'21,July,2004','2004-07-21 00:00:00','    I watch alot of day time television.  I\'m one to just sit down and click away, but I definitely have it on in the background, as my summer days grind away in the most boring fashion available to me.  And there are one group of commercials that not only succeed in making me laugh every time, but also intrigue me far more than they should.  These commercials are for entrepreneurial (a word SHU seems to love for some reason) credit machines that do all the work and give you all the cash.  Not only are they perfect for the wanna-be millionare inside of us all, but they also require us to do nothing but suck in the sweet sweet cash!  You may have seen the commercials yourself.  These wonder machines include ATM machines in fast food chains, internet access points, wireless hotspots, and movie rental machines.  That\'s right, you can rent movies... from a vending machine.  Apparently, these machines do nothing but rake in the dough.  All you do is pick a location, set up the machine, and then stop by every once in a while to get your loot.  It begs the question, \"Why aren\'t we all doing this?!  Why am I going to college?  Why am I even working on graphic design, when all I have to do is get a key, walk down to my giant movie-renting-machine and get my loot?\"  This operation is literally the deus ex machina that will come and end my working career (which has been lazy as best thanks to my Trib job).    What\'s even better about all of this, is the lifestyle I will have once I get my machines.  According to the commercials, I will become an old man with a full head of gray hair.  I will get to wear Hawaiian shirts while sitting by a huge pool, and have an attractive girl in a bikini look over my shoulder as I monitor my money from an old Apple iBook!  Holy shit!  This is the life of a king, I tell you... a king!  I know this is exactly how Bill Gates and Ted Turner live.  I just know it.  And soon, I shall be with them...  well, not WITH them, but in the same class as them.  Getting taxed less (hopefully) than the average man, but instead of giving out more jobs like most billionaires, I will make my billions from the machine.  Perhaps I will hire someone to go around and pick up the money for me... a small Cuban boy.  That way I won\'t look like an American jerk.  I\'m giving back to the community, and letting Pepe\' get some exercise.  And of course, thanks to Global Link (the enigmatic company behind these marvelous machines) and my sex-slave Adonna, I will be the mysterious billionaire that I always strived to be.  I will have no choice but to become a super villain, hiring thousands of nameless henchmen, and that guy with the question mark suit who is always being chased by the feds, and I will take over the world in my zeppelin.  That\'s right... a zeppelin.    But there is just one loose end that needs tied up...  and that is of this mysterious video rental machine.  I just don\'t get it.  Somehow the movie titles will be updated (I\'m assuming by one of my drones) and somehow people are going to return these things?  If they don\'t, Pepe\' will have to break their legs.  But really, is this more convenient than going to Hollywood Video or Blockbuster?  No matter, money will come regardless!  If you too are interested in the exciting business of owning a machine that takes money from people, contact  urlLink Global Link .  They have a helpful website filled with confusing words and charts depicting nonsense!   It\'s like the website says: \"Global Link was launched in 1996 with a simple and yet challenging vision: to look at the world from the desktop of the buy-side and to create efficiencies through the intelligent application of technology to real-world business needs for information, liquidity access, and workflow improvements. In short, the mission is to deliver one \'window\' through which the buy-side can access unique fact-based research and trade any asset class, anywhere, with anyone. In close collaboration with our clients, we are progressively delivering each stage of our vision.\"   Hell yes.  Hell yes...   '),(11989,3496773,'19,July,2004','2004-07-19 00:00:00','    By most accounts - some could even argue by all accounts - my life is far from successful. My problems are many and varied. I am with out gainful employment. My hairline is receding. I live with my parents. I have no luck with women - this, of course, could be a result of the rest of my problems. I am very clumsy, often annoying, and love bad hair bands of the 1980\'s - however, I can\'t stand good hair bands of the 1980\'s, although I can\'t say I\'ve ever heard of any.  &nbsp;  Now, I am aware that I do have some good qualities. Those with an incredible gift of perception may even be able to notice some of them. I am fairly well read, more or less intelligent and occasionally amusing. I\'m a good listener, in many ways unselfconscious, and I make great omelets. Also, I love bad hair bands of the 1980\'s.  &nbsp;  My major problem, I have decided, is that I lack a certain force of will, strength of commitment, courage, intestinal fortitude, balls, etc. I live too much of my life inside of my own head. I lack the strength, desire, or bravery to move ideas out of the intellectual realm and into the physical, everyday realm. I cannot realize my full potential, because - through either fear or apathy - I cannot make myself fully responsive to the world at large. I need to move beyond myself. Take the strength of my mind and apply it to the world. I need to develop my intellectual and social courage in much the same way I have developed my body from flabby to almost ok. I need to gain some strength. I need machismo. I need to jump out of my seat, run out into the street with the crazed look of a blind man who can suddenly see. I need to experience and I need to do. I need to grab life by the lapels, shake it around, and generally make it my woman. In short, I need a mustache.  &nbsp;  Ah, yes, the mustache: the facial hair of the man of action. Mustaches practically breath decisiveness. Quick, think of the mustache. Who comes to mind? Clark Gable. Magnum PI. Rollie Fingers. Cowboys. Every hero of every 70\'s cop movie or television show. Barber shop quartets. Porn stars. Groucho Marx. Every state trouper. All men of action. Men who know who they are, know what they want, and get it. That is the power of the mustache.  &nbsp;  So, is it any wonder that I, weak willed as I am, long for a mustache, would treasure a mustache, would comb it thrice daily, would wax the ends into intricate curly-Q\'s. I quake with desire for a mustache. I have started growing one. I picture it in my minds eye, thick and full, flowing down the corners of my mouth, covering my upper lip in a luxurious facial hair canopy of manhood. I see myself on job interviews, employers falling all over themselves to have me be a part of their company. \"By George,\" they\'ll say, \"just take a gander at that magnificent mustache. We must employ that man. Quick get the benefits package together.\" I can see myself at the bars and the clubs. I can see the women swooning over me. \"By George,\" they\'ll say, \"look at that mustache. He is clearly more man than the rest of these bare lipped fools. Quick, we must go lavish feminine attention upon him.\"  &nbsp;  I am growing my mustache. It is currently little more than stubble at the moment. I am sure that I will have to withstand the horrible, \'I think a caterpillar died on your upper lip\' stage of mustache growth, the stage which most men cannot progress past. I know that some people are going to sneer and mock my mustache. \"By George,\" they\'ll say, \"You look like a pervert (or a pedaphile, or a member of the village people, or Teddy Roosevelt).\" I can handle this. I will simply laugh it off, because I know - deep down I just know - that mustaches are kick ass, and those losers are only jealous of the power and the majesty that is the mustache.  &nbsp;  I also know that some people are skeptical. Not so much toward the power of the mustache - which is undeniable - but toward my ability to handle responsibly the greatness of the mustache. These people may argue that you must first have the mustache attitude then the mustache follows naturally. That to grow the mustache and then hope for the attitude is a fools errand. \"By George,\" they are probably saying, \"this chap\'s mustache will surely wither and die upon his lip. The attitude predicates the mustache, not vice verse.\" Now, this is a very interesting question. It has a number of philosophical facets dealing as it does with issues of causality, individuality, and personality. So, I must be careful in addressing this objection. However, I feel that I am up to the challenge. So, to these people I say: EAT IT!! I\'m growing a mustache and that\'s all there is to it. The mustache is going to be kick ass and so am I. If you losers can\'t handle that, you better shut up and get out of my way. Because, by George, I\'m growing a mustache and I\'m coming through. There is no stopping me now. Nothing can stand in the way of the mustache. Nothing......Unless, of course, my mom doesn\'t like it.  &nbsp;  Shalom   '),(11990,3496773,'18,July,2004','2004-07-18 00:00:00','    My brother has been feeding me these ridiculous-but-true facts all day now (I am guessing he was literally saving them for a rainy day)... and there was one that really stuck out in my mind.  That of the British-Zanzibar War, the shortest war in history.  So I decided to google this so-called war to find out more about it.  What I found was an intriguing journal entry written by the incredibly mysterious Sultan of Zanzibar.  What follows is a series of fateful entries written on the day of the war.  August 22, 6am  Dearest Journal, today I have been appointed the SULTAN of ZANZIBAR!  How sweet is that?  Of course, given the circumstances of my appointment, I must take this job with merely a grin instead of a full-blown crescent moon smile.  My father, the long-time Sultan Tad Storkensburg, died last night of irritable bowel syndrome.  I think there was something in his Zanzibarian Burger that really disagreed with him.  Alas, I mourn my father\'s gruesome and disgusting death, and I welcome my new title of Sultan!  It\'s so much cooler than \"Saltine.\"    My first act as Sultan is to stand up to those awful Brits.  The Brits and the Germans have been fighting over my island for years now, and I\'m damn sick and tired of it.  The Brits keep wanting this island so they can move into Africa, the Germans want this island so they can put a Volkswagon factory on it.  Zanzibar is my homeland, as obscure and ridiculous as that sounds, and I will not let it slip into the hands of the straightbacks from the north!  So, after speaking with Germany, they\'ve decided to back me up as I stand up to the Brits.  My dad had good relations with those limey bastards for years, but now that he\'s out of the way I\'m gonna kick some ass!  August 22, 1896 9am  Dearest journal, I just declared war on the Brits.  I know what you\'re thinking... I\'m in office for three hours and I already declared war on the biggest nation in the world.  Maybe I was a little hasty, but I stand by my decision.  I declared war about twenty minutes ago, and I\'m ready for whatever they are going to throw at me!  I have 2,500 Arab troops ready and willing to be slaughtered for the well-being of Zanzibar.  I also have this really boss gold cannon from 1658.  Back in the day, I hear, we used to use this cannon for everything... we even fought off the Egyptians or something.  But ever since then it\'s been sitting in front of the Zanzibarian Court House just collecting dust.  I also got this sweet old ship...  it\'s like a wooden viking boat.  Only it\'s not a viking boat because vikings didn\'t come down here... often.  August 22, 1896 9:05am  Dear J-Dog, The British are pretty much surrounding the island with their large-ass ships.  Any normal Sultan would be thinking he had made a mistake, but I know I\'m doing the right thing.  I will be the Sultan to end all Sultans!  Let it be known that Reggie Storkensburg, The Sultan of Zanzibar, was the most courageous leader ever.  With the exception of Jesus.  Anyways, right now we are sort of waiting... my troops are just sitting on the beaches playing card games and volleyball and the British keep cocking their rifles at us.  I hope something happens pretty soon, I have a tennis date with the Czar of Uberbeckingstand at noon.  Did I spell his country right?  Oh well, no one\'s gonna read this anyways.  August 22, 1896 9:15am  Shit!  Snap!  Dag!  Holy flaming crap!  The Brits have decided to start shooting cannon balls at my island!  One shot took out 500 of my troops, who were in the middle of a round robin volleyball tournament on the north shore!  I have decided to bring out my golden cannon, which I have named \"Fat Man.\"  I have three cannon balls, which I have named \"Little Boy.\"  Those Brits are going to have some trouble coming there way as soon as Fat Man shoots the Little Boys!    August 22, 1896 9:25am  This isn\'t going at all like I had planned.    August 22, 1896 9:35am  Okay, okay, okay, all of my troops are dead.  That\'s what I get for hiring all these Arabs to do the job... everyone knows Arabs can\'t swim!  My viking ship has been totally destroyed, even worse they spray painted a giant space gun on the side of it before they set it on fire.  Those British sure do know how to make you feel stupid.  Although I must stay that they aren\'t nearly as polite as everyone thinks.  Let\'s just hope my cannon holds out!  August 22, 1896 9:37am  Murphy\'s Law kicked in... my uber-boss cannon just burst into flames!  It\'s made of gold, how did this happen?!  August 22, 1896 9:45am  Dear Journal, you are now looking at the former Sultan of Zanzibar... that\'s right.  I gave up.  I couldn\'t take it anymore!  The Brits had me backed up against the wall here.  And so after 45 minutes of being pummeled by these jerks, I gave up.  I kinda forgot that my population consisted of just 2,500 Arabs, and since they were all dead now... I figured I might as well cut my losses and live to rule another day.  Luckily, my sausage-welding friends from Middle Europe, let me stay at their consulate.  Those Germans are alright, ya know?  I mean, they really know what they are doing, and I wouldn\'t be surprised if they really make a difference in the world.  I\'d stake my Sultan reputation on it.    Anyways, goodbye my lovely journal... the Germans won\'t allow me to bring this into the consulate, so I must leave it on this board and float it out to sea.  Maybe someday those Brits will find it and see all of the stuff I said about em.  Won\'t they feel bad!  -33-  Wow.  That Sultan was a cool dude.   '),(11991,3496773,'13,July,2004','2004-07-13 00:00:00','    Those of you who keep up with the news - I know you\'re out there somewhere - will most likely remember the big story about the Super Strong Mutant German Boy a few weeks back. For those of you who only get your news from \'Entertainment Tonight\' - I know you\'re out there everywhere - allow me to enlighten you. It was recently revealed that a boy in Germany (I want to say he\'s currently 5 years old) was born with a genetic mutation - yes, I said genetic mutation - which has caused him to grow muscles at a prodigious, unnatural rate. He can already lift more weight than some full grown adults. He could probably lift two Mary-Kate Olsens straight over his head, but only one and a half Ashley Olsens. So, the Super Strong Mutant Boy - who I will from this point on refer to as SSMG boy, or better yet, as SSMuG Boy, or even better still, smug boy - being a true genetic freak, was paraded around the news for attention.  Now that a few weeks have gone by and most of the news has moved on to new stories, something about smug boy still bothers me. No, it\'s not that there\'s a five year old out there who can kick my ass. I\'ve been getting my ass kicked by five year olds for years. It no longer bothers me. What bothers me is how the whole thing was covered. It was seen as a medical, scientific miracle. Something to be studied. Maybe this boy\'s DNA could unlock keys to healing those with debilitating diseases, or lead to a safer alternative to steroid therapy. While this is all good and wonderful, I can\'t help thinking the news media has missed the main story. This kid is a genetic mutant - a MUTANT. Doesn\'t anyone understand what this means? The age of the X-Men could very well be upon us. This kid could be a member of the first generation of genuine, real super heroes and super villains. Smug boy\'s a German so, obviously, he\'s destined to be a super villain.  This is huge, and the news media completely dropped the ball. We could be heading toward a war of humans vs. Mutants, or humans and good mutants vs. evil mutants, or good humans/mutants vs. evil humans/mutants, or, most frightening yet, everybody vs. evil shepherds. This could be the most dynamic time in human history as super powerful mutants with a proclivity toward spandex rise to power.  Still, I don\'t want to rush to any judgment. That\'s why I didn\'t write this post weeks ago when the story was fresh. I needed to take some time and see if I could confirm my suspicions. If the age of the X-men is truly upon us, then other mutants must be out there within a generation or two of smug boy. Surely, at least some of these people are already adults and have used their powers to move into a position of power and fame. All I need is to use my keen power of perception and my television remote to weed them out.   I want everyone to sit down. It is time for me to announce my findings. After days of study including extensive reading, monitoring of television broadcast, and three square meals a day, I have determined that the age of the mutant is indeed upon us. In fact, it is much more extensive than I first suspected.  My first tip was Ken Jennings. Ken Jennings, for those of you not in the know, is the man who has been Jeopardy champion for almost 30 days. I\'ve been monitoring him with much interest. He is almost unbeatable. His trivial knowledge and thumb reflexes are clearly superior to the average man\'s. Clearly, Ken Jennings is a mutant who was born with an abnormally large Trivial Cortex - the part of your brain set aside for facts on the court of King Edward II and other such tid bits - as well as genetically advanced thumbflexes - sorry about the technical terminology; it\'s unavoidable. So, now the only remaining question is whether or not Ken Jennings is a super hero or a super villain. Let\'s look at the facts: ken Jennings is Super intelligent, kind of weasely looking, and Mormon. Clearly, he\'s a super villain.  Also, there\'s Saddam Hussein. He\'s been in the news a lot due to his trial in which he has exhibited mutant like abilities. In the face of a criminal court, with a possible death penalty in the mix, Saddam had the guts to defend his actions. He even claimed to still be the president of Iraq. Clearly, Saddam has been born with large - unnaturally, mutantly large - balls. Good thing we already have him in custody. Also, the whole genocidal mad man thing kind of indicates he just might be a super villain, not a super hero.  So, I found 3 mutants, then last night I hit the motherlode: VH1\'s \'I love the 90\'s\'. I\'m sure everyone\'s familiar with the the VH1 specials. You may think they\'re harmless or you may think they\'re annoying, but - after my discovery - everyone\'s going to find them terrifying. All the commentators on \'I love the 90\'s\' are mutants. Every last one of them from the annoying guy from \'Talk Soup\' to the annoying guy from \'Ed\'. This is the first evidence I have found of a conglomerate of mutants. Not only are they all mutants, they\'re all mutants in the same way. They have all been born with the uncanny - that\'s right, uncanny - ability to maintain a career in entertainment with no discernible talent. Clearly, all these people were born with a genetically advanced Zsa Zsa gland - or what scientist often refer to today as the Keanu/Shore gland. This gland controls dumb luck and the ability to be on television with nothing to say. Most frightening of all, these celebrity mutants are all super villains. It\'s obvious. They don\'t have an airplane. All teams of super heroes have a tricked out plane. In absence of a plane, it becomes obvious that this mutant team is truly evil.  Finally, I feel the need to take a closer look at myself. What if I am a mutant? It\'s worth a look. Let\'s see. I\'m not super strong. I have decent hearing, but not mutant hearing. I have bad eyesight. My sense of smell is average. I can\'t shoot any kind of laser from my eyes. When I touch other people the only thing that happens is I get slapped. After jumping off my garage I\'ve confirmed that I can\'t fly. My bones do break. I\'m balding. I\'m unemployed. I live with my parents. I have no real drive to accomplish anything in my life... Oh, my god... I am a mutant. Can\'t you see? I\'ve been born with a genetically superior apathymous. I have powers of apathy far beyond that of the average person. The best part, I\'m not a super hero or a super villain. You actually have to care to be labeled either of those. I\'ve been born with the only great power which does not come with great responsibility. I can live with that...I guess...whatever.  Shalom   '),(11992,3496773,'12,July,2004','2004-07-12 00:00:00','    Tomorrow I turn the ripe old age of 19...  a year that just doesn\'t seem as special as the past three.  Think back, have you ever heard anyone say, \"I can\'t wait to turn 19!\"  It\'s the minor rift after the big waves caused by birthdays 16-18.  These milestones, to which like all kids, have always held things that I have always looked forwards to.  When you turn 16, you can drive; when you turn 17, you can see R-Rated movies; when you turn 18, you can do all sorts of things; but when you turn 19... well.    Turning 19 not only fails to unlock any new abilities (if you want to look at this like it\'s a videogame) but it also leaves you, at least in my family, poor and lonely.  In my family, you have birthday parties until you are 12, and you get presents from your relatives until you are 18.  After that you are considered an \"adult\" and subject to moneyless cards, a birthday dinner and a pat on the back.  That isn\'t to say I\'m some sort of ungrateful punk... but knowing that nothing special is happening makes my birthday, once a sacred event that balanced my year with Christmas, is now a holiday about as important as Arbor Day or Boxing Day in Canada.    They (society?) really should have spaced out everything you can do when you turn 18.  Is it really necessary to make the age of 18 such a gateway to the world, when 19 and 20 have absolutely nothing special going on.  Since turning 18, I\'ve been able to drive after midnight , buy lottery tickets, get into strip clubs, buy pornography, purchase tobacco products, drink and gamble on international waters, drive out of state, and date people only older than 18 (which made for a few laughs when I was dating Aubrey).  But aside from driving out of state and driving past 11PM, I haven\'t really taken advantage of my 18 year old status.  So why not push the porn related abilities back to 20, and the lottery tickets and tobacco back until I\'m 19.  Then I have something to look forward to each year, although that 20th birthday would be sort of weird.    That, or they can just make other things off limits, and have them be available to me after I turn 19.  Like purchasing computer parts.  You don\'t really need to do dabble with the inner workings of your computer until you are out of high school.  Ban eating peas until you are 19, so you are mature enough not to scatter them all over your plate and dining room floor.  You also shouldn\'t be allowed to become a pop idol until you are 19, nor are you allowed to become a sex symbol until that time (I\'ve decided on waiting until I\'m 25 to do either of those things... personal rule.)  And so by now you should be feeling pretty bad for me and my pitiful age of 19... but don\'t!  My birthday is actually going to be a good day, I assure you!  I have decided to overcome adversity and ignore the fact that I can\'t do anything new during my 19th year on this earth.  So instead I am getting fed all day.  Tomorrow I plan on waking up, pouring some water on the floor for my homies, and heading over to my beloved Aunt\'s apartment for french toast.  Then it\'s off to lunch with a friend... and then coffee with another friend.  No sooner will I arrive home will I be eating some more, this time it\'s my birthday dinner cooked by Mom.  So if anything, maybe I will gain some weight the day I turn 19!    I\'m very thankful for my 19 years, and I\'m only sort of thankful for looking like I\'ve only been around for 13 years.  But hey, at least now I can buy lottery tickets and eat peas!   '),(11993,3496773,'12,July,2004','2004-07-12 00:00:00','    Today more than one bird flew into the screen that I have in my bedroom window. My room is not that interesting, nor does it look appealing to birds (no worms or bird porn). Now I realize I\'ve blogged a lot, so that means I\'m bored and need a real job, just like my mom keeps telling me.   If there is anyone out there who knows about birds, maybe they can tell me why birds seem to flock to the roof of my front porch. It\'s very strange. I live in the middle house of five row houses. But the birds hang out on the edge of the roof of my roof only. All the roofs are red. They all look the same. In fact, they are all beveled the same, so what makes my roof different?  I know my house is a party house. It\'s had a history of good parties, not least of which was our halloween party last year. So maybe there\'s a good bird vibe on the porch roof. The thing that kinda creeps me out is the fact that birds come from all over the neighborhood to do it on our roof. I\'m serious! I just noticed something different about our roof. Since we\'re in the middle, we have a little wooden knob sticking up in the center. And I just remembered seeing birds hanging out there a lot. Maybe that piece of wood is like the bird throne. And whoever sits on it gets all the chicks...literally.   Now I\'m no ornithilogical Dr. Ruth, but maybe there\'s something to that place on our roof. I don\'t watch the birds having sex, but on more than one occasion, I\'ve been awakened to loud bird noises close to my head and I slowly open my eyes to witness hot heavy freaky bird lovin\' right on my roof.   That\'s not as bad as the night my bed was right by the window and my head was right by the window. I heard a noise and turned over and I was looking a racoon square in the face. We both screamed and he took off.   I see birds come over and chirp and then other birds join it and the sex is on!  I wish there was a human equivilent to saying \"Hey, I\'m on the bird throne, come over here so I can do you.\" Life would be a lot simpler.   '),(11994,3496773,'10,July,2004','2004-07-10 00:00:00','    Now, most of my blogs are indeed not funny, just sort of ranting. Well this one is no different. I have a friend that went back to school. He\'s a little older than me and last night at about 1 AM we had a discussion on the computer about politics. He went to an \"We hate Bush\" kind of rally in Oakland. He said I would have not liked it since I was \"one of those..you know Rs.\" (republicans, which seems to now be the party of the people strangely enough from what the news says. They want to give the power to you instead of taking it and letting decisions be made by the government, cause apparently the democrats know what I need more than I do! But thatâ€™s another argument)  He then proceeded to say that the whole system needs to be changed. I was like \"What? What? What?\" a la Kyle\'s mom on South Park. He said democracy exploited people. And I laughed and asked what system does not? Democracy is the longest running and most successful political experiment in all of the world\'s history. He argued that monarchies were longer, so I had to point out that every time a new king or queen takes the throne, it\'s a whole different government. We\'ve had the same government for centuries now. How awesome is that? It may not be perfect, but it\'s the best there is.   I asked him for alternatives. He said ... socialism. Which of course can\'t work in a country as large in land and population as the US. There are places like Norway that have socialistic tendencies and they work for the small freezing country of the hot, tall, and all super white.   He said you just have to have enough people to want it, thatâ€™s all. Thatâ€™s all? Thatâ€™s the hardest thing of all and thatâ€™s why it wonâ€™t work. Cuba has been trying for a long time to be a communist state. They have a small population and small amount of land to control, and THEY canâ€™t even get their shit together. There are less minds to change there than anywhere else.  This is why communism is absolute horseshit.  Hereâ€™s an example: Youâ€™re at a store. The clerk says, â€œIâ€™ll give whoever gets to the counter first all their stuff for free, or you can both get here at the same time and Iâ€™ll give you 50% off.â€ Now, some people would want to work together, but most people would be like â€œscrew himâ€ and try and get free stuff. We are ready to screw the other person for personal gain. There can be none of those people at all for communism or true socialism to work. Personally, Iâ€™m glad to be one of those greedy bastards. Thatâ€™s what makes our economy run: the search for more for me.   I said that for any step towards a social society, you have to change the hearts of men and women, not the government. Government and indeed all sovereignty is granted by the people. So unless all the people want it, it doesnâ€™t happen. You canâ€™t change peopleâ€™s heart with force. That will make them resent you and hate your happy, feel good government.   He said it takes time. I told him thatâ€™s impossible. For as many as you change to your mindset, there will be changing back, or switching to the other side for the first time. The only way to change it is all together with a Marxist revolution (to see a great example of what that is, watch the movie ANTZ). You get rid of the people that donâ€™t agree. The workers rise and revolt. And the world becomes a time share, kibbutz style.   Communism is great on paper. Socialism seems even workable on paper. But there is one problem: people tend to be jerks. You go back to Marx and Rousseau and they think people are naturally good. But there is a whole other camp of people that think people are naturally bad. Iâ€™m in the latter simply because if we were meant to be good, itâ€™d be a lot easier to be good than bad.   So while communism is a fun thing to talk about in college, and even a nice daydream. Itâ€™s never going to happen cause someone will always take advantage of the situation. Read Animal Farm by Orwell. All there needs to be in one person to screw it up. With democracy, if one person is screwing up, the people kick his or her ass to the curb. Thatâ€™s why we need to vote. You do have the power, like He Man. People discount that power, but itâ€™s tough, and right now, mostly old people use it.   So get out of your peace and love hippy haze, your happy land of suger plum trees and lollipop lakes and wake up and smell the crusty burnt 7-11 coffee at the bottom of the pot. Democracy is ever changing and slightly flawed but anything else just doesnâ€™t make sense. So yes, communism is horse shit.   '),(11995,3496773,'09,July,2004','2004-07-09 00:00:00','    Now I know some people see me as the epitome of whiteness. I\'m clearly on the bottom rung of middle class. I\'m tall, well liked, can go to the bank and find I have good credit, went to college, did fairly well.  But yet there\'s something else. I mean I severely attracted to women of color. My last two girlfriends were black and puerto rican. In my mind, this begs the question: Am I really as white as people think I am? I mean I can dance. I can understand and use correctly almost all hip-hop slang. I have a very Jewish sense of humor. I love soul food and tex mex. And I adore both Chinese food and little children (Let\'s face it, they are the cutest.)  Maybe this has happened because of my life experience. I mean I did work with a hugely diverse population when I was in Austria. And now I work almost exclusively with African American children and am often the only non-black person in the room. But I liked southern food well before that. Although, thank God for my discovering greens. If you have never had them.....mmmmmm. It\'s a vegetable that tastes like meat. Can it get better? No, I think not. Plus, you can\'t work in Shadyside with kids and parents and not have a little Jewishness rub off. After my second summer there, \"Oy vey\" became part of my regular conversation.   Of course, it might also be because I\'m so German. We got fatty foods and mass genocide to look back on, and no one\'s smiling at that. I do enjoy a good polka or Weiner Shnitzel every now and again, and I was glad for the Austria experience to catch a little of my ancestry.   I think what\'s really going on is I\'m part of a new melting pot in America that is strictly culteral. Our hearts are melting and we\'re sharing customs, foods, games, whole cultures. Plus, we have a new world culture that is based on the internet, video games, anime, and cell phones. It\'s a weird esperonto-like phenomenon that\'s working. the shorthand of instant messaging is becoming a new form of spelling. Kindegarteners are more computer literate than me.   We still label people by their skin color, but we\'re so different inside. It\'s like people that say their outside is male, but the inside is female. Maybe my outside is white and my inside is latino.   '),(11996,3496773,'07,July,2004','2004-07-07 00:00:00','    I am writing to inform everyone that my life has officially bottomed out. Now, I\'m not writing this for pity. I\'m not writing this out of some sick,therapeutic need to purge every little detail from my soul - which, from my limited experience, seems to be all the rage amongst bloggers. No, I write this because it\'s really quite amusing. Amusing in a bitter-my-life-is-going-no-where sort of way, but amusing none the less.  You see, I was turned down for a job at Burger King. I don\'t mean to toot my own horn or anything, but I am pretty sure that I\'m qualified. Heck, some may even call me over-qualified. After all, I am college educated. I graduated from a decent school - Grove City College - with decent grades - (ahem)magna cum laude (ahem). I\'m sure no one else working at Burger King is capable of the following conversation: \"Sir, your fries will be up in a second. In the meantime, why don\'t we discuss Bahktin\'s concept of dialogic?\" True, that knowledge doesn\'t really come in handy at Burger King. As a matter of fact, as I\'ve discovered, it really doesn\'t come in handy anywhere.  The really weird thing is I kinda wanted the job at Burger King. I am practically broke. The money, while far from ample, would be appreciated. The hours would\'ve been alright. Plus, it\'s not like it\'s my long term career goal. My long term career goal is to drive the Oscar Mayer wiener mobile. Ahh, dreams.  And I have other problems too. My car - the \'93 Lumina I affectionately call \'The Silver Bullet\' - is in dire need of brake work. They squeak and squeal and grind. I only turn up the radio. What I can\'t hear, doesn\'t exist. My parents - instead of hounding me to move out - have become far too acceptive of my presence. They now make long term plans assuming I will be living at home. I\'m not talking about a couple of months in advance. I\'m talking about retirement plans. This is somehow much more depressing than having them hound me. I think even my parents have lost hope in me.  I do have some plans, secret untested ways to make some cash. I think I might befriend an eccentric billionaire, get in on his will and then the next thing you know the cops are over and I\'m saying: \"I dunno officer. He musta just fell down the stairs.\" If I can\'t shut my conscience up long enough to do that I\'ve got back ups. For instance, I am currently selling advertising space on this blog. For a small, negotiable fee I will head each post with my current sponsor - I can just see it: \'Today\'s blog entry is brought to you by preparation H, preparation H, for all your hemorrhoid needs\'. I could also slyly insert advertising subconsciously into the blog entries. Hold on while I take a long, slow sip of an ice cold Coke-a-cola. Mmmm, nothing satisfies like Coke on a hot day.   Now where was I? Oh yes, hair brained schemes for money. Now it\'s time for my favorite scheme. You know how people pay good money for those fancy psychiatrist and therapist with their fancy degrees in their field and all they do is listen to you and offer no advice? Well, I figure I could do that. I can listen to people complain about their lives and their parents. I can offer no solutions. And I could do it for a whole lot less money. So, I\'m glad to announce that I\'m now an official unlicensed therapist. For only $10 an hour I will listen to your problems. For $13 an hour, I\'ll pretend to care.  So, even though my life has hit rock bottom, do not despair for me. I\'ve got my ideas and my creativity. I am sure that I\'m going to make it after all, and someday, someday I am going to drive than wiener.  Shalom   '),(11997,3496773,'02,July,2004','2004-07-02 00:00:00','    As I\'m sure all of you know - or at least I hope you do - we are in the midst of an election year. Come November we Americans - or at least the 30% of us who vote - will go out to the polls and cast our lot with either George W. Bush or John Kerry. This is a momentous decision, but if you are like an ever growing number of people I run into, you don\'t really care. This is especially painfully true of our young people. Yes, I know - as I posted on our previous, long forgotten blog - voting isn\'t cool. It never has been nor never will be. So, I think what young people need is a candidate they can really get behind.  With this said, I would like to take this moment to declare my candidacy for the Presidency of the United States of America. Yes, I know according to the so-called \'Constitution\'I am too \'young\' to run for president. To this charge I only have one reply: constitution shmonstitution. I have my right to seek gainful employment and I shouldn\'t be blocked. This rule is clearly biased against the young. This is the same reason high schoolers can play in the NBA. Now all I need to find is some judge to eviscerate the constitution. That shouldn\'t be hard. It\'s what they seem to do.  Now that that\'s out of the way, I know what your next question is: \"But James why should we vote for you?\" A very good question, deserving of a very good answer. So why should you vote for me. Because I\'m super cool, that\'s why. But don\'t take my word for. My mom thinks I\'m super cool too. She also thinks I\'m very handsome and quite a catch. So, there naysayers. You don\'t hear Bush or Kerry\'s mothers saying that.   Still not convinced? Want to know where I stand on the issues, what my platform is? Fine. I now proudly unveil the platform which will carry me to victory come November. I, James Catullo, promise the following:  1) Mondays, according to my careful research of Garfield comics, are the worse day of the week. Therefore, once I am president, I shall declare that every Monday morning from 10:30 to 11 AM be dubbed \'Dance Party USA\'. During this time, according to the law, all business shall cease and all Americans shall dance as they see fit.  2) The band Nickelback will stop recording and touring.  3) I will lower taxes.  4) All Americans will be able to declare 1 personal holiday a year. On their personal holidays people will be excused from work or school. This holiday can be any day and for any reason. If you wish to spend the day at home in observance of Scott Baio\'s birthday, so be it.  5) Women will want to go out with me. I\'m adding it to the constitution. Either they go out with me or they go to prison.  6) I will build more prisons to house the expected millions of women who still won\'t go out without me.  7) I will cut the size of government.  and finally...  8) I will change the national anthem to \'The Hustle.\' This way whenever an American Olympic athlete wins a gold medal, he or she will do a reverent and tearful rendition of \'The Hustle\' on the winners podium. And that\'s funny.   So, come November remember you have another choice besides Bush or Kerry - and no, I don\'t mean Nader. When you vote, America, I want you to remember the only candidate who really knows what it means to be American. I want you to remember the only candidate who speaks to you in the language you understand, the language of old sitcoms. Most of all America, I want you to remember: DO THE HUSTLE!! (do do do do do do do)  Shalom   '),(11998,3496773,'01,July,2004','2004-07-01 00:00:00','    Now, as little mike aforemantioned, there is a new theatre in beaver county.  May I echo that it does indeed beat the pants off the old Carmike Theatre across the way.  To start, it cost 7 dollars for a movie.  Carmike, $7.50.  Now not just the ticket prices, but everything has prices that small, significant ammount lower that you could say, YEA! it\'s better, And Cheaper.   Now i saw Spiderman 2 last night, and I\'ll say i didn\'t like it as well as the first.  Yes, the CGI is a bit cleaner, and the build up was good, but the story was not as... well... \"yea that kicks ass\"-esque.  A fight scene with Doc. Octopus cuold have gone longer and been better (a significant source of untapped territory), and the plot somewhat turns you against spidey, but a scene with a runaway train was done very well. But that is not why i Blog for you today!  The eason is simple...Donnie Iris. As i left work on wed. it was aobut 6:50 when i heard on the radio, DVE nonetheless, that Michelle Michaels was at the new Cinemark Theatre in Monaca, and she was antici...........pating the concert event of the week... Donnie Iris and B.E. Taylor would be performing an acoustic set at the opening of the new Theatre. Holy King of Cool Batman!  I had to get down there!  So going as fast as the legal limit would allow me, I raced over the the Lowes Shopping center to catch the cool King, B.E., and some kick ass guitarist named Rick belting out 8 Days A Week, Rocky Racoon, I wanna Hold Your Hand, Love is like a Rock, Vitimin L, Ah Leia, and The Star Spangled Banner for an encore.  It was so great.  Plus i won a free T-shirt for an upcoming movie called \"Harold and Kumar go to White Castle\"  And an Allman Bros. CD. All in all, a great compilation of events for a wednesday night.   '),(11999,3496773,'01,July,2004','2004-07-01 00:00:00','    \"Hey, a prisoner on death row escaped in his incredible flying machine!  He\'s getting away, quick turn on the spotlights and try and find him!\"    I can only hope that someday, that is a possibility, just so there can be more spotlights in the sky.  I really don\'t think there are enough... at least not around here.  I\'m sure in Hollywood they have millions of spotlights (of course they are harder to see due to the intense smog and their even more intense governor.)  Around Beaver County, however, these things are pure luxury that comes around ever few years.  Let\'s see... I remember spotlights for the new Wal-Mart that opened (so they can let everyone know where to protest, of course), there were spotlights when Lowe\'s opened (I think Lowe\'s actually sells spotlights, actually...) and the Mall had spotlights years ago when they remodeled it.  And of course we have a spotlight just dedicated to the Beaver County superhero: The Cruiser (he tends to concentrate his powers while sitting in front of his \'72 mustang at the county\'s weekly car cruises...)    There were spotlights in the sky last night.  In a weird twist, they were not only buzzing around to celebrate the opening of a Spider-Man 2, but also to celebrate the opening of a brand new Cinemark movie theater.  Two great reasons to have spotlights in the sky.  My area of Beaver County, which houses the tri-shopping mecca of Wal-Mart plaza, Lowe\'s Plaza, and the Beaver Valley Mall, is now experiencing its consumer renaissance.   The trend nowadays has moved beyond strip malls, outlets, and indoor malls and has moved to the \"shopping village.\"  It\'s a set of malls, warehouse stores, and strip outlets along with a movie theater and a ton of over-priced restaurants (the kind that have their walls covered with crazy crap).  And that\'s when that area becomes known only for its stores.  For example, Robinson Twp, Cranberry, Boardman.  Three huge shopping meccas that I only know because I go there to shop and have a good time.  I have heard rumors that people actually live in these places, going about their daily lives and ignoring the frightening things that are operating right down the street (kind of like the folks that live near the power plant).  Is this the fate of my town and my county?  And really... is that a bad thing?  Last night marked the grand opening of a new Cinemark movie theater in the area.  12 theaters, stadium seating, surround sound, and reasonably priced tickets.  This, of course, beats the pants off of our former theater, which could be described as \"malnourished,\" which had eight theaters, sticky and flat floors, ripped seats, and ticket prices that required you to take out a loan.  Did I mention they only got the \"big\" movies?  Hopefully our new neighbors at the Mark will be wise enough to get me some juicy independent movies... or at least the new Woody Allen movie coming out this fall.    Now I haven\'t been to this theater yet, although that will change when I go there tonight, but I hear its pretty nice.  It sits in the newest of the shopping centers in Beaver County.  There is a brand-new Shop n Save and the aforementioned Lowe\'s right next door.  And they are in the process of building a Starbucks (yes!), a Olive Garden, Red Lobster, Home Depot, and Best Buy all within the next couple years.  We are truly on our way...   and I couldn\'t be happier about it!  Sure this may bring a tad more traffic, but these things are far enough away from my house that I won\'t be bothered.  And at the same time, I won\'t have to drive 30 minutes to get to the stores I love.  Most of the people in Beaver County leave the county to go shopping, and now all of that money will stay here!    And so in conclusion, I welcome you new theater.  I thank you for riding in on a horse that looks and sounds wonderful, for bringing Spider-Man 2 with you (a movie that I\'ve not only been looking forward to, but will totally knock Michael Moore out of the box office), and for further murdering live theatre in Beaver County!   '),(12000,3496773,'11,August,2004','2004-08-11 00:00:00','    There is a great and vile evil roaming free upon our land, a pestilence, a blight upon the very soul of our great country. Nothing can divert this rank villainy. What, you might ask, could be so foul? Is it AIDS? Is it poverty? Is it rhetorical questions? No. No. No. I am talking about a plague at once more benign and more devilish than any of these. I am talking about a scourge which wants nothing more than to ruin your good times. I am talking about those idiots who bring guitars to parties and want to \'jam\'.  If you attend a good number of gatherings, get togethers, parties, and the like, there is a good chance you know exactly what I am talking about. You go to a party, and everything\'s A-Ok. There\'s food, people, your choice of fine beverages. People are talking, joking, laughing, having an all around good time. Then, Like a foreboding crack of thunder in the distance, someone brings in his guitar. This is quickly followed by the realization that many people have guitars. Often they drop everything they are doing and run to their car to get them. Eventually, someone suggest that all the guitar players get together somewhere - often in right in the middle of the party - and jam.  For those of you who don\'t know what I mean by \'jamming\', allow me to enlighten you. To jam in the musical sense - as opposed to the raspberry preserve sense - has, in my experience, two definitions. Definition one: What happens when a group of talented musicians get together and play music in a fun and unencumbered way. Definition two: What happens when drunk douche bags get together with guitars - not necessarily in tune - at a party and proceed to annoy anyone who isn\'t playing. As you can see, jamming in the former definition is fine and I have no problem with it. Jamming in the latter sense however is not fine and I despise it. It has yet - even once - to add anything positive to any party I\'ve ever been to. Quite the contrary, it often ruins parties. It makes conversation more difficult. It\'s never as entertaining to the people not jamming as it is to the people jamming. As a matter of fact, jamming has the knack of bringing parties to a grinding halt. People are held captive by these idiots - who always play way too loud. The enjoyment of the masses is negated for the enjoyment of a small group of morons.  Now here\'s the real problem. In every group of jammers - I think I just coined a phrase - there is one guy who is really, genuinely good, and it would probably be okay if just this one alpha guitarist would play by himself. It would be kind of the low rent equivalent of the smooth jazz pianist at high society functions. However, people with guitars always materialize around this alpha guitarist and start jamming with him. This is patently absurd. At the aforementioned high society functions you never hear this said to the smooth jazz pianist: \"Dude, I\'ve got a key board in my car. I\'m totally going to get it and we are totally going to jam.\" It just doesn\'t happen in civilized society.  Of course, the people who join the alpha guitarist are of variable skills. There are usually one or two guys who more or less know what they\'re doing: play things in the right key, keep rhythm, know a bunch of cords. Even this is kind of OK. At least it doesn\'t sound horrible. But this is never how it goes because, invariably, the group is joined by some guy - sometimes guys - who just sucks. This guy always plays too loud, plays in the wrong key, and for some reason can\'t keep his guitar in tune. Now the alpha and beta guitarist - instead of doing the reasonable thing and telling the omega guitarist to get lost - have to play louder to cover up this travesty.  To make matters worse. It always happens that no one knows the same songs. So you get one person trying to show another person how to play something, which they proceed to play poorly. For some reason, they manage to never play a song through the whole way. Often they\'ll spend twenty minutes noodling around on two or three cords which grows monotonous and dull. And through all of this, the other people at the party are just trying to have a good time. They want to talk to friends, meet new people, laugh, and whatnot, but those stupid jammers just refuse to go away. You can\'t talk over them. You can\'t talk to them - and I\'ve had conversations end because all of a sudden my friend just had to jam. You can\'t even dance to them, because they usually really suck. They are like a black hole in the middle of the party, sucking the fun away from everyone else so that they are the only ones really enjoying themselves. At least they don\'t eat all the food.  So, in conclusion, let me leave everyone with a good piece of advice. If you are going to a party and want to take a guitar with you, stop for a second. Now, ask yourself a few questions. Are you playing in a band that has been asked to perform at this party? Are you being paid to perform at this party? Are you going to a party where everyone - or at least a vast majority of the people - will jam with you or not mind if you jam? Are you Jimmi Hendrix? If you answered yes to any of these questions, then go, take your guitar. But if you answered no to all these questions, then LEAVE YOUR STUPID GUITAR AT HOME!!! Try having a conversation for a change. You might like it.  Shalom  P.S. I would like everyone\'s help with a very important decision. I have finally followed through on my promise - made here in this blog a few weeks ago - to grow a mustache. I have not shaved my upper lip for almost a week. As you might imagine, my mustache is in that strange, sleazy, half-there state at the moment. Now this doesn\'t bother me, but...I have a job interview tomorrow. Should I shave my dirt-stache for my job interview or just keep it? Seriously, I\'d like to know what everyone thinks. Even if you don\'t know me, even if you find yourself on this blog on mistake or on a whim, let me know what you think. Just leave a comment and hopefully they will help me decide whether or not to shave. Thanks.  shalom (again)   '),(12001,3496773,'09,August,2004','2004-08-09 00:00:00','    I got a new job! I\'m going to sell insurance. That may not sound like much, indeed it\'s not the thrill a minute lifestyle I\'m used to working with bad kids and all, but it\'s a good job that will pay well if I apply myself. The funny thing is that I will go broke before my job starts. I have to shell out about $300 to get licensed to sell insurance. And the job that I am doing is not paying me until the end of September I recently found out. So that sucks. Yet another reason why insurance will be good for me.  Since I probably won\'t start the insurance job until the end of september, I need money now to pay these bills and so I can eat. So anyone have any odd jobs in the meantime? Very short term. I will clean your house! I will perk up your slacking young marriage if your wife is a looker. Will improv for dinner and $20! I also love charity. I ma not afraid of it.  Excited in a happy way and a nervous way, Mike the Tall   '),(12002,3496773,'07,August,2004','2004-08-07 00:00:00','    Today, I have something to say which I usually do not have the luxury of saying. Today, something exciting happened. No, I did not discover that \'Boy Meets World\' is on twice a day on ABC Family. I\'ve known that for quite some time. I am talking about real excitement. I am talking about genuine heart pounding, pulse racing, bowel loosening, staring death in the face excitement. Of course, it\'s wrapped in the dull and mundane, but such is my life.  Today, I was simply taking a short trip to good old Grove City - which for the uniformed contains no groves and is most certainly a city in name only. I was simply going about my normal activities. Going to the local library, scouring for work and the like. As I piloted my \'93 Chevy Lumina - the very sexy, scantily rusted car I have dubbed \'The Silver Bullet\' - toward a red light everything was going to plan. My mind was hardly on my driving. I was contemplating such lofty ideas as the responsibilities of free will and the uselessness of Aquaman, as I began to depress my brake pedal. Suddenly, to my abject horror, I found my foot and the pedal drop effortlessly to the floor of my car. There was no resistance, and - even more alarming - there was no deceleration. My brakes had given out.  Now, I would like to say that I remained absolutely cool, calm, collected, poised, but I would also like to say that I run a small island nation in the South Pacific. Needless to say I panicked. There was a truck ahead of me stopped at the light. I was going to crash into it. I just knew it. My heart leaped up into my throat as my stomach dropped down into my bowels. Luckily, my liver stayed put. Sweat popped out along my forehead, and soaked my underarms. I actually made this sound: \"DAAARRAARAAARGH!!!\" Of course, this was screamed out like a little girl, with tears streaming down my face. I had never been more sure that disaster and, yes, maybe even death was inevitable.  You may assume that my life flashed before my eyes, that I was treated to a kind of going away montage of my greatest hits. That is what people say happens, but it didn\'t happen to me. I didn\'t see my parents faces, or the lost loves of my life, or my missed opportunities. All I saw was the rear end of a Mack truck, getting progressively and unavoidably closer. Then again, maybe my life did flash before my eyes and I just didn\'t notice it. After all, most people don\'t notice my life. Why should I?  Still, I was not consigned to disaster. I was not going down without a fight. This was not to be the end of James. Oh, no. So, I did the only things I could do: I prayed and I desperately tried the brakes again and again. I pumped those brakes for everything they were worth. Trust me, you could watch all the porn in the world and not see pumping as vigorous or hard as this. I hit those brakes with everything I had and prayed that somewhere, some how, my car had some residual memory of stopping, that there was just one more stop left in the old girl.  I kept pressing harder and harder, until I thought that my foot would burst through the floor and drag along the pavement. Even then I would have remained pressing for dear life. Finally, a miracle happened. As I pressed harder, the car actually began to slow down, and not just from coasting. Somewhere, brakes were causing friction. I was slowing down. Joy of joys, thank you Jesus, I was slowing down. I was actually coming to a stop. I was going to survive. I was going to live. Of course, the braking power was pretty bad. I would still probably hit the truck.  But no, with my new confidence. I had the wherewithal to turn myself onto the shoulder of the road completely avoiding disaster. Then, the light changed. The truck moved and I ever so carefully - never driving above 3 m.p.h - I was able to guide my wearied vessel into a parking lot to safety.  The emergency had passed. I was safe. I was alive. Nothing bad had happened at all. I felt like dancing. I felt like singing. I felt like bursting out of my car and hugging and kissing every person - or any hot girls at least - in sight. Now safe, I felt life rush back into me. I had come so close to death, to tragedy, and I had avoided it by the narrowest of margins. I did the only thing I could do. I curled up into the fetal position and began sucking my thumb. After all, this really was quite a bit for my milquetoast soul to handle.  Now, I could go on. I could tell you about the time spent in the parking lot stretched out on the hood waiting for AAA. I could tell you about Dave - the less than talkative flatbed driver. I could tell you how this caused yet another unforeseen charge to my already overused credit card. I could tell about all of this, but that is only the mundane shell, the ordinary part of this extraordinary adventure. I have ordinary adventures every day, and so do you. So, I\'m simply going to leave it to the extraordinary today. Who knows when I\'ll have this opportunity again? Plus, \'Boy Meets World\' is on.  Shalom   '),(12003,3496773,'07,August,2004','2004-08-07 00:00:00','    Why does every computer come with solitaire? This is an evil, evil thing. Solitaire is a deceivingly simple game. And you think you can win all the time. I usually play until I win. This being the case, I often end up playing for hours! Hours! I could be writing the great American novel, but instead I am moving cards from one pile to another. Not even real cards but pictures of cards. It\'s not like I have nothing better to do. (Well, that\'s not always the case.)  I have had problems in the past with gaming. It tends to take up a lot of my time. But it\'s a cheap addiction. I usually play some video game every day. I\'d say on average I do 45 minutes a day on one of my game systems or my computer. While on my computer, I can do other things like surf the net, chat with friends, or actual work. But I realize that this vice is a cheap one. I figure I spend about $20 a month on this habit. When I went to Dave and Buster\'s a lot, it was probably twice that, but seeing as I\'m close to broke that has stopped. It\'s way cheaper than cigarettes which is like $4 a pack now. It\'s even cheaper than buying a bottle of wine a week. Something else I wish I could do. And it\'s absolutely cheaper than whoring. They cost way too much for me. And they risk of disease and the pending medical bills from that is far more than I\'m willing to spend. So video games is a good vice for me. It\'s controllable.  The real problem is time. I have been a little late for work more than once cause I had to finish a game or get to a save spot. And the days in middle school when I was constantly hollered at for being late for the dinner table for similar reasons. But those were the days when you had to pause the game and go eat and come back. And God forbid something bump that nintendo. I wanted to kill the dog more than once for that.  The other night I was tired and had to get something off my computer around about 11 PM. I began talking to Mike a little, and then I opened the solitaire game. Bad move. I like the Spider Solitaire which most people shy away from cause it\'s complex and stategy out the wazoo not unlike high level chess. And I don\'t mess around with the one or two suit games. It\'s four suit or nothing. The game is hard, but I can usually beat it in a few tries.  Well, it got to be 11:30, I had to work at 8 AM the next morning and I had a long day. My eyelids were dropping. I was fallign asleep and my hand was still moving the mouse to move the cards. What the hell is wrong with me? I was falling asleep and I could not stop playing a dumb inconsequential card game! I was fighting sleep to try and win a card game. I just wanted to win. I just wanted it more than sleep.  I\'ll have to reset my values: 1. Food 2. Sleep 3. Women 4. Video games Okay that\'s better.  I\'m going to get off the computer and shut off the solitaire. I mean I borrowed two games and I need to play them.   '),(12004,3496773,'05,August,2004','2004-08-05 00:00:00','    Now, I\'m sure that no one - not even members of my own family - cares about this, but I just want everyone to know that I have a particular soft spot in my heart and upon my palette for fortune cookies. Even though I can hear everyone yawning at their computer monitors, I am going to proceed to praise the fortune cookie. I know it\'s an odd, not too mention a rather random thing to write about - especially considering that I haven\'t had Chinese in quite some time. I just like fortune cookies. I like the idea of a food that really says something. I like breaking things open only to find a special surprise. That\'s wonderful reinforcement for many of my more destructive desires. As a matter of fact I wish more things were like fortune cookies, and could be destroyed only to gather some greater wisdom. Wouldn\'t it be wonderful if Ming vases or Faberge eggs also contained notes or prizes inside. Wouldn\'t you just love to smash a priceless work of art to the ground only to find a little slip of paper with a message which automatically becomes hilarious once the words \'in bed\' are amended to it?  More than just the hidden prize qualities of fortune cookies, I actually really like their taste. I don\'t think enough people really think about eating fortune cookies. Everyone gets caught up in the breaking and reading, the actually eating of the fortune cookie often becomes a secondary concern. Eating the cookie is seen more as a neat way to dispose of refuse than something to be truly enjoyed. But next time you break apart one of these culinary mystery boxes, I want you too actually pay attention to eating it. I want you to enjoy the crunch, savor the delicate flavor. For me, this flavor will be forever linked to the late evening Chinese food that would be delivered in the college dorm. It\'s just wonderful.  Despite all of this, there is, alas, one thing which can ruin the whole fortune cookie experience: not getting a fortune. You may know what I am talking about. Sometimes you don\'t get a true fortune. Sometimes instead of getting a real fortune, a glimpse into your future - something like, \'You will undertake a great journey\' - you get a piece of advice, a goofy aphorism - such as, \'A man of wisdom is of great value\'. Personally, I don\'t want to be lectured by my fortune cookies. I don\'t want to learn great life lessons. I want to know what\'s going to happen. I want a real fortune.  Sadly, I fear these aphorisms are becoming more and more prevalent in fortune cookies. They are watering down the whole fortune cookie experience. They take away all the mystery of fortune cookies. They turn magic into morality lessons. Plus, they really suck for adding \'in bed\'. In short, the aphorism fortune cookie message must be stopped. It\'s time for some one with a bold vision to step up to the fortune cookie plate and put an end to this madness. That\'s right, you guessed it, I want to write fortune cookie messages.  While this may seem like a fairly easy task. \'Any one,\' you may say to yourself \'can write vague generalizations about the future.\' That is why I do not plan to be vague. I am going to be deadly specific. I feel this is a fantastic way to combat the growing wishy-washiness of the current fortune cookie environment. I also think they\'re much funnier this way. So now, without further ado, I present you with some of my test fortune cookie fortunes.  \'You are not going to score with your current date\'  \'Before midnight of this coming Tuesday you will find a priceless jewel in the restroom of a Greyhound station\'  \'You will be shot in the back of the head at a 7-11 while buying milk which is one day away from expiring\'  \'You will cut your toe-nails on Thursday, 10-ish, when you want to wear sandals but realize that you have disgusting toe-nails\'  \'Look to your left, look to your right. The person who\'s going to give you herpes is certainly in sight\'  \'You will call off work because you get really caught up in a lame VH1 countdown show. You will vehemently deny this, but your cover will be blown when accidentally mention that you think \"The Final Countdown\" by Europe is actually a rocking song despite it\'s place on the worst songs of all time list\'  \'The Canadian government will bug your phones and follow your every step simply because they can and they know no one is going to believe you\'  \'Suddenly, at the age of 78, you will realize that your whole life has been nothing but a sham. You have denied your dreams for too long, but now you are going to finally follow your dreams of becoming a world-renown tap dancer. You will follow your dreams by enrolling in a Performing Arts high school much like the one in the film Fame. While there you will fall in love with an instructor. However, knowing that your love is illicit and can never come to fruition, you sublimate all your feelings into your dancing. You dance like you\'ve never danced before, because you haven\'t. Finally, on the day of the big all school talent show, you confess your feelings to the instructor. The instructor admits that he/she shares your feelings. With your heart full of requited love, you go on to give the most astounding performance the school has ever seen. After all the school has never seen a 78 year old freshman. After the talent show, a Broadway scout approaches you and offers you a part in the chorus in an upcoming big, Broadway musical. You accept. On the eve of the musicals big opening, the star breaks his/her leg. The crusty, but lovable, director chooses you from among the chorus to fill in. You succeed brilliantly. All the reviews gush about you. That evening you will die quietly in your sleep in bed.\'  Now, that\'s what I call a fortune.  Shalom   '),(12005,3496773,'02,August,2004','2004-08-02 00:00:00','    I don\'t know who Allen Jacobs is, but it would be much appreciated if he were to stop using my e-mail address when signing on-line petitions. You see, just the other day I noticed something distinctly odd amongst the penis enlargement and home loan advertisements in my E-mail folder. There was a message entitled, \"petition signature confirmed - I\'m With Busey\". This sent up a red flag in my head. Are they talking, I thought to myself, about that stupid reality show with Gary Busey and that guy who claimed to be a comedy writer even though he showed no sign of a working sense of humor? And, by the way, I have no idea why I use the \'they\' pronoun, I guess I like to think of a vast underground network of people filling my inbox, instead of single, solitary nerds. Anyway, with my curiosity sufficiently piqued I opened the e-mail, and to my abject horror discovered that someone, this so-called \'Allen Jacobs\', has signed, using my e-mail address, a petition to bring back the show \'I\'m With Busey\'.  I was truly horrified. Trust me if \'I\'m With Busey\' makes it back on the air, I had nothing to do with it. I\'m absolving myself of all guilt right now. It was all a horrible, horrible mistake. Whoever this Allen Jacobs is must have mistyped his e-mail address and come up with mine. Please, you have to believe me. I don\'t want \'I\'m With Busey\' back on the air. I would never knowing aid in unleashing such a devastating evil onto the world. I mean, this blog is bad enough, but, then again, I\'m pretty much convinced no one reads this, so there\'s no harm here.  Eventually, after much renting of clothes and gnashing of teeth, I calmed down and became quite curious. What is this petition? Where\'d it come from? Are people actually signing this thing? Shouldn\'t Gary Busey be dead by now anyway? So, I read the entire e-mail message and found that the petition - and the e-mail - originated from PetitionOnline.com. So went to the sight and found the Busey petition. Apparently \'I\'m With Busey\' has - had? - fans, and they cannot imagine why this \"innovative and creative\" show would be canceled by Comedy Central \"without any given reason.\" Apparently, these people are unfamiliar with a little thing I like to call the Nielson television ratings.  Sufficiently satisfied the \'I\'m with Busey\' petition was mostly harmless, I decided to explore the site a little fuller. Let me tell you, a whole world opened up before my eyes. A world populated by obsessive dorks and losers with too much time on their hands and internet access. I felt right at home. People have petitions for everything. Some of them even made bringing back \'I\'m with Busey\' seem like a good idea. Why would anyone in his right mind sign a petition to keep \'On Air with Ryan Seacrest\' true to its name? And that\'s not a rhetorical question. I really want to know. If you can fill me in, please do.  There are a lot of odd petitions. Most of them having to do with bringing back canceled shows on sci-fi and stuff having to do with video games - which gives you a pretty good idea who the average users are. Mixed in amongst these typical internet type concerns where things such as a petition to prove that someone\'s ex boyfriend is a real \'Dick Eating Dick Face\' and someone who desperately wants to see Bob\'s Big Boy make a comeback in the worse possible way - although, I\'m pretty sure, Big Boy can\'t come back any other way. There was one really quaint petition entitled \'Stop the Hate\' which was obviously posted by some well meaning, but utterly clueless liberal teenager. This ranks right up there with putting those stickers that say \'hating\' at the bottom of stop signs. Sure, it\'s a swell message and all, but it doesn\'t really do anything but help the person who put it up feel good about himself. It\'s not as though a Klansman is going drive by one of those stop signs or see a petition against hate and all of a sudden clap his palm to his forehead exclaiming, \"Stop hating, why didn\'t I think of that\" - but then again I didn\'t actually look at the petition, just the name, it could be quite different from what I expect.  Now, here\'s the part which really got me thinking. PetitionOnline.com bills itself as \'a premiere free speech forum.\' While I agree that free speech is one of the most important rights reserved for us in the constitution, and that this site is certainly a forum for free speech, I would be a little wary about calling it premiere. Now, maybe the PetitionOnline creators started with noble causes, to, say, petition the government - which is expressly protected by the 1st amendment - but when the founding fathers added the first amendment I\'m sure creating a petition to get a lame TV show back on the air wasn\'t high on their minds. I\'m not saying that they\'d be against it, but if they knew that people where using free speech to save \'I\'m with Busey\', they would most likely answer with a yawn - of course you\'d first have to explain television to them and then give them some sort of idea what Gary Busey\'s like and then describe the show to them. \"By Gum,\" they may say, \"Sure we fought for freedom of speech, and your silly little petitions are cute and all, but doesn\'t anyone have anything real to say? Aggrievments to the government perhaps?\" People fought and died for serious issues, not so some moron could sign a petition as \'BUTTSEX\' - the same person listed his location as \'BUTTSEX\' and left, as a comment, \'No, seriously, I like BUTTSEX\'.  It just seems that people use the term freedom of speech too loosely anymore. I don\'t think anyone should even mention it until they are seriously being quieted by the government. I\'m tired of hearing people whine about freedom of speech because no one wants to listen to them. You\'re guaranteed the freedom to express, but not the right to a major forum or a willing audience. Although if there\'s one thing I\'ve learned from OnlinePetition.com, no matter how lame an opinion you have, some idiots are going to be with you. Now, if you\'ll excuse me, I\'m bringing back \'Who\'s the Boss\'.  Shalom   '),(12006,3485073,'31,May,2004','2004-05-31 00:00:00','   School is always an interesting experience, today was no exception. We got to paint again in religion class today (I feel like i\'m back in kindergarten), and Mr. O\'Brien (the religion teacher) is still blaming me for the VCR that he broke. It was actually kind of funny how he broke it, he tried to pull the T.V. cart over to watch the movie and accidentally tipped it and sent the VCR crashing to the floor, breaking the faceplate on it. And auto shop was a blast today too. Rodney brought in this push go-cart for his I.S. and, of course, Jason and I took it for a test drive. Man! that thing is awsome! It\'s got two wheels in the front and one in the back that steers. We pushed it full tilt across the parking lot and man does that thing ever corner sharp! You turn the wheel just a bit and you bring it up on to two wheels! Math class was a blast too. Of course, it always is (I won\'t go into detail about that). But the highlight of the day was definitley civics class. We had a sub today, so we played tic, tac, toe and SOS all class. Amanda raped us both at it. then she started to write things on my arm. Man, she broke up with her man on thursday, i\'m thinking about making a move on her...  WILL COLTON EVER ASK AMANDA OUT? WILL RODNEY EVER LET THEM USE HIS GO-CART AGAIN? TUNE IN NEXT TIME AND SEE!  '),(12007,3485073,'30,May,2004','2004-05-30 00:00:00','   It was crazy, today we (me, mike wheeler, and travis ferguson) went nuts today with travis\' car! First, a descrption of his car. He drives an old, beat-up buik skylark that he\'s preping to run n the enduro race. It has only one seat (others ripped out for racing)it\'s very loud, and has no windows. the three of us rode around in the field with Mike stuffed in the trunk, then we started running into stuff (with mike still in the trunk. then Mike and i rode on the roof and hanging out the window while driving down the road. that was fun. then we chased Mike around with the car while he threw stuff at us. Then i tried driving standard for the first time, i stalled it once, then i got the hang of it. ha, ha, good times, good times...  '),(12008,3485073,'28,May,2004','2004-05-28 00:00:00','   WOOT!  MY FIRST POST! I\'d like to thank my mom, dad, and all of my friends for making this possible.(just kidding) As you can see, my mind is very unique and different SO TUNE IN DAILY AS WE DELVE INTO THE SICK TWISTED THING THAT IS MY MIND!!!  '),(12009,3485073,'30,June,2004','2004-06-30 00:00:00','   I\'ve had a shitty week so far. Kev,I have nothing to say to you here, read the e-mail. (That can give the rest of you guys a hint that something happened). But I\'m not gonna dwell on that incident. I went to kev\'s the other day, tried Battle Engine Aquila, pretty good game, I just don\'t share the same enthusiasm as Kev on the game (view Kevin\'s blog for more info). We slept up in his \"barn\" and froze our asses off. The next morning, we were mucha bored until his dad came and picked us up. Then kev drove me home (Yes, you heard right). I feel really guilty for taking those disks now, and i don\'t know why I feel so incredibly crappy about it. Cuz i plan to give them back. Then the next day I did literally nothing all frigging day. But on the bright side, we have a dog now! Rusty\'s the name, being a dog\'s his game (Yeah, I haven\'t slept in a while, so what?). I\'m hoping to hook up with Mikaela sometime this week before she leaves. (Man, if only she wasn\'t moving...) But we should have a good time, maybe invite some other people to come along, just so it doesn\'t seem like more than a get-together (If you know what I mean,) after all, she does have a boyfriend, and I don\'t wanna intrude on that. I\'ve got more respect than that. If you read this Mik, i hope you\'ll never forget us all, read my blog, I\'ll keep updates on local happenings. And i hope you can get lucky and find a guy like Andre or me down in the springs. (Yeah, this was a boring post, so what? you wanna make something of it punk? eh?)  '),(12010,3485073,'27,June,2004','2004-06-27 00:00:00','   Hey guys, sorry I didn\'t post yesterday, I didn\'t have the time to. So I\'m summing up the past two day\'s events with one post. First, Kev, why the F*CK didn\'t you invite me over when Kenan was over? I haven\'t seen Kill Bill either! I\'ll talk to you about that later (ass wipe). But Friday was great! I mean, trav, tony and I all went out for a ride. We got mega muddy. I had to pull down a tree with my winch (again) and Travis got me super muddy riding behind him (thank god for windshields). Going down one trail, I got hung up between two trees, It took us a while to get me out. But my four wheeler never gets stuck in mud (four wheel drive baby!). Then, on Saturday, we went to a friends house (a f*ckin\' nice one at that). Brad and I hung out, watched the football game, watched most of 8 mile (I still havent seen that movie) then we watched Saving Private Ryan. That\'s a long movie! it\'s like four and a half hours long. But some other stuff (if ya know what I mean) was cumming, oops, I mean coming on at 1:50am. But Brad fell asleep on the couch, TL fell asleep right in the chair in front of the computer (hands on the keyboard and all). I was bored so I just turned off the T.V. and the lights, and crashed on the other ouch in the room (the smaller one that I don\'t even fit on). That\'s what hapenned. So till tomorrow, peace out.  '),(12011,3485073,'24,June,2004','2004-06-24 00:00:00','   Whew! Today was very tiring. I\'ve been condemned to clear the bush behing our house. Thats no easy task! Man was I ever tired. Hey, if any of my friends read this tonight, are you doing anything tomorrow, \'cuz I\'m probably gonna be in Cornwall tomorrow. Sorry nothing interesting happened today, maybe something will happen tomorrow. (ZZZZZZZZZZ)  '),(12012,3485073,'22,June,2004','2004-06-22 00:00:00','   For the record Kev, Mark and Kev S were around till 9:30, Mark musta let her out when he went and sat in front of the fire outside. Because I didn\'t let her out.  '),(12013,3485073,'21,June,2004','2004-06-21 00:00:00','   WHA! That was a crazy party. Mark and Kevin S were piss drunk from that Jack Daniels they had. It was so funny to watch those two run into trees and trip over stuff. And the stories they told, and the songs they sung, the dances they danced...good times. Then the hay! i swear, we threw like 3 bails of hay on the fire, the flames were like 10 feet high. At first i didn\'t think i was gonna clear the pit, but they convinced me and MAN! We like burnt all the hair off our arms. Derek L and Kev C burnt the hair on their heads (get rid of the shags boys!). Kev, your bike hasn\'t been stolen, I rode it home. Hey, don\'t be mad, I wasn\'t gonna wait around your house all day, and it woulda taken me all day to walk. It was the only sensible option. Don\'t worry, I\'ll drop it off next weekend. Or you could come and pick it up if you want it earlier...Gotta run before Kevin kills me guys, till next time!  '),(12014,3485073,'18,June,2004','2004-06-18 00:00:00','   Ok, i\"m sorry that I haven\'t posted in a while, but exams screw everything up around here. First things first, after discovering the age of Ashley, I\'d like to withdraw my previous statement and apologize to all of you. Sorry Ashley, maybe when you\'re older.(But that doesn\'t mean we can\'t hang out). Also, EXAMS WERE NUTS!!! I\'m not talking about the actual exam though, it\'s what happened after each one. EB Games is open in Cornwall!!!!!!! YAY!!!!!! I\'m that excited about it. I went there after my second exam to see what they have and ended up buying The Getaway and signing up for their edge card. Third exam, we went cruising in Corey Cooper\'s car. He\'s just hooked up two 12\" subs with a 600W amp. I was getting a massage the whole ride, my ears were ringing. The fourth exam (today) was a bitch. I hate math, and now I\'ve gotta go to summer school to get it. After that, we went cruising in Corey\'s car again today. This time we had the top down, the shades on, and we drove slowly in front of the school jammin\' to the tunes of Guns n\' Roses (take me down to the paradise city, where the grass is green and the girls are pretty...). My sister has been having her birthday party all FREAKIN\' DAY! But we went to go see \"Around the World in 80 Days\" and I ended hooking up with Mikeala (and unfortunately, her boyfriend too) and we all went and watched the Chronicles of Riddick (sorry Kev, you missed your chance). Man! That movie is awsome! So much action...and I\'m not just talking about Mikaela and her boyfriend here either. But I still managed to whoop Mik\'s ass in the arcade before the movie started. (he, he, I\'ve still got the skill) Good times, good times... P.S. I should now be going back to my regular posting habits of near-daily posts, so don\'t you fret, I\'ll have plenty of stories to fill your boring little lives with now that summer is here. Till next time!  '),(12015,3485073,'11,June,2004','2004-06-11 00:00:00','   That\'s all the screen said all last class today. We had a pizza party and we were playing GT3 all class. Rodney and I owned everybody! Also, the Roadmobile is home today! I\'m gonna test it out. If I post again, you\'ll know I survived. Kev, u gotta come out for a ride next weekend! That thing isn\'t so bad, I\'d walk away from an accident on it (in about ten different directions, but that\'s not the point). Well, I gotta go test that contraption. Till next time!  '),(12016,3485073,'10,June,2004','2004-06-10 00:00:00','   I know you guys don\'t like all this mushy, gushy stuff, but a have to do it. Ever since the First Aid weekend, I haven\'t been able to get her off my mind. Ashley, I\'m yours forever, don\'t ever change. See you at Cadets caring 4 Canada tomorrow!  '),(12017,3485073,'10,June,2004','2004-06-10 00:00:00','   Today, we had some strong winds during math class. So of course we all went to try fly our kites. Ours just doesn\'t want to go up, on the other hand, Isaac\'s group\'s kite didn\'t seem to want to come down! It just kept flying higher and higher. I swear to God man, that thing was like 300 feet in the air! they were flying it over in the G.V. field and the actual kite was way over at the far end of our field. When it finally decided to come down, it almost came down on the little kids doing track at our school. Imagine the headlines if it had hit them: INSANE, OUT OF CONTROL KITE KILLS PIPSQUEAK DOING TRACK & FIELD. It\'s not funny man, it almost happend. Cuz when it came down, it came down swooping at our heads (did someone say broken neck?). Yup, that kite rates about a 8.1 on my weird-shit-O-meter. Just another day in the life of Colton McRae.  -P.S. If you read this Ashley, please don\'t get mad at me for the previous posts. I couldn\'t think of much else at the time.  '),(12018,3485073,'08,June,2004','2004-06-08 00:00:00','   Anyway, on the first aid weekend, all the cadets agreed that I was the best group leader out of them all (WOOT! GO TEAM ODHA![O-da-ha])Yeah, it rocked. The test was mongo easy, I was the first one done. Incase you haven\'t noticed, I\'m done with Amanda, Ashley is now in! But the difference this time is that I KNOW that she likes me back. I\'m getting kicked right now, so i\'ll finish today\'s events later tonight (when the parents go to sleep). P.S. Ashley Ceasor is da bomb!  '),(12019,3485073,'08,June,2004','2004-06-08 00:00:00','   First, sorry about not updating the \'ol blog in a while, I was gone all weekend. I was gone to my first aid weekend with the cadets. IT ROCKED!!!!!!!!!! (a lot of hotties). I really got popular with the ladies during the weekend with my quirky ways of acting. About the whole masturbatos thing, it all started with a little mis-communication. In the IMPs (really crappy meals), there are these instant mashed potatos. I hear something else (you know the rest), then we elaborated with that and decided to call it the white, creamy, sticky snack that all women LOVE. (especially Natalie and Mikeala). I really got down with this one chick (one of the hottest ones there.) named Ashley Ceasor. I almost got into the female cabin! But the damn officers wouldn\'t let me. (I walked her to the door and she tried to bring me in, but damn CPL Landry caught us. Yeah, we hung out together a lot, she even went fishing in my pants! (Kev, I\'ll elaborate on that next time I see you.) I\'m too lazy to post the rest right now, I\'ll probably do it later tonight.  '),(12020,3485073,'03,June,2004','2004-06-03 00:00:00','   First, before I start today\'s events, I\'d like to apologize for missing yesterday\'s post, and I\'d like to add that both today and yesterday are going to be in this post. Tuesday was nuts! The storm was insane!(it happened just after i posted tuesday\'s events) I swear, there\'s this tree about 20 feet from my house, the rain was soo heavy that i could barely make out the shape of the tree through the rain. Then the power started flicking on and off until it eventually went right out. Of course, this didn\'t affect me, I was in the living room surfing the net on my laptop. I just kept going for about another hour before my battery got too low. We felt so stupid sitting in the house with no lights when just about everybody else on our road had their generators running. We were without power for about 5 1/2 hours. Yesterday was akward too. I had to explain to my religion teacher that I didn\'t finish the homework because I had no power to do it (Of course, he didn\'t believe me). We also started our kite projects in our math class. As for today, I was lucky. I only had a half day of school. We finished watching this video about HIV and AIDS that explained that sex is the main way of getting it (Yeah, like that\'s gonna stop me). In auto, we made a decision that we\'d never get the rear wheel on the road mobile in time, so we decided to put the track back on and attach it back together by running wire through it. Then I pulled an incredibly stupid move. I left the instructions for my auto I.S. in my locker when I left school (It\'s due tomorrow)! I left school at noon today because we had to go to the burial service for my grandpa. Then we went back home for the rest of the day. TUNE IN TOMORROW TO FIND OUT WHETHER OR NOT COLTON\'S I.S. GETS FINISHED OR WHAT KIND OF TROUBLE HE GETS IN!  '),(12021,3485073,'01,June,2004','2004-06-01 00:00:00','   WHY DOES IT HAVE TO RAIN!?!  It sucks that it rained today (it always messes up the gel in my hair). According to the weather channel, it\'s supposed to rain tomorrow too. Oh yeah, I almost forgot! On to today\'s events. Today started off with civics class, the teacher was back today so fun time is over. Mind you, it\'s not like any of us listen to her anyway. We talked all class and didn\'t get any work done (again). Man, I\'m really starting to fall for Amanda, I gotta make a move before I lose my chance (if anyone has some suggestions, you can post them in the comments section of this post). Math class wasn\'t too much fun (test, Ugh!). Lunch was fun, I brought duct tape to school and went down the halls taping up lockers. We also taped up Mike Beauvais\' shoes (moon walk anyone?). Duct tape is so useful, it even gave me extra grip on my shoes! We spent religion class in the library on the computers doing questions. I had to stay after class to finish the questions (I was goofing off all class). In auto shop I had to call on my trusty D-tape again! We used it to fix the seat on the road mobile, then we pimped it out by spray painting it black. Oops, I almost forgot, most of you should already know about the road mobile. But for those who don\'t, the road mobile is my I.S. for auto, it\'s an old beater of a snowmobile that I brought in to put wheels on, right now we need some parts to finish it, but it\'s gonna be bomb! (oh yeah, I did take Rodney\'s cart for another spin, in the rain! That was interesting.)  '),(12022,3485073,'31,July,2004','2004-07-31 00:00:00','   Hello everybody, I\'m posting from our room in Fredricton, New Brunswick. We\'re in a suite at the Delta hotel, AND THEY HAVE FREE HI SPEED INTERNET! Today was day one of our trip. We left at about 7:45 this morning. Our van has a really nice ride and all the leg room I could want. And as you\'ve probably already guessed, it absolutely poured rain almost the entire way here. The ride was long, we were in the van for about a total of 12 hours. I\'m really glad I brought my minidisc player along for the trip, it kept me sane. Since I didn\'t have a power inverter or a charged battery, my laptop wasn\'t of much use to me during the ride. Hopefully tomorrow if all goes well, we\'ll be able to make it to our house in Nova Scotia where we\'re staying for the week. I don\'t know when I\'ll be able to make the next post (connections aren\'t as plentiful as I had hoped), so I\'d like to say hi to Kev, Mikaela and Ash. So far I\'m enjoying the trip a lot, the scenery is beautiful. Until next time,  This is Colton signing off from the east coast!  '),(12023,3485073,'29,July,2004','2004-07-29 00:00:00','   I hate having chores! I had to get up at 8:30 this morning to do stupid chores. Here\'s a rundown of the list:  -Take out the garbage (the 3 bags in the garage that weigh about as much as I do) -feed, water, and walk the dog (screw walking it! I hate that damn dog) -feed and clean the ginuea pigs (that wasn\'t so bad) -vacuum (all three floors!) -mow the lawn (oh, you mean that huge-ass field outside?) -empty and reload the dish washer (that wasn\'t so bad either) -cater to our every whim and give us money (ok, I made this one up, but its pretty close to that!)  And guess what! There\'ll be more of them tomorrow! YAY! (lights chore list on fire) I doubt I\'ll be posting tomorrow, I might saturday (from a hotel room! I love my laptop)] Till then, this is Mr. Slave signing off  P.S. We pick up our van today!, I can\'t wait to see it, all together now, LEG ROOM!!!  '),(12024,3485073,'28,July,2004','2004-07-28 00:00:00','   Hey everybody, I\'m happy to announce  SUMMER SCHOOL IS FINALLY OVER!!!  I\'m soooo happy! Now I can finally do something Kev, that is, when I get back from the east coast in a week (hehe). I\'m also happy to announce that I passed summer school with an 81% average, the second highest in the class. This is a good thing, this means I get my G1 in September (damn you late birthdays, damn you to hell!). The teacher put on the movie \"Inspector Gadget\" today (ugh!), but I was smart, I brought my laptop to school and was sitting at the back of the class playing Wolfenstein 3D the whole time! My parents have been a little uptight lately getting ready for our trip (we leave on Saturday). We got a new van for the trip because I don\'t have enough leg room in either of our vehicles to sit in the back for 16 hours. So we bought a  urlLink Honda Accord , the biggest in the class. We\'re going to pick it up tomorrow, I can\'t wait! FINALLY, LEG ROOM! Oh yeah, I\'ll be bringing my laptop on the trip so I can keep you guys updated with what happens on the trip (and to keep in touch) Thats all for now, and as Mikaela would say, written with love, sealed with a kiss, Colton McRae (yeah, I know it sounds weird coming from me)  '),(12025,3485073,'26,July,2004','2004-07-26 00:00:00','   As todays title suggests, all we did in math class today was watch Wallace and Gromit. It was fairly amusing, but it got tiring around the third episode. I went to EB Games today (again) and bought&nbsp; something (again). I bought a fourth paddle, now multiplayer games can be fully enjoyed. I know the guys at EB so well now, it\'s \"Hey man, so what\'s new in life\" when I walk in. I\'m getting ahead of myself here. Let\'s start at the beginning of the day, I woke up late this morning and was doing everything at about 100mph trying not to miss the bus. I did make it. That\'s all I can think of for today. Oh yeah, Mikaela, don\'t let life get you down, things will get better, I promise.  '),(12026,3485073,'25,July,2004','2004-07-25 00:00:00','   Oh my god! That road party was great! Kevin, I know you\'re gonna hate me after this post, but it was a killer party. For one thing, I must\'ve put about 10km on the SUV last night driving people around (and myself). I had a fair bit to drink aswell, even my mom gave me beer! We had fireworks and everything, Travis almost blew his face off with one when we thought it had gone out, but it all of a sudden exploded just before he grabbed it! We were mean to Travis too, he went to go sit down (in my seat) and just before he sat down, I pulled (my) seat out from under him and he took a one way trip to the ground. He spilt his beer all over himself, Seleena was laughing her ass off at him. But hey, nobody takes my seat. It was funny too, I was the designated driver last night, and I had about as much to drink as my parents had. But they couldn\'t tell I had been drinking, I kept the truck on the straight and narrow the whole way home. There\'s sooo much more to this story, but my fingers would fall off if I typed it all right now, I just gave you the highlights. Cathrine: Where\'d this baby goat come from, did the goat have babies? Colton: No, the dog did. All: Laughter Man, that was a terrific party. Well, I\'m done for today, see ya!  '),(12027,3485073,'23,July,2004','2004-07-23 00:00:00','   That\'s right, I got money. I went to my grandpa\'s place today after school. I moved a couch downstairs, cut the lawn, and swept the driveway. I made $40! Then I got to drive home. I still don\'t even have my G1 yet, and I\'ve driven sooo many times. I\'m home alone tonight. The stupid dog ran away and I had to go get him with the car, without my parents knowing. But hey, what they don\'t know won\'t hurt me. I\'m also going into my math exam with an 84%, YAY FOR ME! That\'s all for tonight.  '),(12028,3485073,'22,July,2004','2004-07-22 00:00:00','   That title represents two things. One, I beat Spiderman2. The second, and more important one is that I helped two innocent ladies get their car started, I wasn\'t easy either. We tried hooking up to another car, that didn\'t work, so then we had to puch it over to the auto shop and boost it there. It was great. As a reward, I got to take their car for a spin around the parking lot. YAY! Man, I watched the vids from E3 (Electronic Entertainment Expo) for the new Sony PSP and Nintendo DS! They were FUCKING INCREDIBLE!!!!!!!!!! The grapics on the little systems rival those of the game cube and PS2. Plus, WIRELESS GAMING AND WIFI GAMING!!!! For those who don\'t know what that is, basically playing over the net via a wireless network. There\'s soooo much packed into those two little units, I can\'t post it all here. I\'ve become really good friends with the guys at EB Games, I\'m there like everyday. So they showed me the vids. Oh yeah, as far as math class goes, I\'ve got a 79% avg. right now. I keep it up, and Coltons getting some really nice things for his birthday this year. (No, I don\'t always refer to myself in the third person, it just suited here). That\'s it for this post, See ya!  '),(12029,3485073,'20,July,2004','2004-07-20 00:00:00','   Hey guys, I\'ve got some news, I bought 007 Agent Under Fire. I\'m pretty glad, I needed a 4 player FPS (First Person Shooter) for my PS2. I have 3 paddles right now, I\'m gonna buy a fourth soon, maybe the airflow wireless...Today\'s math class was pretty boring, I was done all the work WAY ahead of everyone else and basically sat there for an hour and a half. I swear, if I didn\'t have my Mini Disc player, I\'d go insane in that class. I was nervous today because today the battery was dangerously close to dying (I haven\'t bought batteries in a while). I made $40 yesterday at my grandpa\'s house. I was helping him clean up and set up everything in preparation for some people that came over from Ireland. Man it was hot, I had to work without a shirt all day. (Oh yeah, showin\' off the muscles, j/k). That\'s all for today  '),(12030,3485073,'18,July,2004','2004-07-18 00:00:00','   I say that for lack of a better title. Oh don\'t look at me like that, you read this post and tell me if YOU can think of a better title. That\'s right, you can\'t, because this post sucks too much.&nbsp;But not a whole hell of a lot happened this weekend.&nbsp;I did basically nothing Saturday until I went over to Kev\'s house. Then we played video games all day. I woke up today, packed up my stuff and left. My mom let me drive home again today, I don\'t get my G1 till September and yet I\'ve already driven like 15 times. Ashley, I love you baby. Since this is such a short post, I\'ll give you a little piece of wisdom. Today is the tomorrow you worried about yesterday. Think about that one, it\'ll keep you busy until my next post (hehe).  '),(12031,3485073,'16,July,2004','2004-07-16 00:00:00','   I say that for lack of a better title. Oh don\'t look at me like that, you read this post and tell me if YOU can think of a better title. That\'s right, you can\'t, because this post sucks too much.&nbsp;But not a whole hell of a lot happened this weekend.&nbsp;I did basically nothing Saturday until I went over to Kev\'s house. Then we played video games all day. I woke up today, packed up my stuff and left. My mom let me drive home again today, I don\'t get my G1 till September and yet I\'ve already driven like 15 times. Ashley, I love you baby. Since this is such a short post, I\'ll give you a little piece of wisdom. Today is the tomorrow you worried about yesterday. Think about that one, it\'ll keep you busy until my next post (hehe).  '),(12032,3485073,'16,July,2004','2004-07-16 00:00:00','   I\'m sorry to all of you, I\'ve just been having a bizzarro week, and well, I just haven\'t had the time to post. So, I\'m happy to announce that I\'m currently holding a 71% average in summer school (YAY! that meansI get my G1 in september!). Also, Ash is back from camp (that calls for another YAY!). This week I also discovered a method of playing Spiderman 2 at EB Games. I play the Xbox version, then I just save the game on the HDD and come back and continue from where I left off. God, the guys who sit behind me in my math class are so incerdibly stunned, it\'s not funny. They friggin sit there and attach alligator clips to my shirt. I swear, they have like ADD or something. I walked to the office (Konink Logistics Inc.) in Long Sault today, it took me about 2 hours and it poured on me almost the whole way there. I walked in to the office drenched, my dad was shoked to see me. And the first thing I said was \"Have anything to eat?\". I hadn\'t eaten since breakfast. So he gave me 10 bucks and I went and ate Subway. I\'m sorry I don\'t hav anymore to post right now, but I will Sunday! Till then, I\'m out.  '),(12033,3485073,'09,July,2004','2004-07-09 00:00:00','   Man! Summer school isn\'t that bad, after class is done I get 3 1/2 hours to tour around cornwall. And because of this, EB Games has become my home away from home. So far this week, I\'ve bought two games and traded in for the third: Xenosaga (really good RPG), Intellivision Lives! (oh yeah, old school gaming all the way baby!), and I traded in three of my old games (approx. total value of the three, $20)for Spiderman 2 brand new ($70), EB Games has this deal going on right now, (I\'m talking to you Kev) where you can bring in three used games (for any system) and get either DRIV3R or Spiderman 2 for free. That\'s pretty good considering you usually save like $50. And Kev, I know you\'ve got like 10 games you\'re itching to get rid of, there\'s gotta be at least a couple in that huge collection of yours. It\'s like gaming night in Canada lately. Ash gets back at the end of next week, YAY! Oh yeah, I forgot to mention one thing. I love ashley! So what if she\'s three years younger than me? She\'s really mature for her age, and she looks older than she is. And anyone of yous who has a problem with it, you can go fuck yourselves for all I care. Kind of a harsh way to end a post, but hey, till next time, this is Colton, over and out.  '),(12034,3485073,'06,July,2004','2004-07-06 00:00:00','   Summer school blows! I went in Monday, and even though my mom handed the registration forms herself, I wasn\'t on any of the lists! Fucking dumbasses, I don\'t even think they know the wars over, holy shit. And even though the bus literally drives right by my road, I have to go to Bonville (like 7 kilometers) to get picked up, it can\'t just pick me up at the end of the road. And it doesn\'t stop there either, no, that\'d be too simple, I finish at the half day like most people, and then I have to wait THREE AND A HALF FUCKING HOURS for my bus before I can go home. GOD! Till next time, this is Colton signing out.  '),(12035,3485073,'04,July,2004','2004-07-04 00:00:00','   My head is spinning right now, I thought that I had lost the chance with the girl I love (if that\'s not too strong a word). But now I\'m not sure, she says she loves me, but the type of love she means seems to keep changing, it\'s so hard to tell! She left for her two week camp today, I got to say good bye before she left, that\'s good, and once again she said she loves me. Like I said, I\'m so confused!!! Today, I spent some time with my dog Rusty, we finally untied him, and he hasn\'t ran away! That\'s a good sign. He really is a smart dog. I can\'t think of anything else to post today, except that I start summer school tommorrow (ugh!). At least summer school might bring some interesting stories. I\'m out for now!  '),(12036,3485073,'03,July,2004','2004-07-03 00:00:00','   Man, I got my heart broken (again) today. A couple of more hits and it\'s gonna be completely shattered. Damn, you know what? FUCK LOVE, who the fuck needs it? I think I musta pissed God off pretty bad, he\'s forsaken me to live alone for the rest of my life. But on the flip side, I went into town today, I also stopped by my Dad\'s office and used the T1 line to download the lastest episodes of RVB (yes!!!) That\'ll cheer me up, watching those stupid antics. This was a very sad post, I know, but here\'s a word from to the wise. Don\'t mess with love, all it\'ll ever do is hurt you and scar you. Some of you are lucky, your relationships work out, me, not quite so lucky. Sorry about this guys, hopefully next time I\'ll have something more positive to post.  '),(12037,3485073,'02,July,2004','2004-07-02 00:00:00','   Canada day rocked! I went in a bit late, but after running around like a chicken with my head cut off for a half  hour, i finally hooked up with Ash, Nat, and their friend (Raven I believe). We all hung around and walked around, except for Ash, she ran every where. Man, slow down Ash, you almost killed Nat and them, I thought they were gonna have heart attacks or something. I found a flashy, blinky, thing (Ooh, pretty). The fireworks were pretty cool too, even though they didn\'t even last a half hour. I still can\'t get over the one i saw go off, it flew up in all it\'s shimmering glory, then started to come down, then went back up and exploded! (you heard right, they changed direction in mid air). I also met Anthony\'s girl friend last night, she\'s pretty cool, I really think her personality suits her to Anthony quite nicely. There\'s some other stuff that happened last night too, but I\'m not gonna talk about that here. That\'s all for now!  '),(12038,3485073,'11,August,2004','2004-08-11 00:00:00','   Hey everybody! I don\'t have long, we\'re leaving for home right nmow, post when I get back, see ya!  '),(12039,3485073,'08,August,2004','2004-08-08 00:00:00','   The slogan on my T-shirt. Ok, this is gonna be a long one folks, so grab a drink and sit back. Here\'s the summary of the past week. Ok, when we first arrived, the place looked like shit on the outside (hence the title of the previous post). I thought my mom was gonna freak out and beat someone. But once we got over the outside and decided to step inside, we were pleasantly surprised. It turns out that they had comepletly redone the interior and just hadn\'t got around to the outside yet. We had a beautiful view of the beach and satellite tv (kev, i had G4 techtv there!). The beach was really cool, the tide goes out like a half kilometer. And there was a lot of very hot girls in bikinis on the beach. But away from my fantasies (for now). (Oh yeah, FYI: The name of this town we stayed in is Tatamagouche. We also went to PEI, that place is awsome! there was so much cool stuff. I picked up two cool T-shirts. My fav is the pirate on that says \"beatings will continue until morale improves\". The bridge to PEI is 12km long and we drove the whole thing. We took the ferry back over. The ferry had gambling machines on it. They didn\'t supervise them, they put people on the \"honour system\", big mistake (hehe). I gambled away happily and lost what i gained instead of stopping while i was ahead. We went to a mall in New Glascow. Thats where I picked up the shirt with the slogan \"keep it real\" on it. You need to see the shirt to get the slogan. Man, someone blew a skunk away the night before and it filled the house with the smell. I thought I was gonna die! We\'ve seen these really cool things (i don\'t know what they\'re called). They\'re the amphibious things that are on the first tomb raider movie. Thats all for now, i\'ll finish this tomorrow. bye!  '),(12040,3485073,'04,August,2004','2004-08-04 00:00:00','   I\'ll explain the title later guys. I can\'t give much of an update right now, I\'m on an express terminal and i have 10 minutes. I\'m typing like wildfire right now. Mikaela, I swear to god that I\'ve foung your long lost twin sister in a diner we ate at before we reached our destination. We were shocked when we reached the place we were staying at. The outside looks horrible, but once we got inside, it was amazing. The interior was just redone this year and they have yet to do the outside. I\'m sorry I haven\'t updated earlier thatn this. Kev, I know you\'re probably wondering why don\'t I use my dail-up connection. Simple, long distance connection charges that I\'m not willing to pay. I\'ll give you guys a really long and extensive post when we get to the hotel in Halifax at the end of the week, till then, bye!  '),(12041,4068475,'10,May,2004','2004-05-10 00:00:00','   Well, I just started this blog, but I guess I\'ll add any random crap I can up onto this, right?   How do I add images? Blah...  '),(12042,4068475,'28,July,2004','2004-07-28 00:00:00','   Okay. Didn\'t do much today. Stayed home, played some WC3.   And god, I should stop playing that on Battle.net, actually, just the custom games. People are morons. It\'s like walking into the cesspool of idiocy. Seriously, if you ever get the chance to check it out, I don\'t know whether to advise you to stay away, or to at to prove I\'m not bullshitting you. That place is awful.  Games are bad too, tower defenses, defense of the ancients, I mean, some games aren\'t bad, but try some variety, ya? It gets boring playing the same thing over and over with the handfull of idiots who spout shit like, \"yo lik wtfl ol god dude noob\"  '),(12043,4068475,'27,July,2004','2004-07-27 00:00:00','   urlLink    Picture of my taru character from Final Fantasy XI. I used to use this to kick it with my friend, Christian in-game. But now I just use it to play on whenever I\'m extremely bored. Which doesn\'t happen often now. &nbsp; urlLink  '),(12044,4068475,'27,July,2004','2004-07-27 00:00:00','   urlLink    Another Lil\' John photoshopped thing. This is my mock-up flyer for Lil\' John\'s latest album. :D Rawrl, laugh.&nbsp; urlLink  '),(12045,4068475,'27,July,2004','2004-07-27 00:00:00','   urlLink     It\'s Lil\' John.. or is it?&nbsp; urlLink      It\'s a picture I edited, of Animal from the muppets, what more do you want?   Oh, why I did it? It\'s hillarious. Plus I think Lil\' John is animal. Look at that picture, then try arguing with me. Yeah, that\'s right.    YOU CAN\'T  .  Can\'t argue with facts my friend.  '),(12046,4068475,'27,July,2004','2004-07-27 00:00:00','   urlLink    Alright. This is another recent comic that I made. I guess you won\'t get much unless I do the \"commentary.\"&nbsp; urlLink     Okay. The story behind this is that there was this one idiot on Battle.net. What a surprise, right? Yeah, right. Anyways. This guy, goes on and on about how he hates niggers, and how he\'s going to burn us, even though me and my friend aren\'t black. Then this idiot keeps rambling about how he makes 70 thousand a year, and he\'s 20 years old.  Frankly? I think this guy must be 12, I hope. Or else this leaves little hope for planet Earth if morons like these will inhabit it. Jesus. Just imagine this guy making 70 thousand dollars a year, talking about that stupid shit.  I barely noticed I didn\'t actually describe what\'s going on with the comics.     Well  ... I been watching the Surreal Life recently, that\'s been on Vh1 and has all these old school celebrities and stuff. Mix that in with the above story, and you got yourself a stupid story. There.  '),(12047,4068475,'27,July,2004','2004-07-27 00:00:00','   urlLink    El comic muy chistoso que se trate de mi amigo Christian y los locos de la pelicula, Pulp Fiction. ^_^&nbsp; urlLink  '),(12048,4068475,'27,July,2004','2004-07-27 00:00:00','   Aggh... I accidentally deleted this post, and now I have to rewrite it. So let me rewrite it.  Me and my friend Hunter, had this sprite comic thing going on last summer, and we made some funny comics and all, but then we stopped. Recently though... it has started up again! That\'s right. So check it out, and I\'ll post whatever we come up with, whether it\'s stupid, stupid-funny, not-stupid-funny, or just plain not funny.  If you\'re here, you might as well read the crap I got to say? Right? No. . . Then. . . :(  '),(12049,4068475,'27,July,2004','2004-07-27 00:00:00','   Ah, I\'m scared of this. I\'ll add some stuff. Like.. stuff like, comics, things on my mind.. and uh, pictures maybe? Yes! Pictures.  '),(12050,4068475,'28,July,2004','2004-07-28 00:00:00','   urlLink    What if... Lil\' Jon was Animal? And what if Animal was Lil\' Jon? Well folks, here you have it. Check it out.&nbsp; urlLink  '),(12051,4068475,'28,July,2004','2004-07-28 00:00:00','   urlLink    This is my first comic since work has resumed.&nbsp; urlLink  '),(12052,4068475,'28,July,2004','2004-07-28 00:00:00','   urlLink    Okay. This is another comic. Sprite comic. Yeah. Anyways, this is about.. two things. The Surreal Life, and an idiot on Battle.net.&nbsp; urlLink  '),(12053,4068475,'28,July,2004','2004-07-28 00:00:00','   urlLink    I hope this works. Using a new blog.&nbsp; urlLink     Alright. This comic is about my favorite movie (Pulp Fiction), and my friend (Not favorite). jk jk  Anyways, yeah. It\'s an \"homage,\" \"parody,\" whatever you want to call it, about one of the scenes in the movies. I thought it was funny. It\'s stupid, but yeah. Funny?  '),(12054,4068475,'28,July,2004','2004-07-28 00:00:00','   Alright, first post. Not really, I just recreated from the old one, because well, the title didn\'t really fit. Anyways...  I\'ll be moving the stuff over, and my friend will also add stuff, and post stuff, whatever.  '),(12055,4068475,'17,August,2004','2004-08-17 00:00:00','   No one reads this shit, but if someone did, Luis got mouth surgery or some shit and probably won\'t make a comic for a while.  I made the last comic, so until he\'s made another, I\'ve reached my quota, bitches.  '),(12056,4068475,'08,August,2004','2004-08-08 00:00:00','   Rick James is dead.  :( Let the mourning begin.  '),(12057,4068475,'02,August,2004','2004-08-02 00:00:00','   urlLink    MTV Cribs goes to Christian\'s house.&nbsp; urlLink  '),(12058,4068475,'02,August,2004','2004-08-02 00:00:00','   urlLink    Run run and run!&nbsp; urlLink  '),(12059,3441506,'27,May,2004','2004-05-27 00:00:00','   I\'m here at work, and I was just sent on An Errand. The errand was to pick up a book from another office, the Foreign Languages Department, here on campus. And so I did, and it was a nice walk in the soupy heat over to this building, which used to be called Lakeside. I\'m glad they changed the name because there is no Actual Lake, there is only a pond, and you can tell whoever came up with the building name knew it wasn\'t quite right, but he or she couldn\'t very well call it   Pondside  , for that name would be completely unattractive. (\"Yeah, I\'ve got to my next class over at    Pondscum ,   I\'ll see you later....\")   Walking into the old Lakeside is like entering a freezer. On the second floor, I wander up and down the halls a few times looking for Room 278, feeling like I\'m in a labyrinth, new halls and doorways appearing everywhere, students and university employees peering curiously at me from their desks. And then I find the office, finally, and I step inside, and I look around to see how    nice    it is in there. Look how    nice    it is. Lit by corner halogen lamps rather than overhead flourescence, the dim office contains two Ecko-ish desks facing each other, spotless ceiling-high shelves, new flat-screen computers, carpet that is flat and institutional, yet pleasing (color-wise) to the eye. The two ladies working there maintain pleasant conversation with a very pregnant co-worker, speaking in low tones before breaking off to ask how they can help me today.   \"I\'m here to pick up a book?\"   I feel so young. I feel, incidentally, just like I have in past jobs where I have been sent on errands to obtain Important Things from Older, Wiser People. People with Real Jobs who looked at me, the cute, slightly meek office \"helper,\" college-aged and hourly-waged, one day destined for Bigger, Brighter Things. A college education, even more schooling, amazing job, nice house, beautiful family. This will be me.   Except now I am 26. I have my Master\'s degree (from this very university). There will be no more school. And, at least for this summer, I am still the Office Helper.   What comes next?  '),(12060,3441506,'27,May,2004','2004-05-27 00:00:00','   FEELING BETTER ABOUT ME: AN ACTIVITY LIST 1. drive to Target and wander aisles in manner of old, senile man at airport 2. sing in car as though auditioning for American Idol 3. pretend you are the tetchy British judge Simon, critique your pop star performance 4. use the U-Scan-It checkout line at grocery store 5. purchase odd-looking, exotic new fruit 6. find some leaves and jump in them 7. jaywalk 8. water house plants as part of Let them Live campaign 9. sticky-roll away dog hair on couch 10. wash dishes and put them back in cupboards 11. bake cookies from scratch 12. make some oatmeal and give myself a facial 13. make some oatmeal and eat it 14. decoupage coffee table 15. window shop online  16. design a T-shirt like kids with winged hair in the late â€˜70s and early â€˜80s used to wear  17. make ice cube-tray Popsicles like Pee-Wee Herman did 18. watch episode or two on DVD of Felicity: The Complete First Season 19. ponder sweater collection 20. decide: Noel or Ben? 21. microwave a bag of popcorn 22. play Popcorn Game with Bender-dog 23. watch poorly-plotted, chick-flick movie fraught with clichÃ© 23. devise dissertation-length argument for brilliance, social relevance contained in movie  24. wear white platform shoes and dance on the table like Pee-Wee Herman did  24. eat cereal in a coffee mug while wearing pajama pants  26. pull hair up in a messy ponytail 27. do everything wearing pajama pants 28. do everything wearing hair up in a messy ponytail 29. with Bender as your MTV Theater partner, act out scenes from Sorority Life 30. eat pre-packaged cookie dough with spoon 31. tape back nose and eyes, make faces in mirror like Pee-Wee Herman did  32. watch any episode of Dr. Phil, realize life is not so bad 33. continue letter-writing campaign to Dr. Phil for advice on feeling better about life 34. rearrange furniture 35. get lost in closet, make plans to organize 36. eat small bowl of pickle spears  '),(12061,3441506,'02,June,2004','2004-06-02 00:00:00','   urlLink    The Cat Lady, with doggie Bender.&nbsp; urlLink  '),(12062,3680589,'20,June,2004','2004-06-20 00:00:00','   I was born to the problem:      	A matriarchy full of feminists that bowed and cowtowed to the shuffle and scrape of the masculine boot, but told me I must never do the same. From the cradle I was told not to let a man define me, but taught to serve them til my food grew cold. Never let them think they have you, never sit in their chair. An endless game with the male as the opponent. It needs an opponent.  A more succinct problem: 	I was never beautiful. I had fewer chips to play the odds. I learned to jump through hoops instead. I learned to search faces for one with laughing eyes. One who refuses to play by their rules. Who sees me as an equal. When he sees me, heâ€™ll know me and everything will be well. A larger problem: 	We are a lost generation. We are isolated from each other. Our mothers taught us not to trust the other sex in the playing of the game. We live in one place and work in another. Our isolation increases. The rules of engagement foiled the playing of the game in our work space. I get older. The battle is now fought in crowded rooms where the exchange is stunted. I lose. An unforseen problem: 	I get older, the rules have changed once again. They dictate taking everything that we are, winding it down into a critical mass. A few paragraphs, to read. Because no one goes out anymore. The game is now played on the underground-the internet. I play the odds, write an ad. Titled: The Next 30 Seconds. It reads: I know that I have only the next 30 seconds for you to decide whether to answer this ad. How can anyone decide a whole lifetime of possibilities in 30 seconds? Life is an awesome thing. If I learn in the course of mine to appreciate each moment then I will die happy. As this moment hangs before you for your inspection, what are you thinking? Will you answer? Did I say something that reached you deep at the heart of what you are?   	I receive 100 responses. Iâ€™m very popular with accountants. I go on 30 dates. 20 of them fall in love with me on the phone. I date one for almost a year. But it never works out. Even though I increase the odds. Maybe because I increased the odds. But there had been something lurking-something fundamental. And I think: Am I subverting nature? Have our lifestyles and upbringing preverted nature so that we have difficulty coming together and staying together? 	        I met someone recently who was born on the same day I was, the same hour. We drink the same tea. We read the same books, we think many of the same thoughts.        And heâ€™d always been alone. Itâ€™s what I had always rallied against. At least when there was an opponent, you werenâ€™t alone. But maybe, the real problem is: that I am supposed to be alone. And there would be nothing left to say. '),(12063,3680589,'07,July,2004','2004-07-07 00:00:00','   This is in response to a posting I saw on a website where the girl was questioning what the American Dream was:  I just thought I\'d share this with you because I thought of it when I read your post. I saw my aunt recently and because she is a little too outspoken we had an unfortunate confrontation in front of my fiance and my two children (ages 13 and 8). She basically told me that I was a bad mother because I wasn\'t working and was instead staying home to homeschool my children. I am an artist and writer and I\'ve made a lot of sacrifices to have a career that means that I can be home with them. She said that I would set a better example for them by showing them how to work hard so that I could buy a house and cars (we ride our bikes or take the bus everywhere when my fiance isn\'t around). I told her that I didn\'t care about houses and cars or having money, although at the time I wondered what my children really thought about all that. The minute we left the house my oldest daughter exploded. She said, \"How dare she say that about you, Mom! Who the hell does she think she is? Everyone in the world is supposed to live the same kind of life? I\'m glad that you are home for me. If it means that we don\'t have money, that\'s fine. Having you is more important.\" In my American Dream, people would be more important than the almighty dollar. And it would be reflected in the choices they make. '),(12064,3680589,'01,July,2004','2004-07-01 00:00:00','   Don\'t worry folks, this is NOT about me. This is a song I\'m in the middle of writing. Any suggestions are appreciated.  Heâ€™s in the back of  the Satellite, rolling his cigarettes with the old papers he found in Daddyâ€™s things the sky is black with clouds and the night but thereâ€™s no rain again and I think ainâ€™t that just the way  The road is sweeping us away down the dusty hollows  into the grave 	 And I need to keep moving along escaping the movement of my own mind. And he keeps time with the grind on the AM radio and the click of my teeth and weâ€™re running away.  The road is sweeping us away down the dusty hollows  into the grave  Ainâ€™t it just the way. I can feel the dirt of the grave Iâ€™m one handful away It took me ten years  to get the needle out of my arm and I have nothing left that I ainâ€™t already gave '),(12065,3680589,'01,July,2004','2004-07-01 00:00:00','   I met a man that changed my life. 	     Itâ€™s not just that I love him so much that it fills up every part of me or that he is like a drug, intoxicating and addictive. Everything is made better by him. My whole life and the lives of my children work better because of him. For the first time in my life, everything FITS.      So many times over the last few years I wished that I could just die, so that I would be released from the burden of living.  I was seriously mentally and physically ill for six years. My life gradually became a nightmare of endlessly prescribed pharmeceuticals and struggling just to be conscious and take care of the kids. There were periods of utter hopelessness. For the sake of the kids I remained, went through the motions of life, of planning, but at the core of who I was I believed that my life was really over. And then I met Jodie and everything changed. I am so happy, happier than Iâ€™ve ever been in entire life. There are some people whose prescence is so powerful that they set whole movements of life in motion. The way they look at you, noticing everything and still accepting and understanding you and somehow still LOVING you, create explosions of self-growth. Jodie did this for us. For Bridget, Megean, and me, we have begun to change and evolve into the people that we were meant to be. Love does that to you.  There are no words for how grateful I am for this. He gave me back to myself. He made me believe in life again, and the fact that I knew all along that someone somewhere was meant for me.  All my life I have wanted so much to love someone like this, to believe in someone the way I believe in him. I was a person who was meant to have a partner, to be one part of a whole, but was completely without my other half. Having him love me has given me back my faith in everything.  I wanted to feel connected to someone the way I feel connected to him. But it was so impossible. Now I believe that it happens- that there really is someone for everyone. Jodie says all the time that God made us for each other. But I had become thick-skinned. And blind. I see now that I could never have chosen for myself. I had to be taken by force. But God wouldn\'t let me go. And neither did Jodie. I tried so hard to push him away when we first contacted each other, to not let anyone in. Finally, in an act of desperation to get him out of my hair I agreed to meet him. I figured we\'d meet, he\'d get over whatever he foolishly thought he felt for me, and then I could go on with my life. But the minute he walked in my front door it was all over for me.       I have given him a hard fight. Being in love is not easy. It\'s frightening.  Sometimes itâ€™s hard to take a breath and Iâ€™m overwhelmed by it all. Still, even if itâ€™s all taken away tomorrow, I am so grateful for this moment of happiness. Of being given back my life again. Of being brought back to myself. Because Jodie has had that effect on me. I\'ve been opened up in ways I didn\'t know was possible. And I love that man. Sometimes at night I can\'t sleep. I just lying there, watching him sleep,making sure he is really there and that it wasnâ€™t a dream. Jodie asked me yesterday why I loved him and I replied: I love you because you are the best man I\'ve ever met. You are kind and compassionate and honorable. You are a man of good character. But more than that, you are attractive to me in every way. The quickness of your mind, your ability to memorize and understand and make inferences from complex information and situations, both concrete facts and and personal relationships. I find you physically beautiful. You have strong lines, which I love. I love to stare at you while you sleep. The way you smell is intoxicating to me, it has this strange chemical effect on me.  Your voice is entrancing and soothing. And the complexity of your character and observations on everything are a constant revelation to me.  I feel like I\'m discovering new territory every day.Like I\'m on an adventure each and every day. And I realized that it was just part of your nature and that I would never be bored. And I realized  that I would never come to the end of it. You love my children, you have made them yours and I thought that would never happen. The way that you deal with the world is sublime. You set the best example for me that I\'ve ever had and I love you for that. I love you for all of that. There are no words really that will ever adequately explain why I love you.  This is just a feeble try. The way I feel about you is sacred and deep, an ancient well that is fathomless. '),(12066,3680589,'01,July,2004','2004-07-01 00:00:00','   Today reminded me of a joke I used to tell a lot...  What\'s the last thing that goes through a fly\'s mind when he hits the windshield?   His ass.   I\'m also inspired to quote the great Dire Straits song \"The Bug\"   well it\'s a strange old game - you learn it slow one step forward and it\'s back to go you\'re standing on the throttle you\'re standing on the breaks in the groove \'til you make a mistake  (chorus) sometimes you\'re the windshield sometimes you\'re the bug sometimes it all comes together baby sometimes you\'re a fool in love sometimes you\'re the louisville slugger sometimes you\'re the ball sometimes it all comes together baby sometimes you\'re going to lose it all  you gotta know happy - you gotta know glad because you\'re gonna know lonely and you\'re gonna know bad when you\'re rippin\' and a ridin\' and you\'re coming on strong you start slippin\' and slidin\' and it all goes wrong because  (chorus)  one day you got the glory one day you got none one day you\'re a diamond and then you\'re a stone everything can change in the blink of an eye so let the good times roll before we say goodbye, because  (chorus)   Today I\'m gonna do the  Top Ten Things You\'d Never Want To See At Your Own Funeral:   10. A kiosk full of AOL disks. (They\'re everywhere!!!)  9.  Pews with high speed internet access ports. (Just let it go...)  8.  A one man protest group. (I want a crowd, dammit!)  7.  Casket paid for by product placement.  (It worked for Dale Earnhard) (I think I just earned my crowd...)  6.  A bouncer enforcing a two dollar cover charge. (Instead, have a three drink minimum!)  5.  Any member of Menudo. (New, old, they\'re all icky.)  4.  IRS agents and a state trooper. (Haven\'t you guys got enough???)  3.  Political figures I voted for. (You\'ve disappointed me enough, go away.)  2.  Religious people that normally mug you at the airport. (I\'m dead, no saving me now.)  and the number one thing you\'d never want to see at your own funeral:  1.  Three words... Schlitz Malt Liquor! (Don\'t forget to spill some for me homies!)  I\'m outtie!   Peace!  '),(12067,3680589,'01,July,2004','2004-07-01 00:00:00','   I\'m announcing, right here and now, that  urlLink this  is the ring that I want for my fiancÃ©e and me.  Somehow I\'ll be able to get it for us.  Isn\'t it beautiful?  '),(12068,3680589,'01,July,2004','2004-07-01 00:00:00','   It\'s been a very long weekend... starting with meeting my fiancÃ©e at the hospital on Wednesday, staying there until four in the morning to find out she\'s got a kidney stone, missing work on Thursday, disappointing the kids and cancelling our Disneyland trip on Friday, working on my father\'s computer on Saturday and spending Sunday depressed out of my skull but working up a nice anxiety attack and a huge case of \"dumping syndrome\" anyway.    It\'s my life and right now I\'m at work wishing that I had a bottle of whiskey waiting at home.  Instead, I\'m blessed to be able to return home to my three girls and I have to remind myself that they\'re the most important things in my life.  So, in order to escape without the alcohol it\'s time to come up with...   The Top Ten Reasons Estrogen Is Better Than Alcohol.   10. It kicks your ass and  KEEPS ON KICKING!   9.  No jury ever acquitted because of an  \"alcohol defense.\"   8.  I only feel embarassed at the checkout stand once a month. (think about it)  7.  Brawling, cursing, fits of uncontrollable rage, loud music... once a month it\'s a party!  6.  I love you, get away from me, I love you, get away from me... schizophrenia with no cover charge.  5.  I\'m starting to think the cats are the only ones that actually listen to me.  4.  I can\'t fight off the desire to drink with chocolate.  3.  God is a marketing genius because men fall for the packaging every time!  2.  Hey, no one ever heard of the  bottle behind the man , did they?  and the number one reason that estrogen is better than alcohol is:  1.  There are far fewer people attending support groups to get over it.  We do get over it, don\'t we?  Hello?   Hello?  '),(12069,3680589,'01,July,2004','2004-07-01 00:00:00','   Okay, so today\'s listing is all about an  educational institution  I need to gripe about.  I won\'t name any names but it\'s an online university that comes from the capital of Arizona.   Wink, wink, nudge nudge.   I can\'t say who it is because they are the BIGGEST client for the company I work for.  Who am I kidding?  They\'d be ANYONE\'s biggest client that deals with education.  The organization, as a whole, appears to be reputable and offers a lot of service to their students/clients/customers... whatever.  The material they are studying is valid, granted.  But when I find out that they\'re demanding that we take course material that we designed for a twenty to twenty-two week course at a \"regular\" higher learning facility and cram it into one of their five week courses I cringe.  When I learn that they want their students to accelerate through the curriculum by allowing them to complete topics by answering one question correctly instead of five I\'m aghast.  When I get calls from these very students and they don\'t have the basic computer skills of a five year old child with severe learning disabilities I get very upset.  Upset, hell, I\'m pissed.  These neophytes are getting a sub-level education, paying way too much and waving their pseudo-diplomas around like they\'ve really accomplished something.  I\'ve got news for these knuckleheads.  I\'ve met people that got one of their \"certified graduate\" certificates and I\'ve met people with bona fide diplomas in the exact same subjects.  My money\'s on the ones that went to accredited higher learning facilities.  Sure, it took them five times as long, they had to put up with underpaid, overworked professors who may not actually have \"real world\" experience in their chosen fields, but they seem to have a brain, common sense and the ability to avoid pissing on their own shoes.  The business class students I\'ve talked to from some schools are sometimes pushy but rarely outright rude.  The students from the online circus tend to be so rude their ego needs it\'s own freakin\' atmosphere.  Well, I could go on, and I will eventually, but it\'s my lunch hour and I should probably actually eat something.  One of their \"diplomas\" is currently being used as a trivot on my desk.  '),(12070,3680589,'01,July,2004','2004-07-01 00:00:00','   No, the one on the orange.  Get your fingers outta there, you weirdo!  I\'m still on this spam rampage.  These people are a menace.  Honestly.  I\'m just glad that they\'re only wasting a few electrons and people\'s time.  The rate at which they\'d be killing trees before the advent of the internet is staggering.  So, in order to get it out of my system, off my chest and into your lives I\'ve decided I need to dedicate a top ten to them.   The Top Ten Responses To Spam:   10. Thank you for informing me of my penile inadequacy, but I\'m a woman.  9.  Please file my application for dating services under \"lonely inmates with a photographic memory.\"  8.  If your pharmacy needs any new suppliers I can hook you up with my cousin Fernando in Colombia.  7.  Please send all copies of your software titles to the address of your local FBI office listed below.  6.  My gawd!!  That\'s the best rate ever!  And you can finance people with no job, bad credit and a list of fraud charges a mile long, right?  5.  Um, I love you, too... Who are you?  4.  Mail System Error, huh?  That\'s okay, I\'ll send it via carrier rhino, next time.  3.  Thank you.  You\'ve just sent your webcam invitation to a lonely, violent sex offender with no impulse control.  2.  Please send me more information on how to get rid of you.  and the number one response to spam:  1.  I would like to say thank you by ordering a round of laxatives for your entire organization.  '),(12071,3680589,'01,July,2004','2004-07-01 00:00:00','   Sitting here, at a laptop I didn\'t buy, on a table I didn\'t choose, drinking diet Dr. Pepper from a glass that I  did  buy... at a dollar store.  This is my life.  I work a nine hour day and have to be on the freeway for three and a half hours to do it.  I get up at six o\'clock in the morning and leave for work by seven-thirty to get home at eight and be in bed, hopefully, by eleven.  Just to do it all over again...   Why?   Because this is my house (okay, apartment and it\'s a rental) and it\'s my home.  I live here.  The woman that lived here first invited me into her life and does things for me every day to prove to me that she loves me.  The kids actually  tell me  they love me and they  mean  it.  Teenagers don\'t tell anybody they love them unless they want something.  I should know, I was a teenaqer, once, briefly... a lifetime ago.  My home, my family, my life.  I accept it, I work for it and I need to appreciate it for what it is.  A blessing.  So tonight I\'m grateful.  I\'m loved.  I\'m safe.  I\'m tired and I\'m going to bed.  My bed.  That\'s all I need.  '),(12072,3680589,'01,July,2004','2004-07-01 00:00:00','   I can use this blog to gripe about anything I want.  I finally have a way to bitch about all the spam I get at home and, more especially, at work!  I\'m not joking, here.  I have to filter  THOUSANDS  of junk e-mail every single day.  After a normal weekend my filters caught eight thousand out of ten thousand junk e-mails.  But I spent the first two hours of my day just going through the rest.  I\'m telling the world at large,   right now...   I don\'t want to gain three inches, it\'ll ruin my whole wardrobe. I\'ve no interest in lasting thirty-six hours, that would kill my fiancee. I don\'t want to sign up for a dating service for horny housewives, they need marriage counseling instead. I don\'t need every e-mail address in the world on one CD, I can barely keep track of my address book as it is. I don\'t believe that an internet millionaire wants to share his secrets with me. I don\'t believe that I can\'t share in the inheritance from the death of an African royal. I\'m aware that my computer may contain a virus, website information, cookies and various pastries, as well.  Your product isn\'t needed to delete it.  I know how to clear my cache, site history and use my  freakin\'  delete key. I don\'t believe I sent you an application for a damn mortgage. No one told you that I\'d be fun to talk to and should check out your pics. I\'ve no interest in the key to four hundred porn sites... hell, I don\'t even want one. I don\'t want/need prescription medication for impotence, hair loss (okay, I may  need  that one) or pain from an offshore pharmacy that has the lowest prices. I don\'t want to learn how to stop annoying e-mails from the people sending them to me. I can\'t read Chinese, Korean or Russian and the German I know tells me that the guys sending all the e-mails about German tourists being murdered needs to start his own online zine and leave the rest of the news to the professionals because I don\'t want it in my damned inbox. I invest poorly enough on my own (I own several shares of an investment company that I worked for) and I don\'t need advice from random e-mails. No, I\'m not fat anymore, but you calling me that wouldn\'t entice me to buy your stinking pills. My fiancee has not informed me that she\'ll leave me if I don\'t buy your product that will magically make me a legendary love god... you lied to me and that makes me very angry.  Go away. Your software prices aren\'t the cheapest and I don\'t want to pay to have illegal pirated copies shipped from India, Korea or anywhere else overseas, thanks. I haven\'t even financed the first time so I don\'t need to refinance, besides I get enough crap about my lousy credit from people I trust.  My final observation...   If you don\'t know how to spell I\'M NOT BUYING ANYTHING FROM YOU!   Have a great Monday, everybody, and keep smiling!  '),(12073,3680589,'01,July,2004','2004-07-01 00:00:00','   Today was a good day.  I got nothing done and I enjoyed every minute of it.  My gorgeous fiancee woke me up at five o\'clock this morning for a brief but very intense discussion... yeah, that\'s it.  Then, a couple of hours later, her beautiful face was next to the bed and directly behind her the two sweet faces of her daughters.  Well, our daughters.  Very soon I\'ll be marrying into this wonderful little family and I\'m adopting the girls as my own.  For so many years I\'ve been a father without my children, but now I have found a woman who wants me, needs me and loves me as much as I love her.  Along with her come two incredible kids who want me, need me and love me as much as I love them.  What else am I going to do but love them for the rest of my life.  This year\'s Father\'s Day was the sweetest I\'ve ever had.  I hope they are all this good.  The eight year old made a simple card from printer paper and crayons.  It was a masterpiece.  The thirteen year old made me breakfast.  Because of my gastric bypass surgery I just can\'t eat much at one sitting.  It was simply a handful of whole wheat wafers with cream cheese and half an avocado with a strong cup of coffee.  Perfect.  Of course, after a sumptuous meal what\'s better than a mid-morning snooze.  I slept until eleven, when my fiancee came back into our bedroom and we took up the same topic we were discussing early that morning.  Bliss.  We then went to the store where I attempted to purchase all the things we needed to prepare a large meal for both of my parents, my mother\'s birthday was last week.  Well, it cost more money than I expected and took longer to prepare than I wanted to spend but it was a better than average meal that everyone enjoyed.  Joy  Today\'s top ten is all about being a Dad.  The Top Ten Things Dad\'s Get Out of Father\'s Day:  10. This card is made from 100% recycled paper... and so is your tie.  9.  Happy Father\'s Day... Can I borrow the car?  8.  Sorry Dad, I didn\'t know that you were allergic to shellfish.  7.  Honey, are you going to actually sleep  all  day?  6.  It\'s perfectly normal for a grown man to cry.  5.  It tastes great, (hack) honest.  No, (cough) really, there\'s just (gag) enough pepper, sweetie. (sneeze)  4.  We even got the dog to sign your card... that\'s what that smell is.  3.  A little turpentine, a claw hammer, some chlorine bleach and a hasp should take that right out.  2.  We didn\'t know what your first name was so we had it engraved \"Dude.\"  and the number one thing Dad\'s get out of Father\'s Day:  1. Baby, I\'m all gift wrapped... wanna open your present?  '),(12074,3680589,'01,July,2004','2004-07-01 00:00:00','   This is a regular thing for me.  Whenever I think of something I start breaking it down into categories and sub-groups.  Out of that is born a natural need to turn my twisted sense of humor into a list.  Call it an Obsessive Compulsive Bi-Polar thing.  Call it stupid.  Call it Milo... I don\'t care.  But it\'s my special thing and I\'m going to run with it.  Today\'s top ten is inspired by the fact that it\'s Friday and I\'ve got a whole weekend of nothing important to do.  Top Ten things this dork can waste an entire weekend with:  10. You\'re reading it...  9.  Lord of the Rings trilogy - extended edition (the kids will do Golem for a week)  8.  I must resist the power of eBay!  7.  Wi-Fi, baby.  6.  No credit card required. (if you know what this means seek help)  5.  So many deals at Fry\'s, so little money.  sigh...   4.  Where the hell are my pants???  3.  I have a digital camera, half a tank of gas and it\'s summer in Southern CA... hmm.  2.  I wonder how my Sims characters are doing.  and the number one thing this dorky dork can waste an entire weekend with:  1.  Where the hell are  your  pants???  I\'m going home.  I\'m sure I\'ll have more from the \"apartment of doom.\"  '),(12075,3680589,'01,July,2004','2004-07-01 00:00:00','   Most of the blogs that I read are just one more way for people to complain.  This one isn\'t going to be any different.  However, I tend to register my complaints from the twisted viewpoint of someone with a deep-seated need to point at the rest of the world and laugh.  Otherwise I may start listening to the voices in my head and no one wants that.  I feel the need to customize my new toy.  More later.  '),(12076,3680589,'01,July,2004','2004-07-01 00:00:00','   I\'ve been thinking about the things that couples go through.  Mostly because I\'m going through them... again.  So here\'s a top ten for all you lovers out there.   The Top Ten Rare Conditions Only Couples Contract:   10. MIF: (not Mission Impossible Force) Menstrual Irritation Factor.  It\'s measured in the level of irritation you cause your partner once a month.  9.  DASS: Dig And Scratch Syndrome.  This gets worse as couples are together because guys grow more comfortable scratching places they should only touch privately.  8.  WDYD: Could be \"What Did You Do?!?!\" but it\'s actually Why Don\'t You... Disease.  A condition that forces couples to expect the same things they did to each other while they were dating.  7.  IHR: I Hate Romance.  This one is the cynical symptom of couples who have lost that magical spark... and found it when the fire department arrived.  6.  GAFM: Get Away From Me!  This can occur at any time for no particular reason.  5.  WDYLT: Where Did You Learn That?  This is a disorder that can be both blessing and curse.  Cook something your partner has never had before, say something you told yourself you\'d never repeat or, the most severe case, make the move you\'ve in bed you\'ve never had the courage to try and you could be either hero or villain.  4.  INST: I Never Said That!!!  This one is usually only contracted by men because most women have the unnerving ability to repeat conversations you had with them  years ago,  verbatim.  3.  DTMML: Does This Make Me Look... It\'s a twisted trap as old as time.  Why do women ask this question of a gender that yells at people on a television screen as though they can hear them?  Men will smell their clothing to determine whether or not it can be worn.  These same neanderthals will shove mold riddled food in your face and ask, \"Does this smell like it\'s gone bad to you?\"  You trust their opinion of how an item might alter your appearance?!?!?  2.  YOMH: You\'re On My Hair.  My favorite malady because men, for centuries, have thought so much of their masculine prowess until they realize they\'re just ripping their partner\'s scalp from their head.  Oh yeah, baby...   and the number one rare condition that only couples contract:   1.  WTIIT: Who\'s Turn Is It To... a word of advice to men here... If you\'re asking the question, it\'s probably your turn.  If the person you love is asking... it\'s still probably your turn.    Just get up and do it, you knucklehead!  '),(12077,3680589,'01,July,2004','2004-07-01 00:00:00','   It\'s been a bittersweet day... and I want to get this over with quickly.&nbsp; Today\'s top ten was inspired by something Jackie thought was funny.   The Top Ten Things You Don\'t Want To Hear After Visiting The Gym:   1o.&nbsp;Did those pants fit when you came in?  9.&nbsp; Actually, the limp makes you look kinda sad.  8.&nbsp; No, I don\'t think the Sound Of Music soundtrack is appropriate for working out.  7.&nbsp; Maybe you should consider taking a shower... again.  6.&nbsp; The doctor will see you now.  5.&nbsp; I\'m afraid that we need to refund the rest of your membership fee, less the damages.  4.&nbsp; Poor posture, lack of endurance, unhealthy breathing... maybe you should give up the porn.  3.&nbsp;&nbsp;I\'m sorry sir,&nbsp;I thought you were just crank calling 911 from your cellphone.  2.&nbsp; They call them sweats, not sponges.   and the number one thing you don\'t want to hear after you visit the gym:   1.&nbsp; Sure, you look great.&nbsp; If you\'re an anemic, injured, senior citizen.  '),(12078,3680589,'01,July,2004','2004-07-01 00:00:00','   Fast-food is going to be the downfall of western civilization.&nbsp; You\'ll just have to trust me on this.&nbsp; Where else can you spend $4.99 on substances that can barely be pronounced, encouraged to super-size your ability to clog your arteries at the speed of grease and also get a refill for a quarter?&nbsp; People are so proud of their ability to provide cheap death in a bun that they put their names all over the stuff... Wendy\'s, Carl\'s Jr., McDonald\'s, Tommy\'s, etc.&nbsp; It wasn\'t called Hitler\'s they just call it the Holocaust but the casualty rate wasn\'t nearly as high...  Top Ten Things To Do When You\'re Bored In A Fast-Food Restaurant:  10. Go to the counter and ask if they can \"super-size something that\'s not on the menu.\"  9.&nbsp; Randomly ask people if they find the&nbsp;establishment \"olfactory.\"  8.&nbsp; Wear a cape to the restaurant, jump up from your chair when you\'re done and run back to your car shouting, \"I\'ll be right there!\"  7.&nbsp; When they bring you your food ask if the \"poor, defenseless animal was killed humanely or subjected to the horrors of an anonymous mass slaughter like it didn\'t even have a soul.\"  6.&nbsp; Ask if they offer alternatives to the \"kid\'s meals\" with appropriate toys like \"Buddhist meals,\" \"Hedonist plates,\" or \"Sado-masochistic sack lunches.\"  5.&nbsp; Offer to \"autograph\" people\'s napkins.  4.&nbsp; When people try to sit at the table next to you don\'t look up from your meal but say, \"Stop it.&nbsp; They can sit there if they want to.&nbsp; You can\'t make me do that.\"  3.&nbsp; Write \"Official Straw Monitor\" on a napkin, tuck it in your collar and pass straws out to people.  2.&nbsp; When the person at the counter asks if they can help you hold up four fingers and shout, \"I\'m this many!!\"  and the number one thing to do when you\'re bored at a fast-food restaurant:  1.&nbsp; Leave an anonymous note for the manager saying, \"I don\'t want to alarm you but I think the chair I was sitting in may actually be an exact replica of Malaysian shrine where 500 clowns lost their lives.\"  '),(12079,3680589,'01,July,2004','2004-07-01 00:00:00','   Okay, so tonight I\'m trying to produce two top tens because I want to start doing two a day so that   I can publish.&nbsp; I warn&nbsp;you, I\'m a little jaded because I\'ve had several days to&nbsp;consider the fact that I\'m a recently working stiff trying&nbsp;to compete in a world where everyone wants to be&nbsp;self-employed.&nbsp; The first is one inspired while actually laughing during a very serious conversation I was having with Jackie.&nbsp; Sad, but true...  &nbsp;   The Top Ten Novelty CDs That Just Never Made It Â© :   &nbsp;  10.&nbsp;Bodily Noises From The Motion Picture Deliverence.  &nbsp;  9.&nbsp; The Cast From \"Who\'s Company?\" Sing Barry Manilow\'s Greatest Hits.  &nbsp;  8.&nbsp; Don\'t You Want Me Baby by Sally Fields.  &nbsp;  7.&nbsp;&nbsp;Lounge Songs From The Sixties by The Grateful Dead.  &nbsp;  6.&nbsp; 100 Bottles Of Beer On The Wall by Various Artists.  &nbsp;  5.&nbsp; Ozzy Ozbourne Mumbles Shakespeare.  &nbsp;  4.&nbsp; Binging And Purging To The Hits by Calista Flockhart.  &nbsp;  3.&nbsp; Music To Spasm To by Julia Louis-Dreyfus.  &nbsp;  2.&nbsp; Rick James Sings&nbsp;About His Bitch.&nbsp;  &nbsp;   and the number one novelty CD that just didn\'t make it:   &nbsp;  1.&nbsp; Jailhouse Rock by Martha Stewart   &nbsp;   Okay, and now for Top Ten number two...   &nbsp;   The Top Ten Names For Celebrity Perfumes You\'ll Never See Â© :   &nbsp;  10.&nbsp;Endorsement: Your name is worth more than you are.  &nbsp;  9.&nbsp; Jagged Little Pill: How much can you swallow?  &nbsp;  8.&nbsp; Cleavage: I\'m up here... hello?!?!  &nbsp;  7.&nbsp; Kids And Animals: Hollywood\'s not that bad... is it?  &nbsp;  6.&nbsp; Danger: Put down the crack pipe.  &nbsp;  5.&nbsp; Andy Warhol: \"There\'s nothing behind it.\"  &nbsp;  4.&nbsp; Puke: Try not to get it on you.  &nbsp;  3.&nbsp; Fifteen Minutes: Save for a rainy day.  &nbsp;  2.&nbsp; Purdy: A scent as fleeting as beauty.  &nbsp;   and the number one celebrity perfume you\'ll never see:   &nbsp;  1.&nbsp; Plastic Surgery: I\'m as real as you want me to be.  '),(12080,3680589,'01,July,2004','2004-07-01 00:00:00','   I never met the man but I grieve his passing.&nbsp; I remember summer nights when I was younger being brought half to tears listening to this soul filled blind man, born in a segregated Albany, Georgia, sing, ironically, the national anthem.&nbsp; As a young man, Ray represented the&nbsp;tenacity it takes&nbsp;to overcome the steepest of mountains and still maintain the integrity of dreams and the ability to touch people with unbridled sincerity.&nbsp; What kind of man does it take to bare your soul to millions of people in the painful birthing process of creating a song?&nbsp; Especially when you know that your music will be criticized, scrutinized and rejected out of hand simply because people don\'t know how to believe that&nbsp;a successful&nbsp;black man&nbsp;can still have a pure talent.&nbsp; Ray, you set standards that will remain unchallenged for generations.&nbsp; Did you intend to change music?&nbsp; Did you intend to turn the entire musical world on it\'s ear with your music or did you just write a song because you enjoyed the music?&nbsp; God speed, Ray.&nbsp; You\'ve earned your place in the celestial choir.&nbsp; In fact, I\'m hoping that you\'re enjoying your new found place composing songs that bring God to tears the way yours did for me.  '),(12081,3680589,'01,July,2004','2004-07-01 00:00:00','   It\'s all crude mood and attitude around here lately.  I live in a two bedroom apartment with three women.  I ask you, wouldn\'t any sane man have his patience tested in the best of conditions???  Well, I just went from being a regular working stiff to being home twenty-four/seven.  Hello?  I love each and every one of these women, truly, but when tempers begin to fray and the monthly hormone explosion creates a shockwave that would level the crowd at a World Wrestling Federation match it\'s difficult not to hide behind a mask of indifference.  I\'ve been contributing my own brand of heartless comments without thinking about the sensitive ears listening.  As it is, every person in this apartment has had her feelings hurt, several times, and I\'ve said that I\'m sorry more times than Oliver North.  Family life can be a little hazardous...  it\'s downright extreme when you\'re bipolar.   Today\'s top ten, therefore, is going to reflect my current situation...   The Top Ten Things To Install In A Luxury Doghouse:   10. Drool cloth dispenser.  9.  High-Definition Smell-o-Vision.  8.  Hot & cold running toilet water.  7.  Authentic \"Chase The Car\" treadmill.  6.  A.C.M.E. inflatable leg. (for those lonely nights at home)  5.  Hydrant with infrared flush-o-matic sensor.  4.  Crunchy cat crap vending machine.  3.  \"Just like the real thing\" sun-lamp.  2.  Butt-scent spray and sniff station.  and the number one thing to install in a luxury doghouse:  1.  Ronco meaty bone burying oven.  '),(12082,3680589,'01,July,2004','2004-07-01 00:00:00','   It\'s too true.  One minute you\'re bitching about how stressed out you are about everything and the next you\'re being told by your psychiatrist that you can\'t go back to work... wait, that actually makes sense.  This morning Jackie sat down with me and we told my psychiatrist how difficult things have been for me recently.  A three and a half to four hour commute and working so far from home.  So the doctor told me that he\'s putting me on disability for three months.  Uh, okay, where the hell did that come from?!?!  Yesterday I was worried about looking for a job here in the valley and today I\'m on disability?  Yeah, twenty-four little hours.  I can\'t say that I\'m completely relieved but I\'m not nearly as worried about finding the time to take care of things.  Now Jackie and I can work, in earnest on getting our home-based business things started.  I\'ve got ideas for printing and selling some of my photography.  She\'s going to be making hemp jewelry that we\'ll be selling on eBay, hopefully, and maybe at some of the local \"flea market\" events.  We are going to work very hard on getting the two cookbooks together and ready for pitching to publishing companies.  I\'m going to gather some editorial type material to see if I can sell some of my writing.  Jackie wants me to gather a collection of my top ten lists and possibly put together a book, so you saw it here first!  And amongst it all I have time to find a good job nearby.  Okay, so maybe the disability won\'t be so bad.  Hey, I worry... it\'s how I got in this situation in the first place.  So, in that vain, and because tomorrow is my thirty-sixth birthday, today\'s top ten is going to be about worrying and the art of aging... gracefully... maybe.  The Top Ten Things People Ask When They Realize They\'re Getting Older:  10. Do chocolate chocolate chip cookies come in decaf?  9.  How many calories does sex actually burn, anyway?  8.  Is a Last Will and Testament just a way of avoiding a garage sale???  7.  Does this job make me look fat?  6.  Do my kids need an inheritance as much as I need a red convertible?  5.  Why does my laundry look like a medieval torture chamber?  4.  Do these pill organizers come in a briefcase version?  3.  When is it actually okay to use the words \"When I was your age\"?  2.  How many carbs do vitamins have?  and the number one thing people ask when they realize they\'re getting older:  1.  Is hair really fashion or just lost and found?  '),(12083,3680589,'01,July,2004','2004-07-01 00:00:00','   1. What time do you get up?  6:AM when I\'m working, when I\'m not it usually depends on how long the COMA LASTS!    2. If you could eat lunch with one person, who would it be?  Jackie, because she\'s reading this and will kill me if I don\'t say her.  No, seriously.     3. Gold or silver?  Golver... no, sild... wait... what\'s the question?     4. What was the last film you saw at the cinema?  What the hell is a cinema?     5. What is/are your favorite TV show(s)?  Power off and staring at Jackie...     6. What did you have for breakfast?  Coffee, woven wheat crackers and brie. (It\'s a rough life, but I don\'t have a stomach!)    7. Who would you hate to be stuck in a room with?  Any self-help guru... throttling\'s too good for them.    8. What/who inspires you?  Jackie, God gave her to me and she\'s my favorite, so there.    9. What is your middle name?   Slightly-off-center... and Allen   10. Beach, City or Country?   Country - fresh air, stars and wild, untamed animals... no, wait, that\'s the neighbours...    11. Favorite ice cream?  Until recently it was peppermint, now it\'s no sugar added Klondike bars... cold chemical goodness   12. Butter, plain or salted popcorn?  Lather and sprinkle it on, baby!   13. Favorite color?  Olive green with a hint of \"what is that?\"   14. What kind of car do you drive?  A little white piece of crap that I\'m grateful for   15. Favorite sandwich?  Cucumber, but I can\'t have bread     16. What characteristic do you despise?  What characteristic... nyah, nyah, yada, yada, yada... Uh, mocking... hate it.   17. Favorite flower?  Alstromeria for looks, Gardenia for fragrance and whole wheat for baking   18. If you could go anywhere in the world on vacation, where would it be?  Back to DC... I wanna live in the Smithsonian...   19. What color is your bathroom?  When it\'s clean, tan-ish... when it\'s not it\'s too hideous to open my eyes!!!!!!!!   20. Favorite brand of clothing?  Geoffrey Beene, if you must know.    21. Where would you retire to?  Grass Valley, CA...      22. Favorite day of the week?  Payday   23. What did you do for your last birthday?  *hic*    24. When is your Birthday?  Monday (no, really)   25. Where were you born?  Spiral arm of the Milky Way galaxy, I think   26. Favorite sport to watch?  Football!!!!!!!!!!!!!!!!!!!!!!!!  Go Niners!  Woohoo!   27. Who do you least expect to send this back to you?   (insert your name here)   28. Person you expect to send it back first?  Skippy the wonder hamster    29. What fabric detergent do you use?  The white, powdery one with little flecks   30. Coke or Pepsi?  diet Dr. Pepper (I like being different, too)    31. Are you a morning person or a night owl?  Insomniac, hello?    32. What is your shoe size?  13 wide or 14... I know, I know... Sasquatch borrows my shoes    33. Do you have any pets?  3 cats, 2 kids and one very angry dust bunny that is trying to kill me in my sleep  '),(12084,3680589,'01,July,2004','2004-07-01 00:00:00','   I wanted so much to be able to update this blog on a daily basis. So much for good intentions. I\'m home full time and I thought it meant that now I was away from my job \"rehabilitating\" I\'d have a great deal more time to pursue my writing and photography. Well, I\'ve been  urlLink taking pictures ... but only while I\'ve been out running errands every day. It seems that everything else has become my full time job. Everywhere I look there is something that needs to be fixed, cleaned, put away, reorganized, labeled, folded... you get the idea. Being home is hard work. So, Jackie, the MOST wonderful woman alive, kindly bought me a new desk and is actively trying to get me to finish resurrecting my desktop computer, that I slightly cannabalized to fix my father\'s system, once and for all. Once I get that computer up and running it will be my little corner of the world. A small slice of sanctuary in a crowded apartment. Ah, bliss...    \"Life is funnier when you are anxiously anticipating the punchline.\"   - Right Rev. Rap Masta Cornflake    So, here I am thinking about writing, photography and the age old art of making money without being hired by corporate America. So here\'s a twisted top ten for the blue-collar crowd... we\'re all in this together.   The Top Ten Reasons  NOT  To Kill Your Boss:    10. Orange cover-alls have NEVER been in fashion... especially in the exercise yard.  9. It\'s so important to show respect... without the crosshairs.  8. Because you\'ll simply be replacing them with someone much more annoying and demanding... that  you  pay.  7. No one has ever said, \"Show them how you feel... with explosives.\"  6. It\'s very difficult to forward your mail to a penitentiary.  5. Stalking self-help gurus is much more satisfying.  4. At the office you get cigarette breaks... not broken for cigarettes.  3. Writing a proposal under a deadline is  infinitely  easier than writing a confession under guard.  2. Pictures of your ass, attached to a resignation, can be enjoyed for years.   and the number one reason NOT to kill your boss:   1. \"Yes sir/ma\'am\" is easier to say than, \"Can\'t we just cuddle?\"  '),(12085,3680589,'01,July,2004','2004-07-01 00:00:00','   The wayward eighteen year old has come home to roost for a night.  She\'s got to go to court tomorrow morning to deal with a problem with her driver\'s license.   Don\'t ask.   So with all the girls under our roof I\'m considering things from a different perspective.  Here\'s a one for parents.   The Top Ten Things You  Don\'t  Want To Hear Come Out Of A Child\'s Mouth:   10. (when asked about something broken) It wasn\'t me.  9.  (referring to a myriad of commands) It\'s not fair!  8.  (when asked about clothing they\'ve obviously soiled) I don\'t have anything else to wear!  7.  (when the older child is given permission) How come they get to do it?  6.  (when asked if they\'ve bathed recently) I think so, wait, no, I can\'t remember.  5.  (from a teenager) I hear you say it all the time.  4.  (when caught in the act) I didn\'t know I wasn\'t supposed to do that.  3.  (if asked to do work) I don\'t feel good.  2.  (this one is just icky) I think the cat\'s sick.   and the number one thing you don\'t want to hear come out of a child\'s mouth:   1.  (HORROR!) Mom, I found this in your nightstand.  '),(12086,3886562,'04,August,2004','2004-08-04 00:00:00','   1.é—œé–‰å€‹äººç¶²é å…±äº«  ç³»çµ±åå¥½è¨­å®š-å…±äº«-æœå‹™-å€‹äººç¶²é å…±äº«  åœæ­¢   é–‹å•Ÿphpæ”¯æ´    sudo /Applications/TextEdit.app/Contents/MacOS/TextEdit /private/etc/httpd/httpd.conf  è¼¸å…¥ç®¡è£¡è€…å¯†ç¢¼  ----- å°‹æ‰¾ ----- #LoadModule php4_module        libexec/httpd/libphp4.so  ----- æ”¹æˆ ----- LoadModule php4_module        libexec/httpd/libphp4.so  ----- å°‹æ‰¾ ----- #AddModule mod_php4.c  ----- æ”¹æˆ ----- AddModule mod_php4.c  2.å•Ÿå‹•å€‹äººç¶²é å…±äº«  ç³»çµ±åå¥½è¨­å®š-å…±äº«-æœå‹™-å€‹äººç¶²é å…±äº«  å•Ÿå‹•   å‡ç´š php 4.3.8  1.ä¸‹è¼‰ php 4.3.8 æ‰“é–‹-æ‡‰ç”¨ç¨‹å¼-å·¥å…·ç¨‹å¼-çµ‚ç«¯æ©Ÿ-éµå…¥  curl -O http://tw2.php.net/get/php-4.3.8.tar.bz2  tar -xjf php-4.3.8.tar.bz2  cd php-4.3.8  ./configure \'--prefix=/usr\' \'--mandir=/usr/share/man\' \'--infodir=/usr/share/info\' \'--with-apxs\' \'--with-ldap=/usr\' \'--with-kerberos=/usr\' \'--enable-cli\' \'--with-zlib-dir=/usr\' \'--enable-trans-sid\' \'--with-xml\' \'--enable-exif\' \'--enable-ftp\' \'--enable-mbstring\' \'--enable-dbx\' \'--enable-sockets\' \'--with-iodbc=/usr\' \'--with-curl=/usr\' \'--with-config-file-path=/etc\'  make  sudo make install  è¼¸å…¥ç®¡è£¡è€…å¯†ç¢¼  sudo cp php.ini-dist /etc/php.ini  å®‰è£ MySQL 4.0.20  1.ä¸‹è¼‰ MySQL 4.0.20 http://mysql.isu.edu.tw/Downloads/MySQL-4.0/mysql-standard-4.0.20-apple-darwin7.3.0-powerpc.dmg  2.å®‰è£ MySQL 4.0.20 é€™å…©å€‹éƒ½è¦è£ mysql-standard-4.0.20-apple-darwin7.3.0-powerpc.pkg MySQLStartupItem.pkg  3.è£å¥½å¾Œé‡æ–°é–‹æ©Ÿ  å®‰è£ phpMyAdmin-2.5.7-pl1  1.ä¸‹è¼‰ phpMyAdmin http://belnet.dl.sourceforge.net/sourceforge/phpmyadmin/phpMyAdmin-2.5.7-pl1.tar.bz2  2.è§£å£“ç¸®å¾Œæ”¾åˆ°ç¶²é ç›®éŒ„ è§£é–‹æ”¾åˆ° http://localhost/phpMyAdmin/ æˆ–æ˜¯ http://localhost/~ä½¿ç”¨è€…åç¨±/phpMyAdmin/  ç›®éŒ„å°æ‡‰é—œä¿‚  http://localhost è³‡æºåº«\\WebServer\\Documents  http://localhost/~ä½¿ç”¨è€…åç¨±/ ä½¿ç”¨è€…\\ä½¿ç”¨è€…åç¨±\\ç¶²ç«™ 3.è¨­å®šå¸³è™Ÿå¯†ç¢¼ é–‹å•Ÿ http://localhost/phpMyAdmin/ æˆ–æ˜¯ http://localhost/~ä½¿ç”¨è€…åç¨±/phpMyAdmin/  è¨­å®šrootå¯†ç¢¼ æ¬Šé™ root ä½¿ç”¨è€…åç¨±.local æ˜¯ ALL PRIVILEGES æ˜¯ ç·¨è¼¯ root localhost æ˜¯ ALL PRIVILEGES æ˜¯ ç·¨è¼¯ ç·¨è¼¯-æ›´æ”¹å¯†ç¢¼ å¯†ç¢¼:ä½ çš„å¯†ç¢¼ ç¢ºèªå¯†ç¢¼:ä½ çš„å¯†ç¢¼ åŸ·è¡Œ å…©å€‹éƒ½è¦æ”¹  æŠŠä¸Šé¢å…©å€‹å‹¾èµ·ä¾† åˆªé™¤ä½¿ç”¨è€…åŠé‡æ–°è®€å–æ¬Šé™ åŸ·è¡Œ  æ–°å¢žä½¿ç”¨è€… æ¬Šé™-æ–°å¢žä½¿ç”¨è€… ä½¿ç”¨è€…åç¨±:æ–‡å­—è¼¸å…¥:è‡ªå·±å¡« ä¸»æ©Ÿ:æœ¬åœ°:çœ‹ä½ éœ€è¦è€Œå®š å¯†ç¢¼:æ–‡å­—è¼¸å…¥:ä½ çš„å¯†ç¢¼ ç¢ºèªå¯†ç¢¼:ä½ çš„å¯†ç¢¼ æ•´é«”æ¬Šé™:å…¨å‹¾(çœ‹ä½ éœ€è¦è€Œå®š) åŸ·è¡Œ  ç”¨æ–‡å­—ç·¨è¼¯æ‰“é–‹æ‰“é–‹phpMyAdmin\\config.inc.php ----- å°‹æ‰¾ ----- $cfg[\'PmaAbsoluteUri\'] = \'\';  ----- æ”¹æˆ ----- $cfg[\'PmaAbsoluteUri\'] = \'http://localhost/phpmyadmin/\';  ----- å°‹æ‰¾ ----- $cfg[\'Servers\'][$i][\'auth_type\']     = \'config\';  ----- æ”¹æˆ ----- $cfg[\'Servers\'][$i][\'auth_type\']     = \'http\'; '),(12087,3886562,'04,August,2004','2004-08-04 00:00:00','   é€™æ˜¯ç¬¬äºŒæ¬¡çœ‹é€™éƒ¨ç‰‡ï¼Œé‚„æ˜¯å¿«æŽ‰æ·šã€‚  é™¤äº†å°çˆ¶æ„›çš„æ„Ÿå‹•å¤–ï¼Œæˆ‘æ¬£è³žSamçœ‹ä¸–ç•Œçš„æ…‹åº¦ï¼Œä»–ç¸½æ˜¯å¯ä»¥çœ‹åˆ°æœ€å–®ç´”çš„ä¸€é¢ï¼Œè€Œæˆ‘å»æ˜¯åœ¨å­¸ç¿’å¦‚ä½•æ›´å–®ç´”çš„çš„é¢å°æˆ‘æ‰€è™•çš„ä¸–ç•Œï¼Œè¦ºå¾—è«·åˆºã€‚  ä»–çš„å¾‹å¸«å°ä»–èªªï¼Œè·Ÿä½ åšæœ‹å‹æˆ‘å¾—åˆ°çš„æ¯”ä½ æ›´å¤šã€‚  å‚»å¯ä»¥è®“äººåŸ·è‘—ï¼Œè®“ä½ çœ‹äº‹æƒ…æ›´æ¸…æ¥šã€‚  '),(12088,3886562,'04,August,2004','2004-08-04 00:00:00','   è·Ÿåœ‹å®¶å¹«åŽ»çƒä¾†çƒ¤è‚‰ï¼Œçƒ¤æ²’å…©ä¸‹å°±é–‹å§‹ä¸‹å¤§é›¨ï¼Œå¾Œä¾†åˆ°å±±ä¸‹åŠ æ²¹ç«™åˆç‰Œã€‚  ç”²ï¼šè½å©†å˜´ï¼Œå¤§å¯Œè²´ã€‚  ä¹™ï¼šé‚£ä¹Ÿè¦è€å©†è°æ˜Žã€‚  '),(12089,3886562,'04,August,2004','2004-08-04 00:00:00','    æœ€å¥½çš„æž¶phpBBçš„å…è²»ç©ºé–“  ç¶²å€: http://www.freesites.ws 15 MB ç©ºé–“ æ”¯æ´ PHP æ”¯æ´ PERL æ”¯æ´ SSI 1 MySQL FTP ä¸Šå‚³ftp.freesites.ws ç„¡å»£å‘Š  1. é¦–é ä¸­æ‰¾ - Click Here to Signup! - ä¸¦æŒ‰å…¥ã€‚  2. çœ‹å®Œä¸€å †å­—ä¹‹å¾Œåœ¨ä¸‹é¢æŒ‰ Acceptã€‚  3. ä»¥ä¸‹æ˜¯å¡«è¡¨çš„æ–¹æ³• First Name: å§“å Last Name: åå­— Email Address: E-mail (æ­£ç¢ºèˆ‡å¦ä¹Ÿå¯) Address: æœ‰æ•¸å­—è·Ÿè‹±æ–‡å°±å¯ City: Hong Kong / Tai Wan State/Province: äº‚å¡«æ•¸å­— Postal Code: äº‚å¡«æ•¸å­— Country: Hong Kong / Tai Wan Phone: äº‚å¡«æ•¸å­—  4. å¸³è™Ÿå¡«è¡¨ Account Username: å¸³è™Ÿå Account Password: å¸³è™Ÿå¯†ç¢¼ Domain: åŸŸå  5. ä¹‹å¾Œå°±å®Œæˆç¬¬ä¸€æ­¥çš„è¨»å†Š, ç­‰åˆ°ç¶²é è¼‰å…¥å¾Œæœƒæœ‰ä»¥ä¸‹çš„å¥å­: Please be patient while your signup is processed. This process may take up to 1 minute. [ Customer accountID test successfully added ] Congratulations on your new account! Please check your email for your login information.  6. ä¹‹å¾Œåœ¨å·¦é¢ç™»å…¥, æœƒåŽ»åˆ° CPenal, è¼¸å…¥æ­£ç¢ºçš„å¸³è™Ÿèˆ‡å¯†ç¢¼ã€‚  7. é€²å…¥ä¹‹å¾Œåœ¨ä¸Šæœƒæœ‰ Please update your contact information here æŒ‰å…¥å’Œæ›´æ–°ä½ æ­£ç¢ºçš„ E-Mail.  8. å…¨éƒ¨å®Œæˆ, ä»¥ä¸‹æ˜¯æž¶ phpBBX_SP çš„æ•™å­¸ã€‚   æž¶ phpBBX_SP çš„æ•™å­¸  1. å…ˆé–‹ä¸€å€‹ MySQL Account, åœ¨ Cpenal ä¸ŠæŒ‰å…¥ MySQL Databases.  2. çœ‹ Databases: , åœ¨ä¸‹é¢æœ‰ä¸€å€‹ Db: çš„è¼¸å…¥æ¬„, è¼¸å…¥ä½ çš„è³‡æ–™åº«åç¨±, ä¹‹å¾ŒæŒ‰ ADD Db.  3. Users: ä¸­è¼¸å…¥å¸³è™Ÿèˆ‡å¯†ç¢¼ã€‚  3. ä¸‹é¢æ˜¯å®‰è£ phpBBX SP çš„è¼¸å…¥è³‡æ–™. è³‡æ–™åº«ä¼ºæœå™¨ä¸»æ©Ÿåç¨±: localhost æ‚¨çš„è³‡æ–™åº«åç¨±: ä½ çš„å¸³è™Ÿ_å‰›å‰›è¼¸å…¥ Db çš„åç¨± è³‡æ–™åº«ä½¿ç”¨è€…å¸³è™Ÿ: CPenal çš„å¸³è™Ÿ è³‡æ–™åº«å¯†ç¢¼: CPenal çš„å¯†ç¢¼  CPenal çš„å¸³è™Ÿ: php CPenal çš„å¯†ç¢¼: 123 Db: æ˜¯ test  å°±å¦‚ä»¥ä¸‹å¡«: è³‡æ–™åº«ç¶²å€: localhost è³‡æ–™åº«åç¨±: php_test å¸³è™Ÿ: php å¯†ç¢¼: 123   '),(12090,3886562,'04,August,2004','2004-08-04 00:00:00','    é€£çºŒå¹¾æœŸçš„æ¨‚é€é ­å½©æ²’äººé‡ï¼ŒçŽé‡‘ä¸Šçœ‹å¥½åƒæ˜¯åå„„å§ï¼Œå…¨çœé™·å…¥è¬äº‹å ±æ˜Žç‰Œï¼Œçœ‹é¯‰é­šè·³è™Ÿç¢¼ï¼Œä¹Ÿæœ‰äººæ‹šå‘½åšå¤¢ï¼Œå¤¢æ˜Žç‰Œï¼Œæˆ‘ä¾ç„¶æ˜¯å …æŒä¸è²·æ¨‚é€ï¼Œä¹Ÿè¨±å“ªå¤©å¤§ä¾¿ã„œå‡ºå¼µå½©å·ï¼Œæœƒåš‡å¾—ä¸­çŽå§ã€‚  ä¸€å·åœ¨æ‰‹å¸Œæœ›ç„¡çª®ã€‚   '),(12091,3886562,'04,August,2004','2004-08-04 00:00:00','    1.é—œé–‰å€‹äººç¶²é å…±äº«  ç³»çµ±åå¥½è¨­å®š-å…±äº«-æœå‹™-å€‹äººç¶²é å…±äº«  åœæ­¢   é–‹å•Ÿphpæ”¯æ´    sudo /Applications/TextEdit.app/Contents/MacOS/TextEdit /private/etc/httpd/httpd.conf  è¼¸å…¥ç®¡è£¡è€…å¯†ç¢¼  ----- å°‹æ‰¾ ----- #LoadModule php4_module        libexec/httpd/libphp4.so  ----- æ”¹æˆ ----- LoadModule php4_module        libexec/httpd/libphp4.so  ----- å°‹æ‰¾ ----- #AddModule mod_php4.c  ----- æ”¹æˆ ----- AddModule mod_php4.c  2.å•Ÿå‹•å€‹äººç¶²é å…±äº«  ç³»çµ±åå¥½è¨­å®š-å…±äº«-æœå‹™-å€‹äººç¶²é å…±äº«  å•Ÿå‹•   å‡ç´š php 4.3.8  1.ä¸‹è¼‰ php 4.3.8 æ‰“é–‹-æ‡‰ç”¨ç¨‹å¼-å·¥å…·ç¨‹å¼-çµ‚ç«¯æ©Ÿ-éµå…¥  curl -O http://tw2.php.net/get/php-4.3.8.tar.bz2  tar -xjf php-4.3.8.tar.bz2  cd php-4.3.8  ./configure \'--prefix=/usr\' \'--mandir=/usr/share/man\' \'--infodir=/usr/share/info\' \'--with-apxs\' \'--with-ldap=/usr\' \'--with-kerberos=/usr\' \'--enable-cli\' \'--with-zlib-dir=/usr\' \'--enable-trans-sid\' \'--with-xml\' \'--enable-exif\' \'--enable-ftp\' \'--enable-mbstring\' \'--enable-dbx\' \'--enable-sockets\' \'--with-iodbc=/usr\' \'--with-curl=/usr\' \'--with-config-file-path=/etc\'  make  sudo make install  è¼¸å…¥ç®¡è£¡è€…å¯†ç¢¼  sudo cp php.ini-dist /etc/php.ini  å®‰è£ MySQL 4.0.20  1.ä¸‹è¼‰ MySQL 4.0.20 http://mysql.isu.edu.tw/Downloads/MySQL-4.0/mysql-standard-4.0.20-apple-darwin7.3.0-powerpc.dmg  2.å®‰è£ MySQL 4.0.20 é€™å…©å€‹éƒ½è¦è£ mysql-standard-4.0.20-apple-darwin7.3.0-powerpc.pkg MySQLStartupItem.pkg  3.è£å¥½å¾Œé‡æ–°é–‹æ©Ÿ  å®‰è£ phpMyAdmin-2.5.7-pl1  1.ä¸‹è¼‰ phpMyAdmin http://belnet.dl.sourceforge.net/sourceforge/phpmyadmin/phpMyAdmin-2.5.7-pl1.tar.bz2  2.è§£å£“ç¸®å¾Œæ”¾åˆ°ç¶²é ç›®éŒ„ è§£é–‹æ”¾åˆ° http://localhost/phpMyAdmin/ æˆ–æ˜¯ http://localhost/~ä½¿ç”¨è€…åç¨±/phpMyAdmin/  ç›®éŒ„å°æ‡‰é—œä¿‚  http://localhost è³‡æºåº«\\WebServer\\Documents  http://localhost/~ä½¿ç”¨è€…åç¨±/ ä½¿ç”¨è€…\\ä½¿ç”¨è€…åç¨±\\ç¶²ç«™ 3.è¨­å®šå¸³è™Ÿå¯†ç¢¼ é–‹å•Ÿ http://localhost/phpMyAdmin/ æˆ–æ˜¯ http://localhost/~ä½¿ç”¨è€…åç¨±/phpMyAdmin/  è¨­å®šrootå¯†ç¢¼ æ¬Šé™ root ä½¿ç”¨è€…åç¨±.local æ˜¯ ALL PRIVILEGES æ˜¯ ç·¨è¼¯ root localhost æ˜¯ ALL PRIVILEGES æ˜¯ ç·¨è¼¯ ç·¨è¼¯-æ›´æ”¹å¯†ç¢¼ å¯†ç¢¼:ä½ çš„å¯†ç¢¼ ç¢ºèªå¯†ç¢¼:ä½ çš„å¯†ç¢¼ åŸ·è¡Œ å…©å€‹éƒ½è¦æ”¹  æŠŠä¸Šé¢å…©å€‹å‹¾èµ·ä¾† åˆªé™¤ä½¿ç”¨è€…åŠé‡æ–°è®€å–æ¬Šé™ åŸ·è¡Œ  æ–°å¢žä½¿ç”¨è€… æ¬Šé™-æ–°å¢žä½¿ç”¨è€… ä½¿ç”¨è€…åç¨±:æ–‡å­—è¼¸å…¥:è‡ªå·±å¡« ä¸»æ©Ÿ:æœ¬åœ°:çœ‹ä½ éœ€è¦è€Œå®š å¯†ç¢¼:æ–‡å­—è¼¸å…¥:ä½ çš„å¯†ç¢¼ ç¢ºèªå¯†ç¢¼:ä½ çš„å¯†ç¢¼ æ•´é«”æ¬Šé™:å…¨å‹¾(çœ‹ä½ éœ€è¦è€Œå®š) åŸ·è¡Œ  ç”¨æ–‡å­—ç·¨è¼¯æ‰“é–‹æ‰“é–‹phpMyAdmin\\config.inc.php ----- å°‹æ‰¾ ----- $cfg[\'PmaAbsoluteUri\'] = \'\';  ----- æ”¹æˆ ----- $cfg[\'PmaAbsoluteUri\'] = \'http://localhost/phpmyadmin/\';  ----- å°‹æ‰¾ ----- $cfg[\'Servers\'][$i][\'auth_type\']     = \'config\';  ----- æ”¹æˆ ----- $cfg[\'Servers\'][$i][\'auth_type\']     = \'http\';   '),(12092,3886562,'04,August,2004','2004-08-04 00:00:00','    Blogç‚ºä»€éº¼éƒ½æ˜¯PHPé›£ç”¨æ–ƒäº†ï¼Œæˆ‘ä¸æ‡‚PHPå¾ˆå¤šéƒ½éœ§ç…žç…žã€‚  ç¶²é è¨­è¨ˆæ˜¯ä¹Žç„¡æ³•è·Ÿç¨‹å¼è„«é›¢é—œä¿‚ï¼Œååæˆ‘åˆä¸æœƒç¨‹å¼ï¼ŒåŠ ä¸Šå¾Œç«¯é˜¿å¸•å¥‘è·Ÿæˆ‘çš„è¥¿æ‘³ï¼Œè¶Šæ˜¯ä¸ç”¨éŒ¢çš„è¶Šæ˜¯è¤‡é›œï¼Œæ‡‰è©²æ˜¯æˆ‘è¦ä»€éº¼åŠŸèƒ½ï¼Œæ‹‰ä¸€æ‹‰å°±å¥½äº†ï¼Œæœƒé»žå½±åƒè»Ÿ é«”ï¼Œå¯ä»¥è™•è£¡æˆ‘è¦çš„æ•ˆæžœï¼Œè€Œä¸æ˜¯è®“ç¶²é è¨­è¨ˆå¸«çš„æ™‚é–“éƒ½èŠ±åœ¨å¯«ç¨‹å¼ï¼Œå†ä¾†æ˜¯ç€è¦½å™¨è‡ªé‡æ›åˆ°éº¥é‡‘å¡”å¾Œæ‰çŸ¥é“æ²’æœ‰æŽ¢éšªå®¶çœŸæ˜¯å¯¸æ­¥é›£è¡Œï¼ŒçœŸä»–åª½ï¼Œå¾ˆå¤šç¶²é åœ¨æ²™ç™¼ç‘ž ä¸Šéƒ½è®Šå¾—äº‚ä¸ƒå…«ç³Ÿï¼Œé‚„æœ‰å­—è®Šè¶…å°ï¼Œè¦ç”¨æ”¾å¤§é¡æ‰èƒ½çœ‹ï¼Œå¤–åœ‹äººå¯èƒ½éƒ½æ˜¯æ‹¿è‘—æ”¾å¤§é¡ç”¨é›»è…¦å§ï¼Œå°ç£æˆ‘çŸ¥é“æ˜¯ä¸ç”¨ï¼Œé«”è«’ä¸€ä¸‹å°ç£å§ï¼Œæ´‹é¬¼å­ã€‚   '),(12093,3886562,'04,August,2004','2004-08-04 00:00:00','    åœ¨å…¬åœ’è·Ÿå¥³æœ‹å‹åµæž¶   '),(12094,3753470,'28,July,2004','2004-07-28 00:00:00','    Think I\'ll try black font today.&nbsp; Well the last day or two has had alot of ups and downs.&nbsp; We did the spanish day thing and saw Spider Man 2.&nbsp; It was pretty good but some parts seemed funny when they probably shouldn\'t have.&nbsp; I now owe two people money and have none so I need to get some.&nbsp; OOH, somewhat funny story for you: Erica and Brittany pulled up in my driveway as I was talking to my Mom shirtless and shoeless on the deck...I got ready quickly and found that I had no money (well, like a 2 dollar bill and a whole lot of change, which didn\'t do me much good.)&nbsp; So I asked mis madre for some and she reveals that she has no cash.&nbsp; So we take the unusual approach to it and she writes a check.&nbsp; Rushing out the door we (Erica, Brittany and I) went and picked up Kevin and we stopped by the bank.&nbsp; I thought for a second and said \"I don\'t know what to do...someone come with me.\"&nbsp; So Kevin and I stood in front of the machine densely and said...\"Hmm, well I guess we have to go in.\"&nbsp; Needless to say we were like the most conspicuous looking people in the bank that day and we stood around dumbfoundedly for a moment before Kevin said he\'d just pay my way in.&nbsp; We then went and ate at subway and lots of confusion occured in the ordering process and Kevin ended up not having enough for my ticket and Brittany had to pay my way in.&nbsp; So, anyway, that\'s how I owe people money.&nbsp;   I\'m sad today because my dog had to be \"put to sleep\" (killed) and that sucks.&nbsp; She was bought as a puppy for me&nbsp;from the pound when I was like 4 or 5 and she was the only red pup in a litter of black ones.&nbsp;   I quit air alert 3 last night...I just didn\'t feel like doing it and it was getting to push more towards torture than exercise.&nbsp; I\'m not sure if&nbsp;that\'s a good thing or not because Brandon is still doing it.&nbsp; Speaking of Brandon: I believe I FINALLY&nbsp;decided on his position in the band...he\'s gonna learn guitar and be a rythm guitarist.&nbsp; With the motivation&nbsp;of&nbsp;being in a band and me telling him \"Come on, you\'re holding us up you son of a pig!\" he\'ll probably learn relatively quickly.&nbsp;   That\'s all I can think of&nbsp;right now so, have a good day.&nbsp;   '),(12095,3753470,'26,July,2004','2004-07-26 00:00:00','    Hey!&nbsp; Time for more mindless ramblings of my personal, non-sexual, affairs.&nbsp; I couldn\'t think of a title&nbsp;but I&nbsp;saw my ferret&nbsp;run&nbsp;by and seeing how that\'s a totally random little creature&nbsp;he seemed worthy of being&nbsp;the title of my&nbsp;post.&nbsp;&nbsp;     &nbsp;   I went to Gannon\'s yesterday and saw a bootlegged addition of Anchor Man (is that supposed to be two words?&nbsp; Oh well)&nbsp; It was really funny.&nbsp; We explored all his little barns and stuff&nbsp;that are for some reason scattered throughout his&nbsp;backyard.&nbsp; I got trapped in a little house and like 5 wasps buzzed around in my face when&nbsp;I tried to open a door...miraculously I wasn\'t stung but they blocked my window and&nbsp;door escape routs so I found another window and&nbsp;climbed right into a thorn bush to get out.&nbsp; He also burned Secret Window but&nbsp;either my&nbsp;@%$#&! up&nbsp;PS2 and DVD player or his @%$#&! up disc wouldn\'t work.&nbsp;&nbsp;ANGER!!&nbsp; By the way, whoever you are, you should buy that movie cuz it\'s really good.     &nbsp;   Our&nbsp;(people from spanish)&nbsp; Spanish Day is tomorrow and that should be fun.&nbsp; Holy crap I just saw a transfer truck (well...a really big truck of some sort) back into a driveway across the street...that\'s gotta be hard.&nbsp;&nbsp;     &nbsp;   Ze band name (if my opinion on best-thought-of-thus-far band names&nbsp;makes the final decision) is Gone Astray.&nbsp;&nbsp;I chose this cuz it sounds cool, it\'s catchy, and it could have alot of different meanings.&nbsp; I&nbsp;thought about it&nbsp;and Brandon&nbsp;may have a&nbsp;more minimal&nbsp;role in the&nbsp;band than I was thinking.&nbsp; Of course we\'ll&nbsp;probably have a long uncalled for arguement on the subject that will hurt relations but I&nbsp;believe it\'s for the best&nbsp;so that\'s the call.&nbsp;&nbsp;     &nbsp;   Well that\'s about all for now so...see ya later.&nbsp;   '),(12096,3753470,'23,July,2004','2004-07-23 00:00:00','    My cousin spent the night at my house (yes I kicked my brother and parents out) last night and my brother was driving us INSANE!&nbsp; (if you\'ve ever been to my house you\'ll know what I mean.)&nbsp; So we became shirt ninjas (there\'s a rather simple technique where you turn a shirt inside out and kinda put it on and tie it just so, so that it looks like a ninja mask) and went outside and up the road just walking around.&nbsp; Not wanting to go back home we went up the road to a new housing developement called \"Huntstone\" and walked around up there.&nbsp; Now...I didn\'t think anyone lived there yet....I found that I was wrong.&nbsp; Walking past some houses I saw two different houses with a car in the driveway and one little dog&nbsp;wouldn\'t stop barking.&nbsp; We couldn\'t turn back now because we didn\'t know if it would take longer to go back or forward so we kept going and eventually made it to the main road.&nbsp; As we came over the crest of a hill I saw my dad pull out of the driveway and go up the road, when we got down the driveway we saw him go past the house and by the other way (strange).&nbsp; Brandon and I went to the backyard and were hanging out when my dad came up (didn\'t see us at first and went to the shop) and told us quite demandingly to \"Get in here.\" Here being the shop.&nbsp; Once we were inside he asked where we went and we said up the road and he said \"Well a lady from up at Huntstone called the police and said she saw two young men walking around up there and she was afraid they were gonna tear something up\".&nbsp; Holy crap!&nbsp; But...then my dad told us that he was kidding and the police hadn\'t been called.    &nbsp;  I still haven\'t started reading...and I haven\'t written (book) in several days...DAGNABBIT!&nbsp; And I have English first semester so I\'m screwed unless I get to readin\'.    &nbsp;  Well if I haven\'t told you (which I probably didn\'t) I\'m trying to start a band and my cousin may be in now.&nbsp; My brother rented a snowboarding game with a song called \"Courage\" by Alien Ant Farm&nbsp;and it\'s been stuck in my head ALL DAY.&nbsp;     &nbsp;  That is all.    &nbsp;   &nbsp;   '),(12097,3753470,'20,July,2004','2004-07-20 00:00:00','    Hey, as you can see I\'ve created a blog because it\'s all the rage and it matches my hair.&nbsp; Actually I think it was Kevin\'s relative Josh who had one first, then Kevin copied him and Erica copied Kevin and I\'m doing one because Erica suggested that I do so.&nbsp; I\'ll probably steal a digital camera (probably Gannon\'s) and get some crazy pictures up and maybe I\'ll post some creative literature that I make or something.&nbsp;&nbsp;    &nbsp;  Kevin went to Kansas this morning and I\'m not entirely sure when he comes&nbsp;back but I need to go terrorize his mom, cat and house whilst he\'s not present.&nbsp; I (like the procrastinator&nbsp;I\'ve proven to be) haven\'t started summer reading yet and I have two books to read.&nbsp;&nbsp;I was supposed to go by&nbsp;Kevin\'s and get \"To Sir With Love\" today and of course forgot to.&nbsp;&nbsp;    &nbsp;  Anyway...I don\'t really have anything to post of today so...comment&nbsp;on this&nbsp;holy creation of bloggism and&nbsp;go stomp on your neighbor\'s roof!&nbsp;   '),(12098,3753470,'14,August,2004','2004-08-14 00:00:00','    Well, Gannon and I went to the jamboree yesterday.  We screwed around for awhile and I just walked in without buying a ticket.  A tornado touched down somewhere and it was raining so it got cancelled.  Gannon came back to my house and we did random stuff and talked about Jacob half the night.  The next day (today) we walked up the road to some little old house.  We walked through alot of tall weeds and grasses and stuff to the front porch only to find that the door was locked.  We went around back where the ceiling was caved in and went through that way.  I was caught off guard by a picture of a woman on a wall that someone had drawn on...freaked me out.  We went up stairs and looked through lots of pictures and papers and there was lots of writing on the walls but both the digital camera and camcorder died so...little help there.  Hurricane Charley (haha!) is coming today/tonight....I\'M READY FOR HIM!  It\'ll probably end up as a little breeze but it\'s been raining on and off all day.  Well...that\'s about it.  Later.   '),(12099,3753470,'10,August,2004','2004-08-10 00:00:00','    Hey, actually not much has happened since my last post but I have to keep the flow goin\' in the bloggalicious lair.  Things have been dull outside of school...and in school for that matter.  I talk to all my people and go to classes and things go on as usual.  BUT!!!!  We found a bassist for \"GONE ASTRAY\"!  Some guy named Jake....ok, I don\'t know his last name, and I haven\'t talked to him yet...but he\'s been playing bass for five years and sounds good enough to be in.   Maybe I can do something this weekend to liven things up and make things more exciting for all the blog readers...since you have such scandalous fun reading of my fun.  I kinda suck...I have a poster thingy due in english on Friday and haven\'t started it...but it\'s a drawing thing, so that\'ll be pretty easy; I also haven\'t written anymore on my book!!  It nags at me everyday and I somehow come up with something else to do that is obviously superior in prioritative (that\'s a word now, if it wasn\'t a minute ago) rank in my little bubble of procrastination.   I have to go get ready for bed...after I write in my little computer journal thingy (yeah I keep a journal...cuz I\'m a fag) OOOH!  Aren\'t you so hurt and afraid that I have secret files on my personal life that you can\'t lay eyes on!   That\'s all...go spray paint yourselves blue and streak down the hallways whilst pretending you\'re on hallucinogens.   '),(12100,3753470,'07,August,2004','2004-08-07 00:00:00','    Well...week one of school is down.  Yippee!  only like...175 days left...then 360 more and i\'m done forever, always assuming I don\'t drop out...that\'s not extremely likely though.  Feelin\'  uninspired...think I\'ll start a fire.  I desperately need to write SOMETHING.  I don\'t know...song, poem, book, story, whatever.  I\'ll do that and THEN start that fire.     I got some tricking inspiration by watching Jujimufu in his latest video, and watching the aerial tutorial.  I still don\'t have the guts to try an aerial so I just did some cartwheels and seeing how I can\'t even do a one handed cartwheel I need practice.   If you didn\'t notice I really had and still have nothing to write about...I\'m in a kinda weird mood today.  I think I\'ll go out in the woods or down some trails or something...maybe draw some...write somethin.  Later.   '),(12101,3753470,'03,August,2004','2004-08-03 00:00:00','    I hate double posting on things...nothing big but if anyone didn\'t know I got a new computer like...umm, i dunno, a few months ago.  And we left our old one at ncol cuz it was screwed up.  Just now they figure out that they can fix it and it\'s gonna be fixed by next week, so I\'m gonna have my own computer then...which is cool cuz doing all my randon non-porn-related things on the computer is annoying when my parents and brother are right behind me.  This means there is gonna be some massive rearrangement in my room cuz my room is little and kinda cluttered.  Umm...I  think  that\'s all.  Go be cool expressive unique people and go smell some flowers and pet some kitties and hug some old fellas.   '),(12102,3753470,'03,August,2004','2004-08-03 00:00:00','    Well...back to school tomorrow...suckage.  But that\'s ok cuz i\'m gonna...wait, no...i\'m not gonna see my friends because I have no classes with them.  I have two classes with Matt and one with Emily but that\'s all that I know of.  I\'m gonna take my anger out by being a jackass in school and maybe vandalizing something just for the hell of it...hehe i love insurrection.   I went shopping today and got like 3 pairs of pants and 2 shirts.  Now I have something to wear!  Goody!  OOH, this catchy kinda upbeat music with a hint of techno (and a dash of paprika, hh) was on in Goodys (where I was shopping) and I was being stupid and strutting in tune with the music and doing hand motions and head bobbing, half way to embarrass my Mom.  I came to an intersection where the little tile path splits and there was some big advertisement there and I went around it, and this kid was going all out like it was a disco floor with his arms flying and feet moving.  And it wasn\'t like a little kid...he was a probably 10 or 11 redneck looking kid and he just stopped really abruptly when he saw that I saw him and I just started laughing to myself cuz I was just acting stupid and that kid was doing so much more.  Oh, and I heard \"Another Postcard\" by Barenaked Ladies while I was there and that\'s funny because it\'s a kinda funny song you wouldn\'t expect to hear there.  I finally finished To Sir With Love and I guess it was ok but it wasn\'t my kinda thing.  I read cliffnotes (spelling?) for Siddhartha and that seemed much more like a book I would have liked to read all the way through.     I found another artist that whoever\'s reading this needs to check out: Shinedown.  I\'ve only heard the song \"45\" but it\'s really good and it\'s been stuck in my head.  That\'s like the first good artist I\'ve heard of on MTV...well it  was  the rock countdown though, so I guess there are more decent artists on that.  That\'s about all for now...later.   '),(12103,3618866,'28,June,2004','2004-06-28 00:00:00','    hey guess what...thats right, i know why he says rah now...yup yup yup.  Well see his 2 year old son jack dosn\'t know how to talk yet and so when he\'s confused, he sits on the floor and yells \"rah\" alot. ^-^ aint that cute! Yeay now i know why he says rah....im gonna do that now...maybe  p.s. ears are weird   '),(12104,3618866,'22,June,2004','2004-06-22 00:00:00','    much blah.....too much blah.....so much blah that my head shall explode.......no implode tehehe.....and cows shall wear crowns of satilites....yes, yes indeed.  My cat is insane, yup.  You dont need to know why, you just need to know that she is.  The peach tree is big, indeed.  Many peaches all over.  Big mess.  But good pie.  I mad a pie today.  It was good.  My moms friends brother is a good artist...too good. Im mad now. I want to be that good....well im ok but he is freakishly good.  tisk tisk tisk.  im writing these sentinces weird, yup yup.  oh well, its fun. I finally have lots of books to read. Yeay!!! i like books.  Im reading a funny one right now.  Even Cowgirls Get the Blues.  Go read it now!!! I want to draw something but i dont now what to draw. GRRRRRRRRRRRRRRR!!!!!!!!!!!!!!!!!!!!!!! I have all these ideas floating in my head but i cant put anything together to draw, dag nabit!........actually i have something that im suposed to draw right now.....dogs and cats surfing....its for my aunts job thingumy.....im gonna get paid Yipee!!!! but i dont feel like drawing that right now.  I feel like drawing sci fi or fantasy or whatever ya want to call it that i draw......oh well. tralalalalalallalalalala oh and the title of this is kinda like the description of the cover of the book im reading right now p.s. I just put in the 4th chapter of my story go read it now!!!!!   '),(12105,3618866,'22,June,2004','2004-06-22 00:00:00','    well well well you little invisible platypus, i see your back........even though your invisible....hmmmm.......nevermind. So fathers day has passed and i cant stop eating the spice cake i made ^_^ my dad went out fishing with my mom and brought back some delicious felays (i cant spell that and neither can spell check), but uh somfen wierd happend when i woke up that morning....i looked out side and our little pond was almost empty....it was really wierd and when my parents came home they checked it and they couldn\'t find a leak or anything, the water just went bye bye.  I sugested our dog was reallllllllllly thirsty that night but no one bought it. hmmmmmmm i wonder what it could be................................... Dude!!! i HAD THE CRAZIEST DREAM LAST NIGHT ya know the show queer eye for the straight guy, yes i LOVE that show, well last night i had a dream that i was the straight guy they were doing over, it was sooooo cool, but my dream was doing that dream thing were its not really the reall people but your dream still says it is....yeah well so that was annoying cause i dont remember there ever being a carson in my dream....kinda sucked but whatever.....anyways they were all prettyfing me and i was like cool....then one of them went through my underwear droor and that was kinda embaressing.....but it was cool cause one of the guys i dont remember.....it was probably the interior desighner or the dude who does all \"how to keep clean\" dude, i wish i could remember thier names....well anyways one of them like told me there deepest darkest secret...he was bi(this comes to play later in my dream)so i was like cool and then i told him i wasn\'t completley straight either ya know *wink wink*...so anyways he later told me (while they were showing me how to corectly put on a hat)this horrible thing that happend with him and the other dude, like had gotten married and adopted kids once and one day like they were out walking in the park when one of the kids ran ahead of them and got lost, so they were looking for him and stuff and meenwhile like the dude who owned the park was like hey man better find your kid soon cause im going to blow up this park soon.......ya i know thats wierd but its a dream........so like they are searching and searching but they cant find them and so they are outside the park gates hoping the kid will run out of there before everything blows up when suddenly see the kid and there other kid that was with them is so happy tat she runs in to see her brother but right when they hug the park blows up and little kid body parts and guts are flying around...twas very sad....then the day he tells me this like its the 1 year anaversarie of this tragic moment(the two guys arn\'t married anymore by the way)and for some sick twisted reason they have a renactment of the day and i have to watch it....well its really sad and we cry and i hug the dude and then that dream thing happens were one thing turns into something else like it was never that other thing ya know....so now im a chick and that piece of info the dude told me(he\'s bi) kicks in and we are about to kiss and stuff when i wake up........very odd dream i must say ^_^   '),(12106,3618866,'18,June,2004','2004-06-18 00:00:00','    actually im going to put my reall name on that thingy   '),(12107,3618866,'12,June,2004','2004-06-12 00:00:00','    ya know what my profile only shows some of the stuff that i wrote so im gonna post it all right here because i dont like that okie dokie ^_^ FAVORITE BOOKS:The Sabriel series,False Memory, By the Light of the Moon, Darkfall, Phantoms, the Mask,Hitchhikers guide to the universe, bored of the rings,Eragorn, Harry Potter, Keys to the kingdom series, Artemis fowl series, water series, speak, cut, you dont know me, holes....and many more....does JTHM count cause i love that to. ^_^...and ya knowthe Lord of the Rings is one of me fav movies but the books are sooo boring. FAVORITE MUSIC: Marilyn Manson, Godsmack, linkin park, rob zombie, seether, coal chamber, nirvana, kittie, orgy, tatu, soundgarden,offspring,hoobastank,B-52\'s, lll Nino, dropkick murphys, tool, audioslave, queens of the stoneage, the darkness, jet, that one song flagpole sitta by harvey danger, beatles, and of course my beloved Wierd All!!! FAVORITE MOVIES:LOTR, POTC, Fear and Loathing in Las Vegas, secret window, Edward Scissor hands, Benny and Joone, cry baby, RHPS, Matrix, School of Rock, shallow hall, saving silverman, mr deeds, anger managment, fifty first dates,punch drunk love, Pulp fiction, Party Monster, Queen of the damned, Help, Hard days night, yellow sumbmarine, UHF, Montey Python movies,earth girls are easy,dude wheres my car, arsanic and old lace, princess bride,Van Helsing, heavy wieghts, meet the parents, lier lier, bruce almighty, trumen show, austin powers, wayns world, shrek, finding nemo(i love dori) little nemo(not realted to finding nemo in any way)monsters inc. and anything alfred hitchcock....i could go on and on forever   '),(12108,3618866,'12,June,2004','2004-06-12 00:00:00','    lalalalalala....*blink blink blink* well hello there invisible elephant...you don\'t want to be an elephant....oh well how about a platypus....well it\'s either that or an elephant....ok.....well hello there invisable platypus....now what....but you have to be invisable....because....because is a very good answer.......because....tehehe....ok so your an invisable platypus....hello....anyways back to what i was origanlly going to say...i just got back from camping at hume lake yeeeehoooo twas pretty it was it was indeed, BUT....apparently are site was supposed to be right next to the lake in  a nice shady spot but no it was way up on this steep hill in the sun, and so we had to walk up and down that stupid hill like 5 times a day gggrrr i don\'t think iv ever sweat so much, THEN there\'s this evil camp host lady right and my family had tak\'n a walk around the lake and i stayed in the camper, well while i was  there apparently the lady came by an saw that we had a tiny tiny little fire in are fire pit and so she came over and totally drowned it, then she saw we had a little cooler by the table(it had like one soda or someth\'n in it)well she wound up giving us a fine for the fire and cooler that was like 5 hunred dollars!!!! while i was there but she says she said hello but whatever, well it turned out to only be a warning but still they are soooooo strict like we have to be right by the fire and the cooler has to ALWAYS be in the camper...iyiyi....well things were going good after that i read two glorious deen koontz books HE RULES!!!! i read phantoms and the mask in like 4 days ^_^ very proud of myself, we also had a very odd game of scrabble...my dad is wierd ok...so in the end we had a very odd and obscene little chant that my dad and sister couldn\'t stop saying^_^what fun,THEN as you may nt know hume lake is mostly this huge christian camp right, well when my family decided to go play frisby golf I HAD TO WALK THROUGH IT, this place was filled to the screaming brim with millions of rich little teenage sporty sluty cheerleaders and posers....my worst nightmare, it was horrible, now not because they were christians mind you, im polytheistic but i still believe in god and whatnot, but it made everything sooooo......fakly happy and all these little brats running around in there bathing suits screaming, of course thers were some good people, some of my friends were probobly there but ew it was soooo horrible and to top that off, everything was emensly expensive......it would have been alot better if i had a bunch of my friends there because soomehow that makes everything ok.....well anyways it was still fun even after being chased by fish guts...alot of other stuff happend but i cant remember and im tired of typing ^_^ (my cat is very cute by the way....and she\'s eating my hand right now...wow, here name is mogget so now you must read the sabriel series and you must never name your cat aristophones, now y7ou must read the mask ^_^) (for even more parenthesis, i love this song-45 by shindown)there i think that covers everything for a while......yep   '),(12109,3618866,'12,June,2004','2004-06-12 00:00:00','    well looky here im back and thats was way less then 10 minuites to....cool...oh yeah the reason why i came back....what was it...oh yeah ok like for all you invisible elephants who are reading this, my name not yklail ooo ahhhh indeed it\'s actually kelly gosselin ooo ahhh, but im not gonna change it on me profile, so ya know incase theres a special question ya got to answer on the radio or someth\'n (lol yeah right)ya\'ll know, and all those other peeps wont cause there to stupid to look at this \"secret\" post ooo ahh ^_^ indeed  oh um just incase ya really are a midget, like no offense or anything to ya okie dokie, cause like ya peeps rock!   '),(12110,3618866,'12,June,2004','2004-06-12 00:00:00','    look look look look look....i made a blog ^_^arn\'t you proud of me, let me guess yar not, especially since you probably arn\'t reading this, yeay!!!i feel so loved.  well just for the sake of typing because im bored(actually i should be doing a lot of chores and getting packed for camping but whatever)im going to say why this is called \"the vacuum ate my fish\", well here it is...there is no meaning whatsoever, isn\'t that wonderfull...well it might have a meening at the end of this post and i\'ll say why at the very end but as of ths second it is there because i want it to, so meh ^_^.....meh is good word ya know that, yes very good indeed, it\'s so nice im going to say it 5 more times, meh meh meh meh meh...well that was refreshing indeed. hmmmmm i want to say more but what to say.....well....um......niorpheus, yeah niorpheus...thats a wonderfull word....yeah *sigh* wow im bored.....ya know whats stupid, blogs, yeah they are ya know, i mean basicly they are diarys but we are to cheep to by one and have people make fun of us because we have a \"diary\", so we make one on the internet and think we are all cool...but we arn\'t, were just lonely geeks...yep*sigh* pitiful us, i mean how many people actually read these things(if ya want people to), im pretty sure no teacher would be like \"so billy what book are you reading\", \"oh im reading sexykittie69\'s blog\", \"oh well that\'s nice\"....sexykittie69...a lot of desperste sluts like to have names like that don\'t they, why, do they think that the guys will be all over them just because they think that person looks like a sexy cat who will do a 69?...there probably these fat wanna be cheerleaders to, god i hate them....*sniffle* such poor little idiots...i would pitty them but i hate them too much.....i want to get a reall journal thingy, because i have a really bad memory but im always thinking and hearing these great things...mostly at school, so like if i had a \"diary\", i could keep it in my backpack(his names william by the way)and right down whatever happened right then.  with this i have to wate till i get home and by then i usually forget alot, and im going camping a lot this summer so ya know i wont have internet acces there and it be nice to right down stuff in a little booky thing...oh well.....hmm well i think that\'s enough for right now.....yeah.......ill probably get right back on in 10 miniutes and add something though cause im like that and stuff....so see ya in probably 10 minuites ^_^   '),(12111,3618866,'04,July,2004','2004-07-04 00:00:00','    well what da ya know, it\'s fourth of july today.  I have to make even more peach pies gggrrr and i got to get started on that painting for my aunt *sigh* i should be working but im to bored to work.....lol yeah ok.  There\'s to many good movies out all of a sudden, and i need to see all of them...poo im gonna miss some of them i know it.  Hey the other day coming back from seeing harry potter i found a purple dinosour, yep its cool.......im making a website....but i dont know what to put on it, wich can cause a problem indeed, i should probably just stick with this blog, but i wanted to make one and now i dont know what to do....*sigh* im open for suggstions....(i just relized thats a line from secret window)   '),(12112,3618866,'01,July,2004','2004-07-01 00:00:00','    :( and i hardelly have any of his movies on my profile ): I CANT WAIT TILL WILLY WONKA!!!!!!!!!!!!!!!!!!!!!!!!!!!!     wow this is like the only post with a title that matches   '),(12113,3618866,'01,July,2004','2004-07-01 00:00:00','       '),(12114,3618866,'01,July,2004','2004-07-01 00:00:00','    WHOA!!!! its july?...awesome :), well i forgot to tell ya somthen, yesterday i painted more of my room...yep...but not like normal...if ya can call the other stuff normal...see, the other day my friend was over, and she tended to say \"im confused\"...A lot, but mostly when she was in my room...so i decided to help stop that confusion process by painting some obvious words to point out what things were in me room...so now upon my walls i have painted the words: shelves and pointing to shelves, wall and on a wall, cieling and pointing to ceiling, closet and on closet, clock and pointing to clock, door and pointing to door, window and pointing to window, stars and pointing to glow in the dark stars,floor and pointing to floor, and painting and pointing to the murals on my walls. I didn\'t do bed because the wall space that i would have to paint it on would paint on the murals....so yeah, tis cool   '),(12115,3618866,'01,July,2004','2004-07-01 00:00:00','    howdy doody my platypus friend. wow we just had a lot of company in those four days, and i made a peach pie for almost every one ^_^ delish...before one family came me and me sis were bored and shooting some soft air guns in the backyard, well we just happened to be shooting at a wind chime in the peach tree.......so anyways the next day gabe and becky and there kids came over and gabe went to the backyard and picked himself a peach and bit right into it and onto a beebee thing, yes ouch, but so hilarious ^_^ i knew that was gonna happen.....another thing i knew was gonna happen sooner or later was well we got a drum set in me parents room and as ya know the base has a whole in it like they suposed to and theres a little blanket thing in there to for the sound and whatnot.....well we also have a very adventurous new kitty...yes i know what ya thinkin....well i knew it was gonna happen sooner or later...i walked in there look\'n for mogget(da kitty) and i looked down and there she was in the drum....^_^ she was soooooooooooooooooooooooooo cute, all curled up in the blanket and stuff, but alas i knew my dad would not be happy about that so i went to the back and slamed the little foot peddal on the drum and she shot out of that drum like lightning, tehehe i love her. hm, i got a whole bunch of new books and im read\'n one in a dition to cowgirls.  tis really good, i really didn\'t expect it to be writen the way it is(exile, its called by the way)so far so good....but its HUGE!!! like 800 somythen pages...yes i know thats really not that bad to some of ya but ya know im used to reading 300 to 400 kinda books....besides harry potter lol.......well lets see...i can\'t remember anything else at the moment so toodles and may your pants taste of jello ^-^   '),(12116,3618866,'06,August,2004','2004-08-06 00:00:00','    Yeay...................................duct tape is like the force, it has a dark side and a light side and it holds the universe together ^-^   '),(12117,3380464,'30,May,2004','2004-05-30 00:00:00','   (An excerpt from Kelley\'s journal dated April, 1997. On temporary assignment to the U.S. Consulate in Munich, Germany, Kelley had just caught \"The Tonight Show Starring Jay Leno\" on MSNBC for the first time)  I donâ€™t think I had ever seen â€œThe Tonight Showâ€ with Jay Leno before.  My days of staying up until midnight or later watching television are far behind me now, and in fact between my college days in the \'70s and his retirement from \"Tonight\" in 1992, I seldom watched Johnny Carson.  Even when I was a reporter on radio station KUIC in California in the mid-\'80s, and was coming in from work quite late many nights, I was more likely to tune in the old reruns on Sacramento\'s Channel 31 than  â€œTonight.â€  	 Under Leno, the show has been completely revamped: it resembles the David Letterman show now more than it does the old â€œTonightâ€ show.  Gone are the blue curtain and the spotlight, as well as the NBC Orchestra and the old â€œDA-da-da-da, daâ€ theme music.  The whole thing is very \'90s, with angular shots, the audience right up front, and new, minority faces in the band which have replaced those of Doc Severinsen and Ed Shaughnessy.     Leno has preserved the showâ€™s best tradition, the opening monologue, and his writers are as good as Carsonâ€™s were, although the humor, like the rest of the show, has become appropriately â€˜90s: Carson might have mentioned the â€œHeavenâ€™s Gateâ€ cult, but he would have had the good taste not to wring it for black humor as relentlessly as Leno is doing. Well, thatâ€™s the â€˜90s for youâ€”America has become a nation of namby-pambies, busybodies and crybabies who recoil with righteous horror at tobacco, alcohol, fatty foods, vivisection, fur-wearing and political incorrectness, and who run crying to Mr. Lawyer when their iddle-widdle feelings are hurt, but who nevertheless laugh, snicker and guffaw at mass suicide.     I didnâ€™t stay with the show past the opening monologue; even in the old days I would usually turn the set off and go to bed after the monologue unless there were a guest on the show I particularly wanted to see.  In any case Iâ€™d seen enough.  The â€œTonight Showâ€ is still basically what itâ€™s always been, but revamped of course for contemporary manners and mores.     For me, though, it served as just one more reminder of how this foreign service life, which Iâ€™ve now been living for more than a decade, has cut me off from the American â€œmainstream.â€    I havenâ€™t watched American television with any regularity since the early 1980s, and in fact my habit for the past few years of confining myself in the evenings to videotapes of \"Nick at Nite\" and my â€œgenericâ€ tapes of favorite reruns has really placed me behind the times.  \"Seinfeldâ€ is the current measure of whatâ€™s funny, but Iâ€™m still watching â€œThe Dick Van Dyke Showâ€ and â€œGreen Acres,â€ the shows I watched as a kid.   I did watch â€œSeinfeldâ€ once, and found it about as funny as a root canal, this program of which Jay McInerny wrote not long ago in TV Guide that it may be the funniest show of all time (funnier than â€œI Love Lucy?â€â€”I donâ€™t know.)  But thatâ€™s the way things are drifting, and they have left me behind: Iâ€™m an old fuddy-duddy at 41.  Last year Hal Hinson, writing in the Washington Post, gave the Steve Martin movie â€œSgt. Bilkoâ€ a middling review, in the course of which he slammed the old Phil Silvers TV show with a sidealong cheap shot.  Imagine: the classic â€œSgt. Bilko,â€ one of the funniest programs ever aired on CBS, which had Neil Simon among its writers and featured the flawless comedy acting of Silvers, Paul Ford and Joe E. Ross, getting a condescending sneer from some Perrier-sipping, overpaid Beltway ignoramus whoâ€™s trying to appear cutting-edge hip. I wrote the paper a letter in which I called Hinson a \"moron,\" and then, this being the 1990s, in which even hard-boiled critics are entitled to get their widdle feelings hurt and run crying to Mommy, I got a letter in reply from some gerbil of a sub-editor, tut-tutting at me for calling poor little Hal Hinson mean names. Screw Hinson and his nitwit editor both.     Not that my tastes are entirely mired in the \'50â€™s and \'60â€™s: during my recent two-year sojourn in Washington I discovered--and loved--a couple of more recent sitcoms, â€œThe Golden Girlsâ€ and â€œMr. Belvedere.â€  But again, both of these programs were in production roughly between 1985 and 1992, and they have left the air now.  They may be more recent than â€œThe Beverly Hillbillies,â€ but they are not \"contemporary\" in the sense of still being in production.    Of course television is only part of it.  Whenever I leave the United States for any extended period of time, when I return I find that something has changed in American popular culture while I was gone, and I get a little culture shock.    The first one that comes to mind is my discovery of the â€œinfomercial.â€    When I returned to America in 1988 after my first two years here in Germany, I had never seen an infomercial, didnâ€™t know what one was.  These 30-minute commercials disguised as interview programs had been invented while I was over here.  So, when I found myself in a motel room in Warrenton, Virginia watching what seemed to be a straightforward interview with two guys who had successfully used some hair-restoring product, I ordered a bottle of it by phone.  It smelled nice, as I recall, but I remained bald.    A few years later I was home for Christmas and discovered that big grocery store chains like Vonâ€™s and Safeway had begun staying open 24 hours a day.  That was a big surprise.     During last yearâ€™s presidential election campaign, I kept seeing the term â€œSoccer Momâ€ in the news, but I had no idea what it meant, and Iâ€™ve never actually seen anybody dance the MacareÃ±a.      Armed Forces Network was never a panacaea for this, and even now that we have the Internet, which we didnâ€™t have ten years ago, I donâ€™t think itâ€™s going to keep me â€œin touchâ€ either, because the fees we get charged outside the U.S., both for on-line service and just for using the local telephone line, are prohibitive. It costs so much to log on here that I, for one, donâ€™t use America Online for anything except retrieving e-mail, a quick look at the news headlines, or a very quick trip to the Tower Records website to order a compact disc or two.	   No, it would seem that a foreign service employee is condemned to be, for the tenure of his employment with the government anyway, something of an exile.  I donâ€™t necessarily have a problem with this.  Iâ€™m not all that crazy about stateside life to begin with (I suspect that my Russian-born friend Lena Watt, born and bred in Novosibirsk and currently living in Moscow with her American husband and daughter, probably misses the United States more than I do) and anyhow, the role of â€œexileâ€ appeals somehow to the â€œboring romanticâ€ in me, to quote a 1985 song by John Cougar Mellenkamp.  Surely â€œoutsiderâ€ is a role I have never eschewed, and I donâ€™t think missing out on the latest generation of trashy television shows and garbagey popular music necessarily makes one culturally deprived.  (I take some pride these days in the fact that when the Grammy awards are passed out, I seldom recognize any of the winners, and as often as not that turns out to be the case with the Oscars as well.)     Perhaps itâ€™s not just being out-of-touch with contemporary pop culture that bothers me so much as the fact that having fallen behind the times, regarding television anyway, (Iâ€™m not going to worry about musicâ€”I defected to the timeless Mozart and Stravinsky when I was 16) signifies that I have fallen victim to the old-fogeyism for which our parentsâ€™ generation was constantly getting panned when I was young.      During the Vietnam era it was especially bad.   The media created the famous â€œGeneration Gapâ€ of the 1960s in response to tensions between younger people who opposed the Vietnam war, (because they were being drafted to go fight it), and older people who were unable to see that this wasn\'t the same kind of war that World War II was, tended to support Vietnam as a similarly righteous crusade, and consequently to regard the twentysomethings who opposed the war as both unpatriotic and ungrateful.  Then, having created the â€œGeneration Gap,â€ the media proceeded to give it so much attention that everyone began to take it for granted.  If you were 25 or under you were the object of media adulation and assumed to be the quintessence of hip: up on the latest styles, the latest music, the latest slang. If you were over 30, on the other hand, you were caricatured as hopelessly out of touch, sitting in your rocking chair thinking that Tommy Dorsey and â€œYour Hit Paradeâ€ were still the cutting edge. (Correct me if I\'m wrong, but I believe TIME magazine\'s \"Man of the Year\" for 1966 was \"Anyone Under 25.\") And what the news media wrought, Madison Avenue was not slow to pick up: youth was used to hawk everything from coffee to toothpaste in the 1960s. I do in fact recall a TV commercial in which a \"sensible\" daughter steers her youth-struck mother from a supposedly-frivolous brand of toothpaste back to the sponsor\'s more feet-on-the-ground product.       These stereotypes got a lot of attention of course, not-so-subliminally drilling home the message that it was good to be young and hip and bad to be old and behind the times.  Eventually, as such things do, this us-versus-them mentality filtered all the way down into the elementary schools: amongst my friends there were words you were supposed to use, clothes you were supposed to wear and radio stations you were supposed to listen to. In San Diego, circa 1967, if you deviated from top-40 KCBQ and KGB radio, you were in for it. Once, when a school acquaintance found out that Iâ€™d been listening to KOGO, the NBC radio affiliate in San Diego which played Nancy Wilson and Frank Sinatra rather than Strawberry Alarm Clock and Small Faces, he began asking me derisively where my electric car was--electric cars were something only senior citizens used in those days.    So maybe thatâ€™s whatâ€™s really bothering me when I find myself not laughing at the â€œnewâ€ Tonight Show, finding Jerry Seinfeld\'s wildly-popular program about as funny as dental surgery and having to admit that I would rather watch â€œThe Andy Griffith Showâ€ than â€œRoseanne.â€     Sure, I\'m not alone in  these preferences. Nick At Nite owes its very existence to Baby-Boomer nostalgia for the shows that those of us born during the Eisenhower years grew up with.  But my fellow Boomers back in the States can have both sides of itâ€”-they can switch over to Nick at Nite to watch â€œI Love Lucyâ€ after â€œSeinfeldâ€ is over.  Iâ€™m 5,000 miles from all that, and havenâ€™t even bothered getting cable TV here in Germanyâ€”I didnâ€™t want to pay AEA for it, so I decided to do without it.  And so Iâ€™m watching videotapes that Iâ€™ve had for six, seven, eight years and more, tapes of shows that were old when I made them: â€œThe Addams Family,â€  â€œMake Room for Daddy,â€ â€œCar 54.â€  Iâ€™ve seen them 10,000 times each.   Thatâ€™s called being in a rut, and when I was 12 it was what we expected of people in their 40s. But then, being in our own 40s was something we never imagined.  '),(12118,3380464,'29,May,2004','2004-05-29 00:00:00','   Verrezano Narrows        (for Charles Berigan)  Borough-wide tonight,  you can smell the preparations-- around corners, in cul-de-sacs,  It\'s everywhere: cooking. Over in Manhattan,	 everything\'s \"good to go,\" R.H. Macy\'s cocked and ready for the big moment; campers on the sidewalk, cops cruising, yawning,  anticipating midnight, Rockefeller Center  emblazoned with bells. Gunning the engine after one more annoying toll: almost there now, just this last monstrosity to climb, spidery night-cathedral, and the long drive\'s over.  I\'m not usually afraid of bridges-- There was that one, crossing the Strait of Carquinez, train trestle alongside, and as I put my head out, sampling the summer wind, I suddenly noticed we were  hundreds of feet in the air, and a little too close to the edge.   The toll booth was a lighthouse-- Here again, that giddiness: this isn\'t 14th Street in Washington, D.C. where your biggest worry is low-flying planes, this is a bridge, climbing   to where you understand  why everybody\'s speeding,  and catch yourself wondering, \"What year was this thing built?\" A queasy space beneath  my nervously-rushing two-door, The Narrows are a perfect reflection of starless night, black--could be a mile deep.  63rd Street\'s a brownstone corridor, cold wind scraping leaves along the lined-up  trash cans in the November night, Brooklyn cops shaking down suspects against a  grocery store, beneath the glare and the graffiti, then the climb up the stairs, the dog\'s insistent barking, The sudden caress of warmth and the light from the dusty lamp between the two pianos. Stacked up on the bench of one, American Record Guides collect dust right along with newspapers, CDs,  an old sock the dog was chewing, Rachmaninoff\'s biography in a yellowed edition. At this piano I took  your picture one night, years back, across the river, playing muted Chopin so as not to disturb the neighbors.  My first time in the city: coming in from LaGuardia, the cabbie overheard us discussing Leonard Bernstein, and said \"See you gentlemen at the Philharmonic,\" (though he never did). We got out in the west seventies, and moments later walked past the Dakota--you pointed out  the very spot where John Lennon bled to death. Moments when music defines memory stick up over that jagged, well-lit landscape, interior skyline, shifting,  fluid like burning night in the city always tossing in its celebrated wakefulness. Picking up the yellowed bio, turning a page, I frown: the curse of high-acid paper. It was either that night or the next, we hiked along Riverside Drive, (it was late) to take a look at the building Rachmaninoff owned, with Gershwin living in the penthouse for a while: Imagine: Rachmaninoff as Gershwin\'s landlord; that New York is gone for good.  Years had already blunted the edge of our respective capacities for joy, but looking up where you pointed, as the night wind off the Hudson blew scraps of newspaper around our feet, I could remember a time when it would have been a shrine. The irony always amused me: We met playing football in high-school p.e. class, you fourteen, looking twelve, skinny, pale, long-fingered kid in a white sweatshirt, the slightly-misspelled name of a dissident priest printed across your chest in  black laundry marker, squared off against me, slightly older and larger, truculently unathletic, arrogantly uninterested, determined at all costs  not to be the athlete my father dreamed of. Opposing linemen. Hike--the ball was snapped. I gave you a passing shove just for appearance\'s sake, then walked away, superiorly, whistling the first eight notes of Beethoven\'s \"Eroica.\" You picked up the tune, finished the phrase and made a joke about \"token resistance...\" More than twenty years ago.  Tomorrow will find us Cruising the liquor store  browsing for some wine cheap enough not to strain  our wallets, but pricey enough not to embarrass us later when we hand them to our hostess. Push the books off the bench, and take a seat... To name is not necessarily to command; with ease, you could make sense of those hieroglyphics  like I never managed to, although for years my mother made pin money, playing her own gigs, \"Music lady,\"  first for the Methodists, and then later, twice a week, (Wednesday nights and Sundays) over at the Christian Scientists\'. Some things just aren\'t in the genes. In high school, the choir director (surely you remember him: thin, bald, into Jesus; fed us religious repertoire he\'d never get away with now, fired in disgrace  many years later for  something to do with one of his girl students)	 pointing out the parts: treble,  bass, semiquaver, leger line, F-A-C-E and  everygoodboydeservesfun... though I loved to listen, it was all lost on me: I had to carry my part by ear, following the lead of off-key Keith to my left, or like so many of the rest, remember how it sounded on the radio, only to get it wrong: \"this arrangement\'s different.\" No such problems for  you: if anything it came too easily. (My sister, sometime alto, still laughs about the night, during the \"Hallelujah Chorus,\" when you somehow forgot the rest of us were there: taking the bit, you ran off, oblivious to the director\'s desperate sotto voce: \"CHARLIE!\"--you proceeded with your solo as we watched the top of his head turn red, playing faster and faster until he didn\'t dare hold the pause before that final \"Hallelujah!\" more than one single beat-- you had us all rolling like the Wehrmacht into Poland; someone would have \"broken.\" \"Well,\" when it was over, he said, mopping his head as the veins went down, \"That\'s the first time I\'ve ever heard the \'Hallelujah Chorus\' played as the \'Minute Waltz.\')  Morning. You scratch your head as the dog fusses in circles to be taken outside. \"We missed Macy\'s parade.\" \"Big deal.\"  A stretch and yawn and then the two of you go off for fifteen minutes. With water on to boil, I look over the piles and piles of  compact discs-- by the time you walk back in, we have Kiri singing Kern. A few hours to kill. The things that I\'ll attempt to reconstruct the next bridge over, are still back there, the war that Hemingway deliberately left out of Big Two Hearted River, the watershed that led me up to your door last night; the long weekend alone would have been ill-advised; Dylan himself was surely on an even-enough keel when, declining an invitation one Thanksgiving, he stayed back in his room and wrote Just Like a Woman instead. It\'s a state much to be envied, and elusive at bad times. No, I think, regarding the orange peels on the paper atop the counter among the dust, it\'s best to be here.  The dog shoves her face deep into her food bowl as Kiri gives the bloodshot morning an unacknowledged caress.  Outside, it doesn\'t look much like rain or snow. And  there he is once again: the face on the piano bench: \"A six-and-a-half foot scowl,\" Stravinsky called him once; \"Magnificent Rachmaninoff\" he was to Chaliapin, or so he says in the memoir I was reading that long night aboard the train from Moscow to St. Petersburg--that everything, it appears, is gone for good. Walking along beside the river, water too cold to contemplate chopped by the breeze we\'re bucking, the afternoon sunlight ineffectual like it almost always is this time of year,  we zip up our jackets. I rattle off a poem just completed last week, marking an anniversary: it gets a \"hmm.\"  Yes, it\'s been that kind of year. Far off  up the water, the Statue of Liberty ignores us; did immigrants slouching toward Ellis Island  get the same impression? Here, on Brooklyn\'s most fashionable streets, you almost have to wonder. \"Lotta mafia live here.\" \"Nice houses--\" that cartoon  I saw not long ago in The New Yorker: \"First we take you out for ice cream, then  we throw you off The Verrazano-Narrows,\" yuk yuk yuk--tough town: Memories are equally vivid of the first stroll we took past Carnegie Hall, and down that Saturday-night street out of Martin Scorsese, Ominous dark cars lined up alongside the curb before that infamous restaurant, patrol cars oozing in  and out in slo-mo.  We drive to Coney Island, Under the El, re-tracing  the car chase scene in The French Connection, Vivaldi on the radio, the ocean brilliant, wind-whipped November boardwalk, talk thickening all around us from English to Russian as we hike into Brighton Beach, a quick step back eight time-zones to familiar sights: the sign facing the street, reading \"PECTOPAH,\" the street vendors hawking books: the needlepoint of cyrillic print on wooden  tables. Then all the way back it\'s classic jazz: WNYC:  no limit to the music here. Thanksgiving dinner now: our well-meaning host plays one Tchaikovsky symphony after another, (thatâ€™s  for me--you told him Iâ€™d been in Moscow) but then I request a switch to Mozart; too many familiars come along with Peter Ilyich anymore, not the least of which the one in charge of early memory : I  was just 13, and every attempt of our teachers to get us to listen was as lost on me as on everyone else: music meant top 40; a field trip to the opera was an excuse to throw  paper planes from the mezzanine, music appreciation note-passing time-- (naps for the nascent \"heads\"--) but that morning was different. Mrs. Easton, lagging on her paperwork, no doubt, slapped Overture 1812 on the  phonograph, and told us all to shut up and keep quiet. Nobody did, of course, but the music drowned us out so she could concentrate. The note-passing must have by-passed me that day-- there was nothing left to do but listen.  So  I did. Then my sister had a birthday: among her presents, Serkin Plays Beethoven intrigued me by its cover. The next thing I knew, the \"Pathetique\" Sonata had me by the chin, that second-movement song the perfect accompaniment to the December snow outside the window as I stood, my back to the stove, pouring a cup of coffee, wondering if school would be closed. Christmas of \'68---- space was big that year: first there was 2001  and then Apollo 8. I asked for the soundtrack, and the Tchaikovsky both; surprise--there they were in their places beneath  the tree. As that big blue earthrise flickered across 50 million TV screens, and the crew  of the Pueblo walked across the bridge back to freedom, Tchaikovsky, Strauss, Ligeti and Khatchaturian filled up the basement where my bedroom was, the ground-level windows packed solid white--it turned out to be a severe winter. And to this day, I hear Gayneâ€™s Adagio  and I think \"snow\" seconds before I remember to think \"space.\"  The last of dinner\'s been cleared away.  We audiophiles,  acceding to small slices of pie and tepid cups of afterthought coffee,  repair to compare, taking up positions in front of Hyperion\'s awe-inspiring $5,000 stereo, popping old recordings of Schumann and Mahler, accompaniment to his preparations for work, (\"You gotta work Thanksgiving?\" \"I don\'t mind. I might make two hundred bucks tonight\") strapping his cummerbund around his waist, straightening a bowtie sharp as St. Andrew\'s cross, ironing black pants. But the food and wine have done their work: Bruno Walter\'s phrasing in the fourth-movement chorus of  the Mahler Third Symphony, and then Leonard Bernstein\'s is just sweet sound; I nearly doze, then Hyperion suggests we head downtown with him. He works the Rainbow Room tinkling out tunes for the furry  and the bejeweled--\"Great view from the 69th floor.\" Nothing else on our agenda; What the hell?  Let\'s go. Rockefeller Center emblazoned with bells: that New York  will never die.  Pressed into an elevator with better-dressed people, discussing other, unseen better-dressed people,  our stomachs briefly remind us of that big dinner as we\'re whisked Rapidly up to the softly- lit, carpet-deadened world of self- congratulation three or four hundred feet above the sidewalk, where the wind still blows, and 47th Street suffers the slow indignity of  creeping \"mall-ization.\" The windâ€™s harder up here, colder, too: vertigo. The top of the Empire State, 16 blocks in front of our faces, is a womb compared to this: no glassed-in deck here, just the roof, the edge, the drop--facing that panorama, too much even for Whitman to take whole; my palms start sweating: the windâ€™s awaiting its chance to sweep us overboard, and 69 floors down. This is for the birds-- Let\'s get back to the hubbub.  A quick swing past the White Horse, where that earlier Dylan met his nemesis, and haunts the walls above the rocking din Still, then it\'s back out across the river, home to conjure one more time the spirit of \"that which unifies:\" The lights, the dust, cold beer, late night, past disasters, the present and what might lie beyond morning that lurks beneath the river, and then the sound, that benediction, Rachmaninoff\'s Third Concerto, the Vespers of 1909...         Kuznetski Most       (for Nadya)  Hereâ€™s an idea--the Titanic went down with the band still playing: just how ephemeral is sound?  Waterâ€™s just air, a likewise medium-- echoes like the moon, visible only because of that pale stolen fire The Man  talked about, echo--how long?--a question recently asked in music. Imagination on an errand: I dreamed once, below the equator, of an afternoon in California spring, a dream so real I swear I felt the coastal breeze on my face, saw the dial on the radio in the kitchen. Here again, the idea of the band playing on forever-- numbered snowflakes blowing up and down the street at dusk, around the corner the Bolshoi, lit and filling up for  a silent concert that may or may not ever end, voices crying â€œS novâ€™im godom!â€ over the heads of the everlasting booksellers like ghosts on the other side of a winter night. Itâ€™s still there, fading but there, retrievable in theory, no--in fact, as long as  blood proceeds in its humming orbit of the phenomenal. Somewhere along this street youâ€™re standing underneath a column or a lamppost-- (what joy in discovering which), your face no less than music borne on the wind of this snow that freezes nothing  â€œreal,â€ music vibrating in water, light from stars that may or may not be there anymore. Tolstoy crossed this bridge once; what chance of picking out the reverberations of one diminishing voice from all  the others?  Theyâ€™re here, though, bet on it, and the bricks of Red Square just as much there as when we stood in line, Sunday morning at Leninâ€™s tomb, hustled, hushed through and emerged back into the sunlight to make the rounds of those recently-dishonored dead. This kind of excursion has one advantage over that: instantaneous movement.  Now we drive along the boulevard on a summer night, passing St. Basilâ€™s (lit by the late sun), now Iâ€™m watching out the window, four stories up, the smoke from those eternal factories disgracing the sunset as I think the words, â€œKrim, Klin...â€ You gave them to me to memorize; we ran out of time, time thatâ€™s supposed to be illusion, but plays on the surface like a loud, intrusive voice, like pounding on the wall. Voices, music, chimerical light-- Proust had it right, but how elusive is that tea-soaked  eternity he stumbled across; a one-time shot from forever like the comet slamming  into Jupiter, or the first morning recognized as such. Who would have thought that summoning one snowflake would be the hardest thing of all? Your face? Forget it.  A photograph is tacked to the wall, defiant of seasons, and yet, I hear youâ€™ve cut your hair, changed this and that...Footsteps clicking on cobblestones are scraped or caressed in turn by those afterthoughts of October, leaves that burn in the remembering, their whisper just another  source of  the language that pours unceasingly from above the frozen flames torn from your soul: the gleam in your eyes as they found me in the brightly-lit lobby of the conservatory swims through that water as well. Moscow!  Thereâ€™s nothing like it.  In New York City, In West L.A., Iâ€™ve monitored rhythms peculiar to both, sniffed the ocean, tapped  into the bloodstream of never- sleeping night, sampled perhaps, in some small way, the inheritance, America of Whitman, of Kerouac, a birthright...but there is no cold shower like this: dushaâ€™ means soul in Russian-- Spenglerâ€™s down and out-- do it: go in: dive deep. Rumors say thereâ€™s music down there anyway. I turn the corner, and under a lamppost, youâ€™re standing in the snow. Iâ€™ve seen this before: yes, the stage-setting for Pique Dame, Act III, Scene II, around another corner. But this is no tragedy, you wonâ€™t be flinging yourself in a canal. The waters of this unhappy river have flowed on past Any Act III, any scene II, things bigger than us attending. Screams and bullhorns and flag- waving at the parliament, shouting and shooting, rhetoric in the rubble, some kind of  phoenix trying to get itself reborn against all odds-- we hurried back to the embassy that warm afternoon, the air suddenly stopped like the moment before the clouds tear open on the equator; the bullhorns had been replaced by sporadic gunfire; before it was over, the casualties outnumbered 1917, (there anyway) tanks rolled, bullets flew, the crowd watched from the bridge, eating ice cream.  The dust hadnâ€™t settled from that when your eyes appeared in that crowded lobby in bright November, locked on me, drew me into your own â€œhumming orbit  of the phenomenal;â€ that light, that moment:  a pebble in still water, concentric circles that have left the solar system,  spreading out beyond wavelength, into whatever is meant by â€œeternity.â€ What is?  To be outside of time must be a bore, or so I considered once, writing down notes to a shade who (I didnâ€™t expect otherwise) never answered back. Then again, wasnâ€™t it Faust whose undoing was a request For time to stay its hand? We stood under another streetlight, another night, a poet who had died for love Or politics, take your pick, watching over our heads, and I read to you out loud about a sleeping river. Where itâ€™s forever 2:15 on a bright afternoon in May, a light breeze blowing from the ocean, the sun positioned just so, thatâ€™s where that moment resides--can it be had again for the asking?  As I walk along Pennsylvania Avenue on a Sunday morning, the traces of Saturday night lingering with the sleepers on the steam grates, and the papers crumpled, outdated on the sidewalk, reminders buzz around, gnats by the reflecting pool. Each one a quantum, each potentially broken open, might flower like some lotus sutra condemned to exist through memory, a terrible angel of its own. It drifts back to Zwemmerâ€™s, the English-language bookstore where on a winter day I surfed the stacks, searching for that book you said you couldnâ€™t live without--\"Geroâ€™i Nashevo Vâ€™remeniâ€--a masterâ€™s sleight-of-hand: time shuffled like a deck of cards in some godâ€™s fingers, hiding-in-plain-sight magic, magic that you told me strummed somehow on your heart. I took it home, rode over those Caucasian landscapes  myself as February paled  its way on into March; spring thunderstorms were still only an implied threat.  In your eyes, memory stayed safe in the womb of that Russian night where you sprang out of the light, to take your place along those ever- lasting streets of fire, the streets where you drift, spirit-mist of remembrance over things everyday and â€œimmortal:â€ your chimerical lover who haunted that garden under Tamaraâ€™s  window far to the south, whose imagined image you showed me (I was jealous, but then humbled to learn who he was--who among us would be crazy enough  to resent the intrusion of a visitor from  there?) and the spot on the street where, forevermore,  imaginations will pause the moment it takes to join Pierre beneath his cloak, contemplating the tick- tock of history under the aegis of  the big comet-- â€œIsti Mirant Stellaâ€--a clock and a barometer of strife.  Nobodyâ€™s winter ends quite so abruptly: the arrival of spring here is a much- repeated story.  (If music can reverberate in a theoretical forever, how much more cogently can a clap of springtime thunder?) Thunder -- to Joyce it meant The Fall: for us, it didnâ€™t need to be emblematical of anything; it simply was, no hundred- letter word required; thunder spoke, and the midnight dreams of long winter evenings shattered.  We stood to behold, in the rudest kind of light, the â€œogre doing what  ogres can,â€ rats pacing before a window full of jewelry, our own little ship, more frail than the tsarâ€™s hand-made one, in full retreat before a flotilla of invading icebergs in the blank blaze of afternoon. Now the bridge and the walk across it, and around the corner to the Bolshoi, where  the everlasting concert is still in progress,  are the province of the power that re-creates the candle- flame after itâ€™s puffed out: the smoke that hangs forever on motionless dead air in unfocused space is its domain, heraldic such as all recaptured time  must be, and that goes for music as well: the Schubert symphony through which we sat, holding hands at the conservatory, the slow movement of  Beethovenâ€™s Emperor, Ewig, summoning up the  morning its strains ran through my head as I watched the sun come up in California...who is to call these things any  less real than the table- top across which they and I exchanged mutual glares, or  the piece of paper I  had to sign, that sop to those who wanted the war to go on, because without it they  were utterly lost? This glade they canâ€™t reach: the hiker who gets here is beyond their grasp, like Nietzsche scaling mountains to escape his dead father, like Beethoven the other side of every torment, laughing along with whatever god could hear, in the last movement of Op. 131. The crowd has passed the bridge now; across the river in Park Kulturâ€™i the ferris wheel turns beyond time,  just another wheel within a wheel, the Chinese box  of perception that may or may not have a center at all; it doesnâ€™t matter as long as I see your eyes on every street-corner stolen from the too-insistent sun.  Turn the corner again: Rachmaninoff, all six feet six of him, takes his place at the piano, unsmiling, but â€œmagnificent.â€  He plays, we listen, in the room, or just outside the window, where the breeze coming up from  the awakened river takes the sound down the street, across the fence, over the frozen flowerbed, around the corner, into eternity.  He plays on...      Coronado      (in memory of Michael C. Baker, 1953-1994)  A long time ago you needed the ferry to get over there, but that was before our â€œflowering,â€ when places like Mickey Finnâ€™s and the island were local playgrounds for airline crews laid over in this dead Navy town.   later this useful eyesore went up--it looks lousy on postcards, unlike its over- photographed neighbor to the north.  Today, festooned with signs bearing a number you can call if youâ€™re tempted to jump, it makes an undistinguished lazy-S across what was once a beautiful bay.   I think of you while driving underneath--it was right along here somewhere that you â€œditchedâ€ me that night, because you wanted  to be alone with my sister. To think there was a time I would have tossed you off this bridge without a second thought: that unfortunate first meeting, about which we  always laughed later,  calling it â€˜When Robin Hood met Little John.â€™  This span was going up at the same time we were, during that nasty polyester decade, which some wisecracking journalist named after a pronoun. Nixon.  Brezhnev.  Disco.   yuck.  The layered look. platform shoes found their architectural equivalent launched across the water in tacky splendor from  National City, and a more perfect starting-place I couldnâ€™t  have picked myself.   the lights were nice at night, though, when the lights were all  you could see. We cruised beneath on my graduation night, a night going down the tubes completely, until you showed up at 10 p.m., after everyone else had gone to bed.  I was parked in front of the tube, feeling hugely sorry for myself, and then you came in from work, a â€˜64 Taittinger in hand. As the champagne cooled in the freezer, we went out record-shopping.  It was still the age of LPs--David Bowie and Bette Midler were giant faces in the window at  the hot-dog stand of  Tower Records. Later, after the music,  the champagne, a pizza around the corner, we stumbled on home through the jasmine-scented, soft June night--â€Coffee beans, cha-cha-cha!â€---hard to believe one was ever young enough to get that high on two or three glasses of wine.  Some thought you a hooligan, but I knew that your swagger down the wild side was mostly make-believe: the telephone receiver in the Volkswagen was a stage-prop, the Luger had no firing-pin. One had to be undeceived in order to become your friend. I was, and I did. O those sweet cruising years, from then to the end we had them to live again and again, first not to value enough, to enjoy for a shorter time than it would later seem, and then to remember,  to replay and embellish as hair got thinner and waistlines thicker. Somehow, though, you stayed there, played out the tragedy of--who?  Literature and the movies have only a few prototypes: Peter Pan, maybe, or closer to our own time, John Milner, whom we actually discussed, seriously, while driving home from the movies after seeing â€œAmerican Graffitiâ€ for the  first time.  The cars always remained flashy for you, and the guns eventually became real, though you only popped them off up in the hills on Sundays. But that awful decade accomodated itself  to your dreams: after all those tie-dyed  years of  sit-ins and demonizing, now, suddenly, The Pigs became the new centurions--there were 200 cop shows on the tube, a new sympathy, and you found a whole new crop of heroes. (and either downtown at the movie houses, where the jujubes stuck to your shoes like shit, or at the drive-in, slouched down in your sleek â€˜67 Mustang on  a dead Friday night, from  â€œSerpicoâ€ to â€œThe Super Cops,â€, we saw them, every cheesy one.) Michael, you wanted to be just like that, and maybe thatâ€™s why it didnâ€™t happen: you did persuade the reserves to take you, and Iâ€™ll never forget your pride when you showed me that Colt, new, shiny, quiet, deadly, and O-so-shimmeringly male, and the rights and priveleges pertaining thereto you took so seriously as well. The fourth of July:  we were driving out to Coronado, bored out of our skulls, thinking weâ€™d watch the bicentennial fireworks downtown and maybe split a six-pack. we took the long way around, by-passing the bridge.  Then the traffic slowed.  There had been an accident.  You felt in your wallet for that badge they had given you, then insisted on stopping to help lay out flares and wave traffic on by. I sat in the Dodge, annoyed, the radio for company, while I waited for you to help in the cleaning-up of a mess I couldnâ€™t even see.   We made it to the beach, kicked back on the sand with some Schlitz .  (It was always your brand, after you saw Clint Eastwood pull one from the fridge in â€œMagnum Force.â€) In the shadow of that hotel we remember from â€œSome Like It Hot,â€ while the black ocean talked its talk forever fifty feet in front of us, you replayed  the accident, dreaming of bigger future things: breathless code threes, drug busts; â€œNarcotics, Baker.â€ it didnâ€™t happen.   You had to learn that the Colt was far from the whole story. the city justified the badge, and not the other way around. Not long after, I looked you up. I had to be told  where to find you.   The apartment, the bimbo, the un-cleaned litter box, the drugs...Yes, it was two steps back all right,  and when I left, in a hurry, mumbling excuses, you hardly  looked up from  the screen.  but that time you came back like a written-off cat, and got back to business. Glen Abbey always struck me as a ghoulish place for  weddings: a chapel in a cemetery; talk about tempting fate. (Thereâ€™s a certain lexical justice, I think, in the rhyme of married and buried.) But the place was popular for that; probably still is. That year I drove up from the desert four times for weddings; two of them were there, and one of them was yours.  Memory may be tweaking the facts to jive with the heart, but I think I remember some mischievous look, some almost- eyeball-rolling shared between us two former outlaws, as you slumped out of the  chapel, your bride on your arm. Then life continued its accustomed business,  levelling-off, wearing-down and thinning out. Men now, we watched our days meld one into the next, trying not to notice how they were steadily picking up speed.   We saw each other now only when I was passing through. life had taken me elsewhere, you stayed close to home. But always weâ€™d make a point of an evening, now among broken toys, potty seats that had to be drunkenly removed, carpets scattered with picture books, and walls hung with reminders of the distance into which youth was fading, to tell the old jokes again, replay the heartbreaks, (my sister was yours) and then, with embarrassed side-glances, awash in a mnemonic pool of Schlitz and Cutty Sark,  pop â€œThunderballâ€  into the VCR and fall asleep on the couch.  Some say there are no such things as accidents. A large-busted, brainless girl I once dated would have said that there was a reason I was in town that spring, just happened to be taking a little R&R, and thought I\'d check in. I had in mind nothing more or less than the usual obligatory evening of getting mawkish over beer and darts in the garage.   Your wife answered the phone. \"Are you sitting down?\" she asked me.  No Schlitz and Sean Connery this time-- I drove out to say goodbye. I tried to keep things light, popping wheelies in  your wheelchair as you sat propped up in bed. Morphine blurred the edges of the old jokes, but we told them anyway, and even, incredibly enough, talked about my sister one more time--I didn\'t Have the lack-of-heart to tell you that she\'d gained 100 pounds, not that it would have  mattered much by then. I was back in Moscow, tking deep breaths and smelling shit, when she called from  eleven time-zones away to give me the news. Punching the CD player almost at random that night over a Scotch, what do you suppose I got?  Vespers--his version. I gave it a run-through--not a deliberate memorial; music was never anything to you eExcept window-dressing,  accompaniment for the race downtown, or some moment to be embroidered later, like the night you danced in the parking lot with Sue, throwing open the doors of the Dodge so the car-stereo could bounce Lalo Schifrin off the walls.  I didn\'t see that, but I heard about it later. no, it was an accident (or maybe not) that once again, one more time, Rachmaninoff came drifting by like the H E L Y S guys in \"Ulysses,\" only there if you happened to be paying attention at the time.  Do I believe in angels? Well, maybe not as a regular thing, maybe just that frowning one. home for Christmas again, I watched the ribbon of that ugly thing make like a serpent across black water one more time while riding home from the airport. Night gives it a break, I\'ll admit--at night all bridges are the same, just light. we had to see the place,  it had to be done.  Lynne drove. You\'ll like this, though: we got hysterical: someone had put up Christmas stuff in Glen Abbey, in full view of that chapel, your spot, cornball Liberty Lawn where the World War II veterans rot, and the shopping mall across the valley.  Reindeer, Santas and little creches in the middle of the cemetery. We raced back, in stitches, humming James Bond car-chase music between spasms.   That decade named for a pronoun was about good feelings, after all.  (That and bad taste.) So-- you would have liked it-- we did, I swear: Lynne gunned the engine coming off the light at the I-805 overpass, and we laughed all the way home.  1995-1997  '),(12119,3380464,'22,May,2004','2004-05-22 00:00:00','   Some thoughts Kelley wrote in his journal after reading Solomon Volkov\'s \"Testimony,\" which purported to be the ghost-written memoirs of Dmitri Shostakovich.  This journal entry is dated April 28, 1982:  Last week I finished reading the Shostakovich-Volkov memoirs, and found them profoundly depressing.  They shared a similarity with Solzhenitsyn\'s writings in making an indictment of the Kremlin regime, but there the similarity ended.  For Solzhenitsyn, two steps to the right of Ivan the Terrible as he may be in some ways, at least believes in something. He\'s deeply religious, (which I suppose is a fundamental part of being two steps to the right of anybody) a practicing member of the Orthodox Church.  Shostakovich seemed to believe in nothing--not man, not God, not Stalin, not progress, nothing.  Even music, music which he must have loved and to which he presumably had a deep commitment, seems to have given him only fugitive consolations.  Music was music, it was joy, it was art, but it was not something you could pin hope or faith on; music couldn\'t DO anything; it just WAS.  This seems to have been Shostakovich\'s attitude, and it\'s a strange one, since in the Soviet Union so much music, and even much of Shostakovich\'s music in the war years, was made-to-order for propaganda purposes. 	 Shostakovich, whose refusal to write upbeat, optimistic music, fanfares and choruses of the \"Onward, Communists\" ilk made him a thorn in the Kremlin\'s side, consented during the war to do a certain amount of propagandizing in his music not because he loved the regime, but because he had that deep, utterly irrational mania for \"The Motherland\" that\'s so characteristic of Russians.   \"The Motherland:\" despite his repeated sufferings at the hands of bureaucrats and toadies, the public spanking which he, Prokofiev and other Soviet composers were forced to take from Stalin in 1937 (precipitated by Shostakovich\'s opera \"Lady MacBeth of Mtsensk,\" which so offended Stalin\'s tin ear that he promptly dictated to one of his toadies an article entitled \"Muddle Instead of Music,\" which appeared in Pravda and was the opening shot in the regime\'s ongoing campaign against \"formalism\") and the banning of his works repeatedly, the idea of leaving the USSR never occurred to Shostakovich.    In fact he had the oriental xenophobia about the West and things western which is as typical of the Russian mind as Russians\' equally irrational fanatic patriotism, and probably has the same roots. Did his war propaganda music amount to hypocrisy? (The Seventh Symphony is the specific work I\'m thinking of; there was the famous \"battle\" between Toscanini and Stokowski to see who would get to conduct its premiere, followed years later by Toscanini\'s wondering remark at hearing a recording of it, \"How could I ever have conducted such trash?\"  Glenn Gould has weighed in similarly, calling Shostakovich\'s 7th a \"motoric monstrosity.\") No. It wasn\'t hypocrisy any more than was Copland\'s \"A Lincoln Portrait,\" written at about the same time. Russians, ordinary Russians, as much victims of Stalin as Shostakovich himself, were fighting and dying against Hitler\'s armies, and it was on their behalf, not Stalin\'s, that Shostakovich wrote such music.  Still, he struck me as having a rather primitive mind.  Contemptuous of western \"humanists,\" whom he viewed as stupid dupes of Stalin\'s propaganda, and equally contemptuous of his contemporaries who became dissidents or criticized the regime, Shostakovich comes across in his memoirs (and indeed in much of his music) as an arrested adolescent.    Yes: there is, beneath all of the bitter-old-man grumbling that fills page after page of \"Testimony,\" a strain of the rebellious, pig-headed teenager, the withdrawn loafer-about-the-halls who regards all of his classmates and teachers as idiots, and even resents the intrusion into his choleric solitude of those who may feel similarly about the system. The Fifth Symphony, the Eighth String Quartet and the Tenth Symphony (I\'m using these examples because they\'re the only pieces of his I\'ve heard lately) are all dramatic works, full of mockery, bitter bluster, and in the case of the quartet, some genuine and searingly intense moments of despair.  But the emotions seem exaggerated, raw, again, adolescent. Granted, this could be partly the result of Shostakovich\'s famous affinity for Mahler.  (Leonard Bernstein listed Shostakovich and Britten as composers who \"are at their best when they are most like Mahler,\" and Stravinsky, upon meeting Shostakovich in Moscow in 1962, told Shostakovich that while he, Stravinsky, shared some of the latter\'s high regard for Mahler, he admonished the younger composer to \"go beyond Mahler,\" by which presumably he meant on to Schoenberg, something Stravinsky should have known Shostakovich couldn\'t do, 12-tone music being forbidden in the USSR).    But in a deeper sense I think it stems from the same sort of primitive mentality (and by \"primitive\" I certainly don\'t mean unintelligent) that allowed Shostakovich to hate and despise Stalin and the whole government framework of ratfinks and asskissers, not to mention the secret police, and yet still resolutely refuse to even entertain the idea of leaving his beloved Rooshia for the wicked west, where the intellectual and artistic climate would have been freer, if perhaps somewhat more cluttered and certainly noisier. 	 It\'s noteworthy that Shostakovich sneers at Sakharov, whom he considers a hypocrite for having built Russia\'s H-bomb and then turned dissident against the powers who could use it.  It\'s also worth noting that he apparently didn\'t get along well with Solzhenitsyn, to whom he makes a few oblique references, calling the author of The Gulag Archipelago a \"luminary\" in the most sardonic of terms, and slyly depicting Solzhenitsyn as a pompous, holier-than-thou stuffed shirt, drunk with the feeling of his own importance and determined to lead Shostakovich away from the erroneous path of negativity.  Shostakovich implies that Solzhenitsyn berated him for having no faith in anything, and Shostakovich resented such a preachy attitude, as any teenage rebel with his hat on backwards would.  He might have admitted that Solzhenitsyn had actually done time in Stalin\'s labor camps, while he himself was never even arrested, and therefore Solzhenitsyn, should he so desire, had even more justification for being bitter about the regime and indeed about life itself than Shostakovich, but he chooses not to be gracious here. In fact graciousness doesn\'t come across as one of his virtues anywhere in this volume, though he does decry \"rudeness\" in other people, including Stalin himself and notably, Toscanini. 	 Shostakovich supposedly dictated the material that makes up his memoirs to Solomon Volkov, who made. S.\'s ruminations into prose and got them published, ironically, in the west.  The writing of the book took place very late in Shostakovich\'s life (he died Aug. 9, 1975) and his outlook is as saturated with the fear of impending death as it is with bitterness toward all those who had made his life as a composer in the Soviet Union a Kafkaesque nightmare.    What an abyss to be facing: Shostakovich, a dutiful Communist and Soviet citizen despite his quarrels with Stalin and particularly with Stalin\'s revolting musical hatchet-man Khrennikov, had no shred of religion in him, was as emphatically positive that there is no afterlife as he was unsure, during his life, that he would live to see his next work performed.  What a spot to be in.  After an entire life of the bitterest kind, to contemplate a death that is the end of all, to have no hope of redemption or justification or salvation of any kind beyond the grave.    This actually suggests to me, once again, a kind of shallowness on Shostakovich\'s part, not courage. Not just because he lacked the imagination or the desire to believe in God or an afterlife, but because he lacked the imagination or the desire to believe in ANYTHING.  It\'s reasonable to expect that those who do not believe in God will substitute a belief in something else: John Reed, for example, had a faith in Bolshevism as childlike as any Sunday school child\'s faith in lamby Jesus.  But Shostakovich, while he dutifully liked to identify himself as a \"proletarian,\" a \"worker,\" gives no evidence anywhere in his memoirs of having a faith like Reed\'s that world revolution would indeed miraculously come along someday, turning the earth into a big Marxist day camp.    There isn\'t even any evidence here of the kind of hat-tipping lip-service that the poet Yevtushenko always made it a point to pay Lenin.  (With an eye, naturally, to Y\'s own continued prosperity under the system that made Lenin, in the words of one critic, \"a cross between Abraham Lincoln and Santa Claus\" to most Russians.)    There is in fact a complete absence of Communist buzzwords and Marxist-Leninist cliches in Shostakovich\'s remembrances: no \"people\'s struggle,\" no \"socialist destiny,\" no \"western imperialism.\" (By contrast, you find plenty of this kind of stuff in Yevtushenko.) There is nothing in his book to indicate that his relationship with Marxist-Leninism, his \"ideology\" to use the word the Soviets like to equate with \"virtue,\" is anything more than an acquiescence for the sake of peace, a mouthing of the appropriate words to keep the authorities off his back.  He certainly never became a Kremlin poster boy as Yevtushenko did, although they were friends and Shostakovich used Y.Y.\'s poetry as the backbone of his Thirteenth Symphony.    If Yevtushenko disappointed western conservatives and Soviet liberals by turning away from human rights and more and more toward mouthing the party line in the years following Kruschschev\'s ouster in 1964, Shostakovich was a similar disappointment to Soviet conservatives, who like their art to be positive, up-beat, supportive of the government, easily accessible to the masses (meaning the bureaucrats, of course, with their lowbrow tastes).  What they wanted was of course Sound Effects for the Five-Year Plans, and Shostakovich would not give them that.  It wasn\'t because he wasn\'t an obedient Communist, but rather because he was melancholic, ascerbic and withdrawn by nature and he discovered, though he was slow to learn it, there there is no place in a totalitarian regime for anyone\'s personal foibles. Except maybe at the top. 	 Not everything in these memoirs is bitter, though.  There are many people whom Shostakovich remembers fondly and treats with respect, if not love.  One of them is Alexander Glazunov, the \"Russian Brahms,\" as he is referred to in a footnote.  (If Glazunov\'s stature equals that of Brahms, it\'s only in the USSR; in the west he\'s regarded as a minor composer who lived at the same time as Tchaikovsky and Rimsky-Korsakov.)  Shostakovich liked him; he devotes a great deal of space in his book to singing Glazunov\'s praises, though his praise is for the man, not the music.  He seems to have regarded Glazunov\'s compositions as being every bit as mediocre as most of the world finds them.  Nevertheless, he assures us that Glazunov was a generous and effective teacher, despite being so deeply rooted in the past that he could not manage under the Soviet regime and retired to Paris, where he died in 1935. 	 But his warm memories of Glazunov are one square spot of sunshine on a floor that is otherwise mostly covered in shades of gloom.  To read Shostakovich\'s memoirs is to plumb the depths of a Dostoevskyan soul.  In fact, if I had to make a literary analogy with the two greatest composers Russia has produced in the last century, Shostakovich and Stravinsky, I could draw parallels with Dostoevsky and Chekhov, Shostakovich playing the former to Stravinsky\'s latter.  On the one hand, you have Dostoevsky-Shostakovich dealing in bombastic, highly personalized portraits of suffering.  Nearly all of Shostakovich\'s works, it\'s worth noting, have some kind of autobiographical program; if it\'s not actually printed on the album jacket, it\'s innate in the music.  Like Dostoevsky\'s, Shostakovich\'s art is not subtle, and whereas Dostoevsky is frequently accused of clumsy, ungrammatical, effusive writing, qualities redeemed of course by his raw power and insights into human torment, Shostakovich could be accused of something similar in his refusal to \"go beyond Mahler,\" to reach beyond orchestral mood-painting and heart-on-sleeve expression of personal feelings to an art that speaks more for itself.  On the other hand you have Stravinsky, whose aesthetic was more chekhovian, i.e. a tad more \"refrigerated.\"  Indeed Chekhov\'s own dictum about good writing was \"it must be cold.\"  One of Stravinsky\'s most often-quoted statements was a reply he made to a questioner who asked what it is that music expresses.  \"It can express nothing,\" he replied.  \"It can express itself only.\"  In Chekhov\'s fiction the author is invisible; the characters act on their own, suffer on their own, triumph on their own, all without a trace of sentimentality or a \"tortured glance into the soul\" other than what the character himself might show.  So too, in Stravinsky\'s music, \"emotion\" is kept at arm\'s length.  The emotions are there, but expressed at a remove, through the application of various technical artifices. 	 The composer whom Shostakovich seems to love the most, to relate to on the most personal level, is Mussorgsky.  He even went so far as to make his own adaptation of \"Boris Godunov.\"  It\'s not surprising, really: Mussorgsky was even more a \"Dostoevsky\" character than Shostakovich, not only an original, but a primitive--a genuine product of that Russian Soil to which Shostakovich had such an affective attachment.  And Shostakovich must have seen some parallels between Mussorgsky\'s miserable life (he drank himself to death) and his own (he was nearly hounded to death by the government, or so he would have his readers believe).    Misery is, in any case, the common denominator.  Misery and the love of music, whose own redemptive, regenerative or spiritual value (to use a term Shostakovich would have peremptorily rejected) is highly in question at the end of the book.  Or maybe looking at it that way is to miss the main point.  After all, Shostakovich wrote music all his life, and continued writing it right up until shortly before his death.  This in the face of Stalin, Zhdanov, Khrennikov, the purge trials, the KGB, Hitler...of what indeed seem in retrospect like all the forces of negativity swept into one vast, grotesquely unhappy country that covers one-sixth of the earth\'s surface.  In the face of all that, Shostakovich\'s despair, his utter lack of any belief, religious or otherwise, becomes less surprising.    What becomes surprising is that he produced so much music, and not all of it sardonic, mocking or full of despair.  There is humor in many of his works, the Second Piano Concerto for instance, and in fact one of the poems he set in his Thirteenth Symphony was Yevtushenko\'s \"Ode to Humor.\"  His 24 preludes and fugues for piano have moments not only of humor, but even of profound joy.  How to explain it?  Perhaps there is no explanation.  Art grows in the most unlikely places; I remember reading of little drawings and paintings found in the ruins of Hitler\'s concentration camps, works of art produced by children waiting their turn to go to the ovens.  Art\'s existence isn\'t logical, surely, but then existence itself isn\'t.  But art\'s very existence is therefore a form of hope, a belief in tomorrow.  Maybe Shostakovich knew it and just wasn\'t telling.  '),(12120,3380464,'22,May,2004','2004-05-22 00:00:00','   From Kelleyâ€™s journal, November, 2003  Sometimes a confluence of events comes along and pokes me in a certain direction. It has happened occasionally throughout my life. And they donâ€™t have to be big events, such as the loss of employment, followed by the failure, over the next few months, to find employment, followed in turn by a drive across America, my old 1990 Geo laden with my few pitiful possessions, as happened last spring.  They can be small events, like the reading of an essay following upon the heels of a conversation, followed by the rediscovery of an old, familiar volume. That, in fact, is what just happened, and as a result I feel that my life as a reader has been, in some small way, kick-started. In any case, after having eschewed fiction for a long time, I am reading again, in a tentative way. But throwing tentativeness in a sense to the breeze, I have undertaken a formidable project in that area: \"Little Dorrit.\"  \"Little Dorrit?\" Yeah. For most of my adult life, indeed, for most of my life as a reader, I have had an allergy to the Victorians. All that windiness, all that length, all that prudery, hypocrisy, imperial smugness. Who needed it? Twenty-some years ago, Ray Araiza used to tease me about my proud claim that â€œI donâ€™t read the Victorians.â€ Hemingway and his generation had fought the good fight to liberate American literature from the stranglehold of Britannia. I was their heir, or so I thought. What did I need with Dickens, Thackeray and company?  Well, the journey to \"Little Dorrit\" actually had its earliest beginning in September, 2001, just days before the terrorist attacks of that year. Tatiana Floyd (Russian-born, six feet tall, 150 pounds, and an avid photographer) and I were driving from Baltimore to Boston for the Labor Day weekend.   We took along some books on tape to listen to in the car during the eight-hour haul up Interstate 95, one of which was a collection of short pieces by Tom Wolfe, \"Hooking Up.\"   One of the essays on that tape concerned the brouhaha which followed the publication in 1998 of Wolfeâ€™s novel \"A Man In Full.\" Specifically, the fuss that three famous, and jealous, fellow-writers kicked up over its success. Norman Mailer, John Updike and John Irving all went on television at different times to trash \"A Man In Full,\" Mailer claiming that it was â€œjournalismâ€ and not fiction, (he ought to know about that!) and Updike stuffily asserting that it wasnâ€™t a product of â€œourâ€ literary culture.    But John Irving outdid his two fellow sniffers: in an appearance on Canadian TV, the jealous author so lost control of himself on the subject of \"A Man In Full\" that he was liberally using the â€œFâ€ word. And hereâ€™s the rub in the case of Irvingâ€™s on-the-air tantrum: Wolfe mentioned that some reviewers of \"A Man In Full\" had compared it with Dickens, and that was what really got Irvingâ€™s goat: Irving, it seems, is a great admirer of Dickens, who would like to be compared with Dickens himself. To see Wolfe compared with Dickens drove John Irving into â€œF-wordâ€ convulsions in a TV interview.  This intrigued me. What possible attraction, I wondered, could Dickens have for the author of \"The World According To Garp\" and \"\"The Cider-House Rules?\" I would think that two more dissimilar writers couldnâ€™t be imagined. The idea was tucked away in my memory under â€œcuriosities from the world of book-chat.â€  Then, maybe three weeks ago, I was picking through \"Something To Remember Me By,\" a volume of short fiction by Saul Bellow. Bellow wrote an introduction to this tryptich of novellas, the theme of the introduction being precisely this fact, that they were novellas as opposed to novels. He discussed our general failure in recent years to make as much time for reading as people used to, hence a general trend to make fiction shorter than it used to be. I donâ€™t remember the exact context, but as an illustration of the sort of long novel to which people no longer want to bother making a commitment of time and effort, Bellow specifically cited \"Little Dorrit.\"   Item filed, note with question attached: why \"Little Dorrit?\" Dickens wrote plenty of long novels. Why didnâ€™t Bellow cite \"Bleak House\" or \"The Pickwick Papers?\" It seems to me that they are both generally better known.  Next came my conversation at Marie Callenderâ€™s last Wednesday night with Jan Barnett. Jan told me over soup and buffalo wings that she was endeavoring to stake out some time in her life these days for the things she considers important, not the least of which is reading. When she mentioned that she was reading some of the short stories of Colette, it was like someone had tossed a glass of cold water in my lap. I suddenly found myself thinking back to my college days, or at least to Janâ€™s, when we were all, each in our own way, so interested in literature that we were reading like fiends. In my case the drive was especially strong because I wanted to be a writer, wanted it worse than anything, and of course as my erstwhile teacher Don Baird had said to me when I was 18, â€œAs for writing yourself, keep reading. Sure, all writers are readers.â€   From my teens until I was about forty, it seems I was always reading something, usually something from the â€œwestern canon,â€ e.g. something from the entire pantheon of serious western literature, running the gamut from Homer to Tolstoy, from Thomas Mann to Saul Bellow. (For the most part steering clear of the Victorians, except for Oscar Wilde, who flouted their conventions.) But as my forties progressed and the realization that I was not, after all, going to be a Hemingway or a Henry Miller or even a W.H. Auden began to coalesce in real time like a photograph in the darkroom becoming ever-sharper, my interest in reading great literature gradually began to fade, to extend the simile, like an old Polaroid. Now hereâ€™s Jan, who by the way has earned my admiration for the graceful way in which she has accepted her own version of my experience: when she was in college, Jan dreamed of becoming a great artist. She knows now that she probably isnâ€™t going to be the next Georgia Oâ€™Keeffe, but she has accepted the fact with equanimity, still enjoys drawing, and by the way, is trying to block out time these days for such things as reading the short stories of Colette. Noted and filed.    Next, just yesterday morning in fact, I was reading an essay by critic Sven Birkerts in a book of his that Lucia gave me, \"Readings.\" The essay, \"Against The Current,\" concerned itself with Birkertâ€™s experienceâ€”and, by extension, our experience generallyâ€”of â€œlosing touchâ€ with the world of close reading and the sparks it can cause to fly, thanks mostly to the way our postmodern perceptions have been totally taken over and reshaped, even redefined, by the all-pervading ocean of electronic media in which we spend every moment of our waking lives these days. Using as a starting point his self-described inability to read and appreciate poetry as he once did, Birkerts moves on to a detailed discussion, first of how hisâ€”ourâ€”altered modes of perception have endangered the attentiveness needed for reading, and then to details of some of the small steps he has taken to try and recover some of that, chiefly by making the sacrifice of doing some things in a deliberately slower, less â€œefficientâ€ manner than they are usually done these days: writing letters with a pen rather than a computer, and then taking the time to walk to the mailbox to mail them, noticing things around him on the way. Slowing down, in other words, and tuning in while at the same time tuning out.  All of these little experiences brought me to a decision: I was going to read \"Little Dorrit.\" Yesterday afternoon I got in the car and drove over to the Chula Vista Public Library to see if it was on the shelf. I knew that it probably would be; after all, who reads \"Little Dorrit\" anymore?  And then, as I was entering the library on this mission of reading, another tiny fillip of experience occured, a sort of closing-the-circle gesture on the part of the gods, which, come to think of it, could not have been more perfect had it been scripted for the occasion.   The libraryâ€™s little used bookstore, tucked away in one corner of the main library, is open on Saturday afternoons. I seldom go in there, because they seldom have anything that catches my interest and anyway, in my current living circumstances I donâ€™t have much room for storing books.   But as I wandered into the little shop yesterday, and browsed around the cramped shelves, I spotted an old friend: \"Literature: Structure, Sound and Sense,\" by Laurence Perrine. (Harcourt, Brace & World, Â© 1956, 1959, 1963, 1966, 1969, 1970.) This was the very textbook that we used in Donald S. Bairdâ€™s English 6 class, â€œComposition and Literature,â€ Southwestern College, fall semester, 1973. (MWF 8:00-8:50 a.m.â€”Imagine discussing T.S. Eliot at eight Oâ€™clock in the morning! Still, we did.)  I was 18, it was my first semester of college, and this textbook, along with Bairdâ€™s own curmudgeonly pontificatings, was a key factor in the shaping of my own tastes in poetry and fiction during the years that followed. (Bairdâ€™s greatest gift to me was Yeats. He could be a little prick, but he did me that favor.)   Donald S. Baird is probably dead by now. And there was that book. Did I say â€œthe gods?â€ More likely, Bairdâ€™s own curmudgeonly little ghost patting my butt as I entered the library in search of Dickens, giving his seal of approval to the quest. As I recall, the cost of this textbook in 1973 was $10. I got it back yesterday for 75Â¢. It, and \"Lake Woebegon Days\" by Garrison Keillor, and yes, \"The Mill On The Floss\" by George Eliot, eminent Victorian. Total for all three: $1.75.       The libraryâ€™s two copies of \"Little Dorrit\" were both inâ€”surprise!â€”in fact I had my choice between the one in old blue library binding and the one in old red library binding. Both are slightly yellowed and just a shade tattered. I chose the red one: New York, Dodd, Mead & Co. 1951.  As of this morning, I have read up to Chapter Seven. Itâ€™s going to be a long journey, as Saul Bellow promised it would be; Iâ€™m on page 64 of a book that runs 788 pages. And so thoroughly has the world changed between Dickensâ€™ time and our own that I am already having occasional trouble â€œtaking his sense,â€ not so much with regard to the language as to the sensibilities of his characters. The Victoriansâ€™ shared system of values and beliefs, not to mention their customs, bore little resemblance to whatever shared system of beliefs we have left in the age of the Internet. But no matter, it is giving me a warm feeling in the gut to begin this long journey, and I am determined to see what lies at the other end.   II   Two weeks ago I announced in these pages that I had decided to read \"Little Dorrit\" by Charles Dickens. Iâ€™m just about halfway through it now, a few pages short of finishing Part One. What are my impressions of the book so far? Well, right off, I can see why some reviewers compared Wolfeâ€™s \"A Man In Full\"  with Dickens. An argument could indeed be made that Dickens was the Tom Wolfe of his day, or Tom Wolfe the Dickens of ours. Michael Burgess and I were discussing Dickens and my decision to read Little Dorrit the other day. â€œDickens was a journalist,â€ Michael remarked. Indeed he was, in the same sense that Balzac and Zola were journalists. The landscape of London at the beginning of the industrial revolution was what Dickens painted, and he was always exposing the social ills of his era; \"Little Dorritâ€™s\" particular target is the institution of debtorâ€™s prison. In the custom of that time, his novels were serialized in magazines before they appeared between covers, so he was indeed writing for a popular audience. Novels donâ€™t get serialized in magazines anymore, but Wolfe was a magazine journalist before he began writing books. The parallels are easy to draw, which makes me wonder about John Irvingâ€™s hissy-fit on Canadian TV.  An American reader in the early twenty-first century canâ€™t help but find Dickens a little verbose. It isnâ€™t just because we donâ€™t read long books anymore, either. Both journalism and prose fiction have become noticeably less long-winded in the past 100 years. As a journalist and one who has studied the history of journalism as well as of literature, I can testify that as you progress from 1900 to 2000 in reading newspaper articles, youâ€™ll find them progressively less and less â€œwordyâ€ until you reach todayâ€™s journalism, which is so terse by comparison with earlier eras as to seem like shorthand. When I pick up a newspaper article written at the time of World War I, Iâ€™m aware that Iâ€™m reading prose. Ornate sentences, carefully crafted. Curlicues of simile and metaphor. Itâ€™s obvious that some of these guys were writing with pen and paper, not typewriters. In fact it wasnâ€™t until the 1960s that this kind of thing finally disappeared. As the newspaper market shrank, newspaper writing became less and less distinguishable. Journalism isnâ€™t crafted at all anymore, unless youâ€™re talking about the opinion columnists. Journalism today is churned out as product. Pick up the front page of any major newspaper and the reporting of any two journalists will read pretty much like the reporting of any other two. Formulaic, brief and to the point.  Of course Dickens wasnâ€™t writing newspaper stories, he was writing fiction. But he was writing in a leisurely, mannered style which was the norm of his day and not of ours, whether youâ€™re talking about journalism or fiction. Leisurely, mannered prose fiction was precisely what people like Hemingway, Nathanael West and Raymond Chandler were trying to get away from. They, and their contemporaries, laid the ground rules for the kind of fiction weâ€™re used to reading now: pithy, from the hip. Don DeLillo and Cormac McCarthy, in fact, have taken this postmodern stuff so far that their prose resembles Marlon Brandoâ€™s mumbling. That\'s the stylistic half of the equation: the other half is the length issue, an outgrowth of our \"eight second sound bite\" culture.  I found DeLillo\'s \"Underworld\" a bit of a long haul, but it\'s definitely against the grain: Oprah\'s Book Club features a lot of McBooks that can be tossed off in an afternoon, although some contemporary authors, even some who pre-date DeLillo\'s efforts in the 1990\'s, have gone out of their way to be unaccomodating to our short attention spans: I think of Pynchon, Barth and somewhat more closer to our own time Vikram Seth, whose \"A Suitable Boy\" was such a massive edifice of the now-nearly-defunct \"Bildungsroman\" genre that it drew comparisons with Goethe and Tolstoy from the British critics in when it was published in 1991. (And it sank like a rock nonetheless: Goethe and Tolstoy notwithstanding, it was on the remainder tables within six months.) I think even of Garcia Marquez, who dispensed with paragraphs in \"The Autumn of the Patriarch.\"    But these are acts of intentional obtuseness: guys like Dickens and Henry James were long-winded because that was what their audience expected, not the opposite. They werenâ€™t flying in the face of anything.  And then there is the question of what used to be called â€œsensibility.â€ The 19th century was (I should say, is) infamous for its sentimentality. From the time of Rousseau until the massive global disllusionment engendered by World War I, public taste tended toward bathos and tears. â€œFeeling is all,â€ Goethe said in \"Faust,\" and he may have meant it ironically, but he wasnâ€™t kidding. For a century, novelists, poets and playwrights laid it on with a trowel, which is why we find so many of them unreadable now. \"Uncle Tomâ€™s Cabin\" was credited by no less than Abraham Lincoln with being no less than the spark that started the American Civil War, but today one canâ€™t read it without laughing.  Iâ€™m not comparing Harriet Beecher Stowe with Dickens, but I am saying that the â€œsensibilityâ€ of the mid-19th century tended to favor scenes and characterizations which today we would consider mawkish. Snapshot from memory: my 10th grade English teacher, Mrs. Terry, (who was very much a child of the â€œhipâ€ 1960s) mercilessly ridiculing Longfellowâ€™s \"The Wreck of the Hesperus\" for its gooey sentiment. Again, Iâ€™m using an extreme example: Longfellow is a second-rate poet. But my point is that you do find in Dickens, or I do anyway, some undeniable traces of this pandering to the â€œsensibilitiesâ€ of his time which can make some of his characters seem a little unbelievable to modern readers. Little Dorrit, so christlike in her self-sacrificing, so relentlessly sweet, humble and devoted to her father, looks to me like Mary Pickford hamming it up in a silent film. Arthur Clenham is a painfully nice guy who, in the manner of his time, goes around acting like he has no dick. Even when he falls in love with Pet, he tries to persuade himself that he hasnâ€™t. God forbid that any Victorian should admit having a dick. (Curious, or perhaps not quite so curious after all, is the existence, of which we are now fully aware, of a very active and fecund pornographic sub-culture in Victorian England, of which â€œMy Secret Lifeâ€ and â€œThe Pearlâ€ are two famous examples.)  But having said all that, there is a great deal about \"Little Dorrit\" that has a contemporary ring. England no longer has debtorâ€™s prison, but reading about it reminds me of how thoroughly our American attitudes toward fortune and misfortune have been influenced by those of our sister-culture on the other side of the pond.  Last week I was recounting for our publisher, Linda Rosas, my interview and subsequent e-mail communication with a woman whose life has so totally careened out of control in the past few years that she is now a desolate case, crying for help to the local newspaper. â€œWe all choose our path in life,â€ Linda said breezily, and as far as she was concerned, the subject was closed. I couldnâ€™t help thinking of the scene in Roman Polanskiâ€™s \"Tess,\" which was of course based on \"Tess of the dâ€™Urbervilles\" by eminent Victorian Thomas Hardy, in which some casual passerby remarks of Tessâ€™ misfortunes, â€œItâ€™s yer own fault.â€   There you go. W.H. Auden pointed out in one of his essays that itâ€™s no accident Catholic countries gave us almshouses, while Protestant countries gave us debtorâ€™s prison. Catholic culture is (or was, anyway) untouched by the influence of John Calvin, whose doctrine of predestination added up, in the countries where Protestantism triumphed, to a prevailing idea that you could be as selfish and self-centered as you pleased, while continuing to think that God was smiling on you. Since everyone was predestined, before the Creation, to be either damned or saved, it was a sign of Godâ€™s grace if you were well-off and prosperous in this world, and a sign of His disfavor if youâ€™re poor or down-and-out.   Itâ€™s a short leap from there to the idea that there is something shameful about being poor. There are countries in the world where begging is an honorable profession. But thatâ€™s not the case in England and it is certainly not the case in America. In our Anglo-American culture, even to be unemployed is a kind of blot on your character. I know, Iâ€™ve been there. The sight of the Dorrit family languishing in the Marshalsea, with Dorritâ€™s elder daughter, Fanny, forced to work as a dancer and at the same time hissing and spitting at anyone whom she perceives as casting aspersions on her family\'s status in society, is a sharp reminder of where all of this came from. Admittedly, I cannot relate to Fannyâ€™s, or her fatherâ€™s, anxiety over word getting out that any member of their family has been actually forced to work for a living, she as a dancer and Little Dorrit as a seamstress. We donâ€™t have quite that level of class snobbery here: America does have its rich class, as most countries do, but it doesnâ€™t have an idle rich class, as England once did, a landed gentry that considered it shameful to soil its hands with any kind of labor. But pride and hubris are themes that cut to the bone in American fiction just as they do in Dickens, both fictions growing from societies in which money success (America) or preserving oneâ€™s position in society (England) are the terms that delineate the good life. Small wonder Wolfe made the critics think of Dickens when he created Charlie Croker and company. We might not sing songs like \"Wait Till The Sun Shines, Nellie\" or \"Donâ€™t You Remember Sweet Alice?\" anymore, but Dickens and Wolfe were definitely mining the same vein.  '),(12121,3380464,'21,May,2004','2004-05-21 00:00:00','   I  Homing To The Stream : Ernest Hemingway and Cuba    Born and reared in the suburbs of Chicago, Ernest Hemingway was a product of the American heartland who, once he got out of it, never wanted to have much to do with the American heartland again. Aside from a few of his earliest short stories, which in any case are set in upper Michigan rather than suburban Illinois, Hemingway never published a word about where he came from, nor did he ever go back for long.   There are as many possible explanations for this as there are biographies of Hemingway. Some have suggested that Hemingway, like others of his generation who lived as expatriates in Europe after World War I, found the provincialism and narrow-mindedness of middle America stultifying. Unlike most of his fellow expats, however, Hemingway never went home. After leaving his birthplace of Oak Park, Illinois in the fall of 1921 with his new bride, Elizabeth Hadley Richardson, Hemingway spent the rest of his life either outside the United States or on the periphery of its life, in places like the Florida Keys, the upper plains states, (he enjoyed Montana and Wyoming for their hunting) the Pacific northwest, specifically Idaho, where he ended his life in 1961...and Cuba. Given the enormously complex relationship that Hemingway had with his family, particularly with his demanding and overbearing mother, whom he repeatedly and for years denounced as a \"bitch\" to anyone who would listen, it\'s possible that Hemingway simply wanted to give the midwest a \"wide berth\" because that was where his mother was until her own death in 1951.   Shortly before Ernest and Hadley were married in September, 1921, Hadley had said to Ernest, \"The world\'s a jail and we\'re gonna break it together.\" They were young--he was 22 on their wedding day; she was not quite 30--and deeply in love; in later years, after later marriages, Hemingway would suggest that Hadley was the only woman whom he had ever truly loved. But, although he was at heart a romantic, and conventional enough a midwesterner in his heart to feel that any time he fell in love with a woman, he ought to marry her, Hemingway couldn\'t be monogamous. His friend and fellow author Scott Fitzgerald (who managed to remain married to the same woman until he died, even though his wife, Zelda Sayre, was mentally unstable, Fitzgerald himself was an alcoholic and their marriage was a long, stormy melodrama) predicted quite early on that Hemingway was going to require \"a new wife for each book,\" a remark which turned out to be quite prescient, and the course of Hemingway\'s four marriages also took him to Cuba and lodged him there--he ultimately made Cuba the most permanent home he would ever make of any place.  Hemingway\'s trek south began in 1927, when he and Hadley divorced over his affair with wealthy Arkansas socialite and magazine journalist Pauline Pfeiffer, who promptly became the second Mrs. Hemingway. One thing Hemingway would never do with his wives, until his fourth and last one anyway, was to subject the new wife to living in the venue of the old. Ernest and Hadley\'s marriage had centered upon Paris, where he struggled in the early-to-mid 1920s to get his writing career started.   With the end of his first marriage, Hemingway\'s affection for Paris declined temporarily, and another friend and fellow novelist, John Dos Passos, suggested to him that Key West, Florida might be a good place for him to relocate.   In 1928, when Hemingway and Pauline went to live there in a house that her wealthy Uncle Gus had given them as a wedding present, Key West was a sleepy, run-down fishing village at the end of the Keys southwest of Miami. Prohibition was still in effect, and among Key West\'s attractions for the notoriously-bibulous Hemingway may well have been that its very position on the periphery of America made prohibition little more than a word there.  Smuggled liquor from Cuba was easy to get, and the island had plenty of its own sources of bathtub booze as well. Key West was popular with the artistic and literary crowd already; poet Wallace Stevens was among those who spent time there.  Pauline and Ernest set up housekeeping in Key West and Ernest, who already had one son from his marriage with Hadley, soon had two more.   Every tourist who\'s ever gone to Key West to take pictures of Hemingway\'s house there knows that Sloppy Joe\'s Bar was his favorite Key West hangout. He quickly made friends with its owner, Joe Russell, described by Hemingway biographer Michael Reynolds as \"a salty, red-faced bootlegger who was the real thing.\" Russell had a boat for fishing, the â€œAnita,â€ and he and Hemingway often went out on fishing expeditions in it, some of which lasted for days.   One such expedition took place in the spring of 1932. Pauline had given birth to Hemingway\'s third and final son the previous November. Hemingway wasn\'t happy about becoming a father again, and the birth put further strain on what was already a strained relationship--Pauline\'s family\'s money had been largely responsible for freeing Ernest from the necessity to write journalism for a living, enabling him to devote himself exclusively to fiction, and he was not a man to be comfortable with owing anyone anything.   Partly to get away from the strained relationship and partly to get away from the sound of a crying baby, Hemingway went off with Russell in April of that year on another fishing trip. They sailed to Havana for what was supposed to be a two-day stay. In the end, they stayed four months.   Hemingway had seen Havana once before, when he and Pauline had made a stop-off there on their way to Key West from France in 1928. But Havana had new attractions for him now, not the least of which was 22 year-old Jane Mason, the wife of the head of Pan American Airways in Cuba. They began an affair as Hemingway took up extended residence at what would be his base in Havana for the next several years, the Hotel Ambos Mundos. In the spring of the following year, as his second marriage continued to unravel, Hemingway was in Havana for another extended stay, and once again Jane Mason was in the picture. Pauline, sensing trouble, used the excuse of the annual visit of Hemingway\'s son  from his first marriage, Jack, (known as Bumby) to come over from Key West as she brought the boy to Havana to spend time with his father. But Hemingway made it plain that he didn\'t want her around, and she soon went back to Key West alone.   Hemingway and Mason talked of getting married, but it never happened. Instead, back in Key West in 1936, he met Martha Gellhorn, a vivacious and attractive blonde and fellow author/journalist with whom he would soon share a hotel room in Madrid while they were both covering the Spanish Civil War as correspondents, and who would become the third Mrs. Hemingway in 1940. Pauline was a devout Catholic who did not take divorce lightly, but after much wrangling, she finally agreed to divorce Ernest so he could marry Martha. The divorce became final in November, 1940, and Hemingway and Gellhorn were married that same month.   The house in Key West had been a gift from Pauline\'s family; obviously Ernest and Martha weren\'t going to live there, although he remained generally on good terms with Pauline and they did sometimes visit. The logical place for the newlyweds to settle after their Hawaiian honeymoon was the city that had virtually been Ernest\'s home away from home for about eight years already: Havana.   Aside from the fact that he had already established in Havana what\'s known in the world of bullfighting as a â€œquerenciaâ€â€”-the part of the ring where the bull feels at homeâ€”-there were other reasons for Hemingway to like Havana. During the years of the Grau and Batista regimes that preceded the rule of Fidel Castro, Havana was something of a playground for the wealthy of America\'s east coast (some have called it \"America\'s brothel.\") Havana was a good place to have a good time, and Hemingway liked a good time. His years in Key West had already given him a taste for tropical ambience. Naturally sloppy in his personal habits, (Martha Gellhorn casually nicknamed him \"Pig\") he liked a place where he could go around in shorts and loose-fitting shirts all the time. His outings aboard the â€œAnitaâ€ with Joe Russell had given Hemingway a passion for deep-sea fishing, and the Gulf Stream, adjacent to Cuba, offered the best marlin fishing in the world. Always fond of blood sports, Hemingway discovered and became an enthusiastic follower of cockfighting in Havana, and there was also the less-bloody, if for its human participants more dangerous, sport of Jai Alai, another amusement of which Hemingway became a devoted aficionado.   Martha had joined Hemingway in Havana the year before they were married, but she was not about to share a small, dirty room at the Hotel Ambos Mundos with him. Looking in a newspaper, she found a 15-acre estate about 15 miles from downtown Havana called the \"Finca Vigia\" (\"Lookout Farm.\") The place was badly dilapidated and Hemingway thought it not even worth trying to renovate, but Martha saw it as a challenge and hired workers at her own expense to get the place into liveable shape.   At first they rented the farm for $100 a month, but in December, 1940  after they were married, Hemingway bought the Finca Vigia for $18,500. It was the first home he had ever owned that hadn\'t been given to him by someone else, and he \"settled\" there to the extent that a man as restless as Ernest Hemingway could settle anywhere. The Finca provided a spacious, quiet place for him to work. It had a swimming pool and tennis court, and Hemingway\'s myriad cats and dogs wandered freely about the place. Soon he was a regular fixture at Havana\'s Floridita bar, where he could be seen downing ice-cold dacquiris in his own reserved seat at the end of the bar.   But the Hemingway-Gellhorn marriage was as doomed as his two previous ones. Indeed, Martha had expressed misgivings to a friend before her marriage to Ernest as to how  things would work out between them. Both of Hemingway\'s first two wives had been older than he was, she pointed out, while she was nine years his junior. Martha was less than enthusiastic about assuming the role of housekeeper, and Hemingway\'s mercurial temper, often exacerbated by drinking, increased tensions. Martha later decried her husband\'s \"ceaseless, crazy bullying\" of her. There is also some evidence, perhaps not surprising in view of the fact that Martha was the first of Ernest\'s wives to be younger than he, that they were sexually incompatible.   Then World War II came along and invaded what was already less than Eden. Hemingway had already been directly or indirectly involved in four wars: World War I; the 1922 war between Greece and Turkey; the Spanish Civil War and, most recently, the Japanese war in China, which he and Martha had gone off to cover as correspondents shortly after their marriage. He was inclined to sit this one out, although in 1943 German U-boats were slipping into the Gulf of Mexico to torpedo American and Venezuelan shipping, and Hemingway saw an opportunity to participate in war without having to stray too far from home. By now he had his own fishing boat, the â€œPilar,â€ and with the connivance of the American ambassador in Havana, Spruille Braden, Hemingway embarked on a slightly-crackpot \"U-boat hunting\" scheme--he had the â€œPilarâ€ outfitted with machine guns and ammunition, rounded up a crew and made a number of patrols in the Gulf Stream looking for German subs. (They never found one.) Beyond such Huck Finn shenanigans, Hemingway inclined toward staying home in Cuba as the war in Europe raged.   Martha had no such inclination. She was an ambitious writer who loved her work, and was not going to be content with sitting around the Finca Vigia being Mrs. Hemingway when the whole world was at war. She went off to cover the war as a correspondent for Collier\'s magazine, and Hemingway was eventually cajoled into doing likewise. In 1944 he went to London, his first stop toward becoming Colliers\' front-line correspondent after the Normandy invasion, at which he was present.   His marriage to Martha was already in trouble, and the war finished it offâ€”in London he met yet another woman journalist, Mary Welsh, who would become the fourth and last Mrs. Hemingway. In 1946 he brought her to the Finca Vigia, where she became the somewhat-uneasy but increasingly confident mistress of the place.  Mary  did, unlike her predecessor Martha, go to great lengths to make the Finca a \"home,\" although for the footloose Hemingway \"home\" seldom meant much more than a base for periods between arrivals and departures. Mary put in a great deal of work maintaining flower and vegetable gardens on the property and went to great pains with the upkeep of the decaying house while also having to work around her husband\'s less-than-fastidious personal habits and coexisting with a menagerie of cats, dogs and other animals on the property. Years earlier, Pauline Pfeiffer had tried to keep Hemingway close by giving him a decent home in Key West; Mary Welsh in the 1940s and \'50s made similar efforts in Cuba for a man now in his fifties and aging quickly, with a list of physical and mental ailments that would mount and mount over the years, making their life together on the island as stormy as the previous Hemingway marriages.   Hemingway would make Cuba his base of operations for the remaining years of his life, although, because he craved both audiences and excitement, he and Mary  would often leave their somewhat-remote tropical outpost for sojourns in the United States, Europe and Africa.   Then, in 1959, Fidel Castro came to power. For most Americans living in Havana, the Communist takeover amounted to a klaxon horn sounding Abandon Ship. But Hemingway chose to stay on. He had been fully aware of the corruption and abuses of the previous regimes and, naive about Communism, felt he had no reason not to accept Castro as a welcome reformer. He went so far as to wish Castro \"all luck\" with running the country, an attitude which didn\'t do much to endear Hemingway to either the U.S. government or to the FBI. In fact, FBI director J. Edgar Hoover had been \"keeping an eye\" on Hemingway ever since the sub-hunting adventure of the early forties, and Hemingway\'s mounting paranoia in his later years about FBI surveillance was not entirely without basis in fact. (Castro returned Hemingway\'s compliment, by the way--since Hemingway\'s death, the author has been a much-honored cultural hero in Cuba, the government maintaining his house as a museum and even having the â€œPilarâ€ transported overland to be the museum\'s chief attraction.)   People did sometimes ask Hemingway why he chose to live in Cuba. Late in 1948 he wrote an article for Holiday magazine in which he talked about his life there, offering his readers verbal snapshots of cool mornings at the Finca, cockfights and pigeon shoots, but most importantly the incomparable marlin fishing in the Gulf Stream, which he lovingly called \"the Great Blue River.\" It was not only in journalism, but in his often less-than-successful later fictions that Hemingway\'s attractions toward this part of the world are apparent. In his 1937 \"proletarian\" novel â€œTo Have And Have Notâ€ Hemingway displayed his affection for the seamy side of Havana life, and on one level his late novella â€œThe Old Man And The Seaâ€ could be read as a lyrical tribute to the Gulf Stream. His posthumous novel â€œIslands In The Stream,â€ cobbled together in 1970 from a huge pile of manuscript that he churned out between 1946 and 1950, while it may be unbearably talky in places, is in other places a moving tribute to the Caribbean world--and to Cuba. Hear Hemingway in that novel use his unmatched gift for evoking topography to share with his readers what was, unquestionably, a longstanding love-of-place:   â€œHe got into the car and told the chauffeur to go up O\'Reilly to the Floridita. Before the car circled the plaza in front of the embassy building and the Ayuntamiento and then turned into O\'Reilly he saw the size of the waves in the mouth of the harbor and the heavy rise and fall of the channel buoy. In the mouth of the harbor the sea was very wild and confused and clear green water was breaking over the rock at the base of the Morro, the tops of the seas blowing white in the sun.   It looks wonderful, he said to himself. It not only looks wonderful, it is wonderful.â€   II  Hemingway and Turgenev: A Brief Introduction  One has to be very careful taking Ernest Hemingway at his word. Fiction was his metier, and one of the dictionary definitions of fiction is \"something invented or imagined.\" The impulse to embellish was powerful in him. Hemingway seldom let facts get in the way of a good story, either on paper or while entertaining a crowd at lunch. Add mental breakdown, alcoholism, repeated head injuries and depression to that tendency, and you have some of the reasons why â€œA Moveable Feast,â€ Hemingway\'s memoir of his life in Paris with his first wife during the early 1920s, is undeniably a wonderful \"read,\" (I fell in love with this book, and with Hemingway\'s art, when I was 16) but it\'s also a book that has to be taken not just with any grain of salt, but with a grain of the kind of salt they use to de-ice freeways.   And not only did Hemingway love to embroider around facts, but like many self-educated men, he wore his erudition on his sleeve. We sometimes forget that before the end of World War II and the advent of the G.I. bill, a college education was not considered a universal entitlement in America, nor was it necessarily something that everyone felt they had to have.   Hemingway was offered the chance to go to college--his father had wanted him to attend Oberlin--but he decided he\'d rather go out and get educated in the business of life instead, first as an ambulance driver in the First World War, and then as a reporter on the Kansas City and Toronto Star newspapers. (In the early 20th Century a young man could still walk in off the street with a personal recommendation and begin working as a reporter on a big-city newspaper, but the age of the \"cub reporter\" is as long gone now as is the age of the volunteer ambulance driver.)   Hemingway educated himself in literature not by attending seminars and writing research papers, but by reading, and throughout his life he loved to list the authors he had read, or claimed to have read, for the commiseration of the reading public, although he always got very touchy when anyone dared whisper the word \"influence.\" He had sufficient perspective to generously acknowledge his admiration for the \"big guns\"â€”-his near-reverence for Shakespeare and Tolstoy was never alloyed with ironyâ€”-but when it came to lesser, or God forbid, contemporary influences, friends and associates had to step carefully around him. With his enormous ego and extreme competitiveness, Hemingway was not one to be generous to colleagues, and the surest way to get him to turn on you, as Gertrude Stein and Sherwood Anderson found out the hard way, was to help him out. Hemingway simply wasn\'t comfortable with owing anyone anything, and most certainly not with anything having to do with style. Perhaps that\'s not surprising in view of the way Hemingway\'s style co-opted his personality, (or perhaps it was the other way around.) In any case, identifying the influences on Hemingway has amounted to a full-time job for scholars and critics for decades.   Nevertheless, when he began assembling the sketches for â€œA Moveable Feastâ€ circa 1957, Hemingway\'s need to show the world the breadth of his youthful reading centered largely around the great Russians: Tolstoy, Dostoevski, Turgenev, and also Gogol and Chekhov. Hemingway stood in awe of Tolstoy and was fascinated by the emotional battering-rams in Dostoevski, but ultimately it was Turgenev, among the Russians anyway, who played a role in shaping Hemingway\'s famous style.  Tolstoy\'s impact on Hemingway is as hard to ignore as an iceberg--after all, no one has ever written on the subject of war with anything close to the sweep and depth of Tolstoy, and war was a subject that Hemingway found endlessly fascinating. Admiration for Tolstoy was a \"natural\" for him.   Dostoevski, too, deeply interested him, if only because he reports himself wondering aloud to the poet Evan Shipman how Dostoevski could \"writeâ€¦so unbelievably badly, and make you feel so deeply.\" In the Paris sketch in which he describes his first visit to Sylvia Beach\'s famous bookstore, Shakespeare and Company, Hemingway lists Dostoevski\'s â€œThe Gambler and Other Stories\" as one of the volumes he first borrowed, along with â€œWar and Peace,â€ some D.H. Lawrence and â€œA Sportsman\'s Sketchesâ€ by Ivan Turgenev.   Another name might be mentioned in this context, that of Henry James. Hemingway\'s admiration for Henry James was equivocal; according to \"A Moveable Feast,\" James was the favorite author of Hemingway\'s first wife, Hadley, and Hemingway had sufficient taste and intelligence, self-educated or not, to acknowledge James as a classic. But James\' highly stylized, mannered prose was worlds away from what Hemingway would soon be trying to do with short sentences, and James\' world of elegant drawing rooms, English lawns and upper-class parlors was continents away from Hemingway\'s world of battlegrounds, bullrings and African hills. When it came to subject matter, Joseph Conrad was more to Hemingway\'s taste. But perhaps most importantly, when it came to style, James was a practitioner, in fact the very epitome, of the late 19th-century\'s idea of good writing: the highly-polished \"polite prose\" that Hemingway decided he had to get  away from.   So, ironically, was Turgenev, whom Hemingway embraced as a master.   And indeed, Turgenev and James are similar writers in some ways: although Turgenev wrote in Russian and James in English, both men spent much of their adult lives living outside their country of birth, and both wrote in an elegant, consciously literary style that doesn\'t appeal to everyone. Oscar Wilde once remarked that James wrote prose \"as if it were a painful duty,\" and to this day it is fashionable among bookish Russians to deride Turgenev as \"feminine\" owing chiefly to the elegance and grace of what many regard as overpolished prose. And because I have heard Turgenev accused of being \"feminine\" more than once, I find it ironic that he should have influenced the machismo--obsessed Hemingway, with his laconic tough-guy characters who speak in tight-lipped bursts.   And yet, if you take a look at that very book of Turgenev\'s that Hemingway reports having borrowed from Sylvia Beach in 1922, â€œA Sportsman\'s Sketchesâ€, and then glance through Hemingway\'s first widely-read book of short fiction, â€œIn Our Time,â€ a few of Turgenev\'s fingerprints do indeed seem to appear on Hemingway\'s pages. The most powerful influences on the style that Hemingway evolved in the early 1920s may well have been, as has been often pointed out, the Kansas City Star\'s stylebook, with its emphasis on keeping things short, clipped and free of adjectives, and Gertrude Stein, whose mammoth novel â€œThe Making of Americansâ€ Hemingway helped to type and whose \"cubistic\" approach to prose, repeating phrases over and over with slightly different wordings, doubtlessly influenced Hemingway\'s sense of prose rhythm. But Turgenev, too, is part of the mix, and Hemingway certainly never made his admiration for Turgenev a secret. In 1925 he told Archibald MacLeish that he thought Turgenev \"the greatest writer there ever was,\" and then added \"â€™War and Peaceâ€™ is the best book I know, but imagine what a book it would have been if Turgenev had written it.\"  Tolstoy had dazzled Hemingway by writing dazzlingly about war. Joseph Conrad, like Herman Melville before him, had been a sailor, and many of his tales center around the adventurous and--at least in those days, manly--world of the sea. Turgenev, in â€œA Sportsman\'s Sketches,â€ creates a worldâ€”-albeit a peculiarly Russian one-â€”around a man who spends all of his time indulging his greatest passion, which also happened to be one of Hemingway\'s: hunting. I think that, with regard to Turgenev, this was the hook that drew Hemingway in. But once he was there, within Turgenev\'s world, Hemingway took something else away with him besides the enjoyment of a string of vignettes about a man out hunting and communing with his land and people along the way.  â€œA Sportsman\'s Sketchesâ€ was an unusual book for the middle of the 19th century. Perhaps it could have been written nowhere except Russia, where there wasn\'t much in the way of a literary \"tradition\" circa 1850. At the same time that Dickens, Thackaray and their contemporaries in England were building on the foundations laid by the great 18th-century English novelists such as Richardson and Fielding, and writers in France like Hugo and Flaubert were doing the same with the traditions of Rousseau, Chateaubriand and Stendhal, Russian writers were still trying to create a tradition for themselves. Russian literature is often said to have begun with Pushkin and Gogol, both of whom flourished in the 1830s. Mikhail Lermontov\'s highly idiosyncratic novel â€œA Hero of Our Timeâ€ had appeared in 1840, when Pushkin had only been dead three years. Lermontov himself was killed in a duel, (the same fate that had befallen Pushkin) just a year after his novel came out. Russian authors of the 1840s and 50s may have had to contend with the tyranny of the czar\'s government, but they were working relatively free of the tyranny of a literary tradition.   Turgenev didn\'t bother writing \"stories\" in the accepted sense of the word when he wrote â€œA Sportsman\'s Sketchesâ€ (I\'m using the title Hemingway knew; in more recent translations the book has been called â€œA Sportsman\'s Notebook.â€) There is no conventional \"plot\" or schematic narrative anywhere to be seen. The stories are snapshots of places, characters, situations. Much of Hemingway\'s â€œIn Our Time,â€ a product of \"modernism\" written not long after he had been reading Turgenev in Paris, works in much the same way. It was an uncommon technique for a writer of the mid-19th century to be using, to understate rather than overstate, use simple observation rather than rhetoric, and as often as not, to end on a \"dying fall\" rather than a crashing chord, as at the end of the sketch called \"Raspberry Water:\"   \"Styopushka started up. The peasant sat down beside us. We fell silent again. On the other bank someone started singing, but such a melancholy songâ€¦My poor friend Vlas grew sadder and sadder. Half an hour later, we parted.\"   Now hear Hemingway at the end of \"The Battler:\"   \"Nick climbed the embankment and started up the track. He found he had a ham sandwich in his hand and put it in his pocket. Looking back from the mounting grade before the track curved into the hills he could see the firelight in the clearing.\"   The dying fall, the laconic understatement. Two of Hemingway\'s trademarks. And yet as I re-read â€œA Sportsman\'s Sketchesâ€ a few years ago, I found example after example of the same sort of thing in Turgenev that later made Hemingway\'s reputation. Listen to this bit of dialogue from \"Bezhin Meadow:\"   \"Well, Vanya,\" began Fedya tenderly, \"is your sister Anyutka well?\" \"Very well,\" answered Vanya, slightly slurring the \"r.\" \"Tell her to come and see us. Why doesn\'t she come?\" \"I don\'t know.\" \"Tell her to come.\" \"I will.\" \"Tell her that I\'ll give her a present.\" \"And me too?\" \"Yes, you too.\"   Take out just one word, the adverb \"tenderly,\" and that snatch of dialogue could be Hemingway. Or, perhaps it might be more appropriate to say, the following exchange from The Sun Also Rises could be Turgenev:   \"What\'s the matter with the old one?\" I asked. \"He hasn\'t got any passport.\" I offered the guard a cigarette. He took it and thanked me. \"What will he do?\" I asked. The guard spat in the dust. \"Oh, he\'ll just wade across the stream.\" \"Do you have much smuggling?\" \"Oh,\" he said, \"they go through.\"   Yes, between Turgenev and Hemingway came Chekhov, another master of emotional tautness and the restrained phrase. And Hemingway greatly admired Chekhov too. But there can be little doubt that Turgenev\'s evocation, a generation before Chekhov, of the lovely, vast and brutal Russian countryside and its characters, in prose all the more effective for its restraint, gave Hemingway the cue for some of his own efforts, three-quarters of a century later, to evoke his own landscapes and characters, lovely and brutal, in prose all the more effective for its restraint.  III  Dubious Battles: Ernest Hemingwayâ€™s Journeys to War  In the movie â€œPatton,â€ George C. Scott, playing the role of the legendary-infamous World War II general, surveys a column of tanks advancing over rough terrain and says exultantly, \"Compared to war, all other forms of human endeavor shrink to insignificance!\"   That statement might have been taken directly from the mouth of Ernest Hemingway. While it would be wrong to say that Hemingway approved of war as a thing-in-itself the way Patton seems to have, (and by the way, Hemingway spent part of World War II with one of Patton\'s armored divisions and thought little enough of the general to have one of his later fictional characters all but call Patton a pathological liar) the fact remains that violent death was one of Hemingway\'s lifelong fascinations, and if contemplating violent death is what you\'re after, war is as good as it gets.   Throughout Hemingway\'s life, timing was to be a key factor in what happened to him and in how he reacted. Hemingway\'s timing was always very good, whether with regard to external events or to his own endeavors first at becoming a writer and then at promulgating his career as one. With regard to timing, war was one of the areas in which history accomodated Hemingway-his lifetime spanned four major wars, three of which he saw close-up, though never as a soldier, and he often embroidered his war experiences for presentation to the home audience to make them look as \"soldierly\" as he possibly could. In his youth this would take the form of getting himself seriously wounded and then playing a form of \"dress-up;\" on a later occasion the play-acting almost got out of hand. But young or old, \"playing soldier,\" whether in real-life or vicariously, through his fictional characters, was something Hemingway found nearly irresistible.   When he was young, it was part and parcel of his natural desire to be where the action was. Contemporaries who worked with the teenage Hemingway during his days as a cub reporter on the Kansas City Star reported that he would dash around the city compulsively, wanting always to know where the ambulance went, where the crime had occured, and to get there quickly.   His tenure at the Star was to be a short one for precisely this reason. He had graduated from high school in 1917 and, turning down the chance to go to college, proceeded directly into newspaper work after an uncle pulled some strings and helped him get the job in Kansas City.   But 1917 was also the year that America entered World War I. Hemingway, like many young American males at the time, heard the stories coming back from Europe and was determined to get into the war somehow. He would have liked to enlist as a soldier, but his father was opposed to that idea and in any case Hemingway\'s famously poor vision in his left eye probably would have gotten him a 4-F. But while working on the Star, he struck up a friendship with 22 year-old Ted Brumback, who just the previous summer had enlisted in the American Field Service and spent four months in France as an ambulance driver. Since it was unlikely that he would get into the war as a combatant, Hemingway decided to go this route, and after persuading his father to drop his objections, he, Brumback and another friend, Wilson Hicks, signed on with the Red Cross and by the following spring were on their way to the war. They ultimately served in northern Italy, not far from Milan.   The world knows the rest of this story. On the night of July 8, 1918, while he was passing out chocolate bars, cigarettes and magazines to Italian soldiers on the Piave River near Fossalta, Hemingway\'s dream of being where the action was came true in nightmarish fashion. The Austrians launched a mortar attack and Hemingway was badly wounded by shrapnel. His later recuperation in a Red Cross hospital would include a brief romance with a nurse several years his senior, Agnes von Kurowski. This romance, in turn, would result a decade later in his novel \"A Farewell to Arms,\" and many decades later, in 1996, in the movie \"In Love and War,\" although to the end of her life Kurowski kept denying that it ever happened.   Hemingway\'s natural flair for self-dramatization never had a better opportunity than this one. The relative gullibility of Americans on the home front during the First World War gave him a clear playing field for his love of pose. It was after all an era in which there was no CNN to bring the horrors of war right into everyone\'s living room. Even newsreels were in the future. The war was an ocean away, and the domestic attitude toward it was shaped by patriotic songs, jingoistic slogans and romantic stories of faraway Europe. Nineteen year-old Ernest Hemingway wasted no time in creating the image for the folks back home of himself as Warrior. He wrote extravagant letters claiming, among other things, that his wounds had been caused by machine-gun bullets as well as shrapnel. (This is still being argued.) He got himself all dolled up in a tailor-made uniform and had his picture taken wearing military insignia that he was not entitled to, and when he returned to Oak Park, Illinois the following winter, hobbling on a cane, his romantic tales of war were made more plausible by the soldier\'s cap and boots that he wore, not to mention the custom-made Italian cape.   Hemingway\'s next trip to war was a short one. In 1922 he was Paris correspondent for the Toronto Star and that newspaper sent him to Constantinople to cover the war between Greece and Turkey. He was there only for a few weeks, but the result was two striking pages of fiction, \"On The Quai at Smyrna,\" the \"opening note\" in his first book of short stories, \"In Our Time.\" He would not see war again for 14 years.   To outline the byzantine political and military machinations that brought about the Spanish Civil War (1936-38) would be beyond my purposes here, not to mention my space. Although this is a gross oversimplification, the war between those who supported the left-wing Popular Front in Spain and those who supported the right-wing National Front is often seen as having been a \"dress rehearsal\" for World War II, with Hitler backing one side and Stalin backing the other. Stalin\'s side lost the actual war, but won the propaganda war, not least because it was supported by a solid international phalanx of writers and intellectuals, not the least of whom was Ernest Hemingway.   In 1937-38, Hemingway went to Spain four times, the first three as a reporter for the North American Newspaper Alliance (NANA.) By now an international celebrity with a hairy-chested image to maintain, he sent back reports from Madrid and from the countryside filled with details suggesting that he was constantly in danger from artillery bombardments as well as machine-gun and automatic rifle-fire. He was indeed in harm\'s way his share of times, and the evidence suggests that he always kept his composure in such situations. But as usual, his embroidering is never far away. He boasted that he was even in danger when he went to bed in his room at the Hotel Florida in Madrid, because as he explained to his readers, his room was on the front side of the building, facing the Nationalist guns two miles away. Here we find two of Hemingway\'s biographers offering different stories. In his groundbreaking 1969 biography, Carlos Baker tells us that Hemingway\'s roommate at the Hotel Florida was his old friend the American bullfighter Sidney Franklin. Franklin was indeed there, but Kenneth S. Lynn, in another biography published 17 years after Baker\'s, reveals that Hemingway\'s roommate at the Florida was actually his fellow journalist and soon-to-be third wife Martha Gellhorn, adding that he probably would not have been living there under those circumstances if it were as dangerous as he claimed.   Hemingway\'s involvement in the Spanish Civil War ended certain of his friendships, notably with fellow novelist John Dos Passos. Dos Passos was among those intellectuals who were initially sympathetic to the Loyalist cause, (another was George Orwell) but who, upon observing that the Communists and the anarchists were murdering people with no less alacrity than the fascists, became disillusioned with the cause (and in Dos Passos\' case, with left-wing politics generally.) The politically-naive Hemingway, whose pride moreover dictated that he always seem to have more and better information than anyone else, refused to acknowledge the atrocities being commited on the Loyalist side and reported the war in a one-dimensional, one-sided manner. (Happy to the point of lionizing Hemingway at first, the political left would eventually turn against him when he recovered enough equanimity to write \"For Whom the Bell Tolls,\" in which he not only maligned such loyalist icons as Andre Marty, commander of the International Brigades, but also, in the book\'s \"massacre\" chapter, committed the heresy of depicting fascists as victims.)   But if his first two major wars had served as canvases on which Hemingway drew sketches of what he wanted his own legend to look like, his third and last war was to be the biggest such, and this time he pulled out all the stops as promoter of his own image, to the point where he came very close to getting into serious trouble.   Shortly before D-Day in 1944, Hemingway arrived in London as a correspondent for Collier\'s magazine. He was to cover the invasion of Normandy, and he accompanied a number of armored and infantry battalions as they moved across France.   He was present on D-Day, was in fact aboard an LCVP landing craft as it advanced on Omaha Beach. Having dropped off its troops, the LCVP then returned to its ship with Hemingway still sitting astern. But what Hemingway subsequently wrote not only implied that he had gone ashore with the troops, but that he had played a vital role in helping to locate the beach. The Hemingway legend being what it was by then, few questioned his assertions.   And he was only warming up. Later came his claim to have entered Paris with the first troops, which led to the still-persisting myth that he personally liberated the Ritz Hotel. (Neither is true.)   And in fact this time Hemingway\'s military play-acting caught up with him. Other reporters resented his flamboyant showboating, which included travelling around to all appearances in command of a group of French Resistance irregulars and keeping firearms, bazookas, grenades and other ordnance in his Rambouillet hotel room. A complaint was filed that his activities were violating the Geneva Convention\'s rules regarding the conduct of news correspondents. In what must have been a particularly galling moment for him, Hemingway, at the risk of losing his reporter\'s credentials and being expelled from France, had to appear before a military panel and deny ever having taken part in any actual fighting himself. He offered deft excuses for some of the other things he was accused of, claiming for instance that weapons and ammunition had been kept in his hotel room only because storage space was in short supply.   But such was the Hemingway myth by 1944 that few concerned themselves with such niggling details. Hemingway had done such a good job for so many years as his own press agent, inflating his public persona to gargantuan proportions, that much of the world was prepared to accept his tall tales of war at face value. What few realized at the time was that Hemingway\'s real war, the one between his desire to go on living and his deep-seated desire not to, was even then being lost. Much of what passed for combat- zone bravery on Hemingway\'s part was in fact suicidal behavior, a point he would drive home at the end of a shotgun barrel 17 years later.  IV  The Death of Max Perkins and the Decline of Hemingway  On July 21, 1940, Ernest Hemingway turned 41 years old. He was just finishing â€œFor Whom the Bell Tolls,â€ which was published to great acclaim a few months later. In fact by the time Max Perkins, who was both Hemingway\'s and Scott Fitzgerald\'s editor at Scribner\'s, had returned from Fitzgerald\'s funeral late that year, â€œFor Whom the Bell Tollsâ€ had already sold 189,000 copies. Three days after the book was published, a deal for nearly $150,000 was signed with Paramount for the film rights. The film was duly made in 1943, with Gary Cooper in the role of Robert Jordan and Ingrid Bergman, just a year off her appearance opposite Humphrey Bogart in â€œCasablanca,â€ as Maria.   In terms of his success as a writer, 1940 was in some ways the apotheosis of Ernest Hemingway. Critics and the reading public, who had been tracking and for the most part admiring his work since the publication of his first novel, â€œThe Sun Also Rises,â€ in 1926, felt they had every reason to expect only further triumphs from the man who had become by then the most famous, and in the eyes of many, the greatest living American writer.   It didn\'t happen. In fact, in many ways and for a stew of reasons mixing everything from married life to geopolitics, the decade of the 1940s became Hemingway\'s version of the Lost Weekend. Though busy with a number of projects which included not only covering the war in Europe as a journalist, but also work on a massive, unfinished novel to be called â€œThe Garden of Edenâ€ (published in a heavily-edited edition in 1986) Hemingway produced no significant work of fiction between 1940 and 1950, and when he finally did, he might well have wished he hadn\'t.   As the decade began he was living in Cuba with his third wife, celebrated journalist Martha Gellhorn. However, not content to stay home in Cuba and be Mrs. Ernest Hemingway with a war raging in Europe, Gellhorn went off to cover the war as a correspondent for Colliers magazine, and Hemingway somewhat-reluctantly followed suit after a Huck Finn-like stint of hunting for German submarines in the Gulf Stream aboard his fishing boat, the \"Pilar.\" Hemingway had been an ambulance driver in World War I and a news correspondent in the Spanish Civil War. World War II would be the third and last war he would witness (and once again, ironically, Hemingway, who was so fascinated with war and all of its trappings, would participate only as a non-combatant.)   Hemingway\'s marriage to Gellhorn became one of the war\'s casualties. By the spring of 1946 he was married to his fourth and final wife, journalist Mary Welsh.   With the war over, Hemingway went back to Cuba and resumed the life of writing, boozing and marlin-fishing that it had interrupted. Without a doubt he felt that he had earned the right to \"kick back\" and enjoy himself for while after working hard in Europe as a war correspondent, and certainly he returned to Cuba more famous than ever. That was part of the problem. From the late \'40s onward, Hemingway and the media \"worked\" each other increasingly, and energies that he should have devoted to writing were devoted instead to \"playing the role of Ernest Hemingway\" for a cavalcade of reporters and photographers.   Two separate, highly significant and not necessarily unrelated catastrophes struck Hemingway in the years immediately following World War II. The first was the death in 1947 of his beloved and trusted editor Max Perkins. The second, three years later, was the publication of his first full-length novel since â€œFor Whom The Bell Tolls,â€ the much-panned â€œAcross The River And Into The Trees.â€ Shortly after the publication of \"Across The River\" in 1950, E.B. White, in the New Yorker magazine, published a satire of it called \"Across The Street and into the Grill.\" I don\'t know how funny I would find it now, but when I read it at age 19 I nearly suffocated with laughter. (When told of White\'s satire, Hemingway growled that the next step up from writing parodies was scribbling on latrine walls, conveniently forgetting that one of his own early efforts, \"The Torrents Of Spring,\" had been a savage parody of Hemingway\'s mentor, Sherwood Anderson.)   Max Perkins had been Hemingway\'s editor since 1929, when he had worked with Hemingway on A Farewell to Arms. Hemingway\'s father had committed suicide not long before Hemingway first met Perkins, and from the earliest days of their association, Perkins took on something of a fatherly role with Hemingway, much the same role he assumed about the same time for an ungainly young man from North Carolina named Thomas Wolfe, who had recently dumped an avalanche of manuscript on Perkins\' desk which Perkins would eventually hew and carve into â€œLook Homeward, Angel.â€   Over the nearly 20 years of their association, Perkins acquired the reputation of being the editor at Scribner\'s best qualified to handle the man who was well known as the touchiest author in the firm\'s stable. Perkins knew how to deal diplomatically with the vain, moody and mercurial Hemingway, could guage when it was \"safe\" to offer him honest criticisms of his work and when it was a better idea to soft-pedal. It was with Perkins\' editorial \"ruddering\" that Hemingway brought out all of his fiction from â€œA Farewell to Arms\" to â€œFor Whom the Bell Tolls,â€ a decade which also included Death in the Afternoon, â€œGreen Hills of Africa,â€ â€œTo Have and Have Not,â€ the short-story collections â€œMen Without Womenâ€ and â€œWinner Take Nothingâ€ and a play about the Spanish Civil War, â€œThe Fifth Column.â€   The 1930s had been a busy and productive time for both men. But when Perkins died, Hemingway had not published a novel in seven years.   For many who pursue it, writing is very much like playing a musical instrument. You have to stay in practice or you get rusty. Hemingway loved to apply sports metaphors to writing--his remarks about having \"outboxed\" de Maupassant, Stendhal and Turgenev have been repeated ad nauseam-and indeed he seemed to have been one of those writers who needs to stay \"in training\" to keep his writing sharp, not surprising perhaps in view of the heavy emphasis Hemingway\'s writing always did put on pure sensory experience--seeing, smelling, tasting, touching and hearing. (Although he drank a lot throughout his life, Hemingway never smoked again after his early twenties-he didn\'t like the way it dulled his sense of smell.)   With all of this in mind, Perkins\' death couldn\'t have come at a worse time for Hemingway. He was physically out of shape, overweight and soaked in liquor, (though he did often cut back on his drinking for one reason or another) and his body, including his head, was battered and tired from a long history of injuries and accidents. Journalism had taken up a lot of his time during the war years, and despite continued work on such projects as the ultimately-tabled â€œThe Garden of Eden,â€ Hemingway hadn\'t publicly strutted his fictional talents in a long time.   It was in this atmosphere that Hemingway decided to make his \"comeback.\" Perkins was gone, and between being both out-of-shape and increasingly worried that after all these years he might not be able to perform like he used to, (and hence, disappoint a world that, even without People magazine or eight-second sound bites, had turned him into a superstar) Hemingway was in no condition to be his own best editor.   And so, in 1949-50, Hemingway labored away on his new tale, set in Venice, through a haze of booze shot through with the piercing stage lights of global adulation. Caught up in the work, he managed to persuade himself that he was writing a masterpiece, his best book yet. And there was no Max Perkins around any more to gently rein him in and persuade him to cut here, condense there.   The result was an unqualified disaster, and what underlay the book\'s plot had much to do with it. A hopeless romantic despite his tough-guy posturing, Hemingway throughout his life was constantly falling in and out of love with one woman after another. He married four of them, but there were plenty of other \"infatuations\" along the way, and the older he got, the younger the infatuations got.   In the late 1940s, the overweight, booze-bloated, celebrity-inflated Hemingway, pushing 50, began mooning over a beautiful young Italian woman, Adriana Ivancich. Hemingway\'s romantic stories often had their foundations in his own fantasising, and the character Renata in â€œAcross the River and Into the Treesâ€ is based on Ivancich; Renata\'s doomed love affair in Venice with a dying American army colonel, Richard Cantwell, plays out Hemingway\'s fantasy about himself and Adriana in much the same way that, 20 years earlier, the fate of Frederic Henry and Catherine Barkley in \"A Farewell to Arms\" had played out his wishful embroidering upon his disappointing relationship with 26 year-old hospital nurse Agnes von Kurowski.    It is not my purpose here to write a review of a book that was published more than 50 years ago, but let me just mention the point at which â€œAcross the River and Into the Treesâ€ starts to \"go wrong\" for me.   The book starts out well enough. The duck-shooting scene in Ch. 1 has all of the old Hemingway magic. You hear that ice crack. You see those boats moving in the canal. Cantwell is thirsty; you\'re thirsty.   It\'s at the beginning of Ch. 2 that I detect the first warning signs of Ernest Miller Hemingway fading into the wings and the huffing and puffing \"Papa\" taking center stage. Twice on the same page he refers to Cantwell not as an \"army colonel\" or a \"colonel in the army\" as anyone else might, but as a \"Colonel of Infantry in the Army of the United States.\" This is affectation, and the older Hemingway got, the more he swam in it. And from that point in the book, things just get windily worse.   In his controversial 1986 \"psychobiography\" of Hemingway, Kenneth Lynn raised the enticing question (buttressed by like-minded speculation from Canadian critic Northrop Frye) that had Perkins not died when he did, he might have tried to talk Hemingway into distilling this material into a short story. Frye pointed out, and Lynn agrees, that there is much good in the story, and had Hemingway boiled it down to say, 100 pages, it might have been his version of â€œDeath in Venice,â€ with all the lyricism and dark power of Thomas Mann\'s 1911 mini-masterpiece.   One can speculate forever on such \"fantasy baseball\" issues, but the fact is that Hemingway\'s self-critical judgment had not deserted him entirely despite his catastrophic \"bad call\" on â€œAcross the River.â€ Had Perkins still been around, it might indeed have been a better book. Surely Hemingway\'s ability to cut, compress and keep it short, always one of his strengths, was still there. Having been knocked to the mat (to use a metaphor he would have liked) by the critics over â€œAcross the River,â€ he got back to his feet and two years later issued the short work that would clinch the Nobel Prize for him. Despite what you might think of â€œThe Old Man and the Sea,â€ (and I have little use for its gooey sentimentality myself) it showed that Hemingway could still control his material if he wanted to. He would lose that ability within a few years as his physical and mental problems mounted. But it is tantalizing to wonder about what the quality of the work of Hemingway\'s final years might have looked like had not his good friend and most trusted editor died when Hemingway, to judge from what he subsequently produced, was just about to need him the most.    V  Papaâ€™s â€œGood Eats CafÃ©:â€ Hemingway and Food   One of the first things you realize when you read Hemingway, and this was in fact one of the boldest brushstrokes in the portrait of himself which he presented to the world, is that the man loved to eat and drink. His Rabelaisian appetite for life included, (and quite logically so) a concomitant appetite for the pleasures of the table (not to mention the bar.)   In fact, it has occurred to me more than once that had Hemingway not taken up the writing of fiction as his life\'s work, he might have had a brilliant career as a restaurant critic for the slicker magazines, a sort of foul-mouthed Duncan Hines.   I\'ve mentioned in other places the fact that the self-educated Hemingway liked to boast about the breadth of his reading and his knowledge of books; one of Hemingway\'s best-known affectations was in fact his claim to have \"the inside dope\" on just about everything, and that went for food and wine as well as books, baseball, boxing, battles, bullfighters and trout-fishing streams.   Yes, Hemingway loved to eat and drink. And of all the claims he made about himself which, over the years, various people have questioned, none succumbs to the pin-prick of evidence more quickly than his claim, late in life, that he and his first wife were going romantically hungry in Paris during the 1920s. True, they were living simply and economizing so that they could afford to go on ski trips and such, but starving? I don\'t want to be catty about this, but take a look at a photograph of Hemingway taken on his wedding day in 1921, and then glance at a photo of him taken in 1924 in the courtyard of 113 rue Notre Dame des Champs in Paris, where he and his wife Hadley were living at the time. After less than three years in Paris, the slender bridegroom of 1921 is already getting a bit jowly, and that suit he\'s wearing looks tight on him. If \"hunger was good discipline,\" as he boasted in his late-life memoir â€œA Moveable Feast,â€ Hemingway was getting his discipline somewhere else. I won\'t stoop so low as to point out the irony in the book\'s title. (In fairness to Hemingway, I\'ve been to Paris, and you can\'t not eat there. Those travel brochures which tout Paris as the world\'s capital of good food are only stating an objective fact: it is that.)   I confess that I once ate something pretty disgusting, if not exactly on Hemingway\'s recommendation, then at least because he made it sound so good. No, it wasn\'t a thousand-legger, but in the terms that any gourmet would understand, it might as well have been.  I was 18 and had already been under Hemingway\'s spell for a couple of years. One fine Saturday afternoon I was re-reading one of my favorites among his early stories, â€œBig Two Hearted River, Part I.â€ I read through the following passage. (See if this doesn\'t make you hungry too, even though we\'re clearly not talking here about anything you\'d be served at Michaud\'s in Paris):   â€œNick was hungry. He did not believe he had ever been hungrier. He opened and emptied a can of pork and beans and a can of spaghetti into the frying pan. \"I\'ve got a right to eat this kind of stuff, if I\'m willing to carry it,\" Nick said. His voice sounded strange in the darkening woods. He did not speak again.â€   Then a little further on:   â€œThe beans and spaghetti warmed. Nick stirred them and mixed them together. They began to bubble, making little bubbles that rose with difficulty to the surface. There was a good smell. Nick got out a bottle of tomato catchup and cut four slices of bread. The little bubbles were coming faster now. Nick sat down beside the fire and lifted the frying pan off. He poured about half the contents out into the tin plate. It spread slowly on the plate. Nick knew it was too hot. He poured on some tomato catchupâ€¦He ate the whole plateful before he remembered the bread. Nick finished the second plateful with the bread, mopping the plate shinyâ€¦While he waited for the coffee to boil, he opened a can of apricots. He liked to open cans. He emptied the can of apricots out into a tin cup. While he watched the coffee on the fire, he drank the juice syrup of the apricotsâ€¦â€   By now, depending on the sophistication of your palette, you\'re probably either salivating or getting queasy.   But after reading this vivid passage, I just had to find out what this mess actually tasted like. I walked over to Safeway and bought a can of Chef Boy-Ar-Dee spaghetti, a can of Van Camp\'s pork and beans, and a can of Del Monte apricots. Then, grossing my mother right out of the kitchen, I cooked it all up, sliced some bread and dumped catchup over it, just like in the story.   You know, it actually wasn\'t bad, and sipping apricot juice right out of the can while the coffee was perking was also kind of pleasant.   Give me a break. I was 18.   Hemingway wrote \"Big Two-Hearted River\" in 1923, when he was on the verge of his first triumphs. And although his powers waned with the passing years, the victims of too many injuries and too much booze, not to mention too much fame and adoration, more than 30 years later he had lost none of his touch for making his readers hungry. Check out this mouth-watering passage from â€œA Moveable Feastâ€ in which Hemingway walks into Lipp\'s, a bistro on the left bank in Paris, and orders himself a fine lunch:  â€œâ€¦I sat down on the bench against the wall with the mirror in back and a table in front and the waiter asked if I wanted beer and I asked for a distingue, the big glass mug that held a liter, and for potato salad.   The beer was very cold and wonderful to drink. The pommes a l\'huile were firm and marinated and the olive oil delicious. I ground black pepper over the potatoes and moistened the bread in the olive oil. After the first heavy draft of beer I ate and drank very slowly. When the pommes a l\'huile were gone I ordered another serving and a cervelas. This was a sausage like a heavy, wide frankfurter split in two and covered with a special mustard sauce.  I mopped up all the oil and all of the sauce with bread and drank the beer slowly until it began to lose its coldness and then I finished it and ordered a demi and watched it drawn. It seemed colder than the distingue and I drank half of it.â€   Are you hungry yet? I had to go to the kitchen while I was typing that out and get myself a snack. And by the way, Lipp\'s, like Pamplona, is one of those places that Hemingway \"put on the map.\" Some years ago I was in Paris and thought I\'d drop in at Lipp\'s to sample some hemingwayesque ambience. No way--the place was wall-to-wall tourists, all doing exactly what I was doing. I ended up having my lunch at a tiny Greek grill (not there in Hemingway\'s time, I\'m sure) on a tiny side-street across from the Ile de la Cite and Notre Dame.   Or try this meal, from an earlier sketch in the same collection:   â€œI asked the waiter for a dozen Portugaises and a half-carafe of the dry white wine they had thereâ€¦As I ate the oysters with their strong taste of the sea and their faint metallic taste that the cold white wine washed away, leaving only the sea taste and the succulent texture, and as I drank their cold liquid from each shell and washed it down with the crisp taste of the wine, I...began to be happy and to make plans.â€   Yes, after a meal like that, I probably would too.   As the short-story example cited above makes clear, Hemingway didn\'t confine his culinary passages to memoirs. I always loved this little scene from â€œThe Sun Also Risesâ€ in which his two male characters, who have left a couple of bottles of wine in an ice-cold stream to chill while they go off trout-fishing, sit down to enjoy their picnic lunch and, using the news of the recent death of William Jennings Bryan as the springboard for a bit of typical snappy jazz-age repartee, indulge in a little creative playing-with-their-food:   â€œThe wine was icy cold and tasted faintly rusty. \"That\'s not such a filthy wine,\" Bill said. \"The cold helps it,\" I said. We unwrapped the little parcels of lunch. \"Chicken.\" \"There\'s hard-boiled eggs.\" \"Find any salt?\" \"First the egg,\" said Bill. \"Then the chicken. Even Bryan could see that.\" \"He\'s dead. I read it in the paper yesterday.\" \"No. Not really?\" \"Yes. Bryan\'s dead.\" Bill laid down the egg he was peeling. \"Gentlemen,\" he said, and unwrapped a drumstick from a piece of newspaper, \"I reverse the order. For Bryan\'s sake. As a tribute to the Great Commoner. First the chicken, then the egg.\" \"Wonder what day God created the chicken?\" \"Oh,\" said Bill, sucking the drumstick, \"how should we know?\"   In the much-criticized â€œAcross the River and Into the Treesâ€ (1950) Hemingway seems to spend half of what\'s purportedly a tragic love story going on and on about what his characters are sipping or digging into:   \"Two very dry martinis,\" the colonel said. \"Montgomerys. Fifteen to one.\"â€¦\"The Martinis were icy cold and true Montgomerys, and after touching the edges, they felt them glow happily all through their upper bodies.\" â€¦  \"Just then the lobster was served.   It was tender, with the peculiar slippery grace of that kicking muscle which is the tail, and the claws were excellent, neither too thin nor too fat. \"A lobster fills with the moon,\" the colonel told the girl. \"When the moon is dark he is not worth eating.\"   By the way, calling the martinis \"Montgomerys\" was Hemingway\'s idea of a joke: it was often said of the British WWII general Montgomery that he would not launch an attack unless he had at least a 15 to 1 advantage over the enemy. Hence: martinis made with a recipe of 15 parts gin to one part vermouth were \"Montgomerys.\" This kind of heavy-handed, not very funny humor was typical of Hemingway.  It\'s been speculated that if Hemingway\'s editor Max Perkins had still been alive when Hemingway wrote this novel, he might have tried to persuade the touchy Hemingway to cut it down to novella-length. It would have been easy to do, actually-cut out all the talk about eating and drinking and \"Across the River\" would have been a novella by default.   In truth, Hemingway was anything but a gourmet. To the chagrin of anyone who had to live with him, in fact, one of his favorite taste-treats was raw onion sandwiches with plenty of ketchup. And although he boasted throughout his life of his supposedly great knowledge of wines, true wine connoisseurs who either knew him or have read his works assure us all that much of it was pose. No, Hemingway\'s value to us as a \"food writer\" does not lie in any claim he can make to directing his readers toward the most exquisite and the best-prepared. Rather, it lies where the rest of his value lies, in an art so deeply involved in the world of sense-experience, an art in which the five sensory organs are virtually instruments of thought and expression, that we share in that world, which Hemingway made to come alive at his fingertips. Just as Hemingway\'s chiseled prose brings us to breathe the air his characters breathe and feel their emotions at a remove, so that same prose makes us taste the meals they eat and savor the wines they sip in a world as three-dimensional as good writing can make it. Hemingway wrote well about everything he loved, and he loved a good meal. We\'re all a little richer (and maybe a little fatter) for his having done so.   VI  Scott and Ernest: A Logical Friendship  The most famous friendship in American literature is also, beyond almost any doubt, the most exasperating to write about.   Matthew J. Bruccoli, author and editor of a number of works on F. Scott Fitzgerald (1896-1940) bravely set out a few years ago to write a book about the relationship between Fitzgerald and his younger contemporary Ernest Hemingway (1899-1961). Bruccoli\'s 1994 book â€œFitzgerald and Hemingway, A Dangerous Friendshipâ€ opens on a note that will set the tone for the rest of it: Bruccoli cites Hemingway\'s famous account in â€œA Moveable Feastâ€ of the first time he met Fitzgerald, a meeting that according to Hemingway took place at the Dingo bar in Paris in 1925, and at which Duncan Chaplin, who had pitched for Princeton\'s baseball team when Fitzgerald was a student there, was present.   Then Bruccoli goes on to write that Chaplin was not, in fact, at the Dingo bar that day. In fact Chaplin was not in Paris in 1925. In fact Chaplin was not in Europe in 1925. In fact, Chaplin never met Hemingway. In fact, as Bruccoli proceeds to show his readers, most anecdotes about Fitzgerald and Hemingway, and there are dozens of them, tend to evaporate under the light of investigation. Like a mariner hugging the shore, Bruccoli tries as best as he can to stick to the provable. As a result, the text of Fitzgerald and Hemingway consists largely of letters written by the two authors to or about one another.   After slugging it out with the evidence for 196 pages, Bruccoli comes very close to throwing up his hands in despair. On p. 197, writing of Hemingway\'s appearance in Hollywood in 1937, when Fitzgerald was there doing film work and Hemingway was in town to show and discuss â€œThe Spanish Earth,â€ a propaganda film about the Spanish Civil War which he had worked on, Bruccoli writes:   \"A researcher'),(12122,3380464,'21,May,2004','2004-05-21 00:00:00','   Indian Summer  			 Grief, sir, is a species of idleness. 			 The parked car ticks in its painted slip, 			No breath of air stirs the 			Hundred-million stained-glass windows 			Illuminated in the noontime silence. 			Drought killed the buds of summer early; 			The lingering wasps are getting desperate. 			Sit down to lunch beneath the still tree, 			And they launch airstrikes on your sandwich. 			What is it about this time of year? 			Timeâ€™s teeth are on edge, the very afternoon 			A seemingly-endless con game: 			The Washington Post lies face-up 			On the table, masthead and date 			Magnified through a glass of water, 			By sunlight too certain to be real. 			But their implied threat is as loud 			As the wasp-hum is quiet.  Somehow, 			Unlikely as it may seem, things 			Are going to get moving again, 			Their natural round only slowed 			By these impertinences, these piled-up 			Still-life snapshots which will forestall 			No frozen locks, no iced-over windows, 			No future species of idleness. 			This is when you remember things: 			The first day of trout season on the cold lake, 			The floodtide of epiphany in a first kiss, 			The onrushing of early buds in a 			Whirlwind that seemed to contain all 			Potential in itself.  Itâ€™s at the first equinox, 			And the first only, when all the great 			Romances of the world get written; 			From here itâ€™s all recollection 			And dread.  On the wall, a woman, 			Beautiful and nearly nude, steps down 			Toward the sea, her back to view, 			Everylastingly mysterious, 			Graceful as the rocking, moon-drugged 			Metronome that waits to embrace her, 			A fugitive from memory, returning home.                                                                             Washington, D.C.                                       October, 1995   			To Nadya, During Mahlerâ€™s Third  			The eloquent smugness of April trees: 			Outside, the immensity of returning green, 			As certain as the symphonyâ€™s opening phrase, 			Whispers that carousel of the will 			Which sweeps everything along the street. 			Rebirth is a treachery: 			By rights the dead should be left alone. 			The evening sun is a mockery, 			Grist for the ridicule of nightingales. 			Once we walked along a path of birches, 			Ears tuned to hear their peculiar song; 			Above and around us, spring went along 			Mindlessly with its annual outrage, 			While another awaited, just ahead. 			Seasons away now, your bright portrait 			Somehow resists the sun of timeâ€” 			Somnolent in my steps, I watch 			It hover, sometimes changing shape, 			Sometimes light or depth of color, 			Floating above the open window, 			As snowflakes swirl or pink bud-showers 			Fall to be squashed by passing cars. 			I found you in a bright November, 			(Give me the honesty of autumn leaves!) 			And now recall colors among this riot 			Altogether different from what I see, 			A karma of night, your eyesâ€™ geometry.                                           April, 1995  									 Listening to Wagner and Looking at Kansas  		Some farmer, scratching his butt 		27,000 feet below mine, 		Would never guess whatâ€™s going on 		Up here. 		The fields slide by, criss-crossed, 		Checkerboarded, 		Shades of green fading 		To shades of brown. 		Somewhere down there, 		Someone, no doubt, 		Is worried sick about 		Paying off a loan. 		Up here, the sun makes 		Silver dollars of the Irrigation ponds, And a whole world is falling apart In continuous melody. The plane shakes,  But the painted-on smiles Of the flight attendants Donâ€™t crack as they pass out Fruit and wine coolers. Meanwhile, Brunnhilde launches into Her final act filibuster, Preparing to take her flying leap On to Siegfriedâ€™s pyre. The landâ€™s a mottled pool-table As far as you can see. No wonder tornadoes  Get loose down there and donâ€™t stop Until theyâ€™ve plunked someone down In the Land of Oz. Accepting a drink And a few strawberries, I settle back, wondering If Iâ€™ll know when itâ€™s no longer Kansas weâ€™re over. Landscapes, myths and troubles Have boundaries all their own. The cymbals clash, The trumpets blare, And itâ€™s curtains for the gods Once more.  The plane lurches. I grab my seat, Panicky at turbulence, But itâ€™s OK in a moment, The air is smooth again. And up ahead, beyond the  Horizon even from this height, I know the Mississippi  Flows, a river whose own sagas Are no doubt lost beyond retrieving, Somewhere in the depths Below commercial traffic, Where the waterâ€™s just A bit polluted, And no maidens sing.      November, 1985   		Ocean Logic 	(Sunset Beach, San Francisco)  Descartes envisioned a universe That ran on wheels: itâ€™s easy to believe, Watching the surf crash in. I sit on the sand, unabashedly Reading out loudâ€”who gives a damn? The ocean drowns me out, And besides,  On a beach as seedy as this one, Whatâ€™s one more nut? The ocean beats time In rhythm with my reading. It makes more sense somehow Than the words on the page, Its logic as unperturbed As the merry-go-round of the polar stars. The tugging moon plays on its heart, As a smaller, darker, 		Interior moon plays on mine, 		Infecting its rhetoric a bit perhaps, 		But the message flows on. 		I put my book down and try to tune in, 		Scoping for that music 		Keats thought so divine, 		But it is there?â€”I canâ€™t tell. 		Thereâ€™s a freighter going out now, 		A wet-suited surfer just walked by 		On his way to brave the undertow, 		And Iâ€™ve got a jingle for Coca-Cola 		Stuck in my head. 		Fugues and discourses 		Of god-knows what kind 		Are lapping the continentâ€™s edge 		One hundred yards in front 		Of my face, and yet 		Their language is as lost on me 		As this poem is on you.                                June, 1985    				Sextet for the Dusk		 				    (Benicia, California)   			Secretive, gulls in figure-eight swirls 			Make halos around St Dominicâ€™s domes. 			Arcane in the breath of their holy 			Indifference, as dusk comes down they 			Head for the delta, like the fire-breathing 			Bombers that razed Dresden, and no less one. 			A runner, meanwhile, whose rhythms of blood 			And heart tick-tock him past the Spanish 			Portico, is ambushed there by gnats 			In a cloud, orbiting the nucleus of their 			Common cause.  Three stray dogs, hot for 			A fourth, dance braids of lust 			In the tall grass of a vacant lot. 			 			The Chinese box of our gravities, 			Sun on sun, in layers of night, 			From house to city to planet to brain, 			Wraps universes in quiet submission, 			The whispered non-language of assent. 			The downward-speaking ricorso rain, 			Lullaby and word, rinses the street, 			But blurs the sharp-edged, defining light. 			â€œSleep,â€ it says, then its work goes on: 			It soaks the cemetery at the city limits, 			Gently erasing the names from the stones.                                  November, 1982   				        The Whipcrack  						I  			A tree seemed a natural place for Dylan back then; 			Just in from Minnesota and not yet a superstar, 			He looks down in those old photographs with just 			A hint of the sardonic grimace young America would 			Ape when it clutched him as a banner.  And he looks so 			Right, bearing that bullwhip down the long 			Country road, ready for all comers, an Indiana Jones 			Of the ideal, poised to battle hypocrisy and greed 			Back to the caves of the unconscious.  Crack, 			Went timeâ€™s echo-judgement as the stroke of that 			Whip reverberated for miles and miles 			Over the not-yet famous Woodstock hills.   						II  			The portentious whistle was faintly in the air 			As Strauss lit his candles while the empire faltered, 			Determined to give overwaltzed Vienna what it didnâ€™t 			Want: a magnum opus.  They whirled to his tunes, 			But he could hear a more malevolent kind of 			Acceleration waltz picking up speed above his 			Aging head as they danced and danced.  He prayed 			For rain, the only kind of weather 			That could give him peace, keep the dancers from 			The door.  And so he sat down, pen in hand, 			But time was running out: already the whipcrack 			Heralding the non-tonal storm was in the backswing, 			Gathering momentum in that rainy night.      					III  			Children on the frozen pond, hand-in-hand 			In winter, string out, the wind on their faces and 			Their heads full of Christmas, inscribing joyous 			Nonsense symbols on a frozen slate. 			Arcs, circles and pirouettes scar the surface 			Of what was just a few months ago 			The undisturbed mirror of a summer sky.  Their lives 			Are still andante; a year to them goes on for 			Miles.  Look: they swerve, they swing,  			They crack the whip--scattering like a covey 			Of birds flushed from a thicket, they fly off 			In all directions.  No sound but their inarticulate 			Laughter pierces the deadened December sky.                                 November, 1982    				Prologue in March  			Hushed sunlightâ€”and you remember it again, 			(Here, in places you had nearly forgotten) 			In air that smells of something besides 			Stopped watches.  It wafts through the room. 			The impulse is not to jump up and sing, not quite, 			But to go and seek the echoes in the stairwell 			Of what seemed the cries of demons not long ago. 			Now youâ€™ve bought, with no other currency than 			The accumulated dust on the windowsill, 			One more glimpse through the prism thatâ€™s always turning, 			Each facet presenting the same scene 			A thousand miles from the last. 			Neither hope nor despair make any sense now; 			There is in this breath only a colossal never-to-be-defined. 			The drone of an airplane passing overhead 			Recalls the drone of the electric fan that hummed on the table	 			So many midsummers, blowing in from the window 			Those superheated liquors of honeysuckle and jasmine, 			A breeze that made the blood want to throw off its rhythm, Demand the next day, leap into the river 			And drift downstream, an echo of afternoon thunder.                          March, 1988  Up The Street and Through the Cemetery  (Ann Arbor, Michigan)       for Jesus R. Araiza  The frat-houses, hugged By not-quite eastern ivy, Are old-money smug.  			We mock their pretense: 			â€œGaudeamus igitur,â€ 			As we pass the fence.  The graveyard, you say Is an interesting spot. Why not?  Lead the way.  A water-pipe drips, Phallic at the graveyard gate. I miss it, and trip.    You grin, then pause.  â€œSee?   The city disappears here.â€   I look, and agree.  The rooftops at bay Are only half the story. Time has backed away.   Granite names preside,  Some recent enough, some aged  To the death of pride.  You whisper a word Appropriately humble; The reply of birds  Stirs up the trees, Echoing your sentiment In a sudden breeze.  Empty cans of beer Rattle in the brown dumpster. â€œFrat boys party here,â€  You say.  â€œItâ€™s a great Place for Friday-night bashes.â€ Meanwhile, quite sedate,    Anonymous dust Pays no perceptible heed To frat boys or us.  â€œSo what do we do?â€ I ask as a family tomb Blocks the skyâ€™s pale blue.  â€œNothing,â€ you reply. â€œWhatâ€™s the difference, anyway?â€ So what if you lie  Beneath beer-drinking Gamma Phi Alphans some night? A similar sinking  Awaits them all too. No use brooding about it. Iâ€™ve heard, in Peru,  People come with food And picnic among the headstones, Enjoying the mood.  Multilingual stones Of early Michiganders Stand above their bones:  Hereâ€™s one in Chinese-- There, a Spanish influence Molds beneath the trees.  Dates to when the land Was yanked from the Indians Fade from where we stand.    Whole families past Under common granite names Sleep where union lasts:  No discordant word To question what they lived for Is here to be heard.  One stone reads â€œBABYâ€ For a nameless turnstile life Cut short at â€œmaybe.â€  No pair of dates here: Entry and exit took place The same day and year.  The clock tower chimes. Weâ€™ve circled to the main gate My watch says itâ€™s time  To go home and cook. Bored, you smack the fence-rails with A paperback book.  Buildings reappear Above the whispering trees, Smiling at the fear  That whispers all around The corners of calm habit. Glad, I greet the town  Silently, as two Possible frat-boys come in, Tossing, as they do,  A frisbee back and forth, (they have two) and then the wind picks up from the north.               June, 1982    Wine-Tasting (St. Helena, California)		      First, the chardonnay: A man dressed like a funeral director Waves around glasses Three at a time. â€œItâ€™s very drinkable,â€ he says, Then mumbles something About how long it stood on its head At sixty-eight degrees. A sip, a sidelong glance at you, rolling it around on your tongue. Next, the riesling: Another character (This one looks Levantine) Doles out drops Of the precious stuff While those who didnâ€™t care For the chardonnay dump their dregs  Into a huge pickle jar. As you cross-examine him About sugar content And the time the grapes spent Soaking in their skins, I give my eye to the muscat, (The end of the line) Thinking inappropriate thoughts Of brown-paper bags on skid row. Then comes the cabernet: You donâ€™t care for reds, especially, And tell me so.  Something to do With the acid, you said. But I sniff the blood-colored marvel And let something slip About a well-done steak with mushrooms. This spoils your appraisal As you swallow and giggle. The roomâ€™s suddenly friendlier Now, the Sunday tourists From 400 miles around Just a big befuddled family... And now, finally, we reach  That muscatâ€¦ O, itâ€™s sweet,           Itâ€™s a song,                                              Itâ€™s the color of your eyes.                                              It cries out for a big, roaring                                               Fireplace somewhere, a sheepdog          And other stage props.                                              (Did I mention that                                              Itâ€™s the color of your eyes?)  Then we squint in the daylight, Preparing to drive Another quarter of a mile Down the road, And do all this again. (Champagneâ€™s been promised.) All around, The valley sings your praises As the Sunday sun Gently kisses the vineyards, Working up that sugar content. The newspaper on my lap Says something about Dan White, Just sprung from jail. Would he dare, they ask, Return to San Francisco, The killer of Moscone and Milk? Who cares?  Iâ€™m in love!                  March, 1985    			Weightlessness  Some of the seagulls circling that about-to-explode idol Will be dead within the hour from the noise. But some will survive, and would pass the story on, If there were any way they could, of their enormous Counterpart, which tore the curtain of the morning And then disappeared into the sun.  With the window open, I could smell the dawn, and I turned to watch and hear The sparrows on the telephone line, their notes the right hand, Your breathing the ground bass of the left, As the cat jumped down from the lump that had to be Your feet: this morning- counterpoint was reveille for her, And she went off looking for her breakfast. What were you dreaming of when the firecracker That would piggy-back the shuttle into orbit Exploded three thousand miles away?  I wasnâ€™t aware of it Myselfâ€”I heard about it later, over coffee,  With the radio on and the cat chasing invisible demons Up and down the stairs.  God, what must it be like, The g-forces mounting, The astronauts suddenly finding Themselves upside-down and staring, as they ride away From earth aboard an earthquake, at a sky deepening From blue to black, and then the release, the sudden Absence, the drifting harness-ends? You stood at the sink in your robe, dishing out cat food. When I stood behind you and kissed your neck, You purred just like the cat.  Sunday: we had hours to fill, And no need to hurry the morning along.  The heavens Were doing quite well by themselves, or maybe Assisted just a little by the human presences: Those singing in churches and those loosed from the earth, Chasing broken satellites. I followed you back upstairs, reaching out to run my hand Along your spine as I counted the steps of your bare feet On the carpet.  The news was over with, the radio Indifferently offering Strauss, and we lay down again and Overcame its indifference, your eyes coming to light as Your robe slipped to the floor, the door closed so that the cat, Still jousting with demons, wouldnâ€™t bring her battle  Into the room.  No battle here, just rising g-forces, And the sky I envisioned turning from blue to black, And the universe warming from the three degrees Kelvin of its Birthing moments, to a warmth wherein our births could be Re-enacted.  Listenâ€”was that a knocking I heard just now On the door downstairs?  No, my mistakeâ€¦It was the sound (I like to think) of my own knocking on the door of some Creation, the hammer-blows of longing for a reunion With the one whose messenger you are, warming the clouds As we ride a quieter earthquake to our own escape velocity. And there it isâ€”a sudden hush, like the coasting after the burn, A silence like the space between galaxies when the echo Of that great birthing faded to form the microwave whisper Whose message has repeated all these billions of years With no one listening until just now.  Then suddenly Iâ€™m floating.  No dangling harness straps, no crackling Radio traffic, no need to do battle with the ticking sky  Or anything else.  This is being weightless,  The purest kind to be found between the other two, The floating in the womb and the drifting toward death, But more like the first I thinkâ€”the sparrows have left, But dawn noises reverberate in my head as your arms Draw me in, encircling me in a boundary not unlike Beginnings themselves, just as the layers of whispering forever Encircle the shuttle, drifting 200 miles above.                     March, 1985   			The Minimal Muse  To stretch a point, Newtonâ€™s Third Law Would have us believe That for every Turangalila, There must be a 4â€™ 33â€ Or the other way around. Meanwhile, the poor sap in the middle Shakes his head And switches on the Muzak. â€œItâ€™s lovely silence, But who can dance to it?â€            June, 1985   Voices		  for Lucia  			 			Some follow them to their doom, 			Like the maid, born to burn in 			That more credulous age, who  			Heard in the wind an exhortation 			To drive the invaders out. 			But that was back when the doves 			Came down to take the eyes of the blind 			To heaven, restoring sight 			When they returned to earth. 			Things like that donâ€™t happen anymore.  			If I were to hear voicesâ€” 			I can see it now: the woman behind 			The counter down at County Mental Health 			Would reach for her pink form, 			Pencilling in â€œstressâ€ at line 17b. 			â€œTake this down the hall,â€ sheâ€™d say. 			â€œOtro loco mas,â€ sheâ€™d think. 			What happens, then, when you hear 			My voice?  Are forms filled out? 			Do the doves return? 			Rest assured that though  			The ever-widening sky  May take no notice,  while my love burns, you never will.                           April, 1982    			The Sniper  				I  Someone or something always did Your sacred job.  Uncaring sky And primitive paranoia Were your eldest antecedents. From the Assyrian hell-wheels Whirling fear to the marrow Of subjugated bones, to SALTâ€™s Nervous pulse-checking, you were always there.  				II  Passenger planes over the wheatfield Of infinity wait for you As you watch from near the window, And your various shapes invade The flaming tree-tops and the hills. Everything hinges on your choice For those who fall within the range Of your sight, and are within you.  				III  Einsteinâ€™s face stares from the textbook, Seeming burdened with all he knew. A picture that seems meant for walls Where teachers drone and students yawn. But in another picture, where, Mugging for the camera, tongue stuck out, He takes the smug ones quite off-guard, Iâ€™d like to think he just dodged you.   				IV  Paramedics playing poker At midnight on a Saturday,  Breathing between catastrophes, Inhale that heavy air for life. Not too interested, they look out From the back of the van at one. A nightâ€™s work is just a nightâ€™s work: The shifts they stand are two nights long.  				V  So who can ever go between Your dark intent and our mistakes? Distinguishing malevolence From rainfall is not for our kind. The paramedicsâ€™ van pulls up, Lights flicker over Einsteinâ€™s face, An airplane crashes in a field, Another leaves an hour late.                        July, 1978   			Gumball Universe  Eleven dimensions now, they say. The knobby surface of reality runs Beneath my fingertips, mercurial, unseen. We were just getting a grip on the ones We had, still trying to fine-tune our fears to the ins and outs of matter and Energy either sunflowering in that burst  Of horror thatâ€™s haunted our most recent Long sleep, or smoldering, a mean dog Unaware, but growling as you pass from habit. And then this came along.  What next? Will they fashion some new malevolence out of it, Or will it remain the sole property of the Priesthood, they with the unpronounceable names, Who tend to address themselves in integers? Every day they reshape the unshapeable, While, at the galaxyâ€™s core, something unnameable Is already ordering something unspeakable: A wholly new alphabet, a rewriting and reordering Of all the holy books ever written.                        March, 1985   			Whiteout                 (Vacaville, California)  Itâ€™s the deadest Sunday of the year, 		and the yearâ€™s just getting startedâ€” Consider that. 		Not quite time for the Super Bowl: How will we survive? You can feel it, time,  oozing from a million living rooms, Feel it as you walk to the laundromat, Or watch your breath Materialize. Even the church crowd is hiding out Somewhere. What there is, 			of course, 				       is fog. It drifts and drips                and runs along the windowpanes, 					the icing on silence. You can get lost in it, and somehow Today, I think, everyone would like to. Even â€œ60 Minutesâ€ is miles away; For the moment thereâ€™s only basketball 					and Jimmy Swaggart. The fog gives the percolatorâ€™s bubbling an urgency, makes the pine needles on the patio, (from where the Christmas tree was dragged away) So mutely elegaic, and urges such a General anonymityâ€¦ What it whispers as it 			           			streams down the windows Is clearer than  the streaming windows themselves. The only thing we really need now Is one good snowfall  to wake us all up within view  of the Emerald Cityâ€™s gates.                          January, 1985    To Marina on her Birthday 		 			There was a time when you wouldn\'t 			Have thought twice about them, 			The second looks and double-takes 			Attending each venture 			You made along the street. 			But in these days of the phone 			Off the hook at sunset,  			And the solitary walks around 			The old neighborhood, I hear 			Sometimes you catch yourself 			In the long fluorescent glance 			And think \"No, I don\'t look 			So very much different--still, how 			Could they?\"  They do.  Take my 			Word for it: I\'ve watched, myself, 			(Trying not to get caught) 			As you made your way down some 			Yawn of a corridor,  			A splash of delight, bright in 			Our general sleep-walk, rousing 			That never-aging catch of breath 			That so often becomes song. 			For all the summers you\'ve been 			Part of, all you\'ve seen and heard, 			Smelled, touched and tasted, 			And in the shadow of every word, 			Foreign to me, familiar to you, 			That spells out your other- 			Looking-glass life, still I envy 			Those chimerical, bright-as-you-are 			Mid-Atlantic birds that attend 			Your May morning walks to 			Safeway or CVS...Having seen both 			Them and you, there\'s no doubt 			I think, that grace and beauty 			Have rules of their own, and where 			And when they choose to speak 			Or put in an appearance, 			The rest of us can only smile, 			Scratch our heads, and try to be 			Worthy of the moment.                              July, 1998  During Barberâ€™s â€œAdagio for Stringsâ€  The most longed-for moment is never the furthest away: Imagine the ultimate 2 a.m. under fog, The night as silent and bright as remembrance, The universe illuminated in the orange glow of that Artificial moon,  sleeping city light rising, The cloud-ceiling phosphorescent. Move from the corner room, where the dim sky Illuminates nothing, over the profile of the highrise And down to the street itself. Here there are no Declarations. The clock on the Security Pacific Bank Building chimes an everlasting frozen hour; The police car standing on the corner might be Unoccupied, its engine silent and its headlights all its own. No answers are heard, but the questioner is celebrated With a mute tenderness: footsteps out of the moon and fog; They carry you down to the street where, loud as midday, The rustle of an irrelevant newspaper affronts the stars.                        June, 1990  Six Love Songs         Music: Jesus R. Araiza    					          I  			A falling leaf 			Reaches the ground 			Without a word: 			It makes no sound 			As it flutters 			Down to death 			To disappear 			In a falling breath.  			In my time 			I hope I may 			Be like the leaf: 			Youâ€™ll turn away, 			Your words unsaid, 			Your eyes gone dry. 			Iâ€™ll say nothing 			As you pass by.   				                   II  			Faces follow your footsteps as they come. 			Your fast-approaching cadence fills the air 			With image-blizzards, pictures of yourself 			Torn up and windblown, flying everywhere. 			The cadence stops.  Youâ€™re opening the door. 			A flood of pictures comes in on the click, 			And hidden rains of ancient melodies: 			The fevered music of the deathly sick. 			Yet when you speak, your voice upon my ear 			Is sweet, but not the music that it seems 			When youâ€™re far off.  The sunlight on your face 			Is not, perhaps, as lovely as in dreams. 			The crisis comes: I reach to take your hand, 			And though I shake, the walls stubbornly stand.   					III  			I dreamed my love were dead, 			And I smiled as she lay 			Naked at my feet, 			A vanquished enemy.  			But morning took a hand 			And pulled me back to earth, 			The jacket on the chair, 			The day and dayâ€™s preserve.  			Later, when I saw you 			Sitting at your desk, 			A lion passed the window, 			But then lay down to rest.   					 					IV  			Morning got in here 			Without being frisked. 			No one asked the day 			Its business or intent. 			March flows into April 			Naturally as night; 			Who would ever want to 			Demand its credentials? 			Entropy envelopes 			The entire sleeping earth 			In one-way fluid faith. 			And the elemental you, 			As of the earth as stone, 			But above it as Polaris, 			(Unflinchingly that way) 			Must surely wake to dayâ€™s 			Unfrisked arrogance 			Equal at all times 			To whatever it demands, 			Chosen and in time, 			Subject to a law 			Binding you alone. 			Your little fingerâ€™s ring 			Glimmers with your intent: 			A crack in the sky considers 			Each impaling hour. 			It holds at your discretion, 			And at your will explodes.   				   V  			Ocean, clock and sky conspire 			In their dark complicity, 			Pulling you from my embrace.  			Ocean, large as night itself, 			Comes in, goes out, its rhythm still 			Pulsing like the bloodâ€™s demand.  			Clock, a spinning manmade will, 			Blindly speaks its own intent, 			Ticking fast when we would wait.  			Sky sweeps toward a breaking dawn, 			Hushed around its polar way, 			Threshing through the dark we share.  			But you and I grab stolen gems: 			The seconds that we steal away 			Are ours alone when darkness ends.   					VI  			Now throw aside 			All pettiness 			And come with me 			Into this light. 			Our bodies fit 			Like dovetailed dreams; 			Together we 			Invade the night.    			 			Sadness, defiance, 			Joy between 			The canyon of us, 		 			The valley of you, 			Paint the rocks 			Of cold desire. 			Now climb the cliffs 			And see the view:  			A galaxyâ€™s heart. 			Naked, you are 			The center of all, 			The beginning sun. 			I lie in your core 			As skies explode, 			Hidden in you, 			The completing one.                                                           1979-81     Es ist Genueg		  		In Alban Bergâ€™s great Violin Concerto, 		After a series of tutti hammer-blows, 		And after a quirky glance back at Vienna, 		The woodwinds talk some older talk	 		Still: that \"alter Duft aus Maerschenzeit\" 		--scent of other timesâ€”appears and grows, 		A scent Schoenberg would have recognized, 		And Mahler (had he lived) might have brought home. 		It hints at you, among the rattling 		Dishes and the ringing telephones; 		The lotus of tranquillity in you 		And the chorale are much the same. 		They speak a common language at the bottom 		Of the mind, echoing in the mists of 		Tonal instinct, seeping from the crazy- 		Quilt of time.  Yes, you and the music 		Are of a common kind: two echoes   		Of vintage sweetness in the toneless street 		Where the noises rising from the workshops 		Make that harmony more precious still, 		In the way you have, and doâ€”the way you will.                                       July, 2000    Blake 		 			Challenged to sketch 			The soul of a flea, 			He jumped at the chance, 			â€œThatâ€™s it!  Canâ€™t you see?â€â€”  			Then glanced with reluctance 			At Londonâ€™s gray streets. 			â€œRealityâ€™s nothing 			But a chain of defeats.â€                                       January, 1996    			Near Antalya                        (Belek, Turkey)  				Tongues, faces and weather meet 				Here to argue in harness with 				Dead generations, trucks and Fiats 				Belaboring the road like rhinos or 				Camels criss-crossing the touristy sand.  				You\'d think we\'d be wiser now.  This 				Old \"Mare Nostrum\" unfailingly 				Listens, has heard so much, but still 				Never shuts up;  it talks on and on 				Despite the speedboats and jet-skis...  				Just a hundred feet north, and not 				Quite listening, we play out the 				Old script once more, and then once again: 				The heart\'s pile-up, mirror-drunk mysteries... 				If we can\'t locate ourselves, even here,  				What chance that we\'ll ever find each other? 				The sea either won\'t say or won\'t stop 				Repeating, while the wind, aimed  				At Africa, babbles its mimicry, their 				Common code, unbreakable as speech.                                September, 1997  							  		M51  Old Messier made himself a list Of things that could be lightly missed: Numbered, toss-out points of light, Distractions to his busy night. He was on the comets\' trail, And never thought beyond the pale Of Newton\'s clockwork universe-- These light-puffs were a nightly curse. Now, spinning in a photograph, This whorl, (which Messier thought chaff) Some thirty million years away Appears much like the Milky Way. Old Messier would never know He\'d forced the universe to grow; These \"sands upon the Red Sea shore\" Were not what he was looking for.            July, 1997  Joshua Blues         In memory of Richard M. Bettez, 1957-1980  Running deeper than remembrance in these bones, Richard, Are the fading picturesâ€”remnantsâ€”of what we once were. Look, though: the desert you loved to paint stays  Essentially the sameâ€”the land, the 40-mile horizon, And of course the joshua, that enormous tree Whose portrait you once painted, then brought Around to my house.  We talked as I brewed tea.  Now youâ€™ve joined Picasso, whose vision you once Panned, all those years ago, sketching in Bic pen A crosseyed Cubist man, which you then Showed around for laughs in the high-school halls. Proof yourself that an artist could out-dance An athlete, you went out for sports nonetheless, Chased after girls, loved lifeâ€”a painter one minute, A rocker the next, then again something neither could name.  Fat chance you had, once you decided to choose The mire of that suburb which nurtured us both Like feedlot steers, as a place to come back to. Drifting death, that jellyfish, floating White on the water of our recalcitrant years, Marked your return, watched over the road, Then gently let down his tentacles to brush you As late one night, you drove too fast in the dark.  Once we talked of travelling, partners to Paris, That summer I was pumping gas and mopping bay floors While you copied Constableâ€™s The Hay Wain. We never went, but you, loving landscapes, looked closer to home for what you could secure As your private domain. The desert?  You could have it. And now Iâ€™m living there, watching desolation bear Down--the perimeter of mountains saw-tooths hot sky.    But the joshua tree is out there somewhere, Clinging to its is-ness, stark silhouette of patience, Like the land all around.  If I should come across it While driving in circles, finding nothing else Worthy of note, I may not be sweetly mugged By the muse of your release, but you never know. You went before the rest of us, first to flower, And then to embrace deathâ€™s dirty logic,  And now itâ€™s the joshua as well as Picasso Whom youâ€™re privileged to join in its petrified peace, Whose secret is the mastery of that elusive art, Homing in on the landscapes of the desert and the heart.  				March, 1981    			Dirty Birds  Would you say That thereâ€™s a little of the seagull in all of us? That putting aside all the  Romantic hogwash About complete freedom, And all the freudian nonsense Imbedded in the desire To fly,  Would you say That thereâ€™s a deep-seated yearning In some forgotten corner of the best of us To be tick-ridden, filthy, Despised by sunbathers who donâ€™t like White deposits  plunked in their navels, And forced to live on floating garbage?  Buddhists might say That your karmaâ€™s kicking up, And that maybe, just maybe The next time around, Youâ€™d do well to set your sights A little bit higher.             May, 1985      	 Page One  Itâ€™s not the sound of bursting bombs, But the low whispers of lâ€™infame That shake this card-house in the night, Which rose in artificial light In that unspeakable grinning face. Not the whine of planes in strings As the terrified quartet sings, Remembering the fevered hours Of air-assaults and horrorâ€™s flowers Blooming above the roofs in flames, But rather, those who have no names Under a collective hood: They mouth a common, choral good And follow one who sniffs the breeze And maybe marvel at the ease With which they let themselves be led. O tell me, is the hero dead, And that which killed him in the street? He lies beneath the led onesâ€™ feet As the bewildered mumble psalms And piped-in music calms and calms. 			   	Karma Bum  Sticky night air Hammers at his wakefulness, And in the darkness he listens hardâ€¦ For what? Some blank, redeeming sound In the promise of morning? Noâ€”his ears are tuned To the sound of His own heart, Wrapped in envelopes Of accumulated jet-noise, Radio static And curses long ago spat out And remembered. He strains to hear The birdâ€™s wing on the night air, Frozen in the act of its homing. â€œThe angels mark Your every step,â€ heâ€™s told. â€œThe tightrope may snap At any moment,â€ he hears. Where in the ticking sky Is the mirror he seeks? Those who went before, (How he struggles To imagine their faces!) Far outnumbering  Those who stand around, Line up behind him, Shaking the core of his Fibrous interior galaxy With a universal sigh Of what he hopes is peace.                        Ann Arbor, Michigan                       May, 1982       	Going Back To Ithaca  They tell me that Shelley was disappointed    The day he ran down to the docks and saw The curly-haired Greeks who reeked of garlic    And sweated in the sun as they unloaded Some cargo or other.  They were not the Greeks    He knew, only having met marble Greeks Himself.  Poor Percyâ€”before his startled eyes,    Under a cherished and well-fostered notion The columns crumbled.  What was he out looking for in that harbor?    Perhaps only a link with the source of us, Some reminder of beginnings: a dreaming    Descendant, his eyes deceived by his training, Marble-conditioned, looked for a marble forebear.    (And I am told that even now, young Greeks, When those glorious books are put in their hands,    Puzzle over them like a high-school kid Required to swallow â€œBeowulf.â€)  So what did it matter to my friendâ€™s sister Debra,    That the ghosts of great men moved all around her When she went down to the docked destroyer,    Sons-in-tow in some hellenic port, To greet the man she was to marry?    There was no sad surprise in her eyes When that soulless gray bucket (bought from America)    Yielded him up.  If there were any old statues there, I imagine she stood with her back to them    And smiled.                                             October 27, 1974   Terminalia  â€œDid Rome become ruins, Or do ruins become Rome?â€ Ruins do seem to fit the script: â€œRome,â€ it reads.  â€œCrumbling marble.â€ And now, when people think of ruins, They think of Rome.   Irony: when the theater Of all the Aegeanâ€™s past glory, And the eastern half of â€œMare Nostrumâ€ Were only a corner of what was Romeâ€™s, no patrician in his right mind Would dream that stage-setting, Not even on a morning after, When thoughts of further conquest Could only aggravate his Aristocratic headache.  This occurred to me when, Late one night, yawning every  Fifteen seconds, I thought about The roses Iâ€™d sent to your door. There were twelve, in a glass vase. They were left on the step, And I was angry when I found out, Thinking that they could have been Stolen before you returned. But there they were, on the table, And you came to me, (your hair was wet) And threw your arms around My neck.  I glanced over at The yellow roses, and wished They had been open then. They were embryonic, like your thoughts. They were thoughtless, like your eyes. They bloomed while I was gone, And were opened up The next time I dropped by. Within a week, they had  Turned brown, and you regretted Not having pressed one in a book.  Then they went the way Of all garbage, and I felt sadness, Not for eighteen bucks out the door, But for having given you Something that would die. Tonight, however, I donâ€™t think anything More permanent could have been so right. Disease demands the price It will, and time has moved Only slightly, but slightly  Was enough.  Lent now: A week ago last Wednesday, The reader smeared some ashes On my forehead, and told me I was going back to dust. A thousand years from now, (If they have Ash Wednesday then) Iâ€™d like to think the reader, Smearing ashes borne of me On someone elseâ€™s forehead, Might think of Rome and roses As well as Lentan things; For you, if for no one else, They were loveâ€™s best-chosen gift.                          February, 1977        	 Good Night, Good Night 	         The bass drum pounding in the kitchen sink Goes unnoticed as the moonlight moves A little further westward.  Do you think Youâ€™re equal in heart to what the moonlight proves?  Youâ€™re not alone: the moonlight on your face Silently complements the distant siren-sound As, across the city, two police cars race To where some drunk lies bleeding on the ground,  Having been rolled.  You do not hear: Your contentmentâ€™s unchallenged as the dawn comes on: You dream the right dreams, have nothing to fear As sailors stumble home when the moonlightâ€™s gone.                                                May 5, 1977      	Apricot Brandy  Proust was right: there are moments when time Turns back on itself and, transparent, Pulls back your eyes from the persistent crime Of one-way motion and blindness.  And then the illusion of travelling power Reveals what may be or may not, (In a grain of sand on the beach of an hour) An independent reality.  Four years flew back in his startled face When he uncorked the bottle, And what hurried past was too complex to trace In the odor of apricot brandy.  A midnight, a high wind, a batch of caresses, A moment when a friend now dead Still walked the streets, girls in short dresses: It all raced by, inhaled in an instant,  And was gone.  O was it, I wonder, that quick breath Like the moment of lifeâ€™s unreeling Weâ€™ve all heard about at the moment of death, And the culprit was apricot brandy?                                    January, 1978   Faithful Physics		  A dissonance of voices, both strange and familiar, echoes From every corner: a standing-in-the-middle of things becoming, In combination with things imagined, or as we hoped they were. Pick your corner: on which does the marketplace go mad, On which does the guitar player easily pluck his strings? Are they where we put them yesterday, or did they move While we slept?  Harmonies themselves ring strange, A standing-in-the-middle of colors that shift and skate On the surface of what world?  Do the poplars bend as the exile Imagines, or in a wholly new direction, obedient to The whisperings of possibly-hostile winds?   Compassion holds his hand back as he raises it to strike The spider crawling up the bookcaseâ€”or is it compassion? The uncertainties of the whirlpool come yawning, And in them is imbedded the possibility that he may be Forced to see the world sometime the way a spider sees it. Achor yourself on a chain of the expected: The bubbles in the glass rise for a reason. Then again, what if those molecules were to go into open revolt? The window might become an angry eye staring inward, The world-globe on top of the shelf a fist Looking for something to smashâ€”or is there too much sense Even in that, a hope that the whirlpool can be trusted To go along with the logic of our nightmares? Looks once readable take on a catlike strangeness, A standing-in-the-middle of a most peculiar entropy. Do the sunlight shafts, striking your face this morning, Incline as they did the first day we awoke together, When the doves who had escaped the huntersâ€™ shotgun blasts Rang like bells in the field outside, and the mountains Rimming the desert hummed like buddhas?  Was it even, I wonder, the same sun?  It hasnâ€™t blinked once, but the light Has changed: books unread multiply on the shelf where the Spider crawls, the furniture has gotten more worn, And thereâ€™s a resonance, a dimly-hollow bell-echo, Not fading, but building to an almighty shriek That could only make us run for cover. The boomeranging forces of the whirlpool, growing Like acre-feet of nuclear waste, leave us with the ever-more Desperate search for a place where they can be safely stored, Until their half-lives have run full circle, and their enlightened molecules lie down To hum their own mysterious peace.                                                     March, 1984    	Snowflakes         to a small child glimpsed at Safeway  When I was young and you were unimagined, A compass in a circle, I looked out At a landscape only as wide as I dreamed, And peopled and colored it with what seemed The very best the kaleidescope (my gift at birth) Could whirl into being.  And you were looking At me like thatâ€”you had the same secret Strength, a god unaware, Riding along in that shopping cart, Leaving it to your mother to worry about More earthbound, less-important things.  I waved, you stared.  What background place In the lonely circle youâ€™re building yourself Will I occupy?  The sun, the sky, the circumference As it looks to you are yours alone, As mine were, before that circle widened, And everything shrank to its appointed place. Worlds like snowflakes: within the space You occupy, (as I once did) where trees Could be monsters, sunrises gifts, And holidays lurked beyond the horizon Like joyous constellations waiting to rise, Everything you see, singular crystals, Was there to be arranged as I saw fit, And now itâ€™s your turn.  So build away, And live as long as youâ€™re allowed In the magic circle of that divine neurosis, Doomed to grow until you awake one day To find the process of its destruction Suddenly complete, the boundaries you laid out Nowhere to be seen, the colors dulled, The constellations set, The mysterious noises just distracting sound, The snowflakes melting as they hit the ground.                                            July, 1983    	Any Two Things                      for Holly       At Chartres cathedral (they tell me) you can see God    the Geometer, holding a huge compass, and some of the more mystically-minded still regard    Pythagoras as being on the right track. Probabilities themselves, they say, prove that there is    an Alpha and Omega behind the veil. Imagine, for instance, the odds against two flies in    that very cathedral, buzzing past Godâ€™s nose, and somehow managing to bump into each other.    Now multiply that a few quadrillion times, And youâ€™ve got the odds against two random electrons    finding each other in a blank endlessness. Then thereâ€™s the one weâ€™ve all heard, the â€œtyping monkeyâ€ scene:    Statisticians say if the little buggers Jumped up and down long enough, theyâ€™d manage to write not    only â€œWar & Peace,â€ but all of Shakespeareâ€™s plays, and even the grocery list I tossed out this morning.    That makes us a cosmology in ourselves, I think.  We were born on opposite sides of the world,    six years apart, and the credulous would say that our meeting, when it happened, was no accident.    Then again, what if it were dictated from  nowhere, then taken down by a zillion cosmic monkeys?    How pretty a world does that augur for us? You asked me how I write, and this is how I do it:    (or how it happens, anyway) Iâ€™m a void Where any two things will meet and spark, and a third is born.    I speak, we touch: the word and the touch are things in themselves, their existence outside my volition,    yet they spark, giving birth to angels and suns. Islam teaches that everything began with a sneeze.    I think of that, when I think of how you laughed the time I told you how I sneeze when I get horny.    But when you come right down to it, itâ€™s a lovely thought that each time I desire you, I make a universe.                Frankfurt-am-Main,                May 18, 1986     	Broadway Melody	          for Henry Miller   What kind of night is darker than big-city night? (Big-city night under a billion watts of light, Thatâ€™s the hell of it.) How in the world did you endure that world Of love in smoke-filled rooms, Where the men didnâ€™t take their hats off, And the women didnâ€™t smile? A dance floor in hell: all those Nickels in all those Slots in all those Machines in all those Automatsâ€”ugh, no wonder You wanted to sing five thousand year-old songs, Get drunk on water, Speak any language that never heard a Brooklyn accent, Wag your hard-on at the sun. The time of the assassins has come and gone. Welcome to the hour of the wet ankles, The time-bomb ticking in the blood, The air-conditioned nightmare gone condo. But you carried bigger bombs inside yourself, And I can see you, flinging them left and right, From a bicycle all the better for being imagined, Pedalling along streets whose shrieks of protest To you and you alone  Sounded like Scriabin, Faster than the speed of bullshit, More powerful than a loaded motive.                                    December, 1988   	CELLO  They said: â€œThe sense of this music is like the patterns of the grain in wood.â€ You stand in the corner. The pattern of the grain in wood harbors secrets, And no less the demand of your presence, insistent, by dint of the very eloquence Of mute potentiality, and the unspeakable tautness that you imply into the air: Somehow, someone is compelled to make music here. To make music, the original demand, not the flick of a switch That frees you up to let your thoughts wander as something called â€œmusicâ€ Fills the room. Sorting what you demandâ€¦it isnâ€™t easy, nor necessarily enjoyable: Participating in mysteries not very often is.  The sublime implies catastrophe. The tightrope-walker may smile, afterward, watching the videotape, But not when heâ€™s picking his way slowly forward, under lights and over eyes. This morning I tucked my pocket watch into the change-pocket of my jeans. I take it out now, pop it open, see that itâ€™s early afternoon, the sky outside gray,  The upright coffin over there summoning without a sound, bringing to mind The story of the Russian beauty who told the poor smitten clod to bring her The tsarinaâ€™s shoes, and then, maybe, he might get what he wanted. Now, in all your elegance, you restâ€”but not rest at allâ€”against my shoulder, Between my knees, poised, your strings making that most impossible And at the same time most promising demand against slightly-calloused fingers: Then slowly, in motions repeated that they might be as automatic as breathing, I try to bring out of you (watch ticking, small sweat starting up on my forehead) A Bach prelude, but not just a Bach prelude, no, that wouldnâ€™t be enough-- It has be a singularity, a once-in-a-universe, never-to-be-repeated event, Or it has somehow failed to honor the music in that way we all agree It must, only no two of us can agree on just what defines that quality. Something disturbs the air. Itâ€™s not for me to say if the mysteries have been Well-served; in this ritual Iâ€™m the communicant who waits to see whether The goddess is pleased or not with the offering tossed on the fire. Sheâ€”youâ€”remain mute as the afternoon light when the bow is laid down. Nothing left now but the instant replay in the mind, and the question That cannot, in this silence, find a definitive answer: mystery Remains mystery, despite the point of order poked in the veil by these Tremulous notes that reverberate even now, in fingers that can remember. When I was a child I used to imagine shapes in a dish of strawberry jello: Red  was a threatening color, implying adventure, danger, the unknown. To the woman in the story of the tsarinaâ€™s shoes it implies the beautiful, But not to me at five: â€œThereâ€™s a lion in there, thereâ€™s a lion in there, â€œ I whispered. Now, in the heavy, questioning silence of after, in which the watch And the light might constitute either a mockery or a reply, I find a lion lurking as well in these wood-grain patterns of mystery, Waiting, indisputable, promising only that sheâ€™ll pounce in her own good time.                       November, 1995   	Bach in Mato Grosso 	    Randy guns the Fiat south Toward hills that donâ€™t want Visitors.  Unremembered hills Rise up, through steam unseen, But seeming still An echo of hills inside Yourself, Resonating silences From this noisy earth. The rain streams down: past Faded signs, locals plod Along the road, Oblivious and wet.  The hills Are too far away to walk. I clamp on my Walkman, talk Flagging in the heat, And sound, strangely confident With joy invades this  Pre-flood scenery.  The Second Brandenburg, That miracle which rose From Europeâ€™s own soul- Dampening drizzle, defies The jungle thunderheads, until We park the car beside the road To snap away at wildlife.  There I get into a stare-down with a jacare, immobile as Those thunderheads.  The agate Of his eye encapsulates The jungle afternoon; his perch On the swamp-bank brooks No back-talk from me. Suddenly, mosquitoes Are the roar inside  My head.  That, at least, I manage to subdue with Bach. Thatâ€™s not for him, though: Heâ€™s old as thunder, And doesnâ€™t blink.                      Pixaim, Brazil                    February, 1989      	International Hotel 	      Heavy-metal at breakfast: the waiters donâ€™t mind, But I do, and block it as well as Iâ€™m able, As I bury myself in my free New York Times And strain to hear whether, at the next table,  Portuguese, English or Hindi is spoken. 		Now itâ€™s the lobby, the bar or the street: 		I canâ€™t go upstairs yetâ€”the maids will have broken 		Into my room for the towels and sheets.  		The bellhops, the desk clerks, theyâ€™re all deferential; 		Their English is good, though they donâ€™t know your name. 		Here youâ€™re well cared-for, with all the essentials, 		And each night your room looks exactly the same.                                                    Sao Paulo, Brazil                                March, 1990    The Traffic in Sao Paulo  Two blocks, four blocks, sixâ€” Going to work is like watching Hamburger defrost, and in a din like a video gallery, Where the stakes are much higher Than fake planets or quarters. To miss by inches  Goes unnoticed here, And timetables should be posted On every street sign: Padre Joao Manuel to the Avenida Paulista, ETA 45 minutes, Give or take the eternity between Two shock-enjambed heartbeats, And the mopping-away of sweat. No one but cabbies should invade this whacked-out pinball game,  A slow-motion stampede under a Swamp-miasma of gasohol, Breath of a thousand year-old wino At the most neon hour of the night, And not even a cabbie, until Heâ€™s racked up enough confirmed kills To command the respect of his peers. Honking high noon: the luncheon crowd Sipping espresso at the CafÃ© Almanara Is treated to a weddingâ€™s worth  Of racket when someone stalls A derelict VW at the corner. O dear God noâ€”not a cloudburst, Not nowâ€¦ But it lets up in minutes,  And the streets become A steaming ice-rink in the  Late-November sun.  Now poise at the curbâ€™s edge,  Urban cliff-diver; donâ€™t slipâ€”   Youâ€™re not at home here; You wonâ€™t understand those Last-second curses, And the bus driver? He wonâ€™t even look back.               November, 1988  					 The Island Around the Corner  From the island around the corner, Where dawn is an act of willing, The vicissitudes of midnight Puncture the sailorâ€™s dreams. His unlikely dreams of landfall Satellite-photo that coastline, As time takes the face of icebergs, And the snapshot-clocks are stupid, And the vestals of morning stoke Beach fires that burn without light On the island around the corner.  On the island around the corner, Where the coiled springs all lie broken And the keys to the locks are misplaced Beyond cobweb-ripping light, No codebreaker holds bright vigil, Or boasts of the blueprints to sorrow, And the sailor who hears the waves breaking Wakes up to find only calm sea. No bells ring, nor are heard rising The appoggiaturas of a dawn breeze On the island around the corner.  On the island around the corner, Where the cliff-walls all face westward, And absorb the cries of sea-birds With their backs to the threat of day, The seed of the earthquake that threatens In the tossings of the dreamer Who calls up both island and sailor Shakes no tree, nor this earthâ€™s resolve. The rules have been set down in sorrow, And the treetops will brook no consoling On the island around the corner.                                     April, 2000     Closing the Consulate  				    			Sometimes whatâ€™s been forgotten can speak 			More eloquently than whatâ€™s been remembered. 			Four old, crumbling books, swiped from a box-- 			I canâ€™t imagine how theyâ€™ll ever be missed-- 			Paint pictures of a past no one thinks about now, 			One that might never have been there at all, 			Through glass tinted during an age and a time 			When things were built to last, though nowadays 			We have trouble even grasping such an idea.  			The closed-circuit camera facing the back door 			Makes you pause for a split second, but then 			You smile at yourself--itâ€™s as dead as the rest 			Of this building; no longer does a glassed-in 			Marine eyeball a monitor at the other end. 			Among a pile of trash in the hallway lies a transcript 			Of the speech the secretary made here last month, 			With German newspapers respectfully covering 			The last-time reeling-down of the American flag.  			Itâ€™s everywhere, that official talk, but more articulate  			Are the things now speaking that you know 			Will be forgotten: this afternoon light on the 			CASHIER sign, and on the hours posted for 			Passport business, the empty basement lockers, 			The now-meaningless â€œsecurity check sheetâ€ 			On the propped-open-with-a-brick bulletproof door, 			Which invites the world (which ignores the invitation) 			To watch as we haul an old refrigerator away.                            Stuttgart,                           September, 1996    Extremism In The Defense of Music Is No Vice  In memory of Leonard Bernstein   That which unifies also separates: â€œUniversal languageâ€ my assâ€”you should Have been there the night they booed Messaien.  Then again, you probably were, And whatever you thought or said that night, We should have heard.  Iâ€™ve read the clippings, Seen the video, sensed the question Behind the glance and the well-turned phrase.  Your polytonality was darkâ€”you played The useful idiot when it suited your key, But if anyone had bothered stopping to ask, You could have sorted out the sense for them Of Messaienâ€™s bird-logic, Stockhausenâ€™s way  Of juggling time, Or the hide-in-plain-sight  Wood-grain patterns of Carter.  Beethoven  Chords leave no room for argument; Mahlerâ€™s Appoggiaturas leave nothing but, and beyond Lies a shifting landscape: mountains, glens  And valleys you knew intimately enough  To keep us arguing amongst ourselves for years.  â€œWhatâ€™s this?â€ They say you said at the end. I only hope it was worth your attention, And not another variant of the Stupid Question, Questioning the ends: the ends are music, You said it yourself, with as much finality As it could be said.  The means, theyâ€™re fodder For Newsweek, a noisome racket that dies On the first tremor of a downbeatâ€™s stroke.   	                    Moscow,                   October, 1993    Late Winter Sketches 			 				Taxi Ride  Spinning wheels on the ice, and waving around whoeverâ€™s behind, the cabbie mutters.  Heâ€™s one of those whoâ€™s made a living room of the front seat: beads, sacred heart, pictures of the kidsâ€¦ while a cop watches from the corner, (his partner slipping into the 7-11) shadows in an SUV, rocking past over the deep snow ruts, just barely restrain (you can feel it) the bird. The door swings open,  and as his breath turns to steam, the cabbie lurches out into the street-lamp light, Stumbles around back, throws open the trunk, hauls out a sack of kitty litter and voilaâ€” instant traction.   Now weâ€™re underway again, beneath the eye of the cop who hasnâ€™t moved, and my watch, which I shake.  Itâ€™s stopped.  	Messy Divorce  Shuttle diplomacy, 		lawyers on the line, 		the child a sleepy lateral pass 		between icy sunset at Grandmaâ€™s 		and chilly dawn at the airportâ€¦ 		The telephone burbles in mid-anecdote, 		and suddenly her distracted eyes, 		like Supermanâ€™s, 		shoot past him, miles beyond the glass door 		beside the potted palmâ€” 		Interrupted, he has no choice 		but to make common cause 		with that cheerful door, 		whispering, smiling, nodding himself out. 		Ding-dong, happy as morning itself, 		it ushers him back 		to where the sign says â€œ5th floor.â€  			Gospel Remote   Whale-backs of winter, Bus-and-dog insulted, Die in the gutter While the music blares. (The season in its seed Changes hands, rising Against a dusk for weeks Pushing limits.) Salvation blasts Faces of brownstones, Ears of worn asphaltâ€” The four-lettered van, Ferryslipped at the curb Leaves no doubt Of what itâ€™s all about. (Someone inside is Dial-twiddling probably, While two shadowy figures On the railed stoop, Sway, talk-or-sing; Itâ€™s hard to tell which From here.)  The high- Ceilinged harmony of These city hallelujahs Hums the window screens Beneath risen panes, 	          Up for the first time,  	          Most likely, since these            Mud-spattered patrol cars, 	          Oozing around the corner,           Sullenly endured their threats            Under solsticial light.                     March, 2000  					  Etude  				The sky is friendly because it doesnâ€™t care. 				If the late, first-of-November wind 				Rips loose the screen from my front door, 				Itâ€™s not bad weather; I tighten the screws 				And turn back to what I was doing before. 				The branches whip like womenâ€™s hair. 				And yes, it might be that the wind wants an answer, 				And yes, there are poets who talk to the wind, 				But I think that Iâ€™m out of my depth in the air.                                November, 1999  '),(12123,3380464,'27,June,2004','2004-06-27 00:00:00','   I  This week my subscription to National Review magazine, ordered about two weeks ago, kicked in.   In yesterdayâ€™s mail I received NRâ€™s special commemorative edition on the recent passing of Ronald Reagan.   I was reading it last night in class, to prevent myself from falling asleep out of sheer boredom during the first of a monthsâ€™ worth of lectures on bankruptcy, the final section of a paralegal course Iâ€™m taking at the University of San Diego. Every essay in this issue is dedicated to some aspect of Reagan or his legacy.   I found something in one of the essays thought-provoking to say the least: I reflected on it in the car all the way home from school.   Among the authors who contributed encomiums to the late president was historian Paul Johnson, author of â€œModern Times,â€ one of the best books on 20th Century history I ever read, by the way.   Describing the situation Reagan inherited when he was first elected president in 1980, Johnson wrote:   â€œThe 1970s had seen a president forced to retire in disgrace, and an unelected president with no mandate, beaten in turn by a feeble Democrat from the south who had no obvious policy or coherent view of the world. In Washington, a triumphant but leaderless Congress usurped executive authority, allowing a triumphant Soviet Union, and its surrogates in Cuba and Vietnam, to do what they willed in Africa and Asia. Americaâ€™s apparent decline as a great power was symbolized, in a terrible moment early in 1980, by a shocking military fiasco in Iran.â€   I could add my own lugubrious memories of the 1970s to Johnsonâ€™s: gas lines. The Chevy Vega and the Ford Pinto. Stagflation. Unemployment soaring toward 10 percent. The U.N. transformed into a Third World debating society, with America blamed for every problem on earth. Mason Reese.   And Johnson didnâ€™t bother to mention, but I will, the spectacle of Americans clinging to helicopters to get out of Saigon as the situation in Vietnam finally collapsed on April 30, 1975. (On that same subject, I also remember the 1975 Academy Awards show on television, at which the anti-American film \"Hearts and Minds,\" about Vietnam, won the Oscar for Best Documentary. As if that werenâ€™t bad enough, when two of the hip-and-trendy left-wingers who had been involved in the production got up to accept the award, they crowed to an approving audience of how South Vietnam was about to be â€œliberated.â€ In Hollywood anyway, little has changed.)    As for roller disco, polyester leisure suits and big, ugly medallions, well, to paraphrase Mark Twain, weâ€™ll draw the curtain of charity on that.    Itâ€™s hardly surprising that Jimmy Carter, when he sought re-election as president in 1980, had to resort to desperate scare-tactics: Democrats pitched a vision of Reagan as a bug-eyed, right-wing maniac who would abolish Social Security with one breath and mash his thumb on the nuclear button with the next. There wasnâ€™t a single thing in his own record that Carter could point to as evidence that we should re-elect him, which left his campaign with no strategy except to demonize Reagan, which is all it did, and to no avail, because by 1980 America had clearly had enough.   In November, 1979, when our embassy in Iran was overrun, (it was just a month later, by the way, that Soviet tanks rolled into Afghanistan) things had been sliding steadily from bad to worse for six years, starting with the Watergate mess in â€™73, and there was very little expectation amongst Americans that they would ever get better.   Tom Wolfe famously called the 1970s the â€œMe\" decade, but the catch-phrase I remember much better from those years is â€œlowered expectations.â€ The Oscar for Best Picture of 1976 went to Sylvester Stalloneâ€™s \"Rocky,\" a movie in which a from-nowhere prizefighter is offered a miraculous shot at the heavyweight title, only to realize before the fight even occurs that he has no chance of winning, and announces that he just wants to â€œgo the distance,â€ in other words, heâ€™s willing to settle for second best. Willingness to settle for second best had never been an American core value, and perhaps \"Rocky\" reflected the eraâ€™s malaise in much the same way that the film which followed it as Best Picture the very next year, \"Star Wars,\" reflected a concomitant thirst for pure escapism. In 1933 Americans crowded into theaters to watch top-hatted Fred Astaire dance with slinky Ginger Rogers and thereby escape for a couple of hours the horrors of the Great Depression. In 1977 they crowded into theaters to watch Luke Skywalker dance with Darth Vader, probably for much the same reason. I remember well the fanfare that attended the appearance of \"Star Wars,\" the runaway summer hit of that year: â€œItâ€™s the return of...ENTERTAINMENT!â€ The reviews trumpeted, and all summer long, John Williamsâ€™ stirring orchestral prelude to the movie blared from radios all over the country. In the long view off the caboose of the train, all that hyperbole and fanfare seems a collective sigh of relief.   In short, the 1970s were a very bad decade for America, and Reagan, according to Johnson, affected an almost miraculous turnaround in the nationwide, and worldwide situation during his eight years in office. He restored our national confidence, stage-managed the destruction of the Soviet empire and brought America back to a role of pre-eminence on the world scene. The â€˜70s were the disease and Reagan was the cure.   But hereâ€™s what got me thinking: those godawful â€˜70s were also the decade in which I came of age. And despite all of the introspection Iâ€™ve been doing in my personal journals for more than 30 years now, I have never really made any serious attempt to come to grips with my own relationship to the era in which I grew up, and how it may have affected my entire life. Iâ€™ve written at length of how my father, who came of age during the Depression, let that traumatic experience shape his attitudes and behavior for the rest of his days, but I have written little or nothing of how my own experience of being a â€œâ€˜70s kidâ€ might have profoundly influenced the kind of man I became in the â€˜80s, â€˜90s and right up until today, as I write this at the age of 48.   Surely, some of the most important years of anyoneâ€™s life are the period between their teens and early twenties. Goals, dreams, ambitions and attitudes that will last a lifetime are forged between early adolescence and the time when youâ€™re launching yourself on the great world as a young adult. On January 1, 1970 I was 14 years old. On January 1, 1980, I was 24. I graduated from high school in 1973, from college in â€™77. If anyone can legitimately claim to be a product of the 1970s, itâ€™s those of us who were born around the end of the first Eisenhower administration, circa 1955. Our parents came of age in a time of national economic disaster and psychological pain, and we in turn came of age in a period of national economic lassitude and psychological numbness. (By the way, Iâ€™ve mentioned the fate of the Soviets a couple of times; itâ€™s a curious fact that Russians tend to view the 1970s in much the same way we do: the years of Leonid Brezhnev are referred to in post-Soviet Russia as â€œthe period of stagnation.â€)   Itâ€™s an accepted principle that our capitalist, free-market economy runs in cycles of boom and bust. Perhaps the national mood follows a similarly cyclical pattern. The malaise of the â€˜30s was corrected by World War II, which in turn ushered in a period of such confidence that many were speaking of an â€œAmerican Centuryâ€ beginning in 1945. Our sudden postwar affluence resulted in a burgeoning middle class whose satisfaction with its newly-found prosperity shaped the 1950s, a decade (perhaps unjustly) characterized as a spiritual and intellectual wasteland, whose contrary symbols were the gray flannel suit and the beatnik T-shirt that was a response to it. The confidence in their own and Americaâ€™s future which was instilled into babies born during the war and nourished by the optimism of the Kennedy years, led in turn to the waves of college-campus idealism of the early 1960s. That spirit promptly found fertile ground in the Civil Rights movement, and created a general confidence among the younger generation that it could change the world for the better, which was promptly dashed to pieces by the Vietnam war. In response to the war, the hippies, who were the younger descendants of the beats of a decade earlier, turned on, tuned in and dropped out as the country watched war, assassination and inner-city rioting explode all over its TV sets. My older sister once characterized the 1960s as a period when, for ten years, â€œthe whole country threw up.â€ Pop culture, so often a good reflection of the time that produces it, telescoped the experience of the decade quite neatly as it came to a close: Woodstock engendered Altamont in short order. In the words of one commentator, the counter-culture â€œwent from flower-power to death-tripping in a matter of months.â€   Tired of the constant turmoil and upheaval that the years of John F. Kennedy and Lyndon Baines Johnson had brought with them, in 1968 America turned to Richard Nixon, and the rest is not just history, itâ€™s my personal history, mine and everyone elseâ€™s who was born about the same time I was. Nixon was elected for his first term on Nov. 4, 1968. Reams have been written about what a tumultuous year â€™68 was; I wonâ€™t get into that here. The important thing is, on the night Nixon was elected, I had just turned 13. I was primed and ready to become a â€˜70s kid.    II  I donâ€™t even have to think very long or very hard to come up with an example of how I was sideswiped by the 1970s in a highly personal way.   As the decade began, I was a teenager with understandably little perspective on what was happening with the country or the world at large. We were Republicans and conservatives in my family; on the morning when Nixon was declared the winner of that very-close â€™68 campaign against Hubert Humphrey, my father got up and hugged my mother. It was good news in our household: after eight years of chaos ruddered by Democratic presidents, the country would now get back to normal. So we thought, anyway. I was of course aware of Vietnam as a child, and at our house we supported the war effort; as a Republican family we believed that the crusade against world Communism was rightly Americaâ€™s number-one priority. Social issues could wait until the Communist hydra had been slaughtered. Of course, as the war dragged on, even my parentsâ€™ perspective on it began to change. My father, as staunch a Republican as you could ask for, declared some time in the early â€˜70s that he was â€œBy-God beginning to understand how the young people in this country feelâ€ about the war. Clearly, even his patience with it was beginning to wear thin, and My Lai certainly didnâ€™t help, although my father was with those who felt that William Calley was just a patsy for the higher-ups. I had no inkling of it at the time of course, but I think now that, as 1973, the year in which I would turn 18 and become eligible for the draft approached, my father was beginning to worry about Vietnam getting its clutches on me.   As it turned out, I was spared by inches: the last American to be drafted into the armed forces was inducted on June 30 of that year. I turned 18 on October 12. Draft registration was still required of course, and I dutifully went over to city hall and filled out my draft form. I promptly received a notice in the mail that I had been classified â€œ1-H,â€ a holding category. They were no longer drafting anyone. Two years later the draft was formally abolished.   In that summer of 1975, I was visiting my oldest and dearest childhood friend, Jim Provenza, at his parents\' house in Camarillo, California. These days Jim is a middle-aged lawyer with college-age kids, but in those days he was a young firebrand of the left, ardently loyal to and active within the Democratic Party, who dreamed of becoming the next John F. Kennedy and changing the face of America. I\'m happy to report that just months ago, at age 48, Jim got himself elected to the Davis, California school board. Way to go, Jim. We\'re all proud of you. Hope you\'re enjoying the 16-hour days and the hate mail.   It was Jim who informed me that summer day that I no longer even needed to carry my draft card in my wallet.   We quickly organized a little ceremony. \"Hey, everybody!\" Jim called into the house. \"Come on out here! Kelley\'s going to burn his draft card!\" To the huge amusement of Jim and his family, I took the card out into the Provenzas\' driveway, got out my trusty Bic...and set that sucker aflame. Burning your draft card in 1970 would have gotten you tossed in jail, but by 1975 it was about as inflammatory an act (no pun intended) as saying â€œThe south will rise again.â€   My father not only changed his mind about the war, but as time went on, he changed his mind about Nixon. In fact, Richard Nixon accomplished something that I donâ€™t think anyone else on this earth possibly could have: he turned my father into a Democrat, albeit for a short time. By 1974, with Watergate clearly about to become Nixonâ€™s Waterloo, my father got so mad at Nixon that he went out and changed his voter registration to Democrat. He remained a Democrat for several years, in fact after his retirement from the Immigration Service, my father did a stint as an administrative assistant to a Democratic state senator in California. (That turned him back into a Republican.)   I had just finished my junior year of high school when the Watergate burglars broke in. Again, I watched TV news and read the papers now and then, so I knew what was going on in a general way, but I was too busy being a teenager to worry about it very much. In the summer of â€™72 I was more interested in ogling Olga Korbut, the excruciatingly adorable little Soviet gymnast who was the darling of the Munich Olympics, than I was in anything I was seeing in the papers about Nixon or McGovern. (I signed up to do campaign work for Nixon that fall, but again, being a teenager won out: I think I only showed up at campaign HQ one time, then lost interest.) Come to think of it, the same was true of such issues as the energy crisis and the gas lines of 1974; by then I was using the family Chevrolet to attend junior college, but my dad was paying for all the gas: what did I care if it had just reached the outrageous price of 50 cents a gallon? (I pumped gas in high school for $1.50 an hour. At that time, gas cost about 32 cents a gallon. When I tell that story to todayâ€™s twentysomethings, their jaws drop at BOTH numbers.)   By 1976 I had finished community college and transferred to San Diego State University to do my upper-division studies. I loved history and wanted to make that my major subject, but I could see that my father was uncomfortable with my getting a liberal-arts degree. It was the Old Story of â€œWhat can you do with that?â€ Many of my classmates were struggling with the same dilemma: the Class of â€™77 was loaded with Business Administration majors who would have preferred to be English majors but had to worry about finding a job. I had decided, early in college, not to major in English myself, though I had considered it. At 19, I allowed myself to be swayed by a purely romantic notion: I loved poetry and literature so much that I decided not to allow my love for these things to be poisoned by a lot of academic BS. Somehow I had the idea that a literary degree would take the fun out of reading Yeats, Shakespeare and Tolstoy, so I decided against it. History was a subject I also loved, but not with the level of passion I had for literature. Besides, history is all  about  scholarship: classrooms canâ€™t hurt that.   Still, I had to face the Old Story, and so I struck a compromise with my dad and with myself. Since the age of 16 I had never wanted to be anything but a writer, so I decided that an acceptable halfway measure would be to double-major, in history and journalism. I entered the College of Liberal Arts to pursue a major in history, and then walked across campus to the College of Professional Studies to pursue a major in journalism. I had no notion of becoming a history professor; I simply liked the subject. After college, I figured, I could go to work as a reporter on a newspaper or magazine. What the heck, my 21-year-old self figured. It was all just time-serving anyway, until I managed to explode upon the literary scene with my first big novel. Yeah, right.   Setting aside the nonsense about big (or small) novels, little did I know that my decision to major in journalism had put me on a collision course with history, or at least with cultural trends. Talk about bad timing. And Nixon, appropriately enough, lay at the bottom of it all.   I made the decision about going for a journalism major during my junior year at State: 1976. The year of \"Rocky.\"   But it was also the year of \"All The Presidentâ€™s Men.\" The movie version of Bob Woodward and Carl Bernsteinâ€™s adventures as Washington Post reporters bringing down Richard Nixon hit the silver screen that year, with rugged Robert Redford and handsome young Dustin Hoffman as their nowhere-near-as-good-looking real-life counterparts. (Why does Hollywood always do this, by the way? One thinks also of unbearably-glamorous revolutionaries Warren Beatty and Diane Keaton in \"Reds.\" In real life, John Reed and Louise Bryant were a couple of two-baggers.)   Some contemporary observer of social trends in the mid-1970s remarked that at the moment in the film when Robert Redford bats his beautiful orbs inquisitively and utters the line, â€œWhoâ€™s Chuck Colson?â€ 250,000 college students immediately ran out and changed their majors. Suddenly, with \"All The Presidentâ€™s Men,\" journalism became cool. Everybody wanted to be the hotshot investigative reporter who goes around pulling down the establishment. Little did any of them realize that this phenomenon, of two newspaper reporters taking on the President of the United States, was the ultimate one-shot deal: it had never happened before and in all likelihood will never happen again. Armies of nouveau journalism students who expected to graduate, go to work for the Washington Post and become the next Bob Woodward found themselves covering Rotary Club luncheons in Bayonne instead. But no matter. From the release of the film until the end of the decade and beyond, the nationâ€™s journalism schools were chock full of wannabe Woodwards and Bernsteins, and I, who had no Watergatey pretensions at all, (remember, my plan was to become Scott Fitzgerald, not Bob Woodward) found myself standing in a field swamped by heavy traffic. Everybody and his dog wanted to be a reporter, and every newspaper opening in the country had 50 people (some still fighting pimples) lining up to interview for it.   Needless to say, it took me a long time to get my foot in any sort of door. In fact, it wasnâ€™t until February, 1979, more than a year and a half after getting my B.A., that I managed to glom on to a tiny position with a tiny, independent news service in San Diego that was the very definition of â€œshoestring.â€ (In the meantime, I had worked at a series of minimum-wage jobs, including security guard and 7-Eleven clerk.) The County News Service of San Diego covered the city and county beats, and the courthouse, for subscriber weeklies countywide that did not have the resources to cover these beats for themselves.   How shoestring were we? The era was not only pre-Internet by more than a decade, but pre- computer  by maybe three or four years. I would cover a meeting of the County Board of Supervisors, then type out my stories on an ancient Smith-Corona portable. At the end of the day, whoeverâ€™s turn it was to do the mailing that week would gather everyoneâ€™s copy, determine how many copies needed to be made using a chart of our client list, then drive over to radio station KGB-FM, with whom we had a trade-off agreement: in return for tip service, they let us use their Xerox machine.   Once the copies had been made, the â€œmailerâ€ then had to drive them over to the main post office (not a branch, that would slow things down) and drop them in the mail to our clients. (Yes, we ran a news service using â€œsnail mail.â€) We had complimentary subscriptions to all the client newspapers, and once a month we would all get together with pencils and rulers and go through the tear sheets, measuring in inches how much of our copy they had used. We then billed the clients $1.00 per column inch. Each reporter got to keep 60 cents on the dollar for whatever we managed to get into print. The other 40 cents were set aside for overhead, basically envelopes and postage. No by-line, and 60 cents a column inch: that was payday. (We had a joke amongst ourselves: â€œWelcome to County News Service: 60 cents an inch and all the pride you can swallow.â€)   For the rest of that year, I went back and forth between home and downtown San Diego, by car, by moped and sometimes by bus, to put in eight-hour days for what was usually somewhere between $250 and $300 a month. To supplement my meager newspaper income, I went back to minimum-wage work as a part-time security guard, spending my Saturday and Sunday afternoons and evenings walking a beat at a local tuna cannery. My father, who initially threw up his hands in despair at my 60-cents-per-column-inch gig, went along with this: I think he understood that I viewed this as my last chance, at age 23, to get into the newspaper business, and if I were willing to work seven days a week to do it, that meant I was serious.   By December, (just about the time those Soviet tanks were rolling into Afghanistan) I managed to land my first real newspaper job, on The Imperial Valley Press, a daily newspaper of about 15,000 circulation in El Centro, California. The â€œWoodsteinâ€ crowd notwithstanding, I got the job through an acquaintance who covered the county beat for the now-defunct Escondido Times-Advocate, had worked on the Imperial Valley paper previously, and still knew the managing editor. That inside track helped, as did the fact that nobody in his right mind would want to live in the Imperial Valley, where the average high temperature between June and September is somewhere between 110 and 118. (For my European and Russian friends, thatâ€™s 43 to 47 Celsius.) But I wanted to be a journalist, so off I went.   And thatâ€™s where I was the night in 1980 when Reagan was elected: I was living in a one-bedroom apartment in El Centro, for which I was paying $180 a month rent out of the roughly $850 a month which was my full salary. (I started at $720, but the managing editor liked the cut of my jib, so he gave me a raise.)   And I wasnâ€™t expecting a whole lot more, which is my whole point in telling this story.   My father, who turned 20 in 1934, spent the rest of his life expecting the Depression to come back and pounce on him again at any moment, and he lived his life accordingly. He set his sights low, seldom dared to dream, and even when he did dream, for example of buying a farm, he talked himself out of it. He insisted, sometimes loudly, that job security and the guarantee of a retirement pension were the highest things anyone had a right to hope for. He worked 30 years for the Border Patrol and the Immigration Service, retired and spent his dotage sitting on the front porch watching the world go by, boasting of the size of his retirement income, but inwardly seething with resentment, convinced that life had somehow cheated him.   My outlook at 24 wasnâ€™t as extreme as my fatherâ€™s at the same age, for a couple of good reasons. For one, as bad as they were, the 1970s were not the 1930s, and for another, I had had the advantage of four years of college, something my father never got.   Still, like my father, I had come of age expecting little. The Nixon-Ford-Carter years were conducive to that.   Reagan and his team did, in eight years, gradually manage to turn things around. The change didnâ€™t come quickly, in fact 1982, the second year of Reaganâ€™s first administration, was a severe recession year, one in which I had the experience of being unemployedâ€”for me, the â€˜70s seemed to be continuing. But we all know that economic upturns and downturns both tend to lag several years behind changes in economic policy: the positive effects of â€œReaganomicsâ€ didnâ€™t really start to bear fruit until nearly the end of his first term. His successor, George Herbert Walker Bush, continued Reagan\'s economic policies into the 1990s, ushering in an era of prosperity not seen for half a century. The Clintonistas tried to take the credit for the 1990s economic boom, but it was not theirs to take. They bashed the memory of Reaganâ€™s administration while basking in the benefits of his economic legacy. (Later, Democrats tried to lay the blame for the 2000-2003 recession on George W. Bush, changing the subject when anyone happened to mention that it began in March, 2000, on Bill Clintonâ€™s watch.)   The change in attitude between my generation and the one that followed it is hard to ignore. When I talk to twentysomethings, and even thirtysomethings, nowadays, Iâ€™m amazed at how high their expectations are. Babies born after 1970 have grown into adults who, I would not be at all surprised, scratch their heads in bewilderment at the willingness of Rocky Balboa to settle for second best in 1976. I was a federal employee from 1985 until 1999, and until I left the federal work force and went back to the private sector, I didnâ€™t know what a corporate recruiter was. Iâ€™d never heard of one. Imagine my surprise when a corporate recruiter e-mailed me in 1999 and asked if I might be interested in a job writing for the marketing department of a custom software company. The idea that a company might hire people, and pay them salaries, simply to hire other people, was beyond anything in my experience. And yet when I talk to college students today, their expectation is that corporate recruiters will be looking for them upon graduation. Their expectation (happily unrealistic, even in todayâ€™s world) is that theyâ€™ll walk away from their college graduation ceremonies and be pulling down 75 thou a year the following week. In my first job after college, during the \"Star Wars\" summer of \'77, I was making $2.60 an hour as a night watchman in a Ventura, California lemon packing plant.   Talk about a generation gap. The war babies and their parents didnâ€™t see eye to eye over Vietnam, long hair, rock music and drugs. Now the generations may differ over rap, tattoos and body piercing, but I think thereâ€™s also a divergence on something fundamentally more important, namely, what they expect as regards the quality of life. I donâ€™t have any children, but friends my own age who do are amusedâ€”-and sometimes understandably annoyedâ€”-by the way their high school and college-age children seem to expect so much more coming out of the gate than we did. We â€˜70s kids, for the most part, expected to start out humble and slow. I donâ€™t say that to try and make us sound more virtuous than our progeny, it was simply a fact, a reflection of the era in which we grew up. Todayâ€™s college grads want it all, and they want it now. And they expect to get it, too.   It all came along too late for me, unfortunately. Iâ€™m getting close to 50, and the idea of even owning my own home is something I have only recently begun to think about. When I was young during the Carter years, interest rates stood at 21 percent. I assumed you had to have a huge pile of money to buy a house, and since I never raised a family, I never really thought I needed a house, so my thinking on that subject never changed. By contrast, I have a 26 year-old acquaintance who is aggressively buying and selling one house after another even as I write these words, claiming heâ€™s going to be â€œthe next Donald Trump, but a nice one.â€   A few weeks ago I interviewed an 82 year-old World War II veteran, a man who had been at Normandy, for a newspaper story. While he poo-poohâ€™ed Tom Brokawâ€™s â€œGreatest Generationâ€ applause for himself and his contemporaries, he nevertheless echoed Brokawâ€™s implied message about todayâ€™s young: â€œThey have too much stuff and they got it too damn easy,â€ he said with genial scorn. Well, the grumblings of the old about how soft the young have it are themselves as old as the pyramids and then some, Iâ€™m sure.   But after interviewing this friendly, loquacious old vet, and realizing that it was people his grandchildrenâ€™s age that he was talking about, I surprised myself by realizing that my attitudes arenâ€™t that much different from his. My 22 year-old nephew drives a teal 1994 Honda Civic, fully loaded and souped up for drag racing, (I hope the cops nail him) that his grandfather bought for him for $4,500. My niece, 21, has a late-model Saturn. The only time my father ever bought me a car, it was a â€™72 Chevy Luv pickup that cost $1,300, and I had to pay him back. That same nephew, by the way, is talking about opening up his own business as soon as he finishes college. (At the rate heâ€™s taking classes, this should be around 2023.) I saw a newspaper article a few days ago about students at an affluent high school in California whose parents reward them for getting good grades with Jeep Cherokees, Lexus sedans and the like.   Todayâ€™s kids have too damn much stuff, and they got it too easy.   Like it or not, Reagan lies behind all of this. When he was reshaping economic policy in the early 1980s, the lefty-liberal crowd alternately worked themselves into a state of high dudgeon over those mean old tax cuts, and when that didnâ€™t convince the country that Reagan was evil, they beat their breasts and moaned about how this non-compassionate old grinch was taking money from the poor, unleashing the forces of greed, glorifying selfishness, etc. etc. Yes, this non-compassionate cowboy who didnâ€™t love the poor was determined to change the warm, fuzzy, smiley-face policies that had given us 21 percent interest, 13 percent inflation and 10 percent unemployment. The bastard. Because Reagan was sworn in as president the year he was born, my nephew can now go around talking about opening up his own business when he finishes college.   Oh, and by the way, like many of his contemporaries, my nephew proudly sports a Che Guevara T-shirt. Times may change, and attitudes may change, but kids will always be kids; with the under-25 crowd, what feels good will always win out over what makes sense.   As for me, my editor at the newspaper, who is about the same age as me (and facing his fourth marriage) has his eye on a little fixer-upper over in National City. Three bedrooms, two baths. $350,000, as reasonable a price as youâ€™re going to find in San Diego County these days. He thinks I should take the plunge, too. But Iâ€™m not taking that old fartâ€™s word for anything. I think Iâ€™ll go and have a talk with my friend whoâ€™s planning to become the next Donald Trump. Heâ€™s 26, so Iâ€™m sure I can count on him to give me good, practical advice for the 2000â€™s.  '),(12124,3380464,'06,June,2004','2004-06-06 00:00:00','   In a newspaper column a few years ago, Andy Rooney wrote \"I don\'t know anyone who is good with a computer who uses it for much of anything except being good with a computer.  They understand the technology of information retrieval, but they don\'t have much interest in the information.  I am vice-versa.\" 	 It was something I\'d been trying to say for a long time, but was never able to put it quite that well.    It has seemed to me, since I bought my first PC about nine years ago, that in the midst of the so-called \"computer revolution,\" we\'ve not only lost sight of the forest for the trees, we\'re not even seeing the trees anymore because we\'ve gotten so absorbed in the landscaping.   You don\'t hear computer weenies talking about the information they access on the Internet, but only about the slicko hardware and software they used to access it, and how they\'re all lined up and ready to buy the next generation of slicko stuff due out next week.    	 Dare I suggest that this is all really just another proof of the old saw that women have been throwing around for years, \"Men never grow up, itâ€™s just that their toys get more expensive?\"     Dave Barry: \"Guys like neat stuff, and by \'neat\' I mean mechanical and unnecessarily complex.\"    Bingo.  This is why guys like my old friend Bill Barna, who works for Microsoft these days, can\'t stop buying more memory chips and Pentium upgrades and everything similar that he can get his hands on.  A computer is a really neat toy, and guys do like their toys.  And although women use computers too, they aren\'t quite as single-mindedly fixed upon them for their own sake.  They have other priorities.    Dave Barry again: \"...a woman who would cheerfully use the same computer for fifty-three years will rearrange her furniture on an almost weekly basis, sometimes in the dead of night.\"    	 I am typing this on a computer, an IBM Pentium computer which I acquired a few years ago as part of the package that came with a home-study course in desktop publishing.  Since I bought my first PC, a Packard-Bell 486, way back in 1995, it has long since become \"obsolete,\" in fact they stopped making 486\'s in the late 1990s.  Everything is getting faster and faster and cheaper and cheaper. My long-since obsolete 486 cost me $1,500 in 1995. An IBM like the one I\'m using now can be purchased at a used computer store for about $300.    	 This is the world of \"men and their toys\" run gleefully amuck: computer hardware purchased in June will be \"obsolete\" by Christmas.   But I, for one, am not running out to buy a nuclear-powered laptop with racing stripes and then crate this IBM up for the junkyard  as some guys would.  Despite women\'s relentless generalizations about men, which are sanctioned and encouraged by the sniffy hypocrisy of political correctness, not ALL men are into computing for the sake of computing.   Witness Andy Rooney, myself or my one-time friend, pianist Jesus R. Araiza, who, when I once asked if he were planning to get a computer, replied with a snort, \"A computer is just a very expensive pencil.\"   	 To a limited extent, he has a point.  I keep a computer around mainly for word-processing, and since my stone-age 486 served that function well enough, I didn\'t replace it until it went belly-up on me.  Until just the within last couple of years or so, I seldom used my on-line service for anything except e-mail.  I use it more now than I used to, because in my current line of work it\'s vital for research. But that brings us back to Andy Rooney\'s point: I have a limited understanding of the technology of information retrieval, as most of us do these days; it\'s a necessary part of modern life. But I remain with those who are more interested in the information.  '),(12125,3380464,'30,July,2004','2004-07-30 00:00:00','   An excerpt from Kelley\'s journal dated July 27, 2004:  Yesterday, for the first time in more than 30 years, I walked up and on to the grounds of Chula Vista High School.  I was on assignment to get a newspaper interview with my old homeroom teacher, Gary Chapman, who in the twilight of his career has been named principal at CVHS. Him, and also Wes Braddock, who has been tapped for the same job across town at Sweetwater High. Chapman has been in the Sweetwater Union High School District since 1969: 35 years. Braddockâ€™s been around roughly half that long.  When I first ambled into Chapman\'s classroom in September, 1970, a pimply kid on the eve of his 15th birthday, he was a greenhorn teacher with one year of experience under his belt. Now heâ€™s older and stouter, but heâ€™s still Chapman: I noted the level of jocular locker-room repartee that went back and forth between him and his friend Braddock. Chapman was always a jock, and heâ€™ll die a jock. When I was a kid, it caused friction between him and me, his perpetual gridiron-scented he-man swagger. Now I just find it amusing.  Chapman and I rode over to Sweetwater in his pickup truck, and I spent about an hour interviewing these two guys. Iâ€™ll file my story tomorrow morning, in time for this weekâ€™s edition.  I parked on K Street and just ambled up on to the campus. School was out for the day and there werenâ€™t very many people around. There were the old familiar sights: the senior lawn, (open in my day, fenced, gated and locked now) the band room, the ramp leading up to the administration building, at the head of which my friend Dan Ramet, one-time long-hair, rock-and-roll drummer and god-knows where now, and I stood to have our pictures taken for the school paper in 1973. The photo was occasioned by our partnership writing articles about music for the  Spartan Spectrum , the campus newspaper whose moniker had been my suggestion in campuswide balloting during my sophomore year. It was my pleasure to learn, two or three years ago, that after all these years the name is still being used: my own tiny legacy to a school with which my relationship, when I was a student there, was ambivalent at best. But that\'s one of the funny things about nostalgia: it tends to knock irony flat on its ass.  I entered the corridor of the Administration building, which I do not think I had visited since graduation day, 6/15/73, although I do recall going back to Chula once, during my first year of college, at the request of my former speech coach, the late Jack Nolen, to help out at a speech tournament. (I was supposed to be a judge, but I ended up sorting evaluation slips instead.)  There were the musty old stairs leading up to the second floor, where I suffered Consumer Math with Mr. Oliver during my sophomore year, and English 11 at the hands of Mr. Harris in 1971-72.  There was the nurseâ€™s office, where my best friend Randy Bendel and I used to hang out at lunch time. (Randy was killed in an automobile accident in 1977 at age 22. He was the second of his ill-starred family to go into Holy Cross Cemetery out in east San Diego. His younger sister Robin had preceded him by seven years, at the age of 14. Since Randy\'s death, his brother Fred and both of his parents have joined him and Robin there. The whole family had karma to burn, it would seem.)  There were the old counseling offices, now converted to other uses, where I was summoned in â€™72 to be told that Mr. Bill Weed, the choir director, wanted me taken out of the choir (I sang bass) for my rotten attitude. (I got back in a month later, but in the interim was in Mr. Chapmanâ€™s psychology class.)  The open foyer that used to greet visitors in my day has been walled up and fitted with lockable doorsâ€”the legacy of Columbine and other changes in society since my youth. But I went in there and noticed, on the walls, two glass cases housing the Spartan Hall of Fame: a rogueâ€™s gallery of distinguished grads. Needless to say, Iâ€™m not there, and neither is my lifelong pal, pianist extraordinaire and man-about-Manhattan Charles Berigan, (Class of \'74) but Beriganâ€™s old pal from Rotary Interact, Joe Schilling (also Class of â€™74) is there, as is San Diego County Supervisor Greg Cox (Class of â€™66) and numerous others.  But the distinguished grad whose middle-aged photo caught my eye, and set memory aflame, was a guy I never knew and think I only met once or twice: Bill Roush (Class of â€™70.)  When I was a student at CVHS, Bill Roushâ€™s photo hung on another wall, in another building, part of another rogueâ€™s gallery. He had been ASB president during his senior year, and there his sullen, unsmiling senior picture hung among those of every other ASB president at CVHS, going back to the schoolâ€™s founding in 1947. (If memory serves me, Chulaâ€™s first ASB president was some guy named Marvin Anderson. His was the only photo in which the subject was not wearing a bow tie, and he looked every bit as sullen as his successor from 22 years later.) I remember that distinct feature of Bill Roushâ€™s senior picture: some teenagers smile in their senior pictures and some manage to look either serious or contemplative (Iâ€™ve liked to think, down all the years, that in my senior picture I looked contemplative.) Roush, in his senior picture, looked downright glum.  Overachievers will be overachievers, and although Bill Roush graduated the spring before I started at CVHS, I was aware that he went on to UCLA, and now that I have seen his (smiling, years later) color photo in the Spartan Hall of Fame, I know that he went on eventually to become a professor of chemistry at the University of Michigan.  But who is he to me? Well, it so happened that when I was a sophomore at CVHS, the year after Bill Roush graduated, I met and fell madly in love with a girl named Roberta Hauk (Class of â€™71.) In those days I was capable being madly in love with three girls at a time. Anyway, the since-graduated-and-gone-off-to-UCLA Bill Roush was at that time Robertaâ€™s steady boyfriend. (Roberta was one of those girls who simply  had  to be the girlfriend of the ASB president; it was a status symbol to be dating a Big Man On The Campus.) Also, as speech squad mentor David Hirsch (Class of â€™71) explained to me one night at a motel in Claremont when he had accompanied the speech squad to an out-of-town tournament, he had seen other senior girls do precisely the same thing Roberta had done: claim a boy as their â€œsteadyâ€ who had already gone off to college, thereby exempting themselves from senior-year dating without having to lose face because of it. It should therefore come as no surprise that when Roberta graduated in 1971 and went off to UCLA herself, the first thing she did was dump Bill Roush like a sack of overripe turnips.  All of this would have mattered not one whit to me but for one thing: I was crazy about Roberta. But Bill was a hard guy to be jealous of. He was a shadowy figure for one thing, offstage all the time. The first time I saw him, at a festival of silent comedy films in the high school gym during the spring of my sophomore year, all I saw was the back of his head (and Roberta hugging and kissing him.) Later I saw him once or twice at Robertaâ€™s house. But he was sullen, unsocial, uncommunicative. He apparently just wanted to be alone with Roberta and didnâ€™t want anyone else around. Eventually she began using this against him, when she was setting the stage for his dumping. And he could see the way things were going: the next time I saw him at Robertaâ€™s house, he made an attempt to be friendly. But it was too late. His fate was sealed. Roberta moved on to other boyfriends, eventually marrying some guy named John Howard circa 1976, and thatâ€™s where she disappeared from my radar screen. Iâ€™ve never heard tale nor trace of her again. Howard Freelove, (Class of â€™71) my friend down in the Tia Juana River Valley, told me that he had seen her at their 30-year reunion in 2001, but didnâ€™t get a phone number or an address, so he doesnâ€™t know where sheâ€™s living these days, with whom, or what sheâ€™s doing.  After this all-too-literal plod down memory lane, I fully expected to have that awful recurring dream where Iâ€™m back in high school again, being summoned by the bell to face math, science and social studies, not to mention that ghastly regimentation all the livelong day that made high school such a trial for me. But, whatever her reasons, Mnemosyne decided to spare me that. And so to breakfast.  '),(12126,3380464,'08,August,2004','2004-08-08 00:00:00','   This morning, after my seven a.m. reverie over Harold Bloomâ€™s anthology of poetry in English, (A.E. Housman, Wilfred Owen, Edward Thomas, E.A. Robinson, Stephen Crane and Trumbull Stickney, leading up to Robert Frost, an old friend whom I will revisit tomorrow) I was sufficiently conscience-stung to go and dig up an old unfinished project: â€œ8:45 and Elsewhere,â€ the cycle of poems I began in the fall of 2001 as a response to the Sept. 11 attacks. It has been hanging fire since around New Yearâ€™s Day of 2002.  Iâ€™d hit the worst roadblock in all my years of writing poetry: Lyric XIV hung unfinished for lack of a last verse. I wanted a socko ending and could not come up with one, and thus has the poem hung for more than two and a half years.  I took the cowardâ€™s way out: to hell with socko endings. I rounded it out with a dying fall and then e-mailed the whole mess to Lucia.  It actually came rather quickly when it came. I pulled up the poem and sat here, stroking my chin and tinkering with the text on the computer screen for perhaps 15 or 20 minutes (while Handelâ€™s â€œWater Musicâ€ played on the Windows Media Player) before an idea formed itself of how I could construct a final verse to the poem, and once I had that, the working-out with pen and paper took scarcely five minutes. I wrote the last line first and went backwards.  I have advanced at least one theory about why I have written practically no poetry in the past three years. The â€œno museâ€ theory seems to stand up against the evidence: looking back, itâ€™s pretty clear that when thereâ€™s no woman in my life, as is the case now and has been for three years, I write nothing. Of course thatâ€™s an oversimplification, because some women have inspired me to write poetry while others have not. \"Consuelo\" yes, Jamie, no. Nadya, absolutely; my wife, absolutely not. Diane yes, (despite being fundamentally a moron) Tatiana no (despite being Russian.)  But thereâ€™s been something else at work besides lack of erotic fuel. Lassitude, surely: the old business of why bother when you have no audience to write for? But if you really wanted to do it, youâ€™d do it anyway, so thatâ€™s really just an excuse. No, I think thereâ€™s been some measure of the old adage about the centipede involved. The centipede was asked which leg he moved first when he started to walk, and was thus paralyzed. If he had to think about it, he couldnâ€™t do it. Iâ€™ve just thought too much about it.  And thereâ€™s a \"villain\" in the piece: Garrison Keillor.  Yes, Garrison Keillor, the middle-brow would-be popularizer of poetry who can be heard spouting it on NPR every morning for about five minutes. In one of his books, (I forget which) he has a funny anecdote about an adolescent wannabe poet who gets up every summer morning and, â€œafter reading a few pages of Walt Whitman to prime the pump,â€ gets going. Ha-ha. Although I havenâ€™t done anything quite that egregious since I was young myself, I will confess that quite often itâ€™s been the case that Iâ€™ll get started on a poem after reading someone elseâ€™s poetry.  But you see the problem here, donâ€™t you? Keillor made a joke of it, and I didnâ€™t want to be part of a joke. But thatâ€™s stupid. I donâ€™t have any particular respect for Garrison Keillor. Heâ€™s a buffoon, a midwesterner who successfully transplanted himself to New York City and, proud of that accomplishment, now dines out on making fun of the hicks back home who didnâ€™t manage to become a â€œsophisticateâ€ like he did. Am I to be influenced by Garrison \"Yukking-up-his-sleeve-at-his-own-origins\" Keillor, who tries to gain intellectual respectability by publishing an anthology of middle-brow verse with the aw-shucks title â€œGood Poems?â€ No. I am to be influenced by Harold Bloom, who publishes a magnificent anthology called â€œThe Best Poems of the English Language,â€ which I have been reading, cover-toward-cover, for more than a month, and am now on page 817, about to revisit Robert Frost, author of one of my favorite poems in all the world, but one which I would and will never consciously imitate: â€œCome In.â€ (Josef Brodskiâ€™s brilliant exegesis of this poem, published some 10 years ago in  The New Yorker , is something that I think should be dropped from airplanes.) I can assure Garrison Keillor that he will find nothing of Robert Frost, but perhaps a tiny dash of Whitman, in my â€œ8:45 and Elsewhereâ€ (which heâ€™ll never see in any case.) But what poet since the 1890s has been able to avoid Whitman entirely? Heâ€™s like the boulder in the road: you may walk around it, but thereâ€™s no not noticing it. Slavish imitation and influence are not the same thing, in fact â€œinfluenceâ€ is one of the mysteries that makes poetry exciting. So Iâ€™ll soldier on, despite Garrison Keillorâ€™s funny little joke about priming the pump. For the record, the dozen or so poems I read this morning while having my coffee had no influence that I can see on the final stanza of Lyric XIV. Perhaps years after Iâ€™m dead, some Internet poetry expert will dispute that. In the words of our president, bring it on.  '),(12127,3380464,'01,August,2004','2004-08-01 00:00:00','   \"8:45 and Elsewhere\" is a cycle of 14 poems that Kelley wrote, mostly during the fall of 2001. What they are \"about\" may be open to interpretation, but they were occasioned by the September 11 terrorist attacks. Call them Kelley\'s personal response to that unthinkable morning.    8:45 and Elsewhere a Journal Autumn, 2001    I  These twilit evenings, the twilightâ€™s intense: Looking out over the parking lotâ€™s frown, You feel beyond the trees a humming of sky Not like bees, but like a whisper of intent In a language that you canâ€™t quite place. Where is the morning when that certainty of sky, So blue that its blueness could not be imagined, Glimpsed before a flutter of pigeons flew off, Defined Tuesday morning, a somnolent clock? All of those Tuesdays have been defaced: Long shadows gap-tooth the grass near the fence.  II  The doings of the sun haunted years Of our long, nervous childhood, When the radio antennas Cranked, clawing at the sky; The songs that dreaded the sun Never dreamed a dread mundane Could ignite a billion fears.  Late in the afternoon, As the talking heads buzzed on, My thumb spoke up for silence. A drive down to the reservoir Seemed like the thing to do. Ducks moved in rippling water, Small fishes swam in synchro, The quiet itself spoke eloquently Of smoke on three horizons: The lake fire-danced with sun.  Then bright night, humming loud: White noise drowned out the stars; Sense came down with the cities. Mondayâ€™s clear-text: Tuesdayâ€™s garble. In half a million tons there smoked Remnants of a collective reach, 100-plus channels of helplessness Adding speechlessness to speech. Groping for words from long ago, I think of the belly of a cloud.  The view from the Narrows: Yellow fog drifts over the battery, Then, on the streets of Brooklyn A noisome odor leers. Our job now is to reclaim daylight; Never mind night: that remains The mystery it will always be. But in the afternoon on Canal Street, Language attemped to regroup, Hushed, multi-tongued, in sorrow.  Since Timeâ€™s murderous birth, Whose echo we can still make out, Itâ€™s been obvious to a handful Just what our inheritance is, And now itâ€™s come to consciousness, Not like the sun, but like nonsense Amplified a billion times. This is the search for keys in a downpour, Thought hindered by hammering rain, One more questioning of the earth.  III  A full moon on Halloween. Wonâ€™t be another for 19 years. But it shows up this year, A stage-prop seeking a stage. Yes, thereâ€™s some movement On the streets: little shadows, Clutching bags, shepherded As usual by chatting adults With flashlights, And I hurry Home through the dark, As the wind obligingly Scrapes the street with the dry leaves It shepherds along, Then break out the candy, Pour it in a bowl, and set it By the door to wait. 8:15. Everyoneâ€™s gone now; My doorbellâ€™s only rung onceâ€” The boy next door: his parents trust meâ€”so the candyâ€™s poured Back into the bag, and tomorrow The crowd at the office Will dispose of it, Iâ€™m sure. Now, channel-surfing takes me past The World Series, The Munsters And chattering of anthrax. A full moon on Halloween. Wonâ€™t be another for 19 years.  IV  The gyroscope spinning in the radio tried to right itself As you drove along through traffic this afternoon: The lead story, when the news came on at two, Was the governmentâ€™s done deal with Microsoft. Ah, for a moment there it was old times again: Judges sternly judging, lawyers prevaricating, the prating press. But then the voice moved on to other business: The FBI picked up some telephone calls, And the National Guard was patrolling bridges in California. About that time, the left-turn signal said it was all right to go, And as you pulled back into the office parking lot, You remembered a long-ago collision on the Oakland Bay bridge. The Maverick was totaled; beyond that it was no big deal.  V The Nostradamus industry Has set up shop again. He comes and goes in grocery stores: You see him now and then.  But now heâ€™s on the world circuit Where you canâ€™t see his eyes, And someoneâ€™s claiming he foresaw Fire, raining from the skies.  Fireâ€™s the perfect metaphor For what we encounter next, Before someone corroborates The words of the old text.  Doom brushfires its way around The corridors of space As quickly as the Internet Can make a rumor race.  And by the time that common sense Finally regains the floor, Someone claims Old Scratch himself Appeared amidst the roar.  Itâ€™s true that when things fall apart And the center doesnâ€™t hold, We often go seek order in The awesome and the old.  But must we, even though we must Desperately search for clues, So lose our heads we start to trust In the  Weekly World News ?  VI  One more time, As so often before, Autumn canâ€™t seem To make up its mind. One morning the glass Has to be scraped Before you can move, Then you find yourself Impatiently peeling An unneeded parka As you step outside.  I once wrote of Indian summer, That it seemed as if Timeâ€™s teeth were on edge. How much truer Does that seem now. Less irony around These days (so they say), But Iâ€™m not so sure. These Indian summer mornings Are too much the one Thereâ€™s no escaping, That when the crackle Of strange blue-sky Thunder tore a jagged hole In the face of every clock, And everything froze Into frames that burn With each rewinding, And donâ€™t fade, but grow Sharper with recall, Like the heartbreak That a lover keeps Re-playing, and re-playing.  One wishes the fall Would get on with it, Turn that last key, Clear irony from the air, And bring on the certainties Of winter, whose shades Are a more fitting mirror Of this collective moment. The poets who love Autumn leaves, yes, Theyâ€™ll be here again, And weâ€™ll be glad, But the time isnâ€™t now. I feel, walking along In the blue afternoon With the trees their annual Shout of orange, yellow And red, that this isnâ€™t The time for bright autumn: The thrush music, The dusk outside, Oddly enough, Are holding us back From burying our dead.  VII   Architecture is an act of optimism.  --The New York Times  One Sunday evening in Paris June, Sprawling on a stone bench, half-drunk, (Beer and jet-lag working in tandem) I ogled the west font of Notre Dame And thought...nothing. Sometimes itâ€™s best Just to take a sublime moment for granted, And I was too tired to think in any case. This I once heard said about a great cathedral: That it was conceived as a smile. If it faced the sunset, seeking its own end, Well, that was only part of the scheme of things, And just one more reason to be glad. These also were begun and finished With a smile, though of a different kind: Not arrogance, no, but a confidence That was not common in that era. They arose against a racket so intense That some wondered if it would ever subside. But it did, and on we went, safe in the knowledge Of who our friends and enemies were. And now Notre Dame and that cloudy sunset Lie across a space that in all likelihood Will never again be crossed: the beer-haze Of that summer night gave way in turn To the morning no amount of caffeine Could make sense of, or liquor dull the edge, And in fact, as I watched the split-screen (live) The moment the west font awaited came to mind, And wasnâ€™t quite so easily dismissed As one might normally have expected. It was that kind of morning, As it had been that kind of night.  VIII  In every film Iâ€™ve ever seen That was worth seeing More than once, There comes a defining Moment: Bogey catches Bergmanâ€™s tear as it falls; Henry Fonda heaves The captainâ€™s palm tree Over the side; Gable, eyes cocked Like pistols, freezeframed Forever, resolutely Doesnâ€™t give a damn; Newman and Redford, Plunging 100 feet Into swirling water To the cry of â€œAh, shiiiiiiit!â€ But the movie moment That keeps playing back These dew-wet mornings Is that flock of pigeons. Yes, those pigeons, Those black-and-white  Fail-Safe  pigeons, (Fonda again, â€™64) Fluttering en masse Into New York Morning sky, Like the clattering, Rippling-down timetables In the older airports, Those venetian blinds of Arrival and departure. The pigeons Last a second or two, And then a fusillade Of frozen city sights Leaves you on your own With the blow-back in the mind That you know comes next. Clearer than the implied Soviet missile following Is this: in full Color, (reality doesnâ€™t Usually bother with The touch of Black-and-white) At 8:44:52, somewhere In lower Manhattan, A flock of pigeons Scrambled for the sky, And kept on going, But didnâ€™t fade to black.    IX  An Airbus went down in Rockaway last week. Horror was called for and duly received, But also a collective breath-holding: The stock market nose-dived, and once again We took up our spots to watch CNN And wait for the verdict. Hours went by And as Dominican families wept, (The doomed jet was Santo Domingo-bound) A more general dread filled the quiet. The two flight recorders were quickly found, And the NTSB sent out the word: â€œMechanical failure most likely cause.â€ So. 62 days proved to be too soon. The Dow-Jones took off, screaming for the moon.    X  The turkey carcass is a skeletal wreck, (â€œMy fowl was soupbones,â€ the poet wrote, Conjuring a nightmare of another sort) And Macyâ€™s parade having echoed away, --Uneventful, thank God, as we secretly sayâ€” Cable TV has wasted no time In trotting out the classics it no doubt hopes Will put everyone in a shopping mood. So, as the soap-bubbles tick in the sink, We settle down, sated, not full, as we say, And TNT dishes up  Home Alone , The Christmas number about the little boy Left behind by his family, besieged by burglars, Who outwits the bad guys and makes them pay. Outside, traffic is sparse on the road: A taxi lurches past, and an SUV Whose blackened windows flash, anonymous Under the glance of a gas-station awning As inside, a Pakistani immigrant counts change, And two kids, one with a ring in his nose, Skulk to the counter to buy frozen burritos, Skin slightly green in commercial light. The turkey is ravaged. Itâ€™s Thanksgiving night. The paper still lies on the lawn from this morning, An ignored intruder from ordinary time Bound to re-assert itself come tomorrow. But for now, plastic-bagged, it patiently waits, With everything it promises to contain, For when someone stoops, dew soaking their shoes, On the way to the car, in a ritual As automatic, if not quite as dread-free On this blue-sky morning as it was on that. Meanwhile, we lean back, not full as we say, And laugh in a circle one more time At the little boy, besieged by burglars, Who outwits the bad guys and makes them pay.     XI  The armchair in the corner keeps its back to the window. Outside, Indian summer presses on, and on. Lately, it might have been the storied place Where one looking quietly toward the tall bookcase Contemplated those big events that were none of his doing Or concern; the undusted TV screen, And the spiderweb up near the ceiling Were artifacts of an afternoon long and bemused. The impression still visible could have been any back, Someone watching, barely moving, outrage beyond thought, Wondering as the carping and shouting went on Not so much about them, but of things closer to home, And then, with a stretch and a yawn perhaps, A reminder to remember to take a glance At how the 401(k) might be performing, And to drop off the car on the way in next morning. But now the TV screen, dusted off, has hummed to life, And an old man in a peaked cap is quietly recalling A horribly unquiet moment, 60 years ago, When all eyes were suddenly and cruelly forced outward, And as heâ€™s interviewed, under blue morning sky, The editing cuts to that all-too-familiar piece of film: The teetering tower, black smoke, Arizona listing, Outrage ripping the skin off a long and troubled sleep. (I had a neighbor who was on one of those ships; He dove into the harbor, swam as fast as he could, Eventually came to a sewer pipe, and swam on ahead: â€œI wasnâ€™t going to stop until I came to an asshole,â€ he said.) The old man in the peaked cap is magnanimous: â€œThereâ€™s no comparison. It was our job to be in harmâ€™s way.â€ This is traditionally the time of year when demons come visit, And it might be argued once again that our grandparents Werenâ€™t quite the fools we liked to think they were: Suddenly, the gathering and the pressure to be â€œexcessiveâ€ Seems to make a kind of sense: weeks ago we were told Of a sudden surge in â€œhooking upâ€ and dessert consumption, And now, huddling again in the early and abrupt dark, The only thing differentiating this December from last Is a general nervous pulse-checking, an extra glance Where one glance would have been plenty before. The armchair, its back to the window, faces the door. The spider web near the ceiling has been cleared away. Someone may have just left, and may be returning soon, But whoever was here a year ago has left for good.  XII  Jogging under the aegis of a murder of crows, You recall the courier who survived two plane crashes Back when we took it for granted that such things were Accidents. The white envelope atop the dresser Has been looming there for weeks; louring, one might say, Like the ceiling, or this December sky. You know whatâ€™s inside, having examined it repeatedly. Banal as winter daylight, computer-spat, number-riddled, Perforated along one edge, subject to conditions of contract, The sort of threat you would never perceive as such In other times. You who were never superstititous a day In your life, are now amazed to notice things You never would have bothered to notice before: Baltimore/Washington is harmless enough a POD, But did the plane-change have to be Houston George Bush? (And is this the sort of thing, you ask in quiet alarm, That such as they would notice in plotting their next move?) The obligatory annual homing forces a crisis: Will you, when the chips are down, actually find the courage To go ahead with rituals as youâ€™re being urged to do, And heed the pressure from your family to fly west, Just as you did last year without thinking twice? Out there, the table is set and the kitchen warmly illuminated, Like the tree that confronts the dining room window, Smiling at the street, as it has every year since Vietnam. But you wish, in ordinary time like you never did before, That the tesseract which you remember from great childhood Could become real long enough to swoosh you through this. A dozen time-zones and two satellite-skips away, Theyâ€™re clearing out caves, chasing shadows across borders, While closer to home, the old music makes a game attempt To play again, and the most risible of the doomsayers Seem for now to be those discussing WalMartâ€™s bottom line. Still, you clamp down the hatchback over your carry-on With a bit more queasiness than when air was more benign. Even now, the red sign proclaims no satellite parking. The Gold Lotâ€™s where you go, stacked out in the weeds. Then the packed bus, lurching every foot or two, inching, And in the tension that accompanies the tightness Of clamping your carry-on between your knees, You want to reach over and smack the old lady, three down Yammering into a cellular phone for all to hear. Once youâ€™ve run the bottlenecked gauntlet out of hell, Bossy six-dollar-an-hour help shouting this way and that, (â€œA little authority is a wonderful thing,â€ grumbles one victim) You slip into the gateside menâ€™s room for a slug of resolve, Edgy enough to curse when the stall latch resists closing, Then, facing the bowl standing up, self-conscious, afraid, You see where someoneâ€™s pencilled a message in the grouting: â€œThere is life after September 11â€”move on.â€ One more swallow and itâ€™s there: the backbone to face Jet-blast mobilizing the stroll of the tarmac crows.  XIII  Gathering around and among lights Scattering, doomed: What better return of that familiar Nesting, solsticial light Could possibly be imagined for now? But there it is: only days after Weâ€™ve done it all one more time, The glistening ham taken Crackling from the oven, The family showing up Just late enough to be annoying, Astaire and Crosby dancing across The flickering screen in black-and-white, Bing putting down his pipe to sing  White Christmas,  his hands Fluttering over the keyboard as if Trying to shake drops of water From his fingernails, The menfolk sharing a shot of Glenfiddich Before the table, (always reminiscent Of that awful scene in Joyce) The circling of chairs, then the clearing-away Of all that multicolored, crinkling Paper and ribbon, stuffed efficiently Into a Glad green trash bag For the intrusion of Wednesday morning, Comes this. It breaks, at six Oâ€™clock, Upon a string of open-ended triumphs That may or may not (no one seems sure) Lead to anything like An unqualified conclusion, That the universe may indeed be eternal, But itâ€™s beginning to look like weâ€™re not. The number-crunchers, re-writing The playbook of endgame one more time, Now say the expansion may Go on forever, picking up speed, Not slowing down, (their last good guess) Until, at some time, ten To the string-your-zeroes-here moment From now, Thereâ€™ll be nothing but a darkness, Sublime or not, it makes no difference, As there will be no eyes, evolved or not To look upon it and find it sublime. The object of our mammoth quest Has skedaddled into the hills; Reassuring voices say The future and the answer will be found, But the heavens as always Have nothing to add, As they had nothing to add All those other times either, And what kind of maniac Would be fool enough To think they had? This dead-in-four-billion-years Chunk of ground My fraying Nikes press against As I haul the Glad sack down the driveway, The spiked eggnog still swimming, Fatty and comforting in my blood, Is going with its progenitor, Which just dropped down redly Behind the eucalyptus trees, As symmetrical shenanigans At the galaxyâ€™s core, Patiently, unhurriedly And without purpose, Go about mapping their final score, A new alphabet For all the holy books ever written.      XIV  A quarter past nine: the mirror sheen Shimmers where the candle gutters. New Yearâ€™s Eve, and I settle back Unpacked, swirl the cubes (they clink) And as cable TV once again utters Those commonplaces we hear each year, I join in progress an old favorite: Rod Taylor in  The Time Machine.   I saw it first when I was a boy: A Saturday afternoon in â€˜62: Kids crowded into the school Cafeteria to sit on wooden benches, And marvel, as daylight poked the corners, And the old projector rattled, threatening, At the rickety screen, flanked by the flags Of states none would ever destroy.  Science fiction with its cautionary eye: Oh, he learns a lesson all right, Fast-forwards to high-tech horror, Then dogs on to another kind, Careening through a night of night Toward Weena, and those gargoyles Gnawing on the bones of the innocent Who never bothered asking why.  That world, we know, was where they believed. Thatâ€™s the joke weâ€™re supposed to get tonight. Complacent in their overstuffed chairs, They ask friendly questions of the poor soul; Of course we feel itâ€™s proper and right For the smugness to be blown off their clocks. Okay, he oversteps himself like Phaeton, But in the end, gets timeâ€™s reprieve.  There will be no snowfall tonight. Winterâ€™s staying away with a will When itâ€™s most called-for. Socking over The kitchen tiles to fix another drink, Rounding out a season well-defiled And crying to be buried in something, I finger-spread the slats and look up: Quietly insolent, Orionâ€™s bright.  Wheels are turning, youâ€™d like to think. (Or would that really be such a good thing?) Two years ago, at 1000 mph, What passed for the millennium came in. Around that world a knife-edge flew: Towers exploded and rivers burned, But the real catastrophe stayed its hand. No round-number tonight, just this drink.  There are no firecrackers to be heard As AMC cuts to a commercial break: Some pitchman or woman for Dell Around the corner is selling speed. But it seems we have too much already, As my cheap Lorus hums the chapters on, And I, befuddled under Orionâ€™s sword, Stand swirling cubes, waiting for a word.  2001-2004  '),(12128,3380464,'01,August,2004','2004-08-01 00:00:00','   An American wandering across his own country...what more archetypal image is there in our national consciousness? It\'s Kelley\'s feeling that every American should cross the United States at least once by some conveyance other than air.  Hopefully, however, the circumstances of their crossing will be better than those of the one Kelley made in the early spring of 2003.  Twice-unemployed in 2002, by the following winter, Kelley\'s luck had run out, as well as his unemployment checks. Unable to find employment in the suburban Washington, D.C. area where he was living at the time, Kelley accepted an invitation from his half-sister in Reno, Nevada to clear the decks, load up his car and come back out west.  So he did that, jettissoning along the way a library of more than a thousand books, records and CDs collected over half a lifetime, (donated to the local library) and an apartment full of furniture (Salvation Army.) On March 27, 2003 Kelley emptied out and cleaned the apartment, turned in the key, spent that night with a friend in Arlington, Va. and the next morning was on the road.  Over the next two months, he would gradually make his way from suburban Washington, D.C. back to his family\'s ancestral home on the west coast, with a two-month stopover in Reno, Nevada on the way.  These are the notes he kept on the journey. The \"Blue Shoe\" referred to in the text was Kelley\'s car, a 1990 Geo Storm that served him well from the moment he bought it in Springfield, Va. in 1994 until the moment he gave it away in Chula Vista, Calif. some nine years later, at the tail end of the extended journey described in these notes.     March 29, 2003       Kansas City, Kansas   My westward trek from six monthsâ€™ unemployment to its continuation elsewhere began yesterday morning.  On Thursday, (3/27) I cleaned out my apartment in Silver Spring, continuing to throw things away even as I was making final trips to the storage facility in Wheaton where about 1,000 lbs. of my things will now sit until further notice.  Spent my last night in the D.C. area at Karen Winterbottomâ€™s place. Karen served a supper of broiled salmon with pasta and salad, and I met her son Raymond, whose situation at age 31 isnâ€™t so much different from mine. He was living in Chicago, but his professional and personal lives both jumped the tracks, and heâ€™s now living with his mother in Arlington. looking for a job. In my cups after supper, I lost to him at Trivial Pursuit.  Left Karenâ€™s place Friday morning at about 9:30. In eight hours I drove 500 miles, and by 7:30 was cruising through a blinding, cold rainstorm back into Indianapolis, a city which I swore less than a month ago that I never wanted to see again. I had intended to route myself around Indy and take I-80 across northern Indiana and Chicago and across Iowa. But Karenâ€™s son persuaded me that it would make more sense to take I-68 to I-70 and avoid the Pennsylvania Turnpike and I-80, both of which are toll roads. â€œYouâ€™ll save $30 in tolls,â€ he said. Since Iâ€™d decided to take Rayâ€™s advice and go the I-70 route, which would take me through Indianapolis anyway, I decided I would try to look up Nancy, my little friend from the front desk of the Holiday Inn Southeast in Indy during my ill-fated stay there last month when I attended that horrific truck driving school. I thought Iâ€™d see if she could put me up for the night and save me the fifty bucks that a motel room would cost. On the phone a couple of weeks ago, she said sheâ€™d be glad to.  I found her, now working at a different Holiday Inn, the one in Greenfield, just east of town. I went there and waited around for her 11 p.m. quitting time, which became midnight when her relief came in late. To pass the time, I drank beer and ate fish nâ€™ chips at a nearby Applebyâ€™s, and watched  The Golden Girls  on cable TV in the hotel lobby. By midnight I was tired enough to begin wondering if it wouldnâ€™t have been simpler and more pleasant to just forget about saving fifty bucks and check into a motel somewhere and forget about staying at Nancyâ€™s place. As it turned out, that might not have been a bad idea.  We left at midnight, in a cold rain, and I followed her back to her house, a ratâ€™s nest of trash and clutter in the midst of which her daughter was sleeping on the couch in front of a 52-inch TV set with her boyfriend. Nancy and I went upstairs, and even though it was now 12:30 a.m. and she had to get up at 6:00, (which meant I did too) she proceeded to watch some idiotic country-music awards show on TV, so I had to watch it too. I was restless for the road despite fatigue and only slept a couple of hours, and so was still dead-tired when Nancyâ€™s kitschy alarm clock began an electronic cock-a-doodle-doo at 6:05. But it was definitely time to get moving.  After a stop at McDonaldâ€™s for coffee, she left for work and I hit the road. I took 465 East back to 70 and struck out in a westerly direction. Poring over a road map last night in the Holiday Inn lobby, I decided to just make I-70 my home for the next couple of days. It goes all the way to central Utah, and it runs through Denver, a city I would like to see, or at least stop at on my way west (yes, my readings of Jack Kerouac did influence that decision.) So this morning I took I-70 from Indianapolis through Terre Haute and across the state line into Illinois, where I had to set my watch back.  Central Illinois was pretty much what I expected: an endless panorama of flat farmland, barns, grain elevators, farmhouses and more of the same. A radio station from one of the smaller towns amused me by including local obituaries in the news.  By 11:00 a.m. I had crossed the Missouri River and entered St. Louis. I drove right past the giant arch, the â€œGateway to the West,â€ St. Louis\' most recognizable landmark, which faces east on the western bank of the Missouri. By this time I was listening to a St. Louis jazz station, and the landscape shifted abruptly from agrarian vistas to towns and cities, some of which had curious names: Pine Lawn, Kingdom City, Odessa. Kansas City was my objective, 500 miles seeming a reasonable distance to make the dayâ€™s goal when youâ€™re not especially in a hurry and in any case have only had a couple of hoursâ€™ sleep.  But what with having set my watch on Central Time earlier, Iâ€™d picked up an hour, and when I reached Kansas City it was only about 2:00 p.m., which seemed kind of early to be stopping, but I was tired, and since Iâ€™m not humping any particular deadline, I decided to give myself a breather. I cruised through Kansas City, keeping an eye open for a motel that didnâ€™t look too seedy, and the next thing I knew I had crossed the river into Kansas. I checked into this place, the American Motel, which has a red-white-and-blue motif and is reasonably cheap and clean. Bathed, changed clothes, had a sandwich and a beer, took a nap.  Tomorrowâ€™s objective is Denver, which according to last nightâ€™s road map lies exactly 608 miles west of Kansas City. The Blue Shoe, with 132,000 miles on her, has performed well so far. As of yesterday morning when I crossed from Virginia into Maryland, I have driven 1,157 miles. Despite the pluck that the Shoe has shown so far, and despite her meanderings up and down the east coast and along the Autobahn in Germany, she has never encountered anything like the Rocky Mountains. I wonder how sheâ€™ll do.   March 30       Denver   Well, eastern Colorado was a big disappointment.  After leaving Kansas City at 8:45 this morning, I spent the next 6 1â„2 hours getting from one end of Kansas to the other, finally crossing into Colorado at 3:15 p.m. (and turning my watch back again, now to Mountain Time.)  Kansas was the dullest landscape I have ever seen in all my days, basically 450 miles of nothing to look at. Scary, too, because unlike Illinois and Missouri, where there was another gas station or another town every few miles, Kansas features vast stretches of distance, mile after mile, in which there is nothing, and I mean nothing, outside of farms, grain elevators and trackless waste. I was so afraid of running out of gas, (the Shoe only holds ten gallons) that between Kansas City and Limon, Colorado, I stopped and topped off the tank three times.  Breakfast was Chinese food leftover from last night in Kansas City, eaten at one of these refueling stops somewhere between Topeka and Abilene. Lunch was a tuna sandwich and coffee at my last refueling stop in Limon, about 90 minutes east of Denver.  Geographic serendipity: we non-Coloradans, whose acquaintance with the Rocky Mountain State derives chiefly from the John Denver songs weâ€™ve heard on Boomer-102 FM, expect the entire state to look like a Coors ad. Nope: eastern Colorado is essentially a continuation of Kansas, albeit one that continues a long, slow climb begun somewhere in those trackless wastes that took up much of my day. In fact, eastern Colorado continues to look like Kansas until youâ€™re about an hour from Denver, and the rise in elevation is so gradual that you donâ€™t even notice youâ€™re climbing. Once on the Colorado side of the state line, I cruised through a tiny town called Arriba, and when I saw the sign that read, â€œWelcome to Arriba, elevation 5,228 feet,â€ I couldnâ€™t believe it. â€œYou canâ€™t be serious,â€ I thought. â€œHere Iâ€™ve been driving through nothing but flat farmland all day, and now suddenly Iâ€™m up over 5,000 feet? When did that happen?â€  Closer to Denver, the topography began to resemble the Laguna Mountains of eastern San Diego County that Iâ€™m familiar with, up around Julian: brown, rolling hills, scrub, a few patches of snow from last weekâ€™s big snowstorm. Not until you have almost arrived in Denver itself do you see actual mountains on the horizon.  Denver itself lies on a flat plain at the foot of those mountains and obviously marks the place where eastern Colorado ends.  Knowing nothing of Denver, but noticing as I came in on I-70 West that Peoria Avenue seemed to be â€œthe motel strip,â€ I pulled off there and checked into the Super 8, sticking to my budget. But this is obviously a cheesy part of town, and if I had noticed before checking in that the motel is only half a block from the railroad tracks, I probably would have cruised onâ€”Iâ€™m going to be listening to those trains all night. But Iâ€™m checked in now, so to hell with it: I have whiskey.  Today I drove about 600 miles once again, in about nine hours, fifteen minutes, with two brief stops for eating. I have driven 1,756 miles since Friday morning. Called both Lucia and Karen to let them know I had gotten as far as Denver; last night I called my half-sister Madelon and her husband Bill in Reno, my destination, to let them know Iâ€™d gotten as far as Kansas City. Madelon wasnâ€™t home, but I talked with Bill for a few minutes. He reminded me that Iâ€™m not in a hurry, and advised me, should I run into bad weather, to just â€œpull off the road and wait.â€ But so far, the only bad weather Iâ€™ve seen was in Indiana. Yesterday in Illinois and Missouri it was chilly and windy, and once or twice between St. Louis and Kansas City a few snowflakes blew around, but nothing that I could really call bad weather. Here in Denver, itâ€™s warm enough this evening to walk around in just a T-shirt.  Tomorrow Iâ€™ll try to reach Salt Lake City.   March 31       Midvale, Utah, just outside Salt Lake City   Drove 623 miles today, including a few miles of â€œbacktrackingâ€ when I overshot Salt Lake City while looking around for a motel and had to turn around and come back down I-15 a few miles to the suburbs. Still watching my coins, I checked into Motel 6.  I was on the road today for almost 11 hours. Pulled out of Denver about 8:15 a.m. and pulled up to this motel at 7 p.m. The Rocky Mountains slowed me down (yes, the Blue Shoe, with close to 133,000 miles on her, made it over the Rockies!) as did some mountainous terrain in east-central Utah between Green River and Salina. Also, I drove about 100 miles farther west than I need have. I stayed on the Interstate, only to discover that I could have cut north directly to Salt Lake City on a state two-lane at Green River.  This in fact is what I did do at Salina, on the advice of a cashier at a gas station, rather than push on for another 55 miles to the junction of I-70 and I-15.  The Rocky Mountains, dappled with snow from recent storms which, according to the Denver radio stations, were the heaviest since 1913, were breathtaking, or I should say would have been if I hadnâ€™t had to concentrate most of my attention on driving.  Still, I did succumb to Lynn and Nadyaâ€™s gentle pressure: both had urged me to take pictures, so I stopped at a 7-Eleven store on the way out of Denver and bought a roll of film for my camera. I did stop to snap a few pictures while crossing the Rockies, and also a few times while traversing Utah, to photograph the moonlike desolation of eastern Utah and the mountain vistas that greet you as you get closer to Salt Lake and Provo. Since I was driving  past  these mountains rather than  through  them, it was easier to appreciate their snow-capped majesty.  Utah is as strikingly diverse in its topography as is California. At 2:00 youâ€™re looking at John Wayne movie vistas of desert floor and rocky-brown cliff; two hours later youâ€™re looking at snowcapped peaks that remind you of Switzerland.  Traversing the Rockies did slow me down, but I was past them by about 12:30. The counter of my journey clicked past 2,000 miles just a few miles outside of Grand Junction, where I stopped for a bite of lunch, and then crossed into Utah about 15 miles later.  Most of the afternoon was spent driving through the widest and most utterly desolate landscape Iâ€™ve ever seen: imagine films like  The Searchers  and  Stagecoach â€”thatâ€™s eastern Utah. Did the wide spaces of Kansas make me worry about running out of gas? Yes, but that was as nothing compared with the gas-tank anxiety instilled in me by the wasteland of eastern Utah. When I pulled out of Green River after stopping to take a photo and saw a sign that read, â€œNo Services On I-70 Next Hundred Miles,â€ I was truly spooked. As I mentioned, the Shoe only holds ten gallons. I crept over those mountain passes at 35-50 mph, way under the speed limit, to save fuel. But I still had half a tankful when I reached Salina 100 miles later.  I topped the tank (again) and swung north toward Salt Lake, pulling off the road on the way to photograph one particularly beautiful mountain peak which I somehow knew that Nadya would like (!)  Tomorrow is the last leg of the journey. Interstate 80 lies just a few miles north of here, and owing to my â€œovershoot,â€ which took me as far as Salt Lake City airport and allowed me a glimpse of the landscape between here and there, I know that as soon as you swing on to I-80 from Salt Lake, the signs say â€œReno.â€ By tomorrow night I should be at my sisterâ€™s house, and Iâ€™m not at all sure how I feel about that, but weâ€™ve come this far, letâ€™s see where the rest of the way takes us.   May 26    Monday    dirty motel, Reno   â€œThe rest of the way...â€  That last day of the trip across was a long one. I checked out of the Motel 6 at 8 a.m., had breakfast at Dennyâ€™s and got moving again.  If I thought eastern Utah was desolate, I had a surprise coming in the stretch of Interstate 80 that runs for another roughly 550 miles between Salt Lake City and the eastern edge of the Sierra Nevada mountains which abut the city of Reno. This is country that honestly makes you scratch your head and wonder why anyone would want to inhabit it, and in truth, large stretches of it are uninhabited. This is the great west, not the photogenic west of Monument Valley and the Grand Canyon, but the truly trackless, faceless west which must have held attractions for desperadoes and nobodies of a century and more earlier, people who wanted nothing more and nothing less than to get away and get lost. Getting lost is about the only thing I can imagine western Utah and northern Nevada being good for. A couple of hours out of Salt Lake City, I found myself cruising past the Bonneville salt flats, which were an ideal spot for Craig Breedlove to set the world land speed record in 1964, propelling his rocket-powered car The Spirit of America at 600 mph across an endless expanse of nothing-to-run-into, but I canâ€™t imagine their being much good for anything else.  As I crossed the state line from Utah into Nevada, two things suddenly intruded themselves into my journey: gambling and car trouble.  The moment you pass into Nevada, immediately, even out in the literal middle of nowhere among the sagebrush and the tumbleweeds, casinos and slot machines are suddenly everywhere you look, in every 7-Eleven, every gas station, every restroom.  But I had no time to think about gambling. The moment I crossed the state line into Nevada, I stopped for gas and made a horrifying discovery: the Blue Shoe, which had made it over the Rockies as steadfastly as the Little Engine That Could, was suddenly leaking transmission fluid. I still had 350 miles to traverse, and now I had a problem. I had to soldier on, so I went into a convenience store, walked past the slot machines and bought a bottle of transmission sealant which I hoped would solve things. The car could break down once we got to Reno if it wanted to, but we had to get there first. I poured the sealant into the Shoe, filled her with gas, and got back on I-80.  There is only one word to describe northeastern Nevada: depressing. This is the desert, and they mean it. Even the towns, miles and miles apart and huddled amongst vast expanses of bare brown and relentless scrub, make your heart sink. As a matter of fact, when I decided it was time to make another stop for gas, and to check the progress of my transmission fluid leak, I found that I was making that stop in a spot which the Washington Post, in a 2001 article, had designated â€œThe Armpit of America:â€ Battle Mountain, Nevada.  I had been living in Maryland, beautiful, green, rolling-hills Maryland, when I read the article about Battle Mountain. The author certainly made it sound grim, but somehow I couldnâ€™t believe that any place could be all that bad; surely, this must have been just another case of some snotfaced, hothouse-flower, east-coast limpwrist taking a sneering, condescending attitude toward one of the red states.  Nope. Battle Mountain was every bit as depressing as the Post said it was. Indeed, it was around noon now, and I was hungry as well as worried about my car, but decided not to even stop there for a hamburger; it was just too grim to linger. And I found to my intense dismay, when I stopped for gas, that the Shoe was still leaking transmission fluid. I did the only thing I could do beyond checking into a motel and staying in Battle Mountain while I let a mechanic work on my car (unthinkable!) I bought and threw into the transmission another bottle of sealant, and on I went.  50 miles beyond Battle Mountain, (above which tower, by the way, on the side of the storied mountain overlooking the town, the gigantic white letters â€œBM,â€ and weâ€™ll just let that stand) you come to Winnemucca, which ainâ€™t much better. In fact, during my sojourn here in Reno I was actually offered a job as a reporter on the newspaper in Winnemucca. I drove 180 miles, each way, to meet and chat with the editor of the paper, who took me on a tour of the town, which didnâ€™t take long. From what I was able to surmise, there are basically three things you can do in Winnemucca: gamble, go to church and drink. The town has a population of 4,000, and four casinos. There is a church on every other corner, and no shortage of liquor stores. The nearest city of any size is Reno, three and a half hours to the southwest. If you want to drive for five hours in a northerly direction, you can visit Boise, Idaho. But the point is, youâ€™re going to be looking at pretty much nothing but desolation for the next few hours whichever direction you take. Winnemucca was once a silver-mining town. The silver has long since tapered off, leaving the town without a whole lot to justify its continued existence outside of the fact that itâ€™s already there. Although I had been out of work for more than half a year at the time I was offered this job, I turned it down. I just couldnâ€™t see myself plunking down in such a place to live in the middle of a rerun of The High Chapparal. Winnemucca, close-up, positively reeked of the threat of boredom, which could easily lead one to become a drunk, an addicted gambler, or perhaps worse, a Jehovahâ€™s Witness. Pass.  I made it to Reno late that afternoon. My half-sister and her husband live in a â€œgated communityâ€ on the outskirts of this Nevada city, which is sprawling even as its central industry, gambling, dies out. Reno is in transition: its casinos are doomed, thanks to the burgeoning Indian-reservation gambling industry across the line in California. Why should people drive all the way to Reno to gamble if they can gamble right there in California, simply by driving up on to the local Native American reservation? Well, they gradually wonâ€™t, and Reno, as a gaming spot, does indeed seem to have no future.  But even as its gambling industry dries up, help is on the way: simply due to the fact that Nevada real estate is cheaper than California real estate, industries are pulling up their rootsâ€”many of them Silicon Valley rootsâ€”and transplanting themselves to industrial parks in Reno. During my two-month sojourn here, Iâ€™ve been unable to find steady work, but did do a little bit of â€œtempingâ€ at Horizon USA, which has a massive, warehouselike facility here, in which rows of pixies on telephones take orders from customers. There are many other such industrial â€œwarehousesâ€ on the southeastern outskirts of Reno these days, and more coming.  My initial plan when I arrived here was to take another shot at getting a commercial driverâ€™s license. I even went and talked with a guy at Western Trucking School, explaining that I had not passed the course in Indianapolis. The reason, I felt, was that C1, the trucking school I attended there, was essentially a robot factory: they have contracts with two Arkansas trucking companies which call for them to churn out so many drivers per month: the turnover in the trucking industry is huge, because few people want to drive over-the-road (basically, living 28 days per month in your truck) for more than a short while. At C1 you had two, or at most, three weeks to get your CDL or you were out. Well, Iâ€™m a quick study for some things, but driving a semi isnâ€™t one of them: I apparently needed more than three weeks, and they werenâ€™t willing to play ball with me.  Western Trucking School would have been willing, but my half-sister talked me out of going to another trucking school. She was sure that, somewhere in Reno, I could find a better job than driving a truck.  Well, I couldnâ€™t. The curse stayed on. Finally I hatched one of those harebrained â€œplansâ€ of which Iâ€™ve hatched so many in my life, and which almost never work out: I would attend bartending school, get a job tending bar somewhere along the Reno â€œstrip,â€ and put myself through the graduate program at the University of Nevada, finally getting my M.A. after all these years.  That might have worked, but personal problems with my sister and brother-in-law prevented any such long-term planning. Anyone who has ever gone home for Thanksgiving will understand that, all too often, family members are the worst people you can stay with, especially if youâ€™ve been out on your own for a few years, as I have, and are settled into a routine with which youâ€™re comfortable and familiar.  Bottom line: my half-sister kicked me out of her house. Actually, she didnâ€™t even have the guts to do it herself. She took her grandsons to the park and let her husband kick me out of the house.  In fact I wondered why, after having taken them to the park yesterday afternoon, she would go off again, at nine Oâ€™clock this morning, with the same destination?  Now I know.  My half-sister: a self-proclaimed churchgoing Christian who has Godâ€™s moral rectitude just oozing out of her nostrils.  Never misses a chance to wear her piety on her sleeve. But that doesnâ€™t stop her from talking about other people behind their backs, nor, by the way, does it stop her from bitching about how an influx of Mexicans is ruining the neighborhood.  My churchgoing half-sister had been keeping me under her version of â€œsurveillance.â€  Now, Iâ€™ve been followed around by the best of them: The CIA, the FBI and the Russian Federal Security Service. Who would have thought that the first among my â€œshadowersâ€ to openly admit having gone through the garbage on my account would be, not any of the above agencies, but my own half-sister, who not so long ago told me in a letter, â€œBefore you went off to school, you were my darling little baby,â€ and who also said of my staying at her house in Reno, â€œYouâ€™re safer here than youâ€™ve ever been.â€  Sheâ€™s real good at spreading around marzipan about â€œdarling little babies.â€ Sheâ€™s just as good at sticking a shiv right between your shoulder blades. And hereâ€™s the hell of it: after a busy week of talking about other people behind their backs, â€œjudging others over her tea,â€ so to speak, she goes off to Sunday school, convinced that sheâ€™s Jesusâ€™ little lamb. Kind of nauseating, when you think about it.  The issue, of course, was my private drinking habits. Now, my half-sister is no Baptist teetotaler: she will reach for the brandy bottle after work, and uncork a bottle of wine while sheâ€™s fixing dinner. But whatâ€™s OK for her apparently isnâ€™t for others. She and her husband have both been, for many years, curiously two-faced about alcohol: theyâ€™ll drink themselves, but they look askance at anyone else doing it. I knew this, and so I tried to be very discreet come cocktail time: I kept my Scotch bottle in my room, and if I wanted to have a couple of belts before dinner, Iâ€™d slip in there and close the door, usually turning on The Golden Girls on cable TV as I did so. Now it comes to light that my half-sister was not only closely watching my every move, and giving regular reports to my other sister on the telephone about how she supposedly â€œsmelled alcoholâ€ on me, but was actually digging through her own trash to see how many empties I was throwing out. She actually lied to my other sister, informing her that I was tossing out three empties per week. Thatâ€™s a barefaced lie: Iâ€™ve never drunk that much in my life. Iâ€™d be sick if I drank that much. When I was in the foreign service, we were warned that hostile intelligence agencies would go through our trash to see what they could learn about our personal habits. But I never thought that my own flesh and blood would pull a stunt like thatâ€”and then talk about it behind my back, no less. As bad as the KGB may have been, at least they didnâ€™t gossip.  Well, this presented me with a problem. I signed up more than a week ago to attend bartending school here in Reno. I also pre-paid for the course. To pull out of town now and head back for the old family homestead in southern California would mean forfeiting close to $900. I didnâ€™t want to do that. But staying on at my half-sisterâ€™s Happy House of Hypocrisy was also out of the question, so this morning while she was out in the backyard with her grandsons, hugely relieved, no doubt, that her holy house had been cleansed of Sin now that her husband had handled the dirty work for her, I quietly packed up the Blue Shoe and drove away. (And by the way, learned later that she had had the nerve to complain to complain, on the phone, to my other older sister about my not having said â€œGoodbye,â€ or â€œThank you!â€ Surreal.)  I am, for the next seven days anyhow, staying at the Best Value Lodge on S. Virginia Street here in Reno. To say that this place resembles the Bates Motel would do it too much honor: this is the crummiest fleabag Iâ€™ve ever stayed in. But it has a bed, a refrigerator, a microwave and a TV, and it doesnâ€™t have churchy, two-faced, back-stabbing relatives.  I called home of course, and talked with Dad. He was hesitant about giving me the green light to come on down, but when I assured him that parking on the street and sleeping on the couch in the â€œgranny flatâ€ out back would be fine, he relented. He said he would tell Lynn about it himself, but of course I had already discussed it with her. She canâ€™t wait for me to get there.  The bartending school will be just another waste of time, as I wonâ€™t be staying on in Reno, but I paid for it, so Iâ€™m going to attend. First class is tomorrow morning at 10:30.  '),(12129,3411285,'30,May,2004','2004-05-30 00:00:00','    Who would have ever thought that I would rank as a \"Grammer God\"?  Looking at my posts, I am sure y\'all can see that the standards must be pretty low.      You are a  GRAMMAR GOD !   If your mission in life is not already to preserve the English tongue, it should be. Congratulations and thank you!    urlLink   How grammatically sound are you?     brought to you by  urlLink Quizilla   '),(12130,3411285,'26,May,2004','2004-05-26 00:00:00','    So Friday started off with part of the Syphilis Commercial I\'m developing being filmed in my house.  We had some hanky-panky going on in three seperate locations.  People will drive by my house and say, \"That\'s the Syphilis house.\"  Whatever asshole, just roll on that rubber.  Vamanos to my party; I think I talk too much and say highly unappropriate things in large groups.  But continuing on...there were some cool ass people.  Mis amigos de El Centro vinieren.  Anja\'s roommate from San Fran was here as were her ex-boyfriend and his friend.  Totally awesome people.  I swear, Anja is just surrounded by immensly amazing, interesting, nice people.  Ex-boyfriend works for Lonely Planet...  Also, I wish Allen wasn\'t off to Quito...super cool.  Alguien stood us up, but I think he was getting laid..so I\'ll over look it.  El proximo dia we ventured into Chapel Hill for the first time.  We should have went way sooner.  It has that interesting flair that comes along with breeding a college town and prime realestate together.  We went to Michael\'s dad\'s house (is it proper to use to apostrophes like that?)  Ok, his dad is 73 and is just learning Spanish, has an amazing vocabulary and met a very cool Mexican lady on eharmony.com.  Can I just say I wish my grandma would take some lessons from him; life does not end at 70.  That\'s where we ended up going to Michael\'s reunion...well for 2 minutes.  I am so not going to mine!  All the people I\'d want to see aren\'t in my class...  Oh...FYI check out realrhapsody.com.  It rock\'s my world.  For 10 bucks a month you have access to millions of songs.  I think of my mom\'s song as,  I Never Promised You a Rose Garden .  I got to hear it for the first time in years...   '),(12131,3411285,'26,May,2004','2004-05-26 00:00:00','    I love my little albino kitty, but I really want a puppy or dog. I\'m not picky.  I think Shuggs may hate me forever if I bring a rival for my affection into the house, but really-she\'s not the boss of me. She\'d get over it, right?   '),(12132,3411285,'24,May,2004','2004-05-24 00:00:00','    I thought this quote was a good reminder of how I should live.  Unfortunately, all the good quotes that I would want to put on my signiture file for work come off of people I work with signiture files.    \"SPEAK YOUR MIND, even when your voice shakes.\"                                                       Maggie Kuhn   '),(12133,3411285,'23,May,2004','2004-05-23 00:00:00','    Ok...so it just dawned on me that \"Random Fire\" may sound a little pyroish to those FBI agents putting things like this into my FBI file (And I\'d just like to say to them that they have a rather crappy and boring job--they should come do condom demonstrations like me).  But back to my original point...I couldn\'t come up with anything else and I have used a screenname with \"fire\" in it for the past 8 years (and no, I am not putting what my screenname is on this site) and I\'m a little random.  So put this post into my FBI file while you\'re at it.  Thanks.   '),(12134,3411285,'22,May,2004','2004-05-22 00:00:00','    Ok, so I went here (http://web.tickle.com/tests/uiq/index2.jsp) and they told me my IQ was 131.  Damn, I can never be in Mensa.  The strange thing is I distinctly remember it being 133 at some point, so I am slowly becoming more stupid.  At least now I have some justification as to why I can\'t remember anything.  Argh!   '),(12135,3411285,'22,May,2004','2004-05-22 00:00:00','    Ok..something has got to be wrong with me.  Why don\'t I have any interest in a relationship?  I\'d rather watch Queer as Folk and/or Harry Potter then date someone.  But I do kind of miss kissing...kissing is so nice.  I had a friend in undergrad I would make out with sometimes, just because.  I miss that.  Was that wierd?  I think I am blaming all the men in my life who have screwed me over for my lack of wanting a relationship...they all suck.  And yes, that includes you wimp-ass step-father and you unnamed psycho pervert.  And yes, I can have a self-pity party if I wanna...It\'s my goddamn blog!  Plus, the pity only lasts 24 hours and then I get back to reality(Also the brand-name of the female condom; FYI)  Maybe I\'m a lesbian and I just don\'t realize it.  I\'ve never really been attracted to anyone...I\'ve just wanted the affection/attention.  How sad is that? I don\'t think I\'d care one way or the other what genitals the person had...after all that\'s just cosmetic.  I guess time will tell...   '),(12136,3411285,'21,May,2004','2004-05-21 00:00:00','    Ok.  I am done.  I refuse to watch another Holocaust movie for the rest of my life. Each time I finish watching one of \"those\" movies I become depressed for days and start thinking about what I would do in a similar situation.  It\'s just too emotionally draining.  I can\'t handle it.  However, the American soldiers making Taliban and other Muslim prisioners do all of those humiliating poses etc, should be required to watch every Holocaust movie.  Because apparently the ignorant people that we have in charge of guarding prisioners have no conscience.  Oh, wait, the man we have in charge of running the country doesn\'t have a conscience, either.  American is making me disgusted.  I wonder if Cananda will let me in?   '),(12137,3411285,'21,May,2004','2004-05-21 00:00:00','    This song is so amazing in Y Tu Mama Tambien  Artist: Marco Antonio Solis Lyrics Song: Si No Te Hubieras Ido Lyrics Te extraÃ±o mas que nunca y no se que hacer despierto y de recuerdo mal amanecer espera otro dia por vivir sin ti el espejo no miente me veo tan diferente me haces falta tu   La gente pasa y pasa siempre van y van el ritmo de la vida me parece mal era tan diferente cuando estabas tu  si que era diferente cuando estabas tu  No hay nada mas dificil que vivir sin ti sufriendo en la espera de verte llegar el frio de mi cuerpo pregunta por ti  y no se donde estas si no te hubieras ido seria tan feliz  No hay nada mas dificil que vivir sin ti sufriendo en la espera de verte llegar el frio de mi cuerpo pregunta por ti  y no se donde estas si no te hubieras ido seria tan feliz  La gente pasa y pasa siempre van y van el ritmo de la vida me parece mal era tan diferente cuando estabas tu  si que era diferente cuando estabas tu  No hay nada mas dificil que vivir sin ti sufriendo en la espera de verte llegar el frio de mi cuerpo pregunta por ti  y no se donde estas si no te hubieras ido seria tan feliz   '),(12138,3411285,'21,May,2004','2004-05-21 00:00:00','    \"Wickedness is a myth invented by good people to account for the curious attractiveness of others.\" Oscar Wilde   '),(12139,3411285,'21,May,2004','2004-05-21 00:00:00','    So I was on someone else\'s blog and came upon this website.  Who knew I was such a product of my upbringing?  It actually makes me quite sad.  http://www.selectsmart.com/PRO/beliefnet/index1.html  1. Mainline to Liberal Christian Protestants (100%)  2. Liberal Quakers (93%)  3. Unitarian Universalism (92%)  4. Neo-Pagan (91%)  5. New Age (86%)  6. Secular Humanism (79%)  7. Reform Judaism (76%)  8. New Thought (73%)  9. Theravada Buddhism (73%)  10. Mahayana Buddhism (72%)  11. BahÃ¡\'Ã­ Faith (67%)  12. Sikhism (64%)  13. Scientology (63%)  14. Christian Science (Church of Christ, Scientist) (63%)  15. Church of Jesus Christ of Latter-Day Saints (Mormons) (52%)  16. Nontheist (52%)  17. Taoism (52%)  18. Orthodox Quaker (49%)  19. Jainism (48%)  20. Hinduism (46%)  21. Mainline to Conservative Christian/Protestant (45%)  22. Orthodox Judaism (42%)  23. Jehovah\'s Witness (39%)  24. Islam (34%)  25. Eastern Orthodox (27%)  26. Roman Catholic (27%)  27. Seventh Day Adventist (26%)   '),(12140,3411285,'21,May,2004','2004-05-21 00:00:00','    Ok, so I want to feel safe.  Is that so much to ask?  For work, I had to go teach at the state prison about STDS.  Is it ever good to have a young woman teach hardened criminals about sexually transmitted diseases?  It was a \"low security\" prison.  Which really just meant that there were prisoners milling around every where.  Again, should I ever be milling around a prison yard without an escort?  What the hell is the County thinking?   '),(12141,3411285,'22,June,2004','2004-06-22 00:00:00','    Anja and I adopted a puppy!!!  We are the proud joint parents of a 3 month old girl mix (I refuse to call our sweetie a \"mutt\") named Stack.  She has adorable brindle coloring and a white belly.  Shuggs is a little disappointed, the first time she saw Stack, she attacked her.  However, it seems as if they are easing into being tolerant of one another.     Anja and Stack are already deeply in love, but I am falling in love more each hour.  I think my heart was on the 3 year old Springer mix with heart worms.  I think about him every day even though deep down I know he was not the right choice.  Shelters are very depressing. But I love Stack.  She\'s really cute and we\'re all into the \"Positive Training\" thing utilizing a \"clicker\".  We are taking her to pick blueberries today after work.  Hopefully, she will like the car ride.  Stack went for a walk around the neighborhood yesterday and everyone seemed to be giving her dirty looks.  Finally, a nice man accross the street informed us she\'s part Pitt Bull.  We had no idea.  However, after much research, we descovered that Pitt Bulls can make very nice adoring, loving, sweet pets.  Their agression comes from how they are trained.  But it doesn\'t matter...we are going to make her the sweetest Pitt ever!  In other news...I have no idea what\'s wrong with me. I\'ve been extremely cranky.  I think working for a job and being under paid has something to do with it.  However, I don\'t think it\'s good to quit a job after 3 months.  So I am just going to accept it for the next few months and then continue my search.  The job itself is so easy.  Plus, I have an unlimited supply of condoms.   '),(12142,3411285,'13,June,2004','2004-06-13 00:00:00','    We all knew that I wasn\'t a star;-)           urlLink   Your Broadway Musical Debut (2004 season)        Full Name         Age          You have a part in:      Never Gonna Dance        Your famous (at least on Broadway) costar is:      Idina Menzel        Are you successful?:      The experience is DISMAL, you are fired after a week and never get a job on Broadway ever again              This   fun quiz   by  urlLink  bwaygal128   - Taken 396 Times.          New! Get Free  urlLink Daily Horoscopes  from Kwiz.Biz'),(12143,3411285,'13,June,2004','2004-06-13 00:00:00','    urlLink    Isn\'t my grandma just so cute???  Now if she\'d just get rid of the cigarettes.&nbsp; urlLink   '),(12144,3411285,'12,June,2004','2004-06-12 00:00:00','    urlLink    Oh, North Dakota...&nbsp; urlLink   '),(12145,3411285,'12,June,2004','2004-06-12 00:00:00','    Yay!  I can link! Here is the commerical I organized for work.  I did a lot of the script, organized the actors, locations (Couples 1,3 and 5 are at my house), and much, much more.  So I am pretty proud.  It airs in the Raleigh/Durham area on NBC17. You can go  here  to see it.  Just scroll down towards the bottom of the page and click on the box that says \"Durham County Health Department.\"  You need quicktime to watch it.   '),(12146,3411285,'11,June,2004','2004-06-11 00:00:00','    Ok.  Let\'s see if I can hyperlink.  I am going to try link to quizzila. Click Here For More Quizzes   Here it goes:   '),(12147,3411285,'11,June,2004','2004-06-11 00:00:00','    Sad, but true.    You\'re a Speak & Spell!!  You nerd, you.  Just because you were disguised as a toy doesn\'t mean you weren\'t educational, you sneaky bastard.    urlLink   What childhood toy from the 80s are you?      brought to you by  urlLink Quizilla   '),(12148,3411285,'11,June,2004','2004-06-11 00:00:00','    Anja keeps making fun of me because I came out a \"Christian\" on the religion quiz and all, yes all, the cool people were \"Unitarians\".  It\'s actually quite funny (the jokes), but I\'m really freaked out!  How can a gay lovin\', abortion believer be considered a \"Christian\"?  My family was wielding their power through the quiz; I just know it.  One of my uncles, the one considered smart, tried to argue with me a few years back that Christianity was the only monothiestic religion.  He got ignorantly angry with me when I tried to explain that Jews, Muslims and Christians all worshiped the \"One God.\"  Then my aunt started crying that I was going to go to hell.  However, the funniest part of this whole story is that in all of my 26 years, I have not seen any of them ever attend a church, live a holy life or say anything remotely \"Christian\".  But then again, I guess the whole basis of Christianity is that as long as you ask for forgiveness, you get to go to heaven.  Oh, and tithing doesn\'t hurt.   '),(12149,3411285,'11,June,2004','2004-06-11 00:00:00','    I\'m a Quiz Whore.    Tarzan!    urlLink   What movie Do you Belong in?     brought to you by  urlLink Quizilla   '),(12150,3411285,'11,June,2004','2004-06-11 00:00:00','    Last weekend (yes, I\'ve been too lazy to upate until now) I went to the Chino Latino to watch my friend Oscar perform drag.  The Chino Latino is a huge dive bar with a very interesting crowd.  Well, they are all Latino, for one.  Second it is a complete mix of straights and gays.  The dance floor is the biggest mix of couples I\'ve ever seen.  It kind of reminded me a perfect world (in that straights and gays lived side by side without any prejudices--oh, and everyone speaks Spanish)  Let\'s say my roommate\'s brother, all six-four and white blonde of him really stood out.   '),(12151,3411285,'04,June,2004','2004-06-04 00:00:00','    Life Update: Iâ€™ve been in Chapel Hill twice this week!  Tuesday night we explored with Katka.  We had the worst Margarita ever.  FYI if you go to Banditos just stick to Dos Equis.  Then last night we went to a â€œgoing awayâ€ thingy for Allen.  In front of the Chapel Hill campus is a huge sun dial.  Itâ€™s at least  30 feet in diameter.  Well, Allen, Anja and I are all standing on the sun dial like idiots trying to figure out how the thing works.  Apparently, the students at CH are much smarter then we are; Allen even had instructions printed from the internet.  I definitely feel humbled.  However, we did find some very cool bars.  There is this place called the Lantern and this place called Telulahs.  Telulahâ€™s has a special little room thatâ€™s all Haremish.  I also had a great time talking to Ninki and her husband.  ECH Sandwiches: ECH is nice enough to provide their committee meetings with a snack each month.  So itâ€™s 5:00 and I rush on over to the meeting; famished.  They have these cute little sandwiches, so I take one, get seated and get ready to attempt to save a little piece of the world.  I open up the sandwich, take a bite and nearly have to regurgitate whatâ€™s in my mouth.  I need to preface this by saying that a lot of people there were chomping away on the sandwiches and I mean no disrespect to ECH, but I would never have dreamed of this combination.  There was a bun and on the bun, an inch of mayo and a half an inch of refried beans with lettuce.  Can I just say, WHAT???   '),(12152,3411285,'04,June,2004','2004-06-04 00:00:00','    Your Stripper Info by  urlLink  radioface        first name           age           Stripper Name:    Chi-chi      Specialty:    the booty-clap      Customers say:    \"Nooo, come baaaack!\"             Created with the ORIGINAL  urlLink  MemeGen  !   '),(12153,3668030,'29,June,2004','2004-06-29 00:00:00','    i have serious problems. i am such a fuckin slacker. i have no modivation to wake up in the morning and it is bad news cause i keep skipping the only case i really should be going to. oh well i am sure it will be fine it is always fine.  i had lots of fun last weekend also. i am warn out from it though. the party was fun but being the cup person sorta blows cause you have to be all responsible. i love going on my roof and the night ended with the roof talk so im happy for that. if you havent been on my roof before you should come over and we can kick it together except for im a huge pussy and am rather slow so you have to deal with that when we go on the roof. everytime im a little less inhibited though.   i am at school i have to write this paper for thursday and i need to interview lynn, naomis old roomate but i dont even know how or what im supposed to be asking because i didnt go to class to ask the teacher today. oops. im gonna email the ta and see if she can email in time to find out what the fuck im doing. (oh incase you were curious im interviewing lynn because its a community psych class and i wrote a paper earlier on homeless women and the second paper is to either summarize 4 more papers on homeless women or interview someone who works with homeless women and lynn does.. yeah)  muffy is in town and that is super exciting. she is coming to chicago tonight so if you wre not doing anything you should come  i would like to note that when i say \"you\" i am referring to any of my friends that read this.  well im gonna go email some TA\'s sorry this is so boring '),(12154,3668030,'24,June,2004','2004-06-24 00:00:00','    i am at school and i am super fucking tired. i went to bed fairly late after pretending to study. i really didnt study at all i have winged one test and i have another one in 3 hours that is more important so after this blog im gonna spend the next 3 hours trying to stay awake and studyng for the test. my memory is shit so wish me luck.  as you may know jessica explained how to fix the no one can post problem so now everyone can post and please do cause well it makes me feel good.   did i mention how fucking tired i am. i have a problem i think i may need help this. i am addicted to sex in the city. ginny has every episode available on DVD i am on the 4th season now and its just so addicting i find my self making bullshit excuses why i cant go out so that i can just stay at home and watch sex in the city. there are only 6 seasons so im almost done and than ill move on to movies and twin peaks AHHh!  the last few days have been uneventful due to the fact that i have these midterms today. we are throwing that party though so thats exciting. i hope to see everyone there.   oh wait also my brother hooked it up with tickets to the sox vs cubs game at comisky park on friday. realy exciting. it sucks sorta cause a lot of my good friends had to work but its cool cause i feel special. apparently the tickets are worth more that 150 bucks and i am very lucky to have such an awesome brother. me sarah, josh, and kevin got the 4 tickets that i got but than omar rizvi had kevin call me and tell me that he would buy a ticket for 100 bucks and so i told my brother and he talked to omar directly and sold him 2 tickets. im happy that im going with friends. im gonna be so fucking wasted on friday before i even have the party. bad news bad news or good? i dunno but its gonna be fun.  yawn okay im off to read jessica\'s blog and than go study '),(12155,3668030,'22,June,2004','2004-06-22 00:00:00','    Jessica informed that you can not post annonymous, which just means if you don\'t have a blog you can\'t post. I can\'t figure out how to change this so please be patient cause those of you who don\'t blog will be able to post soon.   I will write tomorrow or later because I am sick of typing :)  I went to Great America though and it was super fun :)  Oh also. I am having a part on Friday so everyone better be there!!! '),(12156,3668030,'17,June,2004','2004-06-17 00:00:00','    the movie i saw was called Route 181: The South. 181 was the resolution created in 1947 to split Palestine giving 57% to jews and 43% to palestinians. Although jews accepted this, palestinians did not because they felt as if the land was theirs and giving 57% of it to jews would be equivalent to giving 57% of your house to the dude who stoll your car last week. yeah so anyway The South was because it was a three part movie based on geography (which I though was a really cool way to do it) it went through the south, north and center of Israel and the unrecognized territory of the gaza strip and west bank giving personal interviews to basically anyone they came into contact with that would talk to them. it was very interesting because you got a feel of how israelis and palestinians think. unfortunately i did not get a chance to see the center or the north but apparently its on dvd and im going to buy it. the south of israel is predominately israeli and therefore the movie was predominately of one opinion. being palestinian i understood where they were going with the movie. an important fact is that there was two directors of the movie a palestinian and an israeli. so its almost like a joke really. \"a palestinian and a jew decided to take a road trip through israel, the west bank and gaza and take interviews..\" anyway in the south they interviewed an eldery israeli man who worked for the israeli water museum. although they didnt discuss politics in the begining they did discuss water in israel. they asked questions hinted at the fact that israelis have all the water they could possibly ever need while palestinians were living in ghettos without water supplies. once any political talk about the oppresion of palestinians came up the elderly man freaked the fuck out no joke like he turned red and screamed it was kind of scary. anyhow even though i havent seen the other movied i know that when they go to the poor palestinian areas they will touch on the fact that they have no water. i might be saying this confusing but i just think its super cool. i cant wait to buy and finish it. unfortunately the movie can be very very boring i wish they could have made it in one movie instead of 3 but i am assuming it is virtually impossible to film the entire country in one movie. part of the movie is landscape because you watch them physically drive while they film outside from inside a car. you never see the interviewers only the interviewed and the landscape of the country through their eyes.  I feel this method of filming the opinions of the poeple of such a controversial area can be the most unbiased. Although there it still can show a biased view based on the questions asked by the interviewers it also shows random people with all different opinions. I also believe that the questions weren\'t very biased but I do feel the directors have an opinion.  When i buy the movie i hope that everyone will watch it. '),(12157,3668030,'17,June,2004','2004-06-17 00:00:00','    so i used to have one of these in a time of my life that was very fun and very painful at the same time. jessica has sparked the want to leave my thoughts for others to read. i attempted at choosing about 15 titles for this damn thing and this is the one that finally was not taken. apparently i am very unoriginal. oh well. anyhow hopefully you poeple will read and comment or else this is just not going to last. that is really how the old one died.  i am in the computer lab at school right now. that will probably be where i am when i write these since i no longer have internet. i have class in 15 and so i need to leave in 5 so this wont be to long.   i am taking 3 classes this summer. i have more school in the summer than i did during the year. i have class from 9-4 on T and Th and class from 1-3:30 M, W, and F. Tuesdays and Thursdays are super rough as you may have guessed. i am not a morning person. i am taking my second psychology class. a part of me wishes i would have done psych. i find it very fun and interesting. the class i am taking is community psych and i love my teacher. he is fun, interesting and makes the class fun and interesting. its not easy but its fun work rather than busy work. everything thursday we have these exercises due that come with some sort of group work its lots of fun. last week was 10 sentences that start with i believe... this week was observing a place that has enclosed borders and writing about the groups of people that were under or not represented. fun stuff. well i am gonna get going to art history now (which by the way is really interesting.. not the class but the fact that the art history kids are SO MUCH DIFFERENT than any other kids in my different classes.. mainly poli sci of coarse but yeah they are hip if you will) im out. '),(12158,3668030,'30,July,2004','2004-07-30 00:00:00','    apparently not updating your blog is a crime--that was the meanest blog comment ever jessica jeesh.  i am volunteering at the abbey and am trying to capitalize on internet as much as possible so i get to post again. for all who dont know i dont have internet so its hard for me to post.  i cant even talk much cause well like i said im volunteering and it is just rude to be doing personal shit but i will wait until scottie yells at me.  so lately i havent been doing much school is over i think i did pretty bad for all you who are wondering but i did graduate or at least im pretty sure i did i have an appointment with an advisor tuesday to find out.  i have been thinking about things and i  think im just going to suck it up and move back in with my parents... i know its horrible and i dont really want to but im trying to way in my positives and negatives here.   living with parents positives     i can get a job and save money which equals stuff, vacations, and paying off debt all the things having a job gets you, you know...and less stress about money  i can work on making myself happy rather than drinking my life away  i can still have my freedoms and sleep at all my lovely friends\' houses right guys?  my parents house is pretty damn pimp--and girls you all can come over and kick it  just because i move back doesnt mean i have to be there forever im prolly gonna go to school in a year and than i will be back on my own with my responsiblity hopefully.   negatives     palos park no friends in palos park no privacy constant bitching constant fighting lack of freedom no chicago apartment no city ahhhh  my decision isnt final so i dunno what to think really but i cant type anymore i will blog soon though jessica. '),(12159,4236023,'22,August,2004','2004-08-22 00:00:00','   Okay, so I just got back from my cousin\'s wedding... Mike Bitner.  He got married to Marsha Edmenson (Or however it\'s spelled), and that\'s cool and everything, but he asked me to film it for him, and I tried... and it might turn out okay, but I don\'t know.  I went to a rehersal dinner yesterday and felt awkward because I didn\'t know anyone.  Well, tonight was about as awkward, only, my family was there, so it was okay, because I just talked to Loren and Erin when I didn\'t know what to do.   The filming went okay, I guess, I felt odd, walking around, taping stuff while there was an official wedding going on and whatnot, but it was outside in the woods near a little beach thingy, so I guess it was a little unorthodox to start with.  Plus, most of the time, other than the obvious parents, pastor, and the to-be-married themself, I didn\'t really know who else they wanted me to capture on film, so I just kinda got everyone.  I figure Mike can just cut out whoever he wants from the video, so it\'s no skin off my nose.  Anyway, I\'m getting behind in this blogging thing... and I don\'t really feel like working on it right now.  Exactly what I was afraid would happen.  So... I guess I\'ll catch up when I do. '),(12160,4236023,'21,August,2004','2004-08-21 00:00:00','   Today I basically wasted my whole day at work.  Usually I don\'t feel a day at work has been wasted, but this paticular day consisted of seven hours of almost nothing.  Slow, tedious, never-ending nothing.  Nothing in a place that nothing is not meant to be.  Not nothing in a nice, air-conditioned room watching meaningless cartoons, nothing in a hot, oven-heated, angry-customer-filled room that smells bad.  (not that bad, but I was going for the contrast).  Also, the other night, I went with some people to go see \"Collateral\".  It was an odd movie staring Jamie Foxx and Tom Cruise.  Basically what happens is this guy makes this taxi driver drive around to all of these witnesses that he has to kill, and the cab driver keeps trying to get away or get help, but it never works.  I won\'t give away what happens, but it\'s a really depressing ending, so get ready to cry for years.  Also, another odd fact of going to see the movie is that Mrs. Eubanks, my 7th grade Reading teacher was there.  I haven\'t seen her since, say... 7th grade.  But I knew who she was.  I didn\'t quite recognize her; she didn\'t have her trademark humungo glasses on.  Or her large cup of coffee that she spent half an hour getting, either.  And she didn\'t assign us to read any books about war, or black city sports, or anything like that.  SO, we had an awkward moment, because she happened to be seeing the same movie we were, and was waiting for a couple of girls outside of the theatre.  We were already late, and I wanted to catch as much of the previews as I could, but she was doing some figuring out of who we were, and she was definetly taking her time with it.  Especially Krystle who, after several minutes, she decided she had never taught.  10 points for Eubanks!  She thought I had graduated, too.  Oh well, if I was a teacher, I wouldn\'t remember my students very well, either.  I will admit to that.  Anyway, it\'s horribly late, and I have to go to work and then a wedding tomorrow, so I\'ll work on this later.  Ciao! '),(12161,4236023,'19,August,2004','2004-08-19 00:00:00','   For LHS\'s National Honor Society, a new crudential has been added.  Ten hours of community service.  Now that\'s all fine and dandy with me, but I\'m not exactly sure what counts as community service.  Apparently, though, I\'ve already gotten roughly 5 hours of it.  Today, I spent about two hours gathering \"first-day\" paper bundles for school with David.  We had as much fun with it as is possible, and got to see Dr. Z, who unlocked some doors for us.  But the real story was yesterday.  David and I were painting the pastor\'s picket fence white, and we\'d just gotten about a third of the outside done, when all of a sudden, these two guys start shouting at eachother.  I was aquainted with one of them, I see him around every now and then, and he\'s dating a girl in my class.  The other guy, I had only just met him, and he was helping paint the fence.  I would think maybe both of these guys are in their early to mid twenties, but I\'m not really sure.  At least that old.  But anyway, the boyfriend introduces himself to the guy, and tells him that they need to talk.  Well, he asks what about, and the boyfriend says thy need to talk about why he\'s all over a 15-year old, or something like that.  And then they start shouting, and the boyfriend\'s yelling, \"Go ahead, hit me!\" and it was really funny and awkward, David and I sitting there painting, trying to keep from laughing.  Very interesting.  Anyway, we got three more hours right there, so we\'re already half way.  Anyway, I\'ve got an orthodontist appointment to attend to, so I guess I\'ll touch base later, ay? '),(12162,4236023,'17,August,2004','2004-08-17 00:00:00','   Well, as much as I\'d like to keep everyone on the edge of their seats, I still don\'t really have much to report... go figure.  I\'m still doing the exact same thing I mentioned in my previous entry every day, and I probably will be until school starts.  The one thing I did do differently yesterday was go out to the covered bridge with David and Krystle.  We had a good time, we scared the whirlygig beetles, and fed cheese twists to a frisky chipmunk, which, after fleeing into his hole, we then attempted to drown with some coke, and we also got rust stains all over us on the swings (at least David did).  It was a nice change of pace, Krystle made us all Turkey sandwiches, and we ate them as the swift currents of the calm river peacefully relaxed our minds... Ah, yes... Then we came back and just hung around, found out I still have some bugs to fix in my hook-up of our entertainment center, and played Simpsons Clue  (I won, it was all too obvious that it was Homer in the Kwik-E-Mart with the slingshot).  Ah, yes, the productive life of me.  It\'s a simple life, but I enjoy it.  Anyway, that baritone isn\'t going to learn how to play itself... and neither am I, at this rate... More to come later! (Aren\'t you excited?) '),(12163,4236023,'16,August,2004','2004-08-16 00:00:00','   Okay, still not much to report. All\'s silent on the western front. But I have to be patient. I mean, as soon as school starts, all kinds of stuff is going to start happening, and that\'s good.       This summer went really fast, but compared to other summers, seemed to consist of nothing. In one aspect, that\'s kind of sad, but maybe it will help me to appreciate this school year more. Sure, school means getting up earlier, having assignments due, having to deal with some people that I\'d rather not, but on the flipside, I get to see a lot more people that I miss over the summer and some of my classes are even fun. An example of this will hopefully be my AP English class.       There are a lot of good kids in that class, kids I can have fun with.  Sam, Matt, Scott, David, and a few more that I can\'t quite remember off of my head.  All of my memories of Mrs. Young\'s 10th grade class are simply of Scott and I doing as we pleased, but still getting the highest grades in that class.  And that\'s pretty cool.  I figure this class will either be really easy and fun, or really hard and fun.  Either way, that\'s a good deal.       Also first sememster, I have some video editing class with Mr. Ehrmantraut for no credit, and I saw some real undeserving people sign up for it, so I don\'t know how that will go.  David\'s in it with me, too, and that\'s wicked awesome, because I haven\'t had a lot of classes with my friends in High School, what with Spanish and Agriculture.  Spanish was at least worth it, because I\'ve met a lot of neat people in those classes, and Mrs. Simmers is pretty awesome.  Now, Agriculture... I can\'t say much for that class.  It was a good class, but the only person I felt I could really talk to in there was Jarrod Pollard, even though Mrs. Carnigie was really nice, too.       Anyway, I had better get back to what I\'ve been doing for the past few weeks: Playing Nintendo, reading English AP books, and practicing baritone.  So, until next time, don\'t spend all of your potatoes in one place. '),(12164,4236023,'14,August,2004','2004-08-14 00:00:00','   Welcome to my blog!  This is, perhaps, my third attempt at one of these \"on-line journal\" type things, and I\'m having big dreams and hopes for it.  Ah, the possibilities of the internet.  For example, I bet Josh Server from All That is reading this right now... I can only dream... '),(12165,3501362,'31,May,2004','2004-05-31 00:00:00','   Disculpen, recientemente he habilitado la opciÃ³n de \"Cualquiera puede comentar\", ahora quien ingrese a este blog puede dejar sus opiniones.  Por favor, colabore con la humanidad y haga un comentario, serÃ¡ muy bueno para Ud!  '),(12166,3501362,'31,May,2004','2004-05-31 00:00:00','   Ok, un nuevo dia comienza...  Debo confesar, ayer estaba un poco cansado de estudiar todo el dÃ­a, que decidÃ­ liberar mi mente por un momento con una buena pelÃ­cula, asi que intentÃ© ir a ver \"El DÃ­a despuÃ©s de MaÃ±ana\". No sÃ© si agradecer la \"hora y media\" que me pasÃ© manejando alrededor del centro comercial por un parqueo, o el hecho de que era domingo y TODOS se reunen para ir al cine, pero lleguÃ© muy tarde y ya no habÃ­an funciones tempranas (al menos no sino hasta las 9:00 pm). Asi que me puse a pensar, mejor veo otra pelÃ­cula (hoy en dÃ­a no hay tantas pelÃ­culas buenas, y para mi gusto, prefiero alquilar una extranjera que ver una pelÃ­cula de acciÃ³n barata). Asi que ahÃ­ estaba en cartelera: \"Alguien tiene que ceder\" con Jack Nicholson y Diane Keaton.  Primero que todo, QUE EXCELENTE PELICULA! Definitivamente la recomendarÃ­a para aquellos que estÃ¡n llenos (o vacÃ­os) de la vida y el amor. (Como que la vida es el tema del momento huh?). Una pelÃ­cula que nos da una gran lecciÃ³n, no importando de que manera lo veas, todo se enfoca a un asunto en especÃ­fico: En una relaciÃ³n ALGUIEN debe ceder.  Este tema no solo le sucede a una persona sino a toda una poblaciÃ³n allÃ¡ afuera, que tienen teorÃ­as inmensas del amor y de \"dejar ir\" y aquellos pensamientos tristes de que \"Si amas a alguien dejalo ir, si regresa era tuyo, sino nunca lo fue\". Ok, parte de eso es cierto, pero que tal si el verdadero amor requiere mÃ¡s esfuerzo que solamente sentarte con tus brazos cruzados con la mirada perdida al horizonte esperando que algo pase?  Es cuestiÃ³n de jugar tus cartas correctamente. El destino sucede, pero para aquellos que buscan su fortuna, sin dejar que la fortuna les llegue a ellos. He tenido estas experiencias en el pasado, y por eso comparto esto. Tu dejas ir solamente porque es necesario, pero el destino viene de vez en cuando y te dice \"OYE ESTUPIDO, AQUI ESTOY TOMAME!\" mientras uno estÃ¡ vagabundeando en cosas sin importancia de la vida diaria. Es cuestiÃ³n de tomar decisiones, de dedicarse, de dar, de alcanzar mÃ¡s alto de lo esperado.  Es chistoso, algunas veces nunca hubiera pensado que el destino a travÃ©s de los aÃ±os regala oportunidades. Es como una loterÃ­a ganadora. Tienes tu nÃºmero, recoge tu premio cuando el tiempo es correcto. Y no solo viene una vez en tu vida, pero VARIAS veces en tu vida, el problema es escoger, y no tener tu mente enfocada en lo que realmente quieres.  Como humanos siempre nos perdemos en nuestros fracasos, nuestras decepciones, y mayormente en nuestros defectos. Nuestras inseguridades estan basadas en estos, y es por eso que a veces sentimos que no es el \"momento correcto\" para dejarse llevar por el amor. Te voy a dar la verdad: NO EXISTE MOMENTO CORRECTO EN LA VIDA. Es justamente de buscar y tomar riezgos cuando te toca tu nÃºmero. Apuestas y le subes el precio, y a veces ganas, a veces pierdes. ESA es la forma para saber correctamente que estÃ¡ pasando en tu vida amorosa.  Â¿AsÃ­ que tienes sentimientos por esa persona sentada al lado tuyo en clase? InvÃ­tala a salir de una vez por todas... Que hay que perder? Integridad? Prefiero perder eso que tener un sentimiento de culpabilidad por un buen tiempo. Y que tal estando en una relaciÃ³n: Todo se torna a rutina, pregÃºntate que hay que perder, que hay que arriezgar para realmente cambiar esa rutina por un breve segundo (que puede durar hasta una eternidad).   Cito a The Darkness \"Creo en una cosa llamada amor\"... Asi que creo que no me voy a quedar aqui sentado y esperar, voy a armarme de estatÃ©gias para buscarla, o buscarla de nuevo (quiÃ©n sabe, a veces el verdadero amor estÃ¡ a la vuelta de la esquina, solo que tu nunca viste a tu alrededor, con los ojos muy abiertos para realmente VERLO).  Es acerca del destino, pero tambiÃ©n de DECISIONES! SeÃ±ores(as)(itas), asegÃºrense de creer en sus decisiones, sin importar que tan malas o buenas sean, siempre y cuando te den una lecciÃ³n (y un premio, quien no desearÃ­a eso?!). O sino te lo advierto, una palabra: PROZAC.  Al final, no seas tan orgulloso(a); alguien tiene que ceder...  '),(12167,3501362,'31,May,2004','2004-05-31 00:00:00','   Hola a todos, y bienvenidos al REAL Taco...  Intersante que esta sea mi primera nota de blog, o bien mi primera nota pÃºblica en el Internet. Usualmente me quedo las cosas para mi mismo, pero bueno, siempre hay un momento en el que uno debe exponerse con ideas locas...  Entonces, comencemos con el tema del momento:  PregÃºntense esto, Â¿EstÃ¡ bien que una persona que ha gastado 5 meses de su vida estudiando algo que definitivamente no es apto para hacerlo, continuar y darle una Ãºltima oportunidad al estudio, cuando son las 10:00 pm, sabiendo que solo tiene un par de horas para revisar el semestre entero, mientras se pasÃ³ el resto del dÃ­a pensando \"Para quÃ© es la vida\"? AlgÃºn moralista por ahi me dirÃ­a \"Tonto idiota (redundantemente hablando), Â¿DÃ³nde has estado todo el dÃ­a? Algunos pensaran que si no vale la pena, no gastes mÃ¡s empeÃ±o en ello, y otros dÃ­rÃ­an \"Hey, 5 meses y no hiciste nada, por quÃ© gastar un minuto mÃ¡s sabiendo que es una pÃ©rdida\"?  Okay, asi que eso es eso, pueden luchar toda la noche/dÃ­a con ello, yo me voy al hecho que si hice mal, pero debo darle un Ãºltimo intento, aunque no sea destinado para mi que gane del todo. ESO es lo que concluÃ­ hoy.  En todos los tÃ©rminos posibles, la vida sucede, y a veces no sucede del todo. No estoy escribiendo para bajar los Ã¡nimos, sino al contrario para dar un consejo; tal vez todo esto se ve mal y confuso pero de esa forma es que la vida nos reparte las cartas. Hay persona que perseveran sin importar la batalla (y aun asÃ­ pelean batallas de otros), y otras personas que saben exactamente cuando terminar todo. Creo que estoy en ese impase.  Asi que aqui estoy, 20 minutos despues, escribiendo aÃºn acerca sobre quÃ© deberÃ­a ser la vida. Pero en lugar de esto Â¿Saben? Voy a darle a la vida un intento mÃ¡s, para poder yo correr el Ãºltimo cuarto de milla. Es mejor que corra como un tonto, pero sintiendo que dentro de mi no existe causa perdida al final, siempre y cuando me dÃ© todas las herramientas y conocimiento necesario para APRENDER.  Eso es una regla general, algo que no nos enseÃ±an en el colegio, ni en la universidad, uno no recibe una maestrÃ­a en Vida, saben? Sin tan solo fuera asi de fÃ¡cil.  Lo que es mejor es la belleza alrededor de ello... El motivo no es darse por vencido y afrontar su derrota, pero reconocer que alguna vez hiciste algo, te propusiste una meta, y a pesar que nunca la lograste, aprendiste cÃ³mo manejaste hacia la Victoria, para que la prÃ³xima vez manejarÃ¡s con mejores habilidades, y tal vez entonces lograrÃ¡s cruzar esa lÃ­nea de una vez por todas.  Es grandioso que la vida SI nos da otras oportunidades, pero a veces estamos muy abrumados por la decepciÃ³n que no nos damos cuenta que ALLI estÃ¡.  Asi que, amigos mÃ­os, palabras sabias: No hay suficiente vida para que nosotros logremos todos nuestros sueÃ±os, pero siempre hay una oportunidad para que lo logres siempre y cuando mantengas esos sueÃ±os VIVOS.  ...Debo regresar a estudiar, los veo maÃ±ana!  '),(12168,3501362,'01,June,2004','2004-06-01 00:00:00','   Experiencias post traumÃ¡ticas de defensa propia cuando estamos enfermos: Inconscientemente nos hacemos un rollo hasta ponernos en posiciÃ³n fetal, para \"sentirnos\" mejor... Ahh no extraÃ±as estar dentro del vientre de tu mamÃ¡? No? Bueno si haces esto cuando estÃ¡s enfermo, entonces al menos tu subconsciente lo extraÃ±a; Es la madre naturaleza dÃ¡ndonos protecciÃ³n.  ');
INSERT INTO `msgs` VALUES (12169,3501362,'01,June,2004','2004-06-01 00:00:00','   Mi Ãºltimo dÃ­a del semestre, y SI, caminÃ© por la plataforma y me hundÃ­ con los peces, corrÃ­ la Ãºltima milla y no pude terminar del todo. Bueno, eso era lo que esperaba no? Al menos si me siento orgulloso de mi mismo por tomar ese Ãºltimo riezgo, en lugar de quedarme sentado con los brazos cruzados. MÃ¡s interesante es el hecho que no me siento desalientado y terminar mi travesÃ­a aqui.  Como ven, hay momentos especiales en la vida en donde se nos dan varias opciones, siendo como un gran buffet de opciones post-fracaso, pero nada estÃ¡ cocinado ni preparado todavÃ­a. Â¿Asi que quieres un omelette? Aqui estan los huevos, el jamÃ³n, el queso, lo que desees en Ã©l. Â¿Deseas mejor que te lo cocine por ti? Â¿SerÃ¡ que tendrÃ¡ mejor sabor? Â¿O no prefieres cocinarlo tu y encontrar tu propio gusto?  La vida es asi. Se nos dan diferentes post-opciones ya sea por Ã©xito o por fracaso. El problema es que no todos son adecuados para preparar un omelette propio con una opciÃ³n tomada por su cuenta. Algunos se van por el camino fÃ¡cil, contactan un \"chef\" experto para que se los prepare lo mÃ¡s cercano a su gusto. Pero Â¿Que pasa si deseas agregarle salsa de tomate o algÃºn otro condimento? No no no, he aqui la forma de hacer tus propios \"omelettes\".  A veces no estamos preparados para lidir con las post-opciones que la vida nos da, pero no hay otra forma de estarlo, ya que la vida continua, sin importar que puedas alcanzarla o que te quedes atrÃ¡s con el resto. Al final, el crecer no importando la edad, se basa todo en experiencia: una palabra clave en el camino al Ã©xito.  Todos tenemos un cierto gusto por nuevas experiencias, sin importar que tan buenas o malas sean, todo se nos queda dentro de nuestra cabezota. Y aprendemos de ello para perfeccionar nuestras decisiones y estrategias para alcanzar metas. Somos muy necios de naturaleza y a veces se nos hace dificil tomar decisiones basÃ¡ndonos en nuestras PROPIAS opiniones. Siempre estamos en una duda persistente por preguntarle a todo el mundo, que hasta le preguntamos a nuestro perro \"Â¿Debo hacer esto o aquello?\". Si sufres de este tipo de sÃ­ndrome, pregÃºntale al espejo mejor, al final Ã©l serÃ¡ el que tendrÃ¡ la respuesta que tu estÃ¡s buscando.  Creo que es mucho mejor si pones manos a la obra con las opciones dadas, obviamente tomando la mejor decisiÃ³n que te enriquecerÃ¡ con conocimiento, y te empujarÃ¡ hacia el camino al Ã©xito. Haz un omelette por tu cuenta propia, Â¿Y que si no eres cocinero? Siempre se aprende fÃ¡cilmente, si tomas la riendas por tu cuenta. Estar al dÃ­a con las tareas requeridas para lograr las opciones tomadas te darÃ¡n un mejor sabor de ti mismo, en lugar de que alguien mÃ¡s haga las cosas por ti. A prueba y error, el omelette tendrÃ¡ mejor sabor cada vez.  Asi que pregÃºntense, Â¿Donde comienzo? y Â¿Donde termina? (Te darÃ© una pista, no termina, solo se modifica).  A comer todo el omelette! Buen provecho!  '),(12170,3501362,'31,May,2004','2004-05-31 00:00:00','   Sorry, I just recently enabled the \"Anyone comment\" issue, now everyone that comes into this blog can leave their opinions.   Please, do humankind a favor and comment, it will be good for you!  '),(12171,3501362,'30,May,2004','2004-05-30 00:00:00','   Ok, a new day arrives...  I must confess, yesterday I was a bit tired of studying all day I decided to let my mind rest for a while with a good movie, so I went to see \"The Day after Tomorrow\". I don\'t know if I should thank the \"hour and half\" I spent driving around the mall for a parking space, or the fact that it was Sunday and EVERYONE gets together to go to the movies, but I got there LATE, and there was no early shows (at least not til 9:00 pm). So I figured, I might as well catch another movie (these days there aren\'t many good movies showing, and to my taste, I rather rent a foreign than see some cheap action movie), so there it was: \"Something\'s Gotta Give\" starring Jack Nicholson and Diane Keaton.  First off, WHAT A GREAT MOVIE! I would definitely recommend it to those who are just full of (and emptied out of) life and love. (life seems to be the subject in matter at the moment huh?). A movie teaching a good lesson, no matter how you see it, it all comes down to that main subject: In a relationship SOMEONE just gotta give.   This is a real issue that just doesn\'t happen to one person, but to a whole population out there, who have inmense theories about love, and \"letting go\" and those sappy thoughts about \"If you love someone let it go, if it comes back it was yours to keep, if it doesn\'t, it just wasn\'t meant to be\". Ok, so part of that is true, but what if real true love requires more effort than just sitting with your arms crossed staring at the horizon waiting for anything to happen?   It\'s a matter of playing your cards right. Destiny does happen, but for those who seek their fortune, not letting fortune come to them. I\'ve had those experiences in the past, and that\'s why I\'m sharing this. You just let go because it\'s needed, but destiny just comes around once in a while and says \"HEY STUPID, I\'M HERE COME GET ME!\", while you\'re just goofin around through unimportant issues in your daily life. It\'s a matter of choice, a matter of dedication, of giving, of reaching higher than expected.   It\'s funny, sometimes I would have never thought that destiny throughout the years gives out oportunities, it\'s like a winning lottery. You\'ve got your number, claim your price when the time is right. And it doesn\'t just comes around once in a lifetime, it comes MANY times in a lifetime, the problem is choosing, and not setting your mind straight over what you really want.  As humans, we are so fooled by our failures, by our deceptions, and mostly about our defects. Our insecurities are based on that, and that is why we sometimes feel that it\'s not just \"the right time\" to give in to love. Let me give you the real deal here: THERE IS NO RIGHT TIME IN LIFE. It\'s just about seeking and taking risks when your number is up. You bet and gamble, and sometimes you win, sometimes you lose. That\'s the way to really get what\'s going on in your love life.  So you\'ve got feelings for that person sitting next to you in class? Ask him/her out for good... what\'s there to lose? Integrity? I rather lose that, than have a guilt trip lasting for a long time. How about while BEING in a relationship: So it all turns to routine, ask yourself what\'s there to lose, and what\'s there to risk to actually change that routine for merely a second (worth lasting a lifetime).   I quote The Darkness: \"I believe in a thing called love\"... So I think I\'m not gonna just sit here and wait, I\'m gonna build up strategies to find her, or just find her again (who knows, sometimes true love is just around the corner, you just never looked around, eyes wide open to actually SEE IT).   It\'s about destiny, but also about CHOICE! People, make sure you beleive in your choice, no matter how good or bad it is, as long as it gives you a lesson (and a reward, who wouldn\'t like that!). Or else consider yourself warned, one word: PROZAC.  In the end, don\'t be so proud; someone just gotta give in...  '),(12172,3501362,'30,May,2004','2004-05-30 00:00:00','   urlLink    I should be a Heineken spokeperson don\'t you think? Anyways it\'s the first picture I post so bear with me...  Posted by  urlLink Hello  '),(12173,3501362,'30,May,2004','2004-05-30 00:00:00','   Hey everyone, welcome to the REAL taco...  Pretty interesting for this to be my first blog note, actually my first public note EVER on the Net. Usually I keep things to myself, but hey, there\'s always a time for one to come out with crazy ideas...  Let\'s start off, topic of the moment:   Ask yourself this, is it cool for a person who just wasted 5 months of his life studying something he\'s definitely not fit to do to actually give it one last chance in studying, right when it\'s 10:00 pm, knowing you have just a couple of hours worth of reviewing a whole semester, while you spent the rest of the day figuring \"What\'s life for\"? Moralists might say \"You dumb idiot (redundant speak), where have you been all day?\", some just might think if you\'re not worth to it, don\'t waste it anymore, and some will go \"Hey, 5 months and you didn\'t do crap, why waste one more minute knowing it\'s worthless\"?   Okay that\'s that, battle your whole day/night with it, I\'m sticking to the fact that I did bad, but I must give it one last shot, even if it\'s not meant for me to finish it at all. THAT\'s what I figured out today.  In all terms possible, life just happens sometimes, and sometimes it doesn\'t at all. And I\'m not writing to discourage, but in fact to give a word of advice; maybe this all seems so wrong and confusing but that\'s the way life deals the cards. There are people who persevere no matter the battle (they even fight other people\'s battle), and some people that just know when it\'s exactly the end of it all. I seem to be in that impassÃ©.   So here I am, 20 minutes past, still writing about what life should be about. But instead, you know what? I\'m giving life a living chance, for once letting me run the last quarter mile. I might as well run like a fool, but feeling deep inside that there\'s no lost cause in the end as long as it gives you all the tools and the knowledge needed to LEARN.   That\'s a general rule, something they might never teach us in high school, nor college, you don\'t get a Masters Degree in Life you know? Hell, if it all were that easy.   What\'s better yet is the whole beauty around it... It\'s not about giving up, about realizing your failure, but recognizing that for once you did a job, you actually proposed yourself a goal, even if you didn\'t make it, it\'s wise to learn about how you drove down that road to Victory Lane, for the next time you will drive with better skills, and maybe then you will reach that checkered line once and for all.  It\'s great that life DOES give us second chances, we\'re just too fooled by deception to actually realize it\'s THERE.   So my dear friends, a word of advice: There\'s never enough life for us to achieve our dreams, but there\'s always a chance you might just make it if you keep those dreams ALIVE.  ... I must go back to study, see you guys tomorrow!  '),(12174,3501362,'25,June,2004','2004-06-25 00:00:00','   Hey everyone... It\'s been a while, a LONG while, so much has happened, happy, sad, good, bad, just about every single flavor in the whole existing emotional rainbow.  Then again while a lot happened, it\'s time to move on forward right? I\'m not gonna go into details of the past, but of a recent event that has happened to me just a few minutes ago...  Has anyone experienced what I would describe as a bump into pure fortune? Perhaps I might have, but one wouldn\'t know unless it\'s serendipity (in case you don\'t know about this word, it\'s defined as \"The gift of making delightful discoveries by pure accident\"). And I do beleive it happens sometimes, we\'re just blind by being unfortunate most of the times, we don\'t give luck a chance. I mean, who would sit in front of his computer at 1 a.m. in the morning dwelling through the Internet when all of a sudden in one of your messenger systems (MSN or Yahoo in case) someone pops up, by accident or ID misspelling and starts off with you on a nice conversation, probably meant to someone else but fits just right into a dialog you would definitely have with someone, or feel to have with that someone that very moment?  How about being too lost in that misfortune that you don\'t easily realize that what\'s happening right there is the medicine to your cures for all the emotional sickness you\'ve been whimpering about for days? I beleive that happens to us every single day of our lives; we don\'t exactly open our eyes to realize the whole scenario, but just fit the details we like about it, and discard the rest.  So the dialog went on, and it was fun, until she realized she had been talking to the wrong person all along (but to me felt right, get my point?). So she just apologized and closed everything off... BLANK... POOF... dissapeared. Why does that happen? How do make the next move after that? WHICH ONE is the right move anyways? I just went ahead and apologized back (through e-mail) to her for not letting her realize she had been talking to the wrong person since the very first minute.  I don\'t really expect a reply (though one would wish... why does fun has an end? The answer to this simple riddle is that if it didn\'t had an end, fun would just be FU... get it?), just being myself makes it all simple, and for the moment\'s worth I sure did take the whole buck out of its value.  Yet I wish I could just grab the whole buck, instead of just 5 cents, out of every single moment I\'d live in my life... that would be interesting, cause then I wouldn\'t have to trip and fall over and over, I would just know exactly all the schematics to a good life. Grabbing the whole buck, getting your money\'s worth out of a good time is not that simple: You\'d need to digest everything that happens through that moment.   So serendipity this... Did she reply an hour later? Yes she did.   Did I reply back? Gimme your best guess ;) (though by now the answer is simple isnÂ´t it).  I\'ll share you the real secret of enjoying a good moment: NEVER generate any expectations over it, just let the moment HAPPEN by itself. If you expect, chances are you won\'t be getting what you\'ve been asking for. Secondly, beware that a good moment has it\'s troubles either at the beginning or at the very end, while the ripe stuff, perfectly fitting for the moment, is in the middle, the edges are very important: CHOOSE always the right ending, with the right beginning.  And let life happen okay? If you don\'t, and you let life go by itself through time, chances are... serendipity will barely happen with you.  So, to that person who gave me a right start thought wrong, thank you tonight for letting me grab the whole buck out of our personal conversation! Maybe it all ends here, it all might go on, I don\'t know. We\'ll just wait.   I hope I did make the start with the right foot.  Check you all on the flipside!  '),(12175,3501362,'03,June,2004','2004-06-03 00:00:00','   Well, back again, trying out my creative writing. I seem to enjoy pretty much what I\'m doing at the moment, writting stuff that comes out of my picked out brain.  Tomorrow (actually later today, it\'s 1 am here) I\'m gonna have the oportunity to go to one of the highest places in Guatemala: TacanÃ¡. Ok so it\'s work duty, but still, leaving that aside, it\'s a long journey, way off road to a far away place in this beautiful country.   I find myself thinking, how many journeys have I taken really? The ones that fall so deep in your heart that every time you just even drive by a certain scene, or dust off some equipment you used somewhere, reminds you of that special time, which you took a trip to somewhere, and memories just start falling into place. I figured, I\'ve had my dose of journeys in my early life.  They aren\'t just \"journeys\" to somewhere specifically, their meaning can be found deeper than just a memory, it lies within the nature of reaching farther, testing your skills and fighting adversity in your finest hour. And what\'s best is that it doesn\'t require measuring time, nor distance, nor resources used, it just requires understanding, comprehending deep within you why was that journey made.  In the end, we all have done this one way or the other. Sometimes in the company of a special one, sometimes on our own. I would guess that on our own is where we test what we really are made up of.   As humans we tend to be more nomads, less sedentaries. Why would that be? It\'s because we try to live with ourselfs in our well settled and restricted environments. But some people aren\'t just satisfied with that. Some people yearn for the journey, it\'s in their blood; curious enough, some might look at them as masochists, getting themselves involved in the most difficult situations in any way you might see it. I respect these people more than settled ones, because as they go throughout their journeys, they develop a higher ambition towards different goals to meet. Few make it until the end, others just let their hopes and dreams die as they walk through, many of them looking back and retracing the wrong steps taken to their apparent \"doom\".   We are left blind by negativity sometimes, specially if by trying to get somewhere you just see a dense fog ahead of you. And this just doesn\'t happen unless we start regreting all the done deeds in the past. Acceptance is a virtue few acquire as they go forward, obviously looking towards making a change for themselves and their environment.  My life has been a very complicated journey so far, and I\'ve tested many times how far can I reach; I just seem to get stuck most of the time in that fog, not realizing that no one has the skill to look past that fog, though the ones that keep on walking through it understand one thing: They live on the moment wisely, for each step they make they are certain is the right one towards the top of the hill.   That\'s what clearly divides the success from failure itself. In life, love, friendships, professional issues and business itself, different thick fogs appear before you. You just gotta beleive in yourself in that moment, that it is all possible with faith, and hope of your own self making it to that desired goal. We even do it naturally without realizing that, and so when we look back we can see just downhill, and guess what? You\'re at the top.   Believe in yourself always, have great humility kept in your soul, accept others and accept yourself through others, forget the fog in front of you, remember the bumps left behind, and keep on a straight walk towards the top of the hill.  The longest journey is still to come, when you see there\'s more past the sky. The sky is not the limit, the only limit that exists is YOURSELF. So let go of the idea that there\'s a limit to reaching dreams, and you will see those dreams appear consequently.   So it\'s time for me to journey into life, I\'ll let you guys know if I reached the top.  Wish me well! -----  PS: Thanks to everyone who has left a comment in my blog, you can\'t imagine how happy it makes me that I can reach someone with my words and my thoughts. You know who you are, and what you are to me so all I can say is that I thank you lots, it means a lot to me.  '),(12176,3501362,'01,June,2004','2004-06-01 00:00:00','   Post traumatic experiences of self defense while being sick: Uncounsciously we curl up to a fetal position, just \"to feel\" better... ahh don\'t you miss being in your mother\'s tummy? No? Well if you do that, then your subcounscience does miss that at least; It\'s the mother nature in us giving us protection.  '),(12177,3501362,'01,June,2004','2004-06-01 00:00:00','   My last day of the whole semester, and YES I walked the plank and sank with the fishies, I ran the last quarter mile and couldn\'t finish at all. Oh well, that\'s what I was expecting anyways right? At least I do feel proud of myself for taking that last risk, instead of sitting in the middle of nowhere with my arms crossed. Interesting enough, I don\'t feel discouraged to end my journey here.  See, there are special times in your life where you are given many choices, it\'s like a whole buffet of post-failure choices, but it\'s not cooked nor prepared already. So you want an omelette? Here are the eggs, the ham, cheese, whatever you like in it. Or shall I cook it for you? Will it taste better? Or wouldn\'t you like it to cook it yourself, find out your own taste?  Life is like that. We are given many post-choices either by success or failure. The problem is that not many out there are fit to prepare a self-made omelette of a choice made. Some just take the easy way, contact an expert \"chef\" to prepare it as close as it would please them. But what if you\'d like ketchup or some other condiment in it? No no no, here is the real deal with making your own.  Sometimes we\'re just not prepared to deal with the post-choices life gives us, but there\'s no other way around it, life just keeps on going, it doesn\'t care at all if you catch up and stay up to date, or be left behind with the rest. In the end, growing up no matter what age, it\'s all about experience: a key word in any road to success.  We all have a certain taste for new experiences, no matter how good nor how dissapointing and bad they are, it all sticks inside that big head of yours. And we learn from that so that we perfect our motives and our strategies to achieve goals. We are stubborn by nature, and sometimes find it hard to take decisions based on our OWN opinions. We are always in that persistent doubt and ask everyone, even the dog \"Should I do this or that?\". If you suffer from this kind of syndrome, ask the mirror instead, in the end he\'s the one that has the answer your looking for.   I think it\'s all better if you do get hands-on with the options you are given, obviously taking the best choice that will enrich you with knowledge, and push you ahead a road to success. Make an omelette yourself, so what if you\'re not a cook? There\'s always easy learning, and that is by taking matters in your own hands. Keeping up with the required tasks for the given choices will definitely give you a better taste of your real self, rather than having things done for you. Trial and error, the omelette will taste better each time.  So ask yourself, where do I begin? And where does it end? (I\'ll give you a clue, it doesn\'t end, it just gets modified).   Enjoy your omelette, bon appetit!  '),(12178,4024747,'30,July,2004','2004-07-30 00:00:00','   Two nights ago was the most wonderful night.  Hannah learned how to kiss, or at least was wanting us to give her kisses.  She cannot do the pucker and smack noise, but she puts her lips together and says, \"mmmm\", and then she waits for us to come in and plop one on her lips.  After that she just laughs hysterically. She gave me and Dan kisses back and forth for 15 minutes.  We tried to get it on camera, but she would only give me one and then wanted to play with the camera as always.  It really was so wonderful.  It is amazing all the new things she learns each and every day!   '),(12179,4024747,'27,July,2004','2004-07-27 00:00:00','   Well I guess I won\'t be writing every single day as a mommy\'s life gets very busy. Today HJ was as cute as usual.  We recently bought her new pink shoes from Target and at many times during the day, she brings them up to me to put them on her.  She loves walking around in them and especially loves the fact that she can walk outside in them.  She is so adventurous and is growing up so quickly now.  Everyday she is sounding out more words to say, like oof-oof for a dog (woof-woof).  Apart from that, here is an unusual tidbit:  We were eating McDonald\'s french fries tonight with a bowl of ketchup.  The next thing I know Hannah is pitching her sock in the ketchup as though that were the exact place it belonged.  It was kind of funny, but I had to tell her that was not where we put the dirty socks :) I don\'t think she really got it, but oh well.  She is my little angel pie.   '),(12180,4024747,'24,July,2004','2004-07-24 00:00:00','   My little girl is getting to be so big and grown-up.&nbsp; I am as proud as I could be of her.&nbsp; Right now she is sitting on the floor reading her Baby Einstein book so intently like she understands everything that is in it- even though she can not talk yet let alone read.&nbsp; It has all sorts of animals in it, which&nbsp;are her favorite things.&nbsp; She also is a spunky independent little girl.&nbsp; She loves looking through pictures, of which her favorite are the ones of herself.&nbsp; I have two large pictures, 8\" x 10\", of her on my dresser, and she always points to them when she wakes up from naps (she naps in our room) saying, \"dat dat\".&nbsp; That means \"that\".&nbsp; She pointed at the fish on our shower curtain today, and I told her those are fish. She said \"ish-ish\".&nbsp; Her favorite shows are Blue\'s Clues (the new ones with Joe), and the Wiggles.&nbsp; With the Wiggles she shakes her hips back and forth to the songs and with Blue\'s Clues she is just glued to the television.&nbsp; Mostly though aside from TV, she loves to just play with her toys.&nbsp; Her favorite is her green wagon for two.&nbsp; She is in and out of that all day long.&nbsp; She is in and out of every room all day long too.&nbsp; She loves her room and our room a lot.&nbsp; Those are her favorite two to explore.&nbsp; Well that is all for today.&nbsp; I thought I would just start out with some basics.&nbsp; Hope you enjoy!   '),(12181,4024747,'16,August,2004','2004-08-16 00:00:00','   Well it has been quite awhile since my last blog.  Hannah and I have been sick- I think we had the whooping cough- but I guess it is pretty prevalent in the south because of the air quality and mold and all that.  Anyway we are finally over the hump and so much has happened, but the biggest thing is that Hannah went poo-poo on the potty the first time yesterday!!  I saw she was trying to go so I rushed her in there to go to the bathroom.  She didn\'t really like it that much.  She fussed and cried but ended up being happy when it was over!  Anyways, the potty training is going slowly as you can probably guess.  We may have to wait awhile.  Also we decided on a name for our new baby girl- Lily Ann Cook.  (Not sure on the spelling of lily yet.) Well feel free to email me if you want to anyone- I would be happy to hear from all of you!  I will write more later!   '),(12182,4024747,'02,August,2004','2004-08-02 00:00:00','   Well I have a few different details about the precious little Hannah to share tonight.  Earlier today she was playing in her green wagon in the sun room.  I was watching tv in the living room.  The two rooms are completely connected just so nobody freaks out.  Anyway, the commercial with the chips ahoy chunky chocolate chips came on.  You know the one that says, \"  I\'m a monkey chips ahoy, I\'m alright...\", and then there is the singing clamation chocolate chip cookie and the monkeys on stage. Well Hannah heard it and perked right up. She got out of her wagon as quick as possible and ran in the the room just to see that commercial.  But by the time she got in there, it was on the tail end of the commercial. When it was over, she headed back to her wagon.  Well then I started singing the song off the commercial, because it is kind of one of those songs that just gets stuck in your head.  She turned right back around and ran in to see if it was on again.  I felt so bad, but it was so cute and funny.  I don\'t know what it is about that commercial, but she can hear it from  anywhere  and come running just to be able to see it.  It is the only commercial she does that with.       We also today were playing in her room. I was looking through all of our baby clothes, to get ready for the new baby girl coming along, and I found this little sleep sack that Hannah used to sleep in.  It is basically a sleeper with no feet and just a sack at the end so baby can\'t kick it off like a regular blanket.  I was putting it up to Hannah to size it up and see how big it was and I decided to put it on her.  Well it was a little tight, so I just unzipped it and left the top part on her arms and let the back go out like a cape.  She loved it!  She had that thing on for thirty minutes.  She was just walking around having a godd old time!  I tried to take it off of her once because I though she was getting tired of it, but when I did she started bawling.  So I quickly put it back on and she was off again!        Here is another one. This one is so funny.  One day a couple days ago, Hannah was walking around just giggling.  I conldn\'t really figure out what was going on until she came a little closer.  She kept lifting up her shirt, pointing at her belly button, and laughing!  It was so funny. Then she would do it all over again.  She has been doing this a couple of days now!  Well I think that is all for tonight!   '),(12183,4109605,'13,August,2004','2004-08-13 00:00:00','   ç”Ÿæ—¥éŽå’—å•¦ã€‚ã€‚ã€‚å’³ï¹åˆè¦ç­‰æˆå¹´å•¦ã€‚å¸Œæœ›å¯ä»¥åˆåŒä½¢åŽ»å”±Kï¹å¥½ä¼¼ä¸Šæ¬¡å’é–‹å¿ƒï¹—ä½†ç³»éƒ½å””è¦‹ä½¢onlineæ—¢ã€‚ã€‚ã€‚å¥½æŽ›ä½ä½¢å‘€ï¹—ç‹¡åˆ°æˆ‘æˆæ—¥â€œç¾…ç¾…äº‚â€ haha...æˆ‘å°‹æ—¥printå’—æˆ‘åœ°å¼µç›¸å‡ºç†ï¹ä»²æ“ºå’—å…¥éŠ€åŒ…åˆ°ï¹—éƒ½æ•´å¾—å·±éšæž¶^^ã€‚æ›´ç³»å•¦ï¹— æˆ‘åŒä½¢å–Žï¹—å“ˆå“ˆï¹æˆ‘ç³»å””ç³»å¥½ç™½ç—´å‘€ï¹–æˆ‘éƒ½å””æƒ³æž¶ï¹å’æˆ‘çœŸç³»å¥½é˜ä»¥å‘€é¦¬ã€‚ æˆ‘å²©å²©å¹«çˆ¸çˆ¸æ•´å’—å¥½å¤šåŸºï¹å¥½ç´¯å‘€ã€‚è½æ—¥å‘€åª½è¿”HKå•¦ï¹—æˆ‘è«—æˆ‘æœƒå¥½æŽ›ä½å§–ï¹—å’³ã€‚ã€‚ã€‚ã€‚ã€‚ä½†ç³»ä¸€å®šæœƒç­‰å§–è¿”ä¾†ï¹ç­‰å§–Dæ‰‹ä¿¡ï¹—  '),(12184,4109605,'11,August,2004','2004-08-11 00:00:00','   ä»Šæ—¥ç”Ÿæ—¥æœ‰åŒä½¢ä¸€é½ŠéŽï¹—åŽ»å’—å”±kï¹ä½¢ä»²åŒæˆ‘åˆå”±å’—ä¸€é¦–æ­Œï¹å¥½å¥½æˆ€æ„›^^  æœ‰ç¨®çªç„¶é–“å¥½æ„›ä½¢æ—¢æ„Ÿè¦ºã€‚æˆ‘åŒä½¢å½±å’—å°‡ç›¸ï¹—å˜©ï¹—ä½¢åŒæˆ‘å—°é™£æ™‚çœŸç³»å¥½è¿‘ï¹ä½¢æ“ºå’—ä½¢éš»æ‰‹å–ºæˆ‘è†Šé ­é“ï¹å¥½ä¼¼ã€‚ã€‚ã€‚å“ˆå“ˆï¹æ‹æ‹•å—°Däººå’^^ ä½¢ä»²å¥½ä¼¼æ¯”å’—D hints æˆ‘å’ã€‚ ä½¢è©±LINAè©±ä½¢ç³»è±¬æ‰’ï¹ç„¶å¾Œä½¢å•æˆ‘é˜å””é˜ä»¥è±¬æ‰’... æˆ‘å°±ç„¡ç†ä½¢å’ï¹é€ƒæ¯”å‘€å˜›ã€‚æˆ‘çœŸç³»å¥½æƒ³åŒä½¢è¬›æž¶ä»Šæ—¥ï¹ä½†ç³»ç„¡å‹‡æ°£æŽ¥å—ç­”æŒ‰ã€‚ç„¡å‹‡æ°£è¬›å¿å•¦ï¹—ä½†ç³»ä»Šæ—¥çœŸç³»å¥½é–‹å¿ƒã€‚ä½¢ä»²è²·å’—é­šè›‹æ¯”æˆ‘é£Ÿã€‚é›–ç„¶æˆ‘é£Ÿå””åˆ°è¾£ï¹ä½†ç³»éƒ½æœ‰é£Ÿã€‚ä½¢é€æ¯”æˆ‘ï¹æ›´ç³»å””æ™’å¾—å•¦ï¹—ä½†ç³»ä½¢é€æ¯”æˆ‘å·±ç¦®ç‰©ï¹ä¸€åŠç³»gigié€å·±ã€‚çœŸç³»è¦ºå¾—åˆæ„›æœ‰æ¨ã€‚æˆ‘ä»¥ç¶“ç•¶æ™’ç³»ä½¢é€æž¶å•¦ï¹—ä½¢æœ¬ä¾†è©±é€å—°å€‹QOOæ¯”æˆ‘ï¹æˆ‘ï¹ä½†ç³»è¢«äººè²·-å’—å•¦ã€‚æ˜¯ä½†å•¦ï¹—ç³»ä½¢é€æˆ‘éƒ½æœƒå’æ„›ï¹—  '),(12185,4109605,'09,August,2004','2004-08-09 00:00:00','   æˆ‘è€ŒåŠ ç•¶ç„¶å¥½é–‹å¿ƒå•¦ï¹—ä½¢è«‹æˆ‘åŽ»ä½¢å±‹ä¼å–Žï¹—ä½†ç³»æˆ‘å²©å²©å…ˆè«—ä½æ”¾æ‰‹ã€‚é»žç®—å‘€ï¹–ç³»å””ç³»å€‹å¤©æƒ³æˆ‘é¢å°ï¹ç„¶å¾Œæ”¾æ‰‹ï¹–å¥½å¤šé‡Žéƒ½å¥½é›£ä¼°ï¹å› ç‚ºä¹œé‡Žéƒ½æœ‰å¯èƒ½ã€‚å¦‚éŽæˆ‘å’è½åŽ»ï¹ç³»å””æœƒçŸ¥åˆ°çœŸç›¸ã€‚è½æ—¥ï¹å¦‚éŽæˆ‘å°ä½¢å¯ä»¥ç„¡FEELï¹å’æˆ‘å°±çŸ¥åˆ°æˆ‘æˆåŠŸã€‚å¦‚éŽæˆ‘åŒä½¢ä¸€é½Šoå€‹é™£æœ‰FEELï¹å’æˆ‘çœŸç³»æ„›å¾—ä½¢å¥½æ·±ã€‚æˆ‘è«—å¦‚éŽç³»å’ï¹æˆ‘æˆ–è€…æœƒå¸¸è©¦åŽ»åŒä½¢è¡¨ç™½ã€‚æ­»å•¦ï¹å°‹æ—¥åŒä½¢å‚¾åˆåˆä»¥ç¶“å’èˆˆå¥®ï¹å’æˆ‘çœ¯åæˆåè¦åŒä½¢è¡¨ç™½ï¹–ï¹—ï¹–ï¹—ï¹–ï¹— å±Œã€‚ã€‚ã€‚ã€‚ã€‚ æˆ‘æœ€å””è­˜åŒäººè¡¨ç™½æž¶å•¦ã€‚oç³»ã€‚ã€‚ã€‚æ­»å°±æ­»å•¦ï¹—  '),(12186,4109605,'08,August,2004','2004-08-08 00:00:00','   æˆ‘åŒä½¢å¯¦åœ¨å¤ªå¤šå·§åˆå•¦ã€‚æˆ‘åœ°å‘€å“¥ç³»è­˜æ—¢ã€‚CåŒVå–ºå€‹keyboardé“ç³»å–ºéš”è£¡æ—¢ã€‚å¤§å€‹æˆ‘åœ°æ—¢å¤¢æƒ³éƒ½ç³»ä¸€æ¨£æ—¢ï¹ä½†ç³»å””çŸ¥ä½¢æœ‰å†‡æ”¹è®Šå•¦ã€‚æˆ‘å’ç¨ç‰¹æ—¢å¯«å­—æ–¹ç™¼åŒä½¢éƒ½æœƒç³»ä¸€æ¨£ã€‚ä¸Šå¹´æˆ‘åœ°Dç§‘ç›®ç³»ä¸€æ¨£ï¹ä½†ç³»å””åŒè€å¸«ã€‚ å””ç³»å–Žï¹å¥‡å¯¦å””ç³»å¥½å¤šï¹ä½†ç³»æ¯ä¸€å€‹éƒ½ç”˜ç‰¹åˆ¥ï¹æ‰€ä»¥æˆ‘ä»¥ç‚ºç³»ä¸Šå¤©æ¯”æˆ‘æ—¢ä¸€Dé¡¯è¦–ã€‚æˆ‘ä¾+æƒ³è¿”ï¹å¥‡å¯¦æˆ‘åœ°éƒ½å””ç³»å¤ªå¤šç›¸ä¼¼æ—¢é‡Žã€‚èµ·é¦¬ä½¢é«˜æˆ‘é­å…ˆå•¦ã€‚å’æˆ‘æ„›ä½¢éƒ½ç„¡æ™’ç¾©æ€å•¦ã€‚æˆ‘ä¾+â€œè‡¨â€ç·Šï¹æ„›å””å¯ä»¥è¨ˆæž¶å‘€å—Žã€‚åˆæ··äº‚æ›¬å•¦-_-â€œ å¥½æ†Žæˆ‘è‡ªå·±ã€‚å•å¤ªåˆç³»æˆ‘æƒ³å‡ºç†ï¹ç­”æŒ‰åˆç³»æˆ‘æƒ³å‡ºç†ã€‚ååå¾©å¾©ï¹ä»¥ç¶“å¥½å¤šæ¬¡å•¦ã€‚é»žç®—å‘€ã€‚æˆ‘ç³»å””ç³»æœ‰ç—…å‘€ï¹–å±Œã€‚  '),(12187,4109605,'07,August,2004','2004-08-07 00:00:00','   æˆ‘å°‹æ—¥å…ˆè©±éŽæˆ‘æœƒæ”¾æ‰‹ï¹å€‹å¤©å°±ååè¦æˆ‘å¤¢ä¸­è¦‹åˆ°ä½¢ã€‚æˆ‘æš‘å‡ç„¡ä¸€æ¬¡å¤¢åˆ°ä½¢ï¹é»žè§£ä¸€è©±è¦æ”¾æ‰‹ï¹å°±è¦è¦‹åˆ°ï¹–é‡è¦ç³»ä½¢é«˜å¤§å¨çŒ›å—°éš»ã€‚æˆ‘å–ºå¤¢é“ï¹å–ºå·´å£«è¦‹åˆ°ä½¢ã€‚ä½¢å¥½ä¼¼å””è­˜æˆ‘åœ¨å…ˆï¹ä½†ç³»ç•¶æˆ‘è¡ŒéŽåŽ»åŒä½¢è¬›é‡Žï¹ä½¢å°±æœ‰å›žèªå•¦ã€‚å””çŸ¥ç³»å””ç³»å€‹å¤©æƒ³è©±æ¯”æˆ‘çŸ¥ï¹ä½¢è·Ÿæœ¬å€‹å¿ƒå…¥é¢ç„¡ä½æ“ºæˆ‘ã€‚å€‹å¤©è¦å¹«æˆ‘é¢å°ç„¶å¯¦ï¹å’å…ˆå¯ä»¥æ”¾å¾—æ‰‹ã€‚å¸Œæœ›ç³»å•¦ï¹æˆ‘æœ¬ä¾†éƒ½æœ‰Då””ä¿¡å€‹å¤©ï¹ä½†ç³»æ‡‰è©²ç„¡å’æ®˜å¿å•©ï¹–ç„¡æ—¢ã€‚ã€‚ã€‚ å””è¬›å•¦ï¹åˆé ­ç—›å‘€ =[ å†è¦‹ã€‚  '),(12188,4109605,'06,August,2004','2004-08-06 00:00:00','   wahh... not thinking abt him... i also hv headache... y is tht? he is not tht important to me la, why do i always get reminded of him? i don\'t know anything about my feelings rite now... it\'s like i\'m a stranger within myself... it\'s like i can\'t tell love from hate, or happy from sad; it\'s all so neutral... the tv has been part of my life ever since i was born... i learn most from it =]. but this time my situation was never portrayed by a tv show, i hv nvr received any info or seen any situation like my own. i don\'t know wat 2 do... it\'s like w/o a fake portrayal, i can\'t do it in real life. i need choices or at least one choice to guide me thru wat is happening rite now. it\'s too complicated... i don\'t kno if i can take it n e more. it\'s all so blurry around me, it\'s all so confusing. it\'s all so detailed, in a way i want to avoid it all. rite now all the simple things seem so much better cuz it takes so much less to solve. btw under all this confusion, i believe i have lost more than i gained. i\'ve lost friends, my life, my courage to face the truth, my trust in other people, my faith in a future w/ him, happiness that i used to have wen there wasn\'t confusion, time that cud be spent happily... i\'ve broke my promises, my heart... but i\'ve also broke down that wall that stopped me from continuing life. that wall that caused me to be stubborn and love someone that won\'t love me back. i don\'t want to waste time on someone that won\'t love me TRUELY. writing all this down, it did take me to another phase in life. i\'ve learned to let go of what i never had, and i believe if i have this principle in my heart, nothing can cause all this confusion ever again. thank you for not receiving my love. i used to be too lucky, now i\'ve learned that without luck i can still stand tall and live a great life. thank you.  '),(12189,4109605,'05,August,2004','2004-08-05 00:00:00','   æˆ‘ä»Šå¹´ç”Ÿæ—¥å¥½æƒ³åŒä½¢ä¸€é½Š éŽï¹—å¥½æƒ³åŒä½¢é£Ÿç•ªé¤å‹ï¹èµ·é¦¬ æˆ‘åœ°å…©å€‹éƒ½æœƒé–‹å¿ƒå‘€å—Žã€‚å””çŸ¥ä½¢æœƒå””æœƒä¾†å‘¢ã€‚æˆ‘å¥½æƒ³ä½¢ä¾†ï¹å› ç‚ºæˆ‘å¥½æŽ›ä½ä½¢ã€‚ æˆ‘å²©å²©ç‡å®Œ Placebo Cure, çœŸç³»å¥½å¥½ç‡ï¹—å‘€Joe éšä»”å‘€ï¹—å˜©ï¹å¦‚æžœæœ‰ä¸€å€‹å¥½chiå‘€Joeæ—¢äººè¿½æˆ‘... ^-^\'\' å’³ã€‚ã€‚ã€‚ é»žæœƒå–Žï¹æˆ‘ç¾åœ¨éƒ½ç„¡ä¹œäººè¿½æž¶å•¦ã€‚è²»äº‹saiå¿ƒgeiå‘€ã€‚ã€‚ã€‚ æˆ‘ç™¼è¦ºï¹æ„Ÿæƒ…å“©D é‡Žï¹å””å¯ä»¥è¨ˆæž¶ã€‚é FEELæž¶ã€‚å°ä¸€å€‹äººç„¡FEELï¹å°±å””å¯ä»¥æ¯”å¸Œæœ›ä½¢åœ°ã€‚æ„Ÿæƒ…ï¹å””æ‹•å¾—ï¹å””ç³»å°±gowå‡ºå€‹å¤§é ­ä½›æž¶å•¦ã€‚å””çŸ¥æ¨£è²Œå¯å””å¯ä»¥å½±éŸ¿åˆ°FEELå‘€å“ªã€‚ æˆ‘è·Ÿæ´¾èˆˆåŠƒç•«ã€‚  '),(12190,4109605,'02,August,2004','2004-08-02 00:00:00','   å°å°å””é–‹å¿ƒå‘€ã€‚ã€‚ã€‚ å””çŸ¥boboç³»å””ç³»é¨™gunæˆ‘å‘€ã€‚ã€‚ã€‚ æˆ‘åœ°13å¹´æœ‹å‹ï¹å””é€šå’éƒ½å””å¤ ï¹–åˆè©±è¬›ç‘Dé‡Žæ¯”æˆ‘è½ã€‚åŽŸä¾†å””ç³»æž¶ã€‚æˆ‘æ ¹æœ¬å°kuiéƒ½å””é‡è¦ã€‚ 2004å¹´å•¦ï¹ç„¡Då’æ—¢æœ‹å‹æž¶å•¦ã€‚ å¥½ä¼¼ä»¤å¤–ä¸€ä»¶äº‹å’ã€‚ã€‚ã€‚ é‚£å€‹ä¸–ç•Œç„¡Då’å¥½æ—¢ç”·ä»”æž¶å•¦ã€‚ æˆ‘éƒ½å¥½æƒ³æ­»å¿ƒï¹ä½†ç³»å¥½é›£æž¶ã€‚ å””çŸ¥ç³»å””ç³»æ¯æ¬¡éƒ½æ„›å¾—å¤ªæ·±ï¹æ‰€ä»¥å…ˆå’é›£æ”¾æ‰‹ã€‚ èµ·é¦¬æˆ‘ç»å¥‰æ—¢ç³»çœŸæ„›ï¹å–å’—æˆ‘ï¹æˆ‘è«—éƒ½ç„¡äººæœƒå—å‚·ã€‚è²»äº‹å•¦ã€‚ã€‚ã€‚åšå’å¤šç›žsai. å¥½å””é–‹å¿ƒå‘€ã€‚ã€‚ã€‚ã€‚  '),(12191,4109605,'02,August,2004','2004-08-02 00:00:00','   è«—è¿”èµ·æˆ‘åŒä½¢ä»¥å‰Då˜¢ï¹æžåˆ°å€‹å¿ƒéƒ½ç”œç‘ã€‚ã€‚ã€‚ æˆ‘è¨˜å¾—æœ‰ä¸€æ¬¡ï¹æ”¾å­¸ã€‚æˆ‘è¦åŽ»orcheastra, ä½¢ç„¡äººé™ªè¡Œè¿”å±‹ä¼ã€‚å’æˆ‘å’ªé™ªä½¢è¡Œè¿”å±‹ä¼å›‰ã€‚å—°æ¬¡æˆ‘å¥½é–‹å¿ƒå‘€ï¹å› ç‚ºä½¢è´Šæˆ‘ï¹è©±æˆ‘å¥½äººã€‚æˆ‘åœ°å¥½å¥½å‚¾æž¶ã€‚ã€‚ã€‚ä»¥å‰ã€‚ä½¢è¬›éŽè©±æˆ‘ç³»ä½¢å…¶ä¸­ä¸€å€‹æœ€ä¿¡ä»»å˜…äººã€‚æˆ‘çœŸç³»å¥½æ„Ÿå‹•ï¹å¥½æƒ³åŽ»æ”¬å¯¦ä½¢ã€‚æˆ‘æˆæ—¥éƒ½ä¿è­·ä½¢ï¹ç„¡è«–ä½¢å°æˆ–éŒ¯ã€‚å””çŸ¥é»žè§£æˆæ—¥è¦ºå¾—ä½¢ç³»çŸ¥åˆ°ï¹ä½†ç³»å””ç•€åæ‡‰æˆ‘ã€‚oç³»ã€‚ã€‚ã€‚ é›–ç„¶ä½¢å””ç³»å¥½éšä»”ï¹ ä½†ç³»æœ‰ç¨®å¥½æƒ³åŒä½¢ä¸€é½Šæ—¢æ„Ÿè¦ºã€‚ä»¥å‰ï¹ç•¶æˆ‘æœ›ä½ä½¢å—°é™£æ™‚ï¹ä½¢æœƒæœ›è¿”éŽè£¡ã€‚æˆ‘æœƒå°ä½ç”œç”œåœ°ç¬‘ï¹ä½¢åˆæœƒç¬‘ã€‚çœŸç³»ä¸€å€‹å¥½é»žæ—¢å¤¢ã€‚é»žè§£ä¹‹ç³»ä¸€å€‹å¤¢ï¹–é»žè§£å””å¯ä»¥è®ŠçœŸï¹–ï¹– æˆ‘éƒ½ç­‰å’—æˆå¹´å•¦ï¹å””ç³»ä»²è¦ç­‰å‘€å—Žï¹– æˆ‘å¥½é©šæˆ‘ç„¡å‹‡æ°£å†ç­‰è½åŽ»ã€‚ã€‚ã€‚ã€‚ã€‚ã€‚ å“©æŽ’æˆ‘ç„¡è¦‹éŽä½¢,å¥½æŽ›ä½ä½¢å‘€ã€‚ çŠä¸Šã€‚  '),(12192,1668565,'11,July,2003','2003-07-11 00:00:00','   test '),(12193,1668565,'25,September,2003','2003-09-25 00:00:00','   urlLink Corporate Gestapo    \"NEVER in my career have I allowed a team which worked for me to think they had a 40 hour job.  I have allowed YOU to create a culture which is permitting this.  NO LONGER.\" It\'s nice to know that these sort of maniacs were running the global economy. '),(12194,1668565,'19,September,2003','2003-09-19 00:00:00','   urlLink Our future lies with eugenics    \"Eugenics, as advocated by kindly figures such as its pioneer, Sir Francis Galton, or its most eloquent exponent, Dean Inge, was simply the notion that the useful and intelligent classes should be allowed, indeed encouraged, to breed, and the murderous morons, who are never going to contribute anything except misery to themselves and others should be discouraged. No one need be killed.\" This was published in the bleeding  Telegraph  mind you. But the author is a kind nazi and doesn\'t think anyone  need  be killed, unless of course it\'s absolutely necessary. '),(12195,1668565,'19,September,2003','2003-09-19 00:00:00','   urlLink How low can Dow go?    \"In a stunning example of corporate insensibility, Dow Chemical, the world\'s largest chemical company, and new owner of Union Carbide, is to sue survivors of the 1984 Union Carbide gas disaster in Bhopal, India. While the site of the disaster lies covered in toxic waste and survivors struggle with continuing ill health and deadly pollution from the site, Dow has decided to add to their woes with an Indian lawsuit.\" Soundtrack: Dead Kennedys \"Kill the Poor\" '),(12196,1668565,'19,September,2003','2003-09-19 00:00:00','   urlLink Crime Pays!    \"In this special report, American RadioWorks correspondent John Biewen explores how some groups with vested interests work to influence public policy - helping to keep more people locked up longer.\" At last, slave labour one can afford. '),(12197,1668565,'19,September,2003','2003-09-19 00:00:00','   urlLink Bosses warn against rights for temps    \"Business leaders have warned that a leaked European Commission draft proposal to extend temporary workers\' rights could be very damaging \'to all concerned\'. \" '),(12198,1668565,'19,September,2003','2003-09-19 00:00:00','   urlLink Human Rights Violations in The United States - Red Onion State Prison    \"The day I arrived I was...told that I was at Red Onion now and if I act up they would kill me and there was nothing anyone could or would do about it.\"  Midnight express... american style. '),(12199,1668565,'16,September,2003','2003-09-16 00:00:00','   urlLink New terror laws used vs. common criminals    \"A North Carolina county prosecutor charged a man accused of running a methamphetamine lab with breaking a new state law barring the manufacture of chemical weapons. If convicted, Martin Dwayne Miller could get 12 years to life in prison for a crime that usually brings about six months.\" '),(12200,1668565,'09,October,2003','2003-10-09 00:00:00','   urlLink Keep \'em working and miserable   Sad workers may make better workers says study:  Psychologist Dr. Robert Sinclair and his former PhD student Carrie Lavis recently conducted a series of four studies addressing the effects of  experimentally induced happiness versus sadness  on work productivity by asking the participants to build circuit boards. In the first study, sad people committed significantly fewer errors than did happy people (approximately half the number of errors) but there was no difference in the number of boards completed. Thus, sad people were more productive.  Experimentally induced happiness/sadness is beyond my grasp anyway. '),(12201,1668565,'09,October,2003','2003-10-09 00:00:00','   urlLink Let us recount the names of the dead   Diary of a Texas death cult. '),(12202,1668565,'29,January,2004','2004-01-29 00:00:00','   urlLink deseretnews.com | Dossier program alarms Utahns   \"It sounds like a sci-fi thriller: a super computer program that gathers dossiers on every single man, woman and child â€” everything from birth and marriage and divorce history to hunting licenses and car license plates. Even every address you have lived at down to the color of your hair.       It sounds surreal, but former Gov. Mike Leavitt signed Utah\'s 2.4 million residents up for a pilot program â€” ironically called MATRIX â€” that does just that. And he never bothered to reveal details of the program to Utah citizens or to state lawmakers who, upon learning of the program on Capitol Hill this week, are now worried the state could be involved in a program that jeopardizes basic civil liberties.\" '),(12203,1668565,'19,January,2004','2004-01-19 00:00:00','   urlLink The Register    \"Once you buy your RFID-tagged jeans at The Gap with RFID-tagged money, walk out of the store wearing RFID-tagged shoes, and get into your car with its RFID-tagged tires, you could be tracked anywhere you travel. Bar codes are usually scanned at the store, but not after purchase. But RFID transponders are, in many cases, forever part of the product, and designed to respond when they receive a signal. Imagine everything you own is \'numbered, identified, catalogued, and tracked.\' Anonymity and privacy? Gone in a hailstorm of invisible communication, betrayed by your very property. \" '),(12204,1668565,'19,January,2004','2004-01-19 00:00:00','   urlLink Workers Assail Night Lock-Ins by Wal-Mart    \"The reason for Mr. Rodriguez\'s delayed trip to the hospital was a little-known Wal-Mart policy: the lock-in. For more than 15 years, Wal-Mart Stores Inc., the world\'s largest retailer, has locked in overnight employees at some of its Wal-Mart and Sam\'s Club stores. It is a policy that many employees say has created disconcerting situations, such as when a worker in Indiana suffered a heart attack, when hurricanes hit in Florida and when workers\' wives have gone into labor.\" '),(12205,1668565,'24,February,2004','2004-02-24 00:00:00','   urlLink UK: Liberty Removal   \"The news that M15 is to increase its numbers by a thousand is merely the latest instance of David Blunkett\'s rampant authoritarianism. The secret state\'s claim that it is losing the never-ending, unprovable war against terror will play its part in the government winning a far greater prize. Across the range of his responsibilities - immigration, policing, the criminal justice system and prisons - Blunkett has either proposed or actually introduced measures whose repressive nature should shock us. That, by now, we may have become inured to them, does not take away from the fact that New Labour is trying to radically change the constitutional environment in which we live. \" '),(12206,1668565,'10,February,2004','2004-02-10 00:00:00','   urlLink A state of siege     \"The threat from suicide bombers was now so great the burden of proof in British courts may have to be lowered in terrorism cases so extremists can be tried before they succeed in mounting an attack, he said. Mr Blunkett said proposals allowing British terrorism suspects to be tried at least partly in secret by a security-vetted judge will be put forward in a discussion paper soon.\"   More :  urlLink The UK will take my baby away    \"Tony Blair hinted the standard of proof in organised crime cases may need to be lowered so that it is easier for police to secure convictions.\"  Ingsoc lite: the Blair project. '),(12207,1668565,'09,February,2004','2004-02-09 00:00:00','   urlLink Joe McCarthy is still alive    \"In what may be the first subpoena of its kind in decades, a federal judge has ordered a university to turn over records about a gathering of anti-war activists... ...In addition to records about who attended the forum, the subpoena orders the university to divulge all records relating to the local chapter of the National Lawyers Guild, a New York-based legal activist organization that sponsored the forum. The group, once targeted for alleged ties to communism in the 1950s, announced Friday it will ask a federal court to quash the subpoena on Monday.\" More from the Des Moines Register: ( urlLink 1 ,  urlLink 2 ).   urlLink The National Lawyers\' Guild responds . '),(12208,1668565,'22,March,2004','2004-03-22 00:00:00','   urlLink Food and Lodging   The UK is fastly turning into a Kafkaesque authoritarian jail: \"WHAT do you give someone whoâ€™s been proved innocent after spending the best part of their life behind bars, wrongfully convicted of a crime they didnâ€™t commit? An apology, maybe? Counselling? Champagne? Compensation? Well, if youâ€™re David Blunkett, the Labour Home Secretary, the choice is simple: you give them a big, fat bill for the cost of board and lodgings for the time they spent freeloading at Her Majestyâ€™s Pleasure in British prisons.\" '),(12209,1668565,'19,March,2004','2004-03-19 00:00:00','   urlLink Storming Denmark\'s free zones   \"Some 200 officers took part in this week\'s operation, and were greeted by barricades and booing residents.  Their 10-hour-long incursion into Christiania resulted in 48 arrests - almost all for the possession or sale of cannabis, a drug banned in Denmark.  Mr Vittrup says the suspects will face trial and possible prison sentences of up to six years...   After decades of defiance, it seems Christiania is finally being brought to heel.\" '),(12210,1668565,'30,April,2004','2004-04-30 00:00:00','   urlLink Winning hearts and minds    \"...The graphic images include one of a hooded and naked prisoner standing on a box with wires attached to his genitals. CBS said the prisoner was told that if he fell off the box, he would be electrocuted.  Another shows naked prisoners being forced to simulate sex acts. In another, a female soldier, with a cigarette in her mouth, simulates holding a gun and pointing at a naked Iraqi\'s genitals...\" '),(12211,1668565,'27,April,2004','2004-04-27 00:00:00','   urlLink Christian Mullah Indoctrination Kit    This is not a spoof site. Really!  U.S. Department of Health and Human Services grant study predicts 50%-60% reduction in teen sexual activity with Choosing the Best... '),(12212,1668565,'06,April,2004','2004-04-06 00:00:00','   urlLink If she plays with herself, is it child abuse?    \"State police have charged a 15-year-old Latrobe girl with child pornography for taking photos of herself and posting them on the Internet. Police said the girl, whose identity they withheld, photographed herself in various states of undress and performing a variety of sexual acts. She then sent the photos to people she met in chat rooms.  A police report did not say how police learned about the girl. They found dozens of pictures of her on her computer.  She has been charged with sexual abuse of children, possession of child pornography and dissemination of child pornography.\" '),(12213,1668565,'04,April,2004','2004-04-04 00:00:00','   urlLink Mengele capitalism    \"Orphans and babies as young as three months old have been used as guinea pigs in potentially dangerous medical experiments sponsored by pharmaceutical companies, an Observer investigation has revealed.  British drug giant GlaxoSmithKline is embroiled in the scandal. The firm sponsored experiments on the children from Incarnation Children\'s Centre, a New York care home that specialises in treating HIV sufferers and is run by Catholic charities. \" '),(12214,1668565,'03,April,2004','2004-04-03 00:00:00','   urlLink Cry Baby Jesus    \"US troops are to be armed with a stun gun that uses a baby\'s high-pitched scream to bring the enemy to its knees.  The gun, which will be issued to marines in Iraq this month, fires \'sonic bullets\' that can be targeted like a torch beam.  Anyone hit with a full blast would suffer excruciating pain, permanent deafness and some form of cellular damage. A prolonged blast could kill. \" '),(12215,1668565,'26,May,2004','2004-05-26 00:00:00','   urlLink The Abu Ghraib recruiting report. Prospects    \"Eleven teenagers have gone on trial in the German town of Hildesheim, accused of abusing and humiliating a classmate and filming his ordeal. \" '),(12216,1668565,'04,May,2004','2004-05-04 00:00:00','   urlLink I was a teenage criminal     \"In the government\'s efforts to protect children from abuse, however, the law also forbids under-16s from engaging in any sexual activity - ranging from \'touching\' to full intercourse.\"  ... but they claim they won\'t actually enforce the law, in such cases. But let\'s criminalize everything just in case... '),(12217,1668565,'01,June,2004','2004-06-01 00:00:00','   urlLink The merry march of progress   Russia\'s poorest face huge cuts in benefits: System supporting veterans, disabled people and pensioners to be swept away as Moscow reforms aim to cap compensation... '),(12218,1668565,'20,July,2004','2004-07-20 00:00:00','   urlLink Second attempt   \"Bush acknowledged that acting CIA Director John McLaughlin had said there was no direct connection between Iran and Sept. 11. But the president said he has long expressed his concerns about  Iraq  Iran.\" '),(12219,3400091,'26,May,2004','2004-05-26 00:00:00','    Alright, this is getting annoying... where is my period?  Pseudo made me lasagna last night and I drank an entire bottle of wine by myself. He emailed me this morning to tell me how great I was last night. Jesus, did I black out?   There is a grain of sand in my gum, it is the current bane of my existence.  In staff meeting this morning, new young boy who works here unintentionally made 3 consecutive farting noises with his boat shoes and I laughed my ass off.   '),(12220,3400091,'25,May,2004','2004-05-25 00:00:00','    Feeling physically ill from 2-digit long bank account balance.  Fun weekend- in sports bar sunday night, eating chicken wings and drinking beer, psuedo pulled out his dick at the booth. Can\'t remember why he did, but all of a sudden, there it was. Totally floored.   Think I set the record for most Sex and the City episodes watched in a row. Or hours spent in a bathrobe.   Saturday night had \"great fucking sex\" with psuedo (I\'ve found when he compliments me throughout the night, our romps are just that much better.), which was unduly interrupted by his dog licking my foot mid-climax.  Got three cute new Juicy dresses from Saks- orange, pink and turquoise- hell yes! (Then got last month\'s Saks bill in mail today and clapped hand over mouth in utter shock. Should be a great anorexic three weeks coming up until next payday.   Fantastic! Buy shoes, no money left for food, lose weight!   '),(12221,3400091,'21,May,2004','2004-05-21 00:00:00','    Phew... things back on track with pseudo:    P: So I pick you up after the game for drinks?  Me: Pick-up sounds nice. Look forward to it. Call first though, so I can usher out whatever hot studs I have over.  P: Miss Brunette - will you entertain the possibility of joining me for drinks atop some building in the ctown area?  I would like that very much. Is 11ish too late for a gorgeous brunette like yourself?  Me: not for such a respectable male suitor as yourself... I will surely be waiting.  P: What does you wardrobe look like now and for the night?  Me: Right now, yellow lace dress and a black cardigan, black heels. Looking adorable I might say. Tonite, maybe whipping out a new skirt, might bring out a new pair of Pradas yet to be seen by *Pseudo*, a tshirt, we\'ll see.  P: I like what I am hearing... You name the place tonight...ill get home probs quick shower then we will rock - close them down and then head on home to crisp ac and sheets?  The last part work?    Ah... we are definitely back to cruising altitude. Looking forward to that first glass of chardonnay and a view of downtown Charleston...   '),(12222,3400091,'20,May,2004','2004-05-20 00:00:00','    Holy shit... I just found a chipped tooth. This fucking hurts.   '),(12223,3400091,'20,May,2004','2004-05-20 00:00:00','    Day halfway over... thank god. Had terrible bantering emails with pseudo-boyfriend all morning. Seems to think my denying hooking up with old flame last night is ludicrous (\"two people of the opposite sex just do not hang out and watch movies drunk... there is no way there is no sexual attraction. If you just admit it, I\'ll leave it alone.\") Anyway, I staunchly denied it (we did make out. dammit.)  Drove home to eat lunch and decided on a little angry Fiona Apple for the ride- stuck at light on Meeting Street with windows down awash in sunshine from moonroof and realized how much I love living in this city...  Wow. Just got phone call from long distance boyfriend, who has no inkling of other hook ups here. Told me it was \"good to hear my voice.\" Cringed, but then cooed something to the effect of \"You too...\" and hung up quick.  Can\'t wait to use remote controlled vibrator again at the bar tonite with psuedo boyfriend. God, he loves that thing!   '),(12224,3400091,'20,May,2004','2004-05-20 00:00:00','    You know it was a rough night when you wake up to a half empty beer on your nightstand. Ugh.   I\'m sitting here at work, dazed and eating some beef jerky. Looks like rat shit in contrast with the crisp white papers and manila folders of my desk...  So I\'m totally avoiding the huge, looming error I made last night; namely, hardcore making out with old flame/best friend. Not hungover, but feeling strangely ill from guilt- this is weird, I haven\'t given two shits about cheating on a guy since I adulterously banged one of my best friend\'s older brother.  My thong is strangling my ass.  Got to start working. No email yet from psuedo-boyfriend. Have a feeling he might be pissed. Hope he didnt notice old flame\'s car at house until 4 am...  oh hang on: he emailed me just now:   You and the drunk have a nice bang last night?   Yeah, he\'s pissed. Fuck.   '),(12225,3400091,'20,May,2004','2004-05-20 00:00:00','    First posting! How exciting.  If this were a diary, I suppose I would say... had another wonderful night with psuedo-boyfriend/neighbor: walked dog on beach, did my best wind-swept, thoughtful expressions. We walked, threw the tennis ball, toes in the sand, dicussed the possible textures of the dog\'s shit after the consumption of large amounts of sea water, and we laughed...  Mexican dinner, chips and cheese and 2 pitchers of margaritas later, I gave \"the best blow job he\'s ever gotten.\" Still getting rave reviews over email today. Something about inflicting paralysis.  email received:  \"Well done...well done last night...ear to ear...\"   Skirt is see-through today. Has gone unnoticed by any attorneys or other such superiors.   Saw hottie attorney from 5th floor in elevator for 3rd day in a row. This time I remarked \"You must think I\'m stalking you,\" in a near purr, and we exchanged some banter until my stop. Hoping to see him again tomorrow; might try to live my elevator make-out dream.  Can I just say that the Washingtonienne is hilarious... makes me wish I\'d thought of it first. However, stay tuned. I\'m only just getting warmed up.   '),(12226,3400091,'29,June,2004','2004-06-29 00:00:00','    Wearing four-seasons-ago Trina Turk plaid mini-kilt today and waiting to hear about it from the bitter, horse-faced human resources woman who has a personal vendetta against me, my legs and my outfits (women around here are spiteful, eh?). Once I wore this miniscule Lilly skort and got called into her office. Horrible day.   As if I needed another reason not to wear Lilly Pulitzer.   My reason #1: the masses of bourgeouis girls who save their menial job paychecks to buy one $180 dress so as to fit in with those to whom that is pocket change; at an event like, say, a steeple chase. Sounds terribly snotty, but it is my pet peeve. Then those who can in fact afford these dresses have to rise above this- which I excel at, being the fashion maven that I am- and can only use their Lilly dresses as closet-corner stuffers (come on, we all have these, items of clothing that merely serve the purpose sitting in the back of the closet with massive price tags still on, to fill the corners so we don\'t have to actually reach back there to see whats there. Or maybe that\'s just me. I found a $350 Rozae Nichols suede shell encrusted halter top yesterday- going to have to wear that one)- but it is annoying anyway.   Wow, that was a venemous rant. Going to have to search my soul to find out where that spit and piss came from. Hope I didn\'t offend anyone.  On to bigger and better things- boyfriend and I had a talk last night- tears of joy involved on my part (slightly tequila drenched), really learning to flex my emoting muscles- ones I didn\'t really know I had. We discussed our future together and how mad about eachother we are and how we are going to move away together. Going to keep things said private (translate: way too syrupy and sappy to put into print) but let me say that if I play my cards right and win this in the end- I\'ll be married to this guy.  (!!!!!!) (already planning outfits for trip to meet parents!)   '),(12227,3400091,'28,June,2004','2004-06-28 00:00:00','    Note to self (and other cat owners): Do not jerk off boyfriend in bed with kitten in room, as rapid movement/swishing sounds will most definitely cause kitten to pounce.   So tired. Sort of sunburned. Going to go investigate new hot law clerk coworker just informed me of.  I can still look, can\'t I?  Email from boyfriend:  \"What do we do tonite- will the moat be closed yet?\" Ah, period times.  p.s. holding some filing for one lady attorney I work for who absolutely despises me (presumably for my adorable outfits and shiny hair. It\'s not my problem you have a FUPA and 3-D moles on your face. Christ.) and she has done some doodling on the side. At first looks like innocent frills and loops, but in one loop she has drawn an angry face- complete with furrowed eyebrows, beady eyes and a mouth with fangs. I am more than a little disturbed.   '),(12228,3400091,'27,June,2004','2004-06-27 00:00:00','    Ooouf, one more hilarious one. Came to me while looking through old photo albums and saw one of my 7 yr old self hanging with an older teenage girl poolside. My mom reminded me her name was Gina and it all came rolling back to me (no pun intended, once you read the following):  I was attending a corporate family cookout with my parents and brother and had resigned to playing with the big kids in the pool until long after dark (and long after pruning- not a fan.) Anyway, you know how it is when you\'re young- no time for getting out and going to the bathroom- especially when it\'s a shit lurking behind your quake hole...  Anyway I did the typical clench and stand still until the urge passed, presumably consequence free, until time to leave. The gracious Gina took the young, helpless (but definitely trying to act cool) shiteful me to the bathroom to change.   Wearing a one piece is pretty tricky, and taking it off poses a dire lack of privacy, but neither of the above holds a candle to when you have a small, marble-sized turd in the crotch of said garment awaiting its escape.  And escape it did. I peeled off my wet suit and out popped a little brown nugget, which rolled aside the toilet and under the sink.  \"OhmiGOD what IS THAT!?! Eeeew..... GRooooooosss!\" Gina wailed in horror, both hands flailing as she backed away.   Playing it cool, I stood there, naked with my suit in hand, and shrugged. Elongated pause as I looked down at the turd, thinking, \'you little fucker. I sure as hell don\'t remember pinching you off.\'   Quelle disaster.  Working with it, I rebounded with a dramatic: \"Huh? Oh that? I don\'t know... I mean, EEEEW, gross!!!\"  Telling this story always brings me to tears of laughter. I could tell a million more. But as of right now, I have terrible cramps (HELLO, PERIOD!!! I love you.) and am off to pop some confiscated pain pills and go to sleep.  Bonne nuit!   '),(12229,3400091,'26,June,2004','2004-06-26 00:00:00','    ha, I just realized that last entry went against the two heavy points I had made in the previous one. If you didn\'t notice the drinking and the peanut butter eating, you weren\'t paying attention to my \"taking a break from boozing\" and \"new Mary Kate diet\" affirmations of yore.  I am such a joke.   '),(12230,3400091,'25,June,2004','2004-06-25 00:00:00','    In an effort to impart some of my history on this flimsy storytelling (in manner of shedding light on current behavioral patterns) I share now a flashback I had after visiting with my best friend for a few hours yesterday in Savannah.  (in kitchen of house in Richmond, VA)  Teetering on 5 inch, orange satin Prada heels, your heroine slurs into a cell phone at nameless caller on other end of line.  After one particularly gut-grabbing laugh, a small trickle of pee down her inner thigh becomes impossible to hide.  (pointing at pee, still on phone, she looks down teetering around, then with hand over mouthpiece of cell phone loudly whispers) \"Oh... ha... I just peed on my Prada shoes. Who does that!??\"   One of many grand tales of taking leaks that I have to offer.  It is absolutely pouring outside and I\'m stranded indoors while my parents are at a wedding boozing it up. I\'m getting into the whole lounge act with my new Juicy orange terry dress (one of... what, 8- holy hell, I DO have 8!!!) and a jar of crunchy peanut butter.  It\'s 7 pm now- I\'m now off to finish watching Gigli. Ordered it from On Demand. Not sure why, just wanted to see what the hype was all about. (ok, reason #2, the much more legitimate one: I have a closet obsession with J Lo. See: lucite clear wedge heels I purchased after seeing her wearing them in numerous photos.) (ahh shit, I might as well offer reason #3 too.... I\'ve been drinking wine since about 5... by myself...)   '),(12231,3400091,'25,June,2004','2004-06-25 00:00:00','    Well I\'m home for the weekend- getting some TLC from the parents and a break from the boozing and spending (actually, boyfriend went out of town to a wedding in DC- coincidentally, the wedding of the guy who took my virginity, and one of his best friends - so I figured I had no real reason to stick around.)   7:00 am wakeup from Dad for a walk on the beach and noticed \"1 missed call\" on cell phone- boyfriend had taken time away from his long lost best friends on a romp through DC to call little me! So adorable. Definitely in love.  Today begins day of my new Mary Kate diet. Something has got to be done (and I really would like to achieve that chic model look in my Joie silk cargos rather than, well, looking like white silk cargo.) I am hoping that my stating this in a semi-public arena (ok, who am I kidding, only I and a handful of others read this,) I will be forced by karma to stick to it. We\'ll see.  Watched Love Actually last night.... loved it. Made me think of boyfriend and how I plan on spending the rest of my life with him (holy shit! - - did I just say that?!? Where doth my bachelor/slut/floozy days go??! It\'s ok... I didn\'t want my \'number\' to get into the 40\'s anyhow.) Off to have some coffee, turn down some wonderful mom-cooked breakfast for GNC shake (whats up, Mary Kate!) and head off to the beach.  Have I mentioned life is good?   '),(12232,3400091,'25,June,2004','2004-06-25 00:00:00','    Observation:   Going along with my previously mentioned pregnancy paranoia was the assumption that my tender boobs were certain symptom. I had emailed my friend as I do once a month (\"Unf, but I just KNOW I am... I mean, how could I not be? Ugh... doesnt it mean you are if your tits hurt? They\'re KILLING me...\") and I had optimistically chocked it up to pre-period pangs... but alas....  I have (like the detective I am) deduced the real cause of said pain. D-cup girls, I implore you- try to remember to wear a bra when getting it from behind. Granted, I am a tiger in the sack (shameless self-promotion,) but they can really get to bouncing around, and I pryed myself away from the rhythm concentration long enough to look down- jesus mary and joseph I had to stifle a laugh- but that is indeed what can be held accountable for the week-long soreness (and this morning\'s soreness.)  Anyway, just had to puff up my feathers, there. I love making these kinds of discoveries.   '),(12233,3400091,'18,June,2004','2004-06-18 00:00:00','    Fuuuuuuuuuuuuck!  On way to breakroom for some caffeine libation I passed a mirror and instictively glanced in it only to find... purple lips.  Not just a few dots, like totally blotched, puffy purple lips- they got the shit bit out of them!  It looks like I have chewed a pen to its fruition and it came all over my lips. I now vaguely remember some furious biting during one particularly out of hand portion of last night.  Have lunch date with best friend and her parents in Savannah today and no shade of lipstick to disguise this catastrophe... This must be chic on some European runway, right? Fuck!  on a lighter note- hand came perilously close to box of krispy kremes in breakroom but I used the force to direct it instead to a coffee cup- phew. That was a close one.   '),(12234,3400091,'18,June,2004','2004-06-18 00:00:00','    So I have been hugely preoccupied this whole week with echoes of last month\'s \"thinking up baby names\" quandary, but as I had to go to my yearly gyno checkup yesterday, it was doubly stressful. I was mentally prepared to have some new scary doctor come in with my cup \'o pee and say, \"And I have some good news- you\'re pregnant!\" Damn that early withdrawal method.  Getting IUD next month- cannot wait. Baby prevention that lasts up to 10 years with ne\'er a thought. Boyfriend calls it \"I get to come in you?\" device. We poured over the information sheet at dinner laughing giddily about our soon-to-be good fortune. Cannot wait. Going to be best $500 I\'ve ever spent.  So everything went well (no bebes- but had thought up Catesby, Charlie and Grady) and I went (baby-free) to the liquor store and got some tequila, went home and laid in my hammock in my bathrobe with a big margarita and a smile.  In keeping with the celebratory chain of events of the afternoon, when boyfriend got home I went down to pay a visit (in new Vince off shoulder shirt- thanks, mom + saks), homemade cheese quesadilla in one hand and a pitcher of leftover \'Ritas in the other, just in time for the NBA draft at 7:30.   So while he watched the draft, munched on quesadilla and margaritas and kicked back, I happily gave him a masterful blowjob, then resigned to a nap. Am I not the most perfect girlfriend? Jesus.  Fastforward to amazing sex for 2 hours- which resulted in \"island bed-\" where we somehow jolt the bed out into the middle of the room. I slept wonderfully, to say the least.  Planning trip to his house in Taos, New Mexico to fuck in the hottub outside while it\'s snowing. I think I might be in love.   '),(12235,3400091,'18,June,2004','2004-06-18 00:00:00','    Last night HBO featured a hugely entertaining Real Sex (what number, I do not remember; #80? #132? Must be up there pretty high, since I\'ve been watching these shows on mute after my parents went to sleep since I was in 7th grade.) It was centered on an anal sex seminar called Butt Sex 101.   My roommate was watching it upstairs and I was watching it downstairs, our laughs filling the apartment simultaneously as various flabby women gave it to their husbands up the ass with strap-ons. Pretty vulgar little program. Never seen so many aholes in my life. There was even a plastic spiderhole to pass around and play with. I guess it was karma for my \'quake hole\' discussions of yesterday.  This morning pseudo didn\'t have to wake up until 11 so I went down to his apartment and crawled in bed with him for some pre-work snuggles.  I was overcome by the mushy-ness of the situation as I lay with his head on my chest, playing with his hair planting little kisses on his head while I softly recounted the HBO starhole special. I listened to him drift in and out of snoring and smiled thinking how this guy is probably my best friend here. However, in keeping with our inability to keep our hands off eachother, with 6 minutes to go he had my pants off, and when I finally left for work, my satin skirt was a wrinkled mess and my knees a little weak.  My computer is finally up and running again, after it crashed this morning, presumably after I looked at a sexual picture yesterday- so in an effort to clean up my tracks before the IT person came to fix it, I did a scanning program and something came up called: SexList.  Needless to say this induced much laughter, now I\'m sure there\'s an office rumor that I look at porn. Which I so do not do (grew out of that phase in college.)  Have mom\'s credit card number; will spend.   '),(12236,3400091,'17,June,2004','2004-06-17 00:00:00','    This day blows.  I have a terrible hangover. And I\'m getting nothing done.  holy shit, ha- roommate/deskmate just now discovered that I changed the radio station to some slow, smooth jams station- being the card-carrying racist he is, he was appalled; me, I will sing along to \"Everybody Plays the Fool\" and daydream about what that Neville brother\'s giant birthmark tastes like.  Anyway, I just ate a box of those veggie chips that a coworker of mine claims she doesnt like. Pretty sick. Girls are so weird. I have a feeling she only said she didnt want them and gave them to me so I\'d eat them and bear the brunt of their caloric quantity- not a big fan of girls.  \"My Girl\" was playing on the radio before roommate and I went to go get cokes from the breakroom and as I shoveled ice in my cup I made up a little ditty to the \"My Girl\" tune:  me: \"Quake hole (quake hole, quake hole!) Talkin\' bout... quake ho-ho-ole... quake hole!  I dont need no quake hole... I got tha month of may...\" roommate: Actually you do need a quake hole, or you\'d probably die.  me: Holy... hell. We\'d all die if we couldn\'t shit.  roommate: My grandmother was once in the hospital cause she couldnt crap. She was literally, \'full of shit.\' me: High five for granny when she finally pinched off that heat stick, eh?  So last night at Gilroy\'s, 5 vodka cranberry cocktails deep, I was pretty sober but drunk enough to make a terrible scene about pseudo staring at this tall, statuesque blond - with one eye- using the other one in conversation with me. Oh, and he was trying to get me to eat the pizza (remember: \'we need to feed these.\') So, in sticking with my ever-present fatal flaw of wicked insecurity in the face of something that actually is a mutual source of love and adoration for me (translate: he loves me how I am, I adore him- so I conjur up said blonde in order to pick a fight, get more reassurance to my own outward appearance, whine and in the end, attempt to destroy something I really want, just for the sake of destroying it- hmm, FEAR.?), I bitched and cried and woke up today to realize the absurdity of my actions.   Can I just say, on a serious note- in a perfect world, no one would be sickeningly beautiful, hugely wealthy, name dropping designers or breathing on the spot on the hood of their car then wiping it away with an elbow. Actually, the only thing that would matter to anyone would be finding that one person you\'re meant to be with, at least the one you can stand, laugh with and who keeps you on your toes- no matter what the inside label of your shoe says. THAT would be the source of happiness.  That being said, when I was lying in pseudo\'s bed this morning, he leaned over to look at me, smiled, put one finger up to his mouth to say \"sssh,\" paused, then farted. If I didnt already know it, I knew then, I\'ve got the one for me.   '),(12237,3400091,'16,June,2004','2004-06-16 00:00:00','    It is absolutely pouring outside. Too bad my pee-on status in this office only offers me a window view when I\'m going to a. kiss someone\'s ass or b. going to pee.  Huge daydreaming for me today. Rain making me want to be scrumping pseudo on a boat out in the ocean somewhere, or rolling around in a sleeping bag in a tent (or any such outdoor venue where rain can be enjoyed.) Email from him really turning me on:  \"Just started to rain very heavily here - really coming down.  I wouldn\'t mind being knee deep in the Madison right now fishing for big browns and rainbows watching bald eagles fish beside me.  Man that was the life.\"  Lord, could he be any hotter?  Rain + hot, rugged fly fishing pseudo-boyfriend causing me to browse online for various, unneccessary outdoor items: North Face fleece, raincoat, hiking boots, Tarponwear fishing shirt, new LL Bean backpack, reminscent of one I had in 8th grade. Saw girl last night at gym with an awesome Orvis tote like my dad\'s... might have to spring for one of those.   Racking brain for last four digits of mom\'s credit card number...   '),(12238,3400091,'15,June,2004','2004-06-15 00:00:00','    So after I go running on the Battery last night, soaking in the huge houses and the Mercedes, Porsches and BMWs, I returned home sweaty and hungry (hungry being the operative word: I am a such an amateur at this anorexia shit) to a phone call from pseudo saying I had 5 minutes to get my ass in gear for a Norm\'s Pizza mission- beers and pool.  In a ploy to appear more adventurous, low-maintenance and a general \"good sport,\" (I am a good sport: have endured hand of ball smell on face, many a dutch oven, post-tobacco-dipping kisses and thorough discussions on how many ass-wipes different varieties of shits require-- ok, I laughed through that one, but...) I went to said pizza joint in running gear, sans shower. Quelle stinkpot!  At pizza place, 2 pitchers of Papst Blue Ribbon deep, a little maharaja of a man showed up and began circling our pool table pointing out shots and yelling at our friend, The Terrier, to quit ignoring his \"ad-wice.\" He was small, he was brown, and he was shouting and spinning about like a hairy, glassy-eyed dradle... Fiesty as he was, I believe he just wanted some friends, but anyhoo we played \"Name That Drug,\", laughed into our beers then finally got him gone with a few faux-threatening swings of the pool cue.  My point is, I had another brush with potent racism last night as the only thing I could think of while watching said bombed brownie was, \"He looks just like the Indian 7-11 worker I knew back in middle school,\" that my best friend and I called Raj-Raj. A man of perpetual cherry red-lips (could be the access to Slurpee, could be some new bindi lip stain very hot in India but not quite into the swing of things in Richmond, VA), Raj Raj might not have even been his name, but it was fitting, and he sold me Lik-M-Aid, Charleston Chews, daily horoscope rolls and The National Enquirer; and I liked him.   Or Habeeb (I think we all had a Habeeb in our underage drinking phases) who sold us Boone\'s Farm at the Penny Saver on the dumpy side of town. A proud owner of a mint-green Toyota Corolla (once again, should not know this) Habeeb facilitated many a wasted night for me. Thank you, Habeeb.  And for that matter, thank you Indians. Without you, there\'d be no Apu, no Yoshi (10th grade boyfriend) to instill in me the fear of an uncircumsized penis, no college algebra teacher that caused me to fail because I just could never get over how he said \"And den, you diwide, de nummbah...\"   '),(12239,3400091,'15,June,2004','2004-06-15 00:00:00','    Today has been chock full of references, jokes and racial slurs towards the Asian population. Brings a tear to my eye.  1. I was trying to shove a huge stack of paper into a hole punch and it was rough going: \"This is like trying to shove Shaq in a tiny asian woman.\" (isn\'t he banging a tiny asian woman? what a huge dong.)  2. Random Asian girl (ok, I at least have the resources to sound more knowledeable and less ignorant/racist on this one: she is actually Korean. But see? Do we care?), who had my position before me, emailed back to say hi to some ex-coworkers.... and attached a few photos. Apparently she moved back to the motherland- photo of her plus 4 Koreans with blond streaks and purple eyeshadow. Said roommate, who sat by her for 3 weeks:  \"Man, now which one is she? I can\'t tell. They all look the same. All Asians look the same. Damn slant-eyes.\" (I love people from South Carolina.)  3. My inability to stop going \"Hiiiiyyy- ohhhhhhh!\", a la Ed McMahon, Chinese chop-on-table chef, an Asian version of one of the Seven Dwarves, etc., etc.....  4. Another co-worker said there\'s nothing he\'d rather see less than two Asian people having sex (to which I responded, or an Asian girl having explosive diarreah on herself, in a bathtub, wearing knee socks....) See link below and brace yourself.  www.tubgirl.com   Have a great day, and.... HIIIIII-OHHHHHH!!!!  Getting no work done. Random outbursts of laughter (followed by either Asian sound, imitation or reference.)   '),(12240,3400091,'14,June,2004','2004-06-14 00:00:00','    Some biatche in a stupid Mitsubishi Galant (embarassed I even know the variety of middle-class, depressing automobile) who just moved into my building has bought the parking space I\'ve been stealing for the past 6 months. I remarked to my roommate/deskmate:  me: \"Fuck that slut in the stupid Mitsubishi.\"  roommate: \"She\'ll probably turn out to be your best friend or something.\"  (after he receives one very hairy, mean look from me...)  roommate: \"Come to think of it, she drives a Mitsubishi, so probably not.\"    Yay for snobbery! Love the little bubble of cleanliness, $80 face cream and premium-octane gasoline I live in.   '),(12241,3400091,'14,June,2004','2004-06-14 00:00:00','    Whenever it rains here in the morning and I\'m walking down the street to my office buildling, I\'m whisked back to my days in France when I was 16. It always rained in that place, and there were always huge buses blowing crap on you from the street and everything smelled like breakfast. What a miserable corner of the world.   There\'s nothing worse than being homesick, fat (and getting fatter) and under the watchful eye of a drugged up, crazy ex-nurse who tells you in broken English- \"Eef you want le lunch, yoouh pahck eet youh-sulf.\"  Little things kept me going on this trip (and kept me eating,) like the multitude of random dog shit deposits along the city streets (friend and I took many photos of these), one giant infected goat anus at a French provincial vineyard (photographed this, too) and a spraypainted grafitti sentence on the wall of the busstop: \"Fuck the Frogs.\"  I guess the France memories can be attributed to more than stank buses and the cafe\'s fragrant bacon wafting across Meeting Street- I have put on a few lbs, here, people. I blame it completely on new love interest who, not without a hefty desire to eat, himself, constantly showers me with flattery about being \"too skinny\" and needing to \"feed these\" (grabs a handful of one of my tits.) Under such conditions, pizza, fajitas, cookies, chips and lots and lots of beer seem quite appropos.  I had some strawberries this morning and am pounding black coffee until I\'m somehow deflated back to normal size.   Morning sex session today solid- rain outside, icy air-conditioning, eyes still crusted shut- really not feeling being at work today.  Have more to share later. Must maintain some semblance of responsibility/productivity or my new powder blue Marc Jacobs suit I wore today will just mock my unprofessionalism...   '),(12242,3400091,'14,June,2004','2004-06-14 00:00:00','    Have not taken lunch break today in feeble stab at being anorexic for the next 2 weeks. Really giving me terrible heartburn.  supposed to go to departing coworker\'s happy hour party at fancy shmancy hotel bar across street after work at 5 but dont see myself making it, as I cant have drinks before I go to the gym and loosely made a pact with myself this morning to not drink beer/ more than a glass or two of wine for the next 2 weeks. I believe it can be done!  Oh- I bought some of that K-Y Warming Liquid the other day- just to see- \"creates a gentle warming sensation\"- turns out you dont need this stuff if you\'re as unbelieveably horny as I am- only makes for a greater mess. Might try it out at work just for fun.  oh- news flash! Psuedo is no longer \"psuedo...\"- I think I have a boyfriend! (DAMN hot gym lover for waiting 4 months to call me again until NOW! Might just keep that number stored in the old cell.)   '),(12243,3400091,'08,July,2004','2004-07-08 00:00:00','    Last weekend was great. Parents came to town and met boyfriend (loved him), successfully avoided ex-boyfriend who was in town (sans knowledge of current boyfriend, so called me 37 times moaning to my answering machine. fag.), met brother\'s new girlfriend, broke ankle in new JLo lucite heels (well, not sure yet, going to get xray during lunch today) and bought a plethora of goodies at Williams-Sonoma.  Confused about life as of late, however. Things with boyfriend escalating beyond any level I\'ve ever experienced before, and without a set precedent to dictate what I should do, I\'m sort of scared and thrilled and nauseous at the same time.  So life starts changing without your even knowing it and then all of a sudden everything is different, or on the parapet of being different, and before you know it you\'re getting a new job, a new apartment, moving somewhere where you have no friends, no life, following the person you think you might be in love with and all the things you held on to and leaned on for so long are let go of like a handful of sand on a windy beach.  Boyfriend and I went out to dinner at Basil last night and sitting across from him in the loud, dark restaurant, he dribbled out excitement-laced monologues about all the things we could do, places we could go, together, in our new town... and he reached his hand across the table and covered mine with his and said, \"You always get quiet when I talk about this. You don\'t want to move. I can tell.\" Which is totally untrue, all I want is to be around him, but I\'m just so scared. Scared cause it could be so fantastic.  Just as scared as I was the one time I was sitting at my desk in college listening to Dashboard Confessional realizing that Major Jones, the most evil heartbreaker, really didn\'t care about me. It\'s those harsh realities, good and bad. They shake you to the core.  I have one hand painted with my new Essie Boathouse polish and the other is bare because people keep on walking past my desk making a nailpainting mission impossible. Hoping my ankle is not fractured; awaiting email from boyfriend; gagging on black coffee; wondering why shoe repair place has not called me back about my new satin Prada heels; pondering the direction of my life....   '),(12244,3400091,'07,July,2004','2004-07-07 00:00:00','    In furthering my quest to be Jessica Simpson (purely with an eye for fashion and accessories) I have finally gotten a Cecilia du Bucourt shawl!!! So excited. Can only find them on ebay. Yips for me.   http://cgi.ebay.com/ws/eBayISAPI.dll?ViewItem&item=3490214021&ssPageName=STRK:MEWN:IT   '),(12245,3400091,'02,July,2004','2004-07-02 00:00:00','    Uh oh.  So I ventured a guess that perhaps more than a few of my buddy secretaries around here had stopped liking me- I could just tell they weren\'t as charmed by me as they used to be when I was a hard-working, ass-kissing little worker bee. Perhaps due to my leaving early and pretty much not even showing up lately, I was right- just sucks to hear it rather than wonder about it.  Anyway, really need a new job bad. Don\'t like being disliked.  Late night of wine and scrabble and I feel like butthole and just want to go to bed. Big weekend- parents coming in town and taking boyfriend out to dinner. First-time meet for dad. Big time.  I forsee myself throwing up at 2:30 today. Creeper hangovers the worst.   '),(12246,3400091,'01,July,2004','2004-07-01 00:00:00','    So I didn\'t come to work yesterday. If I have to hear one more time \"so and so was looking for you\" or \"hungover yesterday?\" or \"you could have called...\" I will unleash my wrath upon these zombies in the form of 7 plagues: 1. permanently extended middle finger 2. evil glares 3. gaseous releases 4. unexplained outbursts of tourette-ish swear words and dirty insults (hairy vagina bastard fat fucker ahole bitch!) 5. \"I will be paying someone like you to be doing this for me in a few years, and I will be paying them better.\" 6. \"Who told you that was ok to wear today? Fucking idiot.\" 7. post-it on forehead that says \"Fuck off.\" (oouf, that was some mean shit.)  Anyway, went to minor-league baseball All-Star game post-party after having 2 gin martinis in apartment before hand. Could not resist having said martinis while in bathrobe with hair up in curlers. Against specific instructions laid out by boyfriend, I wore an extremely bust-exposing little Nanette Lepore number (he said the horny players would be looking if I wore something sexy, so to cover it up. whoops.) and indeed got plenty of attention, which I will not deny facilitating (come on, some of them were really cute.) The final straw for boyfriend was when one who I had been mouthing to all night came up in and said, \"You\'ve got a tight looking woman right there.\" (ego just inflated 13%.)  So I woke up yesterday morning at 1 and decided work wasn\'t going to be an option; met boyfriend and the Terrier for lunch at the Kickin Chicken, watched some Cubs then went to run the dogs at the park. Fun times. Quelle beautiful rainy day.  Afternoon nap preceded by \"quick feel\" as we call it. Quick feel usually involves one of us saying \"Man, I\'d really like to feel a little penis in vagina right now,\" followed by explosive giggles, then one or the other climbing on top and the 2 minute \"quick feel\" turns into a marathon sex mission, segueing into a really solid nap.   I am in relationship bliss. Finally, someone I\'m enormously attracted to, who is my best friend, laughs at farts and they like afternoon pre-nap sex. I mean, seriously. Let\'s seal this deal.   '),(12247,3587934,'30,June,2004','2004-06-30 00:00:00','   so, as you are all aware, we went to play trivia last night at the Raven.  Well, there WAS NO TRIVIA!!!!  So pissed off.  The waitress explained to us that since the Yankees were playing the Red Sox, the trivia guy wouldn\'t be able to use the microphone as that would disrupt the commentary of the game on the TV.  Two things are wrong with this statement:   1) They knew the Red Sox were playing the Yankees LAST week when we were there, yet there was no mention then of there being no trivia.  In fact, the guy made a huge point about inviting us all back next week for more trivia.  It was a ploy to get us to go back.    2) They weren\'t even playing the TV commentary!!!  They played loud music the WHOLE time, so what the hell are they talking about?    Makes absolutely no sense and I am pissed.  I mean, my Dad drove down from New Hampshire for trivia.  So he was not happy either.  Fuckers, fuckers.    Well, we did have a good time anyway.  So there\'s that.  Also - Joel, I believe I saw your friend Rich(?) while I was there.  You know, the guy from Bentley that I never liked.  He is still a COMPLETE asshole, evidenced by the fact that he kept staring and staring at me, and I finally said hi, and he didn\'t respond at all.  what a jerk, I always hated him.   '),(12248,3587934,'29,June,2004','2004-06-29 00:00:00','   Tonight we are all heading back over to the Raven for more trivia!!!!  We have a title to defend here, people.  This is some serious shit.  I hope that I will be able to answer one question tonight, as I did last week.  That\'s enough for me.    Joel has been terribly upset about the course of events this concert season is taking.  First lollapalooza is cancelled, then the Pixies are dicking him around.  Poor guy.  But honestly, what do I care, I\'m going to see the Dead on July 31st.  That\'s enough for me.    Women who wear flip flops with a suit should be shot.  I volunteer for the job.  Is anyone an expert on dog shit?  Audrey seems to be having a slew of problems, with no end in sight.  It\'s either too hard or too soft, going too often or not going at all.  I think Joelle is secretly poisoning her, maybe with arsenic, Flowers in the Attic style.  Incest is next perhaps.   '),(12249,3587934,'27,June,2004','2004-06-27 00:00:00','   urlLink    See how happy the Polak is?&nbsp; urlLink   '),(12250,3587934,'25,June,2004','2004-06-25 00:00:00','   urlLink    Nice face, Megan!&nbsp; urlLink   '),(12251,3587934,'25,June,2004','2004-06-25 00:00:00','   urlLink    Piotr\'s surprise party sure was fun!&nbsp; urlLink   '),(12252,3587934,'25,June,2004','2004-06-25 00:00:00','   Once upon a time there was a man who was walking down a cold, dark lonely road. In the middle of that road lay a very large snake, which suddenly turned his head and began to speak!            \"Help me, Mister!\" the snake said, \"Please help me!  I\'m a cold-blooded reptile and without warmth and nourishment, I\'ll die on this cold, lonely road.  Please pick me up and put me under your coat so that I might warm myself.\"            \"But you\'re a poisonous snake,\" the man said.  \"And if I pick you up, you\'ll surely bite me.\"            \"I wouldn\'t bite you if you saved my life,\" the snake told him.  \"I\'d be very grateful.\"            The man thought for a moment, then reached down, picked up the snake and put him inside his coat.  And when the snake had warmed himself and vitality returned, he suddenly struck out and bit the man.            Ripping open his coat, the man hurled the snake back down on the cold pavement.  \"Why did you do that?\" he demanded.  \"You said if I saved your life, you wouldn\'t bite me!\"            The snake, with a sneer on his face, looked up and the man and said, \"Mister, I\'m a snake.  You knew what I was when you picked me up!\"   '),(12253,3587934,'24,June,2004','2004-06-24 00:00:00','   First off - pet peeve # 2,847,373:      People who get on an elevator, see that the button to the floor they want to go to is already pressed, and then go ahead and press it again anyway, even if it means they have to push people out of the way to do it.  To those people I send out a big fuck you.    Audrey had all sorts of runny shit coming out of her ass this morning.  Nice, huh?  I was late for work because I had to boil her up some chicken and rice and practically spoon feed her because she doesn\'t know that people food is actually food that she can eat.  She thought it was a toy or something.  Kept chewing it up and spitting it out.  And of course, Joel cannot clean up the messes because he pukes on top of it, creating even more of a mess.  Men are such pansies.  I mean, it\'s only shit.    Last night I was at Circuit City picking up a little something for a certain Polak I know.  Well, did you know that if you sign up for a Circuit City credit card you can get a free two liter of soda!!!! Imagine that!  Put yourself and your credit record at the mercy of those deranged facists and get a $1 bottle of soda absolutely free!  I know a deal when I see one people.  You better get your asses down there quick.   '),(12254,3587934,'23,June,2004','2004-06-23 00:00:00','   Dear Joel,  My goodness.  Song lyrics for your first post?  Aren\'t you the one who is ALWAYS making fun of my penchant for song lyrics?  Yet again, I am faced with the undeniable fact that you are a huge, bowl-legged hypocrite.  You are a smoker who complains about second hand smoke.  \'Nuff said.  But alas, Modest Mouse does rock my anemic little world.  So I guess I shouldn\'t be so hard on you.  Nah, I will be that hard on you.  I just can\'t help it.   '),(12255,3587934,'22,June,2004','2004-06-22 00:00:00','   For father\'s day, I took my dad to trivia night at the Raven in Waltham along with the usual suspects.  It was fucking ruling.  We kicked so much ass, we set a record for highest score ever.  I think we only got 4 questions out of 30 wrong.  Well, by \"we\" I mean everyone else sitting at the table.  For those of you who don\'t know, I SUCK at trivia.  really horribly.  But I do enjoy playing and writing down the answers in my type-writer like handwriting.  The only answer I contributed was that Mickey Mouse turned 75 years old last year.  So sad, so sad.  We got a free round of watermelon martinis plus 30 bucks in cash for winning.  Rock.  Didn\'t get home til almost 3:00 though, but I am surprisingly not too tired.    Back to work for a while.  Aimee, you should post on your blog more often.  Your entries are way funnier than mine.   '),(12256,3587934,'22,June,2004','2004-06-22 00:00:00','   Joel seems to have a problem with putting things in perspective, as evidenced by this recent e-mail:  \"fucking lollapalooza has been cancelled.  I\'m so pissed.  All dates have been cancelled, it\'s the entire tour.  This was the only thing I was looking forward to.  This is going to set off a downward spiral in my life that I will never recover from.\"   '),(12257,3587934,'21,June,2004','2004-06-21 00:00:00','   Just as I was pushing \'publish post\' on my last entry, I looked up from my desk to see a figure dancing in the window.  Wouldn\'t you know, it was Joel.  Apparently he woke up, got on his bike, and rode all the way down here just for kicks, and decided to surprise me by dancing in front of my office window.  I told you today was weird.   '),(12258,3587934,'18,June,2004','2004-06-18 00:00:00','   The entire world has stayed home from work today.  Well, enough people so that it seems weird to me.  Let\'s start with Joel.  Since he\'s the craziest and most bowl-legged man-child out there, it\'s a good place to start.    Yesterday, Joel and I went over to his parents house for a little Father\'s Day feel-good fun.  And it was.  After dessert, we all had one cup of coffee.  It was around 6:00 pm.  I thought nothing of it.  But it just so happens that Joel is a total freak, and that one little tiny cup of coffee kept him up ALL NIGHT.  He\'s like an old man.  He was still awake when I left for work this morning.  And I\'m not talking about \"stayed up too late watching TV and couldn\'t stop\" type of awake.  He was laying in bed with his eyes closed all friggin\' night and didn\'t sleep for even one moment.  The boredom alone would kill me.  Why didn\'t he just get up?  Fucking insomnia amateur.  So, he decided to call in sick to work today, go figure.  Well, I do have to say that he NEVER calls in sick, so I guess he deserves to have a day off.    Patricia is also home today.  As she put it, Spencer must sprinkle a delicate mixture of pollen, dust and cat dander in her eyes while she\'s sleeping.  Yes, her allergies are that bad.  And yes, he is evil enough to do something like that.  This is the same guy who got drunk one night and called me fat about 100 times while physically assaulting me.  I was just trying to driving his drunk ass home, see the thanks I get.    Michael Sherman is also out today, which came as a shock.  I love Michael Sherman.  Work just isn\'t the same without him.  I\'m also covered in bug bites, because Joel and his father made me go on a \"walkabout\" (yes they actually called it that, those nerds.  It\'s definitely a genetic thing in that family) with them through the woods in Townsend yesterday.   '),(12259,3587934,'18,June,2004','2004-06-18 00:00:00','   urlLink    See, she\'s so fucking creepy......&nbsp; urlLink   '),(12260,3587934,'17,June,2004','2004-06-17 00:00:00','   Ok, before the women of the world get out of control with their bitching about frizzy hair in the summertime, let\'s just take a moment to remember something - at least they  have  hair.   I would like nothing more than to try to tame a massive bush of free flowing, out of control clumps of hair.  Every morning, as I patiently preen the 5 strands that are left on my head, I think, fuck those bitches who complain about frizzy hair, or about having \'too much\' hair.  If you don\'t like it, shave the shit off and give it to me!    As embarrassing as it is for a man to start balding at the uncommonly early age of 24, how bad do you think it is when the same thing happens to a woman?  For christ\'s sake, I look like fucking gollum.  Pretty soon there will be nothing left.  Anyway, just wanted to give a view from the other end of the spectrum.  The \"I\'m a balding woman at age 24\" spectrum, which definitely does not have enough of a voice in the media, nor enough lobbyists in Washington.    On another note, I TOO think Paris Hilton is unbelievably creepy.  She is one of the reasons I am an atheist.  I think a fairly good horror movie could be made based on my reaction to her existence.  If I ever saw her in the street, I would probably just lose my mind right on the spot.  She is that disturbing.    I just got bitched at by some fucking woman at the Cambridge District Court.  Apparently she is so much smarter than me, even though she has been filing papers in a basement from the beginning of time, and will continue doing so until the end of her miserable little life.  Who\'s the smart one now?   '),(12261,3587934,'17,June,2004','2004-06-17 00:00:00','   Ok, here is a long overdue rant on something that makes me so absolutely disgusted - open toed, strappy (and usually backless) sandals that people wear to WORK.  Tis the season for this type of bullshit, and it just makes me ill.    For example, Patricia and I saw a woman walk past the school yesterday afternoon after we got out of work.  She was an older woman, probably early 40s or something.  She was tall and a bit big, but nothing nasty or anything.  She was wearing a nice linen pant with a similar blouse, which I didn\'t have a problem with.  Then she totally RUINED it all by wearing - the strappy backless high heeled sandals!!!!!    Why, god, why?  I hate them, especially in a situation like that.  Her giant, flat feet were stuffed into the straps, and they were a bit too small so that the back of her heel was kind of hanging off the end.  She looked like the leaning tower of Pisa, like she was about to collapse at any moment.    What makes anyone think that this kind of foot wear is even remotely acceptable?  All that toe cleavage is entirely inappropriate for the workplace, and it\'s just plain nasty.  They make that awful clacking noise when they slap the bottoms of people\'s feet.  And they\'re so minimal that the person might as well be barefoot; it\'s a broken ankle just waiting to happen.    I have come to the realization that there is a correlation between the people who wear these types of horrible shoes and the people who actually go out and get the french manicure looking shit on their FEET.  Utterly wrong.    Next time I see one of these freaks I am going to pull off their sandal and proceed to bash them in the temple with it Single White Female style.  Ok, maybe I won\'t, because that would involve getting near their funky feet in the first place, but I will imagine it in my head.  These people will someday get their due.  Just watch.   '),(12262,3587934,'17,June,2004','2004-06-17 00:00:00','   shut the fuck up.  i can post on your blog, too?   this is nothing short of awesome.  go team!   '),(12263,3587934,'17,June,2004','2004-06-17 00:00:00','   urlLink    Here I am downing my 10th Cape Cod at Aimee\'s apartment this January.  &nbsp; urlLink   '),(12264,3587934,'17,June,2004','2004-06-17 00:00:00','   Is this thing on?  I\'m not really that good with computers.  But Wendy told me to start a blog, and so I have, because I always do whatever Wendy tells me to do, especially in the S&M sense.    Some may say \"Sara, you don\'t want to start a blog.\"  To these people I say \"But I do.  And I\'m sorry, but it\'s not going to stop.\"    All right people, prepare to be disgusted, horrified, insulted, degraded, and many other things you thought you would only feel back when you were an alter boy.   '),(12265,3587934,'17,June,2004','2004-06-17 00:00:00','   Wendy is my best friend in hell.  She is my muse, my raison d\'etre, my sugarplum fairy, my black Bridget Jones.  And, thanks to recent developments here in Massachusetts, she can someday be my wife (it would be worth it just for the look on Donna\'s face).    So this post is dedicated to Wendy, without whom I would never have had a 60 year old stripper slapping me in the face with her sagging, wrinkly tits while I was on the can.  Such are the wonderful things that can happen to someone lucky enough to be friends with Wendy.    Cheers to the girl who introduced me to Popeye\'s, Jeff Buckley, Tony\'s, the Claremont, Community, and Rue.    I love wendy like a fat kid loves cake.  I would tell all of you to go out and discover the wonder that is wendy, but then you would all fall in love with her too.  And that can\'t happen, since I am not willing to share.    Back off sluts, she is mine.   '),(12266,3587934,'17,June,2004','2004-06-17 00:00:00','   urlLink    Some things should never be photographed....&nbsp; urlLink   '),(12267,3587934,'17,June,2004','2004-06-17 00:00:00','   I told Joel about Aimee\'s blog, and that Wendy has told me to make my own blog, etc...  He writes back and says that we should create one together (a joint blog!).  Can you believe this kid?  As if.    First, Joel and I are nothing alike.  He will hate everything that I post here.  In fact, he\'s probably just seeing that picture that I uploaded a few minutes ago and spewing his brains out at work.  He has quite the weak stomach, and I\'m sure that picture is enough to send him over the edge.    Second, who are we?  Sonny and Cher?  We don\'t do things together.  Separate is as separate does.  Joint blog.  Makes me want to retch.  That\'s almost as bad as him expecting me to take his last name.  On the list of shit Sara will never do in a million years, it\'s pretty high up there.  Although, it\'s quite a long list, so maybe that\'s not saying much.      Third, I will hate everything that Joel wants to post.  He will post severe weather warnings, french vocabulary, gardening tips, and new recipes that he has found.  Seriously, watch it happen.    Anyway, I just wanted to share with you all Joel\'s comments after I told him I will absolutely not do a joint blog with him:   \"Fine, I will take this in my own direction.  Besides, I cant let you limit my creativity and fantastic personality.  Mine is going to be so much better.\"    Bring it, Joel.  Bring it.   '),(12268,3587934,'13,July,2004','2004-07-13 00:00:00','   So, Alex got a new car.  A brand new 2004 Honda Accord to be exact.  He apparently got this car yesterday and drove all around in it, basking in it\'s newness.    Alex, you bastard!  I have yet to ride in said car.  I suppose you consider the people you \"live\" with, your \"roommates\", to be more worthy of a ride in the new car than me and Joel.  Granted, they probably are more worthy.  And cleaner.  And more considerate.  And nicer to you.  And they most definitely smell better (I\'m talking about you Joel!).   But Joel and I were just sitting around all night last night doing nothing.  Well, Joel was watching I Love the 90s and I was playing the Sims, so really we were doing nothing.  You should have swung by in your brand new pimp mobile so we could go flossin\'.  Instead, I was forced to play hours of the Sims, which included making you and Seth fall in love, and making Alan marry Sarrrrrrrra (they now have a baby named Kennedy).    So just forget it now.  I\'m not interested in your fabulous lifestyle.  I suppose now that you have this car you\'ll be trying to pick up women like Paris Hilton and the like.  With the way you talk about it, you\'d think that you were going through a mid-life crisis and just bought a Porshe!  So I want no part of this.  No way.  Don\'t even try to offer me rides in a car that actually has power steering.  And power everything else.  And air conditioning.  Nope, I am just too good for that.  So you just stay away from my Hyundai and my Sims-saturated universe.  Good day, sir.    I said good day!   '),(12269,3587934,'12,July,2004','2004-07-12 00:00:00','   this is the link to that article I am speaking of, in case you either don\'t know Aimee or don\'t give a fuck about her blog.  either way, it\'s:  http://www.commondreams.org/views04/0706-08.htm   '),(12270,3587934,'07,July,2004','2004-07-07 00:00:00','   ok, so I read the Robert Jensen article that Aimee provided the link for.  The man is obviously well read and articulate, and I agree with his Anti-Empire political leanings.  However, all that happy horseshit aside, I can\'t really totally buy his article \"critiquing\" the 9/11 film.  I mean, some of the shit might be true, but......    First, who ever said that Michael Moore was trying to make an Anti-Empire movie?  A lot of Jensen\'s comments seem geared towards the fact that Moore should have made an Anti-Empire movie, not an Anti-Bush movie.  While I may agree with that statement in some ways, I\'m not the one who made the damn movie.  Maybe Jensen should make his own fucking movie.  Maybe he should give his time, money, and energy to the making of an Anti-Empire movie.    And by the way, Jensen seems to think that if the movie had been made the way HE wanted it to be made, it would have changed the world!  He makes it seem as if people would rally around in opposition to today\'s two-party system, and there would be a revolution of such force that the government as we know it would be toppled and oppressed people everywhere would rejoice in their new found freedom from tyranical rule!  RIIIIIGHT.  Hello, Mr. Jensen?  This is France.  It didn\'t work for us.  Get the fucking net.  Anyway....         It seems like there is this constant competition among people to see who can be the MOST left wing, the MOST liberal, who can point to and condemn the most instances of the subtle racism that plagues our society.  It kind of reminds me of that episode of the Apprentice where one girl says to the black woman \"that\'s the pot calling the kettle black\", to which she responded \"there you go with your racist comments again.\"  HAH.  And the sad thing is, some people didn\'t even understand why the comment was so hilariously funny!  fuckers.        Although it\'s true that Michael Moore did not emphasize the treatment of Arab americans since 9/11, I fail to see that the mere fact that a white man who was harassed by the FBI for making anti-Bush comments was included in the movie therefore makes it racist.  PLEASE.  Also, about the Coalition of the Willing - Iceland was also listed in that segment.  Iceland is one of the whitest places on the PLANET.  Also, Morocco DID offer to send in monkeys!  How is showing a picture of running monkeys, when Morocco offered to send them, in any way racist?  Just because they are monkeys?  I think THAT is racist.    I think the point of the segment was to show that the other nations who were a part of the coaltion have basically no military at all, and were therefore only useful in Bush\'s attempt to make people think that other nations supported our military actions, when in fact every nation that mattered in the military sense did NOT support our actions.  I did not get the feeling that Moore was pointing to the fact that the other nations weren\'t WHITE as a way of making fun of them.  Paranoid much?  But then again, I\'m about as white as Powder, so I guess I have no right to discuss issues such as these.     I\'m not saying Moore\'s movie is great, or that it couldn\'t have been done better.  What I am saying is that the film has a purpose - to get Bush out of the White House.  And I think it can help with that purpose.  And getting Bush out isn\'t the ultimate solution to all of the nation\'s problems, that\'s true.  But let\'s be real people.    This ain\'t no fucking revolution.   '),(12271,3587934,'07,July,2004','2004-07-07 00:00:00','   I am so honored to be a member of this blog.  The Sara M McMahan memorial blog.   '),(12272,3587934,'07,July,2004','2004-07-07 00:00:00','   So, motherfuckers, how were your weekends?    My weekend got off to an early and unintended start.  On Thursday night, Patricia, Dawn, and I went out for a few cocktails in this lovely city.  While going down the stairs to the Park Street T station, I rolled my ankle.  Really badly.  I heard popping and snapping sounds, I saw stars.  That definitely SUCKED.  I haven\'t hurt myself like that since I was in high school and playing sports.  By the time I woke up on Friday, my ankle was the size of a baseball, and putting on a shoe was just out of the question.  And so, I stayed home from work.  Most of the weekend was spent limping around and not wearing shoes, which I HATE.  I hate being barefoot.  This morning I am back at work.  The swelling has gone down some, and the colors have changed to yellow and brown.  And, I can wear my shoes, albeit uncomfortably.  So that\'s that.    I saw 9/11 this weekend as well.  Aimee\'s comments are definitely noteworthy and I agree with them.  While I wholeheartedly believe that Bush is the anti-christ, the movie definitely suffers a bit from tunnel-vision.  I do not believe that 9/11 occurred because Bush\'s family had financial dealings with the Bin Laden family.  That\'s just crazy talk.  However, I\'m glad this movie was made and that people are seeing it.  I am a card carrying member of the \"shit all over everything and never be satisfied with anything\" phenomena of our generation.  So it\'s easy to find fault with anything, anywhere.  But the portion of the movie dealing with why we are in Iraq and the fucking freedom/fear rhetoric that has been going on since 9/11 definitely rang true with me.  That, in my opinion, is reason enough for anyone to see the movie.  And also, the shit was damn funny.    I hope all of you have been reading Joel\'s blog.  His last post is all about coming \"out\" as an addicted smoker.  For years now he has been saying \"I\'m not addicted, I just enjoy it\", right before he would beg me for a cigarette (since he rarely has his own, cheap bastard).  It\'s just part of his inherent hipocracy (should be hipoCRAZY, I think.)  He actually believes that he has just NOW become addicted to smoking, after 6 years.  BULLSHIT.  Any smoker on the planet will tell you that shit stinks.  It\'s a damn lie.  But, if he wants to live in his fantasy, I say bring it on, since it certainly is entertaining.    Shit, it\'s almost 9:00 am, and I haven\'t done a damn thing yet.   '),(12273,3587934,'07,July,2004','2004-07-07 00:00:00','   Today\'s Tip of the Day is:  When you go outside for a cigarette break, don\'t take your whole pack of cigarettes with you.  Just take one cigarette(or two, if you don\'t get a break very often and therefore become a chainsmoker for 15 minutes, twice a day).  That way, when the soiled masses start coming up to you and asking you for one, you can honestly tell them \"sorry, I don\'t have any more, I only have this one.\"    Seriously, every time I step outside this building it\'s like a plague of assholes decends on me.  Dozens of people ask for a cigarette.  Even when I tell people I don\'t have one, they still act like assholes.  Sometimes people will say \"yes you do, don\'t lie to me.\" or \"yeah, right\" or \"ok, then can I have some money instead?\"  fuck those fuckers.  I am DONE giving out cigarettes during work.  So I only bring one downstairs with me.  And the next one of those dirtbags who accuses me of lying is going to get a cigarette all right: the burning one that was in my hand just seconds previously will all of a sudden come to rest right in the person\'s eye.    I am not cigarette welfare.   '),(12274,3587934,'02,July,2004','2004-07-02 00:00:00','   You know those people who insist on continually talking on their cell phones while driving?  You know, those morons who can\'t possibly do two things at once and end up being a menace to themselves and others?  Well, I hate them.  And I know you do too.    But listen to this - how about a guy on a BIKE that is talking on his cell phone?  Doesn\'t that seem like an even worse idea than driving and talking on the cell at the same time?  I think so.  But apparently, not everyone harbors these same perceptions.  I was outside smoking a cigarette only minutes ago, and I saw a man ride by on a bike.  He was also talking on his cell phone.  And what happened you ask?  The fucker crashed right into the damn curb, and was launched off his bike and onto the hard concrete.  I believe he broke his cell phone.    Just when I think my day is going poorly, something like this happens and makes my entire world brighten.  Now that I have witnessed that man crashing to the pavement, I feel like my demeanor has been scrubbed clean of all the dirt and filth it had accummulated since this morning.  Nothing short of seeing an asshole get his due does that for me.  Ah, what a wonderful world.   '),(12275,3587934,'02,July,2004','2004-07-02 00:00:00','   So, this morning I had to go back to Middlesex Probate for the 3rd day in a row because those people are entirely incompetent.  I get there, I wait in a long line to get the file I need, find out they LOST the file, wait an HOUR for them to find it, wait in another long line to get the will allowed, go back to the desk and wait in another line to give them back the file, then wait in another long line to order a copy of a document.  It took almost two hours.  So I am really really angry when I leave there to go get the bus back to the office.    In case some of you don\'t know, on Monday they began busing people in between Government Center and Lechmere station, while they do construction.  This is supposedly going to take a whole year, and it\'s hugely inconvenient.  It takes about 20 minutes in between on the bus instead of 5 on the train.  So I went to go get this bus.    I had an argument with the bus driver.  I haven\'t bought my new T pass for this month yet, because I am lazy, but I\'ve really only been taking the bus, then using the transfer that I get from that to get home at night.  So, I only brought 90 cents with me for the Lechmere bus back to the office.   The bus driver stops me and insists that it\'s 1.25 for that bus since it is replacing the T, therefore he requires a T fare, not a bus fare.  I tell him that although that may sound logical to him, 1) it\'s still a BUS, 2) I don\'t have any more change on me and I am NOT getting off the bus, 3) I think the extra 15 minutes it takes me to get to where I\'m going because of this mess is worth the extra 35 cent savings and 4) it\'s not my fault that the MBTA has decided to inconvenience me for an entire year with this busing sh*t, so he\'s not getting an extra 35 cents out of me.  This happened right after I had left the courthouse and I was feeling particularly confrontational, since I had to be all peaches in cream at the probate desk in order for them to help me, when I really just felt like jumping over the desk and bashing their friggin skulls together.  I think the bus driver could tell that I was homicidal, so he just let me go without further rebuttal.  That was probably a smart move on his part.   '),(12276,3587934,'30,June,2004','2004-06-30 00:00:00','   yesterday i wore a drant all day.  not only that, but i kind of liked it.  i think that\'s the first sign.  the end of days is nigh.  you\'ve all been forewarned.   '),(12277,3587934,'24,June,2004','2004-06-24 00:00:00','   Joel swears that he saw that fucked up picture of your eye blink at him yesterday.  I believe him too.  That picutre is just too fucking weird.   '),(12278,3587934,'24,June,2004','2004-06-24 00:00:00','   urlLink    same eye, different problem&nbsp; urlLink   '),(12279,3587934,'23,June,2004','2004-06-23 00:00:00','   i\'m trying out this new feature where i can simply send an email to my blog  and it will publish automatically.  this could either be totally awesome or  the worst thing that\'s ever happened.  ....we\'ll see...  not to worry-- i don\'t have pink eye.    yesterday i woke up and my eye was all gross.  the previous night it had been itchy and watery and meagan suggested (half-jokingly, i think) that i might have pink eye.  but, when i woke up and my eye was crusted shut i thought, \"hell, maybe she was right.\"  as soon as i got to work, i looked up conjunctivitis on webmd to see what my chances of survival were.  as can be expected from webmd, one of the first things that popped up was cancer.  fabulous.  but, before i could begin planning for my eminent demise, i thought i might want to get a second (medical) opinion.  so, i called to make doctor\'s appointment for later on in the day.  when i walked into the office, i was confronted with a woman who was essentially my shorter, older, asian twin.  she seemed to have the same general distaste for human interaction that i do.  and the same tendancy to talk to fill the awkward silence-- a silence that only seems awkward due to our own inherent social ineptitude.  as soon as i walked in the door i launched into my explanation as to why i was darkening her doorstep.  she laughed right in my face.  nice bedside manner, by the way.  apparently, not too many people make an emergency appointment because they think they have pink eye based on a diagnosis from webmd.  with that out of the way, we started talking over each other.  both trying to reassure the other that there was actually nothing wrong with me so we could end the agony of conversation.  it went a little something like this:   doctor: so, how\'s your ankle?  giving you any trouble? me: now that you mention it, yes.  it\'s been making a horrible clicking sound, it hurts when i walk, and i can\'t rotate it at all.  but, it\'s probably nothing, right? doctor: oh, that doesn\'t sound good.  maybe there\'s water on it.  or, maybe it\'s nothing.  you\'re right.  it\'s probably nothing. me: oh, good.  it\'s probably nothing.  thanks. doctor: yeah.  probably...  she looked at my eye for about half a minute before issuing her diagnosis: \"well, i think it might be a little infected.  maybe some dust got into your eye.\"        \"dust\".  i believe that\'s the scientific term for \"shut up, crazy, and take your prescription for placebo drops on your way out.\"    i left the office feeling healthier than ever.  i was limping from the wincing pain in my ankle and my eye was watering and itching like hell, but i had a huge smile on my face.  i\'d heard it from a professional that i was sound as a pound.  and that\'s all i needed to hear.  who am i to argue with medical science?   '),(12280,3587934,'18,June,2004','2004-06-18 00:00:00','   sorry, sara.  this may be a new blog post.  but, you\'ve already heard this before.  i thought i would post it here, though b/c it\'s on my mind.  again.  on my way to work this morning, i saw a poster advertising michael moore\'s new movie, \"farenheit 9/11\".  some saavy activist/grafitti artist had defaced the image and it got me to thinking.  michael moore is a total hack.  i don\'t respect him as a human being and as a documentarian or investigative journalist, he\'s unprincipled and irresponsible to say the least.  so, while i mostly agree with his politics, i can\'t in good conscience get excited about or encourage others to see his latest movie.  especially in light of his recent shenanigans about the release of said movie.  please.  i\'m not fooled for a second by mr. moore\'s shameless posturing.  let\'s see, i\'ll get money from disney to make my movie, then talk about how much i hate corporate media, then i\'ll facilitate/fabricate a \"controvery\" about how disney won\'t release my movie...thereby assuring myself a big box office AND maintaing my corporate underdog status.  give me a break!  i\'m not buying it.  why?  because it\'s not for sale.  if any documentarian is going to have enough of a name to get independent and private financing for his film, it\'s this guy.  hell, he could have bankrolled it himself with all of his cash.  but, he   had    to get to get the money from miramax/disney.  why?  for the reasons i\'ve just laid out.  mr. moore is a fame whore who will stop at nothing to get his ugly mug in the papers.  that being said, i signed up to get free tickets to a showing of it at the aclu tomorrow.  so, i\'m going to see it.  but, i don\'t intend to spend one red cent on it.    in your eye, moore!  in. your. eye.   '),(12281,3587934,'18,June,2004','2004-06-18 00:00:00','   ignore everything i just wrote.  jenny just said my hair looks great.  stylish, even.    i wish she\'d come by earlier...maybe i wouldn\'t have spent a billion dollars at fucking urban outfitters trying to make myself feel prettier.   '),(12282,3587934,'17,June,2004','2004-06-17 00:00:00','   I\'m sorry aimee, but on a scale of 1 to suck, your argument is definitely towards the suck side.    How can having frizzy and unmanageable hair be worse than being a bald woman?  It just can\'t be.  Plenty of people have frizzy hair just like yours.  When someone sees you on the street, they may think \'damn, that\'s some big hair, she should use frizz ease\' or some shit like that.  On the other hand, when people see me, it\'s like \'oh my god, I can see her scalp, that\'s really nasty.  I wonder what\'s wrong with her.  think she has cancer?\'  Huge difference.  It also doesn\'t help that I\'m the size of a hobbit, so everyone can look down and see the scalp quite easily.    I too would shave me head if I could, and just be done with it.  But of course, I also have the same reasons as you do for not being able to do it.    So you see, I win.  Or rather, I lose.  In a big, big way.  But on the bright side, at least some part of me is \'thin\', even if it is only my hair.   '),(12283,3587934,'17,June,2004','2004-06-17 00:00:00','   urlLink    hello&nbsp; urlLink   '),(12284,3587934,'17,June,2004','2004-06-17 00:00:00','   sorry, sara...  but, i have to rebut.  at least you have hair?  believe me, if i didn\'t have a head shaped like a troll doll and didn\'t actually need to keep a job, i would absolutely shave my head clean.  anything would be better than the countless useless fuzzy strands i currently have populating my scalp.  as soon as the weather hints at moisture, my hair balloons up to the size of a hula hoop.  i\'m all curls and frizz; no substance.   every day i go through the same tribulations.  i wake up and spend ages doing my hair.  treating it right with shampoos and conditioners and the right amount of indirect heat.  when i\'m done, i\'ve got a coif of enviable, gorgeous hair (well, maybe not that enviable.  it\'s still two-toned like a mo-fo).  then, as soon as i step out into the human soup that is new york city, it starts rebelling.  at first it\'s just the cantankerous ones up front at the crown.  then, word spreads and by the time i get to work at 8:30 in the morning it\'s full-on mutiny.  i look like i just went through 40 hours of shock treatment.  and then was set on fire.    looking my absolute worst every single day might not chafe quite so much if i didn\'t have a glimpse of how good it *could* be-- it almost was...if only i never ever left the house.  damn.  that kills.  why do i even bother?  oh, right.  because i look like shit in hats.  oh, and p.s.-- this applies to my hair year-round, ladies and gentlemen.  year. round.  \'nuff said.   '),(12285,3587934,'17,June,2004','2004-06-17 00:00:00','   urlLink    i don\'t like birds&nbsp; urlLink   '),(12286,3587934,'17,June,2004','2004-06-17 00:00:00','   o.k., i felt bad about that last post.  i shouldn\'t let my seasonal depression seep into my blog.  this isn\'t livejournal, for fuck\'s sake!  i watched the simple life last night.  i don\'t know, those two girls just make me so sad.  the whole time i\'m watching it, i\'m thinking, \"oh, this is funny.\"  they say hilarious things, so it\'s good t.v.  but, then they say really fucked up things like, \"that\'s so ghetto.\" as if either one of those pampered princesses have ever even seen a real ghetto...let alone understand what the word means.  and when i start to analyze it like that, it makes me sick.   i guess that\'s the minefield that is reality television.  human beings are flawed and if we scrutinize them too closely, we\'ll just be saddened.  but, as pure fluff, it\'s pretty damn funny.  i especially like nicole.  although, the whole time i was watching it, i was wondering if this was filmed before or after she went into rehab for heroin addiction.  and that was pretty distracting.    paris is creepy.    i just found out that paris and lil\' john are hanging out these days.  what?!  that\'s so so wrong.  what could they possibly have in common?  except, i guess a shameless desire to garner more and more fame and money.  how very shallow.  and very hollywood.  they deserve each other.  i hear they\'re collaborating on a rap-pop album.  i\'m sure it\'ll be nothing less than pure genius.  paris and lil\' john can do a polycultural jig over the bloated, fetid corpse of what we used to call hip hop.   '),(12287,3587934,'15,June,2004','2004-06-15 00:00:00','   i started writing this blog earlier today.  it was all about human relationships and buddhist philosophy, if you can believe it.  some maudlin bullshit about how we depend on other people to make us happy, when we should really look within ourselves to find peace.  can all that noise.    i\'m in a bad mood today.  it\'s humid as all hell, my hair looks like a brillo pad dipped in animal fat, and i have a splitting headache.  i have to decide between being a good friend to my roommate and going home to wait for her boyfriend to get in from scotland OR being a good friend to someone at work and going to a doctors without borders lecture.  i pretty much lose either way b/c i\'m not in the state of mind to be social tonight at all.  all i really want to do is go to my room and cry.  it\'s a crying kind of day.   '),(12288,3587934,'15,June,2004','2004-06-15 00:00:00','   urlLink    prisoners of the blog&nbsp; urlLink   '),(12289,3587934,'14,June,2004','2004-06-14 00:00:00','   i no longer have a cell phone.  it\'s been exactly 16.5 hours and i can barely function.    but, discussing our culture\'s dependency on digital technology is not the point of this blog.  please.  what do you take me for?  an intellectual?  hardly.  i\'m a cantankerous old fool.  and i\'m about to get grouchy with it--  why don\'t i have a cell phone?  let me explain.  see, when i moved to new york from atlanta, i thought it was too much of a hassle to bother with changing my number.  until about 2 weeks ago when jason casually asks me how much i pay for my cell service.  when i admitted it was a whopping $80, he was dumbfounded and told me that as soon as my contract runs out, i should switch to a more reasonable company.  turns out, my out-of-state number was racking up some serious fees and surcharges.  thanks for nothing, sprint!    contract.  you raise an interesting point, john...  turns out, my contract expired 2 years ago.  i was in the clear!  so, being the genius that i am, i decided to not waste any time and cancel my sprint service immediately.  why the hell not, right?  in my haste, i failed to set up replacement service for when sprint was over.    instead, i thought-- it\'ll take two seconds.  i just walk into the verizon office, put down my cash and walk out with a new phone and plan.  oh, i\'m so naive.  on friday i went to the verizon store to enact my brilliant plan.  i picked out the cutest phone in the store, filled out the requisite 9000 forms, and braced myself for my brand-spanking-new \'917\' phone number.  sadly, it wasn\'t two seconds after the verizon guy took my forms and entered them into his computer that he went from jolly sales guy to disapproving sales guy.  he gave me his gravest face and said,, \"due to your lack of credit, you\'ll have to put down a $1,000 deposit to get a phone today.\"    i responded the way any rational person would in that situation.  i said, \"say what, sucka?!\", grabbed my forms back, and ran out of that store.     so endeth my sad tale.  and, with it, my days of easy accessibility.    not having a cell phone is kind of like being in olden times.  maybe i\'ll start writing letters.    or, you know, get my mama to open an account for me.   '),(12290,3587934,'13,June,2004','2004-06-13 00:00:00','   To start with: two things.  If you don\'t want to read spoilers about Gothika, this is not the blog entry for you. Secondly, I tried to put this into paragraphs, but blogger.com would rather it look like a crazy person wrote it.  How appropriate given the topic...  So Aimee said I could post on her blog.  She\'s pretty nice, huh? She\'s so nice that she agreed to sit down and watch Gothika with me this afternoon.  And what an afternoon it was.  I thought Gothika was going to be one of your sub-par thrillers where no one believes the wrongly accused heroine, who goes through the motions of investigating her case herself (since the police are so corrupt/incompetent) and is eventually vindicated when she gets too close to the truth and the real killer reveals himself.  And, I was right.  But I had no idea that, aside from the last 20 minutes, Gothika was the scariest movie of all time.  I screamed out loud, people...but more importantly, Aimee gasped loudly.  Aimee. Gasped. Loudly.  And there wasn\'t a bird or a clown involved.  You know that\'s some scary shit.  So, anyway, the point of this post is not to talk about Gothika. It\'s to talk about Roger fucking Ebert.  Usually when I watch a movie, I like to check it out afterwards on imdb, and I usually see what good ol\' Ebert has to say about it, too.  There was a time, I\'m told, when he was a respected film critic.  Maybe he misses Siskel, maybe it\'s the stroke, but that guy is not on his game.  He doesn\'t seem to even have the decency to watch a movie before reviewing it anymore. Last week after watching the Magdelene Sisters, I read his review and found that not only had he misspelled a character\'s name, he had atrociously misstated the plot.  Hello, Ebert? Anyone home? You\'re fucking busted, dude. Plus, those were real girls you didn\'t bother to fact-check, asshole.    As for Gothika, he basically tells us to accept the fact that the plot makes no sense and that the end is terribly dissapointing...because that\'s just what you do when you watch a thriller. I\'m sorry, but aren\'t you a film critic?  I\'m fully in favor of suspension of disbelief when watching movies, and I\'m not afraid of an ambiguous ending, but I expect the director/writer/gaffer/best boy to extend me some common fucking courtesy.  In a supernatural movie, suspension of disbelief does not apply to legal matters.  The ghost isn\'t a lawyer, is she? I was totally willing to bite my tounge when the ghost was writing messages on windows like Karen from Watcher in the Woods, but killer mental patients don\'t just get let out of jail because they\'re friends with Halle Berry.  And, by the way, Ebert? We know who killed Halle Berry\'s husband.  Yeah, we do.  The movie clearly wants us to believe that she was somehow vindicated, but I didn\'t know you could plead \"possession\" and waltz out of jail like Maria leaving the nunnery. I\'ll keep that in mind the next time my bus nemesis takes my seat.  I\'ll just tell them that Ronald Reagan took over for a minute there.  Oops! Pesky guy.    It all boils down to: shut up Ebert, shut up Gothika, and shut up again, Ebert.     -Cara B.   '),(12291,3587934,'11,June,2004','2004-06-11 00:00:00','   urlLink    excuse me?&nbsp; urlLink   '),(12292,3587934,'10,June,2004','2004-06-10 00:00:00','   it\'s only day two and i already feel guilty about not posting.  evelyn just called to say that i had to post something-- right now!  yikes.  i called in sick to work today, in the hopes of following monk\'s advice (via oprah) to live deliberately.  and, here it is, with the day half gone and i haven\'t even done enough today to justify a post on my blog.  sorry, ladies and gentlemen.  i\'m a failure.   '),(12293,3587934,'09,June,2004','2004-06-09 00:00:00','   urlLink    say what?!&nbsp; urlLink   '),(12294,3587934,'09,June,2004','2004-06-09 00:00:00','   so, i\'m sitting on the train quietly reading my horrible book (as i lay dying-- faulkner can kiss my ass, by the way) and listening to my headphones when a blind man walks onto the train.  and by \"walks\" i clearly mean, \"rampages\".  he storms into the train flailing his stick at all and sundry, presumably to get his bearings...but, he was swinging that thing with such fury he almost hobbled me.  he came to a stop right in front of me.  as i was reading and couldn\'t hear anything, i didn\'t notice that he was blind.  i just thought he was an asshole and i almost punched him in the face.  when i looked up, it became obvious to me that he was blind and he definitely expected me to get my sight-having ass up out of my seat and relinquish it to him.  which i would have done b/c i\'m sort of a nice person.  but, the woman sitting next to me saw him first.  obviously, to prove her extreme martyrdom, she gives me this exasperated look (like she can\'t believe i didn\'t get up since he\'s standing in front of me) and very deliberately vacates her seat.  excuse me, but my foot is broken.  don\'t i deserve a seat if for no other reason than i got there first?  jackass.  so.  the blind man sits.  and proceeds to make a huge production out of it-- all shifting wildly and whipping around his backpack.  yeah, i know he can\'t see.  but, at this point, he\'s hit me with his cane enough times to know where i am.  not to mention the fact that we\'re sitting next to each other.  where else would i be?     he goes on to pull out this huge ledger-style braille book.  now, i don\'t know much (or anything, really) about braille, but i have to believe that you\'d read braille in much the same way you would read anything else.  this guy was running his hands over the pages with such rapidity i thought he was having an episode.  after every few pages he started rubbing his hands on his face-- up and down, back and forth, and then kissing his hands.    my first instinct was that he was praying.  so, i stole a glance in his direction so i could figure it out.  because, by now, i\'m invested.  upon closer inspection, i realized that he\'s not praying at all.  he\'s just totally fucking insane.    after sitting next to him awhile, i didn\'t really notice the slight jostle i\'d get every time he started in on his facial machinations.  but, i did start to think about what it would be like to be blind.  so, i closed my eyes and tried to imagine what his experience would be like.  then, i got really freaked out b/c i got really disoriented and thought that if i was blind, i would be falling all the time.  and then i felt sad.  it\'s no fun being blind.  especially when dumb bitches like me refuse to give you a seat on the train and then get mad at you for being crazy.  and then make it their first entry in their blog.   '),(12295,3587934,'26,July,2004','2004-07-26 00:00:00','   urlLink    i\'m not really this short.  but, i am usually this pissed off.  so what?&nbsp; urlLink   '),(12296,3587934,'23,July,2004','2004-07-23 00:00:00','   urlLink    the siren music festival: hipsters as far as the eye can see!&nbsp; urlLink   '),(12297,3587934,'23,July,2004','2004-07-23 00:00:00','   i went to the doctor again today.  i spent 4 hours in that office, consulted with two health care professionals for a total of 15 minutes, and came away feeling far worse than i could have ever imagined.  i don\'t think i\'ve heard anyone say the word \"death\" so often in a single conversation.    and then it rained.  i think i\'m going to spend the remainder of the weekend holed up in the squalor of my tiny room.  i\'m done with everything.   '),(12298,3587934,'20,July,2004','2004-07-20 00:00:00','   i\'m sitting here with my brow furrowed and i can\'t even do any work b/c all i want to do is sit in a chair and scream my face off.&nbsp; i\'m back down at my old desk.&nbsp; however, my old desk (and my old job) currently belong to this old woman named sandra.&nbsp; about 3 months ago i was told that i was being moved from my job as assistant to two people to assisting the director.&nbsp; sounds like a promotion, but really it wasn\'t.&nbsp; it was just an annoyance.&nbsp; but, i\'m a trooper, so&nbsp; i agreed.&nbsp; also, i hated my two bosses, so i was psyched to get the hell away from them.  &nbsp;  however, when i moved to my \"new\" position, i spent the first month and a half basically working two jobs-- my old one and my new one b/c the old lady they got to fill my position is totally incompetent.&nbsp; i put up with that, too b/c i\'m a masochistic idiot who doesn\'t know how to assert herself professionally.  &nbsp;  it wasn\'t until very recently that i finally got comfortable in my new role.&nbsp; i wasn\'t working for my old bosses at all.&nbsp; i was almost enjoying the monotony of it all.&nbsp; and then bam!&nbsp; sandra\'s son dies.&nbsp; then, she gets hospitalized for having a seizure.&nbsp; and now she\'s been gone for weeks.&nbsp;&nbsp;i got a call this morning&nbsp;asking me to come back down to cover&nbsp;for her while she\'s gone-- INDEFINITELY!&nbsp; argh!  &nbsp;  so.&nbsp; i\'m back at my old desk.&nbsp; however, it\'s obviously not mine anymore.&nbsp; and it\'s literally covered in filth.&nbsp; there\'s&nbsp;papers everywhere.&nbsp; there\'s a mountain of dust and grime caked up on the keyboard.&nbsp; it\'s like this desk was excavated out of a coal mine.&nbsp; how can one tiny old lady be so disgusting?&nbsp; god.&nbsp; i don\'t need this.  &nbsp;  i love how i\'m always the one to get fucking fucked around here.&nbsp; it\'s just not fair.&nbsp; i\'m back down here cleaning up her mess-- both physically and fiscally.&nbsp; again.&nbsp; i\'m done.&nbsp; this is just it.&nbsp; i\'m currently living my own worst case scenario.&nbsp; i\'m right back in the job i loathe working for the two worst bosses in the history of the world.&nbsp; i can\'t quit because i\'m too poor to live WITH the job...imagine the&nbsp;horror of living&nbsp;WITHOUT&nbsp;it.&nbsp; i may not get the financial aid that will allow me to go to school in the fall.&nbsp; so, i\'ll have to forfeit my acceptance to school which,&nbsp;quite frankly, is&nbsp;the one good thing that\'s happened for me all year.&nbsp; i have yet another stupid boring crush on a boy (which, believe me, i\'m really pissed about.&nbsp; i don\'t want this extra stress in my life right now.&nbsp; my self esteem is bad enough already.)&nbsp;  &nbsp;  &nbsp;i feel like i\'m going to burst into tears right now.&nbsp; i\'m just so frustrated and i feel powerless.&nbsp; yes, this is my life.&nbsp; smell that?&nbsp; yeah.&nbsp; that\'s not success you\'re taking a whiff of.  &nbsp;  i\'m just sick to death of everything.&nbsp; i give up.&nbsp; can i resign from my own life?  &nbsp;  o.k.&nbsp; mid-blog update.&nbsp;  &nbsp;  i feel moderately better than i did earlier.&nbsp; nothing is fixed, nor do i see any solutions to my stupid predicaments...  &nbsp;  but, steve asked me if i\'d like to go for a walk.&nbsp; and i was like, \"hell yeah!\"&nbsp; and then diane was going out for her cigarette break, so i was like, \"yeah.&nbsp; you\'re going to have to give me one of those.\"&nbsp; she was like, \"oh no!&nbsp; i don\'t want to corrupt you.\"&nbsp; please.&nbsp; corrupt this!&nbsp; don\'t worry-- she gave it to me.&nbsp; and it was&nbsp;awesome.&nbsp; even thought it&nbsp;was a virginia slim ultra light menthol (gag).&nbsp; it was the greatest.&nbsp; then, when i was walking back from the delivery/pickup i got catcalled.&nbsp; normally, that makes me feel awful and self-conscious.&nbsp; but, i really like my outfit today.&nbsp; i feel really hip and cool.&nbsp; i have on these great earrings which jason said made me look like a ludlow cokehead.&nbsp; hot.&nbsp; so, i\'m basically strutting down the street in my impossibly high platform wedges (which i\'ve worn exactly once since i bought them last&nbsp; year because, well, they\'re impossibly high and they hurt my stupid ankle.)&nbsp; anyway, i\'m walking down the street and i hear these guys talking as i walk.&nbsp;&nbsp;it\'s on the periphery until i hear someone&nbsp;yelling, \"what?!&nbsp; you can\'t even say hi?&nbsp; god!\"&nbsp; and i turned around because i wanted to know&nbsp;what all the commotion was about and the guy is looking right at me, yelling.&nbsp; so, i screamed out, \"you\'re talking to me?&nbsp; why??\"&nbsp; and he&nbsp;goes, \"because you\'re&nbsp;the only sexy woman on the street wearing a green shirt.\"&nbsp; and i burst out laughing and said thank you and walked on.&nbsp; then, this woman (another total stranger, naturally) walked up behind me and goes, \"well, that was cute.\"&nbsp; and you know what?&nbsp; it kind of was.&nbsp; thanks, stranger!&nbsp; as pathetic as it sounds, you made me feel ever so slightly less crazy and useless today.  &nbsp;  sometimes being sexually harrassed on the street can be the best thing that ever happened to you.   '),(12299,3587934,'19,July,2004','2004-07-19 00:00:00','   i generally hate sundays.&nbsp;&nbsp;it\'s just&nbsp;my least favorite day.&nbsp; it always has been.&nbsp; sundays make me think of homework and stressing about school the next day.&nbsp; they make me think of sitting in a stifflingly hot and humid church with fans pathetically ruffling the hair on the back of my neck enough to make me remember how miserable i am.&nbsp; if sunday had a sound, it would be the blues.&nbsp; or funk.&nbsp; if sunday had a taste, it would be cinnamon.&nbsp; disgusting.&nbsp;  &nbsp;  however, for the past four years, my good friend, t.v., has&nbsp;given me&nbsp;at least one thing to keep me from sleeping the miserable day away.&nbsp; that one thing is a little show called \"six feet under\".&nbsp; and up until last night i thought it was absolutely the bee\'s knees.&nbsp;  &nbsp;  why?!?!&nbsp; why would alan ball perpetrate such a heinous crime on the viewing public?&nbsp; does he have any idea how much money we pay for cable every month?&nbsp; i demand reimbursment for both my time and hard-earned dollar bills, y\'all!&nbsp; he took the one solitary character that i have heretofore found no fault with.&nbsp; i thought michael c. hall was brilliant.&nbsp; i thought the character of david was expertly drawn.&nbsp; in short, i loved him.&nbsp; and now he\'s been ruined.&nbsp; he\'s going to have to deal with the aftermath of this ridiculousness for the remainder of the season, if not the series.&nbsp; grr!  &nbsp;  that episode was just so impossible to watch.&nbsp; it made me uncomfortable and frustrated.&nbsp; two things i don\'t look for when watching t.v.&nbsp; i get enough of that jazz in my&nbsp;own life.&nbsp; &nbsp;  i had a lot more steam when i first started writing this blog.&nbsp; but, then i talked to the monkey.&nbsp;&nbsp;she&nbsp;made some good points.&nbsp; she said&nbsp;that perhaps one day we\'ll look back on this episode and realize that it was genius.&nbsp; maybe she\'s right.&nbsp; maybe i\'m not giving alan ball enough credit.&nbsp; maybe he\'s pushing the envelope right up to the edge...&nbsp;  &nbsp;  OR maybe he\'s pushing the envelope right over the edge of the desk and we just watched sfu jump the shark.&nbsp; in which case, he can go ahead and&nbsp;accept my engraved invitation to &nbsp;kiss my ass.&nbsp;  &nbsp;  obviously, i\'m too invested at this point to bail out now.&nbsp; so, i\'ll be tuning in next week to find out if my favorite show just got that much better...or if i lost one of the only things that kept me going on sundays.   '),(12300,3587934,'15,July,2004','2004-07-15 00:00:00','   i\'m so annoyed.   i just read roger ebert\'s review of \"before sunrise\".   shut up.   as much as i make fun of my brother for LOVING \"before sunset\", i have to admit that i liked it, too. and i was interested in knowing how the stupid story ends with these two insufferably pretentious characters.   unfortunately, i was on ebert\'s page looking to see if he reviewed \"the corporation\" (which he didn\'t, naturally) and i saw that he gave \"before sunrise\" a million stars. before i could stop myself, i clicked on the review.   damn him!   he managed to reveal every aspect of the film that might have been slightly interesting for me, the potential viewer, to find out. how infuriating! not only is he incapable of researching and writing a decent film review these days...but, he can\'t even be bothered to stay away from totally spoiling the plot. good grief.   now, i feel like i\'ve been with ebert through a lot. and, for that reason, i\'ve given him more than his fair share of the benefit of the doubt. i did this because, while i didn\'t always agree with him, i generally took him to be a skilled writer and a well-versed film critic. but, lately he\'s been disappointing me more often than not. he haphazardly awards star after star to total schlock, and, as cara pointed out in her criticism of his gothika review-- the man can\'t even be bothered to watch the movie half the time.   oh, ebert... you\'re pretty spry for a man your age. i didn\'t realize you were physically capable of jumping the shark so soon after having a stroke. kudos to your physical therapist.   '),(12301,3587934,'14,July,2004','2004-07-14 00:00:00','   the fleet bank atm branch is open on broadway and astor place.  my prayers have been answered and my life is now officially   that   much easier.    loves it.   '),(12302,3587934,'09,July,2004','2004-07-09 00:00:00','   urlLink What Michael Moore Misses About the Empire   this article articulates what i was trying to say about this movie so much better than i could ever dream.   '),(12303,3587934,'08,July,2004','2004-07-08 00:00:00','   i went to see \"farenheit 9/11\" tonight.  taking into account my recent post about it, it will surprise no one that i had some issues with the film.  i got a lot of comments after my first post about how i wasn\'t seeing the big picture-- that it\'s amazing that a film like this was not only made, but was also successful.  and that plenty of people wil see this film that may never have gone to see somehing like it before.  it could open people\'s minds and get them to think.    but, the question stands-- thinking about what?  what i got from the movie is that george w. bush is the reason everything in the world is in the shitter.  9/11?  bush\'s fault.  poverty in the u.s.? bush\'s fault.  civil unrest in the u.s. and abroad?  bush\'s fault.  the weight of the world is on his shoulders.  of course, he deserves plenty of the blame.  but, ALL of it?  come on.  that\'s the most naive assessment of u.s. history i\'ve ever seen.  and it\'s a dangerous assumption.  what proof is there that were gore/leiberman in the white house that so much would be different now?  does anyone not remember clinton and his relationship with iraq?  or wha about a little something called plan colombia (i\'m looking at you, too, joe lieberman)?  that was not only A democrat, but THE democrat, dudes.  hello.  it\'s this kind of journalistic laziness that i abhor in mr. moore.    i don\'t know...  it\'s late and that movie clearly made me a little emotional.  i\'ve been feeling more and more disenfranchised (in general) lately and this movie only served to push the point home.  my voice isn\'t being heard.  my views are not represented by my government.  when i look at the faces of george bush and dick cheney and john kerry and john edwards, i can\'t see that they have my best interests at heart.  i can\'t even count on the so-called \"liberal mouthpiece\", michael moore, to present a cogent argument that can hold at least a teaspoon of water.  i\'m not saying that we should all boycott the election.  and i\'m not saying that it doesn\'t make any difference who is in the white house.  i\'m certainly no bush lover.  i absolutely believe that it matters.  and i\'ll definitely be voting and possibly campaigning for john kerry.  however, i don\'t think it\'s particularly prudent to turn a blind eye not only to what has been going on, but to what is in store for the future.    maybe i\'m the naive one.  maybe i\'m the crazy one.  maybe i\'m just being a hater.  who cares?  these are just the thoughts of one sad girl on a balmy summer night.   '),(12304,3587934,'01,July,2004','2004-07-01 00:00:00','   at the bidding of my friend and financial advisor (although, i think he quit after i paid to see \'the day after tomorrow\' for the second time...who can blame him?) and in the hopes of becoming less of a financial paraiah and possibly taking one baby step towards adulthood, i requested a copy of my credit report.   there\'s nothing that can make you feel like more of a total fuckup than seeing in black and white courrier typeface every single one of your fiscal missteps.  and of course, they put the whole damn 5 page letter in bold, all-caps just so you\'re not mistaken that you are a very very bad person.    dear transunion,   i get it.  i know that i\'m a horrible person who deserves nothing more than to be left for dead in the desert, at the mercy of the sun and wild dogs.  i swear, i didn\'t mean to rack up a whopping $149 in debt at macy\'s in 1997.  i\'m sure nothing i bought then is worth the lifetime of pain and suffering i seem to have accrued at the ripe old age of 25.  if i could, i would totally return it.  please forgive me.  love, aimee  yeah.  i\'m pretty sure it doesn\'t work that way.  the road to hell is paved with sales recipts.  and i\'m more than halfway there.  it looks like i\'ve got a lot of scrimping and saving in store for me.  but, hey!  i\'m already saving myself money as i type.  i\'m home sick today (this time i\'m actually sick...if you can believe it.  i hate being home sick when i\'m sick.  what\'s the fun in that?  there isn\'t even anyone around to feel sorry for me and make me tea.)  but, being home sick does have its advantages.  for instance, i\'m not putting myself into harm\'s way by walking past countless shops.  i\'m on the road to recovery here, people...  oh, who am i kidding?  my life sucks.    and i think it\'s going to rain today.  bah!   '),(12305,3587934,'01,July,2004','2004-07-01 00:00:00','   urlLink    i looked so much hotter than this bitch&nbsp; urlLink   '),(12306,3587934,'01,July,2004','2004-07-01 00:00:00','   i\'ve been getting a lot of inquiries since my last post about what exactly a \'drant\' is.  i\'m sorry.  i thought it was common knowledge that a drant is a dress worn over pants (not to be confused with the infamous skant-- skirt worn over pants).  the drant is the province of hippies and faux hippies.  oftentimes you\'ll find loathesome hipsters sporting the drant/skant situation proudly as well...only in a dirtier, more rackish way.    anyone that knows me can see i\'m certainly by no means a hippie.  i hate peace and love and dirt.  i like to wear real shoes and i hate the dead.  no, thanks.  shrugging off the hipster monicker is a little harder b/c, while i don\'t see myself as a hipster...i guess i do like their style of dress to the point of mild emulation, occasionally can be found listening to their music, and i like irony just as much as the next gal.  so, there you have it.  HOWEVER, i have always come out strongly against the drant.  it\'s been second in rantings-per-sighting only to the infamous backless shirt (which i will never approve of).    why did i do it, you ask?  well, the way i saw it, i simply didn\'t have a choice.  i was late to work, had no other clean clothes, and that dress was far too short to wear solo to work.  i\'m not a prostitute after all!  what would you have done in my situation?  that\'s right-- you would have grabbed a pair of black pants, threw them on under your dress, and strode down the street like you knew exactly what you were doing and why.  this is new york city after all.  i\'ve seen people wandering the streets in leg warmers, teddies, chainmail vests...i even once had an encounter with a guy wearing a filthy, tattered ball gown and combat boots.  granted, that guy was a heroin addict proposing marriage to me in washington square park for giving him a dollar.  but, he was still working his ensemble.  in the end, i think i\'m going to have to eat my words.  that outfit was damn comfortable.  and cute, to boot.  i\'m mos def gonna be rocking it again in the not-so-distant future.  so, when i said i hate drants, i guess i lied.  what else is new?   '),(12307,3587934,'10,August,2004','2004-08-10 00:00:00','   urlLink    it\'s my uncle playing sax at the dnc!  how hot is that?  (he\'s the second from the front-- amadee castenell)&nbsp; urlLink   '),(12308,3587934,'03,August,2004','2004-08-03 00:00:00','   i had so many things planned for today.  i was going to set things straight-- really get everything in order.  instead, i spent the lion\'s share of the day sleeping in some pretty awkward and uncomfortable positions.  not going to bed.  literally, i\'d be sitting at the table drinking tea and then ten seconds later i\'m counting zzz\'s.  or typing happily away at the computer and end up taking a 5 minute \"nap\".  it doesn\'t help that the dreams i\'m having are both disturbing in content and startlingly real in quality.  earlier i dreamt that all of my friends and i worked in a sporting goods store.  i was learning how to skateboard.  and in the back of the shop we packaged cocaine for distribution.  funnily enough, in my dream, much like today, i called in sick to work.  i think i may have narcolepsy.Â  can you just randomly develop that?  like a cold...or schizophrenia?   '),(12309,3587934,'03,August,2004','2004-08-03 00:00:00','   the great thing about being home from work is that you can watch all the crappy t.v. shows that you\'re too embarrassed to watch when everyone else is home.  like, i\'m not going to sit here and watch the entire 4th season of 90210 when meagan\'s home.  that\'s just rude.  but, if i\'m home alone you better believe i\'m kickin\' it with the west bev crew.  that\'s just how i roll.  not so, today.  on sunday we were rendered normal people.  our t.v. is braindead.  we can no longer record 2 shows and watch another.  we don\'t have a list of shows stored on the t.v. for later viewing (goodbye, walsh\'s!).  no more pausing live t.v. to make a point or a suggestion or continue a conversation.  gone are the days when we could watch ET and XTRA simultaneiously with the aid of PIP and pause.  instead, we have to sit here like saps and watch LIVE t.v.  oh, the inhumanity of it all...    i don\'t know how we\'ll hold up until thursday.  i guess that\'s why we still have tivo.  and 3 televisions.    it sucks being normal!   '),(12310,3587934,'30,June,2004','2004-06-30 00:00:00','   this is perhaps one of the best satirical pieces I have ever read on The Onion.  Its hilarious.   urlLink  '),(12311,3587934,'29,June,2004','2004-06-29 00:00:00','   On a brighter note, I got a raise today, which is just an all around good feeling.  Looks like I\'ll beat inflation after all.  Fuck you Bush!  '),(12312,3587934,'29,June,2004','2004-06-29 00:00:00','   I really do think that the Lollapalooza cancellation is going to set off a downward spiral in my life that I will never recover from.  The one band that I really really wanted to see has announced a headlining tour in place of the former monumental event.  However, the closest date is in Montreal?  Fuck Fuck Fuck Fuck!!! I hate you Canada, you dont deserve this.  Nique Le Cananda!  This anger may seem irrational to many, but there is good reason, believe me.  Modest Mouse (which is the band I have been talking about) has only played in New England twice since their inception (which is like 10 years).  I was all set to rock out.    Just stab me in the eye with a rusty nail.  '),(12313,3587934,'29,June,2004','2004-06-29 00:00:00','   you\'re damn skippy this blog rocks.  but, i\'d say it has more to do with the fact that i\'m a member than anything else.    i\'m cyberspace\'s number one stunna-- to the max, don!  just call me the spirit killer.  '),(12314,3587934,'27,June,2004','2004-06-27 00:00:00','   I am TOO a contributing member.  I write all the answers down and bring up the answer sheet to the guy who runs it.  I am the all-singing, all-dancing crap of the world.    Why don\'t you come to trivia and show us how much more you can participate than I can?  Oh wait, I forgot, you won\'t come because you have to be in BED by 10:00 pm or you\'ll just be too exhausted to work the next day.  Too bad trivia doesn\'t start at 5:30 so you can be home in time for your precious 9 hours of sleep per night.  Weirdo.  '),(12315,3587934,'27,June,2004','2004-06-27 00:00:00','   I\'d like to think that New Englanders are hardy people, and infact we are.  Speaking for myself, I\'d say I\'m one tough motherfucker.  But honestly, I dont think I can weather the Pixies storm.  I\'ve been left out in the cold for too long now.  Where is the love???  For those that do not know, the Pixies announced back in March or April that they were getting back together, and as a result, kicked off a reunion tour.  This is very exciting news, and from what I hear they are still kicking ass like they used to.  I waited anxiously as tour dates were announced, and wondered why they chose to begin in Minnesota, Saskatoon and Winnipeg?  Fuck Canada, they dont deserve special treatment.  More and more dates were added and still nothing west of the Mississippi.  Dates across Europe were revealed pursuant to the west coast leg, once again ignoring their New England roots.  Finally, after months of waiting, the Mullin Center at UMass Amherst will be graced with their presence, but I\'m still fucked!  There is only one date, in early December, so I have to compete with thousands of drunken college sluts and the rest of the region for tickets.  This sucks!!  I\'m at the mercy of the weather....  '),(12316,3587934,'27,June,2004','2004-06-27 00:00:00','   urlLink    Open the fucking present Piotr!  &nbsp; urlLink  '),(12317,3587934,'25,June,2004','2004-06-25 00:00:00','   urlLink    Kill Piotr Volume I&nbsp; urlLink  '),(12318,3587934,'25,June,2004','2004-06-25 00:00:00','   urlLink    Candid shots are always the best&nbsp; urlLink  '),(12319,3587934,'25,June,2004','2004-06-25 00:00:00','   urlLink    You think Paris Hilton is Creepy...                            Check this bitch out.  If you surf the internet as much as I do, this scary face is on like every website.  Everywhere I turn I see that beak-nosed smirking slut staring at my soul with those creature eyes.  Its really sinister.&nbsp; urlLink  '),(12320,3587934,'23,June,2004','2004-06-23 00:00:00','   Due to time constraits, my first posting must be quick.  Here\'s some song lyrics for y\'all to chew on.    THIRD PLANET by Modest Mouse   Everything that keeps me together is falling apart, I\'ve got this thing that I consider my only art of fucking people over My boss just quit the job says he\'s goinÃ­ out to find blind  spots and he\'ll do it The 3rd Planet is sure that they\'re being watched by an eye in the sky that can\'t be stopped When you get to the promise land your gonna shake that eyes hand Your heart felt good it was drippinÃ­ pitch and made of wood And your hands and knees felt cold and wet on the grass to me Outside naked, shiverinÃ­ looking blue, from the cold sunlight that\'s reflected off the moon Baby cum angels fly around you reminding you we used to be three and not just two And that\'s how the world began And that\'s how the world will end A 3rd had just been made and we were swimming in the water, didn\'t know then was it a son was it a daughter When it occurred to me that the animals are swimming around in the water in the oceans in our bodies and  another had been found another ocean on the planet given that our blood is just like the Atlantic And how The universe is shaped exactly like the earth if you go straight long enough you\'ll end up where you were  '),(12321,3587934,'14,July,2004','2004-07-14 00:00:00','   So I\'ve been very busy these past few weeks, caught up in a little thing we at work call \"Universe Creation\".  Let me explain.  Aside from the Big Bang, which theorists believe was the cosmic event that gave birth to the actual universe as we know it, every quarter the ICC (Independent Consultant Cooperative) creates a universe of client data (what I like to call the mini bang).  (As an aside, I was reading a scientific journal the other day and apparently astronomers have uncovered evidence that disproves the Big Bang theory.  This group of rouge scientists have found red shifts in extremely distant galaxies that indicate these galaxies formed around the time of the Big Bang.  Part of Big Bang theory stresses that galaxies must have formed at a later time after the initial cosmic event.  These extremely old, and distant galaxies do not fit into the Big Bang puzzle, which I thought was very interesting.  However, this discovery could just be useful to Bible bangers and creationists, but enough of this for now...)  Back to Universe creation.  Every quarter, portfolio returns, characteristics and holdings are uploaded to the ICC universe of client data.  After the deadline has passed, all this data is combined, crunched, accelerated to the speed of light, and then sent through a cosmic worm hole where it is then ready to present all nice and neat to our clients.  Actually, for those that are familiar with statistical analysis, we are just creating a population of data, where we are able to compare sample data (specific client info) against all peers.  This is especially useful when comparing returns.  Given the competative nature inherent in all human beings, we want to see how we did compared to everyone else.  Our universe allows this to happen.  We are able to determine mean returns for fixed income, equity, and international equity asset classes, amongst a whole slew of others.  Portfolio returns can also be ranked in their respective universes (parallel universes, he he, I can go on and on with this nerdy humor, somebody stop me!).  And so, my deadline is this saturday, so I\'ve been mad busy, and probably busier after this weekend as reports begin to generate.  Just call me He-Man... Master of the Universe!!!  '),(12322,3587934,'12,July,2004','2004-07-12 00:00:00','   I\'ve added links to the right for all people doing the blog thang.  This should make sharing ideas, fighting, bitching and whatnot for my fellow bloggers much easier.  '),(12323,3587934,'09,July,2004','2004-07-09 00:00:00','   In rebuttal to Sara\'s comments on her blog, I feel that I must back up my previous statements with some more information.  It is true that I have been smoking off and on for nearly six years.  However during this time period, I feel that I have always had my cigarette smoking under control.  What it usually came down to was money.  Cigarettes are not cheap.  If I didnt have the money, I simply would not purchase any cigarettes that week.  On average I would smoke about 5 a day, and I am still not much over that amount.  Smoking is something that I enjoy during the summer months.  In summation, I am far from a hopeless addict.  I am merely acknowledging the fact that cigarettes are now a part of my life.  Hipocrazy maybe, but not hypocracy.  '),(12324,3587934,'09,July,2004','2004-07-09 00:00:00','   After years and years of denial, I\'m finally throwing in the towel.  I am addicted to cigarette smoking, and I really enjoy it.  For those that do not know me very well, this is a huge step for me.  I\'ve been trying to become addicted for years now, without much success, but now things are much different.  Cigarettes are by far not a new thing for me.  I drew my first breath back in my freshman year of college, when smoking was really fun.  It made me all dizzy, and contrary to many, alive.  Yes, cigarettes make me feel alive even though the health problems that cigarette smoking causes are undeniable.  The whole idea of making a conscience decision to partake in an activity that is proven to be deadly is totally invigorating.  I like to compare it to base jumping.  Base jumpers strap parachutes to their backs and jump off of tall things and float to the ground.  This is one of the most dangerous sports in the world, and I\'m sure the adreniline rush must be insane, while at the same time addictive.  Each time I light up a cigarette, take a deep drag, and feel that familiar tightening in your chest, and then the long pleasurable exhale, I think of the base jumpers, except I\'m seeking a nicotine rush instead of an adrenline rush.  Both smokers and base jumpers know that if they continue their habits long enough, they will probably die from them.  However, that wont stop me.  Perhaps I am justifying a dirty habit, but its taking risks in life that make us live.  I\'ll probably just end up quitting during the winter like I usually do because I hate smoking when its cold out, its just not very pleasurable.  But for now, pass me the lighter!  '),(12325,3587934,'07,July,2004','2004-07-07 00:00:00','   urlLink    All Hail Maria Sharapova!  This long legged tanned Russian Vixen is the new Wimbledon champion.  Not only did she crush (and I say crush) Serena in straight sets, she is fucking gorgeous.  She is poised to become the new tennis megastar, and I think I love her.  Hail Hail!&nbsp; urlLink  '),(12326,3587934,'06,July,2004','2004-07-06 00:00:00','   The way my desk is set up at work is quite advantageous.  I have walls on two sides of me and there is a cubicle wall on the other two sides.  There is only a small opening to my desk area.  I call this the \"door\".  The entire area is really private.  Nobody passing by can see my computer screen, which allows me to play hearts, cribbage and partake in any other risky activities at my leisure.  I listen to music freely and liberally.  Strategically speaking I have one of the best seats in the entire office.   However, a certain grotesquely raffish â€œPerson Xâ€ often disrupts my serene little corner.  I can\'t tell you how many times I\'ve been busily working or even worse, playing hearts, when all of sudden I hear booming \"so what are you doing back here\".  I fucking hate you X!  One of these days I am going to wheel around in my sleek n stylish black Aeron chair and stab X in the neck with a freshly sharpened pencil.    Most people consciously announce their approach with a paper shuffle, cough or even a soft knock at my \"door\".  X on the other hand displays absolutely no office decorum.  Sometimes X will even stand behind until I notice somebody is there.  I fucking hate that!    With that said, todayâ€™s french phrase of the day is:         On est tous dans le mÃªme bateau  '),(12327,3587934,'06,August,2004','2004-08-06 00:00:00','   Ok, its about time I posted to my blog, it has been entirely too long. I cant fucking wait to go to Maine in a few weeks, so I thought I would start a discussion thread regarding the trip. Seth had mentioned that there is white water rafting on the Kennebec river. I thought this was a ruling suggestion. I did some research and found a few companies that provide all of the equipment, transportation, etc. Follow the below links to check this out:   urlLink  NC rivers   This place offers trips on the Dead River, Kennebec River and Penobscot River. Looks cool, and rates are posted for comparison.   urlLink  Magic Falls   Sounds gay, but looks ok. Prices aren\'t posted for this one, but it seems like they do host a bbq after the trip which looks cool.   urlLink  Northern Outdoors   This one looks to be the most reputable. The website is good and informative, the equipment looks state-of-the-art, but it is also the most expensive. Damn, there\'s always a catch to everything....   urlLink  Crab Apple   I like this one alot. Looks like they have their shit together, and its reasonably priced (competatively anyways)   urlLink  Three Rivers   This one looks EXTREME!! (in a cheesy, Cancun type way) but none the less is priced fairly and seems credible. Oh yeah, and watch your volume when visiting this site, some idiot talks loudly whenever you click on something.   urlLink  NEOC   I like this one alot too.   There are plenty of companies that offer trips that we could take. I think it would be a good idea if we all agreed on this before we go. We must:    A. Agree if we all actually want to do this   B. Choose a company that provides all the equipment, transportation etc, and possibly a bbq afterwards.   C. Make the reservation before we leave for Maine. We\'ll be in the middle of friggen no where without internet, etc.    Everybody please post comments at the bottom of this message. Once there is a discussion thread going, we can easily get some kind of consensus regarding the trip.  RULE ON  '),(12328,4027719,'31,July,2004','2004-07-31 00:00:00','   RE: urlLink unterseeboot130  \'s posting on 27 July 04:  crying is definitely healthy for us. I cry a lot, too much sometimes. My mom always tell me crying is only for people who are weak. She hates it when I cry. so I tried to control myself all through teenage years not to cry. but i think she is wrong, crying comes from a sensitive heart, sensitive to sin, sensitive to pain, sensitive to other\'s needs. I just had a good cry yesteraday, from my cry, I realized my selfishness and the pain I\'ve caused for someone. My heart is still filled with guilt, regret and pain, but after the cry, after I said sorry for so many times, I know everything will be ok.  '),(12329,4027719,'30,July,2004','2004-07-30 00:00:00','   Have you ever been to the beach alone late at night? It was quiet an experience for me. I was tired of crying at home last night and was in desparate need for some fresh air. I thought I\'d go to the beach where we had one of our fellowship activity nights, but I got lost. didn\'t want to call for directions, \'cause I didn\'t feel like talking to anyone. So by the time I got to a beach on PCH, it was already after 10pm.  You have no idea how many people come to the beach at that hour. There was a long line of cars parked along side of the road. Couples were out by their cars drinking wine and whispering, peolple who came alone just sat in their cars quietly looking out to the ocean. Amazingly, nobody was down at the beach. Maybe it\'s too dark or something, which is perfect for me. I slowly walked down to the beach and found a rock to sit on.  Looking aimlessly at the ocean, I realized I don\'t really know why I am there.  What is it about the beach that attracts me and so many others to come this late at night? I guess I was looking for a place to forget, to get rid of the guilt, to find clearity, to sort things out, to find courage and strength from above. But I ended up sitting there having no thoughts in my head, just enjoying the tranquility of the night. Perhaps that\'s why some people come here, not to think, but to forget thinking. Somehow, the stillness of the night, the vastness of the ocean, the crushing sound of the wave, the mysterious yet clear sky, the reflection of the moon dancing lively on the water, the sparsely spread lights from nearby houses and cars all created a perfect harmony for us to calm our agitated soul, to make us unable to think. This is when your feelings can become one with the nature. All the thoughts, rationalities become so unimportant and minuscule compare to the greatness of God\'s creation.  It got a little too cold after a hour of mindlessness, so I migrated into my car. Put on some light music, it\'s another Yi(4)Jing(4) (ha, I really need to work on my English vocab). It was a very comfortable feeling, like watching a drive-in movie. A movie about light in the darkness, about sources of sound together creating heart warming serenity, about variations seen through stillness. Everthing just seems so perfect.  I didn\'t want to leave, but my eyes were starting to warn me around midnight:\"I have to get up before 8am tomorrow\". alright, alright, I\'ll give you a break, i\'ve already tired you out today.  '),(12330,4027719,'30,July,2004','2004-07-30 00:00:00','   We went to BJ\'s for lunch again today, twice in three days, this time is for our ladies\' lunch. Don\'t you notice that ladies have more needs for social interaction than men? You don\'t see men gather together just to chitchat about nothing. We have a woman\'s affinity group at our company, but I\'ve never heard of a man\'s affinity group. Just an observation.  My old officemate and I carpooled there. It\'s funny how we talked more today than we\'ve talked in the six months we shared office together. I used to be so intimidated by her, I don\'t konw why. But after we moved to our own cube, I actually missed her and want to get to know her better. And she seems to be more friendly and open to talk to me now as well. It\'s really weird. I think it\'s the same with my parents and I. when we lived together, I couldn\'t stand to be around them. I literally escaped here to LA just to get away. But now, even though there are still disagreements and a lack of understanding, I find myself care and worry about them a lot. I am not good at expressing my feelings, but I think they can still feel my change. I don\'t know if it\'s just me, i have to be physically apart from some people to appreciate them. huh.... intersting...  oh, I also had an interesting conversation with one of the QAs. She does horse dressage, also called \"horse ballet\". She can make horse dance. She has already been trained for nine years to compete in the 2008 Olympics. I didn\'t even know there is a sport like that. She is so enthusiastic about it too. She has to fly her horse to Beijing, which cost about $7000. I really admire people like that, they\'d pursue their passion at whatever cost. Not because of any worldly values, but simply because they enjoy it so much. My old Pastor\'s son is like that. He likes playing basketball so much that he would work for a little while to make enough money and he\'ll quit to go play basketball. And when he runs out of money, he finds another job and quit again. I know it sounds stupid, his mom worries about him a lot, but I think it\'s great that he knows what he wants to do and has the courage to do it even though it\'s against all conventional thinkings.  Anyway, we had dessert at the end. It\'s called \"bazuka? bazuki?\". It\' s a huge thin cookie topped with ice cream. The combination of hot cookie and cold ice cream brings out an unique texture in the cookie. It\'s very good. The lunch lasted for about 2 hours and now there is another party going on. I guess I am not going to get any work done today.  '),(12331,4027719,'28,July,2004','2004-07-28 00:00:00','   This is R\'s last week working here at Amgen.  We went to BJ\'s for his farewell lunch.&nbsp;I sat there watch him talking, a sudden sadness came over me as I realized that this could be the last time I get to see or talk to him. It\'s such a strange feeling. I didn\'t know I still&nbsp;care that much.&nbsp;   I know he is&nbsp;just a passanger on my life train, and this is his stop to get off. I see him now only as a&nbsp;friend&nbsp;I&nbsp;truely respect and admire, nothing more. I am glad I got to know him in the short time we worked together and appreciate all the help he offered. My feelings for him went through so many different stages in the last six months or so... but it\'s all good. It worked out for the best. prayer really helps if you are not sure, emotion is&nbsp;a tricky thing, really.  Maybe I shouldn\'t share too much on this, I just want to keep a memory of him and say:\"Thank you for everything and best of luck to you\". Like the AC Sandy said:\"He will be successful whereever he goes\", yes, I am sure he will. And he will surely be missed by everyone here.  '),(12332,4027719,'27,July,2004','2004-07-27 00:00:00','   Sometimes, God makes&nbsp;us&nbsp;wait a long time to answer&nbsp;our questions/prayers. Not this time.&nbsp; You probably read my post about happiness. I was actually a little depressed after posting that, I guess I was confused. The more I think about that topic, the more I feel my life is a mess.&nbsp; I went to bed early last night, around 10pm. But couldn\'t sleep, so I picked up&nbsp;the book Livia had&nbsp;left outside of my door the night before. The book is called \"The Purpose Driven Life\".&nbsp; I know most of you had already read that book, and I heard a lot of good things about it. But I never had&nbsp;the chance to read it.   First chapter, first sentence:\"  It\'s not about you  \". Wow, such strong words. Read on:\" The purpose of your life is far greater than your own personal fulfillment, your peace of mind, or even your happiness...... The search for the purpose of life has puzzled people for thousands of years. That\'s because we typically begin at the wrong starting point - ourselves. We asked self-centered questions like What do I want to be? What should I do with my life? What are my goals, my ambitions, my dreams for my future?\". &nbsp;   There it is, there&nbsp;lies the simple yet&nbsp;definitive answer for my problem. I realized how narrow and selfish my thoughts are. I\'ve been thinking too much about ME. I\'ve made my happiness the goal of my life. If I don\'t know what hapiness is, I have no goal, no purpose. A life without purpose leads to no motivation, no energy, no hope. No wonder I am confused and depressed. Our purpose can only be revealed when we turn to the creator Himself. He has his purpose in&nbsp;each of our&nbsp;lifes when he created us. Our purpose is not self-defined. We can\'t make up a purpose/goal (like I want to be a&nbsp;lead BA in five years) and thinking that if we reach that goal, we\'ll be happy. We can\'t really define happiness either. Happiness is not a goal, it\'s only an illusion.&nbsp; I feel happiness is like a lot of beautiful bubbles in the air,&nbsp;they are&nbsp;only beautiful beacuse of the reflection of&nbsp;light from outside and because it\'s hard to grab ahold of. Each bubble can be each one of&nbsp;our \'dreams\'. But when&nbsp;we actully get close to it, it looses its beauty and we\'ll start to pursue another bubble.&nbsp;The bubble itself is empty, it cannot fulfill&nbsp;us.   It\'s not that I didn\'t know we should seek for God\'s will in us before. I often ask God to lead&nbsp;me to walk on the path He has designed for me. In my knowledge, in my conscious, I know only God\'s will is the best for me. But I think subconsciously, a big part of me is still very self-oriented. I want to be happy, I want the things I like, I want God to help me get what I want. THIS IS VERY WRONG. Another quote from the book:\" Many people try to use God for their own self-actualization, but that is a reversal of nature and is doomed to failure.  You were made for God, not vice versa, and life is about letting God use you for his purpose, not your using him for your own purpose .\"&nbsp;  We shouldn\'t pursue happiness, but we&nbsp;will find true happiness when we find God\'s purpose in us and when we do our best to fulfill that purpose with God\'s help.   I woke up at 7AM today. \"yeah!\"&nbsp;Nothing has changed in my life. But I did feel an excitement in expectation to find more and more God\'s purpose in me each day. I believe God puts me where I am at each moment for a reason. And He has His way to mold me to be used by Him. I need to learn to put my complete trust in Him and to be patient and flexible(like a soft clay). I\'ll try to stop thinking too hard trying to make sense of things myself, and I really need to learn to let go of things that don\'t belong to me.   Today is a beautiful day, a new day, a day filled with hope. Thank you God, thank you Livia.  haha, this was meant to be short, I really need to control myself. it\'s like I am writing essays.  '),(12333,4027719,'26,July,2004','2004-07-26 00:00:00','   Have you ever paid attention to those Latino (usually) workers&nbsp;who mow the lawn, trim the bushes, and spray fertilizers on the side of the road? I&nbsp;see them almost everyday when I go on one of my walks on Hillcrest. &nbsp;Everytime I pass by, they would give me&nbsp;this bright,&nbsp;happy smile and say:\"Hello\".&nbsp; It\'s not one of those phony office&nbsp;smiles, I can feel the sincerity and purity in that smile.&nbsp;&nbsp;Sometimes, I wonder why they are so happy. They have to work under the cruelty of the sun most of the day and receive minimum pay. But as I watch them help each other, play, joke around with each other,&nbsp;I can\'t help but feel a sense of envy towards them. It\'s hard to explain, I don\'t know why.   When I got back to my cube today, the BA/PM of the project I am working on walked in to&nbsp;chat. Once again, she complained about how unhappy she is at work and how unbearable those dirty office politics are.&nbsp;That\'s what she talks about everytime I see her. It\'s so sad.&nbsp;You work so hard to get&nbsp;an advanced degree in order to spend more than 1/3 of your day feeling unhappy and conflicted.&nbsp; I am sure she is not the only one around the office feeling that way. They are not happy, but they all continue to do the job, for what? money? prestige? or perhaps they just don\'t know what else can make them happier.   What about me? Am I happy? I asked myself that. Do I wake up excited every morning to face my day? No..... but why? I have a very nice working enviornment, flexible hours, a understanding boss, very&nbsp;kind, almost protective&nbsp;coworkers who treat me like their daughter/grand-daughter. I have nothing to complain about my job. But it\'s lacking something. Again, I don\'t know what.   I remember one summer in high school when I worked as a waitress, I had this excitement everyday before I go to work. I loved the longing glance of the customers at the kitchen door waiting for me to bring their food; I loved the \"I can\'t wait to try this new dish\" expression when I put the food down on the table; I loved the satisfying smile people have after they finish eating; I loved to say:\"Is everything alright, anything else I can do for you?\"; And I loved the \"Thank you, the dinner was wonderful\" as they walked out of the door; And of course, I loved to see how much tip they left, not only for the money, but&nbsp;as an indication of how my service was. Maybe&nbsp;this is&nbsp;very self-absorbed, but I felt I made people happy with that job, so I&nbsp;was happy.&nbsp;Everyday&nbsp;can start with this desire to make more people happy. Of course, there are difficult customers, and the bike ride home late at night was quiet scary, but it doesn\'t compare to the joy I get from everything else.   I keep thinking about something priskalivia wrote on her blog. something like \"it\'s not about&nbsp;doing what you like, but like what you are doing\". I liked that phrase a lot and I tried to practice to like what I do now. But more and more, I find liking something is a feeling. It\'s not logical, you can\'t force yourself to like doing something you really don\'t have the desire to do. ok, it\'s like that with people too, right?&nbsp;Don\'t you find that you like&nbsp;one person over another for no apparent reason, sometimes even against all reasons? You feel distant and disconnected to some other people, again, for no logical explainations. well, at least that\'s how I am. Is it a conscious choice? No..... i don\'t think so.   ok, ok, I think I am going too far. I am coming back to&nbsp;try to find out what make a person happy. Yesteraday in&nbsp;the sermon, Pastor Micheal said \"Happiess is to follow Christ\". I think this is the ultimate&nbsp;answer for us Christians. But&nbsp;this phrase is too high level and abstract for my little brain to comprehend. what does it mean and how? Maybe what I have now is the best for me ( I know this job is from God), I just need to change my attitude? Maybe He is trying to teach me things at this job, and I am just too absorbed in my own world to see it. uh(sign), I don\'t know. from time to time, i just want to change my career completely, but.... oh, never mind. I am too tired to think now, and very hungry. i can keep rambling on forever with no conclusion. But hey, this is the place to do it, to dump thoughts out, who cares if it makes sense or not.   If you are still awake trying to read this, I applaud you. And thanks for your patience. Now you know what I meant when I say I don\'t know when to stop thinking and writing.&nbsp; Alright, I am going home now.  '),(12334,4027719,'25,July,2004','2004-07-25 00:00:00','   I was sitting at church going through&nbsp;things in my bag&nbsp;in attempt&nbsp;to find tissues, instead, I&nbsp;found my old watch sitting alone in a dark corner. It suddenly struck me, I haven\'t worn that watch for so long.&nbsp; I remember&nbsp;not long ago in college&nbsp;when I&nbsp;simply cannot live without&nbsp;that watch, every minute was so precious to me. I had to plan&nbsp;each day well to fully utilize the 24 hr given to me.&nbsp;&nbsp;I had to know exactly how much time I have for what I am doing and when to start the next activity. I was always very punctual, and always in a hurry.   But look at my life now. I don\'t need a watch. Everyday is the same routine: wake up, go to work, lunch, get off work. The question of what time it is&nbsp;becomes quiet irrelevent in my daily life. What happened? What\'s so different from school and work? I don\'t like this lazy, unenergetic self. It\'s really tiring to be lazy, you know what I mean? I feel I don\'t have motivations to do anything, I don\'t really care any more for the things I used to love.&nbsp;I know there are so much I can do, like reading the bible, drawing, get more connected with family and friends...&nbsp;These are all the things I wished I had more time for in college, but now that I have the time, I am not taking the&nbsp;&nbsp;advantage of it.   I think God must be pretty mad at me now. Everynight before I go to bed, I tell Him I\'d wake up early the next morning to spend more quiet time with Him in prayer and reading His words. But every morning, I wake up telling my alarm clock, five more minutes, ten more minutes until the last minute...... hummmm... I wasn\'t like that, my dad told me when I was little, I used to wake up one hour earlier than I needed to just sitting in my bed waiting for him to call me up.   well, each day is another new day. I pray tomorrow is a new beginning for me to start early with full energy and a grateful heart. I am wearing my watch again now. I don\'t expect it to change me right away, but hopefully it\'ll act as a constant reminder that I need to change the way I am now.... to use my time more wisely in accordance with God\'s will in me.  &nbsp;  &nbsp;  '),(12335,4027719,'24,July,2004','2004-07-24 00:00:00','   I can\'t believe I am finally starting my own blog. I have been resisting this blogging idea in fear of exposing too much of this inner self, not only to the public, but also to myself. There are things&nbsp;in me that even I am afraid to find out. So as you try to walk in the maze of my mind, please use extreme caution,&nbsp;put on&nbsp;protective gears, and most importantly, know when to stop reading. \'cause I have this incurable disease of not knowing when to stop talking, when to stop thinking, when to stop feeling, when to stop writing.....   This is going to be quiet an adventure for me.&nbsp;An adventure for me to&nbsp;discover who I am. I just found out recently that writing is an incredible tool to&nbsp;dig out the deeper side of oneself.&nbsp;&nbsp;That\'s probably why I decided to&nbsp;embark upon this self-discovering journey.&nbsp; It\'s not meant for the pleasure of the readers, only a plain, simple and truthful GAOBAI (sorry, i don\'t know the English word for this).  '),(12336,4027719,'18,August,2004','2004-08-18 00:00:00','   I just had One-On-One with my manager. He came to my cube this time to check out my code. I showed him what I did with the RNALabel page and started to tell him about my confusion and how I think it\'s quiet impossible to accomplish what was proprosed. His immediate reaction was:\"Nothing is impossible. Our job is to make what\'s impossible possible.\" He then asked me patiently what the problem was and started to help me seek a better solution.   I guess I was too into my own way of thinking, I thought going through each probe is the only reasonable way to check and update everything. But everytime I need to check for counter probe in a hyb, my algorithm becomes extremely complicated. He suggested to go through each Hyb and find pairs of probes. \"Ding! Ding!\" Why didn\'t I thikn of that? \"I don\'t know if it\'ll work. You might run into another dead end. And this will make things more complex. It means you need to build very sophisticated data structures. and the checking process might need a major refactoring. But don\'t try to change the requirements. Conquer them instead!\" He said.  His \"Conquer them\" keeps ringing in my head.  \"Programming is all about creativity. \" A friend once said to me. \"No, it\'s all about logic.\" I rebutted back. But after a while, I know he is right. Logic only gets me to run into a brick wall(When I think too deep in one direction). It\'s creativity that turns me around and try to find another way out. It might not work, but that\'s the fun of programming. To be creative, to conquer a challenge. I haven\'t felt this excited to program in a long time.   Looks like a busy day for me. Back to work......  '),(12337,4027719,'14,August,2004','2004-08-14 00:00:00','   Such a teary night.... I watched \'The Miracle Box\' and \'Tuesdays with Morrie\'. Both movies are about love and death that brings out life. I definitely recommend you watch these two movies if you get a chance. Especially \'Tuesdays with Morrie\', I have to watch it again. After that, \'Good Will Hunting\' has to go down one spot.   Try to jot down some notes so I don\'t forget.... -Life is not about longevity, what truly matters is the impact we have on others for each day that\'s given to us. \"We must love each other or die\" What does it mean? What if we never learn that lesson?  -Some people never stop their busi-ness and ask:\"Why am I doing this?\"  -\"Love is the only rational act\"  -What are your fears? fear of death? fear of crying? fear of silence? fear of being afraid? fear of being number two? fear of commitment? fear of being touched? fear of receiving love? fear of losing those we love whole-heartedly so we shy away from love altogether? fear of being touchy-feely?...We are constantly trying to run away instead of facing our fears. Only after Mitch realizes his own fears, he is able to battle and defeat them.  -Even the strongest person, the one you think has everything all figured out, has fear. He has his weak moments, he too needs a hug, a shoulder to cry on, a caring look, a encouraging smile....  -Don\'t take anyone for granted. We never know when we are going to lose them because of our selfishness and fear.  -\"If you can have 24 hours of health, what would you do?\" Mitch asked. Morrie\'s answer is so simple: eat good meals, be with loved ones, enjoy nature, do what you feel passionate about (in his case, dancing). \"That\'s it? it\'s so easy\" I believe was Mitch\'s reaction. Yes, happiness is very easy to find if we treasure everything we have. What would I do if I only had 24 hours? Would I still be a greedy sleeper? Would I feel resentful for something someone did long time ago? Would I watch meaningless TV all day? Would I be hesitant to make that phone call because of stupid \"what if\'s\"? Would I..... How about live each day like it\'s my last day.   -Morrie\'s comment on dependency:\"I depend on others for everything.... culture says I should be ashamed of that..... When we are infants, we need others to survive. When we die, we need others to survive. But here is the secret, in between, we need others even more.\" No shame in needing and depending on others.....   There are a lot more great lessons that touched me deeply. But lessons are dead, I think what I\'ll remeber the most is Morrie as a person. So full of life, even near death, so real and so true, so discerning that he sees right through a person, so soft that you can\'t help but love him.   Special thanks to U-130 for lending us thoes great selections.  '),(12338,4027719,'11,August,2004','2004-08-11 00:00:00','   è©¦ è©¦ å¯« ä¸­ æ–‡  '),(12339,4027719,'09,August,2004','2004-08-09 00:00:00','   Seeing one of my development areas is to gain more domain knowledge from my review, my manager pulled out a book from his bookshelf - \"The Cartoon guide to Genetics\" and said:\"This is your assignment for the day\". It\'s such a fun and informative book. Now I think all introductory science book should be written in cartoon format intermixed with idiotic jokes.   There is also \"The Cartoon guide to the Computer\" by the same author. see if I can find it, should be interesting too....  '),(12340,4027719,'09,August,2004','2004-08-09 00:00:00','   As I was driving back from work today, an alarm went off in my head. It\'s not one of those small alarm clock type which you can snooze it off, it\'s the fire-truck going 70 mi/hour to put out fire somewhere type. I was thinking about something that has bothered me for the whole last week. I just suddenly realized I was so frustrated with the problem because I didn\'t respect a friend\'s feeling, her choice of living.  Just because she confide her emotion and confusion to me, I thought I need to help her to sort things out. The more I think about it, the more I feel her life is such a mess, the more I want her to live the way I thought would be the best. so? so... I am turning into my parents! you see? I screamed in my head. That is exactly how my parents make me feel, only their way is the best for me, there is no respect as to how I want to live my life. And now I know, only if you really care about that person, you would bother to try everything you can to make that person \'happy\'. But who am I to know what makes another person happy? The only way to really care for and love a person is to try to understand that person. Even though I\'ve known her for so many years, I don\'t think I\'ve really made an effort to get to know her. I\'ve always self-righteously thought I was a good friend, I did everything I could to help her whenever she needed. But is it out of love? Or is it just to make myself feel good? I started to question my motives for a lot of the things I did.   I called another friend (call her B)yesteraday to ask her advice to help that friend(call her A). I found myself complaining more about how I don\'t approve A\'s life style than really trying to help A. B just said, \"sometimes, all we can do is listen and pray for the person. God has his way, his timing in each of our lives. We don\'t know what God\'s plan is for A. Just be a friend and let her know that you are always there when she needs you. Sometimes, doing nothing is the best thing to do.\" I was so busy yesteraday, I didn\'t have time to digest what she said. But now, everything starts to link together. I feel pressured and burdened by my parents\' love because I feel they don\'t understand me. But isn\'t that how people around me would feel if I try to \'love\' them my own way? What I think would make them happy might actually bring them pressure and make them feel uncomfortable. I think I did a lot of that recently. sorry... I didn\'t respect others enough, even though my intentions are generally good. ok, next time when I try to do somehting \'nice\' for someone, I\'ll try to think about what the person\'s real needs are rather than what my own plans are. I am very sorry.... for those of you....... I know how aweful it feels, crushed by too much \'care\'. I don\'t want to turn into my parents, at least with that respect, but I think I now know how they feel. It\'s only because they care too much.... yea, I should really communicate with them more often, who knows, maybe one day, when God\'s timing comes, we\'ll learn to love each other the \'right\' way.......  '),(12341,4027719,'03,August,2004','2004-08-03 00:00:00','   Hello, Xiaochen,    Attached is the MAP review form....   You have to fill it out tonight, ok? no sleep before you do that. you have been slacking off too much.... You know why you have a headache now? because you didn\'t do any real work the whole day today and you are stressing out now. remember, you have to finish the new probe page by tomorrow and Scott will do mid-year review with you Wed morning. You better get your act up.... quit day dreaming.     myself  '),(12342,4027719,'03,August,2004','2004-08-03 00:00:00','   oh, man, I feel I am back in school again. My record in college is three days without any sleep and only crackers to fill my stomach. That was the horrible compiler class project.... I remember everyone came to school the next week with red eyes. yeeee.... I had dinner with two college friends last week, I feel college is so long ago now, we couldn\'t even remember our professors\' names. I am getting old, my eyes can\'t stay open, and it\'s not even midnight yet. maybe I should just go home and start fresh tomorrow. I think I am sleep talking now.....who is making all those noise down the hall. somebody else is here too? ok, at least i have company.....  '),(12343,4027719,'03,August,2004','2004-08-03 00:00:00','   \"Hi, Y, I have a question on saving labeled RNAs.\" I walked into the project manager\'s office thinking that was the last bug I need to fix before Thur\'s presentation. \"oh, thank god you haven\'t left yet, you are just the person I am looking for.\" She lifted up her head as she was sorting out a mountain of paper in front of her, \"I just got off the phone with the clients in Seattle, they finally reached a concensus as to how they want the RNA labeling page to behave.\" She then proceeded to explain what changes I need to make. \"oh, man, that\'s a lot of changes in the backend. The logic is completely different from how I implemented it.\" As I was listening, complaints started to flood my head,\"you are the project manager, you are supposed to give me the final requirements before I start coding. It\'s not my fault that you didn\'t do your job. I\'ll never finish if you keep changing the requirements. you lay this on me the last minute, it\'s not fair\".  I think my face must looked very stressed at that moment, she paused a little and said:\"Do what you can, if you can\'t finish, it\'s ok. We can schedule another demo\". \"no, I can do it.\" I said unwillingly.   She looked at me like a mother looking at her child with an unbearable task to complete and started to tell me that she found out earlier today one of the scientists(Sumedha) we worked with was diagnosed with Leukemia. I was so shocked by the news \"What? When? How? Why?\". I just saw him a couple of weeks ago when I was taking a walk, he drove by and asked if I needed a ride. I\'ll never forget that friendly and energetic smile. He gave me so much encouragement when I first started working on his project. Even when I screw up, he was always very understanding. He is very humble and easygoing despite of his high position in the company. We all loved him. I froze there in the office, didn\'t know how to react to that fact.\"You know, H(one of our coworkers)\'s wife also died from Leukemia two years ago.... Life is unpredictable.\" She said, \"you don\'t know what\'s gonna happen tomorrow. but how you want to live today is your choice.\"   She also told me a story about a guy from her church. He fell down the stairs and injured his brain nerves. The doctor announced he\'ll be a vegetable the rest of his life. However, God miraculously healed him and now, he is living everyday for the Lord. Everything he has, he attribues to God, especially his health. I can imagine how the change took place in his life. Sometimes, God will use extreme measures to turn our hearts back to Him. \"What would it take for you to live your life for me?\" This question appeared in my mind after I walked out of her office. \"I don\'t know, God. You know me better than myself, you know I have the heart, but I don\'t know how. Teach me at whatever cost.\" This is my lame yet truthful answer.   I am sitting in my cube trying to work now. I feel so silly that I got upset over such a small thing earlier. just because I have to work late for couple of nights? many people who are out of job would do anything to have the chance to work late. I should be thankful for my job, and for everything else I have. I don\'t know what will happen tomorrow, but I can surely make today count while I still can. And I will pray for Sumedha, pray for healing, pray for courage, pray for wisdom, pray for his family as well. I hope to see you back, Sumedha!  '),(12344,3434961,'24,May,2004','2004-05-24 00:00:00','    Ah,yes.  Well I for one feel much better about my blog.  I spent some time crawling around other peoples blogs and ya know something?  People are boring!  So why should I feel bad for being boring... I am only human right?  Well if anyone every reads this, I\'m telling you.. you gotta read Walden.  It is so delicious.  I\'ve worn the pages of that book out in the last 5 years, but I can still crack it open and sink right in.  I know some people who don\'t read.  I mean, like ever.  Never one book all the way through.  It makes me what to scream at em\'.  Well what\'s it to me anyway.  I\'ve got this friend Stacy, and she is the coolest girl ever.  She was the first girl I was friends with in a platonic way.  I was driving a truck over the road and lost touch a bit, but she called me today and I forgot how much I really miss her.  She has a serious boyfriend now and that could suck, but he is totally cool too.  His name is Andy and he is way cooler than most anyone I know.  Together they are some super couple of coolness.  Well.... ah.. yeah.. cool   '),(12345,3434961,'23,May,2004','2004-05-23 00:00:00','    Good morning my happy little blog.  Today is a day of triumph!  Having been unemployed for two weeks this very day the growing frustration with my goal of getting out of the rent trap and buying a few acres to spread out on getting further away instead of nearer has forced me in to a frightful position.... I am updating my resume.  Yes, folks time to get a job.  Now I have mentioned my lack of marketable skills.  I am a college drop out with experience in retail, restaurants (cook), general construction labor, and driving big trucks.  Ah, how I hate all my previous jobs.  A necessary evil I suppose.  Well, in other news I have decided that perhaps the meaning to this game show of life is to attain perfection.  I mean to do each and every act through  out your day to the best of your ability, the best you can.  To care with your whole heart about what you do, and to not do what you do not care about.  Caring is key.  The thing is I am no perfectionist and am constantly doing things I care very little for.  Ah well, maybe I\'ll just go back to bed... I care about that.  You know I haven\'t really wrote anything except for my dream log in years.  There\'s been no letters, for at least a year, no poems in perhaps 2, no story\'s in maybe 4.  Not that I have other been a prolific writer or have any hopes at being a published work or something, but it seems important to keep on writing for some reason.  Doesn\'t it?  Well, now I got this blogg I guess.  An online journal that people can read and comment on.  You know that\'s kinda weird, but the anonymity makes it OK.  WHO ARE YOU!  Heh heh.  Well, I guess once I break the crust loose this all won\'t feel so forced.  Well, I have made of list of things to accomplish today ( besides updating my crappy resume )so onward I go into the loving embrace of responsibility.... ah so warm and tingley   '),(12346,3434961,'23,May,2004','2004-05-23 00:00:00','    Lets, see where to begin.  Ah, yes at the beginning, I am a 24 year old, un (and frequently under) employed male with few marketable skills.  I enjoy old classy cars, old rusted jeeps, music, art, organic gardening/homesteaders, building with wood and stone, philosophy, cartoons and long pillow fights on the beach.  I am a confirmed book worm brainiac resulting in few meaningfully relationships.  I type slowly, mis-spell words frequently and generally have nothing of interest to say.  I have never \"blogged\" before but, just like drugs I have decided to give it a try.  Today I attended my local church, made a peanut butter and jelly (I also forgot to mention my love of the culinary arts), drank half a bottle of moderately priced wine and enjoyed the final Simpsons episode of the season (worst episode ever).  My favorite authors are Thoreau, Emerson, Pirsig, Eliot Coleman, Helen and Scott Nearing, Rob Roy, and Steven King.  I have large ambitions but a much less than required drive to achieve them.  Some day I hope to own (as much as you can own land you pay taxes on) a small homestead where I can garden, read, contemplate, make music, create art, and learn to do all these things in a holy way so to touch god....or something.  I have managed to save perhaps half of the required money to buy land in the area I am looking and am somewhat excited about making a change.  Well, that\'s a big enough run on paragraph so I will close here and check to see what my blogg page looks like, if this new fangled technology works, ect... thanks for useing a moment of time you\'ll never get back to learn a little bit about me, your friendly neiborhood shut-in.   '),(12347,3573678,'15,June,2004','2004-06-15 00:00:00','   I was reading a very informative article on MSDN. Thought it was worth sharing. This talks about writing Language portable T-SQL statements. Would not the idea of building globalization logic in T-SQL Stored Procedures be interesting...at least for me it is...if you are also interested...to have a look at the page...just follow this link http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dnsqldev/html/sqldev_06112004.asp  '),(12348,3573678,'14,June,2004','2004-06-14 00:00:00','   Multiple result sets can be fectched by using 2 queries separated by a semicolon. If multiple result sets are returned, the DataReader provides the NextResult method to iterate through the result sets in order, as shown in the following code example. [In Visual Basic.Net]  Dim myCMD As SqlCommand = New SqlCommand(\"SELECT CategoryID, CategoryName FROM Categories;SELECT EmployeeID, LastName FROM Employees\", nwindConn) nwindConn.Open()  Dim myReader As SqlDataReader = myCMD.ExecuteReader()  Dim fNextResult As Boolean = True Do Until Not fNextResult   Console.WriteLine( myReader.GetName(0) & myReader.GetName(1))    Do While myReader.Read()     Console.WriteLine(myReader.GetInt32(0) & myReader.GetString(1))   Loop    fNextResult = myReader.NextResult() Loop  myReader.Close() nwindConn.Close()  '),(12349,3573678,'14,June,2004','2004-06-14 00:00:00','   The Hypertext Transfer Protocol (HTTP) listener is implemented as a kernel-mode device driver named HTTP.sys. HTTP.sys is part of the networking subsystem of Windows. However, HTTP.sys is utilized as a core component of IIS 6.0. By running HTTP.sys as a kernel-mode component, IIS 6.0 delivers the following two performance enhancements, which are not included when running HTTP listener as a user-mode component: â€¢	By dispatching directly to the correct process from the kernel, requests are served with less context-switching overhead.  â€¢	By enabling the kernel-mode cache, requests for cached returns can be served without switching to user mode.  How HTTP.sys Works When you create a Web site in IIS, the site is registered with HTTP.sys, which then routes Web requests to the user-mode process that is running the Web site. HTTP.sys also sends responses back to the client. Other than retrieving a stored response from its internal cache, HTTP.sys does not process the requests it receives. Therefore, no application-specific code is ever loaded into kernel mode. As a result, application-specific code bugs cannot affect kernel-mode processes or lead to system failures. HTTP.sys provides other services that IIS uses, including the following: â€¢	Managing Transmission Control Protocol (TCP) connections.  â€¢	Routing HTTP requests to the correct request queue.  â€¢	Caching of responses in kernel mode.  â€¢	Performing all text-based logging for the WWW service.  â€¢	Implementing Quality of Service (QoS) functionality, which includes connection limits, connection time-outs, queue-length limits, and bandwidth throttling.   The role of http.sys is to route the HTTP requests to user mode applications. But it doesnâ€™t process the request. It also caches the serverâ€™s response.  '),(12350,3573678,'14,June,2004','2004-06-14 00:00:00','   Given an XML Schema that complies with the XML Schema definition language (XSD) standard, you can generate a strongly typed DataSet using the XSD.exe tool provided with the .NET Framework SDK. The following code shows the syntax for generating a DataSet using this tool. xsd.exe /d /l:CS XSDSchemaFileName.xsd /n:XSDSchema.Namespace In this syntax, the /d directive tells the tool to generate a DataSet, and the /l: tells the tool what language to use (for example, C# or Visual Basic .NET). The optional /n: directive tells the tool to also generate a namespace for the DataSet called XSDSchema.Namespace. The output of the command is XSDSchemaFileName.cs, which can be compiled and used in an ADO.NET application. The generated code can be compiled as a library or a module.  '),(12351,3573678,'14,June,2004','2004-06-14 00:00:00','   The Update method of the DataAdapter is called to resolve changes from a DataSet back to the data source. The Update method, like the Fill method, takes as arguments an instance of a DataSet, and an optional DataTable object or DataTable name. The DataSet instance is the DataSet that contains the changes that have been made, and the DataTable identifies the table from which to retrieve the changes. When you call the Update method, the DataAdapter analyzes the changes that have been made and executes the appropriate command (INSERT, UPDATE, or DELETE). When the DataAdapter encounters a change to a DataRow, it uses the InsertCommand, UpdateCommand, or DeleteCommand to process the change. This allows you to maximize the performance of your ADO.NET application by specifying command syntax at design-time and, where possible, through the use of stored procedures. You must explicitly set the commands before calling Update. If Update is called and the appropriate command does not exist for a particular update (for example, no DeleteCommand for deleted rows), an exception will be thrown. Command parameters can be used to specify input and output values for an SQL statement or stored procedure for each modified row in a DataSet.  If your DataTable maps to or is generated from a single database table, you can take advantage of the CommandBuilder object to automatically generate the DeleteCommand, InsertCommand, and UpdateCommand of the DataAdapter.  The Update method will resolve your changes back to the data source, however other clients may have modified data at the data source since the last time you filled the DataSet. To refresh your DataSet with current data, use the DataAdapter and Fill the DataSet again. New rows will be added to the table, and updated information will be incorporated into existing rows. The Fill method determines whether a new row will be added or an existing row will be updated by examining the primary key values of the rows in the DataSet and the rows returned by the SelectCommand. If the Fill method encounters a primary key value for a row in the DataSet that matches a primary key value from a row in the results returned by the SelectCommand, it updates the existing row with the information from the row returned by the SelectCommand and sets the RowState of the existing row to Unchanged. If a row returned by the SelectCommand has a primary key value that does not match any of the primary key values of the rows in the DataSet, the Fill method adds a new row with a RowState of Unchanged. Note   If the SelectCommand returns the results of an OUTER JOIN, the DataAdapter will not set a PrimaryKey value for the resulting DataTable. You will need to define the PrimaryKey yourself to ensure that duplicate rows are resolved correctly.  '),(12352,3573678,'14,June,2004','2004-06-14 00:00:00','   MetaData: 	Metadata is binary information describing your program that is stored either in a common language runtime portable executable (PE) file or in memory. When you compile your code into a PE file, metadata is inserted into one portion of the file, while your code is converted to Microsoft intermediate language (MSIL) and inserted into another portion of the file. Metadata stores the following information:  Description of the assembly.  	Identity (name, version, culture, public key).  	The types that are exported.  	Other assemblies that this assembly depends on.  	Security permissions needed to run.  Description of types.  	Name, visibility, base class, and interfaces implemented.  	Members (methods, fields, properties, events, nested types).  Attributes.  	Additional descriptive elements that modify types and members.   Assembly Manifest: Every assembly, whether static or dynamic, contains a collection of data that describes how the elements in the assembly relate to each other. The assembly manifest contains this assembly metadata. An assembly manifest contains all the metadata needed to specify the assembly\'s version requirements and security identity, and all metadata needed to define the scope of the assembly and resolve references to resources and classes. The assembly manifest can be stored in either a PE file (an .exe or .dll) with Microsoft intermediate language (MSIL) code or in a standalone PE file that contains only assembly manifest information.  '),(12353,3573678,'14,June,2004','2004-06-14 00:00:00','   Web Services are not supposed to hold state between calls. Web Services address a branch of architecture which is called the Service Oriented Architecture (SOA). By definition, SOA is based on applications exposing services for other applications or users to consume. When we talk about services, we referrer to an atomic unit of consumable service. In a typical scenario, web services are implemented at a use-case level. A service FaÃ§ade layer would be introduced over the business layer (sometimes over the business faÃ§ade). This service faÃ§ade would be exposed as the service layer (as web service calls). Each service exposed (method in your web service) is meant to be atomic â€“ the service should take in all the details that would be required to invoke the all business functionality in the right sequence. And hence you donâ€™t have the concepts of state sharing. You may now ask how transactions can be maintained across web service calls! Web service transactions is not yet supported â€“ several standards are being proposed but itâ€™s actual implementation is far off.  In conclusion,  1.	You should design your web service in such a way that it does not require to share state between calls.   2.	In the worst case, your application should take care of data management by persisting data in the database. State sharing should not be considered.   Technically, speakingâ€¦ Because Web Services uses the stateless HTTP Protocol, Web Services by default does not support state. Yet you can do this by building additional logic in your code.  '),(12354,3573678,'14,June,2004','2004-06-14 00:00:00','   Validate Request attribute is yet another feature that Dot Net Fx 1.1 offers, to build secure enterprise applications. By default for all the ASPX pages that we design, this attribute is set to true. You can find this setting in the machine level configuration file ( /Framework/V /CONFIG/Machine.config file), under the pages element. This would be set to true. Once this is set to true, ASP.NET scans / checks all the inputs posted back from the page / cookies / query strings for potentially malicious input. This featured can be turned off at page level by setting this attribute to false, if the application demands some kinds of inputs be allowed. Dot Net Fx 1.0 does not provide a similar feature. But the IIS URLScan ISAPI Filter does a similar job. Basically both these features help to mitigate the threat of XSS (Cross-site scripting) by rejecting potentially malicious input.  '),(12355,3573678,'14,June,2004','2004-06-14 00:00:00','   I was wondering, how to find out the actual control, that caused a Page submit (postback).  We can actually use the hidden __EVENTTARGET field, in the Pages request object to find out the control, that caused a post back. The following code will help...  private string GetControlThatPostedBack() 		{ 			if(Page.IsPostBack) 			{ 				//try to find the name of the postback control in the hidden __EVENTTARGET field 				string  controlName = Page.Request.Form[\"__EVENTTARGET\"]; 				// if the string is not null, return the control with that name 				if(controlName != null && controlName.Trim().Length > 0) 				{ 					return Page.FindControl(controlName).ID; 				}                                                           			}  			//the above logic does not work for PostBack events fired by standard buttons. 			//Still, we can find out the control, by looping through the Form\'s control collection 			//and retrieving the control that submitted the page 			foreach(string keyName in Page.Request.Form) 			{ 				Control formControl = Page.FindControl(keyName);				 				if(formControl != null) 				{ 					return formControl.ID; 				} 			} 			return \"\"; 		}  '),(12356,3573678,'10,June,2004','2004-06-10 00:00:00','   A cookie can be marked as HTTPOnly , to indicate that the cookie is  \"non-scriptable\" and should not be revealed to the client application, for  security reasons. Within Windows Internet, this means that the cookie cannot  be retrieved through the InternetGetCookie function. If Internet Explorer 6.0 SP1 detects a cookie marked HttpOnly and some  client side script code, such as JavaScript, attempts to read the cookie  (document.cookie, for example), Internet Explorer returns an empty string,  thus preventing the attack by preventing the malicious code in the XSS  attack from sending the data back to a malicious site. Of course, the cookie  is passed to and from the originating server as normal; the browser using  script code just can\'t read it.  Web browsers that do not support the HttpOnly cookie attribute either ignore  the cookie or ignore the attribute, which means it is still subject to XSS  attacks.  The System.Net.Cookie class does not currently support an HttpOnly property.  To add an HttpOnly attribute to the cookie, you need to use an ISAPI filter,  or if you want a managed code solution, add the following code to your applicationÂ’s  Application_EndRequest event handler in Global.asax:  protected void Application_EndRequest(Object sender, EventArgs e) { 	string authCookie = FormsAuthentication.FormsCookieName; 	foreach (string sCookie in Response.Cookies) 	{ 		// Just set the HttpOnly attribute on the Forms authentication cookie 		// Skip this check to set the attribute on all cookies in the collection 		if (sCookie.Equals(authCookie)) 		{ 			// Force HttpOnly to be added to the cookie header 			Response.Cookies[sCookie].Path += \";HttpOnly\"; 		} 	} }  A future version of the .NET Framework is likely to have an HttpOnly  property on the Cookie class.  _________________________________________________________________ Are you a cricket freak? CDs, books, and more goodies!  http://www.msn.co.in/Shopping/CricketShop/ Available at the cricket shop!  '),(12357,3573678,'08,June,2004','2004-06-08 00:00:00','   The Microsoft caching application block is very useful building block for developing enterprise .NET Applications. I found this very useful, and it reduces a great deal of coding effort. However, I found this piece of information worth sharing, so that it will be useful for other developers.  There is a FileDependency class which is used to set the cache dependency on a cached file. The as-is logic of the application block will refresh the cache every time the file is Accessed. It uses the following logic.  While an item is added to the cache, the App block, makes a note of the LastAccessTime of the shared file.  The App block has an in built File System Watcher, which will scan the file after every â€˜nâ€™ seconds, and will determine.  Â·         If the current â€œLastAccessTimeâ€ of the file is greater than the value, that was determined in step 1 above.  Â·         If yes the app block will refresh the cache.    The problem I faced here is, my application will read the file more frequently and will update the file less frequently. For example, if my application accesses the file (to read certain values) for 10 times in a day, it may update the file only once in a day. So going by the given logic of the App-Block the cache was getting refreshed for all the 10 times. The very purpose of caching was not served.   So I decided to change the code (Microsoft PAG allows us to change and recompile the code). I changed the logic of the FileDependency class, so that it will refresh the cache based on LastUpdated time rather on LastAccess time. The modified code is as below.   //=============================================================================== // Microsoft Caching Application Block for .NET // http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dnpag/html/CachingBlock.asp // // FileDependency.cs // This class tracks a file cache dependency. // //=============================================================================== // Copyright (C) 2003 Microsoft Corporation // All rights reserved. // THIS CODE AND INFORMATION IS PROVIDED \"AS IS\" WITHOUT WARRANTY // OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT // LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR // FITNESS FOR A PARTICULAR PURPOSE. //===============================================================================    using System; using System.IO; using System.Runtime.Serialization; using System.Security.Permissions;  // Used for Handling Exceptions using Microsoft.ApplicationBlocks.ExceptionManagement;  namespace Microsoft.ApplicationBlocks.Cache.Expirations {       ///         ///   This class tracks a file cache dependency.       ///         [Serializable]       public class FileDependency : ICacheItemExpiration, ISerializable       {             #region Private members             private string keyValue;                   private string dependencyFileName;             private DateTime lastAccessedTime;             #endregion              #region Constructor             ///               ///   Constructor with one argument.             ///               ///               ///   Indicates the file name of the file             ///               public FileDependency( string fullFileName )             {                   try                   {                                               string path;                         string fileName;                         #region Throwing Argument Exception                         if(Object.Equals(fullFileName, null))                         {                             throw new ArgumentNullException(\"fullFileName\",                             CacheResources.ResourceManager[RES_ExceptionNullFileName\"]);                         }                         if(fullFileName.Length == 0)                         {                               throw(new ArgumentOutOfRangeException(\"fullFileName\",                           CacheResources.ResourceManager\"RES_ExceptionEmptyFileName\"]));                         }                         if(! File.Exists(fullFileName))                         {                               throw(new ArgumentException(CacheResources.                                     ResourceManager[\"RES_ExceptionInvalidFileName\"],                                      \"fullFileName\"));                          }                           #endregion                          // Validate File                         FileInfo fileDependencyInfo = new FileInfo(fullFileName);                         if (!fileDependencyInfo.Exists)                         {                               throw new FileNotFoundException();                         }                         // Get Path from full file name                         path = Path.GetDirectoryName(fullFileName);                         // Get file name from full file name                         fileName = Path.GetFileName(fullFileName);                         dependencyFileName = fullFileName;                         //////////////////////////////////////////////////////                         //Changed HERE                         //////////////////////////////////////////////////////                         lastAccessedTime = File.GetLastWriteTime(fullFileName);                         //////////////////////////////////////////////////////                         ///   END OF CHANGE                         //////////////////////////////////////////////////////                   }                   catch( Exception genException )                   {                         ExceptionManager.Publish( genException );                         throw;                   }             }             #endregion             #region ICacheItemExpiration Implementation              ///               ///   Event to indicate the cache item expiration.             ///               public event ItemDependencyChangeEventHandler Change;                  ///               ///   This method sets the external dependency key.             ///               void ICacheItemExpiration.Key(string keyVal)             {                   try                   {                         #region Throwing Argument Exception                         if( Object.Equals(keyVal, null) )                         {                               throw new ArgumentNullException( \"keyVal\",                                     CacheResources.ResourceManager[                                     \"RES_ExceptionNullKey\" ] );                         }                         if( keyVal.Length == 0)                         {                               throw new ArgumentOutOfRangeException( \"keyVal\",                                     CacheResources.ResourceManager[                                      \"RES_ExceptionEmptyKey\" ] );                         }                         #endregion                         keyValue = keyVal;                   }                   catch(Exception genException)                   {                         ExceptionManager.Publish(genException);                         throw;                   }             }                         ///               ///   Specifies if the item has expired or not.             ///               bool ICacheItemExpiration.HasExpired()             {                   try                   {                        // Compare the Filedependency object\'s last write time                         // value with the last write time value. If they are not                         // equal return true, otherwise false.                          //////////////////////////////////////////////////////                         //Changed HERE                         //////////////////////////////////////////////////////                         if(DateTime.Compare(lastAccessedTime,                                File.GetLastWriteTime(dependencyFileName)) != 0)                         {                               return true;                         }                         //////////////////////////////////////////////////////                         ///   END OF CHANGE                         //////////////////////////////////////////////////////                   }                    catch(Exception genException)                   {                         ExceptionManager.Publish(genException);                   }                   return false;             }                          ///               ///   Notifies that the item was recently used.             ///               void ICacheItemExpiration.Notify()             {}             #endregion                 #region Serialization functions             #region Public method             ///               ///   This method performs the serialization of members of the              ///   current class.             ///               ///               ///   A SerializationInfo object which is deserialized by the              ///   formatter and then passed to current constructor             ///               ///               ///   A StreamingContext that describes the source of the              ///   serialized stream from where the Serialization object              ///   is retrieved             ///               [SecurityPermissionAttribute(                  SecurityAction.Demand, SerializationFormatter = true)]             [SecurityPermissionAttribute(                     SecurityAction.LinkDemand, SerializationFormatter = true)]             public void GetObjectData( SerializationInfo info,                  StreamingContext context )             {                   // Adds the file name and last accessed time                    // into the SerializationInfo,              // where it is associated with the name key                   info.AddValue(\"fileName\", dependencyFileName);                   info.AddValue(\"lastAccessedTime\", lastAccessedTime);             }             #endregion              #region Constructor             ///               ///   This method performs the deserialization of members of the              ///   current class.             ///               ///               ///   A SerializationInfo object which is deserialized by the              ///   formatter and then passed to current constructor             ///               ///               ///   A StreamingContext that describes the source of the              ///   serialized stream from where the Serialization object              ///   is retrieved             ///               protected FileDependency ( SerializationInfo info,                   StreamingContext context )              {                   try                   {                         // Getting the value of file name and                         // last accessed time                         dependencyFileName = info.GetString(\"fileName\");                         lastAccessedTime = Convert.ToDateTime(                               info.GetValue(\"lastAccessedTime\",                                typeof(DateTime)));                                           }                   catch( Exception genException )                   {                         ExceptionManager.Publish( genException );                   }             }             #endregion                   #endregion } }  '),(12358,3573678,'08,June,2004','2004-06-08 00:00:00','   I was writing a web based RSS reader. Working in .NET, I wanted to write my own reader, rather than download one. Actually the reason for going in for a web based reader was, it could be used by even others who do not have .NET Fx installed. Actually, many of you who have already done this, would know how simple it is to write a RSS reader. Even I found it very simple. But when I tried running the app, I was in for some surprises. My office network is behind a firewall. So whenever I tried running the app, â€œ The underlying connection was closed: The remote name could not be resolved â€ error. After some thought a little research I found out a workaround. Thought I would share this.  I added the following lines to the page load eventâ€¦and lo! It started workingâ€¦  IWebProxy proxyObj = new  WebProxy(â€œIP address of the proxy server:portnumberâ€,true); proxyObj.Credentials = new NetworkCredential( , , ); GlobalProxySelection.Select = proxyObj;  Hope this helpsâ€¦  '),(12359,3573678,'08,June,2004','2004-06-08 00:00:00','   Test Posting :-)  '),(12360,4277394,'19,August,2004','2004-08-19 00:00:00','   once i was sitting in the school hall having the first pe lesson,then i was looking around.when the teachers were preparing the stuff for the lesson,i spotted a girl at the oppsite side of my class.I watched her and cared nothing else,by then i realise she was my girl...once i met her at a bus stop,she was with her stead,but i didn\'t know that that was her stead.Thus i sat at a seat behind them,and heard what are they talking about.then as time flies,each time i looked at her,the more i liked her even more.Then,at mid august,i finally got the courage n asked her for stead,i was that surprised that she accpted me.then it lasted for a week,plus a day.we broke up, she didnt want to hurt me,cuz she had her ex in mind.somewhere at the last day of school,i asked her for stead again then we patched up for 3 days and broke up again.because of the sparse communication..when school reopened,we tried to act like nothing happened.but we still can\'t forget that we had the past together,we were like steads,we act like one,and we got even closer to each other.On the 27th july,we patched up.then,kelly exposed our relationship,and she was her best friend,also my friend,jac scolded her for doing this,i don\'t know what happend,and i actually tried to defend kelly.that dy, i sent her home,we were taking bus 105,we sat in the same sit and held hands,we treasured the time we had,and that was only the time we had together alone.but on the 4 of august, we broke up,she told me that she was a bad stead and broke up with me...then finally on the 7th of august she expained why...she again said that she was a bad stead,and her parents n brother knew about our relationship..she wanted to protect me,and she broke up with me..and she does not bare to break up ,she was crying when she told me all this,and there is nothing i could do to help.i can only see her tears rolling down her cheeks.then she took her bag and ran out of the school,i felt helpless,and started wiping..i hated the world to be so unfair to us..all we wanted i a stable relationship and nothing else..but in the end..we still have to go seperate ways...and each time i thought of she saying...i don\'t want to break up but i have no choice...tears will start rolling down my eyes..i did not give up once,i will wait for the right time..i will always keep her in my mind forever...   '),(12361,3576839,'14,June,2004','2004-06-14 00:00:00','   In the Idaho National Guard we are often told to \"be Gumby,\" or to be incredibly flexible.  Well this last weekend we all practiced our Gumby skills once again as we packed our section equipment and our personal equipment to meet us in Fort Bliss.  The packing of the equipment is not a big deal, you pack up your duffel bags, through them in a pile, and pack up your section equipment in a big box.  My section, the Medic section, has very little in the way of equipment.  Especially when compared to other sections, such as the mechanics.  No, the true test of our Gumbiness came when we were told that rather than leaving on 1 July we were leaving on 3 July.  Why is that a problem?  What trauma could that possibly cause by telling us that we have an extra 2 days to spend with our loved ones?  There is, as so often happens, more to the story than that. Over the course of our \"2 weeks in the summer\" training we were told often and LOUDLY that if there was any information to be told, we would be among the first to know.  Essentially the leadership, specifically the commander and first sergeant, said that media was not \"in the know\" and should information become available, we would be told by our chain of command.  The media was not a good source of information as far as our deployment goes.  So, more than a week ago, the newspaper reported that the unit would be leaving on 3 July to head to Fort Bliss to start training.  We all knew that the paper couldn\'t be right, since that information hadn\'t come down to us through the proper channels.  In light of this, when one soldiers mother asked if she should get her ticket for vacation for a date after 3 July, that soldier said \"No, don\'t waste your vacation time just because the newspaper screwed up on the date.\"  Her concern was to be able to see her son off as he flew out.  Fast forward from then to this last weekend, and on Sunday afternoon, after spending the whole weekend working with the very people that knew all of the correct details, they finally decided that it was important enough to let the rest of us in on it.   Being Gumby gets REALLY frustrating REALLY fast.  '),(12362,3576839,'08,June,2004','2004-06-08 00:00:00','   This is, as you can see, the first entry in what I hope will be fairly regular updates to my weblog.  It made sense to me, as I don\'t know how much time I will have once my unit gets activated to send out letters and/or emails to everyone that I would like to.  This will serve as a forum for people to get a glimpse into life as it unfolds throughout the deployment.  As it stands now, on 8 June, 2004, we are all anxiously waiting out Mobilization date of 1 July.  On that date my unit will move from Lewiston, Idaho to Fort Bliss, Texas.  For those of you who are not familiar with Fort Bliss (as I wasn\'t), it is located near El Paso, Texas, which is also very near the New Mexico border.  We will not, however, be spending much time on Fort Bliss itself.  Instead, the training sites are 60-70 miles out, and consists of three separate sites, each with a different focus.  Each site is a 10 day training circuit, with 3 days in between sites for movement to the next site and rest.  From what I understand it will be some long days.  After Fort Bliss, the unit will attend 1 of 2 training bases: Fort Irwin, California or Fort Polk, Louisiana.  I have spent time at both (4 years at Polk and 2 months at Irwin) and each has it\'s own ups and downs.  After spending a month training there, rumor has it that we will receive leave prior to deployment overseas to Iraq.  Spirits in the unit itself have remained pretty high, but then no one can say how that may change once the shit hits the fan, as it were.  I do have every confidence in myself and my comrades, and this training that we will be attending will only enhance that.  My section, the Medics, will even be receiving extra training for our medical skills.  That is very exciting in itself and I am looking forward to brushing up a bit.   As it stands, I do not know when things will change, or when I will have time to enter again post-mobilization.  I don\'t imagine that there will be internet access at Fort Bliss, nor the time to utilize it.  I will make every effort, however, to update this blog frequently.  '),(12363,3576839,'27,July,2004','2004-07-27 00:00:00','   Now that I have time to add more, I will.&nbsp; Fortunately for me I had the presence of mind to start writing down what was going on for most of the time in training for just such an occassion.  3 Jul 04 - We arrived in El Paso.&nbsp; The flight was very nice, the stewards were friendly and courteous to all of us.&nbsp; Prior to getting on the plane was a lot of tears and hugging.&nbsp; I wouldn\'t wish that day on anyone.&nbsp;   4 Jul 04 - The food at the basecamp (Dona Ana, outside of Fort Bliss, TX) is dismal, but on par with Army chow.&nbsp; This day consisted of a line that lasted for 8 hours to spend 5 minutes at the front getting an ID card.&nbsp; Luckily for us most of the wait was indoors with rows and rows of chairs and desks.&nbsp; On the other hand, we didn\'t know that we would be sitting for so long and didn\'t bring anything to do.&nbsp;   5 Jul 04 - This was a \"down day\"&nbsp; which translates into \"we have nothing to do, so look busy.\"&nbsp; We sat around the barracks all day.&nbsp; The barracks are fairly nice, good air conditioning and new beds.&nbsp; I have spent time in worse places.  6 Jul 04 - Woke up at 5:30 and had a road march with our body armor on.&nbsp; Imagine going for a walk, in long pants and long sleeve shirt, then throw on 50lbs. more gear.&nbsp; Later we were trained in identifying and reacting to UXO (unexploded ordinance) and IED\'s (Improvised Explosive Devices, a big killer in Iraq).&nbsp; I gave a class on identifying and treating heat injuries.  7 Jul 04 - Dad\'s B-day!&nbsp; We attended mandatory briefings - that means sitting through 8 hours of stuff we had heard at least twice before.&nbsp; Again, not so bad, since there was air conditioning.&nbsp;   8 Jul 04 - Another \"down day,\" which again we had to look busy.  9 Jul 04 - Red letter day!&nbsp; We did our SRP (soldier readiness packet) which consists of making sure you are up to date on your home of record, next of kin, life insurance, legal matters (marriages, births, divorces, powers of attorney) and of course MEDICAL!&nbsp; Everyones favorite, since that is where we get poked like little experimental monkeys.&nbsp; I got 6 different things: PPD (TB test), my first Anthrax shot (1 of a 6 shot series), Small Pox (turns into a nasty little bugger and made me feel very ill by about day 9), Typhoid, MMR (measles, mumps, and rubella), and a Hepatitis B shot.&nbsp; We were treated with a trip to the PX (Post eXchange, think Walmart with higher quality stuff and no sales tax) for our troubles.  10 Jul 04 - Went to CIF (Central Issue Facility) where we were issued our DCU\'s (Desert Camouflage Uniform) to include boots, Goretex jacket and pants, hats, and the uniforms themselves.&nbsp; It is always fun to get new stuff!&nbsp; The downside is having to break in new boots.  11 Jul 04 - I have something written down here that for the life of me, I can\'t remember what it is.&nbsp; So here is a joke: If someone speaks three languages they are called trilingual, someone that speaks&nbsp;two&nbsp;languages is called bilingual.&nbsp; So what is someone that speaks one language called?&nbsp; American.  12 Jul 04 - Had to go into Fort Bliss and drive in some vehicles.&nbsp; Managed to stop by the PX again...hard to resist when they have a food court.  13 Jul 04 - Oh yeah, this day.&nbsp; Well, we (medics) were tasked to support a range where some soldiers were zeroing their weapons.&nbsp; The thing is, the active duty component on Bliss is sending out Medics in ambulances to cover all our training.&nbsp; We knew this (since we had found out the hard way 2 days in a row) and told the commander as much.&nbsp; She said to go anyways, since we couldn\'t assume that they would be there.&nbsp; Long story short, we went, they came, we left.&nbsp; WASTE of a day all because our commander didn\'t want to listen to us.&nbsp; But hey, I did manage to make it to the PX again!  14, 15, 16 Jul 04 - This is where I forgot to write stuff down.&nbsp; I can\'t remember what we did.&nbsp; Probably wasn\'t important.  17 Jul 04 - Look at that!&nbsp; Another trip to the PX!  18 Jul 04 - Media awareness class.&nbsp; This is the Army\'s way of saying \"If you can\'t say something nice in front of the camera, don\'t say anything at all.\"&nbsp; An interesting class, but I felt like hell because of the small pox shot.&nbsp; I had flu symptoms associated with the shot.&nbsp; Nothing like having a fever in the heat of the southwest and having to wear all that gear (50lbs., remember?).  19 Jul 04 - This amounted to a down day, which I needed for some recovery.&nbsp; I felt much better but still not quite up to snuff, what what?  20 Jul 04 - The PA (Physicians Assistant) at the TMC (Troop Medical Clinic) gave us a class on the signs and symptoms of the Flu, Cold, and Sinusitis.&nbsp; A very good class, and perfect for me since my sinuses were killing me.&nbsp;   21 Jul 04 - A day at the range.&nbsp; I must have sweat my body weight this day.&nbsp; The guy in front of me that was trying to zero his weapon couldn\'t shoot for shit.&nbsp; As a result, I had to stand in the HOT sun for hours and hours.&nbsp; He must have gone through 50 rounds of ammunition.&nbsp; They finally pulled him off to let me shoot.&nbsp; I was done in 12 rounds.&nbsp; From there we went to the SRM (Short Range Marksmanship) range.&nbsp; This would have been fun, but my head was killing me.&nbsp; I would have thought it would occur to me that it was my sinuses (which I discovered later) but I had sweat so much I thought it was dehydration.&nbsp; Brilliant, can\'t even properly diagnose myself.&nbsp;   22 Jul 04 - Today we did Land Navigation and PLGR training.&nbsp; A PLGR (pronounced \"plugger\") is a military GPS (global positioning system, using satellites to tell you where you are)&nbsp;unit that was the cutting edge of GPS technology 10 years ago.&nbsp; Since that time, they have gotten as small as a cellular phone, but the Army still uses the PLGR, which is&nbsp;several times that size.&nbsp; It was still good training, and a perfect day for it.&nbsp; Not too hot, a nice breeze.&nbsp; We were the first group to finish.&nbsp; Then we had to go and administer the second shot in the Anthrax series to some people.&nbsp; Nothing is as much fun as poking people with needles.&nbsp;   23, 24 Jul 04 - Again, forgot to write anything down.  25 Jul 04 - So there I was, doing laundry in the laundry room.&nbsp; I needed to go use the latrine so I grabbed my helmet and headed out.&nbsp; On the way the First Sergeant (highest ranking enlisted person) saw me and yelled to me.&nbsp; I thought I was in trouble for not having my gear on.&nbsp; As a matter of fact, the first thing he asked me was \"Where is your gear?\"&nbsp; So I thought I was in trouble.&nbsp; He put his arm around me and said come with me.&nbsp; Then he dropped the hammer.&nbsp; \"Your Grandfather died.\"&nbsp; Life\'s little curve ball.&nbsp; I responded by jumping away and saying \"What!?!\"&nbsp; The next part was actually kind of funny looking back on it.&nbsp; He looked at the Red Cross message and showed me and said \"This is you, right?\"&nbsp; You would think that would have been all verified by the time he told me, huh?&nbsp; Well, it was, I don\'t know what the hell he was doing.&nbsp; So, the Army flew me home the next day on Emergency Leave for 12 days, free of charge.  26 Jul 04 - LONG flight, crappy layovers, didn\'t get in until midnight.&nbsp; Still, nice to be home and see my girlfriend and the puppies.  27 Jul 04 - Grampa\'s viewing is tommorrow, and the funeral is the next day.&nbsp; I am a pall bearer.&nbsp; I have been a pall bearer for each of my grandparents that have passed, all within the last ten years.&nbsp; I am quite sure I will cry my heart out, since we were pretty close.&nbsp; No shame there.&nbsp; I wish he would have known I was going to Iraq before he went.&nbsp; Nothing to be done about it, though.&nbsp; I will put in an extra set of dogtags in with him.&nbsp;&nbsp;He is a WWII vet that I am very proud of, so he will undestand the gesture. &nbsp;He went peacefully after a walk in the morning.&nbsp; He had struggled with congestive heart failure for a couple of years now, and we had struggled with his failing eyesight and hearing for much longer.&nbsp; Still, hard to imagine the man that used to take me and my brother fishing and camping is being returned to the earth.&nbsp;  I had hopes that he would hang on at least until I came back from Iraq.&nbsp;  '),(12364,3576839,'15,July,2004','2004-07-15 00:00:00','   So here I am in Fort Bliss and it is hot hot HOT.  I don\'t get much internet (this is the first time) so I will make this short.  We have done some training, but real training doesn\'t start until late July.  Temps reach near 100 every day, and we have to wear about 50 pounds of gear.  We are treated pretty well, and the barracks we stay in are actually alright.  The air conditioning works, so no complaints there.  There is a level of frustration with the amount of flexibility we are supposed to have, but please refer to \"Being Gumby.\"  We finished getting new uniforms (desert, duh) and have gotten cleared through medical.  Of course, clearing through medical meant 6 shots for me, including anthrax and small pox.  None of them hurt much, so it is all good.  That small pox is poxing up really nice, which is disgusting.  It is like a massive whitehead on my shoulder.  It should be healed fairly soon, though.  On a personal note, if anyone emails me, please use my military address.  I cannot check my Yahoo! from this computer, so who knows how it will be when we get \"in-country.\"  I will write more as I can, and I am keeping a bit of a log so I will more fully update it as the opportunity arises.  '),(12365,3196705,'24,April,2004','2004-04-24 00:00:00','    BLOGGING  I just realized that i don\'t know if i can curse or not here.  boy, this is embarrassing.  If only those words didn\'t exist in vocabulary.  i will just watch myself and hope i don\'t slip up   '),(12366,3196705,'24,April,2004','2004-04-24 00:00:00','    COMMERCIALS  I cannot express in words the rage i feel every time i see the pavlovian domino\'s pizza commercial.  i want knock the pizza guy\'s teeth out each time i here him chastise the people making his job necessary.  these are the things that really bug me.   '),(12367,3196705,'24,April,2004','2004-04-24 00:00:00','    MOVIES  I don\'t know whether to blame tony scott or christian wagner for the movie that is \"man on fire.\"  how much say did wagner have in post?  the movie would have been great at 100 minutes, but it rolls to a stop at 146.  was it drama, or action?  i don\'t think scott wanted to pick, and it shows.  two biggest miscues: Radha Mitchell  picking up a southern accent half way through, and the cutesy subtitles.  inovation sometimes fails, as it did here with the dancing, haphazard use of subtitles.  not being a big fan of scott, i still have to say he messed up a good story.  maybe that is his true gift.  at least \"spy game\" was shown out of order when i saw it: that made up for the shabby post-production work.  what was the reason this time?  Man on Fire:  B-   '),(12368,3196705,'24,April,2004','2004-04-24 00:00:00','    COMICS  just read today\'s \"boondocks.\"  i must remember to mark this day down, just to see how close McGruder comes to picking the right day   '),(12369,3196705,'24,April,2004','2004-04-24 00:00:00','    BASEBALL  Following the giants game on espn gamecast, it occurs to me that i die a little bit inside everytime tucker comes to the plate   '),(12370,3196705,'24,April,2004','2004-04-24 00:00:00','    Blog!  the world rejoices as the thoughts of charles melin are made available daily.   '),(12371,3196705,'01,May,2004','2004-05-01 00:00:00','    VENTURING INTO NEW MEXICO  taking a trip to new mexico to deliver a toddler son to his mother.  peace   '),(12372,3196705,'30,April,2004','2004-04-30 00:00:00','    YAHOO WORD OF THE DAY  some say today\'s word applies to me: feckless.  those fools don\'t know what they are talking about   '),(12373,3196705,'29,April,2004','2004-04-29 00:00:00','    CATHOLIC CHURCH  the catholic church is debating whether to let admitted supporters of abortion parttake in the sacrements.  what about admitted pedophiles?  just wondering.   '),(12374,3196705,'29,April,2004','2004-04-29 00:00:00','    NHL  apparently the nhl is still functional.  who knew?  i\'d give a pick to win lord stanley\'s cup, but i really don\'t know who\'s left.   '),(12375,3196705,'29,April,2004','2004-04-29 00:00:00','    SO ARE YOU THE POT OR THE KETTLE?  over at therightwing.com, a poster has taken the democrats to task for supporting a draft dodger in \'92 and a war hero in \'04.  for shame.  the problem:  the republicans supported a war hero in \'92 and a draft dodger in \'04.  how can you critize?  are they saying that we should always go for the draft dodger?   '),(12376,3196705,'29,April,2004','2004-04-29 00:00:00','    TILLMAN HONORED  Tillman was promoted and given the silver star for heroic acts in combat.  maybe this will put into perspective what soldiers must do to earn silver stars.  they are not given out lightly.  let\'s give due respect to  all  recipients of this cherished reward. (read Kerry)   '),(12377,3196705,'29,April,2004','2004-04-29 00:00:00','    TERRY JONES ON SEMANTICS  over at the  urlLink guardian \'s comment pages Terry Jones of Monty Python fame as some words on the United States\' word choices in our campaign in iraq.  he finds it odd that  urlLink US mercenaries are called \"civilian contractors\"  while iraqi citizens are labeled \"insurgents.\"   there does seem to be an unfair slant on terminology thrown around in this conflict, er \"war.\"   '),(12378,3196705,'29,April,2004','2004-04-29 00:00:00','    CANCER IN THE CLUBHOUSE?  A.J. Pierzynski is being called a  urlLink cancer in the clubhouse .  it sounds like the giants are looking for scapegoats to take the blame for a sub-par april.   '),(12379,3196705,'29,April,2004','2004-04-29 00:00:00','    BUSH THE IMPERIALIST  a commenter chastised me for speaking ill of the president due to his efforts to grow the country with at the least economic imperialism, and at most territorial imperialism.  i say that if we\'re going to go that route, why are we picking countries so poorly?  iraq has oil and we got to depose of a madman, but saudi arabia is right next door, why not them?  americans love italian and chinese food, can they be next on the list?  and canada has been asking for it, the way they don\'t protect the border and offer us drugs and other health care related treats.  they want it.  let\'s get going before that other imperialist, france takes all the choice countries.   '),(12380,3196705,'29,April,2004','2004-04-29 00:00:00','    NFL DRAFT REVISITED  a commenter points out that i had stated that bj Symons from texas tech would not be drafted.  i was wrong.  he went in the 7th round to houston.  here\'s to his career working out at least as good as kliff kingsbury\'s has.  eli manning= elway?  only time will tell.  i hate elway, but i don\'t hate eli yet.  he reminds me of jd drew, passing on the phillies to go to the cards.  if you want to ruin yourself i the public\'s eye, go for it.  for his sake i hope he pans out better than rivers.  actually, i kind of hope he fails.  i have to be honest.   '),(12381,3196705,'29,April,2004','2004-04-29 00:00:00','    NFL DRAFT REVISITED  a commentor points out that i had stated that bj Symons from texas tech would not be drafted.  i was wrong.  he went in the 7th round to houston.  here\'s to his career working out at least as good as kliff kingsbury\'s has.  eli manning= elway?  only time will tell.  i hate elway, but i don\'t hate eli yet.  he reminds me of jd drew, passing on the phillies to go to the cards.  if you want to ruin yourself i the public\'s eye, go for it.  for his sake i hope he pans out better than rivers.  actually, i kinda hope he fails   '),(12382,3196705,'29,April,2004','2004-04-29 00:00:00','    THE MAVERICK MINUTE  bill simmons is mostly right on the subject of the  urlLink mavs .  seeing nelson go is a must.  all offense all the time is fun to watch, but painful to route for.  i think it\'s time to let finley go, and hopefully get more young and tenacious talent.  they need a stopper in the biggest way.  rasheed could help at this point.  they need someone that can step up and take charge in the post on both ends of the floor, and they need to move bradley.  maybe he could work for cuban getting things down out of the top shelf and free lance saving kittens from trees, as this basketball thing seems to be alluding him.  all in all, they need to pick big three pieces (nash, dirk, and jamison) keep their role players (najera, fortson, howard, and daniels), and start fresh.  trim the offense fat and build something that resembles a nba team   '),(12383,3196705,'29,April,2004','2004-04-29 00:00:00','    PRODUCTIVE THOUGHT ON POP  check out  urlLink baseball musings  for more and better thoughts on productive outs and those suffering from POS (productive out syndrome) like  urlLink buster olney   '),(12384,3196705,'29,April,2004','2004-04-29 00:00:00','    PRODUCTIVE THOUGHT ON POP  chaeck out  urlLink baseball musings  for more and better thoughts on productive outs and those suffering from POS (productive out syndrome) like  urlLink buster olney   '),(12385,3196705,'29,April,2004','2004-04-29 00:00:00','    HOMERS BY THE BAY  barry is out homering his team 10 to 9.  ouch   '),(12386,3196705,'29,April,2004','2004-04-29 00:00:00','    GIANTS AND PRODUCTIVE OUTS  seeing the giants at the bottom of the productive outs list was not surprising, as i\'d expect them to be near the bottom of most team batting categories other than IBB.  looking at the walks leaders, the top four walking teams are all in the bottom 6 in POP.  looking at batting average, though, all the 6 top POP teams lead the bottom 6 POP teams, allowing for a higher OBP as a group.  this is misleading, as these groups will both regress towards the mean, making the teams that walk more the true threats to score more runs.  also, of the bottom POP teams, the red sox and yankees, can expect significant improvement later on.  boston will improve with the return of nomar and nixon, and the yankees will improve when their shortstops start hitting.  in the top 6 of POP, who can reasonably expect texas to keep their .316 average?  they are playing well above their heads as a whole.  all in all, i\'d say that POP is a POS stat.   '),(12387,3196705,'29,April,2004','2004-04-29 00:00:00','    LOOKING LEFTWARD  looking for a viewpoint from the left?  visit  urlLink http://www.leanleft.com/   '),(12388,3196705,'29,April,2004','2004-04-29 00:00:00','    NEGATIVITY  i\'m all for negativity when it\'s appropriate, but in today\'s University Daily at Texas Tech, Adam Boedeker decided to poison a tribute to pat tillman with a needless shot at other pro athletes.  why make these remarks?  my guess is that adam in a moron.  see?  negativity works when used appropriately.   '),(12389,3196705,'28,April,2004','2004-04-28 00:00:00','    BONDS AND THE GIANTS  speaking as a die-hard bonds fan, these walks are great.  he might get on base every single game this year.  as a giants fan, i wonder if they should just start signing every fifth fan to either pitch mop-up or bat seventh.  i don\'t see how that would work out any worse than the current plan   '),(12390,3196705,'28,April,2004','2004-04-28 00:00:00','    THE O.C.  what was once a guilty pleasure has now become to nice and predictable.  someone needs to die or see aliens or turn to moromism to spice things back up.  come on, i need my junk food for the mind.   '),(12391,3196705,'27,April,2004','2004-04-27 00:00:00','    BARRY BONDS   from and letter to luce.  i see where people are coming from, but no one has produced any evidence.  even the annymous report saying he recieved them is being refuted.  i think most of this is people wanting to hate him because he\'s a jerk to the media.  i.e. the people that report this stuff   '),(12392,3196705,'27,April,2004','2004-04-27 00:00:00','    JOGGING  man-o-man.  i had told myself i would start jogging this week.  yeah, that did not happen at all.  is it that i don\'t want to change, or that i really hate jogging?  maybe i\'m just lazy.   '),(12393,3196705,'27,April,2004','2004-04-27 00:00:00','    DAMN THE  SCHEDULE-MAKERS  i am a tv junkie, and i enjoy 4 shows that come on tonight.  the problem: they all come on in two slots on the schedule, making either miss half of them or tape half of them.  boy, life is so hard.  not like in other, war-torn countries.   '),(12394,3196705,'27,April,2004','2004-04-27 00:00:00','    THE RESTAURANT  wow that intern sucked.  my only concern was that while he was annoying, he made for great reality tv.  i hope some others can step up and lower the bar   '),(12395,3196705,'26,April,2004','2004-04-26 00:00:00','    WHAT I THINK ABOUT DURING THE COMMERCIALS  how much brain damage and/or bribe money it would take for me to vote for bush.   '),(12396,3196705,'25,April,2004','2004-04-25 00:00:00','    CHARLES GOES JOBBING  i hit up a lot of places today, but no luck.  el paso can be a cruel mistress   '),(12397,3196705,'25,April,2004','2004-04-25 00:00:00','    RUSH ON CLINTON  apparently rush links loyalty to clinton as a sign his friends fear for their lives.  could it just be that people are loyal because he\'s a nice guy?  maybe bush should kill some people to keep his friends on the ranch   '),(12398,3196705,'25,April,2004','2004-04-25 00:00:00','    TILLMAN  i\'m sorry, but pat tillman does not deserve a full membership in the hall of fame.  his actions were brave and noble, but the hall of fame should be reserved for actions on the football field.  put him in as a special exibit, and honor him this year.  i like the idea of a black bar to signify loss.   '),(12399,3196705,'25,April,2004','2004-04-25 00:00:00','    NFL DRAFT  i\'m glad to see that Carlos Francis from my ttu got drafted in the 4th round today.  i hope welker catches on as a returner somewhere, but i fear that is wishful thinking. make sure to catch 5 wide sets in oakland this year:  you may see francis   '),(12400,3196705,'25,April,2004','2004-04-25 00:00:00','    ESPN  I love sports, but 5th roud coverage of the nfl draft?  come on   '),(12401,3196705,'08,May,2004','2004-05-08 00:00:00','    HOMERS: BONDS V. GIANTS  three others go deep, bonds gets bb 49  BONDS      10 THE REST   17   '),(12402,3196705,'07,May,2004','2004-05-07 00:00:00','    HOMERS BONDS V. GIANTS  feliz does it again in a giants win.  bonds             10 other giants    14   '),(12403,3196705,'06,May,2004','2004-05-06 00:00:00','    E-VOTING  so what if the system is wrought with insecurities and susceptible to hackers?  so what if there is no federal law to back up the voting with a paper print out?  it\'s the future!   '),(12404,3196705,'06,May,2004','2004-05-06 00:00:00','    TORTURE OR ABUSE?  i don\'t see the problem here.  we\'re at war, you never know what those naked POWs will do.  don\'t they know we\'re their liberators?  we just use chemical sticks to sodomize people as an initiation right for democracy.   '),(12405,3196705,'06,May,2004','2004-05-06 00:00:00','    HOMERS: BARRY V. GIANTS  the rest of the giants widen their lead:  now it\'s  barry      10 the rest  13  tonight\'s dinger comes from feliz   '),(12406,3196705,'05,May,2004','2004-05-05 00:00:00','    SAY HEY  it\'s wille mays\' birthday   '),(12407,3196705,'05,May,2004','2004-05-05 00:00:00','    PAJAMA GRAMS  these are a joke, right?   '),(12408,3196705,'05,May,2004','2004-05-05 00:00:00','    BUSH SAYS SORRY  bush apologizes for the treatment of iraqi prisoners.  let\'s hope we can start treating the iraqi people like we say they should be treated.   '),(12409,3196705,'05,May,2004','2004-05-05 00:00:00','    BONDS V. GIANTS  journaling barry\'s quest to out homer the rest of his team  bonds 10 other giants 12   '),(12410,3196705,'04,May,2004','2004-05-04 00:00:00','    COMMON SENSE AND THE \"PRODUCTIVE\" OUT  people smarter than i have argued against this \"stat\" and people i\'d assume at least as smart as me have argued for it.  this argument comes down to this: there is only one instance when an out would even be on par with the production of a hit: less than to out, man on third, sacrifice fly/hit.  there is no other instance when any right thinking individual would prefer an out to a hit.  how can an out be productive when it is an inferior outcome?   '),(12411,3196705,'04,May,2004','2004-05-04 00:00:00','    AND THE AWARD GOES TO...   urlLink james guillen, a new jersey middle school basketball coach, was fired from his job for given a student the \"crybaby award\"  for complaining about playing time.   at least the kid got an award.  what a crybaby   '),(12412,3196705,'04,May,2004','2004-05-04 00:00:00','    DECLARING WAR ON ERIK  for no good reason i have made erik bates my enemy for at least a few minutes.  check out his  urlLink blog  and see how bad these things can be.   '),(12413,3196705,'04,May,2004','2004-05-04 00:00:00','    ERIK BATES QUOTE OF THE DAY  I learned last night that God is schizophrenic   '),(12414,3196705,'04,May,2004','2004-05-04 00:00:00','    NO MOORE   urlLink disney has put the breaks on releasing michael moore\'s new film fahrenheit 9/11 .  censorship of a movie like this makes it look like there is something to hide.  we live in a country where the free flow of ideas should be encouraged.  how else do we keep people in check?  totalitarian dictators don\'t give the minority a voice.  we chould be better.   '),(12415,3196705,'04,May,2004','2004-05-04 00:00:00','    GODSEND  remember when deniro cared?  this is a silly film full of fabricated scares.  just not good  C   '),(12416,3196705,'04,May,2004','2004-05-04 00:00:00','    ELEPHANT  hypnotic and dreamily paced, this film allows you to slip into the doldrums of high school before showing the brutality of anguished teens.  the movie leaves the answers up to the audience, not pretending to know the warped mind set of a killer.  A+   '),(12417,3196705,'04,May,2004','2004-05-04 00:00:00','    MEAN GIRLS  this was a good movie.  while some parts were cheesy, this was a teen comedy.  meadow\'s killed as the principal.  i must say that every on-screen high school is nothing like mine was.  A-   '),(12418,3196705,'04,May,2004','2004-05-04 00:00:00','    FELLINI FILM CAFE  i\'d give the experience a solid A, food an A, and the drinks a B   '),(12419,3196705,'04,May,2004','2004-05-04 00:00:00','    BUSH AND KERRY 1968-1973  motherjones.com has a timeline comparing the two presidential candidates during this time period.  it is pretty  urlLink revealing .  the sight does seem to back kerry.  fyi   '),(12420,3196705,'04,May,2004','2004-05-04 00:00:00','    TOUGH CROWD 2  this show works for one reason:  the awful collin quinn.  his inability to form coherent speech highlights his inability to say anything he doesn\'t hear on fox news.  he plays as the perfect voice for the people he agrees with, and is the perfect boob for those that know better.  a quick aside:  they changed the tagline from \"4 rounds of smart comedy\" to \"4 rounds of honest comedy.\"  while i do not find most of it honest, it sure isn\'t smart when quinn, dipaolo, or vos talks.   '),(12421,3196705,'03,May,2004','2004-05-03 00:00:00','    TILLMAN USED AS A PROP  i\'m tired of people using tillman as an example of our troops in iraq.  he wasn\'t there.  he was in afghanistan.  there are to battles going on here, and tillman was only part of one of them.  maybe he supported the war in iraq, i don\'t know.  i do know that he supported the war in afghanistan and fought and died there.  keep iraq out of it.   '),(12422,3196705,'03,May,2004','2004-05-03 00:00:00','    TOUGH CROWD  last night they commented on the pictures of our troops embarrassing iraqi prisoners as ok because at least we didn\'t torture them.  so we are only holding our military up to the standards of the people we deemed too brutal to rule?  dont\' we claim them to be blood-thirsty terrorists?  shouldn\'t we have a higher standard?   '),(12423,3196705,'03,May,2004','2004-05-03 00:00:00','    CHARLES\' CRUSHES  1. raquel urban 2. amanda escontredas 3. melissa lugo 4. sze-lyn lim 5. katie moore  you might notice that none of my old girlfriends are on this list.  this illuminates my earlier contention about not  dating the right people.   '),(12424,3196705,'02,May,2004','2004-05-02 00:00:00','    LOOKING BACK  wow.  i was looking at an old yearbook from my highschool, and it brought back some memories.  mostly i remembered how little i knew or even cared about most of these people.  there was also a few pictures of the girl i really liked inn high school, melissa lugo.  the story of my life:  she didn\'t like me back.  i can only get people i don\'t like to like me.  i need to work on that.   '),(12425,3196705,'02,May,2004','2004-05-02 00:00:00','    GIANTS  after sweeping the marlins the giants are 12-14 and 4 games back.  the most destressing stat for me this season is that even after sweeping the fish, the giants are still only 8-9 at home.  luckily, the bats seem to have at least a glimmer of life, and that is more than they had earlier this season.   '),(12426,3196705,'02,May,2004','2004-05-02 00:00:00','    GNOT GNOSTICS, JUST GNUTS    urlLink xxxchurch .  wow.  watch the videos.  wow   '),(12427,3196705,'02,May,2004','2004-05-02 00:00:00','    MISSION ACCOMPLISHED  one year and going strong   '),(12428,3196705,'02,May,2004','2004-05-02 00:00:00','    RIGHT USING SCIENCE?  instaundit.com links to a  urlLink story  bashing the science presented in movies  urlLink the day after tomorrow  and those that would raise questions on the validity and possibility of the events described for political gain.  last time i checked, you cannot die and come back to life again.  you know, scientifically speaking.   '),(12429,3196705,'15,May,2004','2004-05-15 00:00:00','    anyone that wants to contribute to this site on a semi-regualar to regular basis, just let me know. i\'m taking applications for about the next week or two.  don\'t worry if you have differing opinions from me, i\'ll still consider you.   '),(12430,3196705,'14,May,2004','2004-05-14 00:00:00','    urlLink Satan is a pepsi ball .  who knew?   '),(12431,3196705,'14,May,2004','2004-05-14 00:00:00','    urlLink ricky is in trouble for smoking weed .  he says he didn\'t, the nfl says he did.  why do all my favorite players get in trouble?   '),(12432,3196705,'14,May,2004','2004-05-14 00:00:00','    erik is graduating from tech, \"Yeah and just a shade under a decade. All right.\"  i thought i ws there a long time, but e bested my ability to drag feet.   '),(12433,3196705,'14,May,2004','2004-05-14 00:00:00','    urlLink cut to the chase , amoung others has an article describing the defacing of holy relics in iraq by us troops.   urlLink from political commentary by marc rios:    United States forces fighting in the holy Shiite city of Najaf have damaged the shrine of Ali. This is extremely significant: imagine being a Catholic and finding that some nation had damaged St. Peter\'s Basilica. The damage is only four bullet holes, but the symbolism is very very serious.  wow.  there is no respect for the religion of iraq being shown is there?   '),(12434,3196705,'14,May,2004','2004-05-14 00:00:00','    urlLink or so the right is spinning it .  it\'s so simple: have an link to an al qaeda.  move to iraq.  get your buddies to cut your head off.  get your dad to spread \"lies\" about bush.  it is so simple.  thank you chris regan for explaining it to me!   '),(12435,3196705,'14,May,2004','2004-05-14 00:00:00','    \"Despite all of George Bush\'s troubles, the Kerry campaign can\'t seem to get any traction\"  urlLink glenn reynolds   considering he\'s is running at a dead heat with bush, what does that say about the president?  he goes on to say that jimmy carter\'s presidency was not a high point in american history.  not like g.w. bush!   '),(12436,3196705,'14,May,2004','2004-05-14 00:00:00','    is today.  let\'s pay our respects to the good soldiers, they are sullied by the bad.   '),(12437,3196705,'14,May,2004','2004-05-14 00:00:00','    this is  urlLink great atricle  i found via  urlLink bush lies .  it\'s really amazing how bush is never at fault. ever.   '),(12438,3196705,'14,May,2004','2004-05-14 00:00:00','    urlLink doesn\'t this seem like overkill ?  couldn\'t they get a cheaper car and crank up the rpms?   '),(12439,3196705,'14,May,2004','2004-05-14 00:00:00','    apparently america has decided to both leave and not leave if asked on july 1st.  i wonder how that works?   '),(12440,3196705,'14,May,2004','2004-05-14 00:00:00','    i loved the scope of this movie.  it looked great, but the music killed the mood every chance i got.  good preformances all around, but bana steals the show.  isn\'t helen supposed to be beautiful?  it worked overall.  oscar will be making a call for special effects, maybe even a supporting nod.  A-   '),(12441,3196705,'14,May,2004','2004-05-14 00:00:00','    now it will be easier to track my thoughts!   '),(12442,3196705,'14,May,2004','2004-05-14 00:00:00','    check out this sight for religious commentary http://runningoff.blogspot.com/.   '),(12443,3196705,'14,May,2004','2004-05-14 00:00:00','    COMMENTS BACK ON TRACK  i\'ve been having problems with blogspot\'s new comment system, so i went back to haloscan.  i hope this pleases all my loyal minions.   '),(12444,3196705,'13,May,2004','2004-05-13 00:00:00','    HOMERS: BONDS V. GIANTS  bonds goes 1-2 with three walks, 2 IBB.  and the giants lose.  again.  a.j. hits one out and pushes the team lead to 14 over barry.   bonds is frustrated at losing, and who could blame him?  bonds    10 the rest 24   '),(12445,3196705,'13,May,2004','2004-05-13 00:00:00','    LOOKING AROUND THE BLOGS  i have been reading many random blogs and i\'m blown away with how poor the grammar is overall.  for shame   '),(12446,3196705,'12,May,2004','2004-05-12 00:00:00','    MORE MOORE FOR MIRIMAX   urlLink mirimax will buy moore\'s new film, but will have to find someone else to distribute it .  this is crazy.  what is also crazy is disney\'s position that this has been blown out of proportion.   '),(12447,3196705,'12,May,2004','2004-05-12 00:00:00','    CLINTON TO THE RESCUE?  scarborough thinks bush could use clinton to fix the mess in the middle east.  maybe if we had a president that didn\'t screw things up so bad we wouldn\'t have to turn to other, better, presidents to make things right.   '),(12448,3196705,'12,May,2004','2004-05-12 00:00:00','    FRUM SAID IT  arabs amoral.  never mind the racism, he makes points with false info to prop up bad arguments.  on an aside, flavia colgan is pretty cute.   '),(12449,3196705,'12,May,2004','2004-05-12 00:00:00','    FRUM IS RIGHT ON ONE POINT  moron david frum commented that this war (iraq) was forced on us.  i say yes it was.  we have to realize who forced it on whom, though?  i\'d say bush.  frum says terrorists.   '),(12450,3196705,'12,May,2004','2004-05-12 00:00:00','    I GUESS I\'M NOT PART OF SCARBOROUGH COUNTRY  this nut agreed with a nut calling to hate people we disagree with.  i thought that we were the country that wanted to help people.  last time i checked, the god we have to pledge to was the christian god.  i\'m pretty sure jesus wouldn\'t want want us hating people.  i guess i\'m not a good enough person to see the logic in hate.   '),(12451,3196705,'12,May,2004','2004-05-12 00:00:00','    HOMERS: BONDS V. GIANTS  three other giants go deep, including schmidt and feliz twice  bonds      10 the rest   23  bonds was also kept off the basepaths for the first time this season, spawning a new watch:  the barry off base watch.  BARRY OFF BASE   1   '),(12452,3196705,'12,May,2004','2004-05-12 00:00:00','    VELVET REVOLVER  i wounder what e thinks about these guys?   '),(12453,3196705,'11,May,2004','2004-05-11 00:00:00','    WOW  oh the things you can learn that make this president look worse and worse.   urlLink we had a shot at killing the man behind the nick berg killing two years ago, but passed to set up for the war on iraq .  would bombing a known terrorist have pushed back bush\'s war plan at all?  wow.   '),(12454,3196705,'11,May,2004','2004-05-11 00:00:00','    HOMERS: BONDS V. GIANTS  they have almost doubled him up.  feliz goes deep to widen the gap between the rest of the giants and bonds.  could it be bonds will walk 300 times and only homer 30 times?  the way the giants are hitting?  yes, that could very well happen.  bonds     10 the rest  19   '),(12455,3196705,'11,May,2004','2004-05-11 00:00:00','    KEEP THAT BUCK AWAY FROM ME!  is it me, or has bush never done a single thing wrong?  how is it that the leader of the military is always the last to know?  if the aren\'t telling him things, why not fire his advisors?  this president is really a joke.   '),(12456,3196705,'11,May,2004','2004-05-11 00:00:00','    KERRY ON KICKING OUT RUMSFELD  kerry suggests among others mccain.  i say make him kerry\'s vp.   '),(12457,3196705,'11,May,2004','2004-05-11 00:00:00','    WHY NOT RATED \"R\" FOR DRIVING CARS?  cars kill a lot of people, too.   urlLink congress is looking to label any movie with smoking as an R rated movie .   '),(12458,3196705,'10,May,2004','2004-05-10 00:00:00','    T\'WOLVES  i remember a week ago when they had spent all their energy against denver and would be done versus the kings.  they should have told kevin garnett that.   '),(12459,3196705,'10,May,2004','2004-05-10 00:00:00','    THEIR HEART WAS IN THE RIGHT PLACE, BUT THE WALLET WINS THE DAY  idiots write about  urlLink sports has a story about the university of iwoa pretending to care about the plight of native americans .  too bad they couldn\'t really make a stand.  maybe texas tech should take a stand, seeing as we won\'t have to weasel out of our policy due to conference schedules.   '),(12460,3196705,'10,May,2004','2004-05-10 00:00:00','    REPERCUTIONS FROM RADICALS FOR IRAQI PRISONER ABUSE   urlLink this  is the real problem that those u.s. troops have created.  they have given ammunition to the radical groups that claim we are worse than dictators like saddam.  i hope we can eliminate these elements without abusing others.  we need to create some sort of good faith in the middle east, and then get out of there.  i am just sad that nick berg, a bit player in this sick ballet, died for being an american.  a quick aside:  why don\'t we try contracting jobs to all the iraqis that don\'t hate us instead of bringing in americans?   '),(12461,3196705,'10,May,2004','2004-05-10 00:00:00','    HIGHLANDER  i was channel surfing today and caught the theme song to \"highlander\".  that has to be the worst theme song ever and in the top five worst of all time.  i never saw the show and never will thanks to that song.   '),(12462,3196705,'10,May,2004','2004-05-10 00:00:00','    US GUILTY OF WAR CRIMES?  here are two articles of the  urlLink geneva convention  that woudl tend to lead me to say yes.   PART III CAPTIVITY SECTION I BEGINNING OF CAPTIVITY  ARTICLE 17 ...No physical or mental torture, nor any other form of coercion, may be inflicted on prisoners of war to secure from them information of any kind whatever. Prisoners of war who refuse to answer may not be threatened, insulted, or exposed to unpleasant or disadvantageous treatment of any kind...    SECTION II INTERNMENT OF PRISONERS OF WAR CHAPTER I GENERAL OBSERVATIONS    ARTICLE 22 Prisoners of war may be interned only in premises located on land and affording every guarantee of hygiene and healthfulness. Except in particular cases which are justified by the interest of the prisoners themselves, they shall not be interned in penitentiaries...   '),(12463,3196705,'10,May,2004','2004-05-10 00:00:00','    THE PLEDGE  our pledge of allegiance was around for 62 years without the words \"under god.\"  our pledge has had the words for 49 years.  let\'s just call a mulligan and go back.   '),(12464,3196705,'10,May,2004','2004-05-10 00:00:00','    KILL BILL VOL.2  this movie is as good as advertised.  not nearly as violent as vol. 1, kill bill vol.2 plays like a love story.  much more moving than one would expect, this film lends weight to a strait beat-em-up premise.  all the actors are superb, even the typically awful daryl hannah.  carradine makes the film as bill.  A   '),(12465,3196705,'10,May,2004','2004-05-10 00:00:00','    VAN HELSING  this movie is a hodge-podge of movie monsters that works.  a few scenes are obviously and painfully cgi, but for the most part it looks great and doesn\'t get bogged down with anything put witty dialogue.  B+   '),(12466,3196705,'10,May,2004','2004-05-10 00:00:00','    ERIK\'S PLAN TO GET LAID  from  urlLink jabooty.org . \"Apparently, the \"Better than sex\" cake is not, as one would be led to believe by the name, better than sex. I have been charged with the task of sleeping with as many people as possible until I find the one person who having sex with is actually worse than that cake. At that time we will change the name of the cake to \"Better than sex with _______.\"\"  i like this idea.  i hope it works for him.   '),(12467,3196705,'10,May,2004','2004-05-10 00:00:00','    NO NADER IN TEXAS   urlLink ralph nader has failed to get enough signatures to get on the ballot in texas and will not appear on the ballot.  this isn\'t stopping him, though, as he is filing suit against texas for its petition requirements .  he has a point:  independants have to get roughly 20,000 more votes in 15 less days.  that is nuts.  if you can get 45,000 people to sign for you, i don\'t care if you back mass genocide and death to all children, youshould get on the ballot.   '),(12468,3196705,'10,May,2004','2004-05-10 00:00:00','    ANN COULTER THE DEVIL?  i wouldn\'t know, but she  urlLink hates liberals .  that\'s about it.  thats what eveything boils down to with her.  a brief aside, is bombing all the arabs we can think of better than the ideas she rails against?   '),(12469,3196705,'09,May,2004','2004-05-09 00:00:00','    BUSH JOINING THE PARTY?   urlLink apparently bush has stopped getting his national security updates from cbs news and is asking his staff .  you know, the people the president is supposed to listen to.  it only took him 4 years.  his learning curve is improving!   '),(12470,3196705,'09,May,2004','2004-05-09 00:00:00','    NEW LOOK  how do you like it?   '),(12471,3196705,'09,May,2004','2004-05-09 00:00:00','    MORE MOORE  moore defends his honor in this post at his  urlLink website .  disney is nuts.  \"tower of terror\"?  that is in awful, awful taste.   '),(12472,3196705,'09,May,2004','2004-05-09 00:00:00','    ERIK THE LAME  erik bates has a picture of his movie collection up.  lame,lame,lame.  speed racer, stigmata, and uhf?  he needs help   '),(12473,3196705,'09,May,2004','2004-05-09 00:00:00','    BELTWAY JEOPARDY   urlLink jennifer fley was behind the scenes at the \"power players\" celebrity jeopardy tapings .  i must say that al franken sounds like he\'d be fun to hang out with.   '),(12474,3196705,'09,May,2004','2004-05-09 00:00:00','    HOMERS: BONDS V. GIANTS  bonds is 0-15 and the rest of the giants get another homer  bonds    10 the rest  15   '),(12475,3196705,'09,May,2004','2004-05-09 00:00:00','    MORALS AND IRAQ   urlLink richard reeves  has a good op/ed on the morals behind going to iraq.  why is it that bush seems more upset that pictures got than at the act itself?  going on television in jordan and almost apologizing seems a bit morally bankrupt to me.   '),(12476,3196705,'22,May,2004','2004-05-22 00:00:00','    urlLink mexican strippers are receiving work visas to strip in canada .  looks like mexico is losing one of its last natural resources.   '),(12477,3196705,'21,May,2004','2004-05-21 00:00:00','    well, eating people.  another proposal:  hold the conventions on the same weekend, one saturday and one sunday.  this would eliminate any advantages that come with going first or last, and you can market it like a real event.   '),(12478,3196705,'21,May,2004','2004-05-21 00:00:00','    i think all technology has been leading to this  urlLink point .   '),(12479,3196705,'21,May,2004','2004-05-21 00:00:00','    but the single biggest problem in the world by far is religion   '),(12480,3196705,'21,May,2004','2004-05-21 00:00:00','    you just can\'t stop the lakers if shaq hits his free throws   '),(12481,3196705,'21,May,2004','2004-05-21 00:00:00','    i think the lake show ends this round.  wolves in 6.  also, the pistons strap up and get their \"d\" on.  pistons in 6.  pistons are flying high off of destroying the nets, and they match up well with the pacers.  i see rip going off.  the lakers are old.  they won game 5 on a miracle, and i can shoot as well as the spurs did in game 6.  they were not dominant,  and  the other team slipped.  the wolves are riding good vibes and two hot hot players (kg and sammy).  the only thing that can stop them is for shaq to go off for the whole series, or for kobe to play like a man possessed.  i just think the wolves can weather that storm and come out standing.   '),(12482,3196705,'20,May,2004','2004-05-20 00:00:00','    the  urlLink sports guy  as a great plan:  It\'s too bad you could never coordinate this with 40,000 people, but the next time Barry Bonds comes up, if everyone at SBC Park cheered all at once, and then someone in the right-field stands threw a ball into the bay so all those losers in the boats would dive in for the ball ... and then this happened like five or six times ... I mean, that would be kinda fun, right?   i would pay to watch this.  not much, but i\'d pay.   '),(12483,3196705,'20,May,2004','2004-05-20 00:00:00','    claiming torture is ok because it works is crazy.  why not smother a crying baby?  it\'ll shut them up.  just because something can work doesn\'t give us a right to do it.   '),(12484,3196705,'20,May,2004','2004-05-20 00:00:00','    urlLink article III, section 3 of the U.S. Constitution    Treason against the United States, shall consist only in levying war against them, or in adhering to their enemies, giving them aid and comfort. No person shall be convicted of treason unless on the testimony of two witnesses to the same overt act, or on confession in open court.  The Congress shall have power to declare the punishment of treason, but no attainder of treason shall work corruption of blood, or forfeiture except during the life of the person attainted.   can we admit now that no one in the media has committed treason?   '),(12485,3196705,'20,May,2004','2004-05-20 00:00:00','    scarborough keeps using a quote he says is from  urlLink michael moore\'s website , but i can\'t seem to find it.  it\'s about more americans needing to die.  i want to read it for myself, but can\'t find it anywhere.  has anybody seen this?  if so, can you leave me the link?  scarborough wouldn\'t be makng it up would he?   '),(12486,3196705,'20,May,2004','2004-05-20 00:00:00','    scarborough is worried that we are heading to point where a burr-hamilton type duel could happen.  i say do it.  we can pick out two politicians and have them go at.   by the way, this is the first thing i have ever agreed on with anny coulter.   '),(12487,3196705,'20,May,2004','2004-05-20 00:00:00','    with a title like that, i should work at  urlLink page 2 .  two quick thoughts on  urlLink game 7 :   1.  i am happy with the pistons moving on.  i am a fan of rip and big ben.  2.  kidd scoreless?  you mean to tell me that tech alum  urlLink darvin ham  scored as much as kidd in game 7?  i\'m not saying that i could score a point in that game, but i\'m 5\'8 and out of shape.   '),(12488,3196705,'20,May,2004','2004-05-20 00:00:00','    my friend erik is a bush backer, but this  urlLink story  even got him  urlLink upset .  the school\'s actions were bad enough,  urlLink but look at some conservative commentary on the subject .  here is the poem:   REVOLUTION X  Bush said no child would be left behind And yet kids from inner-city schools Work on Central Avenue Jingling cans that read Please sir, may I have some more? They hand out diplomas like toilet paper And lower school standards Because Underpaid, unrespected teachers Are afraid of losing their jobs Funded by the standardized tests That shows our competency When I\'m in detox. This is the Land of the Free ... Where the statute of limitations for rape is only five damn years! And immigrants can\'t run for President. Where Muslims are hunted because Some suicidal men decided they didn\'t like Our arrogant bid for modern imperialism. This is the Land of the Free ... You drive by a car whose Bumper screams God bless America! Well, you can scratch out the B And make it Godless Because God left this country a long time ago. The founding fathers made this nation On a dream and now Freedom of Speech Lets Nazis burn crosses, but Calls police to Gay pride parades. We somehow Can afford war with Iraq But we can\'t afford to pay the teachers Who educate the young who hold the guns Against the \"Axis of Evil\" Land of the Free ... This is the land If you\'re politically assertive They call you a traitor and Damn you to ostracism. Say good-bye to Johnny Walker Lindh And his family. Bye Bye. American Pie. So maybe My ideas about this nation Don\'t resolve around perfection But at least I know Education is more important Than money. Land of the Free . . . If this was utopia We\'d have to see each other naked Before we got married But instead, we see each other naked all the time Because the government has my social security number And the name of my dog! And then we make babies, But don\'t worry, they won\'t be left behind And they grow up saying God bless America! But they don\'t know who Bush is Because they never learned the Presidents. And they will ride the ship Amistad To our dreamland shores Bearing the same shackles as us. I\'m here to say that Generation X Is pissed and we are taking over, Ripping down the American illusion of perfection We are the future generation I have my qualifications I know it looks like Angel Soft paper, But don\'t worry It\'s a diploma Do I look qualified? You can take our toilet paper, But you can\'t take our Revolution.   thanks to the  urlLink progressive review   '),(12489,3196705,'19,May,2004','2004-05-19 00:00:00','    why do i even read that  urlLink garbage ?  my favorite line?  the one about the right not complaining about any news outlet like the left complains about fox news.  i know she\'s heard of the new york times. (the title of this article? \"the other lame \'times\'\")  i have heard my fare share about how biased they are.  again, why do i do it to myself?   '),(12490,3196705,'19,May,2004','2004-05-19 00:00:00','    the debate over ClearPlay dvd players that edit movies of questionable content  urlLink continues .  this is bad.  there is a simple way to stop children from seeing content parents don\'t want them to:  don\'t make it available.  if you don\'t want them watching something you don\'t like, don\'t let in the house.  if you can\'t control your kids, don\'t expect a dvd player to do the job for you.  also, the assault on copyrights and trademarks is appalling.  why have freedom of expression if you can\'t use it?   '),(12491,3196705,'19,May,2004','2004-05-19 00:00:00','    urlLink i\'m not against this idea categorically , but how about after we get out of iraq and fix the economy.   '),(12492,3196705,'19,May,2004','2004-05-19 00:00:00','    remember when bush said he was dedicated to  urlLink lessening partisan politics ?  well,  urlLink where does meeting with only republicans  fit into that goal?   '),(12493,3196705,'19,May,2004','2004-05-19 00:00:00','    urlLink house speaker dennis hastert thinks john mccain needs to respect the plight of our soldiers .  you know, that draft-dodging mccain, that dumb democr-- wait. i think mccain knows about sacrifice.  perhaps hastert should learn to keep his tongue in check.   '),(12494,3196705,'19,May,2004','2004-05-19 00:00:00','    it just doesn\'t seem right that the minority leader is from south dakota, but then again, it is a minor state.   urlLink all that might not matter if daschle doesn\'t get re-elected .   '),(12495,3196705,'19,May,2004','2004-05-19 00:00:00','    urlLink a man at the strip club \"scores\" in new york dropped $28,000 in one night .  who hasn\'t done that?   '),(12496,3196705,'19,May,2004','2004-05-19 00:00:00','    urlLink a story on moral relativism in catholic america .   '),(12497,3196705,'19,May,2004','2004-05-19 00:00:00','    my biggest problem with the current administration is the lack of uniformity on the issue of the value of a life.  a strait life is more than a gay life (at their quality of life), an israeli life is worth more than a palestinian life, and an american life is worth much much more than an iraqi, or really, arab life.  i\'ll concede that bush\'s concern with american lives should be very pressing, as he is the president, but then he places his agenda before american lives.  so my question is, what is a life worth?   '),(12498,3196705,'19,May,2004','2004-05-19 00:00:00','    urlLink that is all i can say .   '),(12499,3196705,'19,May,2004','2004-05-19 00:00:00','    urlLink this guy is suing because they said he might be able to earn a scholarship, then told there weren\'t any .  here\'s a tip:  earn it out of high school!   '),(12500,3196705,'19,May,2004','2004-05-19 00:00:00','    a law designed to kweel inter-racial marriages was repealed in massachusetts after the governor was using it to try to stop gay marriage.  though the law had not been enforced in 40 years, one has to wonder why it has been on the books this long. urlLink story here .   '),(12501,3196705,'19,May,2004','2004-05-19 00:00:00','    this was a monster game by KG.  if that doesn\'t shut up his critics, then what will?   '),(12502,3196705,'19,May,2004','2004-05-19 00:00:00','    this is a forum to educate the masses on anything that comes to mind.  why limit knowledge with set rules of protocol for this noble endeavor?  but questions are fun.  how about this question:  why pick such an obvious movie reference for your handle?   '),(12503,3196705,'19,May,2004','2004-05-19 00:00:00','    So Charles, is there going to be any form of coherence with this blog or is it just a random sampling of your thoughts and the goings on in the news?  Why not ask questions or pose debates to allow for a wider variety of commentary.   '),(12504,3196705,'19,May,2004','2004-05-19 00:00:00','    urlLink maybe they thought it was a gay weddding .  really, though, what did they think the wedding was?   '),(12505,3196705,'18,May,2004','2004-05-18 00:00:00','    urlLink that is their name , not an insult by me.  i concur, though.   urlLink here is a story on their attempts to cleanse their on-line church .   '),(12506,3196705,'18,May,2004','2004-05-18 00:00:00','    anyone that wants to contribute to this site on a semi-regualar to regular basis, just let me know. i\'m taking applications for about the next week or two. don\'t worry if you have differing opinions from me, i\'ll still consider you.   '),(12507,3196705,'18,May,2004','2004-05-18 00:00:00','    urlLink i miss this great show .  i have the dvds, and am excited about the movie, but i miss the days when you had to guess it was on.   '),(12508,3196705,'18,May,2004','2004-05-18 00:00:00','    urlLink nick berg\'s last words a plea for us all to move to philadelphia ?  i think goes beyond stretching it.  not all last words are inspiring.  there is nothing wrong with his words.  we do not need to give greater meaning to them.   '),(12509,3196705,'17,May,2004','2004-05-17 00:00:00','    ok,  urlLink he\'s not randy is .  way to go randy.   urlLink but schmidt threw a one-hitter !  that is just what the giants needed.  now they need to clone him three or four times and maybe they can reach .500.   '),(12510,3196705,'17,May,2004','2004-05-17 00:00:00','    probably.  if you wish there was a voice out there telling me i\'m wrong, go check out the  urlLink liberal antithesis .  note the good choice of blog template, and inform yourself with views from the other side of politics.   '),(12511,3196705,'17,May,2004','2004-05-17 00:00:00','    need proof?  no?  well  urlLink here  is some anyway.   '),(12512,3196705,'17,May,2004','2004-05-17 00:00:00','    i\'m not the only one who thinks  urlLink powell would make a good VP for kerry .   urlLink bell curve at the citizens is of the same accord .   '),(12513,3196705,'17,May,2004','2004-05-17 00:00:00','    urlLink this was one funny guy .  being a tad too young to remember him from his hey day, most of my memories are based on his guest appearances on shows, and not his run on the odd couple.  my favorite memory of him is on snl, where tom hanks is a man with no short term memory.  hanks is on a game show, and at the end of the game how tony randall comes out to be his celebrity partner.  great stuff.   '),(12514,3196705,'17,May,2004','2004-05-17 00:00:00','    on a trek through some blogs, i happened upon a 14 year old that was against gay marriage.  we debated a bit and came this compromise:  she didn\'t like it because she is a homophobic religious nut, and i could see the validity of getting rid of marriage all together.  i went back to her blog to see where she stood today, and her comment section had been removed.  so i win.  in the same vein, i was at a catholic blog that had people debating the same issue with me until i dumbfounded them and they started accusing gay people for all the disease in the world.  so i guess in the battle of open discussion, it\'s charles 2, religious homophobes 0.   '),(12515,3196705,'17,May,2004','2004-05-17 00:00:00','    has any show stood the test of time like  urlLink saved by the bell ?  actually, many shows have stood up much better, but you can\'t beat it a four in the morning.  it is by far the best thing on in that time slot.   '),(12516,3196705,'16,May,2004','2004-05-16 00:00:00','    now i know that in some ways ted rall could be seen as the ann coulter of the left, but at least he uses facts in his articles.   urlLink this is a top-notch article on the iraqi prison scandal .   '),(12517,3196705,'16,May,2004','2004-05-16 00:00:00','    after reading countless attacks by radio personalities, writers, and aveage citizens bash actors and directors for their political views (bashee of the moment, michael moore) i have to ask the obvious question?  why is your opinion better than theirs?  it is amazing how we will listen to the idiot rantings of political pundits and shock-jocks, but once an actor says they favor someone over someone else, well they can go sodomize themselves.  we all have a voice.  none of them really matter, but we all get to use them.  let\'s stop bashing people with opinions just besause they are actors.  we can attack their beliefs, but not their right to state those beliefs.  on that note, rush is a moron.   '),(12518,3196705,'16,May,2004','2004-05-16 00:00:00','    powell was on meet the press discussing many issues relating to the war.  i have to say that of all of bush\'s people i still respect him the most.   urlLink at the end of the interview, questions about powell speaking before the UN in feb. of last year were posed.  at that moment, emily miller, a press aide, attempts to end the interview by shifting the camera .  powell chastises her and the interview concludes.  powell is a class act, answering questions that strike at his very credibility.  that said, how about this:  a unified ticket with kerry picking powell as VP?  thoughts?   '),(12519,3196705,'16,May,2004','2004-05-16 00:00:00','    anyone that wants to contribute to this site on a semi-regular to regular basis, just let me know. i\'m taking applications for about the next week or two.  don\'t worry if you have differing opinions from me, i\'ll still consider you.   '),(12520,3196705,'16,May,2004','2004-05-16 00:00:00','    i don\'t know why people insist on making fun og the south.  oh, yeah,  urlLink yes i do .   '),(12521,3196705,'16,May,2004','2004-05-16 00:00:00','    if i ever have vast amounts of disposable income, you can bet i\'ll order  urlLink this .   '),(12522,3196705,'16,May,2004','2004-05-16 00:00:00','    this daughter of a presidential hopeful was labeled a \"US film director\" by the AP.   urlLink revealing photo here .  i remember her from one of her two acting roles.  (no directing credits)  she was a bartender in the very good \"spartan.\"  she did ok.   '),(12523,3196705,'16,May,2004','2004-05-16 00:00:00','    urlLink we found the weapon of mass destruction .  i knew there was one!  sweet justification for all the lives lost.   '),(12524,3196705,'16,May,2004','2004-05-16 00:00:00','    marc rios (granola snorting nixon)  urlLink was in mass.  for the first leaglly recognized gay marriages in america .  it\'s great to see that there is love in this world.   '),(12525,3196705,'16,May,2004','2004-05-16 00:00:00','    urlLink he has come to save us all .   '),(12526,3196705,'16,May,2004','2004-05-16 00:00:00','    we all love him.  well, we all should love him.  here is how i personally rate all his movies.  they are listed in order of my favorite to least favorite  chasing amy A+  in my top 5.  a real winner, the dialouge and subject matter are classic.  a must watch  mallrats   B+   saved by jason lee, this movie tried to do too much, but in the end it is really, really fun  dogma  A  a great, silly, look at religion that hits what it wants to hit, and only should offend zealots  jersey girl A   a great family film, this movie shows a softness that is not present in his other work  clerks A-   a great film, the dialouge is top-notch.  unfortunately some of the acting is just plain bad, and production values are understandibly low.  all no-budget films could aspire to this  jay and silent bob strike back   C+   smith\'s love letter to himself and his minions, it works as a great clip-show for his other films, but does not stand up well alone  there you go: four great movies, a great comedy, and a film not made for the masses.  if you want opinions on another director\'s work, let me know   '),(12527,3196705,'16,May,2004','2004-05-16 00:00:00','    the pistons even it up, and i hope they win game 7.  the wolves fall and i hope they win as well.  on to the controversy:  the announcers put all the blame for the elbow incident on KG.  the AP, however,  urlLink went the other way with it .  i have to say that both players are guilty, but peeler instigated it and through the dirtiest shot.   '),(12528,3196705,'16,May,2004','2004-05-16 00:00:00','    they are bad.  really bad.  reall, really bad.  but the d-backs are worse.  how did that happen?  by the way, i\'m suspending the homer watch until barry goes deep again   '),(12529,3196705,'16,May,2004','2004-05-16 00:00:00','    wow.  it is amazing how many people are railing against other\'s pursuit of happiness.  i could have sworn it was our right to be happy if it doesn\'t impact another\'s ability.  guess i was wrong   '),(12530,3196705,'29,May,2004','2004-05-29 00:00:00','    that is how we know the picture at  urlLink truth, and other victims of terror  is of a terrorist.  what about the lone ranger?  this is my favorite site to go to for unabashedly slanted reporting.   '),(12531,3196705,'28,May,2004','2004-05-28 00:00:00','    mine is peter cetera\'s \"glory of love.\"  this song just made the karate kid 2.  how did this song come about?  did cetera excitedly tell his friends he just made a great ballad, only to have them laugh at him?  then, as he was throwing away the sheet music, the producer of KK2 calls and asks for a ballad, and cetera figures \"why not\" and sends over a copy.  the movie is ready to ship, but they are still waiting for cetera\'s ballad.  it arrives and they balk, but it is too late for any other song, so it gets used anyway.  i think that is the back story to my favorite bad song, but i could be wrong.   '),(12532,3196705,'28,May,2004','2004-05-28 00:00:00','    that bonds guy is good.  the giants are riding high off bonds\' current hot streak and have won 7 staight.  i\'m getting optimistic about this season now.   '),(12533,3196705,'28,May,2004','2004-05-28 00:00:00','    bill simons again goes after darko:  Darko cheering from the bench with tape over his earlobes ... he\'s either hiding earrings or this is some sort of bizarre Serbian fashion statement. Either way, putting Darko in this series is like putting Corey Haim in \"Saving Private Ryan.\"  i really, really hope darko becomes the best big man in the league in the next five years.  the dude is 18.  why not give him a chance to evolve?   '),(12534,3196705,'27,May,2004','2004-05-27 00:00:00','    they were just the better team tonight.  i like o\'neal, so i\'m glad he\'s ok, but i still hope they lose.   '),(12535,3196705,'27,May,2004','2004-05-27 00:00:00','    urlLink it looks like he put someone else\'s money where his mouth is .  did this town (bartlett) get named after the \"west wing\" president, just as truth or consequences nm got its name from pop culture.  is this new england\'s answer to joe, montana?   '),(12536,3196705,'27,May,2004','2004-05-27 00:00:00','    i am currently enjoying a mix of root beer and irish cream.  i know i\'m not the first person to do it, so i look up the name of the drink.  it\'s a baileyfloat.  but i don\'t drink bailey\'s.  i go carlolans or bushmills.  has anyone heard another name for this drink?   '),(12537,3196705,'27,May,2004','2004-05-27 00:00:00','    i like this movie.  everyone else i\'ve talked to about hates it.  sure it isn\'t the best story, but it treats the source material like literature, which hasn\'t really happened with other comic book movies.  maybe i\'m just too much of an ang lee fan.   '),(12538,3196705,'27,May,2004','2004-05-27 00:00:00','    1.  can you point out where this bias is please?  yes, some outlets skew left.  others skew right.  where are examples of true bias that aren\'t refuted in other mainstream media.  i\'ll look at them all.  i want to be proven wrong.  it seems to me that we buy it because we here it so much.   urlLink here is a study  that shows that in many ways the \"liberal\" media is in fact conservative.  just watch the cycles.  who ever is on top gets the lion\'s share of the negative press.  why?  it sells.  2.  as for the off-topic comment on cnn and fox news, have you tried to watch cnn?  the production values are low, and their people don\'t shout and put people down as much as fox news.  america likes to be entertained, and that is what fox news does.  here\'s a question:  what is more popular, a documentary on, say, anything, or reality programming?  what is more popular on-line:  intellectual debates like this one, or porn?  just because it\'s popular doesn\'t mean it\'s better.  3.  i\'ll give you that fox news is a better watch:  they do a better job, period.  if cnn could match production values, then maybe they\'d be even in ratings.  i\'m an msnbc man myself.  4.  as for the new york times, they know the prison scandal will sell their papers, so thats what they put on the cover.  if kerry was found out as gay, or a junkie, or to be cheating on his wife, what do you think the story one page one would be?   '),(12539,3196705,'27,May,2004','2004-05-27 00:00:00','    that the new york times is covering the prison scandal ad nauseum?  remember oj?  the 2000 election fiasco?  they put up what sells.  it\'s not anti-bush; it\'s pro-profit.  were we sick of oj?  yes.  did people follow it everyday?  yes.  they go for the buck, just like everyone else.   '),(12540,3196705,'27,May,2004','2004-05-27 00:00:00','    wouldn\'t  urlLink this guy  make a fun new character on the oc?   '),(12541,3196705,'26,May,2004','2004-05-26 00:00:00','    look who\'s only 3.5 back.  i swear, as bad as they looked in april, they are still way in it.     oh, and i hate the lakers.  kobe is freaking me out.  one would think these court dates would take their toll at some point   '),(12542,3196705,'26,May,2004','2004-05-26 00:00:00','    urlLink this sounds fine, but was it a problem ?   '),(12543,3196705,'26,May,2004','2004-05-26 00:00:00','    In reality there is the rare occasion that I  don\'t  hate the ACLU.   urlLink THIS  is not one of them.   '),(12544,3196705,'26,May,2004','2004-05-26 00:00:00','    my vote is  urlLink time changer .  its a christian movie that claims the root of all evil is morality without god.  it was just plain bad.   '),(12545,3196705,'25,May,2004','2004-05-25 00:00:00','    sure they let the pacers get within sight of the lead at the end, but they dominated overall, holding a double digit lead most of the way.  i gotta say it:  i love rip.  let\'s hope they can put the pacers on the ropes with another great game in game 4.   '),(12546,3196705,'25,May,2004','2004-05-25 00:00:00','    it there anything dumber than the kfc \"ladies day\" commercial?  it really bugs me.   '),(12547,3196705,'25,May,2004','2004-05-25 00:00:00','    is anyone looking forward to this movie?  it looks ok at best, yet it is opening on 5 screens at the theatre near me.  what gives?   '),(12548,3196705,'25,May,2004','2004-05-25 00:00:00','    On a followup to the \"White Chicks\" thought...  Anybody looking forward to the new sitcom, \"Method and Red\", due out on Fox in a couple weeks?  Didn\'t their movie flop?  What makes them think that a sitcom will do any better?   '),(12549,3196705,'25,May,2004','2004-05-25 00:00:00','    what are the odds that \"white chicks\" will be the worst movie ever?  i\'d say its even.   '),(12550,3196705,'25,May,2004','2004-05-25 00:00:00','    urlLink a school prank that involves manure, dead animals and animal blood, and only mentions manure in the title ?  aren\'t dead animals a little worse?  seeing as the media usually goes for sensationalism over, say, facts, this is just plain odd.   '),(12551,3196705,'25,May,2004','2004-05-25 00:00:00','    Does anybody understand Rugby?  It looks fun simply because it looks like one of those games you would play when you were little and you made up the rules as you went along.   '),(12552,3196705,'25,May,2004','2004-05-25 00:00:00','    one thing i want to clear up:  i don\'t really like the democrats.  i just like them a lot more than this current batch of republicans.  i think the two main parties are inherently flawed in their practices and they both concede too much to their financial backers.  i know i come off as a demo because they tout support for many things i believe in, and i\'m outspoken about my dislike for republicans.  that doesn\'t mean they really speak for me.  i\'m the worst kind of independent:  i sound a lot like one of the main parties, but do not align myself with any third party.  i am merely a liberal.   '),(12553,3196705,'25,May,2004','2004-05-25 00:00:00','    urlLink a member of ringling brothers and barnum and bailey\'s circus has been arrested on child porn charges .  yet another reason not to like clowns or circuses.  at the last circus i went to, i witnessed two deaths, so i\'m biased against them.  vulgar is a reference to the movie \"vulgar\" from view askew.  it\'s not for everyone, but it\'s good if you can take it.   '),(12554,3196705,'25,May,2004','2004-05-25 00:00:00','    i just watched every family guy episode (i have a lot of time on my hands right now) and i have to say they have some real hate for the forgettable \"dharma and greg.\"  i never saw a whole episode, but it never seemed any worse than any other show.  why the hate?   '),(12555,3196705,'25,May,2004','2004-05-25 00:00:00','    urlLink i am all for getting rid of dirty bombs in an attempt to keep them away from terrorists .  the last thing we need are unstable people with \"the bomb.\"  i am also for greatly reducing our weapon stockpiles.  i think we can rest assured with  less but better weapons.  do we really need  urlLink 10,000 nuclear warheads ?  how about we destroy some of the weapons we don\'t need?   '),(12556,3196705,'25,May,2004','2004-05-25 00:00:00','    i just learned chris farley was the original voice of shrek.  too bad they didn\'t finish his dialogue so this could be his last movie.  yes i know, dirty work was his real last movie, but he is forever saddled with the memory of the very forgettable \"almost heroes\" being his swan song.  just why didn\'t that movie work?  a talented director: christopher guest, a great physical comedian: chris farley, a good strait man: matthew perry (as evidenced in \"the whole nine yards\"), and a great character actor: eugene levy.  yet it is pure crap.   '),(12557,3196705,'25,May,2004','2004-05-25 00:00:00','    for topics to write about.  don\'t worry, i\'m not running out of things, i\'m just wondering what people want to hear about leave your ideas with me via email or the comment button below.   '),(12558,3196705,'25,May,2004','2004-05-25 00:00:00','    the pistons are ready to host a block party tonight, or at least i hope so.  while low scoring games are usually a nightmare to watch, game 2 was a delight.  there is nothing better than a game when one team dominates the other at one end of the floor.  i hope the pistons can hold them to 60 tonight.   '),(12559,3196705,'25,May,2004','2004-05-25 00:00:00','    urlLink hofstra booed doctorow for his anti-bush speech .  who will be next?   '),(12560,3196705,'25,May,2004','2004-05-25 00:00:00','    urlLink i don\'t have anything to add to this .   '),(12561,3196705,'25,May,2004','2004-05-25 00:00:00','    bonds went deep, leaving him 20 behind his mates.  if he gets hot again, he\'ll catch him in a month.  bonds    11 the rest 31   '),(12562,3196705,'25,May,2004','2004-05-25 00:00:00','    i caught my foster brother pleasuring himself in the back room.  her\'s the thing:  there is a lock on the door, and he didn\'t even close said door.  that\'s just sad.   '),(12563,3196705,'25,May,2004','2004-05-25 00:00:00','    i\'m just saying.  between those two they have posted one topic.   '),(12564,3196705,'24,May,2004','2004-05-24 00:00:00','    i have one conspiracy theory that i have bought into:  the nba tries to fix games.  this belief comes mostly from my extreme hatered of the lakers.  i know i\'n irrational on this point, but now  urlLink david aldridge thinks smething fishy is going on .  you know, if i\'m right about this, then i may have to subscribe to more conspracies.   '),(12565,3196705,'24,May,2004','2004-05-24 00:00:00','    i am now receiving \"free gas\" spam.  i wonder if it will ever rival the amount of mortgage and penis enlargement spam i get.   '),(12566,3196705,'24,May,2004','2004-05-24 00:00:00','    urlLink snoop doggy dog is getting a divorce .  i honestly didn\'t know he was married.   '),(12567,3196705,'23,May,2004','2004-05-23 00:00:00','    urlLink mass. gov. romney gave the commencement speech at suffolk university and chastised the students that turned their back to him .  how did he chastise them?  ready for this?  he called them biased.  let that sink in.  in case you\'re lost about this, look  urlLink here .   '),(12568,3196705,'23,May,2004','2004-05-23 00:00:00','    urlLink in london they concieved a child from 21 year old frozen sperm .  i say we can beat that by at least a month or two.  we just have to want it   '),(12569,3196705,'23,May,2004','2004-05-23 00:00:00','    urlLink the feds have issued subpoenas to tim russert and matthew cooper  in their efforts to se if the administration improperly leaked info on valerie plame.  how about asking robert novak (the guy who broke the story) and the administration instead of putting the media on trial.   '),(12570,3196705,'23,May,2004','2004-05-23 00:00:00','    well, i was visiting a blog i\'m not going to mention at this time when i got in an argument about when i thought it was right to kill someone.  i said that it is only right when that person gives you permission.  well wasn\'t i surprised when my seemingly innocent comment was removed.  the funny part was that i said much more damning things in other posts, even going as far as backing the side the creators of the sight were against.  i\'m just saying that if you\'re going to censor your blog, you should do a better job of it.   '),(12571,3196705,'23,May,2004','2004-05-23 00:00:00','    have you ever been hit in the gut right when you\'re waking up?  me niether, but that is a lot like the giants\' situation right now.  they finally get durham in the line up, fonzie starts hitting, and bonds might go on the dl.  the good news is they are only 4.5 back.   '),(12572,3196705,'23,May,2004','2004-05-23 00:00:00','    he\'s happy sherk 2 (i give it an A) is the fastest grossing movie ever.  he\'s happy because it is a movie that doesn\'t celebrate violence.  funny, he had no problem with  urlLink this movie .   '),(12573,3196705,'23,May,2004','2004-05-23 00:00:00','    urlLink it looks like everyone will leave bush if he wins .  how about selecting me?  i\'ll bring a different viewpoint, but will know who signs my checks.  i promise not to be a big headache until i resign and get a book deal.  if he makes me attorney general, i\'ll make up silly songs, too.   '),(12574,3196705,'23,May,2004','2004-05-23 00:00:00','    wolves win.  i guess espn, tnt, and fox sports called that series a little early.   '),(12575,3196705,'23,May,2004','2004-05-23 00:00:00','    what should i do?  living in a state that is not up for grabs (texas) has left me with a dilemma.  i don\'t want bush back, but i\'m not salivating over kerry, either.  should i go third party, vote to keep bush out (vote for kerry), or stay home?  my main goal is to oust bush, but i\'m not in a power position at all here.  what to do?   '),(12576,3196705,'23,May,2004','2004-05-23 00:00:00','    i was looking at the  urlLink baseball power rankings , and they don\'t comment on every team now!  come on, 300 comments a week is too much for them?  i\'ll do it for half of whatever the current guy gets.  this is ridiculous.   '),(12577,3196705,'23,May,2004','2004-05-23 00:00:00','    for  urlLink continually baiting an israeli zealot  with pro-palestinian arguments?  i am the one posting anonymously.  yes, i am afraid of posting contrarian commentary on a religious zealot\'s website.   '),(12578,3196705,'05,June,2004','2004-06-05 00:00:00','    urlLink Creed broke up   '),(12579,3196705,'04,June,2004','2004-06-04 00:00:00','    urlLink Story here .  America lost a great president and a fine man today.   '),(12580,3196705,'04,June,2004','2004-06-04 00:00:00','    Following C\'s example, my top 5 movies (in no particular order) are:  High Fidelity Suicide Kings The Blues Brothers Snatch American History X  And a few older movies that deserve the honorable mention:  The Manchurian Candidate 12 Angry Men Gideon\'s Trumpet    Favorite bands:  Stone Temple Pilots Foo Fighters Queens of the Stone Age Radiohead Ben Folds/Ben Folds Five (that counts as one, right?)  #6 -  Ryan Adams   Note:  These lists are subject to change at the drop of a hat.  They\'ve stayed pretty stable for quite some time now, though.    There... maybe that will satisfy Melin... a bit   '),(12581,3196705,'03,June,2004','2004-06-03 00:00:00','    Anybody else out there against the death penalty?  Anybody against it because they think it\'s too soft on crime?  I mean, c\'mon... death is such a cop-out.  Put me in charge and we\'ll bring back the Russian gulag.  There\'s plenty of wide open area in Alaska just begging for some death row inmates to come and break up the millions of rocks into tiny bits of gravel.  I don\'t know about you, but a life of hard labor doesn\'t sound too appealing.   '),(12582,3196705,'03,June,2004','2004-06-03 00:00:00','    Tonight I watched  urlLink Finder\'s Fee .  Good flick.  Fun twists.  Quasi all-star cast.  Actually, it gave me a  urlLink Suicide Kings  feel... just take out Walken and replace him with James Earl Jones  ...but don\'t make him a gangster.  ...and don\'t amputate any fingers.  ...and nobody\'s sister got kidnapped.  ...and throw in that dude from  urlLink Van Wilder .   '),(12583,3196705,'03,June,2004','2004-06-03 00:00:00','    i\'ll be gone until sunday night to the big apple, hopefully getting a job in the inner city, so i won\'t be updating this weekend.  maybe one of my lazy contributors will drop a few words.   '),(12584,3196705,'03,June,2004','2004-06-03 00:00:00','    i have to say that i disagree on the aclu\'s position on this  urlLink case .  after looking at it, i think it is much less a religious issue and very much a historical one.  the seal represents the history of the area, and for that reason should be kept.  on a similar note, things like swastikas and the \"stars and bars\" are symbols not just of history, but of hate.  was the church a good influence in california?  that can be debated, but the are key to the history.  i think the biggest problem i had with erik was his unjustified slam at the aclu.  like them or not, they serve a purpose.   '),(12585,3196705,'02,June,2004','2004-06-02 00:00:00','    urlLink but you wouldn\'t know from this article .  it makes it sound like it is just being taken from stage to screen.  perhaps the writer never saw the film?   '),(12586,3196705,'02,June,2004','2004-06-02 00:00:00','    i have been going through my music and have found my self debating as to who i like the best:  less than jake or reel big fish? fountains of wayne or hootie and the blowfish? gin blossoms or weezer? mxpx or five iron frenzy? miles davis or coltrane? outkast or jurassic 5? pas/cal or marcy playground? tom petty or the beatles?  in the end it doesn\'t matter, i love all of these and many more.  thats what is great about music:  everyone has something they like, and almost everyone can identify with songs and feel better because of it.  that said: country sucks.   '),(12587,3196705,'02,June,2004','2004-06-02 00:00:00','    my top five, in no order:  bottle rocket chasing amy the princess bride swingers the big lebowski  yah, they are are current.  some older movies i like:  dr. strangelove cool hand luke 1776 the great escape all the president\'s men casablanca  movies: the great escape (sans motorcycle).  when the world gets crazy, just pop in a dvd or vhs or laserdisk (anything i forgot?) and let the director paint a picture of other possibilities for you.  i just hope i end up writing a good one at some point.   '),(12588,3196705,'02,June,2004','2004-06-02 00:00:00','    wow.  he has 37 IBB.  the next closest: the arizona diamond backs with 23.  that\'s right, the whole team.  bonds has 10 strikeouts and 68 walks.  he is on pace for over 200 walks and 44 hr.  has he juiced up?  if there could only be one thing responsible, then i\'d say yes.  but with new training techniques, better bats, and personal trainers, all atheletes have been getting bigger.  with bonds, the big deal is: 1. his trainer being in trouble, and 2. his being so much better than everyone.  so i say that other reasons are just as viable, and until it\'s proven, i\'ll assume he\'s not guilty.   '),(12589,3196705,'01,June,2004','2004-06-01 00:00:00','    well, these two teams split their to regular season games, both home teams winning.  the lakers even gave up 106 at detroit.  this was all before rasheed came to detroit.  so how will the series play out?  most people are guessing 4 or 5 for the lakers.  but let\'s look at this thing.  the lakers only work when shaq is killing and kobe gets going.  the wallaces\' can hang with shaq, making his life difficult.  prince is out of his mind right now on d, so put him on kobe.  hope the rest don\'t kill you, but it\'s not like everyone else in detroit are slackers on the d end of the floor.  who can stop rip?  they say kobe, but can he keep up with rip and still be effective on the other side of the floor?  the boggest problem on offense is big ben.  he won\'t do anything against shaq.  so who will win?  i\'m taking detroit in 6.  i think they\'ll slow down and beat up la, and in the end rip will be the only one left with anything in the tank.  at least i hope it happens that way.   '),(12590,3196705,'01,June,2004','2004-06-01 00:00:00','    urlLink that\'s what a baby costs in south korea .  that\'s as good a price as any.   '),(12591,3196705,'01,June,2004','2004-06-01 00:00:00','    i knew they wouldn\'t pull this game out when i saw all the hype espn had on this game.  thanks espn.  you made the d-backs feel like scrubs that had something to prove.  jerks.   '),(12592,3196705,'01,June,2004','2004-06-01 00:00:00','    i got this one right.  pistons in 6 and rip went nuts.  artest is nuts.   '),(12593,3196705,'31,May,2004','2004-05-31 00:00:00','    yes, i\'m optomistic, but even i can see they beat up on three bad teams.   '),(12594,3196705,'31,May,2004','2004-05-31 00:00:00','    urlLink warner is done with st. louis .  being a \'phins fan, i don\'t really care, but weren\'t we all sick of him and his wife?   '),(12595,3196705,'31,May,2004','2004-05-31 00:00:00','    do you think the pacers would just accept walk-ons at this point if they could?  their two best players and starting pg are banged up, so don\'t put any money on them tonight.  go pistons!   '),(12596,3196705,'31,May,2004','2004-05-31 00:00:00','    that if the prison abuse scandal is political, then so was the impeachment?  i personally don\'t see how people could argue that either would have lasted so long without sensationalism (prison photos, starr report) and the guilty party constantly complaining about the need to not focus on the mistake?  yes, they are political, but as they say, the squeaky wheel gets the grease.  the noisy whiner gets the press.   '),(12597,3196705,'31,May,2004','2004-05-31 00:00:00','    there was always something that bugged me about this movie.  well, aside from being a little silly and way formulaic, i hate that train going through the outfield.  no league would allow it, but no one, not even the yankees seemed out of sorts about it.   '),(12598,3196705,'31,May,2004','2004-05-31 00:00:00','    what are core reasons people have been against this guy?  too liberal?  flip-flops?  caring too much about foriegn opinions?   urlLink no good reason ?   urlLink i know my reason , but i don\'t think many people agree with me.  they should, though.   '),(12599,3196705,'31,May,2004','2004-05-31 00:00:00','    i am tired of being lumped in with everyone in my state as a republican.  the texas republicans called and asked me to give them money.  their rationale?  i sounded like a republican, and of course i\'m a texan.  just because i live in this state, i must support the \"red state\" candidate.  i feel bad for the people like me that don\'t vote with their surroundings.  i am sure there is a \"blue state\" candidate many people would rather not be associated with, but are due to geography.  i just feel that my vote has lost some meaning considering there is no chance anyone but bush will win here.  does this same feeling occur  in those states that are predestined for kerry?  i can vote to show my support to the causes i believe in, but it won\'t change anything in this election.  face it:  i could if i could get my whole city (el paso) to vote for, say, local legend don haskins, it would not impact the states electoral college votes going to bush.  i might try to do it, anyway.   '),(12600,3196705,'31,May,2004','2004-05-31 00:00:00','    the streak is at 10.  bonds has 14 dingers, 34 behind the others.  now that the offence looks like it belopngs in the majors, he may not catch them.   '),(12601,3196705,'31,May,2004','2004-05-31 00:00:00','    urlLink i was wrong .  did i mention i hate the lakers?  now that is bias!   '),(12602,3196705,'31,May,2004','2004-05-31 00:00:00','    now they just need the kid from \"powder\" to make  urlLink this display complete .   '),(12603,3196705,'30,May,2004','2004-05-30 00:00:00','    today we remember the sacrifice our men in uniform have made for this country over the years.  our country was built on the backs of these men and women.  i thank every person to have worn our uniform with dignity and honor.  for me, i especially thank my best friend who served a hitch in iraq and got home safe.   urlLink rob of \"the mind of rob\" has an excerpt from a letter penned by a decorated vet .   '),(12604,3196705,'30,May,2004','2004-05-30 00:00:00','    i have to admit that it really bugs me that when addressing the public the administration lumps together operation iraqi freedom and operation enduring freedom (formerly infinite justice)?  they are two different campaigns with different enemies and objectives.  it preys on the inability of some to understand that not all arabs are the same.   '),(12605,3196705,'30,May,2004','2004-05-30 00:00:00','    you have to love rip.  if they don\'t learn how to stop him, this series is over.   '),(12606,3196705,'30,May,2004','2004-05-30 00:00:00','    they are finally over .500.  if they can stay there, they\'ll be in the race all year.   '),(12607,3196705,'30,May,2004','2004-05-30 00:00:00','    in the realm of entertainment is there anything worse?   '),(12608,3196705,'30,May,2004','2004-05-30 00:00:00','    urlLink this story could qualify .  should corporate america be up front about their earnings?  yes.  are the problems in the porn industry their fault.  no.   '),(12609,3196705,'30,May,2004','2004-05-30 00:00:00','    why might bush not want you to see this?  because they made gore the president.  this movie was fun.  i enjoyed myself, but mainly because of how ridiculous it was.  and the special effects.  and emmy rossum.  was it good, though?  of course not.  it was a bad big summer movie.  do i suggest it?  only if you watch movies with a grain of salt.  on a personal note, it was fun to see the local channel 7 anchor man in his cameo at the border.  i would talk about all the things that bugged me about this movie, but i don\'t want to basically recreate the whole movie here.  grade C summer movie grade B   '),(12610,3196705,'30,May,2004','2004-05-30 00:00:00','    urlLink he is backing the packers .  i am not anti-packers, and in fact am anti-patriots (the team, not actual patriots).  i just don\'t think you should betray your team for any reason, especially a silly one.  if i was ever in the witness protection program, you could find me by looking for a blogger that talks ad infinitum about the sf giants and miami dolphins.  yeah, i\'m not voting kerry now.  maybe i\'ll write in for the liberal slayer.   '),(12611,3196705,'10,June,2004','2004-06-10 00:00:00','    i\'m attending the wedding of some college friends this weekend, so i won\'t be updating until sunday or possibly monday.  hopefully erik or even \"d. darko\" will pull their weight.  peace.   '),(12612,3196705,'09,June,2004','2004-06-09 00:00:00','    wow.  i thought the pistons would win, but a twenty point blowout?  they are playing about as well as is possible.  i can\'t wait to start hearing excuses for the lakers after this total domination of them.   '),(12613,3196705,'09,June,2004','2004-06-09 00:00:00','    urlLink ray charles died today .  we have lost a national treasure.   '),(12614,3196705,'09,June,2004','2004-06-09 00:00:00','    they lose their series with the d-rays.  they\'re just lucky the majors aren\'t run like the premier league, or they would be replaced after such a poor performance.   '),(12615,3196705,'09,June,2004','2004-06-09 00:00:00','    urlLink a temporary restraining order that sounds less than kosher, wording that is strait-forward, yet vague at the same time .  this case has movie-of-the-week written all over it.  my take:  just let them get married already.   '),(12616,3196705,'09,June,2004','2004-06-09 00:00:00','    it was behaving oddly, and you can click on the time to get a permanent link if you really want it.   '),(12617,3196705,'08,June,2004','2004-06-08 00:00:00','    he used to be wimp, and now he could kick my ass.  this is distressing.   '),(12618,3196705,'08,June,2004','2004-06-08 00:00:00','    i have accepted a job with a school district here in el paso.  it came down to what i was teaching.  (social studies in el paso or math in new york)  i am not forsaking the idea of going to new york next year, though.   '),(12619,3196705,'08,June,2004','2004-06-08 00:00:00','    urlLink they just signed knight to an extension .  good, he\'s the best thing that happened to tech.   '),(12620,3196705,'08,June,2004','2004-06-08 00:00:00','    sure the lakers pulled out the win, but don\'t act like the pistons are in trouble now.  they came back in that game to take the lead near the end, then didn\'t foul shaq.  they came that close to being up 2-0.  they still took the home court away, so watch out.  the pistons are playing great team ball, don\'t be shocked when they take two of three at detroit and win in six.   '),(12621,3196705,'08,June,2004','2004-06-08 00:00:00','    i wish the loud texan had made it through.  he was loud, dumb, and made light of the things i don\'t like about texas.   '),(12622,3196705,'07,June,2004','2004-06-07 00:00:00','    they beat the d-rays.  but, really, if they didn\'t, why not just fold the franchise?  this was the d-rays chance to see the same giants that spurned then about a decade ago.   '),(12623,3196705,'07,June,2004','2004-06-07 00:00:00','    espn is focusing on the  urlLink wrong part of parcell\'s offending quote .  calling asians \"orientals\" is very insulting to some.  as a friend of mine put it, people are asian, rugs and vases are oriental.  sure the \"jap plays\" line was bad, but the oriental line was worse.   '),(12624,3196705,'07,June,2004','2004-06-07 00:00:00','    he compared ww2 to our current problems in iraq in an attack on michael moore.  this really bugs me.  do we really see everything in terms of ww2, whether it is an apt analogy or not?  do we blame the history channel?  is it the only war idiots like quinn understand?  and finally, shouldn\'t the host of a comedy show be at least a little funny?   '),(12625,3196705,'07,June,2004','2004-06-07 00:00:00','    a day of honoring reagan was both appropriate and very touching, and it all was non-partisan for the most part.  that is until i watched the o\'reily factor.  he had the gaul to compare the current prez. bush to reagan.  the \"great communicator\" to the guy who doesn\'t mind getting \"misunderestimated.\"  the man who helped set up the ussr for its fall, taking out our last great enemy, to the man that tried to make us all fear for our lives from an evil dictator in iraq.  these men may have had the same basic goals, but it is unfair to reagan to compare him to bush.  give the gipper a little respect.   '),(12626,3196705,'07,June,2004','2004-06-07 00:00:00','    rip has an ok game, billups does the offensive heavy lifting, and the lakers forget shaq plays for them.  the detroit pistons, with emphasis on \"d\" are in this series more than any laker fan want to admit.   '),(12627,3196705,'07,June,2004','2004-06-07 00:00:00','    well, i\'m back.  i\'ll know if i have a job in two weeks, as they need to let the school year end first.  the real fun is the trip, so here were the highlights:  the trip going there was uneventful.  once in nyc, i caught a taxi to hotel 17, which is on 225 east 17th street.  only i was dropped off at 25 east 17th street.  at midnight.  after 5 hours on a plane.    to explain hotel 17, let me tell you a story:  my freshman year of high school, my grandfather took me to sweden to see where our family had come from.  we stayed one night in a 19th century prison that was converted into a hotel.  those rooms were bigger than hotel 17.  when i was leaving, i had made it to la guardia three and a half hours early.  good thing.  i arrived to find that american airlines computers were down.  the consequences of which were three-fold, all information for people without tickets or itineraries was lost, the baggage checking system was shut down, and the system they use to identify people to give a terrorist-level search to was inoperable.  this caused a line a fourth of a mile to wrap around the inside of la guardia.  and i\'m not exaggerating.  after about 70 minutes a woman came up and asked if i was checking any thing.  i said no and produced an itinerary, so i was able to skip to the line waiting to be searched.  60 minutes later, i was searched, and on my way to the first of 3 gates my plane was supposed to land at.  but i had it better than a group of elderly people that were headed to florida:  they had to change gates to one in a different section of the airport.  this meant another security check.  luckily, they had someone escort them past the security check.  i hope they weren\'t terrorists.  so now i\'m on the plane, and on time!  but a 90 minute delay takes my optimism and spits on it.  i only had a 70 minute lay over in chicago.  we landed in chicago with three minutes until i had to be at my plane.  i look up and see that the plane to el paso is on time.  the gate i had to get to was about an eighth of a mile from where i had landed, so i book it over there.  and the plane is delayed.  and delayed, and delayed again.  so i call up a friend in chicago to say hi.  about a minute into the call, my flight gets moved up to one of the earlier delay times, and i land in el paso twenty minutes later than i had planned on.  that was my (latest)trip to the city that never sleeps.   '),(12628,3196705,'19,June,2004','2004-06-19 00:00:00','    i\'m watching political commercials, and i notice a pattern:  so far roughly 9/10 of bush\'s commercials are negative, while none of kerry\'s are.  on the flip side, kerry is still yet to make a stand in his commercials.  my question: is this the case everywhere, or just the southwest?   '),(12629,3196705,'17,June,2004','2004-06-17 00:00:00','    this weekend i\'ll be watching my nephew as my sister gets to watch dave chappelle live.  at least i only have to drive four hours this time.  i\'m a good brother.   '),(12630,3196705,'16,June,2004','2004-06-16 00:00:00','    you tend to admonish things in other people that you yourself are guilty of.  example:  i constantly accuse nick from  urlLink net politik  and  urlLink rants of the conservative mind  of being too verbose and needlessly dense.  these are my faults when speaking in public to a group of any size.  it is so easy to throw around intellectualisms to wow people that the sensibility of being conciseness is lost to hubris.  i just believe that i\'d do better to make sure the message got across than to look good doing it.   '),(12631,3196705,'16,June,2004','2004-06-16 00:00:00','    i must say that i never really watched buffy the vampire slayer or the spin-off angel, but having seen the reruns on fx and tnt, i have to say that these shows are really good.  i really missed the boat when these were on.   '),(12632,3196705,'16,June,2004','2004-06-16 00:00:00','    after a day of cleaning up after a toilet overflow and then debating over at  urlLink rants of the conservative mind , i check the scores and see that gem of a score.  it is enough to make me feel good about the whole day, toilet and all.  the giants have moved into second place, and bonds might break his own intention walk record by next tuesday.  if they are playing to potential (with durham \"healthy\" for the first time this year, lets hope) then they have a real chance.  the only thing bugging me?  rooting for the yankees this weekend.  (they are playing the dodgers)   '),(12633,3196705,'16,June,2004','2004-06-16 00:00:00','    urlLink another week, another story on reagan .  pay close attention as she slams reagan\'s son for not knowing his father.   '),(12634,3196705,'16,June,2004','2004-06-16 00:00:00','    urlLink i saw this incident live on television .  i have to say that it might have been the worst case of selfishness i\'ve ever seen.  would i like a foul ball?  yes.  would i steal it from a kid?  no way.  i\'m glad the kid got the hook up.   '),(12635,3196705,'15,June,2004','2004-06-15 00:00:00','    i was wrong.  i thought they would need six games.  they didn\'t.  the pistons throttled the lake show, winning by 13.  it felt like they won by 50.  i have never been so happy about a team winning that wasn\'t one of my favorites.   '),(12636,3196705,'14,June,2004','2004-06-14 00:00:00','    urlLink so some people are afraid that breasts will destroy america .  you know what?  if you don\'t want your kids seeing it, then don\'t let them.  but don\'t pretend seeing breasts will warp them.   '),(12637,3196705,'14,June,2004','2004-06-14 00:00:00','    urlLink it looks like politics might get a little uglier .  leave it to a texan to start a feud.  tact and playing nice are what texans tend to pride themselves on (well, that and a lot of things), but petty bickering has been the norm here forever.  my favorite dick move is how every five years or so the state senate makes a joke about kicking el paso out of the state.  jerks.   '),(12638,3196705,'14,June,2004','2004-06-14 00:00:00','    i never cared for his columns, but i enjoyed his television work.  i appreciated his take on sports, as it was almost always inflammatory yet grounded in some sort of hidden truth.   urlLink 52 is much too young to leave this world .   '),(12639,3196705,'13,June,2004','2004-06-13 00:00:00','    quickly on the nba:  the pistons will sow this up in 5 if they keep up the energy on the defensive end.  the only concern they should have is a monster game from kobe, but i don\'t think that that would be enough to take down detroit.  on to the giants: when they picked up ponson last year i was optimistic.  he was passable in the season, but was pounded by the fish.  well, he leaves to go back to baltimore and then the giants come to town.  what does sindney do?   urlLink he calls out barry bonds .  if only every pitcher was this stupid.  of course bonds goes deep on him.  they are now a game and a half back with the blue jays coming to town.  i just hope they win this series (or even sweep) before the red sox with wakefield and pedro come to town this weekend.   '),(12640,3196705,'13,June,2004','2004-06-13 00:00:00','    urlLink a doctor suggested that lawyers against limits paid in damages be denied medical care except in emergencies .  while he was shouted down by his peers, i find this distressing.  i like the idea of the doctor helping whoever comes through the door.   '),(12641,3196705,'13,June,2004','2004-06-13 00:00:00','    urlLink so coulter takes liberals to task for speaking fondly of reagan .  never mind that if they had spoken against him, then those liberals would have been heartless bastards that poke ill of the dead.  i have a problem with the idea that you have to hate people that disagree with you.  sure, i hate people.  but i do not base it on just part of their personality.  i have many conservative friends, and many friends that think everything but my taste in movies is sub-par.  yet we are friends.  the lesson:  you can like people you don\'t agree with, and failing that, you can say something nice if they die.  accept for chuck.  i wouldn\'t say anything nice about chuck.   '),(12642,3196705,'26,June,2004','2004-06-26 00:00:00','    urlLink how about naming him \"spike\" or \"fox\" or \"lifetime\" instead ?   '),(12643,3196705,'25,June,2004','2004-06-25 00:00:00','    urlLink it costs britons the equivalent cost of two glasses of milk a day to care for the royal family .  so here\'s the plan.  make me american royalty for the low low cost of a dollar a day from each american.  for the cost of a king sized candy bar, i\'ll serve the worthless role of monarchy in a representative democracy.  i\'ll even challenge their royal family to a wrestling match.  it\'ll be great.  so, just tell your friends to send $1 to me, king charles melin.   '),(12644,3196705,'25,June,2004','2004-06-25 00:00:00','    let\'s start with the worst part:  the audience.  i hate audiences that feel a need to interact with the move.  well, this audience was worse than a trekkie at the latest star trek film.  it was almost insufferable.  they cheered when moore\'s name came up, booed bush the first time he was on screen, and yelled \"yeah\" so many times i wondered if a group could catch tourette\'s from those around them.  the booed when the film cut out a few seconds before the credits, then gave a standing ovation to the screen when it finished.  i hate these types of people.  as for the movie, the grade first: A-  ok, moore\'s unfair attacks:    1. the eggs thrown at bush when he was elected. -sure, it was shady and likely should have gone the other way, but don\'t give time to nuts that throw eggs.  2. that bush was on vacation 40% of the time in his first eight months. -he was away from washington, so it looked like he was on vacation.  i\'ll give bush the benefit of the doubt and assume he did some work in his 96 days away from the white house.  3. that it was bush personally that allowed the bin laden family to fly. -this issue is all over the news today, so go find out the latest.  it does seem odd, however, that the fbi would let them go  4. bush as deserter. -i think we\'ve all agreed that deserter is a step too far.  5. bush\'s oil company going under. -have you been to west texas?  it was rather dry to begin with and has been pumped beyond all reason.  i\'m not shocked a company would fail.  6. asking if bush was wondering what would be best for the saudis, instead of america. -this was mostly baseless and totally unfair.  7. the look at the war. -while he did a great job showing what was really happening in iraq, he did not mention how bad saddam was.  this was being unfair by omission.  8. giving the lady outside the white house on screen time. -not really unfair.  i just don\'t like her.  ****** highlights:  1. highlighting the disenfranchisement that occurred in the 2000 election. -this story was buried under chad-mania.  2. the choice not to show the planes hitting  3. the video of bush as he heard about the planes hitting. -i looked more concerned.  4. examining the interference run by the white house at dissuade commissions from looking at 9/11.  5. the bonanza scene  6. look at the terror alerts.  7. small town america being warned to fear terrorists.  8. the patriot act and the wacky stories about it - peacefresno was funny  9. oregan state police  10.  powell and rice stating in 2001 that saddam didn\'t have wmds.  11. rumsfeld calling the coalition of the willing the mother of all coalitions  12. looking again at \"mission accomplished\"  13. the list that the recruiters were keeping of people who did not want to sign up. -what was up with that?  that said, i think this movie is important, and you should see it.  just keep an open yet skeptical mind.   '),(12645,3196705,'24,June,2004','2004-06-24 00:00:00','    i am sick of all the france bashing our country does.  so here is a task:  look at our relations with france from our inception until now.  really, go look.  when you see all of our history together, i then want you to give an honest opinion of our relationship.   '),(12646,3196705,'23,June,2004','2004-06-23 00:00:00','    this guy bugs me.  it\'s the hypocrisy.  two recent examples?  calling for people to stay out of the private lives of politicians.  (remember clinton?  do ya joe?)  and the attack on moore\'s movie, after complaining for about three weeks strait about how the passion was being attacked.  let\'s at least be consistent.   '),(12647,3196705,'23,June,2004','2004-06-23 00:00:00','    urlLink so cheney drops the f-bomb .  like the senate hasn\'t heard it before.  i say they should curse in public addresses.   '),(12648,3196705,'23,June,2004','2004-06-23 00:00:00','    first nba:  tech\'s emmitt is headed to the griz.  i think he\'ll be a good role player.  if he can create his own shot he\'ll stick around.  tennis:  i really don\'t follow it, but venus got screwed  giants:  they sweep the hated dodgers and open up a 2.5 game lead.  i love this stuff.  baseball:  so the astros get beltran?  it won\'t matter.  dotel, however, will greatly help the a\'s.   '),(12649,3196705,'23,June,2004','2004-06-23 00:00:00','    so they pick up stackhouse and laettner, as well as the 5th pick.  seeing as jamison was the odd man out for a lot of the year, i guess this makes sense.  that is assuming, of course, that they make another deal.   '),(12650,3196705,'22,June,2004','2004-06-22 00:00:00','    grade: A-  this movie was a modern fairy tale.  what it loses in plausibility, it makes up for in whimsy.  SPOILER  the fountain was a bit much. END SPOILER  catch it.  you won\'t regret it.   '),(12651,3196705,'22,June,2004','2004-06-22 00:00:00','    grade: B-  this movie was ok.  totally predictable, stilted acting, but fun for what it was.  wait for the cheap theatre.   '),(12652,3196705,'22,June,2004','2004-06-22 00:00:00','    first the grade: B  this movie delivered all the laughs you\'d expect from a movie with stiller and vaughn.  highlights: a pirate, lance (i beat cancer) armstrong, and chuck (walker) norris.  see it if you enjoyed the casts\' other movies.   '),(12653,3196705,'22,June,2004','2004-06-22 00:00:00','    he is about as funny as SIDS.   '),(12654,3196705,'22,June,2004','2004-06-22 00:00:00','    my feelings on him are pretty evident, but i must say that i\'m with him on battling AIDS.   '),(12655,3196705,'22,June,2004','2004-06-22 00:00:00','    The premier hasn\'t even aired yet and I\'m already sick of this damn show.   '),(12656,3196705,'21,June,2004','2004-06-21 00:00:00','    they can take over the top spot with a win.   '),(12657,3196705,'21,June,2004','2004-06-21 00:00:00','    urlLink the only shocking thing is that it\'s only about one of them .   '),(12658,3196705,'21,June,2004','2004-06-21 00:00:00','    urlLink cheating through cell phones is a big problem in some places, so why not block the signals ?  i don\'t think a high schooler or younger needs a cell phone in school anyway, but this makes sure there is no foul play.   '),(12659,3196705,'20,June,2004','2004-06-20 00:00:00','    a pet peeve of mine is when movies are lambasted for paling in comparison to books they are based on.  i think we should judge each as separate entities due to the extreme differences in the art forms.  this isn\'t to say that i on average enjoy translations when i am familiar with the material.  for instance, primal fear becomes unwatchable if you\'ve read the book.  if you haven\'t it\'s a fine movie.  that\'s all i\'m trying to say.   '),(12660,3196705,'20,June,2004','2004-06-20 00:00:00','    urlLink this story blew me away .  the question (at least the first question) i have is this: how does this prevent gambling?   '),(12661,3196705,'20,June,2004','2004-06-20 00:00:00','    riding shmidt\'s one-hitter, the giants take two of three from the red sox and still sit a game and a half behind the hated dodgers going into their four game series starting tomorrow.  on shmidt, of course i love the results, but 133 pitches seems a tad high for the giants only good starter.  the guy has already been on the dl.  how about some kid gloves here?   '),(12662,3196705,'20,June,2004','2004-06-20 00:00:00','    urlLink these pilots were really close to getting it right .  i think we should appreciate how rare these stories are.  it is still rather funny, though.   '),(12663,3196705,'03,July,2004','2004-07-03 00:00:00','    is he a prick for the sake of being a prick?  he was talking to the founder of \'black cops against police brutality\' (a policeman himself) last night, and instead of talking about the problem of police brutality, he touched on it and then attacked the guy.  here are the talking points, with answers in bold:   why black cops and not cops against police brutality?    i created the group and i wanted to call it that.    would you help people that aren\'t black?    yes.  we helped a white guy that got shot a hundred times recently, and help people of all colors.    look at this video  (of a cop hitting a guy on the legs with a flashlight).   is that police brutality?    yes.  if you strike with the flashlight it is automatically police brutality, as we are trained not to use it as a weapon.  (pat:  i can\'t tell if it\'s a flashlight )   so why don\'t you go after the real problem of black-on-black crime?   because my group is committed to stopping police brutality.    so you don\'t think that black-on-black crime is important?    i didn\'t say that.  of course it is.  my group speaks to another problem.    so black-on-black crime isn\'t as important as police brutality?    i didn\'t say that.  they are both important problems.  my group focuses on police brutality.    well i don\'t see why you ignore black-on-black crime.  gotta go.   while not verbatim, this was a good representation of the event.   '),(12664,3196705,'02,July,2004','2004-07-02 00:00:00','    urlLink don\'t you just hate when this happens ?   '),(12665,3196705,'01,July,2004','2004-07-01 00:00:00','    urlLink cassini has made it to saturn .  after studying the ringed planet, it\'s onto some of its many moons.  there was a time when this story would have risen above all the crap we usually cover and lead the news.  unfortunately, we have lost our love affair with exploration.  The world, for the most part, is known.  there are no more continents to discover, no lost cities of gold.  space, though, represents a practically untapped arena to reach out and discover.  will the world care about taking a larger step into space without pressing reasons forcing them to?  i do not know.  i hope we realize that exploration is its own reward.  and yet every time we have ventured into the great unknown we have found riches beyond what we could have hoped for.   '),(12666,3196705,'01,July,2004','2004-07-01 00:00:00','    seeing as i live in texas, i don\'t get to watch many giants games.  yesterday, though, they were on espn2, so i got to see them rip the hated dodgers 7-1.  bonds homer was insane.  giants up 2.5 on the padres, 3.5 on the bums.   '),(12667,3196705,'30,June,2004','2004-06-30 00:00:00','    the grade: A+  -sploilers ahead-  first:  another bad audience.  two kids kept yelling and making a general rukus, a bay cried most of the time, and some morons applauded the credits.  this movie was great, much deeper than the standard summer fair.  alfred molina proves he is one amazing actor yet again, and cameos by bruce campbell and hal sparks add humor to an already funny, yet dramatic plot.  even the cheesy \"united we stand\" scene was much better than the first movie.  was it perfect?  no.  but it sure was close.  this movie shows just how good a comic book movie can be.  the title credits were particularly nice, with the first movie shown in graphic novel slides.  my main problem?  heavy-handed foreshadowing for a sequel.  go watch this movie.   '),(12668,3196705,'29,June,2004','2004-06-29 00:00:00','    they open up their series tonight at the dodgers.  these teams have split the series so far this season, both going 5-5.   urlLink a look at their head-to-head stats  shows that perhaps the dodgers should be in first, but they are in third at 3.5 back behind the padres and giants.  an interesting stat is the giants\' lead in walks at 341.  bonds has 109 of them.   '),(12669,3196705,'28,June,2004','2004-06-28 00:00:00','    urlLink we are getting to the point where the wealthy are feeling the pinch of being too computer literate .  kidding aside, this raises the question, do we need to know how to manually write?  cursive writing today could easily be limited to learning how to write your own name and read all the letters, and nothing would really change.  as far as manuscript writing, this is the only way we really need to hammer home when writing, as there will always be the need to jot things down.  let\'s ditch script writing is all i\'m saying.   '),(12670,3196705,'27,June,2004','2004-06-27 00:00:00','    urlLink fearing losing press to spider-man 2, the u.s. handed over control in iraq two days early .  well, that might not be the reason, but kudos to the new iraq.  i hope it works out.   '),(12671,3196705,'27,June,2004','2004-06-27 00:00:00','    they take the series from the a\'s, bonds goes deep, and he scores his 2,000 run.  they also hold a 3 game lead over the padres, 3.5 over the dodgers going into their trip to chavez ravine.   '),(12672,3196705,'27,June,2004','2004-06-27 00:00:00','    well, i got a letter telling me the fate of the interview i had in the big apple.  it turns out i do not meet the credit qualifications to switch to math from history.  the problem:  i called them three times because i knew texas credits didn\'t correspond directly with cuny scores.  three times i was told that i had no problems, that they checked it out and i could interview.  even at the interview i asked again, making sure i wasn\'t wasting everyone\'s time (and my money).  well, the letter explained that i was not eligible, due to the credits not corresponding.  now, at any point they could have cleared this up, but they didn\'t.  and i\'m out $500.   urlLink at least i got a story out of it .   '),(12673,3196705,'27,June,2004','2004-06-27 00:00:00','    the grade: B-  this film worked.  it didn\'t work phenomenally, but it achieved what it was after.  there really isn\'t much to say about it, though.  i did enjoy the wilson brother\'s cameo.   '),(12674,3196705,'27,June,2004','2004-06-27 00:00:00','    well, we knew that.   urlLink but did you know heinz might be worth $1 billion ?  that\'s a lot of tomato paste.  but it also gives america a choice between a billionaire\'s husband, or a millionaire (bush worth \"only\" $13 million).  i\'m glad the average american has people in tune with their needs.   '),(12675,3196705,'27,June,2004','2004-06-27 00:00:00','    urlLink why does texas attract cults ?  i guess it\'s the land you can buy for cheap and the fact we\'re not a desert wasteland. well, my part is, but most of it isn\'t.  this story makes me wonder, though, about where the line between cult and religion is.  this story is about a cult, i doubt any would argue that.  but if they get enough people, or stick long enough, will they get recognized?   '),(12676,3196705,'27,June,2004','2004-06-27 00:00:00','    urlLink so now using hitler is ok .  well, that\'s what bush\'s campaign is saying.  in an attack add on their website, bush\'s campaign uses clips of people bashing him, as well as a piece of a clip from moveon.org that featured hitler.  now, moveon.org pulled the add because they felt it was in bad taste.  i hope bush\'s campaign will do the same.   '),(12677,3196705,'10,July,2004','2004-07-10 00:00:00','    this is another one of my favorite directors.  i love all three films.  here is my ranking:  bottle rocket.  A+   i love this movie.  this just might be my favorite film of all time, and most of it has to do with dignan.  this is sublime \"dumb\" comedy that is much smarter than most films.  i tend to judge people on how they feel about this movie.  rushmore.  A  this film works the whole time and looks about as good as a movie can look.  there is nothing i don\'t like about this movie.  the royal tenenbaums.  A  this film looks great and is very well done, but seems too unique at times.  regardless, i do love this film.  an aside:  my best friend\'s mother hated this film so much she stopped it about thirty minutes in and refused to finish it.  while these are listed in chronological order, they are also in order of my preference.   '),(12678,3196705,'09,July,2004','2004-07-09 00:00:00','    today i watched all of kevin smith\'s movies, and i have to say that jersey girl is a really great film.  i rated his films  urlLink here , and while i guess i\'ll stand by there rankings, i\'d take any of his films (well, maybe not jasbsb) over almost any film out there.   '),(12679,3196705,'09,July,2004','2004-07-09 00:00:00','    not that i\'m on the fence about bush, but  urlLink bringing up a marriage amendment  again rubs me the wrong way for two reasons: the obvious homophobic nature of the bill, and using amendments willy-nilly.  but won\'t this hurt him with people that are on the fence?  i know many republicans that are not excited about the prospects of changing the constitution, so why mention it now?  i am not seeing any political upside to this move.   '),(12680,3196705,'09,July,2004','2004-07-09 00:00:00','    read the f***ing title of the game, moron!  if you buy your kid a game named slugfest or grand theft auto, then don\'t be shocked if there is violence involved.  also, there is a rating system.  use it.  above all, just think!   '),(12681,3196705,'08,July,2004','2004-07-08 00:00:00','    not only do the giants win, but bonds ties the all-time season IBB record held by bonds.  think about that:  we\'re not to the all-star break and bonds has tied the season record.  wow.   '),(12682,3196705,'08,July,2004','2004-07-08 00:00:00','    i\'ve been kicked out of the  urlLink outlaw republican  blog for basically being a big bastard.  guilty?  anyone that knows me knows the answer to that. (yes)  my only gripe:  how little orange fox (head blogger, then only blogger) kept bashing people she didn\'t agree with in petty and irrational ways, then was shocked when those that disagreed with her sounded off on her.  i can\'t count the number of times she\'d post something only to take it down when it was shown to be a weak or flat-out baseless attack.  where was the accountability?  oh well.  as long as the fox is happy, it\'s in the past as far as i go.   '),(12683,3196705,'08,July,2004','2004-07-08 00:00:00','    as much as i hate mcdonalds, i think anyone  urlLink suing them for having fatty french fries  should be drawn and quartered.  it\'s a french  fry .  what did you expect?   '),(12684,3196705,'08,July,2004','2004-07-08 00:00:00','    the grade: C  this movie doesn\'t come close to working overall, but it does have it\'s moments.  i have never been a will ferrell fan, but i have appreciated him in certain roles.  this role could have worked in a different movie, or this one if the focus had been on the rivalry with other networks instead of inner-office turmoil.  the movie worked best when supporting actors were on screen.  steve carell is by far the best of the main cast, and jack black has the funniest role in the film.  too bad he had less the four minutes screen time.  the other network anchors (vince vaughn, luke wilson, ben stiller, and tim robbins) all work better than ferrell, and even the dog comes off funnier.  my favorite part was in the outtakes when they threw in an outtake from smokey and the bandit 2, or the afternoon delight moment.  this film would be good for the dollar theatres, or maybe a rental.  a disappointment.   '),(12685,3196705,'08,July,2004','2004-07-08 00:00:00','    urlLink my personal website  gets much traffic from that great land to the north.  i have never met a single person from america that shares my surname.  could it be they all went back to sweden (or up to canada)?   '),(12686,3196705,'07,July,2004','2004-07-07 00:00:00','    the grade: B-  this film was long on action and very short on plot.  in this case it works.  the movie bastardizes the traditional tale, but tries to bring to life the origins of the traditional story.  the film is carried by the performances and a nifty scene taking place on a frozen lake.  keira knightley only helps the cause, as does her outfit later in the movie.  some questions:  what the heck are woads, and what was up with stonehenge?  if it looked interesting then catch it, but if the trailers didn\'t get you excited, then nothing else in the film with either.   '),(12687,3196705,'06,July,2004','2004-07-06 00:00:00','    kerry names his vice president ( urlLink edwards ) on george w.\'s birthday.  conspiracy? my father seems to think so. (i don\'t)   '),(12688,3196705,'05,July,2004','2004-07-05 00:00:00','    you\'d think living in a metropolis would keep you free from leopard attacks,  urlLink but you\'d be wrong .   '),(12689,3196705,'04,July,2004','2004-07-04 00:00:00','    the number of people i know in kazakhstan is now officially one.  my good friend shane pettit is there teaching english for the peace corp.  i wish him the best of luck to make a difference and hopefully avoid diarrhea.  i am hoping he will give insights into the kazakhstan way of life while he is there, and if he does, i\'ll share it here.   '),(12690,3196705,'04,July,2004','2004-07-04 00:00:00','    to america.  228 years young.   '),(12691,3196705,'15,July,2004','2004-07-15 00:00:00','    urlLink candace at the christian agnostic has some thoughts on the subject , as well as a post showing how bush contradicts jesus.   '),(12692,3196705,'13,July,2004','2004-07-13 00:00:00','    urlLink nick lewis is tackling the subject as it relates to history .  do i agree with everything he\'s saying:  not really.  but it is worth the read.  really, it is.  go read it now.  go.   '),(12693,3196705,'13,July,2004','2004-07-13 00:00:00','    do women look any hotter than when they\'re wearing glasses?  i don\'t think so.  i ran into an old high school friend and was taken aback.  it lead me to this question, though:  am i the only one from high school that let themselves go?   '),(12694,3196705,'13,July,2004','2004-07-13 00:00:00','    my dad told me one today i hadn\'t heard.  he was very proud of it, so i\'ll share it here:  the democratic party:  so full of crap they need two johns.   '),(12695,3196705,'13,July,2004','2004-07-13 00:00:00','    knowing that everyone is always on the lookout for a great movie, here are some of my favorites (in no particular order) that have not garnered much attention:  the minus man  B+  owen wilson stars as a a serial killer who fancies himself an angel of mercy.  brian cox (i love that guy!) co-stars.  it has a dreamy feel to it and is hauntingly low key.  vulgar  B  a sick, sick movie from view askew (kevin smith\'s production company) this dark film tells the story of the clown from the opening of clerks.  not for everyone, but interesting.  zero effect  B  bill pullman stars as a sherlock holmes-type trying to solve cases without interacting directly with the outside world.  ben stiller plays his representative, a watson-type.  run ronnie run  B  this movie is based on a running sketch from \"mr. show.\"  it follows the life of a man that makes a living out of being arrested on reality cop shows.  scotland, p.a.  A-  it\'s macbeth set in a 70\'s fast food joint.  the sound track is almost all bad company, and it works fabulously.   '),(12696,3196705,'12,July,2004','2004-07-12 00:00:00','    i am giving my friend in kazakhstan a crash course on baseball.  his position is that baseball is a great medium to stay in touch with america, and maybe he\'ll even teach the game to his students.  while he favors the rangers, i\'ll let that slide.  it\'s great to see baseball acting as a banner for america.   '),(12697,3196705,'11,July,2004','2004-07-11 00:00:00','    Many thanks to the  urlLink liberal slayer  for sending me  urlLink this story on the subject .   If two adults want to wed, let them. If they find a cleric to grace their nuptials, hallelujah -- although signing a contract should suffice. Clergy and laity should decide who may and may not marry in America\'s religious institutions.   i have to say that that about mirrors my views on the subject.  another point touched on in the article that has bothered me is the notion that marriage is about having babies.   As for child-rearing, same-sex-marriage opponents should explain why couples who fail or refuse to reproduce or adopt may enjoy matrimonial benefits. If \"the core purpose of marriage is to bind children to their mothers and fathers,\" as Stanley Kurtz argued July 1 on National Review Online, marriage licenses should expire after, say, five years if husbands and wives neither bear nor adopt children.   a fair point.  you should read the whole article.  let\'s look at this subject rationally and figure out what is right for america before we butcher the constitution.   '),(12698,3196705,'11,July,2004','2004-07-11 00:00:00','    urlLink this story  bothers me on two levels:  first, isn\'t this the kind of action that shows that \"the terrorists have won?\" and in the extreme this seems like an attempt to justify holding power in the face of losing an election.  think about it:  bush is set up to lose the election, then word comes of a possible attack and the vote is postponed indefinitely.   '),(12699,3196705,'11,July,2004','2004-07-11 00:00:00','    urlLink you really have to feel bad for the guy .  he had a chance to be the best ever, and now he get hurts every year like clockwork.  while this throws a damper on the 500 hr outfield, i still hope that this all-star game will highlight what a great career this man has had.  on the topic of his injuries, could it be that the reds\' medical staff dropped the ball?  he never had this many injuries in seattle.   '),(12700,3196705,'11,July,2004','2004-07-11 00:00:00','    these guys rock.  they are masters at weaving stories that a quirky yet magical.  who wouldn\'t want to live in the worlds they create?  here is my ranking of their films:  the big lebowski.  A+  i love this movie.  while it gets downright weird at times, this film has moments of pure genius, and at no point does it disappoint.  while some people think there is much too much cursing in this film, it doesn\'t bother me.  fargo.  A+  this is my favorite movie that centers around a murder plot.  from the characters to the settings this film paints a funny and poignant picture of desperation and hope.  a great, great film.  o brother, where art thou?  A+  one of the best literary translations you\'re likely to see, this film is fun and timeless.  it also makes light of texas politics, and that is always fun.  miller\'s crossing.  A  the hardest of the films, this is a great gangster film with solid performances by everyone.  a gangster film that gets slighted when naming great gangster films.  barton fink.  A-  i just love the leads.  goodman is always great with the coen brothers.  the man who wasn\'t there.  A  a great film noir turn for the brothers.  i have a soft spot for black and white films as well.  this film sticks out for being stylistically separated from other coen films.  the ladykillers.  B+  while this wasn\'t there best work, tom hanks is great and i love j.k. simmons.  this movie was fun, but isn\'t as good as most of their work.  raising arizona.  A-  i know a lot of people would put this higher.  this movie is great, but doesn\'t work for me as well as some others.  it was still a great film that really put these guys on the map with most people.  intolerable cruelty.  A-  i like this film.  it was light and fun, but it felt a little like the brother\'s were mailing it in.  it was good, but you get a feeling it could have been better.  still better than 90% of the movie out there.  hudsucker proxy.  B  definitely not for everyone.  this film was stylized within an inch of its life.  many characters came off as a bit cartoony.  despite all this, i like it.  you know, for kids.  blood simple.  B  the only thing that hurts this film is the volume of great films that came after.  a great suspense film that is all the more noteworthy considering this was their first film.   '),(12701,3196705,'23,July,2004','2004-07-23 00:00:00','    grade: A-       this movie was very good at being a spy thriller, which is good because that is what it is. while the twists were obvious, they still delivered. so did the chase scene, at the end of which i realized i was holding my breath. my only complaint is the seemingly total lack of steady-cam. while i appreciate the hand-held feel, it was done to excess. that alone nearly took it down to a B+, but in the end i had too much fun to go that way. as long as you\'re not prone to motion sickness, you should catch this film.   '),(12702,3196705,'18,July,2004','2004-07-18 00:00:00','    my extended family is in town this weekend, and to help them feel better, i attended church.  we looked at 2 kings 13.  my favorite verses of the bible are in 2 kings, so i\'ll share those with you here:     KI2 2:23  And he (elisha)  went up from thence unto Bethel: and as he was going up by the way, there came forth little children out of the city, and mocked him, and said unto him, Go up, thou bald head; go up, thou bald head.    KI2 2:24  And he turned back, and looked on them, and cursed them in the name of the LORD. And there came forth two she bears out of the wood, and tare forty and two children of them.     if i don\'t update regularly this week it will be because i\'m at in-service for the new school year.  peace.   '),(12703,3196705,'31,July,2004','2004-07-31 00:00:00','    grade: B+  i have to admit to be a fan of both the leads in this film, as well as a fan of many of the cameo-stars. this movie is a buddies-on-the-road formula film that works because it knows how far to take the joke and when to play serious for the laugh. neil patrick harris\' part was particularly nice. if you are looking for a good clean fun stoner film, harold and kumar is for you. i recommend it.   '),(12704,3196705,'28,July,2004','2004-07-28 00:00:00','    grade: B  this is yet another remake, a movie based on a screen play based on a novel no less. let\'s face it: denzel is just good. the rest of a cast does well, but knowing what was going to happen put a damper on the experience for me. it is still worth a look, so i\'ll say go see it.   '),(12705,3196705,'28,July,2004','2004-07-28 00:00:00','    urlLink it seems that the electronic age of voting has had some casualties in florida  (of course).&nbsp; while only back up data was lost, it shows that the system is not as secure as is claimed.&nbsp; also alarming is the report that touch-screens register 6 times more no votes than optical scan ballots.&nbsp; take the threat of postponing elections,&nbsp;the knowledge that we might not count all of votes due to electronic oversight, and to that disenfranchised minorities and i can\'t wait for election season!   '),(12706,3196705,'26,July,2004','2004-07-26 00:00:00','    now i can tell the kids that their  urlLink paper is a piece of shit  without getting in trouble!   '),(12707,3196705,'25,July,2004','2004-07-25 00:00:00','    the grade: B   this is a movie based on the summary of a dust sleeve of the book.&nbsp; if that.&nbsp; but as its own movie it is entertaining, and it takes some nice steps in dealing with the three laws of robotics.&nbsp; this movie should have had another name, as it is judged (fairly) by the book.&nbsp; just don\'t name a movie after a book if it isn\'t going to be anything like it, or take the family guy avenue (peter griffin presents the king and i).&nbsp; as a stand-alone flick it left me with good vibes and no regrets.&nbsp; a solid summer choice that you should give a chance if you like will smith movies (though this one is the slightest bit smarter than most of his big-budget work).   '),(12708,3196705,'25,July,2004','2004-07-25 00:00:00','    i don\'t need this right now.   urlLink ricky williams has retired from the nfl, leaving the dolphins beached . while i respect his ability to walk away from the money, i wish he wasn\'t walking away from the team. also, doesn\'t he realize i have to start work tomorrow? on that front, what is more fun than a 30 student classroom? that\'s right, a 40 student classroom!! joy.   '),(12709,3196705,'06,August,2004','2004-08-06 00:00:00','    \"Our enemies are innovative and resourceful, and so are we.  They never stop thinking about new ways to harm our country and our people, and neither do we.\"  preach it, brother man.   '),(12710,3196705,'05,August,2004','2004-08-05 00:00:00','    after polling my classrooms (8th and 10th) i have found that they, too are split on who should be the next president. they are also evenly split on gay marriage, which does surprise me. it is amazing to see how much some of these students know about the issues. my biggest goal is to get them thinking about the issues without besieging them with my views. hopefully i\'ll get them to care about politics more than my generation does.   '),(12711,3196705,'01,August,2004','2004-08-01 00:00:00','    grade: B  this movie was ok, but shayamalan seems to still be in his rut. in other words,you know that there will be a twist and that you\'ll see him in there eventually. i did like the aspect of the 1800\'s as a utopia, even though i am totally against that personally. if you like his stuff then check it out.   '),(12712,2747575,'31,July,2004','2004-07-31 00:00:00','    I watched a whole 2 movies today. The first was  urlLink Eurotrip , which was okay, in a sense. It wasn\'t that funny, but I really, really liked the music from it, which completed me to get the soundtrack (download, of course). Here are two songs that I particularity liked, and if you don\'t like them, well, too bad. I found all the songs on the soundtrack to be quite good, just like the  urlLink Elephunk  album by  urlLink Black Eyed Peas .  Link:  urlLink Lustra - Scotty Doesn\'t Know  Link:  urlLink The Salads - Get Loose   The other movie that I watched,  urlLink Taking Lives , was not that bad. It was quite interesting in some parts. For anyone who wants to know,  urlLink Angelina Jolie  appears nude in the movie, which didn\'t really make the movie any more interesting. If you have the time and the will, you can always download it though, or use one of those  urlLink free Blockbuster rental coupons  to rent it when it comes out on DVD.  Doom III is almost out. It has been  urlLink reported  that some stores in the United States are already selling the copies, illegally. Best Buy is one of them. It will probably be leaked very soon. I, for one, can wait to play it because I don\'t think I will like it very much, but I am, as is the rest of the public, hyped up about. However, only the limit on multiplayer (4 players) restricts the game drastically.  We decided we will be going to  urlLink Playdium  some time an a week or so. If anyone wants to come, drop me a message. Will says he\'s got very cheap tickets, and I can provide the ride some time after my parents come back. He also suggested that we invite some ladies - Ontario\'s biggest shopping mall is located steps away from Playdium. I say that it\'s a foolproof plan.  Oh, and here\'s a poem that Henry sent me. Interesting indeed.   A peach is a peach,     A plum is a plum,      A kiss isn\'t a kiss,     Without some tongue.      So open up your mouth,     Close your eyes,      And give your tongue,      Some exercise! '),(12713,2747575,'30,July,2004','2004-07-30 00:00:00','    Well, my parents have successfully left me for a week, all along with nothing to do. Last night was fun; we watched  urlLink Kill Bill: Vol. 2  and  urlLink The Bourne Identity . I didn\'t even watch the first  urlLink Kill Bill , but ah well, it was fun nonetheless, thanks to Rob\'s  urlLink Sony 60\" Grand WEGA TV . It allowed me to take my mind from other things.  Right now, the weather sucks, so there isn\'t much that I can do. I\'ll probably be watching movies all day while I wait for  urlLink Doom III  to leak out onto the web - shouldn\'t be more than a day from now. And then, I might call her up and see what she says - it\'s always more fun do go places with a bunch of friends, no? I would think and hope so.  I just watched  urlLink Eurotrip . It was an okay movie, but the soundtrack seems so much better, downloading it now. It has all the famous songs from different countries, it seems. And now, it\'s time for  urlLink Taking Lives , so take care all, and wish me luck for all my future endeavors, as I need it. '),(12714,2747575,'30,July,2004','2004-07-30 00:00:00','    Everything is over. Chaos is everywhere.    I have predicted my mark correctly, and ended up with an 87%. I don\'t even care about the mark. Vic is going to  urlLink Wonderland  today, as Will asked her when we were driving by, but we forgot to ask the time that she is going at, or if we could come. I would\'ve went.    Time to move on and celebrate the boredom of summer. '),(12715,2747575,'29,July,2004','2004-07-29 00:00:00','    Well, my high school, English had just ended. I just have to pick up my report card tomorrow and there will be no more English for a whole year, or maybe even more. I feel more relieved and sad at the quick ending of the course - only 18 days of school. I think I did okay on the exam, and I assume I will get around 87% overall, but that is just my guess.     Today, I did nothing. We could not figure out where to go, so we decided to go to the local shopping mall, which wasn\'t quite local actually. After chilling inside for an hour, my friend\'s car wouldn\'t start and we had to wait for about an hour while help came. Turns out that when the workers were replacing the manual transmission a day before, they didn\'t tighten some wire, which got loose and did not allow the car to start. We were actually almost about to leave the car at the parking lot.     Another highlight of the day was when me and my friend Dave (who\'s car wasn\'t starting just minutes before) blocked in another person. It was a 60 km/h road, and we decided we\'d block this person because she honked at Dave before, so we took up the 2 lanes and went at around 30 km/h. That was fun while it lasted.     Tomorrow is the last day I will see Victoria. Although this does sound lame, as I don\'t even really know her, I will miss her. Nothing will probably happen tomorrow as I am too shy to do anything. It will definitely haunt me down in the future. Well, if I really wanted to, I could always call her, but that would be both weird and wrong, for obvious reasons. As I said before, I should smell the fresh air more often. '),(12716,2747575,'28,July,2004','2004-07-28 00:00:00','    Well, our last day successfully ended today. Tomorrow is out exam, and then after, we\'re gonna go do something, however, I am still not sure where we are going and with who. We cannot seem to decide where we would want to go, and I personally am too shy to ask \"them\" to go out with us. Doubt they will even want to go anywhere. Ah well, it was fun when it lasted, although I will not give up hope yet.     In other unrelated news, my parents are leaving for a whole week starting this Saturday morning, so I will have nothing to do. I won\'t have a car either, so someone will have to pick me up and we\'ll go somewhere, or so I hope. Or I can bike somewhere, that will surely be fun.     And now, it is time to study for the English exam. I don\'t really feel like studying, so I will probably do around 15 minutes of review, and consider myself ready for the exam. Although summer school was quite easy, I am very tired from school in general, and feel very lazy at the moment.     Ps. I heard they might be going to  urlLink Caribana . '),(12717,2747575,'27,July,2004','2004-07-27 00:00:00','    Well, nothing new today, as usual. My English mark is constant at 89%, which, in my opinion, is exceptional for this subject. Uh, we didn\'t do much today in class and I was very bored. I felt like sleeping, but my group would kill me. Right now, I have so much homework that I don\'t even want to think about it. I should get started now. Here\'s a picture of a guy enjoying himself in class.     I just went to Wal-Mart to pick up some things for my presentation (candy in fact), and since it was raining, I almost spun out. I was making a turn not that quick, wanted to test the abilities of the car, and I really felt like it was loosing control. ABS might of kicked in, and the car slid a bit. It was a fun experience, but always remember: don\'t do it when there are other cars present.  Tomorrow is the last day. I feel happy and sad. School is over for a whole month, which is a good thing. However, I will probably not see Victoria again after tomorrow (well, maybe for short periods of time on Thursday and Friday). I am extremely mad at myself for that. It is time to get out more often and smell the roses. Will did grab a picture of her today, so, if he sends it to me (and I get his permission), I might show it to whoever reads this blog. She is really sexy. '),(12718,2747575,'26,July,2004','2004-07-26 00:00:00','    Well, it looks like I just finished my individual project - my big CPT. The whole project is 58 pages, with 17 of those pages being completely typed from scratch by me, which is 8436 words of typing. I feel very honoured right now, especially knowing that I have a Unit Test on  urlLink Joan of Arc  tomorrow, and that my homework night tomorrow will probably be much more packed due to the group presentation, which is due on Wednesday.     Anyway, away from all the homework stuff: it is becoming too depressing for me to handle. Nothing really happened today. Will and me are still trying to figure out what to do on Thursday after the exam, and what might be a good way to approach the whole situation of them coming with us, somewhere. In fact, we have not even decided yet upon the place where teenage girls like to go, and where we, as a group, could go at 10 am. on Thursday. If anyone has any suggestions, I\'d gladly hear of them. I am quite desperate here, as school ends in 2 complete days, and after that it is good bye to them, unless Will and me do something right now.     Lastly, my third point is completely off topic, but should prove to be a fun one.  urlLink Henry Chee , my good friend from Montreal (also known as p00kie), sent me a link to a game, which I found very interesting. It is a racing simulation game. Hope you enjoy it, as it is very addictive, and you can even save your progress.     Link:  urlLink Drag Racer v3 '),(12719,2747575,'25,July,2004','2004-07-25 00:00:00','    Well, today went by smooth, except for the fact that I was not able to finish everything I estimated I could. I am still working on my homework, the individual part of the CPT (Culminating Performance Task), so I should be done by around 10 pm tonight. Then there is the group CPT which I still yet to properly start. I blame my laziness, and my lack of self control.     In other news, nothing new happened today, so really, there aren\'t really any other news. But I might as well just write something in here. Uhh, if this is news to anybody, school ends a whole 4 days from tomorrow, and then I will have a whole month of boredom. Obviously, that boredom could easily be turned into fun filled adventurous days, but I have yet to see how I could achieve that. One thing is for sure, since I do not have a summer job, I will be sleeping in and staying up much later than I did for the last 11 months.     Well, now, it looks like it is time to go back to work. I hate English as a subject, although the atmosphere this year is very pleasant. '),(12720,2747575,'24,July,2004','2004-07-24 00:00:00','    Well, today went by pretty well. I got a few things done, less than I hoped I would get done, but a sufficient amount. Obviously tomorrow will be a much more demanding day, since it will be Sunday. Still, I am very lazy. For example, instead of finishing up one of my summaries for a short story, I am typing this message up. Not that it really takes long to do my homework, it\'s just that I get easily distracted. I have two computers, one is a new one and one is old. I purposely do my typed homework on the old one because it sucks and I tend to stray away from my homework less on it, but it doesn\'t seem to stop me from posting a message in my wonderful *cough* blog.   And also, my thoughts about other things never disappear. '),(12721,2747575,'23,July,2004','2004-07-23 00:00:00','    Last night went well. We saw the movie  urlLink I, Robot , which was quite interesting. I almost fell asleep. I can sure tell you that the car in the movie,  urlLink Audi RSQ , was surely very awesome. That\'s about all I can say about the movie. I gotta get more damn sleep before I go that late, and the comfortable chairs make it all even worse.   On another note, I have, as mentioned before, lots of homework. \"Lots\" is just the beginning of it, actually. This is my last homework weekend of summer school, and around 10% of my mark lie on it, as well as an additional 20% of my mark, which is the exam, which I\'ll be writing next Thursday. This means that I will be very busy these last few days, but I will obviously find some time to slack off.   And on a third note...I am still available. '),(12722,2747575,'23,July,2004','2004-07-23 00:00:00','    And damn does it feel nice. I have a shit load of homework to do over this weekend, but at least I can just relax for the rest and the day (and night). Today was a fun day; we didn\'t do much and the hottness factor (of them) does not seem to be diminishing. One thing\'s for sure: they do like driving, and going out on Friday nights. I just gotta somehow tell them that I do do those things too. Next Thursday and Friday would be a nice time to do that, but who knows, I am not big on dialogues. '),(12723,2747575,'22,July,2004','2004-07-22 00:00:00','    On a lighter note, this new pictures are just amazing. These are the pictures of the new 2005 BMW M5. Holy shit is all that I can say. Wow.      More pictures:  urlLink 1   urlLink 2   urlLink 3   urlLink 4   urlLink 5   urlLink 6   urlLink 7   urlLink 8   urlLink 9   urlLink 10   Link:  urlLink BMW M5 Press Release '),(12724,2747575,'21,July,2004','2004-07-21 00:00:00','    Hmm, I probably should not be rambling on about this, oh well. Anyway, I really need to do something about this whole situation. I need to maybe get out more often, or something. I don\'t know. Someone help me, please.   Victoria is hot. She is nice. She is beautiful. She likes talking. She is everything anyone can dream of. The question is: how do I tell her that.   And the answer will probably never be discovered. '),(12725,2747575,'21,July,2004','2004-07-21 00:00:00','    As many of you might know, I am taking a summer school course. Nope, I didn\'t fail anything in day school - my marks are actually quite good. I wanted to take a course so that I do not have to take it next year. I am taking Grade 12 English by the way. There is an upside to my extended schooling: there are quite a few hot girls in the school, particularity in my class. Damn, if only...   Well, I should probably stay away from that route. On a second thought, no. I seriously have to do something, get a girlfriend or something. It\'s long overdue, and Victoria would make a nice one.   Anyway, here are some poems that I have to analyze. Oh well.     Welcome and Farewell    Johann Wolfgang von Goethe       My heart beat fast, a horse! away!  Quicker than thought I am astride,  Earth now lulled by end of day,  Night hovering on the mountainside.  A robe of mist around him flung,  The oak a towering giant stood,  A hundred eyes of jet had sprung  From darkness in the bushy wood.   Atop a hill of cloud the moon  Shed piteous glimmers through the mist,  Softly the wind took flight, and soon  With horrible wings around me hissed.  Night made a thousand ghouls respire,  Of what I felt, a thousandth part---  My mind, what a consuming fire!  What a glow was in my heart!   You I saw, your look replied,  Your sweet felicity, my own,  My heart was with you, at your side,  I breathed for you, for you alone.  A blush was there, as if your face  A rosy hue of Spring had caught,  For me---ye gods!---this tenderness!  I hoped, and I deserved it not.   Yet soon the morning sun was there,  My heart, ah, shrank as leave I took:  How rapturous your kisses were,  What anguish then was in your look!  I left, you stood with downcast eyes,  In tears you saw me riding off:  Yet, to be loved, what happiness!  What happiness, ye gods, to love! '),(12726,2747575,'20,July,2004','2004-07-20 00:00:00','    Hello,   This will be my new blog. I will be posting various different things here, including some of my homework, and other things that happen in my life. I will also be posting some secrets that I have never told anyone else, so beware!  Expect more soon. '),(12727,2747575,'11,August,2004','2004-08-11 00:00:00','    Since I had nothing else to do, I took some high quality cell phone pictures, just for your viewing pleasure. Am I not a nice guy?  For anyone who\'s wondering, it is a  urlLink Fido Siemens M55 . It cost me absolutely nothing, without any contract, but that only happened because I am special guy. And no, it does not have a camera, sadly.  Pictures:  urlLink 1   urlLink 2   urlLink 3 '),(12728,2747575,'10,August,2004','2004-08-10 00:00:00','    Last night, I experienced another tragic power outage. No, it didn\'t last the whole day, like the one last summer did, but it was quite an experience in itself. It happened right at around 12 a.m. I was putting the rechargable mouse in it\'s cradle, as I was about to go to sleep, and boom, the power cuts. It scared me for a second there, as the timing was just perfect. Then the power came back on for another instant and shut off. Deciding it wasn\'t worth to stay up and do nothing in the dark, while waiting for the power to come on, I went to bed.  The power came back on at around 12:35 a.m. I didn\'t bother turning on the computer (which also happens to be the server for this website) as I thought that the power might just shut off once again during the night. It stayed on the whole night after.  What is the moral of this story? Err, I don\'t exactly know, just that Hydro companies should have better reliability. '),(12729,2747575,'10,August,2004','2004-08-10 00:00:00','    Jesus, wow, this site deserves an update. Well, what can I say: nothing, really. Uh, it is Tuesday right now, and there are about 4 weeks of summer left for me, after which I have to go back to school. However, others have to go back in around 2 weeks, like my friend who lives in Montreal, so I should be happy, that is, if I like summer. Well, I do.  Well, we\'re finally doing something tomorrow - going to  urlLink Playdium . It better be a nice day tomorrow, or else I will kill something (notice it doesn\'t say \"someone\"). It\'s been quite boring sitting here at home and doing nothing, and tomorrow will be a chance to kick back, relax, and kick some ass at the arcades. Or not. I will also get to drive - for a very long time, and on the highway, something that I have not done in a while. It should prove to be fun, unless I get some sort of a ticket, or worse. Hopefully it won\'t be too crowded either, however, since it is a discount night, I feel like there will be lots of people present. Oh well, we\'ll have a whole 4 hours to kill. There is even a chance Victoria will come - something that would definitely make my day.  Yesterday, I also got these free  urlLink Toronto Blue Jays  tickets in the mail. I almost forgot I ordered them for free around a month ago. I have 2 tickets for this Saturday\'s game, each one of them worth $40, and I am not sure what I should do with them. Thoughts are floating around though.     Other than that, I hope you can wish for me to have a wonderful day tomorrow. Oh, and I promise (if that means anything) that I will try to update this site more often. '),(12730,2747575,'08,August,2004','2004-08-08 00:00:00','    As you can see, I\'ve been really slow on posts lately, mainly because I have nothing to say. I am just disappointed with myself. End of story.  Look for more details tomorrow. '),(12731,2747575,'05,August,2004','2004-08-05 00:00:00','    Well, I chickened out on my promise that I made yesterday. I did not call Victoria today, however, I will definitely do it tomorrow. No matter how much you trust me, either me or Will (and I am hoping it will be me) will call her tomorrow. I am promising you this time.  Other then that, it is really cold outside right now in Toronto. I just walked my dog and I am freezing, and my hands are numb, so this post will remain short. Period. '),(12732,2747575,'04,August,2004','2004-08-04 00:00:00','    Well, nothing happened today, as usual. Nothing will ever happen with me. I installed a DVD Burner for a friend, and that\'s about it. I didn\'t watch any movies today either. Instead, as pathetic as it sounds, I tried to organize my thoughts and emotions and call Victoria. Well, at around 9 p.m. I finally did just that. I said \"Hi\" to her mom and asked for Victoria. She picked up very quickly. I was just saying \"Hi\" to her, and the phone starts beeping. The conversation was cut. Just my luck, the cordless phone died on me. No, I did not hang up; the battery rang out. I was very disgusted - you cannot imaging how hard it was for me to dial the number, and now, it was almost like for nothing. Wait, no, actually it was not.  I did not call back, however, that does not mean that I will not call her again. I have, indeed, accomplished my goal for today. It did not go as far as it should have, but the cause is beyond my control. I will, and I promise this, call her again tomorrow and speak to her. I will just tell her how I feel about her, and nothing more. I will leave it up to her to interpret - she can do anything with my emotions towards her. The way I see it, the conversation will lift a whole lot of pressure from me. If she denies me sharply, I will be able to forget as I have done everything in my power. If she shows some sign of interest, it will be a bonus, and I will definitely take advantage of that bonus.  Sure, this sounds all lame, but I have to give it a shot. I really do love her that much. '),(12733,2747575,'03,August,2004','2004-08-03 00:00:00','    Just if you didn\'t happen to have enough to read in the post below, I will add another 54 page read for you. Yes, you read it right, I am posting a story of a man by the name Patrick Combs who was written a story that is 54 pages long (formatted in Microsoft Word). It is, as the title states, a story about a bank error. I highly recommend that you read it. It is about 9 years old, but still quite relevant. I didn\'t finish reading it at this time as it is too late, so I don\'t know the ending, but I will after I get some sleep.  Link:  urlLink Bank Error In My Favor (Microsoft Word Document)  Link:  urlLink Bank Error In My Favor (Website) '),(12734,2747575,'03,August,2004','2004-08-03 00:00:00','    Well, summer has been steaming ahead at rapid speed, while I enjoy doing nothing in the comfort of my home. That didn\'t make sense much.  I watched a total of 4 movies at home today, by my myself. So, since nothing else happened in my life (I might be a bit under exaggerating), I will critique every one of those movies.  The first one that I watched was  urlLink Beyond Borders . It wasn\'t a bad movie, especially since one of my favorite actors is in it,  urlLink Clive Owen . The reason that I think he is so great is because of  urlLink BMW Films  that he stars in. If you haven\'t yet caught on, my dream car is a BMW. Anyway, back to the movie. If you want to feel bad about yourself and the world around you, then this is the movie to watch. Otherwise, it is complete garbage.     Next, I watched  urlLink Adaptation . Sad little movie, but it just did not hit me hard enough. I was so boring until around the last 20 minutes when something actually started happening. Basically, the movie is about a movie in which a person is writing a screenplay for a movie based on a book. Yea, something as messed up as that. Anyway, if you like to be bored and like  urlLink Nicolas Cage  acting as a freak, again, then this is the movie for you. By the way, he is 2 characters throughout the whole movie, being on stage twice all the time, which was probably quite hard to actually film.     The third movie that I saw was  urlLink Punch-Drunk Love . It was another one of those very boring movies about a mentally challenged (retarded) person, except in this movie, the guy got everything he wanted: he got love. Well,  urlLink Adam Sandler  did a brilliant job, but this movie was very boring as well. It was hard to see such an incapable characters get anywhere in life. This characters is so out of it, yet he manages to somehow get his life fixed and finds the right girl (well, actually, she finds him). Pathetic, although, I do appraise the movie, as its ratings are very good.     And the last movie that I watched:  urlLink The Perfect Score . Now, this movie really cheered me up. Yes, the plot is not the greatest, but the actors, in particular on of them by the name  urlLink Leonardo Nam  where great. Leonardo, who played Roy in the movie, was the \"comic relief\". He was so funny, it actually made the movie very much worth my time - much better than the other 3 that I watched today. I highly recommend that you watch it: you might not learn much, but at least you will be entertained. Here are some memorable moments from the movie.         That wraps up the movie review for today. Just one more thing that is worth mentioning: I still do like Victoria very much and I am even going to try to speak to her tomorrow, and we will see what happens. Call me crazy, and I know I am, but it is worth a shot. I hopefully won\'t be disappointed that much. Wish me luck. '),(12735,2747575,'02,August,2004','2004-08-02 00:00:00','    Well, I just watched this movie, and boy, is it weird. First, I\'ll start off with quoting the first thing you see on the screen once the movie starts.    It has been said that something as   small as a flutter of a butterfly\'s   wing can ultimately cause a typhoon   halfway around the world.'),(12736,2747575,'01,August,2004','2004-08-01 00:00:00','    I watched  urlLink Shrek 2  today. It wasn\'t as good as I thought it would be, but fun nonetheless. The music was also not bad, however, it wasn\'t as good as in  urlLink Shrek . It was very short as well. Other than that, I\'ve been drinking  urlLink Dr. Pepper  and  urlLink Coca Cola  steadily throughout the day to stay alive. I\'m sort of too lazy to make food; so much for my health.  The  urlLink CNE  opens its doors on August 20th. Yes, that is quite a long time away, but we should probably go. I am sure Victoria would be going there with her friends anyway, so it is another possible location to go. It should be quite fun too; I\'ve gone a few times around 5 years ago, when I lived in Mississauga.  urlLink Ontario Place  might be another fun place to go to as well. I cannot wait.  In case you are still bored, like I am, you might want to check these videos out. Viewer discretion is advised.  Link:  urlLink Dork Vs. Glass  Link:  urlLink Dr. Pepper Is Not Safe '),(12737,3887824,'30,July,2004','2004-07-30 00:00:00','    I went to bed around 11:30PM and got up at 5:35AM.&nbsp; I got to catch the end of sunrise.&nbsp; It was great!&nbsp; Nice and pink.&nbsp; Tomorrow I want to get up around 5:20-5:25AM so I can see more.   Now I\'m doing school, which I have been for the past hour and a half.&nbsp; This is so cool.&nbsp; Yesterday I had a violin lesson which went well, then I went to the library.&nbsp; I checked out some books, also got a cd recording of Les Miserables, and a VHS of Pirates of Penzance.&nbsp; The audio and video are due back in&nbsp;two weeks, the books in four.&nbsp; I think I\'ll bring at least some of the books to camp.   Vickie is coming in FIVE DAYS!&nbsp; YAY FOR YAY!   My days are more productive now.&nbsp; I\'m not just sitting around anymore.&nbsp; Some things are weird, like getting used to the fact that I\'m 16 and I have my drivers license (it\'s like, Whoa- how did this happen?) but I\'m really happy now.&nbsp; I\'m so moody, lol.&nbsp; Two days ago I was feeling really down (Hey, it happens), but now I\'m not, and I have a feeling that I won\'t be for a while.&nbsp; The fact that I\'m in a really good mood Pre-Camp is amazing, so imagine how I\'ll be while I\'m there, and when I come home!&nbsp; Camp is so inspiring.&nbsp; I can\'t WAIT to see people.   School will continue to progress well now, and hopefully I can get a job soon.&nbsp; For a couple of weeks I was just not into anything, but things have turned around... so I am glad.   \"YAY FOR YAY!\" -Vickie.&nbsp; That pretty much covers it.  (Wow, this blog is so pointless, lol)   '),(12738,3887824,'27,July,2004','2004-07-27 00:00:00','    I slept for about 7 hours last night.&nbsp; Yay!   I\'m at a site...  urlLink http://i-am-bored.com .&nbsp; It\'s pretty cool.&nbsp;&nbsp; I found out that I read between 250-300 words per minute (Above Average?).   Here is&nbsp;a quote from Matt Groening:   \"Love is a snowmobile racing across the tundra and then suddenly it flips over, pinning you underneath. At night, the ice weasels come.\"   ????&nbsp; Okay, then.  Here are some quotes from Austin Powers: The Spy Who Shagged Me.   \"Dr. Evil: You\'re not quite evil enough. You\'re semi-evil. You\'re quasi-evil. You\'re the margarine of evil. You\'re the Diet Coke of evil, just one calorie, not evil enough.\"   \"Austin: You\'re one groovy baby....baby.\"   Okay, that\'s enough.&nbsp; Well, one more quote. From The Shawshank Redemption.   \"Warden Samuel Norton: I believe in two things: discipline and the Bible. Here you\'ll receive both. Put your trust in the Lord; your ass belongs to me. Welcome to Shawshank.\"   Yeah, I\'m done now.&nbsp; Really.   '),(12739,3887824,'24,July,2004','2004-07-24 00:00:00','    Today&nbsp;I went to&nbsp;church.&nbsp; Which was cool, I guess.   Last night, I&nbsp;was in bed from 3:30-9:30am.&nbsp; I just do not feel tired at a reasonable time.&nbsp; Still, I had 7 hours of sleep!&nbsp;&nbsp;That\'s what I can consider a \"full nights sleep\" now.  Hmm.&nbsp; I wonder who reads this blog.&nbsp; If anyone is reading, reply.   '),(12740,3887824,'23,July,2004','2004-07-23 00:00:00','    So, I guess blogger is my school/sleep journal.&nbsp; Well, in that case, I finished&nbsp;a science test yesterday, and I picked up some job applications.&nbsp; Today&nbsp;I have school, of course, as well as chores and music.&nbsp; Like everyday.&nbsp; Later I\'m going to&nbsp;hand the applications, and then I\'m going to go to the music store. &nbsp;Specifically I\'m looking for David Lanz, a decent violin and piano duet, and any kind of random pop, such as Vanessa Carlton.   Now onto the sleep portion...&nbsp; I went to bed around 2am.&nbsp; Then my alarm went off at 6:26am, but I pushed snooze.&nbsp; Then I guess I slept through the beeping if/when it sounded again.&nbsp; Around 8am my mom said, \"It\'s getting late!\" so that prompted me to get up.&nbsp; So... 2am to 8am... that\'s still&nbsp;6 hours.&nbsp; Insane, right? &nbsp;Better than 5 1/2 hours...   Updated at 3PM:  I didn\'t do any school today.&nbsp; Around 10 AM we went out though.&nbsp; I submitted an application to Dunkin Donuts, got a book of soprano arias from the music store, plus we went to the grocery store.&nbsp; I got some Ramen noodles, and Mac and cheese.&nbsp; Mmmm... haven\'t had stuff like this in a while.&nbsp; It\'s wonderful...   '),(12741,3887824,'22,July,2004','2004-07-22 00:00:00','    I got up at 6:26am-6:30am this morning, having gone to bed around 1am (or somewhere around that time, anyway).&nbsp; That\'s 5 1/2 hours.&nbsp; Yes.&nbsp; Just an observation.&nbsp; I guess that will come in handy at Camp.&nbsp; Now I just need to throw in some regular exercise so I don\'t pass out after hiking half a mile to the dining hall...   Also, in case you\'re wondering why this journal is so friggin BORING it\'s because the one I use to write my actual feelings in is here: http://livejournal.com/users/oracular_echo   '),(12742,3887824,'21,July,2004','2004-07-21 00:00:00','    Put the first thing that comes to mind with each word/phrase.    ..snow?: white    ..rain?: erotic (yes I\'m weird)    ..tornado?: valley     ..summer love?: happened so fast     ..Jon?:&nbsp;Evans    ..Mike?: Bailey; \"Come along, Mike\"     ..Shea?: California     ..banana?: I think of the taste and texture...    ..dizzy?: low blood pressure    ..Laura?: Delozier    ..Juan?: Hispanic    ..car?: drive    ..white?: color    ..peppermint?: patty   ..New Found Glory?: Good Charlotte (and the tour bus at the hotel!&nbsp; too bad I\'m not someone who is actually a fan...)   ..placebo?: placenta   ..orange juice?: orange   ..candid camera?: good shot   ..sister?: Sue   ..brother?: Trav   ..hate?: dark   ..school?: MHS   ..President?: Bush   ..football?: Patriots    ..rap?: Eminem   ..pop?: music, lol    ..rock?: and roll   ..punk?: limp bizkit   ..sex?: --ual   ..death?: end/beginning   ..baby?: Rowan   ..duuude?: 60s    ..the end?: I picture the end of a book   '),(12743,3887824,'19,July,2004','2004-07-19 00:00:00','    School went well today.&nbsp; I did Physical Science, which is progressing nicely.&nbsp; Tomorrow I will be doing either Sociology, Physical Science, or Algebra 2.&nbsp; Either one of those, or two, or all three.&nbsp; We\'ll see when tomorrow comes.    &nbsp;  I need to be done with English 2, Physical Science, and Algebra 2 by the first week or so of October.    &nbsp;  Note- Yahoo Launchcast Radio rocks my world.    &nbsp;  I finished watching the first four episodes of Tenchi in Tokyo today.&nbsp; Next I have episodes eight through 10.&nbsp; Hmmm, I guess I will have to buy the manga for in between.    &nbsp;  I changed the format and colors for my other journal.&nbsp; I love how it is now!&nbsp; Especially compared to before.    &nbsp;  I need to read more.&nbsp; I mean, I read a lot online.&nbsp; But I should read more of actual books.&nbsp; I usually do... not lately though.   '),(12744,3887824,'18,July,2004','2004-07-18 00:00:00','    A is for - Age: 16  B is for - Boyfriend/Girlfriend: None  C is for - Career in future: Music/Psychology;&nbsp;Music Therapy perhaps  D is for - Dad\'s name: Step dad\'s is Gary, biological dad\'s is Raymond  E is for - Essential item to bring to a party: Humor  F is for - Favorite song at the moment: I really like A Plain Morning...  G is for - Guy/Girls you\'ve kissed: Well, John... and people, often casually. Like on the cheek. That is always \"fun.\"&nbsp; Well, except when you loudly announce, \"I can kiss Yui on the cheek,\" and then proceed to kiss her on the lips.&nbsp; Accidentally. lol&nbsp;Yes, I am eccentric.  H is for - Hometown: Born in Norwich, lived in Glasco until I was 4, then moved to Uncasville.  I is for - Instruments you play: piano, violin, voice, pennywhistle, wannabe guitar  J is for - Job title: Unemployed... Grr  K is for - Kids: I want 2-3.&nbsp; If by 2 I have one girl and one boy, I probably won\'t plan more... but you never know... I don\'t think more than 3 though.&nbsp; (Of course, then 20 years from now I have 5...&nbsp; How nice.)  L is for - Living arrangement: My parents (blended family)  M is for - Mom\'s name: Lois  N is for - Number of people you\'ve slept with: Zero...  O is for - Overnight hospital stays: None  P is for - Phobia[s]: My bedroom door has to be shut when I\'m sleeping... I guess I\'m paranoid that somebody is going to come in at night?&nbsp; Although for some reason I\'m more lenient with windows being open...&nbsp; Also, while I love the water, I have this fear of drowning.&nbsp; If I\'m in a position where I have to tread water watch me freak out.  Q is for - Quote you like: \"The lip gloss is dead for it has fallen over.\" -Brett  R is for - Relationship that lasted the longest: Besides&nbsp;my family...&nbsp;I\'ve been friends with Stephanie Tine since Kindergarten... so, 11 years I guess  S is for - Sexual position: Uh any?&nbsp;lol.. No, I don\'t know. Just normally I assume. T is for - Time you wake up: Between 6:30-7am on a good day  U is for - Unique trait(s): I don\'t know  V is for - Vegetable you love: Black olives!&nbsp; Yay!&nbsp; And carrots, zucchini, eggplant, etc.  W is for - Worst habit: Interrupting people, not making myself do stuff...  X is for - X-rays you\'ve had: Teeth, and that\'s it I believe  Y is for - Yummy food you make: Mac and cheese, lol. And Banana Nut Bread.  Z is for - Zodiac sign: Pisces   '),(12745,3887824,'17,July,2004','2004-07-17 00:00:00','    urlLink http://www.urbandictionary.com/define.php?term=weenus&amp;b=1  &nbsp;  weenus    the extra skin on your elbow when you stretch your arm out.   '),(12746,3887824,'16,July,2004','2004-07-16 00:00:00','    Went to bed a little after midnight... got up about 6:40am.&nbsp; I  still  need to finish up this week\'s bible reading thing, and do some cleaning.&nbsp; Arrgh.&nbsp; School went well today, though.   '),(12747,3887824,'15,July,2004','2004-07-15 00:00:00','    I got a fair amount of school work done today, doing Physical Science, reading about sound waves and such.&nbsp; Pretty interesting.&nbsp; I learned what \"ultrasonic\" meant, which helps because my dad was an \"ultrasonic technician\" when he worked.&nbsp; (He is now Retired/Disabled).&nbsp; So, now I understand a little better.&nbsp; Weird.&nbsp; Anyway, as&nbsp;for tomorrow, I\'ll be doing some more Physical Science and possibly&nbsp;Algebra 2.&nbsp; I also need to make banana nut bread for a bake sale, clean the bathroom, finish up the bible reading this week (thing I\'m doing with Vickie), then work on some music.   '),(12748,3887824,'14,July,2004','2004-07-14 00:00:00','    urlLink       Take the  urlLink What High School Stereotype Are You?  quiz.  Interesting.  Well, when I went to school I  was  an outsider...   '),(12749,3887824,'13,July,2004','2004-07-13 00:00:00','    Your True Nature by  urlLink  llScorpiusll        Username           The quality that most appeals to you:    Loyalty      In a survival situation, you:    Play dead      Your hidden talent is:    Pragmatism      Your gift is:    Artistic talent      In groups, you:    Don\'t fit in      Your best quality is:    Your empathic nature      Your weakness is:    Your passivity             Created with the ORIGINAL  urlLink  MemeGen  !       Whoa... freakish.       Your love is... by  urlLink  ChibiMarronchan        Your name is...           Your kiss is...    mysterious      Your hugs are...    gentle      Your eyes...    twinkle in the moonlight      Your touch is...    heart warming      Your smell is...    amazing      Your smile is...    amazing      Your love is...    unique             Created with the ORIGINAL  urlLink  MemeGen  !            Your Love Situation  by  urlLink  Amberishjewel        Username?           Your Love Is...    Seductive      During Lovemaking You Act...    Like a child, always playing      Your Partner Is...    Your best friend      Your Partner Has Said That You...    Are their favorite person      Your Love is Summed Up In A Quote.    \"You deserve a bed of roses\"             Created with the ORIGINAL  urlLink  MemeGen  !        Hmm... Okay I\'m done now.   '),(12750,3887824,'12,July,2004','2004-07-12 00:00:00','    Definitely! Someones gotta crush on you! It\'s Obvious that this guy likes you, so stop doubting it! If hes a little shy and hasnt said anything, maybe you should be the one to make the first move (If you feel the same way of course). If you dont feel the same way, and he has already told you that he likes you, tell him how you feel in the nicest way possible, after all he was brave enough to tell you.       urlLink Does He Like You?     brought to you by  urlLink Quizilla   '),(12751,3887824,'11,July,2004','2004-07-11 00:00:00','    A connection? Perhaps. We shall see; We shall. A crush? A love. But far away; the Physical distance. You cannot know if it may work out. You are stuck. You strive to like everyone you meet, but sometimes they do not like you back. This one is different; Special. Worth it. There is a connection; Something. But what? Will you ever know? No. ...Maybe. Such is life,  you suppose. Please live to understand that you cannot always get what you want. But if something is meant to happen then it will happen. ...Right? We shall see; We shall. So, A connection? â€¦Perhaps.   '),(12752,3887824,'11,July,2004','2004-07-11 00:00:00','    First of all, this journal is pointless.  Quiz results, general updates.  I will talk about school and such, though.  If you want to read my main journal, go here.  http://livejournal.com/users/oracular_echo  I got up a little before 10am this morning for some reason.  I slept for more than 7 hours!  Wow!  lol  Now I\'m just sitting, and chatting.  I think I\'m going to do some writing, then some school to catch up a little.  I got a letter from Vickie yesterday.  I need to write a couple of letters this week for the pen pal club.  I may be getting a job soon, too.  Well, that\'s all for now.   '),(12753,3887824,'10,July,2004','2004-07-10 00:00:00','    You match best with Brett! He\'s smart and talented and keeps a conversation going. He has a good sense of humor and is usually in a good mood.    urlLink   Which UCG guy do you match with best? (Girls Only)     brought to you by  urlLink Quizilla   '),(12754,3887824,'07,August,2004','2004-08-07 00:00:00','    Vickie is here and we are having a great time.  We\'ve been having long walks and hikes, then going around to a few places.  Plus movie viewing (older Batman movie, episodes of Tenchi in Tokyo, some Fellowship of the Ring, The Village) and playing games (Trivial Pursuit Junior, Bible Trivia, etc.)  Also throughout the day we\'ll do some yoga.  I named a position, \"Head clamp,\" for obvious reasons if you\'d see it.  Fun stuff.  Anyway, it\'s so cool to just spend time with her like this.  It\'s wonderful that she\'s even here.  Tomorrow afternoon we\'re going to the beach, I think, then on Monday to an art museum, and to Leif\'s on Tuesday.  Ideas for activities are still floating around.  Every morning (well, except for this morning) we have and will be walking to Dunkin Donuts.  Yay for half a bagel / donut / muffin.  Mocha swirl latte... yummy... yes, so we will miss things like that at Camp Heritage.  But yeah.  Anyway, just thought I\'d update and give a short summary of Vickie\'s time here and stuff we\'re up to.   Happy Sabbath!'),(12755,3887824,'02,August,2004','2004-08-02 00:00:00','    Friday night sleep: 4:30AM-8:30AM Saturday night sleep: 11:30PM-5:30AM-8:30AM Sunday night sleep: 12:30AM-5:30AM-6:30AM  Now here I am, with striped toe socks. I was all ready to submit work to school, but the site is still down. It will either be up later today or tomorrow, I guess. So at the moment I\'m reading the Bible and awaiting baby sitting a little while from 8:30/ish through ??.  Later today I need to go to the bank and like, Wal Mart, or the Mall or something. Oh, and yesterday mom and I watched Renaissance Man starring Danny DeVito. And last week I saw Cold Mountain, which was a  great  movie! Now I want the soundtrack.  Heehee'),(12756,1143104,'22,December,2003','2003-12-22 00:00:00','   I\'m torn between starting this with a basic gesture that is most often grossly overlooked, or something a little more detailed. So here\'s a little of both:   Tip of the Day   Basic: Open the door for your lady. There is no reason a woman should have to open her own door when a perfectly able-bodied gentleman is standing idlely behind her. Step forward, reach out, open the door, and invite her to enter before you do. In addition to saving her a little extra work, it allows her to make a more stunning entrance...instead of forcing her to deal with the door as she clumsily enters the room.  More Detailed: Greet her in a mature and polite manner. \"Hey.\", \"Wuz up?\",  and \"You\'re hot!\" are not appropriate greetings. Use something akin to \"Good evening\" or \"You look beautiful\". Make her feel less like an object and show her that you appreciate the time she has put into looking good for you (and everyone else at wherever you\'re going). Combinations and different versions of such polite greetings are nearly endless and can add a few extra points. Examples would be similar to: \"You look beautiful this evening\" or \"That is a lovely outfit you\'re wearing\". But make sure you greet and then compliment. If you don\'t greet first, things can seem a little off or moving too fast.  '),(12757,1143104,'21,December,2003','2003-12-21 00:00:00','   This blog is all about the men...well, sort of. It will consist of tips, ideas, and a few charm skills to impress that woman that means so much. Whether its your girlfriend, wife, mother, sister, grandmother, whoever -- this website will (hopefully) help you be a better man. Good luck! :-)  '),(12758,1143104,'23,January,2004','2004-01-23 00:00:00','   Tip of the Day   Return her phone calls. When a girl calls you, it means one of two things: she wants something from you or she cares about you. If she calls \"just to chat\", that means she cares about you. If she calls asking for a favor, obviously she wants something from you. Either way, its probably not a bad thing. By doing the favor, you can impress her by showing her you are reliable. If you just take a few minutes (or hours, depending on the girl) to chat with her, she\'ll appreciate the gesture and your willingness to listen. Just make sure you return her phone call in a TIMELY MANNER. Don\'t wait a few days. Even if you don\'t have much time, you can at least call her back to let her know that. Otherwise, she\'ll think you\'re rude and hate you.  '),(12759,1143104,'22,January,2004','2004-01-22 00:00:00','   Tip of the Day   Drive safely. Normally, like most guys, are you less than law-obiding when operating a motor vehicle. However, when driving your date somewhere, you should make her feel safe and secure with your driving ability. Take it down a notch or two and don\'t....I REPEAT: DON\'T...ever say, \"Hey babe, watch this!\"  '),(12760,1143104,'19,January,2004','2004-01-19 00:00:00','   Tip of the Day   Don\'t interrupt. When the lady is speaking, don\'t try to interject or override whatever she is saying. A conversation should be an exchange of ideas and/or emotions. Therefore, when you interrupt her, it is showing her that you believe your ideas and emotions are more important. Impress her by showing that you value what she thinks and feels about.  '),(12761,1143104,'17,January,2004','2004-01-17 00:00:00','   Tip of the Day   Let her keep her secrets. If she has things she doesn\'t want to share with you, that\'s fine. Don\'t be upset about it and don\'t get annoyingly nosy. There\'s nothing wrong with you wanting to know about her and her life, but if you know she has a diary, don\'t look for it. Finding it can end a relationship.  '),(12762,1143104,'11,January,2004','2004-01-11 00:00:00','   Tip of the Day   Let her sleep. When your lady is tired and needs her rest, let her sleep. You may have had plans or wanted to do something else, but if the woman you love is tired - let her sleep. You can always do things at another time/day. Even if they get cancelled all together, what\'s the big deal? Isn\'t her comfort (and happiness) more important? She\'ll be less grouchy, more rested, more energetic, and more delighted to be with you. A well rested woman is a very happy date.  '),(12763,1143104,'07,January,2004','2004-01-07 00:00:00','   Tip of the Day   Brush your teeth and use mouthwash. We\'ve already covered the bathing part of your hygiene, so now we will cover the cleanliness of your mouth. When you lean in for the kiss, chances are, the girl will pull away if your breath smells like her toilet. If you just ate some nice Mexican food, or woke up and ate a bowl full of crap, please take a few moments out of your life and brush your teeth. Preferably after your done, use an antibacterial mouthwash and rinse your mouth out. That should guarantee (to 99.7% according to Listerine) the cleanliness of your mouth. :-) This will also guarantee that the woman you are leaning in to kiss won\'t turn and run...at least not from your breath!  '),(12764,1143104,'06,January,2004','2004-01-06 00:00:00','   Tip of the Day   Don\'t swear in the presence of a woman. Most women don\'t appreciate the foul-mouthed linguistical skills of your average man. Some men are pansies and don\'t swear, but for the most part, all guys have a sailor-spirit in them. When you\'re around women, keep the profanity in your head and the compliments flowing from your mouth. Of course, the obvious exception to this is if the female company you keep doesn\'t mind a little bad word here and there...especially if it excites them! (wink wink). ;-)  '),(12765,1143104,'05,January,2004','2004-01-05 00:00:00','   Tip of the Day   Take an interest. Learn about the woman you are either trying to impress (a.k.a. - get to know) or the woman you love. Find out what her birthstone is. What are her goals in life? Hell, even the classic favorite color is a good thing to know. You should care enough about her to learn about her. (It also helps when buying gifts!)  '),(12766,1143104,'04,January,2004','2004-01-04 00:00:00','   Tip of the Day   Surprise her. Bring her flowers or meet her when you know she doesn\'t really have much to do and might be bored. The womenfolk love surprises because...well, for a few reasons. First, it shows that you are thinking just a little outside the box. You care enough about her to take time out of your life (at least you\'ve convinced her by now that you have one), and are willing to spend it with her. But they also like surprises because that means more time with you. You should want to spend a little extra time with your girl because she probably wants to spend it with you. But remember, don\'t try and surprise her in the middle of her work day when she has about 1,000 things to do and her boss is breathing down her neck. She\'ll think its \"cute but stupid\". Do it when you know she\'s not particularly busy or maybe a little sad or upset. A nice comforting hug can be the best thing in the world.  '),(12767,1143104,'03,January,2004','2004-01-03 00:00:00','   Tip of the Day   Bathe. Take a bath. Take a shower. Do something to erase the smell of sweaty sports, manual labor, or just sitting in the chair in front of your desk. The ladies don\'t want to smell  all  of you...just the nice parts. They\'ll smell the rest of you...umm...later. hahaha But seriously, make sure you are clean and presentable when seeing your lady.  '),(12768,1143104,'02,January,2004','2004-01-02 00:00:00','   Tip of the Day   Pay for the meal. Show her that you care by saving her the trouble of paying for herself or both of you. In modern society, some consider the first date to be an acceptable time to \"go Dutch\" - each person paying for their portion. But if you\'re looking to impress, pull out your cash or the always impressive Platinum card and pay for her. If it seems like she is trying to order the cheapest item on the menu or orders a salad and water, politely let her know that she should order whatever she wants...let her know there are no limitations. Even if it breaks your bank and the date is already outside your budget, if you care, you\'ll pay for whatever she wants. NOW - if she jumps from water to champagne, dump the gold-digging bitch!  '),(12769,1143104,'08,February,2004','2004-02-08 00:00:00','   Tip of the Day   Respect her parents. They may hate you, they may hate your relationship, they may be racist, they may be stupid, they may be arrogant, they may be pretentious, they may be snobs... But they are still her parents and, especially if you marry her, they will be a part of your life. And perhaps more importantly, they are what made the woman you love. They helped shaped who she is --through good or bad-- she is who she is due in large part to her parents.  Now, I\'m not saying you need to revere them or buy them gifts, but you should at least be civil to them and not make the situation worse for her or your relationship. Put up with them whenever you have to. It will GREATLY help in the long run.  '),(12770,1143104,'18,March,2004','2004-03-18 00:00:00','   Tip of the Day   Compliment her on her progress/goals. This is especially good if your girl is trying to lose weight. Women like to know that you are noticing their progress. Hell, even if you haven\'t noticed a damn thing cuz you\'re too thick-skulled, comment on it anyway. This is actually pretty basic to understand. If they see that you are noticing their progress, they believe you\'re focusing attention on them and actually taking an interest in their life! You should be doing this anyway, but most of us guys don\'t notice things like that. Women, on the other hand, are so jealous, bitter, and backstabbing that they always notice the slightest ounce and go for the jugular with each other. I know - they\'re weird.  '),(12771,1143104,'04,March,2004','2004-03-04 00:00:00','   \"Any dick above 8 inches is just for novelty purposes.\" -- Andrew Ellingwood, March 4, 2004  '),(12772,1143104,'10,May,2004','2004-05-10 00:00:00','   Tip of the Day   Don\'t be insensitive. Sometimes I have to catch myself on this one because men and women view many things in life with emotional differences. Girls can cry about the weirdest, most unimportant things. Guys can often not be emotional when its appropriate. Make sure that when your girl is upset, or happy, or worried, or scared that you aren\'t being an insensitive jackass and making stupid comments about it.  '),(12773,1143104,'30,June,2003','2003-06-30 00:00:00','   ...Sgt. James Russell Ayers V staying alive and continuing to serve proudly in Baghdad.  What are you thankful for?  '),(12774,1143104,'29,June,2003','2003-06-29 00:00:00','   ...thankful people who appreciate what they have, when they have it. ...positive advances in technology. ...puddles. ...my AZHo Pledge Bros!  What are you thankful for?  '),(12775,1143104,'29,June,2003','2003-06-29 00:00:00','   ...Vanilynne and all the support, help, and kindness she has shown me.  '),(12776,1143104,'28,June,2003','2003-06-28 00:00:00','   ...music I connect with. ...the strength of will to be myself. ...Middle East cease-fires. ...technology that I control and not the other way around. ...beatiful female bodies. ...pleases and thank yous.  What are you thankful for?  '),(12777,1143104,'28,June,2003','2003-06-28 00:00:00','   ...not going hungry today. ...spending time with Patti. ...not breaking down on the freeway. ...not being burdened with uncontrollable debt. ...having a family that cares.  What are you thankful for?  '),(12778,1143104,'23,July,2003','2003-07-23 00:00:00','   ...pillows. ...quiet. ...LIVE. ...soft sheets. ...a warm comforter.  What are you thankful for?  '),(12779,1143104,'20,July,2003','2003-07-20 00:00:00','   ...a helpful Dad ...hardware stores ...money (I wish I had more.)  What are you thankful for?  '),(12780,1143104,'19,July,2003','2003-07-19 00:00:00','   ...Robin Williams. ...Ellen Degeneres. ...George Carlin.  What are you thankful for?  '),(12781,1143104,'18,July,2003','2003-07-18 00:00:00','   ...free concert tickets. ...Cookies-n-Cream ice cream. ...well written stories. ...Delta Sigma Pi. ...my health.  What are you thankful for?  '),(12782,1143104,'15,July,2003','2003-07-15 00:00:00','   ...George Carlin. ...epiphanies of truth. ...the ability to read.  What are you thankful for?  '),(12783,1143104,'13,July,2003','2003-07-13 00:00:00','   ...loving parents. ...fireworks. ...my Jeep. ...dedication. ...a cool breeze on a hot day. ...a sense of accomplishment. ...\"mildly high cholesterol\". ROFL!  What are you thankful for?  '),(12784,1143104,'11,July,2003','2003-07-11 00:00:00','   ...parties that turn into really fun kickbacks. ...friends that don\'t add drama. ...friends who contribute to the party supply. ...automobiles. ...innocent girls named Jessica, and not so innocent girls named Ellen. ...James and his unique contributions to every thing. ...Jennifer and her subtle wit and personality. ...Ken and all of his good intentions and true spark for life. ...Mario and his badass persona. ...Will and his \"Root Beer Floats\" (with Smirnoff Twist) ...Patti and all of her unending help. ...parents who go out of town and trust me to have a respectable party.  What are you thankful for?  '),(12785,1143104,'08,July,2003','2003-07-08 00:00:00','   ...chocolate chips. ...friends who don\'t care about color. ...crossing guards.  What are you thankful for?  '),(12786,1143104,'29,August,2003','2003-08-29 00:00:00','   The past two days have been the kind of days I write this Blog for. I\'m not feeling great about much of anything, but I can go back through this Blog and realize that my life could be a lot harder, and that I have a lot to work with and look forward to. With every thing from friends, to Brothers, my love Patti, and my family, I\'m a lot better off than other people. Especially the little things like clean water to drink, the sounds of Metallica, and the taste of chocolate are all worth appreciating.  And this weekend will have a lot of that. I\'ll be hanging out with friends, Brothers, Patti, and my family. I\'ve still got clean water to drink, the sounds of Metallica, and I know I\'ll be having chocolate milk at least once with my girlfriend! (cuz she rocks!) There\'s definitely stuff to be happy about!  '),(12787,1143104,'25,August,2003','2003-08-25 00:00:00','   ...my Dad\'s V.A. benefits so I don\'t have to pay tuition. ...a California school loan. ...a reliable vehicle that is one of my sources of pride and joy. ...Delta Sigma Pi. ...the new Quadratec catalog.  What are you thankful for?  '),(12788,1143104,'24,August,2003','2003-08-24 00:00:00','   ...not having any serious allergies - only hay. ...having food whenever I\'m hungry.  What are you thankful for?  '),(12789,1143104,'22,August,2003','2003-08-22 00:00:00','   ...changing my signature. ;-)  '),(12790,1143104,'20,August,2003','2003-08-20 00:00:00','   ...breath mints. ...silence.  What are you thankful for?  '),(12791,1143104,'18,August,2003','2003-08-18 00:00:00','   ...a nice glass of orange juice in the morning. ...24 hour news channels (usually CNN)  What are you thankful for?  '),(12792,1143104,'16,August,2003','2003-08-16 00:00:00','   ...a comfortable bed. ...cable channels that show great cult classics. ...clean water. ...vacuum cleaners. ...the realization of someone\'s true nature.  What are you thankful for?  '),(12793,1143104,'16,August,2003','2003-08-16 00:00:00','   ...the support of friends. ...the death of Idi Amin. ...the calm and helpful reaction of those caught in the black out. ...talking to Patti after 3 days of not being able to.  What are you thankful for?  '),(12794,1143104,'12,August,2003','2003-08-12 00:00:00','   ...the lyrics and emotions of Metallica. ...Minute Maid and Israeli Sabra products. ...mental control. ...self-honesty. ...self-delusion.  What are you thankful for?  '),(12795,1143104,'04,August,2003','2003-08-04 00:00:00','   ...Bill Maher.  What are you thankful for?  '),(12796,1143104,'03,August,2003','2003-08-03 00:00:00','   ...complete installation directions. ...complete collection of all necessary parts. ...duct tape. ...understanding. ...long-life batteries. ...Patti. ...Dorothy and her brave GCC registration. ...Metallica.  What are you thankful for?  '),(12797,1143104,'30,September,2003','2003-09-30 00:00:00','   ...the support of my Brothers. ...the support of my girlfriend. ...the support of my parents.  What are you thankful for?  '),(12798,1143104,'22,September,2003','2003-09-22 00:00:00','   ...a cold glass of milk when I burn my mouth. ...bosses who react better than expected. ...Metallica and their musical genius. ...patience.  What are you thankful for?  '),(12799,1143104,'21,September,2003','2003-09-21 00:00:00','   ...three phase car wax systems. ...a strong Executive Committee. ...vacations. ...ice cream on a warm calm night.  What are you thankful for?  '),(12800,1143104,'17,September,2003','2003-09-17 00:00:00','   ...supportive Brothers. ...reconnecting with old friends. ...a loving girlfriend. ...heartfelt laughter.  What are you thankful for?  '),(12801,1143104,'12,September,2003','2003-09-12 00:00:00','   ...those moments of inspiration that make the mind feel powerful and strong despite its own acceptance of its infantile importance.  What are you thankful for?  '),(12802,1143104,'09,September,2003','2003-09-09 00:00:00','   ...having a job. ...earning people\'s respect. ...being creative in business.  What are you thankful for?  '),(12803,1143104,'04,September,2003','2003-09-04 00:00:00','   ...cell phones.  What are you thankful for?  '),(12804,1143104,'02,September,2003','2003-09-02 00:00:00','   ...a wonderful girlfriend. ...96.7 KCAL ROCKS! ...Auto Zone.  What are you thankful for?  '),(12805,1143104,'28,October,2003','2003-10-28 00:00:00','   ...having a home. ...Kudos bars. ...my health. ...increased determination. ...self direction.  What are you thankful for?  '),(12806,1143104,'09,October,2003','2003-10-09 00:00:00','   ...realizing that \"short bus\" jokes aren\'t funny. They\'re just cruel. ...studying enough to  probably  get an A on my Russian quiz. ...spending the weekend with Patti.  What are you thankful for?  '),(12807,1143104,'06,October,2003','2003-10-06 00:00:00','   ...cell phones. ...new bumpers! ...AIM ...keys.  What are you thankful for?  '),(12808,1143104,'25,November,2003','2003-11-25 00:00:00','   ...a national holiday dedicated to this blog. ...supportive parents. ...good food. ...a wonderful girlfriend. ...friends who don\'t give up on me even when I\'m out of touch for a really long time (Anjane, Will, Jill & Paul, etc...). ...being (mostly) down-to-Earth.  What are you thankful for?  '),(12809,1143104,'23,November,2003','2003-11-23 00:00:00','   ...Patti.  What are you thankful for?  '),(12810,1143104,'18,November,2003','2003-11-18 00:00:00','   ...I don\'t know.  What are you thankful for?  '),(12811,1143104,'11,November,2003','2003-11-11 00:00:00','   ...self-restraint of jealousy. ...good driving skills. ...a great girlfriend. ...chocolate chip cookie dough.  What are you thankful for?  '),(12812,1143104,'03,November,2003','2003-11-03 00:00:00','   ...computer raffles. ...camping trips. ...cheap windshield replacement. ...objective perspective. ...good sex.  What are you thankful for?  '),(12813,1143104,'24,December,2003','2003-12-24 00:00:00','   ...good food. ...great coworkers like Andrew and Josh. ...Trader Joe\'s truffles  What are you thankful for?  '),(12814,1143104,'31,January,2004','2004-01-31 00:00:00','   ...apple Martini\'s. (Does that make me gay?) ...free drinks. ...good times with a friend. ...Live. ...gifted songwriters. ...crisp evening air.  What are you thankful for?  '),(12815,1143104,'28,January,2004','2004-01-28 00:00:00','   ...helpful and caring Russian professors. ...good friends. ...Disturbed. ...basic automotive knowledge.  What are you thankful for?  '),(12816,1143104,'25,January,2004','2004-01-25 00:00:00','   ...Patti. ...Lambda Chi. ...a new suit. ...Clark\'s 3rd place win in New Hampshire.  What are you thankful for?  '),(12817,1143104,'25,January,2004','2004-01-25 00:00:00','   ...surprises. ...Scottish themes. ...population control. ...laundry washers and driers. ...comfortable shoes. ...friends who keep trying even when I\'m hard to reach. (Good job Will.) ...Patti. (Happy anniversary babe!)  What are you thankful for?  '),(12818,1143104,'23,January,2004','2004-01-23 00:00:00','   ...political awareness. ...one-year anniversaries. ...self-education.  What are you thankful for?  '),(12819,1143104,'20,January,2004','2004-01-20 00:00:00','   ...soft pillows. ...clear night skies. ...crisp cold air. ...Biblical research. ...Freedom of Speech. ...General Wesley Clark running for President. ( www.clark04.com )  What are you thankful for?  '),(12820,1143104,'20,January,2004','2004-01-20 00:00:00','   ...fraternal bonds. ...a good haircut. ...peace. ...warm sunshine on a clear day. ...solitude with my Jeep.  What are you thankful for?  '),(12821,1143104,'19,January,2004','2004-01-19 00:00:00','   ...understanding. ...discussion. ...love. ...Patti.  What are you thankful for?  '),(12822,1143104,'12,January,2004','2004-01-12 00:00:00','   ...vegetables. ...calendars. ...staplers.  What are you thankful for?  '),(12823,1143104,'11,January,2004','2004-01-11 00:00:00','   ...no damage to the Jeep. ...fun weekends of adventure. ...beautiful wonders of nature. ...patience. ...soft towels. ...sunglasses.  What are you thankful for?  '),(12824,1143104,'04,January,2004','2004-01-04 00:00:00','   ...a sense of accomplishment. ...restful days. ...new beginnings.  What are you thankful for?  '),(12825,1143104,'03,January,2004','2004-01-03 00:00:00','   ...safe interstate trips. ...enough spending cash. ...safe vehicles. ...the beauty of America.  What are you thankful for?  '),(12826,1143104,'25,February,2004','2004-02-25 00:00:00','   ...Andrew\'s advice today. ...rain. ...Disneyland. ...West Wing. ...nice cards in the mail. ...class action lawsuit settlements.  What are you thankful for?  '),(12827,1143104,'23,February,2004','2004-02-23 00:00:00','   ...accomplishments for a new business. ...a girlfriend who cares.  What are you thankful for?  '),(12828,1143104,'22,February,2004','2004-02-22 00:00:00','   ...plywood. ...rain and the clouds. ...Queen of the Damned. ...reliable mail service. ...e-mail. ...pens. ...trucks. ...honesty.  What are you thankful for?  '),(12829,1143104,'10,February,2004','2004-02-10 00:00:00','   ...new opportunities. ...chances for change. ...friends who respond to e-mails. ...JA sticker giveaways. ...Patti.  What are you thankful for?  '),(12830,1143104,'08,February,2004','2004-02-08 00:00:00','   ...UV Vodka. ...Emilio\'s Beverage Warehouse. ...naps. ...Jeepaholics Anonymous (www.jeepaholics.com) ...Canadian prescription prices. ...Patti\'s creativity.  What are you thankful for?  '),(12831,1143104,'03,February,2004','2004-02-03 00:00:00','   ...making the right decisions. ...improving my life. ...having a wonderful girlfriend, understanding Brothers, and awesome friends. ...a cool Jeep. ...good health (although I could use more sleep). ...determination. ...General Wesley Clark winning Oklahoma.  What are you thankful for?  '),(12832,1143104,'02,February,2004','2004-02-02 00:00:00','   ...loyalty. ...good first days. ...dirt on friend\'s little secrets. ...democracy. ...America. ...Pert Plus. ...good intentions. ...rain. ...blooming roses.  What are you thankful for?  '),(12833,1143104,'31,March,2004','2004-03-31 00:00:00','   ...relaxing rollerblading excursions. ...cool breezes. ...smooth pavement. ...time alone. ...Ozzy Osbourne\'s music. ...Patti.  What are you thankful for?  '),(12834,1143104,'29,March,2004','2004-03-29 00:00:00','   ...sunny days. ...cool breezes. ...shorts. ...my health. ...Metallica. ...Patti.  What are you thankful for?  '),(12835,1143104,'25,March,2004','2004-03-25 00:00:00','   ...granola bars. ...CNN. ...eagles. ...shoes. ...sleep.  What are you thankful for?  '),(12836,1143104,'22,March,2004','2004-03-22 00:00:00','   ...early morning sun. ...towels. ...screwdrivers. ...second chances.  What are you thankful for?  '),(12837,1143104,'17,March,2004','2004-03-17 00:00:00','   ...TABS. ...a functioning CD player. ...a new cell phone. ...Patti.  What are you thankful for?  '),(12838,1143104,'16,March,2004','2004-03-16 00:00:00','   ...objective viewing. ...parties. (Green Party this Saturday!) ...Billy Joel. ...communication.  What are you thankful for?  '),(12839,1143104,'05,March,2004','2004-03-05 00:00:00','   ....soap. ...evolution. ...natural selection. (Although sometimes it takes a little too long.) ...e-mail.  What are you thankful for?  '),(12840,1143104,'03,March,2004','2004-03-03 00:00:00','   ...cool breezes. ...Band-Aids. ...staplers. ...forklifts.  What are  you thankful for?  '),(12841,1143104,'01,March,2004','2004-03-01 00:00:00','   ...Patti. ...Disneyland. ...the Blue Bayou. ...Metallica. ...tires. ...decent pay checks. ...determination.  What are you thankful for?  '),(12842,1143104,'29,April,2004','2004-04-29 00:00:00','   ...tomorrow is Friday. ...sunshine. ...good sleep. ...Patti.  What are you thankful for?  '),(12843,1143104,'25,April,2004','2004-04-25 00:00:00','   ...fighter planes. ...sunny days. ...socks. ...pillows. ...new brake pads. ...Patti.  What are you thankful for?  '),(12844,1143104,'21,April,2004','2004-04-21 00:00:00','   ...positive and honest evaluations. ...raises (although they are piddly poo money). ...St. Anger. ...Patti.  What are you thankful for?  '),(12845,1143104,'18,April,2004','2004-04-18 00:00:00','   ...money. ...education. ...ambition. ...self-reliance. ...Patti.  What are you thankful for?  '),(12846,1143104,'18,April,2004','2004-04-18 00:00:00','   ...ambition. ...new ideas. ...good gas mileage. ...giant wooly mammoths. ...Neosporin. ...Patti.  What are you thankful for?  '),(12847,1143104,'14,April,2004','2004-04-14 00:00:00','   ...Patti\'s brownies!  What are you thankful for? (If you had these brownies, you would DEFINITELY be thankful for them!)  '),(12848,1143104,'11,April,2004','2004-04-11 00:00:00','   ...sunny days. ...polite people. ...minimal traffic. ...economic recovery. ...bagpipes.  What are you thankful for?  '),(12849,1143104,'10,April,2004','2004-04-10 00:00:00','   ...power steering. ...Patti being cool. ...Patti.  What are you thankful for?  '),(12850,1143104,'09,April,2004','2004-04-09 00:00:00','   ...empty houses. ...evening walks. ...Patti.  What are you thankful for?  '),(12851,1143104,'04,April,2004','2004-04-04 00:00:00','   ...more hours in a day. ...longer weekends. ...productivity. ...laziness. ...Patti. ...Jeepaholics Anonymous ( www.jeepaholics.com ).  What are you thankful for?  '),(12852,1143104,'30,May,2004','2004-05-30 00:00:00','   ...long lasting cell phone batteries. ...long weekends. ...toothbrushes. ...computer disks (the portable kind). ...washing mashines and dryers. ...Patti.  What are you thankful for?  '),(12853,1143104,'21,May,2004','2004-05-21 00:00:00','   ...warm clothes. ...a good book. ...cough syrup. ...the internal combustion engine. ...Patti.  What are you thankful for?  '),(12854,1143104,'18,May,2004','2004-05-18 00:00:00','   ...Patti. ...patience. ...cellphones. ...good driving skills. ...calendars. ...Feuer Frei!  What are you thankful for?  '),(12855,1143104,'12,May,2004','2004-05-12 00:00:00','   ...being an American. ...Star Trek. ...comfortable shoes. ...Patti.  What are you thankful for?  '),(12856,1143104,'11,May,2004','2004-05-11 00:00:00','   ...tin foil hats. ...choclie milkie. ...The Cosby Show. ...loans. ...Patti.  What are you thankful for?  '),(12857,1143104,'03,May,2004','2004-05-03 00:00:00','   ...a supportive girlfriend. ...backup plans. ...savings accounts. ...calm responses.  What are you thankful for?  '),(12858,1143104,'03,June,2004','2004-06-03 00:00:00','   ...website bookmarks. ...pressure. ...water. ...Patti.  What are you thankful for?  '),(12859,1143104,'05,July,2004','2004-07-05 00:00:00','   ...Patti. ...hot tea. ...warm blankets. ...4th of July fireworks intermissions. ...Patti\'s constant taking care of me. ...antibiotics. ...acetaminophen. ...loving parents. ...Rite Aid. ...sweatshirts. ...my Jeep. ...Patti. ...e-mail. ...Patti.  What are you thankful for?  '),(12860,1143104,'05,July,2004','2004-07-05 00:00:00','   ...Patti. ...hot tea. ...warm blankets. ...4th of July fireworks intermissions. ...Patti\'s constant taking care of me. ...antibiotics. ...acetaminophen. ...loving parents. ...Rite Aid. ...sweatshirts. ...my Jeep. ...Patti. ...e-mail. ...Patti.  What are you thankful for?  '),(12861,1143104,'29,May,2004','2004-05-29 00:00:00','   Today Nathan and I took a fun little trip to the North side of town to get prices on different things.  After laying on several matresses and finding that we are too poor to buy the ones we (I) like we ventured on to Target, Linens \'n Things, Pier1 and Lowes.  We found many things we liked and wrote our ideas down so we would not forget.  Luckily while we were at Lowes we decided to check out the tile section one more time before leaving... this led us to find the bathroom tile that we really liked that is super inexpensive.  So jumping on this rare chance, not willing to test our luck and their future stock, we bought enough tile to do our third bathroom (Cryntel - Corsica Blue).  Although now that we have all the tile we are not sure which bathroom to do in blue and which to do in black and white.  Lucky for me this will be Nathan\'s decision!  So yippee for us, we are one more step closer to having our building materials all bought. Woo hoo.   '),(12862,1143104,'28,May,2004','2004-05-28 00:00:00','   Patti and I just returned from Lowe\'s where we purchased our new bathroom tiles. So far, its only for the master (upstairs) and guest bathroom (downstairs). For the master bathroom, we chose the  Congoleum - Triumph Natural Stone . The bathroom will be very \"earthy\" with natural colors and patterns. For the downstairs bathroom, we\'re going for a two-tone look (this one is mine!) and it will be black and white. The flooring is  Armstrong - Granville White .  As soon as my parents move out, we\'ll be painting the bathrooms and laying down the new floor. But this is our first house purchase!   '),(12863,1143104,'24,May,2004','2004-05-24 00:00:00','   To me it seems like the next 50 days are going to go by so slowly!!!  I can\'t wait to move into the house with Nathan!  The wait is torture!  Looks like this weekend is going to be a big plannig weekend for Nathan and I! Woo hoo!   '),(12864,1143104,'23,May,2004','2004-05-23 00:00:00','   Next weekend will be another step in making the house a little more \"ours\". Patti and I will be picking up the new tiles for the bathroom floors and also replacing the curtains on the door to the backyard. The guy who fixed the upstairs carpet (and put in the new carpet border downstairs) tore the curtains! We may be changing the tiles for the downstairs bathroom to something stronger and nicer, but we\'ll decide that when we get there.   '),(12865,1143104,'21,May,2004','2004-05-21 00:00:00','   Well I took the paperwork for the Escrow to our agent today.  We are one step closer to owning the house!  Yippeeeeee!  Woo Hoo!   '),(12866,1143104,'18,May,2004','2004-05-18 00:00:00','   Welcome to those of you that dare to venture into the insane world of myself and Nathan.  By now you are aware that we are starting a life together in just a few short weeks and we want you to be a part of it.  Well aren\'t you special! ;-)  If we didn\'t love you we wouldn\'t have told you! Ok well maybe we just want to gloat a little. ROFL.  But anyway, welcome to our site about our new lives.  Hope you enjoy reading about it!   '),(12867,1143104,'18,May,2004','2004-05-18 00:00:00','   Our names are Nathan and Patti. We just bought our first house together and we\'re really excited about it. Between the two of us, we have at least 15 blogs, but none of our blogs that we share are viewable for the public. So we decided to create this as a journal for our friends (that would be you) to check whenever you are absolutely, completely, undeniably bored. After all, we\'re sure you\'re lives are more interesting than this blog, but nevertheless, we subject ourselves to your scrutiny, adoration, criticism, envy, and pity...and it can continue 24 hours a day because the Internet never dies.  Sorry, back on topic. This blog will be about our adventures in home ownership and our lives as an unmarried couple living together. I already told you that we are excited so I won\'t mention it again. (haha)   '),(12868,1143104,'23,June,2004','2004-06-23 00:00:00','   With her training beginning tomorrow, Patti has moved in with a lot of her stuff tonight. It took less than 45 minutes to get it all in the house and about 15 - 20 minutes to organize and lay out. Before she arrived, I cleaned most of the bathroom (didn\'t have time for every square inch), my bedroom, and did a lot of vacuuming and dusting. I also cleaned off my entire windowseat. \"Ever since I was a little girl, I\'ve always wanted a windowseat that I could work on.\" So now she has one.  I did everything I could to create as much space for her as possible so she wouldn\'t feel cramped for the three weeks until my parents move out. Its not a Marriot, but I can still see some of the floor and have access to all my stuff. Worked out pretty well in my opinion. She\'s already hooked up to the DSL router and her stuff is all over the bathroom. With all four of us under one roof, this is going to be an interesting three weeks!   '),(12869,1143104,'21,June,2004','2004-06-21 00:00:00','   Well with New Teacher Training this Thursday I have tons of things to wrap up here in Downey before I move to Moreno Valley.  I have so many things to pack, just clothing is a ton.  Then there will be the frequent weekend trips to pick up forgotten things.  Ugh.  On a positive note I will be signing the loan papers this Wednesday for the house, so things seem to be moving along nicely. I\'m super excited but super stressed... Oh well, what\'s new! ROFL   '),(12870,1143104,'13,June,2004','2004-06-13 00:00:00','   Now that Patti has a job teaching 5th grade at Jackson Elementary in Riverside, schedules have to change a little. She will be moving in to the house in the last week of June and will (unfortunately for her) be here for my parents\' moving out. She\'ll get to witness the chaos and stress first hand. It also means that my room just got more cramped! Yesterday, we went out and purchased a 4\' tall plastic chest of drawers so she can keep clothes here for work. She and I will probably be going back to her parents\' house every weekend in preparation of moving out here permanently.  I have to get ready to share my space with someone else and not really having the space to do it. I have a bathroom and bedroom meant for one. But it will definitely be fun and exciting! I think stuff like this just brings us closer together. (Ain\'t that sappy?) I\'m looking forward to it. As I see it, I get a preview of living with Patti BEFORE escrow closes.. *evil grin* ;)   '),(12871,1143104,'09,June,2004','2004-06-09 00:00:00','   My parents have bought a home in Amherst, NY. Its more than 300 sq. ft. larger than our current house and is BRAND NEW. That\'s the biggest surprise to me. It isn\'t even done being built yet. Construction is scheduled to finish sometime in August, but my parents are still leaving in July and will stay in a hotel until its done. The concern they have about that is what to do with Vickie. I have a feeling she will be staying here, but my hope is that Sue (my Dad\'s co-worker and Vickie\'s former owner) will take her in.  The house is 2 stories plus a basement, with the top floor being a decent sized loft. The basement that is the full size of the house. There are two bedrooms with the loft acting as a third. There is also a kitchen with a dinette and a separate dining room, living room, two walk-in closets, and a two car garage. It comes with a real porch in the front of the house and their new association will pay for all gardening and water. There\'s also a country club nearby that my Mom has already had lunch in.  So progress continues in this tumultuous process. This means Patti and I are one step closer to living together. :) Oh yeah - my parents will have to buy a new stove, but Patti and I get to keep the microwave. Probably for a garage sale.   '),(12872,1143104,'03,June,2004','2004-06-03 00:00:00','   Patti and I just found out that we will be receiving a new (only used a couple times) queen-sized bed with a couple sets of sheets! Kickass! Finally a bed that we can fit on!   '),(12873,1143104,'02,June,2004','2004-06-02 00:00:00','   As I wrote an e-mail to a friend, I realized that not everyone I know will understand the name of this blog (Welcome Home (Sanitarium)). So here\'s a brief explanation...  Patti and I were looking for a name for the blog and wanted something that would reflect who we are but also be somthing \"home related\". After chatting about random phrases, expressions, and word play, I suggested something musical. Namely, something from the realm of Metallica! I sifted through titles and lyrics, but there was one obvious choice: Welcome Home (Sanitarium). For those of you playing along at home, you can find this musical masterpiece on track four of their third album release \"Master of Puppets\" ( Lyrics ). Patti and I are both certifiably nuts, but we\'re both Metallica-loving homeowners so the song works!   '),(12874,1143104,'28,July,2004','2004-07-28 00:00:00','   Two tin cans and a string would be easier to deal with than the crap I\'ve had to put up with lately.  4 states 3 time zones 5 cities 15 people 2 weeks  The premise of a new reality show? Nope! A cool trip Patti and I are planning? Nope!  This is what it took for us to order DSL! Best of all, the order still hasn\'t been placed! That will happen tomorrow morning and once I get the necessary software and hardware, this process will be finished. In my effort to make the lives of my friends (that\'s you) a little easier and my goal of leaving a mark in history, I make the following recommendation:   AVOID  VERIZON   AT ALL COSTS!   Their customer service is atrocious, their capabilities are nothing more than a Neanderthal at Comdex, and their internal computer system is less reliable than two tin cans and a string.  ...in other news...  Patti and I had our first barbequed meal tonight. It was her famous aluminum foil chicken dinner and I had it for the first time. I will definitely request that again. Its really good! Food aside, I\'m 2/3 done with installing the new tracks on the kitchen drawers and Patti has been busy putting shelf liner in and we\'re cleaning a whole bunch of pots, pans, dishes, and glasses everynight. She broke her crockpot tonight and had never even used it before. We also bought two new cordless phones last night from Target and they work pretty well. I like \'em and that\'s teh important thing because she didn\'t seem to care. Shopping was fun for that!   '),(12875,1143104,'21,July,2004','2004-07-21 00:00:00','   Its been too long since the last update to this blog. As you might imagine, Patti and I have been extremely busy with work and, of course, making our new home reflect what we\'d like it to look like: ours.  One of the most recent things we\'ve done is a trip to Lowe\'s to get some paint and get started on changing the internal colors of the house. Patti got the idea to paint the walls behind the&nbsp;laundry and the refrigerator a bright white in an attempt to lighten the areas. It has worked considerably for the area behind the fridge and the washer/dryer area is still to come. She also had some designs for her closet. After doing a considerable amount of cleaning of the walls and shelves, she and I ripped out the carpet on the floor of the closet and replaced it with a nice piece of pine. Then the whole closet got painted white (same as the fridge area) and it looks awesome. We\'ve decided that the whole bedroom will now be white because it looks so good.  For me, the projects have been somewhat less successful. My first mission was a new dog door for Blacky to fit through. A sawzall would have come in VERY handy for this project. But instead, I went ghetto and had to chisel through wood with a hammer and screwdriver. After hours of back-altering work, its about 75% done but easily usable and Blacky has adjusted very well. He hasn\'t moved yet, but at least we know that when he does, we won\'t have to worry about him using the door. My next project installing a new lighting fixture in the entryway to the master bedroom. I attempted that at 8:30pm and called it quits around 9:30 when things just didn\'t seem to be fitting right. I may end up taking it back to Lowe\'s and getting a smaller one more appropriate for the location.  So in summation, Patti\'s projects are going great and making things look better while mine seem to serve no other purpose than stripping my masculinity and leaving me with failure. (Was that over dramatized?!) But throughout all of these projects, we\'ve been dealing with the details like utility companies, washing the dishes that have come out of storage, working and going to school, and ordering lots of take-out. I\'m also still the Yard Commander and in charge of insect patrol, lawn maintenance, tree trimming, etc.  Its been a lot of hard work so far but I wouldn\'t trade it for anything. I think we\'re having a great time with the projects and living together has been wonderful so far. We will be moving all of the furniture this weekend and my friend Andrew has been kind enough to offer his assistance. Hopefully, once that final phase of major moving is completed, we\'ll be able to put the house together the way we want it.  After all is said and done, and we finally have settled in, we will be sending out invitations to our housewarming party. We\'re really looking forward to having all of you come and visit the new home and drink with us. I\'ve already started preparing the alcohol supply for a diverse and prolonged necessity of thirsts. :)   '),(12876,1143104,'12,July,2004','2004-07-12 00:00:00','   Escrow will officially close tomorrow and my parents will become squatters! I told them we would charge them rent if that made them feel more comfortable and respectable. ;)  Patti and I will pack up her room at her parents house on this upcoming weekend and then have all of her stuff (bedroom and storage) moved in by the end of the month. It will take a few trips, even with the Jeep, truck, and rented trailer, but it will finally be done. Hopefully, the housewarming party will happen by the end of August. We\'ve already come up with a cool contest idea for the party (not sharing it with you yet) and our wish list for the house is enormous.  Patti and I have been putting a lot of time and effort into the landscaping of the house and preparing it for some of our ideas. She\'s done an incredible job working over a lot of the land and cleaning things up. We estimate that we can start our cactus garden in 2 - 3 weeks and we are looking at ways of removing the tons of lava rock all over the property. I\'ve been on bug detail for a couple of months and go around smelling like pesticide. Everytime someone spots an ant anywhere outside, I\'m called in with the sprayer to be John Goodman. Its fun and chemically educational!  Other than that fun stuff, my parents things get packed up on Wednesday and shipped out on Thursday. They\'ll be staying in a hotel until the 21st and then they board a plane for Buffalo. On the 16th, they\'ll be returning to the house to help do a lot of the cleaning (since all the furniture will be gone). Oddly, the house will be nearly devoid of furniture for over a week since we\'re not moving Patti\'s stuff until the weekend of the 23rd. I called most of the utilities today and taken the necessary steps to have the accounts put into my name. But it looks like we\'ll save a nice chunk of change if we switch to SBC for phone service and get their DSL service. I\'ll tackle that tomorrow.   '),(12877,1143104,'01,August,2004','2004-08-01 00:00:00','   Patti and I did a huge amount of shopping for both groceries and home improvements. We started at Target with a list of four things we wanted to buy there. Big surprise - they didn\'t carry a single item. That sucked, but since a couple of the items were for Blacky, we just went next door to PetSmart and got them there. One of the things Patti bought was a new ID tag for Blacky with his new address. :)  After that, we headed to Lowe\'s and just went crazy. We\'re finally getting around to working on the master bathroom and wanted to buy all new stuff for it. That includes lighting fixtures, accessories, and paint. Not only did we get that stuff, we also bought a ton of other stuff for some other household projects and after purchasing the items, realized how poor we really are. LOL. :) Damn houses are expensive!  After that, we did a bunch of shopping for groceries at CostCo. I pointed out to Patti that we had reached our monthly grocery budget and could no longer buy food for the rest of August. We\'ll just have to starve. Damn food is expensive!  The good thing is that we are keeping good records of our purchases and staying within our budget (as best we can). Since this is the first month of budgeting and record keeping, I think its going along nicely. I\'ve created some useful Excel spreadsheets to aid in the process. I strongly recommend doing that for anyone keeping a budget.  NEwayz, the house is coming along nicely and we look forward to getting it \"done\".   '),(12878,3364899,'29,May,2004','2004-05-29 00:00:00','    Hey All  ohhhh LORS you got it all rong, this blog is not nearly pointless, it is absolutly pointless....hihihihi......damb u and ur not going to skool ways.......ohhhhh skool is realy givin me this shits now, with god damb exams and butshit SACs and just overall fuckheads for teachers ( and the fact that i dont do any work is not the reason to their bitchyness.......i just dont know y they dont like me....lol...)   Anyway today was boring as hell we had a Garage Sale with my family and my Auntie, we sat there for 8hrs and sold absolute shit ( i dont know y stupid ppl buy it ).....ohhhh it was bad the sale started at 8am in the morning and there were all these bagans lining up to get in at 6:30am, it was bad when we opened the gates i thought we were going to see some full on bogan fights......ohhhh but we got over all over $800, my family got about $300 and the rest was split between my auntie, another autie got $100 and a friend of the family got $100 ( we sold stuff for them ).....ohh but it wasnt all bad i got to hang out with my cuzs all day....then i went and watched one of my cuzs play netball ( her team won 43 to 13 ) and perv on all the hotties running around ( hell ye )......  Well tomozas going to be ok i get to go see my cuzs again cas its another cuz\'s 21st....so im just going to get smashed wif my cuzs even though it is a luch party thing.....its ok ill just start at 12 and go through untill 4 in the morning ( i hhave to get some sleep coz i have a SAC on Mon but its last period so ill sober up by then...lol ).......   And ohh LORS i dont think we have to worry about HARRY and CHARLIE.......coz that is never going to happen i dont think....sorry HARRY but get some balls and do something....in the words of STIFLER \"locate your dick, remove the shrink wrapper and you use the fucking thing\"........  ok well im off  CYA   '),(12879,3364899,'24,May,2004','2004-05-24 00:00:00','    hey  there goes my perfect record of a msg a day ...... damb week ends..... ok well my weekend sucked coz BRISSY lost and lost half their players while doing it.......ARGGGGGGGGG.........skool sucked as usual and nothing else is happenin in my life..........ok thats it  CYA   '),(12880,3364899,'21,May,2004','2004-05-21 00:00:00','    ohhhh  KILL ME PLEASE!!!!!!! i had a shithouse day at school nothign fun happened there just a hole lot of crapy classes.......ahhhhhhhh.....and then i came home andmy mum was having friends over for dinner, so me and my sis had to go get food and we made 13 trips to different supermarkets because they didnt have the right food at the other ones....argggggggg.....and then i came back and had to look after this lil boy all night, who just wanted to play Harry Potter all night on Xbox but was scared of the things so i had to play the hole game while he watched.....boring and i didnt get paid and was only allowed to have one beer cas the lil boy thaught cas i had it he could so it was Coke for me....i couldnt even get pissed to get rid of the lil shit......and now im tired as hell and cant be fucked doing anything....... so  CYA   '),(12881,3364899,'20,May,2004','2004-05-20 00:00:00','    Hey all  Wow today was a full day...a full day of absolute crap.....maths i hate...bio we sat there and made sure the test were in the test pacs (great work for the exam revision).....accounting i god war craft off LINO...and i.t i made my backgroung for my computer (all you guys are in it)......and thats all....ohh wat fun..... thanx LAURA for the comment, but i know im cute, you dont have to keep reminding me.....my sis went to a 70\'s dress up party and the wig was a lil bob cut thing, but by the time i got to it the next day it was all messed up but i couldnt take it off, i wana get one and just walk round the city and shit with it on......it was soo mad lol...... ok well i got to go  CYA   '),(12882,3364899,'19,May,2004','2004-05-19 00:00:00','    Hey all  well went to skool again and omg JOJO got a date for the formal....i think my threat worked.....you know the one from the other post, the one about HARRY that made DILLON throw up and miss 2 nights of sleep......ohhhhhh and also we read a lovely book in lit today about men that fuck cows and the teacher just had to tell us she was related to a guy that has also fucked a cow ( highlight of the day, no seriously it was ) and then had assembaly and NOONAN came and sat next to me AGGHHHHHH talk about scary, and thanx MATTY for making me piss my self while he was next to me i thought he was going to turn around and kill me.... ok well  CYA   '),(12883,3364899,'18,May,2004','2004-05-18 00:00:00','    Day 2 inside this box.....owwww wait thats just my jumper *pulls head throught the head hole thingy*.....thats better......fuck how cold is it.....im freezing my ass off plu i had to have a double of BIO in that freezing cold room where its already -100 degrees.....Brrrrrrrr......  Wow wat an exiting day today was...1st i woke ( i know that was so cool ) and 2nd i went to school ( double the exitment there ) and now im here...\"wow lets do that again\" Donkey from Shrek.....and if you noobs cant tell im bein sarcstatistic......thats a word like distapeardeded and majigacical.......  so not a hell of alot happened today ohhhh well theres allways tomorow....... AGHHHHHHHHH Not tomorow noooo y god y.........thats a school day.....damb this world we live in....  ok well  CYA   '),(12884,3364899,'17,May,2004','2004-05-17 00:00:00','    Hey Everybody  this is my first post for my journal, before we start if you dont like spelling mistakes, ahem HARRY, you can go get fucked coz ur going to find a hell of alot here... Well today was another boring school day, HARRY\'s favourite, and as useual i did nothing....ohhh but then i got to come on here and get my head ripped off by DILLON for stealin his sayins on MSN.....and im going to the slamer for that coz Marbles has studied Law for a semester at Monash....sooo ill see you wen i get out....  WOW that was quick... So now im going to go on about how good BRISSY LIONS are, this is for you LEESH and how crap COLLINGWOOD are thats for you JORDAN.....hihihihihihi 1-7 whats going on wif em...but brissy are going well but we just have to kick straight or we are going to get killed......  Formals coming up and yes DILLON and NOLA you have to come or i will kill you both...and TIAGO you have to find a date or we are going to set you up with some fat chcik or HARRY dressed in a skirt...so get one  Ok well i\'ll try post most days  CYA   '),(12885,3364899,'24,June,2004','2004-06-24 00:00:00','    Hey People  Do you like the name of this post LORZ....hihihih......by the way, even though i already told you......but damb girl......u do look awsome in those pics......Dung\'s going to be kicken himself ( wink wink nudge nudge )....... Well today i had spiritualalalaittitty day....and apart from Glen Manton giving us a preety kool speach.....it was crap.....and i still have 3 hours of it tomoza......ARGGGGGGGG........ Cant wait till i get to go to BUNDALONG......still about 6-9months before i go but im soooooo exited........cant wait to meet DOC and the Bundy 9.........ohhh and the Bundy Games........YEY.......GO DIL......... ok well i\'ll go  Cya   P.S Dam girl Dam ; )   '),(12886,3364899,'22,June,2004','2004-06-22 00:00:00','    mmmmmm hey  well today i had shity skool wif a twist......crapy classes at the start of they day then fiished wif speeaches from career ppls....i wen to the sports one and cops one...and yey for me i got to get up and say thanks for coming, now fuck off......owwww that was boring.......owwwww another thing yesterday i got a pic of DILLON doing the mclinsky Eatin Leamons And Likeing It face......hihihihi......it is classic.......  and LORZ ur going down on JANES site......  CYA   '),(12887,3364899,'20,June,2004','2004-06-20 00:00:00','    Wow it has been a long time..... ummm lets see wat has happened...... ohhhh had a maths sac that i missed then fucked up the second part of so my parents were cut and i couldnt go to LAURA or HEBA\'s afterparties GRRRRRRRRRRRRRR.....so im dropping maths if i can YEY.......ummm well i gess u know LAURA went to see Harry Potter with you but it was sold out so we couldnt see it =( but then i got to be an adult and a child wif the triple m peoples hihihihihihihi.....ohhhh and then the best knews of all i found out 2.....yes 2....people want to go out wif me....mmmm lucky me......damb shame they are both about 40, 1 is a male and they are LAURA\'s parents hihihihihih.......ummmmm.......went to see Shrek 2 tonight that is hilerious....i love the gingerbreads man and the giant gingerbread man.....hihihihi....NOT THE GUM DROP BUTTONS...... well then......oh ye my exams.....looks like i thaught i went well on BIO but fucked it up and ACCOUNTING i had no chance wif that......sooo ive come to the conclusion that im just going to pass yr12.....fuck me enter score off cas i dont want to do any more schooling......and joing the Police or the Army....probly the police cas im more of a city guy i couldnt gop off to the army by my self i dont think.....i would miss SOME of you guys too much.....enfasis on the SOME.....because i dont give a fly fuck about you LAURA.....HIHIHIH.....just playing bubz......im never going to hear the end of that one but you know LORZ who the SOME are dont you.......thoses guys from that school....you know them.......ok well i betta goesssss.... CYA   '),(12888,3364899,'02,June,2004','2004-06-02 00:00:00','    HEY  ARGGGGGGGGGG, damn you RACHEL making me walk for 4 hrs, my feet kill now and i have about 5 blisters on each foot.....and after all that we didnt get RACHEL any god damn shoes.......for all you who dont know wat i did today i went ova to RACHEL\'s and she doesnt like public transport and likes to walk and she needed to get some shoes for the formal, so we started at the corner of Victoria St and Melvil Rd then walked down Melvil to Morland Rd (2-3km) then up Morland Rd to Sydney Rd(2-3km) then up Sydney Rd to Belll St(4-5km) then back down Sydney Rd to Brunswick St (10-15km) then up Brunswick St to Lygon St (2-3km)then down to the end of Lygon St(8-10km) then back up Lygon St (8-10km) then down to RACHEL\'s house (12-15km) ohhhhhhh i want to kill after that......  ok well apart from that i went to school with a hangova yey lol and then had to do a SAC lol....u got to love 21st....me and my cuz just drank and drank... and now i want to work at the anglers tavern because they have the hottest waitresses ive ever seen......i was talking to one of them at the end of the day about getting a job when at the top of his lungs my cuz yells \"OI MARCUS STOP TRYIN TO PICK UP AND GET UR ASS OUT HERE\" lol i just pissed myself laughin.......  ohhh ye after all that walking we still didnt get any shoes for RACHEL.....  OMG u are like the richest person i know LAURA...and to top it off you got 25c for a colouring in book.....you now how to sell it lol.......seein that i think u are the only one that reads this blog, all this is for you benifit ( or bordom ), so how ya been...... lol  ok  CYA   '),(12889,3364899,'24,July,2004','2004-07-24 00:00:00','    hey  dont you ppl know......I LOVE LAURA.......thats y i write about her and to her on this site......not because shes one of my best mates and not&nbsp;before JANE put my link on her page she was&nbsp;probly the only one to read it.....NA none of that shit its because im obsessed with her......... who eva you r&nbsp;u r A NOOB!!!!!!! anyway....... im at NICKY\'s place at&nbsp;the moment and i happy coz im&nbsp;going to get PLASTED in about and hour....YEY..... ye i had a kool time last night...... saw JESS.....YEY....hihihi....... ye but LAURA (there i go again talking to the appel of my eyes...hihih) so many other ppl have said the oppisite to you......so im starting to worry about u and ur crazy, kinky ways....... i think youve change...... if asked you that in the past you would have been like no dont do it....... but now wif your permisques (i cant spell) ways your like ye go for it...... WHATS HAPPEND TO YOU.....hihihihi...... ye well im going to get drunk&nbsp;now.......for you guys at skool who read this......i&nbsp;might be and hoping that i am hung over at&nbsp;skool....hihihi ok  CYA &nbsp;   '),(12890,3364899,'21,July,2004','2004-07-21 00:00:00','    hey ey hey  lets get this setteled... i didnt steal your crapy MSN name....DILL gave that to me.....and for you to think that your important enough for me to steal something off is just hilerious....your deluded girl......and do i care how to spell Lowther Hall no only u r sad enough to pick upo on that u lost case.....so how ya been anyway....do u wana go do something soon.....hihihi  ye well today haz been boring double SAC then an assembaly....ow wat fun that was.......cant wait for the party it is going to be the best.......  anyway i got nothin else to say so im out of here....  CYA   '),(12891,3364899,'19,July,2004','2004-07-19 00:00:00','    OI LAURA......u suck.....steal all my shit......u betta be ready to get ur ass kicked......ok ok u probly will hurt me more than i hurt u but still \"IM NOT HAPPY JAN\"........... ye but anyway....hey all........ wats happened to me......well i went out and got $200 of new clothes this weekend.......YEY YEY YEY.......that was about all.....and BRISSY killed COLLYWOBBELS and officaly put them out of the finals......YEY......GO BRISSY........ ye well this week is going to be shit coz i got skool but i get friday off.....but then im going out this weekend wif NIKKI the man......to the Lother Hall after party........hihihi thats going to be funny.....im going to piss myself if i see KAIT ( the sluty slut slut HO ) or ELISE ( the sluty slut slut HO\'s SLUT ) hihihihi........as NIKKI advised just go up to em and be like \"Hey SLA LA LA LA LA LA LA TTTTS\" hihihihihi........ye but that should be fun......... well ishould be doing BIO at the moment but cant be fucked and im probly going to be killed by the teach because of it but FUCK IT......... ok well im going to go ok  CYA  &nbsp;   '),(12892,3364899,'12,July,2004','2004-07-12 00:00:00','    Hey you again.....another profile on me.....  LAYER ONE: â€” Name: Marcus Neil â€” Birth date: 27 November 1986 â€” Birthplace: Melbourne â€” Current residence: Strathmore, Melbourne â€” Eye Color: Blue â€” Hair Color: Brown â€” Height: 6\'0 - 6\'1 â€” Righty or Lefty: Right-o â€” Zodiac Sign: Sago  LAYER TWO: â€” Your heritage: 1 16th English ( Manchester ), 1 16th Italian, and theres some Irish in there.... â€” The shoes you wore today: My black school shoes YEY â€” Your weakness: fooooooood â€” Your fears: that im not the sexiest guy alive.....ohh wait i am......fear gone.... â€” Your perfect pizza: Meatlovers with BBQ sauce and extra chicken  â€” Goal you\'d like to achieve: to become a coppa   LAYER THREE: â€” Your most overused phrase on IM: ??? or What was the question  â€” Your first waking thoughts: hey buddy your back again...but i have to get out of bed now so go away..... â€” Your best physical feature: apart from being the sexiest guy alive.....um my arms and legs are toned and i like them ( thanks swimming )  â€” Your most missed memory: i cant really miss a memory coz it always there in my head....  LAYER FOUR: â€” Pepsi or Coke: Coke....mmm Cherry or Lemon or Lime Diet Coke â€” McDonald\'s or Burger King: mmmm its Mac time  â€” Single or group dates: Single â€” Adidas or Nike: nike  â€” Lipton Ice Tea or Nestea: Dont really want either so ill take a VB mate.... â€” Chocolate or vanilla: Chocolate  â€” Cappuccino or coffee: Dont drink the shit....so another round of VB\'s mate.....bar tender \"your the only one here mate\"......ye so give me a round....  LAYER FIVE: â€” Smoke: occasionly â€” Cuss: dam fucking straight motherfucker â€” Sing: all the time.....you are talking to the Aus Idol Wanabe â€” Take a shower everyday: ye i try.... â€” Do you think you\'ve been in love: Yes....im still in love....ohhhh sweet food.... â€” Want to go to college: nope.....cant stand school at the moment so y would i go to Uni.....im going to the police force.... â€” Liked high school: Ye and No...it has its ups and downs â€” Want to get married: someday....long way away  â€” Believe in yourself: ye.....as most now im up myself....hihihi â€” Get motion sickness: No â€” Think you\'re attractive: Well im up myself so.....na im a dog dude.....hihihi â€” Think you\'re a health freak: fuck no.......food gim me gim me gim me  â€” Get along with your parent(s): at least one of em.....sometimes â€” Like thunderstorms: ye i love the cold weather....get to stay home watch movies...oh ye POPCORN...... â€” Play an instrument: Vocals  LAYER SIX: In the past month... â€” Drank alcohol: YEP â€” Smoked: NO â€” Done a drug: alcohol, weed, nicatine, wiz fiz.... â€” Made Out: ye...not the best kiss.... â€” Gone on a date: I guess you could call it that.... â€” Gone to the mall?: Unfortinatly..... â€” Eaten an entire box of Oreos?: No....but oreos are the least of my worries..... â€” Eaten sushi: ewwwwwww â€” Been on stage: yep â€” Been dumped: No â€” Gone skating: No â€” Made homemade cookies: ok i like food...but me cook ur joking..... â€” Gone skinny dipping: Nope.....ohhh the good old days  â€” Dyed your hair: Nope but just got the crapest haircut....... â€” Stolen Anything: *shifty eyes*.....every girls heart...ohh my god am i up myself.....   LAYER SEVEN: Ever... â€” Played a game that required removal of clothing: YEP â€” If so, was it mixed company: YEP â€” Been trashed or extremely intoxicated: Im drunk at the moment i think.... â€” Been caught \"doing something\": almost â€” Been called a tease: no but like every chick is...... â€” Gotten beaten up: Yep....damn you Laura....hihihi â€” Shoplifted: *shifty eyes*....... â€” Changed who you were to fit in: once but it neva works.....i tryed being ugly but to no avail   LAYER EIGHT: â€” Age you hope to be married: 30-35 â€” Numbers and Names of Children: 2....names???? im a dude....i harldy know the name of the guy next to me..... â€” Describe your Dream Wedding: In the prk in the city.....ohhhh then have FOOOOOOOD...mmmmmm  â€” How do you want to die: i dont want to..... â€” What do you want to be when you grow up: COPPA â€” What country would you most like to visit: Bundalong....England...and Ierland....ohhh ye and Italy....  LAYER NINE: â€” Number of drugs taken illegally: 2 â€” Number of people I could trust with my life: are there any.......maby you laura but if you had to save me from falling off something i would probly pull u with me.... â€” Number of CDs that I own: about 30 â€” Number of piercings: 0 â€” Number of times my name has appeared in the newspaper: about 10-15 times...go Sport....oh ye and that murder charge...hihih  â€” Number of things in my past that I regret: to many things  ok well thats me again  CYA   '),(12893,3364899,'08,July,2004','2004-07-08 00:00:00','    SOWY SOWY SOWY...... ye well wat i sed was wrong...and im sowy...... ok well im out of here  CYA  P.S SOWY SOWY SOWY   '),(12894,3364899,'08,July,2004','2004-07-08 00:00:00','    Hey this is all about me....ye i stole it off u LOZ but u stole it off JANE who stole it off SOMEONE......  GENERAL -name? Marcus Adam Neil, Mickey Dee and Dolph Ray (in the porn industry) -d.o.b.? 27 November 1986 -place of birth? Melbourne -places of residence? Strathmore -religion? i dont have one -occupation? a Bernard and a Bogan APPEARANCE -hair? Brown, sruffy but straight  -eyes? Blue -height? 6\'0 or 6\'1  STYLE -clothing? Surfy stuff -music? ROCK, but do listen to anything exept classical and country  -bodyart? a keg and two man boobs....hihihi -wearing? black underwear and a pair of jeans.... -listening to? Maroon 5 - She Will Be Loved -thinking of? why th hell my shoulder is killin me -feeling? sore and happy (talking to DIL...funny man) LAST THING YOU... -bought? a ticket to the movies and a hotdog  -did? broke a POT glass...argggg -ate & drank? a penutslab (like nutbreak chocy) and a pot of beer (hense breaking the glass)  -read? wat DIL wrote to me on MSN -watched on tv? Shane Crawford Uncut....a lil like BB Uncut EITHER / OR -club or houseparty? PUB if that was one of the choises but nooooo.....so club.... -tea or coffee? Dont drink the shit  -high achiever or easy-going? easy going -drinks or shots? drinks (POTS) -cats or dogs? all animals exept small dogs cas they r just shits  -pen or pencil? pencil - to draw stuff  -gloves or mittens? gloves even though only queer guys wear em  -food or candy? CANDY IS A BLOODY FOOD -cassette or cd? cd -snuff or cigarettes? cigaretts -coke or pepsi? coke...Cherry coke and lime and lemon diet coke  -hard or mild alcohol? hard wif coke on the rocks  -sunset beach or the bold and the beautiful? ummm bold and the beautiful for \'THE DRUGS\' song -rickie lake or oprah winfrey? go ricky.....go ricky WHO DO YOU WANT TO... -kill? mmmmmm......dont know -screw- famous? Jess Alba  	non-famous blokie? ohhhh how can i go past Jordy lol 	non-famous girlie? prefer not to say....LORZ u know who....u...hihihi -slap? Steave Erwin -hear from? Jordan -get really wasted with? Jordan -tickle? Rachels fun to tickel....but ill say Jordan...hihihi -look like? Ive benn told that i look like Prince William....but it was an old wog lady who i think was blind...hihihihi -be like? I wana be like Mike...hihihih -avoid? once again LORZ knows that one..... FAVORITE -food? Italian  -drink? VB -color? Red.....but brown for clothes  -album? Songs for Jane - Maroon 5  -shoes? my brown Enties  -site? Janes  -dance? DIRTY.....hihihi -song? She Will Be Love - Maroon 5 -store? Jetty Surf...Strapa Surf.....Genral Pants Co..... -place to visit? i would like to go to Bundalong........neva been anywhere else LAST TIME YOU: -were kissed? by a male or female......lol......2 weeks ago...too long..... -were held? neva im always the holder....im too big i dont think anyone can hold me  -cried? about 5 yrs ago  -smiled? today -laughed? today -cuddled? my new cat at 7 wen it woke me up after 15 mins sleep  -sung along to a song? right now....im an Aus Idol im always singing  -danced? a week ago  -said i love you? ages ago......  ok well thats me  CYA   '),(12895,3364899,'05,July,2004','2004-07-05 00:00:00','    hey  ye \'The Day After Tomoza\' did the same thing to me LORZ.....its freeky......and thank u i knew \'Spiderman 2\' would be bullshit.....I can go rub it in OLLIE\'s face HIHIHIHI.......I get tyo go to both partys now YEY........when do u get back LORZ....coz i tried callin u and u neva pick up GRRRRRRRR........ well my life is boring i just play wif my new kitten i got YEY....its soooooo cute......when i go to bed i opens my door (my door doesnt shut properly and u can just push it to open it (smart cat)) then it jumps on my bed and gets under the covers wif you.....its soooo cute....plus its like having a fury hot water bottle...... ok well im going  CYA   '),(12896,3364899,'02,July,2004','2004-07-02 00:00:00','    mmmmmmm well wat have i done.....well i went to see \'The Day After Tomoza\' with DILLON and then met his phsyco cat that is sooooooo cute and his lil puppy that is crazy......then i went out wif JESS and RACHEL to THE NEXT BLUE, omg did that suck ass, the dj was spaz, he was trying to scratch the record but was doing it in the totaly rong spot......then i went to NOLA\'S and we watched some tv and movies and had some YUMMMMMMMY pizza......then got my new pussy cat JOELY, she is soooooooo cute.........thats about it.......then i might be going to DUMMY\'S party.....but i have another party on the same day sooooo i dont know where to go......i want to go to DUMMY\'s to see JESS ( hihihih ;p ).....but then i really want to go to the other one cas ive known the other person for eva and it should be an awsome party.......... I DONT KNOW......... ok im out of here..... CYA   '),(12897,3364899,'03,August,2004','2004-08-03 00:00:00','    hey you guys well where can i start.....did i write about the afterparty that was yonks ago ye or am i that far behind.....well it was ok....me and NICKY stayed for like 20mins then head back to GEM FRIEND\'s place and did shit all there but it was an ok night.......then i was fucked for skool coz i got no sleeps........and then had too much homework but in my style....it didnt get done.....YEY.......and then had a boring week of skool only higglight was a techer that was pretty good lookin....she looked like she was about 25 then we found out she\'s 38 and we\'re like fuck she awsom (for her age)........then i signed up to backstage for the skool prodution \"Guys And Dolls\"......so i was at taht all day Sunday....that was kool......got to admit there are some hotties there even tho they r sooo young......all the gals look older than they r......there are ones that look older than me and they came and talked to me and i find out they r like in YR 9 &amp; 10.......BAD MAN MARCUS BAD MAN.......its just freaky.........and then had skool Monday......and the production after skool......once again the gals were cool......(ok imtalking bout the gals alot.....but they r sooooo nice even tho the rumer at SBC is that they r bitches.......and i go to an all guys skool its not everyday i get to hang out wif em like this).......the guys are kool to.......but the gals are beta.......then Today.......had skool again......then production.....and guess wat.....the chicks were kool.......honestly there was a thing at DUMMY\'s party where i was talking to TAM and coz i was talking to her everyone thought that i liked her.......if that were true look out female cast members of \"Guys and Dolls\" coz i must love you all.......hihiiihihihi that is such a hoot.........ohhhh and tomoza i get to go the production AGAIN......matenay tomoza so we have the primary skool there....its going to be funny coz the gals have to do a scene in these lil skimpy playboy bunny wanabe outfits.......the kids are going to be like mummy i saw sex today.......lol......lil kids are funny.......ye well i betta go....... ohhhh i forgot something.....i forgot to mension LAURA......ohhh there we go ok CYA   '),(12898,3509141,'31,May,2004','2004-05-31 00:00:00','   This week is gonna be hell, but then look at the bright side, 9 more days of school left.  June 2: English written exam June 4: English ISU June 8: Drama exam June 16: English exam June 18: Business exam June 25: Report cards June 27: My 16th birthday  I feel kinda weird now, because i\'m sitting here waiting for Alex to call, i\'m turning into one of those girls whose life is based on someone else, it\'s sorta a good feeling, i mean, i really like him.  but on the other hand it\'s not, because i love the sense of independence that i use to have, i miss it, but i like how things are like now, i wouldn\'t trade it for the world.    I should be doing my career studies homework, but i can\'t concentrate, and for the first time listening to jay chow doesn\'t help anymore.  i\'m sitting here talking to mike about the one thing that i wish i didn\'t take, but it\'s not helping.  School is acutally getting fun, it\'s probably because i get to see Alex, but i\'m actually enjoying the classes now, well, it\'s about time, i would say.  My lip is getting better, i should thank someone for that.  *wink* Today\'s weather was good, only if it was a little warmer, or i had a jacket or something, i love the sizzling rain, but i got a lot of goosebumps.  i\'m waiting for baseball to come on, yes, i\'m a toronto blue jays fan, but it\'s not on until 10, 20 minutes away.    apparently, i need to get my act together, i\'m getting a 74 in business, and according to my parents that\'s not good enough, not even near good enough, in this house, 80 is passing, and if you don\'t want any lectures, about 90 is what\'s expected.  But hey, i\'m not that smart, and my mom knows taht, i think she\'s just lecture me a bit, that\'s all, hopefully.  now, Alex called, got to go.  my priorities are straight, i think.  '),(12899,3509141,'31,May,2004','2004-05-31 00:00:00','   I finally decided to have a blog... maybe because i have nothing better to do at this very moment.   i think the fact that all four of us got blogs are so that we can stay in touch with each others\' daily life.  I\'m sitting here, seem to be waiting for something, don\'t really know what, don\'t really care either.  I sometimes don\'t understand why people let some of the things get to them, they all seem upset, angry... sometimes i think it\'s not what\'s really happening, just what they think is happening, it\'s all a matter of reception.  now i sound like someone who has no problems, but trust me i do, we all do, but the point is not let the problems lead to other problems, the trick is how to solve it, scream, swearing are all just temporary methods people use, but the real problems still remains, i know i\'m making it sound too easy, but the fact is that it\'s not that hard if you just jump out of the box.    i should be a phsycologist ( if i learn how to spell it first), it hurts me to see people struggling.  '),(12900,3509141,'27,June,2004','2004-06-27 00:00:00','   well well well, it\'s finally my birthday! i finally made it to 16!  i\'m getting my G1 driver\'s lisence next week, then i can start driving, so exciting.  thank your for all the phone calls, emails, hugs to all the people who remembered my birthday.  i\'m waiting for my mom to get home from work, and pick up alex to go to my dinner party tonight at the emporer\'s buffet.  i wish all my church friends can be there tonight, but unfortunately, they can\'t.  the church retreat is the this weekend, and i wish that they had fun.  i can\'t decide waht to wear tonight, i have this really nice top that i bought on wednesday, but i have nothing to go with it.  oh, screw this.  '),(12901,3509141,'19,June,2004','2004-06-19 00:00:00','   Today i watched the Last Samurai for the 5 millionth time, and guess what? for the first time i was emotionally touched, two tears rolled down my cheeks during the final battle.  i think the first few times that i watched it, i didn\'t even feel anything, it was just like, \" oh, whole bunch people died... booo hooo\", that\'s probably becasue there were people around when i was watching it.  today, i just saw it by myself, and i actually cried.  when boys were young, when they fall, most parents say \"be a big boy, don\'t cry.\"  so these people grow up not knowing how to deal with these unpositive feelings.  who do these people grow up to be? when they have a bad day at work, or things just don\'t work out, how do they deal with these violent feelings?  they beat up their spouses and children.   It\'s okay to cry, as long as you know why.  '),(12902,3509141,'17,June,2004','2004-06-17 00:00:00','   I have a business exam tomorrow, so here i am to review some of the business issues.  Why is the interest rate so low?  in the 80s, the interest rate was around 18%, which means for every 100 hundred dollars that you put in the bank, you have 18 dollars by the end of the year, and the 18 dollars will continue to make money for you the next year.  so for those folks who decided to lock their money away at 18%, by now would have doubled their money 4 times. and for those who decide to take a mortgage at 18% and locked it there, they could have bought 2 more houses by the time they finish paying their mortgage.  the interest rate is created to balance the economy, when the economy is in depression, the interst rate is also low.  when the economy is in depression, people lose their jobs, and therefore stop spending unneccesary money and start saving, this cause other industries to lay off worker, and this eventually lead to few people working.  this is when the government lower the interest rate, so people would save all the money in the bank, spend it instead.  the interest has been the lowest in history for quite a while now, it is consistant at 0.25%, which means for every 100 dollar you have in the bank, you make a quarter by the end of year.  but the interest rate is about to go up, the canadian economy is now looking better.  if the interest rate is going up on savings, it is also going to go up on loans, so mortgage rate is also going to rise, if anyone is looking to get a mortgage, this is the best time to do it.  what are some better ways to save money rather than the bank? stock, they are riskier but have a better interest rate.  also bonds, they are not as riskier and they still have a good rate.  buy a house when you have the money, you can buy a house, and wait for the value of it to go up, normally you can sell it for 10% more than you bought it for.  did you know it is more expensive to live in Beijing than in Toronto?  a one bedroom apartment would cost 1.5 million yuans, even if you converted it, it is still very pricy.  so for those of you whose parents have some savings, i highly suggest that you buy a property in china, for those people who bought a house in year 2000, they can sell it for double the money.  especially with the olympic that\'s coming up in 2008, the real estate will double if not triple in the next 4 years.  '),(12903,3509141,'16,June,2004','2004-06-16 00:00:00','   i just finished my english exam today, it was easy and it was multiple choice.  i studies all the terms and all the definitions that we need to know, and how many were tested on the exams? None!  something i actually studied for wasn\'t on the exam, and waht kind of questions did we have? really basic ones.  we had 2 hours to do the exam, it had 25 questions and you pick the best answer out of four given choices.  i finished the exam after 40 minutes, sat there for like half an hour.  we all had to bring our dictionaries, and did we actually use them? no.    tomorrow would be my trip to Cayuga, it\'ll be 2 hours there and back.  fun fun fun.  and night, Rev. Mets and Mrs. Mets had invited me and lin to dinner, fun fun fun.  i have one more exam which i\'ll study for on Friday morning, and then write the exam in the afternoon, that way i\'ll remember more.  procrastination has it\'s own ways.  '),(12904,3509141,'14,June,2004','2004-06-14 00:00:00','   Cinderella!    urlLink   What movie Do you Belong in?(many different outcomes!)     brought to you by  urlLink Quizilla      You are a child\'s kiss.  Completely sweet and innocent and pure. You mean no harm and only love in your sweet kisses.    urlLink   What kind of kiss are you?     brought to you by  urlLink Quizilla      A GAME-BOY. Youre like a tomboy without the love of sports. Reality sucks, but as long as you have your electronics you feel you can cope. Time goes unnoticed when youre locked in your room hooked up to your Nintendo, rocking to your favourite collection of guitar-driven albums.  Your virtues: Intelligence, sense-of-humour, individuality.  Your flaws: Inability to cope with real life, action-freak spirit, reclusive nature.     urlLink   What kind of girl are you?     brought to you by  urlLink Quizilla      you are the \"you suck, and that\'s sad\" happy bunny. your truthful, but can be a bit brutal.     urlLink   which happy bunny are you?     brought to you by  urlLink Quizilla       My inner child is sixteen years old!    Life\'s not fair! It\'s never been fair, but while adults might just accept that, I know something\'s gotta change. And it\'s gonna change, just as soon as I become an adult and get some power of my own.    urlLink   How Old is Your Inner Child?     brought to you by  urlLink Quizilla      Your Heart is Red    urlLink   What Color is Your Heart?      brought to you by  urlLink Quizilla      Aphrodite/Eros    urlLink   ?? Which Of The Greek Gods Are You ??     brought to you by  urlLink Quizilla      You\'re GI Joe with the Kung Fu Grip!!  You\'re strong, tough, and know how to kick some ass. Don\'t forget though, no matter how manly you think you are, you\'re still just a doll. God Bless America.     urlLink   What childhood toy from the 80s are you?      brought to you by  urlLink Quizilla      dependent    urlLink   Which Personality Disorder Do You Have?     brought to you by  urlLink Quizilla      hand holding - you like to be in constant physical contact with your special someone but you don\'t want to take things too quickly.    urlLink   What Sign of Affection Are You?     brought to you by  urlLink Quizilla      You have a free soul! As all the souls go, yours is the most free-spirited and adventurous. You like camping, hiking, or interaction with other people. Your a social butterfly, but not because of your style, but because of your willingness to communicate with everyone. You probably have close friends who can rely on you because you always seem to know whats going on in the world. You love music and are free-spirited and someone fun to be around. A born leader and great explorer-dont ever change-the world needs more people like you.    urlLink   What Kind of SOUL do you posses? (For Girls only) Incredible Anime Pictures!     brought to you by  urlLink Quizilla      Mermaid: Mermaids, similar to Centaurs, have a torso of a human and the body of a fish. You are curious yet reserved in your actions.  You like to have fun but never at the expense of others and you never roughouse.  You love water and the creature in it and feel it is your job to make sure they stay safe.    urlLink   What Mythological Creature Are You (Many Results and Beautiful Pics)     brought to you by  urlLink Quizilla      You\'re Element is Light. You are friendly, happy, social, bubbly, and can brighten up any one\'s day. You are very kind and a real people person because you have several friends (or atleast should). You\'re cheery nature makes you lovable and your stunning looks are sweet and stand out.     urlLink   What\'s Your Element(girls)? (PICTURES)     brought to you by  urlLink Quizilla     urlLink     What Finding Nemo Character are You?     brought to you by  urlLink Quizilla      You are a Siren. More adventurous than all with a voice like no other you sit on warm rocks and sing to the moon and sea. Yet sometimes shipwrecks find you and raving men want you. You are a bottle of talent and power. What the unknown is you seek to find, and a lover. You have the moon and stars as freinds. There are a very few of you, what a rare find.    urlLink   What kind of mermaid are you? (Gorgeous Pics)     brought to you by  urlLink Quizilla      -Perfect- You\'re the perfect girlfriend. Which means you\'re rare or that you cheated :P You\'re the kind of chick that can hang out with your boyfriend\'s friends and be silly. You don\'t care about presents or about going to fancy placed. Hell, just hang out. You\'re just happy being around your boyfriend.    urlLink   What Kind of Girlfriend Are You?     brought to you by  urlLink Quizilla      Congrats! Your a Pure Angel! Angels, as far as most of them go, are all compatabile creatures, but Pure ones simply are symbols of God. Pure Angels always appear when a child is born, when a rainbow is seen, or when someone shares their first kiss. They never grow old, an can appear in the shape of a naked woman with white, bold wings. Pure angels are the carriers of god, and show their love to everyone in the world.    urlLink   What Kind of ANGEL are you? (For Girls only) This Quiz has amazingly Beautiful Pictures!     brought to you by  urlLink Quizilla      You\'re a lollipop!! You\'re known for your coolness, for you are a trend setter. You\'re a natural leader, and are good under pressure.  People often seek you out for advice, for you have great insight.     urlLink   Which kind of candy are you?     brought to you by  urlLink Quizilla      Fun. You Truly Desire Fun. Your always willing and ready to play a game, whatever it may be. But be careful, you cn tend to get a little too competitive.   PLEASE RATE     urlLink   What Do You Truly Desire? *PICS*     brought to you by  urlLink Quizilla      You belong in the world of balance, where everything has two sides and everything is not always what it appears on the outside. Of course, if you have a sense of humor, you may find employment as a sarcastic comedian. If not, enjoy the choices that are presented to you through life, they will always have two sides to them,one which leads you to the light, and the other which entrenches you in darkness. Walk on with hope, my friend.    urlLink   Where do you belong?(ANIME IMAGES)     brought to you by  urlLink Quizilla  '),(12905,3509141,'13,June,2004','2004-06-13 00:00:00','   yep yep, alex came over to my house today after the tutorials, which were completely a waste of time.  i was nervous.  thank god taht i cleaned the house yesterday after all the movies that i watched.  i think he actually liked my house.    he was here since like 10 30, and we watch a movie,  Roxanne  ( Steve Martin), it was just a blast.  and then i made spaghetti, which were delicious, and i think alex enjoyed my cooking.  if you call that cooking.  and then we just talked, my parents were home, they wanted to meet alex and see what he\'s like.  my parents like him, i know how strict chinses parents are about dating, but i think my mom\'s pretty cool about this whole thing.    tomorrow, i\'ll be over at alex\'s house to study for our english exam.  '),(12906,3509141,'13,June,2004','2004-06-13 00:00:00','   sunday morning, 10 o\'clock.  showered, still in pajamas.  feeling a little like a hang over, not really awake yet awake. home alone.  suppose to take the day to cram for exam, but feels too ... crammed already.  i think i\'ll go eat an ice cream bar... okay, now what am i gonna do.  i think i\'ll just watch a movie, something romantic and cheesey, i still have to finish  How to Deal , i think Mandy Moore can actually act, i just watched  Chasing Liberty , again.  i started  How to Deal  last night, but never really got into it, so i\'ll just give it another shot.   *time passed*  i didn\'t just watch one movie, i watch four,  How to Deal  ( Mandy Moore, Allison Janney),  Notting Hill  (Julia Roberts, Hugh Grant),  Two Weeks Notice  ( Sandra Bullock, Hugh Grant),  Kate & Leopold  ( Meg Ryan, Hugh Jackman).  and now it\'s almost dinner time, and all i had all day is ice cream, i declare today Romance Movie Day.  i think my mom\'s pissed at me for not eating, but it\'s not that i don\'t want to eat, but i\'m not really hungry.  i tidied my room, and ironed some clothes, and now i feel a little better.  exams are coming up, and i can\'t believe i\'m not nervous, well, i\'m usually not nervous, but i\'m not even concerned, i guess i\'m pretty confident, just hope that i\'m not over confident.  i think the weather is really good for the pass few days, it\'s not like we actually even notice the weather, all we do all day is studing or trying to study, we are all too busy to even notice the nature\'s wonder.  i check the weather every morning, for what? it\'s not like i actually stay outside long enough to care.  i think we should all take a break sometimes, like today.   just slow down our pace and look around, we might just appreciate all the things that we took for granted.  '),(12907,3509141,'10,June,2004','2004-06-10 00:00:00','   Hatred   Do you ever tink before you speak? Detroying every aspect of me Acting like you know who I really am Telling me who i should and shouldn\'t be  You are hypocritical, undeniable Sadistic mess up I want to let this go Leave me alone and shut up  You really screwed me over What did I ever do to you? you have messed you my lines I want to get over this, don\'t you?  You\'re a real work of art Too good to be true Ironically, you are I always thought there was nothing you couldn\'t do  Too bad you couldn\'t loveme Too bad you couldn\'t feel the same way Too bad you lead me on Then pushed me away  Now i\'m all alone Sitting here with my broken heart You stole it from me And ripped it apart  When I finally move on You come crawling back Telling me he\'s not worth it While he has everything you lack                         -- Gagan Bath  '),(12908,3509141,'09,June,2004','2004-06-09 00:00:00','   \"Dahling, if you luv me, would you please please smile?\"  that is a book title, it\'s a good book, it\'s short, double spaced, large printed, and it deals with a lot of things that i think teenagers should be aware of.  things like peer pressure, drugs, sex, family issues.  i\'m bored.  i haven\'t had homework for a while now, and i\'m not use to it.   i have been working a lot, and i\'m getting my paycheck soon, it\'s nice to make your own money and spend it in any way you want to. it gives you a sense of freedom.  i was just wondering how i\'m gonna spend this summer besides summer school, i think i\'ll just work more, i wouldn\'t mind more cash.  '),(12909,3509141,'08,June,2004','2004-06-08 00:00:00','   Let me tell you somethings i love.  i love bridges, they are so cool. standing on them or underneath them, and do whatever.  it\'s really calm and quiet and soothing.  i also love the woods, it\'s like you are attached to nature once again.  i love the free time i have to spend on the bridge and the woods, i love the people i\'m with.  i love writing on my blog and play math circus and listen to Punjabi music in the background.  i love the weather, i\'m hot, sweating, hey, i don\'t mind an early august.  i love the new harry potter movie, especially when you have read the book, then you can acctually understand the movie, but overall it\'s a good movie.  there\'s one more day of school left, and i just love it.  '),(12910,3509141,'07,June,2004','2004-06-07 00:00:00','   since i know you are gonna read this, so here it is. apologies for those who read the last comments, sorry for the rudeness for those who should die. you want to konw the reason(s) why i hate you... the list is sooo long i don\'t know where to start, but as a whole, i just hate you.  hey, i gonna hate someone, it to be you than anyone else. and if you think my blog is childish, do us both some good and don\'t come.  '),(12911,3509141,'04,June,2004','2004-06-04 00:00:00','   i just realized that a third of canandian girls are not happy with who they are, there seem to be a global standard of waht a woman should be, and if you are, then sucks to be you, becuase you just stupid and ugly.  half the adolences girls are not comfortable with the phrase \" i happy to be me\", and it seems like all the time, girls are told that they are fat, they are ugly, but the truth is it\'s all an act.  some poeple have told me that i\'m fat, but the truth is that she was just jealous and had other better to do.  and in the same situation, i have told someone that she was fat, and it was because i was jealous and bitter.  i know most people who read my blog are guys, and you guys don\'t usually realize the pressure of being a girl.  i\'m very good at ignoring what others say about me, cause i honestly don\'t care, after the b!tch xuan, i think i can take anything, bring it on.  i think i just want to say that looks aren\'t everything, i know people tell you this all the time, but it is not.  i have learned taht super hot looking guys aren\'t worth a penny when come down to what\'s important, and i think this goes out to the guys too, a blonde, big boobs girl may not be what you are looking for.  I\'m happy with who I am, atleast ... so far.    Are you?  '),(12912,3509141,'02,June,2004','2004-06-02 00:00:00','   today was as short.  well, maybe not.  alex wasn\'t here.  in english, we had the first part of our exam, although i finished on time, but the quality of my essay was definitely questionable.  that\'s when i realized taht alex wasn\'t here.    my last class was cancelled.  i came home at about 12 30, and called alex.  he forgot he had an exam  after lunch, me and stew went shopping at square 1, didn\'t buy anything since i didn\'t see anything i really wanted.  came home, and did this.  life is a bore, and so is this, and if you actually read this whole thing, i suggest that you do something useful with your time.  today, the reality of everyone leaving me finally hit me.  i just realized that high school is half way over, and the fun is half way over, and i think it\'s the fun that\'s keeping me in school.  manpreet isn\'t coming back next year, and there\'s many people who might move as well.  it\'s just that everyone is going there seprate ways, and i think i\'m just a little depressed by it.  '),(12913,3509141,'01,June,2004','2004-06-01 00:00:00','   I\'m acctually falling in love with school. all the fun stuff we get to do in all my class.  in business, we are playing jepordy, and in drama, we just practice for our exam, which is turning out to be a blast, in english, we continue to watch movies, now we are watch moulin rouge, and in career studies, we just sit there and chat.  i\'m really hally that there\'s only 8 days left of school, but i think i\'m acctually a little unsettled, i think i\'m gonna miss it.    i love school, everyone gets together, we get to meet different people and make differnet friends.    i know that this sounds really corny, but i find it fascinating that we get to learn something different each day, not just academic stuff, but now everyone gets along, how people react differently and why.    i think high school is one of the last places where in our lives that we get to see so many people with different purpose, interest, background, future co-existing in one school.  i think that high school is suppose to be the highest point of our life where we can have a lot of fun with little sense of resposiblities.    upon the arriving of our exams, i think everyone more or less will be under stress, i just want to point out the fact that you should still go out and have fun once in a while, clear your mind from studies acctually help you study better.  '),(12914,3509141,'13,July,2004','2004-07-13 00:00:00','   \"Your presence in my life brings wonderful smiles and loving thoughts within my heart.\"    i think i\'m in love, i think for the firs time in my life that i\'m in love with someone. i know i haven\'t been updating lately, or even reading anyone\'s blog.  that\'s because i have been so busy lately.  first of all, i\'m taking summer school, so that just takes all the times that i have during the week to do homework.  secondly, this saturday i\'ll be taking my driver\'s lisence test, and on sunday i tutor fandi on the clarinet.  my mom just came in with the photos taken from my birthday, ick.  '),(12915,3509141,'05,July,2004','2004-07-05 00:00:00','   A Love That Lasts Forever by Shane  I sit here today with thoughts of you and I,  the aching feeling inside that I just can\'t seem to hide. When two people have shared as much as you and I have, when they\'ve opened up their hearts, shared their dreams, thoughts, and fears.  We have come to know each other well enough to know if sadness is hiding behind a smile or if happiness is glowing from the others eyes. We have shared so many laughs,  and we know when each others pain at times has triggered some tears.  We have come to believe in one another and are always sincere to the other,  we have trusted one another with the truth that lies within our hearts. Our true love has yet to be seen with the eyes,  but it\'s felt with our hearts, we have gained trust, understanding, secrets, loyalty, and sharing ones deepest passions.  This love is rarely found in life, but when it\'s found it has a profound impact on one\'s well being, strength and character.  A true love does not need elaborate gifts or spectacular events in order to be valuable or valued.  True love needs what I have to give to you now. I want to give you my undying loyalty, unmatched understanding, unsurpassed trust, and the deepest of enless soulful sharing and love that I have for you.  I love you with all that I am and all that I will be for the rest of my life.....  http://library.lovingyou.com/poetry/201/201084.shtml  '),(12916,3509141,'31,May,2004','2004-05-31 00:00:00','   I\'m in a good mood today. Nothing too depressing happened today and I got no hw, except studying for 2 unit tests coming up. No complaints today :).  Awww, that\'s boring... nothing to talk about... Anyways, ... nothing to say...  Oh, I\'m so mellow that I can\'t get mad... Anyways, I might make this a member\'s only blog so that no one will be able to post comments anymore. I no there\'s a lot of good-intentioned people out there, but the recent comment by HiddenAssassin(deleted already) is completely unacceptable. It was complete, unconstructive, pure insulting, flaming. I will not tolerate anymore or this. Since I can\'t do much about it as long as this blog is public, it may not remain that way long. I do wish, however, that everyone continues to visit my blog, even if u cannot comment.  I\'m terribly sorry for the stupid actions of the lesser humans among us, who, I\'m sure will go to hell.  Thank you for ur attentiona and I understand if u never come to my blog again, but don\'t hate me. Hate HiddenAssassin(MSN me for his real name). He\'s ruined a good thing so take him up for it.  I can\'t be bothered with this too much. I can only handle this much insult and still be mellow, so I\'ll just go revel in my mellowieness. Cya.   '),(12917,3509141,'30,May,2004','2004-05-30 00:00:00','   Sigh ... I opened Pandora\'s Box and publicly advertised my blog using my MSN name... Now, I can\'t close it. I\'m glad that people now has a gateway into my mind but I also regret letting in the people who just don\'t understand. And I mean u, creature. Everything u\'ve said has been a judgement on the way I\'ve chosen to live and the way I\'ve chosen to write. Frankly, I don\'t care what u think. This blog is about me for a reason. I made it to be about me. Now, if u have a problem with that, stfu and go to hell.  As for M, thanx. That means alot to me... And I\'m sry about calling u smug... and cute... I\'m just paraphrasing what the chinese parents say... U no, dimples and all... I can imagine getting my ass kicked,  shudders . Anyways, I hope L and H reads it too.  Really though, I\'m sick of it. I\'ve heard it at school so I don\'t wanna hear it here. What I should do and what I shouldn\'t do, what I\'ve done wrong and what I\'ve done right. Get this into ur head: That is not what this blog is about. I\'m tired of getting judged. I made this blog to feel relief from my anxieties and to be comforted, not to be bombarded by judements made by people who have a problem with how I live. From now on, negative comments better be constructive criticism in my eyes, or I\'ll flame ur ass out of here. That\'s right, this is where I have power.  Now, I want to note something about that \"most important thing in life\" thing I wrote a couple of days ago. I just want to make sure that no one confuses that as synonymous towards the \"meaning of life\". The meaning of life is something for each and every one of us to discover for themselves. That journey to discover it is what we call life. Therefore, the most important thing in life is to enjoy that \"journey\" and not regret it afterwards. IOW, be happy.  Sometimes, I do wish that I can control my own happiness, which is impossible since I\'m not even sure what\'ll make me happy. All my fantasies have been about revenge and death is always involved in it. I\'ve never even tried to imagine myself happy. I wish I could just be more optimistist once in a while.  Anyways, I got tons of hw, so I\'m off, cya.   '),(12918,3509141,'29,May,2004','2004-05-29 00:00:00','   Again, I\'d like to prompt people to leave comments. They can be any type of comment. Whether u can relate to the post or not; whether if u agree to the post or not; anything. Be warned though, if u disagree with something I\'ve said and comment about it, expect a rebuttal from me. I\'m like that.  Anyways, I\'ve realized that my posts aren\'t very conventional. Usually, blogs take the form of a diary or journal, telling of the day\'s events and the person\'s reflections about them. In the case of my blog, however, I try to be as vague as possible on what happens to me. Instead, I focus on thoughts that derived from these events and build on them to create questions, hypotheses, theories, and philosophies. This is prolly because recapping the events of the day won\'t help me nearly as much as the analysing of my thoughts. That\'s just something I\'ve realized...  Anyways, just felt like noting this, but Poetry.com wants me to write a poem and biography for their Internation Who\'s Who in Poetry. Now, I never sighed up for this... I\'m even sure if I\'m obligated to do it. It just might be written in the fine print somewhere when I entered my previous poetry entries... What a bother...  Also, I passed up a chance to go to my church friend\'s house. This kinda surpriced me as it sorta contradicts the topic I was gonna talk about today. U c, seeing my church friends on the weekends r usually the highlight of my week since school is always such crap. I\'ve always taken comfort in these friends. Their very prescence has helped me though my period of depression during first semester. I\'ve molded this in this statement: I find strength from without, instead of from within. I don\'t even no if strength is accurate. I might be more like  stability . Just being in the same room as them, listening to their conversations(I\'m the quiet one), makes me super mellow. I forget all my worries. It\'s similar to playing video games and watching anime, both of which mellows me out temporarily, but much more effective. Without these friends of mine, I doubt video games and anime would be enough to keep me from attempting something irraversible... ie, suicide or a mad machine gun sweep of my school.  But me deciding not to go today... Why? I\'m not regreting that decision, but why did I make it? Is it because I\'m mellow enough from seeing them yesterday? Or too mellow from seeing them yesterday? Or perhaps I feel too tired or even too inadequate to face them? I think I\'m growing distant from them. I practically talk to them on MSN everyday about gaming or anime matters, but the serious talks... they don\'t come to me for those anymore. It\'s cause I\'ve been burying myself in computer games recently. They come to me, and I tell them to go away. If it was me, I\'d be discouraged from talking to someone if they told me to go away everytime I tried to start a conversation.  One of them said something about this the other day. He said, although I was the oldest(when one of them isn\'t present), I\'m perhaps the most immature and cared more about computer games than them. I no I\'m immature. I am that way. It\'s sad, but I can\'t bring myself to do anything about it. It saddens me to no how dependent I am on others. I can\'t even ride the bus by myself...  I do care about them though, as much as they care about me. Even though the love is always there in our little group of friends, the relationships I have with each of them is quite different. Let\'s call them M, L, and H. M is a smug cute little boy(only in the eyes of adults) who is three months younger than. L is a big gangsta-like guy who looks exactly like the kind of person who lives as troubled a life he does who is six months younger than me. H is a super-liberal girl who seems like Lori from Gilmore Girls, except with an invisible shroud of mystery around her emotions who is five moths older than me. Let\'s see, I talk to M daily, usually about anime and video games. Now however, I\'m begining to take him for granted and sometimes unintentially talk crually to him. L is being emgrossed in his personal troubles, and I\'m not privy to it because I\'m too busy playing computer games. H, the rest of us almost never c anymore; we\'re begining to doubt we no who she is anymore... The point is, I\'m growing distant from all of my church friends, the ones who I can\'t live without.  Will the day come when they become strangers? I can imagine that day and I no I\'ll be crying. I don\'t want that day to come. So, if ur reading this, M, L, and/or H, I need u, even if I say I don\'t, I really NEED u. Please, stay by me no matter what, and I\'ll try to do the same. God nos I\'ll hate myself if I don\'t...   '),(12919,3509141,'28,May,2004','2004-05-28 00:00:00','   Hey, everyone, just got back from church... Did pretty much nothing except eat slushies and talk to friends... Anyways, I just want to say that I really appreciate this blog thing cause it helps me write out and develop these theories about y I\'m unhappy and how I can be happy. My only regret is that I forget some of these theories before I can write them down.  Anyways, today, I wanna tell about my self-hatred state. I hate myself for the same reasons someone else might hate me. Character defects maybe, possibly for just being plain wierd. But the most dominant reason is that I hurt other people. In my search to find my happiness, I hurt countless others, however unintentionally. That can prolly be attributed to my stupidity. When something seems right or cool to do at a certain time, I do it. And it ends up hurting someone. This has happened everywhere from gym to bonding time I spend with my friends. The worst thing is how the people I care about the most r the ones who get hurt the most. Many of them still carry the physical or emotional scars today. I think about the actions I did, the things I said afterwards and realize how stupid I was. And it was all done because I wanted to be happy...  One of my friends said today, that happiness is the most important thing in life. I agree because without happiness, nothing else matters much. But this important question struck me, who\'s happiness is more important? Ur own or someone else\'s? If they\'re equally important, then what happens if u can only be happy at the expence of someone else\'s happiness? What should u do? And what if u took away someone else\'s happiness? Does that make u responsible for that person\'s happiness, in addition to ur own? Who\'s to make these decisions? Deciding who gets to be happy and who gets to be unhappy is like deciding who lives and who dies. No one is qualified to make that choise. I certainly am not.  So what now? Do I feel guilty about striving for my happiness while unknowningly stepping on others\' happiness? Do I continue to hate myself? I don\'t no. But then, I don\'t no much. Sometimes, like right now, I wish someone who had all the right answers would just tell me what to do. But the most knoledgeble is always the quiet one. In fact the all-knowing one is the most silent one. He nevers tells me anything. I\'ve asked Him every question I\'ve asked here in my blog and so far, I haven\'t got an answer.  I swear I feel tired whenever I write in this blog. I\'m gonna go now. Cya.   '),(12920,3509141,'27,May,2004','2004-05-27 00:00:00','   I\'m out of energy. I can barely keep my eyes open. Not that I really want to since there\'s nothing nice to c. Nothing nice to do, nothing nice to listen to, nothing... Well I have plenty of things to do, but they\'re not \"nice\". I can do research about Paul Bernardo, maybe find an ad to do my spoof ad assignment on. But... I don\'t wanna. As boring sitting here doing nothing is, there r worse things. Namely, hw. Sigh, I\'m too tired to say anything or think anything or remember anything now... I\'m gonna leave today like this... no interesting new topic, no mind-bogling afterthought, just a mental image of my very tired self: an Oriental kid, looks to be around 15, male, sitting with his ass sliding off a black ergonomic chair. His hair is all over his face and he is too tired to brush it away. His eyes r open only a slit, as if he\'s too tired to open them but too scared to close them. He\'s wearing a small black t-shirt with grey and white stripes along the sleaves. He\'s wearing black worn out Khakis with at least one hole in it, at the knee. His socked feet r under the desk in a rest position. His fingers and hands seem to be the only animated part of him, dashing back and forth across a dirty keyboard. In fact, from afar, the kid looks quite dead. He turns to look out the window and sees u... He screams... U scream... U run down the street, screaming in a frantic state. Because of ur current lack of focus, u run into a lamp post and hit ur head hard. U fall... And c the Oriental kid\'s face as u fall unconscious...   '),(12921,3509141,'26,May,2004','2004-05-26 00:00:00','   I would like to thank D.J.Max for his comment and yes, comments r more than welcomed. To c his comment, check under the \"So, tired...\" post on Wednesday, May 26, 2004. I would like to say, however, that this blog is as serious as it gets. What I write here is, in truth, for my personal benefit. It\'s like a reality TV series, where the prize is better self understanding.  Now, Max, u seem like a smart person and ur depiction of how someone like me may end up living is, by all means very plausible. I even respect ur decision in not living the way I\'ve chosen to live. Except, I\'ve never made this choice. And I can\'t change the way I am. At least, not without help. How can I choose the other road if I don\'t no how far back was the last fork? Unfortunately, ur comments recognized many of the things that r \"wrong\" in the way I think and live but does nothing to help me change.  As for whether I want to change, I don\'t no. The one dominant thing in me seems to be a transandant laziness. Not just the kind of laziness that makes me not do hw simply because I don\'t want to go get my backpack from downstairs. It\'s the kind of laziness that shapes my personality and attitude toward an easy lifestyle. However, this laziness of mine contracts my yearning for greatness. One side will eventually win out but I just can\'t be bothered to make a decision right now.  The desperate question is this: When I do choose a path, will it be too late to go down it? Time and the world won\'t wait for me as I let my desires decide for me my fate. I\'ll graduate high school and go to post-secondary. I\'ll grow up. The further u go down a road, the harder it is to turn back. If it so happens that by the time I arrive at a decision, it seems incredibly difficult to turn around, will I just go alone the same path as before?  All this make it seem that it\'s all so urgent. I\'m a procrastinator. I leave stuff off for tommorow. It\'s a bad habit but in a case like this, is it so wrong? I don\'t no what career path I\'m going to take. I don\'t no what kind of car or house or anything I want. These r life-long decisions. I\'ve been told this too many times by too many people. Now, is the decision on how to live my life the same? Or is it so important that it is made now that it will shape the rest of my life?  I simply don\'t care. I don\'t no what to do. I don\'t want to make the decision. I\'ll prolly do the same thing I do with everything else--Leave it for tommorow. Who nos when tommorow will actually come?   '),(12922,3509141,'25,May,2004','2004-05-25 00:00:00','   I was gonna post a follow-up to yesterday\'s blog at school... It didn\'t work out and I\'m too tired to do it now. The basic gist of it was gonna be how my parents don\'t understand me. And how we forget our fights fast but it stays at the back of our minds fueling our anger when we fight again.  Anyways, too much happens at school everyday for me to bother to much with what happens at home. Yup, it\'s almost all school. The way I feel everyday. This lack of a desire to really  live . School does terrible things to me. Maybe I do these things to myself, I don\'t no. Maybe I should settle for a 90% on a major assignment when my friends got 95%\'s and 100%\'s. Maybe I should settle for hitting the ball just once when I made countless errors. Maybe if I just settled, I wouldn\'t feel so bad. Ya, I blame it on school. I don\'t think I can last long if I blamed myself. I strive to be the best yet deep down, I don\'t believe I can achieve it. Maybe if I did, I would actually  try  try. But it doesn\'t work that way. I don\'t feel that much hate anymore. I simply can\'t. Now it\'s the sadness, despair, and simple shame. Y do I suk so much at everything? Y can\'t I be the best at  something ? Just one thing. That\'s all I ask for. I don\'t have to be a Jack of All Trades, just not a Jack of No Trades.  So... What am I? I can\'t maintain a 90 average. I\'ve tried but I no it\'s not possible. I can\'t do well enough in math to beat all the competition. I don\'t even have the guts to enter math contests. And gym. I hate my classmates. They fuel that fire that may one day consume me. I no it and they no it. That one day, I would just explode and either turn suicidal or homicidal. And even the pride I take in writing. The truth is simply that I\'m not that good. I\'ve done peer editing and there\'s way better formal writers in my very school. Even my fiction is weak. I have no bond with my characters. Whatever happens to them, I  feel  nothing. None of that forlorn feeling that u feel when ur favourite character in an anime series or RPG dies. I\'m no one special. In fact, with so little specialty that I\'m not even good enough to be normal.  Nowadays, I only have 3 states. Self-pity, self-hatred, and oblivious bliss. The last of which can be so simply attained through video games and anime that I\'ve gotten addicted to it. This is what one might call a routine of self-destruction. When anyone least suspect it, I\'ll prolly just dissappear. Maybe turn up a few days later down river or something.  Sure, I have my life ahead of me, but what\'s the point. I\'ve got very little satisfaction from my life so far. Who\'s to say that I\'ll get more from the rest of my life? It\'s times like this that I wonder if I should just kill myself.  Then more questions arise. What  if  I kill myself? Who will care? What will my only true friends from church think? What about my friendly classmates? My less than friendly classmates? My parents? Will they hate me for being stupid and killing myself? Will they blame themselves and live with the burden for the rest of their lives? I can\'t answer these questions. And I doubt I can ask them. After all, there\'s few of them would even consider the possibility of me killing myself as long as I\'m alive.  This is the way it\'ll stay. These thoughts will run through my mind again and again until the end. The end of my life or the end of my current state of mind. Who nos... Well, I\'m done ranting... Cya   '),(12923,3509141,'25,May,2004','2004-05-25 00:00:00','   I\'m back, and u guessed it, I\'m alive. After I went DDR downstairs, my dad came and just sat beside my dance pad. It annoyed me. I wanted to be alone. I guess feelings like that don\'t flow easily to parents. Anyways, I told him to go away, repeatedly. Then he got mad and got a broom stick to scare me. I new I couldn\'t fight him so I left for a walk in the cold. On my way out, he said that if I left, he\'d lock the door and not let me back in. I didn\'t even hesitate, I had a key in my pocket. I distanced myself from my house. As I walked, I thought and talked to God. I realized what scared me so much about my parents. I didn\'t no if their love for me was like God\'s, totally and absolutely unconditional, or if they would abandon me as soon as they got news that I failed to get into the university of their choise. The more I thought about it, the more I believed it was the latter. It was cold and since I was wearing a t-shirt, I turned to go home earlier than I would\'ve. Nevertheless, I dreaded going home. All I could imagine were my parents going on about how stupid it was to just walk out like that. On my way back, however, my mom found me. I didn\'t even c her since I was looking at the ground. She brought my sweatshirt and jacket. So... maybe she really loves me after all. I don\'t no. And there hasn\'t been a lecture, at least, not yet. I\'ve been avoiding my dad since I came in so I haven\'t talked to him yet. And I haven\'t talked to my mom either.  So, what now? What else is gonna happen? Will we resolve our problems and be one big happy family? What will happen tommorow? And ya, I was crying. I was crying when I walked around outside and I\'m crying now. I don\'t care about image. Think of me as a sissy who cries a lot if u want. I don\'t care. I\'m just so sick and tired about not noing if it\'ll get better or worse. I g2g again. Cya tommorow.   '),(12924,3509141,'25,May,2004','2004-05-25 00:00:00','   ARG! This is so friggin stupid. My parents r so annoying. My dad screams for no reason. My mom sits beside and and stares at me funny. I\'m gonna go crazy. Can\'t they just leave me alone?! I hate my life! ARGGGGGG! And y do I have to play ahead!? Do I care what I\'m gonna do the coming summer? I don\'t think I do! But they think that I should! I hate this, hate this, hate this! And u no what? This isn\'t as satisfying as I thought it\'d be. Maybe it\'s cause I\'m refraining from swearing too much... Still I think I\'m calm now... No suicidal or homicidal urges anymore... *deep breath... Yup, I\'m fine. Sry for this meaningless post about my stupid meaningless life! I\'m gonna take my agitation out on DDR, just so my mom would stop staring at what I\'m typing...  Anyhow, cya all... Have a nice life in case this is my last post and I kill myself soon...   '),(12925,3509141,'25,May,2004','2004-05-25 00:00:00','   heh, I\'ve done it. I\'ve created my own blog. I\'m now one of them. Or should I say one of  you ? Anyways, life\'s crap as usual. Well it\'s better than sometimes. And it\'s definately better than the lives of some people I no... Ai ya, I can hear my parents sing... *shivers Life\'s not fair... I ramble a lot don\'t I? I\'ve said so much yet I\'ve said nothing at all. Let\'s start over.  I\'m a 15-year-old Chinese Canadian with longish hair. Longish as in longer than any normal guy is ever supposed to have but shorter than that of the guys in Media Garden. I suk at pretty much everything. In fact, I\'m failing Gym class... And I\'m still in a state of shame after losing at air hockey to a girl... It\'s one of the delimmas in my life. I\'m not good enough at anything to be clearly defined as anything.  I do, however, have a passion for writing. It\'s sad cause nothing I write ever gets published. Except what I just wrote, and that, and that, and that... Right, so maybe my stuff do get published. Some of my poems got published at Poetry.com and then they started scamming me for money... Nevertheless, Criff is the main character of my current major project. Well, my first major project anyway. I\'ve got about 5k words into the story so far. Basically, it\'s this fantasy story involving magic and stuff. I\'m really too lazy to give a basic description... But if u want one or want to c what I got so far, don\'t be afraid to ask me through e-mail. Sigh, I\'m so bored that I\'ve resorted to this... I still can\'t belive it.  Well, I supposed I need some love right now. But I doubt I can get any without giving first. So here it is, I LOVE YOU! Now love me back or I\'ll hunt u down and make u give my love back. That\'s it for now. Cya.   '),(12926,3509141,'27,June,2004','2004-06-27 00:00:00','   Retreat was fun... Well, not as much fun as really really interesting and rewarding. I suppose it wasn\'t like last year (to which I didn\'t go, because no one told me about it), which I imagine was more fun than rewarding. Beside the point, anyways, I\'m exhausted, mentally, physically, and perhaps spiritually.  I\'m tired physically because I stayed up late listening to the guest pastor, Rev. Tony Costa and because I played badminton. I\'m mentally tired because I learned a lot, thought a lot, and had little time to sleep and recover. I\'m spiritually tired because..., well, who nos.  Tony Costa is a really interesting man. Like I no he\'s visited our church before but I was never there... Anyways, during the first Message, I was pretty much analysing him, like I do with all public speakers. He\'s confident and definately nos what he was talking about. By the end of the worship service, or third Message, I admired him not only as a speaker, a man of knowlege, and a man of faith, but a human being as well. He\'s a man who does, thinks, and believes what he does so fiersely, passionatly, and absolutely that I can hardly comprehend. He\'s the type of Christian that would\'ve been killed a long time ago if this was J.C.\'s time.  Anyways, I\'ve come to a personal revelation. I\'ve been aware of this for sometime but it hasn\'t been this clear in my mind before retreat. For awhile, I\'ve analyzed, from an objective point of view, Christianity. I\'ve come to the conclusion that Christianity is the best explained, if not only, logical belief. Similarly, I\'ve concluded that a Christian lifestyle, from a completely non-spiritual view, would be, at worst, neutral to my current life. My final conclusion: I\'ve nothing to lose.  I\'ve known this for a long time. But I never really put my real concern into a conherent thought before. I want to be genuinely Christian, but I can\'t. I\'ve prayed and prayed for the Holy Spirit to touch me but it hasn\'t. And God nos, only the Holy Spirit can bring me to truly accept Christ in my heart. It\'s like the devil has me captive and just won\'t let me let go... My solution for now is this: strive to live a Godly life and to keep praying. That and ask for others to pray for me too...  I\'m really a sinful person. I haven\'t really even tried to live up to God\'s standards. Maybe this is y I haven\'t recieved the Spirit. That will change. It has to.   '),(12927,3509141,'25,June,2004','2004-06-25 00:00:00','   I just came back from weeding and mowing the lawn... Housework\'s a million times harder when u own a house and a lawn that comes with it... My hands are dyed permanently green and my foot\'s bleeding from pushing a lawn mower while wearing sneakers with no socks...  Sigh... Sometimes I just wonder, if Chinese food is so hard to cook and taking care of the lawn is so tedious, y do we bother? Sure we\'re born Chinese but that doesn\'t mean we can\'t eat other food and I don\'t c anything wrong with having a plastic lawn... Oh well...  Anyways, even though it\'s summer and i don\'t have anything to do, the simple day-by-day life is actually pretty interesting. Like today, (if u don\'t like sick stuff, skip to last paragraph) my ass bled... I\'ve had this pimple on my ass for like two weeks and it\'s annoying when I sit on it. So anyways, I was scratching it and when I retracted my hand from my pants, there were like red stuff on it. I even tasted it to make sure it wasn\'t ketchup... Well, it sure isn\'t everyday that u can say \"My ass is bleeding.\" =D.  Now... I gotta go cook :\'(. Grr, I wonder if I\'ll survive the summer...   '),(12928,3509141,'22,June,2004','2004-06-22 00:00:00','   I\'m back after soooo long... Ya, after camping (friday and saturday) I didn\'t bother to update. Shows my laziness, huh.  Anyways, since last time, there\'s only one thing new and important. I\'ve taken up a new summer endeaver. I\'m gonna be a household slave for the ENTIRE summer :). It\'s like this: My mom wants me to do something during the summer, like something meaningful. Her list of choises were: Get a job, take swimming lessons, go to China, volunteer at library, volunteer at old-folk\'s home, and doing all household chores for the summer. Since I\'m sociophobic, I can only choose to stay home and be a slave. Btw, all household chores mean EVERYTHING, from cooking dinner to cleaning the washrooms... And believe me, it\'s not easy... Chinese food is crazy... The simplest ones take an hour; U\'ll c me running across the kitchen, battering the eggs, picking the beans, cutting the melon-like thingies, frying the leafy vegetables, boiling the soup... eek, just thinking about it haunts me.  But being a slave gives me some power as well... Being able to prepare food means I\'ll be in the position to sneak in some poison. And being able to clean means I\'ll be in the position to install hidden cameras... The slave will be the master.  Grins evilly .  Anyways, I just saw AFI\'s 100 years, 100 songs. It\'s this list of 100 songs from 100 years of american film, voted in by a bunch of important filmography people... Guess what the number one song was...  Somewhere over the rainbow Way up high, There\'s a land that I heard of Once in a lullaby.  Somewhere over the rainbow Skies are blue, And the dreams that you dare to dream Really do come true.  Someday I\'ll wish upon a star And wake up where the clouds are far Behind me. Where troubles melt like lemon drops Away above the chimney tops That\'s where you\'ll find me.  Somewhere over the rainbow Bluebirds fly. Birds fly over the rainbow. Why then, oh why can\'t I?  If happy little bluebirds fly Beyond the rainbow Why, oh why can\'t I?  I doubt many of us would recognize that. After all, it was in 1939\'s  Wizard of Oz , \"Somewhere Over the Rainbow\".  I spent 3 hours watching the show... Listening and watching songs and scenes from almost 100 movies... It\'s inspired me to want to visit/revisit some of these classics. Right now, I have a tramendous urge to go download the Sound of Music and Westside Story. Maybe I will... hmm...   '),(12929,3509141,'17,June,2004','2004-06-17 00:00:00','   Hey hey. All my exams r done :). Actually they were done on tuesday but I just didn\'t update my blog... Anyways, life\'s gooooood... and boooooring... Yuck. What I do all day and all night is play computer and video games... Boredom all the way. :(. And none of my friends want to play with me :(. It\'s so  sad  :(.  So, anyones. I have a  temperory solution  :). I made a new blog. It\'s a  role-playing blog . I dunno what exactly I\'m gonna do with it yet. But prolly it\'s gonna be where everyone makes up a role and plays it. I\'ll prolly make it loose role-play, i.e. peeps r allowed to change roles, small out-of-role moments allowed, etc. We\'ll c how that works out. Anyways, stay tuned :).  Btw, u can access the role-playing blog at http://rphaven.blogspot.com   '),(12930,3509141,'13,June,2004','2004-06-13 00:00:00','   I could swear i poseted yesterday... Oh well.... Anyways... Exams coming up... well, actually tommorow... And I havent\' crammed yet.... I\'m so screwed... Oh well... Sloth has dug its claws into me, and there\'s little I can do... I don\'t even feel like posting... Nothing to say, nothing to express, when I don\'t have school, I have no feelings. No negative feelings anyway. Computer games make me feel neutral and friends make me feel.... warm and fuzzy inside?! WTF is this?! Wait, nvm...  And this leads me to believe that this is what summer will likely to be... totally, completely, and absolutely meaningless and empty. So much so that I will likely come out of it in such a vulnerable state that my first semester depression period is likely to repeat... WAHH! I don\'t want that to happen...  Y didn\'t plan out my summer properly :(. I have no job, no summer school, no vulenteer program planned.... no computer camp, no army training camp,no CS LAN parties.... no internet cafÃ© package, no computer upgrades, no VB projects, or that of any kind... no new endeavers, no plans of any kind, and definately nothing interesting... I can c myself for the whole summer right now... In front of a computer screen, going to the washroom once in awhile to cry about sucking at online games so much, then giong to bed for a few hours nap, before getting up for the rest of the NIGHT to keep playing with people who live in China and are enjoying their DAY. Pathetic... I no.  But what can I do... I can\'t shake myself loose of Sloth... she is evil and she clings... her fingers around my throat, her nails dug into my heart... she clings to me life, she does. Gluttony\'s got me too... colourful food, tasty food, wonderful food... food aplenty, eat to me heart\'s content, beyond what me body needs or wants... If he had to, he\'d taken me from the stomach, or maybe the kidneys or the liver... I got my dealings with the others too... Pride\'s in my veins and I damn can\'t shake him off, afterall, I\'ll boast till I go hypovolemic. You can c Greed in my eyes. He darts here and there, telling me to take what I can. As for what I can\'t..., he leaves it to Envy, bet u\'d be able to c me complexion turn to a sinner\'s green... And Lust... she stays in my pants, no? I do hope she does, cause jail\'s an unpleasant place... That and that she clips her nails... If u no what I mean. Now, if someone tries to take me from me earthly pleasures and earthly sins, Wrath be upon \'em... Me Warth ain\'t something to recconed with. He be raw and he be vicious, and he plays dirty too...  Now... Do u have an idea what I have to deal with? I take upon my daily sins like another takes up his daily bread. I got not guilt, either. They say the seven deadly sins trips those who seek spirititual salvation and takes \'em up on the route of spiritual damnation... Like I give a damn. Humans cannot exist without earthly pleasures and earthly sins, and I certaingly cannot exist without them. But noooooo, society tells us to feel guilty. Let me tell u, let Guilt be the eighth deadly sin. He\'ll be among friends.  I refuse to conform, but does that mean I\'m rebelling? Really though, who the  orange juice  gives a  peanut butter and jelly sandwich ?!   '),(12931,3509141,'07,June,2004','2004-06-07 00:00:00','   ewww, there\'s stupid people visting my site, so i\'ll just post a post here.  first of all, mike, you can\'t let  your website die, you just can\'t.  there\'s nothing wrong with spending a lot of time on your site, since hey, it is your site.  is there really anything wrong with name calling for those who really deserve it? i mean, i could be creative about it too, let see... nah, i think i\'ll keep this site PG 13.  why do i have to give her a reason? i mean it\'s not like she gave me a reason? i don\'t owe her anything? do i? maybe i do,like a few more insults. but hey, why waste my time?  okay, i\'m done bitching, what\'s up with everyone? updates plz   '),(12932,3509141,'06,June,2004','2004-06-06 00:00:00','   I just went a whole weekend without blogging, soo.... Those people who say I spend too much time on this, in ur face!  Anyways, I\'m short on time so I\'m gonna keep this short.  Let\'s c... I\'ve gotten kinda bored of this blog thingie already... I guess it happens to the best of us, right? I\'m prolly gonna spend less and less energy on each consecutive post and eventually stop... I no it\'s sad, I\'ll definitely miss my blog. It\'s just that..., I have more important stuff to do... Like I\'d rather DO stuff instead of rambling on like a madman about things that really don\'t matter. Especially since posting REALLY takes up energy...  The way I look at it, the blog\'s gonna die sooner or later... I would like to keep it alive as long as possible though... However, I don\'t think I can do it by myself, so I\'d like to ask my three team members to occasionally post here in my blog.  That\'s it for today. Btw, commenting\'s back to registered users.... I don\'t think HiddenAssassin will ever bother to come to my blog again... so... Anyways, if u MUST no something about my weekend.... M, L, and I skipped Sunday school again... Oh, and I went to the Toronto Zoo on saturday. Did u no pythons lays eggs while boas give live births? Never new that...  Hmm, today\'s lesson is: If u can\'t be something, be anti-something. Note: Like in my longer posts, my lesson here is most likely to cause u to crash and burn if u heed it... so, don\'t, OK?   '),(12933,3509141,'03,June,2004','2004-06-03 00:00:00','   Normally I\'d post more but I\'m all \"thought\" out from commenting on my friends\' blogs... No more thought energy left...  Anyways, I got nothing to say at the moment anyway. It\'s the same old damned procrastination that\'s gonna one day kill me...  Ah, I\'m bored down to the level of M\'s blog... I would copy and paste my comments but I\'m too lazy to do that too...  So anyways, let\'s c how\'d I do when I have nothing in mind to talk about. Hmmm.... Well, I\'ve noticed that commenting others is a LOT easier than writting my own piece. It seems that the creation of a topic is something difficult and something to be considered a skill.  It\'s times like these, that u wonder about how creativity is useful in our lives. Certainly, we don\'t need it to stay alive. All we have to do is the do the same thing we do everyday like drones. But we do need it to  live . After all,  living  means going through our lives without acting like drones. It\'s creativity that introduced all the variance in our lives. Even the great minds that we admire would be nothing without creativity.  Then u come to this... Creativity is the source of all excitement in life, but in turn varience and excitement cause us pain. Sometimes, we think that living like drones wouldn\'t be as bad as living in pain. That is when we begin to doubt creativity. If I was never creative enough to up with the idea of putting a gun to a teacher\'s head in my comic assignment, I would\'ve never got in trouble. The same trouble that helped fuel my depression during first semester.  But then, u can argue human nature. Human nature is to survive and advance and it is willing to use creativity as a tool to do that. None of us r above human nature and we can\'t do anything about it.  I prolly shouldn\'t think that way... If I justified all my actions with human nature, I just might do something terrible... Cya and I\'m off to bed.   '),(12934,3509141,'02,June,2004','2004-06-02 00:00:00','   Um... I\'m alive... U might think I\'m dead cause I didn\'t post yesterday, but I\'m alive... I AM! Look, if I was dead, would I be posting this?!  Huh? U\'re not arguing? That\'s wierd... Anyways, my blog is officially Member\'s Only for comments. I might change it back if enough people message me or e-mail me about it on MSN...  I finished my math unit test yesterday and my civics unit test today :). Now, there won\'t be anymore math hw so I just have to review till the exam... Aww, I don\'t remember ANYTHING about analyctic geometry and trig... I got a lotta work ahead of me... I\'ll just study the last day :). As for civics... stupid summative project thingie and final in-class exam r still left... DAMN IT!  As for my other two classes... I\'m more or less failing \'em. Gym is one of those things I have no chance of even getting an 80 on... If u saw me doing beep test, u\'d prolly think I have asthma or something. Which I don\'t. And English...  I WAS getting 90\'s in English until recently... Well, yesterday... I got back 3 assigments... The marks were 65, 67, and 75... Sigh, I\'m a efing idiot... Along with my gym mark, that means I\'ll barely get a low 80\'s average... Which is BAD. No, not bad, just HORRIBLE. I\'ll just die! I need at least a mid 80\'s. Prefferably high 80\'s or low 90\'s. Wah! I felt like dying after english class...  Oh well. I swear, even though it\'s time to study like mad, I\'ve actually amped up my video gaming... I\'m just terrible, aren\'t I? I got myself into this situtation and I\'m too lazy to even try to get out of it...  Oh, and I realized soemthing. I dunno whether this means that I can\'t blame people for hurting me anymore but I\'ve realized that I bring the best and worst out of people. My meek, pathetic, unassertive personality with no trace of self-confidence seems to invoke specific feelings from those around me. They will automatically do one of two things unconsciously. Either A, Take absolute control of me and push me around. The people that get these feelings will make fun of and sometimes physically abuse me. They will force me do things to help them, which is in my nature to do. On the other hand, B, Take pity on me and help me as one would take on a project of taking care of a stray dog. The people that get these feelings are, although rare, absolutely crucial to my survival. They will seemingly help me selflessly but is likely to expact something back from me, which I am over-willing to give.  U c, my very existance influence the way people treat me... So, does that mean I can\'t blame them? Who nos... I gotta take a shower, cya.   '),(12935,3509141,'14,July,2004','2004-07-14 00:00:00','   Hey, I\'m posting again. Didn\'t think I\'d post so soon, did ya? Well, I\'m bored and working in Fairyland so... Anyways, since the last post, there\'s pretty much only one thing new.  I met a really interesting person on Fairyland. Her name\'s Shannon and she lives in Washinton State. I can really open up to her and I find her so interesting that I want to read her like a book. Does that sound strange? Anyways, if I really wanted to, I can go on and on telling u what I no about her. But, I won\'t. =D.  It\'s so strange sometimes, how a total stranger can become such an important person in ur life in an instant. I really enjoy talking to her. More so than perhaps even my closest friends(that doesn\'t mean I don\'t love u anymore... u no what I mean). Maybe it\'s because she\'s a total stranger. If I had to open up to someone who nos me, I\'d have to get them to unlearn what they already think they no about me first. But to Shannon, I can start whereever I want.  I never really went to those online chatrooms, but I wonder if I would\'ve had a similar experience if I did. I don\'t think so though. It\'s o so difficult to find someone who\'ll accept u as u r. I\'ve always struggled with that. That\'s y I thank God for Shannon.  Anyways, almost all my Cotton Yarns are Cloths now. So, cya next time. (Bows and exits)   '),(12936,3509141,'12,July,2004','2004-07-12 00:00:00','   Hey, everyone, I\'m back. I no I haven\'t posted for quite awhile and I\'m sorry about that... But hey, there haven\'t been comments for my posts for way longer than so y should I feel guilty?  In any case, all I\'ve been doing these days r housework, playing FL, SCII, DDRMAX, watching anime, etc. U no, usual boring summer stuff. Oh ya, I got Soul Calibur II and Unreal Tournament for PS2. Both are oldies by now, but who cares... =D  Yes, there is something interesting today. Yet another revelation. I was pretty upset for awhile... Anyways, I\'m just really tired of trying my best and not being good enough. I mean, I spend 45 minutes washing the dishes to the best of my ability only to be yelled at for making a mess. They don\'t even give me a chance to clean it up... Can\'t they c that I\'m doing the best I can? And that the only thing yelling will do is plant a seed in me that will sprout and make me a just-as-bad parent? Can\'t they c that?  It\'s just like gym class. Even the teacher said I tried my best and still gave me a 78. A seventy-eight... U no what that\'s like?! Admitting I tried my best and then giving me a 78 is like giving someone a Christmas bonus and then firing him. It\'s the cruelest thing possible. U no what they teach in the early years of school... It\'s better to aim higher and miss than to aim low and achieve. Well that is total bs. The real world isn\'t a comforting place that u can survive in just by trying ur best. It\'s harsh and cruel. There\'s no conselation prize for trying ur best and failing, just the harsh reality that constantly remind u that u failed. Well, screw this world, I hate it. There\'s no sin in hating the world, since it\'s so evil, right?  That felt good. =D. I miss blogging... Anyways, cya soon. And for my sake, someone please post a comment...   '),(12937,3551359,'02,August,2004','2004-08-02 00:00:00','   For the past months I\'ve been participating in&nbsp;a masterclass about film criticism.&nbsp;The final college of this class took place yesterday, and afterwards&nbsp;I&nbsp;had&nbsp;a nice get-together with some fellow students. We&nbsp;had dinner together and&nbsp;some of us went to a couple of bars in Grooningen.&nbsp;I&nbsp;was one of the last to leave&nbsp;the bar at 05.00 \'o clock, so I got home&nbsp;pretty late.'),(12938,3551359,'02,August,2004','2004-08-02 00:00:00','   While searching for a birthday present for a friend of mine, I accidentally discovered that a dvd of  urlLink  Citizen Kane   (1941) has finally been released. My first screening of this movie took place fairly recently, during a college viewing in a movie theatre. I was immediately hooked by it\'s gloomy opening sequence. For some reason I am always a sucker for movies with a sinister atmosphere. Visually, \'Kane\' is very stunning. It made a huge impression on me, so I wanted to buy it for my collection. Being a poor student, I didn\'t really have the money to purchase this classic. Of course I couldn\'t resist the temptation and so I delved into my savings to buy a copy, along with the last installment of the  Lord of the Rings  trilogy, which was also on my wishlist. This week I\'ll be watching \'Kane\' for the second time. Hurray!'),(12939,3551359,'02,August,2004','2004-08-02 00:00:00','   Here\'s the thing, I am a movie lover, and yes, that means that I\'m a little obsessed with cinema. I have a habit of spending far too much of my money on dvd\'s. I even have this weird dream of directing my own movie one day. Besides this, I\'m also a journalist. I studied&nbsp;for becoming one&nbsp;for five years.&nbsp;During this time, I worked parttime and freelance for&nbsp;a number of newspapers. These&nbsp;jobs&nbsp;were always great fun, but after I while writing about local farmer festivals and missing cows somehow wasn\'t wholly satisfying for me anymore. I would like to spend more time on research before I write my stories, and that\'s why I chose to study some more.&nbsp;I always liked movies, and so I decided to study Arts, Culture &amp; Media in Groningen. I focus on&nbsp;film sciences and I&nbsp;hope this will give me anough basic knowledge&nbsp;and expertise to become a movie journalist or maybe a film critic one day.&nbsp;In the mean time&nbsp;I\'ve got&nbsp;to keep writing, so on this neat little website I\'ll try to keep a record of all my movie experiences. Oh, and I will probably also rave about my life, now and then.&nbsp;&nbsp;Let\'s see how long I can keep this up...'),(12940,3551359,'02,August,2004','2004-08-02 00:00:00','   I got tired of studying for all those tests I have next week,&nbsp;so I decided to take a pause to watch   urlLink Phone Booth   (2001). This movie couldn\'t have been released much later, because the cellular phone will probably have made phone boots obsolete in a year or two.  &nbsp;'),(12941,3551359,'02,August,2004','2004-08-02 00:00:00','   Well, this is it. As of today, the shiny world of television is a no go zone for me. Why? Because the European soccer championship has started. For one month, every TV channel will broadcast matches, hours of boring soccer related television shows and endless bickering by semi knowledgeable soccer analysts. This is not something I want&nbsp;to see.'),(12942,3551359,'02,August,2004','2004-08-02 00:00:00','   One more month of gruesome tests and studying and this summer\'s vacation is a fact. Finally! Except for a short trip to the Greek island of Crete, I haven\'t really been on vacation for years. No wonder I\'m dying to travel abroad. A couple of weeks ago, I got this crazy idea of flying to New York. I\'ve seen that city in hundreds of movies, and for quite some time I\'ve been curious whether it lives up to it\'s reputation of exciting metropolis. Tickets never were as cheap as they are now, but because of my cash-flow problems I\'m still forced to pick a travel destination closer to Holland, which means I\'m going to stay in Europe.'),(12943,3551359,'02,August,2004','2004-08-02 00:00:00','   I just took an interesting  urlLink questionnaire  to determine which movie bad guy I resemble most. According to the test results, my personality seems to be much like that of that \'nice\' guy in  urlLink  American Psycho   (2000), Norman Bates... Hmmm...'),(12944,3551359,'02,August,2004','2004-08-02 00:00:00','   Today\'s only been the second warm day of this summer, and already those high temperatures start to piss me off. I\'ve got some important tests coming up, so I went to the university library to do some studying today. The heat in there made me feel so heavy and slow... I just couldn\'t concentrate on my books. I went back home early. After dinner (Thank god it was my roommate\'s turn to cook, I probably would not have survived the heat of a furnace) I tried watching David Lynch\'s   urlLink Wild at heart   (1990). Unfortunately, I totally forgot about my LCD television. Those tv\'s are nice and all, but they radiate so much heat that an average room feels like an oven in no time. After only fifteen minutes of  Wild at Heart  I was soaked by my own sweat and I started to melt away. Great... Later, I finished watching the movie, but I can\'t say that I really liked it. I don\'t know why, because I liked some of Lynch\'s other productions.  This reminds me of another miserable movie experience. For some reason, I went to the cinema to see   urlLink Pearl Harbor   (2001). What\'s even stranger is that&nbsp;I went to see it&nbsp;on a particulary hot day, a couple of summers ago. During the movie, wiping the sweat out of my eyes consumed much of my attention, so&nbsp;afterwards I could not&nbsp;really tell whether the movie was&nbsp;good or bad. Honestly, I thought I wouldn\'t leave that movie theatre alive. But despite the immense heat I couldn\'t get myself to leave&nbsp;before the movie ended (hey, I paid for that ticket!), so I sat out the entire three and a half hours. Afterwards I had only one word to say to the owner of that cinema: \'airco\'. I guess the summer just isn\'t the right season for watching movies...'),(12945,3551359,'02,August,2004','2004-08-02 00:00:00','   This afternoon I borrowed   urlLink Lola Rennt   (1998) from a&nbsp;friend, who taped it from television. I never saw an eclectic movie like this one before. The flashing images are accompanied by ear pounding techno music and German director Tom Tykwer uses virtually every film&nbsp;technique available. Black and white, color, standard 35 mm film, video, slow motion, single frame shots, jump cutting, animated sequences... You name it, it\'s all in this stylish movie.   Basically,  Lola Rennt  is about a punky girl (Lola, duh...) who runs her ass off for her lover, Manni. This dimwitted guy lost a large sum of drugs money on the subway. Money he was supposed to deliver to a very nasty gangster. Needless to say that Manni nearly craps his pants in fear of the consequences. Lola\'s got a measly 20 minutes to find another 100.000 Deutsche Marken, and so she starts running the streets of Berlin to save her boyfriend from a painful death. The story revolves completely around these frantic 20 minutes, that are shown three times. Each time the events take place in a slightly different manner. These tiny differences have a dramatic impact on the results of Lola\'s actions, however.   The concept of  Lola Rennt  appears to resemble that of movies like   urlLink Sliding doors   (1998) and   urlLink Groundhog day   (1993). These are also flicks that use alternate realities to show how tiny details can change one\'s faith. I was entertained by  Lola Rennt , but I fastforwarded some sequences because I felt they were a little repetitive.'),(12946,3551359,'02,August,2004','2004-08-02 00:00:00','   urlLink Ronald Reagan , former president of the United States of America, died yesterday at the age of 93. I really don\'t remember much of his presidency because I was just an ignorant eight year-old when Reagan was in his second term. However, I do know that in the 50\'s and 60\'s he had roles in a large number of television shows and mediocre Warner Bros movies, mostly westerns. Besides this, he was also president of the American Screen Actors Guild. Strange that a B-actor can become leader of the most powerful nation in the world. I can\'t imagine any Dutch actor being elected as prime minister. But then again, Holland isn\'t America...'),(12947,3551359,'02,August,2004','2004-08-02 00:00:00','   My mother felt generous today, and because my parents will be on vacation when I turn 23, she invited me and my father to a premature celebration of my birthday. We went to a Japanese restaurant. There we got to pick the meat, fish and vegetables of our choice and with the aid of a blazing fire the cooks prepared a meal out of those ingredients. Tasty! A bit too tasty, actually...'),(12948,3551359,'02,August,2004','2004-08-02 00:00:00','   And what an... Erm... Sick week this has been! First of all, I\'ve constantly been in an extremely nervous and nauseous state for reasons that I\'m not going to discuss on this blog. Let\'s say that I tried to give fate a twist and I\'m not sure yet if my actions will give me the results I hope for.  Partly because of the resulting uncertainty, I completely screwed up a test on Monday.&nbsp;Later that day I managed to poison myself and a friend of mine. We agreed she would drop by after work, and so I promised to prepare her a nice meal. The problem is that I\'m not much of a cook. I\'m the kind of guy that ruins even the simplest of dishes. This is probably the reason why my friend felt sick within an hour after we had finished supper. Her sickness must have been caused by the spicy chicken I made... (damn those foul, treacherous beasts!) After she\'d left, my own stomach also started to feel weird (and that\'s an understatement). Because of this, I didn\'t get much sleep that night.   On Tuesday morning I felt so miserable that I couldn\'t get myself to go to the university for the second test of this semester. Great... Well, it was no big deal since I hadn\'t studied for it. I was sick for the remainder of the day and I had a terrible headache. In the evening I felt a little better, so I watched a soccer match (Holland vs Germany) at a friend\'s place. Normally, I am not really a soccer fan (and again, an understatement) but this time I actually enjoyed cursing and swearing (damn those Nazi\'s!) in front of the television. I never realized soccer could be that much fun!   After the match I went home to get some sleep. I&nbsp;had just switched of the lights when I&nbsp;was called by some other friends who asked me to join them for a drink or two (or more). Despite my uneasy stomach and my tired head, I could not resist their invitation. For some reason we ended up in a very unsettling bar, full of freshmen, fratboys and 16-year old girls. It even had a bar which rotated slowly. Usually, joints like these are not my favourite places to hang out, but hey, who cares after a couple of beers? I enjoyed myself, but I forced myself to go home at 02.30. Besides, my roommate didn\'t feel very well either, so I had to escort him home. Enough is enough...   Because of my enthusiastic alcohol consumption I got sick again on Wednesday... No wonder I could not get any studying done this week. I guess I will have to retry all those important exams in August. Yeah, I know I am going to regret this later, but at least I had a great time!'),(12949,3551359,'02,August,2004','2004-08-02 00:00:00','   I saw   urlLink Troy   (2004) last night. I enjoyed it for what it is, but I can\'t say that I was really impressed by this movie. Sure, the CGI armies are nicely done, the action is enjoyable and yeah, the art direction is very good, but  Troy  has many flaws, of which I will describe the ones that annoyed me the most.   What disappointed me most were the massive battle scenes. A couple of years ago, battles like the ones in  Troy  would have been received enthusiastically. But after  The Lord of the Rings  trilogy, it\'s hard to impress people with the image of thousands of marching soldiers alone. The direction and choreography have to be much better than that of director Wolfgang Petersen. Obviously, it was his goal to show us the grandness of the Trojan war, and so every battle starts with long shots of the opposing parties. But when the clash finally begins, Petersen cuts to close ups of individual soldiers too often. Personally, I think this is a bad choice. Those close ups may give a chaotic sense to the spectacle, but together with the rapid editing, they made me forget about the massive scope of the battles entirely. There is only one fight that I really liked. The clash between Hector and Achilles is the best duel I have ever seen in a movie. Here Petersen doesn\'t use close ups and fast cutting, so you can see perfectly how the fight between the two heroes advances. Impressive choreography!   The casting is far from perfect. In  Troy  the role of Helen is played by German Dian Kruger. In Homer\'s story, Helen\'s supposed to be the most beautiful woman in the world. Well, Dian Kruger is definitely not the most beautiful female on this planet. Of course, she has a good looking face and a perfectly shaped body, but to be truely beautyful, she would have to have charisma and personality. Kruger lacks both. I found her performance rather bland. Brad Pitt isn\'t convincing as Achilles either. He trained six months to get the body of a Greek god, and I admit he succeeded in achieving this, but to me he doesn\'t convey the image of a battle hardened soldier. He\'s too much of a pretty boy for that, I guess. Perhaps it\'s also his character that works against him. Achilles is portrayed in a very one dimensional way. All he craves is immortality. Why? I didn\'t understand his motivation. During the movie, Pitt\'s character seems to go through some kind of development, but in the end he still seemed like a whining child to me. Some of the actors were perfectly cast however. I particularly liked Eric Bana in the role of Hector, the honorable Trojan heir to the throne. Other than Pitt, Bana has the perfect look for a movie like this one. Maybe this was also because of his character. Hector was far less one dimensional than those of other characters. His motives are understandable, and I felt some sympathy for him. The same goes for Peter O\'Toole, as King Priam. After Achilles has slained Hector, Priam visits Achilles to ask for the body of his dead son. This scene contains the only memorable dialogue of the movie. In  Troy , O\'Toole and Bana are the only actors who perform better than average.   There is one thing that always annoys me in movies like  Troy . Homer wrote his story about Troy thousands of years ago, so the screenwriters translated it to meet the needs of a present day audience. Because of this, the dialogues and characters in this Hollywood adaptation have a very modern feel that doesn\'t fit the story. Besides this, important elements that were in the original story were altered, modernized or removed completely. For instance, Petersen decided to scrap the gods from the script, because he felt this element didn\'t contribute to the story. Sure, the gods don\'t play a large role in our modern society, so why not remove them from the story entirely? Also, the homosexual relation of Achilles with his cousin Patrokles was transformed into a platonic friendship, because homosexuals are still unaccepted in our society. The relationship was altered because the target audience might feel a bit uneasy about two men making out on the big screen. I know that Homer\'s story isn\'t an exact historical account, it\'s more of a myth. So I can\'t blame Petersen for making a historically incorrect movie. But I don\'t like the fact that Hollywood always twists the facts to make for digestible stories. Historical movies loose authenticity because of this. Is Hollywood afraid that people are unable to imagine what it was like to live in another time, with other customs and beliefs? The more moviemakers make history look like a duplicate of our time, the more it saddens me that people accept the portrayal of history in Hollywood movies as a true account of what happened.'),(12950,3551359,'02,August,2004','2004-08-02 00:00:00','   This morning I was worrying about how I could get the money to pay for my final year at the university. Just when I concluded that I couldnâ€™t afford studying any longer, the phone rang. It was my former employer, offering me a temporary job. From October until January, I can work for the newspapers I worked for before I went to&nbsp;study again. This period is just long enough for me to earn the money that I need to pay for the expenses of my education.'),(12951,3551359,'02,August,2004','2004-08-02 00:00:00','   A decision about my vacation has finally been made! Me and my traveling companion will make a trip to the French alps. The lake of Annecy is our destination. We both had our doubts about this location, because there\'s a good chance the place will be crawling with Dutch tourists. And if there\'s one thing we don\'t want to encounter while on vacation, it\'s our countrymen.'),(12952,3551359,'02,August,2004','2004-08-02 00:00:00','   I read that a teenager has been arrested for recording   urlLink Spiderman 2   (2004) on his camcorder while watching this movie in a cinema. This arrest was possible because of a new law that prohibits the film audience to take recording devices with them into movie theatres. Offenders can be sentenced to a year in prison or they get a 2000 dollar fine for breaking the law. The kid I mentioned was busted by the operator of an LA based cinema. Apparantly, this employee used night vision goggles to check if the audience complied with the rules. This guy was rewarded with 500 bucks for arresting that kid. Distributors hand out these rewards because they hope that this wil convince more cinema employees to aid them in the struggle against bootlegging.  Last week I went to see   urlLink Troy   (2004) in the local theatre, and indeed, before the movie started, the audience was warned not to make any copies. I could not help laughing about this message. I found it so ridiculous... I simply could not imagine why someone would want to copy a movie directly from the silver screen with a camcorder, because the image quality of such bootlegs usually is extremely bad. Are blurred, badly lit copies really worth all the trouble? Is it that hard to wait until a decent version of the movie is released on dvd? Apparantly it is.  It\'s also ridiculous that distributors are putting so much effort in stopping this form of piracy. I mean, Spiderman 2 was released yesterday, and already there are copies circulating on the internet. There is no way to put a halt to it, so the only ones who fare well by the distributor\'s struggle are the cinema employees. If they really receive 500 dollars every time they catch a teenager with a video camera, they can make some nice money. We might be witnessing the birth of a new profession here, that of the cinema bountyhunter.  '),(12953,3551359,'02,August,2004','2004-08-02 00:00:00','   Yesterday the infamous actor Marlon Brando died at the age of 80. I feel ashamed about saying this, but except for   urlLink Apocalypse Now   (1979) and   urlLink The Godfather   (1972) I haven\'t seen many movies in which he has a role. I saw   urlLink The Score   (2001) fairly recently. This one didn\'t impress me at all, and neither did Brando\'s acting. Despite this, his other movies are still on my \'to see\' list...  '),(12954,3551359,'02,August,2004','2004-08-02 00:00:00','   My performance as a student has just been evaluated. The advisor told me that it\'s possible for me to finish the bachelor phase of my study in the first semester, next year. There\'s one \'but\' however... I will have to work like an ox to pull this off, because I\'ve got so many exams to retry. Besides this, I am also going to work two days per week from October until February. Ack! I think I\'m going to need a personal coach for motivation and stimulation when the next college year starts...'),(12955,3551359,'02,August,2004','2004-08-02 00:00:00','   Today I have reached the respectable age of 23. Damn, I\'m getting old and I don\'t like it. But I threw a nice party yesterday and that compensates a lot. A while ago, I promised my friends to organize a barbecue on my birthday (we pyromaniacs love to play around with fire), and since the weather was quite reasonable yesterday, all systems were go.'),(12956,3551359,'02,August,2004','2004-08-02 00:00:00','   I just discovered that I had some succes with one of the tests I had to do last month. I scored 6 out of 10 for the exam of an introductory film class. I suppose this could be interpreted as a very meagre result, but the thing is: I did not study at all for this test... I haven\'t even touched the darn book, so I think it\'s a pretty nice result after all. This is a perfect example of \'minimum input, maximum output\'...'),(12957,3551359,'02,August,2004','2004-08-02 00:00:00','   In the glimmering world of the media, the summer is usually a dull period in which nothing really happens. Everybody\'s absent or on vacation, and the people that do work are lazy and uninspired because of the warm weather. There is a total lack of news, so many journalists spend half of their time eating from their noses. Don\'t ask me why, but the Dutch refer to these sleepy weeks as \'cucumber time\'. News items that normally are considered unworthy of publishing suddenly become headlines in this period. After all, those editorial collumns have got to be filled, no matter what. You can\'t say \'Sorry, no news today\' to readers.  Some journalists complain about the chronical news shortage, but I love cucumber time! It\'s the perfect opportunity to cover less important stories for which newspapers normally don\'t have any time or space. Today I received a phone call from an acquaintance of mine. He works for a regional newspaper. He told me that they were in search of free lancers for a daily serial about ordinary people and their jobs (the typical type of story for cucumber time) and he asked if I had any interest in this job. I really enjoy writing serials like these, so of course I am interested! I\'m going to find out more about the job tomorrow. Finally, some interesting work again!  '),(12958,3551359,'02,August,2004','2004-08-02 00:00:00','   Boredom got a hold on me this afternoon. My friends are not in town at the moment, so I\'ve got to keep myself busy. I decided to get some books at the public library. I haven\'t been reading much lately because I couldn\'t focus myself enough. I had other things on my mind. The autobiography of actor Marlon Brando and the transcripts of the interviews of French filmmaker Francois Truffaut with Alfred Hitchcock should keep me interested long enough, though...  '),(12959,3551359,'02,August,2004','2004-08-02 00:00:00','   I\'m not exactly a sporty guy. Except for my brief membership of a local gym, I haven\'t participated in any sport since high school. Some time ago I got tired of panting like a dog every time I walked up a staircase, so I felt it was about time to beef up the condition of my body. I wanted to feel fit again. Teamsports are not my thing, but I like walking. That\'s why I impulsively bought a pair of running shoes two years ago. For half a year I tried to train myself, but it didn\'t work. When it comes to sports, I have a total lack of self discipline. My condition did not improve and I quickly lost enthusiasm. After six months I got busy with my job and my graduation assignment. The resulting stress was a perfect excuse to leave those shoes where they were: in the closet. When things got calmer I decided that I wanted to give running another try. But I could not push myself to really do so. Yesterday I could finally motivate myself, so I strapped on my shoes and ran for an hour or so. Now I can barely stand up straight because of my aching legs and back... Ouch! Myalgia sucks! I\'m not going to give up this time, though (hey, that sounds familiar). My travel companion likes mountainbiking, so I\'ve got to get into shape before I travel to France. Otherwise I will not be able to keep up with him...  '),(12960,3551359,'02,August,2004','2004-08-02 00:00:00','   According to the annual report of the Dutch association of coffee producers, the Dutch are becoming more and more economical when it comes to drinking coffee. On average, they drink 3,1 cups per day. My consumption is way above average... I\'ve only been awake for an hour or so, and already I had four large cups. Usually this number increases dramatically. Sometimes I drink more than ten cups per day. This is a habit I developed during work. The coffee machine made me an addict! If I don\'t drink enough coffee, all kinds of withdrawal symptoms occur. For example, I get headaches and I don\'t feel awake. In the morning, I really need caffeine to get myself going. I  urlLink read  about some of the effects this stuff can have when used without measure.   -restlessness  -nervousness  -excitement  -insomnia  -flushed face  -diuresis -gastrointestinal disturbance -muscle twitching  -rambling flow of thought and speech  -tachycardia or cardiac arrhythmia  -periods of inexhaustibility  -psychomotor agitation    It seems people can actually die from a coffee overdose... Ulp... Perhaps it\'s time to cut back on my coffee consumption :)  '),(12961,3551359,'02,August,2004','2004-08-02 00:00:00','   All my stuff is packed, the car is ready and so is my travel companion. Tomorrow at 06.30 we leave for France. I\'ll be back in ten days. In the mean time, I\'ll probably be&nbsp;able to update my blog by using my cell phone... Happy holidays!  '),(12962,3551359,'02,August,2004','2004-08-02 00:00:00','   This evening I&nbsp;was&nbsp;driving from Hoogeveen to my appartment in&nbsp;Groningen to fetch some forgotten shoes that I need for my vacation. I clearly wasn\'t paying enough attention, because&nbsp;when I entered the city I was flashed by one of those photographing speed-o-meters...&nbsp;Damn it!&nbsp;I hate those things! On the highway&nbsp;I&nbsp;drove with speeds up to 140 kmp/h (120 is the speed limit on Dutch highways). Lucky enough they caught me when I was driving&nbsp;56 on&nbsp;a road with a speed limit of 50 kmp/h. Otherwise I probably would have had to pay&nbsp;a lot more than than the 30 euro\'s I have hand over to the police now... Normally I\'m not really a speed freak (well duh, I&nbsp;rarely get the chance to drive a car anyway...) so this will&nbsp;actually be&nbsp;the first time I will get a fine for speeding. I think I will hang&nbsp;it above my bed :)'),(12963,3551359,'02,August,2004','2004-08-02 00:00:00','   Yesterday I had to visit Hoogeveen\'s shopping center, because I needed some new clothes and some stuff for my vacation. When I arrived in the main street of town, I was horrified to discover that it was  Pulledag . If I hadn\'t forgotten about this day, I probably would have stayed home. There is no English word that can describe this annual terror, but I\'ll try to explain...  &nbsp;  The summer is usually a&nbsp;worthless&nbsp;period for&nbsp;the shop keepers of Hoogeveen because&nbsp;most customers&nbsp;are on vacation.&nbsp;During the summer, shop&nbsp;owners&nbsp;have to make a living too of course, so they&nbsp;do everything&nbsp;to&nbsp;lure tourists and shoppers to their stores. They organize all kinds of activities to attract people to the shopping center. The result is  Pulledag , which takes place every Thursday of July and August. On these Thursdays, people can sell their old stuff at a cheap goods market, tradesmen sell&nbsp;a variety of&nbsp;junk in&nbsp;their stands,&nbsp;bands and&nbsp;folk dance&nbsp;groups&nbsp;give performances every hour&nbsp;and irritating clowns do their best to amuse&nbsp;visitors.&nbsp;As a kid I used to love Pulledag because there is so much to do.  &nbsp;  Nowadays I try to avoid main street when it\'s  Pulledag . I don\'t like it anymore. Every year it\'s the same. And besides, I don\'t like the people that are attracted to Hoogeveen by this event. Usually I refer to  Pulledag  as \'bumpkin day\', because it seems like&nbsp;the entire&nbsp;yokel community&nbsp;from Hoogeveen and the surrounding villages&nbsp;gathers&nbsp;in the center of town to get loaded with alcohol and fast food.&nbsp;Yesterday one of them loudly&nbsp;belched in my ear when I passed by. A bit further a&nbsp;bunch of other&nbsp;morons had an alcohol induced quarrel. They yelled at each other as if the other people weren\'t there. They even started throwing beer cans around.&nbsp;Really, the average IQ drops like a brick on  Pulledag ...'),(12964,3551359,'02,August,2004','2004-08-02 00:00:00','   Now here\'s a documentary I\'d love to see! In   urlLink Super Size Me   America\'s food culture is discussed by filmmaker Morgan Spurlock. He got the idea for this film after he had heard about a couple of obese girls. These oversized chicks took fast food chain McDonald\'s to court because they felt that their overweight was caused by hamburgers and French fries. A spokesman of McDonald\'s then claimed that this was impossible. \'Our food is perfectly healthy\', he told the media. Morgan Spurlock tested whether this claim is true or not. Is it possible to live on fast food only? Spurlock had many interviews with specialists and obese people, and for one month he ate at McDonald\'s three times per day. He did not allow himself to eat \'normal\' food. Spurlock wasn\'t exactly poisoned in the end, but his health seemed to have suffered from the fast food-only-diet. I am curious about the effect of such an eating pattern. I don\'t visit McDonald\'s or other fastfood restaurants very often myself, so I wouldn\'t know.'),(12965,3551359,'02,August,2004','2004-08-02 00:00:00','   Where are the men in blue when you need them? For the past hour I\'ve been trying to reach the press officer of the police to ask him some questions, but it seems like the entire station has been deserted. No one answers my calls! What if I was robbed right now, and I needed help?'),(12966,3551359,'02,August,2004','2004-08-02 00:00:00','   And of course everything went as I predicted in my post earlier this afternoon. It rained cats and dogs on the way back from my interview. Within a minute after I had taken off on my bicycle, I was soaked. My rain coat did not help to prevent this. I often become angry when it rains as hard as this while I am underway. I hate rain. Usually, when there\'s no one around to hear me, I start screaming and swearing at the weather gods, which makes me look like lieutenant Dan a bit (people who have seen Forrest Gump know what I mean).'),(12967,3551359,'02,August,2004','2004-08-02 00:00:00','   For a free lance journalist, I\'ve got an enormous handicap: I don\'t own a car and I can\'t borrow one either. This is a problem, because I work for the regional press, which means that I have to travel from one place to another most of the time. Today I\'ve got another interview planned near a small town, 16 kilometers north from where I live. The guy I have to talk to lives in the middle of the woods. That means that there aren\'t any busses that can take me to his place. If I had enough money, I would instantly buy myself a car. Now I have to cycle all the way out there. I bet that it\'s going to be raining this afternoon and I will probably also have to clash with those cycling tourists again (see previous post). Great.'),(12968,3551359,'02,August,2004','2004-08-02 00:00:00','   We Dutch love to ride our bicycles. It\'s not so strange that we have a lot of bicycle paths in this country. Especially the rural areas are dotted with them. Drenthe, the province where I was born, is one of those regions. Every summer, thousands of cyclists from all over the country flock to Drenthe to participate in a large scale annual cycling event,  De Drentse Fietsvierdaagse . As of today, thousands of tourists will set out to discover the province for four days, cycling along special routes the organization has planned for them. There is nothing wrong with this event. Basically, I think it\'s a great initiative. Nevertheless I was a very frustrated by some of the participants this morning.   At 09.00 o clock I had an appointment for an interview in a small village, not far away from the city where I spend most of my time this summer. The weather was OK, I had no car at my disposal and I am not a big fan of the local public transit system, so I decided to grab my dad\'s mountainbike. Everything went well. I was there on time, and the interview was very pleasant.   On the way back, I came along one of the routes of the cycling event that I mentioned. For ten kilometers I had to cycle in the opposite direction of this route, so I had to pass hundreds of cyclists. Normally, this would not be a problem, because most of the paths are wide enough for cyclists to pass each other comfortably. Many participants of the event were irritated by my presence, however. These people refused to give me any space, and they often pushed me over the verge of the path. They constantly scolded at me for cycling against the flow. I became very angry when another elderly lady barked at me. I yelled back: \'For god\'s sake, woman, these bicycle paths are not just a tourist attraction, ordinary people make use of them too!\' Somehow, these tourists fail to realize that the world was not made for them alone. I find this extremely annoying...'),(12969,3551359,'02,August,2004','2004-08-02 00:00:00','   After Ocean\'s Twelve, the colorful Dutch capital Amsterdam is used as a background for yet another movie: Deuce Bigalow 2. Principal photography for this film starts next week. I saw a part of the first Deuce Bigalow movie on television. It was so lame, it didn\'t even make me laugh once. The second part will probably be even worse. For some peculiar reason I think that our \'national pride\' (the red light district and the infamous coffeeshops) will play an important role in this production. It would be nice for a change if the filmmakers would not incorporate these clichÃ© trademarks of Amsterdam in their movie, but I don\'t have much faith in them. Hollywood\'s just too predictable.'),(12970,3551359,'02,August,2004','2004-08-02 00:00:00','   Went out last night, met a very nice person, had a great time... And now I\'m tired. Actually, I was already tired before last night, because the past weeks have been very exciting. My free lance activities are suddenly blossoming. I received two long term assignments, and it\'s certain for 99 percent that I will get another free lance job on Monday. All year long I had nothing to do, and suddenly every newspaper in the region seems to be in desperate need of free lance journalists. Weird. But you don\'t hear me complaining!   Anyway, it\'s going to be one of those lazy Sundays. I just woke up with a hangover. Not a very bad one, but still... I\'m also half deaf because of the loud volume of the music, yesterday. Let\'s hope that I don\'t have any permanent ear damage. Have to pay someone a birthday visit later this day, but I don\'t feel like going anywhere really. Can\'t stay home though! Today is one of the rare occasions on which I have access to my parent\'s car. I really need to practice my driving skills before I go on vacation. I haven\'t touched a steering wheel in months. That\'s not a good thing, considering the fact that I have to maneuver on narrow mountain roads next week. I can hardly wait!'),(12971,3551359,'02,August,2004','2004-08-02 00:00:00','   On Sunday evening&nbsp;we found a small campsite in the hills above the lake of Annecy, not far from the city.&nbsp;The site is&nbsp;nothing special, but it will suffice.&nbsp;We had to set up our tent while it was raining, so the next day I was happy to be woken up by the sun, which heated up our tent beyond comfortable temperatures.&nbsp;When we got out of our tent, we were surprised to see some blue sky.   We saw Albertville and Chamonix yesterday. I had my first mountain road driving experience. It didn\'t feel quite right.&nbsp;The palms of my hand were sweating all the time. The car (a brand new Fiat Panda) was not the problem. It drives like a dream. The problem is probably my lack of driving&nbsp;experience. I also&nbsp;know for sure&nbsp;that I am&nbsp;just not&nbsp;a very talented driver. Luckily enough, I managed not to crash the car straight&nbsp;into an abyss. I decided to let Tim handle most of the mountain driving. He knows how to maneuver the car.   My traveling companion wants to see the Tour&nbsp;de France, so today we checked the route for the last mountain stage of the mother of all cycling events, which will pass through this region in two days. It\'s unbelievable, but even now there are already cycling fanatics camping along the road. They really don\'t want to miss&nbsp;anything of&nbsp;this event.&nbsp;These fans have&nbsp;occupied the best spots, so I wonder if we will be able to find a good place when we come back here in a couple of days.  '),(12972,3551359,'02,August,2004','2004-08-02 00:00:00','   Just another 28 km of mountain roads&nbsp;and we will have reached Annecy in 13 hours, after driving 1045 kilometers. The weather is&nbsp;disappointing, just like it was in Holland. The&nbsp;sun isn\'t shining because of&nbsp;thick layers of clouds, and since we passed through Luxemburg&nbsp;there was rain now en then.&nbsp;We didn\'t feel like setting up our tent in the rain, so we decided to head straight for Annecy in one day. Fortunately it\'s not really cold here. We\'re going to search for a campsite now.  '),(12973,3551359,'02,August,2004','2004-08-02 00:00:00','   All my friends are on vacation or on the other side of the country, the sun is shining like the A-bomb flash over Nagasaki and I have no work or studying to do (well, actually I do have to study for some exams, but I don\'t feel like it right now). So how am I going to spend this saturday? The logical thing to do would be to pick up a good book, my mp3 player and a cold drink, for a nice, relaxing session on the garden terrace. But then again, I\'ve already been burned like a crisp by the sun and I am in short supply of good books.  I suppose I could also join my family when they drive to Deventer this afternoon. There they are going to celebrate the 50th anniversary of an aunt of mine. My uncle has planned a special surprise. This special thing probably has something to do with water. Why else would he tell everybody to take an extra pair of socks with them? Hmm... I\'m not sure what choice to make. Will my curiosity conquer my saturday morning laziness? Stay tuned and find out!  '),(12974,3551359,'02,August,2004','2004-08-02 00:00:00','   I just got back from the office of the regional  urlLink newspaper  I started working for a couple of weeks ago. They seemed to be very content with the stories I wrote. Yay! Maybe they will give me a more permanent position in the long run. I owe my colleague and ex-fellow student big time for introducing me to the chief editor of  Het Dagblad van het Noorden . I hope I can write for the northern edition of this newspaper too. Gonna\' make some new appointments for interviews this afternoon.  This morning I also went to the optician to check whether my sight has decreased during the past couple of years. I\'ve been wearing spectacles and contact lenses since I was 10 years old. It took some time before I noticed that I could not see as well as other kids. I remember a visit to a cinema during a birthday party. Me and my friends had seats in the back of the movie theatre. Everyone was excited about the Teenage Mutant Ninja Turtles, except me. I couldn\'t see zip. On another occasion, one of my teachers put me in the back of the classroom because I was behaving like a little jerk. I noticed that I could not see the writings on the big blackboard from a distance. That\'s when it occured to me that something was wrong. Anyway, I started out with -4, and now I wear glasses of -7 and -9, which makes me pretty much as blind as a friggin\' mole when I don\'t wear them. I was glad when the optician told me that my sight has not become any worse.  The glasses I bought three years ago are completely worn out now, so I have to choose a new pair. I took five spectacles home to try which one suits me best. Why on earth are glasses so expensive? I mean, come on: 600 euros for a flimsy piece of plastic can\'t be a reasonable price, can it? Luckily enough my parents are prepared to pay for the expense. I wish I could afford laser surgery, so I would be liberated from glasses and contact lenses.  '),(12975,3551359,'02,August,2004','2004-08-02 00:00:00','   A couple of hours ago me and my travel companion&nbsp;came back from&nbsp;France. We left our campsite in the small village Ouroux en Morvan at 09.00 \'o clock this morning.&nbsp; After a long day of driving I\'m feeling&nbsp;pretty exhausted.&nbsp;This is partially because&nbsp;I&nbsp;had to live through some very stressful moments on the way back.'),(12976,3551359,'02,August,2004','2004-08-02 00:00:00','   Yesterday I wrote that I haven\'t seen a single movie in the past month. But this is not correct. I\'ve seen one movie in july. It was a French one, called:   urlLink Le Battement D\'ailes Du Papillon   (2000) , starring Audrey Tatou from  urlLink  La Fabuleux Destin D\' Amelie Poulain   (2001). Strange enough I saw it on a campsite in France, in the car of the friend that accompanied me. This was possible because he had taken his I-Book with him. On one rainy night we sat in the car to watch the dvd. The people who walked by our car laughed at us. Apparantly, taking high tech devices like notebooks, digital camera\'s and mp3-players on a camping trip isn\'t really a common thing to do yet.   Le Battement  is a sweet movie about coincidence, cause and effect. It revolves around the principle of the butterfly flapping its wings over the pacific, causing a hurricane over the atlantic ocean. Or something like that. In other words: small things we do can have enormous consequences. Sounds familiar? Well yeah. That\'s because this concept has been used in many other films. I like to think of this movie as the French counterpart to  urlLink  Lola Rennt   (1998), a German movie about which I wrote something in a  urlLink previous post .  Lola Rennt  seems to cover exactly the same topic.  Le Battement  was worth my time, but the core concept is very unoriginal. The same can probably be said about  urlLink  The Butterfly Effect   (2004), which is currently programmed in Dutch cinema\'s. I don\'t think I want to see this one.  Anyway, seeing only one movie per month is not healthy for a film student, so I bought some dvd\'s to compensate. After going through the bargain bins, I found four titles worth of purchasing. The first movie that I selected was Stanley Kubricks  urlLink  2001:   A Space Oddysey     (1968). I saw it three times in a row, a couple of years ago. The story is quite difficult and the film is full of symbols that are to be interpreted. I did not understand the movie the first time, so I went through it again and again. The experience was exhausting, because of the slow pace of  2001 . I really don\'t think that this film can be understood completely by anyone, not even by the filmmakers. Co-writer Arthur C. Clarke once stated the following: \"If you understand  2001  completely, we failed. We wanted to raise far more questions than we answered\". Kubrick and Clarke have left the viewer plenty of room to come up with his own interpretations, and to me that\'s an important ingredient of good movies. I\'ve been wanting to buy this film for a couple of years now, but I always opted to get other films first. This is probably because for me  2001  is not an easy movie to watch.  I also bought  urlLink  American Beauty   (1999). When I first saw this one, I really liked the cynical character of Lester, the guy with the mid life crisis. For some reason it satisfied me to see him break out of the cage of his life. This movie is also funny and touching at the same time. I can\'t recall any other film that has the same combination. I heard that   urlLink Lost In Translation   (2003)   also mixes drama and cynical humor, but I haven\'t seen that one yet.  The third movie that I spent money on is  urlLink  Alien   (1979). As a kid I was a huge sci-fi fanatic (Ok, I admit: I\'m still quite a sci-fi fanatic). Along with terminator and Star Wars, this movie was among the films that affected me most. It scared the hell out of me, and I was very impressed by the Alien designs of Austrian artist  urlLink H.R. Giger . A friend gave me a book with images of his artwork some time ago, and I am really happy with it. To me, the biomechanical alien is one of the most effective movie monsters of all time.   urlLink  Sen To Chihiro No Kamikakushi   (2001) (a.k.a as  Spirited Away ) is the last movie I bought. Its a Japanese animation film. I haven\'t seen it yet, but it won a lot of prizes and it also received great reviews. I don\'t think I\'m going to regret buying this one. But only time will tell...  '),(12977,3551359,'02,August,2004','2004-08-02 00:00:00','   Just received the fine for  urlLink speeding  on the Emmaviaduct in Groningen. According to the letter, I have to pay 30 euros for driving 58 kilometers per hour on a road where a maximum speed of 50 is allowed. Could have done a lot of nice things with that money.  '),(12978,3551359,'02,August,2004','2004-08-02 00:00:00','   You can say about insects what you want, but despite the fact that they are annoying as hell, they are pretty photogenic. I was playing around with my digital camera yesterday afternoon, and I decided to try its capacities for macro photography. It works pretty well, as you can see. I took the picture of the fly yesterday. The grasshopper jumped in front of my lens some time ago.  '),(12979,3551359,'02,August,2004','2004-08-02 00:00:00','   I hate mosquitos.    In fact, I despise these winged horrors so much that I wonâ€™t go to bed until Iâ€™ve hunted down every last one of them.    Itâ€™s impossible for me to fall asleep when these insects are buzzing around in my room. They are evil incarnate! I canâ€™t stand the thought of blood thirsty vampires sucking the fluids from my veins while Iâ€™m in the arms of Morpheus.    Parasites!    In the summer my bedroom seems to be swarming with mosquitos, and I feel obliged to carry out a preliminary strike everytime I go to bed.    Every night I wage war against the insects!    I say: kill \'em all and let god sort \'em out!  There are many ways of terminating mosquitos.    Some people like to smash them with a thick book, others prefer to use a rolled up magazine.    My favourite way of sending them to their maker is the sock attack.    Itâ€™s simple and very effective, especially in rooms with high ceilings. Just roll up a pair of socks into a compact ball of wool and voila, you have made yourself a deadly, long range projectile to throw at those flying terrorists!    The trick is to hit them hard while theyâ€™re sitting still. A sock may not be as aerodynamic as an air-to-air missile, but with a little training, you\'ll be able to achieve impressive results with it.    This technique has one major advantage over the other squashing methods I mentioned. When you hit the mosquito, you\'ll break its body, but the crash won\'t result in a bloody mess on your wallpaper because the sock is made of soft fabrik.  You\'ve probably noticed by now that I\'m quite a skilled mosquito anihilator. I\'m proud to announce that I have just perfected my technique.    A couple of minutes ago, I managed to pick off a mosquito in full flight, crushing it between two fingers. You need pretty good reflexes to manage that, I can tell you!    But I\'ve got to go know. Many mosquitos are awaiting a painful death.    I am the master!  (And I also need a good night of sleep. Why else am I wasting my time writing ridiculous nonsense like this?)  '),(12980,3551359,'02,August,2004','2004-08-02 00:00:00','   I got bored of working, so I decided to download some music. For some reason I ended up querying songs that I adored when I was between ten and twelve years old. Hilarious! I can\'t believe that I actually liked overtly commercial stuff like  2 Unlimited  back then. Hearing a song like  No Limit  again brings back a lot of memories of similar eurodance groups, like  24Seven ,  2 Brothers On The 4th Flour ,  Capella ,  Haddaway  and  Captain Jack.  There used to be many of these groups, and I liked most of them very much. The songs of these groups were always made with the same ingredients: a catchy melody, a scantily clad girl singing a recognizable tune and a rapper who accompanied the girl with raps. Back in the days these raps sounded really cool to me, but now I understand English, and the raps like the one below (from  Let The Beat Control Your Body , by  2 Unlimited ) just make me laugh really hard.    The booming system brings bass in places     Move your system, change the spaces     When I\'m on the mic you can\'t refuse     No I didn\'t came to preach, I came to amuse     Techno making no mistaking never faking     Always breaking it down, Hey to a party     Now let my beat control your body !        Whahahaha! Im feeling nostalgic here.  '),(12981,3551359,'02,August,2004','2004-08-02 00:00:00','   After spending five of my precious hours on capturing and editing raw footage from my DV camera, my computer suddenly decided to f*ck up. RAAAH!!! I don\'t know what happened, but I\'ve lost my entire project because the files I saved seem to be corrupted. Blast! I guess I\'m gonna have to start from scratch again. Great...  '),(12982,3551359,'02,August,2004','2004-08-02 00:00:00','   Yesterday I decided to join my family to visit my uncle and his Thai wife, who celebrated her 50th birthday with friends and family. The only reason why I went there was because of the Thai background of my aunt. Since I am half Indonesian myself, I\'ve been to a lot of Asian European birthday celebrations. These are far less static and much more fun than Dutch ones. I try to avoid Dutch birthday gatherings whenever I can, because I find them to be boring most of the time. Besides, they\'re always the same. Have you been invited to a Dutch birthday party? Here\'s a short guide that might help you to survive it.  At a common Dutch birthday bash, the first guests usually arrive between 20.00 and 21.00 \'o clock. We Dutch have a tight schedule to keep, so don\'t even think about showing up earlier or later. This will not be appreciated! Upon arriving at the house of the host, the entry procedure will start. After entering his home, you congratulate the host. You also hand over your birthday present as if it were an entrance fee.  What to take with you as a present? A bottle of wine or another alcoholic beverage is a perfect choice if you don\'t know the host very well. The Dutch also have a habit of giving each other money as a birthday gift. A sum of 12-15 euros is sufficient. Be sure to put the money in an envelope or a Hallmark card, otherwise you will be frowned upon. Money is the ideal gift for people you don\'t really care about, because you\'ll probably receive the same amount of money from the host when he visits your birthday party. This makes the money transfer somewhat symbolic. Of course you can also buy a more personal gift, but don\'t exaggerate. We Dutch have an appropriate saying for occasions like this: \'Acting normal is crazy enough\'. This means that showing up with very expensive presents is not done. Other guests might feel guilty about their inferior gifts!  Congratulations! You\'ve made it to the living room. The gathering takes place in this room only. Of course you can celebrate a brief, personal party in the bathroom, but that\'s it. The rest of the house is off limits, so don\'t start walking around! When entering the living room, be sure to shake hands with the other guests and introduce yourself. After this, sit down on one of the chairs that are put down in a circle. Usually people remain on the same seat for the duration of the evening, so after the introduction, it\'s only important to remember the names of the persons sitting next to you. Yes, that\'s right, during the traditional Dutch birthday party guests don\'t mingle for intimate conversations. You are socially bound to the persons sitting next to you for the remainder of the evening, so be careful about the seat you choose! Have you finished talking to your direct neighbours? No problem! Because of the circle in which the guests take place, Dutch birthday parties resemble group therapy sessions. This means you can always join in on one of the ongoing group discussions.  When everyone has arrived, coffee is being served, as well as slices of birthday cake. Remember, there is only one piece of cake for every guest. It doesn\'t matter whether the host has truckloads of cake in the kitchen or not, it\'s just not polite to ask for more. Between 21.30 and 22.00 \'o clock, everyone has had one or two cups of coffee and a part of the cake. At his point, the host will start serving soda\'s, alcoholic beverages and snacks. These snacks are sometimes passed along the circle of people on dishes, before they are put on a table in the centre.  The food is always the same. Usually there is an assortment of nuts, potato chips, cheese, crackers and salads to choose from. If the food is placed on the table immediately. Don\'t start eating right away. You have to wait until someone else makes the first move, or until the host says something like: \'I don\'t want to be left with any leftovers, people\'. Feeling thirsty? You are not allowed to fetch yourself another drink. Don\'t you dare to ask for one either. The host will offer you a beverage when it suits him. You\'ll get three or four refills, depending on what time you leave the gathering.  You can begin to think of going home around 23.00 or 23.30 \'o clock. Social control prevents you from leaving the party any earlier, unless you have a very, very good excuse, like an early appointment the next day. If you\'re not sure about when to leave, wait until someone else gets up to go home and join him. Before leaving it\'s OK to shake hands with all the other guests, but save yourself the trouble, because this is not necessary. Screaming goodbye to the entire room and to no one in particular will suffice.  Now the exit procedure will commence. The host will accompany you to the hall where you fetch your coat. Don\'t walk out the front door immediately. Take your time to talk to the host personally for a couple of minutes. He\'s been busy serving snacks and drinks the entire evening, so you\'ve probably not spoken much to him. Thank him or her for the party and leave the house. Now inhale the fresh air and rejoice, you have just survived a traditional Dutch birthday party...  '),(12983,3551359,'02,August,2004','2004-08-02 00:00:00','   For some reason I\'ve always had an intense disliking of the seventies. The colors orange and brown give me an uncomfortable feeling, the music from that period sucks, and the ridiculous clothing and furniture designs of that time... Well, they\'re just plain ugly. This is probably why I don\'t appreciate Stanley Kubrick\'s   urlLink A Clockwork Orange   (1971), which I saw a couple of days ago. I like most movies Kubrick made. In fact, I own a lot of them. Many of his works are about people losing their humanity, and I really like this theme.  A Clockwork Orange  is definitely not on my wish list, however. This movie looks and sounds so dated that it makes me feel sick.  I know that it\'s not a good thing that I base my opinion about this movie solely on my feelings about the seventies, but I can\'t help it. It\'s just too much for me. I mean, take a look at the apartment of main character Alex De Large and his parents for example. It\'s horrible! The same goes for the music. There are many classical pieces to be heard, and these are brilliant, but Kubrick also uses a lot of garbage made with prehistoric synthesizers that really made me cringe in embarrassment. Even my parents, who lived through the seventies, thought that this music was stupid. They also frowned upon the nihilistic violence and the explicit sexual content of  A Clockwork Orange . This movie was x-rated for a long time. I guess it still has the power to shock an audience.  There are also good things about this movie of course. Otherwise it wouldn\'t be regarded as a classic by so many people. The screenplay was based on the book by Anthony Burgess. It\'s OK but I did not find it to be particularly interesting. It even bored me in the end. The story poses an interesting question about ethics and free will, though. The best reason to see this movie is the performance of Malcolm McDowell however. He\'s brilliant in the role of Alex. McDowell is scary, intimidating, funny, charming and innocent at the same time. He made me laugh more than once.  '),(12984,3551359,'02,August,2004','2004-08-02 00:00:00','   Man, I got out of bed a couple of minutes ago and I feel totally wrecked. I didn\'t get a minute of sleep last night, probably because I still had far too much energy at the end of the day. I just couldn\'t find my way to dream land. Anyway, I don\'t think I\'m going to be very productive today. First thing on my to do list is to make myself a large cup of coffee...  '),(12985,3551359,'02,August,2004','2004-08-02 00:00:00','   For the past two weeks it\'s been quite hot in Holland. The high temperatures (plus 30 degrees celcius) drive many Dutch people insane. They flee from the cities to the beaches and lakes. I suppose you could speak of an exodus, because some days ago the roads leading to major bathing resorts, like Scheveningen en Bloemendaal, were actually jammed because of all those people trying to reach the beaches. From here it\'s a three hour drive to the west coast of Holland. What a horrible waste of time and sweat to spend all day in a car that\'s hot as a furnace, only to get to a place where temperatures are even higher. Some of us are smart enough to stay home. Take my cat for example. He must be feeling like he\'s in Dante\'s inferno with that thick, white fur coat of his. Luckily enough we have a small pond in our garden, so he doesn\'t have to travel very far to cool down.     Clever cat!  '),(12986,3551359,'02,August,2004','2004-08-02 00:00:00','   Earlier this week I had to made an appointment for an interview with the owner of a bar in Mepppel, a place nearby. First I tried cafÃ©  De Beurs , because this is one of the oldest bars there. The owner wasn\'t there when I made my phone call, so I tried to reach cafÃ©  De Kansel  instead. The owner of this bar was glad to cooperate, so we made the appointment for three \'o clock this afternoon. No problem.  A couple of minutes ago, I heard on the radio that  De Beurs  has been on fire since nine \'o clock this morning. The flames are spreading to other buildings as well, so I imagine that it\'s a pretty huge fire. There won\'t be much left of this bar, That\'s for sure. Weird coincidence. If the owner if  De Beurs  had answered my phone call, I probably would have had an appointment with him, instead of with the manager of  De Kansel . And then I wouldn\'t be able to do an interview this afternoon because the man would have been to busy crying about the loss of his prized posession. I\'m sorry for him, but once again I\'m lucky.  '),(12987,3551359,'02,August,2004','2004-08-02 00:00:00','   \"So if I understand this correctly, you can\'t tell me what those technical terms in your own advertisement mean?\" The froglike face of the salesman shows an uncomfortable grin and he shrugs his shoulders in an excusing manner. \"Well, you know, it\'s impossible for us to know about every technical detail of these digital camera\'s. Four years ago we could test each camera ourselves, but nowadays there are just too many new models on the market. I wouldn\'t know where to start\". He pushes the tiny camera into my hands, probably hoping that the mere touch of the device will persuade me to buy it. \"It\'s a fine camera. Brand new model. You won\'t find it cheaper anywhere else. As you can feel it\'s light as a feather. It delivers fine image quality on standard print formats, and...\" I interrupt his standard selling babble. \"Yeah, I\'m sure it does. Look, I\'ve heard stuff like this many times before, but I still don\'t know anything about it\'s capacities. I\'ve been to four electronics stores now, and I find it rather frustrating that none of you salesmen seem to have satisfying answers to my questions about this camera. So tell me, how are you people able to give decent advice to customers if you don\'t know a thing about the products you sell? For all I know you could be vending nuclear missiles here, without you knowing anything about it\". The man ignores my remarks and makes his last effort to trick me into spending my money on the camera. \"Look, you can take it with you to try it for a while. If the camera doesn\'t please you, you can return it to us within a month and we can let you have your money back\". I sigh at his offer. \"You know what? I don\'t think I want that. In fact, I think I\'ll return it to you right now\". I hand over the camera to the salesman, and I walk out of the store.  '),(12988,3551359,'02,August,2004','2004-08-02 00:00:00','   urlLink  Sen to Chihiro no kamikakushi   (a.k.a.  Spirited Away ) (2001) is the strangest animation movie that I have seen in a long time. The vivid colors, the fantastic settings and the weird characters make this movie feel like a child\'s dream. The magical story instantly reminded me of Alice in Wonderland, The Wizzard of Ozz and similar fairytales. This is because these are also about little girls ending up in fantastic dream worlds.  I particulary liked the first half of the movie, where the main character, Chihiro, discovers the weird, colorful town, full of gods and mythical creatures. The next time I\'m going to watch this movie (and I will watch it again), I think I\'ll have to do some reading about Japanese mythology first. This will probably help me to understand the characters better.  Sen to Chihiro no kamikakushi  seems to be some sort of allegory, and I think that it can only be fully understood by someone with enough knowledge about Japanese culture.  I find it hard to believe that this movie is actually meant for children. Some scenes are pretty disturbing. What child wouldn\'t start crying at the sight of a faceless blob of a monster that chews on other characters?  '),(12989,3551359,'02,August,2004','2004-08-02 00:00:00','   I had another one of those lame Sundays today. I didn\'t do anything worth mentioning. It was far too hot inside (30 degrees), so I grabbed my notebook, my cigarettes and a cold beer and sat down in the garden to write for a while. Luckily some of my friends got back from their vacation today. I\'m glad they\'re home again. Hoogeveen was quiet without them. Tonight I visited them to discuss our travel experiences.  I have to do a lot of work tomorrow. I need to finish another article for   urlLink Het Dagblad van het Noorden   before Tuesday and I have to make some appointments for new interviews as soon as possible. I think I\'ve got to write at least four stories in the days ahead, because the next two weeks I\'ll be working for a second newspaper. I\'m going to have a busy time. I\'m not complaining though. In fact I\'m very happy with the work I\'ve got. Finding a job as a journalist isn\'t easy these days. The bad economy messes things up. Companies are less willing to buy advertising space and people don\'t spend as much time on reading newspapers as they used to. Because of this, most papers don\'t have a large budget for freelancers anymore. Anyway, my other job is going to take two days per week, so I guess that it\'ll be necessary to make a decent planning to combine both tasks in an effective way.  '),(12990,3551359,'02,August,2004','2004-08-02 00:00:00','   Yesterday me and a friend of mine rented  urlLink  Mystic River   (2003). In this film a girl is killed in a violent manner. Cops flock to the crime scene as bees to a jar of honey. The perpetrators are nowhere to be found. Two detectives (Kevin Bacon and Laurence Fishburne) are put on the case and we watch them as they do their thing.  This plotline is only used as a motivation for the other, more important aspect of the story however. Mystic River is basically about three old friends, Jimmy (Sean Penn), Sean (Bacon) and Dave (Tim Robbins). They lost contact with eachother a long time ago because of a terrible experience they went through. While they were playing on the street as kids, Dave was kidnapped by a couple of men who abused him for four days. Many years later, the friends are brought back together only because of another gruesome event, the killing of Jimmie\'s daughter, Katie. Sean is the detective responsible for the investigation of her death and Dave is one of the murder suspects.  It occured to me that the three friends have one important thing in common. They are all driving their loved ones away from themselves. Take Jimmy for example. He loves his daughter Katie, but we learn that she was about to leave her father and family because Jimmy doesn\'t want her to be involved with her boyfriend, Brendan. The same goes for Sean. He pushed his wife away from him, probably because of his work as a policeman, and now they barely have any contact. Then Dave. He is the one who was kidnapped as a kid. He never told his wife about this, and so she can\'t relate to his mental problems. This ultimately drives them apart. I think that this element, of driving away loved ones, has something to do with the central thought behind the movie, but I\'ve not been able to figure it out yet. The main plotline is very well done. The acting is great and Mystic River provides with great drama.  The police investigation, the main drive behind the story, isn\'t satisfying however. Part of the fun of watching police detectives solving a murder mystery usually comes from the fact that the viewers can join the investigators in their research. They can try to solve the crime themselves. If they succeed in doing so, they feel as smart as the movie detectives. I think this is one of the things that makes watching police stories so satisfying. In order to achieve this, it\'s nessecary to provide the viewer with clues. They also have to be able to follow the reasoning of the detectives at all time, because this allow the viewers to become actively engaged in solving the puzzle.  A lot of characters and names are introduced in  Mystic River . Sometimes this makes it hard to follow the logic of Bacon and Fishburne (this might have been so hard for me because I was a bit tired when I saw the movie). Besides this, there is no way the viewer can tell who the real killer is, because there are not enough clues provided. The end doesn\'t come as a complete surprise, but it\'s just not satisfying because you aren\'t really allowed to participate in the research that leads to the conclusion of the story.   Mystic River  tends to drag on a bit, but I liked most of it. It\'s not a movie that I would like to see again, however.  '),(12991,1270372,'28,July,2003','2003-07-28 00:00:00','   Driving my mother\'s car is an experience of delight, joy, exhilaration and a tiny wee bit of frustration. This car has no funny tinkling noises that make you think the engine blots are coming undone. No burps and giggles. No, this car doesn\'t play jokes on me like my old slanted Rose. This baby corners like its on rails. The steering is effortless, not made by Mr. Armstrong. Its sooooooo smooth. Like cornstarch on a baby\'s bottom.  But, like my baby Rose, my mother\'s car sounds like it\'s in the 50\'s. My mother likes to listen to old rock and roll. There\'s an Elvis Cd in. And I blare it. So, there I am driving my mother\'s 89 Honda Civic, blasting Elvis like I\'m on my way to a sock hop, reeling around the corners and I just want to hit the gas.  But for some reason the speed limits in Edmonton are the same they were in the 50\'s as well.  Drat.   '),(12992,1270372,'27,July,2003','2003-07-27 00:00:00','   I\'ve left the farm. My last day there was spent alone. All alone. Naturally I had to do the things I don\'t normally get to do. So what kind of hijinx did I get up to? Photographic hijinx that\'s what. I\'m shooting black and white film right now and the idea of arty and personal photographs came to mind.   My camera can be set up to act on a timer and with a bench and a blanket in the pine trees, I created my first black and white nudes. Me as the subject of course since I was alone after all. I hope they turn out because I managed to get alot of sap on my feet and a few pine needles in places and crevices that shouldn\'t be accessed by pointy prickly things like pine needles. I also ended up with a considerable amount of mosquito bites.  Not an entirely pleasant experience.   The question I have now is how and where do I get them developed. Do I do the one hour experience. Make it quick, get out of there fast kind of thing or the can I have those in a week, I\'ll give you lots of time to make extra prints.  Mmmnnn?   '),(12993,1270372,'23,July,2003','2003-07-23 00:00:00','   This one just fell out and its worth posting here.  Could you send me  a bunch of words?  A bouquet of letters  that will attract certain birds.  Arranged all nice,  with a ribbon or two.   A bundle of b\'s c\'s and a\'s,  all letters will do.  Don\'t worry, don\'t fret,  all will be fine. Any words smell pretty in my little mind.   '),(12994,1270372,'22,July,2003','2003-07-22 00:00:00','   I found this quote today while I was looking at photographers on the net.  â€œWe\'re born naked. We actually should live naked--not literally, but in terms of honesty and openness.â€  I agree. I always want to be stark.  I have an deep desire to reveal myself so completely and starkly. To let people see me. Really see me. To not hold anything back, to let my emotions out. I want to strip away my skin and bones, lay myself bare. If only my emotions had a physical form that people could look at slowly bit by bit as to not scare them. Scan me slowly. But its not like that. Me, I come rushing out like a bull penned up. Startling and unexpected, the spectator unprepared, not knowing that the box in front of them holds a Jack. Turn the little wind up key, listen to the sweet song, and then ca-pow! The lids off and all of me pops out. Stuff my springs back down and close the door.  I feel the anticipation, the tinkely tinkely dee, tinkely tinkely dee, winding up, winding up, waiting for the hinge to spring.  I donâ€™t think people really want to know whatâ€™s in the little box. They say they do. Let me in, let me know you, let me see you. But its really the anticipation, the wind up, that they like. Ooohh its mysterious, ooh what could it be, what does it look likeâ€¦  tinkely tinkely dee, tinkely tinkely deeâ€¦..  Just stop winding the key. Donâ€™t let the door pop open. But, boy do I think I will burst. I need to stretch my springs. Just be careful you donâ€™t stretch them so far that they loose their essence.   '),(12995,1270372,'22,July,2003','2003-07-22 00:00:00','   My sister, my only sister left today. Grandma and I drove Catherine to Calgary. Despite our differences and how badly we annoy one another, I think as the years go on we get closer and the differences slip away and we see more of each other in one another. Sheâ€™s quite spectacular in her ability to be patient, and kind. Always looking to keep things smooth where I was always looking to ruffle things up. Both of these are valuable and need one another. How can you ruffle if it isnâ€™t smooth and what to smooth out if it hasnâ€™t been ruffled.  On the way home, we picked up my uncle Gary. I asked him what he thought my motherâ€™s best quality was and then I asked him what he thought the worst was. We went through this for him and my Grandmother and then for me. What we discovered was one was often necessary for the other. A personâ€™s undesirable quality feeds their best and vice versa.   I suppose it all balances out like sunshine and darkness.   '),(12996,1270372,'21,July,2003','2003-07-21 00:00:00','   Tonight Catherine, Grandma and I went for a drive in the country. We saw a whole variety of animals. First, we saw some cattle near the road. One was down on his front knees (do cows have knees?) and had stuck his head through the fence trying to get to the sweet clover nearby. I got out of the car, grabbed some clover and walked over. They ate it right out of my hand. I got them to raise their heads high and all I could think about was the time I saw cow\'s eyes in little dishes in Biology class and the times I have seen cow tongue saran wrapped to a styrofoam plate in a butcher aisle. I swear I\'ll never eat beef again.   Next we saw 2 female deer standing in the tall hay grass. Did you know that when you see deer standing out in the fields and you scare tham into running, they actually go boing boing boing. They don\'t really run. They jump like kangaroos.  Finally and the weirdest of all..  We\'re driving along, singin a song when all of a sudden, perched there like a hawk surveying the horizon was a small orange fluffy CAT. Like a regular house cat, just sitting there in the middle of nowhere. I went closer and purred hey kitty kitty and sure enough it jumped off the fence post and strolled through the ditch and came up and rubbed itself against my leg. Mangey little thing all scratched up with half an ear missing. Tough little kitty. You could tell it had tangled with something and didn\'t get fed Tender Vittles for dinner. I thought about the city cats I have had over the years and how they wouldn\'t last 2 minutes out here in the wild wild wilderness of rural Alberta.   '),(12997,1270372,'20,July,2003','2003-07-20 00:00:00','   The reunion is over, but I\'ll stay on the farm for another week. Catherine is here until Tuesday and I\'ll drive back up to Edmonton on Saturday with my mother. Tommorrow, Catherine and I are going yarn shopping in Olds and you know what.. I don\'t think it gets much better than that. Shopping for yarn in Olds with your older sister. If I could only knit her some common sense.   Today, she came into the bedroom. I was reading and resting. She asked me how I felt.  Jill? Yes?  Can I ask you something? What? How do you feel right now? Fine. Why? You don\'t feel funny? (She\'s holding her tummy and I start thinking about those pork sausages I saw her scarfing down) No. What\'s the matter? (She lies down on the bed opposite me, pushing all my clothes out of the way. She looks worried and sickly.) I feel all funny. My tummy hurts and I have ringing in my ears and I feel all dizzy. What do you think is the matter? (As though I would know) I don\'t know..maybe it was something you ate. (Hullo. Pork Sausages.)  A moment later she says Frig.   Catherine can\'t swear ever for some reason. I sometimes imagine her getting really mad.. Frig you! Frig off! Jesus Frigging Christ! For Frig\'s sake!  Its dawned on her all of a sudden she hasn\'t taken the anti-depressant medication she takes to stop chewing her fingernails for 3 days. I assure her that\'s probably the culprit to her aches and pains. She gets up and leaves the room. I follow a moment later and find her in the next room on the phone with our mother long distance confirming her suspicions. My confirmation obviously wasn\'t good enough.  See what I mean?  Then I get on the phone with Mom and ask her if she\'s ever wondered what is wrong with her children. I mean we\'re both on anti-depressants. Catherine is overly concerned with her physical state to the point of neurosis and I am overly concerned with my emotional state to the point of mania and did Mom have any thoughts on this.  Well, the day after our blessed family reunion, my mother admits she has felt all along that a terrible mistake happened twice. She was given the wrong babies. Her real children were out there somewhere leading non dramatic lives of a milk toast existence, calmly carrying on. I mean, she says, neither of you look like me.  Isn\'t that nice? Thanks Mom.   '),(12998,1270372,'20,July,2003','2003-07-20 00:00:00','   I don\'t think I really like that poem very much. It fell out when someone commented that I was an early riser.Reading it again it sounds as though I\'m desperatly trying to find peace if mind that is so ellusive. I seem to sound kind of fragile and lost in the poem. It\'s not an accurate representation, and I\'ll make note of that right now to avoid any confusion hence-forth.   So there.   '),(12999,1270372,'19,July,2003','2003-07-19 00:00:00','   Out here on the farm the rooster crows at the crack of dawn coffee\'s on and porrige is cooking  everyone\'s sleeping but Jill is looking where is the wisdom and peace of mind that other ones before her seem to effortlessly find.   '),(13000,1270372,'19,July,2003','2003-07-19 00:00:00','   I woke up this morning, slipped on my jeans, still wearing white tshirt. Still sleepy and unfocused, I got as far as the living room where my sister was still sleeping. I crawled into her bed, snarled Move over and then went in for the cuddle. She acted as though I was some kind of alien leper being who had invaded with poisonous gases and no I didn\'t fart. I got out and woke up from this harsh reality.  Family is flipping flapjacks this morning. I feel like a pioneer.   '),(13001,1270372,'19,July,2003','2003-07-19 00:00:00','   Yesterday was the reunion at the old Hainstock school.  The school my mother used to walk to in the bitter prairie blizzard winter. Its seems so small and cavernous now. The old swing set is still in the yard. Partially overgrown with trees and weeds. Very eerie.   The reunion got off to a start with a car rally. We piled into cars with instructions in hand and set off on a virtual treasure hunt of landmarks, signposts and in search of old farmsteads. Funny and interesting for about a half an hour. I just wanted to stop and take pictures of cows.  When we got back, the place was in full swing. Food production was in overdrive and the potluck parade was on. A room full of strangers tied to me through some connection, I flew the freak flag with my cousin Erin, being the only 2 tattooed anything, especially girls. We huddled on the front step and bitched about the respectability and straightness of the stock we had come from. Alternatively, we relished in the fact that there were no raging alcoholics, overt red necks or havoc and dysfunction. Lucky us to come from a nice, normal functioning family who donâ€™t tattoo themselves, pierce their nipples, noses, belly buttons and other unseen parts, or yell obscenities and drink lasciviously to excess.  We skipped the skits, the presentations, and the PowerPoint demonstration of our family history. At one point I wandered in to find I had won the prize for taking the longest time to get there. I felt like telling the room of nice prairie folk that I had hitch hiked with a bunch of bikers or something. Just to make my mother scoff and my sister scowl. But I came to this with my best behavior packed and left my feather ruffling tools at home for once.   '),(13002,1270372,'19,July,2003','2003-07-19 00:00:00','   Thereâ€™s nothing like a good ride on a quad. A small death machine, spewing gas fumes between your legs, scaring the shit out your older sister who clings for dear life in white linen as you tear around the grass, baseball cap backwards, old lee jeans, tank top exposing tattoos with tree branches whipping at your gentle girl flesh. Dust flying behind you in big plumes as you tear up the gravel driveway leaving bugs in your teeth and dirt in your bra.  Hey Mom! Wanna go for a ride? Well, Ok, but you be careful Jill Ya, Ya. Hop on.  A few minutes laterâ€¦  Jill if you donâ€™t slow down.. I wonâ€™t â€¦..Jill I mean it! Jill, Iâ€™m going to jump off. JILL!!!  Iâ€™m a small, pretty, terror to my family. I scare them, worry them, and leave them wondering what theyâ€™re going to do with me.  If only I would just slow down.   '),(13003,1270372,'19,July,2003','2003-07-19 00:00:00','   There was a ding dong of a door bell I didnâ€™t know existed. The farm has a doorbell? Well whadda know, I thought. â€œArenâ€™t you going to get that?\" Catherine looked at me with impatience and a face that said canâ€™t you see Iâ€™m busy.   Hi. (there are 2 strangers at the door) Hi. You must beâ€¦? Iâ€™m Anitaâ€™s daughter, Jillian. And you must beâ€¦ Iâ€™m Darlene. Iâ€™m a friend of Agnesâ€™. Is she here? No, I think sheâ€™s gone into town. My sister and I woke up to find no one here. (Thinking since weâ€™re having a reunion, they must be distant relatives.) Do you want to come in?  They came in and then Darlene sprung it on me, the real reason she was at the door. They were messengers of God, coming to spread the word. No, they werenâ€™t my â€˜familyâ€™ after all. Not in the sense that I thought. The Jehovahâ€™s Witnesses out here on a farm in the middle of gorgeous nowhere? To my perplexment, I found myself entertaining these 2 nice women. So for the next 20 minutes, I politely listened to the word of God being channeled through them and accepted their offer to read me scriptures proving the world is coming to an end. Then without hesitation I told them that I believed that once there was a God but now he is gone.   Nonetheless, I reluctantly accepted her card and the pamphlet encouraging me to save myself (which happens to be published in 254 languages and tells me what exactly God requires of me. I think the real question that needs to be answered is What is God willing to do that  I require of â€˜himâ€™). I promised, albeit with my fingers crossed behind my back which everyone from childhood knows protects you from the almighty wrath of God for lying, to call. Which I wonâ€™t.   Iâ€™m a bad girl.   '),(13004,1270372,'19,July,2003','2003-07-19 00:00:00','   The day stated at 6:30 am. I awoke suddenly from a strange dream that I no longer remember. The lure of the quiet outdoors was too tempting. I slithered out of bed, still in last night\'s baggy black pants and black tank top, slipped on my shoes and snuck out the door. The sounds of birdsong and silence surrounded my senses. Intoxicating. The crisp morning air brushing up against my newday flesh. I wandered out back near the barn. Looked at the new growth field and waded out in the waist high grass. I felt if I allowed my knees to buckle, I would disappear forever. Swallowed up by the crop.   I took a photo of the old red barn peeking throught the grasses.  Next, I wandered to the garden and looked at the brand new carrot babies pushing through the soil.. I\'m here! I\'m here! They screamed without voice. I layed down on the ground following the advice of a new found friend and listened for the wisdom. I heard the chirper of the morning bird conference. What do they want to say to one another? One bird seemed to be preturbed at my presence. Yakking, screaming at my interference in the morning meeting. I listened, listened...  Hawks crying off far away, Bulls yawning, the flat palm like poplar leaves clapping in the wind, everything was talking and no one was saying anything.   '),(13005,1270372,'04,August,2003','2003-08-04 00:00:00','   My mind is on the subject of will. Excerting your will. Upon others, upon the earth, upon the future. How do we do this safely? Or even productively? How does one excert their will without negative repercussions? How can I? I often seem to get myself in unwanted situations, be  in situations where I am unwanted, or generally cause havoc when I excert my will. How does one excert their will with finess. The delicacy of influencing others without causing quarrel. I am terrible at this. Oh how I plunge forward, leap without looking, jump recklessly with no concern for others.   '),(13006,1270372,'03,August,2003','2003-08-03 00:00:00','   I think I have reached my americano limit. I had abstained. Denied myself the powers of jetfuel. Resisted from the potent brew. Withheld my desire to sip the nectar of the coffee gods. But not today. I gave into whimsy. Released my grip of the word no, blocking the passage down the smooth sides of throat allowing the warmth of a nice americano down. No, indeed, I did not resist. I had not 1 but 2. Not just one shot but 2. 4 god damn shots of rocket fuel and now I pay the price. The sweet cost of coffee elixer.  The silent buzz in the back of my neck, the hum of just the underside of my skin\'s surface, the jerk of my muscles as they spasm underneath all of the sensations of coffee ripping new pathways through my nerves. Unforgiving, unrelenting not willing to let me go just yet.   Can\'t I just sleep this off? Please..   '),(13007,1270372,'03,August,2003','2003-08-03 00:00:00','   Today is my birthday. I slept at my friend Leslie\'s last night and now this morning, my birthday morning, I am here at Leslie\'s. So now I am 32. I feel terribly failed or fantastically successful at not achieving. What will I do? Where will I go from this point on? Where is my future? How will I find it?  And these questions lead me to think of this... '),(13008,1270372,'03,August,2003','2003-08-03 00:00:00','   Fun I had, lots of it, with you.  Silently slipping past theatrical gaurdians.  Striding circularly through chinese chaos. Late night nibbles on words. So many thoughts, that I don\'t sleep as you slumber nearby.  -Jillian    A friend and I snuck into a movie, got seperated in a chinese market, got snacks and talked about poetry late into the night and then had a slumber party. When I couldn\'t sleep barely at all I wrote that and left. '),(13009,1270372,'03,August,2003','2003-08-03 00:00:00','   Skin burning hot Blood rushing gushing madly.....underneath the surface  a thousand lightning bolts  Dance electric and wait rapidly restlessly inpatiently to be grounded  by touch.  I felt like a a live current flared across the the underside of my skin across my back. . '),(13010,1270372,'03,August,2003','2003-08-03 00:00:00','   \"That which is not slightly distorted lacks sensible appeal: from which it that irregularity-that is to say, the unexpected, surprise and astonishment, are an essential part and characteristic of beauty.\" -Charles Baudelaire  this is all about the sublime serindipity. the magic and crazy coincidences, the beauty in ugliness, in the unconventional and deviant. '),(13011,1270372,'03,August,2003','2003-08-03 00:00:00','   This is an extension of the crocus poem. I think its about new day\'s breaking and and going through cycles of the day. And its about being on the beach waiting for the day to pass. And its about spring. And its about rebirth  sunshine yawning rain knocking  under dirt bark breaking bud\'s emerging breath lurking nearby  pebbles rolling waves groaning through sky blue splitting moments lifting beginnings lurking  life starting again '),(13012,1270372,'03,August,2003','2003-08-03 00:00:00','   I feel as though at times I am standing on the precipice of happiness and bliss and in the blink of an eye, a heart beat, a synaspe firing I feel as though I\'m on the ledge of terror and hell. '),(13013,1270372,'03,August,2003','2003-08-03 00:00:00','   Today, I commented to my yoga instructer that the sound of the water spinning round the wheel wells of cars moving past was creating anxiety inside. Ahh, he said, environmental noise. And later, Jabbar, another friend commented on the poetic motion of the statement. I thought, mmn, environmental noise makes for poetic thoughts. '),(13014,1270372,'03,August,2003','2003-08-03 00:00:00','   Ah.the day is crashing in around me.I think too much about the complexities and simplicities and it all swirls around underneath the mundane and the surprising for someone who claims to be thoughtful. '),(13015,1270372,'03,August,2003','2003-08-03 00:00:00','   Favorite Leonard Cohen peoms found by accident again today:    Song  I almost went to bed without remembering the four white violets I put in the button-hole  of your green sweater  and how I kissed you then and you kissed me shy as though I\'d  never been your lover  I like this poem so much for the innocence and simplicity found in the first few lines. There\'s an intimacy found there. Its like being in on a good love. It makes me feel like sunshine on a front porch with creaky floor boards run smooth with time.   and for inspiration:   The Only Poem  This is the only poem I can read I am the only one  can write it I didn\'t kill myself when things went wrong I didn\'t turn to drugs or teaching I tried to sleep but when I couldn\'t sleep I learned to write I learned to write what might be read  on nights like this by one like me  I liked this one for much of the same reason as the first. Because it too shares an intimacy. An inside on the ending is a nice touch in both poems. Right in the last lines, he reveals an identity and immmediatly you feel a belonging inside the poem. Or at least I did. I like L. Cohen. I\'ve liked  Song  for years. I\'ve always thought it was sweet.  Walked around in the grey skied day here in Vancouver. Took some photos of cherry blossums blooming. Made some comments and smiled at lots of strangers. Some days its easy to be available for small-talk and other days its unbearable to get out of bed. Why is it anhilating to think about starting? Anything even, the dishes becomes this monstrous endeavour. And then, thank god, I just start on something and keep going. And off on a roll I go and wont go anywhere until its done. Oh my, I wish it could all be a little more consistent. Routine is another topic on another day that I might didcuss. MMnnnn? '),(13016,1270372,'03,August,2003','2003-08-03 00:00:00','   I wrote this one day (jan 9.01) while pondering the junky wonder of my car   chunky clunky a chunky monkey lucky ducky  swift sleek smooth riders in the storm  rain sparks stall ducky lucky chunky clunky car '),(13017,1270372,'03,August,2003','2003-08-03 00:00:00','   I left you on the street with your drunken friends.  Alluring and rompable,  I found you on a night of boredom and backdoor entrances. Under the 8 dollar cover I slipped Who are you really?  And will our paths trip over each other ever? '),(13018,1270372,'03,August,2003','2003-08-03 00:00:00','   [3/6/2003 11:11:56 PM | Jill Yewell]  pretty pity parties. '),(13019,1270372,'03,August,2003','2003-08-03 00:00:00','   [3/6/2003 11:11:15 PM | Jill Yewell]   Tonight while observing the war news, I concluded that one day the gulf war will be a sitcom on prime time. '),(13020,1270372,'03,August,2003','2003-08-03 00:00:00','   [3/3/2003 3:33:09 PM | Jill Yewell]    Sunshine knocking rain yawning under dirt bark breaking buds emerging life lurking. '),(13021,1270372,'03,August,2003','2003-08-03 00:00:00','   A green hunting cap squeezed the top of the fleshy balloon of a head. The green earflaps, full of large ears and uncut hair and the fine bristles that grew in the ears themselved, stuck out on either side like turn signals indicating two directions at once. Full, pursed lips protruded beneath the bushy black moustache and, at their corners, sank into little folds filled with disapproval and potato chip crumbs. In the shadow under the green visor of the cap Ignatius J. Reilly\'s supercilious blue and yellow yes looked down upon the other people waiting under the clock at the D.H.Holmes department store, studying the crowd of people for signs of bad taste in dress. Several of the outfits, Ignatius noticed, were new enough and expensive enough to be properly considered offenses against taste and decency. Possession of anything new or expensive only reflected a person\'s lack of theology and geometry; it could even cast doubts upon one\'s soul.  -an excerpt from my favorite book in the whole world...A Conferderacy of Dunces by John Kennedy Toole '),(13022,1270372,'03,August,2003','2003-08-03 00:00:00','   I was walking around the farm today remembering all the other times I have been here. You know as far back as 4 years old. Its the same garage, the same trees, the same barn, everything. Today, I wished I had carved something, let a marker, a time bomb, a note written at 8 and tucked away behind a board, behind a pile of junk, forgotten until some wonderful night and the remeberance of having been here before. '),(13023,1270372,'03,August,2003','2003-08-03 00:00:00','   My friend Charles wrote this:   Sunshine morning Blue sky falling around me Sunshine morning  Brave new days. Beautiful people I love your ways Brave new days Come on little sister.   -Charles Andriashek 1965-2000  I found it on a scrap of paper in the garbage. I went to his house when I found out he had committed suicide and found those words. '),(13024,1270372,'03,August,2003','2003-08-03 00:00:00','   Spark died.  Spark was my fire engine red fighting fish who stopped fighting sometime in the night. Not well was spark, I knew. This was coming slowly around the bend. I think I can, I think I can. Except he couldn\'t. Slowly, he slowed down. He stopped flicking he red caboose tail at me. The way he did when playing coyly with the shadows of my presence.  Spark laid down last night on the cool rock bottom, as he has the last few days. Water wing feathers of red fluttering lightly at his sides. At least I knew he was still living. But last night those signs of life I glanced for were getting quiet. Yet consciousness still beat from beneath the scales of crimson shimmer.  I lifted the small flame from the bed of water with the small cold metal spoon. I had chased him with the same spoon days before. Wake up! Keep moving! Just make sure you move every day. Don\'t worry, you\'ll eat later. I raised him near the meniscus, near the food that he wasn\'t eating. Dehydrated worms bloated with fluid, gaining weight until eventually resting silently on the cool rock bed.  I knew it wouldn\'t be long. Life escaping, seeping out, slowing down. Where does it go?   Spark died sometime in the night while I slept and dreamt something..vague I remember barely. I woke up wanting to deny the daylight and its fore brooding presence.I Felt the need to relieve the digested fluids of the evening past. Pass on. Then I remembered the small spark of fishlife up there on the kitchen shelf, who once flicked his fire engine red tail at me while I did the dishes and washed potatoes.  Did I need to flush him too? '),(13025,1270372,'03,August,2003','2003-08-03 00:00:00','   \"Sharp paper edges that slid across my fingertips, slicing under the surface. Not deep enough to draw more than a speck of blood though. The kind of cut that snags on your woolen sweater when I want to wrap myself around you. Damn it, I exasperate and readjust my embrace.\"  The chances of hurting me are not high. And I don\'t feel sorry, stupid or guilty. Mistakes or errors of judgment are just what they are. Kind of like a favorite pair of jeans you ripped, you can mend them and patch them and like it or not they\'re different. Love them for every torn seam or don\'t. Throw them away and swear you won\'t buy that brand again or maybe its the pair that fits you the best. Life\'s like that for me. About acceptance and being comfortable in my skin.(However, I\'m not always good at it.) Loving the things that matter even the bits that hurt, that resonate and hopefully staying on course.   I stayed at my friend Curtis\' last night. He lets me use his computer. I stayed up way too late writing, slept in my clothes on a couch that slowly pulls apart from the slightest movement. Blinding light searing across my face way too early and a groggy half cup of coffee. I write my best analogies and poetic ramblings when I\'m sleepy in rumpled clothes.  Somewhere, latent talents have emerged.  \"When I first saw her it was as though my heart jumped up in my throat, landed on my sleeve and stayed stitched there mysteriously. Held there with the invisible ties that bind. Bleeding from the small pin pricks that held it in place.\"  I seriously have to wake up.  ever, jillian ps. my little burgeoning hat business has started. Sold one already and onto another. There\'s a theory, you become what you know the most. Did it ever evade me that I would become a milliner? '),(13026,1270372,'03,August,2003','2003-08-03 00:00:00','   For those of you who care and for those of you who don\'t and finally for those of you who have no idea what I\'m talking about, I present the following 6 numbers:  1-5-3-6-2-4  The firing order for a slant 6. Mess these numbers up and Rose don\'t run no more. '),(13027,1270372,'03,August,2003','2003-08-03 00:00:00','   Today I engaged in one of my favorite past times, something I can never get enough of. I can spend hours and hours in search of the perfect bra. So, today I ventured forth towards the mountain of lingerie choices facing me in today\'s society.   There are pink ones, purple ones, shocking red ones, glow in the dark white and every color in between. They lift and some seperate. No support or semi support or extra support.  How do I decide? Today I don\'t feel like I need a lot of support but other days I\'m in desperate need of a pick me up.Sizes scanning the first 4 letters of the alphabet and little fraction like numbers. Why didn\'t they tell me in math class my skills could help me one day find the bra for me.  Bras for tshirts and bras for strapless dresses. Some are for the gym and some are for sleeping. Frilly lace things and little embroidered flowers, fancy straps and an array of different closures that are obviously designed to stump even the most dextrous of teenage boys. Bras for little girls and bras for really big girls. And if you\'re not so big but want to be you can get a bra that will help you out in ways that no other piece of clothing can do.   The bras for the boob challenged can mold you into a whole new woman. They\'re actually breasts you can attach to your own. Perfect mound sphericals hiding behind the facade of white lace. They stand up firm and round everyday, no matter how you feel. Molded foam rubber made in such a way to suggest ample fat tissue resides just under neath that thin film of tshirt material. They give the impression of perfect nippleless breasts that somehow silently remind me of my barbie doll\'s boobs. Non-jiggling, non-sagging all the time perfect tits.   But if you want the jiggle, you can get help for that too with your brand new manufactured strap to your body boobs. Some bras even come with a little bag of jiggly jelly like material that makes your firm foam breasts bounce and will probably decieve even further the man attached to the hands as he attempts to answer his internal burning question of \'I wonder what those feel like\'. Well, I\'ll tell ya. They feel like nippleless foam moulds with a bag of jelly in them.   So, after slowly taking in the sights, colours, textures and maddening selection I pulled myself away from my favorite pastime to wander on to pushing my parts into another favorite item of deception and allure. The Shoe.   As I tried on a few pairs of ridiculous, ankle breaking, toe squishing, height adjusting creations of modern man, I passively listened to the following announcement:  Hello Ladies..Is your bra uncomfortable? Find the straps digging in, back aches later in the day? Join us on Friday between 3-5 and consulte with our Wonder Bra Experts and let us help you find the bra that fits you best. But then I wondered why we force ourselves into these uniforms of womanhood, bind ourselves in poor fitting shoes to overstuffed bras, only to need the help of a \'professional\'. Is that what I need? A professional to \'help me\' find my perfect bra?  I\'m getting out of the mall that wants to push me through a machine and make me come out the other side with perfect boobs and longer legs and smoother skin and bigger poutier lips. I like the small boobed, flat footed, makeupless, fashion scentless woman I am already. mmnn? '),(13028,1270372,'03,August,2003','2003-08-03 00:00:00','   The other day my mother remarked that she wanted to sell her dishes for no good reason except to get new and different ones and asked me what I thought. I told her that life was too short to be with a dish you didn\'t like. '),(13029,1270372,'03,August,2003','2003-08-03 00:00:00','   I feel like a seed buried in the deep warm earth. Protected by the pressure on all sides of me. Nurtured by fluid of the tears I\'ve shed in the last couple of months. Trying to get myself to burst through. Break through the thin skin holding me back. Sprout. Push my way up, shove the comforting dirt out of the way. Find light and grow. Soak up the sustenance from the sky and not from within. Become. '),(13030,1270372,'03,August,2003','2003-08-03 00:00:00','   When I finally pulled myself away from the screen of many letters, I found that I was too late for Thai Away. I hope that you did not scheme and plot, knowing full well the operating hours of said establishment, prompting me to stay that one second longer with offers of optical observations.   However, along the way, I found my friend artist extraordinaire Mark Gaskin. Beckoning me to his presence with the help of a black man named Norman, I found him at his usual spot at the bar of Wazubees. A laugh, a gurgle and a mumble later plused with an rather substantial hug, I glanced at the dishonestly priced menu of thrashed fusion fare. This moment of stoppage may in fact have been the deciding factor in my inability to quench the craving for Thai spice. The throbbing bass tones of unrecognizable but incomprehensibly familiar club music making my eyes hurt, I ventured reluctantly on my way. Still in desire for spicy soup, but in debate for funny and ricocheting dialogue, I found myself on the corner talking to 2 street punks who had only moments before been bestowed with gourmet sandwiches on a catering tray. Need food must press on, murmurings from a distant Thai Away were getting fainter.  Upon my finding disappointment from the sight of upturned chairs and a lone employee mopping the floors, I knew my chances were getting slim. I needed nourishment, I was fading. Hallucinations of wet, dismembered limbs frolicking in burnt lasagna from cardboard packages were haunting me. Onward I stalked the wet asphalt beneath my fluvoged feet.  Thai was not to be my gourmet lover tonight. I would have to find another delight. Deserts whispered sweet nothings in my ear but its aromas didn\'t reach me enough to tempt me. Not until I walked on by a former one night stand that I new I had found what I needed to satiate my appetite this evening. Chicken Tikka and me had a round of nibbles and bites.   I chewed lovingly on his tender morsels of spiced breast of fowl. Dipped them slowly through the cooling yogurt, wrapped them carefully with the blanket of Naan and put them to bed on my taste bud tingling tongue. Ahhh. The ménage et toi of bread, sauce and meat lingered and danced, swirling underneath my senses. Slowly they warmed me up from top to bottom, fulfilling my nourishment needs. We were lovers once again, my old friend Tikka and me.  Unfortunatly, his price has gone up and I\'m going home alone. '),(13031,1270372,'03,August,2003','2003-08-03 00:00:00','   The train, like the sky, reminds me of any place any time. The stars are the same no matter where you are and the squeel of wheels on rails never changes. When I\'m on a train, the sway and momentum and my eyes closed and I\'m in another city, another tunnel, a Paris subway sation. Pretend that the announcements are in spanish. the words on those advertisments don\'t look familiar. A subway is a subway is a subway.   As I get off a young girl on rollerskates glides by. Fresh. Alive. And with french accent and that slides my thoughts further and I   am   in Paris.   Walked into a store a few moments later and a lyric about a long red dress is sung and it coincides with my long red coat. A shirt drops to the floor and the sales guy says..I\'ll look after that, they tend to jump out towards people. I say.. I\'ll say, pick up the shirt and it has hand embroidered LOVE written on it and I think all that is meant to be happening is and its always a perfect moment when I pay attention. '),(13032,1270372,'03,August,2003','2003-08-03 00:00:00','   I saw Farenheit 9-11 the other day. I took my friend Heather for her birthday. A beautiful and sunny Saturday and all was well. 2 hours later a hole had been ripped in the cheap veneer of the false reality I had considered to be somewhat true. The fantasy of my percieved reality poured through the tear like the blood through the sheared flesh of my knee after it shredded along the rough pavement of the decrepet east van step last week. Like then, I could do nothing more than wince from the wound, and maybe shed a few tears. Eventually I just had to shrug my shoulders and get up and go on my way. As it was then, I said to myself upon leaving the theatre, \"What can I do? I guess I\'ll find a bandaid later.\" '),(13033,1197086,'06,August,2004','2004-08-06 00:00:00','    Today I finally finished up building this wretched thing.  Last night I glued in what at the time was a perfect windshield.  After holding it down to let the glue setup, however, I noticed that I got two little spots of glue that ran down from the chrome frame that the windshield sets in; I can\'t keep anything nice!  In case you were wondering, in this photo the model is setting on the trunk of my fathers \'57 (see photo in the last post.)  The color is called Tropical Turquoise.'),(13034,1197086,'05,August,2004','2004-08-05 00:00:00','    This is my father\'s 1957 Chevrolet Bel Air Two-door Sports Coupe. Under the hood is a three-year-old 350 cu. in. Chevy crate motor, pushing out 250 horsepower. Setting on top of the motor is a Holley 650 carburetor bolted to a Edelbrock intake. Bolted to the back of the motor is a Muncie 3 speed manual transmission, and the rear end is the only original thing on the car. For more information about my dad\'s car, I have setup a  urlLink website  where I can post photos from trips that we take in the car.  In about 10 or 12 hours I will more than likely be in the garage waxing the front of the car so that bugs won\'t bake into the paint job on the hood when we go to Idaho. We will be driving this car over 2,000 miles to Coeur d\'Alene, Idaho for the 2004  urlLink Bowtie/ Late Great Chevy  International convention without air conditioning, what fun! Yep, that right, no air conditioning because the motor runs \"too hot\" as it is, and adding air conditioning would probably fry the thing. Get this- while driving 5 MPH in a parade our car temperature gage is halfway between C &amp; H, however, when driving 80 MPH down the interstate in third gear the gage is leaning on H and you can smell anti-freeze. My mechanic father and I believe that the shaft in the water pump (which the fan blade is bolted to) is slipping at higher RPMs, thus pumping less anti-freeze/ water through the motor. At lower RPMs, however, the slippage is less, thus causing virtually no change in water temperature.'),(13035,1197086,'05,August,2004','2004-08-05 00:00:00','    Today, or actually yesterday, I finished a devil model car I was building.  The model is a 1953 Chevrolet Corvette Panamericana racecar, however, it ended up being a 1953 Chevy Corvette with engine parts from a 1954 Chevy Corvette because I didn\'t read the instructions.  OriginallyI planned to make it a model that looked like it just rolled off the line, but after screwing up the paint job in two spots on the driver\'s side, and numerous other mess ups, it became the \"Mexico Badger\" Panamericana racer.  I have to admit that with all the race stickers on, it does look cooler, and thanks to a couple stickers on the drivers side, I have a perfect paint job.  I plan to enter this model in a contest in Idaho next week, so I will have photos of it in competition.  If you\'ve made it this far in the reading you are probably interested in the model cars, so here\'s the urlLink  link  to the kit that I built in three weeks.   '),(13036,1197086,'05,August,2004','2004-08-05 00:00:00','    In case you were wondering what the hell  urlLink Internet Explorer  is, it is the computer program that you use to look at web sites on a PC. Well recently my one-year-old 2.0 GHz HP computer decided to start freezing up every time I used Internet Explorer. The thing that pissed me off was that it decided to keep freezing up while I was working on a website for my fathers classic car club ( urlLink Crown City Cruisers ), so I virtually got nothing done after working for almost two hours. Finally I remembered hearing on  urlLink G4TechTV  about an  urlLink Internet Explorer  alternative called  urlLink Mozilla Firefox . After about 10 minutes of downloading, I started up Firefox, and I haven\'t used Internet Explorer since.  I soon found after downloading Mozilla Firefox that you can download thousands of different extensions, which are basically little programs that customize Firefox to your preference. Browser customization is probably the feature that I love the most about Firefox- something that the communist Microsoft Corporation hates.  Mozilla Firefox is what computer geeks call Open Source, which (as Dictionary.com puts it) is a method and philosophy for software licensing and distribution designed to encourage use and improvement of software written by volunteers by ensuring that anyone can copy the source code and modify it freely. This means that if I wanted to, I could write an extension for Firefox, possibly improve the software for some users, and share it freely. Open Source is the future of computer software, and will probably one day take down large corporations like Microsoft.  If you are interested in Mozilla Firefox, or any other Mozilla products check out  urlLink http://www.mozilla.org  for more information.   '),(13037,1197086,'05,August,2004','2004-08-05 00:00:00','    Well since my \"Official Website\" was basically a failure, I have decided to just create a blog and set it up so it is my website.  THX 138- What is it?  It is the California license plate number on John Milner\'s piss yellow 1932 Ford Coupe (Deuce Coupe) in the 1973 George Lucas classic American Graffiti.  The reason that this is the name of my blog is because I am interested in cars,  urlLink American Graffiti  is one of the best movies ever made, and I couldn\'t think of anything better.   '),(13038,4242799,'18,August,2004','2004-08-18 00:00:00','   \"This Child of God\" is a poem that was written by my mother and I\'m adding it on here because it sort of sums up how I\'ve been feeling lately.  Here it is:      This Child of God      I thought today, as time went by,   The many times I\'ve failed....it made me sigh.      Hills I thought were important to climb   Causing nothing, but more wasted time.      A relationship with one to heal this heart   Mislead through blinded eyes, tore us apart.      Another scar appeared on my soul   Falling down too many times takes its toll.      I\'ll stand once more and clear my view   Love comes so seldom...if I\'d only knew. '),(13039,4242799,'15,August,2004','2004-08-15 00:00:00','   My first blog is basically for those females who are married and unhappy in their marriage.  Let me ask ya\'ll a question.  If you\'re so damn unhappy that you flirt with the guys at work, call them on the phone, sneak around to meet them in parks, at restaurants, etc....AND you get pissed off at the females who are single that talk to the men you want....why the hell do you stay married?  Yes, I am the single female that the married females at work are getting pissed off at.  There are 2 guys that I work with that I talk to.  Because I talk to these two gentlemen, there are 2 MARRIED.....females....at work that get pissed off at me every time they see me around these two gentlemen.  Why? Because married female A...we\'ll call her \"Sue\"....wants single guy A....whom we\'ll call \"Paul\".  Well, I\'ll tell you right now that \"Paul\" and I are close and we talk about everything.  \"Paul\" has feelings for me and understands that I need time to be single because of my coming out of a relationship not too long ago.  Well, hell, \"Sue\" can\'t stand that and shoots daggers in my back every time she sees us together.  As for single guy B....we\'ll call him \"Ron\"....I like him quite a bit.  Well....of course married female B....\"Karen\"....acts like she\'s got ants in her pants every time she knows that him and I have been talking or in the same friggin room together.  I know I\'m venting and I apologize to ya\'ll...but my point is this....if you are married and you are unhappy in your marriage....DO SOMETHING ABOUT THE MARRIAGE!!!  And cheating on your spouse is not the solution I\'m talking about.   Be a freaking woman and either be faithful enough to your husband to work it out with him or end the damn marriage.  If you choose to end the marriage...THEN...AND ONLY THEN....do you have the right to flirt, talk to, call, sneak around with whomever you choose.  I mean, come on people, what happened to the sanctity of marriage?  Why the hell get married if you\'re not done dating yet?  This is for the guys and the girls who are married and cheating on their spouse...GET SOME MORALS!!! People who cheat on their spouse make me sick and I have no respect for you.      Thanks for listening to me vent ya\'ll....I\'ll catch ya\'ll another day. :) '),(13040,3828451,'04,July,2004','2004-07-04 00:00:00','   Well today is fourth of July and my family is doing nothing but sitting st home and watching movies and eating a huge meal that my mom cooks like every weekend (its the Italian in her). But I\'m hoping that we get this block party going thats going to happend in a few weeks, well actually next month probably, but i really want this party to happend b/c I\'m gonna be turning 16 soon and want to have a tight party lol!! Well my mom called radio stations to help us with the party and everything and hoping that goes well too. It would have been a lot cooler if i got my licenses on my b-day too, if only i had a car though, hehe. I would go on a small road trip with my friends and not come back untill 2 weeks later (yea right my parents would kill me). Anyway, hope that everyone has a tight Fourth Of July, peace out! '),(13041,3828451,'02,July,2004','2004-07-02 00:00:00','   well well well, i starting a new blog b/c im bored and my friend thought that this would be a good idea. But today was not a good day for me though. I was going to go watch a movie, Spiderman 2 today with friends, so got up around 8:30am today and started my cleaing early so i wouldn\'t have to worry about it later in the day and can get ready for the movies too. So i walked toward my room and this dumb crate was in my way and i stubbed my baby toe on the crate and part of my toe nail came off and i broke it, eerrr! It seems as though bad things only happen to me, oh well. But all day i was in freak\'n pain and didn\'t think much about it untill the sharp pain started getting worse and worse through out the day. So now im on pain killers that are making me feel so weird and in a twilight zone it seems like, that i\'ve been trying to sleep through out the whole day to get my mind off the pain. I even tried to watch BarberShop 2 which didn\'t work b/c my mind was defiantly on my pain. I thought that i can really take pain but i thought wrong lol. But i guess i\'ll get my booty to bed and sleep!! '),(13042,3771668,'30,June,2004','2004-06-30 00:00:00','   For my Beloved Cat:  in the midst of the bustle, two souls collided. disregarding derision, two souls connected. the joy of empathy once found, soured when one refused to be bound. exquisite pleasure became exquisite pain, yet how a soul yearns for both again. how could the healing ever begin, above the deafening ache loss brings. understanding is lost, a bond goes awry. could it ever be restored before a soul dies? the allure of vast pecuniary gain, brings about a loss previously ordained. what is a heart that is being bled? a product of a liaison gone bad? arose from myriad snowballed events? or simply an organ that continues to rant? the anguish unbearable like the noonday sun. those who care matter not, save one.  '),(13043,3771668,'29,June,2004','2004-06-29 00:00:00','   The weather is a little more accomodating than yesterday. The equator can be pretty unforgiving in terms of ambient temperature. Overcast clouds paint a depressing cast to an already drab sky. A brief respite from the heat, but definitely not an environment in which you would want to start every day in. Work is overwhelming, so much to do, yet direction is lacking. Caught a little of the Lee Kuan Yew interview yesterday night. His opinionated ribbing of Malays and the central government in Malaysia, and their racist, biased, treatment of Chinese and Indian citizens, won my sincere applause. Can\'t wait for tomorrow\'s first edition. Eager to see how far the issue can escalate for the press in Malaysia (no shred of journalistic integrity when the government is concerned) will probably be having a field day. Viva la equal rights.  '),(13044,3771668,'28,June,2004','2004-06-28 00:00:00','   My very first blog posting. Feels weird somehow but could be due to the lack of sleep that i\'ve been experiencing lately. I\'m a business development manager for a manufacturer of Car bodykits called ARS. We\'ve been exhibiting at the Singapore Expo Dreamcars Asia show since last Tuseday and haven\'t had a day off yet. The cars in the show were ok, the show was pretty biased towards performance tuning. Racing queens were great tho, chirpy, friendly and hot as red pokers, so if u gals are reading this, u kick mucho ass (will post some pix when i get to it) Writing this blog at work seems wrong somehow but I\'m beyond caring. Slave driver of a boss had me working 14-15 hours since the bloody exhibition started :  '),(13045,3771668,'13,July,2004','2004-07-13 00:00:00','   Just like that, life throws you a curve ball sometimes. Sad thing is it lands in your crotch. Life really sucks.  '),(13046,3771668,'08,July,2004','2004-07-08 00:00:00','   I\'ve oftened wondered as to why we pay such exorbitant prices for Brembo brake kits. Do their calipers offer something so extreme for them to be able to command such a premium? The same goes for struts manufactured by Tein, Koni or Kayaba. Defi Link BF meters would cost you a small fortune, and Sparco bucket seats often have extravagant price tags attached to their designer carbon-kevlar backsides. A GReddy turbo charger would set you back a couple of k. Do the performance of these performance parts really justify what the consumer actually forks out for them? Nevertheless, a plethora of logos on your door panel signifying the \"existence\" of such parts beneath the bonnet of your ride will gain you instant recognition as a force to be reckoned with. Do Nike Shox really make you jump higher, and add pizazz to your game? Does a diamond from Sookees assure you that you are as pretty as the fossilized piece of carbon that you, or some misbegotten sod, spent half a month\'s salary on? Does an Armani three piece insinuate that Donald Trump had better watch out? It\'s funny but unfortunately, I\'m not immune either to such name brand envy. sigh.... Guess it\'s time to accumulate more loot for the meaningless accumulation of more loot.  '),(13047,3771668,'06,July,2004','2004-07-06 00:00:00','   I\'d like to apologize to my blog for the recent hiatus. It felt weird to pen thoughts when I\'m depressed to no end. At this point in time, I\'d like to pose a few rhetorical question. One feels deeply for another, yet knows that he or she is not perfect. Nevertheless, the other demands perfection. Can the process of attaining perfection be, in point of fact, possible? Could the other\'s desire for perfection denote a need for something just a little short of perfection, with all else being unacceptable? Is it possible for a person to emulate the sensitivity and pious characteristics of Saints like Mother Teresa and Francis Xavier? Is it possible to integrate such characteristics into the hustle and bustle of reality (being the constant rat race all of us are somehow involved in unless you are a bazillionaire)? Does the act of loving and caring for another in silent yet extreme distress mean much less than overt expressions of affection? Does thinking and yearning for someone every second of the day signify something not to be taken for granted? Does it indicate something so precious, and thus something to be treasured, like the golden fleece or does it entail that one should drink from the waters of lethe? Dammit my head hurts.  '),(13048,3771668,'01,July,2004','2004-07-01 00:00:00','   Recently, I\'ve made significant progress and clinched a few sizable deals. Most people would be extremely satisfied with fiduciary profit, but I\'m hard pressed to tell all you loyal readers out there that the same applies to me. In truth, food lacks flavor, the air i breathe is rancid, and success in my career provides scant comfort. My relationship with the woman that I adore more than life itself, is tethering on the thinnest of threads. Let this be a lesson to you kids out there. Don\'t screw your relationship up, especially when you truly care for the other. Heed little warnings and admonitions otherwise one day, you\'ll be able to identify with yours truly (a fate that i wouldn\'t want to wish upon anybody else, except maybe the road bully that threatened to smash my head in with his steering lock some time back, but that\'s a story for another time). May the Lord bless all you good people out there.  '),(13049,3572119,'11,June,2004','2004-06-11 00:00:00','   in the name of equality in the name of god people like you piss me off moral panic everywhere fuck you i don\'t need it - i don\'t care  you want censorship  more and more laws like the nazi clone you are you should try to read a book instead of banning, burning them  censorship will solve fuck all you make me wanna fucking puke you make me wanna fucking kill fuck you commie scum get the message now? it\'s: fuck you congratulations - you just won an apocalypse in business class razor blades and broken glass neo-nazis, cops, shit kickers honk-if-you-love-jesus stickers the chief of police says  \'\'hit the lights so that the TV pictures  turn out nice and bright\'\' telling hand picked journalists \'\'this naked man is a terrorist\'\' as they drag him through the street  ...apocalypse...  cops on horses start the old riot thing like the four horsemen when they\'re riding in and some ridiculous SWAT type team is pretending things are what they seem mr perfect hair your TV host smiles and says \'\'good night that\'s all\'\' and he fucking means it stay at home next week: the fall of rome  ...apocalypse...  and it\'s all all over now the western world no more fucking TV junk 112 won\'t help you now so trade your lies for some real life get the message now?  '),(13050,3572119,'10,June,2004','2004-06-10 00:00:00','   i shake my head to the thoughts that prevade when i think of you screaming inside wondering why this all had to hide i wish i would die because this isn\'t worth it this hurt this pain this... excuse for reality makes immortality seem like miles away  i refuse to talk... to think .to know.  '),(13051,3572119,'08,June,2004','2004-06-08 00:00:00','   i think the fact that this site is hidden gives me reason to feel like i can openly express whats going on in my life and not worry that i\'m hurting someone...  summer just started, i\'m excited...but at the same time kind of dissapointed because i can already see myself wasting it away.  i wish i had the motivation every second of every day to accomplish the things that i have on my \'list\'. who knows maybe i will...  life right now...can be described as complicated, but when is life not? to start off i guess i should explain my relationships with people since friends and family are the main excuses for why my life is the way that it is.  family: i just had a long talk with my parents last night about responsibilities and maturity and fun stuff like that, i\'m not really sure if i got anywhere but atleast i tried.  danielle and derek arn\'t very big problems at the moment, were all just trying to survive our own lives...  friends:  justin-  this relationship is complicated enough in itself, justin, my ex...is one of my best friends...and we still act like friends...and sometimes more.  i\'m not really sure what we are, but if i find a name for it i\'ll let you know.  i still care about him...actually i care about him a lot, and this recent break up is still a shock to me even though i know it needed to happen.  i guess what makes this friendship complicated is the fact that i\'m not really sure what justin expects or looks for in me.  i ask sometimes, but he seems to avoid the question, either intentional or not.  not that i\'m assuming that he feels this way or has intended for me to suspect this but i\'m worried that this [insert name of relationship here] is going to turn into some kind of dependency. not for me, but for him....i guess i\'m worried that he\'ll get the impression that because i still love him i\'ll be willing to be \'there\' for him whenever he wants. sounds horrible huh...and inside i really know that justin wouldn\'t intentionally want this to be a result of what were going through.  so thats my motivation to keep hope in the friendship but i guess i can kind of sense a forshadowing of something going wrong in the future.  its like one day we can be happy, hang out...kiss...and just be there for eachother and then the next he can push me away and make fun of me.  then the next time he wants to be a little more intimate he apologizes....its only happened a couple of times, so you would think that i have nothing to worry about.  but imagine the state i\'m in right now...i was in a relationship for a year and two months and we both fell in love.  justin leaves for basic in the summer so we decide that when he leaves we would take a break....something happenes (i\'m still clueless as to what) and he decides to break it off 2 months earlier...so one day we\'re holding hands and kissing and then the next, nothing, we, or i guess i, had to pretend like nothing ever happened and we were just friends.  i wanted to make him happy so...i went a long with it.  now after a week or so i\'m kinda getting used to the idea that, justin isn\'t mine anymore and we start hanging out and having fun as friends, everything seemed to be going well.  so one day...justin invites me over to his house to hang out. were sitting on the couch and he invites me to lay down a snuggle him [this kind of caught me off gaurd because before this there were no intimate or flirting relations] and then the next thing i know justin is kissing me.  i\'m not going to lie, i liked it, actually no i loved it, because here i am sad that i feel like i\'m losing the one person i want and then here he comes wisking me off my feet.  the one thing i learned from what has happened so far was to always be realistic, and i knew deep down that justin didn\'t want another relationship...i mean if i was leaving in 2 months i wouldn\'t want to be tied down too...soo what was going on? i guess it can be summed up that we\'re just two really good friends who still care about eachother a lot and share some intimate moments with eachother.  the question that i\'m facing myself with now is whether this kind of realtionship is what the both of us needs.  i mean for me it did kind of help with me getting over what justin and i had on a deeper level...which sounds weird but stopping our relationship earlier cold turkey for a week was hard to get through, and now that i feel like i still have a part of him with me it helps me stay with reality but also have this \'romantic dream\' with the person i wish i could be with. *sigh* i just wish i knew what was going on in justins head, its nearly impossible to know, he usually resists talking about that kind of stuff...but it is kind of important. no matter how much love i have for him, i\'m not going to stand by and watch myself be used...i just hope that it doesn\'t turn into that.  guess thats all i have time for today....i\'ll post more later  '),(13052,3572119,'07,June,2004','2004-06-07 00:00:00','   summer has just started.... the warm colors emerging filling the air with the scent of freshness like the fresh feeling of a blank new piece of paper a new chapter... .a new beginning.  '),(13053,3719233,'29,June,2004','2004-06-29 00:00:00','   Well, I just finished reading the March 2004 J.C. Blog archiveâ€¦  [*  wipes brow  *]  Boy-howdy, the onset of the Iraq War sure circumvented the nostalgic nature of the place, didnâ€™t it? --- And yet, in reading it now, less than 48-hours after we turned the keys over to the interim government thereâ€¦ [*  shakes head in recognition (literally)  *] â€¦I have to say that I am truly impressed with the degree of deep insight and suspicion that some of my old JC friends and acquaintances exhibited, even as early as the very first days of the war. --- And itâ€™s also damned ironic that Michael Moore somehow managed to be a topical â€˜go-toâ€™ throughout that particular month-long run-on conversation, particularly in light of his most recent cinematic release (and most notably, the subject thereof!)   My sweetheart, Kathy, commented on something yesterday that I thought Iâ€™d mention:  Apparently, reading the J.C. Blog has been very therapeutic for meâ€¦ hell, perhaps invigorating evenâ€¦ because Kathy noticed a certain enlivened shift in my overall day-to-day deportment, and surmised (I believe, 100% correctly) that it is most likely related to my recent reading habits. --- Just simply being reminded that there was a time, once in my life, where I was completely surrounded by high-caliber cogent beingsâ€¦ and of the free-thinking, free-spiritedness that once coursed and surged within meâ€¦ and of who I was and what I was like  BEFORE  I met Kathy, and was married to a diagnosed sociopath (with multiple personalities, a severe prescription drug addiction, reverse-misogynistic tendencies, and a veritable whoâ€™s-who case of nymphomania) â€“ but thatâ€™s a whole`nuther story â€“ It is as though a weight has been lifted from meâ€¦ and I thank each and every member of the old â€˜sceneâ€™ for helping me reconnect with that long-forgotten part of myself.   Unfortunately, even with this newfound (rediscovered?) lightheartedness and optimism, I fear that my well-honed curmudgeonistic tendencies are here to stay!     [*  looks at clock  *]  Yikes! â€“ Gotta go!   Cheers! Gonzo '),(13054,3719233,'25,June,2004','2004-06-25 00:00:00','   Iâ€™m still busy reading ( and I do mean READING - not merely skimming or browsing ) the â€˜Johnson City Storiesâ€™ Blog.  Despite my busy schedule, Iâ€™m now up to the February 2003 archiveâ€¦  WOW! - I had almost completely forgotten what an amazingly talented, thought-provoking, and uncommonly articulate bunch of people were among us back in â€˜the dayâ€™â€¦ and it pleases me to-no-end to see that, despite all self-induced questionable-substance-fueled efforts to the contrary, apparently the vast majority of everyoneâ€™s brain cells somehow survived intact!   Truly; I canâ€™t remember the last time I have so thoroughly enjoyed the act of reading!   So many formative experiences and life-altering (corruptive?) influences derived from my â€˜doing timeâ€™ in JC ( circa 1988 to 1992, or there-abouts ) â€“ I have quite literally laughed to the point of tears upon reading some of the accounts.     Starting the JC Blog e-venue was an EXCEPTIONAL idea!  _____________________________________________    Please forgive the following, but I really have no one else to tell ( least of which my life-mate/almost-but-not-quite-wife of 10-years! )  As a direct result of my recent reading practices, and the subsequent jarred memories that have ensued, I was awoken this morning by a dream that replayed for me, in vivid-technicolor detail, a number of scenes from one of the all-time greatest nights of my lifeâ€¦ and being a tale of wanton lust and debauchery, its romp through my subconscious resulted in some truely momentous physical side-effects.   In addition to a racing pulse and a dead-cold sweat, I also happened to possess  the  be-all/end-all of sunrise statuariesâ€¦ Ladies & gentlemen, Iâ€™m speaking not of mere morning wood here, but rather the sort of remarkably raging erection thatâ€™s (supposedly) unattainable for a man whoâ€™s quickly approaching 40-years of ageâ€¦ `twas a straight-up cast iron coat rack of a boner, I say; worthy of proportional effigy within any-a third world countryâ€™s urban mythology!  ( Let the congregation say \'Amen!\' and brother\'s an sisters, can I get a \'Doy-yo-yo-yo-yo-yo-yo-yo-yoing!\' ?  )   Ahhâ€¦ and to the delightful and breathtakingly beautiful young lady who inspired such a provocative salute ( who, for proprietyâ€™s sake, shall remain nameless ), I offer my deepest and most heart-felt appreciation for choosing me that night to share of herself with such awesome expression and hedonism, without even one solitary thread of restraint.   NOTE: Once I finally manage to read all of the archives, and then submit myself for membership at the JC Blog - perhaps, upon request, I might be coerced into recounting the NC-17 version of that nightâ€™s events for \"posterityâ€™s sake\".   _____________________________________________    On decidedly different note entirely:  To whomever it was that mentioned him, I wish to thank you for finally putting a name to the face; one that I have been completely unable to recall for so long nowâ€¦ Josh Morris - To this day I have yet to find another opponent at backgammon of his caliber (or even pinball for that matter).    Yes, despite all of the small-town banality, they were indeed good times!   OK, I suppose thatâ€™s about enough rear-view yammering for nowâ€¦   Cheers! Gonzo '),(13055,3719233,'22,June,2004','2004-06-22 00:00:00','   What can I say?   OK... Right now I\'m listening to the MP3 â€˜albumâ€™ version of \'Don\'t Eat the Yellow Snow\' by Frank Zappa, and drinking my third Warsteiner, after a long, hard day of pretending to be a Relational Database Consultant.   For my own personal entertainment, I am about to embark upon what will undoubtedly become an outlandish and mystical journey down Remembrance Quay, over at the Johnson City Blog.   BTW: For anyone who might be looking here, solely in the hope of unmasking meâ€¦ Trust me; my identity is not anywhere near as important as all of the crystal-clear, uncannily accurate memories that I can still vividly recall about those wonderfully twisted misspent youthful years.   Oh, and one more thingâ€¦ The first person to correctly put a name to this enigmatic former JC resident wins their choice of either a Free Rolling Rock Beer -or- a dupe of my prized bootleg VHS copy of Stinky Fingerâ€™s reunion show at Websterâ€™s!    Ahhhhâ€¦ The game is afoot!   Cheers! â€œGonzoâ€ '),(13056,3835889,'24,July,2004','2004-07-24 00:00:00','   Hahaxz finally I\'m so happy liaoxz...\'cause I finally get to talk to WanYi liaoxz after so long! Lolxz she very busie mahxz thats y I didn\'t managed to get to tok to her online...My night is her day mahxz...So finally get to tok to her! hahaxz...Anyway I\'m so happi now that even there was slight accident in my room (water spilled over near my laptop and right onto the mouse), it doesn\'t even affect me...hahaxz I\'m still very happi! Anyway I wonder hows my friends studying back in Spore...Priscilla, hows everything huh? Can u handle the studies back home? Hahaxz heres my message to you: ã‚ˆã„é‹! Lolxz I wouldn\'t tell you what that means...hahaxz go find out yourself bahxz! Lolxz  &nbsp;    Anyway I really dunno how to express myself....hahaxz \'cause I am really v happi! hahaxz Lolxz my school going to start very soon liaoxz so I might not hav the time to come online everyday to chat already...I will still try anyway. Nxt week Saturday going up to Arizona&nbsp; (Spore\'s Sunday) and I\'m looking forward to the trip...Anyway I\'m going to sleep now...kinda got sleepy through all the excitement...ì•ˆë…•!'),(13057,3835889,'23,July,2004','2004-07-23 00:00:00','   I woke up this morning v early...To play GB! Lolxz man its a damn fun game and I earn lots of GP today. Anyway I also got nothing to do so I hav lots of time to play the game. Yep that\'s right I am very free until I can come online everyday to add all these blogs. Hahaxz. Its the same day no different like any other day...juz lazing around the house that\'s all.  &nbsp;    Its so damn boring that I don\'t even feel like typing all these blogs. Kinda getting sick of coming here to type all these thingy...Hmm maybe I shld not be coming here everyday to send my blogs bahxz. Anyway me juz had a thick red wine so I\'m gonna go sleep now...Buaixz Buaixz'),(13058,3835889,'22,July,2004','2004-07-22 00:00:00','   Oh man time really passes so quickly...My school is gonna start soon and I am getting kinda nervous. I guess it will be a total whole new experience and friendships when I start school. Anyway I juz realized today that...Its almost the end of July already! I have been here in Tucson for more than 1 mth already! Wow...I do wonder how my friends back in Spore are doin...  &nbsp;    Anyway, its really a boring day and I realli dunno what else to do except playing GB. Yep I juz got promoted again, thxs to the hard work of playing everytime. I guess when school starts, I will not hav much time playing GB already. Hahaxz anyway I passed my driving today already so I\'m free to go...But still muz practice more on parking the car, though I did relatively well...Lolxz better not \'kua\' myself too much or else get a bloated head. Anyway its kinda boring typing all these stuff too, and I\'m oso getting impatient waiting for my mails from my friends. Argh god knows when will they be arriving...I\'m gonna go sleep now after drinking some wine. *Yawnz* (-_-) good night....'),(13059,3835889,'21,July,2004','2004-07-21 00:00:00','   Today is my birthday! hahaxz I received quite a number of e cards, but was slightly disappointed when some of my best friends forgets abt it...hahaxz anyway, its being a tiring day. Know why I said that? Well first when I went for my driving lessons at 9am, the instructor pulled off a stunt on me and force me into I-10 ( A highway similar to Spore expressway, but muz move at a faster speed). I got the hell out of my life as I have to speed up to 75mph (abt 120km/h) but I onli managed to push the car to 72mph (114km/h) and boy, this sure is the first time that I drove a vehicle at such high speed. Wow man it sure is a totally new experience for me. And tomorrow it will be my final lessons, and I will get my graduated license whereby I can drive alone without anybody sitting beside me...     I was so bored that I played GB after my lessons...Then I didn\'t know what to do nxt so I went to the gym...That made me even more tiring...Well I had a birthday dinner wif my family at church\'s chicken ( something like KFC) and then head back home after that. Watched independence day on TV, then come online to type this stupid blog...hahaxz guess I was so tired that I need to sleep liaoxz...Happy 17th Birthday to Me!~ (^U^)'),(13060,3835889,'20,July,2004','2004-07-20 00:00:00','   Lolxz 1 more day and it will be my birthday liaoxz! Anyway if u all still don\'t know, my birthday actually falls on racial harmony day as well as international friendship day...Hahaxz my friends out there, i dun mind abt having any presents anot but i do think that the most important thing for a true birthday present is sincerity. Without it, giving a present is oso useless.     Anyway, today has&nbsp;being a great day...its my dads birthday! Lol and i drove every1 out for brunch together ( breakfast and lunch together). Actually i got nothing much to say abt today \'cause its a sloooow moving day. Kinda bored actually. Guess the best part of everything was that i got 2 bottles of Loreal Paris products for free! Hahaxz its actually because the cashier made a mistake charging my product, which is US$8.09 each and has to refund me $17+ back, as well as&nbsp;giving me the&nbsp;products FOC! Hahaxz this is the system in US. If they make a mistake, they will hav to give the products to us, FOC as well as refund the full amount of money. The reason is simple: we can actually sue them for charging us the wrong price.&nbsp;That means&nbsp;if i bought a bike here and was charged wrongly, they will hav to refund me and give me the bike FOC!&nbsp;Lol i do love US...hahaxz.Btw i hav to sleep soon, \'cause tomorrow early morning i will be having my 3rd and 4th driving lessons...Good night everybody!'),(13061,3835889,'19,July,2004','2004-07-19 00:00:00','   Wow time passes realli quickly...2 more days and i will be going for my nxt driving lessons, as well as my birthday! Hmm but i guess this year my birthday will be very lonely. The first year which i hav to spend my birthday alone without my friends...especially my close friends ( Hey guys dun be mistaken...the gurl i mentioned in my previous entry \'Weird Feelings\' is my closest friend, not my gf!). I guess this year i wouldn\'t be getting any birthday presents from my friends in Spore. But anyway its alright \'cause its the thought that counts. Its the sincere well wishes that tells me my friends still care abt me...   &nbsp;   I realli do not know what to write anymore, but i do hav something i want to say to Wanyi, my good friend: I am always there for you, whenever you need me, wherever i might be. What ever happens, i will always stand by your side. In life, there will be ups and downs, and its up to you to overcome it. I might be there for you, but all i can do is to give you advise, and the rest is up to you. Hmm i guess i\'m getting too emotional with her...Lolxz emotional as in close friendship! Anyway gtg hav dinner right now, then watch WWE raw life...'),(13062,3835889,'18,July,2004','2004-07-18 00:00:00','   Lolxz man today was a great day man...Nothing to do in the morning \'cause so bored...Watched a movie on TV titled Ocean\'s 11, went out to walmart and the post office&nbsp;with my family ( with me driving the car of course! Lolxz!) bought tons of rubbish, then came home and had dinner with my back home food...Curry! Was that delicious or what? If that wasn\'t enough, we invited our neighbours to our house, and then had a few drinks! Then with me gulping down 4 glasses of red wine and going with crackers, popcorn and tibits, I had to sleep soon...Lolxz sry arxz \'cause me now slightly goggy liaoxz...Lolxz...   &nbsp;   Anyway everything went great for these past 2 days...Had no chance to post my entries yesterday \'cause there was a heavy thunderstorm and had to risk a blackout...Man it was exciting...*giggles* and&nbsp; not only that, I had to reboot my computer yesterday too \'cause I need to install my new Norton antivirus Pro and my windows dun allow it. Anyway me too luo suo liaoxz...hahaxz...Me now miss my friends, especially my&nbsp;good friend Wanyi! Lolxz me getting siao siao liaoxz...Anyway Wanyi,&nbsp;dun be too surprise to c your name here arxz...\'cause I really do misses you and your e mails! Lolxz...I guess I really had to go liaoxz \'cause me really damn tired...Buaixz Buaixz (^_^)'),(13063,3835889,'16,July,2004','2004-07-16 00:00:00','   Today went for driving lessons again...My second lesson and was on a busie road...kinda scary at first but when&nbsp;I got use to it, it was kinda fun...But the worst thing was that when&nbsp;I passed a police car, I was speeding at 60mph! Lolxz can you imagine that? Luckily&nbsp;I was not caught because the patrol car was busie on other things...Anyway&nbsp;I can\'t&nbsp; wait for the nxt lesson, which is on my birthday, 21st July...And some more it is a double lesson! And the nxt day 22nd July, I will be able to get my graduated license, whereby I can drive alone on the road! Wow...Can\'t wait man...   &nbsp;   Anyway when&nbsp;I got back home,&nbsp;I was slightly tired. But my dad wanted me to drive my whole family&nbsp;to a post office which is a few miles away...Boy was&nbsp;I exhausted when&nbsp;I got home. I sure is looking forward to another driving lesson, and my new car...Toyota Camry LE...Sure I do...As I never had the chance to own a car back in Spore...Anyway its getting late and I\'m quite tired, so I\'m gonna rest now...*Yawnz*   (-_-)'),(13064,3835889,'15,July,2004','2004-07-15 00:00:00','   I kinda miss my friends back in Spore, even until now... however there was 1 particular close friend which I had, and I will never forget her. I don\'t understand why I was pretty attached to her, and sometimes I still had that kind of feeling I had back then at the Spore Changi Airport.&nbsp; Today, she was back in my mind again...She stayed with me throughout my chalet back then in April 04 and we had some really great time together. I still remembered that we had pillow fights in the room...We brushed teeth together...Toast mashmallows together...Those were the fondest memories I had.&nbsp; Whenever I am feeling down, she will always be there for me... How I wish I could&nbsp;express to&nbsp;her how I feel all along...   &nbsp;    She kept me wondering why I had&nbsp;the heart to leave her behind...Even if&nbsp;the gurl I liked ask me to stay, I won\'t...But if she persuade me instead, I might stay behind and let go all my dreams of&nbsp;furthering&nbsp;my studies...I guess that\'s y friendship is the most important. Without that special person in your life, you might not know how to get on with it...I guess that\'s y even when I had something to tell her before I left, I didn\'t say though...But its all in my heart...And 1 day, I will let her know how I feel...I do miss her alot........'),(13065,3835889,'14,July,2004','2004-07-14 00:00:00','   Its the most exciting time i had today...I can\'t believe it man...I really did drive for the first time in my life on this day! Although this is my first driving lesson, i had already started driving on the road...Guess why? Hahaxz its bcause the driving lessons here in US starts on any road, particularly anywhere. Unlike Spore, driving can only be learned in a driving school...and so unlimited..   &nbsp;  It all started in the residential area, where the traffic is v light. The instructor taught me to accelerate and many other things which includes using the brakes, the mirrors etc etc...Me n my mum learnt together and i started to lose control when i first drove...nearly went head on into a bush...hahaxz i was slightly nervous mahxz no choice. But everything went on smoothly after that...then it was my Mum\'s turn...she almost rammed into a dustbin! Lolxz but no choice mahxz all beginner. However when the lesson was over, the instructor offered me to drive him and my mum home! I was kinda surprise but still went ahead. This was also the first time which i went into a heavy traffic which was kinda like a route which many ppl used...and boy was i glad that i made it home *phew...* Lolxz...I parked the car (anyhow) after that....It was damn interesting man...  &nbsp;  &nbsp;'),(13066,3851365,'19,July,2004','2004-07-19 00:00:00','   Wendy and I went to Dr. Vic today. (psychotherapist)&nbsp; She went in first, then I went in.&nbsp; Dr. Vic said that Wendy\'s bad feelings have definitely intensified.&nbsp; He said she blames me for everything that is wrong with her.&nbsp; This was a terrible shock to me, as Wendy often tells me, \"You\'re the best mother in the world.\"&nbsp; So which is it?&nbsp; He also said that I should not live&nbsp;in fear, and that Wendy\'s problems were bigger than either of us could handle or assess.&nbsp; He did not know what level of threat she presented.&nbsp; I had told him that Wendy was talking more and more often about getting so angry that she wanted to kill people.&nbsp; I don\'t know if this is a figure of speech or whether she would actually do it.&nbsp; And this is a horrible place for a mother to be.&nbsp; I am afraid of what is going on inside her brain. &nbsp; Dr. Vic told me that I must find a psychiatric social worker to assess whether Wendy could continue to live at home.&nbsp; I never even knew there were psychiatric social workers, nor where to find one.&nbsp; So tomorrow I have to start making calls.&nbsp; Craig said to start with calling Dr. S. (our psychiatrist)&nbsp; But ... what if someone tells me that she has to live somewhere else?&nbsp; Wendy would be miserable.&nbsp; She wants to stay at home with us.&nbsp; Dr. Vic said how do I know she will be miserable?&nbsp; What if she enjoys the structure and company? &nbsp; And where will we get the money? &nbsp; Craig has to go to court next Tuesday.&nbsp; His f***ing ex-wife is suing us again.&nbsp; Increased child support for Leo (because she lost her job, and supposedly can\'t find a new one) and resumed child support for Ariel, who moved back in with her.&nbsp; Apparently it doesn\'t matter to her that the MSA specifies that child support is to end when a child drops out of college (as Ariel did, when she was living with her 34-year-old boyfriend).&nbsp; I guess Ariel resumes being a child when she moves back in with her mother and goes back to community college.&nbsp;  &nbsp; Even though Craig says this is a cut-and-dried case (in our favor), I am still terrified about the outcome, because the first time we went to court, we lost, because the judge was an idiot, and ruled that it was okay for the witch\'s boyfriend to live in the house that we were paying for (via alimony and child support) because she \"wasn\'t aware\" that the situation would affect these payments, even though the MSA spelled it out.&nbsp; In other words, ignorance IS an excuse for breaking the law. &nbsp; Saw \"Fahrenheit 9/11\" tonight and cried.  '),(13067,3851365,'17,July,2004','2004-07-17 00:00:00','   Had a comparably relaxing two days.&nbsp; Even pleasant.&nbsp; Last month I had entered a screenwriting competition in which entrants were randomly assigned a genre and subject, and had exactly one week to write a 20-page (or less) script.&nbsp; Last Saturday the genres/subjects were posted to the competition\'s web site, and today was the deadline to postmark the entry.&nbsp;  &nbsp;  My genre was \"comedy\" and subject was \"plastic surgery\".&nbsp; Craig suggested that I use my idea for the sexist guy who goes into the hospital for a vasectomy, and is accidentally given a sex change operation.&nbsp; I originally planned to write this as a feature-length script (and had written the first act several months ago), so the trick was to come up with an ending within 20 pages.&nbsp; Using a lot of what I already had, I was able to pound out the first 16 pages last Sunday.&nbsp; Today I finally had an epiphany about the ending, and finished by 6:30 tonight.&nbsp; So who knows, I might win something.&nbsp; There are only 11 contestants in my category.&nbsp; I figure writer\'s block and bad writing will rule out most of the other 10.&nbsp; Based on the screenwriting competition I used to direct, 99% of what people write is garbage.  &nbsp;  I had a pleasant lunch today with Jim K., former student (from 18 years ago!) and his new girlfriend.&nbsp; Somehow he tracked me down using Google and a search for \"Lisa Wendy Shelby\".&nbsp; I was a little apprehensive about how I would look so much older than the cute youthful teacher he remembered, but thought, \'What the heck, he\'s 18 years older too.\'&nbsp; Well, I was delighted when I walked into the restaurant and he hugged me and said, \"Lisa, you look exactly the same.\"&nbsp; Actually, he looked pretty much the same too, although much cuter than I remembered.&nbsp; And a doctor, no less.  &nbsp;  The conversation was a little awkward as I was trying to pull his girlfriend into the conversation, but she didn\'t have much to contribute.&nbsp; I hope she didn\'t notice that Jim was pretty much staring at me the whole time.&nbsp; He had that same bemused little smile that Rick used to get whenever he watched me.&nbsp; So okay,&nbsp;Jim must have had a big crush on me and maybe still does.&nbsp;&nbsp; When his girlfriend got up to go to the bathroom for the second time, I realized I\'d better&nbsp;make my exit so they could continue their romantic weekend&nbsp;alone.&nbsp; Either that, or she has a weak bladder.&nbsp; Anyway, I was gratified to be, for exactly two hours today, the object of someone\'s long-term admiration.&nbsp; (well, besides Craig\'s)&nbsp; I of course immediately called Craig afterward and informed him that his wife\'s former student still had eyes for her, and he assured me that he was well aware of having such a good-looking wife.&nbsp;&nbsp; All together now:&nbsp; Awwwwww.  '),(13068,3851365,'15,July,2004','2004-07-15 00:00:00','   Back in San Diego after four days in Las Vegas with Wendy, Shelby, and Esen.&nbsp; When you go on a trip with kids (that including teenagers), it is not so much a vacation as a field trip.&nbsp; They still need to be organized, fed, and entertained round the clock.&nbsp;&nbsp; At this point I am regretting that I agreed to take a foreign exchange student for five weeks.&nbsp; Not because she is annoying, but because of every other crisis that is zapping me emotionally right now.  &nbsp;  While in Vegas I managed to go off on my own several hours to visit Rick\'s mother and step-father.&nbsp; Thank God they were home.&nbsp; I knocked on the door and Bob opened it and didn\'t quite recognize me.&nbsp; I said, \"Bob, it\'s Lisa.&nbsp; Is Eve here?\"&nbsp; So he took me back to the kitchen, and I saw Eve, and started crying, and she saw my face and realized I knew about Rick, so she started crying, and we just hugged and hugged each other.&nbsp; We finally sat down and she began the long horrible story of what happened to Rick.  &nbsp;  As I guessed, Rick\'s problem began right after Mother\'s Day, about May 13.&nbsp; (My last email from him was May 10)&nbsp; He was feeling terrible pain in his lower back, so his wife took him to the emergency room.&nbsp; He had a slipped disc, or herniated disc, something like that.&nbsp; They put a pain patch on his lower back and sent him home with his wife, apparently not telling her about the pain patch.&nbsp; Now, this pain patch never should have been used.&nbsp; This was something that was only used on cancer patients and others who had grown immune to every other pain killer.&nbsp; Not only that, but it was the maximum strength.&nbsp; The lowest dose was 25.&nbsp; They gave him 100.  &nbsp;  At home, he slept the next 36 hours, and his wife was only able to wake him for very brief periods.&nbsp; She finally called the paramedics.&nbsp; As they were transferring him from his bed to the stretcher, he went into massive cardiac arrest.&nbsp; It turns out that the pain patch had lowered his blood pressure so drastically that he barely had any circulation.&nbsp; Bear in mind that I am not a physician and I am only relating what I can remember, and I was absolutely horrified as I was hearing this story.  &nbsp;  This is going to be a long post I\'m afraid.  &nbsp;  When they got Rick to the hospital, they put him on full life support.&nbsp; To make a long story short, they told his wife and mother that if he survived he would have brain damage and be blind.&nbsp; However, over the next few weeks he regained consciousness.&nbsp; He opened his eyes and asked his mother, \"What happened to me?\"&nbsp; It seemed as though he would recover.&nbsp; But then, he began having seizures.&nbsp;&nbsp; His immune system was so zonked that he caught the super-flu.&nbsp; And apparently this is what finally killed him.&nbsp; After six weeks, he passed away on June 29, two days before his 47th birthday.  &nbsp;  I cannot tell you how shocked, sad, and angry I am.&nbsp;&nbsp; I consider this murder on the part of the doctor who put the pain patch on him.&nbsp; Rick\'s wife will sue for malpractice and I hope she wins.&nbsp; I hope that doctor loses his license.&nbsp; This never should have happened.&nbsp; The ironic thing was, Rick never trusted doctors.&nbsp; He would always tell me, \"Fifty percent of the doctors graduated in the bottom half of their class.\"  &nbsp;  And I cannot get Rick out of my head.&nbsp; It has been a week, and the memories do not go away.&nbsp; They just get more vivid.  &nbsp;  Rick\'s mother and step-father invited me to lunch.&nbsp; They were always so wonderful and kind.&nbsp; Over lunch, I showed her the pictures of Rick and me that I had accumulated over the three years we were (mostly) together.&nbsp; She was so happy to see those pictures.&nbsp; She told me how grateful she was that I came, and that Rick was such an important part of my life.&nbsp;  &nbsp;  Thank god I did not throw those photos away.&nbsp; I found out several things.&nbsp; A few of the photos were taken at Julia Pfeffer Burns State Park in Big Sur.&nbsp;&nbsp; Eve said, \"Oh, Rick took us there.&nbsp; He told us that you made him promise never to take another girlfriend there, and he hasn\'t.\"&nbsp; Of course, that made me start crying all over again.&nbsp; Another photo was of a street artist in San Francisco who made paintings using spray paint and can lids.&nbsp; Eve said that Rick had begged her on her last trip to San Francisco to please find that street artist and buy one of those paintings.&nbsp; She looked everywhere and even asked other street artists, but couldn\'t find him.&nbsp; Eve and I looked at each other, realizing that Rick had intended to send that picture to me.  &nbsp;  I was gratified to know that he still thought of me.&nbsp; I guess I already knew that, because he kept emailing me.  &nbsp;  I told Eve and Bob about how after we broke up, Rick opened a brokerage account in my name using my social security number and forging my signature.&nbsp; He invested some money and built it up to about $3500.&nbsp; Apparently he intended to build it up to quite a bit, and then turn it over to me.&nbsp; But at the end of the year he got the 1099-DIV and realized that I would be in big trouble with the IRS if I didn\'t declare this on my income tax.&nbsp; So he had to \'fess up and mail everything to me.  &nbsp;  I guess that was the point that I finally stopped ignoring his emails, and we began our correspondence.&nbsp; At one point he asked, \"Will I ever see you again?\"&nbsp; I wrote back with one word: \"No.\"&nbsp;&nbsp;I was to marry Craig in April, and he went on to marry Carolyn in August. &nbsp;  This is going to sound terrible, but the one thing I can take satisfaction in is that Rick died longing for me.&nbsp; I know he loved his wife, but the absent one is always the one who becomes idealized.&nbsp; In a sense I am doing this with Rick now that he is dead.&nbsp; I have to force myself to remember all the rotten things he did too, like cheating on me with Caroline.&nbsp; And I have to remember that Craig is my true love, the one who cares for me AND my daughters.&nbsp; Damn Craig for being gone all this week at that stupid Juggling Convention.&nbsp; If he were here, I wouldn\'t spend so much time absorbed with Rick.&nbsp;&nbsp; I need my Craig.  &nbsp;  To further mess with my head, when I returned&nbsp;to the hotel room after&nbsp;visiting Eve and Bob, Shelby announced, \"Wendy has been talking to Rick.\"&nbsp; Wendy promptly produced a letter that she had done using her \"guided writing\" (which is where one communicates with, and writes down the words of, a dead person).&nbsp; I sat down and started reading the letter and began to cry all over again.&nbsp; Because it sounded like Rick\'s words.&nbsp; Either that, or Wendy has an uncanny memory of details that occurred over six years ago.&nbsp; One phrase that stood out was \"my precious Math teacher\".&nbsp; Rick always called me precious.&nbsp; But&nbsp;this was&nbsp;always during private moments.&nbsp; &nbsp;How could Wendy know that?&nbsp; Coincidence?&nbsp; I have never had a boyfriend or husband call me \"precious\" except Rick.&nbsp; It was always \"sweetie\", \"sweetheart\", \"honey\", or \"babe\".&nbsp; Today on the drive home I asked Wendy, as she was waking up from a nap, \"Wendy, do you remember what Rick\'s pet name for me was?\"&nbsp; She said, \"No.\"&nbsp; I pressed, \"Do you remember what he used to call me?&nbsp; A nick name?\"&nbsp; She said, \"No.\"  &nbsp;  So now, I have to consider that Wendy may in fact be psychic, and not schizophrenic.&nbsp; If people believe in God, including me, aren\'t spirits in the realm of possibility?  &nbsp;  Yesterday I was wandering around the \"Dessert Passage\" shops in Aladdin while the girls were off shopping.&nbsp; Of course, I was lost in thoughts of Rick, when suddenly I realized what song was playing in the b.g.: \"As I lay me down to sleep\" by Sophie B. Hawkins.&nbsp; This was a special song that Rick and I shared.&nbsp; Another coincidence?&nbsp; No, at that point I firmly believed that Rick was sending me a message that he was a spirit now.  &nbsp;  Well, this is enough for today.&nbsp; I\'m glad to be home, even though Skittles peed all over the carpet.  '),(13069,3851365,'08,July,2004','2004-07-08 00:00:00','   Was sad all day today. Last night I asked Craig if he would mind if I attended Rick\'s memorial service today. Craig said that was fine, that he was not the jealous type. But this morning he told me that he thought I shouldn\'t go, that I should focus on the living (Wendy, Shelby, and the foreign exchange student) and not the dead. And that six years had already gone by, meaning this was in the distant past. (To me, it was like yesterday)   I knew that Craig was right, so I made plans to take the girls shopping down in San Diego. But all day I was filled with an unexpected deep sadness, and memories of things Rick and I had done together. We were only together on and off for three years, and most of it was tumultuous, but we had a deep bond. I don\'t know why I loved him, I just did. And I really can\'t believe he is gone. It\'s like someone made a mistake, whether it was God or karma or whatever.   I tried calling the coroner\'s office to find out why he died, but the woman who answered the phone didn\'t understand the lab report, so she told me to call back tomorrow and talk to the doctor.  '),(13070,3851365,'07,July,2004','2004-07-07 00:00:00','   Was reading the paper tonight and came across my ex-boyfriend Rick\'s photo on the obituary page. My first thought was, \"This is a weird joke someone is playing.\" Or, \"maybe they have the wrong guy\'s name and photo\".   I don\'t know how this happened. I read the obituary through two or three times. Mentioned his wife and 3-year-old son, but not how he died. My guess is it was either cancer (because his father died from cancer) or a car accident. I\'m leaning toward the latter. He had a red ferrari and I know he liked to drive like a race car driver. He got in a serious accident about 10 years ago when he was married to his first wife.   The other weird thing is, I used to get an email from Rick a couple times a month, asking me how I was doing, etc. Sometimes he told me about outings he took his son on. Sometimes he attached photos. I would respond within a week, but he always answered my emails the very next day. The last email came May 10. When I didn\'t hear from him for a few weeks, I wrote him and asked what was going on, was everything all right. Still no word. Two weeks later I wrote again and told him I was worried, and to please write and let me know everything was okay. No answer. So I assumed that his wife somehow found out he was emailing me, and told him to stop. So I respected this (apparent) decision on his part, and did not contact him again. I felt it was for the best anyway, as Rick needed to stop thinking about me and dedicate himself to his marriage.   You may think I have a big head, thinking that&nbsp;2 or 3&nbsp;emails a month from a guy meant he was obsessing on me. But I know Rick better than anyone, and I know he was obsessive-compulsive, especially about women who walked out of his life. All the time he was with me, he mooned over his first wife. Then, when I broke up with him, he sent me emails telling me that he finally stopped obsessing on Candace, that now he was obsessing on me. (well, too bad he couldn\'t get it in sync. I figured this was his new girlfriend\'s problem, not mine.) Rick was a deeply depressed man. I tried to repair him by loving him, by being the perfect girlfriend, but it wasn\'t enough. He needed another girlfriend for the weekends I couldn\'t be with him. Endlessly needy.   Yet, he cared about me. It\'s been almost six years since we broke up, and both of us have remarried, and still he wrote to me religiously. I got used to being \"the woman he obsessed about\". So now he\'s gone and I feel the absence, even though I haven\'t seen him in years or heard his voice. There\'s no Rick out there thinking about me. This is totally self-centered, I know. The real tragedy is the wife and little boy he left behind. The tragedy is that this haunted man never enjoyed the good things right in front of him. A sad life cut very short.   I used to fantasize about running into him some day, one of those bittersweet cinematic moments where the ex-lovers cross each other\'s paths, and make stiff, polite conversation, all the while fighting the need to hug one another and cry. I guess I\'ll never have that moment.  '),(13071,3851365,'06,July,2004','2004-07-06 00:00:00','   Today I took our foreign exchange student, plus another foreign exchange student from Austria staying with a different host family, to the Wild Animal Park.  This was the first time the two girls met, and they got along instantly.  What a relief for Esen to finally find someone with whom to converse in German.  I had to take Wendy with us, because Craig had to go to a meeting, and we did not want to leave her alone.  When we got back, Wendy tried doing another spell in the backyard lighting another piece of paper on fire.  The fire immediately got big enough to be worrisome, so she came to me for help putting it out.  She says that she is through lighting papers on fire, and will only light candles.  We stressed that this can only be under our supervision, when one or both of us is at home.  I got upset with Wendy at the Wild Animal Park when she told me that she tried to use her debit card at the gift shop, and it got rejected.  This is the 5th, and she gets her SSI payment on the 1st, which means she managed to spend $550 in less than 5 days.  Wendy couldn\'t figure out how where the money went, so when we got home I called the credit union and it turns out that she had purchased something on the internet, and the company preauthorized too much money, which was released back to Wendy later today, so she actually has a balance of about $78.  I wanted to take Wendy to an introductory Kabbalah meeting tonight, to try to get her into a more \"mainstream\" spirituality - something that does not involve candles and spells - but she tried on some pants and got extremely upset because she could not fit in them, so I told her to take a walk instead.  I have not heard anything from Dr. Castillo\'s office about the biopsies.  I don\'t even know where she took them from, it looked like she was scraping my uterus.  The last thing I need right now is cancer.  '),(13072,3851365,'06,July,2004','2004-07-06 00:00:00','   This is the handwritten letter Wendy brought down to me, after our talk.    Dear Mom   I would like to say, I am really sorry for what happened.   I suffered alot after being depressed.  So I am so happy, I feel I can get away with too much.   I was vain.  I was very impatient.  I wanted my hobbie to work.  I have \"friends\" on the other side.   They tell me I can get friends in the real world - they see me for who I am.   So I talk to them alot.  I asked them to give me power.   I am very upset I hurt you.  I only tried to help myself.    I want other friends.  Very bad.  But you seem to be the only one who cares.   I should not have let this out on you.  I am strong, and will make more friends.  I am shy, thus I need help from the other side.   Only under your supervision will I ever succeed in any thing I do.   Wendy  '),(13073,3851365,'06,July,2004','2004-07-06 00:00:00','   I oscillate between despair, denial, and cautious optimism.  Wendy seems to be in better spirits the last few days, since coming home from her father\'s.  I really thought we got through to her that she must never lie to us again, and that she must obey the house rules.  But tonight she tried to do another spell in her room involving setting a piece of paper on fire.  Of course, the fire got bigger than the dish containing it, and we ran up there when we smelled the smoke.  Again, Craig and I had to reiterate that she is not allowed to burn things in her room, and that if she wants to burn things outside, one of us has to supervise.  What upsets me most is that she knew she was not supposed to do this, but she did it anyway because she wanted the right to \"get closer to her gods\".  She feels we don\'t understand her religion, whatever that is  - wiccan I guess.    I told her that if she lies or goes behind our back one more time, I am sending her back to her dad\'s.  '),(13074,3851365,'09,August,2004','2004-08-09 00:00:00','   Wendy, Shelby and I visited Dad at the hospital tonight.  It took an hour and a half to get to Riverside.  I told him next time to go to the Kaiser in Escondido.  Anyway, he looked fine, but they gave him a lot of tests to figure out why his heart is racing and he\'s getting shortness of breath.  He may come home tomorrow, but probably Wednesday.  I really hope they can figure out what\'s wrong with him.  I chatted with Anna Mae again tonight, and I told her how snotty Ellie was on the phone with me last night.  She said Ellie had no business to talk to me that way.  She also said that when she was visiting my dad last night in the emergency room, she announced she was \"furious with Lisa\".  I asked Anna Mae why she could possibly be furious with me?  Anna Mae said that Ellie thought I should have rushed up there the second I heard my dad was admitted to the emergency room.  Of course I would have if Anna Mae had told me to, but Anna Mae told me that he was fine, and not to even call him for a few more hours.  So, once again, Ellie is trying to drive a wedge between me and my dad.  And for what?  I already bought the house, and his investments are dwindling, so what can she hope to get out of him?  I wanted to call Ellie and give her a piece of my mind for badmouthing me in front of my dad, but Anna Mae asked me please not to, because then Ellie and my Dad would know I found out from her.  And that this was not a good time, because of my dad being in the hospital.  So I promised her I wouldn\'t say anything right now.  Still having problem with the check from ETS that was bounced by Wells Fargo.  ETS\'s bank Wachovia said they already paid the check, Wells Fargo said they never got the money for the check, and ETS says it\'s my problem.  This has been going on since late June.  I left a nasty message with ETS when I got home from school today telling them that I am taking them to Small Claims Court if I don\'t get a new check in two days.  I cannot believe that in this day and age of electronic transactions, that money can simply be lost and unaccounted for.  It\'s only $107 but I want my freaking money.  Craig retained attorney today.  And so the games begin, anew.  '),(13075,3851365,'08,August,2004','2004-08-08 00:00:00','   The past 10 days has been relatively peaceful and stress-free. Shelby spent the week visiting her aunt, uncle, and cousins in TN. Wendy even spent a few days at Popi\'s. All I had to do was a few of the usual errands - getting Shelby registered at VCHS, etc. In the evenings we planted some desert plants in the back yard.  Back to school tomorrow for 7 days of teacher inservice (which seems excessive). One day meeting with my new group (the 10th Grade team), two days of new teacher inservice, and four days of school-wide inservice. I hate these things. They take two hours to step you through a manual that you could read on your own. Maybe this is so teachers can experience firsthand what students feel.  My dad\'s neighbor Anna Mae called tonight to tell me she brought him to the emergency room. He was experiencing severe pain in his chest, and around his back. Plus, his heart has been racing for about a week now. (He saw his doctor on Tuesday, who changed his prescription) So, he has three kidney stones plus they want him to stay in ICU for a few days to monitor his heart. I talked to him briefly and he sounded in good spirits.  Ellie (his ex-girlfriend) just called to tell me to come to his house tomorrow morning to pick up his shaving kit, toothbrush, and mints, to take to him at the hospital in Riverside. (which is an hour drive from here) I told her I had to work tomorrow, but would be happy to bring them up to him when I got off work in the afternoon. In the meantime, he could buy a toothbrush from the hospital pharmacy for $1.99. She immediately guilt-tripped me with \"Well, I\'ll just bring it up to him tomorrow because he\'ll need these things. But you\'ll need to go see him tomorrow night.\" (Thanks, you bitch, for telling me what to do.)  I really hate her. And I hope my dad will be okay.  The judge deferred Craig\'s case with ex-wife til September.  We have decided to retain a lawyer to help him, because even though he does a good job writing down all the facts, it takes a lawyer to know all the procedural crap.  He found a lawyer who said that we are going to deposition the ex so that she can reiterate her lies under oath, on tape.  Cost of Retaining Lawyer: $6,000-$8,000 Cost of Getting Ex-Wife off our Backs: Priceless  '),(13076,3324571,'14,May,2004','2004-05-14 00:00:00','   > > > >     IT HAS BEEN CALCULATED THAT IF EVERYONE IN THE UNITED STATES DID NOT > > PURCHASE A DROP OF GASOLINE FOR ONE DAY AND ALL AT THE SAME TIME, THE  >OIL > > COMPANIES WOULD CHOKE ON THEIR STOCKPILES. > > > >     AT THE SAME TIME IT WOULD HIT THE ENTIRE INDUSTRY WITH A NET LOSS OF > > OVER 4.6 BILLION DOLLARS WHICH AFFECTS THE BOTTOM LINES OF THE OIL > > COMPANIES. > > > >     THEREFORE MAY 19TH HAS BEEN FORMALLY DECLARED \"STICK IT UP THEIR > > BEHIND\" DAY AND THE PEOPLE OF THIS NATION SHOULD NOT BUY A SINGLE DROP  >OF > > GASOLINE THAT DAY. > > > >     THE ONLY WAY THIS CAN BE DONE IS IF YOU FORWARD THIS E-MAIL TO AS > > MANY PEOPLE AS YOU CAN AND AS QUICKLY AS YOU CAN TO GET THE WORD OUT. > > > >     REMEMBER ONE THING, NOT ONLY IS THE PRICE OF GASOLINE GOING UP BUT > > AT THE SAME TIME AIRLINES ARE FORCED TO RAISE THEIR PRICES, TRUCKING > > COMPANIES ARE FORCED TO RAISE THEIR PRICES WHICH EFFECTS PRICES ON > > EVERYTHING THAT IS SHIPPED. THINGS LIKE FOOD, CLOTHING, BUILDING > > MATERIALS, MEDICAL SUPPLIES ETC.  WHO PAYS IN THE END? WE DO! > > > >     WE CAN MAKE A DIFFERENCE. IF THEY DON\'T GET THE MESSAGE AFTER ONE > > DAY, WE WILL DO IT AGAIN AND AGAIN. > > > >     SO DO YOUR PART AND SPREAD THE WORD.  FORWARD THIS EMAIL TO EVERYONE > > YOU KNOW.  MARK YOUR CALENDARS AND MAKE MAY 19TH A DAY THAT THE CITIZENS > > OF THE UNITED STATES SAY \"ENOUGH IS ENOUGH\"        |  |    |  |  Trash Can | Inbox         Get the latest updates from MSN     MSN Home   |   My MSN   |   Hotmail   |   Search   |   Shopping   |   Money   |   People & Chat     Feedback  |   Help       Â© 2004 Microsoft Corporation. All rights reserved. TERMS OF USE   Advertise   TRUSTe Approved Privacy Statement   GetNetWise   Anti-Spam Policy  '),(13077,3324571,'13,May,2004','2004-05-13 00:00:00','   Ha! Maybe that snappy title will spark the interest of the lock step ideologies who support the most evil administration ever to be in the whitewash.  That\'s right you amoeba brained boneheads!  I wrote this just for you so I can enjoy making your blood boil....Rush is a Drug addict, Donald Rumsfeld is a Sado-Masochist, George Bush is a simple minded puppet, Fox news is really Faux news, Bill O\'Reilly is a LIAR, and Condoleeza Skeeza Rice has blow job lips.    Okay enough ranting about how much I hate Bush, now I will rant about how much I can\'t stand the white trash that live across the street in  low income housing.  Of course they have little American flags strung up all over the place, although for the life of me I can\'t understand why they are so patriotic considering most Republicans would like to have them all sterilized so that they would stop reproducing.  There is one who weighs about 500 lbs and has to use a walker just to get around because she has gotten too fat to even walk.  She has 7 kids, and the youngest is 4, I wonder who on earth slept with her?  '),(13078,3324571,'12,May,2004','2004-05-12 00:00:00','   It\'s humid today and I feel like a sweaty over-weight housewife chained to a never-ending cycle of baby chores, laundry, moping, dishes, cooking, making beds, cleaning out toilets, emptying litter boxes, etc...On top of that is our 110 year old Victorian that is turning into a nightmare to renovate.   We should have just moved into an air conditioned storage unit.  I never want to live in a handyman special again.  Just imagine, we had to pay over half a million in New York for a fixer-upper.  '),(13079,3324571,'12,May,2004','2004-05-12 00:00:00','   urlLink    Bathing is a lonely busieness.&nbsp; urlLink  '),(13080,3324571,'12,May,2004','2004-05-12 00:00:00','   urlLink    oh boy!&nbsp; urlLink  '),(13081,3324571,'12,May,2004','2004-05-12 00:00:00','   urlLink    ringmaster&nbsp; urlLink  '),(13082,3324571,'12,May,2004','2004-05-12 00:00:00','   urlLink    The most fabulous baby in the world.&nbsp; urlLink  '),(13083,3324571,'12,May,2004','2004-05-12 00:00:00','   The antipathy I feel for George W.Bush is consuming me.  It simply amazes me that anyone would support such a horrible and inept president.  What really baffles me are the people who vote against their own economic interests i.e. working class people who definitely do not make 200,000 dollars a year.  A typical example of this type of person is my sister in law.  Her husband makes about 30,000 dollars working for the post office in NYC.  They have to live way outside the city because they couldn\'t possibly afford to live near the city with what he makes.  They have lousy insurance coverage, they have to shop at thrift shops, they never go anywhere because they are so strapped....Yet they vote for George Bush.  Do they really think George W. Bush, a man that was raised with servants and mansions cares about them?  I once tried to tell them that Bush never even owned a ranch until Carl Rove made him buy one right before the 2000 election to make him look like a \"regular guy\".  People like my brother and sister in law are hopeless cases, they will vote for \"W\" no matter what because the only news they watch is Fox News, the official mouthpiece of the Bush banana Republic.   I believe history is not going to be very kind to George.  What kind of legacy will he leave?  Under his leadership the worst terrorist attack  this country has ever seen occurred(he ignored the warnings and consequently 3000 people died)  he cut taxes for the ultra wealthy, he cut funding for education, he destroyed our international reputation by pre-emptively going to war in Iraq on manufactured evidence, the armed forces under his command has been torturing and raping Iraqis, the list goes on and on, and I am going to breathe one huge sigh of relief when this jerk is voted out of office, that is of course if the elections aren\'t rigged by Diebold.  '),(13084,3139191,'27,April,2004','2004-04-27 00:00:00','   smells like teen spirit. such a great song.  '),(13085,3139191,'26,April,2004','2004-04-26 00:00:00','   sorry, sira nanaman ang freindster. ayaw magsend ng message. im sorry, now at least you know the events on april 24. ewan ko ba. baliw ako sa iyo. and frankly speaking, i dont even know where i stand. iyon lang ang gusto kong malaman. im just so lost in you and... and.. nawala na sa isip ko yung paulitulit na sinasabi ko sa sarili ko kanina. basta. im in love with you. and i just dont know.. wla na.. ewan ko.. in love ako sa iyo. period. and i dont  know what to do. and.. it just sucks to be this way. that i might never get to know you. naalala mo ba nung nag videoke tayo kay paolo. i felt you were someone i could really spend my life with. someone who is real and, true, and happy, and great and beautiful and, and, and then reality hit me, na hindi mo ako gusto, pero ewan ko, i just want to hear it from you, ayokong sa iba nangagaling. sorry. pero i need to. and want to know where i stand. kung ano ako sa iyo. sabihin mo, dont be polite, just tell it to me, blunt, frank. ano ako sayo?  '),(13086,3139191,'24,April,2004','2004-04-24 00:00:00','   I want to dedicate this song to you, \"thet\". you will be loved. you are loved. by me. and you are missed. i miss you, a lot. come to think of it, i never really got to know you.   She will be loved  Beauty queen of only eighteen She had some trouble with herself He was always there to help her She always belonged to someone else  I drove for miles and miles And wound up at your door I\'ve had you so many times but somehow I want more  I don\'t mind spending everyday Out on your corner in the pouring rain Look for the girl with the broken smile Ask her if she wants to stay awhile And she will be loved And she will be loved  Tap on my window, knock on my door I want to make you feel beautiful I know I tend to get so insecure It doesn\'t matter anymore  It\'s not always rainbows and butterflies It\'s compromise that moves us along My heart is full and my door\'s always open You can come anytime you want  I don\'t mind spending everyday Out on your corner in the pouring rain, oh Look for the girl with the broken smile Ask her if she wants to stay awhile And she will be loved And she will be loved And she will be loved  And she will be loved  I know where you hide Alone in your car Know all of the things that make you who you are I know that goodbye means nothing at all Comes back and begs me to catch her every time she falls  Tap on my window, knock on my door I want to make you feel beautiful  I don\'t mind spending every day Out on your corner in the pouring rain Look for the girl with the broken smile Ask her if she wants to stay awhile And she will be loved And she will be loved And she will be loved And she will be loved  Please don\'t try so hard to say goodbye Please don\'t try so hard to say goodbye  I don\'t mind spending everyday Out on your corner in the pouring rain  Please don\'t try to hard to say goodbye  '),(13087,3139191,'24,April,2004','2004-04-24 00:00:00','   i am still in love with her. god, why is it hard to stop lovingsomeone that you\'ve loved for thepast three years. i loved her before i even knew i did.  '),(13088,3139191,'24,April,2004','2004-04-24 00:00:00','   id always imagine myself with her. that we\'d slow dance to \"balisong\" by rivermaya.  she\'s just so perfect.. so perfect.  '),(13089,3139191,'24,April,2004','2004-04-24 00:00:00','   Patay  Ilang beses na akong humiling sa Diyos na gumawa siya ng paraan na patayin ako. Ayokong mag-suicide, masyadong maraming nagsusuicide ngayon, masyadong common. Gusto ko iba naman.  Okay lang sa akin ang mamatay, wala naming mawawala sa akin, wala naman akong mahalagang gamit, wala rin namn sigurong purpose ang buhay ko, kung mayroon man, sigurado akong isa itong purpose na kahit sino ayang ganapin.   Naawa lang nga ako sa mga iiwan ko, mapupuyat sila, magagastos.  Medyo gusto ko pa ngang mamatay eh, at least malalaman ko kung may diyos nga talaga, kung may satanas, impyerno, langit at purgatoryo, kung wala naman ang lahat ng iyon, walang diyos, walang satanas, walang afterlife, okay rin. Sana naman, at least, may reincarnation, or buddha or allah or something else. Ayoko lang talagamng paniwalaan na pagkamatay ng isang taoâ€™y tapos na ang lahat.  Madalas kong iniisip kung  ano kaya ang pakiramdam ng patay, yung katulad pa rin ba ng dati? O pag sakalign walang diyos, empty lang ang mararamdaman mo, tapos wala na.  Hindi ko pa alam, hindi pa kasi ako namamatay. Pero kung sakaling mamatay nga talaga ako, nasisiguro ko na hindi sa langit ang bagsak ko.  '),(13090,3139191,'24,April,2004','2004-04-24 00:00:00','   The Lost Soul    urlLink   What sign of the Black Zodiac are you?     brought to you by  urlLink Quizilla  '),(13091,3139191,'24,April,2004','2004-04-24 00:00:00','   urlLink   what will your band be?        name:         age:         sex:         do you lke cats?     yes  no  yes, with ketchup  i\'m more of a dog person  i hate cats!!  i feel indifferent        you\'ll call your band:      newxwavexbattlecore        you will sound like:      pedro the lion        lyrics:      the steering wheel, rusted under her tears, became her sun              This   QuickKwiz   by  urlLink  cowboysbecamefolkheros   - Taken 569 Times.'),(13092,3139191,'21,April,2004','2004-04-21 00:00:00','   urlLink   what will your band be?        name:         age:         sex:         do you lke cats?     yes  no  yes, with ketchup  i\'m more of a dog person  i hate cats!!  i feel indifferent        you\'ll call your band:      newxwavexbattlecore        you will sound like:      pedro the lion        lyrics:      the steering wheel, rusted under her tears, became her sun              This   QuickKwiz   by  urlLink  cowboysbecamefolkheros   - Taken 569 Times.'),(13093,3139191,'21,April,2004','2004-04-21 00:00:00','   nobody made me feel this way before. shes all i ever wanted and more.  '),(13094,3139191,'21,April,2004','2004-04-21 00:00:00','   in love with her. she makes me feel like no other person can. she brings me light when no other candle will. she gives me reason to live when there is totally none. im in love with you maria theresa mendiola. and though i may be too drunk, or STUPID to realize what im doing this very instant, know that what i feel for you is true. those four white flowers, they dont mean anything. at all. i prepared this long speech, about you, and love, and things like it, and who your are to me, but you took all the words away. i learned something, it\'s better to have loved and lost, than never to have loved AT ALL. im in love with you, the whole world and everyone else in it comes second to that. goodnight.  '),(13095,3139191,'21,April,2004','2004-04-21 00:00:00','   may scroll down button po ag blog ko, just in case you couldn\'t see them.  '),(13096,3139191,'20,April,2004','2004-04-20 00:00:00','   !@#@#$#@$^$%&%*^()  dsad:SAD\"SA><>A\"AS:{A}{AS?SA:687654321687qwpeipwqoyhsjxzbv,mxbklwehf!@$!~~~  Testing. 1\" \"! \"2\" \"@\" 3\" \"#\"  abakadaboboka..  '),(13097,3139191,'20,April,2004','2004-04-20 00:00:00','   Revamped my Blog  '),(13098,3139191,'20,April,2004','2004-04-20 00:00:00','   Looks nice eh?  '),(13099,3139191,'20,April,2004','2004-04-20 00:00:00','   Wow. Gradweyt na ako. Pasado. Tapos na ang highschool. College na ako. â€˜Steeg! Lupet! Sa wakas ay maiiwan ko na ang halos kasuklaman kong iskwela sa dami ng mga masasamang alaala na ibinigay nito sa akin.  Peroâ€¦ Kahit papanoâ€™y mamimiss ko rin ito. Hindi ang iskwela mismo, kundi ang mga tao.  Mamimis ko ang pagising ko ng six oâ€™clock para laman makapunta sa iskwela, mamimiss ko ang pag wawalis at pag-aayos ni migs sa classroom nila(na katabi ng classroom namin na katabi ng clasroom ng â€œloveâ€ ko na katabi naman ng clinic/faculty room/AVR [oo, isang room lang iyun]).  Mamimiss ko rin ang pagbati sa mga taong dumadating na kakilala ko tuwing naka-trip ako at hindi masyadong galit sa mundo at lahat ng laman nito.  Mamimiss ko ang pagtitig sa mga magagandang dilag ng aming paaralan. Wala lang, mamimiss ko lang talagang gawin iyon.  Mamimiss ko ang pagtawa sa mga â€œjokeâ€ ng mga komedyante sa IV-Newton.  Mamimiss ko ang mga teacher ko sa 4th year, sila ang pinakamatalinong set ng teacher ko, kaso wala lang talaga akong naalala sa mga tinuro nila (derivative of 1 is 0â€¦ okâ€¦ okâ€¦ kaya pa yan.).  Mamimiss ko ang pag â€œfuck youâ€ ko kay Jimjim tuwing wala akong magawa dahil walang kwenta siyang tao (jokeâ€¦).  Mamimiss ko ang pag-iwas ko kay Pating dahil may utang pa rin ako sa kanyang pera.  Mamimiss ko ang pagsuntok sa pinto ng Galilei dahil nababadtrip ako.  Mamimiss ko ang pagsunto sa poste ng building na nasa tapat ng pinto ng IV-Galilei dahil nababadtrip ako.  Mamimiss ko ang paglambitin sa tabingin crossbar sa tapat ng Galilei dahil wala akong magawa.  Mamimiss ko ang over-priced food sa canteen area. Hoy! Walang kwentaâ€¦ Ang mamahal ng mga tinda niyoh!  Mamimiss ko ang Pag-tusok ng pishbol at squeedbol at kikiyam at pagsawsaw sa napakaduming sauce.  Mamimiss ko ang pag-sulat ng kung anu-anong bagay sa blackboard naming na wala naming koneksyon anu pa man sa pang araw araw na buhay ng IV-Galilei.  Mamimiss ko ang pakikipag-daldalan ko kay Ella at pag-tawag sa kanya ng â€œbalugaâ€ kahit alam namin na mas maitim ako sa kanya.  Mamimiss ko sina Shang Cai Chek, Mao Tze Tung, Gary Lising, Mika Immonen, Bernardo Bernardo, Standard, Boy Nigger, Krishna â€œthe nonsense boyâ€ lapitan, dara, cire, dada at wala nang iba. Mamimiss ko sila dahil naaalala ko na â€œstar-studded ang IV-Galilei. Mamimiss ko ang Computer Lab ng ACNHS â€“ SSC dahil sa dami ng games na available duun, daig pa ang Internet cafÃ©.  Mamimiss ko rin ang Math Park, halos lahat ng kalokohan na ngayon ay pinagsisisihan ko nagyon ay dun naganap.  Mamimiss ko ang bread Stix building na binansagan ni Erin, Duon naganap ang confrontation ng mga fourth year sa Einstein at nagsigawan si Thet at Alfranzel at sumingit si cole para magmoderate at kumulo ang dugo ni Francis, Zhel, Patjo, Darwin kay Angelo Angeles. Walang kwenta. Ni hindi man lang ako nakasingit sa group hug. Pero at least nakakatawa, para silang mga batang nagbabangayan in an almost organized manner.  Dun din ginanap ang walang kwentang recollection namin. Bakit may Born Again? Akala ko ba pag public dapat walang religious Bias?! Punyeta. Tsaka baket may bacolaureate mass??? OYYY TAMA ANG SPELLING NG BACOLAUREATE MASS KOHH! Bakit may Pari? Kala ko ba Public. Ano yan? Christian Formation sa isang Public School?! Fuck that.  Oopsâ€¦ There goes my alumni membership.  Siguro magtataka na lang ako, baka minsan, pag naalala ko ang Highschool life ko, mapaiyak ako. Dahil sa sobrang walang kwenta.  Mamimiss ko talaga ang pag agaw ng tingin ko sa mga sobrang transparent na mata ni Ken (valedictorian namin), sa friendly na mata ni ainah, sa napaka cute na mata ni Jambol at sa napaka-apprehensive na tingin sa akin ni roms (bestfriend nung crush koâ€¦ yata.), mamimiss ko rin ang sobrang laking mata ni darwin santos(!@$#%@#$^# ang laking mataâ€¦), at sa wakas mamimiss ko talaga ang napaka warm na mata ni thet.  Mamimiss ko rin ang pagbigay ni Pocholo sa akin ng Dirty finger dahil naka trip yata siya. ..1,,   ..1,,  '),(13100,3139191,'20,April,2004','2004-04-20 00:00:00','   ok, so a couple of days ago, i was at national bookstore browsing through lyndon gregorio\'s beerkwatro book (really nice book, for college folks, buy it.).  so i was reading the book and having a few laughs when the saleslady comes up to me and says,\" pogi, bibilhin mo ba yannnn?\" she gave me the mataray look and waited for a response.   holding the beerkwatro book (by lyndn gregorio) i had with me, i used my other hand, took out the 500peso (oo, may limangdaang piso ako. steeg, noon lang ako nakahawak ng 500 peso bill, yabang...) bill out of my wallet and flashed it in front of her, and then i took another beerkwatro book (by lyndon gregorio) from the shelves and said \"politely\",\"Miss, actually, dalawa eh, isa para sa akin, tapos isa para sa bodyguard ko, nakakatawa kasi oh..\"  she walked away. left me.  of course, i never really planned to buy that book, nor do i have a \"bodyguard\". actually, natapos ko ang buong beerkwatro ni lyndon gregorio sa isang sulok sa national bookstore.  i just hated the fact that people treat other people according to thier physical appearance and clothing.  ie, i was wearing a white t-shirt(the kind i use as an undershirt for school) and voltes V shorts, the ones with the big ass letter \"V\".  sorry miss, naka trip lang ako.  did you know that beerkwatro was written/created/illustrated by uhmm sino kasi iyon?  thank you lyndon gregorio.  '),(13101,3139191,'17,April,2004','2004-04-17 00:00:00','   ...ok, so its been in on the market for a couple of years now... great game... better than streetfighter...  i mean, its a hell of a lot funnier..  oh... hey, someone actually reads [insert karl\'s name here]\'s shit.. cool.  '),(13102,3139191,'17,April,2004','2004-04-17 00:00:00','   nasa pool kami ni pocholo henson yesterday, nag si swimming. medyo maingay kami, magulo, nagbabasaan nang bigalng sinigawan kami ng isang babae \" PWEDE BA?! NABABASA AKO!!!\"... adik. nice pocholo. this was taken from bob ong\'s latest book, ang paboritong libro ni hudas.  may kwento pa ako, bali may dalawang adik sa pool... hahahahahahahahhahahahahhahahhahaha!!!!!  '),(13103,3139191,'16,April,2004','2004-04-16 00:00:00','   yeah, after the swimming and the eating and the swimming and the talking, umuwi na kami, ang sarap mag closing party... even though only a fraction of our class attended, i buimped into an ex-classmate of my, genesis from phs-ssc, it was nice talking to that guy again, he told me stuff that happened after i switched schools... kirstine is mad at me for like, behaving irrationally... i think...  '),(13104,3139191,'15,April,2004','2004-04-15 00:00:00','   after last night\'s karaoke bliss, i couldn\'t sleep, today, we are going to have our closing party. this sucks ass. everything sucks ass, my head hurts, my feet hurts, gawd, my whole body feels like a big chunk of uranium..  '),(13105,3139191,'15,April,2004','2004-04-15 00:00:00','   nag karaoke kami sa SM, bought more than 20 tokens worth 6 pesos each and brought with us lots and lots of pizza, pizza, rootbeer and karaoke with my former classmates, what you put in coffee even when there\'s milk around. hey, read my post ealier this morning... nice huh? i wrote that...  '),(13106,3139191,'13,April,2004','2004-04-13 00:00:00','   Unang Araw  -simula- 03/09/04 9:02:09 PM  Huling taon ko na, bukas ng umaga ang huling periodic test ko sa buhay ko, yung susunod, iba na yung tawag, â€œmidtermsâ€ yata.   Nakakatamad, parang ayoko nang mag-aral, alam kong dapat akong mag-aral, pero, parang, wala nang gana. Gusto ko na lang makatapos.  Periodics na. Pagkatapos nito, halos tapos na ang lahat. Isa lang yung test ko bukas, ayoko nang mag-aral, gusto ko na lang matulog ng mahimbing at huwag isipin ang sandosenang requirements na kailangan kong i-pass.  Tatlong taon din akong nanatili sa paaralang ito. Ang pinaka-matagal ko sa isang paaralan. Madalas kasi, dalawa o isang taon pa lang, transfer kaagad, madalas kasi kaming lumipat ng bahay.  Nakakalungkot isipin na hindi na ako highschool, na maaring hindi ko na makita yung mga crush ko (sila ang dahilan kung bakit masaya akong pumapasok sa eskwelahan). Tatlong taon din ito ng buhay ko, sa tatlong taon na iyon, nasa loob ako ng pader ng classroom, nagaaral, minsan natututo, pero madalas, nakatunganga lang, kaya pagdating na ng test, â€œcrammingâ€.  Pero sa pinakahuling pagsusulit ko sa paaralang iyon, walang cramming, walang kahit ano, itatype ko lang ito, at matutulog na ako, ayoko nang mag-aral. Pagod na rin kasi ako medyo pinapawisan pa. Hindi naman ako nagtrabaho, at konti lang naman ng inaral ko, pero parang pagod na pagod ako, parang buong araw akong nasa construction site. Hindi ko ba alam kung bakit, malamang ito yung â€œstressâ€ na sinasabi nila. Bahala na. Simula na ng test ko bukas. Hindi ako nag-aral. â€œstressedâ€ ako, matutulog na. Itutuloy ko na lang bukas.  -tapos- 03/09/04 9:28:04 PM  '),(13107,3139191,'13,April,2004','2004-04-13 00:00:00','   hmm... i really wish they\'d let us add pictures in this blog for free....  that\'d be really nice...  '),(13108,3139191,'13,April,2004','2004-04-13 00:00:00','   so.. hmm... oh happy days.. im with my former clasmates to day... so yeah, cool. damn.. they should allow us to upload picture for free or something..  '),(13109,3139191,'13,April,2004','2004-04-13 00:00:00','   Hi, I\'m Karl Sampang. This BlogSite is Under Construction. Check Again in like, 10 years or so. Something Actually might happen to my life.  '),(13110,3139191,'26,May,2004','2004-05-26 00:00:00','   karlsampang.friendtest.com  is now online... not that you\'d care  '),(13111,3139191,'22,May,2004','2004-05-22 00:00:00','   it\'s been quite a long while since i last sat in-front of a PC monitor..  very long..  sometimes, i wonderif anyone actually reads this shit.. sometimes.  '),(13112,3139191,'12,May,2004','2004-05-12 00:00:00','   leyte...  i have been in leyte these past ten days. recapping my fucked up na screwed up na buhay. thank you, thet.  '),(13113,3139191,'25,June,2004','2004-06-25 00:00:00','   wow... astig..  kakatapos lang ng motorcade...  so deretso ako sa lib... and voila!  LINUX OS na computers?!  the best talaga ang UP...  so kakatapos nga lang ng motorcade... it rained... pero, that didn\'t dampen our bloc\'s spirit...  okay... shet ang galing... may functioning computer sa lib... sa SSC wala..  '),(13114,3139191,'23,June,2004','2004-06-23 00:00:00','   haha... foundation week namin ngayon.. wheeee...  kasabay ko sa miss micks sa jeep kanina... yehey... wala lang.. ang ganda niyah... sobra..  '),(13115,3139191,'21,June,2004','2004-06-21 00:00:00','   wow... someone actually reads this!  thanks paolo!  Tesa!  :)  okay..  wala pang masyadong nangyari ngayon... first time kaming nagkaroon ng classes sa PE! badminton! asteeg... masarap pala ang badminton...  daming tao!  nandoon sila lolita, jeannie, mga blocmates ko, psyche bloc B, other econ and BM students... masaya naman..  meron akong nakitang isang magandang babae... second year siya yun ang nalaman ko, pero hindi ko alam ang kanyang pangalan.. next week na lang siguro... ganda...  may nabasa ako sa internet:  \"Things happen for a reason, its just that you can\'t understand other\'s reasons for doing some things.\"  '),(13116,3139191,'18,June,2004','2004-06-18 00:00:00','   oh great...napaparagnarok na naman ako... just great...  '),(13117,3139191,'17,June,2004','2004-06-17 00:00:00','   ohh... great.. tesa sent me this message... the message was crap.. but tesa\'s intentions weren\'t... thanks tesa... thanks a lot...  lagay ko picture mo sa site koh ha?  '),(13118,3139191,'17,June,2004','2004-06-17 00:00:00','   mass freshmen orientation sa UP clark... astig... ang daming tao.. noong pumask ako, black ang polo ko, ngayon, White na! white na! ano ba yannnn???  pero okay ang orientation...  it couldve been \"more\" organized, but it was good naman..  '),(13119,3139191,'16,June,2004','2004-06-16 00:00:00','   Walang Title  Nakakaiyak isipin na kahit kailan man, hindi ako mamahalin ng babaeng tatlong taon ko nang minamahal. Ang babaeng sanhi ng daan-daang gabi ng hindi ko pagkatulog. Puyat. Pagod. DI makatulog. Puno ng adrenaline ang katawan. Mabilis ang tibok ng puso. Hindi makahinga. Walang ganang kumain. Walang maisip kundi siya. Iyan ang nararamdaman ko. Siya lang ang laman ng aking isipan. Tatlong taon iyonâ€¦ Medyo mahabahaba rinâ€¦   Isa sa mga taong malapit na kaibigan ko, hinahangaan ko, inirerespeto ko ay may gusto sa babaeng pinakamamahal ko sa buong mundo. Mukhang nagkakaigihan na sila.  Nalulungkot akong isipin na ganito ang sitwasyon ko.   Miserable. Basura. Tae. Ako.  Lahat ng nabanggit ko sa itaas, halos magkapareho lang ang meaning; Miserable, Basura, Tae, Ako.  Pero ako, isang basura, tae, miserable, ay nagmamahal sa isang tulad nya. Parang lupa at langit. Siyempre, ako yung lupa, mukhang lupa, amoy lupa. Lintek. Basura.  Wowâ€¦ Napaiyak ako kaninang nagbabalat ng sayote. Hindi dahil katabi kong nagbabalat ng sibuyas ang kapatid ko (pero factor na rin yun), kundi dahil inisip ko na kahit ayaw NYA (si kwan), at di nya ako type, at may kasama siyang iba (osmeone else) at kahit inaakbayan siya ng isa sa pinakasusuklaman tao ko sa buong planeta (yung kaklase ko) , susubukan ko pa ring iparating ang nararamdaman ko. Dahil alam ko na pag hindi ko ginawa ito, bubulabugin ako ng kunsensya ko habang nabubuhay ako. Mahal ko siya, at kahit sabihin niya na ayaw nya sa akin, â€œthis timeâ€ okay lang baka ikasiya ko pa nga, dahil BAKA sakaling makalimutan ko siya, maalis siya sa background ng isip ko, kahit pansamantala lang. Pero siyempre, pipiliin at mas gugustuhin ko rin na sagutin niya ako, na magkasama kami, dahil, iyon naman talaga ang pangarap ko eh, kahit masaklap, gagawin ko pa rin, wala naming mawawala sa akin. Mahal ko naman siya.  Dahil na rin sa iniisip ko na isinayaw ko siya nung prom at minsan ding nakipag-kantahan sa kanya sa attic ni Paolo, mga masasayang alaala, mga alaalang di mabubura sa aking isipan kahit mabulok na ko si lilalim ng lupa.  Siguro, para sa iyoâ€™s walang kwenta ang pinagsusulat koâ€¦  Pero, umaasa akong maiintindihan mo dahil ito ang effect pag naiinlove ka sa taong alam mong malabong magkagusto sa iyo. Nasisira ang ulo ko(see: di makatulog, yung babae lang yung iniisip part sa bandang itaas.).   Siguro, para sa iyo, nag-sesenti lang ako ditto, nagsusulat, pero ako si Karl Sampang, kung hindi ako nakakatulog sa kaiiyak dahil sa mga problema ko, isinusulat ko lahat ito, para mailabas sa isipan ko, ito ang therapy ko.  Siguro, para sa iyo, isa lang akong tanga, na namamagasang mahalin din siya ng minamahal niya, may phrase para sa sitwasyon ko, at kung tama ang pagka-recall ko, â€œunrequited loveâ€ yata ang tawag.  Siguro, para sa iyo, isa ako sa mga lalaking torpe.  Nagkakamali ka, hindi ako torpe, may pagka-walanghiya pa nga ako eh. Natunaw yata siya sa hiya nuong 2nd year highschool kami, nung maglakad ako sa iskwela na may karatulang naglalaman ng nararamdaman ko para sa kanya. â€œI love you, Thetâ€¦â€ pa nga ang nakasulat, kung â€˜di ako nagkakamali. Oo, gago ako, tanga akoâ€¦ alam koâ€¦  Siguro, para sa iyo, isa lang akong bata. Hindi pa alam ang tunay na ibig sabihin ng salitang, â€œmahalâ€, â€œI love youâ€ at iba pa. Diyan ka nagkakamali.   Siguro nagsasawa ka na sa kakasabi ko ng salitang siguro. Pero alam mo, napagisip-isip ko na, siguro kung sineryoso ko lang ang buhay ko ng konti, mas mabuti ang sitwasyon ko ngayon. Gago kasi ako ehâ€¦ Tanga, Bobo, Gago, Ungas, lahat-lahatin mo na.  Kahit na anong mangyari, mahal ko siya, at hindi ko ito isinusulat lamang, itoâ€™y aking nararamdaman, kung hindi true love â€˜tong nararamdaman ko, ewan ko na lang. (Jolina?)  '),(13120,3139191,'11,June,2004','2004-06-11 00:00:00','   can you say supercalifragilisticexpicalidotious?  i can\'t..  but that doesn\'t mean i won\'t...  ive met a lot of people this past week..  dami sobra.  si pebpeb, and tesa, and lucy, and kathleen dami..  and yung mga upperclassman ko..  and yung mga former classmates ko..  irevamped the site nga pala...  check nyo rin ang friendtest ko..  under \"stuff\" named : \"kaibigan mo ba talaga ako?\"  you should have some fun..  go out....  and visit ka na lang ulit next week  karlsampang.tk  para cute.  '),(13121,3139191,'10,June,2004','2004-06-10 00:00:00','   Yehey!  karlsampang.tk  astig..  tell your friends about it..  para hindi ma-seize yung domain ko...  '),(13122,3139191,'09,June,2004','2004-06-09 00:00:00','   im writing something....  random thoughts...  ill print it out for everyone to see... mebbe later..  \"balisong\" still is in my head... but in a cool uncaring sort of way.  i dont even know what that means...  '),(13123,3139191,'08,June,2004','2004-06-08 00:00:00','   college na ako..  astig..  '),(13124,3139191,'31,July,2004','2004-07-31 00:00:00','   The first one was called â€œ[crush>â€. She was 14 years old when I first met her. Short black hair, fair complexion, five foot five, as athletic as she was intelligent (she was in the â€œstar sectionâ€), infinitely charming, and was a dead ringer for actress Pia Guanio, that was what she was to me. To me, she was the â€œend all, be allâ€ of all the women in the world, nobody else compares. That was three years ago.  I first met her in second year high school. I was stupid, and obsessed with Magicâ„¢ cards, Counter-Strikeâ„¢ and Ghost Reconâ„¢, among other things. I first met her in the grounds of our schoolâ€™s â€œso-calledâ€ Math Park. It was a place where students share assignments, test questions, notes, play cards, and chat their idle time away. And like any other day, it was crawling with these idlers, students who cared little about their studies if in fact they cared AT ALL. I was actually one of these idlers.  I was a transferee that year. I was new, two weeks old, I was fodder for the upperclassmen and I was looked upon as a bumbling ass by my fellow sophomores (of course, I didnâ€™t know that back then). And so, naturally, I was again, for the nth time, in my then 13-year lifetime, I was alone. I was used to that, and frankly, I enjoyed being alone, nobody pesters you with his or her stupid algebra questions or their personal problems, nobody bothers you, it was peaceful; it was nice.  It was at that moment that she, and a couple of her classmates came up to me. I saw her face. There were no words to express for what I felt. I wish there was, but I havenâ€™t found anything yet. They told me their names. There were three of them. I asked her to repeat hers, and she did. I still remember their names just the way they said it. â€˜I wonâ€™t forget her nameâ€™, I told myself. And I wouldnâ€™t, I didnâ€™t want to. What I didnâ€™t know was that her name would haunt me for the remainder of my highschool life.   I was trying to memorize her face. Just so I wouldnâ€™t forget her. She was so beautiful that everyone else, to me, paled in comparison. And what I kept saying to myself that day was that â€œI found herâ€¦ I found herâ€¦â€ just kept telling myself that â€˜til I got home, and the day after that, and the day after next.  Fast-forward to the day that I really, really, screwed up; fast-forward to the day that I made this â€œsuper-fatalâ€ mistake.  It was just like any other Tuesday school day. We started classes by seven, got dismissed by four. The day was very uneventful. IT was supposed to become a day that like many others would rot in the back of my mind. Forgotten, lost, never to be recovered. But it wasnâ€™t, it didnâ€™t become that. I should have done the smart thing, I should have headed straight home. But no, I didnâ€™t do that. What I did was I lingered. Like a bad perfume, I lingered. Walked around the school with my bag. BIG MISTAKE.  	The next thing I did, would haunt me for the rest of my high school adolescent life. Out ofâ€” I dunno, sheer STUPIDITY and boredom, I took a broken arm of an armchair that was lying around, took a piece of chalk and wrote on the armchair arm thingie:   â€œI LOVE YOU, [crush>â€   	But that isnâ€™t the worst part. I laughed at my handiwork, and my friend Erin, who was at the scene of the crime, nodded. Erin was always the radical sort, loved surprising people with this unconventional way of living life. I guess it rubbed on to me. But that doesnâ€™t matter right now. Then, I took my wooden â€œplacardâ€, and walked around the school campus. I saw her, she saw me, she saw the â€œplacardâ€, she read the â€œplacardâ€, and then she turned away. It was at that point that I realized how stupid I was. I dropped the â€œplacardâ€, and it landed on my foot. Very painful, but my pain wouldnâ€™t even compare to the amount of â€œshrinkâ€ bills I would have to pay for after what I did to her. Good thing we werenâ€™t in the US.  I stood there, in the hallway, frozen, dumb, stupid, whatever. One of her friends came up to me; he told me that she was crying. She was crying, I was stupid and ashamed for what I did. Not one of us was happy. Until then, I figured that loved was a really, really good thing. In my case, it isnâ€™t.  	And with that, and a heavy heart, I went home. Straight to bed, hated myself for doing that, and cried myself to sleep.  	She was great, but way out of my league, the truth of the matter is that EVERY girl that Iâ€™ve ever liked is WAYYY out of my league; I donâ€™t even know whoâ€™s in â€œmy leagueâ€.  	But for the next three years I was haunted by another two words:   â€œWhat if?â€   	What if she liked me back? What if she IS in my league? WHAT IF? I wanted to know her better. But that was impossible now, not after the stupid â€œI LOVE YOU, [crush>â€ placard mistake. No. Her friends were always around here, like bodyguards, only, they donâ€™t have guns, they had vindictive stares that turned men into stone and kept idiots like me at bay. Especially her closer friends, by the way they looked at me, I was afraid that theyâ€™d skin me alive and feed my insides to the dogs.          I still am bothered by those two words. \"What if?\" Iâ€™ve found the answer though, â€œShe doesnâ€™t like youâ€. So now, every time my gut asks me â€œWhat if?â€ and it concerns a girl, Iâ€™d tell myself, â€œShe doesnâ€™t like you.â€  '),(13125,3139191,'30,July,2004','2004-07-30 00:00:00','   â€œYou know what your problem is? Youâ€™re in love with this girl and you keep telling yourself that you arenâ€™t!â€  Those words struck me like a steel chair on Chris Benoitâ€™s back. Those words are the painful truth that has bothered me for so many years now.  Iâ€™ve been convincing myself that â€œNo, she doesnâ€™t like meâ€ and â€œNo, I donâ€™t have a chance against whoever her â€œotherâ€ suitors may be.  And then, I find myself listening to Wheatusâ€™ â€œTeenage Dirtbagâ€.  And Duncan Shiekâ€™s â€œHalf-Lifeâ€, and Third Eye Blindâ€™s â€œDeep Inside of Youâ€.  Really pathetic.   Iâ€™m just a guy, whoâ€™s in love with a girl, whoâ€™s way out of his league, and doesnâ€™t have enough guts to even try to ask her out for â€œdinner and a movieâ€. Noâ€¦ Iâ€™m less than that. Iâ€™m just a guy, whoâ€™s IQ is so incredibly high that he walks around the school carrying a huge arm chair arm that read:    â€œI LOVE YOU, [Insert name of my crush here>â€   Some people told me that it was awfully romantic. I agree with the â€œawfulâ€ part.  I thought I was in-love with her. But now, I donâ€™t even know what love is. Is it that weird feeling in your stomach you get whenever you see this one person that makes your heart jump and your butt, clench? â€˜Cause I get that all the time, and I canâ€™t find a way to distinguish love from a stomach cramp.   Old people, they say that when you fall in love, youâ€™ll feel it, in your stomach, in your whole body, they say that the moment you see that special person, your heart will skip a beat, youâ€™ll feel like there were butterflies in your stomach, youâ€™ll feel like youâ€™re floating.   I couldnâ€™t remember whether or not my heart skipped a beat, or if I felt like I was floating. But there was this one thing, whenever I was around her, just near her (I consider the radius of 5 meters near), my heart pounds on my chest like a ~, my brain stops functioning properly, I go into â€œstupid modeâ€ and I cease to receive nerve impulses from my ears.  Stupid mode? Yes, Stupid mode. During â€œstupid modeâ€, my vocabulary will be temporarily reduced to these following phrases:  Ahhhâ€¦ Hi! O! Kumusta ka na? Ok ka lang? May gagawin ka ba? / Busy ka ba? Sino kasama mo? Ehhhâ€¦ Kumusta school? Ganon? Waheheheheheâ€¦ Kumusta ka na? (yes, I really repeat thisâ€¦) Kumain ka na ba? Gusto mong ahhÂ­â€” Kumusta ka na? Wahahahahahaâ€¦ Sigeâ€¦ Ingat ka palagi ahâ€¦  Stupid... Stupid me...  '),(13126,3139191,'29,July,2004','2004-07-29 00:00:00','   I am hurting.. My head feels like a rock on my shoulders, my eyes have involuntary tears falling out of them, my nose, wet and icky and runny, my hands, utterly disoriented, i am miserable.  Have a nice day...  And oh... Check out my Tagboard...  '),(13127,3139191,'29,July,2004','2004-07-29 00:00:00','   urlLink Maddox\'s Site = www.maddox.xmission.com or just look at this links part of my navbar   This guy OWNZ! Bad humor for other people... but some people i know will like this site because it is \"THE BEST PAGE IN THE UNIVERSE\".  Greeat site.  Poch, Pao, GUys! better read this, take a look! really hilarious...  '),(13128,3139191,'28,July,2004','2004-07-28 00:00:00','     '),(13129,3139191,'28,July,2004','2004-07-28 00:00:00','   The truth hit me during Mr. E. Escultura\'s Natural Science 1 class. The truth... is that I\'m cannibalizing my life for this site... Not that there much to cannibalize, I guess there\'d be no point in stopping now... So I\'m not going to stop doing this... No... I still have a lot of \"Fuck You\" left, if you know what I mean. I\'m still going to write, still going to update, about this sad, sorry, pathetic excuse for a life.  You know what?  This is MY sad, sorry, pathetic excuse for a life. And I\'m enjoying it!  '),(13130,3139191,'28,July,2004','2004-07-28 00:00:00','   â€œSweet like candy tooo my soul, sweet you rock and sweet you roll, lost for you Iâ€˜m sooo lost for youâ€¦â€  She was like candy to my soul. Sweet she rocks and sweet she rolls. I was so lost for her. Lost in her. I even lost myself.  Strange that Iâ€™d pick a Dave Matthews song to describe what I felt for her, a song that has so much â€œimproperâ€ sexual innuendo in itâ€™s lyrics. A little green, yes, but those lines say so much about something I wouldnâ€™t even dare speak of.  She used to be the only thing in my mind. That was about two years back. She was just so greatâ€¦ the embodiment of everything Iâ€™ve ever asked for in a woman, lady, girl, you get the idea. I was so lost for her. I was so lost for her in fact that I even lost myself. Came every single school day NOT to learn algebra, or physics or biology or chemistry, but to see her face, to watch her smile.  You should see her smileâ€¦ When she smiles, she smiles with her whole bodyâ€¦ And her eyes are so sincere, not a hint of sarcasm of â€œuntruthâ€ in them.  But thereâ€™s another song. â€œBalisongâ€ by Rivermayaâ€¦  â€œNever in my life, have I been more sure, so come on up to me and close the door, nobodyâ€™s made me feel this way befooore, youâ€™re everything I wanted and moreâ€¦â€  I knowâ€¦ These lines have clichÃ© written all over them. I donâ€™t care. Sometimes, I think of her and everything just â€œfeelsâ€ better, ever when theyâ€™re not. Sheâ€™s probably everything I ever wanted but Iâ€™m stupid enough not to realize that.  â€œShe â€˜sez that love is for fools that fall behind, and Iâ€™m somewhere between, never really know a killer from a saviourâ€¦â€  Shimmer by Fuel. Wellâ€¦ I guess Iâ€™m just one of those fools that fell behindâ€¦  '),(13131,3139191,'28,July,2004','2004-07-28 00:00:00','   Minsan  Minsan nakatitig lang ako sa monitor ng PCâ€¦ Walang ginagawa. Walang magawa. Kaya kadalasaâ€™y ino-open ang file ko na may picture niya. Sabay titigâ€¦ Mangangarap na sanaâ€™y nagging â€œkamiâ€ andami sana naming pwedeng gawin, sine, tambay, nood ng TV, mag-share ng kanya-kanyang kuro-kuro at taste sa mga bagay-bagay, interes, arcade, videoke, asa pa ako... Hanggang dito lang talaga, aasa, mangangarap, maghihintay ng kahit na anong pasabi o bagay na magpapakita ng isang â€œray of hopeâ€ na nagsasabing â€œmay chance ako sa kanyaâ€â€¦ Pero wala, kaya hanggang dito na lang. Walang naiisip kundi siya.  Minsan naman, inuumpog ang ulo at hinahampas ang sarili sa pader ng aking kwarto dahil nagsisisi, dahil naaasar, dahil nagagalit at naiinggit, â€œSana man lang, nakilala ko siya ng personalâ€¦â€.  Minsan naman, magtataka kung bakit padalas at padalas na ang iyong pagkausap sa sarili habang nasa harap ng isang computer.  Minsan naman, makikinig ka sa tape ng Nirvanaâ€¦ tapos, nakangiti ka lang. Kasi mayroon kang bagay na narerealize na ni hindi mo maiisip kung nakinig ka sa kanta ng A1, Backstreet Boys at Westlife (Boy bands=sheiit).  Minsan nakahiga lang, nakatitig sa uhhâ€¦ ceiling, iniisip ang mga bagay na hindi kailan man magiging, mga bagay na hindi kailan man naging, at iniisip mo na rin ung dapat bang kinakain ang garnish ng isang potahe.  Minsan naman, magpapaikot-ikot ako sa aking kama kapag gabi hanggang mapagod at ako at makatulog para lang makahanap ng paraan para.. uh.. matulog.  Minsan naman nakatitig lang ako sa pagkain, almusal, tanghalian, merienda at hapunan, walang gana, mapapagalitan pa ako ng tatay ko. Pero pag gabi na, hahaluhugin ang ref, makakita lang ng ilang pirasong Chocnutâ„¢ at Maxx Honey Lemonâ„¢ na pinapapak habang nakatitig sa PC monitor at pinagmamasdan ang patalbogtalbog na bolang screensaver na makulay na di-maiwasang maihalintulad sa patalbog-talbog ngunit walang kalatoy-latoy kong buhay.  Walang ganang kumain, pero may ganang mang-raid ng ref habang tulog pa ang buong bahay.  Minsan naman, maglalakad lang sa kalye, walang pakialam kung sinong mabangga.. Parang may sarili kang mundo.. Maibabalik mo lamang ang iyong sarili pag nahalata mo na na wala na ang wallet mo.  Minsan naman sa klase, habang recitation, nakatitig lang sa isang butas ng pako sa itaas ng blackboard na marahil binutas noong panahon pa ni Marcos habang pilit na inaalis sa utak ang mukha ng taong nagsisilbing dahilan kung bakit ka pumapasok sa iskwela. Pinapaalala sa iyong sarili na kaya ka pumapasok ay hindi para Makita ang iyong hinahangaan ngunit para matuto kung papaano kunin ang equilibrium constant ng isang reaction sa isang tren na papuntang City A na nakikipag-karera sa isang Kotse na tumatakbo ng 60 Km/Hrâ€¦ Kung ganoon nga, ilan ang pasahero sa tren. Tapos, bigla ka na lamang tatawagin ng teacher, itatanong kung ilang sonnet ang isinulat ni Shakespeare. Patay kaâ€¦ Hindi kasi nakikinig, samantalang kakasabi niya lamang.  Minsan naman maglalakad ka papauwi, bago tumawid ng kalye para sumakay ng dyip muntik kang masagasaan ng isang 100-wheeler truck na may dalang buhangin, pag tawid mo, maglalakad ka lamang, derederetso, parang may sariling mundo, di na bale kung masagasaan ako ng bus ng Philippine Rabbitâ„¢ o mahagip ng sportscar ng isang conio boy, basta nakatatak lang siya sa isipan mo.  Minsan mapagkakamalan ka pang nagwa-wan-two-tree sa dyip dahil sa  sobrang tahimik mo at hindi ka napansing magbayad ng driver, doble pamasahe ka paâ€¦ patay ka pag ganyan. O kaya namaâ€™y sosobra ka sa iyong pagbababaan, kailangan pang maglakad ng napakalayo.  Minsan naman mananatili ka ng isang oras sa isang intersection dahil sa buhol-buhol na traffic. Late ka na naman sa Nat.Sci.1 mo. Patay ka.  Minsan naman kundi kadalasaâ€™y, uupo ka sa harap ng computer at bara-barang itatayp ang mga bagay na nararamdaman mo. Wala kang pakialam, kung sinong makabasa, maging magulang mo pa o crush, o ang lola mo, o ang aso, walang pakialam dahil ito ang iyong nararamdamanâ€¦  Uninvited  For now, I would like to believe that I no longer think of her, that I no longer care or her, that I no longer feel anything for her; that Iâ€™m over her.  But thatâ€™s not true now, is it?  And then Ida told me that she would be going over to Paoloâ€™s with Anneâ€¦ I wanted to go with themâ€¦ Not because I wanted to see Ida, or Anne, or Paolo, or Paoloâ€™s attic, or someone who is not in any way related to me by blood, it was because she told me that Thet would be thereâ€¦ And I wanted to see her. Badly.  Why I wanted to see her? I dunnoâ€¦ I just wanted to. I just really, really wanted to. And I wanted to talk with her evenâ€¦ I wish I could do that.  Because I want to. But I donâ€™t know what else might happen, or who else will drop by uninvited.  But I wanted to come.  But I couldnâ€™t, I just couldnâ€™t.  '),(13132,3139191,'27,July,2004','2004-07-27 00:00:00','   Pretty (daw)  wahehe...          Luci           Tesa           Peb Peb    Pretty, aren\'t they?  yeah... i know you\'re green with envy now...  Okay.. im a stop now.  '),(13133,3139191,'26,July,2004','2004-07-26 00:00:00','   a thief by kimsampang  i am not a poet, am nothing but a thief, who steals the words, from poet\'s mouths, and make thier children weep,  i am not an artist, am nothing but a thief, i mar the canvass, with my strokes, of paint and coal, and dirt and cheep.  i am not a writer, am nothing but a thief, whose words of truth, are nothing more, than lies i write, i sell to eat.  i am not a saviour, am nothing but a thief, who steals your gold, by way of faith, and lies to you, for heaven\'s sake.  I am not a good man,  am nothing  but a thief, who writes some ryhmes, and draws some lines, and speaks of faith, just for me to eat.  '),(13134,3139191,'26,July,2004','2004-07-26 00:00:00','   Sees Through Me by: Karl Sampang  She sees through me, like cat\'s eyes, her piercing eyes, disarming smile, makes me believe, that i\'m alive, im lost in her, so lost in her, ive lost myself, second time this week, she sees through me, those eyes of dark, see my soul, though im without, i have no money, im really broke, can\'t buy no food, no water,im stoked, can\'t buy her flowers, can\'t buy her sweets, can\'t buy her heart, won\'t win that soon, am i making sense? i think not. so why are you still reading this? now, bugger off...  '),(13135,3139191,'23,July,2004','2004-07-23 00:00:00','   This is MY PIECE OF REALITY.         This whole blog.  Is evidence that I\'m real.  That I exist.  That I am not a figment of someone else\'s sick imagination.  This whole blog convinces me that I am real.  That I, with my cheap, worn-out, knock-off clothes exist.  This is proof.  This is MY REALITY.  '),(13136,3139191,'23,July,2004','2004-07-23 00:00:00','   Meow...  Uneventful... i\'m really looking forward to this weekend.  the upepp freshmen night is just around the corner... all the pretty girls are talking about NOT GOING.. so yeah... i\'m having second thoughts about that...  Really really sad about that..  what\'s the cat all about? wala lang... it looks nice...  get a load of this one...                                     DARTH-MOTHER (DARTH VADER\'s MOM) Really...   Now that i\'ve put that aside...  this is nice... finally, a break from my romantic fantasies...  '),(13137,3139191,'22,July,2004','2004-07-22 00:00:00','   \"Beauty queen of only eighteen/ she had trouble with herseeelf/ he was always there to help her/ but she/ always belonged to someone else...\"  She was a beauty queen (well, she could be...), of almost eighteen, i don\'t think she ever had trouble with herself. I\'ll be always there to help her. But i figure she\'s way out of my league. AS always.  \"The Second One\" was named \"Q\". Q, for unknown. She was one of the most popular girls in school. Teh fashionista kind, the RICH fashionista kind, can easily be mistaken for an ice queen, yes. I did, but behind all that \"ohhh-so-expensive\" clothes and those dreamy eyes lies a princess like no other.(what in God\'s green earth am i saying...)  dreamy eyes.  disarming smile.  unfathomable charm.  how the hell did i mistake her for an \"ice queen\".  i have no idea.  although she has \"POP\" written all over her... chuck taylor sneaks, classic blue jeans, lots of jangling things stuck to her wrists... no, so much as i HATE POP, i also love her...  ahh.. great, she walked past me just a few seconds ago, and i wonder if she realizes that i\'m actually writing about her right this second...  i am in the library, using the \"official school computer\" powered by linux... yaeey linux...  she\'s sitting right beside me right now.  and my heart is pounding like hell..  i can\'t feel my chest anymore.. gotta stop.. she might see this...  '),(13138,3139191,'20,July,2004','2004-07-20 00:00:00','   Revamping and fixing my blog, putting more stuff on my navbar thingie... Too lazy to move and do things that i ought to be doing, like studying and the like... My life is really coming back to what it was a couple of years ago... Just a crappy shadow of everybody else\'s. I\'ve tried \"rebuilding\" my life just a couple of months ago, and for a time, i was \"happy\". I guess I got tired of all the hating and the \"classifying of people as if they were music genres\". Read Karina Pomaneg\'s Frontpage column just yesterday. Good luck with that drummer boy Karina.  Which reminds me of this significant other that i oft dreamed about back in highschool. She was a drummer in thier band, \"A\" student, volleyball player, former student council member, deadringer for Pia Guanio... What more could I ask, she was the \"end all, be all\" of all the women in the world to me... That was three years ago... She still is great... I mean, just plain... \"great.  I\'m Gonna write about her one of these days. (Maybe later) I have a copy of here picture, but i figured that i shouldn\'t show it for all the wordld to see, she\'ll hate me even more.  Still fantacizing about becoming a lead vocals/rythm guitarist of my own band. We\'d be called \"The Great Depression\". Angsty, hateful, unconventional, bordering on gothic, but more on grunge. Yep! Just like Nirvana. (God, what am I talking about?)  '),(13139,3139191,'19,July,2004','2004-07-19 00:00:00','   Just finished reading \"The Sicilian\" by Mario Puzo...  Tragic... Beautifully tragic, if there ever was a description like that...  '),(13140,3139191,'16,July,2004','2004-07-16 00:00:00','   Prelims are over... Man, I\'m glad they are... I did \"nice\" on my History 1... But pretty nervous about EnglishÂ 1 and Math 11... Really nervous... Got a copy of Mario Puzo\'s the Sicilian from the school library; Still looking for Machiavelli\'s \"The Prince\" though...  Â   Utada Hikaru is sooo cute... And really talented too!!!   Â      Â   Â   Ohhh... I saw this while surfing the web:   Â         Â   Â      Weird...'),(13141,3139191,'14,July,2004','2004-07-14 00:00:00','     '),(13142,3139191,'13,July,2004','2004-07-13 00:00:00','   Ang book of life ko. Na pinagawa sa akin ni Mrs. Pangilinan, ang English teacher ko.  Marahil, isa sa pinakaimportanteng proyekto na tinapos at ipinasa ko.  Kung sinu-sino na ang nakahiram ng book of life ko. Ewan ko na kung nakaninoâ€¦  Kasama sa .doc na ito ang ilan sa piling artikulo ng librong iyonâ€¦  Sino ba naming nag-akala na maiinspire talaga ako na idocument ang aking mga naiisip/saloobin,/trip/kasalanan/buhayâ€¦ Well, ako actually.  Kahit na sobrang Daming Grammatikal Erors. Ewan. Sabi nga ni Karl Sampang, â€œI write the way I speakâ€.  Ows? Talaga?  '),(13143,3139191,'12,July,2004','2004-07-12 00:00:00','   test tomorrow...  an essay test! wheeeeee!  but really... i\'m going to learn html... really serius about this.. and im going to postpone my dreams of ragnarok online stardom... for now..  '),(13144,3139191,'12,July,2004','2004-07-12 00:00:00','   haha... the pain..  three major tests. one whole week. need to report on conjuctions next week... should have learned about that in highschool... but nooo.... ramayana.... damn...  '),(13145,3139191,'09,July,2004','2004-07-09 00:00:00','   Hindi na ako isang Highschool studentâ€¦ Tapos na ang mga araw na kailangan kong mag-uniform para lamang makapasok sa iskwelaâ€¦ Tapos na ang pagbili ko ng fishball at squidball sa mga overpricing na tindera. Tapos na ang pagsabay ko sa maraming tao tuwing pumapasok para hindi niya mapansin na wala akong I.D. Tapos na ang pag-â€œspoon feedâ€ sa akin ng mga teacher ko. Tapos na ang paglabas ko sa campus tuwing lunch ng madalas hindi para bumili ng pagkain ngunit para mag-internet at mag-laro ng computer games. Tapos na ang mga araw na may definite seating arrangement sa klase. Alam kong wala akong mamimiss sa mga bagay na nabanggit ko, peroâ€¦ Tapos na ang mga araw na nagsasaya ako, pa-easy-easy lang, hindi nag-aaral, ngayon kailangan nang magseryoso. Dahil tapos na ang mga araw na highschool ako. Estudyante na ako ng kolehiyo. Wala nang tutulong sa akin, mag-isa na lang ako ulitâ€¦ Parang tulad ng dati? Tekaâ€¦ Parang highschool lang ahâ€¦ Kaso kailangan na talagang mag-aral, at ang mga teacher dito, mukhang kumukuha sila ng kasiyahan sa pag bigay ng bagsak na marka sa mga estudyateng tinuturuan nila. Ito siguro ang tunay na mundo, hindi masaya, seryoso, at ngayoâ€™y pinapatunayan ko ang isa sa mga teoryang aking binuo habang nagdidiscuss tungkol sa kabutihan ng tao ang Values Ed. Teacher ko. â€œAng tao ay likas na masama. Ang Tao ay Hedonistic, iniiwasan ang sakit at lagging humahanap ng paraan para mapasaya, mapahusay ang buhay niya. Ang tao ay hindi likas na mabuti, dahil kung totoo iyon, wala tayong dapat nakikita na pinapatay, o ginagahasa, o nininikawan, o dinudukotâ€¦ Tayo ang salot sa mundong itoâ€¦ Parang napaka-corny samantalang sa atin ipinagkatiwala ang mundong ito. Tapos na ang lahat ng gusto kong isulat ukol sa parting ito ng buhay ko. Hindi na ako Highschoolâ€¦ At sa ikaapat na araw ng pagiging estudyante ng kolehiyo ko, napagisip-isip ko na mag-isa na naman muli akoâ€¦  Tulad ng datiâ€¦  '),(13146,3139191,'01,July,2004','2004-07-01 00:00:00','   hmmm... pathetic...  three days of pure boredom..  tatlong araw! you know what that can do to a man?!  '),(13147,3139191,'06,August,2004','2004-08-06 00:00:00','   Oh joy!  Oh pure sarcasm!  Oh SHIT!  This is just one of those days. Typical days when i just one to kick the living crap out of my pretentious classmates.  I hate them so much that i wrote something for them.  A prayer for my pretentious classmates.  Dear God, please forgive me for my sins, please help me forgive others for thier sins, please forgive me for planning to murder my pretentious classmates, please help me not to commit that sin again, please help me control myself from KICKING THIER OHH-SO-PRETENTIOUS ASSES to the fiery depths of hell...  AMEN  p.s.  If i am not able to stop myself, please please please save thier souls... they need it.  '),(13148,3139191,'05,August,2004','2004-08-05 00:00:00','   1. WHAT IS YOUR MIDDLE NAME? - Alias  2. WHAT kind of PANTS ARE YoU WEARING?  - Im naked.  3. WHAT ARE YOU LISTENING TO RIGHT NOW? - MY pretentious classmates\' voice singing Manilow\'s \"mandy\"  over and over again... is defiling the song... note DEFILE!  4. WHAT ARE THE LAST 4 DIGITS OF YOUR PHONE? - 5547  5. WHAT WAS THE LAST THING YOU ATE? - Juicy fruit and soda  6. IF YOU WERE A CRAYON, WHAT COLOR WOULD YOU BE? - Black, blue.  7. HOW IS THE WEATHER RIGHT NOW? - partly cloudy with scattered rainshowers  8. LAST PERSON YOU TALKED TO ON THE PHONE? - operator services  9. THE FIRST THING YOU NOTICE ABOUT THE OPPOSITE  - face, complexion...  10. DO YOU LIKE THE PERSON WHO SENT YOU THIS? - who?  11. HOW ARE YOU TODAY? - i am quite satisfied, i controlled myself from STRANGLING MY, guess what, PRETENTIOUS CLASSMATES!  12. FAVORITE DRINK? - extra joss, with water. or just water.  13. FAVORITE ALCOHOLIC DRINK? - VODKA... parang kampit kung gumuhit.  14. FAVORITE SPORTS? - [insert something here]ball  15. HAIR COLOR? - black  16. EYE COLOR? - brown.. yellow from hepatitis  17. DO YOU WEAR CONTACTS? -no  18. SIBLINGS AND THEIR AGES? - i forgot.  19. FAVORITE MONTH? - January, cold.  20. FAVORITE FOOD? - edible  21. LAST MOVIE YOU WATCHED? -THE BOURNE SUPREMACY  22. FAVORITE DAY Of ANY YEAR? - the day after tommorow.  23. ARE YOU TOO SHY TO ASK SOMEONE OUT? - im never shy, im just stupid...  24. DO YOU LIKE SCARY MOVIES OR HAPPY ENDING? - scary happy endings  25. SUMMER OR WINTER? - winter  26. HUGS OR KISSES? - or getting laid?! no... hugs..  27. CHOCOLATE OR VANILLA? - chocolate 28. DO YOU WANT YOUR FRIENDS TO WRITE BACK? to what?!  '),(13149,3139191,'05,August,2004','2004-08-05 00:00:00','     '),(13150,3139191,'04,August,2004','2004-08-04 00:00:00','   1. NAME? >> Karl  2. SINGLE OR ATTACHED? >> Single  3. ARE YOU IN LOVE? >> I don\'t know the difference between love and a stomach cramp.  4. DOES SOMEONE HAVE A CRUSH ON  YOU? >> i don\'t think so..  5. HAVE YOU EVER BEEN HEARTBROKEN? >> constantly  6. HOW MANY EX(ES)? >> is zero a number?  7. HOW LONG WAS YOUR LONGEST RELATIONSHIP? >> 16 years. a love love relationship between me and my parents  8. HOW SHORT WAS UR SHORTEST RELATIONSHIP? >> 2 1/2 hours, there was this girl once na nakasakay ko sa bus..  9. HAVE YOU CHEATED ON SOMEONE BEFORE? >> This is such a \"relationship centered survey...\"  10. HAVE YOU BEEN CHEATED / TWO- TIMED ON BEFORE? >> my patience is running thin...  11. IF YOU ARE SINGLE, DO YOU WANT TO FALL IN LOVE AND GET ATTACHED? >> no... i have too many preoccupations already...  12. IF YOU ARE ATTACHED, DO YOU LOVE YOUR GIRLFRIEND / BOYFRIEND? >> but i am not.  13. HAVE YOU EVER PLAYED WITH SOMEONE`S FEELINGS? >> what are feelings?  14. HAVE YOU EVER BEEN FORCED/DARED TO GO WITH SOMEONE? >> no...   15. HAVE YOU CRIED OVER SOMEONE YOU THOUGHT YOU LOVED/LIKED? >> 365 days a year.  16. HUGS OR KISSES? >> words.  17. DO YOU THINK IT IS ALRIGHT FOR PEOPLE OF THE SAME SEX TO LOVE EACH OTHER? >> sure why not... it\'s a free country..  18. HAS ANYONE EVER TOLD YOU THAT  YOU WERE 2 YOUNG FOR LOVE? >> i think that.  19. WHAT DO YOU THINK OF LONG  DISTANCE RELATIONSHIPS? >> no chance of getting laid whatsoever...  20. WOULD YOU EVER GO WITH SOMEONE OVERSEAS? >> SDGSDFGDFGSDFGDFGSD  21. HAVE YOU THOUGHT YOU LOVED SOMEONE BUT THE FEELING WENT AWAY AFTER A FEW DAYS? >> i takes two days.  22. DO LOOKS MATTER TO YOU? >> yeah.  23. WHAT IS THE MOST IMPORTANT THING  IN A RELATIONSHIP? >> sex drugs and rock and roll..  24. FOR GIRLS: HOW DO YOU FEEL IF  YOUR BOYFRIEND IS YOUNGER THAN YOU? >> im going to ignore this.  FOR GUYS: HOW DO YOU FEEL IF YOUR GIRLFRIEND IS OLDER THAN YOU? >> BRING IT ON.  25. FOR GIRLS: HOW DO YOU FEEL IF  YOUR BOYFRIEND IS SHORTER THAN YOU? >>ignore.  26. for GUYS: HOW DO YOU FEEL IF YOUR GIRLFRIEND IS TALLER THAN YOU? >> are we talkign about an inch or a foot?  26. DO YOU TELL YOUR CRUSHES YOU  LIKE HIM / HER? >> yep... in a weird, idiotic, stalker-like manner.  27. ARE YOU ROMANTIC? >> i am an idiot.  28. WHAT DO YOU THINK OF ONE NIGHT STANDS? >> im going to quote VAn Wilder: \"Don\'t be a fool, protect your tool.\"  29. WHAT DO YOU THINK OF COUPLES  WHO BREAK UP AND PATCH BACK AFTER A FEW DAYS? >>can\'t relate  '),(13151,3139191,'04,August,2004','2004-08-04 00:00:00','   Still no word from Syeriâ€¦ Iâ€™m beginning to doubt that she will be returning my jacket sometime this week. Gave her my number but I just really canâ€™t see myself answering the phone with her voice on the other side of the lineâ€¦.  Really pathetic.  Iâ€™m listening to Nirvana right now, doesnâ€™t matter which songâ€¦  Ahh something popped into my head just a few seconds ago.  I met a guy a few days back.  The turd talked endlessly about â€œFridaâ€ and â€œY tu mama tambienâ€ and Tarantino and Coppola, and Lee like he was some kind of film critic ass. He was carrying with him this tall tumbler of note, STARBUCKS; he drankthe foamy starbucks latte thing inside the starbucks tumbler thing while rambling about his critiques. Like I caredâ€¦  Iâ€™m sorry, I donâ€™t know much about coffee except the fact that it contains caffeine. And I donâ€™t get people for buying coffee at Starbucks and Seattleâ€™s Best or Coffee Overdose for that matter. Itâ€™s probably a cool, rich people thing. I intend to rant about that in a separate entry, watch out for that.  Okay, so there is this guy with a tumbler of whatever they serve at Starbucks, acting like he was some film critic ass. Note the way that I use the word ASS. Thatâ€™s what he is, just some ass. There are some things I would like to cite, and might support my theory that he is a poser.  1.	I saw him in the Angeles city proper. That would mean that the closest Starbucks would be 17 kilometers away from where we were. The closest Starbucks outlet that I can think of is the one along the north expressway, and if there is one in SM Pampanga or Robinsons, which is still 17 kilometers away.  2.	I saw him at 9:15 am. Itâ€™s either he brought his coffee along with him from the expressway for a 50 minute ride, or he traveled at the speed of light to bend the time-spaceakljsfbailsdgfwjklbfd[SCIENTIFIC GIBBERISH HERE], to buy a â€œtall glass of whatever the hell flavor of Starbucks coffeeâ€ and drink it at a cheap-ass mall (NEPO MALL) just to show off.  3.	Or he made himself some coffee and put it in a Starbucks tumbler which he values so much and washes with dishwashing liquid in order to reuse it.  4.	Four? Oh sorry, I already puked because he has POSER written all over him.  5.	Like I didnâ€™t notice his knock-off Girbaud sneakers, or his knock-off â€œpornstar/Am Blvdâ€ T-shirt.  6.	He was talking about how the said movies were so great expressing so much meaning. He also said that the â€œMatrix: Reloadedâ€ and the â€œMatrix: Revolutionsâ€ had so much meaning and deep philosophical content and tackled causality. Really?! Can you tell me what those movies meant?! Do you know what causality and existentialism means? Huh?! I didnâ€™t think so. Dumbass!  I therefore conclude that this idiot is a POSER.  Now DIE!  (stab) (stab) (stab) (stab) (stab) (stab) (stab) (stab) (stab) (stab) (stab) (stab) (stab) (stab) (stab) (stab) (stab) (stab) (stab) (stab) (stab) (stab) (stab) (stab)  Idiotâ€¦  '),(13152,3139191,'03,August,2004','2004-08-03 00:00:00','   1. What is your complete name? Does it have a special meaning?  > Karl Isaac Mari A. Sampang. Yes, I think. Karl, maybe that was for â€œKarl Marxâ€, I think my grandfather really approved of that. Isaac, â€œIsagani from El Filibusterismoâ€? Mari, I am offered to mother Mary. â€œAâ€ stands for Alias, my momâ€™s maiden name. And Sampang.  2.If your life had a soundtrack, what would be the songs and how would they be arranged?  >These are the songs, in autobiographical order:  About a Boy â€“ Badly Drawn Boy  214 â€“ Rivermaya Torpe â€“ Hungry Young Poets  Be Like That â€“ Three Doors Down There She Goes â€“ Sixpence none the Richer Drops of Jupiter â€“ Train Half-life - Duncan Shiek Sleep all Day - Jason Mraz Im on a High - Duncan Shiek Lonely - Rhett Miller  Fast Car â€“ Tracy Chapman Everywhere â€“ Michelle Branch Pare Ko â€“ Eraserheads Tindahan ni Aling Nena â€“ Eraserheads Bizzare Love Triangle â€“ Frente Sway â€“ Bic Runga  Yellow â€“ Coldplay Blinded â€“ Third Eye Blind Deep inside of you- Third Eye Blind The Scientist â€“ Coldplay Androgyny â€“ Garbage Itâ€™s Over Now â€“ Neve How You Remind Me â€“ Nickelback Crash into Me â€“ Dave Matthews Band Shimmer â€“ Fuel Gising Na â€“ Parokya ni Edgar Sampip â€“ Parokya ni Edgar Drive â€“ Incubus Wish You Were Here â€“ Incubus  Balisong â€“ Rivermaya Burnout â€“ Sugarfree I Disappear â€“ Metallica  Here without You â€“ Three Doors Down She will be Loved â€“ Maroon5 Clubbed to Death â€“ Kurayamino mix â€“ By someone I donâ€™t know  But if I were asked to pick only one song to describe my whole life, it would be:  You know youâ€™re right - Nirvana  This masterlist is subject to change, I know Iâ€™m forgetting a lot of other great songs(but if they were great, I wouldnâ€™t forget them, wonâ€™t I?). Songs will still be added.  3. DO you have a girlfriend? Why or why not?  > NO. Because of the following reasons:  They take too much of your time. The Hatid-Sundo Service. And, need I remind you of the PARENTS, most especially, HER DAD, the guy who would gladly rip your ass apart with his [insert anything he uses in his job here, i.e. monkey wrench, calculator, m16 rifle, shotgun, pencil].  They cost too much; you have to pay for their fare when you two go out, you have to pay for what you eat, and for everything else you might need money for, and you also have to buy the presents to assure them of your unwavering love.  Iâ€™m not a chauvinist, so please donâ€™t mistake me for one. Iâ€™m just saying what every other man has in the back of his mind. These are the facts. These are the superficial facts.  As for why I donâ€™t have a girlfriend, this would explain everything, this and the fact that Iâ€™m an idiot.  4. If there was one person in the world, living or dead, you could spend an hour with, who would it be? What will you do to him or her?  > God/Supreme being, I have many questions for Him.  In the sad event (read: sad, because NO GOD is a BAD THING), that there is no GOD, I would very much like to meet Jesus, he is the next best thing, I will ask him the questions that I would have asked GOD.  In the SAD SAD(read: SAD SAD) event that Jesus does not exist and the whole of Christianity is just a hoax to control the populace(which is a really sad thing), I would very much like to meet Ferdinand Magellan and do the following things to him in this exact order:  1. beat the living shit out of him, with my fists 2. beat the living shit out of him, with a 2x2 wooden stick 3. beat the living shit out of him, with a steel chair 4. cut off his fingers one by one 5. scrub all wounds with steel wool drenched in alcohol 6. get some gas and some matches and roast him until golden brown 7. poke his eyes out. 8. rip apart his abdomen and feed his insides to the dog 9. all of this beating is will be done while I pee/shit/spit on him and while insulting him.  I have that much respect for you, FM.  5. Name some people who you really think is intelligent/wise/whatever.  a.	MY DAD. NO EXPLANATION NEEDED.  These people are intelligent in a different level, in my opinion.  a.	Jefferson â€“ His reason is veryâ€¦ reasonableâ€¦ Napabayaan lang. b.	Marthyn â€“ Matalino, napabayaan langâ€¦ c.	Erin â€“ Unconventional but slightly reasonable reasoning ability. d.	Eric â€“ Meaningful, yet utterly meaninglessâ€¦ Napabayaan lang e.	Pocholo â€“ I amâ€¦ at a loss of wordsâ€¦ Napabayaan lang. f.	Jemmabel Viray and Laurence Gueco â€“ The young and the wasted.  6. If there was one word in the dictionary that could describe you, what would it be?  >â€œUndescribableâ€   7.What are you doing at this very moment?  >I am a multitasker. These are the things that Iâ€™am currently doing.  -	beating the shit out of Magellan -	listening to nirvana -	reading Shakespeare -	answering this survey -	not having a girlfriend -	being broke -	trying to sew Magellanâ€™s head back on -	and oh, beating the shit out of the guy who wrote this survey  '),(13153,3139191,'02,August,2004','2004-08-02 00:00:00','   Ext. Sa labas ng campus ng College. Makulimlim, medyo basa ang lupa, hindi pa gaanong maingay, alas nuebe ng umaga.  KARL: (bulong sa sarili) First day of classes, Karl, donâ€™t screw up. (paulit ulit)  JAMES: O! Karl! Kumusta?! (nakangiti)  KARL: Okay langâ€¦ (mapapatingin sa paligid at makakapansin siya ng isang lalaking naka itim na jacket, asul na maong at sneakers)Sina Chino?  JAMES: May pasok na yataâ€¦ Ewan ko. Sige, alis muna ako ah? May pasok na rin niyan akoâ€¦ (tatakbong papalayo, papasok na sa lobby ng school)  KARL: Sigeâ€¦ Ingat. (maglalakad patungong lobby ng school)  Int. Sa lobby ng campus, nakaabang ang mga student council people para tumulong sa mga freshmen at mga transferee.  SOPHOMORE: (lalapit kay KARL) Course mo?  KARL: (nagulat, nagtataka kung bakit  nilapit siya ng estrangherong ito) Ahâ€¦ Psychâ€¦  SOPHOMORE: Oh okayâ€¦ Sama ka sa akin?  KARL: (nagaalinlangan) Bakit?  SOPHOMORE: Walk with me. (habang naglalakad na papalayo sa lobby at papasok sa isa sa mga hall)  KARL: (susunod)  SOPHOMORE: (babagalan ng konti ang lakad para makahabol si KARL, magsisindi ng isang stick ng sigarilyo, tuloy pa rin ang lakad) New blood?  KARL: (nagulat sa pag-ingles) Yepâ€¦  SOPHOMORE: You smoke? (sabay abot ng kaha ng yosi) kuha ka.  KARL: I donâ€™t smoke.  SOPHOMORE: You donâ€™t know how to.  KARL: Itâ€™s not thatâ€¦ Itâ€™s just that I donâ€™t want to.  SOPHOMORE: Donâ€™t worry, thatâ€™s what they all say, next time you see them, theyâ€™d be puffing smoke like a train. (exhales a cloud of smoke)  KARL: Not me.  SOPHOMORE: Not yet. (smiles)  '),(13154,3139191,'02,August,2004','2004-08-02 00:00:00','   Sa isang kalsada sa kalagitnaan ng siyudad ng Angeles (Malayo sa â€œAreaâ€), may isang tineydyer na may labinganim na taong gulang na ang naglalakad.  Ang tineydyer ay nakasuot ng itim na jacket, nakasalaming may makapal na frame na kulay itim, may itim na buhok, nakamaong pants at sneakers. Siya si Karl.  â€œPutchaâ€¦ bakit pa kasi nahulog yung singkwenta sentimos ehâ€¦ AMPâ€¦â€ ang isip ng kawawang tineydyer.  Nagsimulang bumuhos ang ulan galing sa madilim na kalangitan.  â€œBuwisetâ€¦ ano â€˜to? Nawala pa naman din ang payong koâ€¦ Shet, shet, shet, shet, shet, shetâ€¦â€, bulong niya sa sarili habang patalon talon at umiiwas sa mga malalim na lubak ng kalsada na napuno ng tubig-ulan.Lalong lumakas ang ulan at napilitang sumilong sa isang waiting shed na may nakasulat na malaking letrang kulay pula : â€œBawal Umihe Ditoâ€ si Karl.  â€œMay silbe pala itong mga pinapagawa ni Lazatin noâ€¦ Walanghiyaâ€¦ Ang yaman na siguro ng bulugan na iyonâ€¦â€  Kakatapos pa lang ng eleksyon at marami pa ring nakadikit na poster ng mga kandidato, ang iba ay may drowing na ng sungay at bigote sa mukha..  â€œBagay pala ni Jamby ang may sungay..â€, isip ni Karl habang pinakatitigan ang litrato ng kandidato.. â€œHuussâ€¦ kung hindi lang kay Juday, lampaso ka na namanâ€¦â€  Lumakas pa lalo ang ulan at may nakisilong rin sa shed ng ating bida.  â€œUy, babaeâ€¦â€  Basang basa ang babaeng bagong silong sa shed.  â€œLamig noh?â€ sabi ni Karl, pilit na nagsisimula ng isang friendly conversation.  â€œOo nga eh..â€ Sagot naman ng babae habang nagti-text sa kanyang cellphone.  â€œUy.. Shet, chinitaâ€¦ ang puti, cute paâ€¦â€ isip ni Karl habang pilit na ikinukubli ang kanyang ngiti.  â€œWala ka rin bang payong?â€ hirit pa ulit ni Karl.  â€œAh? Wala ehâ€¦ Basang basa nga ako eh.. Nakakaasaaaarâ€¦â€, angal ng cute na babae. â€œInindian nga ako ng imi-meet ko ehâ€¦ bahala siyaâ€¦â€ naasar na tuloy nito.  â€œBoyfriend?â€ tahimik na itinanong ni Karl.. Habang nag-iisip â€œNakuuuu delikado â€˜toâ€¦ may sabit pa yataâ€¦â€  â€œHindi no!â€, ang natatawang sagot ng dalaga.  â€œSyeriâ€¦â€ tuloy pa niya.  â€œHuh?â€ nalilitong sagot ni Karl.  â€œAko nga pala si Syeri.â€  â€œAhhhâ€¦ okayâ€¦ Karl naman.â€ Ang pa-cute na sagot ng ating bida, umaasang makaiskor. Medyo humihina na ang ulan, at habang ganoon nga ay pinagmasdan ni Karl si Syeri, nakanakaw ng tingin dahil sa ibang direksyon nakatingin ang dalaga.  â€œO hetoâ€¦â€ ang bukambibig ni Karl habang inaalok ang kanyang jacket kay Syeri, â€œnakakaawa ka naman kasi ehâ€¦ para kang basing ibonâ€¦â€ sabay tawa.  â€œGanon? Sigeâ€¦ Salamat ah..â€ kanyang sabi habang isinusuot na ang jacket ni Karl.  â€œPutchaâ€¦ buti na lang pala nag long-sleeve ako, at least kahit wala na ang jacket, maporma pa rin..â€ isip ni Karl â€œLamigâ€¦ Putchaâ€¦ Makonsensiya kaâ€¦ Isoli mo jacket koâ€¦â€  â€œNumber mo?â€ tanong ni Syeri habang naghahandang pumindot ng kanyang keypad.  â€œShet 7610â€¦ PAQ! Ang yaman nitoh..â€ ang isip ni Karl â€œ09274135547 yung akinâ€¦ ikaw?â€  â€œAhhhâ€¦ sige, malalaman mo rin naman kung sakaling isoli ko pa ang jacket mo, â€˜di ba?â€  â€œGanon?â€  â€œOoâ€¦ ganonâ€¦ sige, salamat sa jacket, Karl!â€ sabay ngiti at lusob sa patuloy na humihinang ulan.  â€œAng cute niyaâ€¦ Sana isoli niya yung jacket koâ€¦ Ang mahal nonâ€¦â€  Pinagmasdan ni Karl si Siyeri na tumawid ng kalsada at biglang nagpara ng isang Ford F150..  â€œHmmâ€¦ 7210, chinita, weird tagalog accent, girbaudâ„¢ shoes, girbaudâ„¢ pants, really expensive looking beret, F150 na sundoâ€¦ This means only one theengâ€¦ Anak mayamannnnnnnnnnnnn!!â€  â€œSinong kausap mo?â€ tanong ng cigarette vendor na kakasilong lamang sa shed ng ating bida.  â€œHahahahahahahaâ€¦ May Maxxâ„¢ Honey-Lemonâ„¢ po kayo?â€  â€œMeron, tatlo dalawang pisoâ€  â€œPabili po ng six pesos..â€  â€œKuha ka. Philipâ„¢ gusto mo?â€  â€œHindi po ako nagsisigarilyo..â€  â€œMatututo ka rinâ€¦ Marlboroâ„¢?â€ sagot ng vendor.  â€œAyaw ko nga pong matuto ehâ€¦â€  â€œYan din ang sinasabi nilaâ€¦ DJ Mixâ„¢?â€  â€œTenk yu na lang...â€ sambit ni Karl sabay alis at sakay sa pinakamalapit na Checkpoint â€“ Holy na jeep na Makita niya..  â€œHOY!!! Bayad mo!!â€ ang papawalang sigaw ng vendor habang palayo na ng palayo ang jeep ng binata.Hindi na ito pinansin ni Karl, nagabot ng 4.50 piso sa drayber sabay, â€œEstudyante po. UP, University of the Philippines..â€  \"Yabang ko..\" puna niya sa sarili, habang iniisip kung may kahihinatnan ang kanyang ginagawa...  '),(13155,3139191,'02,August,2004','2004-08-02 00:00:00','   No words, can express, the way, im feeling, sad,lonely, scared shitless, although i\'m, all without, regret my, life has, been a,  living hell, it seems, that i, really really, have to, stop doing, this because, it\'s messing, my life, and screwing, my head, but i can\'t, stop now, i\'ve gone, too far, beyond what, is out, of line,  Some words,  might express,  what im,  feeling now,  screwed up,  scared, sad,  lonely, pathetic,  im what,  you can, call a,  perennial dipshit,  who hasn\'t,  got a,  freaking clue,  of what,  is really,  going on,  with my,  life and, i dont, know what,  it all,  means and,  i dont,  know what,  she means, to me,  but i,  wish i,  knew anyway.  Oh crap, im really,  lost now, my mind, is screwed, my heart, is beating rapidly, like a, big-ass drum, and whatever, some other,  things i,  really don\'t, know what, im feeling, for myself, for her, for everyone, im really,  lost don\'t, know what,  i want,  but i\'ll,  update as,  soon as,  i realize, what im,  really really, really really,  here for.  '),(13156,3961625,'27,July,2004','2004-07-27 00:00:00','   Sorry I havenÂ´t been able to post as much, Internet costs a lot more on weekdays... anyway, I will now write about the 24th.&nbsp; I slept tilÂ´ twelve, then ate lunch and dozed around and read a bit.&nbsp; Bed...&nbsp; On the 25th I woke up at nine, did about the same until 6, when we had a 2 hour long and very pleasant barbecue.&nbsp; On the 26th, Steven and I went to the little town of Odenarde.&nbsp; It has a really beautiful town hall with a polar opposite information center.&nbsp; When we got there, we soon discovered that for some reason everything interesting was shut on Mondays.&nbsp; Guess what day&nbsp;the 26th was.&nbsp;     &nbsp;   So we looked at everything from the outside.&nbsp; There were two beautiful Gothic churches, and one incredibly ugly one.&nbsp;&nbsp; Then we visited the adjoining town of Edane (or something like that)&nbsp; which was completely destroyed in the 10th Century by the Count of Wallawalla because he didnÂ´t like itÂ´s mayor or something.&nbsp;&nbsp;&nbsp; So we had fun wandering about the ruins.&nbsp; After that we spent what must have been an hour and a half trying to find a pub (in  Belgium! ).&nbsp; We eventually got ourselves into Brussels, where we finally found our pub...     &nbsp;   When we got back we had a quiet supper of delicious cheesey pasta with ham and capers (devoid of children, which was a pleasant change) and played MONOPOLY.&nbsp; I lost.&nbsp; Shit.&nbsp;     &nbsp;   Today I went into BrÃ¹gge with Nancy and her friend, who dropped me off in the central square and said weÂ´ll meet you here at 12:30 for lunch.&nbsp; So I wandered around BrÃ¹gge for 2 hours.&nbsp; It was really nice.&nbsp; After lunch (salmon salad) I walked for another 3 hrs, but I slept for about a third of it, so that doesnÂ´t matter.....     &nbsp;   ok, at my next destination, which I will be leaving for tomorrow, there is not a computer, so I will not be able to post until August the 11th.&nbsp; I will continue in My Diary.&nbsp; There is a gap of 5 days between the end of my diary (July 13)&nbsp; and the beginning of my blog (July 18)... IÂ´ll fill that in when I get back... &nbsp;  If you want to contact me by post, then please check the comments on this thread... when she sees this, sheÂ´ll put the&nbsp; address up (you will, right Mimi?).  &nbsp;   so tata till I get to see all you buggers in Wilmette!     &nbsp;  '),(13157,3961625,'23,July,2004','2004-07-23 00:00:00','   It is 12:26 a.m.&nbsp; here in Vurste, so I have to write about what I did 1 hr ago as if it were yesterday... but first, let me talk about the day before yesterday... the 22nd.&nbsp; So.. I stayed around Vurste.&nbsp; Read a little, played a little, ate a little, watched a movie, read a little, and slept very little.&nbsp; That having been said, lets get on to today... whoops! yesterday..&nbsp;&nbsp;\" everything I said about the day before yesterday, except add on leaving at 7:30 (we were meant to go at 7 but various children kept on injuring themselves) for Gent, which is currently in the middle of a festival.&nbsp; Jeez... I think only \"Taste of Chicago\" beats that crowd.&nbsp;     &nbsp;  It was beautiful.  '),(13158,3961625,'21,July,2004','2004-07-21 00:00:00','   On the 21st I woke up early (for me) in the morning and got dressed.&nbsp; I ate a breakfast of bread and Patay, then quickly got in the car while Steven drove me to Brugge.&nbsp; We had absoulutely no trouble finding the place, and&nbsp; when&nbsp; we arrived Ben was playing chess with Lotte and Mia was hitting them over the head with a big fuzzy and incredibly ugly teddy.&nbsp; Ine arrived&nbsp; soon aafter from the loo, and so I greeted her and Uncle David with hugs.&nbsp; Their caravan was very nice and I settled down to a game of chess with Ben while the adults talked.&nbsp; This game didnÂ´t last very long because Lotte brought out her equally ugly teddy and the girls took it in turns to hit us over the head.&nbsp; So we left.&nbsp; We drove quickly to Brugge Central and parked the car, where Uncle David drew a small crowd while unloading his bicycle.&nbsp; Then we set off along a crowded canal towards the center of the city.&nbsp; This took about half an hour because the schlock shops worked their magic on the two younger ones.&nbsp; Ben was fairly resilient...&nbsp; We arrived at an old church and had lunch while we waited for it to open at 1:30.&nbsp; When it did, we went inside and had a look.&nbsp; Turned out that this was the church that housed  Madonna and Child , which I admired and photographed.&nbsp; We then headed down to the canal and tooke a very touristy but quite pleasant boat tour.&nbsp; I sat right next to the \"Captain\" who looked rather apprehensively at my camera when I got in, but at the end of the trip thanked me profusely&nbsp;for only taking only one picture.&nbsp; Poor man must be driven to drink by all the clicking!&nbsp; As soon as we arrived in city center it started to rain, so we decided to climb the belfry, except Ine didnÂ´t want to (towerphobia or whatever itÂ´s called) and Uncle David couldÂ´t, so Mia, Ben and I went up by ourselves.&nbsp; This tower was only 364 steps, but the view was good, so it was fun.&nbsp; By the time we got down it was time to go so we sauntered, then walked, the ran headlong for the station.&nbsp; I managed to leap on to the moving train with a quick goodbye yell to the rest of them.&nbsp; Two stations and half an hour later, I was in Gent!&nbsp;  '),(13159,3961625,'21,July,2004','2004-07-21 00:00:00','   Yesterday I went with Steven, Nancy, and Eusebie to the little village  of Ieper, known by the French as Ypres.&nbsp; This is a place that still has 20,000 kilograms of unexploded bombs and grenades in the ground.&nbsp; It was completely destroyed by German bombing raids during WWI (and when I say completely, I mean completely.&nbsp; I have pictures of pictures)&nbsp; and was rebuilt.&nbsp; We took about&nbsp; an hour to drive there and when we got there we parked outside of the tall defensive walls, climbed over and through said walls, and finally came to a massive gate which must have listed at least 500,000 names, all of which had dissapeared and whoÂ´s bodies had never been found.&nbsp; Then we walked through the cathedral, which was beautiful, and then we ate lunch at a place called  Vivaldi  right on the main square.&nbsp; (I had \"Penne the Italian way\")&nbsp; We then went into the Flanders field museum, which is an interesting if somewhat gruesome museum about all the fighting in Flanders during WWI.&nbsp; They gave us little cards with names of soldiers on them, and throughout the exhibit, we placed the cards in little scanners, 4 in total, and it told us the story in 4 parts.&nbsp; My first guy died.&nbsp; Idiot.&nbsp; The second guy got caught and was incinerated somewhere in Western Germany.&nbsp; Oh well.&nbsp; We then went back to the car and took a quick ride to a part of the land around Ieper&nbsp; where the trenches from WWI were still intact.&nbsp; Now I know why war is hell.&nbsp; they were dark, gloomy, and very, very muddy.&nbsp; Besides, there were bomb craters all around us, and the guide seemed intent on telling us every minute or two how many unexploded bombs there were per square meter in the area around us.&nbsp; We then went back to the seaside (1 hrs. drive) and picked up the boys.&nbsp; We had a drink at a little pub, then went back home to NancyÂ´s.&nbsp; There we put the kids to bed, and Steven and I picked up some fries and sausages at the 5th Frituur we came to (all the rest were shut, as it was 9:30 at night).&nbsp; After dinner we tooke the dog on a walk and played MONOPOLY.&nbsp; I won.  '),(13160,3961625,'20,July,2004','2004-07-20 00:00:00','   Today (or as it is, yesterday), I went into Gent with Nancy, Steven, and Eusebie.  It was a lot of fun.  We parked under the same bridge we met Nancy under 2 years ago, walked a short distance to the tram stop, from which we took tram #22 to Vogelmart station, from which we walked a little to a market.  There I spent â‚¬18.00 on ------- for Liah, â‚¬5.00 on an --- for Dale, and â‚¬15.00 on a ------- for Dale.  Then we walked up to the Belfry.  Next to it they were towing cars.  We paid â‚¬0.50 each to climb up a roundtrip 868 steps to the top of the tallest tower in Gent.  Oh, it was so beautiful....  Then we came down, at a lunch of Lasanga (for me) and walked on... we saw witches, guns, rides, canals, castles, and all manner of other things.  Oh, and I spent â‚¬15.00 on a ----- for Mimi.  We saw a show for PolÃ¨ PolÃ¨ and saw a really wierd street theater.... I have pictures.  And that was my day.  BTW: Adrian, u lazy bum; why arenÂ´t you commenting or posting?  '),(13161,3961625,'19,July,2004','2004-07-19 00:00:00','   Today Mimi, Liah, and I went to the Milwaukee Art Mueseum.  However, the exit we normally use was closed due to construction so we had to manuvere our way around downtown Milwaukee to get there.  Once we got there we had a great time, especially at an exhibit with paintings of the northern woods. We got home about 30 mintes ago.  '),(13162,3961625,'18,July,2004','2004-07-18 00:00:00','   If anyone can get me in contact with uncle david, please tell me.  '),(13163,3961625,'18,July,2004','2004-07-18 00:00:00','   Hi  y cant non-travelers post, Rian?  '),(13164,3961625,'17,July,2004','2004-07-17 00:00:00','   Where I sit it is now 6:54 a.m.  That \"where\" is Hyderabad, India.  From my hotel room it looks like a mostly white-painted, sprawling, low-rise city.  I\'ve been traveling with eleven other people, nine of whom have lost their luggage!  Blame American Airlines for that one...  Last night I hosted eight people for dinner at the hotel\'s restuarant and all (even those who\'d be determinedly living on beer and pretzels for the last two days) seemed to enjoy it.  India makes a very good light beer, called Kingfisher, which might have helped :-)  In trying to send you all a picture I\'ve just discovered that my pictures are all corrupt -- blasted computers!  I think the internal card in the camera is less than right.  We will see...  My plans now are to be up and about, getting people towards a 3 p.m. meeting with the first company we are visiting.  Love to all!  '),(13165,3961625,'17,July,2004','2004-07-17 00:00:00','   At 2:00 we left for NancyÂ´s parents house, where I ate a very nice tartÃ©, and read more of the only English book Nancy owns; The Client... they live in a small town owned by a count who demands that all houses be painted a sickly shade of yellow!? That was quite nice, and we stayed there until around 5. Then it started pouring so we lept into the car and drove to the other set of Grandparents, Stevens family... there we quickly saw there ovens (they run a bakery) and picked up Stevens mother. We then drove for about an hour to the coast, where we first took a long and pleasant walk along the seaside, and then ate supper (steak in a mushroom cream sauce) in a mildly fancy restaurant. We then dropped off the boys (thank God) at the apartment Stevens mother owns, which by the way is right on the coast, with a really beautiful view, which includes France and Holland. Then Nancy, Steven, Eusebie and I went for a very long walk, during which we stopped at a pub and got a couple of drinks (hot chocolate for me). Then we drove home. A very nice day.   BTW: If any of you travelers out there wish to post about your days, please feel free to do so.  '),(13166,3961625,'17,July,2004','2004-07-17 00:00:00','   Well, I have a pretty busy day booked today... I will be seeing NancyÂ´s parents at 2:00, and Stevens soon after that... we will be leaving the two boys with NancyÂ´s parents, where they will stay for two days, which IÂ´m not all that remorseful about.  We will then leave for the sea.  Oh, and in a couple of days I might ride horses!  '),(13167,3961625,'17,July,2004','2004-07-17 00:00:00','   Well, itÂ´s 11:21 A.M. here in Vurste, and you lazy bums are still sleeping.... if I`ve gotten any of you confused over what the times are for the time posted, they are set to U.S. Central Time, to make more sense to you lot.  Here IÂ´ll also list the people IÂ´ve invited to be part of this blog..only members may post.  FAMILY  Ian Watt: Dad Lali Watt: Mom Dale Watt: Brother Liah Watt: Sister Pranab Lahiri: Grandad  FRIENDS  Adrian DeSmul: M Matt McClone: M Kelly McAllester: F Herah Kim: F Nick Bonamici: ? Julia Weiss: F Diana Bellafiore: F  '),(13168,3961625,'17,July,2004','2004-07-17 00:00:00','   I will be going to bed soon, as it is 11:27 at night here, but if any of you join during the night, IÂ´ll know by morning...   Night all.  '),(13169,3961625,'17,July,2004','2004-07-17 00:00:00','   Congratulations Rian!  This is a most amazing thing!  WOW!  I am really looking forward to checking this out often.  Well done.  Are you planning to put some or all of your journal up?  '),(13170,3961625,'17,July,2004','2004-07-17 00:00:00','   Hello Everyone....  Wow, what a feeling. Here I am sitting at an old PC in Vurste, Belgium, with an AZERTY keyboard, at 10:54 in the evening, and I can communicate with the world.  I will start to spread the word soon with an email.  But for now, letÂ´s see what this looks like...  '),(13171,3961625,'02,August,2004','2004-08-02 00:00:00','   Hi Guys!  I am now at Jean DeBleeckeres house, where there is internet connection, so I am able to post.  The first day at Rosa\'s, the 28th, I didn\'t do much.  Brit and Yentel came over, and we played a bit, but not much.  The next day, a Thursday, Rosa and I went into Brussels.  We came into Brussels by bus (an hours drive) and then took the Metro to Noord Station.  We first saw the Kings Palace.  Sorry, cameras arent\'t allowed in there so no pics.  Then we walked down a long hill to the Grote Markt, where we had a drink and watched the world go by.  Then we spent about an hour trying to find Manneken Pis (I don\'t think Rosa is very good with directions, we must have asked about 50 people where the stupid thing was, and I\'m not exaggerating).  When we saw it, we went home.  Friday Lorenz and Thalia came back from their fathers\' place.  We all then went over to Brit and Yentels house, where we played in the pool.  I got soaking wet.   Saturday I went to Mechelen with Nancy, Thalia, and Lorenz.  It was a lot of fun.  I bought Order of the Phoenix PB and presents for the pigs.  We saw a church and ate good food.  Fun.  Sunday we lazed around.  Bye for now!!!!!  '),(13172,3536867,'04,August,2004','2004-08-04 00:00:00','   This comic was drawn up during a math class one day because some chick a) actaully said what\'s typed out in the first frame and b) actually left class to book beauty salon appointments when they heard we had a half day one day.   Other Thoughts  Now that exams are over, I\'m going to try and get my shit together, slowly yet surely and get one comic a week.  Than two.  Hopefully I\'ll have 3-5 comics a week if I can get things working out well.  We\'ve not even had 5 comics out already and I\'ve heard bitching about my comic.  I\'m holding back here, but to those people,  nuts to you and don\'t read my comic .  Next comic soon.  '),(13173,3536867,'04,August,2004','2004-08-04 00:00:00','   Yah, I decided it would be best to avoid guest comics for a while. Well since the site just got up and working and there are only a few comics, we decided that it doesn\'t really make sense to post guest comics so early.  So ya, things are starting slow, but that\'s okay! So long as the \"no updates\" habit doesn\'t stick, it\'ll be alright!  '),(13174,3536867,'04,August,2004','2004-08-04 00:00:00','   ...will be up later this week.  There will also be a guest comic from Takkun, maybe from FillerKori someday (if he\'ll let me).  '),(13175,3536867,'04,August,2004','2004-08-04 00:00:00','   OK!   Here\'s the skinny.  Despite the archiving here, both Part 1 and Part 2 of \"Bicentennial Violence\" are one comic.  Sorry about that.  Also, I\'ve not had time to get this week\'s comic done between work and friends but it will be up soon.   Also , exams are coming up for the staff members here, so thinhs might get a little hectic.  So if we get quiet, we\'ll be back soon, don\'t worry.  '),(13176,3536867,'04,August,2004','2004-08-04 00:00:00','   A lot of the features won\'t go up until I\'m finished writing them, and there hasn\'t been much time to do that recently. I have a lot of drama doing on right now, but hopefully it\'ll boil down soon. That, and summer\'s almost upon us, where we will all have untold hours of boredom to make Oddville even better. Look forward to it.  FillerKori out...  '),(13177,3536867,'04,August,2004','2004-08-04 00:00:00','   This comic was inspired by my goofy mind, my love for senseless violence and the movie \"Bicentennial Man.\"  We were doing a science fiction section in english and we had to write a comparison essay between the moves \"Bicentennial Man\" and \"A.I.\".  Oh yeah, I got 82% or something on that essay, in case you\'re wondering. 	Sorry about the quality of this first drawing, I\'m having a bit of a time making the swtich from lined paper to frames.  '),(13178,3536867,'04,August,2004','2004-08-04 00:00:00','   Alright, the site is complete in terms of the design and all the active sites. Glad that\'s over with...  Now all that\'s left to do is to complete the incomplete, underconstruction parts of the sites, primarely the Cast section. As that happens, we\'ll slowly add articles and such to ther \"other\" section.  Whee, fun fun fun.  -Takkun  '),(13179,3536867,'04,August,2004','2004-08-04 00:00:00','   While BlackMage is busy with what he\'s doing, I\'m writing more features for the site. Look for a \"links\" page soon, and be on the lookout for the History of Oddville and Cast features.   The site will grow, and eventually we\'ll have more than one comic up XD.  FillerKori out...  '),(13180,3536867,'04,August,2004','2004-08-04 00:00:00','   Hello.  Ii\'d figured I\'d test this bleeding thing for myself and let you know what  I\'m  up to comic wise.   Comic Editing  Takkun (the web ninja!) has sent me a template that I  think  I can work with and so we\'re looking into mass producing the templates and I\'ll start re-drawing the comics.   New comics  Coming slowly but surely.  I think I might all out stop making new ones at some point in time because a) I run out of ideas quickly and b) editing is going to be a bitch.   The History of Project: Oddville  In the works.  It won\'t be huge, but it\'ll explain a few things.   Cast feature  Takkun has asked me to draw up the cast pictures and those will be done soon.   Other notes  I\'m starting to work now (real life sucks) and so does Takkun.  So if things seem slow around here, we\'re doing our best.  Mad props to Takkun and FillerKori.  '),(13181,3536867,'04,August,2004','2004-08-04 00:00:00','   Oddville on the internet at last! I\'ve been bugging BlackMage/Paul to put Oddville on the internet ever since I saw the first couple comics. Almost a year later, here we are. There\'s no comics up, there\'s a ton of broken links, but we\'re here. BlackMage is the creator, as he said. Takkun is our resident webmaster, so where do I come in? When BM is busy drawing comics and Takkun is busy making the site nice and pretty, I\'m the one keeping you in the loop about the going ons of Oddville, as well as updating site features and handling the PR stuff. You got a question, it comes to me. Have a comment for BM or Takkun? I\'ll give it a quick run over and forward it to whomever. Spammers and flamers bewear, you gotta get by me first. I shall nobely sarifice my inbox to deal with you people, the Internet.  There\'s more to come!  FillerKori out...  '),(13182,3536867,'04,August,2004','2004-08-04 00:00:00','   Well, ladies and gentlemen....1...1...2...3...hello? Excellent! We are now going LIVE! My name is Takkun, and I\'m the web designer, editor and graphics designer of Project Oddville. This site is still under construction, but will still be posting comics...hopefully regularly. In other words, we are active, and taking off! Slowly, but it\'s happening! Wish us luck, everyone!  - Takkun  '),(13183,3536867,'04,August,2004','2004-08-04 00:00:00','   I wrote this comic instead of cleaning my room one day.&nbsp; It seemed like it was more fun anyway.   New comic should be up by end of the week.&nbsp; I\'m going fishing tomorrow which means I have to be up early.&nbsp; Good night all.  '),(13184,3536867,'04,August,2004','2004-08-04 00:00:00','   Corrections made on practically all (not garunteeing anything...I\'m horrible at proof reading things I wrote out...) of the comics done up untill now. The next comic will be posted around 7, or something this morning, so enjoy. It\'s a short one, but that\'s okay. All of the early comics vary in length greatly. Once all of the first year\'s comics have cycled, we\'ll probably set a standard length.    Eventually I\'ll be adding a drop-down listing all available comics. Normally I wouldn\'t add this (cause I\'m lazy), but the archive works in strange ways and doesn\'t list all the comics on any previous comic or, whatever...I\'m not sure if you understand, but if you don\'t, too bad. It\'s not like it\'s gunna effect you or anything.    Anyway, good night...err, morning.    - Takkun  '),(13185,3536867,'04,August,2004','2004-08-04 00:00:00','   OK...I drew this one a long time ago at about 3:30 in the morning one day after watching the movie \"The Cell\" with Jennifer Lopez.  Yeah, I was bored.  I also redrew this comic early in the morning one day, so there\'s a few continuity errors, but you may not even notice them.  If you do, I\'m really sorry and I won\'t screw up next time.  As for lack of updates, I\'m working on it.  I\'ll have to start on the cast pretty soon and I\'ll talk to the other staffers about features and what not.  As for the next comic, expect it soon.  '),(13186,3655970,'29,June,2004','2004-06-29 00:00:00','    Thank God! It\'s all over. Ermz, but thurs still got maths s which im going to heck i think. Going there to play play and relax. Wahah. Anyway, I din\'t quite seem to finish my all my papers on time?! Personally, I feel that it doesnt give u time to think. Once you look at the questions, you must have an idea how to start etc liaoz. Well maybe I\'m just a slow turtle. hehe. This 1 month hol, I dint utilize it properly. I can only blame myself. But I\'ll still have faith in Him that I\'ll do well... :P   Whenever after playing the drums, I will somewhat feel disappointed?! Perhaps I already know the answers but just dunno how to move on from there?! Haha. Ermz, I will not elaborate it further. I guess I just need to seek Him more and pray more about it.  Felt abit ironic this evening. Haha. There\'s this person who came up with this idea/plan and wanted all of us to get involved. I felt that it was kinda weird and maybe ridiculous? Because that person should evaluate the whole situation properly first. -.-  I guess I\'m not part of the \'team\', so I don\'t feel a need to do what you have suggested. : The Lord is still the best because of His unfailing love and faithfulness. I want to keep that fire, love and zeal going for You Lord! '),(13187,3655970,'24,June,2004','2004-06-24 00:00:00','    Guess what? I just woke up!! HAha. It\'s 2.30pm already. SCary manz. Just wasted half a day sleeping. Had a great time with Yanda, Kenny, Engie, Shubert and Terrence at Megapool. After our usual mugging at the CC, Yanda suggested pool. So eventually, we contacted Kenny and yeah..He approached the rest too. However, we had to wait till about 12+ for Kenny\'s van, and so went to his house and watched some movie clips etc. Oh, I got to read the \'I Kissed Dating Goodbye\'. I was so intrigued by it that I wanted to have it etc. Haha. You must be thinking \'Wah..This ben ar..keeps on talking about BGR and love etc..\' :P I guess just for fun and more knowledge on such things?! Hehe. Oh ya..Before we went to Kenny\'s house, Grace phoned and told me that Priz had ran out of her house, crying etc. Therefore, Yanda and I decided to meet her and maybe try to help. But she did not want to tell us anything or let us pray for her. Nevertheless, Priz, if you got anything or need help etc, feel free to call us etc :)  Hmm...then after pool, went to Simpang Bedok to have prata and watch soccer. Germany vs Czech Republic. Wah. I never expected GErmany to lose man!!! Czech had such great defence!! There are a lot of black horses in the tournament manz. HAha. Amidst all these fun etc, as usual, my mum woke up EVERY hour to call me! You should know why. I really feel that sometimes I\'m just being treated like a small kid who will encounter some danger etc. ARggghz. Of course, I can understand why she was doing all these - She was concerned about me. After I got home, she started scolding me and linking issues which are not related at all. I got so pissed off. -.- You know la, when parents scold, they will start a conference like that...talk and talk non-stop. Uncontrollably, I \'rebuked\' =X Then as I was blogging, she came to explain and apologise. Ar!! :D Touched yet \'immune\'? This is not the first time such things happened.  But honestly, I shouldn\'t take her for granted. She did all these because she loves me. I guess sometimes its akin to flying a kite. Too loose, the child feels empty and lonely. Too tight, the child will hate all the nagging and etc.  Dear Lord, I want to thank You for giving me such wonderful parents. Teach me to be more obedient to them and You. And also thank You for everything that You have done for me. Indeed, You are so great and faithful to me. :D  *Cheers* '),(13188,3655970,'20,June,2004','2004-06-20 00:00:00','    Today Yanda released me on an aeroplane! Haha. He said wanted to meet me for main service etc but when I called him at 910am, he\'s just awake. Ya..then I went alone. :| But it was really a great feeling!!! I felt so refreshed and reassured by Him. The sermon by pastor Rony was the answers to my doubts and questions. It was about patience and obedience. I realise that no matter what happen, we should continue to have the faith in Him and wait patiently, just like Noah waited 120 years and etc. Also, we need to do things which are within our capacity to improve our spiritual walk. He wants us to get going FIRST, then He will reveal to us more. Pastor Rony mentioned about Samson who was never forgotten by God. And this applies to us too. Hebrews 11:6.  Besides that, the speedlight\'s sermon also enlightened me. It checked what my focus was and is in the music ministry; whether I was playing for God or myself. Well, I learnt that we have to be humble and just serve Him whole-heartedly. As a drummer, it is very easy to play for self-glory and impress upon others. God doesn\'t want such people in the ministry! Like what Weixiong and Joel have said, we are the frontline people who are to minister to them, not to impress them with our skills. Therefore, it\'s imperative to keep our focus and motive right and pray more fervently. The strong bonds between each other in the ministtry are equally important too. I really thank God for these brothers. :) Oh ya Joel! Thanks for your concern and encouragement! You\'re such a great blessing from Him! :D  HAha. Today actually wanted to study after pract, but was just so tempted to go makan with the gang! Honestly, I had a great time with them, although I always kena suaned like siao!! :( They always bully me. Big size is a hinderance I guess. :P After the practice, the headaches Came back strong! I could feel somthing is pounding on my head! It has been like this for weeks! HOpefully it will be gone when school reopens. I just felt like stop what I was doing just now and take a rest. Even now as I\'m typing, the pain is still there!   Celeste said, \'being humble is where u know that u are weak and nothing when u are before the Lord. That only thru Him u can overcome, and have the strength and wisdom. So now u are in the Lord.. meaning u should know ur identity in Christ.. that you can live life exceedingly....being humble to me is more like telling the Lord.. i can\'t do it without You. But low self esteem is where u say u just can\'t do it.\' THanks cel. YEpx. Was kinda surprised also - something \'good\' happened?! Hope it doesnt turn out to bad or whatsoever.  Lord, thank you for everything. Indeed, You are so faithful. Teach me to be patient and obedient to You. '),(13189,3655970,'18,June,2004','2004-06-18 00:00:00','    Wah today I was pissed manz! I lost my reading club card on way to the CC! Panic-stricken!! So I cycled around the path that I\'ve taken and you know what?! The bicycle chain just BROKE OFF! 1 blessing after another! HAHA. DAmn pissed by then. I pushed the bike to the bicycle shop which is quite far and wasted $12 on a new chain. Finally managed to get into that cosy room with strong air-con! :) And it was already 3plus. Wasted 1 hr plus for all that rubbish things.  Then I began to ponder - was God trying to make fun of me or what? Of couse I know that I shouldn\'t blame or question Him. But arrgghz..just angry?! (Not of Him) Then I recalled the zhai verse, 1 THS 5:16-18. After saying a short prayer and etc, I started to \'fill the mug\'. My mum accompanied me to the doctor because I was having headaches these few weeks. It just came on and off. Surprisingly, the chi doctor just said something\'s wrong with my lifestyle - sleep late wake up late. HAha! Of course, he did not charge me. But find that deduction abit funny?! Maybe he wasn\'t a zhai doctor?! haha.  Talked to Alvin Chia alittle bit on BGR again! Haha. and later Kenneth came along..and Hey Kenneth! I really envy you man! You already have the headstart and God has answered your prayers!! Cool manz. Just relaxed and let Him take charge... This prayer thing somewhat triggers my mind to think whether my prayers are being answered etc. Of course I\'m not referring to the BGR etc, had enough of that. What I\'m trying to say is that somehow my faith is being shaken at times and prayers are not answered?! I really don\'t know.Is it that I don\'t do what I ought to do enough? Did I seek earnestly and fervently?  Need to look for someone who is spiritually mature to answer my questions etc!!  Hmm, meanwhile I think I\'ll just continue to have the faith and not doubt his plans.  Lord, show me your ways as I can\'t survive on my own! Change me to be more obedient to You I pray. '),(13190,3655970,'17,June,2004','2004-06-17 00:00:00','    Went to the reading room to study at 3 plus. On my way there, I listened to class95 and as usual, it really ROCKS! All the old songs that they played were just so NICE! So people, start listening to class95!!! Haha. Yanda and Grace who came later, were there too. You know what? the person who was sitting near us, moved away to other tables...you should know why.   Oh manz!These few days the BGR issue has been preoccupying me...and I talked to people about whom they like and etc etc. I dunno why! But frankly speaking, what matters most now is to maintain a good relationship with God first. If I can\'t even love and commit myself to Him, not to mention the opposite sex. And I can\'t deny that I do admire girls. Sometimes I just want to get to know them better, that\'s all! NO other motives! But you know la, girls are always very paranoid and sensitive...so ya in the end you might even be shunned. Haha!  I think I really need to reflect. Some of you who know me should know that I\'m quite negative when it comes to certain gatherings, outings and socializing. Well, I\'m just afraid of being left out and Xtra. I had tasted such feelings before and it\'s NO FUN! Well, I guess I\'ll just wait patiently for Him to change me and pray more in this area.....  The night is always young and good to have fun! ^_^ '),(13191,3655970,'17,June,2004','2004-06-17 00:00:00','    Our music ministry!!! yeh manz! Haha. Yes...now going to mug hard! Die! I have tons to memorise...May God grant me more wisdom...James 1:5&nbsp; urlLink     urlLink '),(13192,3655970,'17,June,2004','2004-06-17 00:00:00','    Well, I think some of you might be flabbergasted to see me blogging?! Ya, I also don\'t know where the urge came from! haha. I started this thing at 1am till 3pm, encountered lotsa problems manz! So frustrating!Ok looks like I have lotsa things to learn - how to decorate my blog etc?! :P But maybe I\'ll just keep it simple as it is now...  Anyway, I began the day by swimming(training) and was rather drained out at the end of it man! VJC swim team was there too and it\'s rather intimidating! Haha. Because they seem to be great in everything! Oh, I managed to get my tan too. Had a little bit of strawberry flavour..HEhe. At the end of the training, I was totally drained out. The team then went to have some cheap Jap food. Not so bad. ^_^   Met up with some of the music ministry members to watch The Best Bet. IT WAS SO HILARIOUS!!! =D Then at 7pm everyone except Wei Xiong turned up and went for dinner at Fish & Co. HAha...Imagine an army officer being late! Oh ya, I was like being forced to reveal my secrets on who I like and etc...Akin to last sunday when I was having dinner with the big gang. Those who were there should know...After dinner, we went to the coffee club at Hotel Rendevous.   And we moved on to our main agenda. All of us shared quite a lot on our camp experience and the music ministry. Ya..and I learnt quite a bit too. One of them is not to be afraid of making mistakes and constantly be focused on the purpose of playing our instrument - to serve God whole-heartedly and humbly. Yes, we have to be skilful but that should not compromise or affect our worship to God!!! Somtimes we might get too obsessed with the technical issues or our surroundings instead of focusing on God. Besides that, we should always be conscious of what we do, react and behave. Is it to please Him or others?! It\'s rather tough I would say, but we must be disciplined and give our very best for Him.  I hope there will be more of such gatherings or outings, so that we can strengthen our relationships with one another instead of just being a group of musicians. :) I really enjoyed the session! Thanks Joel! And of course, thank God for everything!!! :D '),(13193,3655970,'31,July,2004','2004-07-31 00:00:00','    Went running, cycling and swimming today I felt so refreshed after the whole thing. Well, though I was alone, still I was joyful and satisfied cos I know the Lord was and is with me. During my swim, I saw Douglas who is in NDU now. He told me all the \'sufferings\' that he has gone through and wanted to get out asap.  All the people there are much more fit than the normal fit. Their 2.4 9+ mins and below...pull-ups literally 20 and above..Though it is prestigious but still xiong and terrifying~ Hehez.   Why do you call Me \"Lord, Lord,\" and not do the things which I say? Luke 6:46 . This verse makes me soul-search and reflect on what I should and should not done, be it in my actions, behaviour and speech.  Now I\'ll be at the CC everyday to study, be it rain or shine. I\'m trying to have an agreement with Yanda. Hahah. Pretty lame one - the person who doesnt turn up, will treat. Hehez. I guess that\'s the only way to force us to really study.   God is greater than our heart, and knows all things. 1 John 3:20 '),(13194,3655970,'25,July,2004','2004-07-25 00:00:00','    Alright, just finished a bowl of honey stars. Gosh! It triggered my hunger already, but ok I\'ll control because Mr Fat is here...hahaz.! Im like very particular about this Mr Fat&nbsp;because he\'s&nbsp;soo obnoxious! Can\'t wait to kill him. =X   Ok...I\'m pretty health-conscious. =P    &nbsp;  Our teachers treated the swim team to PIzzaHut. What else? Of course, whack ar!!! Hahaz. Tons of cheese...haha. One of the teachers talked about his days in NUS and career paths etc. And honestly, I don\'t know what course to take in future!!! Don\'t ask me what I like etc, because I like everything? Hahaz. Ermz..I was thinking of civil engineering since I like construction a lot! Being a construction worker rocks! =D Just imagine yourself plastering the wall, mixing the cement, laying bricks etc..Oh if can, I want to operate those mammoth&nbsp; and cool cranes...They simply rox man! Ok...all these sounds peculiar! Back to serious business, they say civil engine is like no future? *shrugss* God will show me His ways...so no fear I guess..hehe.    &nbsp;  Indeed, Time really flies manz! Today, I did a few things ONLY! Supposed to go to my cousin\'s 21st bday party, I went to study instead.    -.- I have tons of homework to complete. Shucks! And Sunday as usual la, is a long day. So&nbsp;Sat is my&nbsp;only day to do work?! Really no life manz! Everyone is counting down to prelims and etc etc...it\'s really terrifying manz! I guess this couple of months is all about eating books, what else? Ok..enough of dwelling. '),(13195,3655970,'22,July,2004','2004-07-22 00:00:00','    Arrrgghz...School...Classmates...Homework etc.. all these are indeed prosaic. I really can\'t wait to go army. Haha. Well, at least you don\'t need to kill that much brain cells and ya..Army takes care of everything - food, sleep, \'play\' etc etc. =X Ok..I know those who are in army will rebut. Alright, enough of that whingeing.    &nbsp;  Anyway, I skipped my sch\'s sea carnival just to go study. Honestly, I wasn\'t involved in it in any ways and I dont have my clique at all. Therefore, no point wasting my time moseying around right?&nbsp;=P I know what you are going to say! But I\'m just trying to catch up and put in more effort in my studies since I\'m lousy and mediocre. Logical isn\'t it? Also, I\'m not a nerd(like some of my friends)&nbsp;who only knows life is studies, right? I do play hard too. Don\'t you agree? Well as you know, All work makes Ben a dull boy. Whahaha. Come and join my club \' The Muggers\' Club\'. Free membership and an upsized kid at your service. ^_O    Had a short yet meaningful sharing with Cindy. Haha this amazing and funny girl. I got this peculiar notion in my head again. If I tell you about it, you\'ll be&nbsp;rather amused and find it absurd?! Nvm about that.    &nbsp;  Yet I am poor and needy;   may the Lord think of me.   You are my help and my deliverer;   O my God, do not delay.  Psalms 40:17 '),(13196,3655970,'18,July,2004','2004-07-18 00:00:00','    What else can I say? Lord is indeed so good and faithful! Well, this week, I was thinking of ways to change and improve my QT with God. And guess what? Today, both the spdlight and main service\'s sermon were on quiet time! So surprised man. God really gave me the answers and solutions man!!! Oh Dr Tan mentioned about personal retreat. That sounds really cool manz, it\'s somewhat similar to camp?! Hmmz..I wonder where got such places...^_^ Anyway, back to QT, ya it\'s not about doing it to please Him or a routine, but it\'s the desire to BE WITH HIM and have THIS INTIMACY WITH HIM. I realised that sometimes many people including me, keep saying that God is the most important, yet it doesnt reflect on us as things like our work, studies or even sleep distract us and took away the priority. Sounds ironic isn\'t it? Well, it\'s never too late to change, right? So let us have more spiritual discipline, to really seek Him daily. =) Must always tell myself not to be a member of NATO (No actions, talk only) HEhe. =P     My father went to church with me today. =D After the service, I asked how everything was etc. The answer was the same as before - \'Aiya, I come for fun only...I know all these... All are the same, no difference...I know what I\'m doing etc etc\' I was really discouraged. Also, as my uncle and his family were at my house just now, we began to talk about religion. My uncle joined forces with my father to argue with me, saying how they felt towards Christianity and etc. Wah, I was really struck dumb by what they said and commented. I didn\'t know how to answer and rebut... Oh my uncle asked me a very simple question which totally threw me off!! Ask me if you wanna know. =X Wah..after all that, I felt so useless and lousy...Now it\'s even harder to convince my cousin who is my uncle\'s son! -.- My mum, another one la, keeps telling people that I\'ve been spending too much time in church, which would affect my A lvls. She said that people in my church should encourage me to stop going to church. Nonsense from her always.&nbsp;WHahhaha. AR! Sick of all these liaoz! I know what I\'m doing... so Pa and Ma...Shhhhhhhhh!!!! HAHA.     &nbsp;   Hmmz...looking back at my life, I think I have changed to a better person?! Well, at least I\'m not so introverted and shy? Ermz although im still alittle bit.. =P And I have lotsa friends now?! Haha. Don\'t laugh ok! Help me to get out of that box if you can! Lolx. Oh yes, any1 wanna go somewhere tranquil, sip a cup of coffee and maybe chat + reading?! :P     &nbsp;   My faith will be strong because of You Lord.    I want to seek Your face always.   &nbsp; '),(13197,3655970,'17,July,2004','2004-07-17 00:00:00','    My beloved pen fell to its death yesterday. The tip kissed the floor twice, leaving 60% of life untouched. Haha. Ok I know it\'s lame and crappy.     &nbsp;   Today went to attend a bible conference by Bill Crowder, the director of Daily Bread. It was on Joseph: Overcoming Life\'s Challenges. Normally, my fridays always end at 5+, so I was abit worried whether I could make it on time etc. But thank God man! The teacher decided to end the lesson at 4+, therefore I had extra time to \'cool down\' before I go for the thing. The speaker rocks. Like what Joanna said, he has kinda brought the passages to life. Somtimes you read the passage without understanding what the reasons are etc, but as he read it, you can really catch the implications etc. Hehe. We should always persevere no&nbsp;matter what happen. Bill said We are always in a hurry, but God is not!! Since you trust&nbsp;Him, why should you be impatient and yearn for immediate results etc?&nbsp;It\'s too soon to quit!!!&nbsp;Joseph&nbsp;became a ruler after 13 years of perseverance&nbsp;and moulding by God.  &nbsp;We are always in good hands&nbsp;of the Lord for&nbsp;He&nbsp;makes no mistakes...     &nbsp;   Oh after the 1st half of the session, they were giving out free copies of ODB. Then, 1 sms came from my classmate, asking me to lend him my copy. I don\'t think it\'s pure coincidence but God who was doing all this! So real indeed! Actually He was a Christian....ya so he saw my DB on my table and begun to browse through it. After a while, I noticed that he was deep in thoughts. Ya...I did not expect the sms to come at all and etc. I think God wants me to help him.&nbsp;Ya...Pray that&nbsp;he would come to know God again. That reminds me of my father. I hope this sunday he will go to chuch with me. He dislikes Christianity and keeps saying he knows what\'s happening in church&nbsp;(like how people brainwash etc etc..) Oh Give me a break!!! Nevertheless, this is the first step, Salvation is still a long way to come. But ya, maybe pray for my father if you can. Thanks. Also, recently my parents are badgering me to stop going to church because they&nbsp;are afraid&nbsp;it might affect my studies etc. Honestly, I think it\'s all crap! --.--&nbsp;They said I\'m too obsessed with all this church stuff. It riled me man! I tried to explain to them, but seems like it\'s to no avail. If argue more, the whole situation will just get more uglier, so I&nbsp;think just leave it to God. Hehe.     &nbsp;   Walked home with Kenny. He enlightened me once again. This time, it was about my ministry and other stuff like showing concern etc. I wont elaborate on it. Ya, but I think this sunday will be the beginning.. =D     &nbsp;   A Christian life is indeed exciting! Don\'t you agree? As you dwell on Him more, He will reveal Himself to you more. Let us keep the zeal and first love going man! Love you always Lord!     &nbsp;     &nbsp; '),(13198,3655970,'15,July,2004','2004-07-15 00:00:00','    Should I say that I\'ve screwed up my week(sleeping a lot! Lazing around..)? Oh man, this really sucks. Must put a cessation to it!  I really need to be more focused in my studies as it seems like I\'ve become more playful this year? Last year was like hugging the mug all the time?! Hahaz. I\'ll just need to hang on for another 4 months and WAH LA!! =P Cannot wait to get out of that junky school! =X I dont like the environment, people, food which is horrendous. Oh, talking about ppl, You know what? I somewhat got shunned by people who r unhappy about me mugging hard? So absurd!!! Well, slacking in my sch is a norm. If you dont slack, you are an outcast. -.- Whatever.  Recently, Grace told me about how she felt about church, the speedlight and etc. It made me ponder too. And I began to ask people for their views and etc. Just some simple \'startup\' for chat I guess? Haha. Anyway, I feel that everything\'s fine. But sometimes, I don\'t quite get the speedlight\'s sermons. Maybe it\'s because the age group ranges from p5 to 20+ which is rather large, so kind of difficult to cater to everyone. Kind of funny?! Hehe. In my opinion, we should not solely depend on church services for feeding, but it is our daily effort in reading His words and doing QT regularly. To put in layman\'s terms, I think it would sum everything up by saying \'practice makes perfect\' or simply think of it as a BGR relationship. If you want to get intimate with Him and know Him better, should you not do something about it? And it should not be a weekly thing only. But because of our daily activities, be it work or studies, we inadvertently neglect Him. Yes, I\'m guilty of it. Ok..Must really put all these into ACTIONS!!!  Oh and yesterday went to a Christian bookshop in Eastpoint with Yanda. Saw some great stuff eg like the daily devotional materials which are akin to daily bread. But I think it is more indept than the latter. Too bad its not free...1 year $23.20. Sounds cool right? Cheap and worthwhile. ^_^ Yepx..I bought this little book titled \'How to have A Quiet Time\'. Im sure this would help me to enhance my learning of His words and having a much effective QT with the Lord! = )  Today met up with Cel to have dinner. Had a great time. Thanks for all the sharing manz. Ya..and it is all the downs that God wants to use to mould us to become better people. So I was wondering how come my life is like smooth? Wahah. Of course, I\'m not implying that I want troubles in my life etc, ya..but..ah nvm.Haha. Nevertheless, He has changed me for the better and done a lot for me. =)     Never be lacking in zeal, but keep your spiritual fervor, serving the Lord. Be joyful in hope, patient in affliction, faithful in prayer. Share with God\'s people who are in need. Romans 12:11-13 '),(13199,3655970,'10,July,2004','2004-07-10 00:00:00','    Had swimming competition on Thurs and Fri. This is my first time because I started swimming again when I was in yr 1! Before that, I was in NCC. So in a way, I\'m like a newbie. Haha. It was really an eye-opener for me. Yepx. Cool People and those Pro swimmers were around, and of course, there were some cocky ppl, esp from those good schools. You know la, Swimming is always dominated by AC and Raffles people. They are really fast!! Oh Ang Peng Siong was there too. Wah He\'s really HUGE man! His calfs are the size of my arms! As for his arms, no need to say liaoz. I want his body!! :P I guess when u have a good body, you will be more confident of youself? =D   I did pray for everything to go smoothly. Well although sometimes things might not happen what we wanted, we should always thank Him for He has the best plans to prosper us!    Lead me Lord I pray  Take me, Mould me Use me, Fill me I give my life to You Hold me, Guide me Lead me, Walk beside me I give my life to You '),(13200,3655970,'07,July,2004','2004-07-07 00:00:00','    Got back my CT results. And only 1 word can describe it - CARELESS!. When I looked at all the errors, I was angry with myself. Anyway it\'s over. No point dwelling on it. Thank you Lord for the results and all that you have done for me.  Hmmz, my classmate who was on the same boat with me, reminded me of my impetuosity. Both of us were considered the consistent ones and we had the same problem - afraid of FM affecting other subjects etc. Teachers and friends persuaded us continuously. Even the HOD of FM reassured and had faith in us. Yet I did not heed their advice! I withdrew from it on the day of the registration. =  I realised that I succumb to small setbacks and failures easily. And this is not the 1st time. It\'s high time for me to change and seek fervently my identity and image in His words. Remember Breakthrough chapter 4 and 5?! Hehe. :P   Rest in the Lord and wait patiently for Him Psalm 37:7 '),(13201,3655970,'05,July,2004','2004-07-05 00:00:00','    Today went for Cg dinner at Fisherman\'s village. When I reached there, Wah! It was so beautiful manz - referring to the beach and etc, although the clouds looked ominous. You might be thinking whether I\'m a mountain turtle or whatever...like never been to the beach before.. haha! There are restaurants and cafes just next to the beach. The ambience was fantastic! Hehe. Perfect place for fellowship/gathering and also lovebirds. Pretty romantic too..HAha.. The restaurant that we went was a \'Zhi Cha\' style. We were sitting in the open, and afraid of the rain coming, as dark clouds covered the skies. But hey you know what? God answered our prayer instantly! The rain did not come at all! Just a couple drops..God is indeed SO REAL when you have faith in Him! Well, it might seem trivial, but I feel that it\'s through all these small little things or situations, then we can learn and apply His words in our daily lives. Don\'t you agree? =)  Oh ya! the whole cell forced me to do a foolish act to Joanna. haha! Always trying to bully me :O Well, I shall not elaborate it. Must keep it as a secret. :D Wanted to stay longer to fellowship etc, but ya tmr got school!!! I\'m still at slacking mode manz! This week, think going to get back all the papers?! Need to pray hard for miracles..Hehe.  No matter what happen, God is with us always. '),(13202,3655970,'02,July,2004','2004-07-02 00:00:00','    As usual, love Sundays always! The day where we can really focus on Him, with no distractions. Hey Grace, nice talking with ya. :P  After dinner with the gang(lesser people yest) at KR, all of us went to have pool and watch soccer at Simpang once again! Hahaz. Initially, I declined it because I wanted to do hw and slp?!? =X Ar! Well in the end, cant bear to forsake having fun with them! hehe. =D Sometimes I simply envy those people who are having REAL holidays. No hw or tests to worry. Can\'t wait for this A lvl year to pass!!! And then *Evil laughter* !!  Sometimes the things that you do might send some wrong signals or ideas to people who would then get irritated/angry/upset...So what\'s next? Leave it as it is? Or voice out? No point complaining this and that, as if all these would really help?! -.- Well, what I want to say here is being transparent and open. Human relations....What a headache!! Hmmz, pets wont give you such problems..haha.   Whatever it is, the Lord is still the only one whom we can truly rely on... *cheers* '),(13203,3655970,'02,July,2004','2004-07-02 00:00:00','    Just came back from pool and supper with Kenny, Engie, Shubert, AlvinChia, Meiling, Peter, Shufang, Yanda and Weixiong. As usual, had a great time! You know what? Actually I didn\'t want to go because I finished my swimming from 4pm to 8pm. I was drained out totally. Hahaz. Well, 4pm to 6pm was for bathing in the sun. Too bad Mr Sun didn\'t come out again!! Arrgghz. But I got an unexpected call from Alvin Chia, and he somewhat persuaded me to go etc. :P Er, I dozed off abit while watching soccer. Hehe =x  While I was playing pool, I received smses from someone who told me some bad news ( It\'s about me and some other people). Gave a call to her to find out more( Thanks a lot). And Wah! I was flabbergasted and dejected by the strong and hurting words that the person used. Everyone does get livid, but is there a need to vent your anger by saying such things?! Are you not aggravating the entire situation and causing more misunderstandings and hurt? If you thrash things out with the person 1 on 1, isn\'t it much better? Rather than going around telling people how bad the person is and etc? I\'m not furious or whatsoever. I just want things to be talked out nicely instead of such things happening! Give me a chance to explain and apologise too. '),(13204,3655970,'01,July,2004','2004-07-01 00:00:00','    Today slacked at home the whole day. Gosh! I could have done something meaningful or interesting manz!! So sad. Couldn\'t swwim and tan because the sun kept hiding throughout the day! Arrrghz. Wanted to run too, but procrastinated it.Hahaz. By the time, I wanted to run, it was already 10pm!  Priz showed me some comments made by a guy in his blog. And it struck me! Was God trying to tell me something?  \'Here\'s what my pastor said just now. You who serve God but never pray, you should consider your resignation.  Simple yet powerful, made me think what I have done so far. Have I prayed in my service to God? Don\'t be surprised when you go to heaven later, the wages for those who prayed for others are bigger than those who seem to serve in the eye of human but never prayed. By the way, today\'s passage was taken from Matthew 13:31-50. What do you think?\'   During my QT, I felt a strong sense of guilt and c. Immediately, I asked for His forgiveness and mercy. Felt so good after the whole thing. Thank You Lord.  Besides that, that \'comments\' made me ponder how I perform etc. And indeed, its egregious! Of course, I need to hone my skills etc, but if I play wrong during service which disrupts the worshippers, then yes, I must admit my mistakes and make sure that it won\'t happen again, and not keep dwelling on it and be upset or whatsoever. As a drummer, I would say it\'s abit difficult to worship while playing, because you have to think of the fills, tempos and etc. Anyway, the bottomline is pray to Him more and commit everything unto His hands. What\'s impossible with man is possible with God. Amen. Love you always Lord. '),(13205,3655970,'14,August,2004','2004-08-14 00:00:00','    Just came back from swimming which was so unproductive! Arrghz! After doing my kicking, I felt like sleeping as I lay down to somewhat tan. So didnt manage to kill Mr Fat. Haha.  Yesterday was Cheney\'s bday. Quite fun. When Isabel and I reached the chalet, we were shocked! haha. Cos not many people were there. It was 9 plus at that time. Well, the rest of the group would be late etc, thats why. So we waited for them to come and Isabel, poor thing...she has test today. And ya...she stayed till pretty late. I noticed something - Cheney and his family members are good-looking!! Hahaz...Good features etc. They are Peranakans..so I guess that makes the difference...?! Oh ya, the food there was nice...And his uncle keeps asking us to eat, too bad I wasnt on form! =P Also, the cake was pink in color!...Cheney = Cherry... Haha!  Planning to cut Armani and gel it up, but my previous \'ordeal\' makes me ponder. Whatever. Should I go and sleep now or study?!  WOooO LaAalAaa...Rock it on Baby!! =D '),(13206,3655970,'09,August,2004','2004-08-09 00:00:00','    Everyone is going out to celebrate National Day, yet Bel, Cindy, Yanda and I actually went to study! Amazing. Well, to me, the NDP is still the same, nothing special. I think the most spectacular one would be year 2000\'s. It had lotsa events and etc. Yepx, so it was a holiday meant for studying.  Arrgghhz. I displayed my stupid golden silence once again! Felt bad about it. I just couldn\'t bring myself to chat or talk. My presence was no better than my absence. I simply don\'t know how to use the \'mixer\' Haha. Well, I guess I\'m still that  shy and introverted guy , as always, unlike those sociable guys who know how to entertain people. Sorry. Haha. I\'m trying.  Went to play basketball this morning. Pretty cool. Too bad some of the people there didn\'t want to play 5 vs 5 which is like so exciting?! It will really make you run like a panting dog. ^_^ Well, 3 vs 3 wasnt that bad, but the opponents were on relax mode. Oh, and I hope that I won\'t grow any taller by playing basketball, if it does, thats it manz, bball-phobia....I\'m not like those tall and BIG people. Imagine a bamboo walking around, eeeewww...haha! Thats why I need more bulk! Btw, is tall a merit? Or issit that I\'m TOO tall?! Scary...Hahaz! Whatever.  I have this small bubble-like blood clot in my mouth. I poked it and tried to remove all the blood, but to no avail. I was thinking of simply adding salt on it! That would really hurt big time I think! Lazy to see the doc. Waste money. Alright folks, take great care, and enjoy yourselves. See you guys soon. ^_^  The song \'The Potter\'s hand\' says it all for me... Love you Lord '),(13207,3655970,'06,August,2004','2004-08-06 00:00:00','    There\'s this foolish person who goes online and simply masters the art of stoning, cos he has nothing to do, unlike others who are having a great time online. =  I suddenly have the urge to play basketball, badminton, cycle and rollerblade! Wanna play bball tomorrow morning, but couldn\'t find anyone. Sad. Ar...Whatever.  Pardon me for being so simple-minded and childish, perhaps. I\'m like that. A kid inside this big body. :-)  Are you hypocritical sometimes?  WHATEVER. '),(13208,3655970,'05,August,2004','2004-08-05 00:00:00','    Went to Bugis just to buy drumsticks. How dumb could I get. Hahaz. I was thinking of who to ask along and have dinner after buying. Just couldn\'t think of anyone \'suitable\'? Cos I\'m a weird guy. Hahaz. Oh, that place is like full of pretty and handsome faces?! Hehez.  Tomorrow is holiday (pon sch) cos my sch is celebrating National Day. So I\'ll have 5 days of holidays?! Cool~ I was telling yanda that maybe one of the nights, we could go pool and supper with Kenny and gang. Haha.  The triathlon is coming soon and I haven\'t gotten a helmet! Shucks! Need to get helmets for my fren, Xueli and Jac(maybe). Hey! If anyone has \'lobang\' for helmets...borrow I mean...pls inform me. Think it would be foolish just to buy it for the event right? We are just going to have fun there, not trying to compete. So...ya..if interested, come on manz! Hehez.  -.- My mind is blank now....Anyway, take care folks! See you guys soon. '),(13209,3655970,'01,August,2004','2004-08-01 00:00:00','    What else can I say? Sundays rox man! The main service\'s  sermon was on powerful truths for effective living, which somewhat links to my previous entry. So amazing and coincidental. Hehe.  Oh and the cell teaching today was enligthening. Joanna talked about Luke 15 which is on the parables of the lost sheep, lost coin, and lost son. I had read it before and the basic inference, or should I say the only one, was we are precious in the eyes\' of the Lord. I did not go into deeper thoughts and analyse the parables. But the last one was really zhai - It has more hidden truths and meanings. Indeed, God is the ultimate story-teller and magician.  Indeed, God plans everything perfectly and let them fall nicely into place. Today the practice was different from the normal ones. Normally, we would pray in the santuary and it\'s a short and rush one I would say. But this time, we gathered in the Jireh room. Alvin Chia and Joel started sharing. It was just so wonderful. Also, we have decided to start early and end early, so that we can share and pray together. Well, all of us somewhat knew that pract was pretty routine - you play, practise...and this cycle repeated itself. But not anymore! This time round, this change will definitely enable us to bond together spiritually and walk closer as a team to God. It\'s not about playing together and being in a ministry anymore!!! It\'s more than that man! *Cheers*  Ok I need to eat more in order to gain more weight. Haha! are you amused? Ok. Whatever. Alright, another exciting week ahead isn\'t it? Hehe. '),(13210,4055095,'27,July,2004','2004-07-27 00:00:00','   Hmmm, after thinking about this whole blog thing, I realized the similarities between the relationship of our lives with the internet and the Borg from Star Trek: TNG.  The internet is this big storage compartment of facts, news, ideas, opinions and so much more.  And with the advent of the blog, it\'s now become this depository of people\'s thoughts.  So is the internet becoming a collective brain for the entire world?       If I need to know something, I go to the internet.  I usually Google it.  Not too long ago, if you needed to know, say, the weather, you had to wait until the 10 o\'clock news came on (or look out the window).  Of course, the Weather Channel changed that, kind of.  You still had to wait until they displayed the weather you were interested in, but it was better, nonetheless.  Now with the internet, I can find out the weather instantly.  The same thing applies to other info like phone numbers, addresses, driving directions, news, whatever.  The internet has provided this collective area to find all this info.       Now, think about personal computing devices and where we\'re headed with them.  I personally use a PDA on a daily basis.  It was a godsend for helping me remember stuff for work.  My PDA doesn\'t have wireless capability, though.  I didn\'t have any use for it and didn\'t want to shell out the extra money for it.  But all of the newest PDA\'s automatically come with wireless capabilities whether you want them or not.  So you can surf the internet anywhere, provided you can find a WiFi hot spot.  You can have access to this collective of information at any time.       I\'m sure personal computing devices will continue to change and adapt to our needs.  With that said, is it too much of a stretch to think of everyone in the world who has access to this seemingly infinite brain (the internet), sharing their thoughts and ideas and experiences, so they all, in essence are sharing the same big life or same big brain?  Sort of brings us all together doesn\'t it?  Gives more meaning to the idea that we are all part of the same whole.  Which is why I compare the idea to the Borg.  They are many parts of a greater consciousness.  All of their thoughts are one.  No, I don\'t expect us to all lose our identities and individual thoughts.  But it is interesting that we can all converge in this one cyberspace.  '),(13211,4055095,'27,July,2004','2004-07-27 00:00:00','   Here\'s the maiden voyage of my blog.  I\'ve heard of blogs for awhile now, but really haven\'t looked too much into what they were.  Of course, I\'m always interested in reading other people\'s random thoughts.  So why shouldn\'t I share all the mundane crap that goes on in my life as well?       Work went well today.  Lots to do and not enough time to do it in.  We have 2 people out of the office, so we\'re short handed.  But we\'re not hurting.  I work in the County Clerk\'s Office and we\'re currently looking at changing the way we archive records.  For decades, we\'ve used microfilm.  That\'s worked just fine except that our microfilm camera is as old as the hills and we are in desperate need of a new system.  So rather than buy a new microfilm camera, we\'re going to start living in the 21st century and switch to \"imaging\".  We\'ll be scanning the documents that are filed with us and storing them as digital images on the computer.  I am so excited because I\'m a total computer nerd and this is right up my alley.       I\'ve been experimenting with a scanning system for the past week and although it\'s simplistic in comparison to other systems that are out there, I am very impressed with the image quality and potential for what we can do with these scanned images.  Right now, if you come into my office and ask for a copy of a document, we have to make a copy from microfilm and it may not look that great because the filmer might not have checked it that well.  Or the microfilm may show some aging.  But digital images will never degrade!  Plus I can make you a copy of a document without getting my backside up out of my chair!  :)  Am I lazy?       Another benefit from going to digital imaging is the storage room we can reclaim.  Right now, our office is overrun with huge books from the late 1800\'s and early 1900\'s.  Then we have many cabinets full of microfilm.  If we can convert all that (eventually) to digital images, we\'d have more room for the new cases that are filed with us.       Upon coming home from a nice, busy day at the office, I was greeted by an almost paved road.  The County is working to have our road repaved and buddy, it needed it!  It\'s had so many potholes for years.  They\'d usually just come out and patch them up.  And we all know what happens to a patch job on a road.  It lasts what, two months tops?  At any rate, we\'re finally getting a brand new re-surfaced road.  Yay!       I also received a message that a friend of mine had her appendix removed today.  Apparently she began hurting yesterday (on her birthday of all days).  She went to a hospital in a city that I know nothing about.  I\'d like to go see her, but I\'m afraid I\'ll get lost.  She\'s supposed to get out tomorrow night, so I don\'t feel so bad about not seeing her.  I\'ll check with her when she gets back.  Poor girl.  She\'s had more bad luck with health.  First she was diagnosed with Crone\'s Disease, then just recently got over a bout with Pneumonia and then some bronchial problem and now she\'s gotta deal with this.  Good grief!  I\'d certainly understand if she were to get depressed.  I know I would be.  But she still keeps going!  I\'m glad she has the will power to keep going.  '),(13212,4055095,'16,August,2004','2004-08-16 00:00:00','   The walls are thin in the apartment complex that I live in.  They\'re not too bad, but sometimes I can hear my neighbor\'s TV and one night he kept me awake until 2:00 am having a party.  He was unresponsive to my beats on the wall.  My neighbor is relatively new (less than 6 months).  He acts as if he\'s the only person living in the building.  He always seems to wake me up somehow, whether it\'s talking on his cell phone right outside my bedroom window or having his friend pick him up at 5:00 am and they honk their horn to announce their arrival.      This evening I heard the typical \"TV noise\" coming from next door again.  But this time, it sounded louder than usual.  So I couldn\'t resist.  I had to see what this guy could be watching so loudly over there.  I did it.  I stuck my ear to the wall.  But I quickly realized that I wasn\'t listening to his TV.  I was listening to him berating his son.  When you live in an apartment, you kind of gather how your neighbors\' lives work just by seeing them come and go on a daily basis.  Although I don\'t know him personally and have never spoken to him (wouldn\'t want to anyway, the jerk), I\'ve gathered that he\'s probably divorced and gets custody of his son periodically.  I\'ve seen them coming back from the swimming pool and getting in to his truck.  His son looks about three or four years old.      As I listened to him I realized that his son was crying.  Then it sounded like the guy was mocking his son.  He kept asking him something, maybe he was asking his son why he was crying.  I heard the boy say something unintelligible and then the guy said something like \"I didn\'t call you a girl, I said you were acting like one\".  There were some words said that I couldn\'t quite make out.  Then I heard the boy crying pretty hard, in that sort of scream/cry that small children do when they\'re upset.  He was crying out \"daddy\" over and over again.  I have no idea if the guy was hurting his child physically, but I certainly don\'t approve of the verbal abuse I heard.  Some people may not think anything of it, but I think verbal abuse (no matter how mild) is just as bad as physical abuse.      I didn\'t like the idea that he was telling his son he was acting like a girl.  Being a gay man and growing up in a society, religion and town that pressured me to hide who I was makes me especially mad when I hear fathers calling their sons names because they might be doing something that they find feminine.  Besides, what\'s so feminine about crying for God\'s sake?  He\'s a child!  That\'s what children do (and newsflash... that\'s was mature adults do) when they\'re upset.      Right after I heard the boy calling out \"daddy\" over and over again, my phone rang and it was my sister.  I unloaded on her (lucky her) and she gave me some reassuring advice.  She\'s a nurse and said she\'s had to call Child Protective Services on many people when she sees certain circumstances in her hospital.  I don\'t know if I should call CPS or not.  I\'m going to wait because I have no proof that anything happened other than the child was upset.  Maybe the boy was cranky and tired and they weren\'t getting along with each other.  I know I\'ve gotten upset with my nephew when he was really young and wouldn\'t stop crying (of course I didn\'t call him names).  It\'s only natural to lose your cool sometimes, but I didn\'t like the way the guy seemed to be mocking his son.  Doesn\'t he realize that children look up to their parents and copy everything they do?  I catch my nephew trying to do everything I do and everything his mom and dad does.  That\'s why we have to try to be good examples for him so he hopefully won\'t develop the same bad habits that we have.  I realize I can\'t save everybody and that verbal and physical abuse go on daily.  But it makes me sick to think about it.  This guy has no business having a child.  I won\'t even go into the fact that the guy is basically a child himself.  He\'s still immature.  He has friends over late on the weekends and they drink and are generally loud.  He\'ll come home at 4:00 in the morning sometimes.  This was probably an unwanted pregnancy and now he feels trapped in this situation and maybe feels resentment toward the child.      I\'m going to keep an eye and an ear out and if I have to call CPS, I guess I\'ll have to call them.  I just didn\'t want to be the nosy neighbor who won\'t mind his own business.  But a child\'s well being may depend on it.  '),(13213,4055095,'01,August,2004','2004-08-01 00:00:00','   Well then, just this past Thursday in my office (remember I work in our local County Clerk\'s Office), we had a man come in who was wanting help with his property.  We\'re familiar with the man.  He comes in periodically and gets copies of his Service Discharge.  I don\'t know what he does with all those copies, we simply indulge him.  You can tell that he has some sort of mental condition.  Perhaps he was in a war and has mental damage from having to cope with it or something like that.  Who knows?  You can just tell, usually, which customers have mental disorders once they begin talking to you.  We don\'t treat these people any differently.  We try to help them the best we can, but a few of them are too far away from reality, haven\'t been taking their medication, or maybe haven\'t been fortunate enough to be able to seek medical help for their condition.  I don\'t feel we should fear people who are like this because I think it could happen to anybody and I wouldn\'t want people treating me badly because I wasn\'t making sense.  I feel like we should treat them with as much respect as possible, but there are lines that can be crossed.      Back to the man who was asking for help with his property...  Everyone in the office noticed he was being much more \"wordy\" than he usually is.  He usually says very, very little.  But then again, we always know what he\'s coming after (his Service Discharge), so he doesn\'t need to say anything except for how many copies he wants.  The girl who was helping him seemed to be trying to find what he needed, but she wasn\'t doing exactly what he wanted.  After he walked out, obviously agitated, the girl told us what had happened.  She said that he was complaining that the co-owner of his property was trying to take the property away from him and he needed to know how to get the co-owner\'s name off of the deed.  Well, our default answer for that question is \"you need to seek legal advice from an attorney\".  The County Clerk\'s Office is not there for legal advice and to explain what documents mean or how to go about transferring titles, etc.  The County Clerk\'s Office is essentially the library for the County.  We store things and keep records and can retrieve what you need if you tell us what you\'re looking for.  But people expect us to know all about property transactions and Probate proceedings.  And they expect us to be able to explain all the ins and outs about how they work, etc.  We know the things that pertain to our office and that\'s it.  If we knew all the other extra stuff, then we\'d be freakin\' attorneys for God\'s sake.  But you don\'t see us charging people $150 an hour while they\'re standing at our counter wanting service.      The man didn\'t like the \"you need to talk to an attorney\" answer.  He also proceeded to tell the clerk about some incident where the police came into his home.  He was mad about that and I guess wanted her to do something about that as well.  When she couldn\'t help him, he said \"then you\'re going to get shot\" and left.      We were on edge for the rest of the day wondering if this man was going to come back into our office with a gun.  We\'ve actually had another man who has a mental disorder come up to our office with a gun and was caught outside in the parking lot with it.  (That guy had a history of threatening people, but usually with stupid idol threats.)  Our courthouse security was on alert and we constantly had an officer within our sight.  Our bailiffs are outstanding for responding to problems and I am thankful that they\'re in the courthouse with us.      With that said, I wonder why our County spent all the money that it did on remodeling the courthouse and putting in extra security measures for the judges, yet didn\'t bother with remodeling our office and added no extra security measures for us other than installing a panic button.  Mind you, I\'m thrilled about the panic button and as I\'ve stated, our bailiffs are great at responding.  But the judges are locked away in rooms that require you to enter a code to open the door.  And in their courtrooms, they\'ve had bulletproof shields installed on their benches so they can duck behind them if a defendant happens to bring a gun to court (or get a gun off of a bailiff).      I don\'t want to be locked away behind some door with a push-button lock.  We\'re supposed to be there for public service.  But what do we do if someone comes in with a gun?  We\'ve discussed this in our office and asked each person what would they do.  Of course, you can never know exactly what your reaction will be if something like that happens (unless you\'ve been trained, like law enforcement).  Perhaps we need to devise a plan in case something like that happens.  I\'ve personally been in a threatening situation where a girl was attacked in my front yard at 2:00 a.m. and she came to my house for help.  I reacted in a completely different way than I thought I would.  I wasn\'t thinking clearly and was scared.      I\'m not trying to diminish the potential threat to the judges.  I know they\'re the ones who dish out the punishment to criminals and who rule against people or take people\'s kids away from them.  So they really piss people off.  I completely understand why they would feel threatened.  I just feel like we clerks aren\'t taken seriously.  As if we don\'t matter.  We deal with the people.  And for the most part, the people are calm, cool and collected and we can help them with what they need.  But our door is also open to the person who\'s not in the best mental state.      I suppose there are no guarantees.  Perhaps I\'m asking for too much.  After all, I\'m not guaranteed to return home safely when I leave my home to go to Wal-Mart.  I could be hit by a drunk driver or someone could go ballistic in the store.  After a few weeks, the threatening incident will be pushed to the backs of our minds and we won\'t be so worried.  But when things like that happen, they make you stop and think about how easily we can lose the life we\'ve been given.  '),(13214,4055095,'01,August,2004','2004-08-01 00:00:00','   Stayed up til almost 2:00 a.m. this morning (I never stay up that late).  I went to my sister\'s house and watched a movie with her, her husband and our nephew (who was in town visiting a friend).  We watched \"Go\" again.  My sister loves that movie.      Ever woke up and felt like someone has beaten the crap out of you?  That\'s the way I\'ve been feeling the past couple of mornings.  I\'ve slept really hard and probably didn\'t move an inch all night long.  But then when I wake up in the morning, I feel all haggard like I could probably sleep some more.  Yet I can\'t go back to sleep.  So off I go in inject my body with Starbucks Breakfast Blend.      After getting out of bed, I decided to pay a couple of bills and get those out of the way.  Rent\'s due today, but since it\'s Sunday, I won\'t get it to them until tomorrow.  I\'m now paying all of my bills online, with the exception of one.  It\'s such a breeze.  And I no longer have to worry about whether or not our wonderful postal service will get my payments to their appropriate destinations.  I don\'t mean to knock the U.S. post office because considering the sheer amount of mail they handle daily and the low cost of mailing a letter, overall they do a great job.  But I\'ve been screwed one too many times by the postal service to warrant my distrust.  The only bill I can\'t pay online is my car payment and that sucks.  Bank One won\'t allow me to do online payments unless I have a checking account with them.  Whatever.      After paying those two bills, I was a little surprised to see my checking account getting kinda low.  I\'m not saying I\'m rich or anything because I definitely am not.  But I like to keep a little bit of a \"padding\" in my checking account, just in case.  Not too much, just a few hundred extra dollars.  (I was able to do this by cashing in my 401k when I quit a previous job.  I was young, what did I know about 401k\'s?)  I think I\'ve indulged a bit too much now.  My padding is slowly wearing away!  It\'s making me think hard about whether I really need to be paying $24.00 a month for Netflix.  Once you get used to living with certain luxuries, it\'s hard to go back.  You begin to feel as though you need them and can\'t live without them.  I\'ve forgotten what it is like to live without broadband internet or movies being mailed to me.  I\'ve become spoiled (or more spoiled than what I was).  '),(13215,3354828,'29,April,2004','2004-04-29 00:00:00','    I hope you guys can read this, I just wanted to say that the WASL, was A LOT easier than I thought. So it\'s ok. Everyone was all making a big deal about it. Maybe the second half is harder we have 3 periods straight taking it. So now we have another 3.   '),(13216,3354828,'27,April,2004','2004-04-27 00:00:00','    Ugh my throat hurts. After we went swimming me Dawn, and Ian went for pizza, then we came home and watched a movie. And finally he went home and here I am. You all know Dawn in my sister right? Just for those who don\'t know.....I was gonna bring Michelle too but she forgot. Then I was gonna bring Carli, but I didn\'t give her enough notice. Soooo yeah. AND I have a dance this Friday, the last one with the 7th and 8th graders combined. Then we have one just 8th graders at the high school. Woo the theme for the dance is like....A night in Paradise. Maybe I should wear a turban and a toga. Lol, that would be interesting. I need to think of some songs to put down on my cd, but every time I ask someone they say songs I already have....DAMN YOU PEOPLE I HAVE ASKED!! Well I\'m gonna go look for some more songs I like. Lata playas   '),(13217,3354828,'27,April,2004','2004-04-27 00:00:00','    Oook Well I\'m in class it\'s 5th period It\'s gonna be lunch in 11 minutes. And I\'m kinda going back and forth between working and goofing off. No one minds though. I have Jazz combo today after I eat lunch which is when I go in for my lunch break and play music with some other select players. I think I\'m gonna join the talent show and play Jazz Improv. On my clarinet. It should be fun. I wonder if I\'ll be able to do it.....oh well! Well I\'m gonna go, lata playas   '),(13218,3354828,'27,April,2004','2004-04-27 00:00:00','    Argh! For some reason some of my posts aren\'t coming up, so either they\'re showing up on yours or on no one\'s I hope Y\'all read the ones before.....Sooo yeah. *gasp* I waved to you-know-who....I forgot his name, but I was really waving to Carli, it just looked like I was waving to him lol and he waved back like a joke and I gave him a really weird look and he started laughing. So at least now he knows who I am. I think I\'m gonna drop the crush and persue my real dream....=D they know who they are, and I don\'t even know if they\'re reading this lately. Soo yeah. Anywho I\'m gonna go watch a movie, catch ya lata playas/playettes   '),(13219,3354828,'26,April,2004','2004-04-26 00:00:00','    Ugh I\'m in such a bad mood. I think I\'m gonna die because everything is bugging me. I need to like....go to sleep and vent on someone. But it\'s so damn hot in here. It\'s like 79-81 degrees F. Ugh....I\'ve been sucking and rubbing ice cubes on me all day. I kept spelling \"cubes\" \"cubs\" too. Lol. I think everyone knew and didn\'t correct me. Oh well I think I\'m gonna go lie down. Lata playas   '),(13220,3354828,'24,April,2004','2004-04-24 00:00:00','    Oook I\'m running out of colors. I don\'t even know if this is a real color! Oh well it should still show up.....I hope. I\'m in the library since I forgot money for lunch....so hungry..... Well nothings going down here there\'s only two people in here....make that three. Ok four. And there\'s another class in a seperate room.....They\'re 7th graders. I never got to go in there for MY communication classes.....How unfair. Well it was also a different teacher. Since I\'m in band, Jazz band and AVID, I can\'t be in social studies or an encore class. Since you only get two. Which is why I don\'t have social studies. Well I have WASL testing on thursay and friday (Washington Assesment Student Learning) Ugh....It\'s the science WASL, I\'m gonna crash and burn on it. My math teacher was going to show us a copy of the 10th grade WASL, but someone was talking........I\'m not saying who....*cough* George *cough*. Oh sorry, I think I\'m coming down with something....:D Well I gotta go. ^_^   '),(13221,3354828,'24,April,2004','2004-04-24 00:00:00','    Everyone is calling me today but my favorite person, I guess their basic phone service went out lol oh well. I went swimming today...I had a ball. Only since I couldn\'t find my bathing suit, I had to wear a sleevless T and some shorts....and the shirt was white and you could see right through it! Good thing I wore a sports bra underneath....phew. We weren\'t the only people there though.....and I brought some friends. *sigh* they couldn\'t stop laughing....until I made them lol, j/k. Well it\'s time for dinner! Peace y\'all, luv yas ^_^   '),(13222,3354828,'21,April,2004','2004-04-21 00:00:00','    Oh no...One of my friends crushes found she likes him, she\'s crushed! I feel so bad for her, but I can\'t get her to vent out on me. She knows I\'m there for her....Maybe she just needs to be alone for a while.   '),(13223,3354828,'20,April,2004','2004-04-20 00:00:00','    Hey Y\'all, I\'m at school and as you guessed...I have nothing better to do than to write here. Ain\'t I interestin\'? Oh I know. Well I\'m actually on my lunch break. And the 7th graders are in the library also....boy are they noisey. But the one I \'kinda\' like is in here. The one I like because they remind me of someone I really really like. Ya know? Am I the only one in the world that does this.....Ah well, it\'s only looks-wise, I don\'t even know the guy....I\'m weird huh? Well I\'m gonna go do other things.....Bye!   '),(13224,3354828,'20,April,2004','2004-04-20 00:00:00','    Yesterday was awesome! I had such a good time! I downloaded X-Chat and it was actually pretty cool...kinda confusing but still pretty cool. And I got AIM, Yahoo messenger, and MSN messenger onto the computer. So it\'s all good. I have everything I need....NOw let\'s just pray that I don\'t get a virus or I don\'t get hacked or somethin like that. Well.....toodles!   '),(13225,3354828,'19,April,2004','2004-04-19 00:00:00','    *sigh* This computer sucks. Spyware sucks, viruses suck. The only thing that doesn\'t suck is Msn messenger, AIM, Yahoo Messenger, Blogger and umm that\'s it. I think....not sure yet. Oh yeah, FFR, and E-messenger. That\'s it. That\'s all you need. PERIOD. Well That\'s all I have to say. Bye!   '),(13226,3354828,'16,April,2004','2004-04-16 00:00:00','    Well ummm I have NOTHING to write here......just thought I\'d tell you that. I never do. Ummm yeah.....*blink* CRAP! I have math homework to do.....but it\'s really easy. Review of ditributive property...so it\'s all good.....well I better get to work on that....soon....not now....but definitly before it\'s due.....= \\ I hope.   '),(13227,3354828,'15,April,2004','2004-04-15 00:00:00','    Oy it\'s the weekand!! Yay!! ^_^ I\'m soooo happy I have a full weekand, I\'m gonna have a sleepover at Amie\'s and Becca\'s house, and if I can\'t I will next weekand, then I\'m going to the movies with Erik, and finally going to play pool with my mom ^_^ We go every other weekand, I beat her twice last time (FINALLY) the first time I won on my own, but the second.....we switched balls lol. Well I\'m gonna go talk to Nick...........uh huh.........weird..........*types* ok he had to go lol. Well.....bye =)   '),(13228,3354828,'15,April,2004','2004-04-15 00:00:00','    *sigh* well it\'s friday, I\'m at lunch and I have about 10 minutes to write this. My next class is Jazz band but in Concert Band our director (Ms. McFeely) Gave up and said she was done for the day....and we waited in class for the last 5-10 minutes waiting for the bell to ring.....It was scary. Now I\'m kinda scared to go back =S. Well....Wish me luck!   '),(13229,3354828,'15,April,2004','2004-04-15 00:00:00','    Btw Astricks or whatever they\'re called the little star thingys (*) By a name mean it\'s been changed for mine or someone elses convinence, thank ya ^_^ See ya lata   '),(13230,3354828,'14,April,2004','2004-04-14 00:00:00','    Omg.....Dude I know no one\'s reading this but this crazy guy keeps walking past my house and throwing his hands in the air like he\'s worshiping god and maybe he is...in some freaked out way I dunno, and like he\'s acting like there\'s snakes all over him or somethin....I dunno it\'s weird....Just HAD to write it somewhere. Well *yawn* I\'m going to bed....not that any of you care ^_^ Peace   '),(13231,3354828,'14,April,2004','2004-04-14 00:00:00','    Oy, today was harsh, the days go by soooo slooow when it\'s close to summer. Jeez....two more months exactly in 4 days. I don\'t know if I can last! =(   Language Arts Well we didn\'t do anything in that class what really went down was passing time (the time I get to go from one class to the next)  Language Arts to Math Passing Time (4 minutes after class ends) Well I saw Andrew* and Josh*. Krista likes Josh * and I kinda like Andrew*, But I only like Andrew* because he reminds me of someone else I really like....only in looks. I don\'t even really know Andrew* = \\ Weird huh?   Math We took notes on this whole other theory, irrational numbers, which we\'ve talked about but havn\'t really looked into. And I finally understand what they really ARE. And we got a really long homework assignment. 6-26 even numbers only. and we have a review sheet. We actually learned all this stuff last year. Mr. Roberts is just trying to get us ready for high school.  Passing Time Math to Science (4 min.) I saw Anndrew again, but he didn\'t see me. I saw Krista look at Josh* too, it\'s kinda funny when she asks me all about him, when I really have no clue who he really is. Odd I know.   Science God this class is boring, well not really boring, the teacher just makes it boring his name is Mr. Moore btw (A.K.A Mr. Snooore) we\'re just filling out the blanks in worksheets and he\'s giving us the answers. But me and Krista actually just wrote notes back in forth since we sit next to each other every day. The notes were mainly about \"Who likes Krista\" \"who Krista should go to the dance with\" \"What she should do\" And I give her a lot of advice, which is weird, since I need to listen to the advice I\'m giving, so I won\'t make fun of her if she can\'t do anything about her crush, since I wouldn\'t be able to do anything about mine. But I can talk to her crushes just fine....maybe she could talk to mine. Maaaaybe not, I don\'t think I want them talking to Krista, I\'m a very possessive person.   Passing time from Science to Concert Band Lol I said hi to Austin, who is a seventh grader that Krista knows, and I only know his name because we went to the same elementry school, only....he doesn\'t know mine lol, so I just mess with him ^_^ Quite fun really.  Concert Band We got a new piece and she can\'t decide between these two pieces one is called Quintero the other is like.....Shining of the Stars or something....I like Quintero better because I like playing in 5/4 time.   Passing time from Concert Band to AVID Well nothing really happened here, I said hi to Krista, Carli, Kelli, Ian and some others that I saw. But that\'s about it  AVID We had a guest speaker today. A congress man for our district, He was kinda cool, and I was a panel person (there was like 10 people out of the class that actually got to talk to him...all hand-picked) Which reminds me, I have to write a thank you letter to him. He talked to me after class, he congragulated me on something only I don\'t really know what it is....ah well.   Lunch I tried to eat fast and get to the library and meet Krista and Alisha, we were using the computers together to check our e-mail. We wrote back and forth for a while until 5 minutes before the bell, when we went outside and waited for Ryan to come out, Only when he did, it didn\'t go according to plan, so we\'ll probably do it tomorrow.  Jazz Band Ohhh this was really fun, we started all new stuff, we got two new songs, One is like Pamalaya. the other is......Pink Panther! Which we begged her to get. WOoOoOo, after that we worked it out, but I had to switch with the other Tenor Sax player (Michelle) Because she can\'t play the \"low notes\" see, I had the first part, but she took it. Oh well I still have a cool part either way, so it\'s cool.   Passing time from Jazz Band to P.E Oy nothing happened here, I just....walked to the gym lol.  P.E All we did here was another pacer test (last of the year!! WOoo) I beat my old score, Krista beat her brother,and ummm Becca and Amie are totally awesome, they seem to think i\'m the funniest person on the planet....weird....well i\'m done woooo   '),(13232,3354828,'18,May,2004','2004-05-18 00:00:00','    Wooo I\'m not going to school today! I\'m supposed to go on this field trip to WWU (Western Washington University) But I told my mom I don\'t want to go. Even though I don\'t have choice since I forgot to turn in my rough draft and I can\'t go anyhew. Oh well....At least I\'ll get to stay home and sleep in.....Wooo! Well I\'m going back to bed, night!   '),(13233,3354828,'16,May,2004','2004-05-16 00:00:00','    urlLink         I am Edea!  Selfless and stong-willed, I have the ability to do a lot of good. Luckily for the world, that\'s usually my intention. If I were ever corrupted by evil, you would all be in serious trouble. On top of all that, I\'m pretty sexy.    urlLink  Take the Final Fantasy 8 Test here!            My #1 result for the SelectSmart.com selector,  urlLink  Your Real Estate Match in Middle Earth  , is  Lorien   '),(13234,3354828,'09,May,2004','2004-05-09 00:00:00','    Ugh, my stomach hurts, I\'m on this new thing with my friends so they aren\'t left out.....lol stupid friends....I\'m just gonna cheat behind their backs....lol no. Anyway..... have two things to put in here, one another poem, and two a comment....first the comment....:        You know what I hate (and this is strictly for girls) I hate it when guys don\'t get the hint.....because after they don\'t you just can\'t say it....so it sucks. Maybe I need to change my ways....        Now....for the poem....:         SISTERS          We\'ve been through a lot We\'re sisters til the end Sticking together, through thick and thin  We\'ll take care of each other Watch each others backs We\'ll take it head on And conqur any task  We\'re sisters and I\'m proud To have you next to me To stay next to you And feel safe, snug and sound         Okie dokie, I\'m done, later!   '),(13235,3354828,'06,May,2004','2004-05-06 00:00:00','    I havn\'t posted anything in a long time...haha I got in a fight with Kirsta\'s brother. He\'s a huge butthole....but forget him. There really isn\'t anything I can put here since only like......3 or maybe 4 people are reading this site. But you know what....that\'s ok because I can vent here. AHH! I lost my Personal Dragon Rough Draft, and I have a rough draft due on the 18th...and I havn\'t even started. And internet explorer keeps freezing....and they\'re so much better than me :\'( and I\'ll never be that good...and I miss her....and I miss him a lot....and I don\'t want to cry but it\'s to late because I\'m so confusing and no one knows what I\'m talking about because I\'m not explaining anything....but it\'s ok.....because I still have music....it\'ll get me through.  Ok well bye.   '),(13236,3354828,'02,May,2004','2004-05-02 00:00:00','    I write the weirdest stuff when I\'m alone.....check this out      Death is better than this feeling Why won\'t someone tell me what\'s going on I\'m to young to feel this way I need help I\'m scared and all alone Help Why aren\'t you helping me? Please....    I could write forever and it wouldn\'t mean anything to anyone but me.    Love hurts They say it\'s not supposed to be painful but it stings.....  True love is not real Not anymore It died along with the GOOD expectations of looks and old fashion ways everything has changed nothing is the same       Woo all done. See ya guys later!   '),(13237,3354828,'02,May,2004','2004-05-02 00:00:00','    Gah it\'s like 1:47 AM and I can\'t sleep. I know why but......I don\'t know how to change it. Tch and I\'m having an allergy attack. Oy......and it\'s so hot......and someone is gonna cook! I won\'t be able to sleep.....I better go now.....lata playas   '),(13238,3354828,'16,June,2004','2004-06-16 00:00:00','    Woo School\'s almost out, Thursday\'s my last full day then I have my \"graduation\" on Friday. It\'s just from middle school to high school...nothin\' big. Ah well! We shall see....your guy\'s comments (not that you leave any) should be how you would sign my yearbook....keep up the good work!   '),(13239,3354828,'22,July,2004','2004-07-22 00:00:00','    &nbsp; I&nbsp;just finished reading Abhorsen (last of it\'s trilogy) And I have to say....Sabriel, Lirael, and Abhorsen are my all time favorite books. *sigh* I\'m kinda bummed out that I\'ve read them and they\'re over....I\'ll have to read them again sometime. Which reminds me...I still havn\'t finished Eragon....I\'d like to sometime. But I\'ll have to get the book from somewhere. I liked that book quite a lot too. Anyway....that\'s all I have to write for now. Catch ya all later. Peace =P   '),(13240,3354828,'21,July,2004','2004-07-21 00:00:00','    &nbsp; Dude...they made this a lot easier, lol. I&nbsp;really didn\'t need it like this but it\'s a lot better.    &nbsp;&nbsp; In case you guys didn\'t know what I was talking about, they now have a drop box for fonts and colors. Instead of having to enter the tags and stuff...And now you can enter pictures directly from your thingy instead of going through an ftp. Which...I think you had to do, but I\'m not entirely sure.   '),(13241,3354828,'19,May,2004','2004-05-19 00:00:00','    Ahh yes, Halle\'s new movie. Not much to say about it. It was just a side thought about the woman\'s outfit. It just looks like it would be....difficult...to take off and put on. Whatddya think? Well just send feedback and I\'ll see what ya think. Later   '),(13242,3354828,'18,May,2004','2004-05-18 00:00:00','    urlLink SelectSmart.com Lord of the Rings Selectors Quizzes Tests Decision Makers Select Quiz Test  These are actually quite fun, there isn\'t a lot of  spam , and they\'re tons of catagories. Two thumbs up on this one if you get super bored, and if that isn\'t enough there\'s always http://www.bored.com Which includes various activites including prank e-mails to send to your friends (which is loads of fun :D) Well, that\'s all I have now, later!   '),(13243,3354828,'18,May,2004','2004-05-18 00:00:00','    urlLink Welcome to Neopets!  This site is kind of childish, and very unrealistic, it\'s fun for a while but grows old very fast. The whole plot isn\'t very intriging, but you get the point.....   '),(13244,3354828,'18,May,2004','2004-05-18 00:00:00','    *sigh* yes, I made another one. This blog is all about teen ideas (just like the other one Teenage View) only this one is for less serious subjects that won\'t cause as much contreversy ;) Enjoy! And please send back some feedback! Thank you =D   '),(13245,3354828,'22,July,2004','2004-07-22 00:00:00','    I HIGHLY reccomend these books if you\'re into realism and fantasy. It had magic but it\'s so well written you\'d believe it =P They\'re very enjoyable and definetly page-turners. Check \'em out and have a blast!!   '),(13246,3354828,'02,July,2004','2004-07-02 00:00:00','    Aaaall right. There\'s some picture going around with Joel and Hilary Duff kissing and etc. Even thought it might be a publisity stunt. The word is that they\'re going out. Which is very very wrong since they\'re about 9 years seperated. Sooo any comments? Pleeeease give me some feedback. Thank you ^_^   '),(13247,3354828,'02,July,2004','2004-07-02 00:00:00','    Actually I usually don\'t reccomend other bands....but this one\'s rather interesting. I\'ve only heard one song, \"Only for the Weak\". And I\'m not much of a fan of the vocals. But everything else rocks. Check it out. *one 1/2 thumbs up*   '),(13248,3560141,'30,June,2004','2004-06-30 00:00:00','   It\'s amazing how the mind can play tricks on you, and even stranger that because of medical school, I can probably deduce that these mood swings are all a product of some tiny neurotransmitter, some tiny hormone that binds to some receptor causing all this havoc in my head...  I\'m in some sort of weird mood, where I even got defensive at the beauty salon...the manicurist said, so...besides your eyebrows, do you think you might want to get your nails done...I snapped in my own stupid way, by curtly saying I like them this way, short, and with dirt under them....  Can it be possible to feel much more inclined to see yourself as retarded (no offense) more so than you did as a teenager?  Someone said to me yesterday something which she may have meant as a quippy remark, but it seemed to hit it on the head...as a teenager we think, \"we know everything..\" in our mid 20\'s we realize we know nothing about life, politics, IRA\'s, or our futures, and in our thirties is when it all comes together...in equal and fine balance...  I am too consumed with establishing myself now, to notice world issues, or even plain and common everyday entertainment news, or even be concerned with our own good green earth...when I was given the opportunity to hear it all beforehand from my parents in my teens my ears held up picket signs in silent rebellion, I am hoping my thirties brings complexities which I feel are manageable, and are worth asking my parent\'s advice on with all the rioting in my ears being over and all....  I\'m just uneasy about something, and I don\'t know what it is???  I wish I could fall asleep really early today, and just wake up happier...  '),(13249,3560141,'29,June,2004','2004-06-29 00:00:00','   I went to baggage claim and retrieved my checked in luggage, not before some old lady had already began unzipping it to make sure it was hers...why did she have to open the one compartment with all my undies in it...I mean the embarrassment was just...well expected because what else would happen to someone like me???  I started to say, the luggage it was lighter I left something big in Miami...huge in fact.. I landed and the humidity in Miami just clung to my thighs as if to say welcome home...He pulled up and ofcourse had a mild confrontation with some pedestrian traffic controller on a head trip..confrontation settled with some fancy manual transmission maneuvering...which trunk would I go to I thought.  How could that be the first thought I have..I thought...I was thinking too much, much the way a puppy gets when you show it a chew toy and a treat...simultaneously...  Oh my god to have his touch meshed with his words after soo long...all my anxiety melted into his maroone leather seats, either that or the humidity sticking to my thighs just condensed...either that or now I was thinking about all the meshing that would end up happening...hahaa  We spoke so openly, maybe influenced by alcohol, maybe I\'m just letting go of these barriers to getting hurt I don\'t know, but I wanted him to know that our distance may be good for us now, that every hurdle he or I ever placed along the track to where we are RIGHT NOW...was worth it...that even if this is where it plateaus...I could never ask for anything sweeter.  We laughed, laughed so hard he almost left with a concussion after knocking his head into the bannister behind our table so many times, we ate and drank at these beautiful places but it was his skin and his lips that tasted like perfection..So many close encounters with the threat of being arrested for lude and the other \"l\" word acts in public, a lot of meshing, exciting, tantalizing, sensual meshing...  I saw my babies too, the whole clan..it was amazing...those women I have as friends they had me rolling on the floor, and I know that no matter where time or planes may take us all away to, I will always see their warmth, they looked at me differently you know like with a lot of respect and maybe even pride...It also felt really army like because well, we all had on creme pants and pastel tops for Divvy\'s bridal shower, we were on a mission for pure fun and quality time, Mission accomplished...on to the wedding...I hugged her with all my strength, I told her in a card ofcourse, because sentiment rolls off my tongue with an anchor attached to it...straight to the floor, my belly says this is a good thing...that feeling we all get when a family member or family like friend meets someone new...you get this peaceful or disturbing rumbling in the tummy, like is he going to hurt her, is he going to respect her...my belly had the fizzes, alka-seltzer style, very relieving the moment Mr. Jefferson looked at Divvy, he was enamoured lost in her, I loved it, I felt she was safe, I didnt have to worry, it\'s not like my worrying would protect her like a bubble, but it\'s selfish I could close my eyes and know she\'s going to be ok forever...and ofcourse I had to include the staple side of sarcasm by saying it could\'ve been the fact that I had diarrhea...but otherwise it\'s all systems go....she laughed her contagious giggle and I remembered that she knows how much I care..  We read the paper together, that was amazing, I\'m the biggest weirdo...these god forsaken old couple things are appealing to me...I like fighting over sections of the newspaper, having him look at me quite perplexed as to why with all my might I could swing and completely miss the golf ball...Brunch was dreamy...he perfectly combined food and flattery...he knows too much, including my ninja status...The proudest moment for me was when he swept everyone off their feet including some potentially fake British guy with his amazing wit and charm...I got annoyed at one point, because I wanted it all for myself...I need to take a stab at sharing...One of the partakers in the being swept off of their feet activity, squeezed my hand and said he is so amazing, look at how happy you are, you guys are beautiful together, a comment which we heard in NY, and also from the Brit...I always let these comments roll off my shoulder (on the outside) right into my heart...I just reached around him and squeezed him like an empire waisted cloak...We discussed in crude terms my needing to take a stab at sharing, and letting go of insecurities...How can he be younger, the farthest from thinking like we\'re in a pseudo relationship, yet he is the one who makes all the sense..who says things that really just make sense???  He has brought something out of me that hasn\'t surfaced since a wild autumn night in 1997...I wanted to share everything with him, wanted to take care of him, not just party my ass off, but make sure he eats well...See, I\'m a natural babier or motherer whichever makes more sense, it\'s a little bit of the humanitarian in me but also a lot of the narcissist...like \"I know what\'s best for you.., don\'t worry I\'ll do it..\", these last few years I have been self indulgent, like fulfilling my every whim for selfish reasons, but for all the jokes I make about him being a full grown baby who doesn\'t have salt in the house :), damnit I almost bought salt at the grocery store...Even if this ends up being the demise of our whimsical friendship/love affair, I am so happy to see that side of me reemerge, because I was banking on being the really cool aunt for my brother\'s kids who orders take out, travels non stop, and drinks martinis after a hard day, maybe now I see I have room in my heart for other people...  Enough..I\'m made myself gag now...  '),(13250,3560141,'23,June,2004','2004-06-23 00:00:00','   Days away from going home...where by the way, I nor my family owns a home....that\'s the only definition I have for a place like Miami, where I spent at least 5/13 of my life...I\'m nervous.  I\'m going to see him, my friends, one of which is scheduled to take the big plunge on the 31st of July, so a bachelorette-ing we will go.  Speaking of which all this talk of weddings (the Armenian is headed for the altar in September, another wedding to go to) has got me in a weird mood.  I\'m starting to look at babies and couples and all the \"Family Circle\" like cover pictures that pop up everywhere.  I am fighting that biological clock with a wrench, I mean I\'m in no position mentally, emotionally or geographically to say I\'m ready to settle down.  I barely know where I will be in one year much less a lifetime, that\'s the catch to this \"last resort\" med school option I\'ve chosen.  My residency could have me in the Tristate area, to S. Florida to Hawaii, if it were all up to me.  But those cheeks and little hands, and the young power couples with the little teacup dogs, pale green walls, wood floors and dark wood and steel furniture, god, this is really starting to appeal to me.  It\'s not a little but a lot weird, because of me because within the last few years I\'ve changed so much.  The Armenian and I were smoking cigs (a true sign that I may not yet be wifey-ready)and he told me that he used to fight this one demonic image that kept popping into his head....waking up grabbing a half put out cig, and a bottle of black label and that\'s how he would commensurate his day, how now, that demon has left his being...Well...my eyebrows raised, I looked at my toes, looked into the house where a pile of papers and books crowded the coffee table and thought....   As recently as 2001, I was struggling with the demon in actuality, how lost I was...I can\'t imagine the same disregard for my future, the same blur of hangovers and strobe lights, overdue bills, just lost...  I want a test run with a baby puppy, I want to see if I demonstrate enough patience, the love portion, let\'s just say I\'ve been waiting a lifetime to give it all away...Anyone who knows me a little too well, would probably laugh at the thought of me driving a van with little kids screaming, actually they probably could picture me driving a bus with little kids screaming, but point being...I get befuddled (new favorite word) because I want to sit on the floor and play with tinkertoys, not by myself with a little kiddie, oh my god my kids are going to have to be raised in a very old school house because technology surpasses my grasps, hopefully their father will know a thing or two about X-cube or game box or whatever....  Oh...the father, well...it would be dreamy...safe would be the word...I don\'t know but I wish I had more than a wrench to stop the seconds from ticking away...  '),(13251,3560141,'21,June,2004','2004-06-21 00:00:00','   What kind of pet would you own or do you own?          Doggies         Kitties         Fish         Hamsters         Chia Pets (just don\'t like fur)          &nbsp;&nbsp;           urlLink  Free polls from Pollhost.com  '),(13252,3560141,'19,June,2004','2004-06-19 00:00:00','   I guess I didn\'t write about this earlier because I needed to process yesterday.  My father drank two beers at Sushi Samba, Asahi\'s mmmmm....22oz. I guess you could say he took down a forty, my old drink of choice a very looooonnnngg time ago, I\'m not embarrassed it\'s the meager beginnings of the alcoholic I strive to be these days..by the way, I\'m so old, I can\'t even drink a glass of white wine, without feeling the \"buzzes\"...  Wow, I even write the way I talk when I get nervous about something, first base, second base, hockey, boxing, polo and then back to home plate...I felt really sorry for him, my father, really really sad, I was so befuddled by these feelings that I chain smoked this morning without pacing, no hesitation...didn\'t help.  In his semi-drunken stuper filled lunch conversation and stroll through the park, my poor father spilled his ego onto the table and then onto the green lawns and then onto me....what I mean is at first I tuned him out, because he started out quite normally as he usually does, my speech went like this, I told these jokes, these people were impressed...but then he said it the sentence that ripped me from the porno channel running in my head (i know...I\'m looking into using the CAGE questions to assess my level of addiction)....He says, \"I\'m nothing D, I don\'t have anything, I make nothing for the amount of work I do, I am a nobody, I will never be in my lifetime what I have always dreamed of becoming...this is what I wish for you.\"  I didn\'t know people like my father had an insecure cell in their bodies, and if they did they would somehow have it transplanted...is what I thought...I processed this sentence over and over, clicking away at any and everything even when my internal memory was full I just kept clicking..indicating my internal memory was overloaded.  He proceeded to say how he has to project this image, sometimes it consumes him, he isn\'t white, he\'s relatively new to the lucrative private practice business..he is within the company of men who own ten clinics worth over 5 million each, don\'t get me wrong D, he says, I make a lot more than the average person, but I will never be worth several million dollars..\"...my father has been eaten alive by the green headed monster who is insatiable...I hope it spits him out alive and whole...  Several million Dollars...oh my god...these are the likes of men who own homes that face this park, that span the city block we just travelled through...I just wanted my father to own a Rolex, drive a really nice car, wear really nice shoes, have money and power that would allow him to take a break finally, not retire, he would collapse from boredome..a home with jasmine bushes, coconut trees, and horses everywhere - you see his is still the proud Indian villager who wants his village to have proper health care schools and basic sanitation - near the ocean, with huge glass windows, I wanted him to be with my Mom, and bring her, her most sought after material wish, him...I wanted him to open a small production studio, where he could make the Indian cartoons he has dreamed about in his head since telling me stories before I slept back in 1989...I just wanted him to work hard enough to be happy...see I know my father better than i let on...I try to create this distance, this war because I am so scared I will become eaten too, if this monster has the taste for the same kindred blood....  He wants to build a dock where I can launch his ship of a thousand dreams, he sees the creative fire in me, the love of nature, the humanitarian, the thirst for the finer things, although I would never buy a purple pin strip suit - our tastes for finer things is where we split paths - He thinks I can be what he never could...doesn\'t sound extraordinary..most parents do....but it\'s the fact that he feels so defeated despite his worth as a father, husband, doctor, and human being.  He feels he can\'t ever stop, because he still falls short, never the holder of more than the \"image\", what he wants is to be holding the camera, creating the image...I just feel really sad, that I haven\'t taken more time out to pat him on the back, to tell him that he doesn\'t need anymore, and that I will make a majority of his dreams come true and I will take care of him, and that he is a great father....  '),(13253,3560141,'19,June,2004','2004-06-19 00:00:00','   As one can see...I bought a new camera, a new really old looking camera, because I was sick and tired of being outbid in the last ten seconds of every e-bay auction.  It started with the Olympus mega outdoor camera with waterproofing and perks galore, well, now I have this Kodak brand camera with barely a megapixel and it holds like ten pics on internal memory, so my adventures through Central Park yesterday would only be captured for the first two hours, because by then the battery died and my camera had stored its last pic...Father\'s Day went well...odds in my favor...  '),(13254,3560141,'19,June,2004','2004-06-19 00:00:00','   urlLink    this was right near the port -a- potty at the street fair&nbsp; urlLink  '),(13255,3560141,'19,June,2004','2004-06-19 00:00:00','   urlLink    reckless abandon, I wish I was historically in tune with Marti..&nbsp; urlLink  '),(13256,3560141,'19,June,2004','2004-06-19 00:00:00','   urlLink    this guy was really goood...&nbsp; urlLink  '),(13257,3560141,'19,June,2004','2004-06-19 00:00:00','   urlLink    who would have thought...&nbsp; urlLink  '),(13258,3560141,'19,June,2004','2004-06-19 00:00:00','   So...tomorrow\'s D-day, Daddy\'s Day and to some doom\'s day too...well, I\'m 50:50 with this one, tomorrow could potentially go really well, or I could end up coming home with that look, and head straight for the Marlboro pot the Armenian and I have created...that look of why me, why didn\'t I go with him to get Indian food at the temple (his absolute favorite is South Indian food) instead of going to get sushi (his second favorite food), now he thinks I have absolutely no organizational skills and have no family values because I didn\'t cater to his FAVorite instead of his second favorite...  I wonder where my neurotic - isms come from?  but how amusing that I know this is what I could be subjected to and this is what I could end up feeling, and that it should be temporary now that \"I know better\", but it\'s still my pops, his approval matters so much...  You know I can\'t be around too much perfection because of this guy.  The more decrepid, confusing, unorganized, or desheveled something is, the more solace I find in it.  I want to see people\'s faults and love them for it, but who can love faults right...swindled again..I still end up being the complainer of all that is not what I expect, meaning I let people be themselves, one of the least judgemental people around but then when they fall short of some weird overachiever psychic expectation, I act like I didn\'t know getting tied to the traintracks might end up getting you run over, so really it\'s a big trick, as a wise he always says...  And..to top it all off, I\'m the biggest hater of flawless game ever...like the Indian girl who is smart, Spiritual AND RELIGIOUS, perfectly blended with Eastern and Western values like an Ohm printed on a Billabong shirt, who doesn\'t drink or smoke, who is athletic, socially apt, and in medical school, well...i wish she would combust too (see previous entry)..because damnit, I have worked too long and hard at not being any of these things, suffered hard stares from many an Indian socialite and spent way too much time in school to make myself infamous to have it all washed away like a sand castle at high tide...I am always trying to make the simple point that \"outside the box\" is way cooler than inside the square..  '),(13259,3560141,'18,June,2004','2004-06-18 00:00:00','   No different than any other day...except it feels longer  This guy who rotates with me told me this story today...He knows a guy in Chinatown who\'ll make him an Omega watch for twenty dollars.  He told me how he spent the more amusing portion of his day convincing a set of residents that it truly was a fake, that the craftsmenship was of superb quality, but it was definitely a fake.  He knows because he was accosted his second day in NYC by three young men and the only thing they took when mugging him was his fake Omega watch.  They punched him in the eye, but took a twenty dollar watch.  My partner relived her experience in Dominica, the ease of a Caribbean breeze, the solitary street light that shone above their apartment complex, and a truck load of locals with machetes who appeared as if they were pirates.  She moved three weeks later, but her other two complex resident buddies remained on their patios always armed with machetes with longer blades and a single door to an apartment with a phone always propped open, set up specifically for the getaway.  I have gotten items of value stolen from me, but I\'ve never been shaken to the core before.  That core, the place that houses respect and hope in fellow man.  If I ever came face to face with an addict who had transcended every ounce of civilized living just to get another fix, or a hungry person who would eat my core alive in order to survive I believe I would curl into a ball and surrender to the idea of faithlessness....Just a thought that invoked serious fear and gratitude  '),(13260,3560141,'17,June,2004','2004-06-17 00:00:00','   It\'s weird how two people can have so much admiration for the other sitting in the lap of...  He says you live a life with qualities that I wish for, I close my eyes when he tells stories so I can drift away from mine...His lifestyle chimes with rockstar - dome, the likes of which he will never admit to.  He lives in a dreamy apartment in the sky and valet\'s his European sports car every evening after a long tortuous day of computer based work, meetings to discuss business ventures, projects with roots in many forests...His hippocampus works on overtime packing new amazing ideas into short therm memory that day by day grow into tangible be it or be it not gambles, he delegates his responsibilities only willing to subject others to work he would otherwise find himself overwhelmed with...he has gadgets like ipods, flat screens, watches that are worth more than my rent, credit card thin cameras...He has his eyes and his savings set on property he wants to buy, like instead of an apratment in the sky a condo in the clouds, he watches the market and knows the lingo, He buys a paper everyday, but he has the Sports Section memorized, he knows the words to every song ever, quotes great authors from the library of books he has read, can actually differentiate between a good and a bad movie these days, has been on stage with lines and everything, knows porn stars personally, takes golf lessons, did I mention he\'s younger than me... and he parties beyond what the average person would call a party (try impressing him) in deserts, on beaches, on boats, in a metropolis or in the Keys, in style with plush couches and bottle holders with ample mixers at his disposal... I can only assume as my life\'s shadow has fallen quite short in recent times....I listened today as he spoke of a project, one that would carry itself into stability, this project which could now possibly carry him to the verge of a bigger gamble, but a gamble with a wager closer to his heart...He could tap that well, full of creativity, spontaneity, full of those one line texts he so quickly sends telling me of new ideas I should write on, use all of that power (I sound dianetics - ish)to make a name for himself...He is never afraid of ideas....  Ok these are the tangible things I admire, what about the brain behind the He...He thinks for himself, not for others to notice, others notice because I pour the spotlight on his achievements at least within my circle...He is unrelenting when his brain is STILL in the mood, no idea is left unturned....He speaks with confidence, with utter confidence...He works the atmosphere, the person doesn\'t even know they\'re being charmed...His assertiveness never oversteps its bounds into cockyland...He will go up to two gay men in a city that is not his and ask for input on an idea he has for a gay friendly gym, utter confidence...He is never afraid of ideas...He will buy a crippled, lonely old man with one ticket to a movie, a soda and a popcorn...a heart with real empathy under the poker exterior...  He says respect amongst my peers, a worthy profession, a so-called talent for writing (I think it\'s like my pouring spotlight thing) or a creative soul, a high intellect, a certain likeable character (it\'s the people pleaser in me), stability...this is what he sees me as sitting in the lap of...it\'s funny how we cloud ourselves with pessimism.....  '),(13261,3560141,'16,June,2004','2004-06-16 00:00:00','   \"Says he wants you, Says he needs you, if he wants you, why not let him wait for you...\"  I\'m not sure how this works, this could be good karma for some drama that recently occurred at school, or this could be bad karma for wishing for too many things to change, disguised as good karma...either way here goes the story..  The Armenian pops his head in the door and says, \"...hey I gotta message from Hat (that\'s what we\'ll call him), he says that you are lookin\' fine and that song dip it low...well, it reminds him of you...\", What??  Great, I think...I don\'t even know what song that is, probably some garbage about booty shaking and dirty dirty sex....but I am, however compelled to find out what is this song...and how does it remind him of me..?  An array of Kazaa searches later, in between all the new pop-ups I now have to close since installing the newer version despite being warned by a wise roommate that warned me, there it was \"Dip it Low\" Christina Milian, downloading in progress..  \"Take your time..feel him out..when he\'s a good boy, a really really good boy, why not let him lay with you, that\'s when you give it to him good...\"  Ok, for those who haven\'t seen this video...SHE IS SMOKIN\', Definitely the next Beyonce or even better...very seductive, very mysterious, totally in control, with some very wicked dance moves.  Since the reference to Beyonce, I was unable to see the literal connection between her, the song and me, but I remember the incident that fueled Hat\'s interest...It was mysterious, seductive, and filled with some pretty cool dance moves...(i.e. a drunken night\'s make out session with plenty of dancing involved, this is where all my trouble starts under those strobe lights)...maybe this was it..  \"...Dip it low, pick it up slow, roll it all around and make your back go..Pop..\"  It\'s not so much that I want him or want anything more from that night, but it\'s more like there\'s this confident, sexy man who knew me as a little goof, but now sees me as this \"sexy little minx..\"  It\'s the fact that HE wants me that makes me hot..God that\'s sick right, I think he\'s sexy too, but noone I would sleep with or date, more like a good friend to have. He\'s intensely strong and un-rattleable, this is too intimidating for someone like me.  If you\'re too smooth, that means eventually I\'m going to get duped, and if there is always an answer I\'m going to run out of questions.  Maybe he would never understand the fact that sometimes these sweaty palms of mine need holding before entering a big building with many, many socialites...I already have this amazing presence in my life who sees everything and brings me this wholesome happiness...he\'s the man I dream about \"dipping it low on..\" (see previous entry)..hat seems to be drawn to \"Drunk D\", my alter ego, confident, sexy, hilarious, never at a loss for words and a huge flirt....I   \"Late at night, he\'s comin\' home..meet him at the door with nothin\' on....\"  You see, I\'ve never really been objectified, so that\'s probably why this feels so invigorating, after I reread this I\'m probably going to confuse myself as to whether I want to be appreciated as an intellect or a \"sexy little minx\", who knows, I know it\'s going to be short lived this novelty of feeling kinda sexy, but I guess one might see me with a little extra strut in my loafers, button down shirt and white coat today...  '),(13262,3560141,'14,June,2004','2004-06-14 00:00:00','   It\'s going to sound weird...why do I always start these things off this way, like noone else thinks this way...if i see this happen one more time I swear I\'m going to get into plastic surgery for all the wrong reasons, not reconstruction, not enhancement I\'m just going to build a nation of taut faced, plump lipped, 26\" waisted, non-saddle bag having, bootylicious, lipo - ed, botox - ed, mountainous boobed having monsters...before I actually get offers for private loans to help me through the rest of school, the reason I feel this way, is because I am not one of these people who is naturally or surgically built this way for one, so if this seems like one big complaint...good at least it\'s honest...and two by default, I automatically get the raw end by being one of the average if not below average people who are being compared to these buxom beauties...  My mother once offered me the chance to get a boob job to enhance my self confidence, but I refused because I work really hard to maintain a natural athletic/actually big boned appearance...Really, I think I refused because from the waist up, I consider myself pretty petite but if I tried to make myself more top heavy (which I have tried with toilet paper and small pillows...damn those drunken girlie sleepovers), I look absolutely mammoth, so I refused...1/2 a chance is better than none...  ok, 1/2 a chance at what one might wonder beyond the world of porn (no knocks against porn, I support this quite avidly, see previous entry)...the world is sold immediately by a beautiful face with a beautiful body...obvious observation, nothing mind blowing here...but what about on the professional circuit...and even more obviously when trying to shape yourself as a person...I mean you just don\'t have to be an overachiever when most things are just handed to you...you just don\'t have to be witty or charming, nor do you have to have the right bag or blouse...If you looked like this you could be picking a wedgie dressed in Hanes and cue ball pajama pants and well...tadaaa you have porn...or better yet, a great job, friendly people all around you, and you would never have to lift anything over five pounds..(oh dont worry biceps can be implanted...)  I have had to overcome this serious demerit I was handed for shoving while I was waiting in the registration line for a kick ass exterior in the Great Beyond, so I worked it the best way I knew how, I developed a smart mouth instead of \"jolie\" lips, a mind filled with useless tidbits about every subject in the book to \"connect\", instead of GGG Cups...Well ladies and gentlemen...for no freaking good reason except to be considered the girl with the great personality...  We all know the grass is greener, and right now I don\'t want to hear about these women who are saying, \"I just wish they would take me seriously..\", or how about the great personality chicks who go for the implants...and don\'t look mammoth...doubly don\'t care...cuz they\'re doubly blessed, and well those who are built this way naturally and were born to a Pulitzer Prize Winner and Bill Gates, well they can just combust...wow I\'m really in a mean spirit today...or everyday just something new to bitch about  I want to be the Video Girl, for five days...just five days...the scantily clad, perfect bangs having, drop dead gorgeous just oogled at for her body and her next appearance in another video..you know the video girl...I want to know what it feels like to not have to say one thing...i\'m making assumptions I know...but it\'s my wish, right...I\'m taking the arrogance thing a little too far...I would just sit, smile, drape off of Jay-Z, cross and uncross my legs, adjust my skirt height occasionally with a random biting my french manicured fingers, and stare into a camera through Dior shades...Just five days, I\'m sure I would be blogging after that again...with fewer words and maybe in this world fewer complaints.  '),(13263,3560141,'10,June,2004','2004-06-10 00:00:00','   Without the shadow of a doubt, even if it makes me seem like the owner of a 1977 life with no headlights, film and documentaries seem to spell out the answers that I have always known, the problems I have always had, and the perverse which kinda hangs out in the background...  I sound like everyone else in this generation but really this weekend a solid 4 hours was spent totally absorbed into a screen, and totally wanting more, almost like motivational speeches from a mentor or a therapy session with a psychologist...I don\'t officially despise my father anymore, and I don\'t want to drop out of school.  Now, we can see the reference.  I watched Dogville, a three hour seemingly unrelenting film about a woman who just seemed like your run of the mill pushover, I related right away....actually, it\'s the last scene that got me, like at the 175th minute, this movie really had me by the &%$!.  Get out of Dogville, is all I could think how simple, just hide out somewhere just as you did 174 minutes ago, you don\'t have to go through this...turns out the victim/martyr role that Nicole Kidman played in this movie and that I play in my own full length feature film running slightly over 180 minutes, was really consumed with arrogance.  The fact that most of us let so much slip past our radars, doesn\'t imply that we really are shiny happy care free people, it\'s more the fact that standards moral, ethical, etiquette, courtesy, academic, professional, financial, emotional that we set for ourselves are well beyond what we expect from our neighbor, so why hold these people to the same flame.  We might actually consider ourselves on a higher level, capable of perfection whereas these \"others\" are only doing what conditions and circumstances have lead them to do...like we should appear so forgiving so easy going only because \"poor guy, he just doesn\'t know what true love is, poor lady she may have had to live in NYC her whole life, so no wonder she\'s so icy and crass, oh poor boy with a family like that no wonder he behaves this way, oh poor girl, her parents neglected her so no wonder she chooses to sleep with every thing in possession of a dick.\"..I don\'t know it made me realize that if i were really equal with everyone else, and it\'s ironic because most would say i\'m a pound or too short of a full bag of confidence, then i should be mad, and should demand my fair share, and should not be expected to be rewarded in some higher power fashion, because well i\'m closer to God...I really don\'t know if I have to say it again, but it\'s a weird neuroticism that I may have introduced into my psyche, but I am arrogant with a cover of being devoid of a backbone....interesting...my father says that I need to be more like him..I resent that, a pompous demanding overachiever....but not anymore..how can I resent someone who I emulate just in a very distorted way, I want everything, but i wont step on some poor person\'s feet, because god only knows the traumas they have suffered these poor people who are so much worse off than me...where do I get off exactly....  Oh..and i\'m not dropping out of school for sure now...not because medicine is where the undercover narcissists go, but because i watched this Discovery Health show tonight about residents, about how each person in this field is quirky, is stressed, is full of self doubt, is just plain guessing, feels overworked and underpayed, always wondering what they signed up for even at the point of residency, full of anguish in their personal lives (i.e. cheating husbands, boyfriends who beg that tonight something too traumatic won\'t happen to end the relationship...again...)coincidentally all the featured residents were women tonight..they were all so content with a day\'s work too, not content like I saved the world, but whoa I didn\'t have a breakdown because I was way too busy helping 20 people today...dude it sounds like my life at this very moment (minus the helping 20 people thing, I\'m way too busy getting their Date of Births and Chief Complaints)...I simply don\'t belong anywhere else...not fabulous enough to be a lawyer, no poker skills to be in business, and simply not happy enough to be a teacher..however i guess anguish always serves as fuel for HOLLYWOOD...if all else fails...here I come....but for now, I love my future job...  '),(13264,3560141,'10,June,2004','2004-06-10 00:00:00','   How do so many people gather the courage to grow up?  I used to be the jumpiest frog in the bunch always a second late to jump on adventure, autonomy and pure ACTION.  Little did I know that, with each leap meant that I would be farther and farther away from a big wet pond with a buffet of flies...  I moved away to school in Miami at fifteen, where i quickly learned that just because you are offered a shiny VISA (and a free T)doesn\'t mean that you have to vacation in the Keys.  You see I thought that this was the adventure, the autonomy the ACTION...Well, a few months later, with no incoming revenue, because of my rockstar schedule not accomodating for a part time job, I pawned my CD\'s and my CD player, for gas money to drive the 210 miles to get home, the big wet pond with a buffet of flies.  I flailed my arms and just complained for those three weeks about having no life, being restricted and having no support with regards to my accomplishments.  I mean it was hard work to get a 2.6 in Freshman year classes, and boy did I play up the age card...\"Do you know how lonely it is for me, how much harder this transition is when you don\'t even have a driver\'s license?..all I really need is your support, how hard is that?  I just wish I was older?\"  I moved off campus in 1996 - post the consecutive below 2.0 GPA grade reports, and decided school wasn\'t my thing and well...I could be an adult, get a job, love the job, make tons of money and never need an ounce of support that I wasn\'t relying on anyway, right and how did I ever pay those credit cards off... I worked, hard I might add, drawing and writing quippy ad campaigns for Air Jamaica and others to print on napkins, pens, etc...my dream job payed 6/hr, so I was slapped with eviction notices almost every other month, because well my rent seemed to be the last priority when it came to food, gas to get to my job and cable, oh and paper plates to eat my ramen noodles on, you see 6/hr just didn\'t stretch that far...another miserable attempt of taking the adult world by the horns.  So, i faired better as a Student, I mean the cafeteria was always open, and you just didn\'t get evicted. Hence the six year bachelor degree plan...I finished and was finally ready to do something with my life, despite how much I would miss the \"OG\" comments I would always hear underneath their breaths, I had to hop on into more school, this time on a sea bound rock.  \"NO, I don\'t need your help to settle in Mom, God, I\'ve been living on my own for how long now..God I\'m 24 years old...\"  Well, after the novelty of the place wore off, seconds after regaining consciousness from the nap on the plane there, I screamed \"MUMMY!\"  I\'m a lot older now, a lot more responsible, I pay my bills when I get them, I put stamps on everything, and I give everyone my real social security number and name, use only debit cards, checks and cash, and the big and....I have a savings account....but I still can\'t figure out why every other day, there are kinks in plans..ruffles in attempts to be an adult...Someone lost the paperwork, We need a roomie ASAP, I have to go the grocery store, dry cleaners, and home depot to cut keys like hours ago, there\'s always gotta be toilet paper, and reorganize and clean closet days, I am so tired...This is the adventure, autonomy, the ACTION....  I\'m here, a few lily pads away from my own pond, a few more hops, a few less hopes....All I need is enough flies here, I just couldn\'t go back to ramen...  '),(13265,3560141,'08,June,2004','2004-06-08 00:00:00','   Where do my days go?  How can I be staring blankly at this site with nothing but wasteful thoughts that play sequel to where my mind wanders during the course of the day.    Scene 1: Crawling out of bed, taking a millisecond look in the mirror to make sure no new blemishes have added themselves to the most despised list, walking past the box of Marlboros many times until I cave, say good morning to the Mercedes Mechanic guy who lives downstairs and probably thinks I spend every night obliterated, considering it\'s 9 AM and I\'m smoking already...spend thirty minutes dancing like a monkey in the kitchen with my favorite roomie discussing how I\'m cutting down, only to realize I have twenty minutes to either go the gym and be late to school or just be late to school, depending on the day I can be found on a treadmill, but I am consistently late to school.  Scene 2:  I am standing side by side with my partner who is equally as blank and shell-like, propped against a faux oil painting that juts into my back.  We nod a lot, and place our hands in our pockets, under our arms, on our chins like clockwork when the doctor faces us to translate the last twenty minutes that has just transpired.  It feels like being yanked on your big toe while in a deep sleep, aware but unwilling to acknowledge this disturbance.  Scene 3:  Interludes to translation:  Sex, lots and lots of sex, old experiences, potential ones, What am I gonna wear tomorrow?  Did I finish writing that paper, Did I get this guy to sign my attendance sheet, I wonder if my wireless card will work today, Do I need tomatoes?  Do my roomates mind beef tonight?  How many cigs have I smoked today?  Has anyone texted me?  Did that loan come through?  sex lots and lots of sex, I\'ve tried to dilute this but really it\'s disgusting, Dr. B turns to tell us that his patient has been feeling fatigued and dizzy probably due to some sort of anemia, and I can\'t detach from the thought of him nuzzling his nose in my neck...please...not close anemia would seem more gripping at this point.  Oh...he...he will probably turn up in one of these blogs eventually...the scenarios aren\'t so explicit, it\'s more like I\'m reaching for spices in my kitchen fiddling between the smoked paprika and the garlic salt, and I feel him behind me, I feel him behind me...His breath is hot and my hairs stand on end...ANEMIA, we can\'t ever cross this point in the fantasy, I\'m way too focused on feeling him behind me and the whole breath description.  There are also flashes of a hood of a car..hmmm.. I wish I had more, but I\'ve become bashful as I\'ve written this...so, for now..ANEMIA  '),(13266,3560141,'08,June,2004','2004-06-08 00:00:00','   So, it\'s not just here apparently that my name has been in use over so many years, it\'s hotmail, it\'s friendster, it\'s even velvetlist.com, the reasons for which I tried to subscribe seem intensely embarrassing and so I will beg no further judgement be made, noone likes to pay cover charges buddy.  I have to add letters, symbols and numbers until my name has become so distorted that I\'m sure I\'ll never again remember a sign in name creating a vicious cycle of sorts.  \"What was my sign in again? Oh, forget it, I\'ll just create another one...what do you mean there is already a dddppaee77#!, damnit..\"  The frustration is not limited to cyberworlds. Let\'s take my present educational situation, taking me from one venue to another over the course of the next year, what does this mean to you as a reader, well, it means that you are probably even more confused as to why this should matter so much.  I have to get new ID badges each time I move from one place to another, because god forbid the fact that I am first and foremost a student, and lastly a lowly scut monkey who is temporarily under the service of the hospital.  NO, it\'s imperative I have six different ID badges with my name presented in so many colorful block, and script letters in, did I fail to mention every morphed version of the spelling of my name...I have a simple name, not Joe or Ann, but i can only imagine what life must be like in the name game for these people, and well..I miss it...I even appreciate when my mother calls me by my full name, every consonant devoid of any underscores or extra P\'s...  You see this foils my plans for ever becoming famous, if someone ever mistakes my complaining and sarcasm for some brilliant work, they\'ll only know my pen name, which is a result of adding signs and characters until clicking on \"Continue\" takes you to the next screen.  It is nothing like the days of old when a pen name was representative of something a conjured pseudo-personality maybe, and well you would be able to write to them and request a manuscript for publishing.  They\'ll never be able to reach me because my e-mail address requires a graphing calculator to store the length of characters and pound signs.  Well, this was a hypothetical problem...very hypothetical, and hardly likely but I just miss my name..  '),(13267,3560141,'25,July,2004','2004-07-25 00:00:00','   å±±ä¸‹ Yamashita (under the mountain) ä¹…ç¾Žå­ Kumiko (eternal beautiful child)&nbsp;   urlLink My japanese name is  &#23665;&#19979; Yamashita (under the mountain) &#20037;&#32654;&#23376; Kumiko (eternal beautiful child) .  urlLink Take your real japanese name generator! today!   Created with  urlLink Rum and Monkey \'s  urlLink Name Generator Generator .  \">your real japanese name generator!  '),(13268,3560141,'24,July,2004','2004-07-24 00:00:00','   I just barely escaped with my life only a few moments ago...I was smoking a cig, supposedly from the last pack me and the Armenian have agreed on buying...when all of a sudden there it was, the biggest yellow jacket I have ever seen...   I\'m not sure if I can be held accountable for what crossed my mind at this point because it was sheer fear...but I had the likes of a ninja, with one eye steadfastedly gazing into the bee\'s six thousand eyes, and the other on the length of the cigarette in my mouth...all of a sudden it seemed like a Marlboro Ultralight 100, grew inches before my eyes...   But I refused to let this yellow jacket make a yellow belly out of me...so I continued with my cross eyed gaze..before i knew it...he (had to be a he...didn\'t have the likes of a queen) swooped into the apartment.&nbsp; I was conflicted, do I put this cig out run inside and warn the Armenian and Yeah man&nbsp;(new roomie),&nbsp; or do I save myself by staying outside on the patio forever?&nbsp; Before I could answer this moral dilemna for myself, God, Murphy, fate brought the answer into the picture for me with the arrival of yet another yellow jacket onto the patio bannister....Fuck...in a city with so much concrete, and not a colorful blossom in sight, we seemed to look like a hive...Fuck   I put it out...the bastards had won...I slid in through the barely ajarred patio door, and then I heard it, the bzzzz the bastard was taunting me...He was just between the glass and the curtain, I trapped him there, shutting the curtains firmly from an arm\'s length away....because you know the curtain is as effective as Raid...whatever...   \"Armenian?\"&nbsp; I said quite loudly, what would happen if I let a bee in?&nbsp; He\'s laughing incessantly and slowly shuttles the bee outside...turns to me and says .\"Everyone\'s ok, no one\'s hurt, break it up, go home everybody.\"&nbsp; I\'m lucky I got away with 3/4 of a smoke and no stinger in my ass...  '),(13269,3560141,'24,July,2004','2004-07-24 00:00:00','   urlLink    It\'s always the little fold that gets me :)&nbsp; urlLink  '),(13270,3560141,'23,July,2004','2004-07-23 00:00:00','   Just random:    &nbsp;  Benefits of ADHD include noticing every detail of your surroundings selectively, not necessarily for the purpose of remembering them,    &nbsp;  Never wear white when you least suspect rain (because it will) and don\'t park ten blocks away from your destination to save 7.00 when your destination is smack in the middle of Shady Grady Hills, Brooklyn,     &nbsp;  oh..and Don\'t think that hugging your messenger bag really close to your white shirt in the rain in Shady Grady Hills looks inconspicuous...likely to draw attention    &nbsp;  Never help anyone...especially psych patients who really want you to get their aunts to bring them KFC because their cafeteria is serving green meatballs,    &nbsp;  Never trust anyone with one of those laughs that make you think, \"Do they not hear what that sounds like...ok Dr. Evil?\"    &nbsp;  Just because your urine tox screen for drugs shows up clean never reveal that you spent a few years growing up in Jamaica and you know that it could cause excessive paranoia,    &nbsp;  Never express any interest in anything, because people will assume that you want to converse about this ONE topic ALL the time...    &nbsp;  Before you hit the Delete All button on your inbox, just double check your selections carefully one more time    &nbsp;  Eat fresh figs with almonds...mmmm    &nbsp;  Never discuss politics with a right wing  MAN  of Islamic faith and an Armenian  MAN  who is well versed in the ethnic cleansing performed by the Turkish in 1915...things could get sticky    &nbsp;  According to a news show on FOX, never not show up for the sentencing of a defendant if you\'re prosecuting...could end up getting the alleged criminal 60 days for raping a 12 year old girl    &nbsp;  Never think you can eat a NY style pizza pie, with olives, tomatoes, mushrooms and green peppers altered by steroids, as a&nbsp; second  dinner, because each slice weighs approximately  2lbs ..    &nbsp;  Don\'t watch Paycheck with Affleck, Uma Thurman looks old, and no I don\'t believe in second chances...    &nbsp;  Blog and Blog some more..    &nbsp;  A new phrase I heard today..., \"Well...him wan ease he body.\"&nbsp; , you know , \"Faat pun me head.\"    &nbsp;  Speaking of which....    &nbsp;  '),(13271,3560141,'22,July,2004','2004-07-22 00:00:00','   Even though I can\'t take back time, and I would never wish for my circumstances to be any different than they are right now, I do wish for this much...  I wish progression didn\'t go hand in hand with expecting so much out of life and loved ones.&nbsp; With all the changes that I see unfolding along with those already experienced, I can\'t help but remember times when my brain was pacified by thoughts of living in a hut by the beach, smoking, and fishing all day long.&nbsp; I used to have such admiration for the simplest of things life had to offer, for instance, if I just got into med school, if I just could lose some poundage, if I could just bask in the sun and never go hungry, if my heart could only be clean and pure, and if my only expectation of people around me was to be happy and be happy with me.&nbsp; Even my choices in men were more grounded based on the mere fact that they possessed ideals that I respected, they really didn\'t even have to respect me...this may account for the jaded product that types before you....  Now...well, now peace comes in different flavors for me...I don\'t want material things, although they would be nice, but the pacifist that once existed in me is faced with my other side, the one that roots the underdog on, the one that empathizes with problematic psych patients who are unruly or undisciplined.&nbsp; My ideals I realize now, are genuinely open for discussion, not open for the sake of being called, \"open minded\" or relaxed.&nbsp; I can see why a child would curse at his mother insistently, after being compared to a sibling or being abused himself.&nbsp; I could see why a group of angry children would want to ridicule an outsider, because he instigates 90% of the trouble these little big boys get into.&nbsp; The budding rastafarian (funny pun if I do say so myself) has been smothered in me...I want to be a survivor, a fighter who doesn\'t take being taken advanatage of lightly, whose pet peeve is a man who shows no respect for me.  I still want a hut on the beach but only after I\'ve worked very hard and deserve a life with no worries...although I can\'t guarantee that complaining won\'t tag along for the ride.&nbsp; If my heart can be free of guilt for hurting another intentionally then this is much more appealing than steering clear of all evil...meaning in life sometimes the right thing just shouldn\'t be done unless you\'re willing to sacrifice your own well being.&nbsp; I\'ve lost my genuine interest in people, for the sake of clarity, I mean I can\'t even see myself attached to very much, tears seem useless, attachment seems useless, and caring to the point where I used to seems useless...  My parents always said with age, practicality would overrun the passion, the drive that used to bring me to tears to hear someone else in trouble...I\'m not sure if I like the trade-off yet..  '),(13272,3560141,'19,July,2004','2004-07-19 00:00:00','   In one amazing shot, vacations, I realize just don\'t suit me.&nbsp; I just got over a three week span of nothingness after being in constant motion since may of 2002. &nbsp; Does this mean that without school, I\'m devoid of interesting things to do?&nbsp; Possibly, considering that \"to do\" always involved a chapter, a patient write up, or a clinic assignment, paying bills, visiting the salon, working out and shopping for groceries for my belly and my feet..shoes are essential...Regardless, I\'m back to the grind and loving it.. &nbsp; &nbsp; Have you ever found yourself, making conversation, like really making conversation..like from a ball of clay molding&nbsp;a sculpture of the most intricate sorts, why has it become so difficult to relate to people..I used to thrive on this the fact that with a few of&nbsp;the right questions I was hooked&nbsp;on a person and what they had to say, now it seems tiring and not worth the effort. &nbsp; I&nbsp;wanted to write this one down before it became a good idea gone to waste: &nbsp; He says imagine a couple so deeply in love, that effortless stuff I\'ve&nbsp;written about&nbsp;before,&nbsp;then one day without notice, reason or fail they end on the ugliest of terms with flying centerpieces and shattered glass.&nbsp; Many a tear filled sleepless night would have passed running playbacks of scenes from their relationship, but as&nbsp;time runs as well, the memory will tire leaving a&nbsp;vague blurry vision.&nbsp;&nbsp;Several years pass, and each in their own respective lives begins to have dreams bringing these visions back into focus, but the dreams are not luscious or loving.&nbsp; They are of wrought iron contorted into malformed objects and steel cages becoming malleable under the force of&nbsp;a guard rail...perhaps an accident of some sort...&nbsp;one of the two of these characters is forced to a hault on one day\'s commute form work, and emergency vehicles have crammed both sides of a bridge...They get out of their cars and with enough courage decide to go and aid the victims of the accident...only to find??? He stopped there....A likely candidate for a great novel by a great author, when he finds the time and the courage... &nbsp;  '),(13273,3560141,'14,July,2004','2004-07-14 00:00:00','   urlLink 20 Questions to a Better Personality        Wackiness: 76/100 Rationality: 64/100 Constructiveness: 64/100 Leadership: 52/100    You are a WRCF--Wacky Rational Constructive Follower. This makes you Paul Begala. You are unflappable and largely unconcerned with others\' reactions to you. You were not particularly interested in the results of this test, and probably took it only as a result of someone else asking you to.   You have a biting wit and intense powers of observation. No detail is lost on you, and your friends know it--relying on you to have the facts when others express only opinions. You are even-tempered, friendly, and educated. Foolish strangers may mistake your mildness for weakness--they will be surprised.  You entire approach to life is enviable. You will raise good kids.  '),(13274,3560141,'12,July,2004','2004-07-12 00:00:00','   How can you dislike someone you\'ve never met ever?  Well, I\'d like to attest to the fact that it\'s possible....  Nothing stupendous or productive happened today, so, I decided as my outing for the evening why not go to the gym and try and shed the few pounds I put on since being forcefed through guilt by my loving aunts.  There it was the usually packed parking lot with an attendant who flags people into obviously empty parking spots and the ones creatively sculpted from the dumpster being moved around by the garbage trucks.  It was empty today, I suppose the rain and lightning makes the treadmill seem like an unlikely choice.  Anyways the usual ritual of searching for the sign up sheet for the treadmill seemed pointless as there appeared to be a total of five people in the gym, so I started stretching and began running...  My routine was coming to an end, the part of the workout where I call myself names and keep picturing life without the extra pounds...when he shows up...that guy, Cud guy, only because he chews his Gatorade gum like a cow.  I had seen him before, he\'s gotta be like 35, Arab or maybe Indian I don\'t know, with a heavy mustache, in pretty good shape.  The thing that stands out in his persona is his attitude, he walks with a towel slung over his shoulder, he pushes his way to everything, despite the five people in the gym, and he rarely makes eye contact with anyone.  My first eye witness account of his cocky attitude was when he tapped some old lady on the shoulder and without saying a word, pointed her off the lat machine...dude, pointed her off, like index finger to her and then thumb to the open empty gym floor, the whole time chewing his damn gum like his name was Arthur Bally.  Then he puts the pin in the hole of the lat machine at a level which would have read \"I really want my neck veins to rupture, really\" or somewhere around 190lbs, either way his gum chewing escalated to beyond cow-like, he got the weight up but believe me I was holding my phone just in case.  This guy had the nerve, anyways I know rules are rules, but I mean with one other person occupying one of the fourteen treadmills I really don\'t see the need for this invisible confrontation.  He stands at the foot of my treadmill, and I catch his eye in the mirror which faces the machine, he folds his arms, chews so loud I can hear it over the motor, whistles...YES...and then nods...YEAH...I didn\'t even get the finger..haha...I got nodded off  So, being as he could smoosh me with his calf, I stop running...gather my stuff, and get off...not even a nod to say thanks I guess....  '),(13275,3560141,'12,July,2004','2004-07-12 00:00:00','   Ok, so I had heard all the warnings, been told that Michael Moore had mastered the art of creating films that could shift whole psyches, that his genius was merely another form of propaganda, and I had made myself aware of the fact that a trade off with being open minded is the fact that you are always impressionable...  So, I saw 9/11, only weeks after most of American youth, I have been disturbed in many other ways but this uneasiness is different.  I had been one of those people who read every piece of literature I could find in print post 9/11 to try to find the reason for the heinous attacks, try to find that one US News article that would compress fifteen years or so and describe the events which lead up to these acts.  I have always found that fundamentalists of any faith, although sitting on the far left or right of a see saw always had profound principles; methodical, philosophical, religious, and steadfast grounds for most decisions that are made. In a far reach, I grew to respect these motives, meaning they were fueled by much more than personal gain or wealth.  I will never understand why besides wealth, people only also respond to violence and humiliation, and therefore my respect stopped at just that respect, not admiration or alliance with these actions.  In debates that would follow, and ones that still sprout every so often, I would say why should we wonder why our country would be so infamous, there are countless caves in which our enemies could be born.  America has shown a disregard for principle, hardly held to its word, and scarcely remained loyal.  This movie was an eye - openner in a different respect though...  As a country we have been blindfolded, spun around and sent to find the pinata.  Wealth and opportunity run rampant here, and to my luck it is even available to me, a minority woman.  The beauty in this land lies here along with freedoms that are neatly trimmed around the edges but for the most part freedoms none the less.  There is a price though...the rich will remain rich, by any means necessary, and the poor will always remain poor.  The closing arguments in the movie described the necessity of this hierarchy to achieve the world power status that we hold as a nation.  Most of our causes are futile ones, our arguments, weak ones, and we are blindly lead to believe that our government is in place to aid the people (including underserved populations) vs. maintaining this hierarchy in the country and the superego complex that we feed in order to rule the world through domination.  There has been so much smoky confusion put in place (loophole after loophole), that it\'s hard to notice the caniving nature of the ruling body and this movie just brought to light the fact that we, the public as a unit, have all been victimized in some way, while others fueled not by principle but by personal gain seem to be winning the race.  The world has suffered by the hand of these manipulations with unnecessary loss of life, destruction of cities, and enforced poverty....  I was just left uneasy, I wanted to try to dupe them back, legitimately, the ideas involved sacrifice giving up any dream of truly becoming a disgustingly wealthy woman with many shoes, so damnit...for now I\'ll rock the vote...  '),(13276,3560141,'12,July,2004','2004-07-12 00:00:00','   So I used to have to take this class where we learned about different personality types, how each patient could present, and how to in turn be able to have a healthy conversation in order to properly diagnose someone...  There were clusters, each category having a particular hallmark you would look for, and disorders that were common to each cluster...  Background being set...since this class I\'ve analyzed everyone I know in order to have healthy conversations, minus diagnosis:  Cluster A: Hallmark includes an avoidance of social relationships, with a \"weird\" nature, but not psychotic, These people generally have psychotic disorders like pshizophrenia  You\'ve got your  Paranoid Type:  Distrustful, suspicious, their problems are generally everyone else\'s fault - haven\'t met anyone like this yet...and if you think I\'m going to devulge my analysis on this thing...you are CRAZY (not one of the documented types listed below) I\'m listing them to eventually describe a textbook case that I met on my last rotation...dude..anyways you also have,  Schizoid:  long standing loners  Schizotypal:  the guys or girls with the garland of flowers on their head with every belief in the world that they are the princesses of a land called Unicorn and they\'ve been married to an alien, basically magical thinking with odd thought patterns or behavior.  Cluster B:  Hallmark being dramatics, emotional or inconsistent, generally these guys have mood disorders and are substance abusers...your run of the mill heroin fiends   Now in this cluster we have your Histrionic - need I say more than this...every guy or girl you have ever met who is just inches away from your face ALL the time, hands are always actively waving (the patented forearm to forehead - poor me - move), and they are not afraid to stick an occasional tongue in your ear.  This is the girl I met over these last few months, and no, she did not stick her tongue in my ear.  Anyways, part of our exam in this class was an oral exam in which on of these personality types would be played by a patient and it was our job was to appropriately handle the situation...needless to say, I never learned how to handle this one, if a patient ever hit on me..I\'d probably turn orange, that\'s the color we turn, and diplomatically refer them to another doctor.  It is described in a textbook as the Don Juan in men and the \"life of the party\" for women...  Narcissistic - the person who assumes we were all born to serve them and this is the person who assumes that noone else\'s dilemna\'s could be half as heart wrenching, so why bother listening at all...lacks any empathy  Antisocial - the kid who burned ants with a magnifying glass and hit frogs with tennis rackets...the arsen, the bully, no concern for social norms whatsoever...associated with criminal behavior in adulthood.  Borderline - the cutter, the erratic, impulsive person who solves every problem with an overdose attempt.  Then...there\'s  Cluster C:  Hallmark is a fearful anxious person usually suffering from anxiety disorders in adulthood ******************ME****************SUBLIMINAL MESSAGE************************************************  These guys include the  Avoidant - This is the person who says, \"Well, since you guys aren\'t paying attention to me, because I\'m of no consequence, I\'ll just stand here and face the wall and pout.\"  Obsessive-Compulsive - Goes without saying...famous for, \"What do you mean, the Alphabet Soup should be first in line in the cupboard, it starts with \"A\".  Dependent - The people who call someone everytime a new pack of gum needs to be purchased, because their choice would never suffice....  Passive-Aggressive - OK ME - these words straight from the textbook, summarize impeccably, \"Outward compliance, but inward defiance\", sure I would love to do that for you no problem ****insert inner thoughts here******stupid bitch, you\'ve never done anything but make me miserable, you\'ve got serious delusions if you think I would....OK, I\'ll just do it**************  And that concludes our class in the Characteristics of the basic personality Disorders...With all this typing...I\'ve forgotten my initial reason for explaining all this, oh yeah...the Glimmerman, that\'s her name...clever manipulation of just one letter would reveal her true identity but...  Ok, so the first day I met her she spoke through me, you know, like asked questions just because she already had a dissertation of an answer ready to fire back at me...I knew this would be fun...she had begun her diet of yogurt and grapes, and she loved telling me how her newborn was doing so well, making his mark on every developmental stage and how she was such a good mother and how this was just her way of getting back into shape...my only response was that she had tried and true dedication to her profession, that to undertake such responsiblities as a student, wife and now a mother were burdens that I respected her for taking on, she was beautiful for just having done what I wish I had the courage and heart to undertake...she blushed and patted me on the head...skipping along to the salad bar..  The weeks progressed we both signed onto the same team for group projects, etc....While most of us were consumed with completing paper work, she got experiences that we see on TV, the entire ER season one....which she didn\'t fail to constantly remind us of....why was she alotted this luck, she left her paperwork till the end (which ended up being a very intelligent move), and gave up her newborn to a nanny and spent countless hours wedged between some intern\'s butt cheeks...with questions like what are you doing now, where are you going now, can I come with you now, show me what you\'re doing now....she was never shy, I watched her in action, she shot winks at the poor newly migrated Indian men who were probably enchanted with the feeling of a set of lips on their asses, regardless, she got her way.  When it came time for our \"group projects\", only one person helped, not her if you\'re quick...one of my partners is a new immigrant from Russia, who used to practice medicine there under the influence of vodka and tatu CD\'s, so he was useless, and she, well, she had now become the owner of bags under her eyes, and a shrunken frame from her misguided diet...she complained of always being too dizzy...too tired...too stressed out...way toooo busy.....see, we weren\'t ALL doing this rotation with ALL the same responsibilities, we were all vacationing in Ibiza until our projects were due...point being she should probably eat SOMETHING....go home once in a while...and stop being so freaking overproductive...but I replied with oh Glimmerman...we probably will never understand the hardships you undergo, ofcourse you don\'t have to do a thing (I was being sarcastic, apparently played the part so well, she thought I wasn\'t kidding) and Glimmerman, gosh...you look so good...so thin it\'s like you\'re swimming in your white coat...woman, what have you lost now 15 no 20lbs?????? she blushed patted me on the head and went off to the salad bar to re-affix her face to some intern\'s butt....  Weeks progressed even more...Glimmerman, had become scarcely available, spending almost twelve hours a day in the hospital...but she did however find me every afternoon at the hour of noon to complain CONSTANTLY...about ANYTHING..how her colleagues (i.e. all of us) were incompetent, how doctors (i.e.  people years advanced to us) were incompetent, how they stressed her out...she would pull me aside, from our colleagues, and whisper like an old woman all these nasty things about anything...and then say well..I\'m just going to skip lunch, you know gotta watch my figure...waiting with an ear perked like Wile E. Coyote, for me to say, Gosh...you\'re wasting away, you know how us girls love being compared to anorexics...  Glimmerman, oh Glimmerman..I ran into her today two weeks post rotation....she pulled me aside again...just to complain, speak of some new intern she has a crush on...and then asked how she looked...I replied...Gosh, Glimmerman...you look so relaxed, so thin, god...how much weight have you lost...she blushed patted me on the head and asked me to keep in touch as she went skipping along to the salad bar....  '),(13277,3560141,'11,July,2004','2004-07-11 00:00:00','   urlLink    his newest talent is...he can roll over...my nephew is only a few developmental stages behind..:)&nbsp; urlLink  '),(13278,3560141,'09,July,2004','2004-07-09 00:00:00','   urlLink    the trek to the arch...&nbsp; urlLink  '),(13279,3560141,'09,July,2004','2004-07-09 00:00:00','   urlLink    the top..&nbsp; urlLink  '),(13280,3560141,'09,July,2004','2004-07-09 00:00:00','   urlLink    pic 3&nbsp; urlLink  '),(13281,3560141,'09,July,2004','2004-07-09 00:00:00','   urlLink    so..I can say I\'ve been to St. Louis&nbsp; urlLink  '),(13282,3560141,'09,July,2004','2004-07-09 00:00:00','   I\'ve gotta get back to NY as soon as possible, oh god when did that become a sanctuary?  I\'m not deprived by any means or even chained to a wall with a dog dish for sustinence as I\'ve made my current status seem to be...I\'m just tired of being politically correct oh who am i kidding I miss never having to pop out the second story screen on the window, so I can smoke a cig...  So, it\'s been weeks, or realistically maybe just days you know since I\'ve had a decent conversation with him....I would normally be typing away privately from some hotmail account, but I\'ve been banned from using this form of communication...for the following reasons..1.)  put it like this if my child was missing, I\'d probably just plan the funeral and skip the search party antics - translation:  I assume the worst, no room for having a bad day irrespective of me, actually nothing to do with me... 2.)  if I was a boxer, I\'d hold up my opponent\'s glove and just consistently plant it square on the jaw, and bite my own ear off - translation:  I create my own dilemnas that seem insurmountable, my own worst enemy...how can anyone else solve my own problems..without medication..hmmmm 3.)  If there were a slight escalation in the wind velocity, I\'ve grabbed my keys headed to the nearest grocery store, picked up bottled water, batteries, and a first aid kit...I\'ve turned off all my fuses and I sit in the dark till the \"storm\" arrives - translation:  in addition to number 1, any small change, minor alteration in routine,  probably means everything\'s gone to shit...to hell...to the heaps of shit in hell. 4.)  If I got a tattoo that said \"Om\", the next week it would say \"mOm\", she would probably piss me off, and I would change it to \"mOmOney\", then I\'d get sick of Puffy dancing the background of everyone else\'s videos and it would change to \"mahOgany\" (I don\'t know I ran out of creative manipulations)..point being, oh I mean - translation:  I AM FICKLE..., so e-mails conveying sentiments are not allowed, the running joke is this week I like him, but who knows depending on the weather, I may not ever speak to him again...ok so maybe it\'s more stable now..like I\'ll never not speak to him (this does not guarantee against foul four lettered words), and I\'ll never not have him in my life (he may be my future investor, or business partner, who knows)...  So, because he reads this...this is my passive aggressive stab at conveying, \"I miss him, and sometimes my sadness translates into fiery e-mails...\"..haha...  Oh yeah, a testament to just how fickle I am, in addition to the other three attributes previously described...read previous entries...  '),(13283,3560141,'07,July,2004','2004-07-07 00:00:00','   Storms brew differently here in the Midwest...  The cliche, \"the thunder rolls\" is damn near pin point precision...The clouds are purple, at sunset at least, and they tumble through the sky..So, I was watching \"Who wants to marry my Dad?\" (because my life lacks reality) when those annoying orange banners creeping along the bottom of the screen insisted on reminding me that we were all going to die sometime between the hours of 9-11PM, due to some tornado watch.  My brother and my cousin run outside so they can play Helen Hunt and Bill Paxton and save Peoria...every five minutes they would update me on the color of the sky...  Unusual for ,my normally, paranoid self I stayed spray painted to their couch, praying that Meredith would pass the lie detector test on the show...she had great hair...Then, the sirens went off, loud tortuous alarms saying,\"wooh woooh wooooooh.\"  I jumped up, looked around, said to myself, where would I anchor myself incase winds exceeding 120mph were suddenly in my face...I asked my cousin where to run in the event of an evacuation..he responded with what siren, what evacuation, scenario could\'ve been compared to car alarms going off in any major city or a woman screaming help in the darkness....  Well, 11PM passed and noone died....rain and lightning continued till morning when the siren went off again at 9AM...I ran downstairs and everyone is seated around the dining table drinking tea or coffee, my aunt smiles and says ,\" OH..that\'s the testing system for the tornado warning siren goes off every first tuesday of every month...\"  what a relief, huh???  '),(13284,3560141,'06,July,2004','2004-07-06 00:00:00','   I saw her gliding through the crowd like she always did, the patented wave and kiss on the cheek to every notable in the room...God..how did she always do it, this time when I reached out to her, it was different..  Our interaction began with the scripted formalities one extends to every Indian person one ever meets (I\'ve decided that if one more person asks so what do you do...I\'m going to tell them I work in underserved communities as a graffiti artist and have you seen my latest work on the Goethal\'s Bridge - it\'s what the baby on Biggie\'s Ready to Die grew up to look like, since he\'s really my kid)...I\'m going to call her Sweat, it\'s not a far reach from her real name..my cousin, the one chick that stood as the standard, so I secretly grew to love hating her.  She never did a thing but be herself, stunning, intelligent, sarcastic, and just plain fun.  Without ever knowing it, I think parents instigate the imaginary feuds that come from constantly comparing two things.  Point being, my dad and Sweat tag teamed the wrestling match I had with my self esteem growing up as a kid and hell even now...  This time it was different when I reached out to her...I hugged her in her beautiful condo which faces both the river and the city of Chicago...stunning..some things do not change...She had the likes of a movie star who just threw on designer jeans and a tank with pieces of jewelry that in certain lights were sure to sear my retina...but the point being, we ate at a crab shack where noone knew how to crack the crab, so we moved on to crab cakes, the work\'s already done, we drank domestic beers, went to eighties bars where trucker hats, fooseball, and get this ping pong kept us entertained till dawn...we laughed so hard, the whole clan is just hilarious...then she talked to me (alcohol may be a reality blocker, but it has openned up so many roads of communication that I would never trade ever), she, the most independent sharp witty woman I know, is looking for a love, that doesnt feel like a burden, like obligation does not exist...her father has been diagnosed with Parkinson\'s and she describes the way her mother without even thinking has taken care of him, with a whole heart and not an ounce of resentment...how many people can do this she asked..I want to love someone like that, and be loved like that, tears began to well in the corners of both of our eyes...I said then with all the other things being exactly as you wished for, don\'t settle for just any functional form of love...live to meet the love of your life...we both cried...tears of the sad reality that this may never happen for either one of us, and my tears had hints of joy for finally letting go of resentment and jealousy...  '),(13285,3560141,'01,July,2004','2004-07-01 00:00:00','   You know when you look in the mirror and you see this one mark, that you try to get off with spit, then you go to the bathroom sink and try your washcloth, and then you\'re completely submerging your whole head in a soak of skin bleach just to see it disappear....I feel like that sometimes, trying to get the monkey off of my back would be an understatement...  I have tried to unravel the roots of this self esteem issue by digging in every forest that I\'ve ever been lost in...I see everything that is beautiful, warm, attractive, positive in so many people, I see it that shine that everyone has and I\'m never afraid of the glittery glare.  For each great thing I find in someone else, I find the one missing memo that completes my own self-loathing file.  It\'s ugly too, the file feeds off of meeting bigger and better people, so that destroys all hopes I have in moving into higher circles and being comfortable there.  I look around in awe with slumped posture at all the amazing people that grace this planet...I try to emulate their actions, their words, their alcoholic beverage of choice, this only seems more pathetic...I have opinions, strong ones, even regarding the shoes I wear, or the state of the nation, but I\'m scared to step away from the crowd with my orange shoes, because me with my stand alone opinions and fashion statements would just be ridiculed...but this is what constitutes leadership and confidence..why should I be afraid to order a rasberry kamikaze, even though everyone is in a mojito/or vodka mood...  I\'ve tried to figure out my faults, list them, methodically (contributing to a very BIG section in the file mentioned before) and \"work\" on them...this approach works because logistically on paper I have no complaints, I\'m professionally, physically, morally, financially, right on the GO on the Monopoly board..but it\'s in my head that needs \"work\"...It\'s like I need to keep playing self help phrases on tape to hear that \"you are ok...you don\'t need to be perfect...you should embrace the shortcomings...you are ok...\"  '),(13286,3560141,'13,August,2004','2004-08-13 00:00:00','   urlLink    one of the main reasons I love emoticons&nbsp; urlLink  '),(13287,3560141,'11,August,2004','2004-08-11 00:00:00','   urlLink    one of the main reasons i love google&nbsp; urlLink  '),(13288,3560141,'11,August,2004','2004-08-11 00:00:00','   Slowly but surely every circle comes back to its beginning...I always end up here, that area in my mind where I rarely want to venture....  It\'s hard for me to come to terms with the fact that my moods can be described as polar, bi- , tri-, or quadruple still yet to be judged....My highs are wicked highs, I can charm a fortune right out of a cookie, and noone or nothing can dent this feeling...My lows are gruesome lows, where I can walk hand in hand with anxiety attacks over the simplest of things, from grocery shopping to cooking, to working out in the gym...  It has been said that each student in this profession has some element of obsessive compulsive disorder...harping over details that eventually compel one to be unmoved from a single task..the sole reason that reading chapter after chapter until the exam is done seems realistic....  I wish I could find the courage sometimes to seek a psychiatrist, not for therapy, but just for their mere power to prescribe.  A mood stabilizer that would leave me baseline all the time...but the costs scare me...granted without insurance the prices can seem infinite, but what I really meant was that in the end...I would trade off the wicked highs too.  The lows at my age are recognizable as surmountable, as ending very soon, but at my age the highs also come with the thought of impending doom all the time.  I am also the type of person who becomes depressed without excitement, wrinkles in my linen, missing pieces in a puzzle, so with medication, I would end up depressed I think.  Imagine if I could never react appropriately, with sadness, anger, happiness, or ecstasy, if it all were somehow numbed.  I can\'t survive in a box without doors or an open top, I would suffocate, I have to be able to escape the norms in order to thrive....  What breeds a being like this...I know my character flaws, but I don\'t know where they come from?  how to prevent someone else from feeling like this?  See, at this very moment, things are so defined down to a timetable that I feel trapped, and therefore depressed.  I can\'t actually believe I did this but this morning I woke up late PURPOSELY...took my time, drank a cup of coffee as if I didn\'t have lips, spilt it on my shirt so I had to get dressed again, smoked a cigarette, and then jumped in my car and sped through traffic, weaving in and out...driving recklessly, running to my floor so as I wouldn\'t be late...slipped in ninja style...rushed through paper work that was over an hour late, but finished it so I could revel in my version of \"living on the edge\"...I could have potentially been yelled at, made to look like a fool, but I risked it so I could be prepared despite being late, so I could get a rush from the challenge...who does this....and is conscious of it??????  '),(13289,3560141,'10,August,2004','2004-08-10 00:00:00','   I don\'t know who that other blogger was who wished that they were oogled at for even five seconds...she must have used my password and hacked into this oh so personal website...  Ok...demonstration of fickle-ness number three thousand and four.... On a construction site in the middle of a fully functional hospital, didn\'t know this filled health codes, but I guess it\'s all within regulations...the old cliche of greasy construction guys cat calling women needs an updated more perverse version...I find myself two feet from some man in kneeling before his century marker on the life ticker, who thinks that incessantly turning his power drill on and off is somehow supposed to make me think of making sweet love to him right there in the elevator.  Oh, but the kicker...he\'s licking the three hour old mash potatoes off of his crusty lips, so what...so I could feverishly make icky squishy in my underpants...NO....he proceeds to place the drill near my ear and tell me that he would love to do things to me...like what build me a bird house, lay some pipe...oh god....I wish if I did have a feverish overwhelming feeling it would have been because I, either slapped the potatoes right off of his face, stuck the drill up his nose hole making him devoid of a midbrain, or lastly jump on him and returned his sexual aggression because we all know half the enjoyment is the power issue and the other half is knowing that an aggressive woman is no longer attractive but psycho....  And yeah man, my new roomate...well, no matter what I try to wear, short of a turtleneck, if I catch this guy fixated on me pouring cereal into a bowl or using the remote control once more, I\'m going to ask him to leave...well, at the end of this month, on a fixed loan income, we all have our price....He\'s never been within this proximity of a girl like ever...poor guy....I should be empathetic and I think that\'s where I stood about two weeks ago...but now, on top of today....I feel like putting pork in his figs and almonds...  And Poopie pants, my patient, the youngest patient on the ward, cute round chubby boy about twice my size, ok...he has a liking for turnovers....tickled me...like reached behind me, a woman more than twice his age, who is his professional therapist...says i\'m cute and have I ever heard of a blow job....I walked away....to find him a few hours later, giggling....again with the blow job business...I flipped..some therapist right?  I said, \"Noone here is nice to you, I try to treat you like a normal kid...like respect you, and you talk to me like I\'m some kid on the streets...I\'m more than two times your age, show some respect...and that\'s it, this is all very inappropriate and I\'m not talking to you anymore....\"   I\'m not talking to you anymore...this is my response when me and my little brother fight...this is my response to kids who called me names back in grade school...this is my response to a psychologically disturbed 12 year old boy who poops in his pants....I apologized...We worked on our \"boundaries\"....but after today....or tomorrow when he\'s discharged...I\'m not talking to him anymore.....  I\'m never wishing for anything anymore, not when a penny falls into my palm near a well, not when candles are being lit on my birthday cake, not when Shaq appears as my genie...never...Oogled by cute boys, with big hearts, and sharp wits, totally different from an oogle just to oogle...I\'m more self conscious...I avoid eye contact...I\'m repulsed to be honest....so, after today.....  '),(13290,3560141,'09,August,2004','2004-08-09 00:00:00','   urlLink    wisdom:&nbsp; urlLink  '),(13291,3560141,'09,August,2004','2004-08-09 00:00:00','   urlLink Five of Wands  (reversed)  in the Lesson position  Position Meaning  The card in this position offers a bottom line message that reflects what you might need to learn or fully comprehend before you can derive maximum benefit from your present situation.   Positional Interpretation   Five of Wands, in this position, suggests the importance of the clichÃ© \"be prepared\". There is a process that needs to be followed in order to become successful in whatever you chose to do in your life. \"Winging\" it, waiting for it, or hanging on the sidelines, canâ€™t offer the same interaction, training, or practice needed to learn to work well, or socialize among your peers, with those of the same gender, or within your field of interest. It is important to get out there and do it, not just talk about it. Getting hung up on petty annoyances only serves to waste or scatter energy, and might tend to make you irrational. Victory comes from skill, direct experience, perseverance, focused energy or adrenalin, healthy rivalry, and readiness, not luck. Excess within control.   Five of Wands. Card Interpretation   The Five of Wands is Strife or Practice, and it can represent the unrest or a turning point that can lead to confrontation, which can result in either altercation or renewal. Specific examples of this can be seen in children whose play has become rough or hurtful, in shoppers at a major sale, in a couple whose roughness turns to passion, or in the mental work that drives you to conquer, overcome, or get carried away. By the same token, it can be seen in the control and efforts to imitate that can foster growth, learning, or maturity, such as the \"play\" seen in young chimps as they try to mimic their elders.  The Five of Wands asks that you recognize that everyone has issues, is different, and may need Grounding from time to time. The question is; do you want to get in the middle of the fray, or can you do something to bring organization, clarity, or harmony to the situation? This card also suggests that turning to outside council or advice could keep you from being caught off guard.  '),(13292,3560141,'09,August,2004','2004-08-09 00:00:00','   So, there might be opportunities for a medical student to make some money on the side... I might be able to say goodbye to fixed incomes and no savings because of limited loan funds....  If I can draw blood, collect urine, and check vital signs of middle aged couples in search of life insurance, I can wave goodbye to all these things...  My life sucks...as of this moment it really really sucks...my gums itch...I\'ve gained like four pounds from merely drinking water with lots and lots of sushi, I\'m surrounded by the negativity of two dueling roomates, one of which who fights with his fiancee every night over wedding details, and then chooses to share these feelings because god only knows that\'s what I really want to hear, is all the matrimonial bliss talk...everyday, I am potentially at risk for being attacked by overzealous psychiatry patients..today it was almost being decapitated by a connect four board....I have no friends outside of medicine, so basically, coversation can only run the gammet of blood draws, urine collection, vital signs, and other students potentially being decapitated, a recap of what I am loathing at this point...I don\'t see the light....not even a small night one...I am drawn to a character who thinks that I\'m way too mushy, and I think my calls are even starting to smother him...I don\'t even know what to talk about anymore, because my days are so blooming boring....I went to a dreamy club in the clouds this weekend, and spent an hour smoking, an hour dancing, an hour surrounded by good looking guys who were all intelligent, and guess what all were in medicine...but what card did I pull the one where I\'m totally turned off by anything medically related, so basically any line these guys would pull...\"I\'m a resident...I\'m in my first year of residency...I really know what it\'s like in your shoes, I could give you some good advice...here\'s my number...blah, blah, blah..\", I refuted with a resounding...\"I\'m dropping out..not interested.\"   I\'m not really dropping out, maybe mentally I am....as in...outside of studying on my own, or being in a hospital mandatorily, or working for this life insurance gig...I\'m not in medicine, I\'m actually a free lance anything else....  I work out, I could train for a marathon, but with my addictive/perfectionist personality...I would begin to think 24 miles and 96 lbs. just wouldn\'t cut it... I watch movies, I could go to more independent film opennings....but with who, a lesbian interested in, \"picking up chicks\",  the Armenian, who thinks being sensitive is beating the other personalities out of pschizophrenics, the new \"doctors\" I just met, so I can listen to them rave about my grave future, or my one female counterpart...who is surgically attached to her boyfriend, and therefore more blissful matrimonial bullshit...maybe the one little guy who had a rotation with me who had great taste in books and food, might be fun, but surprisingly, he\'s a little too philosophical for my tastes...I love to be abstract, but what\'s fun about discussing life\'s truths...my blog is all I really need....maybe the new girls I met on this rotation in Psychiatry...one of whom has a weird conspiracy theory neurosis, involving a serious problem with trust and who always assumes I\'m hiding something...one of whom could be a man, I\'m not sure....one of whom talks about one thing all the time, no seriously, all the time...one of whom has the most perfect life, as previously discussed in another blog, well, she is that perfect Indian girl....maybe she would rub off on me and destroy my plot to be the black sheep forever....who\'s got the neuroses now??  YOu know what..maybe it\'s noone else\'s fault but my own...meaning I used to be the type who could revel in being compatible with all of these aforementioned personalities, but lately, I\'m just tired...tired of trying, so without effort I can only hope to redefine loneliness....  I could go to the beach in these last remaining weeks of summer here in the city...but why would I purposefully want to contract some fungal infection or step on someone\'s used needle, the beaches are a festering microbial playground here....the Jersey Shore is just not that accessible...  I could play sports since playing tennis involves an hour commute and court fees that amount to only thirty dollars an hour, or play golf, at a course also one hell of a commute with huge course fees, but I could join my neighbors for a friendly game of handball....sounds attractive right?  Oh what am I complaining about...I get to go to some obscure openning of a movie this week and I\'m going to see a show next week, and I\'m going to the Park this weekend to partake in some South Asian networking thing...volunteering and hoping this will take the edge off of desperacy when I meet other people outside of medicine, because you us South Asians, so diverse in our career choices...great....  anyone...anyone wanting to trade boring seemingly lonely existence in Queens, for a life unchartered and full of spontaneity....  '),(13293,3560141,'07,August,2004','2004-08-07 00:00:00','   What I hope:   Oceans of Envy Iâ€™ve got a photo booth picture It reminds me of something you said to me If everything you want is so far out of reach If everything you want is so far out of reach Move a little closer to meI did a perfect imitation of someone whoâ€™s alive Before I met you Now colors seem to have a taste and a temperature And everything doesnâ€™t seem so far away Forever feels like itâ€™s never going to be enough Held my breath as the water rushed in I was drowning in the man I use to be A castaway but you were there for me Held my breath as the water rushed in I was falling through a faded memory Castaway and you were there for me   What I could be:   Bark No Bite  I canâ€™t let you go Donâ€™t let the dark Be the only place you see me Iâ€™ll try to find you something new That pulls me back to your arms Little towns are the way weâ€™re living Little white crosses on dangerous curves Everyoneâ€™s just dying to get out They got skills but forgot the nerve Turns me on when you turn me down Everything inside me turned around Youâ€™ve been tried and youâ€™ve been true And I know that pulled me through I donâ€™t mind you donâ€™t know where youâ€™re going I want to go with you All the things you thought you handed down They donâ€™t shine like they shine for you I remember your phoenix shut itâ€™s eyes And scratched it into my shoulder Like everything I wish would die Repeats itself over and over Have I tried? Am I true? Did the things I said Iâ€™d never do All that faith and luck and trust If itâ€™s not real it turns to rust It turns me on when you say your mine All the things I thought Iâ€™d never find Your attention has itâ€™s price But itâ€™s never bark no biteI canâ€™t stand the way you tell it Turns me on when you turn me down Everything inside me turns around Give me faith and love and trust Or just crush me into dust Iâ€™m not afraid to pay your price Because youâ€™re never bark no bite   What I know still exists:   Cumbersome  she calls me goliath and i wear the david mask i guess the stones are coming too fast for her now you know I\'d like to believe this nervousness will pass all the stones that are thrown are building up a wall I have become cumbersome to this world I have become cumbersome to my girl I\'d like to believe we could reconcile the past resurrect those bridges with an ancient glance but my old stone face can\'t seem to bring her down she remembers bridges, burns them to the ground I have become cumbersome to this world I have become cumbersome to my girl too heavy too light, too black or too white, too wrong or too right, today or tonight cumbersome too rich or too poor, she\'s wanting me less and I\'m wanting her more the bitter taste is cumbersome there is a balance between two worlds one with an arrow and a cross regardless of the balance life has becomecumbersome   What it is:   She\'s got a ticket but you need more than a ticket to get you there You need a string of good intentions and an easy burn Maybe some good luck with the cash tucked in the back of your pocket Maybe it\'s still not enough to ease your pain you get tired and say fuck it Each little mystery is safe in your heart you\'ve got to wonder, is moving on the best that you can do? Need to have a little conversation Find the truth behind the questions that are haunting you She says there\'s motivationtucked into the crease between the seats I looked through the box for cigarettes, maps, keys There\'s so much room in hereyou can put your feet straight out the dash She turns the radio on pulls me close to hear me laugh Each little mystery is safe in your heart you\'ve got to wonder, is moving on the best that you can do? Need to have a little conversation Find the truth behind the questions that are haunting you She rests her hand along the space between my neckline and my back I can feel her fingers runnin\' through the feelingsI didn\'t think I had. Likemaybe this road we\'re on is the very thing that\'s come between us maybe it\'s the highway to the very place that might just save us Each little mystery is safe in your heart you\'ve got to wonder, is moving on the best that you can do? Need to have a little conversation Find the truth behind the questions that are haunting you Each little mystery is safe in your heart you\'ve got to wonder, is moving on the best that you can do? I hear you\'ve got a little education It\'s not enough to satisfy what\'s haunting you  '),(13294,3560141,'04,August,2004','2004-08-04 00:00:00','   What I fear:   Made to be Broken Some things are made to be broken Some things are made to be kept Weâ€™re in the space between whatâ€™s over and whatâ€™s left If I tried too hard to keep you I did it so I could reach youI had to re-arrange the only things that worked for me Part of this happened slowly and part of it overnight What happens when everything you want never arrives? Some things are made to be broken Thereâ€™s a face in here from the past Itâ€™s and ocean under the glass I hang it up on the wall and never look back Little scar shaped souvenirs Are the only witnesses here To a faded alibi that used to be mine It used to be mine   Seven Mary Three lyrics  '),(13295,3560141,'04,August,2004','2004-08-04 00:00:00','   I was watching ESPN this morning, a perk of living with two testosterone filled beings...and I saw the SF Giants uniform...I remember being about 12, and having a major \"poser\" disease, my cousin who was so cool because he got to use gel and wear real Polo shirts, had a poster of Will Clark that was about three feet by two feet, your standard poster. Every morning he would stretch, pretend to pull his baseball cap up and down, and pick up his hands like he was practicing his swing. He would look over his shoulder after each trial swing to see if he looked just like the poster. Little did he know that his dweeby cousin was looking over her shoulder to see if she looked anything like him. I bought a Will Clark poster, I don\'t even know who Will Clark is....I never used gel, but I did manage to score a Polo shirt.      Onto other topics, Parker, a social worker that is on my unit at the hospital bought all the kids dunkin\' donuts today...every kind, chocolate (1), chocolate frosted (5), coconut (3), crullers (2), apple cinnamon (2), boston creme (2), and good ol\' jelly filled (3)...she says before the herd gathers, grab your favorite....      I began to think out loud, ...\"Chocolate is my favorite, but there\'s only one...there\'s five chocolate frosted, so maybe I should take one of those, noone would notice because there\'s four left over...but if I take the one chocolate, noone would fight over it, because noone would know it was even there, hey parker which is your favorite..\" At this point she may as well have pulled up a couch and picked up a pen and pad to analyze my overanalytical behaviors....Without a word being spoken, I realized how I treat everything in life like this, instead of grabbing my chocolate doughnut, which I know I would enjoy, I am way too worried about being caught red handed making myself happy, without concern for the herd...the funny thing is it used to be something I would force myself to do, in order to please other people...now I think it\'s just conditioned behavior to overanalyze till I don\'t even remember what my favorite doughnut is anymore...and the fact that I\'m writing about it too....a day later, yes....overanalytical...  '),(13296,3560141,'04,August,2004','2004-08-04 00:00:00','   Weddings are beautiful, receptions are fun, but the toasts are the highlight for me...it\'s the perspectives of a couple, how they\'re seen in the eyes of those closest to them...or in some cases, those of us who really soak up the spotlight, me I prefer the lost candles and the speed walking ring boy action...my sentiments, I\'ve mentioned just seem to come out of my mouth like an oversoaked spit ball...  Oh yeah, what i really wanted to write about..well, my choice in using simple scenarios fit for reality tv to decide if I think this is all a dream has proven successful...  first..i thought there\'s no way we could hang out without other people, for longer than a day, without getting bored, irritated, or smothered...the pilot for this show is so appropriately named...\"Life after Elimi-date, When there\'s just ONE\"+++++Ratings were ****, easily...not just four asterisks, but four stars....fabulous city things, we did from Broadway to the Yankees, to NYU bars and cute restaurants and shopping...station executives predicted a good first season...good ratings were in the future...I was still skeptical..because...after the fairy dust lands, how does the fairy tale end? After the flames would there only be ashes?      second...i thought even with the fabulous could he share his personal space, and more importantly, could we share our personal space (i.e. our city...Miami), could we potentially just fall back into old habits of him taking me for granted and me creating drama which would end things abruptly...the premiere episode also appropriately titled, \"CSIO - Miami - Can\'t Slide Into Old Miami)...On it\'s debut, well, ratings remained in the ****1/2  realm....stellar reviews from critics, for romancing me in a city I thought I already knew...sushi, saki bombs, champagne, within the first ten minutes of this hour long show...the climax (haha) was reached during the scene on his pier, the resolution left the viewing audience waiting for more, but had the season started and ended with its premiere episode?  Had they already reached a plateau in the first week?  Now...      thirdly....remove active agent causing disinhibition (alcohol),  remove almost all outside contact, never go anywhere where entertainment or food exceeds 50 bucks (i.e. fairy dust), introduce the dreaded \"my friend\'s wedding\" into the intinerary, and I was almost fearful of the show being cut from the Fall lineup on CBS opposite Friends...This episode was called \"The Simple   Life\"...without Nicole Richey and Paris...or the pink pick up truck...it was replaced by a gray Dodge DURANGO...kid can you say, \"HEMI?\"....without the bikini clad buxom bleached blondes, the studio execs were fearful of the show\'s staying power, would old fans appreciate the characters maturing so quickly, and would new fans be able to appreciate the sobriety of the whole episode?  Ratings shot into the ***** 5 asterisk range, with over one million viewers.  The simple life gave me time to appreciate him as a whole person, never at a loss for words or laughs and finally tears, \"What every good TV drama should milk from its audience...\", raves  Rolling Stone  The sex scenes did not lack luster or passion since being moved from it\'s premium cable to basic cable status...critics wondered would sex in the city be just as HOT....not a worry ladies and gentlemen...the scenes have greater weight when you can replay them over and over again without empty blanks in memory due to alcohol, it\'s like you Tivo\'ed this episode....there were no rifts or dissent sensed among the characters, everything just flowed smoothly, and the scenes where they just looked at each other, and talked about the next episode, trying to keep the audience guessing as to whether to expect a SEASON or SERIES finale...were amazing....altogether the best episode yet...I can\'t wait to see how this show holds up in the Fall lineup on CBS against \"Friends.\"....With over one million viewers thus far this season...wonder if we could get some sponsors...some other investors who believe in the success of this show????  '),(13297,3769478,'31,July,2004','2004-07-31 00:00:00','   PADERBORN   (Panzerland)  -- After nearly three weeks in offline design and development, the forums for Panzerland came online on the 28th. Kaiser Von Wulf happily announced that the forums would now be open to the public, after a private testing phase on various browsers was done to insure compatability. The forums will prove to be a major part of future Panzerlander development, serving as the communications hub of the online-based Empire. Yet the number of sub fora have, thus far, been kept to the bare minimum. A general discussion area, a beer hall for entertainment, the kaiser\'s court and several other government branches have been established. The Kaiser noted in an interview that a forum for every singular category of government would be wasteful at the present time, but further noted that additional fora may be added if and when they become necessecary. In the meantime, activities will be focused on the present fora and their requisite discussions.  Design of the forums was inspired, in part, by the Kingdom of Gotzborg. Their own astonishingly well-designed forums, designed by the talented HRM King August of Gotzborg, share both cultural and graphical qualities with the designs used in Paznerland. Yet in order not to plagarise the work of His Majesty, only the basic premise of the design was used, such as transparent buildings/cityscapes adorning each forum. Attention of the nation will now be focused from forum and website to more pertinent matters, such as development of the constitution, the national history and gaining more citizens to fill vacant government offices. The Kasier noted, however, that he hopes to hold off on constitutional development until such time as more citizens are available to provide input on its contents. He has instead begun work on the detailed history of Panzerland, of which only the first two chapters are complete. In addition, he has begun an ad campaign to gain more citizens.  Meanwhile, the Kaiser has opened the beer hall to the public and invited anyone who wishes to, to stop by and say hello. Likewise, the Ministry of State is also now open and accepting foreign dignitaries for those who wish to establish diplomatic relations with the Empire. Either way, we hope to see you in Panzerland.   Related Links  Forums  urlLink http://p205.ezboard.com/bpanzerland   Website  urlLink http://www.geocities.com/evilsnapple/ '),(13298,3769478,'25,July,2004','2004-07-25 00:00:00','   The Empire of Panzerland was formed today in an official ceremony headed by His Majesty, Kaiser Ryszard Von Wulf, the first Emperor of Panzerland. The Empire was formed out of the ashes of the failed Virtual City of Netropolis project, which was cancelled due to unforseen obstacles, most unfortunately. In lieu of this project, Netropolitan Duke Ryan was re-christened with a heritage&nbsp;German name at the Cathedral in Aachen as Ryszard Von Wulf. With the Von Wulf assets established and other business agreements brought to closure, Ryszard purchased a division of Panzers with his own money, a parcel of land, and established the foundations of Panzerland. Embracing the noble military teachings of such great men as Erwin Rommel and Manfred Von Richthofen, a basis for the Empire was drawn up and organised.  Panzerland will fall under an administration echoing military organisation and a chain of command similar to that of an army. The Emperor rules absolute, with a close circle of advisors and staff. This is not to say, however, that Panzerland will be entirely without a voice for its people. The Emperor has assured the public that there will be a legislature established in which the public can make a difference in the affairs of government. Military and the art or war will, though, remain the dominant moving force behind Panzerlander society. After all, when literally translated, Panzerland means \"Armor Land\", and indeed, Panzerlanders are an armored people. More details on the progress of Panzerland will be posted as it becomes available. '),(13299,3769478,'13,August,2004','2004-08-13 00:00:00','   PADERBORN   (Panzerland)  -- The Republic of Slobovia officially became Panzermannsch Slobowien this week when Kaiser Von Wulf concluded the deal with Slobovian President Troy Thompson to make the nation a protectorate of the empire.  The ruling was made in the form of a decree issued by the Kaiser, which sets the rules and stipulations by which Slobovia will be governed whilst it is part of Panzerland. This includes guarantees of maintaining territorial integrity and recognising Troy Thompason as the leader-in-absence of the nation. Upon Mr. Thompson\'s return to micronationalism - of which the details have not been specified - Slobovia will be allowed to determine its own path. Be it a resumption of independence or in maintaining its affiliation with Panzerland.  In the meantime, the Panzerlander government has assumed control of all Slobovian state assets and government offices. These will be incorporated into the imperial government  and Panzerland will attempt to rejuvinate Slobovian foreign affairs and treaties, upholding them as if they were signed by the Kaiser himself. Future foreign affairs will, however, be directed by Paderborn. Kaiser Von Wulf has chosen Don Diego de Almagro to hold the position of Slobovian Steward while it remains a protectorate. Mr. Amagro has been hired by the government via a worker visa, and thus does not hold citizenship. But his past record shows him to be an experienced leader and he is sure to prove an apt choice to guard the Slobovians while their leader is away. '),(13300,3769478,'03,August,2004','2004-08-03 00:00:00','   KREISAU   (Panzerland)  -- The Republic of Slobovia entered negotiations this week with the Empire of Panzerland over the possibility of becoming a protectorate. With most of its citizens having gone inactive, the fate of Slobovia has been questioned abroad. Slobovia, which has been declining in activity since it seceded from the Arminian Community over two months ago, culminated in its slow-down with the announcement by President Troy Thompson that the nation would be \"making serious changes\". These revolved around Mr. Thompson\'s own departure, albeit temporary, from micronations for an unspecified amount of time.  This prompted Kaiser Von Wulf, at one time a Slobovian himself, to speak with President Thompson about the future of the Slobovians. As it would turn out, Mr. Thompson was seeking a friendly state of similar interests to establish a protectorate over Slobovia while he is away. Such a protectorate would have to, of course, guarantee the territorial integrity of the nation and make assurances that upon the return of the President, Slobovia be allowed to determine its own future. Von Wulf has made these assurances and has entered into stronger negotiations with Thompson in order to finalise the deal.  These negotiations are continuing and are expected to culminate in a formal protectorate treaty being signed between the two nations and Slobovia becoming the Panzermannsch Protectorate of Slobowien within the next few weeks. More information on this event will be posted as it becomes available. '),(13301,3767918,'27,June,2004','2004-06-27 00:00:00','    Hi folks, today I create my first blog. I\'d never thought of making a blog before, I made this one just to get a Gmail account :P but hopefully it will be fun.   What I\'ve Been Doing:   Well yesterday was my girlfriend\'s birthday and she came over for the whole day - most of which we spent naked ;) I bought her a necklace, a union jack thong (that she asked for and looks great in :D), a rose and, using the Far Cry engine, I made her a map - a collection of desert island that spelt out \"Simon [heart] Kirsty\". She loved it all and I got a very... nice thankyou ;) Oh, and I made her a cake and dinner too.  I had maths homework to do, but I\'ve forgotten how to do most of it, so whether I will get round to it is questionable.   What I\'ve Been Playing:   I\'ve been playing Hitman Contracts on PC the last couple of days and I\'m liking it. I seem to resort to violece a bit too much for the genre but its fun and that\'s what counts.   In the News:   According to Chinese state media, dozens of drug dealers have been sentenced to death. Not because China is that tough on drugs/barbaric, but to celebrate International Anti-Drugs Day! How quaint...  Michael Moore\'s new film, Faranheit 9/11, has made $8 million on the first day. I\'m quite looking forward to seeing this film [expect my opinions on it soon] but I fear most people will see it because it\'s the thing to do, and not really watch the film the way it should be watched.   Other Stuff:   All should be on the look out for a new worm being spread by popular websites. The code exploits a hole in Internet Explorer and redirects you to a Russian site that tries to install the code. To check if you are infected:  Click the Start button and then click on Search   Make sure you choose the option to look through all files and folders   Search for files called Kk32.dll and Surf.dat   If infected use up to date anti-virus software to remove malicous code  That\'s all for now, see you... whenever.  cvx  Thought of the Post: Is mockery a form of ignorance or a form of enlightenment?   '),(13302,3767918,'28,June,2004','2004-06-28 00:00:00','    Hey ho. Back from school, felt like a very long day and I\'m uber tired so I must get to bed early tonight. Unfortunately this means I will have more homework to do tomorrow which means I won\'t be able to play Planetside until Wednesday which I really need to [see  What I\'ve Been Playing ]   What I\'ve Been Doing:   School today. My English teacher tells me that she forgot to right on my coursework draft that I had to put in a load of stuff on camera angles (media piece on Bowling for Columbine) so I have English coursework to do :( I also have to revise for my maths exam on Wednesday and do loads of science homework for then. Not... cool...   What I\'ve Been Playing:   Hitman Contracts last night, got quite far and got another \'Silent assassin\' rating so all is well. Also, I have been up for a promotion to Legion Officer in my Planetside outfit the Renegade Legion. This, the second highest rank available would be totally awesome but I have little time to play right now. I was hoping to get a to of commanding done this week but I may not have time :(   In the News:   Two days ahead of schedule, the US have handed power back to the new Iraqi government and Prime Minister Iyad Allawi. The bringing forward of the hand over was probably to prevent any large scale terror attacks that may have been planned to mark the occasion.  At the NATO summit in Istanbul this week, the 26 NATO states have agreed to offer training to Iraq\'s new Security forces. Hopefully this will see Coalition troops leave Iraq sooner as their own forces can take more of the weight.  Also at the summit, police clashed with protestors - protesting about pretty much everything - and used water cannons, batons and tear gas to disperse the crowd.  To fight of Google.com\'s Gmail service, currently in the beta testing stage, Hotmail are boosting the free email storage size to 250MB. This however is still far short of Gmail\'s 1GB.   Other Stuff:   I\'ve decided to request an increase in my allowance. I currently get a meager Ã‚Â£5 a week which I can\'t live on. Ã‚Â£10 or Ã‚Â£15 would be much fairier I feel and bring me inline with my friends. Whether I get it is questionable... Was hoping to take Kirsty\'s cake round today so I could see her, if only for a few minutes but my family ate it all :( So I won\'t see her until Saturday evening which pisses me off because I miss her soooooo much!  cvx  Thought of the Post: If one experienced everything it is possible to experience, would they still have desire?   '),(13303,3767918,'28,June,2004','2004-06-28 00:00:00','    Hi folks. I am currently sitting in a business studies IT lesson. I\'m rather annoyed because Lizzie wanted to stay the weekend in Bournemouth with me Kirsty and Courtney, but my parents (for the first time in years) refused to let me. So I\'ll only be going for the day - sometime in the summer.  I can\'t see Kirsty until Saturday which is annoying and it will only be in the evening. She\'s going shopping with her birthday money and Lizzie won\'t let me come because her boyfriend Courtney can\'t come and they were meant to be fucking all day but Kirsty dragged her along lol.  So I\'ve got this week free. I\'ll probably go into town with some people and spend as much time as possible sleeping - I\'ve been very tired recently.  Also, I\'ve got sports day on Thursday which sucks arse. Hopefully I won\'t be doing anything though.  My mates Ben and Vicki broke up over the weekend. They had been going out for at least a year I think so that\'s a real blow. Apparently he just stopped fancying her...  But I think he still does. I think, after a certain period of time, the awe of how beautiful someone is just wears off and you kind of take it for granted how amazing they look. I\'m not sure how involved the were sexually, but maybe it just got boring. I mean, if you do the same thing day after day it must just get boring and you stop looking forward to it and it seems like you don\'t fancy the person anymore.  cvx  Thought of the Post: How can women understand those weird whispering sounds they make to each other?!   '),(13304,3767918,'29,June,2004','2004-06-29 00:00:00','    Sooooo tired... I very nearly fell asleep in science but was prevented by the uncomfortable chairs.  I need sleep but I never seem to get to bed early...  What I\'ve Been Doing:  School and being very tired and missing Kirsty an unbelievable amount. Hopefully I won\'t be in sports day - my team captain says he won\'t put me in unless he has to, so fingers crossed.  Episode 37 of Red vs Blue came out last night. Best one in a while. Quite long and lots of story too.  Kirsty is depressed :( She\'s got lots of coursework to do, can\'t see me all week and might not be able to go on a boat trip she\'s been wanting to go on for ages. If she can\'t go then I\'ll have to do something really nice for her (dunno what lol) and I\'ll try to make Saturday as nice as possible (next time I see her).  In the News:  Saddam Hussein is set to be handed over to the Iraqi authorities. He will be held in Iraqi cusotdy and tried by an Iraqi court for various crimes.  Militants in Iraq have released three Turkish men after they promised not to work for America again. It seems the militants are trying to keep in Arab nations\' good books by letting Muslims go free whenever possible.  A woman has, for the first time, successfully conceived a child after having an ovary transplant.  A bodged terrorist attack resulted in the injury of three Turkish cleaners... A booby trapped bomb planted on board a plane exploded AFTER passengers had left the plane, injuring three cleaners.  Other Stuff:  Grrr... Still no email about Gmail. I need a new email address sharpish...  My mother laughed at the suggest of an allowance increase so I shall procure the money from my Grandmother\'s will every week. Tis my money anyway.  Apparently I am not allowed to use the phone upstairs anymore... Looks like my parents are trying to enforce rules again. I\'m not having that. I shall continue to use it upstairs but I will now strive to phone mobiles instead of landlines and use it off-peak. Mwuhahaha. Also I haven\'t spoken to my friends in Denmark and the US in a while... I think a very long phone call to each one is in order...   ');
INSERT INTO `msgs` VALUES (13305,3767918,'30,June,2004','2004-06-30 00:00:00','    What I\'ve Been Doing:  Maths exam today. Was fine, very easy apart from one question which I didn\'t have a clue about.  What I\'ve Been Playing:  Just about to play Planetside. We\'re having a King of the Hill tournament which should be fun.  In the News:  A tube strike proposed by the RMT has gone ahead today causing widespread disruption in London.   Two militants and one policeman have been killed in gun battles in the capital of Saudi Arabia - Riyadh.  A new system is being trialled in the UK to stop plagiarism in universities. The software will search for blogs of text or entire essays lifted from websites. It seems unlikely that it will be able to detect altered text e.g.  \"America and the USSR were the main rivals during the cold war.\"  changed to  \"Russia was the main rival of America during the cold war.\"  However, it should prevent simple \"copy & paste essays\".  Other Stuff:  Sports day tomorrow... It might rain but otherwise, its quite likely I\'ll be running the relay :( oh well.  Now I\'m off to play Planetside :)   '),(13306,3767918,'01,July,2004','2004-07-01 00:00:00','    What I\'ve Been Doing:  Twas sports day today. I thought it was going to be rained off, but it wasn\'t. Luckily all my team turned up and were healthy so I didn\'t have to run :D Watched a film in English which was ok. Going to be playing Planetside tonight hopefully. I\'ll get some screenshots to make a better signature for Vasquez (my character\'s name).  What I\'ve Been Playing:  Planetside. Lead a platoon last night. We resecured Amerish after a few hours and I got a few Command Experience Points. We were going to play a king of the hill game at a tower on the test server but it was just crap so we didn\'t bother.  In the News:  Saddam Hussein had his first court appearance today. He mocked the court and said he was still president. He refused to sign his legal papers and called the court a \"theatre\".  Construction of the barrier the Israeli military is constructing on the West Bank has been halted by an Israeli court. They say although the wall is legal, some sections cause too much disruption to the lives of Palestinians and must be re-routed.  Other Stuff:  Sony are working on a new Walkman model. It\'ll be a digital music player (an MP3 player that...Dosn\'t play... MP3s... lol, all will be explained) instead of an antique tape player. It is set to be Ã‚Â£55 cheaper and with considerably more space than the top iPod model (which it is settot rival). However, there is a catch. It will only play music in sony\'s own special format - so u have to buy your music digitally from the sonyonlinee music store. However I\'m sure some genius will make a program to convert MP3s to theSonyy format.  Can\'t wait until Saturday when I can see Kirsty!   '),(13307,3767918,'03,July,2004','2004-07-03 00:00:00','    What I\'ve Been Doing:  Pretty average day at school on Friday. Saw Kirsty today :D haven\'t seen her for a week. She went shopping and blew her birthday money, then we went back to hers. Went on her new (huge) trampoline and... Other stuff ;)  What I\'ve Been Playing:  Been leading platoons on Planetside, getting some CEP but not much :( good fun though :)  In the News:  I forgot to read the papers/websites/watch the news on Friday, so you\'ll just have to make something up for yourself.  Brazil has passed a law banning the carrying of firearms in public and making it illegal to own an unlicensed gun - with licensing rules similar to those in the UK. There is also to be a referendum to on whether to ban guns outright.  A car containing four Rumanians and bearing German number plates, has crashed into the security fence at the Houses of parliament. Two men have been arrested and police are trying to determine if the crash was deliberate or not.  Other Stuff:  Microsoft have upgraded their search engine to come close to the quality of commercial search engines like Google. They are also planning to release their own search engine software later this year.  The Hubble telescope has discovered around one hundred new planets orbiting distant stars. This makes the total of known planets to two hundred and thirty.   '),(13308,3320342,'26,May,2004','2004-05-26 00:00:00','    ...Because Jimmy and I had our first marital fight this morning.  It\'s about our financial situation.  I don\'t make enough money to help pay the rent and stuff so he basically threw it in my face.  Needless to say, I ended up crying and telling him if he didn\'t like it, then he knew where the door was.  So he threatened to watch what I say.  This made me even more upset.  So that\'s how he left for work; pissed off at me.  I\'m not really happy today.  I also have a injured foot.  I think its an ingrown toe-nail and it hurts like a mother!  I went to call into work today bc I can\'t stand for long periods of time and I waas told I had no choice but to come in.  My shift got changed from 12-6 to 3-6.  So I made a doctor\'s appointment to get my foot check out at 11:20.  I just recieved a phone call from the assist. manager saying I  had  to come in at 12 - that the cover from 12-3 had to work at her other store.  I told her I couldn\'t come in bc I had a doctor\'s appointment and she sighed and was like, \"fine, 3 then.\"  And then hung up on me!!  I swear, I want to quit so bad.  I have someone working on getting me a job at her office so I hope that goes through.  I need to get out of Dunkies.  They treat people like shit.  But anyways...I promised I\'d upload some pictures.  So I kept true to my word.  These are what I have right now.  I\'m getting more soon...hopefully in the next week or so.  But, enjoy and feel free to comment on them.  I uploaded the pics to my picture site.  To view them, click  urlLink here  to see them.   '),(13309,3320342,'23,May,2004','2004-05-23 00:00:00','    ...and I\'m an official married woman   It feels weird.  Being married and all.  You think being with someone for 3 years it would feel natural.  But it doesn\'t.  It feels like everything that we\'ve been through is new again.  It\'s not bad though.  I\'m really happy.  But our honeymoon could\'ve been better...  Our wedding day was just as any young girl would picture it, no clouds in the sky, clear blue skys, nice and hot (well...nice and warm).  The perfect wedding day.  We have some nice pictures which I\'m gonna have to upload somewhere.  Our honeymoon was on a cruise to Bermuda.  There wasn\'t much for us on the way to and from for us, but we managed by sticking ourselves in our stateroom the whole time, lol.  We did go to the casino a few times - spent a $150.  Bermuda was beautiful.  Hot and wonderful.  Clear waters, nice people.  I\'ll write more later.  I have many errands to run to get re-settled into my home.   '),(13310,3320342,'14,May,2004','2004-05-14 00:00:00','    ...1 more day until my wedding   I don\'t know what to think right now.  I never knew how much work really goes into a wedding until I did mine.  We got engaged a month and a half ago and pulled everything together for tomorrow.  It still amazes me how well we did.  But, we did freak out this morning about the cake.  We never heard from her, so we thought she forgot about us.  I called her this morning and everything\'s on track and I\'m going over there at 2 to make my payment on it.  There is still so much more to do, it feels, but everything is right about done.  I just have to stop over my mother\'s to get the DJ\'s # so I can call them and see what\'s up.  They need some songs from me.  Overall, I think I better log off because I have to run to the bank, go all the way to Brockton to pick up my Maid of Honor (she doesn\'t drive), go back home to meet my cousin, go out to lunch with my family, go pay for the cake, go to the spa to get pampered with Christina (the MOH), and then go to my mom\'s house to get some sleep.  Somewhere in there I\'m going to get my check from work because my car insurance is due on Monday and I need the check to clear.  ::wipes head::  So yeah.  I\'m logging off now.  See you in about a week.   '),(13311,3320342,'13,May,2004','2004-05-13 00:00:00','    2 days before I\'m married!   Am I scared yet?  Hell yeah.  Do I care, no!  I\'m going to make this nice and short.  Today I worked 7 am till 3 and everyone was in a horrible mood.  I got nagged so many times that I wanted to walk out so bad.  I had to bite my tounge from swearing at my manager because she mad fun of me!  Yes, that\'s how mature my co-workers/boss are.  But anyways, other than that, I went tanning, went food shopping, AND had my final appointment with our wedding coordinator.  I\'m such the busy girl :P  Anyways...we\'re cooking hot dogs and beans so I\'m gonna go.  I\'ll blog more later.   '),(13312,3320342,'12,May,2004','2004-05-12 00:00:00','    Hopefully I will use blogger.  I use to swear by it.  My domain doesn\'t interest me anymore.  Plus, I\'m about to start a new life on Saturday.  And to honor it, I want to be somewhere to blog about my life - other than a domain or a journal.  No one will visit most likely, or comment for that matter, but feel free to if you read something you feel you  have  to reply to.  I don\'t mind.  You\'ll probably make me feel special anyways.  I\'m gonna need to get a layout up for this thing - possibly.  I\'ll check around and see what I can do.  I\'ll be back.   '),(13313,3320342,'15,July,2004','2004-07-15 00:00:00','   Here it is...chapter 1.  Remember to comment with feedback when you\'re done.  Thank you.   &&&    Chapter 1   My body jumped as the familiar buzzing in my purse made me grab my cell phone subconsciously.  I held the phone in front of my face and my fingers began to push random buttons.  I had a text message.  I opened the text messaging folder and gasped at what the snippet line read.  I couldnâ€™t be.  I opened the message and finished reading the short sentence present on the screen.   I want to see you again . How could I let this happen?  How could I have even considered what I was doing?  It wasnâ€™t like I was single either.  I was a happily married woman.  I loved my husband more than anyone could ever know.  To any normal person, what I was doing was officially wrong, a sin even.  Under the eyes of God I was in a binded relationship that couldnâ€™t be broken without serious paperwork involved.  Yet I had another man text massaging me, calling my cell phone when he found free time, even emailing me. I know it sounds strange to anyone that doesnâ€™t know the whole story.  If Iâ€™m  so  in love with my husband then where did I find the time to meet someone else?  And why did I even attempt to?  But itâ€™s not like that, I swear.  I met him at a Hanson concert a few months ago.  And since then, my life has totally changed.   ------    â€œHoney, my mother bought me tickets to see Hanson,â€ I said excitedly.  â€œThe tickets should be here in a few weeks.â€ â€œGood for you,â€ he said with the roll of his eyes. 	Kyle knew how much I loved Hanson.  I made it very clear to him that he was not to try to change my love for the guys.  And I wouldnâ€™t change his love for sports. â€œAt least now this balances out with you going to New York in October,â€ I said.  â€œNow I have something special to look forward to.â€ 	I had been a fan of Hanson since the â€œMMMBop days.â€  My favorite Hanson was like every other girl on the planet; Taylor.  But when young Zac went through puberty, grew taller, and grew more manly, my feelings changed.  For awhile, before my desire for Zac, I thought Isaac was pretty hot; but he just never stuck to me the way the other two did. Alas, it had been my dream to meet the guys and tell them how much their lives as celebrities had helped me surpass so many obstacles over the years.  Two months before I went to my first concert in June of 1998, my grandfather passed away to cancer.  At the concert they sang â€œWith You in Your Dreamsâ€ in which I had a difficult time getting through.  I felt the boys love for their fans that night.  It made me respect them even more than I already did.  That night made me believe they truly meant it when they said they cherished their fans. The one thing I always wanted to do was to meet them; to try to convey my gratitude to them being one of the biggest supports in my life.  I sometimes wonder if that is why Iâ€™m still a fan today; if Iâ€™m trying to hold onto the past.  But I donâ€™t believe it because I know my love for them is as real as their love for me is. â€œWhen is the concert?â€ Kyle asks, glancing away from the baseball game that blares from our TV. â€œItâ€™s the day before your birthday.  The twentieth,â€ I said.  He says nothing and turns back to his game. 	I roll my eyes and wonder how I can love this man so much.  I met Kyle three years ago in January of 2001.  His younger sister had offered to hook us up and I agreed; just to see what he was like.  Two days later we became an official couple.  He is Italian, and you can definitely tell it when you visit his grandparentsâ€™ house.  I knew the night I met him that he would become my husband someday.  It was something my mother had always told me, â€œYouâ€™ll just know,â€ she would say.  I never understood until Kyle stepped into my life. 	He was different, and he was definitely not a child.  He had morals, values, and wanted to be taken seriously; yet he wanted to have fun while doing it.  And he isnâ€™t bad looking either.  Standing 6â€™2â€ with brown hair buzzed neatly to his head and eyes blue of that of an ocean in New England, heâ€™s something to look at.  Iâ€™m surprised when he choose me as his companion.  Compared to my short 4â€™11â€ frame, people must think we make an odd couple.  But we love each other, and I guess thatâ€™s all that matters. 	Ever since our relationship started, I knew it wasnâ€™t going to be like my past relationships.  Kyle moved in with my mother and me only three months into our relationship.  Three years later, this past March, I decided I didnâ€™t want to live with my mother anymore.  Kyle had been itching to move, but I knew I couldnâ€™t afford it.  But something inside me snapped, and I wanted to make it work.  I wanted to know what it was like to be on my own.  For Christ sakes, I turned twenty-three a month prior. 	Kyle and I applied to an apartment complex down the street from my motherâ€™s condo.  We got approved for the apartment three days later.  Because of the move, my health insurance would be canceled since I wouldnâ€™t be living at home anymore.  At the end of March Kyle and I mutually decided to get married so I could fall on his insurance.  We had two dates in mind.  May 8th or May 15th.  I was leaning more towards May 8th because it wouldâ€™ve been two years and one month after Taylor Hanson and his wife Natalieâ€™s wedding date.  But that date wasnâ€™t happening; it was Motherâ€™s Day weekend.  After a week of debating, May 15, 2004 was the date etched in stone that was to be when I would become a wife. 	Though there wasnâ€™t a proper proposal, I didnâ€™t really care.  All that mattered was Kyle and me.  On the weekend of April 2nd, I traveled to Maine with my father to visit my family there and to start picking out wedding gowns.  When I returned home my father handed us $60 to go out to dinner and left us bewildered.  So we went to dinner at a nice little Italian place up the street, came home and changed into our scrubs, and began watching baseball.  During a commercial, Kyle stood from his lazy sitting position on the arm chair and started a little speech. â€œAs you know, weâ€™re getting married,â€ he started off.  â€œAnd as you know, there is something that hasnâ€™t been done yet.  I have something I want to give to you that belongs to your family.  They wanted you to have it.â€ 	He continued with saying that it was an heirloom passed down to my fatherâ€™s older sister.  After he finished talking, he grabbed my hand and stood me in front of him.  Taking a deep breath, he said â€œI already know the answer, but itâ€™s nice to hear it anyway.â€  I could tell he was shaking as he got down on one knee and looked into my eyes as he asked, â€œWill you marry me?â€ in a childish tone, the â€œKyle toneâ€ as the game came back on in the background. 	Now I was shocked.  I had wanted for two years to have those words uttered from his lips.  I never thought this moment would happen.  And it couldnâ€™t have been more perfect.  It was how I dreamed of it with Kyle; some sort of sport happening in the background.  It was just the way it was meant to be, and it was nothing sort of what I expected. â€œYes,â€ I said as he prepared the ring for me.  I removed the rings I had already on my finger; an opal he gave to me our first Christmas together, and a tiny heart shaped ring with an amethyst (my birthstone) in the center.  I set them aside as he took the tiny diamond from its case and took my left hand in his hand.  The ring fit perfectly.  That night I called everyone in my family.  It was then that our unity was official.  It was really going to happen.   -------    	I re-read the message on my screen.  I wanted to see him too.  I didnâ€™t even know he was in town.  But something was holding me back to reply to him.  I couldnâ€™t do it to my husband.  But I wanted to. 	It had been a few months since I last saw him, at that concert in July.  It was now October, and the air around me grew chilly as I thought of what to write back.  Finally I settled on  I want to see you too .  I closed my eyes and bit my lip as my finger hit â€˜send.â€™ 	 What are you doing?   I silently scolded myself.  Then a thought hit me suddenly.  Did I really love my husband or had he always been a sort of backup plan? '),(13314,3320342,'15,July,2004','2004-07-15 00:00:00','   ...Welcome to Pieces of Me.  This blog space will be used to post chapters to my new Hanson story creatively titled \"Pieces of Me.\"   &&&   Leading Lady:  Nevada Blaire Ace-Thomas Leading Hanson Guy:  Isaac Hanson Genre:  Romance, Drama Rating:  R  [language, sexual content] Summary:  None.  I\'m not giving away anything this time :p   &&&   Please read and comment for me.  It means a lot when people do give feedback.  It makes me want to write and update more.  So hang on tight - chapter 1 is almost ready!  x Kerrie '),(13315,4219169,'16,August,2004','2004-08-16 00:00:00','    ::entry for yesterday::    omg so weird thing happened 2dai!! lol i was on ma way to ma aunties n i passed sum houses on the opposite road of where my house was .. and to my suprize .. I SAW SARAH!! haha!!! damn that was really odd .. she was like playin basketball wit her cuzzies lol i saw bo n i was like \'MOM!!!! ITS SARAH AND BO!!!\' lol i screamed so loud that my dad (he was drivin) thought he hit a kid and he thought that\'s why i was screamin lol!!! haha soo funni so we slowed down n i jumped outa the car n ran towards sarah n she ran towards me n we kinda collided .. HURT LIKE A MOTHER! hahaha pretty funni .. haha!!! well ne wayz ..............       ::entry for 2dai::    lol!! hmm soooo~ wat happend 2dai .. NOTHING!! haha!! went to cross country practice this morning from 8-11 omg that was sooo funni!! so im all like on the forest track with all the seniors and juiors and im like the only freshman still runnin cuz all the others were all tired so they were just doing drills ..and so like .. wow .. i got lost!! haha!! soo funni!! i was like lost for 10 minutes!! haha!! well ne wayz, yahh that\'s basically all i did this whole day besides registering but that isnt all that funn to talk about haha!! well yahh i guess im done for now! i\'ll update later!! mwuah love yohz!!                                     xoxo, joy*CEE'),(13316,4219169,'14,August,2004','2004-08-14 00:00:00','    haha! yahh, so wat happend 2dai?! lol well i went to darien for this lil shin*dig party for ma ateh estephanie because she left 2dai to go back to nj .. it was okay i guess .. but yahh .. ma rents went to this other party around like 6 or something but i didnt wanna go so i hung out with ma ateh rachel -- OMFG!! HIGHLARIOUS!!! fun fun times!! oh god! first we went to z\'pantry to get a pack but the guy there wouldnt belive she was of age and the guy was like lookin @ her ID so carfully to make sure its not fake!! haha! then we went to her house to do our nails right? and we wanted a s***e right? oh god sooo funni she\'s like .. u cant light it cuz ur nails r gona catch on fire cuz of the asitone or something like that cuz its like famible lolz! soo great! binky and max (the dogs) were outside n we were all yellin at them cuz like they were eatin the grass omg that was funny! well ne wayz, we ordered a pizza ate a few slices (oh god i was stuffed! haha) then went to walgreens to buy more asitone for our nails (which we ended up not using haha!) and while we were there, we saw this waxing thingy! we were all like \'omg! we need to get this for our brows\' so we got it .. then wen we went to my place oh god how fun! she waxed my eyebrows lol! it hurt like a mother i mean, ive gotten them done before but not by her haha! oh god!!well yahh, i gotta getta goin now! haha! love yohz! mwuah i\'ll prolly update tomorrow or later! mwuah!!!!!                                                        xoxo, joy*CEE'),(13317,4219169,'13,August,2004','2004-08-13 00:00:00','    heyy er`body .. here\'s just a few of the pix i took while on vaca in st. louis for our family reunion (it was from friday till wednesday) it was pretty funn .. here\'s a few pix .. ENJOY!    ~ma cuzzie elise holdin ma lil cuzzie danelle in line for tickets to the \'Gateway Arch\' .. dont ask me wat they\'re lookin at cuz i have no idea lol!!!~       ~danelle rockin in ma pimp rockin hav`n ass hat! haha!! she\'s sooo cuuuuuute!!!!!! lol!~      ~this is a pic of the shadow of the \'Gateway Arch\' .. i took it while in the arch looking out the window~   ~this is of the people in our family that acutally went in the arch (not many of us went because most was afraid of hights) .. sorry .. this pic is kinda blurry~        ~ma cuzzie\'s elise and kristine and me on the \'Gateway Arch\'~   ~me and ma cuzzie kristine \'tryin\' (key-word being trying lol) to act ghetto in the back of my van while on our way to the zoo~         ~me ma cuzzie kristine and my cuzzie berns also in the back of my van on our way to the zoo~           ~awwwwwwww!!!!!!! its alex rockin in ma pimp ass hat n ma aunites ghetto fab. sun glasses!! haha!! so cute!!~   ~this is the most cutest pic ive ever seen!! lol its ma lil cuzzie alex bein hugged by ma other lil cuzzie natalie!! awwwwww!!!!!!!!!!!! soooooo cuuuuuuuuuuuute **smiles enormusly** hahha!!~   ~every night we had a party up in our hotel room .. this is me @ the last one .. dont worri .. its ROOT BEER!! hahaha!!~   that\'s all for the pix for now .. i\'ll upload the others and put em up when i get the chance!! lol!! love yaz!! mwuah!!                                                                  mahal~ xoxo, joy*CEE'),(13318,4219169,'13,August,2004','2004-08-13 00:00:00','    ahummm ... blogger .... i heard it was a good place to do blogs and ma xanga was being  hav`n ass gay lezbo  on me so i decided to join here ... hence ...  \'My Fresh Template\'  ... feel free to help me out here because i have no idea wat the fuckin im doin LoL! IM me if ya\'ll wanna!!  empyr xtacy5  !! iight payce out .. ima go try to figure this out!! holla!!                                                                           mahal~ xoxo, joy*CEE'),(13319,3556477,'03,July,2004','2004-07-03 00:00:00','   Karen Kunawicz        What is the sound of one heart breaking?  It is the sound of someone curled up in a tiny ball crying softly in the night,  the sound of the first unwanted teardrop touching your skin,  it\'s the sound of a telephone that doesn\'t ring,  the sound of regret pounding inside your brain with every heartbeat,  it\'s the whispers of the toy animals he gave you.   It\'s the shuffling of feet walking away from you,  the sound of your soul shattering into a million pieces  at recognizing the word \"goodbye,\"  it\'s the soundtrack of memories torturing you,  it\'s the sound of feeble hands trying to push back the obstinate hands of time,  it\'s the sound of cherub\'s dying breath,  the sound of all those years disappearing in the vortex of Cupid\'s kitchen sink,  it\'s the unrelenting, plaintive baby meows  of an abandoned kitten outside an ignoring door.   It\'s the sound of the rain that doesn\'t ever stop,  the sound of all the doors in the world  shutting and closing in your face at the same time,  of raging, howling storms in the night when there\'s no one there to hold you, the sound of your voice as it screams back at you,  the echo of \"I love you\" burning holes in you,  the sound your heart makes as it tells you to lie still  because nothing you will ever do will matter without love. The sound of the waves at the polluted beach you went to  as it moves from the shore and crashes inside your mind,  of the sniffles that make up your pathetic \"SOS-to-the-world,\"  the cracking of the brittle black-red petals from the sidewalk vendor roses he gave, the sound of the music he used to make going to your gut. The sound of things in your room being thrown around and landing on the floor,  the caress of sharpened kitchen knives on skin,  the sound your throat makes as you swallow your saltiest tear.  It\'s the sound of your own voice calling out to someone who isn\'t there,  of winged creatures dying and falling on a city pavement,  of terms of endearment used a hundred times a day  struggling to crawl into a vacuum of forgetfulness,  it\'s the sound of your own sobs keeping you company,  it\'s the cold, uncaring stillness of the air you share your space with.   Destruction isn\'t always as noisy as bombs exploding.  Sometimes the ultimate catastrophes are as quiet  as feather falling on the floor of a Zen monastery.  No one else can really hear your heart breaking except you. '),(13320,3556477,'03,July,2004','2004-07-03 00:00:00','   (Author Unknown)      Many people are afraid of growing old.  I\'m afraid of growing old and boring.  Many people are afraid of growing old, alone.  I\'m afraid of growing old, insane.  Many people are afraid of losing their looks.  I\'m afraid of losing my dreams.  Many people are afraid of losing their youth.  I\'m afraid of losing my soul.     When you\'re 15, 35 seems ancient.  When you\'re 35, 15 seems juvenile.  A turnaround in a split second - two decades zoom past and before you know it, it\'s only a mile to the next millennium.  Don\'t\' fear age - it\'s a right of personhood.  Don\'t fear death - it\'s God\'s greatest jest.  Don\'t grow old - you don\'t have to.        Don\'t date because you\'re desperate.  Don\'t marry because you\'re miserable.  Don\'t have kids because you think your genes are superior.  Don\'t separate because you think it\'s fashionable.  Don\'t drink because you have troubles.  Don\'t gamble because you think winning is inevitable.  Don\'t philander because you think you\'re irresistible.  Most likely, you\'re not.        Don\'t associate with people you can\'t trust.  Don\'t cheat.  Don\'t lie.  Don\'t pretend.  Don\'t try to buy your way into the kingdom of God.  Don\'t dictate because you\'re smarter.  Don\'t demand because you\'re stronger.   Don\'t sleep around because you think you\'re old enough and know better.  Don\'t hurt your kids because loving them is harder.  Don\'t sell your self, your family or your ideals.     Don\'t stagnate.   Don\'t regress.  Learn a new skill.  Find a new friend.  Start a new career.  Don\'t live in the past.  Time can\'t bring anything or anyone back.  Don\'t put your life on hold for possibly Mr. Right.   Don\'t throw your life away on absolutely Mr. Wrong because your biological clock is ticking and you can\'t afford to have your eggs harvested before the new millennium.  There\'s always a mad rush to something, somewhere but victory does not always belong to those who finish first.  Sometimes, there is no race to be won only a price to be paid for some of life\'s morehasty decisions.  You can\'t always go with the throng who could be wrong.  Sometimes, you have to be alone to be enlightened.  To terminate your loneliness, reach out to the homeless.  To feed your nurturing instincts, care for the needy.  To fulfill your parenting fantasies, get a puppy.  Don\'t bring another life into this world for all the wrong reasons.  To keep yourself warm, buy a jacket. In the long-run, it will be less complicated and less costly.   To make yourself happy, pursue your passions and be the best of what you can be.  Simplify your life.  Take away the clutter.  Get rid of destructive elements - abusive friends, nasty habits and dangerous liaisons.  Don\'t abandon your responsibilities but don\'t overdose on duty.  Don\'t live life recklessly without thought and feeling for your family.  Be true to yourself.  Don\'t commit when you\'re not ready.  Don\'t keep others waiting needlessly.  Fall in love - it\'s the greatest thing on earth.  But take care and remember, after the fall must come the rise.  Go on that trip.  Don\'t postpone it.  Say those words.  Don\'t let the moment pass.  Do what you must even at society\'s scorn.   Write poetry.  Love deeply. W alk barefoot.  Hold hands.  Dance with wild abandon. Cry at the movies.  Take care of yourself.  Don\'t wait for someone to take care of you.  You light up your life.  You drive yourself to your destination.  No one completes you - except you.  It is true that life doesn\'t get easier with age.  It only gets more challenging.  Don\'t be afraid.  Don\'t lose your capacity to love.  Pursue your passions.  Live your dreams.  Don\'t lose faith in God.    Don\'t grow old. Just grow-up. '),(13321,3556477,'03,July,2004','2004-07-03 00:00:00','   (Author Unknown)     I choose to love you in silence... For in silence I recieve no rejection  Nor should I bother of what it is that prevents us from being together - its just that we can never be together.  I choose to love you in loneliness... For in loneliness, no one owns you  but I.  I choose to adore you from a distance... For distance will shield us from pain.  Still, I love you and I just can\'t help but do so...  I choose to kiss you in the wind For the wind is gentler than my lips  I choose to hold you in my dreams... For in my dreams, you have no end.  Forever.  That is what you are to me.  In all honesty, I find it hard to believe that I will ever love somebody the same way I loved you.  Thanks for the experience, I loved it while it lasted. I can\'t help but smile during those times that you threw me out of focus - those times that you made me panic as to what I could do just to appease your call.  Thanks for those times that you almost destroyed my sanity. I always knew it was done for anyway.  Thanks for all the phone calls, all the emails, and all the small things that you did. Without them, I may have not even been sure that I really loved you.  And most of all, thanks for being honest. Thank you for putting me back on track. You would be a constant reminder of a God\'s gift.  Maybe you were right:  Love is an attachment... to someone who\'ll never be yours.  Love is something we will never have... something we\'ll never experience  And I really meant my reply.  The sun maybe up, but the moon does not vanish. It may wane, yes, it may  even seem to fade. But one thing is certain. It remains in its sky... '),(13322,3556477,'03,July,2004','2004-07-03 00:00:00','   (Author Unknown)      I\'ve learned.... That the best classroom in the world is at the feet of an elderly person.  I\'ve learned.... That when you\'re in love, it shows.  I\'ve learned.... That just one person saying to me, \"You\'ve made my day!\" makes my day.  I\'ve learned.... That having a child fall asleep in your arms is one of the most peaceful feelings in the world.  I\'ve learned.... That being kind is more important than being right.  I\'ve learned.... That you should never say \"no\" to a gift from a child.  I\'ve learned.... That I can always pray for someone when I don\'t have the strength to help him in some other way.  I\'ve learned.... That no matter how serious your life requires you to be, everyone needs a friend to act goofy with.  I\'ve learned.... That sometimes all a person needs is a hand to hold and a heart to understand.  I\'ve learned.... That simple walks with my father around the block on summer nights when I was a child did wonders for me as an adult.  I\'ve learned.... That life is like a roll of toilet paper. The closer it gets to the end, the faster it goes.  I\'ve learned.... That we should be glad God doesn\'t give us everything we ask for.  I\'ve learned.... That money doesn\'t buy class.  I\'ve learned.... That it\'s those small daily happenings that make life so spectacular.  I\'ve learned... That under everyone\'s hard shell is someone who wants to be appreciated and loved.  I\'ve learned.... That the Lord didn\'t do it all in one day. What makes me think I can?  I\'ve learned.... That to ignore the facts does not change the facts.  I\'ve learned.... That when you plan to get even with someone, you are only letting that person continue to hurt you.  I\'ve learned.... That love, not time, heals all wounds.  I\'ve learned.... That the easiest way for me to grow as a person is to surround myself with people smarter than I am.  I\'ve learned.... That everyone you meet deserves to be greeted with a smile.  I\'ve learned.... That there\'s nothing sweeter than sleeping with your babies and feeling their breath on your cheeks.  I\'ve learned.... That no one is perfect until you fall in love with them.  I\'ve learned.... That life is tough, but I\'m tougher.  I\'ve learned.... That opportunities are never lost; someone will take the ones you miss.  I\'ve learned.... That when you harbor bitterness, happiness will dock elsewhere.  I\'ve learned.... That I wish I could have told my Dad that I love him one more time before he passed away.  I\'ve learned.... That one should keep his words both soft and tender, because tomorrow he may have to eat them.  I\'ve learned.... That a smile is an inexpensive way to improve your looks.  I\'ve learned.... That I can\'t choose how I feel, but I can choose what I do about it.  I\'ve learned.... That when your newly born child holds your little finger in his little fist, that you\'re hooked for life.  I\'ve learned.... That everyone wants to live on top of the mountain, but all the happiness and growth occurs while you\'re climbing it.  I\'ve learned ... That it is best to give advice in only two circumstances; when it is requested and when it is a life threatening situation.  I\'ve learned.... That the less time I have to work with, the more things I get done. '),(13323,3556477,'03,July,2004','2004-07-03 00:00:00','   Lord,     Make me an instrument of your peace.  Where there is hatred. Let me bring love.  Where there is injury, pardon.  Where there is doubt, faith.  Where there is despair, hope.  Where there is darkness, light.  Where there is sadness, joy.  Where there is discord, harmony.  Where there is error, truth.  Where there is wrong, the spirit of forgiveness.   O Divine Master.  Grant that I may not so much seek  To be consoled as to console.  To be understood as to understand.  To be loved as to love.  For it is in giving that we receive.  It is in pardoning that we are pardoned.  It is in dying that we are born to eternal life. '),(13324,3556477,'03,July,2004','2004-07-03 00:00:00','   (Author Unknown)     A time comes in your life when you finally get it... When in the midst of all your fears and insanity you stop dead in your tracks, and somewhere the voice inside your head cries out - ENOUGH! Enough fighting and crying or struggling to hold on. And, like a child quieting down after a blind tantrum, your sobs begin to subside, you shudder once or twice, you blink back your tears, and through a mantle of wet lashes you begin to look at the world through new eyes. This is your awakening.   You realize that it\'s time to stop hoping and waiting for something to change or for happiness, safety and security to come galloping over the next horizon. You come to terms with the fact that he is not Prince Charming and you are not Cinderella and that in the real world there aren\'t always fairy tale endings (or beginnings for that matter), and that any guarantee of \"happily ever after\" must begin with you; and in the process a sense of serenity is born of acceptance. You awaken to the fact that you are not perfect and that not everyone will always love, appreciate or approve of who or what you are ... and that\'s OK (They are entitled to their own views and opinions). And you learn the importance of loving and championing yourself; and in the process a sense of new found confidence is born of self-approval. You stop bitching and blaming other people for the things they did to you (or didn\'t do for you) and you learn that the only thing you can really count on is the unexpected.You learn that people don\'t always say what they mean or mean what they say, and that not everyone will always be there for you; and that it\'s not always about you. So, you learn to stand on your own, and to take care of yourself; and in the process a sense of safety and security is born of self-reliance.   You stop judging and pointing fingers ... and you begin to accept people as they are, and to overlook their shortcomings and human frailties; and in the process a sense of peace and contentment is born of forgiveness. You realize that much of the way you view yourself, and the world around you, is as a result of all the messages and opinions that have been ingrained into your psyche. And  you begin to sift through all the crap you\'ve been fed about how you should behave, how you should look, and how much you should weigh; what you should wear and where you should shop, and what you should drive; how and where you should live, and what you should do for a living; who you should sleep with, who you should marry, and what you should expect of a marriage; the importance of having and raising children, or what you owe your parents. You learn to open up to new worlds and different points of view. And you begin reassessing and redefining who you are and what you really stand for.  You learn the difference between wanting and needing and you begin to discard the doctrines and values you\'ve outgrown, or should never have bought into to begin with; and in the process you learn to go with your instincts.   You learn that it is truly in giving that we receive. And that there is power and glory in creating and contributing; and you stop maneuvering through life merely as a \"consumer\" looking for your next fix.   You learn that principles such as honesty and integrity are not the outdated ideals of a bygone era, but the mortar that holds together the foundation upon which you must build a life. You learn  that you don\'t know everything, it\'s not your job to save the world ... and that you can\'t teach a pig to sing.   You learn to distinguish between guilt and responsibility, and the importance of setting boundaries, and learning to say NO. You learn that the only cross to bear is the one you choose to carry, and that martyrs get burned at the stake.   Then you learn about love. Romantic love and familial love. How to love, how much to give in love, when to stop giving, and when to walk away. You learn not to project your needs or your feelings onto a relationship. You learn that you will not be more beautiful, more intelligent, more lovable or important because of the man on your arm or the child that bears your name.   You learn to look at relationships as they really are and not as you would have them be. You stop trying to control people, situations and outcomes. You learn that just as people grow and change, so it is with love...and you learn that you don\'t have the right to demand love on your terms ... just to make you happy. And, you learn that alone does not mean lonely. And you look in the mirror and come to terms with the fact that you will never be a size 5 or a perfect 10, and you stop trying to compete with the image inside your head and agonizing over how you \"stack up.\" You also stop working so hard at putting your feelings aside, smoothing things over and ignoring your needs. You learn that feelings of entitlement are perfectly OK.... and that it is your right to want things and to ask for the things that you want...and that sometimes it is necessary to make demands.   You come to the realization that you deserve to be treated with love, kindness, sensitivity and respect; and you won\'t settle for less. And, you allow only the hands of a lover who cherishes you to glorify you with his touch ... and in the process you internalize the meaning of self-respect. And you learn that your body really is your temple, and you begin to care for it and treat it with respect. You begin eating a balanced diet, drinking more water and taking more time to exercise. You learn that fatigue diminishes the spirit and can create doubt and fear. So you take more time to rest.   And, just as food fuels the body, laughter fuels our soul. So you take more time to laugh and to play. You learn, that for the most part, in life you get what you believe you deserve ... and that much of life truly is a self-fulfilling prophecy. You learn that anything worth achieving is worth working for, and that wishing for something to happen is different from working toward making it happen. More importantly, you learn that in order to achieve success you need direction, discipline and perseverance. You also learn that no one can do it all alone and that it\'s OK to risk asking for help. You learn that the only thing you must truly fear is the great robber baron of all time. FEAR itself. You learn to step right into and through your fears because you know that whatever happens you can handle it, and to give in to fear is to give away the right to live life on your terms. And you learn to fight for your life and not to squander it living under a cloud of impending doom.   You learn that life isn\'t always fair, you don\'t always get what you think you deserve; and that sometimes bad things happen to unsuspecting, good people. On these occasions you learn not to personalize things. You learn that God isn\'t punishing you or failing to answer your prayers. It\'s just life happening. And you learn to deal with evil in its most primal state - the ego. You learn that negative feelings such as anger, envy and resentment must be understood and redirected or they will suffocate the life out of you, and poison the universe that surrounds you. You learn to admit when you are wrong and to building bridges instead of walls.   You learn to be thankful and to take comfort in many of the simple things we take for granted, things that millions of people upon the earth can only dream about; a full refrigerator, clean running water, a soft warm bed, a long hot shower. Slowly, you begin to take responsibility for yourself by yourself; and you to make yourself a promise to never betray yourself and to never, ever, settle for less than your heart\'s desire. And you hang a wind chime outside your window so you can listen to the wind. And you make it a point to keep smiling, to keep trusting, and to stay open to every wonderful possibility.   Finally, with courage in your heart and with a higher power by your side you take a stand; you take a deep breath, and you begin to design the life you want to live as best as you can. '),(13325,3556477,'09,June,2004','2004-06-09 00:00:00','   and   The Moon sees me God, bless the Moon  and  God, bless me.  '),(13326,3556477,'08,June,2004','2004-06-08 00:00:00','   Infallible as you are to Einstein I cannot help but regard you With such wariness.   Understandably so... You, which I cannot see, smell, and touch Only feel.  The sun seizing Moments of life Dancing rapidly.  The shapeless moon  Tiptoes in the dark Leaving shadows.  Raindrops wooing Leaves golden yet shy Snow on a postcard.  Fallen friendships  Ceased breathing  Withered youth.  Love forgotten Unrequited Gone.  Love resurrected Untamed Still.  Pastâ€¦  Presentâ€¦  Futureâ€¦ You, without hands Bare of emotions.  Are you an ally? An adversary? Or just here to stay?  '),(13327,3556477,'08,June,2004','2004-06-08 00:00:00','   One day a young man was standing in the middle  of the town proclaiming that he had the most  beautiful heart in the whole valley.    A large crowd gathered and they all admired his  heart for it was perfect. There was not a mark  or a flaw in it. Yes, they all agreed it truly  was the most beautiful heart they had ever seen.    The young man was very proud and boasted more  loudly about his beautiful heart.    Suddenly, an old man appeared at the front of  the crowd and said \"Why your heart is not nearly  as beautiful as mine.\"    The crowd and the young man looked at the old  man\'s heart. It was beating strongly, but  full of scars, it had places where pieces had  been removed and other pieces put in, but they  didn\'t fit quite right and there were several  jagged edges. In fact, in some places there were  deep gouges where whole pieces were missing.    The people stared - how can he say his heart is  more beautiful, they thought? The young man  looked at the old man\'s heart and saw its  state and laughed. \"You must be joking,\" he  said. \"Compare your heart with mine, mine is  perfect and yours is a mess of scars and tears.\"    \"Yes,\" said the old man, \"Yours is perfect  looking but I would never trade with you. You  see, every scar represents a person to whom I  have given my love - I tear out a piece of my  heart and give it to them, and often they give  me a piece of their heart which fits into  the empty place in my heart, but because the  pieces aren\'t exact, I have some rough edges,  which I cherish, because they remind me of the  love we shared. Sometimes I have given pieces of  my heart away, and the other person hasn\'t  returned a piece of his heart to me. These  are the empty gouges -- giving love is taking a  chance.    Although these gouges are painful, they stay  open, reminding me of the love I have for these  people too, and I hope someday they may  return and fill the space I have waiting. So now  do you see what true beauty is?\"    The young man stood silently with tears running  down his cheeks. He walked up to the old man,  reached into his perfect young and beautiful  heart, and ripped a piece out. He offered it to  the old man with trembling hands. The old man  took his offering, placed it in his heart and  then took a piece from his old scarred heart and  placed it in the wound in the young man\'s heart.  It fit, but not perfectly, as there were some  jagged edges. The young man looked at his heart,  not perfect anymore but more beautiful than  ever, since love from the old man\'s heart flowed  into his. They embraced and walked away side by  side.    How sad it must be to go through life with a  whole untouched heart.  '),(13328,3556477,'07,June,2004','2004-06-07 00:00:00','   YoU are the RiDdLe of the MiNd  that pierces the brains and inhabits the skull a sanctuary to your mystery.  How long will you stay? How deep will you probe? How much will you remain -- A marked question?  Words without rhyme Fragments void of logic Sentence unadulterated Truth naked and fragrant.  '),(13329,3556477,'06,June,2004','2004-06-06 00:00:00','   Eternal goodness, with hearts of light Bringer of truth and justice with might  Slays the demons that wrought havoc and despair To those faithless, hopeless, without care   Unending battle for the undying light Forever vigilant to what is right One day the Spirit will prevail Death to the torments that it entails   The spirit of light resides in us all It is within, the battle in our core Ourselves is the final ground to stand For we are the temple blessed by God\'s hand   Living temple of God, Masters of creation We are key to the triumph of the Fallen\'s liberation It\'s not written but it can be done The fallen as the victor and the salvation of none   Never weary, forever awake Demons of sin exists to desecrate Spirit of light, Sentinels of the temple Defend us, we are weak, be our example   Without the Spirit we are gone Our temporal will amounts to none Just a sweet whisper from the fallen Will sever our line to heaven   Minions of the dark, unseen evil of the world Torments the desecrated with horrors untold Constantly is the time and the struggle for all Ours is the burden, are we destined to fall?  '),(13330,3556477,'06,June,2004','2004-06-06 00:00:00','   Did I say platonic? I cannot recall What matters most is the present, not was said before  I was seized by your wit and entangled in your charm Calmed by your words taintless of harm  Your prized presence, attention and thought Like the others from you, I have fervently sought  I thought it was fleeting, your addictive hold on me Waiting for it to pass thinking Iâ€™ll be free  I realized that time was not truly potent I have to break my way out from your enchantment  My reason as my sword and my will as my shield I have broken your spell, my quest fulfilled  The final blow that brought my struggles to an end Was the fear of losing you as a dear friend  Did I say platonic? Now I recall What matters most is the present and what will be from now on  '),(13331,3556477,'13,June,2004','2004-06-13 00:00:00','   Malamig na hangin kumot sa katawan nitong mandirigma, sahig ang kanlungan.  Nakapikit ang mga mata mga labing nagpipigil ng marahang pagsambit ng libong panalangin.  Pilitin mang matulog diwa ay kusang gigising lilisanin ang panaginip may himig ng uyayi.  Malakas na pag-ungol bakas sa mukha pakikibakang pasakit kalaban ay sino?  Pait ng nakaraan asin sa sugat malaya at sariwa tuluyang naghihingalo.  Impit na pag-iyak hikbing kumakawala gustuhin mang lumimot hapdi ang ala-ala.  Bangungot ang pag-asa iyon ang pamuksa sa katawang pagal sa kaluluwang patay.  Ngunit angkin niya dugong marangal hindi pasusupil anuman ang wakas.  Maliit man ang apoy sa kalungkutan ng gabi itoâ€™y magniningas kasabay ng tala.  Yakap ang tapang siyaâ€™y mahihimbing gigising sa umaga sisilay sa araw.    Siyaâ€™y mandirigma Dugo ang sandata.  '),(13332,3556477,'13,June,2004','2004-06-13 00:00:00','   I was not thinking of Eden or an island ever virgin, neither of a palace in the heavens.  I was not looking at the sunâ€™s rays blinded by brilliance or the glowing boldness.  I was not hearing chimes of yin and yang or classical music.  I was not smelling cocoa on a rainy afternoon seducing the wind.  I was not tasting marshmallows vanilla and vainly sitting on my ice cream.  I was not touching the face of Venusâ€™ perfection; nonchalant to Cupidâ€™s tickle.  When I said I loved youâ€¦ I was not shouting at the top of my lungs I was not even listening to my voice.    It was only a whisper of a heartbeat that probably got drowned in the ocean of bliss.  '),(13333,3556477,'21,June,2004','2004-06-21 00:00:00','   Sometimes, Iâ€™d like to die And kiss the day goodbye Sometimes, Iâ€™d like to lie down And lay still on the ground.  Sometimes, Iâ€™d like to weep Until I fall asleep Sometimes, Iâ€™d like to think Until my body sinks.  Sometimes, Iâ€™d like to pray That I be happy and gay Sometimes, Iâ€™d like to wish That I swim like a fish.  Sometimes, Iâ€™d like to live All the things that I believe Sometimes, Iâ€™d like to breathe All the things that go shit.  Sometimes, Iâ€™d like to start And fade away in the dark Sometimes, Iâ€™d like to end And follow how they bend.  Sometimes, Iâ€™d like to whisper â€œFarewellâ€ Sometimes, Iâ€™d like to shout â€œIâ€™m gone.â€  '),(13334,3556477,'21,June,2004','2004-06-21 00:00:00','   â€œSo this is  Barrio Mapagkumbaba .  Hmmm, I wonder where the party isâ€, Sophia Spaghettiâ€™ s noodle strands stretched as she tried to peek in at the place.  Bright green, red, blue and yellow triangle-shaped objects danced over the big crowds.  There were people everywhere.  Children came running one after another.  Different sounds filled the air â€“ bells ringing, men laughing and women chatting endlessly.  But the loudest of them all were the trumpets and drums of silver-uniformed men marching.  Leading the parade were ladies in golden dresses gracefully twirling batons and wowing everybody watching.  Sophia Spaghetti knew then that she was to be a big hit in this merry happening.  So she twirled in her thick tomato sauce, showing off her cheesy bits.  â€œMy, I always look best in this velvety red gown.  Everyone will say  bellissima  when they see me â€“ because I am truly beautifulâ€¦ mmmwahâ€.  Next thing Sophia Spaghetti knew, she was placed on a very long table covered with a huge banana leaf.  She quickly swirled her slightly misplaced meatballs and looked at her companions. What she saw surprised her.  â€œWhoâ€¦ what are you!â€  Sophia Spaghetti screeched.  The others stared at Sophia Spaghetti for a moment.   Sapin-Sapin  who glowed with her violet, yellow and red colors squealed with delight, â€œWelcome to our fiesta!â€  It was the thin noodles with lots of vegetables, which spoke next.  â€œHi, â€ she curled up shyly, â€œIâ€™m  Pasing Pancit Bihon  and I am very pleased to meet you, miss --? â€    â€œWho is she to ask my name as if she does not recognize me!â€  Sophia Spaghetti could only mumble.  Aloud she said, â€œI am So-phee-ya Spaghetti from the great Pasta family!  So, is this the party?  In this small and dusty place?â€   Pasing Pancit Bihon  shamefully hid under her cabbage bits and meat chunks.  Her brother  Cardo Pancit Canton , who was short, stout, and yellowish, answered instead, â€œThis is a fiesta!  We give thanks to our patron saint who is very good to us.  And during fiestas, we have banderitas, those colorful triangles that you see hanging around.â€   Cardo Pancit Canton  was proud of his town.  Sophia Spaghetti was annoyed because she does not know anything about a fiesta.  â€œWe have none of these foolish things in the big city.  What we have are plenty of colorful and blinking lights â€“ everywhere, every night.â€   â€œWow, you came from the big city!â€ cried the white round-shaped  Putos  as they bumped into one another.  â€œWeâ€™ve always wanted to go there,â€ they went on excitedly.  As she curved to look at them, her noodles gave out a loud swishing.  â€œHuh, what on earth are you!  You are totally black, like mud!â€   Beside the  Putos  was the deep bowl of  Dinuguan , which turned even darker.   Dino Dinuguan  grew sad over the rude remark about his color so he kept silent.      Then Sophia Spaghetti smelled something bad.  â€œWhatâ€™s that dreadful odor?  Oh, I canâ€™t bear to be on this table anymore!â€  The pot of orange  Kare-Kare  bubbled up in anger, â€œThat is my good friend  Bong Bagoong .  Never judge him by his smell because once you taste him, you will surely like him.  We make a good team.â€      â€œ Mare Kare-Kare , control yourself.  Miss Sophia Spaghetti here is our guest.  We should make her feel at home.â€   Bong Bagoong  was not at all offended by how Sophia Spaghetti treated him.  â€œLet us all introduce ourselves so that she will know us better.â€  But before anyone could start, Sophia Spaghetti sneered.  â€œI bet  Bong Bagoong  and  Dino Dinuguan  will be the champion leftovers here because they are the ugliest!  Shee-hee-hee!â€  The others remained quiet but Sophia Spaghetti did not seem to notice.  She continued talking and scoffing at the same time.  â€œWhy do you have such funny names?  But then again, they suit you well because you are all so funny looking.  I bet the visitors wonâ€™ t eat you at all. Their stomachs will ache from laughing hard simply by looking at you.â€ Sophia Spaghetti shook with laughter until her meatballs and cheese scattered.  The roasted pig lying on the middle, the King of every fiesta table, has been observing Sophia Spaghetti.  In all his shining glory, he declared, â€œA-hoinkâ€¦A-hoink.  I am the honorable Judge  Listo Lechon .  As everybody know, â€ and he looked at the others meaningfully, â€œit is my duty to restore the peace and order on this table.  Sophia Spaghettiâ€¦â€  Sophia Spaghetti hastily interrupted and spoke in her most dramatic voice.  â€œI, the fairest of all noodles â€“ the whitest and firmest of them all â€“ is expected to listen to an old dark-skinned hog like you!  My, why donâ€™t you just nibble on that apple in your mouth.â€    The rest got angry.  To disrespect the wise and honorable Judge  Listo  was the last straw.  They all prepared to attack Sophia Spaghetti with their vegetable and meat weapons.  A food fight was about to take place when, like the true leader that he was, Judge  Listo Lechon  shook his head slowly and said, â€œNow, now, please calm yourselves.  Do not ever forget that we are here for the same reason.  Our duty is to feed the hungry and make everyone happy.â€  Sophia Spaghetti, who grew scared at the sight of  Pasing Pancit Bihon  ready to hurl carrot sticks at her, now sighed with relief.  She jeered at the others once again, â€œDonâ€™t be so jealous.  Iâ€™m sure that after they finish eating me, they will turn to you.â€  â€œSometimes, itâ€™s just so hard to be the best,â€ she whispered to herself.  Finally, somebody shouted  -- â€œItâ€™s eating time.â€  Sophia Spaghetti felt the heat of excitement.  â€œNow, these dishes will certainly die of envy,â€ she snickered.  All the people flocked to the long table.  Each one wanted to taste the dishes served.  All the dishes, on the other hand, desired to be tasted.  Elders and children gathered near  Pasing Pancit Bihon .  Everyone liked her crisp vegetables and delicious chorizos.  Pasing Pancit Biho n was glad that she pleased the guests.    Everyone praised  Cardo Pancit Canton  because he was tasteful with all the large chunks of vegetables, pork, and chicken.   Cardo Pancit Canton  exchanged smiles with his sister.    The combination of  Dinuguan  and  Puto  was naturally a fiesta favorite.  So was the popular pair of  Kare-Kare  and  Bagoong .   And, of course,  Sapin-sapin  was the number one dessert.  Each dish had fun except Sophia Spaghetti.  Nobody had tasted her sweet red sauce or noticed her cheesy dress.  She felt like crying.  She never felt so embarrassed in her entire spaghetti life.  She asked herself, â€œWhy donâ€™t they want to eat me?â€   Pasing Pancit Bihon  glanced at Sophia Spaghetti and saw that she was sinking in her thick sauce.  She coiled into a smile to cheer her up, â€œDonâ€™t worry.  Theyâ€™re going to like you, too.  Itâ€™s not everyday that people here see something like you.â€  Sophia Spaghetti then felt sad as she realized how vain and proud she was.  She remembered how she mocked  Dino Dinuguan  and  Bong Bagoong .  She was deeply ashamed at the way she looked down on the other dishes as if she was the only one beautiful.  She spoke softly, â€œI am very sorry.  Please forgive me for being mean.  The honorable Judge  Listo Lechon  was right.  We were made beautiful to give happiness to everybody.â€  All the dishes gave Sophia Spaghetti a warm smile.   Bong Bagoong  then  remarked, â€œHey, arrange your meatballs and grated cheese now.  Your guests are here.â€    Quick little footsteps rushed to the table.  Sophia Spaghetti recognized that one of them was the cute girl who grated cheese for her.  The girl asked her two playmates, â€œHey, why donâ€™t you both have spaghetti?  It tastes good.â€  One of the playmates answered, â€œWeâ€™re not used to that kind of food.  Is it like pancit?â€  The other made a face, â€œYuck, is that blood all over?â€  The girl urged her playmates, â€œCâ€™mon, just try it.â€  The playmates grabbed forks and scooped up Sophia Spaghetti.  They fell silent for awhile and then exclaimed, â€œYouâ€™re right, itâ€™s delicious â€“ itâ€™s sour-sweet!â€   The three children called on their other playmates and relatives to taste the spaghetti.  Everybody loved her.  Sophia Spaghetti was so happy.  She did not even notice that only a few noodle strands were left.  She twisted in a wide grin at the honorable Judge  Listo Lechon  and he could only wink in approval.  '),(13335,3556477,'10,July,2004','2004-07-10 00:00:00','   A voice saying something about a P12,000 salary...     July      To dream of July foretells unusual gains in all undertakings. Use caution in your affairs.    Money     Commodity. Profit or loss. Security. What do you value in yourself? '),(13336,3556477,'09,July,2004','2004-07-09 00:00:00','   I am eating a mocha cake heartily.     July       To dream of July foretells unusual gains in all undertakings. Use caution in your affairs.    Brown      This is an auspicious color to dream about:signifies freedom, success, money and happy and long-lasting union.    White      People feel they can rely on you. You have an abundance of energy and vitality.    Cake      Celebration. Sometimes treat, sometimes indulgence.    Eating      Satisfaction. Pleasure. Happiness is assured. '),(13337,3556477,'09,July,2004','2004-07-09 00:00:00','   My buddy and I are in a car one summer night driving up a road for a vacation.  I control the steering wheel while she is seated beside me.  We talk a lot during the trip.  At the highest point of the pavement, I stopped the car beside a tall gate.  I feel vulnerable.     April      It signifies that much pleasure and profit are in the horizon,but if the weather is miserable,it is a sign of distress.    Summer     Summer foretells joyous events and prosperity; denotes the maturity of ambitions and signals unbound satisfaction.    Night      Darkness. Mystery. Unconscious contents. There is a mystery that you want to penetrate.    Car      Personal power. Ego.    Driving    Work on energy and power. Looking for the desired destination.    Moving      Change. New start. Ready to leave something behind.    Friend      Joy and consolation. Aspect of self ready for integration.    Vacation      Rest or relaxation. Freedom. Looking for a change or maybe some adventures.    Travel      Journey. Freedom. Change. Change in attitude.    Talking      Conversation. Communication. Seeking for a way to express something.    Road      Achievement of something. Direction. Ask yourself about your life\'s path.    Climbing    Aspiration. Growth with effort. Achievement. Unexpected difficulties.    Brakes      Feeling secure with power. Control of slowing movement.    Gate      Opening or closure of something.    Vulnerable      Unprotected. Penalties. Need for a safe position or control over something. '),(13338,3556477,'09,July,2004','2004-07-09 00:00:00','   One fine summer day, the White Light joins the Dark Glow in a log cabin for a vacation.  They are in the dining room.  With feet on the wooden bench, the Dark Glow seats on the wooden table and watches the television placed on the high shelf beside the kitchen sink.  The White Light comes to Dark Glow.  They start to chat and tell sweet nothings to each other.  They laugh.  The White Light kisses the Dark Glowâ€™s lips but the latter hesitates and just turns back to the television.     April      It signifies that much pleasure and profit are in the horizon,but if the weather is miserable,it is a sign of distress.    Summer      Summer foretells joyous events and prosperity; denotes the maturity of ambitions and signals unbound satisfaction.   Vacation       Rest or relaxation. Freedom. Looking for a change or maybe some adventures.    Cabin       Domestic troubles. Love affairs.    Kitchen       Nourishment. Productivity. New developments in life.    Table       Place of activity. Workplace. Ready to start with your plans.    Chair     Position, style or attitude. Being conformable with something or someone.    Television       Movement. Images. Virtual reality. There is something that you wish to observe.    Old       Long time. Mature. Sensible. Former times. Something could be completed or ready to be replace.     Friend       Joy and consolation. Aspect of self ready for integration.    Lover       Vanity. Pleasure. The idealized inner self.    Talking       Conversation. Communication. Seeking for a way to express something.    Laughing       Relaxation. Volatility. Need for spontaneity.    Flirtation     Affection of love. Intimacy. Suitable time to pursue a courtship.   Kissing     Affection. Joy. Coming success. Wish to be close to someone. '),(13339,3556477,'03,July,2004','2004-07-03 00:00:00','   The Blue is a rapist and a murderer by profession, who lusts after a 10-year old girl on a fine morning in a lush and green forest.  She wears a long black hair and a long white gown.  She struggles to free herself from the brutal Blue.  She starts to run when she had a chance.  But the Blue continues to pursue the helpless victim.  She reaches a large, dark house.  The Blue corners her.  Her mother comes to her rescue.  They all sense something and look at the terrace on the upper floor.  There is her father.  She realizes she is a ghost.        April      It signifies that much pleasure and profit are in the horizon,but if the weather is miserable,it is a sign of distress.    Forest    Natural forces. Social activities of a happy nature. Ready to explore.    Green      Growth and serenity. There are projects which you are enthusiastic about. Great pleasures from simple things.    Homosexual      Union or fear of union with aspects of self. Passion.    Child or children      Innocence. The new self seeking to develop. Part of self nature is childlike.    Girl      Learning to be receptive. Happiness and tranquility.    White      People feel they can rely on you. You have an abundance of energy and vitality.    Lust    Eagerness for possession. Unfulfilled in life and need to be satisfy.    Rape Crime.      Profanation. Fear of being forced to unite with something.    Murder    Violent end.Loss of control over a situation.Sorrow and failure.    Fight    Violent resolution. Release of energy. Building or releasing conflicts.    AFRAID    Yourself; Disappointment is around you.    Fear      Unexpressed love. Self-doubts. Courage.    Running      No restraint. Freedom. Distance. There is something that you want to escape from.    Home    Center of being. Spiritual self. Shelter. Basic need fulfilled. Happiness within the family.    Mother    Approval or disapproval. From what do you need to feel nourish and protected?    Father      Authority. Control. Guidance. Recognition. Happiness. Virtue.    Ghost    Spiritual aspect of self. Memory. Past coming back. Beware of enemies.   Mystery     Unexplained. Secret. You are curious about something going on around you. '),(13340,3556477,'03,July,2004','2004-07-03 00:00:00','   Childhood home...  Relatives from the past...  Present home...  Loved ones of today...     January   To dream of this month,signifies financial gains,a mystery will be solved and beware of a false friend nearby.      HOUSE      Consistent metaphor for the self. Upper floors and rooftops symbolize higher functions, such as consciousness, thinking, and spiritual perspective. Basements and cellars represent subconscious feelings and awarenesses. Large houses with many rooms may be metaphors for extended family lines.   Discovering new rooms in a house-typically a pleasant dream-reflects the rediscovery of \"lost\" aspects of the self. This dream is common in women who have sacrificed personal hobbies and passions (painting, music, desires to own a small business) for the responsibilities of parenting. It\'s time to reconnect with your passion!  You drive a car, but typically you LIVE in a house/apartment. Dreams about a HOUSE symbolize a larger aspect of your Self, and the aspects of Self which make the whole. Each room is said to symbolize a different aspect of your Self; for example:    AN ATTIC  symbolizes your Higher Self, and your spiritual development & progress. Look at other symbols in the attic of your dream, and try to evaluate what they mean. Also pay attention to the feeling you experience in your dream...is it pensive, enlightening or what, exactly? All these things are clues for you.    A BATHROOM  would symbolize the need for cleansing/purging/elimination of something in your life that isn\'t quite working, or that has served it\'s purpose and now it\'s time to move on.    A KITCHEN  would symbolize the need or act of supplying nourishment or food for the body/mind/soul...whatever is currently \"cooking\" or developing in your life. If the food is plentiful, you have what you need. If the cupboard is bare, time to go shopping for new nourishment, and you need to figure out what is needed for that \"shopping list.\"    A DINING ROOM  is similar to the kitchen, but has more to do with immediate needs for supplying & utilizing nourishment, and less with the preparation or taking stock of those needs.    THE MAIN ROOM or LIVING ROOM  symbolizes your daily interactions with others, and often you will have other people appearing in your dreams in this room. Remember, they represent aspects of YOUR Self, and not themselves. (See PEOPLE, above)    BEDROOMS  symbolize the unconscious mind aspect of your Self, rest, dreams, sometimes sexuality issues in your life.    THE UPSTAIRS  symbolizes your spiritual awareness aspect of Self, or the Higher Self that holds all the keys or knowledge to this life\'s role you are acting out, and always has your higher good looked after, no matter how it might seem otherwise.    THE DOWNSTAIRS/BASEMENT  symbolizes your subconscious mind aspect of Self, which deals with habits, old coping skills, automation, ego. That\'s usually the part of our Selves that makes us feel \"torn\" between knowing we should do one thing, but inexplicably ending up doing the opposite. (Don\'t you HATE that??) Old belief patterns & fears have to be corrected, if that is the case. Tackle & overcome it, and you will feel much more peaceful about your life.    THE GROUND FLOOR  of a house represents your daily agenda; what\'s currently going on in your life.    REVISITING OLD HOUSES FROM CHILDHOOD OR EARLIER TIMES:  This points to issues that probably are resurfacing in your current life, and need to be looked at, analyzed, and healed so you can move forward and not look back. If you find yourself repeating the same old tired mistakes, or dealing with the same old tired fears, chances are you will have this dream.    A HALLWAY  symbolizes that you have reached an area that is necessary to journey through in order to get to the other side, and it may be a narrow path that has to be traversed with care and awareness. If you have that \"closed in, claustrophobic feeling\" then you need to expand your awareness/open your mind to more possiblities for completing this phase of your journey.    A  PORCH  symbolizes perhaps a metaphor of being undecided, such as the saying, \"I\'m still on the porch on that decision.\"  It could be symbolic of being contemplative, or uncommitted or withdrawn.  It could be something entirely different such as being on the threshold of a new sense or an extension of Self.  It depends on the entire context of the dream.  Back or Front Porch? Screened-in or open?  Messy or neat?  Sparse or furnished?  All these details add to the overall symbolism.    House   Financial security. Happiness within the family. Honor and dignity. Being.    Home   Center of being. Spiritual self. Shelter. Basic need fulfilled. Happiness within the family.    Relatives   Relationship. There are parts of you that feel the need to be acknowledged.    Child or children   Innocence. The new self seeking to develop. Part of self nature is childlike. '),(13341,4078761,'31,July,2004','2004-07-31 00:00:00','   funny - I could stay out and do whatever, but the first thing I did when I got home was clean the kitchen.  '),(13342,4078761,'29,July,2004','2004-07-29 00:00:00','   Don\'t know why I decided to create a blog today.  I\'ve thought about it for a while.  Why not - everyone else is doing it.    God is faithful today.  even though I got caught in the rain three times.  and the pain level is at about a four.  with throbs to six.    faithful, faithful, faithful.  Liz leaves in the morning.  Molly\'s going back to South Africa.  School starts in three weeks.  not ready!!!  '),(13343,4078761,'10,August,2004','2004-08-10 00:00:00','   It\'s amazing how the days after a perfect vacation can be so miserable.  No reason really.  well, lots of things to think and stress about.  but wow, that was a great trip.  We got to Buxton Thursday night, after driving through a lot of standing water on Hwy 12.  Hurricane Alex had definitely left its mark.  We got to say hello to the beach while we waited for others to arrive.  We found out later that it would only be 11 of us, but in the end that seemed perfect.  It was a great bunch - Alexa and Pete, Chris and Tana, Mark T, Mark E, Philip, Nisrine, Sarah, Susan, and myself.    Friday, we started the day with a devotion and life story.  I went first and they got an ear full.  It was supposed to be 10 minutes and I think it was at least 45.  Lots of tears.  show and tell of the nieces and nephews.  I had just been diagnosed with fibromyalgia on Thursday afternoon, so that was my chance to let everyone know.  Then they all prayed for me.  Very powerful prayer time.  I was really thankful for that.  Then, we headed to the beach.  And what an experience that was.  First, Philip creamed me on a boogie board.  Major wipe-out.  thanks a lot.  He pinned me down, and I yelled at him, \"Get off of me!!\"  I was so mad, I turned around and dove back into the next wave.  That was the next major mistake.  Within a few seconds, I was hundreds of feet down the beach, unable to stand up.  The current was so strong after the hurricane, I couldn\'t believe it.  Between the pounding waves and that current, I really started to get scared.  I kept trying to get back to the shore but I couldn\'t make it.  Finally, I started crawling out. Nisrine got there just as I managed to get out on all fours.  She helped me to the shore where I gasped for breath for a while (and didn\'t even feel a crab crawling on me).  Then she and Alexa helped me back to where everyone else was.  I literally had about 3 cups worth of sand in my bathing suit.  Everyone was pretty worried about me.  not half as worried as I was about myself.   We gave up and went back to the motel for lunch.  Then the Marks went to check out another beach - that\'s the nice thing about Hatteras - you get to chose which side of the island has the best surf.  So it turned out that the south-facing beach was absolutely peaceful.  We headed over there and I finally got to float like I love.  Then we packed up the cars and headed for 10cent shrimp, only to find out that it was too late.  So we headed back and had our cookout instead.  Then we went for frozen custard at Uncle Eddy\'s, and Isaac offered us a free round of golf.  I got one real whole in one and two honorary ones.  My best round of putt-putt ever - a 42.  Not to bad considering my usual showing.    I started to get really tired that night.  My feet were really hurting.  Hard to stand after a while.  Had a major confrontation with Philip after that, but it was good in the end.  I was very glad to get all that off of my chest.  It\'s amazing how creepy it is when he tries to touch me.  I don\'t even want to shake his hand.  Saturday was the perfect day.  Perfect.  Started with Mark waltzing into our room while I was still in my pjs.  He started cooking breakfast, and I just stayed on the couch until I could make a beeline for the bathroom.  He messed up the pancakes and smoked up the room while trying to cook bacon, but it was pretty cute.  Darn it, really cute.    Then we moved to our new room, and two more folks shared life stories.  Mark shared his, and I just wanted to hold him afterward.  It was so sad.  I had a vision about God turning the charred remnants of his heart into diamond incrusted gold.   I hope it\'s true.    Then the news came back that the surf was perfect, including for me wanting to float.  So we went out, and sure enough, I got to float for at least 3 hours.  Very very happy.    I had a hard time accepting that it was time to come in.  that was when God gave me the refrain: it is enough.  It is more than enough.  Thanks be to God.  I feel like I have to write a Psalm about that, but I\'m not sure what it needs to say yet.  then we went for 10 cent shrimp again, and that was wonderful.  sat next to Mark.  happy.    then, the most amazing - kayaking in the sound at sunset.  wow.  incredible.  so peaceful, so beautiful.  I wish I could do that every day.  okay, so I\'d settle for once a month.  silence.  easy to be with people.    Mark talked to Sarah a lot, but they talked about ethics and I don\'t have to know everything about ethics.  she does, because she\'s getting a PhD in it.  she\'s jealous of me for my overseas work.  I don\'t have to be jealous of her.  will she and mark get together?  After kayaking, we headed to dinner.  the Outrigger?  FABULOUS soft shell crabs.  note to self: if that\'s available, get it every time!  sat next to Mark again.  felt like we connected on a few things.  including the fact that we ordered exactly the same meal.  he has dreamy eyes.  did I mention that?  sad hazel eyes, but the laugh lines aren\'t tanned, so they stand out more now.  sad eyes with big bright laugh lines.  laughter through tears, my favorite emotion.  then the bonfire.  singing great songs.  I got to sing.  then I went and watched the moon over the water while the others sang.  feeling solitary but not lonely.  beautiful to watch the flicker of the bonfire on the sea foam as the waves came in.  nothing like it.  saw a few shooting stars.   mark and I walked up to the motel together.  I was hoping to talk to him, but his car was there and he had to drive it back.  have a good night.  so I went to my room and stared at the stars a while.  then I went in and saw his travel mug there.  had to bring it back to him, right?  but when i went into his hotel room, he was changing.  oops.  thought you might need this.  oh, yeah, thanks.  good night.  but left feeling happy.  I tried.  I was strong.   decided if I woke up by 7:30 I\'d go to church. I woke up at 7:25.  went to church at Little Grove.  I think I surprised him.  sweet folks there.  told him I came because it reminded me of Concord Primitive Baptist.  that\'s partially the reason.  partially because I really just wanted to be near him.    back to the motel, packed up, grabbed breakfast, realized that Alexa has had a hard life and is not perfect, went to Hatteras for church #2.  good to be there.  God gave me the verse Hebrews 11:11 - she considered him faithful who had promised.   I can do that.  I know God is faithful.  I know that without a shadow of a doubt.  why do I feel like I have so little faith otherwise?  mark and sarah?  a little weird as we left.  sarah rode back with me.  lunch, Mark E and his haircut.  Philip\'s life story.  I hope people understood a little more about him then.  everyone else left.  finished up Mark E\'s haircut while MT and I talked.  thought about telling him.  but didn\'t quite fit.  then Nisrine joined us.  he asked about the diagnosis.  was very kind about it.  encouraged me to get into a support group.  probably will.  need a ride to the airport?  he looked at me like I was insane.  well...  bye sweetie, thanks for everything.  at least I did get a hug out of it.    drive home with Mark E and Nisrine.  I\'m glad he was taking care of her.  I slept some.  hard to be home.   I would have rather been at the beach.  for another year or two.  '),(13344,4078761,'02,August,2004','2004-08-02 00:00:00','   Must get my ordination paperwork done by tomorrow afternoon.  Which is worse, paperwork or cleaning mouse droppings out of the pantry?  between the two, I just want to stay in bed.    How much of my procrastination is related to FM?  Not being able to think straight, or too deeply?  or is that more fear or laziness?  It hurts to have to think that hard.  I have a tremor in my left arm today.  Spasm in my ear last night.  Pain right now about a 3, but I\'m back on Tylenol.  It does make it more bearable.   God, I know you\'re here.  I know you\'re using these trials to shape me.  But I feel pretty darn well shaped already.  Can\'t we lighten up for a little while?  Whatever happened to restoring the years the locusts have eaten?  '),(13345,3907553,'14,July,2004','2004-07-14 00:00:00','   Beware of drinkers! For if you drink to much this Saturday then your merriment wil turn to woe on Sunday.  '),(13346,3907553,'13,July,2004','2004-07-13 00:00:00','   Why must people here in portsmouth refer to all and sunder as mush. This is starting to get right on my tits now. Rant over. In other news seeing as it is my birthday soon I have decied that kira Knightly would be a nice present. *cough*Chris*cough*  '),(13347,3907553,'13,July,2004','2004-07-13 00:00:00','   Even without Lifehouse Jason Wade writes great songs. Like this one from the shrek soundtrack.  You Belong To Me   See the pyramids along the Nile  Watch the sunrise from the tropic isle  Just remember darling all the while  You belong to me   See the market place in old Algiers  Send me photographs and souvenirs  Just remember when a dream appears  You belong to me   And I\'ll be so alone without you  Maybe you\'ll be lonesome, too   Fly the ocean in a silver plane  See the jungle when it\'s wet with rain  Just remember \'til you\'re home again  You belong to me   And I\'ll be so alone without you  Maybe you\'ll be lonesome, too   Fly the ocean in a silver plane  See the jungle when it\'s wet with rain  Just remember \'til you\'re home again  You belong to me    Its great...trust me. But not as Good as Everything by Lifehouse now thats a love song, maybe one day all good men will find someone whos their \'Everything\'. Happy listening.  '),(13348,3907553,'12,July,2004','2004-07-12 00:00:00','   I\'m bored so if you have an amusing rant/story then stick your comments here please!  '),(13349,3907553,'12,July,2004','2004-07-12 00:00:00','   Well I\'ve been playing with the Blog code today and added the stuff to the right: books, films etc. If you\'ve go a suggestion for things to go on here then just mail me.  '),(13350,3907553,'12,July,2004','2004-07-12 00:00:00','   Bilious says that finding someone to be with is not unlike archery...except its uphill, thers a strong cross wind and you drank 13 pints of gut ripper the night before. If you aim to low you\'ll never be happy with what you hit. But if you aim to high you\'ll never even hit the target. So get drunk be merry and I\'ll see you with a hangover in the morning  '),(13351,3907553,'12,July,2004','2004-07-12 00:00:00','   Well here it is at last, I\'ve given up with social chameleon the website as dreamweaver was pissing me off! So for the moment I\'ve gone with Social Chameleon the blog. Keep an eye out for pics, reviews, recipes and rants on this page. Have fun  '),(13352,3907553,'24,July,2004','2004-07-24 00:00:00','   My legs hurt...\"why?\" I here some of you cry. Wellactually none of you probably give a stuff but i\'mgonna tell you anyway. Basically its because ofthis...I have been on a course for 3 weeks and itended yesterday at about 3pm \\o/ so i went for a quietsocial drink with my fellow course mates :o) Aftersome lunch and a few pints me and some of others movedonto jugs of cocktails. It was about this point youngVictoria Salter arrived to join me in my Soiree and ibegan on the vodkas. Now I\'m not exactly how many Ihad but after an hour or so went got the tube totemple as some of the peeps off my course where goingto a pub on a boat. I think it was called the queenmary, however due to the queue we couldn\'t be arsed sowent to motion for another vodka before going toocamden and munching through a kfc. Basically the restof the evening goes: meet tom in pub -> 1. aftershock-> 2. vodka -> repeat actions 1. and 2. several times-> go to underworld -> do point one again and repeatseveral times -> lose tom as he runs off with somebird that isn\'t his girlfriend -> leave with vic at2.30ish -> walk to trafalgar square (now anyone goodat geography knows there not exactly close) -> try andnot let vic stumble into road -> have mcdonalds to tryand sober up -> get night bus -> get home -> sleep. Now can anyone guess why my legs hurt? Well I ain\'tgonna spell it out :op it was a fun night tho and thewalk was quite nice actually. I went to  urlLink www.rupissed.com  and it said this:&nbsp;281.87g alcohol ingested...201.37g still left in me and the comment was \"You should be dead by now\". In the words of Samuel Beckett \"Oh boy\".  '),(13353,3907553,'21,July,2004','2004-07-21 00:00:00','   Yay another fun weekend was had in Portsmouth, 2 nights of drunkeness followed by the feeling that I should sleep for a week. Does anyone know how to put pics on blogs?  '),(13354,3907553,'31,July,2004','2004-07-31 00:00:00','   a kitten and a place of my own! No make that two kittens! Disco    and Tilly  '),(13355,3907553,'30,July,2004','2004-07-30 00:00:00','   I have two new blogs though one may get deleted soon for numerous reasons. However, for those amoungst you that know what I\'m on about here\'s my  urlLink Adrenalin Monkey Blog  Anyway enjoy your saturday!  '),(13356,3907553,'30,July,2004','2004-07-30 00:00:00','   Well this friday eveningwasn\'t overly exciting, but it was quite leisurely chilled. 2 Litres of orange reef and a film with my housemates (Grosse Pointe Blank). Could\'ve been worse :o) Well enjoy your weekends people!  '),(13357,3907553,'28,July,2004','2004-07-28 00:00:00','   You know women are a strange breed...in fact I\'m pretty sure they\'re not actually the same species as men. But anyway enough of that, I was asked the question yesterday in an email, \"do you fancy any one?\" and the answer was no and honestly so. But today however the answer maybe somewhat different, maybe indeed. Now I\'m pretty sure no one (except maybe John as he\'s a moderator) reads this blog so I\'m quite safe in posting the fact that I may fancy someone from work. However, I will only say who if you email me :op  Have fun  '),(13358,3907553,'26,July,2004','2004-07-26 00:00:00','   It seems official the only night with any good TV on is Monday\'s and then not till 9pm and only on BBC 2. 9pm Coupling, 9.30pm Dead Ringers, 10pm The Keith Barrett Show. The rest of the week is utter toss! Well it is if you only have 4 channels! Anyway rant over, watch these programs!  '),(13359,3907553,'05,August,2004','2004-08-05 00:00:00','   urlLink George Bush\'s Blog courtesy of the Onion.  Look theres even the nuclear launch codes!  '),(13360,3907553,'04,August,2004','2004-08-04 00:00:00','   I hate English summers! When it finally gets warm its all icky and muggy. why can\'t we have a nice 35 degree dry heat!?!?!  '),(13361,3907553,'03,August,2004','2004-08-03 00:00:00','   Yeah its another rant about women...*sigh* Today I felt really old (and I\'m only 24) I found out that all the women I slept with or went out with while I was at Uni are now in very serious relationships. Even the ones I had serious relationships with! Ok I do honestly wish them all the best, but here I am doing the career thing and some of them are married and more than a couple live with long term boyfriend now.  Kinda makes ya feel old and lonely at times, I\'m enjoying my life but sometimes things like that make me feel like I\'m being left behind. Ah feck it! It ain\'t gonna change in the next 5 mins so will try and stop thinking about it. Have fun y\'all  '),(13362,3907553,'01,August,2004','2004-08-01 00:00:00','   Well is it? I\'m half way through series 2 of 24 and so far between the two series the only really treacherous characters have been women. The presidentâ€™s wife, Nina Myers and Marie Warner are all involved in various grand conspiracies. They all betray people they supposedly loved, are involved in each one persons death and don\'t generally get what they may deserve. Now is this some deep-seated view that Sutherland has? Or is just that men (on the bad/evil side) in these things tend to be murdering, torturous bastards and the women are backstabbing nymphetâ€™s. Go figure.  '),(13363,3907553,'01,August,2004','2004-08-01 00:00:00','   Thats it really, we\'re now on blogwise. Click on the button in the right hand menu or go  urlLink here  '),(13364,3907553,'10,August,2004','2004-08-10 00:00:00','   Yay! I ordered 2 cd\'s from Alice records last Thursday night and ta daa! They turned on saturday. I got   urlLink Breaking Benjamin\'s  We Are Not Alone; its really fucking good! Though not as heavy as Saturate. I also got 3rd Strikes Lost Angel; which is also good. Have been listening to a lot of   urlLink Rob Dougans  Furious Angels lately too which is different and very good. Go! Go buy them all! Well at least Breaking Benjamins and Rob Dougans  '),(13365,3907553,'16,August,2004','2004-08-16 00:00:00','   Well I picked up my free ticket and trottered off to cineworld today in anticipation of the latest M Night Shyamalan film called  urlLink    The Village     and well, hmmmm. It was O...K...it wasn\'t bad...but it wasn\'t great either. The idea was a good one but the tense moments seemed to run a bit dry sometimes. While some parts that could have been made tense and full of suspense just weren\'t. Another problem was that once you\'d been sat there for 20 minutes you could kinda guess the entire film...that never helps. Maybe I expected to much, maybe I\'m just being fussy seeing as I go to blockbuster films expecting entertaining tosh and go to the more original films to see something new. However, constantly I find myself disappointed by American cinema and I turn to Asian cinema to remind me that films can be more than just seat fillers for rich people.  On the whole though the village is a good film. The idea is good, and to a large degree plausible. The directing was good (as always) and some of the characters were amazing (Although seeing Sigourney Weaver in it was a surprise!). However, I felt there was something lacking...unfortunately seeing Joaquin Phoenix made me just wanna shout \"swing away Merill! Swing away!\" as well. All in all The Village gets a very entertaining 6/10.  '),(13366,3907553,'15,August,2004','2004-08-15 00:00:00','     '),(13367,3907553,'15,August,2004','2004-08-15 00:00:00','   Well it does if this is posted Joker  ===== \"Could it possibly get any uglier?  I used to be a highly respected Watcher, and now I\'m a wounded dwarf with the mystical strength of a doily.\"   		 __________________________________ Do you Yahoo!? Yahoo! Mail - Helps protect you from nasty viruses. http://promotions.yahoo.com/new_mail  '),(13368,3907553,'15,August,2004','2004-08-15 00:00:00','   Go on try it -  urlLink Slime Volleyball  '),(13369,3907553,'15,August,2004','2004-08-15 00:00:00','   urlLink   What Is Your Best Sexual Skill?         Name:         Age:         Sex:     Male  Female       Sexuality:     Straight  Gay  Bisexual       Flirting Skill Level -  20%                                                                            Kissing Skill Level -  32%                                                                            Cudding Skill Level -  75%                                                                            Sex Skill Level -  79%                                                                             Why They Love You      You can do amazing things with your tongue.        Why They Hate You      They can\'t bend the way you want them to.              This   quiz   by  urlLink  lady_wintermoon   - Taken 540393 Times.          New! Get Free  urlLink Daily Horoscopes  from Kwiz.Biz'),(13370,3907553,'15,August,2004','2004-08-15 00:00:00','   Am bored and doing uni work. So while attempting to not get distracted (AND FAILING!!!) I realised i hadn\'t posted in a while so here it is! A brand spanking new post...oh if any of you know why my legs bruised after friday nights drinking then please reply to this post. Ta  '),(13371,3907553,'17,August,2004','2004-08-17 00:00:00','      '),(13372,3907553,'12,August,2004','2004-08-12 00:00:00','   So.  Not even arrived in Oz yet and already having nightmares about spiders and even bigger Brit eating spiders in Oz.  Think I might have to get Hypnotized or something before I go as I dont think I can take it lol.'),(13373,3907553,'04,August,2004','2004-08-04 00:00:00','   This blog has been made for two of my friends, Vic and Nat. They\'re off travelling to Australia from the beginning of 2005. So what I\'ve done is make them this blog so they can post their travelling rants and stories. This way we can all read them without the need to send millions of emails :oD   '),(13374,3669958,'27,June,2004','2004-06-27 00:00:00','    urlLink    For those that have read further down into my blog, you\'ll understand why this is so ironic! Thanks Google for making me sadder about not going to SCI!!   Posted by  urlLink Hello '),(13375,3669958,'23,June,2004','2004-06-23 00:00:00','    Hello.. I\'ve been quite depressed lately, and I have no idea really why. I guess that\'s the downfall of being a woman. Sometimes you get really depressed or really happy, and there\'s no explanation as to why you\'re so up or down. Well, boys and girls, I hate it! So here\'s what\'s been going on in my life: Friday morning my mom dropped my kitten Stitch, and Cat Sibby off to the vet, and later on that day I called in to check on them, and my favorite nurse is joking about how badly she wants my kitten. Well, At first I\'m really excited because no matter how much I love my kitty, we\'ve just got way to many animals. GC\'s friend was moving out of state and left us Stitch, and we--especially me--fell deeply in love with Stitch. But GC has been complaining that 3 cats and a dog is too much work, and even though I cant even think of parting with any of them, He is right. I have to make a compromise though. We cant give away Sprite, nor Harrah. Harrah has only been with us 7 months, but this baby is my angel. I love her and Sprite with my whole heart and soul. I feel such a connection with them both. I think GC is going to keep Sibby, his baby.  So anyways, Here comes the most perfect opportunity for Stitch. CJ is the director of the vet office, and she is truly the only person I\'d trust whole-heartedly giving Stitch to. She has showed so much love to me and my babies for years, and I know that if she takes stitch, it will be the best thing for her and him. Me and GC have talked, and we\'ve decided we\'ll give Stitch to her. Now, I just have to keep my broken heart from spilling into too many tears when I say goodbye or I\'ll keep him, possibly make things harder on GC, and take away Stitch\'s one REALLY great chance for a truly, good home. My poor Sibby, Sprite and Harrah the most, are gonna miss him terribly. But sometimes you have to do what you have to do.. Or atleast that\'s what I\'m telling myself to try to keep my emotions from taking over.. The worst thing, last night he curled up on my neck while I was asleep, he put his little head on my face and slept with me a long time. He made me feel so loved, and I guess he made me feel like the right thing is to give CJ the little kitty she now loves as much as we do.  Now I have to go before I start to cry. Too late.. '),(13376,3669958,'23,June,2004','2004-06-23 00:00:00','    urlLink    I was bored, so I decided--for your entertainment only--that I\'d draw how my tummy feels! Bet my poor GC feels the same! ) =  Posted by  urlLink Hello '),(13377,3669958,'21,June,2004','2004-06-21 00:00:00','    Hey guys, I\'m gonna get right to the point this time... I\'m never eating ARBY\'S AGAIN!  Two nights ago, GC and I shared a Turkey, Bacon, & Ranch Marketfresh sandwich from there and last night we both got sick. Here\'s what happened... Here\'s the Dilly-O! LOL! We just ordered food last night--Of course GC ordered Chinese like always, and I ordered from a local pizza place--we had just sat down to eat, and taken a few bites each and the next thing I know GC is in the restroom. We had both complained of our tummies growling a few hours before, but we both didn\'t really feel like eating. So, of course, I ask him, \"Are you okay, baby?\" He says he\'s not, and that he\'s sick to his stomach. RIGHT after I asked him that, Suddenly I got sick! Lemme just say it wasn\'t puke we were feeling... *hint hint!* LOL!!! **Embarrased as hell** So we only have one bathroom, and I start screaming at him saying that I have to go NOW! It was the worst feeling ever!! So poor GC comes to the rescue, and carries me to the restroom, and I\'m in there no more than a minute and he\'s saying he has to go again! It was horrible! As soon as I felt a little better, I went right to bed. I remember having the same annoying dream over and over and I didn\'t sleep well. I wake up at 5 a.m., and again I\'m sick. It started for us again. Poor GC had to go to work today, since he started a new job, and as he was walking out the door we agreed that both our tummies were burning bad! I never feel like eating again!  Now.. Let us proceed to the next issue of the morning! **Clears throat** So, GC is getting stuff ready for me as he\'s about to leave for work. He goes to the fridge and he yells, \"Shit!\" He told me the fridge had popped open in the night and all our food got warm!:( Our take out which we only took a few bites out of was gone, and everything else, eggs, milk, you name it. I was so angry! Why you ask? \"It\'s just a fridge! It\'s just food!\" You might say. I\'m angry because it seems like everything is broken or not working properly in this house! The stupid fridge will close but within a few second--every now and again--the damn thing just pops open all of a sudden! Stupid fridge... On to he next hot issue... **clears throat and bangs on chest** Ouch.. Sheesh! See, when GC is gone I\'m usually at the computer since I\'m waiting on a new wheelchair. So, I\'m pretty much left in one spot. Let\'s hope for no fires. LOL! Okay.. That really isn\'t a joke.. kinda scary actually... but ANYWAYS!!!!!!!! So My computer monitor just stopped working today. *POOF!* likes its unplugged from the computer. So he twiddles with it, fixes it, and right as he was walking out the door it does it again. I\'m panicking of course, because if it goes out in the middle of the day, I\'M SCREWED People!! But luckily it hasn\'t gone out yet... **knocks on wood**  Well.. Okay, So, um, Yah, I\'m done ranting on my blog for now. Thank you for sticking with me through all my silly posts. Place a comment, I have so many stories to tell LOL, lemme know what kind you wanna hear! Once I think of something to write, I type FAST! So Help me get started!  Well, Since my first hour of the day was horrible--Yes it all happened within an hour of me waking up--I think I\'m gonna try really hard to have a good day while my tummy goes on hurting.  Hope you all have a good day too. '),(13378,3669958,'21,June,2004','2004-06-21 00:00:00','    This weekend was one of the laziest ones I\'ve had in a long time. I hate when this happens because usually when I have nothing better to do I end up doing the following... 1. Breaking a rib 2. Hurting bad from arthritis because nothing to keep my mind occupied 3. Rebreak something else that just healed..  So you see, I tried really hard this weekend to be REALLY careful as I did not what to rebreak my collarbone that has just finally stopped popping and moving around. I have been REALLY careful I think. I hope atleast, the next few days will be the real test! GC started a new job today, he seems to really like it. Its good we need the extra money.. Shoot EVERYONE needs the extra money! Haha! I\'ve been really excited lately about getting back to school. I was going to go to a world renowned cooking school in Scottsdale, Arizona. I was even accepted, actually but I think I was only really kidding myself. Yes, cooking is my passion. Mainly Pastries and Baking, but its so much easier to do that at home than to do that full time in a commercial style kitchen. One thing, commercial style kitchens for those that have never seen one are TALL. Everything is standard size, meaning 36\" or higher. Its almost impossible to reach anything in those type of kitchens. We had this master plan all sorted out, but I realized I was only kidding myself. I\'ve decided to turn down the chance to go to the Scottsdale Culinary Institute and instead I\'ll go with my 2nd passion. Graphic Design. I\'ve been doing amateur design myself at home with Photoshop and the like, but when I started looking into the schools I got really excited. First of all I\'d be getting a real degree, B.A. instead of a certificate at SCI. Second, Graphic design also lets me be creative which I am all about. I am a real crafts/artsy girl. I like doing all the things housewives like to do! LOL. I love to sew, cross-stitch, knit, paint, draw, etc. I\'d get to take lots of drawing and painting classes, and of course all general education which I\'m really excited about. I\'ve been out of school for 3 years now, and I\'ve been really depressed about that. I really enjoy learning and seeing all my old friends going to school and getting degrees already made me feel pretty worthless. I had a pretty bad experience in high school and its REALLY made me scared to go to college even though GC tells me its so different. I wanna be like GC, I want a B.A. I wanna be smart like him! hehe! I\'m looking into a lot of schools right now, but there is one that I REALLY want to go to! The Art Institute. I love the offered classes, but you wouldn\'t believe the price to go!! Yikes! I just pray every time I think of school that my parents will help me pay for it since I\'d really be doing something I CAN do.  Well.. I\'m getting too excited, and now I\'m off to look at some more schools! I\'ll post my art on here a little later! Or just make an ART blog. Either way show you some of my art soon! Take care everyone! '),(13379,3669958,'19,June,2004','2004-06-19 00:00:00','    urlLink    Sibby is 3 1/2 years old, Sprite is 4 years old.   Posted by  urlLink Hello '),(13380,3669958,'19,June,2004','2004-06-19 00:00:00','    urlLink    Harrah (left) and Stitch both 7 1/2 months old, are best buddies!   Posted by  urlLink Hello '),(13381,3669958,'19,June,2004','2004-06-19 00:00:00','    Hello again!  Seems like I sometimes dread Friday nights. There is nothing to do here EVER! It sucks. I hate living in the Midwest. There are no big cities to party in, no beaches to hang out at. All the clubs feature country music, and trust me the list goes on and on. Its kinda cute when my mom and dad tell me of all the stuff there is to do here in BORINGVILLE but I guess me and GC have different preferences when it comes to having fun! I wanna party!! I wanna go out, listen to loud music, drink, and have fun! Now I should say, I am only a social drinker, and smoker. I know, I know, Smoking is bad, I\'ve heard it time and time again from GC and seriously after this pack I\'m done. I\'ve kept this pack now for over a month, so you can see I\'m clearly not addicted. I just like to smoke when I\'m having fun!? Nothing wrong with that! Anyways.. So here I am sitting on my couch cross-stitching a present for my mom, looking like some old lady doing her hobby, and the doorbell rings! Thank god, LD and MB came over to hang out. We just chilled, watched TV, then they went home about 30 mins ago to get to bed early before we all head out for a day at the park tomorrow with our dogs, while the boys play basketball.  I\'ll post a picture of my dog, Harrah. She is 7 1/2 months old, she and my 3 cats are the loves of my life, including GC of course! :) So I\'ve sat around all night, cross-stitching, playing with Harrah, watching TV, and posting on my blog.  What a great frickin\' Friday! *sigh* Oh well.. I guess I\'ll get back to finishing up my mom\'s gift. Sorry for such a boring post!  I\'m out! '),(13382,3669958,'17,June,2004','2004-06-17 00:00:00','    urlLink    Cast Kit--Some items are a little dirty since they haven\'t been used in years, thank god!  Posted by  urlLink Hello '),(13383,3669958,'16,June,2004','2004-06-16 00:00:00','    I\'d like to first start off telling you about my night last night. My best friend, LD, recently moved back to town, and I think I\'ve been the happiest with her return. She and her boyfriend came over last night and we had the best time. Her boyfriend, MB and my boyfriend GC were playing ESPN NBA 2k4. LD and I on the other hand were playing Skip-Bo, a game we used to play even as kids when she used to come over my house. It was a blast. I think I laughed so hard all night that I broke a rib! LOL! Fun times, fun times! Okay, so I\'ve been thinking for the last few days about what to write about. Its actually been pretty difficult to think about this since I have so many things I have to tell. I guess I\'ll tell you about my broken bones, and stuff like that! When I break a bone, the average person thinks I am the strongest person with such a high pain tolerance--Except my dad, I think he\'s the only person I know that has felt as much pain as me. He\'s broken almost everything, Huh Dad? :) I may have the yr wrong but the last time I remember going to the doctor for a cast, was when my dad picked me up from school in... 5th? Grade after I broke my arm at lunch time. After that my mommy became my doctor. We slowly started this collection of splints, braces, slings, collarbone braces, Ace bandages, etc. (I\'ll have GC take a picture of my current set for you guys.) My mom was the best \"Doctor\" I\'d ever had. She was so gentle, and if I told her it hurt she\'d stop. I\'d usually take a day or two in bed or relaxing at home before I was out and about again.  EXCEPT, when I broke a leg. That was painful! I remember having to lay in bed the entire time for it to heal. I\'d usually have to go on home studies from school to continue my education and not fall behind. One time I broke my femur, and of course I stayed in bed for a few weeks, and after a while the pain went away, but I could still feel this funny feeling in my leg, like it was moving in a way it shouldn\'t be moving. Well regardless, I felt NO pain whatsoever and I went back to school (My leg was still splinted up because of the movement.) and I even went to stay the night with friends, and my friend used to pick me up and sometimes help me rewrap my leg and everything. Well after about a month of this my mom decided it was time to get it X-Ray\'d. Well, imagine the surprise to find out my leg WAS still broken and the bone was separated about an inch apart!! My friend who came with me almost started to cry, and my poor mom felt so bad. I couldn\'t help but laugh, I just couldn\'t believe that I had gotten so used to the pain I didn\'t even feel it anymore! Sometimes I think my pain is really bad. But when others see how I\'m dealing they tell me I\'m so strong. I guess I have a pain tolerance that I really don\'t see.  Well, I\'m gonna end this now because GC, is setting up the picture of my \"Do it yourself Cast Kit\" Lol. I\'ll post it as soon as he takes it! '),(13384,3669958,'16,June,2004','2004-06-16 00:00:00','    Hello again! I woke up not too long ago.. Well maybe an hour or two.. Yah I stay up late, cant seem to get my sleeping schedule right when I\'ve got a broken bone! (Currently have a broken collarbone!)   Okay so anyways, I don\'t know what it could possibly be, but I always have the most vivid, and sometimes very scary dreams. They are so strange. They jump around so often at times, that I cant even tell someone what it was about because the story changes 10 times. I have always wondered why some people can have such types of dreams, yet others cant remember if they\'ve had dreams or not? Is it something we eat? Is it because we don\'t go to bed with a full stomach? Are we just more gifted into seeing these dreams for what they are? I honestly have no idea.  My most scary dreams are the recurring dreams. And its not just one, but several that I get every now and then. Like me and my sister Ashley out in the country walking on a long dirt road. Then all of a sudden there are little tornados everywhere. We have no where to go. Its so scary! They move around like those little ones in the movie twister! I guess I\'ve seen and been around enough tornados in my life that it just kind of stays with you if you are scared enough of them.   Another really scary one is sort of of like the movies  28 Days Later , and  Gremlins  (I think). We\'re at this hospital, me and about 10 other people. And for some reason the majority of it is empty, and trashed. We cant figure out why? We\'re on like the 30th floor of the building. And someone screams and tells us to hide. So we all do, and there is this huge beast, like a cross between a werewolf and gremlin trying frantically to find us. He has these HUGE claws, and huge teeth covered in green slime. Its so realistic looking though. He has incredible strength, and we keep moving from place to place trying to hide from him.   But the weird thing that I cant quite comprehend at all is the way I can change my dreams any way I wish. This only happens if it is a recurring dream. I have to have the dreams several times to know the outcome. But once I have it like the 4th or 5th time, I can change it any way I please! Its great. I don\'t know if anyone understands what that\'s like, but to have power over your dreams... Its unbelievable!   And the last thing I\'d like to point out is this. In 98% of my dreams, I am walking. It feels so realistic. Sometimes after a really vivid dream, I can wake up and for a split second I almost feel like I can get up out of bed. Its so odd when I have a dream when I\'m in my wheelchair, and it usually only lasts a small part of the dream before I\'m up and walking again. I wonder if walking in my dreams fulfills some of my desire to walk, because right now in my life, if someone could wave a magic wand, and I would be cured, I would NOT want to walk. Yes, I\'d like my OI cured, and I\'d sure like to have my arthritis pain go away, but walking.. Nah! I\'ve only ever known how to be in a wheelchair, and only recently have I\'ve been able to accept my disability 100%!  Well, damn, its lightening and thundering bad outside! I better go before it a tornado comes and sweeps me away to Oz! (or before the lightening takes out my power and this is lost!) '),(13385,3669958,'15,June,2004','2004-06-15 00:00:00','    Hello everyone. Thank you very much for taking the time to read my blog. I hope you find it interesting, and I hope you will keep coming back and checking up on me. I guess let me start off by telling you all a little about me. My main reason for this blog is  to share things with my friends and family that I normally wouldn\'t tell them. Also, I want others to know what its like to live with OI, and I also hope to meet others with OI. I feel bad telling them about my pain, sorrow, depression, or even happiness but I have a hard time telling people how I feel when it has to come from my mouth. I have a much easier time telling them when I can write it down. I can put more truth, and emotion in a letter... if that makes sense. I am a 21 year old, living in the USA. My name is Amanda. I was born with a rare genetic disease called Osteogenesis Imperfecta. When I was born back in 1982, this was actually a very rare disease. I cried constantly from the moment my mother gave birth to me. My mom was worried when I cried instead of eating, and soon she told the doctors that she felt something was wrong with me. She was right, and upon examination, and several x-rays later they found out I had broken both my legs.     Of course the accusations went around, and they tried to say my mom had abused me, and when that didn\'t hold, they tried to blame the nurses that were in charge of my care. I am not quite sure how the test came about, but they ended up doing a test on me, and it proved I had Osteogenesis Imperfecta, a brittle bone disease. They instructed my mom and dad to put me into an institution saying that I would never be able to live the life of a \"normal\" child. I am never quite sure if I\'ve taken the time to say thank you to my mom for not sending me there... I\'ll take that opportunity now.  Mom, Thank you so very much for knowing that you could raise me to be \"Normal\". You did the best job anyone could have done, better than some damn institution. You and Dad (my step-dad) are the best parents! I love you both so much!      Growing up was hard, I really don\'t have many memories being out of a cast. I was always breaking something. I know how hard that must have been on my mom, always having to see me in pain. It used to break my heart when I was hurt because I could see in her eyes that her heart was hurting as badly as my bones were, if not worse. As I got older, the breaks started to lessen, and at one time I went 3 years without a broken bone--not incuding ribs, those break all the time. I was in a car accident when I was 13, and I broke 13 bones, that was the worst 4 months of my life. I laid in a hospital bed the entire summer.      Things have gotten better over the years, and I\'ve done a lot of research on the  urlLink Official Osteogenesis Imperfecta website . I found out that even though I have suffered through so many breaks, and so much pain in my life, I actually have Type 4 OI, which is the least common type. It means that I do have the short-stature, and the bent or bowed bones, and of course many breaks as a child, but as I get older, the breaks become less frequent, and usually only happen to ribs, collarbones, wrists, and toes. Of course, you can imagine my happiness to learn that I WASN\'T as bad off as I thought I was, but I was sad at how bad most people with OI are.      There\'s a lot that goes on in my body due to OI, and I\'m sure I\'ll share more of that in other posts, and there is a lot that goes on in my day to day life that I feel is influenced by my disease. My decisions in life, education, love, etc. You\'ll hear more about all this later on.      Well, I think I\'ll end this for now. This is a lot to read, and I think I\'ll leave the rest of the info and stories in their own posts! Thank you so much for reading this.      .o*o.Amanda.o*o. '),(13386,3669958,'05,August,2004','2004-08-05 00:00:00','    urlLink    I WANT THIS!  Posted by  urlLink Hello '),(13387,4204123,'20,August,2004','2004-08-20 00:00:00','   Well, my sister had a beautiful, BEAUTIFUL baby girl this morning at 7:21 am, and i love her so much. I dont want to sound like im not glad or anything, but my sister is only 19 years old, and hasnt even been to college, has no job, and her \"boyfriend\", or whatever you want to call him *we all hate him*, had no job, and is starting school this monday. He hasnt had a job for like three months and my sister has been supporting his lazy @ss for the whole time she has been pregnant and on her own, since she gets help.  She doesnt want to move home and let my mom help or I, but she loves the freedom of living on her own, and \"supporting herself\", i just wanted to yell at her today and say \"You love supporting yourseslf, it\'s kinda hard to support yourself when you get help from the gov. b/c you\'re pregnant, have no job, and your boyfriend doesnt either, and your mother is still supporting you too, since you are nineteen, and she has to take care of you until you are 21.\"  It just bothers me so much, she is nineteen years olf and is wasting her life with a guy that doesnt appreciate her, and the only reason she is staying with him is b/c he\'s the father of \"her\" child.  The way he acts, it seems like he doesnt even want the baby, i think he needs to grow up.  Im afraid for her, and her baby, i dont want my neice to end up like her father, or mother, i hate to say it, but it\'s true. My sister could\'ve gotten someone alot better, and that would treat her alot better, but she had to choose him. I dont want to end up like my sister, or my mother, they chose their road in life and that is not one i want to travel. I dont want to have a baby when im 18 or 19 and have no way to support it, and then have it have a dead beat dad. My dad was a dead beat, always druck always in jail, never cared a thing about us. I never new my dad at all, he left when my brother and i were 9 months old, and never came back. I did finally get to see him though, it the coffin they were burying him in, never really felt anything on that day. Well enough of that, i hate, well i guess, it\'s my past, and my future. I hope i can make it far away from here, i hope so much. '),(13388,4204123,'13,August,2004','2004-08-13 00:00:00','   K8erz28: update your blog you poop heavenly0587: no heavenly0587: you update my blog K8erz28: give me your password K8erz28: and don\'t tell me it\'s 12345  ^^Recent AIM convo^^  So it turns out, less than 1 week after our dear Heavenly made her first blog, she is too lazy to update it.  That\'s where I come in.  See, my ambition is so great that I can manage to update other people\'s blogs as well as my own.  I rule.  Not really.  I just don\'t have anything else to do.  Sooo..yeah.  I guess I\'ll tell you a little bit about Heaven.  She likes reg-GAY music.  She writes poetry sometimes.  She\'s tall.  She\'s good at art, unlike me, because cool people aren\'t artsy fartsy like Heavenly.  Just kidding.  She loves hamburger meat.  She thinks vegetarian hippie treehuggers are stupid.  She often refers to me as \"gay\", \"fag\", or \"crystal clear\" (hence my blog address, which she reluctantly decided to NOT include in her blog).  She likes those scruffy long haired haven\'t-showered-in-a-week kind of guys.  Oh, did I mention she loves meat?  Medium rare steak covered with mushrooms and onions is her favorite dish. '),(13389,4204123,'11,August,2004','2004-08-11 00:00:00','   Title says it all '),(13390,3443825,'30,May,2004','2004-05-30 00:00:00','   Pain  why do I exist, on my own  it hurts like a fragmented bone  The worst black nightmare I scream blue  Without your touch I feel untrue  I need your touch, for this I dream I need your soul, to hold you near. with you near, my soul will beam without you here, my soul does tear  I know you are the one for me to be with you, will set me free life is not life, without you here and to lose you, my greatest fear.  '),(13391,3443825,'26,May,2004','2004-05-26 00:00:00','   When i love you   And you love me   You make my heart turn to gold   You make my time with you golden   Our love is a gold crown   And when I am with you   Time stops  '),(13392,3443825,'25,May,2004','2004-05-25 00:00:00','   From the rising of the sun To the rising of the moon I wish i was with you From the dreams of my sleep To the thoughts of my days always are of you  '),(13393,3443825,'25,May,2004','2004-05-25 00:00:00','   I dream of the day of glorious you Like a goddess, my dreams worship Such beauty, men lose their breath your green eyes makes me frozen your brown hair, like flowing silk  such soft words, they float and stun your ambient air of love entraps me intoxication, liberation, purification my soul is free, my heart complete to be complete with you, a dream  '),(13394,3443825,'25,May,2004','2004-05-25 00:00:00','   Each breath i take  Is a breath for you  Each thought i make  Is a thought for you every pain and ache  is from missing you  Every dream i take  Is a dream of you  every move i make  is a move for you  every time i love  it is love for you  '),(13395,3443825,'25,May,2004','2004-05-25 00:00:00','   water washes with water like my soul with yours mixing, into a glorious sea filled by rivers of feelings and tears of joy  stirring a passionate storm that gives the sea a heartbeat as the waves slide together into a brilliant calm  '),(13396,3443825,'25,May,2004','2004-05-25 00:00:00','   I told the river i loved you the river rippled to the sea the sea whispered to sky and the sky wept tears of joy  '),(13397,3443825,'25,May,2004','2004-05-25 00:00:00','   If my love for you was a stone  It is the strongest stone  It is the richest stone  It is the most beautiful stone  It is not a ruby  It is not even a diamond  But a stone of pure love  for that is what you deserve  It is unseeable by man  It is uncreatable by the gods  And I give this stone to you  as part of my love  '),(13398,3443825,'25,May,2004','2004-05-25 00:00:00','   walking in the fields of bliss in the plains of tranquilty I noticed your flower the scent of your love that made me grow roots that mixed with your roots stalks growing together stalks wrapping round  and bonding forever  '),(13399,3443825,'25,May,2004','2004-05-25 00:00:00','   When I look at you My breath turns to perfume My soul springs from the desert Roses spring from the heart  When I love you The deserts have rain Flowers bloom, birds sing I was in pain before  When I touch you Sand turns to silk  souls merge together and together we sink into  The sea of silk  '),(13400,3443825,'25,May,2004','2004-05-25 00:00:00','   Force   If my love for you  could be converted,  into another force it would turn the tide and stop the yangtze  '),(13401,3443825,'25,May,2004','2004-05-25 00:00:00','   How I desire my sweet red wine And how you make me feel The wine tasted from your lips to touch and feel your neck to feel your full body  to merge as one  drunk from your presence happy from our harmony How I desire my sweet red wine  '),(13402,3443825,'25,May,2004','2004-05-25 00:00:00','   I sit and cry, your soul is missing  Nothing to touch, nothing to feel You are so far away, like gold from the rainbow, from eldorado shadows before me Glimpses behind me I hold my head in dispair sad that you, are not here like a frame with no picture I hang my sorry head wishing for your touch like the pollen needs the  wind  '),(13403,3443825,'25,May,2004','2004-05-25 00:00:00','   Want   I need to feel your love  Like the wind feels the trees  to feel your warmth  like leaves feel the sun  to feel your soul  like the ghosts of ancient lovers  to feel your loving touch  Like the sea strokes the sand  to feel complete  like two perfect souls  in embrace, forever.  '),(13404,3443825,'24,May,2004','2004-05-24 00:00:00','   Leaves    Words for you are not enough cliche written words of love  The words I want for you  are greater than love itself If I say,  I love you These words are like a  leaf in a forest,  I want to give you a tree I want to give you the forest My words fail me My love for you is a forest  '),(13405,3443825,'27,June,2004','2004-06-27 00:00:00','   Spring  when our hearts first spoke The sleepy sun it awoke which pulled back the grey curtains  causing tears from the mountains  which made the trees grow to the sky this made the weeping willows cry this made the rabbits bound the birds sing beautiful sound young deer come to play That chase the grey days away all this growth, love and life when our hearts first spoke  '),(13406,3443825,'26,June,2004','2004-06-26 00:00:00','   Her reflection, gazes at her,  in tribute to her admiration she wants, but she cannot have tempted, by that on display longing to have, like the apple her face floats, over glass ice  Ghosting from view to view  on a grey sunday afternoon Walking with windows  '),(13407,3443825,'13,June,2004','2004-06-13 00:00:00','   I am the flower Your smile is the wind Your love is the earth Your soul is the light I am the flower - without  Earth, wind and light I am nothing  '),(13408,3443825,'06,June,2004','2004-06-06 00:00:00','   I dream of you and me  under the cherry trees in full spring glory can you take me there? And love me there? to live in your eyes and sleep in your heart trees blossoming us in their crowing glory to be with you forever like the oyster with the pearl  '),(13409,3443825,'05,June,2004','2004-06-05 00:00:00','   If I were, born a bird of my love, I would sing of my hopes, I would dream of life, i would give for you. Of search for you, i would go to fly over river and vale to glide over mountain and rock to hover over desert and sand to sail against tempest and tide to sing of my love to you  '),(13410,3443825,'25,July,2004','2004-07-25 00:00:00','   I flip another coin  Into the well wishing you were here too long i have done this day after day week after week month after month sad time spent alone each day a lonely heart throwing another coin into the deep black well listening, watching it fall never hitting the bottom like these days never end  '),(13411,3443825,'10,August,2004','2004-08-10 00:00:00','   when I see  the letters you  send to me I forget the paper I forget the pen and I dance with the words  when I see  your eyes  looking at me  I forget the sky I forget the earth and I dance with your soul  '),(13412,3443825,'10,August,2004','2004-08-10 00:00:00','   My words fail you for when i say you are beautiful I mean to say a whole lot more for you deserve a whole new language for when i look at you rome is built in a day and the gods found new words of love in their forgery for these are the words I would give to you  '),(13413,3443825,'10,August,2004','2004-08-10 00:00:00','   the wall fell and we kissed this felt like reaching the life giving oasis at the end of the desert  '),(13414,4049195,'29,July,2004','2004-07-29 00:00:00','   Lydia, oh Lydia; have you met Lydia? Lydia the tatooed lady.  Homework is yet to be done. This post only exists to remind myself I have a blog that still exists and \'needs\' regular updating. After having several blogs on rival websites that have failed miserably, I feel I should start with something on the site that more or less started it all.  May start transferring thoughts online. These have already been compiled with papier mache and twigs in the garden shed, and simply need a lick of paint before they are released on the world.   They will be stolen. This frightens me. Well, at least I\'m the only person aware of this page thus far. This makes me joyous.  '),(13415,4049195,'27,July,2004','2004-07-27 00:00:00','   As the tennis ball howls at the moon, an elderly teenager sits at his laptop, considering starting the English homework he should have done last week. Alas, it has not been done; it will be put off until the very last minute like everything else this year.  He has joined the mob, jumped onto the bandwagon. He is now a blogger. What the hell is he doing?  He is  Daxiel Kahnsworth  and requests that anyone reading this portion of faeces does not judge him too harshly on the disgusting mediocrity of this first entry. Why does he care about approval from faceless online entities? The answer will not come with time.  '),(13416,4049195,'04,August,2004','2004-08-04 00:00:00','   Well, as expected, the matinee yesterday went pretty crap. The crowd was composed of the usual suspects; old people and primary school children. The elderly members of the audience either didn\'t hear or understand the jokes, and the primary school children left halfway between Act Two. More than a slight debacle, but we kept going through it all and survived. My microphone was falling off my head consistently through most of the songs I sing, especially giving me trouble through the choreography. It\'s bad enough to wear Madonna-style microphone headsets, but if they don\'t perform their intended function, it goes beyond a joke.  Tonight all was amended. The microphone no longer sucked because the tech crew worked out how they are meant to sit on your head. After I got over the embarassment of finding I was wearing the thing in entirely the wrong fashion, I was intensely relieved that I could relax on stage without fear the microphone was about to eat my face or fall on the floor. The crowd was pumped, laughed at all the right places and enjoyed the bits I love performing. I\'m truly having an awesome time.  The surreal nature of time\'s behaviour when one is on stage goes beyond explanation. Time simply races, and the whole experience is entirely out-of-body goodness. I will miss  Hot Mikado  by the time Friday night comes, but having my life back is something I will not take for granted!   Homework I have not done because of  Hot Mikado  and blog:    Preparation for English Creative Response SAC on  The Quiet American   Maths sheet  Public speaking preparation (more on this later)   Well, that\'s it for now. The mundane rubbish about homework was posted because only I read the blog at this stage.   If you come back and read this, if the foolish idea of advertising my nonsensical ramble to the world occurs, do not blame me, for you were not meant to read this. It is your own fault and I shall hear no complaints.  Hail Mary full of grace, help me find a parking space. I\'m not Catholic.  '),(13417,4049195,'02,August,2004','2004-08-02 00:00:00','   Hot Mikado  dress rehearsal today went totally shit. I\'m shit. The chorus is shit. The microphones are shit. It\'s all shit and that\'s the way it will probably stay until Friday night when we will put on a powerhouse mediocre performance to end a shitty run of shitty shit.  Those worries discussed in \'The butterflies attack\' were not without reason. I shall write again after the matinee, simply to tell you how poorly it went. Hmm. Argh.  '),(13418,4049195,'01,August,2004','2004-08-01 00:00:00','   Is it worse to be the man who intentionally isolates himself from others by pretending he is weird, or the man who is unconscious of his weirdness and wonders why he remains distant from the world?  This week, Kahnsworth appears in an \'academic\' version of a disgusting bastardisation of Gilbert and Sullivan\'s  The Mikado . Named  Hot Mikado , its world is a suprising accurate caricature of Japan today. The main Japanese elements remain, but all else has been replaced with American features - accents, clothing, all reminiscent of 1940s America.   The nervousness would be easy to deal with if I could actually justify my involvement in the musical. I am continuously ambivalent about my part in it, consistently revising why I choose to be in it in the first place, and why I continued my involvement to the very end.  It is a school musical, so the bar is not raised very high. I want to exceed the bar, impress all, a triumph of form and meaning. I love chucking that in writings all the time, because it sounds like I know something about art.  Is my entire personality a facade? Although at times the answer would certainly be yes, no, no it isn\'t.  The people I have met and got to know through the musical are superb. But have I actually gotten to know any of them? Continuous superficial conversations on my part have led to my knowing I can talk shit with many people, but not many new friends. The amount of friends I have at the moment is not alarming at all; I am happy, I have several.   But, as the adage from  Full House  goes: \"Friends are like jello, you can never have too much.\"  So it\'s either poor grammar or poor semantics if you are to quote  Full House  correctly. Oh dear. How American. Nevermind.   Post Script:  Eric Bana has been announced as the new James Bond by  News of the World , an exceedingly unreliable tabloid. It would be extremely interesting if it is true, but I will continue to doubt the reliability of this information until I see Bana in the next Bond film.   Why, when I write \'Bond\', do I consistently make a typographical error and write \'Bong\'? I\'m not addicted to marijuana... yet.  '),(13419,3977680,'30,July,2004','2004-07-30 00:00:00','   hiAz.. sO siaN.. bUt h3e.. Lat3r gOin watCh moVie ( bRothErhOod ) wIf my SisTaz, aManDa.. :) Miss hEr sIa.. My faUlt.. LasT sUn wOrk So cAnT waTcH wIf heR.. yEaH 2Day gOin tO c Her l3.. d lAst dAy we gO oUt iZ tT daY wHeN we ceLebRate Jia\'s bDay.. haHa.. u All duN tHinK taT i \'like\' her Hor!!! we sIsTaz wAt.. Miss d TimE we All gO ouT 2gEtheR.. k Lah.. i stOp h3rE le.. bY bY c U gUys.. StuDy haRd.. Hop3 i caN mAke iT fOr my TerM teSt nxT wK.. hee.. :) gd Luck 2 mI n evEry1 eLse..  '),(13420,3977680,'26,July,2004','2004-07-26 00:00:00','   yAwn.. siAn.. nw In lAb dOin nThiN bCoz fIniSh sCh @ 10 l3.. :P tIs f3w daYs tIred siA.. woRk + scHliN.. nO tiMe oNlin3 Or wAtcH tv.. hee.. wah 2dAy i feeL so n3rd n TooT bcOz i weAr spEc 2 sCh hahA.. how? Nxt wK tErm teSt le sCaRe faiL leH.. dUn reAllY knOe sUm (or raTheR mOsT) oF d ThinS teAchEr teAcH.. Haz.. hoPe cAn go Out wIf my sIsterS soOn.. mIss thEm oTz... :)&nbsp; bTw sOrrY aMandA laSt sUn caNt gO ouT wIf bCoz wOrKIn ReaLly SoRrY!!! sO sIan.. nThiN 2 sA l3.. By ev3ry1.. Tk cAr3.. sTuDy haRd!! :)  '),(13421,3977680,'22,July,2004','2004-07-22 00:00:00','   hiy3e.. hee 2day nv listen in carc lec.. :P so sian jus nw gt cmsk  urlLink email   urlLink test .. cant find 10 errors leh.. haha only can c abt 6 to 7.. hope can pass d  urlLink test  though iz nt tt impt.. remember d  urlLink job  interview i mention last nite? Yip3e.. i gt in.. startin work on tis sat.. hope iz nt a bery tough 1.. xi wang wo ke yi cope work n studies well.. i muz try 2 really revise my work le.. if nt i really sure fail my term  urlLink test .. no! i dun 1.. dun wan go retake n study while other ppl r enjoyin themself.. hee.. k tat all 4 2day.. wan go find research 4 cmsk le.. by by.. StuDy hArd wOr!! :) n last but nt least.. jIa HappY bIrtHdAy 2 u!! :) aL d b3st 2 u.. mUacKs..'),(13422,3977680,'21,July,2004','2004-07-21 00:00:00','   sianx.. tis wk i keep takin cab to sch sia.. waste my $.. shuld try to wake up earlier sia.. hope i&nbsp;from nxt wk onwards wun keep tan shui..&nbsp;:p&nbsp;if nt really bcumin broke le.. xiang get a part time job leh.. but arts laser nv call mi.. :(&nbsp; tomolo goin espirt job interview.. thnx to sally 4 rememberin i wan a job too n ask mi along.. :) hee hope tomolo d job interview will b ok!'),(13423,3977680,'16,August,2004','2004-08-16 00:00:00','   Yoz.. i so long nv cum online le.. almost 1 wk+.. bo pian.. pc gt prob.. sian.. how u guys? our term break finish le.. today back to sch study.. sian.. die ar.. all my 4 tests sure fail de.. hiaz.. shuld try to listen in lectures n do hw.. if nt ar end of 3 yrs cant get diploma thn wah.. time pass so slow sia.. still gt 1 more hour for lab.. bored.. k nthin to write le.. by by.. hope u guys done well in ur tests.. Tk cAre!  '),(13424,3732960,'27,June,2004','2004-06-27 00:00:00','   how odd are we all? keeping an online journal so anybody who\'s anybody can find out what we\'re doing on a day to day basis. id say we all never got enough attention when we were younger....  im bored... i could really go for a smoke, but its raining outside and that sucks.  lets see... whatd i do friday... i met up with kels at 2:25ish to go to work the phish show at 3 at alpine. so we saw kreger at the show and did nothing (and got paid for it) for like 2 1/2 hours. then my team leader forgot about me, so i didnt get a break till 10ish. i bummed a cig from some chick and it was a lite, so that lasted me all night until after everything was done and we were goin home at like 120 in the morning when i went up to some other girl and asked to bum one of her camels which she did give me. so then i went home and my sis was home from college and we sat in our driveway and smoked a cig and she told me how horrible her date went. then we went inside and talked about Bonaroo (cuz she went) over a beer. then i went to bed cuz i had to work at steins the next day.  which brings us to saturday. i went to work from 1230 - 9pm and during the day me jamie and katie decided we\'d hang out. so after work i came home and took a shower and called them up and they came and picked me up. so we drove around and smoked 2 bowls of katies dank ass nugs, and then went to dennys. in the car, the 3 of us were geekin. so when we got into dennys, jamie went to the bathroom and when the waitress came, me and katie geeked for like 10 minutes (which was bad) cuz it was so obvious we were fuckin ripped. then one of our bosses from steins walked in with his girlfriend around 11 and we were like oh fuck. but nothin bad happened. later when i started getting tired and we were all done eating, katie went to the bathroom and jamie was like \"youre tired arent you\" and i said yea but i still wanted to hang out with em cuz i didnt wanna seem like i was having a bad time or anything, which i wasnt, but jamie was like \"dude its no big deal, cuz i know you have to work at 9 tomorrow\". my ripped ass was like ok. so they dropped me off at home and me and my dog sat around till like 130 watching shitty TV.   which gets us to sunday. i went to work at 9 (till 1) and it went by surprisingly fast. when i first got there jamie told me about how her and katie accidentally ran somebody off the road cuz jamie was fuckin around in her car and not looking and she ended up in the wrong lane, which made me laugh. then i went and worked till i got to leave.   cool thing happened this weekend. i found out jamie likes horror movies and reading about true murderers and stuff. which was really cool, cuz i like that shit too... i love scaring the shit outta myself for some odd reason. so anyway, jamie let me borrow a couple of books she has about scary people that were murderers. then i went home and took a shower and me and my brother went to best buy, circuit city, and soundwaves. i bought dead or alive 2:hardcore cuz its one of the only videogames i actually like, Gia, and donnie darko. and that brings me to this very moment... queens of the stoneage \"No one knows\" is playing. i like that song.   '),(13425,3732960,'24,June,2004','2004-06-24 00:00:00','   PRETTY SHITT!   lets see, everyday i have to wake up at 9 to get ready and go to summer school. then every couple of days i get to come home, take a 2 hour nap, then go work at Steins. and i have to devote a large handful of weekends to working at alpine valley (music theater) and thats not really fun, cuz you get there at like 11 in the morning and dont get to go home till 1 or 2 AM the next day! last year was pretty fun cuz we got to do whatever we wanted there, but now they changed everything and you dont get to walk around everywhere, you have to stay in one parking lot, and you get switched around, so sometimes you\'re working in the show but the only plus to that is you KINDA get to see the show, but the shitty thing is youre checking tickets and standing ALL NIGHT LONG.   so its not even worth trying to go out and party or anything cuz if its during the week i have to get up at 9 and go to school and then go work at steins, and if its a weekend im working at either steins or alpine!   now i know what youre all thinking.. yea but youll have money... WRONG! i still owe my dad like $450-$500 for car insurance and a large chunck of the rest goes into savings... FUCK SCHOOL, FUCK WORK, FUCK DEBT....this summer sucks   '),(13426,3732960,'23,June,2004','2004-06-23 00:00:00','   the life i used to have. my freshman year, i had two or three really close friends who would come over EVERY weekend and drink and smoke at my house. seriously for 6 months straight all we did was go in my basement and take shots and then around midnight we\'d head outside and do stupid shit all night. we\'d sit in cait\'s car and smoke or take walks down the road and smoke out of our homemade bongs and shit (occasionally we had a pipe) but thats besides the point. now everyone is stupid and doesnt know how to have fun. one of the close friends i had cant smoke or drink anymore because his girlfriend wont allow it and the others all graduated and have work everyday and shit... and im only a senior.  I started talking to my one of my friends again today. Monica AKA monular. we went to school together freshman year, then she moved and now she goes to a different school. but hey at least we started talking again.  this year i had to \"buckle down\" and worry about grades and stuff cuz i want to go to college and be able to support myself when im older, but i just missed constantly being high and drunk. ive only smoked three times in the past 2 weeks... if this were a couple years ago... i couldnt have gone 2 weeks without smokin at least once a day, and im not talking one bowl, im talking fucking RIPPED off my ass. and one of the best nights of my life was when i was on x.... everything was so fucking awesome, i felt sooo good and everyone was having a good time. i miss shroomin too...  you know youre good and fucked up when your living room starts getting smaller and smaller. heh.   last year was kinda like that. but i always make friends with older people and then they go and graduate. i had a good share of my drunk and high times, and now my BEST friend in the world is moving away and getting married aug. 21st and the rest of them are all going to school and working everyday. this sucks...im going to bed.   '),(13427,3732960,'23,June,2004','2004-06-23 00:00:00','   im bored.  i hate all that lord of the rings shit. granted, ive never seen or read any of the stories... its all just so \"dungeons and dragons\"-ish and all that shit is stupid (in my opinion). Im not a big fan of the whole harry potter craze either, same with all that Pokemon shit... never liked it at all. I played it once, and had absolutely no idea what anything meant, so i stopped. Fun things are easy to follow. No complicated story lines or any of that shit. Unless youre talkin about the movie memento... now there was a good complicated movie. ive seen it like 5 times now. im listening to \"Control\" by Puddle of Mudd.... just thought id share. im headin out for a smoke.... be right back (11:22 pm)ok im back (11:26 pm) well that was fun. i\'ve got \"Broken\" by seether playing now.. good song. My friend Becky is hilarious. she always tells me that the first time she heard me talking was when i was talking to this other girl about drugs and she thought to herself \"i hope that kid never talks to me\" now we\'re best friends. I think we get along so well cuz we\'re both really sarcastic.. about EVERYTHING!                            *AIM CONVERSATION* me: i hope your watching TV er something, cuz i dont have one with cable in here, so if it says somethinglike hey get in yer basement tornado to the fullest... i wouldnt know becky: lol becky: tornado to the fullest becky: is that like the BET weather channel  BET is black entertainment television for all you out there who don\'t know that. *wink*   i hate living in a small town... and i hate being in high school (only one year left thank god) you\'re so sheltered from everything, and i hate being naive.   '),(13428,3732960,'23,June,2004','2004-06-23 00:00:00','   urlLink    Me at my friend Terri\'s house&nbsp; urlLink   '),(13429,3732960,'23,June,2004','2004-06-23 00:00:00','   I planned on starting this thing on monday, but that never happened, so ill go back to the beginning of the week.  fucking summer school started on monday... god damn shit fuck bitch cunt.... ok ive vented. kathy (my better half) left for virginia beach on monday too.. she\'ll be gone for a week. so yea i woke up and took a shower and got ready for school and i realized i was kinda early. so i called my friend katie up and asked if she wanted to smoke. eventhough she\'d just waked n baked she said sure. so i picked her up and we smoked. then i had to go to summer school, so i drove out to bumfuck mukwonago and sat in school for 2 hours... the whole time fidgeting cuz i needed a smoke. after school came home and slept, cuz schwag does that to ya. (Dont get the impression that me and my friends are dirty schwizzo smokers, she got it for free, so what the hell, eh?) then i woke up and called my friend erin and we had to go to the library for a book for my summer school class... then we went to her place and watched the end of Futurama and Family Guy, then i decided to go home  tuesday was shitty. woke up, went to school, came home, watched tv for awhile, then worked 4-9, came home, watched 50 first dates, and went to bed.  that brings us to today. (listenin to the beatles, \"Love Me Do\" is on) so i had to get up early and get my work permit for alpine valley, and i forgot how long that shit takes and i cant be late for summer school, so i left at 9:15 to get that shit done and got to the place at 9:25ish and realized that all she had to do was sign and photocopy some stuff... so i was going to be hella early to school. i decided to entertain myself and drive out of my way on a cigarette roady.. too bad that only took up 5 minutes of my time. i got to school around 9:40 and sat and (\"From me to you\" is on now)read from this book SMACK by Melvin Burgess. Its about running away from home and getting addicted to what else... heroin. sounds good, ill keep you all posted. then i came home and watched tv for awhile then listened to music.   '),(13430,3732960,'07,July,2004','2004-07-07 00:00:00','   yea so my dad\'s been being a penis as usual. he always walks around like we\'re all ruining his life, when in actuality none of us are doing anything wrong. the other day he accused my mom of hiding his bills.... explain this to me somebody!! why would my mom hide HIS BILLS, that she has NOTHING to do with!? (which is exactly what she said to him). Then i hear he\'s bitching because me and my brothers dont do his yardwork anymore. my ma says he said something along the lines of \"Im not doin all this yard work myself, those boys dont do anything around here\" or something to that effect... ok, me and my brothers are no longer 10, 12, and 16 years old anymore! we\'re 17, 19, and 23! we all have more important things to do LIKE OUR JOBS than his yard work. I have summer school every day, and then most of those days, i have to go to work a couple hours later! and then on the weekends im either working one job or the other, or both! i dont have time to do that shit and he knows it. seriously, he just likes bitching and acting like no one is useful except himself, and its really fucking annoying. Im sick of him telling me how im useless and getting a \"free ride\" because when i was younger, we busted ass helping him with stuff. Sorry dad, that we have lives now.   '),(13431,3732960,'03,July,2004','2004-07-03 00:00:00','   Yesterday me and kathy went to lunch after summer school... yes i know, subway is so romantic. Then i went home and slept and she went shopping and to get Sally fixed. When i woke up and got ready, we headed to summerfest. we\'d panned on seeing switchfoot. then we decided to keep our options open, cuz we\'d both heard \"meant to live\" 100000 times on the radio, so we were debating between switchfoot, dashboard confessional, and trapt. What did we end up seeing??? NONE OF THEM! We just walked around the whole time. I saw some cool people. Brie, Heikinnen, Menghe, Christina and Sidlarz, Britney, Dan, Chiello, Hannah, and my friend jessica, so that was cool. But the plan was to meet up with some of my other friends and get shit-faced. turns out they decided to leave their cell phone at home, and i didnt find \'em all night. Then we got to the car to go home and got stuck in the parking structure for about 45 minutes, and there were stupid people who kept laying on the horn cuz there was a huge ass traffic jam and it just started to give me a headache, but when their stupidness died down, i was fine. Then me and kathy decided to say fuck you to the huge traffic jam getting on the freeway, and we went the other way.... only problem: we didnt know how to get home that way. Yea Turns out, kathy\'s a spectacular driver and navigated her way from there to the freeway and we got home around 1:15ish. I made myself a drink (which i\'d wanted to do all night) and then kathy decided she should go, cuz she had to work this morning at 9... poor kathy. **Dumbest License Plate: \"GV THNX 2\"  Thursday I went to school, came home and slept, went to work, then came home and did nothing.  Wednesday I honestly can\'t even remember what I did.  Tuesday I went to school, then went to kathy\'s right after and she made me lunch. Then we were bored and decided to go to Soundwaves, then to the mall... i made 4$ (kathy made 6$) take a survey at the mall... it was pretty worth it, cept the lady asking the questions was a fucking crackhead.   '),(13432,2645566,'26,January,2004','2004-01-26 00:00:00','   I know that I haven\'t really been writing lately...but alot has been going on.   So this post is just to tell Tori thanks. You\'re there for me when I need you most and I don\'t even have to ask. Especially today, you took your time and spent it with me just because I was having a bad day. I can\'t thank you enough. You are the bestest.   '),(13433,2645566,'25,January,2004','2004-01-25 00:00:00','   turns out that \"holding steady\" and \"stabilized\" means \"in a coma and going quickly.\" my great-grandfather died on thursday night.   molly got here on friday. i had so much fun with her. she really needs to move back. she\'s one of my best friends and i hate the fact that she lives so freakin\' far away.   '),(13434,2645566,'22,January,2004','2004-01-22 00:00:00','   recieved a call at 10 til 7 this morning. it was as follows:  \"I just wanted to call to tell you that grampa is holding steady\"  That was the important part. I wish he would just die so he was no longer in pain.   '),(13435,2645566,'20,January,2004','2004-01-20 00:00:00','   They took my grandpa off the hydration tubes. They are going to give him morphene to alleviate the pain...that\'s all they can do...they\'ve given him less than a week.  I HATE THIS!!! Who are \"they\"?! Who are they to let my grandfather die?! Don\'t they know that he\'s important to me, to my gramma, and great-gramma? To my entire FAMILY???  It\'s not fair.   '),(13436,2645566,'18,January,2004','2004-01-18 00:00:00','   My great-grandfather is in the hospital again. This time they say that it is inevitable. He will soon pass.   This all started because of his damn seizures...why God? Why do I have to loose someone so close to my heart? If he hadn\'t had that seizure, he wouldn\'t have broken his hip. It wouldn\'t have resulted in this mess.   He can\'t swallow anymore. They are keeping him hydrated through an i.v. His living will demands that a feeding tube not enter his system. The doctors diagnosed that we will loose our loved one in a matter of no more than a month.   I hate not being there for my grandma...I hate not being there for my great-grandma. I hate not being there for my Grandpa Fred who only has limited time left on this world.   There will be no funeral. Grandpa Fred will be cremated and his ashes will be spread at sea. They are going to try to spread his ashes at the same longitude/latitude that they spread my aunt from. I love my grandpa. I wish I was there.    Another soul was lost yesterday. Dustin...Deborah\'s son. Family friends. Dustin shot himself in the head earlier Saturday morning. It\'s horrible...he was trying to sell that very gun one day prior. He was only 16. That\'s my age for goodness sakes.   In closing: This is a cruel and hateful world that we call home.   '),(13437,2645566,'13,January,2004','2004-01-13 00:00:00','   This weekend was...a blast. Thursday afternoon we left for Mid-state and honestly, I didn\'t want to go. But boy am I glad that I did. The ride up to M\'boro was interesting to say the least. Drew, wow, and Alex...man. Rooming with Leah and Andrea was great. It allowed me to get closer to two people that I considered friends...but weren\'t that close to. The first night we watched a movie and kaci came in (our rooms were connected..FUN STUFF). The next we stayed up until the wee hours of the morning just talking about whatever came to mind. It was great. I got to see people from other cities...friends...that I never get the opportunity to talk to. And that was just the beginning.  Rehearsals were hilarious, tedious, tiring, fun, exciting, etc. I don\'t know what was more funny: the fact that our conductor was old and didnt get a pattern and got excited over little things or the people that I was sitting around making fun of him. Either way, the music was great (minus one piece) and it was worth going. I\'m glad I went.  Explanation of my name: On the way up to Mid-state, I had for some reason touched Drew\'s hair...a very rough experience. He said that all it would need was some conditioner..but he doesn\'t bother with that. Then he proceeded to say that a \"friendly\" shower would help as well. That would include a shower with a girl...hmm...interesting. So we joked about that for the whole weekend. Finally we concluded that \"conditioning\" meant having sex in the shower. So, \"Man I need to condition!\" would mean, \"Man I need to have sex in the shower!\"...good times...not having sex in the shower, but talking about that whole process...shutting up now.   '),(13438,2645566,'13,January,2004','2004-01-13 00:00:00','   Wake up. Go to school. Leave school. Come home. Sleep. That\'s my day in a nutshell. You would think that after so many years of the same thing that something would change. I have a wonderful boyfriend, but I don\'t get to see him due to the fact that he lives in Manchester. It costs a little too much for us to drive 20 miles every day so we can see each other. So that leaves barely any options for the afternoon. Spend time with friends, do homework, come home. That\'s what I usually do but you can only do that so much. I find myself devouring my time with books and stories (usually stupid little things that will not benefit my being at all). Maybe leaving for mid-state on Thursday will have a positive effect on me. At least I will be spending time with different people from all over middle Tennessee. I vow to make the most out of Thursday, Friday and Saturday...my midstate experiance for this year. Now I will just go off and read...hoping for time to hurry and pass along so I can see Paul tomorrow. Then I\'m off....   '),(13439,2645566,'10,January,2004','2004-01-10 00:00:00','   What do you do when your whole day is centered around your baby cousin? Okay..so maybe it wasn\'t my *whole* day but it was close enough.  Last night I had to go over to Brookie\'s (4 years old) house to see her \"Ch-ch-ch-chia\" pet that she was making. The sweet little thing wouldn\'t let me leave until I promised her that I would take her out to lunch today. So now that the scene is set...here\'s the story:  I arrived back home from Tori\'s house at around 9:30. Two hours and counting until I had to pick Brookie up at her house for lunch. Finally got over to her house at around 11:45 to have an anxious little girl waiting for my arrival. She ran to the door, opened it and practically dragged me into the house so we could leave. Brookie\'s choice for lunch -- good ol\' Mickey D\'s... and of course, like every little kid, she insisted on playing. So I spent a good 2 hours in McDonald\'s today. After that escapade we returned to my house. After many protests from the munchkin, I finally get to take her home. *end scene 1*  So now I\'m home...resting...before I have to meet everyone at China Wok. I fell asleep on the couch. Smart me. I woke up at 4:35 and I had to be at the restaurant at 4:45 (no one ever credited me with being the brightest crayon in the box). And (just my luck) on the way there, I get stuck behind a truck going 25 mph in a 35 mph zone. I was too close to shooting myself. Dinner, once I got there, was a great experience. It was Ashley, Tori, Emily Em., Kerry, Cody, and I. I think it was even better because Emily was actually able to come. It\'s really nice to be able to see her again. So we are sitting around the table and good ol\' Tori decides to call Paul for me (which is really funny because she just screams his name into her phone and it automatically dials it)...He actually wanted to do something! so the plan was that he would meet me at my house sometime after 7. ~later~ everyone has a great time and departs on good terms. *end scene 2*  I pull into my driveway and there\'s an extra car at my house. Relatives. *enter Brookie* Yes, they were at my house to watch the Titans game...but whenever MY family is together, the process is relatively the same: Torture Cassi. Tonight was 100% lighter on the adults behalf (too engrossed in the game)...but sweet little Brookie took an immediate liking to Paul. So when Paul and I went into the other room to watch a movie...we have a tag-a-long. Paul and I also make the mistake of kissing in her presence (\"I was getting ready to leave the room because you guys were kissing!\") \"Psst...Cassi...are you in love with Paul?\" (she\'s very intelligent...no one knows where she got it from) \"Yes, Brooklyn, I am,\" I responded. \"Good, \'cause Paul is in love with Cassi too.\" (he\'s such a sweetie) ~short while later~*enter Lori, Brookie\'s mom* \"Psst...Mommy...Cassi is in love with Paul.\" \"Really?? And is Paul in love with Cassi?\"  \"Yup!\" So naturally, now my whole family knows how head over heels I am for this guy...and vice versa. \"how sweet\" you may say? You don\'t know my family. I will *never* live this one down.  \'Night.   '),(13440,2645566,'10,January,2004','2004-01-10 00:00:00','   Okay, first of all...I just want to see if this will work. I hope it does because I am going to proceed in telling about last night.  Last night was...crazy. Instead of sitting around and waiting for Paul to call (like I usually do), Tori and I took matters into our own hands and decided to go out on a date. Ruby Tuesday\'s here we come! Sitting there, we begin to talk about whatever happens to come to mind. There was lots of laughing. Then our waitress comes to get our drink order...2 suicide lemonades. Not 2 minutes later, she returns...\"Can I see your i.d.\'s?\" What?! We just wanted a mixture of kiwi, blackberry, and strawberry lemonades. *insert mad laughing* Then comes our dessert...Chocolate Tallcake. *yum* Throughout the rest of the night many jokes were made. \"That guy\'s tall and lanky...Oh D*mn!\" \"STAND UP and PLEASE tell me there\'s a truck there!\" (Tori thought she was hallucinating...there was a pillar blocking my line of vision to where the truck actually was...turned out to be quite humorous) \"Look at me Weena!\"\"I thought he was just a retarded kid.\" (The family next to us had a baby...the elder kid was trying to get the baby\'s attention for a picture. The baby\'s name was Hannah or Anna...but we heard \"Weena\")...Many great memories came from Ruby Tuesdays...Stuff that, at least for now, Tori and I can keep laughing about. And you may not think it\'s funny...but I guess you had to be there. Or maybe it was just the fact that we were spending that time together that made it better.   Either way, I\'m glad that I got to spend that time with Tori last night. It was definately better than sitting at home. And in a way I feel as if it has drawn us both together...more so then we were prior to that experiance. I\'m glad that I have a friend like Tori that I can open up to about whatever is troubling me..or whatever is on my mind...or even for the times that I am *truly* happy and need to tell *someone* about my experiance...someone that won\'t hold it against me. Thank you Tori for being there for me. I appreciate it more than you will ever know.   '),(13441,2645566,'29,February,2004','2004-02-29 00:00:00','   Now folks, don\'t let the title fool you. I still am in a wonderful relationship with the man I love. Now that that is understood, I will explain. Everytime we are together we always end up tickling one another. He will say \"HEY NOW!\" and then I will chime in with it as well. So we were sitting around one day and he said \"Hey now\" and I said \"hey nooow\" like in the song. But we didn\'t know what song it was. Now we do. That\'s the story. You probably think it\'s stupid, but it means a lot to me.   Yesterday was...well, I don\'t know what it was exactly. In the morning, Paul and Chris came over and Chris watched Jordan while Paul, my mom and I went to Winchester to see The Passion of the Christ (the movie was indescribable; I can\'t believe one person would give up so much just to cleanse those yet to come...wow guys, wow.) After that was over we came home and had lunch. Then Paul went home and the rest of us *attempted* to go to M\'boro. You see, we were gonna go get a prom dress for me (didn\'t work out) but my car kept dying on the freeway. My car sucks. So now my car is going to have to go to the shop and they will have it for who knows HOW long. And it really sucks being dependant on people for rides once you know what it\'s like to have that freedom. Suxors bigtime.  Today was a whole lotta nothin. Went to Donut Palace then to church. After church I went out to eat with Paul and his dad and little brother. That was interesting. It was also kinda sad because I don\'t get to see Paul until next weekend. *maybe* So that really stinks. I usually get to see him on Wednesday nights for a good 3 hours at least, but the stupid Colemans had to go and schedule concert festival forthat day. And it\'s just my luck that we wont get home until 10 or 10:30 AT NIGHT! Ugh...  So all in all...this weekend was interesting..boring..dull...gosh I hope I have a better week than I did weekend.   '),(13442,2645566,'27,February,2004','2004-02-27 00:00:00','   Tonight was great. I got to spend time with Emadian. Now there is something that I never get to do...mostly because in the past we didn\'t really have a way to communicate (now we have a class together). So I was supposed to do something with Paul tonight, but Emily and Allie came in this morning and told me about our original plans being sprung back into action. I had to go to Em\'s house. Plan was to mess around with her new keyboard (which is totally awesome!!) and then eat some grub and watch a movie. So the three of us were playin\' around and enjoyed dinner together (pizza and buffalo wings *yum*). Then Allie had to leave for Spiderweb. So Em and I did what any sane person would do: pop some popcorn, grab the M&M\'s and these new TollHouse candy bar things, and pop in a good ol\' Disney movie. The movie of choice was The Lion King 1 1/2. Talk about an adorable movie!!! Lots of laughs and sublte humor. They really tied in The Lion King well. I give it 2 thumbs up.  The most fun I had tonight was after the movie. Em and I just sat in her room and talked for a good hour and a half. We talked a lot about how our \"friends\" have changed. A lot of them have. Some for the better and some for the worse. It was nice to know that someone had the same view point on it as me. We aslo talked about how much certain people get on our nerves and how the past (3rd grade) was so much easier. Ya know...the simpler things: kissing on the cheek on top of the jungle gym *~!EWWW COOTIES!~* hehehe...those were the days.   Point: I\'m glad that Em and I are closer now. It seems that as everyone is changing, our group in particular, certain people are drawn together...I\'m glad that I can tell her anything and know that it\'s safe. I\'m glad that I feel the same way about Tori and Chris. I don\'t know what I would do if I didn\'t have my friends there to keep me rational in the times when I need them the most. I just want to say thanks.   '),(13443,2645566,'19,February,2004','2004-02-19 00:00:00','   So when I went to the doctor\'s office 2 weeks ago, they scheduled an appointment for today to make sure that I am recovering/recovered. They made the appointment for this morning at 9. So I get there and am waiting patiently. THEY FORGOT TO PUT MY APPOINTMENT INTO THE FREAKIN\' COMPUTER!! So my visit became a matter of whether or not I was supposed to be there. I went to my car and brought back the appointment card (because I\'m good like that). They apologized and rescheduled my appointment for next week. Good thing happened though- they gave me a doctor\'s excuse for the whole day. So I don\'t have to go back if I don\'t want to!! But I am gonna go for fourth.   Tomorrow--eye doctor. Woot!   '),(13444,2645566,'08,February,2004','2004-02-08 00:00:00','   Wow, I suppose that I should put a little update on here. On last Thursday I left for Tech. That was fun stuff. I got to see a lot of people that I never get to talk to let alone see. Jesse, Tara, Kayla, Vaughn, Darren, Logan, Andrew, and Tanner were all there. Plus I made some new friends. All in all, it was good.   Saturday was Valentine\'s Day. Since I was still at Tech, Paul was supposed to come to my concert. He called me on Friday night and informed me that he wasn\'t going to be able to come. His parents wanted to go on a date of their own...which is all swell and fine in my opinion, but it was the first Valentine\'s Day that I have been in a relationship. So anyway...he felt really bad and we were both sad about him not being able to come. We decided that we would just celebrate on Sunday.  Sunday it snowed. Sunday it snowed...a lot! During church (a mere 1 1/2 hours) we got around 3 inches. First real snow of winter. And it comes in FEBRUARY! Well, Paul and I had planned on spending the whole day together, but because of the snow, the play we were going to see was cancelled. So after some lunch at Fazoli\'s...he came over. He brought me 4 roses: one pink, one yellow, one white, and one red. It was really sweet because we have been together for four months. I gave him \"The Fighting Temptations\" on dvd. It was the first movie that we saw together in theatre on our first date. So it meant something to both of us. ....After Paul left, Chris came over and stayed the night. It was good times.  Monday we were out of school and I went to devotions. That was nice because we didn\'t DO anything--no one showed  up. There was talk of merging the two devotions (i.e. bryce\'s idea) and somehow I think we are going to try it. >_  Yesterday, Paul and I decided to skip church. So he came over and we played around and had some jolly good fun. Woot. I gave him some brownies (the rest of his V-day present) and a letter. He gave me a stuffed monkey (SO cute!! it has a belly button!!!) and a homemade card. Good times. Then we went to church and only showed up for the last 10 minutes or so. So yeah. Paul was going to say something to Bryce..but I think he is going to wait until Sunday-ish. After church Paul actually came to the Tech concert. The symphony band and the jazz band came down here for a concert. I hope that he enjoyed it, because I did. *And that\'s the way we do it when we drink a lot of beer!* Thanks to Keith for that one. (oh yeah, keith was at tech too). Well, I think that is it for now. So cya!!  FINGERS!!! AHH!!!   '),(13445,2645566,'08,February,2004','2004-02-08 00:00:00','   I don\'t know what was better, waking up and being able to breathe or spending the entire day with Paul...wait...that was a dumb comment. I SPENT THE WHOLE DAY WITH PAUL! Wow. It was amazing. For starters, I should say that my plan for the day was 1) go to church, 2) come home, 3) take a nap, and 4) work on my huge load of make up work. I am much happier that it turned out differently.   After church, Paul and I went to Taco Bell...simple, delicious, and something we both love. After Taco Bell was Wal-Mart, then finally my house. When we got to my house we kinda couldn\'t do anything because Slectues and her mom came by. They still want me to go to Florida with them this summer!!! Anyway...once they left Paul and I went and \"watched\" Fight Club. Good movie, but it was better just laying in his arms again (not to mention...*sigh*). I\'m not going to forget the, \"you better not get me sick again...\" \"don\'t worry! I can\'t. I\'m on antibiotics.\" *grin*kiss* That was nice. And then there was when we were messing around with each other (play fighting):  \"YOU NOT MY FRIEND NO MORE!\" -me...I leave the room \"That\'s okay...I\'ll find new friends.\"-Paul \"WHAT DID YOU SAY?!\"  That was followed by me pushing him into the other room (\"OOH!! You\'re in trouble now!!\" calls out my mom), pinning him on the bed (\"I like being in trouble\"), and kissing him senseless. Good times...Good times. So now \"trouble\" is on ongoing thing. Paul does everything he can to get in trouble. I really love him. He\'s so sweet when we are being serious.   I have no good way to end this...so ... peace.   '),(13446,2645566,'06,February,2004','2004-02-06 00:00:00','   Ya know, I haveta admit, being sick stinks.  Tonight is Winterfest. Okay, so I wasn\'t really planning on going to the dance or anything, but I really wanted to go to the game. I didn\'t care about the school activities throughout the week. I didn\'t care about \"spirit week.\" I just wanted to go to the game. And normally I wouldn\'t have cared. I just wanted to go because my friend is Queen!!! QUEEN!!! And I can\'t go...because I am sick. I mean, I even have friends that went tonight. And it wouldn\'t have been so bad if I didn\'t go because I had other plans. But no, that\'s not the case. I didn\'t go because of this crap.   I know that I will be up to my head in make-up work. I missed a good couple of tests, including a state-wide test that I don\'t know if I will get to make-up. I am really glad that it\'s not worse though. I mean, I could have had mono and Lord knows I am glad that that wasn\'t the case. Although, it was a good kick when I told Paul that I was planning on telling him that that was the case instead of the other stuff. I got a good laugh outta that. But I just wanna get better so I can spend time with my friends, actually go to school, and spend time with my boyfriend. Hopefully, I can get outta this jail tomorrow.   '),(13447,2645566,'05,February,2004','2004-02-05 00:00:00','   I would just like everyone to know that being sick  *SuCkS* . Yeah. So I\'ve been sick all week. The only day that I have gone to school is on Monday. Some would enjoy this, but not when you are *miserable*. And I didn\'t think it was really anything bad. Just a head cold and some nasty cough...but no. I have an ear infection in BOTH ears, some sort of throat infection with a technical name that I have never heard of, and bronchitis. Wowzers, let me tell you that it\'s a ton of fun! I guess the only good thing out of it is that I do *not* have mono. (hahahaha...now That was a funny conversation.)   Next week is a big week for me. On Wednesday, Paul and I will be together for 4 months. By far my longest relationship. AND! Then there\'s Valentine\'s Day on Saturday. That is going to be a new experiance for me. This is the first relationship that I have been in during any *big* holidays that you spend with the people that you love. I am really excited that I get to partake in that experiance this year. But of course, there is a downside to the upcoming week. I will be leaving town. So this is how I *hope* it will work out. By Wednesday I will be well again and will be able to spend time with Paul. Which will be a good thing because I leave Thursday afternoon. I will be gone Thursday evening, Friday (all day), and Saturday(most of the day). So that just means that Paul has to come to my concert on Saturday if he wants to see me on Valentines. If not, then we can always celebrate it on Sunday. Whichever he wants to do. I don\'t want to make him come to the concert if he really doesn\'t want to (even though I would be sad if he didn\'t come because it\'s Valentine\'s Day). I just want him to be happy because I love him.   Now for sooner times, I have a feeling that I need to spend more time with Chris. We usually have Monday nights and some time on Wednesdays. But I know that I haven\'t been putting as much effort into our friendship that I should. And in all honesty...I feel horrible about it. Chris and I have been friends since 7th grade. Yes we have had our ups and downs...but that is normal in teenage girl friendships. Yes, I have wanted to kill her at times and I know she has wanted to strangle me at some point or another...but I think that is what makes our friendship so unique and inspiring in ways. I\'m sorry that I haven\'t put in full effort. I will make it up to you Chris.   '),(13448,2645566,'21,March,2004','2004-03-21 00:00:00','   I got my prom dress altered today....well just pinned. But that means that it will be altered soon and I can bring it home! Woot! Ashley (yes ashley) went out there with me. That way she could see it. She said it is Beautiful. ^.^ Now my dress is long enough that it still brushes against the ground but I won\'t step on it when I walk. So it\'s all groovy.   After getting that done, Ash and I went to Sonic...in Manchester. Fun stuff....I guess. Well, it was fun when we went to Claires (back in T-town) and Dillards. I got a new skirt. It\'s uber cute. I wouldn\'t have gotten it but Ash said it was really cute so I did. And it was super cheap. Originally $34.00-Marked down to $8.50. GOOD DEAL EH?? Then I got a pair of OP pants there on Friday...originally $44.00- marked down to $11.00. Good stuff Man.  Well...I guess that\'s it for now. I don\'t really have much else to say...except I am UBER excited about prom!!!!!!   '),(13449,2645566,'20,March,2004','2004-03-20 00:00:00','   We have finally decided on prom where-abouts. Well, everything except the whole after-prom scheme. I EVEN GOT MY DRESS!!!!!!!!! AND SHOES!!!!!!!!!!! it\'s Great! Anyway....before dinner I have to make an appearance at the Wright\'s house because Daddy Wright is expecting Paul and I to come for pictures. Then we will *hopefully* get picked up in the limo and go to The Butler\'s Pantry for dinner (we are getting a GREAT deal on that!). After that I suppose we will go to the dance out at UTSI. Nice little drive..but I don\'t mind as long as we have that limo. hehe. But even if we don\'t, it\'s all gravy. I don\'t know what we will do afterwards....hopefully Tori will ask her mom...but if not, oh well.   '),(13450,2645566,'14,March,2004','2004-03-14 00:00:00','   Spring break has started...Whoop...\'cept I\'ve been working! At least I am getting paid for it. Paul got back from Gatlinburg today and called me. That was nice. His little brother wants me to come to their house on Friday for cake because it\'s his birthday. Paul said that the downside would be that his grandparents would be there. I dont know how that is a downside, but then again, I only met them for a few minutes. I will haveta think about that one.   As for the whole friendship being screwy thing. I\'m putting it in the past and never thinking of it again. Its bringing up way too much mess and junk. I just haven\'t seen Ashley to be able to tell her that....*sigh* hopefully she\'ll read this and accept my lame apology.   '),(13451,2645566,'09,March,2004','2004-03-09 00:00:00','   In...out...in...out.   A nice, calming rythm. I find myself to be doing things like that a lot lately. I\'ve been so stressed. I feel like I\'m neglecting my friends (those of them that still like to spend time with me that is). I feel as if no matter what I do, I\'m failing my mother and my religion. I think I\'m going to fail Chemistry...and no one will help me with it. It\'s all just too much for my poor pitiful brain to take in. Yes, Poor, Pitiful.   I wish I could jump ahead to the summer. That would be nice. I want to do a lot of things this summer. Mostly just spend time with my friends and Paul. Just hang out. I guess that\'s what I\'m gonna be doing over spring break though. Laying around. Paul has school the first week and my friends will more than likely have other stuff to do. *sigh* this bites.     Okay, I have to get this off my chest. WHAT THE HELL ARE YOU SUPPOSED TO DO WHEN SOMEONE THAT YOU THOUGHT WAS DECENT TELLS YOU THAT YOU SHOULD GO KILL YOURSELF!?!?! How fucking rude right? I sure as hell thought so. Especially since I had JUST finished saying that I was having a HORRIBLE day!!!! \"why don\'t you just go kill yourself\"...grrr....Yeah, real mature there right?? I wish I could leave this place forever...or fast-forward life.   '),(13452,2645566,'30,May,2004','2004-05-30 00:00:00','   A lot has happened this past week. Paul broke up with me, got accepted to SCMI, spent a lot of time with friends, read a book, took exams, caught up with an old friend, got back together with paul, got ready for SCMI, etc. I want to elaborate on 3 of those topics.   1) Caught up with an old friend. At one point in time he was more than a friend, but that was a long time ago. I figured that it was only natural to want to spend sometime with him considering our past. The time finally came when Paul and I split. He decided to take me to lunch and cheer me up. I was fine with that. But something seemed off. He insisted on picking me up at my house and buying my lunch...then took me to his house. Urgh, while we were there we decided to watch a movie. He kept trying to hold my hand and asking me to go swimming with him in his pool. I would have been fine with just a visit and friendly chat over a light lunch. I wasn\'t planning on being hit on. HE HAS A GIRLFRIEND!! And at the time I was heartbroken. Goodness people!   2) Yes, Paul and I broke up. He needed time/space to think. I regretfully had no choice but to give it to him. But that\'s in the past. Now we are back together. We had talked for a long time on Wednesday night. Then tonight I went over to his house. It was like we were never apart. And now, we\'re officially back together. I couldn\'t be happier. I love him with all my heart and I just hope that he knows it.  3) Thirdly and finally I want to express how much I will miss some of the seniors this year. Last night was graduation and throughout the whole ceremony I was thinking how I will miss so many of them. Some more than others. There\'s one senior in particular who I love joking around with. I have so many memories with him from so many different occasions. It\'s easy being around him and I will miss him dearly. Then there\'s others who I have always expected to see around. I can\'t imagine how hard it will be to leave next year myself if it is half as bad as seeing these folks off.  I think that is it for now. Tomorrow after I sing in choir at church Tori and I are leaving for Cookeville. Time for SCMI! A whole week in Cookeville with no mother, brother, or evil step-father to pester me. And I must say, I CAN\'T WAIT!! Just kind of sucks that I have to leave the day after I get back together with Paul...go figure.   '),(13453,2645566,'22,May,2004','2004-05-22 00:00:00','   Okay...first off...It\'s been a helluva long time since I posted last. And the only thing that I can really think of that has happened is when I contracted mono. Ah...Mononucleosis...can\'t wait to get rid of that.   Anyway...I just want to say that I NEVER called you a fucking bitch, Tori. I\'m not that shallow for starters and it hurts to think that you thought I would go \"that low.\" Secondly, I\'m not waiting for you to apologize. I\'m not apologizing either since the only thing that I said was, \"Thanks...\" and walked off. You try not having a choice in whether or not you are *allowed* to go to school everyday. I WANT TO GO!!! I would love to be able to get out of the house before 1:30 everyday just so I can go to band. But frankly, I\'m not supposed to even be doing that. And the only reason that I was mad in the first place was because of the attitude that you used when you told me that I was \"going to fail all your [meaning my] classes\". That really boosts one\'s self-esteem. I was and am quite aware of the chances of me not passing. I am also aware of the fact that the STATE requires the school to give me incompletes for the semester until I have the opportunity to make it all up. I hope that you realize that I\'m not mad at you and that yes, I still want you to go to California with me. But this is bullshit.   '),(13454,2645566,'30,June,2004','2004-06-30 00:00:00','   So I haven\'t written in awhile, but I have good reason. Last week I was in Michigan. I have to say that it was definately a new experiance. The weather was great. Leaving Tennessee where the temp was around 97 with 100% humidity and going to Michigan where it\'s only 65...words can\'t describe how great it was. Okay, now I seem really dumb for talking aimlessly about the weather, but it\'s my favorite kind and well...I\'m dumb I guess.  Anyway, I was there to see Molly. I can\'t believe that she has lived up there for over a year now. The memory of her leaving seems so recent and I still hate to dredge it up. So on with the good stuff. I got to Detroit on Tuesday (I flew Northwestern...seemed like the flight was abandoned it was so empty, and it was really short in comparison to the California trips that I am used to making.) Becky, Molly\'s sister, picked me up at the airport (oh, this airport, yeah...it kicks ass. It has a freakin\'   train   in it!). Well, since I don\'t really feel like giving a day by day, I\'ll just say that I went shopping every day (ranging from garage sales, to Outlet Malls, to Old Navy, to Molly\'s drawers...lol). I came back with more stuff than I had ever expected to get while I was there. 2 purses, new underwear, a t-shirt, 3 pairs of jeans (all from Molly! How often do you get Banana Republic, Lucky Jeans, and Gap clothes from your friends??), a hat, a pair of flip flops, a keychain (courtesy of Molly\'s aunt), and earrings. I think that was it, but I\'m not guarantee-ing it.  Sunday night Molly, Becky, Emily (Becky\'s 8-month old daughter...she\'s SOO silly! Such a ham. hehe), and I drove back into Tennesee. We left Ann Arbor around 8:30 and got into Nashville around 4:30. So we made really good time. Molly and I went to Smyrna on Monday where my parents picked us up. We had Toot\'s for dinner which was definately an experiance as it was just Molly, my mom, and myself. From there we went around Big Lots...but then there was the Firewroks a.k.a Fireworks stand. Good stuff. Okay, so you won\'t get all of this, but oh well.   Moral of this entry: it\'s been a long time since I have written and in the mean-time I had a really great time with a really great friend.   '),(13455,2645566,'19,June,2004','2004-06-19 00:00:00','   I keep forgetting to write in this thing. And considering my last post, an update is in need. I apologize for to anyone who actually DOES read this and was confused by my last post. It was just a rambling that I needed to get out and I had no one to really talk to about it. Anyway...since the last real post, so much has happened. I got to go to Bonnaroo 2004 which was an experiance in itself. It was great because I got in for free, and I got to work which wasn\'t bad at all because it was soo much fun. The highlight of being there was I got to see my favorite band: Guster. Man, that was the best show ever. The energy was so high...and so were the people. haha.  Anyway...the only other thing that has really happened is babysitting. I HATE KIDS!! And I don\'t really have a choice in the matter. Yes, I am getting paid, but these kids are horrible! They never have anything good/nice to say, they don\'t behave, they don\'t do what they are told, and they are always talking back. I mean, how hateful can ya get guys?? It sucks even more that they are all related to me. Arrgh! It\'s okay though because it has made me realize that I never, EVER want to have children of my own. Paul backs me up 100% on that, so we don\'t have to worry about the whole kid thing. Woot.  On a final note...I want out of this house. Out of this city would be nice. And out of the state would be ideal. What I thought was going to be the best summer of my life is turning out to be pretty damn lame.   '),(13456,2645566,'07,June,2004','2004-06-07 00:00:00','   Has that ever happened to you? That one moment (or several in some cases) where you thing about the exact same thing as someone else. It could be simple and you both said the same thing, or it can be deep and personal. I don\'t know how many times that\'s happened with me...The unimportant ones at least. It always happens; most people call it coincidence. But what about the other kind. Those times when you \"happen\" to be on the same page with someone. Especially if they are close to you. I can say that I can count those times on one hand. That those times mean the most to me. That one of those times happened in a conversation last night. How can someone know what is deep in my heart and bring up the same question? How can someone understand me so well and yet not at all? It\'s all so puzzling. What I need is time to think...but it\'s hard to think when you have underlying questions. I need that one underlying question answered, but only one person can do so. Will the result be jeopardy? Am I willing to give up myself if I don\'t ask the question? Is it time to take the next step in my life? Should I even be questioning such a thing? All these questions and so little answers. I need to find the one piece of the puzzle...that\'s all I ask. Answer one question, find the one piece, and the rest falls into place.   '),(13457,2645566,'06,June,2004','2004-06-06 00:00:00','   Last night was amazing. Simple as that. And even though I won\'t go into detail, I need to say this. It\'s brought up questions. Questions inside me that I need answered...Hopefully they will be sooner rather than later. And hopefully, it will be amazing again.   '),(13458,2645566,'04,June,2004','2004-06-04 00:00:00','   \"What\'d they do to you up thur?\" That\'s definately the understatement of the year. Since last Saturday I was in Cookeville for Southeast Chamber Music Institute...so much happened! \"Oh gawd\" (as Russell would say) it was so much fun!! Apparently I am \"married\" to Jesse because we have a love-hate relationship. We even had a divorce while we were there (thanks to good ol\' Vern) and got back together...stupid Jess-dash-e claiming he got the car...Anyway it was good stuff. Lindsey Jo was awesome. So was \"I AM BETTY CROCKER BITCH\" and \"I CAN\'T READ!\"...let\'s not forget \"COMPTON!! WHAT?!? COMPTON! WHAT?!?\" and Russell\'s butt dance. Yeah, this stuff is going to take a lot of explainin\'...and it sucks that I have picked up on the whole country accent thing...oh well...It was good stuff Man...Good Stuff.   '),(13459,2645566,'13,July,2004','2004-07-13 00:00:00','   Today has been, in one word, horrible. And I really don\'t like that. So let\'s start from the beginning. I woke up with a phone call from a friend. No biggie there. I just wasn\'t really conscious of what I was saying, and that\'s not ever good. When I finally got my lazy butt out of bed I remembered that I had to get my pictures taken today. Now how can someone forget that? I don\'t know. So I did some wash. Oh, I should probably mention that I was babysitting. So, I\'m doing wash, watching kids, and making sure they don\'t make a mess while they make lunch. Then I have to start getting ready to leave. The little heathens decided that since I was a little bit preoccupied that they could do whatever the heck they wanted. BIG MISTAKE ON THEIR PART. So now I\'m doing wash, watching kids, making sure they don\'t make a mess (in general), trying to get ready for my 2:00 appointment for senior pictures in Manchester, and trying to control the mongrouls. By the time that I was FINALLY ready to leave I had to turn around and come back because I forgot my clarinet. Then I got caught at a train. Here comes the part that I am   so   happy about. My freakin car broke down again! By the time that I finally got to my grandparents house, my car had died twice. And I still had to go to Manchester! It died two more times before I even got out of Tullahoma. It dies when I go fast, it dies when I go (period), and it now dies when I slow down!! I have to say that the best part of my day was when I was getting the pictures taken (which took a whopping 2 and a half hours)...except the fact that my mom didn\'t come like she was supposed to. Then I came home. My car *only* died 2 times by the time that I got home. It\'s been a long day. My car won\'t get looked at until Thursday at the soonest, and band camp is on Friday. Oh poo...I just found out that I\'m going to have *another* kid to watch tomorrow. I think I need a break.  God Bless.   '),(13460,2645566,'12,July,2004','2004-07-12 00:00:00','   I get my senior portraits taken tomorrow. Formal and informal. It\'s kinda freaking me out because it means that I will really be out of the high school next year. Sudden realization sucks. Anyway, I got a really cute new outfit tonight for some of the pictures. I\'m nervous and excited. More on that later...  God bless.   '),(13461,2645566,'09,July,2004','2004-07-09 00:00:00','   So now I feel really bad about the whole \"hating on mothers\" thing. When my mom got home from work today she told be about how bad everything has been. She stopped taking her anti-depressants a month ago; they cost too much. She had a breakdown today. Called the doctor crying. Had to leave work to go to the doctor. She said that the only thing that has kept her alive this week (literally) is me. She said that she would have killed herself this week if she would have known that I had someone in life. She realizes that I don\'t. And I really mean   someone  . Jordan, my brother, has his father (that I can\'t stand), grandparents, and everyone here. The only person that I have, besides her, is Paul...and frankly, I couldn\'t go live with him if something happened. I would be sent away to California and that is the reason that she didn\'t do something drastic. She didn\'t want me to have to give up my life here. Now I am living in a land of guilt. How can I be so selfish? I vow to take into consideration the way others are feeling and what they may be going through. I ask for help in keeping this vow. So all I have left to say is...sorry.  God Bless.   '),(13462,2645566,'03,July,2004','2004-07-03 00:00:00','   So at this point in time, I think I just need to type. There has been so much going on lately that all of my emotions are completely out of line.  So how about scenario #1:  We were supposed to leave for Gatlinburg this morning. Well, since Monday my mother hasn\'t been able to decide whether or not we are going. One day the answer was \'yes\' and the next the answer was \'no\'. There was a problem with this. On the second day there, we were supposed to go to the water park. Well, I don\'t own a bathing suit!!! For goodness\' sake I know that that is odd (but in all honestly, mine from last year is too big). That was the only reason that I kept asking her if we were going. I needed to buy a bathing suit if we were! Oh!! And it didn\'t really help that everytime she was on the phone she was talking to the family that we were supposed to go with! She would sit there and tell Marsha over and over that we were going, then get off and tell me no. What the heck people?! And then last evening she got really mad at me because I had asked her again while she was on the phone with Marsha. Can you blame me? I would kind of like to know what is going on in my life. Arrrr...  Scenario #2: Last night I\'m sitting on the couch reading a book and all of a sudden the phone rings. Nothing abnormal about that. The nature of the call, on the other hand, could have gone over a LOT smoother. You see, it was a babysitting call...for me. But I didn\'t get to talk to the lady on the phone. My mom just volunteered me for the job without even asking me in the first place. What if I had plans for today? Which I did. Well it obviously didn\'t matter because I had to babysit. And while we are on that topic, I\'m going to elaborate. The kids that I have no choice but to watch today, I have watched before. I had no problem with watching them then. My sudden change of heart was due to the fact that their mother never came to pay me. So she has owed me $50 for over a month now. Why should I do her a favor and watch her kids if she still hasn\'t paid me for the last time? I just don\'t understand that.  Scenario #3: I have to say that this one is very different from the others. This one makes me sad. Things are great with Paul. Getting better with every day that I see him actually. On Wednesday night I went to his house after my senior portrait meeting. That was so much fun. The power went out and we went to \"investigate\". We sat on his bed and talked for a really long time about nothing in particular. It was really nice. Then yesterday I went over to his house again. We watched a movie and played with his little brother. Once again, it was nice. The whole leaving thing got me down though. And I know how pathetic that is. But he\'s leaving for vacation sometime today, and I\'m going to miss him. I don\'t even know if I\'m going to talk to him while he\'s gone...and he\'s going to be gone until the 19th. Then I have band camp. So in all honesty, I don\'t know when the next time I will see him will be.   Scenario #4: I AM REALLY EXCITED ABOUT BAND CAMP THIS YEAR!! Not only am I a senior now, but our show is going to be really awesome! Okay, so our theme is pirates (Arrr matey..hehe) and we are doing songs from Pirates of the Caribbean, Song of the Sea Maidens, and something from DCI (Drum Corps International). I can not wait!   And now that I have said everything that I can think of at the moment, I am going to end this post. Sorry if you actually read this...I know that it\'s almost pointless. I just had to say some of this stuff before I burst. I feel much better now!   God Bless!   '),(13463,2645566,'03,July,2004','2004-07-03 00:00:00','   Just in case anyone wants to know...my mother really pisses me off. I don\'t really care to go into detail because I don\'t want her to find this and get all pissy with me. But the fact is she is infuriating. Yesterday she was trying to be all buddy buddy (because of something that happened earlier in the day that I don\'t think she realizes I know ALL about). Today she is being completely stand-offish. Not only that but when I try to   talk   to her she closes up and the only thing she says is \"mmhmm\" and \"oh\". Great conversation eh? And the last weekened that I was here she was being super nice. \"Mother-daugher this\" and \"Mother-daughter that\". Why can\'t she be NORMAL sometimes?? And it really bites because I always hear \"your mom is so cool\" and \"your mom kicks ass\" and all that junk because she   isn\'t   a conventional mother. But every once and awhile it might be nice to know that she is a tad bit more mature than I am.   '),(13464,3471722,'27,May,2004','2004-05-27 00:00:00','    I\'m in eighth grade i\'m 14 and am graduating june 7. i\'m in love with a boy named antwaun and he is in love with me. we aren\'t exactly in a relationship yet but,it will be official(i\'m pretty sure) this weekend. i trust him so much and he respects me. my life at home is just okay. my parents aren\'t together and i dont get along with my mom. She is mean to me. I hate her so much that is why this summer i get to move in with my dad. usually i go to his house every other week but now i get to move in full time after i graduate.this is my first blog an di dont know i fi will write another. i probably will. i\'m Italian and i love Italian food, especially the desserts. i went on a cruise with my aunt and my cousin jen. i got to kiss and swim with a dolphin and i had so much fun. it made me feel good about myself i dont know why. It was like i had no worries and it just felt liberating. My family is big. i have ten aunts and uncles just on my dad\'s side of the family. i love my dad\'s side of the family it is so much fun and we always get together for dinners and parties, and everyone eats and drinks and has a good time. i go to private school now but when i get to high i will be in a public school. private school is ok but i can\'t wait to experience public. i have to wear a plad jumper and i have been in this same small school all of my lif. i have never gone to school without a uniform, except for jean days which dont happen often. i am short have light brown hair brown eyes and i weigh around 100 lbs. i love basketball and was already invited to play varsity basketball my freshmen year of high school. i love shoppin\' and i always like looking good. i like getting my nails done but usually can\'t afford it. i hope you lik emy blog and now i\'m an official bloger.  p.s i just learned about blogging and how it is addicting...for whoever is addicted to blogging right now think if you have neglected a family member or friends or anyone who is special to you, and get off your computer now and spend some of your time with them. In my researches i read that 2.5 million ppl are addicted to bloggin.   '),(13465,4174428,'22,August,2004','2004-08-22 00:00:00','   \"mocking the world as it beats our ass.\"  Josh and I are walking home from a perfectly splendid evening. In front of us are some boisterous boys talking about their evening\'s exploits, and one of them comes out with \"she totally wanted to give you a blow job.\" Josh, walking only a few feet back, answers as only Josh can, \"no she didn\'t.\"  They were decent sized guys, not big, but bigger than Josh and I, and there were four of them. Luckily, only the one behind looked back to see us snickering, and I don\'t think he heard what Josh had said. '),(13466,4174428,'21,August,2004','2004-08-21 00:00:00','   urlLink This  is by far one of the coolest things I\'ve seen in a long time. I suppose that only Josh will care, but I just found a ton of bloggers that live in my neighborhood. One of them linked to  urlLink pictures from a target opening  that happened in my neighborhood. I\'m pretty sure that\'s Lizzy Grubman there, but can anyone recognize some of the others?  Is there anything similar for Boston or TN? '),(13467,4174428,'18,August,2004','2004-08-18 00:00:00','   Two movies that caught my eye recently, neither of which is probably safe for work.  urlLink The first  is about the difference between the internet and real life, and  urlLink the second  is about the Republicans coming to New York. '),(13468,4174428,'17,August,2004','2004-08-17 00:00:00','   Fox is running 2 hours of Family Guy in prime time Wednesday night. This means something is coming. Or it\'s just competition for the Olympics. Well, that was a quick emotional roller coaster. '),(13469,4174428,'16,August,2004','2004-08-16 00:00:00','   I don\'t know if anyone else used to play games on an old Apple IIE or IIGS, but  urlLink Peasant\'s Quest  really took me back.  In case you guys don\'t know the background, this was made by the guys who created  urlLink Homestar Runner  and are responsible for weekly  urlLink Strong Bad E-mails  which are pretty entertaining. '),(13470,4174428,'16,August,2004','2004-08-16 00:00:00','   Do you realize that you all left a full, unopened box of Nerds here?  This is the best thing that\'s happened to me all summer. '),(13471,4174428,'14,August,2004','2004-08-14 00:00:00','   and she\'s bringing him to New York.  I mentioned that she was visiting in a comment below, but forgot to include the boy. For anyone who hasn\'t seen it yet, my new place is a studio with one queen sized bed, one couch, a little bit of space on the floor and that\'s pretty much it in terms of sleeping space.  Me. Julie. Ansel.  There really is a lot to be said, but I think I\'ll leave it alone. '),(13472,4174428,'14,August,2004','2004-08-14 00:00:00','   so i got into a car accident last night.  just saw harold and kumar with tina and bhaumik and we were on our way back from the theater around midnight when the fool in front of me decided that upon seeing the light change from *green to yellow* (not yellow to red, mind you), he would slam on his brakes despite the fact that we was **already in the middle of the intersection**.  of course, having seen the yellow light and having realized that i was definitely within range of making it through before it turned red, i kept going at full speed.  didn\'t quite have enough time to come to a halt after i slammed on my own brakes.  luckily we were all fine--wearing seatbelts, airbags deployed, etc. (have you ever smelled airbags?  it\'s awful and i\'m pretty sure i\'m going to die of lung cancer from all that crap in inhaled).  anyway, we were ok, but the front end of my car isn\'t looking so good.  thank god i have insurance.  anyway, i hate driving in this city.  to end on a less serious note--i have a crapload of essays that i need to edit for my job so i decided to stay in tonight and do that.  what have i done all day since tina and bhaumik left for home?  read that damned blog!!  thanks for helping me out srinivas--i had, of course, already read all the comics because i\'m obsessed with this site.  what will i do when i finally get caught up?  oh yeah--all the stuff i\'m supposed to be doing instead.  4 days to go until school starts... '),(13473,4174428,'13,August,2004','2004-08-13 00:00:00','   I\'m pretty sure the blogspot ads at the top of your blogspot blog is supposed to be tailored to the content of your blog in some fashion.  What does it say about us that the ad I\'m currently looking at (in another window, of course) says, \"Help us cure  urlLink Crohn\'s  and  urlLink Ulcerative Colitis .  Learn about these diseases.\"  Maybe Coco had ulcerative colitis.  Maybe they know about UHS\'s incorrect diagnosis of my chronic appendicitis as a possible ulcer.  But it\'s probably just Josh\'s crazy med school talk. '),(13474,4174428,'12,August,2004','2004-08-12 00:00:00','   I counted four gray hairs in the mirror just now.  Must be aging and not stress since I don\'t feel very stressed right now.  But maybe they sprouted during spring finals, and I\'m just noticing them.   Going gray isn\'t so bad.  It\'s distinguished.  Last month, a cab driver told me I look 18 or 19.  Maybe if I had gray hair, he would\'ve treated me with the respect I deserve. '),(13475,4174428,'12,August,2004','2004-08-12 00:00:00','   Wow, this really is turning into the extended slow motion version of our room. '),(13476,4174428,'12,August,2004','2004-08-12 00:00:00','   ok, so it\'s like 5 am.  i am *still* reading this girl\'s blog ( urlLink http://theunderweardrawer.homestead.com ).  in the 4th year now.  late 2002.  just gotta get to 2003 and then i\'ll sleep.  i swear.  thought i would share this funny little story:  \"All of a sudden, I woke up out of sleep with a start.Â  Some noise, or some subtle motion in my body, had set me awake.Â  I wasn\'t quite sure what it was.Â  In a semi-stuporous daze, I tried to figure out what it had been.Â  And then I realized. I had farted.Â  I had farted, and the sound combined with the vibration had shaken me awake.  I HAD FALLEN ASLEEP IN THE MIDDLE OF GRAND ROUNDS AND FARTED MYSELF AWAKE IN FRONT OF THE ENTIRE CARDIOLOGY DEPARTMENT.  Now let us never speak of this again.\"  ha!  oh that michelle...what will she do next?? '),(13477,4174428,'12,August,2004','2004-08-12 00:00:00','   Hey Sal,  We\'re going to start calling you Coco the Wonder Monkey from now on, \'k? The relevant Seinfeld episode popped into my head for some reason, and the phrase just sort of came together. I realized I had to assign it to someone, and if not you, then who? If you think about it, it\'s sort of a compliment. You know, \"Wonder,\" etc., and who doesn\'t love a monkey, am I right?   In the interest of fairness, I\'ll let you decide how you want \"Wonder\" and \"Monkey\" spelled. \"Coco\" is non-negotiable.   Peace out, Coco. '),(13478,4174428,'12,August,2004','2004-08-12 00:00:00','   just saw silence of the lambs again.  hannibal is totally going to eat my liver with some fava beans and a nice chianti while i sleep.  sorry to show up fashionably late to this thing, but you know, i\'m a fashionable guy.  shankar--hope you didn\'t die in the hailstorm. sal--tina\'s coming up today--we\'ll probably be hanging out in the city--you should come. srinivas--um, i hate you and your memphis weather...(torrential rains in the city today...ruined all my plans).  later gentlemen.  i mean assholes. '),(13479,4174428,'11,August,2004','2004-08-11 00:00:00','   If you\'re looking for something besides the  urlLink official kerry gear,  I\'m going to link to the two best Kerry shirt sites that I\'ve found:  urlLink one  and  urlLink two.  There are plenty of others, but these were my favorites.  I\'m planning on getting one or two if I can afford it when I get paid. '),(13480,4174428,'10,August,2004','2004-08-10 00:00:00','   I think my windows might cave in. '),(13481,4174428,'10,August,2004','2004-08-10 00:00:00','   SAL! Check your email! '),(13482,4174428,'10,August,2004','2004-08-10 00:00:00','   This is an open letter asking that you please accept my sincerest invitation to become a member of the blog that I started called QuincyB21 (reference  urlLink here ) . Shankar and Srinivas have already accepted my invitation and summarily been made administrative users. We humbly request your presence and await your reply.  Faithfully yours, Sal  p.s. You know we\'re going to talk about you, right? I mean, come defend yourself, man. '),(13483,4174428,'10,August,2004','2004-08-10 00:00:00','   Hey asshole, aren\'t you going to make us all administrator level users? '),(13484,4174428,'10,August,2004','2004-08-10 00:00:00','   So after a weekend of very college-like activities, I decided to actually start the blog that we mentioned in passing. I\'m positive that Shankar and Srinivas know more about this than I, so let me know if we\'d like to get a different site, change the look, etc.  I imagine we\'ll delete this first post, but let me know what you guys think.  Also, how do we get rid of the blogspot link at the top? Do you guys get a link button on your page (I think that\'s a problem with macs)? '),(13485,4185342,'19,August,2004','2004-08-19 00:00:00','   My laundry machine kept flooding my bathroom, and when Ben took a look, he said that when my bathtub was replaced and the machine got moved around, the exhaust hose was put in the wrong hole, but that a piece had been removed so it would fit and I would need a new hose.  I mentioned this to my supervisor (who had done the fixing) and he insisted on coming to my home to look.  I told him it was okay, but I have a feeling that he and Mastui, who tagged along, were eager to get away from the office for a bit during the day.  It was pretty funny to see them both trying to figure out how to fix my little washing machine, with one tilting it and the other, kakhis rolled above the ankles, sqatting below to see what was happening.    They told me to put in some laundry as a test run.  I did that, and when they decided it was working, they said that it was time to go back to the office.  Of course, this was mid-laundry cycle.  I did my best to waylay them (I gave them tea and chocolate), but there was no way I could stay and wait until the end of the cycle.  My clean clothes were doomed to wrinkles and mildew.  I asked Matsui if he had ever done laundry before, and he said he had, but informed me that really, you don\'t wash your own clothes, the machine does it for you.  (Ba-dum ching.)  A note on Matsui: Matsui, to me, is hilarious.  I told him that I didn\'t have much experience cooking but that I planned to make stir fry (and had to explain what that was) his response was, \"Oh, wow, you really can\'t cook, can you!\"  He then proceded to guide me to some leeks, saying that they were better than scallions because they were the same thing but bigger.  (I assumed he was right, until I ate it and recognized the flavor - it did not taste good in the stir fry.)  He also thought I bought too much garlic, even though I explained that it doesn\'t really go bad and that I planned to use it pretty much every time I cooked.  He\'s married, and I have heard from the usual (unreliable) sources that he\'s probably never done any housework in his life.  '),(13486,4185342,'17,August,2004','2004-08-17 00:00:00','   I\'ve been printing out the New York Times\' \"Metropolitan Diary\" on a weekly basis in hopes of presenting some of the blurbs to my English Conversation group one day.   This is from the August 18th Metropolitan Diary:   Dear Diary:  One recent afternoon, I was waiting in line at the silver counter at Tiffany. A woman ahead of me had just purchased a bracelet and was filling out a gift card. She looked up and asked the salesclerk, \"How do you spell \'bar mitzvah\'?\" The salesclerk didn\'t hear her. I intervened. \"Bar mitzvah?\" I asked.  She smiled and nodded. \"Didn\'t you buy a bracelet?\" I asked.  \"Why, yes I did,\" she answered.  \"So it\'s for a girl?\" \"That\'s correct,\" she said. I explained: \"Well, bar mitzvah is for a boy. Bas mitzvah is for a girl. So you should say \'Happy bas mitzvah.\' \" She thanked me, then I asked, \"Do you know if they are Sephardic or Ashkenazic?\" Her face dropped. \"Oh my, I have no idea. Does it matter?\" she asked. I replied: \"No, not for the purpose of a gift. But if they are Ashkenazic, it\'s bas mitzvah, Sephardic is bat mitzvah.\" \"So how do I spell it?\" she asked. I told her. She smiled and said: \"I\'m visiting from Milwaukee. Thank you for all this information, it\'s so interesting.\" She looked a bit sheepish and said, \"I don\'t know any of this; I\'m a Catholic.\" I said, \"So am I.\" Surprised, she asked, \"My goodness, how do you know all this information?\" I responded matter-of-factly, \"I live here.\"   -Brian Honan   I\'ve often tried to explain this phenomenon to non-New Yorkers, and sometimes to Japanese people, who seem to only half believe me.  Ironically, though, it has relevance here.  My predecessor, Ben, was Jewish, and in his monthly column in the town newsletter, he often introduced Jewish holidays and customs.  Many people in the town smile enthusiastically when I mention the now familiar words \"Rosh Hashannah\" or \"Hannukah.\"  Even funnier, though, is that Ben and I are both living in Shitara, we figure that out of 5,000 people, we in Shitara might very well have the highest concentration of Jews in Japan.  '),(13487,4185342,'15,August,2004','2004-08-15 00:00:00','   I was dying for some basil, but there was none to be had, either fresh or dried.  I decided that, despite its being the end of the summer, I would plant some.    At the first store I went to looking for seeds (they had dirt and planters out front), there were no seeds, but a friendly old obaasan (granny) who introduced me to her daughter, who was a modern artist with an upcoming exhibition in Nagoya.  She showed me her work, and she told me how her daughter lived in London, so we were able to speak some English.  Who woulda thunk that there was a house full of award-winning abstract art down the street from me?  Basil was a bust, but at least I got a dose of culture.  I gave up on basil for the day, and went to meet my friend (and predecessor) Ben.  When I met him, he was having a conversation with a very petite obaasan who was talking a mile a minute.  I introduced myself, and she bantered on without the usual pauses to make sure that my foreign mind was absorbing it all.  She went on about how it was really too expensive to bring presents all of her friends in nursing homes, or even to get to the ones in Hokkaido, but that now was the O-bon holiday, and all the dead relatives were going to come in to visit tonight anyway.  Then, very abruptly, she started talking about how foreigners eat a lot of meat, but it makes them tall.  \"And, I watch those movies where the foreigners kiss, and it\'s so wonderful!  The foreigners definitely have us Japanese beat in the kissing department!\"  That is exactly what she said.   Back to the great basil hunt - yesterday morning, Ozzie neighbor Hamish and I walked through the entire downtown, but not even a suggestion of basil seeds.  Finally we were walking past a florist near my work, and Hamish pointed to the plants lined up outside.  \"These are all basil!  Why plant when you can get a fresh pot!\"  I walked into the unmanned, darkened store.  The inside seemed to be some sort of stationary store.  \"Hello!  Excuse me?\"    Finally, a middle aged woman appeared from the back.  I told her I wanted to buy one of her basil plants.  She looked shocked, then laughed, \"Oh, those aren\'t for sale.  But I can give you one.\"    I tried to refuse, but she picked out a plant that had BOTH basil and the Japanese green, shiso.  I was rather embarrassed, but now I have my own, very much alive and fresh, basil plant.  '),(13488,4185342,'10,August,2004','2004-08-10 00:00:00','   urlLink          urlLink DSCF0102      Originally uploaded by  urlLink Yage Danchi .'),(13489,4185342,'10,August,2004','2004-08-10 00:00:00','   urlLink          urlLink DSCF0004      Originally uploaded by  urlLink Yage Danchi .'),(13490,4185342,'10,August,2004','2004-08-10 00:00:00','   urlLink          urlLink At the Nanaharas\'      Originally uploaded by  urlLink Yage Danchi .'),(13491,4185342,'10,August,2004','2004-08-10 00:00:00','   Well folks, I\'m finally settled and ready to send all the news I can.  Not totally sure about this blog thing, but hey, I\'m still waking up at 5 am every day, so might as well be productive... I\'m living in a small town in a region I\'d though no one had ever heard of, but apparently it was made famous by a photographer named Shinzo Maeda.  Most of the \"town\" is literally in the mountains - houses along windy roads with no guard rails, terraced rice fields and tea and veggie farms - you name it.  Absolutely gorgeous.  When I first got here, I was staying in the home of one of the town\'s muck-a-mucks. Now, I live in low-income housing in the downtown.  Downtown is called \"Taguchi,\" and the danchi (apartment building or complex) is called Yage Danchi. I\'d better get ready for work.  I finally, after almost a week, have my very own shower.  More pictures (and explanaing) to come!  '),(13492,4127857,'14,August,2004','2004-08-14 00:00:00','   ok, so last night i went to the Lance Armstrong concert downtown with Jessee, Sadie, Zsofie, Painter, Amar, Brandon, Jake, and some of their parents...and it was AWESOME!!!! omg i want to marry lance...he is my hero!!  and Steve Miller sang! yea he did!! ABRACADABRA!!!!  but yea, ya\'ll should deff. wear your LiveStrong bracelets if you have one...b/c they don\'t just represent lance...they represent all of those who have survived cancer or are going through it right now.  and it shows that dreams do come true...i mean, hello! look at lance!! And dancing in the middle of the streets last night was so cool! everyone was watching us b/c they wish they were us! yea they did! and Brandon and Amar were fighting over me b/c they love me soo much haha..but what could i do? i already have a boyfriend...but i know i always have options.  :) haha ANYWAY, after the Lance thing, we all ate at Whataburger (well, everyone but me) and then we went bowling at like 1 in the morning and i got a strike!! my first ever!!  it was awesome..and then i didn\'t get home until like 2:30 am...but thats ok, b/c the concert was AWESOME!!!!! omg...this week has been awesome!!  So today,  i went by the school to see if i could talk to my choir teacher b/c she messed up my schedule..but she wasn\'t there...so i went home and my parents and i went to Highland Mall for Best New Talent auditions, but i didn\'t actually audition b/c i went last year to California and so we were just wondering if we could skip this year since i was invited back, and come next year.  (we don\'t know how my mom will be feeling by the time the real auditions in Cali. happen.) so yea, they said yes but they will work with us so i can come this year and they asked me to come to the second audition at the Sheraton Hotel tomorrow and talk to everyone and tell them about my experiences in LA! yea, totally telling them i met AJ Trauth of Even Stevens!!! yea, and then we went shopping at J.Crew and i bought a pair of really cute jeans!! they fit perfectly!! i really love them..and then we went to Barton Creek Mall to shop for some un-mentionables and then home, and now i\'m visiting the hotel and writing in my blog! yay! haha but yea, now it is time for me to leave...i have to get ready for tomorrow b/c its my last day of summer! ahh!!!  i\'m sooo not ready for school!  If i don\'t write in my blog tomorrow,  GOOD LUCK ON EVERYONE\'S FIRST DAY OF SCHOOL!!! or whenever you start!!!  Monica.  '),(13493,4127857,'12,August,2004','2004-08-12 00:00:00','   i forgot one more thing!  if you live in austin, go to Hills Cafe on South Congress Wednesday nights!! i\'ll more than likely be there at the KVET booth, hopefully with the spin wheel!  and you\'ll meet SCUBA STEVE!! (aka-Steve-O) and Matt!!! woohoo!!!!  '),(13494,4127857,'09,August,2004','2004-08-09 00:00:00','   Hey! ok, so yesterday was the best day EVER!!!! ok, so i\'ll start from the very beginning.....  I woke up early to get ready and go meet Tim and we met up at the mall and then i took him to a cheap place for him to get his hair cut (he looks so good now!) and then we went and saw The Village.  Wow, thats a creepy movie.. and then we went walking around the mall and ran into a friend of mine and her mother and they met him and thought he was very cute (funny, i totally agree with them!) and then he had to leave :(  but then, Jessee invited me to go to Hills Cafe with her and her family and i\'m sooo glad i did!!!! ok...here is the second part of my day...  We get to Hills, and Jessee starts flirting with the two guys that work with KVET at the \'free stuff booth\' (it was only free because Steve was a genius and lost the spin wheel!) and by the way, Django Walker was singing there last night and if you don\'t know who he is, Jerry Jeff Walker is his Dad.  and if you don\'t know who he is then, YOU SUCK!!!! b/c JJW is AWESOME! but yea, i heard that Django went to LIPA University in Liverpool, England, where i am wanting to go so i told the guys at KVET and they pushed me to go talk to him...and i did...and i\'m glad i did! so this is the convo with Django: Me: did you go to LIPA Univ.? Django: yea i did M: was it like, an awesome experience? D: omg yes, their vocal teachers are amazing...it was awesome M: ok, cuz i\'m looking to apply there D: oh really? what are you going for? M: singing D: oh really? M: yea, i\'m a singer D: thats awesome M: yea, is it hard to get in? D: no not at all..as long as you have the pipes, you\'re in! M: oh, thats great! D: yea, i\'ll tell you what..email me and i\'ll tell the guys there to keep a heads up on your audition tape M: omg, thank you! D: yea no problem, enjoy your night M: thanks yea, you too...good luck up there!  OMG YES!!! HE\'S WATCHIN OUT FOR ME!!!!  THANK YOU DJANGO!!!!!!!   ok...so instead of watching the concert, me, jessee, her sister, and their foreign exchange student were all hanging out with KVET and got really cool FREE stuff!  and they all call me \"Carry Around Portable Monica\"...yea...and i un-oficially sing the KVET jingle! \"Its the genuine austin original...98.1..KVET!\" lol oh and yea...if you are from austin and listen to 96.7 KISS FM, the week after next you have to listen weeknights from 10 pm-5 am and sunday nights from midnight-6am!! b/c Steve-O is the DJ and he\'ll be filling in for someone for 2 weeks! and he\'s gonna have me on his show and i\'m going to sing live on the radio!! woop! woop! but yea...it was great...KVET IS AWESOME AND THEY ARE MY NEW BEST FRIENDS!!!!!!!!!!! (i\'ll have pics of Steve and Matt, the 2 guys we were hanging out with, posted soon!)  HAVE A GREAT NIGHT YA\'LL!!!  '),(13495,4127857,'09,August,2004','2004-08-09 00:00:00','   On Friday i got to go shopping with my boyfriend, Tim, and i helped him pick out all his clothes!  it was so much fun..he even bought a couple of Polo shirts because i like them! (he likes them too...but its more fun acting like he bought them for me, hehe!)  He\'s a really great guy.  he makes me laugh constantly...But anyway...yea so this past weekend was Tax Free Shopping...yea, totally bought ONE thing!! go figure right? i\'m so disappointed in myself.  Oh well, now there are going to be MAJOR sales!! woohoo!! (i\'m all about sales) ANYWAY... i don\'t really have anything else to say.  feel free to comment!  Have a great day!  '),(13496,4127857,'04,August,2004','2004-08-04 00:00:00','   hey ya\'ll today was pretty interesting day..i was woken up around 9 and then fell back asleep and then woken up again around 12:30 by my mom b/c she got a call saying that she needed to go to LoneStar Oncology and have blood drawn for some tests to see if she needs a bone marrow transplant (for those of you who don\'t know, my mom was diagnosed with Non-Hodgkins Lymphoma Cancer stage 4- in December of \'03 and has gone through 6 months of heavy doses of chemotherapy...she\'s in remission now! but is scared to death of the bone marrow treatment...its a life or death situation..and i think she\'s not going to go through with it) ANYWAY after that she, my sister, and i went to lunch at Schlotzkey\'s (spell check?) and then to my sister\'s hair appointment b/c her hair was way too long for her..and the hair stylist was so nice that she even trimmed my mom\'s wig for her! free of charge! (it had kinda looked like a mullet and needed to be cleaned up) and then we went to the mall and did some shopping then to barnes and nobles and then finally, home.  But yea...that is what has happend so far..oh yea! i\'m getting ready to perform in San Antonio (i\'m a singer) and i\'m really excited! i think that\'ll be happening at the end of August..ok.  thats all i have to say.  Have a great day ya\'ll!  '),(13497,4127857,'03,August,2004','2004-08-03 00:00:00','   Hey thanks for visiting my blog..Kudos to Sabrina for helping me start this up! Have a great day Ya\'ll! Bye!  '),(13498,3828283,'29,July,2004','2004-07-29 00:00:00','   iPodlounge has reported that some 4th Generation iPods have  audible static and noise interference in the earphones that is most prominent whenever a new iPod\'s hard disk is accessed. Similar interference was not detectable through line out (Dock Connector) output iPodlounge reports.    Users of affected iPods will be able to hear a hard disk-like whirring sound  in their earphones , coupled with several seconds of light static at the start of a song that has just been loaded. This should not be confused with the quiet hard disk loading sounds that an iPod makes, which sounds are not audible through earphones, or with normal static-like compression artifacts in your audio.  The units known to be affected were manufactured in China, shipped from Shanghai, and ordered directly from Apple. If you have a new 4G iPod read the  full article  and the  update article  to learn how to test to see if your iPod has the malfunction. '),(13499,3828283,'19,July,2004','2004-07-19 00:00:00','   urlLink    Here it is, The 4th Generation iPod. The new iPod looks a bit different, operates more efficiently, has a few more features and costs less. Here are the highlights.   Design.  This one is about a millimeter thinner and, more significantly, eliminates the control buttons that sat under the display screen. Instead, it uses a \"click wheel,\" where the controls are placed on the compass points of the circular touchpad that lets you scroll through menus. This is an innovation carried over from the diminutive iPod Mini.   More efficient menus . There\'s less thumbing required to get to your favorite stuff. \"Music\" is a first-level entry, and now a single click initiates the popular technique of shuffling your library for playback.New features. You can create multiple on-the-go playlists and delete songs from those ad hoc mixes. And audiobooks are not only easier to find, you can listen to them at normal speed, slower or 25 percent faster, without its sounding like a Munchkin.   Longer play.  Coast-to-coasters rejoice: the new iPods are rated for 12 hours of rockin\' between charges?a 50 percent boost in battery life. This is accomplished, Apple says, not by a heavier battery but diligent conservation of power.   Lower price.  The top-of-the-line iPod, holding 10,000 songs (40 gigs, as geeks will tell you), now costs $399. The lower-capacity model, with room for 5,000 songs (20 gigs), costs $299. That\'s a $100 price reduction for each. (There\'s no more 15-gig model.)   Color.  Fuggedaboutit. Despite rumors to the contrary, the wide-bodies are still as pure as the driven snow.   Information Provided By Newsweek. Pick up you July 26 edition on shelfs Monday July 19, 2004. Or read the full story   urlLink here . '),(13500,3828283,'17,July,2004','2004-07-17 00:00:00','   urlLink    Newsweek Cover for July 26 Edition which will be shipped on Sunday and will be on shelfs Monday and subscribers will recieve it on Tuesday. This confirms the new iPod release on Monday but as you see there is no color screen. '),(13501,3828283,'14,July,2004','2004-07-14 00:00:00','   I got a gmail account. I am so happy, the account has 1 gig of storage!  It\'s still in beta but it is a great email client or so I am told. I  will tell you what I think once I have some experience with it. '),(13502,3828283,'14,July,2004','2004-07-14 00:00:00','   ThinkSecret has reported that new white iPods will be released on  Monday, July 19, 2004. They are reported to have a click wheel like the  iPod mini. Best of all they are reportedly going to be cheaper!! I just  hope they have color screens like previously rumored. Read the article  urlLink here . '),(13503,3828283,'14,July,2004','2004-07-14 00:00:00','   Via webcast conference call Apple stated its financial earnings today over fiscal 2004 third quarter ended June 26, 2004    Highest Q3 in 8 years      - $61 Million profit on revenue of $2 Billion      - Highest quartely shipment of macs      - iBook 240K, iMac 220K, 416K desktops      - Apple Retail Stores now total 80 with $270 million revenue      - Cash up to $4.96 billion. Apple is a debt-free company      - 2004 total revenue should exceed US$8 billion      - iMac G5 CONFIRMED by Apple CFO Peter Oppenheimer ( urlLink Previously Rumored )      - 860,000 iPods sold in Q2 - up 183% from same period last year      - Demand on iPod mini expected to \"visibly outstrip supply\" through end of the year      - iTunes Music Store generated a small profit      - hPod: HP will ship \"this summer\"      - Office Depot authorized to sell online. No plan to extend to retail stores      - Best Buy and Apple work well with iPod sales. No plans to increase working together on CPU sales      - Xserve G5s sold 13K units - a quarterly record for Apple      - As Euro iTunes Music Store launched, Euro iPod sales increased as well      - Over 13,000 retail outlets for iPod. More people want to sell it than Apple has supply to provide   Information Provided By Mac Rumors '),(13504,3828283,'12,July,2004','2004-07-12 00:00:00','   Apple has released a list of the winners who bought each 100th thousand song from 95 million to 100 million and the grand prize winner who purchased the 100th million song on iTunes. Check out  urlLink Apple\' list  to see if you are a winner. I already checked it and I am not a winner so I\'m not to happy but I congradulate all the winners. '),(13505,3828283,'11,July,2004','2004-07-11 00:00:00','   100 million legally download songs have been purchased and downloaded on Apple iTunes. Apple will have a list of the winners in a few days so just wait it out and pray you were one of the winners. Although this is all great I have been submitting every time it gets close to every hundred thousand song and I had been submitting but I didn\'t check the the number of songs downloaded counter and totally missed the 100 million song. '),(13506,3828283,'08,July,2004','2004-07-08 00:00:00','   urlLink    View of San Fransisco from Alcatraz Prison '),(13507,3828283,'08,July,2004','2004-07-08 00:00:00','   I\'m back home and will try to post pictures using Hello the chat software from Picasa that makes it possible  to upload pictures to your Blogger blog. Have you been buying songs on iTunes to try to win the awesome prizes? If you haven\'t this is probally going to be the last day. I forgot to tell you all that you don\'t even have to buy a song to enter you can use the Tell a Friend feature in iTunes and type itunes100@apple.com for the email address and you will be entered, it is equal to buying one song. I know it is a little late but you still have the opportunity to win the grand prize. '),(13508,3828283,'08,July,2004','2004-07-08 00:00:00','   At the time of this post Apple iTunes has sold 98,181,799 songs so get downloading. '),(13509,3828283,'07,July,2004','2004-07-07 00:00:00','   At the time of this post Apple iTunes has sold 97,642,750 songs so get downloading.  If you don\'t already have iTunes  urlLink download  it. You won\'t regret it.   I will continue to update the number of songs sold at least twice a day untill the 100th million song is sold.  I have a counter that shows the number of songs sold. I can\'t figure out how to get it to just show up but here\'s the  urlLink link . '),(13510,3828283,'06,July,2004','2004-07-06 00:00:00','   At the time of this post Apple iTunes has sold 97,230,307 song so get downloading.  If you don\'t already have iTunes  urlLink download  it. You won\'t regret it.   I will continue to update the number of songs sold at least twice a day untill the 100th million song is sold '),(13511,3828283,'05,July,2004','2004-07-05 00:00:00','   As I mentioned in a post on July 3 named Late Announcement Apple will be giving away different things. \"Good news for those people who actually use iTunes Music Store -- in celebration of the approaching sale of the 100 millionth song, Apple will be giving away 50 special 20GB iPods to the purchasers of every 100,000th song. If you\'re the lucky person who purchases the song number 100 million, you\'ll win a 17-inch Powerbook, a 40GB iPod, and a gift certificate for 10,000 iTunes downloads. Oh, and your own Celebrity playlist too. This all starts when Apple iTunes sells 95 million songs.\" At the time of this post Apple iTunes has sold 97,089,788 song so get downloading.  If you don\'t already have iTunes  urlLink download  it. You won\'t regret it. '),(13512,3828283,'05,July,2004','2004-07-05 00:00:00','   Democrat John Kerry has chosen his running mate and it is .....   urlLink SEN. JOHN EDWARDS . I think this was the best choice because John Edwards is a hard working  Average American who worked his way to college to get a law degree with top honors then worked as a trial attorney where he made his fortune. Edwards is an extremely good debater and very photogenic who had to work for a living. Visit the  urlLink campaign website . '),(13513,3828283,'05,July,2004','2004-07-05 00:00:00','   This website is so funny. It\'s called  urlLink FUH2  and is dedicated to flipping off the Hummer H2. Even if you like the H2 you should check this out. Read the home page all the way to the bottom and then check out the pictures section. '),(13514,3828283,'04,July,2004','2004-07-04 00:00:00','   Happy Fourth of July Everyone. Hope your having a great time. Just remember try not to hurt yourself while lighting fireworks '),(13515,3828283,'03,July,2004','2004-07-03 00:00:00','   I just created a  urlLink Mac and PC  section on my  urlLink website . My first article and the only one right now is focused on informing everyone about the Mac and PC. I know not everyone will agree with my opinion, so check it out and form your own. '),(13516,3828283,'02,July,2004','2004-07-02 00:00:00','   Good news for those people who actually use iTunes Music Store -- in celebration of the approaching sale of the 100 millionth song, Apple will be giving away 50 special 20GB iPods to the purchasers of every 100,000th song. If you\'re the lucky person who purchases the song number 100 million, you\'ll win a 17-inch Powerbook, a 40GB iPod, and a gift certificate for 10,000 iTunes downloads. Oh, and your own Celebrity playlist too. This all starts when Apple iTunes sells 95 million songs. At the time of this post Apple iTunes has sold 95,805,796 so get downloading.   If you don\'t already have iTunes  urlLink download  it. You won\'t regret it. '),(13517,3828283,'02,July,2004','2004-07-02 00:00:00','   I decided that I would just change my  urlLink website  template for right now and maybe next month or later redesign my website. If anyone wants to create their own website you should host with  urlLink Freewebs . They are the best free web hosting company ever.   I also added a domain to, so instead of having to go to www.freewebs.com/ipodsrock, I can go to www.zachkarst.tk. The .tk domains are totally free. You should get one  urlLink yourself . '),(13518,3828283,'02,July,2004','2004-07-02 00:00:00','   Visit my  urlLink website . I have recently updated it and planning on even more content. I might even decide to totally redesign my website using Dreamweaver or Frontpage if I find time to do that. '),(13519,3828283,'02,July,2004','2004-07-02 00:00:00','   I just can not make up my mind. I have changed once again first it was Live Journal then it was Blogger and once again it is Blogger. I was trying to change my theme and make new posts but it was not working so I decided to just create a new blog. Hopefully I will stay with this blog. '),(13520,3828283,'13,August,2004','2004-08-13 00:00:00','   I just installed Windows XP Service Pack 2. I was on vacation so I couldn\'t install it last week so when I got home I did it. I found a  urlLink direct download link  to download SP2 on  urlLink kevinrose.com , so you don\'t have to wait for Microsoft. '),(13521,3828283,'03,August,2004','2004-08-03 00:00:00','   I just got back from Seattle. It was great I visited the Seattle Flight Museum where I boarded a Concorde and got to see the amazing cockpit and cabin. I also boarded Air Force one as it was called when the president rode it. It carried Eisenhower, Kennedy, Johnson and Nixon. The plane has a area for press and staff, a lounge type conference room and the presidents suite complete with a lavatory with shower. The seats for the staff and some other seats have leg rests that pop out for their comfort. This plane is full of history Eisenhower on his first jet flight in 1959&#8212;an experience he termed \"exhilarating.\" On that trip he met with Chancellor Konrad Adenauer in Germany, conferred with Prime Minister Harold Macmillan and the Queen in London, and then touched down in Paris to talk with President de Gaulle. President Kennedy used the plane to attend a conference with Macmillan on the crisis in Laos and to meet for the first time with Russian Premier Nikita Khrushchev. Vice-President Johnson flew our VC-137B to Love Field in Dallas on the day President Kennedy was assassinated. Secretary of State Henry Kissinger used the Museum\'s plane for his famous \"shuttle diplomacy\" flights during the Nixon administration. It took him to Paris for secret peace talks with the North Vietnamese, to China to arrange for Nixon\'s historic first visit there, and repeatedly to the Middle East to conduct peace talks between Egypt and Israel. There were also many other exhibits that I found interesting. While I was there I took a tour of Boeing\'s Everett plant where they make the 747 and the 777. I have to say it was the lamest thing I have ever done. But overall it was a good trip. '),(13522,4257430,'20,August,2004','2004-08-20 00:00:00','   Man I am already trying to work on my thing. It\'s really hard to find all the stuff. Today my sister-law got this post card from her aunt. So she gave that to me. I got lucky guys! Is anyone else already having probs finding stuff? lol. Well good luck to you guys, I know I will need it.   ~Manda '),(13523,4257430,'17,August,2004','2004-08-17 00:00:00','   Oh i love my blog, and it loves me. My name is Manda. I am 16 as of right now. I got to a high school in Kentucky. I live in a house with a cat named Jasmine. I have a mom and dad. They\'re pretty cool. They take care of me. I work at the local DQ. I hate work. I am sure it hates me too. And I really don\'t know what to say. so later      ~Manda '),(13524,4257430,'13,August,2004','2004-08-13 00:00:00','   I\'m setting up this blog for my second period creative writing students.  I am hoping that they will fiind it interesting to keep personal thoughts, ideas, and writing considerations in this blog.  Likewise, we can post what we write and respond to each other\'s ideas.  Welcome and good luck blogging! Ms. H  '),(13525,4257430,'18,August,2004','2004-08-18 00:00:00','   OK gang, I\'ve decided to give you some homework this weekend--and lots of it.  After all, that\'s what you would expect isn\'t it???  Seriously, I\'m sending you on a scavenger hunt.  In order to complete some activities I have planned for a    P O  E  T  R  Y   E  X  T R A  V A G  A  N  Z       A    next week, I need you to have some materials ready.  Your assignment is this:  Each student in the class needs to find each of the following items.  The goal is to find the objects which will be used in class (so you shouldnâ€™t need to have them in the same form to take back home), and not spend any money on them.  Look around the garbage can for the junk mail or a \"stash-it-all\" drawer for the postcards or letters, for example.  Items due in class on Monday and I will give you credit for all of your hard work.  Good luck and happy hunting!    Item list   One permanent ink marker (preferably some dark color) One t-shirt (white or light color) that fits you An old post-card (blank) An old post-card with message A very, very, very sharpened pencil At least four used postage stamps (at least 1â€ x 1â€--the larger the better)â€”on the envelopes or off Ten pieces of junk mail An old magazine (school appropriate :) Some sort of blank book A piece of wood (no larger than 1â€™ x 1â€™ x 1â€) An old calendar page (any month--bring extras if you find them) An old letter A map (can be local, national, world, etc.) A family recipe A shoebox to hold all of it (unless your piece of wood is too big--try to make it fit!)    Point #2 You should begin e-mailing me your drafts that you are beginning to finish.  Mail to:   urlLink lhager@mercer.k12.ky.us  '),(13526,4257430,'17,August,2004','2004-08-17 00:00:00','   Given that your goal is to write at least two entries a week into this blog, I think I\'ll set things up like this. Each week, I\'ll give you a prompt (for example, the one from yesterday) and for the other prompt, you will be invited to either post your own writing for people to read/respond to AND/or respond to a second prompt. We\'ll see how this works. Even though things can be read by anyone searching through the blog, I believe that keeping the response contained will make things run much more smoothly.  Thanks. Ms. H  '),(13527,4257430,'16,August,2004','2004-08-16 00:00:00','   I\'m trying to figure out why some of you are on the blog and some aren\'t.  From what I can see, it looks like those of you who followed the email link are on and those who didn\'t aren\'t.  I will try to figure out how to fix this.  Meanwhile, could those of you who DIDN\'T follow the link go back and try to follow it.  It might be that the system simply has to make the connection.  Try this and see what it does.  Thanks!  '),(13528,4257430,'16,August,2004','2004-08-16 00:00:00','   Trot right on over to  urlLink http://www.writersdigest.com/  to find loads of new writing prompts.  You can even sign up to have them emailed to you each month automatically.  Now what could be better than that :)  '),(13529,4257430,'16,August,2004','2004-08-16 00:00:00','   Hi all, See, this blogging thing isn\'t too hard after all! Even I can do it. I hope you\'ve made your way around online and found this journal. My intention is for us to be able to keep it as a class journal much in the way we would do if we were keeping a written, bound journal. The way I understand it (yes, I\'m still trying things out here), is that after I put your email in you have to \"join.\" You should have received an email \"invitation.\" Then, you are able to post to this blog.  I\'d like to start out by setting up a first prompt for you to respond to: how do you perceive class going so far? I\'m interested, particularly, in knowing how you are responding to some of the \"freedoms\" that writing workshop allows like being able to define your own writing tasks--things that you want/need to write about, etc. How do you feel about being given writing time in class rather than it all being given as homework? What would you like to see as a feature of your class??  The other thing I\'d like some feedback on (and I\'m going to be asking you in class about this, too) is what you might want our webzine (and print magazine) to look like. What should its title be? We have to start thinking about how we will invite other writers to join our little party, so it\'s going to be important to make some of these decisions soon.  I can\'t tell you how long I\'ve waited to teach a creative writing class. I am really enjoying what you\'ve done so far. You just have to promise to push me as hard as I push you (well, maybe not that hard :)  Post your thoughts! I think I\'ll stumble over to the blog for portfolio_princess and journal my thoughts for the day. . . Let me know if you set up an online blog for yourself (or if you already have one you want to keep working in). I\'ll have to make note of the URL. Ms. H  '),(13530,4257430,'22,August,2004','2004-08-22 00:00:00','   We are getting ready to dive into our poetry extravagannnnnzzzzzaaaaaa this week. I\'d like to invite you to post something you write during the week to get feedback on. Have fun with it!  '),(13531,4257430,'22,August,2004','2004-08-22 00:00:00','   For this week, I\'d like you to consider this question. What is the best piece of writing you have ever done? Describe it. In what particular genre did you write? Why do you consider it your best?  '),(13532,3343256,'31,May,2004','2004-05-31 00:00:00','   At this time, we are not able to identify a match for you.  We require that any match we provide be highly compatible with your values, beliefs and attitudes. This standard means that thousands of eHarmony users are often not matched right away.    We are optimistic that we\'ll find the right matches for you and hope you will remain optimistic too. We don\'t intend to lower our standards and you shouldn\'t either. We ask for your patience during this process and assure you that the right match will be worth the wait.   Translation: We\'re sorry, but there is no one compatible with you, even with the search area set on \"The Entire World\".    Ouch.  '),(13533,3343256,'30,May,2004','2004-05-30 00:00:00','   urlLink GuruNet  is an interesting little program I\'ve found for those, like me, that sometimes look at a word and go \"huh\"?  All you do, is load this little program onto your computer, then you hold alt, click a word you don\'t understand, then release alt.   Then the program pops up and gives you a definition.  It\'s very nice, and very useful.  Hope its useful to someone out there.  '),(13534,3343256,'29,May,2004','2004-05-29 00:00:00','   ...its because I don\'t understand it well enough myself.  So, if you read anything, and go \"huh\"? Be sure to leave a comment on what you don\'t get.  That will help me focus on what to learn more to make clearer.  I\'m looking more for concepts that I either seem to take for granted, or anything else that I haven\'t explained clearly, rather than words you are simply unfamiliar with.    Thanks, this will help me lots. :)  '),(13535,3343256,'29,May,2004','2004-05-29 00:00:00','   How to review a book of such a powerful message in such a short space, I have no idea.  This book is one I would recommend to anyone who is serious about doing something with their life.  It provides the tools necessary for a person to get past the issues that hold the vast majority of us back from reaching our potential.  The prime power of this book is its ability to create a paradigm shift in your way of thinking.  For those who aren\'t familiar with that term, here is how the dictionary defines it:  A set of assumptions, concepts, values, and practices that constitutes a way of viewing reality for the community that shares them, especially in an intellectual discipline.   We are all brought up with a certain way of looking at the world, of there being certain truths to things.  This book helps people to shift their paradigm of the world to be able to see situations and problems in a whole new light.  This paradigm shift if the whole basis, in my opinion, behind this book\'s power. A quote from Einstein in the book reads:  The significant problems we face cannot be solved at the same level of thinking we were at when we created them.   The way we view the world significantly impacts our ability to affect the world.  Very briefly I will try to run through the 7 steps listed in the book.  1.  Be Proactive : It sounds almost cliche to tell people to be proactive, but it\'s an important basis to begin at.  Many people let others control their lives, either through expectations, or hatred of other people determining your actions, or simply doing something you\'ve always done just because that\'s the way you\'ve always done it.  In short it means being responsible for your life.  Which simply means being able to choose how to respond to situations instead of just reacting in a pre-programmed way.  2.   Begin with the end in mind :  Why do you do what you do?  What drives you? What do you want to accomplish, what do you want to be remembered for?  What makes you different from everyone else around you?  Whereas the first habit tells you that you  can  have control over how you act, the second habit tells you where to direct that control.  Or more accurately, it gives you the tools to let you determine the direction.  3.  Put first things first :  This habit is basically the time management chapter.  But not in the traditional sense of scheduling everything out.  Rather it teaches you to do those things that are important yet not urgent.  Think of it as preventive maintenance.  It helps to remove the constant panic of crisis management, and gives time to build relationships and plan your vision for your life.  With this habit you put the first two habits into action.  Together these 3 habits work to make you an independent person, able to determine your own future.  4.   Think Win/Win :  This habit is the first of those for dealing with other people.  It\'s a strong message to give, and one that I personally fully agree with.  The message is to, in every situation, see a path that is mutually beneficial for all of those involved.  Instead of just looking out for yourself, look for what is good for everyone, because you never deal with a person just once.  If you question the validity of this habit, just consider for a moment how you feel towards someone who wants you to do something and that you feel has no interest in your own well being.  5.   Seek first to understand, then to be understood : This is the chapter on  urlLink Empathy  that I wrote about earlier.  I can\'t really add much to what I said in that post, other than that if you want to get the most out of all your relationships, learn to look through the other person\'s eyes first.  Once you do that, they will then be willing to look at the situation through your eyes.  6.   Synergize :  A fancy word that simply means that the whole is greater than the sum of two parts.  This habit is strongly related to the previous two.  By looking for a mutually beneficial solution, and understanding the other\'s point of view, you can accomplish so much more together than simply adding together what you could do separately.  7.   Sharpen the saw :  This habit is about taking the time to step back, and keep yourself healthy and on track.  With this habit you learn to take the time to take care of your needs so that you can more successfully perform the other habits to their highest perfection.  With this habit you renew your ability to go on.  Well, that was rather long, but hopefully it will inspire some of you out there to pick this book up.  If you are serious about personal growth and maturity in your life, I promise you won\'t be disappointed.  Oh, and you can click the a link at the right to get it from Amazon and help me out in the process. ;)  '),(13536,3343256,'28,May,2004','2004-05-28 00:00:00','   Are individuals, and developed countries in general, over insured?  Everything from cars, to jobs, to health, so interest rate changes and currency flutuations seems to be insured these days.  Now, don\'t get me wrong, insurance is a nice thing to have.  Most people can\'t afford to dish out tens of thousands of dollars if their car is smashed, or if they need major surgery.  But in those cases the insurance works in such a way that small payments over time balance out the large bills that happen rarely.  It\'s like paying on a loan before you need it, and perhaps never needing it, yet at the same time being able to draw on it multiple times.  That\'s classic insurance, and its designed to handle big problems.  What is appear more and more however is full insurance, that is designed to cover all the little problems that pop up.  That is where the problem lies.  As every little problem problem surfaces, the insurance takes it away, and no behavior is changed.  Normally when something bad happens, say getting a ding on their car, the people pay for the repair if they can afford it, and want it, and if they don\'t, they don\'t.  With \'full\' insurance that ding is repaired whether it\'s really necessary or not.  This analogy can be extended to most about anything.  People no longer exert any personal control on their spending, because it is not  their  money that is paying for it.  At least, that\'s what they think.  In the end, the money that is payed into an insurance plan has to at least match what is paid out (plus overhead and profits) in order for an insurance company to stay in business.  This relates directly to my  urlLink previous  post.  There are things called derivatives in the financial world.  What they are generally used for is to blunt the impact of price changes through-out the economy.  Including exchange rates.  Many economists recently have remarked that changes in the dollar\'s value is having a reduced effect from what is normally expected, due to the fact that so many companies have used these derivatives to insure against changes.  As a result they are not reacting to changes in the economy as they should be, as they are too well insulated to notice.  Health insurance is another big area where over insurance is taking its toll.  With no incentive to pay attention to where a person\'s money is going, people are more likely to spend it unwisely.  Purchasing prescriptions that they may not need because the doctor is being pressured to push them.  Getting unneccesary procedures done, because the cheaper plan is not covered under the HMO\'s plan.  Overall health is allowed to deteriorate because if someone gets sick, they can just go to the doctor, no worries.  As a result, health costs continue to surge year after year.  All in all, I think insurance is starting to get out of hand, and we are starting to pay the price for it.  Side note:  urlLink HSAs  are now available to help blunt the cost of medical care, while still letting the individual control what is going on.  '),(13537,3343256,'26,May,2004','2004-05-26 00:00:00','   urlLink Previously  I discussed the U.S. current account deficit and how it related to the federal deficit.  In that post I stated that the federal deficit is simply a reflection of the current account deficit shift from private into public hands.  This rationally makes sense, due to the fact that much of the deficit was created by tax cuts.  Those tax cuts allowed the public to increase or maintain their current purchases (there by maintaining the current account deficit) while the government footed the bill.  This spells bad news for the economy in my opinion.  The existance of a large current account deficit means that we are buying more than we are selling, in regards to the rest of the world.  This cannot continue indefinatly.  It is an indication that the economy is not sufficiently producing products domestically to match the demand for such products.  As such we are purchasing those products from abroad.  Now, that does not mean we should not purchase those foreign products.  Thanks to the market, we can simply make products that are more valuable to the rest of the world to reduce the deficit.  How can we do that?  Normally the market will do that on its own, provided it isn\'t fiddled with in the mean time.  So who has been fiddling with it?  Answer: Everyone.  Japan and China, along with most other East Asian currencies manipulate the exchange rates to keep their currencies, and therefore their products, cheap.  With those currencies cheaper than the market would set it is artificially making their products more desireable than the U.S.\'s.  Now, there are other factors, including maintaining stability in China, on why those countries are manipulating their exchange rates.  I am attempting here to simply analyze, not place blame.  Additionally, the Bush Administration\'s tax cuts have fiddled with the economy.  Around the beginning of the decade the economy was slowing as, I believe, it was attempting to correct the current account deficit.  I like to call this a purification cycle, others call it a recession.  At this time the tax cuts were being pushed through, and many people were remarking on the excellent timing of it to avert the recession.  And indeed, it worked.  The recent recession is remarked as being one of the mildest in history.  However, at what cost?  Already it is visible that the recovery was slow in coming.  That however may be attributable to other circumstances, such as the enhanced gains in productivity.   What about the future though?  We still have a large (5% of GDP or so) current account deficit that the market will keep trying to correct.  One way to correct is for the exchange rates to adjust, with the dollar falling.  However the dollar needs something to fall against.  East Asia won\'t let their\'s fall.  Europe is simply unattractive, and with the potential of interest rate reductions, will become more so.    The other, more likely option is for the previous solution to be attempted by the market again: purification cycle.  '),(13538,3343256,'26,May,2004','2004-05-26 00:00:00','   Thats what I need.  Four, and just four hours a day.    Why? Because that is, I\'ve discovered, about how long during each day at work my mind is able to come up with and process new ideas and theories.  So thats what I need.  A job that doesn\'t require me to think, and lets me just do repetitive mindless tasks for 4 hours.  Then I can come home, sleep on the ideas, and then write about them.    Which is what I\'m doing right now.  '),(13539,3343256,'26,May,2004','2004-05-26 00:00:00','   Yesterday, while reading  urlLink The 7 Habits of Highly Effective People  I was deeply moved by the chapter on the 5th Habit: Empathy.  The core principle that the author was trying to convey was:  Seek first to understand, then to be understood.   What that means is empathy is not sympathy, where you relate your own experiences or understanding to what someone else is going through.  Rather empathy is taking the time to see and understand what is going on from the other person\'s point of view.  I was moved by this because it struck me with such force that  this  is what I was missing growing up.  I never considered my childhood happy.  Oh, it was better than many, I know that.  There was no abuse, plenty of food, two parents.  But there was always something missing that I couldn\'t quite put my finger on, and that no one would listen when I tried to express it.  That something was empathy.  No one understood me.  It wasn\'t until about a year ago, when I discovered my INTP nature, that I even made the connection that I wasn\'t being understood.  I had just accepted the fact that I was wrong, and I should be happy.  But I never was, and never felt that close to anyone in my family.  Reading the sample conversations in the book I felt like I was going back in time, to the point that it actually brought tears to my eyes.  Trying to explain what was wrong, and the people I was talking to trying to fix the problems without really taking the time to listen to what was  really  wrong.  Even today I get that when I talk with people about what I\'m going to do for a job or career, and someone just says \"Go to college!\" as if that will solve all my problems.  I\'ve actually felt offended when someone says that to me, and now I know its because they didn\'t really take the time to understand.  They just wanted to fix the problem from their perspective, from what they would do, or think  they  should have done.  What I am really looking for is someone to take the time to understand what  I  want.  Now, obviously thats a pretty big want.  Most people don\'t care at all about another person\'s problems.  I guess I am fortunate at that, to an extent, I do take the time to care.  I even made two friends that way, by sitting there and listening to them talk about the problems in their marriage.  Oh, I wasn\'t, and am not, the perfect empathizer, I know that.  However I found it rather amusing that one of the suggestions for action at the end of the chapter was to catch yourself when you arn\'t listening anymore, and are trying to judge the situation by your own autobiographical stance.  It was amusing because I recall many, many times when I would try to offer a solution to a friend, and realize that I wasn\'t listening, and actually say \"Oops, trying to fix the problem, shutting up, continue\".  So, keep this in mind people.  If you really want to have a good relationship with someone, learn empathy.  Understand what is going on from their point of view, don\'t try to relate it to your own experiences.  Take the time to actually care.  They will appreciate it, and so will I.  '),(13540,3343256,'25,May,2004','2004-05-25 00:00:00','   Browsing through  urlLink U.S. trade data  I\'ve discovered the primary driver of the current account deficit.  Oil, cars, and electronics.  Oil products and vehicles alone make up approximatly 1/5 of the total increase in the size of the U.S. current account deficit from 1998 to 2003.    Where is the rest coming from? No where specifically.  All across the board imports are up, while exports are mostly steady.  The result? An ever widening deficit.  We are essentially borrowing to drive.  An important note to remember though.  This is the U.S. current account deficit, not the U.S. governments fiscal deficit.  The difference is that the current account deficit reflects the total of both public and private debt.  Additionally, the current account deficit grew the most between 1998 and 2000, almost doubling in fact, while the government\'s fiscal deficit was actually a surplus by the end.  Interestingly, the current account deficit grew by only about 1/4 between 2000 and 2003, while the government\'s fiscal deficit grew massivly.  In 2003 alone the federal deficit was $374B while the current account deficit was $535B.  What does that mean?  It means that although the size of the current account deficit has continued to grow, the source of its growth has changed from private to public hands.    Which begs the question, should the government seek to reduce its burden, and if so, what will happen?  I will attempt to address that question in a later post.   Update:  I\'ve recently recalled that Bush has ordered the Strategic Petroleum Reserve to be completely filled.  That most likely is a prime cause of oil\'s substantial increase in imports.  '),(13541,3343256,'23,May,2004','2004-05-23 00:00:00','   Ok, Missa just dropped me a note saying she couldn\'t leave comments here for some reason.  I went into the settings, and re-saved it as \"Anyone can Comment\".  I even did a test by logging out.  Seems to work now.  If you wanna have your name appear, instead of it being from Anonymous, then feel free to register.  So far I have received exactly 0 e-mail from Blogger.  So I think its pretty safe regarding spam.  Happy commenting people.  And let me know again if it still doesn\'t work.  '),(13542,3343256,'22,May,2004','2004-05-22 00:00:00','   I\'ve been reading a lot of books lately, and keep either getting hung up, or bogged down with the reading.  In other words its not going as fast as I\'d like, so I\'m looking for ways to improve my reading performance.  I\'m currently looking at two books, and am wondering which to get.  Or even both.  They are:   urlLink Breakthrough Rapid Reading   and   urlLink Power Reading : The Best, Fastest, Easiest, Most Effective Course on Speedreading and Comprehension Ever Developed!   Both have very good reviews.  Just reading the reviews I think I\'d get the first one.  Any suggestions or alternate recommendations would be greatly appreciated.  '),(13543,3343256,'20,May,2004','2004-05-20 00:00:00','   Ugh, I\'m certainly understanding right now what all those personality type pages are talking about when they say that working with less developed functions is stressful for people.  First some background:  The four letters in the personality types translate into four functions that describe how our minds work.  The first and last letter describe the details of the middle two letters.  I\'ll explain the exact details later, but for an INTP that means that Ti (introverted Thinking) and Ne (extroverted iNtuition) are the dominant and auxillary traits.  There are how we interact with the world the most.  Right now I\'m working on a program design to simulate how an economy works.  And that is forcing me to use my Si (introverted Sensing) function to flesh out all the details to make sure it operates correctly (Sensing means details, iNtuition means big picture).  So while I can invision how the integrated whole is supposed to look, my brain is getting fried working out the details of the implemention.  And its giving me a headache.  Incidently I think that the stress works both ways.  Working in lower functions causes stress, and stress can cause you to work in lower function.  A deadly cycle.  '),(13544,3343256,'20,May,2004','2004-05-20 00:00:00','   Exactly how similar are the cultures of Europe and America?  Many people seem to shove us all together into one big entity called \"Western Society\".  The Aussies are shoved into that conglomeration as well.  But is it true?  Are we really as similar as many people think.  Sure, when compared with the rest of the world, we are very similar.  As is expected, considering the fact Europe was the launching ground for the building of the Americas and Australia.  But even with those similarities there are so many differences that people seem to overlook.  Certainly recent events can attest to the differing views on how to handle global issues.  Local issues seem to me to be an even greater source of differences.  Social welfare is one particularly large example.  European cultures, along with Canada, place a higher premium on such spending, to the extent that they seem to believe Americans don\'t care about each other at all.  Which any American would respond that that couldn\'t be farther from the truth, we simply have different ways to doing things, such as chartiable giving.  Even within America and Europe there are vast differences.  This post today was inspired by an article in the Wall Street Journal about an EU-Swiss banking deal.  The Swiss rejected the more sweeping tax-evasion rules that the EU wanted, saying:   \"Privacy is one of our core values...We like to trust the citizen instead of assuming he\'s a cheat.\"  A book I am reading,  urlLink The Wealth and Poverty of Nations , has helped me understand some of the unique characteristics of \'Western\' society that helped it grow into the world power that it is today.  Hopefully light will be shed, as I continue reading, on the differences within the cultures as well.   Such an understanding that we are in fact seperate and distinct cultures I suspect would help to at least calm some of the tensions and misunderstandings among us.  Along the divide of right and left cultures, the left seems to look down on others as having unacceptable values, while the right parodies left\'s hipocracy saying \"We believe everyone has the right to live in their own way, as long as it agrees with our view of what that way should be.\"  Perhaps if more people understood that we are all different cultures, there would be less hate among us \'allies\'.  A vain hope I suspect.  '),(13545,3343256,'18,May,2004','2004-05-18 00:00:00','   Note to self: Making a draft article will cause the article to be posted under the date that it was first saved, not the date it was posted. (See health care article at bottom).  Note to readers: Missa confirmed that she blushed. :P  General notes:   I have a site counter now....so I KNOW if you don\'t come back!  Learning how to blog is rather interesting, as I am getting lessons in HTML as well.  Largely due to the fact that you simply can\'t click a button to add a feature.  You have to go into the source code on the page to edit stuff in there.  A little more work and I can change those nasty colors that are the blogroll default.  That is all, carry on.  '),(13546,3343256,'17,May,2004','2004-05-17 00:00:00','   Known her for a while online.  Talked to her a little bit recently.  Had no clue she was INTP.  She sent me a pic today at my request.  More good news, she\'s cute.  I\'m IMing with her right now, and am trying to see if I can get her to blush. :P  *winks at Missa*  Damn, I probably should have checked if she had a boyfriend first.  '),(13547,3343256,'14,May,2004','2004-05-14 00:00:00','   Yes, you heard me.  I cut my own hair. I\'m cheap.  And, it means I don\'t have to deal with people who seem to have an obsessive-compulsive need to talk.  You know...hair dressers, or barbers, or whatever they wanna call themselves.  The right side of my head is a lil more jagged than the left though.  I think I need to clear it up tomorrow morning after I take a shower...that should do it.  Hopefully I can aim good so it all lands in the garbage without me having to vacuum again.  I probably should have taken my dad\'s advice to cut every week, so if I screw up, no one can notice.  But....I\'m too lazy to cut my hair more than once a month.  Ah well. :P  '),(13548,3343256,'14,May,2004','2004-05-14 00:00:00','   What excellent timing.  Browsing The Economist website and I find this  urlLink article .  A good sign for my first day with a blog.  Turns out birds, ravens at least, have what is called the \"theory of mind\" which basically means that they have an understanding that other creatures have minds of their own.  Here\'s a bit from the article demonstrating a bird\'s intelligence and awareness:   Dr Bugnyar...noticed strange interactions between two males, Hugin, a subordinate bird, and Munin, a dominant one.   The task was to work out which colour-coded film containers held some bits of cheese, then prise the containers open and eat the contents. The subordinate male was far better at this task than the dominant. However, he never managed to gulp down more than a few pieces of the reward before the dominant raven, Munin, was hustling him on his way. Clearly (and not unexpectedly) ravens are able to learn about food sources from one another. They are also able to bully each other to gain access to that food.   But then something unexpected happened. Hugin, the subordinate, tried a new strategy. As soon as Munin bullied him, he headed over to a set of empty containers, prised the lids off them enthusiastically, and pretended to eat. Munin followed, whereupon Hugin returned to the loaded containers and ate his fill.   At first Dr Bugnyar could not believe what he was seeing...Hugin, he is convinced, was clearly misleading Munin.   As it happened, Munin was no dummy either. He soon grew wise to the tactic, and would not be led astray. He even stooped to trying to find the food rewards on his own! This made Hugin furious. â€œHe got very angryâ€, says Dr Bugnyar, â€œand started throwing things around.â€ Perhaps ravens have something else in common with peopleâ€”a hatred of being found out.   Hmm, yeah, I get upset when my attempts at fooling someone go off track too. :P  '),(13549,3343256,'14,May,2004','2004-05-14 00:00:00','   Ok, first post.  Don\'t tell everything about it in the title.    Too late.   urlLink Take the test.   Here are my results:  Extraversion&#160(E):  14&#160&#160 54 :(I) Introversion Sensing&#160&#160&#160&#160&#160&#160&#160 (S):   3&#160&#160&#160 69 :(N) iNtuition Thinking&#160&#160&#160&#160&#160&#160(T):  72&#160 18 :(F) Feeling Judging&#160&#160&#160&#160&#160&#160&#160(J):   9 &#160&#160 63 :(P) Perceiving  You scored as an INTP.  Assuming that you are an INTP,  Your DOMINANT function is Introverted Thinking. Your AUXILIARY function is Extraverted Intuition. Your TERTIARY function is Introverted Sensing. Your INFERIOR function is Extraverted Feeling.  Extraverted Thinking / Introverted Feeling   : 16 Extraverted Feeling / Introverted Thinking   : 53  Extraverted Intuition / Introverted Sensing  : 21 Extraverted Sensing / Introverted Intuition  : 35  Rationality (Dominant Judging Function)      : 41 A-rationality (Dominant Perceiving Function) : 3  Fits with the other tests I\'ve taken before.  The questions were much, much more useful on this quiz though, as they let you weigh your responses or even choose not to answer an specific question.  '),(13550,3343256,'14,May,2004','2004-05-14 00:00:00','   Read about this the other day in the Wall Street Journal.  Found it online  urlLink here .   Mr. Kerry, for his part, has a sweeping plan that tries to cover all uninsured children and most uninsured adults without the kind of fundamental structural change that doomed past Democratic proposals. It would cost $650 billion over 10 years, his campaign estimates, and would be financed by rolling back the Bush tax cuts for those earning over $200,000 a year.  The most unusual part of Mr. Kerry\'s plan would have the federal government pick up 75 percent of the cost of the most expensive medical cases - those of over $50,000 a year - if employers guaranteed that they would pass the savings along to their workers through reduction of premiums. This is intended to ease the burden on businesses, especially small ones, and provide cost relief to Americans with insurance.  I\'d say this is a good start.  I\'ve considered an idea similar to this for a while, where an ultra high government funded deductable was in place.  This plan has several good points.    One, its  a  plan.  And not a huge massive government controls everything plan.  Something like this would remove the need for such a centralized system that many people, myself included, believe could only cause spending to spiral ever more out of control.  Second, it will indirectly encourage the general trend towards moving people onto private health care.  With the government catching so much of the tab above a certain level, the \"cherry-picking\" problem some people cite will be largely eliminated, or at least severely reduced.  I have little problem with this being government funded, due to the fact that I see the government as something that should only do something either when a) Certain programs are not forced on those people who do not want them, or b) The indirect effects of such programs creates such side advantages to outweigh the drawbacks of less economic freedom.  All in all, its a good start to the health care debate.  '),(13551,3343256,'30,June,2004','2004-06-30 00:00:00','   I love mind puzzles.  They are a great mental workout.  Fortunately, I\'ve found a blog that have a new one (or more) up every day.  It\'s called  urlLink think again! .  Check it out if you like that sort of stuff.  Also, I ran across this animated puzzle game called  urlLink Chasm .  You are some sort of creature, that must fix the town\'s water supply, which has mysteriously broken.  Through a bunch of flood gates that you discover you can manipulate the flow of water to get it to flow where you need it to.  Took me perhaps an hour total.  Probably less.  A fun diversion in any case.  Enjoy.  '),(13552,3343256,'27,June,2004','2004-06-27 00:00:00','   Ok, this is a bit late, as it happened a couple months ago.  But, basically the biggest stop for our bakery was a grocery store just down the street.  They had been slumping in sales, every year worse than the last, for about the last 5 years.   Why?  Because the super big grocery stores kept popping up.  It severely cut into this store\'s business, as they were more service oriented, and so higher prices.  Our products were more high end kind of stuff, so we fit in nicely there.  Lots of rich old folks who don\'t like the horribly soft bread, or rock hard cookies.  But, our sales there finally slumped enough that the owner kicked us out in favor of his own product, which he got through his store affiliation (Spartan stores).  Of course, his cookies tasted like crap, and I never even tried his bread.  I used to do my shopping there, but don\'t any longer.  See the post  urlLink below .  So, big stores brought about by globalism cost me my job.  But, as in my post below, I am not upset.  Especially as I did not particularly like my job anyway.  I\'ve also already figured out a likely  urlLink alternative  for work.  In any case, I still got one month left before I am laid off.  Then it\'s free time galore!  And no money...but lots of free time galore!  '),(13553,3343256,'27,June,2004','2004-06-27 00:00:00','   It\'s so useful at making things cheap.  Just yesterday I went to  Meijer  and bought a 12pk of bar soap.  For less than $6.  I don\'t remember exactly how much I was paying at the old grocery store I went to, but I\'m pretty sure it was more than 50 cents a bar.  Which is actually a big deal, as I need to save all the money I can, due to the fact that I am soon to be unemployed.  Right now my grocery budget is about 20% of my expenses.  Definitely room for improvement.  Globalism is also nice in bringing down the prices of just about anything.  I love it.  Then again, maybe I\'m thinking of free markets?  Free market globalism...hmm, yes, that works.  So, some people can whine all they want about it, but as far as I am concerned, it\'s a plus, despite the fact that it cost me my job.  Oh, wait, I never wrote about that, did I?   Hmm.  '),(13554,3343256,'27,June,2004','2004-06-27 00:00:00','   I just  urlLink stumbled  across this rather amusing  urlLink short movie .  It\'s a combination of Star Wars and Office Space.  If you are familiar with both, likely this movie will at least get a chuckle out of you.  Especially when you see which Star Wars character plays the \'stapler guy.\' You know, the one that mumbles, obsesses about his stapler, and was moved to the basement, with his paychecks finally cut off.  Yeah, that guy.  So, check it out.  P.S. The movie is long, over 4 minutes, so those without high speed internet, you were warned.  '),(13555,3343256,'27,June,2004','2004-06-27 00:00:00','   Just talking to one of my friends today, the self appointed editor, and she is totally distraught.  She is turning into a girl.  Pink purse, with polka dot lining.  And she bought it herself!  Disgusting, isn\'t it?  Another friend wrote this:   You know you\'re in deep trouble when...  ...all the sappy songs on the radio seem good and true. OMG. :O  This, from someone that hates the concept of sappy!  What is it with girls?  I mean, some of them are always girly, but others are all tough like usually, including both of these friends.  So, why do they turn girly?  Not that I mind of course, I find occasional girliness rather attractive.  Note the occasional part though.  All the time, and it\'s generally a sign of witlessness.  Ah well, they are still two friends that I love, sappy girliness and all.  If either starts wearing froofy pink dresses though, I\'d have to kill them.  For their own sakes, of course.   Update:  Example  '),(13556,3343256,'25,June,2004','2004-06-25 00:00:00','   Was at my parents today, for our family\'s monthly dinner thingy.  After everyone left, I was in the office, and asked my dad if he ever used his scanner.  He said he didn\'t, and that it didn\'t even work right on his computer.  So I asked him if I could have it, and looked it up on eBay.  There were two, for $5 and $10, neither with any bids.  He just told me too take it.  Cool, eh?  I sure thought so.  Now...just gotta figure out what I can use it for.  '),(13557,3343256,'25,June,2004','2004-06-25 00:00:00','   Yes, just one bee.  It lives in my back sliding door, between the main door/window and the storm door/window.  See, the storm door doesn\'t close all the way, so there is about a one centimeter opening that the bee can fly into.  It\'s using it\'s time making nests, for some reason.  I hope there isn\'t a queen around it is planning to invite in.    It has 3 nests so far.  Fortunately it seems to have abandoned the only one that is in the part that opens.  The other two are on the side that is permanently closed.  So, I am content to leave him.  He has a little spider friend next to his new nest.  Maybe they will kill each other.  I can hope at least.  Oh sweet! I just checked, and apparently he is cannibalizing the older two nests.  The one that I can access is almost completely destroyed already.  More observations on him later.  '),(13558,3343256,'25,June,2004','2004-06-25 00:00:00','   Hellooooo Googlers!  Just in the last couple days google has found my site.  Currently, a search for INTP will apparently show me as a result in the 80th entry!!!  That like...cool.  Not too shabby for less than 2 months blogging.  Oh, and in case any of you were wondering, yes, this is the stupid post I promised.  '),(13559,3343256,'25,June,2004','2004-06-25 00:00:00','   Here are things I want in a job:  Flexible schedule. Using my mind. Enough to make ends meet. One on one interaction.  Things I don\'t want:  No idiot boss. No idiot co-workers. No jerks in general. No menial labor. No groupwork.  A lot to ask for, eh?  Well, I think I discovered the perfect job: Tutoring!  I can set my own schedule, will only work with people that want me to help them, will work one on one.  I could earn $10, $20, or even up to $50 an hour, although higher amounts are not likely anytime soon.  I can start out on college students.  There are plenty where I live, one university, and one community college.  Heck, half the town is college students.  What kind of tutor? Math of course.  Most people despise it because they don\'t understand it.  I do understand it.  I was told in high school that I was a good teacher when I helped my classmates understand how to do something.  I even taught a teacher once how to do his own assignment! (Granted, he was kinda stupid.)  Einstein said there are only two infinite things: the universe, and human stupidity.  He wasn\'t sure about the first, and the second is going to make me rich! :D  '),(13560,3343256,'23,June,2004','2004-06-23 00:00:00','   That last post took too long, and I still wasn\'t fully pleased with the results.  Brain hurts now.  Must rest.  Stupid posts soon.  '),(13561,3343256,'22,June,2004','2004-06-22 00:00:00','   Generally when we think of the environment and the economy, we think of two competing goals.  While usually true, it is not always that way, especially now.  First, let\'s look at some figures.  With gasoline making up  urlLink 45%  of all consumption of oil, and oil counting for about  urlLink 3.5%  of economic growth, that means over 1.5% of our economy each year is spent solely on purchasing gasoline.    Looking at the  urlLink environmental side  it doesn\'t get any prettier.  In 2002 the U.S. emitted almost 5.8 billion tons of CO2 (table 4).  Of that, gasoline consumption for transportation amounted to over 1.1 billion tons(table 9).  Unlike most other countries in the world, the U.S. spends the majority of its oil consumption on  urlLink transportation , largely due to the fact that it is less densely populated, and as such requires more fuel to get where one needs to go.  Moving on to trade, I wrote before that   oil seems to make up a disproportionately large part of our current account deficit.   urlLink One fifth  of our current account deficit in fact is due to oil imports.  As stated by The Economist  urlLink recently  the deficit seems to have a life of it\'s own, regardless of whether there is a government deficit or not:   The OECD economists point out that private saving tends to fall when public saving increases. Between 1992 and 2000, for example, the Clinton administration turned a worrying budget deficit into a handsome surplus. But this only helped to unleash a private investment boom. Public saving was offset by private dissaving, ensuring that the country\'s trade deficit continued to deteriorate. America\'s budget and trade deficits may be twins but one, it seems, can survive without the other.  While you  undoubtedly  found this very interesting, you are likely wondering what my point is.  My point, simply, is that contrary to what some think, there is something we can do.  Recent spikes in the price of gas, have  urlLink shown  the American consumer\'s willingness to change to more efficient vehicles.  Hybrid vehicles produced by auto makers are showing an increase as a result:   The Prius...is attracting U.S. consumers after regular-grade gasoline prices rose 39 percent on average this year, to $2.051 a gallon. General Motors Corp. and Ford Motor Co. will start selling their first hybrid models this year, seven years after Toyota introduced the Prius.  With higher fuel prices people now have the will to demand more efficient vehicles.  Where there is a will, there is a way.  I propose that the U.S. begin instituting an increased gas tax, to be phased in over the next decade.  And that it do so very publicly.  An increase of, say, 1 cent per gallon per quarter year, would give a lasting, and growing incentive for consumers to demand more efficient vehicles.  Fuel efficiency on such vehicles is so much greater than what is in use today, that we may in fact be able to inspire such a shift that would have virtually no negative effects on the economy.  This would be due not only to the fact that consumers will consume less, but also that the higher prices will not be going to other governments that pump the oil, but rather to our own.  Even those cynical about the efficiency of government spending are unlikely to dispute the fact that the benefits are much greater with gas revenue in our own government\'s hands.  Looking roughly at some  urlLink fuel economy data  it is easy to see that the most efficient cars are roughly twice as efficient as the least efficient.  This means that, especially as the technology improves, it may be possible to double the overall efficiency of cars on the road.  Extrapolating that out to the previous observations, and we see that it will: free up 3/4 of a percent of GDP to be spent somewhere other than on gas, every year; reduce CO2 emissions by about 10% overall; and that it will shrink the current account deficit by about 10% as well.  All are pretty major improvements, with the exception of the deficit, yet even that is a major plus as any decrease in it right now is something that is definately favorable.  Best of all, due to the fact that the price will increase by less than the increase in fuel economy, as an end result we may see an overall decrease in the cost of fuel/mile.  Some deeper analysis will definately be required, but a steadily increasing gas tax is most definately something that should be considered due to the wide range of positive benefits and comparatively few detriments.  Your thoughts are appreciated.  Edit: I just realized that the corrective effect on the current account deficit IS still quite large, at somewhere around .5% of GDP.  The fact that it is only 1/10th of the deficit now is only due to the fact the deficit as a whole is so large.  The improvement in the balance will be relativly constant as a % of GDP, as opposed to a % of the deficit.  '),(13562,3343256,'22,June,2004','2004-06-22 00:00:00','   I found  urlLink this  a week or so ago, and it\'s finally working now.  Likely it got linked to too much recently and so was bogged down.  Anyway, it\'s an interactive map of the USA, showing breaking down of what languages are spoken where.  Rather interesting, if you are into that sort of thing.  The analysis for where I live is interesting.  Here are a few notable statistics:  English outranks other languages by a factor of 10 to 1.  156 people speak a language I have neither ever heard or, nor likely can pronounce: Gujarathi  Only 24 people speak the Scandinavian languages.  Which is a pity, as right now I am taking the time to learn Swedish.  But, I guess that means I am consistent in my pursuit for totally irrelevant knowledge. :P  What is interesting about where you live?  '),(13563,3343256,'21,June,2004','2004-06-21 00:00:00','   I finally made an appointment with an oral surgeon to have my wisdom teeth taken out.  All four of them.  Guess how much it\'s going to cost.  Here\'s a hint: Look at the title of this post.  Ouch.  Fortunately, this is just the high end.  Also fortunately, is I have the money for it, although it\'s going to eat up 75% of the wad of cash I had sitting around that was helping me make it from month to month.  Well, at least it\'s done now.  I\'ve been putting this off for about 5 years.  No more things to remember to initiate on my part.  Two months from now I head in, and have them yanked.  Oh!  And good news, they use a \"Twilight Sleep\" on me while they do it.  I get to be knocked out!  That will certainly  make the procedure itself more bearable.  I wonder...will they try to collect the money before I am drugged up, or after.  I think I shall have to be careful.  So, anyone else ever have their teeth taken out?  '),(13564,3343256,'20,June,2004','2004-06-20 00:00:00','   Well, right now is the summer solstice.  Yes, right this very instant.  At least this instant that I am posting this.  What is the solstice?  It\'s the time of year when the sun reaches it highest point in the sky.  For those in the Northern Hemisphere at least.  It\'s the lowest point for those in the Southern.  If you happen to live on the Tropic of Cancer, the sun should have been directly overhead.  It is, also, the longest day of the year, as a result.  Then again, I somehow doubt I have any readers in Mexico, the Sahara, Saudi Arabia, India, China or Taiwan.  What the summer solstice means in astronomical terms, is that the North Pole is now pointing in the direction of the sun.  No, not directly, that would just be weird.  In order for that to happen the Earth would have to have a tilt of it\'s axis 90 degrees off in relation to its orbit around the sun.  It would make an interesting SciFi story at least though.  With the Earth having a 90 degree tilt, the North Pole would have the sun be directly over head.  All day.  Can you say hot?  No ice caps here.  At least not during summer.  The south pole would be in total darkness the entire day, along with everywhere else south of the equator.  It would make for one wicked weather system.  I wonder if any planet tilted like that could naturally evolve life.  Or rather, intelligent life.  I suppose such life would have to develop underground.  That would be the only way to survive long enough to wait out the cold, cold, winters, and hot, hot, summers.  The Planet of the Dwarves.  Has a nice ring to it.  Let\'s see, what else would this planet have.  Ice caps?  Not permanent if they were at the poles.  Based on an article I read on Mars, I would suspect we might find permanent glaciers between 10 and 20 degrees north and south of the equator.  Why there?  Well, the equator will be the spot on the  PotD  that has the sun overhead for the least amount of time.  It would pass directly overhead twice a year though, on the equinoxes, just like on Earth.  The rate that the Sun would move is the fastest during that time.  On Earth, the days get shorter after the summer solstice.  Slowly at first, then more rapidly, until the autumn equinox arrives.  Then the rate of change in the length of days slows down, until the winter solstice, when the process reverses and the days start to get longer.  The same thing happens for any planet, regardless of its tilt.  On the Planet of the Dwarves, this would cause the Sun to be overhead the poles for an extended period of time.  As the days get further away from the solstices, the sun will move towards the equator, but will quickly brush past.  This would make being near the equators the best place for permanent ice caps to develop.  Although not actually on the equator, as it would still get sunlight all year round, even though it would be essentially a days long sunrise/sunset at that solstices.    So, with ice caps 10-20 degrees away from the equator, the sun will spend a very short time each year directly overhead. Most of the ice melt will occur at this time, and it will happen twice a year at the equinoxes.  The Sun will then move towards the poles, dropping in the sky from the point of view of the ice caps until its only 10-20 degrees above the horizon, although never setting.  During this time the ice will be able to stabilize and add to itself enough to survive the return trip of the Sun.  As the sun passes overhead again, the days will get shorter and shorter, until a extended night sets in.  During this time any buildup will be completely protected from melting.  Hmm...ok, so I got a little off subject, but not really.  I like to ponder strange things, get used to it. :P  Update: Here is a nice  urlLink link  showing the movement of the sun over the course of a year.  '),(13565,3343256,'19,June,2004','2004-06-19 00:00:00','   I just got finished watching  urlLink Patch Adams , and the movie inspired me to write this post.  Patch Adams is an ENFP.  I could explain what those individual letters mean, but that doesn\'t explain what an ENFP is, nor can it explain any other type.  The personalities are too complex for that.  ENFPs are those energetic people with vision.  That was readily visible in watching the movie about Patch.  They can see things, and dream, better than any other type, in my opinion.  The pages I have read about ENFPs state that they have visions to change the world, or at least parts of it.  They don\'t let can\'ts and not possibles deter them.  They see beyond that, to what is possible.  They are essential to the advancement of the human race.  And I don\'t mean in economic terms, although it can be in that way.  Rather they are those people out there that seek to find the greatest potential in everything and everyone, and then try to find a way to achieve it.  With all that enthusiasm and energy comes their weakness too.  An ENFP that is discouraged, depressed, repressed, or oppressed is a sad sight to see.  Their visions turn dark, they hide from other people, and they give up.  If you\'ve seen the movie, you saw that happen to Patch after his girlfriend Carin died.  Everyone needs a friend who is an ENFP.  But even more so, every ENFP needs a friend.  Someone who is willing to listen to, and even participate in their visions.  Someone who won\'t let them just quit, or hide, when things get to tough.  Someone who is there for them.    ENFPs can\'t work on their own.  They need other people to make their visions become reality.  Take Patch.  He didn\'t try to think in a small cubicle, by himself.  He went out and got involved doing what he wanted.  He then recruited others, even if they resisted, to join him.  He didn\'t start his clinic alone, but got others to work with him, and share his dream.  And today, due to his work and others, it\'s becoming a  urlLink reality.   Are you up to the challenge to help the world, one ENFP at a time?  The future of the human race is counting on you.  '),(13566,3343256,'18,June,2004','2004-06-18 00:00:00','   Yesterday I discovered something rather interesting about fusion.  I already knew the basic process of protons fusing together to make helium and release energy.  (If you thought hydrogen fused together to make helium, you are correct, as hydrogen is simply an atom with only a single proton for a nucleus.)  What I didn\'t know, was how the actual fusion process took place.  Apparently, before the advancement of quantum physics, classical physics could not explain how it was possible for nuclear fusion to take place.  This was due to the fact that the sun was not sufficiently massive enough to make the protons (which are positively charged, and as such repel each other) fuse.  Reading a  urlLink website  yesterday, suddenly made it make sense.  Everything in the universe is not made up of particles, but is rather in a state that is both particle and wave.  The wave portion is a probability wave that determines where an object is.  Instead of having a definite positions, it  could  be here, or it  could  be a little ways over there.  It turns out, that although the intense pressure and gravity of the sun is not able to overpower the repulsive force of two protons, it  is  powerful enough to push them close enough together so that their probability waves overlap to an extent, so there is a small chance that at any one moment they will both be in the same position.  And when they are, they fuse.  Now this got me thinking about gravity.  Just looking at what I just described it looks similar to what gravity does.  Two objects are mutually attracted to each other.  What if gravity works on the same principle?  What if the probability waves of matter interacted in such a way, that they had an  additive  effect.  Imagine two objects in space.  Both have probability waves that describe their location.  Good chance here, small chance there, really small, but still a chance that its way over there.  Now, as these two objects got close to each other, they would start increasing each other\'s probability of being in the space between the two of them.  As a result, over time, and more quickly as the mass increased or distance decreased, the objects would move closer.  Sounds just like what Einstein\'s gravity does, doesn\'t it?  It needs a lot of work though I think.  For one, I realized while writing this that photons would be affected by this, but could not add to the effect.  The other problem is that this seems so blindingly obvious to me, as someone who knows very little about physics, quantum or otherwise, that someone would have already thought of this and crunched the numbers.   Update:   Another big problem, I forgot all about time dilation.  Duh me.  '),(13567,3343256,'18,June,2004','2004-06-18 00:00:00','   No, seriously, it\'s a very interesting topic, as long as you have a good teacher.  Fortunately, I did in the book  urlLink The Fabric of the Cosmos: Space, Time, and the Texture of Reality  by Brian Greene.    In it, he explains in very down to Earth terms how the universe works, according to quantum physics.  My aroused my interest the most, was an example of the fact that observing something changes its outcome.  Now, most people nowadays will have heard something to that effect regarding quantum physics.  What most people  don\'t  know however, is that if you erase that observation, the particles you observed return to the state they were in before you observed them.  Freaky, huh?  Let me give an extremely simplified explanation.    Photons are shot from an emitter, into a beam splitter, towards two mirrors, which both bounce the photons back to the same spot on a detector.  This creates an interference pattern, due to the wave-like nature of the photons.    To understand an interference pattern, just think of a wave in a tub of water, it is split in two, bounces off the sides, then converges again further along the line.  Where it converges there are points where two crests meet, making the wave twice as high, where two troughs meet, twice as low,  and  points where a trough and crest meet, canceling each other out.  This pattern of highs, lows and nothings is the interference pattern.  Photons (light) do this because it is made up of probability waves, which interfere with each other when it is split, and it makes a pattern of alternating light and dark on the detector.  Attempting to detect whether a photon goes towards one mirror or the other after going through the splitter, will cause the interference pattern to disappear.    However , if you were to set up a splitter along each path, then bounce the new paths along to the same detector, thereby making it impossible to know which path it came from, you can restore the interference pattern. *  This means that it is  not  the act of observing that is causing the change.  Disturbing the photons is not what is causing the interference pattern, but rather the fact that we  can  observe which path they took that is causing the pattern to disappear.  Once the observational data is destroyed, completely and irretrievably, even in principle, the interference pattern returns even though we are still messing with the photons.  Freaky.  * EDIT: A little clarification on the second to last paragraph, where I bolded it.  When you set up a splitter along each path it is a different kind than the first splitter.  The first splitter causes the photons to go along one of two paths, the bolded sentence splitter actually creates two entangled photons, one which continues along the original path, and another that is sent along to the detector.  Entangled photons, for all intents and purposes, act in unison as if they were the same photon.    If a detector is set along each path, to determine which direction the photon took, the interference pattern is destroyed.  If however the photons are directed from the splitters along each path to the same detector, instead of separate ones, then it is impossible to determine which path the photon took, then the interference pattern is restored.  Hope that\'s clearer.  '),(13568,3343256,'16,June,2004','2004-06-16 00:00:00','   Browsing  urlLink Marginal Revolution  today, I noticed Tyler Cowen has a post up that nicely compliments my post from  urlLink earlier  today.  Some of the figures in his article differ from those in my data, however that is likely due to how I choose to interpret \"Foreign official assets in the United States\" as central bank holdings in  urlLink my data source .  His analysis is considerably more optimistic than mine, as mine seems to fit largely into his last explanation for why foreigners are buying so many Treasuries:   7. They are just plain, flat-out stupid. I call this the uh-oh scenario. They won\'t stay stupid forever.  More the uh-oh part than the stupid.  Of additional interest to me is that Tyler seems to have  urlLink once again  changed the attitude of his post from a negative view, to a more neutral one.  How can I tell? Check the link name, it\'s based on the title of a post when it is posted.  Hmm, it appears this is another of his posts  urlLink I blogged about .  I think I need more blogs with trackbacks to visit, as all of my site hits seem to come either from my own pimping, or from Marginal Revolution trackbacks.  Tyler might think I am stalking him.  Although he has piqued my interest.  His habit of changing post titles, overall competence, and  urlLink other indications  suggest to me that it is possible he may be  urlLink INTP .  Hmm, 7 links in one post, I think that\'s a record for me.  '),(13569,3343256,'16,June,2004','2004-06-16 00:00:00','   Ha, a right-winger saying that, now that\'s a sight.  After all, what\'s illusionary about it?  Over a million jobs in the past few months, phenomenal growth, stock market is up significantly verses a year ago.  That\'s the current picture that most people are seeing.  Here is what I am seeing however.  During the boom years (1998-2000) we saw an increase in both imports and exports to the U.S.  However, imports increased by approximately twice as much as exports did.  During the subsequent recession and slow recovery (2001-2003) we saw both imports and exports decrease at first.  Towards 2003 however, imports had fully recovered, while exports were still down from the highs in 2000.  For that to be able to happen, we must have money coming in from somewhere to balance the gap.  For that we look at the financial account during that period.  Foreign investment increased rapidly during the boom years, and tapered off somewhat afterwards, but not anywhere near to where it had been in 1998.  It was in fact still almost double in 2003 of what it had been in \'98.  It\'s composition had changed dramatically however.  Most of the inflows during boom times were into securities, direct investment, and private debt.  Comparatively little was in public debt (Treasuries).  In the following recession and recovery, inflows shifted heavily into public debt, comprising more than half of the total inflows.  Outflows were comparatively unchanged except for a large reduction of private lending to foreigners.  As a result, we went from a large trade deficit at the end of the boom times financed by foreign investment, to a larger deficit at the end of the recovery period financed in large part by foreign lending (foreign purchases of Treasuries).  About 2/3 of this debt is being financed by private foreign investors, the other 1/3 by their governments.  That is a cause for worry in my opinion, especially as interest rates are set to rise soon.  As rates rise, the value of the debt drops.  If foreign investors decide they do not want to continue taking the risk we will end up seeing a large outflow very quickly.  This is not even counting what foreign governments will do, as they will want to protect their investments as well.  In fact, China is already moving some of its large assets out of dollar assets and into yen as reported by the Wall Street Journal on Monday.  What really worries me however, is something I wrote about  urlLink before .  After the boom times, the trade deficit was large, and the economy went into recession as a result.  Normally this should have caused imports to drop, until exports could be increased to pay for them.  Tax cuts were quickly passed however, which allowed consumers to keep importing large amounts.  Between 2000 and 2003 the drop in government revenue (aka tax cuts) amounted to $243 billion.  Had there not been any tax cuts, the recession would have been much worse, however the private sector would have shifted its behavior to adjust to the times.  Instead, because a significant part of the deficit was a result of tax cuts, private sector behavior remained little changed, hence the result that imports recovered quickly without exports.  So, while the administration has said that it believes market forces should correct the current account deficit, its very actions have directly fought against any correction!  Now the current state of things is looking decidedly unstable.  As  urlLink Buttonwood  put it in an opinion article I read yesterday:   In his idle moments, Buttonwood wonders what would happen to the American economy were property prices to tank, at the same time that oil prices soared and the fiscal stimulus ran out. It wouldn\'t look pretty, that\'s for sure.   His article (worth reading) talks more about domestic debt, and how so much more of it is vulnerable to rate changes today.  Combined with foreign debt, and no wonder investors are nervous about stocks.  Combined with my belief that the economy never got a chance to naturally correct its trade imbalance, and I am keeping most of my assets out of U.S. stocks.  My primary sources of trade data is  urlLink here , and my budget data  urlLink here .  '),(13570,3343256,'16,June,2004','2004-06-16 00:00:00','   A strange request, no?  Well, a little background first.  I work at the family bakery with my dad.  A little over a year ago my dad told me that he could only afford to pay me part time.  Finances were tight, and things weren\'t looking good.  So, as a result my work schedule dropped from 4 days at 10-12 hours each, to 2 days.  My income dropped accordingly.  From that point I went from having a nice surplus each month that I put into investments, to having just barely enough to break even each month.  If I could go back to work full time now though, I wouldn\'t.  I have come to greatly value my free time.  I have used the time to learn many new things, ranging from personality types, to quantum physics.  This current setup is almost like a dream for any INTP.  The problem though, is now my dad has put the bakery up for sale.  I am looking at, not 2 days a week of work, but 0.  That means new job.  And damnit, I don\'t wanna.  What can I do though? I have about $1000 a month in expenses I need to cover.  Here are my options:  1. Get another job.  Not ideal, because a) I likely would get a lower pay rate than I am receiving now, meaning I couldn\'t work just part time.  And b) any job that I would be well suited for mentally, will likely require a college degree of some sort.  I hate physical labor, not because I am lazy, but just because I can feel my mind rotting from disuse.  I will have to do some searching, although I\'m not sure where to begin.  2. Get a roommate to help with expenses. Actually, I\'d rather work than do that.  3. Move to a cheaper place, as half my expenses are rent.  I might, except for the fact that I am renting from my sister, which means I have an excellent landlord.  Also, it\'s her policy not to increase rents on current tenants.  That means my housing expenses are immune to inflation.  4. Shop more frugally.  Ha! The only place to save money is likely on food.  Although I may be able to squeeze a bit out of there, as I\'m currently shopping at an above average price store.  5. Generate income from various sources.  This is the one I like, and it\'s where you come in.  With a bit of luck, and some work, I can get multiple revenue streams from such sources as my investments, this blog, eBay, and even perhaps an online game I am working on.  All of these are long shots for income, but I will never know if I can if I don\'t try.  So, if you like reading my blog, and you are financially well off, feel free to donate.  Now, I don\'t really expect to get much, or even really anything, but as I said, I won\'t know if I don\'t try.  So...help me stay unemployed!  Click the appropriate link on the sidebar to help me in my quest to keep blogging for you people.  '),(13571,3343256,'15,June,2004','2004-06-15 00:00:00','   Yes, you remember  urlLink Lakisha , don\'t you?  Well, just the other day she initiated communication with me through  urlLink eHarmony.com .  To respond though, costs $50!  I don\'t know about you, but right now I\'m not really willing to spend that much of my own money just to initiate a conversation.  I am however willing to spend your money.  Huh?  Look to the right.  I\'ve discovered Amazon\'s Honor System.  It lets people send me money through the website, without me having to handle any payment information.  Do I expect anyone to give me anything? No.  However, if anyone suggests I should actually try to look for a girl, I now have a link to give such people, so they can put their money where their mouth is.  Clever, eh?  So, donate, and help find me a girl!  Oh, and if you donate, but I don\'t collect enough to subscribe, the Amazon system lets you get a refund of your money for up to 30 days after you give. I\'m going to leave this up a maximum of two weeks.  If I don\'t have enough by then, I\'ll take the pay button down and let you all know to request refunds from Amazon (if anyone actually donates anything that is).  '),(13572,3343256,'15,June,2004','2004-06-15 00:00:00','   Praise their partner.  No, seriously, it works.  I used to joke that I helped break up two marriages.  It was an amusing correlation that I drew.  I start talking to a person having marriage problems, but is sticking it out.  Then later they reach the point that they break up with their spouse.  Now, I didn\'t actually blame myself for it, as in both cases the spouse (both men) was not being a good husband, even to the point of being emotionally abusive/manipulative.  More so it was as I said, an amusing connection.  However, it lurked in the back of my mind that maybe I did affect things somehow, and something I read recently solidified that feeling.  The book I was reading is called  urlLink How to Win Friends & Influence People .  Doesn\'t quite sound like a book to get ideas on breaking up couples, now does it?  Bear with me.  Part III, Chapter 1 is titled: You can\'t win an argument.  The reason for this, is that when you argue with someone, you are putting them on the defensive.  Even if you succeed in the logic, and defeat them, you still likely haven\'t converted them to your way of thinking.  Why? People are emotional beings.  As such:   A person convinced against their will, is of the same opinion still.  So, how do you convince people? The book recommends agreeing with them.  Once they are no longer feeling the need to defend their positions, they are open to your ideas.  Ok, so how does this relate to breaking up relationships?  Simple.  Most people that want to break two people up, will go to their friend and try to point out all the bad things about their partner.  This instinctively causes the friend to become defensive and try to rationalize and play down the things their partner has done.  Now, I\'ve been accused of being a nice guy, and as such I try to look for the good in people, and try to look for ways to help people work out their problems, where I can at least.  Looking back, what I now believe I was doing was giving my friends a chance to, for themselves, see the bad traits in their partner.  They\'ve always known they were there, they likely even told other people about those traits.  But usually those other people will likely agree with their friend\'s assessment, causing the friend to second guess, and make excuses.  With my habit of not attacking my friend\'s partner, and even trying to point out the good, my friend would finally have a chance for their brain to process the bad traits without being defensive.  It\'s an interesting theory at least.  Now, don\'t read what I\'ve written the wrong way.  First, I don\'t blame myself for breaking them up, as in both cases I believe the relationships were unhealthy.  Second, this technique will not work on a fundamentally healthy relationship, due to the fact that the negative traits must actually be present before they can be acknowledged.  Side note: One of said friends came up with the second half of the title to this post.  '),(13573,3343256,'12,June,2004','2004-06-12 00:00:00','   For those of you wondering how you can figure out your personality type, here are a few links to some tests.   urlLink A short one.    urlLink A medium one.    urlLink A long one.    urlLink A blog oriented one.   If you are unsure what exactly personality types are, or want more information, you can read a  urlLink short description  that I wrote, or check out one of the many sites below.   urlLink http://typelogic.com/    urlLink http://www.pitt.edu/~bav11/intro.html    urlLink Life Explore    urlLink Personality and Careers    urlLink The Personality Page  '),(13574,3343256,'11,June,2004','2004-06-11 00:00:00','   Ok, some of you have probably looked at the name I use to sign my posts, and wondered \"WTF does that mean?\"  Some of you already know me from elsewhere on the internet, and so recognize the name.  Others though have followed a link here and likely have no clue.  Well, at first I wasn\'t sure how to explain it, as I have a different explanation depending on how I know a person.  For those I know  IRL  I say Birdeye is the name of my favorite donut I enjoyed while growing up.  With the family business being a bakery, certain perks go along with it.  Including my dad being willing to bring home my favorite donut when he gets home from work...which is around the time Saturday morning cartoons started.  The donut itself is a chocolate covered raised dough roll, with a big dab of frosting in the middle.  For other people, namely those that I\'ve met through an  urlLink online community  the name has a different origin.  This site is actually where I invented the name.  It was inspired by a quote about my favorite character in the book series that the site is based around.  Shortly put, the quote mentioned how this character\'s eyes look like a \"bird studying a worm.\"  That gave me the inspiration to make the name Birdeye.  Incidentally, later I determined that this character, Verin, is an INTP.  But, what does the name mean here?  Well, being not overly attached to the idea of having to have a single explanation for what my name means, I\'ve created another one.  It\'s how INTPs like to look at things, all at once, from a distance to get a good perspective.  Getting a bird\'s eye view of the picture.  So, yeah, that works.  '),(13575,3343256,'09,June,2004','2004-06-09 00:00:00','   Well, I\'ve discovered how I can post pictures to my blog now.  I found out how while reading my copy of the  WSJ .  It\'s a lovely little program called  urlLink Hello .  Silly me, after I read the name, I realized there was a  urlLink link  explaining it on the main page of  urlLink Blogger .  The nice thing about this program is it\'s not restricted for use in blogs.  It\'s actually an instant messaging program that is designed to be able to rapidly share pictures with other people.  The blog part just works by sending the pics to a IM buddy named BloggerBot.  So, in any case, I can add pictures now.  And you can look forward to seeing some from time to time.  Oh, and I\'m using this post to test out the queuing feature where you can save a picture and have it automatically added to your next post.   urlLink      Didn\'t work quite like I had thought, but at least I\'ve figured out how to manually move the pictures around.  '),(13576,3343256,'09,June,2004','2004-06-09 00:00:00','   urlLink Google Answers  is a project of Google\'s that looks absolutely fascinating.  People can ask questions, and then others, called Researchers, answer them for a price.  Already I found  urlLink one question  for $200 for which I was able to find the answer rather quickly just by looking in material I already have.  I even responded with a comment.  If my answer is good enough, I just saved this guy $200 from having to wait for a Researcher to find the answer.  Unfortunately, they are  urlLink not currently accepting applicants  for new Researchers.  However, once they start again, I am most certainly going to apply!    Wish me luck.  '),(13577,3343256,'08,June,2004','2004-06-08 00:00:00','   I was inspired by a recent post by Tyler Cowen at  urlLink Marginal Revolution  to figure out how, and if,  urlLink personality types  relate to political leanings.  In my analysis I found some relationships that made perfect sense, and some that puzzled me.  For those interested,  urlLink here  is my source.  Since the quantity of types, and type combinations is huge, I will only be highlighting those that I found significant.  My source was a website stating the results of questions asked of about 1200 people.  Those people chose one of 7 options to describe their political leanings.  The options were: Democrat, Republican, Not Political, Middle of the Road, Liberal, Conservative, Libertarian, and Other.  Keep in mind while viewing the figures, that since those polled can only choose one of the 7 choices, it is entirely possible for any one category to show an increase in two polarly opposite political views.  The first that surprised me not at all, was that ISTJs are 41% more likely than the average responder in this survey to consider themselves Republican.  Reading from any of the descriptions around the web you can easily see why:   ISTJs...are hardworking.  They honor their commitments...They have a strong internal sense of duty...They place great importance on honesty and integrity and can be depended on to do the right thing for their families and communities.  All hallmarks of how Republican tend to think of themselves.  ISFPs also tend to depart sharply from the normal, being 30% more likely to be Democrats, and more than 40% less likely than average to say they are Republicans.  A shorthand way of describing the ISFP is as the \"Stop and smell the roses\" type.  Not shocking to see their general political views lean as they do.  For each full type I have calculated the percentages and their deviations from the survey\'s \'normal\', however due to the fact that the sample sizes are so small for each type, the results for most tend to be skewed.  The remainder of the results are those from partial types, that include more than a single full type.  Looking broadly, just one letter at a time, the one that stands out the most is the iNtuitive, who scored as 24% more likely to be Liberal than average.  As Liberals tend to take a larger world view than what they believe others take, it makes sense that the type that sees the big picture when dealing with things would correspond with that political position.  Intuitives also score lower for both political parties, showing that they are less likely to be willing to label themselves so narrowly, which further corresponds to their nature.  Looking at the four main  urlLink temperament  types the biggest deviation appears in the  NT  type, showing them as 44% more likely to call themselves Liberal than average, mostly for the same reason as Ns in general.  Another big deviation is in the  SP  temperament, which scored as 35% more likely to declare themselves as Middle of the Road.  As  SP  types are \"If I can\'t see it, I don\'t believe it\" it is likely that this is a result of politicians failing to fulfill their promises.  In other words,  SPs  never actually \'see\' an political path that they can believe in.  The  IJ  subtype is the one that gave me the strangest results of all the types.  They scored as almost 24% more likely to be Republican and 13% more likely to be Conservative than normal.  The corresponding scores for Democrat and Liberal are either flat, or slightly down.  What could cause this political leaning, I am honestly unsure.  The prize of the \"Most Different\" goes to the  EN  types however.  These types scored a whopping 60% more likely to consider themselves Liberal, and eighty and seventy percent more likely to have chosen Other or Libertarian, respectively.  Due to the exceeding small sample size though, the last two results are less significant.  The Liberal title however fits very nicely with the iNtuitive tendency in general to lean Liberal, for the same reasons as stated above.  They also were about a third less likely to choose either the Middle of the Road, or Not Political options.  In general they appear less likely to just let things just pass them by, and so take a stand in large numbers.  It took me probably 4 hours to compile and calculate all these statistics from my source, but it was worth it.  The results were incredibly intriguing to me.  Hope you found them interesting too.  '),(13578,3343256,'04,June,2004','2004-06-04 00:00:00','   Specifically, the  MBTI  that I refer to here regularly.  This typing system is fairly useful for understanding the different \'types\' of people out there.  To call them personality types is a bit of a misnomer however, as culture, upbringing, and other environmental factors all contribute.  The types are excellent however in describing how our minds work, in terms of taking in information and making decisions.  More so than anything else I\'ve read about.    The four preferences:  Extroversion vs. Introversion Sensing vs. Intuition Thinking vs. Feeling Judging vs. Perceiving  The first preference does not mean shy vs. social.  It is simply the default and most comfortable way that people act.  Generally speaking introverts get their energy from solitude, while extroverts get it from other people.  The second preference has to do with the forest and the trees.  Sensors like details, and so are good at focusing, but also getting stuck not seeing the forest for the trees.  Intuitives like the big picture, and dealing with issues more abstractly, but can lack practicality by not seeing the trees for the forest.  Both these are considered \'Perceiving\' functions.  The third preference is likely the easiest to understand.  Thinking tends to deal with logic.  Feeling does deal with emotions, but more so it refers to a persons values.  Both of these are referred to as \'Judging\' functions.  The fourth preference has to deal with which of the second or third preferences is extroverted.  If it is Perceiving, then the second preference (whether Sensing or Intuition) is extroverted, making the person laid back and flexible.  If it is Judging, then the third preference will be extroverted, making the person more focused and strict.  These four preferences actually break down into 4 different functions, but that is more complex than I want to get into right now.  Likely more so than you want to hear either, so I will stop here.   Update:  I now have  urlLink a post  up explaining the function breakdown.  '),(13579,3343256,'04,June,2004','2004-06-04 00:00:00','   I arrived home yesterday, and found her just lying on the floor, not willing to move.  Adding to the fact she was in the same condition that morning, I decided I couldn\'t wait until her appointment today to have her put down.  So, I called the vet, and got them to let me just drop her off, so they could put her to sleep when they had the time.  Expensive place, $65 to have her put down and to bury her.  But, what can you expect from a vet in a town called \'Richland\'?  I miss her, and it still hasn\'t fully sunk in that she is gone.  '),(13580,3343256,'04,June,2004','2004-06-04 00:00:00','   Poor Tiger.  She\'s 15 years old, and has stopped eating.  Took her to the vet today.  Two hours, and $136 later, they told me she has kidney failure.  She can\'t be cured, but might be able to be gotten back to normal via periodic fluid injections to help clear stuff out.  That\'s what I am trying right now.  They did a fluid injection (which Tiger hated) and now she is back home with me.  We go back Tuesday to see if it helped at all.  If not, I\'m having her put down.  Sadly, so far, Tiger is acting no better. :(  '),(13581,3343256,'03,June,2004','2004-06-03 00:00:00','   ...you buy a game that has been on the market for a year, and it still barely runs.  I just bought  urlLink Warcraft III  off of eBay the other day.  It took a defrag, and shutting off all my  IM  programs in order for there to be enough memory for it to run.  Fortunately, it at least  does  work.  Fun game too.  So far I\'ve just completed the first few Human chapters.  Currently on level 6...the game actually has me killing townspeople so they don\'t turn into the undead.  Amusing.  I lost the first time through this level, going to give it another go tonight.  Just another distraction to take my mind off stuff.   Update:  Turns out much of the problem was some scratches on the CD.  After sticking in the expansion, which seemed to work fine, I discovered the problem wasn\'t my RAM at all.  So, I contacted the seller, and he let me ship it back to him for a refund.  '),(13582,3343256,'03,June,2004','2004-06-03 00:00:00','   A few weeks ago I made one of my  urlLink first posts  on this plan.  Just today I found it finally making the rounds, when I saw Tyler Cowen  urlLink post  about it.   If the plan interested you at all, follow the links back to see what he, and others wrote on the subject.  I don\'t have anything really new to add to what I had written before.  So why am I making another post about it?  Simple, to exploit the trackback function to get some people that care about economics to read my other stuff.  And to see that I reported this plan almost 3 weeks ago.  But, as new kid on the block, can\'t really expect anyone to have seen it. ;)  Hey, at least I\'m honest my motives.  '),(13583,3343256,'02,June,2004','2004-06-02 00:00:00','   I\'ve recently added all the following to my blog:  Product Recommendations - I\'m such a sell out Guestbook - Say Hi or something Mini Poll - Vote!  So, yeah, that\'s about it for now.  More stuff later for sure.  Especially as I am learning HTML.  Oh, and make your presence known here people.  Somehow I\'ve gotten about 20% of all my site hits in the last few hours..and I have no clue why.  Likely some bots ran across my blog, but if you are real people, make sure to use one of the above features so I know you  are  real.  '),(13584,3343256,'01,June,2004','2004-06-01 00:00:00','   Can\'t blame you.  An article in the  WSJ  today states:   A good historical novel...transports readers, as though by a time machine, to a world where not only the curses, hairpieces and grog are authentic but also the beliefs and prejudices of ordinary people.  Historical fiction lets you look back in time through the eyes of its participants, instead of through the eyes of a historical.  The great thing about the world today is, thanks to the internet, we no long have to wait years or decades for someone to write about how somewhere else in the world really was like.  We have blogs.  For those that have not wandered around at all on my little blog, I point your attention to the upper right area, where there is a list of other blogs.  The (currently) second from the top one is called  urlLink Iraq the Model .  Clicking here will bring you face to face with 3 brothers who live in Baghdad, writing about their views and experiences, and hopes for their country.  Reading that blog, and the other Iraqi blogs they link, you can get a more personal view of events in the world.  In addition, all but the most stolid America-hater I suspect will be moved by the posts made there to, if they were not already, become more interested in seeing Iraqis win, than seeing America \'lose\'.  So, check it out, its like historical fiction in there here and now.  Without the historical or fiction parts of course.  '),(13585,3343256,'01,June,2004','2004-06-01 00:00:00','   Recently, I discovered an  urlLink article  at  urlLink Marginal Revolution  on whether there is any truth to the idea that social transfers hurt growth in welfare states.  Most people\'s intuition says yes, this article challenges that intuitive conclusion.  To put the paper\'s conclusions\' simply, the answer is no, there is not a significant loss to GDP.  How can this be so?  The answer is in the form of a variety of reasons.  First is that models used to estimate social spending costs, are not actually practiced by any states today, or that the estimates are extrapolated beyond that which has actually be practiced.  Second, believe it or not, but welfare states have a more regressive tax policy that is more efficient in terms of economic growth than in non-welfare states.  Additionally, disincentives to work are designed in such a way as to prevent a negative impact on GDP growth.  The third main reason is the core of why the previous two are true.  It is simply that whenever a bad choice is made, it becomes very obvious very fast in a high tax welfare state.  This is realized on both the tax side and the recipient side.  On the taxing side, it gives the example of the double taxation of dividends.  Welfare states tend to be much more capital friendly, and do not tax businesses twice as a result.  The United States on the other hand, has only recently reduced the double taxation.  As a result, welfare states are able to attract capital much more easily that one would assume.  The overall mix of taxation is more efficient to growth as well.  Higher income tax along with higher consumption taxes result in a broader income base, while also tending to tax those who are receiving the services, so the next transfer from one income group to another is reduced.  On the recipient side there are multiple external effects from social transfers.  Three areas that the paper touched on were retirement, health, and education.  The externalities of health and education are fairly obvious, as the better people perform at their jobs will of course has a positive effect on GDP.  As for retirement, the paper theorized that by removing older, less efficient workers from the workplace, the overall worker productivity improved.  This view sits well in my mind, as many businesses are known to discriminate against older workers for that precise reason.  All in all I found this paper quite informative, although I still have my reservations.  I have, for a while, observed that the least efficient economies in the world seem to be those that have both larger populations and larger welfare states.  As a result, there can be major cultural differences of opinion on what should be provided by the state, within these states.  The point that the article made about democracy creating a form of cost control is certainly evident in the political history of the U.S.  In 1994, when faced with the possibility of a major shift in the structure of health care, there was a major revolt against the Democratic party, from which it has yet to fully recover.  Essentially a system that was perceived to be inefficient was blocked before it ever could take effect.  Additionally, if there is no loss to GDP, why do welfare states tend to have a noticeably lower GDP per head?  Note, this does not mean they have a lower standard of living.  Such a comparison would be much more difficult to quantify.  Also, states that have a lower GDP tend to benefit from being indirectly subsidized by those with higher GDPs, sometimes in the form of adopting another\'s technological advance, sometimes by gaining by having an economic model based on exporting to those \'richer\' countries.  All in all, I agree with what I see as the author\'s conclusion: With democracy forming a kind of cost control, a country can increase the size of its welfare state, and still remain efficient as the political trade winds will blow more fiercely as reality makes itself evident.  In other words:  To each their own.  '),(13586,3343256,'01,June,2004','2004-06-01 00:00:00','   Took them a little while, but they found someone that\'s compatible:   Lakisha   Nice name   Lakisha\'s friends describe her as:  Sweet  Affectionate  Modest  Good listener    Sounds like someone I could get along with.    The most important thing Lakisha is looking for in a person is:  A good heart and a kind personality.    Hmm...my friends I think would describe me as that.  Getting better.   Three of Lakisha\'s best life-skills are:  Entertaining in my home  Communicating my innermost thoughts and feelings  Being a good friend and companion     Oh dear, they\'ve found me an ESFJ. *shudder*   The things Lakisha can\'t live without are:  My car  T.V.  Music - any kind  Pet\'s  Friends     Can\'t live without T.V.? And she was starting to sound so good.    Except for that ESFJ thing, of course.  '),(13587,3343256,'31,July,2004','2004-07-31 00:00:00','   Guest blogger Bryan Caplan over at  urlLink Marginal Revolution  has an  urlLink interesting post  on a  urlLink paper (pdf) he wrote.  I finally managed to get around to finish reading the paper today.  It is definitely worth a look for anyone interested in economics.  Essentially it works like this:  1. Good policies cause good growth.  2. Good growth causes good ideas.  3. Good ideas cause good policies.  Note that the ordering of which is first is largely arbitrary on my part.  Step 3 directly affects step one, as it is a continuing cycle.  In this context the ideas are the public\'s perception of what works.  Step one is pretty intuitive.  Step three is also intuitive, as if a public supports a good idea it will likely be implemented or sustained as a policy.  Step two is more complex, as there can either be a positive or negative feedback at this step.  Basically, if there is positive feedback then when there is good growth the public recognizes it and encourages the policies that created it to continue.  The same is true if there is bad growth.  The people believe the ideas  should  work, they just need to be implemented more aggressively, so bad policies can also receive the public\'s support.  A quick note on terminology.  Positive feedback means self sustaining, not producing good results.  Negative feedback means self defeating, and does not refer necessarily to bad events.  Point in fact, there can be negative feedback, instead of positive, in the step two above.  An example of this is that if there is good growth, the public decides to change it\'s direction, perhaps to implement more \'equal\' policies, and so causes economically worse policies.  This then causes slower growth.  If growth slows enough, the negative feedback kicks in again, and the ideas of the public about what is good policy changes to cause better policies.  From this, economies that have a negative feedback in step two (the only step that can have negative feedback, the other two always have positive, meaning self reinforcing) always tend to result in having an overall mediocre performance.  When things get good, growth is diverted away for other purposes, when it gets bad, the people  urlLink become more open to change .  For economies in with positive feedback in all three steps, these are those that are stuck in the \'Idea Trap\'.  When things are good, they stay good.  When things are bad, they stay bad.  Only random shocks apparently can change the state an economy is in, as there is no self correcting mechanism for those with positive feedback from growth to ideas.  Overall, interesting concept.  '),(13588,3343256,'28,July,2004','2004-07-28 00:00:00','   Or at least google\'s search rank.  Currently a search for \'intp\' will yield my blog as the 27th listing.  Just a few weeks ago I was somewhere in the 80s.  On more specific subjects, I rank even better.  Currently 11th, up from 12th a few days ago, for searches for  urlLink Robert Shrum .  Not that I know much about the guy other than that one post.  Or really care.  I used to get some hits from my  urlLink personality and politics post , but that appears to have dropped off google\'s radar.  Ah well, times change.  So, anyway, I\'m just happy.  Higher rankings = more hits.  It\'s been a while now since I could visit my blog and see the hit counter not move at all.  Rare now are those days.  Yay!  '),(13589,3343256,'28,July,2004','2004-07-28 00:00:00','   Browsing  urlLink Mahalanobis\'s  archives of old posts, I stumbled upon this  urlLink Morality Quiz .  Here are my results:   Your Moralising Quotient of 0.17 compares to an average Moralising Quotient of 0.22. This means that as far as the events depicted in the scenarios featured in this activity are concerned you are more permissive than average.  Your Interference Factor of 0.00 compares to an average Interference Factor of 0.13. This means that as far as the events depicted in the scenarios featured in this activity are concerned you are less likely to recommend societal interference in matters of moral wrongdoing, in the form of prevention or punishment, than average.  Your Universalising Factor of 0.50 compares to an average Universalising Factor of 0.37. This means you are more likely than average to see moral wrongdoing in universal terms - that is, without regard to prevailing cultural norms and social conventions (at least as far as the events depicted in the scenarios featured in this activity are concerned).  Their results are largely consistent with how I think of myself.  If you intend to take the quiz yourself, you should stop reading this post first, as my reasoning may temporarily influence your own.  Mostly I tried to look at the questions based on whether a person was internally consistent.  I believe what I believe, but don\'t believe that other people should be held to my standards or judged by my beliefs.    I  do  however believe they should be accountable to their own values.  Few things irk me more than people violating their own stated beliefs and values.  If you don\'t intend to act in a certain way, don\'t pretend that you will.  In one example from the quiz, it appears I am definitely in the minority in this respect:   You judged the failure of the son to visit his mother\'s grave once a week as he promised to be wrong. 12% of the people who have undertaken this activity to date agree with you.  In that same light I believe that people should not judge others just because they have different values and priorities.  I am tired of those who hate people, or entire cultures, who share different values than themselves.  As such they develop a strong hatred towards those people or cultures that seems to make them look for everything wrong with their beliefs.  They then readily accept any negative suggestion and dismiss out of hand any positive result of those beliefs.  No way of doing something is perfect.  Some ways are distinctly better than others.  Promoting hatred however is no way to get people to look at things differently.  Hatred creates stubbornness, and makes people resistant to change, not more accepting to it.  '),(13590,3343256,'27,July,2004','2004-07-27 00:00:00','   As most people who get around on the internet will remember, a couple or so weeks ago, an amusing little video was posted making fun of Bush and Kerry. The site that made it, and where you can find the video if you are unfamiliar with it, is called  urlLink JibJab . Basically it mocks both of them, and gives a very good, short example of why Americans always seem to consider their choice of President as the lesser of two evils.   The way most people found this video was through links from blogs. A blog somewhere likely got it from being subscribed to JibJab\'s newsletter, then put it on their blog. As happens in the Blogsphere, other bloggers visited that site, found it funny, and linked it from their own. In likely a very short time this link made its rounds to the biggest of the blogs, where it quickly spread to just about everyone on the internet that doesn\'t live under a rock (It was linked 3 times on a message board I visit, and was in many of the major blogs I visit.)   What is interesting though is that this reminded me of what I read in the book  urlLink Linked: How Everything Is Connected to Everything Else . This book is about networks, and how things, well, link together. From what I remember, there are two parts to a linked network. Links and nodes. Links connect nodes to each other, and the number of links a node has is determined by its fitness. It\'s fitness is determined by essentially how good it is. Nodes tend to attract links according to what is called a power law.   Blogs provide a good example of this. There are a very few very  urlLink well known blogs  that get linked by hundreds or thousands of blogs, such as several of those within my BlogRoll on the right. There are also very many blogs that are linked to by virtually no one, such as this blog. In networking terms there are a few nodes with a large majority of links, and there are many nodes with very few links. This, is the  urlLink power law .   So what does JibJab\'s movie have to do with this? Well there is a remarkable connection the author,  urlLink Albert-LÃ¡szlÃ³ BarabÃ¡si , and his students discovered.  They discovered that networks can act just like a  urlLink Bose-Einstein condensation .  Here is a short snippet describing the condensate:   ...the condensation occurs when individual atoms meld into a \"superatom\" behaving as a single entity at just a few hundred billionths of a degree above absolute zero.  What was discovered while studying networks was, that if energy levels were assigned to nodes, and individual atoms were assigned to links, then:   ...in terms of the laws governing their behavior, networks and a Bose gas are identical.  Because they obey the same laws, they can both result in the same dramatic transformation into the condensate under the right conditions.  I believe, from what I understand on the subject, that that is what happened at JibJab when their \"This Land\" video was released.  The way the condensate happens in networks is that all the links point to the same single node, in a winner takes all fashion.  In terms of blogs, massive numbers of blogs out there all pointed to the JibJab site at once, creating the effect that everywhere seemed to point to this one site.  It took its toll on the JibJab site as well.  Completely swallowed their bandwidth.  At the time of this posting they have taken down basically everything on their site, and are using an offsite free service to host their video.  When a network condensates, it is a powerful powerful thing.  On a side note, I suspect there was condensation on some of their clothing as well when they saw the massive inflow of traffic at their site, if you know what I mean. ;)  '),(13591,3343256,'23,July,2004','2004-07-23 00:00:00','   I\'ve been waiting for several days for different things I\'ve ordered online, and a whole slew came today.  It totally filled up my mailbox (and it\'s a rather big box).  I got:  3 GameCube controllers ~ For the gathering I am going to in a few weeks. 4 CliffNotes books ~ To assist in my attempts to be a tutor. 1 National Geographic Magazine ~ Wooyeah! 1 Vehicle registration form ~ Ok, have to pay for this one, so it\'s not all good.  Overall, that\'s a pretty good haul if I do say so myself.  Much fun to be had.  Best of all, one of the Cliff Notes books is on economics... I\'ve never actually taken any classes above the very basics, so it should be helpful.  '),(13592,3343256,'23,July,2004','2004-07-23 00:00:00','   Ok, this is something that has bothered me with movie makers for a while.&nbsp; Why do writers feel the need to cram an irrelevant love story into a perfectly good movie?&nbsp; Shamefully, I can\'t think of any specific examples at the moment, but you know what I am talking about, right?   Movies, that have their own story, then tacked onto is a totally irrelevant love story subplot that has nothing to do with the main story.&nbsp; It\'s annoying.&nbsp; Totally ruins the experience.&nbsp;   One movie, that I only half watched, was Pearl Harbor.&nbsp; It was presented as an action movie, on the previews that I saw.&nbsp; But the movie itself seemed to be some sort of love triangle thing, with the story of the battle being the irrelevant subplot.&nbsp; I don\'t really like Ben Affleck though, so maybe that\'s it.&nbsp; Another example I was just handed was The Day After Tomorrow.&nbsp; The love interest between two students was not the best executed.&nbsp; Then again, the entire movie was poorly executed.   Some of my favorite movies have no love subplot to mess up the experience.   Shawshank Redemption  The Green Mile  Cool Runnings   All are great movies that don\'t need a gimmicky love story to try to attract audiences.   Now, don\'t misunderstand me now, I am not against love stories.&nbsp; I am against those that try to say they are one type of movie, and then not be.&nbsp; Some stories did it well.&nbsp; For example, in the Spiderman movies, Mary Jane is an integral part of the story.&nbsp; She is not just some light skirt slapped on to try to draw in a few women viewers.&nbsp; No, you take out Mary Jane, and the entire movie loses itself.&nbsp;   Another is Titanic.&nbsp; Now, it would fall into the tacky love story category if it tried to sell itself as a story about the ship\'s disaster.&nbsp; But instead it is honest from the get go that it is about Rose and Jack.&nbsp; With no love story in this one, and you are reduced to a documentary.   So, movie makers, listen up.&nbsp; If you have a love story, write a movie about it.&nbsp; If you have a story that isn\'t a love story, please, please, don\'t try and tack on a worthless subplot that only serves to distract from what is going on.&nbsp; We are not that stupid.  '),(13593,3343256,'22,July,2004','2004-07-22 00:00:00','   I wrote two articles earlier,  urlLink here  and  urlLink here , about Europe\'s growth potential in the coming years.&nbsp; (Careful, the second one will link you to the first again, don\'t get lost.)&nbsp; Reading at  urlLink Marginal Revolution ,  urlLink Tyler Cowen  posted about a shift in attitudes in France.&nbsp; You will recall that changing attitudes was one of the major things I mentioned that would be needed to bring about change on the continent.   The attitude change that TC posted about was small, but it was a start.&nbsp; He linked to a  urlLink quote  that was appropriate.&nbsp;&nbsp;Personally though, I like the earthquake metaphor better.&nbsp; There is a little rumbling going on in Europe, and few people are noticing it as of yet.&nbsp;  urlLink Progress  continues in Germany as well.&nbsp; With the two biggest economies in continental Europe starting to change course, others will follow.&nbsp; Especially as these two also tend to be the furthest behind of the developed economies.   All in all, I am getting very excited about the potential for progress. I am not overly sure why, as I don\'t live in Europe.&nbsp; Perhaps its just the potential to actually witness a major revolution in the making.&nbsp; I sure hope I don\'t get disappointed.  '),(13594,3343256,'21,July,2004','2004-07-21 00:00:00','   I talked  urlLink earlier &nbsp;about Europe\'s potential for growth.&nbsp; One of the key things that was necessary for Europe to change direction was the need&nbsp;for more consumerism.   As I mentioned in that earlier post, I am not advocating the mindless acquisition of worthless things.&nbsp; I am talking about a ready willingness and ability to purchase the kind of things that one needs and wants.&nbsp; It is especially important that they are able to produce things locally for each other to buy.&nbsp; This does not mean protectionism, which restricts outside supply, but rather working to boost local production by itself.   On a macroeconomic scale, it is  urlLink pretty clear  that Europeans as a whole are better savers than Americans.&nbsp; That in turn makes them&nbsp;better lenders.&nbsp; What that means is that although Europeans believe they are saving their money, they are actually likely to simply be loaning it to Americans to spend.&nbsp; So why shouldn\'t they spend it themselves?   Is it even possible for Europeans to spend more?&nbsp; At first glance, there are a lot of things working against them.&nbsp; The culture itself seems to promote more savings, which by itself can be a good thing.&nbsp; The tax systems leave individuals with less money in their pockets, and places more decision making for what to buy in the hands of the government.&nbsp; The government, while able to sometimes get advantages due to the scale of their spending, will not be able to draw as easily on the unique contributions of smaller businesses.&nbsp; On top of all this is a significantly higher level of protectionism that helps keep prices high for even everyday goods.   I was speaking with my Swedish friend the other day, and we compared the cost of mouth wash.&nbsp; For her, it was the equivalent of $4 for 300ml.&nbsp; I took out my bottle, and had 1.5L for less than $3.&nbsp; She cried.&nbsp; Or at least whined.&nbsp; I get more than 5 times as much mouth wash for my money.&nbsp; That same situation repeats itself to varying degrees over many other products.&nbsp; It all adds up to reduce the amount of spending money that Europeans have, at least relative to Americans.   So is there any hope?&nbsp; Most definitely.&nbsp; For one, with the continuingly large current account deficit for the U.S., the dollar is under constant pressure to drop.&nbsp; Eastern Asian countries are likely to at least try to mitigate the dollar\'s drop relative to their currencies.&nbsp; That leaves Europe as the only place for the pressure to be released.&nbsp; Over the next several years at least I believe there will be severe upward pressure on the value of the euro, and other local currencies.   This causes two big things to happen.&nbsp; One, is bad, at least from many people\'s point of view.&nbsp; That is that with the euro rising, European goods get more expensive, and so exports will tend to drop.&nbsp; I suspect this will be the initial reaction to a rising euro, and the EU area will see itself back in the doldrums.&nbsp; This in turn will help to give the powers that be the people\'s support to change things.&nbsp; Right now, especially in France and Germany, the governments are fighting an uphill battle against the will of the people to get things done.&nbsp; Even high unemployment isn\'t enough to make the people willing to change.&nbsp; Another recession occurring before the recovery from the last can even materialize may be the catalyst needed however.   This brings us to the second big thing that a rising euro will cause.&nbsp; The flip side to exports becoming more expensive, is that imports become cheaper.&nbsp; As goods become cheaper, obviously people will be able to purchase more of them.&nbsp; This will also give more competition to local producers and give them the incentive to become more efficient.   A rising euro is not the only thing that can help trigger higher domestic spending (consumerism).&nbsp; The EU, for all its faults, is becoming more effective in its ability to reduce trade barriers with the rest of the world.&nbsp; Here is  urlLink one article  on the subject that shows a greater willingness on the EU\'s part to advance trade talks.&nbsp; Lower protectionism, means lower prices, more competition, and greater consumerism.&nbsp; That in turn will lead to more jobs and growth.   There are many pieces of the puzzle starting to align in Europe\'s favor.&nbsp; It will take time, but&nbsp;I believe that within the next 5 years a major revolution could grip the European continent.&nbsp; I am so confident of this, that just today I purchased shares that reflect the value of the EMU\'s various stock markets.&nbsp; I may be a little early to the party, but it sure will be fun to watch.  '),(13595,3343256,'20,July,2004','2004-07-20 00:00:00','   The right one.&nbsp; If it\'s still broke, you should be able to tell, as the entire sidebar has been shoved down to the bottom of the page.&nbsp; It doesn\'t have to do with the width I have allocated to each column on the screen, as I set the main column down to just 10%, and the sidebar was still shoved down.  &nbsp;  The problem only appeared after I posted my recycling post.&nbsp; If you click on a previous post, the problem isn\'t there.&nbsp; So, if anyone happens to know, or even have a guess, as to what is wrong in that post (or elsewhere), please, let me know.  &nbsp;  Thanks.    Update:  It is in the post itself that is the problem, as clicking on the permalink for this particular post shows the sidebar working.&nbsp; I choose to blame Blogger\'s new tools they implemented.    Fixed:  Not quite sure why it messed up, but I took a look at the source code on each page, before and after it was fixed.&nbsp; The only difference I saw was a double line break after my post on the fixed version that wasn\'t on the broken version..&nbsp; So, if it happens again, I\'ll just try that to see if it fixes it.  '),(13596,3343256,'16,July,2004','2004-07-16 00:00:00','   Huh? What the?&nbsp; I must be joking, right?&nbsp; Well, that\'s what I initially thought when I read  urlLink something  at the  urlLink Adam Smith Institute Blog .&nbsp; There are very good points made in the  urlLink resource  that the blog linked, however.   Some points that were made I found a little iffy.&nbsp; For example, the assertion that more fossil fuels are consumed to collect recyclable materials.&nbsp; But if people threw away recyclables then the amount of waste would be increase, resulting in more waste transportation...overall, I suspect it would be largely a wash on that front.  &nbsp;  Other points are good however.&nbsp; For example:    Recycling is a manufacturing process, so it has environmental impacts. An EPA study found more toxic materials in recycling paper processes than in virgin paper manufacturing.  And:    Using less of one resource often means using more of other resources. When the total costs of recycling and disposal are compared, on average curbside recycling is 35 to 55 percent more costly nationwide than conventional disposal.    There are also good arguments that trash isn\'t so bad.&nbsp; For example, we are  not  running out of landfill space.  &nbsp;   All the garbage we will generate in the next 10 centuries will require less than 35 square miles to a height of only 300 feet. While people rarely want landfills in their backyards, they can be excellent neighbors, contributing wonderful parklands to our communities. Landfills are no longer a threat to the environment or public health. State-of-the-art landfills, with redundant clay and plastic liners and leachate collection systems, have now replaced all of our previously unsafe dumps.  But what about saving resources?&nbsp; The articles make some good point about how current processes use fewer and cheaper (read: more plentiful) materials than in past years.&nbsp; Still, I find this point to be a little more questionable.&nbsp; Especially with China developing so quickly, any savings on material is going to start being more necessary.&nbsp;  &nbsp;  On the other hand again, most resources are made from renewable or nearly infinate sources.&nbsp; Glass is simply sand, which composes much of the Earth\'s crush.&nbsp; Plastics are a byproduct of petroleum, but even now we can start making them from plants we produce.&nbsp; And most paper apparently is made from trees that were planted specifically for the purpose of being harvested.  &nbsp;  That last statement I\'d like to find some more detailed information on, especially as a very good point was made about CO2.&nbsp; Since trees absorb CO2 in their growth, recycling paper means fewer trees are harvested, and so fewer trees are planted to absorb the CO2.&nbsp; If we all threw away our newspapers, they would go to a modern landfill, permanently removing the CO2 from the air.  &nbsp;  All in all, it\'s an interesting concept to look at.&nbsp; The counter-intuitive nature of it means we have to look past the feel good activities, to find out what actually ends up helping the environment.&nbsp; After all, that\'s the whole point of recycling, isn\'t it?  '),(13597,3343256,'16,July,2004','2004-07-16 00:00:00','   Plogs, are apparently a play on blogs.&nbsp; They are automatically generated (like most things at Amazon.com) based on what you have purchased in the past.&nbsp; One insightful recommendation that they gave me is:   Tuesday, June 22, 2004    urlLink My Life  was released today; We thought you\'d be interested because you bought  urlLink The Fabric of the Cosmos: Space, Time, and the Texture of Reality .    Well, yeah, of course.&nbsp;&nbsp; Why hadn\'t I ever realized how much I would enjoy the autobiography of Bill Clinton?&nbsp; After all, I enjoyed reading about quantum physics so much, and both subjects are likely equally dull and incomprehensible to most people.*&nbsp;&nbsp; I am shocked I hadn\'t thought of it myself.   I also love their description of a Plog:  &nbsp;    The Plog Service is a personalized blog. A blog is a straightforward and now widely adopted method of posting a reverse chronological diary on the Internet.&nbsp;  &nbsp;Your Amazon.com Plog is a diary of events that will enhance your shopping experience, helping you discover products that have just been released, track changes to your orders, and many other things. Just like a blog, your Plog is sorted in reverse chronological order. When we think we have something interesting or important to tell you, we\'ll post it to your Plog.   Excellent, now those people who didn\'t know what a blog was, and so missed the obvious connection between blog and Plog, have had their confusion completely cleared up.&nbsp;   It appears that Plogs are another absolutely brilliant addition to  urlLink Amazon.com\'s  site.   I addition, it appears my cynicism and sarcasm are both on the rise.    *Addendum :&nbsp; From someone who exceeds my levels of cynicism  and  sarcasm: Both are subject to debate about how realistic what they describe is.  '),(13598,3343256,'14,July,2004','2004-07-14 00:00:00','   For just $25, it was a pretty good deal.&nbsp; Most airbeds I was finding were running close to twice that on eBay, and three times that in stores.  &nbsp;'),(13599,3343256,'14,July,2004','2004-07-14 00:00:00','   Twice now in the last few days I have found people who believe that there is no progress being made in Iraq, in both reconstruction and security.  I\'ve recommended it before, and I recommend it again, read Iraqi blogs, especially  urlLink Iraq the Model .  One  urlLink post  I found at a new blog called  urlLink Common Knowledge  said:   Nothing will happen in terms of reconstruction until stability is achieved. An initial showing of force by the interim government is needed  after discussing new figures on the size of the \'insurgency\'.  Yet I find  urlLink this post  over at Iraq the Model listing many various advancements in reconstruction, just within a single 1km circle.  Here are some examples of what he listed:   The first building you will see is that of the law institute for higher studies which was destroyed ... this institute was rebuilt few months later and now it functions in a way better than before and help to create a new law system that is honest and independent and to me this is another and more important addition.  [T]he publishing house that used to publish Udayâ€™s hateful newspaper (Babil) ... we all remember how the thieves left it as mere walls with the vents missing their robbed air conditioners. Ok, what happened later; itâ€™s working now with full power after it was reconstructed and now itâ€™s printing Al-Sabah news paper that you buy every morning.   [O]ne of the former evil military institutes that was used to train professional butchers; it ... was robbed by some sick people who left it as ruins, and you know uncle how the American unit in charge of this area turned it into a residency for students who come from other governorates to study in Baghdad. Now they have air conditioning, hot and cold water and decent bedrooms.   I didnâ€™t exceed an area of one kilometer-that we all see and pass by every day-in this round. I have mentioned a lot of things but I also forgot a lot more like all the schools and public clinics that were rebuilt within this area. Now, do you still have doubts about where our oil money went? And Iâ€™m not talking about the whole country here, have you asked yourselves why the exchange price is 1500 now instead of 2200 in Saddamâ€™s days? And do you think that a 100 times raise for some employees salaries doesnâ€™t represent a part of the money you were talking about? Donâ€™t you notice the private businesses that are starting and those that are booming?  There is more, and I strongly suggest that people read it for themselves.    The point of that post in the Iraqi blog was not just about reconstruction advancements, but also about making it a point that it is not at all what the media wants people to believe is happening there.  He ends the post nicely:   Then I told myself: if some Iraqis, living in Iraq, watching the changes being done under their noses fail to see the truth and still depend on the media to tell them whatâ€™s going on in their own country then how can we blame people who live thousands of miles away for thinking similarly!?  Indeed, as so many people who live those thousands of miles away apparently do believe the lack of progress.  As for the security situation, check out  urlLink this post .  Or, if you prefer to have a \'real\' news source,  urlLink here  is the BBC article on it.  The situation in Iraq is far from ideal, however it is equally far from the image most major media projects of it.  '),(13600,3343256,'13,July,2004','2004-07-13 00:00:00','   Reading a title like that and you are probably wondering if I am off my rocker.  Well, if you\'ve spent any amount of time on my blog, you know the answer to that: Yes.  But that\'s besides the point.  A couple of times now,  urlLink Tyler Cowen  has mentioned a book called  urlLink Growing Public  by  Peter H. Lindert.  It\'s apparently a historical look at government spending over the last couple centuries, and how social spending added to economic growth, as opposed to the intuitive belief that it reduced it.  This subject greatly interests me, and I have even  urlLink written upon it in the past .  While the basic concept makes sense, I still find it difficult to believe that spending on social programs has little to no negative effect on growth.  After all, a quick glance at the economic health of Europe vs. America, easily shows America far in the lead compared to the largest countries in Europe, and to the EU as a whole.  Even Europeans seem to be fully aware of this gap.  In a recent article I found, sub-titled  urlLink Who Minds the Gap?  the various leaders, while seeming to be aware of the shortfall in growth compared to the U.S., do not seem very willing to actually tackle the problems they face.     Each government awarded itself good marks for its wise and decisive policies. In fact, with the exception of Spain, Portugal and the Netherlands, (Britain was already ahead of the rest thanks to the radical Thatcher reforms of the 1980s that Labour has preserved and built upon), progress by most of the others consisted of two steps forward and two steps back. In France, \"progress\" has consisted rather of three steps back. Romano Prodi, the outgoing President of the European Commission who, told them to their face that he wished they would stop pretending that they are even trying to implement the Lisbon Programme.   This perspective contradicts that of Peter Lindert\'s book.  Unless of course it\'s not the social spending that is holding back Europe.  So what other problems could Europe be facing that is not directly related to social welfare?  In a book I am reading,  urlLink The Wealth and Poverty of Nations , I have gotten a better idea of the sources of growth that made the Industrial Revolution possible.  Of the several things needed for growth, a few of them seem to explain Europe\'s problems quite nicely.  First, the transportation infrastructure desperately needs improvement.  I read recently in the Wall Street Journal about Germany\'s problems with getting new Autobahns built.  The gist of the article was that the country is being bogged down in a massive number of appeals by special interests that is delaying the completion, or even the start, of roads connecting to the newest members of the EU. Since Germany is geographically the hub of most of the trade between east and west Europe, this is a major drag on the entire continent\'s ability to move goods freely.  The second major problem area for Europe, appears to be in its resistance to productivity gains.  I have read in multiple places that Europeans seem to view productivity as a job killer, and so believe it is logical to shun it.  Sadly, as a result the competitiveness of European businesses suffers, and so results in a net effect of having  urlLink unemployment across the Euro area at about 9% .  This is compared to well under 6% in the U.S.  The third, and largest problem area in my opinion, is a lack of consumerism in Europe.  Why is consumerism good?  First, I need to clarify what consumerism is, as it is not to be confused with materialism.  Basically, the ability of one individual to consume the product (or service) of another is consumerism.  The major difference between developed countries, and undeveloped, is that in developed countries people tend to make products for each other.  In export driven countries, the people try to make products for others to buy, usually because the people themselves cannot or will not purchase their own products.  Because there is no local demand for goods and services, then the true potential of individual workers cannot be realized.    Case in point, if people were not willing to spend their own money freely, I would not even be able to consider becoming a tutor.  Because of a consumeristic attitude, there are people who will be willing to purchase my specialized contribution to society.  It should be noted however, that I am not advocating mindless buying of \'stuff\'.  That helps neither the economy nor the environment.  What I am saying is that people should be willing to spend the money that they earn wisely, instead of just defaulting to putting it in savings.  That savings tends to end up being lent to those that will spend it anyway, and perhaps not as wisely as those who decided to save it.  Fortunately there is hope for Europe.  Better transportation can be built. The per hour worker productivity in Europe is as high as, or even higher than America\'s. And the attitude towards consumerism has can be adjusted.  Granted, there is little motivation towards change at the moment, however that will not last forever.  There is only so much lack of progress that people will put up with until they declare that they\'ve had enough.  It happened in the late 70\'s and early 80\'s for the U.S. and Britain.  It is only a matter of time before continental Europe gets sick and tired of its situation as well.  When it does, its potential for growth is phenomenal.  '),(13601,3343256,'11,July,2004','2004-07-11 00:00:00','   urlLink Earlier  I reported that I was having trouble with my computer.  I figured it was just getting old, as I couldn\'t find anything specifically wrong with it.  Adaware: No problem.  AVG Virus Scan: No pro#%&*... system crash.  Ok, so I knew at least that it wasn\'t spyware.  Or so I thought.  Reading a Marketplace column in yesterday\'s Wall Street Journal, I found that the free spyware detectors don\'t always find everything.  And, more importantly, you don\'t need to do anything \'wrong\' to get spyware anymore.  So, I took the column\'s advice, and looked up  urlLink Pest Patrol  on the internet.  Although their software costs about $40, they offer a free system scan.  I took them up on that, and found spyware.    The spyware had an innocent sounding enough name, which I suspect is part of its success in hiding itself: PowerReg Scheduler.  Fortunately, despite the fact that Adaware couldn\'t even find it, the Pest Patrol manual removal instructions were easy to follow.  Very easy.  Essentially, all I had to do, was click Start >> Programs >> StartUp >> move over to PowerReg Scheduler >> Right-Click >> Delete.  It was that easy.  Whereas previously my virus scan was crashing my computer, it now works fine.  I just finished running it a short while ago, no viruses.  I am also hoping that perhaps removing this spyware cleared up a few other problems I had been experiencing lately.  Namely, that when I close all but one IE window, that last window tends to freeze.  I will experiment later.  So, if you are having trouble, or just want to check to make sure, head over to  urlLink Pest Patrol .  But, if you have Adaware, run that first, that way all the tracking cookies are removed ahead of time, which makes it easier to sort through Pest Patrol\'s results to see if there are any spyware programs.  '),(13602,3343256,'11,July,2004','2004-07-11 00:00:00','   Ok,  urlLink earlier  I posted some musings about how gravity may work.  A few problems with that idea though.  So, I pondered some more!  This stuff has its own flaws, but at least it should be interesting to think about.  Now, we know, according to quantum mechanics, that everything in the universe has a dual nature, being both particle and wave.  The wave is a probability wave that describes where an object actually is.  To visualize this, take a pebble dropped into a pond.  The point where it is dropped creates the strongest waves, which exponentially decrease (that means really, really quickly) in intensity the further from source you get.  Now, I suppose that for particles, the waves are generated continuously, describing where the particle is located.  So, we now have a continuously generated wave spreading out from the pebble point.  Let\'s add another pebble.  Now we have two, with their waves coming out, and overlapping.  Let\'s concentrate on where the waves overlap.  Just like in a  urlLink double slit experiment  the waves will add together.  As a result there will be high crests, low troughs, and flat points.  I am unsure exactly how to interpret each of these, but I got some guesses.  The crests, and perhaps troughs, are increases in the probability of the particle being at that point.  This will cause an attraction between two objects, as the points between them will be more likely to contain the particle than other points.  The flat points, and this is important, are where there is no chance for the particle to exist.  It is important, because this may change the overall probability of the particle.   Starting with the assumption (which may be wrong) that a particle\'s probability wave adds up to 100% when it is undisturbed, the interaction with another particle may change the overall probability of its location.  What this means, is that when disturbed by other particles, a given particle may have less than 100% chance, in total, of being in any one place.  An object therefore has the chance of  not existing at all .  Because an object may not exist at all during certain moments, it will appear to act slower than objects that appear to be disturbed less.  The higher the disturbance from other probability waves in the area, the more disruption, the there higher the probability of an object not existing at all at any given moment. This is the time dilation I was missing from my earlier musings!  This is all purely speculative of course.  I need to read up a lot more to really be able to understand and actually do calculations.  But, as I said earlier, it\'s interesting to think about.  '),(13603,3343256,'10,July,2004','2004-07-10 00:00:00','   Here I was, sitting here just minding my own business, writing a blog post about Spiderman 2, which I had just seen...and my computer crashes.  Again.  That\'s probably the 5th time today.  The first three times were while trying to run a virus scan.  As you can probably guess, I\'ve stopped trying to run the scan now.  The other two times were just seemingly random crashes.  Everything just locked up.  Worse of all, I had to hit the power switch on the surge protector to shut off my computer.  The stupid off button won\'t work if the system has crashed.  Excellent design I must say.  Make a button that only lets you turn off the computer when everything is running fine.  Sheer brilliance.  Oh, and you can\'t even use it for when it\'s time to shut down the computer, because when you exit Windows, it automatically shuts itself off.  So I have a totally worthless power button. Gah.  *Saved post as Draft*  There, now I won\'t have to totally start over from the beginning.  Now, where was I?  Oh yes, new computer.  So, if anyone has any suggestions about what to do, I\'d love to hear them.  I\'ve been considering building my own computer, but as I\'ve never done it before, I am somewhat apprehensive about it.  On the other hand, I could buy another cheapo computer that will hopefully last me a few years.  This one is running right at about 3 years now.  For $400 for the tower part, I suppose that\'s not too shabby.  *sigh* I hate shopping.  '),(13604,3343256,'07,July,2004','2004-07-07 00:00:00','   Need it for a gathering of people I\'ve met online that I am attending in August.  Fair price too, only $26.65 off of  urlLink eBay .  Always a good place to get stuff cheap.  Currently am shopping around for a sleeping bag and air mattress, then I\'ll be all set.  I made sure to check it out, to see if everything was okay with it, as it was a used tent.  Sure enough, I got it up and stable...after a few tries.  Turns out that vacuums are a powerful source...so make sure the door is open before you try to set up a tent, otherwise you end up trying to lift the entire thing off the ground.  Live and learn.  Glad I figured that out now, before anyone is around to watch me mess around with it when I actually go camping.  Although, I use camping rather loosely, as it will just be in someone\'s backyard.  Yeah, I\'m a wimp, but at least I\'m a comfortable wimp. :P  Hmm, I need to get a camera too.  I wonder how cheap those come.  Time to search eBay some more.  '),(13605,3343256,'07,July,2004','2004-07-07 00:00:00','   One of the spiders got a  urlLink bee .  It\'s just hanging there, nice and dead-like.  Poor bee.  On the flip side though, this means that spider now has nourishment, and will live longer and have lots of baby spiders.  Damn.  Stupid bee.  I think I need to write another note to myself about getting bug spray.  '),(13606,3343256,'07,July,2004','2004-07-07 00:00:00','   This post is being written with the intent to introduce the Myers-Briggs Type Indicator (MBTI) to those who have no experience with it, or with personality typing in general.  What the MBTI is is a classification of different ways in which people prefer to operate.  It is not a system for classifying what skills a person has.  Nor is it a means to label someone with certain traits, although within each type similar behaviors do tend to develop.  Rather, personality typing is a way of understanding people.  It is a way for those that fall into the less common groups to understand that, despite what others around them may say, they are not \'strange\' or \'wrong\'.  This, in fact, is one of the main reasons why I\'ve taken the time to learn about personality types: to help people understand that there are many different ways of acting, without one way being the \'right\' way.  Few people want to be told that they are defective, or that they need to be fixed.  Most people, if not convinced otherwise early in life, are happy with who they are.  I have even found one  urlLink autistic  person who seeks to inform people of how his mind works, why he acts like he does, and most importantly, to  not try to fix him .  There is nothing for people to fix, he loves how, who, and what he is.  This is not to say that some of the types do not fit into society as easily as others.  That however is due largely to the fact that the 16 types in the MBTI do not break down evenly in the population.  In fact, the single classification that provides the most communication problems, also shows the largest disproportion in how they are represented in the population.  The intuitive, global, abstract mindset comprises only about 25% of the overall population, based on  urlLink figures  I have seen, yet accounts for half of the 16 types.  The sensing, detailed, concrete mindset, makes up the other 75% of the population.  Due to this disproportionate representation, the Sensing mindset is that which tends to be considered the more \'normal\'.  I have little doubt, that the majority of those diagnosed with mental \'disorders\' (ADHD, for example) are simply iNtuitives living in a Sensing world.  Well intentioned parents and doctors seek to diagnose the behavior of iNtuitives as some sort of defect, and then attempt to find ways to correct it.  A quick search on the internet will find many, many opinions about the overdiagnosis of diseases such as ADHD.  This does not mean that only iNtuitives are not understood.  Any type can be made to feel an outcast when there are no others of their own type around.  Especially if raised in an environment where strict adherence to a set behavior is enforced.  What is important for people to realize is that there is no good or bad type.  There is no one way of doing things.    We live in a world where there are acceptable ways of acting, but at the same time the world is as we make it.  It will change as the people in it become aware of and respect the uniqueness of others, without trying to dictate so called proper behavior to them.  We each have our strengths and our weaknesses.  Sometimes they are one and the same.  Just as some people are socially oriented, they should not consider those who are more quiet as defective.  Those who are more decisive should not considered the patient to be less intelligent.  In both cases the tables could be turned, and in many times are.  Introverted types may view those that are extroverted as somehow defective in their constant need for attention.  Those who take their time in matters may assume that those who jump to conclusions to be the ones who are less intelligent.  Hopefully I have given those of you new to personality typing enough information not to dismiss it out of hand as being a way to put people in boxes, and rather to see it as a way to help let people be themselves.  If you want to learn more, read my  urlLink introduction to personality types , although from here on out it will steadily get more technical.  Don\'t worry, it still starts out easy.  '),(13607,3343256,'06,July,2004','2004-07-06 00:00:00','   Well, the big news out today is of course that Kerry selected John Edwards as his running mate in the presidential election.  What is less apparent is the fact that it appears Robert Shrum may have played a large part in that decision.  Who?  I\'d never heard of him before today either, however there was a front page article in the  WSJ  about him.  Reading the article, along with another  urlLink article  that Google found for me, and I\'ve discovered this guy is shaping up to be the Democratic equivalent to Karl Rove.  From the WSJ article:   Officially, Mr. Shrum\'s job is to produce Mr. Kerry\'s campaign commercials.  But after three decades in presidential politics, where most stars burn out far more quickly, Mr. Shrum has an unusual amount of sway.  \"I\'ve been doing this longer than anybody,\" Mr. Shrum says.  Aside from Mr. Kerry himself, no one person will have more influence on the Democratic ticket\'s message between now and election Day.  That influence also apparently includes the choice of who is on the ticket.  Shrum has a long history as a populist figure, and has developed many connections within the party as a result.  He was a speech writer for Edward Kennedy\'s campaign against Carter in 1980, he created the ads for Richard Gephardt\'s primary campaign in \'88, and was a consultant to Al Gore in 2000.    My initial impression of Edwards was that he was too populist for me.  Reading a post, with multiple links, at  urlLink Marginal Revolution  didn\'t improve my opinion either.  Frankly I find populism in general distasteful.  The old Slate article I stumbled upon summed up my opinion of it nicely:   Populism is one of the more romantic and less admirable American political traditions. It purports to represent the interests of the little guy-the people, not the powerful, to use the Shrum-Gore bumper sticker-but more often than not it has manifested itself as a witlessly reactionary bundle of prejudices: nativist, protectionist, isolationist, and paranoid. The central assumption is that the little guy is so aggrieved that he can only be roused to citizenship by an appeal to his basest suspicions.  All this however appears to be the hallmark of Robert Shrum.  According to the WSJ article he has strong ideological convictions that he brings with him and was apparently a principle architect of Al Gore\'s people vs the powerful rhetoric.  Which explains why he helped to pick John Edwards, who ran a rather populist campaign himself during the primary.  It remains to be seen however how much overall influence the populism has on Kerry\'s campaign and potential presidency.  Regardless of that however, if Kerry wins, I suspect Robert Shrum will become as well known, and as hated by Republicans, and Karl Rove is by Democrats now.  '),(13608,3343256,'02,July,2004','2004-07-02 00:00:00','   Any grammar Nazi\'s that visit my blog will likely have noticed this.  I\'ve been using it\'s wrong.  Silly me thought it was always possessive, like:  The man\'s dog. The dog\'s bone. It\'s whiteness.  But noooo, they gotta change the rules on me.  It\'s: Its whiteness.  No apostrophe for possessiveness, only for the contraction of \"it is.\"    Worst of all, my first instincts were to do it the correct way, but I kept correcting myself and going back and adding the apostrophe all the time.  Ah well.  Should be better now.  Hopefully.  '),(13609,3343256,'02,July,2004','2004-07-02 00:00:00','   A while back I wrote  urlLink a post  introducing what personality types are.  Here I will continue the introduction, by explaining the different functions.  Be sure to read that earlier post first, if you have no background on what the four preferences are.  Each personality type breaks down into four distinct functions, based on the four preferences of that type.  What is a function?  A function is how a person interacts with the world.  It is very important not to confuse preferences and functions.  Preferences are the four letters that make up a type\'s name.  Functions are how those preferences combine to explain how a type, well, functions.  Each function is made up of one of choices for the middle two preferences in a personality type\'s call letters.  That means Sensing, iNtuition, Thinking, and Feeling.  These are represented as capital letters in the shorthand notation.  Each of these also has a lower case letter, signifying it\'s direction: whether that function is introverted or extroverted.  The four functions for INTPs are: Ti Ne Si Fe.  In that order, from dominant, to inferior.  Translating, they are:  introverted Thinking extroverted iNtuition introverted Sensing extroverted Feeling  Using my type as an example, you can notice a few things.  First, is that the dominant (top) function, is the same direction as my 1st preference: Introverted (I).  You will also notice that the top two functions are determined by what the middle two preferences are: iNtuition, and Thinking (N and T).  Finally, and less obvious, is the fact that as the 4th preference is Perceiving (P), that means that the 2nd Preference, iNtuition in my case, is extroverted (Ne).  If my 4th preference was Judging, then the 3rd Preference would be extroverted.  To see what happens when just the 4th preference is switched, lets look at INTJs.    Ni Te Fi Se  introverted iNtuition extroverted Thinking introverted Feeling extroverted Sensing  Again, the 1st preference is the same as the 1st function: Introversion.  This time though, as a Judging personality, the Thinking is extroverted, and as such is the 2nd function in INTJs, where the iNtuition is the primary function.    You will note that as the top two functions switched places and directions, the bottom two did as well.  This is due to the fact that the first and last function, are always from the same preference.  Either the perceiving preference, or the judging.  The middle two functions also follow this rule.  Also, you will note that the direction of the functions always alternates from one to the other as you go down the function list.  Introverted, then extroverted, then introverted, then extroverted again.  It is the opposite, but still alternating, order for those with the Extroverted preference.  Here are a few examples to see if you correctly understand how the function order is determined.  ESFJ: Fe Si Ne Ti ENTJ: Te Ni Se Fi ISTP: Ti Se Ni Fe ESFP: Se Fi Te Ni  Later, I will explain what each of those functions mean, and how they work in each type.  '),(13610,3343256,'02,July,2004','2004-07-02 00:00:00','   Earlier I ran across an interesting  urlLink article  about the speed of light speeding up.  What was really interesting though, was where it talked about a  natural  nuclear reactor.  Yes, natural, as in occurring in nature.  I found  urlLink this  page on the internet, which did a very nice job of describing the process of what happened.  Apparently 1.8 billion years ago there was a sustained nuclear reaction in what is now Oklo, in Gabon, Africa.  They way this was discovered was due to the composition of the uranium found there.   Natural uranium is always of the same isotopic composition--99.27 percent Uranium-238, 0.72 percent of Uranium-235 and traces of Uranium-234. Only the Uranium-235 isotope can take part in the chain reactions necessary for power production or for explosions.   The uranium harvested in Oklo however, was severely depleted, compared to natural uranium, showing that it had already undergone nuclear fission.  It wasn\'t technologically advanced cavemen who did it though, it was bacteria.  What happened was uranium dissolved into the water of flowing streams becoming one of the many trace elements within the water.  Downstream bacteria collected and concentrated the uranium, and when the bacteria died, caused it to be deposited in significant quantities.  Eventually the quantity of uranium was large enough, that critical mass was achieved, and a nuclear reaction started.  It wasn\'t an explosive reaction however:   Provided that the number of neutrons produced balances those that escape, or are absorbed by other atoms, the reactor continues. This kind of reactor is not explosive; indeed it is self-regulating. The presence of water, through its ability to slow and reflect neutrons, is an essential feature of the reactor. When power output increases, water boils away and the nuclear reaction slows down. A nuclear fission reaction is a perverse kind of fire; it burns better when well watered. The Oklo reactors ran gently at the kilowatt power level for millions of years...  An interesting note is the fact that uranium cannot dissolve in water without the presence of oxygen.  This is very fortunate for the bacteria involved, for the concentration of the fissionable type of uranium when life began, and therefore before oxygen was abundant in the atmosphere, was not 0.7% of the whole, but rather 33%.  That would make for some very sparkly bacteria,  while  the bacteria was still alive.  All in all, a very interesting event.  '),(13611,3343256,'02,July,2004','2004-07-02 00:00:00','   Many new friends.  Earlier today I found four bees, where before there had been only  urlLink one .  I suppose I shouldn\'t be surprised, after all, that\'s why they make nests, is it not?  There are also a few more visible in the central nest between my sliding glass doors.  They have not \'hatched\' yet, however they are still distinguishable within the nest.  All the active bees appear to be working on the righthand nest.    I think I need some bug spray.  The handful of spiders that share the area between the doors are rather lazy.  Although I did spot one of them lunging after a bee when the bee trampled through one of the lines of it\'s web.  I can be patient though, as the bees have not yet reactivated the lefthand nest, which is the only one I have to deal with.  Frankly though, I\'d rather have bees than spiders anyway.  '),(13612,3343256,'07,August,2004','2004-08-07 00:00:00','   I just got my games back from my cousin who had borrowed them a while back.  One is Mario Kart: Double Dash.  It\'s an excellent multiplayer game, and a mediocre single player game.  Battle mode especially kicks butt.  The Grand Prix race though is what you are left with for just single player though.  In it, 16 characters, on 8 carts (paired in teams of 2) race around the various tracks to win the race.  There are usually 3 laps, and there are four tracks in a Cup.  Each race is scored based on how you placed, and then the overall score of the 4 tracks in the Cup is added together to determine who wins the trophy.  It\'s not simply Go-Karts though.  There are items to pick up along the way.  Actually, calling them weapons would be more accurate, although items is better as some are bonuses for your own kart.  Along every track there are masses of these item blocks, and each character gets a chance at a special item unique to them and one other character.  For example, Yoshi and Birdo are a pair, and they get eggs.  These eggs race along the track until they impact the kart ahead of them and spill 3 items onto the track.  Whenever you get hit, you spin out, or are otherwise momentarily slowed up, making it easier for those behind you to catch up and pass you.  Explosions are one of the worst things to happen to you.  Not only do you get blown high in the air, you are blown straight up, so unlike some other items you can\'t still move forward a bit and maintain some of your momentum.  Also, you lose any items you were holding at the time.  Generally, the game is well balanced, except for one item that I know some have complained about before.  The blue koopa shell.  This nasty little bugger is only given to players who are not in first place.  And for good reason.  This shell will fly along the track, totally avoiding anything that might stop it, and seeks out the first place player, and explodes around them.  Very annoying to have the lead, and lose it because someone behind you managed to get this item.  There is absolutely no skill involved, and except for invincibility, essentially no way to avoid it.  I realized something today while playing though that I think explains why the game designers put it in.  When you are racing, and you take first place, you are in a very very safe position.  There is no one ahead of you to lay stuff on the track to slow you down, and those behind you can do little to you after you managed to get a small lead.  Basically, if you are a good driver and manage to take first, the race is all but wrapped up.  On the other hand, if you are in any other position, you are under assault from other karts both ahead and behind you.  This is incredibly annoying, and makes it difficult to catch the first place player as all the karts further back are competing with each other, and so fall further behind the first place kart.  The blue koopa shell evens the odds a bit.  It makes it much more dangerous to be in first than it otherwise would be.  Overall, I think it is a good addition to the game.  Makes it considerably more interesting to play.  And more frustrating.  But that\'s what it is all about, isn\'t it?  I\'m going to have a chance to play with a bunch of people in a couple of weeks.  Four player races is going to be loads of fun.  Can\'t wait.   urlLink    Blue Koopa Shell  '),(13613,3343256,'02,August,2004','2004-08-02 00:00:00','   It seems like everything stressful in my life today is a result of a lack of direction.  A paper I am writing, my job future, even a little project for a friend I\'m working on.  I know I can do whatever I need to, I just don\'t know  what  to do.  This paper I am writing is for a contest, with a top prize of $20,000.  Nice chunk of change, although I highly doubt mine will score high enough.  Especially where I stand now.  I managed to write a first draft...that basically said nothing.  It was too vague, and had no focus.  The topic of the paper is supposed to be related to jobs, and those lost overseas.  The papers are judged based on bring new ideas, perspectives, or judgments to the discussion.  Yet the criteria is so vague, I couldn\'t figure out what to write about.  I still haven\'t, and I\'ve only got about 9 days left to finish a final draft.  For my job, as some of you know, I\'m looking into becoming a tutor.  I\'m totally clueless as to whether it will work, and how exactly it does work.  Do I put up lots of flyers?  What about pricing?  Can I charge the \'standard\' rate of $20, or should I provide a discount since I don\'t know what the heck I\'m doing?  What do I do if I get a student and know remember or understand how to do what they are working on?  Granted, I\'m teaching math, and it has always been easy for me, but even with that some things have eluded me.  That is the nice thing about this blog, I don\'t pretend to have a main subject and so can write about whatever I want.  The lack of focus is a great plus, it fits my personality perfectly.  A lack of focus in the real world is less than ideal.  Sometimes I wish I just had someone to point me in a direction and say \"Do this\", and then I will attack it, and do it.  I can do almost anything I set my mind to.  It\'s not arrogance, but rather simple fact.  I was one of those annoying kids in school that got good grades without even trying.  It\'s the setting one\'s mind to something that is the problem.  I remarked to someone not long ago that the fact that I never had to work very hard in school made me lazy and unambitious.  I think it is finally starting to catch up with me.    Time to focus on a direction.  I\'ll start tomorrow.  Or maybe Monday.  Eventually at least.  '),(13614,3343256,'02,August,2004','2004-08-02 00:00:00','   It is practically accepted fact to most people that prescription drugs are substantially cheaper in Canada than in the U.S.  What people don\'t know, is that generics are substantially more expensive there as well.  While searching google I found this  urlLink paper(pdf)  on the issue.  The difference in prices is stark.  The paper found that a weighted mean of the drugs surveyed were 37% more expensive in Canada than in the U.S.  Some were significantly more expensive, one over 10 times as much.  That translates to almost C$400 million in extra costs vs the U.S. prices when extrapolated across all the various generic drug purchases in Canada.  Certainly nothing to sneeze at.  So what causes this difference?  The same thing that causes lower prescription prices in Canada: price controls.  With the Canadian government (and its provinces) controlling drug prices, the price for generics is set centrally, and therefore is practically immune to reduction by competition.  This, in large part, is because instead of many manufacturers trying producing a generic after the drug\'s patent expires, many times only the two largest will start producing it.     While most top-selling generics are produced by several suppliers, the market for each product tends to be dominated by one or two companies. Apotex and Novopharm are the two largest generics manufacturers in Canada and account for more than half of the total Canadian generics market (in terms of both prescriptions and sales revenues). These two companies dominance of the market for many products may make it difficult for other manufacturers to compete and for new generics firms to enter the market.  Certainly not a recipe for price wars.  This makes one wonder several things.  For one, as the paper also expressed interest about, does this hold true for Europe as well?  Countries there have health care systems much more similar to Canada than the U.S.  For another, since price controls cause both lower prescription prices, and higher generic...what is the overall benefit of price controls?  While I have little doubt that when looking at prices the benefit is still there, at what cost does it come?    Definitely something that should be considered in the overall health care debate.  '),(13615,3879492,'27,July,2004','2004-07-27 00:00:00','   I Apoligize for a coupleof my posts which have been RUDE,Crule,and hurtfull Nick i am sorry for slamming on u i know ur not a pothead as for camron i no u havent done anything wrong and plus we dont even see each other anymore but as&nbsp; i said im sorry for my rudeness and the cursing but as for the ones who know me they do know im a great person with much love towards life and people im not a bad person i just had to speak my opinion. But this is the usa aint i aloud too ? Im sorry i didnt want to offend anyone or give a wrong impression!&nbsp;&nbsp;So for everyone i ask for forgiveness! I ask for all to accept this appology and reconsider what u think about me and give me a chance. Out Aj Buben '),(13616,3879492,'25,July,2004','2004-07-25 00:00:00','   EVERYONE GO TO SKATEWORLD yes thats what i said this friday everyone should go me gots a big plan in stare for everyone nothing&nbsp; bad we will all have fun&nbsp;to other things&nbsp;i havent had any sleep im in love and im lonely and bored scary huh well im serious everyone needs to go to skateworld its very important everyone that comes will see something or be amused at something you have never seen before but im soooooo tierd and i cant remember other things i was gonna say so love you all   out   tony the bengal tiger buben '),(13617,3879492,'25,July,2004','2004-07-25 00:00:00','   yea that\'s what I am sooo what im bored and I\'ve got no one to talk to or anyone that I can see without a fucking car. soo if any one that reads this and wants to come over or pick me up from my dads go ahead please. Well I cant remember wtf I was going to say soo im gonna go ill write stuff later peace to all and gooooooood night '),(13618,3879492,'15,July,2004','2004-07-15 00:00:00','   Now jessica has changed her mind she wants to stay with camron OK i guess its good i did not bother to tell him he was dumped well OK then just thought id update u with some news  Chow TONY B '),(13619,3879492,'14,July,2004','2004-07-14 00:00:00','   Jessica Broke up with camron last night though he doesnt know yet she told me to tell him but im not going to call him or walk down the street to tell him if hes my friend he shall call than i shall tell him but he cant blame jessica because its his fault he never called her or made efforts to spend time with her or to see her that sounds like his last realationship with mindy what im trying to say is both brakeups are his fault but it seems as always i am stuck in the middle so this is my advice for me friend camron grow the fuck up and be a man and dont blame them for your mistakes!!!!!!!!!!!! As for me today bored as hell waiting to die lol jp but i am bored just sitting here singing songs and downloading away im bout to finally go the fuck home peace yall  TONY BUBEN OUT '),(13620,3879492,'14,July,2004','2004-07-14 00:00:00','   I cannot say how much im soo glad to be goin back to kettering tomorrow because im soooooooooooooooooo bored sooooooooooooo lonely and sooooooooooooo depressed no one has been calling me ya know its getting to be hell tryin to maintain some sort of social life when my girlfriend\'s ankle is broke and no one else but her bothers to call me  so when i get home i is goin to call and see as many people as i can and do things every day and i hope to see my girlfriend afore she has surgery but as i can i will try to write afore i leaves here and i will try and right at my moms sooooo if ya read this give me a call so i iznt bored  MOM # 256 2375 DAD # 432 0145 ME CELL # 286 2183 if i doesnt answer this # leave me a lively message  I IS OUT TONY B '),(13621,3879492,'09,July,2004','2004-07-09 00:00:00','   This Poem is A poem from one of my friends who u say im not tellin u   Forgotten Sitting here not intouch with the real world Trying to understand what went wrong Constantly wondering my mistakes What did i do for you to treat ne this way Ever time I told you to go ,Why exactly did u stay? Was it a test to see if Id give in, take all the things I said back and let you win? Inside I feel like giving up I cant take the stress that comes with this love stuff No longer will I sit here wondering what went wrong I will learn from my mistakes and everything I have I will appreciate Sitting here knowing my future and wishing that you\'d be gone......... Author  I shall not tell '),(13622,3879492,'09,July,2004','2004-07-09 00:00:00','   Hey this song i like  alot if u dont ohh  well violent j: Ive Lived one hell of a life, if your to take it tonight, Ive lived the life of a king, but so much to darkin my light Shaggy 2 dope: Looking 300 degrees, two hundread thousand desease, how can I lay back and chill, so many down on there knees j: But then who again who am I, Im just one Typical guy, I aint no fuckin hero, I Just been wondering  why..... Chorus: I seen some children crossing the bridge, what kind of life did they get to live? What kind of choice do they get to make, what if it was a childs mistake? J: I\'ll never question the book, but lets say i lived by the book, i never ate meat on tuesdays, how much bigger would my wings look Shaggs: I try to pray everyday, but sometimes get lost on the way, I seen the holiest spirt, so much to misslead the way J: i gave visions I got, Ive been told im gonna rot, inside the devils intestines, im still holding my spot, Ive just been wondering why..... Chorus: I seen some children crossing the bridge, what kind of life did they get to live? What kind of choice do they get to make, what if it was a childs mistake? Break down: COME ON UP AND OPEN UP YOUR DOOR, WHAT IF ITS A DERAGHTORY DOOR, COME ON UP AND OPEN UP YOUR DOOR, WHAT IF ITS A DERAGHTORY DOOR, fuckin take me instead! Chorus: I seen some children crossing the bridge, what kind of life did they get to live? What kind of choice do they get to make, what if it was a childs mistake? I seen some children crossing the bridge, what kind of life did they get to live? What kind of choice do they get to make, what if it was a childs mistake? I seen some children crossing the bridge, what kind of life did they get to live? What kind of choice do they get to make, what if it was a childs mistake? '),(13623,3879492,'08,July,2004','2004-07-08 00:00:00','   No one fucking realizes I care ok sorry had to get that one out Jenny u have to know that I really care a lot about u  I cannot help that  I cannot  get over to ur house and I know that u know I love u but the thing is I really really love u like the love u know u hear about in  story fantasy books come on realize it people its finally my turn to speak up now to another big subject ummmmmmmm lets go at it again Amanda realize u got at least one friend that cares me and  to another person  camron okay u need to fucking realize that we are best friends and all but  dont  run ur mouth to the little fucking punks u hang out with lets see umm Jeff Phil bobby and dwyane  but  I dont have no problems with u  you just need to get some better  other friends that are actually ur friends not people who just hang out at places and smoke cigs and weed and fuck there best friends girlfriends I know u dont do none of those things just fucking realize these are bad people now to my old best friend nick u wanna run ur mouth to cam and zach and then say ur my best friend the dont call except  every three weeks when u get into ur dads pot  stash and try to get me to smoke and  never call unless ur parents are botherin u and  use me as a get away fuck that sorry ill be back with more later TONY B REMEMBER I LOVE YA JENNY '),(13624,3363036,'16,May,2004','2004-05-16 00:00:00','   I have attended thus far, 5 theory lessons and i must say the best was conducted by this guy named Mr Woo. I enjoyed his lesson coz he related it to real life driving disasters haha..   Altho he has this \'i-take-no-nonsense\' kinda look,he is actually an interesting guy! Anyway, i juz wanna share with U some stories he told us during the lessons.    STORY#1: Animals Ahead    Mr Woo was doing his job as usual as an instructor with his learner driver. They were driving along and he spotted a bird on the road ahead. Hence, he warned his learner. The learner said, \"Where? Where? Got meh?\" and before Mr Woo could say another word, they heard crushing sounds of bones beneath the car. Eeiyer, some of u say.. but the learner say \"Where got? Dun have wat..\" Pekak-badak ke budak tu sampai sound camtu pun tak dgr! Anyway, they made a U-turn to go back to the site. The poor animal was like pancake and it made the learner totally freaked. At the traffic light junction, Mr Woo said the learner was perspiring all over altho the aircon was on. \"Eh, got animal ahead!\"Mr Woo said after they continued their drive. Immediately this learner stopped the car (nasib takde kreta lain pat blakang)\"How now?\" said the learner who spotted the animal on the road. \"Shoo it awaylah,\" said Mr Woo. \"Shoo! Shoo!\" the learner went (complete with the hand actions). Mr Woo said to us in the class as he related the story, \"When my learner did that, I laughed like hell lah in the car. She was shooing the animal from inside the car!\" Macamlah binatang tu leh dgr with all the windows of the car drawn up, kalau iya pun use the horn-lah.. haha. \"Lucky thing she didn\'t get out of the car to shoo it away, that would have made me laugh harder and perhaps even the passing drivers,\" said Mr Woo. Apparently, this girl selalu kena animal related incidents after that. Once, they drove along Bt Timah and they had to stop for a tortoise to cross. Tu kalo horn pun tak guna, tak leh speed nye! hehe.. Muz be the curse of the pancaked bird at work gaknya.    STORY#2: The Zig-zag cyclist    This cyclist seems to enjoy making things difficult for learner drivers from this driving centre i enrolled in. Mr woo tells us of this man who would immediately ride his bicycle in a zig zag fashion whenever he spots cars from this driving centre. The worst part is he would do his zig-zag cycling across the whole lane, making it impossible for the learner driver to proceed without hitting him. A while back, this man did as he usually would when he spotted one of the driving centre\'s car. The driving instructor in the car told the learner, \"Juz horn once and overtake on the right.\" The learner did as he was told and the as soon as he had overtaken, the cyclist tailed them all the way to the driving centre. At the drivin centre, he created a commotion in the instructor\'s room (sebab he cdnt tell who was who, all instuctors wore the same uniform pe!)\"Wei! Whose the fella who horn me huh?\" To settle things, the supervisor was called and the instructor apologized. Howver, the apology failed to appease the angry zigzag cyclist. \"U all better watch out, u dunno where i came from.\" After that day, whenever the instructors spot this zigzag cyclist ahead, they will tell the learners to either right turn, left turn or U turn to avoid him. Of late, this zigzag minded cyclist is no longer seen around. Perhaps he has returned to where he came from -Changi Prison. ;p    STORY#3: The Taxi Driver    There was once this taxi driver working the late night shift. He drove along Lim Chu Kang area and a lady dressed in white flagged the taxi at one of the deserted bus stops. Being the smart taxi driver that he is, this guy stopped his cab a slight distance ahead of the bus stop. He did it to get a good look of the lady from the rear view mirror.. juz in case.. u know..He looked at the rear view mirror and noticed that this lady who was in all white outfit had messed up hair. He didnt feel too good about it. So apa lagi, he engaged his gear and terus drove off-lah! Next morning.....   \"Hello, Mr XXX. We\'re caling from LTA\"  That lady at the bus stop had filed a complain against him.  '),(13625,4279253,'23,August,2004','2004-08-23 00:00:00','   We went to Knoebels Grove amusement park today.  Left late, Dave (not I) drove and got up there at about 2.  We went right to the pool and spent most of what Dave has set aside for the day.  This was okay - our dinner was in the cooler - hamburgers with all the fixings, watermelon and dessert.  We ate no lunch - I got hungry and asked Dave how much we have for food.  He said $1.50 - not enough.  So I fished my purse out of our locker and pulled up pocket change - about $8.00 worth.  I got myself a bag of chips and we stayed in the pool area for about 4 hours.  Then we left to eat - hungry.  Dave hadn\'t taken the stroller into the pool area, so Elizabeth had to be held onto on the way to our car to get those picnic fixings.  She was tired, and wanted booboo.  Dave suggested he run for the car, repark it closer to the tables and I wait at the table with the girls.  He came back about 15 minutes later hauling the cooler chest in the stroller.  On the way up he had turned on the lights for road construction (the law here) and had asked me to remind him to turn them off.  I thought the lights being on was stupid - the day was bright and sunny and lights don\'t help when the day is already as bright - so I forgot to remind him to turn them off.  Guess what - and you already know.  So, now we need a jump - and have no jumper cables to get one with.  Dave starts the fire in our antique hibachi - dispite the fact I laid some serious cash down on a portable grill for his birthday 2 birthday\'s ago - and has a hard time getting it going.  I pray - and he gets it going fairly decently.  Then he opens the cooler - no hamburger - none.  And not much cash.  So - I pull out that pocket change and Jenny and I walk back into the park and pick up the cheapest dinner we can find - hotdogs.  We get back - give Dave the hotdogs and I mention how it\'s good I had it, and not to worry as only 1 thing has gone wrong with this Knoebels trip (we never have completely good luck at Knoebels since the kids came, or since our marriage for that fact - Knoebels was strictly a good dating spot for us).  Dave says 2 things went wrong, and that he didn\'t want to talk about that second thing.  I start to think he\'s still mad at being up late last night, when infact what he\'s talking about (that he hasn\'t told me about yet) is that dead battery - 1-1/2 hours away from home.  We eat, we do the only thing we have money for (thanks to Dave\'s packrat behavior, and $3.70 of tickets from old trips), we hit a stand for coffee and 1 (and only 1) cone of cotton candy - and go to the security hut.  Really quickly we are in a service cart heading for our car, and on the road (Knoebels packs their own portable jumper kits).  Guess what - it turned our okay for the most part anyway - the kids had a blast, and so did we - thanks to that new baby area (one of those pool areas that has those funny fountain things) and the small bit of change.  I\'m tired.  Elizabeth is asleep.  Dave is falling asleep on the couch and Jenny is trying to copy him.  Goodnight.  '),(13626,4279253,'22,August,2004','2004-08-22 00:00:00','   I went a surfing and came upon a site under \"next Blog\" that played music and featured weird messages in Microsoft Internet Explorer (my computer is a microsoft) in upper/lower case words while music played.  I couldn\'t get out of it, so I just had to log out my computer (the messages wanted me to hit that button underneath the weird message) rather than go into one of those weird sites attached to this whole thing.  Anyhow - excuse my lack of knowledge here on computers - some site in this Blogger journal thing looks like it\'s a virus.  So I logged out, ran my virus protection and I\'m good to go (also - I dumped a bunch of cookies).  But somebody needs to do something about this site - delete it maybe - that is acting so weird.  Okay - done.  '),(13627,4279253,'22,August,2004','2004-08-22 00:00:00','   urlLink    Dave, Jenny and Elizabeth at Cape May zoo&nbsp; urlLink  '),(13628,4279253,'22,August,2004','2004-08-22 00:00:00','   We\'re off to Knoebels Grove today.  This is Jenny\'s last monday before school starts and Dave\'s day off.  Essentially - it is our last chance.  Dave freaked out at the kitchen last night and was up until 3 more or less doing dishes, laundry and etc.  Anyhow - Elizabeth fell asleep on my boo last night (after a day at Hersheypark pushing a stroller I wonder why I just couldn\'t stay awake) and Dave woke me with his banging around.  He said \"go to sleep, you\'ll be driving\" so I\'m on the internet getting directions.  I wish I wasn\'t such a dissappointment to him.  Anyhow - today I drive, and for lack of coffee - I am still shot.  Elizabeth nursed all night, Jenny woke us up.  '),(13629,4279253,'22,August,2004','2004-08-22 00:00:00','   I love to read poetry - but given my age it wasn\'t cool enough to admit to such a thing when I was young.  Stupid - but I don\'t know alot of really good poems as a result.  Here\'s one I just found on a Robert Louis Stevenson site.   Dear Lady, tapping at your door, Some little verses stand, And beg on this auspicious day To come and kiss your hand. Their syllables all counted right Their rhymes each in its place, Like birthday children, at the door They wait to see your face. Rise, lady, rise and let them in; Fresh from the fairy shore, They bring you things you wish to have, Each in its pinafore. For they have been to Wishing-land This morning in the dew, And all your dearest wishes bring -- All granted -- home to you. What these may be, they would not tell, And could not if they would; They take the packets sealed to you As trusty servants should. But there was one that looked like love, And one that smelt like health, And one that had a jingling sound -- I fancy it might be wealth. Ah, well, they are but wishes still; But, lady dear, for you I know that all you wish is kind, I pray it all come true.   It\'s called \"Dear Lady, Tapping at Your Door\" by Robert Louis Stevenson.  '),(13630,4279253,'21,August,2004','2004-08-21 00:00:00','   urlLink    From Hershey Rose Garden&nbsp; urlLink  '),(13631,4279253,'21,August,2004','2004-08-21 00:00:00','   We went to church - and I\'d have paid attention to what the preacher said if I hadn\'t been so upset at my sister Jill.  We went out to lunch at a place called Don\'s Pier and paid too much for too little.  We went to Hershey Park - where we were treated to a long drawn out argument by Jenny called \"why I\'m unhappy - because you won\'t by me any toys here\" that lasted 3 hours, when we finally told her we were leaving if she continued her complaints.  She stopped - and we stayed another 2-1/2 hours.  Now - my feet hurt and I\'m tired.  We have thin steaks thawing and the kids are getting a bath.  The day was beautiful - but exhausting.  done.  '),(13632,4279253,'21,August,2004','2004-08-21 00:00:00','   urlLink    This is Yankee Lady - I thought there needed to be another pretty rose shot here.  So - here.&nbsp; urlLink  '),(13633,4279253,'20,August,2004','2004-08-20 00:00:00','   Dave has just left for work.  Before he left, before we sat down to eat what I had been making when the fight started - we went into a fight.  He\'s comparing me to other men\'s wives.  I feel totally hopeless about this - once they start compairing, you\'re in trouble.  I told him - after we ate an icy dinner of soft tacos and discord - to let me know when he\'d like me to start compairing him to other women\'s husbands.  He apologized. I hope this gets settled without him walking out on me.  I hope it gets settled without him cheating on me.  But he has to come to the point whereby he no longer puts me up for comparison.  And I don\'t know when that will be.  I do know he\'s been doing it for years.   Help.  I need this to stop - because I\'ll always be able to be topped.  Always.  But I\'ve never done it to him - ever.  Will his love ever be true?  '),(13634,4279253,'20,August,2004','2004-08-20 00:00:00','   I\'m having problems relating to Jenny and I really shouldn\'t be.  We are almost dead on the money similar in personality.  But I\'ve been making mistakes with her - ones I should really not be doing because they wouldn\'t work on me (though they are what were used on me when I was a kid).  My parents yelled - that\'s YELLED - and yet I didn\'t change because of it.  Now - I yell - that\'s YELL - and go figure - Jenny doesn\'t even listen.  The instructions said clearly - DON\'T GO EMOTIONAL - go logical, and as I recall when my dad finally did go logical (on a job search issue), and lowered his voice and talked true to me - I LISTENED - go figure.  So - why am I using a tactic that didn\'t work on me on my daughter who is so very similar to me?  Why??????????  Because I didn\'t know better.  Now I do.  Having problems with your kid.  Check out this link from  urlLink Focus on the Family .  You\'ll have to sign up, get a number - then go in with that number to take the test.  done.  Now I just have to look at this really really carefully - and make some changes.  bye  '),(13635,4279253,'20,August,2004','2004-08-20 00:00:00','   I went online - I took a test. I came up with these recommendations to follow. I need someplace to copy them as my printer didn\'t print them out right. Here it is    Recomendation for Jenny  You\'re doing great! We need just a little more information before we give you a Parenting Action Plan. We would like to make some recommendations based on the information you\'ve given us about yourself and Jenny. The following list of Do\'s and Don\'ts can provide helpful guidelines for effective communication with Jenny. When communicating with Jenny Respect her quiet demeanor. Remember that quality work requires quality instructions. Give reason for a change in rules. Support your ideas with facts. Allow time for her to be alone if she wants. Keep your distance. Allow about three feet between you and her. Take your time. Be relaxed and informal. Allow time for questions. Be prepared. She is not influenced when you present your ideas in a random order. Use the correct buzz words for her age and interest. Follow through on your promises. Use an unemotional approach. Ask \"how to\" questions. When communicating with Jenny, avoid these things . . . Don\'t overuse hand or facial questions. Don\'t yell or scream. This creates tension which will affect what she hears. Don\'t present your case with shallow ideas. Don\'t criticize the quality of her work without analyzing the quality of your instructions. Don\'t overuse authority. Keep communication at the discussion level. Don\'t be vague. Be specific about instructions and make sure you allow time for questions or clarification. Don\'t be disorganized in thought or ideas. Don\'t make statements you cannot prove. Don\'t stand too close - allow space between you when talking. Don\'t get overemotional. She is not influenced by emotion. Logic is more important. Don\'t overwhelm her with compliments. She will feel manipulated if you give too many or fail to be sincere. Don\'t be giddy, casual, or informal if you\'re serious about what you want her to hear.    Okay - now there it is.  '),(13636,4279253,'20,August,2004','2004-08-20 00:00:00','   I\'m dumping some stuff that had to do with my old journal - as well as some illustrations I used for it that weren\'t my own, but someone elses and pulled off the internet.  If you see it here - I took it.  Jenny is bugging me to go onto the Barbie internet site and sports news is on.  They just shared a bowl of ramen noodles, leaving the poptarts I put out for them sit.  I\'m going to do some knitting today - I have the pieces for Jenny\'s size 8 red wool sweater almost done.  Canada is apparently going through an early fall - so we\'ll be going through one too (as the jet stream runs right over our state).  She needs that sweater.  I have soft tacos happening for them later.  Dave didn\'t get off until 3:30 AM last night so I\'m going to let him sleep until about noon.  fini.  '),(13637,4279253,'20,August,2004','2004-08-20 00:00:00','   I just deleted my Bravenet account and now I\'m with this blog provider only.  You can\'t go and have 2 blogs successfully - and I can be more honest in this blog than I could be in my other.  So I carried the blogs I like to read from the journal\'s links to my toolbar and now I\'m gone.  fini.  You don\'t do this to gain friends - you do this to help handle your own life.  I used my first journal to help handle my own life - but my second carrier was mixed with trying to get acceptance from \"the community\" - because there were perks to doing this - a little bit of annonymous fame so to speak.  If I wanted fame I\'d have tried to be an actress or something - this isn\'t about getting something like that.  So - now I\'m honest about when my day is bad and I\'m honest without apology too.  No deletes here.  When my day is good I\'ll be spending it doing something besides looking at a computer screen typing.  I\'ll be enjoying it.  But when my kids are making me go nuts - I\'ll dump those feelings into here.  That\'s the purpose of keeping a journal.  done.  '),(13638,4279253,'19,August,2004','2004-08-19 00:00:00','   We just had a major thunderstorm go through that lasted a good 2 hours.  We were getting so many strikes that I was sure we\'d loose our electricity - and I lit a good 8 candles (in jars or sconce candles on our walls).  Anyhow - the electricity possibly going out caused Jenny fear and in fear she wants to crouch under a blanket - so she starts pulling out sheets to hide under - and Elizabeth follows suit and now I have 2 kids throwing around sheets in a house with 8 candles going and the possibility of loosing our lights.  I flipped.  I yelled.  The thunderstorm is gone, and the kids are settled down watching a Barbie video and Jenny is calling out to me to watch this and that on the video.  I had to pay close attention on what was happening and now I\'m just glad it\'s over.  Elizabeth has been flying into the refrigerator to hide (thank God we have shelves on both sides that keep her from being able to close herself in it) and then wandering off leaving the refrigerator door open.  The light is burned out now.  Jenny pulled out more juice boxes - but I was just too stressed to drink.  Thank God that storm is over - so many strikes - and we didn\'t even loose out lights. I don\'t live in Florida - we live in Pennsylvania - and in Pennsylvania just before Charley wracked havoc on the folks down south - we had some tornado scares - touch downs that were relatively close.  No - the second one wasn\'t like what they are used to in the mid west - but the first one was - and happened about 20 miles away from here.  The second weaker one happened closer.  Anyhow - 2 little kids not listening, candles a necessity, sheets flying, my Elizabeth climbing around our stone fireplace and in the kitchen, Jenny running by the candle in the hall and Elizabeth having to follow suit (like most little sisters do) - I\'m just calming down now.  That\'s why I\'m in here. Dave\'s working overtime likely tonight.  And Elizabeth is showing no signs of slowing down (wish she was - I could use the break).  If the time is off we are looking at 11:05 PM Friday night and Dave shouldn\'t be home for another 5 hours.  I\'ll be putting the girls to bed solo as usual - and with a bit more difficulty as I had to yell at them (safety required it) and threaten them (more safety) and they are likely seeing me as a \"tiger mama\" tonight. I saw some poor soul on TV tonight who has tried to make himself appear as a tiger at the cost of $100,000.00.  The man had a cleft lip put in - a collogen upper lip - an abundance of tattooing on his face and arms - numerous piercings on his face for whisker \"earings\" to be put in, grew huge long fingernails he sharpens on both sides and sports a mechanical tail.  Get this - he wants more surgery - that would require him to walk on all fours.  What surgeon would do this to such a madman? Okay - well I\'m calmed down a bit.  Goodnight.  '),(13639,4279253,'19,August,2004','2004-08-19 00:00:00','   I don\'t believe in home schooling under normal circumstances - such as having a normal kid in no need of any special education, or them living at home all the time - not moving around etc.  I don\'t believe in it for the reasons people usually state for homeschooling their kids.  But the school district refusing to send parents information about bussing or their teacher or what time to pick them up from the bus stop or drop them off at the bus stop - I think now is one really really good reason to homeschool.   to homeschool - one must need a kid in need of special attention at home and be lacking in those qualities kids usually consider \"friendship\" worthy.  to homeschool - one must be having to deal with the worst possible school principal or teacher - a Nazi principal - a warden teacher - or the most idiotic school staff possible - an \"animal house\" set of idiots running the place.   Can you tell I\'m just a little upset at the lack of information our school district has seen fit to pass along to me - at this late date - school for us starting in only a little over a week?  '),(13640,4279253,'19,August,2004','2004-08-19 00:00:00','   and time to play on the computer doing things I shouldn\'t to find out if I can.  Nothing illegal, nothing immoral - just stuff that might make this journaling burp.  I\'m still learning stuff about this new blog system - bear with me.  to go color I hit this color thing  and type with   a different   color -   cool.  To link to another journal or anything else   first I go back to black then I type a word - such as  urlLink Grandpa Chuck  then I hit that link thing writting down the link\'s address before I hit the link thing but I highlight the word first and voila - I get a hyperlink (catch that - I\'m hyperlinking now - coooool) and on this journal too.  I already know I can\'t insert pictures into a single entry with text - and maybe that has it\'s plusses.  What happens if I hit one of those funny looking things here I get a billet style with either numbers or dots.     Now when I hit that parenthesis thing I get another indentation.  hit the italics thing and I go italics  hit the bold additionally and I go bold italics.  Hit the straight edge line thing and this lines up for me.    Send it all back the way it\'s supposed to be and I almost go normal.  let\'s hit enter'),(13641,4279253,'19,August,2004','2004-08-19 00:00:00','   Because I don\'t know how to use this yet.  Can you do posts with pictures incorporated into them - or do the pictures have to be done seperately?  Obviously - for this one - seperately.  Dave\'s out with the kids in the kiddy pool.  I have tarragon chicken in the oven.  I still haven\'t had any coffee - IN MY BRAND NEW COFFEEMAKER!.  Yes - that\'s right - my BRAND NEW COFFEEMAKER!.  So stupid.  Time to end post.  '),(13642,4279253,'19,August,2004','2004-08-19 00:00:00','   urlLink    pool time for Dave and the kids&nbsp; urlLink  '),(13643,4279253,'19,August,2004','2004-08-19 00:00:00','   Had something in it more than Dave\'s saying.  All he\'s admitted to is 2 recorders and 1 tuning instrument, plus some guitar strings.  Guitar strings . . . the wheels are turning . . . but I found no guitar downstairs.  I did however find Dave\'s old guitar case . . . did he just get a new case?  I have my tarragon chicken in the oven, Dave has the kids out in the pool . . . Okay - this is too weird.  I thought I lost this when I added that picture.  Move on time now.  '),(13644,4279253,'19,August,2004','2004-08-19 00:00:00','   urlLink    Hawkeye Belle&nbsp; urlLink  '),(13645,4279253,'19,August,2004','2004-08-19 00:00:00','   Dave is off getting me my new coffeemaker - to replace the one that died after only 2 weeks of use.  Jenny and Elizabeth are eating their lunch of silly spagetti (ramen noodles) because the macaroni and cheese Dave picked out for them was buggy (yuck).  I\'m online trying to pull up stuff for some fairs Dave wants us to go to (hate that Bloomsburg one - yuck) and I am hanging on waiting for him to get home with my new coffeemaker so I can make some and so that we can eat our lunch of hot wings (yum).  The ups guy just dropped something off for Dave - from Musician\'s Friend - and it\'s in a huge box (a huge guitar carrier sized box - huuuum?) and I\'m just a little upset because we already have 3 guitars (3!). I\'m glad to have found this - from that crochet girl who keeps one (Kariholtz) - because I can type into it without Jill having a clue as to this.  Jill is my sister.  Don\'t keep a blog with the same carrier as your sister.  Those sibling rivelries continue into adulthood.  Just don\'t. Okay - about me.  I\'m married for almost 10 years to a man with the post office, I\'m a SAHM with 2 kids - daughters aged 2 and 6.  Dave\'s home - time to eat, see my new coffeemaker and most importantly - see what\'s in that great big box!  '),(13646,4279253,'19,August,2004','2004-08-19 00:00:00','   first entry.  Just checking here.  '),(13647,1223044,'16,April,2004','2004-04-16 00:00:00','     '),(13648,1223044,'14,April,2004','2004-04-14 00:00:00','     '),(13649,1223044,'10,May,2004','2004-05-10 00:00:00','   The abuse of Iraqi prisoners in Iraq is disgusting, abhorant and near vomit inducing. American and British soldiers, albeit a minority of them have behaved terribly and individually they should be punished, but at the very top heads must role for these despicable war crimes.  The Red Cross has reported severe concerns of breeches in POW conditions as stated under the Geneva convention, and yes all ye idiots, America has signed up to the Geneva convention. In the British segment Red Cross advice was acted upon, the fact that the abuse ever took place is appaling by itself, but in the American segments abuse continued.  The Arab world will use this again against the west and in particular America, a major figure must go to show that these acts of abuse are simply not tolerated, for the good of all of us. RUMSFIELD MUST RESIGN.  '),(13650,1223044,'04,May,2004','2004-05-04 00:00:00','   Exam stress, a stubborn attitude and a slight arrogance on my part have made the last week difficult to say the least. I have been, in the words of one of my \'friends\' (you\'ll see why I use inverted commas later) a jumped up backstabbing dick. I think this email, sent to me after a pathetic dispute over... well the email explains all:   Dear Richard   You may be correct, it may be us that owe you an apology, and i am sure that we were all prepared to talk about this like sensible adults. However, you may think that you are being mature by sending around an email telling us that we are pathetic, but for once could you maybe take a step back from your seemingly bottomless pit of knowledge, and the attitude that you always know best, and think to yourself \"Maybe i am the pathetic one?\"   For example how did this whole situation spawn, well i can tell you quite easily, it began when we very kindly invited you to come on our holiday with us in the forthcoming summer, what a nice gesture. Obviously not. Somehow you managed to fabricate the idea in your own mind that we didn\'t want you there, which we obviously didn\'t considering that we invited you! You then went on to moan about the cost, location, and company which would also rather piss you off somewhat if anyone else had gone about acting in the same \'pathetic\' manner. When asked if you would be attending the trip by myself and Richard you went on to kick a chair, slam a door and then go and tell Adam and Simon that they wern\'t invited. Very mature. I feel that what you did to them was the biggest backstabbing action which you could have taken, very noble considering that you were supposedly good friends with all of the people on the trip. As you can tell this pissed all of us off just that little bit more, but still i was prepared to hear your side of the story so i came and found you in your history room where YOU found it interesting to bombard Amy and myself with abuse, yet again another honorable action.   Then i attended Science today happy in the knowledge that i would see my good friends Richard and Polly. However, as you may know the mood turned rather sour when you found my very funny joke somewhat unfunny and went about hitting me in the face.(Which i must add was a pathetic punch)   I feel that the trouble with you Richard is that you are somewhat insecure about yourself, your exact words were \"I\'m not really bothered by what you are saying, it means nothing to me\" if you not bloody bothered then why hit me in the face you idiot if your not bothered then why get in such a shit about everything, why do you make everything worse for yourself squaring up to my best friend storming off in a rage, we are not trying to upset you, it is called a joke if you wern\'t so pathetic and took it all with a pinch of salt people wouldn\'t know that it got to you, so wouldn\'t drill the same old insult, personally your poncy little spack today just made the joke twice as funny, as did it for everybody else in the square after you walked off, hopeing to leave in some sort of dramatic fashion when actually all you did was make yourself look like a complete twat. You wonder why people call you boring, or moody it seems inconceivable to me that you can\'t see why, this time last week i still thought that you were a genuine decent bloke but you just went and made yourself look like a complete idiot by trying to make somesort of noble stand, resulting, as per usual in everybody else laughing behind your back.   Richard C was right from the very beginning when he told me that you could be a jumped up backstabbing dick, i gave you the benefit of the doubt, but you had to go and shove a number of friendships right up your own ass. Pathetic.   Written by Nik   agreed by Richard    The joke he is refering to by the way, is writing nazi on a photograph of me and drawing a swastika.  I do have to admit he has exagerated these facts ever so slightly, yet it remains me who is the one who needs to apoligise. Considering my actions of the past few days, weeks even, I can see and I do realise that I have behaved in a manner that isn\'t appropriate, isn\'t friendly and a manner I simply should not have behaved in. I have apoligised to these people, but I doubt my apoligies will mean much.   I need to consider to myself, what do I want to get from this life? Happiness of course, and if this is what I want, why I am going about it like this, clearly the wrong way. Why am i ruining friendships that took an age to build, with people I actually do get on with, despite thier frequent need for racist comments. I have decided the best strategy by far is too keep my head down until way after exams now, but I\'m not sure they\'ll have the same idea. It seems to me that I have taken a feeling of self belief too far and stepped in the realm of feeling some kind or arrogant self importance.  Things need to change, and they need to change now, I can only hope that I can change myself before I lose even more friends.  '),(13651,1223044,'22,June,2004','2004-06-22 00:00:00','   Perhaps a slight exageration, but tomorrow is my last GCSE exam! This momentus event in my life so far will be a 2 hour history exam on the subject of Germany between the wars, now you don\'t get any more exciting than that.  So this is the end of my secondery education, the end of an era you might say. After this point I shall be moving on into the world, tomorrow marks a conerstone in my life.  Okay, so now I\'ve written a lot of waffling nonsense about that lets see whats been going on today. Um, not a lot, football tonight has been funny, final games of group C in Euro 2004. Italy played Bulgaria and Sweden played Denmark. Before the game it was widely publicised that both Sweden and Denmark would progress to the next round if thier game ended in a score draw of higher than 2 - 2. This was of course followed by the famous Azzuri excuse machine going into overdrive, \"it will be a conspiracy\" they shouted, \"we worry they might fix it\" vieri was quoted as saying. So guess what the final score was? Yep 2 - 2, never have I laughed so hard at a football result. So Sweden top the group, Denmark come second and Italy and Bulgaria are going out. This match certainly wasn\'t a stitch-up, but it is a very, very funny result. If Jonsson hadn\'t scored that late equaliser for Sweden, Italy would have been spared.  Alas, tomorrow night is a much more important game for myself, it awaits to be seen whether Germany can actaully become good at football suddenly to beat the Chezch Republic. I shall be out celebrating the end of exams with my phone constantly buzzing with score updates.  Which brings me on to the proposed events of tomorrow night. A piss up in a field. Yes, thats right, we shall be going out and doing the most stereotypical of teenage taboos, we shall be going to a field in the evening with some beers and some cider and getting sloshed. Oh the joys.  For now, I must flee, for there is much last minute, panic revision to be done.  '),(13652,1223044,'19,June,2004','2004-06-19 00:00:00','   urlLink Germany 0 latvia 0 , not a great game at all.   As promised I shall explain my reasons for my disent towards Ukip. Ukip are a party controlled by a idiot, spun by an American and whose lead candidate in the European elections is a bigoted racist with a permanent sun tan.      To move on from the childish reasons, the Independent on June 16th put it nicely, describing just what leaving the EU means, it  urlLink truly is a Â£23Bn question . Not only is it sheer lunancy to leave the EU, the largest free trade area in the world and the greatest single benifit to UK interests, the alternatives proposed by Ukip are terrifying to say the least.  They want Britain to leave the European Union, but they are in favour of economic globalisation, and a stronger role for the US dominated World Trade organisation (WTO). The WTO which is already forcing genetically modified US goods into British supermarkets; the WTO that puts global free trade before environmental regulation; the WTO that undermines national laws to push its own economic rules on nation states. While Ukip say they are concerned about independence from Brussels paradoxically they are perfectly prepared to surrender our sovereignty to the WTO and the US.  Similarly Ukip dangle before the electorate the enticing possibility of joining the North American Free Trade Association (NAFTA) to replace the EU, cementing Britainâ€™s place as the 51st state of America. Unfortunately however this would allow American companies to be free to sue British companies for loss of profits if they tried to impose health and safety or environmental laws that according to NAFTAâ€™s frankly bizarre definition of free trade discriminates against these companies.  Rather more worryingly, Ukip have almost identical policies to the BNP on migration and asylum; many UKIP parliamentary candidates have been former National Front or Northern League of New Britain Party activists, hence UKIP have been dubbed \"the BNP with smiling faces.â€ I only have to point your attention to Robert Kilroy-Silk, who made horrifically stereotypical comments about Arabs describing them all as â€œsuicide bombers, limb amputators and women repressors.â€  Ukip is an unrealistic political party composed almost entirely of white middle-aged men who seem determined to try and return Britain to an imagined 1950s golden age. It\'s not going to happen, Britain is stong when in Europe, strong when at the heart of Europe, strong when she can speak alongside her European allies to provide a counter-balance to the lunatic US foreign policy of late.  Recently after Ukip\'s supposed election breakthrough (15% of the votes from an uninformed and uninterested electorate), they have vowed to \'wreck\' the European parliament, so they don\'t just want to \'wreck\' the UK by withdrawing it, they want to ruin the institution that is bringing economic, social and political benifits to the rest of the EU. Selfish? Go and play your mindless political games elseware Mr Kilroy-Silk and let the people of the EU enjoy the peace and prosperity that a stable and united Europe has brought to us.  '),(13653,1223044,'19,June,2004','2004-06-19 00:00:00','   Exams are all but over now (I have food technology and History left but they really arn\'t important), so the long summer has all but begun. This can only mean one thing, time for a lot of drunken parties, lounging around doing nothing in particular and eventually and very reluctently accpeting that I shall have to get a job once I begin to run out of money.  As I write this its half time in the Germany vs. Latvia match in Euro 2004, I expect I\'m one of very few people in England cheering on Germany, often in a match like this the whole nation become honourary supporters of the opposition. The English have always disliked the Germans and as far as I can see always will, no doubt there are very deep physchological reasons for this none of which I can be bothered to venture into now.  Throughout the country theres a lot of rather messy upheaval regarding Europe and Britains place in the EU, something I feel quite strongly about as a person who considers himself to be \'European\' rather than \'British\' (this is because one of my grandparents is Irish the other German) and I feel particuarly strong ties to both these countries. I\'m personally disgusted by the gains of the far right UK independence party at the European parliament elections last week for reasons I shall venture into after the second half of the football.  Caio for now.  '),(13654,1374061,'06,July,2004','2004-07-06 00:00:00','   Went over to Seamus last night for dinner; it was just the 3 of us. Seamus has been doing his house in Collinswood up recently; new wooden floors and all the rooms have been painted. Also he had taken delivery that day of a new Ford Mondeo. After dinner we watched Switchblade(?) a movie written, directed and starring Billy Bob Thornton (I can never take that name seriously) - really very good, thought the chef didn\'t enjoy it much. Seamus will be 40 next week, and he\'s having a large party at his home the following week; P. is doing the food.  Got up early and brought Scott for a walk on the beach; the boiler has packed up (again) and it\'s quite chilly; very grey and overcast in Dublin today.  I want to read more of \'Memoirs of The Future\', then get going on organising my thoughts on Utopianism. I may even start a new blog to marshal my thoughts.  Trim tomorrow to see Barbera on stage - not only has she done all the scenes, artwork & posters for the new production, she\'s starring as well! What a gal!  '),(13655,1374061,'06,July,2004','2004-07-06 00:00:00','   So, this evening I (finally) get time to explore Blogger a bit more. I\'m an occasional diarist, but I\'m also a note-taker and an archiver. I have lots of different diarys/logs in lots of different media, going back to the mid 1980\'s.  Lately I\'ve begun to get back into Utopianism, and I\'m reading \'Memoirs of The Future\'. Once again I feel like talking, talking, talking about the future, and of our (I think) desperate need to figure out better ways of local and global government.  If nothing else, the current invasion of Iraq by the USA goverment shows us that any wealthy group will stop at nothing to secure and continue it\'s fortunes. Not only does represenatational government fail at national levels, it\'s lethal at international levels.  The best way to deal with any problem is to solve it. All systems (as I know from work) continually evolve. However the question is so broad, covers so many questions, is based on so many suppostions, it needs a total re-think and re-org at several levels. I want to get a bit more success this time (not sure what I mean by success, though), so I think that a ground-level examination of the entire subject would be a good starting point. Hmmm.  '),(13656,1374061,'06,July,2004','2004-07-06 00:00:00','   I presume everyone sends a quick \'Hello World!\'-type post to see how it works and what it looks like?  '),(13657,1374061,'06,July,2004','2004-07-06 00:00:00','   Just watched the director\'s cut of \'Cinema Paradiso\' on DVD with P. What a movie. I love it (nostalgia), but Peter really, really loves it - it says and means a lot to him. I made us a nice dinner to have before it; tomorrow I\'ll go home to Trim for a few hours (Salvatore returns!)  '),(13658,1374061,'06,July,2004','2004-07-06 00:00:00','   Well, I finally got Mum & Dad down to Kilkenny over the week-end, Sunday in fact. Jentilla & I went in my car from Dublin, and Mum & Dad took John from Trim. I had a table booked in Rinnuccini\'s for 1330. We had a really nice meal, then headed over for Clashers.   Gran must have heard my requests, because we had beautiful weather. So, they had a quick look around the house (Dad checked out the sheds whilst Mum looked the house over upstairs & downstairs. Tara Mines had had some major problem over the week-end, so Dad was being really good to come down; his boss was on the line trying to get Dad to come back to solve some production problem.  Anyway, they both liked it but thouoght it was way too much work to take on. Dad worried about the electrics (which probably do need a good overhaul), whilst Mum fretted about gypsies murdering us. John was Spy-Gor in the back of the car on the way home, and he sent us an email summary of what was said.  So by the time we were leaving Clashers it was already 1700. We went on to Kilfane with the plan of seeing the gardens, but there was a cloud-burst, and so we called it a day. Once again, thanks to all those who helped organise the sunshine for when I needed it most; appreciated!  The chef was dying to know what was said, and drove me cracked for the last few days waiting for news.  Work is mad at the moment (Moodys Terminal Server installation), plust we\'re going on strike Friday, so looks pretty eventful. Bar-B-Que Sunday, then New York Tuesday. Yaay!  '),(13659,1374061,'06,July,2004','2004-07-06 00:00:00','   Well, I\'m really not using my new blog, am I? When I kept a diary as a teenager, I wrote in it every day (how gay is that?!)  It\'s been a week of unrest at work. In fact, last week my laptop hard drive went ka-boom, which was perhaps a small fore-warning of what was to come. The Bank published a list of the employee benefits that will and will not be transferring over to HP. For the stuff that can\'t go across, they offered a measly €400 (gross) per year per employee. Because this is the first cold-cash amount up fpr the grabs, the union balloted it\'s members for strike action, and got it.  The last few non-members of the union in my dept. quickly signed up, sans moi, naturellment. Gary tried to pressure me into joining up, the chump. So the chef and I went over and over it, but in the end it was my decision. Really, I agree with the union position, and I\'d never cross a picket line anyway. My main problem with this was the support I\'ve gotten from Colin over the last few years; I was afraid that if I signed up to the IBOA, somwhow it would be a slap in the face for him. However, having spoken with him in private, I discovered this would not be the case. So I signed up (really, the only thing to do given I don\'t care for the pressure from wither side, and I can\'t become a manager), but I did so in secret; let whoever finds out find out. The whole thing reminds me of the stuff that went on in Tara in 1996.   Also I was speaking with Conor, who\'s working on some of the stuff on the HP end. He\'s begining to find out what the Bank can be like, especially the attitude of ITS. I\'m sure it\'ll all work out, things have a funny habit of doing so, but just not in the way that one thinks it will.  The skip arrived Thursday (after a mix-up earlier in the week), so Thursday evening P & I shifted all the heavy stuff from 31 (the cooker, washing machine, dryer, sofa, etc) and the same from 29 into the skip. Tomorrow (well, today really, as it\'s 0039 12.07.2003) I\'ll just dump all the garbage that\'s been hanging around the house here.  I noted that as a HP employee I\'ll have the right to work from home (if posible) and also P will be covered on my VHI; should be handy when we move to Clashwilliam.  That\'s it really; the tickets for New York arrived this week. I\'ve been looking forward to this for so long I can\'t really focus on it any more. It seems to unrealy.  Did a little bit more work on \'Walking With Scott\'. I got loads done on the ending; just need to fill out the middle, really.  '),(13660,1374061,'06,July,2004','2004-07-06 00:00:00','   So here I am in New York with the chef. I\'m using the hote TVs sucky access, so I have no real dea if this is getting though or not.It\'s 0035 EST Monday 11.08.2003. We\'ve been having agreat time; walking arund is very tirying, toug. We\'vemet up with Joe, seen 2 shows (Chicago  The Producerw), beent the Met, The Frik, The Whiney, Emire State Building, Atlantic City (P won $200!), we\'edn it al, really. Tomorrow is the Planetarium, and shopping. I miss Sco a lot, and am looking forwardto coming home. I\'ve been here before, so y trip highlight has been seing the portrait of Thmas Mooe holbei the Younger) in the Frick. That\'s it the spllingis  disastr caus of th junk remote-control keybord. Love NYC, but hav had enough for thenext few years! The W=A is ice.  '),(13661,1374061,'06,July,2004','2004-07-06 00:00:00','   Industrial Action!  There\'s nothing I hate more than taking sides. Of all the stupid, ignorant, primitive ways of humans, this is the worst. It\'s our \'get out of morals free\' card, and it stinks, stinks, stinks.   Anyway, the facts are that today we\'re all on a 24-hour stoppage because the IBOA (bank union) want to get the best deal going for their members. Whilst I\'m a member on paper, I have ZERO respect/regard for the union (or the bank, for that matter), so I\'m at home. Monday is a bank holiday, and P & I are off to NYC Tuesday, so this is an early start to my holiday.  So far, I\'ve stayed up until 0330 playing \'The Sims\', didn\'t get up until 1000, brought Scott for a big-walk at Malahide Castle (it\'s a beautiful sunny morning in Dublin), and now I\'m webbing and abount to have a long shower. Then I\'ll head for The Pavillions. So not so bad, really.  '),(13662,1374061,'06,July,2004','2004-07-06 00:00:00','   OK, here we go!  We\'ve now officially moved to Clashwilliam! The last 4 days have been a Big Bang of boxes and bags - but we\'ve finally done it - we now live in Kilkenny!  Very late and tired as I write this, so that\'s all I\'ll write for now. Lots to do tomorrow on Day 1 of being a Kilkenny guy (!) - but as I\'m on leave for the week, I should have time to write over the next few days. Whilst packing this week-end, I found my old dairy that I started in 1985 (as a very highly-strung 15 year-old), ansd continued up to age 26, just before I came out. So I really want to do some catch-up work on that!  I love the fact that tomorrow is 01.10.2003!  '),(13663,1374061,'06,July,2004','2004-07-06 00:00:00','   Just time for a quick blog; settling into Clashers, very tired (stayed out till very late last night to study incredible stars with telescope). Also met with Tom Duncan, and had a meal in Rinuccini\'s.  Lots of stuff going on, but we hope to be settled down within a few weeks - we\'re still lacking some of the basics like a cooker and a shower! I have to get on to the electrician tomorrow and see if he\'s going to come at all or not.  Busy, busy, busy...  '),(13664,1374061,'06,July,2004','2004-07-06 00:00:00','   I just learned that The Galileo probe will be sent on it\'s final encounter with Jupiter on Sunday. I\'ve been busy following the life of this probe since it\'s launch in 1989 (I was 19!!!), and more especially since the web came in the early 1990\'s.  It\'s funny - I really feel for the little guy; a trusty little probe (well, not very big) that outperformed it\'s own mission criteria by a factor of 4. It\'s designers and operators deserve the biggest pat on the back going.  In order to protect Europa, and I suppose to have a spectacular end to a spectacular mission, it\'ll go down in a blaze of glory, in the freezing, gargantuan clouds of the Jovian atmosphere.  Thanks Galileo!   urlLink Galileo End of Mission -- Webcast  '),(13665,1374061,'06,July,2004','2004-07-06 00:00:00','   Clashwilliam, Gowran, Co. Kilkenny.  My first blog post in Clashwilliam! I brought my laptop with me from work Friday, and as our phone line is now working, Clashwilliam is now connected to the farthest reaches of the web!  It\'s a beautiful September morning as I write. Pete\'s working again today, so once again Scott and I have the place to ourselves. I washed my hair in the sink (roll on power shower!) and then Scottlesburgh & I took a walk. I love this time of the year - blue skies, the leaves turning autumnal, still some heat in the sun. Scot munched down some grass then barfed it back up again (euuh!), so he\'s miffed that I won\'t give him his dish. As we recently took down the stud wall in the hallway, the room is filled with sunlight from the entry door. The laptop is on Pete\'s old writing desk (one of the items that got moved last week), and Lyric FM is playing in the kitchen - what taste!!!  Yesterday I went into Kilkenny at got a few things for the kitchen. I also got a new pair of shoes for work, and a zip-up wardrobe (to keep our clothes in some semblance of order whilst we hunt for some nice wardrobes).  Tom the electrician called round @ 1500, and we went through the work we\'d like him to do. I made a bit of tea, then Pete rang to say the PC in the boss\'s office was in trouble - could I come and help? So I went over, met Pete\'s boss (and his very handsome son), quickly established they had the WormBlaster virus, and cleaned it up. Whilst I was working they pumped me as hard as they could to find out who I was, how dod I know Peter, etc. I didn\'t really say anything, but when they quizzed me on why I was down with Peter, did I know the house, etc... Peter\'s boss (the wiley old so-and-so!) then went to Peter and said \"Alan says he moving down to the house in Gowran as well...\" which I didn\'t say at all, but he picked it up and tried it see what Pete would say. Anyway Pete had already decided that he would use the next available opportunity to out himself, and decided that was it. So it went down OK, no screaming fits, and they gave him a big tin of Roses to give to me, as a thank-you for fixing the computer.  David called over for a while, then we just settled down for the evening.  I have to decided if I\'m going to town this evening, then come back tomorrow, or I could head for town now, do all my laundry, bring down clothes, etc, then come back tonight. I should finish the painting in the guest room as well, of course. Plus there has to be a major de-webbing... bleh.  I love it all!  '),(13666,1374061,'06,July,2004','2004-07-06 00:00:00','   This week-end has been one of the greatest and happiest of my life. Yesterday we borrows a van from a guy who used to delive fruit & veg to Pete when he worked in Malahide. We went over to Pete\'s mum & dad, loaded the van and the Jeep with stuff from their house. We moved some of the couces and chairs from their house up to 31 Kinsealy, then loaded the van & Jeep with more of our stuff to move down to Clashwilliam. I was a bit nervous about driving the van (large Ford Transit), but I soon got into the swing of things, and so we were off!  I had a real sence of being home when we approached Clashwilliam - the the first time it\'s turned into a place to my place, for me anyway.  So, we did an enourmous amount of lifting, & moving, much to the entertainment of Pete\'s dad (who came with us). It was a beautiful sunny day, a real September day, my favourite. I\'ve never seen Clahers looking so good. Barbera, Jentilla & John arrived later in the evening, and had a look around the house. I\'ve never seen the house so sunny or full of people (Ben Quinlan & his daughter turned up as well). Pete was *so* pleased with how his great-great grandfather\'s table looked in the sitting room, he was so happy.   We then drive back to Artane, took even *more* stuff from there and moved it to Kinsealy, then took all of our heavy items (chairs, the sofa, and endless, endless boxes), and loaded it all up, and down we drove again. We finsihed around 0200, and headed for bed. Had a great lie-in (!) and got up around 1300 Sunday. We went for a walk with Scott, and moved evrything from the van into the house.  So I\'m tired (and filthy! We really need a shower there), and I\'m back in Dublin. I\'ve just changed my address on all the utilities I\'m registered with, and I want to chill a bit before next week (which is going to be murder with Moodys). I won\'t get to see a lot of Pete, as he\'ll be working in Kilkenny, and I\'ll be working late in Cabinteely to finish everything off for Moodys.  '),(13667,1374061,'06,July,2004','2004-07-06 00:00:00','   Saw the Aurora Borealis for the first time last night! I noticed what looked like flood-lights far-away on the way home in the car. The night was clear, with a crescent moon almost fully set to the west. As the sky was so clear, I knew it wasn\'t fog; then I remembered John had taken pictures of a huge sunspot last week, so I twigged it was the Northern Lights!  I called everyone I knew, and when I got home I went out to the lane to have a look. The glow was bright enough to make out the time on my watch! Most of the time there was a faintly-green glowing band, to the north, stretching west to east. However, from time to time, I could see bands, or ribbons, with perpendicular sharp spikes - wonderful! I watched them for about 30 minutes before the cold and the lure of my tea got me back inside!  Later we took Scottlesburg for a trot, and they were still bright in the sky, though not as clearly defined.  So I\'ve finally seen the Northern Lights - life goal achieved! Yaay!  '),(13668,1374061,'06,July,2004','2004-07-06 00:00:00','   Well, it\'s really late at night (or really early in the morning, which ever way you want to look at it), and P is in Dublin and I\'m in Clashwilliam. He  had to do an out-door this afternoon, then his parents are coming down to Kilkenny tomorrow, so he\'s staying with them and I\'m down here.  It\'s my first night on my own here - it\'s oK - I\'ve spent the evening online, playing games & talking to Thomas. T & C are going through the mill at the moment, thought it seems to be drawing to some kind of conclusion. Even if T does go with C to Melbourne, I\'m not convinced that he sees why he should. It really makes me wonder. Anyway, really tired, so off to bed!  '),(13669,1374061,'06,July,2004','2004-07-06 00:00:00','   So today was our first real, full days of being real, full Kilkenny residents!  Pete went off to work, so I got up after him and took Scott for a walk up the laneway. It was a beautiful sunny morning - slightly misty, so there were golden sunbeams everewhere. It was so clean, and quiet - beautiful October weather, though warm enough to be like early September. The sky was white, white, white around the sun, then changing to high blue to the west - a fabulous morning, and one that had me in high spirits for the rest of the day. I just wish Pete could have been around, though of course if he wasn\'t working, he would have been in bed anyway! I\'ve always been an early riser - surely a sunny morning is the best time of any day - the world is fresh, and sunlight like promises, spilling everywhere and filling the whole world with magic. How wonderful to be in such a beautiful place on such a beautiful day. I should note that I saw a rat by the gate near the main road - it didn\'t try to run when Scott ran over to investigate - it just shivered. On the way back, it looked as though it was trying to turn around, but couldn\'t. Of course I left it alone (so did Scott), but I guess it was poisoned, and I didn\'t think that land was poisoned anymore?  Anyway, back to the house, and after a quick bit of dish for Scott, we headed into Kilkenny. I left Scottlesburg in the car whilst I went to the County Council offices for details on refuse and recycling. Next I got paint for the bathroom, frying pans, groceries, over trays... everything that was on my list.  Back to Clashwilliam, we had a bite of lunch, then went on a long walk down the avenue. I called into Liam & Mary, who were busy painting their living room. I think they\'re delighted that we\'ve moved down at last.  After we came back for our walk (today, the sun never stopped shining), I tackled the kitchen, and basically unpacked, put away, and organised everything.  Next I called John, and by the time we were finished talking, I had run out of energy and enthusiasm for doing any bathroom work at all! So I finished cleaning the kitchen, washed and hung out some tea towels. Pete rang me @ 1800 to say that Mary and Des were over in Shankill (about 10 klicks away) to look at a cottage that Mary is thinking about buying.  Finally, back to Clashwilliam with P where we had our first home-cooked meal (a real mix of chicken & sweetcorn soup, rashers, sausages, beans, toast & cake!)  So that was Clashwilliam day no 1! Couldn\'t be happier!  '),(13670,1374061,'06,July,2004','2004-07-06 00:00:00','   Yesterday Pete had a disaster day at work - his supervisor read him the riot act, and criticisms him very unfaily about stuff she had never mentioned before as being a problem. Major hassle ensued, and Pete threatened to resign. All sorted out in the end, with phone calls from both the lady in question (Anna) and the owner of the business Tommy Roache, who begged Pete not to leave.  We went out for dinner to Cafe Sol (I had booked a table for P\'s birthday), and we talked it through. Pete decided he didn\'t want to leave yet, and told the owner in a subsequent phone call.  So today he was back in work, and I did a fair amount of house work, including some DIY plumbing (fitted a new exterior drain-pipe to take bath and sink water.)  We\'re heading over to Dublin tonight for Bernard\'s 40th birthday, and tomorrow is hang-the-bathroom cabinet day!  Jentilla headed to London this afternoon for 4 months in McCann\'s office there.  Tonight one if the NASA Mars rovers lands on Mars - here\'s hoping they have better luck with their landers than the Beagle2 group did! I may get up early to watch the broadcasts live via the Net.  '),(13671,1374061,'06,July,2004','2004-07-06 00:00:00','   Well, I see that my last posting was back in October, when the aurora was glowing above. As today is New Years Day, I\'m full of goodie two-showes resolutions, all of which I take seriously. It\'s just a shame that past performance doesn\'t auger well!  Still working away in BoI and commuting daily. The \'work-from-home\' thing was a bust - bit of lie, really. Given that I don\'t have any respect for my employers, and I no longer care about my career except in monetary terms, I don\'t really care!  I did out a mind-map on my plans for 2004 - quite a list, but I hope to get through it.  Pete is 37 today - he\'s working of course, so we have to wait until tomorrow evening before celebrating.  The house is coming on OK - we had to get a new concrete floor in the living room in order to put down wodden floorboards. That work should resume next week. There\'s a number of household tasks on my list ofr 2004 - if we accomplish all of them, we\'ll be dpoing well!  Made a few on-line donations today; Red Cross appeal for the Bam earthquake survivors and SightSavers International.  Pat is here today, so we\'re going to head into Kilkenny for a few hours, bring the dog, etc. Back to work tomorrow - boo-hoo; I want o play Sims!  Looking forward to 2004!  '),(13672,1374061,'06,July,2004','2004-07-06 00:00:00','   Just a quick check-in. It\'s a grainy Sunday, but things are pretty good, all told.  I suppose the main news is that I\'m very seriously hoping to get redundancy from the bank, as we\'ve decided to go for the outdoor catering / b&b in a serious way. We have to decide the best way to fund it, but I just want to do this now. I can\'t figure out if I just want to be out of the bak, or if I really want to stop IT work. I\'m so tired of it, my little bit of self-confidence has been gone for years, and I know it\'s not the right thing for me to do any more. It\'s hard to see how I\'ll get the VSP (Voluntary Severance Package), but if I do; whoo-hoo!  We talked about marriage last night; I\'d love to marry Pete - it\'ll probably br 10 years before the topic even comes under the national forum though!  Did quite a bit of domestic and house-keeping yesterday. Today is personal development, shelve designing, bathroom cleaning & grocery shopping.  Went to \'Cold Mountain\' with Pete last night; average. Jude Law; phew! What a bod! Yummy!  '),(13673,1374061,'06,July,2004','2004-07-06 00:00:00','   Star  urlLink  '),(13674,1374061,'06,July,2004','2004-07-06 00:00:00','   Tendrils  urlLink  '),(13675,1374061,'06,July,2004','2004-07-06 00:00:00','   Sphere of Life  urlLink  '),(13676,1374061,'06,July,2004','2004-07-06 00:00:00','   Had a good week-end, thought the chef is very stressed out. After his outdoor on Saturday (which went very well, the Jeep broke down on the way home, near Carlow. As luck would have it, I had just put him on my AA membership, so he was able to get the think towed as far as Gowran. Howeverm he\'s got a bid outdoor tomorrow in Dublin, and now he\'s no car. So I\'m going to give him my car, we\'ll go up tomorrow morning, and I\'ll make my own way to work from the Red Cow.  Beautiful week-end weather-wise, check out the pictures.  '),(13677,1374061,'06,July,2004','2004-07-06 00:00:00','   urlLink Saturday 05 June Murphy\'s Cat Laughs 3-7 June \'04   Tom & Stephen are coming over next week-end, so hopefully we\'ll be able to catch a show!  '),(13678,1374061,'06,July,2004','2004-07-06 00:00:00','   So the ideas are coming in thick and fast - a really tumultuous week as far as Mr. Brain goes. prot, the tadpoles / vegetarianism / veganism and now I\'m finding out much more about global warming, and the consequences of climate change.  Of course, Shrubbie (that\'s George W. Bush) has banned all government scientists from commenting on \'The Day After Tomorrow\' movie, the cur. We seem to be headed for global fry-up, rather than freeze-up, but the issues is the same.  Jentilla mailed me this link this morning. There\'s loads of food for thought in it, and I want to get going on estimating our CO2 emissions from travel / power & heating-oil usage. I heard a scientist from the UK on the radio last week explaining that we really, really are in big trouble weather-wise, and therefore everything else-wise.  Other news; Thomas is 33 today (he bought a new PC), plenty to do tomorrow, but I may be helping Pete with his 2 outdoors tomorrow instead - he\'s working until 0000 to prepare.  Feeling very unsettled - almost like I\'m going through my mid-life crisis (about my 9th!)  '),(13679,1374061,'06,July,2004','2004-07-06 00:00:00','   Quiet day today; my replacement is working really well, they getting a good guy! Checked the tadpoles when walking Scott tonight - saw three (worryingly, the three I put in the \'quiet\' part of the pool - no sign of the guys who went into the stream side. Told Pete I wanted to become a vegan - I think he thinks it\'s just another fad - hopefully it is\'nt! Had chicken enchilladas for tea, not a great start I suppose.  '),(13680,1374061,'06,July,2004','2004-07-06 00:00:00','   The remains of a fly (?) seep oil into the \'sheep-dip\' by the stream. The stream eventually pulls this away.  urlLink  '),(13681,1374061,'06,July,2004','2004-07-06 00:00:00','   Sunday; hard to see, but there are 80 - 100 tadpole at the bottom of this small puddle.  urlLink  '),(13682,1374061,'06,July,2004','2004-07-06 00:00:00','   Well - what an evening. Not sure if I mentioned the tadpoles in a previous post or not. Anyway, there was a small puddle close to the \'sheep-dip\' resevoir by the stream, and this week-end I noticed between 80-100 tadpoles in it. (See the picture) I even topped up the puddle with water from the stream at the week-end.   Anyway this evening when I got home I saw that a tractor had been, driven straight into the puddle, churned it all up, and to cap it off, it had flushed out diesel into the mud. The top of the mud was thick with dead flies (I think) - but I noticed some tiny ripples in the mud - a few of the tadpoles were still there and alive.  So I got the empty salad-box & water bottle from lunch & topped up the sticky puddle with stream water. It was pure brown ooze, but I could see tadpole-shapes, wriggling.  I spent the next hour getting the tadpoles into the bottle and (as carefully as I could) transferring them to the little pool of the \'sheep-dip\'. The water became really oily from the ooze that went in along with the tadpoles, and some of the dead flies floated on the top. Rainbow-swirls of diesel Catherine-wheeled from them as the oil seeped away from their bodies and the little current in the sheep-dip made them slowly spin.  I think I rescued between 3 and 6 of the original 100 or so. Three of them swam straight down the bottom of the pool, wriggling away in what looked a healthy manner (given what had happened to them today, and that they were covered in oil.)  Given recent thoughts & communications, and thinking about this rescue mission (which has upset me) I\'ve decided I *need* to be a vegetarian; I can\'t behave like this then tuck into a steak sandwich, can I? I may as well go the whole hog (N.P.I.) and become a vegan - why go half way?  Poor little tadpoles - I hope at least one makes it to froggy-hood; also I need to learn more about tadpoles & vegans.  '),(13683,1374061,'06,July,2004','2004-07-06 00:00:00','   Domestic day. Did dome webbing (email from Gene Brewer!), then off to Carlow to buy internal doors. Looks like we\'re going to have problems with sitting-room and kitchen doors, as they are so old and small. Spoke with a nice assistant called Barry who explained door sizes to me. So it seems that we may have to speak with a joiner and see what the cost of getting some (or all) of the doors made to order.  It was a beautiful Saturday, so I drove back to Clashwilliam, dropped off some groceries, then Scott & I went on to KK. Town was packed - I browsed a bit in Meubles (drool drool), but refused to pay €700 for a TV stand. So in the end I got one in Argos, then spent 3 hours putting it to gether at home. As Pete says it\'s a bit \'Wayne & Waynetta\', but it\'ll do fine for the time being.  Then off out to watch the ISS overhead @ 2306 - beautiful evening, the moon a slender crescent setting in the west.  '),(13684,1374061,'06,July,2004','2004-07-06 00:00:00','   Al  urlLink  '),(13685,1374061,'06,July,2004','2004-07-06 00:00:00','   urlLink    Just trying out Hello & Blogger - took this picture Sunday 16.05.02004 at Graiguenamanagh  '),(13686,1374061,'06,July,2004','2004-07-06 00:00:00','   So last night I saw Bushie & that gom Rumsfeld basically leering and the entire world; the smuggest, most evil and sly creatures to ever wriggle out of the rot-wood of greed.  If I could have the ear of all Americans, I would say this; Look! Look now! Your (restricted) democratic system only allows for 2 parties. The leaders of those parties only get to where they are by having money, big money, and by promising others the chances of getting even more of it. You are now ruled by the most ruthless of people who can only have the interests of their bank-rollers, period. American government is now too privatised - it\'s only about personal gain. The ideas of leadership are totally gone - the most vicious beast gets to the top.  The same was true in Iraq - a dictator and his two sons got it all by fear, control, extortion and torture. The cruelty of that torture was infamy itself. Now The President of The United States of America is publically thanking the Secretary of Defense for showing the same level of brutality. It\'s hardly been 1 year. And no-one, NO-ONE is looking around saying \"we all know there was no valid \'War On Terror\' to be fought in Iraq.\" No-one is saying \"this was for oil, not \'thank you, Secretary of Defence, for defending our country (by allowing low-level soldiers to torture low-level Iraqis)\".  It\'s all over. So let\'s just be sure we all know where we all stand when presidents stand tall, and people stand low.  '),(13687,1374061,'06,July,2004','2004-07-06 00:00:00','   Well, it\'s been like 100,000,000 years since my last post, but I\'m still here.  Well; where to start? Clashwilliam is coming on well, the house is now fully painted (inside, anyway). We got a load of furniture, and the last major item (sitting-room sofa) is on the way - perhaps even this week. Now that summer has arrived the house is looking great, and the painting is really helping the inside light levels. During the week-end I realised I was feeling more familar and \'at home\' than ever, malking me realise that I wasn\'t as settled as I thought (or perhaps have ever been).  Pete is well (except for a small problem he told me about this evening). He\'s trying to let 29 & 31, and it\'s a major pain. I think he\'s considersing selling 1 and renting the other. We had quite a good week-end; the twins from malahide came down Saturday morning, and Joe arrived in the afternoon. the weather was great - it\'s summer! Joe, the chef & I headed into Kilkenny for a meal in Rinuccini\'s - beautiful. Lithuania was having it\'s accession into the EU celebrations in the town, so there was lots of activity.  Was speaking with Barbos earlier today. She\'s liking Australia very much, but I don\'t think she\'ll stay somehow - I hope not. I hope she can come stay with us for a while when she gets back, probably Septmeber.  Had the worst 2 weeks ever in work; really really bad. Major over time though, especially now that I\'m contracting. May be getting a place down in Kilkenny - pie in the sky at the moment.  Enjoying Sim City 4 Rush Hour (and listening to the extra sound tracks at the moment) - also doing quite a bit of photography, which I\'m really enjoying.  This is me, Dale, signing out! A.  '),(13688,1374061,'06,July,2004','2004-07-06 00:00:00','   Earlier today I came across a visual recording of the Cassini launch in 1997. I remember watching that launch - the first live event I watched via the Internet - and wondering what I would be up to when the probe arrived in 2004, then 7 years into the future.  I thought I\'d be living in Canada by then. I was in the middle of applying for a working visa at the time. Having visited Canada 2 years before, I had visions of me in a skyscraper looking out on grey Vancouver skies or the glass elevator that brings tourists up the CN tower.  Regardless of where I was living, the ideas of   when   I\'d be living were even more exciting. So many things were up-and-coming for the 21st century; the web would be everywhere (it is), mobile devices would be in everyone\'s pocket (they are) and global warming would continue to cause major weather events (it did).  2001 had always been a goal for me (I   still   think those 4 digits look beautiful; and exciting). So trying to imagine what it would be like when that year was 3 years in the past was really too much - I couldn\'t do it.  And now here I am and here Cassini is; less than 72 hours from SOI (Saturn Orbit Insertion) - the end of a journey that took 7 years and over 3 billion kilometres. The JPL/NASA site for Cassini gives a time-line for Cassini\'s voyage, which makes we wonder what my time-line for the last 7 years would be like. And here\'s a question that I shows the different ways I look at things; I wonder if I will ever do the same trip, and look at Saturn the same way that Cassini is doing, as I write. It\'s massively unlikely I won\'t do it corporeally; but I wonder in spirit?  October 15, 1997 -- Cassini-Huygens launches from Cape Canaveral Air Force Station in Florida October 15, 1997 -- I watch the launch and check the status of my Canadian visa application  April 26, 1998 -- Cassini-Huygens flies by Venus, picking up a boost from the planet\'s gravity.  April 26, 1998 -- I\'ve spokwn with Trish and decoded there\'s no point running away to Canada. I decide to go contracting.  June 24, 1999 -- Cassini-Huygens flies by Venus again, getting another \"gravity assist.\" June 24, 1999 -- I post a card for my brother\'s 18th birthday. I\'ve already met Pete, and can\'t get him out of my mind.  August 18, 1999 -- Cassini-Huygens gets a third celestial push when it flies by Earth. August 18, 1999 -- I\'m finishing my contract for CIE, and hoping to get into the Bank. Pete & I have met for a coffee, but he\'s seeing someone else.  December 30, 2000 -- Cassini-Huygens flies by Jupiter, snapping photos and getting a final boost. With Galileo still orbiting the planet, it\'s the first time two spacecraft have explored the gas giant simultaneously. December 30, 2000 -- Pete & I are together almost 1 year. We\'ve bought Clashwilliam, and are planning to renovate & move down.  June 11, 2004 -- Cassini flies by Saturn\'s moon Phoebe, uncovering evidence that the moon may be a frozen artifact of a bygone era, some four billion years ago. June 11, 2004 -- I\'ce taken reduncance from the Bank, but am working an extended 3 month contract with them. We\'ve traveled around Europe & USA. We\'ve moved to Clashwilliam, and I\'ve been commuting +160 miles a day, hence the main reason for leaving the Bank.  July 1, 2004 -- Arrival at Saturn (12:12 a.m. EDT). July 1, 2004 -- Leave full-time employment for the 1st time in 15 years.  '),(13689,1374061,'06,July,2004','2004-07-06 00:00:00','   Rose, Manderston formal gardens.  urlLink  '),(13690,1374061,'06,July,2004','2004-07-06 00:00:00','   Manderston, where we went on our last day (the itinerary chnaged a little). We had lunch here with Lord Palmer.  urlLink  '),(13691,1374061,'06,July,2004','2004-07-06 00:00:00','   Dryburgh Abbey, beside the hotel. Field Marshall Haig & Sir Walter Scott are buried here.  urlLink  '),(13692,1374061,'06,July,2004','2004-07-06 00:00:00','   Traquair Castle.  urlLink  '),(13693,1374061,'06,July,2004','2004-07-06 00:00:00','   Tom at the same spot!  urlLink  '),(13694,1374061,'06,July,2004','2004-07-06 00:00:00','   Pete & Scotland. Taken near Newton St. Boswells, Scottish Borders.  urlLink  '),(13695,1374061,'06,July,2004','2004-07-06 00:00:00','   Nacho & John check their pictures in Glasgow.  urlLink  '),(13696,1374061,'06,July,2004','2004-07-06 00:00:00','   Statue wearing traffic-cone-hat outside Glasgow gallery  urlLink  '),(13697,1374061,'06,July,2004','2004-07-06 00:00:00','   Pete @ Mellerstain  urlLink  '),(13698,1374061,'06,July,2004','2004-07-06 00:00:00','   Mellerstain castle. Evelyn is on the right.  urlLink  '),(13699,1374061,'06,July,2004','2004-07-06 00:00:00','   If an eagle just landed on you, you\'d be surprised too!  urlLink  '),(13700,1374061,'06,July,2004','2004-07-06 00:00:00','   When eyes meet...  urlLink  '),(13701,1374061,'06,July,2004','2004-07-06 00:00:00','   Me and the barn owl!  urlLink  '),(13702,1374061,'06,July,2004','2004-07-06 00:00:00','   Pete meets a barn owl.  urlLink  '),(13703,1374061,'06,July,2004','2004-07-06 00:00:00','   Being welcomed at Paxton.  urlLink  '),(13704,1374061,'06,July,2004','2004-07-06 00:00:00','   Alan at Floors  urlLink  '),(13705,1374061,'06,July,2004','2004-07-06 00:00:00','   Fountain detail, Manderville  urlLink  '),(13706,1374061,'06,July,2004','2004-07-06 00:00:00','   As we did so much on this trip it\'s hard to know how to lay it out; a simple day-to-day, I think! I\'ll just start by saying this was the \'Treasures of The Tweed Valley\' by our good friend and ciceroni, Tom Duncan (http://www.ciceroni.co.uk/tours/Tours200304/Tweed%20Valley.htm)  We flew to Glasgow on Sunday and were met by John & Nacho at Glasgow airport. They brought us back to John\'s apartment in Glasgow, which is beautiful - very spacious indeed! We had a good walk around the centre of Glasgow, then had a meal in Cafe Rogano; which was beautiful. Next day (Monday) John drove us to Edinburgh airport, where we met up with Tom & some of the tour group (the rest were arriving by their own means).  Off to the hotel (Dryburgh Abbey, http://www.druburgh.co.uk) where we unpacked and had some light refreshments then headed to Floors Castle (http://www.roxburghe.bordernet.co.uk/floors) where we had a very informative tour, and some more snacks. The main feature of this house which I remember is the unpolished beech furniture everywhere. Historically, the house was important because of it\'s connections with rebellion & strife during the 16th & 17th centuries.  In fact, we did so much over the week (touring, learning, dining with lords, ladies, dukes & duchesses), I can\'t write it all down. Check out http://www.ciceroni.co.uk/tours/Tours200304/Tweed%20Valley.htm for an accurate itinerary.  I will say that we met some lovely people, especially Sophie, Ina and Evelyn. My favorite paintings on the trip were by Francesco Guardi and Steenwijk. Check out the pictures!  Next trip is a week sailing on the River Shannon in August with Tom & Stephen.  '),(13707,1374061,'06,July,2004','2004-07-06 00:00:00','   OK, this is  alittle strange, I suppose. Just before going to sleep we were talking about the strangest things we\'ve ever seen. I told Pete about the UFO I saw in 1992. SO I drifted off to sleep, and about 30 mins later P is waking me to look at a light that was moiving in the sky.  So. we saw a yellow-ish light, looking a bit like Jupiter but yelloer, and much more of a disc apparant. I knew Jupiter had already set, so it wasn\'t jupiter. Located in the North West, about 30 degrees, and bobbing away, gently.  We tried to film it with the camera, and ended up getting up, getting dressed, getting my telescope and heading off down the laneway to get a better view (right about now the movie audience would have been screaming \"Are you guys crazy?!?! Don\'t go!!!\" So off we went, and just watched the thing. Inthe telescope it was just a yellowy disk, still just little more than a point. Definitely not Jupiter - no atmosphere bands or Jovian satellites. Over about 15 minutes it slowly went from 30 degrees to about 10, until we could no longer see it over the trees.  And that was that! This reporter thinks \"weather balloon!\" - though they\'re not usually so low.  '),(13708,1374061,'06,July,2004','2004-07-06 00:00:00','   It\'s a crap emotion, but because I\'ve always taken everything personally, I can\'t help but feel personally diminished when I do something careless.  So I had to apologise to both the iTSIS (\'HP\', whatever...) Project \'Manager\' and the BoI guy today for not answering my phone yesterday (Sunday). In fact, the problem turned out to be some work that was going on the mainframe, so I wouldn\'t have been able to do anything anyway - it would have been a total waste of a journey and a Sunday, so I suppose in the end I didn\'t do too badly.  Beautiful hot day today - hopefully some clear skies tonight; the first in ages. Did someone say \"telescope...?\" :)  Read a bit about \'Longhorn\', the next version of Windows. Me want.  '),(13709,1374061,'06,July,2004','2004-07-06 00:00:00','   Just a quick line to say this is coming from my new desk - took from 0730 until 1700 to build (with about 2 hours of a stop) - but the desk looks great. I still have a room of garbage to sort out though, and it has to be done before Joe arrives Friday.  The entire week was fairly manic work wise, as I was doing tickets with Keith, and 1426 suddenly blew up, all over again. We do seem to have gotten it all sorted, though.  Texts from Barbera - she\'s back in Melbourne (I think), and Thomas, who has a new PC and is addicted to Civ3. I myself had a Sims all-nighter last Sunday (bank holiday) - Io swear I\'ll never stay up all night again - this time I really mean it!  Pete has been under major pressure too, so Saturday I sorted out No. 31 whilst he did a day for Mary in the castle. I guess I\'ll be helping a lot more with stuff like this whilst I\'m *UNEMPLOYED* from 01.07.2004! His mum & Dad and Owen too came Monday night to stay - first time we\'ve all been together under our roof! We gave them our bed whilst we slept in the guest room - Owen had a sleeping bag on the floor.  Not a whole lot else - studying orbital mechanics (kinda for the interest, kinda for \'Orbiter\' simulation, and kinda for the K-PAXiness of it all!  Hopefully going to see \'The Day After Tomorrow\' later on with P.  '),(13710,1374061,'06,July,2004','2004-07-06 00:00:00','   urlLink BBC NEWS | Science/Nature | Planet Venus completes transit   Well, I forgot my telescope (that\'ll teach me, see next entry...), but it was totally overcast here all day anyway. I did get to see the sun briefly on the way to work (at the dolmen, actually).  Oh well, 12 years time then!  '),(13711,1374061,'06,July,2004','2004-07-06 00:00:00','   Well, it\'s been a busy week - though still no word on the job front. I was supposed to hear (yea or nea) by Friday, but no word at all - maybe no news is good news?     &nbsp;   Today (Saturday) was pretty busy - helping P get an outdoor ready for tonight; plus doing loads of work *on my own* to prepare for Sunday\'s outdoor for Fingal Council. Today I learned how to dice an onion! I then proceeded to make 2 loafs of bread, Greek salad for 50, potato salad for 50, *34* chicken fillets, beef for 50 (gasp), plus preparing all the cutler and loading the van!     &nbsp;   There\'s more to do, as when Pete gets home around 2300 we\'ve got to wash all those plates, bowls, cutler, then get them ready for an 0700 departure tomorrow morning. I\'m wrecked - definitely not used to standing all day.     &nbsp;   On top of all this prep work we had David, Aneka &amp; Aaron down yesterday - entertaining &amp; prepping at the same time.     &nbsp;   So - where am I? At the start of the year I determined to try to join a futurist group, but I\'ve not done that. OK, I could subscribe to the World Future Foundation, but that seems like just another magazine.    I totally want to give my thoughts on the future, science &amp; society, a frame-work; but what??!! I don;t have an idea for a book (though it doesn\'t have to be a meisterwerk, I suppose) - I\'d love it to be more a novel, which explains my ideas about the future without being a treatise.     &nbsp;   What do I think about the future? Ultimately, time will continue, without me, and with or without the current human race from Earth. Period.&nbsp;Eventually of course the sun will expire, and we\'ll either have moved on enough so that it doesn\'t matter&nbsp;for our survival, or we\'ll have figured out a way to&nbsp;keep it as is.&nbsp;     Why  do I think about the future - ahh, the ?64000 question.    The Universe is everything - and everything is the Universe. That means that the Sun is a bit of the Universe, your fillings are little bits of the Universe, the Virgo super-cluster is a bit of the Universe, your smile is a bit of the universe, pi is a bit of the universe,&nbsp;George Bush is (unfortunately) a bit of the Universe and the collection of neurons in my brain that made&nbsp;me add the \'(unfortunately)\' comment just now are also bits of the Universe.    If I\'ve learned anything in my 34 years, it\'s that everything that is, both large and small, young and old, is,  together , the Universe. If you\'ve ever studied anything, you will surely realise that arriving at this conclusion is as&nbsp;incredible as any revelation.       So here we are, and to me it seems that this simple truth just doesn\'t make it out into the world, in terms of everyone\'s day-to-day business. It\'s like shopping and kinds and work and TV and travel and lunch and cars and pamphlets and advertising and money is more immediate, so it gets more coverage - personally, locally, nationally and globally.     &nbsp;   I think, IMHO etc, that we\'ve learned a lot, and it\'s a shame to not allow our knowledge to lead us on. We seem to think we need leaders, and those leaders should be  other people ; why is this?     &nbsp;   Am I so cynical; do I really believe that  human nature  is what stops us from being eternal? I do; I do I do I do.     &nbsp;   I\'ve ranted enough; the day you start to belief that the world would be a better place if everyone thought like you is the day&nbsp;to turn into a tyrant, right?  '),(13712,1374061,'06,July,2004','2004-07-06 00:00:00','   Well, we\'ve been having a fine old time of it. We spent the week getting a bit of a garden in order, reading, cooking, relaxing; it\'s just the best!  My folks are coming tomorrow (Sunday), so we\'re trying to get the house in order before they arrive. Hopefully they\'ll like the place - they haven\'t been before Christmas, so the place is vastly changed from when they were here last. Poor P. is a bag of nerves!  Jentilla is having a blast in Oz; she really, really likes it there. That\'s about it! I\'ve been panicked, not-panicked about working. Yesterday I mailed my friend Alanna, and I have to admit I felt a little, well, like a sell-out because I was looking for another job. But we need the cash!  Happy :)  '),(13713,1374061,'06,July,2004','2004-07-06 00:00:00','   Well, we\'ve done it. Not only are we both living done here, but yesteday was my last day at work in Cabinteely. As would friend Gary would say, I\'ve cut the cord! Actually, yesterday was momentous not just becase I left 15 years of continuous employemnt (for one firm or another), but also Pete traded in the Jeep! We now have a silver Ford Courier van (which I was driving around Kilkenny all this morning!). Also, Mary called over to the house, with a beautiful antique wall clock as a gift!  As if that weren\'t enough, Cassini made SOI perfectly in the wee small hours of this morning - that\'s a real milestone for me because I\'ve been watching the mission progress with considerable interest since launch in 1997 - see other posts about Cassini!  So, Pete was in Archersrath today, so I brought Scott for a walk, cleaned the kitchen, loaded the van with our domestic waste & recycling. Off to Dunmore with that lot, then back to town to retrieve Pete\'s West Link barge which he left in the Jeep. Then I ordered prints from our photolist, and got the making of a lunch before heading back home. I carried the bed for the guest room upstairs and assembled it - though my drill ran out of charge, so i had to stop that for a few hours.  Now I\'m off to have lunch before settlingdown to... not sure. Finish \'Angels & Demons\'? Sims? Ah, un-employment!  '),(13714,3051898,'24,July,2004','2004-07-24 00:00:00','   Wow I\'m up early... look at the time of post if you don\'t believe me... I woke up not long after 5am, but didn\'t get out of bed till 5:45. And then I came online, but nobody is online to say bye to. I did cry in bed last night, but I\'m feeling okay now. Not super happy again I\'m afraid, but I\'m alright. Wow, this really will be my last post before I go on holiday.... I\'m going to see if I can ask Chris not to be online when his mom goes out at night so that I can call him... maybe I\'ll be able to finally tell him that saying he loves me and then saying he doesn\'t is actually really upsetting for me. And maybe he can cheer me up because I\'ll ask him if he\'s been thinking about me (which I reckon he will :) ) I bet my parents are glad that I\'m awake.... they were going to wake me up at 6:30, but for reasons not even I understand I woke up completely on my own not long after 5am.... Odd. I\'m tired but I\'m not going to get back to sleep... I guess I\'ll probably just have an early night tonight.  Nobody\'s going to read this for a while except maybe Lauren, so Lauren I\'m warning you now there\'s loads of repetitive crap written on some days!! I hope you enjoy finding out how things between me and Chris are seeming different.... we act more like a real couple, even online last night we were asking each other if we\'d miss each other and saying how we wish we were together and doing the /skiss thingie on msn and wishing they were real kisses. Actually thinking about this is cheering me up more and more... Also the fact that when we\'re alone we usually kiss and hug and stuff (in fact we\'ve kissed a little bit with Louise in the room..... but I usually try to resist the temptation when she\'s around). I guess just because he doesn\'t love me it doesn\'t stop that things are great between us and that I reckon he\'s on his way to loving me (or possibly does but doesn\'t realise as he doesn\'t know what love is). Ok, now I know I\'m going to be okay, I\'ve reassured myself.  When I come back off holiday I might try to put some holiday photos on here.  Bye for 2 weeks! (unless I find a computer in Spain... *has hope*) ~Hannie xxx  P.S. A message for Lauren: Lauren my parents still don\'t like the idea of Shell Island... so I don\'t think I\'m coming. I\'m really going to miss you over these holidays.... When we\'re all together we\'ll have to see loads of each other to make up for it!!! And I won\'t moan at you too much for wearing revealing clothes! lol. Buy some credit so you can text me, as I will text you. I\'m already missing you, I hope you miss me! Bye for now, luv Hannah. xxxxxx  '),(13715,3051898,'23,July,2004','2004-07-23 00:00:00','   I\'ve been thinking.... please can everyone not mention the not loving me thing to Chris? It\'ll probably only cause problems if anyone does mention it... I would\'ve been fine if he hadn\'t said he did love me. It was raising my hopes and then dashing them that makes me cry. Louise had actually told him to say he loved me anyway.... but she told him to only say it if it\'s true. It\'s really mean of him to say it and not mean it. It\'s cruel. I hope I stop crying soon, and that I don\'t cry in front of him tomorrow when we pick Louise up. I suppose he is only 14... he is too young to be in love really.... But I would\'ve been okay. I was happy thinking he didn\'t love me until he gave me hope. I then went really happy knowing that he loved me, but then, suddenly he says he doesn\'t know what love is. Which means he doesn\'t love me. Which means he made the mistake of giving me hope and then taking it away. So now I\'m crying.... I suppose it\'s probably just hormones again. You know what? Maybe I should just put the convo we had up here so you can see for yourself what happened... I will cut out parts that are too irrelevant. It was hours earlier that he was saying he loved me so that\'s not in the convo below.  \"My House, My\"&nbsp;is Chris (full version of name is My House, My Rules, My Way) \"I may be an\"&nbsp;is me&nbsp;(full version of name is I may be an idiot, but at least I\'m a happy idiot)  [21:24:20] My House, My: do u think im ur true love  [21:24:46] I may be an : maybe... i dunno if i can be sure at only 16 tho  [21:24:57] I may be an : why, do u think i\'m ur true love?  [21:25:11] My House, My: well to be fair i dnt kno wat love is  [21:25:14] I may be an : ah  [21:25:43] My House, My: id like to think u r tho [21:27:01] I may be an : so basically ur sayin u like to think i love you but u dnt kno if u love me cuz u dunno what love is?  [21:27:11] My House, My: yeh  [21:27:33] I may be an : i wish you did know  [21:27:41] My House, My: i try to keep out of emotion at my age  21:28:47] I may be an : so are u saying u dnt want to love me? :S [21:29:05] My House, My: i do wanna love u  [21:29:10] My House, My: just not yet  [21:29:16] My House, My: cus skoll  [21:29:24] My House, My: and house  [21:29:26] I may be an : :( [21:29:38] My House, My: i do wanna love u  [21:29:48] I may be an : you knowloves not meant to just happen when u want it to  [21:30:03] My House, My: i dnt kno wat it is  [21:30:11] My House, My: so i cant relli say  [21:30:28] My House, My: all i kno is that i want to love u  [21:30:33] My House, My: and be wiv u  [21:30:39] My House, My: and tlk to u  [21:30:43] My House, My: kis  [21:30:44] My House, My: u  [21:30:49] My House, My: and hug u  [21:30:56] My House, My: and tlk and kiss and hug  [21:31:23] My House, My: and love u to bits  [21:31:28] I may be an : yeh wen ur in love u just wanna be with the other person and u think abuot them all the time....  [21:31:44] My House, My: how do u kno  [21:31:51] I may be an : cuz i love you   At this point he had to go. I don\'t think he any idea how upset I am.  Anyway by the time I\'m back off holiday hopefully I\'ll be happy again. This is why it\'s probably better if nobody says anything to Chris... I guess I\'m cheering up a tiny bit... I mean, he says he wants to love me... (but not yet :(:(:(:( ) I don\'t know, 2 weeks away from him might help me sort out how I feel about things properly. Hopefully. Maybe.  On a different note, Lauren I miss you!! Buy credit for your mobile so you can text me please!!!!!!!!!  Bye for 2 weeks! ~Hannie xxx  P.S. ALWAYS REMEMBER THAT NO1 CAN KNOW ABOUT THIS BLOG UNLESS YOU ARE LAUREN OR LOUISE OR ALISON. NEVER TELL ANYONE ELSE OF MY BLOG\'S EXISTANCE PLEASE! I ESPECIALLY DON\'T WANT CHRIS TO FIND OUT ABOUT THIS, BECAUSE IF HE FINDS OUT HE\'LL WANT TO SEE IT, AND THEN HE\'LL KNOW IF I EVER SAY BAD THINGS ABOUT HIM...  '),(13716,3051898,'22,July,2004','2004-07-22 00:00:00','   Chris lied. He doesn\'t love me. He doesn\'t know what love is. He says he wants to be with me, he says he wants to hug me, he says he wants to kiss me, he says he wants to talk to me, he says he wants to love me. But he also doesn\'t want to love me yet because of school and whatever. Love isn\'t optional. He can\'t just love me later on when it\'s convenient for him. Why did he say he loved me earlier if he really he doesn\'t know? Didn\'t he realise I would be upset later? He says he likes being with me, he says he always thinks about me. I take those as signs of being in love.... but if you want to be with someone and you think about them a lot does it mean you are in love? I don\'t think it\'s fair to say people will know when they are in love, I think that often people don\'t know... It seems really unfair that I love Chris and he doesn\'t love me.... and he got my hopes up and everything. :( I don\'t know how to feel. He still cares about me, he\'s still going to really miss me when I\'m gone.... but I thought he loved me and he doesn\'t. I wish somebody could explain love to him. I guess my best explaination is when you think about the other person all the time, you always want to be around the other person, they become the person who means the most to you in the whole world... also other people interest you less. Plenty of people have crushes on loads of people, but I think a sign of being in love with a person is when your feelings for other people become weaker, and your feelings for the one you love become stronger (although I guess this isn\'t always the case...). You\'re in love when a simple kiss can make you feel happy, but if they aren\'t interested you become miserable. When you\'re in love the other person\'s happiness is more important than your own, and you would do just about anything to stop them being sad. Also if you can see yourself having a future with the person and being happy, that\'s another sign. Anyone who reads this, please comment on how you think being in love feels like. Maybe Chris does love me and just doesn\'t realise... he did say he doesn\'t know what love is, so how could he know if he loves me or not? If you think he does love me, feel free to try to convince him. I just have one very important rule. NEVER LET HIM FIND OUT ABOUT THIS BLOG.  Bye for now! (or probably for 2 weeks) ~Hannie xxx  '),(13717,3051898,'22,July,2004','2004-07-22 00:00:00','   Today I was really really really happy, but then I felt pretty sad. Then Chris came online and told me he loves me so I\'m happy again! He loves me! He loves me! He loves me! He loves me! He loves me! He loves me!  Can you tell I\'m happy? Heehee. He\'s great, he can cheer me up so easily. :D  Life is good again!  That\'s all I wanna write right now  Bye for now! ~Hannie xxx  '),(13718,3051898,'22,July,2004','2004-07-22 00:00:00','   Life is good, life is great. I\'m in a happy mood at the moment. My family is great (well, sometimes :P), my friends are great (I would prefer they were happier), Chris is great, I\'m great.  urlLink Click here for a webpage that annoys Alison!  I find it funny though...  Alisons happy today which is nice, Louise is in quite a good mood (or at least was earlier), Chris is happy because he\'s&nbsp;thinking about the house he\'s going to buy in the future (would be good if he really did have this house), I\'m in a good mood because Chris makes me happy. Lol.  Tomorrow\'s my last day before I go on holiday... Which means I won\'t see Lauren for ages (I leave here on Saturday, she comes back here Sunday) and that I won\'t see Laura for ages and I wont see Alison for ages and I wont see Chris for ages. Would be nice to get a goodbye kiss off him but I don\'t think it\'s possible. Louise had a plan to get her mom out of the room but I don\'t think it will work out so perhaps we shouldn\'t bother. Firstly theres a good chance my dad will come with me into the house (if I say he can stay he\'ll prob be like, \'no i dont mind coming\' argh), and secondly Louise probably won\'t get her mom out of the room for long (and Louise is gonna hate readin this, but I\'d really prefer a goodbye snog....). Also even if the other stuff works out then Louise would be in the room with me and Chris. So I\'d feel a bit weird kissing him with her there, and also she\'d probably be upset for the rest of the day. Overall, (that word reminds me of history essays...argh) its probably not worth attempting. Pity really.  Well even without a goodbye kiss I\'ll probably be in a good mood, so I that means I don\'t need one! Wooo holiday in Spain! It\'ll be great, swimming everyday, Universal Medditerannia often, nearby neach, visiting Barcelona. Plus I\'ll be with 2 great friends, Louise and Stephen. It\'ll be great!!! So why aren\'t I actually that excited inside? :S Ah well. I\'ll enjoy it when I\'m there.  Bye for now! (Tomorrow will probs be my last entry before holiday) ~Hannie xxx  '),(13719,3051898,'21,July,2004','2004-07-21 00:00:00','   Stopped the night at Louise\'s last night and I spent most of today there. Things are going absolutely great between me and Chris. I don\'t think things have ever been so good! I\'m dreading the fact we\'ll probably have an argument to ruin it all sometime soon... I guess I\'ll just have to try really hard not to get upset by him doing something stupid and also to not start an argument. If I sense an argument coming I must back down immediately!!!  I could tell you the gory details of the time spent with Chris, LOL. Though already the details are slipping from my mind... Also I don\'t think anyone really wants to know. Actually Louise tends to want to know to reassure herself... hmm. Well she shouldn\'t get so upset because everything is bound to collapse soon. :(   Its a shame Louise was getting upset today... In fact her crying interrupted Chris\' attempt at making me horny (he was horny so he wanted me to be, but I just sorta wasn\'t... I basically just wanted be cuddled and kissed...)  I think Louise is a little happier now (dunno if she will be after reading my blog) but I felt sorta guilty that I was spending so much time with Chris (but I loved every minute with him!). Unfortunately I was annoyed that me and Chris had to stop... I felt like I was being a really horrible person, but I really did just want to stay with Chris and keep kissing and stuff. I wanted Louise to be happy, though. She was so upset I did recognise that cheering her up was important.... Even if it did make me feel sorta frustrated to have to jump off Chris to go look after her.   I liked the discussion we had about marriage... Me and Chris actually discussed the idea of getting married! LOL!! Nah don\'t worry I don\'t mean we decided we were going to get married or anything like that...  If we are both still together and happy in the years to come then marriage is entirely possible. He even pointed out that technically we could get married in two years time! I did mention that we would be too young then, but then later on he mention the fact that we could in 2 years time again.... Perhaps he\'s hinting he would like to marry me? LOL, I realise that he was just mentioning it cuz that\'s sorta scary. It scared me anyway.  He seemed to want to discuss the future a lot. It seems his life plan includes moving to America (something I\'ve wanted to do for years), living in a big house so friends can visit (a great idea!), owning 2 dogs or possibly a dog and cat (he said 2 dogs but then he knew I wanted a cat so said a dog and a cat so now I don\'t know what the plan is), being really rich (this has always been a life goal for me actually), and having kids (I\'ve always wanted kids until more recently when I\'ve thought it would hurt and it would make me feel old). In fact he wants a boy and a girl, so 2 kids. Sounds reasonable to me. Actually when we were discussing it later on (when Louise wasn\'t in the room) he even asked me what I\'d name my kids! I\'m beginning to think he would like us to have children together.... (don\'t worry Louise, we won\'t get this far... I bet we\'ll have that terrible argument that breaks us up for good any day now :( )   I should be happy. I\'m not smiling though. I think I feel guilty cuz of Lou. I think I feel sad that Louise and Alison are sad. And at the back of my mind I\'m concerned for Rachel. And I miss Lauren (I hope she\'s having as much fun in France as I am here when I\'m with Chris :D).  Apart from those things, I\'d say my life is great at the moment. Things have never been better with Chris, and I\'m not arguing with my family (having a tidy bedroom helps). I really wish my friends could all be happy. Although it certainly seems like I\'m very preoccupied with Chris, I do care a lot about them.... They\'re all really important people to me. Please can you guys just give me a hint how to make you happy? (hopefully without stopping me and Chris enjoying ourselves)   I feel I\'ve written enough to keep Louise depressed for a while. She really shouldn\'t read my blog, it\'s bad for her health. With any luck she\'ll concentrate on me saying how I want her to be happy and how I care about her as&nbsp;I do all my friends.   Btw, you probs guessed but my mission to stop talking about Chris has been failing really badly. Perhaps I\'ll succeed more on holiday (I leave Saturday for 2 weeks, so&nbsp;no blog updates for ages). I\'ll let you know how it goes (if I remember).   Bye for now!  ~Hannie  xxx  '),(13720,3051898,'20,July,2004','2004-07-20 00:00:00','   I think I\'ll start by saying I am very happy with Chris lately. He has been noticably nice to me and I have to say I\'m very pleased with the result. I do sorta miss Lauren... I hope she\'s having a good time in France. I said some mean things last time about being glad she\'s gone... and I do regret that. Although admittedly since she\'s gone I have had better moments with Chris.... hmm. I can\'t remember if I mentioned that we were having a garden party for relatives... it was last sunday. I was worried I\'d be stuck looking after all the kids but I wasn\'t in the end, plus my room is still amazingly tidy! Oh yeah, I almost forgot about my wasp sting ordeal, I have to write about that. This happened Saturday, so the day before the garden party. I woke up at 10:30am, which is strange because I normally would wake up at least an hour later, and also because I had fallen asleep much later than usual. But the reason I woke up was because I got stung. A wasp seemed to have stuunng me while I was peacefully sleeping (I didn\'t actually see the wasp, but I heard buzzing so I\'m guessing). I would\'ve just stayed in bed as I would usually every morning when I wake up - but it hurt more than I expected. So I got up and called my mom and ran downstairs, but I started feeling really dizzy. I collapsed onto the floor and was carried into the living room where my parents layed me down on the sofa. I finger really stung and I went really pale (except for the finger that got stung - that went bright red) as I had almost fainted (my hearing went weird and everything). I stopped feeling dizzy but then had REALLY bad stomach pains. I felt sick... and then I was sick... quite a bit actually. After this I got better with just a painful finger. Sunday night I stopped at Louise\'s house. I had a great time with Chris as we got some time alone... I won\'t go into to details abut what happened though. In fact tonight I\'m going to stop over again. I can\'t be bothered to type cuz I wanna day dream about Chris just to annoy Louise, heehee. &nbsp; Bye for now! ~Hannie xxx  '),(13721,3051898,'15,July,2004','2004-07-15 00:00:00','   Today I went to the cinema to see Spiderman 2 with Louise, Lauren, Rachel, Alison and Chris. It was nice that we could all go together. Unfortunately Louise and Alison got all upset... Anyway, the film was quite good. I sat by Chris and he held my hand through the WHOLE film. My hand felt sorta numb by the end though because of this... I hadn\'t wanted to let go as he doesn\'t often hold my hand, so it was nice. After this Rachel went home and the rest of us went to McDonalds, where we could see Alison and Louise were still sad... we decided to all come back to my house. On the bus Chris told me he wouldn\'t come to the cinema with us again becuase people were just being sad (Louise and Alison)... which I was dissapointed about. I like going to the cinema with him... At my house Louise and Alison were still sad. Chris kept trying to get Laura to come over, and Lauren spent a lot of time with us. Louise and Alison sat in my room on their own crying together... then the rest of us went upstairs, too. Chris kept messing with my webcam...(which reminds me, I need to delete lots of bad pics of me). I kept trying to get Chris to hug me... but I got the impression he didn\'t  really  want to. Louise felt miserable... she felt like Chris was stealing her friends away. I feel that there wasn\'t anything wrong with me spending time with him as I spend way more time with her and I\'m his girlfriend. Lauren, however, doesn\'t have an excuse. I think Louise and Lauren should talk about this... Actually that\'s a good point. Why is Lauren spending so much time with me and Chris? I guess nobody could really tell that I actually wanted to be alone with him, to see if he was more likely to do stuff this way.... But again, she\'s supposed to spend more time with her friend than her friend\'s brother! I realise people could say the same to me, but I still have a perfectly good reason to be spending time with him, and it\'s unfair for people to get upset because of it. So how do I feel right now? I would\'ve expected me to be happy... but by now I feel sortof down. Perhaps it\'s because everyone was embarassing me about my unnderwear that was all over my bedroom floor, or perhaps it was the insulting picture Chris drew of me (which he hadn\'t actually intended to be insulting....). On the one hand it does seem like Chris was being nice and was making the effort I wanted to treat me nicer than everyone else, but at the same time I do feel sad. I think it might be because I was hoping for more intimacy... but I spose what can I expect from a 14 yr old? Also when I\'m in this sort of mood I don\'t think there\'s any pleasing me. I think I\'m sorta bummed he didn\'t kiss me when he had the opportunity... I mean we had moments when we were alone and I was wishing he would take advantage, but he never does. Actually at one point when we were in front of the computer in my room he seemed to be motioning that he wanted a kiss... which I just didn\'t want to give. It seemed unfair because Louise was right there in the room... but I wonder if it would have changed things if I had gone for it? Either way he has plenty of way better moments to ask, so that makes him just silly. I just hope he doesn\'t get the wrong impression and think I\'m becoming uninterested in kissing... I wish Louise didn\'t get so upset... if she feels as though she\'s losing us why doesn\'t she talk to us more and steal us back? Or maybe she should convince Lauren to stay with her as Lauren has no reason to spend time with MY boyfriend. (Heh, I guess the amount of time Lauren and Chris are together annoys me...) Oh yeah, today I got told I smell. Apparently it\'s just a Hannah smell,&nbsp;and not a bad smell.... but I think everyone was just being too kind to say that I stink. Apparently sometimes after I leave Louise\'s house Chris can still smell me. Wow. I must reek. Actually thinking about it this way it\'s no wonder I never get the attention from him that I want, as he would probably rather avoid the smell. I guess this is yet another thing making me sad. &nbsp; Well I\'m getting repetitive as usual so I\'m gonna stop typing and hope Louise doesn\'t get too upset whe she reads this. As usual this&nbsp;blog is not intended to be nice about other peoples feelings, it is intended for me to be honest and frank about mine. This is a blog for me, not for my friends. If they read it and get upset, it\'s their own fault and they should learn that maybe reading this blog isn\'t the best idea. I guess I did sorta want it to be personal anyway as I knew all along I would say things that people wouldn\'t want to hear... &nbsp; Bye for now! ~Hannie xxx  '),(13722,3051898,'14,July,2004','2004-07-14 00:00:00','   HAHA TODAY STUFF HAPPENED. I\'M SOOOOOOOOO NOT IN THE MOOD FOR THIS. I\'M HYPER!!!!!!!!!!!!!!!!!!! (JUS DRANK DR PEPPER)  HAHAHAHAAHA  TODAY HAD UPS AND DOWNS BUT I\'M IN A PRETTY GOOD MOOD   TO FIND OUT MORE OF WHAT HAPPENED TRY  urlLink LOUISE\'S BLOG . OR MAYBE  urlLink LAUREN WILL TELL YOU ON HERS .   BYE FOR NOW!  ~HANNIE  XXXXXXXXXXXXXXXX  '),(13723,3051898,'13,July,2004','2004-07-13 00:00:00','   The best word to decribe today is dissapointing. It was just full of dissapointments. First I\'m trying to work out how to travel to Alton Towers, but I\'m just not finding anything that does exactly what I need it to. Considering the rest of my day this was the least of my worries... Louise invited me and Lauren over, so I went and got dressed, made sure I looked decent, and then her mom comes home and tells her that actually we can\'t come over. So obviously I was quite dissapointed to lose an opportunity to see Chris. Right after this she asked again about going to Alton Towers tomorrow, and her mom said no. A little while after this, me and Lauren decided to go to Bearwood so I could buy Rachel a card (I already have her present) and so she could buy Rachel a present (she already has her card) as today is Rachels birthday (no she hasn\'t invited us to do anything... I don\'t think she wants to see us :(. As usual there was limited choice of cards saying 16... I thought it would be nice to get her one with people holding footballs. A boys card, but surely Rachel would be happy with this as she loves football.  While we were shopping Lauren told me how she didn\'t have enough money to go to Alton Towers, so I said I could lend it her, but she said she didn\'t want to owe me money. And then she told me how that morning her mom told her not to make plans for tomorrow as she wanted to go shopping with Lauren. Lauren hadn\'t even told her mom we might be going to Alton Towers!! ARGH! When we finished shopping we realised if we were quick we might see Chris after school, but then we saw the bus just slightly ahead of us... and then it zoomed off to the bus stop where Chris would\'ve got on... so I didn\'t even see him in the distance, but missed him by about 2 minutes. Aftfer this I figured it might be nice to surprise Alison by meeting her after school, so I waited outside the school gates (on my own, Lauren had to go pick up her little brother from school). I was waiting half an hour and she still didn\'t show up, so I left. I later found out that she went to her friends house. ARGH. I went home expecting Chris to be online, he wasn\'t. :( He did come on a bit later though. He was being nicer than usual, so I was actually quite happy talking to him. Until Laura came online. He and Laura were talking for ages. They didn\'t tell me what they were talking about, but they were both too busy talking to each other to really say anything to me. It was really upsetting. I felt really left out. They only know each other because of me!!! Admittedly Laura did give me a chance today to say I didn\'t want her to talk to Chris... I didn\'t tell her that I didn\'t want her to though, because I know it makes me look bad. I really don\'t like them talking... especially on days like today when they ignore me to talk to each other more. Laura then went on to block Chris... this resulted in him talking to me again. He was all annoyed and upset, though... it seems Laura ruined his previous good mood (at this point I was thinking how this proves talking to her is bad for him). I\'m worried because Laura was quite mean to him and it might have been because of me. I really really hope it wasn\'t, but it might have been. So at this point I could now be dissapointed that Laura made Chris in a bad mood, and dissapointed in myself in case it was my fault. And then I was dissapointed cuz Chris neva sed bye to me (when he dusn\'t it always gets on my nerves). Then I was dissapointed because Lauren went offline... and then finally because I found out that none of Louise\'s family like Spiderman... so Chris probably wouldn\'t want to see Spiderman 2 with me :( I\'ve now asked Louise to see if she can come to Shell Island (there\'s a chance I can go, but I better not mention it to my parents today cuz the answer will dissapoint me more). I also asked if I can come over tomorrow. If I get answers to either of these requests tonight, I think I know what they\'ll be. Well I\'m not talking to Louise much... sorta naughty of me in fact. I guess I\'m not in a talkative mood. I\'m feeling sad. I\'m usure if it\'s hormones or just the dissapointment. Either way I\'m not a happy bunny. Oh yeh, Louise is making her own blog so I\'ll have to look at that soon. Hopefully that\'s something to look forward to.  Bye for now! ~Hannie xxx  '),(13724,3051898,'12,July,2004','2004-07-12 00:00:00','   I made a group blog for me and my friends to all post in, but for some reason the invitation emails weren\'t sending. Now I\'ve sorted it all out, but only Louise has joined so far. I\'m sure Laurenwill when she comes online, and I can probs convince Alison (who may not come online again tonight), but I have my doubts abuot Rachel. I want her to join but I don\'t think she\'ll want to. Anyway, if you want to see our group blog  urlLink click here . I don\'t really have anything particularly interesting to say today. I think I am improving on the Chris thing, but I have slipped up several times today but Lauren was here to point it out. It may be annoying, but she\'s helping me! It\'s for my own good! So thanks Lauren :D  Bye for now! ~Hannie xxx  '),(13725,3051898,'11,July,2004','2004-07-11 00:00:00','   I changed the layout... I did want a pink layout but Lauren changed hers to the one I was gonna have so I decided I wanted a different one (it\'s boring to have the same!). I chose this one because the archive thingie is on the left, on most templates it\'s on the right. Also it has pretty writing where it says Archives and Recent Posts. Actually I think I just have it cuz I reckon it looks sorta odd, so it\'s like me ha ha. Perhaps it will make my blog that little bit more interesting. Or not. Whatever.  Bye for now ~Hannie xxx  '),(13726,3051898,'10,July,2004','2004-07-10 00:00:00','   Why is it that so many people out there actually have interesting blogs? They also communicate their emotions better than the rambled garbage I come out with. It\'s actually quite depressing, lol. I was talking to Lauren about the possibility of me going to shell island with her, but now Lauren is all quiet and ignoring me (probably not at the the computer) and so I\'m bored. I\'ve just realised that I haven\'t put my quiz on here!! I have made a quiz to find out who you are most like out of me, Lauren, Louise, Rachel and Alison. Most people are Lauren, which is probably because Lauren is more normal than I am. Anyway, my quiz includes pictures of us! Yay! If you wanna take it  urlLink click here . Lauren\'s back now so I won\'t write more.  Bye for now! ~Hannie xxx  '),(13727,3051898,'10,July,2004','2004-07-10 00:00:00','   Today I woke up in a much better mood (I had managed to fall asleep not thinking haunting thoughts). I wasn\'t super happy or anything, but not miserable. Also I asked Louise about yesterday and she helped reassure me a bit, so now I\'m not worrying abuot him and Lauren. I really was just crazy and full of hormones. The worst thing abuot me today is that I seem to be a bit irritable. Today\'s been boring... I guess this post is just to clear up a few things. I\'m talking to Laura on msn now, I feel like I can\'t talk to her as much as I used to, but we are still talking sort of acting as close as we were. I hope we both get over this soon so I can go back to telling her every detail of my love life, lol! I was kidding, I need to stop remember. So no more mentioning Chris unless others bring him up. I am also allowed if I\'m talkng about something else and I have to mention him, but then I do NOT get into a lengthy discussion about him. I can do this. Maybe. I might post later...  Bye for now! ~Hannie xxx  '),(13728,3051898,'10,July,2004','2004-07-10 00:00:00','   I know what it is. The whole damn problem is those bloody hormones. Surely they\'re whats making me feel so depressed when I was happy not so long ago. I mean, a few minutes ago I was crying, wtf is wrong with me? I still feel upset... but I think I realise that the cause is most likely hormones. Chris and Lauren at party in the park were simply triggers (I know when Lauren reads this she\'ll point out she didn\'t do anything, which is true. I\'m just putting it this way because it\'s sooo much simpler for me to explain). So I\'m going to sit here feeling sory for myself, everyone can just ignore me because it\'s probably my own stupid paranoid hormonal fault I\'m like this. Bye again ~Hannie xxx  '),(13729,3051898,'10,July,2004','2004-07-10 00:00:00','   Well since last time Laura has got mad at me for telling Chris about her wanting me to dump him... she said he was mad at her, but I asked him and he said he didn\'t get mad. Laura reckons I told him because I\'m jealous of her, and she convinced Chris of that (I have now told him otherwise, I hope he believes my side of the story). Laura has calmed down since I think, but she still seems sorta mad. She gives me the impression that she does like him quite a bit. I\'ll admit I would be a LOT happier if they stopped being in contact with each other, but that won\'t happen unless they get bored of each other... which could take forever to happen (but is much more likely to happen if they don\'t see each other in person). By the way, my atemp to stop talking about Chris has failed. I now find myself hoping someone else will bring him up so I can talk about him, or else I just bring him up anyway. I think I have improved very slightly, as everyone else was talking about him. No noticable changes, though... I\'ll keep trying. Today it was brmb\'s party in the park. Lots of music artists were at cannon hill park basically. I went with Lauren, Louise, Chris and Louise\'s mom. It rained. A lot. Some of the music was good, though. I liked Mark Owen and Sugababes... Nelly Furtado and McFly were quite good, too. I\'m convinced that it was Ronan Keating who made it rain cuz he sucks (Mark Owen made it stop cuz he\'s great). Actually I also liked Sam and Mark... Now the part you know I want to talk about where I mention Chris. A lot of the time he was talking to Lauren,  which I did my best to ignore... he also hugged her (twice!) and not me (if he can get away with hugging her why can\'t he get away with hugging me??). He did carry me lots so that I could see the stage better cuz I\'m so short...that was sweet of him. Except he also picked up Lauren, and she could see anyway. At the time I was fine with this, but later after discussing it I now feel pretty shit. Also after reading Lauren\'s blog she mentions him quite a bit, why does she feel she needs to say so much about him? She didn\'t even mention the music! Doesn\'t she realise it only makes me think about it more and more and feel worse? I\'m not angry or anything like that... I just feel sad. Lauren thought I was fine with everything, but I was actually trying to make excuses to reassure myself and also I was hoping she would change subject... okay so I sortof do want to talk about it, but only if it\'s going to make me feel happy, which it doesn\'t. Actually, I was pretty happy with today until I thought about everything he did with Lauren. Why did he hug her? If he\'s not allowed to hug me in front of his mom why would he be allowed to hug her? Also why did he pick up Lauren? I mean before I thought about it and discussed it I was a bit unhappy, but I tried not to overthink it... because he had just been saying I was heavy and I said Lauren was heavier it was like he was testing this out to find out how much heavier and how long he could carry her for, which made sense at the time. Lauren told me he kept asking if she could see though and that she thinks this could mean he wanted to carry her. I hadn\'t thought of that. I assumed he was asking because I couldn\'t see and he was just sorta wondering if she could see or not. See in most situations I would assume I\'m right, but what she says makes sense and so concernes me. There\'s no reasonable explaination for the hugs. Lauren says they were just messing aroud hugs, but still the point remains that she got 2 hugs without really wanting them and I asked for a hug and I got none. I can understand why I got none, but I can\'t understand why she got 2. Ok I should stop now. It\'s not fair, I\'m going to think about this more and more and then while I try to get to sleep it\'ll be worse (it always is - i think too much in bed) and so I won\'t be able to get to sleep and make myself depressed and tomorrow I\'ll probably be moody. Guess I\'m not gonna stop afterall. In fact I could go on abuot this all day. Maybe it\'ll get it out of my system, though I doubt it. Thinking about it Chris wasn\'t that nice to me today actually. He made fun of the way I eat, the way I drink and my height. He talked to Lauren more than me, hugged her and not me, and picked her up for no apparent reason. Ok, really I\'ve probably said enough now... I\'ve probably pissed Lauren off by writing this stuff (I don\'t mean to...it\'s just the way I feel). I\'m gonna end by pointing out to Lauren that no, she does not act like she dislikes Chris AT ALL! So therefore yes, you could act like you dislike him more because you actually act like you like him. You act closer than to him than I do for goodness sake... I realise me saying this will make you angry... I dunno it just does annoy me sometimes when you claim you don\'t like him but when you\'re around him you NEVER act like it. Perhaps really you do like him but don\'t want to admit it for fear of me becoming paranoid (as I do get VERY paranoid as I\'m sure you can tell from today). If that\'s the case too late. Ok, I\'m not saying I think he fancies you or you fancy him... just it wouldn\'t surprise me.  I doubt you have a reasonable explaination to calm my nerves, so I\'m left panicking and assuming and worrying. It\'s horrible. It\'s times like this I really hate being me.  Lauren, I\'m sorry I posted this. I\'ve said things that will annoy you, no doubt. I suppose I created this blog for my feelings, I didn\'t create it for you to get offended or annoyed by, so please try not to be. I\'ve probably exaggerated it on here anyway. I\'m just gonna go and be miserable somewhere and hope that I find something else to think about while I\'m trying to sleep (I CAN\'T clear my mind - just doesn\'t work - being stuck in my head is like HELL at times... count yourself lucky you don\'t have my thoughts).  Bye for now ~Hannie xxx  P.S. I\'m also sorry for being overly repetitive and probably not making what my point is clear. In fact it would probably be best to ignore the post completely.  '),(13730,3051898,'06,July,2004','2004-07-06 00:00:00','   Ok so instead of loading on sims I thought it would be more fun to do some online tests. YAY. Heh, some of these results include things which are SO inaccurate. Yet others include things which are scarily accurate...    You are a child\'s kiss.  Completely sweet and innocent and pure. You mean no harm and only love in your sweet kisses.    urlLink   What kind of kiss are you?     brought to you by  urlLink Quizilla     E.T.!    urlLink   What movie Do you Belong in?(many different outcomes!)     brought to you by  urlLink Quizilla    You are totally in love.You try to hide it but you cant.You are nice and pretty in your own way     You should be dating a Gemini 21 May - 20 June This mate is inquisitive, entertaining and charming, liberal, broad-minded and youthful. Though Gemini has a tendency to be impatient, gossipy and sometimes irritable, this  twin has the ability to expresses his or her pent up emotions during sex!    urlLink   What Zodiac Sign Are You Attracted To?     brought to you by  urlLink Quizilla      A GARAGE-GURL. Youre into loud music, hot guys and wild fashions. Youre most at ease when you\'ve got all your mates around you and you like to party. Boys are a game and youre always on the ball because you make sure you\'re always number one.  Your virtues: Confidence, fun nature, sociability.  Your flaws: Loudness, jealous tendency, need for attention..    You almost certainly wouldn\'t like this game, because it\'s not your thing:  www.life-blood.vze.com    urlLink   What kind of girl are you?     brought to you by  urlLink Quizilla      You are the Secretive Mermaid. Perpetual beauty that longs for legs to walk by the side of men. You spend your time gazing at the stars and whispering to the moon. You have little to no freinds that breathe. Your freinds all missing. You are sweet as syryp and kind as cake. There are a handful of people and mermaids like you. Would you rate my quiz I will keep it a secret?    urlLink   What kind of mermaid are you? (Gorgeous Pics)     brought to you by  urlLink Quizilla      Congrats! Your a Pure Angel! Angels, as far as most of them go, are all compatabile creatures, but Pure ones simply are symbols of God. Pure Angels always appear when a child is born, when a rainbow is seen, or when someone shares their first kiss. They never grow old, an can appear in the shape of a naked woman with white, bold wings. Pure angels are the carriers of god, and show their love to everyone in the world.    urlLink   What Kind of ANGEL are you? (For Girls only) This Quiz has amazingly Beautiful Pictures!     brought to you by  urlLink Quizilla      Well your Happy, yes thats right your so happy, so pretty and witty and if i go on any more i will have to pay for the rights of this song so, great job your one of the lucky ones.    urlLink   What Emotion Dominates you?      brought to you by  urlLink Quizilla        My inner child is ten years old!    The adult world is pretty irrelevant to me. Whether I\'m off on my bicycle (or pony) exploring, lost in a good book, or giggling with my best friend, I live in a world apart, one full of adventure and wonder and other stuff adults don\'t understand.    urlLink   How Old is Your Inner Child?     brought to you by  urlLink Quizilla      you are the cute but psycho happy bunny. You adorable, but a little out there. It\'s alright, you might not have it all, but there are worse    urlLink   which happy bunny are you?     brought to you by  urlLink Quizilla    Hmm I\'ll probs do more but this\'ll take up plenty of space.  Bye for now! ~Hannie xxx  '),(13731,3051898,'04,July,2004','2004-07-04 00:00:00','   Today I just feel frustrated. I\'m not entirely sure why... I think it\'s something to do with Chris. Today I explained to him how he upses me... well sortof. There\'s more that I didn\'t want to add. I don\'t feel like I\'ve accomplished anything, though. Nothings changed. I think I\'m also frustrated with myself because he explained that he does like me even if he doesn\'t always act like it. I\'m annoyed that I seem to be moaning at him whenever I talk to him, which I don\'t want to do because it will annoy him. I also told him how Laura wants us to break up, and he said it was my own fault for discussing my love life! He missed the point completely... the point was that Laura isn\'t such a great friend to him as he seems to think and also that if she thinks we should break up he isn\'t treating me right.  Something else I\'ve decided, I\'m going to try to stop talking about Chris. I\'ll still write about him here... but when tlking to people I\'ll try to only talk about him when other people bring up his name.  Actually talking about Chris yesterday he told me about how he wasn\'t happy with life... I could give details but I promised I wouldn\'t tell anyone (except I was allowed to tell Laura) and Lauren might read this. I feel sorry for him, but I\'m still dissapointed that he never seems to consider my feelings... I care about him so much and it often seems like he doesn\'t care about me at all.  Anyway, moving on, today Laura told me that she\'s going to go out with her friend Matt! I wish them the best of luck in their relationship... I want Laura to be happy. I know she can annoy me at times, but Laura is pretty safe to me really so she deserves happiness.  Yesterday Alison told me she was really upset... it seems her sister has been cutting her wrists because she\'s depressed and can\'t decide who she likes best out of a college friend and long term boyfriend Will. She\'s been arguing with him a lot lately, and I guess I finally know why. I like Rachael (Alison\'s sister)... so I hope things become better for her. I hope Alison cheers up, too...  (I would like to say sorry to Alison because I said I wouldn\'t tell anyone and Lauren will probably read this, but I\'m sure Lauren won\'t tell anyone else and technically I haven\'t told her, Lauren read it...)  Lisa (can\'t remember if I told you, she\'s my sister) is ill today, poor girl. It\'s only a cold, or hayfever or something, so she\'ll be fine. I guess I\'m just showing that I do care about her even if I do think she can be a right bitch at times.  My mom\'s losing her hearing in one ear, it\'s funny to laugh at her and say it\'s old age, but my mom is worrying it\'s not old age and that it\'s some other serious problem, so I\'ve decided to stop laughing at her. It was mean of me to laugh anyway...  Well this is all I\'m gonna write (though it\'s still plenty :P). I want to load on the Sims (which Lauren has kindly let me borrow :D)  Bye for now! ~Hannie xxx  '),(13732,3051898,'04,July,2004','2004-07-04 00:00:00','   Do you ever feel like the world is picking on you? I\'m sure most people know the feeling of when everything seems to be going wrong and it seems like everyone just wants to make your life that little bit worse. Okay so things aren\'t THAT serious. But today felt pretty crap. I have a boyfriend who doesn\'t trust me and thinks I lie all the time (I\'m actually very against lying) and it feels like my friends are turning against me. I could explain everything about me and my boyfriend (as I mentioned in my last post, called Chris) but I\'ll probably ramble on about him enough in other posts. One thing that\'s really been bothering me lately is Stephen... I miss him. No he hasn\'t gone anywhere. He lives just down the road from me, yet I haven\'t seen him in quite a while. He\'s coming online less and less often and when he does he doesn\'t really talk to me. He spends more time with my next-door neighbour (Laura, we\'re both friends with her) and I just feel left out... It also annoys me that the few times Stephen does come here he wants to call for Laura. I know that the reason he spends so much time with Laura is because she plays football with him... and football doesn\'t interest me. Even so, I can\'t help but feel rejected that he calls her and makes plans and he never calls me anymore. I end up seeing them outside togther... and he never wants to call for me. It\'s upsetting, me and Ste used to be really gud mates, but I\'ve never felt so distant to him as I do now.  Similarly, Laura is talking to me a bit less (on msn). In fact she\'s talking to Chris more. She tells me to dump Chris, even though they are friends (only because of me...) and even though if I did she would still talk to him on msn. She wants me to dump him because he can be really mean to me considering he\'s supposed to be nice as I\'m his girlfriend! Depressingly enough he\'s probably nicer to Laura than me... Something that annoys me about Laura is the way she thinks she\'s right all the time. Actually that\'s similar to Chris. They both reckon they\'re always right. They don\'t even properly admit they\'re wrong when I have presented them with evidence that proves them wrong... it\'s so irritating.  Another person who keeps telling me to dump Chris is Lauren. Don\'t get me wrong, Laurens a lovely friend... I was annoyed yesterday when she signed in on msn as me though! She really shouldn\'t be doing that. I could\'ve got really mad but I didn\'t. I changed my password anyway so she won\'t be doing that again. Today Lauren got all upset and started crying... all starting from me observing Alison was wearing a short miniskirt so I said it was a bit slutty... Lauren defended her saying it wasn\'t (Lauren often wears short skirts and revealing tops herself - yes we say she dresses sluttily). Anyway, we ended up not talking for a while... I felt as though they were both picking on me moaning at me, as I hadn\'t meant to offend anyone, I did mean it jokily. It was playing on the sims that really brought things back to normal, I\'m glad me and Lauren are getting on okay. Though it was her who said sorry to me, and now I feel it\'s my turn to say sorry. I should\'ve said this at the time, but I didn\'t... I guess I didn\'t want to admit this in person. I\'ll say it now because I think Lauren will read this.  Lauren, I\'m really sorry for the way I say you dress like a slut. Of course you aren\'t a slut, I know you\'re not like that. When it comes to boys you\'re a good girl (usually :P)! So why do I say it then? As I explained at the time, sluts would also wear revealing clothes... but there\'s a bit more to it than that. I could say something like I\'m concerned guys will get the wrong idea and something bad will happen, but the chances of that are slim. So I\'ll tell you a more real reason... I\'m jealous. You have great legs, a nice stomach (no matter what you say you are NOT fat!), and a generally great body. The truth is that I know that if I wore the clothes you wear I wouldn\'t be able to look anywhere near as good! I don\'t even have such clothes to wear, so I can\'t even try to compete anyway. I\'m really sorry. When I do say you\'re slutty I really don\'t mean to upset you... If it really does upset you that much I\'ll stop... it\'s just when you\'re dressed like that I feel like I must look terrible next to you. Please forgive me?  Moving on... I\'ll talk about Alison, who was taking offence at my comment today... I was sure she knew I was joking. Either way she said sorry so I\'ll also say sorry. I don\'t know if Alison will read this because it\'s so long, but if she does then i want to say this to her... Sorry Alison I didn\'t mean to make you feel upset. I really did mean it in a jokey way.  And now Louise, she\'s been great lately. Sometimes she gets on my nerves when I\'m in a bad mood, but I know she doesn\'t mean too. Louise, if you read this then I want to say to you that I\'m sorry I get so mad over stupid things, please don\'t get upset when I\'m like this as I just need a bit of time to let off steam.  I now realise I\'ve mentioned a lot of people... and I guess I\'ve also realised that the world isn\'t picking on me. As usual it\'s just a few things upsetting me and I\'m blowing it all out of proportion and taking it out on everyone else. Well I need to finish this post soon, so I\'ll finish by mentioning Rachel. With everything thats been going on, I still feel most sorry for Rachel. She might not always be nice, but you can still have fuun with her. I know she spends a lot of time feeling upset as she thinks many things are her fault that aren\'t... I realise me and Rachel have had a good share of arguments, but I really do wish her all the happiness in the world. It\'s not fair that she should be so upset all the time and I wish I could help her. Whenever I talk to her I always feel so helpless because there\'s nothing I can do to cheer her up. I don\'t think Rachel will read this... but if she does hopefully she\'ll feel just a little bit happier to know I really do care about her.  That\'s all for tonight then. I didn\'t mention my family (who were being nice today until my sister, Lisa, discovered I accidentally cut my leg while shaving and I was using her razor...) but perhaps next time.  Bye for now! ~Hannie xxx  '),(13733,3051898,'04,July,2004','2004-07-04 00:00:00','   Well now I\'m annoyed. I\'d spent ages writing the perfect first post and I lost it all. I might try to type it out again... maybe. For now I\'ll just say hello my name is Hannah and this is my blog. I\'ll probably talk about Chris (my boyfriend), Louise, Lauren, Alison, Rachel, Laura and Stephen (all close friends, all amazing people) in future entries. I would write more but I\'m annoyed because I lost my entire entry (it was so long!).  Bye for now! ~Hannie xxx  '),(13734,3426788,'31,May,2004','2004-05-31 00:00:00','   I\'m having busy days. A bad experience with .net and I was wondering, how am I going to fix my website? Well for newcomers, I already have a complete flash website. Iâ€™ve uploaded it 2 days ago & marketed it. SMS, Mail, Instant Messages, Chat Room advertising everything Iâ€™ve done to tell the world that I was online. It was a good experience. My mood was pregnant. It is like buying a home in downtown New York or Sunset Boulevard. lol even Upstate New York is not a bad idea.    I was dying to see my guest book & No. of visitors. The first one to sign in was my dear bro. Heramb .. Well he is a nice guy. Always funny â€¦ yet somber at times but even that solemnity has a tinge of laughter in it. Whenever I watch Will Smith , I find Heramb so much reminiscent of him .. lol like I did , yesterday, when I was watching MIB II on HBO. He was online that night and provided me with good hints n tips regarding my website. Well he was kinda busy wid damsels â€¦. I know itâ€™s high time for him to get some one â€¦ (hope His parents r not reading this:P)   The one comment I was keenly waiting for was of Ashutosh. Now He is my dear bro. too but exactly opposite of me. A Very Shy Guy , as he loves to call himself , skinny , always wearing the naÃ¯ve specs of intelligence. As expected he sent me a long mail very next day with heapz of tips n tricks. It was a good mail to read. He diagnosed the websiteâ€™s present condition well & presented new ideas. I replied him immediately expressing my joy for an apt reply & thanking him for his program (which hotmail deleted cuz an .exe file as an attachment is considered as a potential threat.)   Before him, I met a good friend of mine. Anu, as I call her wid love, I met her few months down the line in chat rooms n sent her summer of 69 song in Mail :P â€¦ Since then I never chatted wid her in Private Messages. As the juggernaut of time rolled, we met again tht night. Iâ€™venâ€™t seen her in person, but I can sense her Suave nature & Sprightly presence around. I guess u donâ€™t have to meet people to know them, Its something inside that ticks. Or else peepz live together for number of yrs n feel astonished every now n then bout their partners. The best thing : SHE LOVES GAMING LIKE ME. I Dunno, How good she is in CS (Counter Strike) as she proclaims. Let me have a game wid her someday... n let me see who get the most FRAGS.   Back to the 31â€™st may â€¦. I was thinking the whole day bout nature of my website. I had a very crude, rookie design online. My main intention was to put up a website for game programmers, loaded wid tuts n games. Some sections for vb/vc programmers and discussion forums should be handy.   To be frank, I was happy when I designed my flash based website. Later I found it very spiteful. A failed attempt of a Greenhorn as u call or a Neophyte (Thanks to Barronâ€™s :P). Suddenly I found myself visualizing bout the website more and more.   I needed a sleek, super sleek design wid minimum of mess n loadz n loadz of useful links & tutorials.  I will definitely have some flash part in it. Huh thaâ€™s the best way to show ur knowledge bout the damn thing. As a hard ass perfectionist Iâ€™m, I believe certain things in life represent U. 1> Your resume 2> Your matrimonial Ad and 3> Your website. If you have spelling mistakes, disorganized content, perplexed design then it shows negligence on ur part & not the finesse you want to depict.   '),(13735,3426788,'23,June,2004','2004-06-23 00:00:00','   Arete, Sophrosyne, Sumpsimus, Serpiginous, Gaminerie and Autochthonous  Can anyone guess meaning of these words: O â€¦..? Lol... forget about meaning. How about spelling them correctly in Single chance â€¦ in front of a big audience in Grand Hyatt Washington. Hats off to you. If youâ€™ve guessed it. No â€¦ I ainâ€™t talking about GRE and TOEFL and BARRONS or CAPLAN Tests. This is National Spelling Bee 2004. The winner was David Tidmarsh, a 14 yr old Indiana boy. David, an eighth-grader at the public Edison Intermediate Center in South Bend, won the top-prize package of $12,000 and an engraved cup from the bee, plus an additional $5,000, encyclopedias, a $1,000 savings bond and a reference library from other sponsors  Whoa â€¦ Nothing get bigger than this. I was flipping channels ,trying to find out some value based entertainment. Discovery was showing the re runs. National Geographic was a passÃ©. Suddenly I saw this competition on ESPN 2. I couldnâ€™t resist myself. I put aside all the work I was doing (lol How bout David J. Kruglinksiâ€™s Inside VC++). My eyes were glued to idiot box (Very often I find myself in this situation if Baywatch is not playing or Swimsuit illustrated Pageants are not showing).   I was loosing it totally â€¦ Couldnâ€™t spell better than them. At the end when David hide his face behind his placard then I found myself standing and clapping with the same enthusiasm as Davidâ€™s mother.   These contests are amazing. Intel science fair or National Spelling Bee or Any other talent examsâ€¦. It shapes up your mind, character. It strengthens you to face any challenge, overcome any adversity. Talk about prizes, name, and fame â€¦ Itâ€™s a serious Addiction â€¦  Let me tell you Addiction of fame is the dangerous amongst all types of addiction. You start to feel more n more hungry or thirsty for that share of limelight â€¦. The worst situation to be in than any Man Eater Leopard/Tiger.  I know it better than many of my peers. Once you reach to such plateau it is very hard to come down or thinking of descending. Once youâ€™ve been there for years and years then it is very hard to fathom the defeats, it is callous to accept the failures.  But every trough has a crest to follow and vice versa. Thatâ€™s why life is interesting and it is not monotonous.   I wouldnâ€™t ever trade my multidimensional (add dimension of time also) life to any other dull, monotonous, droning lifestyle. I think to feel the zephyrs of success whirling around you & then face the calmness, coldness of failures and carrying the torch of hope to find that warmth shanty of success is the essence of life. Mansions of successful lives are built upon the foundation of failures and sagacity and erudition derived from them.   '),(13736,3426788,'06,June,2004','2004-06-06 00:00:00','   For last 4 days I was frantically waiting for something to happen. Related to my life or at least related to my website. It was true â€“ my first dig at personal website creation was terrible than a neophyteâ€™s practice work. A Bad design in Flash was the culprit.  I wanted to put everything in my website. Programming articles, Game designing tutorials, Several Links & Hundreds of utilities â€¦ My present website wasnâ€™t even accommodating my info. On 5â€™th night, I was searching for good layouts for my website n suddenly Voila! I found one. It was a sleek, sophisticated online curriculum vitae layout by Ubooty (A company based in Norfolk,UK). By the way it was free.  I downloaded it enthusiastically and created my home page with minimal efforts. This time by hopes were centered on Flashâ€™s elder brother Macromedia Dreamweaver. What a Name !!! Dream Weaver. Hats off to them.  As I was going strong & creating my web pages at a speed of nauts, I confronted a IceBerg straight ahead. The web page layout by UBooty Had a complete table in a web page & all components arranged in different table cells & rows. So if u want to delete extra components & resize the table -> IT WAS IMPOSSIBLE â€¦ It resulted in Long web pages even when the content was merely 2 lines long â€¦..  I tried relentlessly, only to find wall straight ahead n a head bang â€¦.  I was getting frustrated â€¦ Never ever I was so much annoyed with myselfâ€¦ Damn it... I thought â€¦ It was a simple web page â€¦ and I used to deal with Goliaths of C++, Visual C++ and .Net. I shut down my computer & relaxed on my bed. Deciding to forget everything bout that website.  I couldnâ€™t concentrate anywhere else. Not even my favourite Yahoo Chat â€¦ Where I love to play All Star Counselor. Self Proclaimed Shrink. Iâ€™m chatting for last 6 yrs now & Iâ€™ve seen all kinda peepz here. Straight forward, Backstabbers, weak hearted, Wicked, Timid, Brash, Liars, Candid but there is a line of similarity which passes through all of them. They all need A Life Saving Support System.  Chatting is an addiction like any other addiction. When you get weak, When u loose control over ur life, When situation overpowers u, When you pass the reigns in the hands of Destiny n decide to enjoy bumpy ride in the back seat â€¦â€¦. You need someone, something, some way somewhere to rekindle past or Just to tickle yourself with the past without accepting the present tense hence trying to be someone who could have looked like/behaved like/been like this one in the past galore of glories.  Chatting gives you the same satisfaction. Freedom of being someone you are not, you were, you want to be, you always wanted to be or someone you never can be. It answers an impeding question, why do peepz visit Dance Bars or such filthy places. There for a piece of currency, they can enjoy the pre-eminence.  Power, Control, Freedom â€¦ Addictions give you everything for those split seconds â€¦.  Maybe Life is like a Carrot hanging in front of a donkey... Like u n me ... So we run in the Great Optimistic mood in the Hope of catching the carrot which furthers with every step.  My website creation lead me to the point where I felt I couldnâ€™t do anything to improve my situation. Suddenly, a thought kissed my cerebrum. Iâ€™ve done so many things before & toyed with all these tools. Why canâ€™t I do this again?  I studied the whole layout again. Understanding the base concept... I created my own web page using layoutâ€™s graphics. WOW â€¦ It was piece of cake then :P  It was already morning of 6â€™th June.  Yesterdayâ€™s French Open Ladies Final was a Drag. Two inexperienced players playing in historic grand finale for the first time n one awed  by the situation, second one in the same seat but holding both the handles. Thank God , It ended. Anyways, Congrats to the Anastasia Myskina. Well Now her name is taken from Russian domain. She looks like a Mink though.   I started developing my website rapidly. My information, scorpio graphics .. all were in place. I wanted a big page for all links section which I placed categorically. Linkin those hundreds of websites was a neck breaking task, still I managed it. Due to sheer anticipation of website completion & a Gentleman named Mr. Ralf Luve.  If you know Mr. Ralf Luve, congrats. Let us draw some inspiration from this german. He works as a Systems Administrator in an Internet Service company in Germany. He is 32 yrs old. He is married and HE HOLDS 5 GUINESS RECORDS !!! Huh .. Lemme tell u bout his latest one â€¦ Have you ever heard the term â€œ Rubic Cubeâ€? Imagine solving it in 5 minutes â€¦ What? Simple? Piece of Cake? Rookieâ€™s Delight? â€¦ Well How bout solving it in 5 minutes â€¦ BLINDFOLDED â€¦HA HA HA â€¦. See thaâ€™s the gr8 thing bout him. He did it. Gosh â€¦ I was spellbound, enthralled,  awestruck, agape â€¦.. everything at the same time â€¦ Mebbe I wonâ€™t act such even wid Ms. Pamela Anderson Live In Person n Naked â€¦.  Talk bout Mind over Matter & this guy was amazing. It proved Mr. Albert Einstien right, who used to use 40% of his brain. We are better than any super computer. Only we should know ourselves.  Remember the Dialogue from THE MATRIX ->    Donâ€™t think you areâ€¦ Know you are â€¦ !!!     It pushed me & finally I was able to finish my website. Still one last glitch was waiting. Counter on my main page. As I moved my site from yahooâ€™s geocities (provides 25 mb) to 125mb(provides exactly same 125 mb), my counter was gone. It was important to me in a sense of contentment. I searched again n again â€¦ found several free counters â€¦ registered with them .. result -> A counter with an ad. Again I was looking towards the window of melancholy. But as I was already stirred by Mr. Ralfâ€™s otherworldly record, I decided to write my own counter. Which I precisely did with the help of My favourite Microsoftâ€™s website. Thaâ€™s why Iâ€™m an Ardent Fan of Microsoft.   Long Live Microsoft.   '),(13737,3426788,'02,June,2004','2004-06-02 00:00:00','   Despair, Depression, Desolation, Dereliction, Delinquency â€¦â€¦ I was suffering from all of them at the same time, in different forms.   I looked out & saw the glistening Full Moon above. It was clear in my mind. I ought to feel like that way. It was on of those days of the month again â€¦   Full Moon day & No Moon day .. are so close to my life. If I found myself solemn than required & mulling over everything again n again â€¦. Itâ€™s gotta be those days in month. Iâ€™m not lunatic. Iâ€™m not ruled by any extraterrestrial creatures. Though Iâ€™ve faith in the God & supernatural powers, Iâ€™m skeptical enough not to take things on their face value.   After all Iâ€™m in technology field where everything happens due to something or everything has a valid reason. Everything is causal in this world. Well not everythingâ€¦. Iâ€™m a bit psychic. All Scorpions are â€¦. By the way no marks for guessing .. Iâ€™m a scorpion â€¦ On the border of Libra and Scorpion but I identify myself more with a true good Scorpio with itâ€™s The Eagle variety (Read Sun Signs: Linda Goodman).   But I know the exact point when this process of thinking n getting poignant was triggered. It was the conversation with my lovely friend, Anu, over 3 hours and 39 minutes. She is the exemplary whacko, punk, care free, unabashed teenybopper. The best part is she knows how she is n why she is. Well we all were in our teenage. To know that precisely is the best part bout her.   I still remember the time when I was in 10â€™Th or in 11â€™Th or in 12â€™Th. The world was outright black or white to me. Nowadays the world certainly is not a rainbow but smeared, blurred, smudged combination of everything.   Morales are something yesteryearâ€™s people used to care for (n how foolish they were)â€¦ Value is something only currency notes have. Virtue is the word somewhere in dictionary.  Why different games like Soccer, Rugby, Boxing, Tennis, and Cricket are interesting? Because they all are played with certain RULES. Or else , a night club brawl, a street fight, a riot can be considered as various forms of those games. Same way â€¦  Life is interestingly worth when it is bound by some rules. Nowadays the only rule is to break the rule or bend it.I ainâ€™t acting like a moral inspector or an eighty yr. old guy who has seen enough n enough scenes.  Suddenly everything you are not supposed to do are means of enjoying life. Everywhere I go , I get the same answer â€¦ I want to enjoy my lifeâ€¦ I want to experience everything in my life once â€¦. Huh... Hail... Lame excuses for smoking, boozing, doping â€¦ lol do I say something bout SEX??? (Scorpions are not supposed to say anything against SEX)  Forget bout Conversation with Anu, she has to experience so many things. Consider my good old friend Sanjay, who is at least 10 yrs elder to me. I hear the same lines in the same tone every time â€¦ then may be it is in bar or a pub or at any other damn place. I keep on smiling to them n for myself â€¦ Cuz nowadays Iâ€™m being seen as someone who is the biggest fool alive â€¦ goodness gracious reason : I donâ€™t smoke. I donâ€™t drink. I donâ€™t dope and Iâ€™m always the only one to drive all of my friends home when they are intoxicated above all permissible levels.  Life is getting fast & instant â€¦ So Iâ€™ve decided to write blog once in 2-3 days & decided to stop for today.   '),(13738,4008178,'17,July,2004','2004-07-17 00:00:00','   Well its me again, doing random crap again. Uhm yeah. So this will be my basic blog. I\'ll type whatever i wanna, so yeah, this is where i dump my sadness, my political crap and my rants. maybe ill tell you what i did during the day. MAYBE. But I have to sleep now. Im going to a concert tomorrow. So I\'ll see ya.    Bye!    &nbsp;  No! Wait! theres something else. Its a small paragraph rant about  urlLink www.gaiaonline.com     &nbsp;  I really want to crush every skull on gaia. I mean, every single guy is a nerdy motherfucker who tries to act tough. GODDAMMIT I WANNA KILL THEM ALL!    &nbsp;  No, i didnt get flamed, but their stupidity makes me angry. thats it. probably ranting about it some other day. anyhow, goodbye.  '),(13739,4008178,'19,July,2004','2004-07-19 00:00:00','   w00t w00t. so uhm Ill just make this blog in english cause i think most people will read it in english. or... uhm... yeah.     &nbsp;   Concert crap:     &nbsp;     Chivo - Chivo - Chivo - Chivo -  Fecha: Domingo 18 de Julio. Hora: 5:00 p.m.Lugar: Gospel Music, Outlet Mall, San Pedro.Valor: Â¢1.000.Grupos: Osho, No Resolution?, Up Shock, Neginot, DÃ­as de AgonÃ­a.     &nbsp;    &nbsp;   Ok, so this was my first time at going to Gospel. I shouldve known this before, but people there told me the concerts at Gospel generally start at about 6:30. i got there at 4:50. The concert started at 7 pm. *gonk* well, that was a lowdown on the cafÃ©, since most people got there early. Not at 7.     &nbsp;   I only got to see Neginot, Up Shock, Dias de Agonia and No resolution?.     &nbsp;   The concert started out with neginot playing. I basically went to the gig to check out neginot, since i know the bassist dude and wanted to listen to them. So basically you could say the play death metal? You know, a hardcore type of scream, and some changing riffs... everyone in the band can play really good. The singer was constantly jumping up and down the bandstand, blah. Sorta got everyone in the mood, and the sound guy was using them to try out noise. everyone was just getting seated when they played.     &nbsp;   next was... no resolution. I cant really remember these guys, i liked them but not that much. their music could get better anytime. sorta crappy riffs and crap that made me hungry. yeah, hungry. anyways, ill give this band 6/10, just because i cant really remember them.     &nbsp;   Then came dias de agonÃ­a. i had never heard this band. First of all, every guy in the band was straight edge, from what i could see. These guys play hardcore. Just hardcore. Didnt really get the lyrics because of all the screaming. The singer dude looks gay. i mean, maybe he was trying to look tough, but he really looks gay. like hes coming at you to rape you. the drummer got his drum loose and they had to stop playing. everyone clapped except me, who... was maybe too tired. the singer really needs to know how to sing. i mean, yeah, screaming sounds cool but if all you hear is GIAR GIARR YEARHG! for like 20 minutes it gets boring. overall good band. but not worth buying their cd.     &nbsp;   then came upshock. ok these guys i had already heard, but i dont like them that much anymore. its general punk that you would get by listening to pennywise. the singer was constantly wooing out the lyrics and then screaming a bit. Really retarded. and he\'s like 40, i mean, what the fuck. ok this band sucks ass, dont even pay for one of their concerts. get a guitar and make quick noises and there you have upshock.     &nbsp;   so basically lets talk about the place. Its your general internet cafe, a 20 by 20 room with red dim lights. The sound guy sucks. He plays the music so loud i got bored and left. it starts hurting after a while. specially when they play hardcore. The room is so packet up with chairs theres no room for skanking/moshing. besides, all the people gather around the center like if they cant get to see the band from 10 meters away.     &nbsp;   There was this slipknot fan dude who kept making coreography to the music, looked kinda like a dj with no turntable. the only difference between the dj and this dude would be that uh.. yeah.  BEWARE  you might find this guy at a concert. he wears bike glasses so that he can aerodinamically soar to the bar, wears a slipknot hoodie and finger cut gloves. I dont know why everyone knows him, but to say the truth i really couldnt care less.    The pit consisted of 4 people. The slipknot dude and three other kids. well there was a big guy but whatever. The reason of why you cant make a pit in a 1x1 four people space is that you cant move around. these guys were waiting for the other guy to finish so that they could punch the air like a moron. Then there was the human windmill. who moves his arms so fast you can actually see them.     &nbsp;   basically the concert sucked. i did enjoy listening to neginot though. I told the bar dude to get some ear plugs or he\'d go deaf.     &nbsp;   There was also this little 5 year old girl in there. apparently i was the only one who cared, because all those other motherfuckers just laughed at her. i mean, she was with her mom, who worked at the cafÃ©, but what would happen if you see a hardcore show everyday since you were 5? i mean come on, how can a little girl be in a place like this? yeah anyways. i saw like 6 ppl from school. blah.     &nbsp;   like, everyone at the concert was wearing like 50,000 colones in clothes. what the hell, arent they supposed to be the working class? anyways ill rant about this tomorrow or some other day when i have the time. and i also have to talk about the art exhibition i went to today. it was.... lame. yeah. but now its 11 pm and i needa sleep some. bye bye'),(13740,4008178,'20,July,2004','2004-07-20 00:00:00','   i cant believe where everything is going. this new media, this new thought. yes, be free, be unique, but guess what? everyone is thinking the exact same bullshit that you are, and everybody is a fucking freak. not saying it is bad, but they rub it on your face and it feels like shit. and if you had been in my exact same position you wouldnt be laughing. you would be crying or drugged up somewhere in the country. i made a difference and this is what i got to see: how everyone has gone down to a level where it isnt even WORTH trying to change a thing. anyways. i think i should talk about something nicer, but there isnt anything nice to talk about. oooh how great, youre a fashion victim! and you, you spend your money like there is no tomorrow. and how many people actually care about suffering? no time for things we need while people starve and people bleed. thats what the song says, and none of you will ever move your picture because your explanation of the world is a label with some clothes on it.    and drugs. fuck i hate drugs. whats so cool about doing them? oooh look i drink this liquid that looks like dehydrated person pee and tastes like crap just because everyone accepts it.&nbsp; whats so cool about that? \"oh i heard that smoking makes you forguet your problems\" it doesnt. it only delays it a bit and then fucks up your health. and guess what? youre a fucking sissy piece of shit if you cant even control your own emotions. ooooh im so tense i need a cigar. you lame piece of degenerative crap.    second, everyone just acts like a total asshole. im getting bored of you all trying to act to some stupid rule of society, and deep inside i know that none of you are like that. stupid two faces, bunch of motherfuckers. heres a little secret of my own: I hate you all. im just waiting for the day when you will die so that i can live my life happpily ever after without having to worry about your stupid prejudices. sick of your dirty looks, of your inching away, ASSHOLES.    so yeah, the world is going downhill and im staying on top of it to watch it fall. maybe its just for the kicks, i dont know, but i surely dont wanna be part of this thing you call \"being cool\". jesus.   &nbsp;  '),(13741,4008178,'20,July,2004','2004-07-20 00:00:00','   Okay. i wanted to talk about the taxes? Ive noticed that the government is always acting like \"oh jebus we dont have money whadda we do? i mean, i really wanna buy that car and the rest of costa rica! MONEY!\"'),(13742,4008178,'20,July,2004','2004-07-20 00:00:00','   So lets talk about the art expo. In a word, you can say it was... lame. It was a glass-fusion, wood carving, sculpture, painting expo. O.o yeah. It was three artists, dont remember their names but who cares, i mean, they sucked! ok lets rate the painings:  Annoying. The guy just grabbed a brush, dipped it in his six color staedler kids paint, splotched a bit on the paper and \"there you have it, its a volcano!\" bleh. and then he folded the paper to make a \"visual effect\" on the art. on the volcano. i mean, come on,  the paintings sucked .  &nbsp;  ok now the wooden things. its basically a piece of wood that has een shined out. then you make three little holes: the boobs, and the stomach. then in the holes put a stone in them and it looks like a pregnant person. you know what the worst part is? that  every single wooden carving had the exact same shape !&nbsp; Then there was this piece of wood that had like writen \"ove, love, lo ve, lo ve?\" written on it. then there was a dick right below it. this was probably the worst thing in the WHOLE exhibition. i mean, i think the guy who wrote this crap really wins the prize.  &nbsp;  urlLink http://www.funfreepages.com/prize/yourock.jpg  &nbsp; So what the heck, i realy wrote that on the guest book.  &nbsp;  uhm now there comes the glass things. ok its pieces of pearlescent crushed between two glassplateswith different colors. it was generally a triforce with different colors inside a circle. if thats art, well good riddance.  &nbsp;  some of the sculptures were cool, but they  were all the typical \"girl coming out of a rock\" thing . there were about 5 of these. if i was gonna make a scuplture, id make batman climbing out of a stone with a hotdog in his hands and name it \"scotch tape\" and it will have a better name than evryone else\'s art. and i will become leader of the world and paralize you with my lazer beams.  &nbsp;  but thats just me.  '),(13743,4008178,'21,July,2004','2004-07-21 00:00:00','   ok im \"esteban says: IMBECIAL HANLFABETO\" and im talkingto esteban.   im also \"jaj... a las cholas de canal once... ojala las atropellen. viejas polas.\"      jaj... a las cholas de canal once... ojala las atropellen. viejas polas. says:  ya bengo   esteban says:   *vengo  jaj... a las cholas de canal once... ojala las atropellen. viejas polas. says:  si callese   esteban says:   IMBECIAL HANLFABETO     esteban says:   XD     esteban says:   entendio   esteban says:   esta malo     esteban says:   pero vale    (...)     &nbsp;    jaj... a las cholas de canal once... ojala las atropellen. viejas polas. says:   ai amback mal    jaj... a las cholas de canal once... ojala las atropellen. viejas polas. says:    [quote] esteban says:  IMBECIAL HANLFABETO     [/quote]   esteban says:   q andaba haciendo     jaj... a las cholas de canal once... ojala las atropellen. viejas polas. says:&nbsp;  &nbsp;JAJAJAJAJJA     jaj... a las cholas de canal once... ojala las atropellen. viejas polas. says:   mae nada     jaj... a las cholas de canal once... ojala las atropellen. viejas polas. says:   no jamando     jaj... a las cholas de canal once... ojala las atropellen. viejas polas. says:   jajajaj   esteban says:   XD    jaj... a las cholas de canal once... ojala las atropellen. viejas polas. says:      [edit] there was a crappy \"XD\" smiley here, which i cant post right now. the eyes were white [/edit]     esteban says:   mae q mierdas es esa cara     esteban says:   parece q esta llorando     &nbsp;   [nick change!]       esteban says: IMBECIAL HANLFABETO says:   esque se le borrÃ³ el negro     esteban says: IMBECIAL HANLFABETO says:   callese no pregunte   esteban says:   XD  esteban says: IMBECIAL HANLFABETO says:   xD        &nbsp;   &nbsp;   &nbsp;   &nbsp;   &nbsp;  this is by far the best convo ever.'),(13744,4008178,'21,July,2004','2004-07-21 00:00:00','   and i wonder, how many so called \"punk rockers\" do something?  because singing about it is certainly not gonna do anything. id like to see a singer climb down the bandstand and start a march. but they wont do that. all our heroes are dead or corrupt. - international noise  '),(13745,4008178,'22,July,2004','2004-07-22 00:00:00','   These last months&nbsp;I had being watching the news about the the situation between Mexico and the&nbsp;United States. Like most of you may know the U.S. don\'t wants&nbsp;to allow&nbsp;people to pass through the border between these to countries.&nbsp;I really comprehend this action made by de U.S. since, among the average foreign Mexican in the U.S., there are&nbsp;drug&nbsp;dealers and people you wouldn\'t like&nbsp;in you country if you are a straight president&nbsp;with a high and strong moral. The problem I have with them is that they generalize making everyone look like a drug dealer. They complain about us being the ones that sell the cocaine, marijuana, and so on, but I wander who are the stupid that buy it? Where ever you go in California or Texas or any of those states that are close to Mexico, they see you as just another Mexican that came here because, either we cross the border to see&nbsp;what the land of the free is all about&nbsp;or ether because they are so merciful that saw&nbsp;you and saw the misery&nbsp;you were living in&nbsp;your tree house with a roof made out of palm trees\' leaves. Mexicans&nbsp;went&nbsp;there because the U.S. crashed Mexican economy with their NAFTA&nbsp;starve-to-death-agreement in the early 90\'s. Whether you are Brazilian, Venezuelan, Ecuadorian, or Nicaraguan they see you just as another pathetic-ignorant-country-music-lover-Mexican with no&nbsp;right to look at their pair of omnipotent eyes. These discrimination and indifference to latinos irritates us creating a situation just like the one you have with the black people. I wander why is it. Your politicians, they are making a fascist country out of yours and that\'s because&nbsp;you discriminate,&nbsp;you are afraid of everything that is different from you, black, Latino, Chinese, Indian, all those so called \"minorities\" are just what you are afraid of because they aren\'t minorities and you know that.'),(13746,4008178,'22,July,2004','2004-07-22 00:00:00','   Â¿QuÃ© es la paz? La verdad es que no sÃ©. En un mundo donde los polÃ­ticos no son nada mÃ¡s que fascistas enmascarados con hambre de poder y guerra, uno no puede distinguir entre blanco y negro. Â¿QuÃ© clase de personas elegimos para que nos representen me pregunto? Tenemos lideres obsoletos que sirven para nada mas que destrucciÃ³n la tan llamada â€œsociedad civilizadaâ€, que no es mÃ¡s que un circo de payasos y malabaristas que de intelectuales y humanistas. El mundo en el que vivimos no es nada mÃ¡s que la decadencia del ser humano. Hoy no se puede sembrar y cosechar frutos para maÃ±ana. La&nbsp; sociedad de hoy es una banda de cuervos que en el momento mÃ¡s inoportuno arrancaran el ojo del dÃ©bil y el necesitado. Muy pocos se preocupan de los demÃ¡s. Esos tienen que ser nuestros futuros lideres. Seamos solidarios, compartamos ideas, seamos consistentes con las ideas y miremos hacia un futuro unido, porque en el momento mÃ¡s insospechado, va a venir la maquina a destruir todo para fortalecerse y en ese momento no habrÃ¡ mucho que hacer. Todo el mundo estarÃ¡ entontecido con las ropas de marca, autos de lujo, y en sÃ­ serÃ¡ una sociedad consumista.  '),(13747,4008178,'22,July,2004','2004-07-22 00:00:00','   gueeeeno. hace unos dias estaba en el parque de daiversions aca en costa rica. lo linda que es mi costa rica y su bonito monopolio de parque de diversiones. :)  aunque por un lado es bueno. osea, quien querria hacer un parque de diversiones en costa rica? O.o pero bueno. estaba en uno de esos chunches de agua.. como se llama? el que no es el reventazon, el... ese. mis contactos aqui me dicen que es el pacuare. mushas gracias. bueno. llegando al final de las escaleras en el pacuare, ustedes pueden observar que detras suyo, volviendo a&nbsp; ver para las afueras de nuestro afamada construccion, se ve un caballito. si muy bonita la yegua esa, pero lo que uno no se da cuenta es QUE EL COCHINO CABALLO ESTA DEBAJO DE UNA ANTENA ELECTRICA INDUSTRIAL.   y no tengo nada contra la electricidad. no soy como el mae \"rebeldon que no prende el microfono porque esta tan enojado con el sistema que no usa su corriente electrica\". no no, mis segnores, no estoy aqui para discutir esas... \"cosas\". solo les queria decir: como se va industrializando/ poblando/ \"desarrollando\".  osea, di si, no me voy a poner en varas de que es malo para el ambiente, porque aca ya todos lo sabemos, y si, entre mas edificios y menos arboles tenga un pais, mas \"desarrollado\" lo es. pero que raro que la gente necesite esto. di, no les propongo la solucion porque soy muy estupido e ignorante como para averiguarlo (y la verdad me da mucha pereza, no joda, toy en vacaciones) pero les meto el problema aunque sea por la nariz para ver si algunos de ustedes (ya que son todos tan \"inteligentes\" y estudiosos) saben que hacer. yo no. pido una patada en el culo.   y como my compita aqui abel pachuco dice: muchas gracias.  '),(13748,4008178,'23,July,2004','2004-07-23 00:00:00','   Meh. Been feeling a bit depressed and empty lately. My parents are starting to lecture me everyday. \'Tis a real bitch. They\'re always telling me I\'m too hateful and angry, and they\'re so damn nosy. Searching my room, always looking at me suspiciously. I know it\'s only because they care about me, but I can\'t help but feel annoyed and slightly pissed off.  Anyway....I got a phone call from someone I didn\'t even know today. It was....weird. She knew who I was, but I had no idea who the fuck she was. =\\&nbsp; The conversation went something like this... Me: *picks up* Hello? Her: Is&nbsp; Angela there? Me: This is her. Her: Oh hi! This is blah blah blah. (Note: she didn\'t really say \"blah blah blah\", but I was distracted and didn\'t hear what her name was) Me: Ummmmm....hi Her: Do you know me? Me: ....I don\'t think so.... Her: I\'m one of Olympia\'s friends. Me: Oh, cool Her: Have you seen me? Me: ......I don\'t think so.... Her: Oh. Are you going to Timpview? Me: Yeah... Her: Cool. I might go. Do you know anyone that\'s going there? Me: Errrr...just some of the people from our school. Her: Oooh, do you know blah blah blah? (Can\'t remember who she said) Me: No.... Her: What about Cecelia? Do you know her?  Me: No.... Her: Do you know blah blah blah? (you get the point) Me: no.... Her: dang, you don\'t know any of the people Olympia hangs out with! Me: Heh. Her: Do you know Cruz? Me: Yeah.... Her: Cool. You know Javier right? Me: Yeah..... Her: *launches into long story about how Javier is moving to Springville* Me: Fun. I know people who live there. Her: Cool. Are you Mexican? Me: No.... Her: I always thought you were Me: Nope, sorry. Her: Are you Mormon? Me: No.... Her: Are you Catholic? Me: No.... Her: So you\'re nothing, eh? Me: I suppose you could say that... Her: Well, I have to go! Me: Bye. o.o  God. That was an awkward conversation. Considering that I had no idea who she was, and all. I hope she didn\'t think I was rude. ^^;;;;  Ugh, anyway, I\'m off to post in my LJ, then study for my English exam (WHY ME?! *screams at the sky). Night all.  '),(13749,4008178,'23,July,2004','2004-07-23 00:00:00','   today i was buying swimtrunks cuz were going away to the beach tomorrow. so i was in the dressing room,trying to put on the trunks (i also realized my pants weight a lot? O.o) and there was this really crappy/commercial reggae on the radio... and the music ended and then the dj started accepting phone calls.   most&nbsp; of them were rants about the government, so here\'s one i snatched into my memory:  why does the government spend all that money on tourism and not on rebuilding the goddamned roads that we have?  i mean comeon, the roads have like huge holes in them and what does the MOPT do about it? they start drawing hearts on the street?! HWY DO YOU HAVE THAT STUPID GUY IN CHARGE?!   and what would happen if they actually got to fix the streets? more tourism! because the gringoes get bored of crappy roads, dont you get it?! how many gringos have ever set foot on say, [insert name of remote place in costa rica here]? why? because theyre not gonna spend four hours in a grav-- i mean, swamp type of road, just to get to a goddamn beach/thing/whatever.  so i say, get better roads, get better tourism. invert on tourism, and you only get a bunch of gringos infesting the local mcdonald\'s up.  *licks the reader*  '),(13750,4008178,'24,July,2004','2004-07-24 00:00:00','   If youve read the last two blog entries from esteban and messeguer you may already know that i (gurdian)&nbsp;am being hunted down.&nbsp; People in 07 are too proud.&nbsp; They think they are so united but still they exclude and pick on people.&nbsp; I have started my own little revolution against it and&nbsp;the AGUANTE&nbsp;constantly tries to crush it.&nbsp; It all started when i got drunk at a party and insulted the kiddies there and pablo and manu tried to beat me up.  '),(13751,4008178,'24,July,2004','2004-07-24 00:00:00','   First of all I will have to explain what does&nbsp;\"aguante\" means&nbsp;and what does the \"aguante la&nbsp;\'07\" means as well. \"Aguante\"&nbsp;it\'s&nbsp;like to confront something with brave, so you can guess what does \"aguante la&nbsp;\'07\" means (for the ones that couldn\'t guess it, it is like \"you can do it generation \'07\"). Generation \'07 claims to be a united and powerful generation in which everyone is an&nbsp;important part of the generation \'07 society, kind of like communism, including the dictatorship part of it since if you are not part of it they want to kick your ass up to your face as you could see with the rant Felipe did about Friday\'s party in Pablito\'s house (if you didn\'t then do it). Now the truth is that this unity is as true as George W. Bush being a \"peace president\". On our generation (or class) many people is bullied everyday (in school days), like Gay J (he deserves it since he is annoying) Patricia, Mow Mow, Tatiana, Oscar... And the list goes on and on. I personally have being bullied for about&nbsp;two and a half years, which means I have being suspended and have had to fight some of the persons who bullied me. The curious thing about it is that some of the people who bullied me now are trying to be nice and sometimes it pisses me off, because I think I have more friends in other generations than in this generation \'07, and when someone who bullied me&nbsp;found&nbsp;me talking to someone from other generation they start being hypocrite and start acting like they have always being my friends. Now you can like have an idea of how is our generation. These doesn\'t means there are good persons in the class, but unfortunately there is more bad than good in my opinion, like for every 5 persons 2 of them are cool. This ratio is not objective since it is my opinion. Now back to the main theme. If this generation is united then why the bulling? That\'s because it\'s not true. If someone deforms your face no one will care about it in fact they will just stare a you wail they do it. Our proud generation also comes with its own home page:  urlLink  www.tuleparty2nite.4t.com  . At the beginning it was owned by Tufo, Fabro, and Chris (friend of us). Then Fabro left and came in Carlos, there it start the reing of stupid polls. They first made a poll which said: Who is the most freak in our generation: in which featured Felipe, Gurdian, Tatiana, Daniela Montealegre, some other persons I don\'t remember and me. This poll was made by our hypocrite (according to what people have told us) friend. Then it came another poll in which it said: Did you went Patricia\'s party, options,I wasn\'t invited, No, Yes, I don\'t fucking care. In the comments they start to making fun out of her. Finally I post a page long comment complaining about this generation and our friend Chris delete the poll. Now the web page is mainly owned by Chris since Tufo doesn\'t care about it and Carlos left the web page. I think mainly that\'s the main stuff about the problem and with that, you reader, can see our opinion. Thanks.  '),(13752,4008178,'24,July,2004','2004-07-24 00:00:00','   So! Review of last night\'s \"party\"!   *ahem* ok first of all, whoever wrote down those instructions and directions on how to get the is like an idiot who shouldnt have been born (pablo?) i mean, a brick colored house with two palms on it? thats like so totally perfect, considering that every  goddamn house in the place was brick colored and had at least two goddamn palms on it.   Even if you DID follow the instructions, you would\'ve ended like two blocks down the road, looking at the guardhouse. probably in the park with the rocks (park with rocks? wtf?). but i got there! after like 8 phone calls. xP&nbsp; but pablo is an idiot. its like \"keep going! now turn to the left, turn to the left again, again, and turn to the left! youre there! (if you noticed, i just made a turn around the block) god!    So i was there in the terrace/livingroom/thing next to garden/bar and kinda bored. like, chris making out with manuel\'s sister (o.O) and blah. and everyone talking about how they wanted to beat gurdian and estebitan up. its like   pablo: \"DID YOU HEAR?! i wanna kill gurdian! he\'s such a loser! i mean, he actually told me that i wasnt a true 07\'er! can you believe it?!\"  manuel: \"DUDE! what an asshole! do you wanna beat him up?\"  pablo: \"yeah! thatll teach him!\"  daniel: \"When he comes, i join! and i wanna beat estebitan up!\"   and hte convo goes on, but you get the idea.   but i mean, what the fuck, even chris got mad at me for saying \"he\'s kida tright about it\" and he like gave me a strange look and stuff. O.o;   so yeah, if any of you get into a fight, i get in, im not letting everyon beat you two up. (ill get my ass kicked too, but not so badly haha)    but the party was kinda boring. i mean, what the hell? i arrive there and people (drunk) start talking about my edge. what the fuck? and its all like: \"what do your fist say?\" \"uh... drugfree\" \"WOW!\" \"uh.. yeah.\" and then they start talking to like each other, and these two \"punk\" kids (who seem like they just got out of the quiksilver store, with beanie and all) start talking to me about punk. O.o;   one word: i dont care! one more word: I cant count!   who the hell wants to talk about punk? its like:  kid: since when are you edge?  me: like 3 years ago.  kid: cool.  me: uh yeah.  kid: you like hardcore?  me: uhhunh. you?  kid: yeah..  me: like what bands?  kid: uh... crass and uh.. green day.  me: ok.  kid: you?  me: [dump bandlist here]  kid: oh. ive heard minor threat!  me: good. i uh.. gotta go.  [gurdian and estebitan&nbsp;arrives] ::relief::    and so when gurdian and estebitan arrived, i told them that ppl wanted to beat them up. and estebitan starts talking about it out loud. JESUS MAN, SHUT UP! i mean, you couldnt be inciting the fight even more XD   so we got bored of just standing around in there, and decided to go for a walk around the place. because.. uh... we dont really familiarize with most of the people there. at least not me.   so we sorta walked around.... late night... uhm, there was another party going on in one of the blocks in which people started whistling around at us. i hate people. lol. anyways, we kept on walking and heard some music on my radio then stopped at the park with the rocks and started talking about pr0n. yeah. kinda bored,eh? i also took a piss like a meter away from their faces. xD   i cant piss when people are watching me. its hard to.   at eleven i had to call my parents to tell them to pick me up at one because \"the party was great and theres a lot of people\" while gurdian made a body image of \"youre talking crap\" which sorta consited acting out crapping and hten taking the shit into your mouth and spilling it all over. yeah.   we came back into the party like at 11:20. bo-ring. a lot of people had come. well no. but anyways, we ate some cold hot dogs we found resting on the bar. we&nbsp;gulped them down with some water. yeah. then we like randomized and helped cleaning the place up. poor maid, everyone\'s an asshole.   but maybe ill rant about the maid some other day. and about the 07 generation. but someone else might wanna do it. what about you estebitan? im gonna get ready to leave now. be back in three days!    urlLink www.hotelpuntaislita.com  (thats where im going :O)   &nbsp;  [edit] oh, to end last night\'s convo, if i found daniela on my bed with her legs spread open, i would pull her eyes out and fuck her eye socket. im a sick fuck, but daniela is a hpocrite [/edit]  '),(13753,4008178,'26,July,2004','2004-07-26 00:00:00','   I was bored today.&nbsp;&nbsp;         LOVE ME!  '),(13754,4008178,'26,July,2004','2004-07-26 00:00:00','   Esta misma noche del&nbsp;25 de julio del 2004&nbsp;a las 0:18 am. Felipe GurdiÃ¡n y yo (Esteban Alvarez) comenzamos el plan para un nuevo movimiento llamado â€œUnidos sin HipocresÃ­aâ€. Este movimiento tiene 7 pasos fÃ¡ciles de seguir si nos da el apoyo que necesitamos para que funcione (Messeguer esto tambiÃ©n es para usted).&nbsp;     &nbsp;    1) Todos tenemos que ponernos el mismo nick en MSN:  -Â¡Â¡Â¡Â¡Â¡VAMOS A UNIFICAR ESTA GENERACION Y DEJAR DE SER HIPOCRITAS!!!!!  Seamos honestos tenemos una generaciÃ³n dividida y no podemos sacarnos este nick por nada  2) Explicaremos este plan a dos de nuestros amigos (Robin y Jose) y pediremos su ayuda  3) Pondremos propaganda que incite a la uniÃ³n de esta generaciÃ³n alrededor del Colegio Lincoln  4) Pediremos a la gente que defienda a los molestados y los ayude  5) Hacer reuniones o â€œfiestasâ€ y vamos a poner a la gente a q se siente y hable de este nuevo cambio y de la forma hipÃ³crita q muchos hemos actuado entre nosotros  6) Vamos a mantener esto con constancia y con fuerza y con esto tambiÃ©n buscamos q la gente no se pelee y q la gente se apoye si no es mucho pedir  7) Cada paso que se cumple se tiene que hacer con ganas y con fuerza para que la gente vea q este plan no es cualquier estupidez que no va funcionar como el anterior. TambiÃ©n vamos a tratar de mejorar este plan con propÃ³sito de mantener unida esta generaciÃ³n   Estos pasos necesitan de constancia y de empeÃ±o asÃ­ que por favor el que lea esto sea fiel a nuestra idea. Y para Messeguer, Jose y Robin: tenemos que mantenernos en  contacto todos los dÃ­as , absolutamente todos los dÃ­as y comentar sobre esto para ver como van nuestros planes. Gracias. OjalÃ¡ esto no sea para ustedes mÃ¡s que letras al azar y que de verdad esto signifique algo para ustedes, porque al final esto nos va a ayudar a todos.&nbsp;        A final word, wait it\'s not a call to action. We ain\'t no sect we ain\'t no fucking faction. Unity unity youâ€™ve heard it all before. This time it\'s not exclusive we want to stop a war. Unity as one stand together. Unity evolutions gonna come. â€“ Operation Ivy  '),(13755,4008178,'30,July,2004','2004-07-30 00:00:00','   okei ya estoy harto de toda estra mierda de que si los mae playos se pueden casar o no.  primero que todo, me gustaria acalarar que TODA ESTA MIERDA sobre estar casado o no es un monton de picha. que significa estar casado? todacia no se para quÃ© el registro civil quiere saber si uno esta casado o no. me parece una imbecilada. es como que en su cedula saliera si ud tiene tele o no. a quien le importa? toy seguro de q a mis apreciadas figuras de autoridad no les importa.  ahora, para que tanta mierda de que los playos no se pueden casar? okei, diganme PORQUE PUTAS NO!?  porque va contra la iglesia? y que?! entonces para que dejan que hayan nightclubs por toda la ciudad? porque matan gente? porque es que los pedazos de mierda que se estan quejando TAMBIEN HACEN PECADOS?!  porque desde mi punto de vista es lo mismo serplayo que cometer un pecado: sigue siendo un acto contra dios.  ahora, no es que tenemos que mantener la distancia entre el estado y la iglesia? ahora, QUE PUTAS ESTA PASANDO AQUI?! porque lo que veo no es gente discutiendo el tema con un apunte objetivo, son un monton de playos de la iglesia quejandose de que dios no quiere esto y dios no quiere aquello y si no lo hacemos nos quema la casa. saquen a los padres de la tele y de los periodicos! porque con uno de esos cogechiquitos aqui no va a cambiar nada.  y que pasa con la gente que nisiquiera es catolica? esa gente q se casa con omo 20 zorras porque les da la gana? esa gente que?! no es que se merecen un voto en esto? AH PERO NO! los unicos con derecho a discutir son el padre minor y mister bush. no me jodan, al chile, si van a empezar con esta mierda al menos hagan algo con otros temas como \"porque es que todos los padres son unos violadores?\" no se pongan a discutir con los playos. hasta donde yo se, ellos no pasan violando a los menores como ustedes.  bueno, y mi punto final:  en conclusion, yo quitaria la vara de poder casarse \"civilmente\" porque estaria enfocando solo un lado de las religiones. no todo el mundo es igual. que se quite la posibilidad de casarse! y ya no hay mas problemas!  p.s.: a los playones que se quieren casar bajo el nombre de la iglesia: no pueden. metanselo en la upa, monton de estupidos. eso es como querer unirse a una secta satanica porque le gusta dios. es como hacerse sXe porque le cuadra emborracharse. es una imbecilada. callense y den una buena discusion, no una mierdita sobre jesus.  '),(13756,4008178,'30,July,2004','2004-07-30 00:00:00','   Hoy fui al corazon de lo que es \"pipi\".  hoy fui a multiplaza. vi a dos carajos que andaban muy orondos con su nuevo look \"punk\".  Una faja de studs camisa quik gorra camionera quik y un rristban (wristband).  MAE ESAS GORRAS HACE DOS AÃ‘OS USTED LA HUBIERA CONSIDERADO \"CHOLA\". que asco USTED ES TOTALMENTE LO OPUESTO A LO QUE SU \"MODA\"DICE. mae usted si es cool quiero ser como usted otro hijueputa en esta sociedad decadente.  '),(13757,4008178,'31,July,2004','2004-07-31 00:00:00','   I\'ve decided to post the tab for part of one of my favorite guitar riffs. It\'s from the Goldfinger cover of 99 Red Balloons. I\'ve never seen it tabbed this way, on any sites with guitar tabs, but I prefer playing it this way.  Goldfinger plays it with heavy distortion, which is why it sounds spiffy.  E------------------------------------------------------------------ B------------------------------------------------------------------ G----------------------------------------------------------------- D--------------------4h6p4------------------------------------------ A----7-7-4h7p4-------------7-7-4h7p4-------------4h7-------------- E-----------------------------------------------7-7--------------------  Click  urlLink here  for the tab of the whole song. There are mistakes in the solo, and possibly in other places, so watch out for those. Other than that, it\'s a pretty good tab.  Note: blogger won\'t let me post the vertical line-things, which is why the tab looks messy.  '),(13758,4008178,'03,August,2004','2004-08-03 00:00:00','   so what the heck are people thinking.  actual news in costa rica:      (transcribed from my memory)      \"some random guy from the states wants to come into costa rica to search for gold in somewhere blablabla. this will give costa rica more money and more jobs. bla bla bla\"      ok i mean, WHAT THE FUCK.      id like to point out that this guy\'s \'gold search\' is going to be sponsored by the govenrment. BY THE GOVERNMENT. JEBUS.      and theyre not even fucking sure if there is gold in the place. what if there\'s no gold in the goddam area? well the government, not the guy, will have spent millions of dollars to find out that theres no gold, they just fucked up the environment, and they have an external debt because of some weird treaty the country made. (like always)      what jobs is it gonna give? because they always say \"oh, its gonna give jobs\" but its bullshit. they usually do this: ok lets bring 40 locals and make them work 10 hours a day at minimum wage for a year. after that, they fire them all. jesus man, that was hell of a job oportunity, specially when the next people you hire are stupid gringo tourists who propel sexual tourism. go figure.      out of the gold that is extracted, how many of it is the country gonna keep? this are the things that my beoved president would probably do:      1- donate the gold to the US for no apparent reason   2- share the money among themselves   3- spend the money on building roads, which translates to having Ignacio Santos buying cars with the money, or making random amounts of cement poke out of the roads, ike you can see near plaza del sol.   4- give it to the guy who promoted the gold thing.   5- give it so that we can \"boost tourism\" which means sped it on nightclubs and more advertisements of night clubs along the road   6- give it to the church, so that priests can become richer   7- give it to me. ill make good use of it, trust me. xD   8- give it to the MEP, so that THEY CAN ACUTALLY PAY THE TEACHERS   9- im out of ideas.      well you get the idea.      second, why would i want to help them make this proyect? so that the president can become richer and NOONE noticing? fuck off, im not paying my taxes for that.      can you imagine the enviromental damage? ok lets see:      -the biggest erosion that costa rica has ever seen, apart from the teca wood.   -lots of dead fish because there\'s no water   -less frests, more dirt roads   -houses having entire mountains falling on them      well you know. read mamita yunai. whats going on right now is the same as to what happened 70-100 years ago, in the book. (i donno when it was. sue me.)      i say, fuck the proyect. thats not the goddamned way to go. and if youve got a problem about it, im fucking waiting.          NO REFORMATION WITHOUT  REVOLT  '),(13759,4008178,'04,August,2004','2004-08-04 00:00:00','   me siento agradecido de por lo menos tener 5 buenos y confiables amigos en una generacion de 90 hipocritas.   y creer q en algun momento pense unificar a esta genercion.  '),(13760,4008178,'04,August,2004','2004-08-04 00:00:00','   la verdad es que no me importa si me odian. eso es lo q menos me interesa. tengo otras cosas mas importantes que discutir. -messeguer      esteban - Â¡Â¡Â¡Â¡VAMOS A UNIFICAR ESTA GENERACION Y DEJAR DE SER HIPOCRITAS!!! says: mae metz es una mierda todo IT STILL LOOKS PRETTY UGLY TO ME! :@ says: *sigh* IT STILL LOOKS PRETTY UGLY TO ME! :@ says: sip IT STILL LOOKS PRETTY UGLY TO ME! :@ says: pero yo no quiero unirme IT STILL LOOKS PRETTY UGLY TO ME! :@ says: porque mini nunca va a pensar igual esteban - Â¡Â¡Â¡Â¡VAMOS A UNIFICAR ESTA GENERACION Y DEJAR DE SER HIPOCRITAS!!! says: mae yo ya me desanime IT STILL LOOKS PRETTY UGLY TO ME! :@ says: jaj. esteban - Â¡Â¡Â¡Â¡VAMOS A UNIFICAR ESTA GENERACION Y DEJAR DE SER HIPOCRITAS!!! says: todos creen q yo les huelo el culo a ud y a gurdian IT STILL LOOKS PRETTY UGLY TO ME! :@ says: vale. hay mejores varas por quÃ© pelear. esteban - Â¡Â¡Â¡Â¡VAMOS A UNIFICAR ESTA GENERACION Y DEJAR DE SER HIPOCRITAS!!! says: y q soy un wannabe punk IT STILL LOOKS PRETTY UGLY TO ME! :@ says: vale picha IT STILL LOOKS PRETTY UGLY TO ME! :@ says: a quien le importa lo q otros digan esteban - Â¡Â¡Â¡Â¡VAMOS A UNIFICAR ESTA GENERACION Y DEJAR DE SER HIPOCRITAS!!! says: di si esteban - Â¡Â¡Â¡Â¡VAMOS A UNIFICAR ESTA GENERACION Y DEJAR DE SER HIPOCRITAS!!! says: la bara es q yo creia q podia confiar en ellos IT STILL LOOKS PRETTY UGLY TO ME! :@ says: mae, hay q pelear por otras varas, un carajillo muriendose de hambre y nosotros hablando de que hay q unir la generacion. esteban - Â¡Â¡Â¡Â¡VAMOS A UNIFICAR ESTA GENERACION Y DEJAR DE SER HIPOCRITAS!!! says: mae si al chile en este momento yo me paso por el culo a la generacion \"07\" IT STILL LOOKS PRETTY UGLY TO ME! :@ says: jaja IT STILL LOOKS PRETTY UGLY TO ME! :@ says: a la mierda con esa \"union\" lo unico q hacen es pelearse  '),(13761,4008178,'04,August,2004','2004-08-04 00:00:00','   okei, todos sabemos q eso es caca. es algo en lo que yo no decidi unirme. asi que aqui esta la lista de esos elitistas qe se creen unidos:  manuel arce manuel alvarez mini daniela montealegre karina ruiz paola ruiz pablo ramirez hernan cordero daniel vivero roberto ortiz  a los maes q escriben aca (tebitan y gurdi) mae sigan updateando la lista. Ni se metan con esa gente, dejenlos vivir su mentira, la verdad esque no necesito amigos como esos basos.  '),(13762,4008178,'06,August,2004','2004-08-06 00:00:00','   long have you awaited it... maybe the original is gone, but it was written by stupid racist people so here you have... THE HITLIST! (to add something, just edit this post) if youre gonna post something, make sure it hasnt been posted. use ctrl + f to find a word.      - abel pacheco   - fascists   - george bush.   - george bush jr.   - goths   - punks   - sk8ter bois   - posers   - hot topic   - marylin manson   - bad music   - U.S.A.   - Avril Lavigne   - McDonald\'s   - American chocolate   - people who don\'t take a stand   - nazis   - people who can\'t debate   - fast food   - sellouts   - capitalism   - people who piss me off   - rumors   - ravers   - DJ\'s   - one man bands   - leaders   - GG Allin   - Taco Bell   - fashion   - people who act like theyre different but are just like everyone else   - modern artists   - mindless people   - bad news reporters   - morals   - corruption   - meat-eaters   - cheap cd-r\'s   - stores   - stupid teachers who made my life hard    - labels   - hentai   - good cars   - people who look down at you   - bad hair gel   - keyboards in spanish   - fat people   - anorexic people   - pop punk   - quiksliver and all of that crap  '),(13763,4008178,'06,August,2004','2004-08-06 00:00:00','   CONFLICT LYRICS  The Guilt And The Glory (How Many More Lies)   Do you know that at least 70% of the human population of this planet starve and many more do not get enough to eat? You must now be familiar with reports of the famine In Ethiopia, where the weather has been so dry that crops fall, where the people are so hungry they have to walk many miles to the nearest relief station where there might be food If proplr have been generous. In any one station at least 40 people a day die. totally shriveled. They haven\'t the energy to brush away the flies that are waiting for THEIR meal. These poeple are the victims of our western governments. In Britain alone there are massive great food stores where surplus grain sits, In the EEC there are food mountains, In Amrerica farmers are paid to destroy part of their harvest, while big corporations plunder the 3rd world countries at the expense of the natives. Why Is this - economics, the science of greed. That which says people must pay western prices for their food. Natural starving countries can\'t afford this especially when their government takes the best fIrst. The western government have plenty of money. They tax us on our income, then they tax us on what we buy. They take the money we have worked for and use It to buy arms. They would rather be able to destroy the world 12 instead of 10 times over than transport surplus food to those who die. Indeed, they become 111 through obesity. Must we tolerate these obscenities? When soon even we may be starving because of their taxes. Some people don\'t see the connections between big bussiness and destruction. The biggest murderers are MacDonald\'s. Not only do they breed and slaughter millions of cattle, great rain forests are pulled down to make way for grazing ground. This destroys hundreds of unique species each time not to mention the fact that we need trees to produce oxygen for the planet. Many birds In North America migrate to these forests, on arriving and finding no forest they die. The drop In population of birds means more insects and the Americans only answer to this is pesticide. We all know about factories who manufacture pesticide, you know about Union Carbide. They\'re an American Company who set up dangerous plants. It doesn\'t matter, nearly 3,000 Indians are now dead due to their callousness, thousands have strokes and injuries that will never heal. Innocent people who had done nothing, whose life is not easy anyway, and why? So America can steal food from the starving millions. Nearly all big bussiness, is based on explanation of the worst kinds. Cosmetic companies use dead animals to make their products, then poison many more by feeding it to them with the end result of encouraging sexism by convincing men and women to look the same as each other. Advertising promotes this and advertising of any kind usually portrays women as a sex object only, there to be raped. Drug companies (like Roche &amp; ICI) claim to Invent new drugs to cure Imagined diseases. Thousands die at the hands of their useless research. They try to make drugs to cure heart disease which was originally caused by the sufferer eating meat and dairy products to excess. Booker McConnell own everything from butchers to health shops, as long as the cash rolls in who cares? Other companies who make machinery and domestic appliances (like Thorn EMI.) also makes defense gadgetry. British Aerospace specialize In fighters in favour of passenger jets. Governments pay more for weaponry that domestic consumers pay for domestic items. That\'s why many many comapnies world wide clamour for defense orders, some would go bankrupt without them. The final Industry, uranium mining, sponsored in Britain by RTZ plunders the landscape, kills the wildlife,exploits the natives and gives them cancer. The product makes bombs to kill all and 15 used for nuclear fuel. The kind of power that produces waste which can be pumped Into the sea In Cumbria. Deformed babies and children dying and leukemia are coincidences we are told. how many more lies must we listen to?  '),(13764,4008178,'07,August,2004','2004-08-07 00:00:00','   What the fuck is everyone thinking? i mean, they get mad when they find out this psycho guy kills cats, but theyre all too happy eating their fucking hamburguer. for god\'s sake.  i showed this link:  urlLink http://www.textfiles.com/anarchy/MISCHIEF/1001cat.txt   and people got pissed off. why? because they couldnt believe i would kill a cat.  \"you woud kill a cat? what did the cat do to you in your life?\"  at least in my eyes, its the same to kill a cow, than to kill a cat, than to kill a squirrel, than to kill a fish, than to kill a human. youre still a killer, you took someone\'s life away.  so why do these people complain so much? i mean, maybe i\'m not vegan, but if i went vegan it would be for the good of it, not so that only one animal loses its right to live.  anyways, my dad just talked to me so i lost interest in going on writing. i guess you get the idea.  bai.  '),(13765,4008178,'07,August,2004','2004-08-07 00:00:00','   what the fuck is everyone\'s problem? everyone i see in a random forum on intrenet is all like \"WOW MAN I GOT FUCKED WHEN I WAS 12 I RULE! EVERYONE I KNOW IS A BIG SLUT! I ROCK! I FUCKED A SLUT AINT I COOL!?\"  it pisses me off. specially the way these people look at girls. \"sex object only\" come on, you couldnt get worse like that. a girl is not a fucking prize. i could go to any fucking place and get a blowjob whenever i cant, but its just not cool. wow your best friend is a slut so you brag about it, but deep inside you know she has fucked basically every other guy you know. thats not fucking cool. it makes you sound like an idiot actually.  anyways, if what you wanna do in your life is be a \"fuck machine\" go ahead, but youre no better than anyone else around here, and i dont fucking need you scrubbing it in my face, asshole.  thank you.  '),(13766,4008178,'07,August,2004','2004-08-07 00:00:00','   okay im sick of people trying to sound smart when they couldnt possibly be. this is an actual e-mail: (after responding to another one)      Just trying to figure out how massive amounts of chaos equals setting things straight, but that\'s alright.'),(13767,4008178,'08,August,2004','2004-08-08 00:00:00','   In response to one of Felipe\'s previous posts:  Sex is supposedly \"in\" right now. Take MTV for example -- nearly everything revolves around sexuality. If mtv says it\'s cool, half the teens in the world will think it\'s cool. The minds of teenagers and children and easy to manipulate, and brainwash. There\'s really not much we can do about it. I think it\'s safe to say that more people will believe mtv/abercrombie/etc rather than a few kids who can actually see the truth, or at least something near the truth.  If everyone wants to fuck around, it\'s their choice. In fact, I have a friend who lost her virginity earlier this summer. She\'s only 13, and she did it with some older guy. The guy, I\'ve heard, only likes girls who are pretty and easy. I lost a ton of respect for her.   We can hope that they\'ll understand someday (although hope is usually just a waste). It just sucks how so many people lack common sense these days.  '),(13768,4008178,'08,August,2004','2004-08-08 00:00:00','   What are vegans in a short definition: people who try to change the world without thinking too much. The simple fact of not eating any meat is stupid. You say you don\'t want to become an animal killer, but I bet you eat and wear child-made products and you don\'t give a fuck. I\'m sure you don\'t think about it when you buy your tasty Milky Way with processed chocolate in a poor African country, or when you buy your Quicksilver shirts or your Nike shoes, I can understand that because you also have to be worried about what people say about you in school. Let me tell you something Tilio told me: \"\"School\" is just a fraction of your life\" and that is true, unless you are going to die at the age of 20 and you know it, then you are really fucked up.  I hate those \"persons\" who call them self humans and claim to be charitative, while they spend 50 cents for every $5 invested in their mascots, to charity. That is hypocrisy not compassion. Showing off their $100 shoes and their father\'s brand new Porche Carrera GT. I think that\'s the worst combination in a human being: hypocrisy and superficiality. If you want to be hated for the rest of your life, this is the key for your success, either this or killing the pope and twenty babies, but that is sick and difficult. Today the average human has became blind, shallow, and ignorant. Just like Felipe or Susi (Felipe\'s alter-ego) put it: Who the fuck cares about you having sex with a bitch! Good for you, but you may end up with herpes or aids or even a child, so now, does that worth it? Now, everyone cares about what\'s good for him. If you are combulsing in the middle of the street they see you as an animal with nothing to lose and probably most people think you are crazy. Only few are the ones that stop and help you, few are the ones that really care about you and the million persons around them. Those few are the ones that difference us from ants, they don\'t walk in one direction and fuck everyone who is on their way. They stop and talk to you and gives a meaning to what does being a Human is. If you enter a store nobody cares of saying \"Good Morning\" or \"Thanks\", everyone\'s is thinking \"Dude, that\'s cool, I want it and nobody is going to take it away from me, even if don\'t buy it ever. Yes! I am a soul-less consumist, BUY! BUY! BUY! MC DONALD\'S, QUICKSILVER, CELL PHONES, MORE!!!.\" Today it is like if you are   the only human left in the world and everyone else is just inferior.  Consumism is another thing is making everyone look like an ant. I say an ant because ants keep walking without caring about the other ants, they only fulfill their mission and nothing else. They don\'t talk they only work. The only difference between ants and humans is that ants are more consent with what they do and how they do it. We harass with the environment killing the animals and lately killing our selves. The good thing is that at least we are trying to improve and to correct our mistakes, but still not all the countries do that, for example U.S., Kyoto, does it tickles something? If it does then fuck you for doing nothing!!! Continuing with the subject... consumism is what is making us all the same. Disney is the ultimate consumist empire. There you can pay for a hotel and you can buy food, toys, shirts, toys, play in the theme park, and buy toys! You pretty much can to everything there, and since everything is Disney\'s monopoly in there you just consume. You may be asking your self... What about it? And that is exactly my point. Nobody cares. They just accept it, like if its normal, and certainly it is normal. Everyday we read advertisements, watch advertisements, and some people are even human advertisements. That is not bad at all, what is bad is to buy things that we don\'t need. Things that are just unusable is what is stupid. You could invest with that money in other things, either for needed people or just throw it at the bank. Other thing that comes along with this problem is than people feel insecure about them selves because they don\'t fit with the stereotype companies show to the consumers in the advertisements. Many fat people end up having problems for using products and pills that promise you you will lose all the fat in your body, and the ironic thing is that after those advertisements finish Mc Donald\'s, Taco Bell, Subway, KFC, commercials come in. What the fuck! Corporations are using us and the only thing that can help us decide what we want or what we need is us.  End    PD.: I don\'t know how I started talking about vegans and finished talking about self determination.  '),(13769,4008178,'09,August,2004','2004-08-09 00:00:00','   School starts in a week. Hopefully, I will be able to finish my summer English assignment before then. I really need to get it together, I\'ve only finished about a quarter of the novel. DDD: Schedule: 1. French 1 2. Biology 3. Geography Honors 4. World Civilizations 5. Honors Art 6. English 9 Honors 7. Algebra 2  Blah, I can\'t believe I\'m taking every honors/\"advanced\" class available for my grade (9). >. I really hope this school year will be better than the last. I got into a ton of trouble this past school year, and I hope my record doesn\'t carry over to the new school. I\'m kind of annoyed because I have Geo Honors and World Civs with a really annoying Star Wars fan who loves to argue with teachers over stupid topics. It would\'ve been nice if 9th graders could take US History, since my expertise includes The Civil War, but World Civs and Geo Honors will be fun too, I suppose. Yeah, yeah, enough of my ranting over school subjects. I\'ll just end this now, since I seem to have nothing else to write. Ta.  '),(13770,4008178,'09,August,2004','2004-08-09 00:00:00','   this is a current event i have it due this friday for US HISTORY so its is full of shit.  Article:   urlLink  http://news.bbc.co.uk/2/hi/business/3549408.stm                This article is about the rise of oil prices in the world setting a new record in history. This is caused by the menace Iraqi militant cleric Moqtada Sadr set after rejecting calls to end the fighting. This implies that Mr. Sadr may sabotage refineries and pipelines in Iraq since Sadr followers had taken over the cities of Basra, where 90% of Iraqâ€™s oil exports go through. Iraqis officials shut down the plants to avoid a possible attempt of a terrorist attack. This reaction caused the rise in prices. In the article also talks about the impact Yukos may have with this. Analysts say it will take about six weeks for the effects to come.             It is pretty obvious that this is a great problem the world is undergoing. The simple fact that oil has made a new record in prices alarms everyone. If oil prices are high, then, considering that much of worldâ€™s oil comes from the Middle East, countries around the world will suffer its effects. Another problem within the subject is the effect it may cause to Yukos, which is one of the worldâ€™s biggest oil companies, since if it haves a big impact on Yukos economy, then prices will go even higher aggravating the problems.  '),(13771,4008178,'09,August,2004','2004-08-09 00:00:00','   FAWK! ok so im getting bored of vacations, and im all like, man! i wanna go back to school and talk to friends blablabla. I really miss school. going back to school is gonna be a blast!  BUT NO! it had to be another way! i had to get there and stand under the sun for like 2 hours! and listening to the principal talk crap!  and guess what?! I GOT US HISTORY TAUGHT BY MY PRINCIPAL! WOOHOO im gonna fucking kill you, asshole.  jesus. and all my teachers are stupid except for the math teacher and the computer teacher. think im going to die now. tada.   ps: wE ONLY HAVE AN HOUR OF RECESS! NOO  '),(13772,4008178,'13,August,2004','2004-08-13 00:00:00','   hoy me conto mi amigo felipe de ese comentario q puso ud. me parece bien q piense de esa forma, pero me gustaria q escuche al menos mi opinion antes de saltar a concluciones.     primero me voy a presentar. me llamo esteban alvarez y estoy en el hueco de la lincoln. si, seguro ud esta pensando... otro hipocrita q juega de ser revelde sin causa... pero dejeme decirle q yo no soy asi. agradesco a dios todos los dias por haberme suituado en una familia q tiene el dinero para darme d q comer todos los dias y para darme la educacion q \"nesecito\". esto y mas siempre pasa por mi mente y ...pienso q ojala pudiese ayudar en algo servir para algo, ser algo mas q un digito en el registro civil... tambien ha de pensar ud q yo soy un hipocrita por qjarme del colegio licoln y a pesar de ello seguir alli, pero sabe por que lo hago? porq yo pienso q si yo estudio y me esfuerzo lo suficiente podre ser alguien para cumplir mi anelo. suena como q si me interaza la politica pero no. yo qiero ser medico y no un medico vendido q nada mas le interesa ganar mas dinero, yo qiero ser un verdadero medico. a lo mejor no le interesa saber en absoluto esto, pero dejeme decirle algo... con esto qiero q vea q yo no soy cualqier otro superfulo de la lincoln, yo ya encontre mi proposito en la vida.  a mi personalmente no me gusta conciederarme un \"punk\" porq eso es ser un \"poser\" y a mi no me gusta rajar en ninguna forma, pero como ser humano y por mi imperfeccion como tal, a veces lo hago, y despues pienso lo tan imbecil puedo ser. yo no busco su aprovacion ni tampoco su amistad, si la consigo de alguna forma me es indiferente. lo q qiero q qde claro es q yo no soy un \"punk\" ni \"rude boy\" por q es una \"moda\", a mi me gusta esta musica y la lirica en ella q me hace pensar... yo tengo el poder para hacer una diferencia. tal vez sea solo un sueÃ±o nada mas... tal vez sea nada mas estupides... pero quiza sea posible, y gracias a esta ultima yo soy qien soy.  en nuestro \"prestigioso colegio\" nosotros somos odiados. quiza sea porq nosotros qeremos q la gente como nuestros compaÃ±eros de clase habra los ojos y vea lo q pasa en el mundo, pero no la gente es cerrada y discrimina. a ellos les gusta vivir en su burbuja y nada mas. yo personalmente solo tengo 4 amigos en toda esa muchedumbre y lo ironico de todo esto es q esa gente nos odia por ser diferentes y qjarnos del factor comun en ese colegio, la superficialidad, y tratamos de ser simples lo mas posible, y en la otra mano los proletarios nos odian por \"tratar de comprenderlos\".  me da pena q ud haya tenido q vivir todo lo q vivio en su infancia. pero creame yo trato de cambiar el mundo en q estamos, y si.... ojala algun dia nos encontremos... para ver que.  '),(13773,4008178,'15,August,2004','2004-08-15 00:00:00','   the cafta agreement destroys the environment, american companies come to third world countries and destroy the land. when the soil wont produce any more, they leave the country, sell the land back to the government, and go on destroying other countries.  the companies that come here do not produce any jobs. sure, the government says \"oh this new hotel they are building will bring jobs to the area\" but what the companies do is to hire locals for a month, fire them and then hire americans to work there.  the companie\'s wastes are thrown into rivers, which then spread an epidemic across the country. the municipality or however you call it in english wont do anything about it, because the local company is constantly giving them money, so they wont do anything. have you ever been fired because you didnt give out a permit to some random conpamy you dont even know? because giving out the permit would be illegal? therefore, if any problem arises, you are the guilty one. but the mayor keeps acquiring money, and if you dont give the permit, he wont get money, and so they will fire as much people as needed just to get 100 bucks  people do not need an american company to come into a 3rd world country to fuck around with the poor. did you know that if we want CAFTA to exist, all 3rd world countries would have to close every company currently on a monopoly? did you know that almost every country depends on these monopolies to exist? why? well most of the money goes directly to the government. the government, in exchange, gives free services to the lower classes, and people without economical advantages. if there was no monopoly, the country would probably go un-communicated, most parts would have no more electricity or phone lines. the american countries that come would also come in with really high prices, and probably the middle class wouldnt be able to afford it. The high class would be the only ones getting any services, therefore they would be the only ones aware of what is going on in the country.  what for? so that the gap between rich and poor can widen? so that the people with money can puppeteer what goes on inside the country?  did you know that after the treaty is signed, any copyrights made in the country will belong to the united states\' government after ten years? what does this mean? that if i write something, the author will be the united states. that the united states can change whatever they want to change, whenever they want to.  did you know what happened to mexico after the agreement was signed? they had an economical deficit. sure, the u.s. says \"oh mexico\'s economy rose like hell after the treaty was signed\" bullshit. we all know what really happened to mexico. after walmart came into mexico to exploit the comercial areas in the country, the country lost mos of its money, and then the economy \"rose\" when they had to sell the national bank... to the U.S.  for what do we need any american countries? doesd america really need to steal more money out of poor people? do they need to come in here to destroy the country? why do i have to say yes to anything that a pissy gringo tells me to do? since when do i follow orders from someone who enjoys killing people? if theres someone who needs the be the US\'s yes-man, let him be, but im gonna stand against him with everything ive got. so that the americans can laugh at me and spit on my soil? no thank you.  '),(13774,4008178,'16,August,2004','2004-08-16 00:00:00','   1. what are you doing now? What do you think, fucktard? 2. what were you doing before this?  Wouldn\'t you like to know...  3. what are you thinking now? No comment. 4. what time is it? 1:07 pm 5. what time did you sleep last night? Around 1am 6. what did you do last night? Lurked on Gaia, read. God damn, I hope I finish that novel before Thursday, I might get kicked out of my English class. D: 7. what did you last say to a friend last night? Uhh, me and Charmbracelet were discussing the Olympics. O.o 8. what attracts you to the opposite sex? *shrug* 9. what makes you happy? Playing my sexeh guitar, slash fics, some other stuff. 10. what are you wearing right now? Sleeveless shirt. Jeans. 11. what will you do after answering this? Answer the next question... -.- 12. who was the last person you talked to on the phone? My mom. 13. what do you want to tell someone but you don\'t know how to tell him/her? No comment. 14. when is your birthday? Up yours. 15. when is your bf/gf\'s birthday? n/a 16. when did you last cry? Few days ago 17. when did you last smile? This morning. 18. why did you answer this? Whatthefuck. 19. when will you marry? NEEEEVVVERRRR!!! MUWAHHAHAHAHAAH!!! o.o; 20. when did you last receive a gift? Two months ago. 21. when did you last drink alcoholic beverages? Two weeks ago. 22. who will you vote for president? Not old enough to vote yet. I don\'t think my parents are voting in this election either. They don\'t like any of the candidates, and I don\'t blame them. 23. who was the last person you talked w/yesterday? Myself. O.o 24. who makes you happy? Blah. Don\'t know. 25. who is your enemy? Humans. >. 26.Qualities you like in a guy/girl? Intelligence, good taste in music, understanding. And there\'s something sexy about guitar/bass/drum players. :B 27. who do you hate most among your friends? No comment. 28. who\'s your closest among the females? Charmbracelet. 29. who has the cutest smile you\'ve ever seen? My doggie  30. who do you talk on d phone with often? Charmbracelet and myself. Gah...I just realized how pathetic I am for calling myself. D: 31. do you belong to a team in your school? No. *social anxiety* 32. how\'s school? Shitfuck. 33. any regrets? Yes, many. T_T 34. do you smoke? drink? Nah. Smoking blows. 35. are you a clubber? HAHAHAHHAHAAH. 36. do you have someone in mind that you want to marry? No. 37. do you like swimming? It\'s okay. 38. do you wear eye contacts? Yeah. 40. do you kiss your parents good night? *snort* 41. does your crush often send you messages in friendster? n/a 42. how about in e-mail? O.o 43. how often do you use the internet? Hours and hours everyday. DDD: 44. how is your hair styled right now? Down. 45. how is your feeling right now? Empty. 46. how is the weather? Cloudy-ish. 47. how do you treat your loved one? ------ 48. how did you learn about friendster? *pretends to know what that is* 50.do you miss some one? Probably.  Ah, so now my life has been reduced to wasting myself away in front of my computer, filling out idiotic surveys. Oh, and photoshop and guitar. :P One of these days I\'m going to sex up that jason ellis showmaster guitar. I hope. o.o;  '),(13775,3792567,'30,July,2004','2004-07-30 00:00:00','   Someone asks you for a letter of recommendation.&nbsp; The job she seeks will probably be in the same field.&nbsp; How do you do it when you have nothing to say?&nbsp; I can say she\'s a good Mom, she has a great heart.&nbsp; She is reliable and she has good organizational skills.   What she doesnt have is the ability to work with others.&nbsp; Patience.&nbsp; Zip on the problem solving skills.&nbsp; She has a short fuse and a heavy mouth.&nbsp;   Maybe this is only because she hates her job.&nbsp; Maybe its the reason why she has decided to leave.&nbsp;   But - I cannot in good heart and mind put that in writing for the next employer.&nbsp; I would be lying.&nbsp;   Thats what she has given me - and it is what I have seen.  If she was so awful - why did she stay employed for so long?&nbsp; She\'s not awful or maybe we have just learned how to get around her moods.&nbsp; In the end things get done - but at a price.  I am confused and feel badly that I cannot give a gleaming review.&nbsp; I don\'t want to hurt anyones feelings or chances at gaining employement for themselves.&nbsp; But - I don\'t want to be the reason someone hires someone.&nbsp; My word is good in this business.&nbsp; Good as gold.&nbsp;   Does that mean that any employees we hire in the future you keep&nbsp;at arms length?&nbsp; Can that be done in an office so small?&nbsp; Is that what \"I am the boss\" means?  It\'s a very sticky situation and honesty is always the best policy.&nbsp;&nbsp; Uggghhhh!!!!  I need a beer.  &nbsp;  '),(13776,3792567,'29,July,2004','2004-07-29 00:00:00','   Ok - a couple of posts back I said Hope - its Help.&nbsp; So that I don\'t misrepresent the DNC Candidates.&nbsp; God Forbid.  '),(13777,3792567,'29,July,2004','2004-07-29 00:00:00','   I don\'t think anyone has been&nbsp;interested in my&nbsp;\'going\' since I was a baby.  You know when your Mom makes sure you \'go\'?&nbsp; Makes comments about it. Or, when you don\'t \'go\'&nbsp;they get all concerned.&nbsp; \"She has a bellyache today - she hasnt \'gone\'.\"  But my \'going\' has become common talk.&nbsp; Why, because I suffer from IBS.&nbsp;   IBS stands for Irritating Bull Shit (not really - but I seem to scream that when running for the can).  If you\'ve never experienced this \'syndrome\' your really missing out.&nbsp; It\'s a joy.&nbsp; Things that you would never think of become regular staples in your car.&nbsp; A bucket, some toilet paper, baby wipes&nbsp;and a change of clothes.&nbsp; You become a major share holder in Imodium AD Extra Strength.&nbsp; You know every&nbsp;can available from here to there and you dont even care if it hasn\'t been cleaned in a year.&nbsp;   Your family and co-workers know that when you say pull-over you mean NOW!&nbsp;  You consider diapers - but you still have enough pride left to not get them  There is no cure.&nbsp; They blame diet, stress, lifestyle.&nbsp; So one by one - you eliminate all the things they say that trigger you.&nbsp; Nothing works.&nbsp; So you work around it.&nbsp; You learn to live with it. Its on your mind - every day - every hour.&nbsp;&nbsp;&nbsp; These are things no one should have to worry about except my Gramma.&nbsp; She worries about stuff like that.&nbsp; My girlfriend with the newborn, worries about stuff like that.&nbsp;  I may be seeing a light at the end of the tunnel (no pun intended).&nbsp; It\'s been six years I have been dealing with this.&nbsp; I seem to have been getting a break recently for whatever reason (maybe someone is taking pity on me and my family and friends).  Whatever it is, I am grateful - though it\'s made for some great comedy.  Now that I don\'t have \'going\' on my mind as much - I am gonna have to \'come\' up with something new to worry about. (Ba Dum Bump)  &nbsp;  '),(13778,3792567,'28,July,2004','2004-07-28 00:00:00','   Ok - I will admit last night I watched some blurbs of the DNC (that name kills me everytime).&nbsp; I briefly saw a girl talking about her mom - a mom talking about her husband.&nbsp; Then the husband (who looks more like the son) gave a speech very similiar to the one Jesse Jackson gave a few years back. (Ya know the one -&nbsp; Keep Hope Alive)  Edwards mantra was - Hope, is on the way.  Is it me? That the platform they all use - is&nbsp;same shit - different day? Doesnt matter which party.&nbsp; I am not going to discuss which party is better.  Tax cuts - cut prescription cost - medical care (blah, blah,blah)&nbsp; Hey - but hope is on the way.  Edwards made his money from suing companies for millions.&nbsp; He is a lawyer.&nbsp; His profession is the reason our insurance rates are so high.&nbsp;   Some will say - \"yes, but the little guy he represented that was hurt made millions too\".&nbsp; I understand that - but its going to continue. No matter what party is in the house.&nbsp; We had Democrat in for 8 years.&nbsp; Why didnt anything occur about the issue while he was in office?&nbsp;   Well, I will tell you.&nbsp; I will refer to Saturday morning cartoons (which everyone loves)  The President is a figure head.&nbsp; He does not have power over much (ok, he can veto stuff - but he can\'t make laws).&nbsp; The people that hold the key to anything being done are in the Congress and the Senate.  urlLink http://www.school-house-rock.com/Bill.html &nbsp;&nbsp; (Remember this when you were little?)  These are the folks you need to pay attention to when voting.&nbsp;   They say yes or no. They make things happen.  The majority of the time nothing gets thru them because they are so busy bickering and taking sides and having fillabusters so that nothing gets done (remember that little lovely thing last session?)  I learned when I bought my home - Washington didnt effect me.&nbsp; The City Council did.&nbsp; My taxes and where I can dig on MY property is the decision of a 6 person panel.&nbsp; People don\'t need too look far to see the people that really effect them.&nbsp; There are so many levels of Goverment before you EVEN&nbsp;get to the President.  When you vote for President - your voting for the \'voice of the USA\'.&nbsp;&nbsp;He represents America.&nbsp; If you don\'t like the way he represents you - then vote him out.&nbsp;   But - if your primary concerns are the issues they speak of - you need to vote the PROPER&nbsp; people into the Congress and Senate.&nbsp;  In November - you are voting for the mouthpiece.&nbsp; I think my mouthpiece is fine.&nbsp; Sure, he may mispronounce a word or two.&nbsp;Then the media has a field day.&nbsp; But most of America&nbsp;didnt know the proper way to say the word anyway.&nbsp; So&nbsp;making fun of him is funny but useless.  I was happy that Bush was in the White House when terror struck.&nbsp; He stood in front of those cameras and said what every American was thinking.&nbsp; \"Watch out Terrorist, we will hunt you down and eat your balls for breakfast!\"  BUT - he had to get it \'APPROVED\' first.&nbsp; He knew this and did what he had to do (even if he said he didn\'t - whatever).&nbsp; He knew to go after the Terrorists he had to pass it thru.....(anyone? You Sir, with your hand up in the back)&nbsp; Thats right...Congress!  People don\'t agree with the war - I feel safer today than I did on Sept 11.&nbsp; I was terrified.  People blame him for oil prices - Its not our oil.&nbsp; You won\'t let him dig - remember?!?  So then it all comes back to - health care, prescriptions and tax the rich, give to the poor.  I am telling you - &nbsp;if you keep re-electing the same old men year after year as your Congressman and Senator&nbsp;(don\'t play dumb - some of these folks are pushing 80 - I wouldnt trust them in a car and on the road but you keep re-electing them to make your decisions!!!&nbsp; What are you thinking???) nothing will get done.  We will continue to put in a handsome man (All of our Presidents look good - they are mouthpieces).&nbsp; Wonder what his wife is wearing.&nbsp;   The old - bloated guys in office will maybe come in if they don\'t have a doctors appointment.  Hey, but Hope is on the way!  &nbsp;  &nbsp;  &nbsp;  &nbsp;  &nbsp;  &nbsp;  '),(13779,3792567,'28,July,2004','2004-07-28 00:00:00','   Last nite was a good one.&nbsp; Reality shows galore.  Too bad I didnt have any popcorn (its has carbs - right HB?)&nbsp; J and I sat there all night for a TV extravaganza.  The 2nd installment of Trading Spouses was on last night.&nbsp; I don\'t know if anyone caught it.&nbsp; But the white lady came off horrible.&nbsp; I hope she sat there and watched it with her family last night and was embarrassed.&nbsp; It\'s going to be an ongoing show - it will be on again next Tuesday with a new family.&nbsp; The next family previews look like they are gonna be good!  Big Brother was on at 9.&nbsp; I love this show too.&nbsp; Where the casting directors got these people I will never know.&nbsp; They are some of the most horrible people on the planet.&nbsp; A few seasons back there was a man named Will who won the game.&nbsp; Will was a skunk, but not a mean in your face skunk.&nbsp; He played the house like a fiddle.&nbsp; Deserved to win because he was a great player. The men on this program \"The Four Horsemen\" are arrogant, evil hearted and mean\". They have apparently gotten thru life on their looks because their personality sucks rotten eggs.&nbsp; I am hoping with last nights show that there will be a turnaround for the other half of the house.&nbsp; Cross your fingers that the tiebreaker (Head of Household) will make the right choices.&nbsp; He comes off as a weenie but he may be a really good player.&nbsp; We will see Thursday.  The Amazing Race was also on last night.&nbsp; Regretfully - it was up against Queer Eye.&nbsp; I had to make a choice and Carson the Fashion Diva won.&nbsp; They did a great job on the man they were working on (as usual).&nbsp; I think every man needs to make a gay friend. It only seems to be a benefit.&nbsp; Even the most manly of&nbsp;men on the show end up in tears because of how grateful they are to the Q.E. crew.  The DNC (fitting name don\'t you think? For the men out there - its a procedure for women) was also on.&nbsp;  But guess what?&nbsp; I didn\'t watch it.&nbsp; I will wait for the debates.&nbsp;   &nbsp;  &nbsp;  '),(13780,3792567,'27,July,2004','2004-07-27 00:00:00','   I am either really cranky or I have a short fuse this week.  Let\'s talk about lines.&nbsp; I am talking about lines that are formed when people are waiting to get to one specific area.&nbsp; Lines are formed to create order.&nbsp; One by one - people form a line - one behind the other.&nbsp; Each person is taken care of in order.&nbsp; Lines may be long, but if you WAIT YOUR TURN you will be waited on too.&nbsp; See simple.  Well - I live in a&nbsp;place where many people have migrated from some third world country where apparently they do not have these things called lines.  We have lines at stores, banks and even those merge lanes on the highway can be considered lines.&nbsp; Lines work.&nbsp; Lines prevent others from screaming that they were there first.&nbsp; Lines make life fair and everyone is happy and can move on with their day.  Some people don\'t believe in lines at all.&nbsp; My father used to make fun of a commercial that said \"stay in the lines - the lines are good\" refering to coloring.&nbsp; I am all for someone showing their creative side.&nbsp;&nbsp; But - not while I am waiting in line.&nbsp; Do it on your own time.  I think that if you come into this country - you need to learn a few things.  First - To speak the language (thats English just in case your confused)&nbsp;Even if just a little.&nbsp; At least I know you are trying.  Second - To drive.&nbsp; I think this has a lot to do with lines again.&nbsp; There are lines on the road for a reason.&nbsp; Stay in your lane.&nbsp; If you are going to come in my lane.&nbsp; Warn me...I dunno maybe you can use your turn signal provided in the car?&nbsp; I am just sayin.   Plus - the honking thing as soon as the light turns gets old.&nbsp; Take a chill pill.  Third - To make a line - a single line.&nbsp; I understand that some of you come from worlds where its first come first serve. You have to scramble, crawl and push to get what is rightfully yours.&nbsp;  But - America has so much to offer.&nbsp; We have quite the bounty of everything you could possibly ever want.&nbsp;  If you just wait in line.&nbsp; That bag of Yucca could soon be yours too.  '),(13781,3792567,'26,July,2004','2004-07-26 00:00:00','   My daughter and all her infinite wisdom told someone that we would be more than happy to take a dog that needs a home.  I lost my dog of 13 years in February to old age and various ailments and I am taking my time to decide if I even want to care for another doggy right now.&nbsp;&nbsp;&nbsp; Why would I say \'I\'?&nbsp; Because we have become doggysitting central since Luckys passing.&nbsp; During that time though I have enjoyed playing and loving and feeding these animals.&nbsp; What I havent enjoyed is the bathroom time.&nbsp; These dogs large and small leave little presents which I need to pick up or spray down.&nbsp; They need to go all the time.   J (as I will refer to her) has loved having the company of a dog.&nbsp; She will drag them into her room so she can sleep with them.&nbsp; But - when they start to stir and need to go out she staggers to her door opens it and lets them out...into the living room.&nbsp; Once they are there - you can here the \'tap,tap,tap\' of their toes on the tile.&nbsp; \"Where did the human go?\"&nbsp; Then the \'tap,tap,tap\' in my room.&nbsp; \'I\' have to get up and let them out to pee - smell a flower and take a poop.&nbsp; \'I\' have to pick it up.&nbsp;   So - what type of dog has she agreed to adopt?&nbsp; A german shepard.&nbsp; A&nbsp;HUGE beast that has huge poops.&nbsp; What the ...  I have mentioned a little doggy - a Bichon or something from the teacup family.&nbsp; Little poops.  She\'s not happy at all.&nbsp; (whatever)&nbsp; When I actually see her get up and let a dog out then pick up the mess that occurs then she can have a say on what dog we can have.&nbsp;&nbsp;   A german shepard - What was she thinking???  '),(13782,3792567,'26,July,2004','2004-07-26 00:00:00','   On my way in this morning I listen to news radio.&nbsp; I dont use it for traffic reports as my work is only down the road from where I live.&nbsp; I listen to it because its news, straight and to the point.&nbsp; I can hear whats going on in the world in the 10 minutes my drive takes me.  We today - Kim Commando (who is some computer guru) has a blurb about blogs.&nbsp; Did you know that some blogs are paid in excess of $5000 a month by advertising companies?&nbsp; Apparently it will continue and maybe one day you and I will be rich, rich, rich (wild laughter).  So - as I am logging in I am pondering what type of advertising company would be interested in me.&nbsp; I tout FOX and Marlboro and Starbucks. I can think of alot more - I havent even mentioned but it would be stuff like booze, and Imodium and marital aids (uuhh -kidding Mom). But seriously....I can sell Starbucks.&nbsp;   Well back to ...as I am logging on...I am pondering....I already have an ad on the top of my blog/diary.&nbsp; Does anyone care to guess what someone is selling on my Blog/Diary???  NO SMOKING AIDS!!!&nbsp;   Irony.  '),(13783,3792567,'26,July,2004','2004-07-26 00:00:00','   I am tired of watching what I eat.&nbsp; I want to eat whatever I want.&nbsp; I am tired of picking and choosing.&nbsp; Making good food choices.&nbsp; I am tired and bored of it.  I don\'t want to think and make changes in the menu.&nbsp; I want to eat that potato that comes with the steak and I want to eat the entire bag of fries.&nbsp; I want to order a girle drink loaded with sugar and I dont want to have to just have Vodka because it has 0 net carbs.&nbsp; I miss pita bread - I miss pizza.&nbsp; I miss garlic bread and mac and cheese.&nbsp; I miss a Whopper with the bun and a 12 foot sub actually in the roll.&nbsp; I want hashbrowns with my eggs and toast too!!!  I don\'t want to be reminded that today will be the day we will start back on the strict no carb regimine because some else is feeling fat.&nbsp; If I feel&nbsp;like taking a break and so what if it cost me a few pounds.&nbsp; I miss carbs!!!&nbsp;   Its Monday HB and \"I want\" in -&nbsp;this -&nbsp;order....A whopper with cheese, large fries and a chocolate milkshake.&nbsp; If you want yours without the bun and a glass of water please feel free to have it.&nbsp;   But if your hand gets in my way - I will eat that too!  &nbsp;  '),(13784,3792567,'23,July,2004','2004-07-23 00:00:00','   My daughter will be attending private school this year for the very first time.&nbsp; Her old public school was horrid and if I kept her in there - they would have swallowed her up and spit her out.  But - with private school comes, uniforms.&nbsp; Now in the grand scheme of things you would think uniforms are great.&nbsp; Not having to think about what your going to wear.&nbsp; Having to keep up with the Jones family.&nbsp; Or avoiding that fight in the dressing room over whether the pants are too tight or the shirt too short.&nbsp; But - uniforms come with issues all thier own.  They are not made for every body type.&nbsp; I am not even saying they are made for Twiggy.&nbsp; The model that they use has a very long crotch and no butt to speak of.&nbsp; Now&nbsp; I do have to say they are unisex.&nbsp; What fits the girls must fit the boys (I guess thats where the long crotch comes into play).&nbsp;   The employees just stand outside of the dressing room and say the \"thats how its supposed to fit\" thing.&nbsp;   All the girls were&nbsp;mortified.&nbsp; Whining and complaining.&nbsp; Hissing noises could be heard thruout the store.&nbsp; Teeth were sucked and comments were said under their breath.  I decided then and there - this is how alcoholic mothers are made.  &nbsp;  &nbsp;  &nbsp;  '),(13785,3792567,'23,July,2004','2004-07-23 00:00:00','   Monday.&nbsp; It comes around week after week.&nbsp; Usually the day after a Sunday.&nbsp; Everytime after a weekend that you dont want to end.&nbsp; Mondays, bleech!  **On the Reality TV note.&nbsp; I spoke about a Wife Swap show.&nbsp; Its really called Trading Spouses.&nbsp; Its on FOX.&nbsp; It will have an encore presentation tonight and then the second show will be on Tuesday as regularly scheduled.&nbsp; I really liked it.&nbsp; I hope you make a point to see it.&nbsp; **  The ciggie commies were in full force this weekend.&nbsp; As most of your know - Florida residents have voted to keep smokers out of restaurants.&nbsp; Most establishments freaked because they knew that it would seriously hurt their bottom line.&nbsp; Soooo...most (not all) have decided to make outdoor seating areas for the smokers. Some of them not acceptable AT ALL&nbsp;.&nbsp;But, some places have made little terraces of paradise.&nbsp;&nbsp;&nbsp; I will admit - I would much rather be sitting in the A/C but I have been banned to the patio - so I will make the most of it.&nbsp; Well apparantly&nbsp;so have all the ciggie commies that voted that I cannot smoke inside.&nbsp; They are taking up tables&nbsp; that are SUPPOSED to be for ME!&nbsp; Or making little comments like, \"I dont want to sit here - she\'s smoking\".&nbsp; Do you see where this is going...???&nbsp;   I am gonna eat at home.  &nbsp;  &nbsp;  '),(13786,3792567,'23,July,2004','2004-07-23 00:00:00','   The best piece of&nbsp;advice that I can give you today is open your mail.&nbsp; I am not talking junk-mail.&nbsp; I am talking about those standard things you get from you car note or your insurance company.&nbsp;   I have become lazy and rip them open with my finger.&nbsp; Peek inside (not even removing from the envelope) then throw it in a pile.&nbsp; Well - its insurance time.&nbsp; When all the policies have come due.&nbsp; I have become lazy because I don\'t pay them.&nbsp; I pay my mortgage company and they pay them.&nbsp; So - when I see \"this is not a bill\" I say - \"ok\" and pile them away.  So - I am looking into refinancing so I have become interested in how much it cost me to insure my stuff.&nbsp;   Well first off - I just got a new car.&nbsp; I received a statement for the old one.&nbsp; I would have been inclined to throw it away.&nbsp; I dont have that old car anymore.&nbsp; Well - it was a cancellation notice with a little \"check of love\" in it.&nbsp; I almost threw it away.&nbsp; And - State farm would have never said anything.&nbsp; Or placed the credit towards my new car, because duhhhh they already sent me a check.  Second mail flub - this is a biggie.&nbsp; Some of you know I live near a LARGE body of water.&nbsp; So they MAKE you get flood insurance.&nbsp; Just in case a huge wave flys over the big condos and floats my way.&nbsp; Never paid attention.&nbsp; I dont pay it (well I do - but since I dont write a check I don\'t ever look at the bill).&nbsp; Well - I open - scan - jaw drops to the floor.&nbsp; If in fact it rains and the ocean decides to do a back flow into my home I only get 10% of my homes value to fix it and you know the furniture and stuff that would be the first to get ruined by the litter box they call they ocean...$1000.00.&nbsp;   \"Hey, Rooms to Go guy, can you outfit my home after this flood I had for one grand?\" (can you hear the laughter??)  So I call them - \"Uh, did you miss a digit or two or three?\" \"No, its been like that since the beginning (going on 4 years!).&nbsp; But - call this man and for a small fee he will come over and we will figure out what we can overcharge ( oh - I am sorry charge) you to increase the number to fit your needs.  (Smacking myself in the head - whap, whap whap) There goes my \"check of love\".  Then the final one was a HUGE package from another insurance agent - Its the policy they send you every year with all the changes that they made (you can understand it so off it goes in the pile).&nbsp; So it starts..&nbsp;Its legalease - blah blah blah. \"Sign here and return as soon as possible because your WINDSTORM (hurricane policy) has been REASSIGNED and you will get a CREDIT if you return this to us so we can notify your Mortgage company.&nbsp; FAILURE&nbsp; to comply will&nbsp;leave your policy \"NULL AND VOID\" (which means no hurricane coverage).&nbsp;   (Smacking myself in head again)  Read your mail!!!  '),(13787,3792567,'22,July,2004','2004-07-22 00:00:00','   You know the song?   Friday...(do doo do do) Thank God&nbsp;it\'s Frii-daayy.   Friday...(do doo do do) Thank God it\'s Frii-daayy.   (I know your singing it to yourself now - your doing the music in your head...do do doo do).   Guess&nbsp; what? Now your stuck for the rest of the morning.&nbsp; Just thought I would put it in there for fun.   Friday. (do doo do dooo) Thank God it\'s Friii-daaayyy.   Friday.&nbsp; (do doo do dooo) Thank God its Friii-ddaayyy.   &nbsp;  '),(13788,3792567,'22,July,2004','2004-07-22 00:00:00','   I dont share this Blog with you know who (H-B).&nbsp;   He thinks this is mindless drivel and thinks I need to get back to work (whatever).  But - he shares with me today the blog he reads and then goes on to rant and rave about how the \'adults\' are running the country....blah, blah snooore. I should read more about the world(zzzz....sorry were you saying something)  I think useless rambling is a good thing.&nbsp; That\'s what women do.&nbsp; We gab.&nbsp; We can talk for hours.&nbsp; We can talk over and over again about the same crap - day after day.&nbsp; Mindless is good.&nbsp; Stupid is good.  That\'s why I like magazines.&nbsp; Cosmo, Lucky, Elle and People.&nbsp; Just a bunch of ads - random thoughts and sometimes those free samples of perfume too (extra bonus)!  Men on the other hand read...National Geographic, Sports Illustrated, Time (do you see what I am getting at?).&nbsp;   We flip - they read.&nbsp; We browse - they sit holding our purses and stew thinking of all the things they could be doing to be more productive to this great world of ours.&nbsp;  But - let\'s think what mindless acts men do.  HB can sit and smoke a cigar on the back for an hour.&nbsp; Not reading just looking into space.&nbsp;&nbsp;  HB can make an afternoon out of watering the lawn (also done while smoking a cigar).&nbsp; I am not speaking of setting up the sprinkler either.&nbsp; He will grab the hose and stand there in one spot seemingly forever.  He will flip. Channel after channel. Over and over.&nbsp; Not slow enough to see whats on that channel he just passed - just flip.  Now he will tell you - he\'s contemplating.&nbsp; Thinking and planning.&nbsp; Mind always working.  But then how come when I ask him what he\'s thinking he says \'Nothing\' ?  &nbsp;  &nbsp;  &nbsp;  &nbsp;  &nbsp;  &nbsp;  &nbsp;  '),(13789,3792567,'21,July,2004','2004-07-21 00:00:00','   Ok - I just got another call from my Gramma.&nbsp; Can I say something? (I will anyways because she doesnt own a computer).&nbsp; You know how the 9/11 report is coming out today?&nbsp; Well, apparently they think we need another level of intelligence.&nbsp; Forget the CIA or the FBI or Homeland Security guy.&nbsp; No -&nbsp; We need someone else.&nbsp; We need...spies!!!  Well - I would like to nominate Gramma.&nbsp;  She knows everything about everyone.&nbsp; Somehow in her Gramma way she gets everything from you.&nbsp; Even if she doesnt know you.&nbsp; If you are in the WinnDixie and you see her coming towards you.&nbsp; I suggest you run.&nbsp; She will smell you out, hunt you down and know your whole life story before you even know what hit you.  She won\'t tempt you with cookies.&nbsp; Stale bagels and old coffee cake&nbsp;(maybe).&nbsp; But she has this power to bomb you with question after question.&nbsp; You don\'t even think your answering her.&nbsp; But at the end of the conversation she knows EVERYTHING!  When I have something to hide.&nbsp; I avoid her. It\'s best that way.  But she always catches you - she makes friends with your co workers.&nbsp; They are like \"what a sweet old lady\"&nbsp; \"Shes so funny\".&nbsp; But what they don\'t realize is she sucking their&nbsp;brains for info because by the time they pass her to me she knows stuff about them I didnt even know.  I used to think I was tired after talking to her because she talks so much.&nbsp; I figured it out - Shes not talking - she\'s sucking my brain for info - I just \'think\' &nbsp;its talking as there seems to be a buzzing in the air.&nbsp; See that\'s how she gets you.  She\'s Americas number one spy.&nbsp; No man can resist&nbsp;Gramma.&nbsp; Those terrorist won\'t even know what hit them.&nbsp; Forget the language barrier.&nbsp; She\'ll just talk louder and s-l-o-w-e-r and soon we will know all thier secrets.&nbsp;   Believe me - I have seen it in action.&nbsp; My brother ex in-laws didnt speak a word of English - but she was doing hand gestures,&nbsp; that mime thing and soon they were all hugging and singing Kumbaya.  With Gramma in power - we will rule the world!!!!  &nbsp;  '),(13790,3792567,'21,July,2004','2004-07-21 00:00:00','   I have a new car.&nbsp; Hunny-Bunny says no smoking in the car.&nbsp; It\'s my car.  Whats with that!?!  Everyones against me today.&nbsp; Ciggie commies.  &nbsp;  '),(13791,3792567,'20,July,2004','2004-07-20 00:00:00','   I now have a staff.&nbsp; Sounds funny- I know.&nbsp; But I am thinking about it and I have a staff. I have a maid (Hunny-Bunny says they are Housekeepers - whatever).&nbsp; I have a gardener and I have a pet-walker.  I laugh to myself about this because I always thought...\"I am woman hear me roar!\" I could do my own damn laundry , mulch my own yard and feed and&nbsp;walk my own dog.  But - I have decided that I don\'t want to.&nbsp; I want to come home and chill out.&nbsp; On the weekends I want to do fun stuff and I am tired yelling about the condition of the house and yard. So I hired so people who take care of it for me.  Now - being the boss of these people has it\'s downside.&nbsp; I have now gotten very picky.&nbsp; More picky than I ever would have asked from myself.&nbsp;   I will give you an example.&nbsp; The maid was almost brought to tears working on my house the first time but she did a fine job.&nbsp; But now I have come to expect excellence.&nbsp; Why?&nbsp; Because I am paying her.&nbsp; She cleaned my fridge yesterday and it sparkles.&nbsp; She didnt dust the top.&nbsp; I find myself shaking my head and saying \"Tsk, tsk, tsk\".&nbsp; I have a gardener - she has made my yard a thing of spectacular beauty. I have flowers everywhere and I love it. But God forbid she doesnt get that weed that I have been eyeing all week.&nbsp; You mean \"I\" have to bend over and pull it?!?  The dog walker - she\'s ok for now except for that alarm fiasco - plus isnt perfuming the dog after a walk an option?&nbsp;  These people are not making a fortune from me by any means.&nbsp; They are working for well&nbsp;below minimum wage for what they do and doing the dirty work I have given them.&nbsp; But - somehow they all show up week after week and take pride in thier work.&nbsp;  My shirts are folded perfectly.&nbsp; Like they were ironed one by one.&nbsp; You should see my linen closet - I kept opening the door over and over in amazement.&nbsp; Yep - still pretty.&nbsp; Everyday I drive up and its like Christmas.&nbsp; What did my staff do for me today?!?!?  I love my staff.&nbsp; I need to relax and I need to pull that weed!  &nbsp;  '),(13792,3792567,'20,July,2004','2004-07-20 00:00:00','   I get smoke breaks at work two 15\'s and lunch.&nbsp; My boss is an ex smoker.&nbsp; He should understand. It takes 5 minutes max to smoke one cigarette.&nbsp; So - I should get 6 breaks total at 5 minutes each if I don\'t use my 15 minutes at once.  It makes sense to me.  '),(13793,3792567,'20,July,2004','2004-07-20 00:00:00','   I looooove reality TV.&nbsp; I loved Real World when in came out in 92.&nbsp; Now its 2004 and I can\'t wait till the next season starts.&nbsp; I like Joe Schmo and Survivor.&nbsp; I liked Survivor so much that a few years back that when I was traveling on business I asked my client if I could follow&nbsp;him home with his partner just so I could see the Final Episode of Survivor (Tina won that year - I so thought Cody had it in the bag).&nbsp; Granted, I had&nbsp;to watch it between the commercials of Will and Grace.&nbsp; But - hey, he let me watch it!  My hunny-bunny calls it Stupid TV.&nbsp; So I gently remind him of the other TV located&nbsp;in a different room.  Now is it reality?&nbsp; No.&nbsp; I don\'t even know exactly why they call it that.&nbsp; It\'s not like any reality I know.&nbsp; And then - somewhere in the show they have a part where they are going to try to hurt and humiliate them.&nbsp; Like the show....can\'t remember the name...it was on&nbsp;Bravo.&nbsp; It was a gay guy looking for love - the catch, some of the suitors were straight. He cried.&nbsp; It made for good TV.  The newest one out is WifeSwap (or something like that).&nbsp; The first 10 minutes grab your attention.&nbsp; You have a black lady with 3 kids and a hubby.&nbsp; The both have a job in the medical field.&nbsp; They live day to day and she rules that home with an iron fist and humor to boot.  The second wife is a waspy blonde.&nbsp; Married to a Japanese guy.&nbsp; They are loaded.&nbsp; I do not know what he does for a living but he put the word&nbsp;bling in bling bling.&nbsp; She lives with her Nana and two kids.&nbsp; They show her eating a really nice dinner in that first 10 minutes at home ( these little things are important to note).&nbsp; Then they are whisked away to live the life of the other.  Well - everyone is really nice at first.&nbsp; The waspy blonde is embarrasing - \"I want to meet everyone on this earth - I want to make the color of a rainbow with the people I come across.&nbsp; Hey kids - your white mama is home!\"  Well the black lady is amazed and gracious.&nbsp; She went right into her role.&nbsp; She made breakfast.&nbsp; She cleaned up after.&nbsp; She offered to make lunch for the daughter when she returned. She played games with the son.  The white lady - bitched and whined and complained.&nbsp; She rolled her eyes.&nbsp; She yawned and was bossy and demanding.&nbsp; Her shared family made their own breakfast&nbsp;(SHE DOESNT COOK)&nbsp;- ran to the store for her (SHE DIDNT DO INSTANT COFFEE) &nbsp;and one of the sons sat at a fuse box and kept flipping a blown fuse at least 40 times so she could dry her hair.&nbsp; Why she didnt change outlets is beyond me.&nbsp; The thing that killed me about her is she would not shutup about the shared daughters weight.&nbsp;   So whats the catch? They get 50k.&nbsp; The shared mother has to decide how the other is going to spend the money.&nbsp; The black family has about 30k in debt&nbsp;- she wants a used corvette and a louis v purse.&nbsp; I have no clue what the white lady wants or needs.&nbsp; It seems she needs nothing just looking at her lifestyle - but who knows.  My opinion - the white lady who has been harping on this 13 year old about her weight is going to do something about it.&nbsp; Not that the 13 year old asked for her help - she seems uncomfortable and on the verge of tears everytime white lady says something about it. But - they showed scenes from next weeks show where the real moms are reading the letters and the 13 year old is crying.&nbsp;   Thats my bet for the week.&nbsp; I will throw my trusty pillow at the screen if this happens and then I will curse the families for putting their kids thru this.&nbsp; Some people - I tell ya.&nbsp; Amazing.  If you watched it - whats your views?&nbsp; What do you think is gonna happen?  &nbsp;  &nbsp;  '),(13794,3792567,'20,July,2004','2004-07-20 00:00:00','   Yes- my dear little brother has outed my deep thoughts page.&nbsp; The pain - the embarrassment that now shall befall him....hee hee hee. &nbsp; As I had mentioned in a previous post.&nbsp; I would call this Diary.&nbsp; &nbsp; Just like a little brother to tattle and tell all his friends&nbsp;(Rasberries to you).&nbsp; All because I mentioned his ummm...shortcomings.&nbsp; I think his body fits his rather large&nbsp;head perfectly. (Don\'t laugh - its a small defect and he has feelings you know). &nbsp; I say this all in jest you know.&nbsp; He\'s a pretty great guy.&nbsp; We are a few years apart in age.&nbsp; So I remember when they brought him home.&nbsp; He was the little brother so he was picked on a little (only a little b-cuz&nbsp;I could have made him totally drink the pee if I really wanted). But - the good part was when he was old enough to kick my ass I was already out of the house. &nbsp; Recently - he has been going thru some changes in life.&nbsp; (I know its a little late for puberty to hit - but I told you, defect).&nbsp; And, he has come to me to let me know.&nbsp; I was floored and flabergasted as we never really shared.&nbsp; But it made my heart smile (just a little - don\'t think for a minute I was getting soft). &nbsp; He&nbsp;is bringing two little big heads into the world (this is where we start feeling the pain for momma).&nbsp;  &nbsp; Oh yeah = I only want to remind you (and Momma) &nbsp;that our parental unit&nbsp;had to get knocked out totally to pass your watermelon head. &nbsp; &nbsp;I think he will do a great job.&nbsp; My daughter digs him and thinks he\'s very&nbsp;funny (whatever). So there will be laughter in the house.&nbsp; He is the smart one (because his brain is bigger because of the head thing) and he can play an instrument (so the home will be filled with music - I am sure they are familiar with Metallica from the womb at this point). &nbsp; He has all the mannerisms of my father down (its spooky) so he will be the one that will make the kids tremble before he has said a word (Its all in the eyes) when they are in trouble. &nbsp; He has a good heart (that - he gets from me) and a good soul (uh - me again) and he a big softy ( he cries watching chicks flicks). &nbsp; He will do well -&nbsp;he will make pretty babies and he will be a great&nbsp;Daddy.&nbsp; I hope he has a boy and a girl. &nbsp; Just like us.&nbsp; &nbsp; &nbsp;  '),(13795,3792567,'19,July,2004','2004-07-19 00:00:00','   Well - not exactly twin.&nbsp; I think she\'s more like her Dad - but she will swear we are like each other. &nbsp; Its weird - when I look at her how I see her Dad.&nbsp; So spooky that she doesnt spend anytime with him for one reason or another and her characteristics are his.&nbsp; I only thought genetics had something to do with they way you look.&nbsp; Not the way you act.&nbsp; Your gestures and so on. &nbsp; But - the twin part came in yesterday.&nbsp; I like to shop.&nbsp; I don\'t even care if I buy anything.&nbsp; I like to look and oooohhh and ahhhh.&nbsp; My fave thing to say is...\"Thats cute, how cute, you look cute\". My daughter actually got dressed up to go out with me.&nbsp; Looked around and ooohhed and ahhed and said the word \"cute\". &nbsp; We didnt discuss politics or religion and when a nasty-ass song came on the radio she quickly changed it. &nbsp; As I had mentioned before she has been away visiting her NeeNee for two weeks.&nbsp; She has returned a girly-girl.&nbsp; My mom isnt girly-girl.&nbsp; What the heck happened!!!! &nbsp; I can only say - I hope that she maintains this characteristic that I adore.&nbsp; I think girls are meant to be just that.&nbsp; Sure - I want her to work just as hard as any man.&nbsp; But - you should always look \"Cute\" when you walk out the door. &nbsp; &nbsp; &nbsp;  '),(13796,3792567,'16,July,2004','2004-07-16 00:00:00','   I miss my youth.&nbsp; I miss spouting off the so called injustice I noticed in the world.&nbsp; I miss seeing everything in black in white.&nbsp; Why?&nbsp; Because then it would be easier to deal with the words that spew from the youth of today. &nbsp; Things are not black and white.&nbsp; They are grey.&nbsp; They will never be black and white because of the lack of responsibility people take for thier actions. &nbsp; This isnt a poor vs rich thing either.&nbsp; This spans humanity. The CEO&nbsp;who signs everything states he didnt know that the CFO was fixing the numbers.&nbsp; Its his job to know.&nbsp; The media mogul who refuses to say she\'s sorry for her actions.&nbsp; Instead, she blames everyone else for getting involved in her \"personal matters\".&nbsp; The young man who was high on drugs and wouldnt go with the police willingly so they zap him with the taser.&nbsp; He dies.&nbsp; According to his family he would have lived if the police would have left him be. So he robbed a few people.&nbsp; So what.&nbsp; Now they are suing the PD. &nbsp; I readily admit that I didnt take advantage of the education that was available to me.&nbsp; But, somewho I managed to work hard and make a living for myself.&nbsp; I am not going to blame the school system for not being able to name the 50 states without looking them up.&nbsp; I am sure that was a lesson.&nbsp; I chose not to pay attention. &nbsp; If I get a ticket.&nbsp; I am sure it\'s because I have done something wrong.&nbsp; You will never see me fight with a police officer on that one.&nbsp; They don\'t do it because they are bored or have a quota.&nbsp; I was breaking the law. &nbsp; I had a friend who blamed his family for his misfortune.&nbsp; His lack of anything good in his life.&nbsp; In fact it was his own lack of work ethic.&nbsp; His own lack of humility.&nbsp; That makes life hard for him until this very day. &nbsp; I am a victim.&nbsp; Poor me.&nbsp;&nbsp; Boo - Hoo - Hoo. &nbsp; I have always been fairly spirtual.&nbsp; I thought I was very&nbsp;religious but when I found that I broke too many of the rules in the code of ethics upheld in the churh that I was no longer welcome I decided to live by the 10 commandments.&nbsp;&nbsp;I truly believe that those 10 \"statements\" are what the laws of this land were originally based on and should be based on to this very day.&nbsp; I believe that if you live by these 10 \"rules\" you will be ok. &nbsp; Many people are fighting to have these \"rules\" removed from courthouses.&nbsp; I think that even though they are winning - I give kudos to the judges that are telling them to \"stuff it\". &nbsp; If you run \"The 10 Commandments\" on your MSN search page.&nbsp; You don\'t get the 10 Commandments.&nbsp; You get articles on how evil they are and how they need to be removed because someones eyes began to bleed when they saw them in the courthouse.&nbsp;  &nbsp; But - Is it so wrong to live by... &nbsp; Honor your Mom and Dad. (No matter how much of a tool you thought they were - you try being a parent) &nbsp; Dont kill. (No brainer here - whats so wrong with dont kill?) &nbsp; Dont steal. (Another no brainer - having people steal from you sucks!!!) &nbsp; Dont have sex with others while your married (oh this is the toughie - Romeo - get a divorce first) &nbsp; Don\'t use the Lords name in vain (Is is so hard NOT to say G-d Dammit or Jeeesus C-rist? - Get creative - There are much better bad words out there). &nbsp; Take a day and rest - rejoice in the beautiful world thats out there.&nbsp; (This means - get outside and take the time to see a sunset - the TV and work will be there when you return) &nbsp; Dont bear false witness against your neighbor (This means when Mary comes over for coffee one day - dont talk about Fran and say the only way they got that big new boat was because her Hubby sells drugs and shes a Ho.) &nbsp; Now the one - I have the most problem with is the&nbsp;thall shall not&nbsp;covet one.&nbsp; (Who doesnt look at the new pools and wish they had one - or the new car and dream?&nbsp; Or the hottie new hubby and Fantasize?)&nbsp; I am guilty of this one but you know how I remedy it?&nbsp; I wont steal it or kill anyone to get it or talk bad about the one that has it.&nbsp; So it equals itself out. &nbsp; See when you look at it in these terms its simple.&nbsp; Not black and white.&nbsp; But a shade of gray. &nbsp; &nbsp;  '),(13797,3792567,'16,July,2004','2004-07-16 00:00:00','   Three in one day.&nbsp; Bored - probably - but I have made another observation from my lame-o world. &nbsp; I was looking into other Blogs and I don\'t mean to be critical and I would never tell them so in the comments section, but.... (this is where I give my opinion though no one asked or cared to know). &nbsp; Where did all these very articulate folks come from?&nbsp; Come on,&nbsp; no one talks like this in real life. &nbsp; Plus - What normal Moe (that means Normal Moe in Geek speak) reads it and finds enjoyment out of reading something that needs a dictionary to translate? &nbsp; All I keep thinking about as I read the textbook-like blog...\"Look at the brain on Brad!\" &nbsp; Some of you are a little too deep for your own good.&nbsp; Yawn. &nbsp; &nbsp; &nbsp;  '),(13798,3792567,'15,July,2004','2004-07-15 00:00:00','   As I get older - I am now getting the most dreaded thing in the world.&nbsp; Uncontrollable farting.&nbsp; You know the kind.&nbsp; The kind that when your Gramma bent over to take something out of the oven it escaped from her rear.&nbsp; The kind that just ripped out of your Great Grammpas butt while he was watching the news?&nbsp; They never apologized.&nbsp; They would always blame it on the dog or a frog in the room. &nbsp; I have always been very concerned about being embarrassed by farting.&nbsp; So - I learned how to be \"silent but deadly\".&nbsp; The kind that sneaks out, ever so quietly and by the time the smell makes it to someone elses nose you have walked (quickly) away - never to be smelled again. &nbsp; I am losing the silent part.&nbsp; What is it?&nbsp; When you get old does your butthole loosen&nbsp;too?&nbsp; Is it necessary for exiting gas to make all that noise while your trying to escape? &nbsp; But then again - maybe its not me making all that noise.&nbsp; There was probably a dog or frog in the room.  '),(13799,3792567,'15,July,2004','2004-07-15 00:00:00','   I love Starbucks.&nbsp; I am such a fan - its disgusting.&nbsp; I spend on average about $1500.00 per year with them.&nbsp; Maybe more because I will take anyone there who hasn\'t experienced this lovely nectar and introduce them to the Starbucks Experience.&nbsp; In my neighborhood - they are about 10 blocks away from each other.&nbsp; I know each and everyone.&nbsp; The baristas know my drink of choice and I always leave with the best cup of coffee in town. &nbsp; I know I do because I have tried them all.&nbsp; I live in a town thats trying to revamp itself with neat coffee houses and spectacular restaurants and cute clothing stores and relaxing spas.&nbsp; I believe you should try to keep the money in the \'hood.&nbsp; So, I gave them a shot. Bleecch. &nbsp; Some people said I should make my own.&nbsp;&nbsp; Once again, I have tried - I failed.&nbsp; Double Bleech. &nbsp; What is it about Starbucks?&nbsp; Is it because they have thier own language?&nbsp; I like a Triple Grande Almond Latte.&nbsp; This means a medium expresso - three shots&nbsp;with steamed milk and some almond syrup.&nbsp; If I ordered it as I explained it - it would then be shouted to the Barista as Starbuckanese.&nbsp; Like a Frappachino.&nbsp; Its sounds good, tastes good - but its really only an iced blended coffee. &nbsp; My boyfriend doesnt get the language.&nbsp; He thinks somehow it makes him \"gay\" to say the words Triple Grande Almond Latte.&nbsp; He says that - because of the Starbuckanese that gives them every right to charge me $4.23 for a .99 cent cup of coffee.&nbsp;  &nbsp; I am starting to wonder.&nbsp; Could he be right? &nbsp; So - last night we venture into Cold Stone Creamery.&nbsp; Guess what they have...?&nbsp; Yep,&nbsp; a language all their own.&nbsp;I will call it Creamanese. &nbsp;I got a LoveIt Germanchocolatakake.&nbsp; Description they take ordinary chocolate&nbsp;icecream and them mix alot of goodies in it.&nbsp; Taste great - good portion and lots of stuff in it.&nbsp; Well - we get to the cashier.&nbsp; The total for two...over $10.00.&nbsp; I could have gone into Winn Dixie and bought all the stuff needed to make it and still have leftovers for days for $10.00. &nbsp; oh geez - Now I just had a moment of panic.&nbsp; Remember the Eddie Murphy movie where he is making fun of his Mom because he wanted McDonalds?&nbsp; She said she could make the same thing with stuff she had in the fridge?&nbsp;&nbsp;So she made a meatball with green onions and placed it between two slices of white bread.&nbsp; This does not a cheeseburger make.&nbsp; But it was less costly. &nbsp; I feel the Mommy gene kicking in - I actually thought for a moment that \"I\" could in fact duplicate the \"Creamery\" experience.&nbsp; Didnt I learn with the \"Starbuck\" making my own coffee fiasco?!? &nbsp; Ok - breathe....Every Mom reading this - please raise your right hand.&nbsp; &nbsp; I do solemnly swear that I will never, ever, ever&nbsp;try to recreate, make, or substitute something for the real thing.&nbsp; No matter what the cost.&nbsp; Amen. &nbsp; You and I both know its never the same - anyways.  '),(13800,3792567,'15,July,2004','2004-07-15 00:00:00','   Bill Cosby (that is). I have never been as big of a fan of Mr. Cosby as I am right this very second. Well - I liked him as Dr. Huxtable in the 80\'s. He was a funny Dad!   As you know, Bill has come out in public urlLink http://slate.msn.com/id/2103794/?GT1+4244  to say what was only said in secret behind closed doors and very low whispers.   The reason I like Bill right now is because of the culture that has totally soaked into teenage America. I think his cry was mainly for the African American citizen. But I think that it can carry over into White America. I know that as a mother of a teenage daughter who has teenage friends. Or, walking down the streets or in a mall you hear the same thing coming out of the youths of any race. I am always correcting them and I do that not only out of love, but because the words that are spoken now will not carry over into adulthood. It will never be acceptable unless you make your living speaking and acting in that manner.   It\'s not ok - and it never was or will be ok. I don\'t expect every teenager to act like little ladies and gentleman. I know I was very obnoxious with my \"gnarly, barf and like gag me\" verbage.  But - I didn\'t listen to music about women being bitches, how they act like ho\'s and their sexual acts graphically described.   I like music - always have. Great beat. But when I hear the lyrics it reminds me of that thing Mom\'s always used to say...\"Leave a little to the imagination\". Nothing was better than trying to figure out what we thought they were trying to say. If they were trying to say anything at all.  '),(13801,3792567,'14,July,2004','2004-07-14 00:00:00','   A place that is located so far south that its heat index is equivalent to the South American Rain Forests? A place where the bartender cannot afford to live and has to drive in from another place because the land is stupid expensive?  A place where you may have a hard time distinguishing the men from the women? A place so laden with ghosts and ghouls that all your film from the weekend has these tiny white spots (orbs)?  This place is Key West and my Spidey Senses were tingling.  Why?  Ghosts.  Yes sir-ree I do believe in ghosts.  I wouldnt say I am I am a seer.  But I think that I am ultra sensitive to spirts.  I believe that ghosts are like vampires.  A vampire cannot harm you in your home if you don\'t let him in.  It\'s only once he is welcomed - he can cause havoc on your soul.  Ghosts are the same way.  They are there.  I know they are.  For all those naysayers I have an example.  Your sleeping all cozy in your bed.  Then you feel someone sit on the bed - you feel the presence.  That, my friend, is a ghost.    I went as far a buying sage and \"cleansing\" my home with the smoky bush.  Why - I don\'t mind living amongst them.  I believe my Grandpa Louie-Louie comes and hangs with us sometimes.  I believe that because I know he came and said goodbye to me. What I dont want are the mean ghosts.  I dont want the playful ones either because they never seem to know when enough is enough.  But - if your a good ghost and you want to come chill out at the \"Casa del JacMariesmom\" grab a drink and keep the burglars at bay.  I like having Spidey Senses sometimes.  Two times in particular  on the trip made the tour worthwhile.  The first time was when I was walking thru Ernest Hemingways home and sitting in his gardens.  I went up to his studio and you could actually see him working - walking around in thought.  Heading to the bathroom.  The second time was at the Fort.  The Civil War never made it that far - though they were staffed to the hilt and people died at 15 a day.  But looking over the courtyard - I could see the soldiers busy at work.  The Commander at the desk surrounded by maps.  Some may say I only have a good imagination.  I say...they are ghosts.  '),(13802,3792567,'13,July,2004','2004-07-13 00:00:00','   My daughter and all her computer skills has found my Diary. Your a snoop o\' little one.  That\'s ok - she knows a lot about her Momma.  Warts and all.  My lil girl will be coming home this weekend.  I am truly excited.  I miss her more than you can imagine.  My parents have been gems though to give me this opportunity for us to have some \"alone\" time.  I think it has given us the opportunity to miss each other.  I hope she is as excited to see me as I am her.  I will cry when I see her.  She will roll her eyes and say...\"Mooooom - stop it!\" But - after I wipe my slimy nose on her and she laughs.  Its worth the ridicule.  In the end its me and her.  I tell her that all the time.  I am her biggest fan.  She is an excellent student and she has a brain bigger than most.  She has had some ups and downs in her life but she tries not to play victim.  She takes those things that hurt her and it makes her a stronger person.   I hope she takes advantage of all the opportunitys given to her. I hope she doesnt squander them and then regret it in the end. I hope she grows up and adds to society - that she makes a difference and doesnt live off the system.  I wish her only a good man when she\'s older that treat her with respect and honesty.  I wish for her a good job that makes her feel worthwhile.  I wish her only good fortune and never a black cloud.  Why all these hopes and wishes?  My friends daughter is in the hospital with Kidney Renal Failure. She is a very sick little girl and I send to her my prayers.  I send them to the little girl and to her Mother and Father so that have the strength to deal with the long road ahead.  The fact that they have to see their daughter so frail and in a dangerous time.  I wouldnt wish that upon anyone.  I have these hopes and wishes for my daughter - so that she will realize she has her health, she has the opportunity and she has the mind and the will to make anything happen.  Life is short.  Too short.  Never waste it.  '),(13803,3792567,'08,July,2004','2004-07-08 00:00:00','   Tony Bologna. Thats my pet name for my honey-bunny.  My shnook-ums.  My sweetie pie.  I know, I hear the gags already.  But - I don\'t care.  He\'s a good man.  I met him many years ago.  He was my Manager and I was a recent HS Grad and really didn\'t have much going for me in the workplace except my sunny-disposition (yes - I have one - well, every once in a while).  Well - he saw something in me.  Behind my Sally Jesse Rafael glasses and payless shoes. He gave me opportunities that no other employer would even dream of placing me in.  For this, I am grateful.  Time passed and we went our separate ways to other businesses.  Always keeping in contact as we both stayed in the same field.  He was someone I always called friend.  When money was tight for me or when we really needed some guidance and help, he was always there.  He taught me how to play golf (a hobby I miss) and how to portray myself as a serious businessperson.  Well, as luck would have it about 6 years ago we both found ourselves in a place where we were in failed marriages and our friendship flourished into a romantic whirlwind.  It wasnt easy.  Speculation, peering eyes over the eyeglasses and whispers were heard.  Blatant finger pointing and blame was screamed.  But - here we are many years later. Together and strong - while our \"others\" have been thru a string of relationships. Why? We still have that core of friendship.  He is my biggest fan - I am his. He is proud of me everyday and tells me so. He makes me feel so safe when I feel small and fragile.  He gives me the strength every day to get up and keep on moving ahead.  He tells me I am pretty. He is a good role model to my daughter and although he is not her \"real\" Dad you would think he is when she starts talking about boys and he makes a real funny face like my Dad used to make. Most importantly - He loves her as much as he loves me  Why am I vomiting my soul on this thing?  Who knows.  No one knows it exists.  I guess it\'s because when I was younger - I didnt keep a diary.  I couldn\'t.  I couldn\'t because my parents were very into what I was up to.  I can\'t blame them.  As a parent now I totally understand.  But - my most personal thoughts had to stay in my head because I would have been mortified if they would have read them. I was a good kid.  I never was pregnant or arrested. I always held a steady job. I kept the house clean and I graduated from school. But - they wanted to make sure I was REAALLLLY ok, so they snooped.   So with that said - I will not call this Blog.  I will call it Diary.  '),(13804,3792567,'07,July,2004','2004-07-07 00:00:00','   I am feeling old and ugly today.  Though I look cute (I really like my outfit).  My face is really bothering me (I know, its bothering you too).  I went to the beach this weekend. (Yes Mom, I wore sunscreen!) I even sat under a beach umbrella.  Now- what kind of day at the beach is this?  Remember when you were a kid and you piled into the car with your friends?  You would, sit at the beach all day.  Drank whatever concoction your freinds were able to make from the bottles in their parents bar.  Lube yourself up with baby oil and bake. For hours.  No care about skin cancer.  No worries about wrinkles.  You earned the burn.  Blisters the next day aside - Tan lines were an art.    Our tan lines were not sprayed on - rolled on or spread on. We didnt look a funny shade of orange from the chemicals that we rubbed on our skins.  This was sun damage all the way!!!  Well - I am seeing what damage I have done today.  You know Samantha (my hero) on Sex in the City had a face peel done.  It didnt look to great once she did it.  But eventually - she has skin that I would love to have again. Baby butt skin.  You know the kind.  I am really thinking about financing some beauty.  The whole shebang. Now- I am not gonna get a face lift or those horrible lips (Courtney Love, darling, stay away from the bee stung lips.  They don\'t suit you).  I am talking scrape the face - suck out that extra chin.  Lift those child riddled boobies. And suck enough fat out of the front and give me a JLO ass.  I know I have enough.  Dont tell me it cant be done.  I too - will have baby butt skin.  Expense, who cares.  I will be young.  But - does this mean that I will have to become my Gramma at the beach? Long sleeve everything and a big hat?  I guess it all has a price.  '),(13805,3792567,'06,July,2004','2004-07-06 00:00:00','   Jacmariesmom.  That\'s me.  Lost my real name 8 years ago.   How did I get that name?  Well, it speaks for itself and its a name I do wear proudly.  It comes from the little girls at the brownie troop meeting.  The ones that couldnt remember that I was \"Ms. Something...\" but they knew my daughter.  So - my name came from the seven year old girls wailing \"Jaaccmariees Mom\"!  The reason for my name is gone for a couple of weeks. I know, I hear the envy from other Single Moms out there.  But - she\'s been gone for 4 days.  I miss her.    Don\'t get me wrong.  We are a tough bunch of chicks to hang with.  We bicker and fight about everything constantly.  From the time we wake up until the time we go to bed.  Sometimes, in between there we giggle, snuggle and kiss.  I like the snuggle part best.  But - its too few and far between. Mainly - it\'s war.  Why you ask?  I know why - I would never admit it in a public forum.  But I know why.  And it kills me.  Thinking about it in my head causes my eyes to swell with tears.  But, its one of those things Moms must get over.  I will post daily if I can.  My rants, my gushing over something sweet or just daily poop. Please feel free to add whatever you\'d like.  I\'d welcome the company.  '),(13806,3792567,'17,August,2004','2004-08-17 00:00:00','   Ok - I am leaving to go to the Greater Northeast.  I will be visiting Boston, Cape Cod, Newport and New York.  All in one week on a whirlwind tour.  Not staying very long at any one place but apparently long enough to get the jist of what it\'s all about.  Many people when traveling in the above mentioned states - talk about shopping.  Its been a  frequently asked question since I have told people my travel plans.  My plans are..to eat.  I plan on having a lobster dinner on Boston Harbor - Clams in Cape Cod and Clam chowder in Newport.  I plan on eating a slice of NY pizza, tasting a bagel and eating pastrami at the Broadway Deli.  I am heading to a real NY Yankee game in Yankee stadium.  I plan on having a hotdog and a cold brewski.  I think I will also buy a hotdog from a street vendor.  Maybe a pretzel.  I also want a cheesecake for dessert.  I feel you should never go anywhere without a plan.   Talk to you all in a week!  '),(13807,3792567,'16,August,2004','2004-08-16 00:00:00','   Today is the 1st day of a (hopefully) positive journey in Miss J\'s life.  Today was the first day of a new school - new clothes and new friends.  She looked cute in her little uniform.  Freshly scrubbed face and new shoes.  Little bible in hand.  Shes very excited and is really looking forward to making some new friends. The cute buys don\'t hurt either.  Everyone cross your fingers and say a little prayer.  NeeNee is coming into town for a week Tuesday night.  Thankfully she is a morning person.  I was barely awake walking thru the dark house trying to find the kitchen in hopes of coffee!  Jacquis very excited and knows she\'s in good hands - they will be planning dinners and making schedules.  She\'s NeeNee\'s girl.  Two hours down and five more to go.  '),(13808,3792567,'13,August,2004','2004-08-13 00:00:00','   Ok - is it just me or are we still a little freaked out by Andrew circa 1992?   The weather channel and all the local stations are a constant mantra - please evacuate.  So, people are.  Then there are photos of long lines of cars on US1 tring to get to the mainland.  Or shall I say - long lines of cars on the 7 mile bridge.   I would much rather be in a house located in the Keys (most of them dated pre Andrew) than on a bridge when a hurricane comes a calling.  Is it just me?  Give me four concrete walls and land beneath my feet.  Somehow - I think I will be ok.  Give me a concrete platform with nothing but ocean below me with miles between me and the land.   What are you-  freakin\' crazy????  '),(13809,3792567,'12,August,2004','2004-08-12 00:00:00','   Ok - last night on BB5 - the most horrible man was voted off last night. Not one of the houseguests was sad to see him leave. Well, one but I think hes gay.  Nah, nah, nah, nah. Nah, nah, nah, nah. Hey, hey, hey, Goodbye!!!  So during one of the interviews - Nakomis (Jennifer) says the many reasons shes voting him off is that he pees on stuff.  Did I miss that episode?  '),(13810,3792567,'12,August,2004','2004-08-12 00:00:00','   Ok - BB5 fans.  Is tonight the night Jase is gonna go?  Has he been able talk his way off the block?  Has Marvin gotten a case of cockilockitis and rubbed people the wrong way?  Find out tonight - on Big Brother 5 CBS @ 8 EST.  '),(13811,3792567,'11,August,2004','2004-08-11 00:00:00','   Ok - on the J private school front.  Today was kid orientation.  This is where no parents are allowed.  Just new kids and team leaders.   What is a team leader?  A team leader is another kid - that is well respected in the school for some reason.  Be it sports or brainiac.  The other kids admire and listen to these few kids. The team leaders are the go-between between other students and the administration.  For example - if a team leader sees a fellow student do something not cool.  The team leader then will say something to the kid.  It is said to work better than the teacher telling them to do stuff.  Well, apparently not only are they respected.  But, from what I understand...very cute too.  So, needless to say they got Ms. J\'s attention rather quickly.  They were also very nice to her because her Uncle is the football coach.  Which never hurts when your the new kid.  So - she left the party in smiles and giggles of glee.   We get to go Friday night too.  WooHoo.  I have already been told that it is EXPECTED of me to look cute.  I am not allowed to attend the function if I am grubby in any way.  AS IF!!!  '),(13812,3792567,'11,August,2004','2004-08-11 00:00:00','   **Mom - if you didn\'t watch the episode last night - please pass this post**  Trading Spouses.  If you read NeeNee\'s post yesterday you understand the premise of the show and what has been occuring in the recent episode.  Well - last night was part two and the end of this families visit together.  Having seen the first episode with another family - I figured that this show would end like a storybook \"and they all lived happily ever after\".  This show tugged at my heartstrings and I felt a lump in my throat the size of California.  It was mainly due to this one mother named Lisa.  Now, Lisa wasn\'t a bad person.  She had feelings just like the next guy.  But her tough love views and demeanor just didnt go off well at all.  Her children (and husband) were very nice to their visiting Mom and they gladly welcomed the change.  I think they are going to miss the blonde visitor that made everything \"positive\" in their world.  They told the visiting mother they \'loved her\' as she left which in kidspeak is \'thank you\'.  Lisa\'s new family couldn\'t get her out of the house quick enough.  There were no warm words.  No goodbyes.   Now granted - no one made Lisa feel welcome or comfortable thru her stay.  Lisa was brutally honest with the children.  She was too honest for the cameras about the children.  Hurt feelings and bruised egos were left in her path.  So - the saying what comes around goes around was in full force during this episode.  Hopefully, Lisa watched this episode and learned that words hurt.   There was a scene where the girlfriends of her host family just tore her up.  I think that they were unfair and mean at times.  They made fun of her - and said horrible things.  At the end of the episode - when  Lisa returned home it only took her family 3 hours to walk away from her.  She was on a roll.  Lisa needs to learn from this if she can.  Regretfully, she will probably say...\"see how awful they were to little ol me?\"  The one good thing that came from this sad ending is hopefully that Lisa\'s kids will take that short week of \"positive\" and not let the \"negative\" bring them down.  '),(13813,3792567,'10,August,2004','2004-08-10 00:00:00','   BB5 was on last night. It was a nail biter til the very end. I have to say something about CBS. They have finally gotten it right. The cast. The rules. The editing. Its wonderful!!!  So - lets discuss.  I talked about the 6 finger plan. Remember when I said IF something could go wrong it will. Well - Jase wasn\'t as stupid as I thought and he figured it out. But - the genius part was that he didn\'t figure out the WHOLE plan as he was still trying to get the remaining horsemen to give him a shot. Is he crazy?!? Even if he is the best person to be up against in the end (hes a true ass) if you saved him this round your stock would fall in the den of females (oh and Will).  So Drew, who is having a little kissy face time with Diana (Ok - this is a couple that needs to be separated quickly. Along with those twins) was wondering if he should even use the POV at all. He figured Diana was safe. No one would vote for her. Its just a week earlier for Marvin - they will get Jase next time. She AGREED!!!  At this point I have my trusty pillow in hand - shredding it to bits. I keep getting up and pacing. Js saying \"Mommy your scaring me\".  They show Drew in the Diary room. Talking about lying to Jase. Jase did so much for him.... Blah, blah, blah. (Diana - Drew is a pussy. Drop him now before it\'s too late).  Everyone is counting on him. He seems so pained about a plan that was devised weeks ago. He doesnt realize that this decision is not up to him. This is something he must do. If he fails this mission both he and the lovely Diana will pay.  So - as the clock is winding down...he stands in front of the group. And....  Uses the Power of Veto.  JASE IS ON THE BLOCK!!!!!!!!!!!!!  '),(13814,3792567,'10,August,2004','2004-08-10 00:00:00','   Boy - I am a jabberjaw today.  I mentioned the AH in another post and I thought I would just share the distain I have for the AH.  The AH is a miserable little worm.  Whos (it thinks) tortured life sticks with it till this very day.  So it is the AH\'s mission to climb out of its little hole every few months and make everyone as miserable as it is.  The AH says mean and belitting things about everything  and everyone in its path.  Its the only way that the AH thrives.  To hurt feelings and cause little girls to cry.  The AH giggles at the pain it causes.  AH is a true masochist.  AH is very tricky - because it knows it can sweet talk anything.  Its what has gotten it thru life. So - if you are not as experienced dealing with the AH - you may be lured in.  Only to get vomited on by it\'s vile bile.  Be careful because that\'s how it feeds.  The AH truly doesnt bother me anymore.  I am very used to its ways.  I no longer can be controlled by its wrath.  I am lucky I escaped.  But my reward dealing with the AH has been a lovely.  To me, dealing with the AH has been worth it when I look at my prize.  The AH is breeding again and hopefully its spawn will not have to endure the mental torture the AH spews daily.  But it will happen again - the AH can\'t help himself.  '),(13815,3792567,'09,August,2004','2004-08-09 00:00:00','   Last night in my reality TV search - I came across a show called Family Plots. Its the reality show version of one of my FAVORITE shows on HBO...Six Feet Under. Some people might think that my TV watching is mindless. But - I do have to say that after the shows it make for good group discussions. What would you do? How would you handle it?  The first common thing we all stated was...what would make someone want to be a mortician? To surround yourself with sadness and heartache. Death and destruction. We have decided that most funeral homes are family and its just the call of order. This is what you do. This is your future.  And with that, on both the movie version and the reality version. These family\'s don\'t seem so happy. Death of children effect them daily. The loss of the much loved spouse and father when they too are trying to attain such adoration and failing miserably.  Then there is the discussions with the families on the ways to proceed with the burial/cremation or whatever.  This has always been a touchy subject with everyone as they don\'t want to think about their demise. But - I think its important to tell your family your wishes and HOPE that they follow them. It seems that that is a constant source of fighting during a most sad time. What to do with the body.  Some believe in burial. Or the Mos (so they aren\'t in the ground). Some want to be cremated others frozen for another day. There are burials at sea and some thrown into the wind.  On Six Feet Under - Nates wife died. Her family wanted her cremated and then placed in the family Mos. Nate knew his wife\'s desires and shared them with the family. They denied him and told him to proceed as they wished.  He filled a box with unknown ashes and gave them to the family. Nate then took Lisa (his wife) out to a beautiful field and underneath the tree he buried her.  I thought that was beautiful. Sad scene. Illegal in most states. But still beautiful.  I have always said I didn\'t want to be buried in a cemetery - though I find them to be very interesting places. Most of my vacations include a stop in a cemetery or two. But - because of my stops - I decided that as these people have come and gone. So have their families. Most headstones are old in disrepair. Weeds replace flowers. Neglect, neglect, neglect. Each one was a person with a very interesting story that has been buried with them. No one visits - except the occasional cemetery lookieloo (like me).  I deciced I wanted to be cremated. I then want everyone to dress in light colors and rent a pretty boat. I want them to serve shrimp and horse ovaries. I want everyone to drink and I want a great soundtrack coming thru the sound system. I then want everyone to kiss my urn and say something nice about me. Then plop me in the ocean and continue the party dancing to\"Brown Eyed Girl\".  So shall it be written - so shall it be done.  '),(13816,3792567,'09,August,2004','2004-08-09 00:00:00','   I have recently tricked the system and now have an ad for fart machines.  I don\'t need to buy a machine as I have two homegrown machines that compete with each other at the worse times.  Last night - J got a call from someone I will fittingly call the AH.  It never fails when the AH calls to start some drama.  Well, after she got off the phone with the AH I told her to come and sit on the couch with me and HB.  We watched Gotti  - then we ate cookies and milk (they always make you feel better).  After the show HB said he wanted to retire for the evening but J and I still wanted to watch TV.  So we all piled into the family bed.  Well - we are by no means a petite crowd - so needless to say it was a little cramped.  But J was laughing so that made it all worth it.  So while we were watching Airplane (a reality show that follows Southwest Employees around) the two of them start expelling gas.  Sadly - I am in the middle and trapped!!!  Its a dutch oven times 10!!!  The bed is shaking from giggles and the air unbreathable - there is no where to run.  There is also the knowledge that if you try to make a break for it - you will get pulled down by the booby-trap blanket that gets thrown over your head forcing you to smell the stank even more.  So - I don\'t need no stinkin\' fart machine.  I have my own!  '),(13817,3792567,'09,August,2004','2004-08-09 00:00:00','   Somehow anything with six fingers makes me nervous. It\'s not natural and its not right. So somewhere in my gut - I am scared with fear.  What is the 6 finger special? Its a weird and wacky plan that has been devised by Nakomis on Big Brother.  She is going the long way around nominating and voting off someone. It sounds good in theory but you know the saying...if something can go wrong, it will.  The power of veto (POV) ceremony will be on Tuesday night. This is where her plan is supposed to come together. Everyone trusting everyone else.  You know the saying - dont trust the guy with the 6 finger discount.  I think someone is lying.  Find out - Tuesday at 8 on C.B.S.  '),(13818,3792567,'06,August,2004','2004-08-06 00:00:00','   Some of my most favorite bands have taken to the streets and are having concerts to raise funds for the Democrats.  I am in a quandry about this - as I am a huge fan of two of them.  I know their songs are about equality - justice and choice.  I enjoy the music and have my favorite songs that I know word for word.   But as with some great actors - I really don\'t want to hear what you think.   Why is it that you feel that because your profession allows you to be placed in front of a microphone that I want to hear you speak about politics?  Why do think I would pay to attend a concert where not one nice thing will be said about the current administration and turn into a bashing of America.  For all the Dem\'s out there - let me say two words...Whoopi Goldberg.  Oh wait...how about, Alec Baldwin.  Lest not we forget about the...Dixie Chicks.  Top that with Susan Sarandon and her paramour Tim Robbins and an outsider would think America really stinks.  So I say to the Pearl Jams and the U2\'s (Hey your from Ireland, what you say doesnt count)  Shutup and sing!  '),(13819,3792567,'06,August,2004','2004-08-06 00:00:00','   I was listening to a news report the other day on the way home. It stated that some larger people were very upset because anorexia is becoming a common trend. But - there never seems to be any sympathy for anyone who has a obsessive eating disorder. Heavy people are usually labeled lazy and told to put down the fork. Seems like an easy order. In a world where most actors are a size 2 and struggling (not eating) with staying that weight for fear of losing roles. Why would the world want the general public to follow suit?  I have always been a big girl. I don\'t blame anyone for my weight. Most of it is genetic. Most of it can be tapered with watching the food intake mixed with some movement.  I am just tired of hearing about it all the time. I grew up with a Gramma who was obsessed with the weight thing. She mentions weight everytime she sees you. Also a Grandfather who pointed out that you were bigger since the last time he saw you.  To this day - I have friends that monitor everything I put in my mouth. They want to discuss the fat content and cholesterol level. Whenever HB is feeling fat - everyone needs to follow suit and eat the way he deems acceptable.  It becomes hard to put a stitch of food in your mouth when you are around these people. Because you don\'t want to hear it. It gets old - really fast.  Its a shame because I know a lot of people that have been bigger all their lives (me). We have come to accept it and actually relish in our beauty. We can look in the mirror and feel comfortable in our skin. Is that so wrong? Is it wrong to feel good about yourself - no matter what your size?!?  The majority of the females in the USA are a size 12-14. I know this to be a fact - because if you are that size - trying to find that size sometimes is a feat. There are plenty of 2 and 4s on the rack. Why - because most of America doesn\'t fit in that size. Actually I laugh at times. I think that even if I became skin and bones - I still wouldn\'t fit in that size.  Sadly - I came across this very subject this weekend with someone I hold very dear. Regretfully - she is too young to embrace her beauty and is feeling the same shame I did when I was younger. I have made it a rule for myself - that I would never ever mention the things that so bothered me when I was young to someone that young.  Like I said - bigger people look in the mirror. They know exactly what they are. Pointing it out doesn\'t make people change. You won\'t see someone leave the Doctors office and go buy a treadmill. It\'s only when the person it is effecting wants change.  I understand that most people say things out of concern - health reasons. But - if you don\'t know my medical history and bloodwork - then you need to butt out.  If it\'s effecting me because I cannot get out of bed, into the shower or walk 5 steps. That is a medical emergency. Then its your duty to get me some help.  We don\'t need a play by play everytime we choose to eat. All that does is make us want to eat alone. We eat when no one is around so we can enjoy our meal in peace. This never makes for healthy eating. Why must I eat standing in the kitchen alone when I could have eaten the same thing sitting at the table enjoying your company?  We are also not just a number on a scale. \"Gee, that\'s an awfully high number\" \"You have such a pretty face if only...\"  So when you come to visit my home - I am going to treat you like my father always did to us and company.  I will feed you till you burst. I will have enough drinks to go around. I will always have dessert. And the meal will be fun and enjoyable.  If you mention your diet or cholesterol while your there.  I will personally stick a bread roll in your mouth.  '),(13820,3792567,'06,August,2004','2004-08-06 00:00:00','   Another weekend passes and more money goes flying.  J starts school next week.  I thought I would get off lucky because of the uniform thing.  But - now I am investing in accessories.  New glasses - ching.  New shoes - ching.  Lets not forget her hair and nails - ching-ching. I guess in the scheme of things I did get off lucky but my wallet is feeling a little empty.  The big event starts next week - everyone cross thier fingers and toes.   I hope it all goes well.  '),(13821,3792567,'06,August,2004','2004-08-06 00:00:00','   Now thats more like it!!!!!  ROFL!!!!!  '),(13822,3792567,'05,August,2004','2004-08-05 00:00:00','   Big Brother was on last night. Scott is gone, gone,gone!!! He was as ass clown. He was mean and arrogant. I detested him.  He was actually shocked. \"I was nice to them this week. Why didn\'t they fall for it?\" Boo hoo hoo  Then - the Head of Household Competition. Jennifer (Nakomis) won!!! She\'s a little odd - but I think she will make the right decision and put up Jase and Marvin next week. Hopefully, Jase will not get the POV this week and then he can be voted out. He\'s the blonde Scott. Bleachhh.  Did anyone notice that when Jen won the HOH not one of the \"Horsemen\" got up and congratulated her?!?  Also - in a twist in the game. \"A\" Adria and her twin sister were introduced into the household. I don\'t know how that is all gonna work out.  It will be interesting indeed.  '),(13823,3792567,'05,August,2004','2004-08-05 00:00:00','   I like punch. The summer is here in full sizzle and I have become a pro at knock you on your ass punch. Its very tricky. It seems all sweet and wonderful and then POW!  HB is starting to go no carb crazy. So he \'suggests\' that maybe I need to switch to something not a sugary. I say bah-humbug to him.  Its the summer. I am enjoying the beach and the sun and the long days. Soon - it will be too cold to swim and the nights will come quicker.  Life is passing you by ol man. My auntie had a great saying on her blog yesterday. I am going to print it and share it with him.  I have started Yoga. I go 3-4 times a week. It\'s very hard. HB thinks that I lie there and meditate for an hour everytime I go. Its actually a series of poses and postures that take all your muscles to hold.  If you are not in shape - you twitch and shake like a crackhead.  I am only a semi-crackhead at this point - but I am getting better every day. I come home sore and shaky, take a shower, sit in front of the TV and want to go to sleep. So when the weekend comes I want to enjoy it. Thats where the magical punch comes in.  Beer bloats me and makes me burp. Wine gives me a headache at the beach. Vodka and cranberry is a sipping drink - nothing to fun about that.  Punch is the way to go. And - because its so fruity and juicy you can start drinking it early in the morning.  Hell - its happy-hour somewhere in the world.  '),(13824,3792567,'05,August,2004','2004-08-05 00:00:00','   Please bear with me.  Anything has to be better than those ads.  (I am not saying the word because apparently thats whats getting me in trouble.)  Starbucks - latte - coffee - and expresso.  Potty breaks and s___ breaks.  diapers and buckets  reality tv and real world.  Queer Eye.  Gotti.  Maids and gardeners and pet sitters.  poop and farts  dogs and cats  smelly boys.  Teenage daughters.  Toothpaste?  (I told you I am trying)  Nelle  '),(13825,3792567,'04,August,2004','2004-08-04 00:00:00','   So last night I am flipping the tv.  I come across a very interesting show.  Amish in the City.  It was on UPN.  Here\'s a little useful knowledge for you.  Amish do not go to school past 8th grade.   Annnnd - Amish are allowed out into the world for one year.  They then have the option of returning back to Amish life never to return to city life again.  Or - they can choose city life - never to return to Amish land again and be shunned by their families for the rest of their life. Amen.  This show is depicting just that.  Its like 5 Amish and 5 regular city folk living together in LA.  They are learning about each other.  The Amish don\'t know much about regular culture - but they know carpentry and God.  But they are learning how to live together and the regular city folk are showing the Amish kids a good time.  Its pretty neat to watch because the Amish kids are from very sheltered homes.  I think that so far watching it - that maybe 2 of the 5 will leave the Amish.  The other 3 - though they seem to be having fun learning new things really miss the Amish ways.  The next show was called Player.  Its set in South Beach.  So I had to look.  Its like 12 guys who think they are players trying to swoon this girl who claims she is a player.  Its kinda lame.  But - the ending phrase was kinda catchy.  Dont hate the player - hate the game.  Big Brothers on tonight.  There was a huge twist again on Tuesday - but there are two juicy players on the block tonight.  I think Scott should go.  No strategy - I just think he\'s an ass and he needs to put his shirt back on.  There is also a very another great twist that will occur tonight.  I can\'t wait!!!!  Oh - I need to mention... I heard this morning on the news that Growing up Gotti (mentioned it a few posts back)  was the highest rated and viewed reality show to date.   See - I know my TV!!!  '),(13826,3792567,'04,August,2004','2004-08-04 00:00:00','   I don\'t know if you know that my neighborhood is a HOTBED for terrorist.  I will give you one fact.  Sept. 11 I am driving home from out of town.  Drive into my neighborhood and the amount of police cars and tv crews that were blocking the street was startling.  Come to find out, they (the terrorists that hijacked the plane) lived in a little boarding house down the road.  They blend in.  Plus where I live there are so many transients and tourists.  No one every thinks twice.  Well- this morning in Albany the FBI raided a mosque.  The arrested two leaders of the mosque for trying to buy missles.  I live near a large mosque.  Once again - it blends in.   If you didnt know any better you would think it was a Middle-eastern restaurant ( I did - and pissed them off when I asked for  table for two).   Its in a strip mall in a middle class \'hood.  Set far off the road - you could pass it and never know it was there.  They all go to the 7-11 during prayer breaks.  It\'s a mosque - I swear.  I know that following Sept 11th. the police were all over that place too.  They cried foul - claiming racial profiling.  They said they knew the terrorists very well and they knew nothing of the plan.  I say...whatever.  I am telling you...my town will be on the news again.  I can bet you that.   I hope its not too late.  '),(13827,3792567,'04,August,2004','2004-08-04 00:00:00','   Ok - I have written to Google and told them to take off that stupid ad for non smoking aids at the top of my blog.  I am sure they will pay attention.  NOT!  So - I am opening up my blog to any advertising that I deem fit in my dark and skanky world.  My mothers blog ad is for swimsuits and my brothers is for Wonderwoman undies (Dont ask me but apparently they know something we don\'t).  I would like the following companies that I use in my everyday life to please come forward and I will hawk your wares.  This will be like infomercial central.  Have you seen what I have achieved with my following of Reality TV junkies!!!!   My international recognition on the wonderful wordfest called Blog!?!  Ok - well maybe thats a bad example.  But - pay me and they will come. (I think)  '),(13828,3792567,'03,August,2004','2004-08-03 00:00:00','   Apple martinis are yummy.  They are becoming one of my favorite things.  They fall in the fruit food group so I am having a well rounded meal.  I am glad I added them to my diet plan.  With the vodka - which is made from grain - I am having my veggies too.  There are so many things that may look like they arent good for you but if you look deeper and really think about it.  Almost anthing can work.  For example - Hotdogs.  People think that their bad for you.  But with a little creativity.  Hotdog - Meat Family Bun - Bread Family Sourkraut - Veggie Family Ketchup - Fruit family.  All wonderful goodness wrapped up into one healthy meal.  Hershey almond bars - add the nuts for the health nut in you.  Lets talk about icecream - milk and fruit.  Your getting your calcium and you won\'t get scurvy.  Pizza is a wholesome meal.  Pepperoni and sausauge - Pile it on.  Meat group.  The rest is dairy - veggies and bread.  They all work each other out.  One of the best meals is Pb&J - You have your bread, your nuts,your fruit.  If you drink it with a big glass of milk - there is your dairy.  Add a little chocolate syrup to your glass of dairy as a little treat because you deserve it.   Eating so healthy and all.  '),(13829,3792567,'03,August,2004','2004-08-03 00:00:00','   Regarding the convention I am attending in a few weeks.  I have just gotten word that one company will not be attending due to the terror warnings.  There will probably be more.  I expect it.  Don\'t people know that NY has been on yellow alert since 9/11?  Not only that - but I can say that I bet it\'s the safest place to be right now.  You see all the policemen out?  They are there to protect you!  The terrorist want you to be afraid.  They want people to show fear.  They get off on that stuff.  Soldiers in Iraq and all over the world are fighting so that you can feel free.  They are dying for your freedom.  Please don\'t let them down by cowering in your homes.  That means the terrorist won.  Look at what they did to our industry (aviation).  It effected us the most.  All of the companies and people who went out of business and or lost their jobs.  You should be spittin mad.   Make a stand and show them you are not afraid.  Go to NYC!!!!  '),(13830,3792567,'02,August,2004','2004-08-02 00:00:00','   New reality show last night folks.  Growing up Gotti. It was on A&E at 9:30. A day in the life of V. Gotti daughter of John Gotti (now deceased Mafia boss - oh I am sorry he was in plumbing) and her three boys.  Another one I would recommend just to see the train wreck of a family. This is one that I will keep watching because I am waiting for the cameracrew to get beat up and their stuff stolen just so the kids can fence it.  Seriously guys, check your gear.  I bet something valuble is missing.  These kids are a whole lotta trouble.  Check your papers in a few years.  Remember that name...Gotti.   Tonight is a special one.  Reality tv galore.  I love Tuesday TV.  Don\'t forget we have Trading Spouses on Fox (New family - tears shed - woo hoo).  Then see what happens at 9 on Big Brother (I see the times a changing. Last weeks twists and turns were wonderful).  The Amazing Race and Queer Eye and Jessica Simpson all compete at 10.  Get those tv clickers programed.  Thank goodness for commercials.  I never miss a show!!!  Oh - dont forget to try to catch \"Things I hate about you\" on Bravo @11.  I know its late - but if your up its worth peeking at.  '),(13831,3792567,'02,August,2004','2004-08-02 00:00:00','   ok - I think I am in lust with George W.  He winks at me everytime he talks to me on the TV.  Swoon.  '),(13832,3792567,'02,August,2004','2004-08-02 00:00:00','   Went on a road trip this weekend with two boys.  This is a little story I wrote to remember the trip.  Boys smell.  Boys fart and burp.  They eat your chips when your sleeping.  They push their seats all the way back and you have no leg room.  They want to borrow your pillow.  Your magazine.  Even though they made fun when you packed it.  Boys drive way too fast when it rains. Then yell at you to stop talking so they can concentrate.  They hog the radio and make you listen to sports.  Boys get mad when you have to stop and pee.  Boys smell.  '),(13833,3792567,'02,August,2004','2004-08-02 00:00:00','   I heading to NYC in the next couple of weeks. I have never been there and I am really looking forward to my trip.  I am staying in the middle of Times Square. In the same hotel the Republican delegates that are attending the RNC. I will get over my fear and head towards the highest building and head up. I will dance and taunt you in front of the NYSE. I will wave at the Statue of Liberty and I plan on taking the subway to go see a NY Yankees game.  Why? Because I am not scared of you!!!  You towel headed, boy-butt loving terrorists will not ruin my day.  But- how I wish I could ruin yours.  '),(13834,4015188,'25,July,2004','2004-07-25 00:00:00','   urlLink    This is Tenor_Man looking as goofy as possible......&nbsp; urlLink   '),(13835,4015188,'23,July,2004','2004-07-23 00:00:00','   We all have to start somewhere.   &nbsp; And of course, we also know where ultimately we are headed...well in a gerneral sense anyway.&nbsp; I have done my best to avoid this whole blog thing.&nbsp; The problem I have with it is that most people are boring, stupid, or both.&nbsp; I know that sounds arrogant but I am not putting myself on high.&nbsp; I am simply stating a fact.&nbsp; I am a singer.&nbsp; Specifically, I am a classically trained opera singer.&nbsp; A tenore.&nbsp; I try to avoid saying opera singer really, mostly due to&nbsp; the numerous associations people have and the pictures they get in their minds.&nbsp; Like I said most people are stupid and they tend to generalize (\" You\'re not fat\" ).&nbsp; I call myself a singing actor beacause I am an actor.&nbsp; I am an actor that happens to sing.&nbsp; As i said i am a tenor (to be a real opera-geek I am what is&nbsp;called a Lirico-Spinto tenor by Italians, and a Jungendlicher Heldentenor by the Germans....in boxing terms I\'m a middle weight).&nbsp; I live in NYC (Astoria, Queens) and am trying to get a performing career underway.&nbsp; I\'d like to make my living perfoming in opera, musical theater, theater, or film.&nbsp; I don\'t have to be a mega-star or the 10th fourth tenor or anything of the sort.&nbsp; I simply want to be able to do what I am on this earth to \"do\" (aside from being a Dad, Partner, Son, Brother, Recovering Addict, etc).&nbsp; I want to be able to create characters and go through my \"process\" and inspire people, to move them to consider love, hope, compassion, beatuy, and other higher human truths.&nbsp; I have a need to express my humanity and I have something to share.&nbsp; But it ain\'t easy.&nbsp; Thus this blog is called a Tourney (for those of you who do not know what that mean please look it up&nbsp;here:   urlLink  http://dictionary.reference.com/   .&nbsp;) And let me tell you this performing thing AIN\"T&nbsp;FREAKING&nbsp;EASY.&nbsp; This blog will be a a series of post about my progress or lack there of in trying to do this thing.&nbsp; This idea just came to me so I am not sure what my format will be exactly...it\'ll just be.&nbsp; I&nbsp;will stick mainly to the present or the very near future but I may also on occasion have a flashback or two(yes,&nbsp;THOSE kind as well).&nbsp; So have a look see, take a seat.....the show\'s about to start.&nbsp; Toi, toi, toi&nbsp;    &nbsp;!   '),(13836,3957657,'27,July,2004','2004-07-27 00:00:00','   I just saw a Michael Moore film, which I had been meaning to see since its release entitled \"Bowling for Columbine\" This documentary tries to find out why Americans are so...well....such assholes who kill each other. It all comes down to, nobody knows why....The Media keeps everyone in fear, and that doesn\'t help much,  I REALLY suggest you see this film, Absolutely amazing.  '),(13837,3957657,'26,July,2004','2004-07-26 00:00:00','   I bought a new hat today....It\'s pink....so what? A picture to come, but I\'m telling you....Pink hats > You.  '),(13838,3957657,'23,July,2004','2004-07-23 00:00:00','   OK, This list will be updated as I see fit, This is a list of movies you are required as a human being to watch.  * indicates that males are the only gender REQUIRED to see this film   Bowling for Columbine Casablanca Pulp Fiction Reservior Dogs Die Hard* The Matrix* Swordfish Punch Drunk Love Schindler\'s list Nightmare Before Christmas The Last Samurai Philadelphia The Green Mile Top Gun* Rainman Mission Impossible* Vanilla Sky When Harry met Sally 12 Monkies Bullet Proof Monk* Con Air* Alice in Wonderland (NOTE: This film has a much deeper meaning, Alice in Wonderland |the book at least| can be seen as a young girl who is using LSD or Acid. Wonderland in entirely in her head....She is on an acid trip) The Wizard of Oz    This list is subject to being updated whenever the fuck I feel like it. If you have not seen these movies...you are in danger of being made fun of by me. ESPECIALLY if your a guy who hasn\'t seen Die Hard....thats just gay....TOTALLY GAY  '),(13839,3957657,'19,July,2004','2004-07-19 00:00:00','   No updates recently...life drags on and on, as boring as ever. I bought Painkiller, it is fucking great, I don\'t get the FPS i want, but whenever I get my new videocard it should be good. Saw some girl at a store today with pink hair, made me miss MY pink hair :P BUT Taco Bell tends to frown upon pink hair.... I still need like 4 things before I can get that job, but I\'ve pretty much got it...I hope. I reformatted my computer, so it runs better, but I lost all of my music :( I feel like I\'ve done a good deed, I introduced someone to Atreyus wonderful music :) Hah, wow this is a boring boring boring post.....boring....WOAH WHAT A COOL DOG! be back later, I have to chase a dog down the street!  '),(13840,3957657,'19,July,2004','2004-07-19 00:00:00','   :) Well, at 3 I am going to Taco Bell for an interview. Sadly I have to get a job :-\\ Hopefully it goes well. My french Fries are burnt :( My oven sucks. And I forgot salt...  '),(13841,3957657,'18,July,2004','2004-07-18 00:00:00','   urlLink    Grrrrr, it\'s me.....the Vampire from the Dungeons and Dragons Monster Manual GRRRRRR!!!!!!!! Seriously though...isn\'t that dude cool as hell?&nbsp; urlLink  '),(13842,3957657,'17,July,2004','2004-07-17 00:00:00','   urlLink    Me as Gordan Freeman from half-life. Clifford \"BuZzSaW\" Rose sent it to me as a joke.&nbsp; urlLink  '),(13843,3957657,'17,July,2004','2004-07-17 00:00:00','   urlLink    A great desktop Background given to me by a friend of mine from the UK by the name of Kurt, or \"KeneticKangaroo\" Check out his Condition Zero team Played the game.&nbsp; urlLink  '),(13844,3957657,'17,July,2004','2004-07-17 00:00:00','   urlLink    Not sure where this picture came from, but thought it was worth a mention. See ladies, I\'m sensitive :)&nbsp; urlLink  '),(13845,3957657,'17,July,2004','2004-07-17 00:00:00','   urlLink    The \"Nurse\" Project after I was finished :)&nbsp; urlLink  '),(13846,3957657,'17,July,2004','2004-07-17 00:00:00','   urlLink    A Photoshop Coloring project I did.&nbsp; urlLink  '),(13847,3957657,'17,July,2004','2004-07-17 00:00:00','   urlLink    Samurai Kittens.....trust me...it is worth mentioning....&nbsp; urlLink  '),(13848,3957657,'17,July,2004','2004-07-17 00:00:00','   urlLink    A Banner a made that I am pretty proud of. &nbsp; urlLink  '),(13849,3957657,'17,July,2004','2004-07-17 00:00:00','   urlLink    Me with my uber cool pink hair and pwning with my deagle at CS :)&nbsp; urlLink  '),(13850,3957657,'17,July,2004','2004-07-17 00:00:00','   urlLink    Me again, caught off guard by my sister and her camera, I was walking home from school, the medical mask is something I did at school, running around screaming things, was pretty funny.&nbsp; urlLink  '),(13851,3957657,'17,July,2004','2004-07-17 00:00:00','   urlLink    This is a bad photo of me :-\\ But its one I just found. my sister took it as I came out of my room, this was before I moved out. In the background the the picture frame is me as a baby :)&nbsp; urlLink  '),(13852,3957657,'17,July,2004','2004-07-17 00:00:00','   I am watching southpark and eating cashews and M&Ms, and being the person I am, decided to mix the 2 in my hand and down them at the same time. I was interested to find how it would taste, and might I say, Cashew M&Ms would be lovely.  '),(13853,3957657,'17,July,2004','2004-07-17 00:00:00','   Well, I started my radio empire :) HaveN Radio with HaveN (Bryden) Ghostkodiak (Chavez) and RoninAfro (Jacob) can be heard every Friday via shoutcast. more to come, weeee.  '),(13854,3957657,'02,August,2004','2004-08-02 00:00:00','   urlLink    This is KIKIKIKIKIKI! I am kikikikikiki prenounced key-key-key-key-key-key (ALWAYS SAID 6 TIMES AND NO MORE!) I often add teh kikikikikiki onto my name in various things. IE. fL ^ HaveN teh kikikikikiki or sSquezz [!] hvn teh kikikikikiki kikikikikiki comes from lol, i made kik mean the came thing as lol, because they are directly to the left of the keys  lol, thus kik was born, rather than lololololololol i said kikikikikiki and I kept the word. :) THUS is the history of kikikikikiki&nbsp; urlLink  '),(13855,3957657,'02,August,2004','2004-08-02 00:00:00','   urlLink    WHAT KIKIKIKIKIKI IS! For the first time EVER I will annouce WTF KIKIKIKIKIKI is!&nbsp; urlLink  '),(13856,3957657,'02,August,2004','2004-08-02 00:00:00','   urlLink    WTF IS KIKIKIKIKIKI? GRRR!&nbsp; urlLink  '),(13857,3957657,'02,August,2004','2004-08-02 00:00:00','   urlLink    WTF IS KIKIKIKIKIKI?&nbsp; urlLink  '),(13858,3957657,'02,August,2004','2004-08-02 00:00:00','   urlLink    No it is not.&nbsp; urlLink  '),(13859,3601759,'27,June,2004','2004-06-27 00:00:00','   I need to remember.  Feeling my strength being sabotaged with doubt, I get really fed up with disgust and go outside and pedal on my bike until my muscles burn and my breaths are deep, drying my throat.    Drops of sweet drip from my body like tears  as I cry with my whole being  crying as I worship this creation  feeling the pace and control take over me  leaving me a passenger on the vessel.   I\'m seeing nothing  but noticing everything in detail  taking note of the balance and necessary breath  and I grip with ease  comfortable to be riding a machine  that carries me to another world with each passing mailbox.   Flicking gears  stressing my anterior then my posterior  long methodic strokes  feeling the size of my legs  the shape of each quad flex and then release  -pushing, pulling mile after mile  holding speed  approaching you, passing you.   My body isn\'t solid  but a porous entity that is atoms filled with space  air passing through it  and so I\'m connected and surrounded  one with existence  giving and taking from the air and earth.   I\'m touching the unseen  I\'m reaching out catching more  sweating and hungering for more.    The pleasure of the attachment  and yet separated  as my spirit soars far above you  it reaches the clouds   And the clouds rain down  dropping big tear drops  splashing the road  caught in my tires  spraying the back and my front  -rain soaking my clothes  trailing down my shaved legs  beading on my arms  collecting in my shoes  and now my tears of sweat  drip with the rain from God  and we cry together  we weep for closeness and eternity.    Together we will mix like water  and run a long river down a pass  cutting the rocks of tribulation and hatred  swelling the pond of love and forgiveness   The animals will drink us  and be renewed for the long frolic to being  -to existing as one.   '),(13860,3601759,'24,June,2004','2004-06-24 00:00:00','   I\'ll spray out words like paint onto the wall.  I\'ll fill up the space with thoughts, words lots of words flowing to the floor making meaning out of the nothingness that I feel when I live on the floor waking from a deep dream where god is my counselor, my psychologist telling me my problem is not my family.  He wants to be seen as unobstrusive and just a friend (\'I\'m a whittler\') like he has just formed me into a tiki statue from wood that illuminates the truth that resides in my heart.  If I could just grasp the infinite inside me the long lived and long loving core of who I am.  It is reaching out with hands -touching my core wanting to come out and be itself the true creation walking on concrete dodging bullets and seeing the clouds.  The sky rolling over head and the earth spinning around exposing all things to the star and the stars far away, like god feels at times.  His light shed on me from millions of years ago and the luster has faded into a dim memory of his face in a dream that tumbled around like myself in search of a lover and a beloved my true heart loving only one cherishing God\'s wisdom and fumbling disorderly chaos.  That two people should meet and become one, but for me it is more amazing that I have a purpose, something to be in this dreary mess of a world with air unbreathable, water undrinkable, food inedible -thoughts unthinkable.  Where do I belong?  In the grave right before I die I\'ll focus on the beat of my heart, slowing down and the rhythm and force will move my soul to focus on this emptiness that lives, my soul harnesses the faith of eternal life I\'ll know the heart will stop but the existence of my mind will not stop and I\'ll feel all new things and the flesh will not hassle me as I move on to the unthinkable and unknowable life after the encasing of the soul is removed what sights will be seen will it be on earth is space a new place?  No eye has seen or ear has heard of the world made to honor the soul. But life is not a pure thing or an absolute -there are variances to good and bad and the senses are weak it can\'t see.  So I \'see\' differently I would know more clearly. My cries of the soul that can\'t be distracted by the flashing brilliance of the world the beauty only skin deep will not faze me anymore I will be picky for the real beauty that is in a person in touch with their soul and consciousness grappling to stay here and wanting to move on full of health and life and understanding of the imprints of the thoughts and the inconsistency of man and his acts when justice is not served the ability to be totally unamazed that black is judged as white and one is locked away unfairly.  So when it happens to me I won\'t be ashamed for myself I\'ll be ashamed for those who act on external knowing and cares -not connected to their soul- their god -haphazardly moving from one sale to the other not thinking about why they have to drive cars and sit in traffic working in order to drive to work unsatisfied with the purpose of making widgets -collecting information, harboring information, organizing information, injecting information into their veins getting totally high off of the rush of words and meshing the names into a force that blasts apart all inner voices and guidance for we fear the long silence and the missed opportunity to buy cheap while caring not at all that we crack the whips at the backs of people with out any connection to us for could we look them in the eye at all and say thank you for green cashmere sweaters or tight ass blue jeans that make me look so hot that girls will want to make out with me because I only fear not getting laid and when I\'m old Viagra will ensure my virility but lower my love of a person for whom I have sex with.  But that little kid with broken hands and no free Lasik surgery for the myopia eyes originating from the American diet for which we pass around the world like the plate of cheese at our socials that take up the space in our empty lives of mass consumption and broken hearts do not mend over warm cups of coffee they only mend when they say I\'m sorry for leaving my petroleum in the yard, the chemicals in the air, for your body and my painted food, and to prove my contrition I\'ll dismantle the car and find another way to get around and I\'ll grow my own food and while I\'m at it I\'ll find purpose and have useful energy and make funny friends and I\'ll be a real person loving the hand that feeds me and giving more than money to the people that need help.  Racing to the top will be seen as running away from the earth and the love it gives to each and every soul unconditionally.  Love from the earth is unconditional not choosing to love just the beautiful, not having pity on the weak, just the same giving wisdom to people who hid the legacy from one another and place the love that isn\'t in a book on a tree averting our attention to the sin and rope of the unloved and corrupt and you realize that the dominos fell a long time ago and when it is on a roll you will fall too but the fall is more than just your purpose and the inner life will cry out and drive you mad and in death there will be silence for the dark souls and they will come back and the selfless and you too goody-two-shoe once wronged a man and broke a child and so you are back to do good.  We have all killed ourselves and so we honor those lives we meet by giving our own deep heart and not being closed minded and wretched.   '),(13861,3601759,'18,June,2004','2004-06-18 00:00:00','   Hey Megan  When it rains here people are frustrated and relieved.  The area hasn\'t been getting enough rain so everything, mostly the lakes, are drying up.  There are water restrictions allowing people to only water their lawns on Wednesday and Saturday.  So of course everyone is praying for more rain, but what a bitch that it rains on Wednesday or Saturday because that was the day that  I  was supposed to water  my  yard.  And there\'s the rub Megan, for people who are in such a routine, or the routine is their meaning in life, that even when something good happens, something painfully necessary, they don\'t want it if its not at the right time.  My life has been improving.  I\'ve been working on rising out of my funk, my work is constantly running into roadblocks.  My summer consisted of school (anatomy/physiology) and two jobs, Personal Training and Pet Sitting.  My roommate Gene owns a business in Pet Sitting, www.vippetsitting.com, that I started working for in July for extra cash.  Its ok, really, I like seeing pets throughout the day.  I guess its good animal energy in my life.  About this same time, Gene\'s girlfriend moved in, which basically means one thing, they will get married.  Shawn, is 35 and Gene is 34, so its that age for people to get hitched.  Surprisingly, they wanted me to stick around and make me part of the house.  Gene truly appreciates me and I care much about him.  There is this ridiculous humor about him that cracks me up and gives me the giggles.  I think he likes having someone around to joke with.  Shawn is starting to like me, perhaps because of my sensitive side.  I tell her how I like what she has done with the house (very \"girlish\" but this place is shaping into a home).  I have met both sets of parents and I think they approve of me too, which is surprising because I feel I have some shy tendencies that could be affront to someone else.  I honestly stay very busy and your letter made me think about how I don\'t like that .  For example -what am I reading and thinking about?  I read science stuff (school related) and nutrition.  I think about what it will take in order for  me to pay my bills and keep me from stressing about being poor.  Your letter inspired me to exam my life a little.  So for kicks I picked up a book that wasn\'t so grounded in the here and now, in the present senses of sight, but more emotional in text, like Of Human Bondage, by W. Somerset Maugham.  It is very good.  I like these books, but the simple pleasures are easily forgotten when life is turning at a mile a minute.  Just so you know, there are times when you won\'t read, or write, or ponder or grow in your scholastic way, and its ok.   When you feel too smart for your own good, remember that there are people who run labs and heal people.  But you megan can do whatever you set your mind to, of course if the path is blocked, that will be a tad difficult to travel.   Anyway, some thoughts of mine from your letter.  I have never tried to make a point or conclude that I became a Christian because of you.  I have thought, though, I stopped being a Christian for a girl, but that isn\'t a real answer and unfair to say all that I believed and cherished -everything I did so fervently- could be toppled by one girl.  When the truth is that many factors played on that change in me.  Your brother Jason fascinates me.  Even more, the two of you talking fascinates me.  I wonder what you ask him and what you say when his reply is so different (or the same) to what you live around.  You seem frustrated, is this the case?  Let me tell you that a boy can get terribly wounded.  My insecurity has been growing lately.  I realize that I am struggling only when that girl Kim calls.  Nothing going on, but when I was such a good friend to her and now it is gone, I realize an empty spot.  Why is it so hard to strike up a friendship with a girl?  It is the only way something powerful, meaningful will ever develop.  My grandma says I\'m just insecure, unconfident and that I\'m letting some family thing get to me.  But I meet girls or attract girls easily.  However, I feel they just want me for my looks or body or whatever.  I don\'t know of a decent \'girl\' that wants to get to know me that I also think I can trust.  So your question about the \'Christian girl\', I haven\'t seen her much, but its not worth the effort when I know how I can and want to be.  I\'m still growing up and figuring things out, but as one friend told me, guys don\'t realize that the girl gets a piece of you too.  So I can flirt and be very confident about me, but at the end of the day I go home and plan the next day and do the things to get me out of this position I\'m in, so when I\'m 26, 27 I won\'t feel behind the 8-ball and can take care of a woman that I can be friends with and grow with.  --What a feel good thought all that is, but what is my reality, megan?  Reality is that I have 20, 30, 40, 50 year old women that would love to take a bite out of me, but for better judgment they don\'t.  It\'s fun to flirt, but you know what, I wish they would.  That is a pathetic but true feeling I have.  However short term, I wish these ladies I know and talk to would make a move, but they won\'t because I won\'t also egg that crap on.  I wish my world was filled with just friends, but sex is hanging over me and I miss it.    Maybe you know what sex is like, maybe not.  I don\'t assume anymore because I have seen married women come on to me, Christian girls -whatever- nothing is as it seems.  Generalizations are not good ways to think.  There are no blanket statements.  You can\'t be solved by me just because of your family, school, friends, age, sex, religion - you may not be able to solve yourself.  You wake up and think about doing something that yesterday you swore you wouldn\'t think of again.  But life will rain on you when its \'time\' to water your yard and you will have to do something else when it rains on your parade, or else I\'ll come by your home, see the sprinkler system on and take you as a joke.   '),(13862,3601759,'15,June,2004','2004-06-15 00:00:00','   DAD  I donâ€™t need you  To help me feel alive The man inside comes out By his own rite You donâ€™t see how The love is gone You donâ€™t pretend To want to know  I cry over you  Or at least I did You never gave yourself  A chance to feel me When did you reach out to me?   Why would have I seen It? How could I tell?  What makes the heart break  makes the pain sting, inside And pushes the tears out  But I put myself  In a vulnerable place At the disposal of Your deserted emotions I would find you â€“come to you Hoping you would look for me And I was satisfied by Finding you; seeing you But I donâ€™t know why  It never seemed important Enough to you Because now it doesnâ€™t seem Important enough to me I wonâ€™t be strung along Or hurt by my own expectations  of what you should be.  Do what you need to do. Iâ€™ll accept that little you have  and give in like manner  no longer hurting myself wishing you would give more.   '),(13863,3601759,'13,June,2004','2004-06-13 00:00:00','   In the beginning of the trip, as the sun began to appear over the pines and maple trees, Fletcherâ€™s emotions were about as mixed up as his favorite pancake matter, complete with extra wheat germ.  He had just said goodbye to his old client JB and for the first time during the days preceding his move he had cried.  Jonathan was with his partner, M, and there excitement and love was just too much to look at and not realize what was happening; he was leaving.  Jonathan made him some eggs and toast with a tall class of juice and Fletcher found himself saying out loud, â€œIs this happening?â€  The experience was a dream.  Here he was having breakfast made by J, like many times before, but this event would not be happening again anytime soon.  Within a day Fletcher would be about 800 miles and 7 states away.  But that was the point of leaving; He would be away from  there .  So what was there that scared the hell out of Fletcher?  Well, hell, in the theoretical sense was not a problem for him anymore.  He had given up that scare tactic and idea years ago.  No, what was happening in Durham was a mixture of an embrace from a straight jacket and the tiling before the planting of a seed.  Or perhaps it was best thought of as the parable from jesus and the seed that fell on rocky soil.  Fletcher was ready to grow, but if he was unleashed here, he would sprout with no root system and squelch in the sun.    Less than a year before, the idea occurred to Fletcher to go somewhere else.  Around September, when money became about as hard to find as a Nun in full garb in downtown Carrboro (except perhaps on Halloween) Fletcher had feelings things for him, a life for him was not to be found in NC.  So he thought a lot about going home, home to Indiana.  He even visited in late September, mostly for the chance to see his good friend ZW and some other friends at BSâ€™s wedding to BB.  But when he was there he thought about moving back.  He asked questions about the job market and even pinned down a promise with Bob that he would have a job if he moved back.  The town seemed to have relaxed like a pillow to cushion the fall that Fletcher knew was about to occur back in NC.  His landlord Gene was going to have to give him sometime to pay rent so he could find another job.  The strain in the house only grew with the extension of grace from Gene on rent.  Not from Gene, Fletcher never really felt it from him as a source, but from Shaun.  This mid-30s couple was just hell-bent on getting married and Fletcher knew that his welcome was wearing out.   The talks of moving out, the day to move out, was pushed further and further away from the present moment of absolute emptiness in Fletcherâ€™s life.  He was trying to find solace in a new acquired friend, Dave.  He was also from Chapel Hill, graduated from the school and so Fletcher had an immediate buddy.  In a sad way, Fletcher didnâ€™t know much of anything about Chapel Hill and the bars on Franklin Street.  He had barely even been on the street to do any drinking back during his recently passed college years.  Well, he had gone and seen a movie at the Varsity once.  So David was a way for Fletcher to learn Franklin Street.  Perhaps there was something he was missing there, and once he found it, all the confusion of the Dark Times, the poor times, the eating eggs as the only source of protein times, would somehow just change, like the wind blowing over a leaf reveling the concealed shiny side.  The nights became really long during those times with David.  He would return home just dizzy drunk, or just crash out on Davidâ€™s couch.  What was happening?  Just drinking and looking for women.  Neither one had any luck, at least in Chapel Hill.  So when David couldnâ€™t deliver or be a conspirator to Fletcher picking up a girl, Fletcher called his brothers and went to Raleigh.   And once he got there, He would take some shots of Tequila in his car and then go meet his brothers at the bar, usually Bogartâ€™s.  Being in Raleigh was like walking up to the plate and the ball from the pitcher is moving slower and looks bigger and on every swing the ball hits the sweet spot.  Fletcher was in the zone.  Every time he went out, he would find some girl to kiss or get a number.  He would go out by himself and pick someone up without even doing anything but standing around watching Sportscenter.  The seduction of women and the allure he had on them was becoming a real powerful realization in his personality.  At times he was seeking something innocent, a girl to fall in love with perhaps.  But if the girl he meet lived in Raleigh then there was not hope for something to materialize.  The one girl that could have been a potential candidate couldnâ€™t ever return a call on time making Fletcher assume the worst.  Without fail for about 2 months she would call on the last possible day before Fletcher dumped her to the recesses or waste heap of girls that I once met in Raleigh.  On a last ditch effort, Fletcher accepted her call on New Yearâ€™s eve.  He was still sad, just moping around about a recent girl that broke his little bit of fragile hope in an innocent girl and love.  They went out, drank, kissed, got naked and then Fletcher pulled away. She was not happy but Fletcher was beginning to sense something:  he didnâ€™t want a commitment.   In October Fletcher found a third job, and the pet sitting job really picked up around the winter holidays, so much so that Fletcher paid Gene off of all bills and moved into his own apartment a week before Christmas.  He began to settle into the idea that he would be here for a while.  The realization came to him that he really loved nutrition and that he could get his RD status from Central, the historically black college, within 2 Â½ years.  With the idea of being grounded in Durham for that much longer, the entertained thoughts of love should have arisen, but they didnâ€™t.  With three jobs taking up 12-12 Â½ hours of his day, there was no time to give to anyone, but himself.   The need for a girl grew and manifested itself with a perspective of him as a toy for them.  That pull he had began to be used for that need, which was a need for touch, or affection.  However he went through dry times, dark times of just working his ass off and hating his life.  An accident occurred and he felt that push from the semi pushing his car and crunching him into the median.  Sitting in the car waiting for the police sentence, cold, wet and broken, his spirit began to feel the push in a different way.  He was not crushed, or dented, but he yielded to the push and the force was strong enough to push him away from there.  It was no accident that the next day he had found Bastyr and Seattle.  With a new place live, the vigor returned.  He visited the school and town within the month and within a day of being there, he knew the decision was made.  It didnâ€™t hurt that he also got laid while he was there, which broke a 6 month streak.  He didnâ€™t expect to even talk to that girl again, and that was after the first night they met.  But he called her the next day.  They hung out, shopped a little, had dinner and returned to his hotel room.  In the background was Moon Struck; in the bed was a slow test of who wanted to do what?  Would it mean anything?  Who had done this before?  The questions were answered quickly and once the voice lingered off into the background, the bodies of passion finished the questions and pushed it further toward a completion.    Why is it that people think they donâ€™t act like animals?  The moment a decision to leave Durham was made, girls were not hard to find.  Perhaps there was this pheromone released that other girls could pickup on every time he went out.  The first weekend back, Fletcher is out with his brothers, Jon and Bryan.  There are some other girls out, but one in particular was just gorgeous to Fletcher.  This was Jonâ€™s dance partner from last year (and they are professional dancers) but she was warming up to him.  Of course he could turn her as the night wore on. The bros each had someone to talk to, as naturally as the coupleâ€™s paired off, they would end up togetherâ€¦and so they did in her car.  Perhaps there were really only three significant things about this girl to remember.  One she was 38, and as such the oldest person Fletcher had kissed sexually.  Two, she played J. Lo music and that just seemed like a wonderful touch to Fletcher.  Three, she had fake breasts (no way a 38 year old had breasts that firm).  What was wrong about this girl was the refusal on her part to let him crash on the floor, at least.  So driving home at 3:00 am the inevitable happened, Fletcher was falling asleep at the wheel.  But, his cell phone rang.  He answered it but it was broken up and cut short.  A couple of minutes later it rang again and this time it was her, the girl from Seattle, Kelly.  And she does what was already agreed upon not to do, ask details.  So like the tired semi-drunk guy he was, Fletcher said he kissed on some girl and her response was â€œOh, thatâ€™s just great for you!â€    She was there, he was still here.   The feeling would surface later during their phone conversations about her significance.  Was she special?  Was it just a moment?  At work when his coworkers would pick on him and the eventual marriage to the girl he became really confused and then scared.  Scared of being tied down and repressed in a new big city, a city with a lot more girls than Durham, Raleigh, and Chapel Hill put together.  But if she was, wouldnâ€™t that be nice?  Well, no it wasnâ€™t worth chancing.  Fletcher made it clear that any future happenings between them, physical or practical (or arenâ€™t they the same?) would be as friends and nothing more.  No thoughts of possibilities, just the moment.     The â€œmomentâ€ was that elusive energy that arose from the connection of two people and their desire for it.  It would lift you up and hold you like water does the buoy.  And Fletcher had a way of creating the dark matter, or dark energy that would intoxicate a proper recipient.  His appeal was more apparent when in one night he kissed 2 women (one married) and got 3 numbers.  One of the girls came from Penn and lived in Greensboro.  She was very pretty, but lived to far away.  And besides, she was still looking for something innocent; Fletcher had moved on to a new realm.  He had become sure of himself and all his glory.  What was needed was a moment and by definition that meant there had to be someone else, and that they wanted it too.  When you look at a happy couple of 30-40 years and condense the love of that you get the moment.  He couldnâ€™t handle the idea of a stretched out love over years because there are no promises that after 25 years she wonâ€™t leave you with nothing but half the shit.  It became easier to just take the â€˜pillâ€™ form of a long content life.  And the dose came in unsatisfactory flavors.  The girls that appeared after that night of charm were ill fitting stand-ins for what Fletcher really wanted.  He knew the facts of the situation and the girls did too:  He wasnâ€™t going to be around much longer.  So there was no other thing to do but enjoy the present.  There was no moment with those girls.  He was with them for that time, and that was all.  Fletcher was turning a corner soon but it seemed that he had already been around this block because he ran into an opportunity, an old opportunity.  When the details became clearer about departure from Durham and need of arrival in Seattle, Fletcher realized he had sometime.  So he figured he would quit his jobs early, take a little bit of time getting across the country and hang out with Kelly, since there were lines drawn (though he was unsure if they would be crossed).   The time getting across the country would be taken up by his family in Indiana that was until he turned the corner.  Her name was Megan.  Years ago the name conveyed a girl he loved that he couldnâ€™t have.  A Holy Grail.  Now she was mature, a free thinker and needing some comfort.  Fletcher was a comfort for her, and to his half-surprise, she was a comfort to him.  Well, she asked him to visit her but he declined, said there wasnâ€™t enough time and it was out of the way.  They talked about it more and she wrote him an email that changed his plans.  Without any love in his life, no home, or job, and a desire to see the little girl that caused so much grief and curiosity in his life, Fletcher decided to visit Megan.  He was honest to her â€œI want to kiss youâ€, but he was nervous about what he would find out by being near her again.  There were questions about faith she was dealing with; ones that he dealt with that froze him because he didnâ€™t want to say anything that might ruin the time with her.  All he wanted was to enjoy the time with her.  There was never a time before like the time that Boston afforded.  She lived on her own and was enjoying her own friends and ideas.  They would get to drink together, cuss together, smoke together and these were certainties.  But Fletcher wondered about more, like sexually things.  He imagined her having an orgasm as he went down on her.  Would she touch him?  The daydreams seemed silly after awhile and it became more realistic to wonder if it would even feel right to see her again.  And if they kissed would it be a â€œkissing my sisterâ€ emotionless thing.  The possibility of nothing happening was a huge fear.  He would rather wonder the rest of his life that there was something magical between them than find out that there was more solidness in the clouds that portrayed dreams.    The timing was off.  On the day of his arrival to Boston, Fletcher was 15 minutes early and had to call Megan home.  In the little spot of Beverly, she was only a few blocks away and called him to meet her at the library.  With the directions to memory he walked that way and then had a call, from her.  â€œHey, are you wearing a black shirt?â€  â€œYesâ€¦where are yâ€¦.â€   She was just ahead on the sidewalk.  Her hair was the first thing he noticed.  It was limp and straight on the sides of her face.  Her figure and gait was as he remembered, so he was not taken aback by any weight gain.  No matter what, it fit her, and on the hug, no doubt it was Megan.  An afternoon ensued of driving around Boston, getting lunch and just checking out sites.  It was a different time and place, but familiar feelings.  Fletcher, who had come there rather reserved and patient felt something for Megan.  In the evening, as her house was filling up with people, Fletcher made a break through that he could hang out with her Christian friends because they could laugh about the same stuff. Megan and Fletcher went to get a movie and toilet paper.  In the grocery store Fletcher walked slower, behind Megan.  She wanted him to walk by her side, but he just dragged along as a feeling was building up in him.  He clutched her tightly after picking up the Charmin and said â€œoh Nothingâ€ to her inquiry of â€œWhat?â€   In that second, it was clear that Fletcher wanted this time with Megan to be a moment and he assumed, before they made it to the car, that she did too.  So before he went stark raving mad, and they made it back to the house, he desperately said â€œWill you just kiss me, I canâ€™t stand it anymore.â€  In the parking lot of some school with other people mulling about, they kissed.  The first time since 1997 and unlike that last kiss, this kiss was stirring.  The stage was set for an extended moment between two friends.    The next day a group of Meganâ€™s friends left, three girls that had been sleeping in her room. Fletcher slept on the couch next to a couch that had a guy snoring so loud he couldnâ€™t sleep.  Once the girls left though, Meganâ€™s room was empty.  Without communicating, Fletcher went into her room, sat on the bed, looked at the books on her shelf and waited for her to return.  She entered the room and after rummaging around crawled under the covers and closed her eyes.  There was a big lack of sleep the night before, but Fletcher understood as clearly as she did that this was what both wanted to experience on this visit.  How would he feel in my arms?  Would she be comfortable with me, would she embrace me?  Reasonable enough questions for both of these kids to ask about an old boy/girlfriend.  However, it was the answers that threw things for a loop.  The plan for that Sunday was to see Harvard and Boston.  They were not in any hurry.  Being in bed with Megan was like nothing he had experienced.  It had little to do with the physical as it did with the emotional.  Fletcher felt more passion from Megan than anyone he had been with before.  There was no guilt.  Neither one had to look over their shoulders.  Perhaps he was fooled, but he sensed something deep from her rising up and clutching him to her body.  He even considered it being spiritual.  Their kissing was riveting and they would linger to each otherâ€™s mouths wanting to touch and taste more of each other, drawing in long breaths, drinking it in.  They were present in the moment and then went to Harvard.  With the apparent first hurdle cleared, Fletcher thought that the trip could be no better.  He felt the love, knew that she was good and the next morning he would leave and live his life without a change.  But in the Boston public, something happened.  They always got along well.  She would make gestures or hand him things and without thinking he would take them.  They cooperated well.  But there is nothing like the presence of a stranger to validate the obvious that neither wants to realize.  So they get on the subway on their way to Harvard and both notice a guy eyeballing them both with a look of surprise: he was mesmerized.  On the way up the stairs out of the tunnel the comment was hinted at, but Megan knew that Fletcher thought they looked great together, and that strange dude made it clear they did too.  They walked around Harvard talked about Meganâ€™s future and possibility of attending school there and went downtown.  Downtown was packed with people.  It was a beautiful day and people were everywhere.  They bought coffee and shared some thoughts and met up with Meganâ€™s roommate from Gordon, Karen.  The introduction was priceless.   â€œFletcher, Karen, Karen, Fletcher.â€  To Karen â€œThat pretty much says it allâ€ or was it â€œSpeaks for itselfâ€ Either way Fletcher blushed but appreciated the vocal complement.   They walked and talked for the next 2, 3, 4 hours.  Took some memorable pictures and shared a bottle of wine at dinner in the Northend.  Karen left, after Fletcher shared some feelings about Megan in a very candid conversation.  He thought he showed too many of his cards.  Over a couple of draft beers, Fletcher began to tell a story about his transition from faith in Jesus Christ to his present condition of being â€˜spiritualâ€™.  Megan smoked a cigarette, Fletcher took a couple of puffs and the story was short, but it contained a huge revelation, something Fletcher himself didnâ€™t realize until then.  Megan was asking questions of faith and morals; not too different from what Fletcher did two years ago.  With his perceptive eye the idea came to that perhaps she would fall like he did, and if so he wanted to catch her, be there for her.  And he said to her that he wanted her to fall, because he wanted her.  He wanted her like him because he couldnâ€™t be with someone so closed minded as the typical Christian was.  She showed promise though and he wanted to believe that perhaps they could be together, even if they really couldnâ€™t.  Fletcher drove home.  It was cute to him seeing Megan a little drunk.  He wanted to kiss her mouth and taste the smoke on her breath.  He wanted to touch her, but she retracted.  He didnâ€™t push it and felt fine just feeling her pull his head to her so she could kiss him.  Fletcher massaged her hands and shoulders.  He relaxed her and he was calm feeling the warmth of her body.  That night they slept together for the first time.  The bed was just a twin and way too small, but sleep wasnâ€™t all that important.  Fletcher decided to just hangout the next day as Megan went to work.  He chatted with her roommates and friends that passed through her home like a bank with a revolving door.  And like a bank all came to deposit something or withdraw something.  Sometimes it was emotional and sometimes it was for food.  They even had a landlord that took a weed eater to the yard, since a lawnmower wasnâ€™t necessary for an 8 x 8 yard with missed placed weeds as â€˜grasses.  Karen came over later that afternoon to take Fletcher to the mall, get some food and ask him questions about why he was there.  Did he come just to kiss Megan?  What would be the point of that?  He could kiss any girl, but it was Megan after all, and to him that was special.  But he came to see Boston, to talk to Megan face to face, show that he cared and if it could be possible, enjoy the moment.  Karen and Fletcher met up with Megan and a guy working with Megan name Jon.  Jon was funny.  There was just no way that this guy could be serious and usually he wasnâ€™t.  He was an actor and a play writer.  The ideas he had challenged the norm in the Christian world, and he liked being the antagonist.  He showed Fletcher a lot of promise and he told him so.  Megan dropped the guy off, picked up a housemate and they went home.  There was only one night left and what were they going to do?  They went to an ABC store to buy liquor, a bottle of wine for David the housemate and a pint of Amoretto for herself, â€œits candy alcohol.â€  Fletcher was mesmerized by the new traits and actions that Megan had.  They talked over pizza at dinner and when she rose to get a cup to drink the Amoretto with before they went home, Fletcher caught a glance of sexiness.  She pivoted on her right foot and when her butt turned, Fletcher knew then he wanted to take her home.  It was getting dark and they drove to a park and, well, parked.  They drank Amoretto and kissed while listening to music.  She was good, thinking, lowering her driverâ€™s seat all the way down making it easier for him to float above her, kissing her yummy lips and he found his hand going down, under her pants but she stopped him.  â€œYou love me for itâ€ and In a way he did like it that she stopped him, but he wanted her.  Her smell and the way she communicated her wants, even if they where to stop him just excited him.  He hated guessing what a girl wanted from him or for herself, but Megan knew what she wanted.    She was beginning to sense the moment more and more as they settled in her bed.  She read a story that was good, but being talked to sleep was one of Fletcherâ€™s ultimate weaknesses and so he fell asleep.  Coming to after the story Fletcher knew he had let her down but she enjoyed herself by acting annoyed so he would cling and touch her skin, pressing through to her muscles and moving her to a place that she hadnâ€™t felt for years.  They fought the sleep, again and again returning to each otherâ€™s lips, intoxicating each other until sleep came late in the night.  In the morning, before the alarm clock went off they laid by each other, Fletcherâ€™s head around her shoulder, perfect place to whisper drops of affection.  â€œI have a song in my head.â€   â€œBack to you, it always comes around, Back to you.  There is something about you, itâ€™s just the way you move, the way you move me.â€  But rather than being silent, like he had crossed a line with his sentimental voice, she came back with, â€œYou are fun to touchâ€ and it was true.  Her touch was passing energy through him like a massage.  And he began to melt and wish he didnâ€™t have to leave, though the need to go was so important to him and his future.  Karen had made a comment to that, â€œyou never know stranger things have happened.â€  What would they do?  How crazy to wonder about it; them being together?  In less than a week they would be 3000 miles apart, what the hell was the point of getting all emotional?  But there they were looking into each otherâ€™s eyes before they said goodbye, seeing oceans.   '),(13864,3601759,'11,June,2004','2004-06-11 00:00:00','   Over Xmas I became extremely fixated on F.K. (again, this fixation has happened before).  I woke up in the night from a dream about possibly being able to reach out to her and bring her back into my life.  Like we had something great back in the dorm Ehringhaus -so my memory had me enthralled into believing and remembering. But I was enraptured into planning on how I would try and get in touch with her once I got back home.  Inevitably I did call.  The first call was something of a hope that the weight of the world would be removed from my shoulders.  Knowing that school wasn\'t in session I dialed her number and just like I thought I got the answering machine, but my reason for calling was not left on the answering machine.  I never leave messages on answering machines anyway, except when I do my job at the gym.  So after that first call I thought this pursuit for a long lost time and event- not necessarily the girl Far--- as much as for the chance to rectify my standing that had probably long crumbled, would vapor away, but it remained strong like a vice and I had to call her repeatedly.  I thought that she or H------ would freak out once they saw their caller id, with about five calls from a \"gene c------  489-****\", but on a Friday, midmorning, roads too bad to drive on from snowfall, I got in touch with Far--- for the 1st time since August.  All the imagined conversations I had already had with her couldn\'t match the one I did have with her.  Once I identified myself, I was hit with a \"What\'s new?\"  Weakly I said \"I stopped going to church\" and I only say it was weakly because I think I could have, no, I wanted to just tell her that what I used believe hand changed.  God, if I could just erase the terrible memory of pounding my fanatic faith at her and effectively punching myself in the stomach and her away.  I used to be so sick over that girl, I mean sick.  I would listen for her steps on the dorm balcony and know when they where hers.  I would study on the floor she worked on in the library so I could run into her.  Every girl I saw with a pony tail and brown hair reminded me of her.  I was sick and crazy for her.  Perhaps in a round about way I lost my faith because of her.  That\'s quite a responsibility to throw on somebody, but her skepticism about religion, Christianity, began to erode away at my own beliefs about the world and my faith in Jesus.  Yet, once it was over, that is, once she said, \"We are going in two different directions,\" I think it was then that I wanted to go in the same direction as her.  She was a strong, poking, intelligent person and I couldn\'t imagine meeting someone else like her.  Such good times we had, me the pursuer and her the flee-er.  But I won, at sometime in our hide-and-seek, out of all the girls I could have, I choose her and she accepted, much to my surprise and joy.  But I was torn, bitterly and wrenchedly between my \"friend\" matt b., who exhorted the Christian \"don\'t be unequally yoked\" and I believed him and the Bible.    Back then, this is what I told her and showed, I was never as open and honest about my faith than with her in all my life.  I showed her verse and chapter on my yellow Goodwill purchased couch.  And I was happy she didn\'t turn then, but later she did and I hurt bad with regret because I really drove her away.  During this conversation I explained my change of heart in Christianity without the deep specifics and she seemed very interested, in a \"good for you\" sort of way.  But after some jolly laughs and pestering comments, it was time to say goodbye.  She said \'call me\' again and I jumped to asking her out for lunch or a drink.    I was so happy -I made a connection with Far---, my long lost soul mate (which essentially I\'m thinking this).  Two days later I think I blow it, yet it was as inevitable as the Great Gatsby, which I was reading during these days.  The story The Great Gatsby mirrored my own life.  Gatsby pursued an old, short lived bright flame that snuffed out once the wick that it lived on was spent.  And his pursuant destroyed him.  He did nothing that didn\'t lead him to that all too glorious dream of being with Daisy again.  Well, Far--- began to turn into a Daisy for me and in my excitement of being so near to her, I called two days later and lost my cool.  \"Not much has changed in two days\" she said.  I stumbled for words, something to say.  \"Well, I was thinking about you and didn\'t think you had anyone to talk to, unless you know your neighbors that live around you.\"  -This was more of a confession than an actual thought. I was the lonely one and by the end of the conversation she had flexed her individual might at my attempt to reconnect with something disassembled and I think I communicated to her that I could be whipped.    She toyed with me and didn\'t have the energy to mess with me anymore (like I know her state of patience to be last year) so she broke it away in a fast good-bye, click and no mention for me to call her back.    Then there was just the hum of a computer and a long Sunday night ahead for me to grow more and more enraged with myself for actually thinking I could make a leap back in time.  The reality of the situation was painfully obvious to me during the moments following the talk.  So I left the house and did the one thing that I am realizing gives me pure joy and surreal experiences for my isolated existence and mind -working-out.  I go and do my workout with anger and fear, fear that someone would see my wounds, busted open again.  To deter looks and hind I wore a hat tight and low around my face and I went to WG where I would not be a FW employee but another person working-out.  I was satisfied, tired and went home, only that I was nice to Kelly, or at least made a gesture of cordiality (I gave up a good treadmill so Kelly could run on it)  \"your welcome\" I said and moved on.  That Sunday Night I finished the Great Gatsby and saw what a fanatic life would do. It was awful that Jay didn\'t know anyone, didn\'t see the need to know anyone.  He wanted to realize his dream in a vacuum, in a far off place where no one could reach him.  He had the parts pieced together but one, he didn\'t have Daisy and the years had removed him from her and nothing, not even the glamour he adored, could make a last midget in Daisy\'s commitment to Tom.  This taught me something.  I learned that if I let my fantasies get the best of me then I too would be gone in the deep end.  With new found resolve to venture and learn in this world, other than what can be done in a relationship, I go through my workday looking forward to that time that I could workout.  And then it came and went.  By the time I was gone from the gym I was more puffed up than a farm roaster in front of the hen house and I felt great.  I floated through the gym, in front of members whose glances and flashing eyes said more than words could communicate, unless it was a girl saying \"Man Fletch, you\'re Jacked!\"  I\'m lingering around and as I spied on a girl that is the most amazing young girl I have seen in the club I realized something; I was untouchable.  This presence and the self awareness of it (most certainly) hadn\'t been there before, but as I saw this girl, who avoided eye contact and conversation with pretty much anybody (though she could attract anyone) I felt something transverse the space between us and I become untouchable.  No man could threaten me and no girl could whip me.  I was my own man free from earth\'s laws, suspended before air and I wanted to stay that way (though all things do fall).  Something told me that I could have control and the right girl would come to me and then I would pursue.  Live the life of now, not the life of the past.  It\'s not easy to be in the present.  Far---- and I will probably never see each other again, but I don\'t think I would mind if we did.   '),(13865,3601759,'10,June,2004','2004-06-10 00:00:00','   You Knocked my plant over You blew my house down You threw rain in my face You blew snow down my shirt You messed up my hair You turned the page of my book before I finished the page -again. You did nothing and I was hot as hell.  Then you were just right.  You pollunated the plants You hummed by the house You made the chime make beatiful unscripted music You made the snow form chilly stealthish banks You made the hot day cooler You made all things stronger.   '),(13866,3601759,'10,June,2004','2004-06-10 00:00:00','   from the Denver Post Kobe Bryant Gives the Facial and Wins the Case by Bob Sporty  In the trial that has become the watershed for all further cases involving superstars and lying beat-up women, Kobe Bryant was found innocent of all charges of sexual abuse by his accusing hussy, Kimberly Girth.  This verdict however did not surprise anyone in the court, since three weeks ago Kobe Bryant lead the Lakers to their fourth NBA title in five years, many have been outspoken that someone that glides through the air and makes key 3-point buckets could never beat up some two-beat tramp because they don\'t need to fight and plead for sex.  Kobe\'s attorney, speaking on behalf of his client said, \"Today, Kobe Bryant has proved to everyone that having rough sex is not a federal crime.\"  Evidence in the case turned on text messages that once believed to be erased where uncovered by AT&T techies.  These text messages, sent after the incident, protrayed a bitch that just got big nasty from a black dude.  Later testimony, from a jealous boyfriend didn\'t fair well for Mrs. Girth\'s case.  Though medical records showed that perhaps she was beaten up by both Kobe and her ex-boyfriend, the winnings that everyone recieved from gambling on the Lakers seemed to out weigh justice being served.  Mrs. Girth\'s attorney planed an appeal.  \"We are deeply troubled today, and so should our country, because sport stars are given preferential treatment they don\'t deserve.\"     Outside the scene was a carnival scene, with women\'s right activist and Kobe Bryant supports jostling over the verdict.  Susan Meanie, from the National Organization for Women (NOW), came from Boston, MA to protest.  \"We are deeply troubled that a guy who plays sports really well can become bigger than the justice system.\"   Many women, though, expressed different opinions in the crowds.   Alexia Ray, self-proclaimed NBA groupie said, \"That girl should be greatful.  Her problem was that she just wasn\'t good in bed.  I don\'t blame Kobe for anything.\"   In what was the most surprising follow-up to the verdict, Kobe Bryant\'s wife was seen wearing a spaghetti-strap shirt imprinted with a glass of orange juice that read, \"Thanks to OJ, I\'m able to walk free.\"   '),(13867,3601759,'10,June,2004','2004-06-10 00:00:00','   No longer do I care â€“nov 2002   No longer do I care But I will embrace you No more do I care About your looks Nor my own Come over and see me Naked in my bed My table my chair My life w/out a care Come and find me See yourself walking Down a hallway you dreamed of With wood panels and dirty Floors aroma of piss and despair Donâ€™t turn away You are all I have To bring me out of this shit Donâ€™t go into that day Stay until night  paint your face And be yourself No longer do I care   '),(13868,3601759,'10,June,2004','2004-06-10 00:00:00','   I read an article in the wall street journal today and it persuaded me to start doing a blog.  Usually I have lots on my mind so this should be fun to have as an outlet.  Look forward to lots of postings, most from a journal that is from the last couple of years.  later.   '),(13869,3601759,'14,July,2004','2004-07-14 00:00:00','   Have you ever sat in an empty field?  It feels like being in the middle of an ocean -looking around and seeing nothing different as far as the eye can see.  Walking through a cornfield makes me feel the same way -lost.    So I wonder to the middle of a field and image I\'m hovering over the ocean to face my situation.  Trying not to be scared, trying to focus on floating.  I\'m sensing confusion forming, losing, lost in my thoughts.  I act brave -I move courageously but I have fear -its painted on my skin and wrapped around my tepee.  It stares at me everytime I try to go inside my tepee, inside my heart.  There is warmth inside the tepee, there is faith in my heart.  I try not to be scared of faith -I\'m looking into the cosmos and seeing the life for me, not happening with me.    And I\'m absent -where is me?  Who is me?  She makes me feel different -you make me feel different -Who makes me feel like me?  (the answer is up to me)  In the field, there is space, but its not empty.  The essence of the space is god and the energy contracts around me and my space.  The energy from me expands to fill out the space, and then I retract to feel the space.  I\'m lost in the space.  I can only be found in the space.   '),(13870,3601759,'14,July,2004','2004-07-14 00:00:00','   I have this warm feeling and I can\'t sleep The darkness lets my mind see you you are fearless and bright The light warms me and I can\'t sleep  When can you be near me Our auras mixing together colors blending in swirls blown by the Wind pushing you near me giving me this warm feeling and I can\'t sleep   '),(13871,3601759,'14,July,2004','2004-07-14 00:00:00','   I chased a girl down and up the stairs.  I felt silly for I would catch the lovely flower only to let her go and go she did.  But I haven\'t forgotten her and much I still wonder about; about her and missing myself in love and in pursuit of her.  I wonder about love with someone and I think there is a woman somewhere tonight that I will someday meet and be enraptured with.  She may be thinking about me and doing the samething I\'m doing; writing about a future love.  But I can feel her love in my heart.  She\'s not beyond me.  She\'s similar in passion and beauty.  A raven haired, perhaps out casting shadows with her being.  Flowing from my dreams I\'ll find her in a marketplace finding the fruit of the Garden together and the knowledge shall be made for our liberation from Eden into the banishment that will be us together for God will watch over us granting favor in the toil of existence, with health and wonder in the life that reigns around us - relationship to all things by the creation.  Filling each other up from the taking of food, imparting it to us for the melding of our bodies into one.  Returning to the earth we\'ll depart from the body and play in the space that is in all things, where God resides, playing forever   '),(13872,3066719,'29,July,2004','2004-07-29 00:00:00','   At the end of the day, am left with this thought ... what have i done today for others, to this world...??!!!  It was just another day, doing the daily chores ... doing this - doing that, solving a problem, browsing, talking, eating, driving etc etc.  half our life time goes sleeping, another half split between adoloscence and aged and whatever rest we are left out with we freak out, play, argue, enjoy and do what not!!!  but, whats that difference am i making here? am i a part of this ant colony, just another worker ant, doing its routine - if only to a larger extent?  that\'s a kind of thought, which goes undigested. mmmm.  each one of us can make a difference to the world around us... yes and i want to something! but where do i start? and in what way? I certainly don\'t want to do something for the sake of making a difference but  for what it is, wholeheartedly!  Thousand ideas, thousand thoughts go zooming past - but i need action. i need focus. i need more patience. i need to do something more ... to find that missing link!  let me see - try to sort out the spaghetti of unceasing thoughts, just atleast to find a focus - to square on something that would make me feel better in this aspect.  food for thought, challenges to the will.  '),(13873,3066719,'29,July,2004','2004-07-29 00:00:00','   urlLink Greek Mythology  This site has some good illustrations (at times comical), but i liked it. Interesting read...  '),(13874,3066719,'29,July,2004','2004-07-29 00:00:00','   Discovered blogBuddy!  After browsing through  the Blogger API, the first thought that struck my mind was to create a small tool and there i go, getting the XMLRPC implementation for Delphi and halfway through my search i bump across blogBuddy!   The project seems to be long dead, but what the heck, blogBuddy works! check it out!    its cool! urlLink  '),(13875,3066719,'29,July,2004','2004-07-29 00:00:00','   If you like contemporary art, don\'t miss this.    Even otherwise, its pure delight to watch these objects d\'art  PS: How i envy the artist, let alone the girl on the painting  '),(13876,3066719,'29,July,2004','2004-07-29 00:00:00','   Am not alone, in this virtual world... shocked (should i be???) to find the rantings of these geeks   urlLink Revaz Musinz   urlLink The Rock  '),(13877,3066719,'29,July,2004','2004-07-29 00:00:00','   I Discovered that am an INFP! (for those who want to know what\'s INFP, i suggest to check this out : http://www.humanmetrics.com/cgi-win/JTypes1.htm)  Here\'s a list of famous INFP\'s ----------------------------------------------- Homer Mary, mother of Jesus St. John, the beloved disciple William Shakespeare, bard of Avon A. A. Milne (Winnie the Pooh) Helen Keller, deaf and blind author  Carl Rogers, reflective psychologist, counselor Jacqueline Kennedy Onasis Julia Roberts John F. Kennedy, Jr. -----------------------------------------------  Whew, thats a nice note to start off my BLOG. lol.  Thanks to [sriram, my friend], for triggering the spark to BLOG, which i\'ve lazily discontinued, after a brief start, earlier in 2003.  '),(13878,3066719,'04,August,2004','2004-08-04 00:00:00','   Intimations of Immortality       Ye that through your hearts to-day      Feel the gladness of the May!   What though the radiance which was once so bright     Be now for ever taken from my sight,   Though nothing can bring back the hour     Of splendour in the grass, of glory in the flower;   We will grieve not, rather find     Strength in what remains behind   (William Wordsworth)  '),(13879,3066719,'02,August,2004','2004-08-02 00:00:00','   urlLink     Cat on the Move!!!  Finally! Revaz ventures out of the den. :o)  Way to go, buddy! Its a long road ahead...  Feeling so very happy for you. :o)  '),(13880,3066719,'02,August,2004','2004-08-02 00:00:00','   \"The year is 2050, and you are standing in front of a vending machine. What form of payment will you put into it?\"   -perhaps nothing- ? you just walk in, select what you want and walk away...! it could be your fingerprints, voice or ur retinal print or perhaps an embedded bio chip!  There might not exist a vending machine at all, standing out there, in the corner of the street!  mmm, what would be offered for sale on a vending machine?  well, well, that already leads to a chain of thoughts, tickling the creative bone...  Do we have the time to spend a moment and think about it? why isn\'t there any more of those ground breaking inventions, as of the early 19th, 20th and 21st centuries?  Where are we heading to?  Well, thankfully there are still some inventors out there... Check this...   urlLink The Impact Lab...  '),(13881,3066719,'02,August,2004','2004-08-02 00:00:00','   Had been to the FORUM yesterday. Cool place. It has been modelled on the lines of a typical US Mall.  Quite colorful and too much of activity. nice to hang out...  will be back again  '),(13882,3066719,'02,August,2004','2004-08-02 00:00:00','   Too much of action this weekend... Whew! Finally, decided to tie the knots! Gave the green flag, for the hunt!  Got to wait and watch, whats on store.  \"I want to know God\'s thoughts; the rest are details.\"  - Einstein  '),(13883,3066719,'31,July,2004','2004-07-31 00:00:00','   This post stems from an attack of altruism. Faced with the sudden freedom of being unemployed, I think that some of this time could be best spent in helping fledgling souls find their way. (I define fledgling souls as beings that have finally discovered that they are a soul and not a body:-)). Or rather help them not lose their way. So here are a list of books that any soul worth its salt should avoid at all costs.  1. Illusions - Richard Bach - Richard Bach tops my list of purveyors of kitsch in the garb of neo-spiritualism. Bach attempts to peddle Advaita to the masses and succeeds for maybe 5 minutes. But if you are serious about wanting to know about the self, find a different Guru.  2. Conversations with God - Neale Donald Walsch - What put me off most about this book, is the author\'s claim that God ordered him to write the book. Don\'t read this book if you want a personal relationship with God or you feel you have one.  3. Notes to Myself - Read this book for a prime example of how you can convert an ego trip into a profit making enterprise. First you had Notes to Myself, then Spiritual Notes to Myself, Notes to Each Other and then Notes on Letting Go....yeesh!!!!  4. Autobiography of a Yogi - The first half is good, the second half is an exercise in excess. Reading the second half, especially the revelations of Yogananda\'s guru on life after death temporarily destroyed any interest I might have had in the spirit. It can confuse the daylights out of you.  5. A Bridge Across Forever - Richard Bach - This book, contrary to other Bach books, seems to go on forever. Maybe that explains the title. It is a long and tedious saga of Bach\'s quest for his soulmate. What compounds the tedium is not just that Bach divorced his soulmate but also subsequently explained it in suitably spiritual terms of being temporarily separated in his journey and all that blah. This was on his official website which now seems to have been taken over by his publisher.  There is a single thread that runs across all books (except the Yogi) - they are all pulp fiction in the garb of true spiritual experiences. They try to claim legitimacy for purely materialistic intents by portraying their authors as spiritually advanced beings who have been asked by the Divine to enlighten poor souls like you and me (and also make millions in the process). They make a mockery of what is essentially a personal and totally individual quest. There is a palpable lack of sincerity in the words. In essence, they lack \'Om\' - a union of the spirit, body and the mind:-).  If you want a list of books that do inspire, start off with Gandhi. '),(13884,3066719,'29,July,2004','2004-07-29 00:00:00','   The  Me cury  S urface,  S pace  EN vironment,  GE ochemistry, and  R anging mission (MESSANGER) would study Mercury. This is the closest we can get to the  Sun &nbsp;(for now that is :)).         The MESSENGER would be launched on  August 2nd, 2004  and it takes 7 years to reach -  March 2011  it would reach Mercury !!.     &nbsp;'),(13885,3066719,'28,July,2004','2004-07-28 00:00:00','   - OM - in Hindu culture signifies an union of mind - body - spirit. This blog simply aims to unite the thoughts of (minds) different people (bodies) to acheive (the spirit) an unison of action. &nbsp; '),(13886,3982753,'31,July,2004','2004-07-31 00:00:00','       urlLink    urlLink     Not a very nice picture. Taken 2003 when my patient died prematurely of Hydrops fetalis. This is the autopsy. Note the chopping board. It was not my fault!! Honest!!     urlLink    The autopsy  urlLink      urlLink    The autopsy2&nbsp; urlLink      urlLink    Post autopsy&nbsp; urlLink     A while ago, I tried out the mms feature of my phone wherein I experimented in sending this archived pic.  When looking at the picture: 1. Remember that before you barf or faint, this was a person albeit a premature one, who died because of something we can control. Well, the mother can quit her drinking, smoking, sex and other vices that can harm the fetal development. Plain cause &amp; effect. But sometimes, due to a cruel interplay of random selection which we cannot control, tragedies such as this can happen.  2. Just because something is as assaulting as this pic, doesn\'t mean it has lost all of its moral ground. Autopsies are just a normal procedure in the medicolegal field- it\'s not advertized, that\'s all.  3. The picture is to document the pathology involved so with more research on the disease can decrease its incidence. This is neither to mock nor to dehumanize the dead patient, but rather to scientifically record his short existence without malice or prejudice.  4. Respect for human life begins in understanding its secrets.      '),(13887,3982753,'30,July,2004','2004-07-30 00:00:00','       Le Chatelier\'s Principle: \"Complex systems tend to oppose their own proper function.\"     I, Robot  is the new movie starring Will Smith involving him in the \"man against the machine\" theme melded with the question of Darwin\'s evolutionary survival. The twists involved are obtuse, while the plot is linearly predictable which makes this film not that \"spectacular\" or memorable but with it\'s flashy visual effects and cgi animations, this makes the movie a notch below  Minority Report . Though linear, the mystery and the fast pacing keeps you at the edge of your seat. The dialogue is cliched and very very hollywood with all their pregnant pauses and their saccharin expressions- well what do you expect with a high-budgeted sci-fi schlock from tinsel town? However, the advantages outweighs the downsides.  The action and the thrill of the chase keeps your eyes on the screen.   I give it  7.5/10 .      '),(13888,3982753,'27,July,2004','2004-07-27 00:00:00','       The Cardinal Conundrum:           \"The Optimist believes we live in the best          of all possible worlds.  The Pessimist         fears this is true.\"      '),(13889,3982753,'27,July,2004','2004-07-27 00:00:00','       I went to Robinson\'s Place today with Jorin and watched \"King Arthur.\" I just realized the following:  1.  The higher the budget for any historical fantasy Hollywood film is directly proportional in terms of the cheeziness, crappiness, corniness and stupidity of its acting and characters.  2.  That any Euro-historical film done in English will sport that bloody English accent. No exceptions noted.   3.  The more fantastical the story is, the more fantastical is the logic of its actors. Imagine just staring at your enemy when the ice beneath your feet has already cracked off. ... Or indulging in a piteous caterwauling of a speech about freedom, friendship and other stupidly profound subjects just as when the enemy is about to attack your fort.   4.  You see one Roman-themed film, you\'ve seen them all. If Maximus in Gladiator sported an impassive poker face, so will Artorius (Arthur) but more crappier. If you\'ve seen Legolas, the elves, & Helms deep, then you will find it refreshing to find them recreated in King Arthur.  5.  The \"wear & tear\" of costumes of historical films (especially Roman films) corresponds to the number of days of shooting. Roman armors after \"so-called\" countless battles remain spotless without dent or wrinkle. It\'s already bad enough that the acting is so fake.  Don\'t believe me? Go watch it!  Also,  6.  The person two seats beside you clad in sando will intermittently expectorate into the floor fronting him. And you think these things do not happen. Only in the Philippines. People think that their P75 ticket aslo bought them the right to trash the place. ***sigh      '),(13890,3982753,'26,July,2004','2004-07-26 00:00:00','       urlLink   I am  A Shit Brick  .   The Shit brick is usually the kinda person that doesn\'t like doing much. Turning down a nice romp in the park to his or her favorite television show and a nice big Mac or a box of chicken nuggets. Though shit bricks have their anti social flaws, what they lack in social values they make up for in laziness... But I guess that isn\'t a good thing either way.. Ok, shit bricks are just lazy fucks...   urlLink What Kinda Shit Are You?'),(13891,3982753,'25,July,2004','2004-07-25 00:00:00','       I definitely have the colds. My eyes can\'t stop watering and my sinuses are clogged with a mountainful of mucus. I already took Tuseran and Biogesic but it seems it\'s not that effective. Well, I should know, shouldn\'t I? ***sigh*** Mostly it\'s viral so all treatment is&nbsp;therapeutic and symptom-based. Good thing, there were Sylvanas in the fridge. Mmmm!  Got a pirated DVD from ol\' trusty Quiapo. It\'s called \"Mishima\" which is primarily based on the life of Japan\'s most prolific and distinguished post-WWII author, Yukio Mishima. I thought at first, \"Ah.. Japan.. my homeland... anything from there is good.\" I guess I was mistaken, \'cuz the film itself was too \"poetic\" that watching it akins to appreciating paintings by MirÃ² or Basquiat in one sitting. One, it was boring and bereft of anything \"samurai\" unlike the films of Kurosawa. But pettiness aside, the film is a wonderfully made and sincere portrayal of the author-schizophrenic. The music which was composed by Richard Glass is very haunting and appropriate for Mishima.      '),(13892,3982753,'22,July,2004','2004-07-22 00:00:00','       urlLink      '),(13893,3982753,'21,July,2004','2004-07-21 00:00:00','       I dunno what the hell\'s happening but suddenly (more like the whole afternoon) I feel as if&nbsp;I was&nbsp;going down with&nbsp;a cold. Good thing there\'s Paracetamol around because at least in this way, the symptoms won\'t be as bad. Most&nbsp;likely, this is just a viral&nbsp;infection so antibiotics&nbsp;are out of the question (for the time being.)&nbsp;But damn, that feeling of fatigue, heavy eyes, raspy throat, and increased frequency of sneezing, makes the condition worse than it seems.   Nana, Raprap and I went to Robinsons Place today to buy some groceries. Goodness, the prices have skyrocketed so much that our one cart-full of groceries (mainly veggies, meat,&nbsp;toiletries &&nbsp;household items) costed us like 3k. When I was still in college, say 1998, P1,500 worth of goods are already sufficient for the monthly use. But now, our peso value has sunked so much that it is not worth anything at all. Your P100 can now only buy a couple of items, unlike then, when that kind of money can already buy a family-size pizza.   I also bought an internet card from Mega store near National bookstore. I was rudely awaken last night by the sudden disconnection of my presumably unlimited PLDT Vibe postpaid internet line.&nbsp; Reason? Well, that fucked-up company said that I overused over P2k worth of internet hours, consequently, my account was blocked until the 1st of the next month REGARDLESS of PAYMENT&nbsp;of the&nbsp;BILL FOR THE CURRENT MONTH! So, I asked that since I paid my bill for this month, won\'t the usage also revert back to zero? The guy said, \"No sir, regardless of payment made, usage is charged from the 1st of the month.\" Huh? So, what\'s the use of paying on time? And furthermore, why don\'t they&nbsp;inform the users that such can occur and will occur regardless of loyal payment of bills? Just to show you that logic and reason here in the Philippines is and ever will be skewed.&nbsp; It\'s not very customer-friendly of them to treat its&nbsp;customers this way.&nbsp; So right now, I have to contend myself with a trusty anonymous internet card and an upcoming sniffles. Pass me the Tempra, will you?   Ahh... be sure to check-out this satirical site (Thank God for small miracles!):   urlLink http://www.pldt.com      '),(13894,3982753,'21,July,2004','2004-07-21 00:00:00','       urlLink      Ready for the kill!!     urlLink    All smiles from here!&nbsp; urlLink      ahem ahem ahem... allow me to clear my throat... the photo above was taken when? last may? yes, it was.. last may. anyways, jan paul hormillosa of new york arrived here last may for a long vacation. and during his last leg, he stayed near gb4 in makati and during our last gimmik with him, we went to bizu, that swanky dessert resto in greenbelt 4. the place was painted ochre and purple like cheeze and ube, perhaps to stimulate the appetite. there were lots of yuppies and buena familias there... obviously oblivious to the dessert. ohhh... the dessert... those sinful delectable mounds of sweetness. i ordered vanilla shake to go with my cheezecake/mousse and at first i was a bit disappointed at the proportions given. well, this was a big irony. it\'s because the cake was so heavy and laden with cream, chocolate ganache, and syrup that after a few bites, i surrender to its prowess... so there were: Terai, Jp, Me, and Robinson all prepared to indulge into the most sinful desserts imaginable. mmmmmmm.... the chocolate was intoxicating,... so rich... so sticky that it clings to the palate and the cream so ....well, creamy. anyways, the service is such that is expected from those swanky gb4 restos. crisp, clean cut and somewhat efficient....but not that creative and attentive to the customers      Ratings: (5* criteria)     Interiors: ***    Food:****    Value for Money:***    Service:***    Overall: ***    urlLink  &nbsp;      '),(13895,3982753,'21,July,2004','2004-07-21 00:00:00','       Are you feeling tired? Down in the dumps? Need a laughtrip?   Then, go to:     urlLink http://www.blurty.com/users/caffeinerush/     urlLink http://schadenfreude.blog-city.com/read/390719.htm     Geez, these coÃ±o pricks really have delusions of grandeur.      '),(13896,3982753,'21,July,2004','2004-07-21 00:00:00','       Filipinas          Filipinas, Filipinas,  Where art thou, my countryâ€™s bride?  Have you gone deep into slumber?  In your shame you do now hide?       &nbsp;   For today your childrenâ€™s ember  hast lost their glorious past.  For none shall burn nor remember  In the season of your last.'),(13897,3982753,'21,July,2004','2004-07-21 00:00:00','       I am really not that fond of artfilms at first because of the lack of exposure and the availability of materials in local video rental shops. It\'s true! Go to any decent video shop and ask if they have, let\'s say,  Amelie , most of the time, they would only shake their heads faster than a tree shedding it\'s leaves in autumn. It was the Hollywood schlok that ruled the nest. Even during the vcd era, there were not a lot of artflicks other than the asian horror genre.&nbsp; Back then, it was what?  Ringu, the Phone, Pang Brother\'s The Eye . But that was it. &nbsp;It was only after I got hooked on dvds that I learned to appreciate artfilms as a whole. The first film i bought was  Cuaronâ€™s â€œY Tu Mama Tambien\" , followed by â€œ Crime of Padre Amaro \", and the rest was, well, history. The amount and type of dvd artfilms available depends on what the quiapo bootlegged dvd market supplies. There, you can buy any artfilm for P65-70, which is also the price of watching a hollywood film in, letâ€™s say, Robinsonâ€™s mall. If you go there, just ask for the foreign films. They would say, â€œboss, poren? poren?â€ Because foreign films are hard to come by, supply for it is small even if the demand is huge, so a full supply of  pasoliniâ€™s â€œDecameronâ€  dvd now will be no more in less than a week. I vascillated, so I wasnâ€™t able to get a copy. damn. I mean, if you want to go treasure hunting for artfilms & classics, quiapo is the place to be. Itâ€™s what people in internet threads call â€œthe dvd lair\", so most probably youâ€™ve gone there too. =)  &nbsp;  A week ago, i went to  Quiapo  (Ahhh... that veritable ghetto for everything illegal) and bought criterion collectionâ€™s â€œ Seven Samurai \", â€œ Lâ€™Appartement â€ and even the disgusting â€œ Salo .â€ Talk about luck. and you wonâ€™t find any originals in Astrovision or Odyssey. So, who says we must not to patronize piracy when in fact there are no original counterparts for artfilms being sold in legitimate stores? And besides, though I donâ€™t advocate pirating Philippine films, piracy of arthouse films brings movie appreciation to a higher level wherein the masa can now appreciate reading subtitles. And besides, collecting these films is a good way of building my own filmlibrary or archive since the pirated dvd copies are quite good and the quality seldoms depreciate over time unlike vhs or vcd.&nbsp;Suddenly a thought of guilt ran through my brain, but thankfully, looking all those neatly stacked up dvds pushed that thought away. What a relief! Long live DVD piracy in the Philippines!!      '),(13898,3982753,'21,July,2004','2004-07-21 00:00:00','       Thank God I found this site.... a new blog a new beginning. There have been blogs for which I am a habitue of like Demented  urlLink http://demented.so-phobic.com &nbsp;where a small circle of close friends update themselves with their daily grind, and a group blog in Yahoo groups composed of retired CAT officers also whooping about their daily grind. Ugh!  &nbsp;  But now! Mwahahahahah! Mwahahahaha! I have my own blog where I can shout to the world! I can now hear Dr. Frankenstein in the black&white background say, \"Igor! It\'s Aliiiive! It\'s Aliiiive!!\" And so, without much further ado, I hereby declare this blog officially open!  &nbsp;  clap! clap! clap! Thank you! Thank you!      '),(13899,3982753,'18,August,2004','2004-08-18 00:00:00','       During my  Bacolod  years, my grandma whom I shared the room with (yes, hers was air conditioned so \'twas better that I sleep in her room!) sometimes reprimand me when I tinker with her TV saying, \" Nunca pahilabti ang TV basi mapierde ina karon !!\" Any Ilonggo can understand that statement perfectly well which reads, \" Don\'t even try to disturb the TV for it may \"lose it\'s efficiency\" later on .\" For us in the province, we were not acutely aware that our tongue was peppered with  Spanish words  or  Hispanicized Ilonggo.  We just took it for granted and assume it\'s vernacular Ilonggo.  Coming to  Manila  however made my senses more acute since I have to shift to  Tagalog  when speaking with people. (I hate code switching!!!) It seems that many words here are of  Malayo-Austronesian  in origin while the same counterparts in Ilonggo are usually Spanish loaned. For example, the word for \" open/to open \" in Tagalog is \" bukas/buksan \" while in Negros, it\'s the Spanish \" abri/abrihan .\" Another is \"meanwhile\" wherein in Tagalog it\'s \" habang \" while in Ilonggo, it\'s \"mientras (while) or mientras tanto (meanwhile).\"  Some of them are as follows: [English- Tagalog(vernacular)- Ilonggo(vernacular)- Spanish] 1.  Eyeglasses -Salamin- Antipara,Gapas -Antipara,Gafas 2.  Gun -Baril- Pusil -Fusil 3.  Teacher -Titser,Guro- Maestro,Maestra,Titser -Maestro,Maestra 4.  Lose -Talo- Piyerde -Pierde 5.  Cheap -Mura- Barato -Barato 6.  Change -Sukli- Kambyo -Cambio 7.  Dance -Sayaw- Bayle -Baile 8.  Key -Susi- Liyabi -Llave 9.  Clean -Malinis- Limpyo -Limpio 10.  Until -Hanggang- Asta -Hasta  I don\'t know whether I should feel proud or happy or ashamed about having a more colonial tongue. It doesn\'t diminish though my love for the soil of my forefathers. Yeah right!! Soil, my ass!! I guess this matter is inconsequential and moot, not worthy of killing my remaining neurons. But it makes me wonder though. Is development or the dynamism of a language determined by the progress of its users? Has Ilonggo stagnated itself into the nostalgia of the past? I hope not. Tagalog seems to have more English incorporated into its vocabulary than any other Philippine language, probably due to the fact that it\'s the country\'s major port- an entrepot of global influx. Whatever. Basta, it\'s moot and academic. Urgh!  Oh well...that\'s it.... no more neurons to burn.      '),(13900,3982753,'16,August,2004','2004-08-16 00:00:00','       Wow. This reality show sure is fun!! Not only you get to see the  Pyramids, Catherine\'s Palace in Russia, the Nile, Kilimanjaro , but you also see the pairs struggle their wits as they pass each challenge whether it may be physical (pulling a huge block of Pyramid stone 50 meters using no wheels), gastronomic (eating 1-kg of caviar in a hurry) or mental (block ice hockey pucks 5 times). The tests are ingenious and never boring and the contestants are really funny (with all their sarcasms, double-dealings and insults) but not to the point where you get irritated by them. This is so unlike  Survivor  where you were stucked in a never changing island full of whiny, lazy, spoiled inconsiderate ugly Americans who do not know how to cook decent rice doing nothing all day except scratch their folliculitis-infected bottoms and conspire to smilingly backstab their groupmates.  When you watch  Amazing Race , you heart goes tachycardic (>100 bpm) and your skin become diaphoretic (sweaty) from watching the hurried & harried faces of the contestants who are all under time constraints. The visual &amp; aural stimuli are tremendous, resulting with you wanting to watch more. Better than Survivor.    Update:   They\'re on the way from  Tanzania  and going to  Dubai, UAE  next week. And the whiny cousins  Mirna and Charla  (people call her Shmirna) are thankfully eliminated. Note how loooong Mirna\'s hug with Phil! And Charla\'s speech made  Phil Keoghan  choke! hahahaha...  ;P I pray that somewhere in Dubai,  Colin & Christie  will be ambushed by Taliban Ninja Fighters or something.  And I pray that cool  Chip & Kim  wins this so amazing race!!! They have the class and the decency that no other team has. Isn\'t it ironic that it\'s the white pairs who are showing how ugly Americans are?  Cheers!    Next Week:   Colin goes to jail!!! Yahoo!!      '),(13901,3982753,'16,August,2004','2004-08-16 00:00:00','       A Midsummer Night Lullaby by: Julsitos  Now rest your eyes, my dearest child, lay upon your mother\'s breast, and feel the warmth, so sweet and mild, your peaceful face gently pressed.  Hush little one, the day is done, for you are in my keeping. The ills of yesterday now begone its secrets now asleeping.  Can you hear the crickets playing, hidden \'mong the grasses there? Can you feel the wind whisp\'ring blowing through the midnight air?  And when you yawn, your face alights without a care or worry. To dream in splendid perfumed nights, is paradise and glory.  In the darkness of midsummer not a stirring, child and I, for in sleep we sleep together, and dream beneath a moonlit sky.'),(13902,3982753,'16,August,2004','2004-08-16 00:00:00','       Don\'t you just hate it if you have a  writer\'s block ? I mean, when your head is out of ideas to churn out or the words are lost in the air. There is some grain of truth to what \"spiritualists\" say about writing. They theorized that creativity (writing, painting, composing, etc.) comes from inspiration, and this inspiration is possible because our \"spirit guides\" or guardian angels send these signals or ideas for us work on.  I wonder if seasoned authors get such block. If they do, here are some of my take on their works:  Can you imagine  J. K. Rowling  of the  Harry Potter  series if one fateful day she was stricken by the bug? She might chuck it all up ala the painter Paul Gaugin and decide that she is getting tired of Potter that she instantly have him and Hermione elope for good. Well, if she wishes that Ron shall enter the fray, then they could have a menage-a-trois. If I were Rowling, I would suddenly have the wicked urge to end the saga in the middle chapter where Harry will have a heart attack and due to an earthquake, Hogwarts shall crumble down and kill everyone to smithereens. Yes, even incontinence-stricken Dumbledore dies because the dumb Dumbledore was on the act of changing his adult diapers in the lavatory.  The Horror!  How about the  Da Vinci Code  by  Dan Brown ? Doesn\'t he get the writer\'s block from time to time? He might have been tempted to end the chase then and there at the villain\'s house but since he is so well into the novel, we get a hundred pages more to read. Wouldn\'t it be funny that instead of the couple finding all those clues in the Louvre, the sprinkler system kicked in and washed away those written messages leaving Langdon and his mate clueless as ever. Or what if the zealot monk decided to hijack the private jet and as a protest to Bush\'s War of Agression, he rammed the plane into Big Ben just like Sept. 11? Such would bring a smile!  So, after writing this rant, is my spotless mind still blank? Positively so!      '),(13903,3982753,'11,August,2004','2004-08-11 00:00:00','       There was an article about a week ago wherein the government announced that the  air in Manila  is at par with world health standards. What tha? Have they gone that stupid from testing air quality here? To think they are \"college\" graduates\"! I guess a degree does not gurantee intelligence, that\'s fer sure! There are only two conclusions to their disinformation: one, it\'s just the government\'s run-of-the-mill white propaganda (more like  puro pampaganda ) or two, their air quality gathering equipment is so decrepit and obsolete that any result coming from it is off-tangent. Probably, the air quality they were investigating were those of Mars or Jupiter. Most probably from Uranus! Geez. They think Filipinos are that dumb not to notice their celestial results.  If their quotations cannot be trusted, whose then? Simple. Our boogers!  Yes! Our booger, kulangot, pong-it, or snot is the best indicator of air pollution levels in Metro Manila. Not only it is free and easy to get, it\'s natural!!  How? Just at the end of day, pick your nose and collect a sample of your daily remnants. Then, rub it in between your fingers until you get a nice homogenous consistency in which you then compare it with a corresponding color chart that shows the present pollutants in the atmosphere. Easy, no? It\'s better than to trust some ambigious result doctored by the DENR. =)      '),(13904,3982753,'10,August,2004','2004-08-10 00:00:00','       Louie, Enna (both batchmates & interns at Manila Doctors) and I watched Collateral at Robinson\'s Place Manila yesternight. I just settled for White Cheddar popcorn and Ice tea from Tater\'s to make my viewing pleasure more palatable.  Watch out! Random spoilers ahead!  For a complete synopsis, check out this site:  urlLink  Collateral from themoviebox.net      To make the whole gamut short, the movie was too fast and furious. Tom Cruise\'s timely end was anti-climatic but there were good chase-and-run scenes in the end. The premise as the assassin is satisfactory but the premise on why he\'s killing those people is at most, murky. His acting was Hollywood-like as always, with posing-posing, pregnant pauses and cowboy dialogues, but he managed to pull everything off in his favor. But I have to praise Jamie Foxx (Max) who played the cab driver to was coerced into servicing Vincent in his game of \"sutukil\" (shoot-to-kill). Max\'s acting was brilliant- the way he was panic-stricken by all the lightning speed events happening around him and his change of character from the optimistic coward to the fatalistic hero. His emotions were neither melodramatic nor exaggerated but were on the other hand, befitting for someone in the edge of being shot, not to mention that his complicity to Vincent\'s job made him a \"black\" accessory to the crime. I simply enjoyed the look on the driver\'s face when he was told by Vincent to retrieve the list of victims from Felix. His face became white from fear. Hahahaha.  Collateral, while a good cat-&amp;-dog film, gave me migraine. I don\'t know if it was from the popcorn or from the drink, but I\'m sure all those exploding car chases, blair-witch carmera shots, and those endless running contributed to that throbbing feeling inside my head.  Soooo... I rate this movie: One thumb up!!! Yipee!! (Cuz the other thumb is on my cell\'s keypad texting!)      '),(13905,3982753,'08,August,2004','2004-08-08 00:00:00','       I just arrived from  Malate Church . It\'s one of my favorite churches and attending their Sunday masses is sheer joy. More like inner joy.   The priests there are mostly  Columbans  who primarily are  Irish . I find their presence very humbling (imagine leaving their homeland just to spend the rest of their lives here in the Philippines) and inspiring (for they\'re very active in community works.) Their experience in poverty-ridden Negros during the 1980s\' \"years of famine and insurgency\" strengthened their faith and resolve in propagating the Catholic faith through example. This in turn, makes their words effective in touching the heart. They seem to be surrounded in an aura of holiness which is very effusive and infectious.  Their sermons are conscience-hitting steeped in suffering and hope, love for the poor and the sick, and most of the time they talk about our obligations as Christ\'s children to our fellowmen and our duties to Christ himself. Their sermons are based on their experiences with the people of Negros, and not on lofty abstractions that only theologians can relate. What they have is  heart  and hearing their words can bring about the conversion of one\'s soul.  (The question is: Is that change going to last as soon as you leave the Church?)  Because of their zeal and their holiness, the Columbans of Malate are highly respected and loved by their parishioners. A very good example can be seen right after the mass. It\'s only in their church where children, as young as 3 and as old as 30, approach the priest to make the \"mano\" sign. You can feel your heart being tugged as you watch the priest being delayed from leaving the aisle by hordes of people jostling each other so they can all make \"mano.\" Such a sight is far from the hatred and animosity felt by Filipinos for the Spanish friars- but that\'s another story.  However, I am disappointed with some of the parishioners. Why?  They leave at the middle of the mass.  They prefer standing up and taking a bow right after the sermon or after eating the holy host. What on earth people think this is? A movie house? A play? A theater? Do they think this is not worth finishing at all? Do they think they can just leave without saying goodbye? It\'s as if you simply left the banquet table without letting the host know that you\'re leaving. I don\'t want to sound sanctimonious or oh-so pious but finishing the mass is the least you can do as a sign of respect for the faith and the religion you are affiliated with. If you really believe your faith can save you, then at least, give it the proper attention it deserves.   Sure, sure, there\'s that clause which says, we humans have the \"freedom of choice.\" Well, freedom of choice indeed! Blah! Blah! Blah! Rationality sucks sometimes.  Oh well, just another Sunday blues.      '),(13906,3982753,'07,August,2004','2004-08-07 00:00:00','       Angels & Demons (Dan Brown, 2000):  Hmmm... still on the first chapter. But nonetheless, it\'s as good and fast-paced as the The Da Vinci Code.    The Decameron (Giovanni Boccaccio, 14th Century):  To think that Medieval Literature is boring, this should dispell any notions. Thanks to  King Arthur, Chaucer, Thomas More, St. Augustine , etc., we tend to think of Medieval lore as painstakingly boring and full of old English yarn. We conjure ideas of princesses with pointed hats being rescued from dark castles by prince charming, or the Robin-hood like chivalry and cavalierism during those times. However, we must remember that the Black Death ravaged most of Europe during the Dark Ages. Not everyone is in the church praying and kneeling before the pews feeling pious.  This classic shows the bawdy, humorous, satirical and farcical side of the Medieval Era. It\'s not about noble knights searching for the holy grail or battles fought for love (ala Romeo & Juliet) or philosophical monarchs going murderously mad (ala Hamlet & Macbeth), but rather it\'s the daily activities of the common people of Italy. I have to admit, most of the population then were poor. Most of the time they eat, feast, get laid, blackmail, cheat, murder and get drunk... just like today. You agree, GWBush? I can see you nodding your head Erap.  After seeing  Pasolini\'s The Decameron , I simply have to check if the film was faithful to the original manuscript. Well, mostly yes.  The blurb from  Dover Publications :  \"While the Black Death rages through 14th-century Florence, a group of young people retreat to the countryside and amuse themselves by telling tales of romance and adventure.  This is the premise of Boccaccio\'s Decameron... Vast in scope, teeming with colorful characters, and rich in worldly wisdom.  Folk tales, ancient myths, fables and anecdotes range from earthly and irreverent satires of hypocritical clergy, to gripping tales of murder and revenge, to stories of passionate love, both adulterous and faithful.\"   Some of the stories from  The Decameron :   Ninth day, Story II : An abbess rises in haste and in the dark, with intent to surprise an accused nun abed with her lover, thinking to put on her veil, she puts on instead the breeches of a priest that she has with her: the nun, espying her headgear, and doing her to with thereof, is acquitted, and thenceforth finds it easier to forgather with her lover.   Third day, Story I : Masetto feigns to be dumb, and obtains a gardener\'s place at a convent of women, who with one accord make haste to lie with him.   First day, Story I : Ser Ciappelletto cheats a holy friar by false confession, and dies; and having lived a very sinful life, is, on his death, reputed a saint and called San Ciappelletto.      '),(13907,3982753,'05,August,2004','2004-08-05 00:00:00','       urlLink    The hidden location of the pirate\'s lair. Just go into Arlegui Street.; urlLink     Hmmm... And so the plot thickens. This is one of the entry points to the infamous  Quiapo DVD market . From the  Quiapo church , you have to cross the underpass to the other side of the avenue. There, you can ask for directions or better yet just follow where most people are going. Here,  Arlegui street  poses as one of the less crowded access to the fabled stuff. Just go straight. Do not panic if you see a cop because as far as anyone with an IQ higher than 90 knows, they are frequently batting a blind eye to this seemingly illegal trade. I think their philosophy is that as long no one is hurt, killed, maimed or kidnapped, then everything\'s well.    Caution:  Do not bring lots of cash, or credit cards for that matter. Do not show off your cellphone in public, or use it in the streets.  Do not wear expensive watches or jewelry for they can become targets for street pickpockets. Do not wear formal clothes because Muslim merchants are wary of giving discounts to coÃ±os. Furthermore, do not make any snide remarks ala GWBush about Muslims because you won\'t get any discount that way. (I was about to write \"you won\'t be able to escape Quiapo alive if you do insult them.\")   urlLink    Inside one of the many alleys in DVD country; urlLink     There are lots of places to start. And you have to be discriminating with your choice of DVDs. Just don\'t take it as it is because if something is wrong with it, it can be very difficult to return it back for exchange. There are rows and rows of DVDs and with luck, you can stumble on rare titles (like the works of Paolo Pasolini & Kurosawa), or better copies of unreleased films. The DVDs are stack on shelves upon shelves extending from the floor to the ceiling. When they say DVD copy, it means it was copied from the original, but when they say \"Clear copy,\" it means it\'s not worth buying the disc yet.   My routine:  1. Check for the titles. 2. Check for scratches and defects on the shiny side of the DVD. 3. Have you choice be tested on their DVD player. 4. Haggle. Currently, it\'s P70 ($1.30) per disc. If bundled into three purchases, you can get them for P65 ($1.20) each. 5. Make sure they stamp/mark the DVD title insert, so if you return them, you can argue that it was from their store that you\'ve bought the disk from because if not, they can play that Shylock argument that you \"might have bought it from other stalls.\"   urlLink    An otaku shopping for anime; urlLink     There are lots of films to choose from. Just remember to have the presence of mind on what to buy and a limited budget for your trip, so that you won\'t indulge on impulse buying. It\'s difficult to control oneself once you\'re there, especially if you have been afflicted with the shopping bug.   Many of the films being peddled in Quiapo are:  1. Unreleased Hollywood films      a. copied from a promotional disc     b. copied inside a theater 2. Released Hollywood films (copied from the original DVD) 3. Classics (Hitchcock, Kurosawa, war movies, etc.) 4. Arthouse films (infrequent, scarce supply, rare titles) ex. Criterion collection 5. Asian foreign films (Korean, Japanese)  Raids are frequent in Quiapo. However, if Edu and his cohorts have a scheduled raid, more or less, the sellers have already been tipped off and are extremely vigilant during that particular day. You can see them talking about the impending raid, having boxes readied, and their supplies are not 100% displayed. Besides, if a raid ensues, the news spreads like wildfire and within a span of five minutes, all of the shops are closed. Should you encounter one such raid, then simply head to the nearest convenience store and take cover.  So, be sure to send a postcard on your next trip to Quiapo.      '),(13908,3982753,'02,August,2004','2004-08-02 00:00:00','       Yesterday, I accompanied Dennis to the  National Museum  as a sidetrip in his ongoing vacation here in Manila. We rode the jeepney and went down at the old Congress building. We were surprised to see that the Museum was under renovation. The question is: for how long? We were informed that the exhibits are currently housed in the nearby Finance building. So, as intrepid tourists, we walked over to the new museum and we were surprised at the state of the National Museum, which should be the repository of our culture and heritage.  When you go there, please bring an ounce of sanity because the information desk is as daft as a lobotomized chicken. She smilingly informed us that there are no available brochures or maps as to how the exhibits go. She cannot elucidate if the Spoliarium is on exhibit. Don\'t the employees know what on earth they\'re displaying there. Any well-respected museum should have more foresight than that! If you go to the  Hong Kong Museum , you\'ll be all agog and be swept away in the depth of their exhibits.    urlLink    Look at all those plates!!&nbsp; urlLink     So, we headed first for the San Diego Exhibit. Wow! It was worthy of praise. The exhibit was well made with the ship\'s planks laid on the floor; the cannons and shots all meticulously arranged. The nice thing is, it was informative and interactive- meaning, research was very thorough and the ojects (cannons) can be touched.  And get this- there was no guard around.  Unfortunately, there were a handful of visitors to the museum. This is to show how we appreciate our cultural heritage: by not going to the museums.   Then there were lesser exhibits like the evolution of the Philippines, the exhibit of the Filipino people showcasing indiginous crafts and antiques from the Ivatans, to the Tausugs, from the Igorots to the Mangyans. There were swords, krisses, baskets, handlooms, dresses, cooking pots, utensils, head gears, etc. For school children, this was a perfect place to start immersing them into Philippine history & culture, but for the avid collector & art afficionado, the place is bereft of anything interesting (except for the San Diego wreck).  We were looking for Lunas, Amorsolos, Hidalgos and old Spanish gold and ivory. But due partly to the renovation, only the peripheral specimens were exhibited. I\'m sure, though, that when the old Museum will be opened it will be a prime tourist attraction. But if they shall procastinate and delay the opening of the old one, then as far as I\'m concerned, this will be the last time I will go (lest bring anyone) to this \"replacement\" museum.      '),(13909,3982753,'02,August,2004','2004-08-02 00:00:00','       urlLink     MMM. The Pandora\'s box of goodies!   urlLink      urlLink     MMM... donuts. Seen here are the Peanut Butter frosts and Berry Full Stuffits   urlLink     MMM.... what can I say? These donuts from  Gonuts Donuts  are really really delicious. They are like  Krispy Kreme  of the Philippines. \"Insanely delicious\" is what the product blurb tells us. The dough is soft and moist- similar to those no-bake brownies. There is no lemony taste or the sickeningly sweet sensation in the dough. It\'s not that sweet. And this delight is carried by the yummy taste of the glaze or the filling. The filling is really really good. For example, I love their peanut butter donut where the frosting if reheated tastes like real soft delectable peanut butter. So far that\'s my favorite, though some of the flavors like the Chocohazelnut and the Berryfill really stands out.  The line in  Greenbelt3  is continuous and you can see all those donuts disappear before your eyes. I\'ve heard that at the main branch at the Fort, you need to wait for 2 hours. The turnover is so fast that by the time you reach the counter to tell the cashier your orders, they sometimes will regretfully (but happily) inform you that your chosen flavor is already gone. It\'s ironic because it was only 10 minutes ago that you saw a whole tray of them.  And people cannot stop buying this stuff. It\'s as if the whole world turns on this donut alone. You can see families going to the movies, lovers strolling by, children waiting with boxes upon boxes of Gonuts Donuts. It\'s amazing to see hordes of people gorging and stuffing their faces with these donuts- and still keep a smile.  Is this just a fad? I don\'t think so. Compared to the price of a single donut from  Dunkin or Mister Donut , the difference is  a mere peso . It\'s very affordable, even if the product is coated in exclusivity. It\'s a status of good taste to have them. So rather waste your money on some mass-produced donut that you can\'t even finish a single one, sample and get addicted on Gonuts Donuts. I just can\'t help raving about this stuff. You\'ll never go wrong on this one!  ***oh yeah, I was paid by Gonuts for the article. Psyche!      '),(13910,3982753,'01,August,2004','2004-08-01 00:00:00','       Actually, the SONA of President Arroyo was not the original speech she intended to deliver. It was a hastily written draft made by Dinky Soliman replacing my speech. Damn, that Dinky! She must have sensed that what I\'ve written can capsize this administration. Capsize my foot! So, she crumpled and threw away my 11\"x8\" piece into the paper shredder, and with an evil glint in her eyes, she mocked, \"Now, your nefarious plan, whatever it is, will never ever happen.\" The last thing I remembered was her elephant feet tramping away from the Office of the President.  This was the jist of my original SONA, since I didn\'t save any drafts:  1. To curb the unbridled  population rate , we shall adopt a 4 children policy in which a parent or parents can have a maximum of only 4 children regardless of custody. For which after the fourth child has been delivered, attending physicians are hereby authorized and deputized to  ligate  the mother\'s fallopian tubes so as to prevent unnecessary pregnancies in the future. This will be with consent signed by the husband and by the mother-in-law.   Should the couple have  more than four children  on or after the promulgation of this bill, the parents of the unfortunate spawn can have two options:    a.  put the child into  adoption     b.  the family can raise the child with a proviso for payment of an annual  Nuisance Child Tax  which amounts to 10 times the parents combined income taxes.    c.  the family who cannot comply with the Nuisance Child Tax for a maximum of three years, will force the child to commit  seppuku  with Dinky Soliman acting as a \"second\" and such ritual will be placed under the auspices of the DSWD.  2.  All  public utility vehicles  including but not exclusive to jeepneys, taxis, buses, FXs, calesas, Tricycles, Trisikads, etc., will be confiscated. Sixty percent of which will be nationalized and refranchised for public use while the rest will be disassembled and smelted into recycled steel by the National Steel Corporation.  3.  All  police officers and personnel  shall undergo 2-months physical and skills retraining at specified military installations in compliance with the 32-inch waist line memorandum. Certification from such retraining is obligatory for reentry into the police service. All police officers and personnel shall have a  tamper-proof portable electronic BlackBox  to record all apprehensions of law violators.  This device shall be turned over to an NBI representive who will electronically feed it into the NBI database for future reference.  4.  All  residents of Mindanao  with Muslim descent shall be  transmigrated to designated cities and municipalites all over the Philippines, with the provision of land parcels of equivalent land values of their former residences. Farms, buildings and other Mindanao-based assets shall be confiscated and recompensated by means of high-yielding government bonds of equivalent market value. Each city and town must not have a Muslim population more than 15% of its total population. This act shall be made under the authority of the AFP.   Non-compliance to the act shall be deemed subversive, therefore, subject to forfeiture of any civil liberty and protection of justice.  Liquidation shall be executed through the modus operandi of the newly deputized  Davao Death Squad .   Muslim Lands and assets  remaining in Mindanao shall be put up at a nation-wide public auction under the auspices of Christie\'s and Southeby\'s. Proceeds shall go to government agencies, specifically the DPWH, DOH, DSWD and DAR. Mosques shall be transformed into Christian churches pending auction by the Iglesia ni Christo, Mormons and by the Catholic Church.   5. There will be a creation of a body for the shift from presidential form of government to a federal type. A federal state is currently defined as a state comprising of not less than 3 provinces sharing enthnolinguistic similarities and can generate enough revenue for self-sustainment. Rules and regulations on this matter are still pending.  Current nominations for federal states are: a. State of Ilocandia (Region I + Batanes, Cagayan, Isabela, Kalinga-Apayao, Ifugao) b. State of Bicolandia (Region V + Northern Samar) c. Ilonggo Republic (Region 6 + Palawan, Romblon) d. Republic of Cebu (Region 7 & 8)      '),(13911,3982753,'01,August,2004','2004-08-01 00:00:00','       In Thailand, there is an open market called  Chatuchak . It is famed all over SEAsia and probably all over the world in which tourists, young and old, rich and poor, go there to purchase cheap knock-offs and overuns and to rub shoulders among the natives. It must be one exhilarating experience to enter and be devoured into the cavernous maw where your senses are frequently assaulted with exotic sweaty smells, colorful sights, raucous noise and the temptation to dive into the fray and haggle with the merchants. There your budget is on the red and if your common sense does not reign in your impulse, then your wallet can go bye bye sooner than you think.  An experience in  Greenhills  is probably as near the one in Chatuchak. The only difference is the air conditioning. I observed there that the place was set-up according to the merchandise.  The  atrium  was filled with clothes of every kind- from sarongs to printed T-shirts, from pants to socks.  The place was so jampacked that you have to squeeze yourself right through the crowd. I even stepped on the foot of one of the saleslady- that\'s what she got for wearing open-toed sandals. The  mezzanine  meanwhile is filled racks upon racks of cellphones of every brand and color. This is compounded by Muslim urchins prodding you to buy bootlegged DVDs harking \"DBD, boss? DBD?\" In another corner, there were countless stalls of cheap jewelry- most of which looked like plastic beads anyway. Most of the sellers are Muslims. I don\'t know why that is so though. In Virra Mall side, the whole place is full of CDs, pirated software of any kind, computer parts, and collectibles which can bring anyone with the shopping bug into virtual orgasm. Everything your imagination desires is here.  The whole place looked like a big termite colony. You can see flow and counterflow, traffic sans altercations and bumper to bumper congestion. A claustrophobe nightmare. The funny thing is that there were a number of foreigners haggling over bags and jackets, trying on rip-off Nikes, rummaging through the clothes and rubbing elbows with the Filipinos. It\'s as if they turned our shopping playground into another secret tourist destination.  For adventurous \"National Geographic\" tourists and foreigners out for a bargain, then, this is the best place to be- all in one roof. Some even brought their children. Must be expats. Just imagine them being so sensitive to their travel advisories that they ignored it got enough spunk to go into Greenhills. To think there are Muslims in every stall and that there is a chance that a terrorist can blow up the place or hold them hostage. But like life in the fatalist Philippines, hell may come over  basta  we get to shop.  The government should take notice about this sight, so they can be encouraged to put up more shopping areas emulating Thailand\'s premier entrepreneurial paradise- Chatuchak. If not, I hope the private sectors like travel agencies hopefully shall put the shopping experience of Greenhills on their brochures. If not, then we shall remain in the belly of the beast.      '),(13912,3878306,'30,July,2004','2004-07-30 00:00:00','   I went to my cousinâ€™s funeral. Her name is Maria. I never met her and I never met almost anyone there but there all are my family. I met my great uncle Agusto, My grandfatherâ€™s brother. He had gray eyes. My family genes include a whole bunch of different colored eyes like gray, blue, green, and brown.  Anyway I felt a little odd meeting all these people but it was nice to know hey this my cousins. But like always there either too old or too young for me to approach them. A couple of cousins were crying. No one really knew me and if I had to introduce myself it had to be: The son of my grandfatherâ€™s daughter. Why you ask? Well people only knew you as the head of the family, example some of my cousins introduced themselves as kids of other people who are more above in the family tree which I call the heads. My grandfather and his siblings, which are about ten and they had about ten more kids. Just imagine how many kids that is. This was my momâ€™s fatherâ€™s family. Yes complex like any other family tree. I never knew this side but I was part of it. I knew my momâ€™s mother side Iâ€™m always with them. I know my fatherâ€™s fatherâ€™s family and I even know my fatherâ€™s motherâ€™s side but never my motherâ€™s fatherâ€™s side until know. Zapata, Cepeda, Escoto, and Henriques thatâ€™s what I am. Later we went to Charroâ€™s place. (Charro is a girl; itâ€™s some weird nickname.) They made a lot of food but we already ate. So they drank coffee, smoked cigarettes and talked. Leytta is a family favorite. Marthaâ€™s introduced her to everybody already and beside Leyttaâ€™s pretty so it all works out. I came with a nice sense of the people that no matter what Iâ€™m a part of. Iâ€™m really happy that I met my great uncle Agusto, because heâ€™s my only connection to my grandfather since my grandfather died two years ago and I never met him. Well I met him when I was so little that I donâ€™t remember.  Rest In Peace Maria HenriquesMy beloved cousins whom I would of loved to meet for her charm and loving attitude. The heavens will accept you with out a doubt. Because that is where god wants you  '),(13913,3878306,'29,July,2004','2004-07-29 00:00:00','   Listening to: Black Eyed Peas Song: Lets Get It Started Rating: 5 out of 5 Comments: I just love this song it makes me want to dance but I bet you didnâ€™t know that itâ€™s original name is â€œLets get retardedâ€ and instead of saying â€œIt startedâ€ they say â€œretardedâ€ they probably did that for the video and radio edit.  '),(13914,3878306,'28,July,2004','2004-07-28 00:00:00','   Listening to: Ashlee Simpson Song: Unreachable Rating: 4.5 out of 5 Comments: I really like this song itâ€™s has a really nice guitars and nice drums but with a sad medley and nice vocals.  '),(13915,3878306,'28,July,2004','2004-07-28 00:00:00','   I came from Kevinâ€™s house right now and my mom I think didnâ€™t expect me because she locked the door and she also asked me what time I would be back, I really was unexpected sorry to crash your party. The guy who calls my mom is here at this moment. Itâ€™s all. All right I like him for her. I wish they wouldnâ€™t be so shh shh about it. But what can I do? Anyway the guy doesnâ€™t look me in the eye. He thinks I donâ€™t like him, which I didnâ€™t but heâ€™s All right now. I feel so excited for my mom. Maybe there are just friends but Iâ€™m not sure what they are. I still think itâ€™s all exciting. I like the idea of being a group of three. Anyway I ordered lasagna. Iâ€™ll be all right for now.  '),(13916,3878306,'28,July,2004','2004-07-28 00:00:00','   &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; I slept over at Wanfiâ€™s house yesterday and Helen woke me up and we had to go to my house. But of course there were other stops then home. But on the way there Helen talked about Tarah. I know some things but I donâ€™t know what Tarah feels if I know. Did she want me to know? She didnâ€™t mention it. So you I know. I wish Tarah would of told me but all right. &nbsp;I went to the dentist office today and in two weeks my braces comes off. Iâ€™m so happy. Today they broke the metal things that were on the top back teeth. Drilled those babies right off. And broke the teeth guard on the back of the front teeth. Then I had to get a new imprint. This means tasteless pink stuff pushed on your teeth to make a make this way they can make a retainer. Good thing this time it was only on the top teeth. My mom and me went to Bedford cafÃ© and ate club sandwiches. (Yummy) then we crossed the street to the supermarket. When we got home my I decided to choose my clothes for Walt Disney. I have a nice luggage. 7 shirts 5 pant 5 boxers 1 sneakers 1 sandals 1 bathing suit. Thatâ€™s about it. Since my mom doesnâ€™t let me put on new clothes until the trip she noticed she had too much clothes so I can put clothes Iâ€™m not taking with me. Crappy thing Michael left today to Florida. I mean he could have lived here but he rather roams free in Florida. I talked to Jessica today sheâ€™s really annoying me, Iâ€™m not calling her anymore. I talked to Kristina today instead. I miss talking to Kristina, her and her boyfriend and their problems. Not really problems but anyway we talked for a long time mostly about nothing but it wax nice. Sheâ€™s leaving to D.R. Monday and sheâ€™s a little afraid of planes and turbulences and this is beside the point she doesnâ€™t want to leave her boyfriend but heâ€™s going over there too so I donâ€™t see whatâ€™s the problem. I really like Tarah and Iâ€™m telling her in an email. I think she likes me too  '),(13917,3878306,'24,July,2004','2004-07-24 00:00:00','   After the long wait. (O.K I wasn\'t really waiting but)    I got Ashlee Simpson\'s Cd Autobiography!'),(13918,3878306,'21,July,2004','2004-07-21 00:00:00','   Well today I had an appointment and it went fine last time I had to get (naked) now play the drum music. Anyway It went fine I was told I had a odd heart beat that went ta dum ta dum ta dum ta (whistle) dum. But he said it normal for adolescents. So I\'m alright. Anyway I need to get a blood test tomorrow ad I\'ll be fine. I went to Michaels house today (both actually) and here I\'m, typing. I post my story it\'s forty pages. And well i guess my blogger lookers can see it. I\'ll post it in four sections. 10 pages each just to leave you wanting more. (well it\'s not great but I think it\'s o.k) anyway my highlight of my day&nbsp;I got an Email from Tarah. I\'ll make sure to keep in touch. with her.     !Happy Birthday Kevin!'),(13919,3878306,'20,July,2004','2004-07-20 00:00:00','   Me and My cousins (Micheal, amanda, Quenia, and her boyfriend Randy) went to six flags. first we went to the superman ride after being split up me micheal and the rest. I found Audriana, her sister, her cousin, her mom, her father, kevin, amanda, randy, and quenia all in one ride at the same time and trust me this was not planned or even expected. but it was good. Now time for my ride review!   Superman    Time waiting: 1 hr and a 1/2     Fun:4/5  Fast:5/5 Puke Factor:3/5 (the lower the better) Coftablity: -2/5 (and why you ack I got a cramp and it hurt really bad) Notes : Good and fun hate the lines    Scream Machine   Time waiting: 1/2 hour  Fun:3/5 Fast:3/5 Puke Factor:1/5 Coftablity:3/5 Notes: It\'s allright but really I won\'t go on it twice...   Nitro   Time Wating: 40 minutes (we had fast passes by now)  Fun:5/5 Fast:5/5 Puke Factor:1/5 Coftablity:4/5 Notes: This one is the best ride and trust me we went back for more! It\'s really good I mean  good.  Best ride ever.   Skull Mountian   Time Waiting: 7 minutes (This was a fast pass thing)  Fun:4/5 Fast:4/5 Puke Factor:2/5 Coftablity:2/5 Notes: besides the fact that this ride is just dark and that I had to run from nitro to get there it was alright   Batman the ride   Time waiting: 55 minutes  Fun:3/5 Fast:5/5 Puke Factor:4/5 Coftablity:2/5 Notes: I got a horrible headache after this ride it\'s alright just let this one go and try chiller.  We later saw a show Batman Catfight.   Batman:Catfight   Time Waiting: at 7 pm  Fun:5/5 Fast: well it\'s not a ride Puke Factor:alright I should just cchange my system  New System  Time Waiting: at 7pm Fun:5/5 Realisim:3/5 (They really weren\'t hitting each other.) Action:5/5 Eye Candy:4/5 (So many things happeing at once) Notes: I thought it was good especially the begging I won\'t tell because that destroys the intro but they were lip syncing.  &nbsp;  '),(13920,3878306,'18,July,2004','2004-07-18 00:00:00','   &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp;I went to Ivan and Saraâ€™s House last night and well today we went Belmonte. (Itâ€™s some beach thatâ€™s in jersey but itâ€™s kind of far.) Let me tell you what happened in the ride there. First we were riding relaxing and Iâ€™m looking to the right and there is these two white guys listing to Lloyd Banks â€œFireâ€ and theyâ€™re really into it. Moving there body with it but it was so funny because they act so ghetto but in such an â€œwell you live in your motherâ€™s house and have two cars but stillâ€ ghetto way. Itâ€™s hard to explain letâ€™s just say anyone else would not do that like them. I start to stare and then move my body down and start laughing. They get pissed and stop doing it. Second we were almost there and Ivanâ€™s talking about his job. Ivan does something do to with moving boxes and organizing and stocking Iâ€™m not exactly sure anyway he works with inmates. Well not with inmates but at the end what he piles and moves goes to inmates. And then the radio goes â€œlocked upâ€ we all start laughing. Anyway we get there and we walk a bit. The place is about 3.5 miles long one way. We walk to some point and then Ivan goes by himself and runs in both directions, 7 miles or more. Sara and me go to subway, which is right there, and order sandwiches. Then we take a quick seven eleven grab and get drinks. We hang around the car but every time we were next to the car someone would ask for parking. So we sat in a bench. Sara had to pee and right and that moment Ivan just comes back. We go get tickets to et in the beach. And we chill for a while and eat. There were a lot of condoâ€™s and nice houses. Anyway later we decide to go in the water. It was really wavy. Sara couldnâ€™t go in because well you know why. Sheâ€™s pregnant. (Oh yeah thatâ€™s the pregnant woman from the title duh) anyway Ivan and me got in. the waveâ€™s were so much fun. It would come and we would have to swim away fast it was great. Sometimes we just let it hit us but sometimes they were so big you had to run. After we went back and talked relaxed got tanned and we were deciding what to do but we just decided to go back because it started to rain. I have to admit this was really fun and would die to do it again, even just the same thing all over. (Well Iâ€™d prefer just going again) Sara wants to name the baby Nathaniel if itâ€™s a boy and Ivana if itâ€™s a girl. They donâ€™t exactly know if it is a girl or boy. I really donâ€™t mind what the baby is but Sara said that Elaine had a boy then a girl and so did Ariel. This trend might just happen again. Crappy thing though is that Nathaniel or Ivan is going to be young compare to the rest but the baby always get the most love so its not that bad. Iâ€™m really happy for the grandkids though because they have 5 first cousins and if they have another one and Iseli have one thatâ€™s 7 first cousins. That would be so cool. I have two first cousins. The Lâ€™s Leytta and Lana. Bad thing though is that Iâ€™m so old compare to them. Leytta is 6 and Lana 6 months but thatâ€™s not so bad you get to have small baby cousins. Well when we went back to the house we watched some TV and my godmother called and it she said she was going to take me home in the morning and had to be there by tonight. I left Ivanâ€™s house kind of sad that we didnâ€™t spend more time with him and Sara but I guess it was back to my world. Well I got there I learned we were leaving tonight just because of traffic. Nikki, Her dad and my godmother took me home. We bought pizza on the way back. I noticed something though. When I was in Ivanâ€™s house he said that in his momâ€™s house that there was too much greasy food. And my godmother said that they donâ€™t eat much. Iâ€™m sorry I donâ€™t know whoâ€™s side Iâ€™m on here because my godmother does it wrong but Ivan eats much healthier. O.K I got Ivanâ€™s side. When I got home I said hello to my room. My dad left me some shoes. There is Nike Tennis with deep green color on the side. Thanks Dad. My mom had finished her butterfly puzzle, which itâ€™s oddly shaped like a butterfly. Iâ€™m happy to be home.  '),(13921,3878306,'16,July,2004','2004-07-16 00:00:00','   I have a stomach ache and man it\'s horrible. I\'m just little stressed out. Erik is (pause) really a tough kid deal with... He\'s only six but damn I hope I wasn\'t that annoying when I was young I mean he\'s not that bad but sometimes I can\'t stand him. Anyway I love Ashlee Simpson there I said it. I\'m buying her CD when it comes out July 20. I saw the show and man it\'s a great idea for promoting herself. I already have an idea of which songs she is going to make singles. First \"Pieces Of Me\"&nbsp; then \"Shadow\" and finally \"Autobiography\".&nbsp; We\'re going to this water park place today and tomorrow I\'m going with Ivan and Sara to the mall or beach who knows. (But they are Beach freaks so..) Now that I think about it Duke did a hole bunch of songs for Freq games. He did about 10. That\'s about 10/55. (you do the math on the percent) Anyway I was thinking about Terrah. I think destiny showed me her and was thinking \"Here Steven here\'s someone who doesn\'t like the same music as you doesn\'t have to many things in common but this is the person you will like\" and man it worked. Destiny is grinning saying \"There you go, you\'re likeable you don\'t need Jessica\" and I agree. It was only about four days or so but I really liked her. (I mean I  really liked her)    We ended up going to the pool anyway. I was relaxed and got my swimming skills back (Somewhat) anyway later we went home took baths and got ready to go to Keysburg it\'s a fair like place. Me and my cousin Nicole rode all the good ones but she didn\'t want to ride the one in which you fly up and tit takes you back down really fast. Anyway by then she was sick. There was a arcade there and there were some really craze DDR (Dance Dance revolution)&nbsp;fans. I\'m trying to say a gay guy with extensions a midget with a amazing DDR skill and a freaks tall guy with a funny laugh. I didn\'t play DDR I really want to but it\'s fine. I met this girl on the last ride I didn\'t want to ride by myself&nbsp; so I got to sit next to two girls. She told her friend on the way up the ride the second time. \"I like him more than then the other guy we met\" we were on the Power Ride and if was just a little loose you\'d fly over. It was fun but it made me a little home sick. It reminded me of my friends. I wanted to be there with my friends.&nbsp; I imagined what my friends would say.  I\'m in Ivan/sara\'s house now and we\'re going to the beach tomorrow.  '),(13922,3878306,'15,July,2004','2004-07-15 00:00:00','   They upgraded the blogger tools and now it\'s even better!        &nbsp;   Well first I awoke and watched \"Real World S**T they should of shown\" for the thrid time. Today nothing was planned&nbsp;which was good. We ended up going&nbsp;for a ride around the car. We went to a lawers office to thank her for getting my&nbsp;uncle papers and got mcdonalds. This Mikki D was new and&nbsp;the bathrooms were clean. I ate a ten piece chicken mc nugget meal with some lemonade. (Minute Maid of cource). Getting there was a blur since i was sleepy and needed to pee. Anyway when er got back I got on the computer and registerd myself to Ashlee Simpsons&nbsp;website. (She\'s hot) My&nbsp;cousins I&nbsp;have no clue what they were doing but i was looking&nbsp;some sex info on the net. I like this site they give reliable facts and answer guestions. It was  urlLink  www.teenwire.com     .&nbsp; I went downstairs I became a kid again. I played with my cousins&nbsp;for a while then my cousin\'s neighbor&nbsp;told me his computer was&nbsp;going slow. I went there and twicked it a&nbsp;little to try to fix it. It\'s a great computer&nbsp;it\'s just really slow. First to many downloaded crap. (You now click here free download crap)&nbsp;It has great&nbsp;hard drive but it needs a faster DDR. I guessed it was 256 and i was right he needs at lease 512. &nbsp;      &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; I left around nine and went to play a good old fasion cops and robber.     &nbsp;  '),(13923,3878306,'14,July,2004','2004-07-14 00:00:00','   urlLink \"Lamb Gabriel (Symbion Project Remix)\" Symbion Project   urlLink \"Synthesized\" Symbion Project  '),(13924,3878306,'14,July,2004','2004-07-14 00:00:00','   urlLink     Posted by  urlLink Hello    Here is another version. I\'ll use this one for my profile since the other one looks blurry  '),(13925,3878306,'14,July,2004','2004-07-14 00:00:00','   This is the beat cartoon picture of myself I could get cool huh?  urlLink     Posted by  urlLink Hello  '),(13926,3878306,'14,July,2004','2004-07-14 00:00:00','   urlLink http://illustmaker.abi-station.com/index_en.shtml  '),(13927,3878306,'14,July,2004','2004-07-14 00:00:00','   Some bands I like:   urlLink www.freezepop.com    urlLink www.garbage.com    urlLink www.symbionproject.com    urlLink www.ladytron.com    urlLink www.lifestyleland.com    This is where I get my amplitude remixes:   urlLink www.freqremixes.com    some Bloggers I like to check out:   urlLink http://standingonthebox.blogspot.com/    urlLink http://www.bizstone.com/   &nbsp; some other links:  urlLink www.teenwire.com  '),(13928,3878306,'13,July,2004','2004-07-13 00:00:00','   Today I woke up at around nine I think. We went to Chukie Cheese today (not my favorite place I prefer a place with DDR) I was with my cousin Nikki, My Godmother, Elaine, Erik, Brianna. I had my 360 song CD in my CD player and was looking at my CD\'s auto ranking and it showed me this: 1.\"I Love You\" Faith Evans 2.\"Stupid Girl (Techno Remix)\" Garbage 3.\"Bills, Bills, Bills\" Destiny\'s Child 4.\"Breath\" Finkl 5.\"You know\" ???: It\'s the song that Overnight Celebrity stole the beats from 6.\"Can You Keep A Secret?\" Utada Hikaru 7.\"Everytime I Look\" Blink 182 :I\'m not sure on the title but it\'s the song in American Pie 8.\"Bomb\" Kim Hung Yung 9.\"Queer\" Garbage 10.\"Excuse Me Mr.\" No Doubt  Anyway after we left chukie cheese and thought man it\'s hot lets go to the pool. It\'s been raining all week and the sight of the sun was a beautiful thing. So we went home, changed and went to the pool. Bad luck strikes revenge! The pool had a dead rat in it. The pool staff took it out and threw it away. There was nobody there since most people come only on weekends. We still got in. (Yes in the rat pool.) but let me tell you that it was really cold! But after you wither in it you get use to it.   Bad Lucks not done.  After about an hour and a half it starts to cloud up but good thing we were on the way out. We went home and my cousins friend/neighbor came over. Her names Karina. I think my cousin is really mean to her but I try to tell her not to be so mean and i asked her why doesn\'t she like her she tells me \"She\'s one year younger than me\" I say \"well my best friend is three years older than me.\" she looks away. Oh well i tried right? Then later they get bored so they want me to make a game. I make up a time challenge game with me as a dark voiced host. They had to find three items one was a cellphone, a baby doll, and a spongebob ball. They had 2 minutes to find them. I asked them if they want me to go easy or hard they said hard so that is what they got. Nikki found one item and ran out of time. karina found two items and ran out of time. Erik did the same as Nikki so i had to make a tie breaker. Erik won by finding one item in plain site in 27 seconds. Then I made them spot multiple items of certain colors in the basement and because nikki lost she had to up the stairs. (Not really I just said that with the dark voice) Karina won she had to find 8 Purple items, 12 black items, 6 blue items and 55 green,pink and gray items in 2:34:04 (minutes,seconds,milliseconds) minutes. They got bored again and decided to rent a game. They rented Lilo And Stich for PS and then went to dunking donuts. (really in NJ your always hungry  unlike in NY I was never hungry like this all I want is water in NY but in NJ it\'s food and juice...) I bought the new caramel swirl latte with two French crullers and said latte like ladday. The caramel latte was fine maybe it\'s because I\'m not hooked on Coffee...   I like it here in NJ (and not just the fact of DSL internet) I feel more cofmtable than the city. It\'s surburban and calm but you need a car to go any where. I like just being in a house, but don\'t get me wrong I don\'t want to move here or anything but it was nice just being somewhere different for a while. Oh yeah   my mom passed the drivers test!!!!!!!!    finally she gets a car and I can drive it yeah hurray. O.k enough excitement.  I got the picture with me and terrah. It\'s a nice picture except that i got headphones around my neck...(I forgot...) man I was really happy in that picture. Terrah why do you live in Canada?  '),(13929,3878306,'13,July,2004','2004-07-13 00:00:00','   Well today\'s was a long day. First I ate some Cinnamon Roll things damn they\'re good (I\'m going to be fat...) anyway I later I took a bath to go to the mall. I went to my second home. Hot Topic. I\'m thinking of buying a wallet maybe the Nintendo one or the three chain one who knows.  Second we went to the Skating place YMCA. (Trust me their are a lot of those everywhere including NY) I had fun for the only hour that we (me and my cousin) we\'re there. And they have Dance Dance Revoultion there for two tokens. Tokens are four for a dollar so it\'s only 50 cents can you believe that, in NY it\'s 1.00 to 1.50 what a rip off in the City. I haven\'t skate in a while but I picked up quick and got skating away with only two falls! (This one is a record) the first fall was me trying to avoid a kid. (There was some summer camp there) and the second time I made a sharp turn. (but the fall was fun) and by the end I already mastered the skates once more and was going preety fast and as good as some of the people that worked there.  Thrid We came back ang started downloaded some music and then thought about the Freezepop Remix Contest and I started to look for the Sony Acid to download but all I got was a demo but it was fun I\'ll buy it later.  Lastly I played outside with some neighbors (see this is what I mean by the oldest of the youngest and the youngest of the oldest it\'s really an advantage) I got tired and started to play with Acid program for about two hours and that was my day.  '),(13930,3878306,'12,July,2004','2004-07-12 00:00:00','   Listen to this freezepop song  urlLink Plastic Stars . I just love this song so listen to it.  '),(13931,3878306,'12,July,2004','2004-07-12 00:00:00','   Well What um did I do?...Ph yeah I slept late so give me some time to recuperate.           I woke up at 9 or so and well I was at my cousins house and they woke up because there was thunder but I didn\'t remember so it doesn\'t matter what did matter was that it was raining and it is going to rain the entire week. (Bad luck just likes me). Anyway I tried that spongebob game with the nose as the controler. And then found Final Fantasy 7 and noticed that those graphics are so bad after looking at the recent Final Fantasies. (Man they don\'t even play it they should give it to me... I feel bad for the game... it wants to be played) well it was fine anyway...  '),(13932,3878306,'11,July,2004','2004-07-11 00:00:00','   So I\'m in the Great Garden state of New Jersey. My cousin\'s house right now. My summer has been normally boring but now I see the light called not-boredom. We\'re going to go to the pool later this week. hey Maybe I\'ll get a tan.  '),(13933,3878306,'08,July,2004','2004-07-08 00:00:00','   I got pictures to work!   urlLink    Here\'s My Cousin Ellaine. (Dazzeling!!)&nbsp; urlLink  '),(13934,3878306,'08,July,2004','2004-07-08 00:00:00','   I knew about Garbage for a while but I wasnâ€™t addicted yet. Then amplitude came out (Sequel game to Frequency) and I got more interested in all the artist in the game but garbage had something more appealing (I donâ€™t remember exactly) anyway I went to their site and got quickly hooked. Then I was in Sam Goody and their was Beautifulgarbage and I thought what the hell. And bought it and loved it. Later I got more into old garbage songs and I liked them even better. (The old songs are great!). Garbage is making a new album working title â€œHands On A Hard Bodyâ€ Iâ€™m defiantly buying that when it hits stores.  Here\'s A Pic Of Garbage   urlLink     &nbsp; urlLink  '),(13935,3878306,'08,July,2004','2004-07-08 00:00:00','   I just graduated from middle school and these are my final grades: English: 94 Social Studies: 98 Math A: 92 Earth Science: 97 Music: 95 Physical Edu: 90 FINAL: 94.3  Iâ€™m in an advanced class and I took the regents this is what I got: Math A: 87 (The highest score in my class was 97) Earth Science: 79 (Highest Score In my class)   State Test Scores: (these go by 1-4) three being good and four being great English 3 Math 4  The Awards I got this year are the following: Certificate of achievement In Academic Excellence in 03-04 Certificate of recognition In Science Certificate of recognition for excellence in English Membership for Arista Society Certificate of award Betsy Gotbaum Award (oh this oneâ€™s fancy)   And of course there was a diploma but you can imagine what thatâ€™s like  '),(13936,3878306,'08,July,2004','2004-07-08 00:00:00','   Listening to: Nirvana Song: Smells Like Teen Sprit Rating: 4.4 out of 5 Comments: One of my favorite Nirvana songs and one of their best songs.  '),(13937,3878306,'08,July,2004','2004-07-08 00:00:00','   I heard about Freezepop first in Frequency and fell in love with their electronic song â€œScience Genius Girlâ€. On the game frequency I would play their songs so many times and eventually went to their site and heard other great synth-pop songs. I emailed them and they responded (I was so happy that day!) I still have that email copied to a word file. I continued to hear more of their music and downloaded some. (Sorry Freezepopâ€¦). I always want to buy their CDs but theyâ€™re hard to find. (Theyâ€™re in an Indie Label) Iâ€™ll order them eventually  Here\'s A Picture Of Them!    urlLink     &nbsp; urlLink  '),(13938,3878306,'08,July,2004','2004-07-08 00:00:00','   Listening to: Freezepop Song: Bike Thief Rating: 5 out of 5 Comments: Itâ€™s really a five not because Iâ€™m a Freezepop fan.  '),(13939,3878306,'08,July,2004','2004-07-08 00:00:00','   Listening to: Utada Hikaru Song: Deep River Rating: 4.5 out of 5 Comments: Its soothing sounds are great and peaceful.  '),(13940,3878306,'08,July,2004','2004-07-08 00:00:00','   (My Best friend) Kevinâ€™s Birthday is on the 21 of this month. He made a list of games he wanted and he left it in my house. I got to get him something. Hereâ€™s the list  -Legend of Zelda. The Wind Waker -Onimusha 3: Demon Siege -Resident Evil Outbreak -Onimusha Blade Warriors  -Max Payne 2: The Fall Of Max Payne   -Dead Or Alive Xtreme Beach Volleyball  -Project Gotham Racing 2 -Final Fantasy Crystal Chronicles -DDR -DDR Konamix -DDR Disney mix -DDR Ultra Mix  The ones I highlighted in are the ones I might buy for him. Only one not allâ€¦  Iâ€™m leaning towards an Xbox Game  '),(13941,3878306,'08,July,2004','2004-07-08 00:00:00','   From: fanmail@freezepop.net  View Contact Details   To: \"Steven Zapata\"    Subject: Re: I can\'t find Ultra Funky Fresh...  Date: Fri, 9 Jul 2004 22:33:10 -0400        Hey Steven,  Yeah, I doubt any big record stores are going to be carrying our CDs  since  we\'re just a little indie band. Your best bet would be to order it from  our site  (you can always mail-order it if you don\'t want to do the PayPal  thing), or if  you know of a record store that\'s a bit smaller, you can ask them to  special- order our CD from our distributor, Darla.  And yes, the Duke is Symbion Project.  cheers, Liz E.  ------------------------------------------------------------------ Well freezepop didn\'t let me waiting for long and they E-mailed right back. Only problem is this CD maybe I\'ll money order...  '),(13942,3878306,'08,July,2004','2004-07-08 00:00:00','   My cousin\'s name is Lana. She\'s about 6 months old and she is so cute. Anyway today was her baptism. She cried when they put water on her head but then she fell asleep which was good because we were able to take pictures without her crying. We took a lot of pictures (digital of course...) and then we went to the reception it\'s real close so no worry. They played a bunch of old music but it soon got better and I danced. I talked to Terrah the most. She\'s really cool. Some of the little kids were annoying and I locked one in a room. (Well he started to cry and had to let him go, but you don\'t know what he did so don\'t go cussing me out, o.k I\'m sorry...) Well he started saying perverted things and being a pain for a nine year old... Anyway.        Well I continued to talk to Terrah she\'s actually really nice. (My friend will kill me for using \"nice\") and well I like her ..........But she lives in the great  Canada!  well there\'s not much I can do about that. But it was nice to know that I could like again after the 2001-2004 tragedies. I\'ll get more into detail later about what I\'m trying to say.  '),(13943,3878306,'08,July,2004','2004-07-08 00:00:00','   urlLink    O.K this is Plural who sang a song in the Ampltiude game called \"Cool Baby\" Now I love that song and Plural looks nice in the game but this picture she looks like a hooker&nbsp; urlLink  '),(13944,3878306,'08,July,2004','2004-07-08 00:00:00','   ------------------------------------------------------------------------Date: Sun, 4 Jul 2004 12:59:40 -0700 (PDT)  From: \"Steven Zapata\"    View Contact Details   Subject: I can\'t find Ultra Funky Fresh...  To: fanmail@freezepop.net        Dear Liz,  It\'s zapata2424. (I doubt you remember...) anyway I want to but ultra funky fresh but i can\'t order it and i\'ve tried to find it on tower records but so far no luck...   Do you know where i can buy your cd without having to order it?                                Your fan for ever and ever and ever,                                                     steven zapata   P.S I loved the sampler! and If you could, could you put my new  email on the mailing list. garbagefan2424@yahoo.com (i tried putting freezepopfan2424 but it din\'t work...) oh just one more thing is Duke also symbian project? ------------------------------------------------------------------------ well this is my first email i\'ll share if you want to learn about Freezepop go to  urlLink www.freezepop.com  '),(13945,3878306,'06,July,2004','2004-07-06 00:00:00','   I\'m posting my Emails for you to read so far i sent one to freezepop.  '),(13946,3878306,'06,July,2004','2004-07-06 00:00:00','   I spelled my name wrong somewhere on the main page and i tried fixing it in the Profile but no luck... it\'s  Steven Zapata  '),(13947,3878306,'06,July,2004','2004-07-06 00:00:00','   I made this Blogger so I can Convey thoughts and meet people. So Hiya rest of the Blog community!  '),(13948,3878306,'06,July,2004','2004-07-06 00:00:00','   Well I started my Blog. (What youâ€™re reading right now). Iâ€™ve decided to tell limited amount of people that I started a Blog just people I wonâ€™t see again and a couple of my in touch friends that way they now what\'s up with me when weâ€™re not in touch  '),(13949,3878306,'06,July,2004','2004-07-06 00:00:00','   My mom found a person who teaches music.  I called her today. Her name is Lori. She lives near my house and it is really cool to learn how to play keyboards and violin. I told her my name. She a little expensive but music lessons is expensive. She sounds really cool and she knows to play everything and teaches various things. Anyway here are her rates: 	Musical Instruments: 	1 Hour		$35 	10 Hour	$300 	350 Hour	$350  I asked about multi learning like Violin and keyboards but I wonder if I can just learn both for one month. Now Iâ€™m hooked on it and I want to learn so badly! I hope this works outâ€¦  Iâ€™m calling Lori back at around seven.  '),(13950,3878306,'06,July,2004','2004-07-06 00:00:00','   I was still in my uncleâ€™s house today. I played with my cousin. Her name is Lana and sheâ€™s six months old. Wanfi got a new computer and Iâ€™ve been copying all my CDs to Atrac 3 plus format to fit lots of songs in one MP3 CD. I ate two croissants with some orange passion fruit juice. (Yummy). The tourist wanted to go to the Bronx Zoo, which is a small walk from where we were. I got ready and talked to Terrah and her mom. Her mom is so different from her daughter. On Sunday she passed some man who was smoking a cigar. She went saying â€œBAD BAD BREATH, BAD BREATH, BAD BREATHâ€ and the man wasnâ€™t even that far ahead. He kept looking at the group. Helen told me that they were thinking that the man might throw them something or follow them. Well none of that happenedâ€¦ 	When we left the Bronx Zoo we came with seven and left with six. Josephine was upset and tried and took a shortcut. (Not really a shortcut but whatever) she got home later that day and we went to palisades mall. We didnâ€™t buy much it was more for the tourist. I really like Steve & Barryâ€™s I mean not as much as I like Aeropostale.  '),(13951,3878306,'06,July,2004','2004-07-06 00:00:00','   I did nothing much today. The tourist left real early I reckon (Maybe you got thatâ€¦) anyway I stood in Wanfi base and hung around. We went to Costco I think thatâ€™s how it spelled and well did nothing interesting. Later that day the tourist arrived. (Drum roll please). I talked to Terrah some more and then went to sleep.  '),(13952,3878306,'05,July,2004','2004-07-05 00:00:00','   I went to my uncleâ€™s house. There were tourist there from Canada my aunts family. Let me describe them. Josephine is the oldest. Sheâ€™s quiet and well intellectual. Her husband is Rowdell. (I think thatâ€™s itâ€™s spelledâ€¦) Heâ€™s a web designer. Their kid is Robert heâ€™s like an anime with funny impressions and Josephineâ€™s sister is Rosy. She is really funny. Iâ€™ll tell you some of the things she does later. Her daughter name is Terrah. (Say it like Terra from FF5) Sheâ€™s really different. Quiet and fun. On Monday we went to Liberty Island. (Where the statue of Liberty is) We couldnâ€™t go upstairs to the crown but I still had fun. (Iâ€™ll post pictures later). I tried talking to Terrah during the boat ride there sheâ€™s quiet. It was my first long boat ride. I had been on boat rides but there werenâ€™t that long. That night in my aunt/uncleâ€™s apt I tried talking to Terrah. We talked and told stories sheâ€™s really cool. I kind of like her but I donâ€™t know + sheâ€™s leaving on the sixteen of this month. So forget about it  '),(13953,3878306,'06,August,2004','2004-08-06 00:00:00','   Well today I didnâ€™t do much. You know listen to music, play games and work on my RPG maker game. But really thatâ€™s all I did. I talked to Jessica yesterday we really havenâ€™t been talking. Sheâ€™s mad at me and well I let her be for a couple of days I donâ€™t think she wants me as her friend, I called her earlier today she said sheâ€™ll call me back so Iâ€™m not planning to call her back and I think she wonâ€™t call back. Jessica and my cousin Michael (garbagefanscus.com) are talking and I think he likes her. I mean not in love with her but I just fell badly because I donâ€™t want him to waste his time on her because sheâ€™s not worth his time. Itâ€™s hard to explain but it wonâ€™t work. (Or so I think). On much better news, I told Tarah I like her. (Crappy though since she lives in Canada) but I donâ€™t know if she likes me back. Oh one more thing for people who read my Blogger feel free to send me email too but just label something like â€œAfter I read your Blogger I started to hate the color tea green autumnâ€ something that letâ€™s me know that youâ€™re a reader. Which just to let you know thereâ€™s not a lot of youâ€™s (yes improper grammar) but hey I want to know what you think. Give me advice. I like any feedback. O.K next order of business, Iâ€™m playing Dark Cloud 2 again. Itâ€™s not too bad but let me tell you I really wasted my money on Megaman anniversary collection. Please save yourself and DO NOT I repeat DO NOT get Megaman anniversary collection It was a waste of money.              I had the funniest dream last night. Jessica went to my house and she hated my room and I laughed. (yeah laughed) and her whole family came over and wanted her, her cousins and me to go to some ice skating place. I told them that I didnâ€™t want to go. Iâ€™m not exactly sure what this means but it was kinda weird but hey.              I know Iâ€™m jumping from topic to topic but I canâ€™t just canâ€™t wait for UTADA to make her make here in the States. I just love her so much and now in English Utadaâ€™s even better then ever. If you want to listen to her music go to  urlLink www.utada.com  but the server sometimes doesnâ€™t work. You can also try her Japanese site  urlLink www.utada.jp  and click on now playing to hear â€œEasy Breezyâ€ already a hit in Japan. Her original site is  urlLink www.toshiba-emi.com/utada.com  the first two were for her English CD. Her CD comes out October 8 here in USA. Iâ€™m so excited but if your desperate like me you can download her first English single on iTunes. Too bad I donâ€™t have iTunesâ€¦ ((Weep, donâ€™t worry Iâ€™m (blow nose) fine)).              Iâ€™m trying to get my cousin Alyssa to go to my middle school. I loved it there. Well letâ€™s get some things right. My program I loved, some of the other ones are kinda crazy. Anyway she has to be my protÃ©gÃ© there keep good grades and eventually make it to a specialized high school. One of six, anyone really. I just want her to go to a good school cause the schools around her place (zone) is really bad. But um I think sheâ€™ll get in.  '),(13954,3878306,'03,August,2004','2004-08-03 00:00:00','   Well today I didnâ€™t do much. You know listen to music, play games and work on my RPG maker game. But really thatâ€™s all I did. I talked to Jessica yesterday we really havenâ€™t been talking. Sheâ€™s mad at me and well I let her be for a couple of days I donâ€™t think she wants me as her friend, I called her earlier today she said sheâ€™ll call me back so Iâ€™m not planning to call her back and I think she wonâ€™t call back. Jessica and my cousin Michael (garbagefanscus.com) are talking and I think he likes her. I mean not in love with her but I just fell badly because I donâ€™t want him to waste his time on her because sheâ€™s not worth his time. Itâ€™s hard to explain but it wonâ€™t work. (Or so I think). On much better news, I told Tarah I like her. (Crappy though since she lives in Canada) but I donâ€™t know if she likes me back. Oh one more thing for people who read my Blogger feel free to send me email too but just label something like â€œAfter I read your Blogger I started to hate the color tea green autumnâ€ something that letâ€™s me know that youâ€™re a reader. Which just to let you know thereâ€™s not a lot of youâ€™s (yes improper grammar) but hey I want to know what you think. Give me advice. I like any feedback. O.K next order of business, Iâ€™m playing Dark Cloud 2 again. Itâ€™s not too bad but let me tell you I really wasted my money on Megaman anniversary collection. Please save yourself and DO NOT I repeat DO NOT get Megaman anniversary collection It was a waste of money.              I had the funniest dream last night. Jessica went to my house and she hated my room and I laughed. (yeah laughed) and her whole family came over and wanted her, her cousins and me to go to some ice skating place. I told them that I didnâ€™t want to go. Iâ€™m not exactly sure what this means but it was kinda weird but hey.              I know Iâ€™m jumping from topic to topic but I canâ€™t just canâ€™t wait for UTADA to make her make here in the States. I just love her so much and now in English Utadaâ€™s even better then ever. If you want to listen to her music go to  urlLink www.utada.com  but the server sometimes doesnâ€™t work. You can also try her Japanese site  urlLink www.utada.jp  and click on now playing to hear â€œEasy Breezyâ€ already a hit in Japan. Her original site is  urlLink www.toshiba-emi.com/utada.com  the first two were for her English CD. Her CD comes out October 8 here in USA. Iâ€™m so excited but if your desperate like me you can download her first English single on iTunes. Too bad I donâ€™t have iTunesâ€¦ ((Weep, donâ€™t worry Iâ€™m (blow nose) fine)).              Iâ€™m trying to get my cousin Alyssa to go to my middle school. I loved it there. Well letâ€™s get some things right. My program I loved, some of the other ones are kinda crazy. Anyway she has to be my protÃ©gÃ© there keep good grades and eventually make it to a specialized high school. One of six, anyone really. I just want her to go to a good school cause the schools around her place (zone) is really bad. But um I think sheâ€™ll get in.  '),(13955,3878306,'03,August,2004','2004-08-03 00:00:00','   My cousin\'s blogger  urlLink www.garbagefanscuz.blogspot.com  Sometimes I post things that happenned in the past and post it under the day that it happened so you might see some new post that happen in the past. Get it right? Like I post July 24 on july 27 but july 24 will be on july 24 so you might want to llok back every now ans then  '),(13956,3330628,'14,May,2004','2004-05-14 00:00:00','   Fun of the day: our 2E mangosteen \"gatherin\"... Suckiest thing of the day: eh... dunno... none i tink? Friend of the day: err... WanYi... =)  Yay~ so much things to crap about today.. this morning... we had our Maths Paper2, then after tt which was abt 10am.. we went to tampines interchange... and i got m00dswing again. suddenly felt left out so i tried to left... but i turned back when i \"recovered\"... i\'m glad i did... or i\'ll miss out all the fun! Then we decided to go to Bugis.... Parco.. to shop around... and to pass time... we went to pretty-in-tokyo first to take neoprints, a thing we wld always do whenever we go out... we took the wadeva song machine and it was great this time! All the pics were so nice! Then we chose, decorated... cut... so nice!!! really didnt bear to stick it in mi album... but... i did aniways... so we proceeded to shop ard.... and the shops in edge had quite a great change.. sonme extentions were made and it rocked more... then last time... hahz... a badge which had \"Have a nice day. F*ck someone\" made us laugh lyk dunno wad... then we went to the *ladies*... its so funni thr! Not that we did anything ^weird^, we left and hid in the stairways... then KaiLing was left alone inside the toilet. She tot that we would be scaring her outside the toilet so she let out a fake screeching scream... then when she found out we wasnt there... she so malu!!! betcha everyone was looking! then we show ourselves and she looked quite embarrassed... hahaz... thx lingling... for givin us a gd laugh... hee... then we walk ard as usual... to kinokuniya... blah.. lots o shops... then heArd tt the malcolm oso coming to meet ur there so i quite i mean quite... happii bah. quite depressing too... neways as i kept whining to go back to tampines... we went back... ard 2pm... then we went to tm... 4th floor and tt shaun tan stds butterfly was also thr! in orange tight orh-piank tee... quite \"beng\" la... got dragon one i think? neways... wanyi slapped his head... and he retaliated. as usual... wen and ling joined in the fun. not me. i dun like to beat boys... just got a feeling why they want to lyk... cheapen themselves and let girls bully them? Like kena slapped by girls, kena hit by a sigg bottle.. blah... lots! What can i say is that, dun zuo jian zi ji la. Hahaz... then we saw malcolm... he sponsor ur to take neoprints loh...thank you hoh if u happened to see this... hee... then after tt we went to mangosteen... our hsh...hahz... of cos malcolm tagged along..then he saw our vandalism... =Pthen we saw tian wei, isabel, kang wei, whitney, mei hong, yun wei, eugene, wei zhi, star, lin kong... lots o ppl thr...then the 2E ppls joined us... climbed at mangosteen... blah... lots o fun!!! Then nt much things happenend after tt... as it was near to 6pm we went home... tt\'s all... wow... siaoz.. i wrote alot lei... okie.. i going to end here.. haf fun reading... thanks... bye... =)   '),(13957,3330628,'13,May,2004','2004-05-13 00:00:00','   Fun of the day: water splashing... hahaz... Suckiest thing of the day: some sia lan kia who irritate mi... read on to noe more... Friend of the day: (it\'ll be ur turn soon... different ppl everyday.) Wennie   Wa seh~ today so sway... went to mangosteen today... but halfway we went to the bball court... tt jm and stds butterfly oso there... the suddenly one sia lan kia dunno wad name jason or jack... came ji siao us. Tink he wanna jio wenni or waD ba... then suddenly want sit between us the mangosteeners. we dun want let him so lohz of cos, then he bhb so SQUEEZE and SQUEEZE. I nvr offend him right, then he want to make a joke out of me. Fark la. And he tinks it\'s soooo darn funni. didnt listen much cos i was just trying to control mi emotions, then suddenly he say wad $20 la, wad u want i want. So damn sia lan. So damn pissed off. Then he tinks its funni and laugh laugh laugh. better off dead man. wad? i looK soooo innocent and easy to be bullied isit? Dun judge a person by its appeareance ok. na beh... 16 yrs old very big issit? wish u kena retain for another 5 yrs? see u 21 yrs old still sec 5 so funni HAHAHA. Then i finally cant control my emotions anymore and broke into tears. Then he tio kia loh. serve him rite. Then we proceeded to mangosteen, our hsh. Then stds butterfly and jm come... see i cry lyk nvr see b4 liddat. wa lao got tissue oso nvr give.hahaz... theN we play play lohz... wif lingling bsktball. Then dunno why we start to play water fight. p00r stds butterfly as usual kena bullied by the mangosteeners. so sad. hahaz... then he was so drenched? all invisible? like r(a) like tt... hahz... then not much happened after that... and we played volleyball too... wanyi and butterfly one group, me and wennie one group. then i cant shoot properly one. look at butterfly shoot to wanyi. look at wanyi shoot to butterfly. weird~ but still thanks to Isabel for lending us the ball... =) then we all went home... sorrie all reading this cos i use sooo much vulgaRities. cos today darn buay song!!! hahaz.. signing off here... bu4!z and hope tmr will be a better day.   '),(13958,3546432,'06,June,2004','2004-06-06 00:00:00','   Mirrors are meant to be fractured.   I look at my reflection and fracture it in my mind to what I wish it would be--a flat belly, a small waist. The only part of my body I wish to fracture and re-build.  Would she do that for me?  An operation to fracture and rebuild.  They are out there.  Expensive.  To fracture.  The mind is lost.  I am 47 and wish for a flat belly.  Otherwise, I am content with the reflection in the smooth looking glass.  Content....  '),(13959,3546432,'05,June,2004','2004-06-05 00:00:00','   and then leaving like summer sky azure. No clouds today. Only the hot breath of God making us beg for the cool rain of monsoon. When I think of sky, I think of feathers, I think of wings unfurling pale cat\'s paws claws retracted ready to arch and scratch at moment\'s notice. Give pause to this thought. Then move on.  There is nothing else.  Naught.  None  Non  no  '),(13960,3546432,'05,June,2004','2004-06-05 00:00:00','   Smarty, ah, my dear, what a blow to your beautiful ego--I\'m sure you will still be much-loved by your folks and trainers, as well as your jockey. Smarty, I love you yet...so sad...  '),(13961,3546432,'04,June,2004','2004-06-04 00:00:00','   I shall print out vouchers and itineraries for the morrow. When it is five, I will turn off the computer, check out, and leave. I will drive home and I will not arrive until after six. I may or may not swim. I may or may not drink tequila. I may or may not have a heavy dinner. I may or may not visit with my friend before she leaves tomorrow for a trip. I may or may not sleep. I do not know what I may or may not do; I only do. These are my thoughts.  I am water woman, born near a river, swam in a river, but now there is not water, only desert sand and sky and it is dry and hot and I wilt under the azure sky curved like an angel\'s wing blanketed over her blank face that reveals nothing for she has no eyes nor mouth nor nose only a fresco of emotions that run through her veins as the dark blood runs through mine. Water. How I long for it. To drink it. To pour it over my face and my body. To splash and delight in its essence, its crystals that change under the microscope of desire and heart beatingpoundinggrinding until it one day ends with single bleep, perhaps on a screen, perhaps not.  And thatis how I will end with a single bleep on a screen, miniscule. Once here, now not here.    Gone.  '),(13962,3546432,'04,June,2004','2004-06-04 00:00:00','   and I am sitting here counting down to 5 p.m. I have six hours to go, and six days left to work before a weekend of freedom arrives.  Our friends leave tomorrow and will probably be back next Saturday. We are to feed their kat. They leave us a key and we are free to eat their food, drink their liquor and beer, swim in the pool, and sit in the sun.  It\'s a glorious life.  Thoughts run through my head today like thugs out for revenge. Azul sky (cielo) climbs like a curved wing ever-expanding, unfurling like crescent moon rising over Pichaco Peak....  '),(13963,3546432,'04,June,2004','2004-06-04 00:00:00','   Ugly, ain\'t it? A beautiful Saturday and nothing on....  '),(13964,3546432,'04,June,2004','2004-06-04 00:00:00','   or at least those in cooler climates might view Arizona sunshine to be paradise, but when the temps are well over 100 every day, it seems that Milton\'s Paradise Lost might be a more appropriate way to view the month of June in Tucson AZ. The mornings are pleasant enough--that\'s when I like to go to my friend\'s pool (on my days off, which are few these two weeks--the ole\' dollar calls when bills are to be paid)--anyway, back to the pool, and lie in the sun toasting. Of course I use sunscreen--only idiots from elsewhere refuse to block their tender skin. Five minutes and sizzling may occur. Went over there last night after work and dove right in. No hesitation. 105 I believe the temp was. Had a couple of Coronas with lime, had dinner, and went home, and slept so well that I didn\'t want to rise an shine. And now it\'s Saturday and I\'m at work (not a usual gig, but it\'s okay now and then). I will work until 5, then it\'s to home I go. Perhaps another dip in the pool. Who knows? But it is an hour drive...thank God I\'m moving soon. Only five minutes from work then, but an hour from my friends.  So is this a boring post or what? I don\'t care--it\'s just a way to release some pent-up anxiety and boredom....  Pissed off about the war, but what to do? A little late to pull out--not another Viet Nam on our hands. Blood shed for nothing. Double-speak is what we get. And how dare our government \"confuse\" the plight of Afghanistan with Iraq--how dare Bush link the Taliban with Hussein...sneaky, eh? Not so much, unless you\'re a sheep without brains and, apparently, without wool...  '),(13965,3546432,'04,June,2004','2004-06-04 00:00:00','   That all I can do is sit around and type on my blog. I wonder how I can find others...like that Japanese man who became an internet wanderkind with his blogging interests and so forth....Sure like to talk to some other bloggers...Anyone interested in a dialogue about whatever????  '),(13966,3546432,'04,June,2004','2004-06-04 00:00:00','   Today has been a thug day, not a thought day. And now you ask, what is a thug day? It is a day in which my thoughts are scattered and non-cohesive, like rain water sheeting through a Tucson sky (usually azure). There are days when thought days are worse than thug days because thought days lead to obsessing about tomorrow, over which I no control.   And I am new at this, so revealing too much at one time is not an option.  I wonder how to read others\' journals...  '),(13967,3372184,'21,May,2004','2004-05-21 00:00:00','   urlLink    DANGER!&nbsp; urlLink  '),(13968,3372184,'20,May,2004','2004-05-20 00:00:00','   My sister gave me a copy of  urlLink A Fine Balance  for my birthday and nearly three months later I\'m finally finished reading it (I kept stopping to color in the pictures). So here\'s my review:  In the book, one of Rohintin Mistry\'s characters, Vasantrao Valmik, says:  \"You see, you cannot draw lines and compartments, and refuse to budge beyond them. Sometimes you have to use your failures as stepping-stoens to success.  You have to maintain a fine balance between hope and despair .\"  Well there\'s far too much despair and not a bit of hope in the book. And when the characters try to use their failures as stepping-stones, their hopes are equally dashed. In fact, it seems the book should be titled after the main characters mantra,\"Everything Ends Badly.\"  Mistry is an excellent author and over the course of 500 pages he introduces you to more than a dozen detailed characters and makes you feel for all of them from a place deep in your heart. In fact, I was nearly late for work yesterday, because I had to finish the last 50 pages of the book before I got ready. But it turns out Mistry is a con-man, expertly pulling our heartstrings, then clipping them one by one.  If you like Dickens, you\'ll feel very comfortable with Mistry, but David Copperfield seems like a bedtime story compared to this tragic train wreck. If Mistry had left even a sliver of hope, a half-twisted grin on one of the characters faces, a twinkle in their eye instead of a tear, I would have put the book down satisfied. Instead, I felt forlorn, grief-struck and spiritually and emotionally demolished.  Of course, that\'s probably Mistry\'s goal: to make you feel the sense of desperation and hopelessness that pervades Indian life. His scathing political rebuke of Indira Ganhi\'s term as Prime Minister was an eye-opener to my unschooled mind and his harsh social commentary rung true to scenes I recall from my travels in India in 1995. But a picture of a gray horizon, no matter how masterfully painted, is not a masterpiece in my eyes, unless there is a hint of a coming break in the storm.  '),(13969,3372184,'20,May,2004','2004-05-20 00:00:00','   Having wi-fi internet access has changed my life. Maybe that\'s oversimplifying or maybe I just was that stagnant. Anyway, as a music freak, I\'ve really taken on to streaming internet radio.   The two best sites I\'ve found are  urlLink 91X  and  urlLink WNRN    I grew up listening to 91X in San Diego and I\'ve never found another alternative/modern rock station that\'s quite as good. The DJs know their stuff. They play a good mix of old, new, hard and softer stuff, plus they play a lot of independent and local music as well. My only complaint about their streaming radio is the commercial breaks get tedious because they have generic filler music during local ads.  I just discovered WNRN last year. I was driving from Roanoke to Washington DC and surfing the radio when I found this station playing great songs one after another, after another. And it wasn\'t the same cheesy stuff on most other stations, it was stuff I hadn\'t heard in a while (or had never heard at all). I think this is where I heard \"Do You Realize?\" for the first time. The station turned me on to the Flaming Lips and for that I am forever in their debt.  The best part of WNRN is their a public station, so they never have commercials (though the DJ will throw in a plug from time to time). And it\'s not all rock. The Boom Box plays a great mix of rap and hip-op at nights. Check it out. I see that their Board of Directors is considering getting rid of the audio stream so log on and  urlLink tell them to keep the audio stream    Enjoy!  '),(13970,3372184,'18,May,2004','2004-05-18 00:00:00','   I managed to waste my days off fooling around on this Interweb. Among the more entertaining things I\'ve seen is the  urlLink Sitepal  service, where you type something and can see and here a cartoon say it. Just click on the link, then click \"try now\" and then click \"add audio.\" Enjoy!  I also have been catching up with old friends and found out that not one, but two old college classmates are now engaged. To Marcus and Whitney, congratulation, you\'ve officially made me feel old and unwanted. Truly, I am really happy for them both. It\'s very exciting, but it is weird.  With Whitney I got that same feeling I had back at my college graduation. Where you realize that this particular moment a chapter in your life is completely and fully wrapping up. It\'s happiness, but bittersweet, because of it\'s immense finality.  With Marcus it was just a reminder that the old college days are indeed past. That the guys hanging out, will not be the same, because now there are significant others. But I hope it will be \"the more the merrier.\"  '),(13971,3372184,'17,May,2004','2004-05-17 00:00:00','   I\'m putting you all on notice: I\'m now a Blogger. Believe this!  Why now? Well....  I\'m beginning my fifth incarnation.... a new life, in a new city with a new job and a new chance to make friends, prove myself and take the world by storm.... plus I now have wi-fi internet service and a laptop.  See life, like a book is made up of chapters. I seperate mine by location:  Preface: Montreal Chapter 1: San Diego Chapter 2: Chicago Chapter 3: Arizona Chapter 4: Roanoke, Virginia  and now I\'m beginning to write Chapter 5: Richmond.  I recently began working at  urlLink NBC 12  in Richmond as a news reporter. I knew 3 people when I moved here, and had been to town once. So I have to start fresh.   Enjoy the journey...  '),(13972,3372184,'15,June,2004','2004-06-15 00:00:00','   Do you ever watch \"The Simpsons?\" Do you know that character who plays the flamboyant moustached clerk and always over-stresses the word \"Yes?\" Who is that guy?  Well it turns out he\'s based on an old character actor named  urlLink Frank Nelson . Nelson acted in a ton of TV shows ranging from I Love Lucy to Sanford and Son and was known for that odd effeminate overpronunciation of the word \"yes.\" Why would such a random icon have such a recurring role on \"The Simpsons?\"  Well it turns out the actual Frank Nelson had a recurring role as a clerk in another TV cartoon family: none other than \"The Flintsones.\"  Apparently some believe the character is based on Vincent Price, but that makes absolutely no sense whatsoever. Others say Hal Perry, but I don\'t know who that is and a google search turns up nothing.  '),(13973,3503201,'23,June,2004','2004-06-23 00:00:00','   Beware the storyteller, to often a liar does it conspireto make the soul cold, in lies does it enfold. Let no love shine your essence is not divine, forget, forget, f..o..r..g..e..t.   I have forgotten but I\'ll not believe the echoing lies swarming as flys to a carcass. GET OFF ME! I am free. Yet, I have forgotten, and more lies are begotten. Dead I am yet I am not caught between what was and what will be I am not free for these things exist conceptually only. Yet are very real when I hold on to what\'s gone or dream of dawns to days in which I cannot play. I have forgotten today. Forgotten today? Forgotten today!   Now! I remember the glowing ember that beats and repeats, and heats, it treats me with the gift of rhythm. Remind me of the moment of movement, with out premeditation, pure inspiration, an installation of authenticity, felicity. Bring me back to my center, I enter gratefully dead to what was and what will be, now I write history! Instead of history writing me,  now my future is not just a projection of my past, now I write free at last!   steer clear of me fear for if you draw near the piercing sound of loves laughter shall echo in your ear. Go ahead and lear fear do your judgments bring you cheer. My vision remains clear, your darkness is just the back drop for my light for as I see with loves sight, there is an end to the dark nights.   These times ain\'t about hocus pocus or dooms day locusts it about staying focused on our feelings. To climb past glass ceilings and set minds reeling, peeling back layers of lies that left us lying with flys feeding on the wounds to our truth our life, our love.  '),(13974,3503201,'21,June,2004','2004-06-21 00:00:00','   Aloha, translated literally means: In the presence of(Alo) the divine breath(Ha). This statement is used as hello, goodbye, and I love you. Aloha Now.  '),(13975,3503201,'16,June,2004','2004-06-16 00:00:00','   I soften my gaze to see clearly the haze, blurring formerly dividing lines inspiring rhymes of unity. Where rigidity turns to hilarity where if we just change our viewpoint then we can anoint ourselves with new understanding, demanding different behaviors, because in this instant you\'ve become your own savior.   So rip yourself from the web you\'ve woven. Isn\'t it time you claim youself stolen, from a world that keeps holden, old thought forms passed of as norms unconsciously we conform, awoken by a storm of our own creation, a synchronized place, to taste divine grace, witness the true face, like mace it will burn, illusions it will turn causing us to learn of our connection to life. Break free from the strife.   We\'ve gotta take time to unwind our mind and find the first knot that got us caught in this perpetual tangle. I have come at this from so many angles, callin on all the angels. I have dangled of cliffs, repaired psychic rips, busted of clips into my own ego trips. Flipped scripts until it clicked that life is a circle, no one point on it is better or worse we are all equidistant from the center, the symbolic sun where all life begun and because of which will end only to begin again.   Mental Conflicts keep sending us sailing on a turbulent ocean, perilous waves of pitching emotion, no solid shores of which to be holden. Till we return to the shore that is golden. Turning trash into treasure pain into pleasure. Remember, remember?   Only ways things stay the same is that they change. Duality is just a mind game. Awareness burns through both opposites cop to it, and realize you will that you just swallowed the conscious pill.   Soon your sittin at the souls windowsill free to chill in your Zion I. Becoming a being in the world but not of it, knowing how to love it wholly. Hack back in but being aware of the agents vibing invisible contagents, memes based in fear and illusion, schemes created to continue the confusion, elude them, and transmute them, we will uproot them, We are the ones we\'ve been waiting for. Can you feel the sands shifting are you listening.  '),(13976,3503201,'10,June,2004','2004-06-10 00:00:00','   Well this is just another movement of the mind, another space another time to find my inner divine. Alowing that light to shine and bear witness to the quickness with which it shapeshifts my reality. Transcending duality, the sacred fallacy we are but aren\'t living in.  See there are levels to what is true quite often differing from me to you. On the level of unity no opositions exist, over this truth your mind may trip, for to grasp this you must rip from the illusion that there is something that you are not. In this world so many are caught.   Identifying with the lying of the ages believe the body cages, the soul. But in truth does the soul enfold the physical. Now were getting mystical.   Knowing no boundries it confounds thee, because we identify with another lie that we are our \"mind\". The mind much like a mirror the less is on it the clearer, thus we draw nearer to the light of our source. Illuminating our course. The soul is the horse we are riding, but mentaly denying, and so we start crying feeling disconnected when we could be conntented to be in ease at one with all we see. All we have to do is be.   In line with the souls perspectives guided by divine directives, Illusions we will rout, collecting our clout, we will shout our love from above and below and let people know that peace never left us.  '),(13977,3503201,'07,June,2004','2004-06-07 00:00:00','   Hello...Glad you could be here. What is that wieght your carrying on your shoulders. Don\'t know well, that would be the wieght of the world view you carry. Let go of what you can\'t control, namely everything outside yourself. What is yourself? Remember to breath this too will pass. Remember to breath this too shall come again. Another reminder from the mirror.  '),(13978,3503201,'06,June,2004','2004-06-06 00:00:00','   Another breath I take make no mistake I know where I have faked myself out. Want to scream and shout about how I have been lost and found again. How the mind does bend from enemy to friend.   Where we tend to focus is what we\'ll find flowing to us. So I put my mind in the spaces where love can bloom light can zoom to cast out doom cause the two can\'t occupy the same place.   I give thanks to grace for the clarity to see, allowing me to be free to create consciously. Thank you love for clearing the clouds, illuminating the shroud that is so loud as to drowned out the still small voice.   The on going choice I see I remember I be.  '),(13979,3626700,'23,June,2004','2004-06-23 00:00:00','   i have a lot of stuff on my mind lately that i just have to get done. More like priorities. But i dont want to rush them, because i dont want to grow up to quickly.  But at my age, there becoming a neccessity. I need ot get a job. Ive applied. Its been one day. So im applying at rite aid today. I need a car, but in order to have a car, i need to have my license, and in order to have my license, you guessed it, i need to know how to drive. All these things are starting to become neccessities in my life. Need a job for income so i can buy a car to drive myself places when my mom refuses to, but i need a license to drive, which leads back to the whole driving thing, which mom also refuses. Ive driven once. woohoo. I need more. So i just have all this on my mind today....blah blah blah....So you know what im saying right about now? \"HERE YOU ARE, GOD, TAKE CARE OF THIS FOR ME.\" just have faith is all im saying. Another thing thats bothering me is that my report card hasnt come and i need to know my gpa and my final scores....THERE YOU ARE TOO GOD, DO THAT FOR ME TOO. hehe. FAITH>>>>>>   HAnging on to life by faith,  sam.   '),(13980,3626700,'21,June,2004','2004-06-21 00:00:00','   Ever since about a few hours ago i didnt understand why my father was the way he was, and why he was my father. What did i do to deserve him as my father? Was it just bad karma? Or does things really happen for a reason? Well in my case things really do happen for a reason. There are always lessons behind actions, and lessons that i have to learn at this time in my life, and God never fails with that. He shows me why everything the was it is, and to look beyond whats right in front of me, just by certain actions that take place in ordinary life. This lesson that i learned yesterday took place at church. And lets just say it really hit home.   The subject of the sermon yesterday was Seeing God as a Perfect Father, \"perfection vs. reflection\". Tom, our lead minister,(who is awesome by the way ;) ) gave this very heart-felt sermon which left me in tears from the beginning all the way to the end. Tom talked about how we all want and need our earthly father\'s blessing. We yearn for that. And in my case, i knew from that moment that i would be a wreck. We want our fathers love, blessing, active participation, and encoragement. We also want he\'s spoken belied in who we are and what will will become. Our Perfect Father is the Lord God himself. He is our real father. He makes up for everything our earthly father doesnt have. This is the reason why we have fathers. For the very reason. we need the love , we need the support, and we need their blessing. Its our very instinct that we rely on that blessing to get us through life, to build our self-esteem, to make us know that someone loves us and supports us through life. And when you don\'t get that blessing, life gets tough. And thats what happened to me.   A.W. Tozer states thats \"the most important thing about a person is waht that person thinks about when they think about God.\" This is one of my standards that i live by. I try and put this into everyday life. Applying it to the people i meet and how i can change their view on the father himself. Show them what an awesome God He really is, and show them what he has done in my life alone, and show them what He can do in there life. Show them that the Lord is our Perfect Father, and that his love for us is endless beyond eternity.  \"How great is the love the Father has lavished on us, that we should be called the children of God! And that is what we are! the reason the world does not know us is that it did not know him. Dear friends, now we are children of God, and what we will be has not yet been made known. But we know that when he appears, we shall be like him, for we shall see him as he is.\"- 1 John 3:1-2  There are five types of Earthly fathers models. These models are similar but they have there own distinctive qualities. They are: The empowering father, the abusive father, the absent father, the possessive father, and the performances based father. They are basically self explanitory. My father was a mixture of all really. HE was more a abusive and performanced based father though. The performanced based father is when you have to live up to his standards in order to get his love and his blessing, and if you dont succeed at this, then you get the grunt, the cursing and the abuse. (This is so hard for me to write but i feel like i should be, so therefore i am.) You see my father was always like you show me you can do this, and ill give you my blessing. BUt if you screw up along the way your in for a nasty treat. And basically to him, my whole existence is a screw up. But Tom talked about that too. And he showed us that in the bible it actually says you are not accident, that the Lord planned our existence. Basically the core truth in Tom\'s words is that God is not the reflection of your earthly dad; he is the perfection of your earthly dad. God will always give me the love, the blessing, the forgiveness that im lacking from my earthly father. He will never frown upon nor be distant and angry with me. He knows you even before you were concieved. He made you. To me, this is what i had to hear. I had to know why. why was my father the way he was, why was i his supposive daughter, why did it always feel like it was my fault. Well i know now that all i need from my earthly father is everything i can get from my Perfect Father. I will still keep up everything im doing with my father(e.g calling him, trying to get him to listen to me if i can actually get a hold of him, that type thing..and eventually try and get him to God...it will be a challenge, but im up for it.)  \"Ask and it will be given to you; seek and you will find; knock and the door will be opened for you. For everyone who ask recieves; he who seeks finds; and to him that knocks, the door will be opened. Which of you, if his son ask for bread, will give him a stone? Or if he ask for a fish, will give him a snake? If you, then, though you were evil, know how to give good gifts to your children, how much more will your Father in heaven give good gifts to those who ask him!\"- Matthew  7:7-11  This whole sermon was soo overwhelming but i really, really needed it. I needed to know that the Lord was there for me,(i really dont know how to phrase that, but to me its sounds odd) and that it wasnt my fault for the way my fahter was. I needed to know that i was a punching bag for a reason, that there was all a reason for his madness. I needed that closure that i wasnt a screw up, and that i really am cared for. The qualities of a perfect father are; Everything a Perfect father says is   true. everything the perfect father does, is   right. Everything the perfect father hears, he   understands. Everything the perfect father promises, he   delivers on. The perfect father    isnt going anywhere. He will always always be there for you, no matter what. And not all dads are bad. Many of them fit the father profile. But there are many that don\'t. And there are reasons for that. Just remember the perfect father is always there, with open arms.   \"even if my father and mother forsake me, the Lord will   recieve me.\" psalms 27:10. This passage was the one that hit home with me. It was the closure i needed to get through life. I will always always have that one in my heart. that is the verse that should be publicized around the world for all to hear, Because once i heard it, i knew that all would be alright, for a while. Tom\'s message really got me throughout the whole thing. He ended it with \" The miracle today is that you can leave this place having a BETTER blessing- a blessing from a Perfect father.\"  Tom had found a \"fathers love letter\" online and used it in his sermon. It was a letter from God himself, straight from the bible. Its a \"cry of a Father\'s heart from Genesis to Revelation\". If you want to check it out go to www.FathersLoveLetter.com   There was really a hidden message in it too. Well for some, they didnt see it. But i saw it. I saw it because i think it was only there for me to see. I saw it because it was actually happening at the same time. And it stood out because of that. God purposely did it for me to recognize it. He wanted me to see that there was someone who physically cared for me and that loved me and wanted to be there beside me through my rough times. And I just want to thank her for staying with me, because believe it or not, shes helped  me keep my cool. Just that feeling of someone sitting next to you and caring for you, its calming. It showed me how much that one person cared, and i needeed it at that time. I know my mom or someone would have hugged me or something, but then i also now there would have been a smart comment that would have ruined the moment. And i just needed one of those long, long hugs that last for life, long after you let go. One of those reassuring its alright hugs/i know that was painful type hugs. I want to let that perosn know that i appreciate her for being there with me through my rough times and to let her know that she is cared for on the opposite end of the chain that bonds us together, and that if she ever needs to talk or just someone to listen that im there for that. She is truely loved for life, and not just by me, by many.   I hope this message gets out to all, and that those people who have that rotten relationship with there father, or jsut really need to hear that there is someone else out there that cares for you. God will always always be there for you, through all times. ever.    Loved by my father,   sam.   '),(13981,3626700,'18,June,2004','2004-06-18 00:00:00','   I run my fastest But still get beat i land on my head when i should land on my feet i try to move forward but im stuck in reverse  Why do i keep at it? i won\'t be left behind  The harder im thrown, the higher i bounce. I give it my all, and thats all that counts.  In first place, myself, i seldom find. So I push it to the limits- I wont be left behind  Some people say you can\'t. Some say don\'t. Some simply give up. i reply, I won\'t.  The power is here, locked away in my mind, My perseverance is my excellence. I won\'t be left behind.  Make the best of each moment. The future is soon the past. the more i tell myself this,  The less i come in last.  Throughout my competitions i learned what winning is all about,  a plain, clear lesson: Giving up is the easiest way out.  So every night before i go to bed i hope in a small way i have shined tomorrow is a brand new day and i wont be left behind.    I wrote this poem while i was living at my dad\'s house. It got me through a lot. And sometimes to this day, it still gets me through...piece by piece. but for the little things now. I automatically remember it when im being challenged. From just test to other things that i have to go through in life. But now i have the Lord for my stength to get through. That, and i finally have love, that love that gets you through a lot. The Lord has donnne sooooo much in my life, i am soo grateful to just be alive. He has saved me and made me conquer all my quest. HE can do the same for you to...all you have to do is have faith.    \"He gives power to the weak, and to those who have no might He increases strength. Even the youth shall faint and be weary. And the young men shall utterly fall, But those who wait on the Lord shall renew their strngth; They shall mount up with wings like eagles, they shall run and not be weary, they shall walk and not faint.\" Isiah 40:29-31  Powered By God\'s strength,   Samalamadingdong.   '),(13982,3626700,'16,June,2004','2004-06-16 00:00:00','   I dont really understand what makes my dad the way he is. I think it is that he is lacking a relationship with Christ.(He claims to be catholic, but going to church is a hassle for them and they already claim there going to hell anyways. i dont really understand why he is like that. he supports me in my \"church business\" as he would call it. He doesnt like how i do it every weekend though, how i love it too. HE thinks church is torture.) I can understand that somehow i\'ve hurt him by moving out of his home but i could not take the unholiness and the uptight \"dont laugh your not funny atmostphere, your to hyper for our dry personalities.\" But what i dont understand is that when i call him, he doesnt feel the need to answer. I mean we started talking again after almost a year in february but before then he Refused to talk to me. I felt horrible. when he finally called me on that special valentines day this year, we agreed that our relationship would get back to normal. We even saw each other on that same day. It was a bit uncomfortable but I managed. I felt like the whole time i was being judged on everything, but i just remembered thats how it was at my dads house. So i started out calling him every other day. And that went well in the beginning, little did i know that he was trying to change the person that i have become. Trying to pull the carpet out from underneath my finally stagnant feet. My foundation was nearly constructed, with tidbits here and there willing to be glued for perfection. So, willing to fill in those holes, i started to listen to him, and change back to the person that i was, little by little. Finally when i realized what he was doing to me, i was in awe of what had happened. I had trusted him. He could not see my happiness through the what he was doing.But the remark from him that made me notice what he was doing was this very phrase: You cant hide behind religion. That killed me. I knew from then on he was trying to murder me and use my body for \"the perfect child\" in his mind. trying to tell me i wasnt good enough, too fat, and that i lowered my standards, whatever that means. he didnt notice that i finally loved the person who i had become.they couldnt see the change in personality that i have now. I have the Lord. He makes up the light of the world. My Father doesnt understand that everyone needs to have a relationship with Christ. So one day after he got through \"the list of everything i need to do in order to become the person who i was in a past life\", i just told him; that i didnt like how he was trying to run my life because for once i was on the right road and that i finally found my home in life. I didnt want to change my hair color or lose weight, or kill my personality. I also told him that i wasnt hiding behind my religion, and that it wasnt a religion, it was a relationship with christ. I told him i forgive him for everything he had done to me in the past and that he was only trying to figure out what was wrong with me. and then i told him why i was acting out: I wanted out. I could stand the restraint. I needed God.( i didnt notice it then but i really did need the lord. and i believe he caused me to act the way i was. and i believe he didnt want me there anymore.) He works in extrodinary ways. I knew it then but it didnt really click. The whole time i was being tortured at my fathers house, i was always praying to the Lord to help me and let me find the way out of trouble. Just the little things. Ive never done anything drastic before, so no need to worry about much. just like talking back and disobeying...i never hurt anyone by it...i just needed out.. i was yearning to be loved, and just to be hugged. I was basically in solitary confinment.couldnt talked to anyone in the house without having consequences and the only time they talked to me was when i was \"in trouble\". no hi or bye, no i love you, no hugs, no love. That when i started to believe that the world for me was over and that no one will ever love me again. ( I think that is half the reason why i am the person who i am now.) My Dad and stepmom took me to my \"normal\" counsling appts to try to figure out what was \"wrong with me\". The counselor saw nothign wrong with me. So they decided to talk to my parents(Dad and stepmom, and occasionally my mom) and they\'d basically put words in everyones mouth, meaning not being able to fend for myself because they believed my parents over me. (We\'ve had some good times before all this though, so you know my life before wasnt strictly hell. ) And dont get me wrong i love all these people to death, i just dont see why they wouldnt let me be, just listen to me once in my life. LET ME OUT was all i wanted to say. i couldnt understand why they were refraining me back from this world that i wanted to live in. So anyways, back to the story. They told the counselors that they thought i had a variety of disorders such as OCD(obsessive Compulsive Disorder), Bipolar, ADHD, PDS (Post Dramatic Stress),and on top of that ODD(something difiant disorder), when really....all it came out to be was jsut my personality. I believed them for a little bit until i moved in with my mother and noticed that none of that stuff was really true. I just needed out. So that part of my life was  very restrained. That was a bit of background in my life. But really what im trying to get at now is that he wont answer my phone calls. Everytime i call, the answering machine picks up. And im no good with messages. So one day he answered the phone after a few weeks and told me i was going down hill and all that jazz..(when my grades werent good) and that when i leave a message to not say \"hey its me\" because he has a name. But im not directing the message to just him its basically to everyone. So yeah, im just going to leave it up to God to work on my father and our relationship. Just have to have faith and not rush myself because these things take time right? right. HEy ive notice that this thing is a good counselor. :) maybe i should tell them that i found one....theyd probably put me in for sessions again. ANNNNNYYYWHHHOOO... So im just going to live life the way i have been. Maybe he thinks that not talking to me will bring me down....hmmm... never thought of it that way. I have no clue what really brought this on to write about this but somehow it all just poured out. Dont really know what to do..there isnt much. I just keep calling and praying.   Well...   \" Then Peter came to Him and said,\"Lord how often shall my brother sin against me, and i forgive him, up to seven times?\" Jesus said to him, \" I do not say to you, up to seven times, but up to seventy times seven.\"- Matthew 18:21-22. (great Parable: The parable of the unforgiving servant--Matthew 18:21-35)  Loved by Many,  Sam.   '),(13983,3626700,'14,June,2004','2004-06-14 00:00:00','   Today i had finals. I have been praying and praying all week for them to be easy enough for me to pass and for me to have the wisdom to understand what im doing and to be able to let this week pass by comfortly without stress. I totally put everything in God\'s hands, leaving it up to him to carry me through all of the exams. Well ive gotten through two today, actually 1 4/5 but i can finish that up tomorrow because of my charming witts--and of course her not knowing how long she made the test. but all in all, i feel pretty good about them. there are 2 advantages to Giving everything to Christ to handle and having faith that he will pursue. One- It is A LOT less stressful then holding the world up on your own 2 shoulders. and two- It pleases the Lord.  \"But without faith it is impossible to please Him, for he who comes to God must believe that He is, and that He is a rewarder of those who diligently seek him\"- Hebrews 11:6  The bible talks a lot about having faith and about people who had faith. In the world, you don\'t often see much faith. In fact, some people are negative almost all the time. Any time you presented them with a big idea, they will start rattling off reasons why it wont work automatically. That\'s not faith.  In order to please God, you\'ve got to have faith. You\'ve got to be sure of \"the substance of things hoped for\" and to believe in what you can\'t see. God does give people dreams and ideas; and most of the time they seem bigger than your abilities and resources. He wants you to be absolutely positive that he can make it happen.  Start by believing in the bible. Faith can really grow as you begin to get confident in  His Word. Dont just memorize some verses. Get them into your heart. Try them and let God show you he means business. He\'s good for His Word. Believing Him is crucial, and no one else can do it for you. You have to develop your own faith and determine not to be shaken. You\'ll see it work and Gow will be pleased.  \"Now faith is the substance of things hoped for, the evidence of things not seen. For by it the elders obtained a Good testimony. By faith we understand that the worlds were framed by the word of God, so that the things that were seen were not made of things which are visible.\"- Hebrew 11:1-4  \"behold the proud, His soul is not upright in him; But the just shall live by his faith.\" - Habakkuk 2:4  \"Have faith in God.\"-Mark 11:22  \"Your faith has saved you. go in peace.\" Luke 7:50  \" watch, stand fast in faith, be brave, be strong.Let all that you do be done with love.\" - 1 corinthians 16: 13-14  \"For we walk by faith and not by sight.\" - 2 corinthians 5:7  \"We are bound to thank God always for you, brethren, as it is fitting, because your faith grows exceedingly, and the love of everyone of you all bounds towards each other, so that we ourselves boast of you among the churches of God for your patience and faith in your persecutions and tribulations that you endure.\"- 2 thessalonians 1:3-4  \"Fight the good fight of faith, lay hold on eternal life, to which you were called and have confessed the good confession in the presence of many witnesses.\"- 1 timothy 6:12   Carried by christ,  Samantha.   '),(13984,3626700,'13,June,2004','2004-06-13 00:00:00','   The One Person...   Underneath all your tears there is always that one person who heals your insides they help you mend your open wounds And are there to comfort you through the times those painful severs cant be bonded Even when you think the pain wont go away that person will be with you and will never let you go astray.  With just there soothing trustful words and there serene comedic charisma they help with even the littlest innocent issues  that one person will always be  that one person who will never not be there for you  in any situation  Theres is that one single person  who when all say failure she says success  She will believe in you through every task you will undertake she will always build up your self esteem just by saying hi. for it is such an honor  to have that person in your life  That one person is the one of the two who is or will be  pulling you through  So i pray now for you to find that one person in your life that person will be a godsent.  How do i Know all this?  Because i have found that  person and she is truely  sent from the Lord God Himself.    The person who inspired me to write this poem is that special person. She has been and i know she will always be there for me even when i go off to college and move to austrailia :P(shes probably thinking that will never happen...hahahha) you just have to know me to understand what that comment was about. but anywho. she knows who she is and so therefore i wont mention her name. (IN any case, \"she\" can be replaced with \"he\", but in my life it was a she so that influenced the poem) Till next time....  \"Let not Mercy and truth forsake you; Bind them around your neck, Write them on a tablet of your heart, And so favor and high esteem in the sight of God and man.\"- Proverbs 3:3-4   Keep the faith strong,  Sammie.....   '),(13985,3626700,'27,July,2004','2004-07-27 00:00:00','   well..nothing has really changed...just a bit confused more.... my friend amy had her baby on the 24th @ 5:30 pm (finally...) Went out with \"M\" a few times( good times every times...one ackward time tho shannon falling down the sstars...taht whole day was a wreck...but a good one...) Lava lava kids camp was last week (19th- 23rd) a spirirtually moving experience for all... went out on the 14th...im thinking a bad idea...(ill email you) no job yet...(god must have a reason for this) probably growing up to fast..i dont know..rushing myself...  uh.... thats a bout it basically...on a different computer because mine is crap...  Love ...  me...   '),(13986,3821176,'25,July,2004','2004-07-25 00:00:00','    JULY NEWSLETTER  2004    Hi everyone. It finally feels like summer, a bit too much like summer if you ask me. Itâ€™s been in the 80â€™s which isnâ€™t so bad until you remember thereâ€™s no nice, cool, air conditioned building in which you can find refuge. I wonâ€™t complain too much because Iâ€™ve had a fabulous summer. We havenâ€™t had any visitors this summer so we had to find ways to amuse ourselvesâ€¦..and we did.     MY MISTER MAKES MASTER    Joe isnâ€™t the type to brag and will never toot his own horn (or saxophone) but I will do it for him. He tested for Master Sergeant in March and just found out that he passed and will pin on his new stripes on August 1st. This may not seem big news to you but it is very important to our future. Joe will be retiring in 2 years and our retirement pay is based on the last three years of service. The higher the rank, the higher the pay, the happier we are. So, Master Sergeant Ott has lots to be proud of!     INDEPENDENCE DAY (sort of)    The fourth of July has always been a fun holiday for most people. It usually includes family or friends, barbeques and of course fireworks. This year was the first time in nineteen years that Joe has not worked on this holiday. We were so excited just at the idea of this and had no idea what to do. Just a reminder that we live in Germany. Independence Day is not so important to the Germans. There are celebrations on the military bases but the crowds are unbelievable and we didnâ€™t feel like the fight. So, a friend had a barbeque for those of us that didnâ€™t have any exciting plans. Joe was one of two men there due to the fact that the others were working. He was the barbeque master and we had a fun time even without the fireworks.     SOUP, SALAD AND COFFEE    I recently went out to lunch with some friends from school. All are American except one. As we were placing our orders, this German friend of mine began to giggle when it was my turn.&nbsp; I assumed that I was pronouncing something incorrectly so I just joined in with the giggling and asked what it was that I said that was so funny. She said my pronunciations were perfect but what I ordered was funny to her. I wasnâ€™t getting the joke at all. I had ordered a salad, some soup and water. Apparently, Germans donâ€™t consider this a meal and somehow find it funny that we Americans eat this way quite often. She recently went to a party with a theme. When I think of themed parties I think of a luau or Mexican food or even a costume party. The theme of this party was based on a supposedly American idea: Soup, Salad and coffee. Everyone had to bring a dish that fit that theme. They had great fun with this idea.     MOM, LOOK!    Have you ever been to the zoo and looked at the animals, perhaps the chimpanzees, and pointed at them, gawked at them, made noises at them, etc.?&nbsp; I think everyone has done this sometime in their life. But have you ever wondered what it must be like to that chimp, the one thatâ€™s being pointed at, gawked at, had rude noises directed to him? Probably not. Well, I think we have, indirectly.  Joe and I were recently walking along the fussganger in Kaiserslautern . By the way, a fussganger is simply a pedestrian zone where there are no cars. Itâ€™s fun to walk, shop and sit outside at the cafes having a drink. I heard a boy say something in broken English that shocked me just a bit. He seemed to be about 10 years old, not a toddler whom you might expect inappropriate behavior from. He pointed at one of us, weâ€™re not quite sure which, he gawked at us and he said, â€œMom, lookâ€¦itâ€™s an American.â€ How did he know that? Why was it such a big deal to him? What makes us so spectacular to look at? We get what we call the â€œDeutcher Stareâ€ aka â€œGerman Stareâ€, quite often but this crossed the line. I donâ€™t like to be pointed at, gawked at, nor have rude comments directed at me.   &nbsp;    BMDCVZ    B is for Bavaria  MD is for Munich/Dachau  C is for Castles  V is for Venice  Z is for Zugspitz   This year we decided to stay in Germany for our vacation. We spent most of our time in Bavaria and it has to be one of the most beautiful places in the world. This was a vacation of contrasts. Read on and youâ€™ll understand what I mean.  On our first day we arrived at our hotel at about 1:30pm. We checked, breathed in the clean air and took in the breath taking view of the Bavarian mountains, including the Zugspitz, and headed straight for the Linderhof Castle 30 minutes later. King Lugwig II had this castle (among others) built and it is the only one that was completedâ€¦â€¦after 9 years of construction! We went into the grotto where Wagnerâ€™s Tannhauser played-this was an indulgence on Ludwigâ€™s fantasies about the music of Wagner. One thing that I thought was funny was how to tell if the king was in his castle. He loved peacocks and swans and they were placed all over the castle. If you wanted to know if the king was in the castle you just had to look for two peacocks outside the entrance. If they were out, he was in. If not, he wasnâ€™t. No flags for this king!  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; We still had some time so we stopped into a town called Ettal. This town is known for the Kloster, or seminary. We went into the church and were lucky enough to see the beginning of the mass. There were about 25 monks and they were singing beautifully. We both felt a spiritual vibe while in there. These monks also make their own beer called Klosterbier. We figure thatâ€™s why monks always have such a nice, peaceful look on their faces. In fact, have you ever seen a skinny monk? Probably not, because they all make their own beer. This tradition started many years ago but not for the reason you might think. It was necessary because it had all the nutrients of brown bread but kept much longer without going bad and so the people almost always had beer with their meals.     MUNICH    Our second day was spent on a tour of Munich and Dachau. As we drove through this area, Joe called it the land of the Porsche. Thatâ€™s his favorite car. However, we passed a Mercedes corporate headquarters and every car they make was on display and every car was silver. To me, it looked like a giant matchbox car display. We only had a short time in Munich so we made sure we saw the highlights. We first saw the glockenspiel which has a mechanical marionette type of performance. The clock rang at 11am and then the show started. The story is of how the people were dying from the plague and they were fighting amongst each other. Finally the plague was gone and the people dance in the streets. This is an abbreviated version of the story and may not be exact. I didnâ€™t get to hear it very clearlyJ We had lunch in the famous Hofbrauhaus. As we sat down the waiter very clearly announced to the others that we were Americans. I donâ€™t know why. Perhaps, so weâ€™d get menus in English. Not so. Once again, we stand out. Anyway, the food was good but typical German and the beer was excellent. It was rather quiet at the time which surprised us until we realized that we were in the restaurant and not the beer garden which is where Oktoberfest is held. Oh well, maybe next time.&nbsp;  We went on to Dachau. I never really realized that Dachau is a city. I always associated it with Hitler and the concentration camps. It is actually a lovely city and the inhabitants would like you to know it. Unfortunately, I am among the majority in my ignorance of this city. All I know of Dachau is the horrific events that took place. This was a rather sobering part of the day. I took pictures but somehow found it difficult. It didnâ€™t seem right to be photographing evidence of such horrors. I did take some pictures but Joe and I arenâ€™t in any of them. We had an excellent tour guide who had some insight that others didnâ€™t. He had a friend who was a prisoner in Dachau. In fact, he was a prisoner in two camps and was freed while in a third. He has since passed away but a book was written about him. His name was Marvin and he was Czech. The book is called, â€œThe Last Survivorâ€. Unfortunately, I didnâ€™t get his last name. Iâ€™m going to search online for it. Anyway, Marvin would come to Dachau and interrupt tour groups when he heard the guides sharing information. It seems that the Germans have tried to somehow sanitize the whole thing and have even left out information on the signs. Just recently they have admitted and put on the signs that there were gas chambers but they deny that they were used for mass killings. Marvin would tell you otherwise. He also knew of places where people were told to stand up against a concrete wall where they were shot in the head. Again, the Germans have covered this up by growing vines on the wall, but if you know they are there, you could see the bullet holes in the wall. It was difficult to imagine. Just next to the wall was a ditch. It looked like any other grass covered ditch until you found out it was used as a blood drainage ditch. We wondered why they drained the bodies. They did this because they piled all the dead bodies into a truck and transported them somewhere else. They did not want people to know what they were doing so they drained the bodies of blood so that it would not flow out of the truck when they were driving on the highway!&nbsp; Across from the ditch were some trees that were used for hanging. The prisonersâ€™ hands were bound behind their backs and they were hung by their wrists. If they made any sort of noise, they were tortured even more. The guards would swing them like a pendulum until their arms were torn out of their sockets.  Other things that were cleverly covered were the electric guard fences. This has been landscaped to look like what youâ€™d see growing wild in the forest. There were 30 barracks at the time of this horror but only 2 are left standing. We walked through the gas chamber. I asked Joe how this made him feel and he said guilty. â€œHere I am touring and looking at something these people died in,â€ was his response. We also went into a cell for â€œspecial prisonersâ€. These prisoners were supposedly treated better than the others. Many were doctors and clergymen. This also made us feel weird about walking through the cells. We felt the spirit of what those prisoners must have felt. We saw standing cells. These cells look like upright coffins and the prisoners were bound at the legs so they could not bend their knees. They were unable to move at all. They were kept in these cells for days, weeks and sometimes months. &nbsp;The gas chambers and the ovens were the most difficult to look at. If you saw the picture of the oven now, you might think it was a brick oven for cooking bread or pizza. But the realty is much worse. Many times, the prisoners were hung just to make sure they were dead. The hooks are gone or rather removed by the Germans but the pulley system is still there. We could see how they were hung, killed and lowered to the ground and unto the ovens. Two to three people could fit into the ovens. The gas chambers looked like a large showering area and this is what the prisoners were told it was. But once the gas poured through the â€œshower headsâ€ they knew the truth. It was eerie to stand in such a horrific place and we have no intention of going back but think it was important to see.  We ended that day with a drink on our balcony overlooking the mountains. We needed to wind down from our hectic and emotional day. Following are some quotes from the one and only, JoeJ  â€œIf I were king for a day, this is where Iâ€™d be.â€  â€œLook at this long and hard. Youâ€™ll never see it again when we get back to the states.â€  â€œFirst you see valleys of green and above you see that! Youâ€™d have to be dead not to appreciate this beauty.â€     Neuschwanstein Castle    &nbsp;The Neuschwantein Castle is another of King Ludwigâ€™s castles. This castle is the most famous in Germany and is used in many advertisements. It is known as the fairy tale castle and was Disneyâ€™s inspiration for the castle in Disneyworld. Three castles were built and a 4th was planned but never built. This one was incomplete and had only three floors completed. The Throne Room was decorated in the Byzantine style including crown shaped candelabras and chandeliers, all solid brass with glass cutouts. The throne room was missing one thingâ€¦â€¦ the throne! It never got done. When you looked out the windows you saw incredible view of the Alps, waterfalls and snow.  This took the better part of the day but we still had some time to explore Garmisch, Germany. We did some window shopping when we passed a beer garden and an oompah band. What more does one need?  We also found the Olympic Ice Stadium where Joeâ€™s hockey team had a tournament last year. This was the year that he broke his wrist and was unable to play. We found a professional hockey team from the Czech Republic practicing. This was thrilling for Joe. One of the coaches kept looking at Joe. We think he thought Joe was from another team and was checking out their style of play.&nbsp; As we entered the arena I noticed a remarkable change in Joe. His entire body took on a more upright position. He heard that first slap shot and it was like he was home. &nbsp;As Joe says, â€œI coulda been a contenda!â€     The Zugspitz    &nbsp;The Zugspitz Mountain is the highest point in Germany. We continue with Joeâ€™s therapy on his fear of heights. Why not go to the highest point in Germany and the success of his therapy? Hee, hee. Well, he did quite well. I think the beauty overtook his fear. We drove to a lake called the Eibsee. All the water around here is a bluish-green color due to the high concentration of minerals especially calcium and the fact that it is glacier water. We took a train through the mountain. I do not mean around the mountain. I literally mean through the mountain. This took about 45 minutes and was mostly dark except for the entrance and exit from the mountain. We were let off several feet below the top of the mountain. As everyone was deboarding the train, they were putting on sweatshirts and jackets, preparing for the 32 degree temperature. As we walked outside, we were greeting with an intense orange sun, a perfect blue sky and whiter than white snow. How odd to see snow in the middle of July! Although this is a ski area and the rail runs through it, it remains relatively untouched. We sat on a rock next to a mountain chapel and just took it all inâ€¦.the absolute beauty of nature. I could have sat their all day. It was a perfect meditation spot. We finally resolved ourselves to the fact that sooner or later we would have to move on. We took a 4 minute glacier rail car to the top of the Zugspitz. There were several terraces for taking pictures from different vantage points. Thank goodness for digital cameras! We noticed quite a difference in the temperature. It was about 25 degrees at the top and quite gusty. We enjoyed a nice bratwurst and pretzel. We decided to be monk like and enjoyed a bit of the klosterbier, made by those wonderful, happy and peaceful monks. While we were eating, we were able to see four countries: Germany, Italy, Switzerland and Austria. But the clouds began to roll in and suddenly some of the mountains and other countries seemed to disappear before our eyes.  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; It was time to go back down to the Eibsee via a 10 minute cable car ride. As we looked down we saw the real athletes, hiking up! We thought about doing this ourselves but when we found out it takes 6-8 hours we thought better of it. We were definitely not prepared for that and didnâ€™t have Zoe, the dog, to help pull us upJ     Venice     &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; As I mentioned before, this has been a vacation of contrasts. We saw the beauty of the mountains and the horrors of a concentration camp. We saw the simplicity of the monks and the decadence of King Ludwig II. We saw the greenery of the valleys below the mountains and the snow atop them. On this day, we took a day bus tour to Venice, Italy. Venice is only 5 hours from Bavaria so we decided to take advantage of the tour. This was a marked contrast, simply because we traveled to another country where the temperatures were about 85 with no relief in sight. It was difficult deciding what to wear because it was cool in Bavaria but would be hot in Venice. I decided on a pair of long jean, a tank top and a sweater. I will explain the importance of my ensemble later.  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Venice has no cars in its city. You can only get to it by boat. We took a taxi, a water taxi to the city. The waters were incredibly busy, even scary with all the crazy Italian drivers. Once we were dropped off at Saint Markâ€™s Square we were on our own. Our guide offered to organize a gondola ride through the canals. Although you may have a romantic notion about this, itâ€™s not all itâ€™s cracked up to be. Venice is not a clean city and the odors in the narrow canals can be stifling. We chose not to do it for this reason as well as the high cost and our limited time. Joe has been here so knew certain spots to show me. We went into St. Marks Basilica. Here is where my outfit became important. There was a specific dress code for entering this sacred building. No shorts (men or women), and no tank tops. Several people had to step aside and were given shawls to cover themselves. Hereâ€™s another one of those contrasts. Italy is know for its style and the women are dressed beautifully, if not scantily. They are quite open and comfortable with their sexuality but donâ€™t be a wearinâ€™ that outfit in the churches! They wonâ€™t let you in. The Basilica is in the Byzantine style with mosaic floors as well as the mosaic artwork in the atriums. The ceilings are pure gold. It looks completely different from any other cathedral we have seen and weâ€™ve seen plenty.  &nbsp;Time did not allow for lots of museum tours so we did lots of walking and looking. We saw the Bridge of Sighs which connects Doges Palace with the prison. Apparently Dogeâ€™s Palace takes about a half day to tourâ€¦.maybe next timeâ€¦.when you come visitJ We walked/ran along the Grand Canal. No strolling allowed at this time. We also saw Rialto Square, famous for its market, vendors and shopping. The market had beautiful fruits and veggies and not so beautiful meatsâ€¦.rawâ€¦.with heads and fins and feathers. EEEKKK!!! The outdoor vendors are jammed between the narrow streets. You have no choice but to look. We were literally surrounded by Murano glass, Venetian masks, leather and any other trinkets a tourist might buy. Joe has been here a few times and bought a murano vase for me so this time we bough a mask. Iâ€™m not a collector and wasnâ€™t interest in the real thing but did want something unique to Venice. I found a neat one along with a red leather shoulder bag. Joe goes to the same vendor every time he tours to Venice so he tried to talk the price down but was pretty much ignored. Supposedly, Italians find this insulting.  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; I think I can vaguely relate to mice and rats. Venice is a city of narrow canals and even narrower streets. There are plenty of signs to the cityâ€™s highlights, museums and churches but it can still be confusing, especially if you miss a sign. Even though Joe has been here, he still gets lost. We would be crossing a canal on a bridge, walk down a narrow streetâ€¦ By the way, let me explain narrow. Generally speaking your height is the same as your wingspan (your arms stretched out shoulder height. I am 5â€™2â€ and could spread out my arms while standing in the middle of the street and nearly touch the building on either side. The Rialto market was twice this width which is still quite narrow. Now back to the rats and mice. Weâ€™d take that street which only led on way. So like a rat or mouse in a maze, we followed it around several bends only to find a dead end. Weâ€™d turn around thinking we could retrace our steps only to wonder, â€œDid we turn here or did we go further?â€ Weâ€™d just chuckle when we saw others doing the same thing. After a lime and lemon gelato (Italian ice) we re-boarded the water taxi and the 5 hour bus ride back to Bavaria.&nbsp;   &nbsp;'),(13987,3821176,'23,July,2004','2004-07-23 00:00:00','    urlLink    King Joe at Linderhof Castle.&nbsp; urlLink   '),(13988,3821176,'21,July,2004','2004-07-21 00:00:00','    urlLink    Lunch at the Hofbrauhaus in Munich. Beer and horseradish!&nbsp; urlLink   '),(13989,3821176,'21,July,2004','2004-07-21 00:00:00','    urlLink    Joe\'s therapy continues atop the Zugspitz.&nbsp; urlLink   '),(13990,3821176,'21,July,2004','2004-07-21 00:00:00','    urlLink    Suzanne trying to cool off near a canal in Venice.&nbsp; urlLink   '),(13991,3821176,'21,July,2004','2004-07-21 00:00:00','    urlLink    Snorkeling in Mallorca&nbsp; urlLink   '),(13992,3821176,'21,July,2004','2004-07-21 00:00:00','    MAY-JUNE NEWSLETTER  2004  Hello Summerâ€¦â€¦.Spring?????? Iâ€™m not sure based on the weather but I am sure based on the fact that school is out for the summer. Iâ€™ve already made the most of my summer and Joe has toured to England and Bulgaria and is now back here in Germany. We have plans to go to Garmisch, Germany where the Zugspitz Mountain is in July. That will be quite a contrast to my trip to Spain. Read on if you like.    HEXENNACHT- WITCHES NIGHT  Hexennacht translates to witches night and is similar to our Halloween without the trick or treating. Some villages partake more than others. Our village is tiny yet we still had the parade of children on hayrides partying through the streets. This exciting night also offers the opportunity to have your house, car, trees, etc. toilet papered or sprayed with shaving cream. We have so many kids under the age of five that it is funny to hear them. They get so excited to participate in this silly tradition. You can hear them outside laughing and trying to be quiet as they throw a roll of toilet paper at a bush or ask their parents to help them push the button on the shaving cream can. Itâ€™s all in good fun and the best part is that they clean up after themselves the next day!   NO SHIRTS NEEDED  We recently took our 4th and 5th graders on a field trip to a planetarium and park. Special arrangements were made to have the presentation at the planetarium given in English. I mentioned to some of the other teachers that perhaps we should ask the kids to wear their school shirts to identify ourselves as a group. The German teacher happened to overhear me say this and said, â€œOh, this is not necessary.â€ I asked why only because in my previous experience at other schools we always had name tags at the very least. She explained to me that everyone will know we are together because we are Americans. I still havenâ€™t quite figured out what it is about us that makes us stand out so much in a crowd, but apparently we do. We needed no identifying shirts, nametags or anything else. Simply being American was enough.   LAZY GOATS  I recently was visiting with my landlord in her beautiful garden when I heard our neighbor mumble something. You may remember me telling you about Frau Bauer. She is an 83 year old lady that works constantly. She is always cleaning, gardening, climbing apple or cherry trees and she rides her bike to the market. In fact, sometimes I get tired just thinking about all the hard, physical work she does. She was very kind to my parents when they stayed with her and of course, loves Zoe (the dog). Well, as Karla, my landlord, and I were visiting I heard Frau Bauer say something. She often talks to herself so I didnâ€™t really pay close attention to her. Karla started to giggle and asked me if I understood what she just said. All I got was something about the goats. Karla does have two goats so I still didnâ€™t think much of it. Then she explained that Frau Bauer had just called us two lazy goats because we were not doing any work! I can honestly say that I donâ€™t believe I have ever been referred to as a goat and donâ€™t think I like it. Have you seen a goat lately? They are not the most attractive animals in the world. I then realized that Frau Bauer does not appreciate the finer things in lifeâ€¦â€¦like sitting on a bench overlooking fields and mountainsâ€¦.sipping a cool glass of mineral waterâ€¦â€¦.chatting with a friendâ€¦..possiblyâ€¦.maybeâ€¦â€¦.kinda likeâ€¦â€¦.a â€¦..ummmmmâ€¦.hmmmmâ€¦.okâ€¦â€¦a lazy goat!   SYMBOLS OF AMERICA  When you think of symbols of America Iâ€™m sure you think of the flag, stars and stripes, the national anthemâ€¦.maybe even baseball and hotdogsâ€¦..but do you think of McDonalds, Burger King and Walmart? I donâ€™t. At school, we recently received a warning that terrorists are targeting Americans overseas. The places they look for us is in these American icons and we have been cautioned not to go to these places. I know these are all known as American places of business but do we really have to have our identity defined by fast food and one stop shopping?   GARDEN-RING  As you may remember, our landlords are avid gardeners. They have flowers, vegetables, fruit trees, everything. Karla was working in her garden and they had decided to move some flowers from one place to another. As she was working she was sifting through the ground removing rocks and stones when she hit upon a ring. It was solid gold ring and quite a surprise. She cleaned it off and realized it looked familiar and then it hit her where this ring came from. Reinhold was engaged before he met Karla. When the engagement was broken he buried the ring. This occurred over 30 years ago! None of us is quite sure why he buried the ring and we havenâ€™t discussed it. I think there is more of a story to the burial of that ring but I donâ€™t think Iâ€™ll ever know what it is. The more you know the more you want to know. But I wonâ€™t.   ITâ€™S ALL SET IN JELLO  Itâ€™s all set in jello. This is not what you want to hear when you are trying to keep your job. I was once again on the chopping block for my school. I think I told you that I am now considered permanent and guaranteed a jobâ€¦â€¦world wide. My hope was to stay right where I am. However, that is not to be. Change is good and Iâ€™m getting it. I am going to be teaching at the same school but not 4th or 5th grade. Iâ€™ll be teaching Gifted Education to 4th-8th graders and one class of 5th reading skills. Iâ€™ll be working both ends of the spectrum as far as ability goes. I like change and I like a challenge. Iâ€™m ready for both! However, all of this is set in jello. That is a direct quote from the principal. So no matter what I do in the fall, it WILL be a change.   MALLORCA  Mallorca is an island of the coast of Spain and is sometimes known as Germanyâ€™s 15th state. I went there to celebrate the end of the school year with four other teachers. The Mediterraenean Sea is the most beautiful water I have ever seen. It is blue, itâ€™s turquoise, itâ€™s green and itâ€™s clear. The highlight of the trip for me was snorkeling. I have always wanted to try this but never had the opportunity. We took a short boat ride out to some caves and after a mini lesson, jumped in and saw some incredible sights. I saw an octopus, a sea cucumber, held a sea urchin and more. The plant life is like another world under water. I felt like I was in the middle of a National Geographic special.   Another exciting thing was the hiking along the cliffs of the sea. I had brought some good hiking sandals and just had to check out all those trails. I hiked for about an hour to a castle and found a restaurant at the top. I saw two ladies who must have been 75 years old if not older. They had also hiked up to the restaurant and at 10am were enjoying their first beer of the day. I decided that I liked their spirit and have decided Iâ€™ll be doing the same when I get to be their age. It was inspiring.   We enjoyed the nightlife, sort of. I know you probably think we were party animals, but somehow all that sun took the party right out of us. But we did try at least once. We had dinner then went out to the walking area. No cars are allowed in this area and it is fun to just walk around and see the shops, restaurants and people. The specialty of this area is Sangria so, of course, we had to partake in the culture. We then decided to go into a pub. We found one that was an English pub. We thought it would be fun and refreshing to hear a bit of English. As I mentioned above, Germans come to Mallorca for months at a time and so most vendors speak Spanish and German and not much English. We were ready to be able to understand something if only for an hour or two. We walked into this pub and it was filled with Brits, Irish people and Germans. There wasnâ€™t an American to be found, not even a Canadian was there. So we were listening to the DJ and after a bit, we all looked at each other with question marks all over our faces. I finally burst out with, â€œI know heâ€™s speaking English, but I have no idea what he just said!â€ The others agreed and we just laughed it off, had another Sangria and went home. If thatâ€™s a party animal, then I am oneïŠ   '),(13993,4033738,'31,July,2004','2004-07-31 00:00:00','   As I walk through the valley where    I harvest my grain   I take a look at my wife and realize she\'s very plain   But that\'s just perfect for an Amish like me   You know, I shun fancy things like electricity   At 4:30 in the morning    I\'m milkin\' cows   Jebediah feeds the chickens and Jacob plows...   ( fool)   And I\'ve been milkin\' and plowin\' so long that   Even Ezekiel thinks that my mind is gone   I\'m a man of the land, I\'m into discipline   Got a Bible in my hand and a beard on my chin   But if I finish all of my chores and you finish thine   Then tonight we\'re gonna party like it\'s 1699   We been spending most our lives   Living in an Amish paradise   I\'ve churned butter once or twice   Living in an Amish paradise   It\'s hard work and sacrifice   Living in an Amish paradise   We sell quilts at discount price   Living in an Amish paradise   A local boy kicked me in the butt last week   I just smiled at him and turned the other cheek   I really don\'t care, in fact I wish him well   \'Cause I\'ll be laughing my head off when he\'s burning in hell   But I ain\'t never punched a tourist even if he deserved it   An Amish with a \'tude?   You know that\'s unheard of   I never wear buttons but I got a cool hat   And my homies all I agree I look good in black... fool   If you come to visit, you\'ll be bored to tears   We haven\'t even paid the phone bill in 300 years   But we ain\'t really quaint, so please don\'t point and stare   We\'re just technologically impaired   There\'s no phone, no lights, no motorcar   Not a single luxuryLike Robinson Caruso   It\'s as primitave as can be   We been spending most our lives   Living in an Amish paradise   We\'re just plain and simple guys   Living in an Amish paradise   There\'s no time for sin and vice   Living in an Amish paradise   We don\'t fight, we all play nice   Living in an Amish paradise   Hitchin\' up the buggy, churnin\' lots of butter   Raised a barn on Monday, soon I\'ll raise another   Think you\'re really rightous?   Think you\'re pure in heart?   Well, I know I\'m a million times as humble as thou art   I\'m the pious guy the little Amlettes wanna be like   On my knees day and night scorin\' points for the after life   So don\'t be vain and don\'t be whiny   Or else, my brother, I might just have to get medieval on your heinie   We been spending most our lives   Living in an Amish paradise   We\'re all crazy Mennonites   Living in an Amish paradise   There\'s no cops or traffic lights   Living in an Amish paradise   But you\'d probably think it bites   Living in an Amish paradise   Yuck '),(13994,4033738,'01,August,2004','2004-08-01 00:00:00','   Ridin\' in a bus down the boulevardAnd the place was pretty packedCouldn\'t find a seat so I had to standWith the perverts in the backIt was smellin\' like a locker roomThere was junk all over the floorWe\'re already packed in like sardinesBut we\'re stoppin\' to pick up more Look outAnother one rides the busAnother one rides the busAnd another comes on And another comes onAnother one rides the bus Hey He\'s gonna sit by youAnother one rides the bus There\'s a suitcase  pokin\' me in the ribsThere\'s an elbow in my ear There\'s a smelly old bum standin\' next to meHasn\'t showered in a year Well, I think I\'m missin\' a contact lensI think my wallet\'s gone And I think this bus is stoppin\' againTo let a couple more freaks get on Look outAnother one rides the busAnother one rides the bus And another comes onAnd another comes onAnother one rides the bus HeyHe\'s gonna sit by youAnother one rides the bus YeahAnother one rides the busAnother one rides the bus, ow Another one rides the bus, hey, heyAnother one rides the bus, heyThe window doesn\'t open, and the fan is brokeAnd my face is turnin\' blueI haven\'t been in a crowd like this Since I went to see The Who Well, I should\'ve got off a couple miles ago But I couldn\'t get to the doorThere isn\'t any room for me to breathe Now we\'re gonna pick up more, yeahAnother one rides the bus Another one rides the bus And another comes on And another comes on Another one rides the bus Hey He\'s gonna sit by you Another one rides the bus No No Huh? Yeah Yeah '),(13995,4033738,'01,August,2004','2004-08-01 00:00:00','   I\'m not the brightest crayon in the box  Everyone says I\'m dumber than a bag of rocks  I barely even know how to put on my own pants  But I\'m a genius in France (yeah), genius in France, genius in France  Hoom chaka lakaHoom chaka lakaHoom chaka I may not be the sharpest hunk of cheese I got a negative number on my SATs  I\'m not good looking and I don\'t know how to dance But nevertheless and in spite of the evidence I am still widely considered to be A genius in France, a genius in France, a genius in France  People say I\'m a geek, a moronic little freak An annoying pipsqueak with an unfortunate physique If I was any dumber, they\'d have to water me twice a week  But when the Mademoiselles see me, they all swoon and shriek  They dig my mystique, they say I\'m c\'est magnifique When I\'m in Par-ee, I\'m the chic-est of the chic They love my body odor and my bad toupee They love my stripey shirt and my stupid beret And when I\'m sipping on a Perrier In some cafe town in St. TropezIt\'s hard to keep the fans at bay They say, \"Sign my poodle, s\'il vous plait\"\"Sign my poodle, s\'il vous plait\"Hemenene humenenehimenene homenenePoodle... poodle...Folks in my hometown think I\'m a fool Got too much chlorine in my gene pool A few peas short of a casserole A few buttons missing on my remote control A few fries short of a happy meal I couldn\'t pour water out of a boot with instructions on the heel Instructions on the heel Instructions on the heel  But when I\'m in Provence, I get free croissants Yeah, I\'m the guy every French lady wants And if you ask \'em why, you\'re bound to get this response(He\'s a genius in France! Genius in France!)That\'s right(He\'s a genius in France, genius in France)You know it(He\'s a genius in France, genius in France, genius in France )I\'m not the brightest bulb on the Christmas tree But the folks in France, they don\'t seem to agree They say, \"Bonjour, Monsieur would you take ze picture with me?\"I say, \"Oui, oui\"That\'s right, I say, \"Oui, oui\"\"Oui, oui\"He says, \"Oui, oui\"I\'m dumber than a box of hairBut those Frenchies don\'t seem to care Don\'t know why, mon frere But they love me there I\'m a genius in France Yeah, I\'m a genius in FranceGonna make a big splash when I show up in CannesGonna make those Frenchies scream\"You ze man! You ze man ! You ze man!\"Like a fine Renoir (waa), I\'ve got that je ne c\'est quoi (quoi!)Like a fine Renoir (ooh la la), I\'ve got that je ne c\'est ...Quoi quoi quoi quoi quoi, oo-we-ooQuoi quoi quoi quoi quoi, oo-we-ooBow diddy bow di bow di bow bow diddyBow diddy bow di bow di bow bow diddyBow[snort]I\'m a taco short of a combo plateBut by some twist of fate, all the Frogs think I\'m greatOh, the men all faint and the women screamThey like me more than heavy creamWhen I\'m in Versailles, I\'m a popular guyMy oh my, I\'m as French as apple pie (apple pie)They think I\'m awfully witty, a riot and a  halfWhen I tell a stupid joke, they laugh (haw haw haw haw haw)And laugh (haw haw haw haw haw haw)People in France have lots of attitudeThey\'re snotty and rude, they like disgusting foodBut when they see me, they just come ungluedThey think that I am one happening dudeBowm ba ba bowm ba bowm ba bowmI\'m about as sharp as a bowling ballBut they like me better than Charles de GaulleEntre nous, it\'s very trueThe room temperature\'s higher than my IQBut they love me more than Gerard DepardieuHow did this happen; I don\'t have a clueWell, I\'m not the quickest tractor on the farmI don\'t have any skills or grace or charmAnd most people look at me like I\'m all covered with antsBut I\'m a genius in France (yeah), genius in France, genius in FranceAnd  I\'m never goin\' back, I\'m never goin\' backI\'m never never never never goin\' back home againI\'m tearin\' up my return flight ticketGonna tell the folks back here  where they can stick it\'Cause I\'m never goin\' backI\'m never goin\' backI\'m never goin\'  backThe girls back home never gave me a chanceBut I sho\' \'nuff got them Frogs in some kinda tranceAnd I\'m aware that it\'s a most improbable circumstanceBut \"Great Googily Moogily!\", I\'m a genius in FranceEvery Frenchie that I meetJust can\'t wait to kiss my feetGet in line, pucker up! Tout Suite!Bowm diddy bowm diddy bowm diddyI\'m gettin\' even more famous by the hourI\'m stuffed with pastries and drunk with powerNow they\'re puttin\' up my statue by the Eiffel TowerA little more  to the left, boys, a little more to the leftA little more to the left,  boys, a little more to the leftI\'m the biggest dork there is aliveMy mom picked out my clothes for me \'till I was 35And I forgot to mentionI\'m not even welcome at the Star Trek convention But the Frenchies thinkThat my poop don\'t stinkI\'m a genius in France Say, would you pass the Grey Poupon?Merci beaucoup '),(13996,4033738,'01,August,2004','2004-08-01 00:00:00','   Ooh, my little hungry one, hungry oneOpen up a package of my bologna Ooh, I think the toast is done, the toast is doneTop it with a little of my bologna Never gonna stop, eat it up Such a tasty snack I always eat too much, then throw up  But I\'ll soon be back for my, my, my, yi, yi, wooM-m-m-my bolognaSpreadin\' on the mustard now, show me howSpread it on a litle of this bolognaHopin\' that we don\'t run out, don\'t run outIf we do I\'m sure that I\'ll miss  bologna Never gonna stop, eat it up Such a tasty snack I always eat too much, then throw up But I\'ll soon be back for my, my, my, yi, yi, wooM-m-m-my bologna  M-m-m-my bologna(belch)Goin\' to the market now, market nowI\'m the city\'s biggest bologna buyerWalkin \' down the shopping isles, shopping isles Filling up my basket with Oscar Meyer Never gonna stop, eat it up Such a tasty snack I always eat too much, then throw up But I\'ll soon be back for my, my, my, yi, yi, woo M-m-m-m-m-m-m-my, my, my, yi, yi, wooM-m-m-my bologna  M-m-m-my bolognaM-m-m-my bolognaM-m-m-my bologna '),(13997,4033738,'01,August,2004','2004-08-01 00:00:00','   La-la-la-la-lasagnaYou want-a some-a lasagna magnificoOr a-maybe spaghettiAy, you supper\'s a-ready now, where you go Mama  mia bambinoMama mia bambino, \'samatta you\'Samatta you, \'samatta  you You should-a taste my lasagnaAy, you no like-a lasagnaThat\'s okay tooHow about-a calzoneSome-a nice minestrone, ats good for youHave-a some marinaraHave -a some marinara, I know-a you likeI know-a you like,  I know-a you likeLa-lasagnaLa-lasagnaLa-lasagnaWould you like some-a zucchiniOr-a my homemade linguini, it\'s hard to  beatHave-a more fettucciniAy, you getting too skinny, you gotta to  eatAy, mange, mange(gargling)Ay, you-a pass the lasagnaA-don\'t you get  any on ya, you sloppy pig Have-a more ravioliYou-a get roly poly, a-nice and-a bigLike you cousin LuigiLuigi, Luigi, capisce paisanCapisce paisan, capisce paisanLa-lasagnaLa- lasagnaLa-lasagnaLa-lasagnaHeyHey(laughing) '),(13998,3656035,'16,June,2004','2004-06-16 00:00:00','   Hello everyone. Justin gave me this website to create a blog. he said it was easier then Xanga and ill understand it better so yeah... These past few days have been pretty fun. Ive had to babysit during the day while my parents are at work but its all good. I dont have to do the whole day cause my cousin is here and she babysits ahlf the time which is fine by me. I went mini golfing with Matt yesterday. It was fun but he cheated!! he kept kicking the ball! and he tried throwing me in the pond. *Tearrr* So yeah. Right now im just sitting around being lazy. Waiting for the phone to charge so i can call Byron back so yeahhh... Then im going to the movies to see Shrek 2 with Maria, Matt, Ryan and i think Mike but im not sure... so yeah that should be fun stuff.   Lisa leaves for the rest of the summer in Paris tomorrow... Im going to miss her sooo much! Weve gone back since the 5th grade!! Weve gotten through some pretty tough shit and shes sooo strong. All the memories weve had.. and shes leaving me all summer. whats with that?? hehe no im kidding, im really happy that things are going good for her and shes happy.. thats all the matters. I love her.  So yeah... Life has been pretty lonely.. I met this guy though.. hes really nice and sweet.. i cant wait to just spend time with him and maybe be more then friends but i doubt it cause im like anti-girlfriend material... oh well.  So far this summer me and Maria have been having the best fucking time!! Hanging out EVERY day and just having fun and being us. Summer times too short to be mature so we said screw it and are just going to live it up and not care what people say or think about it. I LOVE YOU MARIA!!  Matts amazing.. what else can i say?? We have been through soo much together and hes helping me so much its not even funny. i love to just lay around outside on the trampoline and just talk to him about everything and anything. he doesnt care what i say, he just tells me what he thinks but doesnt judge me and thats the best part. I have told him more in the past week then i told anyone in my entire life. and the best part is, im not the only one who tells how i feel. he tells me how hes feeling and everything going on with him too and thats just amazing that someone is willing to open up to me.. thanks matt,  i love you cutie.  well im going to go get a shower, get something to eat, give byron a call and then go to the movies.. ill write more later. leave me comments so i feel loved.  -Tyra  '),(13999,3413492,'27,May,2004','2004-05-27 00:00:00','    Joyous is the day that you find out that one of your recently completed projects at work has caused a problem... and that the factory is full of this new \'problem\'.  I think that I am lucky in that this is not causing a real yield hit, just causing some end of line testing to shift.  Still, I had to break out the mad skills to change my crap back to the \'old way\' and now start up some old work that was un-necessary b/c of my previous \'triumph\'.  I still got it, I won\'t need it in a few months when I leave, but I still got it.  On another front  urlLink TheMickster  has joined the ranks.  Props to her.    I can\'t watch the NBA playoffs for any duration.  The thuggery on defense and the travelling on offense is a horror for any purist to watch.   '),(14000,3413492,'25,May,2004','2004-05-25 00:00:00','    When I go back to school, my wife and I will be living apart.  This will be difficult at times, but it is for the better good of our future.  What will be difficult is that we have to divide all of \"our stuff\" up between us.  This basically comes down to who gets the Tivo unit.  She will be getting it most likely b/c I won\'t have DirecTV in my graduate dorm (we have a DirecTV Tivo unit - very nice).  Now I must either buy another or go without.  Anyone who has Tivo would understand that going without is pretty damn hard.  So I will be buying another.  Looking around on the web, I should be able to upgrade whatever I buy with a second or a bigger hard drive.  you of course void the warranty, but cheap Tivo with lots of space = good.  Anyone else with experience hacking Tivo?    This also lead to me thinking about hacking DirecTV, but a quick google search turned up info that it is time consuming, capital intensive, and hacking DirecTV carries more fines than hacking Tivo I\'m sure.    Bottom line, Tivo rocks.  If you ain\'t got it, get it!   '),(14001,3413492,'24,May,2004','2004-05-24 00:00:00','    Since I am leaving my job in the next few weeks to return to school, I am ramping down at work.  My replacement is here and I am bringing him up to speed on big picture issues.  Those of you that have ever left a job or hung onto a job when you knew that you were leaving can probably relate easily.  I wonder if everyone goes through periods of low motivation or am I just a big ass slacker?  There are times when it pains me to read email or deal with day to day issues and other times when I just crank through it all.  It seems a lot like \'senioritis\' but you would think that people mature and push through all of that crap.  Am I slacker or does this happen with everybody?  I have no desire to tell my boss off or anything extreme like that.  My group has been good to me and I will leave on good terms.    Everyone thinks that they work hard and are truly a worthwhile addition to their workplace.  Have you ever asked yourself how easy are you to replace?  All my company needs is another intelligent body and two months of training from me.  The last part could be picked up in three months of work by the new employee with some help from his/her new peers.  I am quite easy to replace as I am sure is the case with 95% of workers.   '),(14002,3413492,'21,May,2004','2004-05-21 00:00:00','    urlLink The Phantom of the Opera  is coming to Indianapolis Nov 10 through Dec 5!  Tickets go on sale June 14th!  Excellent.  I saw Phantom for the first time in Feb 2004 during the Austin, TX stop and it was fantastic.  I\'ve been on a big Phantom kick recently and this show stop will work out great.  Hope to see it during Thanksgiving weekend.  Once again I will get to enjoy the music of the night...   '),(14003,3413492,'21,May,2004','2004-05-21 00:00:00','    What an unproductive weekend.  While I did do some good by volunteering some time on Friday and Sunday with  urlLink The CenTex Starzz girls high school AAU team .  I keep wondering if I was this stubborn when I was in HS.  It feels like the coaches have to continuously push to get the girls to play hard all the time.  Several of them are really good.  Maybe not Div I, but several could get scholarships somewhere.  It is a blast to coach and hang out with some people who know what basketball is all about and more importantly, how to teach it.    I did take a few breaks to catch some flicks this weekend though.  First up to bat was \"The Karate Kid\".  I paid money to watch this crappy 80\'s movie so that I could also hear the fine gentlemen of  urlLink Mr Sinus Theater  tear the movie a new one.  The Mr Sinus folk are much like Mystery Science Theater 3000 except they focus on modern bad movies and they do it all live in the theater as you watch.  Bonus points for them because the theater in which they perform is the  urlLink Alamo Drafthouse .  A theater where you can order dinner and DRINK BEER.  I have seen Mr Sinus rip up Terminator, Top Gun (hilarious) and KK.  Up next is Dirty Dancing.  I\'m counting the days...  I also saw Shrek 2 this weekend.  It was pretty good, but not as amusing as the original in my mind.  LOTS of good one liners and visual gags.  You have to pay attention or you will miss it.  I wonder how many Starbucks cafes were in the movie...  It\'s worth it and I got my laughs worth.  On the future front, I am anxiously awaiting my return to the Midwest where I will start business school at Northwestern\'s Kellogg School of Management in September.  It will be a great change of pace and more importantly it will move my wife and I back to the Midwest where all of our families live.  No more flying for Thanksgiving and Christmas.  Can I get an amen?  The only big task left to complete in Austin is to sell our home.  Keeping the axe to the grindstone is difficult.  I have at most 10 weeks of work left.  Not that I am counting...   '),(14004,3413492,'21,May,2004','2004-05-21 00:00:00','    Uhh, a friend\'s  urlLink blog  had this posted and I had to try it out.  If the shoe fits...    urlLink    Quiz by  urlLink buntz   '),(14005,3413492,'21,May,2004','2004-05-21 00:00:00','    Taking the Nestea plunge and starting a blog...  My wife, KC, will probably just role her eyes and say, \"Hopefully someone out there will finally listen to your babbling.  Someone might even find you amusing.\"  Yes, maybe.   '),(14006,3413492,'22,June,2004','2004-06-22 00:00:00','    These last few weeks before our move have been hectic.  The relocation company was a \"pleasure\" to work with.  It could have gone much faster if people were face to face, but we got to play phone tag for two weeks.  I am now in the process of disconnecting current utilities and setting up new ones.  DirecTV seems to be the biggest challenge (so far).  Disconnecting will be easy, but we will have to go through a slightly different route to hook up new service.  Normally, DirecTV will just coordinate the setup of your dish at your new address.  But, because our new home is in a NRTC area (anti-monopoly type creation) we will need to work with the local service provider.  It doesn\'t look too bad from this end, but it could be deceiving.      On top of all of this utility moving is the address changing portion as well.  Cell phones, credit cards, student loans, car payments...  Ugh.  A web page should be created that is called www.immovingandhereismynewaddress.com.      Thank goodness that we don\'t\' have to pack, load and drive our crap ourselves.  My lovely wife\'s company will pay for all of that :)    I did see Dodgeball this weekend.  It is pretty funny, but not as funny as Zoolander.  See it during the matinee if possible or just wait for rental.  Not a Saturday night flick.   '),(14007,3413492,'04,June,2004','2004-06-04 00:00:00','    Wife has been out of town since Wednesday but of course these would be the busiest two days I have had at work for a while.  I got to enjoy the rest of my \'bachelorhood\' by giving cats antibiotics, running silly errands and finally relaxing by watching an edited replay of the 1993 Indiana @ Penn St Men\'s basketball game.  IU had waxed Penn St (who was \'enjoying\' their first year in the Big 10) at IU by ~50.  Penn St played tough and pushed the then #1 Hoosiers to 2OT.  In the end, the bad team found a way to lose by not scoring when they could push the lead to 2-3 possesions (even though IU\'s two big men - Alan Henderson and Matt Nover had fouled out and Big John Ameci (sp) ruling the paint) and not playing zone defense worth a crap.  The game will always be remembered for a controversial call. A Penn St player was streaking towards the hoop (with pass on the way) Chris Reynolds of IU grabbed his jersey and the player retaliated with an push off elbow.  The official called the push off foul, but did not call / see the jersey grab.  The Penn St coach went ballistic and IU (down by two) tied the game to go to OT when Greg Graham was fouled while shooting a 3.  He made 2/3 and in the end the bad team won...   '),(14008,3413492,'02,June,2004','2004-06-02 00:00:00','    My wife and I headed into Memorial Weekend with good news in that we received an offer on our home.  We made a counter and they accepted.  This is great news and timing as we were starting to get worried about getting rid of the home so we could move worry free.  With that taken care of, we only need to square away the details with the moving company and figure out how we can entertain our two cats for the long drive North.    This weekend is setup nicely.   urlLink Robert Randoph & The Family Band   at  urlLink Stubb\'s  tonight and an engagement party for some friends tomorrow.    I should be able to fend off the wife from seeing the latest Harry Potter movie, but she might have to see it late Sunday night like we saw Shrek (read - no kids...)   '),(14009,3419991,'31,May,2004','2004-05-31 00:00:00','   This has been a busy weekend! On Saturday was Erwin\'s 75th birthday celebration. First, we had a mass in the afternoon for him. The church choir sang as a surprise, and Johannes and I decided to cantor everything about an hour before mass, so we had to learn everything really fast. I\'m discovering that singing during pregnancy is rather difficult because I can\'t take deep breaths. The baby must be squishing my diaphragm. The mass was nice though. In the evening, we went to his party. There were tons of people there, and even more food. There was a buffet with vegetables and seafood and fruit and lots of other stuff. Johannes and I ate so many appetizers that we barely had room for the main course. After that, there was a dessert buffet. It was so popular that we ran out of small plates and had to use dinner size plates. After I finished my dessert, Johannes was still hungry, and came back with a dinner plate filled with dessert. I was surprised because he had already had 2 plates, just like I had. He ate it all though. Then came the entertainment. There were lots of musical things and a couple skits. Ludger, Brigitte, Johannes and I played something together- viola, guitar, recorder and flute. There was a funny skit at the end that Helga thought of. It was about saving time during ironing. It involved cutting up a shirt that Michael was wearing so that only the parts that show when you have a suit on were still there. Then Johannes, Ludger, and Christoph came up at the end, and they all took their jackets off and they all had these strange looking cut-up shirts on. It was pretty funny. I think I have a picture of Johannes in his.  On Sunday was Pentecost, and the choir sang again, so I had to get up early to go to practise. I was really tired from the night before. It was a nice mass, and I was glad to be up in the choir loft, away from the incense. Afterwards, Johannes and I were both feeling pretty lazy, and we took some leftovers from the buffet from the night before out of the fridge, and put them on the table outside and ate as much as we could. There was a lot of food leftover, and not enough people took leftovers with them, so Johannes and I ended up with a lot. It was a nice day, so Johannes and I enjoyed the sun and then went for a swim at his parents\' place.   Today was also a holiday here. It\'s Pentecost Monday. It\'s also our anniversary, so it\'s nice that we are both home to celebrate. We had church this morning, and afterwards went downtown to eat in a Mexican restaurant. The weather isn\'t that nice today, and Johannes was falling asleep on the bus, so we decided not to walk anywhere, so we came home and slept for a bit. It started pouring during that time, so it\'s a good thing we have a nice tenant who brought in our laundry from outside. All my pants are out there at the moment, and I don\'t want to wear a dress to work tomorrow. Later, we ate cheesecake and watched the wedding video and lit our wedding candle. I sang along with everything and did all the church responses too. Johannes was laughing at me. I also had to pause the wedding video because I had forgotten what my wedding shoes looked like, and I had to find them.  So, it\'s been a good weekend. Coming up this week is our first birth preparationn class, my first water gymnastics class, and the new Harry Potter!! I\'m hoping to see it next weekend, but I\'m not sure when it will be availabe in English. I might have to wait until June 15 to see it, but if I\'m lucky, I\'ll be able to see it this weekend. That\'s all for now, I think.   '),(14010,3419991,'22,May,2004','2004-05-22 00:00:00','   Welcome to my first post. When I woke up this morning, I had no idea I would be creating a blog today, but that changed when Hannah came online and talked me into it. So here I am. She has one now too.It\'s at http://hanniverse.blogspot.com/. While making my profile for this blog, I noticed that my website had been deactivated because I haven\'t updated it in so long. I think I will work on it today, because my name is wrong and the picture of me isn\'t very good.  Time for a pregnancy update. Everyone asks me about that, so it must be interesting.I\'m 30 weeks pregnant now, which means there are 66 days, or 10 weeks until the baby is born. It sounds like a lot, but I think it will go by fast. I only have 3.5 more weeks of work, and then I can start organizing the mass of baby stuff which is taking over the basement.This morning, I organized the cupboard full of baby stuff in our rooom because it wasn\'t shutting  properly. Yesterday, I bought 4 cute baby outfits for a really good price. I hope the baby is a girl because they are all girly. On Monday, I have my doctors appointment, and I will be getting an ultrasound, so hopefully I will find out the gender this time. I also have to get a shot because my blood type is negative. Other than that, I have no idea what will happen. Hopefully my iron level has improved.  Let\'s see...what else is happening? Next week is Pentecost and Erwin\'s birthday party. Johannes, Ludger, Brigitte, and I are playing some music for that. My flute is broken, so I have to borrow one for next week. I hope it\'s not too different from mine. Our wedding anniversary is also coming up soon, and it happens to be a holiday here (Pentecost Monday), so Johannes and I are going to go out somewhere to eat, then do something that we haven\'t decided on yet, and later watch the wedding video. Then it will be June. The first week is exciting because we start our childbirth preparation class, I start water gymnastics for preegnant people, and the new Harry Potter movie comes out on the fourth. I stop working on the 16th, and then Ludger and Brigitte get married on the 26th, and soon after that, it\'s July. Two very exciting things are happening in July. Well, one for sure. Anne is coming to visit on the 26th. Or the 27th. I have to check. The baby will be born too. She\'s due on July 28th, but she might decide to come in August, or earlier in July. We\'ll see.  The cats went to the vet this week. It was the first time they ever met a German vet. She lives on our street, and a few days ago, we found 2 ticks on Sara and couldn\'t get them off, so Johannes went down the street and rang her doorbell, and she told us where her office was. We went the next day, and the cats got vaccinated and Sara got his ticks removed. They didn\'t enjoy it much, but they did enjoy the free food sample that the vet gave them. We also bought tick/flea stuff that they get every month, and a tick remover.  Anyway, that\'s enough for now. I\'m hungry and lunch is almost ready.   '),(14011,3419991,'13,June,2004','2004-06-13 00:00:00','   It\'s been a pretty quiet week this week. My cold still isn\'t gone, so I\'ve been feeling lazy. The weather was really hot for most of the week too, which only increased my laziness. It got so hot that Johannes dragged the extra bed up from the basement, so we have 2 single beds to sleep on, so that we don\'t have to sleep in our extremely hot loft. The only problem is the beds are right next to the window, and as soon as it gets light, I wake up and can\'t get back to sleep. I don\'t really like waking up before 5. The curtains don\'t seem to help much.It cooled down on Friday though, so it is possible to sleep in the left until it heats up again. I think it\'s supposed to be hot tomorrow. The kids at work will like that, because we turn the water pump on for them when it\'s really hot, and they play with water and sand and make a huge mess. They really enjoy it. I only have three days of work left. Wednesday is my last day. I\'m looking forward to sleeping longer in the morning and having time to organize stuff around the house and to actually have time to cook interesting things.  Let\'s see...what else did I do this week? Wednesday was childbirth prep. We practised sitting on the birthing balls and breathing. I didn\'t enjoy that much because she kept saying to breathe through my nose, which wasn\'t possible and I was supposed to sit so far forward on the ball that I could barely keep my balance. We saw some pictures too, and they were black and white, and illustrations and not photos, so Johannes couldn\'t find anything to complain about. They all had english captions, which was helpful so I had some idea of what the teacher was talking about. There are still a lot of medical words I don\'t know and Johannes doesn\'t know how to translate them all. We also did a couple of relaxation exercises, but I kept relaxing too much and falling asleep. I don\'t think that will work during labour, so I should learn how to relax and stay awake at the same time. Thursday was Corpus Christi, which is a holiday here, so I didn\'t have to work. It was a very hot day, and we had mass outside, and that was follwed by a procession around Marienborn with lots of singing. I skipped the procession though, because it was very hot and my legs were hurting. Poor Johannes had to wear an altar server robe and carry a heavy speaker. He invited all the altar servers to his parents\' house in the afternoon for a swim, since they were away for a day, but it started pouring 10 minutes before everyone was supposed to come, so no one came. There was thunder and lightening, and it was a really cool storm.Johannes and I went swimming by ourselves, and it was fun to listen to the rain pounding on the roof while swimming. The lightening made Johannes a little nervous though. On Saturday, I finished up the baby shopping. My bus pass expires on Thursday, and I won\'t be getting another one. Since it takes me forever to decide on baby stuff, I decided I should finish it before my bus pass runs out. The only thing left to buy now is cloth diapers, and Johannes and I are planning to get those on Thursday. I haven\'t done anything interesting today, only a lot of eating and being lazy. Anyway, time for bed now because I have to work early tomorrow.   '),(14012,3419991,'06,June,2004','2004-06-06 00:00:00','   I went to the doctor today, and everything went pretty well. My iron is still a bit low, but at least my thyroid is normal. I had an ultrasound to check how the baby is growing, and she\'s grown a lot in the last 2 weeks. She\'s only 1 week too small now, which is pretty good. She weighs about 3.7 pounds. I have 2 new pictures, which I will scan and put on my website soon. Not much else is new. I only have a week of work left.I\'m looking forward to having time off so I can get all the baby stuff together. It\'s really hot today. It\'s about 32 degrees, or maybe more. I have to be outside a lot at work, and there isn\'t much shade, so I get really hot. Johannes and I went swimming at his parents\' place after work, but the pool was over 30 degrees as well, and it felt like a bathtub. Time to go cook now.   '),(14013,3419991,'05,June,2004','2004-06-05 00:00:00','   The new Harry Potter movie came out this week. After weeks of waiting, it\'s finally here! It came out in Germany on Thursday. There was actually a showing right at midnight, but that wasn\'t very practical since I had to work the next morning. I couldn\'t go Thursday night either, because I had to go to my water gymanstics class. So,we went last night, which was Friday. It was my first time ever in a German movie theater. The one we went to is the newest one in town, and the biggest. It has at least 10 theaters in it, and I think it\'s bigger than Silvercity in Victoria, but I\'m not sure. They sell lots of food there, and there\'s even a restaurant so you can eat before you see a movie. Right before the movie starts, someone comes in and sells ice cream so you don\'t have to eat it before the movie starts. The most exciting thing about the theatre itself was the chairs. They are huge and comfortable, and the best thing is they aren\'t all individual with hard plastic arms. Some are big enough for two people, and the arms are all soft, and the back of the chair goes all the way up to your head. They were really comfy. The screen was really huge too. German popcorn is really weird though. You can buy either sweet, which is sort of like caramel popcorn but not really, or salty. I wanted butter for mine, and no salt, but that didn\'t work. There was about half an hour of advertisements before the movie actually started, and after they were done, they turned all the lights on again, and the ice cream person came in, and then a few minutes later, the movie actually started. There was a break in the middle of the movie. It was kind of annoying because I wanted to keep watching, but it wasn\'t too long, and it gave me time to go to the bathroom without missing everything, so it wasn\'t so bad. Very practical for pregnant people.We saw the movie in English, mainly because we could get better seats there than in the German one. Another neat thing about German theatres is that you can buy the tickets ahead of time, and pick your seats. Also, you can buy a booklet with 5 movie coupons and use them whenever you want and it\'s cheaper than just buying two tickets whenever you feel like going. I really liked the movie.It went really fast, and they left out a few things I wanted to see, but it was still good. I think the ending especially went too fast, because I think if I hadn\'t read the books, I would have been a bit confused about some things. I won\'t say too much more because I don\'t want to ruin for anyone. After the movie, we had to wait forever for the bus, and met this drunk guy who was complaining to us about a dying pigeon that he found and he wanted to give it to us. Then the bus finally came, and I was watching all the crazy drivers out the window, and this one car turned a corner and cut the bus off, and we nearly ran into it. Mainz at that time of night is a little too exciting for me, I think.  The movie was probably the best part of my week, but I did do a couple of other interesting things. One was getting a bad cold, which is more annoying than exciting. I\'ve been all stuffed up since Wednesday,and my throat\'s sore too. I stayed home from work on Thursday because my head hurt so much that I didn\'t want to be around screaming kids. I went back on Friday, but I wasn\'t feeling all that much better, but I don\'t have all that many days of work left, so I don\'t want to miss that much.   On Wednesday, Johannes and I went to our first birth preparation class. It was a bit hard for me, because there were lots of new German words that I didn\'t know, and we had to do strange exercises and I couldn\'t figure out what I was supposed to do, but it was interesting. I  think Johannes liked it too, except for when we were doing all the stuff where I had to lean on him. I think I\'m a little too heavy for that. We also practiced breathing, which worked pretty well, except that I couldn\'t breathe through my nose. On Thursday, I went my first water gymnastics class. I really liked it. The swimming pool was super warm, and it feels really good to be in water. I feel so much lighter there. We threw balls around, and had bike races on noodles, and did strange leg exercises.  I had a bit of trouble understanding instructions, but I eventually figured everything out. I sure learned a lot of new German words this week.  I didn\'t do much at all today. I\'ve been really lazy, hoping to get rid of this cold.Johannes and I also started working on picking some music to play at Ludger\'s wedding reception. Yesterday afternoon, I explored a music store downtown and found a book of flute duets that would work for Johannes and I. He has a tenor recorder, and I have my flute. It\'s hard to find duets because flutes usually play high notes, and the tenor recorder doesn\'t go that high. I finally found a book with duets from the 1800s that works. We picked a couple of them, and then worked on a piano duet, and then Johannes pulled out one of his old organ books, and I played the top part on my flute, and he played the rest on the piano. It was interesting sight reading- tons and tons of sixteenth notes that were really high for me to play. I think it would have worked better if I could breathe through my nose. I don\'t know how much of that stuff we\'ll actually play for the wedding, but it was fun to work on.  Anyway, that\'s enough for now. I should really try to get some sleep.   '),(14014,3832385,'22,July,2004','2004-07-22 00:00:00','   Do not go gentle into that good night, Old age should burn and rave at close of day; Rage, rage against the dying of the light.Though wise men at their end know dark is right, Because their words had forked no lightning theyDo not go gentle into that good night. Every so often I think about it and I shudder. Itâ€™s a feeling that haunts me. I donâ€™t like to talk about it. I just shrug it off and try to forget as soon as I remember. I soldier on. What else can you do? I wake up each morning and tell myself I canâ€™t go on, that I must go on. So I go on. And yet sometimes I embrace it. I look down into it and realize that it has to be, this ending, this finality. At these times there are few things that I can do besides want to hide myself from others, so that they canâ€™t see the utter terror on my face and the heaviness in my heart. Itâ€™s a feeling we all have and few want to share. And even fewer share in constructive ways that make others feel better about the situation. These few are wise men (and women). And their words sometimes bring me solace. Good men, the last wave by, crying how brightTheir frail deeds might have danced in a green bay, Rage, rage against the dying of the light.Wild men who caught and sang the sun in flight, And learn, too late, they grieved it on its way, Do not go gentle into that good night. And I guess my greatest worry is that sometimes that Iâ€™ll never be able to capture the same sorts of feelings and emotions in my writing. To ever bring to the fore the real issues behind certain pieces of music, whether the artist intended it or not. There are flashes and glimmers every once in a while, I feel. But for the most part it lies flat on the page, leaving impassive readers to their impassive reactions. Even my most accomplished writing, dancing upon the page, leaves meâ€”in those desolate momentsâ€”feeling weak and worthless at getting the point across. And even the writers that I most admire sometimes have these feelings. They just hide them better. Wild men, gonzo journalism, concepts rising above their imposed structures to say something moreâ€”there are moments of doubt everywhere to be found if you can look hard enough and care to engage with a writerâ€™s entire oeuvre. But usually you donâ€™t. You read the highlights. And the lowlights. And draw your conclusions from there. Grave men, near death, who see with blinding sightBlind eyes could blaze like meteors and be gay, Rage, rage against the dying of the light. Which is why I donâ€™t fear the present. And I donâ€™t fear the past. And you, my father, there on the sad height, Curse, bless, me now with your fierce tears, I pray. Do not go gentle into that good night. Rage, rage against the dying of the light. And itâ€™s why I choose to engage with moments. Perfect moments. Moments that confirm life. Moments that make me smile as I head to an office after a few hours sleep, after working on what I really believe in. Only insincerity would make you believe otherwise. And itâ€™s in these moments, moments that send OTHER shivers down your spine, moments that destroy the monotony of everyday life with their blinding light and humanity and goodness and utter rightness that I canâ€™t believe that they exist. Junoâ€™s â€œThe French Letterâ€ has one of these moments.  todd burns '),(14015,3832385,'18,July,2004','2004-07-18 00:00:00','   id swim across lake michigan id sell my shoes id give you my body to back again in the rest of the room to be alone with you you gave your body to the lonely they took your clothes you gave up a wife and a family you gave your ghost to be alone with me to be alone with me you went up on a tree ive never met a man who loves me '),(14016,3832385,'14,July,2004','2004-07-14 00:00:00','   feel meter (1-10) current state:7 happy current position:jakarta '),(14017,3832385,'12,July,2004','2004-07-12 00:00:00','   ive said before that life is much easier for me these days that when you already know what you want in life  the dissapointments that came becomes much bearable easier to live with well is it? i want to go to spain but i didnt  i was rejected the dissapointment still lingers on why wouldnt it go away? '),(14018,3832385,'11,July,2004','2004-07-11 00:00:00','   Statement: It\'s not really that insane is it? Response: No, it\'s not.   TO HAVE BREAKFAST, THE BEST PLACES  first of all, breakfast is kind off a luxury for me,only have breakfast 2 or 3 times a month, is it really possible to go on without breakfast forever?  mcdonalds, anywhere in singapore most probably the only country where you can get a big breakfast meal, my favorite spot is the one opposite marriott hotel, just a little bit on the side of shaw building, quite noisy but all worth it.   dome cafe, burjuman, dubai the meal itself is so so, but the view is just worth it, great place to spend from 7 to 11 am  lotus, pondok indah, jakarta dim sum galore  gran melia, jakarta the compromising choice, a good family spot  nasi kuning, jalan veteran, bandung the tastiest cheepest breakfast ever  i wish i was there to see scissor sisters jake shears stripped himself at this years t in the park  its a gloomy day and i couldnt enjoy it '),(14019,3832385,'11,July,2004','2004-07-11 00:00:00','   puncak  urlLink '),(14020,3832385,'10,July,2004','2004-07-10 00:00:00','   bdg  urlLink '),(14021,3832385,'10,July,2004','2004-07-10 00:00:00','   urlLink    bdg '),(14022,3832385,'10,July,2004','2004-07-10 00:00:00','   pi  urlLink '),(14023,3832385,'10,July,2004','2004-07-10 00:00:00','   as we come close to our time we cant help to think that it could be a quick one '),(14024,3832385,'10,July,2004','2004-07-10 00:00:00','   what goes around comes around when you reject someone you bound to get rejected yourself reject = eject ejecting the cassete when youre thru listening just eject it when you dont need it anymore act-ion reject-ion fabricat-ion demolit-ion react-ion fact-ion dict-ion select-ion i could go on and on when you dont want to get rejected then dont reject dont reject someone onto your life my life i have my right to reject anyone from coming all over affect-ion the rules of attract-ion wanna have it wanna know it wanna open up wanna be understood we are because we want to we have because we just had to do we love because we want to we need to we have to or we forgot to do it ??? way out  just all the way out of here wanna be out of here ions away '),(14025,3832385,'05,July,2004','2004-07-05 00:00:00','   when i was reading a review on pan sonic\'s latest effort the sprawling 4 discs \"kesto\",i came across a rather unique word,even more peculiar is the fact that the word suggests a deeper understanding for \"kesto\", that to really understand \"kesto\" which was said to be heavily inspired by this, we have to know what triptych really is. references to francis bacon was an obvious on my google search result, it seams that he is creator of this triptych word. said is the most arresting and major work from francis bacon is the triptych work from 1962 called the three studies of crucifixion. ive seen all along many triptych works, its interesting how i finally came across what probably the original, never thought that those works that i saw was called a triptych. there are probably many triptych words that i worte here, but i still not come close to any conclusion or any understanding of what the word really means. funny huh. i started with a curiousness towards the word. but what i really want to write is about bacons work. i will promised you that in the future i will explain a little bit more about the word triptych. yes, Francis Bacon Three studies for a Crucifixion 1962 Oil on canvas each panel 197.2 x 147 cm Staatsgalerie Moderner Kunst, Munich odd enough, three work that titled as a study on crucifixion didnt have any crucifix at all. when seeing this for the first time, one could easily came to think, and feel the violent nature of it. for somebody who grow up with too much tv and movie violence like me, it obvious that i would come to a conclusion of a murder scene in the making, or sort of. not the most beautifull sight, but it is beautifull in a sense of a story if there is a story in this work in the first place. triptych are supposed to be based on a singular idea, then, that idea was put to work into three diferent canvass, and if those three canvass are put align, then you could find a flow, more sense to it, sort of like a frame by frame study of a film, or an print from a lomo action sampler, sort of. it seems that this perticular work is suposed to be seen as a flow from left to right, although im not sure weather or not its what the artist expect us to see it that way or not. where i see it, it flows from left to right. the left part shows two person, seen from a view of a third person the victim i supposed, the two person viewed, was viewed from the right side, one person seems laughing as if proud of what he has done, the other was anxious to leave. The scene is inside a circular like room, the victim iying with its back upright, on one side, the opposite side saw a window and a door. the first one seems to establish an aura of curiosity, some questions that come up when looking at it are, what is happening?, where are they going?, where am i?,why is he laughing? and so on... the first one actually suggest us to see the next one. the second one, the one in the center, not always have to be the central piece, but shocking none the less, really up the tension level, a body, disorganized, crouching, naked, bloody, deformed, the head is shown in skull form, while the rest of the body is still flesh, some internal parts are also visible, although not clear which is which, its lying on a bed, white sheets with splatters of blood here and there, the windows in the back are covered with blinds, dark ones, purposedly shut, its hard not to feel a little pity towards it, is it dead? the third and last one up the level of tension another notch, the most appalling, and even more violent than the center one, half human half animal, or is it wholly animal? with an open ribcage, one cant help but think what is this thing? and what is that in the center? is it a head? why is the head placed there? trully uneasy and unaccessible... this account are based on first person first time seeing basis, deffinitely there will be some more in the future '),(14026,3832385,'04,July,2004','2004-07-04 00:00:00','   i dont want to vote but i vote anyway why people forced other people to do what they dont want to do is a question that lingers along until forever damn you presidents '),(14027,3832385,'04,July,2004','2004-07-04 00:00:00','   if i am alive this time next year will i have arrived in time to share? '),(14028,3515252,'04,June,2004','2004-06-04 00:00:00','    Hey, I just realized that to blog on this site would require everyone else to jump hoops in order to comment, which i have already stated I hate, so \"good bye blogspot\"  Do not expect to see anything more here from me.  I couldn\'t put my adoring fans through that.   '),(14029,3515252,'01,June,2004','2004-06-01 00:00:00','    So, I have already said how much I don\'t like having to jump through hoops just to comment on somebody\'s blog.  But, look, here i am with a brand new blog of my own just because I wanted to comment.    Perhaps this is a good thing.  Maybe I should divest my interest in Envision\'s website and give it back to Davin?  We\'ll have to see.   '),(14030,3515252,'12,July,2004','2004-07-12 00:00:00','    So i like the privacy of this spot.  I know the concept of bowing to blogger and \"logging in\" every time you want to comment or anything has been prohibitive for me, but I like the hoop jumping now.    A little distance is a good thing sometimes.   '),(14031,3515252,'11,July,2004','2004-07-11 00:00:00','    Well, i like having my own place.  Especially  such a quiet place.  Not much to say today.  pretty sick of the tears thing.  But, oh well. saw   about a boy   last night.  That was a fun movie.  I\'ve got one for tonight but I think I\'ll go to bed.  gotta work in the morning.  Babysitter is flaking out on me.  Wants fridays off.  At the outset said fridays were a work day. I wish I had a good alternative.  But. . .that really cheeses me off.  anyway gonna pee, gonna read, gonna sleep.  and hop back on that hamster wheel tomorrow.    Yee-Haw.   '),(14032,3515252,'11,July,2004','2004-07-11 00:00:00','    so, here\'s the deal.  I\'ve been thinking about divesting at envision and moving my blog here.  i think i will do it, but not tell anyone.  So here i\'ll say whatever I want.  No one will ever find me.  I fucking depressed.  It follows me around like a cancer, like a smell.  I might shower but it always comes back.  I might find some diversion, a movie, a friend, but I can\'t run away from this fucking disaster of a life.   I am so sick of crying and not being able to answer the question, \"how are you?  How\'re things going?\" and I can\'t figure the damn thing out.    I can\'t pray.  I can\'t guess what this fucking beast wants or else I\'d give it.  Right now.  I wish I could tell someone.  I need help.  Oh my God.  Is there anyway out of this?  Dan\'s talk about God\'s will just makes me think I need to handle this on my own, that the will is revealed in the word and whatever is there goes.  I\'ve set up a meeting with Paul on Monday.  I don\'t know why.  I don\'t know what else there is to say.  I really want to die.  If I could without scarring my girls I would do it.  And I don\'t give a fuck if that is narcissistic.  I\'m past caring.  So I suck, big fucking deal.  I always have sucked and tried to fix the suckage.  Who gives a shit, now?  I want to die.  I\'m past being useful and over being a good friend.  I just maintain for my kids.  and I\'m losing the ability to do that.  Can anybody help me?  can anybody hear me?  Fuck me.  The ship\'s going down, man.  Who cares?   '),(14033,1675646,'31,July,2003','2003-07-31 00:00:00','   My final night in this apartment.  This is the first place outside of my parents house of course where I actually made it to the one year mark.  I must say I am feeling oddly emotional about the whole thing.  I have never been one to revel in the idea of hard physical labour either.  I hate moving.  In fact I already hurt my back because stupid me, yet again(I\'ve done this everytime I moved), put almost all my cd\'s in one box(which is 200 something) then tried to lift really fast.  So that didn\'t work out so well, and now my back hurts.  Well what can you say besides I am an idiot.  And before anyone gets all excited: If you ever come visit me remember that I have rather eclectic taste in music and my collection if you can call it that deserves to be made fun of.  Although don\'t make fun of it to my face or I will hurt you.  My apologies but moving makes me violent.  Just be happy that I am moving into my own place away from my roommates and their swords and fighting sticks.  I do however have a baseball bat, tennis racket, and a Darth Maul double edged lightsabre.  Why yes I am a big geek, thank you for noticing.  I also found out today that my new computer has arrived...hooray...although my hookups for phone, net, and cable aren\'t until Aug. 7.  But I will do my best to get to the internet where I can, most likely at the home of my loving parents.  My eldest sister went to the \"I love SARS\" concert.  At least that\'s what she called it...because of course if it weren\'t for SARS the event never would have happened.  Hence the I love SARS.  But I digress.  I\'m hardly jealous, I mean if I wanted to see a bunch of old men try to be young I\'d....well I\'d something anyway.  It was like the geriatric empowerment sessions or something.  ACDC???  The Rolling Stones??  The Guess Who??  Blue Rodeo??  Sass Jordan??  Who let these people out of the home??  Really they should be sitting in a chair eating honey mush, not playing music for the Gen X croud.    As you can tell I am going to miss my internet for the next little while so I am ranting on majorly to make up for it.  But all good things must come to an end, and I must rest up for my move, and final cleaning.  Well there will always be the memories, oh yes, I\'ll always have those.  '),(14034,1675646,'28,July,2003','2003-07-28 00:00:00','   Ewwwww oven cleaning.  Okay I don\'t actually gross out that easily, nevertheless the oven is not pleasant.  Sigh, still waiting to know when I can move in to my new place and getting impatient with it.  My roommates can start moving into there place tomorrow, and I naturally am helping with that.  The irony there is that it\'s likely that when it comes time to help me they won\'t be able to help me.  It\'s just the way things will probably work out, I don\'t hold it against them it\'s just annoying.  And I once again must remind people, that if you don\'t hear from me for a little bit I am alive and doing fine (I hope), just without phone or internet for a while.  What a vapid entry!    I shall make it my goal after the move to actually make some interesting entries...maybe even thought provoking.  Maybe...you never know...anything is possible...  '),(14035,1675646,'27,July,2003','2003-07-27 00:00:00','   I should have packed today but I didn\'t.  Oh well, I\'m sure I\'ll get it done.  I\'m still waiting to hear from my new landlords about movie in time.  Here\'s hoping I can move in the same day I leave here, the 31st I hope.    I ordered my new computer (as the one I have written all this on isn\'t mine), which will hopefully arrive Aug. 2nd.  And since I am not entirely clear on what my new address is it\'s being shipped to my parents house.  Which is only 5 minutes drive away from me anyway.  But without knowing the address I can\'t get an appointment with eastlink to set up my phone cable and internet.  And even if I made an appointment I can\'t be sure when the computer is coming.  Shipping is supposed to be two weeks, and I placed the order Tuesday.  I am looking forward to my move, or at least the move being over with.  It\'s way too hot in this apartment and it\'s interfering with not only my comfort level, but my sleeping pattern as well.  I could probably be sleeping right now, but instead I\'m writing here.  So I guess if there\'s anyone who may need to call me next weekend it\'s unlikely I\'ll be availabe...I can still check my email at a couple different places.  I\'ll likely hang at my parents, as I need to sort through alot of my old stuff that\'s still sitting in their house.  Let\'s see...today is July 27 (my time it is anyway)...I turn 24 in 2 weeks.  Time to put myself out to pasture, because I\'m an old mule.  '),(14036,1675646,'22,July,2003','2003-07-22 00:00:00','   Oops not writing enough.  Over the course of the past few days I have come up with many great things to write about.  But now, as the time comes to write, I\'ve got nothing.  This is like writing an exam or something.    If I were to test the people who know me, I would bet the results wouldn\'t be very good.  Although I\'m also willing to bet that I would fail a test on myself as well.  What is this drivel I spewing?  Apparently I\'m in moos to be cyclical.  That\'s no fun.    So allow me to pose a philosophical question.  If I am my own worst enemy.  How do I fight myself?  I know all my tricks, I know all my moves, and all my arguments.  So how do I do it?  I\'d like to know...but then once I know how, I won\'t be able to pull it off because \"I\'ll\" know.  '),(14037,1675646,'17,July,2003','2003-07-17 00:00:00','   Is it wrong to enjoy a movie that is thoroughly awful??  Well although that\'s a rhetorical question, if it did have an answer it is certainly no.  If it were wrong to be entertained by bad movies, I would have a very small movie collection.  Actually I do have a small movie collection but that\'s beside the point.   So yes, League of Extraordinary Gentlemen, terrible.  It is based on a graphic novel.  Which, for those of you who are unaware, is like a big comic book.  Anyway the graphic novel is actually British in origin.  But as we all know, American audiences tend not to understand anything British, either by choice or by genetics.  So in they bring the allegedly attractive Shane West as Tom Sawyer.  And aside from one good joke at the expense of the American character, it just takes away from the dynamic of the GC.  In fact the GC is actually incredibly light on action.  But I don\'t consider it worth writing a review of, you can go see it or not.  I think it has something to offer, whether it\'s a good or bad thing would be up to the viewer.  Just keep your expectations low.  Bottom line, Sean Connery still rocks, even though he hasn\'t made a good movie in 20 years.  '),(14038,1675646,'15,July,2003','2003-07-15 00:00:00','   Going to see a movie tonight.  At least I think I am.  Reviews will follow.  I know this is dated, but anyone who though The Matrix: Reloaded sucked donkey balls say \"aye\".    So let\'s see....it\'ll either be the loveable Yeti Sean Connery in League of Extraordinary Gentlemen.  Or Johnny Depp and uber-hottie Orlando Bloom in Pirates of the Carribean: Curse of the Black Pearl.  As long as it\'s good summer fare I care not.  I am hardly expecting a cinematic masterpiece.  I mean hey, a movie like Demolition Man is only made once in a lifetime.  How that got shut out at the Oscars I have no idea.  Speaking of terrible movies.  I bought Independance Day on DVD today...hooray for bonus features.  Oh well, I never bought it on VHS, so I\'m allowed to get it now.  As I am currently too poor to repurchase me VHS collection (not that it\'s large) on DVD.  I don\'t think I have anything else to comment on right now.  Talk about sad.  If I wasn\'t going to pack boxes all night I\'d watch the news and write a brilliant response to some story that was on.  Oh yes....and the hands down best use of a cowbell in a song is most defiantely \"Don\'t Fear the Reaper\" by Blue Oyster Cult.  Just so you know.  Close second award goes to \"Hey Ladies\" by the Beastie Boys.  Man this blog is vapid beyond belief.  Yay.  '),(14039,1675646,'15,July,2003','2003-07-15 00:00:00','   Why is it that when I try to write bad poetry is turns out great and when I try to write good poetry it sucks.  I\'d write it on paper, tear it up and throw it out, but there\'s no point.  Birds won\'t even use them in their nests.  How long have I been doing this blog, and I\'ve already stolen like 10 Simpsons jokes.  Matt Groening is gonna have my head.  '),(14040,1675646,'14,July,2003','2003-07-14 00:00:00','   What can you do with a day?  What will you wake up and see?  The farther you get the closer to me.  -Dar Williams \"Closer to Me \"  I am not sure what the point of putting those lyrics there is.  It would make more sense to put something that I wrote there.  But to be honest, most of the poetry I write is so dark it frightens even me.  Therefore it remains locked away.  Today has been a day like any other really.  Working in the morning, and a long nap became necessary.  And so after drueling plenty and leaving my throat dry enough to sing Bonny Tyler songs I am now awake trying to figure out what to do with my evening.  Packing should be on the agenda, as it\'s basically 2 weeks til the big day and the more prepared the better.  Ah, screw that being prepared is a boy scouts thing.  And I ain\'t no f*cking boyscout.  The unfortunate thing about napping in the afternoon is that I am now going to be groggy for the next 5 hours.  So anyone who is on MSN with me right now keep in mind that my brain no work good now.  I\'ll write something else later...because I don\'t have anything to say at the moment.  And because I think I\'m going to give into temptation and get an insanely large slice of pizza from accross the street.  '),(14041,1675646,'14,July,2003','2003-07-14 00:00:00','   I suppose you\'re all wondering where I got the name for this blog.  Well perhaps not but I\'ll tell you anyhow.  One day while feeling disallusioned, as I am apt to do, I made up a second hotmail address because I was going to use it for junk mail and such.  That never did happen and I\'m sure that account has closed.  During this time I was a student at St. Mary\'s University in Halifax, which is essentially the capital of the generic.  I have never been anywhere that I felt more left out than while attending SMU.  And since is seemed as though my uniqueness was something to be ashamed of I tried to remain a cog in the gritty machine.  This of course did alot more harm than good for me.  Now I have had a theory for a number of years of: \"Unique in mind, not in body.\"  That is not to say that I look down upon people who take it upon themselves to look different on the outside.  Far from it.  My belief is simply that for me, I can be very a unique person who thinks differently from the majority and still look like an unassuming dork.  Who needs crazy piercings and tattoos when you have thought?    I still often feel as though the \'majority rulez\' as I don\'t seem to fit in anywhere.  And indeed I guess I can say that I subscribe to the theory that sheep go to heaven and goats go to hell.  Goats being people who are different.  Bla bla bla...it\'s also a great song by the band Cake.  Do I believe the majority always rulez??  No.  However with all the pressure in our society for people to \"fit in\", it can certainly be a big factor in peoples lives.  '),(14042,1675646,'13,July,2003','2003-07-13 00:00:00','   I hate food.  Food is the beign of my existance.  Because to say I hate food is a lie, I love food, I can eat all the time.  I eat when I\'m hungry, I eat when I\'m bored, I eat for the sake of eating.  This food obsession is gonna get me in trouble some day.  My arteries are probably just as gross as Halifax Harbour.  Although I don\'t think I have condoms floating around my arteries.  Stupid food...I\'m completely hooked.  Only by the grace of some unknown force am I not 300 pounds.  I remember the days when I used to be an athlete, ah those were the days.  The days before I had a job and made my own money.  The days when I couldn\'t eat out all the time.  I swear there must be addictive additives in fast food.  Because nobody in there right mind would want to continuously eat it if they could choose not to.  I don\'t want to eat fast food, and yet I do.  It practically brings me to tears sometimes...well not quite, but perhaps someday.  Because it seems as though I can\'t stop myself, like something makes me do it.  Maybe they put mind control devices in the hamburgers, and there is a massive control center that targets people with technological voodoo, forcing them to eat their greasy disgusting food.  Why am I whining about this anyway?  That\'s not a rhetorical question, somebody needs to tell me.  And the first person who says it\'s because I\'m a native Maritimer is gonna end up with a broken limb or three.  Now if you\'ll excuse me I am going to go purge and read a non-threatening Boys magazine.  '),(14043,1675646,'12,July,2003','2003-07-12 00:00:00','   Ah I see how this works.  *Sigh* I am about as computer savvy as a blind lemur, and am equally as likely to jump off cliffs.  So yes what an exciting day.  I went to my parents house as I do on most Sundays, to visit and make use of the free laundry facilities.  My parents are renovating the house and it\'s coming along nicely.  I ended up moving furniture, which was completely unexpected, and has led me to the epiphany that I had better hit the gym before I move into my new apartment at the end of the month.  My roommates are packing, it\'s all so very exciting really....I mean it....it\'s just soooo exciting.  Of course I am aware that I had better start packing lest I leave it to the last minute and break all my stuff.  Of which there is very little.  Does anyone else think that Paul Martin needs to have an anvil dropped on his head?  I know that just comes out of nowhere, but I mean seriously, the guy can\'t be trusted.  You don\'t trust someone who is rich to handle your money, they tend not to have a great sense of fiscal responsibility.  Not that I do.  I mean, I eat out more that I sleep it seems so I guess I\'m not one to talk.  Stupid bickering roomies...Man, I am going to revell in my own place.  I love these two but they drive me nuts.  If I am ever in a relationship and I act like that someone kill me in an absolutely horrible fashion.  Perhaps force me to listen to Avril Lavigne while forcing me to watch wrestling.  That\'s a pretty horrible way to go.  OK, I\'ve actually disturbed myself.  Oh well, I will go vomit now and then I\'ll come up with a plan to write something more intelligent.  '),(14044,1675646,'12,July,2003','2003-07-12 00:00:00','   I don\'t actually know what the word firmament means.  It is simply the title of a song by Dar Williams.  I suppose I should look it up before using it to describe something, but I don\'t own a dictionary, so I guess I\'m stuck.  And I don\'t think that saying \"man that guys ass if firmament\" is correct usage.  This being my first post I suppose I should write something of myself.  So that way if anyone was actually reading this they could follow along comfortably and understand what on earth I am talking about.  First things first:  If you are going to follow this site keep in mind I have no life so you will likely be able to look in every couple of weeks and not miss anything.  But I suppose you never know perhaps ranting over the internet will become my new favourite passtime.  My name is Dane, I am 23, I live in Halifax, Nova Scotia.  I was born and raised here etc.  I am gay, I likes men etc.  Perhaps I\'ll discuss that later who knows.  I like sports, I like movies where things get blown up, ditto for video games.  I\'m a big fan of the Dynasty Warriors series because nothing is more cathartic than hacking hundreds of people to death.  I am manic-depressive, but I get by.  I am currently a student at Mount Saint Vincent University in Halifax.  I wrote for the school paper \"Jargon\" this past year focusing on being editor of a section on \'Sexuality\'.  This coming year I will be co-editor of the paper and will continue to write about sex and sexuality as it is something I am interested in.  Moreso academically than in practice, honestly.  I don\'t enjoy sex all that much and don\'t tell me it\'s because I\'ve never been with anyone good because believe me that\'s not the case.  Oh and I should mention that my grammar sucks.  Taking two and a half years off school tends to weaken some of those skills.  But hopefully in time it\'ll all come back to me and my nightmare of red marks all over my essays will end.  Damn comma splices, I hate them so much.  So anyhow.  I am a huge fan of The Simpsons.  Indeed I love cartoons, especially Saturday morning cartoons from the 80\'s, and many of the modern sitcom animated shows.  Everyone write to MTV to make sure the produce another season of \"Clone High\".  That\'s a great great show.  Well that\'s me in a nutshell I guess.  There\'ll be more to tell I\'m sure, and all of it as equally boring as this.  '),(14045,1675646,'31,August,2003','2003-08-31 00:00:00','   Actually to her credit though...that is the girl in my prior entry...She was the first person to ever ask me if I was gay.  I should have told her \'yes\'.  I wish I had said yes to the other few people who asked me that when I was closeted.  Not as though things would have been better that way.  Just because I am curious as to what would have happened.  '),(14046,1675646,'25,August,2003','2003-08-25 00:00:00','   So I\'m desperately waiting for my counsellor to come back from vacation because when he gets back he\'s getting me into the clinic so I can get some meds, praise jesus, because only the lord knows I can\'t go on this way especially with school starting.  Haha I am the king of the run-on sentence.  I\'m still trying to figure out how I am supposed to co-edit the school paper when I have no grammar skill whatsoever.  Well that\'s probably the least of my problems in terms of that situation.  How about that I did nothing about it all summer, the other editor was in Prague all summer, I don\'t know if we\'ll have the money to publish, at least by the first publishing date.  I don\'t know whether we\'ll have anyone writing for the paper, and more importantly I don\'t even know why I\'m doing it.  I suppose it has something to do with the former editor believing that I was the right person for the job.  Although personally I still believe it\'s a big conspiracy because people just want to watch me crash and burn.  But I mean if that\'s the goal just sit me down and make me listen to \"It Won\'t Rain All The Time\" by Jane Siberry and see what happens.  You know, I\'m just doing the whole...\"I don\'t know how I\'ll ever make it\" thing.  Honestly, I have no idea how I made it through last year.  I had meds for most of that, and that helps.  But still, and especially the 2nd semester the meds started failing and I really stopped doing well.  Looking back it\'s amazing I passed a couple of my classes and wrote for the paper, and worked, let along finished the year with a not incredibly terrible GPA.  Mind you the lack of structure in that sentence should knock me down a grade point.  Thank goodness this isn\'t a school project.  My entries are too long, nobody wants to read this stuff.  But there\'s one story I feel like telling.  I\'m working the other day, and this girl comes up, and I know her and she knows me, she\'s friends with my older sister.  And she wants to buy a chair, it\'s a cool chair, good for her.  But then this girls sister, who is my age comes in.  Words can never express the contempt I feel for this person.  I went to school with her for 10 years and not once did she make me feel like I was anything even close to human.  I despise her, and I always will.  I suppose it\'s a good thing I don\'t consider myself catholic anymore, otherwise I would be forced to forgive her.  I\'m sure she\'s quite the mature and wonderful person now.  I mean we all grow up.  I haven\'t even talked to her in 8 years, aside from the hello I gave her when she came in the store.  However, sometimes I think people are forgiven too easily for past transgressions.  Or at least it\'s passed off as \"oh well that\'s what kids do.  They just need to grow up.\"  Do you want to know why I hate guns so much??  It\'s because I\'m afraid of what I would do if there was one in my hand.  '),(14047,1675646,'25,August,2003','2003-08-25 00:00:00','   Speaking of DVD\'s, why is the stuff I want so damned expensive.  I mean, hello, the Star Trek series dvd\'s are like $150 per season.  Plus I would love to get all of the Dr. Who stuff.  I need to find a way to make some money.  It\'s a shame I know nothing about html programming, otherwise I could just start a crummy porn site and get cash from silly horny people who buy memberships without thinking with the upper brain.  Hmmm that\'s got me thinking.  There is much to do.  Or actually not do, but I\'m trying to make you think I am doing it.  I suppose I\'m ruining it by telling you.  Oh well.  '),(14048,1675646,'22,August,2003','2003-08-22 00:00:00','   Why did Fox cancel \'Futurama\'.  Those morons!  Season two DVD set in stores now.  '),(14049,1675646,'22,August,2003','2003-08-22 00:00:00','   I have so much to do, and so much time to do it in.  So why is it I\'m not doing any of it?  Rhetorical questions are fun my friends.  True dat, as some may say.  Am I unhappy?  No.  Comtemplative would probably be the best description.  I\'m like a little boy, cowering in a corner.  Partly fearing my imminant failure, and slightly fearing any possible success.  OOoohhh rhetorical nonsense...anyhow don\'t misunderstand this brief entry, it\'s just a couple of sentences with no real meaning.  I want to add more but I should go to bed, and I ain\'t not making the sense anyways.  '),(14050,1675646,'18,August,2003','2003-08-18 00:00:00','   I guess it was time for a change.  That\'s the kind of thing I have been telling myself for a long time.  Time for a change.  \"Farewell to the Old me\", as they say...or at least my favourite singer/songwriter Dar Williams says in a song of the same name from most recent CD, \"The Beauty of The Rain\".  Now I praddle on.  My ex was in twon on Friday night and dragged me out to NRG.  It\'s not that I had a bad time, but if I\'m not dancing I get really uncomfortable, and since there was nothing at all played that Iwanted to dance to I was left to be uncomfortable.  That not withstanding my ex prancing around like he was the Queen of Siam.  Not that there is anything wrong with that.  But thanks to Dallas who allowed me to keep from freaking out and being utterly miserable while I was there.  There were some other friends of mine there but they were drunk, so it\'s not quite the same as when they are sober and can relate to me.  Plus they were tickling me, which as everyone knows ticks me off.  So I actually put my friend Woody in an arm bar breifly.  That was amusing, he was quite surprised that I did it actually.  But hey, as much as my inferiority complex argues otherwise...I am full of surprises.  Or maybe it\'s full of shit, one of those two anyways.  '),(14051,1675646,'14,August,2003','2003-08-14 00:00:00','   I officially declare Mitsou to be the greatest musical artist of all time.  Someone has to say it.    Now with that aside.  Yes of late I have reindulging myself in music...I love music for those of you who don\'t know.  Essentially I really like just about anything.  But especially songs from my past, the kind that tend to illicit positive emotion, or an emotion of some kind.  There are some songs which just remind me of various moments in time, and to listen to them almost puts me back in that moment.  Despite what I may have led people to believe, not all the moments of my life have been negative.  Quite a few mind you, and chemical imbalances tend to make those bad moments seem like great tragedies.  Who reads this site anyway??  I know of two people for sure, and of course those are the only two people with my email address too.  I want to know what kind of direction people think I should take with this blog.  I really don\'t know.  My mind is usually all over the place so it\'s hard to focus on one thing unless I know what it is going in.  See I\'m not making any sense.    \"It was that cocky janitor Glen!!!\"  '),(14052,1675646,'12,August,2003','2003-08-12 00:00:00','   Thank you everyone for your outcries of support.  But as I believe I wrote(I\'m too lazy to go check) I did not write that blog because I was looking for support or sympathy.  Regardless thank you for showing you care.  But let it go, I was in a pretty horrid mood that night.  I suppose I still am, and yes I am afflicted with this condition and have yet to find a way to cure it or aleviate it.  But that doesn\'t mean I\'m going to give up.  Because I know that if I stop trying to make myself better...sometime down the road, I won\'t be around.  That\'s not meant to sounds grim but it\'s the truth.  So anyways....yeah...don\'t stop caring, but I\'ll be fine.  '),(14053,1675646,'10,August,2003','2003-08-10 00:00:00','   On another note, after reading one of Johnny\'s posts I was inspired to go take some quizzes at thespark.com which is a site I hadn\'t visited in a while.  And I am not tryin to steal your thunder by posting my results Johnny, but after the last entry anyone reading may want a lark.  So yes, I am 50% gay, which is 7% more gay than the last time I took the quiz...I\'m not sure why.  I am 66% dateable.  According to their personality test I am a Mastermind(Submissive Introvert Abstract Thinker).  I am 48% sexy, which is giving me credit I think.  And I will have sex with 28 people in my life.  I\'m curious as to what their definition of \"sex\" is.  Because the number of partners I have had varies depending on that....and I very much doubt I would have \"sex\" with 28 people in my life.  If that were the case I\'ve got a long way to go.  ');
INSERT INTO `msgs` VALUES (14054,1675646,'10,August,2003','2003-08-10 00:00:00','   So.......  So yes apologies for not updating in a long time.  Aside from moving I have been putting it off because I honestly have nothing good to say, and by that I mean I\'m likely to write something horribly depressing.  Today is my birthday by the way.  Thank you Dallas for my wonderful card.  I hate birthdays.  Not because of getting older, or anything like that.  Just because it\'s a day that doesn\'t warrant celebration.  Although sometimes I think that the fact that I managed to live another year does warrant some form of celebration.  Recently a friend of mine pointed out something about my behaviour.  He said \"Do you think that by trying to make people stay away from you, you are in fact doing the opposite and making them want to know you more.\"  I\'m paraphrasing, he was more articulate than I am being.  When I think about it, it\'s a valid point.  I guess I come accross as being mysterious to alot of people, and this is intentional in that I don\'t like giving alot of myself up to people.  And yet at the same time I am forthcoming with people, but usually only with things that I hope will drive them away from me.  There are few people on earth who know me well.  Indeed even my best friends would not be able to claim that they know me all that well.  Because I do not talk about many things with people.  Subjects that aren\'t talked about in detail:  1. My Depression 2. My hatred for myself 3. My lack of will to live 4. Severity of depression 5. Suicide 6. My hopes to die everyday  I\'ve got more but you get the idea.  And no, typing those out for all to see provides no catharsis whatsoever.  It honestly makes no difference if people know or not, because everyone offers the same help and nobody really gets it.  Although I fault nobody for that because at least they try, and I can\'t explain my feelings sometime anyway.  I had a Psycholgist promise to cure my depression.  He doesn\'t return my calls.  Hahaha, I think I\'m too far gone for him.  And as I get older I am actually getting worse.  Time was that I could hide it well enough, I never used to have problems getting up and going to work or whatever.  Now over the past year it\'s nearly cost me my job because I don\'t want to get up in the morning to face it.  I had gone to see another Psychologist once and he was absolutely shocked that I had a job, and left my house.  He was also shocked that I was even still alive.  And do you want to know why I am still alive??  Because I have a conscience.  I cannot hurt my family like that.  So I live for them, because it\'s the only reason I can think of.  I hate being alive.  I hate getting up in the morning.  I hate going to sleep at night.  I hate inhaling, I hate exhaling.  But most of all I hate myself for being this way.  There now you know that I\'m nuts :)  So I suggest that you re-evaluate whether or not you truly wish to know me.  '),(14055,1675646,'27,September,2003','2003-09-27 00:00:00','   It\'s funny that I am so afraid to talk to my parents.  Indeed I just don\'t want to disappoint them, although I invariably do things to that affect.  Even though they would never say they were disappointed in me.  They just want consultation.  I think the time has come to confess everything, and to let them into what my life is like, starting from way back and coming up to the present.  And I thought coming out of the closet was going to be hard.  Imagine having to tell your parents that 90% of your waking thought has been devoted to wanting to kill yourself, and that said thoughts have existed for almost 20 years.  And imagine even more having to explain that the only reason you haven\'t followed through on said thoughts is that you would never want to hurt them, and thusly you stay alive for them.  It\'s not likely to go over well.  And I\'m not going to do it to soften the blow of decisions I have made recently.  But rather, I think that everyone has been kept in the dark long enough.  I make allusions to certain things, but I rarely blatantly speak of my thoughts or feelings.  And even to those people who are reading this thinking that I have opened up...well...I still keep quite a bit to myself, actually almost everything to myself.  The glimpses I give don\'t tell the whole story.  Anyhow I needed to type something because I have been trying to sleep and failing miserably...as usual my thoughts are eating me alive.  Lousy carnivorous negative thoughts...curse you and the horse you rode in on.  To the sky I look, The distance and the awe, But rather is seems to me,  that the sky is close at hand, hanging above, falling perhaps, graying clouds, thunder strikes, I cry crimson tears, and then the sun rises, and all is calm and innocent once more.  Really bad grade 11 poetry...I\'m not sure I remembered it all right, but I\'m not about to go searching through stuff to find the correct version.  '),(14056,1675646,'23,September,2003','2003-09-23 00:00:00','   So I went for a very late walk.  It lasted about an hour.  I like it when the world is quiet, in fact, I think my mind may have even quieted itself for about five minutes.  Although it does need to be said that boats on the dark water of the harbour at 3:45am are damn creepy!!!  So why did I go for a walk?  Good question, I\'m glad I asked.  Of course I can\'t answer it.  I just wanted to take some time to think, somewhere that wasn\'t this apartment, or work.  I\'ve been rather preoccupied all night, trying to think of what to do, and trying desperately not to do certain other things.  OH I am trying to be optimistic people...I\'m trying so hard...but the effort is straining, in fact it\'s exhausting.  I may have a plan, I\'m not sure there\'s alot to think about.  And if there\'s one thing that I have come to hate it\'s thinking.  Oh how I hate my brain...contstantly questioning, telling, tricking....talking!!!  It never stops!  But I suppose I\'m not the only one who suffers that...except that my brain is constantly telling me what a miserable fuck up I am...and then it floods me with memories that should be forgotten, and my defenses weaken and fall and I give in.  And then I type nonsense like this.  '),(14057,1675646,'17,September,2003','2003-09-17 00:00:00','   I was trying to make a nice pastel work for my friend, because I owe it to him...let\'s just say my first effort has turned out a miserable failure.  And I\'m not being overly harsh either, I know a little something about miserable failure after all.    \"Yeah and now I\'m runnin\' down this camp town drive, with your laughter in my head, and I\'ll have to block it out some how to survive.  Cause those dreams a dre dead, and I\'m alive\"   That\'s from a very very good Jackson Browne song called \"I\'m Alive\"...taken out of context that sounds really depressing eh.  Of course I think I have the camp town drive part wrong I\'m not sure what kind of drive he says he\'s going down.  Rest assured that drive is correct though.  I mean, really probably the most depressing song for me to hear is Duncan Sheik\'s \"Reasons for Living\" which is a song that I usually only listen to if I feel an overwhelming need to cry...and please keep in mind that crying can be cathartic is doesn\'t have to be a bad thing.  \"Even my wonder, even my fear.  Only amount to a couple of tears.  There is a rythm, it\'s near and it\'s far and it goes to the heart of us.  It\'s already in you, it\'s already there.  You may disagree but I don\'t really care...\"  Of course I haven\'t unpacked my CD\'s so I can\'t find the CD inlay to get the lyrics and I\'m not in the mood to find them on the net.  Well I will but I\'m not gonna post it here...nuts to that.  haha....oh Burton Cummings...hahahaha...fine state of affairs...is a great song just so everyone knows.  '),(14058,1675646,'15,September,2003','2003-09-15 00:00:00','   I am so weak, once again succumbing to that which will bring inevitable misery.  What am I thinking??  Why??  If I know what it will do to me why do it??  The last thing I need at the moment is stupid things like this.  I\'m missing classes, not getting organised...I\'m gonna be so screwed by the end of the month.    My visit to the clinic at the Abbey Lane is on the 26th...which I have been thinking all this time is a Monday, but it\'s actually a Friday which means I had better get the time off work in the morning, otherwise I\'ll probably be working...and I can\'t afford to miss any work.  And I certainly can\'t afford to miss this appointment.  Ah well it should work out.  Anyhow I had best go off to bed.  I must get sleep so I can get up and make up for the jack and shit that I did today.  Indeed I need to make up for alot of things, and not just lost time.  '),(14059,1675646,'15,September,2003','2003-09-15 00:00:00','   I have have taken that right turn at Albequerque    I can\'t spell that word for the life of me, and I don\'t care.  I was thinking tonight about how I got here.  I mean to this time in my life.  I wonder if in some alternate universe I made all the right decisions and am living happily, without despair.  And hopefully with Elijah Wood.  If indeed the theory is true that all things that can happen do happen.  Which if you really put some though into that means that in some other realities I am probably dead, in others I am living and happy, maybe in others I\'m still closeted and engaged to a woman.  Well you never know I suppose.  Bah, nothing else to say.  '),(14060,1675646,'14,September,2003','2003-09-14 00:00:00','   Are you happy my friend, who\'s name starts with a D, ends with an S and has alla in the middle?  That means nothing it\'s late and I don\'t know what I\'m talking about.  Are you sure I\'m not Fugly??  '),(14061,1675646,'02,September,2003','2003-09-02 00:00:00','   Why is it that sleep resets everything.  When I go to bed at night I resolve to do better, to get my shit together.  And I know that I can.  But then when I wake up in the morning it\'s all gone, and I can barely resolve to move my flabby ass out of from under the quilt.  And then I go to foolish things such as get taken in by sweet words from someone who has about as much relevant things to say as Colin Powell.  Why oh why???  I should know better than that, I can see right through the language and I knew what he wanted.  He just wanted company, and he still does.  Not relationship company he just wants someone to cuddle with.  Which is fine, everyone wants companionship and everyone expresses that in there own way.  Now personally, I like to be able to have a conversation with the other person.  That\'s just me.  And then my weakness kicked in...because I am lonely...and so long story short (pardon the pun) I have officially seen an 8.5 inch penis.  Please tell me Signa that you did not give this URL to Mom and Dad.  That would be quite traumatising.  So as usual, I am now who knows how far behind in school.  And I don\'t seem to care much.  There is so much red tape and politics going on with the school paper that I just want to let it die.  It\'s not really worth it, I\'m not even sure what I am going to writing in it this year.  The co-editor has suggested that maybe I should do something about depression...write what you know as it were.  And that\'s probably not a bad idea.  If I can write about it on the net, why not in a student paper.  Oh well something to think about.  '),(14062,1675646,'02,September,2003','2003-09-02 00:00:00','   The thought just occured to me that perhaps I really do need to move home.  I have no doubt that things will start to turn around soon anyway.  But perhaps adding that safety net of ever present unconditional love will aid the process.  Of course there are many things which would have to take place.  I would have to inform my landlord of course.  I am not bound to a lease here, but they were kind enough to give me this apartment, and I am not out to betray that trust/kindness.  But on the other hand, perhaps I need to start being a little more selfish, and deciding what will make my life better.  Then of course, Mom and Dad would have to say yes.  And it is fair to say that most of my stuff would have to go into storage, and I would be happy to pay for that in place of the rent I am paying here.  It was just a thought I had.  In fact I have it alot.  Part of me just doesn\'t want to admit that my old psychologist was right.  That I am better off at home.  It is true that I hate being in the environment where people do things for me.  But I also hate being in the environment where I have all the control, it is because of that that I have spent the past 72 hours holed up in here fearing to go outside and to have to face people.  Aside from getting horribly unhealthy food I have stayed in, slept the long days, ate the awful meals, frustrated at my inability to stop myself, and yet pleased that I am succeeding in my own destruction.  Are there two me\'s, I sometimes think to myself.  Are there more?  What is it like to have more than one personality?  Must they both manifest in ways that are obvious to other people?  I\'m not a psychology student.  I have no idea.  When you sit alone at 4am thinking these things, there must be something truly wrong with you.  I live out every day hoping that that one event will happen, that one thing that will change everything for me.  Of course life doesn\'t work that way.  Epiphany\'s are hard to come by, especially for those who seem to not want to have one.  Paradox, you say?  Well perhaps, although likely not in the truest meaning of the word.  But since I don\'t know the truest meaning of the word I can use it in this way.  So I call myself a paradox, either because I believe it to be true or because I think the word sounds cool.  One of the two.  The first thing that any counsellor or psychologist or what have you will ask you is \"How long have you felt this way?\"  To which I answer always.  Or at least as long as I have memory, but it has been getting worse as I get older.  The follow up question is usually \"Why do you think you feel this way?\"  To which I respond I don\'t know.  Because I truly don\'t know.  I hardly have the right to be this way, nothing particularly tragic has ever happened to me.  I don\'t know how I became this way, and I don\'t know why, and if I did know why...why on earth would I be seeking counselling trying to find the answer to that question.  More inane ramblings.  I\'m trying to figure out if I should just stay up all night or should go to sleep and set my alarm early as though I were just going for a nap.  The problem there is that I know what will happen if I go to sleep.  I will set my alarm, and it will wake me up, and I will simply get up and turn it off and crawl back to bed, after having reset it for the time I need to be up to go to work.  And it will go off again, and I will turn it off, and I will rest my eyes, and low and behold I\'ll wake up with only 20 minutes to get to work on time.  That is what happens.  I am so predictable.  My eyes feel tired but that might be only from looking at this screen.  I slept for a very long time yesterday.  And there are things I need to do, and if I try to sleep, I\'ll end up doing none of them.  '),(14063,1675646,'31,October,2003','2003-10-31 00:00:00','   So yes I am finally writing again.  I made another addition to my home today.  Yes I now have another roommate.  The Care Bear known as Grumpy has come to live with me.  So now I have 4 roommates...Grumpy, Topper the Penguin, the stuffed cat which still has no name and my other personality Peyso.  Peyso rarely surfaces, because he is the confident organised and incredibly smooth version of me.  But he\'s around, and he says hi.  I need to start writing some articles for the school paper, which is still going forward despite my complete incompetence and neglect.  If Peyso were in control I certainly wouldn\'t be in this predicament.  I feel like I could use some Ritalin right about now.  I have been on it before, and it\'s very helpful to me...because I can\'t concentrate on anything for very long.  I get bored with this awfully quick, and the only reason I haven\'t gotten bored with it already is because I\'ve got some people messaging me as I type.  So I can break concentration on this and move to something else.  And then come back to type some more but lose my train of thought.  So anyway I cloned a sheep.  Wait a minute.  Where was I??    Invisible Robot Fish.  Hmmm.....When I can\'t stops my fiddelin\' I just takes me Ritalin I\'m poppin and sailin\' man!!     Special thanks to The Simpsons for always knowing exactly what to end an episode with.  '),(14064,1675646,'20,October,2003','2003-10-20 00:00:00','   It has occurred to me over the past couple of days that I love music.  This is hardly an epiphany, but it is funny how I don\'t listen to much music when I\'m feeling down.  A special friend of mine who happens to be named after a city in the fine state of Texas made a CD for me which helped remind me of that.  I am currently listening to a CD of my own ecclectic taste...and It\'s putting a smile on my face.  So the next step is to try to make a CD for this friend of mine.  Turnabout is fairplay after all.  I finally had the talk with the parents about leaving school.  They weren\'t upset at all, except when I alluded to things I\'m sure they didn\'t want to hear about.  So I didn\'t continue that line of the conversation.  Really I don\'t think my parents have to hear all about my inner pain to be able to feel compassion for me.  Indeed the last thing a parent wants to hear is that one of their children hates being alive, and has wanted to die for a long time.  And thusly I will not say these words to them.  But rather just make sure that they understand that the inner workings of my mind are not all smiles and sunshine.  But of course they only want me to get better, and they support my decision to leave school for the time being.  And obviously support my getting help, and also wish to help any way they possibly can.  I am blessed to have the family I have, I truly am, and I know it\'s what keeps me here.  So I have pretty much decided that I have a major crush on this guy that I met.  The problem is that I haven\'t heard tell of him in a week and I am starting to become paranoid.  This is the problem that comes hand in hand with feeling better.  When I am depressed I don\'t mind being alone at all.  However when I emerge from that I really want to find someone...then the question becomes \"do I actually like this guy, or am I just desperate.\"  I honestly don\'t know.  I want to find that out, but I can\'t do that if he\'s avoiding me or not interested in talking to me anymore.  I am very confused by this current situation, and bothered as well.  I mean, I want to see him and talk to him so much I even went to NRG briefly on Friday night hoping he would be there...and for those of you who know me....You know what a big deal that is!!  Haha, I\'m being such a silly boy...it\'s pathetic, but cute and funny at the same time so I guess all I can do is smile and try not to focus on it too much.  '),(14065,1675646,'16,October,2003','2003-10-16 00:00:00','   Wow I can\'t believe I haven\'t written anything all week.  I suppose I should be changing the date and time on these as it seems it is never correct.  For example it is now Thursday Oct. 16, 2003 at 4:00am, I\'m sure it\'ll say something different after I finish posting.  I cannot sleep.  I do not know why I cannot sleep, but I couldn\'t even settle my mind or body for more than a minute whilst attempting to sleep so I thought perhaps I would come on here type something out then try it again.  Type what out I\'m not entirely sure.  My night at work just plain sucked.  There\'s no good reason for that it just did, the last part of it anyway.  I have been letting that bother me, among other things.  As I constantly do, I allow past events to bother me...making up hypothetical \"if I had done this\" situations in my head thinking about past events that I can\'t possibly go back and change.  And even if I could the likelihood of them happening the same way again are very good.  I have met someone that I believe intrigues me.  Whether or not this person is interested in me even for friendship is uncertain.  However there is something interesting to me about a person who pays no attention to politics and doesn\'t watch The Simpsons.  I never said this intrigue was a good thing.  But maybe it is, who knows...it seems that once again my old arch nemesis Time will be the deciding factor.  Damn you Time, you\'ve won again!!!  If time is of the essence.  Then wouldn\'t the present be of the convalesance (can\'t spell that word)??  '),(14066,1675646,'08,October,2003','2003-10-08 00:00:00','   For those that would like to know I now have power.  It finally came back on tonight almost a full 10 days after it first went out.  Thank goodness for family and friends with power, because I can only stand the sound of my brain so long.  Nevertheless the hurricaine was good for me it would seem.  If I may be so selfish as to say so.  Almost as if my own storm had passed.  Of course there is much I need to get sorted out, and a couple of hard things to decide and to say to people.  The other thing I would like to note as I write this, is that this is the happiest I have felt in as long as I can remember.  I don\'t know what it is, but for some reason all day...or most of it anyways I have felt comfortable in my own skin as you might say.  That was one of the questions the psychologist asked me on Monday, \"How comfortable do you feel in your own skin?\".  Interesting question eh?  I thought it was.  I constantly fidget, whether I am calm or nervous I can\'t sit still unless I am holding myself in some way, arms crossed, legs crossed etc.  Anyhow I thought that was interesting.  A girl at work asked me today \"What\'s with you tonight.\" and I said \"I guess you could almost say I\'m happy, if I\'m cured somehow I\'ll probably be like this all the time.  Do you think I\'m acting weird.\"  She said \"No it\'s just nice to see you this way.\"  And it probably is, for those who are witness to my day to day stupor the happy me must be a stark contrast.  I\'m practically dancing a merry jig right now.  It won\'t last, I know that...it will probably reset in the morning.  But I want to mark the date, so that I will know at least, that it\'s possible for me to step out of my shadow and to feel the light on my skin.  '),(14067,1675646,'02,October,2003','2003-10-02 00:00:00','   I still have no power.  But I\'m sure it\'ll be coming soon.  And yes once again I could allude to the fact that not having power can be taken in more than one way...but once again let us go for the literal sense.    I cast a magic spell to allow myself this brief time to update my blog.  Indeed I am a powerful sorceror, conjuror if you will.  They call me Granger the master of Gray Magic.  At least they do in my head.  Damn I just realised that I need a number that I left at the apartment...I also have no phone service...and I should probably call and confirm the appointment that I have on Monday...oops...oh well there\'s still time.  Anyhow gotta go for now...yes yes...but I\'ll be back....yesss precious....I\'ll be back.  '),(14068,1675646,'24,November,2003','2003-11-24 00:00:00','   I for some reason want to cry.  There\'s no reason for it whatsoever but I\'m almost thinking it\'s necessary.  Just for the sake of purging.  I\'ve really been doing a number on myself the past few months with my diet.  I eat crap and now I\'m paying for it.  I feel like crap.  It\'s so hard for me to get out of bed.  And there are so many things I need to be changing.  The most painful thing is that I know this, and yet nothing has changed.  What\'s up with that.  One of the main things I don\'t like about me, is the rather large schism between knowledge and action.  If I know I need to change certain things than you would think that I could easily change them.  But I don\'t work that way.  So on the other hand American Thanksgiving is this week, the day that usually marks the beginning of the Christmas season.  At least it does in my eyes.  Specials should start airing around December 1st.  Not like it makes a huge difference because I have them all on tape anyway, but still I like to watch them when they air.  The quality of picture is better off of the TV than off of a tape made a couple years ago.  Third paragraph.  I want to have a relationship with this person.  I realise it has been a little over a week and I am looking to far ahead.  However he is talking about moving home in January until April, May, or June.  I personally don\'t like this prospect.  And am admittedly disheartened at the thought.  But I would wait, and I would hop on a bus to go visit.  I say this now of course.  There are at least 36 days until January and many things can change.  But then again, there are 6 hours until 8am, and many things can change.  '),(14069,1675646,'19,November,2003','2003-11-19 00:00:00','   I guess my last Blog was censored.  I can\'t say I blame them due to the rather large amount of swearing that I did.  I apologise to the webmasters for that.  And apology...How Canadian of me.  So yeah I don\'t know what to say.  I feel like I should be dumping alot of stuff on this site, and yet I can\'t seem to do so.  I\'m more empathetic than apathetic of late, I really care what\'s going on in other people\'s lives.  As things don\'t seem to be going to badly for me.  Although I have realised that I have screwed up a could of things big time, and once I come down from my high I am bound to punish myself for it.  But I digress.  For those of you who don\'t know, I guess I\'m dating someone.  The reason I say I guess it because I find it hard to believe he wants to go out with me.  I mean, to me anyway...he seems at least superficially to be way out of my league.  Although time may teach me otherwise.  I do really like him, and I really hope this one goes somewhere.  I haven\'t met too many people (at least not in a romantic sense) who make me smile when I simply think of them.  Weird eh??  I don\'t really remember what it\'s like to be enamoured, it really only happened once before, and I am far removed from that feeling.  But anyways I shouldn\'t really talk about it much.  Because things are just starting out and things could change very quickly, very quickly indeed.  '),(14070,1675646,'14,November,2003','2003-11-14 00:00:00','   First things first.  How did I ever live without the song \"Shiny Disco Balls\"???  I simply can\'t imagine life without it now.  I haven\'t felt this way since \"Finally\" by Cece Peniston. Has anyone ever noticed that she has penis in her last name??  Anyways....so yeah the past 24 hours have been a ride that would rival any rollercoaster.  Highs and lows, loops and spirals.  Anxiety, fun, contentment, anger, frustration, calm, serenity, rage....it\'s all been there and more.  Why??  I guess I\'m just getting my hopes up, because something felt right.  But I shouldn\'t be too optimistic.  In fact in my experience it never pays for me to be optimistic because the feeling gets destroyed so easily.  For example I can even predict this behavior.  I said to my boss this morning, \"I\'m in a great mood, do you want to place a bet on what\'s gonna ruin it??\".  Something always happens to ruin my good mood somehow...Today I made such an ignorant mistake and it\'s eating me up inside but I can\'t put it out of my mind.  Some of my co-workers have been trying to figure out what annoys me.  And if they ever read this site I\'m about to give it away.    I HATE MAKING MISTAKES!!!!  Even small ones, I can\'t stand it.  Especially at work because I know my job so well, it shouldn\'t matter if I don\'t sleep at all the night before a shift I shouldn\'t make stupid mistakes.  But I do, and it\'s annoying.  Fucking hell!!  I\'m gonna head for a nap.  '),(14071,1675646,'12,November,2003','2003-11-12 00:00:00','   Man am I ever lethargic.  I sleep wayyyyy tooooo much.  I think my diet is completely fucked up.  I mean, of course it\'s fucked up.  I\'ve been doing it to myself for a while.  So now it\'s time to change.  And although I did just finish eating McD\'s, it is officially the last one I am having until the New Year for sure.  No fast food unless it is Subway or a Salad.  And I need to ask my Dad to take me grocery shopping tomorrow as well.  I need to have good food here to eat if I\'m going to do this.  I wake up in the morning feeling like shit.  So this needs to change, so that I can use my days for something constructive.  I didn\'t realise it has been so long since I\'ve written anything.  Not that much happened.  Johnny came to visit, that was nice.  Ummmm....I ain\'t got nothin\'...that\'s what I mean with the \"lethargic-ness\", I should have stories to tell but I ain\'t got shit.    But at least I can say that this is by far my most curse filled blog to date.  Shit fuck fuck ass boobs crap pubes bush fuck shit.  '),(14072,1675646,'03,November,2003','2003-11-03 00:00:00','   Have you seen me???  Have you truly seen me???  Thoughts pervade my waking time.  I crave to find myself.  I suppose it\'s no less then any other person wants.  I still don\'t honestly know what it is I am capable of.  Indeed lately I have been so spastic I can barely sit still for 5 minutes.  At the times when I feel like I have grown, when I feel happier and more contented...these are the times when I am probably the most scared.  It feels to me almost as though if I were to actually come into my self, and simply realise what it is I am....that if this happened I could destroy the world.  That is how powerful I sometimes feel.  I don\'t know why.  I certainly don\'t think I can destroy the world, nor can I change the world.  Indeed I realise that at this rate I won\'t even make a contribution to the world other than to gobble up it\'s rapidly reducing resources in my personal quest for self-destruction.  THere are times when I envy Religious people.  Although I do believe in a higher power and in many other things that science can\'t explain.  I don\'t believe in them enough to be able to turn to it in my darkest moments.  Which begs the question of course:  If you don\'t have faith in yourself, how can you have faith in anything??  I\'m sure you\'re reading this thinking I\'m depressed again.  This is not the case, although I would say please keep in mind that I am never \"not depressed\".  Just depressed to lesser degrees.  I honestly don\'t think I\'ll ever be cured, mildly satiated perhaps but a cure isn\'t likely.  It is so deeply rooted within me to feel this way.  Maybe I am too smart for my own good??  That is possible I think.  I question things that don\'t need to be questioned.  And I analyse things that don\'t need to be analysed.  Give a little love, and you get a little love.  But if you give alot of love baby, you can get alot of love.  Some words to live by from Ken Tobias :)    Goodnight all.  '),(14073,1675646,'19,December,2003','2003-12-19 00:00:00','   I lost a blog in here somewhere.  Well that is to say the last one I wrote seems to have disappeared.  I\'m sure it was my fault but it still upsets me.  So I went to see Return of The King tonight.  It was great, I could nitpick about the story changes forever but I don\'t because it works anyway.  As most people know I find Elijah Wood to be super duper attractive.  But really there is a more existential reason for my like of him.  Don\'t get me wrong, I find him sexy, I respect his acting ability, and enjoy his personality in interviews.  But it\'s those eyes of his.  I look in his eyes and I can see my salvation, not as in my salvation is Elijah Wood.  Moreso that they are windows into wonder.  I feel when I look that somewhere I can shake my burdens, where all is well and I am happy.  I know it sounds strange.  That\'s because it is.  But the feeling always fades when those big blue pools aren\'t there to gaze into.  The clouds move in and it\'s impossible to see clearly again.  Man I hope he doesn\'t read this, I\'m making myself sound like a big ole freak.  How very un-Christmasy I feel right now.  How very lost I am.  I guess you could say I feel like the sword that was broken, and I am waiting for the Elves to reforge me.  But alas, since Elves aren\'t real, I suppose I will remain shattered.  '),(14074,1675646,'08,December,2003','2003-12-08 00:00:00','   It\'s always funny how the strangest things upset you.  Unfortunately of late, in the past couple of days anyway, I have been more open to negative thoughts than I have been.  This is not a good thing.  The last thing I need is to get depressed before Christmas, since I usually have alot of trouble getting through the season without feeling sad that Christmas is over, or feeling trepidations over the pointless year to come.  It is also funny to realise what a tenuous hold you have on things.  I probably could have gone to bed happy tonight, but I saw one thing and now happiness will not occur until the bliss of sleep has taken me.  Oh how I hate my waking hours.  I don\'t even think falling in love could cure that (no, I am not falling in love that\'s just an example).  And yet I am aware of many things.  The joy that I have been feeling can remain if I stay positive and follow the course of treatment.  I also realise that if I fail this time at ridding myself of such negative thoughts and emotions, I probably won\'t get another shot at it.  I will never be able to articulate exactly what it\'s like in my head.  I don\'t think I would want to either.  Nobody needs to know that stuff.  People shouldn\'t have to hear about it, but then again, I shouldn\'t have to live with it.  But I do.  For now, at least.  Live with it.  '),(14075,1675646,'06,December,2003','2003-12-06 00:00:00','   Wow I honestly hadn\'t realised how long it has been since I\'ve posted anything.  The crying remark is now funny due to something that happened earlier today.  So I am still dating guy without a name.  He shall remane nameless until I\'m sure he actually feels something for me.  No jinxing at the moment thanks.  Well anyways my Christmas spirit is about as good as it gets despite my current financial problems.  I can\'t do any shopping until the 18th, and probably will need to finish on the 24th.  My January rent is going to be sooooo late, but I am also moving out of that apartment at the end of January.  Yes I gave my notice, so now another thing I will have to decide is what to do with myself come that time.  These things will probably be discussed at the parents tomorrow.  Ah, it has been a very bad day.  Things aren\'t so bad at the moment.  But the day won\'t be over for another 90 minutes or so, so anything can happen I guess.  '),(14076,1675646,'25,January,2004','2004-01-25 00:00:00','   Oh man I don\'t feel good right now.  Laura nad I went to Vortex on Saturday night.  She was sick on Sunday, apparently Adam isn\'t feeling great and I can assure everyone right now that I am also not feeling great.  I haven\'t thrown up or anything yet, but the operative word is yet.  This sickness will delay my writing about the story of running into the first boy I ever had a crush on.  Because I just can\'t sit here, I need to go lay down.  But I will say, that since I am sick I needed something better to watch than tv so I rented two movies.  I haven\'t watched the second one yet...but as for the first I\'ll just say I need to stop renting weird gay movies from Europe.  Bizarre!!!  '),(14077,1675646,'20,January,2004','2004-01-20 00:00:00','   When I close my eyes I don\'t see any darkness.  There is a low and subtle light that shines before me.  It dangles there, sometimes near or sometimes far, but always straight ahead.  An omniscient prescence in my own vision.  And so I look forward to sleep, for it is the only time there is light.  And I can see it shine, and I can know that all is well, at least until my eyes open once more to the light of day.  '),(14078,1675646,'16,January,2004','2004-01-16 00:00:00','   Why am I here at 4am??  I\'m not even all that sure.  It\'s damn cold here by the computer and I am only in my boxers.  What fools us Dane\'s may be.  I for whatever reason felt the need to write out this brief passage.  This is not from me, although I occasionally come up with my own rather clever things.  It is written in At Swim, Two Boys...although the author is not named.  My grief on the sea, How the waves of it roll. For they heave between me,  And the love of my soul.  And my love came behind me, He came from the south. His breast to my bosom, His mouth to my mouth.       And there you have that little piece of poetry.  Damn beautiful if I do say so myself.  '),(14079,1675646,'13,January,2004','2004-01-13 00:00:00','   What is love?  Baby don\'t hurt me, don\'t hurt me no more.  Let\'s face it now, Haddaway did not have a way with lyrics.  He must have had some pretty messed up relationships.  Hmmm I don\'t think I came on here to talk about that.  I\'m not sure what I came on here to write about.  I was listening to music in bed and it always makes me feel like coming out here to type something out.  What that thing is I have no idea.  Coolest chick in the world who wants to become an Egyptologist Laura has introduced me to the utterly bizarre show that is Aqua Teen Hunger Force.  The first cartoon I have ever seen that was definately written by members of the Marijuana Party.  Anyway check it out if you can, it\'s hilarious and so weird you\'ll forget what dimension you\'re in.    Right now I\'m just stuck in a state of mass confusion.  I just don\'t seem to know anything.  I need to figure alot of things out but I seem unable to even figure out what the options are.  I\'m very frustrated.  Hopefully after tomorrow night at least one thing will be sorted out.  I\'m trying to figure out whether or not I am still together with \"you know who\" who is now in Fredericton, I\'m not sure for how long.  I would very much like to know where I stood there, I just want to know whether I am single or not essentially.  Not like people have been beating down my door or anything.  I\'m hoping I don\'t choke in my upcoming job interview.  It\'s not like it\'s the be all end all of jobs, but I really need to move on and I would like to do it before I get fired from my current job.  I really want to be able to walk away from this job with my head held up, instead of down.  But I do need to walk away, and I can\'t do that until I have another job.  Anyways....  '),(14080,1675646,'04,January,2004','2004-01-04 00:00:00','   So I\'m listening to a cheesy song.  Well that\'s what I\'m doing now.  Before I was reading the great book I got for Christmas; \'At Swim, Two Boys\' by Jaimie O\'Neill.  It\'s a hard read for me, not just because there seem to be so many words that I don\'t understand.  But moreso on my own psyche it\'s pretty difficult.  If the end of the book isn\'t what I\'m hoping it will be I\'ll be very upset.  And the other main thing that bothers me, is that the book is so romantic and beautiful, and it makes me realise that I have never experienced anything of the sort in my life.  And I really don\'t expect I ever will.  Now if I were to play the blame game, I would probably place my wish to fall in love and have a great romance on my friend Natalie.  Sorry, N-Babe...I\'m not saying this is a bad thing but I wonder sometimes if I would actually be bothered by loves abscence if it weren\'t for the queen of romance herself.  As I have alluded to before I consider myself to be a big ole puzzle.  The big piece in the middle that is missing is my love and my soulmate, and everything will fall into place around that if only I could find it.  And yet, it doesn\'t seem like it will fit perfectly even if I do find it.  On top of this stuff a customer called me ugly today.  On top of that I made a couple of judgement errors Friday night, so now I am left desperately trying to hold onto any remnant of positive mood.  As anyone knows nothing quite bothers me like making a mistake.  Ugly??  Imagine that, eh.  I mean, I think I\'m unattractive, but ugly my be a tad harsh.  Certainly if you go more than skin deep I might be construed as being ugly.  Mind you in that sense I could also be construed as being beautiful.  I am nothing if not empathetic, I try my best and sometimes how much I care for others can affect me negatively.  In case any of you wonder why I don\'t show my empathy much that would be why.  But when it comes to myself I don\'t allow much \'self-pathy\'.  I\'m not sure if there is a word for empathy of the self.  I don\'t give myself a break.  I care very little for myself and my value.  In fact I value myself alot less than I am sure most of you would begin to think I do.  So you ask the question, am I depressed?  I would say no.  This is the way I live my life, constantly in this purgatorey of mood, I am neither here nor there but I am more likely to fall then to ascend.  So it goes.  '),(14081,1675646,'22,February,2004','2004-02-22 00:00:00','   And for the record, I don\'t give a rat\'s ass about my spelling and grammar mistakes.  This isn\'t a forum where that is really necessary.  I\'ll fuckin\' screw up my sentences all I damn well please.  And that ain\'t not bad.  I ain\'t not a dorkus, and so on.  '),(14082,1675646,'15,February,2004','2004-02-15 00:00:00','   Dammed Fry, I can\'t teach.  I\'m a Professor.  That about sums up the University experience right there doesn\'t it?  Don\'t get me wrong.  I have had some very good professors, but, I have also had a couple that have no business in a job that pays well over 70,000.  If you want to know why tuition is so high why don\'t you take a look at what your school is spending money on and quit looking for money the Government not only doesn\'t have, but realistically doesn\'t need to be giving to schools in the first place.  Say what, you say.  Government gives money to universities more as a courtesy.  The Government knows that this is where the next batch of workers come from and we want a well educated work force not only for domestic reasons but also to improve ourselves on the world stage.  Every university in Canada is in effect, a private organization.  And thusly, should not technically be given public money.  Obviously for the greater good they are given money, which is fine.  But I really think that if you want tuition costs to freeze or go down there is nothing the Government can do to help.  They have enough things to poor money into.  You know like sponsorship and things.  I\'ll make that point now before anyone gets all high and mighty on me.  And another thing.  What\'s the big fucking deal about that??  This is hardly the first time that tax payer money has been used for inappropriate things.  I\'m willing to bed that almost every Government since Confederation has lost money somewhere.  Are we going to collect from the corpse of Sir Wilfrid Laurier?  Probably not.  A Government is a large institution.  All the more reason why it should be downsized.  Hello Senator *glock glock*...See you in hell you rich do nothing bastard.  How do you like that for downsizing???  Again, please don\'t get me wrong.  I am upset about this.  But this is part of what living in a Capitalist society is...Business will always be grabbing at the pockets of government, and sometimes they\'re going to get something.  I find it hard to be outraged.  I would rather have my Government misuse 150 million dollars, then misuse billions.  You know what I\'m talking about.  *Cough cough* war in Iraq *cough cough*.  Well that\'s enough ranting...And I didn\'t even talk about anything personal.  Not that there\'s much to say.  I mean...Considering how pathetic I view myself to be, no wonder nobody ever gives credence to my opinions.  As true as they may be.  '),(14083,1675646,'15,February,2004','2004-02-15 00:00:00','   I am supposed to go to lunch with my friend Michelle today...Well it\'s technically Sunday at 2:55am...Later today.  I mean, I will go to lunch with her I forget when she said she was going to call though so I\'ll have to get up and get ready around noon.  No sleeping in for me, plus I have to come home to put the ham in the oven.  Some of few are probably snickering at that.  Get your minds out of the gutter you perverts!!!  Someone wrote to me; \"Happy VD Dane\".  To which I reply, \"Happy Venerial Disease what the hell kind of friend are you?  You know I\'m very sensitive about my STD\'s.\"  I know it\'s a bad joke.  I don\'t have any STD\'s for the record.  I don\'t know what I want to write about.  I am surprisingly stunned by something I was watching on TV.  And despite the obvious social message it is delivering I think I was more affected by how useless the teenage girl was.  I mean, she enters into a competition where you have to kill the other contestants but she does jack shit besides get beaten to death by an old man.  Jeez.  Pathetic.  And yet very disturbing because her family and boyfriend watched it happen.  It was called \"Series 7\" or something to that affect.  Anyways interesting....I also caught the tale end of John Carpenter\'s \"Ghosts of Mars\".  That was much less interesting.  So...\"Fellow\" has reappeared on my MSN list.  I suppose he\'s expecting me to be all excited and message away.  I don\'t think I\'ll do that.  Don\'t get me wrong, I do really want to talk to him and everything.  But I want to know if he actually wants to talk to me.  The only vibe I ever got from him was that he was using me for something and didn\'t really give a shit about me.  And the fact that he would ignore me over so small a thing really points to him just needing any excuse to remove me from his life.  As if moving to another province wouldn\'t do it??  Anyways...I the I am going to go listen to ABBA or something.  Ain\'t got nothin\' better to do until the morning comes.  '),(14084,1675646,'11,February,2004','2004-02-11 00:00:00','   You know it\'s a week since the last entry and although there were lots of ups in the past week I am right back to where I was.  It is an inevitability of being me I guess, I\'ll always come full circle, especially where the circles origin is negative feeling.  I\'m not making any sense.  Sometimes I get the feeling that I never do, my way of thinking doesn\'t seem to fit within the reality of the world that we actually exist in.  Not to mention the people who inhabit said world.  I care but I don\'t.  I\'m not sure whether to think of it as apathy, or as a form of hypocrisy.  I want to be better, to get better.  But I am either unwilling or incapable of making necessary changes.  Malaise forever.  I went to bed last night at...hmmm...I\'d say 3:30am, although didn\'t fall asleep until 5 at least.  And despite getting up for a couple bathroom breaks, and missing 2 appointments in the process I stayed in bed until 6pm.  Really I didn\'t even want to get up then.  I didn\'t feel bad or anything until after this...That\'s the problem.  I hate it when I stay in bed all day, but I still don\'t get up.  And negative feelings are an unavoidable result of this.  It\'s the same as when I have a one night stand, or brief relationship, I know I\'m going to feel guilty afterward but I do it anyway.  Do I really hate myself that much?  I guess I would have to say I do.  Although I\'m not sure what I did to deserve the hatred.  Maybe it\'s more things I haven\'t done, or that I should have done.  A lot of my self-deprecation is to keep people at a distance.  It doesn\'t work most of the time, it just makes people more interested in me.  At least it roots out all the shallow people.  Right now there are three people who are interested in me in some way.  Two are in the navy, one of whom is out at sea the other who is French-Canadian.  The other is a cute Taiwanese student at Acadia.  Now as of right now, I am not interested in any of these people.  I\'m not even sure if I want to be friends with any of them.  Two of them seem a little needy, and the other one seems like a whore.  Two I would have problems communicating with, and two likely lack the kind of intelligence I require in a person.  Such is that situation.  We\'ll see how it plays out I guess.  I still miss \"fellow\".  He still isn\'t talking to me, but I guess it\'s really showing me what kind of person he is.  Which is \"not a very good one\", as sad as that is to say.  '),(14085,1675646,'03,February,2004','2004-02-03 00:00:00','   How pathetic do I feel right now?  Well I don\'t have an answer for that but certainly I am feeling about the apex of patheticness at the moment.  I got up this morning...key word is morning...but after eating breakfast, I was cold and got under the covers once again and slept until 4.  How absolutely wonderful it is to sleep for so long and to never accomplish anything.  It\'s almost like my life, with the exception that in most of my life I have been awake when I don\'t accomplish anything.  And while I\'m in this mood I may as well go on about my new theory of how my Dad likes my sisters boyfriend more than he likes me.  I am not saying love here...love is unconditional I know he loves me.  But on Sunday he didn\'t ask anyone if they wanted more garlic bread, there were three pieces left and he brought them all down and gave one to my sister, one to her boyfriend and the other for himself.  I realise it was just the mood I was in that day but it really bothered me.  He didn\'t even ask me if I wanted it.  And I did want it too.  Then it seems like I just get looks...and I know what they (my parents) must be thinking, I mean, I\'ve been back for 4 days and haven\'t done jack shit.  I mean yes my back hurts but that\'s no excuse to spend the bulk of my time sleeping.  But somehow I need to get the idea of my head that I can hide from the world while sleeping.  Really that\'s what it\'s all about is hiding.  When I\'m asleep I don\'t have to think of any of the things that plague me during my waking hours.  It was bascially a guaruntee that I would become depressed after moving.  In the period of adjustment there is always that weakening of my mind and will.  But I really don\'t want this to happen, I have so many things that need to get done around here...well not just around here, I have alot of stuff that needs to get done.  But the only things I have any attention span for are sleeping and staring at the tv...now that\'s pathetic!  '),(14086,1675646,'01,February,2004','2004-02-01 00:00:00','   Well there\'s two things I suppose I need to talk about.  My apologies for no updating of late but I\'ve had to move and all that crap.  So first thing....\"fellow\" is still not talking to me.  And although he\'s explained why I can\'t help but be frustrated.  I don\'t think he understands how hurt I am over the whole thing.  I feel awful for saying something I didn\'t mean, and I realise I pretty much brought it on myself but we should be able to work it out.  I am sick of coming online for the sole purpose of hoping he\'ll talk to me.  I wouldn\'t be going on gay.com at all if it weren\'t for this situation, I am quite tired of that room and all that goes with it.  *Sigh*...all I want is to be friends again...or to make him understand that I really care about him.  But that seems so hopeless.  Second the first boy I had a crush on story.  I\'ll mention this person by name because the chances of them reading it are slim, and even if he did I would hope he would be flattered as opposed to offended.  So yes...it has been 10 years plus since the last time I had seen him, as a rather small blonde haired, green eyed boy.  Tyler Field is his name.  We played basketball together one year, and I so desperately wanted to make sure he played again the next year that I managed to get his phone number and call him a couple days before registration.  His eyes are as close to perfection as I could imagine, that\'s really where the appeal is.  Don\'t get me wrong, he\'s a good looking guy, but they eyes are distinguishing.  I so wanted to be his friend, to hang out with him....just to be around him...and yes I always hoped he was on the skins team in the scrimmage.  So I ran into him at the store....and I looked into his eyes, and was taken back to that time.  A time of awakening I suppose.  Desire awakes, and the mind flourishes with thoughts of romance, love and the ever elusive happiness.  There\'s a part of me that really wants to ask him if he\'d like to hang out sometime...realising he isn\'t gay...I don\'t want to go on a date.  But I always thought he and I could be good friends, and I didn\'t sieze the opportunity back then.    Such are things in this crazy mixed up world.  My back hurts, I miss \"fellow\", and I love that boy who first brought out my gay self.  '),(14087,1675646,'24,March,2004','2004-03-24 00:00:00','   Some one explain to me what the hell just happened.  I was innocently searching for jobs, when all of a sudded the life was sucked out of me.  I have been in a good mood.  But it has floated away, as though it was only a vestige that had briefly taken refuge within my mind.  I need to get a new job, or a second job.  But what am I going to get.  I am not qualified for anything really, and potential employers don\'t look at how capable you are to do a certain job.  I like that sentence because at least it implies I do believe I am capable of something.  I try so hard to be able to help others through their insecurities.  It\'s hard for me to do it, having plenty of my own to deal with.  But that gives me a good insight on what other people have to work with.  Which is why I can safely tell people to pull their heads out of their asses and smell the fresh air.  Guess what folks??  You\'ve got talents.  How unfortunate for you that you can\'t live in this dream world where you are useless, and going nowhere.  That world doesn\'t exist.  You live in the real world.  Where if you wanted to accomplish something, you have the tools to do it.  But of course, you\'ll never believe what I say.  You\'ll just continue to make small obstacles into gigantic life movement blocking structures of complacency.  And I know what you\'re thinking...I\'m a big fucking ass hypocrite for saying all this shit about others and not myself.  Well guess what??  I love my friends and family...but they don\'t see the things that are lacking in me.  And there are many many things.  I don\'t possess the life skills to go far, not right now anyway.  I have no skills to get another job that is better than what I have.  And even if I did, I likely don\'t have the dedication to do it.  I don\'t want any of you to think you are being singled out here.  Over the past few days I have had similar conversations with at least 4 people.  With talk of \"lack of accomplishment\" etc.  Well what you define as accomplishment and what others may see you have accomplished are two different things.  Personally I feel that anyone I know who has fallen in love, regardless of whether they fell out of it has accomplished a great deal.  It is certainly more than I have done.  To me the idea of making over 20,000 dollars a year is a pipe dream.  I\'ll always say that I don\'t want to argue about who has the worst issues.  I am working on mine, and I feel that if you have similar issues you need to work on them.  The sooner the better, hopefully before you get to the point where I am.  Which is essentially my last ditch effort to save myself.  Well i think I have gone on long enough and not really done much except possibly offend some friends.  Which is never my intention.  Nor do I wish to make them think they can\'t talk to me about these things.  You really can.  I am a good friend, or at least I try to be.  It\'s about the only thing that I do try to do well.  '),(14088,1675646,'07,March,2004','2004-03-07 00:00:00','   It\'s been a while since I have posted anything.  I am not sure why that is.  Who wants to read what\'s on my mind anyways?  I know I don\'t.  It is always fun trying to live your life resgined to the fact that you can never be anything.  I always see people I know my Dad knows...and would love to say, hello I am Rick Butler\'s son.  But I don\'t, and not because I am not proud of my father.  But rather, because I don\'t want people to lose respect for him because I am the son he had.  Let\'s face it, I am pretty emberassing.  I have little education, I work a part-time job that I will get fired from more likely than get promoted.  I sleep 12 hours a day, I have no self-esteem and seem hell bent on making sure I never get any.  And I can\'t change.  What is more pathetic than a lack of ability to change and adapt.    My counsellor doesn\'t agree with this assesment, but that\'s easy for someone with an MSW to say.  Living your life just because you love people enough to realise they might miss you if you weren\'t around is very tiresome.  It really is a good thing I don\'t take my own feelings into account much because lord knows if I did I would probably be picked out of the harbour by a net and a pool skewer.  I mean really if you look at what I am good at.......the only things I am good at are having no self-esteem, and being pathetic.  Well three cheers for me on accomplishing that much.  Otherwise I seem always to be caught in the in between:  I am too ugly to be attractive, but I am too attractive to be ugly I am to intelligent to be stupid, but I am too stupid to be intelligent I am too straight to be gay, but I am too gay to be straight I am too good to be bad, but I am too bad to be good  I could do that all night but what\'s the point really...it\'s not like I am making any sense.  I\'m sure many people just think I am being self-depricating for the sake of it.  Alas, no.  For I am self-deprication incarnate, no compliment can pass my walls to attack my self-image.  Pleasant thoughts must be purged forever lest they create a stable mental environment.  '),(14089,1675646,'26,April,2004','2004-04-26 00:00:00','   Why is it telling me I have an unsaved post.  Lord I suck with computers.  '),(14090,1675646,'19,April,2004','2004-04-19 00:00:00','   Ooops...I meant to put my email address in the last post.   joes_mon@hotmail.com  there ya go.  '),(14091,1675646,'19,April,2004','2004-04-19 00:00:00','   Get this people I was actually in bed tonight before 1am.  And I don\'t work tomorrow until 4pm.  That\'s a rather odd thing to happen.  Of course now it\'s 1:45am and I have come in here to write and entry on ye ole blog.  Aside from it being Sunday, which seems to be my main update day.  I suppose I was prompted by a couple of things I was reading in bed.  I was reading my book \"Disturbance\" by Jaimie O\'Neill, who also wrote \"At Swim, Two Boys\" which I can\'t possibly rave about enough.  The other was a letter I came accross on a day where I actually unpacked a little.  I found it a while ago but haven\'t bothered to read it.  I know what it says.  I suppose I didn\'t want to read it due to how utterly complimentary toward me it is.  And as we know I am not a huge fan of a compliment directed at me.  The letter I recieved last year as a result of the writing I did for the school paper at MSVU.  Sure the guy is probably a crack-pot.  That certainly remains my opinion.  Especially considering his high praise for someone whose grammar skills are  elementary school level at best.  However at the heart of it is the real question.  Am I a good writer??  Aside from my poor grammar of course.  Could I become a good writer??  What would I write about??  What is my voice??  My cat Agatha just spontaneously looked up at me and now wants pets.  I often wonder about animals and the way they read me.  Especially after a certain incident with my now deceased cat Spooky that happened a few years back.  Animals just seem to have that sense.  Truly amazing if you think about it.  Anyhow I have lost focus on this.  I am not sure if any of the questions above are meant as anything besides rhetorical.  But if you feel like answering I\'ll listen.   urlLink   That\'s me...although perhaps I will take up this offer to try out the google mail.  Tell me if I should do that too.  '),(14092,1675646,'19,April,2004','2004-04-19 00:00:00','   Another day, another failure.  Does it matter what the failure was?  At this stage of my life I doubt it matters much.  Just add it to the pile.  Anyhow, on an entirely different note...I caught a very odd thing tonight.  On the Family Channel starting at 11:50pm was a program I have not seen in years.  In fact I am the only person I know who even remembers it.  It\'s called Earth Star Voyager, it was made around 1987 or so and doesn\'t really star anyone you may have heard of.  I am certain it was the pilot for a show that never got picked up, and was just released as cheap network filler material.  As I recall it had aired on ABC and/or CBC during the Wonderful World of Disney.  Is it cheap??  Oh yes.  Is it cheesy??  Beyond a shadow of a doubt.  Does it remind me of good memories??  Indeed.  I pretty much was raised by TV.  That was more of a voluntary thing.  Trust me my parents certainly tried to get me away from it.  And for many years succeeded.  I only had vague recollections of this thing.  I knew if I ever saw it again it would be horribly lame, which it is.  Although it does have some redeeming qualities.  But oh, how I wanted to the child genious character....and oh how cute I thought he was.  And for the record at the time this show originally aired that actor was probably older than I was.  Oh how I wanted to travel the galaxy to find a new home for our civilization.  I guess that\'s feeling that was brought out later on again by other space epics, like Star Trek and Star Wars.  I called my friends Adam and Laura as soon as I discovered it was on.  They were off watching Kill Bill 2 so I left a silly message as I am apt to do.  Here\'s hoping they watched a little of it.  Just so someone knows what the hell I am talking about.  Incidentally Kill Bill 2...excellent...I could say more but it seems moot.  Chances are if you\'re someone that knows me or finds me at all interesting then you already know how good it is and don\'t need me to give a more detailed opinion.  That\'s all for now.  Let\'s see what this week brings.  '),(14093,1675646,'12,April,2004','2004-04-12 00:00:00','   As Nat said, \"this blog is far too self-depricating\".  Well that it is my friends.  Which never seems to be my intention, but does always seem to be the result.  So is this one going to be different?  Not as far as I know.  But we\'ll see how it ends shall we?  One of the reasons I guess I have been so slack when it comes to unpacking is as follows;  I don\'t like to be reminded of what my life is.  Or for that matter who I am.  I don\'t like to be reminded of my uselessness.  Even though I go to bed every night and tell myself I am useless regardless.  So I know, if I feel I am unaccomplished I should go out and do somethign about it.  Well, things are not that easy.  Or maybe they are.  It\'s not like I have tried anything.  I realise this Blog tends to get more personal than I ever meant it to be.  I suppose that\'s because I have a hard time talking about this stuff with even my best friends face to face.  Not that I necessarily want to.  But at least this way if you want to know the inner workings of my mind you may occasionally catch a glimpse.  It\'s hardly Babylon I know.  If only the things that other people saw in me were actually there.  What a person that would make me.  But nobody can live life in \"if only\'s\".  \"If Only\'s\" are basically regrets.  But also, in a way, we all need to have our regrets.  Do I ever make any sense???  Am I even capable of making sense???  Can I ever use less than three question marks???  '),(14094,1675646,'09,April,2004','2004-04-09 00:00:00','   It is a shame I am not a teacher, for there are so many lessons people need to learn.  I sometimes wonder if most people comprehend the gravity of the word \'genocide\'.  As I recall nobody leaped to Rwanda to stop all the killing.  In fact well over 500\'000 were killed before anyone batted an eye lash.  I mean the people who were being killed no doubt batted their eyelashes right before they were hacked to death with a machete.  Where was the outrage???  It is 10 years later, and we honor the memory.  Which is good.  We should never forget.  But I have had people tell me they didn\'t realise how big a deal it was.  Granted at the time I don\'t think I did either.  I did learn much sooner than now, however.  Does nobody comprehend that over one million people were senselessly butchered???  Now we have the most powerful nation in the world hunting down those for killing 2000 people.  It\'s a shame not all people are allowed that kind of justice.  Actually, forget shame.  It\'s not fair.  Of course life is not fair.  That may even be the meaning of life, at least for some.  Bah!  I lost my train of thought.  Which happens all too often and explains why I can talk big but can\'t back it up.  Thoughts are one thing, and actions are another.  Although it is true that words can be powerful.  It will take a lot more than I am capable of to make people care...to make people see.  '),(14095,1675646,'05,April,2004','2004-04-05 00:00:00','   Our weaknesses are as much a part of us as our strengths.  We all have the ability to be fragile as glass, and yet solid as granite.  I am just realising that my last post is gone.  I wonder why that is?  There are few updates to give of course.  I officially am the heaviest I have ever been.  Will this prompt me to change the way I eat and live my life?  Probably not.  It really should, but it won\'t.  I have finally started talking to \"fellow\" again.  I am looking for jobs.  Wish me luck, because my unqualified ass is going to need it.  I have met a couple of interesting people.  But it\'s not like that interaction makes me feel any better about things or myself.  And it\'s not like I am meeting people who even live in the same city as me.  My friend Jay moved to Toronto this past Friday.  Good for him is all I can say, and I wish him the best of luck.  He has joined both the gay soccer, and basketball leagues there.  You want to talk about things that make me jealous.  Ah well, I\'m too fat to play sports now....Six foot one and 185...I may as well be a derelict, stuck in a hammock all the long day with five fans around me cooling the air and spreading my foul stench throughout my trailer in Eastern Passage.  I hate Sundays.  No matter how good my weeks are, everything always resets.  And then I have to start anew, trying to get the week to go my way.  That is until the next Sunday when it resets and I must begin again.  This is also why I hate the word cyclical.  Well I am going to try to find out what happened to my last entry.  Hopefully I will start updating more than once a week.  And even more hopefully will write something that\'s not quite so self-centered.  '),(14096,1675646,'29,May,2004','2004-05-29 00:00:00','   I\'ll keep that last entry in mind as I write.  Of course it is very early Saturday morning now...umm 1:22am that is, I have to get up for work in about 4 hours.  Now considering the time of week of course I am feeling fairly positive.  Well, positive for me at least.  That is to say I have no major interest in crawling into a hole to die at the moment.  Hopefully I can make it through the weekend with this mood in tact.  Went to see Shrek 2 tonight with Laura.  Hilarious.  I especially enjoyed the Fairy Godmother doing a ballad version of Bonnie Tyler\'s \'I need a Hero\'.  In fact, I found that so hysterical that I had to grab onto Laura lest my laughter float me off the planet or something.  Apparently I have resorted to clutching to women in times of joviality.  Time to throw out the lube and the boy porn???  Also worth mentioning is that Laura seems to have a rather nice bicep.  What\'s up with that?  Keep pumpin those guns girl, I\'m gonna buy that heavy bag eventually.  Keep pumpin\' those guns girl???  What the hell am I saying.  I think there was something weird in that Cinnabon I ate.  '),(14097,1675646,'23,May,2004','2004-05-23 00:00:00','   Well it is Sunday.  Or at least it was.  It is after midnight so I guess that makes it Monday.  Regardless, it is Sunday so we can all guess what kind of mood I\'m in.  You\'d think I would be able to get over this whole \"life is pointless\" thing by now.  How did I ever end up this way?  If poor self-image and depression were strictly the result of environment then I probably wouldn\'t be in this state.  I  think it really comes down to letting go.  Letting go of negative thoughts, memories, traits and images.  That\'s easier said than done of course.  But I still let things that happened when I was 3 years old bother me.  You know, I kick and I scratch and I fight everyday to take some semblence of positivity away.  And yet at the end of the week it is inevitably lost to some unknown force that seems hell bent on ruining me.  Then again, perhaps it is not so unknown.  I would dare say it\'s me.  It has always been me sabotaging my own efforts to improve and progress.  I also don\'t understand why I always make entries like this one.  Yeesh.  You\'d think I could write something light hearted and fancy free for once.  I am capable of that with some effort.  When did this become a tome to my self-loathing???  Time to go to bed, hopefully to not ponder any further questions of major severity.  Wishful thinking I imagine.  '),(14098,1675646,'15,May,2004','2004-05-15 00:00:00','   Oh lord what am I doing up???  I have to get up for work in three hours.  I was just laying in bed with a million thoughts coming to me, many of them surprisingly clever.  Mainly focusing on an idea for a book I had with a title along the lines of \"Opinions that don\'t matter: A guide on how not to think.\"  Which would essentially be a book of some of my opinions and theories.  Most of which many people don\'t know.  It is this laying in bed and thinking that has made me realise that someone should invent software that will take the thoughts I am thinking and transfer them to the computer.  I am so much more clever when simply thinking as opposed to speaking or writing.  Everything is so much more cohesive in my head.  Well, at least in terms of certain things.  Well I guess it\'s time to try sleeping again.  One more day of work to face then a day off.  Which of course is Sunday, traditionally my worst day of the week.  Marvellous.  But you never know how this week will be.  We\'ll see how it goes.  '),(14099,1675646,'13,May,2004','2004-05-13 00:00:00','   I have been playing around with the features on this thing.  Here\'s hoping I didn\'t somehow wipe out my whole history.  Mind you would that really be any big loss.  It\'s not like I\'m changing the world here.  The argument that this provides some form of catharsis for me is also a rather specious one.  I\'m not even sure what I want to write about.  This week does not seem to be going my way, for whatever reason.  Mind you my state of mind is akin to a porcelain vaise on a thin table in the middle of a huge party.  ie; it\'s very precarious.  Prescription drugs or no, I still seem to be tottering between positive lucidity and my always popular frigid dark of complacency.  Hmmm I like that...frigid dark of complacency.  I\'m goint to remember that one.  I\'m at a loss right now.  Ummm...well I finally finished my handwritten journal.  It took me long enough.  It was started in November 2000 for chrissake, but I took a few long lay offs from it.  Still it\'s interesting to read back on some of that stuff.  Now I just need to fill up the other 3, or is it 4 books i have waiting in the wings.  Mind you, I can still remember the emotions around many of those entries.  Any entry involving say...the name Trevor tends to make me very angry.  Entries involving Christmas tend to brighten my perspective.  The end of a year entries make me feel hopeless.  The really old entries cause me to reminiss (sp?).  It\'s odd how for some reason I want to be locked back into the time of living in the crack house.  For those of you who don\'t know, the crack house was just the run down place that was my first apartment.  I do miss those days.  I even have CD\'s that remind me specifically of that time in my life.  It\'s not that I\'m living in the past per say.  More like I am not willing to let go of that time, those people, and those experiences.  I guess that\'s pretty pathetic.  Pathetic might be a bit harsh.  But as we all know I don\'t really give myself breaks.  Especially since I don\'t deserve them.  Ho ho ho I\'m not in a good frame of mind right now.  That means it\'s time for bed.  '),(14100,1675646,'08,May,2004','2004-05-08 00:00:00','   Well tonight was speed dating.  As you may recall that was supposed to happen back in April but didn\'t due to lack of interested parties.  This time around there were enough people signed up, but very few came.  Five to be precise.  So needless to say it didn\'t go ahead, but I did sit with these people and talked for a few hours.  I had a good time.  Plus the goal was accomplished; That being the going out and meeting strangers in a public setting.  Take that social anxiety!!  Although my somewhat lack of social skill did allow me to make a fool of myself a couple times.  Or so I believe.  I didn\'t mean to talk as much, or as openly about myself as I did.  Oh well I can\'t take it back so there\'s no point dwelling on that.  I also won a coupon for a free brunch at Vortex.  I believe I will enjoy that.  There is a display of art there by Trevor VanDen...ummm....spelling this wrong but here goes....VanDeneiyden...I don\'t think I\'m even close.  Very interesting pictures, some including hot boys, and one including  \"fellow\" as he is referred to in this blog.  For those who don\'t know \"fellow\" is the last person I dated for any amount of time, and I am still left with some unresolved feelings over the whole situation.  The picture makes me smile.  I wish it didn\'t but it does.  I was always smiling when I looked at him.  Oh well.  So now I end up dating yet another nice guy who doesn\'t seem to give a shit about me.  But not in a bad way.  More of just an impartial way, as though, it\'s good that I\'m there.  But it wouldn\'t make a lot of difference if I weren\'t.  My priorities are all wrong I know.  I don\'t know why I feel this need to be in a relationship.  Or why I seem to think it is the link in the chain of my life that is keeping everything from coming together.  Wow that metaphor was terrible.  The one thing I have always wanted is to stay up all night talking to someone...just talking...perhaps walking on the waterfront when it\'s later, and quiet.  I\'ve never really done that with anyone.  Well I walked along the waterfront with that professor from SMU that time, but there was no relationship like intent there.  But i have definately never stayed up all night talking to a guy about life, or even generic things like music or movies or what have you.  I don\'t want to work tomorrow.  And I have this rather strange pain in my right pectoral area...I\'m not sure what\'s causing it.  Perhaps it will flare up and I\'ll have to go home early.  I\'m not sure why I\'m hoping for such a selfish thing.  I suppose because it\'s much more likely that I will be injured and sent home from work then something more selfless, like world peace.  World peace???  Talk about pipe dreams.  Alright enough rambling for now.  I guess I am trying to make up for a lack of updates with a long entry.  Well if you read it, then good for you.  Your attention span is much higher than mine.  '),(14101,1675646,'24,June,2004','2004-06-24 00:00:00','   Let\'s play a little game shall we?  I have been making tallies of a couple of different things over the past few days at work.  Your job will be to guess the acromym for what I have been observing.  I am not giving any hints right now.  However you can ask for one through email, or post a message.  However you must ask it in a form such as \"are any of the words not english?\" or something to that affect.  It\'s sort of a Front Page challenge thing.  The first 2 days I only did the 1st group the OGG.  So here are the stats.  Fri. shift: 4pm-12 OGG=17  Sat. shift: 11-5 OGG=34  Mon. shift 3-12 OGG=30                 GLB=61  Wed. shift 630am-12 OGG=1                     GLB=14  Thurs. shift 10-5 OGG=15                   GLB=32  Now I will say this much; The two groups can potentially relate, and the may be counted in more than one category.  Meaning that when I tally something as an OGG, I may also tally it as a GLB although that\'s only happened a couple times.  Notice that the time of day also has an affect on the stats, and that Saturdays number of OGG\'s was highest.  So...guess the acromym, play along, it\'ll be fun :) For me at least.  And I deserve a bit of fun.  email is joes_mon@hotmail.com by the way, if you don\'t have it.  '),(14102,1675646,'19,June,2004','2004-06-19 00:00:00','   So after I got off work tonight I ended up getting bugged big time to go out with some current and former coworkers.  Now don\'t get me wrong, I love these people.  That was why I agreed to visit them on my way home.  But if I say I\'m not going to go out, then I\'m not going to go out and trying \"estrogenically whine\" me into it won\'t work.  But I still love you all.  What else is going on?  Ummmm, nothing.  Well...Euro 2004 has started, and I\'ve been watching as many games as I can.  This is a soccer(football) tournament for those of you who aren\'t in the know.  The other thing I have been doing is reading the Harry Potter books.  I have just started the 3rd one \"The Prisoner of Azkaban\".  I won\'t be giving reviews on it, I\'m simply mentioning it as something of interest.  And trust me, when it comes to me actually being able to read something all the way through it is something worth mentioning.  I really have nothing interesting to say right now.  I should be changing so many things but I still don\'t seem to want to do it.  I weighed myself again, which was a big mistake.  Of course I can\'t complain because I know I am willingly doing this to myself, and also that I do have a gym membership and could nip this decline in the bud if that were my wish.  It is my wish.  But I also wish to be fat and ugly, so at least I have a tangible reason for the fact nobody wants me.  I\'m doing pretty good on that end.  190lbs.  You may not think that\'s much, I am 6\'1\"...however I am now officially well overweight, and I am now 10lbs over the heaviest I have previously been, ie: 180 back in late April.  For one of the first times in my life I really feel as though there are some good options open to me.  The catch is that right now I don\'t seem very interested in pursuing any of those options.  I could move cities, I could move out again, I could just go visit some people, I could quite my job, I could visit Europe, I could go back to school, I could take teach english overseas classes and do that.  There\'s more to that list too, I could write, I could take singing lessons, I could learn how to play the guitar.  Many of those aren\'t life options, but they all are within my current means to do.  So what will I choose?  Knowing me, I will choose?  Knowing me I will choose complaicency and all of these opportunities will pass without being given any due thought.  '),(14103,1675646,'09,June,2004','2004-06-09 00:00:00','   Let\'s see now, what time is it?  Ah yes, 4:42pm, and I have been out of bed now for about, well 42 minutes.  I have got to do something about these destructive sleeping habits.  Of course I say that, but I\'m not actually going to so I suppose it\'s a moot point.  It is not like there\'s anything terribly interesting going on at night.  I usually just sit at the computer, hang out in the chat room, surf around.  Not very exciting stuff.  I really should hang out in the tv room and watch Conan, I haven\'t watched him in a long time.  I need to finish getting ready so I can go meet up with Laura.  I feel bad because I am not in the mood to spend time outside, and that\'s what she wants to do.  If it were cloudy and dank I\'d be all for it.  I\'m not one of those people who gets all cheerful and empowered over the sunshine and warmth.  As usual I am the opposite.  I don\'t like being outside under the sun and the blue sky anymore, because it is beautiful, and I am not.  '),(14104,1675646,'07,June,2004','2004-06-07 00:00:00','   I would like to start this blog by pointing out that my friend Laura is not nearly as bad a speller as she appeared to be in the comment she left after my last entry.  In fact she is intelligent, beautiful, and all other nice things you could ever say about one person.  Having said that however, let me add this special note to her: \"Seriously were you drunk or something when you wrote that???  I think we need to have an intervention.\"  Ahem...so yes...I have invented a few terms in my day.  Wait, did I talk about \"Queek\" already???  I can\'t remember if I did.  That is the term referring to a queer geek of course.  Which thusly is a term I can use to refer to myself.    So it was just Sunday, and I\'m not in an utterly terrible mood.  Although I was a while ago, and there\'s no telling how I\'ll feel when I go to bed and the sound of my brain kicks in.  Damn brain.  I believe that my own thoughts drive me nuts, but if I tried to share these thoughts, or if someone were reading my mind all they would hear is \"AFLAC!!\" I\'m not actually sure what that joke implies or if it\'s even funny.  Well I tried at least.  I think it\'s about time I went to bed.  It\'s another week.  More potential for change.  More likelihood things will stay very much the same.  '),(14105,1675646,'04,June,2004','2004-06-04 00:00:00','   I don\'t know where to start.  Not that anything of major interest is going on.  Rather, it is simply because I am at a loss as to what to talk about.    Tonight was the 3rd Speed Dating event I have gone to.  It is also the 3rd Speed Dating event I have gone to that did not happen.  It\'s starting to piss me off.  I\'m trying to put myself out there at least.  Apparently the all women event was full and was a big success.  I\'ve said it before, and I\'ll say it again.  Gay men suck, and not just \'that\' way you dirty dirty people.  It\'s not like I am expecting to find \'the one\' at these events.  Hopefully I could meet someone I would at least be willing to hang out with now and then.  It has been a long time since I have had a friend of the same persuasion to hang out with.  I miss Zack a lot.  And yet I make very little effort to spend any time with him.  There are others who have offered, and although I am interested I seem to resist.  I don\'t know why.  Maybe I\'m afraid of gay people or something.  The one thing that does piss me off about the speed dating is as follows; It take a great deal of effort to make myself look presentable.  But it ends up that there was no reason to put forth the effort.  And whoever says that it\'s not important how you look on the outside, it\'s the kind of person you are on the inside, was obviously not a gay man.  I know it sounds cynical.  It is cynical, but this also applies to me.  I have met some very nice people, who would probably be good to date, you know...if I actually found them attractive at all.  I truly believe that most people are inherently superficial.  Unless of course they are completely ugly, in which case they don\'t feel the need to judge others, those lucky ducks.  That paragraph makes me sound like a bad person.  If you know me, that means you know that I often am attracted to men who are not conventionally good looking.  This is where the \"geek quotient\" applies.  That being an equation that measures the appeal of a man or woman based on their geeky attributes.  Geekiness is totally sexy.  Nerdiness on the otherhand is highly unfortunate.  And dweebiness....puhleeze.  As for spaz...well they usually have big dicks but otherwise not attractive.  Somewhere along my way I have lost my train of thought.  Maybe my train of thought was big dicks all along.  That would be rather odd, as I am certainly not a size queen.  I have a feeling I\'ll be editing some stuff out of this post tomorrow.  But until then.......See what a big freak I am ;)  '),(14106,1675646,'29,July,2004','2004-07-29 00:00:00','   So here I am at last to update my blog; And despite the fact that it has been weeks since I have posted anything I\'ve got nothing to say really.&nbsp; I promised myself that I wouldn\'t simply write about how much I hate myself, and life...but without that fodder I\'m left with very little.&nbsp; At least when you consider that I will turn any topic into how much I hate life and myself.&nbsp; It is inevitable.  So yeah, there\'s nothing to say then.&nbsp; I asked for 2 weeks off work and that time has officially started as I just returned from my final pre-vacation shift.&nbsp; Ummmm....Go see the movie Napolean Dynamite.  This is all I\'ve got for now.&nbsp; I\'m not in a very clever mood so there\'s no way for me to come up with anything good to put in..........  My name is Shake-zula the mic rulah the old schoola!  '),(14107,1675646,'05,July,2004','2004-07-05 00:00:00','   I\'m not sure where my last post went.  The site doesn\'t seem to have updated.  Perhaps they are just behind or something.  Not that you\'re missing anything if it doesn\'t get posted.  I\'ve got nothing to say.  I actually met someone new.  Which of course is quite a feat for me.  But naturally I have my hopes a little too high as to what may happen with it.  But I digress.  It\'s my own fault I feel so inadequate compared to other people; That I feel so unworthy.  It is this attitude that I need to change, because I\'ll never be happy unless I figure out a way to do it.    Sometimes perhaps I place too much emphasis on the feeling I get when I look into someones eyes.  However I really do believe that in most cases I can tell if there\'s a connection with a person through eye contact, coupled of course with some other factors.  I don\'t know what on earth I\'m talking about, or what I even want to say.  God I just wish I could get my head together.  I just wish I could be the person, be the things that, at least in an alternate universe I am capable of being.  I know I\'ve got potential.  But potential is nothing.  Potential doesn\'t bring happiness.  '),(14108,1675646,'04,July,2004','2004-07-04 00:00:00','   It\'s about time I posted something isn\'t it?  I don\'t really have anything to say.  I\'ve finished reading the Harry Potter books.  So now I, like millions of others, must wait until the next one comes out.  I met someone knew tonight.  Someone from online obviously.  Had a good time blah blah blah.  I don\'t put much hope into anything coming out of it.  He\'s too attractive for me and I think he knows it.  Plus I made myself sound like a total dumbass, which is something I tend to do when I meet people who are of equal if not superior intelligence.  So he\'s got brains and he\'s got looks, he\'ll make someone very happy and maybe we\'ll be friends or something.  Who knows what the future holds?  I really don\'t know what to write.  I never intend to make entries about my negative feelings.  But it seems that as soon as I sit down to write and entry I\'m completely flooded with negative thoughts.  All the things that happen to me in the run of the day and at the end I always feel the same.  Now this would be a good thing of course if I went to bed happy and satisfied.  But that\'s not the case, and I can\'t see it ever becoming so.  Ooooh I got home in time to catch Harvey Birdman and Aqua Teen Hunger Force....jesss!!!!  '),(14109,1675646,'09,August,2004','2004-08-09 00:00:00','   I feel funny.  Likely owing to the fact that I have yet to eat today.  My parents always ask me whether or not I have eaten.  As though I am not eating or something.  I eat plenty, and plenty of the wrong food while I\'m at it.  If they are wondering if I\'m eating then all one would have to do is look at me as I leave the bathroom after my shower with the towel on.  Then you can see how utterly \'jiggle-rific\' I have become.  There\'s yet another term that I have made up.  I shall add it to the Dane dictionary.  I believe I am up to a whole page now.  So this week I have been on vacation.  I was also supposed to be on vacation this week however I agreed to work on Tuesday and Wednesday because they needed people.  I had forgotten that Tuesday was my birthday, but I\'m off at 4 so it really makes no difference.  Usually there\'s a big birthday dinner.  That\'s what my parents do for us kids...take us out to dinner.  We don\'t give presents in my family, just cards and dinner.  I like this arrangement.  I am a much bigger fan of food than I am of sweaters that don\'t fit properly.  This year however I really don\'t want to have anything.  Mind you I\'m pretty sure I feel this way every year, and my parents usually get me to do it anyway.  This year I probably could have convinced them not to do anything...However my cousin whose birthday is August 12th is in town and so the families want to have a one night two birthday thing, and I can\'t very well say no to that.  If you\'re wondering why I don\'t want to celebrate my birthday I\'ll run down as many reasons as I can.  Although my head is sort of swimming with lack of sustinence I can try.  I don\'t like getting older of course, that\'s everyone\'s generic reason.  Another year lacking accomplishment, lacking improvement, lacking my self you could even say.  I also hate people spending money on me.  I don\'t feel deserving of it.  Not when the money could go to much better use on either themselves or some charity.  This past year in particular is nothing to celebrate.  The only thing I can really take from it is that I\'m still alive.  That is a good thing, that I\'m still alive.  However that seems to be all I ever take from any year.  I can never find solace in the fact that I am living.  I have never learned to create my own happiness, and it seems so improbable that I ever will.  So when it comes this time of year it seems that no matter what the coming year will always seem draped in utter futility.  '),(14110,1675646,'05,August,2004','2004-08-05 00:00:00','   Have you ever recieved news and for some reason didn\'t react the way you believed you should have.  It\'s an odd feeling isn\'t it.  I was told something today which I suppose I should be angry about but I\'m not.  I\'m not sure how I\'m feeling about it.  So yeah...unfortunately my mood from the weekend has kept me from many things so far this week.  In fact I have done absolutely nothing with my vacation time so far.  In fact I have barely even left the house save for the occasional jaunt out to the deck.  It seems like any control I had over my thoughts or direction has flown away....and I\'m waiting for Charleton Heston to show up and shoot it down for me.  Is he even still alive?  The old bastard.  I hate not knowing what to do.  I hate that I\'ve never had any goals.  I hate that I don\'t know what makes me happy.  Haha...well it seems I\'ve gone back to the ole whining about my mental problems.  It\'s the old stand by when there\'s nothing else to talk about since I\'ve shut myself indoors all week.  Which I\'m sure is a metaphor for how I handle life...shutting in, never going out...getting carpal tunnel from sitting at the computer too long.  Well, scratch the last one.  Anyways...there\'s nothing new...continuing along the road of jagged rocks that we call life.  I just wish I could find a pair of shoes.  '),(14111,4042606,'09,July,2004','2004-07-09 00:00:00','   Well my plane takes off at 9:50 am tomorrow morning, so this will be my last post for a while. I am going to be in the air for 4 hours so I will have time to write some drafts of posts to come. Like I said earlier, Ian is going to take over some posting! Yay! He finally gets responsibility! Anyway, We are heading in early so I must finish packing, and I will see you all in a week!  Kisses for Marty and Ian '),(14112,4042606,'08,July,2004','2004-07-08 00:00:00','   Well I guess Jackie was trying to get in touch with me all of last night. She finally caught me today after work. The South Park Characters, are not going to be up atleast until sunday. For some reason the website will not let me upload the files. I will crack the code and get those pictures up. Anyway, as for Jackie going to Tampa Bay, that is brand new news to me. So I guess it\'s up to me to update the website at the moment otherwise like Jackie said \"Everything will be normal\". '),(14113,4042606,'08,July,2004','2004-07-08 00:00:00','   Well I finally got a hold of Ian, and I sent him the pictures. He said he was going to work out the problems while I am gone. If I haven\'t told you already, I am going with my dad to Tampa Bay, Florida. Finally going on vacation! So, even though I will have my laptop with me, I probably won\'t be posting much so you\'ll have to rely on Ian for posts! Anyway he will probably post the south park characters later this weekend. Otherwise, things will be normal! '),(14114,4042606,'08,July,2004','2004-07-08 00:00:00','   I am trying to make south park characters of me and Ian. But, at the moment the website won\'t let me post the pictures. The pictures will probably be up sometime next week, or whenever I get Ian\'s attention. He seems to be quite the little party boy, and is never answering his cell phone. '),(14115,4042606,'07,July,2004','2004-07-07 00:00:00','   A freak rainstorm has blanketed Burlington with a layer of water. I myself had arrived just a hour before hand from downtown. I then was about to walk out the door to go skateboarding. All of a sudden it starts raining, this is the second time this has happened to me this week. It is also starting to anger me, Hopefully the ground will dry up so I can go out and do what I originally intended to do.  On another note, I have a cell phone. On my cell phone I get charge a certain amount of money for the amount of minutes I use. Now that you understand this, let me say that I don\'t want to here about some dork starring at your funbags for fifteen minutes. I don\'t know how Jackie did it, but she was able to turn that one simple event into an hour long lecture about how I shouldn\'t stare at her boobs. I not only lost an hour of my life, but, I also lost money. Something I will never be able to get back.   There are somethings in this world you will never get back. '),(14116,4042606,'07,July,2004','2004-07-07 00:00:00','   Weird things occur during rain storms don\'t they? Here I thought it was supposed to be partly cloudy, when I am outside talking to my mom when she is stepping out the car. Rain starts falling like no tomorrow and it is still perfectly sunny. I was totally soaked and had to change out of the brand new shirt I just bought that day. I didn\'t even get to show it off to anyone. I guess I will just have to wait until tomorrow to show it off. '),(14117,4042606,'07,July,2004','2004-07-07 00:00:00','   About an hour ago, I got back from my friend Tim\'s house with a borrowed copy of  The Elder Scrolls III: Morrowind  for PC, I know I am a bit late on the morrowind train, but I have realized how good of a game this really is. I asked Ian to post a review of the game and he said he would think about it. I seriously won\'t be going anywhere for about a month because of this game. My wood elf scout is on her way to fort frostmoth to solve some problems!, if you are unsure what this game is about then  urlLink click here  and go to it\'s official website. And to tickle your fancy I have posted a picture of morrowind in action. Ian says I post to many pictures but I don\'t give a shit. '),(14118,4042606,'07,July,2004','2004-07-07 00:00:00','   The Burlington bike path has been my main mode for transportation for the past week. Taking me downtown or into Leady Park. But, there is one problem I do have with the bike path. The bitches that think they are the only one there and take up the whole damn stripe of pavement. I\'m talking about the people that jack knife their crappy ass bikes in the middle of the path, then act like you are some kind of burden when you try and get around them. The topper is when they let out a big sigh when you ask them to move over a bit, then they stare at you until you are out of their sight. This just proves my theory that humans are the worst thing to happen to this planet, of course, not including me. '),(14119,4042606,'07,July,2004','2004-07-07 00:00:00','   As I was in EB Games down near my moms house, I noticed a nerdy kid starring at my boobs. I know some people are attracted to big chests (hint, hint) but he didn\'t need to make it so obvious when he was starring. It was like he has seen them for the first time and I was standing looking at a copy of super monkey ball. I turned to him finally and asked him what the fuck he was looking at, he just said \"nothin\" and walked away while still starring at my chest. If my boyfriend or Ian were there, they would have probably punched the guy out. I decided to let it go and just went and got an ice cream.  On another note I bought super monkey ball, and I love it! The only thing bad is I feel bad bouncing the monkeys around, it looks like they would die because they were spinning so fast. It is still a fun game though! '),(14120,4042606,'07,July,2004','2004-07-07 00:00:00','   urlLink    Hey! Guess what everybody, the new Ipod Mini is going to hit shelves on July 24 so start saving money so you can buy it! I know I will. '),(14121,4042606,'06,July,2004','2004-07-06 00:00:00','   After Ian posted about the old guy, I have decided to share my experience with you all too. After hanging out with Ian downtown (and seeing hitler) I went home to chill. I jumped onto my dad\'s couch and flipped on my tv when I heard yelling outside. I walk over to see what\'s going down outside, and I happen to see a shirtless fat guy screaming at his poor wife. He was yelling at her because she never respected him, So I yell down and tell him to shut up, and his wife starts yelling at me for talking shit about him. I thought I was helping to, So I quickly shut the window. '),(14122,4042606,'06,July,2004','2004-07-06 00:00:00','   Today I decided that I would go for a quick skate session down at my favorite spot, also known as hanny\'s cirb. Well, as I was walking back down the sidewalk, I noticed a old man facing one of the baby tree\'s that were growing along the sidewalk. Unfortunately I realized what the old man was actually doing. He was urinating all over the tree. I am aware that nature calls, but, there was a bathroom right inside Hannafords, and also not right in the middle of public viewing. Later, it was brought to my attention that this old man should be allowed to piss where he wants because he is old. Yet, I am not allowed to piss in someones closet on account of revenge.   This world is clearly a little off '),(14123,4042606,'06,July,2004','2004-07-06 00:00:00','   Let me tell you what really happened with \"Mommy Hitler\". We were walking downtown when we saw these dorkish looking girls playing their violins (poorly too). We were walking down Church St. when we see these people, and Jackie turns and says to me \"Why don\'t we just give them a dollar?\" At first I thought that, those kids must be pretty poor to be playing a cardboard violin on church street, when while Jackie is tossing the money in, I catch a glimpse of the kids mother. It was like she was Steve Urkel\'s white cousin.  Jackie tossed the dollar in and started to stand back up when the Hitler lady gave Jackie the look like the money was an insult. It probably didn\'t help that she was wearing my \"I don\'t drink water, fish fuck in it\" t-shirt at the time either. So, Jackie decides to toss in a 10, luckly I was there to stop her and we were able to continue on our way to Pac-Sun.  The mom probably was pulling some hustling gig on her kids... Clearly humanity is in trouble. '),(14124,4042606,'06,July,2004','2004-07-06 00:00:00','   I saw a really sad thing downtown when I was with Ian today. 2 little girls were playing their violins while wearing those little tap shows and standing on white tile. I was happy to see them busy, but then, I look and their mother is standing behind them, judging them. I thought it looked like their mom was forcing them to play for money. I felt bad for them so I tossed a buck into their violin case, then their mom looks at me with those eyes like \"I hope you were planning on giving more\". I was about to when Ian pulled me away, cause I was planning on giving them more. '),(14125,4042606,'06,July,2004','2004-07-06 00:00:00','   Soon I will be posting pictures of me and my friends skateboarding. Well, as soon we can get everyone together, and get a digital camera. Those are the only real issues we have at the moment, so look for some pictures to arrive shortly. '),(14126,4042606,'06,July,2004','2004-07-06 00:00:00','   It has been brought to my attention from many people, that skateboarding is no longer a okay thing to do with your free time in Burlington. Everywhere I go now has some kind of sign that says I can\'t skateboard there. I am sick of this, because when I do complain. They just tell me to go to the local skatepark down at the waterfront. I hate that fucking place, it is filled with assholes that think they are better then you. Well, let me tell you something, I don\'t give a shit if your better than me, I just want to skateboard in peace. This is nearly impossible now because of those jackasses in suits telling me that I am going to break something if they let me skateboard in their parking lot. '),(14127,4042606,'06,July,2004','2004-07-06 00:00:00','   John Kerry finally has made his decision tonight about his running mate. Luckly, he did not choose Howard Dean, but, instead went with the popular North Carolina senator John Edwards. We can only hope that his attitude of liking everone will not make him weak in points where harsh words are needed. And Edwards is now pitted against the Cyborg Vice President Dick Cheney, They said he would provide a striking contrast to him, though Republicans countered that a debate between the two would be a face-off between sizzle and substance, particularly over the Iraq war, when John Edwards\'s relative lack of foreign policy experience would be an issue. '),(14128,4042606,'06,July,2004','2004-07-06 00:00:00','   \" The National Institute for Media and the Family, a critic of violent games, reports that 79 percent of stores it surveys prohibit selling mature-rated games to kids. Many major chains â€” including Wal-Mart Stores Inc., Best Buy Co. and Electronics Boutique of America Inc. â€” will have a carding policy in effect by the end of 2004, though the titles are available over the Internet.\"   This what the article (written by NICK WADHAMS) says about the subject of video game violence and what is being done to prevent young children from playing these games. Let me just say that if a 9 year old child has not seen someone get raped, murdered, fucked, shot, stabed, strangled, pistol whipped, gang banged, or paddled. Then that child parents have not been doing their job. I had seen all of that on television or heard it from music since I was 9 and I turned out only mildly retarded. The kids that do commit crimes and blame it on video games, are not victims, they are stupid and should be ass raped in prison. Otherwise I would like to see a little more of this (picture below) in video games rather then a character shiting flowers and hugging rabbits. '),(14129,4042606,'06,July,2004','2004-07-06 00:00:00','   Okay, here is the deal. I started a WMD Donation account over at Rum and Monkey so if you want to make the Wings of Kakutama stronger, I suggest you donate a weapon or two. You can do so by clicking the freaking awsome button below.  If you do not understand what a WMD Donation Account is, well, I will explain it to you. A WMD Donation Account is a little thing they started up that lets people donate \"weapons\" to a group the user names. If that group gets enough weapons they could appear on a top 10 list of most dangerous users. So do me a favor and donate to my worthy cause.   urlLink '),(14130,4042606,'17,July,2004','2004-07-17 00:00:00','   Ian wanted me to make another post on his website. So I will, since it is important to him that I do. I am still not clear on what I should put in this post though. Maybe I will put a picture of something I like, or of something I have encountered today. I do not know yet, I will ask Ian what he wants me to do, I don\'t want to ruin his website with my bad topics (just kidding Ian). I will figure something out though, maybe I will try and talk about what Ian does for me, like&nbsp;Ian was being really nice and said he was going to send me a birthday present,&nbsp;you probably don\'t know it but it is on&nbsp;monday. Ian, you really don\'t have to. &nbsp; It is almost 3:00 in the morning, and I&nbsp;have to go to sleep soon, this is the latest I have been up for a very long time, and if I don\'t get to sleep, I will get into trouble at work tomorrow. I really love my job, so I guess I need to sleep. See you (whoever is reading this) later and maybe I will be able to post tomorrow night.&nbsp;  I think everyone should also visit this  urlLink website , It has some cute things on it that I like. '),(14131,4042606,'16,July,2004','2004-07-16 00:00:00','   Your probably asking yourself why the fuck I am posting on a Saturday. Well because, to me it is still Friday. I haven\'t slept yet because I have been talking to Michelle all night and I just uploaded my new profile picture because the old one was gay. Anyway, This is the only post I will be making today, Hey wouldn\'t you know it? Right at the stroke of midnight, but, now I have to catch some z\'s before I head into work in the morning. That\'s right, my whole night was ruined and I couldn\'t go over to Becky\'s because I probably wouldn\'t of been to work until 5 at night. &nbsp; I told Michelle to post something tomorrow, I hope she does. '),(14132,4042606,'16,July,2004','2004-07-16 00:00:00','   I have no idea what the hell is going on. First Ian is talking to me, then he isn\'t, then he is, and now he isn\'t again. Plus, this Michelle chick is now a member of our website when Ian didn\'t even ask me if another person can join or not. I thought he was never going to talk to Michelle again either after what she threatened to do. I guess Ian just can never make up his mind. It is always like Ian to flip flop positions faster than the actual situation is made. I don\'t even understand him anymore, I guess I will just have to have a serious talk with him when I get back on Sunday. &nbsp; One more thing, If anyone gets a hold of Marty at all, tell him to call me. I don\'t think his cell phone is working everytime I call him the phone goes out. '),(14133,4042606,'15,July,2004','2004-07-15 00:00:00','   Guess what people? I finally got my fucking Mountain Dew hat that I was supposed to get like 3 weeks ago. Anyway, it\'s awesome we got a fresh face on the website. Kind to break the tension that has been building lately. Anyway my hat is the shit, and everyone knows it. If you don\'t you should be stoned to death, with really big rocks.     urlLink '),(14134,4042606,'15,July,2004','2004-07-15 00:00:00','   Hello, this is my first post on Ian and Jackie\'s website. My name is Michelle and I am currently in Ohio. Ian wanted me to do a favor for him and post once and a while on this website so he can see what is happening with me. Sometimes it is hard for him&nbsp;to get a hold of me on my phone, so I guess I will do him this favor and post. I can\'t post a lot because of my living situation with Adam and I am not always near a computer. I am mostly outside most of the day because I got a job working on yards and stuff like that. I hope this makes you feel better Ian. It kind of makes me feel a little better too. &nbsp; I don\'t really know what else to say, I am not really good at stuff like this. I don\'t always know what the right thing to say is so I guess I will just end my post now. '),(14135,4042606,'15,July,2004','2004-07-15 00:00:00','   I am going crazy, I just can\'t find that perfect look for the website. If you haven\'t realized, the look changes a lot. It\'s because I need change every day, else I go crazy. Ian on the other hand will leave things alone for ever and that just makes me angry. I also want to know who this&nbsp;new member&nbsp;is, and what is going on with the website. Since Ian is no longer talking to me I am completely in the dark. So if anyone has any information on this subject, please let me know. '),(14136,4042606,'15,July,2004','2004-07-15 00:00:00','   Earlier I was under the impression that I had the day off. Yet, Now I am working 11:30 to 8:00 so I don\'t know what is going to happen. Someone quit so I guess I have to step up to the plate. Anyway, my plans for tonight have to be put on hold, because, I don\'t want to be going into work with a major hangover. Then after I leave at eight, I am going to be coming in again at 3, so that gives me no time to sleep. That also gives me a good oportunity to hang out with some friends. Then I\'ll go home and sleep for the entire Sunday. So I might be taking Sunday off as well as Saturday. It\'s all cool though because someone else that I won\'t mention will pick up the slack. Also there could be a brand new member to the Kakutama Sign crew, if she is up to joining. '),(14137,4042606,'15,July,2004','2004-07-15 00:00:00','   Tomorrow is our days off from updating so you\'ll have to wait until sunday. Anyway, I also have the day off from work so it should be turning out to be a party day. Oh, I am hoping to get out and see I, Robot. Tonight, with who, I do not know. But, I will see it eventually if I can\'t today. At the moment I am feeling tired so I am going to sleep for the rest of the day, then go out and party all night. Yep, my life is an odd one. '),(14138,4042606,'15,July,2004','2004-07-15 00:00:00','   I am going to see I, Robot tonight. So tomorrow I should have the review all nice and neat for you guys. It will be longer than my anchorman review and it will be better too! Anyway I just wanted to let you all know, and Ian you should talk to me again. I have so much to tell you. '),(14139,4042606,'15,July,2004','2004-07-15 00:00:00','   Ian\'s on mute to me again. For some reason he refuses to answer my phone calls, and he won\'t reply to my emails. I don\'t know what I did to him but, it can\'t be as bad as what he did to me. I am not sure what is going to happen to this website. It may just end up splitting into two separate websites. I hope not though, I may not show it but I really do care about Ian. '),(14140,4042606,'15,July,2004','2004-07-15 00:00:00','   I just saw Anchorman staring Will Ferrell, with my dad and I loved almost all of it. It was funny and I laughed&nbsp;out loud, which is something I haven\'t done since Rat Race. Although at times&nbsp;Anchorman feels like an extended \"Saturday Night Live\" sketch from the show\'s treasure chest of Will Ferrell characters, co-writers Will Ferrell and Adam McKay&nbsp;created enough inspired&nbsp;insanity to sail past the&nbsp;stupid stretches and&nbsp;give&nbsp;a welcome splash of funny kickass movie-tude.  &nbsp;  This would be the first  Alright Alright  review. I hope you got something from this short short. I didn\'t want to blow all my creativity on one review! '),(14141,4042606,'14,July,2004','2004-07-14 00:00:00','   Ian&nbsp;has his short stories and his list of hatred, while, I have  Alright Alright .  What is  Alright Alright  you may be asking me. Well it is my title for the reviews I do. So there Ian, your not the only one with a cool section. Plus, mine has a&nbsp;uber cool&nbsp;name. I will be posting my very first  Alright Alright  topic in about 10 minutes so hold on there! '),(14142,4042606,'14,July,2004','2004-07-14 00:00:00','   I got my jacket just now. Let me tell you something, it is the coolest thing you would ever see in your life. To bad I won\'t be wearing it until November... '),(14143,4042606,'14,July,2004','2004-07-14 00:00:00','   I also should be getting my kick ass jacket tomorrow. If you want to check out what it looks like,  urlLink click here . '),(14144,4042606,'14,July,2004','2004-07-14 00:00:00','   Last night was not good, instead of going to the waterfront and ranting. I decided to play it safe and stay in my house. That wasn\'t the case because I was on my computer all night IMing people and posting random crap that shouldn\'t have been posted. Hence that long section about how much I hate people. Right now, I have a horrible headache and I want to sleep. When I was about to go to bed a spider fell off the ceiling and onto my arm. It set me off balance, because, I was freaked out and fell on my bed. Luckly I fell asleep.  Anyway, Jackie was right last night. Part 3 of the story A Figure In October will not be posted because I haven\'t finished yet. I have to write a couple more pages then I will post it. Right now though, I am not really in the mood to do anything. I really had to force myself to even write this. I just barely woke up too. '),(14145,4042606,'14,July,2004','2004-07-14 00:00:00','   I am about to head out the door and go down to breakfast, I just wanted to leave a message that I hope Ian gets. Stop drinking you crazy redskin. I think everyone in Burlington, and now everyone that can get on the internet, has seen what happens when he drinks. I also want to let everyone know that I am personally feeling a lot better, and as you can see Ian is doing better. When I get back I will deal with the other problems. Anyway I am hungry and I am going to get some breakfast, so catch everyone later. '),(14146,4042606,'14,July,2004','2004-07-14 00:00:00','   Part 3 of A Figure In October won\'t be up tomorrow, because to tell the truth Ian won\'t want to do anything. Plus, I don\'t think he\'s finished it anyway.     Just to let you know it took me a while to get the Smiley face on because of hotlink issues. '),(14147,4042606,'13,July,2004','2004-07-13 00:00:00','   I am finally seeing the old happy Ian again. Unfortunately I think he is a little out of it right now, and keeps trying to mess with the website when I am trying to post my messages. I have already tried to post two, but, there was some retard interference. There is something about a drunk American Indian/Irish guy that really brings you into a happy mood.  Ian\'s greatest math equation ->   Irish blood + American Indian blood = Biggest drunk on earth   I wouldn\'t doubt that his head is laying on his computer keyboard right now. '),(14148,4042606,'13,July,2004','2004-07-13 00:00:00','   These people I know have been screwing me over my entire life. Now I am about to tell them all off in a horribly written one paragraph of doom. You know who you are you crack whores. It\'s wonderful news to find out while your dating your second girlfriend; someone mentions to you that they walked in on her getting tea bagged. You suck Kim. Anyway you need to leave me out of every stupid conversation you piggy pig. Next on the list is that fucker Jared, who thought it would be funny to try and blow my toe off with a bottle rocket, instead shooting himself in the stomach and telling me I had to buy him a new shirt. I hate you fucker. Another one is that bitch sister of Koki. She was so ugly, I felt ashamed to look at her, because if I saw a dog that looked like that, I would shoot it because I felt bad for it. I hate you to Koki, but, not as much. On a scale from 1 to 14 your about a 5.  New paragraph of hate, lets continue the list onto all those idiots I go to school with, and the idiots I work with, and the idiots who say \"Lets put all the computer keys really close together so it takes Ian five minutes to hit it right without smacking an h or a j or some other gay ass letter. People blow.  Know who I hate the most though? Ronald McDonald, that fink-rat bastard. In every fucking commercial he is passing his rat burgers off on some little kid while he finger pops that fucking purple blob. You never see him eat the burgers himself do you? you know why, because if he did he would probably die. What a fucking stupid red head.  While were on reds, here is another person I hate. Joan you stupid freckly red headed retard, I thought you were just a little off, but I was mistaken. Your a fucking retard and I hate you. No wonder you smoke pot, If I looked like you I would too.  And who can forget Ryan, I don\'t know why I even bother mentioning him, your a jerk Ryan. If I wanted help changing a tire I would use your face as a tire changing device.  Ray\'s okay in my book.  Whatever made Georgia think I like her is gotta be a joke, I wouldn\'t touch that chick with a giant stick with little green stuff poking out the sides. Or Bamboo, because bamboo is fucking awsome.   Gimli rocks.  Vineee is chiller, like a freezer, actually now that I think of it, he may be trying to kill me. I think he is going to hire someone to accidently shoot me in the throat hole.  Jackie is an emotional wreck like the titanic. She needs to shut it and make me a sandwhich or sandwitch, however you spell that fucking word.  I need a nap '),(14149,4042606,'13,July,2004','2004-07-13 00:00:00','   Looks like Jackie beat me to the posting punch. Yes, infact there was some problems with the servers, and hopefully they have been fixed. An hour ago I attempted to post something and the website gave me the run around.  On to more interesting things though. My jacket should be here tomorrow, or Friday, since it is still Wednesday. Anyway that jacket is the coolest fucking thing I have ever seen in my life.  Hopefully no one that talks with my mother is reading this right now, but, I don\'t really give a shit. I am drunk, and in about five minutes the whole world will know how much drinky drinky Ian took in. That\'s right, the list of hatred will be posted. '),(14150,4042606,'13,July,2004','2004-07-13 00:00:00','   I do not know what is going on right now, but, I think there is something wrong with the servers this site is on. I was trying to post a review I wrote, when in the middle of uploading; the whole workstation locked up. Thank you Bill Gates, or whomever runs these servers into the ground. '),(14151,4042606,'13,July,2004','2004-07-13 00:00:00','   These couple of days have been a real downer. I have been in fights with two people I care about. I don\'t want to break up, and I want to keep my friend. The only way I can get through to anyone now is this website; since neither one is talking to me. I am so sorry, I didn\'t mean to hurt anyone. I am going to change, I promise you that. I just hope I will have my best friend, and my boyfriend when I get back... I just want everything to go back to happier days. '),(14152,4042606,'13,July,2004','2004-07-13 00:00:00','   Lately I have been having a hard time getting out of a rut of depression. I don\'t understand it but, it seems as if the whole Michelle threatening to kill herself has brought me down a notch. As soon as I got home from work my mind just sunk a level. The whole day I was in a good mood. I think I will probably go see a movie tonight, something funny, and see if I am able to pull myself out. The only thing I did find amusing today was when I threw have my bagel out the car window and hit a SUV. God damn soccer moms. '),(14153,4042606,'13,July,2004','2004-07-13 00:00:00','   The morning was crisp, as if the storm the day before had never occurred. Yet, there were hints that another was on the way. As the clouds in the distance were forming, not the usual dark gray that a normal heavy thunderstorm would have; the clouds were pure black, as if someone sucked the very light out of them. There was no light no shine, no hope that a normal cloud would bring. The only thing that these clouds brought was an ominous feeling, and a sinking in your stomach. 	As the sun was beginning to be covered by the black sheet of clouds, a police car moved down the empty road towards the beach houses. There were calls the night before of screams coming from Isabel Katertilâ€™s summer home. In the car there was Sheriff Wallace Gleyson, the most honorable man in town some called him. Sheriff Gleyson had always been liked ever since he arrived in the town 6 years ago. He moved up here when his mother became deathly ill, as he was up here he needed a way to pay the bills, so he joined the Sheriffâ€™s department. His mother died 4 months after his arrival, he just never left. 	In the seat next to him was Deputy Mark Carls. Deputy Carls was not only Sheriff Gleysonâ€™s partner but was also his best friend, he was particularly worried, because calls were coming in that there were screams coming from his girlfriends beach house. That is where he wanted to go the night before, but Sheriff Gleyson needed him down at the docks to help tie down boats. Isabel was always worrying that Mark would get hurt; he never thought it would be the other way around. 	â€œYou shouldnâ€™t worry Mark, you know Isabel knows how to take care of herself, she was probably just frightened that the power was outâ€ Sheriff Gleyson said as he turned and looked at Mark. 	â€œI never thought I would be the one worrying about Isabelâ€¦ She was always the one telling me to be careful, and now we get a call to her houseâ€ replied Mark without even showing a hint of emotion. 	â€œMark, she is alright. Trust meâ€ Sheriff finally said after a period of silence. And he then turned into Isabelâ€™s drive way, and parked right beside her station wagon. 	The two officers stepped out of the car only to see no motion in the yard. Not only that, but, there was not a person on the street or any sign of life in the entire neighborhood. 	â€œItâ€™s nearing noon, and I donâ€™t see anybody Wallaceâ€ Mark said with a worried look on his face as he quickly looked around the street. 	â€œI donâ€™t want to make you worry, but, this whole week has been strangeâ€ replied the Sheriff as he walked up the steps of Isabelâ€™s summer home. 	â€œIsabel! Isabel! Are you in there?â€ Mark screamed as he banged on the front door. 	Sheriff Gleyson decided to take a look around the back of the home. Wallace slowly walked across the deck and looked down the side of the building, where the deck continued until it reached the back of the house, were it ended in stairs down to the beach. 	The Sheriff slowly made his way across the wooden patio and peeked in the windows as he made his way out back. As he was walking he could here his Deputy Mark screaming out front, still banging on the front door. 	Wallace looking around for any sign of anything found his eyes wondering to the wooden floor as he made his way to the back porch. What he saw, was a dark stain, as if a puddle of some sort had laid stagnant for a extended period of time. 	â€œMark! Mark! You might want to come back here and see this!â€ the Sheriff yelled as he kneeled down to get a better look at the stain. 	â€œWhat is it? Did you find Isabel?â€ Mark frantically yelled as he ran around the house. 	â€œI couldnâ€™t even break through the front door, itâ€™s like something is leaning up against it!â€ Mark continued to say as he reached the Sheriff on the back porch. 	Mark looked down at the stain, and kneeled down with the Sheriff, as he looked, he face went pale and he looked at Wallace. 	â€œThatâ€™s blood!â€ Mark exclaimed as he quickly stood up and looked at the porch door. 	â€œHold on Isabel! Iâ€™m coming in!â€ saying that, Mark jumped into the porch door and busted right through. 	Deputy Carls went through the door surprisingly easily, as if the door had been broken down earlier and had been placed back in its original spot. As Mark went through his feet hit something, and he hit the kitchen floor hard. Mark lay on the floor moaning for a brief moment, and then realized where he was again. He put both hands on the floor and propped himself up and looked around the kitchen. He then felt something on his hands, like it was sticky or just had been mopped. He looked down at his hands to find a red fluid all over him. The focus of his eyes then went from his blood soaked hand to the rest of the kitchen floor. 	As Deputy Carls went through the door Sheriff Gleyson quickly moved in after him and got a good look at his surroundings. The first thing that he noticed was the entire kitchen floor was covered in blood. This sparked him to withdraw his gun from its holster. 	â€œMark stand up nowâ€ Wallace whispered as he pointed his gun towards the hallway. 	â€œItâ€™s blood Wallace! Whatâ€™s happeningâ€ Mark said in paranoia, usually Mark was very professional about situations. Today though, he was not ready to handle anything. 	â€œGet a hold of yourself Mark, Iâ€™m sure there is a reasonable explanation for thisâ€ Wallace explained, but, in his own mind he had no hope that Isabel was still alive. 	The Sheriff and his distraught Deputy slowly made their way down the hallway with their guns drawn; their destination was the front door. To discover why Deputy Carls was unable to break through, and hopefully find Isabel alive. 	They made their way down the hallway, sometimes looking at each other. Sometimes checking doorways to see if there was anything there, but, as they made their way further down the hallway, they knew they were not alone. As the Sheriff was leading he put his hand straight up, signaling that they were to stop. He pointed his handgun at the front door which was not very well lit. The shades were all pulled, and a gloomy yellow light was radiating through the whole house. It began to get darker as the sun was finally blocked out by the looming black clouds. 	Mark looked over Wallaceâ€™s shoulder to see what he was pointing his gun at and why they had stopped moving, when he saw a figure standing at the front door. Facing them, the figure did not move. Only standing there looking back at them; Mark couldnâ€™t make out who it was, or if it was male or female. All he knew was the whole situation was scaring him, and he just wanted to find Isabel. 	As they stood there in silence Sheriff Wallace finally spoke up and said â€œHey you! Put your hands on your head and identify yourself immediatelyâ€. 	The figure made no acknowledgement of hearing the Sheriff and continued to stand there. 	â€œI am not going to say this again! Put your hands on your head and identify yourself now!â€ the Sheriff yelled now apparently irritated. 	â€œDo what the Sheriff says!â€ Mark then broke his silence and yelled after the Sheriff. 	The figure then made a slight movement, as if it was growing tired and had to shift its wait from standing in the same spot for to long. 	â€œI will shoot if you do not follow my orders!â€ the Sheriff finally yells as he stares at the figure with extreme interest. 	Mark then pointed his gun at the figures leg and readied himself to fire upon it. 	Then as the stand off continued, and Marks finger started to press on the trigger, a bolt of lighting hit somewhere close by and in the flash the two police officers closed their eyes from the bright flash, and when they opened them, the figure that had been standing at the front door for five minutes was gone in a flash. '),(14154,4042606,'13,July,2004','2004-07-13 00:00:00','   I don\'t understand why Jackie continues to try and disect my mind. I am fine, I always have been fine, and she needs to stop trying to make me feel bad about things that don\'t effect me. Otherwise I would talk to her again; after the whole thing with Michelle, I had to take some time to sort things out, but, I don\'t need more than a couple of hours to do so. She on the other hand will dwell on things for what seems like an eternity. Yes, I did start writing  A Figure In October  after I got the news, but, it is not a \"funnel\" of sorrow. Jackie stop trying to figure me out. Unlike what Jackie says, I will continue posting regularly as well as posting the short stories. '),(14155,4042606,'12,July,2004','2004-07-12 00:00:00','   As of right now, I have calmed down from the horrible day of yesterday. I thought I was the only one depressed that day, but, apparently Ian was feeling down as well. I always worried about how he\'d handle things, and I still do. For some reason though he is capabable of pulling himself out of a hole of depression with remarkable speed. I think  A Figure In October  is his funnel to get rid of some of that depression. That is why he is still around. I don\'t think he will actually be posting his thoughts for a while, posting the story everyday is kind of his little vacation, a vacation I think he deserves. I just hope he keeps busy, I know what happens when he sits around and begins to think. Sorry Ian, for sharing a piece of your mind with the world, but, people need to understand. While some fall apart, Ian just shovels it into the back of his mind. I am still sad from yesterday, but, I am learning from Ian, and I to will try and find an outlet for my sorrow. Sorry this post is so depressing, but, it is something I needed to get off my chest before I became even more depressed. I know Ian will probably not talk to me for a while. He has already started to break off communication with me. '),(14156,4042606,'12,July,2004','2004-07-12 00:00:00','   Itâ€™s nearing the end of October in this quiet little village on the coast of Maine. As winter rolls in, the storms become harsher and harsher, until the snow storms finally begin. This year though, the storms have seemed to be harsher than usual. People in the town were beginning to worry; yet, the mayor reminds them that they are just storms they should have nothing to worry about. On this night, the storm seems to be the worst than it had been all week. The waves were higher, the winds were stronger, and the moral of the town was fraying. The dangerous waters were sloshing back and forth on the decimated beach side. As the water would back off into the sea, you would be able to see dieing fish lying on the beach. Then the waves would come back in and wash the filth away. As a vicious cycle, the water then leaves filth again. The filth would momentarily go away, but, you could count on its return. Looking up from the beach are several beach houses. Paint worn and shutters flapping, the storm was giving these houses a run for their money. As the large gusts of wind blew in, the shutters would quickly smash into the sides of the houses. The wind would stop and suck back out to sea like a warrior taking in a breath before lunging forward again. Inside the pale green beach house, two women are sitting at a table, lit by only a single candle at the center of the table. The older of the two women is Mrs. Emily Thompson, a 62 year old woman, whoâ€™s out look on any situation, is that of utter disgust. While the young woman of 22, Ms. Isabel Katertil, had a lighter look on the world, in which everything had some kind of purpose. â€œYou know Isabel, these storms have been getting worse ever since Julyâ€ said Mrs. Thompson as she sipped on the freshly brewed tea that was made to calm their nerves. â€œWe all agree that Mrs. Thompson, but, they will get better. You just watchâ€ Isabel replied. â€œThe mayor is just saying thatâ€¦ You havenâ€™t heard what is really happening around here do you?â€ said Mrs. Thompson, she leaned closer to Isabel, as if someone else was there to here her. â€œOh, come on now Emily. You have too much intelligence to be believing in those silly stories. With all those years under your belt, youâ€™d think you would have more sense then thatâ€ chuckled Isabel as she poured Mrs. Thompson more tea. 	â€œYouâ€™re too young to understand Isabel, when I was your age I didnâ€™t believe in no stories like these either. I smartened up, and so should youâ€ replied Mrs. Thompson as she leaned back in the chair and drank some of her tea. 	As she sat in the dark, the lights flickered once and then remained dead as if they used up their last breath of air. Mrs. Thompson looked up at the lights in hopes that they would suddenly jump back to life. Isabel looked out the windows and starred at the beach. 	â€œYouâ€™re right about one thing Mrs. Thompson, these storms are getting worseâ€ Isabel said those words and had a look of sorrow on her face, she then looked at Mrs. Thompson for a second then broke eye contact and glanced around the room. 	â€œI know youâ€™re worried about your boyfriend, Mark, and you probably donâ€™t want to here what I think. Everyone is always telling me I say the wrong things at the wrong timeâ€ said Mrs. Thompson as she got up and slowly walked to the door. 	At this time the two women both starred out the window to see that there was a figure on the beach. Out far enough to get hit by debris but close enough to not get dragged out. The figure was of something they had never seen before, it gleamed as a crack of lighting hit the water, and it was just standing still. 	â€œNow, what does that person think their doing?â€ Isabel said to herself as she looked out at the figure. 	â€œI never seen them around before, and why are they standing out in the water like that?â€ Mrs. Thompson said as she started to open the window. 	â€œWhy are you opening the window?â€ Isabel seemed scared but, she tried to hide it. 	â€œWell, I am going to find out what the hell they are doing out thereâ€ Mrs. Thompson said as she looked up at Isabel. 	Mrs. Thompson open the window and grabbed onto the scarf that she was wearing, the wind was so bad, that if she didnâ€™t it would have blown away. As she was about to yell something, the figure began to move. Slowly, but, just enough to notice, it was slowly walking up towards the houses. Mrs. Thompson squinted to be able to see it but with her age she couldnâ€™t see much. 	â€œIsabel, where did they go?â€ Mrs. Thompson whispered as if someone might over hear. 	â€œTheyâ€™re still there, but, they have stopped walkingâ€ Isabel said as she was holding her hand over her face so the rain would not hit it. 	â€œHey! Hey you out there! You best be getting off that beach, your going to get hurt or worse!â€ Mrs. Thompson yelled out as she cupped her hands around her mouth. 	Isabel kept her eyes on the figure, the thing gave her the feeling that if she looked away, it would disappear. At that moment the figured turned and ran down the beach, away from Isabelâ€™s beach house. Then it stopped, turned, and looked straight at Isabel. She knew it too, that chilled her to the very bone. A bolt of lighting struck somewhere off in the distances but was enough to light up the figure for someone with good eyes to see. 	â€œDid you see him Isabel?â€ asked Mrs. Thompson 	Isabel said nothing; she stood there leaning against the window sill. Looking back at the figure, as if it suddenly intrigued her so much that everything else in her mind had been shut out. 	â€œIsabelâ€ said Mrs. Thompson as she took Isabelâ€™s hand in hers and began to shake it. 	There was still no answer, and Mrs. Thompson began to get worried. She started shaking Isabel violently, as if to shake someone into consciousness. Yet nothing she did worked. Isabel just stood there immobile starring out the window. 	â€œIsabel! What is going on? All you alright? Do you need a doctor?â€ Mrs. Thompsonâ€™s questions rang out with no answer. 	Then Mrs. Thompson turned to look out the window, to see what was so interesting, only to see that her view was blocked, by the figure standing only a foot outside the window. Starring back at her, she let out a scream that was silenced by a roll of thunder. '),(14157,4042606,'12,July,2004','2004-07-12 00:00:00','   I just woke up from my nap and hopped on real quick to see how Ian was doing. I noticed he started ranting about how I am always on my laptop during my vacation. It\'s alright though I can understand why he was angry. I also wanted to thank him for posting all my messages, anyway, I think posting short stories is an awsome idea and I am going to join in when I get back on sunday. Anyway, I need to take a shower because for some reason I smell funny. It is like a mix of sweat and the smell of masturbation. I won\'t comment on either. '),(14158,4042606,'11,July,2004','2004-07-11 00:00:00','   Well as I was looking at the length of Part 1, of  A Figure In October  I noticed that it is fucking long. When I say long I mean, as long as one of my normal short stories combined with another one of my short stories. So, I have made the decision to split it into four parts. Making this run longer than I originally expected. I will be posting the first part of Part 1 before I go to work tomorrow. '),(14159,4042606,'11,July,2004','2004-07-11 00:00:00','   Get ready everyone,  A Figure In October  is being written as we speak. I have already said that. Jackie and I are beginning a project of short stories that will be posted on our website.  A Figure In October  is my first project, and is different from any other short story I have ever written, that\'s right, this is the first time I am actually focusing on the details of people rather than the enviroment. We will see how it goes, I may switch back to my regular style afterwards. I think Part 1 of  A Figure In October  might be able to be posted tonight if I can finish it before 10:00 pm, and if the website allows such a giant upload. I actually may end up spliting Part 1 in to, two segments.   We will soon see '),(14160,4042606,'11,July,2004','2004-07-11 00:00:00','   I just thought of the greatest idea on earth. You\'ll all love me for it too. I am going to start posting my short stories, and I will make Jackie post hers too. I will split the stories up so they last a little while. I got the idea when I was talking about Jackie\'s bitch of a post. Anyway, I am going to start writing the first set of stories tonight, since I have nothing better to do. '),(14161,4042606,'11,July,2004','2004-07-11 00:00:00','   Jesus Christ, I have never seen anything run so slowly. I finally got Jackie\'s novel up on the site. Just to tell you that time is wrong, It wasn\'t posted at 2:30, it was posted about five minutes ago. It took me from 2:30 to post the giant fucker. Thank you Jackie for writing such a short little story. Oh yeah, just to give you a heads up I wrote this at 6:30, that quick server I tell you. '),(14162,4042606,'11,July,2004','2004-07-11 00:00:00','   Dear Marty, I am trying to find a way to say this, without seeming like a mean person. I stay awake at night trying to think ways to bring this subject up, without it seeming scripted. While I got help from friends in writing this, (as you can see it is not really my style of writing) I just couldn\'t find a way of saying it right.  I have decided that what I am about to tell you, has been needed to be said for a while now. We have know each other for a while and you know I am really close to all my guy friends. You always seemed to show a little bit of jealousy. You also were untrusting when I would hang out with them rather then hanging out with you. I saw the look on your face when I would leave with them. It looked as if you were being shot in the heart. I am so sorry that you feel that way, but this is just a piece of my personality that will never go away. I am a flirt and I will always be that way.  I don\'t want to here you saying things behind my back also. I know you see me and my friends treating each other like that, but that is because they are just my friends. Not someone I care about dearly. It hurts me when one of your friends reminds me that you said I was a \"cock tease\". I never would say things like that about you, I don\'t know why you say them about me.  I hope when I get back from Tampa Bay, we can talk about this more. You haven\'t been answering my calls, so I thought something might have made you angry. I promise you that I will try and be with you more.   You also need to be more trusting of me, I know you think somethings are happening but they are not. I had never been untrustworthy, you should believe me when I tell you I haven\'t been sleeping around on you.  I hope Ian posts this since I am having trouble with updating from all the way down here.So I hope you get this before it is to late for us. Ian has been acting weird to me lately so I am not sure this will ever be posted. That is a problem that I will deal with later, this letter is to you. I guess I just need to get it off my chest and out into the open. '),(14163,4042606,'11,July,2004','2004-07-11 00:00:00','   There is no explaining that, she is a dick. I was waiting a hour last night to post something, because she is apparently writing a novel for her next post. It seems that she was logged in for a couple of hours typing the damn thing, locking up updates. Let me say this, It had better be one damn good update for making me wait all night. Also for being on vacation, she sure is posting quite a lot. And no Jackie, I am not going to post your letter.   But I am just ranting '),(14164,4042606,'11,July,2004','2004-07-11 00:00:00','   Tampa Bay is beautiful, and I couldn\'t be having more fun. I decided to make a post because I just got back from tanning at the beach. After I finish posting this I will probably go and take a nap. Hopefully Ian will be able to post this. For some reason I am only able to type it, then I will sit and rot. Luckly Ian has been doing the uploading for me. I heard about Michelle, I don\'t know her so I won\'t make any comments. I know though that Ian and his friends will be able to handle it, their tough.  On a much lighter note, I have heard that the south park characters are causing much more trouble then they are actually worth. So, I am not even going to deal with them anymore and just delete them. I worked hard on those too... '),(14165,4042606,'11,July,2004','2004-07-11 00:00:00','   Sometimes life doesn\'t go your way, your friends are falling around you and everyone breaks their bonds of friendship to pursue a wild fantasy that is never going to come true. Such things can drive a person mad, yet, I will put them out of my mind. As these events unfold I have a message for Michelle:       \"Don\'t be stupid, you learn from most mistakes. But, that mistake will take your life and damn you to a hell that is of unimaginable irony and torture\"   That is all I have to say on the subject and I will say no more about this unexpected turn of events. I will wait for a phone call, but, I have little hope of it.  Goodbye Michelle, we will never speak again. '),(14166,4042606,'11,July,2004','2004-07-11 00:00:00','   Want to find out if your going to hell or not? Just click below and you can take the test. I did and I guess I love Atheism, which I do. Anyway, Click below now and take the test sinner.   urlLink     urlLink Are You Damned?    You can laugh at the silly superstitions of the religious, safe in the knowledge that we are only dust and lies. All that will be left of you after you die is a slow decay and some fading memories in the minds of your friends. Hope you\'re enjoying your life at the moment- there\'s nothing better to come. '),(14167,4042606,'11,July,2004','2004-07-11 00:00:00','   I have officially decided that saturdays will be my day off from posting. So you\'ll have to just curl up in a ball, and cry. Because I know how much everyone loves the website. Let me also note that for this and the other post I have been on a mental trip of my life... Bitch. '),(14168,4042606,'11,July,2004','2004-07-11 00:00:00','   Your probably wondering \"Hey Ian, why didn\'t you update your website yesterday?\". Well, you should get out of my face before I break it. I was busy all night and did not get home until midnight last night, and I wasn\'t in the position to update the website. I hope Jackie is enjoying her time at Tampa Bay, because when she gets back, she is taking over update duty for a while.  Also, I want to know why I haven\'t gotten my fucking Mountain Dew hat yet. They shipped it on June 16th and I have yet to see it on my door step. I am getting mad, and when I get mad, I eat ham... Only replace eating ham with throwing things... and mad with pie... Because pie is fucking awsome. '),(14169,4042606,'24,July,2004','2004-07-24 00:00:00','   Sorry for the lack of posts, this entire day has been extremely busy. I barely have time to type this message. This will be the only post for today, Jackie\'s packing, I\'m going to be at the movies, and Michelle has a very long day of work at some park. So I have to go and eat before I head out again. There will be more posts tomorrow hopefully. '),(14170,4042606,'23,July,2004','2004-07-23 00:00:00','   Well,&nbsp;I made my desicion and I am heading out to New Hampshire with my dad. I have realized that there just wasn\'t enough here to make me want to stay. I am getting a fresh start in New Hampshire. Don\'t be worried though, I will still be posting on the website, and on a regular basis too. I just won\'t be able to post for the first couple of days that I am there, because of all the unpacking and such. So I will be out of Vermont by August. '),(14171,4042606,'22,July,2004','2004-07-22 00:00:00','   It\'s my day off and I am doing nothing except watch old episodes of Walker Texas Ranger while I jam my fat face with pringles. I also will actually go and buy a new video game today, something I haven\'t done in a very long time. Anyway I am off to watch more Walker Texas Ranger, so leave me alone. I also turned my cell phone off so I can deal with none stop Walker Texas Ranger action. '),(14172,4042606,'22,July,2004','2004-07-22 00:00:00','   Here I am again, posting right on time, or maybe a little late. I am sorry to hear that Jackie must make a decision like that one. Sometimes things work out for the best though, so I am sure whatever choice Jackie makes, it will be a good one. I already knew about Ian\'s hurt foot, because we talk every night for a while. It seems like it has been bothering him for a while, I suggested that he see a doctor, but I don\'t think Ian likes doctors very much; as he refused to go no matter what. Anyway I don\'t have to work tomorrow, so I will sleep in and probably call Ian early, since he doesn\'t work either. So I guess I will go to bed, and wake up early. Maybe sit on my porch and sip some coffee, and stare at the pond in our backyard. Goodnight everybody, especially you Ian! '),(14173,4042606,'22,July,2004','2004-07-22 00:00:00','   I guess I will not be skateboarding for a while. My left foot was damaged after a incident in which I will take to the grave. There is only one other person that knows of this event and I know she isn\'t talking. Anyway, because of my foot injury I will be unable to skateboard for a while, which blows ass because now I have no outdoor activity to do. '),(14174,4042606,'21,July,2004','2004-07-21 00:00:00','   I still have not come to a good conclusion; I have reasons to stay in Williston, but, there is also a couple of very good reasons that make me want to leave forever and never come back. That was something Mute and I would always talk about when I spent the night at his house. We would stay up for 1 or 2 hours just talking about what we would do if we just left and never come back. Now I have the chance, but, I am unsure I want to take it. '),(14175,4042606,'21,July,2004','2004-07-21 00:00:00','   This is a record for us, We have had the website up for more then 3 weeks. It seems like it has been longer, but, it hasn\'t. We have also switched the archiving system from daily archive to a weekly archive. The list of days would become way to long. So anyway, the three week mark has been hit, and that is all I have to say. '),(14176,4042606,'21,July,2004','2004-07-21 00:00:00','   I could have sworn that my week was getting better, then I find out that my Dad is going to be moving to New Hampshire, so I have to make a decision by next Sunday, whether I want to live with him or my mom. I am not sure who I am going to live with. I don\'t think there is much left for me here, and I don\'t really want to start over in New Hampshire, so I have some thinking to do, I will post my decision as soon as I think of it. '),(14177,4042606,'21,July,2004','2004-07-21 00:00:00','   It has seemed that I am only posting at midnight everynight. This is because I never have time during the day to get on the computer. I am either working on someone\'s yard or garden, or Adam is on the computer working on things that I can not understand.&nbsp; So the only time I can post on this site is at midnight. I also talk to Ian on the phone a lot so there isn\'t much left to post here. I love the South Park drawing Ian did of me, and I am sorry Jackie if me posting on the website is making you angry but Ian made me promise that I would as often as I can spare. '),(14178,4042606,'21,July,2004','2004-07-21 00:00:00','   I have been digging through some old boxes of mine, and I just found my original Gameboy. This thing is so big it makes it look like I am smuggling a brick in my purse. I should start using it just so I can start a new fad. Think of it, everyone throwing their SP\'s in the garbage, instead, they whip out a 500 pound gray monster.&nbsp;Just to make things clear I am talking about the original Gameboy. '),(14179,4042606,'20,July,2004','2004-07-20 00:00:00','   I don\'t think Jackie is planning on posting a South Park version of Michelle, so I will do it myself. Now introducing:  South Park Michelle , I guess I have to describe what we did so, She is just wearing her red shirt, and she also has short black hair. I am done&nbsp;talking about&nbsp;this,&nbsp; Jackie should have done this instead of me.'),(14180,4042606,'20,July,2004','2004-07-20 00:00:00','   Well, remember us talking about South Park in Vermont? I finally got the program working, and the pictures are now up and running. If you look below, you\'ll see  South Park Mute , and in a couple seconds I will have  South Park Jackie  posted too, I have to say I look wicked cute in my pigtails!   Anyway, here is  South Park Mute  (I have gotten into the habit of calling Ian that now too). Wearing his usual dress shirt and I tried giving him his wild messy hair and his cargo pants.      Now here is what everyone wants to see; It\'s  South Park Jackie ! Look at me with my gorgious black pig tails and my white tank top (which now has bird shit all over it). I am also wearing my jeans. '),(14181,4042606,'20,July,2004','2004-07-20 00:00:00','   I tried going without a hat today, attempting to kick it up a notch. I thought it would be something different, but, I just thought it was stupid so I put my hat back on later. If you were to think I look good without a hat, then I would have to punch you in the mouth. Otherwise, keep up the good work. '),(14182,4042606,'20,July,2004','2004-07-20 00:00:00','   Did you know that having big breasts can sometimes be entertaining? Let me explain, today I was down at the University Mall shopping. I was walking down the food court when I noticed a couple of younger teens starring at my breasts. All of a sudden one of them walked into one of those giant cement columns that they have throughout the mall. Of course I started laughing and kept walking, but it is those little things that make my day. '),(14183,4042606,'20,July,2004','2004-07-20 00:00:00','   I hope Ian had fun at the baseball game he went to tonight. I also have noticed that the nickname I gave him from when I first met him has been reappearing as of late. I have decided I will let everyone know why he is called \"Mute\". When I first met Ian, we were with five other people, only one, Ian knew. His name was Rich, and he introduced Ian to everybody. We went out and walked around an talked all night, yet, Ian would walk beside us and remain silent, but, It would seem as if he was in the whole conversation. My friend made a comment about something, and I looked in at Ian, and he gave a brief reply about the subject. When Ian spoke everyone remained silent to listen. Ian has always been like this, when he first meets people or is around people he is unsure of, he will remain silent or mute, and take in all the information that he can get about people. Then always at the needed time, he will speak, only for a short time. And when he does, everyone listens. That is where Ian\'s nickname \"Mute\" comes from. After that night everyone called him Mute. '),(14184,4042606,'20,July,2004','2004-07-20 00:00:00','   Whomever thought of Baseball as the American pastime should be dragged out into the street, striped of their clothes, and beaten with large pointy sticks. The game was so bad that&nbsp;I actually died a little inside as I was watching it. The only reason I was having fun was because I was hanging out with a friend, otherwise it would have been a total waste of time. Also listening to that dumbass  Tarheel  behind me spouting off about how the  Expo\'s  were going to win, which they were clearly not going to. It even got to the point where Krystal bet&nbsp;Chad a dollar a run that the Expo\'s would lose. Chad now owes Krystal a good deal of money.&nbsp;We finally decided to leave at the 9th inning because the game just got unbearably sucky. '),(14185,4042606,'19,July,2004','2004-07-19 00:00:00','   I have been sitting in my house all day doing nothing, while Ian has been hanging out with some girl at a baseball game. I don\'t know why I wasn\'t able to go, but, I guess I didn\'t need to go with him. Anyway, I just finished watching The Fellowship of the Ring, and I need something else to do. I hope Ian gets home soon and starts posting again because this day\'s posts are all from me and that looks sad. '),(14186,4042606,'19,July,2004','2004-07-19 00:00:00','   I just stopped home&nbsp;for a couple of seconds to drop off some things. I decided not to buy anything while I was downtown this morning. Like Jackie, I too landed a copy of  urlLink Picasa , I have yet to try it out yet though. '),(14187,4042606,'19,July,2004','2004-07-19 00:00:00','   I am sitting in my dad\'s living room right now trying to figure out the new program called Picasa. What it can do is organize photographs and use another program called Hello to upload photos we need to the internet. So that we can then post them here. I am trying to figure it out at the moment, so, when I do there will be a whole bunch of new pictures on the web. '),(14188,4042606,'19,July,2004','2004-07-19 00:00:00','   Well I heard that Ian\'s hand got mangled in his window this morning.&nbsp;I guess Ian was opening his window again and the top halve slid down fast and whacked into his fingers. He showed me this morning and it looks like the top of his middle finger peeled apart. Anyway, the reason Ian isn\'t posting this himself is because he was on his way to FYE to buy some music. Plus Ian has that baseball game to go to tonight, so he asked me to post about it for him. If you haven\'t noticed yet, Ian is talking to me again. '),(14189,4042606,'19,July,2004','2004-07-19 00:00:00','   I am hooked on online shopping. Everyone knows stores are not open at 2 in the morning; so I have been going crazy with my debit card and ordering everything I can find. I found this wicked awsome belly shirt on the  urlLink google store  which has the words \"Feeling Lucky\" written on the back. I am thinking about buying the beanbag chair also. '),(14190,4042606,'19,July,2004','2004-07-19 00:00:00','   I will be sleeping all day, until five comes around when I need to head down to the Vermont Expos game. I know they will lose but, it is better then getting high or drunk at a friends house... Wait a minute no it isn\'t, yet I am still going to the baseball game, because I don\'t want to ditch my friend Krystal. I hope the woman I will not mention updates the website while I am out. Or I hope Michelle will take some time and drop a note or to on the site. '),(14191,4042606,'19,July,2004','2004-07-19 00:00:00','   I want to thank Ian, for wishing me a happy birthday today. As I was talking to him on the phone, he gave me the best present ever. I hope he does not get mad at me for telling everyone what he did for me. It is a long story but, when Ian and I used to date a couple of years ago. The first song that we kissed to was an odd one; the song was called  Push Me, Pull Me  by  Pearl Jam.  As we were talking on the phone I started to hear a song in the background on Ian\'s end of the line. I asked him what was happening over there and the song got louder and he wished me a happy birthday. I never thought you would remember that, thank you Ian.  &nbsp;  Ian thought I was getting sad again, because I was crying a lot. But, it was because was very romantic and&nbsp;I was just so&nbsp;happy. So Ian doesn\'t have to worry. After I finished speaking with Ian, Adam also gave me a nice birthday cake and gave me a present. When&nbsp;I opened it up I found a copy of Harvest Moon, for Super Nintendo. I have been looking for a copy of this game forever and now I have it. I think this is one of my best birthdays I have ever had.  &nbsp;  Thank you Ian and Adam. '),(14192,4042606,'18,July,2004','2004-07-18 00:00:00','   The Butterfly Effect, is probably one of the best movies I have ever seen. It has been a while since a movie has made me think as much as this one does. I won\'t go into extreme detail, as I am really busy at the moment but here is a website where you can  urlLink read all about it . '),(14193,4042606,'18,July,2004','2004-07-18 00:00:00','   After speaking with Ian on the phone, I have realized that I shouldn\'t of posted about how Ian\'s mind works. That has always been a nerve for him, and I was wrong for even posting anything about the subject on this website. If only Ian will forgive if I leave him alone for a while. '),(14194,4042606,'18,July,2004','2004-07-18 00:00:00','   Don\'t think I forgot Michelle! &nbsp; Today is Michelle\'s 22nd birthday, so happy birthday Michelle. '),(14195,4042606,'18,July,2004','2004-07-18 00:00:00','   I think I have Ian worried, because I have not posted in a little while. I just came home to change my clothes. I had dirt all over them from working hard today. I went out and did some tree planting at this old womans house. She was very nice to me and the other workers. I am now going to go and call&nbsp;Ian and check up with him. I would think Ian would get sick of me by now but, I think he is starting to forgive me. '),(14196,4042606,'18,July,2004','2004-07-18 00:00:00','   I have just heard from my friend Vince has seen Fahrenheit 9/11 and he said that it was a movie that would make me hate everybody. Well, since I already hate everyone I don\'t think I need to see it. Then he claimed that I would hate everyone more, I was intrigued. So I will probably go and see this movie of hatred. Not anytime this week, because everyone I know is busy and people know how much I hate seeing movies alone. '),(14197,4042606,'18,July,2004','2004-07-18 00:00:00','   I hate my life, I come home only to be dumped by my boyfriend Martin. I thought we would be able to work things out but, he had his mind set. I also want to know why Ian still isn\'t talking to me. I have tried calling him like four times today, and everytime I get his answering machine. I know he is working right now but he should still have his cell phone on. I am sick of this shit, I want to know what I did to everyone else that made them hate me. '),(14198,4042606,'18,July,2004','2004-07-18 00:00:00','   Guess what? I finally had an encounter with a Black Widow Spider. This morning as I was getting the black grapes out onto the display, this Black Widow Spider just fell out of the grapes. It freaked me out but I was able to deal with it. If you haven\'t seen a Black Widow Spider before, then your completely retarded but here is a picture anyway.  &nbsp; '),(14199,4042606,'18,July,2004','2004-07-18 00:00:00','   Well, as soon as I finish posting this, I will be heading home. I will have to deal with all the problems that will be waiting for me. I am not sure when my plane will land but I will post again as soon as I finish unpacking and shit. '),(14200,4042606,'31,July,2004','2004-07-31 00:00:00','   As I am still a member of the website, I see what is going to be posted before it actually enters the web. Luckly I was able to catch a certain post before it was going to be loaded onto the site. I knew I should have kept those two apart. I knew exactally what was going to happen, they just thought that I was freaking out like usual.  I have a direct message for Abby,  \"Don\'t you dare think about posting what you just wrote. I know what you are thinking about doing, and your just setting yourself up for falure. I know what is going to happen, your going to scare him, and he will run away from you if you do it. So I suggest you really think about what your going to do, before you go through with it.\" That is the only thing I will say and I will not go any further into detail, because Abby knows what I am speaking of. '),(14201,4042606,'30,July,2004','2004-07-30 00:00:00','   There will be a few minor delays in updating the website this week. Let me explain the situation for each contributor of this website:  Michelle will be unable to post for quite a while. The computer virus that attacked her house-mate\'s computer days ago has seemed to do more damage than the originally thought. The entire harddrive is destroyed and it will take time and money to fix the problem. That means, no Michelle for about another week.  Jackie is still not communicating with anyone that is a contributor to the website. The only person that she has talked to since she has left was her mother. We all know, I am not on speaking terms with that woman either.  Spinner (Abby) is on her way back to Pennsylvania tomorrow. That gives her two more days of catching up with old friends. So you can bet she will not be making many (if any) posts tomorrow. I am unaware of how long the car ride back to Pennsylvania is so she may not be posting for a period of time.  Myself, will be leaving at noon tomorrow. I will be heading out to Williston until Wednesday, and I am not going to be wanting to post at all. I need a vacation from this place so I really don\'t want to be responcible for anything, including the website.  On another note, our other website went under today. I really didn\'t have the time to take care of two different websites. I am a busy person you know.  Other than the problems above, the website will be going good. Just think of this week as a vacation for all of the contributors on it. I will probably be making a couple more posts tonight, as well as making some more tomorrow before noon, if I have time. '),(14202,4042606,'30,July,2004','2004-07-30 00:00:00','   Guess what, everything went fine. Spinner got out of my house without my mother finding out so I am safe for the time being. It was at around noon before Spinner left, we woke up at around 9 but we sat there talking and waiting for are moment to get out. My mother had to take a nap, so that is when I got Abby out of there. '),(14203,4042606,'29,July,2004','2004-07-29 00:00:00','   Alright, we are about to head off to bed again. I just decided to make a quick post while Mute was in the shower. I know many people sleep in the same bed together. Yet, people find it odd that two good friends sleep in the same bed while nothing happens. We are just friends and Mute and I are just sleeping nothing else. I am saying this to make Jackie stop screaming at me and calling me a whore. She is one to talk in the first place. So we will probably get caught when we wake up in the morning, but, we can always hope that Mute\'s mom has to go out somewhere so I can get out without anyone seeing. '),(14204,4042606,'29,July,2004','2004-07-29 00:00:00','   Mute and I are in trouble at the moment. We want to spend the night together again, remaining in secrecy from certain people. When it is brought to our attention that Mute\'s mother is going to have the day off tomorrow, making it almost impossible for us to get up in the morning without bringing unwanted eyes. If there is a way to pull it off I would do it in a heart beat, I am actually screwed at the moment because I am at his house right now. My ride won\'t be here until around 10 tomorrow. So we are just going to enjoy things while they are good, and worry about that little problem that will occur in the morning. '),(14205,4042606,'28,July,2004','2004-07-28 00:00:00','   Here\'s the story, I slept over at Mute\'s last night. I didn\'t get home until 1 in the afternoon, and now I am at Mute\'s again. Jackie, apparently called home at like noon. My mother picked up, and I specifically told her if Jackie called, not to tell her that I was with Ian. Guess what? My mom said this \"Oh Jackie honey! I\'m sorry Abby is with her friend Ian\". Thank you mom for fucking up an already fucked up situation. I called her back when I got home and big suprise Jackie is flaming. I got a long lecture on how Ian is a jerk and is just hanging out with me so he can get into my \"pants\". Let me say Mute is nothing like how Jackie is describing him. Mute, in his entire life has nevered asked for sex, or anything assosiated with it. I know him much better than Jackie ever did. She needs to shut the fuck off. I am nothing like her either, I will not sleep around like Jackie is known for doing. Mute and I are just really good friends. We have been so since we were around 9. So I think I have a better insight on Mute, then my dumbass sister. '),(14206,4042606,'28,July,2004','2004-07-28 00:00:00','   I am not sure if anyone has noticed, but, Michelle hasn\'t posted in a litte while. That is probably going to stay that way for some time. As I was talking to her on the phone today, She explained to me that Adam\'s computer got hit with a virus and is going to be down until they can get it to a computer repair store. I hope it get\'s fixed soon so she can add to the site. So for now it is just going to be Spinner and I. The only reason Spinner has been able to post is because she has hung out at my house for the past day or so. Anyway, I hope Michelle is able to get her computer fixed. '),(14207,4042606,'28,July,2004','2004-07-28 00:00:00','   The goal that day was to get a hold of Spinner without Jackie knowing. Well it worked, very well. Abby and I have hung out for like 6 hours everyday since then. Last night she sleep over (without my mother knowing ofcourse). It seems that we have picked up exactally where we left off when we parted ways about four years ago. Tonight we will probably get together again and possibly sleep over again. To bad she has to go home on Sunday, but, she is now a member of two websites with me so we\'ll keep in touch.&nbsp;She actually just left my house at about noon, before my mom got home. I don\'t think it would have looked very good if I were caught sneeking Abby in and out of my house at midnight. '),(14208,4042606,'28,July,2004','2004-07-28 00:00:00','   It\'s a long story, and I don\'t really know how to explain myself. I never really got close to anyone since I moved to Pennsylvania. Mute was the last person I was really close and open to. When I woke up this morning I looked beside me in the bed to see Mute sleeping peacefully. I laid there for like an hour just starring at him sleep. I don\'t know what the feeling was, but, it was a good one. It was like all my problems just sort of drifted off into the night. I know we are just really good friends, but, it is nice to know that there is someone out there in the world that cares about me. This is the first morning in a long time I haven\'t woke up either pissed off or depressed. It is something I could get used to. '),(14209,4042606,'28,July,2004','2004-07-28 00:00:00','   Mute just made a comment to me as we sat here that pissed me off. Let me clarify on the subject. When any of us mentions the \"Emmon\'s Curse\" we are talking about large breasts. Jackie has&nbsp;them (everyone already knows that). I wear baggy clothing so that I don\'t draw attention to myself, unlike my sister whom barely wears anything. Now I usually will always wear cargo pants and a hoodie. When I am with close friends I\'ll just wear the t-shirt under my hoodie. So they know my \"endowments\". If I ever hear Mute say my breasts look a lot like my sisters again, I will kill him. Plain and Simple, they are no where near as big as Jackie\'s so shut up. '),(14210,4042606,'28,July,2004','2004-07-28 00:00:00','   I would like to make a comment on how lady like Spinner is being at the moment. She is really classing things up by shoving as many Oreo\'s as humanly possible in her mouth. Trust me they didn\'t stay in there for long. She was in the bathroom 30 seconds later throwing them all up in the toilet. Just wanted to let everyone know that the serious Republican just made a complete ass out of herself, leaving a stain of spit up oreos on my floor. '),(14211,4042606,'27,July,2004','2004-07-27 00:00:00','   Yo everyone, this is my first post and I\'m here to say that green is the new color of the website. Jackie picked that lame blue template, but, she isn\'t here to force it on anymore so I am making it green. If you ever would of thought that the little teen princess would throw the biggest fit on earth, well you&nbsp;missed it. I was there when she flipped out, it was totally awsome. She finally realized that I wasn\'t some little girl and I could take care of myself. She was just pissed that she could never have Mute. She thought I was trying to steal him away from her or something, even though we are just really good friends. That\'s why she got all pissed and left early, because she couldn\'t pressure Mute into things like she can with most people. And another thing, Mute was wondering if I still went by the nickname \"Spinner\". I hadn\'t been called that since I left for Pennsylvania. I am getting used to it again, as Ian is probably to me calling him \"Mute\" all the time. '),(14212,4042606,'27,July,2004','2004-07-27 00:00:00','   I find it sad that Jackie left on such a bad note. I hope she did not do anything she would regret, I also wish her luck in New Hampshire. I know I also left without really letting anyone know. I have realized though that you must make up with the people that you hurt. So I hope Jackie stays in contact with Ian and her sister before it is to late.&nbsp;I guess I will have to start posting more and more on this website now. It is just going to be Ian and I for a while until Abby gets the chance to start posting. '),(14213,4042606,'27,July,2004','2004-07-27 00:00:00','   I was seeing how popular each person was on Kakutama and the results are not very suprising:    Jackie Sign  Mute Kakutama  Michelle  Spinner Abby   As usual Jackie Sign is the most popular person on the website, probably because she tries and sucks all attention from everyone else on earth and pins it to herself. I was astounded that I actually beat Michelle for second place. I usually don\'t do well in popularity contests. I guess it has to do with Michelle never posting on the sight anyway. The only reason Spinner is in last place is that she hasn\'t made one post yet so at least I have bragging rights. '),(14214,4042606,'27,July,2004','2004-07-27 00:00:00','   Well, as Abby and I were talking more and more today; we decided we should have our own section dedicated to us. So Abby and I introduce  urlLink A Butterfly Effect , a website that is for our combined thoughts of the world. It is more seriously oriented than Kakutama, yet humorous in it\'s own way. '),(14215,4042606,'27,July,2004','2004-07-27 00:00:00','   Well when I got back from hanging out with Abby, I hopped on my laptop to see if Jackie posted anything new. I found that Jackie actually left, so I am not sure if she is ever going to post on the website again or what is going on. Abby will start posting as soon as she get\'s back to Pennsylvania, which will actually be next Wednesday. She didn\'t want to leave right when Jackie left, because she still had things to talk to me about. So I will say to Jackie, that she completely over reacted, and we could of hung out one last time before she left. But, being the little drama queen that she is, she left with a big fight. Abby says she has been that way her whole life, an I didn\'t see it until now. Anyway, with the loss of one member, we gain a new one. And another new one might be coming around, Michelle dropped a little hint that a friend of mine, Sarah, was asking to join the website. She is heading off to college soon and she doesn\'t want to be left in the dark. I am thinking about it, but, there seems to be a lot of women on this website these days. '),(14216,4042606,'27,July,2004','2004-07-27 00:00:00','   Well, looks like I\'m leaving in 10 minutes. Ian and Abby have decided to hang out with each other instead of seeing me before I leave. I don\'t know why they feel that getting pretzels at the U-mall is better than seeing me for one last time, but, I guess their priorities are all fucked up, so see everyone later. I might post once I get settled in New Hampshire, but if Ian keeps being an ass, then this will probably be my last post. '),(14217,4042606,'27,July,2004','2004-07-27 00:00:00','   I am in love with those cute little Pikmin so much, that I got hold of a beutiful wallpaper that you can download.    urlLink '),(14218,4042606,'26,July,2004','2004-07-26 00:00:00','   I just got a copy of the new Nintendo Power, and I read the article on Pikmin 2. I can not wait until the game is released. The graphics look absolutely beautiful, and they are the best water effects that I have ever seen. The Pikimin also have their same cuteness, adorable looks as they did from the first game. Now there are two new species of Pikimin joining the crowd, a white and purple breed. They are always so happy go lucky and skipping behind Olimar, then when they face a foe, they turn into these dangerous monsters that behead the enemy. It is very disturbing, but that is what gives the game it\'s unique charm. '),(14219,4042606,'26,July,2004','2004-07-26 00:00:00','   Well, I didn\'t get home until 4:00 in the morning. Abby and I were out walking around Williston. You may be wondering how the hell I got down to Williston without a car. Well, I would just like to thank a friend of mine for giving me a lift down there when I really needed it. I met Abby near Best Buy at around Midnight, and we walked around down there talking until I had to leave at around 3:30, It was good to see her again, she has changed a lot. Not only physically but also mentally. She said I have changed a lot too, but I don\'t really see it. Anyway, I am finally taller than her. When we were friends a while ago she had about an inch on me and she would always bring it up. Now I am much taller so, I guess I win in that aspect. Luckily Abby did not inherit any of Jackie\'s bad personality traits, instead she has whole other set of them. One thing I did notice though is she dresses a little more modest then her big sister. Last night she wore and Independent sweatshirt and some baggy cargo pants. Something Jackie wouldn\'t be caught dead wearing. The reason she says is she doesn\'t need to flaunt the Emmon\'s family curse to get what she wants. She is right too, Abby is a very opinionated person, who sounds sort of like a feminist.  On another note, Jackie says she is leaving early so she will never see me again. This in turn would force Abby back to Pennsylvania. Good thing that is not the case, because I am still going to spend time with Abby next week, because she is going to stay for another week. She will end up getting a ride back to Pennsylvania from her cousin who just got up here. So I guess Jackie\'s little plan isn\'t going to work and she is just going to deal with the fact that Abby and I are good friends again. '),(14220,4042606,'26,July,2004','2004-07-26 00:00:00','   I was going to stay until Monday, because Ian was going to come down and wish me goodbye. Well, I really don\'t ever want to see Ian again. So by Wednesday I will be in New Hampshire, also because I am leaving early; Abby will no longer have a reason to be here and she will go back to Pennsylvania. So it\'s a lose-lose situation for Ian. All because he kept ditching me for a stupid rich college bitch and a retarded red headed idiot. If he ever wants to apologize to me, he knows to leave my sister alone, an he shouldn\'t hang out with those two Burlington sluts anymore. Then maybe I will talk to him again. What pisses me off even more is that Michelle girl is actually encouraging Abby to hang out with Ian. I never liked Michelle and now she just game me more of a reason to hate her guts. If she thinks she still has a say in what happens in Vermont, while she is sitting comfortably in Ohio, well, she has another thing coming. So anyway, I probably will never be posting on this crappy website again, unless I want to check up on what that idiot Ian is doing. '),(14221,4042606,'26,July,2004','2004-07-26 00:00:00','   I feel really bad for Ian, he is the only male on the website. I also just found out that Abby is joining the website, and that Sarah girl also wants to join. So that means there will be 4 women on the website and only 1 man. I really do feel bad about this, because, it was Ian\'s idea to make this thing, and now he is being pushed down by all the women. I really like Abby though, so I do think she really should join. '),(14222,4042606,'25,July,2004','2004-07-25 00:00:00','   Well, Looks like I am sneaking out of my house tonight. I have to see Abby again, it is driving me crazy. So I am leaving now and meeting her at a certain location so we can chill. I am not saying where I will be because we don\'t need Jackie getting on our asses and making us split up. I keep reminding Jackie that Spinner and I were just friends and always will be friends no matter how much meddling Jackie tries. '),(14223,4042606,'25,July,2004','2004-07-25 00:00:00','   As Jackie is still pissed, as if you couldn\'t tell. Abby though called me when I got home from work. The plan worked and she got my cell phone number. So, we have been talking about the good old days and talking about what has happen since we all split. Let me tell you, she has defiantly grown up. Jackie doesn\'t know this, but, Abby and I are meeting at the mall tomorrow before I have to head down to work. Jackie has no chance in stopping us from getting together now, so she is just going to have to live with it. I also am getting pissed at her, if she calls any of my Burlington friends sluts one more time, I am going to lose it. She also has no right to keep Spinner and I apart since Spinner is the one who introduced Jackie and I. Spinner was one of my closest friends since I moved from Jonesville to Williston. So Jackie can just shut it. '),(14224,4042606,'25,July,2004','2004-07-25 00:00:00','   Abby is not going to join the website, I can see personally to that. I am sick of Ian thinking he is completely running the show. Why doesn\'t he just get one of his whores to join the website, and not my sister. I know exactally what is going to happen and he is just going to end up hurting her. Abby may have known Ian longer than I did, but, I still know more about him. Abby is just being stupid when she says she wants to hang out with Ian again. I would have posted this earlier, but, somehow Abby got Ian\'s phone number and hasn\'t gotten off the phone since 4 in the afternoon. I want Abby to stop calling Ian, and I want Michelle to butt out of everything, she has no right talking about us when she is way out in Ohio, and has no idea about what is really going on. '),(14225,4042606,'25,July,2004','2004-07-25 00:00:00','   Well, I am still talking to her, I just need to explain what is going on real quick. Abby is going to join in on the website. She just won\'t be able to post until next tuesday, when she goes back to Pennsylvania. Her profile is already up, and you can finally see a picture of her. She sure has changed after four years. '),(14226,4042606,'25,July,2004','2004-07-25 00:00:00','   Jackie thought she could out smart me, by staying on the internet all the time and not answering her cell phone. But, Mute & Spinner have once again joined forces. That\'s right, I hate to break Jackie\'s bubble; I have gotten in contact with Spinner thanks to a good friend which will remain secret. I also think Jackie needs to really think, Spinner and I have been friends much longer than Jackie and I. Yet, Jackie is actually trying to tell Spinner and I that we are not allowed to talk to each other. No one takes on us and wins. We just proved that, and it wasn\'t even that hard to do. So if Jackie wants to apologize then I am waiting, otherwise I don\'t want to talk to her. Anyway, I have to go and talk to Spinner about things, and hopefully she will join us with this website. '),(14227,4042606,'25,July,2004','2004-07-25 00:00:00','   I just can\'t stay away from posting today. Well, we have finally reached the one hundred post mark on the website. I think that is great that we were able to keep the website going for a while. I think we will continue to post up until 200, I believe it is a special occasion. '),(14228,4042606,'25,July,2004','2004-07-25 00:00:00','   I have just gotten back from doing some things, I will probably start cleaning soon. I haven\'t done that in a while. I heard that Abby was asking about me also, along with Ian. I don\'t think she knows I did a little soul searching for a while, and ended up living in Ohio. I think she knows now though, I can not stand that I won\'t be able to see her again. The last time I saw her, she was so small and adorable. I hope Ian gets to see her! '),(14229,4042606,'25,July,2004','2004-07-25 00:00:00','   Isn\'t it weird that I am posting at this time? I think so, If you have not heard, I have the day off today. I am sitting here in my pajama\'s sipping my coffee, and looking out at the front yard. It is a nice day, I hope Ian got in contact with Abby last night. I would really like to see those to reunited. I now need to go take a shower and I will probably go to the store to pick up a couple things, so I will post when I get back. '),(14230,4042606,'25,July,2004','2004-07-25 00:00:00','   Well everybody, I am getting back to posting only at midnight. I have heard some people call me weird because of my late night posting. I am sorry, if I could I would post during the day, but I am just not home enough to do so. I was luckly home early today, so I was able to post more. Today (July 26th, 2004) is my day of relaxation. As a birthday present, Adam said he would work in my place, any day of my choosing. So I chose today, when I wake up later I will post on this website all day. I will also try and figure out what is going on with Abby, Jackie, and Ian. Ian also said to be prepared to have another member to the website. I think he is going to try and get Abby to join in on the fun! I can not wait to see a picture of how much Abby has grown. She is the same age as Ian so I want to see if she is still a perky happy girl like she used to be when I was around. She was so cute when I first met her, so let\'s see how much she has changed. '),(14231,4042606,'25,July,2004','2004-07-25 00:00:00','   I don\'t know why it is such a big deal that Abby is home for the week. She is only my half sister so it isn\'t like we are extremely related.&nbsp; And I am not sure I want&nbsp;Mute talking to my half sister anyway, I can\'t really trust him anymore with other girls because he is always sleeping around with random girls in Burlington. So if he really wants to get in touch with an old friend, he can just visit \"Sarah\" and have sex all night with her. It doesn\'t matter if Abby wants to see Mute, I say he is an asshole and he\'ll only hurt her if she gets near him. '),(14232,4042606,'25,July,2004','2004-07-25 00:00:00','   Alright, it has now hit the one hour mark of me trying to get through to Jackie\'s mom\'s house. Apparently they have taken a brake from packing and gotten on the internet. Thank you Dial up for that, and Jackie is mad at me for some reason and is not answering her cell phone. I would like to say that I would really like to talk to Spinner, so I know Jackie may be mad, but that is no reason to keep old friends away from each other. Once again, narrowband has caused anguish in the world. '),(14233,4042606,'25,July,2004','2004-07-25 00:00:00','   I just got off the phone with Ian, and he told me all about how Abby is going back to Vermont for a week. You wouldn\'t believe how excited Ian is, it\'s to bad that I won\'t be able to see her though, maybe they will take pictures to send me. I really would like to see her again. She was actually one of the people whom introduced Ian and I. After we were done hanging out with Ian and Rich that night we all went back to Rich\'s house. Abby was there and that is where we saw Ian really talk. That is how I kind of got a little crush on Ian. So in a way Abby caused Ian and I to start dating. She was happy about it too, when she found out we started dating. She was also really upset when she found out we were breaking up before we all went our separate ways. Ian asked me to explain why Abby is called \"Spinner\", I usually don\'t call anyone by their nickname when I write, because it is confusing for people sometimes. I do though call people their nicknames when I am talking to them though. I just wanted to clear that up, so people didn\'t think I was weird or something. The story of \"Spinner\" is very cute, and I think everyone will like it. I was not there when she was first being called \"Spinner\". I guess there was just one event that gave her the nickname \"Spinner\". The reason though that she has that nickname is because her mood would reverse really fast without anyone knowing. It is hard to explain, like we could be having the worst day on earth and all of us in a really sad mood, including Abby. Abby, would all of a sudden become extremely happy, like her mood would just spin around. She would then cheer everyone up. That is the easiest way of explaining why Abby was called \"Spinner\". '),(14234,4042606,'25,July,2004','2004-07-25 00:00:00','   Well, I guess Jackie wasn\'t going to tell me, but, Spinner is in town. Michelle knows whom I am talking about. My old friend Spinner, her real name is Abby Emmons, and she is Jackie\'s half sister. Spinner as we used to call her used to be extremely good friends when we were younger. Until I had to move to Burlington, that is when Michelle and I broke up. Abby and I were both pretty down, so instead of sitting in Williston without me, she told her mom she wanted to live with her dad. Did I mention he lives in Pennsylvannia? Well he does, so I went to Burlington and Abby went to Pennsylvannia, all in the same month. Before this time I had started to become friends with Abby\'s older sister Jackie, who\'s dad lived in Burlington. Jackie would switch every week, from her father\'s to her mother\'s house. So when Spinner went to Pennsylvannia, Jackie became my new Spinner. Anyway, I was forced to Burlington, and I lost contact with Michelle and Spinner. But, Michelle and I started talking again, after certain events occured. Spinner on the other hand seemed as to disappear completely.  Now that you know the background story, here is the deal now. Spinner, is coming up to visit her mom, and to help her sister, Jackie pack. I got word that Spinner was asking about me and Michelle. This is where we are now, so when I go down there to visit Jackie one last time (if she lets me). Spinner and I will be united once again, and we will be back to our old games in no time. That\'s my plan anyway. I hope Spinner\'s has the same idea.  Anyway, I need to let Michelle know about this little blast from our past. I am not sure if Spinner is the same as she used to be since we haven\'t talked for about 4 years. Also if you want to know how Abby, came to be called \"Spinner\" you should ask Michelle, because, I don\'t feel like typing anymore.  Just another thing on my mind... '),(14235,4042606,'25,July,2004','2004-07-25 00:00:00','   Hello everyone, I hope I did not worry anyone by not writing anything for a couple of days. I had a big project at work that had taken up all my time. We were working at this park, and we were planting all sorts of trees. It was a lot of fun, I can not believe I get paid to do something I love so much. I also started a garden at home, I know it is late in the year, but, I think of it as starting early for next year. I also heard about the situation with Ian and Sarah. Ian told me all about it today, since we did not talk last night. We usually will talk on the phone every night for an hour, Ian was at the movies with his friend Georgia, while I was at a plant warehouse until midnight trying to get the right kind of tree for the park. So we didn\'t have a chance to talk at all. I think it\'s cute that Ian talks to me every day, even though I am way out in Ohio. I really appreciate it, and I love every minute he talks to me. '),(14236,4042606,'25,July,2004','2004-07-25 00:00:00','   You know what I don\'t understand about Mute, how he always hangs out with other girls, yet he can\'t make the time to hang out with me. I know he is coming to spend a weekend with me but that isn\'t enough. This is one of the reasons that I am not staying in Vermont, none of the people I care about spend any time with me at all. First he ditches me to hang out with that red head, now he is hanging out with that rich college girl, and he might be dating her? I don\'t understand, I always thought old friends were more important then some Burlington sluts. He never would go see movies 2 days a week with me, or spend an entire day at my house watching a movie I wanted to watch, like when I invited Mute to come and watch 28 Days Later over at my dads house, Mute just says he had already made plans to hang out with the red head. Well, don\'t even bother coming down to say goodbye Mute, just stay up in Burlington with all your slutty bitches. '),(14237,4042606,'25,July,2004','2004-07-25 00:00:00','   I don\'t know why, but, all I have been smelling today is Georgia\'s hair. I am unsure of what shampoo she uses, the smell is driving me insane, I think I am going to go and try and cut off my nose if it doesn\'t go away soon, it probably doesn\'t help that I am still wearing the same clothes from yesterday either.That&nbsp;still doesn\'t say much for Georgia though, if the smell is that&nbsp;strong&nbsp;that it can become intertwined with the fibers of my shirt, says that you could definitely smell her from a mile away. I can only say this: Georgia stop smelling so much, it\'s not that it is a bad smell, it\'s just that after smelling it for 12 hours straight you start to go a little crazy. '),(14238,4042606,'25,July,2004','2004-07-25 00:00:00','   I am at a emotional crossroad at the moment. A friend of mine, that I only thought of as so; asked me if it was possible for us to start dating. Everyone may think \"Go for it\" but there are unseen conditions that could make things harder. I have been thinking about this for about two days now, and I still have not come up with a conclusion. Everytime I run the&nbsp;situation through my head, it would always lead be back to where I started. So, I am still unsure if I want to date this person or just stay friends. I think I need a lot more time to run this through my head. I hope she is not expecting an answer anytime soon. '),(14239,4042606,'25,July,2004','2004-07-25 00:00:00','   I have been sleeping since I got home at 10:30 AM. The night before I was out with my friend Georgia watching  The Bourne Supremacy . Which happens to be a very good movie, even if neither of us has seen  The Bourne Identity . Anyway, I didn\'t sleep at all because I got home at midnight, and I had things on my mind. I went over to Sarah\'s at about 12:40 and stayed there until 2, then snuck back into my house at around 2:30, I then proceeded to go to work at that exact time. So that didn\'t leave me much time to sleep, as you can clearly tell. I was about to pass out at work, so I decided to leave&nbsp;a half hour early and get some sleep. I am glad I did, no one realizes how no sleep can drive someone crazy. '),(14240,4042606,'25,July,2004','2004-07-25 00:00:00','   I have realized that packing is not as fun as everybody makes it out to be. Because I am moving, I also have to get everything I have at my moms house also, which is quite a lot since I spent more time there. Anyway, I have to get all of my stuff on to a sort of cargo truck type car. This is taking me forever and I am going to still be packing well through the week. I leave next Monday, Ian said he was going to come down on Sunday and see me so I hope he does. Now I have to continue packing, away I go. '),(14241,4042606,'07,August,2004','2004-08-07 00:00:00','   There is going to be no more posts for a while. Some issues between all of us need to be resolved. As they are not going to be concluded anytime soon, we have decided it would be best if we didn\'t post about anything until we stop arguing. So for a while, there will be like a forced vacation for all of us. '),(14242,4042606,'06,August,2004','2004-08-06 00:00:00','   What I known for, and as my nickname states: I am a big believer in Astrology. It is always true, I even got Ian in the habit of reading his horoscope all day. And I was interested in what the stars say about Ian\'s little problem. I discovered that I now (in very rare events) agree with my sister. Here is what the stars say:  \"The merging of Libra and Sagittarius can be paradise found for both Signs; this combination is a harmonious one, to say the least. Signs that are two positions apart in the Zodiac tend to have a very deep, special connection and understanding of one another. Sagittarius, the journeyer, is constantly in search of wisdom. Libra has a mind keen for art, beauty and aesthetics; Libra makes a stimulating companion for Sagittarius on their travels and adds much to Sagittarius\'s experience. The relationship always feels new because these two keep it exciting--and yet their natural understanding of one another makes them feel like intimates from the beginning. Together they will reach new horizons in love and in life. Libra and Sagittarius make great friends as well as lovers. They share a certain fresh-faced optimism in the world; they\'re both enthusiastic about finding truth and beauty in their experiences. Problems rarely occur. Occasionally Sagittarius may hurt sensitive Libra\'s feelings by speaking without thinking; conversely, Libra can sometimes be a bit too emotionally controlling for Sagittarius. Since Libra is a born diplomat who abhors conflict, they\'ll be able to smooth over any ruffled feathers. These two have the gift of forgiving and forgetting quickly. Libra is ruled by Venus (Love) and Sagittarius is ruled by Jupiter (Luck and Philosophy). Venus\'s femininity and Jupiter\'s masculinity make these Signs compatible and balanced. Sagittarius is constantly on a quest for knowledge of the big things--the truth, the meaning of life and death--and whenever Libra comes up with a new idea (as they are wont to do) Sagittarius is eager to go along with their partner and become familiar with this new concept. Libra is an Air Sign and Sagittarius is a Fire Sign. This partnership is full of kinetic energy. As air spreads fire and makes it grow exponentially, this relationship is heated and arduous. Sagittarius thinks they\'re steering the rudder, but Libra controls the flow of events with a quieter hand. These two are energetic in the extreme and can run together far and wide. The relationship runs most smoothly when Sagittarius can feel they have plenty of freedom and independence within the relationship.Libra is a Cardinal Sign and Sagittarius is a Mutable Sign. Libra is an initiator; they think up new ideas and new places to go--but can change their minds at a moment\'s notice. Sagittarius is flexible and adaptable, amenable to any changes Libra wants to make in plans.What\'s the best aspect of the Libra-Sagittarius relationship? Their mutual interest in cultivating knowledge and utilizing intellect. They are well-matched and will go far together, both emotionally and geographically!\" So as you can see, they are definately compatable, and I think Ian should go for it now. '),(14243,4042606,'05,August,2004','2004-08-05 00:00:00','   It looks like Ian was even more upset than I originally thought. After talking for a while with him, he decided he needed a little break from the website. So he can clear his thoughts. So he asked me to take over, and handle everything on the website. I will just say this:  \"There is no room for two Emmons by Ian\'s side, and it looks like once again, I win Ian\'s trust and not anybody else\". I also can\'t thank Ian enough for forgiving me for what I did. I think he has realized that I am right in a couple of ways. It may have been better if Ian and Abby never did get back together. '),(14244,4042606,'04,August,2004','2004-08-04 00:00:00','   Where the hell is my cell phone? I just noticed it was not in any of my bags, or any other box that I had packed. Usually I make sure I have it before I go anywhere but I thought I just tossed it into a box. Now I found out that my sister has it in Pennsylvania, and if she does not send it to me I will have her beheaded. '),(14245,4042606,'04,August,2004','2004-08-04 00:00:00','   I would have thought that I would be pissed at Ian even more. But, I can\'t help but agree with Ian. Nobody, not even me, has the right to force him to date anybody. I have realized that I was wrong and now I see that Abby and Michelle are also wrong. That is why he isn\'t talking to those two. Because they are trying to make decisions for him. That would piss me off too. So I am going to call him right now and see what the hell is going on. '),(14246,4042606,'04,August,2004','2004-08-04 00:00:00','   Mute is getting very defensive on the subject of not dating the \"red head\". I guess Abby really made him mad when she kept telling him to ask her out on a date. Then when Mute sent me the copy of the conversation, I agreed with Abby. That made Mute even more angry. I think he is just acting like he doesn\'t want her, because, he is afraid if he asks that she will turn him down. That is just my opinion, I know Mute doesn\'t want to hear it either. I think he has asked four other people about it, and they all agree with Abby and I. I would like to ask Ian if we could post Abby\'s conversation on the website, just to see what everyone else thinks. Unfortunately I know the answer, he is going to say no. He isn\'t talking to either of use since the whole conversation has happen. But, I just think he is embarrassed. '),(14247,4042606,'04,August,2004','2004-08-04 00:00:00','   Well, Abby has gone back to Pennyslvania. I am not sure what is to come of our relationship. I think we will continue to be friends, maybe trying something else. Abby seems to have gotten the idea that Georgia and I are dating which we are NOT. So if anyone else there is thinking the same thing, well there is your answer. '),(14248,4042606,'03,August,2004','2004-08-03 00:00:00','   Looks like I am about to spend my last night in Vermont. I am about to head home from Mute\'s house at around 3:00. Luckly my friend, Tom has been my ride around town all week. My last night here will be with Mute, and this is also the last time I will see him for a long time, so I am going to make it the best night of both our lives. '),(14249,4042606,'02,August,2004','2004-08-02 00:00:00','   That last post was the only time this week I have heard from Jackie. I am unsure if she is going to start posting, or she is still pissed that she can\'t have Mute. She knows what is now going on between Mute and I, and I am sure that has made her even more angry. Not just that I have made the offer, but, that Mute is actually considering it. I guess she is just mad that Mute and I are connecting better than she ever did. '),(14250,4042606,'02,August,2004','2004-08-02 00:00:00','   You would think being caught walking around on the waterfront by a down pour would piss someone off, yet, it didn\'t. Abby and I were hanging out downtown when it started down pouring. Then it started thundering and lightning\'s. We just didn\'t care, it is our last day together for a while, and no rain was going to put a damper on our party. She bought me a new CD, while I bought her a CD player.  I am unsure what are relationship is at the moment. We had a very long talk this past week, about our current \"situation\". I am still not sure where we both stand. Well, that isn\'t completely true. I know where Spinner stands, yet, I am unsure where I stand. Anyway, we might be spending tonight together. Tomorrow morning she will be getting on a bus and going back to Pennsylvania. '),(14251,4042606,'02,August,2004','2004-08-02 00:00:00','   Guess what everyone? I am still in Vermont. Sunday was going to be my last day here, and that was going to be the last time Mute and I were going to see each other. Things change though, and I remain in Vermont with Mute. I have actually no idea when I will go back home, I will probably leave tomorrow or Thursday sometime. I have to ride one of those Greyhound buses so I have to figure it out soon. Anyway, I am spending the day again with Mute, we are probably going to head downtown later if it doesn\'t rain again.  As for Jackie\'s suprise threat to me. I have decided not to post what I was going to post. Instead I printed it out and gave it to Mute when I saw him again. That is one of the main reasons I am still in Vermont. We still have a lot of things to discuss. '),(14252,4042606,'02,August,2004','2004-08-02 00:00:00','   Well I am back, after being offline for a couple of days. Ian let everyone know what was going on, so no one was really worried when I didn\'t post for a while. The virus Adam\'s computer had was not as bad as we originally expected. It was actually a quick fix, it would have been fixed faster if I had Ian around. I hope Ian and Abby had fun together yesterday, their final day together was probably very emotional. Anyway, I am back and I will get back to posting every night. '),(14253,4042606,'13,August,2004','2004-08-13 00:00:00','   If you haven\'t noticed, we are not posting on the website very often anymore. Maybe because of a lack of interest. Or maybe because we haven\'t talked to each other in over a week. All I know is the website is on it\'s last life. It won\'t be around for very much longer, yet I do want to create a new website. I will be in talks with some other people to create a more goal oriented website that actually has a purpose. '),(14254,4042606,'10,August,2004','2004-08-10 00:00:00','   This is the greatest joke I have ever heard in my life. Read it and enjoy:  There once was a bear and a rabbit that hated each other. One day, they found a genie in a lamp who said he would grant them each three wishes.The bear went first and he said,\"I wish to be the only male bear in this forrest.\" And he got his wish.  The rabbit said, \"I want a motercycle helmet.\" And he got his wish.  The bear went up and said, \"I wish to be the only male bear in the United States, and all the rest to be female.\" And he got his wish.  The rabbit said, \"I wish I had a motorcycle to go with that helmet.\" And he got his wish.  The bear said, \"I wish I was the only male bear in the world, and all the rest were females.\" And he got his wish.  It was the rabbit\'s turn, and he said, \"I wish that bear was gay.\" '),(14255,4042606,'09,August,2004','2004-08-09 00:00:00','   I got home from the hospital yesterday, and I am fine. Thanks to everyone that called to see how I was doing. It was even more suprising to hear that Ian was ready to hop on a plane to come down and see me. I thought that was very sweet, and there is no way I can put into words how much I appreciate that.  I will be out of work for a little while, I am on medication that will make me a little unsteady on my feet. They thought it would be best if I stayed home, and I also am getting paid for it. So I will be posting more than I usually do. '),(14256,4042606,'09,August,2004','2004-08-09 00:00:00','   Well, Jackie just wanted to talk to me on the computer all day. Yet, that became very boring, very fast. So now I am going to head downtown if it doesn\'t start raining and pick up a couple of things for my skateboard. I also may pick up a new cd, not at Border\'s though, they over price everything in that fucking place. '),(14257,4042606,'09,August,2004','2004-08-09 00:00:00','   There is absolutely nothing to do today, I didn\'t make any plans to go outside or anything because the news said it was going to rain. I haven\'t seen any rain at all and now I have nothing to do. I guess I will just sit on my computer and chat with Ian, because he didn\'t make any plans either. Hopefully Ian will wake up soon, for some reason he always sleeps until noon these days. I don\'t know how he got into that habit. '),(14258,4042606,'09,August,2004','2004-08-09 00:00:00','   The lightening went as fast as it came, and the two officers were astounded that the figure disappeared right in front of their eyes. As the lighting quickly went back to the dim yellow, Wallace began to regain his senses. He turned to Mark and gave him a nod. Wallace pointed to the stairs and motioned Mark to follow close behind.             â€œI think he went up the stairsâ€ Wallace whispered as he placed his hand on the railing of the stairs.              â€œHow do you know? It disappeared right in front of us!â€ Mark said, in a clearly disturbed voice.             â€œThat isnâ€™t possible Mark; he must have gone up stairs. He couldnâ€™t have moved passed us down the hallwayâ€ Wallace explained.             Wallace and Mark moved up to the very top of the stairs to find more blood on the floor. Yet, the stairs did not have any; moving down the upstairs hallway they stopped as one of the doors was slightly opened.             â€œIâ€™ll check inside you go on aheadâ€ Mark said as he opened the door and peeked inside.             Wallace continued down the hallway, while Mark disappeared behind the jarred door. Sheriff Wallace slowly and quietly moved down the hallway and finally stopped at the end. Where there was a single door, light was coming from inside.             â€œWhy is there a light on? The power is out all over townâ€¦â€ Wallace said to himself as he put his hand on the door knob.             As he pushed the door inward, he placed his hand on the safety of the gun and switched it off. Slowly, Wallace entered the room. The light was so bright that Wallace couldnâ€™t really see what was giving off light.             As Wallaceâ€™s eyes began to finally focus on the source of light, he noticed that it looked like the figure that Mark and him saw downstairs. He still couldnâ€™t make out what the figure looked like.             â€œWho are you? What are you doing in this house?â€ Wallace said as he pointed the gun at the figure.             As he said those words a loud thump came from the hallway that drew Wallaceâ€™s attention. As he looked over and then looked back, the light had faded and the figure completely disappeared.             â€œWhat is that thing?â€ Wallace whispered to himself.             After standing there looking around for evidence, as if still looking for the figure; he quickly turned around again as he heard the sound that occurred just a few moments before.             â€œMark! What is going on out thereâ€ Wallace yelled, yet when he went silent to listen for an answer, he heard none.             â€œMark! Are you hearing me?â€ Wallace repeated again, hoping this time that his call will be returned, but it was not.             Wallace quickly swung around and exited the room that had once contained the figure they were so desperately looking for. As he exited he noticed that the door that Mark entered was shut. He stealthy moved down the hallway and placed his hand on the door knob.             Appling pressure to the door, he slowly went to push it in. Wallaceâ€™s pressure was then met by an equal pressure from the other side. It seemed as if something was holding the door shut.             â€œMark, if youâ€™re trying to play a joke on me, it isnâ€™t funnyâ€ Wallace said as he gritted his teeth. Wallace then knew that Mark wasnâ€™t playing a joke, and that Mark had never played a joke on anyone in his life.             With a worried look on Wallaceâ€™s face, he decided to ram the door inward. As he backed up to prepare for a charge into the door, it suddenly swung open. Standing there in the doorway was Isabel.             â€œIsabel! Youâ€™re alright!â€ Wallace said in complete surprise.             Wallace then realized that she wasnâ€™t alright, she was obviously crying and she was soaked in blood.             â€œIsabel! What happened in here? There is blood all over the kitchen! And I saw a strange figure in the house!â€ Wallace jumped from one statement to another, not giving Isabel any time to respond.             â€œWallaceâ€¦ where is Mark?â€ Isabel quietly responded without emotion.             â€œI donâ€™t know, we split up, I thought he went into the room you just came out ofâ€ Wallace then replied, but reluctantly as if he did not trust her enough with the information.             â€œAre you sure you are okay?â€ Wallace then added as he moved to put his hand on her shoulder.             â€œDonâ€™t touch me!â€ Isabel suddenly screamed as if she suddenly awoken from a deep sleep.             Isabel began to rock side to side, and mumble to herself. Isabel turned to the stairs and began to walk down them. As she reached the bottom, she stopped moving and starred at the closet door. Wallace was eyeing her carefully as he began to follow her. Wallace looked down and waited for her to move, but could tell she had zoned out again.             â€œIsabel you stay right there! I am going to check around upstairs more!â€ Wallace yelled, but knew Isabel hadnâ€™t even heard him speak.             Wallace turned back from the stairs and walked towards the door that Isabel just came out of. Walking in, he noticed that the floor (like the kitchen) was covered in blood.             â€œWhat the hell?â€ Wallace said to himself as he looked at the floor, continuing to walk in.             He pushed the door open a little wider, and then stopped when he noticed there was something behind it that was causing resistance. With fear, Wallace began to pull the door back so that he could see what was behind it.             â€œOh my godâ€ Wallace said in horror when he saw what was on the floor, behind the door.             Wallace looked at the back of the door, only to find two hand prints on it. Hand prints that had been there, trying to keep him out a couple of minutes before. His eyeâ€™s looked down and saw that Mark was lying on the floor. Dead, his head looked as is someone had hit it with a bat. Blood soaked his uniform, and the entire floor below.             Wallace glanced up in the air as if he just realized something and turned and ran out of the room, slipping on the blood as he went.             â€œIsabel donâ€™t you move!â€ Wallace screamed as he pulled his gun from itâ€™s resting place once more.             Yet, when Wallace reached the stairs, he saw that Isabel was no longer at the bottom. He suspiciously crept down the stairs glancing around to make sure there was no one waiting for him at the bottom.             As he finally reached the bottom, he noticed the exact spot in which Isabel was standing. There were imprints of her feet in the blood that had covered the floor. He then began to follow where the prints were heading. As he reached the end of the tracks, he saw that they went directly into the closet.             â€œIsabel! Get out of the closet right now! Step out with your hands in the air!â€ Wallace screamed as he took a step back and pointed the gun at the door.             â€œWally? Helpâ€ Isabel screamed a muffled yelp.             Wallace stood there confused for a minute, and then began to hear nails scratching on the door.             â€œIsabel step out of the closet right now!â€ Wallace remained standing and steadied his gun.             The last noise Wallace heard from the closet was a soft thump and then silence. He quickly walked forward and pulled at the door, yet it wouldnâ€™t open. He placed the gun back in its holster and grabbed the knob with both hands and began to pull frantically.             â€œIsabel! Isabel! Answer me!â€ Wallace yelled as he pulled one last time before losing his footing on the floor and falling.             Wallace lay on the floor starring at the ceiling, when he heard something in the closet once again. He heard the door open up, as he tried to move, he noticed that he couldnâ€™t. His heart beat quickened as he heard someone stepping closer and closer to him. He tried lifting his head, yet the pain was so great that he was unable to. The stepping stopped and he looked up at what was looking down at him. It was Isabel; she was looking in his eyes, as he was looking in hers.             â€œWhere is Emilyâ€¦? What have you done to Markâ€¦?â€ Wallace managed to squeeze out in between the shots of pain in his back.             Isabel did not answer though, and she continued to look into Wallaceâ€™s eyes. The pain began to take the better of Sheriff Wallace, he lost consciousness and everything went black. The last thing he was looking at was the lifeless look in Isabelâ€™s eyes.             Wallace stirred back and forth on the floor. He opened his eyes and closed them again, trying to make the pain in his back go away. It was no use though, he knew something was broken and that he would not be able to move.             â€œIsabel where are you?â€ Wallace finally broke the silence in the house, expecting to see Isabel near by.             â€œWhat is going on in here? Please call 911, I think my back is brokenâ€ Wallace pleaded to an uninterested ear.             He began to move his head around a bit, just so he could see if anyone was around him. That is when he noticed that the closet door was wide open. And when he looked down, he saw the Isabel was no longer in it. He looked at the back of the door, and there were those bloody hand prints again. Just like in the room upstairs. He could only study them for a second when a shot of pain went up his back, and he had to put his head on the floor again. '),(14259,4042606,'08,August,2004','2004-08-08 00:00:00','   My day off was pretty fun actually, I went over to Sarah\'s early in the morning and did something that I was not expecting to do. Later I went skateboarding and then returned to have a barbecue with Sarah\'s family, I then went skateboarding again. Now I am sitting home at eight (something I rarely do) and I am bored out of my mind. But, this boredom has given me the chance to finish up on that short story I started back when this website was first created. '),(14260,4042606,'08,August,2004','2004-08-08 00:00:00','   Don\'t think I haven\'t forgotten about my cell phone. I wanted Abby to send my cell phone through the mail to me. I called it earlier today, and she answered it. I bet she hasn\'t been yelled at like that since she was a little girl. I think she needs to send my fucking phone back to me right away, and stop making long distance calls. '),(14261,4042606,'08,August,2004','2004-08-08 00:00:00','   I heard about what happened a little while ago. I really hope Michelle will get better soon, I heard that she looked like she was 3 months pregnant because it was so swollen.  Other than that news, things have been pretty uneventful for me. I know Ian\'s weekend was much more confrontational then he would have liked. Mine including sitting on my couch on a Saturday morning watching Sonic X. It is very embarrassing, but there was nothing else to do. '),(14262,4042606,'08,August,2004','2004-08-08 00:00:00','   Late last night, it was brought to my attention by Adam that Michelle was rushed to the emergency room. What happened was that she was working somewhere when she accidentally stepped on a Mud Wasp nest. She was stung 20 times on her side, did I mention Michelle is deathly allergic to bee stings? Well, she is.  They found her passed out on the ground. She was then taken to the hospital, and that is when they called me. I was thinking about heading down there on an airplane, but they said she would be fine.  Anyway, she is going to be in the hospital for a couple of days and she will not be able to post at all. I am not sure if Abby has heard about this yet, but I am not going to tell her. Jackie already knows though and she will be posting soon too. '),(14263,4042606,'19,August,2004','2004-08-19 00:00:00','   I never finished the drawings I was supposed to do last night. My plan was to draw them and then post the pictures on this website, but plans changed. Instead I drank a lot of wine and watch   urlLink Cool Runnings   twice before I finally passed out on the floor. I awoke this morning in a pool of my own drool, and my dad was not to pleased to find me there, on the floor unconscious, with   urlLink Cool Runnings   playing in the background. We then had a little talk about my drinking habits and how I am underage and shouldn\'t drink. I told him that he need to trust me more, and that I only get drunk when I am alone. He didn\'t like hearing that either. So I think I am in a little trouble for today, but I will be back in a little while. '),(14264,4042606,'18,August,2004','2004-08-18 00:00:00','   This was possibly the worst night of my life. Starting off with squeezing into that tight dress, that showed off way more then I wanted too. Then I had to listen to Mute and Jackie taunt me for 2 hours, before I even left for the dinner.  As I sat in the car, I couldn\'t breath because the dress was to tight around my chest region. I didn\'t care for that at all. My stepmother also made the comment that I could use to show off myself a little, she also said that I am to pretty to hide it under all that clothing. Well, she is wrong and I don\'t need to dress up fancy for people.  Getting back to the story, we got there at around eight. To see a whole shit load of people already inside. All I wanted to do was sit down somewhere, where no one could find me. That didn\'t work out to well, because my dad and stepmom wanted to introduce me to every single person in the building. Guess what, There was no eye contact the entire night, unless you consider my chest eyes. It was ridiculous, and people wonder why I dress modestly. I cannot stand this shit.  Well, as I went from meeting each person, I continued to look for a way to get the hell out of there. I found my exit, it was the bathroom. I said that I had a stomach ache and then went to the bathroom. I stayed in there for about 15 minutes, but when I came out my dad had more people for me to meet.  When I finally sat down to eat, I was luckily sat against the wall, so I was no longer in the center of attention. I thought the hell that was the night was finally going to calm down. Not exactly; When dinner was served I caught this one old guy starring at me the entire time I was eating. If that doesn\'t make someone uncomfortable, I don\'t know what does.  We left that hell hole at midnight. I got home at about 12:10 and I am finally out of that tight dress I had been wearing since seven. I am still trying to get my normal breathing patterns back. I am completely exhausted from this night, and hopefully I don\'t half to go through it again. '),(14265,4042606,'18,August,2004','2004-08-18 00:00:00','   Abby is off at the banquet, Ian is in what he calls \"the zone\", and I am left bored out of my mind. It seems as though everyone else has something to do but me. Usually it would be the other way around. I guess I can\'t complain though, this is the first night I have been able to just sit around and do nothing. I finally can have sometime for myself, in which I can reflect on my life. Actually I think I am going to get drunk and watch  urlLink Cool Runnings  until I finally pass out on my floor. It sounds like a plan to me, so I won\'t be posting until sometime tomorrow. '),(14266,4042606,'17,August,2004','2004-08-17 00:00:00','   I have lost Ian for the night, apparently he finds Kingdom Hearts more interesting than anything else in the world. Even, dare I say? Women. That\'s right, I have discovered that Ian actually enjoys playing Kingdom Hearts more than talking to me, or anyone else in the world. I hope his house doesn\'t catch on fire because he would surely die. Blissfully but dead none the less. Anyway, I am finishing the picture drawing fiasco that I got myself into. Ian was going to help, but I don\'t think that is going to happen now that he is playing with Goofy and Donald. '),(14267,4042606,'17,August,2004','2004-08-17 00:00:00','   I have had the urge to play some Kingdom Hearts tonight. So you people will probably not see me until Friday. I will probably be sucked into it like I was yesterday, so I will be up until 1 in the morning playing it. Then I will be going to work until eight. So yeah, my time is short. Typing this now is even taking up precious time that could be going to Kingdom Hearts. So I am going to stop talking now, and go save Atlantica from the evil clutches of the Heartless. '),(14268,4042606,'17,August,2004','2004-08-17 00:00:00','   I  am currently working on making a picture for \"Team Kakutama\" which basically consists of Abby, Ian, and I. I hope you guys will like it even though this is usually Ian\'s department. '),(14269,4042606,'17,August,2004','2004-08-17 00:00:00','   I sit here on my computer, before I go to leave for this banquet that I was forced to go to. I look like I am some rich snob. I also happen to notice that the dress I am wearing was probably worse than the one that showed off to much cleavage. It is pressing my breasts together so much that it hurts, and it makes them look even bigger than they actual are. It will be a mirical if I get through the night without having someone grab at me. As for Jackie and Ian\'s idea of taking a picture and posting it on the website, you two can go to hell. I want this night to be over, and I don\'t want any evidence of it actually happening. It\'s bad enough that I am going to have this dress hanging in my closet. '),(14270,4042606,'17,August,2004','2004-08-17 00:00:00','   Abby has continued to complain for about 3 hours now. The banquet is at eight, and she has to start getting ready at five. It is funny, because she hates wearing dresses so much, because she can\'t hide her \"endowments\". I know for a fact that this is going to be the most uncomfortable night for her in her entire life. I also wish that Ian and I were there to see this monumental event.  Ian and I have been ridiculing Abby for two days now, and we won\'t stop until the night is over. '),(14271,4042606,'17,August,2004','2004-08-17 00:00:00','   My stepmother\'s boss has invited her and her family to a banquet tonight. Now this may seem nice, except for the fact that I am apart of her family. That means I am not only going to have to show up at this dinner, but I also find out that I am going to wear some crazy evening dress, because my style of dressing is unacceptable at this place.  I\'ll let everyone in on a little secret, I don\'t like dressing up for anybody. I am fine with my sweater and cargo pants. Yet I was going to be wearing some black dress that shows a little to much cleavage. I told everyone I wasn\'t going if I had to wear that dress, luckily my stepmother found a dress that was a little more in my taste. So now I have to go.  When I told Mute about this, all he did was start laughing, he added after the laugh \"Have fun with guys staring at your boobs all night\". While Jackie may not have a problem with showing a little breast, I do. This is going to be possibly the worst night of my entire life. And Mute and Jackie are getting a kick out of it. Good thing they are having fun, because next time I see them, they are going to get their asses kicked. '),(14272,4042606,'17,August,2004','2004-08-17 00:00:00','   Well Sarah left today to go to college. I still didn\'t figure out what college she was going to. I think she told me once before, but I guess I just didn\'t remember. Somewhere out in West Virginia I think. '),(14273,4042606,'17,August,2004','2004-08-17 00:00:00','   In 1994, The Elder Scrolls: Arena was released to the public. It was amazing to say the least, it\'s non-linear gameplay and nicley developed RPG system was raved about by everyone whom played it. Here is a description of The Elder Scrolls: Arena by the people whom made it,      Arena started, as the name might suggest, as a medieval-style gladiator game. You had a team of fighters and went around the world fighting other teams in each city\'s arena until you became grand champion in the Imperial City. The world used for Arena was Tamriel, the fantasy world created by a few members of the staff for use in their weekly D&amp;D campaign. During development of Arena, more and more RPG elements were added -- what if you could walk around these cities? What if you could take your team into a dungeon? And soon it was clear what Arena needed to be -- a full-blown RPG.  Inspired by such games as Ultima Underworld and the unheralded Legends of Valour, Arena was now seen as a massive first-person RPG -- the game that recreated the pen-and-paper RPG experience -- be who you want and do what you want. For a long time Arena was a party-based game, where you lead a group of adventurers. This style of play proved less fun in first person, so Arena became a single character game.   Arena was to be the first chapter in an ongoing series of games, so the series took its name from Tamriel\'s mystical tomes of knowledge that told of its past, present, and future -- The Elder Scrolls.   Now why am I talking about a 10 year old game? The answer is simple, the people over at  urlLink Bethesda  have re-released The Elder Scrolls: Arena. Not for sale mind you, but for free. You can download it at their Elder Scrolls  urlLink website . I suggest you do, because I already did and I love it. '),(14274,4042606,'15,August,2004','2004-08-15 00:00:00','   I forced a copy of   urlLink Start Something   by the   urlLink Lostprophets   on Mute today. Mute came home today from his friends house to find that the mailman dropped it off in his mail box. Mute said he didn\'t like it at first, but after listening to it for a while, he told me they had some good qualities. Of course I knew this, that is why I sent it to him in the first place.  The   urlLink Lostprophets   are probably one a band that I enjoy quite a lot. It is very embarrassing to admit this though, because of their \"fake punk-rock\" image they acquired. I don\'t want to like them, yet I am compelled to. Just like Muteman is. '),(14275,4042606,'15,August,2004','2004-08-15 00:00:00','   If you haven\'t heard, on the  Viral Cartoon Show  Michelle announced that she was leaving both websites. She asked both Adam and I for our advice, and we both told her the same thing. That if she didn\'t want to be on the websites anymore, that no one was going to force her to do it.  I guess she agreed with us, and now she will no longer be posting on the website. '),(14276,4042606,'15,August,2004','2004-08-15 00:00:00','   After hanging out with Sarah for a little while this morning, I returned home and sat in my chair deciding what I was going to do with the rest of the day. As I sat there contemplating my next action I saw my copy of   urlLink Kingdom Hearts   sitting on top of my Playstation 2. I decided I had nothing better to do, so I started playing it.  6 hours later, my foot was asleep, and I was two inches from my TV screen. I some how was trapped in a little world in which nothing but saving the world from the Heartless mattered.  Apparently while I was trying to save everyone from the Heartless, Abby was trying to get a hold of me. I should have heard my cell phone ring, when it was sitting right next to me, but I was so busy with the keyblade that I tuned every other noise out except the sounds of Daffy Duck yelling \"Sora!\". '),(14277,4042606,'15,August,2004','2004-08-15 00:00:00','   Okay, so now the other website is gone, after lasting about a week. Abby has become friends with Ian once again, and I wanted to only focus on one thing at a time. So here we are again.  It\'s funny, no matter where we go, or threaten to go. It looks as if we always come back here, as if it is a secret mental meeting place for us. Anyway, that site is gone and now the three of us will be focusing on Kakutama. The only thing that hasn\'t changed in the past day, is the name of this website. '),(14278,4042606,'15,August,2004','2004-08-15 00:00:00','   Looks like I am running Kakutama now. Ian has gone crawling back to Jackie like a lost puppy. So I guess I will run this myself, Michelle is not posting for some reason too. I think everyone is going against me. Oh well, I guess things were bound to go this way eventually. '),(14279,4008161,'24,July,2004','2004-07-24 00:00:00','   urlLink    this is him&nbsp; urlLink '),(14280,4008161,'23,July,2004','2004-07-23 00:00:00','   This is the boy I am obsessed with.&nbsp; This is how I think of him: genius guitar god&nbsp;exuding joy and love that crowns him like&nbsp;a halo as he bends over, intent upon creating something beautiful.&nbsp; This is me: the unseen person behind the door,&nbsp;peering through the window for a glimpse of&nbsp;what I can never have but want with all my being.&nbsp;&nbsp;I can do nothing but obsess over him from far away, listen to his music and imagine that it is for me,&nbsp;wait for him to look up and spot my eager face, framed by the&nbsp;jail-cell window.&nbsp; Will he catch my eye and grin?&nbsp; Will he come out to meet me?&nbsp;&nbsp;Will he invite me inside?&nbsp; Or will he just smile sadly, pick up his guitar, and&nbsp;climb&nbsp;up the steps, even further out of reach?&nbsp;   This is new to me - this longing, this&nbsp;lowly position of dependence.&nbsp; Is this payback for always being the bitch?&nbsp; The tables have turned, and now I am the one crying, wishing, waiting.&nbsp; The thing is, I don\'t even mind.&nbsp; I know I am pathetic, but that knowledge does not&nbsp;create any feelings of shame or remorse at my&nbsp;revelations of&nbsp;undying love and my&nbsp;desperate yearning for reciprocation.&nbsp;&nbsp;Surprisingly, I do not feel sad.&nbsp; This is his power;&nbsp;the mere thought of Him, even if it is not&nbsp;a happy one, makes me smile.&nbsp;&nbsp;   I analyze his every word, his every punctuation mark, searching for clues -&nbsp;clues to Him.&nbsp;&nbsp;Calling forth&nbsp;the obsessive analyzation skills built into&nbsp;every female, I ponder the millions of possible meanings in his&nbsp;IMed sentance fragments.&nbsp; Long afterwards, I replay&nbsp;our conversations over and over again in my mind.&nbsp; Why did he say it that way?&nbsp; What if he meant this?&nbsp; No, he meant this.&nbsp; But wait!&nbsp; Remember when he said that?&nbsp; Oh yeah, so he must have meant&nbsp;this!&nbsp; The choices are mindboggling, and I drive myself crazy.&nbsp; But I love it.&nbsp; I love him.&nbsp; I love everything connected to him, even if it is my doubts.   He hasn\'t talked to me in a while.&nbsp; But I don\'t mind - I\'ll wait.&nbsp; A month ago, I would have&nbsp; ruthlessly mocked any girl who let herself be used by a boy in this way.&nbsp; But now, I don\'t care -&nbsp; I have been injected with this new sense of hope and optimism.&nbsp; Any embarassment I feel is immediately erased by the prospect of talking to him.&nbsp; I try to wean myself off of him.&nbsp; I am not going online today, i say.&nbsp; I refuse to call him.&nbsp; But it doesn\'t last.&nbsp; What if he is waiting for me?&nbsp; What if he thinks I am mad at him?&nbsp; What if he gets mad at me?&nbsp; What if....and the infinite number of what-ifs drives me to the computer screen, where I log on to my desktop - a photo of him hugging me.&nbsp; In the picture, I am ecstatic, and staring at it as I wait for the internet connection makes me realize how happy I am now - I\'m going to talk to him.     &nbsp; '),(14281,4008161,'23,July,2004','2004-07-23 00:00:00','   I&nbsp;don\'t know why&nbsp;I created this blog -&nbsp;I do not flatter myself into thinking that anything I write is important or original enough to influence others.&nbsp; I am your typical, white, upper-middle-class teenage girl who likes to think she is special.&nbsp; I do keep a journal - a small, black notebook.&nbsp; I never let anybody read it, not because it protects deep, dark secrets that have the potential to&nbsp;ruin my reputation (if I have one), but because I have never written for anybody but myself.&nbsp;&nbsp;   The notebook is full.&nbsp; Yesterday, I filled the last pages exactly two years after I bought it and wrote my first entry.&nbsp; My two year anniversary.&nbsp; Should I celebrate?&nbsp; Celebrate two years\' worth of embarassments, fights with parents, random crushes, and bad, experimental poetry?&nbsp;   In an attempt at nostalgia, I tried to read over some of my past writing, but I couldn\'t.&nbsp; Honestly, I felt so embarassed at the raw emotion that came through that i slammed the book shut and stuffed it back into its place at the corner of my desk.&nbsp; All the things I was so passionate about a year ago, a month ago, I don\'t care about anymore.&nbsp; I\'ve moved on, and I cringe as I read my dramatic words and remember things I would rather forget.&nbsp; And then I wonder, all these&nbsp;things I&nbsp;am experiencing now -&nbsp;all the things that are&nbsp;really important to me now -&nbsp;in a few weeks or months, will they have&nbsp;assumed the status of embarassing trivialities that I once obsessed over?&nbsp; It is reassuring yet frustrating at the same time.&nbsp; While it is a comfort to remember that&nbsp;my present worries will soon pass, I resent the idea that my present worries and obsessions are not as important as I think they are.&nbsp; I would rather get pissed over these tiny little problems than go through life chanting, it could be worse, it could be worse.&nbsp;&nbsp;   By the way,&nbsp;I hate that whole philosophy of putting things into perspective.&nbsp;&nbsp;Every time I complain about school or illness or&nbsp;a disagreement with a friend, my parents always tell me to Put It into Perspective.&nbsp; It could be worse, they say, you could be starving in Africa, living in a tiny hut with ten siblings.&nbsp;&nbsp;How can you complain about your mild case of strep throat&nbsp;when&nbsp;millions of people are suffering from AIDS? they ask me.&nbsp;&nbsp;Well, because I don\'t have AIDS.&nbsp; And is it really my fault that I have plenty of food in my fridge, or that I live in a nice house?&nbsp; I don\'t see why I should pretend to be completely happy all of the time&nbsp;just because somebody else\'s life is worse.&nbsp;&nbsp;  I admit&nbsp;I&nbsp;sound like a&nbsp;brat, but if you think about it, Putting Things into Perspective basically guarantees that you won\'t experience&nbsp;raw emotion.&nbsp; Unlike many others, I think of emotion as an asset rather than a weakness.&nbsp;&nbsp;So many people distrust their natural instincts.&nbsp; they tell me not to let myself get carried away by my feelings - don\'t make a decision until you have calmed down.&nbsp; But they are kidding themselves - all those people who profess to be objective -&nbsp;we are all human, and we all let our feelings influence us.&nbsp; And that is not necessarily a bad thing.&nbsp;   So...yes, this&nbsp;webpage&nbsp;will be filled with my random thoughts and&nbsp;adolescent&nbsp; angst.&nbsp; I apologize, but these are my problems; they may seem petty and routine, but they are all&nbsp;mine. '),(14282,4008161,'18,August,2004','2004-08-18 00:00:00','   It was one of those Matrix-like scenarios where everything you think is real turns out to be an illusion. In the dream, my parents were throwing me a grad party, and I was at the airport to pick up one of my friends. In the US Airlines terminal, Antonio Banderas, wearing a long black leather coat and dark glasses approached me and told me I was adopted. Of course, I laughed at him and told him to leave, but then, he showed me Proof (do not know what that proof was, but it was sufficient to make me believe his claims). So then, he told me that if I wished, my real father would explain everything to me at my grad party. Of course I consented, though in real life, I wonder why I would ever consent to let this man and my real father into the house inhabited by my imposter parents.  When I arrived home that day, my house was not at my driveway but had been replaced by my middle school as if a forklift had just picked it up and moved it. My family was nowhere to be found, but my party had already begun (even though it wasn\'t scheduled for another week). I met Antonio Banderas inside, and he directed me into a back room, where he set up a laptop computer. He then proceeded to show me webpages and articles that explained who my father was - a member of death row who had been in federal prison all of my life. Antonio Banderas informed me that the authorities have no idea who my mother is, and they only recently tracked me down. My father was a murderer. The pointer finger of his right hand was longer than his middle finger, and he was such a ruthless killer, his nickname was \"Trigger Finger.\" I stared in horror at the photos of my father\'s mutated hand before screaming profanities at Antonio Banderas and running from the room.  I ran from my party and continued to sprint as far away as possibble. I thought I heard somebody chasing me, so I kept running until I couldn\'t anymore. I ran until I fell, gasping, onto my knees. Clutching my side, I stumbled to my feet only to recognize the ruins of Alcatraz Prison, just off the California coast. When I looked down at my muddy hands, the pointer finger of my right hand was longer than my middle finger.   I jerked awake at 5 a.m, sweating profusely. Trailing my hand across the walls for support, I scrambled into the bathroom and flipped on the light switch, ignoring the cry of my pupils as they shrunk. I fingered my right hand, measuring the fingers over and over again. In the mirror, my eyes were red and moist, my skin pale beneath my tan. I splashed myself with cold water and returned to bed, clutching my pillow. '),(14283,4138081,'04,August,2004','2004-08-04 00:00:00','    \"To fear love is to fear life, and those who fear life are already three parts dead.\"- Bertrand Russell. This is a very interesting quote I found in a book I\'m reading. To fear love, a big part of life, would be the same as fearing life. You would fear the only things you care for and care for you hopefully. Fearing life would be fearing all things including the good with the bad. Happiness is an emotion that keeps us alive and without love, it\'s not an emotion you will feel. Do not fear love. It may hurt and tear you apart, but you musn\'t hide from it. You\'ll never know if the feeling is mutual or not unless you find out for yourself. Show your love and express it with the greatest enotion possible. Show people you are happy and you live and love to the best you can. This seems like a hard thing to do. Trust me, it is. But you can\'t live in fear for your whole life. That\'s not what inspires people to want to be around you. Understand that life will bring you happiness when you want it. Not when you need it. Looking for happiness and living for happiness are two different things. Live to find happiness. Show you are happy and people will follow.   '),(14284,4138081,'04,August,2004','2004-08-04 00:00:00','    \"Never regret. If it\'s good, it\'s wonderful. If it\'s bad, it\'s experience.\"- Victoria Holt. This is an interesting quote that I found the other day. The purpose of this quote is to recognize that everything happens for a reason. Every choice you make can either be bad or good. The bad ones allow you to learn from this mistake and apply what you\'ve learned to your everyday life. The good ones give you some type of happiness that you don\'t have to regret doing. Life has twists and is going to throw some crazy ones your way so your going to make some bad decisions, but hopefully you\'ll learn from them and fix it so that you will live happy. Regreting something is the same as feeling bad for yourself. Don\'t do it and you can\'t be depressed. Forget the past and learn from it.   '),(14285,4138081,'04,August,2004','2004-08-04 00:00:00','    Writing is more than an assignment by a teacher or just making a book. Writing is a state of mind that allows you to use your mind as a tool and understand things better. You endulge in your own world and make your own rules. Noone can bother you because you are so focused that your mind blocks everything out allowing you to continue. The best thing about writing is the feeling you get after you write. You feel complete because you were able to tell your true feelings. Even if you write by yourself and allow noone to read your writings, that feeling is still shown. My writings will be based on truth and my feelings so I hope you enjoy them. I will show facts about life, love, the world, and many other subjects that purtain to you and me. Hopefully both you and I will learn something from my writings. Quote of the day \"I never think of the future - it comes soon enough.\"- Albert Einstein. My explanation for this quote is that you must live life one day at a time. Don\'t worry about the future to come. It will come and you will see what happens either way. Forget about fear of death and think about life. The future is unpredictable so stop trying to figure it out. Life will give to you what\'s rightfully yours. This will in turn make you much happier because fear will be a lost emotion.   '),(14286,3519860,'29,June,2004','2004-06-29 00:00:00','    Your eyes which first held me captivated where I stood.  Your smile to dazzle the sun and warm every corner of my soul.  Your voice like a sparkling mountain stream which flows into my heart.  Your walk and the way your gracefulness takes my breath away.  Your hair about which I dreamed cascading into my face as you leaned over me.  Your hands whose caress I crave to hold my face in their tenderness.  Your arms I long to have around my neck as you pull me close to your warmth.  Most of all everything you are changed the way I feel about my life.  I love you.   '),(14287,3519860,'11,June,2004','2004-06-11 00:00:00','    at night I dream of you, Coming to be by my side. I see your silhouette at first; I watch your calm, quick glide.  Vivid images; Pictures in my mind. My night\'s love, my passion; All in you I shall find.  Once again, you\'ll whisper to me, \"Hello, my dear. \" Your touch eases me, Removing any fear.  Your gaze meets mine; The warm look in your eyes, It melts my heart. From now on Nothing but blue skies!  Tonight you romance me; Our bodies entwine. You enter; I gasp! Our souls combine.  A touch here, a tickle there, It enthralls me; I am captivated. Our bodies move, the motion constant; My body\'s completely activated.  Throughout the night Our passion flows, Coming together, My emotion grows.  For you, in dreams, Oh, how I care. But when morning comes, I am left in despair.  Anther night gone and passed. I awake in ecstasy all alone. For you visit at night; In dreams you come. And I dread the coming of a new dawn.   '),(14288,3519860,'04,June,2004','2004-06-04 00:00:00','    The first time I saw you, I knew you were the one. There was no question in my mind, after having that much fun.  There\'s just something about you, that I\'ve never found before. And now that we\'re good friends, I find myself wanting more.  I never thought it possible, to care for one so much. But everyday that passes, I long to feel your touch.  To look into your eyes, would make my day complete, cause every time you look at me, my heart just skips a beat.  Everything I\'d wished for, every dream I\'d hope come true, I never thought would happen, until the day that I met you.  I know that youâ€™re not ready, for the kind of commitment that I am, but for you I\'d wait forever, there\'s nothing I can\'t stand.  Although it hurts to know, that you don\'t feel the same as me, hopefully one day you\'ll realize, one day you\'ll wake up and see.   '),(14289,3519860,'01,June,2004','2004-06-01 00:00:00','    Behind Blue Eyes\" limp   No one knows what it\'s like To be the bad man To be the sad man Behind blue eyes And no one knows What it\'s like to be hated To be fated to telling only lies  [Chorus:] But my dreams they aren\'t as empty As my conscience seems to be I have hours, only lonely My love is vengeance That\'s never free  No one knows what its like To feel these feelings Like I do, and I blame you! No one bites back as hard On their anger None of my pain and woe Can show through  [Chorus]  Discover l.i.m.p. Say it [x4] No one knows what its like To be mistreated, to be defeated Behind blue eyes No one knows how to say That they\'re sorry and don\'t worry I\'m not telling lies  [Chorus]  No one knows what its like To be the bad man, to be the sad man Behind blue eyes.   '),(14290,3519860,'06,July,2004','2004-07-06 00:00:00','   Tomboy    urlLink   What\'s your sexual appeal?     brought to you by  urlLink Quizilla   '),(14291,3519860,'03,July,2004','2004-07-03 00:00:00','   Nightmare Before Christmas!    urlLink   What movie Do you Belong in?(many different outcomes!)     brought to you by  urlLink Quizilla   '),(14292,3519860,'03,July,2004','2004-07-03 00:00:00','   A GARAGE-GURL. Youre into loud music, hot guys and wild fashions. Youre most at ease when you\'ve got all your mates around you and you like to party. Boys are a game and youre always on the ball because you make sure you\'re always number one.  Your virtues: Confidence, fun nature, sociability.  Your flaws: Loudness, jealous tendency, need for attention..    You almost certainly wouldn\'t like this game, because it\'s not your thing:  www.life-blood.vze.com    urlLink   What kind of girl are you?     brought to you by  urlLink Quizilla   '),(14293,3519860,'03,July,2004','2004-07-03 00:00:00','   You should be dating a Gemini 21 May - 20 June This mate is inquisitive, entertaining and charming, liberal, broad-minded and youthful. Though Gemini has a tendency to be impatient, gossipy and sometimes irritable, this  twin has the ability to expresses his or her pent up emotions during sex!    urlLink   What Zodiac Sign Are You Attracted To?     brought to you by  urlLink Quizilla   '),(14294,3519860,'01,July,2004','2004-07-01 00:00:00','   Your soul is WILLFUL. You are determined and a little reckless, and you do whatever you want to do. You have strong opinions and are not easily swayed, and your headstrong resolve is not easily countered. You have few regrets. People find your refusal to go down without a fight formidable, and they respect you for it. You are a proud and content soul.    urlLink   What Is Your Soul\'s Trait?     brought to you by  urlLink Quizilla   '),(14295,3519860,'01,July,2004','2004-07-01 00:00:00','   Your a goddess! Forget angels, its all about you. Angels serve god, right? So why not be a goddess? Hold on there, girl, and easy with the fame. A Goddess is absolutley the most beautiful, powerful, and magic being alive there is, but they can tend to be quite obnoxious. Some goddesses go a little to far with their power, and turn greedy and shallow. Good ones are pure and full of life, and if on earth, flowers sprout at their steps. Others, decided to live in a palace in heaven, with such beauty never knwon. They are very kind and good leaders, but can tend to be a little lazy, jelous, greedy, and obnoxious.    urlLink   What Kind of ANGEL are you? (For Girls only) This Quiz has amazingly Beautiful Pictures!     brought to you by  urlLink Quizilla   '),(14296,3519860,'01,July,2004','2004-07-01 00:00:00','   You\'re Elemant is Wind. You\'re light-hearted, care-free, kind, sensative, and mysterious. You have friends and most absolutely love you. You can be calm and soothing one minute and ragging in anger the next so no one wants to get on your bad side. You\'re beauty is inspiring and magical.     urlLink   What\'s Your Element(girls)? (PICTURES)     brought to you by  urlLink Quizilla   '),(14297,3519860,'01,July,2004','2004-07-01 00:00:00','   You\'re Skittles!!!  You have a very interesting personality, you\'re so unique.  You\'re the kind of person who always thinks outside of the box. You\'re also a very accepting individual, and believe in inner beauty.    urlLink   Which kind of candy are you?     brought to you by  urlLink Quizilla    my favorite candy!!!   '),(14298,3519860,'01,July,2004','2004-07-01 00:00:00','   You have a heartsick soul! Youre the type of girl who always has a crush and is writing their name on all your books. You are a hopeless romantic. Waiting for that prince charming, you take love seriously, but still play any chance you get. You can have a lot if boys who are friends, but waiting for that perfect boyfriend. Sometimes you are discouraged because there are no sparks but even if the smallest thing happens, youre on Cloud 9.  You believe in true love and wait for it. Just dont be afraid to take a chance. Love is all about risks.    urlLink   What Kind of SOUL do you posses? (For Girls only) Incredible Anime Pictures!     brought to you by  urlLink Quizilla   '),(14299,3519860,'01,July,2004','2004-07-01 00:00:00','   hand holding - you like to be in constant physical contact with your special someone but you don\'t want to take things too quickly.    urlLink   What Sign of Affection Are You?     brought to you by  urlLink Quizilla   '),(14300,3519860,'01,July,2004','2004-07-01 00:00:00','   Your: Wondering eyes. Your not quite focused and your quite the day dreamer. Your a bit odd and as many say \"Your head is in the clouds.\"     urlLink   What type of eyes do you have?     brought to you by  urlLink Quizilla   Wondering eyes. Your not quite focused and your quite the day dreamer. Your a bit odd and as many say \"Your head is in the clouds.\"   '),(14301,3519860,'01,July,2004','2004-07-01 00:00:00','   You\'re Exotic Dancer Barbie.  You have some moves, and will do anything for a few bucks.  Take it off girl, but keep it PG-13 please.    urlLink   If You Were A Barbie, Which Messed Up Version Would You Be?     brought to you by  urlLink Quizilla   '),(14302,3519860,'01,July,2004','2004-07-01 00:00:00','   your fuck.    urlLink   What swear word are you?     brought to you by  urlLink Quizilla   '),(14303,3519860,'01,July,2004','2004-07-01 00:00:00','   You have a mysterious kiss.  Your partner never knows what you\'re going to come up with next; this creates great excitement and arousal never knowing what to expect.  And it\'s sure to end in a kiss as great as your mystery.    urlLink   What kind of kiss are you?     brought to you by  urlLink Quizilla   '),(14304,3519860,'01,July,2004','2004-07-01 00:00:00','   BEAUTIFUL ICE PRINCESS/PRINCE .You need distance between you and your partner in your relationship. You are very difficult to get. You have big requirements and this one you love must try hard to get you. But after she/he melt your heart she/he will be the most happy person in the world. You need someone who shoes you that you are special  and it makes you feel good to see that you are loved. She/He shall know that you could easily get another girl/boyfriend but you wont  as long as you love him.  when she/he hurts you you will hurt him too, but in general you dont get hurt. If your partner cheated you ,you would react cold and immediately (try to) forget him PLEASE VOTE, I want to know what you think about my quiz, I worked hard on it.You can always message me or tell me how I can improve that quiz. Ill sure write back.    urlLink   	~THE big LOVE TEST!! What do you need? With PICS! For girls and boys!~     brought to you by  urlLink Quizilla   '),(14305,3519860,'01,July,2004','2004-07-01 00:00:00','   Timeless moon Weeps. Tears of saline  for the should have known.  How could it sit up there full,  round,  and bright. Not to have seen?  Not to have felt, the uneasiness. Or did it just ignore the signs?  Judas moon, will your color  be red on the \'morrow?  Has the last soul been used? Will the emptiness bring about my demise?  I beg the answers. It is I that needs a new  soul.  It is I that weeps for the unknown.  It is I that ignores. It is I  that will stand rigid as the blood of a moon drips down upon my head.   '),(14306,4100934,'02,August,2004','2004-08-02 00:00:00','   For the first time in my life I am realizing how little I am. Not only am I vertically challenged, but I just  feel  small. When I was a child, I associated growing with \"growing up.\" Now that I\'m done growing (physically) I associate growth with realizing how insignificant I really am.  Is this what is known as \"maturity?\" I think maybe so. I attribute this epiphany to my increasing awareness of world and political events. To put it bluntly:  there\'s a whole lot of seriously fucked-up shit going down and there isn\'t much that I can do about it.  Before, I just accepted it and never questioned. Now, I think, \"WHY???\"  Maybe there  is  something I can do. What? I wish I knew. Maybe, one day, I will find something that I can do.  Sometimes, I feel scared. I have these moments where I realize how completely alone I am in the world. I have friends. I have wonderful friends and wonderful family. I feel so loved.   I\'m not sure if I love them as much as they love me. I mean, I do love them, but they love me a  lot . They never give up on me.  But even knowing that, there is still this sense that I\'m really just alone. There isn\'t anyone that completely knows what\'s going on in my little head except for me.  This scares me. I\'m the type that wants a person to completely know me.  It doesn\'t matter how much I let them in on--- it never feels like enough.    I attribute all of the above to my lack of faith in God. Many people pray.  God supposedly knows everything about you and then some, so in theory, from conception to death, you are never truly alone.  I\'m way too skeptical right now to accept the existence of God, or at least the existence of a God that actually listens to us in the way that another person does when we confide in them. I wish I could believe this idea, but I just don\'t.  So, yeah...little me.   '),(14307,4100934,'02,August,2004','2004-08-02 00:00:00','   Something I\'ve learned that I think is worth sharing here:  Relationships evolve. You can\'t force a relationship to be what you want it to be, and sometimes, you think a relationship is something that it isn\'t. When you give your love, trust, and honestly to someone, that doesn\'t automatically mean that they can give it back. Sometimes, it\'s not even personal. If there is a lot at stake, sometimes, they simply can\'t do it even though they do care about you. This realization is hard to cope with. It\'s very disappointing, hurtful, and embarrassing, but for big-hearted people it just comes with the territory.  Another thing to consider: You can either pine for the friends you  wish  you had, or you can appreciate the ones you  know  you have.  I also wonder sometimes if trust and honesty in a relationship doesn\'t just happen in its own time. In relationships where I truly feel that I can trust a person and that we are both very honest with each other, I don\'t really think about the fact that this honesty and trust is present. I don\'t even remember the day when I realized that we have trust and honesty; it just snuck in somewhere. I think it automatically happens when you are both ready for it. Forcing it can cause a lot of pain, guilt, and even more distrust than would normally exist. I think that being angry with someone for not completely trusting you is a huge mistake because then they will probably never be able to trust you completely, even though they might really love you and want to be able to.  Man, the older I get, the more I hurt. It  sucks.   '),(14308,3976199,'30,July,2004','2004-07-30 00:00:00','   Ok, so this is going to be my little venting session because I know my sister can\'t read/probably won\'t ever see this. Can you guess what its about??? Yeah.   My sister.  &nbsp; If any of you have an older sibling you know what its like. I love her and all, but lately all she has been doing is just critisizing me to hell. Everything I own, everything I do, anything that relates to me in any way, even if I can\'t control it, she puts down. And I\'m getting really sick of it. It\'s gotten to the point where I just want to get away from her, go off to college so I won\'t have to see her anymore. So I won\'t have to deal with her shit.   I was discussing this a little with Randy *her bf* a few minutes ago, and she(and he)&nbsp;seems to be mad at the fact that my mom likes Matt better than Randy. Which is not true. My mom doesn\'t even know Matt. The only thing she has ever said to me about him is that he is attractive. She doesn\'t talk much about Randy, but I can understand what false impression she might get of him by my sisters constant complaining about things he does. But she doesn\'t hate him.  I\'m a lot like my mom was in high school, never having any boyfriends, but having lots of friends that are guys. *minus the small factor that the majority of my guy friends are gay* No that I have a ton of guy friends.... but nonetheless thats not the point. HER sister, on the other hand, constantly had a boy. I don\'t get boyfriends. Angela\'s had one for going on 4 years. I think that my mom just knows what its like to be alone, under the influence of a constantly attatched sibling. And she\'s happy that I have a boyfriend now. Its new and exciting and she\'s glad for me. Not that I ever talk to her about my relationship/Matt anyway, but thats irrelevant also. But she can relate to me, and I am so much like her its scary.  So I think Angela\'s jealous in a way. She wants mom to be excited for her and Randy like she is for me. But, they\'ve been together for so long that its nothing to be excited about anymore.   &nbsp; And she also is angry at the fact that I don\'t really \"follow the rules\" and I\'m not such a goody goody ms. perfect like she is. I lie to my parents, I go out till all hours of the night, I have a life and friends and will do anything to keep it that way. I don\'t get in trouble, and its not like I\'m doing anything illegal. But she, doesn\'t really have any friends. Not around here anyway. And she doesnt really have a life.   She told me how stupid she thought it was that we all cried at graduation.  I wanted to punch her in the face and be all \"Fuck you, thats just because you didnt have any friends in high school!!\" which she didn\'t. She had Randy, and Ian, and Putnam and Cam. And thats it. But I have so many and I love them all. She\'s just jealous, I know. High school sucked for her, and I\'m sorry, but my high school was awsome and I loved it. I\'m going to miss it a lot. I want to say something to Randy, and I probably will eventually, but I know it\'ll just get back to her. Because that stuff happens with them. So I have to be real careful what I say to him nowadays.   She doesn\'t accept me and accept my life. It pisses her off for some reason, and I\'m not trying to seek her approval on anything so I really don\'t care. But at the same time I don\'t want her to critisize everything in my life. I don\'t even ask her opinion and she throws it in my face. I don\'t think she realizes how much it affects me. I don\'t go around trashing everything she thinks and&nbsp;does. I keep my opinions to myself unless necessary.   She doesn\'t approve of Matt and our relationship, she doesn\'t like my clothes, or my choice of bedding for college, or the color of my room, or my music, or anything else. Can you possibly imagine why I don\'t talk to her about my life? I listen to all of her bitching and complaining and tales of her college life or drama with people, but I don\'t tell her a single thing about what is going on in my world. She knows the basics, yes, because some things you\'re going to learn when being sisters with someone. But she isn\'t even a part of my personal life. I try to tell her things and she just doesn\'t listen. So I gave up on that a long time ago. Thats what my friends are there for, to listen and to comfort me and to be there for me. Family is supposed to do that too, but my family pretty much sucks at being close and at all family like, so I have to overcompensate with my friends. I think I\'d kill myself if I didn\'t have you guys.   Versus, why I\'m going away to college and never coming home. *at least, not very often*&nbsp;I love my family and all but, they aren\'t really a family. They\'re like distant relatives that I happen to live with. I feel like I can\'t really fix things now. Its not worth the effort. My friends are my family. They are all I need to survive.   '),(14309,3976199,'29,July,2004','2004-07-29 00:00:00','   This just came to me: You know how your parents always make a big deal about the first word you eve said? Could a person\'s first word have any effect on the type of person they turn out to be? I wonder...    &nbsp;   &nbsp;  I know for a fact that my first word was \"shit.\" No joke.    &nbsp;  What was your first word? Let\'s compare!   '),(14310,3976199,'28,July,2004','2004-07-28 00:00:00','   Today I hung out with Jenna and had a really good time. We went to stalk Amy as usual, but as usual se wasn\'t home and didn\'t answer her cell. It\'s really starting to irritate me. I mean, she tells me to call her, so I do. The only problem is that she NEVER PICKS UP, and even when she is home she stil doesn\'t return my calls. It\'s called common courtesy people.    &nbsp;  Anyway, getting back to the whole Jenna thing- we went to Al\'s on a hunch that she\'d be working. She wasn\'t. That didn\'t stop us from eating 9 dollars worth of french fries and \"onion nuggets\" (SOOOOO GOOD!) Fortunately, I didn\'t eat all day just in case, so I didn\'t have any guilt ;^p After that, we went back to her house and watched  Bring It On.  OH! and we changed Jenna\'s message on her phone. It\'s awesome. Every single one of you MUST call her phone and hear it. I\'m not joking.    &nbsp;   &nbsp;  Well, I must be off. I have to go pick up the sushi I\'m having for dinner. Sushi=Joy.   *hugs*   '),(14311,3976199,'27,July,2004','2004-07-27 00:00:00','   Our new appliances just got here, and the fridge is too big. The salesman gave us the wrong measurements. My mom is   pissed!   This on top of the fact that my two litle cousins ae spending the week her driving her NUTS. I think she might spontaneously combust, so I\'m gonna run. Talk to you later! *hugs*   '),(14312,3976199,'26,July,2004','2004-07-26 00:00:00','   I just pulled my 4th all-nighter in a row&nbsp;this week (or last week. Whatever.)&nbsp; This time I became bored around like 4:45. So I got up, put my contacts in, and did what I&nbsp; pretty much always do when I\'m bored. I took a&nbsp;walk at 5 in the morning.    &nbsp;  It\'s been so cool in the morning. When I went out side it smelled like fall.&nbsp;I saw one tree already with turned leaves. This made me very afraid. Then&nbsp;I reasoned that it was WAY too early for me to be thinking about school, and kept on going. now that I think about it, I want school to start, and yet I don\'t. But I\'m sure you\'ve all been there before...    &nbsp;  I actually just got back from said walk. It\'s a little cool out for late July, but it\'s relatively tolerable. It was nice walking down 27 because everyone was so friendly, especially when you consider the ungodly hour that we were all out. The road was pretty much deserted with the exception of the occasional&nbsp;jerk going 60 ( MAN I hate that!) I ended up walking all over town. It was nice to just be able to walk down the quiet street, not worrying about unfortunate tourists asking for directions, traffic, etc. It just gave me time to think, as cliche as that sounds.    &nbsp;  In other news, my mom and I decided that in a week or two when my dad takes my brothers to Boy Scout Camp *shudder* we are going to go to New York. We\'re most likely going to take a train down there, and stay for a week or so. I\'m very  very    very   happy about this. *Does little jig around the room* OH! which reminds me- I think I\'m going to go with the vintage black and white, somewhat vogue-ish New York theme in my room. I can\'t wait!    &nbsp;  Now I shall leave you all, as my mom just asked if I\'d keep her company while she goes on her walk around town and such. Talk to you all later! *hugs!*   '),(14313,3976199,'25,July,2004','2004-07-25 00:00:00','   Italy\'s surprise birthday party ( that I helped plan) was today. Overall I\'d say it went well. I don\'t know, I guess I\'m not really that close to all of them anymore. Whatever. It happens...    &nbsp;  I really do NOT want to spend the night at home. Someone save me from this Hell. I Just found out that my aunt and uncle are going to be staying through tomorrow (&nbsp;I was pleasently surprised that my uncle\'s 4 kids didn\'t end up coming!)&nbsp;and maybe going to Water Country. I just don\'t know if I want to go.    &nbsp;  I\'m desperately craving caffeine right now. You know that&nbsp; it\'s-summer-and-I-just-spent-the-entire-day-swimming-in-the-sun tired?&nbsp;That\'s what I\'m feeling right now. It will wear off in like 30 minutes&nbsp;or so, but&nbsp;I really don\'t want to wait that long.&nbsp;    Oh! Remember how I asked you all your opinion on the room theme thing? Does anyone have any&nbsp;input whatsoever? I\'m kinda stuck b/c&nbsp; I like them&nbsp;both, so I\'m looking for what you all have to say.     &nbsp;  Another random note: Congratulations to Krystal and Amanda! You guys are so cute! *beams*    &nbsp;  I\'m off to think of something to do/someplace to go tonight. Give me a call if you already ahve an idea, hahaha. Talk to you later. *hugs*   '),(14314,3976199,'23,July,2004','2004-07-23 00:00:00','   I had this dream last night where I was watching myself look down on my own funeral. It was weird, because I was watching myself look at all the people there (&nbsp;which was odd, because I know there wouldn\'t be a lot of people in actuality) and I could see them crying, but I could tell that the version of myself that I was looking at didn\'t feel anything. Does that mean anything? Am I going to die alone and jaded? God, I hope not.  I\'m sure everyone thinks at some point how they\'d like to die. I know I do. I read somewhere of this 87 year-old who died while dancing. I think I\'d like that. Just going along, without a care in the world, then just letting go. I know for a fact that I don\'t want to drown, burn to death, or die slowly in any way. Like cancer. Cancer would suck.  &nbsp; Anyway, I just thought I \'d share this with you all in the hopes that I\'ll find out what it means sooner or later.   &nbsp; Talk to you later! *hugs!*   '),(14315,3976199,'22,July,2004','2004-07-22 00:00:00','   I\'m going to redo my room, but I\'m caught between two themes, so I thought I\'d ask you all. I was thinking of either doing an English gentleman\'s library (the type where the guy goes on safari and brings all these cool things back) or a vintage, jazzy,&nbsp;black and white NYC theme. Any preferences? Just to help you out, here\'s what&nbsp;I have in my room:    &nbsp;  2 Drexel dark mahogany sleigh beds,    &nbsp;  2 dressers, one with a mirror, one in a \"highboy\" fashion,    &nbsp;  1 desk, with all my assorted computer stuff on it,     &nbsp;  2 white&nbsp;bookcases (i can replace these),    &nbsp;  1 nightstand (matches the beds).    &nbsp;  &nbsp;and 1 mahogany drum table (this can leave if necessary)    &nbsp;  My carpet is a neutral, so that won\'t affect anything, and I have stained wodwork, which goes with both themes. All my accessories can be changed in a heartbeat, so don\'t let that stop you either. I hope to hear from you soon!   &nbsp;   &nbsp;   '),(14316,3976199,'22,July,2004','2004-07-22 00:00:00','   I\'ve determined that my Uncle Andy would be the PERFECT James Bond villain. He\'s practically a European, he loves fine food, he\'s a wine expert, he has EXTREMELY expensive taste in things like cars, houses and furniture, and he has this great social laugh/chuckle thing. Oh, and my cousin Alan (the one you\'re all going to meet) thinks so too. So there!    &nbsp;   &nbsp;  Muahahaha, I love my family....... *gag*   '),(14317,3976199,'22,July,2004','2004-07-22 00:00:00','   Leslie\'s pool party was so much fun! I had a really great time. Ricky and&nbsp;I sang the \"I\'ll fuck you up\" song (All State joke), and Leslie kept chucking hot dogs at everyone. After all the oudoor fun, we went inside and watched the most pathetic-yet-funny show I\'ve ever seen. HOORAY FOR POOL PARTIES!   '),(14318,3976199,'22,July,2004','2004-07-22 00:00:00','   This is Hero to Sexkitten.&nbsp; I repeat. Hero to Sexkitten. Do you copy, Sexkitten?!   '),(14319,3976199,'20,July,2004','2004-07-20 00:00:00','   I forgot to mention the job (!) that I got.&nbsp; my organist called me 2 days ago and asked if I was free on August 10th. I said yes. Then she goes on this whole thing about how apparently this really influential and extremely wealthy woman called her up and asked if she knew any really good singers, and she recommended me. Then this woman told her that she\'s having a remembrance ceremony for her brother, who at 47 committed suicide. So my organist called me to give me a heads up forthis lady. I had never heard er name before, but the organist (who goes by the name Mrs. Goss) who has lived here for 50 years said that she\'s loaded and extremely powerful in NH and Massachusetts, and a HUGE patron of the arts. So i\'m like  whoa  and wait by the phone. Sure enough, this woman calls and introduces herself and describes the situation that I had already learned from Mrs. Goss. Apparently she just wants someone to sing backgrond music while she has a reverend read a few psalms and whatnot and then follow her acros the road from her HUGE Rye mansion to the ocean while she scatters his ashes. She told me it would only be like 4 songs. Then she asked me, \" Your fee is $100, correct?\" She was indeed correct. ( Lemme give you a run down. The standard fee is 100, but the standard tip is 25) She said in reply, \"I\'m sure that could easily be arranged.\" (NOTE: Mrs. Goss told me after that the rich lady wold DOUBLE my pay if I did a good job!)    &nbsp;  So basically on August 10th I\'ll sing 4 5-minute songs and get paid at LEAST $100. Does that rock or what? Share in my joy, people!&nbsp;The best part is I already know exaclty what I\'m going to do with that money: put it in my checking account! After that I\'ll be all set for another SERIOUS day of shopping! Anyone interested?   &nbsp;  &nbsp;   '),(14320,3976199,'20,July,2004','2004-07-20 00:00:00','   Ohhh yeah. I\'m officially like, really cool.    &nbsp;   &nbsp;   &nbsp;  But not as cool as some people.    &nbsp;   &nbsp;  LIKE DONNIE FLYNN!   '),(14321,3976199,'20,July,2004','2004-07-20 00:00:00','   Today was fun. &nbsp;I hung out by a pool and got a&nbsp;nice tan (YES!)&nbsp; Then I came home and ended up going to the beach and seeing the fireworks with some friends of the family.&nbsp; It\'s rather amazing how many people we\'ve met since we moved here. (I make it sound like it was only yesterday, even though it\'s been over eight&nbsp;years. It\'s weird because I can still remember Clifton Park perfectly. What\'s wrong with me?)    &nbsp;  Anyway, getting back on topic, the fireworks were mediocre at best. Big surprise there. I actually was more entertained by the form of the smoke as the wind blew it inland. The one time I didn\'t have my phone, dammit.    &nbsp;  Oh, and has everybody seen the Bush/Kerry animation clip at  urlLink jibjab.com  ? I heard about it on the news, and went to go watch it. I found it to be quite funny. I mean, I personally strongly dislike both of them and have no draw to politics whatsoever, but I thought the two candidates pulling each other apart to be pretty humorous. If you haven\'t seen it, i suggest you check it out.    &nbsp;  &nbsp;I just remembered: Tomorrow is my friend\'s&nbsp;pool party. I\'m excited. Hope to see you all there! ;^p    (I\'m sorry my entries are so random.)   So yeah, I really can\'t think of anything to write about right now. My parents have  Top Gun  on TV in the room over the garage, and they just played the Righteous Brothers\' \"You\'ve Lost that lovin\' Feelin\'\" so now I\'m finding myself paying less and less attention to my typing and more and more to the song that is now continuing in my head. Curse this musicality of mine!    &nbsp;  I bid you all adieu. *hugs*   '),(14322,3976199,'20,July,2004','2004-07-20 00:00:00','   OK, I officially    &nbsp;   Goodnight to you all! *hugs*   '),(14323,3976199,'20,July,2004','2004-07-20 00:00:00','   Well hello there everybody! um... I don\'t really know what to type. This is my first blog on this site obviously, so I hope to update often.    &nbsp;  I guess I\'ll just talk about my day, because yes, I\'m just  that  cool.    &nbsp;  Today I decided to sleep in. I\'ve been getting up at 8 every day, and I guess I just didn\'t feel like it, so I ended up waking up at 11:30. my mom was just about ready to order out Chinese for lunch, and she asked me if&nbsp;I wanted anything before she left. I was like,  Whoa, she\'s actually ordering out!  and quickly put in my order. I wasn\'t very hungry, as I had just woken up, so I snacked on it throughout the day. Chinese food is my life, except for the fact that 2 hours later you\'re ready to eat again, hahaha.     ANYWAY, then I went to the oral surgeon\'s office to schedule my next procedure (August 12th, people) So I\'ll be out of commission for a couple days&nbsp;after it\'s all said and done. I\'m really not looking forward to it, but whatever. The guy was actually pretty cool, and he kept saying how with the drugs I won\'t remember anything, which actually after a while got a bit unsettling. I was glad that he wasn\'t all quiet, boring,antisocial, and dry as normal doctors normally tend to be.    &nbsp;  So when I got back I hung out with Krystal and ran errands for my mom downtown. On the way back we got caught in a freak rainstorm. It felt really good to just not care about staying dry or anything. That\'s one thing I love about summer. The rain actually turned out to be a thunderstorm, so Krystal and I were even happier, because we both absolutely love thunderstorms.    &nbsp;  Now I\'m sitting here, realizing that I\'ve run out of things to say. Peut etre it\'s a good time to stop typing. Peut etre...    &nbsp;  Oh well, I\'ll guess I\'ll talk to you later guys! *hugs*   '),(14324,3976199,'17,August,2004','2004-08-17 00:00:00','   All of a sudden I feel extremely dizzy and thoroughly exhausted. Just thought you\'d like to know. ( All I do is GIVE!)   '),(14325,3976199,'16,August,2004','2004-08-16 00:00:00','   Swelling is bad. Vicodin=good. Thank God I\'m actually feeling normal again!      I miss you all. We have to hang out before I go to Florida and then have to come back and do all my summer hw!      *hugs!*   '),(14326,3976199,'11,August,2004','2004-08-11 00:00:00','   I took my brothers to Marston to hang out on the playground today, and on the qway back they asked to stop at Marelli\'s. While I waited for them, who should walk up to me but Angela DiMillo! Holy George Washington Christ, I haven\'t seen her in 2 years! We talked, uch to the annoyance of my brothers, for a good solid hour, and she told me that we have to hang out this week, she\'s supposed to call me this afternoon. HOORAY FOR SINGING PARTNER REUNIONS!      ...That is all...   '),(14327,3976199,'09,August,2004','2004-08-09 00:00:00','   Thursday =  ewwwwwwwww. Call me crazy, but I don\'t think knives go very well on me at all.   '),(14328,3976199,'08,August,2004','2004-08-08 00:00:00','   *Happiness* I\'m going to Florida!  That\'s all for now. Talk to you later! *hugs*   '),(14329,3976199,'05,August,2004','2004-08-05 00:00:00','   I\'m back!  OMG Clifton Park and Fall River was SOOOOOOO much fun! I got soooooooooo many new clothes! HOORAY FOR HOT CLOTHES! *ahem*  I have to go pick up  my brothers today. Grrrrrrrr. I wish i could leave them up in Manchester in the woods. Oh well...   *GASP* I found the most incredible house EVER in Clifton Park. I want it. Put it in a box right now and send it to me. This thing is HUGE!!!!! I have a picture on my phone. WANT WANT WANT! I saw it when Jeanette and Dean showed us all the new neighborhods tha are springing up in town. These things are HUGE! Each neighborhood is the size of Portsmouth. It\'s crazy. I mean, what the Hell is there in upstate New York that draws people!? The world is a screwy place...   So yeah, I\'m off to Manchester. I haven\'t seen my brothers in a week, and I\'ve really gotten used to peace and quiet. Pray for me ( if you do that sort of thing...)   Talk to you later! *HUGS!*   '),(14330,3976199,'01,August,2004','2004-08-01 00:00:00','   Home state-ward bound! I\'ll be back in a few days. Heads up to Clifton Park! Donnie\'s back! ;^p         I Love being a New Yorker :)   Talk to you when I get back! *hugs*   '),(14331,3976199,'23,July,2004','2004-07-23 00:00:00','    So.   Today.   Hm.   Well....   Was awoken at 9:00 sharp. Told I had a half hour to get up and get ready. Ready you may ask? Well, your answer will have to wait. I\'m not there yet. So. 9:05 rolls around. I\'m too cozy. 9:15. I\'ll only take a few minutes to get ready. 9:18. There\'s still plenty of time. 9:20. Fuck, I really should get up. So quick as a bunny on cocaine, I groggily (got to love the&nbsp;disagreeing adjectives!) made my way to get dressed. Down stairs I go, expecting to be hauled out to my dads \"workshop\" [corner of a storage space facility] to lift heavy things, I realize he\'s left. (by the way, that was what I had to get ready for.. if your aforementioned question was still anxiously awaiting an answer) Needless to say, I wasted very little time going back to my air-conditioned room to read, happily.   10:30 rolls around. NOW it\'s time to go haul around heavy things. Ugh.   Blahblahblah   We go down toll farm rd (Is that the right road?) make it to the storage space and I am put to work. And by put to work, I obviously entail standing around while my dad mumbles to himself trying to figure out what needs to get done first, or rather, what I\'m capable of helping with). After much standing around, in a moderately cool warehouse might I add, I\'m handed an electric drill. Hah! Drilling!    Through much painstaking aggravation I succeeded in taking apart a shoddy workbench/table/ thing  my dad had installed. Aggravation? Yes. Who installs screw SIDEWAYS into the least accessible corner of workbench?! ...   Anyways. After drilling for a period of time... the bit fell out. For no reason. So I had to struggle, with what little drill knowledge I have, in reinserting and maneuvering the drill bit into workable condition. I am not meant to handle wood and tools.    LATER    After a trip or two back and forth from the house and warehouse (Ooo! I forgot to mention! I got to dust off lights! ... Carbon dust is disgusting to deal with on a humid day.) my dad decided to rig up his big-expensive-what-the-hell-does-it-actually-do-?! machine with many red straps. Once done, he decides to borrow the fork lift at the warehouse. With a lot of vrooot-beep-beeep-beep-beep-rrrrraaaaa-vrrroot-vroot-vrooot he maneuvered over to the machine... followed by more fork-lift noises and we had it to the infamous  loading dock!  That\'s when things got worrisome. Putting the straps to the test. [Ahem. Because the dock is so high, and my dads truck rather far below it, it was only logical that he lift the several thousand pound machine up by the little red straps and place it in the truck.] Aside from the straps stretching about 2 and a half feet when trying to lift the machine of doom, nothing exciting happened. *sigh*   After that there was just much more lifting, grunting, pulling, shoving, strapping and unstrapping to be had once we got the thing back home... with no forklift to help anymore.   Then I spent my day looking at stuff in JoAnn Fabrics. [I\'ve begun constructing a mahvelous orange corduroy bag. i\'ve got a new thing for  orange.  It\'s weird.] So there\'s that. My mom and I proceeded over to WalMart thereafter and found a guy in the most (intriguingly) revolting shirt ever. Perhaps some of you have seen Chris Giorgio\'s highlighter yellow shirt? This was highlighter orange... It must have had battery packs to shine as brightly as it did. GOD. It was a sight to see, I suppose. And what made it worse (two things, really) was the fact that the man wearing it was in his 60\'s, and the shirt had those tacky fake-wood buttons on it. Not good combinations, let me tell you.    &nbsp;  People need to be less uptight about porn and sex.   My labelmaker is the man. Anyone want labels?    Playing spit over a low coffee table results in poor back conditions.   Sprite Remix: BerryClear is the essence of skittles.   I want to go to Boston. I want to visit Kristina.   I want to go to AZ (sooner). I want to visit Chad (longer).   My travel toothbrush..was dusty.. even though it was in a plastic case.   Next week will be boring.   I am going to buy presents. For a certain someone. (Wink) So secretive!   I\'m liking purple less and less by the day.   I need  real  chocolate. Not pansy commercial franchise chocolate.   I am  so  demanding.       &nbsp;   '),(14332,3976199,'22,July,2004','2004-07-22 00:00:00','    Here it is...almost 3 in the morning...my time.&nbsp; And I\'m awake...almost six your time. Le sigh I have found my new love in Sweeney Todd.&nbsp; I need to get more than just the highlights cd.&nbsp;   &nbsp; ****************************************************** Now it\'s 3:02...I don\'t know why I\'m still up.  I truly believe that everybody can make there dreams come true.&nbsp; Just never give up.  \"One day I\'ll fly away, fly, fly away...\" Sorry that was just on and I thought it fit.&nbsp; I realized just how trusting I am with other people.&nbsp; I never believe that they will hurt me, mostly because I haven\'t really been hurt all that badly by people in the past.&nbsp; That and I am optimistic in nature...ooow...suddenly I am tired.&nbsp; That came on pretty fast.&nbsp; jeeze..ugh I don\'t think I am going to be able to correct my typung errors any more.. I \'m nust hoing to type ant type...withour corrections and well see whereit goites.&nbsp; ught My comp isaoijan asshole..I can\'t stand it it keeps like not letting me tyupe ot simthing...oow, I\'m going to try tying wtih me eyes closed...almost asleep. its a good thing I\'m listening to music. ptjerwose O I would fall aslwwp.&nbsp; hmm, what to weite?I ac tually likke these little indent things that help you keep your fingers on the\"home row\" ewww, typing vass, what was it?computers and a waste of time?&nbsp; that class was woethless I really didnt learn anu thing.&nbsp; I justthought I&nbsp; would share that.&nbsp; O\'m not really a very fast typer. I tealiced that just now...no I lied I alwasy knew that.&nbsp; Now lets see what happens when Chad types with his etes closed amd fpes veru fast. ready?&nbsp; Buckle up...take some aspirin and get ready to decipher.  I had forgoot en that I had added austin Powers stiff to my ipod it just soetr oc came up on me and I awwas all. .aakjj Ot a[...mpow tO tjomd P amd going to change the spnmgsong chanfed to...sometin I fon\'t remembe...I thinkyep its Rahmstein,,,fun!&nbsp; Late night angey Gernan miustic uippe I think ebneyt boud should listin to this ebey niow and then, ushf this tis more fiddicyltk thatn I would habe realiced I\'m sure your all amaced ast how horrible I am at typigbn buit then againg I am apmos tfalling asleep ,&nbsp; Wouldn\'t it be just so amacing if when I fell asleep I kept typing aahfand typed some pergect book.&nbsp; lke the acount of how the world was avtually ctrayed and teh I posted it and was all cool.&nbsp; That\'s kinds how the urantia book was created.&nbsp; Dome giu fell asleep and was talking in hid slwwp.&nbsp; Finally somebodu fecided to copy doew&nbsp; waht he was saying and it turned out it was an entire book.&nbsp; LIke perfectky orgabuzes and evertgugbf///altight I\'m goint to fo now... goodnight all.&nbsp; I love yoiQ&nbsp;&nbsp;   Live and love   '),(14333,3976199,'22,July,2004','2004-07-22 00:00:00','    Bon jour, mes amis.&nbsp;   Finalmente yo puedo...postar cosas en este blog.&nbsp; (I\'m assuming the noun blog is masculine) Por eso no necesitan Uds. preocuparse porque yo estoy aqui.&nbsp; Yay!!  Okay, nough Spanish.&nbsp; (I know it\'s often&nbsp;\"nuff\" but it really should be nough, seeing as how the \"word\" is \"enough\" minus the \"e.\")  &nbsp;Does anybody else find it kinda rediculous when guys sing in falcetto?&nbsp; Like...on cd\'s and stuff.&nbsp;   K...I\'m gone 4 now!&nbsp;   &nbsp;  Live well and love lots!   '),(14334,3976199,'22,July,2004','2004-07-22 00:00:00','    I say no, nay, never! *clap-clap-clap-clap* No, nay, never, no more! *clap-clap* Shall I play wild rover! *clap* No nay, never, no more!    &nbsp;  Medieval Manor last night was wonderful.   The peasant man, Squint, was attractive. He reminded me of Jay Buckley. Coincidentally, Maggie turns to me and says \"He looks like your type!\".... Moments later \"Who\'s my type?? I have ears like a hawk!\", followed by me laughing and blushing profusely.   Twas funnier this go around. The crowd got more into it.    &nbsp;  A six course meal is just too much, really.    &nbsp;  This time I got a doggie bag to go. The bag had dogs on it, and a printed song. Haha.   Let\'s all go there when we\'re 21!     &nbsp;  I\'ll fix this up later, I\'m out to look for fabric and patterns for a skirt... G\'day to you all.   '),(14335,3976199,'21,July,2004','2004-07-21 00:00:00','    Oh be-still my beating heart!   He  called me  by phone at 11:30!    He never calls!     EVER!    My life is complete now.    &nbsp;  And why is it that the later hours of the evening bring out&nbsp;energy?    &nbsp;   Tonight at dinner:    &nbsp;&nbsp; Dad: \"Remember when I bought you that bird wing?\"   Myself: [over a mouthful of corn] \"Mmph.\"   &nbsp;&nbsp; Dad: \"Do you know where that is?\"   &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; [upon leaving the table I saunter up to my room, find the owl wing, and place it on my dad\'s desk]   &nbsp;&nbsp;&nbsp;Dad: \"I walked into my office and thought there was a dead rat on my desk!\"   Myself: [smirk]    &nbsp;  Alright.   It was amusing for  me .    &nbsp;  I was just going to type out an irish-style song&nbsp;I&nbsp;made up&nbsp;in my mind about being perky and happy... but then realized you\'d never get the jist.    &nbsp;  I have a  very  large bug trying to fly around my desk lamp. In what manner will I take care of this matter? Get my dad to swat it to death. Huzzah!   '),(14336,3976199,'20,July,2004','2004-07-20 00:00:00','    Another tedious day.   Type, pause, read, repeat.    &nbsp;  Quite simply this has been the wettest summer to my memory. It really is very depressing, and I\'m not even a summer person.    &nbsp;  I feel so out of sorts. I stayed up until 3 o\'clock last night. ( When was the last time  that  happened? ) Then, to my utter suprise and dismay, awoke at 10:30. My body refuses to let me sleep regardless of my mind repeating over and over \"There\'s nothing to do in this house!\". So I continue to alternate between transcribing, reading, and staring at my pleasant beeswax candle.    &nbsp;  I\'m beginning to dislike how boredom brings greater expectations for a new day. Then that day arrives, and somehow, hopes aren\'t accounted for, and your mood is left indistinguishable from the pensive thoughts of the day before.    &nbsp;  So my hopes are sitting in the back of my mind, creeping up whenever I stop to sit and wonder what there is to do next. Time works in odd ways.. sometimes too lethargic, sometimes too hasty.&nbsp; Too eager for the future and dreading the shadowy implications of it all, while in the meantime I sit and write entries about the present being unsatisfactory. Impossible to please-- Yes, yes I am.  &nbsp;   '),(14337,3976199,'19,July,2004','2004-07-19 00:00:00','    Another uneventful day. How grand! What better way to celebrate the passion of boredom than to create a blog? You tell me.    &nbsp;  I\'ve sat in this chair for the better part of the day, aside from a brief and fleeting trip to the supermarket to buy light cream ( How can one live without the proper accoutrements for coffee? ) and a bottle of lotion ( Burt\'s Bees; Milk and Honey. Absolutely divine ).     &nbsp;  Also, I cleaned my room. And by clean, I mean straightened up. Things shifted from one shelf to a drawer, and from desk to box. Orderly looking on the outside, but really only hidden piles of junk. Not to say that all my junk is stowed away, but at least now it\'s... slightly better.    &nbsp;  Tomorrow looks to be as interesting as today.    &nbsp;  I would highly recommend the book  The Enchantments of Flesh and Spirit  by Storm Constantine, but I\'m only a short way into it. So far it proves to be excellent reading, though.    &nbsp;  I wish I had more to say for a first post.. Oh well, I suppose.    &nbsp;  Thank you and goodnight.   '),(14338,3976199,'29,July,2004','2004-07-29 00:00:00','    This just came to me: You know how your parents always make a big deal about the first word you ever said? Could a person\'s first word have any effect on the type of person they turn out to be? I wonder...  &nbsp; I know for a fact that my first word was \"shit.\" No joke.  What was your first word? Let\'s compare!   '),(14339,3976199,'28,July,2004','2004-07-28 00:00:00','    Billy Joel\'s \"Zanzibar\" is the perfect song to get one motivated on a cool July day. That and looking forward to New York ;^p    &nbsp;   &nbsp;   The inning isn\'t over yet for me...       &nbsp;  Talk to you all later! *hugs*   '),(14340,3976199,'26,July,2004','2004-07-26 00:00:00','    Why do I constantly feel let down?   Are my expectations too high?   Or have you never been as perfect as my mind portrayed you?    &nbsp;  The streets are uneven and my eyes won\'t raise.    &nbsp;  Why are the meanings to actions never interpreted correctly?    &nbsp;  Why don\'t questions have answers?    &nbsp;  One of those days... (weeks? months?) where nothing is sufficient and everything only seems to get worse.   Every day spent leaning towards the next.   Is one solid hope worth more than a carload of certainties?   Tonight I\'d beg to differ.    &nbsp;  Nothing is sustainable    I\'m only delaying inevitability.    &nbsp;  Why is it we only miss the past?   [why can\'t heartfelt kisses make another person open their eyes]   I wish I could move my eyes elsewhere   '),(14341,3976199,'12,August,2004','2004-08-12 00:00:00','    Well kids, let you be the first to hear,      Jordan and I are no longer a couple. Again.         Being worthless isn\'t so bad, right?   '),(14342,3976199,'10,August,2004','2004-08-10 00:00:00','    I was pleasently surprised to find out that the service that I had to sing at today was only 30 minutes, but I still got paid 100. This has added to my happiness level. YES!      Anyway, it\'s been a while since I last typed something here. Just letting you all know I\'m alive.  I know you\'re all disappointed, sorry ;^p.      Psyched for Florida!         Anyways, I\'ve got to go. I\'ll talk to you all later.   '),(14343,3976199,'17,August,2004','2004-08-17 00:00:00','    Aren\'t we just the talkative group.         Schedule has been flip flopped.. changed around.. I finally have to start worrying about college.      It\'s funny how things catch you off gaurd. The things you\'ve expected the most. I wish this were as easy for me.      And of course.. Senior Pictures...       I wish this were easier.    But then again, I wish for a lot of things.      If only I weren\'t just being shrugged off...   '),(14344,4115894,'01,July,2004','2004-07-01 00:00:00','   I\'m so tired of this shit.     I thought you were different,     thought you were it.     I didnt realize you didnt \'get\' me too.     Just like everyone else,     You\'re making me puke     up my insides all over the floor.     I\'m saying it right now.     I want you no more. '),(14345,4115894,'18,August,2004','2004-08-18 00:00:00','   I dropped to my knees silently alone in the dark room, no one there to face except myself and.....him... Although my mind was scatter-brained my concentration remained. Slowly as I stared blankly into the reflection in the water.. I reached out to touch the image that I had built with my mind. That I had worked so hard for, Something I knew I could find. An image in the water that I knew I would not see again until my 18th year. Although I had been here for 16 years already, I had only been born at fifteen. My teeth pressed down against the corner of my bottom lip softly. I had been concentrating so hard that the blood that dripped from my lip went unnoticed into the black sea that lay before me. Straight onto the image of his nose, my favorite feature. The ripples were his demise. I felt my heart jump into my throat, as I was crouching over the pool of black water, gazing into his eyes.. I hadn\'t noticed that I was gone, lost somewhere. My rude awakening had come abruptly. Just like the last time we had to part. Rude and abrupt. I felt the same pain in my stomach that I had before.. As if the whole downward spiral of our tragic love had been played out again. I gasped as his image faded from the water. Gasping for air. I could not breathe again, just like before. My knees weakened and I could no longer sit on them. So I slid onto my butt slowly, reaching out to touch the cold darkness inside the water, maybe hoping that his hand would come out and touch mine. I gripped at nothing, cold wetness. I had to feel again.. I stumbled to my altar, picking up an old friend. I knew it as I sliced the wrong vein. This was the end.'),(14346,4115894,'17,August,2004','2004-08-17 00:00:00','   Dear you know who-  Today is your birthday.. I wish that I could call and say. How much you mean to me. I wish that I could put into words how just your voice alone weakens my knees. But I am limited to this note, and a five minute call. Im yearning to hear those Three Dots. I\'m begging you mentally not to stall. I want to reach through the line and touch you. Caress your whiskered face. But the phone is the only thing there that i have left to graze. Oh and I forgot to tell you. There are some things you left behind.. The scent of your cologne on your pillowcase The memories in my head Little fragments of your heart and that lingering thought of the way you used to look right through me when you said ...dot dot dot... '),(14347,4115894,'15,August,2004','2004-08-15 00:00:00','   Hateful.      Why are you so FUCKING hateful?     Because I Hate you.     I hate what you\'ve done.     I hate what you\'ve abused.     I hate your coping skill.     I hate your substance of choice.     I hate the way you guzzle it down     Just to drown out my screaming voice.     Do you know what its like to have every tavern\'s phone number memorized when your five years old?      Its just another story that shouldn\'t have to be told.     I didn\'t ask to be born.     I wish they would\'ve given me that choice.     Instead of having to sit here crying tears of scorn.     Listening to your slurring voice.     You disgust me.     Just like Scarlett.     So leave me be.     And take with you your precious Harlett.     I never needed either of you.     You never taught me shit.     I wish you would both get a clue.     Instead of sitting here, making me watch over you.     Just so that you dont pass out and die.     But never did you appreciate     The fact that 80% of the tears I cry.     Are from worry.     The anxiety that troubles me so.     Is actually from the thought of watching you both go.     I am a walking contradiction.     When in fact I am pure human.     Always wanting you to quit your addiction.     The only thing that you love.     So cling to that bottle.      And wonder why...     You have no friends.     Or true love in which to confide.     But I will always be here, codependant little me.      Always here, to take away your keys.'),(14348,4115894,'13,August,2004','2004-08-13 00:00:00','   You Have no Idea     How much I Yearn     to watch you Die.          You make my insides turn.     Like a bad taste in my mouth     that refuses to subside.          But why is it..     that everytime I try to hide     You come seek me out     and apologize.          But everytime im crying     You\'re too busy to care.     Lost in your precious television     and how could I dare     inturrupt as my spirit is dying.          And everytime I fall for your act     Your so great at making me believe     You mean it this time.          But I hate your inner soul     And I will dance on your grave.     And everytime, I think you\'ve changed.     But tragically still your hostility remains. '),(14349,4115894,'11,August,2004','2004-08-11 00:00:00','   It\'s Like Air.     I dont want him, I need him.     It\'s like air.     And it\'s like I\'m watching a movie of me drowning..     But I\'m helpless to do anything about it.     It\'s like air.     What do you do without oxygen?     You suffocate.     That is the worst way to die..     By heartbreak.     I\'m falling, But, God, I just want to hit..     The bottom of this never-ending pit.     I\'m reaching out for something to grab onto.     But all that is there     is this cold wall of authority     that seperates.     and kills.     and suffocates me.'),(14350,4115894,'09,August,2004','2004-08-09 00:00:00','   Blue Red and Yellow Pills drop With each one my heart flips And my stomach flops I love you Doctor Forever and Always You are the perfect man You leave me staring at the screen in a daze. With your cute little green hat And the way You look at me and say.. Your lips parting ever so gently, as your voice gets deeper, so much lower. Whispering into my ear Game Over.'),(14351,4115894,'03,August,2004','2004-08-03 00:00:00','   Is that all I am?   A fucking number on your phone?   Push a button and forget me.      You said you were going to be the one who cared.   You always said you\'d be there.      Hang me up and tell me my feelings matter.      Then tear me down passively, never to my face.      You never COULD stand to see me frown.      But that frown is the only friend I have left to chase..      -All the rest of the people in this world away like you.      I dont want to attract any more of them just so that they can pull me down      Just like you.      So as I frown...      Go on and push that button on your phone.      While I sit here,   Clutching my blade and my sanity.   Left all alone. '),(14352,4115894,'02,August,2004','2004-08-02 00:00:00','   Hate to Interrupt your perfect life.     Just so you can listen to me cry     -a few times     while you roll your eyes.          I know thats what your doing     locked yourself into Euphoria     what a horrid place to be     Let\'s all feel pity on you     But, God, Never me.          I want to slam your face into the cement          But, fuck, its not fair     that i get accused and you get consequences.          Yeah, I slit my wrists for you.     bled for you     cried for you     died inside for you          You leave me with a movie,     But i\'ll leave you with something you\'ll never forget.     A broken heart and a burnt out cigarette. '),(14353,4115894,'02,August,2004','2004-08-02 00:00:00','   I\'ve been holding my breath for too long.     Drowning in this bittersweet nightmare that is you.     Everyday I wake reluctantly     Just to be forced to look through these eyes.          Once again, like everyday     Forced to hear their lies.     \"Everything will be okay\"     one day...     they say.     But I dont want it to be okay.     Unless it means having him back.     right now.     today.          Nothing will ever mend, truthfully.     Time does not heal, only allows you to make it part of your tedious routines.          So here I go,     turning blue.     You know I would only do this for you... '),(14354,4115894,'02,August,2004','2004-08-02 00:00:00','   I\'ll be your Catalyst.   I\'ll be the trigger on your gun   that makes you fly   like a bullet   into all of their eyes.   to make them see what they want to block out   They cant have me.   They wont take me without a fight.   So keep my heart.   Hold it tight.   Or just throw me away.   whatever you like.    In a world of Prevention...    I\'ll be your catalyst   I\'ll make things happen   things that they dont want.   things that you dont have the guts to say.    But I\'ll step up   and I\'ll play.   I\'ll be your linebacker in this risky game.   The only thing that matters now is that you\'re okay.   But i swear on everything that I have.   I\'ll be with you someday. '),(14355,4115894,'02,August,2004','2004-08-02 00:00:00','   Falling Deeper Into this Act     Hiding behind my Beautiful Mask     I\'m Beginning to Think that This Monster I\'ve Created is now me.     I\'ve barricaded myself in so tightly.     No one will ever get in,     But that is how I want it.     Not because I\'m afraid of you.     But, because im afraid of the things that I might do.     I would kill myself if I ruined it now.     So that\'s why I\'m posting my sign:     No Boys Allowed     Because what I gave to you,     will never again be mine. '),(14356,4240336,'21,August,2004','2004-08-21 00:00:00','   Tomorrow\'s the Nightwish show.  And after throwing the idea around of maybe meeting the band, in the end it\'s going back to the original plan of just seeing the show.  We\'re probably gonna get their around 3, and we\'ll end up meeting Rod and all his friends.  Which reminds me, I must tell Niki when the train is.  I\'m becoming  very  paranoid over when this whole thing will end.  The last train is 12:37.  It can\'t end a  second  later than 12 midnight or we\'re fucked.  I just realized how else we\'ll be fucked since on the Nightwish forum there\'s talk on how the show is probably oversold.  It only holds about 1000.  If we get caught in the middle  A) that would suck lol   and  b) we\'d be stuck.  I need to talk to Niki.  She\'s not online right now though.  I hope she feels well.           Anyway after that on Monday we leave for Massachusettes (hence why I MUST get home).  So I won\'t be updating for awhile.  I get a break from using the internet =D           Now off to find Niki.  We really need to plan this better :/   '),(14357,4240336,'19,August,2004','2004-08-19 00:00:00','   I know, I\'m getting  very  ahead of myself when talking about Halloween, but Katy got my mind thinking.  She said something about us all hanging out for Halloween.  Then someone mentioned something about dressing up.  And that\'s when I decided, I shall pull the Steve Harris thing =P  Plus, I just realized, AG said he\'s trying to come up in October.  If I remember correctly, he said around the last week in October.  So maybe he can do the metal thing with me??? I think I can pull off the Steve Harris thing :p  It\'s the hair - I can do it with my hair.  I just need a jacket (lether one possibly) and find cloths close to what he is very notable for whereing like from a tour or something.  Then I need something that looks like a bass guitar and a strap close to his.  Hm....looks like I got some planning to do with that strap :/  And some sewing =D   '),(14358,4240336,'18,August,2004','2004-08-18 00:00:00','   I\'ve just gotten disappointed one too many times.  AG knows it...he\'s the only one that understands.  I wish he still lived up here.  Then I\'d feel better.  Sucks that he\'s out there.  I may be forced to move out there now.  I just can\'t wait to leave here now.  I really can\'t.  I just want to  leave.   Never come back.  AG understands me, and Tennessee is a Republican state.  I should really live there.  I hate it here.   '),(14359,4240336,'16,August,2004','2004-08-16 00:00:00','   I met Laura from the IMBB today =) She\'s so nice!! I\'m so glad I got to see her! We were talking about the Rush concert and a bit about the Dance of Death tour. Then we started talking about the Crew and how she wants to basically get me and Bryan in to chill with them all at Twins Pub before Nicko\'s show. That would be great. She said she doesn\'t see a problem in them letting us in, as long as we don\'t drink. She figured this since the show is all ages for one, and one of the other guys on the BB, LordIffy I think she said it was, is actually bringing his kids to the show so he\'s gonna bring them to the Pre-show get together at Twins. I hope Bryan doesn\'t mind hanging out with them at Twins. If he doesn\'t want to, then we\'ll hang out together longer, and then we\'ll just wait for them all at about 8-ish outside of Twins and we can all walk to Time Sq together. This is gonna be so great!!! Laura\'s so amazingly awesome and I can tell the rest of the Crew is too =) I feel so special since they don\'t mind that we\'re are 17 and don\'t mind hanging out with us. They\'re all so great and loving.  I love Iron Maiden fans [heart]   '),(14360,4240336,'16,August,2004','2004-08-16 00:00:00','   Sorry about not updating yesterday.  Yesterday I worked from 10-6 and then I went over to Sara\'s house for her birthday.  I didn\'t get home till about 11PM.  Sara got her license though, I\'m very happy for her =)  For her birthday, we went to Red Heads, then we went to play pool.  Lauren and I lost the first game, but we won the second game and I got the shot =)  Whooo I felt special.  Wish I had a pool table to play on -_-              Other than that, I feel kinda...like I\'m fading.  Just fading from people.  I was thinking this earlier but then I realized that I didn\'t update from yesterday -doh- so I decided to put these both in one entry.  I feel like I\'m still close to  my friends, but on the other hand, not as important to them anymore.  Maybe I\'m not -_-  I\'m disappointed in myself for letting it get like that though -_-         And I also think I gained weight again this week.  Some of my cloths didn\'t really fit yesterday.  I\'m disappointed in myself for that two.  If only my fat would fade away like how I feel I\'m fading away from some of my friends, that would be great =D  Should just switch the situations.  -sigh-         On a happy note, I think Laura from the IMBB is gonna come and visit me today =D  We\'ve been planning this all summer and I think it\'s finally gonna happen today.  I can\'t wait to finally meet her.  Then I get to see her in Sept at the McBrain Damage show.  Yay.              I guess that\'s all.  Oh, wait....lemme say one thing:      About the friend thing, I\'m not  trying  to drift from any of you.  I just feel as if I\'m not really there with you anymore :/  I feel left out, or just not wanted if I\'m talking to you.  I\'m not saying I don\'t want to be friends with you, because I do.   '),(14361,4240336,'16,August,2004','2004-08-16 00:00:00','   As I\'m sitting here having a bit of an artist block (or just pressured to use my extra paint) I think I should type up what happened today at work.           I was at work from 10-6.  Maren was coming in at 5 so I could get off the register for my last hour and do the put-backs.  At about 430PM I was alone at the register while Heather was in the foyer.  She was looking out the front door while trying to put up a sign.  She turns around and goes to me, \"there\'s a clown in a wheelchair in the parking lot\".  I just looked at her with that \"what are you talking about\" look upon my face.  I\'m thinking it\'s one of those things when there\'s business and they want to promote something, they have someone dress up in an animal suit or something with the promotion.  It wasn\'t.  Here\'s how our conversation went:      Heather: There\'s a clown in a wheelchair in the parking lot   Me (I just have that look on my face): What?   Heather: There\'s a clown, in a wheelchair, in the parking lot      (At this point I leave the register (like there\'s anyone there anyways) and I go look out the front door).      Sure enough, there was a man (fat man might I add) in a yellow clown suit, in an electric wheelchair coming toward the store.  I wish I thought of taking my phone out earlier than I did (I have a camera phone) because when I thought of it, he was getting a little too close to the door and I didn\'t want him to see me take a picture of him.  Now I really regret not doing it.      Heather and I then run back to my register, and Heather tells Robyn to dial my register.  By this time, the clown was in the foyer of the store.  We can\'t hold back the laugher, we start laughing.  I never saw Heather laugh so hard in my life till this very moment.      Robyn calls my register.      Heather then tries to say, \"there\'s a clown in a wheelchair in the store\" but fails to do it due to laughter.  You can tell Robyn asked, \"what?\" through the phone because Heather tried to repeat it.  She was laughing  so  hard that she actually passed the phone to me and went \"do you wanna tell her?\"  I was laughing so hard myself that I couldn\'t so I just shook my head.  Finally Heather was able to gather herself so she could tell Robyn.        Later Robyn comes up to the register (and by now the clown is just in the store somewhere) so Heather could ring her up.  I could tell Heather was trying to talk about the clown.  And now at this time, I had customers.  I don\'t know why...but when I heard Heather try to explain it again to Robyn, I started laughing.  I started laughing the second I saw Robyn come toward the registers because I knew Heather would tell her.  I\'m sure the two customers on my line probably thought I was crazy.  Since I was laughing at nothing, through their eyes.           People, you don\'t see this everyday.  You don\'t see a clown, in an electric wheelchair, go through the parking lot and come into your store.  You might have seen weird things...but in my 4 months of being at the Rag Shop, I never saw such a thing.  I\'m still getting a kick outta it.  I\'m still amused and I\'m still laughing at the thought.  It was definitely funny.  Maren laughed too when I asked if she saw him when she came in.  I didn\'t even have to  finish  my sentence.  She did for me.  It was great.  Him coming up through the parking lot to the front door is one of those things that are scarred in ones memory for life.            No, I wasn\'t on drugs, nor am I on them right at this moment.  This actually happened.   '),(14362,4240336,'15,August,2004','2004-08-15 00:00:00','   I\'m starting to have that feeling where I don\'t belong again. Prolly the age thing. Or probably because of the person that I am, I really  don\'t  belong anywheres. Maybe that\'s just me though. I\'m sure people will fight with me with this feeling. They\'d just be like \"that\'s how you\'re supposed to feel at 17\"and all that bullshit. That\'s all my mom ever gives me. Like that day she decided to  lecture  me on life and crap. And how I have to do better in school. And then she went, \"life isn\'t easy\" so I just went \"Whatever. I hate life then\". I think I really do. I find no purpose. It\'s to feel like this till you\'re like...21 or 22, then probably feel like this again once you\'re outta college and can\'t find what to do with your life because you\'re life just failed you and you can\'t find a career that you\'ve been looking for for years. So life really isn\'t that grand. I hate it already.   '),(14363,4240336,'15,August,2004','2004-08-15 00:00:00','   You know what, fuck it.  It\'s not worth it.  Nothing is anymore.  Right now, I just wanna talk to either Lauren, or Susan.  The only two people who\'d fucking understand.   '),(14364,4240336,'15,August,2004','2004-08-15 00:00:00','   Today was a lazy day really. Kinda. I had to go food shopping, which wasn\'t that bad since I bought a lot of shit (never go food shopping when you\'re hungry =x), then my dad took me to CD World. I really should have driven myself....all well. I hate driving. Anyways - I got 2 cds. I got Rush - Vapor Trails and I got Symphony X - V. I like the Symphony X so far =) Then I came home and I ordered some shirts. I got a Symphony X one (it has the masks on it...it\'s from their official site) then I got an Opeth shirt, and 2 Iced Earth shirts =] I can\'t wait to get them. I spent almost 87$ on band T\'s =X That\'ll suck once the bill comes.       Now tomorrow I have to work. Yuck. I gotta wake up at 830AM for the next (and last) 4 days. I was gonna hang out with Sara tomorrow night after work, but I\'m really not sure now. I may tell her we\'ll just hang out Thursday night since that would be my last day. If we hang out tomorrow, I wouldn\'t get to bed till  probably  1030/1045. I don\'t go to sleep right away. Usually I\'m up for about an hour once I lay down in bed. So, if I go to bed at 1030, I\'ll hopefully fall asleep by 1130 and get 9 hrs of sleep. That\'s if I don\'t have a problem sleeping in the night :/ Blah. I hate sleep. And yet, every morning when I wake up I just think about going back to bed at the end of the day >P       Now I\'m just rambling. So I\'m gonna go.   '),(14365,4240336,'15,August,2004','2004-08-15 00:00:00','   I decided to come back to blogger.  Only because I just want a simple journal site.  The other journal sites make things too complicated ><>!!   For fair warning:  I will speak my mind without caring about what others may think when reading it.  This is my little blogger so I\'m gonna say my own little thoughts.  Just fair warning =)   I won\'t always rant and rave about debatable,  controversial little topics but I may if I get frustrated enough about something.  So I guess I shall update later =)   '),(14366,3469176,'30,May,2004','2004-05-30 00:00:00','   4:50pm 5.30.04   tis time again... tis time for me to replenish my strength in the sanctuary... tis moment where singing songs of love and adoration and \"searching\" and worshiping one\'s Pillar of Strength will rejuvnise first your spirit then your soul then the flesh... im beginning to realize once again that i am nothing if i am not attached to the Vine. Beginning to realize my mistakes... beginning to realize that success is not in my hands. I am not even in MY hands.  time of refreshing is at hand ...  '),(14367,3469176,'28,May,2004','2004-05-28 00:00:00','   8:36pm 5/28   I believe in the plans that He has for me. He is my inspiration. My Rock and my Salvation. He is everything. But though His promises are in your hands, why do you keep walking in the counsel of the ungodly? You wretched soul... ask for forgivness so that you might become acceptable in His Presence...Pray that you might not be tempted.  '),(14368,3469176,'27,May,2004','2004-05-27 00:00:00','   8:33pm 5/28  got to get hold of reality... time is of the essence...  '),(14369,3469176,'27,May,2004','2004-05-27 00:00:00','   5.27.04   8:30pm  school started 5 days ago....a fresh start......Grace is all i need to do this right....Grace.  '),(14370,3469176,'13,June,2004','2004-06-13 00:00:00','   10:58am 6.13.04  ok... here i am again... why is it so hard to tread the path of the righteouss? i ask even though i know the answer... its really wierd to know the Truth and yet still go against it, at times maybe deny it even. its hard to do things now because of the present situation... trials and tribulations are aplenty.   School has been great, generous even. by saying so, i mean that the Professors are not that sadistic anymore...or maybe its just me. i dont know...   relationships are going stronger... now i can see those that are true from the those that are merely acquaintances... the genuine from the \"japex\"! sapphires from the.... rubies??? i dont know... both are precious but , its not really a matter of importance or value... but a matter of those I chose that will be more valuable to me...or \"usefull\"(hmm)... valuable...usefull... i think both are basis in choosing friends... but the first more than the latter, ofcourse.  i REALLY NEED to rub myself off this wap thingy... its taking over my whole life!!!! i dont know what to do.... But God is soooo Good.... he answered me... my prayers... He spoke to me thru Ptr.Jo last sunday....i have to go back to the heart of Worship(like the song).... Its really not that easy....by yourself...once again...Grace is all i need...and the Source of Grace ofcourse....   He is God Almighty, so we dont have to be.\"  '),(14371,3469176,'03,June,2004','2004-06-03 00:00:00','   11:00am 06.03.04   Isaiah 9:6   There is a profound truth at work in the life of every Christian.  Regardless of your situation you have hope and deliverance because you have  Jesus. You have a wonderful, eternal, Father who is able to advise you.  Guidance, direction, healing and counselling are all found in his  words. He gives peace to all who will trust in his abosolute ability to work  all things for good. He carries the burdens of all who will allow him  to do so.   He is God almighty, so you don\'t have to be.   No matter where  you find yourself today, as long as you have Christ you have something  to smile about.  '),(14372,3469176,'08,July,2004','2004-07-08 00:00:00','   19:44 7/8/04      the past few weeks were a total emotional roller coaster ride... since she was admitted to the hospital, the mood around me seemed to fluctuate like i was in a...or was it me that changed... i still dont know...aahh, sumthing to ponder later while i lie myself to deep slumber.i have sumthing to share, since i didnt find my physical \"listener\" yet, i resort to bloggin\' it   here it goes....    i have a sister. born as patrisha anne(my fave name), on the 26th of january \'95. she is my parents\' only daughter. shen is a beauty to behold. sum people told me that i look just like her. even though she was still little, you can already see that she will be beautiful when she grows up...   she was diagnosed with cancer at the age of 1 y/o. that time was kindof historic for my family since both sides, my mother\'s and my father\'s, didnt have a history of cancer or sumthing... she went thru chemotherapy for a couple of years...at that time i can still remember my mom, looking for a head gear, sort of a bonnet(?), for trisha to wear bcoz she lost all her hair due to the chemicals from the therapy... before she was even 5 y/o, she already had 4 operations...you could just imagine....the nurses can no longer see the veins where the I.V. would be placed bcoz all the veins in her arms are already, what the doctors would call, already crushed bcoz of injections that the hospital people should give her.....   june 27, 10 days after my dad\'s birthday, 4 more days till my mom\'s bday,it was a sunny sunday afternoon, trisha was in the hospital bcoz of a heart attack she hadd a few weeks back. she had heart failure bcoz of the chemotherapy, she also had a kidney failure as a reaction to one of her heart medications....   it was june 27, it was a sunday, i woke up late bcoz saturday night,i went out with my barkada, twas Cha\'s bday at gilligans,coffee at antipolo, strolling along roxas blvd... i got home approximately 6am june 27....woke up late....i got a call from sum1 whose voice i didnt recognize, because \"she\" was crying and shouting something at the same time....i was like wtf!!?? i yelled back asking her who she was...? but all i understood was,\"Your sister, trisha.... she\'s dead....(mumbling, didnt understand)....\" i was dumb struck!! i felt like sumthing with cold fingers touched the back of my neck, it was a moment i cant forget, with the voice, and everything......   july 2 came. i had a sister and she was everything for my parents.... my parents came home, their eyes&nbsp;seem to say that everything is ok.....  but it was not.  i had a sister and she was my parents\' only daughter, flower among the thorns  ika nga.  she passed away last june 27,2004. it has been 3 weeks now since the burial... we all know that she is now in heaven... and knowing this is enough for me.  God is good. that is a known fact. He didnt fail to give us the comfort thru friends and other relatives.... and i believe with all my heart that i will see her again one day.... a day that i too will be in heaven.   \"He is God Almighty, so we dont have to be...\" '),(14373,1550687,'21,July,2004','2004-07-21 00:00:00','   &nbsp;>>> a letter sent to me by Melissa in one of her glory and glowing days ...    &nbsp;   Hi Bes...  &nbsp;  It\'s a Sunday, i\'m the only phone rep here tonight save for BB&nbsp;and JQ whor\'ll busy answering emails. And me? My Eshare\'s all fucked up so I can\'t answer emails.  &nbsp;  I want to text you and ask for the link where you can write what-have-you\'s but I guess you\'re sleeping so i thought otherwise.  &nbsp;  I\'m watching Possession, the movie i love because I can actually relate to it.  &nbsp;  I love what&nbsp;Dodo did for me yesterday, Saturday. I was chatting with him at the start of my shift. I mentioned that I was the only rep here. Next thing i knew, he went offline and he showed up at the office.  &nbsp;  He brought some pandesal, peanut butter and chicken spread, and grapejuice (which he personally prepared..di masyado matamis, pero ok lang!). How sweet!  &nbsp;  I\'ve always imagined someone doing that for me while on my weekend shift. Really nice of him.  &nbsp;  He kept me company til&nbsp;5am. He left coz he had to pick her sister up at the bus station.  &nbsp;  I think I\'m genuinely starting to like him. &nbsp;I didn\'t know there was a sweet side to him. If only I get to see more of that...a bit of consistency.  &nbsp;  Haaay...i was grinning stupidly while I was on my way home.&nbsp;I kinda missed the feeling. I am, as what I\'ve been ranting about for a couple of weeks now, succumbing to what my vulnerable side has been telling me....that I am getting tired of being alone.  &nbsp;  To have someone I can share my day with. Perhaps travel to places ive never been to.. discover things together...hohum..I\'m starting to need somone to care for me.  &nbsp;  Whatever.....  It\'s just a thought I\'m dying to get out of my system.      ~ Meme'),(14374,1550687,'20,July,2004','2004-07-20 00:00:00','   These were&nbsp;some letters from Nemo. We\'ll be celebrating our 3rd year anniversary this December and I couldn\'t help but reminisce on old times and read old letters...  &nbsp;  &nbsp;          Happy 6 mos.        &nbsp;   and yes we\'re BOTH still here.  I am happy we have officially become employees of SYKES.  and of course of MACROMEDIA #1.  &nbsp;  6months and it just went by with no warning.  life here has been funfilled and adventurous.  full of love and tears.  &nbsp;  what I am most happy about is our friendship.. which has grown ...  we don\'t need to say a single word and we would knwo how each other feels...  Our trials&nbsp;and petty fights has made us more strong and more undefeatable.  &nbsp;  My life has been much easier with you here. I hope in my own way I was able to touch your life,  and made you relaize what a great friend you are.  &nbsp;  I wish you only the best in life. only the best. nothing but the best.  &nbsp;  Imagine    6months.marlon.jonnel.anna.doc.rv.marky.galera.porch.whistlestop.'),(14375,1550687,'19,July,2004','2004-07-19 00:00:00','   &nbsp;     My heart ached when I saw your almost throbbing hand yesterday. I felt your anger but decided not to rub salt on it by giving you unsolicited comments about what happened.   I resorted to just listening but hurting at the same time.'),(14376,1550687,'10,August,2004','2004-08-10 00:00:00','   hi bes..   i\'m so glad we were able to set this blog up so we can still stay in touch. We can still say what we couldn\'t say to each other verbally.   Yep, I\'m in a different shift and it\'s like I\'m detaching myself from you guys. I see you all scrambling at the start of your shift while I\'m already on my way home. It\'s nice because at least I get to miss all of you.   Wow, you\'ve started going to the gym. That\'s good. I wanted to tag along but i thought otherwise. Pagkatapos nyan, beach na tayo.   And about the guys who\'ve been clamoring to get to know you...good for you, bes. But I do know that you\'ll make the right decision coz we both know that we deserve the best. And I\'m glad you\'re starting to like a certain someone although you haven\'t told me much about him, I\'ll respect that and prolly wait for the time when you\'ll finally tell me.    What\'s up with me?    Well, I\'ve been battling some financial issues at home. Katrina\'s been bugging me to let her ride on my credit card. I told her that I\'m actually thinking of shelling out some moolah for Paolo\'s tuition fee that\'s due at the end of this month but she wouldn\'t hear it. She actually texted some of our tita\'s asking if she will be allowed to ride on their credit cards! Que horror!  Guess i learned something new about my sister (oh, wait! Macoy approached me to ask a question. I had to stop typing so I could cover my nose. ok, where was i)  . She will get the thing she wants one way or the other.  Hohum...sometimes I spend hours getting ready for work because I would always end up staring at the wall for almost half an hour just trying to figure out what i\'ll do with my life and my family. I\'m afraid i\'ll end up following kaje and get a job in Dubai just so i can give enough for my family.  I\'m scared because I might lose myself in the process. I\'m afraid I won\'t be able to fulfill my dreams because of all these shit. I try to be selfish at times but everytime I go home and I see the empty cupboards, I am reminded of my neverending responsibilities.  I just wish I could get away for a couple of days and just enjoy myself.  I miss the old times we had. I miss the simple life we had but of course, with Ada entering the picture, one could never complain.    O well, life\'s a bitch. I guess I just hafta wait for my turn to wither and die.     love you bes! Always take care! Get enuf sleep if u can, hah. mwah   '),(14377,1550687,'09,August,2004','2004-08-09 00:00:00','   Hi there Bes ...       you\'re on a different shift .. So am I .. oh well... we do need time just for ourselves.. to think and just muse around ..       Yes .. the badminton incident was like a dose of cold water ... I guess that\'s the thing about having a child ... when all your friends are single and able .. I got hurt ... well that\'s that ... I realized a lot after that ...       I guess that\'s why I choose not to attend parties anymore... or wait for an invitation to be extended anymore.. one rejection is enough ...       I love my 8pm to 5am shift .. I\'m starting my post natal exercise program .. and hopefully it will be fruitful... did I mention that I was gonna be attending a wedding and I was a part of the entourage? well ... yes I am .. always a bridesmaid never a bride... hohum ...       I get messages from all over asking me for my number and other stuff .. thru friendster .. but I\'m not really wasting my time over it... I don\'t get thrilled like I used to. I guess I\'ve truly matured... I don\'t know bes I want someone I\'m comfortable with ..       I think I\'m liking someone ... but I\'ll keep it to myself for now .. I wanna straighten things out with Reker first ...       You\'ll be in that shift for the whole month ... I was thinking ... why don\'t we check into a hotel and I don\'t know .. brood??? we\'ll plan it ...       this will be for now..    Chin   '),(14378,1550687,'06,August,2004','2004-08-06 00:00:00','   You wrote:    I miss Summer..we seldom spend time together.. she belongs to a different shift...I miss going out with her...coffee and smoking..reminds me so much of her..She seems to be fine with GT <> ... well, I guess she has to be in her own element again .. Boracay comes to mind...    ...I love to spend time on my own..and just be by myself..walking around smoking, eating and other stuff..I guess that\'s why Summer and I are classic.   And I say to you:  I miss you too..so much more than you\'ll ever realize. I miss hugging you. I miss the comfortable silence only the two of us can really hit if we\'re together. I miss just being with you and that doesn\'t include the talks. You\'re one rare gem. I can be myself when I\'m with you. I can be grouchy, I can talk about nonsensical stuff as well as really really profound ones and you\'ll counter it with your own views. I miss that about you.  The \'badminton\' incident was really awkward. My mind was out of sorts and to tell you the truth it did cross my mind to ask you but I thought otherwise. I guess I underestimated you. I do apologize for that. It\'s just that there are times when I really have to shake and remind myself that you\'re with us again.  I guess I got so used to not seeing you around when you were carrying Ada in yer tummy. I have yet to snap out of it. You\'re already here with us. And you\'re ready to start a new life with Ada. Sorry if I could be selfish like that sometimes.  Maybe it\'s the price I have to pay for being friends with a lot of people. I really have to struggle to get out of their circle just so i could spend some time alone. That\'s also the reason why I chose to work on a diff shift for the remainder of August. I want to sort my life out and spend time with people I really want to spend time with. And with no questions asked: you are no.1 on that list.  I miss you bes. Wish we can go out some time and be autistic.  Love you!   '),(14379,1937270,'30,June,2004','2004-06-30 00:00:00','   You can always count on The Onion to cover the big stories... urlLink REAGAN PYRAMID NEARS COMPLETION!  '),(14380,1937270,'29,June,2004','2004-06-29 00:00:00','   Bush & Kerry in 2004 face the strong undercurrent in Ohio of Jerry Springer\'s likely 2006 gubernatorial run.  It would be best if both avoided the arrogant condescension of the mainstremam Ohio media on Springer\'s suitability.  The Plain Dealer\'s Brent Larkin,  urlLink displaying the sophisticated analysis of a foot stomping kindergarten teacher , claims that it would be yet another disaster for Ohio Democrats if Springer ran in 2006.    Hey Brent...get over yourself.  (while we\'re at it, Dick Feagler needs to get over himself, too...he\'s stomping his foot just as hard as Brent.)  Yes, Ohio Democrats have an inexhaustible supply of incompetence.  Yes, it looks desperate to court Springer.    None of that means Jerry Springer shouldn\'t run for governor.  Springer is a legitimate Ohioan, who has held office in Ohio before, is a very good public speaker, and has been doing a lot of the right things to build a campaign.    I used to be in the \"no, Jerry, no\" crowd, used to think that we Democrats should adhere to some intellectual purity that demands that we \"recruit\" candidates as if we were holding a job interview.   The California recall changed my mind.  If Arnold Schwarzenagger and Jesse Ventura can be governors, based on little more than celebrity, Jerry Springer can run for governor of Ohio.    If Springer is able to distance himself a bit from the show (and it won\'t take much to do that), but still display the uncanny ability to relate to average schmucks which makes the show work, then he will win in 2006.  Period.  And a Democratic Governor Springer is better than any Republican.  Too often, Democrats spend more time in search of some silly Aristotlian ideal of politics while Republicans fight like junk yard dogs to win at any cost.  Bill Clinton knew that.  Brent and Feagler don\'t.  '),(14381,1937270,'29,June,2004','2004-06-29 00:00:00','   It\'s been brought to my attention that my blog sounds a bit to hard on Democrats in Ohio.  Given the state of Ohio Democrats, I\'m not sure that\'s really possible...but never let it be said I don\'t respond well to criticism...a good turnout for a literature drop this weekend (20 people...in   June   for god\'s sake), reminded me of a story.    â€œâ€¦anybody home?â€    The old metal screen door looked open.  Lights were on inside.    â€œâ€¦hello?â€  Still no answer.  I took my hand out of the front pocket of my waterproof jacket to look at my watchâ€¦shoved it back into my pocket, and thought for a second out in the cold October rain.  Looked around the parking lot for other carsâ€¦nopeâ€¦just mineâ€¦my old blue Ford Escort, packed with campaign literature, looking very lonely in the wet parking lot behind a strip mall along the main drag in Euclid, Ohio.    Might as well walk in, door looks open, better than standing here in the rain.  The squeaky door slammed behind me as I walked through the narrow back hallway toward the old storefront.  â€œHello?â€ I said.  â€œYeah,â€ an old voice crackled from near the front window.  â€œAnybody home?â€  I walked into the well lit storefront office space to see an old man sitting in a brown metal folding chair, looking out the rain splattered front window.  â€œGâ€™morning,â€ he said.  â€œHiâ€¦Frank Chukayne?â€ I asked, walking over to him.  â€œThatâ€™s me,â€ he answered, turning to look at me from his chair.  â€œHi, Tim Russo from the Hyatt Campaign,â€ I said, reaching out to shake his hand.  Frank smiled.  â€œHow ya doinâ€™?â€ he asked through a meager, unlit cigar butt lodged in the corner of his mouth.  We shook hands.  â€œA bit wet,â€ I said.  â€œYourself?â€  â€œCanâ€™t complainâ€¦â€ Frank gave the cigar a chomp.  I launched straight to business, turning back toward the parking lot.  â€œIâ€™m just gonna get the lit from my car.â€   â€œSuit yourself,â€ Frank said quietly.  I rushed out to the car, grabbed a couple boxes, 3,000 pieces each.    â€œHow many people we got cominâ€™?â€ I said as I threw them onto the floor next to the table across from where Frank was sitting.  â€œWeâ€™ll see,â€ Frank replied.  I ran back out to the car, got two more boxes.    â€œAnybody call to cancel?â€ Threw another two boxes on the floor.  â€œPhone hasnâ€™t rung,â€ Frank noted.    Another few trips to the car later, and all 20,000 Hyatt for US Senate lit pieces were on the floor, at the Euclid Democratic Party Campaign Office, ready to be delivered as soon as the twenty volunteers showed up.    â€œJust one more tripâ€¦,â€ I said, â€œâ€¦brought donuts.â€  â€œThings are lookinâ€™ up,â€ Frank smiled.  I put the three dozen donuts out on the table amongst the array of campaign literature from every campaign in the area, little stacks of leaflets in all shapes, colors, and sizes, for every race up and down the ballot for the November, 1994, election.  â€œRight,â€ I said as I clapped my hands, took off my wet waterproof jacket, and grabbed a cup of coffee.  I sat down and looked at my watch again.  â€œPeople should be showing up in no time.â€   Frank just sat in his brown metal folding chair, looking out the fogged up window into the rain, cars racing by along the busy main street, throwing up spray.  About 70 years old, Frank looked every bit the old Slovenian blue collar Cleveland Democrat.  The white dress shirt well past its prime, brown polyester dress pants, dress shoes, thick forearms belying years spent in a steel mill or a factory, the slight build, the thin grey hair slicked back with a bit of brylcream, the Buddy Holly horn rimmed glasses.  And the thin, stubby, Garcia y Vega cigar.   â€œWant a donut?â€ I asked.  â€œNo thanks,â€ Frank said, â€œmy wife would never forgive me.â€    I took a sip of the coffee.  The quiet of the empty office just then began to occur to me, so I nervously filled it with chit chat.  â€œHowâ€™s the campaign goinâ€™ here in Euclid?â€    Frank smiled again, his cigar crawling back and forth from one corner of his mouth to the other.  â€œNot badâ€¦â€  His tone was half sarcastic, half knowing resignation.  Frank was the Democratic Partyâ€™s ward leader in the city of Euclid, the partyâ€™s most senior person within the Cleveland suburb.  He ran the little storefront office every election, and had come in this Saturday to open the office for the canvass Iâ€™d organized.  He knew just how bad the 1994 campaign was going.    â€œHowâ€™s the Hyatt campaign?â€  Frank asked politely.  â€œNot bad, either,â€ I lied through my teeth.  I looked at my watch again.  â€œWeâ€™ll see how many people turn up today.â€      November, 1994, was the year that the Republicans swept to a national landslide victory,   taking both houses of Congress, making Newt Gingrich speaker.  In Ohio that year, the landslide would be particularly disastrous, Democrats losing every statewide campaign plus a few congressional seats, including the US Senate seat Democrats were fighting to hold onto, the race for which I was a paid local coordinator, whose 20,000 lit pieces were sitting in the Euclid campaign office waiting to be sent out.  Frank and I couldnâ€™t possibly know how bad it would finally become, but we sure felt it coming that morning.  â€œHow many pieces you got there?â€ Frank asked.  â€œTwenty thousand,â€ I answered.  Frank nodded.  I looked at my watch again.    The afternoon dragged on and on and on.  The rain kept falling.  Frank and I sat there waiting.  And waiting.  Iâ€™d spent the entire week calling volunteers, setting up the lit drop.  I thought I had twenty signed up.  A half hour passed.  None of them showed.  â€œThink Iâ€™ll give it another hour,â€ I said.  â€œSuit yourself,â€ Frank replied.    And there we sat, one young, excitable, eager campaign staffer, feeling like a damn fool, and an old timer chomping on a cigar, who could have said â€˜I told you soâ€™ but mercifully did not.    Our idle chit chat continued, my utter embarrassment keeping me quiet, listening to Frank, who started to tell a few stories.  Frank had been the ward leader in Euclid for decades, involved in Euclid politics almost since he arrived in Euclid as a kid on a train in the Depression.  Weâ€™ve all heard the romanticized stories of people in the Depression hopping onto boxcars and hopping off where they might find a job.  Frank was one of those people.    â€œYou were a  hobo ?!â€ I laughed.  â€œYou could say that,â€ Frank replied.  â€œSo howâ€™d you get involved in politics?â€ I asked.  â€œI mean, you just jump off a train and show up in Euclid in the middle of the Great Depressionâ€¦â€  â€œWell,â€ Frank smiled, â€œ thatâ€™s  an interesting story.â€  On return after fighting in the second world war, settling into a neighborhood, starting his family, Frank decided to run for precinct committee.  Precinct committee in the American system of government is the very first level of elected government at the grassroots.  The precinct committee is the body responsible for organizing and implementing any election that will occur in the precinct, a tiny geographical area, usually no more than a thousand or two thousand people in each..  To insure partisan balance, they are made up in equal numbers of Democrats and Republicans; in Ohio, one each per precinct.  Therefore in each precinct, there is an election within each party for a precinct committee person from that party for every precinct, who then recruits the people to hand out the ballots, manage the lists, count the ballots on election dayâ€¦and hopefully be the backbone of the campaign at the grassroots.  But mostly, they donâ€™tâ€¦wellâ€¦do much, really.  They get old.  They get lazy.  They save their effort for election day and their basic statutory duties of organizing the polling places, their involvement in party politics limited to monthly or yearly meetings, if that.  For the most part, they are names on a list that someone from a campaign, like me, would call to ask for help who would almost always decline; either too old, or too busy, or too uninterested to bother, especially if the cause seems lost anyway.  I didnâ€™t bother to ask Frank if heâ€™d made any decision along these lines, and if so, why.  I just let the silence in the cold empty storefront speak for itself.  Frank was happy to come into the office and open the door on a wet rainy Saturday morning.    In 1994, the tectonic plate-shifting of national politics clearly evident  , he wasnâ€™t about to do much more.  Which left it to me to get my own volunteers from my own list to deliver the leaflets for my candidate   Which I tried to do.  Got twenty people to commit.  Had my list with me, even.  That was before the weather failed to cooperate, as it almost always does.  Combined with the apathy and resignation taking hold throughout the country, it was inevitable Frank and I would be sitting alone in the Euclid office, surrounded by undelivered lit.    So story telling about post-war runs for precinct committee was the order of this cold, wet day.  â€œWasnâ€™t there already someone holding the seat?â€ I asked.  â€œThere sure was,â€ Frank said.  â€œAn old woman had held it for years.  Years.  And people couldnâ€™t stand her.â€  â€œWhy?â€  â€œShe never did anything in the neighborhood.  Never went to any of the party functions, never campaigned for the ticket, always expected people to work for her on election day, but never did anything else.â€  Frank got a bit excited all of a sudden.  â€œSo I started my campaign.â€  I looked at the thousands of pieces of lit on the floor for a second, spending a final, short, and resigned amount of anxiety on them, and grabbed another cup of coffee.    â€œHowâ€™d you do it?â€ I said, now intrigued.  Unseating an incumbent, be it a precinct committee person or a countryâ€™s president, is rare and difficult.  â€œWell, first I counted all the houses that were in the precinctâ€¦â€  â€œO.Kâ€¦â€  I sat back down with my fresh brew.  â€œAnd figured out how many houses I could visit in a dayâ€¦â€  â€œRightâ€¦â€  â€œDivided the total houses in the precinct by the number I could visit in a day, and figured out how long it would take me to visit all the houses.â€  This all seemed pretty standard.  â€œRight, rightâ€¦â€  Then Frank got a glimmer in his eye.  â€œNext, I took a map,â€ he said, â€œand looked at where her house was on the map of the precinct.â€  â€œOâ€¦K...,â€ I said.  I leaned over in my chair.  â€œAnd I drew a line from her house, to the house  farthest  away from hers on the map.â€  Frank was getting into it now.  So was I.  â€œYeahâ€¦â€  â€œAnd started knocking on doors at that house.â€  â€œThe one furthest away from hers?â€  â€œThatâ€™s right.â€  I paused, thinking it through.  â€œAnd you just kept knocking on doors every day?â€  â€œEvery day,â€ he said, â€œgetting closer and closer to her house.â€  I got it finally.  â€œ Ah hahhhhhhâ€¦,â€  I smiled, leaning back in my chair.  â€œUntil the day before the electionâ€¦,â€ Frank chuckled.  â€œWhen you knocked on   her   door,â€ I said.  â€œThatâ€™s right.â€   â€œThe day before the election?â€   â€œThatâ€™s right.â€  Frank smiled again.  I fell back in my chair.  â€œShe must have shit her pantsâ€¦â€  â€œShe had no idea what was happening,â€ Frank said with a triumphant chomp on his cigar.  â€œShe was so out of touch with the neighborhood, she hadnâ€™t heard a thing.â€  â€œHolyâ€¦,â€  Frank started laughing like a little kid.  â€œShe opened the door, and I was there with my campaign leaflet.â€  â€œOh my godâ€¦â€  â€œAnd I said, â€˜Hi, Iâ€™m Frank Chukayne, candidate for Democratic precinct committeeman,â€ Frank said through another chuckle.  â€œWhat did she say?â€  Frank leaned over in his chair, his face next to mine.  â€œShe said,  â€˜You canâ€™t run for precinct committee! Iâ€™m the precinct committeewoman in this precinct!â€   â€œWhatâ€™d you say?â€   â€œâ€˜Not for long, you ainâ€™t!â€™â€     We both burst into laughter.  â€œYou must have kicked her ass,â€ I said.  â€œGot 80%,â€ Frank said, sitting back in his chair, hands in his pockets, cigar chomped over to the other side of his mouth again.    â€œShe never knew what hit â€˜er,â€ I said.  â€œNever saw it coming,â€ Frank saidâ€¦never saw it coming.    Frank and I never got out the 20,000 pieces of lit, but his story about a precinct committee election in post-war Ohio got out to someone who needed to hear it during a tough election.  '),(14382,1937270,'25,June,2004','2004-06-25 00:00:00','   On seeing â€œKerry Legal Team Updatesâ€ on the Thursday, June 24, Cleveland, Ohio Volunteer Meeting Agenda I was initially encouraged.  Good to see the Kerry campaign gearing up for potential election day shenanigans, I thought.  Boy, was I wrong.  Apparently, the â€œKerry Legal Teamâ€ held a national conference call with all 68 field staff in the Kerry campaign to deliver instructions on certain â€œwell meaning people and groupsâ€ engaging in activities on behalf of John Kerry that were illegal.    Activities like selling John Kerry bumper stickers at house parties or â€œmeetingsâ€, or selling homemade John Kerry bumper stickers, anywhere, like at parades.  Soliciting funds for the Kerry campaign.  Passing the hat at a Kerry house party to raise funds for local party or candidate activities.  â€œRepresenting yourselfâ€ as part of the Kerry campaign while engaging in such activities.  All illegal.  Unless, of course, you are doing it under the direction of the â€œofficialâ€ Kerry staff, for the â€œofficialâ€ Kerry campaign.  Phone banks at your house?  Illegal.  Phone banks at the friendly local law firms who we know and love, who we funnel people to all the time?  Perfectly fine; in fact weâ€™d like you all to volunteer for those phone banks.  The room of about 50 volunteers seemed a bit confused.  Anxious.  Wondering what the hell these people were talking about.    The Kerry staff continued, asserting that anyone running a Kerry operation on their own, opening an office, or running their own phone bank, risked legal action.  In fact, the Kerry Legal Team in DC was preparing to shut many such operations down.  Realizing that this recently condemned illegal activity sounded a lot like what the Kerry Campaign Internet Meetups have been doing for the last year and a half, some volunteers present who were active in the Meetups asked in humorous futility, â€œIs it o.k. to go the Meetups?  Theyâ€™re â€˜unauthorizedâ€™, does that make them illegal?â€  The Kerry staff instructed that it would be o.k. to go to the Meetup, take some forms, and sign them up to work for the â€œofficialâ€ campaign.  Meetup regulars started getting restless.  â€œWhat   is   the campaign?â€ one asked in frustration.  â€œIâ€™ve filled out numerous volunteer forms, and have yet to hear from anyone with the â€˜officialâ€™ campaign.â€  â€œYouâ€™re here tonight, arenâ€™t you?â€ the top staffer responded.  â€œWe direct the campaign from  here .â€  â€œWeâ€™ve had a house party structure in place for over a year,â€ shot back the volunteer.  â€œAre we supposed to stop that operation?â€  â€œWe welcome that into this group, here,â€ replied the staffer.    â€œYou have to understand,â€ half joked the other staffer, â€œwe donâ€™t want to go to jail.â€    Right.  After all of this discussion about illegal in-kind contributions or activities from â€œunauthorizedâ€ groups or people, the staffers then issued a blanket call for in-kind donations of office equipment; tables, chairs, phones, laptops.    That probably made the confusion unanimous.  Iâ€™ve been working in campaigns for many years, as volunteer and staff, at many levels.  I have never seen a volunteer meeting so thoroughly dominated by a rambling, hamfisted, and mostly ignorant discussion of federal campaign finance laws and their dire consequences for volunteers.    Issuing such threats seemed to be the number one priority of the Cleveland Kerry staff.  Everything else was an afterthought...recruiting, organizing, targeting...all of it took a back seat.  Out of an hour long meeting, more than half of it was devoted to a legal debate that no one in the room was qualified to hold, but that the Kerry staff was eager to pound home with the gracefulness of a soviet politburo member pounding his shoe on the table.  Let\'s clear up one thing.  Campaign finance laws are complicated, yes.  They do not result in volunteers ending up in jail, ever.  It is important to make volunteers aware of the rules they must follow when campaigning.  But that does not appear to be the intention of this new found, suddenly top-of-the-agenda emphasis on the FEC and how lawyers are about to hunt down â€œunauthorizedâ€ or â€œillegalâ€ campaign activities.  Whether or not there is an intentional, national effort within the Kerry campaign to squash Meetups and the organizations theyâ€™ve built, it is certainly manifesting itself as such on the ground.    What may have been a legitimate conference call on minor legalities is being used, at least in Cleveland, Ohio (read â€“ top targeted state) as a weapon in a petty turf battle over who controls the campaign.  Itâ€™s ugly, counterproductive, and likely to scare off a lot of the new people brought into the party by the unprecedented use of the Internet to re-energize the party through Meetups...which may be the intention of the â€œauthorizedâ€ campaign.  Whatever the intention, the Meetup crowd carries on.  After the 5.30 pm meeting on Thursday, at 7 p.m. a far more organized Kerry Meetup was held just across town, where the confused discussion of what might or might not be legal was continued reluctantly.   Amazing, isn\'t it?  Let\'s go after GW.  But first, let\'s get after each other.  '),(14383,1937270,'23,June,2004','2004-06-23 00:00:00','   George W. Bush was in Cincinnati yesterday, the third time since March.  This is a good sign for Democrats.  Greater Cincinnati is the hardest of hard core Republican areas in Ohio.  It makes sense for Republican candidates to go there late in the election, to rally the base and get them out to vote.  If GW is going there nearly once a month at this point in the campaign, it means hard core Republicans still need presuading...which is not where Karl Rove would like to be at this point, I am quite certain of that.  It also means that the places where GW can go in Ohio and get a warm enough welcome are dwindling.  Where GW needs to go is a place like Stark County, the national media\'s darling bellweather county.  Unfortunately, with plants closing there once a month, I\'m guessing Canton isn\'t on the White House itinerary.  '),(14384,1937270,'21,June,2004','2004-06-21 00:00:00','   In the biggest \"...yeah, so anyway\" of all time, the media follows up a bipartisan week of Reagan worship and the nauseating sight of George W. Bush drooling all over the Clintons at their White House portrait unveiling, with a return to the good old days of discussing where  Bill Clinton\'s Penis  may or may not have been.  How does this matter to Ohio in 2004?  I asked my mom, who as everyone knows is   the   Ohio swing voter, what she thought of the 60 Minutes Bill Clinton interview.  It was like entering a time warp back to 1998.    She said, \"What do I care?\"  End of conversation.    Which of course is why Bill Clinton survived.  Real people never cared where  Bill Clinton\'s Penis  may have been.  Only Ken Starr and his disciples cared where  Bill Clinton\'s Penis  may have been.    On November 2, George W. Bush will find out precisely what real people care about.  '),(14385,1937270,'18,June,2004','2004-06-18 00:00:00','   Ohio Democrats have found many new and exciting ways to lose statewide elections in the past decade.  Here\'s my personal favorite.  Election day, 2000, about 2 p.m.  Despite the Gore campaign having written the state off 10 days out, everyone knew it was close and getting closer.  I was dragging a group of seven volunteers into the most important Democratic congressional district in Ohio, Stephanie Tubbs Jonesâ€™ 11th Congressional District, to help knock on doors and get out votes in the heavily African American east side of Cleveland.  Parking our cars, we marched to the front door of the lead campaign office on Euclid Avenue, prepared to do whatever anyone asked us to doâ€¦but quite unprepared for the sight which greeted us.  The office was completely empty.  No tables, no chairs, no phones.  No boxes of literature, not a single doorhanger, sign, button, leafletâ€¦nothing.  Some guy was carrying a computer out the back door to a U Haul truck.  I walked back outside, checked the address on the front doorâ€¦this was the place, alright.  â€œWait here a second,â€ I told my befuddled friends.    I charged right back in.  â€œWhat the hell is going on here,â€ I said to the guy with the computer.  â€œAsk her,â€ he said, pointing at the campaign staffer in the corner, Betty Pinkney, the Congresswomanâ€™s handpicked coordinator, who was standing next to a bank of charging mobile phones that were not in use, directing the movers.    â€œBetty,â€ I shouted, charging toward her, â€œIâ€™ve got seven people here ready to go to work.â€    She looked at me like I was some kind of insane person.  I continued in futility.  â€œWhat do you want them to do?â€  â€œYouâ€™re too late,â€ she said insistently.     â€œToo late?â€   â€œWeâ€™re movinâ€™ out.â€  I paused and squnited.    â€œWhat?â€    â€œLease is upâ€¦weâ€™re moving out.â€  â€œYou must be joking,â€ I said.  This was a new low.  â€œNope,â€ Betty replied.  â€œWeâ€™re done here.â€  â€œItâ€™s two oâ€™clock!â€ I shouted.  â€œThe polls are open for another five hours!â€  â€œYou see that U Haul truck?â€ Betty finished.  â€œThis office is closed.â€    And that was that.  Ohio is a tough state to win for Democrats.  For John Kerry in 2004, a lot of things need to go right, a lot of breaks need to break his way, and a lot of activity needs to be executed with total perfection.     If the main campaign office in the most heavily Democratic congressional district in Ohio makes its top priority on election day packing up the damn U Haul truck, that means that very little has gone right, a lot of activity was poorly executed, and that George W. Bush will be President of the United States for another four years.  Itâ€™s that simple.  Democrats in Ohio have not been known to learn from such mistakes, but rather, repeat them regularly.  Early signs in 2004 don\'t look good.  If I were John Kerry, Iâ€™d plan to make a call to that office in Cleveland this November 2, about 2 pmâ€¦just to be sure.  '),(14386,1937270,'16,June,2004','2004-06-16 00:00:00','   If anyone still questions Ohio\'s importance in this presidential election, just ask John Ashcroft.  The latest Bush/Cheney campaign tactic in Ohio is to claim Al Queda is targeting Ohio shopping malls.  Is anyone else just a  little  bit suspicious at this?    The case of Mr. Nuradin Abdi looks pretty flimsy to me.  He alledegly lied on political asylum forms, gave \"material support\" to terorists, \"conspired\" (a classic catch all charge) and was in the \"early stages\" of a mall blow-up plot.    What Al Queda operative is going to   apply for political asylum  ?!?!?  Hard to think of a better way to draw unwanted investigative attention to yourself.  Don\'t be surprised if the real story changes.  The Jose Padilla case, the alleged \"dirty bomber\" who has never seen an attorney despite being an American citizen, is now no longer accused of plotting a dirty bomb attack, but of planning to attack apartment buildings.  They arrested this guy on the immigration charges last November 28...why are we just hearing about this   Ohio-20 electoral votes-targeted-swing-state mall blow-up plot   now, 5 months before the election?  You can just see Ashcroft sitting in his office, cooking this up.  \"Let\'s see...swing voters in Ohio, how do we target them...I BET THEY SHOP AT   THE MALL  !  Al Queda hates malls!  Let\'s round up a brown skinned Muslim looking fellow, trot him out, and say he wants to  BLOW UP THE MALLS IN OHIO!   Yeah, that\'s it!\"  I\'m not usually a conspiracy theory guy, but I wouldn\'t put it past Ashcroft to use the justice system to target anti-terror arrests in swing states this presidential election, hold breathless press conferences, and start counting electoral votes.  In fact, I bet there\'s a bomb plot unfolding in Missouri, Wisconsin, Nevada, Florida....  '),(14387,1937270,'15,June,2004','2004-06-15 00:00:00','   Ralph Nader spoke at The City Club in Cleveland yesterday after a visit to Toledo.  The City Club calls itself \"a citadel of free speech\"...in the interests of full disclosure, they should have billed the speech a Bush/Cheney event.  Nader received 117,000 votes in Ohio in 2000.  The last time a Democrat beat an incumbent president in Ohio was 1992, when Bill Clinton won Ohio by 75,000 votes.    Nader needs 5,000 signatures to get on the Ohio ballot.  If you see someone circulating petitions for Nader, DO NOT SIGN.  Instead, offer the circulator a cup of coffee, take him/her to lunch, WASTE THEIR TIME.     urlLink Plain Dealer story on Nader in Cleveland  '),(14388,1937270,'14,June,2004','2004-06-14 00:00:00','   The ham handed imposition of the Kerry Campaign into Ohio continues with typical imcompetence.  Anonymous sources tell the OhioCountdown that Kerry Campaign staff in Cleveland, headed by recently hired coordinator Vel Scott, held a meeting with the heads of the volunteer committees developed by Meetup organizers over the last few months.  At the meeting, Scott and Cuyahoga County Democratic Party officials simply replaced the existing committee heads with new people who\'ve never been involved in the Meetups, ignoring the list of hundreds of volunteers recruited, organized, and active under the Meetup leadership.  Kerry Meetup organizers are continuing to work as they did before, which means that there are now two structures; two volunteers committees, two visibility committees, two speakers bureaus, two fundraising committees, two voter registration committees, etc.    Two Kerry campaigns in Cleveland...one that\'s existed for months and has functioned on its own without direction with new people, energy, and enthusiasm...and another run by Vel Scott\'s appointees.  Guess which one will actually do the work........   urlLink Compare the \"unofficial\" Kerry website...    urlLink ...With the Cuyahoga County Party website  '),(14389,1937270,'13,June,2004','2004-06-13 00:00:00','   Only Democrats could take an article whose main theme is \"Kerry Will Win Ohio\", and come up with a reason to fight over it.    Case in point...in response to my Salon.com article  urlLink (Salon Article - you need to sit through an ad to see it) , a Jennifer Palmieri, Communications Director for Kerry in Ohio, left a testy message on my machine, calling my attention to the fact that the Kerry campaign is, in fact, in Ohio, and asking to discuss my article.  Which is funny, because upon returning her call, there was (a) no answer at the Columbus number she left, and (b) no response to the message I left on her cell phone.  That was a week ago.  Is the Kerry Campaign here in Ohio or not?  '),(14390,1937270,'11,June,2004','2004-06-11 00:00:00','   Partisan Democrats like me aren\'t real fond of spending a week and countless network broadcasting hours gushing about a Republican in the middle of an election year.  But the man gets his due from me on one note in particular.  I was one of those kids growing up who was like, \"Why can\'t we all get along?\" who greeted the Cold War as some giant misunderstanding, thinking that communism was just another philosophy, as legitimate a way of organizing a society as any.  Until I lived in the former Soviet Union for a couple of years.  The ruin that 70 years of tyranny, murder, opression, and misrule communism left behind in the places where I lived is astounding.    Ronald Reagan was wrong about a lot of things, but he was right about the evil empire.     urlLink Click here for a more raw, current, and oh-so-Tim take on Who Lost Russia from The St. Petersburg Times in Ocotber, 1999.  '),(14391,1937270,'10,June,2004','2004-06-10 00:00:00','   urlLink    Reagan didn\'t get \'em all...Lenin statues still stand in Bishkek, Krgystan, where I trained political parties in 1999.&nbsp; urlLink  '),(14392,1937270,'10,June,2004','2004-06-10 00:00:00','   My mom, known as the zeitgeist of the American electorate, speaketh the truth as only an Ohio swing voter can.  More on my mom from the  urlLink Salon.com article on Ohio.  '),(14393,1937270,'10,June,2004','2004-06-10 00:00:00','   Sometimes itâ€™s hard to know whether to rejoice or weep for my political party, the Democrats here in Cleveland, Ohio.  The unprecedented energy and internet based activism born out of the despise for George W. Bush is cause for most of the rejoicing.  The Meetup phenomenon â€“ like-minded people using the internet to connect, hold meetings, and become politically active, is drawing more first time volunteers into the political process than Iâ€™ve ever seen in a lifetime spent in politics.  Despite the end of the primary, the Meetups continue, now gaining more energy from the decision on John Kerry as our candidate, and the emergence of Bush as the only target.  Cause for weeping can be found in the Cuyahoga County Democratic Partyâ€™s reaction to the monthly Meetups.  On my Meetup adventures, Iâ€™ve always wondered why I never saw anyone representing the party.  The other day, I learned why.  The party hates â€˜em.  I visited the party office on St. Clair, just to pop in, see whatâ€™s going on, get my name on a volunteer list.  I said something about Meetups, and their eyes bulged in shock as if presented with blasphemy.  â€œDonâ€™t get me started,â€ one of them said, shivering and groaning like she just stepped in manure.     â€œNot authorized,â€ said the other, a newly arrived Kerry staffer from out of state.  â€œThat website they have, thatâ€™s not an authorized Kerry website.â€     urlLink (by the way, itâ€™s www.neohioforkerry.com).    â€œThey could get into a lotta troubleâ€¦â€    Trouble?  â€œNothinâ€™ but trouble.â€  I noted that there are at least five different Meetup groups ongoing, including Meetups for Kerry, John Edwards, Wesley Clark, and Howard Deanâ€™s morphed organization, Democracy for America.  They became even more shocked.  â€œNot authorized by us,â€ they kept saying.  â€œTroubleâ€¦â€  I had thought that the local party didnâ€™t attend the dozen or so Meetups occurring in Cuyahoga County every month out of mere laziness.  Noâ€¦apparently, a strategic decision has been made to ignore the most significant recruitment of new people into the Democratic Party in decades.    Meetups are â€œunauthorizedâ€.  If youâ€™ve ever attended any â€œauthorizedâ€ Democratic Party meeting, in this county or any other in Ohio, youâ€™d wonder at this strategy.  The average age of attendees at monthly meetings is about 72.  Average turnout is about six.  Average amount of activity between meetings is precisely zero.  The only thing missing from a monthly ward club meeting that would make it a nursing home is an ambulance on call outside.  The precinct committee structure of the party, locally elected Democrats in each precinct, is dead from the inside out, many precincts unfilled due to death without replacement, the other seats occupied by unmovable dinosaurs whose last meaningful political activity may have been in the Carter administration.  Meetup attendees, on the other hand, are new to politics, average age about 40, people from all races, educational backgrounds, careers, incomesâ€¦brand new activists just itching to get at Bush.  That night, I attended the Meetup calling itself the â€œDemocratic Partyâ€ Meetup.  About fifteen people attended to hear a presentation from a retired Case Western Reserve University Professor who has finally gotten involved in politics, first time in his life, because of the issue of electronic voting security.  People were there to recruit volunteers to register voters, speak on behalf of John Kerry, work on campaign visibility.  Two attendees were actually disgruntled Republicans who were now voting for Kerry.  Stuff was actually going on. I didnâ€™t have the heart to tell them that they were all getting into really big trouble.  If any organization ever needed new blood, literally, it is the Cuyahoga County Democratic Party.  Through the Meetups, new blood is practically oozing out of the ground like oil in the Beverly Hillbillies backyard, and the local party officials shun the whole thing as â€œunauthorized.â€  Somethingâ€™s going to have to give here.  Either the party will reach out to the burgeoning political activity around it, work with the new people, and with them reinvigorate itself, or the new people becoming active will one day realize that they are more important than the party.    Some enterprising person in the Meetup crowd will figure out that down at the party HQ, thereâ€™s no â€œthereâ€ there â€“ that it wonâ€™t take much to just assume the precinct positions that now lay empty or are irrelevantly occupied by aging place holders.  Between now and then, thereâ€™s an election to win, and itâ€™s all hands on deck.    Authorized or not.  '),(14394,1937270,'10,June,2004','2004-06-10 00:00:00','   Started this thing off with an article in Salon.com.  It\'s an answer to the NY Times Sunday Magazine Article about the Bush effort in Ohio.     urlLink Salon Article - you need to sit through an ad to see it  '),(14395,1937270,'10,June,2004','2004-06-10 00:00:00','   urlLink    Me with Governor Clinton in 1992, introducing Clinton to Ohio\'s Republicans for Clinton Chairman, Charlie Dawes.  Note stupid look on face indicating \"I can\'t believe I\'m standing nxt to Bill Clinton\".&nbsp; urlLink  '),(14396,1937270,'10,June,2004','2004-06-10 00:00:00','   urlLink    Bill Clinton at Columbus Airport in 1992, right before the election.&nbsp; urlLink  '),(14397,1937270,'10,June,2004','2004-06-10 00:00:00','   urlLink    Bad hair day, great campaign - Tim at a late 1996 Ohio Clinton Gore staff meeting...cool maps, eh?&nbsp; urlLink  '),(14398,1937270,'10,June,2004','2004-06-10 00:00:00','   Ok, this was all Lori\'s idea.  '),(14399,1937270,'30,July,2004','2004-07-30 00:00:00','   Since blogging about the ill-concieved Bush attack piece on abortion that appeared in Cleveland media markets almost two weeks ago, I\'ve only seen it appear once...could it be that it backfired?&nbsp; In exactly the way that I predicted?&nbsp; Hmmmm....  '),(14400,1937270,'28,July,2004','2004-07-28 00:00:00','   The internet has changed political fundraising so dramatically, it now seems quaint to hear about a major presidential candidate holding a big dollar fundraising dinner.&nbsp; GW Bush is still spending a lot of time doing just that.   His visits to Ohio have almost always included a many-thousands per plate closed dinner at a fat cat\'s house... urlLink like today in Kirtland Hills at Eddie Crawford\'s house .   D\'s certainly do this too, but it may be a dying breed of fundraising.&nbsp; I haven\'t seen the figures lately, but the rate at which Kerry is raising money on the internet seems to open up a lot of his time for more important matters, like talking to voters.&nbsp;  '),(14401,1937270,'28,July,2004','2004-07-28 00:00:00','   The Kerry-Edwards ticket aren\'t the only folks on their way to Ohio.&nbsp; Tourism to Ohio will spike this fall... urlLink at least according to a PD story today , which notes from the convention in Boston that people from all over the country are planning bus trips to Ohio to help the Kerry-Edwards ticket, because one said, Ohio\'s where the next president will be chosen.  '),(14402,1937270,'28,July,2004','2004-07-28 00:00:00','   Been making a point to catch Jerry Springer\'s dispatches from the convention in Boston on Channel 19 here in Cleveland.&nbsp; Needs work.  They keep hyping the coverage as \"not the same old stuffy political coverage\", or \"we\'ve got Jerry\", etc.&nbsp; The product, however, is so tame it might as well be&nbsp;Jim Lehrer&nbsp;reporting from the floor.&nbsp; You get the feeling Jerry\'s being so careful with his first foray into the non-tranvestite-Nazi-on-set-brawl world that he\'s holding himself back.  Case in point - he interviewed Cleveland Mayor Jane Campbell this afternoon, which was hyped by the lead in as \"you don\'t wanna miss this\".&nbsp; Take my advice...miss it.&nbsp; It wasn\'t an interview so much as it was Springer schmoozing Jane for his future gubernatorial run.&nbsp; He actually asked Jane, \"So, you know Cleveland pretty well...will it go for Kerry?\"&nbsp; IS THIS FOR REAL!?!?!&nbsp;  Some advice for Jerry...loosen up a bit.&nbsp; You look tight, nervous, and like you\'ve got an agenda.&nbsp; Oh and another thing...you don\'t need to schmooze Jane Campbell...she won\'t be mayor when you run for governor anyway.  '),(14403,1937270,'27,July,2004','2004-07-27 00:00:00','   Is it just me, or is the commentary from network analysts in Boston for the convention this year simply awful?&nbsp; You can understand why these guys are so easily spun...they\'re morons.  In particular, I haven\'t seen a single discussion of any issues among the pundits.&nbsp; They seem hard wired to discuss people\'s hair dos, the size of the stage, whether or not Ted Kennedy is embarrassing.&nbsp;   Theresa Heinz Kerry had it right...they should all just take their commentary and shove it.  '),(14404,1937270,'27,July,2004','2004-07-27 00:00:00','   Jerry Springer isn\'t the only Ohio Delegate getting major airtime this week in Boston.&nbsp; The youngest delegate (at 17) happens to be from Ohio, happens to be cute as hell, and is plastered all over the 24 hour news cycle.&nbsp;   My favorite, though, was a CSPAN fly on the wall documentary of an hour or so in Renee Lipson\'s life at the convention.&nbsp; Renee is a delegate from Cuyahoga County, and CSPAN followed her around for a day.&nbsp; The 73 year old dynamo kept those whipper snapper CSPAN camera guys hoppin\', from women\'s caucus event to Emily\'s List, to Ohio Lunch, to the convention floor, bouncing all over town like a super star.&nbsp; Renee, you made my day.  There\'s even a special cameo appearance by Madeline Pellish, Renee\'s roommate at the Boston Sheraton, and freelance reporter for the Cleveland Jewish News.&nbsp; Always workin\' it, baby.  '),(14405,1937270,'26,July,2004','2004-07-26 00:00:00','   Jerry Springer is doing a good job of branding this year\'s convention in Boston.&nbsp; Last night,&nbsp; Springer made his&nbsp;first broadcast from the floor for Cleveland\'s CBS affiliate Channel 19.&nbsp; He was on stage at the Rock The Vote event last night near Fenway.&nbsp;   Saturday night at a wine bar in Cleveland, I talked with a friend about Springer\'s prospects for governor in 2006.&nbsp; She\'s a Kerry voter, not real tied in to politics, college educated, runs her own business.&nbsp; She thinks Dennis Kucinich is an embarassment, was so-so on Bill Clinton...probably thinks of herself as an independent voter.  The very idea of Jerry Springer runnning for office makes her visibly cringe.&nbsp; There\'s more than a little hypocrisy in such responses - people who dismiss a Springer candidacy out of hand because he \"exploits\" people with his TV show are also&nbsp;nightly voyeurs, tuning in to the&nbsp;most humiliating of the reality TV shows.&nbsp; After pointing this out to my friend, she came around a bit...but only&nbsp;a bit.  Bottom line...she says she needs to see some \"seriousness\" out of Jerry Springer.&nbsp; Not sure my friend will be watching Channel 19 this week, but note to Jerry...she might.  '),(14406,1937270,'25,July,2004','2004-07-25 00:00:00','   Much is being made of the \"toning down\" of the convention rhetoric in Boston.&nbsp; As this blog has shown, I like a good Bush bash as much as the next guy...but staying positive, especially in a place like Ohio, does work.  In fact, of all the ads that I\'ve seen aired by the Kerry campaign in Ohio, not a single one of them could be characterized as negative, or attack, or a hit piece.&nbsp; Conversely, just about every ad that the Bush campaign has aired in Ohio has been a negative&nbsp;attack piece.&nbsp;   The McCain Feingold campaign finance reform law has its problems, but one little provision has proved very useful...the acceptance of personal responsibility for the message.&nbsp; Attacks on Bush are coming from the 527\'s - MoveOn.org, The Media Fund, and others, hammer away at Bush on the airwaves.&nbsp; But there\'s no tag line saying, \"I\'m John Kerry and I approve of this message.\"&nbsp; Thanks to McCain Feingold, Bush\'s fingerprints are all over his hit pieces,   making Bush the only candidate using negative ads in Ohio.    Voters in Ohio are hyper tuned in for the slightest negative attack.&nbsp; Dirty politics regularly backfires in Ohio, probably because the state is so targeted so often with so much negativity, the voters have had it.&nbsp;   &nbsp;  '),(14407,1937270,'25,July,2004','2004-07-25 00:00:00','   Some readers complained about the non-collegial, fictional, often profane&nbsp;satire&nbsp;tone of the blog lately...now that I\'ve vented, let\'s try polite on for size, shall we?   How about a look at the latest Ohio polling, my friends?&nbsp; What a jolly good idea!   I say,&nbsp;the race is&nbsp;certainly freshening up in the Buckeye State, my dear colleagues!&nbsp;  urlLink One poll shows the right honorable George W. Bush with a lovely 5 point lead , another shows our honorable friend John  urlLink Kerry with a perfectly agreeable 2 point lead. &nbsp;&nbsp;  urlLink USA Today has Kerry with most encouraging 5 point lead .&nbsp; All were cheerfully conducted&nbsp;last week.  &nbsp;  '),(14408,1937270,'24,July,2004','2004-07-24 00:00:00','   I have a lady&nbsp;friend who\'s married to an intransigent non-voter...let\'s call him Squiggy.  Squiggy is at the top end of the&nbsp;typical non-voter demographic.&nbsp; White male, 30-something, married, no kids, home-owner, works in financial products, is college educated.&nbsp; The TV remote&nbsp;is programmed only for the 4 ESPN Channels and FOXSports Ohio, and possibly the Golf Channel...he plays a ton of golf.&nbsp; &nbsp;  Squiggy has never voted, and has been quite proud of that.&nbsp; He\'s cited all the cliche reasons; one vote doesn\'t matter, no one listens, what\'s it good for, blah fucking blah.&nbsp;   Squiggy\'s wife...let\'s call her...hmm...Laverne...dragged him to see Farenheit 9/11 one day.&nbsp; Nothing else was playing, so Squiggy didn\'t put up much of a fight.&nbsp;   Squiggy came out ready to vote against George W. Bush.&nbsp; He\'s not really voting \"for\" John Kerry, but he is going to vote.&nbsp; Wow.  '),(14409,1937270,'22,July,2004','2004-07-22 00:00:00','   My hyper-predictive Ohio focus group, i.e., my Mom, watches a lot of TV.&nbsp; Typical middle America stuff.&nbsp; News.&nbsp; Weather.&nbsp; Soaps.&nbsp; (she tapes those)&nbsp; Tribe games.&nbsp; The occasional episode of 60 Minutes.&nbsp; Wheel of Fortune.&nbsp; Everybody Loves Raymond, Trading Spaces, Emeril....you know the drill.  In all this viewing, my Mom doesn\'t recall seeing a single political commercial since the primary season ended.&nbsp; Not one.&nbsp; The Cleveland, Ohio media market is&nbsp;flooded with&nbsp;political ads, and my Mom can\'t remember seeing any.  I was shocked, so I interrogated her.&nbsp; She watched the 60 Minutes which aired the Bush abortion hit piece...doesn\'t remember seeing the abortion ad.&nbsp; She watches the Today Show, perhaps the epicentre of all political advertising in Ohio, but has no recollection of any political advertising on the Today Show.  NEWSFLASH....The Ohio swing voter is TUNED OUT.  She  does  remember seeing \"on one of those news shows in the morning\" the satire animation by Jib Jab Films, a parody of Woodie Guthrie\'s This Land.&nbsp; ( urlLink which is hysterical, you should view it ).&nbsp;   By the way...she\'s still undecided.  Lesson for campaigns?&nbsp; You are certainly targeting my Mom with these ads.&nbsp; In fact, you may even think you are reaching her.&nbsp; You ain\'t.&nbsp; You need to reach these people in some way other than TV.&nbsp; Get creative.&nbsp; Funny seems to work.  '),(14410,1937270,'22,July,2004','2004-07-22 00:00:00','   OhioCountdown2004 has received exclusive access to the terms agreed between John Kerry and Dennis Kucinich.&nbsp; Dennis drove a hard bargain.  In exchange for the endorsement of Congressman Kucinich, who represents Cleveland\'s west side and western suburbs, and for the support of Kucinich\'s powerful block of 70 delegates, John Kerry agrees to the following -  1.&nbsp; In future comments, Kerry will not refer to Dennis as insane, unstable, megalomaniacal, bonkers, loopy, or confused in any way.  2.&nbsp; Kerry will avoid references to psychic seers, Shirley McLain, or the zen of trees, plants, rocks or other inanimate objects when referring to Kucinich.&nbsp;   3.&nbsp; Kucinich\'s endorsement does not apply to any past or future lives that he may or may not have lived or will live, either as a human being or other life force.&nbsp;   4.&nbsp; Kucinich is released of any obligation to stay pro-choice, which he has been for this entire presidential campaign, after having been pro-life for his entire political career.&nbsp; Use of the word \"flip flop\" or derivations thereof to refer to Kucinich\'s position on abortion,&nbsp;whatever that position may be in the future,&nbsp;is strictly prohibited.  '),(14411,1937270,'21,July,2004','2004-07-21 00:00:00','   In yet another sign of Ohio\'s importance in this year\'s election,&nbsp;a plane load of cheap hookers,&nbsp;its fuselage painted with \"Bush Cheney 04\",&nbsp;landed at Cleveland Hopkins International Airport last night.&nbsp;   Over 400 prostitutes pranced out&nbsp;of the jetway&nbsp;and immediately started targeting undecided voters in the concourse.   \"I\'ll take the free blow job,\" said Frank Climonsky, a Parma, Ohio voter, one of the first to be hit with the new persuasion tactics outside the Starbucks in Terminal C.&nbsp; \"But I think I\'m still...un...decided...how many girls got off that plane?\"   Heather Kopke, spokeswoman for the gaggle of whores, defended the controversial practice.&nbsp; \"This is a good way to defend marriage,\" Heather said as she mounted an undecided voter on the baggage carousel.&nbsp; \"A little variety here and there...\"&nbsp; When asked why she supported the Bush Cheney ticket, she responded, \"It\'s a gig.\"   RNC Chair Ed Gillespie noted to CNN this week that no expense will be spared in Ohio.&nbsp; \"If we have to send a call girl slut to every undecided, pro-life, pro-gun, obese and jobless white male in Ohio, we\'ll do it.\"&nbsp; When asked how they might target undecided women, Gillespie noted, \"These girls go both ways...saved us a bundle.\"  '),(14412,1937270,'21,July,2004','2004-07-21 00:00:00','   Thanks to everyone who\'s posted comments.&nbsp; Some responses...   On the Bush abortion ad running in Cleveland...   If you are a pro-life Catholic in Cleveland, and you voted for Bush in 2000 because of it, you are in the Bush base.&nbsp; In fact, you are also in the Reagan base, the Bush the Elder base, the George Voinovich and Mike DeWine base (our two undistinguished senators), etc.&nbsp;  It\'s called being a Reagan Democrat. &nbsp; If you are being targeted with ads on abortion, Bush thinks you\'re leaving that base.&nbsp;   You may also vote for Dennis&nbsp;Kucinich (until recently a pro life&nbsp;Democrat), and other D\'s on the ballot, while you vote for R\'s elsewhere.&nbsp;  That\'s called being a ticket-splitter .&nbsp; Happens all the time.&nbsp; &nbsp;  As one reader pointed out, it is impossible to find a single Gore voter (pro life or otherwise) who will vote for Bush in 2004.&nbsp; Gore\'s pro life&nbsp;voters may also be ticket splitters, but this year, they aren\'t going anywhere.&nbsp; Bush\'s pro life&nbsp;ticket splitters are in play.&nbsp; That\'s a base problem.   On Dick Fuck Cheney...   One reader posted ...\"Interesting assertion, given that the right has been referring to the long faced one as \"John Fucking Kerry\" for about a year and a half, after his little bout of foul mouthed rhetoric in Rolling Fucking Stone. I guess that makes your whole premise . . . I don\'t know . . . lame?\"    &nbsp; Lame?&nbsp; You know what\'s fucking lame?&nbsp; The fucking VP of the fucking USA telling a Senator on the Senate floor to fuck off.&nbsp; What\'s more lame is the fucking right leaping to his defense with their best hypocritical sancti-fucking-monious self fucking righteousness.&nbsp;   Don\'t you have a marriage to defend somewhere?  '),(14413,1937270,'19,July,2004','2004-07-19 00:00:00','   Word is the Kerry Campaign has hired a new person to run the operation in center-of-the-political-universe Cuyahoga County, (the Cleveland area for you non-Ohioans).&nbsp;&nbsp; Is it another false start?   urlLink Mark Griffin, currently a partner at the law firm of Hahn, Loeser & Parks , will be added to the staff in Cuyahoga County.&nbsp;&nbsp; Mark joins Vel Scott, whose appointment and tenure has been lackluster at best, counterproductive at worst.  Is this a staff shake up?&nbsp; Some say Mark will be&nbsp;coordinated campaign director (which would mean he\'s running the operation for the whole ticket), some say he\'s Kerry\'s west side guy, which would mean Vel Scott is responsible for the East Side.&nbsp; Some say he\'s reporting to Vel.&nbsp; Some say Vel reports to Mark.&nbsp; Some say Vel reports to no one, other than her chief sponsor Congresswoman Stephanie Tubbs Jones.&nbsp;   Why is this important?&nbsp; The balkanization of Cleveland politics (white v. black, west v. east) manifests itself every 4 years in staff decisions like this one, laying the ground work for paralysis by October in&nbsp;the most important voter turnout operation in Ohio, the Democratic Party, the USA, and therefore, the whole fucking world.  Bet you never thought you\'d care who Mark Griffin reports to, did you?  '),(14414,1937270,'19,July,2004','2004-07-19 00:00:00','   An ad on 60 Minutes is the most expensive piece of advertising real estate in prime time.&nbsp;&nbsp; If you place a buy on 60 Minutes, it had better be effective.&nbsp; On that measure, the latest Bush ad buy in Northeast Ohio is a complete waste of money, in more ways than it is possible to count. &nbsp; This Sunday night, Bush Cheney aired a hit piece on abortion on 60 Minutes.&nbsp; The ad attacks Kerry for opposing parental notification laws, and alleges he supports distribution of the morning after pill to high school students. &nbsp; Whatever the truth or merits of this argument may be, to air a pro-life&nbsp;attack piece on abortion in the most highly Democratic (i.e., pro choice) media market in Ohio is suspect logic.&nbsp;  &nbsp; The only possible explanation is that Bush continues to have a base problem.&nbsp; Pro-lifers need to be revved up, get back into the Bush base, and get mobilized.&nbsp; How ever many of them there are in Northeast Ohio, they need to have some gasoline thrown onto their fire.&nbsp; Makes sense, right? &nbsp; It would...if it was done right.&nbsp;&nbsp;If you are going to rev up pro-lifers, you don\'t complain about pansy-ass issues like parental notification and RU-486.&nbsp; You go after Kerry on&nbsp;partial birth abortion.&nbsp;&nbsp;&nbsp;That\'s like throwing raw meat at Pat Robertson.&nbsp;&nbsp; This ad is milk toast in comparison. &nbsp; Yet another sign that Bush Cheney is in complete panic.&nbsp; They are losing their base, trying desperately to regain it, and doing so with complete incompetence, wasting money on the most expensive TV time available with an ad that will not work.  '),(14415,1937270,'17,July,2004','2004-07-17 00:00:00','   The first signs of true panic among Republicans are starting to show among members of Congress and Senators up for re-election.&nbsp; Ohio\'s Senator George Voinovich, who up to now has been seen as a shoe in for re-election in 2004, this week  urlLink blasted the Bush administration on trade, from the Senate floor itself.  &nbsp; Why?&nbsp; Because according to Voinovich, the Bush administration has not been doing enough to stop Ohio from \"bleeding jobs.\"&nbsp; John Kerry couldn\'t have said it better himself. &nbsp; Pretty startling.&nbsp;  &nbsp; It\'s July.&nbsp; The election is more than 3 months away.&nbsp; Voinovich is as safe a Republican Senator as they come.&nbsp; And he\'s jumping off of GW\'s sinking ship as fast as a rat off the Titanic. &nbsp; Maybe Voinovich isn\'t so safe, after all? &nbsp; &nbsp;  '),(14416,1937270,'15,July,2004','2004-07-15 00:00:00','   Score one for the OhioCountdown2004...fuck yeah!  Paul Fucking Begala on CNN\'s Fucking Crossfire has started referring to \"Dick F-Word Cheney\" as often as fuck.    You will recall, we at OhioCountdown2004 began the campaign to christen the fucking VP as Dick FUCK Cheney some time ago.   urlLink (See the post)   Glad to see it\'s catching on.  Hey Begala...grow some fucking balls!  The name is Dick FUCK Cheney!  It sounds better if you use the actual fucking word...which is \"fuck.\"  C\'mon...they\'ll bleep it out, don\'t worry...that\'ll actually be funnier, anyway.  If FUCK is good enough for Cheney to use on the Senate floor, FUCK is good enough for bloody fucking CNN.  '),(14417,1937270,'14,July,2004','2004-07-14 00:00:00','   urlLink As covered by The Onion ...\"Neil Armstrong\'s Historic First Words On Moon: HOLY LIVING FUCK!\"  '),(14418,1937270,'13,July,2004','2004-07-13 00:00:00','   Jerry Springer, Ohio Democrat of The Year, and delegate to the convention in Boston, will be covering the convention for the local TV news version of the National Enquirer.   urlLink Story here.  '),(14419,1937270,'13,July,2004','2004-07-13 00:00:00','   The utter impotence of labor unions in political campaigns might be the Democrats\' best kept secret.  Thanks to the importance of the 2004 election, we are beginning to measure just how worthless labor support actually may be.  Dick Gephardt\'s entire campaign for president was based on labor unions.  We now have a precise measurement of what that gets you...a distant fourth place in the Iowa Caucuses, and an early exit.  He still road that wave of supposed labor muscle to become a front runner in the veepstakes, analysts even going so far as to say that his labor credentials would help deliver Ohio.  Laughable.  Why?  Ask anyone who\'s organizing labor union members this year.  Unions focused on getting out the vote among their memberships this year have been conducting internal polling to target their efforts. (Teamsters, SEIU, AFSCME, etc.)  Talk about an eye-opener.  Turns out that on average,  only about 40% of all union members are registered to vote.   Of that, only about half actually vote Democrat.  Which means that there is a  20% chance that a union member will vote Democrat this year.     If you whittle it down a little more, focusing on likely voters, you come up with the headline that a   union member is less likely to vote Democrat than the entire population as a whole.     Exactly what labor clout are we actually talking about?  '),(14420,1937270,'13,July,2004','2004-07-13 00:00:00','   As the Cavaliers/Boozer fiasco descends into the shameful, one wonders where it lands among the all-time Cleveland sports nightmares.     urlLink ESPN has a forensic examination of the top 10 most tortured moments in Cleveland sports.   It is a must read for John Kerry and George Bush if they want to win Ohio, a litany of misery so profound and head spinning that it lands Cleveland at THE TOP OF THE LIST OF TORTURED SPORTS CITIES.    We finally win something.  Yay.  A good sign for Kerry...when his wife introduced him in Cleveland, she noted she was from Pittsburgh...which drew an enormous \"BOOOOO\" from the huge crowd.  Kerry tenderly placed his hands on her shoulders, and turned her around toward the Cleveland Browns stadium standing behind her.  Nice touch.  '),(14421,1937270,'12,July,2004','2004-07-12 00:00:00','   Want to have your intelligence insulted?  C\'mon!  It\'s fun!  Read Elizabeth Auster\'s column in the Sunday PD, and you will literally feel yourself getting dumber by the second.  Ms. Auster...I\'m trying not to laugh here...spent an entire column waxing philosophic about John Kerry\'s hair.  The length of it.  What it means about America\'s struggle to come to grips with 60\'s era war activism.  How smart Kerry was to get it cut short....STOP ME!  I CAN\'T WRITE THIS ANYMORE!   urlLink Here...read it yourself.  She\'s serious.   This isn\'t some joke column...she really means to say that Kerry might have offended the bald vote by saying he and Edwards have better hair than Bush Cheney....FUCK!  I DID IT AGAIN!  STOP ME NOW!!!!  Read it and cringe.  Profanity will result.  I guarantee it.  I can\'t wait for the column lamenting the lack of focus on \"real issues\" in this campaign, how petty it\'s all become, and how Ms. Auster is so, so, offended by it, and how voters should pay more attention to columns like hers to inform themselves in such an important election season.  Hey Lizzy...we aren\'t watching a football game at your sister in law\'s house.  You can\'t choose who you like based on the color of their uniforms, or the length of their hair.  Get with the program, sweetheart.  '),(14422,1937270,'09,July,2004','2004-07-09 00:00:00','   At the Kerry-Edwards rally on Wednesday in Cleveland, everyone had to go through airport-like security...unless you showed up late.  At the last minute, just before the candidates began speaking, the huge line of people waiting to get in was simply waved through the bank of metal detectors and rushed in, the metal detectors going off like slot machines.    It appears that this eventuality was planned for.  A pen had been laid out with barriers and was quickly filled with people pressing to the front for a better view.    None of these people were searched or screened.  Most did not have a ticket.  Most had a clear view of the podium, from a distance of about 30 yards.    I\'m all for tight security at presidential rallies.  But if you are going to have a multi-tiered ticketed event, with multiple banks of metal detectors, individual searches of attendees, even to the extent of not allowing umbrellas into the rally, then YOU HAD BETTER DO IT FOR ALL THE ATTENDEES.    Hundreds of unsearched, unscreened people were essentially allowed...no...encouraged, to breach security.  In a post 9/11 world, this is simply not acceptable.  '),(14423,1937270,'08,July,2004','2004-07-08 00:00:00','   Got an interesting email from the DNC today.  They\'ve seen this blog, and after observing the Meetup saga played out in these venerable pages, decided to contact me with some good news.  Apparently they strongly support meetups, even have  urlLink a meetup webpage ...they even sponsor the one in Cleveland.  Hmmm.    You may recall, the Cuyahoga County Democratic Party and Kerry Campaign staff in Cleveland have previously asserted that Meetups are \"unauthorized\", threatened that they might be illegal, and have ignored the existing organization of more than 400 volunteers in a turf war over control of the campaign.  Can we please figure this out?  We just had a great rally for the Kerry-Edwards Ticket (security lapses aside - see previous post), momentum is building, and we still have a pissing contest going on within the party.  Fix it.  '),(14424,1937270,'08,July,2004','2004-07-08 00:00:00','   All the focus groups in the world will not help political campaigns understand the Ohio voter (especially the NE Ohio voter) better than standing in the shoes of a Cleveland sports fan for 5 stinking minutes.  Cleveland sports heartache is constant, repeated, severe, and mind numbingly predictable.  It\'s most virulent strain is the kind where you think Cleveland sports glory is at hand...only to watch it snatched away in the most conviluted, bizarre fashion.  Latest disaster?   urlLink The Cavs will probably lose Carlos Boozer thanks to a moronic decision by Cavs management.   You probably don\'t know who Carlos Boozer is.  Probably don\'t care, either.  If you\'re running a campaign here in Ohio, might be a good idea to find out.  '),(14425,1937270,'08,July,2004','2004-07-08 00:00:00','   Remember Al D\'Amato?  Former GOP Senator from New York?  Foul mouthed partisan?  Got his clock cleaned by Chuck Schumer?  Yeah, that one.  Apparently, Dick Fuck Cheney has gotten too offensive even for Al D\'Amato, who wasn\'t the most tactful fella in his day, and who now suggests GW take Cheney off the ticket.   Hey Dick Fuck...stay right where you are.  We love you.  We need you.  America needs you as Vice Fucking President.  Please stay.  '),(14426,1937270,'07,July,2004','2004-07-07 00:00:00','   I\'m stealing one from Dan Savage.  You may recall, when Rick Santorum went way overboard with homophobic idiocy, sex columnist Dan Savage decided to create a new and appropriate definition to fit the senator\'s distinctive last name, in an attempt to immortalize Santorum in homosexual lore. Check out the result at  urlLink Spreading Santorum. ; the campaign was so successful, that website is now the first hit when you do a Google search on Santorum.  In that spirit, my new name for the vice president, reflecting his tact, diplomatic skills, and overall value as a human being, is Dick Fuck Cheney.  Nice ring to it, eh?    Try it out...works best if you are in, say, the halls of Congress, one of those corridors in the Cannon office building with lots of echo and reverb, emphasis on the middle name...  '),(14427,1937270,'07,July,2004','2004-07-07 00:00:00','   The Plain Dealer\'s coverage of the Kerry-Edwards rally defaulted to low rent cynicism over the planning of the event.  Diane Solov, displaying Sherlock Holmes-like insight, called it an \"orchestration of electoral grandeur, designed to the last detail...\"   Ms. Solov clearly has spent many years honing a blood hound instinct for news, uncovering that \"getting the right play on the evening news was the campaign\'s grand plan...\"  Genius.    The real news is how something the size of this event is able to even come off at all.  The ticket was announced 24 hours before the rally.  The security for a presidential rally in a post 9/11 world was tighter than an airport.  Anyone who\'s ever been involved in planning such an event knows this rally was no small task.  The Kerry campaign takes a lot of heat in Ohio (including from this blog), but this event was as good as it gets.  Instead of informing readers certain to be curious how such a feat can be pulled off on 24 hours notice, the PD knee jerks into cliche.  Nice touch, guys.  '),(14428,1937270,'07,July,2004','2004-07-07 00:00:00','   Click on any picture for a larger version.   urlLink    Tim outside the rally.   urlLink    Tim inside the rally.   urlLink    Senator John Glenn and the candidates.   urlLink    The candidates.   urlLink    Senator Kerry at the podium.   urlLink    After the rally.   urlLink    Senator Kerry.  '),(14429,1937270,'06,July,2004','2004-07-06 00:00:00','   The more this campaign goes on, the more it looks like 1992 all over again.  CNN\'s coverage of the Kerry-Edwards Ohio visit included a debate between Ohio Democratic Party Chair Denny White, and Ohio Republican Party Chair Robert Daddy-Warbucks Bennett, on InsidePolitics.  Bennett looked like an 1897 political cartoon robber baron, pried into a chair by a shoe horn because he\'s so obese he can\'t fit himself into it unaided.  The only thing missing was the shiny top hat and monacle, and a few ragged peasants at his feet begging for a shilling or two.  Laughing and smiling with that arrogance only Republican fat cats can summon, he declared that all is well in Ohio...what recession?  Apparently, he\'s missed the weekly...repeat...WEEKLY...plant closings all across NE Ohio.  The Bush Cheney Team continues to talk right past the voters, out of touch with reality...just like 1992.  '),(14430,1937270,'06,July,2004','2004-07-06 00:00:00','   Republicans are amazing.  Not 24 hours go by, and GW is already wallowing in the gutter.    When asked how John Edwards compares to Dick Fuck Cheney, GW said, \"Dick Cheney can be president.\"  Hey George...last time I looked at the constitution, the guys with the most votes get to be president and vice president.  Maybe they have the fix in already?  '),(14431,1937270,'06,July,2004','2004-07-06 00:00:00','   Had to be the biggest rally of the campaign season to date.  Cleveland always knows how to put on a rally for a Democrat.  In 1992, one of the first stops on the post convention bus tour for Clinton/Gore was only a few blocks away from today\'s rally...had a similar feel to today\'s rally, too.  Lots of optimism, confidence, and determination.  Bush is so fucked, man.  '),(14432,1937270,'06,July,2004','2004-07-06 00:00:00','   How will the Kerry-Edwards ticket play in Ohio?  As usual, check with my mom.  A super-predictive Ohio voter, she never voted in a Democratic primary in her life...until this March, when she asked for the Democratic ballot to vote for John Edwards.  Ever since, I\'ve asked her repeatedly what she would think if Edwards got the veep spot, and she\'s been coy.  Hard to get anything out of her on this, other than a wry smile.    But she does smile...and yesterday when the word started to leak out, she almost seemed...dare I say it... excited  by the news.  Stay tuned.  '),(14433,1937270,'04,July,2004','2004-07-04 00:00:00','   On his fucking visit to fucking Parma Saturday,  urlLink Dick Cheney sat down with some fucking reporters .  When asked about MoveOn.org and America Coming Together, two 527 organizations that are just fucking hammering Bush/Cheney in Ohio, Cheney declared,   \"I think these 527s are - I\'m trying to think how to say this in a statesmanlike fashion - are somewhat disreputable, obviously. . . . I think they are loud and obnoxious...\"  The Plain Dealer noted that \"Questions and answers were edited for clarity and brevity.\"  Wonder what fucking words they had to take the fuck out of Cheney\'s fucking remarks.  '),(14434,1937270,'03,July,2004','2004-07-03 00:00:00','   Declaring Ohio \"perhaps the most closely watched swing state this year,\"  urlLink The New York Times  again has placed the Buckeye state front and center in yet another Sunday issue.  You may recall that Ohio was featured in a cover story in their Sunday Magazine a month or so ago.  Reporter James Dao also writes a bit on Bob Taft\'s weakness and the looming thermonuclear primary in the Republican party in 2006.  '),(14435,1937270,'03,July,2004','2004-07-03 00:00:00','   Ever hear of the  urlLink Concerts For Kerry  tour?  Great stuff.  Went to the second Beat Bush Bash at the Beachland Ballroom, saw some great bands (The Mammals, Pat Humphries & Sandy Opatow, Big Black Africa and Harvey), with a like minded group of lefties...all done to raise money for John Kerry.   urlLink Harvey Pekar  was there (Cleveland\'s own ambassador of reality - you might have seen him in   American Splendor  ).  His remarks were hilarious.  In typical Pekar-speak, here he is, in a room full of people who are attending a  Kerry fundraiser , and he ends his short remarks by saying, \"You know, just don\'t vote for Bush, man...\"  Greatness.  As usual, he may be revealing something about Cleveland and Ohio...you can\'t take any vote for granted.  '),(14436,1937270,'13,August,2004','2004-08-13 00:00:00','   Mom\'s postings are totally unprompted, unedited, and unscripted...which makes sense, because she\'s UNdecided. She gets a piece of mail, looks at it, then types what pops into her head. I\'ll type a bit more on each piece...that technical strategic mumbo jumbo all you political junkies crave.  Her blogs will be identifiable by font color...green...her favorite.  Also trying to make sure she writes, \"FROM MOM\" in each headline.   Questions, comments to Mom are encouraged. Also, feel free to forward Mom\'s thoughts to the candidates...they claim to be interested.  '),(14437,1937270,'12,August,2004','2004-08-12 00:00:00','   ACT goes after Bush on the prescription drug plan and other health care issues, with the theme \"Ohio Real Stories\", featuring a picture of a very Ohio looking middle aged woman on the front who may or may not be from Ohio, but who\'s been laid off. It also solicits contributions. The piece arrived on Thursday, Aug. 12. It was a square glossy, folded, four panel, full color piece...not cheap. Paid for by ACT.  Targeting could either be at women or geographically based on a ticket splitter index.  If anyone got this piece, please comment. We\'re also looking for volunteers to scan their Ohio pieces and link to them...scannning was above our pay grade.  '),(14438,1937270,'11,August,2004','2004-08-11 00:00:00','   It is sad that we Medicare patients have to go to CANADA.... If Bush and Kerry would contribute their campaign money to the prescription plan instead of traveling all over the country we Medicare patients would not have to go to CANADA....  '),(14439,1937270,'11,August,2004','2004-08-11 00:00:00','   Someone must be reading this blog...only days after noting that the perfectly predictive Ohio swing voter, Mom Russo, has yet to see a piece of campaign literature or direct mail, THE FIRST PIECE ARRIVES!   OhioCountdown2004 is taking this opportunity to launch a new series of posts with a new contributor...none other than Mom Russo herself!  (we hope...computers and such are not Mom\'s cup of tea...this could start and end rather quickly)  Mom is going to comment (again, keep your fingers crossed) on   all the direct mail that comes to her home, from both campaigns.    We may even try to scan them in so you can see them yourselves...technology permitting.  If things get real crazy, we\'ll ask Mom to comment on TV ads, phone bank calls, maybe even door to door campaigners (if they show up).  Meanwhile, though, as we await the first post from our super-swing-voter, some thoughts.  It\'s very early in the game for a persuasion piece to hit; usually these things start going out in October.  And a little teaser...it\'s paid for by America Coming Together (ACT), the granddaddy of all 527\'s.  '),(14440,1937270,'11,August,2004','2004-08-11 00:00:00','   Trying out a new template on the blog...the white background is a bit easier to read and print.  Got some complaints that the blue background was unhelpful for printing.  Comments welcome.  '),(14441,1937270,'11,August,2004','2004-08-11 00:00:00','   If the Three Stooges did an episode where Curly ran for President, it would look a lot like the Cuyahoga County Democratic Party.   We are now three months away from the election, and still, in Cuyahoga County (the Cleveland area), the single most critical county in the state for Democrats, callers to the party office face full and disconnecting voice mail boxes, unanswered phones, unreturned phone calls, and an incompetent paralysis that would be comic if it weren\'t for the fact that the entire election may be effected by it.  It sure   looks   like a campaign...the Kerry campaign finally got the hell out of the black hole of the party office and opened it\'s own office in Shaker Square on Cleveland\'s east side, complete with new staff, signs, bumper stickers, etc.  The phones seem to work there.  But volunteer forms were hard to come by because...you\'d better be sitting down for this one...as one volunteer explained it, no one ever calls the volunteers after they fill out the forms anyway, so they might as well not hand out forms to fill out, because when the volunteers fill them out and never get called, they get frustrated and complain about how nothing is going on in Cleveland.   Shoot me now.  Some volunteers do get called...not to do any phone banks, or lit drops, or to call other volunteers...but to show up at a Bush rally and heckle.  It seems the only time the volunteer forms see the light of day is when GW Bush is in town.   The sheer weight of endless volunteer walk-ins, their numbers higher than at any time since the 1992 Clinton campaign, does appear to be forcing some activity...a phone bank was sighted.  Thank God for the momentum of Bush hatred...it\'s overcoming organizational incompetence of mind boggling proportion.  '),(14442,1937270,'10,August,2004','2004-08-10 00:00:00','   urlLink The Organization for Security & Cooperation in Europe (OSCE)  has been invited by the State Department to send an observation mission for the Nov. 2 election in the US.  It\'s tempting to see this as a cave in to pressure from D\'s in Congress.  Think again.  The State Dept. jumped at this because it isn\'t a threat.   Anyone who has participated in any OSCE election observation mission (I\'ve been on three) knows that the verdict of the mission will be written by OSCE member state ambassadors, who have political agendas to grind.   Guess who is the most important member state of the OSCE - the US.  Guess who appoints the ambassador to the OSCE - GW Bush.  In January, I wrote a piece for the  urlLink London Sunday Times  on the impending OSCE election observation mission in the former Soviet Republic of Georgia, warning of a whitewash (which happened.)  It\'s a good primer on the OSCE\'s internal processes.  I pasted it here (you can\'t get it online unless you subscribe to the Sunday Times.)   Georgia vs. the election observers.   By Tim Russo       \"You\'re joking,\" my British friend whispered to me as he came to a stop after racing into the building with urgency.    We stood at the back of an old auditorium, packed with international press, government officials, embassy staff, and observers of the 1998 presidential election in Armenia.  The head of the Organization for Security and Cooperation in Europe (OSCE) international observation mission sat at a table in the front, calmly reading the delegation\'s preliminary statement on the conduct of the election the previous day.  I looked down at the ground in stunned silence.  \"I just got back from my count,\" my friend said in disbelief.    \"Have you even showered?\" I asked.  \"Haven\'t showered in a day and a half,\" he said.  \"We went to the OSCE office, they told us to come here,\" he said.  \"I just got out of the car after an eight hour drive,\" he finished, panting.   I feared what was coming.  He\'d just been an observer at a late night vote count in a remote, mountainous border region where military manipulation of the election was expected, and turned out to be particularly egregious.  \"Where do I take these?\" he said waving a stack of notes and documents in the air, his voice now frustrated.  I just looked at him and shook my head, having spent the previous night at an equally farcical vote count myself until 9 a.m. that morning.  He started looking around for someone in charge.  \"What is going on here?  Why is the statement being issued, I haven\'t even reported from my assignment?\"  I just rubbed my forehead as my friend kept charging about the back of the room with his documents, his desperation telling it all about how the election must have gone where he\'d observed it.  Then we heard from the podium, \"...a step forward for democracy...\"    My friend lost it.  He stopped in his tracks.  I walked over to try and calm him down, but he just looked at me in shock, shoving my outstretched arm away.  \"What a fucking joke...,\" he said with one last wave of his notes, then he threw them to the floor angrily, turned around, and marched out of the building in disgust, as the head of the delegation calmly continued reading his verdict of approval.  Since the soviet collapse, it seems every new election in the post soviet republics deteriorates into more of a sham than the one before it, 2003 proving a banner year for post soviet electoral fraud.  This year, the OSCE, post soviet international election observers of record, condemned various elections in Armenia, Azerbaijan, Georgia, and even Russia, whose December 7 parliamentary election it described in its preliminary statement as `overwhelmingly distorted\'.   The people of Georgia, however, finally refused to accept yet another fraudulent election.  Helped along by the condemnation of election observers (OSCE as well as domestic observers), the people power in the streets of Tbilisi in November forced out the president, Edward Shevardnadze.  Enough had become enough, and now there is great hope for a new era in Georgia of elections free of manipulation, beginning with the presidential election January 4.  As my British friend would agree, Georgia need look no further than its rugged southern border for guidance in the days ahead.  In 1998, Armenia faced very similar circumstances.  A president who relied on electoral fraud to stay in power was forced to resign.  A new election was scheduled, the OSCE descended en masse to observe.  And a similar hope for a future of free and fair elections was in the air.  But Georgians should watch the OSCE as carefully as they watch their election processes, for the OSCE\'s recent boldness in the aftermath of this year\'s many sham elections is very recent indeed.    Run by member states, OSCE observation missions exist in a diplomatic no mans land; on the one hand dedicated to promote integrity in the democratic process, but on the other, representing the political and economic policy interests of the member states.  The result is that far from being disinterested guarantors of democratic integrity, OSCE election observation missions, and especially the statements they make in the immediate aftermath of an election, are not unblinking verdicts by a referee, but more often are delicate balancing acts of high diplomacy.  It is a balance the OSCE has often gotten exactly wrong at precisely the wrong moment.  As the new democracies of the former Soviet Union were finding their feet, and member states of the OSCE were assessing the new geopolitical realities, OSCE election observation missions found ever more creative ways to gloss over the ever more fraudulent elections.  Whatever the competing interest, the commitment to helping these new democracies create fair electoral processes regularly took a back seat, with the perverse result that messy elections were not only ignored, but made even messier by the OSCE and the increasing caricature of their increasingly predictable whitewashes.    Armenia\'s presidential election in 1998 was a prime example.  As an observer within the OSCE mission at the time, I had a front row seat to not only observe the pervasive electoral fraud, but also the OSCE\'s delicate balancing act between the compelling observations the observers kept reporting, and the member states\' interest in ignoring them.   The resulting OSCE verdict, a masterfully acrobatic navigation of statements and reports that managed the least amount of honesty when it was most needed, rubber stamped an election widely described by experienced observer delegates at the time as perhaps the most sophisticated electoral fraud they\'d ever witnessed.    The timing of the OSCE\'s abdication of its responsibility in Armenia in 1998 could not have been worse.  After years of preceding electoral manipulation, as a chance arose for democracy to change the course of a troubled country, the OSCE hailed a thorough fraud as a `step forward\' for democracy, writing off the blatant joke of precincts reporting 300% turnout as not causing them `to question the result\'.  The result in the years since has been the rapidly accelerating rot of Armenia\'s democracy.  The OSCE should be applauded for its honesty in 2003, but perhaps with a slow hand clap rather than a standing ovation.  Its talk this year is quite cheap.  For when it mattered most, when a credible international organization\'s verdict on the conduct of an election could have affected democratic processes for the better, the OSCE too often has blinked.  Elections in Armenia, Russia, Georgia, or the other former soviet republics, did not suddenly become `overwhelmingly distorted\' overnight and in secret; they\'ve been deteriorating predictably and in full view for more than a decade.  The OSCE\'s sudden, too little, and far too late honesty about electoral processes which are rotten to the core -  processes the OSCE itself has had a hand in perpetuating - might actually be comic if it weren\'t so tragic.  Georgia in 2004 may be different.  The OSCE did have an effect on the November election.  And if they are willing and able to stay as honest about the next one, Georgia may finally find the courage to not only refuse their legacy of electoral fraud, but work to eradicate it.It is likely, however, that the OSCE balancing act is in full swing already.  The likely next president, Mikhail Saakashvilli, is a friendly protÃ©gÃ© of the OSCE member states.  He will likely win in a landslide, whether or not there is any fraud, and the OSCE will want nothing to taint his victory.  Clean reporting of electoral manipulation may already be taking a back seat to the rise of a friendly new ally in a difficult and unstable region.  Fraud there will certainly be.  It is tempting to think that the people power in Georgia that refused the latest fraud can herald a new interest in fair and clean processes, that somehow the disappearance of a corrupt figurehead will result in the disappearance of the deep rooted structure of electoral manipulation that kept him there.    But the practices of looking the other way while a ballot box is stuffed, or taking advantage of miserably inaccurate voter lists, forging a signature here or a vote total there, bribing or blackmailing voters with meager amounts of money that are many times their monthly salaries, have a way of sticking around and metastasizing into a permanent cancer if ignored for too long.  The OSCE will then have a choice between issuing a statement, the first draft of which is likely already written, that reports what actually happens, or one that will predictably refer to these practices as not having `affected the outcome\', that despite these shortcomings, the result will have `reflected the will of the people\'.  It is a crucial moment for the OSCE, an organization that appears to be attempting to rehabilitate its credibility.  Perhaps the OSCE will finally get the timing right.  For Georgia, though, much more is at stake than institutional credibility.  It will soon have a new president who will likely bring hope for a new future.  Whether or not his election will herald a new, freer democracy, is an open question.  '),(14443,1937270,'08,August,2004','2004-08-08 00:00:00','   Sunday, August 8th\'s Columbus Dispatch has Bush at 47, Kerry at 44, Nader at 2, with a hefty 7 percent undecided.  Aside from the relatively large group of undecideds, here\'s the good news - more than three-fourths of undecided respondents said the country is on the wrong track, and more than 7 of 10 disapprove of the way Bush is handling the economy and the situation in Iraq.   The unorthodox Dispatch methodology does it again.  The poll was based on returns from 3,047 registered Ohio voters who said they intend to cast a ballot Nov. 2, with a margin of error at 2 percent.  AND...get this...it was taken July 14-23... before  the Democratic convention.  '),(14444,1937270,'07,August,2004','2004-08-07 00:00:00','   Been getting a little flak for taking shots at D\'s as well as R\'s.  Interestingly, one of the big blogs on the block,  urlLink mydd.com,  commented on the lack of moderate blogs in the upper echelons of the blogosphere.  Chris Bowers writes,  \"Blogging appears to be following a pattern of other pundit-heavy media outlets (talk radio, newspaper columnists, TV political opinion shows). In all of these cases, partisanship sells while mixing your views or straddling the fence does not.\"   Unfortunately, he\'s right.  Fellas...when you live in a state whose Democratic Party is as profoundly inept, for as long, with as much futility, as Ohio\'s, it becomes easy to be a moderate.  It took a ranting and raving Howard Dean to remind us all that Democrats need to be called out when they are fucking up.  In Ohio, Jerry Springer is the one coming to the rescue...need I say more?  It is our duty to identify and fix our party\'s problems.  If we don\'t, be assured, the Republicans will fix them for us...by beating our asses over and over and over.  '),(14445,1937270,'06,August,2004','2004-08-06 00:00:00','   The PD has compiled a list of appearances and major rallies in Ohio by Kerry and Bush... urlLink you can see it online , but like much of what the PD does, the story as it appears online is pretty lame.  In the \"actions speak louder than words\" category, the list tells quite a tale.  Bush indeed has a base problem in Ohio, at least according to his scheduler.  Of the 48 stops listed for Bush, 24 of them, half, are in the Cincinnati area, Republican heartland, and another in Lima, also R heartland.  Bush visited the key quarter moon area (see  urlLink How To Win Ohio ) 16 times, with 7 other visits split between Dayton & Columbus, key swing counties.  Of 34 total Kerry stops, Kerry visits the key quarter moon area 22 times, including 12 stops in Cleveland (the Democratic base), with 4 stops in Cincinnati, 4 in Columbus, and 4 in the Dayton area.  '),(14446,1937270,'05,August,2004','2004-08-05 00:00:00','   Ohio\'s Democratic US Senate candidate, State Senator  urlLink Eric Fingerhut , is embarking on a tour of Ohio...on foot. Eric  urlLink will walk from Cincinnati to Columbus to Cleveland . You can track his progress at  urlLink www.walkohio.com .  Eric is a great candidate, good speaker, smart guy, certainly deserves to be Senator...but this little trick has an unpleasant history to it. Last guy to try it was Dennis Kucinich (yes, he of the quixotic presidential run), back in 1986 or thereabouts, running (or shall we say, walking) for governor. He stopped about a third of the way through, for various reasons, then dropped out of the race...after having ridicule heaped upon him.  Despite this sad legacy, it\'s always been a good idea. It might work to get Eric on the evening news, might even capitalize on any potential Kerry momentum, if Eric manages one thing ...finish the walk . Eric...YOU MUST NOT STOP MID-HIKE. I don\'t care how many blisters you get on your feet, how many times your wife calls on the cell phone, you have got to finish the thing, or you will find yourself in the Kucinich universe of Democratic silliness and futility. In Ohio, that crowd is already big enough.  OOPS...forgot one thing.  You know what, ditch the cell phone for the duration.  Many unpleasant things will flow from pictures of you on the trail constantly pressing a phone to your ear (which will happen, if you have it with you).  '),(14447,1937270,'05,August,2004','2004-08-05 00:00:00','   The perfect predictor of Ohio voting patterns, undecided Ohio swing voter Mom Russo, caught a story on the news about Kerry and Bush holding events at the same hour in Davenport, Iowa. The story noted that the mayor of Davenport, a Republican, is still undecided.  The reporter said to the mayor, \"You\'re one of the most prized commodities in the country right now.\" Always ready for opportunity, Mom responds, \"So am I! I\'ll take a million bucks for my vote!\"  She asked me to let the campaigns know she\'s available to the highest bidder, and can\'t understand why they haven\'t contacted her yet.  This is interesting...given all the effort both campaigns are putting into targeted swing state undecideds, my mother still can\'t identify any campaign ad or lit or mail piece that has reached her.  Swing voters are maddening.  '),(14448,1937270,'04,August,2004','2004-08-04 00:00:00','   Websites about Ohio\'s role in the 2004 presidential election are popping up like dandelions. Two that have caught my eye are  urlLink MakeOhioBlue , and  urlLink Bring Ohio Back .  Make Ohio Blue is the latest incarnation of the JerryForOhio.com website. Yes, that Jerry...Springer. Make Ohio Blue is essentially the Springer website and blog, re-tooled for 2004. Bring Ohio Back is a 527 organization enlisting celebrities to raise money for campaign ads in Ohio.  It gets interesting when you go to the About section of each site.   About Make Ohio Blue      Make Ohio Blue is sponsored by Jerry Springer and Jerry For Ohio.com. Jerry has worked tirelessly for the Democratic Party, and has contributed funds to the Democratic Party and its candidates for decades. He is determined to help revitalize the Democratic Party in Ohio.    About Bring Ohio Back      (after a paragrpah-and-a-half of spin)...Northeast Ohio-based professional strategists, with decades of success in this region, will direct these activities. We will use our first-hand knowledge and expertise to craft the right message for Ohioans, and develop the best tools to reach the public.     Hmmmm....\"professional strategists\", the \"right\" message. The Plain Dealer recently reported that Bring Ohio Back is a brain child of Burges &amp; Burges, the Cleveland political consulting firm.  Where does the money go? The 527 (i.e., B&B) is raising money mainly for broadcast ads in Ohio, the PD story citing $1 million as the goal. The \"professional strategists\" will keep 15% of that for media buy commission, i.e. $150,000 to B&amp;B\'s pocket. At Make Ohio Blue, if you donate, it doesn\'t go to Jerry Springer\'s campaign...it goes directly to the DNC.  More interestingly, the Bring Ohio Back site is already taking shots at Springer on the front page, linking to an article headlined \"Re-Elect Bush And You Might Get Jerry Springer, Too.\" I\'m guessing B&amp;B has made a professionally strategic decision to work for one of Springer\'s likely primary opponents in 2006...I\'m sure they\'ll have the \"right\" message then, too.  '),(14449,1937270,'03,August,2004','2004-08-03 00:00:00','   It is really quite amazing how much early polling is readily available this year for Ohio.  Just spend 10 minutes roaming the internet for polls...you will be overwhelmed.   urlLink There\'s a good site that brings swing state Zogby polling together, check it out.     In Ohio, only one poll really matters (aside from the actual voting).  The Columbus Dispatch does a poll the Sunday before the election (which this year is October 31, Halloween...make of that what you will).  It is eerily predictive...down to tenths of a percentage point.  The Dispatch poll\'s precision is a mystery.  It is never wrong.  Plus, it is a mail in poll...yes, the US Postal Service.  No one really knows how it works...but it does.   So when you\'re looking at polling data for the next 3 months, just hold on to your hat...and log onto  urlLink www.dispatch.com  on Halloween...that will tell the tale.  '),(14450,1937270,'01,August,2004','2004-08-01 00:00:00','   Many among the landed punditry have been busily prescribing how Kerry or Bush could win Ohio.  Who\'s right?  There\'s really only one way.  Take a pencil and a map of Ohio.  Start at Toledo.  Draw a line from Toledo, eastward across the southern shore of Lake Erie to Cleveland, then diagonally, passing equidistant between Akron and Youngstown to East Liverpool on the West Virginia border, then along the Ohio River to Steubenville, Marietta, and then start back westard again to Athens.  Stop.  Your drawing should look like a comma, or a quarter moon.  If you are able to drive through every county on that line without touching Republican ground, Democrats win the state.  Period.   It\'s easy to get caught up in the major D counties along that line - CuyahogaCounty, containing Cleveland, the Democratic mother lode, Lucas County (Toledo), Summit County (Akron), Mahoning & Trumbull Counties (Youngstown area).  But as more second tier counties along that line go D, the closer the result gets, and the greater chance for a D victory.  The more trouble you have connecting that line with D counties, the likelier it is that an R wins.   You hear so much about Stark County (Canton) and it\'s predictive value because it doesn\'t sit on that line precisely...it\'s just to one side of it.  Whichever way Stark goes, that means the other counties straddling the line (which look a lot like Stark County) are trending that way too.   Much is made of Franklin County (Columbus) and Montgomery County (Dayton).  Don\'t be distracted, they aren\'t predictive so much as a lagging indicator.  It has recently become possible for a D to win both Franklin and Montgomery Counties and lose the state.  However, if the key line goes solid D blue, nice and thick to both sides, you may carry those two swing counties as gravy.  Watch the candidate schedules.  This weekend, Bush and Kerry were in counties bordering or on that line the entire weekend.  Plotting their visits, one will find the stops start to fill in along that comma or quarter moon.  '),(14451,3247172,'31,May,2004','2004-05-31 00:00:00','   Off today - hope you are as well. We\'ll be back tomorrow. Have a safe and happy Memorial Day.  '),(14452,3247172,'28,May,2004','2004-05-28 00:00:00','   This morning I wanted to briefly discuss crafting a job description (needs analysis) when working with recruiters.  To help you - as a recruiter I will definitely want to know about the core competencies required for the job, experience level required, educational background etc and why the vacancy exists.  But I will also want to know who else is on the team and for how long, what are their strengths and weaknesses?  What would happen if one of them left...what is the next most critical skill set required?  A good recruiter will ask a LOT of questions.  We want to know and understand what you really need.  Sometimes you won\'t know and that\'s ok.  We work together to find out.    So the next time you have a vacancy begin to think about other members of the team, their core competencies and strengths, so the new hire is a complement to the team instead of just the new hire.  Have a safe and happy holiday weekend.  Lucia  '),(14453,3247172,'26,May,2004','2004-05-26 00:00:00','   I spoke to a VP of Human Resources for a large company last week who had signed on 15 contingent recruiting firms to assist his in-house recruiters to fill quite frankly, a lot of new jobs.  He was becoming frustrated because of the amount of time updating all of these firms was beginning to take.  On top of that, many of the firms were overlapping with one another and beginning to submit the same candidates.  How is this helping the organization that hired on these firms?  I don\'t think it is. When trying to decide whether to go to a retained search firm or a contingent one...you have to think of many things - especially the value of your time.  Many firms offer a hybrid approach - we call them engagement fees.  You put a deposit down and pay the balance upon the new hire being made.  Our firm charges for example, a modest $295 down and it\'s applicable toward the total search fee upon placement.  Why do we do this?  We do it primarily to get your attention, to engage you in this process which goes very well provided there aren\'t 15 other firms involved.  I understand the need to work with more than one firm in the event we\'re talking about filling multiple positions. And some firms specialize in different areas as well.  But in selecting search firms, remember to work with those recruiters who work well independently - provide good updates and don\'t need to talk with you every day for a status update unless there is a need to do so.  By the way, as recruiters, it\'s our job to find out where our candidates are interviewing and where they\'ve been submitted whether we are retained or contingent.  It wastes an awful lot of time (worth money!) when we don\'t do our jobs well.....and it doesn\'t help you, the hiring managers in the least.  If this is happening to you - it might be time to talk with your recruiters or account managers.  Have a question?  Just ask... http://www.askhireworks.com  '),(14454,3247172,'25,May,2004','2004-05-25 00:00:00','   Morning,   I realize it may be difficult due to volume to confirm receipt and/or acknowledge each applicant who contacts your company.  But you MUST do this.  Your IT folks should be able to set up autoresponders to work with your ATS to make your life easier.   However, please put some thought into the verbiage used in these acknowledgements.  I\'m going to assemble some of the worst I\'ve seen and publish them soon.  Remember, some of these applicants will be people you are interested in and you leave a lasting impression on people with not only your first contact/correspondence but all of them.    Those of you with access to your marketing departments should really get them involved.  Strong verbs, empathetic - caring language - personalized messages (even addressing them by first name in autoresponder thanking them for their interest) should be used here.  I\'ll get some examples to show you and post soon.  Thanks for reading and make it a great day!  Lucia  '),(14455,3247172,'24,May,2004','2004-05-24 00:00:00','   I receive emails about once a week from LabSitesOnline.  They offer a $125 gift certificate usually to Walmart, Home Depot etc. if you purchase a job posting.  If you decide not to elect a gift certificate you get a discount on your posting and it\'s only $170 instead of the $295.    I find this interesting.  I don\'t think this appeals to most corporate clients, and wonder if this isn\'t having more success with independent recruiters (like myself).  I haven\'t personally posted a job and will not do so. I don\'t believe posting jobs on traditional or even niche boards is effective anymore.  Posting to a job board is like playing darts.  You have no idea really who will see your ad and how well the site is actually optimized. And you are dependent upon people contacting you. Most candidates who will contact you as a result of a posting will generally be those with issues.  Maybe it\'s a work history (lots of jobs all the time).  Maybe a license has lapsed.  Maybe they were terminated at one time.  I\'m not saying this is always the case....but usually there is something.   The best way to source candidates is still in my opinion by referrals from those who are excellent, above average performers.   But if anyone out there has tried this board...please share your feedback and I\'ll post for the group.    Best wishes,  Lucia  '),(14456,3247172,'21,May,2004','2004-05-21 00:00:00','   Morning!  He who interviews best gets the job? Hmmm, not sure I totally agree.  Let me explain.    I found a great site today by accident really.  A former pharma rep who became a recruiter is really offering a great value to potential candidates who want to break into the pharma sales field. I applaud Pat Riley (although I don\'t know him) for providing valuable information to interested candidates and putting in place a system for his business.  However, as a hiring manager, do you really want to hire someone who interviews well or who WILL DO WELL on the job?  That\'s not to say a good interviewer won\'t perform well on the job....but it can go the other way as well.  When interviewing a candidate, be sure you are asking questions that will give you an idea of future - predictable behavior.  Be sure this individual is motivated, excited and driven. Ask for specific examples of how they handled challenging situations in the past.  What happened, how did they react...if they didn\'t handle it well - but would have now ask why.  What did they learn as a result and what would they do differently now.  Ask them how a colleague and/or a customer would describe them.    What we are after is high performers - strong leaders.  Sometimes they may not always be the best interview...at least initially.  After a bit they will open up, relax..ask good questions and show you who they are.   Wishing you more fun and success with interviewing...  Lucia  '),(14457,3247172,'20,May,2004','2004-05-20 00:00:00','   urlLink'),(14458,3247172,'19,May,2004','2004-05-19 00:00:00','   I talk with candidates looking (and sometimes not looking) for new and interesting career opportunities every day.  Many of the candidates I talk with are leaving a current post because of a conflict or unhappiness with a manager they report to.  What does that mean to you Company XYZ?  Your most talented staff members may not be quitting you when they leave, but they are quitting their manager!    Of course this topic runs deep and is one we can discuss further so you can remedy it. But as turnover increases (and it will!) be sure you are conducting exit interviews to see WHY people are leaving.  You can outsource the function to make it confidential or have HR staff conduct them for you.    In the era we\'re in with labor shortages rapidly approaching (and worsening e.g. healthcare) you need to be able to invite former staff back into your organization.  Be sure to take the time to find out why they left in the first place and see if it\'s something that can be fixed. Asking questions is still the best way to learn what we do not yet know.  Feedback and comments welcome....  Lucia  '),(14459,3247172,'19,May,2004','2004-05-19 00:00:00','   urlLink The Hiring Insider  '),(14460,3247172,'19,May,2004','2004-05-19 00:00:00','   urlLink    HireWorks...helping you make better hires faster.&nbsp; urlLink  '),(14461,3247172,'19,May,2004','2004-05-19 00:00:00','   New, Improved - Hooray!   Hot Job!  Business Development Rep - CRO Experience Required - Work from home!   Iâ€™m pleased to be working with a client who was named one of Fortune\'s 100 Best Companies to Work for in 2004.    Weâ€™re seeking an innovative and motivated Business Development Representative to sell the services their research division has to offer clients worldwide.  Your role?  Expand a current customer base as well as establish new business in the field of clinical research.   Qualified candidates should have a minimum of three years experience in a CRO environment selling services (Phase I-IV). Strong project management skills and proposal writing skills are a must. The team includes 4 Business Development Associates each reporting to Director of the department. For profit organization, campus/academic setting.    Would you like to join a team of professionals in clinical research where there has been very little turnover, excellent support for research to prepare bids, quotations and proposals? Interested in experienced clinical staff to travel with you to sell your services?   Please send resume and cover letter to my attention at lshaw@hire-works.com.  Tell me about the types of projects you sell and service, therapeutic areas of interest, as well as what preferences you may have.   Compensation plan includes generous base PLUS bonus and full benefits package.  Should you choose to relocate to client site, relocation assistance is provided.  If you have a friend or colleague who may be interested, please feel free to have them get in touch with me.   Lucia Apollo Shaw Recruiter lshaw@hire-works.com   Have a question? www.askhireworks.com  HireWorks, Inc. www.hire-works.com   Toll Free: (866) 960-3411  Phone:    (919) 960-3411  FAX:       (919) 960-3413  '),(14462,3247172,'18,May,2004','2004-05-18 00:00:00','   Niche Boards vs. the Big Boards  I received an email today from topusajobs.com stating niche boards hiring more than 10x the number of candidates than do the big boards (e.g. www.hotjobs.com).  Well I should hope so!  The big boards account for LESS than 6% of hires made on average.  That figure is so low it\'s not worth the time you take to set up an account and post the job.    This email was promotional however, it did attempt to list a diversified list of sites, among them www.latpro.com, a site focused on diversity recruiting - specifically Latino candidates.  Please submit your favorite niche sites to me at lshaw@hire-works.com.  I\'d like to do a post and survey to publish and share with readers.  Make it a great day!  Lucia  '),(14463,3247172,'18,May,2004','2004-05-18 00:00:00','   Job Posting - Do I Dare?  In today\'s mail I received a coupon from monster.com offering me 2 free job posts if I post 5.  The promo copy compares the value of a job posting on monster.com to a posting you might place in your local or a regional newspaper.  Monster applauds its screening tools and comments that in a newspaper ad anyone can apply and you cannot screen them out.  Well, screening tools on the internet are a far cry from screening anyone out.  I have used several job boards (in the past...I never use them now) and anyone really can still post a resume online and apply to your posting causing you HOURS of work filtering through all of the unqualified candidates.  I can almost guarantee you that within 1 hour (yes 1 hour) of posting a position to monster.com you are likely to receive over 50 resumes and that number will surely be closer to 100 by the end of the 1st 24 hour period. Ask yourself, do I have time to post a position and deal with all that comes with it?  Sites like monster definitely have their place in the search space.  I think they are great places for image advertising (good value really for that) and posts here are a great tool to see if folks really can follow directions, use email, attachment features etc.  Posts are great for retail and hospitality positions when it\'s also important to you that future team members are computer literate.  I welcome comments and questions from readers...email me at lshaw@hire-works.com.  Make it a great day!  '),(14464,3247172,'17,May,2004','2004-05-17 00:00:00','   Getting \"Linked In\"  I\'d like to invite those of you who have not checked out http://www.linkedin.com to do so.  It\'s an excellent way to search for candidates if you are a recruiter.  It\'s also a great way to network and do research for companies you think you\'d like to do work for.  If you are looking for a new position, job, career, you can post a resume on-line as well so we can find you!    I\'m a Spoke (http://www.spoke.com) user as well, but must admit that I much prefer the ease of use on LinkedIn.  Spoke\'s support however is fantastic.  I emailed with a question and got a call back the very same day from a member of the support team.   Remember, if you are looking for work - job boards are not the only vehicle.  Do research in your field.  Use boolean search strings, look for blogs like these (and others).  If you are a life sciences professional I recently discovered an exellent resource www.dnajobs.com.  Very organized list of positions by specialty and state.  If you have a great resource you\'d like to share, email me at lshaw@hire-works.com so I can post it for others.  Have a great day!  '),(14465,3247172,'13,May,2004','2004-05-13 00:00:00','   Free Webinar, The Future of Hiring, sponsored by Deploy Solutions will be held May 18th, from 2-3pm EST.  Guest host is Lou Adler, industry expert and author of \"Hiring with your Head\".  Lou is a great presenter, speaker with over 30 years experience in the field.  Not to be missed!  Register at www.deploysolutions.com.  Lou always has a Q&A session at the end of each call....and usually a download of the webinar...great value...!  See you on the call.  '),(14466,3247172,'06,May,2004','2004-05-06 00:00:00','   urlLink Applicant Tracking, Talent Management - Deploy Solutions  '),(14467,3247172,'06,May,2004','2004-05-06 00:00:00','   urlLink Pharmiweb.com | The Heart of Life Sciences on the Web  '),(14468,3247172,'06,May,2004','2004-05-06 00:00:00','   urlLink Blog Search Engine - Blog Directory, Blog Search, Blog Directories - Add Link  '),(14469,3247172,'06,May,2004','2004-05-06 00:00:00','   Good morning!  After much work, the first issue of our monthly ezine is now available.  The Hiring insider, helping you make better hires faster.  To subscribe please visit http://www.hire-works-forhire.com and sign up for our newsletter.  Or, send us an email at info@hire-works.com.  We look forward to your feedback.  Make it a great day.  '),(14470,3247172,'06,May,2004','2004-05-06 00:00:00','   urlLink Pharmiweb.com | The Heart of Life Sciences on the Web  '),(14471,3247172,'06,May,2004','2004-05-06 00:00:00','   urlLink Pharmiweb.com | The Heart of Life Sciences on the Web  '),(14472,3247172,'06,May,2004','2004-05-06 00:00:00','   urlLink Google Toolbar  '),(14473,3247172,'03,May,2004','2004-05-03 00:00:00','   Hot Jobs!   We\'re pleased to be working with a client who was named one of Fortune\'s 100 Best Companies to Work for in 2004.   Weâ€™re seeking an innovative and motivated Clinical Trials Account Manager to sell the services their research division has to offer clients worldwide.  Your role?  Expand a current customer base as well as establish new business in the field of clinical research.  Qualified candidates should have a minimum of three years experience in a CRO environment selling services (Phase I-IV). Strong project management skills and proposal writing skills are a must. The team includes 4 Business Development Associates each reporting to Director of the department. For profit organization, campus/academic setting.   Would you like to join a team of professionals in clinical research where there has been very little turnover, excellent support for research to prepare bids, quotations and proposals? Interested in experienced clinical staff to travel with you to sell your services?   Please send resume and cover letter to my attention at lshaw@hire-works.com.  Tell me about the types of projects you sell and service, therapeutic areas of interest, as well as what preferences you may have.  If you have a friend or colleague who may be interested, please feel free to have them get in touch with me.   Lucia  lshaw@hire-works.com  '),(14474,3247172,'03,May,2004','2004-05-03 00:00:00','     '),(14475,3247172,'03,May,2004','2004-05-03 00:00:00','   INSIDER TIP 1:  LEARN HOW TO THROW A BOOMERANG   Welcome to the Hiring Insider.  Let\'s begin to build you a Hiring Toolkit  you can use .  We\'ll archive all posts and assemble tips and tools into a toolkit you can access, so don\'t worry if you are trying to figure out where to store these ideas!  Besides, you must admit youâ€™re curious about the proper way to throw a boomerang?     â€œThe boomerang effectâ€ is a relatively new trend of inviting back talented former employees into the fold.  Surveys indicate 12% of employers cautiously admit they would re-hire ex-employees (alumni).  An additional 21% say they welcome back alumni without hesitation because of the time and dollar savings.  In fact, the Gartner Group estimated that in 2003, 60% of midsize to large companies will have routinely hired back former employees.  Here are four simple strategies you can consider to reclaim lost talent swiftly:   	Keep your alumni on your press release distribution list.  Keep them in the loop!   	Send current and former employees alike a list of your job openings!   	Invite alumni to company picnics, holiday parties, and other casual, festive events!   	Keep lines of communication OPEN.  At the very least, you want these people might to refer other talented candidates your way.    Remember, theyâ€™re going to talk either way. Why not help your own cause?    Watch for a new tip or tool each week.  To contact me directly, please email (me) Lucia at lshaw@hire-works.com.   To ask us a question, please visit our new ASK page (complete with images and sound!) http://www.askhireworks.com.  '),(14476,3247172,'30,June,2004','2004-06-30 00:00:00','   Part II of your search plan should be a summary or approach of the process.  Possibly the first step will be an initial review of job description (s) and maybe even the complete life cycle of the current recruitment process.  Next you will want to identify who will be involved in the interview process and why.  Map the process out on a whiteboard with targeted deadlines (or desired dates for completion). Will you need someone to start by a drop dead date for cross training (before the outgoing ee leaves)?  Or is there a manager the person must meet before going to Japan for a month? This is a time to check everyone\'s calendar and avert disaster whenever possible!  Be sure that you and your recruitment team are clear on who is responsible for what.  If travel arrangements and interview times need to be scheduled don\'t lose time figuring out who is responsible for the workflow. Likewise, who will be extending the offer to the candidate and checking references; will you need a background check, health screen, credit or licensing check?  I usually present my written search plan to clients for a sign off just after we sign a contract to proceed with the work. That way before going forward we\'re on the same page about what will happen next.   Part III to follow tomorrow.....make it a great day. Lucia  '),(14477,3247172,'28,June,2004','2004-06-28 00:00:00','   This week I\'d like to review the 5 elements in a solid search plan.  I thought I\'d break it down and do one part each day for a series this week. On Friday I\'ll post a full mock plan.   When working with an executive level recruiter you will have an  initial phone consultation  which is generally followed by a written search plan so both parties are on the same footing and in agreement about the search process going forward.  The phone consultation should review why the vacancy exists, who else is on the team and what strengths are we seeking in a new addition to the team. Is the position actually approved and what will the interview process look like (who will be involved, who will make the offer and when do you need someone to start)?  In a written Search Plan HireWorks, Inc. (my company) first reviews your needs summary.   I. Needs Summary   This should summarize your information about the position, location, title, number of positions, salary range, work environment etc.   Skills and competencies required should be listed next.  I typically use adjectives and phrases listed in a series to be sure I\'ve hit on all areas of interest to my client.  This includes  minimum  requirements such as degree required, years of experience, software knowledge etc. But it also lists for example, critical thinking skills, ability to work independently or from home.  I\'ll speak to motivation levels and attitude required as well.    Tomorrow I\'ll review Part II of the Search Plan - Suggested Work Plan (Process).  Questions and comments are welcome.   Lucia lshaw@hire-works.com  '),(14478,3247172,'17,June,2004','2004-06-17 00:00:00','   Bear with me today - I\'m reprinting an article in its entirety today...good read. Enjoy.    Make Use Of The Calm Before The Storm    By: Bill Larkin    This article originally appeared on HR.com.    The primary motivator for employee loyalty over the last three years has been fear â€“ fear of being downsized, and fear of not being able to find another job. But as signals of an improving job market continue to flash, thousands of employees who have been rooted in their jobs are plotting to free themselves. The market tide is turning in favor of the workforce, and experts say few employers are ready for it.  If you donâ€™t believe it, consider this: in a survey conducted by Accenture, 40 percent of U.S. workers say they plan to leave their current job within the next five years. A recent AOL survey paints an even grimmer picture; in which 58 percent of more than 5,000 participants said they â€œmayâ€ or â€œwill definitelyâ€ start looking for a new job when the economy improves. Why are these numbers so high? The need for financial cutbacks forced many organizations to neglect programs that are essential to employee retention. The Accenture survey says it all: respondents cited lack of compensation (71 percent), advancement opportunities (58 percent), and professional development or training (51 percent) as the primary reasons for seeking new employment.    Despite the renewed focus on employee retention that 2004 is expected to bring, many experts say the damage has already been done. â€œWe are entering a dynamic period for the employment market,â€ says Ed Jensen, a partner in Accentureâ€™s Human Performance Service. â€œEmployers need to protect themselves from raids on their talent in order to remain competitive.â€   It stands to reason, then, that there are two main challenges looming on the horizon: acquiring only the best of the candidates you will need sooner than you think, and keeping the best of those you have. Ironically, these burdens will fall upon one of the hardest-hit departments in your organization: human resources.    Pan for gold Itâ€™s hard to believe that a short time ago, HR professionals â€“ and the organizations they work for â€“ were used to shelling out 30 percent search fees to agencies in the hopes of securing a high-quality candidate to fill one of their many open requisitions. Now, there is no room in many budgets for outside help in the hiring process, and thereâ€™s less time and manpower to do the legwork internally. For those HR departments that find themselves short-handed due to recent downsizing and attrition, the thought of a renewed focus on recruiting is surely a cause for concern.    Despite the popular assertion that the current market is still flooded, finding the right candidate is harder than it may seem. â€œThe writing is on the wall,â€ asserts Scott Cohen, national practice leader for talent management at Watson Wyatt Worldwide. â€œOrganizations will increasingly compete for the smartest and the best individuals.â€ And theyâ€™ll have to hunt pretty hard for them, like the proverbial needle in a growing haystack. A recent report issued by Challenger Gray & Christmas Inc. maintains that 92 percent of recruiters and hiring managers were inundated with irrelevant responses via the Internet, and 71 percent said a majority of the resumes they received in response to an online job posting did not match the job description. The increasing demand for resume and recruiting management is poised to overwhelm an HR department that has already been reduced by layoffs. How does the already overburdened HR professional handle and evaluate resumes to fill critical positions from a shrinking pool of contenders with less help on a tighter budget? Most major and many smaller companies have already automated their recruiting and applicant tracking processes. Those that have not need to make it a top priority.    The ideal solution should provide integrated applicant management capabilities that not only streamline the hiring cycle, but also enable hiring managers to identify the best candidates more quickly and cost-effectively. The idea is not to replace human judgment, but to support decision makers by automating repetitive tasks and increasing the efficiency of those high-touch activities that cannot be automated. Managers may even use self-service to hire applicants on the spot, an important capability in environments where demand can dictate rapid bursts of hiring, such as retail.    Do your homework In desperate times, people may take desperate measures: some people will do anything it takes to land a job, up to and including fabricating degrees, work experience, previous earnings, and qualifications. Even worse than including phony accomplishments, they may leave out negative information that would typically disqualify them. In an era that has uncovered â€“ too late â€“ a high-profile college football coach who exaggerated both his college and football experience and an armed robber turned board chairman of a major firearms manufacturer, itâ€™s frightening to imagine how many covers have yet to be blown. While some of these stories are benign enough, the wrong person in the wrong job can cause real harm, and when that happens, the damage to everyone involved can be devastating. How can organizations avoid hosting the next big scandal, or worse? By understanding how to legally and accurately verify candidate credentials, and making the effort to do it consistently.    With less HR staff in place to perform extensive background checks, is it worth the time it takes? Absolutely. Even one bad hire can do some serious damage. LeRoy Robbins, CEO of recruiting specialists IIRC, claims that the hard costs are estimated at â€œ50 percent to 200 percent of that personâ€™s salary and benefits,â€ not to mention the soft costs of an increase in workload and its accompanying decrease in company morale and customer service. â€œAs you move up the line, bad hires become more significant because of the impacts of their roles. Having to replace that person is enormous,â€ explains Robbins.    Beyond the responsibility to the organization itself, there are legal duties to consider, such as taking the proper steps to avoid hiring an unfit individual who causes harm. Negligent hiring processes, discrimination, and invasion of privacy can all lead to costly grievances and lawsuits. At the 21st Annual Employment Law & Legislative Conference, Wendy Bliss, J.D., SPHR discussed effective pre-employment screening practices. Depending on the requirements of the position being filled, all of these methods may be appropriate: multiple reference checks; verification of the last 5 to 7 years of work history; criminal background checks; a social security number trace, examination of education and motor vehicle records; and credit, professional license, certification, or registration checks when job-related and validated. She also recommends checking the applications over for any of the following red flags: blanks, unexplained gaps in employment, questionable reasons for leaving prior jobs, illegible writing, no signature, information that is inconsistent with the applicantâ€™s resume, and multiple people listed who may not be contacted for a reference. Because every stateâ€™s laws differ slightly, it is vital that employers familiarize themselves with the specific requirements governing reference immunity, service letters, use of criminal history, and other related issues for any and all of their locations.    Stay ahead of the pack As the job market continues to improve, employers will find themselves once again catering to the best and brightest talent. â€œEmployers have been less careful about how they treat their employees because they know there is a ready supply out there,â€ says Sharon Koss, SPHR, president of Koss Management Consulting in Seattle, WA. â€œThey have forgotten what turnover is, how expensive it can be, and how hard it can be to find good people.â€    Theyâ€™ve also forgotten how much good people can cost in a competitive market. To avoid getting burned, employers need to first make sure that all current employees are fairly compensated, and when a position comes open, consider hiring from within. In many cases, the best person for the job may be lurking a few cubes down. This is another scenario in which an automated system can make a world of difference. With a comprehensive in-house talent pool just a couple of clicks away, recruitments costs are significantly slashed. This makes it easier for organizations to promote top talent, fostering higher employee engagement, and improving the retention of key employees.    Then, organizations should start putting in place a system by which they can ensure the compensation packages they offer are competitive, but also fall within the parameters of their budget requirements. For example, are you able to quickly measure the value of a candidateâ€™s total compensation package (salary, benefits, bonuses, etc.) against your budget for the position? Can you compare packages you are offering against industry compensation data, salary surveys, and internal jobs before an offer is made to the candidate? Can you easily take into account the geographic pay differentials that exist based on the candidateâ€™s work location?    Itâ€™s a lot of work, but a system of compensation â€œchecks and balancesâ€ is critical to preventing your company from overpaying or under-competing for talent. Fortunately, there are applications available that can automate much of the legwork that needs to take place in order to assemble sensible but competitive compensation plans. In fact, some organizations have used applications and the accessibility of data to take a less traditional approach to compensation.    According to Edward Lawler, a professor at the University of Southern Californiaâ€™s Marshall School of Business, many organizations are using the availability of compensation data combined with automated employee performance management to develop competency-based compensation systems. It attracts employees who like the notion that, if they work hard, they know how and when they will be rewarded. It also prevents the loss of talent, which is perhaps even more important. â€œA competency-based system helps an organization achieve its goals,â€ explains Lawler. â€œWhen it is structured right, it rewards top performers and keeps many of them from leaving.â€ Companies that do not provide the right rewards and opportunities, however, wind up priming their talent for the competition.    Watch where youâ€™re going Many organizations are blissfully unaware that they are currently suffering from a false sense of security; in fact, most organizations believe that when the employee exodus does happen, it will happen to some other company. It wonâ€™t be long before that illusion is shattered. As the pendulum begins to swing back in the opposite direction, savvy employers who anticipate its movement can survive, or even thrive, by taking preemptive actions to retain their proven performers while they attract and groom the best of those candidates who have begun to test the waters of the recovering job market. By providing the appropriate rewards and opportunities to seasoned in-house talent and offering competitive compensation plans to meticulously screened prospects, your organization can ride the wave of the impending upheaval while your competition struggles just to weather it.   About the Author:  In his role as Vice President of HRMS Solutions, Bill Larkinâ€™s current charter is to drive Kronosâ€™ success in the Human Resource Management System (HRMS) market. In June 2002, Kronos entered the middle tier of the HRMS market with licensed HR and payroll products, which complement Kronosâ€™ existing suite of employee resource management solutions.    Larkin brings to Kronos extensive consulting, sales, product development and marketing skills honed during a 20-year career in human resources, in which he held positions with Automatic Data Processing (ADP) and Genesys Software Systems.    Prior to joining Kronos, Larkin founded LCS and Associates, a consulting firm specializing in HRMS end-user requirements definition and vendor selection. Larkin also has helped a number of software vendors develop and expand their HRMS strategy. Larkin was a co-founder of the eClipse Group, a consulting firm formed as a joint venture between the Human Resource Management Consultants and LCS and Associates.    Larkin is an active member of the International Human Resource Management Association and the Northeast Human Resource Association. He holds a bachelorâ€™s degree in business administration from Iowa State University.  '),(14479,3247172,'15,June,2004','2004-06-15 00:00:00','   The Triangle business journal reports that a recent survey conducted by Manpower states employers are looking to hire big in the 3rd quarter.  The paper reported:  \"From July to September, 47 percent of the companies interviewed plan to hire more employees while 3 percent intend to reduce their work force, the survey says. Fifty percent expect to maintain their current staffing levels.  Nationally, 30 percent of the 16,000 U.S. employers polled said they planned to add new employees, with 6 percent expecting a decline in the number of jobs.\"  Call your recruiting partners and prep us with projected needs and timelines.  The more we know, the better able we are to assist you.  Looking forward to hearing from you!  '),(14480,3247172,'07,June,2004','2004-06-07 00:00:00','   Good morning,   Many people want to know just what is the difference between HeadHunters, Exec Recruiters - third party recruiters and so on?   Well, typically we are all self employed and do not work for a corporation we\'ve been hired to staff - we\'re a third party provider. Some recruiters don\'t like to be called HeadHunters, doesn\'t bother me in the least.  We are hired to do what companies cannot or will not do for themselves.  Research positions, the market and target candidates who will become strong achievers for that organization.    I did a quick google search this morning and many searches are being done each day for all of the above titles and keywords daily.  In fact, the daily budget google recommends for an AdWord campaign is $750/day.  So that tells you about the level of interest and the intensity of searching for recruiters to call most days of the week.  If you are looking for a recruiting partner, give us a call.  We\'d be happy to do an initial 30 minute (free) consultation to see if we might be a resource for you.  Best wishes,   Lucia  '),(14481,3247172,'02,June,2004','2004-06-02 00:00:00','   Good morning!  Is your recruiting process going well? Is your recruiter(s)responsive and do they understand what you need? Are you happy with the flow of candidates, scheduling of feedback and interviews?  What is missing from the process and what can we (recruiters) do better to serve you?  Recruiting is a  service  you pay for.  You pay us to make it go away!  I tell my clients they should feel a huge weight lifted from their shoulders once they sign my fee agreement.  At that point I own the process and update them along the way. It is no longer their primary concern or worry.  They can move on, perform other duties and know that they\'ll hear back from me with status updates.  We want your feedback!  Is recruiting what you think it should be? Where do we have room for improvement?   Suggestions and feedback welcome.  Look forward to hearing from all of you out there!  Make it a great day.  Best, Lucia  '),(14482,3247172,'01,June,2004','2004-06-01 00:00:00','   I subscribe to ERE Exchange - a group/forum for recruiters.  Each day there is a feature article and while most are quite good, written by Dr. John Sullivan, Mr. Lou Adler and other professionals, today\'s article was awful.  For many reasons, articles like today\'s and recruiters like the one featured, is why I remain a recruiter.  Recruiting is tough work and some days even life insurance salespeople are treated better than we are...but articles like those found today are why!  Actually if you can get past the first paragraph of the article I do agree with most of what Mr. Adamsky says.  Sales is service...if you believe that and act that way then there is no need to be shark-like and use terms like Eat what you Kill. Just God Awful if you ask me.  People do business with those they like, know and trust.  I think the words you choose to represent yourself and those in your profession should be chosen carefully.  Tip of the day:  Be sure you LIKE, KNOW (or get to know) and TRUST the recruiter you are working with.  I promise it will be much more fun!  Best wishes,  Lucia  '),(14483,3247172,'29,July,2004','2004-07-29 00:00:00','   So you\'ve now got a sexy, exciting job posting written...   Where do you place the post to get your hire?   First of all, be sure that you comply with corporate policies and procedures and if required to do a 3 or 5 day internal posting for existing employees that you do so.   Then you can have some fun placing the ad/post where your target population is hanging around.    Ideas include:    User groups/forums  Google AdWords or Yahoo Overture (PPC) text ads  Flyers for bulletin boards in coffee shops, wireless cafes etc.  Blogs!  Audio Postcards Niche web sites (don\'t forget the diversity sites to target women and minority groups).  Job Boards (depending upon the type of position).  Print job announcements on pay stubs, in company newsletters and more.   There are many more ideas of course - brainstorm and track your results.   We\'ll cover tracking and ROI tomorrow....  '),(14484,3247172,'27,July,2004','2004-07-27 00:00:00','   Writing Job Descriptions that  pull  - that folks  want to read ... That\'s today\'s topic.   Job postings, position descriptions, ads whatever you\'d like to call them should be written from a sales or branding perspective. What are you trying to accomplish with these posts? We want folks to call or email us a resume right? Well then STOP posting job requirements that top candidates will never respond to.      Job posts should be interesting and lead the reader like a good title does a news story.&nbsp;&nbsp; Your post should be full of action verbs and questions...   Example:   Seeking energetic multi-tasker who likes to interact with others! Does this sound like you?  Issue a challenge in the post:  Would you get excited to be charged with the task of decreasing downtime by 10% in 60 days? Would you get even more excited charged with increasing productivity by 20% in 30 days?  Top candidates, those who have the competency to do the work and the motivation to do it better than it\'s been done before WANT you to challenge them! Why not start with a challenge in your ad or posts and distinguish your brand or work culture?  Have a story you\'d like to share? We want to hear about it!   More tomorrow.....  '),(14485,3247172,'26,July,2004','2004-07-26 00:00:00','   Good morning!   I\'m back after 10 beautiful days of vacationing on Cape Cod. It\'s the longest vacation I\'ve ever taken and I highly recommend it to anyone who needs a good break.  I\'m going to cover a 5 part series this week on how to hire. We\'ll begin at the beginning...  So you have a vacancy to fill. Why? Where? What\'s the work to be done, how and when?!  Take this as an opportunity instead of looking upon it as a burden.&nbsp;   What work needs to be done and with who?&nbsp; Who will supervise the effort and why? How can you measure good work? What skills are required or can be learned on the job?  Take a good look around the team make-up and \"re-engineer\" if it\'s time to do so.&nbsp; Become friends with your compensation department and help them to understand what you need and why.  Begin to make a good list of action verbs for the position description...which we\'ll tackle tomorrow.  Make it a great day. lucia  '),(14486,3247172,'15,July,2004','2004-07-15 00:00:00','   We\'ll be on vacation the 16th returning on the 26th.....\"see\" you again soon.   p.s. if anyone has posts they\'d like to share for future time off we\'d love to review your ideas.  Lucia/HireWorks  '),(14487,3247172,'12,July,2004','2004-07-12 00:00:00','   Good morning,  CareerXroads just reported that only 1.8% of hires were made as a result of posts to www.hotjobs.com.  Last year they reported that less than 10% of all hires made came from the big 3 (monster.com, hotjobs.com and caereerbuilder.com)... Feel free to click on the link above and read the full article.  They\'ve got a ton of information on their site - really a must read!  These are not great numbers folks.  As I\'ve said in the past, I do believe job boards have their place.  However, if you are reading this entry - they probably aren\'t going to be the best tool for you based on your needs.  Corporate sites (if maintained and working well) outperform boards by almost 40% on average...and of course working with recruiters really increases your success rate.   Remember customized search plans are 20% off (regular price $297) through July 31st.  Email or call us today and tell us \"I want to save 20% today!\"  '),(14488,3247172,'08,July,2004','2004-07-08 00:00:00','   Usually after receiving a signed fee agreement and completing a customized search plan for a client I email the a Work Environment Questionnaire to complete.  I can email the PDF file to you - just email at lshaw@hire-works.com  It\'s just 2 pages but gets right to the heart of the matter.  We\'re trying to find out if there is turnover, what can we do? Is it a fun place to work? Do coworkers like and trust one another? Is management open to listening to new ideas?  It\'s a good exercise...  Make it a great day.  '),(14489,3247172,'07,July,2004','2004-07-07 00:00:00','   Calling All Hiring Managers! Take 20% off your initial consultation and customized, written search plan (value $297).  Simply subscribe to our e-zine, The Hiring Insider by July 31st.  The Hiring Insider, a monthly e-zine, provides hot tips and tools to help you make better hires faster.   Email us at info@hire-works.com and in the subject line type â€œI want to save 20% today!â€.    We look forward to serving you.  Lucia  '),(14490,3247172,'06,July,2004','2004-07-06 00:00:00','   Part IV and V  Part IV of the search plan is usually where we include a bit of information about us (and I include my Bio).  I do this since the person with signature authority might not know much about the process to date and we want them to have a clear document with an articulated path that is easy to understand and sign off on.    Part V.  Proposal of Services and Terms.  1.  Planning 2. Execution 3. Working Status 4. Expenses 5. Rate or Fee (depending upon the type of work or project engaged in). 6. Termination Notice  7.  Related and on-going services   I\'ll put this all together and publish a live search plan to illustrate these points below:   Example of Live Search Plan (some information omitted to protect client confidentiality)   Recruiting Plan for \"client\"   HireWorks has prepared a proposed Recruiting Plan to serve as the blueprint to meet client\'s recruitment and retention goals.    I. Needs Summary:   Company name:	 Title of positions:	Customer Service Representatives Location:	 Number of employees:	40 + (high-volume recruiting needs) Orientation:	Nine-week training program Salary:	 Bonus:	 Benefits:	 Environment:	    Skills and competencies:	Strong multi-task abilities 	Flexibility in work scheduling (The facility will be open from 8:00 am to 8:00 PM Monday â€“ Friday) 	Clear communication skills 	Problem solving orientation 	Critical thinking skills 	Previous experience working in a call center environment preferred 	College degree preferred 	Works well independently as well as interdependently (in a team environment) 	PC literate (ability to type minimum 35-40 WPM) and possibly other minimum requirements as determined by client. 	Positive attitude, patience, strong commitment to excellence with fortitude to see a case through to completion.     II. Suggested Process   An initial review of both job descriptions and the complete life cycle of the current recruitment process (including metrics and related hiring and retention statistics) shall be conducted.  As part of this process, input from key managers and call center reps (staff) should be solicited.  All comments and suggestions should be considered.    The existing process should be mapped out so that process improvements can be introduced and rolled out.  For example, is client conducting exit interviews?  If so, why are people leaving?  Can any imrovements be made to increase retention?  Finally, a timeline for current hiring needs shall be defined, as well as a master plan for overall hiring needs as of January 1, 2005.  Note:  Be sure to secure buy-in from the management team as part of this process!   Projected timing and planning:  Since this call center is a benefits call center for client base, and open enrollment is July through October, all new hires must begin their nine-week training program by the last week of April.  Therefore, offers must be extended to determine the actual level of staffing by late March or early April.  Of course additional flexibility will be built into the plan in the event additional contracts are accepted, and staffing levels need to increase or decrease accordingly.     III. Sourcing   I would suggest the engagement of  multiple, simultaneous sourcing activities to keep the center staffed properly.  While I have not listed all possible methods, below are a few ideas to get started:  1.	Boomerang campaign (â€œWe want you back!â€):  Who left the center that youâ€™d like to return?  2.	Sourcing from competitors (locally and regionally). 3.	Sourcing from like industries, based on specific skill sets required (i.e. hospitality, airlines etc.)   4.	Campus recruiting (two and four year schools).  Sell the management as well as the global opportunity with client, and promotional opportunities therein.  Does client assist with college education for staff  as an incentive?  5.	Diversity Recruitment (on-line, career fairs etc.).  Is there a need for bilingual phone reps? 6.	Job Boards:  Monster, HotJobs, local boards (e.g. trianglejobs.com and use of MetaPosts.) 7.	Internet Sourcing (AIRS Oxygen, Boolean Search, etc). 8.	Open Houses (to be held at call center) followed by an Ambassador Program. 9.	Career fairs (locally and nationally). 10.	Online communities (e.g. Topica, AOL etc.) 11.	Media outlets: Print advertising, billboards, possible use of radio and internet radio shows in to reach the appropriate target market.  12.	Employee referral program (ERP).     IV.  Luciaâ€™s Bio   Lucia Apollo Shaw is President and CEO of HireWorks, Inc.    She has been helping her customers for nearly nine years, working both as a third-party recruiter for CDI Corp. (staffing customers like IBM), Trilogy Consulting (now Venturi Partners), staffing the Biotech and Pharmaceutical industry, and in places like Duke University where she was a corporate recruiter and Team Leader for recruitment for Duke University Hospital.   Prior to her recruiting career, Lucia was among the first group of trainees hired when Phoenix and Home Life (PHL) Insurance merged and created a new call center in New York.  Lucia was promoted to Team Leader, responsible for hiring, training and managing a group of 12 staff members.  At that time, the reps averaged approximately 100 incoming calls per day (in a four-hour time frame).  Lucia was employed by PHL for a period of four years before relocating to North Carolina.  Lucia earned a B.A. from the State University of New York (University Center at Albany) and pursued Graduate Studies in Public Administration at the Sage Graduate School in Albany, NY.   V.  Proposal of Services and Terms   Per our conversation on, I understand that the current need is for a senior recruiter to work for a period of  three to six months.  I am incorporated, and currently have an office, equipment, etc. in Chapel Hill.    I propose the following (and welcome further discussion):  1.	Planning:  I will create a specific recruitment and hiring plan as outlined in this document (which will include feedback and input from client) for implementation.  2.	Execution:  I will work from my office in Chapel Hill three days per week, and report to client site two days per week (with the understanding that there may be occasions where additional on-site time is required).  When off-site, I will be available via telephone, fax and e-mail.  3.	Working status:  The contract that both parties (client and HireWorks) will sign will recognize that neither HireWorks, nor Lucia Apollo Shaw, will be an employee of client.  4.	Expenses:  HireWorks will not seek reimbursement for costs of operation relating to phone, internet charges, or travel expenses to and from the client Site.   5.	Rate:  $35.00 per hour, on 40 hours per week, for a period of three to six months.  Invoice remuneration to be submitted to HireWorks on net 30-day terms, which shall be inclusive of  actual hours of service performed.    6.	Termination notice:  Minimum 30-day notice required to HireWorks.  Regardless of the timing of the termination notice, all billable time for services rendered, but not yet paid, to HireWorks, Inc. shall be recognized as the final payment to Hireworks.    7.	Related and on-going services:  HireWorks welcomes any future discussion of on-going  contract services beyond the anticipated three to six month initial requirement.  Additional recruiting services and searches for specific positions can be initiated upon request.  '),(14491,3247172,'01,July,2004','2004-07-01 00:00:00','   Customized, written search plans continued...  My favorite part - sourcing candidates!  Most recruiters will suggest the engagement of MULTIPLE, simultaneous sourcing activities to keep the flow of candidates moving for both the short and long term.  Remember, you are also always seeking profile candidates for positions you don\'t even have vacant now as well!  Here is an example from a live search plan I\'m currently working for a client:  1.	Boomerang campaign (â€œWe want you back!â€):  Who left that youâ€™d like to return?  2.	Sourcing from competitors (locally and regionally). 3.	Sourcing from like industries, based on specific skill sets required (i.e. hospitality, airlines etc.)   4.	Campus recruiting (two and four year schools).  Sell the management as well as the global opportunity, and promotional opportunities therein.  Do you assist with college education for staff  as an incentive?  5.	Diversity Recruitment (on-line, career fairs etc.).  Is there a need for bilingual staff? 6.	Job Boards:  Monster, HotJobs, local boards (e.g. trianglejobs.com and use of MetaPosts.) 7.	Internet Sourcing (AIRS Oxygen, Boolean Search, etc). 8.	Open Houses (to be held at call center) followed by an Ambassador Program. 9.	Career fairs (locally and nationally). 10.	Online communities (e.g. Topica, AOL etc.) 11.	Media outlets: Print advertising, billboards, possible use of radio and internet radio shows in to reach the appropriate target market.  12.	Employee referral program (ERP). Does it need a refresher? Do staff members know all there is to know?   Tomorrow we\'ll close this up and put it all together.  Questions/comments??? email lucia at lshaw@hire-works.com or post a comment on the site right here.  '),(14492,3247172,'28,June,2004','2004-06-28 00:00:00','   I have a question for all of you candidates out there.  How are you finding/locating recruiters to send your resumes to?  Are you doing searches on the internet or using on-line recruiter directories?  I am receiving a LOT of unsolicited resumes lately (which is fine!) but do not use any of the job boards for position advertisements.  So we\'re in the office getting all of these resumes from people in our space (relevant resumes) wondering how it is happening!  Email me and tell me how you heard about HireWorks, Inc and we\'ll post our results.  P.S. Keep the resumes coming - we look forward to serving YOU in your search.  Best wishes,  Lucia lshaw@hire-works.com  '),(14493,3247172,'17,June,2004','2004-06-17 00:00:00','   Calling all PT\'s, OT\'s PTA\'s and Speech Pathologists!  We are staffing rehab facilities throughout NC and VA and need your help to spread the word.  Excellent multi-disciplinary team environment, great benefits and continuing education program.  Please call Lucia directly at 919-960-3411 or email resume to lshaw@hire-works.com.  '),(14494,3247172,'16,June,2004','2004-06-16 00:00:00','   Senior Sales Professional - Online Patient Recruitment Services   This position requires  significant experience in selling patient recruitment services  to biopharmaceutical clients.  We require an energetic, motivated individual who can sell an array of patient recruitment services, is agile/flexible and adept at thinking creatively and can communicate effectively with a variety of audiences.    Our client provides solutions through three distinct product lines: clinical trial listing service on the Companyâ€™s industry-wide clinical trial information exchange (veritasmedicine.com); clinical trial specific online recruitment campaigns with screening websites; and client specific web-based recruitment systems.  Base salary plus commission, Medical, dental, 401K, and stock options.   Please send resumes to Lucia at lshaw@hire-works.com.  '),(14495,3247172,'14,June,2004','2004-06-14 00:00:00','   I\'m working with a new client with multiple locations in the SouthEast and we\'re seeking Therapists to join our team. We\'d prefer direct hires, however will consider per diem hours and contract staff.  Feel free to send resumes to Lucia at lshaw@hire-works.com or give me a call at 866-960-3411 for a chat.  Have a great day!  '),(14496,3247172,'10,June,2004','2004-06-10 00:00:00','   We have multiple openings nationally for cytotechnologists...certification desired but not required.  Experience preferred, but new grads are welcome to apply.  Please contact Lucia directly for additional information:  toll free 866-960-3411 or lshaw@hire-works.com  '),(14497,3247172,'09,June,2004','2004-06-09 00:00:00','   Most recruiters will be happy to help you revise and customize your resume for jobs we\'re helping you interview for.   I must admit though, I cannot stand writing and updating my own resume, must less anyone elses.  Remember to lead with strong action verbs, list major accomplishments and please be sure to spell check your resume (including getting rid of extra spaces).  You would not believe the typos I see that I would rather not.    Also, there is no need to list your date of birth, marital status and names of children.  Even hobbies are questionable in my mind on a professional CV.    There are good resume writers - no question...but they cannot invent the content and experience..you\'ve got to have it!  Questions and comments are welcome.  Look forward to hearing from you.    Lucia  '),(14498,3247172,'07,June,2004','2004-06-07 00:00:00','   Morning!   I need some feedback from those of you paying to have your resume \"Blasted\" to recruiters.  Are you getting phone calls and job offers as a result?   I\'ve recently subscribed to all 3 services as a recruiter and am getting some good looking resumes on paper.  But I\'m curious to know from you, the users (and payors) how is this working? What is its biggest strength and weakness? Do you have to pay multiple times to get through all of the \"noise\" on the internet?  Do you feel like you hear from more experienced recruiters than you do if you just post free on hotjobs.com or monster.com?  ...inquiring minds want to know!  we look forward to hearing from you.  Lucia  '),(14499,3247172,'04,June,2004','2004-06-04 00:00:00','   I receive hundreds of unsolicited resumes via email each month.  It never fails - several per week come to me \"To whom it may concern\".   Please do not continue to do this.  If you have taken the time to research recruiters and send your resume to them unsolicited why don\'t you address it to them personally?  I must tell you that this is such an irritant to me that going forward - any resume I receive in my inbox \"To Whom it May Concern\" will be deleted immediately.  Tip: take the time to research your contact, spellcheck your email and resume before sending to anyone.  Have a great weekend.  '),(14500,3247172,'03,June,2004','2004-06-03 00:00:00','   Like many professional recruiters, I like to work with candidates and set up phone appointments for pre-screens.  I let my candidates know we\'ll need 20-30 minutes together for them to share with me their background, likes, dislikes etc.    The purpose of this call is not typically to discuss a particular job.  Candidates I talk with often are not looking for a new job the first time I talk with them.  They make become a client first and sometimes I may never place them at all.  There have been times where the placement did in fact occur sooner than both of us thought it would and I\'ve often refrerred candidates I speak with to colleagues who have helped them if I could not.  But my tip today is to KEEP that initial phone appointment with the recruiter.  If something comes up and you need to change the appointment, email and/or call the recruiter.  But do not blow it off.  We\'ll probably never call you again and I for one will be sure to make a note of it in my database where it will forever remain.   Keeping appointments is professional courtesy and an indicator of future behavior.  Many recruiters ask themselves, if they blew me off, will they blow off the interview with a hiring manager  I introduce them to ? Remember, we\'re doing our homework for our clients.  Keep those appointments and follow up with your recruiter.  Have a great day!  '),(14501,3247172,'02,June,2004','2004-06-02 00:00:00','   To all of you sending out resumes via email......please do not email PDF files of your resume or CV.  Most of need and want an MS Word file we can import into our database to search and parse data later on.....  Just a tip!  best, Lucia  '),(14502,3247172,'01,June,2004','2004-06-01 00:00:00','   Iâ€™m pleased to be working with a client who was named one of Fortune\'s 100 Best Companies to Work for in 2004.    Weâ€™re seeking an innovative and motivated  Business Development Rep  to sell the services their research division has to offer clients worldwide.  Your role?  Expand a current customer base as well as establish new business in the field of clinical research.    Qualified candidates must have a minimum of three years experience in a CRO environment selling services (Phase I-IV).  Strong project management skills and proposal writing skills are a must. The team includes 4 Business Development Associates each reporting to Director of the department. For profit organization, campus/academic setting.    Would you like to join a team of professionals in clinical research where there has been very little turnover, excellent support for research to prepare bids, quotations and proposals? Interested in experienced clinical staff to travel with you to sell your services?   Please send resume and cover letter to my attention at lshaw@hire-works.com.  Tell me about the types of projects you sell and service, therapeutic areas of interest, as well as what preferences you may have.   If you have a friend or colleague who may be interested, please feel free to have them get in touch with me.   Lucia Apollo Shaw Recruiter  lshaw@hire-works.com  Have a question? www.askhireworks.com  HireWorks, Inc. Chapel Hill, NC  www.hire-works.com  Phone:    (919) 960-3411 FAX:       (919) 960-3413  '),(14503,3247172,'01,June,2004','2004-06-01 00:00:00','   Welcome to the first entry of The Job Blog.  I am a life sciences recruiter (just over 9 years now). What does that mean? I\'ve been recruiting for Health Care Institutions (clinics, research centers, hospitals) and for small to mid pharma/biotechs and CROs.   Hope that helps....we recruit for all disciplines as long as they are in the spaces mentioned below.  So if you are an IT professional, RN, Pharmacist, MD, Rad Tech etc. we are looking for you if you are looking for a new career.  Be sure to read daily and post your comments. If you are ready to send me a resume feel free to email it to me at lshaw@hire-works.com.  But if you want to lurk around and read and get to know me - feel free to do that as well.  Best wishes,  Lucia Apollo Shaw www.askhireworks.com/healthcare.html www.askhireworks.com  www.hire-works-forhire.com  919.960.3411  '),(14504,3247172,'12,July,2004','2004-07-12 00:00:00','   Cover letters should be short, strong, and interesting so the reader moves on and reads your resume.  Cover letters should not be a recap of your resume.  Keep the cover letter to 3 or 4 short paragraphs.    Paragraph I.  Let the reader know what you are applying for and how you heard about it.  Paragraph II. Use 3 or 4 bullet points with strong action verbs demonstrating how what you\'ve done elsewhere could be done at the job you are applying for.  Talk about how much above goal you are if you are in sales or how far below budget you came in at if you are a project manager or how quickly you accomplished XYZ etc.  Paragraph III.  Invite the reader to contact you for more information about how you can help them.  Paragraph IV is optional...if you need more room.  Folks reading resumes get a LOT of them.  Readers will really appreciate your ability to get to the point quickly and highlighting what they need to know.  And be sure to SPELLCHECK your document!   Good luck!  '),(14505,3247172,'09,July,2004','2004-07-09 00:00:00','   HireWorks is expanding.  We recruit healthcare and clinical research professionals nationally and need more help.    We are seeking Recruiters and Researchers who will work from their own home.  You should have access to high speed internet and a dedicated phone (biz) line.  Part-time/Flexible hours are ok.   Please send resume and letter of interest to Lucia at lshaw@hire-works.com  '),(14506,3247172,'02,August,2004','2004-08-02 00:00:00','   Afternoon!  Today I read a stat which was very interesting.  According to the Cornell University Career Services Department research show that up to  80 percent  of all job openings are  unadvertised .  As a recruiter I know that many jobs are never advertised for many reasons.  But I\'ll be honest, I did not think the number was that high...50% or more maybe..but 80%.  Wow!  My advice:  Find a recruiter you  like, know and trust  and they\'ll help you find the best job for YOU.  Make it a great day.  '),(14507,4200309,'19,August,2004','2004-08-19 00:00:00','   I just saw this movie... Twilight Samurai, and I\'ve got to say it\'s one of the best movie I have ever seen. It\'s a man content with himself without an ambition to build a career that can ultimately destroy his humanity and family. It\'s the perfect portrayal of life and fate. I love it. It\'s messages are narrated beautifully without a hint of persuasion. So if you ever visit your rental or dvd store, buy this gem! I highly recommend it!  I\'ve also been listening to L\'Arc-En-Ciel\'s latest masterpiece: Smile. Although the Album\'s only got ten songs, every one of them is worth listening! '),(14508,4200309,'18,August,2004','2004-08-18 00:00:00','   I plunged.  I was excited. And I dreamed again of roses and berries.  And like so many in my past....  A dream is only a dream... '),(14509,4200309,'18,August,2004','2004-08-18 00:00:00','   Ismiaji Cahyono, This isn\'t the gentle kind of day that you Librans may dream about having, but it\'s time to step outside of your regular role. Playing the perfect host or hostess is not what the planets have in mind for you now. Itâ€™s time to get down and dirty, or maybe even a bit crazy. Jump into your fantasies and let them come to the surface. Of course, you donâ€™t have to express every single thing you imagine, but it is okay to say more than you usually share.  That, by the way, was my astrology of the day.  Hahahahahahaha... how ironic!  By the way, I saw The Village yesterday. Despite of the critics, I liked it.  This Bali photosession is beginning to get on my nerves... now my boss wants to come along... shit!  More postings tomorrow.... '),(14510,4200309,'16,August,2004','2004-08-16 00:00:00','   At the end of the day I heve neither the strength or wit to write a clever entry. Well that\'s not the point is it?  Today has passed by so fast I didn\'t notice anything coming. Maybe I wasn\'t supposed to. Does it matter? Probably not.  Tomorrow is another holiday. Should I look forward to it? Perhaps. I\'ll be meeting my friends. Something I haven\'t done in quite a long time.  I\'ll be leaving to Bali on Monday. Yipee. Maybe I can work on that tan I\'ve always wanted. '),(14511,4200309,'12,August,2004','2004-08-12 00:00:00','   Here I am alone again at the office... One of the first to arrive and the last to leave. I hope that it won\'t conclude my life.  I could describe my routinity but then it would bore you...  You know what\'s weird. How you are so busy, pushed to the limits by piles of overlapping work that you\'re body automatically reacts to it like a machine, you\'re mind succumbs to the pressure, and goes blank... and before you know it, you\'re work is done. I was working on a company profile, minutes away from printing it. Then a disaster happened. The boss\' secretary delivered me an image via network without clearly notifying me. My crummy machine crashed. I was able to recover my file but it wasn\'t printable... Apparently my software program was \"shocked\" at the abrupt meltdown and ruined my file. Shit. I was speechless. I couldn\'t get angry at the secretary. It was an honest mistake. A freak accident. Two days worth of work gone. Now I have to rebuild everything. Luckily, my colleague kept my original file, and so I would pick up from there. Somebody up there must like me. Can you imagine an overnight work redone in just an hour. My elbows felt like rubber at the end.... Sigh!  The only thing interesting today was casting models. Yup, an entourage of beautiful models lined up to be casted in this super exclusive apartment project. Not that I haven\'t seen beautiful women before... It\'s just interesting to see and hear people who make a living being an object of promotion.... Things like this fascinate me. I\'m curious as what goes on through their mind. What is behind that beautiful facade? I\'m also amazed at how well maintained they are... but women nowadays are like that. I guess I\'m not in tune to such concept. Nontheless it\'s always exciting to discuss....  It looks as though we might have to delay the photoshoot. Great! Now I can go to that interview!  All in all it was fun. I just wish I\'d hear from some friends... but I guess today is not the day.  I\'m still smilin though. '),(14512,4200309,'11,August,2004','2004-08-11 00:00:00','   Just a perfect day... Lou Reed  It\'s cloudy out. The air conditioning is blasting. I\'m freezing. I have an awful headache.  My interview at Fulbright went pretty well. At least I thought  it did. I hope I get this one. Can\'t wait to get out of this hell hole. Don\'t get me wrong, I love this country, but more and more each day, the future here seems foggy.  I just put myself in a dillema. I agreed to an interview this monday. A teaching post at Binus. Nice right? The catch is that on the same day I have a photoshoot. Shit. I can\'t afford to postpone this...  Gosh it\'s awfully lonely.  By the way, this is my first post. The first of many to come. This is also my first real interaction with the vast NET. I hope my stuff  will blow you away... or bore you to death. Either way, I am pleased to be able to share my feelings, thoughts, my philosophy, and my troubles with the digital community... and hopefully you! Yea you...  So welcome to ismeagain.blogspot.com  Add me and check me out. Place your comments. I\'ll try and get back to you as soon as I can. Have fun reading my miseries. Hah! '),(14513,3661106,'18,June,2004','2004-06-18 00:00:00','   Try not to cry Don\'t let the pain show When your fears hurt the most Just let them go No regrets You told me with a smile But I didn\'t listen I still cry  You live a life Doubtless and free of pain and worry \'free of misery\' I wish I could live by your wise creed but I can\'t block emotion It won\'t let go of me  I look up to your wisdom I admire your strength But I can\'t follow your footsteps For I am still week A fledgling of truth Yet living in lies this is the lifestyle you so despise  This is written for my cousin Harrison, who though he is 6 months younger than me, has always been a sort of mentor.  He\'s there for me in tough times and gives me advice on what I need advice on the most, life. '),(14514,3661106,'16,June,2004','2004-06-16 00:00:00','   Well, this isn\'t exactly poetry, but I\'m just not feeling very inspired at the moment.  Of course, that could be due to the fact that it is 2:33 AM where I am currently located.  Anything I try to type that isn\'t planned will most likely not be coherent.  I\'m very tired, but I don\'t really feel like sleeping.  Strange.  I\'ll probably sleep in tomorrow until 12 and then be a pain in the ass all day.  Such is life.  Anyways, I\'m going to attempt sleep.  I\'ll put up a poem tomorrow or something. '),(14515,3661106,'16,June,2004','2004-06-16 00:00:00','   The Fall of the Pure  A tear falls from my eye Into the sea Where has it gone? This thing called humanity My tear is lost In the great abiss of purity where dwells a style so sweet and free The swells of sweetest revelry  where friends are not lost and love doesn\'t end where the signs of the time are not peoples\' sins where hatred and darkness are not known where the beautiful are not alone  it lightens up each soul with truth and all are hearkened to the light of youth such in these souls  is purity a neverending  and lovely chastity  but while man reigns so does hate it floods our souls and swells over the gates the purity wanes the truth wavers the hate holds fast and we await a saviour  When will he come to take away our doubt For all this pain Can\'t be what existance is about What light he giveth me I pass to you And hope that your soul Will to hope hold true '),(14516,3371580,'26,May,2004','2004-05-26 00:00:00','   WOW!!! That was interesting... Well I kissed her for the first time in the five years I\'ve known her and it was great well we started by median up at the hotel, You know the usually parent see us together and tell us how cute we are together and we did that, then we went to go see shrek 2 and we sat there and held hands and just flirted and everthing, then we found a place to sit and then we sat and talked awhile but then I leaned over and kissed her!!! And it was like fireworks when we kissed...But anyways she cut her hair its cute but I liked it better when it was long, But I still think she looks good...But there will be pics of me and her on my picture album...                                                 well its late so laters,                                                                 Shaun  '),(14517,3371580,'25,May,2004','2004-05-25 00:00:00','   Its time to leave im typein as i walk out the door but anywayz im kinda nervous and everything and im realy gonna miss all my friends but ill be back in no time but im signin off the net now so i will post wen i get back from the movies tonight....                                                    laterz....  '),(14518,3371580,'25,May,2004','2004-05-25 00:00:00','   I just setup a photo album to go along with this blog but there are a few pics that are from before summer but im gonna go crash ill update in the AM laterz... MuCH.......         Just Klick the title and you get there pretty quick....  '),(14519,3371580,'25,May,2004','2004-05-25 00:00:00','   Well im Sutton here adding to my sitting and im actually kinda nervous cause I haven\'t seen Jenna in two years and miss her so much.  But anyways im getting packed right as im typing this, but anywayz im cooking lunch right now. Man its just like bam I leave and i cant see anyone for like a month and a half and im not gonna see the ocean or anything but nice grassy feilds and tractors but anyways im gonna post tomarrow morn and tomarrow night so catch yall on the flip side                                                           much laterz,                                                               Shaun  '),(14520,3371580,'17,May,2004','2004-05-17 00:00:00','   Here I am sittin on my ass just starting this blog for this summer while im on the road so my friends can see how much fun im havin, not!  But anywayz i will be posting any time i can get a net connection so i will post at my first stop in Griffin GA wen i visit my first and last love Jenna so i will post on the may 26 after i get back from the movies with her.                                                    lataz,                                                      Shaun  '),(14521,3371580,'02,June,2004','2004-06-02 00:00:00','   Well im probly missin out on everything in middleburg but its pretty gool up here in \"Dead\"wood SD Its been fun on the road, i got to drive across the state today and i have no cell service wat so ever witch sucks nuts but i am stealing my net so i can only get on ever once in a while... but ill talk to yall in a while so laters....  '),(14522,3376440,'18,May,2004','2004-05-18 00:00:00','   This useful and handy little tome on radio marketing for local/community stations - and a useful quick guide for stations on the cusp of \"re-engineering\" is available (all but a few chapters) on this link:  urlLink Participative Marketing for Local Radio (1) Marketing and radio  It\'s worth ordering from the publisher in Australia, too.  '),(14523,3376440,'18,May,2004','2004-05-18 00:00:00','   Click on  urlLink i95.5 FM  '),(14524,3376440,'18,May,2004','2004-05-18 00:00:00','   A list (not exhaustive on Caribbean English radio stations currently streaming, frequently updated. Click on  urlLink Caribbean Radio Stations On The Internet  '),(14525,3376440,'18,May,2004','2004-05-18 00:00:00','   urlLink Talk Radio / Radio Talk : History  '),(14526,3376440,'18,May,2004','2004-05-18 00:00:00','   While Caribbean broadcasters shut down their Medium Wave (AM) radio transmitters one by one, reports of the death of AM in the United States, whose economy we ape, are greatly exaggerated... AM RADIO YESTERDAY, TODAY, AND TOMORROW   Every day, Americans spend more time with radio than TV or newspapers. Radio is everywhere. If you want to buy a car without a radio youâ€™ll have to special-order it. Talk is radioâ€™s most-listened-to, most Sales-opportune format. For 30 years in this industry, Iâ€™ve had lots of fun, worked real hard, and made a great living. But radio isnâ€™t all I do. Iâ€™m also in retail, and am myself a radio advertiser. Iâ€™m a big, big fan.   Every industry is changing. Like airlines, banks, and the family farm, life is very different in radio today than it was â€œwhen we got into the business.â€ And we can already see that radio will change even more in the near future than it has in the recent past. Itâ€™s human nature to dread change, but I am very optimistic about radioâ€™s future.   What wonâ€™t change is whoâ€™s winning. Itâ€™s been said that if the government confiscated all wealth and issued every American the same amount in new currency, the same people would end up rich and broke a year later. Radioâ€™s winners will continue to be the stations that are most alert and resourceful, that have the most fun, in short, those that care. Places where employees high-five each other when something great comes out the speaker.   When Marty Riemenschneider invited me to speak at this convention, I asked him what I could say that would have the most value. He sent me some specific questions, and Iâ€™ll give you specific answers. But what kind of consultant would I be without explaining â€œthe latest trends?â€ So first, The Big Pictureâ€¦   Yesterdayâ€¦   Music on FM was either Percy Faith or Buffalo Springfield, music on AM was everything in-between.   Double-digit shares were common. There were fewer radio stations, and fewer other competitors for Time-Spent-Listening, as the Internet and hundreds of TV channels have become today.   Other than Paul Harvey, most of what you heard originated locally. This was a very redundant industry. There were always a half-dozen young men sitting between two turntables, even in medium markets like Springfield, Massachusetts, where I grew up.  Still, it was a buyerâ€™s market for help. Even with Viet Nam draining the demographic that produced most radio talent, there were more applicants than on-air jobs. Talent felt lucky to be working, and management often treated them accordingly.   Stations with the highest ratings usually got most of the radio ad dollars.   The Internet was â€œthe future.â€   Todayâ€¦   Listeners think of FM as music radio, AM as non-music radio.   Super-successful stations have single-digit shares. The aggregate cume of many clusters is now greater than the declining circulation of the metropolitan newspaper, though print rates still dwarf radioâ€™s.   As self-serving as the following observation sounds coming from a vendor who sells systems, this historic deregulation has endowed owners with more stations than they have the think-time to manage. And itâ€™s typical for more attention and resources to be allocated to defending topped-out music FM shares than to growing AM shares that are sooooooooooooooo much more Sales-friendly.   Most of the programming on non-music AMs comes from another city, and lots of it has franchise value. But at any given moment, there are no human beings in AM control rooms, even in big markets. I knew that The Talk Radio Revolution was over when visiting a client station and seeing them learn that one of their AMs was off-the-air only when a concerned listener phoned the switchboard.   Recently, some AMs who talk-the-talk about being the marketâ€™s information button failed to walk-the-walk when weekend how-to shows went uninterrupted by live coverage of the John Kennedy missing plane story. Station-in-a-box just chugged on.   The dark side of an 11,000 Dow is that service has deteriorated in a sellerâ€™s market for help. As a consumer, you suffer this every day, when receptionists who ask you to hold disconnect youâ€¦or you stand at a retail counter waiting to be waited on while employees chat with each other. Radioâ€™s version of this is the uninformed Talk host, or the local newscaster reading the newspaper on-air and telling you what happened yesterday. Want to stand out from the other stations? Act like Nordstrom! Over-deliver! Hire a great receptionist! After all, this person is responsible for 99% of your stationâ€™s off-air PR.   With so many mediocre career radio performers lowering the bar, some of the biggest on-air salaries now go to people who havenâ€™t spent their lives on-air. Before becoming Minnesota Governor, Jesse Ventura parlayed his wrestling persona into a Minneapolis morning show; and WJR/Detroitâ€™s Mitch Albom and WRKO/Bostonâ€™s Howie Carr established themselves in other local media before hosting their successful afternoon shows. Are there â€œreal peopleâ€ in your market who should be on your air?   Two consultant buzzwords youâ€™ll be hearing a lot are â€œrepurposingâ€ and â€œconvergence.â€   Example of repurposing: That gardening show is only on your air 2 hours on Sunday morning. Why must that be the only time you make gardening information available? How about reiterating your expertâ€™s shtick as a point-and-click menu of gardening topics on your stationâ€™s web siteâ€¦sponsored by your local advertiserâ€¦and paid for by their national co-op or vendor bucks?  Examples of convergence: ABCâ€™s NFL Monday Night Football Pre-Game Show from originated from Baltimoreâ€™s ESPNZone restaurant. In the process, several Disney-owned brands and assets cross-promoted each other. Dr. Laura plugs a catalog of â€œNow Go Take On The Dayâ€ sportswear; and rather than giving away cheap T-shirts, smart stations sell high-end logoâ€™d denim shirts.   The Internet is no longer â€œthe future.â€ Though most stations donâ€™t yet make real money on-line, many GMs will soon awake in a cold sweat realizing that they canâ€™t meet Melâ€™s goals without this revenue stream. Rather than the station web site being a distraction or sideline, smart stations make it someoneâ€™s job.   Converge your on-air and on-line programming, and dollars will follow. Example: The day after the USA women won that World Cup soccer game, America Online was polling members:   Why do you think Brandi Chastain took off her shirt?  To show off her fabulous abs  To get a sports-bra endorsement  She was hot  Male athletes do it, why shouldnâ€™t she?  As one of the worldâ€™s best soccer players, she was celebrating an amazing victory.  Do that on your web site. Put a client ad panel on that page and name the poll after the client.  Lots of stations are still peddling dots-and-spots, bidding-down the value of inventory by selling against other stations to â€œget on the buyâ€ made by agency radio buyers who are too young to understand AM radio. But if you remember one thing from this whole rant, itâ€™s the mantra I chant to my client stations, and it applies to both Programming and Sales: Do what only you can do.   View Programming this way, and youâ€™ll exploit the value in your franchises, like play-by-playâ€¦or Rush Limbaughâ€¦ or being the marketâ€™s Weather Channel radio affiliate. â€œNOW YOU CAN GET THE LATEST WEATHER CHANNEL FORECAST EVEN WHEN YOUâ€™RE IN THE CARâ€¦OR IF YOU DONâ€™T HAVE CABLE!â€   Ditto for Sales: When I managed USA TODAY Sky Radio, we created avails with a unique listener profile: business travelers listening live aboard Delta, United, and Northwest Airlines. Those spots werenâ€™t shotguns, they were laser beams. But what got us the Toyota Avalon buy was our value-added: We included an ad panel on the monthly Sky Radio play-by-play schedule you found on your meal tray. This is what cops call â€œaccess to restricted areas.â€ What doors can you open for advertisers?   Tomorrowâ€¦   AM stations that feel lesser than FMs today will get the last laugh in the digital future, when In-Band-On-Channel multiplexing turns WXXX into WXXX1, WXXX2, WXXX3, WXXX4, etc. Itâ€™ll take a few years for new receivers to proliferate, as with UHF TV. But when it does, AM will sound as good as FM, and can offer Americans whose lives keep getting faster and faster just what they wantâ€¦instant gratification.   Some TV stations have already shown us how, swapping Retransmission Consent rights for a second cable channel they use to constantly replay newscasts, or show logoâ€™d weather graphics. So radio, which will always be the #1 in-car medium, and AM (the information band), will be even more handy and relevant to busy people. While Rush is pontificating on WXXX1, you can be constantly playing and updating your Weather Channel forecast on WXXX2; with neither interrupted by the afternoon baseball game on WXXX3, or the replay of that great interview your morning show had earlier when many listeners might not have been tuned in.   Build key images and encumber programming franchises now, and youâ€™ll be known for them later.   And the answers to your specific questions areâ€¦   Q: â€œ3 person morning shows: a good idea?â€  A: Hard to generalize, since opportunities vary locally. One advantage of an ensemble: separate vacations. With only one person gone at a time, youâ€™ve never got the â€œBâ€ team subbing.   Q: â€œWhat to pay on-air people?â€ and â€œGood incentive/bonus plans for on-air people?â€  A: Peg rewards to Sales rather than ratings results. Example: a smart GSM I know preps for a new business pitch by inviting all on-air people to submit spec spots. Then, rather than asking the prospect yes-or-no, the AE asks which spot works best. And the winning talent gets a cash award. Some stations pay talent a small fee for live spots, or a bonus for new accounts renewing. Results can be dramatic.   Q: â€œAre we playing too many commercials today?â€  A: Read The Impact of Higher Spot Loads on Radio Listeners, FREE at www.EdisonResearch.com   Q: â€œWhat other incentives do you recommend, other than money?â€  A: Whatever you award is most meaningful when itâ€™s recognized as an award, rather than simply an entitlement. But what gratifies on-air people most is so obvious you might overlook it: listen to them. Hey, they talk for a living! Bring them into brainstorming meetings that they otherwise feel left out of. Even if you donâ€™t do everything they suggest, theyâ€™ll feel very included; and youâ€™ll get better execution.   Q: â€œHow do you see AM radio competing against digital radio (i.e., XM, CD Radio, etc.)?â€  A: National media can never do what solid local radio can: be local.   Q: â€œHow will â€˜virtual radioâ€™ (radio on the Internet) impact us?  A: Cyber-surfers have already demonstrated that theyâ€™re not as interested in hearing distant broadcast stations as other on-line audio (i.e., the LAX tower talking to pilots). But thatâ€™s a novelty, compared to how day-to-days useful a hustling local station can be.   copyright 1999 Holland Cooke  '),(14527,3721842,'29,June,2004','2004-06-29 00:00:00','   urlLink    well...its almost 1 in the morning. cant seem to fall asleep tonight so i thought i should update the good ole blog. still no gmail invite in case you care...  as for today\'s picture, this one is from my trip to the beach this past weekend with two good ole pals, shell and boyle. as you can see, im having lots of fun at the beach. can you find me? if you can locate me in the photo you will be eligible for a special prize.  see you all soon.  otto&nbsp; urlLink  '),(14528,3721842,'27,June,2004','2004-06-27 00:00:00','   urlLink    well... its been a good weekend, went to the beach yesterday and you\'ll see more from the beach adventure this week. today read the newspaper in the hammock, went for a bike ride and tried out the new clipless pedals. only fell 1 and a half times so i think that\'s pretty good. now i\'ll head off to church, maybe barbecue my dinner, then spend some more time in the hammock... ahhhh life is good.  see you all soon otto&nbsp; urlLink  '),(14529,3721842,'24,June,2004','2004-06-24 00:00:00','   urlLink    so heres today\'s picture. its a picture of me on the tower of terror in disney world. thas my dad in the center in the green coat. as you can see, he is totally terrirfied by this tower. i am to the left of him. i am equally terrified to be in the tower. but look at the sistah in the front row in red. shes totally cool about being in the tower of terror. shes all like \"yeah... what... this tower is my beeeotch. lets roll out...\" contrary to that woman\'s expression, i must tell you, my friends, there was much terror in that tower... that tower of...  terror.&nbsp; urlLink  '),(14530,3721842,'23,June,2004','2004-06-23 00:00:00','   ok everyone... thanks for the couple of comments... indeed alan, the people who search for \"Barely Legal Teens Naked Celebs Pics Videos 100% XXX Asian\" are just the type i want to attract here. so yeah... if you\'re one of those looking for such stuff, theres plenty to find here and its all real and theyre all really really good looking. just send me an email with your credit card number and exp. date.  as per shel\'s request, everyone take a look at this  urlLink article ...  thrashers fries are so good. right now im thinking about how this weekend i will eat them until im sick. and then after i get sick how i will still eat more as if i didn\'t learn my lesson.   kevin, still no gmail invite... im thinking ill get one in a week if i keep on posting everyday.  anyways... check back later for today\'s picture and the ensuing heated discussions.  see you all soon. otto  '),(14531,3721842,'23,June,2004','2004-06-23 00:00:00','   urlLink    ok... 1 blog and 1 comment later, thanks alan... i\'m gonna go ahead and post what i said i was going to do. this here is a picture that i took during my first ride on my recently purchased bike. i would say that \"i like this picture\" its the washington monument... theres some pretty trees... they have flowers. all around very nice. except for the chick in the corner. she totally ruins the scene and i get quite frustrated when i look at the picture and absorb in all its goodness and then remember that the girl is there in the right corner. i\'m sure she\'s really great in real life but in this picture, i must say i dislike her with all my might.  anyways... back to work. please post a comment so ill know if people are reading this. you dont have to register or do anything funny to post either.  talk to you later.  otto&nbsp; urlLink  '),(14532,3721842,'22,June,2004','2004-06-22 00:00:00','   hi all, this is my first blog.... i dont know what im gonna do with this thing. i mainly signed up because i hear if i post a couple of times to this thing ill get a gmail invitation.  anyways... i think maybe ill put pictures here... and if anyone here looks at them then they can say something about them... like \"i like this picture\" or \"i dont like this picture\" and then ill reply and post here something like \"i agree with you\" or \"i disagree with you\" and it should be a grand time for all.   what do you think? what should i put here? (other than porn) if you dont know what i should put here at least say hi to me and let me know that people will actually look at this. i promise to post much more witty, insightful, intelligent, thought-provoking, compelling, smart, fresh, funny, original, and witty, and intelligent, and insightful stuff if i know people are looking at it.  talk to you soon otto  '),(14533,3721842,'22,July,2004','2004-07-22 00:00:00','   hello all, well i broke my promise and didnt post pictures sunday, but here they are for real this time. in celebration of this coming weekend\'s beach trip and hopefully crab fest, i will finally post the rest of the pics from my first beach trip with shel and boyle...     thrasher\'s fries are so good. as you can tell both shel and boyle think they are money. shel is a nice guy. he will probably blog here soon about food that is good. trust me, shel is probably one of the world\'s preeminent authorities (along with myself) in classic, as well as, modern foodology. im thinking if you ever meet shel, you would have to at least refer to him as dr. shel. anyways... hopefully that was enough shameless plugging of shel that he will continue to read this blog and not cut my readership by 33%. oh by the way, did i mention that shel is very good looking and very much what the ladies want?      anyways, back to the beach trip. the boardwalk is always tons of fun at the beach. here you can see me having tons of fun spraying pirates with water. haha... silly pirates....     well... as you can see after all the fun at the beach, one can get really tired, especially after eating thrashers, grotto pizza, a couple of lemonades, water, a soda... yep its a tough life at the beach.  so that was my beach trip. maybe next week you\'ll get to see some pics of guys eating crabs. oooh... doesnt that sound delicious?  talk to you all soon. otto  '),(14534,3721842,'17,July,2004','2004-07-17 00:00:00','   sorry the posts aren\'t coming as frequently as before. once a week is still pretty good right? first off let me respond to some comments... no alan, gmail doesnt mean gay mail or girly mail or g string mail or... galapagos islands mail. it is a free email account from google who owns blogger.com. you get 1gb of storage which is a gigando amount of storage which i intend to fill with pornography, bootleg movies, and illegal mp3 downloads - the only 3 useful things to use the internet for. this is contrarty to what most people believe the internet is useful for.  in fact, practically no one uses the internet for meaningful research, doing online banking, or keeping in touch with old friends. even if people do use the internet for such activities there are vastly more people using it for downloading porn, and illegal movies and music. anyways... thats what gmail is.  right now not everyone can get it. you must be invited to receive it and i hoped that starting this blog might get me one of those coveted invites. contrary to what jd said, you are basically not cool if you do not have a gmail account. it is a known and proven fact that people with gmail accounts are money. they are more likely to be cool, successful, and have a much higher percentage of scoring with insanely beautiful women. so you see jd, it is very important for me and also for you, and for that matter, everyone to get gmail accounts. if we dont get one soon or have to wait till gmail gets released to the general public we might as well shoot ourselves in the face.  ok, enough with that now on to today\'s pictures... actually i have to go now and take a shower, so i promise tomorrow i will post pictures.  talk to you all soon. otto  '),(14535,3721842,'12,July,2004','2004-07-12 00:00:00','   sorry for not blogging in so long. things have been sort of busy and i have switched to a new computer. its also become clear to me that i will probably not get gmail by doing this so these past couple of days ive been deeply pondering what to do with this blog... and after many a torturous night, wrestling with my inner demons, and searching the deepest depths of my soul... i have decided that.....  the blog must go on....  so loyal readers (sasso, shell) fear not, for more witty, interesting, compelling, funny, light hearted, serious, intriguing, enlightening, intelligent, gripping, and witty, and light hearted and interesting commentary on the issues that matter most to us will be coming soon. there\'s also been massive interest from numerous readers (sasso, shell) to open up my blog so that others may try their best to provide witty, interesting, funny, serious, compelling, enlightening, gripping, witty, and funny posts. however, this blog shall remain otto\'s blog because... i have no idea how to change the name of it. in any event... if you want to blog about stuff email me and ill figure out how you can do that. either way... if you post a comment that gets automatically posted on the blog.   so yeah... look forward to guest blogs about pirates and dinosaurs and why they are so money and why ninjas are big fat hurtings... and also blogs about pictures ive taken and maybe blogs about food i\'ve eaten along with blogs about food shell has eaten.  sorry no pictures today. i dont know how to do that with this new computer.  see you all soon otto  '),(14536,3350341,'31,May,2004','2004-05-31 00:00:00','    This really pisses me off!  I fork out $180 for my pool membership, look forward to spending Memorial Day lazing in the sun with a good book, and lo and behold, it\'s pouring with rain!  I guess it\'s still better than being at work...  Last night we saw a god awful movie called  Raising Helen  - with Kate Hudson. As charming as she is, she could not redeem this boring, soppy, stupid movie. Americans really don\'t have a clue how to make a good movie...occasionally there is some fluke and they bring out one good one, but movies should be left to the Europeans!  My weekend definitely was not the same without The Sopranos! HBO decided against screening the season finale on Memorial Weekend, so we have to wait until next week to see what happens!  I am positively plutzing to find out what transpires - it\'s undoubtedly the best show on TV!  The Mouse Spouse thought he had to work today (he is an engineer for CN8 cable TV station) but he called earlier to say he had been incorrect, and CN8 was not doing live shows. Yippee so we get to spend the day together!  He is due to arrive soon bearing tasty treats from Giovannis - triple deck sandwhiches, cheese hoagies and other delights. Weather is so shitty I guess we\'ll just chill, play our newest Play Station Hockey NHL pro game (I beat him yesterday!) and relax.  Tomorrow it\'s back to work, urgh!   '),(14537,3350341,'30,May,2004','2004-05-30 00:00:00','    It\'s almost 1am but I\'m not sleepy, so I thought I\'d pen a few words before I push off to sleep.  I just returned from \"the mouse spouse\" (my boyfriend) after a fun evening of playing hockey on playstation and watching the movie  21 Grams . I didn\'t know what the fuck was going on for the first half hour - I hate movies that go back and forward in time so you are left confused and irritated. There were interesting moments, especially enjoyed seeing Sean Penn\'s ass - he\'s a great actor. Loved him in  Mystic River  too.  Today was the unveiling of the new WW2 Memorial on the Mall in DC. I find it a bit odd that it took sixty years to pay tribute to the men and women who fought in the war, but nevertheless, I guess better late than never. It was moving to see all the old toppies in their medals and regalia, and trying to imagine the horrors they had been through so long ago. I know this is going to sound trite, but war is so fucking stupid! All the death and injury and misery it causes, why don\'t people learn from history and stop trying to constantly destroy one another. Doesn\'t every human being at the end of the day really want the same thing? To have people/family to love, to eat, to work and to get on with their lives in peace?   But every once in a while, far to often, some prick somewhere (apologies to bona fide pricks - I don\'t mean to cast dispersion on penis\'s everywhere) arises which stupid fucking ideas on world domination or destruction, masses get sucked in by this asshole\'s promises or grand aspirations, and BOOM, innocent young men and women go marching off to war and either return in body bags or return emotionally scared for life or (and this never ceases to amaze me) return spewing the glory of war - proud to have fought for land and country.  SM   '),(14538,3350341,'28,May,2004','2004-05-28 00:00:00','    I am full of wonder at the amazing life of the little cicada. Here in DC, particularly where I am, in an Arlington suburb, there are zillions of cicadas buzzing around. I didn\'t know anything about the amazing lifecycle of the cicada until coming here. Apparently every 17 years, the creatures emerge from the earth, fly up to the trees, mate (male dies immediately - typical! no hugging or holding post sex) and then the female lays hundreds of eggs in the tree and then she too dies. When the eggs hatch, the little larvae go back underground and stay there for another SEVENTEEN years! What is so enchanting about these bugs is the incredible noise they make. When you go outside here or drive in wooded areas, you hear this deafening buzz!  I have grown to really like it, it\'s somehow soothing.  Washington is saturated in cicadas, you are constantly stepping on little corpses...there were even some in the Holocaust Museum yesterday! Symbolic? (By the way, if one should fly into you - they are blind - while rather unpleasant, they don\'t sting or bite.) Here is a picture of one:   .   I\'m off now to Union Station to return to Philly and my sick, much beloved boyfriend who has had some horrible virus for a week.  SM   '),(14539,3350341,'27,May,2004','2004-05-27 00:00:00','    I\'ve been visiting a friend here in DC the last few days, my first time here. Interesting city, but I chose a crazy time to visit - it\'s memorial day weekend and there are thousands of veterans and others in town to unveil the new WW2 memorial that was 16 years in the making - never mind that the war ended half a century ago!  Washington is really full of memorials honoring poor men and women who went off to horrid wars for god knows what reason or purpose, and never came back alive. The Vietnam Memorial was very moving, that black granite, full of names of young men and women who lost their lives, and for what? I also visited the Lincoln memorial. Can\'t say I knew much about Abe, I\'m horribly ignorant of American history, so the museum adjacent to the monument educated me and I learnt to appreciate old Abe and his sentiments against slaverly etc which seemed way ahead of his time.   Today I spent 5 hours at the HOlocaust Museum - very powerful and draining. My head is full of horrific images, brutality and suffering that I will never ever understand or fully fathom. I found the exhibition a lot more powerful than Yad VAshem in Jerusalem, somehow the museum here seemed more holistic, comprehensive and had a greater impact on me. I could have spent a few days there, there was much I still would have wanted to see, movies to watch etc....  What was most moving for me were the images of the shtetl in Poland - a tower of photos of lovely people enjoying their lives in Pre-War East Europe, a shtetl that was reduced to ZERO after the war. I kept looking for faces of my own family who were murdered in the forests of Riga - only my grandfather left Riga for South Africa in the early 30\'s - the rest of his family - his parents, 7 siblings, their spouses and children (my grandpa was the youngest) were all slaughtered by Latvians - only my uncle/cousin Julius survived after years in the camps. He passed away last year, I never did learn of his war story, how he survived etc... I wonder whether he saved his memoirs on tape, I would very much like to hear it.   Anyway, I\'m totally exhausted from the Holocaust museum and walking my feet off around the mall - we had to park a mile away...  Off to sleep though i\'m sure my dreams will be full of horrific images....  SM   '),(14540,3350341,'15,May,2004','2004-05-15 00:00:00','    May 15, 2004   Greetings, and welcome to The Scarlet Mouse Blog! This is my first foray into the Blogsphere!  I\'m pretty psyched.   It\'s a muggy Saturday morning here in Center City Philadelphia. Thank God the weather has warmed up, I cannot tolerate the east coast winters!  I am used to warmer climes as I hail from sunny, tempestuous South Africa. I grew up in Johannesburg and left in 1995 at the age of 25, after completing my law degree. I traveled to Israel where I thought I\'d hang out for a while, but it turned into a 6 year sojourn - a mostly great time in my life when I got into computers, met great people and lived the Jewish dream for a while. I left in July 2001 mainly because I became an anxious wreck living in Jerusalem and constantly fearing becoming a terrorism statistic. I set off for Florida, USA where I had an ex-boyfriend. Dumped him but adopted Florida. Stayed there for a year and then landed a job with visa sponsorship in Philly, so off I went, to a city I\'d never spent more than 4 days in...  Mostly my life here has been really good. I work for a Jewish non-profit, live in the city and have a wonderful boyfriend who we will call the Mouse Spouse.  The big news today - or should I say positive, happy news as opposed to all the ongoing shit with the Iraqui detainees and American baddies - is that South Africa won the Soccer World Cup bid for 2010!  This is so great, South AFricans are crazy soccer fanatics and the whole country is celebrating wildly!  This is the first time the competition will be on African soil - about time if you ask me!   '),(14541,3350341,'28,June,2004','2004-06-28 00:00:00','    Every day I read the news I find myself feeling more and more angry at what is going on in the world. Sometimes I think maybe I should stop being such an insatiable newshound (mouse?), wouldn\'t I be happier if I just kept my head buried in the sand, blissfully oblivious to the daily parade of horrors around the world? I have a few friends who live like that...it\'s not that they set out to do this, they are just genuinely not very interested in the daily news. Growing up I always felt different from my friends in this way - long before the Internet I was reading the morning and evening newspapers. Part of this was an indescribable feeling of social responsibility to know what was going on in my own country under Apartheid. I knew shit was going on that was hidden or masked in the media. Every Friday I would religiously read the  urlLink  Weekly Mail and Guardian   - the so-called leftist publication that attempted to reveal what was going on in the townships and elsewhere. When I would inform my schoolmates that the SA army was murdering people just 30 minutes from our safe, white neighborhoods - they\'d look at me as if I was crazy, label me \"the liberal\" (ha!) and repeat the crap their parents would tell them. In truth, my dad would tell me the same stuff, but I had my older brother B around to educate me about the real truth. For that I will always be grateful.  Anyway, today it occured to me that I needed my own color-coded barometer akin to Rumsfeld\'s pretty Homeland Security Advisory System - only mine would reflect my level of \"pissed off-ness\" based on daily events around the world.        Today is an elevated green day. A 49 year-old man and a three year-old child were killed in a missile attack next to a kindergarten in Sderot, near Gaza (the child\'s mother was critically injured). The attack on Sderot is the first in which one of the hundreds of rudimentary rockets fired by Hamas from the Gaza Strip in the past three years has proved fatal. This came after an attack on Sunday in which Palestinian terrorists dug under an Israeli army post in the Gaza Strip and blew it up, killing a soldier and wounding five others.  Then CNN is now reporting that the American Army Pfc. Keith \"Matt\" Maupin, held hostage in Iraq since April has been executed. Video released by his murderers show him being shot in the back of the head.  In the next two days, five more innocent hostages (three Turks, a Pakistani and another American soldier) held in Iraq will probably be beheaded as no country is willing to give into the demands of these scumbags. (The handover of power from the US to an interim Iraqui government took place today - BFD - like the violence is going to end now!)  The American soldier, Cpl. Wassef Hassoun is a Muslim Arab marine who worked as an Arabic translator. Hassoun\'s father pleaded from his home in Tripoli,  \"I ask them for the sake of God, Prophet Muhammad and their children to release my son, and I thank them and they will have a great reward from God.\"   Wake up Mr Hassoun, these bastards are not going to spare your son\'s life because he is an Arab Muslim! He is an American first. An American-Arab Muslim - no doubt perceived as a traitor to Islam and his people. And with all due respect Mr Hassoun, these guys won\'t be getting any rewards from God, even if your son is spared...  SM   '),(14542,3350341,'23,June,2004','2004-06-23 00:00:00','    Tonight I met the son of Julius and Ethel Rosenberg - Robert Meerpohl (he was adopted a few years after their execution). He spoke at Robin\'s Bookstore about his parents, his traumatic childhood (he was only 3 when his parents were arrested), post 9/11 America (McCarthy-era lessons for Ashcroft\'s America), and the amazing work he does now as head of  urlLink The Rosenberg Fund for Children . The fund was established \"to provide for the educational and emotional needs of children whose parents have been harassed, injured, lost jobs, or died in the course of their progressive activities and who, therefore, are no longer able to provide fully for their children.\" This was an eye-opener for me, I niavely believed there were no \"political\" prisoners in America, or that there would be no kids who would need this kind of assistance in today\'s America.  I bought Meerpohl\'s book, \" urlLink An Execution in the Family \" and look forward to reading it.  This morning an amazing child passed away.  urlLink Mattie J.T. Stepanek  the young poet whose four volumes of work courageously detailed life and death, love and loss, faith and hope, has lost his own battle with the rare form of muscular dystrophy that had taken his three older siblings lives as well. He was thirteen years old. I have often seen Mattie on TV and marveled at his amazing bravery and optimism despite having to live with his awful muscular disease that kept him confined to a wheelchair, hooked up to oxygen, respirator and trach tube. As someone who was born with a very mild neuro-muscular disorder (Spinal muscular atrophy), I have always identified in some way with Mattie and other kids with muscular diseases. I always think how fortunate I am to have been born with only a very mild atrophy, able to live a normal and productive life save for having had scoliosis and dealing with respiratory challenges. Rest in peace Mattie, you were an inspiration to many. You will not be forgotten. The Muscular Dystrophy Association has established a  urlLink memorial fund  in memory of Mattie.  And now to the absurd. Saudi Arabia promised Wednesday that terrorists in the kingdom will be safe if they surrender within a month -- but after that they will face forceful consequences.  \"If they give themselves up without force within one month maximum from the date of this speech, we can promise them that they are going to be safe.\"  Crown Prince Abdullah said all such people would be dealt with fairly, in accordance with Islamic law.  Now that just cracked me up!  I laughed out loud - \"in accordance with Islamic law!\" Last time I checked Islamic law (Sharia) you could be stoned to death for adultery. Sharia calls for punishments such as the severing of limbs, stoning, hanging and flogging, or beating. Countries that do practice Sharia include Saudi Arabia, Pakistan and Malaysia. Saudi Arabia is perhaps the most well known practitioner of strict Islamic law ( urlLink source ). So come on all yea terrorists, turn yourselves in and face the \"fair\" penalties under Islamic law.   My guess is there won\'t be a queue of terorists outside the palace...  SM   '),(14543,3350341,'22,June,2004','2004-06-22 00:00:00','    Iraqi insurgents have killed the South Korean civilian they were holding hostage - his beheaded body was found earlier today. The murderers have the audacity to call themselves the Monotheists! How bloody preposterous - thugs commiting atrocities in the name of God.  In another show of moderation, tolerance and humanity so common in Iraq, Saudi Arabia and other Islamic outposts - CNN reports that the dean of law at Mosul University and her husband were found dead Tuesday morning outside their home in the northern Iraqi city of Mosul. The dean was the only woman to hold such a prominent position at the school and voiced moderate opinions, she therefore deserved to die.  What is Bush doing with himself instead of setting out to destroy these militant extremists who are cutting off heads with gay abandon, killing US soldiers and murdering civilians? Let me guess, he\'s out playing golf! It\'s time to stop sitting pretty and merely spewing pathetic condemnations every time someone gets beheaded or murdered. It\'s time to fucking do something! (Sorry mom, I know you don\'t like my swearing, but I\'m just fucking furious!)  SM   '),(14544,3350341,'20,June,2004','2004-06-20 00:00:00','    There is a lot to be said for dying young.  Don\'t misunderstand me. I am not suicidal or ready to exit this world. Not by a long shot. I love life and living every day. I just don\'t want to ever be old and helpless.  I spent the afternoon today visiting the Mouse Spouse\'s 91 year old grandfather at probably one of the nicest Jewish old age homes around. It has manicured lawns, caring staff, decent accommodations, daily activities and entertainment and none of that obscene stench of \"decrepitude\" you usually encounter in hospitals or old age facilities. However, anyway you dress it up - it is still a place where people go when they are too old or frail to take care of themselves. It is full of people with blank or dazed expressions on their faces, people in wheelchairs or with canulas of oxygen up their noses, and sadly, some forgotten people left alone to die while their children occupy themselves with other things.  As I sat there, listening to the afternoon\'s entertainment - a young women singing golden oldies to an assembled crowd of ambivelent looking old people, I was filled with a deep sadness. It is impossible not to imagine oneself living in such a place. I pictured myself, 3 foot tall, frail and helpless sitting in a wheelchair listening to some young woman singing Madonna songs...It was too painful to think about, and I could tell the same thoughts were running across the minds of the other visitors there to celebrate father\'s day with their aging parents/siblings/family.  By the looks on their faces, they too were thinking, just shoot me. I don\'t want to ever end up here.  SM   '),(14545,3350341,'18,June,2004','2004-06-18 00:00:00','    Let\'s start with the sublime. Last night I trekked off to the Free Library to hear Helen Fielding - author of Bridget Jones\'s Diary - speak. It was great - she is just as witty as you\'d expect. Although she denies vehemently that she is Bridget Jones, it was clear to me that she IS indeed her character - it shone through in her humor and funny stories.  I met my new friend Veronica at the talk - the first women I\'ve met through Craigslist. What a pleasant surprise - a lovely, intelligent woman. After the library we headed to the Rose Tatoo Cafe for drinks and dinner. We hucked like old friends and had much to talk about. Cool.  And now to the horrific. I had just sat through 5 hours of meetings this morning (no, that was not horrific - tiring and tedious perhaps...) when I came back to my computer and invariably clicked home to see if I had any new Yahoo mail. I caught sight of the news headlines to read that Paul Johnson (of NJ), the Lockhead Martin guy kidnapped in Saudi Arabia, had been beheaded. First Berg, now Johnson.  I just went cold. I thought of his family. I thought of how terrifying it must be to know you are about to be beheaded...But mostly I thought of what fucking bastards, scum, these murderers are. How does one human being do  urlLink this  to another  (Warning: clicking on \"this\" will take you to the horrific pictures of a beheaded Johnson.  Sometimes I believe we need to see horror to really understand who we are dealing with and how urgent it is to wipe these vermin off the face of the earth.) I\'m sick of Bush\'s cowardice, I\'m sick of the world condemning these terrorists but doing fuck all against them. I\'m sick of the hypocrisy!  Fifteen of the nineteen hijackers on 9/11 were Saudis and Saudi money has always funded Al Qaeda.  And yet America still calls Saudi Arabia \"our friends\"? Don\'t you find it rather convenient that the Saudis claim to have killed  urlLink Al Qaeda chief (and murderer of Johnson) in Saudi Arabia Abdul Aziz al-Muqrin  as he was dumping the body? Hmm, sounds just a wee bit suspect to me!  Michael Moore is on Letterman tonite - I\'m going to stay up and check it out. I\'m not too fond of Moore the person, but I do have great respect for his work. I thought  Bowling for Columbine  was fascinating, and I can\'t wait to see  urlLink  Fahrenheit 9/11   in which, among other things, reveals the connections between Saudi Arabia and terrorism - and the dubious relationship between Saudi Arabia and the Bush administration. I dig the movie poster of Bush and Moore holding hands like great pals of yore! Brilliant.  SM   '),(14546,3350341,'17,June,2004','2004-06-17 00:00:00','    Tomorrow night Helen  Fielding , author of  Bridget Jones\'s Diary  is speaking at the Free Library. I absolutely loved that book, it was hilarious from cover to cover, and I\'m pretty excited to have the opportunity to meet Helen and hear about her new book. This is one of the reasons I love being in the USA - in Israel or South Africa I never had the opportunity of meeting \"famous\" people. Well not exactly true. I guess there aren\'t many people who can say they\'ve seen Nelson Mandela in person (I was at his graduation at my alma mater, The University of the Witwatersrand) - can\'t get much more famous than that!  Not only will I be meeting Helen, but I am also hooking up with one of the women  friends  I\'ve met through my ad on Craigslist. She sounds lovely and interesting, and it should be a cool experience meeting someone new!  Today I recieved sad news that my friend lost her father to pancreatic cancer. He was only 61,  funeral  was today in NJ. My friends, his daughter, live in Israel where I met them both although they are both American. My heart goes out to my friend, one of my deepest anxieties/fears is losing my own dad. I mean I worry about both my parents, but my dad is 13 years older than my mom so I naturally worry more about him. He turned 80 in March. It\'s hard to live so far away from my family, I constantly worry about something happening there and having to fly out suddenly. Joburg is so frigging far (16-22 hour flight), it was so much easier when I lived in Israel (9 hour direct flight). I\'m hoping to go home (SA will always be home) with the mouse spouse in January.... I love seeing my friends and fam, but it\'s such an intense, unnatural thing, to see people you love more than life itself for 2 weeks and then to not see them again for a year or two. God it sucks.  Anyway, back to my friend, I was contemplating making a sympathy call, but they are in North NJ and there is no train from Philly, and I\'m not a long distance driver...So I thought until I figure that out, I should express my sympathy somehow (seems to early to call...they are probably still in shock etc) - and in America, that is done primarily via a condolence basket. No such thing exists in Israel or SA - I only discovered this phenomenon when the mouse spouse\'s grandmom passed away a few months back. All these baskets of food and goodies arrived at the house -- the house was positively drowning in cookies, cakes and candy. In South Africa food is also a part of a shiva house - but people tend to bring meals not send snack baskets... I found the whole practice rather weird. Similarly, at the shiva I was at, all the mourners ate a big dinner BEFORE prayers. Another major departure. Practices may differ globally, but one thing is true everywhere in Jewish house\'s of mourning: Death and Food are inextricably linked!  When in America, do as the Americans do. I sent a sympathy basket.  SM   '),(14547,3350341,'13,June,2004','2004-06-13 00:00:00','    Phew, what a busy weekend, it\'s 10pm Sunday and I\'m ready to hit the sheets. Yesterday me and the Mouse Spouse went to Rittenhouse Square to browse at Barnes and Noble, and then we walked 18 blocks to the Ritz Cinemas to see  urlLink Super Size Me  - a really interesting and disturbing documentary on the causes of obesity in America. The movie focused on the fast food industry with the director/writer undertaking a fascinating experiment: eat McDonalds for 3 meals a day for 30 days straight. The results were horrific - significant sclerosis of the liver, huge increase in cholestral levels, impaired energy and sexual function, and gain of 25 pounds!  Other disturbing facts quoted include the estimate that 60% of kids born post 2000 will have diabetes - and a dimunition of their lifespan by 17-27 years!  A really thought-provoking movie, and one that every American should be forced to watch!  (I have to confess however, after the movie I had a craving for a Happy Meal...:-))  Today I went with the Mouse Spouse\'s sister and niece to Willow Grove shopping mall -- in the two years I\'ve been in Philly, I have been to ONE mall, ONCE - so this was somewhat of a novelty. Unlike most Americans and most women, I hate malls, and don\'t particularly enjoy shopping. Perhaps this is because I seldom leave a mall laden with exciting purchases!  Being a super petite, really skinny (read Freak in America) woman, finding attractive clothes is always a challenge, finding shoes is nothing short of a fucking tribulation! Clothes shopping is predominantly undertaken in kids\' clothing stores - Limited Too, Kid\'s Gap, etc...but of late I find myself almost fitting into a size zero petite. Indeed, today I was thrilled to fit into a size zero slim fit jeans at The Gap. V-I-C-T-O-R-Y!!  Sophisticated Petite is also a store I often find clothes that  almost  fit (ie. need some tailoring).   But shoes, alas, is a hideous nightmare!  I am saddled with size 1 1/2  feet - forever sentenced to shop in kids\' stores only!  Yes, American kids are pretty sophisticated these days, but I think more 12 year olds now wear adult shoes...so I\'m stuck with a pretty poor selection.  Sometimes, when I pass adult shoes stores, I get this painful pang...what must it be like to be able to choose any shoes your heart desires? I guess I\'ll never know... **sigh**  SM   '),(14548,3350341,'10,June,2004','2004-06-10 00:00:00','    It\'s been a few days since I last wrote. I am happy to report that my niece Noa has made a miraculous recovery. Four days after major surgery, she is  HOME !  This is all credit to her amazing strength and fortitude - she has flabbergasted not only all of us, but her doctors as well who had expected her to remain in hospital at least a week, followed by a stint in rehab to treat her lungs.  Kol Hakavod Noa! Thanks also to everyone who mentioned Noa in their prayers - I truly believe that this also led to this postive outcome.  Since I last wrote Ronald Reagan died. Ray Charles died. Brittney Spears had knee surgery. Tony S killed Tony B. Two people who I\'ve never heard of won the French Open. And, researchers say they\'ve found a dog that understands more than 200 words and learns new ones as fast as many children. In other words, it hasn\'t been a terribly exciting week.  As I\'ve stated before, I\'m very content with my life in Philadelphia. The only thing I\'ve felt lacking every since I got here was the absence of female friends to hang out with. I have the greatest group on girlfriends on the face of the planet. But that\'s just the point - they are dispersed all over the planet - London, Israel, Canada, South Africa, NY, Atlanta, LA, DC - and not a single one of them resides in Philadelphia!  I have long whined about this predicament and feel it\'s rather pathetic that in close to two years I haven\'t managed to find some girlfriends. The truth is I haven\'t made much effort. At 34 I haven\'t really felt like investing my time to nurture new friendships especially when I have so many close friends who have known me for years, who have shared my childhood, high school and the ups and downs of life.  I have some lovely female colleagues at work, but we seldom socialize - most of them grew up in Philly and are not lacking in family, and a social life.  This week, I decided to stop whining and get proactive. No, I didn\'t walk around the city ambushing hapless females!  I did what you would expect an obsessive techie in 2004 to do. I placed an ad on  urlLink Craigslist  website announcing my willingness to meet intelligent women interested in friendship. The response has been wonderful - I have had great replies from interesting, educated, worldly women from a myriad of different backgrounds. Many are also not American, who found, like I have, that they don\'t have terribly much in common with the Americans they\'ve met. I hope to meet some of these women next week. It\'s sort of weird, it\'s like online dating in a way, only different of course. Hopefully meeting them won\'t be as hideous as blind Internet dates!  Watch this space for more...  SM  PS Happy Birthday to my wonderful friend Croc in South Africa!   '),(14549,3350341,'06,June,2004','2004-06-06 00:00:00','    It is after midnight and thousands of miles away, in Jerusalem, Israel, my beloved 14 year old niece Noa is about to undergo a spinal fusion operation to correct her scoliosis (curvature of the spine). Seventeen years ago, in January 1987, I too had a spinal fusion operation to correct my advanced state of scoliosis. It was a horrible period of my life, but like all of life\'s hardships, one gets through them and soldiers on. My thoughts are with Noa right now. I am unable to stop imagining her lying in her raised hospital bed preparing to be wheeled in to the operating theater for her 5-7 hour surgery. I know just how she is feeling - it was a mixture of fear, weary anticipation and a sense of just wanting to get it over with. I think harder than undergoing the surgery and recovery, was watching the suffering my parents endured seeing me suffer so. That and the helplessness, being exposed and unable to do a thing for myself for about 10 days after surgery. I couldn\'t move, I had to lie on my back and rely on nurses to turn me and give me painkillers. It was bloody awful, the memories still linger.  I pray that Noa will have an easier time, medicine has advanced in the past 17 years...kids tend to be up and walking after a day.         B\'Hatzlacha Metukah, B\'Ezrat Hashem hakol yehiyeh beseder.     Update (Sunday, 3:50pm)  Thank God the operation was a success and Noa is on the road to recovery. She is in a lot of pain and discomfort, but hopefully in time this will subside as her body recovers from the surgery.  SM   '),(14550,3350341,'02,June,2004','2004-06-02 00:00:00','    Reuters reports today that  \"Philadelphia, the City of Brotherly Love, has launched the first U.S. television advertising campaign to promote the city as a gay destination. The spot emphasizes the city\'s historic attractions and its friendliness to gay travelers and is part of a three-year, $1 million campaign using the slogan \"Philadelphia - Get Your History Straight and Your Nightlife Gay.\"   Hopefully this will bring greater economic prosperity to the city, and attract not only prospective tourists, but residents too. While Philly\'s many top universities attract hordes of students nationally and beyond, for some reason, the great majority fuck off after getting their degrees... I\'m sure Philadelphia isn\'t the most beautiful city in the USA, but it\'s a pretty cool place to live, the standard of living is more affordable than other east coast cities like NY and DC, it\'s diverse and has a lot to offer.  News that saddened me today concerns the  urlLink closure of the synagogue in Vilnius , which celebrated its 100th anniversary last September. No, it was not the victim of some anti-semitic attack. No, this was not some communist plot. No, it is not because there are no more Jews in Vilnius courtesy of Hitler. The synagogue was  \"closed after fist-fighting broke out during Shavuot services between two groups of worshipers who wanted to have the otherâ€™s rabbi removed from the shul. The service also was interrupted by police, who were called in by one of the groups.\"   This situation really disgusts me. Who needs the Nazis when we can fight eachother and destroy our very own religious life? The Jews of Vilnius need to do some serious introspecting me thinks!  SM   '),(14551,3350341,'01,June,2004','2004-06-01 00:00:00','    urlLink The View from Here blog  (written  by my friend Harry in Israel) alerted me to more craziness in the Holyland. In the religious enclave of Bnei Brak, they are now  urlLink segregating streets  - women on the one side, men on the other, and never the two shall meet! Let alone look at eachother. Of course the word \"segregation\" automatically gets my hackles up - segregation, apartheid...etc...and the joys of my childhood in South Africa.  Sometimes I feel that religion is the curse of the world - it\'s a tyranny - zealots killing, wreaking terror in the name of God. How many people have died in the name of Religion?  Still on the topic of religion, I had to give a chuckle at this article in the  urlLink NY Times  (I\'ve given the Maariv link so you don\'t have to subscribe). It seems there are tiny crustaceans floating around in the NY water supply. If you know anything about Judaic dietary laws, then you know that all crustaceans are not Kosher - no shrimp, lobster, crayfish etc for we Jews!  Anyway, this latest discovery has the collective tits of all NY Jews in a tangle - and water purifying system sales are soaring!  I particularly laughed out loud at the statement describing these organisms as \"little armadillos\" - I guess armadillos are not Kosher either? Hmm, I should check! Anyone know if they chew the cud or have cloven hooves? They do have long shnozzes..does that help? :-)  Back to the topic of things South African. Ousted Haitian President Jean-Bertrand Aristide is now living it up in South Africa with his family - at the cost of South African taxpayers! I have to agree with this from the SABC (SA Broadcasting Corp):   The Democratic Alliance says South Africa cannot afford to host a leader whose country is \"far removed from our national interest\". Objecting to the arrival of Aristide in Johannesburg, Douglas Gibson, the DA spokesperson, said the money spent on hosting the deposed leader would have been better spent on poor people\'s needs.   With one in five South Africans HIV positive, and in dire need of medical services, the idea of hosting Aristide at the country\'s expense is offensive and unconscionable! Go home Aristide, wherever the fuck home is or pay your own way!  Finally, today\'s  Brave Woman Award  goes to  urlLink Ines Perez , who is believed to be the only woman known to have performed a successful caesarean section on herself.   '),(14552,3350341,'29,July,2004','2004-07-29 00:00:00','    So, yawn, the Democratic Convention is underway in Boston.&nbsp; I have watched snatches of it - and I say snatches because it nauseates me to watch more than about 5 minutes at a time.&nbsp; I don\'t have the vote here as I\'m not a citizen nor a green card holder (kind of ironic that&nbsp;I now get to experience what it is like to be disenfranchised like so many of my fellow-South Africans were for decades)&nbsp;- but as usual, I am fairly interested in what goes on around me, so I\'ve tried to learn what I can about Kerry and Edwards.&nbsp; I\'m still not sure what to make of them - I am quite put off by their pretty-boy looks (especially Edwards) and their huge personal wealth. How can someone with millions of dollars truly appreciate the needs of the average American?&nbsp; I sat and listened to speeches, the promises, Edwards tonight declaring \"there is HOPE\" and I just don\'t buy any of it.&nbsp; Has there be any improvement in access to healthcare here in the last 50 years? Is homelessness dead? Are the impoverished less impoverished? Are American children better educated now? My guess is that nothing really changes&nbsp; here, not for the masses.   It nauseates me to see the rapture on the crowds\' faces as they take in the speeches, the loud music, dancing, glitz and glamor of flashing screens etc....I think about what this circus must cost - couldn\'t the money be better spent?&nbsp; I hate the fact that you have to&nbsp; have money to get into office in this country...   If I\'m sounding communist or socialist, so be it. I don\'t see either as&nbsp;a dirty word. There are up sides to all economic systems, and downsides too.&nbsp; Lately I\'m seeing a whole, ugly side to capitalism, and it upsets me profoundly.&nbsp; Take for example how the quality of one\'s whole life in this country is dependent on a number. Your credit score.&nbsp; Take for example how credit card companies continually try to entice people into obtaining new credit cards, then get you into a hole of debt that you cannot get out of because of crippling interest and other \'hidden\' charges. I think it\'s criminal!&nbsp; And then you have me who can\'t get a fucking credit card because I have no credit history due to only having a social security number for 2 years.&nbsp;&nbsp; So I have to do ridiculous things like give my bank half my savings in return for a bank loan that I don\'t want or need, purely for the reason of trying to establish my credit. So I gave the bank x amount, they then gave me the same amount, and now I have to pay off the loan at interest of course, so I\'m losing money doing something I really didn\'t need to do, but for the chance that it MAY help me establish credit. The only up side of all this crap is that it is practically impossible for me to get into debt cos I can\'t get a credit card!   I am happy to be in America, and like many aspects of life here, but the above just sickens me to the core.   SM   '),(14553,3350341,'25,July,2004','2004-07-25 00:00:00','    Since celebrating my birthday last week, life has been more exciting than usual!   On Thursday night I went with my friend Veronica to see Cirque de Soleil. V won a CityPaper contest and landed two ace tix to the Cirque, free dinner at Christians and a limo to shlep us about. What a great evening it was. It was so cool to have a long, black and sleek limo waiting to take us to our various destinations - we got to sample a bit of the celeb life, and we liked it baby!  Cirque was unbelievable, and anything I write here really cannot do it justice.&nbsp; Many of the acts defied what we would expect human beings able to do,&nbsp; and gawking at the fabulous male specimans was a treat.&nbsp; Besides these two irritating blonde idiot woman sitting next to us who didn\'t shut up for one minute, the show was fabulous!   Other excitement this week revolves around progress in my writing endeavors.&nbsp; I saw a freelance job ad on Craigslist for a Grant Writer at U of Penn, and decided to apply.&nbsp; Turns out the job is for the biochem dept and the man in charge is quite an extraordinary character!&nbsp; His name is Dr. Britton Chance and if you google him, you\'ll see that he is the pre-eminant authority in everything to do with bio chemistry. He has three Phds, a gold Olympic medal, and at the ripe old age of 91, still works 6 days a week from 7am-7pm and cycles to and from work!&nbsp; It was an honor to meet such a man, he has the gentle manner and wry sense of humor, and one cannot but be inspired by his dedication to his field. When I asked, amazed&nbsp;how he manages to work six days a week at his age, his eyes lit up and he exclaimed, \"I love it!\"   So I landed the gig and spent much of yesterday editing a 48 page grant proposal on a hand-held breast cancer detection gadget.&nbsp; Besides the fact that I am the most unscientific gal around, I really enjoyed the work. I think deep down I am a wannabe school teacher, I get a kick out of correcting people\'s mistakes! I am hoping Dr. Chance will like my work and this will open up more opportunities with him and with the U of Penn in general.  Thanks to Craigslist I saw another post this week about a new neighborhood newspaper coming soon to various areas of Philly. They were advertising for writers, so I replied.&nbsp; Yesterday I got the email I have been wishing for ever since arriving in the States. The first line read, \"Daniella, I really like your style of writing, I want you to write for us!\"&nbsp; Woohoo!&nbsp; The first edition comes out in September and I\'ll be covering my Art Museum neighborhood - my first article will be restaurant reviews - rather ironic given my physique!&nbsp; This is the answer to my prayers, the opportunity to write freelance for an American publication - a means to get experience and get my name out there - the money, if there is any, is secondary!   So, so far, 35 is turning out to be great, full of new and exciting opportunities!  Onwards and upwards!   SM   '),(14554,3350341,'16,July,2004','2004-07-16 00:00:00','    I can\'t believe it. Tomorrow I will be 35 years old!&nbsp; It\'s not that I feel or look old (some people still mistake me for a teenager - why is it that people always assume that petite people are young!?), but I do feel a bit down at the thought of approaching my 40\'s. Maybe I would feel less down about it if I felt I\'d accomplished a lot in my 35 years. I know I know, I\'m being hard on myself...there aren\'t a lot of people out there who have survived SMA, scoliosis, and two immigrations. I should be pretty proud of myself. It\'s just that in mainstream society accomplishment is measured at my age, in being married with a few snotty nosed kids in tow, a house with a picket fence, a golden retriever and a successful career.&nbsp; I\'m not too depressed at not having the huge responsibility and 24/7 burden of caring for kids, but I would like to be settled down, own a nice comfy home (with or without the picket fence) and the dog. I want the dog more than anything, lol. &nbsp; AT least I have the fabulous Mouse Spouse in my life, being alone at 35 would really suck. I have great family and my health. I have a decent job, great friends (albeit not in Philly) and a good quality of life. What more can a gal ask for at age 35?&nbsp;  &nbsp; Happy Birthday to Me - I look forward to a great year ahead. Wow, blogging is great, I managed to cheer&nbsp; myself up!&nbsp; This is better than therapy! :-) &nbsp; SM   '),(14555,3350341,'11,July,2004','2004-07-11 00:00:00','    Last night me and the Mouse Spouse went to the Keswick Theatre to see  urlLink Johnnie Clegg and Savuka  - probably the most well known South African band. It was so unbelievable to sit there and hear all my favorite songs, to hear South African accents in the audience and to get up and dance and scream!  He played many old favorites and some songs off his new album - it was so cool!  For those short two hours, I felt connected to my South African roots and it felt so damn good!  I will definitely see him again when he comes back next year! Rock on!  A few days ago, the International Court of Justice ruled that Israel\'s security fence is illegal and should be dismantled. The ruling was no surprise, to me anyway, the UN minions are consistently anti-Israel. I have tried to define my own feelings about the security fence, but I find myself \"sitting on the fence\" in this issue. I can totally appreciate the wisdom of Israel\'s trying to stop terror attacks in Israel - and largely, until today (a woman was murdered in a bomb attack in Tel Aviv), the fence has succeeded in significantly halting terror. What bothers me about the Fence is how it impacts on the lives of ordinary Palestinians - cutting them off from their land, schools and making their lives very difficult.  Should I care that \"my enemies\" lives are inconvenienced and made intolerable? Should I give a fuck when the bottom line is the Fence seems to be saving Israeli lives? This is what I struggle with.  I always try to distinguish between terrorists and ordinary Palestinians -- surely there must be Palestinians who do not support terror, who just want to live ordinary lives with family, jobs and food on the table? Sometimes it\'s difficult to believe their are \"good\" Palestinians, but to believe otherwise would be counter to my humanistic beliefs. I always think of Anne Frank\'s famous quote, that despite everything, she still believed that people were basically good.   Finally, congratulations must go to one  urlLink Phillip Rabinowitz  of Cape Town, South Africa, a 100 year old who set a new official record yesterday in a 100-meter sprint.  Way to go old geezer!  SM   '),(14556,3350341,'04,July,2004','2004-07-04 00:00:00','    In my last post I expressed how content I am living in America. After just having returned home from the July 4th Parade on the Parkway (literally a few steps from my door), I feel really euphoric to be in the USA. What I so love about living here is the diversity, shown so beautifully today in the Parade.   First of all, I had trouble deciding whether to look at the parade participants or just the hundreds of people milling about the Parkway. Both were rivetting. During the parade, representatives from every imagineable community strutted their stuff in their unique national or festive get up. There were the Irish, Polish, Philipino, Sikh, Indian, Pakistani, Chinese, Falun Gong, Bolivian, Italian communities to name a few. They all danced or pranced or marched to the general joy of all gathered. The Parkway was overrun with families and kids -- especially African-American kiddies with their hair in elaborate braids decorated with red, white and blue barette (American for clip fellow South Africans) - you can just eat them. Also in evidence were many many obese Americans, young and old, dressed in skimpy shorts and tank tops, sans a care in the world. I could sit out there all day and just people watch.   America fascinates me!  And I love to be fascinated :-)   '),(14557,3350341,'02,July,2004','2004-07-02 00:00:00','    I just realized that three years ago today I arrived in America.        When I landed at Miami Internation Airport I had no idea that I would be permanently relocating from Israel to the USA. Initially I came just to check it out, to visit my ex-boyfriend who I was still close to, intending to stay two months. But I fell in love with south Florida - the tranquility, the beautiful beaches, the sparkling ocean and the complete lack of fear, anxiety and terror that was a part of my everyday existence in Israel. I also fell completely out of love with my ex. Despite the latter, I decided to return to Israel, sell or pack up my belongings, and return to the USA, full of hope that I would find a job and start a new life.  Things didn\'t exactly work out that way. 9/11 happened. Could not find a job in south Florida - luckily I still had work doing web stuff for an Israeli company, so it kept me going. I lived a vacation-like existence in Florida for a year. It was a wonderful time, but also a worrisome and lonely time too. I missed my friends in Israel and worried about my very tenuous legal status in the States. With two weeks left on my visitor\'s visa, I made a trip to Philly to visit family, and totally lucked out, finding a job at a Jewish non-profit. And so once again, I packed up and moved North to a city I was totally unfamiliar with, where I had no friends and little family. But two years on, I can look back and feel really proud of the life I have made for myself here. I really like living in America - I feel happy, secure and safe, I\'ve met a wonderful guy and his equally amazing family, I like my job and enjoy everything this city has to offer!  Happy July 4th!  SM  PS If you don\'t hear from me again it means I\'ve won the  urlLink megamillions  lottery - $290 million - and gone on an extended vacation! :-)   '),(14558,4053312,'31,July,2004','2004-07-31 00:00:00','   I want to thank everyone that has posted comments.  Thank you for your words of wisdom.  I have had a lot of questions about how long the surgery will last?  Who is going with me?  Who is watching Brandon and DeeDee?  What hospital?  How long I will be in the hospital?  Here are some answers:  I have been told by someone that has had the surgery that it takes about four hours.  In all of my excitement, I forgot to ask my doctor.  Duh!!!  I have to be at the hospital at 6:00 on the morning of Tuesday, August 3, 2004.  I am leaving from home on Monday night and stay in a hotel in Ocala to be close.  Anyone that knows me, knows that I am not a morning person and this is the best option.  ;-)   I will not be alone.  My very dear, supportive, and loving friend Carlos will be with me from the time I leave home until the time I return home.  He has been going to Ocala for doctor\'s visits up to this point.  He has been in this with me from almost the very beginning giving me informaiton to help me decide what procedure I wanted to pursue.  I feel that I am in capable hands.  With GOD\'s help he will take care of me.  So, don\'t worry.  My mother will be at home with Brandon and DeeDee.  I have not told them yet.  I am still not sure if I will, but knowing my very nurtring mother, she will think of something if I don\'t.  I just want them to know how very much I love them, and that they are my motiviation for doing this.  Here is the hospital informaiton: Ocala Regional Medical Center 1431 SW 1st Avenue P.O.Box 2200 OcalaFL  34478 Telephone: (352) 401-1000  I am suppose to be in the hospital for three days.  I am hoping to be back home by Friday or Saturday.  If you want to know how I am doing, you can call the hospital and talk to Carlos or me if I am up to it or call my mother at home.  I am excited.  I made contract with someone that had the surgery two years ago.  She was very helpful in answering my quesitons.  I feel a little better.  She is going to help me as much as she can.  I ask that you all continue to pray for me and my family.  I love you all.  Tina  '),(14559,4053312,'29,July,2004','2004-07-29 00:00:00','   urlLink    WOW!!!! 5 days til surgery!!!!! The days seem to just be flying by!!!! I am having a lot of mixed emotions right now. I am scared, excited, and anxious all in one. I can\'t wait for Tuesday to get here, but at the same time I wish that it will never come. For those of you that have not seen me lately. I have cut off all of my hair, and am now growing locs. See, my transformation started a little over a month ago.&nbsp; urlLink  '),(14560,4053312,'27,July,2004','2004-07-27 00:00:00','   urlLink    Tina, Brandon, & DeeDee on vacation in NC - July 2003&nbsp; urlLink  '),(14561,4053312,'27,July,2004','2004-07-27 00:00:00','   I am 33 years old, I am 5â€™8â€ tall, and I weigh a whopping 300 pounds.&nbsp; For my friends I know that this may be a shock.&nbsp; It is a shock to me.&nbsp; I have been overweight for about 10 years, but never as heavy as I am now.&nbsp; I have always been what people call \"big boned\" (What ever that means).&nbsp; I canâ€™t even blame my weight gain on child birth because my weight started to rise out of control when I stopped playing basketball and I graduated from college.&nbsp; About a year and a half ago my uncle (Ronnie) whom I love very much passed away about a year and a half ago.&nbsp; I decided that I had to do something drastic to make me lose weight.&nbsp; The hurt and pain that I went through at that time helped me to see how precious our time on this earth is, and I was wasting mine.&nbsp; My family has a history of diabetes and heart problems on both sides.&nbsp; On my motherâ€™s side of the family I have had one uncle pass away with diabetes and cancer; two of my surviving four uncles have diabetes, as well as both of my grandparents.&nbsp; On my fatherâ€™s side of the family I have had one aunt pass away from cancer, my one surviving aunt has diabetes, and so does my grandmother.&nbsp; I want to do what I can to make sure that I do not have these problems.&nbsp; I have a 6 year old and an 8 year old, and I want to be here to see them become adults and have their own families.&nbsp; Heck, I want to be here to see my grandkids start their families.&nbsp; My surgery date is scheduled for August 2, 2004 (exactly 7 days from today).&nbsp; I have just become very scared.&nbsp; Not because I am having second thoughts about the surgery, but because of the unknown.&nbsp; I know that  GOD  will bring me and my family through this.&nbsp; I will just continue to pray and ask the  LORD  to watch over me and my family.&nbsp; I know that he has a master plan for us all.'),(14562,4141669,'04,August,2004','2004-08-04 00:00:00','   Are You Called to Short-term Missions?         My name is Bobby. Welcome to my small piece of cyberspace. I work for a church in a north Dallas (TX) suburb and help facilitate trips to fascinating places all over the world. This is a recent change of assignment for me as my work for some time has been in the arena of local outreach and evangelism - working with the homeless, the incarcerated and people in very needful situations. I have, however, always had a passion for people in other countries and longed to travel to bring the Word of God to them.    Last year I had the privilege to speak on the streets of Diriamba, Nicaragua and to a men\'s prayer gathering in Puerto Vallerta, Mexico. The desire I held in my heart was brought to fulfillment a couple of months ago when I was transferred to our Short-term Missions Department. I am preparing for a trip later this month - again to Nicaragua, but this time we will also spend a few days in beautiful Costa Rica. In November I will be going to the Cairo, Egypt and then to Mexico City in December.  My prayer in keeping this blogspot is to be a source of encouragement, empowerment and blessing to others who may feel the desire of the heart to venture out into the rest of the world. God places in each one of us gifts and abilities. And for many of us, He also grants a special desire and willingness to use those gifts and abilities in a foreign country.  If you find yourself drawn to another land - perhaps you just keep thinking of another country, perhaps your ears perk up when you hear it mentioned on the news, or your heart breaks when you hear about oppression, flooding, violence or starvation there - then seek the Lord in prayer to determine if He is drawing indeed drawing you somewhere, out there. If so, pray for an open door to explore the wonder of His will there.  It is amazing and He is due so much praise for it, as He will provide a way to get you there, if you will open up to His will.  Short-term missions are a way to test the waters.  Are you called to be a missionary somewhere?  Or, to use your gifts and abilities on a short-term assignment.  You also will test your own limits. Can you do without the creature comforts of home? Can you adjust? Do you hold up well in an environment of change?  Can you pack light?  Can you eat THAT?  The second amazing thing is that He is, indeed, already there. This is evidenced in the help you will find along the way, the people He will put in your path, and the favor you will be shown as you travel.  The third amazing thing is that language is not that big of an obstacle when you are on a well-planned short-term mission trip.  Yes, you will not be able to carry on a conversation like you might like but that is actually a good thing.   For example, on my most recent trip to Nicaragua we were helping a local church put on Vacation Bible School for a couple hundred youngsters and then helped repaint a local Christian school.  With language out of the way, the only thing you can communicate is the love of Christ.  It shows in your smile, in your work, in the way you carry yourself.  It speaks louder than words.  The first day we were painting it was our team and a couple of teachers from the school.  By the end of the evening some other parents had joined in.  By the middle of the next day an entire army of parents, students and neighbors had joined us, many of them bringing their own tattered paint brushes from home.  We had started the job.  We had purchased the paint and supplies.  We were the catalyst that ignited a team effort to complete a very large task.  At the end of the second day, the work was completed and all of us, young and old, American and Nicaraguan, English-speaking and Spanish-speaking, toured through the building and celebrated as the school name was reapplied, carefully lettered by a local artisan.  Even if you only go once on such a trip I encourage you to go.  '),(14563,4236769,'16,August,2004','2004-08-16 00:00:00','   Hello world!   Yes, well... umm... I\'ve just started this journal, as you could probably tell, and I\'m aiming to use it to record life as it is now for me. As for most journals, I haven\'t put this online to \'show it off\' per say- I\'m using Blogger because it\'s very accessible and so I can come back to memories and thoughts that I might well have forgotten in years to come. However, comments and everything are very welcome!   I hope that I\'ll be welcomed here well enough- I\'m currently working on a layout of some kind that might come sooner or later... We\'ll have to wait and see!   Well, I suppose that first journal entries include a bit about the journaler, so I\'ll introduce myself, shall I? My name is Jennifer, but on the internet I go by the alias of Myth, I\'m currently 16 and studying in 11th grade at a school in Brisbane, Australia. I\'m more of a nerd than not, I guess you can say, but that\'s mainly because I need to study really hard to get where I want in life (in terms of a career) as I would like to become a physiotherapist, but I\'m quite good at a variety of different things- I can write stories and poetry, draw to some extent and even defend myself, as I learn a martial art called  Ju Jitsu . My parents are separated, but I see them both equally, so it isn\'t that bad a deal, and I have two brothers- one twin brother and one half brother, who\'s currently six. He (the latter) is really cute, but he\'s also terribly annoying and (in m opinion) a little spoilt.   I like to RP (RolePlay), mainly the online kind, so you can often find me in the RP sections of forums, or MUSHing on sites like  Elendor , but most of my time is lately spent in study or at Ju Jitsu training, which is really very fun, despite all the bruises.   Well, that\'s about all I can think of at the moment, so I\'ll be off and I\'ll post sometime later! I plan to try a very different style of journal writing, so I\'ll see what I can come up with! '),(14564,3422339,'31,May,2004','2004-05-31 00:00:00','   heylo.. Today was...well... nothing really THAT special happened. Had to do a presentation is French. Me=hating talking infront of people i dont usually associate with. no, im not a snob. just a shy person at times. I had to do a test in science, wow. think i failed that one. Lunch was kinda..yeah...nothing special...i cant remeber..cept Cody and Erin kept farting...jk jk...and then cody and a bunch of guys read my girly magazine..? no comment. then tim came to where we were in a muscle shirt. and shorts. lana thought he looked hot...sorta.:) In gym we had to play cricket. it was sorta like baseball. but not at all...it was wierd.  Then English was frikin boring as usual. Tim, yet again, took his shoes off and kept touching me with his sock feet. i didnt like it too much. sorry man. After school, went to work...it was boring, nobody came in. excpet some old lady i didnt know who gave me a cookie. :) wow..then the guy i work with bought me a pizza bun and cake. hes trying to make me fat-er. lol  Then i went to dads. for like 30 mins. went outside...talked to some people...[heart] ..now im home at my moms:(...i hate this. -later nesh  Its been awhile and i havent forgotten the way you made me feel like  im on top of the world and i wish i could feel it agian i wish i could love you agian, and do anything to make you happy. to make you love me. I love you. still. forever and always.  '),(14565,3422339,'30,May,2004','2004-05-30 00:00:00','   Heylo people. Havent blogged in days. well, yeah, i was at my dads this weekend and his computer has a virus. so that would be why. So Friday night Erblina came over to my dads. we went out for a walk...ummm...i cant remeber...lol...went to Tim\'s house. called Em. went to her house, walked to Macs. Emily made me cry. seriously. i was sad. then we went to her house. ate. chilled and played nintendo. then me erblina and tim ran to his house all linked arms.. um went to my house. erblina got picked up and tim walked home. i went inside and watched Deep Blue Sea on TV till like 1 then went to bed. Saturday...came here to my moms to get my baseball unifrom then went to get my baseball pics taken. My dad was convinced it was at 1. i said 12 45. got there at 12 45, dad and phill went for a walk. ended up being taken at 12 45. took exactly 1 minute. was tons a fun! then i went back and stodd beside my dads car for 27 mins. doing nothing. wow fun. Went home, helped my dad with yardwork...was very boring. but i got paid 10$...played catch with amy..that was awesome.:)...then saturday night was the party at kates! wow it was sooo fun! i love kates house. really. and her butler Gordon is awesome.  Ppl showed up around 7ish. when i say 7ish i mean like 7 15. got a tour of the house. evan and mike were determined they could walk across a narrow ledge high up above the stairs. they=really couldnt. then in the basement they had sprints back and forth, cuz her basement is huge obviously. we hung out for ahwile till Ray came. then we went outside for a walk down this long long long long path. all the guys ran down it. except mark cuz his pants kept falling down. we went to this treefort thing, but it really wasnt, it was kool tho. Ray got stuck...it was really funny. kept walking and came to this rope swing. then there was this donavan guy. i yelled at him. and he made random bird noises.then he came and stood with us and laughed with us. yet we didnt know him. kay, then we went to the park...and ray got stuck, for a change, on top of the park thing. with the little table underneath. the guys tryed to push him off. it was funny. ray=gets stuck. ALOT!then back to kates and watched GOTHICA...IT was soooo scary!!! omg!! we were just sitting there. and i didnt even noticed some of the guys had left. and just at a scary part, they started banging on the widnows right behind where we were sitting!!! it was sooo scary! lol, but good times.  the movie was so scary tho. i barely watched it! seirously. i kept screaming i was so scared. im so pathetic. after the movie we just kinda hung out. and ate...was lots sof fun. awesome party kate:)...im updateing this now. i just read kates blog. and i must say, i laughed by butt off for a whole minute! So today, sunday, i did like nothing. woke up late, did homework, worked out...did nothing. watched the movie spiderman for english..and yeah. thats about it. so i better go finish my hoemwork now. :@ and im outti -later Random quotes: \"Hey i think i know you, kid\"-mark \"oh...\"-kid \"Whats your name?\"-mark \"Donavan\"-donavan(kid) \"nope, dont know you\"-Mark haha then we were sittin on the couch: \"Hey Ray you have to see the huge basement!\"-Evan *Evan runs into study* \"Wrong room!\"us ...then *Evan runs around with umbrella in one hand and umbrella basket on head making awesomely funny noises* *All guys poke out thier heads from behind walls and make weird noises* \"Hey guys. oh my god. guys oh my god. theres a little girl sitting in the backyard!\"-evan \"....AHHHHHHHHHHHHHHHH!!!!!!!........ur kidding right? yes, your kidding\"every1 else. \"no GUYS! im NOT kidding! there really was a girl there! right mike..? ..mike?\"=evan this one is right from kate: Kate PROPS TIMES 758927985678934....i luv yaaaaaa! *Donovan makes loon sound.*  \"sorry, i don\'t speak loon.\"-evan *All guys plus erin sing \"Oh what a wonderful world!\" in manly voices* ...wow good times man, good times right there. -nessHA!  '),(14566,3422339,'28,May,2004','2004-05-28 00:00:00','   Wow! what a busy day! aww it was soo funny. kay, so this morning got up at frikin 6 a.m. damn. so got ready, the usual picked up erin, and went to GC at 7 30, for the softball tounrament! took the bus there, blah blah, umm played our first game. we=lost 11-6. i had to play CF the whole game. frik. The second game i picthed 3 innings. and we tied. 5-5. so we played another inning. tied again, 7-7. played agin. lost by 2. frik agian. shit shit. oh well. mom went and got me a sub as me and erin watched another game. then mom picked us up and we went home. Funny quotes (inside jokes avec erin): \"[crouching guy] is going to cut the grass]*crouches*\" lol and: \"The Flying Dog Restuarant!\"-radio \"hahahahahahahahhahahahahahahahhahaha times 7589764832\"- me and erin \"wtf? flying dog? BONE!\"-me \"ER!\"-erin \"...?\"-me \"get it? u said \"bone\" i said \"er\"\"-erin \"hahahahahahhahahahahahahahahahaha\"us wow. So next, went home, at 12ish. did nothing till 12 20. then called erin and told her to come to GC with me for the assembly. she agreed..cuz she wanted to see Cody. me=got showered dressed and ready in 30 mins. that is a record! picked up erin at 1....went to GC..assemeby..haha aww it was funny. that guy. wow. hes funny. and the old people sining about cheese wiz....i loved it. good times im tellin ya. so THEN I went and got a popscile and went to Gateway! yay!..haha NO! i waited outsdie for a while. Saw [hottie]:):):). talked to amy....i was waiting for my bro to come home from camp:(..oh well. he finally got home. 30 mins later. he got off the bus and i ran and gave him a hug. he ran away. lol jk. i swear he went thru pubery in 4 days! while he was at camp. no joke. his voice sounds crackly! like a man!...he says he just lost his voice tho. lol. B.S.! then we were walkin towards the van. and his pants were fallin down alomost too his knees. wow. i have a ghetto brother. and hes a man. wow.  ok so thats my stroy! im prob missing sytuff. i laughed so much today. it wasgreat! i almost cryed tho, my cousin natalie came to see me after school. shes like, 8 months. i held her. she=cryed. me=cryed. :(..but then julia (other cousin, 3 yrs old) came and held my hand and we went for a walk:)...meh, one of them loves me so i g2g pack for my dads. he\'ll be here in 10 mins! FRIK!  -later  '),(14567,3422339,'27,May,2004','2004-05-27 00:00:00','   Hellooooo! I have so much eneergy right now. sad, because ive been runing around since gym, then GC softball practise, then a softball game tonight. It was awesome! Erin caught amazing, let me just say! lol i picthed, 1st inning, 3 up 3 down:) then 3rd inning, 1 walk, then 3 up 3 down..lol..and 5th, yet agian:) wow. that was  a shocker, out team played awesome! but lost, for a change{not} 11-6. i saw one of my caoches from teeball. i went over to talk to my mom, and  shes like \"oh, do u remeber Hgheainaughw?{not. just cant remeber her name}\" \"...\"me \"oh well, she was your first teeball coach\"-mom \"oh Yeah!\"- me. not kknowing whats going on \"yeah, remeber we had the pool party?\"-lady coach dude \"OMG YES! and mom got drunk and had to get carried home and alll she kept saying was \'oh i left my show...uhh...the POOL\"-me. finally realizing whats going on. \"hahahahah\"mom and lady \"well i reconoized you right away\" lady \"oh yeah?\"- me=assulted. thanks, cuz i look about 5 years old, the age i was when u coached me. right. thanks. \"yeah, but your so pretty! and i see you filled out nicely\" lady \"...\"-me (thinking= first she insulted me. nows she comenting on my chest.wow...) \"hahahahah~!!!!! HAHA!\"-mom \" tap..tap swish\"-me, walking away silently towards team bench to bat wow, what a great stroy! oh but it gets better The coach took 5 of us girls out to get ice cream, rom The Cone Shoppe where my sister works. it was great. thanks! lol..ohh so us girls were standing outsdie with our cones when megan has a great idea. \"Hey GUYS! lets sit on the ground like Homos!\" \"HAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHA\" -US ALL LAUGHING! BECAUSE MEGAN SAID \"HOMO\" INSTED OF \"HOBO\"...haha! \"Now your being rasist!\" \"haha,..wait a minute!\"..haha aww i guessyou had to be there, we all laughed. for 10 minutes. no joke. it was amazing! lol, but then! *Megan licks her ice cream cone, and the ice cream fall sonto the ground\" \"HAHAHAHAHAHAHAHAHAHA!\" Brianne \"HAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHA\"- us girls \"hah aww megan!\"- me \"wait! i need a spoon\"-me \"YES!\" stacey *me and stacey run into cone shoppe and get a spoon** *Stacey then gets down and eats the ice cream off the ground! then passes the spoon to megan, and then to me! we all eat ice cream, the part that didnt touch tthe ground, but its still sittin on the ground. people around us=thought we were crazy. so i yellwed at them haha.* WOW! you really really had to be there. wooo...good times, right there. well its late, plus i have a tounrament tomorrow. me=going to sleep now. and good night xox nessh  '),(14568,3422339,'26,May,2004','2004-05-26 00:00:00','   ugh. hello. just got in from mowing the lawn, wow. that was a first. i now have wavey lines on my lawn. it looks pretty cool though. i like it. and i didnt hurt myself. Bonus times 2. Had practise tonight. was good was good. Wasnt too intence. pitchin a bit. went pretty well i think. coach said i amazed her. cuz of my blindness. yay me...frik. i hate it. So anyways, today at school we had cake. it was great. then we got to go thru a blowup maze thing! it was intence! it was so frikin fun tho! I went twice. cuz kate was too nervous to go, when she was in line. she told me to go. sweat. i beat Lana. twice...she tryed to push me off the thing tho! yay. we had fun! then TIM FELL OUT! it was hilarious. i laughed... NOW i have to go study for science=test tomorrow. frik again. but tomororws my last day of schoool this week. Friday=softball tourny! sweat!!!! and im done. -nessa  '),(14569,3422339,'25,May,2004','2004-05-25 00:00:00','   heloo. Just got home from baseball practise. was great, pretty boring tho. and wet. but it was ok. My house is so quiet! My bro (phill) left for Camp Wahanowin this morning:)=4 days of peace! I went to Gateway to say bye to him this morning. i managed to get a hug outta the kid. as he kicked dirt on my white pants. love you too bro. frik. Oh well, had like the worst day today. i swear. everybody is really out to get me. thanks. my arms sorta (extremely) sore from pitching yesterday, and my shoulders are killing me. like a bitch. oh well. i guess it can be worth it. even though it isnt. im no longer complaining.. Goin to dads now. havent seen him in 675432 days. he went up north. to visit my family. frik, cant wait to go thur this summer...well now im gettin boring and blabbing on. so im out. -later props2kate.frink. there goes your paralyzing eyes, there goes your tantalyzing smile, there goes my heart. falling for you.(8)...country music. yay!..no.  '),(14570,3422339,'24,May,2004','2004-05-24 00:00:00','   cuddle and a kiss on the forehead - you like to be close to your special someone and feel warm, comfortable, and needed    urlLink   What Sign of Affection Are You?     brought to you by  urlLink Quizilla  '),(14571,3422339,'23,May,2004','2004-05-23 00:00:00','   i just got home from softball clinic. learned how to pitch a beauty low strike. its sexy. I pitched straight for 30 mins. thats like, 2 every 10 seconds. like 56762878 times. my arm=sorta sore. oh well, it helped. now im pro. even more pro than before. HA! anyways, i better go do some home work. me=not done any all weekend. yay! later.  '),(14572,3422339,'23,May,2004','2004-05-23 00:00:00','   gr. Good morning. even though its 1:20...i just woke up. go away. I hate the people who call your house and want to piss u off. i.e. this morning: \"ring ring ring ring ring..times 768592\"-telephone. \"njhndmmmhmmm hello?\"-me. being woke up:( \"Hello!(birng cheery voice) is this the deboer residence?\"-lady. on coffee high. \"...\"-me. i hate how my mom has a different last name than me. its too much to handle in the morning! im not a deboer. im a DiPasquale. thank you. \"yeah...\"me \"mwhp whwp hfugs ghsud gbuih gnufdsgh nbuifdh\"-Lady, blabing about some thing i wasnt interested in\" \"sounds great!\"me. faking it. \"okay, we\'ll give u  call back later in the week to confirm\"-lady. \"ok..\" \"Have a GREAT day\"-lady \"...click\"...me...JUST KIDDING! im not that mean of a person..\"ok, you too, bye\"-me \"click\" ..wow i bet your glad i told you that one. it was the highlight of my mornning.. okay. now im in a better mood:)..well..actaully..nah. im great really, im gonna go now. my mom\'s bf is yelling at me. great. now he hates me too. damn it. i hate life. -later.  '),(14573,3422339,'23,May,2004','2004-05-23 00:00:00','   \"What kind of kiss are you?\" - Results:    You have an entrancing kiss~ the kind that leaves your partner bedazzled and maybe even feeling he/she is dreaming. Quite effective; the kiss that never lessens and always blows your partner away like the first time.      You have an entrancing kiss~ the kind that leaves your partner bedazzled and maybe even feeling he/she is dreaming.  Quite effective; the kiss that never lessens and always blows your partner away like the first time.    urlLink   What kind of kiss are you?     brought to you by  urlLink Quizilla  '),(14574,3422339,'22,May,2004','2004-05-22 00:00:00','   haha. tonight was great. kate came over. then we went to silvias house...then we walked to tim hortons.. it was soo funny. Kate=3 donuts Me=1 donut Silvia=1 danish..which she later ened up throwing on the ground cuz the bag got wet? It was funny..we were talking there as it was raining,  \"I hate you, but then this one time\".... good times.. Then we walked around in the rain! like and got soaked! we went to call on ppl, but nobodi was home. thanks guys. joking then we went to my hosue agian.sat in my room, i got changed like 5647 times...or twice..cuz my clothes were soaked. then we walked to Macs. we each got exaclty one candy...walked home in the pourin rain again. watched some kids fireworks and hollered at them. funfun. came back to my house. ate popcorn, talked...girlie stuff. anyways. me=tired times 758 night* night* x n o -nessha-  '),(14575,3422339,'22,May,2004','2004-05-22 00:00:00','   hello there Just got home from the mall. with my mom! it was aweosme got: Skirt (turquiose its so \"different\", and thats why i like it) Tank top (yellow, tie up) Tank top (White pink and turquoise...got luv that colour!) Shorts (white with yeollow flower) and socks...thuir pretty hott too.. so yeah, trying deperatly to make plans for tonight...with kate and erblina...who isnt home right now. frik. (sorry, kayte, ur word...props.) so yeah, im going now.  '),(14576,3422339,'22,May,2004','2004-05-22 00:00:00','   helllo.. Just woke up. Eating breakfast. Talking on msn to Erblina Josh Amy and Erin. Yet another boring day. Hopefully we\'ll get a thunderstrom. i love thunderstorms. who doesnt?!? i mean reallly. -later  '),(14577,3422339,'22,May,2004','2004-05-22 00:00:00','   Good evenin! Just got home from tim hortons. Tim Em and Kate came over tonight..we watched part of Jarasic Park=only thing on TV, Em played with my webcam.lol. then we walked to tim hortons (me and kate) saw her bro there. It was funny. Me=1 donut Kate=2 donuts Then Kates bro drove me home. James\'s friend=sexy times 7283 it was funny, we were drving down my street: \"This is the right street, right?\"-James \"No James\"-Kate *sarcastic voice* \"Kate im just saying that because she has two houses i didnt know if it was this street\"-James \"Ok fine\"-Kate *sarcastic agian* ----Car screeches to a stop---- \"I  dont need that atitude Kate\"-James \"..ha...ha\"-me \"Ok..hahaha...ok\"-Kate \"Get out and walk\"-James \"NO ok im sorry\"-Kate \"ok good\"-James -Car goes again, about 1 foot in distance- \"thanks, idiot\"-Kate ---Car stops agian!-- \"Kate! Whatd i tell you!\"-James \"....HAHAHAHA...(hmahaham)\"-me laughin histerically, hhen tryin to hide my laugh and try not to embarass myself infront of hot friend LOL...guess you had to be there. wow im done now. Good night.sleep tight. -Nessha xox P.s. \"Tim you didnt go to the dance yesterday eh?\"-Em \"hey i didnt go to the dance yesterday...\"-Tim (A.k.a. mr. obvious)  '),(14578,3422339,'22,May,2004','2004-05-22 00:00:00','   Today was bo-rag. I went shopping with my mom. She said i could get a shirt sometime this weekend... Then we had to go lokk at gardening/plans/plant stuff/patio stuff..i swear, if i hear any of those words one more time, somebody will get hurt. So tonight im hangin out with Kayte and Em...we were going to go to timmy hortans...but its startin to rain... later.  '),(14579,3422339,'22,May,2004','2004-05-22 00:00:00','   Yes, i got plans. Im making a pond in the backyard. Hey its better than cleaning my room. Thanks to Erin and Kate for ditching me, just kidding. they went to see Mean Girls. I was invited=couldnt go. Cuz im not as rich as them. especially Kate. Im joking. i really am rich. You just cant smell it... Just read that over. i am realy a loser. If anybody reads this, who doesnt have plans, give me a shout. im desperate.(jokes...) -later  '),(14580,3422339,'22,May,2004','2004-05-22 00:00:00','   Heyy  Kate, thanks for the help. i made a blog. cant you tell? Anyways, hello everybody. This is my blog. sexii eh? So my life... Last night i hung out with Erblina. went to the mall with Tim...then went to his house, and to StFrancis...on bikes...no comment except my ass hurt. Erblinas soo cool...we had fun...i love her...  Oh and Thursday. (Thursday during gym i got major chest pain. it hurt like horribly.  Then i missed baseball cuz of it...to go to the Dance, oh well i had like 74389 practises this week anyways)So Thursday was the dance. wow it was the bestest. All cuz Erblina and Lana. You girls are so fun! ERBLINA I LOVE YA!!! And Erin. wow for those of you who werent there. you missed out big time. lets just say, she had too many.( i only had a sip!!hey its not my fault i didnt drink it! it was strong times 16537. im not the one who poured it out either..lol..cough*erin) \"Your my best friend\"-Erin...said to people she didnt know. or people who ran away crying. joking, erin, u made me laugh so hard. The other best part was the free gel..they were yummy, and the Crispers. ew?  Oh yeah, and Wednesday.  Had practise. Did this thing where u held ur glove above you, and \"bounced\" a ball to all teamates. We did it, after numerous tries. Won pizza and ice cream. My coaches rock.  Today is boring, i like just woke up, fucking 12 00. oh well. i needed sleep. Last night i was on the phone till like 2 anyways, it was great. Chrsitians friend has the sexyest voice ever...so no plans today. i was gonna go to the movies, to see Mean Girls, finally, yet im broke, no dice there i guess  So going to get food now Later xoxNessa   For the record, The Reason, By Hoobastank is the bestest song ever. i dont care! i luv it. and so is Ocean Avenue...wow..like seriously. and im done.  '),(14581,3422339,'29,June,2004','2004-06-29 00:00:00','   Whew Helllllo! Today was one of the bestest days: Woke up at 8:45. Slept till 9. Showered and got ready. Recived a very odd phone call (prank?) around 9-20-ish... \"Hello?\" \"Hello. is this Vanessa?\" \"yes\" \"oh hello, how are you?\" \"good\" \"thats good. i am wishing i was there with you so i could feel your bum\" \"..\" \"i bet its nice and....\" *click* Haha i hung up. it was odd...anyways Got picked up at 9 30 and went Strawberry picking with my Gramma. Was fun. Sorta dirty, but got lots of berries [heart].  Came home, got ready, msn, went to Erin\'s. We Played basketball...21 actually. I won! yay! 2 outta 3. lol, it was nice. cept the net was wayyyy to high. then we saw ian so we went to talk to him for about 10 mins. it was nice. then he left. hugs* Went to Matt g\'s house...saw him in the pool. erin and i are going swimming there tomorrow. yay.! lol Loud music eh erin? ahaha. Went back to erin\'s..msn..i curled her hair ..lol im pro..minus the part about me being pro. it did look pretty cute tho. cuz erin is sooooo mega gorgeous! Then she did my makeup. It was hott stuff.  Soooo then i went home on my bicycle...lol i think im on that thing so much. Anyways, i went home quick cuz my mom said she had a surprise for me. so me=excited. Got home. she said i had to wait till Phill was home. he finally dragged his late ass home. Our \"Conversation\" \"So whats the big surprise?\"-me \"i just wanted to talk about our vacation\"-mom \"oh....*boring*\"me \"i was thinking about where to go\"-mom *mom leaves to get the brochure of where we are going* *i talk to phill* \"so what do ya think? the usual? Collingwood...Port Dover?\"-me...PS..those are all in ontario. \"i hope Blue mouuntain so i can go biking...\"-phill \"i hope Sauble so i can go swimming!\"-me *mom comes back downstairs* \"look\"-mom *i look at pics of facy resort and beautiful sandy beaches...clear water* \"...what!?\"-me *points to top of page*..mom \"...CUBA!!!! omg OMG OMG!!\" times 65787574 for the next 5748952 minutes so going to cuba in 51 days...woooooo! Anyways. went to the t-agent. booked it. came home...here i am. anyways, better roll out yo. chow. PS I luv amy! shes the bestest!!! X n o yo. Nessssssaaaaaa Quotes!  \"Have you seen smokey eyes?\"-erin \"...no?\"-me \"you havent seen them?\"-erin \"no, who is it?\"-me \"NO!! hahahhaha No! like you havent seen them in magazines?\"-erin \"..no? are they famous?\"me haha its funny u had to be there... pps...smokey eyes is a way of doing makeup.  then. \"Im going to Cody\'s tonight for some spooning\"-erin \"...whats that?\"-me \"you know, spooing is  like cuddling\"-erin \"ohh okay, i get it...so whats the difference between spooining and forking..??\"me hahha \"......how aboout knive-ing!?\"-me haha. and im done  I dont understand changes of heart. i dont know how your love can change so suddenly, and you dont understand hows it breaking me up inside. you didnt hear the truth.  love always.. ...heart broken  '),(14582,3422339,'28,June,2004','2004-06-28 00:00:00','   heyy there. Today was a fairly boring day... Woke up around 9-ish. watched tv...ate..and unpacked my duffle bag, and that was the highlight of my dad. yep, thats about it folks! I left for work around 3 45-ish. It was sooooo boring today. the best part? i made coffee and sat reading old people magazines.NOT fun. dad picked me up...went to his house, had a TV dinner...went to my moms house... Erblina came over at 8 30-ish. We biked down candlewood because she said she saw some HOTT guy. But no. Then we biked to Tims, and he came to the park with us...we just kinda hung out, went back to his house. His Bro, Matty was there. with Robert Chelsea Josh Shauna Ian Emily...etc. We talked to them for a few minutes, then we rode our bikes to my house. went to ben\'s to get tim\'s bikes. went back to tim\'s house. walked around looking for those people...went to the park. tim pulled around a wagon. saw the guys agian behind tim\'s house. talked to Ian. He wanted tim. in the pants. Went back to Tim\'s backyard-ish. saw austin. introduced him to erblina, like we introudced her to ian. they thought they knew each other from somewhere, which was funny. Shes now known by them as \"erby\"..ahah i like it. so then we went into the front yard...and said bye to Ian, cuz prob wont see him again for awhile. *hugs. Then Tim\'s dad drove Erblina home, and i biked home, and yes, here i am. Well i think im going to bed. ive been sooooo tired lately. sooooooooo ttyl.  /X/ /N/ /O/ Vanessa.  '),(14583,3422339,'27,June,2004','2004-06-27 00:00:00','   heylo. this blog is for my friends, including amy. Ive been somewhat down lately...my friends know the deal  anyways, Vicki is awesome. i told her what was up, and she asked me to han gout and go for a wlak with her so we could talk, so thats what we did on saturday night. it was awesome. and then today, i was talking to her on msn, still sad, and she told me to be strong, and sent me a funny picture of her, lol, looking sorta confused in frotn of a tv with her winter jacket on and saaid it would make me smile. it=did. shes so awesome. And then emily. who has been so sweeeeet..lol and is helping me so much. shes great and im so glad shes around. even when her lover isnt..just kiddin em...you\'ll be fine without him..and even if you miss him, it proves how much your love is and it will make you guys stronger. And then Amy. i can talk to amy about anything. i bet she gets so annoyed with my thoughts and stroies, but shes always there for me. like tonight, i was...tears..and i tell her soemthings wrong, but its a long sotry over msn, and the first thing she says is \"call me now\"...shes the best, and we talked for like an hour! shes such a good listener and she gave me the best advice. i love her so much! amy your so awesome. well my day was liiike this: Woke up to the sound of my brother yet again. \"VANESSA!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\"-phill. \"wtf!? mmm goodnight!\"-me \"WAKE UP DUMASS WE\'RE GOINGGGGGGGGGGG FISHING!\"-phill \"catch me a biggie. goodnight\"-me nice eh? oh plus the whole: \"fine. dont spend time with us. bye.\"-phill *walks outta room, turns around, runs back. farts in my face. laughs. runs away.* aha. nooooot funn! so i got outta bed cuz i oculdnt sleep. well, i guess you could actually say i got outta the couch..thats where i slept. haha..ohh it was nice. watched some tv, took a shower...ate...went to play baseball with megan and lisa and her friend. was fun. then i went to tims house and made a cookie pizza with him and em. it was soo funn!! i enjoyed it. highly. then we went outside...went to em\'s. went back outside and lay in the back field..made plans to go to the beach this week. yay...cant forget about  that one. threw grass around. then a tennis ball. went home...talked to erin on msn...decided to go play catch with her and matt g. lasted about 10 mins. ate dinner...made plans for the night.hung out with erblina...talked....saw adam. went to the park..haha..ended up eventually going to tims house and hanigng out...lol im too tired to remeber anymore..ugh tim walked me home at 11.....umm that was it. saw ian. hes a good guy! so here i am. just got off the phone with amy. shes the bestest. number one. wow...its awesome how i can talk about nothing of importance in my bloggs...im such a loser..oh well.......i have no cares. and im done xno vanessa d  '),(14584,3422339,'26,June,2004','2004-06-26 00:00:00','   Wow. Today was a pretty badd today. Like awful. minus 647832965. Got up. Unfortuantly. i really wanted to sleeeeep. got ready. Christian came by to say hi. played basketball. he kept whipping the ball at me..ouch thanks. Went in, got ready. Mom picked me up. Drove to the family reunion. took like 6234789 hours...well 1 and a bit...but i slept in the backseat, aw well as my little babies cousins. im cool. not. Got there. said helllo...even though nobody really talked to me. because i was being \"shy\"..my mom even yelled at me for it. ah. so just kinda say around. ate chips...drank fruitopia.  It was sorta bad, because all the people who were around my age brought thier boyfriends, and were holding hands and kept walking around to make out..and all the people/kids younder than me wanted me to be thier mommy, while they playyed \"pretend\" and made birthday cakes. wow...to have an imagination. So i ended up holding my baby cousin..shes a cutie...and helping my other cousin, whos 3, pump water outta this old rusty well...it was hard..and then she just dumped out alll the water..haha good times. So then i went for a walk...went into the hay-barn. there was rope swings and like 3 levels. it was HUGE...i almost wanted to play..but i was alone and haha. yes. Anyways. started the BBQ...but i didnt eat..cuz i wasnt hungry. got yelled at by mom agian. ate some strawberrys...pakced up and lefft! ohh and i got 3 hugs by there girls..who were like...12, 6 and 3. thanks guys. i feel loved..and im done Took the van home..i even got to drive part way. that rocked! came back to my dads. called em. had a sorta bad conversation..wont go in to detail..but tears...i needed to talk to erin. went for a walk. came home. broed. went for a walk with vicki. got popsicles...talked..went to the park and talked..came home..ate and here i am. anyways...im sorta in a bad mood. i hate how guys think they can treat girls however they want, pick them up so high and drop them so suddenly like nothing happened...i just dont undersatand. -vanessa quote: \"hey you. pick up 2 horse shoes and put them in both your hands\"-some 6 yr old...talking to me. \"okay! why?\"-me \"JUST DO IT!.....NOW!\"girl..and she said it with so much attitude. thanks.  '),(14585,3422339,'25,June,2004','2004-06-25 00:00:00','   wow. what a bad day today...Lana\'s party was awesome tho...: My day. Wok eup by phill leaving for school. tried to sleep. couldnt. got up watched tv...talked to vicki and megan on our usual 3way calling..its our tradition now i think. showered. went biking with vicki megan and lisa when they camer to call on me. went to get jess and jodi..but they couldnt come cuz their bikes were outta air. so we all went home. ate lunch. hung out..bored...cleaned room...packed..and went to my dads. we were debating what i should get for my birthday. Our conversation: \"well i think a bathing suit.\"-dad \"..i have a bathing suit already..and i like it\"-me \"yes. but its not something your father approves of...\"-dad \"what do u mean? i like it...?\"-me \"well. its not something your should wear on ur weekends here. and you cant wear it when we go to visit your grandparents in Sault Ste. Marie this summer\"-dad \"well ill just wear a t-shirt over it\"-me \"no..they go see threw.\"dad \"what your point...ill wear a black one. ha i win\"-me \"...no.\"-dad \"so what u want me to get a one piece\"-me \"no..but something very close to it, unlike yours\"-dad *vanessa walks away* ...so yeah. got ready for lana\'s party. picked up em and went. it was hot. played tennis with sofo and mark..for about 0.46 seconds. it was kool. hung out. ate pretzles. then the band started. it was awesome...cuz there was only 8 \"groupies\"watching. except these lil like, 7 yr olds came up and asked to watch. we said ok. * then another lil boy came up, and hugged one of the girls.* \"AWW!! thats so cute! they\'re so young, and that girl has a boyfriend! look at them hugging\"-me \"...\" \"This is my brother!\"-little girl hugging boy \"Lmao!\" haha it was awesome so Delenda played. it was soo awesome...cuz with the garage door closed, we were about 2 feet away from them. lol it was nice. and loud..! at one point, the walls shook, and a hollow lil barrel thing fell off a shelf in fron of evan and ray. they were in mid-preformance..so evan stopped singng and laughed. it was veryy funny. so they played Bloody Romance and 187 by request. thanks guys! they rock my socks so far off, they\'re so pro. so then we took a break and had pizza.   thats when em and erin and kate and i had a little talk. em told me soemthing that totally changed my opinion..long story short..i suck at the whole \"guy\" thing. yah. i have the worst luck..again! haha but u dont care so im done.  thanks em. really. i love ya. thanks so much..honestly. So then we had bday cake for Lana! i love lana. she rocks. the cake was good too. then we watched happy gilmore. that movie makes me laugh..even tho i felt/feel like shittt... anyways. left at 11. lana gave me these pictures..and they made me smile Lana is so awesome. shes such a great person and she makes me happy. shes so sweet and fun...! :) so came home. to my dads. here i am...i dont feel like talking anymore..sorry..im so done. bye.  quotes:  \"Goose \'Pobfdufhjkfd\'\"-Evan \" Goose Problems!?\"me you really had to be there. we laughed hard and long.  '),(14586,3422339,'24,June,2004','2004-06-24 00:00:00','   Heylo chicks and dicks. wow. ive been so busy..its pretty awesome. so..back to tuesday, when i last blogged: Tuesday night i went to My bro\'s hockey. hes a pro. but they lost and now they\'re done, but their goalie sucked major fat ass. im sorry. but yes.  Wednesday ...uh..OH! Vicki called me in the morrning...and we had a \'3way call\' as always with megan. we met as Zehres around 10 50 ish...did our stuff...went to macs and gotr slushies!!! [heart] then we walked to the back diamond at Springdale Park and play baseball. we pitched to each other. which was veryyyy amusing..and it was very hot. then i went to get Erin..played more ball..then me and erin and megan went up to gateway. i wanted to find my sister to say happy birthday. which i did. HAPPY BIRTHDAY AMY!!! LOVE YA HUNNNYY SISTER! lol, so then...me and erin played more ball, vicki and megan left and me and erin went to her house and chilled..watched wierd things on her comp..lol like this freak thing where faces pop out..it  was scary..i was under the blankey..im such a freak. then we went out to play alittle bit more catch. saw Josh. erin read his report card and me and him played catch. he left. erin and me played more catch. cept she kept making me move and making them role down the streets. thank you so then i biked home! got ready and biked to softball practise. then it got cancelled cuz our coach hurt herself...so we went to megans house to have a party. (we=jay (ray!), jodi, jess, megan, lisa, angelica) it was awesome. we sat in her backyard and ate and talked...about everything from guys sex and baseball....then megan called our coach. shes cool, cuz shes onli 19!! lol, so then jess started yelling random things up to megan\'s bro\'s window...it was great. so megan makes the best cupcakes. for sure..and shes a pyro. major. lol she stuck everything into the misquito candle. then we went inside cuz it started to rain! haha...but yes, we saved the cupckaes...watched tv..erin got there..went to megans room and chilled. looked at her scrapbooks..containing pics of me in kindergarden. wow...no comment. then vicki showed up awhile later..it was fun. lisa drew a pic of a flower and we all wanted one. lol...anways..then i got a ride home with jay..and slept!  Thursday..today! whew! hmm woke up early.. iwas sooo tired but i couldnt sleep!! shit shit! umm sat around in my pjs...did luandry..realized i had a doctors appointment so i biked downtown. found out im getting tested for asthma..and i got a puffer/inhailor. yay. no. biked home...took down laundry...got ready..biked to erins...and went to see my sister amy before her grad. she was soo nervous, it  was soo cuute!! shes awesome., went back to erins, went swiming for like, 2 minutes!! yay, it was so nice. and kates swimwear was hott. because it was alll pink! woo props.  so then i got out to get changed...realized i forgot \"underwears\" and had to wear my bikini [soaking wet] under my clothes..lol crap. met josh josh at the park. went to play some baseball. we pitched. and he trying to pitch under hand which was sorta funny.. haha...cuz i was soo better!! lol it was soo fun. then my bro showed up? after going to his girlfriends house..lol..so phill and josh talked about girls...which was a dirrty conversation. Phill left..and we played somemore ball...then we left!! haha..it was fun tho:) :) :)..xox. so then i went back to erins house..for like 20 mins. then i biked home. got changed and ate dinner. then daddy came and picked me up and took me to his house for 5 mins so i could see amy all dressed up. wow, AMY LOOKED BEAUTIFUL! shes always pretty, but tonight she was stunning. wow. i loved her dress and her hair and her shoes...wow. shes so awesome!! [hearts]. so then daddy droveme back home. mom dropped me off at batting practise at exhibition. which was cancelled part way threw due to the rain. so we watched the 10 yr olds rep game. and we cheered for them. which they found so much pleasure in. wow..it was nice. cept i got soaking wet!!! and i was onli wearing my baseball shorts, socks and a t-shirt! haha so jay let me wear her Track and Field sweater. which i loved cuz it was soooooooooo cool and warm and comfy!! anyways..i bet im boring you now came home, and here i am well i actually better go. im veryy tired!! x n o\'s! vanessa vanessa baby.  PS> To Amy! Aw I know you\'ll be reading this! so i just wanna say im proud of you! and your so beautiful! and ur soo awesome! and i luv u alweez! enjoy grad! sisters for life! blondies forever! cia!  '),(14587,3422339,'22,June,2004','2004-06-22 00:00:00','   hiio. Im all done Grade 9! wooo. feels very crazy actually!  Anyways, My Life: Monday...well Sunday night i slept at daddys. got up Monday and went to Gateway (yes!) with Erin to coach thier senior girls 3-pitch team. it was awesomely intence. Erin and I coached 1st and 3rd. and really, we\'re pro. My Sister Amy did awesome. i cheered for her. She got a Homerun!! It was awesome. they won all the games that we were there for.  So then at 1, my mom picked Erin and I up and took us to Waverly, to watch the boys play. wow. they kicked ass. major. they won the tourny...and they played awesome. I think they all got homeruns. it was that good. Me and Erin watched 2 of thier games, they rocked!   Anyways, then i drank so much that day i had to GO so badly. it was insane. Erin kept mentioning water as i fantasized of deserts. wow, thanks. then she laughed at me..! AND I GOT VERY SUNBURNED!! I HURT. OUCH.  So then, we took the bus..she went downtown and i went to work at the Hospital..which was veryyy boring...but im up to 49 hours!! yay! so then went to my dads, then to my moms and studyed for science all night.  Tuesday...today...had my final science exam. sucked, but wasnt too hard. but everyone who said it was very easy was insane, because it wasnt at all. but now im all done, so its all good..took the bus home all alone. ate. watched tv. then i went to em\'s house i changed into my bathing suit, in an attempt to tan..haha.   we sat in her backyard, and the sun was barely out. but everytime it came out, we stripped down. lol..it was funny. then we mande strawberry smoothies, which were incredible. and 3 bugs landed on em. i laughed. so we sat and talked for ahile on the blanket in her backyard...it was nice. then we decorated her basement for her lil party thing tonight. i was pro...i wrapped the long paper stuff around the banister, then backwards..its hot. and we got ballons to. so then we walked up to gateway. i saw amy. and we walked home. finished decorating. then em decided to open up the noise makers she bought, to try them out! so she blew in it....AND IT MADE NO NOISE! you shoulda seen the look on her face..omg it was soo funny, cuz the noise makers made no noise...none of them did... it was awesome...then she went to walk into the laundry room, and pushed the door open when it wasnt open...so it didnt open and she just kinda hit it...wow i laughed. i told her to turn the handle..haha good times.. so then i biked home, and here i am. woooo. so i better go...i dont know what my plans are for tonight. i kinda wanna get outta here....anybody!?...? please.  forever vanessa.  Good times: *Me and Erin watch some wierd girl run onto the baseball field just to give Blair a stuffed horse, and then she gets yelled at by mr. c* \"hahaha, what was that about\"-us ...a few mins later... *Blair comes in and tosses us the horse* \"To Blar from your lovers!?\"-me reading tag. \"hahaha, BLAR!\"-us \"She looked kinda like a horse\"-Austin \"...thats why she gave him a horse...she had to give him something to remeber her by\"-me haha it was funny, then at their next game: \"wow, Jason (H) is veryy hot\"-erin \"..haha\"-me \"Id bang him!\"-erin \"so would I!\"-all the guys on the bench \"of course! even id bang him! even jason would bang him!\"-me  ..so i guess you had to be there. we laughed pace.  '),(14588,3422339,'20,June,2004','2004-06-20 00:00:00','   Well, its late and i should be in bed. I\'m not. I decided to make a BLOG about Amy. because shes my #1 fan. haha, oh and i luv her! So first off, Amy=blonde. no, not just pretty hair, (and pretty toes! lol) but extremely BLONDE! wow. heres justa couple examples  One day we decided to go swimming at the West End. so we get outtta the car, amy\'s carrying her towel, and she looks down and she says:\" Hey guys! where are my clothes!?\" The funny thing is, she was wearing them. overtop of her bathing suit! haha aww that was a funny one. My favourite was this one: I was on msn, and she msg\'s me: \"hey vanessa, can u print something for me? MY PRINTER ISNT WORKING!\" \"yah sure\"-me *I spend 35 mins trying to open up the file she sent me on my computer. i try in Word, on Paint and in the Internet, over email and over msn. but it just wont work. so we;re bothing going insanely frantic...! *amy msg\'s me agian* \"...its working now!\"-amy \"oh thats good, what did u do to fix it?\"-me \"I turned it on\"-Amy WOW! i laughed so hard.  Me and Amy have also had our good times too. like the wedding of our switching parents last year. yup, they switched. weird huh? So we wer eboth to enrvous to walk down the aisle by ourselves, so we walked together. then her mom and my dad were saying thier vows, and my and amy were standng up with them, across from my brother Phill. he was making faces and imitating the minister making US laugh. it was embarsing! but funny. then the dance was the BEST part, right amy? lol and the golf carts..and Doug[heart] Which reminds me of yet another moment.  Amy and i were sitting in the drink/bar area with Doug. [remeber, the wedding is at a golf course] *Amy looks onto the wall at a plaque, as she debates what to get to drink* \"i Think ill have a drink of \'Eric Puss\' \"-Amy \"hahahahahaha\"-me and doug \"what!?\"-amy \"That plaque has the name of golfers. not drinks\"-me HAHA! get it? wow!..if u dont...\'Eric Puss\' is a golfer...amy thought it was some sort of fancy drink. wow then there is our looooooong talks about everything. lovers, school, hot guys..lovers..guys..haha, yeah. well there is probably sooo much more i could say. but im so tired. and im going to her baseball tomorrow, so i better sleep.  AMY! heres your blog! i love ya hunni! always here 4 u! *Step sisters for life, blonde sisters forever* XOXO Vanessa.  '),(14589,3422339,'18,June,2004','2004-06-18 00:00:00','   hio. My Weekend: Saturday~ Baseball Tournament. \"Twisty\"...all teams were switched around so there was people from all different teams on 4 teams. Which was the lamest idea ever. The point was to make \"new friends\" and be less competitive. my ass. Nobody had any effort. Oh well. Lost first game. went home, to tim hortons, BBQ, back to the Diamond. Lions Park. Tied second game. agianst the team my coach was coaching. played well! best game..friendly. Lost third game. 14-11. sorta sucked. I was playing second in the last inning with 2 out. Erin was catching. The runner on first was stealing second. Erin threw the ball to me for me to get her out. i went to tag her. she pushed me over and landed on top of me. landed on my elbow and got dirt everywhere...by the way, i got her out. im still in pain. thanks. was cold and windy all day! So Then came home. msn. dinner. saw Chrsitian!:) showered, went to Tim Hortons with Josh and Em. Hung out. was funn!;)..hugs. Sunday~Woke up, went to Church. Came home. went to mall avec Erin. Bought a new tshirt from American Eagle. and a Father\'s Day Card. Came home. Relaxed. Ate Dinner. Now I\'m going to my dads in an hour for the night. we\'re going for ice cream. im buying. happy fathers day. Then Erin and I are going to watch the MMVA\'s. yay. were supposed to swim. but i doubt it now.  anyways better go study for Science. frik frik...onlii one more exam! yay. Latar... v. quotes.:. Today at The Mall: \"Hey Christian! you look so clean!\"-Erin. since this is the first time she saw Chrsitian in a few months. \"thanks!\" \"..hey didnt you wear that shirt 2 days ago!?\"-me \"Yup!\"-Christian ~ *I Get off the bus, some guy is walking in front of me* *Guy turns around randomly* \"hey you see this ring! it a mood ring! changes colours with you mood!\"-man. points to his finger. \"...neat\"-Me *Man just walks away* Thanks.?..wow that was random. ~ *Sitting around dinner table, after i tell my sister about the mood ring incident* \"I remember i used to have a mood ring as a kid\"-mom \"haha\"-amy:) \"i did too. but i think it was broken. because it was always black!\"-me. \"hahahahahaha\"-everybody \"mine was always green!\"-amy. ~ *at Dinner* \"hey dad where are we going on summer vacation?\"-amy asks wayne. \"well, its T.B.D.\"-wayne \"..to be determined?\"-me \"to be Decided!\"-wayne\" \"...oh\"-me \"no, but really where are we going?\"-amy \"G.T.H.\"-wayne \"Going to Hawaii?\"-Amy \"GOING TO HELL!\"=me. ...niice eh?...cia.  '),(14590,3422339,'18,June,2004','2004-06-18 00:00:00','   wow. that was the topic of my english essay today. wow if i have to write that word [love] one more time, ill die. well, maybe. yuk. sooooo ill back up. Last night i had a baseball game. Vicki Lisa and me biked there. yes, from the west end, allll the way to Victoria road and St. James..or St Johns...or something like that. yeah. it was fun!!! it only took 40 mins. wooo, and Lisa is pro. So we were the first ones at the diamond. played catch warmed up. the game was amazing. cept the ump, who sucked ass. seriously, like wow. second and third inning i pitched. went well went well...we kept going into extra inning\'s. it was a veryyy good game. 7th inning i caught the ball, as it was almost at the other teams bench=way outta play., i ran from first and just made it. yes, OUT! lol and then the bottom of that inning, i was up agianst like a REALLY fast pitcher. like wow, shes pro. so it was full count...and i fouled the ball...times 9! 9 fouls in a row, behind the fence. on the plate, nbehind me, ya, 9 in a row up to bat. anyways, she threw an inside ball and i walked. got to first and told matt it was fun. ended up scoring along with Lisa and we won 20-19. Lisa hot an awesome hoomer. woooooo and we played well concidering we were down 12-6. anyways, i bet your bored with that so ill shutup about my life now.   So today i had an english exam at 9 am. yuk, too early. just wore sweat pants, which were 758943278 sizes too big. they kept falling down. im sorry. Anyways exam was very easy, but my hand felt like it was gonna fall off after i was done, and i was freezing cuz the air conditioning was blowing right down on me. thanks!  so i satyed an extra 2 hrs in the caf with nikole. missed 3 buses..haa! so we ate and talked and some people studyed. wrote lots of things on the table. [Ben (heart) Penis.] lol it was fun.then i took his pen and thought it was niki;s...blah.. icant remeber what else happened. ben put a hole in his water bottle. water ended up on the table. i wiped it up with his penicil case and nikole used a peice of paper wow im really boring. im gonna stop now.. tonight mall with erblina. saw a realll hotttt guy. wow. got talked to by wierd guys in a van. bought stuff at \"La Senza\" .and stuff. ate ice cream which torry owed me. finally. well i g2g. got a baseball tounry allll day tomorrow!! TWISTY!  xno ness.d.  '),(14591,3422339,'17,June,2004','2004-06-17 00:00:00','   hio hio. just got home from writing french exam. wasnt as bad as i could possibly imagine...im pretty sure i did ok. *fingers crossed* stayed up studying last night till 12 30. couldnt sleep till like 1, my sauna was too hot. so i have baseball tonight. vicki lisa and me are biking there. all the way to the other side of town. st. james diamond. wow...!! fun!!:) so im going now. stuyd for english tomorrow...oh well. that wont be hard. yessss.  =x=n=o= vanesssha me=wearing no makeup today.first.  '),(14592,3422339,'16,June,2004','2004-06-16 00:00:00','   hio. so im sitting here taking a break from studying, listening to \"Flip, flip , flip, game ovar!\" on the radio, and the LOUD toe-truck outside. eating peanuts in my baseball uniform and shorts. wow. im so cool. not. Today was the last day of school! last day of grade 9 ALREADY! wow. it was sorta crazy!! In french we did work! like seriously! who does that!? In Science we did a lil exam review, then ate chips, timbits and drank coke. lana and me talked, and she made a card for ms ditchfield. we all signed it. gave it to her. she=cryed. then she wanted to give us all hugs as we left!!! \"Aw! shes SO cute!\"-Lana, about ms. d. every day!  During lunch we took lots of pictures, and ate. wow. it was kinda fun. i got a freezie and candy. i love freezies!!! and i love Erblina!!!! shes always there for me when i need her. i love ya!!  After lunch, had english. we did exam review on A Midsummer Night\'s Dream. it was soo boring. then we got the last 20 mins to take pics. the guys played cards. tim stole my camera. wow, thanks.  Then in gym we did a fitness assignement thingy. it was gay. then we played British Bulldog...aka British konrad when used with Chickens. lol long story. got changed, and school was alllll over. walked home with lana. talked alot. was good times. wow, so then i got home, went on msn, went to baseball for an hour. we played a scrimage. wow my hitting sucked ass tonight. all the balls i hit were caught. and then i caught alot, so it was allll good. biked there with lisa vicki and erin. biked home with erin. we had a truck full of middle aged men stare at us. wow, no comment.  came home. studying my ass off. but im sick of it. je suis hate francias. tres much. il est should die. je suis done. wooo.  anyways, better go. the toe-truck is stilloutside. shit. it sounds like a grabage truck. meh.  .x. .n. .o.  Vanesha babe.  Funny quotes today: \"Vanessa i love your earrings\"-everyone! thanks guys. but lana said it \"thanks!\"-me *lana looks at one side of my head* \"omg turn around i wanna see what the other one looks like!!\"-lana OMG! hahahahah what a blonde.  \"hey vanessa, has anybody ever told u, that u r ugly?\"-nick Thanks!  '),(14593,3422339,'15,June,2004','2004-06-15 00:00:00','   hio. short blog today. not much time, and tired. must study!  This morning my sister left for Ottawa. im so jealous. last year it was so fun. wow, luckkyy her.  Today was nice and slowwwww. french we studyed. science we learned about space. gym we did health writing thing. english we watched the rest of A Midsummer Night\'s Dream.  Lunch was the bestest. kate brought her digital, so we took lots of pics, and made a lil movie thing. i love it. cuz mark looks retarded in it. which is sweet.  Only 1 more day!! its soooo exciting, sorta sad and CRAZY! wow. good times, good times.  |x|n|o| cia.  *make these moments last, make this time into something, dont leave this lifeless.*  '),(14594,3422339,'14,June,2004','2004-06-14 00:00:00','   wow, today was one of the worst days. Boring day at school. except english, b/c we got to watch the drama classes preformance. yay! it was fun. kinda funny, sorta kool. i liked it. Then i took the bus bus downtown with kate. saw the guy she has seizures for. wow. hes special. saw maria too. i miss her. then i took my bus bus to the hospital, for work. which was very boring.: *i walk in, to sign in, wearing my shorts and tank top* \"hey Danielle!\" Danielle! DANIELLE!?\"- my boss, as she yells as i walk away. i am still wondering if she ment to say vanessa, or was just yelling names randomly cuz i was the only one around..hmmm.. so i get into work. do the usual. get a call from my mom: *ring ring* \"hello?...oh yeah...ok one second...vanessa, PHONE!\"- the guy i work with. \"hello?\"-me \"hey vanessa, just checkin to make sure u got to work ontime and alright\"-mom!! \"no...yes..i did\"-me \"oh okay, and ur boss came into my office and pulled me aside and told me that you got 2 complaints for wearing shorts to work last week, apparently its not allowed as part of the dress code\"-mom \"..ohh...ok?\"-me. as i fucking stare at the sign on the wall saying to wear pants or shorts in the summertime. wtf!? \"so i just thought id let u know, bye.\" nice eh? so then, cuz work was so boring, i watched the coffee pot brew new coffee for 10 whole minutes, then i re-aragned the gum packages. wow. intence.  the only good part was when this HOTTTTTTTTTTTTT guys came in. i call him Orange Crush. *Orange Crush walks in, goes to pop and gets an Orange Crush* \"heyy there\"-o.c. \".........HI!\"-me. in awwe. \" one dollar please\"-me \"here\"-o.c. \"thanks\"-me \"see yah :)\"o.c. wow. it was nice. then i went to my dads, then came to my moms. got kinda wet. went on the treadmill. now i should be studying.  but im not. im actually in a very bad mood. i got my hopes up when i really shouldnt have. i dont know what to think, and i dont know what to do. i wont explain it, because you really dont care. but im sad. i just really am.  '),(14595,3422339,'12,June,2004','2004-06-12 00:00:00','   goooood evening. im pretty tired now...buttttt for all u losers, heres my day...joking about the losers.. Woke up to the sound of the lovely fuckin phone. i swear, dillon wakes me up everyday. thanks. \"hmmhmhmhmpp hullo?\"-me \"hi! is phill there!?- dillon \"ugh, no. hes sleeping. and so was i, until u woke me up\"-me \"ohh..................................... ..................................is this vanessa\"?-dillon \"...yes.\"-me \"ohh hi!\"-dillon \".....yeah, ill wake phill in 20 minutes, so call back..in 20 minutes ok?\"-me  \"k, bye\"-dillon *click* *5 MINUTES LATER!!* \"hullo\" \"HI! is phill there\"?-dillon FRIK!  so yeah. after that, did the shower/get ready thing. made some plans went out back and played catch with the faw-shaw. attaemped to pitch, but it wasnt too grreat today. oddly, i find when im nervous, i picth strikes. wehn im not, i suck big time. hm???!! then i cahnged and went to meet nikole. came back here. Then we put a blanket on the front lawn, got into our bikinis! and yes! we tanned. well...i sorta burned..but hey, red is better than white. no im not racsit. lol..k. so we just kinda were out there for like an hour. i painted my naails, witch look awful. and then my toes. we ate ice cream...ummm she was admiring a certain somebody on my sreet/stalking him. it was verii funny actually. lol you had to be there. \"ohh there he IS! oMG OMG OMG! i am so going to marry him. that shirt looks soooo hot on him!!\" -us. well mainly her. but i agreed.  then we walked down to macs and got some slushies. came back hurr. sat outside ...watched mr. hottie. ate dinner, then walked to her house to get a sweater, then to my moms hosue to get a sweater.apparently my mom had a garge sale today. thanks mom. so on the way back and forth, we had like 18476723 guys staring at us in cars. no, not that we\'re cauky. we looked at every car, then they looked back. except for this guy, who i duno how hard he tryed, couldnt make it more humanly possible that he wanted us. lol..aww yes.  came back to my house. sat outside watching this guy yet agian.  played some basketball. cuz we=pro. then kate came, she went on msn and made plans.  met up with ray mike c and evan at TE and walked to tim hortons. ate...blah blah blah...went to the gas station...came back here to my house.[only because they think my sister is hot] the guys sang kareokee, which was probably one of the must humourus things i have ever seen. seriusly, cuz they sucked sorta. then ray kept complaining about his sore knee. it was fun. then evan broke one of my picture frames, and glass went everywhere. shit shit. thanks. then they bit my pink candle!? weeew...nice? so anyways, im tired, and im going now. see yaw. x.n.o. vaness. [red/white.flowered.hawaiian.bikini]  Quotes today: wow there were so many..cant remeber tho. cept this one: *we all walk up stairs to see the huge buttons on my phone, just after evan broke the picture frame* \"HEY EVAN!why dont u tell vanessas dad what you did!? yes you! in the ORANGE shirt!\"-nikole (my dad is like, 2 feet away) \"ugh NO!!:)...can i have a donut!?\"-evan \"no, tell him first\"-me \"yeah! dude in the orange shirt! tell him what YOU DID!\"-nikole *evan almost takes shirt off* \"heyy man holding the orange shirt-!\"...me \"hey mike wanna switch shirts...?\"-evan ahha wow. hada bee thur.  '),(14596,3422339,'11,June,2004','2004-06-11 00:00:00','   heyo. I swear i have a memory problem. cant remeber anything that happened today. got tibits in scence, cept they were gross, cuz all the good ones were taken. watched A Midsummer Night\'s Dream in English. wow. it was actually a veryy kinky movie. it was sorta kool...mabe. Tonight Kate came over. Em brought us her straightener, so i spet like an hour straightening my hair. i had fun! cept it looked bad. darn. then we were on my sisters Kareokee Machine. it was aweosome! i swear. i think we should have a kareokee party sometime. i had fun, lots. Then we went out for a walk, thing. did nothing really, cept talked. i made a new word. frik i cant remeber what is was...shit shit. So NOW me=bored. i have no life. seriously. well i do. but it is not the greatest at the moment. oh well. Phill needs comp. must go. late-ar. xno ~vanessa baby. QuotersToday: none that i can remeber, man i hate this. \"back back back up, reverse reverse\"-Tim *Tim walks by Me and Kate in the hallway* \"ciz i, i liked you, the way you were\"-me and kate at the same time *walk by tim* .....it was so cute/funny/amusing.  '),(14597,3422339,'10,June,2004','2004-06-10 00:00:00','   Hio! Today was not as hot. which was sorta nice, plus sorta not...woa dude i rhymed. i cant remeber anything outstanding happening. except the very unpleasent gym classs. which was sorta bad. but oh well. haha everybody is slamming my door. yesterday my mom was mad and slammed it in my face. 1 minute ago my bro came in begging me to do his homework. i said no. *slam*..thanks . So i had baseball tonight. our team played awesome. and that one cheer was so loud, my ears rang. way 2go ladies! i pitched 2 innings. one of them was good. no runs. the other..ugh..practise this week! yes?.. We ended up losing in the last inning. by one run. frik. we were so good tonight! But the sun was so bright, and it got in my way a couple times. damn. o well. Got donuts after the game. cuz stacey\'s beday. Happy sweet 16th! ;) Got home, went for a nice long jog, i still had energy to burn. woooooooooooooo. anyways, i bet im boring u now right? yes okay.  no plans this weekend.! yet. anybody whos free, please give me a call. if u want to chill wit me yo. haha woa that was gangster. :) X N O -Ness vanessa vanessa baaaaaaaaaaaaaaaby  P.s. Kate makes me smile:)  '),(14598,3422339,'09,June,2004','2004-06-09 00:00:00','   Heylllo. Today was so hot. i thought it was gonna rain so i wore jeans=big mistake. The classrrooms were so fucking hot. almost died. almost then the popele from gateway came. i snuck outta science to say hi, figures i didnt find any of them ,including my sister! frik. i love freezies. Then this afternoon was just hot. took the bus bus home.  then i biked. had baseball practise, but our coach cancelled it after 30 mins. cuz it looked like it was gonna rain. she also got incredibly pissed off at us. for NO reason. long story. biked home, went to erins, chilled. \"on the BIKE!\"-erin. ...haha aww we laughed for 10 whole minutes. wow, what fun! you had to be there well i better go.  x  n  o -Vanessa. baby.  ...why do u look so sad? the tears are in your eyes dont be afriad to cry, let me see you through i see the dark side too. let me love you now...  \"Im Blind! So i cant speak English!\"-Mark  ...haha i laughed.  \"hey Erin, you kno- oh fuck omg omg omg!! EW EW EW EW!! ahhh! yuk!\"-me, as a bug of some sort flew into my mouth biking home. \"...\"-erin ..haha what a good day. frik. life is so confusing. it confuses me. mwah.  '),(14599,3422339,'08,June,2004','2004-06-08 00:00:00','   bfuidgha bvhubh hbufdsghgjfkyh gfjdk? hei. im so tired. tonight i went on a looooooooong bike ride. then, i biked to my dads, went on his treadmill for 45 mins. straight. sooooo i got my exercize! haha. yes anyways, still pissed i couldnt go tonight. oh well, its over;) better go get some sleep. im done. x n o  *These are my confessions Man I\'m throwed and I dont know what to do This by far is the hardest thing I think I\'ve ever had to do*  '),(14600,3422339,'07,June,2004','2004-06-07 00:00:00','   wow. its soo hott out. its insane. like 32 or something. im sitting in my sauna room. its that hot. and the window is wide open, so im feeling a little bit of a breeze. on top of feeling hot and sticky, i am officialy allergic to warm weather and summer. it sucks. i havent sopped sniffing all day, and my allergy medicine all the suddn decided to go on strike, the day it gets hotter than ever. wooo life, gotta love it. So my day? well i wore my favourite skirt. its kinda nice. so in french i had my inclass exam, and reading and answering questions in french. i think i did ok, thatnk gosh. oui oui. Had a pretty boring day other wise. i feel sorta bad latelu. im sorta confused. In gym we did the beep test. im thinkin i have athsma. cuz i sorta had a huge loss of breath, and got like frikkin 5.5! yes, wow, last time i was like, way mroe than that! seriously, this time sucked. but ill shutup now. then to top off my day, we sat in english for the ENTIRE 1 HR AND 15 MIN. PERIOD IN THE UN-AIRCONDITIONED ROOOM SITTING LISTENING TO THE TEACHER READ SHKAKESPEAR. what torture, i almost died. almost. so came home on the bus bus, went outside with a popsicle and attemped to tan. now, im doing nohitng. i really wanted to go to the Athletic bacnkwet tonight, but heather took her ticket back from me. shittttt. im so pissed!my life is sucking right now. :( anyways, going to go sleep in the friut cellar. mmm coldness... im such a dork!!!!:) :) :) yay!!! x n o bored vanessa.  '),(14601,3422339,'07,June,2004','2004-06-07 00:00:00','   hiughfg hey. im feeling kinda sick and i hate it. hug? anyone? no...? today was yukki. nothing special. school, then work. which was boring. excpet for when some guy bought me a peice of cake. and some lady was giving me her money: \"okay, that will be $ 2.45\"-me \"okay, 1 sec, i think i have a 10.\"-lady *Pulls out a pad!* \"...hehehe hhmmmphmhmhp\"me laughing then trying to hide it \".....\"lady, 9acting as if nothing wrong and keepin a staight face) well i laughed. then dad pciked me up and i went to the gaylic gig. there was nobdy there. but the bands were good! yay! then i got the bus home. i had to run for it. i felt like a loser. got home, did hw, went on msn, watched the hockey game. fuker. calgary lost. shit shit. NOW im debating whether or not to go to the athletic bankwet tomorrow. heather gave me her ticket, but i dont know if i should go...hm..any thoughts? i wanna, but i dont know if should...but i wanna..should i? lol...hmm frik i have no idea. but i feel sick. so i better go, cuz i feel sick. and liek shit :)byee xno  '),(14602,3422339,'06,June,2004','2004-06-06 00:00:00','   heyyllo there Im having a moment right now... lol. i was just readin my firneds bloggs, which rock. and most of them are mentioning the next 2 weeks. and the last 8 days. Only 8 more days of high school! 8! then 3 exams, then summer. it went by so fast. it seems like a month ago we were walking down the halls for the first time...getting lost! lol. but now, its crunch time. 8 days of school and its all over. already. grade 9 will be done forever. wow, thats a crazy feeling. its almost scary.  \"make the best of the moments, cuz they wont all last make them last as long as you can make something of them, make them worth forver, cuz when its all over you\'l regret not smiling one more time\"..song. anyways. my life today? could have been better, but i was happy!:) i went shopping with my mom. bought a red and white hawaiian bikini. so about that exercize and healthy food...:P then i did homework, hung out around the house..relaxed, hopefully well better go study. gonna fail a french tomorrow. *yay*(sarcasm..)  blahhhhhhhhhhhh  .x.  .n.  .o. vanessa dipasquale.  Todays ONLY funny quote \"i like ur necklace, is it new?\"- my MOM! \"ugh no...does it look like a guys? or a girls?\"-me \"umm both..whats that called...Bisexual?\"-mom \"HAHAHAHAHAHHAHAHAHAH! omg, no! Unisexual!\"-me \"...........hahahahahahahhahahahaha!!!\"-mom Yay!...no.  '),(14603,3422339,'04,June,2004','2004-06-04 00:00:00','   Heyllo. Just doing some French homework, decided to take a break to blogg, since i havent in 7589475 days.  So..Saturday what did i do saturday? ohh wait, i got up at 9ish. fell asleep till 9 20. somebody was in the shower, fell back asleep till shower was empty at 9 30. got ready, went to dragonboat races at Guelph Lake with my mom. it was sorta fun. watched the races, helped mom work, ate pizza and a drumstick...that was the best part. finally came home around 2. Then i hung out with Sylvia and Kate. we went to zehres to get Sylvias pictures, i had to buy hotdog buns, then sylvia bought us milkshakes (they were really good, thanks Sylivia (heart)) then we were walking home. sylvia still had here milkshake cup. she all the sudden had the urge for tadpoles. so we walked into the ditch and into the fron place of zehres, where it all dirt, and there is a tiny pond beside the road/sidewalk. since i was the one wearin flip flops, i walked over, and got her 2 tadpoles. one was a baby. i caugh them in a cup then we left. Kate named them Ethuce and some other wierd thing, like a combination of bruce and buddy and ethel. wow kate, creative! then we all took turns holding the cup, cuz we are so cool, and lame. and i swear, the tadpoles were having sex in the cup. frik. get a frikin cup!.. so then i dropped them off at my house, got yelled at by my mom for not going mini golfing with them, and went to sylivias. we hun gout for a couple hours, went on her roof,looked at yearbooks and magazines, ate pringles, went on msn and talked on the phone. left around 7 10. me and kate went to meet em at st francis. walked around, saw some guys, but then got hungry so went to my house. kate had my chicken burger,, and licorise for em and me. and a banana. then we went tback to the school, and guys were playing basketball. we hung out. and went to the park, and the swings. kate lost herphone, so we looked for it. i found it. yay. called home and found out the score of the game was 2-1. frik. then the guys left. me and kate walked back, and mike decided to come with us! we satyed at st francis till 11. we slid down the pole thing. i sucked at that. yay. then we talked and watched the stars. ben came back eventually too, and kept falling off his scooter. him and mike raced down the railing pole things. and me and kate did. woooo it was fun! then we walked home! kate got pikced up and i watched the end of the game. Calgary lost in double overtime, 33 fuckin seconds in. i lost a bet now with my brother and dad. crap.:( then i came on msn  and talked till like 1 30 to mike!:) good times. well i better go do homework again. then im going shopping with my mom for a new bathing suit:) YES! latar- x n o  '),(14604,3422339,'04,June,2004','2004-06-04 00:00:00','   Hello everybody who is reading this. and to everybody who is not. including you. the one staring at your computer screen .yes YOU! wow what a greeting. so today was semi- ok. it was friday thank goshhh! i dont know what happened really. i think i have short term memory loss. so ill skip my day, and talk about tonight. ;) by the way, i was such a dork today. i have bad timing. frikkkker.:P So Erblina, Kate and Jenelle came over around 7ish tonight. we ate food, and watched Euro trip. it was a veriii funnyyy movie! i luved it. i laughed alot too. laughing is fun. you should have been there. so after that, we went to hang out outside. we sorta walked around in no certain direction, which was kinda niice. after awhile we walked back to Jenelle\'s house. and she left us. then we went to this guys house. and stoddoutsdie talkin about who would go up to the door. the guys dad looked at the window and began to watch us so we all ranin different directions. then we went back 5 minutes later, and he was still watching. so we sat on one of those green boxes and watched the stars. sorta romantic you could say. except there was no guys there, so it wasnt romantic. mabe next time guys? ok. This whole time i brought a bottle of \"stuff\" outside and was walking around drinkin it. unfortunaly it was just pop:(! lol. so then went back to my house. kate went on msn forever. erblina left. then kate left. i cleaned up, and now im very tired. i like sleeping. its nice and neat. so i think il go do that now, and ill leave u with these evenings quotes: \"k mom, i got the movie\"-me in Blockbuster \"lets see..hm...wait, this doesnt look that nice\"-mom (Commenting on \"Eurotrip\") \"oh well, no big deal, come on\"-me *read back of moive case and raises eyebrow at refeerence to \"sexual things\"* \"...ready to go? okay!\"me, and yes. i got my way. IIIII rented it.  and then.... \"hey Ladies!\" - hot guy in car driving by *guys yell things at us from car and van as we walk down the street...sex related of course! lol* \"heyy guyys\"us \"wooo\"-guys/ then they drove away \"OMG OMG OMG!\" us. especially kate. \"omg! that was s**** f*****!\"-kate. she as freaking out cuz the guys \"hit on her\" it was amusing!:) kate they wanted you! haha, and then THE BEST QUOTE EVER! *guy from eurotrip goes into random places. example: jumping thru a bush into a hot tub and a random girl* \"oh!...this isnt where i parked my car!\" haha that was the best.  X N O Vaness  There are easier things then finding a good BF... nailing JELL-O to a tree, for intance.. (haha, i just found that. kinda funny? no? didnt think so.)  ~you make me the happiest i ever want to be, and all i can be. you give me butterflies, and they take me away now all i need is you to be here, thats all i ever want~ {hearts}  '),(14605,3422339,'03,June,2004','2004-06-03 00:00:00','   heyyylo:) Today was soooo fun. went to the science center...on a science trip. wow, it was great. the bus ride there was kooool dude. i sat with lana. read magazines. didnt get to listen to music. great. talked. was fun, got there, split into groups, and had to go to these exhibits that were dum, cuz we had to fill out 67565863 worksheets. then had lunch. i ate. it was nice. then i bought an ice cream oero for like 2 75$..ripoff man! so then we went to another exhibit...then we got to go wherever we wnated. naturally to the sports place. got to see how fast we threw, did a rowing thinger againing lana! and a balance beem. which was fun, cept smelled like stinky footsies! then gabe took everybodies name tag and lana wrote on them and stcuk them too him. he had like 10. and they were all sexual. hehe!!eah. on the bus home, erin was maing up random songs about cars and laughing her ass off. an then i said something not remotely funny. and she was hysterically laughing. lol, then she spit while she was talking. in slow motion it went on her arm. you had to be there. it was sooooo funny. and after that, she sung more about cars. to the turn of every fuckin song! so then lana switched seats with me, and she sat beside erin and i sat beside adam. then em kept falling asleep/closing her eyes and we were luaghing. then there was some guy in a red car, who had been beside our bus for about an hour. on the busy highway, and he was starig at us! so Lana waved and he waved back. then we got some paper and adam wrote his phone number on it and put it agianest the window. then the guy turned off the highway. it was pretty funny. so then emily got a lovely show by a couple behind her. ;) she liked that tongue. haha. well i cant rember anything else. shit shit. so i had baseball tonight! AND WE WON! 13-12!! GO RAGTOP! we had OVERFLOWING team spirit. we cheered the entire game, no joke! we sung thisone like 657482657438567 times for real: \"lets get pumped up!\" \"LETS GET PUMPED UP!\" \"lets get get pumped up!\" \"LETS GET GET PUMPED UP!\" \"lets get get get pumped up!\" \"LETS GET GET GET PUMPED UP!\" haha. and the other team got sick of it. lol suckers! then we made an origional cheer. they said it was fucking gay. haha. go team! i liekd it! but they said it was gay. but we liked it so yeah, i pitched. meh, i guess it was okay. been better, been much worse! then got home. talked to kate, i (heart!) miss kate. she is awesomely awesome! wow. i cant htink of anything esle to say. except im veryyy happy. wow. and my head hurts like madly ...oww..ok.and im done;):)  X N O!! Nessssshahhaha!!  ...FUNNYIEST QUOTES! evr. \"Hey, these escalators are like walking, except quicker!\" some blonde on the escalor..haha and no, not me.! \" Bonjour hfuighfj ghfuidgh uifahngfjknb gfduigh fuihgfja\"-french person \"bvyudhgfd hfuig hfuihg dfghufdig hdfuhg\" -french person \"hfiuhgjkfhgj bhuif hgfjdhg hguf ds\"-french person *French speaking people are having a conversation* \"hey guys they\'re speaking french\"-me \"ohh sweet! BONJOUR!\"-erin \"Bonjour!!\"-me \"Bonjour!\"- everyone else. \"hahah thats so cool! are they from france!?\"..wow  you had to be there!!  \"Stace sOle sTecond\"-Jodie ..haha.insted of \"sTole sEcond\" ..wow what a good day:) goodnight!  sole stecond  '),(14606,3422339,'02,June,2004','2004-06-02 00:00:00','   Tonight was shit. haha im sorry. had practise 6-8. i was picthing. it=sucked MAJORLY! wow... i was actaully mad at myself. cuz comparaed to last week...woa. Then to make it worse, coach was yelling at me. and lecturing us. and i felt like a loser the whole time. Then i got hit with the ball. twice. cept nobody cared..haha..thats nice! one time somebody threw it at my body, then someone hit my hard in the head with the ball in thier glove to get me out. thanks team, i still love ya. needless to say, im in a baddd mooood dude!  But yeah, going to go let out my frustration with a bowl of ice cream. hey, if im in a bad mood, might as well get fat-er while im at it! Science center tomorrow! wow. it would be niice. minus the fact that we HAVE to do some sorta asssssignment. frik. life is out to get me, im still trying to find a good place to hide. even tho i dont want to..woa, im done. ICE CREAM HURRRRR I COME! insanely, .x n o. ONLY good quote today: \"Fill Your hole!\"-Coach (refering to the holes or \"open speaces\" on the field\" \" *Silence* \" \"hahahahahahahahahhaha!!!!\"-me \"hahahahahhahahaha\" erin \"ahahhahahahahahahahahahhahahahahahahahhahahah\"-coach. realizing what she said. \"...zzzz\" other coach. thanks man. \"hehehe\"-team. ..get it? no?, k.  '),(14607,3422339,'01,June,2004','2004-06-01 00:00:00','   heyyoo. Taday was peeeachy good. We had an assembly first period. got to miss some of French:) Erin kept dissing the girl who was \"slightly overwieght\". Erin kept saying she was getting exercize by going up to get awards. meany! haha. i onlii laughed a liitle. Then Mark won award! yay! but Erin yelled out \"MARK YOUR HOT!\"...Thankfully he didnt hear. lol, jokes. So then in Sciencer we had to watch the boring-est movie ever, about the universe and how the sun will explode in 5 billion years. man that sucks. but then: \"what would you do if you only had 3 hours left to live?\"-Ben \"...\" \"Sex!\"- Ben Jenelle and Me! \"haha, that was the first thing that came to mind\" ..lol...so you had to be there. Lunch was kinda yummy. cuz i bought Nachoes:) they were very yummy. i love melted cheese and nachoes! Then went to english hall. and everybody kept taking Mikes shoe. cuz he kept taking them off. and they almost got thrown in the garbage can. twice. damn..so close! English was boring as usual. it always is. and in gym we played doctor dodgeball(goalie!) and soccer/baseball/cricket/bowling...was verii funn! So just got home. bus was free...and my bro is such a freak, cuz hes riding all around town, \"bus hopping\"..you know you really have no life when.. So tonight i have baseball practise...PLEASE NO RAIN!...haha. thanks. its greatly apprisaited.* (never been able to spell that word) and im done X N O nesha  '),(14608,3422339,'01,June,2004','2004-06-01 00:00:00','   Hey guys. Was bored yet again. Decided to answer some of those random blog questions on the profile thing. and you dont kno wwhat im talking about. laugh anyways. i tryed. and i amused myself.:)   You have to dig a hole to China. Where do you start?  The very very top on the CN Tower.   You\'ve written a hit musical! How will you avoid having fame go to your head?   I will build giant walls around just my head that have signs that say: No Entry. (Warning to all Fame)   You forgot your mom\'s birthday! What can you make out of super glue and olive pits?  Super Pits!  ...i have no life:) see yahhhh:)  '),(14609,3422339,'01,June,2004','2004-06-01 00:00:00','   Hello Again. Im Sorta bored. Obviosly. I should be doing my science hmwk. i have a major assignment due tomorrow. Oh well, im almost done. Anyways,  Im talking to Tashina and Daniel (Tash and Dan) on msn. I miss them soooooo much! (In case you dont know! We met last year during the Sevec student exchange...we went to Vancouver!..and thats where they live:()anyways. i miss them. [tears]. and i wanna see them soooooo bad. because they are amazing. Tash shutup! Your pretty! And Dan! man i (luv) ya! I still sleep with the stuffed animal ya won me!! XOX!  Anyways, moving on...I have 0 better to say, so im going to do my homework now!! .x n o.  :you can close your eyes to things you donâ€™t want to see, but you cant close your heart to things you donâ€™t want to feel. [loveless]  '),(14610,3422339,'01,June,2004','2004-06-01 00:00:00','   GoodDay! Today was a happy day! haha wheew...Im not sure why it was so great! we played baseball in gym. was fun...i finally understand science, which is always nice. not that it was hard to understand. i was just away and fell behind. all good now!:) and lunch was great..:) haha kate..wow.! lol i wish.. anyways, just finished dinner, better start my scince hw...big project due tomorrow...and i think im going to a hockey game later too..(Phill\'s)... cya! -later  \"i dont miss you. i miss who i thought you were\" \"No matter how hard you hate him for what he did to you--you\'ll never stop lovin him.\"  '),(14611,3422339,'28,July,2004','2004-07-28 00:00:00','   Heyy  So busy lately.  Today went to movies with Jess[puddles] Jodi[crip] Lisa[lilo] Angelica[angel] Vicki[stitch].  Saw A Cinderella Story. Was the best....so close to tears! was the cutest. CMM is definatly the hottest ever....wow. like seirously. wooooooooooooooooooooooooow.   Came home, played basketball with thomas. was beating him 8-7. he won 11-10. came home, got changed, uniform. biked to practise.   Was the best practise. Coaches batted and we were in field. Brianne, Adrian and his friend Matt.  Was the bestest. They pulled up both thier cars and put the radio up, so we played baseball with music. it was hott.   Biked home, got changed, all that. went to Lisa\'s house for a partayy! we hung out, and played \"i never...\" and it was just fun. we talked soooo much. ...oh yea..congrats Staci!:)  Went outside, to the park. realized after being there for 1 minute there was a man 10 feet behind us on a bench. ran.   Back to Lisa\'s....chilled more. then i left...:(  So now im home and bored....sweat.  Later.  love vanessa. \'pimpette/squale\'  '),(14612,3422339,'23,July,2004','2004-07-23 00:00:00','   OLA!   Had a realli good past few days..   Wednesday my baseball practise got cancelled (we realli needed it)! So my mom and i went to the mall. i returned a shirt i got and exchanged it for a smaller one, and got a new white denim skirt. its hot. and i got 2 more holes in my ears!!!! it hurt onli a little, and it looks so sweeet!    &nbsp;  Thursday was like the best day..ever. I had a baseball game at night, and i pitched and got 3 strikeouts in a row in one inning...3up 3 down:) and then i hit a double and it was hot. and after that amy showed up:)&nbsp;haha it was my best game. lol it started raining in like the 4th/5th inning! so we all got soaked. it was so nie tho cuz the sun was out and we were all standing out in the field, and it was like a millon degres and it was so hot, and it just started raining. i loved it. so we took a 10 minute break and went under this big tree with our whole team, and they all sung happy birthday to lisa and me. :) Vicki was also there after being gone for like ever. she brought us all postcards and key chains and she brought me these candy on a stick things, which i love. so then we started playing again. we ended up losing but it was the ump\'s fault seriously.&nbsp; he always ump the team we played no matter what diamond its at just cuz his step daughter is on that team, and they always win and we get shitty calls. but it was the best game still, so fun. and i won MVP:) Thanks guys;) So then at one point Adrian came to our bench and told me and vicki and megan to cheer like \"yeayyeah\"-it sounded really gangster, and it was so funny, cuz megan actually did it, it was funny. we laughed.   So after the game erin thomas and cody and me went swimming at erins. it was the funnest. we played \"pool games\". haha, awwh good times. me and erin went on thier shoulders! and we tryed to push each other off. haha and then we played marco polo. except i got caught cuz thomas was holding me out in front of him:   \"marco\"-erin   \"polo\"-thomas, with me in front of him so he couldnt get caught   \"cheeter!\"-me   \"whos marco cheeter!?\" haha   So we got out at 10 30 ish. and tom and cody walked me home.   Then i go thome, 15 minutes late. shittttt. but surpirsingly phill asked my mom if me and him could go get slushies at 11 30 and she let us...hm? wierd    &nbsp;  So today i went to st francis and met amy there and then we went to Deerpath park played grounder and hung out with this hot mexican guy who she met.. aww they\'re so cuuute togther!! Marco..no...Maresio! it was soo funn!! hes the sweeetest guy, hes so cute, and awesome:)   then me and amy came back here and talked for an hour eating ice cream sandwiches.   So now im gonna go pack cuz im going to my trailor in Orrilia for the weekend.    &nbsp;  Ciao. xoxox   love vanessa.   Adious!    &nbsp;  '),(14613,3422339,'20,July,2004','2004-07-20 00:00:00','   Heyyy There!!    Im 15! wooo!!!    Had a really nice day today.    Woke up at 8. won this birthday thing on the radio. score.     &nbsp;   Chillied all by myself in the morning.    Alyssa came by on her way home from camp around 12 30. brought me a milkshake:) and a neato card. thx hunnnny. i luved it. she just stayed briefly and we watched the sitcom movie we made last year, which was SO funny. anyways     &nbsp;   Then my gramma took me and Phill out for lunch at the usual, Jose\'s Noodle factory. My Favvy. Was sooooooooo goood:) But our waiter was even yummyer than my capeletti. mm mm good. got to spin the birthday wheel, which has been my dream since birthday wheels were invented. anyways.     &nbsp;   Came home and got a few thiings ready for my BBQ.     &nbsp;   Kate came over at 5. i attemped 4 times to hang up ballons, but only one green ballon worked (yes, it did, yet tim broke it when we were waiting outside later...:()     &nbsp;   Erby came then and so did erin. so we went in the basement. and erin was facinated by my webcam. haha. then tim came over.    We went in thr backyard to chill on my dancefloor deck. its so neato. listened to some music. watched my pond. haha.    Ate BBQ dinner. i loved it...yummmmy.    The we had ice cream cake...it was soo good. at one point we were gettin all hyper or whatever, and erin smashed her head off the deck thing and we laughed for 10 minutes staright about everything. it was in-tents. then i got some pressentS:)    tim gavce my the best card ive ever seen...thanks...sexy!! 101!..what LOL! hahah good times. and then of course got some cute ones too, from the girlies. thanks guys:) tear. got a kool shirt:) earrings...and nice and easy...lol...see \'quotes\'.[hard..]     &nbsp;   Then we decided to dance. and i got out the video camera. and we made this little movie thing. BEST MOMMENT EVER:     &nbsp;   Erin was dancing around, and shook her bott-ay for the camera.    She finished, and i wanted to. so i said \"look, im shakin my botttyy!\"    so i danced and as i am, \'shakin ma thang\'    Hiii there vanessa. happy birthday..what are YOu doing!?    said my GRAMMA WATCHIN ME FROM INSIDE THE WINDOW. lmao ohh we lauighed sooo hard. too bad we ran outta film for that one.     &nbsp;   so we watched that. then we got food and watched part of Pearl harbour. soooo sad movie...haha look at the quotes.[anti-thrust]    SO THEN. THE&nbsp;HIGHLIGHT OF MY DAY. Phill comes downstairs and says i have a phone call. i pick it up and someone starts singing me happy birthday:) its from PEI! it was emily. i almost cryed. well wait, i did cry:) it made me soo happy. so i talked to her for awhile, then i talked to ian! wow. that really made my day. thanks guys. love yas.   &nbsp;so then we decided to wait outside for everyones ppl. erby left first.     &nbsp;   then we were sittin on one of those green electrical boxes. tim was throwing rocks on it and we were talking/laughing loudly.    then a lady comes outta her house, and talks ..in a russian accent:    You kids are way too loud. next time u loud i calling the cops! its past 11 and everybody is sleeping! you should go sleeping! now!     &nbsp;   hahahah i laughed. and they did too. then we were quiet. at this pouint tim popped my beloved green ballon.:(    so then they left, andi cleaned up , here i am.     &nbsp;   10 more minjtes of birthday. tear. i hate endings.     &nbsp;   ciao    love vanessa.     &nbsp;   quotes:    *Erblina give sme nice and easy hair dye*    \"is that nice and easy!?\"-mom    \"..no, hard and fast\"-erin    \"ERIN!\"-mom    hahahahahahha     &nbsp;   *people havin sex on movie.*    \'THEYre not even moving!! gosh they\'re just lying there, geez\"-tim    *girl gets prenant*    \'well i guess that shows&nbsp; u can get pregnant during anti-thrusting sex\'-me   ILOVE YOU GUYS!!!:):):)[hearts]  '),(14614,3422339,'19,July,2004','2004-07-19 00:00:00','   Heyyyyyy!!!   Hows it rollin\'?   Today was a really good day.    &nbsp;  Got woke up at 6:30 by garbage truck #1.   Fell back asleep.   Got woke up at 8:30 by garbage truch #2.   So showered and all.   Adam R. came over at 9:30ish looking for Phill. He came in Cuz Phill was in the shower. so we chilled;) lol.   We just hung out for like 3 hours. watched TV. kicked his ass, literally like a millon times. Dillon came over, and the guys had lunch.   Then they went biking and i hun gout with Adam....(PS....hes not Phill\'s age. no worries. hes&nbsp;older)   So then the guys left. and i kinda hung around here, did some good olde laundry. woooo. fun stuff. Then mommy came to pick me up at 3:30 for work. took me to the hospital for workin.   I got a pretzle when i was there. cheese flavoured....i tryed it and it was soo good.   \"there\'s this girl at my work who reminds me of you\"-my \'co-worker\' talking about his other job.   \"..oh really?\"-me   \"..yeah, your hieght, skinny, curvy..\"-him   \".....mhm...\"-me. scared. haha thanks bud.    &nbsp;  So it was an alright shift. only 1 more:) score.    &nbsp;  came home, daddy dropped me off. went to the Boathouse with my mom and Philly for ice cream. met my gramma and auntie and cousins julai and natalie there.&nbsp; First we went and sat down for presents. Phill and Julia went to scare geese and look in the water. I got 50$ from gramma:)....and.... A CELL PHONE FROM MY MOM AND AUNT AND UNCLE!!! plus 30$ calling card thingy!! yay!!!!!:):):):):):):):):)   Its so neat, cuz its the only one like that in Canada cuz its a new model. its \'champagne\' coloured and it has the coolest ringtones and pic things!! its so high tech.. [hearts].   So then got ice cream...came home...called erin and erby...set up my phone thing...and im gettin ready for the part tomorrow. shall be good times. im so happy.   ciao!   love forver vanessa.   ....thats iight. forever.    &nbsp;  ~KISSES AND HUGS~  '),(14615,3422339,'18,July,2004','2004-07-18 00:00:00','   Heyy There    Havent blogged in awhile to any of you keeping up with my crazy life.    &nbsp;Friday I went to Erin\'s house for a swimm. Ended up having some \"sneaky peak\"...tipsey. Haha i fell on the stairs and Erin fell on top of me. Sweet. Then we swam. Matt g came over. cuz we were naked!!! (loonngg story) and then he left. then cody came over. and then we all swam. and then i left and went home. then later on Erin came and picked me up and i went with her Family to Cody\'s football game. Go Bears! ....they sorta lost big time (47-7) wooo but it was fun. lots of guys...:) and cheerleaders that looked like they were like 12 years old. haha...losers*. Si after the game we got swarmed by 75897328657 misquitoes. then we drove home. went to erin\'s. her and cody spooned, and i went on msn. then he left and We watched Sweet Home Alabama. which is sooooo cuute.    \"Why would you wanna marry me?\"  \"So i can kiss you anytime i want\"    Awwwwwwe!     &nbsp;   So then we fell asleep. and then Saturday morning i got picked up around 11:07. went home, got ready, and had to go babysit at 11:25. So went to my aunts, and babysat Julia allllllllllllllllllll day. 11:45 a.m.-12:45 am. Yayy...We went to the store to get chocolate popsicles at one point. she ended up wearing it alll over. shirt shorts hair arms hands face legs everything. woo funn. We also ordered pizza. except the guy thought i said \"55\" mont street, insted of \"65\" so we saw a guy gettin pizza a few houses down...and&nbsp;thought&nbsp;it was wierd. then they called and i told them the house number. so he walked down&nbsp;the street and we got our pizza. except he was checkingme&nbsp;out and givin me the elevator look and stood and watched me walk away...CREEP!    &nbsp;   Today i slept. it was nice. then i finally plannedmy party. BBQ. bought all that stuff.    it was nasty tho. we went to get hamburgers from this meat store, and we walk in and the guy is about to slice a cows leg. ugh, im so NOT eating those burgers.     &nbsp;   So now i just had Cheerios for dinner. cuz i dont want pork:)     &nbsp;   Later.    love vanessa.     &nbsp;   Oh yah. I found out when i go to Cuba (1 month) We get to go on this cruise boat thingy for one day and they drop us off at an island and we get to go swimming with dolphins. which is my dream. :) im so happi.     &nbsp;  '),(14616,3422339,'15,July,2004','2004-07-15 00:00:00','   Good Morning..ahh its like 12:15. i woke up today, yet again, to Dillon calling. frik i swear he gets nosleep. so him and phill are gone biking, i have the house to myself...and im bored. i watched Dawson\'s Creek for awhile this morning. i almost cried, it was a good one.:):(  So Yesterday i babysat my cousins Natalie(..11 months) and Julia(3 years). I took them for a Lonnnnnnng walk. Then we went to thier house, ate and played. yayyy. then i took them to the park. Then i told Julia it was time to leave and she started crying and everybody looked at me. ouchhh. im sorry.  So Last night, me and tim went to Erblina\'s. we swam in her pool, in the rain!! it was just us, plus this annoying kid trying to ruin my life. it was sooooo fun. then we went in the hot tub, which was the koolest, cuz u couldnt move or it would overflow setting off fire alarms. then the sauna. it was sweeeeet. then my mom picked us up and took us to tim\'s house, where we watched Pirates of the Carribean. i had never seen that movie. it was hot stuff. then erblina left and me and tim went in the basement with his bro and robert and austin. i tryed to play a song on the guitar, but i sucked. so all the guys played guitar, haha austin had the acoustic. so then we all left, and i walked home.   so thats my life. now im bored like mad. bored to pieces. and im done.  love vanessa.  Erblina has the coolest belt ever!  *haha Erblina and tim are coming to my game tonight. and they\'re wearing shirts that say \"Go Vanessa\". they rock my baseball socks off.*  Quotes: \"Hey vanessa, lets go downstairs to see austin..powers\"  '),(14617,3422339,'13,July,2004','2004-07-13 00:00:00','   Heyy. i am bored. Im grounded for coming home 10 minutes late last night. yay. Today was soooooo hotttttttt. like burning.  Got in trouble for not taking down the laundry. and then yelled at mom, and she told me to shutup.  \"okay well i dont wanna go out for hotdogs, who decided that!?\"-me \"fine then you\'re not coming, but you\'re grounded for the night\"-mom \"well if i came with you would i still be grounded??\"-me \"yes\"-mom \"...frik\"-me So i stayed home. ate dinner in my backyard by our pond. was eating and realized we have 3 dead fish in there...how nice. isnt that special? ha. Then i talked to daddy  on the phone. then erin called. and we talked for awhile. all-star baseball game is on TV...gonna go now. Latarr. love vanessa.  Quotes: *talking about a guys mustache* \"Its about time he shaved it\"-Erin I heard* \"At bedtime he shaves it!?\"-me \"hahahahahhaha\"  '),(14618,3422339,'12,July,2004','2004-07-12 00:00:00','   ahhh. hello. This afternoon I went over to Erins house with Kate and Em for a little swimm. It was pretty funn! The water was so warm. i reaaaaaaaaaaaaaalllly wanna pool. seriously, now now now. erin\'s pool used to be mine. and my hot nieghbours...but it got passed around..frikkin hill in the back yard!!!  anyways, we swam, and got funny hair..haha remmeber? lol. we had some fun. dried off..took some pictures, which were sooo hottt. and then erin straightened her hair, which looked encredibly beautiful, wow. and em did kates makeup..and then i left for work. fruk   At work, the guy i work with told me to eat the last 2 peices of cake (..gladly!) and then he took out a frozen one and put in on the thing to sell. minus the part where it has to de-thaw cuz he just took it outta the freezer...wow hes a smart one.  Came home to my dads. worked out for like half an hour. went for  a walk for an hour with daddy and kim. came home to my moms, went for a walk with erin, but ended up going to Cody\'s. we hung out there for awhile. erin yelled at cody. then they did thier thing, and i stodd there...hmm..awkward!!!!  went to codys back yard. his dad came out, hes soo frukkin funny its so..funny. and he pretends to be gay...haha i love it. and erin kissed me on the cheek too....na na na na boo boo. ha. im done.  came home 10 mins late and got major trouble........frikkkkkkk i hate my life.  later. vanessa  \"i cant form sentences\"-kate \"whos courus spentences?\"-me? haha.  '),(14619,3422339,'11,July,2004','2004-07-11 00:00:00','   Hellllllooooooooo.! I hate misquitoes alot. i got bit 36 times by them this weekend. 36. frik, then a \'horse fly\'. they funny thing is that i was wearing pants, yet the bites are on my thighs...perverted misquitoes went up my pants! Ugh!  SO this morning, i had to go to the hospital for a test. i had to do this thing for asthma, where you basicly give the air tube a blow job. it was sorta disturbing..lol but it was funny. i really enjoyed that one...and i still dont know if i have asthma...more testing! yay *sarcasm*  This afternoon Em, Kate, Erin and I are getting together, ahh just like the olde times. we migth be having a waterfight! yay, or swimming, thatd be niiice..hmmm.  Then at 4 i have to work...fack. i dont wanna, almost done though, so i am happy.  Well i better go,  Later. love vanessa.  Quotes this morning:  *breathing questionnaire* 7) Have you drank alcohol in the past 4 hours \"....\"-me..thinking about it. and going to check the yes box. \"Vanessa! *dirty look*\"-mom \"...joking!\"-me  *Dillon comes over and him and Phill are about to leave for biking* \"..do you have your key\"-me \"yeah...cya\"-phill \".alright, bye\"-me \"I hate you.\"-dillon \"...good morning to you too sunshine.\"-me  '),(14620,3422339,'11,July,2004','2004-07-11 00:00:00','   wow. i really hate hate. its depressing isnt it? I dont like how when somebody you care about so much is trying to understand how u feel, and you dont let them. you dont bother explaining yourself because your afraid it will only make things worse. it also really sucks how you have to live on, after you feel like you don\'t know what your living for. its like trying to find the light in a dark room youve never been in. thats what it feels like. and when you do finally think you have found the light, and your so happy to finally see whats going to light up your life, theres nobody else in the room.  '),(14621,3422339,'07,July,2004','2004-07-07 00:00:00','   I Hate how you smile when i want you to. I Hate how your jokes make me laugh. I Hate how you give me butterflies. I Hate how i get excited to see you. I Hate how your touch makes me feel. I Hate how your kisses make me so happy. I Hate how you look in my eyes and make everything in my world alright. I Hate how i want to be with you. I Hate how much your hugs mean to me. I Hate how i love you.  '),(14622,3422339,'07,July,2004','2004-07-07 00:00:00','   Ah i havent blogged for a while. miss me? anyways. hmm..thursday i had a baseball game. against Gatto. We only had 8 players, so it was a little bit of a struggle you could say. at one point we were losing \'20-4\' (Reason being, we hadent played/pracised in 2 weeks) but then we came back and only lost 19-11...haha yeah, im still trying to figure that one out? hm. anyways, megan got MVP for an awesome catch she made in the field when i was pitching, it was such a good ctach, she got mid season MVP (VMP! lol megan) and got this baseball phone thingy. \"and it works!\"...lol. So came home, and ended up going for a little walk...it was nice;):)  On Friday i just sorta cleaned...and attemped to burn a CD...which failed miserably 3 times. Sorry Josh! Ended up getting picked up by my dad at 5:47 to go to Orrillia for the weekend, to our trailor! Amy and i watched her grad video and part of her SEVEC video on the way up, cuz we have one of those VCR/TV things in her moms van, which is very handy. so we watched that, and by the time it was over, we were there..lol 3 frikkin hours later..cuz there was tons if traffic/accidents. frik.  So when we got there, it wasnt really dark yet so me and Amy went for a walk around the park. it was my first time being there since last summer. we got a new golf course, and it looked cleaner! haha, so we decided to walk down to the lake..i think its Lake Couchiching? Hm, anyways it was so funny, because we were walking down there, and there would be these huge packs of misquitoes every 5 steps, and one time me and amy both covered our faces and waved our arms around as we ran threw them...and we Ran into each other!!!! Ha! it was soooo funnnnyyy!!!!!!! BLONDES! So we went and sat on this rock right on the lake for like, ever. we talked so much, about guys, high school, and basicly evetyhing. i love how much i can talk to her:) so we just sorta stayed there, until it was dark: \"...heyy its dark?!?\"-me so then we went back to our trailor...ATE! and hung out. we ended up going to bed in the Florida room around..12:18? or thats the last i heard....cuz Kim told us to be quiet and she said it was like 2...but it wasnt. hm. anyways  Woke up to Uncle Bob talking to amy about history and politics the next morning, i could tell amy was reallll enjoying it!! lmao. so i put my walkman on. eventually got up and showered, and went shopping. Bought: SWEET pink blouse thingy? u know those cool ones? Jean Capris from Bluenotes, earings, ankelet, pink campuscrew tshirt, jewelery from clairs...and yeah., i was happy. Went back to the trailor and Cindy and Pat were there (kims friends) as well as Deb and harry (thier trailor is behind ours) so they all hung out, and me and amy went swimming. *me and amy are walked to the pool* \"yes i know im hot\"-some random guy passing us \"hmm what the?\"-us So we swam for awhile. it was warm. supposely a few hot guys came in, but without my glasses i couldnt see!! fruk. Dried off and went back.. LAYYED ON THE HAMMOCK!! WHICH IS THE BEST THING IN THE ENTIRE WORLD. umm had dinner, watched 10 Things I Hate About You...cute movie. Then daddy and Phill went fishing. me and amy went to see what they caught[nothing]..then we went to the park. We were on the swings, then this guy who we had seen like sooooo many times came and sat there with us. there was a bunch of ppl with him too. he asked us how we were doing, and we talked about whre we were from and all that kinda stuff. he was John, from Woodbridge? or something like that. he was kool. then we said bye and walked home. the adults were having campfire..and we went inside and listened to music and ATE!!! ahh white powder donuts own me like nothing else...yumm. and nibs too. i swear, i learned i could live off those donuts, nibs and bagel bites for the rest of my life and be totally satisfied. mmmm.  So Today woke up to phills boxers in my face. Thank you very much. got ready and all that jazz. went to the lake with amy. saw dad and phill fishing. saw tadpoles... We also saw John agian. siad hello. went back to the trailor packed up and left.:  *driving outta the park* \"hey vanessa are you gonna watch a movie on the way home\"-phill \"...............vanessa?\" \'*Amy and I stare at John as we pass him in the van* \"bye John!!\"-us *waves* \"cya!\"-john \".............\" \"Phill, you can ask them your question agian now. thier distraction has passed us\"-dad  So we watched a Walk to remeber on the way home...thats the best movie. got home around like 4ish? and now im back at moms. apparently i have a big lecture coming on tonight. fruk...cant wait..not. latar. ness.  Quotes: *i ssit right up in bed* \"guys i need money to buy my ticket!\"-me \"what ticket\"-amy \"my ticket for the swings!! \"-me \"....?\"-amy \"ugh i need money for my ticket\"me *i fall back in bed*  LMAO! APPAREBTLY THAT HAPPENED..AND THE FUNNY THING IS..I DONT REMEBER ANY OF IT! IT WAS IN MY SLEEP! I DIDNT WAKE UP! haha.. \"hey vanessa, isnt that the guy u like\"?-phill, every single time we passed a guy, and john, this weekend...and of course he said it loud enof so even that person could hear *we\'re all in the pool* \"hey Vanessa! isnt that the guy u like!?\"-phill. when John walked into the pool. \"Hey Amy! Theres the guy u like!\"-phill, as John passed us walking... and you get my point \"Dissed and dissmissed\"-phill, every 5 seconds.  *amy and i are sitting on the swings* \"i cant wait to go to Cuba. have you found out where you;\'re going this summer yet?\"-me \"no i wanna go to P.E.I.\"-amy \"why?\"-me \"so i can get out of Canada\"-amy Buh-Londe! lmao \"Score!\" \"YES!\"  \'i want youuu to want me. i neeeeed youu to need me. i loveeee you to love me\'  '),(14623,3422339,'07,July,2004','2004-07-07 00:00:00','   Heeeeeeeyy. So today was fun. went to Lana\'s:) hehe, on the way over there, it was pouring rain and there were these naked kids on my street! like they were 6-7 years old? and there was a boy and a girl? brother and sister? and they were playing some sorta of wierd hopscotche game. anyhow, they scared me. cuz they were naked. so i went to the other side of the road. yuk, what is it with you people?! CLOTH YOUR CHILDREN!  Just got home from baseball. i was supposed to have a practise tonight, but my coach said it looked like it was gonna rain, so cancelled. then sahe called me at 5:56 and told me she wantd me to be a call-up pitcher in the midget league( like 18 yr olds.) and i was like \"....yay\". so the game started at 6:00. so we rushed over to the diamond, and played some ball. i pitched, and well. hm, they were strikes for sure, but too slow, so they hit off me and the outfield was kept busy. frik i need to speeeeeeeeed up. then it started to rain HARD, so me and the girls from my original team who were also call-ups went to this little park thing. and jay came with a clear garabge bag over her head that her coach had given her. i laughed at her, then coach came and literally Put it on my body. almost straglling me.  \" Okay, quikley...CAN YOU BREATHE!?\"-coach \"....mhmm\"-me  it was fun tho. we looked so kooooool. then the game started again. and then i hit once, between short and 2nd. yay, got home. then it started to rain. and the wierdo ump saw lightening, so the game was cancelled. frikker, it didnt even count...damn.  so yeah, going for a shower now. latar. xx nn oo vaness  Quooote: *i was playing 3rd, and there is a girl on 3rd base talking with her coach about stealing home* \"okay now, 1 out, if its a pop-fly, hold up here, then run hard, ifs its a grounder watch first then run home, but-\"-coach \"Shutup. I\'m PMSing. just tell me when to fuckin run.\"-chick  haha, i tryed to hide my laughing. this chick was a tank.  '),(14624,3422339,'06,July,2004','2004-07-06 00:00:00','   Good Morning  Im so tired...but i have been sleeping so much. hmm...? Anyways, just woke up. About Wonderland yesterday, Phill  and Sean (his friend) came too. I did all the Rollarcoatsers they did, i was sooo proud. Exact opposite of last year, when i was the one on the ground watching them on every single one. no joke. Ive been missing out. haha  So first we did that new Tomb Raider one. where you lie down the whole time! it was kinda different, but fun. sorta like the Fly. Then they convinced me to do Cyclone! Wow that ride is one of my favourites cuz its sooo fun! i screamed tho...then Sledgehammmer, thats my other favourite, cuz its crazy fun...Then we went over to Cliffhanger, which was probably the freakiest, because your upside down most of the time! I had my eye closed for most of it, regretably. and we got wet. So had lunch. Went to the waterpark...for like an hour. it was Very warm. we were in the lazy river and they guys pushed me off my tube, thaaaaaaaaaanks. Came back out, went on this standup rollarcoaster, which was so in-tents. but i liked it. Hmm...then we watched that Spongebob thingy, with the Pickle. i laughed, it was cute. ahh i cant remeber what else we did...i think the guys had dinner with mom, we went on Top Gun which was sooo rough, but reallllly fun. and Dragon Fire, sweeat..and the Vortex. twice..then i felt sick.  Then just as we were leaving, mom phill and sean all playing that game where u wack the moles as they pop up and try to win something. Sean lost. Phill lost. mom lost. wooo failiures!! haha so then we left. i fell asleep on the way home, and thats about it!  *OH! the best part, was seeing a little boy with a MULLET! it was soooo looooooooooooooooooooong! the boys talked about it for the rest of the dayy, cuz it was so funny!! ahh good times.  later  *walk by dog kennel* \"there ya go phill, see ya later! go in there\"-Sean \"haha shutup...hey vanessa, lets go look for the elephant one for you\'-phill  '),(14625,3422339,'06,July,2004','2004-07-06 00:00:00','   maybe i should hate you for this, never really did ever quite get that far  '),(14626,3422339,'04,July,2004','2004-07-04 00:00:00','   Good niiight. Had the bestest day today. went to wonderland. did rollarcoasters, like real ones, for like the first time. I\'ll \"blogg\" agian about it another time, but i just had to get one thing out.  I really like this person. its funny how i thought about him all day, i wish i could be everything he wants...i just dont know what to do, and if i should even go there.  love sucks big time...:P  and im out  x n o. coaster babe! vanessa  '),(14627,3422339,'04,July,2004','2004-07-04 00:00:00','   Hello Hello!  Yesterday me and Amy made diinner=Tacos. They were delisious. It was funny though, 2 blondes making Tacos eh amy? haha wheres the beef!?  Last night  i babysat these kids who lives behind me. They are so cuuute. I love them, we played games and watched Spiderman. it was great. yay,. Came home at like 2a.m....and sleep.  Today i woke up around 11:30 a.m. yay. sleep. Lazed around, woke up my brother to say happy birthday!! hes 12 today. wooo. i remeber when i was 12..yuk.lol So then me and him watched \"My Boss\'s Daughter\" with Ashton Kutcher( sooo hot) and Tara Ried (sooo pretty). it was a randomly funny movie. at some points i wanted to leave, but it had a happy/kinky ending, so it was all good. Then i came on msn....finally dragged my ass in the shower at like 2 :00...and cleaned my room sorta, and now im soooo tired...but yeah, i have to finish cleaning, so im out  Family is coming over for Pizza and Cake tonight, for Phill\'s borthday. yay.. Later vanessa  My Goddess Name: Vivaciously Alluring Notable who Encourages Seriously Serene Admiration  Randomly Awesome Website: www.toothpastefordinner.com  Quotes: *Luca falls asleep during Spiderman* \"..Luca..you ready to go upstairs to bed?\"-me \"...no\"-Luca  Poem I Wrote for Phill\'s Bday:  Happy Birthday Philly! On your birthday, When you turn 12 years old, I hope you are happy, And you donâ€™t have a cold. Eat lots of cake, Have pizza with that, Make sure you donâ€™t have too much, Or else youâ€™ll be fat. So have a good day, Make it the best, And when you get tired, Then go have a rest. :)   Love your sistar, !~.:â€¦Vanessaâ€¦:.~!  Haha! Nice eh? haha chow.  '),(14628,3422339,'03,July,2004','2004-07-03 00:00:00','   good day.  Friday: Got up around 7:45ish. Showered and all that. Got picked up at 9...picked up Erblina, then drove to Rockwood with Tim, Kyle, Em and Erblina.   We were the first ones on the beach. it was sooooooo nice. We srtaed by tanning of course. Then i went in the water first! woo. then Kyle came in. Then Me and Him picked up Tim and dragged him into the water in his tshirt and shorts and dunked him. haha, he wasnt too pleased with us, but it was hillarious. Then tamsen came. A little while later,  we took a \"hike\" through this little trail rocky path thing. came to a cliff. set up our towels there and ate lunch, while the guys cliff jumped. i was still to chicken, but next time for sure. after that we walked to the waterfall, and the mill, which is really pretty. then we went back to the beach. swam, got freezies and ice cream, tanned more. i finally tanned for once, im so happy, but my whole back and the back of my legs feel like they\'re burnt. but im tanned so its allll good:). Tamsen barried me in the sand, and tim and kyle helped, except they got so much sand in my eyes/mouth/lips/ears/bikini. thanks. then tamsen gave me a sand penis and \"genital area\"...i loved it. she took a picture.then we all went in the water one more time. Em almost drowned me. almost, i couldnt breathe, i was honestly scared. then i got her back..haha!! i win. dryed off and tanned more, packed up and walked to the gas station where we were meeting tims dad to go home.   came home and showered yes! had a BBQ for dinner with my whole family. Then i walked to Zehres with my mom to get my Grade 9 pics. they are awesome, for once my camera worked:):):). i love it.  Came home and watched \"A Walk to Remember\" with Amy. We both cryed...it was the sweetest movie. i think it is my favourite i love it soooooooo much, i could watch it everyday. its so sweet, and Shane West is the hot-test evR! wow...i love that movie.  Then i came on msn. chat-ted it up...found out me and mike\'s dads both grew up in sault Ste Marie and they know each other, kool stuff eh...got tired and got some sleep, because i love sleep.  Woke up today around 11 15:) yay! i slept in! Ate strawberries that my sister picked. yummy yummy. showered and got ready  Went to the mall with my mom and phill. i got new funky brown flowery sandals, sunglasses, makeup, and razors! yes! haha...Then went to the Boathouse for some ice cream on the way home.  So now im going to gooooo...its sooo nice outside. warm...yay.  :) xoxo :) Vanessa.DiPasquale.  *Ashley tosses phone onto couch and walks outside after talking on it* \"....hello?\"-amy haha! there was nobody on the phone! ashley was just putting it back. hahaha.  '),(14629,3422339,'01,July,2004','2004-07-01 00:00:00','   Heyloo  Today was a good day. it was nice. Came Home from the Sleepover. Showered and all that jazz. got ready. Attempted to tan this afternoon. for about 5 minutes, haha. then i played catch with my mom. yes.got ready to go to the fireworks.   Erblina came to pick me up at 5 15 and we drove to Canadian Tire, where Kate was waiting with mark and \"them\". we walked to Riverside. Had to pay the family rate of $5 to get in. which mark payed. thanks.and plus we got stcikers, which is the coolest of cool. so we walked around the park for the evening. Saw a couple hotties lol Erblina. red shirt[heart]. Then we walked over the bridge thingy and saw all these DIRTY KIDS playing in the water, like where it was going down and it was like slimey. the kids were so yukki. the guys even threw a golfball in and kept making them get it. and it was sorta yukki.   Kate bought me a hot dog. yum it was good. with no tax..! so we ate, saw chrsitian, walked around some more. Kate and Sylvia went on the ferris wheel thingyer. me and erblina walked around, then watched. i wanted to do it. it would have been so romantic...awwe. Ugh i have the wost memory i swear...ummm i bought a bunch of freezies:) and water and a drumstick. so we then went to where everyone set up thier blankets. it was soo pretty.  I SAW MORGAN! YES AFTER LIKE 2 YEARS!!! woo. lol she still looks the same. hugs. awww morgan it was soo nice to see you!! morgan is the coolest. almost as cool as stickers. well, maybe just as kool.  so we hung out for a while. took pictures!! listened to old music stuff. ew. it wasnt the greatest... lol we all..like all 20 of us just sorta chilled. it was soo nice. sylvia, sofija, kate, morgan, erblina, anna, more girls, all the guys...it was so sweet.  So finally the fireworks started. they played to music which was one of the prettiest things ever. it was sooo nice. it could have ever been romantic. COULD. we all lay down. it was cute. the fireworks were awesome. one ever looked like a butterfly. i liked it. they played a bunch of songs. then the last one, the \'finale\' was Oh Canada. it was so awesome.  When Oh Canada was done, it was like, mad Chaos. like wow it was insane to find our rides...it was soo many people!!  anyways, it was a great night, so nice, and i highly enjoyed it. minus the romance of couse.  ciao. VaNesssAA  Quotes: *Oh Canada stops, the firework finale is over, and people start standing up* \"..........................oh!....is it over???\"-Erblina haha i laughed   There he goes. There he goes agian.  racing through my brain and  i just cant contain this feeling that remains.  '),(14630,3422339,'01,July,2004','2004-07-01 00:00:00','   Hey there. just got home from a sleepover sooo i am tired. but amy was desperate to read my blog..here it goesss.. Wednesday: Got up early...did nothing...did laundry...biked to my dads. me and amy watched part of her SEVEC video. was fun. the side of my head was in it. yay!  Then me and Erin went swimming at matt g\'s house. it was fun stuff. sorta cold. but they splashed me...lol...so we tryed to run and dance in the pool...which was sorta fun...just kinda hung out. me and erin flashed him, but he just kept missing it!! lol!! So then we had to get out, so we went to Erin\'s. we talked and played basketball. i pantsed or skirted? erin. lol it was so funny...haha ohh u had to be there. i kicked matt\'s ass in basketball. like 12-4. rematch eh.   Came home and went to a call-up game with erin. for her sister\'s team. was fun. we won. scored 2 runs each. yay...!  Came home. got changed and went to jess and jodi\'s for a sleepover. we had fun. dvd player was broken, so no movie. we played this game tho, where u say something you\'ve never done, and if someone has they had to eat a chip. lol it was nice. then we just sorta hung out...made really funny prank phone calls. one of them, i actually had a 5 minute conversation with somebody i didnt know!! wow itw as awesome. his name was andy. and mine was melissa. hahaha omg it was so awesome. everybody at our sleepover was staring at me and gettin me to ask him stuff. it was soo funny. then i hung up...tried to call back but it didnt work. frik. he was a cool guy.  At 1a.m. we snuck out. walked down to Mac\'s. it was soo warm outside. i bought candy, and Jes bought us sparklers. so we used them walking down paisely road. wow it was awesome. went back to thier house. watched tv, fell asleep around 2:30ish after friends was over.  Woke up this morning around 10 40...and yeah. came home...now im going foir a shower and to EAT! YeS!!  chow. vanessa  Quotes: \"Hi is Mike there?\"-me \"...no\"-person on the other end of the conversation oh the phone \"are you sure? No Mike? Mike Hunt?\"-me \"no, not here\"-person \"oh thats too bad. dont u like Mike Hunt!?\"-me lol do u get it? Mike Hunt? say it out loud...lol!  *erins wearing a skirt* \"hey i see your \'haha\' \"-matt every  3 seconds.  \"Where is your dream place to have sex?\"-lisa \"Beach, sand, hottub, pool, public beach\"-us \"An old phone booth!\"-lisa. \"you, me. phone booth. 3 minutes\" \"Me, you. hottub. 3 minutes!!\" 3 minutes!! haha  \"i wanna have a little girl.\"- me \"me too. that way i can scare her boyfriends\"-matt \"...oh you can just have a gay son\"-me.  '),(14631,3422339,'11,August,2004','2004-08-11 00:00:00','   heyy.   Had a really kool night.   went for a jogg with em at 8 45ish. adam stood us up!! ahh!! we went for 20 minutes ish. 2 KM. it was awesome, it was hott.   went to her house and changed then went to tim\'s house with him and kyle and em. watched tv in the basement. then when em went into the bathroom, she came out and sat on me and so did tim. i=couldnt breathe. tim gave me a 10 minute charlie horse!! owee! and they tryed to  pull down my pants.. in the process seeing me \"panties\"..hehe:) then i like sat on tim for awhile. very amusing/comfy!   Then tim and kyle walked me partway home, so that i could still have the plessure of bring raped. thanks guys:)   anyways, was a good night:) ahh i wont see emily for 2 weeks!! ahh, ill die. well ill be in cuba for one of them. :) yay!   Tomorrow=baseball game!! you really should come ya know:) hehe, impitchin 3 innings. score.   and im outtie like a belly button.      love vanessa.   ps: tims sexy.  '),(14632,3422339,'10,August,2004','2004-08-10 00:00:00','   Heyyy. Today was hott. I tryed to tan all afternoon. turning on the music and running outside when the sun came out behind a cloud! yep! hope it worked! Went up to the mall with mom phill dillon to pick up our Cuba tickets! 9 days baby!! Came home. Dillon stayed for dinner. \"do you want apple juice\"-dillon. \"....\" haha, had soup and grilled cheese and it was yummyyy!!  Anyways i better go, tonight em and i are going for a run with adam! should be enjoying!! love vanessa.  \"have you seen thoose bugs!?\"-dillon \"nope..what\"-me \"they make the same sound as crickets, fly like butterflies and look like grasshopers\"-dillon \"....\" \"i think they had a 3some to form a new animal.\"-dillon  '),(14633,3422339,'09,August,2004','2004-08-09 00:00:00','   heyy im home! long time no...blog..eh? Hmm well i had a wicked sweet vaca. i wont bore ya with details, but this is basicly what went down..  last Saturday afternoon  drove to the trailor in Orrilia. That night went for a campfire at John and Steve\'s site with thier Itailian Family. it was so good times. drank a little...golden something? i duno, was lemony and veryy strong! laughed lots, amy and i came back at 12.   Sunday morning  left at 7:05 for Sault Ste Marie...A.k.a. \'Ste Sault Marie\' (john!). was along drive (7 hours) but wasnt as bad as usual cuz we had the TV/VCR in the van. watched austin powers and sports funnies.  Clever Car Games Phill Came Up With:  1) Reading out loud every sign we pass. lasted about 3 signs. 2)Counting the numberof cars we pass on the highway. lasted like 107 cars. stopped for breakfast at 9 30. saw a HOT guy at the French River Trading Post (where we ate). Daddy and i had bacon and pancakes. phill had burger and fries...no comment.  Driving thru the middle of nowhere:  \"where are we\"-phill \"....\" \"i just saw a sign that say chip stand.\"-me \"....riiiight. we\'re in chipstand. population=2\"-daddy.  Funny:  \"what day are the English people coming to stay at your house?\"-dad \"sunday\"-phill \"...no\"-me \"saturday\"-phill \"nope\"-me \"oh friday\"-phill \"uh no.\"-me ... \"theyre coming thursday dad\"-me \"ohh dad they\'re coming thursday eh!? didnt i tell u!\"-phill. ya had to be there!  Driving By Bondhead Restuarant:  \"wtf! who would eat at bone heads!\"-phill.  Driving By Place We Stopped Last Time Going To Sault:  \"hey wait isnt that the place with those bears?\"-me \"what bears\"-phill \"the alive stuffed bears\"-me \"....ohhh vanessa\'s a smart one..\"-dad.  Other Jokes:  \"Hello, Welcome to the Indian Trailor Park. To find your site, follow the arrows!\"-uncle johny. \"What does a mexican fire chief call his children?....Jose and Hose B.\"-uncle Kev. \"heyy look its minivan-essa. or car-essa...truck-essa?..no im just kidding wow youve grown up...full size van-essa.\"   ahh imtired. ill blogg later...i still have like my whole trip to talk about..this blog only had like 1 day in it. meh,later. had an awesome trip tho:) didnt wanna come home, i love going places. ciao love vanessa. Car- essa. PS M.D. is hott.  '),(14634,3623073,'24,June,2004','2004-06-24 00:00:00','   Ohh mY god!! mY money is in!!!... finally a sigh of relieve.... '),(14635,3623073,'22,June,2004','2004-06-22 00:00:00','   I actually finally found the way to change the time... wat a total SelenGe!!!! SheesH... so dumb... hahaha... My eyes r failing on me... But i\'m still inseparable from the comp!!! Wahahaah... First day back at work... Was really fucking bad at first... The staff there treated us as tho\' we\'re non-existent... Full of shit i swear... BUt hack them ah... Plus this morning, my mother actually scolded me for bringing maggi me to work cos i got no more money... Early in the morning at 5am... & this is wat she does.. *speechless* I can\'t wait for my money to come in... Then again, half of it will b gone... WHY?? Cos my hp bill is 244... plus my piano fees is 250... VONDERFUL!!! ahhaah... *GOSH! I\'m acually laughin\' abt this... wat a mad ass* Hahhahaha... oh well, Music is my only Cure... *La la la...* '),(14636,3623073,'21,June,2004','2004-06-21 00:00:00','   The moment you are in tension, You will lose your attention. Then you are in total confusion, & you\'ll feel total irritation. This may spoil your personal relation Ultimately, you won\'t get cooperation, & get things into complication. Then your BP may raise caution And you may have to take medication Why not try understanding the situation? And try to think about the solution. Many problems will be solved by discussion Which will work out better in your profession Don\'t think it\'s my free suggestion, It is only for your prevention. If you understand my intention, You\'ll never come again into tension. '),(14637,3623073,'20,June,2004','2004-06-20 00:00:00','   What a hot hot day! I\'m practically perspiring in front of the comp!!! ARGH!!! aahahahaahah... Im going back to work tmr ward 42*sigh*... I\'m just dyin\' to get back to school!!!! Im missin it so so so much!!! LITERALLY! oh well, enough of complaining... I had good nasi padang at zion rd today... Afte which i went to my grandmama\'s hse to plan for my grand dad\'s birthday buffet next sun... MMMmmmm.... Yummy Food! GloRious Yummy DeliCious FooD!!!*do i sound like a Glutton???* sheesh... haahah... Gonna go cycling with ma lil\' bro lata... *aint i just so sweet????* hahaahah... ya right germaine... whatever sia... ahahhah... AHHH!!!!! I miss my BOO!!!! EEE!!! '),(14638,3623073,'19,June,2004','2004-06-19 00:00:00','   Oh my!!! I am so broke i can die!!! ARGH!!! Me n adeq went to Zee\'s work place - cosi cafe... N her manager is actually quite cute... ahahha... *gatal!* hahahah... I got to see ma boO bOo today!!!! SO HAPPY... My mama went for an interview today... Hope she gets the job... Just for 2 weeks that is... Had tom yam instant noodles just now... Gonna go to bed now la... Cos gotta wake up at 9... So miSSy is gonna take her beauty sleep now... NiGhts! '),(14639,3623073,'18,June,2004','2004-06-18 00:00:00','   WOhOO!!! 10 days of attachment is finally gone! Will be at a medical ward next week *oh no! CHAOS!* Im here listening to \'Right here waiting\'... & boy is it such a love-lay song... wahaaha... heehee... Ooh Ooh! I also did well for my attachment... that\'s just \'vonderful\' la ah... I was dozing off at work today.. again that is... haha... I had a good SHIT today man i tell ya... hahaha... its like 3 days worth of shit sia!!!! haaha... *aint i just so crude?* oh well... I was so sleepy today & BROKE that i just simply slept during ma breaks & have only eaten 2 chicken pies till now... Can\'t wait 4ma dinner!*HOOGA HOOGA* Am so hungry i can eat a COW!! '),(14640,3623073,'14,June,2004','2004-06-14 00:00:00','   another day at work... last time i\'ll work with tt stupid assholic filipino staff nurse... hahaha... but otherwise, its really fun worki with another saff nurse.. weird name tho\'...- magnolia!! I KNOW!!!! ICE CREAM!!! hahahah. but she\'s just da best la... Oh Oh! When i was on the way to work in the lift at SGH, there was this cute guy!! hahaha! I opened the lift door for him n he smiled at me!!! HEHEEHE!!!! That was a good start... hehehe... N then adeq n zee started sayin i \'gatal\' n \'miang\' la... sigh... whay nice girlfriends i have... ahaahah! anyways, me gonna talk to boo boo now... Taaa!! '),(14641,3623073,'13,June,2004','2004-06-13 00:00:00','   I finally have a blog!!! wohoo!!! hahaha... not much to yak about today... went to indoor stadium with da family for \'CORPUS CHRISTI\'... had yummilicious \'wanton mee\' at old airpot road... MMmmm... & now i\'m home on this half broken chair, typing about ma day.. gonna be back at SGH tmr... with a fucking irritating filpino staff nurse... ARGH! But i\'ll just tolerate la ah... ahaha.. ok then... thats all for tonight... '),(14642,3623073,'30,July,2004','2004-07-30 00:00:00','   Fi is in malaysia... N i\'m here al alone!! Im missin him so badly. Y must his auntie get married in malaysia? Y? Singapore water so much cleaner somemore... xiao ding dong... sigh... Thank god i\'m workin on sat n sun... If nt i\'ll go insane!!! N oh ma god! I just managed to read his blog& i have to admit. I\'ve not realised how much he loves me till i read it. i\'m so touched n moved tt im moving now!!! wahahahaah....&nbsp; Its not that i dont know that he loves me... But to think that he is so affected of what happened t me even before we met... Words arejust not enough.... I\'ve no met anyone who cares so much for me... So much that he was oing outta his mind.... Know one thing my dear... I\'ll always be here for you... I\'m so truly, madly , deeply in ove with you... My mother isaging at me to go down n eat now... but whatever... i\'m waitin syg... so get  &nbsp;your ass i singapore quick!! I LOVE YOU! '),(14643,3623073,'04,July,2004','2004-07-04 00:00:00','   Oh well, sch\'s startin... N im xcitd n yet scared... Cos we\'ve got 6 modules... I HAVE so many things to juggle dis ime ard... studies, work, piano... argh... irritatin\' sia... I just want ta have more hols... (cries out loud) At leas i can lk forward to seeing ma HuNny on tuesday... God... Plus this stupid menses cramps is KILLING ME!!!! So SHITY!!!! AAAAAAAAAAHHHHHHHHH!!!!!!!!! I am broke already... Y can\'t money just fall from the sky???? WHY?? sigh... I\'m so tired now... Wanna lie down & yak to ma boy...  We accidentally fell in love... N now i\'m so much in love with you ma boy...  Pardon me for ma shity lang n attitude today... Not iN A Good feeling... d pain is just killin\' me... :( '),(14644,3623073,'04,July,2004','2004-07-04 00:00:00','   Oh well, sch\'s startin... N im xcitd n yet scared... Cos we\'ve got 6 modules... I HAVE so many things to juggle dis ime ard... studies, work, piano... argh... irritatin\' sia... I just want ta have more hols... (cries out loud) At leas i can lk forward to seeing ma HuNny on tuesday... God... Plus this stupid menses cramps is KILLING ME!!!! So SHITY!!!! AAAAAAAAAAHHHHHHHHH!!!!!!!!! I am broke already... Y can\'t money just fall from the sky???? WHY?? sigh... I\'m so tired now... Wanna lie down & yak to ma boy...  We accidentally fell in love... N now i\'m so much in love with you ma boy...  Pardon me for ma shity lang n attitude today... Not iN A Good feeling... d pain is just killin\' me... :( '),(14645,3206585,'26,July,2004','2004-07-26 00:00:00','    urlLink    This is mainly to help Michael navigate his way in Malaysia. I will try my best to spell out the location of where the photos were taken. For starters, I am situated in Kuala Lumpur, the capital. The Husband is in Miri, East Malaysia, whereas my parents are in Johor Bahru and in-laws are in Kuantan. '),(14646,3206585,'25,July,2004','2004-07-25 00:00:00','    urlLink    Ok... same scene. They gave us huge t-shirts, reaching to our knees. I might give mine to the Husband, but the shade of grey is such a yuckie colour. '),(14647,3206585,'17,July,2004','2004-07-17 00:00:00','    urlLink    Waiting for the treasure hunt results to be announced. It was really one long wait. '),(14648,3206585,'17,July,2004','2004-07-17 00:00:00','    urlLink    I know, this is so lame... I wanted to snap a picture of a site we needed to list down during the treasure hunt, but only managed to take a snap at the side view mirror. I know, I am not much of a shutterbug. You don\'t have to point it out. '),(14649,3206585,'17,July,2004','2004-07-17 00:00:00','    urlLink    A view of the Kuala Terengganu rooftops on an extremely cloudy day. You can\'t see much, really. These pictures were taken sometime in June, when the region was experiencing monsoon showers. '),(14650,3206585,'17,July,2004','2004-07-17 00:00:00','    urlLink    Another one. The Husband was in Kuala Terengganu attending a course, but due to torrential rain, he was cooped up in his hotel room most of the time. '),(14651,3206585,'17,July,2004','2004-07-17 00:00:00','    urlLink    The view of the beach from Primula Hotel, Kuala Terengganu on a very windy and cloudy day. '),(14652,3206585,'17,July,2004','2004-07-17 00:00:00','    urlLink    Early morning yoga by the poolside. The picture was taken by the Husband who declined to join, but preferred to recline in bed. '),(14653,3206585,'06,July,2004','2004-07-06 00:00:00','    urlLink    And here is the Cape Rachado Lighthouse. '),(14654,3206585,'06,July,2004','2004-07-06 00:00:00','    urlLink    On the way down, a few of us discovered this fantastic looking bug. The Husband dubbed it the Cape Rachado bug. Unless you see it moving, I\'d think it\'d be a bit difficult to tell which end is head or tails. '),(14655,3206585,'06,July,2004','2004-07-06 00:00:00','    urlLink    A triumphant looking lot who sweated it out to climb to the top. The guide tricked us, there was actually a paved road to get to the top. However, going through the forest was fun. '),(14656,3206585,'06,July,2004','2004-07-06 00:00:00','    urlLink    It was quite freaky being up on the cliff, making our way to the lighthouse. But the view down below was fantastic. '),(14657,3206585,'06,July,2004','2004-07-06 00:00:00','    urlLink    Yeah... was supposed to post this eons ago but didn\'t quite manage. Sensei is front row, second from left. '),(14658,3206585,'06,July,2004','2004-07-06 00:00:00','    urlLink    The view from my previous 6th floor apartment. So cooling and refreshing. To get to the pink-roofed condominiums, you need to travel on a winding and narrow road uphill. '),(14659,3206585,'06,July,2004','2004-07-06 00:00:00','    It\'s now or never. I have to start with these gloomy pictures, to get the ball rolling. '),(14660,3206585,'03,July,2004','2004-07-03 00:00:00','    urlLink    Right, the building there would be my object of desire. It\'s currently still under construction and all units have been sold out, so I am waiting for people to start selling. Keeping my ears and eyes open. '),(14661,3206585,'03,July,2004','2004-07-03 00:00:00','    urlLink    Yet another view showing the city... I need to put up more showing the greens. '),(14662,3206585,'03,July,2004','2004-07-03 00:00:00','    urlLink    No, no, no, sorry to disappoint but that isn\'t our place. '),(14663,3206585,'03,July,2004','2004-07-03 00:00:00','    urlLink    The area has been well-planned; but a bit too many houses for my liking. Nevertheless, beggers can\'t be chooses, as I am often reminded by the Husband. '),(14664,3206585,'03,July,2004','2004-07-03 00:00:00','    urlLink    You can see the minaret of the mosque from the balcony. The call for prayer is loud and clear; especially in the crisp morning air. '),(14665,3206585,'03,July,2004','2004-07-03 00:00:00','    urlLink    This is the view we get from the balcony of the new apartment. On good days (without the haze) you\'ll be able to see KL Tower. '),(14666,3206585,'03,July,2004','2004-07-03 00:00:00','    urlLink    Whoa... grab on to what ever you can darling. '),(14667,3206585,'03,July,2004','2004-07-03 00:00:00','    urlLink    Me - in a state of loss and confusion; as always the case. '),(14668,3206585,'03,July,2004','2004-07-03 00:00:00','    urlLink    Trying my best to keep up. '),(14669,3206585,'03,July,2004','2004-07-03 00:00:00','    urlLink    That\'s me sprawled on the floor. I don\'t feel too bad cause I see others are in the same predicament - though much more graceful at it. '),(14670,3206585,'03,July,2004','2004-07-03 00:00:00','    urlLink    I think the Husband is thrilled to see me being flung to the floor. '),(14671,3206585,'03,July,2004','2004-07-03 00:00:00','    urlLink    Ah... a little help from KB Sensei does the trick. '),(14672,3206585,'03,July,2004','2004-07-03 00:00:00','    urlLink    Err... quite unsure what\'s happening here really. '),(14673,3885215,'31,July,2004','2004-07-31 00:00:00','    I was talking to my mother on the phone on the way home from work. I came soooo close to telling her about the baby. I want to .. but at the same time I am terrified that if I start telling people that I will jinx the baby and something will go wrong. '),(14674,3885215,'27,July,2004','2004-07-27 00:00:00','    Between work, all the funeral stuff for my granddaddy, and being in the first trimester well lets just say Stephanie is one tired out puppy. I won tickets to a movie preview this past week at work. The movie is showing Thursday night so I would have had to switch off on Thursday and work tomorrow instead .. I was actually upset to find out I had won the stupid tickets. I was like can&nbsp; I give them to someone else that wants them. I am just so exhausted from this weekend\'s events that I really need my day off tomorrow. They were kewl with that. They don\'t know I am pregnant yet. No one does. Charles and my cousin Tina are the only people in real life that know I am pregnant. I am still scared to tell my family just yet because of last year\'s loss. Charles gave me a great lead in last night at my mom\'s and couldn\'t believe I didn\'t say anything. But I am just too big of a chicken right now. I am scared that if I start telling people that I will jinx it and that it will all go away. One good thing on the tired ness factor is that my mother and step-father are \"kidnapping\" Chip (they love and adore him and he adores them) and taking him back to Georgia with them for a week or two. Being that I am so exhausted right now I jumped on the offer when they said that. Plus that gives Charles a chance to finish remodelling the house. He has the living room to paint and our bedroom left. It\'s impossible to remodel with a 2 year old underfoot. So we take advantage of my mom\'s offer to keep Chip when we can so he can work on the house. The kitchen and Chip\'s room we did the last time they had him. '),(14675,3885215,'24,July,2004','2004-07-24 00:00:00','    So I have one more angel watching over me now. It\'s so hard. In 1996 I lost my two grandmothers 6 months apart. In 1999 I lost one granddad and then my dad a month apart. Two months later I lost my great grandmother. Then last September I lost my second child at not quite 12 weeks. And now Granddaddy Calvin.        Granddaddy Calvin at his last birthday July 1st 2004      Calvin was my dad\'s dad so atleast now I know he is in good company with my dad and my grandma and my baby that I lost. I got the call last night. My cousin was hysterical. I couldn\'t even understand what she was saying to begin with. Then it dawned on me what was going on. I said \" Do you need me? \" She continued to sob and said \" I don\'t know \" I told her I would be there as fast as I could. And I was there as soon as I could be. I drove like a bat out of hell to get there. Traffic was horrible. A normal hour drive took an hour and 15 minutes.   When I got there I realized the funeral home\'s hearse was blocking the drive (along with all the vehicles of my aunt, uncle and many cousins). I&nbsp; parked 2 houses down at my great aunt\'s and ran across her yard, my cousin Virgil\'s yard and field and raced to granddaddy\'s house. I was so scared that they would have already taken him by the time I got there. I got there just in time. Had I been 5 minutes later I would have missed them probably. My cousin Pam saw my face and stopped the people from the funeral home and said that I needed some time with him. I said my goodbyes, cried and patted his hand and rubbed his arm. He didn\'t look like my grandfather any more.   My cousin Tina who lived with granddaddy and will now be renting the house was scared to stay there by herself after they took granddaddy away and all the aunts, uncles and cousins all went home so I spent the night with her. We were all stressed out so we relived a tradition from our past. Whenever we were upset about something we would spend the night over and one another\'s house and we would bake. Cupcakes, muffins, cakes or cookies&nbsp; .. you name it we baked it. So last night we went to the grocery store and got cake mix and icing and made and decorated cupcakes and then watched (okay half watched because we ended up crashing about 2am) Along Came Polly. Which the part I saw was really funny.   This morning I had to get up early, drive home, order flowers (I ordered an arrangement that me, my mom, and my brother went in on .. a beautiful bible made with mums with red roses accented. My granddad loved roses) and then got ready and headed straight to work. I worked 7 hours then came home. I am sooo exhausted.   Tomorrow is the family visitation (or wake whatever you want to call it). I will spend the night at my mom\'s that way I don\'t have to drive all the way back home and then get up bright and early and drive back on Monday for the funeral.   Please keep us all in your prayers if you pray. It\'s going to be another couple of stressful days for us all. '),(14676,3885215,'22,July,2004','2004-07-22 00:00:00','    Well today July 22nd 2004 exactly one year after my last BFP I got another one. I get another chance at this Mommy thing :) or atleast I sure hope so. This baby will never replace Stella, but I know that I have a hole in my heart that needs filling. I am cautiously due April 3rd 2005 which is exactly 1 month and 5 days before my son turns 3. I am so excited and so scared. I am praying to God that everything goes perfectly this time and we get a healthy baby girl or boy to bring home with us say 8 months from now. If you pray .. please pray for us and our little unborn baby. What a difference a few days can make!!! 2 days ago I was in the depths of dispair .. today I am on cloud 9. '),(14677,3885215,'20,July,2004','2004-07-20 00:00:00','    And what a depressing birthday it has been. My last remaining grandparent is my granddad Calvin (my dad\'s dad) well he is dieing with lung cancer. He took a major turn for the worst yesterday and isn\'t expected to live but maybe a couple of days. Then I decided to take a HPT test today hoping and praying that 13 dpo for once would be enough to get that beautiful 2 pink lines I am so waiting to see. No such luck. I sat and cried before going to work. My husband tried to reassure me because the same thing happened when I was pregnant with Stella our angel baby. I tested early got a negative and cried my heart out only to get a positive 4 days later. I so hope we will get a positive soon. But it\'s so hard to think positive right now.&nbsp; I wish I could remember what it was like to not have to wonder if I would ever get pregnant again. When we first started TTC with Stella everything was so exciting and we just knew we were going to get pregnant and we did .. first month trying. We just knew our little family was going to be complete and everything was going to be perfect. God were we stupid or what? Okay not stupid so much as ignorant. Ignorant of the statistics .. that one in 4 documented pregnancies ends up in miscarriage. Ignorant in the fact that even after seeing the heartbeat you aren\'t out of the woods. 3-5% of women still miscarry after seeing their baby\'s heart beat. It\'s funny I used to comfort myself with that statistic!!! That statistic sounds great until you realize how many MILLIONS of pregnancies there are&nbsp; and that multiplied by that statistic its staggering to know how many women still lose their child after seeing it have a heartbeat. That statistic is no longer comforting when you are in that 3-5%. I also was ignorant of the fact that 1 in 130 births ends in still birth. I have friends that learned that statistic the hard way. I so wish I could be ignorant again. I wish I could have that assurance that I was going to get pregnant again and carry a baby to term and go home with a healthy baby. Not the fear that I live in now. The fear that I won\'t ever get pregnant again, that even if I do get pregnant again who says my baby will make it when Stella didn\'t. Oh how I wish I could have instant amnesia and forget all the horrible statistic I have learned in the last year and just be that woman I was last year waiting on a pregnancy test to turn up positive .. at almost the exact time. I got my positive test with Stella 2 days after my birthday last year. What a horrible year 26 was for me. I am hoping and praying that with the start of this new year I can find better out comes and hope than I have found in the last year of my life. If you made it this far .. thank you for reading my rambling thoughts. '),(14678,3885215,'17,July,2004','2004-07-17 00:00:00','    :bfn .. I pretty much expected it because I have never gotten a :bfp this early in either of my pregnancies. The cool thing is that the test I open was supposed to have one test in it .. but it had two. So I didn\'t lose anything by testing so early. I will probably test again at 15dpo and 17dpo if :af doesn\'t show up. '),(14679,3885215,'17,July,2004','2004-07-17 00:00:00','    OMG .. the last two days the most minor thing totally sets me off into the worst mood!!!! OMG I have flipped out on Charles last night over him not wanting me to fix the kind of Pizza I wanted. Then today in the story purchasing a new tv&nbsp; I just totally flipped out when their was a minor issue. Totally unlike me. Charles is convinced I am pregnant. I hope he is right .. if not I must be in the middle of some hellish PMS.&nbsp; Which AF isn\'t expected for another 7 days.&nbsp; My temps are shooting up really well too. So I am leaning towards the Pregnant theory. Pregnancy for me is like being Manic Depressive for 9 months. '),(14680,3885215,'16,July,2004','2004-07-16 00:00:00','    But it\'s sooo hard!!! My temps have been super high the last 3 morning in a row. I know it\'s possible that it\'s because of sleep interuptions and such. Oh but I sure hope they are legitimate temps and mean that my chart is really triphasic this month. I would so love to be pregnant. Not that the TTC hasn\'t been fun (haha), but come on&nbsp;already. Let\'s bring on the puking and peeing all the time!!! I think women who have been ttc for a long time or who have had miscarriages are the only women that pray for morning sickness. But really .. I am just ready and raring for some morning sickness. '),(14681,3885215,'15,July,2004','2004-07-15 00:00:00','    My nephew Evan Ray Currie came into the world screaming his head off after about 25 minutes of pushing on my beautiful sister in law\'s part. She had an incredible labor. She refused pitocin, refused&nbsp;an Epidural&nbsp;and only had a shot of Stadol to help take the edge off. Evan was 7lbs 2oz and 20 inches long. Which is huge for a baby 3.5 weeks early.&nbsp; Here he is with his loving Aunt Steff . Isn\'t he precious???    &nbsp;  I am totally in love with my wonderful little nephew. I can\'t wait until he is old enough to play with his cousin Chip (my 2 year old).     My sister in law is going to be the best mother ever!!!&nbsp; Not to mention how great she looked after a 12.5 hour labor.&nbsp; Like a model!!!     Here is a photo of their little family. That\'s my big brother Steven in the Orange hat. The proud Daddy!!! (Not to mention very tired!!!) '),(14682,3885215,'13,July,2004','2004-07-13 00:00:00','    My sil is right now in labor. Her water broke at 3pm today. I just got off the phone with my brother. She is about 3-4 centimeters dialated and his contracting on her own but they don\'t seem to be too strong. I know she wants to go as natural as possible so I hope and pray the Dr. doesn\'t stick her on pitocin. I wish I could be there so bad, but I have to work in the morning so I can\'t leave tonight. It\'s killing me sitting here and not being there. I asked Steve if they were giving Sharon any antibiotics since she hadn\'t had her Group B Strep test. He asked Sharon and she said they had given her the Group C Strep test today but haven\'t gotten the results back. I told Steve to make sure that they mention it to the nurse when she comes back in. I know how horrible Group B Strep can affect a baby and I don\'t want them taking any chances with my precious little nephew. It\'s amazing how losing a child instills fear in your heart. Before my miscarriage pregnancy and birth didn\'t scare me. Now it terrifies me. '),(14683,3885215,'12,July,2004','2004-07-12 00:00:00','    Ugh .. all last cycle I went without buying any HPT or having any in the house. So I didn\'t test at all last month and just waited for AF to show up on her own. I was so proud of myself!!! It was like being on the testing addict wagon. Well guess what ? I fell off the wagon tonight. We were at the Dollar tree and DH made the mistake of pointing out that they \"had pregnancy tests coming out the wazoo\" and I saw them .. and automatically my addiction kicked in and I grabbed some. I couldn\'t help myself. It was like a crack addict needing a fix. I was good though. I limited myself to two. But I promised myself that if I need more I was come back. SICK SICK SICK!!! Where can I sign up for Pregnancy Test Addict Anonomous??? I need an intervention. DH says I should just take a dollar bill out and pee on it and then flush it. He thinks it does pretty much the same thing. And I can\'t say as I blame him. It\'s pretty much true. '),(14684,3885215,'11,July,2004','2004-07-11 00:00:00','    So today makes me 4dpo. Now I know in my head that there is no way I could possibly know at this point whether I am pregnant and will be getting a BFP this month. Even if the sperm and the egg met up the chance of it having implanted this early is extremely small. In my head I know implantation takes place normally within the 6dpo-10dpo window and that\'s when the HCG starts building and that\'s what makes you start having pregnancy symptoms. Since I know this, why is it that when something as minor as my breast starting to itch does the phrase \"Oh my goodness .. my boob is itching .. is it sore? Am I finally pregnant again???\" immediately pop into my head? TTC I think has made me into a hypechondriac, not with illnesses .. but with pregnancy symptoms. And believe me it\'s not intentional. Because I totally hate the getting excited and hopeful each month only to be greeted by a negative HPT and the followed shortly there after by a visit from my least favorite relative Aunt Flo. My body just can\'t seem to help itself. Each month it builds me up with these phantom symptoms and against my best intentions I forget everything but the idea of being pregnant again. I feel like this dog an old friend of mine had. About every 6 months or so the dog would go through a fake \"pregnancy\" and then apparently give birth to imaginary \"puppies\". We all talked about how crazy the dog was and we would laugh. Now I am starting to be able to imagine how the poor dog felt. '),(14685,3885215,'09,July,2004','2004-07-09 00:00:00','    I almost got to be an aunt last night. My mom called about 10pm last night saying she had just got off the phone with my brother and that my sister in law was having contractions and had been spotting blood all day. Well the contractions were like 9 minutes apart and they called the Dr. and he told them to monitor them for an hour and if they kept on coming, got any closer or got more intense to give them a call. Within an hour they were headed to the hospital because they were 6-7 minutes apart and were getting much more intense. My brother had told me that Sharon was a day or two shy of 37 weeks so I got ready and headed to their hospital which was about 2 hours away. Apparently Steve had gotten the weeks wrong because Sharon was just at 36 weeks last night. The nurse put Sharon on the monitor and checked her and she was in real labor it wasn\'t just Braxton-Hicks. She had started dialating and was 90% effaced and Evan (my nephew\'s name is to be Evan Ray after my brother and my dad)was engaged and ready to go. The nurse called the Dr. on call and he decided that he didn\'t want to let the labor to progress so they gave my sister in law a shot of Tirbulen (sp?) and her labor stopped. They ended up discharging her and gave her 3 days of Tirbulen pills. She has to take it every 4 hours for the next three days until she sees her Dr. on Tuesday morning. I am pretty irritated with her Dr. because he didn\'t even come in to see her while she was there. She will be seeing a different Dr. on Tuesday so hopefully he will show her more attention. The good news is according to all the monitoring Evan is doing outstanding. I was so worried something would go wrong. After my loss and having so many friends who have gone through such terrible heart wrenching losses I am on paranoia over drive until this baby get\'s here safe and sound. I love my brother and his wife so much and I already love my nephew to be sooooo much that the thought of anything going wrong .. it\'s just too terrible to think about. My little nephew is a miracle baby. My brother and his wife tried for like 2.5 years to get pregnant with him and I know they were beginning to worry that they may never get have a child. I am so thankful that they are getting to experience all this wonderful stuff. I just hope and pray that little Evan gets here safe and sound. I told Sharon that if she could wait 9 more days Evan could be on  my birthday. That would be so awesome. '),(14686,3885215,'09,July,2004','2004-07-09 00:00:00','    My DH and I have been on the TTC Rollercoaster for over a year. We started TTC #2 back in June of 2003. I got pregnant immediately. Too much celebrating on the 4th of July apparently. Our little firecracker baby we called our unborn child. Well atleast until all hell broke loose. We found out on the 10th of September that our baby\'s heart beat had stopped. I had a D&E 2 days later. The dr. told us it would be best to wait 3 cycles before TTC again to let my uterus recover from all the turmoil. So we did. We couldn\'t wait to TTC again I mean it only took 6 months to get pregnant with our son Chip, and then it only took a month of trying for our second child (we named our angel Stella Rhiannon which means Star Goddess .. we thought it appropriate for an angel baby) so surely we would be pregnant shortly.   Well here we are 8 cycles later, still waiting on the Fertility Gods to say,   \"Hey those people right there deserve to have another child.\"   My dad (he died in April of 1999.. I know he is in Heaven taking care of my angel baby) used to tell me that times of tragedy and strain are charecter builders.   I have lost my father, 3 grandparents, a great grandparent, and my child all in the last 8 years. Oh yeah and my only remaining grandparent has been diagnosed with lung cancer and has only 3 months to live. Gee do you think I have enough charecter yet??? '),(14687,3885215,'15,August,2004','2004-08-15 00:00:00','    We are trying to get him used to the idea of mommy having a baby growing in her tummy. And he has been so sweet about it. He will pat my belly and then he says \"Baby\" then he will kiss my belly. It\'s Sooooo incredibly sweet. It makes me want to cry!!! Then some times he changes it a little and will pat my belly, say \"Baby\" and then when I think he is going to kiss my belly he instead blows a ZERBERT in stead. It makes us laugh. He is such a little goober. I love him so much. I can\'t wait to see him get to have a little sibling to have as a friend and companion. The baby is due the month before Chip hit\'s 3 years old so I think it will be a great age difference. '),(14688,3885215,'12,August,2004','2004-08-12 00:00:00','    They were happy. I ended up not doing anything special .. shame on me .. but I have felt like crap lately so haven\'t done anything much except work and sleep. We were all watching Chip and I was like so do you like him enough that you think you are ready for another one? and My mom looked at me and goes .. are you pregnant? and I said yep. I told her how far along I was and she was like you kept this a secret for this long. I explained to her that I had found out about the baby the day before my granddad died and I didn\'t think that that was a good time to break the news .. and that I hadn\'t wanted to tell them over the phone I had wanted to tell them in person. They were excited. Mom told Charles and said he had better had made sure that this time was a girl .. we all laughed and said .. it\'s probably a boy. She said oh well if he is anything like his big brother he will be the perfect child.  After they left I talked to my SIL on the phone and I went ahead and spilled the beans to her so that mom wouldn\'t have to worry about keeping it a secret when they went to see them. Not sure how my brother is taking it. He stills see me as this 13 year old girl that I was when he moved out and went to college. He had a hard time getting used to me getting married and having Chip .. and now this is my 3rd pregnancy .. so I am sure it\'s a little uncomfortable for him to think about. The kewl thing is I got pregnant with this little baby about a week before his son was born .. so Evan and this little baby (either Skyla or Cage) will be about 9 months apart .. give or take. '),(14689,3885215,'10,August,2004','2004-08-10 00:00:00','    Okay .. I go to the Dr. to get my official test and everything on Friday. Not sure when I will actually go in for a real appt. after that. But for some reason today it\'s been driving me crazy .. what if the pregnancy isn\'t going well .. what if my HCG has dropped. I kept thinking .. I haven\'t puked yet .. my boobs aren\'t quite as sore as they were last week. What if what if what if. It\'s enough to drive ya batty. So today I ran and grabbed another pregnancy test. THANK GOD for Dollar Tree tests!!! This is the 4th one I have taken so far. So I take it .. I swear that is the fastest positive pregnancy test I have ever seen!!!! I mean the second line was DARK PINK before the first line was even wet. So that reassured me and made me feel like a ninny at the same time. So the news for today .. I am STILL PREGNANT !!! Woo hoo. This is gonna be the longest first trimester in history. I say a little prayer on a regular basis every day. \"Please God let my baby be okay. Let him or her be healthy and a sticky baby. Please let me go home from the hospital in 8 months with a healthy baby.\" This is my daily prayer.  Oh yeah and my new favorite friend is GINGER ALE .. This stuff is the only thing that seems to help me get rid of the \"I need to yack\" all day feeling. And what\'s with the runny nose and extra saliva? I took a shower this morning and ever since .. my nose has been running like a race horse. '),(14690,3885215,'07,August,2004','2004-08-07 00:00:00','    My last post actually posted!!! YEAH ... I am so glad whatever was going wrong the other day is now fixed. '),(14691,3885215,'02,August,2004','2004-08-02 00:00:00','    I woke up this morning and was freaking because I wasn\'t nautious. Now couple hours later I am so nautious I can barely contain myself. But that\'s okay. I know it\'s weird to pray for nausea but after having a miscarriage .. any pregnancy symptom is something to cling to.  Today I am 6 weeks pregnant. If I can make it 6 more weeks and everything be okay .. I think I will calm down some. I found out at 11w4d that my baby no longer had a heartbeat, after having had one for 3 weeks, my last pregnancy. So my goal is to make it over the 12 week goal line and hopefully this pregnancy will be a sticky one. '),(14692,3885215,'02,August,2004','2004-08-02 00:00:00','     '),(14693,3504445,'03,June,2004','2004-06-03 00:00:00','   well happy belated birthday to mee!! finally i;m a year older. 15. i guess i;m already old enough to think abt my life what i want out of it and change. i know theres one thing i need to change in my life. well no gonna write it heree. anyway ytd woke up with a small pink baggiee beside mee on the table. i know what that is. my present from my parents. cos they gave the same thing to all my sisters so it wasn;t really a surprise for me. but i don know what is inside. all i know it is a jewellary. soo i open it and there is a small cutee silver earrings. aww i love it. and i straight away wore it. lols. and my sister, who are at home!, sms-ed me asking me how to celebrate my bday. then i;m like  ok lets go to breeks.  a wonderful dinner i thought it mitee be turns kinda terrible. it;s the food. the horror! i rather eat next door which is seoul garden. breeks has no variety of food. and nothing special is on the menu. i ordered seafood platter which consists of everything fried and it makes me wanna puke. have to admit the seafood platter at the fish & co is much much much better. i really regret eating there. sort of. well at least i\';ve food to eat ritee.   oh yah thanks to all the peeps  vina dillys liping nigel arifin maryann asri simone fazilah  for the birthday wishes. and thank you liping for the cutee birthday e-card. well my this year;s birthday celebration isn;t over yet!! going to have a proper celebration at downtown east with my family my relatives. there;s gonna be cake, bbq and lots of food. sure thats gonna be better than ytd. ok enough about my birthday. chao!   '),(14694,3504445,'01,June,2004','2004-06-01 00:00:00','   i don know what is happening now? is it me or is it really the end of our friendship? i don know. never call never sms never email. no news nothing. i wonder if whatever i confess to him causes all this to happen. but it can\'t be. i already ask him to promise that whatever i told him will not affect our friendship. he promised! but what is happening now? it\'s been exactly one month since he last called me. can he be that busy that he can\'t call me? well who am i to him for him to call everyday ritee. i\'m nobody. nobody who loves somebody. i really treasure this friendship. he\'s the only one whom i still keep in contact with. he\'s always making me laugh and make my day. i still remember everytime he calls my heart will beat faster. after that last call, i\'ve been like a fool. running to the phone every time it rings but in the end to be heartbroken to find out it\'s not him. even if he doesn\'t want to talk to me anymore, i\'m kinda fine with it but couldn\'t he just open his mouth and tell me?? keeping quiet is more hurting then saying it. i\'m now confused. can\'t reach him via his mobile. wonder if he purposely change his no. probably. i din say it is true that he changed his no. well wanted to call his house alot of times but i\'m just afraid. afraid of him. afraid that he doesn\'t want to talk to me. but all i wanna know is  are we still friends? are you still talking to me?   plss i want to know. don keep silent. silent kills. it hurts soo much after you left me years ago. but you manage to heal my heart when you decided to keep in contact with me. i;m so faithful. i\'m still liking you. but you just can\'t see it. now don break my heart again with your silence. well i let destiny decide..       LOVE ME NO MORE (by Bardot) i toss and turn in bed can\'t get you out of my head even though you\'re so far away i need you here with me oh boy, why can\'t you see that i can\'t live without your love when i close my eyes i think of you well i wish i had you here with me but there\'s nothing i can do, oh oh  chorus counting every day that goes by and the tears that i cry (tears that i cry) you don\'t wanna love me no more wish that you could hold me tonight i\'m hurting inside cause you don\'t wanna love me cause you don\'t wanna love me love me no more  i tried to call your phone but you ain\'t been at home i need to find out where you are so i can make you see that you belong with me for me there is no other love when i close my eyes i think of you i wish i had you here with me but there\'s nothing i can do, oh oh  repeat chorus  all i know is that i can\'t live without you ooh, but i wanna know what made you feel this way i\'ll be right here for you (right here for you) and everything that you do (thing that you do) but how can i get close to you when you don\'t feel the way i do, ah hah  repeat chorus to fade   '),(14695,3504445,'01,June,2004','2004-06-01 00:00:00','   i am still sick today. but went out to look for fun. lols. went to watch troy with my cousin, anum. at causeway pt. for free of cos. lols. got complimentary tixs remember? the one i won from that sms contest. yeah. i notice something during the show. the troy necklace that i won which i thought was ugly is the same necklace as what brad pitt wore when he was acting in troy. woah cool eh?? and the show is so sexy. ahahah!! with the gorgeous guys wearing \"short skirts\" [thats what i call them. lols. :P] and some ahem* scenes which small kids should. NOT watch. yah and orlando. hes so yummyy!! melting* but i hate his character in troy. a young but stupid prince who fall in love with the wife of the king of sparta. haiz. and all the killings made me cried. especially when hector died. pity his wife and baby. i cried you know. also when hector\'s father went to see achilles and ask him to give bak hector;s dead body. his words touched my  '),(14696,3504445,'01,June,2004','2004-06-01 00:00:00','   well. i feel like i\'m fated to be with an aquarius. look at this   test no one. took it at www.astocenter.com    Gemini & Aquarius  The Aquarian ability to devote body and soul to the cause of justice attracts the respect and love of the Gemini, who will vie to conquer this rare love. Once the two Air signs are united, they will merge into a shared pleasure in action, excitement, and rebellion against stuffy conventions. Together, they believe life gives them the right to freedom and independence. Nothing stops or separates them. Friends as well as lovers, they will unite their souls in making their life into a masterpiece, a great and exciting adventure.    test no two. took it from http://web.tickle.com  Rafidah, the best Zodiac Match for your personality is Aquarius   Aquarius, the Water Bearer (January 21 to February 19): This friendly and thoughtful partner is just your type. Initially, an Aquarius may catch your attention by showing you their intellectual side and optimistic outlook. But as you get to know them better, you\'re likely to be even more drawn to an Aquarian\'s idealistic resolve and focus on making positive change. People born under this sign are frequently capable of seeing both sides in any dispute. This can make them quite talented when it comes to finding creative solutions in your relationship. Your Aquarius may be a bit headstrong at moments, but they\'ll also tend to look for the bright side in times of trouble. In the bedroom, you\'ll likely find the Water Bearer both discrete and cautious. An Aquarian will ordinarily make sure that all areas of your relationship are strong before plunging into sex. Overall, Aquarians are solid, unwavering partners and complex people. So, it may take you a little extra time to get to know your Aquarian match, but it\'s apt to be well worth the wait.   and now some free zodiac compatibility thingy at http://astrocenter.astrology.msn.com//msn/deptlove.aspx?af=-1000  Gemini & Aquarius    The Aquarian ability to devote body and soul to the cause of justice attracts the respect and love of the Gemini, who will vie to conquer this rare love. Once the two Air signs are united, they will merge into a shared pleasure in action, excitement, and rebellion against stuffy conventions. Together, they believe life gives them the right to freedom and independence. Nothing stops or separates them. Friends as well as lovers, they will unite their souls in making their life into a masterpiece, a great and exciting adventure.   '),(14697,3504445,'01,June,2004','2004-06-01 00:00:00','   just made this blog yesterday. will not be updating this blog. and probably will update my od. maybe. holidays are making me busy with activities, hols assignments and the making of my blog layout. its not so nice la. but nevermind im a freshie at this. not so good like some of my peeps. before i forget. tmr is going to be my big dayy. lols. ya anybody wanna join me at downtown east from 4-6 of june, contact me ok? well got to go. chao.   '),(14698,1694057,'31,July,2003','2003-07-31 00:00:00','   is that the terri rosenblatt i went to high school with?  urlLink jetset records is featured in cmj.   '),(14699,1694057,'30,July,2003','2003-07-30 00:00:00','   urlLink digital piracy can be funny.   '),(14700,1694057,'28,July,2003','2003-07-28 00:00:00','   urlLink miss melody nelson and diego elefant?  who knew? just kidding. that girl does look like audrey though.   '),(14701,1694057,'28,July,2003','2003-07-28 00:00:00','   urlLink the new friendster is tribe?  [link via  urlLink affordable justice ]  and  urlLink stereogum  says,  urlLink friendster is a spam sham .  oy givalt!   '),(14702,1694057,'28,July,2003','2003-07-28 00:00:00','   WALKMEN TOUR DATES!   8/2   Summer on Southport - Chicago, IL 8/3  Magic Stick - Detroit, MI 8/4  Horseshoe Tavern - Toronto, ON 8/5  La Sala Rosa - Montreal, PQ 8/6  Paradise - Boston, MA 8/7 Pier 54 - New York, NY (**this is a free, all ages show) 8/8  Black Cat - Washington, DC 9/7   Maxwells - Hoboken, NJ 9/8   Maxwells - Hoboken, NJ 9/11 Casbah - San Diego, CA 9/12 Henry Fonda Theater - Los Angeles, CA 9/14  Bottom of the Hill - San Francisco, CA 9/15  Slims - San Francisco, CA 9/18  Richards on Richards -  Vancouver, BC 9/19  Graceland  - Seattle, WA 9/20  Dantes -  Portland, OR   The Stills Hook Up with Interpol!  August: 21 Rock Feed Back Party @ Buffalo Bar London 22 Leeds Festival  24 Reading Festival 25 NME Night @ The Monarch 26 Middlesborough Cornerhouse BBC 6 Live  with Steve Lamacq support to the rapture  27 Manchester MDH with interpol 28 XFM Boat Party with the rapture  September (on tour with interpol): 16 Montreal with interpol 17 Toronto with interpol 18 Detroit, MI with interpol 19 Chicago, IL with interpol 20 Cincinnati, OH with interpol 22 Lawrence, KS with interpol 23 Denver, CO with interpol 24 Salt Lake City, UT with interpol   '),(14703,1694057,'25,July,2003','2003-07-25 00:00:00','   heard new morrissey songs today from some demo sessions. where did the irony and cynicism go? what is this \"i like you, please hold me\" bullshit? oy, morrissey, someone needs to punch you so that you feel some sarcasm again.   sweetness, sweetness i was only joking when i said i\'d like to mash every tooth in your head...   '),(14704,1694057,'24,July,2003','2003-07-24 00:00:00','   urlLink the morning news  on the key to a successful freelance career: slacking off. genius.   '),(14705,1694057,'24,July,2003','2003-07-24 00:00:00','   it\'s a fucking glamorous life again!  urlLink sheila e.  reunites Prince\'s backup bands. woo-a-hoo.   '),(14706,1694057,'24,July,2003','2003-07-24 00:00:00','   urlLink salon  rips off  urlLink the modern age , as far as the MOB project is concerned. shame on you! can\'t you try to be original anymore?   '),(14707,1694057,'24,July,2003','2003-07-24 00:00:00','   wish  urlLink jack white  well. xoxo, nme.   '),(14708,1694057,'23,July,2003','2003-07-23 00:00:00','   urlLink Poke the Penguin.  You know you want to.   '),(14709,1694057,'22,July,2003','2003-07-22 00:00:00','   Part deux of the  urlLink Prefix  yeah yeah yeahs interview with karen o. she talks about her korean heritage, living in jersey, her dark days sloshing down beers in odessa bar, and how brian chase gets marriage proposals all the time. and by god, don\'t forget the  urlLink audio clip !   '),(14710,1694057,'22,July,2003','2003-07-22 00:00:00','   urlLink grandaddy  interview...   '),(14711,1694057,'21,July,2003','2003-07-21 00:00:00','   urlLink friendster tees!  that\'s scary. i don\'t want my face on someone\'s boob. [ urlLink gawker ]   '),(14712,1694057,'21,July,2003','2003-07-21 00:00:00','   if you read the new york times today, you\'ll see an article about siren music festival on page e5. look at the modest mouse photo and play \"where\'s jinners?\"  word! i\'m famous!   '),(14713,1694057,'21,July,2003','2003-07-21 00:00:00','   yep,  urlLink sir mix-a-lot  is back with BIG songs that are sure to be hits like \"big johnson,\" \"big ho,\" \"big screen,\" and \"poppi grande.\" i\'ve got the album, do you?   '),(14714,1694057,'21,July,2003','2003-07-21 00:00:00','   cuz we know  urlLink true hipsters  write for new york magazine, a rag for yuppies.  yay,  urlLink grandaddy  comes to new york in october.  Be careful of what you download!  urlLink RIAA Goes After File Traders With Hundreds Of Subpoenas    urlLink Ryan Adams  likes water. No, really.  I still think  urlLink  28 Days Later   won\'t be scary. I don\'t care how much your eyes bugged out.  I think  urlLink Gawker  has a crush on David Amsden. She transforms into Gusher when she\'s around this \"hipster\". C\'mon, he would never make it on Hipster Bingo.   '),(14715,1694057,'18,July,2003','2003-07-18 00:00:00','   weezer\'s  brian bell  gets spacey. what\'s new?  was that adam green i saw on the graham avenue stop of the \'ell train?  i have this feeling that the witnesses are gonna be featured in new york magazine very soon... especially since they had a photog following them around siren festival.   urlLink russell crowe  -- father? poor kid.  bad news for people who like  urlLink internet porn .  will  urlLink lollapalooza  finally die?  guitar stuff for  urlLink radiohead\'s \"there there\" . thanks  urlLink stereogum !  coldplay\'s  urlLink chris martin  goes apeshit on photog. bloggers beware.  i already know who  urlLink jin  is. you don\'t have to email me anymore about it.  the entertainment weekly dot com code this week is: \"entertainweb\"   '),(14716,1694057,'18,July,2003','2003-07-18 00:00:00','   wow,  urlLink underwear party  in greenpoint tonight. amazing. chilly. stupid... but if you\'re going, take lots of photos! at least there\'s a clothes check. how thoughtful of them. this is going to be like that jimmy eat world video, eh? not very novel if you ask me. [link from  urlLink gawker ]   '),(14717,1694057,'18,July,2003','2003-07-18 00:00:00','   our favorite ibiza deejay  urlLink erick morillo  and p. diddy hook it up.   '),(14718,1694057,'18,July,2003','2003-07-18 00:00:00','   urlLink prefix magazine  interviews karen o. she\'s NOT a morning person. this is part one of the interview.  and by golly, don\'t forget about the  urlLink audio clip  at the end of the interview. karen o is sweet.   '),(14719,1694057,'18,July,2003','2003-07-18 00:00:00','   R.I.P.:  urlLink Celia Cruz  and  urlLink Benny Carter . both written by jinners   '),(14720,1694057,'18,July,2003','2003-07-18 00:00:00','   urlLink gameboy  for cats! [link from  urlLink les.Yper-sound ]   '),(14721,1694057,'18,July,2003','2003-07-18 00:00:00','   when  urlLink porn  goes too far, we tend to criticize it even more. go figure.   '),(14722,1694057,'18,July,2003','2003-07-18 00:00:00','   urlLink liz phair  says, \"eff yoo.\"   '),(14723,1694057,'18,July,2003','2003-07-18 00:00:00','   from entertainment weekly:   Celebs plan \'\'Simpsons,\'\' \'\'King of the Hill\'\' cameos. Everyone from Brad Pitt to Tony Blair will appear   Brad Pitt apparently hasn\'t let the lackluster box-office for his \'\'Sinbad: Legend of the Seven Seas\'\' deter him from doing more cartoon voiceover work. He\'ll be leading a Texas-sized contingent of celebrities doing voice cameos next season for \'\'King of the Hill\'\' and \'\'The Simpsons,\'\' Fox announced Thursday.  Pitt will bring his pipes to \'\'King,\'\' as will Ben Stiller, Christina Applegate, Jamie Kennedy, Laura Dern, and Fred Willard. Several country stars will visit Arlen, Texas, for the Christmas episode, including Trace Adkins, Brad Paisley, George Strait, Deana Carter, and Travis Tritt. Other musicians slated to appear are Kiss\' Gene Simmons and Texas\' own Kelly Clarkson.  Over in Springfield, the 15th (!) season of \'\'The Simpsons\'\' opens with an episode featuring the return of Glenn Close as Homer\'s fugitive mom. Another returning parent: Jackie Mason, as the rabbi dad of Krusty the Clown. The Halloween \'\'Treehouse of Horror\'\' episode will feature Jennifer Garner, Jerry Lewis, and boxer Oscar De La Hoya. The already much-touted trip to England (in an episode called \'\'The Regina Monologues\'\') will include appearances by British Prime Minister Tony Blair, Ian McKellen, J.K. Rowling, Jane Leeves (as Grandpa Simpson\'s old flame) and token Yank Evan Marriott (of \'\'Joe Millionaire\'\'). Also visiting Springfield will be Simon Cowell, Mr. T, \'\'Bowling for Columbine\'\' director Michael Moore, and the late Robert Stack.   '),(14724,1694057,'18,July,2003','2003-07-18 00:00:00','   complimentary  urlLink butterfly chairs  for your pussycat.   '),(14725,1694057,'18,July,2003','2003-07-18 00:00:00','   urlLink mike d.  says fannypacks are hot. so does  urlLink daily candy .   of course, i\'ve always liked  urlLink tote bags .   '),(14726,1694057,'18,July,2003','2003-07-18 00:00:00','   another article on  urlLink friendster ... but so true!   '),(14727,1694057,'18,July,2003','2003-07-18 00:00:00','   idlewild almost makes the big time, says  urlLink the new york times .   '),(14728,1694057,'29,August,2003','2003-08-29 00:00:00','   urlLink product shop nyc  makes a triumphant return. you should put the  urlLink On the Moon Music  shows on your calendar!!  dj dop discovered this  urlLink anti-hipster shit . oy vey. can\'t we let the trucker hat thing go already?   '),(14729,1694057,'27,August,2003','2003-08-27 00:00:00','   urlLink ms. kittenpants  interviews  urlLink michael ian black  and  urlLink the polyphonic spree\'s tim delaughter . i\'m not sure whether mib was trying to be funny or jerky. but i still love him. and i still love kittenpants!   '),(14730,1694057,'27,August,2003','2003-08-27 00:00:00','   The Strokes shot a video last night in Brooklyn, N.Y., for \"12:51,\" the first single from their sophomore album, \"Room on Fire,\" due Oct. 21 from RCA. As with videos for prior tracks such as \"Last Nite,\" \"Someday\" and \"Hard To Explain,\" the new clip was directed by Roman Coppola. \"12:51\" will be delivered Sept. 9 to U.S. radio outlets for airplay consideration and released commercially Oct. 6 in the U.K.  Like its predecessor, 2001\'s \"Is This It,\" the new album contains 11 songs but at just over 32 minutes in length, is about three-and-a-half minutes shorter than the debut. A number of songs underwent last-minute renaming, with \"Raga\" changing to \"Automatic Stop,\" \"Supernova\" morphing into \"12:51\" and \"Never Needed Anybody\" being altered to \"Between Love and Hate.\"  [ urlLink billboard ]   '),(14731,1694057,'26,August,2003','2003-08-26 00:00:00','   urlLink julian drinks with gideon.  \"\"We\'re not super serious artists, but we take what we do very seriously.\" â€” the Strokes\' Julian Casablancas [mtv.com]   '),(14732,1694057,'26,August,2003','2003-08-26 00:00:00','   RADIOHEAD video for \"Go To Sleep\"... see our favorite whiner/crooner thom. you may like it. you may hate it. whatever. [rolling stone]   '),(14733,1694057,'26,August,2003','2003-08-26 00:00:00','   BECK GETS LOUD After thirteen months on the road plugging last year\'s Sea Change, Beck will head into the studio next week to begin recording his eighth album, expected next summer.   \"It\'s pretty aggressive,\" he says of the new material. \"I\'ve been working on it for a while, but I ended up doing Sea Change first. I\'ve been wanting to do a record with loud guitars for a long time.\" [ urlLink rolling stone ]   '),(14734,1694057,'26,August,2003','2003-08-26 00:00:00','   you thought ambulance were playing with the stills at sin-e? uh uh. the ambulance boys bailed. you heard it here first.   '),(14735,1694057,'22,August,2003','2003-08-22 00:00:00','   check it:  urlLink  stellastarr* interview   by jinners.   if you are a stellastarr* fan, you\'ve probably wondered who the hell coco, jenny, and moongirl are. during my conversation with with the stellastarr* lead singer and songwriter  shawn christensen , i got him to talk about the women who inspired his music and lyrics...   he also talks about how he went from songwriting to singing his lyrics. yep the two rock star notions didn\'t always coincide!   '),(14736,1694057,'21,August,2003','2003-08-21 00:00:00','   urlLink ON THE MOON MUSIC FALL PREVIEW!   HEY KIDS!!!  It\'s back to school time, baby! So put yr pencil back in yr pocket and listen up! This summer has been super fun--what with all the RAD SHOWS and parties, countless cups of HELADO DE COCO, and that pesky lil\' BLACKOUT! But it\'s almost time to retire them white shoes and stanky SWEATBANDS, and get down to business!   ON THE MOON  is here to show you just how you can get to the head of the class! JINNERS and ANNABIOTICA are your headmistresses!  urlLink info@onthemoonmusic.com    AUGUST REQUIRED COURSES:  +++DANCE 101  Friday, August 29th @ Sin-e *  urlLink CHERRY  (featuring e*vax) *  urlLink DE-REGULATOR  *  urlLink SAINTFACE  *  urlLink THE EAVES   Wallflowers take note, these bands will have you on the dance floor faster than spiked punch! Guaranteed social improvement through movement. Listen to  urlLink Cherry\'s \"17 Years\" mp3 .  Doors @ 8 // Show at 9 $8    SEPTEMBER REQUIRED COURSES:  +++FILM 101 Wednesday, September 3rd THE STILLS @ Sin-e   THE STILLS VIDEO PREMIERE PARTY The Stills will play one New York show before going on tour with Interpol. Not only will they perform all of your faves, but they will also premiere the video for their single, \"Still in Love,\" at the end of the night with some killer giveaways!  Stay tuned for more information!  Doors @ 8 // Band @ 10 // Video Screening & Giveaways @ 12  +++COMP 101 Wednesday, September 10th @ Sin-e *  urlLink KIND OF LIKE SPITTING  *  urlLink KILOWATTHOURS  *  urlLink COLD SIDES  *  urlLink THE NEIN   Kind Of Like Spitting has released six previous records on small DIY labels and done fourteen tours, taking a staunchly independent approach to making records and performing. Ben Barnett, the core element of Kind Of Like Spitting, is an astounding guitarist and songwriter, and has a punk rock heart and mind.    +++ECON 101 LUXX ROCK \'N ROLL GARAGE SALE Sunday, September 14th @ Luxx  Twice as nice! This is our second ever Rock \'n Roll Garage Sale at Luxx! A portion of proceeds from this event will be donated to the  urlLink Brooklyn Animal Rescue Coalition  (BARC). Shop local! Shop for a good cause!  Free admission // Cool A/C // Refreshing Drink Specials  BUY, SELL & TRADE! MUSIC GEAR, CLOTHES, RECORDS, ART!  Vendor booths are now available to rent for the day of the sale. Lot sizes come in Small $10, Medium $15, or Jumbo $20 (sizes start at 3\' x 4\' but depend on availabiity and type of merchandise). Call Anna at 718-387-2007 or email  urlLink annabiotica@onthemoonmusic.com  for more info.   YOU MUST RSVP BY FRIDAY, September 12th to guarantee a space   +++DRAMA 101 Thursday, September 25th @ Sin-e *  urlLink ON!AIR!LIBRARY!  *  urlLink PELA  *  urlLink SAY HI TO YOUR MOM   ON!AIR!LIBRARY! create visceral aural textures you thought possible only in movies, or your dreams.  Doors @ 7:30 // Show @ 8 $8  See you soon! xoxo, J & A   '),(14737,1694057,'19,August,2003','2003-08-19 00:00:00','   did you, uh, hook up during the blackout?  urlLink this  timeout  writer  wants to know. [craigslist]  johnny rotten  urlLink disses  avril and kelly oz:  â€œAnyone who\'s parading under a hundred-thousand-dollar-plus video is not free from corporate. That\'s just the MTV advertising agency. I find them all to be just a bit of a sham. I mean I don\'t mind, I don\'t care, that\'s for some people who are into that. But don\'t be telling me what my life is. Because in the likes of Kelly Osbourne and this silly Lavigne thing, yapping on about what punk is or isn\'tâ€”it\'s just unbearable. At their age, if they\'re calling me old, then I want to know why they haven\'t done nothing of their own. Nothing. It\'s all borrowed from my life. Therefore, they have no word of any value in it.â€   urlLink ever wonder what happened to that nirvana  nevermind baby ?  i didn\'t either. [rolling stone]    ew, does someone seriously want their  urlLink diesel jacket  autographed by  american  reject justin guarini? i think not. now the jacket is totally worthless. is anyone really entering this contest? and why is diesel partnering up with  american idol  contestants? i just don\'t get how the two mesh...   '),(14738,1694057,'19,August,2003','2003-08-19 00:00:00','   urlLink test your new york wits.  how you like dem big apples? btw, i totally failed this quiz. i am not a true new york fanatic/denizen/whatever. what was your score? i\'ll surely bet money that it was higher than mine. who the hell wrote this quiz anyway? [daily candy]   '),(14739,1694057,'19,August,2003','2003-08-19 00:00:00','   i love this  urlLink  eels   cd:  shootenanny . why did i not spin it before today? it speaks to the softer side of jin (yeah, you didn\'t think it got any softer, did you?). i am grooving to \"love of the loveless.\" for some reason i thought these guys sounded more hard core. perhaps because the tagline under their cd title says \"everybody knows thse are rock hard times.\"  btw,  urlLink hootenanny  means:  1) An informal performance by folk singers, typically with participation by the audience. 2) Informal. An unidentified or unidentifiable gadget.   '),(14740,1694057,'19,August,2003','2003-08-19 00:00:00','   urlLink Recoys CDs coming soon!  [affordable justice]   '),(14741,1694057,'19,August,2003','2003-08-19 00:00:00','   the new york times  urlLink reviews   stoked .   '),(14742,1694057,'19,August,2003','2003-08-19 00:00:00','   Coming October 21st on Off Records:  urlLink  Wig In a Box: The Songs from Hedwig and the Angry Inch     This CD was put together by Off Records, John Cameron Mitchell and Steven Trask to benefit the Harvey Milk School, a part of the Hetrick-Martin Institute... (You know that place you walk past every now and then near Astor-Disaster -- aka Astor Hair Styliststs -- and B&N)  Here\'s the amazing tracklisting:  Rufus Wainwright - \"Origin of Love\" Sleater-Kinney & Fred Schneider - \"Angry Inch\" They Might Be Giants - \"The Long Grift\" Frank Black - \"Sugar Daddy\" Robyn Hitchcock - \"City Of Women\" Imperial Teen - \"Freaks\" The Breeders - \"Wicked Little Town (Hedwig Version)\" Bob Mould - \"Nailed (LoudBomb Mix)\" The Polyphonic Spree - \"Wig in a Box\" John Cameron Mitchell & Stephen Trask - \"Milford Lake\" Stephen Colbert - \"Ladies & Gentlemen\" Spoon - \"Tear Me Down\" Yoko Ono & Yo La Tengo - \"Hedwigâ€™s Lament/Exquisite Corpse\" Ben Kweller with Ben Folds & Ben Lee - \"Wicked Little Town (Tommy Gnosis Version)\" Cyndi Lauper & The Minus 5 - \"Midnight Radio\" Jonathan Richman - \"Origin of Love (reprise)\"   '),(14743,1694057,'17,August,2003','2003-08-17 00:00:00','   urlLink knife skills  get a GLOWING review from popmatters. that\'s right, mofos!   '),(14744,1694057,'17,August,2003','2003-08-17 00:00:00','   the strokes announce tracklist for their album, due out Oct. 21. so not only is justin timberlake channeling michael jackson, so are the strokes:  Guitarist Albert Hammond, Jr. also told NME that band favorite \"Nightmare\" has a backbeat hearkening back to Michael Jackson\'s 1983 smash hit, \"Billie Jean.\"  next thing ya know, julian will start grabbing his crotch  while  falling down drunk. ouch! [pitchfork]   '),(14745,1694057,'16,August,2003','2003-08-16 00:00:00','   cult of yellow submarine: in a twisted state of lucy in the sky with diamonds,  urlLink living colour cover the beatles  on their first album in a decade. they also cover ac/dc. i think my boss is good friends with vernon reid. perhaps she already has the album? i\'ll ask. [rolling stone]   '),(14746,1694057,'15,August,2003','2003-08-15 00:00:00','   urlLink the fiery furnaces are glowingly reviewed by the ny times!  a special kudos to eleanor and matthew for being way cooler and way more rockin\' than meg and jack. you didn\'t have to read kelefah to know that. on the moon music loyalists know what we\'re talking about! british sea power and jeffrey lewis were kickass too, but i have a special place in my heart for the fiery furnaces. they have been the soundtrack to my days, especially in the past couple weeks. i keep  gallowsbird\'s bark  blasting in my cd player constantly! i especially love matthew on \"inca rag\" and eleanor on \"up in the north.\" another thanks goes out to mike reddy who introduced me to this band\'s music a long, long time ago.   '),(14747,1694057,'15,August,2003','2003-08-15 00:00:00','   someone at work said, \"when stuff like this happens in england, we go to the pub.\" apparently the same is true in new york.  urlLink in times of black out crises, there\'s always drinking for new yorkers.  [ny times]   '),(14748,1694057,'14,August,2003','2003-08-14 00:00:00','   it was a kind of eery and lonely walk from lincoln center (66th street and broadway) to williamsburg, brooklyn. there were thousands of people around me, many of them walking in two and threes and even fours. i scurried past them.   i opted not to put on my headphones. i wanted to be aware of everything going on around me, sirens from cop cars and fire engines, the screaming babies, the radio news blasting from random cars parked on the street, the screeches of car brakes as they tried to make their way through intermittent deadlocks.   i stopped by the entertainment weekly building to see if there would be any familiar faces there. there wasn\'t.   i continued on and made my way downtown via second avenue. there were all sorts of wacko traffic activity because of the lack of traffic lights and traffic cops. conscientious civilians began to take matters into their own hands, helping to direct traffic. that was nice to see.   i kept walking.  many stores were closing. delis were also closing down. many people perched outside of their office buildings to wait it out. i wanted to get out of the hectic mess that consumed midtown. i trudged on faster, getting annoyed at all the slow pedestrians who happened to wear their stilleto sandals to work. suckers. i had my trusty converse.  i couldn\'t help my memories of sept. 11, which flooded back to me. the chaotic scene of thousands of people on the street, similar to now. i knew we were going through a huge blackout that spanned the north eastern united states, but who knows what would happen in our incredibly vulnerable state? i remember walking across the 59th street bridge on sept. 11 with several other friends. most of us walked in silence as we saw the empty, smoky downtown skyline from the bridge.   i got down to the east village and thought i would check to see if my friend rachel was home. she wasn\'t. i saw one of her neighbors, and they kindly offered me some bottled water. i took it. i started to feel like i was going to pass out, not only by the physical exertion and dehydration but because mentally, i just wanted to talk to and relate my feelings to another person and there was no one. my cell phone continued to not work, and i felt so frickin\' irritated.   i continued to walk.     i got near the entrance to the williamsburg bridge, and i actually saw people i used to work with at alloy -- will, john, bill, and tim. i was overjoyed. i began to take some photos of what was going on around me. i smoked a cigarette. i trudged on and chatted about what the hell was happening around us. i felt the 59th street de ja vous again. it made me woozy.               we ran into deron who was on his way into the city to find anna. i thought to myself, that is totally crazy. he should just stay home and wait for her to come home! but it wasn\'t my business so i stayed quiet and wished him luck on his search. when i turned around, the alloy peeps were gone. they hadn\'t waited for me. jerks.  then i ran into chingyi and her friend. they were heading to bed-stuy. we separated.  i continued to walk, again on my own.   when i finally reached my house, keith was there waiting for me. he had been terribly worried, and i was glad that i had someone who still cared about my whereabouts. it was a nice feeling, and i didn\'t feel so alone anymore. the blackout was not the most pleasant experience, but at least one good thing can still result from it.   urlLink \')\"> getCount(\' \')   '),(14749,1694057,'14,August,2003','2003-08-14 00:00:00','   urlLink a strokes remix.  thanks, nick! it\'s so weird, but kinda fresh. and not as sexy as d\'angelo\'s ripped bod.  i\'m also enjoying the bootylicious sounds of the dirtbombs (from detroit). \"the thing\" is da bomb diggity.   LinktoComments(\' \')    urlLink  \">Comment   '),(14750,1694057,'13,August,2003','2003-08-13 00:00:00','   urlLink damon albarn  goes solo. well, that only makes sense because blur is basically \"the damon albarn show.\" [nme]   '),(14751,1694057,'13,August,2003','2003-08-13 00:00:00','   for the fashion forward:    Sample Sale Alert   What: Intermix When: 8/13-8/16; Wed.-Sat. 11 a.m.-7 p.m. Where: Metropolitan Pavilion, 123 West 18th Street, between Sixth and Seventh Avenues (212-243-3177). Why: Chloe tanks, were $450, now $99; Catherine Malandrino dresses, were $420, now $119; Diane von Furstenberg tops, were $118, now $29; Theory cropped pants, were $195, now $69; Seven jeans, were $148, now $49 ... You get the picture.   What: Norma Kamali When: Through 8/30; Mon.-Sat. 10 a.m.-6 p.m. Where: 11 West 56th Street, between Fifth and Sixth Avenues (212-957-9797). Why: Because Kamali\'s dropped her prices to below $250 on vintage-line accessories and swimwear collections.   What: Find Outlet Warehouse Sale When: 8/19-8/24; Tues.-Sun. noon-7 p.m. Where: 361 West 17th Street, between Eighth and Ninth Avenues (212-243-3177). Why: Up to 90 percent off trendy labels such as Mon Petit Oiseau, Jill Stuart, Joe\'s Jeans, Emma Black, Julia, and others.   What: Pumpkin Maternity When: 8/7-8/21; Mon.-Sat. noon-7 p.m., Sun. noon-5 p.m. Where: 407 Broome Street, at Lafayette Street (212-334-1809). Why: Because pregnant women have more to show off. Dot dresses, were $108, now $60; V-neck jersey dresses, were $108, now $60; capri pants, were $98, now $50; silver shalvar/tunics, were $119, now $60.   What: Malia Mills Swimwear When: 8/16-8/23; Mon.-Sun. 12 a.m.-7 p.m. Where: 199 Mulberry Street, between Spring and Kenmare Streets, (212-625-2311) Why: Bikinis, baby! Tops and bottoms are sold separately in any size, style, and fabric for up to 80 percent off retail.   What: Leigh Bantivoglio When: 8/13-8/14; Wed.-Thurs. noon-6:30 p.m. Where: 648 Broadway, between Bond and Bleecker Streets, suite 301 (212-420-9994). Why: Camisoles, panties, boy shorts, and thongs sold at Barneys, Bergdorf Goodman, La Petite Coquette, Saks, and Neiman Marcus are 80 percent off.   What: Barneys Warehouse Sale When: 8/14-9/1; Mon.-Fri. 10 a.m.-9 p.m., Sat. & Sun. 10 a.m.-7 p.m. Where: 255 West 17th Street, between Seventh and Eighth Avenues (212-593-7800). Why: Unless you\'ve been living under a rock, there\'s not much explanation necessary for this one. (And if you\'re agoraphobic, you might want to stay under that rock.) Designer men\'s and women\'s clothing, footwear, accessories, baby stuff, and housewares are up to 80 percent off.   What: Michelle Roth & Co. When: 8/19-8/20; Tues.-Wed. 10 a.m.-6:30 p.m. Where: 24 West 57th Street, between Fifth and Sixth Avenues, suite 203 (212-245-3390). Why: Getting hitched? Designer bridal gowns were $2,500 to $10,000, now $695 and under.   What: Totem When: 8/14-8/31; Mon.-Fri. 11 a.m.-6 p.m., Sat. 11 a.m.-6 p.m., Sun. noon-5 p.m. Where: 71 Franklin Street, at Broadway (212-925-5506). Why: Modern designer home furniture is up to 60 percent off. Pastoe bookshelf, was $1,535, now $471; F-seat sofa, was $2,274, now $1,364.   What: Rafe When: 8/12-8/17; Tues.-Thurs. 11 a.m.-7 p.m., Fri.-Sat. 11 a.m.-8 p.m., Sun. noon-7 p.m. Where: 1 Bleecker Street, at Bowery (212-780-9739). Why: Handbags adored by Cameron Diaz, Alexandra von Furstenburg, and Helena Christensen are up to 75 percent off retail.   What: Samantha Root When: 8/19-8/20; Tues.-Wed. 10 a.m.-5 p.m. Where: 1133 Broadway, between 25th and 26th Streets, suite 708 (212-255-0017). Why: Cute-as-can-be bags and cosmetics cases are 60 percent off.   What: Billion Dollar Babes When: 8/23; Sat. 8 a.m.-6 p.m. Where: The Metropolitan Pavilion, 110 West 19th Street, between Sixth and Seventh Avenues, North Pavilion Entrance. Why: Wake up early, because stuff goes fast. James Perse, Joe\'s Jeans, Ella Moss, Kate O\'Connor, Oliver Peoples, Anya Hindmarch, Blue Cult, Miss Sixty, YA-YA, and more are up to 80 percent off wholesale. And there\'s loot for boys, too. Log on to bdbinvite.com to R.S.V.P.   [daily candy]   '),(14752,1694057,'12,August,2003','2003-08-12 00:00:00','   By popular demand,  WET HOT AMERICAN SUMMER  returns to the big screen in New York City...what a perfect way to celebrate the end of summer, by remembering the last day of summer in 1981.     LABOR DAY WEEKEND ONLY  August 29, 30, 31 at MIDNIGHT AMC Empire 25 (42nd & 8th) -Cast members will appear live. -Special give-aways, music performances and more -details as they develop will be posted at  urlLink www.wethotamericansummer.com   If you\'ve only seen it on video, you HAVEN\'T SEEN IT!    * * *   Also --there may be seats left at late shows, just added for  SOUNDTRACKS LIVE  - August 14 & 21  - an evening of live scenes and songs from  PRETTY IN PINK .  Directed by Amy Poehler, music directed by Amy Miles.  Cast includes Rachel Dratch, Michael Ian Black, Michael Showalter, A.D. Miles, Matt Ballard, Jack McBrayer & Amy Poehler.  Musicians include Craig Wedren, Peter Salett, Amy Miles, David Wain & Fred Armisen.  At UCB Theater.  More info -  urlLink www.soundtrackslive.com   '),(14753,1694057,'12,August,2003','2003-08-12 00:00:00','   urlLink liz phair  reviewed in the  ny times . oh kelefah.   '),(14754,1694057,'12,August,2003','2003-08-12 00:00:00','   urlLink writers sound off  on the new and strange \"pop of king\" column in  entertainment weekly  featuring commentary from... stephen king? [morning news]   '),(14755,1694057,'12,August,2003','2003-08-12 00:00:00','   urlLink these days it pays to be indie.  our market continues to thrive amid industry slumps and one of the worst years for layoffs. it\'s \'cause we are intensely loyal and know that supporting one another is the only way to survive. we prefer insound to amazon. we shop at other music, kim\'s, and newbury comics rather than virgin megastore. word.   '),(14756,1694057,'12,August,2003','2003-08-12 00:00:00','   urlLink get your name inscribed on a british sea power cd sleeve.   '),(14757,1694057,'12,August,2003','2003-08-12 00:00:00','   urlLink riaa wants to woo college students back.  napster-gate has no end, does it?   '),(14758,1694057,'12,August,2003','2003-08-12 00:00:00','   urlLink snoop gets all r.kelly on us.  tsk tsk. you should never believe girls who say they are 18. c\'mon, you\'re just asking for a lawsuit!   '),(14759,1694057,'12,August,2003','2003-08-12 00:00:00','   i guess it\'s a day for asians in the press. asians not only rap (see link below),  urlLink they rock . [modern age]   '),(14760,1694057,'11,August,2003','2003-08-11 00:00:00','   props to rapper jin for stealing my name and  urlLink getting published in the  ny times   before me. damn you!   '),(14761,1694057,'11,August,2003','2003-08-11 00:00:00','   urlLink pete libertine reaches for help:  \"When I was arrested for burglary, they found traces of opiates and crack cocaine in my blood. I\'m all cut up, unsure, I don\'t know what I\'m doing, I need a good kicking and I need some help.\"   '),(14762,1694057,'11,August,2003','2003-08-11 00:00:00','   a special thanks to  urlLink catherine\'s pita  for linking to me three times. how come i didn\'t see you backstage at the walkmen? oh yeah, i don\'t know what you look like. you should introduce yo\'self next time! we could\'ve been drinking the free booze together!   urlLink new york doll , thanks for the link. you are not wrong when you say i am the number one walkmen fan in the city. i think everyone who talks to me says, \"so you like the walkmen, eh?\" at least their not saying, \"jin, like gin and juice!\"   '),(14763,1694057,'10,August,2003','2003-08-10 00:00:00','   urlLink fashion fashion fashion.  [from educated community]   '),(14764,1694057,'10,August,2003','2003-08-10 00:00:00','   urlLink derek  is running for some kind of political office. if you\'re in his area, vote for this dawg. he\'s down.   '),(14765,1694057,'10,August,2003','2003-08-10 00:00:00','   hard glitter reviews the  urlLink saintface show at the mercury lounge  on saturday. her comments reflect my sentiments exactly. that\'s why i booked the band at sin-e for august 29. they will be playing with the eaves, de-regulator and cherry. mark your calendars, peeps! this band is for all the people who love moz, pulp, and blur!   '),(14766,1694057,'10,August,2003','2003-08-10 00:00:00','   want to go on the tinkle booze cruise but don\'t have any money?  urlLink try writing lyrics for the new les savy fav mp3, \"get me now,\" and you could win two free tickets.  i tried doing it today, and didn\'t get anywhere. i wasn\'t cut out to be a rock star, but this song is gonna be amazing. lsf are so fucking good!   '),(14767,1694057,'08,August,2003','2003-08-08 00:00:00','   aeri (grosse contemp) has  urlLink an amazingly beautiful write up  of the walkmen/french kicks show in dc at the black cat (the day after the hudson river rocks show). it almost made me weep. she also has posted a  urlLink slideshow of photos  from the show.   vanessa (vive/le/roi) has also posted her  urlLink detailed write up  of the walkmen/unicorns show in montreal. she also has wonderful photo galleries on the  urlLink walkmen ,  urlLink tangiers , and  urlLink the unicorns .  oddly enough, both aeri and vanessa report poor sound problems at the walkmen shows... am i crazy when i say i want to see this band again? aeri also says hugh was mia at the french kicks performance. i wonder what is going on with that. i am intensely curious. anyone got details they can share with me? i\'ll keep it, like, secret.  divestar also has a  urlLink small photo gallery  from the walkmen/adam green pier show. bidi has posted her  urlLink review  of the pier show.    urlLink !!! work on dirtier album ...whatever that means. dirt dirt dirt? mud mud mud? mold mold mold? expect the dirrty album in early 2004, touch and go. and here\'s a piece of good news for soviet panda --  !!! will appear Sept. 26 at the Wire/Adventures in Modern Music festival at Chicago\'s Empty Bottle. A host of other forward-looking acts will also be on hand, including Black Dice, Adult., Lightning Bolt and Michael Gira. [billboard]    urlLink gregory hines. r.i.p.  he was only 57!!   urlLink l.a. tomorrows parties finally rescheduled.  now you can purchase your plane tickets! the event will take place nov. 8 - 9 on long beach. expect cat power, built to spill, spoon, modest mouse, magic band, mission of burma and the boredoms, and more! it\'s all ages and bands will perform at two venues on docked cruise liner the Queen Mary, now a luxury hotel and convention center.An East Coast edition of ATP, to be curated by Stephen Malkmus & the Jicks, was originally slated to take place this fall in New Jersey, but has been indefinitely postponed. bummer. [billboard]   urlLink the angelika film theater gets dissed  for poor sound quality, thin theaters, and unreliable a/c. the best part of the article? they quoted mike albo!!! jco will appreciate that. remember the good ol\' days when we used to intern at  urlLink the moth  and knew people like mike albo, jonathan ames, george dawes green, hazelle goodman, sheila ratner? crazy!! feels like DECADES ago. remember the last time we went to the moth to see gideon yago and kimya dawson (moldy peaches) tell  urlLink rock and roll stories ? that was pretty nifty. [ny times]   urlLink the kills get a 7.8 from pitchfork.   urlLink startime darlings the joggers get an 8.0.  boo ya!  dennis realistic sings an acoustic version of  urlLink \"why didn\'t you stay?\" . classic. makes me weepy.  eben has graciously put me on his guest list for the  urlLink tv on the radio  show tonight at luxx... free shows are good so i\'ll have to pass up the grandaddy warsaw show in favor of rockin\' goodness and perhaps a free drink or two at luxx. meow.   '),(14768,1694057,'08,August,2003','2003-08-08 00:00:00','   for 12 rheingold caps, you can get your very own  urlLink authentic rheingold trucker hat . uh... ok!   '),(14769,1694057,'06,August,2003','2003-08-06 00:00:00','   well well well. chico has posted a wonderful jinners story on his blog,  urlLink Rock n Roll in the Real World . folks, i think that deserves a spot on my list of peeps. thanks chico! also i realize i need to add more people to that list of peeps. word.  in other blogger news, there is  urlLink an indiscernable PR piece on blogger gurls  in that uppity alternative nyc paper,  new york press , featuring jasper (ahem, gurl?), laura and audrey. i couldn\'t figure out what this story was about -- \"vicious\", the modern age, blogger \"gurls\" or how blogs have slowly become yet another PR tool for pushy publicists? all i have to say is that i didn\'t care about  new york  rock until i heard the walkmen. and for that i thank miker, who gave me their cd before it was even released. smooches, miker!  and kudos to isaac for having the coolest fucking record label ever. shuddup, naysayers.  were you at siren wearing your best pointy shoes and polka-dotted tube?  urlLink then you\'re probably one of these people.  thanks, nina. click for slideshow  urlLink here .  the strokes get all new-wavey on us [ urlLink billboard ]:   In particular, tracks like the hand-clapping \"Supernova\" (a \"first single contender,\" according to Ralbovsky) and \"I Can\'t Win\" feature more of a new wave-vibe than \"Is This It,\" thanks to bright guitar lines that replicate frontman Julian Casablancas\' vocal melodies.     urlLink the shins make basement tapes... due out in october . and to answer the mystery, that top model girl was dating the keyboardist. i wonder if singer james mercer is single? tickets for the shins tickets at the bowery ballroom on sept. 2 go on sale today... it\'s going to sell out quick. i have one ticket for sept. 3. anyone want it?  who needs a  urlLink photobooth , when one has a digital camera and a  urlLink fotolog ?   '),(14770,1694057,'05,August,2003','2003-08-05 00:00:00','   the walkmen show previewed in the  urlLink new york post . you should note that the walkmen haven\'t quite risen to \"superstardom\" status (at least not until hamilton marries j.lo after she divorces ben), and adam green is still in the moldy peaches so he\'s not exactly \"ex-Moldy Peaches\". ugh.    August 5, 2003 -- TO prolong that summertime feeling, pray for no rain, have some pink drinks al fresco in the Meatpacking District, then head to Pier 54 (West Side Highway at 14th Street; no phone) for the Walkmen on Thursday at 7 p.m..   They first rose from the ashes of Jonathan Fire*eater and then rose to national superstardom with their Saturn ad-tune, \"We\'ve Been Had,\" but the neo post-punk band isn\'t resting on its laurels.   The New York quintet, fronted by the theatrical Hamilton Leithauser, has a follow-up to the vintage-piano sprinkled \"Everyone Who Pretended to Like Me is Gone.\" Expect new tunes. Adam Green (ex-Moldy Peaches) and the Oranges Band open.   '),(14771,1694057,'01,August,2003','2003-08-01 00:00:00','   excerpt from my  urlLink kings of leon interview . post it on your blogs!! merci!   '),(14772,1694057,'01,August,2003','2003-08-01 00:00:00','   WOW! DON\'T YOU WANT ME?  THE HUMAN LEAGUE FEATURING: PHILIP OAKEY JOANNE CATHERALL SUSANNE SULLEY  w/Guest DJ MICHAEL T (of MOTHERFUCKER)Spinning before and after THE HUMAN LEAGUE  August 5th $25.00 Advance $30.00 Day of Show SRO * Doors @ 6PM 8 PM  Synth-pop\'s first international superstars, the Human League was among the earliest and most innovative bands to break into the pop mainstream on a wave of synthesizers and electronic rhythms, their marriage of infectious melodies and state-of-the-art technology proving enormously influential on countless acts following in their wake.  Their breakthrough hit was the classic single \"Don\'t You Want Me,\" from the album Dare!; both topped their respective charts in England, and went on to become major hits in the U.S. as well.  In 1983, they scored a pair of hits with \"Mirror Man\" and \"(Keep Feeling) Fascination.\"  To the surprise of many, the Human League resurfaced in 1986 with Crash, produced by the duo of Jimmy Jam and Terry Lewis; the plaintive lead single \"Human\" soon topped the U.S. charts.  [info from Saryn]   '),(14773,1694057,'30,September,2003','2003-09-30 00:00:00','   the triumphant return of french kicks\' wwfkd?  urlLink doggie style my best friends girl a larry situation sex in the city   and then there were four... i mean, five -- who is this mysterious  urlLink \"dan\"  on drums?   '),(14774,1694057,'29,September,2003','2003-09-29 00:00:00','   hmmm, i just remembered i bought tickets to see margaret cho at the beacon theater on oct. 10. where the hell are my tickets?? anyone else going?   '),(14775,1694057,'29,September,2003','2003-09-29 00:00:00','   in defense of liars i\'ve heard a lot of people griping about the \"secret\" liars show on saturday night at the knitting factory. ok, here\'s my response: if you didn\'t like any of the four songs that liars played on saturday night, then you probably never really liked noisy post-wave rock. that\'s ok, of course. so they\'ve become more experimental -- so what? my guess is that all that touring with sonic youth has heavily influenced the band, and they\'re trying out a new sound. i have to admit, i was curious as to what the band would sound like minus two original members, but i wasn\'t disappointed by the band on saturday -- i was disappointed by the fans, most of whom were stiff as a board during the set. and for some reason i was surprised that the guys didn\'t play any oldies and sad that their set didn\'t last longer, but that doesn\'t mean i didn\'t like what they played. this band has never been about following what the rest of williamsburg has been churning out. keep that in mind, dearies. the new york rock revolution will soon skew towards more experimentalism, and mtv rock will fade into the  trl  countdown.   '),(14776,1694057,'29,September,2003','2003-09-29 00:00:00','   the new  urlLink hardglitter . i like.   '),(14777,1694057,'29,September,2003','2003-09-29 00:00:00','   i\'m going to start a weekly  urlLink OTM  newsletter that will include nyc indie show listings and a gathering of relevant interviews and features. if you want to be part of the mailing list, please sign up by emailing  urlLink info@onthemoonmusic.com . everyone already on the OTM mailing list will already be included. newsletter will come out every sunday to start your week off right!  urlLink email me  if you have a show coming up and want to be included in the listing.   '),(14778,1694057,'29,September,2003','2003-09-29 00:00:00','   so many friends and friendsters playing this week. what\'s a girl to do?   09.30.03 brilliant mistakes  @ galapagos   10.01.03 ambulance ltd, aerial love feed  @ mercury lounge   10.01.03 qatsi, read yellow  @ pianos   10.02.03 turing machine, hominid  @ lit (open bar from 9-10)   10.03.03 dirty on purpose  @ grand central (in the \'burg) no cover   10.03.03 sonic mook experiment: panthers  @ sin-e   10.03.03 divestar and melody nelson: the realistics, the flesh  @ bowery ballroom   '),(14779,1694057,'29,September,2003','2003-09-29 00:00:00','   very cool photos of:  urlLink o!a!l!   urlLink dirty on purpose   urlLink pela  a la jasper   '),(14780,1694057,'29,September,2003','2003-09-29 00:00:00','   urlLink sup mag  spells stellastarr* wrong. on a brighter note, they do spell the notwist right. i like that cameron cook, fun journalist extraordinaire.   '),(14781,1694057,'28,September,2003','2003-09-28 00:00:00','   the natural history   urlLink music video . i can\'t get it to work on my computer, but perhaps you can tell me what it looks like?   '),(14782,1694057,'28,September,2003','2003-09-28 00:00:00','      '),(14783,1694057,'28,September,2003','2003-09-28 00:00:00','   just surfing the ny times:   urlLink aclu vs. music industry   urlLink lower cd prices could hurt the yyys   urlLink marg cho still rules   '),(14784,1694057,'28,September,2003','2003-09-28 00:00:00','   urlLink  liars   played a \"secret\" show at the knitting factory last night. i have been in liars withdrawal for quite a while now, so i was excited to see them play again. there was a pretty large gathering of people there for the show. asif saw me holding tickets to the show and he yelled at me. \"bitch, i put you on the list, you\'re jackey\'s plus one!\" uh, jackey didn\'t tell me! it was funny cuz he called her up right then and there and yelled at her. ha. then he bought me a drink. sweetheart. love love him! i told him i didn\'t mind buying tickets and supporting the bands. he called me a dying breed. then he introduced me to katie from  young people  who was knitting at the knitting factory. priceless!  i didn\'t see cameron or bidi around, but mike d tells me they were there. i did run into Say Hi To Your Mom. chatted with those boys for quite a while. it was nice to run into them again. i loved their set, and i thought that lead singer eric had such a cute stage presence. i love running into people at shows these days. i feel like going to shows is so much more of a community event now. a couple years ago for me, it was just going to a show and seeing a band. now it\'s like seeing many of my friends.    liars  played songs from their forthcoming album, which includes tracks that are more experimental. and angus gets on the guitar! i didn\'t get to snap any photos, but i did rock out. i have to say though that the crowd was pretty lame. it\'s probably the tamest crowd that liars have ever played for. why don\'t you dance, people! start a little ruckus. by the time i got into the front and center area, the boys had just finished their final song. that was the shortest set ever!  naf gave me a bootleg of  dirty on purpose  and  on!air!library!  at the sin-e show. i think i might post a couple of the tracks online here for a limited amount of time -- maybe when i have the photo galleries up. for the rest of the night i lost track of the friends i came with (they ended up in chinatown) and so i hung out with mike d and kevin. alcohol was just not agreeing with me that night.   the next liars show: october 30 at the bowery ballroom. buy your tickets now, suckas.  also tonight, go see  sea ray  at the bowery ballroom with  clem snide !   '),(14785,1694057,'27,September,2003','2003-09-27 00:00:00','   new  walkmen  west coast tour  urlLink photos  a la  urlLink affordable justice .   '),(14786,1694057,'26,September,2003','2003-09-26 00:00:00','   tonight i went to luxx -- you know that glorious rock club of old that used to have punk/dance/electroclash shows and hipsters hanging outside smoking and bullshitting? yeah, that one. there\'s also a new \"carribbean bar\" that opened up a couple doors down from the club. if you haven\'t heard yet, luxx is turning into a gay dance club beginning oct. 25. but right now there is a late night gay dance party on friday -- after the rock shows. so during the  25 suaves  set (a husband and wife team like the white stripes only harder, crazier and better), i stepped out for a cigarette and some alone time. that was probably my big mistake. while i\'m smoking, i overhear this conversation.   guy one:  so you\'re gay right?  guy two:  no i\'m straight.  guy one:  oh, i\'m straight too.  obviously though guy one does not believe that guy two is straight. he continues.   guy one:  things are so different now. back in the day, i could batter you.  needless to say, guy two soon snuffs out his cigarette and heads back into the club. unfortunately, of course, that means that guy one moves his interrogation on to me.   guy one:  you probably don\'t want to talk to me after you heard that conversation.  me:  [silence, puffing cigarette]  guy one:  can i ask you one question?  me:  no.  guy one:  just one question. c\'mon!  me:  no.  guy one:  just tell me, what makes you want to be with the same sex?  me:  excuse me? what are you talking about?  guy one:  what makes you want to rub up on another girl?  me:  what the hell are you talking about? i don\'t.  guy one:  you don\'t? ah man, i fucked up didn\'t i? but what makes you want to have a relationship with another girl?  me:  i don\'t.  at this point he must have lost his train of thought or was thoroughly confused. he continues.   guy one:  oh you\'re chinese!!  me:  [silence, puffs cigarette faster]  guy one:  you\'re chinese right?  me:  no.  guy one:  you\'re not chinese? ah man i fucked up again, didn\'t i? you\'re not chinese?  me:  no. what are you?  guy one:  puerto rican. if you\'re not chinese, what are you?  me:  korean.  guy one:  oh. i fucked up. i know you don\'t want to hear my story but let me tell you anyways...  me:  no, don\'t tell me your story.  at this point, he tries to light his cigarette with mine and i back away, saying, \"i have a lighter.\" this somehow perturbs him and he says, \"keep your lighter.\" and finally, walks away. i went back into the club, but left shortly after. this kind of weird harassment does not bode well for the gay disco and its carribbean neighbor.   '),(14787,1694057,'26,September,2003','2003-09-26 00:00:00','   mark your handy dandy calendars:  tuesday, october 7  urlLink  kenna   performs for free @ the supper club doors 8 p.m.  plus, an intimate discussion with gil scott-heron -- the revolution will not be televised!   '),(14788,1694057,'24,September,2003','2003-09-24 00:00:00','   currently listening to  the stills  -  logic will break your heart . in stores october 21. [fawn]   '),(14789,1694057,'24,September,2003','2003-09-24 00:00:00','   r.i.p robert palmer.  ;-(   '),(14790,1694057,'24,September,2003','2003-09-24 00:00:00','   thanks to all who came out last night to the show at sin-e. i thought  urlLink  on!air!library!   was amazing. the new drummer kevin adds depth to their dark, moody songs. pee and ali\'s voices sounded wonderful. and i love claudia on the farfiza!   as announced on  urlLink melodynelson  yesterday, i have begun to manage  urlLink  dirty on purpose  . so i\'ve been going to the last few practices as they ran through their set list for last night, and i knew it was going to be killer. the set was pretty heavy, and they debuted a new song tentatively titled \"spider eyes\" where joe sings in a voice that draws me in every time because it is so distinctive. many of their songs flowed really nicely into each other. and they ended with everyone\'s favorite, dynamic instrumental, \"monument.\" it was totally hot.   urlLink  pela   also sounded wonderful live, and the boys in that band were such a joy to talk to! as sweet as the music they were singing. lead singer billy has this unique voice that rings with heartfelt, uplifting melodies -- a voice that really carries the beautiful music they create. my faves are from their ep -- \"episode,\" \"latitudes\" and \"venom.\" and the live performance was just as good, if not even better, than their recording. i was so happy!  surprisingly, i knew the keyboardist in  urlLink  say hi to your mom   already! so when alex walked into soundcheck yesterday, i was like, \"hey, i know that guy!\" funny how that happens, eh? i had never seen them live, but i\'ve been listening to their mp3s online for a while now. so i was very excited to see shtym in the flesh! and i must say they delivered. i hold \"the fritz\" dear to my heart. i love how the songs are so smooth musically and have this really cool, almost jazzy vibe to them -- eric\'s vocals have a mysterious edge, and then the lyrics have sarcastic, witty bite, which gives the songs even more appeal. can\'t wait to see them again!  it made me so happy to see so many of my friends at the show -- christian and justin from bishop allen, naf from OTM Tokyo, Knife Skills, deron, Jenna, Jeff Sea Ray, Asobi Seksu, rachel, jennifer, the Red Comes Up, June Fool, the jean genie, Divestar, Miss Modern Age, melody nelson, jasper jenyk, new york doll and more that i am probably missing cuz my brain is so foggy (but you know who you are!). thanks for coming and making a girl like me so very happy.  photos from the show coming soon...   '),(14791,1694057,'24,September,2003','2003-09-24 00:00:00','   urlLink     like a broken record, i know.   '),(14792,1694057,'23,September,2003','2003-09-23 00:00:00','   urlLink  the realistics   get  urlLink bizarre .  come support the realistics at the bowery ballroom on oct. 3 -- it\'s their first headlining gig, and the show is 18+, for all of the kiddies crushin\' on that keyboard player.   '),(14793,1694057,'23,September,2003','2003-09-23 00:00:00','   who is  urlLink  leona naess  ? her voice is as soft and velvety as a rose petal, and her new self-titled album will bring out the cheeseball romantic in you. check: \"he\'s gone,\" \"calling\" and \"how sweet.\" but literally: She\'s the daughter of Arne Naess, former husband of Supreme Diana Ross. Born and raised in England but presently based in New York, Leona Naess is an introspective singer/songwriter whose alternative pop-rock draws on influences ranging from Edie Brickell, Tori Amos and Joni Mitchell to the Cure, David Bowie and Joy Division. -- allmusic.com   '),(14794,1694057,'23,September,2003','2003-09-23 00:00:00','   urlLink another \"which stroke are you?\" quiz  [nme]   '),(14795,1694057,'23,September,2003','2003-09-23 00:00:00','   urlLink bands against bush include sonic youth, liars and erase errata.    BAB has set up more than twenty regional chapters in the U.S. and abroad, with a directory of groups willing to participate in various events to raise voter awareness. Among the events planned is an October 11th international day of action, with concerts planned for New York City; Olympia, Washington; Seattle; Los Angeles; Boston; Washington, D.C.; Portland, Oregon; Paris, France; and Dublin, Ireland. The day of action will be preceded by a Rock Against Reagan Memorial show on October 9th.   [rolling stone]   '),(14796,1694057,'23,September,2003','2003-09-23 00:00:00','   urlLink   '),(14797,1694057,'22,September,2003','2003-09-22 00:00:00','   cds i\'m fawning/yawning over today:   liars  -  they were wrong so we drowned  * very experimental, but hot nonetheless. only liars can pull this off [fawn]   rufus wainwright  -  want one  * love his voice... but what\'s up with that \"vibrate\" song? [fawn]   leaves  -  breathe  * coldplay part deux from iceland [yawn]   '),(14798,1694057,'22,September,2003','2003-09-22 00:00:00','   tonight: acoustique: mary lorson the last town chorus jason trachtenberg the io\'s june fool @ sin-e  7 p.m. // $10   '),(14799,1694057,'22,September,2003','2003-09-22 00:00:00','   urlLink finally someone agrees with my opinion that  meg white of the white stripes  is boring!     Seeing the duo in concert is like watching a gifted artist cut a demo. It\'s as if he says to Meg, \"Hey, can you just keep time for me while I get this stuff down on tape?\"  ...The songs blurred like the years and spoken words slurred into one another, and the whole thing slowly turned into a bunch of electro pseudo-blues gobbledy gook.   ...Meg\'s simplistic skin-pounding, her child-like delivery and coy posturing resemble a stunt more than they suggest unique technique or talent. Even the White Strpes\' positioning onstage -- Jack frequently singing incoherently into a mike behind the drums, Meg playing directly to him at all times -- suggested as much. The only thing that made her a better choice than a drum machine was her partner\'s ability to control her tempo without physically pushing a button. Well, she\'s cute, too.   [rolling stone]   '),(14800,1694057,'22,September,2003','2003-09-22 00:00:00','   spoiler tee-shirt tagline contest from  urlLink drewbacca : Hey hep cats,  SPOILER 3 is just a month away, and one of the myriad tasks that must be done between now and then (apart from actually finishing our films), is that t-shirts must be made!  The t-shirt, as we see it, will consist of the SPOILER logo and a tagline.  Rather than stick with the boring \"A Comedy Film Project\" tagline, we thought we should drum up a new one for the shirts.  Something with moxie and, preferably, razmatazz.   We\'ve come up with a few ideas already, but thought: hell, why not get as many ideas as we can?  So, without further ado, I\'d like to announce the SPOILER T-Shirt Tagline Contest!  Email us your ideas for the SPOILER t-shirt tagline by the end of the day WEDNESDAY, and if yours is the lucky winner, we\'ll send you a $20 gift certificate to Amazon, and a SPOILER t-shirt.  That\'s right: you can have your bullshit immortalized on a t-shirt, AND get that DVD of Daddy Day Care you\'ve been salivating over.  For those who don\'t know what SPOILER is, please check out our website at:  urlLink http://www.spoilerfilms.com   Cheers,  Drew  urlLink spoiler@drewbacca.com   '),(14801,1694057,'22,September,2003','2003-09-22 00:00:00','   what do you think of the keyboards in  urlLink the new  strokes  song, \"12:51?\"  is it a lil\' electro-clash or what?   '),(14802,1694057,'21,September,2003','2003-09-21 00:00:00','   kudos to:   urlLink  Bangin\' Brooklyn   for her gawker mention. you go, girl! sex, drugs, and heavy metal karaoke prevail!   urlLink  the fever   for landing that opening night gig for the cmj music marathon...   urlLink  jasper   for this  urlLink fiery photo . lovely...   '),(14803,1694057,'21,September,2003','2003-09-21 00:00:00','   urlLink  The Darkness  reviewed in  The New York Times .   '),(14804,1694057,'21,September,2003','2003-09-21 00:00:00','   Save the Date:  Thursday, September 25 On the Moon Music Presents...   On!Air!Library! Dirty on Purpose Pela Say Hi To Your Mom   @  urlLink Sin-e  148-150 Attorney Street 8 p.m. // $8   '),(14805,1694057,'19,September,2003','2003-09-19 00:00:00','   my fave  urlLink  queer eye   quote to date: \"shake it! don\'t break it! it took your mama 9 months to make it!\"  when are they going to make bravo on demand so i can watch queer eye any day of the week?   '),(14806,1694057,'18,September,2003','2003-09-18 00:00:00','   from \'burg to \'burgh   my friend stephanie who lives in pittsburgh wrote me this email: hey, so, it looks like some of your new york bands are finally making it here to the other burgh (you know the one that\'s not williamsburg). last night i saw stellastarr* here and quite enjoyed the show. next weekend the fever is coming again, and i really wanted to see that show, but i\'m going to be out of town that weekend! maybe someone cool will be in d.c. then. anyway, you should keep me updated on who\'s good and shit so i\'m informed when bands come to town. also, if you have any connections to any of these guys that i could get in on, that would be cool too.   '),(14807,1694057,'18,September,2003','2003-09-18 00:00:00','   \"The city itself and the scene feels jaded now. I can remember seeing Interpol a few years back and there was hardly anyone there. No one really cared about local bands back then. You could never get a gig. The only way people would book you is if you could bring an audience. Now a NY band can go from opener to headliner in a month and after three months they have a recording contract and a manager. It\'s crazy. It seems so hyper-aware of the outside attention. The heartfelt passion and the authenticity to just f*cking rock out just because you need to,  with your friends. I think that\'s really tainted.\" --  Nick Zinner  [from annabiotica] Hmm, wonder what bands he\'s talking about?   '),(14808,1694057,'17,September,2003','2003-09-17 00:00:00','   damn, panthers are opening for the darkness tonight, and it\'s completely sold out! bad for me because i don\'t have a ticket, but hella good for them!   '),(14809,1694057,'17,September,2003','2003-09-17 00:00:00','   urlLink derek  reports on the walkmen\'s west coast tour. i love this photo:     yo, derek, we need more photos! more more more!   '),(14810,1694057,'17,September,2003','2003-09-17 00:00:00','   urlLink ooooh new  kylie  album on february 10th!  [billboard]   urlLink new  dido  song, \"white flag\" ... this sounds like all of her other songs, but i still like it. shuddup. [rolling stone]   urlLink join the  slate  book club.  are they going to be like the next oprah?   urlLink  har mar  gets  karen o  to get all jiggy on our ass.  \"Plans for a collaboration with Har Mar are in the making when we both find the time,\" she said. \"I think itâ€™ll be a little more R&B than the Yeah Yeah Yeahs usually are!\" [nme.com]   '),(14811,1694057,'17,September,2003','2003-09-17 00:00:00','   whatever.   '),(14812,1694057,'17,September,2003','2003-09-17 00:00:00','   urlLink breadsoul  cappuccino is awesome.   '),(14813,1694057,'17,September,2003','2003-09-17 00:00:00','   last night, annabiotica and i rushed downtown to the knitting factory to see three of our favorite guys in the new york rock scene:   urlLink jude, luke and dAve  . when we got to the box office we found they annointed us with nicknames on the guest list: anna \"kicking ass and taking names\" barie and uh, jinners. the box office guy smiled and asked, \"how am i supposed to ID you gals?\"   we snuck into the green room and greeted the boys, thanked them for putting us humble and po\' girlies on the list. then minutes later,   urlLink qatsi   took the knitting factory stage and totally kicked ass on that tiny stage (was it always that itty bitty?).   this new york trio gets tighter and tighter every time, and they are a joy to watch on stage. dAve really gets theatrical on stage... at times it seems like playing guitar is musical masturbation for him, and he continually flirts with the crowd. we wanted more luke (aka more drums!) \'cuz he\'s a giant drummer in more ways than one, and we wanted the music to pound more. and jude\'s voice was in stellar condition as he got all hot and sweaty under those stage lights. and we ate it up when he jumped off stage for that song we can\'t remember the title of. what\'s the title of that song???     ***  anyhoo, a coupla my friends work for the PR company that represents narciso rodriguez. isn\'t his  urlLink spring 2004 collection  totally hot? not very indie or hipster, true, but if you walked into a club wearing  urlLink these numbers , the rock stars would be drooling all over you, fo sho.  ***    why are there praying mantis on the subway? it freaked everyone on the train. nature gets too close to us nyoo yawkers again!   ***  happy belated birthday, cameron. may you find your gravy train!   '),(14814,1694057,'17,September,2003','2003-09-17 00:00:00','   urlLink the new york times asks, \"why do we like music so much?\"  i think it\'s cuz we\'re aching to connect with each other.   '),(14815,1694057,'16,September,2003','2003-09-16 00:00:00','   it was heartwarming to see   urlLink bishop allen   take the stage at the bowery ballroom, under the red and blue glow of the stage lights... the cheering fans up front, us loyal followers who have been waiting for the group to return from their ambitious, long summer tour. and finally we get our fill of a band that\'s totally changing the sound of brooklyn. and they are getting lots of radio play on kexp.org, too, so i\'m totally siked for them. john in the morning knows what\'s up!  they played my favorites and did a cover of talking head\'s \"psychokiller.\" bonnie says they\'ve been playing that cover all summer -- they did a great job of giving the song their own twang. some of my fave bishop allen tunes include \"quarter to three,\" \"eve destruction,\" \"busted heart,\" \"another wasted night,\" \"little black ache,\" \"charm school\" and of course, \"things are what you make of them.\" yep, they played all my faves -- though i have to say i love ALL of their songs. they are the shiznit, and you betta recognize. and a special hello to new bishop allen member christian o, the girl with mad skills on the xylophone.  after their show i chatted for a while with  urlLink a great photographer , and then the band and blake downstairs. then i pulled myself away to head over to vicious at sin-e. talked with jasper and audrey. said hello to luke, yannick, bill jr., hugged dennis, heard the candy butchers canceled the bowery show (fuckers!), chatted with some asobi seksu, wanted to see the fiery furnaces, but decided i couldn\'t cuz i was so tired. so tired! so now, folks, i\'m off to bed.   '),(14816,1694057,'16,September,2003','2003-09-16 00:00:00','   there is so much to do tonight! i\'m going to be running from a party at serena\'s to a couple of shows in the LES. come join me and rock out with:   urlLink  bishop allen   @ bowery ballroom 9 p.m. // $15   vicious: the fiery furnaces  @ sin-e 11 p.m. // $8  also... panthers/a.r.e. weapons @ lit 9 p.m.   urlLink plus, IM is all about flirting and gossiping.  not working. duh! [from fifi]   '),(14817,1694057,'11,September,2003','2003-09-11 00:00:00','   i really liked the kings of leon after i saw them live... more on their set later (they were joined by ben kweller on stage for one song)....   the kings of leon afterparty was downstairs at the bowery after the show. it was pretty crazy. i met ben kweller who has a mighty grip. the kings in the flesh. i saw julian and the rest of the strokes. i stood next to nick valensi for part of the time during the kings set in the balcony. i met gina gershon. it was an insane night of meeting people. and now my feet hurt.  here is the one photo i got of julian and the back of nikolai. i kept wanting to take a formal photo of julian and other rock stars, but he sensed that and ran away downstairs. har har. whatever -- i didn\'t want to be like the papparrazzi around us. that was actually quite sickening.   i have come to hate the papparazzi and thank god i am not a celebrity! flash flash flash. how annoying to stand there talking and feeling like you\'re going to have an epileptic seizure. this one guy seemed so normal. he came up when we were talking to gina gershon and he acted like he knew her. gina was like, \"when have we met?\" and he was like \"at some premiere for some film, i\'m sure.\" then he asked her something like \"did you make out with some chick and could you recreate that moment?\" d\'oh, it was papparazzi. he kept saying it as we walked away. shit like that must happen to celebs all the time. i was so weirded out by the experience. also in attendance were liv tyler, drew barrymore who was making out with fab quite a bit, that mooney suzuki guy, the donnas, jet, and a host of other rock stars i\'m sure i didn\'t recognize.   along the way i ran into justin from bishop allen. bloggers. nina and matthew. jco! and the red comes up. more news later...   and now a moment of silence for 9-11.  apparently, there are bloggers including me in this mtv you hear it first  urlLink stellastarr* piece . co co co! i learned from reading hard glitter and listening to dennis that the realistics will return to the bowery ballroom oct. 3 in support of the candy butchers (that guy who sang \"that thing you do\"), jonny lives and the flesh. erik -- that\'s all you!   '),(14818,1694057,'10,September,2003','2003-09-10 00:00:00','   for bidi:  urlLink free  david bowie  concert - sept. 18 at 7 a.m. at Rockefeller plaza.  aren\'t you glad you live in the city now?   '),(14819,1694057,'08,September,2003','2003-09-08 00:00:00','   The Boggs  to play Tiswas on Sept. 20. come out and support! slip on your new heels and dance!   congrats to marisa b on her illustrious new gig!   urlLink LME  has a new home. check it. and then help mike d. redesign it! he\'s got some hot pics of the stills and the prosaics...  why is everyone posting that Britney  RS  cover. stop it, please. doesn\'t  RS  get enough press for that bad cover choices?   '),(14820,1694057,'08,September,2003','2003-09-08 00:00:00','   urlLink TMN party in the \'burg.  HO!   '),(14821,1694057,'08,September,2003','2003-09-08 00:00:00','   Tell your friends, family, clients, neighbors, anyone you think would be a contender for this show.    CASTING CALL!   Saturday, September 13, 11am - 3pm  Jai Rodriguez will be in attendance noon - 1pm   Paramus Park Mall  700 Paramus Park  Paramus, NJ 07652  201.261.8000  www.paramuspark.com     We\'re out to make over the world one straight guy at a time, starting with the NYC metro area!   Will you be next?   We\'re looking for straight guys, both single and married, of all ages and backgrounds to help become more stylish, more cultured and more desirable straight men.    If you live in the greater New York City area and have a big event coming up where you want to reveal a \"new and improved\" you to your wife, girlfriend, family, friends or co-workers, download and complete our preliminary candidate questionnaire    and email it with a recent digital photo to info@thequeereye.com  .   Note: candidates must be real people - no actors will be considered   Maria Zoccolillo  Manager, Bravo Marketing  NBC Cable Networks  818-840-2534   '),(14822,1694057,'05,September,2003','2003-09-05 00:00:00','   urlLink strokes and the kings of leon tour!  the strokes are promoting their upcoming release  Room on Fire , in stores Oct. 21. the kings of leon, who have a very strokes-ish buzz about them, are hyping their debut LP,  Youth and Young Manhood . can\'t wait to see them at the bowery -- that sept. 10th show is completely sold out, btw. hope you snagged your tickets early.   quote of the week: Britney Spears: \"I think we should just trust our president in every decision that he makes and we should just support that.\" addendum: cuz i can\'t think for myself. it hurts my wig.   '),(14823,1694057,'31,October,2003','2003-10-31 00:00:00','   saintface     the national   happy birthday, starfoods! thanks for being a moon-er!   '),(14824,1694057,'30,October,2003','2003-10-30 00:00:00','   jenna is awesome. that is all.   '),(14825,1694057,'27,October,2003','2003-10-27 00:00:00','   december, baby    i could have gone to see the strokes/kings of leon show, probably hanging somewhere in VIP (through work, of course, not jinners) but my boys and my girl were playing a show at the mercury lounge, and like a good manager, i went and saw one of the coolest nights of rock i\'ve seen in a long time, including cmj.      i am so so so very proud of  urlLink  dirty on purpose   for pulling off a spectacular, amazing, wonderfully, exhilarating set last night, pulling in the second most people of the night, second only to the headliners,  urlLink  the decemberists  . i have a major crush on colin meloy, who reminds me of colin firth. i want to travel to portland to see this band play again!! dirty on purpose did a fucking awesome, tear-jerking cover of \"summer of 69\", and i was totally floored.       urlLink  earlimart   also blew me away with a robust, energetic performance that was both beautiful and whimsical.  urlLink  pela   performed probably their most perfect set yet, playing all of my favorites! i can\'t believe that every band on the bill tonight was FUCKING AWESOME. it was a totally sold out event, packed to the brim on a wednesday night, days after cmj madness. how cool is that?   '),(14826,1694057,'27,October,2003','2003-10-27 00:00:00','   post cmj goodness   october 29  the decemberists earlimart dirty on purpose pela @ mercury lounge   october 30  the national saintface @ mercury lounge   october 30  liars  young people @ bowery ballroom   '),(14827,1694057,'26,October,2003','2003-10-26 00:00:00','   tonight: last night of luxx  come party with free beer and free vodka.  in williamsburg   '),(14828,1694057,'25,October,2003','2003-10-25 00:00:00','   cmj 2003: friday  and yes, i missed all the cool events on saturday, including  brendan benson  replacing  earlimart  at mercury lounge. fuck. and anna tells me that  we ragazzi  was amazing. shit. and i missed an opportunity to hang out with the talented and dapper caz from  pas/cal . crapola. well, there\'s always next year, right? here\'s what i saw on friday, as i promised.  no photos of jasper\'s exhibit and elbow at pianos because i was watching said show on a crappy projected screen. boo hoo!!     jason darling      pas/cal      gravy train      slumber party      hella      the gossip      ropstyle from semiautomatic      slim and jin moon!   '),(14829,1694057,'25,October,2003','2003-10-25 00:00:00','   saw jasper\'s photo exhibit at pianos. elbow from a projected screen. afterwards pas/cal, gravy train, slumber party, hella and the gossip. photos to come. i\'m going out of town so i\'ll have to hear about saturday\'s cmj events from others!   '),(14830,1694057,'24,October,2003','2003-10-24 00:00:00','   cmj 2003: thursday  i made it to even less last night. oy!     colin meloy of the decemberists      steve reynolds      all night radio      fiery furnaces      british sea power   at this point, jenna and i were planning to see the explosions at cbgbs, but we soon learned the venue was no longer honoring badges. we met up with steve o and headed for food. ended up at the milkshake company on st. marks. oh my lord. that was a delicious grilled cheese sandwich.   '),(14831,1694057,'23,October,2003','2003-10-23 00:00:00','   cmj 2003: wednesday  ok i didn\'t make it to everything on my list, but here\'s what i did catch...     dresden dolls      mommy and daddy      black car      lovedrug      the killers      menlo park      prosaics   '),(14832,1694057,'22,October,2003','2003-10-22 00:00:00','   De-Luxxed This Sunday night the 26th is the last night of Luxx. There will be FREE BEER and cheap drinks all night long. It starts at 8 and ends when you leave. Lets rip the place to shreds! bands, djs, your ass. suprises....   Bluebird Pearlene Lion Fever From Monument to Masses,  Miracle Chosuke   urlLink     r.i.p. elliott smith, a wonderful needle in the hay.   '),(14833,1694057,'21,October,2003','2003-10-21 00:00:00','   urlLink listen to the shins new album, chutes too narrow   '),(14834,1694057,'20,October,2003','2003-10-20 00:00:00','   revised more-realistic cmj schedule:   Wednesday  Noon Pick up pass at NY Hilton Hotel 1 p.m. Turin Brakes @ NY Hilton Hotel  2 p.m. The Stills @ NY Hilton Hotel  7 p.m. Dresden Dolls @ Mercury Lounge 8 p.m. Mommy and Daddy @ Sin-e 10:15 The Killers @ Don Hill\'s 11:45 The Prosaics @ Don Hill\'s 12 am My Morning Jacket @ Webster Hall   Thursday  3 p.m. Elbow @ NY Hilton Hotel  4 p.m. Trachtenburg Family @ NY Hilton Hotel  6 p.m. Cocktail Party @ Lit 7 p.m. Stellastarr* @ Virgin Megastore 9 p.m. British Sea Power @ Bowery Ballroom 10 p.m. The Thrills @ Mercury Lounge 11:45 Mates of State @ Knitting Factory  12:15 Folksongs for the Afterlife @ Living Room  12:45 Rainer Maria @ Knitting Factory   Friday  1:15 p.m. Jet @ NY Hilton Hotel 2 p.m. Soviet @ NY Hilton Hotel 4 p.m. Joggers @ NY Hilton Hotel 8 p.m. Pas/Cal @ Living Room  9 p.m. The Rapture @ Roseland Ballroom  9:45 Rogers Sisters @ Knitting Factory  12 am Ted Leo & The Pharmacists @ CBGB  12 am Mink Lungs @ Pianos   Saturday  3 p.m. Vexers @ NY Hilton Hotel  8 p.m. The Eaves @ Lit Lounge   9 p.m. Panthers @ Knitting Factory  9 p.m. Echo and the Bunnymen @ Webster Hall 10 p.m. The Koreans @ Don Hill\'s 10:30 p.m. Dance Disaster Movement @ Knitting Factory  11 p.m. Earlimart @ Mercury Lounge   12 a.m. Kaito @ Mercury Lounge   1:00 a.m. The Decemberists @ Luxx  1:15 a.m. The Gossip @ Knitting Factory   '),(14835,1694057,'19,October,2003','2003-10-19 00:00:00','   are you still in love? in addition to their many cmj shows, to celebrate the release of their cd,  logic will break your heart ,  urlLink  the stills   will play at the vice store in new york city around 7 p.m. tomorrow.  and don\'t forget this hot show coming up:  October 29 the decemberists, earlimart dirty on purpose, pela  @ mercury lounge   '),(14836,1694057,'19,October,2003','2003-10-19 00:00:00','   urlLink the rapture interviewed in cmj.    i have figured out my tentative ambitious-as-hell cmj schedule with jenna:   Wednesday  Noon Pick up pass at NY Hilton Hotel 1 p.m. Turin Brakes @ NY Hilton Hotel 2 p.m. The Stills @ NY Hilton Hotel 7 p.m. Dresden Dolls @ Mercury Lounge 8 p.m. Mommy and Daddy @ Sin-e 10:15 The Killers @ Don Hill\'s 11:45 The Prosaics @ Don Hill\'s 12 am My Morning Jacket @ Webster Hall   Thursday  3 p.m. Elbow @ NY Hilton Hotel 4 p.m. Trachtenburg Family @ NY Hilton Hotel 6 p.m. Cocktail Party @ Lit 7 p.m. Stellastarr* @ Virgin Megastore 9 p.m. British Sea Power @ Bowery Ballroom 10 p.m. The Thrills @ Mercury Lounge 11:45 Mates of State @ Knitting Factory 12:15 Folksongs for the Afterlife @ Living Room 12:45 Rainer Maria @ Knitting Factory   Friday  1:15 p.m. Jet @ NY Hilton Hotel 2 p.m. Soviet @ NY Hilton Hotel 4 p.m. Joggers @ NY Hilton Hotel 8 p.m. Pas/Cal @ Living Room 9 p.m. The Rapture @ Roseland Ballroom 9:45 Rogers Sisters @ Knitting Factory 12 am Ted Leo & The Pharmacists @ CBGB 12 am Mink Lungs @ Pianos   Saturday  3 p.m. Vexers @ NY Hilton Hotel 8 p.m. The Eaves @ Lit Lounge 9 p.m. Panthers @ Knitting Factory 9 p.m. Echo and the Bunnymen @ Webster Hall 10 p.m. The Koreans @ Don Hill\'s 10:30 p.m. Dance Disaster Movement @ Knitting Factory 11 p.m. Earlimart @ Mercury Lounge 12 a.m. Kaito @ Mercury Lounge 1:00 a.m. The Decemberists @ Luxx 1:15 a.m. The Gossip @ Knitting Factory   '),(14837,1694057,'19,October,2003','2003-10-19 00:00:00','   the strokes & the rapture: new york rock\'s latest hangoverâ€¦    Albums by the Strokes, the Yeah Yeah Yeahs, the Walkmen, the Liars, Oneida, the Ex Models, Enon and Animal Collective and the collection \"Yes New York\" proved that the music still sounds exhilarating offstage. But now comes the hard part. Every celebrated local scene has a half-life, from Detroit and Memphis in the 1960\'s to New York City and Athens, Ga., in the 1970\'s to Minneapolis in the 1980\'s to Seattle in the 1990\'s. To survive beyond the initial buzz, the music has to point someplace new.   [ urlLink new york times ]  yay!  urlLink  fiona apple   is about two weeks from finishing her next album. the album is slated for a february release.   Among the songs guaranteed to make the final cut of the yet untitled 10-song LP are \"Extraordinary,\" a Tin Pan Alley-esque blend of Tom Waits and Vaudeville, and \"Better,\" an OutKast-like deluge of beats.   [ urlLink rolling stone ]   urlLink  fischerspooner   go wild on  #2 .   The new wave/electro sound of their debut is still intact, but Fischer said the album will also feature songs like \"Downup,\" which he described as an \"anthemic shoegazer\" track that mixes influences such as Jesus and Mary Chain and My Bloody Valentine with a dash of the Ronettes.   [  urlLink rolling stone ]   '),(14838,1694057,'19,October,2003','2003-10-19 00:00:00','   i saw  lost in translation  and wasn\'t as impressed as everyone else seems to be. sofia coppola had the potential for an interesting story, but she didn\'t quite develop anything too much. but the movie was beautifully shot. i really liked the mood of the movie, but it dragged at times without ever really going anywhere. my favorite scene was indeed the karaoke scene.  i also went to see  kill bill  for the second time this weekend. it was just as great the second time around. i am totally into the gogo character. anna tells me that this is the same girl from battle royale. i heard that quentin tarantino saw gogo in battle royale and decided to cast her in this role. how awesome. i am dying to see this flick. must surf ebay.  drank pumpkin ale and saw  willard  tonight with george, dj, gary and joe. that movie was terrible. awful. patooey! joe totally called the fact that crispin glover sings \"ben\" at the end. it was a movie-filled weekend, where music hardly played a part. very rare in my world. but i guess it will all balance out after a crazy fun filled week of concerts during cmj. woosh!   '),(14839,1694057,'17,October,2003','2003-10-17 00:00:00','   urlLink choose your own new york   i got: You love the Rolling Stones, so you decide to walk over and get his autograph. At the same time, you think you\'ll try to alter his destiny. Maybe you can make use of this time-travel thing by convincing him to change his self- indulgent ways, altering the course of history and preventing his eventual evolution into a scary lizard-man. Unfortunately, just as you are on your way over to talk to him, Mr. Jagger\'s delivery of cocaine from Bogota unexpectedly arrives, burying you alive in an gigantic avalanche of Colombian blow!You gasp for air in vain in the sea of nose candy. As the cocaine enters your bloodstream and you begin to lose consciousness, you notice the faint but unmistakable melody of Bee Gee\'s \"Night Fever.\" This is the best nightclub ever!   '),(14840,1694057,'16,October,2003','2003-10-16 00:00:00','   The pile of CDs sitting in front of me right now includes: erase errata, the fiery furnaces, the dandy warhols, pela, dirty on purpose, regina spektor, mogwai, the sleepy jackson, pulp, the darkness, spiritualized, scene creamers, the raveonettes, rufus wainwrights, the rentals, stellastarr*, liars, and damien rice. and those are just the CDs in cases. I need a better system of organization.  joe, here\'s a haiku made \'specially for you: talent is never wasted on your deft fingers so play us a song.   '),(14841,1694057,'16,October,2003','2003-10-16 00:00:00','   OK, I didn\'t watch the Yankees-BoSox game. I did however learn how to crochet (thanks, shara!!), and I also went to see  urlLink Bishop Allen   last night at the Mercury Lounge. If you didn\'t know, they changed the lineup a little â€“ put Jack on drums, new girl C.O. on keyboardsâ€¦ I like the new combination a lot. What I find particularly appealing about this group is the fact that they are so full of character and quirks, and write intelligent songs-stories about being young, in the city, in and out of loveâ€¦ with sense of humor and heart. Those are the qualities that charmed me when I first saw them perform in their loft a year ago, and those are the qualities that still bring a smile to my face when I see them now. That and, oh, Justin\'s hokey pokey dance moves. Adorable! Hehe. And yes, I did love their new song: \"Making friends is so damn easy, Just smile and act all pleasing!\" Classic.  Bishop Allen were followed by headliners,  urlLink  Metric  . I had to admit I was curious to see what this band was all about because so many people were singing their praises ( urlLink Sarah ,  urlLink Marc ,  urlLink Eric ,  urlLink Jo ). I thought that the lead singer, Emily Haines, had a great voice and an OK stage personaâ€¦ but then she started acting like she was on drugs -- totally bizarre and high as a satellite. she started doing jump kicks and shoulder twitches and hair flinging and making eyes with  urlLink fans . it was funny, but very distracting. so, in short, I almost liked metric. maybe I have to be on drugs, too, to enjoy it. or at least drizzunk. neither of which I wereâ€¦  the other night, I inadvertently went to see  urlLink  oxford collapse  . I was hanging out with my friend, saryn, who I haven\'t seen for ages, and we wandered into lit. I suddenly realized that OxC were playing downstairs, and we went down to watch their set. the boys have a new bassist to replace Yong Sing, though I don\'t know his name. mike started rockin\' out on stage like crazy, as the room sweltered in body heat and cigarette smoke. they were a lot harder than before, and their stage performance has improved quite a bit. if you haven\'t seen them, they are playing the vicious/crashin\'in cmj showcase at sin-e on October 22. go see them, along with  urlLink  mommy and daddy   (8 p.m.).   '),(14842,1694057,'16,October,2003','2003-10-16 00:00:00','   tonight: spiritualized @ north six $20   '),(14843,1694057,'15,October,2003','2003-10-15 00:00:00','   i cannot stop listening to this song:  urlLink the killers - \"mr brightside\"       urlLink read  what nme has to say about  urlLink  the killers  .     this band is also very hot,  urlLink  snow patrol  . listen to samples from their album on their site.   '),(14844,1694057,'15,October,2003','2003-10-15 00:00:00','   urlLink the strokes + damon albarn  = ???  does anyone have this track? please send it to me.   '),(14845,1694057,'15,October,2003','2003-10-15 00:00:00','   someone told me that the links to the photo galleries of  urlLink  the walkmen   and  urlLink  say hi to your mom   were not right. sorry! i have corrected the problem. now check that shit out!    urlLink the walkmen   urlLink say hi to your mom    shtym  are playing the mercury lounge on tuesday, october 21.   tonight:   urlLink  bishop allen    metric  mercury lounge 9:30 p.m.   '),(14846,1694057,'15,October,2003','2003-10-15 00:00:00','   urlLink listen to the whole stills record   '),(14847,1694057,'14,October,2003','2003-10-14 00:00:00','   urlLink  black car  , a really cool band that is playing the ASCAP CMJ showcase on Oct. 22, sounds pretty fucking hot. poetic. thoughtful. wondrous. listen to:  urlLink \"asleep at the wheel\"   urlLink \"st john\"   '),(14848,1694057,'14,October,2003','2003-10-14 00:00:00','   thanks,  urlLink jonathan purvis  for this lovely shot.   '),(14849,1694057,'14,October,2003','2003-10-14 00:00:00','   urlLink jennifer  does free research for you. i asked her a bunch of questions like: Who is taller, Nick from the French Kicks or Hamilton from the Walkmen?  Can dogs hear better than cats?  Is it true that you can\'t regain your hearing?  go and look at the answers. they are rad.   '),(14850,1694057,'14,October,2003','2003-10-14 00:00:00','   starting to make a list of bands/artists that contain my name: jin tha mc jinrai   '),(14851,1694057,'14,October,2003','2003-10-14 00:00:00','   \"the little birdie\" says that vice party has been scrapped... but with the release of the stills\'  logic will break your heart  coming up on october 21, don\'t count a surprise gig out of the question...   '),(14852,1694057,'14,October,2003','2003-10-14 00:00:00','   urlLink  conor oberst   sounds off on clear channel...   '),(14853,1694057,'14,October,2003','2003-10-14 00:00:00','   another  urlLink  rufus wainwright   album so soon after  want one ? how excellent! expect it possibly in the springâ€¦   Following the release of Poses in 2001, Wainwright hit a personal rock bottom in October 2002 after years of alcohol and pill abuse. He entered rehab and on the last day wrote \"Want,\" a song he says \"encapsulates something I realized, which is that it\'s all about the simple things.\"   That song helped open the door to a double album that was to bear its name, Want, but due to the record industry\'s lagging sales, the record was cut in half, leaving Want Two for future release.   \"I was listening to it as a double album set,\" Wainwright says, \"but the record company wasn\'t too keen. And also in listening to all of the stuff, I realized that no human mind is capable of really properly digesting that much overwrought emotion. It was just too much for one sit-down, and I realized that certain things would get lost. I tend to really demand for the listener to pay attention, and I didn\'t want them passing out.\"   Wainwright says the songs for the Want albums are among his most autobiographical, addressing subjects including his personal struggles, his show business life and his rocky relationship with his parents (singer-songwriters Loudon Wainwright III and Kate McGarrigle). \"I\'m not trying to be moralistic or tell other people what they\'re doing,\" he says. \"I wanted to go down on record, so no matter what happens in the future, I can always look back on this period as a sort of intense journey.\"  [rolling stone]  from the  urlLink  spiritualized   interview [filter magazine]:   Jason Spaceman is that illuminating science professor who blew your mind and taught you to dream. Jack White is the cuddly but ultimately unstimulating history teacher holed up in his office down the hall, reminding you to memorize all those presidents.   sure, The Yeah Yeah Yeahs, the French Kicks, Oneida, Rye Coalition, the Ex-Models, The Seconds, The Prosaics, The Narrator, Trans Am, Songs: Ohia, Josh Ritter and Liz Phair all have  urlLink Oberlin connections â€¦ and so do Folksongs for the Afterlife. donâ€™t forget them! [freewilliamsburg.com]  a little birdie told me that  urlLink  the stills   are playing a free show with andrew wk at an anti-cmj party by vice at that venue that used to be limelightâ€¦   urlLink alexander chats with nick zinner.  hey, whereâ€™s that interview you did with kenna?? [freewilliamsburg.com]   urlLink more  on that \"do daughters cause divorce?\" article [slate.com]     you asked for it (ok, maybe you didnâ€™t), but here it is anyway:  urlLink the jimi Hendrix  doll, complete with purple haze... just kidding.   '),(14854,1694057,'14,October,2003','2003-10-14 00:00:00','   had such a good time at vicious last night. if you haven\'t wished  jasper  a good one, please do so  urlLink now .   pela  were great as usual last night, and that \"latitudes\" song always gets me so revved up! and they are still some of the nicest guys in rock and roll.  asobi seksu  also were completely adorable and more rockin\' than i expected!  i can\'t wait until  dirty on purpose  play vicious in november. it\'s going to sizzle.   '),(14855,1694057,'14,October,2003','2003-10-14 00:00:00','   on the rhode  wow i had this big write up of dirty on purpose\'s trip to providence rhode island. but then something got fucked up, and it disappeared. arggggh. will have to muster energy later to re-write it... until then, check out the photo gallery.      urlLink rhode island photo gallery   '),(14856,1694057,'13,October,2003','2003-10-13 00:00:00','   urlLink  british sea power   interview on prefix magazine.  and... part two of the  urlLink  BSP  prefix magazine interview .   '),(14857,1694057,'12,October,2003','2003-10-12 00:00:00','   the tracklist for  urlLink  the walkmen  \'s upcoming album,  bows and arrows .  01 What\'s In It For Me 02 Little House of Savages 03 My Old Man 04 No Christmas While I\'m Talking 05 The Rat 06 138th St. 07 The North Pole 08 Hang On, Siobhan 09 New Year\'s Eve 10 Thinking of a Dream I Had 11 Bows & Arrows   '),(14858,1694057,'12,October,2003','2003-10-12 00:00:00','   urlLink  corn mo  is on mtv.com.   '),(14859,1694057,'12,October,2003','2003-10-12 00:00:00','   oh,  urlLink  miss modernage  , i will miss thee.     urlLink peter bauer of  the walkmen   has decided to play matchmaker. and it sounds like he\'s starting a blog of some sort! who else is going to their three knitting factory shows in one day on nov. 1? should be lotsa fun!  just came back from  dirty on purpose \'s show in rhode island. details and photos to follow!   '),(14860,1694057,'11,October,2003','2003-10-11 00:00:00','   urlLink  margaret cho   looked so hot last night. she wore this icecapades-like outfit that was brilliant, sparkling white. she looked so happy. this is her third touring comedy show, and i was happy to be seeing it. i\'ve seen all of them live, starting from \"i\'m the one that i want\" to \"notorious c.h.o.\" and now to \"revolution cho.\" her act has gotten considerably more political and socially conscious, which explains the show name. she talks about social inequalities, racism, crazy recent political events (aka gov. arnold).   she focuses less on the gory sexual details, which was prevalent in \"notorious,\" and more on how we all can start a revolution by recognizing the silly, \"sacred\" traditions we all have come to accept as normal. marriage does not have to be sacred between just a man and woman. the government shouldn\'t have to ask you to die for your country, but tell you not to talk about being gay. and my favorite is when people ask you what kind of asian you are (\"i can\'t tell any of you apart!\"). marg cho: \"uh, why do you have to tell us apart? are we going to be separated and put in internment camps?\" i couldn\'t stop laughing when she said that. it\'s so true! why DO people have to tell us apart? not even  we  can tell us apart!  she had so many new jokes and interesting perspectives to add to \"revolution,\" which made me happy because a lot of the material in her first two shows overlapped. \"revolution\" was packed with lots of new material that talks about her more current life experiences, and it\'s good to see that she\'s living in the past less and less and enjoying her present more and more.   go,  margaret cho !  her opening act, bruce daniels, was pretty funny too. he talked a lot about the same types of issues -- racism, social inequalities, stereotypes...  marg and bruce and some other chick (can\'t remember her name) did a rap at the end that wasn\'t very good, but i had to give her props on her effort. mc mc!   '),(14861,1694057,'10,October,2003','2003-10-10 00:00:00','   urlLink popmatters girl, devon powers, talks about music journalism.   '),(14862,1694057,'10,October,2003','2003-10-10 00:00:00','   urlLink i wish i had seen this band, peelander z. i think they recently played luxx. goddamnit. they look so fun!!   '),(14863,1694057,'09,October,2003','2003-10-09 00:00:00','   urlLink the national - \"slipping husband\"   '),(14864,1694057,'09,October,2003','2003-10-09 00:00:00','   urlLink dirty on purpose - \"spider eyes\" (live)   thanks, naf!    urlLink vhs or beta - \"disco paradise\"  i hear the asian guy in the band is singing now... cool!   '),(14865,1694057,'09,October,2003','2003-10-09 00:00:00','   the newly redesigned  urlLink  knife skills  . wow, nice!   '),(14866,1694057,'09,October,2003','2003-10-09 00:00:00','   urlLink laura is not only miss scenester (hot), she\'s also new york city\'s sound byte queen for music in general.  [ny post]   all i have to say is, if kid rock hates radiohead, they must be damn good.     urlLink  stellastarr*  on carson daly\'s last call on october 21. reserve your seat now.  oh man, don\'t they look so rock \'n roll?  more places to catch those stella starrs* In Store Performance @ Virgin Megastore in Union Square  BLENDER MAGAZINE & BEN SHERMAN PRESENTS: stellastarr* Thursday, October 23rd 7:30pm Virgin Union Square New York  Opening Reception of â€œ,The Robin Series,â€ select paintings by Shawn Christensen  â€œThe Robin Seriesâ€ 6-8pm @ Cassis  ***Free Wine, delicacies and with a Special Guest DJ  Cassis 52 Stone St. b/w William & Pearl 212.425.3663 Take 2/3 Subway down to Wall Street Stop (walk 2 blocks towards Hanover Square)  Music:  Return of the Rentals   '),(14867,1694057,'09,October,2003','2003-10-09 00:00:00','   school of rock the movie is better than most music today, says  urlLink  the lefsetz letter  . [from jon bahr]    damien rice beats out the yyys, cat power, sigur ros, interpol, the streets and more to win the short list prize, the shorty. i didn\'t  urlLink  damien rice   was  that  unknown. but i suppose it is true he doesn\'t get played on the radio... yet.  everyone who was nominated played except for the yyys and sigur ros.  Paul Banks, the typically taciturn singer of Interpol, said during the band\'s set that if he were in the audience, he probably would have considered the show the best he had ever seen.    urlLink eddie izzard gets sexie.   '),(14868,1694057,'08,October,2003','2003-10-08 00:00:00','   urlLink groupies vs. music journalists.  [annabiotica]  i was telling the knife skillers this last night, i could never date a musician. all the musicians i\'ve met are self-absorbed and needy. they need you to support their project, but hardly ever is the favor returned. but most of us just need to accept the fact that it\'s not a two-way street and move on in life as art philanthropists and that music karma will take care of things itself.   '),(14869,1694057,'08,October,2003','2003-10-08 00:00:00','   stopped by the mercury lounge to drop off copies of this awesome poster joe made for  urlLink  dirty on purpose  \'s next show -- they\'re opening up for  the decemberists  and  earlimart  on oct. 29. i saw modest babs at the merc, and remembered she started working there a few weeks ago. she\'s totally sweet, smart, and into good music. LOVE her.  dirty is also playing  urlLink vicious  on nov. 18. it\'s going to be totally hot.  to support some work peeps, i stopped by the sidewalk cafe for the \"writers at night\" series they have every month... but i was feeling kind of tired and decided to head home da \'burg.   anna was meeting up with fellow  urlLink  knife skillers   fon lin and jerry (turing machine), and so i decided to stop into union pool and say hello. of course, i ended up staying there for a little bit... but i just drank water. we had so much fun despite the weirdly, out-of-place romantic music.  after a lot of research on  urlLink  jason mraz  , i knew that he wore a red and white trucker hat every now and then... so i think i spotted him again at union pool tonight, hanging with some friends. but i\'m kind of blind, so i can\'t be 100% sure. it did look like him a lot.   '),(14870,1694057,'07,October,2003','2003-10-07 00:00:00','   i did the jason mraz interview, and he\'s a lot smaller in person. we were both so tired, we sat at the table and whispered to each other. he was nice enough. he had just gotten his hair styled because he was on the view this morning. michael moore was on it, and some stupid guest host that was on called  bowling for columbine  a staged mockumentary. of course, he got so angry, and an argument ensued. they quickly cut to starr jones who said, \"we need some music up in here. stay tuned for jason mraz.\" hahaha.   '),(14871,1694057,'07,October,2003','2003-10-07 00:00:00','   urlLink kenna\'s \"free time\" video.  nice kicks.   '),(14872,1694057,'07,October,2003','2003-10-07 00:00:00','   urlLink  kenna  is officially awesome.  i saw him perform last night, and he blew me away! the energy emanating from him and his band was just brilliant. his music reminded me of depeche mode. he told me people have told him he sounds like flock of seagulls on crack. HOT! and he wore a tee shirt for  the kills  on stage and wore  the the  and  johnny cash  pins on his jacket. how fucking cool is that? there were a gaggle of girls from fischerspooner (kenna just finished touring with them) screaming their support for him. i love when bands support each other like that. and outside of the club, kenna complimented my beat up chucks and gave me a big hug! he\'s my new crush. **sigh**  also i met this R&B artist who was wearing levis pins with  the realistics  and  the boggs . i traded my  dirty on purpose  pin for the boggs. he wouldn\'t give me the realistics pin because it was all beat up. how funny! i can\'t wait to start wearing it around. they are fabulous!  congrats to laura for her  urlLink scenester award .   '),(14873,1694057,'07,October,2003','2003-10-07 00:00:00','   the other night i met up with george from DoP and we headed over to the knitting factory to support doug marvin (DoP drummer) in his other band  an ideal for living . they completely took me off guard. amazing! i wasn\'t sure what i thought about the music in the beginning. it seemed kind of all over the place, but in the middle of the first song, it started to take shape and form into this awe-inspiring noisy, experimental set that featured double drumming and intense feedback. doug played both guitar and drums. this guy is such an impressive musician -- creative, dark and unpredictable.  finally saw  urlLink  hula  , too. it was the end of the night and i thought i was going to fall asleep during their set because i had heard it was mellow music... but i was far from sleep. their songs had such cool, groovy beats and notes that were pure and clean -- no distortion, no fancy pedal work, no feedback -- just good, clean music.   '),(14874,1694057,'06,October,2003','2003-10-06 00:00:00','   urlLink behind the lyrics to songs by coldplay, beck, violent femmes, and yes, sir mix-a-lot.    urlLink you can kill rock stars but you can never kill  kill rock stars ,  says slim moon.   urlLink carissa\'s wierd splits.    urlLink oh those strokes.    The problematic cover art has been revealed online-- nothing as racy as the Suicide Girls-ish UK sleeve for Is This It, but more in line with the colorful geometric dabblings of the last few singles. There are subtle references to gambling \'n\' guns (the four suits in a deck of cards, revolvers held aloft by both black and white hands, and an abstraction of a roulette wheel). There\'s no peelable fruit on the cover, if that\'s what you were asking.    urlLink r.e.m. covers interpol\'s \"NYC.\"   '),(14875,1694057,'05,October,2003','2003-10-05 00:00:00','   finally more photo galleries!   urlLink a slew of new photo galleries , including  dirty on purpose ,  deerhoof ,  semiautomatic ,  the walkmen ,  ex models ,  qatsi ,  say hi to your mom ,  pela ,  on!air!library! ,  kings of leon ,  jet ,  kaito  and  bishop allen !!!! surf to your heart\'s content!   '),(14876,1694057,'05,October,2003','2003-10-05 00:00:00','   any movie featuring scenes in north six and with a poster for  urlLink  the walkmen   in it deserves props from me.  school of rock  was cute, but what was up with the mooney suzuki cameo?   urlLink the drunk dictionary.  this dictionary translates male drunk talk:  guy says:  so, i\'m in a band...  translation:  please have sex with me because you think i\'m cool.   guy says:  i\'m pretty drunk. will you walk me home?  translation:  will you go home and have sex with me?   guy says:  sure, you get this round -- the next one is one me.  translation:  we won\'t be here long enough for another round, dumbass.   urlLink more pick up lines.  some of the ones that made me laugh were: 1) hey did you fart? because you blew me away! 2) i may not be the best looking guy here but i\'m the only one talking to ya! 3) ok i\'m here. what\'s your next wish?  and this one is for doug in DoP -- use it on annie: 4) Hi, my nameâ€™s Doug. Thatâ€™s \"God\" spelled backwards with a little bit of \"u\" in it.  need to break up with your boo? try these  urlLink brutal break up lines.  i like this one: it isn\'t you. i\'m in love with your brother.   hahahahaha!!   '),(14877,1694057,'04,October,2003','2003-10-04 00:00:00','   surfing cnn.com...   urlLink ouch.  poor roy of siegfried and roy gets mauled by a tiger -- in the neck and arm! oh what will siegfriend ever do without him? he\'s in critical condition.   urlLink poor frances bean.  first her dad commits suicide. then her mom is a crazy drug addict cracking people\'s windows and shit. what the fuck is wrong with these people?? i was talking to someone, and they were like, if you can\'t get it together for your kid, then you are totally beyond any help. courtney, get yourself together! frances bean is gonna be one fucked up child.  oh weird, i\'m now supposed to do a lunch/interview with jason mraz on wednesday. what will i ask him? when do i get to do lunch/interviews with julian of strokes fame? julian, swing by the office, why don\'tcha?   urlLink revamped billboards.   '),(14878,1694057,'03,October,2003','2003-10-03 00:00:00','   urlLink  queen latifah  signs to the new cool label, vector recordings.  ken and jack also own the company that handles that really hyped tennessee band,  kings of leon .  their reach is growing so quickly! when i went to the KoL show at the Bowery, I had heard that jack was looking for an assistant at his label. now I know what they were talking about! too bad I like my job and don\'t want to leave it!  the free  urlLink  dirty on purpose   show at  urlLink grand central bar  went better than we all thought it would. walking into the bar and realizing some people had to play around a pool table is kind of like a big train wreck, but only a band like DoP could pull it off and draw everyone in with their amazing music. it was beautiful. at many points, i didn\'t even feel like i was deathly sick from my miserable fall cold! oh, how their music warms me... after the show, the band was packing up their equipment and these guys across the street, who were throwing bottles of beer at people who i\'m guessing are their enemies, kept yelling, \"gentrification! fucking gentrification!\" i kept thinking, perhaps these guys should use their anger for better things than throwing beer bottles at people and yelling big words from the  urlLink dictionary . sometimes i hate gentrification too, but i don\'t go around screaming it in the middle of the night, do i? i mean, i can barely afford to live in my neighborhood, and don\'t feel like i can move because apartments are so friggin\' expensive now in wburg. but i certainly don\'t go around yelling at every hipster and hipsterette on the street. that\'s what crazy people do. they also try to kill people by hurling beer bottles. that was kind of scary, yo.  i\'m waiting to hear the reports of the realistics show at the bowery from the rest of the bloggerazzi. how was it, gals? what\'s the good word? i\'m sure the boys did a fantastic job of rocking that supercalifragilisticespialidocious venue.   '),(14879,1694057,'03,October,2003','2003-10-03 00:00:00','   a tip from  urlLink ed2010 : A little-known voting fact Ed learned last night: In order to vote in the New York State Presidential primary on March 2, 2004, you MUST register your party affiliation no later than Friday, October 10th. Do so by calling 1-800-FOR-VOTE or go to  urlLink www.elections.state.ny.us/voting/voting.htm . Do it today!   '),(14880,1694057,'02,October,2003','2003-10-02 00:00:00','   urlLink this is some funny shit on hbo.  it\'s called sketch pad, and you can\'t miss the super-dirty sketches. they are hilarious!   '),(14881,1694057,'02,October,2003','2003-10-02 00:00:00','   urlLink hamilton of  the walkmen  gets grumpier during a mess of a show.  poor kaito could not cross the maple leaf border.    Lead singer, Hamilton Leithauser seemed annoyed â€” angry even. He seemed to be trying to force himself to get into the show. What was up? Was this part of their act? The hipsters from NYC laying down some serious Nooo Yawk indifference for the crowd?   [link from vive/le/roi]   '),(14882,1694057,'02,October,2003','2003-10-02 00:00:00','   everybody loves the  urlLink planetarium . i think it\'s time to take a work field trip over there. anyone else in? it will surely make me feel like i\'m back in middle school, watching the laser light show that featured primus, nirvana and the red hot chili peppers. scary!   urlLink will child porn end file-sharing?  ew, gross. but i love how they have a photo of britney spears with this article!   duran duran  set a  urlLink fall tour  that doesn\'t include new york city. snobby bastards.   urlLink will the video ban damage indie filmmaking?  even though i graduated from tisch, i still don\'t know the answer to this question.   urlLink do daughters cause divorce?  what kind of stupid question is this? so what is the cause of this wacky hypothesis? anyways, we should all take some kind of survey, but i know a lot of guys with divorced parents. i can\'t remember if they have any sisters though...   urlLink a nausea-inducing hipsterette train wreck walks into odessa bar.   '),(14883,1694057,'02,October,2003','2003-10-02 00:00:00','   justin is officially superman  justin successfully pulled double duty, rocking out with  panthers  and sweatin\' with  turing machine  at lit last night. and guess what? the mofo is gonna do it again tomorrow night, playing plaid with turing machine at 11 and then running over to sin-e to play with the panthers at 12:30. how effin\' crazy, dude.       anyways, the show at lit was the going away party for turing machine\'s bassist, scott, who is leaving for mother england for a year. it was packed and stuffy. and fun. finally met marisa of sup and rheingold fame in the flesh. she told me she thought that i had been a boy this entire time. i guess i have a very ambiguous name!  we ran into a film crew who was making a documentary on a student filmmaker who is making a movie on the new york music scene. fon lin was mic\'d for most of the night so we kept saying stupid things into the mic.    totally random  earlier in the evening i was having drinks with christian, christian and justin of  bishop allen  at a bar called b-sides. randomly, i bumped into a guy named jay who had emailed me a couple days ago. he recognized justin first. i thought that was pretty cool. and yes, very random.   get dirty  if anyone is hanging in the wburg tonight, come see  dirty on purpose  play at the new bar opening up called grand central -- near grand street and graham avenue. they hit the stage at midnight. and it\'s freeeee!   or, get real   the realistics  are also playing at the bowery ballroom -- so go see those bodacious babes rock out in the pretty ballroom lights.   other interesting shite  omigod, like, marg cho has a  urlLink blog . copy cat! she talks about music, too --   \"My newest obsession is The Neptunes song \"Pop Shit\". I have never heard anything so beautiful in my life...\"   '),(14884,1694057,'02,October,2003','2003-10-02 00:00:00','   urlLink CMJ Music Marathon schedule.   '),(14885,1694057,'01,October,2003','2003-10-01 00:00:00','   From our favorite new filmmaker Andrew:  As you probably know, we\'re up and running on this new film project--   This coming Tuesday, the 7th, we\'re shooting what will presumably be by far the most complicated scene of this mostly uncomplicated film, a performance scene in which the makeshift supergroup of  urlLink Justin Rice  on guitar + Kevin Micka on drums plays a couple of songs at the Northsix club in Williamsburg. (Where, we are told, they shot a scene for School of Rock...?) 66 N. 6th St., near the Bedford stop on the L. We need extras, folks who look like they might conceivably come out to a rock show. We don\'t need a lot of folks, Justin + Kevin are supposed to be a pretty underattended opening act.   Most of you have jobs and cannot hang around a rock club from 10am - 5pm on a Tuesday, which is what we need, but if you\'re considering quitting your job, now would be a good time. What\'s in it for you, I guess is (a) we buy you lunch + there\'s some snacks around, (b) the glamour/creative process blah blah blah, (c) you get to see a free concert, sort of, albeit a concert in which the same two songs get played over and over again...It\'s an adventure, an adventure.   And if you know anyone who you think would be up for this, please do forward this along, I\'d appreciate it tremendously.  If you want to do it, e-mail me back asap and let me know. If you can\'t, no need to respond-- Thanks so very much, Very excited,  Andrew  ***  if anyone is interested,  urlLink email me  and i\'ll pass your message along!   '),(14886,1694057,'01,October,2003','2003-10-01 00:00:00','   urlLink how to be polite.    urlLink how to be rude.    urlLink how to be neutral.  but notice that quirky political editorial at the bottom. that\'s definitely not neutral.    is anyone else scared by freaky photos of  urlLink  arnold schwarzenegger   on the front page of  the new york times  today? is he really going to win???   '),(14887,1694057,'01,October,2003','2003-10-01 00:00:00','   while going to see qatsi play last night at pianos, i saw another pretty amazing band called  urlLink  read yellow  . they were throbbing noisy punk rock with lots of sharp guitar licks and hard drum kicks. they are from amherst, mass., and i think they\'re gonna blow up, considering their next couple shows are with mission of burma and radio 4 -- a pretty impressive move, though i\'m not sure their music gels with those bands -- they\'re much harder. listen for yourself:  urlLink \"the association\"  and  urlLink \"fashion fatale\"    urlLink  qatsi   also brought the house down last night. i\'m pretty sure dave was playing guitar between some girl\'s legs at one point. they always bring something a little new to their performance. those guys are awesome. if you haven\'t heard them, here\'s a song:  urlLink \"City Son\"   and i wish i could go to sleep tonight, but there is so much going on!! if anyone is around the east village, they should stop by lit and see  turing machine  -- it will be their last show for quite a while as their bassist is going abroad for a while. justin will be pulling double duty as he will rock and sweat in both  panthers  and  turing machine . that guy is a maniac! oh yeah, there\'s open bar from 9 to 10.   '),(14888,1694057,'01,October,2003','2003-10-01 00:00:00','   what i\'m listening to -- mp3s from  the decemberists :  urlLink \"the soldiering life\"   urlLink \"Here I Dreamt I Was an Architect\"   more kickass kill rock stars and their songs:  urlLink  semiautomatic  - \"this place does not exist\"   urlLink  the gossip  - \"fire/sign\"   urlLink  deerhoof  - \"gore in rut\"   urlLink  gravy train  - \"hella nervous\"   urlLink  the bangs  - \"i want more\"   '),(14889,1694057,'01,October,2003','2003-10-01 00:00:00','   oooh  urlLink korean people  on broadway!   '),(14890,1694057,'01,October,2003','2003-10-01 00:00:00','   urlLink strokes  room on fire  e-card . this includes a special  urlLink \"in the studio\" video .   '),(14891,1694057,'01,October,2003','2003-10-01 00:00:00','   happy rocktober!   '),(14892,1694057,'01,October,2003','2003-10-01 00:00:00','   stellastarr* \'s bowery ballroom show with the raveonettes  urlLink reviewed  in earlash by modest mint.   '),(14893,1694057,'30,November,2003','2003-11-30 00:00:00','   vomit launch   urlLink the new york times says that interns are the new hipsters.  every young person in america has at one point or another heard that getting an internship is a valuable and necessary way to get ahead. it is all about the people you know and the hands-on experience you have. when i was in school, i certainly took that to heed and did a bunch of internships. looking back, i think that all meant horseshit. not to say that my times at some of those major magazines, non-profits and other miscellaneous organizations weren\'t valuable at all, but they just don\'t mean much anymore. i just think about them fondly as experiences i had a long time ago.    so how were they valuable? all of those experiences cannonballed and led me to my current aspirations as less a writer and more a behind the scenes type in the music industry. working at major magazines made me realize that i don\'t like writing about things that are huge and commercial. and i don\'t really care about fashion and beauty as much as i enjoy  urlLink making fun of it . my enjoyment lies in the indie culture that is unfortunately fast becoming huge and commercial enough to be written about in the new york times.    that kind of commercialization makes me shy away from professing my love for bands like the strokes and pavement, and instead, professing my love for  urlLink smaller, up-and-coming bands . i get a rush from trying to change and grow with my community. thinking of ways to improve the community that is soon in danger of imploding on itself. it\'s like how malkmus said in the slow century documentary on pavement when their videos started spinning on mtv and a wholly different group of kids started coming to their shows. the fans and the band started to diverge, making gigs less enjoyable. suddenly they were playing for kids at lollapalooza, getting booed and mud thrown at them. hopefully living in the new york indie community won\'t become less enjoyable with its recent commercialization. [link from  urlLink tma ]   '),(14894,1694057,'26,November,2003','2003-11-26 00:00:00','   top 500 albums, my ass  as soon as i saw rolling stone\'s cover story of  urlLink the 500 greatest albums of all time , i sighed and said, \"let me guess, the beatles are number one.\" lo and behold, i see this article online and the fucking beatles are number one! and number three! and number five! and number ten! and number 14!   '),(14895,1694057,'25,November,2003','2003-11-25 00:00:00','   phantom planet  ew, i feel really weird saying this (especially because i am losing my fanatical ways), but i like this  urlLink  phantom planet   song \"big brat.\" i feel like i want to shoot myself because up until now, i thought i absolutely loathed phantom planet and their teenage fan club of haughty middle schoolers. listen for yourself here:  urlLink audio stream   in other news, i do NOT like  urlLink  the killers   song \"indie rock and roll.\" every time it spins in my cd player, i have to skip forward. and there are only four songs on this cd, so every 10 minutes it gets kind of annoying.   '),(14896,1694057,'25,November,2003','2003-11-25 00:00:00','   Dulli Noted   urlLink greg dulli interview on prefix   '),(14897,1694057,'25,November,2003','2003-11-25 00:00:00','   53 Things Eminem Uses To Wipe His Ass   urlLink check out the brilliant list on kittenpants , including MTV movie awards and Brittany Murphy\'s eyelids.   '),(14898,1694057,'25,November,2003','2003-11-25 00:00:00','   KaitO (UK)    urlLink kitty magik interviews  kaito .   Dave Lake, guitar/vocal:  The name KaitO comes from the name of a cat that Nikki had Which brings me to ask a question! Why if your magazine is called Kitty Magik are there no questions for us about cats???? I have 8 and am only too willing to answer any questions you have about cats!!!!    urlLink KaitO - \"Should I\" mp3   visit this \"top pop band\"\'s website:  urlLink  KaitO (UK) website   '),(14899,1694057,'23,November,2003','2003-11-23 00:00:00','   i fucking love margaret cho   can i just say that i love  urlLink  margaret cho  ? and i love her  urlLink blog . and i love that she just does whatever she wants and doesn\'t take shit from anyone. she\'s not out there trying to impress all the bling bling hotties at the vibe awards. she\'s just there, she\'s a comedian, she wears an outlandish outfit to get reactions and laughs from people. she\'s goofy. she\'s used to using her physical self into comedy routines.  urlLink so fuck all the haters!!   '),(14900,1694057,'23,November,2003','2003-11-23 00:00:00','   as real as it gets  dennis of  urlLink  the realistics   will be at  urlLink sin-e  on monday, doing his solo thang. shake ya ass on over to the LES and watch this talented man sing. i will crawl out of hibernation and see him. you should too. he steps on stage at 8 p.m. be there!   this weekend the L train was out of service. the whole fucking train!!??!! it\'s not breaking news that us williamsburgers get shafted with train service late nights and on weekends, but this is just getting completely ridiculous. not all was lost, however, as i stayed home for most of sunday watching  bridget jones\'s diary  with  and  without audio commentary. oh how i love mr. darcy. sigh.  oh, and i randomly found this  urlLink interview  with  urlLink  les savy fav  \'s tim harrington. the interview refers to my  urlLink  steal this indie look   on him! this reminds me that i have to update that section and another fun game i came up with last week. you are gonna love it!   '),(14901,1694057,'22,November,2003','2003-11-22 00:00:00','   the winter pageant  i said in a  urlLink previous post  that i missed seeing   urlLink the winter pageant   live. but at the last vicious i ran into mike, the guitarist from the band who graciously gave me the band\'s cd. it took me a couple days to work it into my crappy work computer rotation, but when i did i found songs as soft and pretty as gentle flurries during a pennsylvania snowfall. i will have to try and see them live when they play luna lounge on december 9. you should too.   '),(14902,1694057,'21,November,2003','2003-11-21 00:00:00','   franz ferdinand     urlLink darts of pleasure mp3   i\'m going to try to see them at mercury lounge with the fiery furnaces. list action is hawt!   '),(14903,1694057,'20,November,2003','2003-11-20 00:00:00','   what i have listened to thus far today   urlLink lovedrug   urlLink pavement   urlLink the killers   urlLink mommy and daddy   urlLink the winter pageant   urlLink the head set   urlLink dresden dolls   urlLink sukilove  brian christinzio   '),(14904,1694057,'20,November,2003','2003-11-20 00:00:00','   chi-town  might have to go to chicago for work in mid-december... in doing some research, look what i found:  urlLink daisy knows rock . she has  urlLink ypersound  listed so maybe they know each other. what are the cool spots in chi-town, peeps? what are some good music sites? what are some good venues? please, give me the skinny.   '),(14905,1694057,'20,November,2003','2003-11-20 00:00:00','   tim harrington is getting married!   urlLink omigod  from  urlLink catherine\'s pita    Met: May 1999 Engaged: Jan. 26, 2003 Projected Wedding Date: July 3, 2004  When Tim Harringtonâ€™s future mother-in-law first saw him perform in his punk-rock band, Les Savy Fav, she had a most unmother-in-lawish reaction. \"She was like, â€˜Heâ€™s the Messiah!â€™\" said his fiancÃ©e, Anna Schwebel.   \"Sheâ€™s European,\" Mr. Harrington explainedâ€”specifically, Swedish. A Web designer by day, he makes an impressive sight onstage, jumping around half-naked with his bushy golden beard and bristling eyebrows to match. \"I like to examine space,\" he said.  Ms. Schwebel, 28, an abstract painter who works under her motherâ€™s maiden name, Killander, first saw him perform at Tramps (a pal of hers was dating the bandâ€™s scruffy drummer). \"I was just like, â€˜Yeah, heâ€™s a kooky guy. Whatever,â€™\" she said. But Mr. Harrington, 29, was quickly smitten by the cherubic, ponytailed blonde, a Columbia grad. \"I thought she was stone-cold fox,\" he said.  They became friendly, and when Mr. Harrington had an empty spot in a five-person share in Williamsburg, Ms. Schwebel moved in. Ah, la vie bohÃ¨me â€¦. Mr. Harrington, simmering with desire, would sing Bad Companyâ€™s guttural \"Feel Like Making Love\" around the apartment, hoping sheâ€™d take the hint. But she remained oblivious until this woman Ivy began calling the house for him. Turned out Ivy was every New York womanâ€™s nightmare: a gorgeous Rockette. \"I got really jealous,\" Ms. Schwebel said. \"I hadnâ€™t understood how great he was, but then one day I woke up and was like, â€˜Oh! Oh my God, Iâ€™m in love with Tim!â€™\"   They began sharing a loft bed, eventually moving to a private apartment in a Williamsburg townhouse, and it sounds like theyâ€™re having a wild time behind closed doors. \"Heâ€™s really good at pretending to be animals. Really good,\" Ms. Schwebel said. \"Like Gorillas and tigers. And lions, too!\" Grrrrr.   Mr. Harrington consulted his female kinfolk for advice on how to propose. \"They gave me all these cockamamie ideasâ€”straight from the WB,\" he said. \"And I did them all.\"  But the chunky diamond ring he ordered was all wrong. \"It was like it belonged to the Bride of Frankenstein,\" he said.  And the week before they jetted off on a surprise trip to Paris, Ms. Schwebel broke her wrist snowboarding.  Dying to get the damn thing over with, Mr. Harrington dropped to one knee in the middle of the night at a Rive Gauche hotel when she slipped out of bed for some water, offering up the ugly ring and promising heâ€™d replace it with a nicer one later. Ms. Schwebel responded by erupting in giggles.   Theyâ€™ll be married at an inn near her familyâ€™s country house in Vermont. The bride will wear a Reem Acra gown she got from a sample sale and a new rose-gold ring with 10 small diamonds. The groom is writing a love song for the receptionâ€”something along the lines of \"Weâ€™ll Make a Lover Out of You,\" Les Savy Favâ€™s only existing love song. \"Itâ€™s about how we have enough love,\" he said, \"so that everyone else can get some of it, too.\" Groovy, man.   '),(14906,1694057,'20,November,2003','2003-11-20 00:00:00','   there\'s no place like texas  i found out today that i will be heading to  urlLink  sxsw   next year. of course, i will be working a lot of that time, but i will also have the opportunity to party a lot and see lots of great shows for free. the more i think about it, the more i\'m getting excited about it. hopefully,  urlLink  dirty on purpose   will join me. the band is getting ready to record some more songs this month, and we\'re all as excited as can be! everything seems to be moving smoothly, and everyone\'s working very hard.    vicious night   dop\'s performance at  urlLink  vicious   was good. i was nervous before they went on stage, but after the music began, all of my nerves melted. their songs washed over the crowded room like a blissful breeze of pretty atmospheric notes that shimmered. the set list was really exciting. they debuted a new song called \"mind blindness,\" and then they followed it with a kick-ass version of \"monument.\" i feel like they get stronger and stronger every time i see them. they probably won\'t have another show lined up until mid to late january, so to tie you over, buy a dop cd from  urlLink  truthflies.com  .    i was also impressed by  urlLink  scout  \'s performance. they had catchy songs that build nicely into poppy rock songs. i liked ashen keilyn\'s husky vocals and her tomboyish vibe won me over. i still have their song \"before you\" in my head. it\'s on their website if you want to listen to it! oh yeah, and boy could that guitarist do high kicks!    urlLink  ps   were also quite good, though their set felt really long. i really liked their one song that had the lyrics, \"i used to think you had a sense of humor.\" don\'t know the title of that one... that sensitive song stood out the most to me, but overall ps gave off a thom yorke-ian vibe that was less technical and full, but still really cool and heartfelt. i heard they used to have a cellist, and perhaps the music would have sounded more full with some strings.     urlLink  aerial love feed   also performed with their strobe lights and sunglasses intact. they definitely have a placebo vibe, complete with rock star poses and haircuts. their music throbs and bounces off the walls, but it\'s not really the kind of thing i\'m into, though i think technically they\'re a good band. sometimes i don\'t know why i am attracted to  urlLink some bands  more than  urlLink others . music is mysterious in that way. there\'s something inside of us all that draws us to different chords, different lyrics, different intonations. and that\'s why being in a band is such a gamble these days. who knows what will work and what won\'t?   however, the rocks, a highly charged band from london, apparently is fairly confident that rockin\' out, jumping on bars and falling off the stage is what they want to do. though i wouldn\'t say their music was overwhelmingly brill, their performance was quite outstanding and fun to watch. it certainly made me whoop and holler. it also seemed like most of their cohorts (including the lovely miss vivian from mommy and daddy) were in the audience supporting their english buddies. also the girl in the band had these phosphorescent yellow heels paired with royal blue stockings -- all with a dress that had the poofiest sleeves ever. joe, dj, erika, jenna, jason and i had dinner at the pink pony with the band\'s bassist who was very nice and charming. dj lent him some bass stuff (i know, very specific, eh?) and joe lent their guitarist some guitar stuff, and the band was quite happy to play with dop\'s equipment. i didn\'t mean that in a dirty way. haha. anyways, the bassist was leaving back for england the next day and couldn\'t play with the rocks at plaid, so vivian was subbing on bass for them. the drummer also couldn\'t play at plaid and kevin from on!air!library! had to sub on bass for them. anyone go to that show?   throughout the night  urlLink dennis realistic  played the best list of songs ever. he played the ronettes with the raveonettes and the realistics with old school kylie minogue. and the best new strokes song - \"the end has no end.\" so good! dennis is always full of energy and is the biggest sweetheart always. he also donned a tres cool fedora. love that boy.   '),(14907,1694057,'19,November,2003','2003-11-19 00:00:00','   from the darkness to enlightenment   seems like these days, everyone likes  urlLink the darkness . me? i\'m not sure. i\'m too busy trying to understand why all of a sudden everyone thinks bands that sound like  urlLink those hair bands that my heavy metal boyfriend listens to  are suddenly considered the next big thing after the death of electroclash and the slowdown of post-no-new-wave. granted, i do dig the justin hawkins\' falsetto, wavering voice, and i do like \"i believe in a thing called love\" and \"love is only a feeling.\" jesus, am i starting to like  urlLink the same bands  as my boyfriend??? this is freaking me out.  on a different note altogether, someone sent me this  urlLink scary link .   '),(14908,1694057,'19,November,2003','2003-11-19 00:00:00','   songs to start my day   urlLink bishop allen - things are what you make of them   urlLink the realistics - why didn\'t you stay?   urlLink the decemberists - shiny   urlLink joggers - hot autism   '),(14909,1694057,'19,November,2003','2003-11-19 00:00:00','   british sea power   urlLink an underwelming interview with british sea power.  hopefully, they\'ve got more to say when they\'re not about to step on stage and be all crazy... but here\'s a good  urlLink joggers review . [earlash]   '),(14910,1694057,'18,November,2003','2003-11-18 00:00:00','   meow  vicious was good last night, but now i\'m so friggin\' tired. i didn\'t get home until 3! will elaborate later, peeps, but it was good to see all of you. mwah!   '),(14911,1694057,'18,November,2003','2003-11-18 00:00:00','   walkmen walkmen walkmen     urlLink i came. i saw. i wrote. about the triple walkmen/roger sisters show at the knitting factory. check it!   '),(14912,1694057,'16,November,2003','2003-11-16 00:00:00','   kurt cobain comic book   The book combines biographical fact with a bit of fiction (perhaps it\'ll give readers a nice version of Courtney Love?), and chronicles everything from Cobain\'s childhood and Nirvana\'s rise to fame to the tortured frontman\'s suicide in 1994. The cover features a beefier-looking Cobain with angel wings, which should incite curiosity as to what his Nirvana bandmates Dave Grohl and Krist Novaselic\'s cartoon alter-egos look like (one can only hope that they\'ll be clad in skin-baring leather loincloths). [ urlLink filter ]   '),(14913,1694057,'16,November,2003','2003-11-16 00:00:00','   sea ray - stars at noon     urlLink sea ray gets stellar review in rollingstone.com!   '),(14914,1694057,'16,November,2003','2003-11-16 00:00:00','   get vicious tonight!    photo: jenyk.com  well kids, it\'s about that time again. vicious is again upon us at sin-e with a super kick-ass lineup featuring:  urlLink  aerial love feed    urlLink  ps    urlLink  dirty on purpose    urlLink  scout   @  urlLink sin-e  (148-150 attorney st) 8 p.m. // $9   '),(14915,1694057,'16,November,2003','2003-11-16 00:00:00','   the strokes of art    i actually found myself reading the liner notes and fine print on  the strokes  CD. i was curious to find lyrics in the cover (d\'oh, you\'d think i\'d look for those first), and yes, in fact, julian casablancas again wrote the words and music of all the songs, save one -- \"automatic stop\" -- which apparently everyone\'s favorite curly-haired guitarist albert hammond helped pen.   i was looking for the cover art credit -- and i was surprised to find that one of the inside artworks was by german artist  urlLink  anselm kiefer  . go figure! i remember going to the met to see a kiefer exhibit in connection to my german film class. beautiful dream-like images that were endearing in the seeming gentle glide of each stroke.   '),(14916,1694057,'15,November,2003','2003-11-15 00:00:00','   bishop allen are looking for a new female singer!   have you got a sweet voice? do you love bishop allen and would love to join their band? you\'re in luck. these cool kids are looking for a new female singer to be part of the bishop allen family. yes, yes, bonnie is sadly leaving the nest -- but this is your opportunity to join a kick-ass, witty, smart, funny band!  urlLink  email christian   to audition for the band now!   '),(14917,1694057,'15,November,2003','2003-11-15 00:00:00','   happy birthday, ropstyle!!   '),(14918,1694057,'13,November,2003','2003-11-13 00:00:00','   congrats, miss modernage! this is her good news: Our limey-savvy friends have informed us that our formerly cool site appears in this year\'s NME Big Book as the number 2 best site on the Internet. What\'s number one? The now technically retarded Friendster. This only goes to show you, if you want to be number one, you\'ve got to sleep with more high-level editors. We\'ll try harder next year.   '),(14919,1694057,'12,November,2003','2003-11-12 00:00:00','   saw  urlLink  ben lee   at the bowery ballroom tonight. he is one fine AUSSIE (sorry). he sang covers of christina aguilera\'s \"beautiful\" -- quite beautiful, in fact -- and began one of his songs with an acoustic version of outkast\'s \"hey ya.\" now that\'s FUN. finally met dani. she\'s such a sweetheart!   '),(14920,1694057,'11,November,2003','2003-11-11 00:00:00','   oh geez,  urlLink stereogum , why do i like that britney  urlLink \"toxic\"  song? it sounds like it came out of outkast\'s big ol\' butt. thanks a lot. good seeing you last night.  tonight i head to the  ben lee  show to meet up with the infamous dani. good times! but i have to say, i don\'t really like phantom planet. there are going to be too many screaming 16-year-olds up in this joint.   '),(14921,1694057,'11,November,2003','2003-11-11 00:00:00','   urlLink the most shocking jacket ever.   '),(14922,1694057,'11,November,2003','2003-11-11 00:00:00','   winter makes me numb  i\'m not sure if it\'s this winter weather or the fact that i go out every night that makes me so tired. are you feeling like that too? i am finding it very hard to kick myself out of bed in the morning, but once i am up, i am ok. it\'s just those few seconds when my alarm beckons that i am ferociously exhausted.   anyhoo, i stopped by  urlLink product shop nyc \'s first  dead hookers social reading club  last night, and it was a really great lineup. i was there to see dan kennedy, who i think is one talented, funny mofo. he was funnier last time i saw him read, but not less talented. before him, there was a newbie reader named randy stulberg, who i found out is a friend of scott. she wasn\'t bad for her first reading, but stumbled here and there. she definitely has the potential to grow as she becomes more confident. i wanted to stay to hear marc spitz and jonathan ames, but had to head over to some other commitments...  then jenna and i jettisoned over to mercury lounge to see  urlLink  aeon spoke  , a band i think is from california, play. they were very atmospheric and moody, but i felt that they were a little too spiritual or over-the-top emotional -- if that makes any sense. however, the crowd there loved it. jasper will be sad he missed their very pretty girl bassist! be on the lookout, jasp. the band was like radiohead but with more whispery vocals.  later i finally met alexander (not andrew) laurence. he was really nice and knows a lot about music so it was fun to go back and forth with him for a little bit. he was heading over to tonic, and i was hungry so we parted ways. chatted a little with yannick and bill jr at sin-e. i actually missed talking to them. they\'re such nice peeps! afterwards, i walked jenna to sidewalk cafe where she was going to see ian thomas. i headed for home and ran into daniel from  urlLink  circle and square   (i run into him everywhere!). we talked for a bit, and he told me C&S would be performing at the knitting factory soon. later closer to my subway i ran into steven mertens, who had been on tour with adam green. he was leaving agan to tour with another band that i can\'t remember the name of... (a.m. something?). anyhoo, it\'s good to see all my peeps so busy and making music and touring and being so happy!   '),(14923,1694057,'11,November,2003','2003-11-11 00:00:00','   cover me  yesterday was the night of covers in the basement of north six. it was dingy down there, and someone was sucking on tea tree oil toothpicks (mmm, minty fresh!) and some band was practicing very loudly in the rehearsal space. plus, the sound in the basement was terrible, but we were all there to support and have some fun...  first  urlLink  the debutantes  , a gothy raveonettes in stereo, sang the go-gos \"our lips are sealed.\" their own songs were cute and spunky. the lead singer definitely had a belinda carlisle-esque voice that was nice to listen to. they had two guitars and drums and were rocking out. some of their songs kind of sounded the same, but i think this band is still evolving nicely...   then  urlLink  kudzu wish  , a les savy fav-ish group from athens, ga, sang \"web in front\" by archers of loaf. god how i love that song.  \"all i ever wanted was to be your spine!\"  the lead singer hung by his feet from the rafters (very tim harrington, eh?). then he fell off and landed on his head!! that has gotta hurt. yow. also this band used a cowbell in one of their songs!    \'64 World\'s Fair , composed of annie and sung bin with doug dop sitting in on drums, covered the pixies\' \"gigantic.\" because the sound was so bad, i couldn\'t hear the vocals too well but i know they sang songs about how beer is better than soda and thomas the tank engine. yeah!! i was so tired and didn\'t stay for the last band,  urlLink \" target=\"new\"> the winter pageant  , but i hear they are good. will catch them next time!   '),(14924,1694057,'10,November,2003','2003-11-10 00:00:00','   detroit rock city  went to detroit unexpectedly last week for work, but i was siked to go and check out detroit rock city. i had never been before, and i found it kind of dirty, but dirty isnâ€™t necessarily bad. unfortunately, i donâ€™t think i got the proper tour of the city, and as a result, my impression of the area lacked. i was staying at a hotel in downtown detroit, where there were hardly any restaurants, though we eventually found a german biergarten that had decent burgers.   i was helping with a sold-out (free) showcase sponsored by ascap and heineken  at detroitâ€™s independent rock venue, the magic stick, featuring some of detroitâ€™s up-and-coming bands: the detroit cobras, the paybacks, the detroit city council (members of the dirtbombs), and even the odd.    the detroit cobras  had a singer who had a stevie nicks vibe going on from her streams of long hair to her throaty voice. the band could do down and dirty rock â€˜n roll to soulful gospel.  the paybacks  had a singer with andy warhol hair and a massively wild personality to go along with it. the crowd ate her up. her band was quite good drawing from all-over-the-place 70s stoner rock â€˜n roll, but i didnâ€™t really dig the singerâ€™s voice (though it wasnâ€™t bad, just not my thang). i kept thinking that if the paybacksâ€™ band teamed up with the singer of the detroit cobras, they would be amazing. something to think about.    the detroit city council  had a total bluesy, motown feel to their music â€“ something youâ€™d expect from a band composed of members of the dirtbombs. one of the guys in the band (canâ€™t remember his name) was responsible for founding kid rock and eminem. the lead singer had this chubby bono thing going on, and he kept shuffling around the stage like james brown. quite entertaining.  even the odd  were a harder band with screaming vocals and more deafening music. again, not really my thing, but they werenâ€™t terrible. while this was going on in the main part of the magic stick, there was hardcore slam dancing going on downstairs!   the band room was gross with lude graffiti all over the walls. it stank of vomit and cigarettes. but there was a stellastarr* sticker in the midst of all the ugliness. how funny! at one point, i was in the room, and it was full of groupies. i was rolling up one of the banners, and one of the groupies threw her beer bottle across the room at me. it shattered against the wall (damn the free heinekens!), and i didn\'t react at all. i just finished rolling up my banner and left. the room was silent. i think they were amazed that i didn\'t react at all. c\'mon, kiddie, flying beer bottles don\'t faze us new york rockers. plus, i don\'t have to sleep with a band to be their friend! haha!   the magic stick , i think, is like the mercury lounge or bowery ballroom of detroit. the space fits about 800 people, and has pool tables, bowling lanes, food and video games to occupy you in case you donâ€™t like the music on stage. the night before our showcase the rapture played. i looked on the schedule and in the near future the shins and james chance will be passing throughâ€¦  overall, i think that detroit really wasnâ€™t my scene, though i adore  pas/cal  who are from detroit. tried to get in touch with caz while i was in town but wasnâ€™t able to reach him. itâ€™s not meant to be! maybe next time iâ€™ll get a proper tour of detroit through the eyes of pas/cal. is it hipper than i think it is?   '),(14925,1694057,'04,November,2003','2003-11-04 00:00:00','   urlLink junior senior interview  [freewilliamsburg.com]   urlLink stellastarr* review  [popmatters.com]   '),(14926,1694057,'04,November,2003','2003-11-04 00:00:00','   urlLink the stills  on tour with ryan adams:  Dec 4  -  NEW YORK CITY  - Webster Hall Dec 6  -  PHILADELPHIA  - Electric Factory Dec 7  -  WASHINGTON DC  - 9:30 Club Dec 9   - ITHICA, NY  - Cornell University Dec 10 - TORONTO  - Opera House Dec 11 - ANN ARBOR  - Michigan Theatre Dec 13 - CHICAGO  - Riviera Dec 14 - MINNEAPOLIS - First Avenue   '),(14927,1694057,'04,November,2003','2003-11-04 00:00:00','   urlLink we need more cowbell!   '),(14928,1694057,'04,November,2003','2003-11-04 00:00:00','   i\'m off to detroit rock city!  see ya when i get back!   '),(14929,1694057,'02,November,2003','2003-11-02 00:00:00','   Apparently,  The Walkmen  have changed the tracklisting for their upcoming album,  Bows and Arrows  (this is different than what i posted before so check it):  Whatâ€™s In It For Me The Rat No Christmas While Iâ€™m Talking Little House of Savages My Old Man 138th Street The North Pole Hang on, Siobhan New Yearâ€™s Eve Thinking of A Dream I Had Bows & Arrows   '),(14930,1694057,'02,November,2003','2003-11-02 00:00:00','   three time\'s a charm    i saw  urlLink  the walkmen   three times on saturday night. so much fun in so little time. thank you, ham, paul, pete, matt, walt. you rock.  le sigh.   i have a review of the night written up for  urlLink  PopMatters  . i\'ll post the link once it is live. so yeah, now i\'ve started to write concert reviews for PopMatters. now i actually have a \"real\" reason to go to lots and lots of shows and cover all those bands that  NEVER  get covered. however, please don\'t ask me to review your band. i only want to write reviews when i feel compelled to write one.  by the way, i love how british sea power (aka that band that sounds like david bowie) start \"fear of drowning\" --  \"jesus fucking christ, oh god no\"  -- that is the best lyrical start to a song to date and i can\'t stop playing it on repeat repeat repeat. it perfectly reflects my mood lately.   '),(14931,1694057,'01,November,2003','2003-11-01 00:00:00','   in case you were wondering, YES, i will be at the knitting factory tonight. all night. reviewing the bizarre walkmen experiment of three shows in one night with the roger sisters at the same venue .... for some music site. anyone care to join me?   '),(14932,1694057,'01,November,2003','2003-11-01 00:00:00','   the day of the dead mixtape in my head:  british sea power - fear of drowning the shins - gone for good the strokes - the end has no end dirty on purpose - air show disaster folksongs for the afterlife - reunion the decemberists - here i dreamt i was an architect pela - latitudes dresden dolls - coin operated boy the shins - those to come saintface - hudson & day the walkmen - french vacation french kicks - destro brendan benson - life in the d   '),(14933,1694057,'31,December,2003','2003-12-31 00:00:00','   happy new year!   DoP finished up in the studio last night and we went over to antique lounge for some celebratory drinks. so much fun! the jukebox was wonderful. during van halen\'s \"jump\" a group of two guys and a girl started strutting their stuff. one guy gave the girl quite a lap dance while us jerks whooped and hollered. then antique lounge closed up and we headed for greenpoint tavern. that joint was closed too. in desperation we went to the abbey, and surprisingly to me, they were open. we sat down and joe was ridiculously drunk by that point, playing pool with thugs and hippies from the nabe. greg also played pool with the thugs and on his last shot, looked the thug in the eye, and fucking made the shot. that\'s some cool as pool playing right there, dude.   '),(14934,1694057,'30,December,2003','2003-12-30 00:00:00','   s-s-studio    spent hours and hours in the studio with  urlLink  dirty on purpose  . they are recording a new EP for release early next year. it\'s going to be hawt. fucking hot! if you haven\'t voted for that suicide girls contest,  urlLink please vote for dop now ! scroll down to new york city, and then vote for dop, the second band over. the cut off for votes is december 31.    i also went to see the  urlLink  dresden dolls   at joe\'s pub. they were amazing. i was in jaw-dropping awe as their magnificent sound filled the room -- which incidentally was packed to the brim. this is an amazing feat if you consider that they were the only band on the bill and the cover was $12. the band sounded 10 times better than they did at mercury lounge. the acoustics at joe\'s pub were mindblowing. i won\'t say too much here because i want to review the show for  urlLink popmatters , but let\'s just say i was enthralled. anyhoo, jenna and i also ran into  urlLink daniel and robin , as well as glenn and yuki of  urlLink  asobi seksu  . there was an entire boston contingent at the show, which i am only connected to by jenna. she\'s da bomb, yo.   '),(14935,1694057,'29,December,2003','2003-12-29 00:00:00','   spend $12 for some fun     urlLink dirty on purpose   head into the recording studio today...   also, i know you\'ve become bored during vacay, wondering what happened to new york city night life. well, at least i have. but don\'t despair, hon\', here are two cool things to do tonight:    urlLink dresden dolls  @  urlLink joe\'s pub  7:30 p.m. // $12   boz boorer (of the smiths fame) w/  urlLink saintface , my favorite @  urlLink sin-e  8:30 p.m. // $12   '),(14936,1694057,'23,December,2003','2003-12-23 00:00:00','   aw shucks  so much has been going on the past few weeks that i feel like i haven\'t had a chance to breathe. i took today off from work to do just that. breathe. i guess today is my day to get my shit together a little. the holidays are such a drain. work has been crazy as well, but i like that better than having downtime. if you haven\'t noticed, i like to keep busy.   this time of year is also a time for reflection and correction. that whole new year\'s resolution. i always make one or two or three, and then sometime in the middle of the year, i forget i made any. typical.   there are certain things i\'d love to accomplish: build a higher tolerance to alochol so i\'m not puking in some restaurant on the steps like a crazy person. hmmm, that\'d be lovely.   other things i\'d like to accomplish: meet colin firth and fall madly in love and he\'d leave his wife and his kids for  me ... oh wait, wife and kids? scratch that. alimony is a bitch.  another accomplishing thought: read more books on the subway coming home from a rock show when i\'m either tired or buzzed or both... oh wait, lately i\'ve been terribly wasteful and taking cabs back to wburg, so nevermind. maybe i should try to stop doing that and instead invest said money to smoke more cigarettes.  sounds like i\'m going to have to think more about this.   '),(14937,1694057,'20,December,2003','2003-12-20 00:00:00','   finally   urlLink i finally wrote up my crazy, drunken night last week. check it.   '),(14938,1694057,'19,December,2003','2003-12-19 00:00:00','   my top 10 of 2003   urlLink check out my list here.  and while you\'re at it, add your own. thanks,  urlLink drewbacca !   '),(14939,1694057,'19,December,2003','2003-12-19 00:00:00','   pas/cal goes jingle bell rock   urlLink pas/cal - \"jingle bell rock\"  -- download this christmas jingle from the only band in detroit not getting beat up by jack white.   '),(14940,1694057,'19,December,2003','2003-12-19 00:00:00','   uh huh  yes, i know i missed the walkmen show at north six the other night. shocker. whatever.   '),(14941,1694057,'12,December,2003','2003-12-12 00:00:00','   wha tha?  geez, where the hell have i been? yesterday i had been drinking since like 1 p.m. it was a day of holiday party after holiday party after holiday party. i totally lost my cookies at virgil\'s for my work\'s departmental holiday dinner. that was after 4 glasses of bubbly. i shoulda slowed down the pace because then i had to go to  urlLink another holiday party  and i was just such a wuss. my head was aching and i couldn\'t find any water! go figure. just more alcohol. so shortly after greeting several of some  urlLink lovely   urlLink lovely   urlLink lovely  peeps, i bid adieu and cabbed it home.   the other weekend, i saw anna\'s band  urlLink knife skills . the only word to describe them is BADASS. they are fucking badass. with all of their practicing and rehearsing and recording, they fucking ruled that small sin-e stage, and seemed destined for bigger halls of music. anna\'s vicious vocal and guitar skills. fon lin\'s sexy voice and basslines. jerry\'s machine gun drums. watch out, people. you have no idea!  at some point i also went to see my friend miker\'s jokey robert palmer cover band, dr. doctor, and it was fun times singing along to those songs. i kind of wish that they were dressed up with slicked back hair and red lipstick. that would have been fucking hawt. isaac kept playing songs by people like crystal waters (\"she\'s just like you and me but she\'s homeless!\") and jody watley?? (\"real love! i\'m searching for a real love!\") i found out last night that i had missed brian long\'s set at the startime holiday party. he was the first DJ, but i\'m sure i\'ll catch him another time.  oh and for those of you who missed the panthers show at \"the lounge\" the other night (me included), paris hilton and seth green was there. woopty-doo. i hope she was at least flashing people and saying, \"i\'m not spoiled.\"  jesus, i honestly don\'t remember what else i\'ve been doing. i know there was something else that i did? please feel free to remind me.   '),(14942,1694057,'12,December,2003','2003-12-12 00:00:00','   saintface review  check out my review of  urlLink saintface  at the mercury lounge on  urlLink popmatters.com . can you believe the band is opening for  urlLink boz boorer  at sin-e on december 29? crazy! fantastic!   '),(14943,1694057,'11,December,2003','2003-12-11 00:00:00','   will fill in the blanks soon    ok so the other night i went to the vector holiday party at joe\'s pub with daniel -- he plays drums in  urlLink circle and square  and also works for  urlLink inland empire . i met him way back when i booked his band and the boggs for those \"on the moon\" shows. anywho, we walk into joe\'s pub and lo and behold there are waiters walking around with trays of alcohol. it\'s the kind of party where you don\'t even have to ask for alcohol. it\'s right in front of you! so as soon as we walk in, i grab a glass of red wine. 10 minutes later, i see a guy walking past me with a tray of tequila shots. i grabbed one. then 10 minutes later, he came by again, and yes, i grabbed another one. by this time i was really trashed. damien rice, who is on vector records, was supposed to perform at the party. daniel and i only caught one song -- but i heard the rest of the night was simply amazing... will have to check out damien rice again.   so daniel says he has an extra spot on the guest list for the rapture show that night because someone canceled at the last minute. on our way to bowery, we ran into a guy who was dressed up as tony clifton -- the alter ego of andy kaufman. daniel started calling his name, and so i joined in. we didn\'t expect the guy to actually come over, but he did. i took the photo above to capture the moment. as we continued on to bowery, i was worried that i wouldn\'t get into the rapture show. but as it turned out, we didn\'t even have to check at the list because we were waved in. how awesome is that? i wasn\'t planning on attending this series of nights at the bowery ballroom, but i was intoxicated and it was free, so why not? we arrive, and i see miss melody nelson et al faithfully in the front, as we (we also met up with another c&s dude, ed) made our way to the middle. the rest of the night was kind of a blur of dancing bodies, but i remember being amazed that the rapture had improved so much since i had seen them at irving plaza. wowzers.    damien rice, with yes, the worst sweater in the world, performing earnestly at joe\'s pub. whatever, i still love his music, despite the fact that he obviously needs a stylist.    some other night, i also caught  urlLink the winter pageant , who actually impressed me with their great live show. i was only there for three songs, but i distinctly remember loving, \"daggers and nails,\" a song that i was disappointed to find was NOT on the CD i have. record this ditty and send it to me, mike! the band filled the small luna lounge space, and people were totally into it.       i also met up with bsiii for the second time in my life at a bar/lounge called happy ending. it was so weird because i felt like i was at some kind of underground fashion show and there was a runway of some sorts with the strip of space between the band and the booths. very fashionably conscious people were strutting down that strip, workin\' it harder than missy elliott. bsiii had described cousin to me as \"steely dan meets radiohead.\" this was not aan accurate description. they were like some kind of neo jazz ensemble -- i didn\'t get the whole radiohead reference, but i suppose it was because the singer\'s voice was kind of high like thom yorke\'s. i can\'t remember at the moment, but the band had kind of wack, pseudo-funny song titles... though i could totally picture myself listening to that music on a quiet wednesday night of sorts...   '),(14944,1694057,'10,December,2003','2003-12-10 00:00:00','   the lush life of ultragrrrl   urlLink check out her photos with celebs   '),(14945,1694057,'09,December,2003','2003-12-09 00:00:00','   startime holiday party   startime\'s first annual holiday party  Tuesday, December 16, 2003   urlLink The Joggers  Doctor Doctor (Robert Palmer Cover Band) DJ Eucho @ The Pussycat Lounge 96 Greenwich Street 9 p.m.-2 a.m. Free brooklyn lager and vodka from 9-10 p.m. $3 cover  >>> listen to  urlLink \"hot autism\" mp3   '),(14946,1694057,'09,December,2003','2003-12-09 00:00:00','   dresden dolls    boston-based band,  urlLink  the dresden dolls  , came into the office today, and i got a copy of their CD. they just got signed to a label, don\'t know which one... they will be playing at joe\'s pub on december 29. grab your tickets while they last!  urlLink listen to \"coin operated boy\" mp3   '),(14947,1694057,'09,December,2003','2003-12-09 00:00:00','   MTV CASTING  Find out if you have what it takes to be on MTV!  AE and MTV/Granada Television USA invite you to the American Eagle Outfitters stores in Menlo Park Mall and Livingston Mall in New Jersey from 2 to 8 p.m. on Saturday, Dec. 13, for a special casting call.   Guys and girls between 17 and 25 who live in the greater New York/New Jersey area are invited to try out for a new reality series set to air on MTV in 2004. You must be single or not seriously dating to participate.   Plus, buy an AE trucker hat customized any way you want it! New York\'s top graffiti artists will be tagging hats to fit your one-of-a-kind style from 2 to 8 p.m.  No appointment necessary. If you\'re interested, just stop by either AE store between 2 and 8 p.m. Saturday and talk to the casting staff. For more information, call 212-941-7200 Ext. 207.   '),(14948,1694057,'09,December,2003','2003-12-09 00:00:00','   VOTE FOR DoP NOW!  OMIGOD,  urlLink Dirty on Purpose  made the finals for New York City band for the Suicide Girls Indie Band search.  urlLink Vote for them here and now up until Dec. 31.  Spread the word and post this on your site too!   urlLink    also, i am in love.  urlLink double love . [link from  urlLink annabiotica , but the heads up from  urlLink fly life\'s tricia romano ]   '),(14949,1694057,'08,December,2003','2003-12-08 00:00:00','   thank you  thank you to  urlLink  the rogers sisters   for linking to my article. rad!   '),(14950,1694057,'08,December,2003','2003-12-08 00:00:00','   walkmen download   urlLink fluxblog  has a walkmen song to download:  urlLink \"new year\'s eve\" mp3   on the contrary to his opinion, i think the new walkmen album will be absolutely fabulous and hot as shit!   '),(14951,1694057,'05,December,2003','2003-12-05 00:00:00','   here\'s to you, mr. green   urlLink the skinny on StarTime International.  by the way, StarTime\'s christmas party will be on December 16, and while i can\'t yet disclose the location, i will say that my friend miker\'s band,  doctor doctor , nyc\'s coolest robert palmer cover band, will play and be awesome! there will also be another athletic band playing, so i hear. it\'s the same night as the next  urlLink vicious , so i may find myself running back and forth. boy, next week is going to be crazy!   '),(14952,1694057,'05,December,2003','2003-12-05 00:00:00','   awards awards awards   urlLink in case you didn\'t know, the grammy nominations were announced today.  someone told me this is the first year the grammys have had the \"best alternative music album\" category, of which nominees include:  Fight Test   The Flaming Lips  [Warner Bros. Records]    Hail To The Thief   Radiohead  [Capitol Records]    Untitled   Sigur RÃ³s  [Fat Cat/MCA Records]    Elephant   The White Stripes  [V2/ThirdMan Records]    Fever To Tell   Yeah Yeah Yeahs  [Interscope Records]   wow, can you believe the yeah yeah yeahs got a grammy nomination? that\'s a fucking huge step for indie kind. i don\'t really consider the other music \"alternative.\"  maybe  sigur ros... out of those people though, i think the flaming lips have the tightest album. word!  i also attended the deems taylor awards yesterday. i thought it was pretty interesting. that  spin  editor, sia, was there. the deems taylor awards honor writers who cover music in journalism and books. i missed the first half of the awards ceremony, but i found that most of the awards went to blues and jazz genre writers. it was cool though that  urlLink allmusic.com  received an award. i never knew that the whole allmusic.com operation is based in michigan, though they have 600 writers who contribute. the a&e show  urlLink \"live by request,\"  which was produced by tony bennett and his son, danny, also received an award. that\'s a show where the big names in music take requests from their fans via phone or email, and many times they actually interact with their listeners. people who have been on the show thus far include bowie, elton john, elvis costello, tony bennett (of course), gloria estefan, etc. sometimes, even other stars phone in, just to rag on the featured guest. i\'ve never seen this show... but now i\'m interested in seeing it.   '),(14953,1694057,'04,December,2003','2003-12-04 00:00:00','   braving the cold    lately i\'ve been in complete hibernation mood. i\'ve been totally repulsed by this biting cold weather, and have retreated to the warmth of my apartment and lots of carnivale and da ali g show on HBO on Demand. the key word is \"comfort.\" but every once in a while, i can be compelled to leave the comfort of my home and venture out... like when bishop allen play a show.   yes, i went to see  urlLink  bishop allen   tonight at sin-e. this was the first time i saw them without bonnie, their former bassist. they put C.O. on bass duty now and have no keyboardist and no main female singer. if you are adept at the keyboard and have a sweet voice, please email them and try out for the band! they write catchy pop/rock/alt-country tunes that charm my socks off. it sounds like they are still trying to figure out how to transition with the absence of bonnie, but the essence of the band is still intact in both the new and old songs. i am still giddy as a school girl when i hear them play \"quarter to three.\" i really liked the new song that they started to play but never finished, \"remember monday.\"    side note: yannick got a cute new hair cut!    starfoods, oh how i love thee    afterwards, rachel and i headed over to  urlLink  starfoods   -- just like old times! starfoods has the best comfort food in this city. there is no doubt about that. and they have the most handsome, nicest staff i\'ve ever met. the owner, byron, greeted us with hugs and then he brought us champagne with casis on the house! i was so excited. i love me some bubbly, but this drink is beyond DIVINE. it\'s my new official new fave, and i may be wandering back to starfoods every weekend just to have one. byron, you got me hooked!!! then while the cars were blasting from the speakers, i had a fabulous grilled cheese with tomato soup that left my tongue craving for more and more. i was so stuffed though because i also gobbled down the star-shaped buttermilk biscuits that were freshly baked warm from the oven. then byron and our delightful waiter adrian surprised us again with free dessert. i was completely in shock. could my tastebuds handle more goodness? they brought us an ice cream sundae with caramel and chocolate syrup AND a banana pudding with vanilla wafers and whipped cream. it was fucking decadent as hell. i felt like i had won the comfort food lottery. thanks starfoods! and thanks for playing the new rapture album! if you haven\'t been to this place yet, you must go and check it out. it is the best cure for the winter blues.  OH AND HAPPY BELATED BIRTHDAY,  urlLink MIKE D !! YOU LEGAL BASTARD.   '),(14954,1694057,'03,December,2003','2003-12-03 00:00:00','   wowee zowee   urlLink tim harrington\'s wedding registry  sent to me from someone else who is getting married.   '),(14955,1694057,'02,December,2003','2003-12-02 00:00:00','   A Must Not Miss Event  Tonight those lovely peeps from Bishop Allen are performing! Here\'s the skinny:   urlLink  Palomar    Laguardia   urlLink  Apollo Sunshine    urlLink  Bishop Allen   @  urlLink Sin-e  (148-150 Attorney St) 8 p.m. // $8   '),(14956,1694057,'02,December,2003','2003-12-02 00:00:00','   ah, dios  listening to the new  urlLink  dios   record is like listening to something really familiar and comforting. like a warm blanket. the lead singer\'s voice is a gentle coo -- accompanied by fluid piano and enunciated drum beats. so though their music may be remarkable calming, the  urlLink photos of drizzunk animals  on their site is not.   urlLink listen to \"you make me feel\" mp3   urlLink listen to \"everyday\" mp3    the simple life  i love that new paris hilton show,  urlLink the simple life . severely bored? play dress up with paris and nicole  urlLink here .   '),(14957,1694057,'01,December,2003','2003-12-01 00:00:00','   the best and the worst   urlLink freewilliamsburg\'s best and worst of 2003.  my list is not yet complete. what\'s yours?   '),(14958,1694057,'01,December,2003','2003-12-01 00:00:00','   the end of the world   urlLink here it is  [link from jenna]   '),(14959,1694057,'31,January,2004','2004-01-31 00:00:00','   wowee zowee!    thanks to everyone who came out last to the mercury lounge to see  urlLink  dirty on purpose  . the turn out was overwhelming! i saw many familiar faces and  tons  of new faces! the band was on first, and had the biggest crowd -- their performance was simply illuminating (not that i\'m biased or anything...) but my high spirits were confirmed by several people who came up to me afterwards. it\'s so nice when you get immediate feedback like that. it\'s too bad the band didn\'t get to play \"mind blindness,\" one their new ones that just consumes you, but their performance exploded into wild gusto when they ended with crowd favorite, \"monument.\" i couldn\'t stay still. i was dancing and just drowning in the intense musical finale.  i love dop, not only for their good music, but they\'re just awesome people. they\'re good natured and appreciate everything that comes their way. they love their fans. they love their peeps. they love music. it makes all of the work i do for them extremely rewarding. it sounds a little sappy, but that\'s the truth, ruth. hanging with them at practice and recording and just hanging out is time i never feel i\'ve wasted because i get so much back. a special thanks to dop!  also another thing i\'d like to announce: DoP\'s first EP is officially sold out at  urlLink truthflies.com ! thanks to all who purchased a copy. i think truthflies.com will be taking pre-orders of DoP\'s second CD at some point. we\'ll let you know when it all goes down. i also did get ed from truthflies to come to the show (granted it was after the band went on) to sell me a walkmen  bows and arrows  cd. word! you know i\'ll be buying a bunch of copies of that on the official release date to give to my peeps.    anyhoo, while we were hanging around the mercury lounge\'s downstairs band room, a small, smoky space covered with graffiti and band stickers, we got to know the guys from  urlLink  swell  . they were ultra nice, and we were all siked to see them on the lineup. their music had a wonderful groove to it that was more laid-back than dop\'s set. it was music that just made you say, \"yeah this shit is coool.\" extra \"o\" intended. i really liked them and look forward to seeing them again soon! they have an upcoming show with the wrens at sarah lawrence, so if you missed them this time, you can check them out again soon!    i was seeing  urlLink  the blondes, inc.    last night for the second time. what\'s immediately striking about this band is their look. they had really cool lighting that came up from the stage floor. they were going to put up their signature black-and-white polka-dotted curtain, but at some point decided against it. i think this was a good move. instead of the audience focusing on their stage set up, they were focusing more on the music. the result is rewarding because i completely disagreed with the village voice review that the blondes don\'t play good music. their music is a throwback to the old school. and i saw many girls swaying their hips like they were at a rolling stones concert or something.  tonight i\'m going to see  urlLink  the go station  , a band i\'ve been hearing many good things about. they\'re playing tiswas with  urlLink  aerial love feed  . i\'ve never seen them live before, but you should come check them out with me...   '),(14960,1694057,'30,January,2004','2004-01-30 00:00:00','   show show show!   urlLink  dirty on purpose   return to the mercury lounge after a three-month break! they\'ve been recording some hot new stuff, and they\'ll be playing old and new faves tonight!   listen to some mp3s: *  urlLink \"go back to sleep\"  *  urlLink \"monument\"   jan. 30 --> tonight!!  dirty on purpose  w/ the blondes, swell, int\'l orange @ mercury lounge 8:30 p.m. // $8   '),(14961,1694057,'28,January,2004','2004-01-28 00:00:00','   hot poster  i just love this poster so much.  click it to see it in all its  urlLink garrett morlan  glory.   urlLink   '),(14962,1694057,'27,January,2004','2004-01-27 00:00:00','   dearest dennis and eric,  please forgive my absence last night at your sin-e and mercury lounge shows, respectively. i hope that they went well.   dennis,  urlLink i love your mp3s  and was really looking forward to hearing you, but i got stuck at work until uber-late and then i was stressing out so much my eyes wouldn\'t stop twitching. i went home and i slept like my life depended on it. but i guarantee i will catch you at asobi seksu\'s release party in march. mucho mucho sowweeeee.    urlLink eric , i listened to  say hi to your mom \'s new cd and it rocks with poppy goodness! everyone should  urlLink buy it .  love jinners   '),(14963,1694057,'27,January,2004','2004-01-27 00:00:00','   congratulations to you  i chose two winners in a random draw for  the stills contest . congratulations to the winners! be on the lookout from an email from moi about the contest. if you haven\'t received any notice from me, you weren\'t among the lucky bastards. sorry!   '),(14964,1694057,'26,January,2004','2004-01-26 00:00:00','   walkie talkie  the next  urlLink walkmen  show: april 1 @ irving plaza. wheeeeee!   '),(14965,1694057,'25,January,2004','2004-01-25 00:00:00','   fred flare, etc, etc   urlLink  fred flare   is having a major sale. lots of cute valentine\'s gifts, too, for the romantic in you. their stuff is always adorable, fun, colorful and hip. check \'em out!  another link sent to me by my friend katy freaked me out.  stevie nicks  fans probably already know about  urlLink this one , but c\'mon!  one of the cooler experimental bands in new york city,  the new humans , are doin\' their thang on jan. 27 with another new york band,  cheeseburger . while the new humans will play under their huged quilted sheet, cheeseburger will play simulataneously... all of this happens at  urlLink gavin brown\'s passerby art gallery  in chelsea. 10 p.m.  also playing on tuesday is  dennis from the realistics  at sin-e.  urlLink he\'s posted some mp3s on his site , so check em out and go see him! 10 p.m.  also playing are  urlLink  say hi to your mom   -- it\'s their record release party at the mercury lounge! 9:30 p.m.   '),(14966,1694057,'25,January,2004','2004-01-25 00:00:00','   sleater-kinney   Sleater-Kinney  have just launched their official web site. You can find it at:  urlLink www.thisissleaterkinney.com   ALSO Janet and Corin made a celebrity playlist in  urlLink iTunes .  To find it click on the link to Celebrity Playlists on the iTunes Music Store front page.   '),(14967,1694057,'23,January,2004','2004-01-23 00:00:00','   +1 +1 +1 +1  the launch of jonny kaps\' new company,  urlLink +1 Management and PR  last night at don hill\'s was crazy packed. it was difficult to maneuver through the crowd that was shuffling about with their discounted vodka drinks. i walked into the venue as  inouk  was on stage. i liked what little i heard as i headed back to the coat room, an arduous task that took longer than it should have. i\'d have to see them again to really formulate a solid opinion.  my spirits were high , and i couldn\'t help but break into lots of smiles. later  ambulance ltd  took the stage and impressed me. they\'ve gotten a lot better since i last saw them -- and as you know they also have a new keyboard player. there were definitely parts of songs i liked better recorded than live, but they were tight overall. marcus\' voice was also stronger.   on tv screens in the bar,  stellastarr*  videos were playing on a loop as if to foreshadow to their \"surprise\" performance. they took the stage, looking exactly how i remembered them the last time i saw them. they started out with \"in the walls\" -- those opening guitar parts are always so engaging and ominous of a spectacular performance. i\'m not sure what it was, but i think they sounded better at don hill\'s than in any other venue i\'ve seen them in (funny, because don hill\'s is the first venue i saw them in). it was like they had come full circle in my eyes. the crowd exploded when they began to play the anthemic single, \"my coco.\" it was good to see that people that weren\'t that impressed with their recorded stuff, were impressed with their live performance. i always think of stellastarr* as a band that is fun to watch live because guitarist michael jurin will undoubtedly be writhing on the floor in dramatic flails as will lead singer shawn christensen. the drummer arthur kremer also broke out into his trademark barechests, with his trademark asterick nipple. he\'s probably the most entertaining performer to watch on stage because his personality is so pronounced. bassist amanda tannen also seems like her stage presence has improved, her stance on stage seems to have more hamony with the other members than it had in the past.   '),(14968,1694057,'23,January,2004','2004-01-23 00:00:00','   adrian tomine   ooooh adrian tomine has a new  optic nerve  out, finally! he\'s one of the few storytellers out there i really identify with. and it\'s not just cuz he\'s asian. he understands alienation, loneliness, confusion, small joys...  here\'s a blurb from the housing works used book store site: Adrian Tomine, creator of Optic Nerve #9, will appear for a Q&A. The graphic artist will also be available to sign books.   About Adrian Tomine: The California-based cartoonist has been producing the Optic Nerve series for more than 12 years. His latest, Optic Nerve #9, presents the fictional story of a troubled twentysomething Japanese-American guy\'s cross-country quest for the perfect girl. Excerpts from his earlier work appeared in the critically acclaimed graphic novel Summer Blonde. Tomine\'s illustrations have also appeared in The New Yorker and Esquire. \"Reading a comic book suddenly becomes as rewarding as reading good contemporary fiction...Tomine has both talent and a writer\'s eye for the truth.\" --The New York Times   i can\'t wait for his book signing:  wednesday, feb. 4 @ 7 p.m.  urlLink the housing works used book cafe   '),(14969,1694057,'22,January,2004','2004-01-22 00:00:00','   scary shit  you know what\'s kind of freaking me out? i looked at my list of referrers today and guess what was listed as a referrer?  urlLink riaa.com  what the hell is up with  that ?? anyone experiencing this? email me!   '),(14970,1694057,'22,January,2004','2004-01-22 00:00:00','   killing penguins is not cool    urlLink penguins penguins penguins  click on the penguins and play the game... the trick is to get the penguin to slide when it lands on the snow. you get more distance. i got 530 ft.   don\'t let your co-workers see you playing this game. it will give them nightmares. in fact, it gives me nightmares. and maybe it will give jo divestar nightmares. you know how we both adore penguins.   i wonder who created this sick game? i have to confess, it\'s damn fun. it\'s damn addictive. and i love the \"whee.\" i don\'t think i can write enough crap to fill this space on the side of the picture. oh well. deal with it!  penguins! penguins! penguins!  ok i\'m done now. sorry about that. i\'m loony for penguins.   '),(14971,1694057,'22,January,2004','2004-01-22 00:00:00','   omg  shuddup!  urlLink the walkmen  are going to be on conan o\'brien feb. 6. hot diggity dawg! that\'s the same night dirty on purpose is opening up for the suicide girls burlesque show at the knitting factory. set your tivos/vcrs, peeps, and come to the knitting factory instead! i wonder if they will play \"the rat\"? from what i hear, it will be their first single and video. you\'ve got a nerve, baby! also, i wonder who\'s taller, hamilton or conan?   '),(14972,1694057,'21,January,2004','2004-01-21 00:00:00','   french kisses   '),(14973,1694057,'21,January,2004','2004-01-21 00:00:00','   walk on air   urlLink jasper\'s  new venture  urlLink truthflies.com  is having a special pre-sale for the air album. rock on! it\'s a limited edition cd for all of you  urlLink air-heads . it also comes with a free poster, bonus! truthflies will also have a drawing for some autographed air gear -- check out their site for deets.   '),(14974,1694057,'20,January,2004','2004-01-20 00:00:00','   dop dop dop    i have taken lots of funny  dirty on purpose footage  during the past few weeks. they are amazing. their new music is KICK ASS. they are lovely.   urlLink george dop beats walt walkmen on stillness   and i think i\'ve been having way too much fun with my digital camera and its ability to do multiple exposures. a couple of these photos were taken by joe. booyaaaaa:   '),(14975,1694057,'20,January,2004','2004-01-20 00:00:00','   get boggy   the boggs  have a  urlLink website  finally!   '),(14976,1694057,'20,January,2004','2004-01-20 00:00:00','   stuff to do, and you know, stuff  i feel sort of strange not going to vicious tonight, but there was so much shit i needed to get done. and the range of upcoming shows is stellar. check it:  wednesday, jan. 21  urlLink au revoir simone  +  urlLink coup fourre  FREE @ knitting factory  thursday, jan. 22 mahogany, prosaics, on!air!library!, sea ray, medicine hat (what an orgasmic combination of bands) @ sin-e  friday, jan. 23 starsailor, tim burgess (i love the charlatans, tb\'s new album is pretty good too) @ irving plaza  saturday, jan. 24 +1 launch party // tiswas with ambulance and inouk @ don hill\'s  qatsi, the fever, the affair @ mercury lounge  read yellow, circle and square @ north six   '),(14977,1694057,'16,January,2004','2004-01-16 00:00:00','   urlLink check out my dresden dolls review on popmatters.   '),(14978,1694057,'16,January,2004','2004-01-16 00:00:00','   vaz + the amber smith  on the same night i saw the unicorns play, i saw  vaz  for the second time and the amber smith for the first time. i remember being really impressed with vaz the first time i saw them play with knife skills back in the days when luxx was still a cool rock venue. the lead singer, paul, is a sweaty agile performer who spits frequently and attacks the mic viciously. so much fun to watch. i became fascinated with just staring at him and his wild, but strong performance. whoa. his drummer was like a crazy homeless person with severe schizophrenia. his head would swivel wildly like he was having a seizure. i\'d be caught staring helplessly at him, too. then suddenly his eyes would rise from the drums and quickly fall back down. it was rather scary. and i liked it.   the amber smith  were the exact opposite. not scary. not threatening. their music had a cheerful tone. it reminded me of the strokes meets  that thing you do!  i swear. one singer sang while chewing gum and playing guitar at the same time. the other singer looked a bit like michael ian black, kind of. some people describe them as a girly version of the strokes. i basically agree except that some of their lyrics were on the sarcastic, snide side (which turned out to be their saving grace) -- i think one of their song lyrics were \"i\'m having an affair with your girlfriend.\" if there wasn\'t that iota that makes them stand out from other poppy rock bands, i would say they were good but nothing special. in this case, they might grow into quite a catchy band if they work on charisma and diversity in tempo and pitch.   '),(14979,1694057,'16,January,2004','2004-01-16 00:00:00','   marg cho show    From  margaret cho \'s mailing list: Episode #2 of our new Internet Show will be up tomorrow. In this episode, Margaret and Bruce visit the Human Sexuality Collection at the Cornell University Library. You can view it from our Home page:  urlLink http://www.margaretcho.com  or Margaret\'s blog:  urlLink http://www.margaretcho.com/blog . We\'re going to try to put up a new episode every other week or so. We have some fun things planned and it\'s all free to view (at least for now).   '),(14980,1694057,'15,January,2004','2004-01-15 00:00:00','   pela download  this new  pela  song,  urlLink \"all in time,\"  is amazing. check it out, you can\'t deny it! producer  rod \"goldenrod\" sherwood  sure does have the midas touch. he produced pela\'s songs as well as the upcoming  dirty on purpose  cd that is mindblowing. all i can say is wowee wow wow!   '),(14981,1694057,'15,January,2004','2004-01-15 00:00:00','   we\'re the unicorns and we\'re people too   urlLink    even though it was snowing, i went with deron and anna to southpaw (they drove) to see  urlLink  the unicorns   play at southpaw. one of the first things out of one of the unicorns\' mouth was, \"this show is so much better than last night. pianos sucks.\" hmmm, i heard it was crazyfest! but let me tell you, southpaw was jampacked with people, and y\'all know that venue is a pretty big size. the buzz was\'a buzzing.   urlLink CHECK OUT THE PHOTO GALLERY !!   also check out real player video clips of the unicorns: *  urlLink unicorns 1  *  urlLink unicorns 2  *  urlLink unicorns 3  *  urlLink unicorns 4   i\'m writing up a formal review on popmatters.com, but my overall impression of the unicorns was first amusement, pleasant surprise to later yawn-eriffic to then amused again to confused to amused to physically tired of watching their \"grody roadie\" dance around spitting beer, throwing condoms all in a unicorn uniform. all in all a good show, but way too long. however, there was some kylie thrown in there. laura will love that. thank god they didn\'t do \"hey ya.\" aren\'t we getting tired of people covering outkast? maybe not.  right before the unicorns we ran into susan. she said that she liked the unicorns after seeing them at mercury lounge. i was kind of surprised, but after i saw them, i felt the same way. i also met doug who books at lit, does version city records, blah blah blah. he was nice. his label is putting out the amber smith cd... i also saw  vaz  and  the amber smith  but i\'ll talk about them in a separate post...   '),(14982,1694057,'14,January,2004','2004-01-14 00:00:00','   my pet fat  can we say, ewww?  urlLink my pet fat . thanks alot, anna.   '),(14983,1694057,'13,January,2004','2004-01-13 00:00:00','   the stills ticket giveaway   urlLink   '),(14984,1694057,'12,January,2004','2004-01-12 00:00:00','   hammy!!  ok, he was super nice and even, jocular. weeeeeeee. here are a couple of new walkmen photos:   '),(14985,1694057,'10,January,2004','2004-01-10 00:00:00','   some fun links     urlLink  Pete\'s Corner - Mr. Match contestants and their answers   wouldn\'t it be funny if walkmen fans around the world started dating each other because of pete?   The Stills on Last Call with Carson Daly   urlLink Real Player   urlLink Windows Media - low    urlLink Windows Media - medium   urlLink Windows Media - high    The Stills on  Rolling Stone /MTV2 special   urlLink Real Player   urlLink Windows Media - low   urlLink Windows Media - medium   urlLink Windows Media - high   '),(14986,1694057,'10,January,2004','2004-01-10 00:00:00','   boys, inform me     photo: rachel lovinger   i had this incredible post about  the shins  show at the mercury lounge and blogger ate it up. argh! here\'s the short version:   starfoods , nila wafer banana pudding, mac and cheese, bourbon and chocolate pecan pie, i heart byron who is a total sweetheart every time!   todd barry  +  the shins  is funnier than  david cross  +  the strokes . i was skeptical about the whole comedy before rock thing, but todd barry pulled it off.    tigers and monkeys  had jeff from  sea ray  in it. incredible female singer. surprisingly gutsy... we\'ll have to see how they develop. jeff plays the drums standing up!  roland, ok = awesome the shins = awesome  the last time i saw the shins was at siren fest a couple years ago. they were just OK back then. so i had medium expectations this year... jeez, was i blown away. their performance at the mercury lounge was dynamic... almost explosive at times. the guitars and keyboards were jacked up and LOUD. it was nice to see the keyboards showcased a little more. they are kind of buried in the recording, and that marty kind of goes wild on stage. NICE showmanship! james mercer was kind of the logical foil to his wild antics. his voice was so much better this time, really crisp and that sweet ring of his vocals was heightened and really made the songs pop. they sang all the usual suspects from the old and new album -- no  flake  songs, much to rachel\'s dismay. rachel can tell you why they are reluctant to play flake songs.   what is roland, ok?  the city and state where the shins got arrested for pot and drug paraphanelia. marty seems very bitter about this and kept making mary jane jokes throughout the set. he was a hoot to watch...  people i saw and talked to: jeff from sea ray, eric from  say hi to your mom , dawn, noah from  sam champion , daniel from  circle and square  -- he pinched my cheek (the one on my face) like i was an infant, todd barry, rachel, justin and c.o. from  bishop allen , jo divestar who told me the latest shins video had penguins in it!!  people i wished i talked to: david cross (i hope he smokes everyone in celebrity poker showdown!) watch it tuesday on bravo. james mercer and marty... all i would have been able to say woulda been something dorky like, \"uh i really thought you guys were great.\"   '),(14987,1694057,'10,January,2004','2004-01-10 00:00:00','   hammin\' it up  i am going to be interviewing hamilton of  urlLink the walkmen  on monday over the phone. i\'m prayin\' he won\'t be surly towards me. i don\'t know why he still intimidates me. maybe it\'s the perpetual scowl on his face or his considerable height. the new walkmen album is DOPE. it\'s so weird to be hearing the songs on a CD though because i\'ve been hearing them live for so frickin\' long. but i likes it.   '),(14988,1694057,'08,January,2004','2004-01-08 00:00:00','   why did i just read this?  normally i wouldn\'t post anything on britney spears, but this is too funny:  urlLink mommy and daddy broke the britney spears wedding story.   '),(14989,1694057,'08,January,2004','2004-01-08 00:00:00','   dissing the bloggerazzi  dude, it\'s just not right to  urlLink diss  people who aren\'t there to defend themselves. [link via miss \"modern\"] gotta at least leave the option open for people to respond to your \"low blows.\" that\'s just my opinon though.   '),(14990,1694057,'08,January,2004','2004-01-08 00:00:00','   some more shows     urlLink  the stills   have added a second night at the bowery ballroom for jan. 29. it\'s a special bbc 1 broadcast and is 16+ for the kiddies. there might be a possibility that i\'ll be giving away a pair of tickets to the show so stay tuned. are they playing before or after  the killers ? i just saw that ultragrrrl posted on her site that the killers were gonna play the bowery on the same date...  while perusing, i see that  the von bondies  are coming to the bowery ballroom on feb. 25. hope those bruises have healed, jason. ouch.   the national  have been added to the bill for the feb. 21  elefant  bowery show, too. does anyone else think that\'s kind of a strange combination? on one end you have kind of dark serious music with smart introspective lyrics, and on the other, well... you know. let\'s not kid ourselves. he\'s hot but nowhere near deep.   '),(14991,1694057,'08,January,2004','2004-01-08 00:00:00','   french kicks    what\'s cooler than cool? the  urlLink  french kicks  .   the first french kicks show at the mercury lounge will be february 4 with  the shout out louds , that band that  urlLink jo divestar  raved about recently...   '),(14992,1694057,'08,January,2004','2004-01-08 00:00:00','   lord of the rings  greg texted me to tell me that the movie theater at lincoln center will be showing the  lord of the rings  trilogy back to back all day. well, day-am, it takes almost half a day to just show the whole string once, don\'t it? he is going to go see it, bless his soul. hope there\'s some sort of intermission so people can take a pee break or something.   '),(14993,1694057,'08,January,2004','2004-01-08 00:00:00','   ambulance  the emergency sirens are going off as  urlLink  ambulance   announces the band will release their full length finally on march 2. this full length features the single (a brand new version of \"primitive (the way i treat you)\") along with brand new tracks! well, let\'s hope this new version of \"primitive\" is a lot better cuz that old track was kinda wack. i have to say though their shit is catchy.   '),(14994,1694057,'08,January,2004','2004-01-08 00:00:00','   many good things to come    woo boy, i\'ve been swamped at work and wasn\'t able to update at all for the past couple days. day-am! after whooping some ass in an intense game of scrabble, i headed over to sin-e with jenna to see  urlLink  au revoir simone   for the second time in 7 days. let me tell you, i really love the concept of this band. to create a band that revolves around keyboards is simply awesome. and three keyboards, no less! i can tell erika, annie, sung bin and heather are heading in the right direction. their songs are pretty and hip. i really enjoyed the instrumentals (just like  urlLink jenna  did), and there were a couple songs with super-soft, melodic harmonies -- pretty heavenly, you might say. i also loved their covers of heavy d & the boyz\' \"now that we\'ve found love\" and liz phair\'s \"girls, girls, girls.\" there will be many good things to come for this band down the road if they keep at it. their next show is in the old office at the knitting factory on jan. 21 with  coup fourre . okay, that is a mighty strange combination of bands, but the show is free, so go check it out!  so what\'s coming up? i\'m sure you already knew, but, well  the shins  are playing under the moniker \"roland, OK\" at the mercury lounge with comedian  todd barry  on sunday. my friend rachel talked to todd barry about it, and he said something like, \"i hope it\'s not a blood bath\" in that wry tone that pervades his stand-up. dude is funny, so i\'m looking forward to this show very much. it sold out at some point this week so i hope you have already bought your tickets. apparently, the show is a warm-up for the shins\' before they appear on letterman or one of those other talk shows.  on tuesday, jan. 13,  the fever  and  the unicorns  are playing a FREE show at pianos. word is that the show is going to be insanely packed so get there early, kiddies. you know how frickin\' small pianos is.   bishop allen  is playing the mercury lounge on wednesday, jan. 14.   on jan. 15, there is one juicy show going on:  prosaics ,  mahogany  and  the winter pageant  at sin-e.   on jan. 16, another juicy fruit:  asobi seksu  and  moonraker  at mercury lounge.  i promised annabiotica i wouldn\'t post this on my blog, so i\'ll be vague. one of my favorite brooklyn bands might be breaking up because their drummer is moving outta town!!! i am heartbroken. argh!!   nyc bars closing at 1 am!  urlLink jo divestar  forwarded this message: Apparently City Hall and the Department of Consumer Affairs have been getting complaints about noise late at night from neighborhood activists and other lame ass people like them, so much like the smoking ban, New York City is about to be ruined once more.  Only this is much worse.  They are planning an initiative to force all bars to obtain a license to stay open past 1am.  It\'s somehow tied into the repealing of the cabaret laws, but apparently it\'s a big joke because they can revoke the license at any time based upon the dumbest criteria and without due process.  If you care at all about keeping NYC...well, NYC and keeping bars and clubs open past 1am, then I suggest you go to these links and write to your state congressman and Mayor Bloomberg himself.  Tell them what bullshit this is, and that you\'re opposed to it.  For those of you who are musicians, this is a potential threat to your way of life. Congressman Joe Crowley:  urlLink write2joecrowley@mail.house.gov  Mayor Bloomberg:  urlLink http://nyc.gov/html/mail/html/mayor.html   '),(14995,1694057,'07,January,2004','2004-01-07 00:00:00','   jennarator.com  my friend jenna has a  urlLink site , finally! check out the fabulous life of jenna!   '),(14996,1694057,'06,January,2004','2004-01-06 00:00:00','   au revoir simone  stereolab + postal service =  urlLink  au revoir simone    or something like that... anyways, the girls are debuting on the lower east side: january 7 @  urlLink sin-e  11 p.m. // $7   '),(14997,1694057,'05,January,2004','2004-01-05 00:00:00','   french kicks  a little birdie tells me  urlLink  the french kicks   will have a residency at the mercury lounge in february. how hawt is that?   '),(14998,1694057,'04,January,2004','2004-01-04 00:00:00','   blonde redhead  can anyone tell me why blonde redhead are not listed under the bowery ballroom for jan. 22 and 23? have those dates been canceled? if so, BIG DISAPPOINTMENT for me and hundreds of other fans in nyc. jeez.  that said, there are tons of awesome shows coming up this month and i plan to be there. back in the rock saddle!!   '),(14999,1694057,'03,January,2004','2004-01-03 00:00:00','   congratulations, DoP!   urlLink  dirty on purpose   has won the opening slot for the  urlLink suicide girls burlesque tour  when it comes to new york city. mark your calendars! it\'s going to be february 6th at the knitting factory. thanks to all who voted for the band!   '),(15000,1694057,'02,January,2004','2004-01-02 00:00:00','   au revoir 2003    i stopped by jco\'s dinner party at the powers house on new year\'s eve. there was a lavish spread of food, but i was wary to eat anything cuz i thought i might end up puking it all up when i get boozed later on. then later on i rushed over to graham\'s loft to see erika\'s band au revoir simone play. au revoir simone are four lovely ladies, three keyboards, one drum machine and four beautiful harmonies. have you ever heard of that before? three keyboards! afterwards, we all headed up to the roof to ring in the new year. we could see the city skyline, the empire state building and all that good stuff. when the clock struck midnight, the fireworks in times square went off as well as in other places within eye sight. everyone was chugging their own bottle of bubbly and giving each other good cheer. soon afterwards, keith and i headed home because he wasn\'t feeling well, and we passed out. happy new year!!   '),(15001,1694057,'28,February,2004','2004-02-28 00:00:00','   elephant woman    i adore the new  urlLink  blonde redhead   song, \"elephant woman.\" i was listening to it on repeat whlie walking around the city last night. the emotion in kazu\'s voice is startling. mind blowing. inviting you into the sadness. embracing and gentle and fragile... can\'t wait for the album to come out on march 9... the album is more produced and pops a lot more than their old stuff. i bought a ticket for each of their upcoming bowery concerts, and i just can\'t wait to lose myself in the dizzy spin cycle of their music.  while walking around, i started off feeling very cold, but after a while of just wandering around and seeing the night life, my body warmed up with a flood of memories. after almost 7 years of living here, i have so many memories... so many different aspects of my life lived here... and suddenly here i am today, a product of all those experiences. motherfucker, where did the time go?  sometimes i take for granted the fact that i live in new york, and what a fucking great city it is. how fortunate am i? i have always known that i would end up in new york, even before i knew what new york was like. and once i came here i never left. some people go home during their summers in college, i was emphatic on staying and working in the city. some people spend time with their family during vacations and breaks, i stayed in the city and lounged in the buzz of new york -- those experiences i wouldn\'t trade for anything else.   earlier in the night i was at pianos to see  the blue sparks  and meet up with jo jo. i have to say that i had had my doubts about  the blue sparks . i\'ve seen them twice before and i didn\'t remember them at all. the only song i knew was from the truthflies compilation. so when i saw them at pianos that night, i didn\'t know what i was in for. it could be good. it could be really bad. but luckily, it was really good. they play their music fierce, sometimes even getting in fights with each other on stage. the tension was definitely thick. but those spats add fuel to the electricity in the air of their performance, and actually made them quite interesting to watch. plus the girl in the band has one of the biggest girl fro\'s i\'ve seen in new york, minus macy gray. that\'s quite an accomplishment in my book.  then we headed over to see  urlLink jo jo  deejay at black and white. boy was that shit packed to the brim. i\'m very claustrophobic so the place was really freaking me out. i had to go outside every 15 minutes or i would get very short of breath and annoyed at all the people brushing violently past me. i was determined to stick around though and see jo jo deejay. that was my entire goal for the night. she finally went on pretty late in the night... i think she started with  stellastarr*  \"my coco\" and then she played  the walkmen  \"the rat\". i was overjoyed, but probably the only one besides jocelyn who was rocking out to it. i\'m the biggest dork when it comes to the walkmen. their fast stuff makes me want to dance and pop a vein in the process. can\'t wait to see them in april!   '),(15002,1694057,'27,February,2004','2004-02-27 00:00:00','   new order  re-read this old article about  urlLink organizing your records ... why? because it\'s friday. and i like this quote:  I love order. It\'s my dream. A world where all would be silent and still and each thing in its last place, under the last dust. -- Samuel Beckett  i bought  the cooper temple clause  cd,  kick up the fire, and let the flames break loose . i\'m supposed to interview them on tuesday... i really dig track six, \"blind pilots.\"   '),(15003,1694057,'26,February,2004','2004-02-26 00:00:00','   damien  i am taken today with  damien rice . jesus. i feel like i can\'t breathe when i listen to him. my heart hurts. you know?   '),(15004,1694057,'26,February,2004','2004-02-26 00:00:00','   practice makes...  ... awesome. practice went well yesterday. i think i was in a haze for most of it, and i felt like i was dreaming. my head was swimming in music that really comes alive, swirling and twirling around the rehearsal space like chinese dragon kites on a windy day in central park.  afterwards we went to kayrock\'s speakeasy. i had never been before, but he turned his place into an amazing lounge. i didn\'t feel like i was in someone\'s apartment at all. i ran into steven\'s roommate and luke. very random... i was gonna try to drag my ass into the city to see jo jo spin at pianos, but i was starting to feel deathly ill. so instead i made the long, cold trek home and completely passed out.   '),(15005,1694057,'26,February,2004','2004-02-26 00:00:00','   this winter is killing me  you know it\'s cold when you have to prance around like an elf just to keep warm while you\'re standing outside of a building to smoke. or something. well, that\'s what i found myself doing all night. plus, the late night after late night gig was weakening my defenses. but i told mike from the winter pageant i would come see his band, and i was determined to keep my word. i mean it\'s a pretty exciting gig to be opening up for french kicks. i was only sad i had to leave to go review the fever/von bondies show at bowery... peeps, that is what happens when you make plans in advance. i am going to try to never do that again.  the night at mercury started off with a sweet trio from kentucky called  shopping . as a girl, i have to tell you that i like the name... but i can\'t explain why three guys named their band shopping. i didn\'t have time to talk and investigate further. i wanted to pinch their cheeks they were so cute! they definitely had the startime vibe exuding from their pores, especially one guy who was nerded out in glasses and a v-neck sweater. i had no idea kentucky could birth indie-ness as such. the other singer was a cute one as well, and boy could he sing charismatically. it wasn\'t so much that he had a beautiful voice or anything like that (the other guy had a prettier voice), but that he commanded a lot of attention when he sang and sang it with spastic energy. they started off strong and upbeat, and at many points reminded me of french kicks in fact. i really liked the song they dedicated to isaac, mostly because of they screamed this one line in the middle of the song in unison and it sounded awesome -- it totally brought a smile to my face.   after shopping i ran into jerry from  knife skills  and  turing machine . he\'s the coolest dude ever. always has something cool and interesting to say. a very engaging person, not to mention an excellent drummer who seems to make other drummers scream like girls. ;-)  he and i both were there to see  the winter pageant . he\'s friends with paul the bass player. you might have heard me mention this band before on this blog/site/thang. they were and are still auditioning for drummers. but i saw them a long time ago at luna lounge, and was pretty amazed by the pretty sleepers they were offering up. the lead singer\'s voice is almost like an instrument itself -- gliding along the top of the music like a cool current of coos. they\'re the kind of band you want to lie down on a carpeted floor, close your eyes and listen to. plus they have some kind of xylophone instrument that makes their songs ring like icicles dripping and freezing on the ground. excuse me if that doesn\'t make any sense. it\'s quite early in the morning for me.   then i headed over to the bowery to see  the fever  and  the von bondies . said hi to rachel as she headed into the merc for the kicks. ran into jo jo also, and we and her friend headed over together to bowery. i ran into matthew at the bowery, who is a total sweetheart. found out a little more about the band he is starting. i have to say, i am dying to hear him sing. you can read my review of the rest of the show on popmatters in the somewhat near future. if i don\'t die of pneumonia before then.  and no i wasn\'t at the after party at starfoods. before that point i texted jo jo and said, \"i\'m dying here.\" that\'s when i know i gotta go to sleep or else i\'m gonna pay sorely.   '),(15006,1694057,'25,February,2004','2004-02-25 00:00:00','   opening the locked doors  as you probably already know, i started off today listening to  urlLink  neil young  . what a great songwriter... except for that song \"a man needs a maid.\" i don\'t care if he wrote it in the 70s -- that song was quite comical in a \"is he serious?\" kind of way. that really made me laugh, but i hear there is an acoustic version of the song that will erase my memories of the dramatic musical version. on the other hand, i love love \"pardon my heart.\" it appeals to the sobbing sap in me.   when i listened to that song five billion times, i got into the mood suddenly to listen to  urlLink  the dresden dolls  . they are dark and emotional at times... i craved the song, \"half jack.\" it was weird how i began to listen to it on a whim because it\'s one of my faves of the band, and then i started to feel intensely sad. the lyrics are vague enough to where i could relate to them and the pain that is encapsulated in amanda\'s voice and those powerful words. the song opened a flood of thoughts and emotions -- opening the doorway to stuff i don\'t like think about. not too many songs do that these do they?  most of the time we listen to music to escape and forget our troubles. but the dresden dolls love to open up and put the spotlight on their troubles. wallow, even. or poke fun of. their music is exposed, and listening to songs like \"half jack\" is like picking a scab over and over to watch it bleed -- until finally it forms a scab and you can move on. quite cathartic.   '),(15007,1694057,'23,February,2004','2004-02-23 00:00:00','   old man look at your life  my boss would faint is she knew i was listening to  neil young  right now. in fact, i think i\'m about to faint.   '),(15008,1694057,'23,February,2004','2004-02-23 00:00:00','   the art of storytelling  anna\'s story about our run in with the split pants guy is up on  urlLink really small talk . her perspective of the sitch is funnier than mine.  also tonight:   urlLink watch dan kennedy as he hosts the moth  -- in another life, jco and i were interns at  urlLink the moth  for about two years. i like to think of them as the golden years of the moth where we had amazingly talented storytellers like mike albo, shelagh ratner, jonathan ames, frank mccourt, hazelle goodman, romy ashby, edgar oliver, sarah jones, candace bushnell, albert maysles, george plimpton, jerry stahl, joey xanders, reno, timothy \"speed\" levitch, Philip Gourevitch and many many more. now i think the only person i know at those events is george dawes green, the founder and funder. i haven\'t been to one since kimya dawson, dan kennedy and gideon yago told rock and roll stories!  also, i think i am in love with this  urlLink cass mccombs  ep,  not the way  and his full length,  A . love love!   '),(15009,1694057,'22,February,2004','2004-02-22 00:00:00','   bushwick blast  for two weekends in april, brooklyn is going to rock during the first annual  urlLink bushwick blast . some of the most interesting, experimental bands from all over the country will converge at the asterick art space in bushwick to change the way you listen to music. oooh i can already tell it\'s going to be phenomenal...   '),(15010,1694057,'22,February,2004','2004-02-22 00:00:00','   dennis cahlo   urlLink dennis of the realistics  has new mp3s up on his site. check out his next show -- it\'s free!  Dennis Cahlo Friday, Feb. 28th @ The Living Room 11 p.m.   '),(15011,1694057,'21,February,2004','2004-02-21 00:00:00','   more lit shit  went to see another friend\'s band  urlLink circle and square  play at lit last night. i was really exhausted from partying the night before, but got myself siked up for the show anyways. two nights hanging out at lit is kind of too much. but such is life. circle and square is a band that has been around for a while. they have their shit down. so when they\'re on stage, all four members perform tizzight. it just looked so natural. to me they kind of sound emo-ish, punk... very catchy.   other bands performing were  kilowatthours  and  soundtrak . there was one song that kilowatthours played, it was the last song of their set -- an instrumental -- that was just really pretty and magical. i was actually surprised. it was gentle at heart. i like! i like! by the time soundtrak got on stage, i wasn\'t really paying attention too much. but from what i heard, they are very strokes-ish. i will have to see them again some time i think to really form a solid opinion. what d\'ya\'ll think?  some peeps were hungry so we went foraging the cold city for food. i was a dumbass and wore a skirt -- so i froze my ass off all night. we all headed over to the new  urlLink misshapes  party in the west village at a bar/lounge called luke and leroys -- a two story mini-club. we got there kinda late -- 2-ish, so the party was actually kind of winding down. we hung out for a little bit -- it was good to see all the bloggers in tow. and the most outrageous part of the party was the last song i stayed for -- the walkmen\'s \"the rat.\" ultragrrrl instantly pulled me on to the dance floor and a bunch of us freaked out, screaming those hammy lyrics. love love!  spent the rest of the night at a friend\'s place just listening to loads of really cool music that was new to me. it was pretty awesome. then i completely passed out, in a drunken haze that had me in and out of consciousness. i hadn\'t stayed up that late in forever. it was a crazy, but memorable night. thanks to all.   '),(15012,1694057,'21,February,2004','2004-02-21 00:00:00','   all lit up  i went to see anna\'s band  urlLink knife skills  last night. it feels really good to support a friend\'s band, especially because you get to see them grow and bloom as a performer. i feel like every time i see this band, the assault is more massive, more killer. their songs are short, to the point but just as deadly, like sharp, quick jabs to your racing heart. if you haven\'t seen this band, you haven\'t seen tough, sexy girls kick ass on the stage. i\'m not talking about girls who LOOK tough. but girls who really let go and lead with their aggression. if you haven\'t seen this band, you are missing a phenomenal drumming maniac. i\'m not talking about some guy who just bangs hard on the drums. i\'m talking about insane beatings that would make the most vicious domestic abuser blush with envy. in short, get thee to a knife skills show soon! if you think you can handle it.   the forms  played afterwards, and they were pretty good through the first three songs. energetic, swaying, melodic, notes elliding into others... and then they began to sound the same on every song. they had a lot of energy and oomph, but could probably stand to fill out more diverse forms to break up their set a little bit. but perhaps the point was to reject form and standard lyrics for the more abstract vocals and the kind of music that melts into itself... a very interesting concept if you want to intellectualize it. i could be taking my thoughts too far though!!  i am so totaled now. i thought that you couldn\'t get hangovers with liquor, but i proved myself wrong by drinking the most i\'ve drunk in a while: three makers and gingers and a shot of tequila. serious shit, man. and i think i\'m ready to do it all over again tonight at the circle and square show. watch out for this girl. she is on a roll. she is talking in the third person... ok i\'ll stop now.   '),(15013,1694057,'20,February,2004','2004-02-20 00:00:00','   be our friendster  i finally went ahead and did it. i created a friendster profile for dirty on purpose.  urlLink add us as your friend!   also i will be at lit for the next two nights:  friday, tonight!    urlLink knife skills   quick fix kills the forms @  urlLink lit  9 p.m.  afterwards, head over to see  urlLink jo divestar  deejay at black and white! (10th Street, betw. 3rd & 4th Aveâ€™s in Manhattan)   saturday, tomorrow:   urlLink circle + square  kilowatthours @  urlLink lit  9 p.m.  then head over to greg\'s new party,  urlLink MisShapes  in the west village at luke + leroys (11 p.m. to 4 a.m.). deejays are gonna be hawt:  urlLink miss modernage ,  urlLink ultragrrrl ,  urlLink aleix + inge (girlie action pr)   '),(15014,1694057,'19,February,2004','2004-02-19 00:00:00','   The Cooper Temple Clause  â€œPromises Promises\"  urlLink Real Player   urlLink Windows Media   â€œBlind Pilotsâ€  urlLink Real Player   urlLink Windows Media   '),(15015,1694057,'18,February,2004','2004-02-18 00:00:00','   mall-o-riffic  today anna and i ventured into the big time warner mall on columbus circle. we are happy because there is a  whole foods  and  whole body  in this mall. there\'s also a sephora and the borders where the  urlLink fab five  signed books yesterday. i feel like i\'m back in suburbia. lunch time has become a little bit more fun. i\'m eating melons like crazy.  on our way back to work, a guy dropped something on the sidewalk, picked it up and then said to me, \"i split my pants!\" it was a total line but i didn\'t realize it at the time. he wanted me to look backwards at his ass. i didn\'t look, but anna had the pleasure. she said he looked back and waved. the whole funny sitch kind of brightened up my day!   '),(15016,1694057,'18,February,2004','2004-02-18 00:00:00','   the show to be at  according to me and y\'all: tonight @  urlLink southpaw  pela, tincantelephone, moonraker $8 // 8:30 p.m.  word to your moms.   '),(15017,1694057,'16,February,2004','2004-02-16 00:00:00','   bling bling  still haven\'t seen bling kong yet. looking forward to seeing them tomorrow at vicious @ sin-e. should be good!   '),(15018,1694057,'15,February,2004','2004-02-15 00:00:00','   winter be gone!    saw the video for  urlLink the boggs \' song \"plant me a rose\" on  urlLink cliktrax . it was set in new york during the winter with snow on the ground. i don\'t know about you but i\'m so over new york winters. yet i still like the old school feel of the boggs video. it felt like a century old reel of normal life in the city.   '),(15019,1694057,'15,February,2004','2004-02-15 00:00:00','   the winter pageant need a drummer!  from  urlLink the winter pageant : the winter pageant still really needs a new drummer.  temp to permanent.   bear with my many posts, this is a rough setback to some fantastic momentum - we really do need all the help we can get, because john is so so so so so good.....and this music calls for a really unique player and individual.  i\'ve gotten lots of response so thanks to those who have helped so far......  please forward to anyone who you think might possibly know someone.  again, we\'re doing so well, and this is a great project to be involved in.   we just got a friday showcase at sxsw - this is huge.   -- byron music @ www.thewinterpageant.com  urlLink notbrian@earthlink.net  917.676.6201  p.s. upcoming shows: we\'re playing with french kicks at mercury on the 25th of feb, and then a fringe benefits party at pussycat lounge with the occasion and mountain men on march 3.   '),(15020,1694057,'15,February,2004','2004-02-15 00:00:00','   Blur - There\'s no other way  You\'ve taken the fun Out of everything You\'re making me run When I don\'t want to think You\'ve taken the fun Out of everything I don\'t want to think at all  There\'s no other way There\'s no other way All that you can do is watch them play  You\'ve taken the fun Out of everything You\'re making it clear when I don\'t want to think You\'re taking me up when I don\'t to go up anymore I\'m just watching it all  There\'s no other way There\'s no other way All that you can do is watch them play There\'s no other way There\'s no other way All that you can do is watch them play  (Ahh, ahhh, watch you play Ahh, ahhh, watch you play..)  There\'s no other way There\'s no other way All that you can do is watch them play There\'s no other way There\'s no other way All that you can do is watch them play There\'s no other way There\'s no other way All that you can do is watch them play There\'s no other way There\'s no other way All that you can do is watch them play   '),(15021,1694057,'14,February,2004','2004-02-14 00:00:00','   more photos from vegas  i was in the worst mood until keith bought me this las vegas trucker hat to wear. then my old, goofy self came back and i was able to enjoy las vegas despite the cheesiness. i played black jack for an hour, and left only five dollars down. not bad! it was fun. i felt like i was in one of those celebrity poker showdown episodes, except i was playing black jack.   '),(15022,1694057,'14,February,2004','2004-02-14 00:00:00','   mr brightside  walking around las vegas is like being in a pedestrian terror trap. unlike new york, las vegas is anti-pedestrian. but after hearing from the hotel clerk that there was a 45-minute wait for cabs, i was like FUCK THAT I\'M GONNA FUCKING WALK, in true new yorker fashion, of course. i started walking over to the strip, in a cynical kind of mood. but it was just so nice outside. i swear it felt like it was in the 70s, a pure joyous contrast to the east coast weather i was trudging through only a couple days earlier. warmth = good mood. i ambled on the streets directed only by the huge \"harrahs\" hotel silhouette on the skyline.   when i got there, i followed some people walking into the harrahs. of course i saw the casino floor and just sort of walked through the entire thing. the ka-ching of the slot machines were kind of enticing -- reminding me of the time i went to detroit and went to canada to gamble in the windsor casinos. i had lost $20 that day. today i wasn\'t really in the mood to play much of anything. soon i realized that there wasn\'t much to do in las vegas if you weren\'t in the mood to gamble. but seriously, who gambles on their own, right? i decided i would do that tonight when keith is around. today i would just walk around, soak in the gaudiness, take photos of things like the tourist i am, and go shopping! i\'ll spare you the details but here\'s some eye candy. i\'m hoping to feel the energy of a song like  the killers \' \"mr brightside\" while i\'m here in vegas. so far, no go.   '),(15023,1694057,'13,February,2004','2004-02-13 00:00:00','   colorful evening, sleepy days     urlLink joe and george  and i went to see  urlLink  the color bars   last night at sin-e. i was listening to their cd earlier in the day, and wasn\'t sure i was into the pop-alt-country-funny songs. but what was particularly delightful about the color bars in their live performance is their beautiful three-person harmonies. it\'s pretty amazing that three people can sing so in synch. it was like watching an alt-country version of hanson!! while their lyrics were kind of silly (intentionally, i\'m sure, but still not too impressive or enticing), some of their songs were really well-assembled rock songs with beautifully swaying guitar lines. it was weird because the band would go from plain goofy to really intricate, sweet music. it was also the kind of band that switched instruments a lot, which was cool and brought in different dynamics to their performance. i also loved it when they gave away valentine\'s day cookies that one of the guy\'s girlfriend had made. sweet.      originally we were at the show to meet up with some peeps for band business shite, but couldn\'t find them in the crowd. i was kind of annoyed because i missed seeing  urlLink  sam champion   at luna lounge -- sorry about that noah and jay. i didn\'t miss you guys on purpose!! and i promise i will catch you next time. sorry sorry sorry.     anyhoo while i was there at sin-e, i talked with bill britelle of  the blondes inc  for a little bit. and he introduced me to this guy who\'s an artist -- a pretty amazing artist, in fact -- very basquiat. i think his artist moniker was fa-q -- pronounced like \"fuck you.\" i saw his portfolio of work, and was instantly taken in by the complex emotions and raw display of pain and anger in his work. this guy should not be working at sin-e. he should keep on with his art and have shows again in new york city.       it was kind of a late night, and j&g and i headed over to my place. i had like almost an entire pizza in my fridge from the night before that i needed to get rid of, so i heated it up in the oven, and we all had a late night munch. it felt like i was in college or something. we ended up watching elimidate -- a show i\'ve never seen before, and soon found it was kind of fucked up!! all the girls on the show kept bitching at each other and the jockish guy was totally grossing me out. \"i wanted to squeeze her butt like they were georgia peaches.\" YUCK-O.  after j&g left, i had about an hour to pack and clean up the house. i didn\'t have any time to sleep. before i knew it i was heading to laguardia. i thought i was gonna pass out!! but the music list on my ipod helped me get through my 8 hour plane ordeal to vegas. here\'s what i was listening to: \"mind blindness\" - dirty on purpose \"monument\" - dirty on purpose \"latitudes\" - pela \"all in time\" - pela \"what\" - brendan benson \"quarter to three\" - bishop allen \"don\'t hold your breath\" - calla \"mother sky\" - calla \"lovefool\" - cardigans \"delicate\" - damien rice \"love is only a feeling\" - the darkness \"here i dreamt i was an architect\" - the decemberists \"coin-operated boy\"- the dresden dolls \"in this city\" - enon \"crying just for show\" - french kicks \"close to modern\" - french kicks \"1985\" - french kicks \"trying whining\" - french kicks \"the devil wears earplugs\" - the joggers \"mr brightside\" - the killers \"rocknroll\" - lovedrug \"down toward the healing\" - lovedrug \"to hell with good intentions\" - mclusky \"tomorrow\" - morrissey \"happy kid\" - nada surf \"inside of love\" nada surf \"big brat\" - phantom planet \"1st things 1st\" - phantom planet \"nothing better\" - postal service \"lipgloss\" - pulp \"common people\" - pulp \"disco 2000\" - pulp \"collide\" - rachael yamagata \"house of jealous lovers\" - rapture \"samson\" - regina spektor \"waiting\" - the rentals \"wonderwall\" - ryan adams \"gone for good\" - the shins \"the end has no end\" - the strokes \"inoculate the innocuous\" - the unicorns \"the rat\" - the walkmen \"my old man\" - the walkmen \"little house of savages\" - the walkmen \"138th street\" - the walkmen \"maps\" - yeah yeah yeahs \"y control\" - yeah yeah yeahs  the hotel room is nice. but i totally crashed and slept all day. i got to vegas around 10 a.m. (vegas time). so crazy. completely exhausted. i ordered room service, ate heartily and then went to sweet sleep. keith had his laptop all set up with internet connection for me (he knows me so well -- i\'m completely addicted to the internet and would go crazy if i didn\'t have a computer on this short trip). tonight we\'re supposed to go to some party that his company and ibm are having. then we have the rest of the night to do whatever we want. originally he had told me that friday night would be stripper night... i dunno if i\'m totally down with that, but i guess we\'ll see what the night will bring!   '),(15024,1694057,'12,February,2004','2004-02-12 00:00:00','   the walkies talky   urlLink read and link to my interview with hamilton of the walkmen  love jinners   '),(15025,1694057,'11,February,2004','2004-02-11 00:00:00','   joke of the day  what did five fingers say to the face?  slap!  somehow it\'s funnier when joe tells it.   '),(15026,1694057,'11,February,2004','2004-02-11 00:00:00','   rough trade  this week has been kinda rough. some good things, some bad things. there has been this huge anxiety building up within me these past couple days. i\'ve been trying to fight it off, but it\'s kinda weighing me down emotionally. though i always try to be my perky lil\' self. think happy thoughts. happy thoughts!  ...  urlLink pela  . whoa, they totally stepped it up a notch or 10 at their mercury lounge show on tuesday. the volume of their show has gotten considerably louder, and billy is as passionate at the mic as ever. these are the kind of guys who wear their heart on their sleeves. love love them all! they also have this great new drummer who makes rock faces while he\'s pounding the beats. i love bands that look like they\'re having a lot of fun while their songs are bursting out of the monitors. full force. a good performance needs great songs, great spirit, and charismatic players. all was had by pela. also had lots of fun listening to their manager, lee, tell his biz stories. this is what i imagine touring life must be like -- sitting around in musty rooms with my peeps, smoking and drinking and telling stories that draw laughter and bring us closer together. sorry i don\'t have any photos. i brought my temporary digicam with me but then realized there was no smart media card in it. d\'oh!!   '),(15027,1694057,'11,February,2004','2004-02-11 00:00:00','   forget simon cowell. judge for yourself:   urlLink pitty sing - \"radio\"   urlLink denali - \"hold your breath\"   urlLink on!air!library! - \"bread\"   urlLink south - \"loosen your hold\"   urlLink young people - \"el paso\"   urlLink call and response - \"rollerskate\"   urlLink the vines - \"ride\"   urlLink usher   '),(15028,1694057,'10,February,2004','2004-02-10 00:00:00','   sin city  ooh it\'s official! i\'m leaving FOR a very special valentine\'s day in las vegas. wheee!!! don\'t worry. i\'m not going to pull a britney or anything. but i might get lost in translation. gaudy hottie tottie!   '),(15029,1694057,'09,February,2004','2004-02-09 00:00:00','   french kicks on carson daly  .... again. in fact the taping is today. you better act fast if you want to go.  urlLink click here   '),(15030,1694057,'09,February,2004','2004-02-09 00:00:00','   yeah yeah yeah, no grammy    from nick,  urlLink catchdubs.com  unfortunately, the white stripes won. what\'s up with that?   '),(15031,1694057,'08,February,2004','2004-02-08 00:00:00','   no week for the weak  what a crazy week it\'s been! i started off friday morning at work, and there was an explosion in my building. the entire place was evacuated and eventually people were sent home. this was fine with me because i had a lot to do to prepare for the dop show at the knitting factory... but it was kinda scary seeing thick black smoke coming out of the side of my buliding. i think the first thing people thought it was was a terrorist attack. all they smelled was smoke and then when they looked out the window, they just saw black smoke. scary eh? people were running out of the building before anyone even announced an evacuation.  anyhoo, the  dop  show with  suicide girls  was great. the place was packed, and our fears that they\'d get boo\'ed off stage in favor of naked girls was forgotten once the band started playing. there were a lot of new faces in the crowd and people were into it. their set was 25 minutes long -- short and sweet. it was cool because after their set, people in the crowd asked for more. and they finally got to premiere their hot new song, \"mindblindness.\" that song just drowns in emotion and builds up to an epic finish. sigh.  i really liked the people who worked at knitting factory. thank you especially to johnny the night manager. you are a total sweetheart!!! mwah!  so good to see bsiii in the house. i was happy that he was really into dop -- i had given him their cd at the last  cousin  show i went to at happy ending, and he came out to support. talked with him for a little bit but then lost in him the crowd later on. he\'s going to have a birthday party coming up at sin-e this month on feb. 22. cousin will play. it will be rad. come out and celebrate!  arthur from  stellastarr*  also came out to check out the show. one of my friends told me she didn\'t recognize him with his shirt on! it\'s ok though because he looked just as cute with clothes on. ironically, later in the burlesque part of the night, there were many topless girls with tape on their nipples. that\'s such a stellastarr* rip off! haha. saw daniel from  circle and square  at the show, and i was very delighted to hear that he liked the show. ed from circle and square was also in the house representing. we also ran into doug from  the go station , jeremy from  autodrone , heather from  au revoir simone , and many more that i can\'t seem to think of at the moment. forgive me, it\'s been quite a week.  on a side note, lately the band has been calling me \"pickles\" despite the fact that i keep telling them i\'d rather be called \"penguin.\" i think i\'m going to have to just accept this new moniker. but i thought another nickname for me uttered later that night was even funnier: \"america\'s sweetheart\". just as long as no one calls me \"all-american girl\" i think we\'re a-ok.   the band\'s show with pela last night at b61 in carroll gardens was super fun. it was so comfortable and free with lots of warmth. the bad part was that i got more than tipsy, dropped and broke my camera. while that\'s getting fixed, i will be using my bf\'s little digi-cam. cute and just as effective. pela was really amazing. it\'s been so long since they\'ve played a show so we were excited they asked dop to play with them and even happier to see them play. i think the last time they played was october! they played my old favorite: \"latitudes\" and my new favorite \"all in time.\" there were some technical difficulties -- the band played so hard that they blew out the power! but that\'s awesome in my book. the more passionate, the better.  dop also played a fun set that included one of my faves that still hasn\'t been recorded yet: \"air show disaster\". it\'s this huge instrumental that always makes me want to cry. it sounds cheesy, but when i listen to it, it affects me so deeply i want to weep or something. what can i say? i\'m a sensitive sucker. george broke a string and the band, like pros, went into an impromptu performance of erika\'s new keyboard song. it\'s a very low-key song with a beautiful flow. hopefully they\'ll play it at another show soon...  here are some new photo galleries: *  urlLink french kicks @ mercury lounge  *  urlLink dirty on purpose with pela @ b61  *  urlLink dirty on purpose @ mercury lounge  *  urlLink dirty on purpose with suicide girls @ knitting factory  *  urlLink knife skills @ north six    and some cool upcoming shows this week:  feb. 10:  pela  @ mercury lounge feb. 10:  steven mertens - solo!  @ cinema classics feb. 10:  dresden dolls  @ joe\'s pub feb. 11:  french kicks ,  peelander z  @ merc lounge feb. 12:  sam champion  @ luna lounge feb. 12:  ambulance ltd  @ pianos feb. 13:  ex models ,  roger sisters  @ merc lounge feb. 13:  asobi seksu  @ sin-e feb. 14:  regina spektor  @ tonic feb. 14:  my favorite  @ the hook feb. 14:  the wau wau sisters  @ bowery ballroom  check my  urlLink nyc shows calendar  for more!  buy  the recoys  cd at  urlLink realisticrecords.net !  see the amazingly cute stills  urlLink aeri  took from her video footage at the triple walkmen knitting factory shows!  omigosh, i almost forgot to mention that i saw  the lilys  downstairs at the knitting factory the same night of the dop show. they were amazing and surreal and beautiful. i was captivated with their low key set. i want to see them again soon!   '),(15032,1694057,'06,February,2004','2004-02-06 00:00:00','   strip it down, honey  it\'s official. the  suicide girls / dirty on purpose  show tonight at the knitting factory is completely sold out. not even any tickets at the booth. try your hand with the scalpers! but for those of you with tickets, i will see and party with you tonight! drinks! rock! and probably a lil\' more than janet\'s nipple.  by the way,  the walkmen  perform on conan tonight!  i just saw an snl commercial where meghan mullaly ripped clay aiken\'s shirt off to reveal a nipple star. weird! how long are they going to be \"milking\" this joke?   '),(15033,1694057,'05,February,2004','2004-02-05 00:00:00','   cuts like a knife     urlLink knife skills  ravaged the north six stage last night. they opened the night with their fast-paced, dangerous riffs. anna let loose on the mic and guitar. fon lin exuded sex appeal. jerry beat the hell outta those drums. and they owned it. so fun! it was good to see my dop crew in tow. lowbrau was in the house representin\'. and the lovely katy. and her friend jenny. i saw a little bit of  parts and labor , and really dug what i heard. experimental, loud but in a strangely catchy, dancey way. the music moved.   '),(15034,1694057,'05,February,2004','2004-02-05 00:00:00','   awful plastic surgery  i don\'t know where katy finds these things...  urlLink awful plastic surgery   '),(15035,1694057,'04,February,2004','2004-02-04 00:00:00','   the walkmen on truthflies.com  if you buy the new  walkmen  album,  bows and arrows , on  urlLink truthflies.com , you\'ll also get an autographed poster! this is a limited offer so scoop it quick! it\'s only $12, but you get BIG LOVE. ed, you better save me one since i hooked you up!   '),(15036,1694057,'04,February,2004','2004-02-04 00:00:00','   fast suicide  word on the net is that  urlLink the dirty on purpose suicide girls show  is sold out, at least online. hope you got your tix!  tonight:  knife skills parts and labor lungfish  @ north six   '),(15037,1694057,'04,February,2004','2004-02-04 00:00:00','   french kicks!!!     urlLink  the fucking french kicks   made a triumphant return to new york city with the first installment of their residency at the mercury lounge. fucking hawt, people. FUCKING HAWT! i was amazed when they took the stage and wowed me with wondrous, melodic, pure-joy-filled new songs. my hips were swinging and my head was bobbin\' with jaunty jubilee... whee!  i know i sound silly but who the hell cares? drummer #2, the post-hugh drummer, quit the band so they temporarily used jake from the joggers, who apparently only practiced with them twice before the show. i was quite impressed. he hit those beat, and he hit them with sassssss! flirting with the crowd! the keyboard and guitar parts were as catchy as ever and josh\'s voice was featured on one new song. some good quality shit here, peeps! they sounded waaaaay better than they did at irving way back in the day. that venue just plain sucks. mercury is the bomb! no doubt! the whole time i felt like i was a little too close to nick\'s crotch. but at a french kicks show, i guess that\'s not a big surprise. that mofo is one tall drink of water. a special thanks to  urlLink isaac . isaac, you are the man. awesome possum and whatnot. matt! lawrence! nick! josh! jake! love love.  met jay from flavorpill, and had a blast! we talked about all sorts of music and had great conversation. he introduced me to the man of the night -- johnny beach who books mercury. johnny was the nicest guy ever. i had so much fun jabbering away, i had to apologize for talking so much. sometimes i get in these weird moods where i have insane motor mouth. it just won\'t stop! but i promised we would go out drinking and have lots of fun after he returns to the city. wonderful totally cool guy. love love!  anyhoo, also chatted with peter bauer about all things  walkmen ! i had heard from a friend that the \"listen to this\" supplement in entertainment weekly this week was going to have a walkmen drinking special in it... so i asked pete about it and he said that the writer just drank with them and asked them what they drink. pretty lame stuff apparently... we\'ll see how it turns out. he said  weird war  may or may not join them when the band attacks irving plaza. i wish i had gotten a chance to say hello to hamilton. i never know if people connect jinners to jin, you know? but apparently hamilton does. i told pete i interviewed hamilton recently, and he told me that he knew because hamilton had told him. the whole time i thought i was interviewing all incognito and shit! no more, jinners, no more. mental note: come up with new incognito shit.  earlier in the night, stewart lupton\'s band,  the child ballads , were actually quite impressive. stewart has the kind of voice that makes you think of crazy legends like tom waits or bob dylan. i liked the folkish feel of his band. it was weird because i was expecting jonathan fire*eater redux! but got some quality rock music. straight from d.c.! it\'s kind of a shame because i think a lot of people didn\'t give it a chance. there were some lulls, but for the most part, quite dramatic.  i still don\'t know what i think about  the shout out louds ... my initial reaction was a non-reaction. but i think i was so distracted i couldn\'t form an honest opinion. will have to see them another time. i know jo divestar loves them so they must have some merit!  at the end of the night a bunch of us went to starfoods! of course. i had my standard makers and ginger. i chatted up a storm with new friend jess. shared a cab with new friend justine, who i found out lives a couple blocks from me. it was good to see jco and rachel back in action. i had so much fun! i might end up going all four nights at the merc! crazy? could it border on walkmen love? we\'ll see...    before all of the musical chaos i went to see adrian tomine at the housing works cafe. great place to hang out, i realized. i briefly saw jason from  urlLink product shop nyc , but i don\'t think he saw me waving to him. i got adrian to sign my book and i was a stupid flustered nincompoop. damn!!   '),(15038,1694057,'04,February,2004','2004-02-04 00:00:00','   things to do tonight   adrian tomine  signing optic nerve 9!!! @ housing works cafe 7 p.m.   french kicks ,  shout out louds  @ mercury lounge 9 p.m.   qatsi  @ sin-e 10 p.m.   '),(15039,1694057,'03,February,2004','2004-02-03 00:00:00','   love and lost  ok, so... flavorpill fixed the fact that dirty on purpose wasn\'t mentioned online. whatever. i still think they\'re journalistic dunces. not in a harsh way. just from one journalist to another. now i know i can\'t exactly trust their reviews. moving on...  anyways, back on to the music i\'ve been listening to. on saturday night  urlLink  the gostation   played tiswas. the crowd was full and lively. it was weird how i was there to see and support singer doug levy, a recent acquaintance. then i realized i knew another member of the gostation, their guitarist, from college. small small world! the gostation is brit pop gone rock\'n\'roll on u.s. soil. relatively new to the scene, i was digging their oasis ways... it was weird to hear them also cover an oasis song, but doug had the lead singer moves all down pat. watch out, liam! their set was energized and emphatic. looking forward to seeing them again soon...   urlLink  aerial love feed   headlined the night. i had seen them once before at vicious. the first time i wasn\'t really sure what i thought of them. they were loud. they were stylish. and they were strobin\'! and geez were they strobin\' again at tiswas. however, my opinion of them changed for the better. strange, eh? the only thing i didn\'t like was the fact that their set started to sound more and more like one big song. they had energy and the rock star poses (which are great, actually), but not enough differentiation between songs. and way too much strobe. please stop the strobe. it\'s great for finales, but not for a 35 minute set!  also, in case you were wondering, my interview with alex greenwald of  urlLink  phantom planet   went well! he seemed a lot younger than i expected, but he was a talker and easy to chat with. i can see why the teeny boppers like him. while listening to the new phantom planet album, i felt like dancing. i think that\'s a good sign. who cares if their last album was a pop-tumor? though i did like that \"turn shift repeat\" song. anyways back to the fact that it makes me want to dance. dance!   '),(15040,1694057,'02,February,2004','2004-02-02 00:00:00','   the good and the bad  happy  walkmen  bows and arrows   day! go buy your very own copy of their record today!  I was really underwhelmed with the Flavorpill review of the  Suicide Girls  show this Friday. Could everyone  urlLink email them  and let them know they forgot to mention the featured band,  Dirty on Purpose ?? I will love you forever if you do me this favor.  in other news, popmatters finally posted  urlLink my review of  the unicorns  at southpawâ€¦   also, I am interviewing alex from  phantom planet  today over the phone. we\'ll see how it goesâ€¦   '),(15041,1694057,'02,February,2004','2004-02-02 00:00:00','   the walkmen are A+++++  entertainment weekly give  the walkmen \'s  bows and arrows  an \"A\". fucking hawt. they were the featured review --  urlLink full page review!  i kind of don\'t like the comparison to the strokes though. they are NOT the strokes! get over it! they are the walkmen! thanks.  i read that time out interview you\'re all talking about. can\'t they get through one interview without mentioning neil hagerty?? the answer is no. haha.  all i have to say is IT\'S TRUE!!    also: Reprise Records says it will begin selling the debut album from rock act  Secret Machines , \"Now Here Is Nowhere,\" as a download three months ahead of its physical release. The move pushes the envelope on the lead-time major labels are giving to online sales of new releases. as part of the reprise offer, those who buy the album online will receive six free additional downloads, a bundled collection called \"sympathy for the download 01.\" the digital sampler will feature a secret machines bonus track and cuts from the upcoming von bondies, the walkmen, the sun, head automatica and johnathan rice. fans will also receive a free CD-R to burn the tracks on to. the cd-r is a limited edition featuring exclusive secret machines artwork. [billboard.com]  i also got a chance to listen to the von bondies. that album rocks! can\'t wait to see them when they come to nyc feb. 25, bowery ballroom...   '),(15042,1694057,'02,February,2004','2004-02-02 00:00:00','   damien rice revised  New Remix of Damien Rice\'s \"Cannonball\" released: Windows Media Version:  urlLink low  //  urlLink high   urlLink QuickTime  damien rice will be coming back to nyc on april 16, beacon theatre...  i saw  the go station  and  aerial love feed  on saturday night at tiswas. full review coming soon!   '),(15043,1694057,'29,March,2004','2004-03-29 00:00:00','   the walkmen show is sold out  Oh.my.god. Thank gosh i bought tix for my friends yesterday!   '),(15044,1694057,'28,March,2004','2004-03-28 00:00:00','   eternal sunshine of the spotless mind  Yesterday my new movie buddy mike and I went to see one of the most amazing films I\'ve seen in a long time...  eternal sunshine for the spotless mind . We were both in the perfect mood to see this film. It was rather sad, but what a beautiful story, full of just the right amount of romanticism and whimsy balanced out with the gloomy, cynicism so many of us feel about love. I cried, and wiped my tears with a stupid movie theater napkin. I immediately wanted to see the film again! Hope it comes out on dvd soon so that I can watch it over and over! This could be my next  high fidelity  or  me without you . The movie caused mike to not finish his icee, which apparently is an event that has not occurred ever. So there u have it. Go out and see this movie because it is  that  fucking badass hot shit (not \"nice\").  I also convinced mike to \"make party\" with moi and meet tom at  urlLink the hook  in red hook. We hopped in a cab and made our way over to the club, which is pretty easy to get to with good directions. We didn\'t know what to expect when we got there but was instantly struck by how beautiful and elegant and cool the club looked. As mike would say, \"it is the jam,\" or simply, \"hot shit\"! Ha, and he wonders why he gets carded for R rated movies!   Tom gave us the full tour of the place, which has this amazing downstairs \"green room\" and then a back outside area where I can already tell will be amazing to hang in during the summers with a juicy bbq set up. I was so excited when tom asked mike and I to deejay the club on may 7. So come see me and dj hot shit human-jukebox! I\'ll let you know deets as soon as I have \'em...  So we went to see the bands  the dears  and  sons & daughters . I had been hearing about the dears for quite some time now. In fact they were the band I got once on an indie rock band quiz. I had pretty high expectations -- looking forward to getting engulfed in a wall of what I call depressionistic music. And they were dark and depressing at times, but would also often times roll into a psychedelic haze of long, long jams...  The band I was really excited to see was sons & daughters. I have been curious about this band, the kind of band u hear about for a while and suddenly u see them and instantly think, Amazing! I believe they are from glasgow... With a guy and girl singing and harmonizing lyric lines that spoke of unrequited love and dreary states of mind: \"you made the first move when I\'m near you, don\'t walk away.\"  The guitar parts were folkloric with incredible, passionate strumming.... I thought for sure they would break a string or two. Some of the songs also swaggered with cool struts, moments that made me take notice of how tight this band\'s performance was. I think they play knitting factory soon, so go check \'em out if u can!   '),(15045,1694057,'26,March,2004','2004-03-26 00:00:00','   a night of dirty furries  i\'ve been moving into my new digs, located centrally in the wonderful east village. i am really excited about this change because i feel like the east village has been my hood for quite some time now, and now i\'ll be able to walk home from a rock show. save money on cab rides, yo!!  mike came over late afternoon with a bottle of wine, and his friend anna who i met at sxsw joined us. it was fun just sitting in the living room and having a glass of vino. then i was on my way to pianos for dop\'s soundcheck. everything was swell in the air. i was light on my feet, breathing in the city.      the rapture and the furries @ knitting factory  tom had invited me to a private party at the knitting factory for cartoon network. his gf lexi got the hook up, yo!  har mar superstar  and  urlLink  the rapture   were playing. i missed har mar, but after all the sxsw drama with har mar, i wasn\'t too sad. i haven\'t written about it yet, but har mar beat up some guy who threw a big penis prop at his head. he got all huffy and refused to finish his set... after only performing three songs and stripping down only twice.   i got to the knitting factory right before the rapture went on stage. there were all of these people in animal and pirate costumes walking around. very bizarre. i thought maybe it was a cartoon network thing? but then i also heard that the rapture recently did a photo shoot with a bunch of people in animal costumes, so maybe it was a rapture thing. i dunno. either way it was lots of fun! the open bar was a complete open bar, meaning  you could order anything you wanted. so i went all out and ordered two grey goose and red bulls. i wanted to get some caffeine in me for the long night. i was totally double fisting it, and it was awesome. as i finished one drink, the rapture were stepping on stage.   i quickly ran to the front. the thing i love about seeing the rapture live is that i feel free to dance like a crazy person. there are few indie rock bands that i feel comfortable doing that with. but the rapture is one of them. something about that cowbell and luke\'s crazy robert-smith voice, it just makes my legs move and my feet jump. i was also really totaled. i started dancing with the polar bear. i was singing along! we were dancing for joy! they danced with furries. luke jumped into the crowd and danced. he jumped on amps and rocked. the cowbell/sax dude was going all dancey. everyone was dancing! it was crazy. at one point, luke sang to an owl bobblehead. it was really hilarious! man, the had to have been totaled to be so goofy. luke gave me a high five after their set ended. i was so happy!! i think all of our dancing really hyped up their performance.   and afterwards the polar bear came up and introduced himself to me, without the costume. apparently, he\'s in a band too -- his band rahim is playing  urlLink misshapes  on saturday at 10:30. i don\'t know what they sound like or what he plays in the band, but any guy who dances with me at a rapture concert in a polar bear suit is fucking ROCKIN\' in my book. word! plus  urlLink dennis cahlo  (realistics) and audrey of  urlLink melodynelson.com  are also deejaying so there\'s three awesome reasons to go!  i think the pictures will tell you more than my words:  urlLink check out the rapture photo gallery   i ran into jason from  urlLink  the boggs   at the show, and he told me his band was playing a show at tommy\'s tavern in greenpoint on saturday night with tall boys and the double. it\'s a todd p party, and the boggs are one of my favorite new york bands, so you should check \'em out!     dirty on purpose @ pianos  then our crew zoomed on over to pianos to see  urlLink  dirty on purpose   play. shark sandwich had canceled last minute and a band called other passengers (formerly medicine hat) played. they were really good, and had a somewhat similar sound to dirty... the difference being the vocal arrangements and harmonies and basic quirky aspects of dirty like the guitar bow and the fact that dop also have four singers and five songwriters. dop really digged them, and was siked that they were added to the show.  when dirty took the stage i was really excited. they started off with \"monument,\" which was a crazy, dynamic way to start the set. they always get really hyped up for that song. i was still high from seeing the rapture too so i kept jumping and bouncing around. it was pretty awesome. they also played some songs on the upcoming ep, which will be released april 22. we\'re all really siked about the way the CD is shaping up, and goldenrod has been mixing like crazy! thanks to everyone who came out to pianos last night. i had the best time!   urlLink check out the dirty on purpose photo gallery here.   i hung out with miss jo jo dancer afterwards, it was really chill at a bar where her friends were dj\'ing. we were dancing to punk rock and pulp. how hawt is that? and we took silly photos like this one:    tonight i\'m going to see  urlLink a place to bury strangers  at pianos. come and join me in a night of LOUD rock. this trio is rawkin\' and i\'m sure they\'re gonna blow up very soon. i saw them at sxsw, and was completely entranced. plus, they\'re very swell peeps too. and i\'m all about supporting nice, talented musicians! and then over to black and white to see jo jo dancer deejay. i told her to save the walkmen\'s \"the rat\" for when i get there!!!   '),(15046,1694057,'24,March,2004','2004-03-24 00:00:00','   DOP + GBV = Pianos This Thursday  Jinners (that\'s me talking in the third person, sorry) would like to cordially invite you to see Dirty on Purpose this Thursday. Come party with me and Dirty, and we\'ll give ya a smooch. Here\'s the info below:  March 25 / Thursday / 8:00 / $8  @  urlLink Pianos   UP -- Late Night. Kick Ass. Rock n Roll Presents...  12: The Boxes  11:  urlLink Dirty on Purpose   10: Shark Sandwich (members of Essex and Guided by Voices)  09: Skycam  08: Function (Australia)   10:00/FREE!/upstairs lounge  Guest DJ\'s Kit Chaps (VHS or Beta)  2am Drink Specials    a coupla other things wholly unrelated to sxsw...  **  urlLink my review of the von bondies and the fever at bowery  **  urlLink photos of dirty on purpose at southpaw  by jasper at jenyk.com **  urlLink photos of dirty on purpose at sin-e  by jasper at jenyk.com  right now i am listening to  swearing at motorists \' \"i\'m still in love with you\" on repeat. good song!!!!!!!! good band!!!!!! they are from my hometown, philly.   '),(15047,1694057,'22,March,2004','2004-03-22 00:00:00','   SXSW: march 20, saturday daytime  it was one of those off-and-on gray days in austin, where the sky was somewhat shaky. i didn\'t know if the sun was just peeking out at me shyly or what, but i was ready to start my day early. i have to say i was in a sour mood when i woke up. i think i almost felt like crying, i was so emotionally exhausted. this past month has been a trying time for me. and i went downstairs to smoke outside the hotel, and then do some writing in the hotel lobby.   as i was writing, i saw this familiar looking guy on the couch. i suddenly realized this guy was dan wilson, the lead singer of that band semisonic, except now i think he only does solo stuff. he had played the ascap sundance music cafe in january, so i recognized him from those photos. you might think i\'m sorta bizarro, but i used to have a huge crush on this guy back when my favorite song used to be semisonic\'s \"secret smile.\" that song still makes me happy... it\'s the kind of sentimental happy that never goes away like when i hear that new radicals song, \"you get what you give.\" it reminds me of my romantic mishaps over the years, and i\'m overcome with whimsy.   so i had to do it. i went over and sat next to dan wilson and introduced myself, explained i worked for ascap and he did the sundance music cafe we produced and blah blah blah blah. he was supernice. i didn\'t tell him i used to have a crush on him. that will be my secret smile for now...     the winter pageant @ the tillery street theater  [winter pageant write up coming soon]   SXSW: march 20, saturday night    rachael yamagata @ la zona rosa  most amazing rachael lyric: \"not gonna shed one more tear for you until sunday afternoon\" [full write up coming soon]     matthew caws of nada surf  ...so when i had the chance to see matthew caws and ira elliot play a stripped down nada surf set, i was jumpin. it was pure. it was beautiful. lemme tell you, not too many songs can withstand the test of a simple acoustic strip down. most would run away from embarrassment after being so naked because they tend to be empty. but nada surf songs have tons of substance in their storytelling and the strength of their lyrics. plus, they are romantic, which always draws me in. the stage was short so matthew asked everyone to sit on the floor, and it reminded me of the intimate setting that the very first  on the moon music show  i threw last year in march had with the moldy peaches side projects at luxx, except the floors were cleaner.      ira elliot of nada surf  matthew is a gifted storyteller, and we were moths drawn to the flame of his heartache and openness. beautiful. they are working on recording new material, i hear, so i can\'t wait to hear that!! i was singing along to pretty much every song, except for the new ones (for obvious reasons), and at one point, i had to consciously lower my voice because i felt like i was singing louder than matthew!!! i didn\'t want to have a  rockstar  moment, if you know what i mean... i told ira about it later, and he laughed and said, \"that\'s so awesome!\"     michelle from read yellow @ club deville  [read yellow write up coming soon]     bishop allen  after read yellow, ran into the bishop allen kids everywhere at sxsw. love those kids, but every time they had a show, i was working at the trade show. sometimes things like that get me so down, but luckily i\'ve seen this kids perform so often, i could just sit at the trade show and think fondly about them and just smile. i imagined justin singing one of my favorite bishop allen tunes, \"quarter to three,\" and my face would light up. if you don\'t know these kids, get to know them now. they are the sweetest bunch with the most surprisingly good music.      ezt  [ezt write up coming soon]     ???, brian, daniel  ok, i\'m not sure who that girl in the photo is because i wasn\'t formally introduced to her, but that is brian, the booker from the knitting factory l.a. and daniel next to him. we were outside of the cass mccombs show at one of the many monitor records blowouts. they had amazing fajitas outside of the venue. texas is good for mexican food, which i ended up scarfing down a lot. my stomach is still recovering!     cass mccombs  cass mccombs is just one of those songwriting wonders that astounds me. i pay so much attention to lyrics when i judge a band, especially with people like cass. substance over style, any day kids. his recorded music is much more wondrous and simple than his live performance. when he\'s on a stage, cass has a full band with a full sound. when he\'s on the recording, it feels like it\'s just him and you in a room and he\'s telling you all about the sorrow in his life. it\'s pouring out of his mouth and flowing directly into your heart until your stomach wrenches with empathetic pain. ouch ouch ouch. he didn\'t play my favorite song, \"so damn pure,\" though, but i forgive him because he did an amazing \"aids in africa.\"     ira from nada surf  while we were gathering enough passes for all of our group to get into the kemado/vice party on saturday night, the doors suddenly closed. we were like, what are we gonna do now? fortunately, we had been talking to ira outside for a while and he saved our evening. after he showed off his \"national waffle association\" button, he solved our problems, and invited daniel, tom, brian and i to his friend\'s house party. fortunately, tom had a car and drove our asses over to this guy\'s house. i felt like i was in suburbia! a real neighorhood! a real backyard! no supertall buildings with glass windows in sight! it was a nice laidback change. we rolled into the house and a gaggle of comedians (i think including eugene mirman) were jamming and goofing around on instruments.      just chillin\' in texas on swings and stumps      swinging texas!   we walked to this beautiful backyard with home-made christmas lights chandeliers hanging from the tree-tops. their was a bench swing sweetly framed in lights. the setting was perfect and instantly comfortable. our host and hostess for the evening were instantly hospitable, offering us beer and friendly conversation. i LOVE texans. they are proud, but they are damn friendly. they also had a fireplace in the backyard that looked like a bong, no doubt. check it out below.     daniel, that is NOT a bong.   i had this really long conversation with ira and his maplewood bassist jude about drumming and what drummers would like to be asked in interviews, as they rarely do the interviews... then we talked about the infamous nada surf tune, \"popular.\" i don\'t care what anyone says, i love that song. and if you listen to my launchcast station, you will notice that it comes on the station every single time!! yes, it\'s very highly rated. spoken word. making fun of high school brats. the crescendo and sarcasm. genius!!! but of course i also really loved their other songs, especially off of the latest album...  the house party was pretty much the perfect, laid back way to end our sxsw extravaganza. we all gave each other a group hug and parted our separate ways, promising to party with each other again soon. hopefully i\'ll make it over to l.a. at some point to visit brian. and i\'m seeing tom again this thursday. i feel like i met some really good friends! the bonding that goes down in austin is one of a kind and i can\'t wait to go back next year.   '),(15048,1694057,'19,March,2004','2004-03-19 00:00:00','   SXSW: march 19, friday daytime    the killers @ the spin party, stubbs outside stage  i had the best time dancing with ultragrrrl to my favorite killers song, \"mr brightside.\" that song gives me such a high with its dancey beats and its lovely melodies by a mr. brandon flowers. the guy was wearing a full-on suit in sweaty, sun-searing weather. you gotta give a guy props for that alone to suffer for the sake of fashion, but the bigger props come from producing such danceable, poppy tracks that make a silly girl like me dance like a dork in front of hundreds of non-dancing hipsters, probably staring at me and thinking, \"what the hell is wrong with that girl?\" whatever, i still love that song. it makes me happy-go-lucky.     the constantines @ insound party, club deville  while you walk into club deville, the beer is just sitting in bunches at the bar for you to grab on your way to the backyard area, where the stage is set up for you. the insound.com party is awesome in that way. ok, i didn\'t know much about the constantines before seeing them at sxsw at the insound party. and that\'s what this whole experience is about, right? discovering new music? that\'s why i tried not to see too many of the bands i frequent in new york. i was in austin, with a no-holds-barred amount of new rock around me. and i was hungry to see something novel and new and stunning. vocally, the constantine\'s music hark back to the days of old school singers like dylan and neil young. the singer\'s voice is classic, but the music is forward rhythmically forward-thinking. let\'s just say there were a lot of bobbing trucker caps and rocking knees in the crowd. including mine, of course.    electrelane @ insound party, club deville  [electrelane write up coming soon]     panthers @ insound party, club deville  [panthers write up coming soon]   SXSW: march 19, friday night    the detachment kit  [the detachment kit write up coming soon]     all night radio @ blender bar  i saw all night radio for the second time. the first time i saw them was at the ascap cmj cocktail party at lit. they were just a two-piece, fairly new to the subpop scene, and i totally dug what i heard then. they are members from beachwood sparks, and their sound is pretty out of this world, including lots of points of recorded announcements: \"all night radio\" -- and in response, i kept saying, \"in case you didn\'t know, we\'re called....\" --- \"all night radio\". jimmy\'s drumming style is so loose, yet so solid, and his hair is well... like cousin it. long and flowing and rockin\'. farmer dave\'s voice is just wondrous, drawing you into the psychedelic haze.... their music is better than drugs and alcohol. it\'s the shiznit. if you don\'t own their album, buy it now!!!!!     robin phillips prays for the baptist generals      guy from the lilys dancing to all night radio!  when i saw them play the subpop showcase at sxsw, i was all goofified (meaning, all goofy and shit) and apparently so was the singer of the lilys who was dancing all dodgy (and drizzunk?) in front of the stage. they added a third member on bass guitar, this awesome asian dude with the world\'s biggest funkified sunglasses. it made me instantly adore him. can\'t wait to see them at the mercury lounge! come to the show and see the wonder for yourself.     swearing at motorists @ buffalo billiards  [swearing at motorists write up coming soon]   broken social scene @ stubbs  [broken social scene write up coming soon]   kinski @ blender bar  [write up coming soon]     the east-west coast rivalry between daniel and steve heats up with a rumble at foosball at the blender bar... it\'s a tie.  so the story goes, steve and daniel are the same people, except they dwell on different coasts. the rivalry was built by mutual friends who couldn\'t decide who was hotter. daniel or steve? what do you think? foosball says it\'s a tie.     a couple in the rain... fighting or making out?  right before i went to see the walkmen at the love collection party, it started to rain lightly. it had been doing this on and off the entire week because the air was so thick and humid. i couldn\'t believe the saturation of heat that was engulfing me, making my skin perspire and sweat. i couldn\'t believe it had been snowing in new york the day i left for the festival. and now i was encapsulated in hotness. as we were walking along this fairly empty street, i saw a couple standing in the rain at the end of the street. with the way the light was striking their silhouettes, i was instantly drawn to wonder if they were fighting or just being romantic and lovey dovey. i hoped it was the latter, and decided i wanted to capture the moment. snap snap. and here it is above. doesn\'t it just make you wanna sigh really big?     me in the rain...  me? i was enjoying the rain too. i took steve\'s awesome white sunglasses and took a photo of damp self. at this point though i was getting tired and impatient to hop into a cab and head over to the love collection party. the two girls we were hanging with ran off somewhere to do coke or something and we couldn\'t find them. finally we decided to just head to the party ourselves. it was a drive to somewhere in the dark boonies of austin, where even the cab driver was like, \"are you sure you are going to this address?\"     the mean reds @ the love collection party      penis prop guy @ the love collection party  finally, we arrive at the love collection party! there are a lot of people already there. the mean reds are on stage in short 70s style running shorts and no shirt on. they are skinny young\'uns with a huge penis prop being thrown around stage. i headed for the free beer in the back, hoping their set would be over soon. i was looking forward to seeing har mar superstar... yet scared at the same time. he has a bad reputation of getting huffy on stage...     har mar superstar @ the love collection party  [har mar superstar write up coming soon]     new kid and steve      la steve and moi      the walkmen @ the love collection party  and finally the walkies! i was pretty slick and didn\'t tell peter or any of the other   urlLink walkmen   that i was coming to sxsw. so imagine the look of pleasant surprise on peter\'s face when i was there, front and center, as they walked on stage at the love collection party? steve, who used to write for filter, daniel and i hopped in a cab and made the long, scary trek over to where this party was -- basically an empty warehouse about 5 minutes from the main drag of sxsw. even new yorkers couldn\'t walk that far. the look of surprise on peter\'s face was priceless. it said, \"what the fuck are you doing down here? that\'s awesome!\" and my look was like, \"no you\'re fucking awesome!\" oh how i love the walkies.     the wildest walkmen fans ever... yes, wilder than me.  i got a bunch of people around me to dance like crazy throughout their whole set, and i think that amped up their performance. my photos basically suck ass because the light was beaming straight at the audience and would drown out the walkmen. so this is the best i have for ya\'s. i think every time paul saw a bunch of us dancing, he couldn\'t help but flash that adorable smile of his. their performance was electrified, and i was singing along to every friggin\' song. those guys are the soundtrack to my life, no doubt! \"what in it for me?\" \"the rat\" \"old man\" \"thinking of a dream i had\" \"bows and arrows\" ALL FUCKING AMAZING SONGS. if you do not own their albums, what are you doing? you\'re a slacka, babe. i bet if you ask me and i have some extra cash, i\'ll buy it for you. no worries.  some random guy took my camera from me and took a photo of me and daniel. then he wanted to take a photo of me and myself. it was like the most bizarro thing to do, but i was so high on music that i was all giggly and into it. saw dawn barger at the show. i say it every time and i\'ll say it again, she\'s one of like three women in new york that i admire more than anyone. if i ever needed a woman to look up to, she\'s it! she manages three amazing bands (the walkmen, the national and the decemberists), and she\'s fucking awesome at it. a strong role model indeed! she says she\'s in talks in adding one more band to her management roster, but wouldn\'t tell me who... hmmmmmmm....     me and peter from the walkmen  the lovely peter came up to me after the show and thanked me for coming. uh, thank the walkies for playing such a wonderful show! they have grown so immensely and intensely over the last two years, that i felt like i was seeing brothers grow up and find their place in life... and in this case, find their place on stage. hamilton is one of the most intense lead singers around. he really gives his all to the performance... his veins pop, his voice get hoarse with emotion, and his face perspires in red huff. paul starts roaming the stage and smiling at the audience. walt was as still and calm and collected as ever. matt was bumping up and down in his throne and slamming those sticks on those drums. and peter was convulsing and sweating in his very distinctive bass waver. together they just work. the equation is a full-on bullseye! and you know what? i can\'t wait to see them again at irving plaza.   '),(15049,1694057,'19,March,2004','2004-03-19 00:00:00','   SXSW: march 18, thursday daytime  I was working the ascap booth at the trade show during the day but my lovely friends would come visit me and keep me company.... Tom! Jo! Mike and jared and byron from  winter pageant ! Thanks so much for coming by for the free beer. I also randomly talked to the guys from the national, who were supernice! I gave them all beers. I also chatted with keith girard, the eic of billboard. He asked me what the hot shows of the night were and I of course raved about  the walkmen , and I told him he missed the amazing  cass mcccombs . And I told him about a band called  I love you but I\'ve chosen darkness . Best band name ever. They are from austin and are phenomenal. See my photos and write up below...     the pleased @ the fader/levis party  Before that I was with jo jo at the fader levis party to see  the pleased , a band from san francisco. The girl in the band was absent and I forget why -- but they were still really good. I just love the way the music felt like drugs were surging thru my brain. I swear it wasn\'t the free red stripe. Their music just wooshes and unfurls like a trippy kaleidoscope.    SXSW: march 18, thursday night     robbers on high street  i also decided to see  robbers on high street  with jo jo. now i have to say they didn\'t impress me much the first time i saw them, but i decided to give them a second chance because i find that bands tend to get better over time... and robbers have definitely improved their stage charisma tenfold. i really liked the last song of their set... don\'t know what it\'s called but if any of you can enlighten me, please do!! is it on a record or what? i couldn\'t stay for the sounds because it was beginning to remind me of the nightmare experience of seeing them at luxx -- crowded, smoky and sweaty. i jetted out of that joint and wandered over to try and see stellastarr* -- tom was supposed to be over there.      i love you but i\'ve chosen darkness  on the way, i ran into the winter pageant boys who were standing in line to see  i love you but i\'ve chosen darkness . so i stood with them for a while because i heard that the line for ss* and ambulance was too insane to even attempt. so tom met up with us at red eye fly to see chosen darkness. everybody\'s been raving about this band like crazy. and i hear the bbc did a huge write-up of the band as well. they are going to be huge!! their songs were so moody in an upbeat way. it made me feel like i was soaring and diving into something quite extraordinary...  Check these dudes out. Met their manager thru daniel. She was a total sweetheart, and I could tell she felt the same way about her band as I did with  dirty on purpose . Its nice to see that. I was also at the show with tom, mike and byron. We were all floored. This band made my jaw drop. I can\'t wait for them to come to nyc. Nota bene: aurelio from  calla  was there too. along with chris newmyer from self starter and corey from absolutely kosher records.     modey lemon @ beerland  Then we couldn\'t get in to see  the walkmen  (when i ran into peter later, he said not even people from warner bros could get into the show -- uh, that\'s crazy!!) so we went to see  modey lemon , who were surprisingly fun to watch.... Actually the asian dude with the bleached hair who was rocking/making love to the guitarist was fun to watch. They thrash their hair around a lot and the music was much like that. I liked the venue b/c it was called beerland and had empty prescription pill bottles on fish nets on the wall. I drank vodka and red bull to keep me awake. I was still not anywhere near drunk so I think my tolerance was way up... Tom left and told us to call him when we left for the next show.   the stills @ emo\'s  So then we went to emos to see  the stills . Daniel had a special pass that gets u into the venue through the back way. It was funny. I felt like we were incognito celebs sneaking in the back way to avoid the papparazzi. We met up with tom there. We were so tired we sat on the bleachers, but ended up leaving halfway thru the stills set b/c of utter exhaustion.      dinner with ma\' peeps @ iron cactus  Earlier in the night I had arranged an amazing dinner at iron cactus for mexican food with jennifer from the ascap l.a. office, jojo, tom, justin and jack from  bishop allen , mike and jared from  winter pageant , and jennifer\'s friend brian, who books at the knitting factory l.a.     It was a lot of fun just hanging and getting to know each other a little better. Tom said the table was full of jin\'s friendsters! So true! It was funny and bizarre how friendster is in our normal lexicon.            We ran into  ultragrrrl  at the restaurant. She is one of the coolest chicks in the music scene (besides the obvious other queens of nyc, you know who u are) and always friendly. an entire table applauded her when she walked in and she did a little curtsy. Word!   '),(15050,1694057,'18,March,2004','2004-03-18 00:00:00','   SXSW: march 17, wednesday daytime  Don\'t have time for much of an update but here\'s kind of a general overview of what happened last night... Set up booth at trade show... Was in the elevator talking about going to see max from  natural history  play the fader/levis party... And one guy said, hey I know max! And the other guy in elevator said, I know him too! Apparently one of those guys was britt daniel of  spoon  (bidi, isn\'t that funny?) ran into isaac at the hotel.... He\'s staying with dawn.... Briefly saw  the walkmen  on the street but didn\'t have time to say hi! Had lunch with tom the booker at the hook (dirty on purpose is playing there on saturday with dennis cahlo and surefire -- go see them!) saw christi from sxsw who is apparently  rhett miller \'s sister. She was supernice.... Went to soundcheck....    SXSW: march 17, wednesday night    sea ray  Saw  sea ray  who were phenomenal! they have improved a thousand times since the last time i saw them. with visuals going on in the background and a wash of sound flooding my ears, i found the experience quite overwhelming. the songs are so strong, and the atmospheric surges of music swirled around the air like the flutter of leaves on a windy, autumn new york day. i practically ran to the venue to see the band play sxsw. i hadn\'t seen them in ages, and was dying to see them, especially in texas! the place was packed.     cass mccombs @ la zona rosa  Saw  cass mccombs  who was wonderful, but not as dreamy live as on the album. I stilll swayed a lot. and i think i missed some of his set! but i bought a tee shirt to support, and still listen to \"so damn pure\" all the time on my ipod. what a great fucking song... Daniel got stuck in dallas so we couldn\'t hook up for his showcase. He got me into cass, who I want to see and sing along to all the time!    the go @ hard rock cafe  [the go write up coming soon]     the dresden dolls @ hard rock cafe  Tom and I ended up hanging all night. He was a good partner in rock. We tried to sneak over to  the von bondies  but the line was insane. We went back to the ascap showcase and saw  the go ,  the dresden dolls  (love love! Amanda is a total sweetie), and tim\'s band,  a place to bury strangers . justin, the drummer totally pulled an arthur stellastarr* by whipping off his shirt and sweating up a storm on the beats. the funny part? arthur was there. i turned to him and told him, and he laughed. so talked with arthur and michael from  stellastarr*  who showed up for aptbs, and lio of  kanine  for a little bit. aftewards, i was sooooo tired and headed back to the hotel. i had been wearing heels all night, and just couldn\'t walk anymore. i vowed never to wear heels to sxsw again. this girl has learned her lesson, peeps, so don\'t yell at me.     a place to bury strangers @ hard rock cafe  [aptbs write up coming soon]  I feel like I\'m forgetting funny things like how tom and I ran into gaggles of birds who would swarm the tree tops and poo on the ground like some weird bird cult. Anyhoo I\'m off to work the trade show and drink more. What a night!   '),(15051,1694057,'17,March,2004','2004-03-17 00:00:00','   SXSW: march 16, tuesday night  So I\'m addicted to listening to \"all new friends\" -- the fucking coolest song that is on repeat on my ipod.... Yes it is a dirty on purpose track but who fucking cares? Its one of those tracks that make u wanna jump up and sing.... And dance especially.... Excuse me if I\'m using \"interweb\" language like \"u\" and lack of apostrophes. I\'m a bit tipsy from too much free red wine tonight...   So far I have met some really cool people at sxsw.... And I\'m totally loving it. Had a great dinner with some peeps including the cool kids from a place to bury strangers (tim, ollie, justin) and karen from violator -- a fucking depeche mode cover band! How fucking hawt is that?   Tomorrow I have a lot of shit to do for work but I plan on seeing ambulance ltd at the vice/jane party... The wrens and clearlake at the 33 degrees instore.... And ma\' boys in the natural history at the fader levis party... and maybe charlie mars at the bmg new times bbq... And cass mccombs at the inland empire showcase and of course dresden dolls and a place to bury strangers at the ascap showcase....   I just came back from some film and tv party.... And the bar was called \"maggie mae\" .... Do u know that pietasters (?) song \"maggie mae bring us another beer take away our trouble dear.... Uh neither did ANYONE in texas. What the hell didn\'t anyone listen to ska back in the day???   '),(15052,1694057,'16,March,2004','2004-03-16 00:00:00','   what goes on in austin, stays in austin  That\'s my mantra for this week. Let the debauchery and corruption begin! Trouble come find me! Labels, come sign  urlLink my band !   '),(15053,1694057,'12,March,2004','2004-03-12 00:00:00','   i feel like reposting this so...  i\'ve been reading my old blog, and it\'s been so funny. like a trip down memory lane. and i thought i\'d repost this funny article i wrote over a year ago:   How To Get Over My Band Obsession:  1) Get one of the members of the band to offend me in some unforgivable way. Like, don\'t you hate it when a guy asks you about your gorgeous best friend?  2) Every time I hear the band\'s music, I can zap myself with a taser to the beat of their drum until I have negative associations. It worked for Pavlov, it can work for me, damnit!  3) Just get over my shyness and make out with the keyboardist alreadyâ€¦  4) Go all out and make out with all of the band members.  5) Stop being such a drug addict and snap out of my star fucker phase. Besides, guys rarely look as sexy when you\'re sober. Am I right, ladies?  6) Pretend they\'re all 12 years old. Jailbait is so passÃ©. Pass it on!  7) Take some scissors to their next show and cut off all their irresistibly dirty indie rock hair. Remember, Samson and Delilah?  8) Give the rock star the worst pick up line ever and run away in horror, never to go near him again. Just like the time I ran into Steve Malkmus in a libraryâ€¦ \"Er, I, like, really like your musicâ€¦Derâ€¦\"  9) Keep telling myself they all look like Keith Richards naked (or at least, they will sooner or later).  10) Meet another band to obsess over. Ooooh, shit!   '),(15054,1694057,'11,March,2004','2004-03-11 00:00:00','   do you remember...  when this blog used to be funny and good?  urlLink check out my old school shite   also funny link about  urlLink moz getting a job . [link from anna]  tonight: head to motor city to see dear mario\'s band,  low brau  then off to see  the new humans  [ urlLink info here ] then off to lit to see the gang of  bishop allen  then party party party?   '),(15055,1694057,'11,March,2004','2004-03-11 00:00:00','   bands against bush  i went to see emma la reina and shellshocked at the bands against bush show at sin-e. it was quite fun. saw joel (sp?) from the blondes, inc. at the door, he\'s so sweet and nice and always has a smile.   met matt, the drummer from  emma la reina , a band whose former incarnation was called telenovella. they played with dirty on purpose at one of belvy\'s gotham parties at luxx last year. i thought that the band was overall good... but was unsure if it was really my kind of thing. i liked the music for the most part, but the songs weren\'t as strong vocally... but you know what? people were definitely into the music, cheering on the band, and that\'s a good sign for them... plus, matt seems like a really nice guy with a passion for drumming.  also saw anna\'s friend\'s band  urlLink  shellshocked  . they were quite good. the lead singer is reminiscent of tim harrington\'s wild style, and i instantly fell in tune with him. he jumped on the bar. he jumped on the counter. he jumped on chairs. he jumped into the crowd. i was fully enthralled and entertained. i really liked the girl singer in the band too. watch out for this band! they really know how to get the party up in arms.  then anna and i jettisoned off to north six. like old-fashioned nerds, we chose the subway over cabbin\' it. we arrived just as  cheeseburger  was going on stage. i had never seen them before, and just heard dj raving about them. i think they were good, but not exactly my steez. they were good at what they do though, rambling about the stage like drunken indie party boys who infuse a little bit of blues into their music. it was like being in detroit. said hi to christy (sp?) from the band, who was dancing like crazy during mommy and daddy\'s set. that was awesome, dude!  then of course we saw  urlLink  mommy and daddy   perform spectacularly too (listening to their fantastic demo this morning!! good way to wake up.). i love watching them rawk together on stage. the interplay and sparks between them is dreamy and fun. every girl\'s fantastic voyage! vivian is a girl who\'s not afraid to just let loose and dance and sing her heart out. edmond, from the first time i saw him, has always been crazy scary behind the mic and then the sweetest being off stage. their energies mixed together is the perfect concoction for the punk dance music they produce -- and i don\'t even usually like that genre of music. viv and ed just know how to rock it the right way.  caught a little bit of  bad wizard  afterwards, but my knee is all fucked up and i was passing out on the bleachers of north six, so i grabbed my bag and made the long trek home. it had been a long night, and i still can\'t believe i made it through in one piece. i\'ve been going out every night for the past two weeks, and now next week i\'ll be heading off to sxsw. life is so crazy, i feel like i\'m dreaming. the bliss that surrounds me is tingling my nerves. in a good way.   urlLink   '),(15056,1694057,'11,March,2004','2004-03-11 00:00:00','   peanut butter jelly time  dj sent me a couple really funny links: **  urlLink peanut butter jelly time  **  urlLink other equally as funny doodads   if you\'re having a bad day like we are, please play these on your computer and crank that shit up!   '),(15057,1694057,'10,March,2004','2004-03-10 00:00:00','   gallery opening    for all of you who love art  metalstone gallery presents....  recent works by Colin Burns opening reception Thursday, March 11thÂ  from 7 - 10 p.m.  sponsored by Red Stripe!   metalstone gallery 175 Stanton Street @ Clinton 212.253.8308Â Â Â Â Â Â Â Â   hope to see you there! Wendy & Nina for more information on the artist visit  urlLink www.vivagritona.com  flip through the \"dirty books\" section of colin\'s website -- genius!!  afterwards i\'m gonna head over to sin-e for the bands against bush show. then jet on over to north six to  mommy and daddy . come along for the art-slash-rock adventure!   '),(15058,1694057,'10,March,2004','2004-03-10 00:00:00','   superstar!  this  urlLink mp3  of me singing the carpenters\' \"superstar\" is so friggin\' hilarious!!!!! i sound so terrible, it\'s not even funny!   '),(15059,1694057,'10,March,2004','2004-03-10 00:00:00','   wine is silly  yep, folks, wine is silly. but i can drink the stuff like nobody\'s bidniz! and what a wonderful haze it is. jackey and i went to her roommate\'s gallery to preview works they are premiering tonight. i\'ll post details on the opening here a little later, so that you can show up for the free wine and the wonderful pencil drawings that are worthy of hip screened tees or framed on your purdy lil\' wall. i want to buy one or two or three of them, but can\'t afford to shell out the $75 at this point in time... otherwise, i\'d be all up in there because they are  that  good! fans of  urlLink yoshitomo nara  will love colin burns\' simple, quirky pencil drawings because they are infused with the same wit, sarcasm, cynicism and humor... but he also has these amazing paintings too... well, you\'ll see.   '),(15060,1694057,'10,March,2004','2004-03-10 00:00:00','   the crotches    link from  urlLink the modern age   i couldn\'t resist re-posting this. so funny. read the interview  urlLink here .   '),(15061,1694057,'09,March,2004','2004-03-09 00:00:00','   morning playlist  don\'t mean to sound cocky, but i made the most amazing ipod playlist this morning, and i just wanted to share. make it yours:   calla  - \"televise\"  dirty on purpose  - \"mind blindness\"  au revoir simone  - \"through the backyards of our neighbors\"  ben kweller  - \"make it up\"  damien rice  - \"delicate\"  dennis cahlo  - \"crushed\"  the cooper temple clause  - \"blind pilots\"  the decemberists  - \"here i dreamt i was an architect\"  sonic youth  - \"superstar\"  the stone roses  - \"i wanna be adored\"  the walkmen  - \"french vacation\"  doug and i were talking about how we\'re both people who can zone out of the world with their mp3 player... and how we also sometimes just enjoy the sound of the city. sometimes (most times) i just don\'t want to hear a crying baby on the subway on my morning commute. word? besides, music is cheaper than therapy.  doug also told me that he recorded me singing, \"superstar\" at winnie\'s. omigod. i might feel compelled to post it online... but i started singing the words that were the on the screen and realizing they were wrong, so i closed my eyes and just sang my heart out, remembering the words suddenly like jagged edges of my nervous finger nails. sigh.   ');
INSERT INTO `msgs` VALUES (15062,1694057,'09,March,2004','2004-03-09 00:00:00','   the revolution will be televised   just when i was beginning to think that aurelio\'s hair and eyes were prettier than calla\'s music, something completely changed my mind. wow, my mouth dropped wide open when   urlLink calla   played the last song of their set last night, \"televised\"... then i couldn\'t stop talking about how amazing that song was to doug. how amazing it was when they softly played the song like it was going to end and then how they crescendo\'d right back into the song. i was completely blown away, and that one moment made up for all of the duller ones. it put me into an instant daze. i loved the way aurelio moved on stage. it looked like he was walking through the crowded streets of tokyo, manuevering his guitar around people -- like his guitar was swimming around the stage. and i loved how he was making out with the microphone. what a fucking sensual being.       urlLink the cooper temple clause  ... someone yelled out, \"the cooper temple clause, the cooper temple pilots: same thing!\" and it was kinda true, but not really. i love this band\'s ability to delve into so many different elements and mix mish mashes of things together to create a wall of sound that could be taken as a lotta rock, a little bit of hip hop and a little bit of jazz, they were great. i love tom bellamy, the guy with the stache who reminded me and doug of a shorter angus andrew from liars. i love guys like that. he was wild on stage, just letting it all loose and i enjoyed watching him. you could probably say i have a new crush, folks! turns out he\'s the one i interviewed last week, and he was indeed nice, but definitely has a girlfriend. my favorite song of the night was of course \"blind pilots\" -- the song i have playing on my ipod endlessly in all my playlists. fucking love that song. you only need one really good one to win me over, at least for now, in this temporary moment of blogginess.   earlier in the night, i stopped by vicious to see   urlLink the gostation   again. i had seen them play before with aerial lovefeed at don hill\'s and was curious as to how they would sound at sin-e. i had recently befriended doug the lead singer of the band, and found he has a great presence on the stage. i like watching people who look confident when they are behind the mic. it might be sort of an obvious point, but rock stars have to look like they want to own the stage, and doug definitely has that going on, without being annoying like some people can be... anyways, they covered new order, and that\'s always fucking cool in my book. it was fun, and i was definitely glad doug went with me. incidentally, we ran into the other doug (the gostation) again at the cooper temple clause, who loves tctc. rock on! audrey and dennis were dj\'ing, and she informed me that i missed \"the rat.\" rats!!!!!! next time i will have to arrive earlier so i can dance my ass off to that, eh? i was in good spirits, dancing all around with jo jo and jocelyn and whoever. it was fun being a lil\' tipsy. so we added a fourth member to  \"excuse, moi\"  -- miss jojo dancer on drums! how hawt is that? ok, you think i\'m silly. but really, i\'m just happy.   even earlier in the night, i was having after-work drinks with ma\' girls at pianos while  surefire  was soundchecking. we ran into justin from   urlLink bishop allen   there, who was cracking us up with his chipped tooth. we nicknamed his adorableness \"snaggletooth\" and drank and talked about boston. (everyone had gone to school in boston \'cept for moi.) we were all giggling because surefire were so cute, and they kinda sounded like the monkeys or something -- i love the monkeys! it was cool... now i want to see surefire again. fo\' sho! wouldn\'t it be funny if everyone started callin\' them sho\'fire! that would be awesome and completely snoopified. unfortunately i will miss the band when they play with dirty on purpose at the hook, but if you\'re all around, you should go support!! march 20 @ the hook. mark your calendars, babe. also, dennis cahlo is performing, and we all know how good that boy is. and he gives the best hugs of any singer \'round town! so ask him for one.  i told justin i would come see bishop allen at lit on friday. you should come too. it will be loads of fun! and probably the last chance you\'ll get to see them for a month because of their crazy tour schedule. justin told me the band would be at sxsw the whole week next week, so i\'ll definitely run into him again down there. every day i meet someone who\'s going to the festival. it\'s going to be fucking insane!!!!!!!!! insane!!!!!!!!   '),(15063,1694057,'09,March,2004','2004-03-09 00:00:00','   trashy or trashed  anna and i met up with vivian from mommy and daddy today for lunch at whole foods. viv gave us loads of make-up, and now i feel like i\'m high from my new pink lipgloss. thanks viv!!! this lipgloss is called \"juliet\" and it makes me want to wear trashy lingerie. this is what tyra and gisele must feel like every day.   '),(15064,1694057,'08,March,2004','2004-03-08 00:00:00','   questionaires  decided to go ahead and fill out jocelyn\'s questionaire...  Favorite band in New York City:  the walkmen, dirty on purpose  Top bands I\'m excited about right now (in no particular order):  dirty on purpose, au revoir simone, knife skills, cass mccombs, damien rice, pas/cal, pela, dresden dolls, bishop allen, qatsi, the winter pageant     Rockstar crush:  matt from french kicks -- the cutest, nicest boy ever to grace this planet. always with a smile and a hug! also, slight crush on sweeney from trouble with sweeney... talked to him for a little bit at southpaw and was very taken with him onstage. he\'s like what jack white would be if jack white were a sexy, geeky indie rock boy   Rockstar whose head I would like to bash in with a brick:  i can get along with almost anyone. and i\'m not that violent. but if you give me \'tude, i\'ll give it right back to ya.    Rockstar I wish I could be best friends with:  ed harcourt. i was once alone in a room with ed harcourt for an hour. we got along so wonderfully, just chatting away like best friends. when i went to see him later that night at fez, he was the sweetest to me... and then i never saw him again. so sad. i think we\'d get along splendidly!!  Rockstar I am best friends with:  anna barie from  urlLink knife skills   Soon to be rockstar:  jinners and audrey and dennis in \"excuse, moi!\"  Band that all the hipsters love to love, but I think is actually total crap:  ha, wouldn\'t you like to know?! i don\'t like to diss bands because everyone has a right to their rockstar dreams in my book.  Band that I love to love, but all the hipsters think is actually total crap:  hmm tough one... most hipsters like what i listen to.   Band that I wish I could be in:  blonde redhead or enon  Band that I used to wish I could be in:  belle and sebastian -- don\'t you love \"take me away from here i\'m dying\"?  Best band to photograph:  the walkmen -- hamilton is intense. INTENSE  Band I wish I could have seen back in the day:  the recoys  Band that I will be SAYING I saw back in the day:  dirty on purpose    Rockstar I want to be right now:  sweeney from trouble with sweeney   '),(15065,1694057,'08,March,2004','2004-03-08 00:00:00','   come on let\'s go  so i get home from rod\'s place last night and get ready for bed when... i hear sweet caz\'s voice on the fucking telly. i was like, what the fuck!?! and today\'s email from caz confirms my suspicions:  Jin, Here is some oddball news... urlLink  PAS/CAL   have followed the path of the likes of Cameo, Brendan Benson, and the Walkmen and are now featured in the latest Saturn ad campaign. The nice thing about the dealio is that it will enable us to record our debut full length instead of doing a 3rd EP. Braggadocio notwithstanding, Caz  congratulations, caz! i still love you.   '),(15066,1694057,'08,March,2004','2004-03-08 00:00:00','   oooh wah ooooh wah oooh wah ooooh wah  new  urlLink  au revoir simone   mp3 called \"through the backyards of our neighbors. like joe said, so good. they play pianos april 10 with somnambulants and boyskout. check it out yo! i\'m so proud of those gals.  also, i have a wonderful video of goldenrod singing \"moon river\" at winnie\'s.  urlLink check it here.  goldenrod produced both the au revoir simone track and the new dop stuff. as well as the new pela and the new jen chapin. so if you need a fucking awesome producer, you can  urlLink email him here .  on the table tonight:  the gostation  @  urlLink vicious   the cooper temple clause ,  calla  @ bowery ballroom party party party???  maybe another run to 31 flavors. rock and pralines. so good!   '),(15067,1694057,'08,March,2004','2004-03-08 00:00:00','   head gear studios    if you\'re a band looking for a great place to record your songs... well look no further than  urlLink head gear . they were awesome for  urlLink dirty on purpose ! they\'ve also recorded the yeah yeah yeahs, tv on the radio, liars and the boggs... DoP\'s new recording will be done hopefully this month. i\'ll have copies of rough mixes for sxsw next week! word! you have no idea how excited i am!   '),(15068,1694057,'04,March,2004','2004-03-04 00:00:00','   moby  apparently,   urlLink moby   was at the sin-e show on saturday night to see   urlLink dirty on purpose  . he bought a cd, watched their set, and then left!! saryn will appreciate that.   a big thanks to all who showed up for the band. i love you all. shout out to all the amazing bands that dirty played with this weekend:  urlLink the eaves ,  urlLink the caufield sisters ,  urlLink trouble with sweeney ,  urlLink dennis cahlo ,  urlLink asobi seksu ,  urlLink the hong kong . what a fucking awesome weekend of music! we were grateful to be playing with such great bands. thanks also to doug at southpaw, jasper and audrey and asobi seksu for asking dirty to play. also thanks to  urlLink aeri  and her friends for filming both dirty shows. you are amazing, and i\'m so happy we finally met in person! mwah! also thanks to the  urlLink misshapes  peeps for putting the afterparty together.  i ended the big weekend with a karaoke party at winnie\'s for goldenrod\'s birthday. i was so tired, but made my way over to chinatown... after a couple drinks, i was ready to sing my standards: \"killing me softly\" and \"superstar.\" doug pointed out, \"both those songs are depressing and both talk about crushing on rock stars.\" well, doesn\'t that perfectly describe jinners? haha. ;-)   '),(15069,1694057,'04,March,2004','2004-03-04 00:00:00','   NOTA BENE: DIRTY DIRTY DIRTY!!  hey peeps,   despite whatever i said before, go with the listings.  urlLink  DIRTY ON PURPOSE   goes on second at  urlLink southpaw  (friday, march 5, 10 p.m.) and  urlLink sin-e  (saturday, march 6, 8:30 p.m.). i\'m sorry for any confusion, but we want to make sure everyone gets to the show at the right times -- so we\'re going to go with the advertised times that have already been released. the band\'s going to debut a couple, hot new songs. i\'m still dizzy from their spell.  also, Special After Party @ Misshapes! GO TO:  urlLink www.misshapes.com  for more info.  xoxo, jinners and dirty   midtown is the new downtown ... today has been the day of running into people in bands. it\'s all up in the air. i ran into jason from  urlLink  the boggs   this morning on the subway -- he was on his way to band practice... and now that i think about it, that was really early for band practice, eh? he\'s the sweetest. and then anna and i ran into vivian from  urlLink  mommy and daddy   at whole foods. we ended up lunching with her and jason from  urlLink product shop nyc  and their friend. who will i see next?   '),(15070,1694057,'03,March,2004','2004-03-03 00:00:00','   white lace and promises...    saw  urlLink miker  and john from  urlLink  team usa   yesterday. they came into the offices, and brought me a copy of their new cd. let me just say, i simply adore the cover -- don\'t you? and the music is great too!   '),(15071,1694057,'03,March,2004','2004-03-03 00:00:00','   HAPPY BIRTHDAY,  urlLink ANNABIOTICA !  as part of her birthday present, i took anna out on the town to see dresden dolls. this band  sleepytime gorilla museum  opened for dresden dolls -- we had no idea what we were in for! they were kind of cliche experimental pagan rock of sorts, interesting at first, but in the end, they were simply tiring and entirely predictable. snore! yawn! headache! i did like the girl in the band though. she could go solo and be quite successful. creepy, but in a good way.     dresden dolls  were great, but not the best i\'ve seen them. maybe amanda was tired from playing the night before. maybe she was bummed half the place emptied out after sleepytime (bizarro). but me, i just can\'t get over \"half jack.\" what an amazing song. some goth girl yelled at me for taking photos. damn, that was the first time someone ever yelled at me for snapping a few photos. what the hell?  currently spinning: cooper temple clause seafood damien rice mono folksongs for the afterlife french kicks brendan benson dresden dolls dirty on purpose  finally decided i hate the  franz ferdinand  ep (enough to want my money back!!), although \"darts of pleasure\" is completely danceable. don\'t hate me, ultragrrrl. on the other hand, i can\'t get enough of  the cooper temple clause . \"blind pilots\" is amazing. i\'ve also been consuming  damien rice  in large quantities. \"delicate\" is the bestest.   '),(15072,1694057,'02,March,2004','2004-03-02 00:00:00','   blog new world   urlLink village voice article on bloggers   some of that is really true!   '),(15073,1694057,'01,March,2004','2004-03-01 00:00:00','   cooper temple clause  in researching for my interview today with  urlLink the cooper temple clause , i am caught up with their obsession with poet philip larkin:   It must be admitted that there is a strong current of skepticism running through Larkin\'s poetry. In \"Kick up the fire, and let the flames break loose\" from The North Ship Larkin writes about the human need for companionship; he suggests that when we are with a friend, we \"prolong the talk on this or that excuse\" in an effort to avoid being alone. Unfortunately, \"when the guest / Has stepped into the windy street, and gone, / Who can confront / The instantaneous grief of being alone?\" Although we find some comfort and meaning in the company of a friend, as soon as we are left alone, the reality of our personal isolation crashes down upon us with frightening implications. [ urlLink source ]   '),(15074,1694057,'01,March,2004','2004-03-01 00:00:00','   black sabbath = art  dj from dirty on purpose took part in this art project... and it is screening tomorrow at the passerby art gallery. come with friends, it\'s going to be a treat: DON\'T MISS THIS-IT\'S A REAL GEM!!!!!! Tues March 2nd @ Passerby... If you were at our Halloween party 2 years ago you may remember this. If not, you may have heard about it. And many who were there never saw it due to the viewing room being too crowded! We took 40+ artists and divided the first  Black Sabbath  album into 40+ 1 minute clips. Each artist was assigned a clip and did a film, video or animation based on their clip. We then reconnected the clips to create one continuous soundtrack with a whole bunch of shorts over it. Some of the artists include..... Tunde Adebimpe, William Henry Beck IV, Warren Bennett, Joe Bradley, Crystal Bahmaie, Richard \"DJ\" Boudreau, Ben Coccio, Chris Coccio, Bjorn Copeland, Benji Cossa, Luke Crotty, Jim Drain, Madaline Davy, Caleb Emerson, Andy Estep, Dan Gorman, Andrew Freiband, Kaori Hamura, Sharon Hardy, Christy Karacas, Larisa Laivins, Richard \'Dicko\' Mather, Ben Murray, Keith McCullock, Amanda McCorkle, Matt Keegan Rapheal Lyons, Takeshi Murata, Rich Porter, Jason Rothenberg, Ryan Sovereign, Francine Speigel, Kevin Umbrict Stephen Warbrick, Aaron Warren, Sara Welsh, + a couple more I\'m sure I left out!!! (sorry if I left anyone out or misspelled name!!!!)  This video is really awesome!!! hope to see you there!!! screening @ 9PM...followed by heavy partying and rock videos...  Passerby  is located at 436 W15th street (10th ave) 212.206.7321   '),(15075,1694057,'01,March,2004','2004-03-01 00:00:00','   word association with the walkmen  from the  urlLink prefix mag interview  with  the walkmen  lead singer hamilton leithauser.   prefix mag:  friendster  hamilton:  what?  classic.   '),(15076,1694057,'01,March,2004','2004-03-01 00:00:00','   all in time  you already know this, but i really dig that new  urlLink  pela   song, \"all in time.\" so damn catchy! caught the band again at sin-e this weekend and had such a blast. billy was extra active on stage, bashing and crashing around. and nate, eric and josh were so awwwn it as well. i especially love nate\'s catchy-as-hell guitar parts. their new drummer was also rockin\' out and is quite fun to watch... but you\'ve heard me spill time and time again for pela.  after their set, i SPRINTED to the living room to see  urlLink dennis cahlo  from  the realistics  do his solo thang. let me tell you how sweet his voice is... it\'s so soft, and his falsetto just melts into the air, swirling around and unraveling like an intimate secret whispered into your ear. it\'s a dizzying spell, like almost kisses and hands accidently brushing up against your arms. his acoustic songs made me giddy. and his tender cover of my favorite realistics song, \"why didn\'t you stay?\", just blew me away -- to the point i was sitting in my chair, singing along, my heart tingling with each wonderful lyric. the girls are gonna swoon big time when his solo album comes out. and i thought it was totally sweet how he dedicated pretty much every song to  urlLink my favorite frenchie !   '),(15077,1694057,'29,April,2004','2004-04-29 00:00:00','   girl, you know it\'s true. oooh ooooh ooh, i love you  yes, it\'s true. i will be deejaying for the first time ever, and boy, is it gonna be a doozy of a show. hot bands and the other deejays are hawt too! let\'s turn it into a party. if you cab it over to the venue, you can turn in your receipt at the bar for a free drinky drink. yummmy. mike (greensleeves) will be my deejay partner-in-rock -- we will be reactionary deejays -- wheeeeee!    urlLink DJ JINNERS     May 7  @  urlLink The Hook  (Land\'s End: A Showcase of local bands and DJs) w/ DJ Sam (Interpol), Greensleeves (The Winter Pageant), Audrey (MelodyNelson.com)  with HOT bands: Feast  urlLink Other Passengers   urlLink Qatsi   urlLink The Big Sleep   '),(15078,1694057,'28,April,2004','2004-04-28 00:00:00','   read yellow = good spastic rawk   urlLink read my review of read yellow on popmatters.com   '),(15079,1694057,'28,April,2004','2004-04-28 00:00:00','   two reasons to like the french   urlLink audrey   and   urlLink concert and co   '),(15080,1694057,'28,April,2004','2004-04-28 00:00:00','   my alphabetical playlist  while on the subway i got up to regina spektor, will add more later  not over you yet - kevin devine make it up - ben kweller quarter to three - bishop allen pink love - blonde redhead pleasure seeker - brendan benson almost crimes - broken social scene step on me - the cardigans not the way - cass mccombs delicate - damian rice crushed - dennis cahlo twice a day - dirty on purpose a mistake - fiona apple the falls - french kicks i\'m on standby - grandaddy ballgame - kevin devine can\'t take my eyes off of you - lauryn hill down toward the healing - lovedrug inside of love - nada surf david - nellie mckay bem, please come home - pas/cal latitudes - pela all in time - pela nothing better - postal service worn me down - rachael yamagata samson - regina spektor   '),(15081,1694057,'27,April,2004','2004-04-27 00:00:00','   au revoir simone + thurston moore = awesome   urlLink au revoir simone  will be playing with thurston moore and jim o\'rourke and friends at bard college on saturday. whoa! congratulations, girlies!   '),(15082,1694057,'27,April,2004','2004-04-27 00:00:00','   i\'m the world\'s saddest dj ever   my ipod playlist today  i\'m on standby - grandaddy backyards of our neighbors - au revoir simone make it up - ben kweller quarter to three - bishop allen like someone in love - bjork equally damaged - blonde redhead in particular - blonde redhead pleasure seeker - brendan benson almost crimes - broken social scene so damn pure - cass mccombs   '),(15083,1694057,'26,April,2004','2004-04-26 00:00:00','   homeless at nyu   urlLink Homeless at NYU    Dear Steve: jinners - email - url  i just want to say that i went to nyu too, and i feel so much sympathy for you. being poor and not being able to sometimes afford a hot dog or to be able to make the suggested donation at the Met during class trips can make a student feel really bad about themselves. i\'m so glad that you were able to get free housing! Tue 27.Apr.2004 @ 04:19:29   urlLink what is up with the big Stern ad in this ny times article??   '),(15084,1694057,'26,April,2004','2004-04-26 00:00:00','   i heart kevin devine    i can\'t stop listening to \"not over you yet.\" listen to it on  urlLink kevin-devine.com  over and over and over!!!   '),(15085,1694057,'23,April,2004','2004-04-23 00:00:00','   siren festival -- finally a hot lineup   Death Cab For Cutie Blonde Redhead Har Mar Superstar TV On The Radio Constantines The Fiery Furnaces The Fever The Ponys Your Enemies Friends (more to be announced)   urlLink http://www.villagevoice.com/siren   thanks,  urlLink drewbacca ! i\'m gonna try and make june 29  urlLink spoiler film fest !!   '),(15086,1694057,'23,April,2004','2004-04-23 00:00:00','   playing catchup through photos    i saw  urlLink the boggs  last show last night at bowery. it was great to see them on a big stage, and they really shined like the blues punk rockers they are... so sad that jason is moving to germany!     as you may or may not know, brad (drummer) has joined the  urlLink on!air!library!  lineup, and will rock on with those!ethereal!kids!     hopefully zeke will keep on with his slide guitar, because he\'s got mindblowing skillz. i hear his solo stuff is amazing -- hello, record labels!     urlLink dennis cahlo  has the voice made especially for soaring hearts that swoon for the hope of love gained and love lost. here he\'s playing at his residency at pianos. soft lighting, soft heart.    urlLink LUKE\'S NEW EVENT POSTING SITE! POST YOUR SHOWS AND STUFF HERE!!      according to one boy,  urlLink mary timony  seems to be going through a cute phase.     according to several other boys,  urlLink mary timony  is  definitely  going through a cute \"i love you, will you marry me?\" phase.      urlLink hightower  plays the dennis cahlo residency at pianos, and looks good while singing his lovely songs...     urlLink the sleepy jackson  are not as sleepy as you\'d think: \"i love to watch good dancers talk. my heart is stronger than you all!\" one guy kept making out with his gf on the roof of bar 13 in the freeeeezing cold. that\'s love, people.     urlLink q and not u  know how to be political and dancey at the same time. i liked it, and even found my feet (one of which i pulled the achilles tendon) dancing.   more to come:  the winter pageant. circle and square. au revoir simone. qatsi. autodrone. a place to bury strangers. hotbird 7. etc etc etc.   '),(15087,1694057,'23,April,2004','2004-04-23 00:00:00','   GBV to Break Up at end of the Year  \"We are the kings of indie rock. When we quit, indie rock will die.\"   Beloved indie rock act Guided By Voices will split at the end of the year, frontman Robert Pollard announced on stage last night (April 24) at New York\'s Bowery Ballroom. The group\'s final album, \"Half Smiles of the Decomposed,\" will be released Aug. 24 via Matador. A full announcement is expected from the label on Monday.  urlLink Billboard.com   '),(15088,1694057,'23,April,2004','2004-04-23 00:00:00','   oxford collapse cd release party    the boys in  urlLink oxford collapse  are on the rise with their debut long play cd, some wilderness. they party tonight at the knitting factory tap bar. the show starts at 9 p.m. and OxC goes on at 11 p.m. come support the boys in their release!     urlLink >> listen to \"1991 Kids\"   '),(15089,1694057,'23,April,2004','2004-04-23 00:00:00','   thank you, everyone!!! xoxo, DoP and Jinners    wowee wow wow. i was overwhelmed by everyone\'s support last night. what a great night of rock, and a very happy  urlLink Dirty on Purpose  CD release party! thanks to all the bands for playing, and thanks to all the fans for paying!  a special thanks to greg and jonathan and jonathan for picking up the CDs and bringing them to the venue. this is why you are so friggin\' awesome!          what\'s doing tonight?  first off, happy birthday,  urlLink jay !  what are you all doing tonight? i\'m going to see   urlLink the winter pageant   at mercury lounge -- 11:30. then come to the after party at black and white (10th St, between 3rd and 4th Aves) -- it\'s also the after party for metropolitan and the gostation, who are playing at plaid tonight. fucking rawk! and hopefully swing by jay\'s extravaganza.   '),(15090,1694057,'21,April,2004','2004-04-21 00:00:00','   Dirty on Purpose on Riot Radio!    The upcoming  urlLink  Dirty on Purpose   CD,  Sleep Late for a Better Tomorrow , will be streamed live on  urlLink Riot Radio  on April 22 at 10 a.m. EST until 12 noon.  There will be another broadcast of the show for those who missed it. It will be from 4 to 6 p.m.  Check it out, kids, and then come to  urlLink the big party at Pianos !!   Hear the Buzz:   From  Time Out New York :  Local five-piece Dirty on Purpose\'s stunning EP  Sleep Late for a Better Tomorrow  is full of strange, sweet melodies that had us wishing it was more than 20 minutes long.   From  urlLink Sugar Zine :  This Brooklyn five-piece has been making some serious waves since they began playing out in late 2002. One of the more interesting aspects of this group is their trading vocalization. Although their drummer Doug Marvin is considered the lead singer, three other members share time on the microphone while all five contribute to the song catalog. DoPâ€™s sound can change from delicate to dynamic in a matter of seconds with soaring instrumentation and intense riff interplay. With a self released Ep due out in April (with some help from On the Moon Music), DoP is ready to win you over.    From  urlLink www.rheingoldbeer.com/backyard   You\'ll be lulled into those dreamiest of states like a baby rocked in the cradle when you catch this Brooklyn based quintet. Their soothing, glistening sound creates climaxes for the smart ones who really listen with raspy guitars and unique plays on volume. Less is more in their game...DOP\'s winning but understated complexities will make your cup runneth over...lets just hope its full of Rheingold!    From  urlLink jenyk/dot/com   With the able assistance of producer extraordinaire Rod (of Pela fame) - fused and synched and flowed their 5 distinct personalities into a sound that after having sat down with the songs on eternal repeat for a good solid week I realize aren\'t dark like I had thought they were when I first heard them live, but have taken on a youthful so-not-typical optimistic childlike wonder way of seeing the world - a feeling that if you knew Doug, Erika, George, Joe, and DJ you\'d understand just how and why they\'d feel that way.   From  urlLink Flavorpill.net :  Brooklyn\'s Dirty on Purpose open with sprawling soundscapes that induce dreamy reflection. The pensive lyric-writing duties are split between all five members, while only four of them share time behind the microphone, anchored by the strong vocals of drummer Doug Marvin.    From PlayLouder.com: urlLink    Kicking off the night were New York rock scene newcomers Dirty On Purpose, a five-some causing a stir with their swirls of atmospheric pop, combining soft, smooth melodies that seem to caress like warm summer waves, and introspective lyrics delivered in a childlike croon by none other than the bandâ€™s drummer.   From  urlLink Escaping Words :  I heard about dirty on purpose last summer and remember checking out the tracks on their first ep. I remember them being cool. They vaguely reminded me of Explosions in the Sky...The previews got me re-excited though. They sound check-them-out-worthy if you like layers and don\'t mind lots of repeated notes.    From  urlLink Mo\'Time :  Dirty On Purpose meanwhile are well matched with Say Hi To Your Mom; they too have a mellow and fluid sound. However DOP rely more on rock-based structures, heavy on the echoing electric guitars and feedback. But they can also tone it down to an acoustic hush with such songs as \"Go Back To Sleep\", an imperative song whose command is easy to follow.   Do you have a  Dirty on Purpose  blurb?  urlLink Email me  and I\'ll post it on here!!   Recent Radio Playlists:   urlLink Mount Holyoke   urlLink FBi   '),(15091,1694057,'20,April,2004','2004-04-20 00:00:00','   those kids  you know, those kids you always run into at shows, on sidewalks on sunny days in new york, in dingy, smoky bars in the wee hours of the morning? yeah, those kids. love them.  urlLink you know who you are.   in other news, i finally saw noah from  urlLink sam champion  do his solo thang at the mercury lounge this past weekend. sure it was a stripped down version of the cd i\'ve been listening to... but i really liked what i heard -- the kind of shy guy behind the mic who sings quirky lyrics with humor, wit and a bit of charm. stephie and i liked him so much we gave him our drinks. i especially love \"TV Fever\" and that song that Jay calls his \"cuddle song\"... the band\'s next show is: may 20 @ pete\'s candy store  709 lorimer st, williamsburg, brooklyn, ny 10:30 pm // 21+ // free.   we also went and supported bsiii in the off-broadway play he\'s in,  urlLink the moonlight room . the play is set in a hospital waiting room -- it\'s the kind of play that has a lot of witty repartee, along with moments of silence that almost act as other characters in the story. as the play unfolds, and you begin to figure out the complexity of the characters, you see that there\'s always more going on with people than it originally seems. this is a lot like life, we all hide behind painted exteriors that can slowly peel and chip away in floods of hardship and reality checks. and as the play progresses, you see the characters aren\'t always what they seem. if you get a chance, you should check out the play. call 212-239-6200 or 800-432-7250 to reserve tickets. use discount code (MRGR645) for $10 off ticket price!   '),(15092,1694057,'15,April,2004','2004-04-15 00:00:00','   angel i can see myself in your eyes    angel won\'t you feel for me from your heart... blonde redhead = intense, phenomenal   '),(15093,1694057,'13,April,2004','2004-04-13 00:00:00','   preview of the dirty on purpose EP   urlLink      Dirty on Purpose CD Release Party!  sponsored by UP,  urlLink jenyk/dot/com ,  urlLink melodynelson.com ,  urlLink On the Moon Music :  April 22, 2004 @  urlLink Pianos  (158 Ludlow Street) w/  urlLink Qatsi ,  urlLink Autodrone ,  urlLink A Place To Bury Strangers , Hotbird 7  Special Guest DJs:  urlLink Miss Modernage ,  urlLink JoJo Dancer ,  urlLink Phil (Blue Sparks)   8 p.m. // $8   '),(15094,1694057,'10,April,2004','2004-04-10 00:00:00','   march 25: dirty on purpose -- up party @ pianos  I think I already told you about this show... but it was loads of fun, and the bands on the bill were amazing... kari and belvy\'s party on Thursday nights at pianos are always loads of fun with all of the amazing people that show up from week to week. They really have their shit together, and promote some of the coolest bands in the city. We had some problems with the sound at the venue, but overall it was a fun night of hanging with peeps in tow.  urlLink Dirty on purpose  are happy to have them host the cd release party on april 22 at pianos with  urlLink jasper  and  urlLink audrey  sponsoring.    march 26: a place to bury strangers @ pianos  it was tim\'s birthday, and we celebrated with cake and congratulations. I first saw tim\'s band,  urlLink a place to bury strangers , when I was in Austin for sxsw this year, and was impressed by their jesus and mary chain roots. Their music often swerves from subtle romanticism, muscled with a tough masculinity that just pulsates with sex. The droning guitars and bass are a total turn-on, the shirtless drummer is magnificent and sweaty. Can\'t wait to see them again when they play at the dirty on purpose cd release party!   march 29: paper mag party @ blvd & all night radio @ mercury lounge  I was going to spent the night working on my writing, but carol called me up and invited me to go to paper magazine\'s party to celebrate the beautiful people issue. So I put on my dirtiest jeans and grabbed my g-fab gold lamee purse and headed over to some club called blvd. On the way over, I walked past mercury lounge and saw the chalkboard say, \"all night radio.\"  Wha??   urlLink All night radio  is playing? Why didn\'t anyone remind me? You are all officially dorks for not doing this.  I continued on to the club, thinking maybe I could convince carol to come with me after the paper mag party. I got to blvd and carol and I strolled in. That place was  massive . We walk in and there are people everywhere drinking skyy vodka and cranberries. Talk about advertising! But the drinks were free, so who were we to complain?  We partied with carol\'s vidal sassoon peeps and one guy named clint seemed to take to me instantly. He was a white, super fashionable guy who was adorable and bubbly. It made me miss mikey and scotty. Clint works as an asst art director here in new york now, but apparently in Toronto he used to work with a lot of Korean hairstylists who taught him some of the language. It was amazing. He sang me Korean lullabies, songs I haven\'t thought about since I was like 10 or something. It was wonderful. Then he sorta flipped my hair with his hand and declared that he wanted to cut my hair. I felt like his diva. Mold me! Mold me! My hair has been looking pretty rough!! He gave me his business card, and I was giddy.  Carol agreed to check out  all night radio  with me. What I love about her is that she\'s always game for anything. She just moved here from boston, and is one of those smart women in this world I so adore and admire. We zipped on over to mercury lounge and I dragged her to the front of the stage because all night radio is a band that is both a visual and aural spectacle for me. My voice cracks with joy when I\'m watching jimmy drum all crazy and sweep the wind chimes connected to his drumset. He\'s become one of my favorite drummers, aside from Jerry from Knife Skills and Turing Machine. Farmer Dave and Jimmy are from beachwood sparks and they have a new bass player playing with them, this skinny asian dude with googly sunglasses. He is very rad.   As a band they create these fantastical soundscapes that are trippy and weird and wondrous. I adore their song, \"oh when,\" and I have it playing on my ipod all the time. it\'s a total pick-me-up-from-my-rotten-mood song. The band also has this electronic voiceover that says \"all night radio.\" It\'s totally tongue in cheek, which is another aspect I adore. They don\'t take themselves too seriously. That\'s the trouble with a lot of bands these days. They are waaay too serious... get over it, buddy.  Ran into noah from  urlLink sam champion  outside of the venue and tim from a place to bury strangers. Sam champion is a band I\'ve recently listened to a lot because I love pavement. They are pavement reconstructed or revisited. Whichever is the case. I love the twang and the steve malkmus vocals. I haven\'t seen them live yet, but definitely want to check them out soon!!   march 30: delightful dinner and mixing at rod\'s  I had a delicious dinner with deron and anna at angelica kitchen where I scarfed down a marinated tofu sandwich. Yummmmmmmy. And drank apple cider. Yummmmmmy!  Then I ran over to Brooklyn and hung out at rod\'s while we were doing some more mixes on the upcoming dirty on purpose CD. Watching rod work is like watching Einstein solve complex equations. And then all of a sudden it\'s like \"E=MC2\" and you\'re like, what a total fucking genius! He\'s the best producer, and just has a natural talent that can\'t be explained.  urlLink Email him  if you want him to work on your CD, and y\'all can discuss rates. This is what he does for a living, folks!   march 31: robbers on high street @ mercury lounge  already I\'m beginning to see that I\'m hanging out at mercury lounge soooo much. I still think it\'s one of my favorite venues in the city besides Bowery. The sound is always amazing there, and the people who work there are delightful. In fact one of my fondest show memories is when I saw natural history, panthers and the walkmen play there... And I was a dumbass and didn\'t stick around for liars. Boy did I regret that.  I remember being really impressed with how much  urlLink robbers on high street  have improved when I saw them open for the sounds at sxsw. There was this one song they ended with at sxsw that I didn\'t hear them play at mercury lounge. I was kinda of in a crappy mood at the venue, and I could have definitely benefited from hearing that song. It\'s ok though. The first thing that someone said to me when the band started playing was, this band sounds exactly like spoon! I was like, yeah they are similar but I think robbers on high street have more lush build-ups in their songs than spoon do. Spoon is more about the sparse, open land of music (at least on Kill the Moonlight), and britt daniel\'s vocals are a bit more closer to the mic... more intimate with a bit more drawl. Robbers in contrast have good vibes shining out of their prettily constructed guitar lifts and pulls and melodic keyboard parts.   april 1: the walkmen @ irving plaza  there is something terribly incredible that happens to me when I hear  urlLink the walkmen . My body starts to quiver and I start to full-on dance. It was amazing. I have to thank peter bauer for putting me on the list for this show, and I have to especially thank him for the vip pass, which I didn\'t know I was getting. It was one of the coolest gestures a band has ever offered me. And I am wholly grateful. I didn\'t use it very much because I like being up front and center so I can dance with all the other die-hard fans. I danced like crazy!  You know how we are so connected to music and the emotions and experiences and people associated with certain kinds of music? What recently happened to me was that one of my favorite bands was sorta ruined for me because of the fact that they were associated with a certain being. It was insane. Now I feel like I can\'t listen to that band anymore because it will always make me think of awful emotions!! However, that will never happen to me and my Walkmen music. Why? Because the Walkies only remind me of my exciting entree into the New York music scene. They remind me of times at shows with friends and swaying in front of Hamilton crunched over the microphone. How times have changed! Now Hamilton owns the stage with scary intensity. His in-your-face performance just blows my mind because I saw the band grow so much from their early days. From shy boys to rock stars. The transformation was inevitable, and it is just awesome to witness. I never get sick of seeing them live -- this is a girl who has seen them four nights in a row, and probably could have gone for a fifth. And then went and saw them three times in one night!  I feel like I am a part of history. The walkmen have a special place in my heart that is only associated with me, and for that I am grateful. Currently loving, \"Thinking of a Dream I Had\"!! But adore \"What\'s In It For Me?\" and of course, \"The Rat.\" Angry, vitriolic, bitter, reflective... step off!  Bsiii called me afterwards and we met up at the walkmen afterparty at hi-fi. We hung out for a bit, and I talked to Peter and Hugh for a little while. JoJo and I also hung and partied. I randomly ran into Fiona, one of my former bosses who I am quite fond of. She will forever be 17! I also saw jennifer, preeti, rachel, who I haven\'t seen in the longest.    april 2: dirty on purpose -- bushwick blast @ asterick art space  I\'ve never really ventured much into bushwick, that place often dubbed these days as \"east williamsburg,\" but I strolled in to see  urlLink dirty on purpose  play the  urlLink bushwick blast  with some of the coolest, noisiest bands out right now. The band was honored and siked to play this festival, which was inaugurated and organized this year from dr.k of  the drayton sawyer gang  She did a stellar job!   And we all pretty much agreed that the show was the band\'s best performance ever, so if you missed it, you missed a historic moment. We didn\'t expect it, but because the band just let loose and experimented with different things on stage for their songs, the creativity just juiced up their performance. They completely changed the end of \"mind blindness\" and they revamped the beginning of \"monument.\" They also dared to play a quiet song, \"cheat death,\" which is probably my favorite song from the upcoming EP. It is so pure, and george\'s bowed guitar part makes me weepy and heart-achey. Can\'t wait for y\'all to hear it recorded with the vibraphone parts! So pretty!  We enjoyed the bad-ass antics of  the Drayton sawyer gang , and the delicious pancakes produced during the set of  tungsten 74 , a band who improv\'d their album during the recording session and then went back and learned the songs they just made up on the blink. How awesome is that? And having food made in front of the stage while they played definitely made me have pleasant associations... I mean coconut and chocolate silver dollar pancakes are fucking awesome! I also think  services  (flux information sciences) were banging on their drums and cymbals with knives... fucking hardcore. And scary.   april 3: coney island at 2 a.m. is a ghost town  we spent another night finishing up the mixes for the upcoming dirty on purpose EP. Yes, you\'ve heard me say that a million times, but this time it was for real and we worked on it until midnight. Then spontaneously we decided to drive down to coney island like we were high school kids driving down to the beach for a picnic. Except it was 2 a.m. and freezing, so we just drove there and came back. Good times! Dirty on Purpose and Rod are like my second family. Good brothers and sisters, like I\'ve always kinda wanted. I\'m lucky to have befriended them all.   april 4: dennis cahlo and hightower @ pianos & the sleepy jackson -- shout @ bar 13  dennis started his residency last weekend at pianos. That kid\'s songs break my heart. I was yearning to hear me some more  dennis cahlo . So I went over to pianos to check him out. He played with a guitarist who was a lil\' too loud, but I liked the dynamic element he added to dennis\' purely beautiful songs. Truth be told, I prefer the stripped down dennis, but that\'s because the bare essence and power of his music is in his angelic, tear-inducing vocals and those wonderful constructed lyrics about love and heartache. I\'m a sucka for a beautiful love song.  Then we went to see the  sleepy Jackson  at the shout party at bar 13. I didn\'t know what to expect really. At that point, I had listened to their CD twice, and each time I kept thinking, this is NOT sleepy. I want sleepy.  But they amazed me at Shout. Their music was rollicking and playful. They were earnest and down-to-earth. There were no antics (well not too many anyways). I wish I had listened to the CD more to sing along to the smooth melodies. It was a mad-house free party. I was kinda feeling intensely anti-social though so jo and I jetted on outta there real fast. I\'ve been kinda anti-social lately. I\'m sorry if you think I\'m ignoring you or hating you. I\'m just not feeling very well. So stressed and incredibly ill. There will be no other apologies.   april 6: passion of the chris: thermals, say hi to your mom, dirty on purpose @ north six  this night is one of the finer moments in new york city rock, where three incredibly awesome bands in their own right came together and totally rawked Brooklyn like there was no death cab for cutie show going on at irving. Haha.    urlLink Dirty on Purpose  played another stellar set, filled with old and new songs. They are constantly writing new music, which is exciting for me, a person who\'s seen them a thousand times now. I still get chills when I listen to them.    urlLink Say Hi To Your Mom  started with Eric playing two songs on acoustic by himself. I love him for his sarcasm and smart lyrics. He and I are like the same person in that way. We like to make fun of a situation that pains us, rather than deal with the harsher reality where humor turns into hurt instead. His songs might sound happy and poppy (though not all the time), but the underlying message is that of confusion and sadness. Dude is mad-talented. I still love \"Hooplas Involving Circus Tricks\" and \"Super.\" I also adore the other members of the band, who I got to know better during the night... I think at one point I tackled the drummer while Dan from SHTYM tickled him. They all rock personally and musically, and it\'s nice to keep company with such awe-inspiring talent.    The Thermals  fucking WOWED the shit out of me. One of the most amazing surprises of the evening. They RAWKED hard. So much energy and sound spouting forth from this three-piece, and Cameron and I were dancing like mad with a few handfuls of other people who were equally as rawked as I was. Holy mother of god. Lordy Lordy Lordy. I took loads of photos. Hopefully one day I\'ll be able to post some. I\'m updating this blog from Alt Cafe. So East Village of me, I know!  Big love to  urlLink Chris Newmyer , the coolest, nicest jesus look-alike I know. I totally adore him!!! Passion of the chris forever!   april 7: the pleased @ mercury lounge & detachment kit @ north six  then I went to see  the pleased , a band from san francisco that draws comparisons to echo and the bunnymen and u2. I got to interview noah the lead singer right before their set. Noah was a sweetheart. His powerful vocal abilities kill me. He gets to into his performance, arching and swiveling the mic stand around with passionate gusto. I met the rest of the band after their set, and they were the nicest peeps. I love when bands are like that. They just want to chill and hang with peeps. They aren\'t snobby or up-in-their-britches. They are basically a DIY band, and I dig that down-to-earth approach to music. Don\'t be cold and distant, working through however many layers of publicists, managers, agents and whatnots. Be warm and into creating a community.  Then I met up with my DoP peeps at the  Detachment Kit  show, another night celebrating Chris Newmyer\'s birthday. I saw Detachment Kit at SXSW and loved them so much. Their music has so many shifts and turns, my head was dizzy with delight. I love the way they sort of arch forward as the music makes a dramatic shift. I love how they clatter around the stage like it\'s some playroom with plenty of disposable toys to kick around. But it was late night and I had to get to bed. Thus now I am sick with a terrible feverish cold.  But I like to support my friends despite my illness. So I\'m going to see  urlLink Au Revoir Simone  at Pianos tonight. They go on at 10:30. Then off to see JoJo Dancer deejay at  urlLink Misshapes . 11:30. Cutting it close, I know.   '),(15095,1694057,'08,April,2004','2004-04-08 00:00:00','   mea culpa  still no internet at home, so i haven\'t really been able to do an update, but here are some more interviews that i\'ve done with new york bands!   urlLink on the scene: williamsburg, brooklyn -- part 2   '),(15096,1694057,'06,April,2004','2004-04-06 00:00:00','   dirty on purpose tonight  tonight is the  dirty on purpose  show at north six! it is also the dirty thirty birthday party show for dj, so bring a bottle of jack or something and drunken him up. fo\' shizzy.  they are playing as part of  urlLink chris newmyer \'s birthday festival, passion of the chris! and we are soooo excited to be playing with  say hi to your mom  (whose album i\'ve been listening to nonstop!! hooplas involving circus tricks!!) and  the thermals  (their album is fucking called \"fucking a\").  dirty on purpose goes on promptly at 9 p.m. see ya there, sweets.  also sorry about not updating yesterday. it will happen soon though, i promise, and then you\'ll hear all the weird shit that has happened in my life for the past two weeks. who knew it could get this weird?   '),(15097,1694057,'05,April,2004','2004-04-05 00:00:00','   happy birthday,  urlLink DJ  ! hope all your birthday dreams come true, sweets.    also, sorry for lack of updates, but i\'ve been without internet connex for the past week... and still don\'t really have any. but there\'s a huge update coming tonight about my crazy past two weeks...  in the mean time, listen to this wonderful  urlLink  nellie mckay   track:  urlLink \"david\"  from her performance at npr  xoxo jinners   '),(15098,1694057,'02,April,2004','2004-04-02 00:00:00','   thinking of a dream I had  I\'m waiting for a dream to arrive! Walkmen last night = fucking hawt. The best performance yet, you knows its a good show when peter\'s smiling from the stage during the set. I was sooo waiting for a second encore. Apparwntly they were thinking about it but by that time everyone had run out for a ciggy. Damn! But at least they played \"what\'s in it for me?\" of which I know all 10 lyric lines oh so well.... Ohmyheart, they are so sweet and fierce at the same. Rawr!   '),(15099,1694057,'29,May,2004','2004-05-29 00:00:00','   the winter pageant and au revoir simone tonight!  two great shows tonight! one in brooklyn, one in manhattan. choose your borough or surf between the two!     urlLink the winter pageant  @ rothko 11 p.m.     urlLink au revoir simone  @ asterick art space early-ish?   '),(15100,1694057,'28,May,2004','2004-05-28 00:00:00','   support the walkmen!    here\'s a friendster message from  urlLink affordable justice :  Go to  urlLink http://www.1077theend.com/playlist.asp   and vote for \"The Rat\" by  urlLink The Walkmen .  This is one of the first commercial stations to add the song and it\'d be fantastic if The Walkmen made it into their countdown...  Feel free to forward to bored co-workers and underemployed friends.  Thanks so much!  ---------------------------  btw, if you don\'t know The Walkmen, then you must be new. welcome, click, vote!  ++ check out  urlLink the walkmen in brazil . ok, no, for real,  urlLink here .  ++ check out  urlLink peter\'s latest reviews  from the road.  ++ then buy their stuff  urlLink here  or  urlLink here .   '),(15101,1694057,'28,May,2004','2004-05-28 00:00:00','   my acronym      J     Joyful       I     Influential       N     Nerdy       Name / Username:       urlLink Name Acronym Generator  From  urlLink Go-Quiz.com       J     Joyous       I     Industrious       N     Natural       N     Normal       E     Exhausting       R     Radical       S     Stunning       Name / Username:       urlLink Name Acronym Generator  From  urlLink Go-Quiz.com   '),(15102,1694057,'28,May,2004','2004-05-28 00:00:00','   the new steve malkmus!     hutch harris = steve malkmus = swoon!      kathy foster = cool as shit!      jordan hudson = cool as fuck!      basically, they\'re all fucking adorable!   also... nice to finally meet you last night,  urlLink catherine ! thanks for all the drinks, beach! thanks for the list action, dt! thanks for the happy stuff, tom! thanks jay for not culkin\'ing me. and thanks to everyone who danced like crazy! the footage on my camera during the hold steady is fucking brill. thanks to  urlLink panthers  for rawking out. too bad people were stupid and left before your set ended, but i guess it was kinda late and rainy. overall the night was stupendous... though the hold steady are just a tad bit on the jammy side! also, sorry  urlLink noah  for calling you sam...  also, john, if you don\'t remember our first meeting, here\'s a reminder of when you took the camera out of my hands and snapped a photo yourself! fun times!   '),(15103,1694057,'27,May,2004','2004-05-27 00:00:00','   tonight: panthers and the thermals @ mercury lounge!    are you down to get totally totaled and sweat to some hardly art, no culture icons and thank me with your hands? thought so. join me tonight in getting smashed at the mercury lounge with  urlLink panthers  and  urlLink the thermals  and  urlLink the hold steady  and  urlLink blood on the wall . it\'s going to be so bad ass!       plus, happy birthday,  urlLink johnny ! multa basia!   plus, a funny  urlLink rodeohead link  from  urlLink adam b . thanks for the laugh! it reminds me of every time joe sings a  urlLink dirty on purpose  song in a country twang that annoys the shit out of me during band practice! i have a recording of him doing \"spider eyes\" like he\'s a retarded dwight yoakam... such a jokester!   '),(15104,1694057,'25,May,2004','2004-05-25 00:00:00','   the big sleep tonight!    ever since i saw  urlLink the big sleep  at the hook earlier this month, i\'ve been obsessed with their amazing wash of sound... very similar to the secret machines, but better and catchier and awesomer! so go check them out with me tonight at  urlLink rothko  (on suffolk street). they go on promptly at 9, followed by  urlLink aerial love feed  and  urlLink asobi seksu .   '),(15105,1694057,'24,May,2004','2004-05-24 00:00:00','   loose record party photos and video    wowee zowee! check out the  urlLink photos and video  i have of the fun times at  urlLink jay \'s  urlLink loose record listening party   and check out what i played  urlLink here . or right below:   May 20, 2004 @ Loose Record Listening Party  belle & sebastian - \"sleep the clock around\" scene creamers - \"better all the time\" seafood - \"pleasurehead\" dirty on purpose - \"all new friends\" weezer - \"islands in the sun\" suzanne vega - \"tom\'s diner (DNA remix)\" pavement - \"shady lane\" the thermals - \"no culture icons\" enon - \"disposable parts\" the realistics - \"why didn\'t you stay?\" the rentals - \"waiting\" seachange - \"glitterball\" the walkmen - \"thinking of a dream i had\" garbage - \"#1 crush\" liz phair - \"fuck and run\" dirty on purpose - \"mind blindness\" french kicks - \"one more time\" bishop allen - \"quarter to three\"   '),(15106,1694057,'23,May,2004','2004-05-23 00:00:00','   The Fever @ Bowery Ballroom  New York Rawk City, a mishappen land very near and dear to my heart, has many a band hoping to not only be the next big things but also grow to be legends or classics. It\'s a common formula for groups here to tag along and ride the wave of whatever rock is popular NOW. It\'s a habit that New York, and in particular, the Williamsburg scene, has afforded bands that don\'t have much vision or originality. Here, bands like A.R.E. Weapons can make the  Spin  magazine spread for hot band lead singers, but we all know they suck and are only hyped because of the celeb connection.... But nonetheless the media feeds and dresses up these lackluster, hype-filled bands with nary a wink of lost slumber. Blogs included.   So I wasn\'t surprised when I got to Bowery on Friday night, hoping to have missed the ARE set altogether, and found people yelling at said band to get off stage. The hostility that the common rock fan feels toward this group seemed in unison that night, summed up in one simple phrase: \"You suck.\" It was that phrase repeated by many that finally got ARE and their weird posse off the stage. Tell me again why their friends were all standing on the stage? Their brand of hopeless noise was enough to make my intestines twitch. Yikes.  The Fever finally got onstage and proceeded to play their album in order of the tracklisting. As lead singer Geremy Jasper stepped on stage, I could see something glimmering at me. Holy shit, he had a fucking g-fab earring on his left ear! Then he turned around and I saw even more glam shining at me -- a huge golden sequinned butterfly that seemed like it could just flutter all over the sold out crowd at the Bowery Ballroom. How fantastic! That alone got me pumped for their set, which started off with \"Cold Blooded\" -- whose title reminds me of a Paula Abdul song -- but seriously, the Fever are retro in such a fun, party hardy way... Then sometimes they surprrise with more heartfelt songs like \"Put It On You\" with parts that scream with emotion. What I kind of missed were the days when I used to go see the Fever and get doused in water or beer and get hit in the head with microphone stands. They have such a raw, unabashedly crazed energy to their music that makes them charming and almost gleeful. The spirit of fun times in New York is certainly within the Fever, and I hope they go on to charm the rest of the world as well, if they haven\'t already.  Then Mike and I headed over to Rothko, where the afterparty included free Sparks energy alcohol drinks (oxymoron???) and performances by Diamonds and Cheeseburger, two bands that probably should have played Bowery along with the Fever. I quickly snatched up a Sparks... And about halfway through I felt like I was going to have a heart attack. I put that shit down right away, but at least now I was awake. This was my first time at rothko, a new Lower East side rock venue that\'s just a bit bigger than sin-e, with three floors. The inside was all red with a spinning disco ball. And the DJ dared to play Jackson 5 back to back with Michael Jackson. Jacko love? Always. It was fucking packed in there, and I didn\'t exactly like the crowd. I don\'t know why but some of the peeps there were a lil\' too trendy and swank-faced for my taste.   Diamonds finally stepped on stage and I thought they were better than I\'d thought they\'d be. I know Jo has been raving about them forever, and I trust her taste in music... But something in me was resistent until I saw them live. The sound at rothko was great. But the vibe was sucky. I didn\'t feel like anyone in the room except for the couple people going crazy in the front were really into them, which is a shame. I definitely want to check them out again. They\'re the kind of band that isn\'t exactly doing anything that hasn\'t been done before (and not many bands these days do), but they made good rock and weren\'t annoying or pretentious about it.   Christie\'s band, Cheeseburger, also had a similar vibe -- fun rock music about partying on Saturday night. The lead singer had on his captains hat and strutted the stage barefoot like a cross between Mick Jagger and Tim Harrington from Les Savy Fav. They were better than the last time I saw them at North Six when they played with Mommy and Daddy. I recalled visions of Christie, a dude with shoulder length brown hair, thick glasses and some kind of tropical printed loose button down, dancing like a joyful maniac during M&D\'s set and thinking, that guy is so fucking cool! That is a rare occurence for me in this city, but it happened that night -- and in fact happens every time I see him. Man, I hope I am spelling his name right...   '),(15107,1694057,'20,May,2004','2004-05-20 00:00:00','   thank you, peeps!  thanks to anyone who came out and enjoyed free pina coladas and my second dj set ever. it was fun fun fun playing my favorite songs like  pavement \'s \"shady lane,\"  enon \'s \"disposable parts,\" yes, two  dirty on purpose  songs,  rentals \' \"waiting,\"  walkmen \'s \"thinking of a dream i had,\"  french kicks \' \"one more time,\"  thermals \' \"no culture icons,\"  liz phair \'s \"fuck and run,\"  suzanne vega \'s \"tom\'s diner,\"  realistics \' \"why didn\'t you stay?\" and  weezer \'s \"islands in the sun.\" etc etc etc. maybe i\'ll have time later to post my complete playlist somewhere... though i don\'t remember what order i played it in. sorry. and thanks to  urlLink jay  and  urlLink hawkins  for putting the soiree together. fucking hot shit! thanks to the guy in the suit who complimented my set. no thanks to the other guy in the suit who tried to dance with me. thanks to everyone else who just was there to have wicked good times! let\'s do it again sometime!  doug, dj, george, annie, sung bin and i were also jumping on the small stage at the tiki room and dancing to dennis and audrey\'s dj sets. they were playing some great dance tunes... i jumped out of my skin when dennis played pizzacato 5\'s \"twiggy.\" i was going to play their song \"magic carpet ride\" but ran out of time. next time, folks, next time!   eric from  urlLink say hi to your mom  did a lovely solo set for a packed bar. it was awesome. he didn\'t play \"hooplas...\" but i still adore him. i\'ll hopefully have photos to post at some point for all of you who love to look at drunk people chugging from free beer and stuff. oh and do i have some footage of eric grooving to some of  urlLink dennis cahlo \'s tunes? yes, i do. and peeps, audrey spun some wonderful tunes as well. they both rock in the dj booth. jay also spun some fun stuff as well. i was so siked to hear him spin the walkmen\'s \"the rat!\"  during the coarse of the night, i also met the dude who does  urlLink casa de punk , and then he just walked away from me... thanks for the chat, dude!   and then i talked with the boys from  urlLink other passengers  for a while and we all danced to that song from  dirty dancing  -- i think it\'s called \"be my little baby\"? it was awesome. we were so wasted. i haven\'t gotten that drunk in quite a while... you know, my tolerance is not that high but those pina coladas were sweet!  and then after the party, a bunch of us crammed into jay\'s car like we were clowns at a circus. i was so embarrassed but so tired that i didn\'t care that i was sitting on jay\'s lap in the front passenger seat hoping and praying that i wouldn\'t fall out of the window or that the cops wouldn\'t stop us kids in our tracks and ask us why we have so many fucking people in the car. hawkins is a wild driver. and dancer. s\'wild!   '),(15108,1694057,'20,May,2004','2004-05-20 00:00:00','   watch asobi seksu\'s video!  here\'s a link to  urlLink asobi seksu\'s \"walk on the moon\" video  on mtvu.com. yo, they look so hot and trendy! that red lipstick is like \"rawr!\"   '),(15109,1694057,'19,May,2004','2004-05-19 00:00:00','   andy kaufman back?  so i hear from  urlLink jay  that andy kaufman is back...  urlLink  . if this is true, then i wonder if the guy dt and i ran into at fez was andy kaufman. if so, he totally kissed me on the lips. that guy was skinnier than the real tony clifton/andy kaufman pictured below...   THE REAL TONY CLIFTON/ANDY KAUFMAN      THE FAKE TONY CLIFTON/ANDY KAUFMAN     however,  urlLink this link  makes it seem false... it\'s all a hoax, people!   '),(15110,1694057,'19,May,2004','2004-05-19 00:00:00','   knife skills cuts you till you bleed    ohmygosh.  urlLink  knife skills   were so fucking tough last night. i have seen every one of this band\'s shows and last night was probably the best i\'ve seen them. no joke. their songs are so tight with many swift shifts that my head was left dizzy. their new song, \"vessels and streams,\" particularly surprised me. there were so many cool changes within the song that rawked so hard... and built up into a fierce bomb, exploding into my ears. the band played my personal favorite, \"summer of teenage lust,\" which i plan on playing for y\'all tonight when i try my hand at deejaying for the second time in my life. that song is so good!        afterwards, george and i headed over to jay\'s house and watched part of  24 Hour Party People . i didn\'t fall asleep this time like the first time i watched the film. perhaps because i only saw like 30 minutes of it or something.   '),(15111,1694057,'19,May,2004','2004-05-19 00:00:00','   do you like free stuff? me, too!    that\'s why you should come to  urlLink j good times \' first  urlLink loose record  listening party tonight! here\'s some of the free swag you\'ll score if you come:  ++ free music from bands like Dirty on Purpose, Say Hi To Your Mom, The Thermals, The Catheters, The Fever, The Streets, Trouble With Sweeney and Washington Social Club. plus, i got a peek into the goody bag last night, and there\'s some fucking hawt shit in there that i know tons of people will want to walk away with including g-fab chains, buttons, stickers and tee shirts!  ++ free Red Stripe and pina coladas  ++ there\'s no cover! just bring yourself and all your friends!  ++ listen to me, audrey and dennis spin lots of good music!  ++ free solo performance by eric from say hi to your mom!  the party begins with me human jukeboxing at 10 p.m. and goes on till whenever. the tiki room is located at 22nd street between 5th and 6th avenue. word, i hope to see y\'all there!  and afterwards if you\'re not stumbling home drunk, check out ma\' boys in  urlLink the winter pageant ! they are playing at pianos at midnight, sharp.   urlLink knife skills  were so fucking tough and bad-ass last night. update on their performance at pianos last night coming soon! and a special belated birthday shout out to deron and katy. much love, peeps!  also,  urlLink buy tickets  for  urlLink the fever  record release show on friday at bowery ballroom! it\'s going to be hot! party party party! and the afterparty is at rothko with diamonds. whoa. be prepared to rawk!   '),(15112,1694057,'19,May,2004','2004-05-19 00:00:00','   donate money to make online scrabble for real!   urlLink donate money to this worthy cause   anyone who knows me knows that i am insane for scrabble... and i\'m hard to beat, supposedly. i dunno. i took five years of latin in high school. whattaya want?   '),(15113,1694057,'18,May,2004','2004-05-18 00:00:00','   living in new york is $$$$$  read about  urlLink new york\'s ubermiddleclass  from the observer today. thanks, anna for sending this along. it\'s really true. living here, you walk out the door and automatically without thinking spend $20. last week i actually went to my ATM and it said, \"insufficient funds.\" that was an eye-opener to my spending habits. now i am much more careful on how much money i withdraw and spend. that extra quarter for movie tix is always makes me huff.   '),(15114,1694057,'18,May,2004','2004-05-18 00:00:00','   you need knife skills in your life    let me tell you that i can\'t wait to see  knife skills  tonight at pianos. what i love about this band is immeasurable bad-ass guitar parts, deadly and threatening drum beats, and sexy, throbbing bass lines that slither and pounce on the audience. they make me want to scream. they make me wanna drive into oblivion. don\'t miss them tonight! it\'s one of their last shows for the summer. plus, the trio will debut a kick-ass new song! hawt hawt hawt!  TONIGHT, May 19 @ 9 p.m. sharp  urlLink  KNIFE SKILLS   @  urlLink Pianos  (Ludlow and Stanton on the Lower East Side)  oh and i forgot to mention that i can\'t stop listening to their song, \"summer of teenage lust.\" i particularly love the lyric that gets stuck in your head: \"if this isn\'t desperation... then i don\'t know what is.\" then the catchy-as-hell guitar part comes in and i\'m totally hooked!!   '),(15115,1694057,'18,May,2004','2004-05-18 00:00:00','   come on, let\'s go to sin-e to see pas/cal!    i got to see  urlLink pas/cal  for the second time in my life last night, and they were oodles of adorable splendor. i love everything about them, from their pink accented outfits to their nerdy glasses to their love of tambourines to caz\' sweet and dear voice to their often goofy storytelling. this band makes this girl goofy and geeky, in a good way.    plus they have great matching shoes. word.  i was quite baffled by the huge turnout for openers,  urlLink elkland , who were like handsome boy modeling 16-year-olds covering erasure songs with their cute, swept-over \'dos... it seemed they brought their entire high school to the party as sin-e quickly filled up with people whooping and hollerin\' for more more more... i swear one song sounded very similar to \"total eclipse of the heart\" as the lead singer held dearly onto the mic stand and sang, \"i really need you tonight...\" oddly enough, the crowd asked for more and the band gave a rather lackluster encore. i think, however, if i was 15 and starry-eyed and a total hag again, i would have LOVED them. i can totally picture them playing the roxy with a bunch of people dancing like crazy under the disco ball.    the more disappointing part was that most people left after elkland and missed  urlLink shelby , a musically superior band that swayed from gentle vocals to fierce rock-out explosions. they surprised me with their beautiful violin parts and the singer\'s exceptional, soaring voice. i want to see more of this band.  i didn\'t get to stay and see much of  urlLink soundtrak . someone told me that they were like U2, but i didn\'t see that comparison at all from the one song i heard. what did everyone else think?   '),(15116,1694057,'18,May,2004','2004-05-18 00:00:00','   pas/cal is like belle and sebastian meets the beach boys!   urlLink   '),(15117,1694057,'16,May,2004','2004-05-16 00:00:00','   MORE LIVE STREAMS OF PERFORMANCES AT COACHELLA  i know most of you have the pixies performance at coachella, but here\'s a new one i haven\'t seen before:  urlLink beck and radiohead live at coachella    a coupla interesting referrals...  i don\'t pay too much attention to my site stats... but today i was curious and looked at some of my referrals. i forgot that i sent  urlLink the unicorns  my photos from their southpaw show and they have posted them on their site... and they also link to my live video of their performance there under their \"links\" section... also,  urlLink lost at sea  links to my unicorns video too!  oddly enough, a couple people\'s resumes came up as referrals... +  urlLink steve goldstein  +  urlLink joshua ross  +  urlLink david blum   '),(15118,1694057,'15,May,2004','2004-05-15 00:00:00','   skateboarding... am i back in high school?    saturday was starting to be one of those days. i woke up with a hangover... throbbing headache. then i went shopping in the thick, sticky heat that was swelling feet in the city, and finally decided to make the trek into williamsburg to hang with ma\' peeps. the catch of course is that this weekend the L train wasn\'t running at all. the mta should give us a refund on our unlimited metro card because it means nothing to me if i can\'t travel into wburg. wtf is up with that? it\'s such an isolating feeling.    as i was chilling out to my summer mix in my apt, joe called me and i decided to meet him at andreas\' art opening at kcdc, a skateshop in brooklyn. anna and ryan decided to come with too. they were going to another art opening, where the theme of the night was anti-war. i don\'t know how i feel about the art i saw there. it seemed reactionary in a childish way that didn\'t really say much more than, \"hey, you suck for making us go to war.\" i didn\'t feel inspired... i didn\'t feel  anything  other than  bored .     then we headed over to kcdc, and it started raining a lil\' here and there. andreas\' work is charming; perhaps i\'m biased cuz i know him. but there were photos of friends, his mom and dad, and screens or paintings of sorts...and there was a live cockroach crawling on his \"broken arm\" piece. that was classic. his work at least made me smile.    i have to say, i haven\'t ever really been to a skate bowl before. when i was in high school, i really liked the idea of skateboarding and skateboarders. i wore big baggy pants and sneakers and band tee shirts... kind of like how i dress now actually. i was a total tomboy. so when i was watching the skaters on the ramp, i was completely fascinated, watching them go back and forth... trying to figure out what the skater ramp etiquette was. it was pretty entrancing to see a subculture like that at this point in my life. i felt like i was surrounded by 16 year olds. i got to see  urlLink andreas  and  urlLink chris  skate, though joe refrained. it\'s too bad, i wanted to photograph him on the ramp. thanks for the free beer, kcdc.  anyways it started pouring, and anna, ryan and joe and i headed over to pete\'s candy store where anna\'s friend alice was playing and goldenrod\'s new country cover band was playing. that was cool and chill. joe and i had walked in the rain, so we got there soaking from head to toe. but it was a fun walk nonetheless. at least for me. joe said he ruined his skateboard in the rain (apparently that is not good for skateboards). we hung at pete\'s for a while and listened to the music, which was good... country covers of rock music by the likes of journey... and other people. i recognized the songs but couldn\'t tell you what they were called or who they were by.   ok, the end.   '),(15119,1694057,'15,May,2004','2004-05-15 00:00:00','   the giraffes @ sin-e    The night started out innocently enough. I met up with dave, who I now want to call pfeff, at a bar somewhere down the street from pianos. We hung out and drank for a little bit and then headed for sin-e to see  the giraffes , a band dave can\'t stop raving about. I have to say, there is something very intriguing about the mustached singer with the slicked hair. I got pulled into a van where the band was hanging out and drinking jim beam choice. The whiskey was smooth and strong. I think they made me take one too many sips. I remember saying hi to chris, dj and sung bin... The rest of the night was a drunken haze of loud loud LOUD music. But I have photos somewhere that I will hopefully post at some point. Hehe.   '),(15120,1694057,'14,May,2004','2004-05-14 00:00:00','   oh my gosh. super scrabble!  anna, deron: it\'s awwwn. i\'m gonna kick ass in  urlLink super scrabble! . thanks,  urlLink cameron  for alerting me. i fucking love scrabble. i challenge everyone to a game!   '),(15121,1694057,'13,May,2004','2004-05-13 00:00:00','   diesel wha?    ok i went to see  urlLink au revoir simone  play at pianos last night -- met up with new friends dave and chris. the ARS girls were in top form last night and covered bruce springsteen\'s \"dancing in the dark.\" hot shit!   more au revoir simone photos:            dave, chris and i headed over to the diesel party after their set, and on the way got the call from  urlLink cameron : there were too many people at the party and no one can get in. too bad we were already in a cab heading over. whatever! if you went to the party/show, tell me how it went? i also said hello to the  urlLink bishop allen  kids standing outside the show. they couldn\'t get in either. boo!  afterwards, dave went home (boo!) and chris and i hung out at alt cafe where mike was working. i had an iced coffee, and after i got home, i went right to sleep. coffee has no effect on me anymore. so sad!  this morning i put on moz\'s \"do your best and don\'t worry\" again and it made me feel so good. it\'s probably his most uplifting song ever, eh?  right now i am entranced with the  urlLink on!air!library!  cd. there\'s always been something in their moodiness that has caught my attention and my heart. ever since i saw them in that dank basement of cb\'s gallery with their million-part instrumentation and their haunting vocal melodies... and eery chords and sounds, i have been swept up in the rush of o!a!l!   '),(15122,1694057,'13,May,2004','2004-05-13 00:00:00','   get loose!   urlLink   '),(15123,1694057,'12,May,2004','2004-05-12 00:00:00','   you give me fever    i went to  urlLink the fever  listening party last night for their upcoming album,  red bedroom . it was packed full of peeps, and you can\'t really beat $2 beer, unless of course it\'s free. songs like \"labor of love\" and \"hexxxed\" get me energized -- so much so that i felt my body moving, despite the comatose state my brain was in. as  urlLink miss modernage  has already stated, the album illustrates the band\'s diversity, and if possible, their songs are even catchier this time around than on their  pink on pink  ep. bravo, kids! if you haven\'t bought  urlLink tickets  to their record release show at bowery ballroom yet, you should grab them quick. i hear they\'re going fast!  i talked for a little bit with brian long, who manages  urlLink vhs or beta . you might also know him as the guy who put together the  yes new york compilation . he got me siked for the new vhs or beta album,  night on fire , where craig pfunder debuts his vocal artistry. he also talked about his experiences in the music industry and how he got into managing a band, which i thought was cool. i love to hear stories like that because it inspires me. the album comes out in a couple weeks... i\'ll let you know what i think of it when i get the cd.  holy mother of gosh. i am drinking a berry-lime  urlLink jamba juice  from the whole foods on columbus circle. it is so good! but a and were saying cherry-lime would have been infinitely better. just something to consider...   '),(15124,1694057,'12,May,2004','2004-05-12 00:00:00','   sea ray, metric, the stills @ irving plaza    whoa, when  urlLink sea ray  explode on stage, the audience is swallowed up in a warm wash of intense sound. just look at their cellist above. c\'mon, tell me that doesn\'t look intense.   '),(15125,1694057,'12,May,2004','2004-05-12 00:00:00','   the shining with bunnies   urlLink watch this 30 second clip.  i love bunnies. thanks, barbara!  i have decided that the best way to start each day is to listen to morrissey\'s \"do your best and don\'t worry\" on repeat. it\'s a daily affirmation that life ain\'t that bad, even when you have $2.50 in change in your wallet, along with a lot of old atm receipts.   '),(15126,1694057,'12,May,2004','2004-05-12 00:00:00','   the new york dating game  i think  urlLink dave , a fellow  urlLink walkmen  fan, should date me.   '),(15127,1694057,'11,May,2004','2004-05-11 00:00:00','   ENTER THE CONTEST     urlLink The Diesel-U-Music Contest  has born three winners this year:   urlLink The Fiery Furnaces  ,   urlLink Prefuse 73   and   urlLink Madlib  . And now what? Theyâ€™re having a big-ass party on Thursday, May 13â€¦ and you and a friend can go.  urlLink Just email me.  I have five tickets, each with a plus one! Winners will have to meet up with me before the show and claim their tickets. Oh yeah, and the party is at @ Splashlight Studios (529 W. 35th Street) at 8:30 p.m. Word!   '),(15128,1694057,'10,May,2004','2004-05-10 00:00:00','   movie aside  if you thought for a second that getting stoned and watching  urlLink  van helsing   was going to be a good idea, i can re-assure you ahead of time that it is NOT. thanks, joe...  after that horrible movie experience, george, joe and i meandered to the east east village to rue b, a saucy lil\' bar on avenue b complete with a classy piano player. the place was kinda small, and i bumped into the piano player and made him mess up his song. sorry, kiddo! there we met up with a man by the name of daniel sasse (pronounced like that teen magazine of yore). i kid you not. what an amazing name to live up to... and peeps, he certainly did live up to that name. his talk was amazingly involved and strangely entrancing. i fell asleep on his couch for 15 minutes. but he made us amazing coffee in four minutes with a french press. we were impressed. and awake.  then george and i, newly awakened, wandered over to pianos, where the shins and any number of other comedian and celebrity hipster types were all sipping on clear cocktails with limes and chillin\' after the big bowery ballroom show. i heard their gig with moz on friday night was splendid! as was SNL with the olsen twins and their bowery gig with the hold steady -- as DT put it, \"their best show i\'ve seen them in ever which includes about 20 shows or so.\" wow, big praise!  my dj gig went fine, and it was so much fun to spin music that i like listening to. i\'ll be spinning again with  urlLink miss melody nelson  at a huge listening party with  urlLink dirty on purpose ,  urlLink the thermals ,  urlLink say hi to your mom , catheter and more that jay good times is putting together. that kid has so much love for music, it makes me inspired. save the date: may 20! check the  urlLink on the moon music  calendar for details as they develop!   '),(15129,1694057,'07,May,2004','2004-05-07 00:00:00','   Fever Listening Party!  Wha?! You haven\'t heard the new  urlLink Fever  album yet? Well, you don\'t have to wait until it hits record stores on May 17, my peeps. You can hear it this Wednesday with lots of cheap beer and fun video games. Wha?! Then, if you like it, and haven\'t bought tickets yet, you should buy tickets to see the band massacre the Bowery Ballroom stage on May 21 with their unstoppable beats, and that incredible Geremy Jasper \'do. The debauchery starts here, peeps. Yep, read the message below:  Friends, drunkards, fellow degenerates...  Come down to Hi Fi (169 Avenue A) and listen to the new  urlLink Fever  album  Red Bedroom  (out May 18th) and drink cheap Brooklyn Beer.    Can you beat The Fever at pool?  Pac Man?  Trivia? Contests throughout the night to win tickets to their concert at Bowery Ballroom on May 21!!    The Fever Red Bedroom Listening Party Wednesday, May 12 Hi-Fi 169 Avenue A 9PM â€“ 11PM All Brooklyn Beer pints and bottles 2 Bucks!!   '),(15130,1694057,'07,May,2004','2004-05-07 00:00:00','   so what am i doing tonight? hmmm....   well for the first time ever, i\'m going to try my hand at deejaying. i\'m really terrified, but i have this need to share the music i love with people. and since i don\'t have a cd burner anymore to make and send mix CDs, i figured DJing was the next best way to do it. what is cool is that i\'m getting to do it with one of my faves,  urlLink mike greensleeves ! the other day we came up with an amazing playlist (just an outline, i\'m planning on throwing in a few surprises here and there), and i realized mike and i are PERFECT dj partners because our tastes in music are so similar, but our catalogue of music is just different enough to make it interesting for the both of us.  if you don\'t happen to be going to the moz/shins extravaganza, come party with me! despite what you might have heard,  urlLink the hook  is incredibly easy to get to by way of car or cab --  and if you turn in your cab receipt at the bar, you get a free drink!  the hook is also a wonderful place to see some good rock, the sound is amazing and the space is really impressive. here are the deets for the gig tonight (this is what i posted on  urlLink friendster :   COME SEE ME DEEJAY, PLEASE!   For those of you NOT attending the Moz/Shins extravaganza on Friday, let me tell you about another exciting New York rock event... Uh, yeah, the debut of me as a deejay. Are you titillated yet? Thought so.  Yes, I\'m still a deejay virgin, but gosh darnit, don\'t you want to be there when I pop my turntable cherry? I feel like you\'ll like listening to the music I\'m going to spin with my partner-in-rawk for the night, DJ Greensleeves, or as I like to refer to him, Mike from Winter Pageant. Ahem.  We\'re also deejaying with Sam from  urlLink Interpol  and Audrey from  urlLink Melodynelson.com . Plus, some really cool local bands are playing:  urlLink Qatsi ,  urlLink Other Passengers ,  urlLink The Big Sleep  and Feast.  We plan to bring it awwwn, so put on your rock shoes and cab it over to the Hook ( urlLink www.thehookmusic.com ) this Friday night, May 7. Come on, we know you wanna. Mike and I start the music from 9 to 10 and again around midnight right before Feast go on.   DIRECTIONS:  Here\'s how to get there from the city. Save your cab receipt for a free drink at the bar! + Take the Brooklyn Bridge + Make a right onto Atlantic under the BQE + Left onto Columbia + Right onto Degraw + Go left and stop + Left on to Van Brunsik + Left on to Commerce + and it\'s between Richards and Columbia.  See how easy it is? Plus, you will get some free hugs from moi for coming.   xoxo jinners   '),(15131,1694057,'05,May,2004','2004-05-05 00:00:00','   coffee and cigarettes  lately, i\'ve really taken to hanging out in coffee shops instead of bars. the bar scene gets tired after repeated exposure, though i still love going out for a smokey smoke. i\'m also tired of drinking, after a week-long respite from alcohol, last night i had a drink at 2a with  urlLink joe  and gary, who made their first appearance in my new hood! i was actually about to go to bed when i got the call at 1 in the morning from joe. and i was like, \"when is joe  ever  in my nabe?\" never! so i need to hang with him while his jokey ass is here. hanging with joe is like laughing in spring time. joyous as he\'s a primo comedian who likes to greet me in a lovely british accent: \'ello, jin! where\'s my gin? gary is his best friend from erie, who is probably one of the nicest, down to earth guy i\'ve met. we all talked about how in erie, $5 for a beer is simply preposterous. yeah i said it, preprosterous. let\'s just say that they tip with quarters in erie -- that\'s how cheap the beer is!  there we met up with  urlLink jay  and  urlLink noah  (happy double-two birthday!) jay had just seen the bouncing souls and was wearing a cute green cap, his lip piercing still in place. i\'m surprised the ruffians at the concert hadn\'t ripped that off. noah was sufficiently drunk, and told me he was going to the moz/shins show tonight. i\'m utterly jealous.   '),(15132,1694057,'05,May,2004','2004-05-05 00:00:00','   !!! Music Video Taping Tonight  !!! are taping a music video tonight at  urlLink North Six  and need lots of ladies ready to shake booty at 6 p.m. I think you can just show up.   '),(15133,1694057,'04,May,2004','2004-05-04 00:00:00','   From the East River Music Project  Our first summer show at the amphitheater is just around the corner, so here\'s where our roster stands as of right now.  All shows are on Saturdays, all from 2-6, all FREE.  - 5/22: White Magic, Measles Mumps Rubella + more tba - 6/26: Cat Power, Double Leopards, 2/5 BZ, Valley Of Ashes, DJ Brian Turner - 7/24: The Natural History, Sea Ray + more tba  If you\'re down to volunteer your time or just donate some cash (we\'re a freshly incorporated not-for-profit/501c3, so all donations are tax deductible), write to donations at eastrivermusicproject dot com.  ERMP opperates without any corporate sponsorship of any kind, so we appreciate your support.  For more info (including directions to the amphitheater), look to http://www.eastrivermusicproject.com.  Expect updates about summer line-ups as well as a benefit show this month at free103.   urlLink http://www.eastrivermusicproject.com   '),(15134,1694057,'04,May,2004','2004-05-04 00:00:00','   dear peeps  Sorry I have been slack on the updating of my adventures in nyc. Does anyone besides my friends even read this site?   Anyhoo, there\'s been TONS of stuff going on lately in my night life except when I\'m at work I can\'t do a proper blogger update cuz I\'ve been swamped with award show writing at work. Believe it or not, part of my job is to write bios of award winners and also even writing the entire script for the different genre award shows we have at work. I\'m no bruce vilanche, but I do learn a lot of stuff about various artists along the way. Working for a songwriters organization is definitely a learning experience for me, plus it\'s been fun getting to know the ins and outs of the biz.  What\'s bad is that in my outings I\'ve taken to smoking a lot more. This is a result of hanging out more with smokers. I\'m what u might call a social or stress smoker. I smoke in social sitches or when I\'m super stressed. Lately I\'ve been both stressed and social -- so double the smoking! U know what though? I like it.  Hanging with  urlLink DoP  has also caused me to love bbq\'ing. When u know someone with a backyard and they like cookouts and the weather is warm, then it\'s all good. This weekend is gonna be a double picnic extravaganza with erika\'s birthday on saturday in mckarren park and deron and katy\'s birthday on sunday in central park. Cross your fingers for sunshine! I\'m reading to chill out on blankets and grass.  I\'m also getting really excited about deejaying this friday at the hook, even though I think some of my close friends and others I haven\'t seen in a while won\'t be able to make it. Things like scheduling things in support of each other is really tough, especially if your friends work in the music biz. I know I constantly feel bombarded with shows and dj gigs of friends and since I can\'t be in two or three or four places at once or sometimes I\'m just plain tired, I miss a gig here or two. Please forgive me if this happens. I still love u and your band, and will keep trying to catch you. Many of my friends have band practice or are in the recordiing studio or have to catch their other friend\'s gig or have some kind of birthday party to go to. Life in new york seems to revolve around everything like that and the buzz of the city is built on this \"busy as a bee\" syndrome.   I think I have come up with a bunch of key songs that I want to spin on friday. I will post more details on the dj gig, complete with directions to the venue. I would love to see your adorable face there to eat up the fantastic mix of delectable music morsels I plan on feeding u. Woohoo!   '),(15135,1694057,'04,May,2004','2004-05-04 00:00:00','   i love....  listening to ella fitzgerald. what do you love?  i am very excited about the fact that my peeps in  urlLink pas/cal  are playing the next vicious!  i\'m also excited about the  urlLink sea ray ,  urlLink metric ,  urlLink stills  show next week at irving plaza!!   '),(15136,1694057,'28,June,2004','2004-06-28 00:00:00','   sweet and gentle music    right now i\'m kinda digging on this track from the chicago band,  urlLink the changes .  urlLink listen to it here   '),(15137,1694057,'25,June,2004','2004-06-25 00:00:00','   the anti-bush online adventure  my friend caspere sent me this:  urlLink the anti-bush online adventure   '),(15138,1694057,'25,June,2004','2004-06-25 00:00:00','   the paper chase & we ragazzi @ north six while most people were probably getting totaled at the les savy fav craziness at volume on friday night, anna and i zoomed on our bikes over to north six to check out her friend\'s band the paper chase from dallas/denton texas. isn\'t everyone from denton these days? we got there in the middle of their set, but just in time to hear my favorite song, \"don\'t you wish you had some more.\" unfortunately though, there weren\'t that many people there. i still enjoyed watching them create their spastic set, complete with sound effects, guitar, bass, drums and keys. it was hot and sweaty, just the way i like my friday night rock. if the band isn\'t dripping with sweat, they didn\'t try hard enough, you know?  i had never seen we ragazzi before either, and this three-piece was definitely interesting. anna says the lead singer sings like prince, but what fascinated me about the band was the kick-ass drummer girl. not only did she hit those drums hard, she did it with such a unique style. it was tough and forceful without being out of control. though their songs started to sound more and more the same, i did like what i heard... and am curious to hear more -- hopefully more diverse songs from this band.   celebrate brooklyn! neil young tribute @ prospect park on saturday it was a beautiful day. i got a call from annie asking if i wanted to ride our bikes over to prospect park for the neil young tribute. i was game. annie brought the scrabble board and made this delicious picnic for us. we were zooming along, and about three-fourths of the way to the park, the chain on my bike broke. it was so sad. i abandoned the chain on the street (r.i.p) and we walked our bikes the rest of the way there.   we were quite early and stood in line to get into the concert. half of the time we were trying to figure out how we were going to sneak our wine inside. they were checking bags... shit! annie stashed the wine in the bushes near the fence... we thought we would go get it later when we were safely inside. doug and his friends from north carolina met up with us a little bit later. we found out they snuck two bottles of wine in. and we went to fetch our stashed wine. i had kind of a headache and decided not to drink. i was quite impressed that at the end of the night, annie, doug and the gang had finished up the wine! drunks!  annie and i slowly played scrabble as people walked by and commented on our game. people would say thinks like, \"is \'liven\' a word?\" we\'d reply, \"why yes it is.\" the concert was pretty bad. though cat power, iron & wine and ron sexsmith i thought were good, most of the people who played were just plain awful... kind of a massacre.   while there, i ran into dt, noah, josh, sean and robin. they were all there just hanging and of course representing for iron & wine. hotness. dt said he never heard sam beam sing the way he was singing at the tribute. it was more of a country style of singing. it was good though. as that crew rolled out, i stayed and hung out a lil longer with annie and co. we took the train home and luckily doug\'s friend helped me carry my bike home (thanks dude!). overall it was just a nice day spent in a beautiful park. i heart new york!!!   lazy sundays i woke up the next morning to go and fix my bike chain. there was this obnoxious texas family in line in front of me, just completely boasting their richness. the dad was buying bikes for his two daughters, who were in their 20s. they kind of gave off spoiled vibes as their dad paid for their $450 bikes. the cashier guy was like, it will probably take a couple weeks to get the bikes made, but don\'t scream if it\'s a little longer. the dad replied, \"oh don\'t you worry. they\'ll definitely scream.\"  after i got my bike fixed, i rode over to mckarren park. met up with doug and annie again. we tried to go over to pete\'s candy store, but it wasn\'t open yet. so we bought a bunch of stuff to make dacquairis (spelling?) and also a kiddie wading pool to fill with water. it was pretty awesome. we dipped our feet in the pool and got drunk on dacqairis and mojitos. yummy! somehow i rode home drunk in the dark, totally buzzed on the wonderfully relaxing sunday. i love weekends like that!   a little extra something-something   urlLink if you\'re bored and feel like starting a restaurant, here\'s how to pitch your restaurant to the mayor.   '),(15139,1694057,'25,June,2004','2004-06-25 00:00:00','   my awesome roommate anna    i read a lot of blogs... and everyone always talks about their cool roommates, so i figured, i should start doing it too, so i can be wholly awesome.   ok, so let me tell you about my roommate, anna. i met her about two years ago, so unsuspectingly. i used to be the music editor for  urlLink a teen website , but i was also responsible for posting content on other sections of the site, like a fashion column called \"dig or dis,\" where we took clothing from our delias-like catalog and posted it on the site with some cutesy question like -- floral tank: wilted or in bloom? and the kids would post their thoughts on the java board. most of the time they just talked about how the model\'s boobs were too big or too small or too fat. it was the stupidest thing ever, but i had to take one of my skinny coworkers each week and take like 7 photos of her wearing various clothing from our catalog. anna was working as a receptionist on the floor where all the fitting rooms were. we instantly connected as we laughed about my ridiculous job. she suggested i write for  urlLink venus zine ,  urlLink popmatters  and  urlLink kitty magik  instead... and eventually i did.     hearing i was the music editor of the site, she talked about her band,  urlLink knife skills . they were going to play their first show soon. i told her i would check them out. around that time i was throwing a mod party at my apartment, and invited her to come check it out. i didn\'t think she\'d actually come. i didn\'t think anyone would come. it was raining horribly that night. but people did, anna included. from then on, we became really great friends as we started to hang out more and more.   her boyfriend, deron, now her hubby, was working sound at a club called luxx in williamsburg. i kept reading these huge articles about luxx and the whole electroclash thing. it was around the same time i started to dig the walkmen and basically the entire  urlLink startime records  catalog because my friend miker was doing the website and an old college acquaintance isaac ran the label. i started to hang out with her and deron at various rock shows.     i started to meet tons of people, including members of the asian rock jury (pictured above). i started listening to a bunch of cool bands i\'d never heard of before. i started taking photos at shows. then i started a blog. she introduced me to music like turing machine, vaz, les savy fav, bishop allen, the new humans, the big sleep, on!air!library!, semiautomatic... most recently paper chase.   she introduced me to vegan food, and i dig her tofu and her vegan Magnolia-like cupcakes. she makes me more politically aware than i would be on my own. she helped me start my own monthly  urlLink On the Moon Music Presents...  party at Luxx, which I stopped doing soon after Vicious took off because booking shows stopped being fun. She encouraged me when I decided to start managing  urlLink a band .     she introduced me to my long-lost cousin,  urlLink slim moon !  usually people are never around for my birthday because it\'s two days before christmas, but anna (along with my ex) managed to round up all of my friends and throw a karaoke party for me... two years in a row. last year, she and i went to chinatown together to buy dumplings for the party. it was pretty spectacular, and made me feel loved.   she\'s also one of the only girls i know who can do high kicks. she also sings that \"thank you for coming to loews. sit back, relax, enjoy the show!\" song and dance better than lindsay lohan. she\'s also the toughest female rock star i know. just check out this photo. too many girls these days just want to look fucking cute and coy. anna is badass. she screams. she writes menacing guitar lines. she tells jokes in between songs. she writes great lyrics. she\'s BADASS and not afraid to show it.  And then when I had basically nowhere else to go, she let me move into her apartment. i\'ve since come to adore lucy, the stray cat she and deron adopted. and this weekend, she and deron are adopting a dog named nikita from mighty mutts. are there people our age who do caring things like that and are so giving of their time and love? not many. plus, she plays scrabble with me. She\'s pretty fucking awesome, and one of the coolest people i know.   '),(15140,1694057,'24,June,2004','2004-06-24 00:00:00','   don\'t you wish you had some more    i went to see  urlLink french kicks  play bloomies soho yesterday... it was kinda weird. but my love for them is so strong, i ignored all of the weirdness, danced, sang along and had a great time. free shows are always awesome. they played mostly stuff from the new album, including my favorite, \"one more time,\" and also a couple oldies, including \"close to modern.\" here\'s a  urlLink video clip  of them playing \"close to modern.\" btw, their drummer is amazing. mike, byron and i tried to drink the non-existent \"cocktails\". after some foraging, we found stella served in wine glasses. some people were walking around with trays of bread. tons of cute indie rock kids showed up. i think only one girl was full-on dancing and singing. indeed, we need to teach the indie rock kids how to dance. stop thinking you\'re too cool to dance. how can you not dance to french kicks? they\'re so damn catchy. shame on you.          cass mccombs and iron & wine @ maxwell\'s i got a last minute invite to the sold-out cass mccombs and iron & wine show at maxwell\'s. the last time i went to maxwell\'s (in hoboken) was to see the walkmen. i swore after my treacherous commute to and fro, i would never go back to maxwell\'s. but there i was, taking the path train, heading out to jersey on a thursday night. but c\'mon, cass mccombs canceled all of his new york dates and this was his only area show for a while. dt and i got there in the middle of cass\'s set... cass was wearing some kind of eyeshadow and his brand new band had a guy wearing a dress like he just came from some hippie festival. i thought, uh oh. while they did mostly play fast songs, cass did play a couple of my faves: \"aids in africa\" and \"not the way.\" so beautiful.     i have to say that this is my first time hearing or seeing  urlLink iron & wine  in concert. and boy was i looking forward to it. the band did not disappoint. their songs were so soft and sweet and pure -- probably the most sincere music i\'ve ever heard. ever. the lead singer, sam, had this voice that was like a soft whisper in your ear, incredibly intimate. it gave me shivers. i just closed my eyes and swayed to the music. i found out that i had seen the drummer\'s other band last week. he also plays in a band called teen wheat who played with circle and square last week at lit. iron and wine and teen wheat are total opposites as far as musical styles, so it just goes to show you that you can use music to express your creativity in oh so many ways.   '),(15141,1694057,'24,June,2004','2004-06-24 00:00:00','   what am i doing tonight? FREE STUFF!  6 p.m.  meet up with mike and byron for the french kicks performance at the bloomingdales in soho. FREE drinks!   8 p.m.  move on to some event at webster hall that my friend susan helped put together for some kind of music technology product... uh FREE drinks! even the top shelf stuff. sweet!   10 p.m.  sam champion and these bones perform at luna lounge for FREE! good rock/n/roll. woo hoo!   '),(15142,1694057,'23,June,2004','2004-06-23 00:00:00','   can we say, astrobi seksu!?    wow,  urlLink asobi seksu  put on a great performance last night. i was impressed with their live show -- their songs really come alive... especially with their strobe light endings, though that got a bit epileptic after a while -- for one or two songs, it\'s a great effect, but after four or five, i start seeing dots and dots and dots... yuki looks like she\'s in the throes of ecstasy on stage. she just kind of throws her head back and closes her eyes. i wish there were more songs with james on vocals too. he\'s got a great voice that kind of reminds me of the mink lungs. hot damn!      urlLink mellowdrone  exceeded my expectations. i had only seen jonathan bates do his solo mellowdrone thang before, but this time it was pretty awesome -- the sound is obviously much fuller with a band backing him and his wild stage antics. his voice travels from sweet jeff buckley-like notes to low sexy drawls -- and it all seems so effortless for his skinny ass. i\'ve never seen someone that skinny. he kind of threw his slim frame all over the stage, pushing mic stands over, stumbling over himself and then finally at the end just falling onto the floor in front of the drums. very mellow-dramatic.    and for those of you who stayed for  urlLink au revoir simone , i know you were not disappointed. the girls put on a solid show with their dreamy soft melodies. and they looked pretty cute while showing off their catchy fast songs and then falling into their more numbing, romantic electronic ballads. they get better and better every time!   sidenote:  dude the line from the killers outside mercury lounge was just ridiculous. it was a mile long, and full of \"hipsters.\" are you a hipster? get in line, right here. haha.   oh my, that billboard scandal ok i\'ve met keith girard a couple times through my job (most recently at sxsw), so when i\'m reading about  urlLink this huge billboard scandal , it\'s even crazier!   According to the formal complaint, \"gender and race-based discrimination, sexual harassment, retaliatory firings, intra-office sabotage, and other wrongdoing ... took place in the offices of the music industry\'s leading trade publication.\"  The complaint adds that Girard and Chang \"found themselves in a newsroom culture in which corporate management readily sacrificed editorial integrity for the sake of financial interests, [and] sexual philandering among senior and junior-level staff members was not only condoned but encouraged.\"  \"Schlager frequently and repeatedly made inappropriate remarks to Chang of a sexual and/or erotic nature. ... During all relevant times, Schlager maintained a green phallic vibrator, of the type normally intended for sexual use, in full view on a shelf in his office. On at least three occasions, Schlager removed the vibrator from the shelf and showed it to Chang, and on at least one of these occasions he turned the vibrator on to show Chang that it still worked.\"   '),(15143,1694057,'23,June,2004','2004-06-23 00:00:00','   wtf is wrong with this world?   honestly, right now, i can give two shits that lollapalooza is canceled. i couldn\'t care less about bill clinton\'s best-selling memoir. why? because this world is just  urlLink totally fucked up . terrorism is spreading to attack all parts of the world. and we\'re boo-hooing about not being able to see moz and the pixies while some mother and father across the world is crying over the beheading of their son. i feel so helpless. i feel sad. wtf? anna says this is the beginning of a new world war, and that thought scares the shit out of me. why is there so much hate in the world? can\'t we all do something to make each other feel good? god, it\'s so frustrating.   '),(15144,1694057,'22,June,2004','2004-06-22 00:00:00','   is clear channel evil enough to pull the plug on lollapalooza?  from a listserv somewhere: We had a couple rentals for Lollapalooza (our business specializes in renting vans to touring musicians) and of course this is a big setback for us. From what I have heard the main problem wasn\'t ticket sales (Major markets were doing just fine, and Baltimore was already sold out). It was that Clear Channel decided they didn\'t want to support what is basically a liberal festival. With organizations like MoveOn.org doing major promotions over the whole tour (which in turn were being documented by VH-1 for a series) Clear Channel decided to pull the plug. Without Clear Channel there was no place to play, since as we all know, they own or control almost every large venue able to accommodate a festival of this size. The promoters aren\'t saying anything because they have to do business with these guys, but that\'s the back story as I heard it.  Sharky Laguana (ex Creeper Lagoon) Co-Founder Bandago Van Rentals   '),(15145,1694057,'22,June,2004','2004-06-22 00:00:00','   the morning after   urlLink jason productshop\'s new band the morning after   '),(15146,1694057,'22,June,2004','2004-06-22 00:00:00','   the paper chase & foreign sounding bands  i can\'t stop watching  urlLink this video  for  urlLink the paper chase . can\'t wait to see their nervous, twisted rock at  urlLink north six  on friday (june 25). they\'re playing with  urlLink we ragazzi , another band i\'ve been meaning to check out, but my plans just never work out. all the stars are aligned this friday though. you should come out too! that is if you\'re not planning on  urlLink les savy fav  and  urlLink q and not u  at  urlLink volume  or  urlLink iron & wine  at  urlLink bowery  or  urlLink enon  and  urlLink mellowdrone  at  urlLink rothko  or  urlLink saintface  at  urlLink sin-e . holy shit! too many choices.   UPDATE:  I forgot to mention another great show on Wednesday, June 23 @ rothko. Some of my favorite bands are playing it: Local faves  urlLink Asobi Seksu  (which means \"play sex\" in Japanese -- rawr!) and  urlLink Au Revoir Simone  (i believe that is a peewee\'s playhouse reference? anyhoo, these four ladies are dreamy electronic pop) and  urlLink Mellowdrone  (who i\'ve only seen solo, but he\'s amazing and i can\'t wait to see the full band play!)  let\'s also not forget  urlLink sam champion  (very reminiscent of one of my faves, pavement) at luna lounge (free show!) on thursday, june 24, with  urlLink these bones , who do a funny cover of britney spears\' \"toxic.\" show starts at 10...  or if you get a chance, my friend jumee\'s friend, miho hatori, who used to be in cibo matto is playing solo at pianos at 9 on thursday, june 24. one of anna\'s friend\'s band,  urlLink the quick fix kills , are also playing that show.   '),(15147,1694057,'21,June,2004','2004-06-21 00:00:00','   is this what touring life will be like?     urlLink dop  did a double show this weekend, one in  urlLink jersey , one in  urlLink nyc . as we are currently booking a tour right now, we got a little taste of touring life with the trek to good ol\' passaic. passaic is in the middle of nowhere. we all hopped in a van with  urlLink the girlfriends , and happily zoomed on over to jersey. we found out that, as expected, dj is a swervey and aggressive driver. i thought we were going to die. but as the trip went on, he got better, so i guess practice makes perfect.   we ate dinner at some sporty bar that had like 10 big screen TVs, and one of the nicest waitresses ever. there was a dj who was pumping all this techno dance music from the 90s. i felt like i was in some kind of time warp. even the CVS was playing wilson phillips from the radio. sung bin ran around with a rubber lobster on her head, and annie and i perused the makeup aisle. yes, suburban life is this boring!    when we played the dead disco party at loop lounge (which is like the tiswas or misshapes of jersey), we were happy to see that the crowd was really receptive to us, including the local band that followed us,  urlLink mynuskris  (pronounced \"minus chris\" -- as in minus this guy kris who left the band). the party promoters todd and geoff were also totally sweet guys, and that made us feel good about playing a show in some area that definitely needed some good music to pass through. we were pumped to play our set, thinking it would be a great warm up show for the mercury lounge show the next day in new york. we were right. the set flowed beautifully, and dop were charged. and we left jersey. home sweet home!   the next day, i was feeling pretty rough. riding around in the back of a van with a bunch of amps and people isn\'t exactly my idea of comfort, but it was fun. my back was aching and my head was feeling heavy. i tried to chill out for most of the day to psyche myself up for the evening\'s show. it was part of a  urlLink karipalooza  celebration of one girl\'s awesomeness in the music scene. hot stuff. i especially love supporting other girlies in the scene, we all gotsta stick together, ya heard? the mercury lounge show promised to be excellent with its hyped up bill: dennis cahlo\'s new band, the sons of sound, benzos, the band kari manages, and the bravery (or rather, flock of seagulls!). the early part of the night boasted kick-ass rock, while the latter part of the night summoned tripped-up dance music.    dennis\'s new three-piece band,  urlLink the sons of sound , completely blew me away. i had no idea what to expect, but i didn\'t expect the bold pairing of gentle jeff buckley vocals with such dark, badass music. wow! the textured sounds of two guitars and one forceful drummer is just explosive. i can\'t wait for their next show! i can\'t help but say it though... i will miss how dennis\'s solo stuff made me want to cry and yearn for sweet, beautiful things. is that cheesey? sorry i\'m so completely bridget jonesin on you.    i love anytime george breaks out the bow in dop songs. i don\'t know how he makes it sound so beautiful and pure, but his bowed guitar parts simply soar over the crowd and makes my heart pitter patter with such sadness. i also loved that doug stood up front for \"cheat death\" three part vocal harmonies.   the band started with such a rockin\' song -- \"mind blindness\" with a special extended ending that gets so fucking loud without losing any momentum or catchiness. they also debuted that brand new song, \"your summer dress,\" which joe sings on -- his voice is such a trip, but i love its strange, sweet quality. it\'s almost other worldly.  there were these kids dancing up front like it was a fatboy slim \"praise you\" video. it was totally cracking me up. those dances were worse than elaine\'s from  seinfeld . yikes!!! all in good fun though. all in good fun. ok this post is getting way too long.  i finally met  urlLink mr. coolfer  himself! i was so excited when i found out his blog name was taken from slang used by his friends, as in when your friends eat half of your meal, you say, \"coolfer eating my meal, dickhead.\" hotness!   another slang term to add is my friend katy\'s new one. she was walking into our apartment building one day and the kids outside pointed to her and said, \"that\'s my type. smell it.\" ahahahahaha!   '),(15148,1694057,'18,June,2004','2004-06-18 00:00:00','   hooplas involving circus tricks  apparently rachel trachtenburg couldn\'t make the show last night at north six because she was being interviewed by vh1 or something. after dop practice, we all went over to say hi to eric and the  urlLink say hi to your mom  gang. we thought we\'d slip in just in time for shtym\'s set, but unfortunately that wasn\'t the case. we suffered through the last half of  jason trachtenburg \'s solo show... the absence of rachel\'s cute face was thick in the air, as we all tried to patiently listen to jason mutter out long, drawn out, seemingly directionless stories... he seemed to nervous, i felt so bad for him... but it also pained me to listen to it. so much so that i couldn\'t help but laugh out of nervousness for him. oh.my.gosh. poor guy.  then some kind of funk band whose bassist had this blue light up flashing ring came on stage... and we all knew we had to leave. pizza run. we ran into billy from other passengers there, chomping away on a slice. this town feels more and more like family as time goes by.   we eventually said bye to billy and wandered back to north six. eric finally went on stage super late, but we enjoyed the set nonetheless. i was barely conscious because of my lack of sleep these days. but hearing eric sing \"super\" is nice and reassuring that someone else in the world holds the same sarcastic sense of humor that i do. word.    back to band practice though. i don\'t know why, but i am addicted to going to band practice. i think that i miss dop when i don\'t see them at practice for like a week. having such access to their music and their creative process is an amazing privilege that i take advantage of as much as possible. i see them develop. i see them grow. i see them create amazing songs. i feel like a mother, proud of her children\'s accomplishments. the beautiful sounds start to envelope me, and i can\'t help but smile. it\'s so fucking good! i took some video of them practicing dj\'s new song, which joe sings on. his voice is so sad and beautiful. the song is so goddamn catchy.   urlLink watch them here and see for yourself!   [warning: it will take you a while to download, even on high speed, but it\'s worth it.]   '),(15149,1694057,'17,June,2004','2004-06-17 00:00:00','   everyone wants a drummer  The following is a letter from Max from  urlLink the Natural History . It seems everyone is looking for a band member who likes to tour. Ain\'t that a bitch???? If only I was a bad-ass drummer. I\'d be in bidness! Help him out, peoples.  Hey Everybody, The Natural History is looking for a new drummer. Someone who likes to tour a lot and would want to be in a full time band. Well, at the very least would want to tour this fall. We\'re in the middle of recording our new record which will be done at the end of summer (our current drummer will be finishing up the record) and released early next year. f you know of anyone in the NYC area who might know someone, please forward this to them! If you someone who has always wanted to move to NYC and be a drummer in a rock band, please forward this to them! our help is appreciated. Thanks  urlLink Max   '),(15150,1694057,'16,June,2004','2004-06-16 00:00:00','   surrender, surrender! rock and roll still exists in nyc    it\'s been quite a while since i\'ve been to a rock show and had a beer bottle thrown at me, but it happened last night at  the blue sparks cd release party  by miss alanna. it came out of nowhere and splattered all over near the bar of sin-e. craziness! all in good rock/n/roll fun! it\'s always a good night when you spend some time at home washing the beer from your face, and the ink from your hands.  in case you don\'t know,  urlLink the blue sparks  are down and gritty rock and roll that splatters (like a hurled budweiser) with tough spastic fun. it\'s pretty amazing how tight they are, and i especially love when phil rocks it out on the guitar. he seems so innocuous, but that boy just rips out the demon when he\'s on stage. in an adorable way, of course! they just pound their guitars, and i don\'t know how their strings aren\'t all in pieces right now.   i really liked their entire set, but the old song they pulled out at the begging of john martin was pretty amazing. jo tells me it\'s an old blue sparks song, but a classic, called \"tell me whatcha want.\" that song is so good, i thought it was a cover song! it\'s the kind of old school rock song that makes you want to go crazy and scream along, \"tell me whatcha want! i want i want i want! tell me whatcha need! i need i need i need!\" sooooooo fucking good. i\'m officially an addict.     wolfbait  (which of course was  urlLink the fever  playing cover songs) played some goodies too, including: ** hall and oates\' \"maneater\" ** r. kelly\'s \"ignition\" ** cheap trick\'s \"surrender\"  i\'m also happy to report that  urlLink geremy \'s hair is lookin\' better than ever. word.  this was all  urlLink witnessed  by a lovely group of  urlLink sweet   urlLink hot   urlLink familiar   urlLink peeps . so much raw goodness. i fucking heart new york.    another great band i saw last night was  urlLink other passengers . this band just gets better and better each time i see them. their dark, despondent lyrics and overwhelming compositions swallow me up into some kind of abysmal black hole. just amazing, and quite heart-stopping live.   '),(15151,1694057,'14,June,2004','2004-06-14 00:00:00','   free french kicks show   urlLink     FREE  urlLink FRENCH KICKS  SHOW  Bloomingdale\'s Soho Brought to you by Puma & Blender Thursday, June 24 6:00 - 8:00pm  urlLink rsvp@blender.com   Did I mention the free drinks? Yes, free cocktails. Yum!   '),(15152,1694057,'14,June,2004','2004-06-14 00:00:00','   new slang (for me) some people i\'ve recently met have been saying some funny slang that is new to me. i thought i\'d share them so that you could laugh with me:  mika, christy\'s gf, is originally from japan. when we were all hanging out at snacky\'s, she suddenly turned to dj and said, \"why you so dick?\"  my co-worker jeremy (you may see him spreading around ascap earplugs at the vicious events) has a couple great sayings. if he wants to ask you if you\'re interested in certain things, he\'ll say, \"do you fuck with that?\" and when he wants to communicate that he\'s tired, he will say, \"i\'m so assed out.\"  when my friend mike likes something he\'ll either say, \"hot shit!\" or \"that\'s the jam!\"   so anyhoo, what kind of funny slang phrases do your friends sling around?   '),(15153,1694057,'11,June,2004','2004-06-11 00:00:00','   just as an fyi  urlLink    the walkmen  and  urlLink dresden dolls  show at  urlLink volume  has been canceled. volume is still not re-opened yet...   save the dates! there are some great upcoming shows this week, and you don\'t want to forget to attend! of course, these are all me-related events or my friends\' so i\'m sorta biased. but you can still trust my taste in music, eh?   ** 6/15:   urlLink other passengers  @  urlLink vicious  -  urlLink rothko . i haven\'t seen the other peeps on the lineup yet, but i\'m sure they are real and spectacular.   ** 6/15:   urlLink blue sparks  cd release party @  urlLink sin-e  (w/ special HOT guests) and  urlLink jo  is deejaying. wheee!   ** 6/16:   urlLink circle and square  @  urlLink lit    ** 6/18:   urlLink bishop allen  +  urlLink adam green  @  urlLink mercury lounge    ** 6/18:   urlLink dirty on purpose  @  urlLink dead city disco party  -  urlLink loop lounge  (NJ)   ** 6/19:   urlLink dirty on purpose  @  urlLink karibomb \'s birthday party -  urlLink mercury lounge  ... also with the bravery, benzos, the sons of sound!   or if you don\'t like music...  Helen Fielding Reading   ** 6/15:  The  Bridget Jones  author reads from her new spy thriller for the chick-lit set, Olivia Joules and the Overactive Imagination. 7 p.m., Barnes & Noble, 33 E. 17th St.; 212-253-0810.   PS. I heart  urlLink Nerds .  PPS.  urlLink French Kicks play \"One More Time\" on Fearless Music TV.  le sigh. love those boys.   '),(15154,1694057,'11,June,2004','2004-06-11 00:00:00','   booking a tour!  i\'m currently working on booking a tour for  urlLink dirty on purpose . we have booked dates in new york, DC, pittsburgh and philly so far. we still need to fill the following dates: august 1, 2, 5, 6   we want to travel to places like baltimore, detroit, chicago, cleveland, chapel hill (carrboro), richmond, providence, portsmouth, burlington, saratoga, boston, portland, and anywhere else along the way! if you have contacts or are in a band in those areas and can hook us up with a show, please  urlLink email me !  check out samples from  dirty on purpose \'s debut ep  urlLink here .   '),(15155,1694057,'10,June,2004','2004-06-10 00:00:00','   it\'s friday! wheee!    hey kids. guess what? it\'s friday! woooohoooo! here are a couple places to get some good music going on to, uh, get you in that friday mood:   urlLink  radio indie pop     urlLink  kexp     urlLink  cherry blossom clinic with terre t     urlLink  jinners radio     urlLink  damian on riot radio... i think around 2 p.m. EST?   '),(15156,1694057,'10,June,2004','2004-06-10 00:00:00','   a shady lane, everybody wants one  this past weekend was not only filled with lots of weary moments, but my housemates and i watched two documentaries:  spellbound  and  capturing the friedmans .    spellbound  was entertaining, but i thought it would be more intense. the kids in it were so cute, and some of those words were damn tricky. the sweat pouring down their foreheads, the biting of the lips, the utter fear of blanking out cringing out of their eyes.... we always knew that vocab tests in school were always so damn nerve-racking, but imagine taking that test in front of cameras and hundreds of parents. yikes! the one lesson i learned from this movie:  bee happy.    capturing the friedmans ... i really didn\'t know what i was in for when i decided i wanted to watch this film. and to think that it all started because the filmmaker was profiling new york\'s #1 clown... who turned out to have some not-so-funny family drama going on. as the story unfolded, i became utterly horrified. then i was confused. then i was like, wtf is wrong with these people? hello, your father is a child molester.  geez us!   in the unbearable heat, there is one song that always makes me happy. i love pavement\'s \"shady lane\". sing along with me:  Blind date with the chancer we had oysters and dry lancers when the check when it arrived we went dutch, dutch, dutch, dutch a redder shade of neck on a whiter shade of trash And this emory board is giving me a rash I\'m flat out You\'re so beautiful to look at when you cry Freeze, don\'t move You\'ve been chosen as an extra in the movie adaptation of the sequel to your life.  A shady lane -- everybody wants one A shady lane -- everybody needs one Oh my god, oh my god, oh my god, oh my god Oh my god, oh your god, oh his god, oh her god It\'s everybody\'s god, it\'s everybody\'s god, it\'s everybody\'s god, it\'s everybody\'s god The worlds collide, but all that we want is a shady lane  Glance, don\'t stare Soon you\'re being told to recognize your heirs No, not me -- I\'m an island of such great complexity Stress surrounds in the muddy peaceful center of this town Tell me off in the hotel lobby right in front of all the bellboys and the over-friendly concierge  A shady lane -- everybody wants one A shady lane -- everybody needs one Oh my god, oh my god, oh my god, oh my god Oh my god, oh your god, oh his god, oh her god It\'s everybody\'s god, it\'s everybody\'s god, it\'s everybody\'s god, it\'s everybody\'s god The worlds collide, but all that I want is a shady lane.   '),(15157,1694057,'09,June,2004','2004-06-09 00:00:00','   calling all book worms!  i wish there was a book club called \"great books for people who love music.\" i would totally join that club.    the last book i finished reading was called  urlLink  drinking: a love story   by caroline knapp. it was one of the most depressing books i\'ve had to page through, and as i was reading it, i felt like i was also going through many of the trials and tribulations with drinking recounted in the book. the book is actually caroline\'s memoir on her bout with alcoholism. now i know we all like to consider ourselves alcoholics, but caroline went through some major drinking episodes that make mine seem like some sort of pathetic joke. she would drink before going to work, right after work. hide bottles in back porches. drink entire bottles of wine and replacing them so that no one would know... she drank to deal with pain. she drank to not feel anything. she drank just for the sake of drinking. it was like a nervous habit. some of us bite our nails. caroline fucking drank until she would pass out and not remember anything. reading this book made me realize that i was definitely not an alcoholic, though my love affair is quite coquettish.  now i\'m thinking of picking up caroline\'s other book,  urlLink  appetite: why women want  . why? because i wanna find out why.   i\'m also trying to get through  the corrections . but sometimes it\'s so fucking boring. granted it certainly has moments of brilliance, but then i get to another chapter and the story goes from somewhat interesting to just plain dull. i was quite attracted to the story however. it\'s about a family from pennsylvania that is totally dysfunctional. the absolute struggle to get family together, even at the most stressed times. the basic outline plot reminded me of my own family and how disparate everyone is. i think much of my frustrations as a person is partially due to the fact that i don\'t have much of a family structure to fall back on. i don\'t have the close bonds. i don\'t often feel the compulsion to call my familly. not to say that i don\'t love them, but some of them drive me up the wall. they are quick to criticize me, and i often think my somewhat chronic bouts with depression are because of all the fucked up ideas they have instilled in my head about myself. i was never skinny enough. sometimes not smart enough, though most of the time my academic life was the only thing they ever showed approval of. i was never girly enough. blah blah blah.  the corrections  covers all of that fucked up nonsense about families and their ability to both support and tear apart each other. hopefully i will plow through this soon and move on to another book... but i\'m kind of infamous for reading multiple books at once. judging by my past posts, maybe i should read jonathan franzen\'s  how to be alone . ha.  any other recommendations?   '),(15158,1694057,'08,June,2004','2004-06-08 00:00:00','   to lost souls everywhere  i feel like i\'m perpetually caught in limbo. i feel inexplicably sad. i feel lonely. this is partly due to my recent singleton status, i know. when you\'re in a couple, you don\'t mind other couples so much. you don\'t mind the fact that all of your other single friends constantly hook up with other single friends or make out with random strangers, whatever. you don\'t mind being on your own every once in a while. then you read things like  urlLink this , and it makes you kind of sad. girls are fucked up. boys are fucked up. and together, they manage to fuck each other up real good. what the hell?  in my new single life, i\'m finding that being alone sucks. and what\'s even more interesting is that in my new digs, i find myself living with a married couple, who are nice peeps and all, but hell, they are newly married. they are a unit. it\'s sweet, but weird for me.   my new bedroom is what i call \"the bat cave.\" i climb up these creaky, wooden stairs to get to my bat cave. at night when the door is closed, i am swallowed in darkness. and i pass out and dream strange dreams, most of which i can\'t recall anymore. this past weekend, i found myself in a terrible mood, and i just couldn\'t get out of bed. i slept all day saturday, then finally ventured out of my bedroom around 7 p.m. granted i went out to support au revoir simone at trash and bishop allen at the tank, but then the next day, i walked around like a zombie yet again. i felt incredibly woozy and unwell. my head was pounding. i was nauseous. i passed out around 7:30 p.m. and didn\'t awaken again until well past noon the next day.  i was suddenly sick.  steadily since then i\'ve been feeling better, and i appreciate everyone who has shown concern. i\'m alright. it\'s kind of amazing how powerful our minds are. if we are unhappy, then we sort of react physically to our emotions. i feel sort of lost right now. it\'s probably the norm for all people in their 20\'s but in this town, who has time to fuck around like this? not i. and so i solider on, trying to ignore everything that makes life so miserable for us young ones. in the grand scheme, things like that don\'t matter much. and like i said, i\'m trying not to be emo. i\'m trying to look on the brighter side of things. i\'m trying to be optimistic and avoid cynicism.   '),(15159,1694057,'08,June,2004','2004-06-08 00:00:00','   snap and clap!  i gots a photo gallery published on cameron\'s fun review site,  urlLink snap and clap . he reviews  urlLink dirty on purpose \'s performance at north six... cameron is zee best!   '),(15160,1694057,'08,June,2004','2004-06-08 00:00:00','   midnight cravings  you know when the clock is nearing midnight, and then a friend calls you and asks you to go for a drink? yeah, that\'s mike. he was done band practice and was in my hood. i went to meet him at snacky\'s. i later found out he thought i said spanky\'s. teehee.   he\'s one of those dudes who are really easy to talk to and totally sweet and honest about everything. he\'s the first guy friend i\'ve had in a while who i felt like i could just talk about everything with. he\'s the best fucking friend! i haven\'t seen him since i moved out of the east village. so it was good to see him in my hood! i heard that he got into an accident, and it was totally freaking me out! i had not heard from him for about a week, and already something crazy, life-changing happens! time is fucking crazy like that sometimes.   we had so much fun just talking, drinking sake and beer, eating korean bbq beef and peanut chili wontons and watching the kung fu movies on the tv. snacky\'s is the best! thanks for the good talk, mike. we gal pals certainly appreciate.   '),(15161,1694057,'08,June,2004','2004-06-08 00:00:00','   y\'all got jacked!   urlLink this song, \"jackin\'\",  is cracking me up. it\'s by some group called rhymefest. what it is, boy!   '),(15162,1694057,'07,June,2004','2004-06-07 00:00:00','   knife skills have a hot new site  knife skills, one of my favorite local hard-asses, have  urlLink brand new site ! Check out the buzz and those tough-as-hell mp3s. whoa.   '),(15163,1694057,'07,June,2004','2004-06-07 00:00:00','   Favorite Songs About Love  I did a search on my iPod for songs with the word \"love\" in it. Although I must admit that my iPod collection is nowhere near complete, I thought I would share the top 20 songs that came up, based on a random shuffle.  What\'s your favorite love song?   1) \"Feeling Called Love\" - Pulp 2) \"Tropics of Love\" - Black Heart Procession 3) \"The Love I\'m Searching For\" - The Rentals 4) \"Is This Love?\" - Cake 5) \"Don\'t Love You\" - TV on the Radio 6) \"Ten Story Love Song\" - The Stone Roses 7) \"House of Jealous Lovers\" - The Rapture 8) \"I Believe in a Thing Called Love\" - The Darkness 9) \"Last Night I Dreamt That Somebody Loved Me\" - The Smiths 10) \"I\'m in the Mood For Love\" - Louis Armstrong 11) \"Inside of Love\" - Nada Surf 12) \"To Your Love\" - Fiona Apple 13) \"Lovefool\" - The Cardigans 14) \"Love\" - The Sundays 15) \"I\'m Still In Love With You\" - Swearing At Motorists 16) \"Loved Despite of Great Faults\" - Blonde Redhead 17) \"Goodbye to Love\" - American Music Club (Carpenters cover) 18) \"Between Love & Hate\" - The Strokes 19) \"Be Be Your Love\" - Rachael Yamagata 20) \"Love Cats\" - The Cure   '),(15164,1694057,'07,June,2004','2004-06-07 00:00:00','   dearest peeps  i\'m sorry i\'ve been such a fucking downer on my site lately. i\'m really not that emo. and frankly, i\'m really fucking tired of being sad and depressed, and just want to begin enjoying life again. even though i haven\'t been feeling well this past week, i\'ll try to always greet you with a smile and not a sigh.  xoxo, jinny jin jin   '),(15165,1694057,'07,June,2004','2004-06-07 00:00:00','   kitty magik\'s new website!   urlLink kitty magik  launches a swank new redesigned site that is all hotness. and guess who is under their contributing photographers? woo hoo! it\'s kind of funny since i\'ve done mostly writing for them, but i\'m about to submit my blonde redhead photos, and hopefully a photo gallery will follow.   '),(15166,1694057,'05,June,2004','2004-06-05 00:00:00','   the slow century  i feel like these days, i\'m totally into slowing it down. for the past couple of years, i\'ve lived fast. heck, for the past 7 years in new york, i\'ve lived fast. But night after night of hanging out at rock clubs with earplugs and the search for free liquor can be quite taxing. i\'ve never been one to limit my experiences like that, but these days nothing compels me too much. except for things that are free because i\'ve become quite broke.   going to a rock show? probably not unless i\'m on the list. this makes me sad as i desperately want to attend lollapalooza and see moz and the pixies. i\'m talking,  a lot . but not only would i have to take two days off of work, i\'d also have to shell out like $90 on tickets and probably another however much on transportation. rock and roll is like my drug and all, but seriously, this habit can get quite expensive on a big scale. perhaps that\'s why i try to keep most of my show experiences on the smallish, local level.   with all of the work that i do end up doing in the music industry, sometimes i feel like i don\'t get enough back to sustain the lifestyle. i envy all of the rich kids who have the support of their parents and can afford to do whatever their little hearts desire. i, on the other hand, need to balance everything out with a full-time job. it\'s a tough thing to juggle in this twisted economy. but c\'est la vie! c\'est la rock/n/roll!   i\'ve also always thought that i wanted to be a writer, and a part of me still does. but things can get discouraging in this day and age. i feel like many others get the chance to do things that i can only dream of doing... for one reason or another. mostly because my tastes are not very mainstream. that was one thing i learned from interning at  entertainment weekly . i despise mainstream stuff most of the time. it\'s too manufactured and not raw enough for me. thus my life is not mainstream, and the opportunities that present themselves to me are all lowly funded indie projects. but the awesome thing is that i believe in everything i do, and i don\'t think that i get swayed into writing and believing things that are simply trendy or buzzworthy. anything that i support really does have some weight (in my opinion). it\'s not all dressed up in purdy bows and ties. it\'s real. it\'s meaty. it\'s something precious to me. it\'s not a paid-for-commercial. i hope you get that sense of me.   '),(15167,1694057,'04,June,2004','2004-06-04 00:00:00','   tangmonkey knows where it\'s at  here\'s a cool review from sean of  urlLink tangmonkey.com . thanks for the kind words, dude!  Dirty on Purpose - \"Mind Blindness\". Vibrant shoegaze pop from a NYC band that knows to blend haze with clarity, summer somnolescence with a young autumn\'s skip-run-and-fly. (cf. my jeans feel a bit tight, think i washed them a bit too high. i was gonna be late, so i picked up my pace to run.) There are sighs and interlaced vocals, dream guitars, a touch of Yo La Tengo earnesty, but then the brilliant bristling sprint of Doves or \"Where The Streets Have No Name\". Lots of bands strain to reach this kind of place - Torrez, Tiger Saw, heck, Mercury Rev, - but here\'s a group that not only has the mood, it\'s got the soaring, spaceship hook. Marvellous. (To top it off - I\'m told they like Os Mutantes.) From the sleekly-packaged  Sleep Late For A Better Tomorrow  EP.   '),(15168,1694057,'03,June,2004','2004-06-03 00:00:00','   stop fucking around things have been quite hectic for me this week, in more ways that one and for all of my friends, i dearly apologize. sometimes i get so friggin\' busy with work, music, etc that my head explodes and i become comatose. this week has been one of those weeks where i think my mind just went blank. but i can\'t swear off rock/n/roll like  urlLink some people can . i feel like i physically and emotionally need music to feel like myself. it\'s worse than crack.  i still feel really out of it -- and still feel like a nomad with all my shitty moving around. it\'s been sort of an adventure, but right now my life doesn\'t need more drama. today\'s culture is going more and more for minimalism. take the impact of the film  super size me . also consider ordering a small coffee for 60 cents with one less lump of sugar, and maybe no cream at all. frequenting bars only once in a while. cooking more at home and eating half as much. twice the leftovers. mini skirts. shoes that don\'t have a back on them, i think they\'re called mules. jeans that expose more than they should because they are ultra-ultra-low-rise. this is steadily becoming the norm. i don\'t really remember what my point was... see, what i mean? even our thoughts are downsized... or rather, shorter-lived.   the winter pageant @ rothko the other night i went to see mike\'s band,  urlLink  the winter pageant  . they were so good! there is this incredibly sad, delicate haze that surrounds their entrancing music, and listeners beware, if you are depressed, the winter pageant will only induce more tears. and perhaps, even if you are happy, you will walk out of their show wholly depressed. think of a lost love. think of loneliness. think of disappointments. think of mean girls (not the movie). think of everytime someone ever betrayed you or kicked you in the shins or made fun of you. every time your parents criticized your life, your looks, your low salary. you get the point. incredibly atmospheric (especially when mike breaks out the e-bow) the winter pageant have the same mood as elliott smith songs but they come across as a bit more emotional, more swirly. a lot less bitter. and yes, equally as beautiful.    need new body @ knitting factory   in contrast, the band i saw last night thanks to some list action from  urlLink dt and rt , was energized, ridiculous, sarcastic and spastic --  urlLink  need new body  . HOLY SHIT. i had no idea what i was in for, but i got a clue when rt was like, \"you\'ve never seen need new body? oh man. i guarantee this will end up on your blog.\" and so it is! i think i spent the entire show squealing in delightful shock and amusement.    re-enactment:   aha! oh.my.gosh. aha! wtf?!? oh.my.god.   the band started off doing a few ditties that sounded like a re-interpretation of the theme from  all in the family . dude, i\'m totally serious. it was obviously some kind of joke (and admittedly funny in a freaky sort of way that someone\'s voice could be so friggin\' high normally -- think geddy lee or joe jurewicz), but then the band suddenly broke into their real set list, which was amazingly dancey and funky, with punk ideals. and whenever i hang with drummers like dt, i am always made to watch the drummer on stage. and i\'m glad i did -- that dude was really fucking crazy. his eyes were bugging out and he was thrashing those drums like he was some insane asylum resident gone hogwild. it was kind of scary actually. the main lead singer had a very redneck look to him, and he kind of reminded me of a  urlLink tim harrington , but with less theatrics and more emphasis on just pure dancing.  also, thank god,  creed  finally broke up. woo hoo!   '),(15169,1694057,'03,June,2004','2004-06-03 00:00:00','   good news for people who like bad news -- that\'s me!   is it just me or is there an air of  pixies  power on the new  urlLink  modest mouse   album? i can\'t stop listening to it. i love the jubilant squeals bursting forth from singer isaac brock. i love the high-pitched, swooning guitar parts. it makes me inexplicably happy, even during dark times. this will definitely be one of those albums i throw in the cd player whenever i\'m feeling down and need an instant pick me up. sigh.    favorite lyric:  \"we were done done done with all the fu-fu-fucking around!\" -- from \"black cadillacs\"   '),(15170,1694057,'02,June,2004','2004-06-02 00:00:00','   that joke isn\'t funny anymore  Being in New York is definitely a privilege. Not only is it expensive to live here, it is one of the most competitive cities in the world. And at times when I think that everyone seems to have lost a sense of community, something comes along to prove me wrong.   Struggling to recover from Thursday night\'s drunken Makers & Ginger debauchery, I spent most of this weekend kitten sitting for my boss\'s friend who lives in Nolita. Walking out of that neighborhood was quite refreshing, the warmth of the sun and the myriad of shoppers buzzing along the sparkling sidewalks, happy because they don\'t have to go to work the next day. I was one of those shiny happy people.   I walked from Nolita to Alphabet City, where I was moving the very last of my stuff out of my two-month sublet. I was kind of sad to be leaving what I think is the heart of the city for me. I spent more time around Avenue A and B in the past two months than in the entire time I\'ve been in the city, which is coming up on 7 or 8 years. I will miss the convenience of crossing through Thompkins Square Park and visiting Mike at Alt Cafe. Or heading over a block to some listening party at Hi-Fi. Or breaking down at 3:30 in the morning and eating San Loco tacos (evil!) Or meeting peeps for drunken fun at B-Side or 2A and walking home afterwards. That is the shizznit.   Now I am back in Williamsburg, a community forever labeled by rock mags everywhere as the birthplace of electroclash and New York\'s burgeoning indie rock scene. And I have to say, I have missed Wburg a lot. Many of my friends live in the \'burg and steadily I have missed that slower-paced community of friends. The city is immediate and swift. The borough is lounge-worthy and calming in a strange way.   '),(15171,1694057,'01,June,2004','2004-06-01 00:00:00','   two cute bands, one great show   urlLink    we are scientists! i was immediately attracted to their geeky name. i heard about them through my friendship with another awesome brooklyn via boston via texas band, bishop allen (pictured below). then someone at work gave me their cd and i was hooked. dude, i\'m so excited to check them out this saturday! @  urlLink the tank ! times square! be there or... well, you know.   urlLink    bishop allen! one of the most charming and unpretentious bands in new york today, these cats are solid peeps and solid songwriters. I am forever in love with their song,  urlLink \"Quarter to Three.\"  Favorite lyric: \"Now let me tell you I\'ve been up since a quarter to three / And I\'ve been pacing back and forth through the hall / I\'ve been thinking \'bout the first time she took my hand / And I don\'t understand it at all\".    UPDATE: By the way, Bishop Allen\'s \"Things Are What You Make of Them\" was used in the movie,  Saved! . Who wants to go see it with me?   '),(15172,1694057,'13,July,2004','2004-07-13 00:00:00','   the big sleep + helio sequence tonight!   hey kids, my peeps in  urlLink the big sleep  are playing with another really awesome band called the helio sequence, who are on subpop have have the old/new modest modest drummer in the band. i\'ve been listening to loads of both the big sleep and  urlLink the helio sequence  and have siked myself up for this incredible show tonight. let\'s get wasted at  urlLink north six  tonight. show starts at 8 p.m. and also has comedian  urlLink eugene mirman  making us laugh our skivvies off. wha?   '),(15173,1694057,'12,July,2004','2004-07-12 00:00:00','   new earlimart tracks   **  urlLink heaven adores you  **  urlLink the hidden track    new allmusic.com whoa,  urlLink look , it\'s all pretty and shit. wow.   i can\'t stop listening to...  urlLink     urlLink cocorosie   '),(15174,1694057,'12,July,2004','2004-07-12 00:00:00','   so wha happened?  ok i\'ve been getting a lot of questions about what happened on saturday. well,  dirty on purpose  didn\'t back out of the show as people were being told even on saturday evening... we did get a phone call from  the new year  telling us they were sorry about what happened -- apparently knitting factory overbooked the night, including some kind of weird party before the new year show, and one band had to get dropped from the bill. because of the party, the new year didn\'t even get a soundcheck, apparently, which is totally bogus. fucking bedhead! of course it sucks that all of this had to happen 2 days before the show... hopefully everyone was able to get a refund if you were seeking one. thanks to everyone for their support! the next dop show will be on july 30 at office ops for superfun rock/n/rollerskating fun! bring your skates or rent one! au revoir simone and circle and square will also be playing. it\'s the start of the dop tour, so come and celebrate with us! word!   '),(15175,1694057,'09,July,2004','2004-07-09 00:00:00','   party on friday  from ryan the girl: George Bush doesnt want you to read the 9-11 commission report.... Dick Cheney doesn\'t want you to be allowed to look into the people, agenda, or notes from his secret/closed energy commission meetings.... Donald Rumsfeld doesn\'t want you to look closely @ the pics of him laughing it up w/ Saddam... & John Ashcroft really doesn\'t want you to have the right to privacy, to choose or to dance...   BUT DANCE WE WILL!!!!!!!!   in honor of the cancers @ one-forty (FYI: sam, megan, lida, & ryan), there will be a fuckin dance party this friday night in williamsburg. you have been officially summoned to shake your ass off @ the   BIRTHDAY BOOGIE AGAINST BUSH!!!!!   (it\'s gonna be a long summer preparing for the RNC and the election...but this night will be all about celebrating......so get your dancing shoes on!!!) the mojitos will start early, the dancing will go late! 140 Metropolitan Ave...1st floor (Between Berry & Whythe)   And please......NO PRESENTS! take the $$ you woulda spent on a cd or bottle of vino and throw it into the collection jar. the moola raised will go to Act For America, a grassroots group that is registering voters in swing states and the shepard fairy anti-war poster campaign   '),(15176,1694057,'09,July,2004','2004-07-09 00:00:00','   dirty on purpose interview on crashin\'in     urlLink check it out, lio interviewed dop.  joe is such a goofball. word.   '),(15177,1694057,'09,July,2004','2004-07-09 00:00:00','   my little danish friend  a bunch of us went to see  metallica: some kind of monster  last night... it was one of the funniest movies i\'ve seen in a while. probably funnier than anchorman. my favorite line was when dave mustaine was in a therapy session with lars...   dave: do you know what i went through? lars: i can imagine what i put you through. dave: you\'re not hearing me. do you know what i WENT through? lars: [rubs his face] dave: i miss my little danish friend.  or the exchange was something like that. seeing a band dissected like that was crazy. it was crazy when they were like, rob, please be our bassist -- here\'s a million dollars in advance! shit!! it was crazy to know their therapist was earning more a month than i earn in a year. beejezuz.   '),(15178,1694057,'08,July,2004','2004-07-08 00:00:00','   Other Great Saturday Happenings  Let\'s keep it on the positive this Saturday and support some amazing local acts like:  @ Mercury Lounge Other Passengers, Izzys  @ North Six Blood Brothers, Kill Me Tomorrow  @ Sin-e Grand Mal, The IOs  @ Rothko Chinese Stars, Mommy & Daddy, Mahi Mahi   '),(15179,1694057,'08,July,2004','2004-07-08 00:00:00','   cliktrax goes online with fuse!   urlLink now you can watch cliktrax online!  This week\'s featured artist on the site is Franz Ferdinand. Music videos by The Thermals, The Walkmen, The Streets, and TVOTR. Holla.   '),(15180,1694057,'08,July,2004','2004-07-08 00:00:00','   hello, we did not cancel!    no matter what you hear about saturday,  urlLink dirty on purpose  did not cancel the show for saturday... why would we do that? the new year are one of our heroes. anyhoo, we\'re not sure what is going on right now, but we\'re no longer on the bill for the new year show on saturday at knitting factory. we\'ll keep you posted on how you can get refunds if you already bought tickets for the show. check back here or  urlLink dirtyonpurpose.com  for details as they unfold.   UPDATE:  Knitting Factory will be sending out an email through Ticket Web to everyone who bought tickets, instructing people on how to get a refund if they wish to do so. Also if you bought tickets at the KF Box Office, you can go get your refund there too.   '),(15181,1694057,'07,July,2004','2004-07-07 00:00:00','   things i love right now  ** listening to internet radio ** riding my bike to greenpoint ** iced lattes ** drinking red stripe ** making care packages for out-of-town friends ** demeter perfume in gingerale ** epitonic ** shootin\' the shit with friends ** reading emails from around the world (hello vanessa from spain!) ** happily ever afters   '),(15182,1694057,'07,July,2004','2004-07-07 00:00:00','   radio experiment  a few weeks ago i sent the  urlLink dirty on purpose  cd to  urlLink kexp.org , one of my favorite internet radio stations. so help me with a little experiment today...  urlLink email kexp.org  and request a dirty on purpose song! if you don\'t know which song to request, try \"all new friends\" or \"mind blindness\". let\'s see if they really play requests! merci!   '),(15183,1694057,'07,July,2004','2004-07-07 00:00:00','   dirty on purpose with the new year   urlLink    hmmm... here\'s a little self-promotion. dude, it\'s my blog. it\'s all about self-pimpin\'.   Hey Peeps,   urlLink Dirty on Purpose  is playing on Saturday, July 10, with the New Year (x Bedhead), VietNam and Mazarin at NYC\'s Knitting Factory (79 Leonard St).  We would love to see your purdy lil faces in the audience, front and center, whoopin\' and hollerin\' and yelling \"We Love DoP!\" ever so raucously. Come celebrate this awesome bill with us.  Show starts at 10 p.m. and it\'s $12. We know, it\'s a lot of money, but you probably spent about that much to see a movie this past weekend, and we think this show will certainly entertain you more than  Spider Man 2 .   urlLink We think this show is gonna sell out... so buy tickets ahead of time at knittingfactory.com.   '),(15184,1694057,'05,July,2004','2004-07-05 00:00:00','   i love you golden blue  this past weekend was like being in some sort of weird comatose state. or drifting in a deep pool of water... scissor-kicking my way to nowhere in particular. it\'s like listening to sonic youth\'s \"i love you golden blue\" on repeat whle the sun warms my cheeks to a nice shade of rose.  on saturday night i hung out with anna, george, katy, stephanie of va, greg and ryan the girl. some of us checked out vaz at north six. what a fucking awesome show, including the end where paul knocked the microphone stand off the stage and mimed his thank you. vaz is heavy rock that foams at the mouth but over the last couple years, they have developed really catchy guitar lines that swerve around like race cars. how can a drummer and guitarist make such a full sound? i dunno, but that drummer is fucking outta his mind. he looks so possessed. blink and you\'ll miss way too much.  then we headed over to ryan\'s apartment... which is crazy. they have a swing and hammock. they usually leave their double doors open and people riding their bikes by think it\'s some sort of tropical bar. but no, it\'s a fucking apartment. so stop walking in! while we were lushing on puerto rican rum, a couple people just walked in. one of ryan\'s roommates were like, \"who are you?\" and they were like, \"isn\'t this a party?\" hahahahaha.  i had been wanting to ride my bike all weekend. finally got to a little bit on fourth of july. after making peach daiquiris with katy and stephanie of va, i rode my bike to greenpoint to meet up with george. george, joe and i ended up watching the sonic youth dvd that george bought. i still can\'t get over the video with macaulay culkin in it. disturbing to watch thurston moore bobbing heads and playing guitar with the red-lipped wonder.  then george and i headed to the rooftop party. we ran into christy and eventually greg and joe graced us with their presence. i got kinda smashed on makers and gingers. we watched the fireworks in awe. smiley faces with whiskers! we danced around with wimpy sparklers. we ate the bbq food that christy brought. then i was so tired i went home early and passed out. how was your J4 celebration?  then i re-read the  griffin & sabine  trilogy, which chris helped me locate. that story is so romantic. love those books!   '),(15185,1694057,'05,July,2004','2004-07-05 00:00:00','   the retropod... it\'s waterproof too     urlLink apparently this is no joke  [from my coworker jon]   '),(15186,1694057,'05,July,2004','2004-07-05 00:00:00','   dj annabiotica guests on wbar with the notorious b.i.d.\'s electric avenue!     urlLink here\'s the wacky playlist    when i figure our where the archived stream is, i\'ll post it here!   UPDATE:  here is a  urlLink link  to the stream of the show... (nota bene: the mics in the studio weren\'t working right so if you don\'t hear anything between songs, just imagine bidi and anna goofing it up and screaming things like \"who rocks the party that rocks the body?\" the link will only be live for about a week.  ps. we did not play the strokes. that was bidi. har har har!! (ok, she says it was a request. fair nuff!)   '),(15187,1694057,'03,July,2004','2004-07-03 00:00:00','   happy fourth of july!   urlLink click here for some fireworks!   '),(15188,1694057,'02,July,2004','2004-07-02 00:00:00','   your wish is my command    ok peeps. sorry i\'ve been lax on the updates this week. indeed, it is true. i\'ve been assed out. it\'s been a rough one! but now i\'m back on track and want to share some fun stuff with you... if you\'re still into it.    i spent much of last night at band practice, preparing for the impending  dirty on purpose  tour coming up... if anyone can help us fill dates on august 5 or 6 (we currently have some leads in boston, rochester, providence and possibly chicago),  urlLink let me know . word! george\'s brother, greg (pictured above), came to practice to let the music just soak in.   i also went to this new williamsburg bar after practice to drinky drink with dj. the bar is called  funhouse  and it was pumping some kind of weird mix of music -- anywhere from !!! to bad techno music. but the british bartender was nice and mixed the makers and ginger strong. and they showed these projections on the white walls -- and guess what? there was a whole segment with penguins! fyi to people not in the know, but i fucking adore penguins. jo told me that when they cry, they can turn their salty tears into fresh water. or something like that. plus, i heard they mate for life. plus, they\'re damn cute, so...  tonight i think i\'m hitting a couple birthday parties... and then i\'m not sure what i want to do for fourth of july. there are several rooftop parties that i\'ve been graciously invited to... but it\'s funny when you think about how many of my friends don\'t peep together. so it\'s probably impossible for me to see all my lovely friends. boo!  in other news, noah and dt were remiss in informing me that  modest mouse  and  the walkmen  are touring the west coast together. for once, i am jealous pf the left coast. c\'mon walkies, come back to nyc. we need some \"RAT\" up in herre, ya heard?  i also saw  farenheit 9/11  on monday night. it was horribly depressing and i found myself crying several times. the movie was also kind of funny, but i think i was more annoyed with all of the wacky sound effects and musical emphases that michael moore was using to overscore his points. yeah, some of it was smart and witty, but other times, i felt like some obnoxious child was directing the documentary... er, film (it was far from an unbiased documentary). while i agree with a lot of what michael moore has to say, i think the way in which he expresses his opinions can be outlandish and over-the-top... which will turn a lot of people off. on the other hand, sometimes i think one has to be extreme to make an impression in people\'s minds. unfortunately, we americans are often too complacent and need to be shook. shook hard.   '),(15189,3457789,'31,May,2004','2004-05-31 00:00:00','   Yup, i\'m gonna start from scratch outta crappy blogger! it seriously sucks! ALL my javascripts dun work here! %^#!%#!@$%!@#%$  lolz...  '),(15190,3457789,'30,May,2004','2004-05-30 00:00:00','   urlLink    yes, its me after my nationals... *gosh* i so roasted... i think i was laughing at ana\'s lame crappy jokes... haha...&nbsp; urlLink  '),(15191,3457789,'30,May,2004','2004-05-30 00:00:00','   urlLink    g n pui tee doing something horrifying lolz...&nbsp; urlLink  '),(15192,3457789,'30,May,2004','2004-05-30 00:00:00','   urlLink    caught red handed again... lolz...&nbsp; urlLink  '),(15193,3457789,'30,May,2004','2004-05-30 00:00:00','   urlLink    neo neo up to no gd again... haha...&nbsp; urlLink  '),(15194,3457789,'30,May,2004','2004-05-30 00:00:00','   urlLink    juz me n sam sam... \")&nbsp; urlLink  '),(15195,3457789,'30,May,2004','2004-05-30 00:00:00','   2 words: i\'m clueless....  so dun ask me how i did... i really have no idea...  *goes off to search for javascript codes*  '),(15196,3457789,'30,May,2004','2004-05-30 00:00:00','   just had tuition from 2-5.30... so siong ar!  i\'m kinda worried bout tomorrow... but its too late to do anything of significance i suppose... I\'ll just be pointlessly battling oppoents in neopets lol...  '),(15197,3457789,'29,May,2004','2004-05-29 00:00:00','   wah, got so many words 2 study... can die ar!!!  i slept last night at 3am cuz i was studying chi... haiz... i guess its 2 lil, 2 late...  k, i gotta eat lunch and rush off to tution...  '),(15198,3457789,'29,May,2004','2004-05-29 00:00:00','   urlLink    look wad the chao g-boy put on his blog! grrrr!! cant u ppl ever leave me ALONE??!!&nbsp; urlLink  '),(15199,3457789,'28,May,2004','2004-05-28 00:00:00','   its quite scary to think that the chi o lvl\'s so so close... n i\'m so so dead... i think i\'ve woken up a bit too late so i\'ll just do wad i can and keep my fingers crossed...  just realised that even 1 mth is not enuf to catch up on the lost time i wasted away...  and oh yeah, my percentage points slipped a massive 6 pts and my lvl position got doubled... i can eat dirt wif my L1R5 at 21... *gosh* i cant even qualify for Poly like tt...  Guess i really gotta study... I\'ve finally learnt that one needs 2 study 4 the exams or get seriously raped...  kz, i\'m off to study my chi...  '),(15200,3457789,'28,May,2004','2004-05-28 00:00:00','   urlLink    Ana, \"sure of course wee keat darling! And you were saying?\"&nbsp; urlLink  '),(15201,3457789,'28,May,2004','2004-05-28 00:00:00','   urlLink    Wee Keat, \"i\'m getting a sore back from turning back and flirting with you ever so often! May i stand up and flirt with you?\"&nbsp; urlLink  '),(15202,3457789,'28,May,2004','2004-05-28 00:00:00','   wah lau.. mrs ho so inflexible! dun let me postpone tuition to study 4 chi... haiz...  basket.. my L1R5 got 21... really can go n die liaoz... goes to show tt one cant take an exam w/o studying or confirm tio rape...  well, 2day\'s the last day of sch n now its time to face the horrible june hols (i dun think i can even call it a holiday)...  kz, gotta rush off 2 chem tuition liaoz...  *shrugs* i wish i can go class chalet man...  '),(15203,3457789,'27,May,2004','2004-05-27 00:00:00','   yupz, 3 days 2 chi o lvls... haiz... sure tio rape ar...  hiaz, it\'s like i\'ve spent a day wasted in sch... did practically nothin... at least i complete my band of bros... basket, i 4got 2 bring my water bottle 2 sch... thirsty like siao...  hehe... i skipped chi tution 2day...  juz plated frozen throne wif sam sam and he got a hella scolding from the pro(me) but still refused 2 follow my instructions. In the end, we lost (of course) 8 stages to the end...  great, i just discovered i have 4 or 5 chi exam papers that i havent touched... guess i\'ll try 2 finish it by fri nite so i can let my tution teacher mark on sat...  kz, cyaz...  '),(15204,3457789,'26,May,2004','2004-05-26 00:00:00','   Well, eddy the g-boy threw down the gauntlet and challenged me to do up a blog so i gladly accepted cuz i know that i\'ll beat him thumbs down in casading style sheets (CSS), Javascripts and HTML... Haha!  I\'m gonna start edding the effects only after the chinese o levels on 31 May so please bear with the basic templete for now...  One thing that leaves me clueless is why people will wanna waste their precious time by reading about other people\'s lives lol...  Anyway, here\'s my entry 4 today:  Stupid xh whipped me wif his pencil box strapped and now my skin\'s *ahem* swollen and bulging up... *haiz*  haiz, 2day must stay back 4 the extra chinese lesson! make me lose me afternoon nap...  kz, i gotta rush off to my a math tuiton liaoz...  chaoz...  '),(15205,3457789,'22,June,2004','2004-06-22 00:00:00','   well this blog thingy\'s novelty is like wearing off... i dunno... it has become a chore to like update it... well, i really dunno...  '),(15206,3457789,'11,June,2004','2004-06-11 00:00:00','   *THROWS CONFETTI*  p.s. mine\'s tomorrow lolz...  '),(15207,3457789,'10,June,2004','2004-06-10 00:00:00','   urlLink    the two love birds... aint they sweet?&nbsp; urlLink  '),(15208,3457789,'09,June,2004','2004-06-09 00:00:00','   well, come 2 think of it. i think God has really been good... i cant even remember the last time i was sick. so i\'d like to thank Him for keeping healthy for much a long time... \")  but my backache\'s killing me!! \")  '),(15209,3457789,'06,June,2004','2004-06-06 00:00:00','   well, i\'m horribily sick! got inflammed tonsils, fever and a flu all at the same time... sianz ar... wad a nice early b-day prez i got eh?  '),(15210,3457789,'06,June,2004','2004-06-06 00:00:00','   sry bout not updating the blog cuz i\'m quite busy at my neopets forum...  anyway, i had an awesome time at festival of praise... i finally got to see hillsong up close at a concert... i hadn\'t jumped during worship 4 quite some time but i couldnt help it cuz hillsong totally RAWK!  n marty sampson was totally cool lol!  and ppl, rem must wear uniform 4 tmr\'s eng talk! - confirmed by mrs gan.  '),(15211,3457789,'04,June,2004','2004-06-04 00:00:00','   urlLink    edwin reveals his secret... haha...&nbsp; urlLink  '),(15212,3457789,'01,June,2004','2004-06-01 00:00:00','   Just realised i\'ll spend my bdae in Singapore! Finally my annual church camp doesn\'t clash wif it!!! yay!!  *throws confetti*  '),(15213,3457789,'01,June,2004','2004-06-01 00:00:00','   i apologise for lousy service of Eddy Technologies, sometimes the G-box wont appear so just hit the f5 key until it does...   sorry bout the trouble...  Edwin aka G-Boy CEO  '),(15214,3457789,'20,July,2004','2004-07-20 00:00:00','   You can trust Him!  &nbsp;  My friend, you can trust God today  He is good all the time  He is unparalleled and unprecedented  He is the centerpiece of civilization  He is the superlative of all excellence  He is the sum of all human greatness  He is the source of divine grace  His name is the only one able to save  And His blood is the only power to cleanse  His ear is open to the sinnerâ€™s call  His hand is quick to lift the fallen soul  Heâ€™s the eternal lover of us all  Who else can we trust but Him?  &nbsp;  He supplies mercy for the struggling soul  He sustains the tempted and tried  He sympathizes with the wounded and broken  He strengthens the weak and the weary  He guards and He guides the wanderer  He heals the sick and cleansers the leper  He delivers the captive and defends the helpless  And he binds up the broken-hearted  Surely we can trust him  &nbsp;  Jesus us the key to all knowledge  Heâ€™s the wellspring of wisdom  Heâ€™s the doorway of deliverance  And Heâ€™s the roadway of righteousness  Heâ€™s the highway of holiness  Heâ€™s the gateway to glory  And yes you can trust him  &nbsp;  Jesus IS more than enough  Heâ€™s the all-sufficient King  Heâ€™s the king of the Jews  Heâ€™s the of all Israel  Heâ€™s the King of righteousness  And Heâ€™s the King of the ages  Heâ€™s the King of Heaven  Heâ€™s the King of glory  Heâ€™s the King of all kings and  Heâ€™s the Lord of all Lords  And yes again you can trust him  &nbsp;  And rejoice in this my friend  He is a sovereign King  Thereâ€™s no gauge to measure His limitless love  Thereâ€™s no barrier to block His blessings outpoured  He is enduringly strong  And He is entirely supreme  He is eternally steadfast  He is immortally faithful  He is imperially powerful and  He is impartially merciful  He is Godâ€™s Son  And you can trust Him  &nbsp;  I wish I could more accurately describe Him to you  But Heâ€™s indescribable  Heâ€™s incomprehensible  Heâ€™s invincible  Heâ€™s irresistible  You canâ€™t outlive Him  And you canâ€™t live without Him  The Pharisees couldnâ€™t stand Him  But they found they couldnâ€™t stop Him  Pilate couldnâ€™t fault Him  Herod couldnâ€™t kill Him  Death couldnâ€™t conquer Him  And the grave couldnâ€™t hold Him  &nbsp;  My friend  Heâ€™s the Alpha and Omega  The First and Last  Heâ€™s the God of the present  Heâ€™s the God of the past  Heâ€™s the God of the future  And we rise to speak His name again and again  Jesus Jesus  He is Jesus  He is for us  So who can be against us  We can definitely, absolutely and totally trust Him&nbsp; ------------------------------------------------------------- &nbsp; Okay, this is a lame one i did for an english assignment some time back... &nbsp;  Daryl the bully&nbsp;  &nbsp;  Daryl was a bully,  The worst of all there is,  nobody liked him  for he was so awfully naughty.  &nbsp;  He tormented all the boys,  and broke all their toys.  Upside down he turned the neighbourhood,  &nbsp;no doubt nobody wanted to play with him.  &nbsp;  He was really nasty,  sparing no one.  Even the feeble old lady was left soaking wet,  Completely drenched in lemon tea.  &nbsp;  He would sneer, â€œYou little wimps,  come and challenge me to a fight!â€  Nobody dared to talk back to him,  Even though he was outnumbered.  &nbsp;  He laughed at the children,  mocked them and called them names,  He would tease them,  â€œstand up and oppose me if you dare!â€  &nbsp;  Even the adults were at a loss  of how they would deal with him.  The children werenâ€™t the only ones that Daryl terrorized,  The adults also testify of his unruly behaviour.  &nbsp;  Daryl took Mr. Moyâ€™s fax,  which he received from the insurance company,  and covered it with wax.  â€œLetâ€™s see how you are going to read that!â€  &nbsp;  He was a horrible tyrant,  That ruled with an iron hand.  everyone was wondering  when his reign would come to an end.  &nbsp;  All this was about to change  when Germaine moved into the neighbourhood.  All the children realized that Daryl behaved himself  Whenever Germaine was around.  &nbsp;  The children decided to tell ask Germaine for help  After a long discussion,  she agreed to do her part  To try to convince Daryl to stop all of the bullying.  &nbsp;  Germaine walked up to Daryl,  Who towered over her.  She uttered, â€œPick on somebody your own size!â€  Daryl was shocked and hung his head in shame.  &nbsp;  &nbsp;â€œIâ€™m sorry, I promise I wonâ€™t intimidate them any more.  But now nobody wants to befriend me,â€  The big bully replied,  almost in tears.  &nbsp;  â€œIâ€™m sure they want to play with you,  only if you are nice to them,â€  Germaine consoled the once great-and-mighty big bully,  â€œIâ€™m sure they will accept you.â€  &nbsp;  Daryl tried to be nice to those he once terrorised,  after he had seeked everyoneâ€™s forgiveness,  he was graciously pardoned and played together with the other children,  who gained a new friend in return for pawning a bully.  '),(15215,3457789,'18,July,2004','2004-07-18 00:00:00','   i write&nbsp;poems!!!! =) &nbsp;  The love i leaned on pierced my heart,  the hands i clung to loosened themsleves from mine.  the moon was darkened by a misty doubt,  and my candle\'s little flame went out.  &nbsp;  While i swayed, weak, trembling and alone,  the everlasting arms held my own.  My burdens did the Lord Almighty help me bear,  while generously showering His tender loving care.  &nbsp;  Despite all of my immense pain and sorrow,  even though i am left without a purpose or aim,  I&nbsp;know for a fact that yesterday, today and forevermore,  My Lord God will be the Same!  &nbsp;  -------------------------------------------------------------------  &nbsp;  Dear restless heart,  dont moan and be sorrow so.  Be still! for peace is found just by looking at His smile.  He has a meaning for the chilly winds that blow.  &nbsp;  Just hope, and hope and hope,  until in Him you braver grow.  God is always there! He will never, ever flee!  Just pray, and pray, and pray till\' you have faith to see!  ------------------------------------------------------------------  &nbsp;  Everything seems so so bleak,  \"I\'ll put my trust in him\",  my bruised heart faintly murmured,  as over me rolled a huge crushing boulder of woe.  &nbsp;  I am singing without the sound of melodious music,  without the sound of tuneful strings.  I am crushed and defeated but i will continue to trust  and ride on his caring wings.  '),(15216,3457789,'06,July,2004','2004-07-06 00:00:00','   the p-boy edwin (formely known as g-boy) keeps on bugging me to update my blog so i\'ll please him la...  okay, 2day\'s the 1st x in my life i got the highest in class 4 a test... i also dunno how i did it la... haha...  '),(15217,4257468,'13,August,2004','2004-08-13 00:00:00','   I\'m setting up this blog for my second period creative writing students.  I am hoping that they will fiind it interesting to keep personal thoughts, ideas, and writing considerations in this blog.  Likewise, we can post what we write and respond to each other\'s ideas.  Welcome and good luck blogging! Ms. H  '),(15218,4257468,'18,August,2004','2004-08-18 00:00:00','   OK gang, I\'ve decided to give you some homework this weekend--and lots of it.  After all, that\'s what you would expect isn\'t it???  Seriously, I\'m sending you on a scavenger hunt.  In order to complete some activities I have planned for a    P O  E  T  R  Y   E  X  T R A  V A G  A  N  Z       A    next week, I need you to have some materials ready.  Your assignment is this:  Each student in the class needs to find each of the following items.  The goal is to find the objects which will be used in class (so you shouldnâ€™t need to have them in the same form to take back home), and not spend any money on them.  Look around the garbage can for the junk mail or a \"stash-it-all\" drawer for the postcards or letters, for example.  Items due in class on Monday and I will give you credit for all of your hard work.  Good luck and happy hunting!    Item list   One permanent ink marker (preferably some dark color) One t-shirt (white or light color) that fits you An old post-card (blank) An old post-card with message A very, very, very sharpened pencil At least four used postage stamps (at least 1â€ x 1â€--the larger the better)â€”on the envelopes or off Ten pieces of junk mail An old magazine (school appropriate :) Some sort of blank book A piece of wood (no larger than 1â€™ x 1â€™ x 1â€) An old calendar page (any month--bring extras if you find them) An old letter A map (can be local, national, world, etc.) A family recipe A shoebox to hold all of it (unless your piece of wood is too big--try to make it fit!)    Point #2 You should begin e-mailing me your drafts that you are beginning to finish.  Mail to:   urlLink lhager@mercer.k12.ky.us  '),(15219,4257468,'17,August,2004','2004-08-17 00:00:00','   Given that your goal is to write at least two entries a week into this blog, I think I\'ll set things up like this. Each week, I\'ll give you a prompt (for example, the one from yesterday) and for the other prompt, you will be invited to either post your own writing for people to read/respond to AND/or respond to a second prompt. We\'ll see how this works. Even though things can be read by anyone searching through the blog, I believe that keeping the response contained will make things run much more smoothly.  Thanks. Ms. H  '),(15220,4257468,'16,August,2004','2004-08-16 00:00:00','   I\'m trying to figure out why some of you are on the blog and some aren\'t.  From what I can see, it looks like those of you who followed the email link are on and those who didn\'t aren\'t.  I will try to figure out how to fix this.  Meanwhile, could those of you who DIDN\'T follow the link go back and try to follow it.  It might be that the system simply has to make the connection.  Try this and see what it does.  Thanks!  '),(15221,4257468,'16,August,2004','2004-08-16 00:00:00','   Trot right on over to  urlLink http://www.writersdigest.com/  to find loads of new writing prompts.  You can even sign up to have them emailed to you each month automatically.  Now what could be better than that :)  '),(15222,4257468,'16,August,2004','2004-08-16 00:00:00','   Hi all, See, this blogging thing isn\'t too hard after all! Even I can do it. I hope you\'ve made your way around online and found this journal. My intention is for us to be able to keep it as a class journal much in the way we would do if we were keeping a written, bound journal. The way I understand it (yes, I\'m still trying things out here), is that after I put your email in you have to \"join.\" You should have received an email \"invitation.\" Then, you are able to post to this blog.  I\'d like to start out by setting up a first prompt for you to respond to: how do you perceive class going so far? I\'m interested, particularly, in knowing how you are responding to some of the \"freedoms\" that writing workshop allows like being able to define your own writing tasks--things that you want/need to write about, etc. How do you feel about being given writing time in class rather than it all being given as homework? What would you like to see as a feature of your class??  The other thing I\'d like some feedback on (and I\'m going to be asking you in class about this, too) is what you might want our webzine (and print magazine) to look like. What should its title be? We have to start thinking about how we will invite other writers to join our little party, so it\'s going to be important to make some of these decisions soon.  I can\'t tell you how long I\'ve waited to teach a creative writing class. I am really enjoying what you\'ve done so far. You just have to promise to push me as hard as I push you (well, maybe not that hard :)  Post your thoughts! I think I\'ll stumble over to the blog for portfolio_princess and journal my thoughts for the day. . . Let me know if you set up an online blog for yourself (or if you already have one you want to keep working in). I\'ll have to make note of the URL. Ms. H  '),(15223,4257468,'22,August,2004','2004-08-22 00:00:00','   We are getting ready to dive into our poetry extravagannnnnzzzzzaaaaaa this week. I\'d like to invite you to post something you write during the week to get feedback on. Have fun with it!  '),(15224,4257468,'22,August,2004','2004-08-22 00:00:00','   For this week, I\'d like you to consider this question. What is the best piece of writing you have ever done? Describe it. In what particular genre did you write? Why do you consider it your best?  '),(15225,4257468,'23,August,2004','2004-08-23 00:00:00','   hey again.  I guess i\'m supposed to write about things that happen in my life, what i think about things. . . to be honest, not a whole lot happens! that\'s ok though:)  Actually, this past weekend my boyfriend, Erik, moved into EKU--- so i\'m kinda nervous about that (eek!) We\'ve been together for about 9 months (which is a long time for the both of us lol). . . so hopefully i\'ll get to go visit him on the weekends.  For those of you who don\'t know me, i guess i could at least explain the whole \"stiletto Fran\" title.  First of all, if you know me you know i love shoes with a passion-- heels, flip flops, boots sandals, blue ones, pink ones, furry ones--- i think you get the picture lol.  And, well, the name Fran is a nickname that a couple of my crazy friends gave me-- it\'s a long story which evolves getting kicked out of fast food places and moms that can\'t hear, so i won\'t bother to explain!  I\'m a senior and i feel like i should still be 12, so it\'s kinda crazy.  I\'ve always dreamed of being a fashion designer, but now i\'m not so sure-- talk about changin your mind at the last minute lol.  I either want to go to U of L or SCAD, which is an art school in Georgia, it just depends on a lot of different things.  I have about two hobbies which include shopping and sewing.  Other than that, i can\'t think of a whole lot to tell at the moment. . . so, when i think of something i\'ll get back at ya! later ~luv~ cara   '),(15226,4257468,'19,August,2004','2004-08-19 00:00:00','   I am so tired and I am so glad that it is almost the weekend again. . . just one more day.  It seems that everytime Monday comes around again the countdown starts all over again.  But anyways, we\'re in the computer lab working on our blogs right now, I finally got linked to the class blog, and I am still trying to figure out what to write about.   '),(15227,4257468,'17,August,2004','2004-08-17 00:00:00','   I finally got my journal set up for my creative writing class, but now I don\'t know what to write in it.  So I just thought I would say hi to everyone.  So until I understand this thing a bit better, this is it! C-ya   '),(15228,3523319,'05,June,2004','2004-06-05 00:00:00','   Wow. It\'s like the old BOA, but slightly more confusing.  '),(15229,3523319,'04,June,2004','2004-06-04 00:00:00','   don\'t have scales,  urlLink it turns out . i was as surprised as anyone.  '),(15230,3523319,'03,June,2004','2004-06-03 00:00:00','   damn you all. i\'m still here for an hour yet. but i have read the review, and pronounce it Unforgiving.  '),(15231,3523319,'03,June,2004','2004-06-03 00:00:00','   I\'m going home.  '),(15232,3523319,'03,June,2004','2004-06-03 00:00:00','   It works for me. Shall I just copy and paste it?  '),(15233,3523319,'03,June,2004','2004-06-03 00:00:00','   Yeah, OK, it\'s not working and I can\'t be arsed to fix it.  It\'s been a bit temperamental recently.  I\'ll e-mail a draft of it to you instead dave.  '),(15234,3523319,'03,June,2004','2004-06-03 00:00:00','   Christ I\'m having a stressful day.  I\'m pretty much legging it now to get my scarecut before I leap on the train at Broadstairs to bum Dave.  The TDAT review should be working, if not, it\'s not my fault and I have no inclination to do anything about it.   urlLink So there  '),(15235,3523319,'03,June,2004','2004-06-03 00:00:00','   I\'ve done shockingly little work today. I have, in fact, spent a large proportion of it on the interweb.  I also downloaded the 14 day trial of Star Wars: Galaxies last night, all two gig of it. I\'m moist with anticipation. I think I may create a wookiee exotic dancer.  '),(15236,3523319,'03,June,2004','2004-06-03 00:00:00','   i never got to see it. is your site working again yet, uncle fonz?  '),(15237,3523319,'03,June,2004','2004-06-03 00:00:00','   watch me as i monkey around with your post. tee hee.  '),(15238,3523319,'03,June,2004','2004-06-03 00:00:00','   Ahh. I was just trying to find an appropriate size, but you done beat me to it.  '),(15239,3523319,'03,June,2004','2004-06-03 00:00:00','   that\'s you, that is.  '),(15240,3523319,'03,June,2004','2004-06-03 00:00:00','   I broke it. But tee hee anyway.  '),(15241,3523319,'03,June,2004','2004-06-03 00:00:00','     '),(15242,3523319,'03,June,2004','2004-06-03 00:00:00','   Whilst and amongst are archaic versions, apparently. But still equally valid.  '),(15243,3523319,'03,June,2004','2004-06-03 00:00:00','   this is becoming some kind of necessity here too - this fecher is just a bit too bright and obvious. i\'d like to leave this template for posterity if possible and come up with some alternatives for the more sensitive among us, so does anyone know how to take the blogger output and run it through another page? anyone?  christ, customising the boa was hella-easy. i miss that boa. well, in a way.  and what\'s the difference between among and amongst? anyone? and while and whilst? someone should know.  '),(15244,3523319,'03,June,2004','2004-06-03 00:00:00','   or poor eyesight. you\'ve probably got great big godzilla eyes too, haven\'t you?  '),(15245,3523319,'03,June,2004','2004-06-03 00:00:00','   I see. The other its. Piss off.  Why do I sometimes see the XML buttons, and sometimes not?  '),(15246,3523319,'03,June,2004','2004-06-03 00:00:00','   As in \'it is the future\'. Or were you talking about something else?  '),(15247,3523319,'03,June,2004','2004-06-03 00:00:00','   you need a rest, guy. why has no-one complimented me on my nice buttons? or the johnson quote about yoav at the top? why am i always doomed be ignored, and my art overlooked in favour of crude populist hackery?  '),(15248,3523319,'03,June,2004','2004-06-03 00:00:00','   That\'s because your reader is Too Straight, like you. You need to loosen up. Hey relax, fella.  '),(15249,3523319,'03,June,2004','2004-06-03 00:00:00','   my reader doesn\'t. anywoo, you can just click on one of those XML links over there, and things will look hella different. honest.  '),(15250,3523319,'03,June,2004','2004-06-03 00:00:00','   Most feed readers will work with Atom, anyway.  '),(15251,3523319,'03,June,2004','2004-06-03 00:00:00','   blogger uses atom xml, rather than rss. but fear not - i\'ve just converted it to rss with feedburner. take a look at the cheeky XML link on high on the right. there you go. that\'s not so bad, is it?  '),(15252,3523319,'03,June,2004','2004-06-03 00:00:00','   It\'s the future, man.  Fonz, you could get yourself a feed reader (Google for it), and then look at the BOA using it\'s RSS feed. It\'ll just look like you have an email program open, or somesuch.  '),(15253,3523319,'03,June,2004','2004-06-03 00:00:00','   there must be another way to get at the content of the site without all the botting. i\'ll have a luke.  '),(15254,3523319,'03,June,2004','2004-06-03 00:00:00','   It\'s no good, this site is drawing too much attention.  I can\'t look at this at work, I\'ve already got in trouble once today.  If I don\'t get something as inconspicuous as PISS back, I\'ll have to stop using it.  '),(15255,3523319,'03,June,2004','2004-06-03 00:00:00','   Having misspelled \'Fonz\' in an earlier post, I have decided to complete the job. From now on, I shall refer to him as \'Diddy Bot Fonx\'.  '),(15256,3523319,'03,June,2004','2004-06-03 00:00:00','   i am a pulling 66%er. that makes me a little better than ross. which makes sense to me. i am also 30% gay, which is too straight.  i think that these tests are indeed \"balls\".  '),(15257,3523319,'03,June,2004','2004-06-03 00:00:00','   I didn\'t half panic when that needle went all the way up to 100% gay, til I realised it was just calculating my score.  I am 26% gay, which is apparently &quot;Too Straight&quot;.  '),(15258,3523319,'03,June,2004','2004-06-03 00:00:00','   Drunk Fonx would get 4% and irreperably stained trousers.  '),(15259,3523319,'03,June,2004','2004-06-03 00:00:00','   I was brutally honest in the pull-o-meter, and I got 54%. I am a Shy Guy.  Seems about right.  Of course, that was Sober Fonz taking the test.  If Drunk Fonz were taking the test, who knows what score he\'d get?  '),(15260,3523319,'03,June,2004','2004-06-03 00:00:00','   I\'m 43% gay, and apparently a \'happy and well adjusted hetero man\'. See, it must be true if Channel 4 says it.  I also have a 62% chance of pulling, and it says:  \'Always in control, and that often works out the best on a holiday night out, But stop playing it so safe! We know its not cool to get wasted, but you need to let down your hair and live a little! Perhaps a bit of dutch courage will give you that pep you need to go for it a bit more.\'  I\'m a little confused by that, as I chose to go without food and went for the beers, so would be legless by about half way round.  '),(15261,3523319,'03,June,2004','2004-06-03 00:00:00','   I took one of those gay tests before, where it reckoned everyone was at least 37% gay on average.  I think I was 1% nmore hetero than average, which surprised me as much as anyone.  '),(15262,3523319,'03,June,2004','2004-06-03 00:00:00','   See if you are  urlLink good at pulling ...I got 81%. Go here to see how  urlLink gay you are  Fonz anything less than 50% and I know you are lying! I got 29% which is just the right amount of gayness. If the above links come out stupid apologies...I just copied fonz\'s code and pasted  '),(15263,3523319,'03,June,2004','2004-06-03 00:00:00','   If you use the \'post a full message\' link, and use the interweb link button to insert the hutchtemel for you, it doesn\'t add the target=blank shiznit by default. Is there a way to change that, or am I going to have to learn something, and do my linking myself?  '),(15264,3523319,'03,June,2004','2004-06-03 00:00:00','   makes the target of the link open in a new window. thing is, my work pc usually has about 9 ie windows open at any one point, and takes 30 seconds to open another one. whereas i don\'t mind hopping forwards and backwards in my \"leisure\" window. hence my behaviour.  no problem to watch the botter; it\'s shorter than the other two, and apparently better - i watched the second one the other day, and pronounced it \"meh\".  you know, like ross would have done. only not as well.  '),(15265,3523319,'03,June,2004','2004-06-03 00:00:00','   It makes the link open in a new window, standard practice if you are linking to a page on an external site.  i.e. &lt;a href=&quot;http://www.bbc.co.uk/&quot; target=&quot;blank&quot;&gt;BBC&lt;/a&gt;  I think.  It\'s not easy typing that in html.  '),(15266,3523319,'03,June,2004','2004-06-03 00:00:00','   What does that target=_blank thing do? I don\'t know much hutchtemel.  '),(15267,3523319,'03,June,2004','2004-06-03 00:00:00','   I demand when you post a link next time dave, that you use a target=&quot;_blank&quot; thing.  Do it.  Now then, to importnat matters.  I think I need to see Larry Botter and The Prisoner of Azerbaijan this weekend, for reviewing purposes.  What are the chances of us catching a showing, maybe you\'ll see me at the mattinee on Sunday?  '),(15268,3523319,'03,June,2004','2004-06-03 00:00:00','   they\'re considered a delicacy in many cultures, i\'ll have you know.  '),(15269,3523319,'03,June,2004','2004-06-03 00:00:00','   I don\'t think that\'s a problem you need worry about, Dave.  Unless pots of jizz become a valuable art form.  '),(15270,3523319,'03,June,2004','2004-06-03 00:00:00','   someone find me a bittorrent/e-donkey/something link to download this for my zboz. it\'s not out here until august, and i\'d never pay for it anyway.  i wonder whether, should i ever produce anything of any artistic merit whatsoever, i\'ll still be so blase about other people\'s intellectual property rights?  '),(15271,3523319,'03,June,2004','2004-06-03 00:00:00','   matters dick. i\'ve got an xp pid changer, and i can\'t be the only one. there are always pure pids out there.  mmm. pure pids.  '),(15272,3523319,'03,June,2004','2004-06-03 00:00:00','   urlLink Microsoft bars Windows pirates   I don\'t want to have to give Billy Gates my money.  '),(15273,3523319,'03,June,2004','2004-06-03 00:00:00','   If putting my name in the title had been deliberate, it would have been acceptable. But it wasn\'t. I just spacked it. I feel so Yoav.  '),(15274,3523319,'03,June,2004','2004-06-03 00:00:00','   I got six out of seven on the BBC news quiz, because I always have to guess at the last question, about the ages of people I don\'t know.  I actually knew the people today, but misread the question, and put them in order of oldest first. Tits.  Dave, your eye is bloodshot because you burst a blood vessel through overly vigourous wanking.  '),(15275,3523319,'03,June,2004','2004-06-03 00:00:00','   What\'s the fucking deal with all this work I have to do?  It sucks.  '),(15276,3523319,'03,June,2004','2004-06-03 00:00:00','   me, up the knightley. hypothetically, unfortunately, but the theoretical evidence is strong. and i\'m looking forward to comissioning a full practical study.  again and again and again and  '),(15277,3523319,'03,June,2004','2004-06-03 00:00:00','   CHOO CHOO CHOO CHOO CHOO Choo Choo Choo Choo choo choo choo choo choo  That\'s my impression of a train nearby getting further away  '),(15278,3523319,'03,June,2004','2004-06-03 00:00:00','   i just got me 7 out of 7 on the bbc news quiz this week. i don\'t think i\'ve ever managed that before.  '),(15279,3523319,'03,June,2004','2004-06-03 00:00:00','   is really bloodshot today. why would that be? i want ross to give me answers.  '),(15280,3523319,'03,June,2004','2004-06-03 00:00:00','   that\'s no problem. last night, i accidentally missed the gym and went home and played kevilisation for hours. i\'m not proud.  a mighty wind is a good film, though.  '),(15281,3523319,'03,June,2004','2004-06-03 00:00:00','   On a scale of one to 10 of getting completely wasted, dancing like a fool and then pulling womens I\'m going to be hitting a fat 10 tonight...I\'m very much in the mood and it\'s only just 9.  '),(15282,3523319,'03,June,2004','2004-06-03 00:00:00','   Right, Dave, I\'m warning you now, I\'m coming to your house later, and I might require food and water.  I\'ll let you know what time I\'m arriving when I\'m on the train, but it\'ll be between 7:30 and 8:15 I should think.  '),(15283,3523319,'03,June,2004','2004-06-03 00:00:00','   I\'ve been at work for about half an hour, and already I\'m bored shitless. I want to go home.  '),(15284,3523319,'03,June,2004','2004-06-03 00:00:00','   I still make that only 14.  And that Kingston one obviously doesn\'t count.  Although now you mention it, I\'ve been to Kingston-upon-Thames too.  '),(15285,3523319,'02,June,2004','2004-06-02 00:00:00','   screw you guys, i\'m going to the gym.  '),(15286,3523319,'02,June,2004','2004-06-02 00:00:00','   although i\'ve been to kingston upon thames, rather than kingston jamaica... does that matter?  '),(15287,3523319,'02,June,2004','2004-06-02 00:00:00','   Paris  Rotterdam  Mauritius  Kingston  Kentish Town  Felixstowe  Brixton  Antwerp  Edinburgh  Tokyo  Amsterdam  Manchester  Dublin  Tooting   howzat?  '),(15288,3523319,'02,June,2004','2004-06-02 00:00:00','   I don\'t believe you have been to 15 of those places.  I dispute it.  List the ones you\'ve been to.  '),(15289,3523319,'02,June,2004','2004-06-02 00:00:00','   Wow, that took quite a while to turn up didn\'t it?  '),(15290,3523319,'02,June,2004','2004-06-02 00:00:00','   It seems to me  that we humans are trying to divorce ourselves from the other animals on the  planet as much as possible.&nbsp; We buy our food in supermarkets and cook it,  we sleep in well built houses on beds.&nbsp; We speak to each other and wear  clothes.         &nbsp;     However, it seems  to me that we are still linked to the animal kingdom, because we shit.&nbsp; It  is a distinctly animal trait to just wander off to a corner somewhere, squat and  curl one down.&nbsp; It produces no useful output, it\'s waste we have to deal  with, and it\'s a waste of our time too, so my question is  this:         &nbsp;     Do you think a  time will come when technology advances enough that humans don\'t shit  anymore?  '),(15291,3523319,'02,June,2004','2004-06-02 00:00:00','   i was talking to him only a minute ago, and he was about to sign up properly. but then i had to sign out of messenger, and i lost him. goddamit. if anyone sees him lying around, give him an encouraging poke with a sharp stick.  '),(15292,3523319,'02,June,2004','2004-06-02 00:00:00','   I\'ve been to 8 places on that Rambling thing.  '),(15293,3523319,'02,June,2004','2004-06-02 00:00:00','   Yes, I\'m not sure why it\'s doing that. I need to Sort It Out, but I don\'t seem to have time right now.   I\'ll fix it at home later.   The film is precisely that bad.  My browser is crashing left right and center.  I can\'t wait for a new computer.  You know, one that works.  '),(15294,3523319,'02,June,2004','2004-06-02 00:00:00','   again. we need a poll or something.  '),(15295,3523319,'02,June,2004','2004-06-02 00:00:00','   Yes, I\'m not sure why it\'s doing that.  I need to Sort It Out, but I don\'t seem to have time right now.  I\'ll fix it at home later.  The film is precisely that bad.  '),(15296,3523319,'02,June,2004','2004-06-02 00:00:00','   I got to the review eventually, but everything else I touched turned to server error.  '),(15297,3523319,'02,June,2004','2004-06-02 00:00:00','   the film was better than HTTP Status 500 - Internal Server Error  '),(15298,3523319,'02,June,2004','2004-06-02 00:00:00','   Who\'s is that site, Fonz? Why is so much of it wanged up?  I disagree with your review. The film is Not That Badâ„¢.  '),(15299,3523319,'02,June,2004','2004-06-02 00:00:00','   kentish town and tooting  '),(15300,3523319,'02,June,2004','2004-06-02 00:00:00','   I have been to precisely two of those places. Guess which ones.  '),(15301,3523319,'02,June,2004','2004-06-02 00:00:00','   I\'ve been to 15 of the places mentioned in the  urlLink Ramblin\' Man  lyrics. I bet that\'s more than Ross.  Fonz, your web site isn\'t working at me.  '),(15302,3523319,'02,June,2004','2004-06-02 00:00:00','   Comments are cack, turn them off now.  I can\'t believe how long it took me to do this, but here\'s the review:   urlLink Fonz tells it like it is   It didn\'t help that my computer kept crashing.  '),(15303,3523319,'02,June,2004','2004-06-02 00:00:00','   I posted this from my mobile. I wonder if it will work?  '),(15304,3523319,'02,June,2004','2004-06-02 00:00:00','   looks familiar. i demand that we change it every so often for the purpose of our own amusement.  '),(15305,3523319,'02,June,2004','2004-06-02 00:00:00','   This is a regular email, like they used in the 1800\'s. Not a fancy mobile one like nowadays.  '),(15306,3523319,'02,June,2004','2004-06-02 00:00:00','   I tried to email post from my mobile, and it seems to have spectacularly failed.  '),(15307,3523319,'02,June,2004','2004-06-02 00:00:00','   but if no-one likes them, we\'ll get shot of them again. i thought we might need them to comment on fonz\'s review, though... ho ho.  '),(15308,3523319,'02,June,2004','2004-06-02 00:00:00','   and yes, because it will all come out of our mouths when we talk. this is a condition known as being an MBA student.  tee hee.  '),(15309,3523319,'02,June,2004','2004-06-02 00:00:00','   post your emmerich-slaying review, whydontcha?  '),(15310,3523319,'02,June,2004','2004-06-02 00:00:00','   i\'ve set it to limit the front page to three days\' worth of output, but at this rate that\'s still an offal lot. there\'s a link on the right hand side to jump straight down to the posty bit, though. the one that calls you a whoo-er. you whoo-er.   urlLink click me and jump southwards  '),(15311,3523319,'02,June,2004','2004-06-02 00:00:00','   Cause I can post stuff.  Im not going to though.  So there.  Oh yeah. make the posting box at the top so in a years time we don\'t have to spend a day scrolling to the bottom of this page.  '),(15312,3523319,'02,June,2004','2004-06-02 00:00:00','   It seems to me that we humans are trying to divorce ourselves from the other animals on the planet as much as possible.  We buy our food in supermarkets and cook it, we sleep in well built houses on beds.  We speak to each other and wear clothes.   However, it seems to me that we are still linked to the animal kingdom, because we shit.  It is a distinctly animal trait to just wander off to a corner somewhere, squat and curl one down.  It produces no useful output, it\'s waste we have to deal with, and it\'s a waste of our time too, so my question is this:   Do you think a time will come when technology advances enough that humans don\'t shit anymore?  '),(15313,3523319,'02,June,2004','2004-06-02 00:00:00','   Hmm, I tried another e-mail post, it totally didn\'t work.  Poo.  I\'ll post it the regular way instead.  '),(15314,3523319,'02,June,2004','2004-06-02 00:00:00','   and that\'s not all  i invited dave, too  '),(15315,3523319,'02,June,2004','2004-06-02 00:00:00','   when i post by email. like this.  '),(15316,3523319,'02,June,2004','2004-06-02 00:00:00','   You go to settings, then e-mail.  Then there are two address options.  As dave says, the bottom one has a box in the middle of an e-mail address for you to put a password in.  Then you e-mail to that address, and bob is etc..  '),(15317,3523319,'02,June,2004','2004-06-02 00:00:00','   i tried an email post and nothing happened. bah and double-bah.  '),(15318,3523319,'02,June,2004','2004-06-02 00:00:00','   plus a min four-character .code in it, to stop damo guessing it and posting as them. then all you need to do is send something to that address, and bob\'s your proverbial.  mine is dave_mf.lovesross@blogger.com. or is it blogspot.com?  '),(15319,3523319,'02,June,2004','2004-06-02 00:00:00','   El Finz, what address did you send to? I just looked at the settings, and it didn\'t seem to be set up yet. It also has my name in the address, unfortunately. Sac du ballons!  '),(15320,3523319,'02,June,2004','2004-06-02 00:00:00','   hmm. it worked.   maaaaaaaaaaaaaannnnnnnnnnnnxxx!  '),(15321,3523319,'02,June,2004','2004-06-02 00:00:00','   It does mean clever people can do it via telemophone.  '),(15322,3523319,'02,June,2004','2004-06-02 00:00:00','   This is me posting  via e-mail.         &nbsp;     Can\'t see the  point of it myself.  '),(15323,3523319,'02,June,2004','2004-06-02 00:00:00','   maaaaanx!  '),(15324,3523319,'02,June,2004','2004-06-02 00:00:00','   I tell you what Dave, you stand over there and call manx, and I\'ll stand here and call him, and who(m)ever he comes to first, he must like the best.  Manx!  '),(15325,3523319,'02,June,2004','2004-06-02 00:00:00','   Someone should try posting via email. Go on, try it.  '),(15326,3523319,'02,June,2004','2004-06-02 00:00:00','   I, for the most part, concur.  Especially about the Brocolli thing.  I demand Dave make this page... you know... different.  And move the posting box at the bottom of the page over to the left. I don\'t like it centered. [If any mofo mentions the spelling of that word, I will have to open up a can of whoop-ass on them]  '),(15327,3523319,'02,June,2004','2004-06-02 00:00:00','   The page that is Any idea what time we are due in Epson? I will be nursing a hangover undoubtedly so need to mentally prepare myself for an hour long train trip from Angel. I just ate a load of broccoli...I remember a few years ago I was like \"what the fuck is a broccoli?\" not believing anything could be so disgusting.  Gradually over time it\'s become my favourite vegetable. I think we should feature up this page a bit so we can see some distinct advantages over the old one. Maybe an option to upload pictures etc. to the right of the message entry box  '),(15328,3523319,'12,June,2004','2004-06-12 00:00:00','   \'m scared for england. watching greece bum portugal has reminded me how easy it is to lose a football match.  shall have more wine and try to forget.  '),(15329,3523319,'11,June,2004','2004-06-11 00:00:00','   The incredible feeling of excitement I\'ve had all week about the England game tomorrow is slowly starting to be replaced with dread, fear and anxiety.  I hate that.  '),(15330,3523319,'11,June,2004','2004-06-11 00:00:00','   I\'m only getting about three a day at the moment, but it still annoys me.  Plus, it can\'t be long before I\'m getting a lot more.  SPAM = GAY  '),(15331,3523319,'11,June,2004','2004-06-11 00:00:00','   I\'m only getting about three a day at the moment, but it still annoys me.  Plus, it can\'t be long before I\'m getting a lot more.  SPAM = GAY  '),(15332,3523319,'11,June,2004','2004-06-11 00:00:00','   Yeah...spam is a fucker and there is nowt you can do about it really. There is one reasonably effective method which is a plugin for outlook which my old man used for a wile.  (Can\'t remember what it is called or where I got it from so you\'d have to research it a bit!)   It\'s kind of similar to adaware in a way though.  First it aloows you to specify basic rules (which you can do in outlook already) i.e filter out emails with abnormal characters in the address field/emails with certain attachments etc. The best bit is log file that you download every day from the plugin server which contains a relatively up to date list of spam addresses/domains/subject titles (new ones are created every minute so it doesn\'t catch everything) He was getting somewhere in the region of 100 spam mails coming from jap domains through to standard \"porn\" domains...when he slammed this on he got it down to about 10 a day. I spoke to him about this already as I need it but he doesn\'t have it anymore...anyone tracks this app down or something similar let me know as I also want to break things at the moment.  '),(15333,3523319,'11,June,2004','2004-06-11 00:00:00','   outlook 2003 is pretty good at detecting and removing spam, but not 100%. about the only thing you can do is tell it to accept mail only from people in your address book, like on hotmail.  '),(15334,3523319,'11,June,2004','2004-06-11 00:00:00','   That was not the answer I was looking for.  It seems to me, the smart spammer could send out an html e-mail, and the server would register the e-mail address was valid when the user read the e-mail and the html got served.  Which is pretty much what I expect they do. So as you say, you can\'t stop the flow of spam.  I thought there might be some sort of blocker I could get.  '),(15335,3523319,'11,June,2004','2004-06-11 00:00:00','   if you\'re being spammed, you\'re effectively bummed. there is no escape. if you click the Remove Me links, you just confirm to them that your email address is active, and they redouble their efforts. the only solution is to close the account and start another.  it annoys the shit out of me because i had a perfectly good email address for 6 years, then some cunts got hold of it through a whois record and now it\'s 50 spam for every proper email i get. it gives me the rage.  '),(15336,3523319,'11,June,2004','2004-06-11 00:00:00','   Yes, I don\'t think anything will appear until we start scoring.  Either that or we\'ve fucked it.  Why aren\'t I drinking beer yet?  '),(15337,3523319,'10,June,2004','2004-06-10 00:00:00','   When i go to \'my mini-league\' in the fantasy football.  Theres mostly nothing there  '),(15338,3523319,'10,June,2004','2004-06-10 00:00:00','   Jubus CHRIST, I\'ve been working ALL AFTERNOON on a FRIDAY.  and i had to work late as well.  I need to serously chill out with some whisky and badgers  re Ledley King: 5 Caps comprise one start and 4 second half appearances in Svens 22 man friendlys.  Truth of it is  (22 per team that is.  not 22 on the pitch, cos thats quite standard)  '),(15339,3523319,'10,June,2004','2004-06-10 00:00:00','   is this working now  '),(15340,3523319,'10,June,2004','2004-06-10 00:00:00','   Here\'s what I keep meaning to ask: one of my e-mail accounts has started getting the shit spammed out of it.  What can I do to stop this?  '),(15341,3523319,'10,June,2004','2004-06-10 00:00:00','   dear me, i pushed the envelope a bit with the amazon \"things you might like\" bug, and then went to The Page I Made.  i was surprised, to say the least. and a little scared.  '),(15342,3523319,'10,June,2004','2004-06-10 00:00:00','   i had a shave this morning. i resented it as a waste of 5 minutes, but at least i looked less like ross as a result.  the bearded, hairy godzilla scot.  '),(15343,3523319,'10,June,2004','2004-06-10 00:00:00','   I just went to the bog and looked in the mirror.  It seems my ratty whiskers, over the course of the day, ahve gone from &quot;I really should shave, but i\'ll just look ratty&quot; to &quot;I definitely should have shaved I look like Gary Neville\'s best attempt at a \'tache and goatie&quot;.  This embarasses me.  '),(15344,3523319,'10,June,2004','2004-06-10 00:00:00','   Have any of you seen this? Go to Amazon, and search for \'Sennheiser HD202 Headphones\'. Then look at the \'you may also like...\' panel.  Apparently Amazon have been having quite a few problems like this, recently.  '),(15345,3523319,'10,June,2004','2004-06-10 00:00:00','   manx looks like he\'s been added to the fan foot league OK now.  '),(15346,3523319,'10,June,2004','2004-06-10 00:00:00','   a wise man once said  \"the yoav is smelly.\" it\'s an eternal truth, i think.  '),(15347,3523319,'10,June,2004','2004-06-10 00:00:00','   Is that the &lt;blockquote&gt; tag?  Whatchoo talkin\' \'bout?  Apparently Ledley King has had five England caps.  I only remember one.  '),(15348,3523319,'10,June,2004','2004-06-10 00:00:00','   england: lasers and ninjas and gold wales: sheep and dirt  goal, it\'s a goal, ooooooh, it\'s a goal  '),(15349,3523319,'10,June,2004','2004-06-10 00:00:00','   now quotes are italic   that\'s not one i use much.  '),(15350,3523319,'10,June,2004','2004-06-10 00:00:00','   It does that quote thing...indents the writing a bit and puts a line space in...I was doing that hutchtemel thing  '),(15351,3523319,'10,June,2004','2004-06-10 00:00:00','   what did it do?  '),(15352,3523319,'10,June,2004','2004-06-10 00:00:00','   Ah ok, furry muff  what does this do me wonders?  '),(15353,3523319,'10,June,2004','2004-06-10 00:00:00','   I think yours is in damo, because you can see Manx pending right?  Attack of the Scones.  It seems we won\'t appear in a table until we\'ve scored some points.  '),(15354,3523319,'10,June,2004','2004-06-10 00:00:00','   is my team in ok fonz?  '),(15355,3523319,'10,June,2004','2004-06-10 00:00:00','   is this working now, it stopped working for a while from work as well  '),(15356,3523319,'10,June,2004','2004-06-10 00:00:00','   I have no idea why manx is still pending in the league.  Let\'s hope that when we start scoring, it all sorts itself out.  '),(15357,3523319,'10,June,2004','2004-06-10 00:00:00','   i\'m sure glad that damo is several miles away.  '),(15358,3523319,'10,June,2004','2004-06-10 00:00:00','   I feel a bit bundersome still...my coke tastes like whisky...I may leave it. I had a pro plus for the first time ever an hour ago...not sure if it works or not - I am shaking a lot and feel a bit more aware but these signs are easily confused with alchoholism.   I feel like a big poo and because I\'m in a nasty mood there aren\'t going to be any courtesy flushes either...SMELL MY PAIN!  '),(15359,3523319,'10,June,2004','2004-06-10 00:00:00','   Regarding posting from home, I can sign in and it recognises me as me but just doesn\'t have the Book of Arse listed under \'Your Blogs\' or whatever it says.  Oddd  Someone try sending me an invitation to my hotmail and i\'ll see what happens when I try to use that from home  '),(15360,3523319,'10,June,2004','2004-06-10 00:00:00','   I\'ve pressed submit and my team is listed as fully built but when i look attt the mini-leage I appear to be the only one in ittt  '),(15361,3523319,'10,June,2004','2004-06-10 00:00:00','   I had something similar, Dave - someone had jammed a paper clip under one of the control keys on my keyboard, which caused me no end of problems. I had to pry the key off, eventually.  '),(15362,3523319,'10,June,2004','2004-06-10 00:00:00','   I only see one other team inthat league? Ok I\'ll get then newer civ game tonight  '),(15363,3523319,'10,June,2004','2004-06-10 00:00:00','   Manx, I think you still need to submit your team, you appear to be  pending  still.  Ross, you\'d better sign up when you get home or I\'m going to muck up your lens.  '),(15364,3523319,'10,June,2004','2004-06-10 00:00:00','   actually, it\'s not that odd now i look at it again. there just seemed to be far too many letters.  my control key keeps sticking. you would not believe how strangely a computer with a sticky control key behaves. go on, try it.  '),(15365,3523319,'10,June,2004','2004-06-10 00:00:00','   that\'s the oddest spelling of \"definitely\" that i\'ve seen for a while...  '),(15366,3523319,'10,June,2004','2004-06-10 00:00:00','   That submit button was definateley not there before.  According to my latest battery of tests I\'m a peace loving hippy like Dave and my theme tune is Back in Black by AC/DC  '),(15367,3523319,'10,June,2004','2004-06-10 00:00:00','   my kev is the III, and is a bit out of date, although i only found that out once i was pretty far into my game. they\'ve release a couple of expansion packs and a raft of patches, so i thought that if i were to have another shot at it, i\'d want to get all the newest stuff together first. if you really want the one i\'m using, i\'ll see what i can do, but you\'d probably be better off getting hold of a newer version yourself, as that\'s what i\'m intending to do.  i\'ve read a couple of guides to playing on the harder levels of kev, and it\'s super-hard and mind-blowingly technical. i\'m not sure i have the patience to calculate the precise number of turns it will take my worker unit to cross a river, irrigate on square, and mine another and rationalise that against controlling my tax rate so that a new science advance arrives exactly when i need it, while all the time managing my city population\'s working pattern and growth to alter levels of production to synchronise with an anticipated event in 40 turns\' time.  for christ\'s sake.   urlLink a good bomb or a bad bomb?  '),(15368,3523319,'10,June,2004','2004-06-10 00:00:00','   It\'s like this:  Name: Don\'t tempt me, Figo Password: iloveross  I also noticed just now that Desailly had made it back into my squad, despite me taking him out ages ago.  I think some of the changes I made originally didn\'t happen.  Hence Beckham not being in originally.  '),(15369,3523319,'10,June,2004','2004-06-10 00:00:00','   What is the league name/password I need to add my team to the league now...  '),(15370,3523319,'10,June,2004','2004-06-10 00:00:00','   Have any of you been watching The Mighty Boosh on t\'BBC? I find it bizzare yet amusing.  '),(15371,3523319,'10,June,2004','2004-06-10 00:00:00','   I cannae look at that link, Fonz. The category \'games\' is filtered at this time. Bastards.  Dave, is this Civ III you keep banging on about ad nauseum, or is there a new one?  '),(15372,3523319,'10,June,2004','2004-06-10 00:00:00','   I am the gayest man ever to walk this earth...wait a minute...I\'m not the gayest man ever to walk this earth...wait a minute...I\'m not gay...so confused...  I will make a team now...I\'m booking last place in the league...give me kevinisation...do it now Dave  '),(15373,3523319,'10,June,2004','2004-06-10 00:00:00','   i\'m not convinced that that submit button was there before. winkers.  '),(15374,3523319,'10,June,2004','2004-06-10 00:00:00','   i finished my game of kevinisation last night, most unexpectedly. it was about 1865, and i was launching the second of my glorious wars of righteous liberation deep into the indian hearland, with wave after wave of modern tanks and mechanised infantry pouring into their borders, when all of a sudden a box popped up in the middle of the screen that said \"You Win! You have achieved a Cultural Victory\" and the game was over.  i felt a little let down, but it at least got me to go to bed before 12.  '),(15375,3523319,'10,June,2004','2004-06-10 00:00:00','   Ross, build a  urlLink fan foot team  just for shits and giggles.  And you damo, or I\'ll believe this letter about you being the gayest man e\'er to walk the earth.  '),(15376,3523319,'10,June,2004','2004-06-10 00:00:00','   That is gay.  Listen up homos:  I was just perusing my beautiful Euro 2004 Fantasy Football team when i noticed a little hyperlink at the top that said &quot;Submit&quot;.  It turns out, that although I\'d fully built my team, the system doesn\'t do anything with it until you press the submit button.    I sincerely hope I am the only spaz that didn\'t notice this, but if I\'m not, you guys might want to submit now.  '),(15377,3523319,'10,June,2004','2004-06-10 00:00:00','   I fucking hate my interweb connection. I can\'t stay signed in to anything.  '),(15378,3523319,'10,June,2004','2004-06-10 00:00:00','   I too am digging that sweet savings action.  I\'m hoping they rocket up to 7.5%.  And then the housing market will crash too.  Brilliant.  '),(15379,3523319,'10,June,2004','2004-06-10 00:00:00','   stoopido yoav  '),(15380,3523319,'10,June,2004','2004-06-10 00:00:00','   stoopido mortgate holders  '),(15381,3523319,'10,June,2004','2004-06-10 00:00:00','   dynamic ip address allocation. it\'s used on big networks with flexible clients.  if anyone hasn\'t seen popbitch this week, they should check  urlLink this shit  out.  i\'m glad interest rates are going up. i might make some money on my savings for once. stoopid mortgage holders.  '),(15382,3523319,'10,June,2004','2004-06-10 00:00:00','   What the hell is a DHCP?  '),(15383,3523319,'10,June,2004','2004-06-10 00:00:00','   I doubt they are.  There are a number of reasons you can drop it - Most likely in your case would be firewall configuration or the fact you are a bummer. It shouldn\'t affect you too much anyhow unless you have some sort of bummer setup...  '),(15384,3523319,'10,June,2004','2004-06-10 00:00:00','   Alright, you compscis, solve my problem for me: why does my ISP keep giving me a one hour DHCP lease instead of the usual three days?  '),(15385,3523319,'10,June,2004','2004-06-10 00:00:00','   I\'d quite like to go home now. Work is boring.  '),(15386,3523319,'10,June,2004','2004-06-10 00:00:00','   If you still can\'t post from home and run a firewall on your computer...turn it off just before you hit publish post button...then turn it back on when the publish completes.  '),(15387,3523319,'10,June,2004','2004-06-10 00:00:00','   What on earth makes Gibbo think i\'m remotely interested in interest rates  weirdo  '),(15388,3523319,'10,June,2004','2004-06-10 00:00:00','   for the beeb,  urlLink this  is very good. it\'s probably nicked from somewhere.  i saw bad education. it\'s pretty good, but... dude...   gayest   film   ever.  '),(15389,3523319,'09,June,2004','2004-06-09 00:00:00','   i can\'t believe ray charles is dead.  oh wait, yes i can. he was very old.  '),(15390,3523319,'09,June,2004','2004-06-09 00:00:00','   You could just test that for me by answering the questions to the very best of your ability.   Oh yes, I went there.  '),(15391,3523319,'09,June,2004','2004-06-09 00:00:00','   You could test that by doing it again and answering wrongly.  I bet you wouldn\'t come out as a \'goddamn fucktard\'.  '),(15392,3523319,'09,June,2004','2004-06-09 00:00:00','   Apparently my Super IQ test is 147.  That\'s more like it.  I am an  Intuitive Investigator , which makes me  truly brilliant .  I can\'t help but think it\'s just being nice to me so I\'ll buy stuff.  '),(15393,3523319,'09,June,2004','2004-06-09 00:00:00','   Blimey, I took that ink blot test, and I was quite impressed with its standard summary of my personality.  Has anyone else taken that one? It\'s good.  '),(15394,3523319,'09,June,2004','2004-06-09 00:00:00','   Yes, all he will be.  I did the IQ test just now, and got the same score as dave, which i am assuming means we both got them all right:  &quot;Your IQ score is 136   This number is based on a scientific formula that compares how many questions you answered correctly on the Classic IQ Test relative to others.   Your Intellectual Type is Visionary Philosopher. This means you are highly intelligent and have a powerful mix of skills and insight that can be applied in a variety of different ways. Like Plato, your exceptional math and verbal skills make you very adept at explaining things to others â€” and at anticipating and predicting patterns. And that\'s just some of what we know about you from your IQ results.&quot;  I think someone should try and beat this score.  You might not be able to post at home because you are not logged in.  Try logging in and making sure cookies are turned on in your browser.  Also, check your Internet Options to make sure your SSL Security boxes are checked.  '),(15395,3523319,'09,June,2004','2004-06-09 00:00:00','   OK, good  How come when I\'m on my \'puter at home this won\'t let me post anything.  Also, who else thinks that Bobby Pires is going to be diving his ass off on sunday in an attempt to get the little ginger lad sent off.  Just a thought  '),(15396,3523319,'09,June,2004','2004-06-09 00:00:00','   is this working yet  '),(15397,3523319,'09,June,2004','2004-06-09 00:00:00','   i\'m in a bad mood today. and my guts are wrong. it\'s a super-good day.  '),(15398,3523319,'09,June,2004','2004-06-09 00:00:00','   because they\'re true  '),(15399,3523319,'09,June,2004','2004-06-09 00:00:00','   These  urlLink Gary Neville Dairies  are hilarious.  Well, funny anyway.  '),(15400,3523319,'09,June,2004','2004-06-09 00:00:00','   young man  '),(15401,3523319,'09,June,2004','2004-06-09 00:00:00','   I know no hutchtemmel. I just copied the link from Dave\'s post.  '),(15402,3523319,'09,June,2004','2004-06-09 00:00:00','   Has Ross finally learned some hutchtemmel?  Dave, can\'t you just explain to your FM that you\'re not in a client-facing role, and therefore can dress like a sack of shit if you feel like it?  And also explain why it would be a good idea for him to put his face in a blender.  '),(15403,3523319,'09,June,2004','2004-06-09 00:00:00','   That\'s one sleepy Bambos.  I have a Linksys BEFSR41 v2. All I had to do was upgrade the firmware. Unfortunately, Linksys have designed their site such that it\'s actually impossible to find anything. Consequently, I was under the mistaken belief that the last upgrade was this time last year (as that\'s what the site told me), when in reality there had been two since then.  Routers are very interesting.  Also -  urlLink BT finally decides to ditch the shitty old switching phone network .  '),(15404,3523319,'09,June,2004','2004-06-09 00:00:00','   \"I have nothing to say of my working life, only that a tie is a noose, and inverted though it is, it will hang a man nonetheless if he\'s not careful.\"    Yann Martel, Life of Pi'),(15405,3523319,'09,June,2004','2004-06-09 00:00:00','   tuesday? okidoke. i reserve the right to flounce, though.  i\'ve finished the innernet too. i might read the Komplete Kevinisation Game Guide, but i\'m not sure i have the energy...  '),(15406,3523319,'09,June,2004','2004-06-09 00:00:00','   I\'ve spent all morning reading the internet, which was great.  But now I have nothing to read at lunch time.  What a miscalculation.  What will I do this afternoon too?  My website has stopped working.  This does not please me.  '),(15407,3523319,'09,June,2004','2004-06-09 00:00:00','   Dave...next Tuesday...Angel...lots of Booze...me an AJ and a Lexie.  Be there or be a raving homosexual.  (You can come along and still be a raving homosexual if that makes things easier for you)  '),(15408,3523319,'09,June,2004','2004-06-09 00:00:00','   what kind of router do you have, and what did you need to do to it?  bambos set up our router, and it works pretty well as far as i know. i think if it had been left to me and yove, one of us would be dead by now, and the other would be sobbing in a corner somewhere. the router would still be in its box, i imagine.  mind you, that\'s the same bambos that fell asleep on the victoria line at euston and woke up in seven sisters. and it was going south when he got on.   urlLink see where he went wrong  '),(15409,3523319,'09,June,2004','2004-06-09 00:00:00','   Today\'s headlines are indeed funny. I also like the Statshot. \'Who\'s number 1?\' - \'Number 6\'.  Hur hur hur.  I\'ve already typed this message once, but my stupid gay interweb connection keeps logging me out of Blogger, which seems to wang things up. Bastards.  '),(15410,3523319,'09,June,2004','2004-06-09 00:00:00','   I am here. I have been working solidly all morning. Bah.  I have found that, since I gave my router a seeing-to, all my interweb games work properly. This pleases me, in a very compsci way.  '),(15411,3523319,'09,June,2004','2004-06-09 00:00:00','   Ties are gay.  I like being able to come into work in whatever I like, although it makes it harder to distinguish between weekdays and weekends.  I think most of Britney\'s knee problems could be improved by letting me massage her.  Of course, sometimes the area that is sore isn\'t the area you need to massage, which I would explain to her as I fondled he boobs and diddled her nubbin.    I\'ve gone over to the bad place.  '),(15412,3523319,'09,June,2004','2004-06-09 00:00:00','   i see britney\'s had knee surgery again. i wonder how big her knees will get this time?  i\'m listening to rage against the machine to fuel my misdirected adolescent angst.  '),(15413,3523319,'09,June,2004','2004-06-09 00:00:00','   I also had Italy beating France in the semi\'s.  But then losing the final narrowly to the the Portugays  '),(15414,3523319,'09,June,2004','2004-06-09 00:00:00','   i just got told that i have to wear a tie at work again. it really, really annoys me. i\'m not sure why, but it does.  i wunt to be a rebellious teenager again. not that i ever was in the first place.  '),(15415,3523319,'09,June,2004','2004-06-09 00:00:00','   According to my score predictions, it all goes swimmingly in the group stages.  Then England go out to Spain in the Quarter Finals (on penalties after a 2-2 draw) and Italy go on to win, having narrowly beaten France in the semis.  '),(15416,3523319,'09,June,2004','2004-06-09 00:00:00','   I especially like \'Leno\'s Voicemail Message Pauses For Laughter\'  '),(15417,3523319,'09,June,2004','2004-06-09 00:00:00','   the onion headlines this week are hilarious. i doubt the text of the articles can now match my expectations, so i don\'t think i\'ll read them.  how\'s ross today? any progress with the lynchmob?  '),(15418,3523319,'09,June,2004','2004-06-09 00:00:00','   I\'ve utilised the BBC score predictor.  I now know everything.  England out in the quarters unfortunatley  '),(15419,3523319,'09,June,2004','2004-06-09 00:00:00','   there\'s a little score predictor thing on the bbc sport euro web thing. it\'s quite interesting, in a way my job totally isn\'t.  '),(15420,3523319,'09,June,2004','2004-06-09 00:00:00','   The personal league thing doesn\'t seem to work on it though...oh well  '),(15421,3523319,'09,June,2004','2004-06-09 00:00:00','   I\'ve already got my wallchart up and on Saturday I will purchase many beer  Theres no games during the day, This pleases me.  The earliest any of them start is 5pm and thats only really a handful of games.  The rest are all at sensible o\'clock  '),(15422,3523319,'09,June,2004','2004-06-09 00:00:00','   Too late for you guys but I\'ve just joined a league with mates on this site http://en.fantasy.euro2004.com/  It\'s got much better scoring with assists, clean sheets, cards, defensive clean sheets, etc. Only missing thing is choosing a manager.  If anyone wants to start on this one let me know as I\'ve started a seperate league we can use when I was mucking about.  '),(15423,3523319,'09,June,2004','2004-06-09 00:00:00','   I made the change.  Did you notice all three of us have opted for the same manager?  That should ensure some pretty bad luck, right there.  I put up my World Cup Wall Chart the other day.  I know technically it\'s a Euro 2004 Wall Chart, but that doesn\'t have the same ring to it.  So I shall be refering to it as my World Cup Wall Chart anyway.  '),(15424,3523319,'09,June,2004','2004-06-09 00:00:00','   i don\'t reckon van horsetelrooy will score very many goals in the tournament.  whoops. i\'ve only gone and said it. i might as well fedex him the golden boot right now.  i\'m trying to drink coffee but i\'m no good at it.  '),(15425,3523319,'09,June,2004','2004-06-09 00:00:00','   The way I see it is that Becks is more likely to score than Gerrard full stop. (Frank Lampost is more likely to score than Stevie G) Becks is also less likely to get broken/sent off in a rage of red-misted headwrong  I\'ve decided than I\'ll keep my team as it is, downgrading my Wesley to a Latvian or a Stelios would only allow to upgrade my Torres to a Morientes/Trezu-Gay and not all the way up to a Ruud/Thierry.  Thats not a good enough trade off  '),(15426,3523319,'09,June,2004','2004-06-09 00:00:00','   i had a ross-inspired dream once. it ended with jets firing rockets past skyscrapers into this enormous pair of legs that was terrorising tokyo.  i finally found  urlLink yoav\'s purpose .  i stopped playing kevinisation at half ten last night, and managed to get to bed on time. and then i had a phone call at about 12.45. it was a cheerful greek, who said \"i fell asleep on the tube and i\'m stuck in seven sisters. i can\'t get back to balham. can i come and stay at yours?\". i said yes.  then there was another call a minute later: \"do you know any taxi numbers? there\'s only one driver here and he\'s a mentalist who doesn\'t know where camden is\". i gave him the only taxi number i know.  then another call: \"the taxi number\'s no good, there\'s no-one this side of london. none of the bus drivers knows which bus i need. help!\". i gave him the number for the transport helpline.  at 2.45am, i had a bambos on my doorstep.  this morning, i mostly want to die.  '),(15427,3523319,'09,June,2004','2004-06-09 00:00:00','   The way I see it is that Gerrard\'s more likely to score from open play, and is the braver choice, but Beckham takes all our penalties and free kicks, and with Mickey on the pitch we\'re bound to get a penalty or two...  I\'m erring on the side of swapping Gerrard for Becks.  I haven\'t seen your formation manx, but you do need midfielders who are going to score.  They also get an extra point for scoring over a striker.  So the decision is yours really.  Have you signed up to the league?  I had a dream that Dolly\'s house kept getting burgled last night.  It got burgled once, and she was too afraid to stay there, so she left the house and stayed at someone else\'s.  So the burglars broke back into it again.  The next day the same thing happened, but this time she left her keys in the door, and the burglars took EVERYTHING.  I suppose it was a Ross-inspired dream on some level.  '),(15428,3523319,'19,June,2004','2004-06-19 00:00:00','   I feel a sort of rosy glow.  and while that may be partly due to sunburn and wine I think its partly due to some good footy  '),(15429,3523319,'18,June,2004','2004-06-18 00:00:00','   Yeah, who needs decent defenders?  I can\'t decide who looks more alike: Pavel Nedved and Patrick Swayze from Point Break, or Japp Stam and Jan Koller...  '),(15430,3523319,'18,June,2004','2004-06-18 00:00:00','   Blinkin heck.  footballs alright when its like that  '),(15431,3523319,'18,June,2004','2004-06-18 00:00:00','   I\'ve finally scored a goal in Euro Fantasy Football!  The scorer?  Filthy Jan Koller.  '),(15432,3523319,'18,June,2004','2004-06-18 00:00:00','   I\'ve been outside all day and due to the sun am now literally a redneck  '),(15433,3523319,'18,June,2004','2004-06-18 00:00:00','   I was just cruising the JobServe website, and I saw a job that included this line:   You will also have extensive knowledge of Netscape Navigator and Microsoft Internal Explorer.   Tell me that\'s not a real browser.  '),(15434,3523319,'18,June,2004','2004-06-18 00:00:00','   Go and meet an aristocrat for drinks then go out to a club and get hammered...that\'s what I am doing...nice  '),(15435,3523319,'18,June,2004','2004-06-18 00:00:00','   Well, I\'ll be sure to check that one out.  You know, when I run out of better things to watch.  I feel like I\'m wasting my weekend, but I simply can\'t be arsed to do ANYTHING.  I sat in my chair for so long earlier my arsecheeks started to ache.  '),(15436,3523319,'18,June,2004','2004-06-18 00:00:00','   It even has someone known as \"The One\"...actually it\'s fucking shit but I was too hungover to move  Warriors of Virtue: The Return to Tao  '),(15437,3523319,'18,June,2004','2004-06-18 00:00:00','   I\'ve been toppled from the top of our mini-league...goddamn...I\'m not making any changes for the next couple of matches now...change of tactics.  '),(15438,3523319,'18,June,2004','2004-06-18 00:00:00','   has anyone seen  urlLink this   nice to see where our license fee is going.  those of that pay it, dave  '),(15439,3523319,'17,June,2004','2004-06-17 00:00:00','   urlLink here .  '),(15440,3523319,'17,June,2004','2004-06-17 00:00:00','   No, it\'s gays.  '),(15441,3523319,'17,June,2004','2004-06-17 00:00:00','   i need you to post that riddle here again. i think i may be onto something, but i need to check the semantics.  where is everyone this afternoon?  '),(15442,3523319,'17,June,2004','2004-06-17 00:00:00','   ooh, now it\'s emule at the top.  '),(15443,3523319,'17,June,2004','2004-06-17 00:00:00','   all the ads at the top are advertising gayness at me. we\'ve broken google. i knew this would happen.  '),(15444,3523319,'17,June,2004','2004-06-17 00:00:00','   in every way  '),(15445,3523319,'17,June,2004','2004-06-17 00:00:00','   From the BBC\'s coverage of France/Croatia on Thursday...   Ian Wright: \"Prso works hard for a big man.\"   Following on from Andy Townsend\'s insightful \"He\'s quite sleepy for a big man,\" Mediawatch is happy to learn that not only are our lanky friends insomniacs, but workshy as well.  Daves a big man  '),(15446,3523319,'17,June,2004','2004-06-17 00:00:00','   hur hur  entry level  '),(15447,3523319,'17,June,2004','2004-06-17 00:00:00','   I recognised about half a dozen words in that. I fully deserve my entry-level Japanese certificate.  '),(15448,3523319,'17,June,2004','2004-06-17 00:00:00','   i could tell you, but i\'d have to invade poland.  Suba-ra-shi chin chin mono Kintama no kame aru Sore no oto saru bo bo Iye! Ninja ga imasuuuuuuuu Hey hey let\'s go kenka suru Taisetsu no mono protect my balls! Boku ga warui so let\'s fighting... Let\'s fighting love! Let\'s fighting love! Kono uta chotto baka Wake ga wakaranai Eigo ga mecha-kucha Daijobu - we do it all the time!  '),(15449,3523319,'17,June,2004','2004-06-17 00:00:00','   What does racht mean?  '),(15450,3523319,'17,June,2004','2004-06-17 00:00:00','   it does tend to create little CVS folders around the shop, but i think there\'s a way to configure it to stop it. ultimately, it\'s the open source industry standard, so it\'s free and there will be people out there who know an offal lot more than i do.  es ist zeit fur racht.  '),(15451,3523319,'17,June,2004','2004-06-17 00:00:00','   hmmm, fish dumplings  '),(15452,3523319,'17,June,2004','2004-06-17 00:00:00','   Dave, a CVS Question.  My techie man says its good but it leaves little hidden files everywhere. What are your opinions on this ? I can see someone deleting a local file and everything getting screwed up.  '),(15453,3523319,'17,June,2004','2004-06-17 00:00:00','   i have also swotted like a monster. it\'s colder today, which is good.  hang on, we\'ve not had the weekend yet. bonus.  my eclipse workspace has buggered itself, which is less amusing than it could have been.  '),(15454,3523319,'17,June,2004','2004-06-17 00:00:00','   i did a lot of sweating this weekend  '),(15455,3523319,'17,June,2004','2004-06-17 00:00:00','   I\'ve got a new computer at work, and no time to set it up.  And I\'m not at work for the next two weeks, so when I get back I\'ll probably find both my old and new computers have been taken away by the CIT people.  Bah.  '),(15456,3523319,'17,June,2004','2004-06-17 00:00:00','   there\'s definitely something wrong with this pc. it\'s all over the place, and it won\'t let me close excel.  '),(15457,3523319,'17,June,2004','2004-06-17 00:00:00','   Is it me - or is everything a bit wrong today... discuss.  '),(15458,3523319,'17,June,2004','2004-06-17 00:00:00','   Well, that\'s made it interesting.  You know, if we draw with Croatia and Switzerland beat France (unlikely I know), I have no idea who goes through...  '),(15459,3523319,'16,June,2004','2004-06-16 00:00:00','   It\'s for the tears man...for the tears  '),(15460,3523319,'16,June,2004','2004-06-16 00:00:00','   hee hee hee hee  laminated book of dreams  tee hee hee hee hee hee hee  '),(15461,3523319,'16,June,2004','2004-06-16 00:00:00','   Yeah, Dave often does that.  '),(15462,3523319,'16,June,2004','2004-06-16 00:00:00','   Billius Ballius was very funny...his pwartishead take off was bwilliant I have to say but lost on 90% of the audience.  There was some idiot in front of us who clearly didn\'t get half the jokes...he tended to laugh about 3 or 4 seconds after everyone else started.  '),(15463,3523319,'16,June,2004','2004-06-16 00:00:00','   I\'m glad you like it.  It will be even better when Dave turns his designer\'s eye to it and improves it.  That film 7 thing is a bit of a fuck up.  It\'s not supposed to do that.  Hmm.  '),(15464,3523319,'16,June,2004','2004-06-16 00:00:00','   I quite like your site, Fonz. Make some more of it work, so I can comment.  I especially liked the review of film 7, as well.  '),(15465,3523319,'15,June,2004','2004-06-15 00:00:00','   i dun read the reviews earlier and forgot to mention it. they weren\'t as Contentious as i\'d hoped. or were they?  look at me, i\'ve bagged me a boamboz. who has the Best Profile Ever.  '),(15466,3523319,'15,June,2004','2004-06-15 00:00:00','   is this going to work ?  '),(15467,3523319,'15,June,2004','2004-06-15 00:00:00','   Right, my website has started working again. I JUST looked at it.  So, can someone go to  http://www.topfivefilms.com/sampledb.jsp?filmid=  that address, and put any number from 0 to 6 on the end, and it should work.  I\'m reluctant to put a proper link on here because it seems to shaft my site for some reason.  '),(15468,3523319,'15,June,2004','2004-06-15 00:00:00','   i demand that ross play that game until he is intel\'s mindslave. shouldn\'t take long.  '),(15469,3523319,'15,June,2004','2004-06-15 00:00:00','   Dude, that games pretty bad, but nowhere near as bad as a game Oracle made about Toplink portals.  (Un)fortunately, I can\'t find that game now, it may have been binned.  '),(15470,3523319,'15,June,2004','2004-06-15 00:00:00','   i\'ve found the least fun game on the innerweb:   urlLink intel have wasted an offal lot of money on this  '),(15471,3523319,'15,June,2004','2004-06-15 00:00:00','   you know what would be really funny? if we killed bill after the show tonight, your colleagues wouldn\'t be able to see him, and it would be like that film, about killing bill, whatever it was called. so it works on two levels, see?  and that\'s why i\'m called \"comedy genius\" dave.  '),(15472,3523319,'15,June,2004','2004-06-15 00:00:00','   Quite a few people from work are seeing Bill on Friday and Saturday.  I\'ll be able to spoil it for them. ha ha ha  '),(15473,3523319,'15,June,2004','2004-06-15 00:00:00','   ilk.  '),(15474,3523319,'15,June,2004','2004-06-15 00:00:00','   But I\'m only 26% gay! I blame damo and his ilk for letting the side down.  '),(15475,3523319,'15,June,2004','2004-06-15 00:00:00','   No, of course you didn\'t get up before me Dave. You missed by about half an hour.  I predict that you will enjoy Bill Bailey. It is oolarious. He does a very credible impression of Portishead, too.  I\'ve already typed this message once, only to be unceremoniously logged out from Blogger, and have my message lost. I hate Websense more than Jesus.    And he hates it a lot.  '),(15476,3523319,'15,June,2004','2004-06-15 00:00:00','   it\'s finally  urlLink here . mmmmmmmgoodyumyesnice.  i never look at the ads, but i suppose they\'re google context-sensitive efforts. and we\'ve mentioned mckellen and his ilk a fair bit here, so god knows what the old googlometer\'s made of it all.  '),(15477,3523319,'15,June,2004','2004-06-15 00:00:00','   Why is there an advert for gays up at the top of this page?  I bet this is really why Ross wanted to move to a blog.  '),(15478,3523319,'15,June,2004','2004-06-15 00:00:00','   I\'m looking forward to bill...he will be good  '),(15479,3523319,'15,June,2004','2004-06-15 00:00:00','   The overnet connection isn\'t of much use unless you are sharing over 20 odd gigs...overnet is a series of hubs which have minimum shares.  Anything of worth is usually only found on servers over 20 gigs...hence if you aren\'t sharing enough...the hub won\'t feed you anything. Overnet is pretty pants anyway...I used it a year ago or so for a couple of months...fastish connection speeds but you always get cut off by the people you download from if you ever get going too fast.  Not a problem for small files but for films and isos etc. it\'s a real pain.  '),(15480,3523319,'15,June,2004','2004-06-15 00:00:00','     '),(15481,3523319,'15,June,2004','2004-06-15 00:00:00','   For the record, eDonkey2000 came first, and connects to Overnet and another ed2k server.  eMule came second, and seems to run better and look nicer, but it only connects to one ed2k server and not overnet.  Og course, they both have updates released regularly, so neither of them are older and therefore worse.  I\'m not sure which is better, though.  Is it lunch time yet?  '),(15482,3523319,'15,June,2004','2004-06-15 00:00:00','   I wouldn\'t say that Greece knoebbed Portugal.  Portugal just played really badly and Greece gave them a good Germaning.    Mind you, they could well do the same to England  '),(15483,3523319,'15,June,2004','2004-06-15 00:00:00','   If you did, wouldn\'t you know which one came most recently and which was better? I put it to the honourable gentleman that he knows as little as Dave or I...either way, they both smell of poo and wee and stuff. I am phoning Aiden now to tell him to turn up at Hammersmith now...not now...at 7.30...you know what I mean...if you don\'t...it will be ok...I promise  '),(15484,3523319,'15,June,2004','2004-06-15 00:00:00','   Having listened to your collective eDonkey/eMule ramblings, I\'ve come to the conclusion that I am the most knowledgable on the subject.  I\'m still not sure who I WANT England to face if we get to the Quarter Finals.  My inkling is Greece, but then they gave us a run for our money a few years back and dicked Portugal the other day, didn\'t they?  I don\'t really want to come up against Spain, even though they bottle it a lot.  Maybe Portugal is a handy middle road...  '),(15485,3523319,'15,June,2004','2004-06-15 00:00:00','   So now we meet in Hammersmith rather than \'somewhere central\'  Thats fine by me to be honest  '),(15486,3523319,'15,June,2004','2004-06-15 00:00:00','   You made me leave the pub with half a jug of beer still left - A sacrilige that you will be made to answer for sometime soon! Emule/Edonkey...one came before the other the latter being based on the prior.  Find out which came first and you will have your answer...  '),(15487,3523319,'15,June,2004','2004-06-15 00:00:00','   did i get up before you today, ross? i demand answers!  i\'m super-glad that i have tomorrow off. i think i almost passed out twice earlier, presumably from fatigue. although i did drink half a beer last night, so that could be it.  '),(15488,3523319,'15,June,2004','2004-06-15 00:00:00','   i\'ve used emule, and found it prettier. i\'ve no idea how performance compares, though.  we face the winner of group a (is it? the one with greece, spain, russia and portugal), assuming we come second in the group. so it looks like spain at the moment, although if greece beat them, it may well be bambos\' buddies.  i\'m finishing work at half five tonight, going home, getting changed, pouring some kind of food into my mouth, then heading for hammersmith about 6.45, i imagine. should arrive a little before half seven, in time to pull aiden out of whichever establishment of ill-repute he\'s accidentally wandered into, pick up the tickets, resuscitate damo (with those little heart-paddles that george clooney uses) and take our seats in time to hear that the show\'s cancelled because it\'s too hot or something.  that\'s my plan.  '),(15489,3523319,'15,June,2004','2004-06-15 00:00:00','   You know...M-Games have probably been summoned to the courts in America as a result of Yoav\'s failure to answer any phone calls from Wireless Developer/Airbourne but unfortunately the letter has been lost in a pile of papers on Wobu\'s floor.  You may think I\'m joking...I\'m not.    Living in ignorance is not intelligent...and yet it is those who think themselves intelligent who are most ignorant.    I just made that up...nice.  '),(15490,3523319,'15,June,2004','2004-06-15 00:00:00','   I think thats a promise that Joe just can\'t make, or at least keep.  To get to Hammersmith my research tells me you go to Hammersmith tube (which was as big a shock to me as i\'m sure it is to you)  This lies on the Hammersmith & City, Piccadilly, and District lines.  I expect the Apollo should be pretty clearly signposted from the tube.  So, whats the plan  '),(15491,3523319,'15,June,2004','2004-06-15 00:00:00','   I moved over form the eDonkey to the eMule last night.  I am undecided about the change so far. eMule doesn\'t connect to the Overnet server, does it?  So who do we want England to face in the quarter finals if they ever get that far?  I demand answers.  '),(15492,3523319,'15,June,2004','2004-06-15 00:00:00','   david james = lexie when it comes to football. a massive case of the retardation.  our tickets will be waiting for us at the box office tonight. they sent them out in may. i bet woab\'s swallowed them accidentally. does anyone know how to get to the hammersmith apollo? we\'ll need to get there for 7.30 or so, i imagine. someone ask aiden.  i\'ll be going to brum for the cricket day, as long as joe promises not to try to put his finger up my bottom when i\'m asleep.  '),(15493,3523319,'15,June,2004','2004-06-15 00:00:00','   Yeah, I\'ve seen a clip of that on TV.  I have no idea what James thought he was doing standing in the middle of the goal there. Becks, as it turned out, knew best.  '),(15494,3523319,'15,June,2004','2004-06-15 00:00:00','   Dave, check  urlLink this link  out   Half way down on the right is a picture of Beckham in the wall pointing to the right - The fateful moment I mentioned last night where he shouted at Cole to get back on the line and every time Beckham turned his back, stupid David James sent him back to the wall...need to find a movie of it now.  '),(15495,3523319,'15,June,2004','2004-06-15 00:00:00','   The World Health Organisation are going? Crazy.  I can\'t go, I have to be in court at that time.  Yes, you heard me.  '),(15496,3523319,'15,June,2004','2004-06-15 00:00:00','   Who\'s going to Joeys Big brummie piss-up.  WHO  '),(15497,3523319,'15,June,2004','2004-06-15 00:00:00','   there is a shower. but only one, and it turns out that people use it all the time. i\'m not sure i should be nipping off for a shower after 9.30, but seeing as i arrived in plenty of time before work and was denied my opportunity, i don\'t see that i have any choice. it\'s either shower now or smell bad until lunchtime.  no sign of the tickets, but i have a nagging feeling that the envelope containing them is sitting under a stack of papers on yoav\'s desk somewhere... like that old insurance letter that they still talk about in hushed tones at the back of pubs. i\'m calling the customer service line as soon as it opens - at half ten.  '),(15498,3523319,'15,June,2004','2004-06-15 00:00:00','   Is this a shower at work?  Christ, you have it easy.  I wish I got paid to have a shower and go to the gym.  Anyway, here\'s the real question:  IF England don\'t fuck up and manage to get to the quarter finals, which team from Group A do we want to come up against?  I need to decide so I know who to support this evening.  '),(15499,3523319,'15,June,2004','2004-06-15 00:00:00','   portugal are going to get knocked out tonight. we shall see, say i.  the shower is still in use. i\'ve been here since five past nine, and there\'s been some MOTHER-FUCKER in the shower every time i\'ve tried it. it\'s enough to make me very angry indeed. and even smellierer.  '),(15500,3523319,'15,June,2004','2004-06-15 00:00:00','   and i mostly stank of beer. i\'m going for a shower now.  '),(15501,3523319,'15,June,2004','2004-06-15 00:00:00','   Yeah, that was my favourite bit too. Gotta love that compski.  '),(15502,3523319,'15,June,2004','2004-06-15 00:00:00','   I done finally looked at that 16k thing.  for n=0 to 2   Those were the days next n  Now that\'s some funny compsci shit right there.  '),(15503,3523319,'15,June,2004','2004-06-15 00:00:00','   You got them tickets yet Dave?  '),(15504,3523319,'14,June,2004','2004-06-14 00:00:00','   He\'s looking to score some streaker poontang! Nice!  '),(15505,3523319,'14,June,2004','2004-06-14 00:00:00','   We were losing the game 1-0. She was on there a long time and nobody seemed to want to do anything about it. In the end it was a case of get over there and push her off the pitch...     ...the stewards and police were letting her get on with it and they seemed to be quite happy to stand there and look at her. I\'m going to send her a letter to say thank you because she fired me up a lot and a few minutes later I scored our equalising goal!   Gillingham manager Andy Hessenthaler thanks a streaker for her part in the Gills\' win over Wimbledon.   Yes, that\'s the kind of manager we have.  '),(15506,3523319,'14,June,2004','2004-06-14 00:00:00','   i\'ve had enough, i\'m going to the gyme, and thence to angel, to eat, drink and be miserable in a corner somewhere.  '),(15507,3523319,'14,June,2004','2004-06-14 00:00:00','   I just carefully negotiated my right to leave early on Thursday so I can get home in time for England\'s match, and them I remembered I\'m sharing a car with my mum this week, and she\'ll probably have to work til 5:30.  Dude, that is so fucking weak.  '),(15508,3523319,'14,June,2004','2004-06-14 00:00:00','   i just found a muddy little stain on the front of my shirt, and i\'ve no idea how long it\'s been there. i bet people have been staring at it all day, and thinking to themselves that i must enjoy dripping poo on myself in the morning.  i bet that\'s exactly what they\'ve been thinking. either that, or \"look at the spistock who can\'t drink coffee properly\". my money is on the former.  '),(15509,3523319,'14,June,2004','2004-06-14 00:00:00','   I did give a Mighty Boosh explanation, but then my stupid gay interweb connection lost it for me. Thanks, Websense.  Anyway - Boosh: Funny, BBC3 tonight at 22:30, or Monday, 22:30 on BBC2.  '),(15510,3523319,'14,June,2004','2004-06-14 00:00:00','   Be the one!!!!!!!  '),(15511,3523319,'14,June,2004','2004-06-14 00:00:00','   i concur  '),(15512,3523319,'14,June,2004','2004-06-14 00:00:00','   I have no idea what a mighty boosh is, and I am not impressed with your efforts to sell it to me.  I would at least expect a short description, the channel it is on and the time it airs.  Also, too.  '),(15513,3523319,'14,June,2004','2004-06-14 00:00:00','   i also think so too.  '),(15514,3523319,'14,June,2004','2004-06-14 00:00:00','   Also, I think I also put too many \'also\'s\' and \'too\'s\' in that last post, too.  '),(15515,3523319,'14,June,2004','2004-06-14 00:00:00','   It is indeed Helen\'s birthday today. You must be some sort of memory man. Let\'s all have a rousing chorus of that song... you know, the one to wish someone a happy birthday... what\'s it called again?  I hope some of you bummers have been watching The Mighty Boosh. It\'s veddy funny. Also, I Am Not An Animal is good, too. Although it\'s now finished, so I guess it was good.  '),(15516,3523319,'14,June,2004','2004-06-14 00:00:00','   it must have broken the moment you set up a Conduit of Evil from this page, which would have swamped it with computofilth in seconds.  '),(15517,3523319,'14,June,2004','2004-06-14 00:00:00','   I don\'t understand this at all. I just checked my site, it was all working beautifully.  I pasted a link on to this page, and when I clicked on it, my site had stopped working.  It\'s not working now either.  What the fuck is that about?  '),(15518,3523319,'14,June,2004','2004-06-14 00:00:00','   how\'s it going, wee man? any luck with the insurance claiming?  stop that working nonsense. people will just expect it of you in future.  is it helen\'s birthday today? if so, happy birthday helen. if not, happy birthday ross.  '),(15519,3523319,'14,June,2004','2004-06-14 00:00:00','   I looked at that 16k thingy. It looked entertaining, but I have no speakers, so I\'ll have to have another look when I get home.  '),(15520,3523319,'14,June,2004','2004-06-14 00:00:00','   I\'m not the Paedo! It\'s Joey...I\'m the one with the sick jokes...Joe\'s the one that talks about shagging children seriously.  '),(15521,3523319,'14,June,2004','2004-06-14 00:00:00','   i find myself in the same boat as damo. it\'s a bit like life of pi, only with a paedophile instead of a tiger.  '),(15522,3523319,'14,June,2004','2004-06-14 00:00:00','   Neither of those links work for me...I feel cheated and lied to...I may well have a little cry  '),(15523,3523319,'14,June,2004','2004-06-14 00:00:00','   Yay, my website is working again.  My  urlLink Larry Botter review  is on it if anyone is interested too. And a really old review of  urlLink 8 Mile .  '),(15524,3523319,'14,June,2004','2004-06-14 00:00:00','   Well you wouldn\'t until all chance of meeting England again in the championship were gone....wait a minute  '),(15525,3523319,'14,June,2004','2004-06-14 00:00:00','   i see bobby pires has very kindly decided not to rub his arsenal colleagues\' faces in that consummate french victory. he\'s a generous chap, isn\'t he?  i appear to be short of a wee pseudoscotsman this morning. and to think i went to the trouble of making that link for him. where be he?  '),(15526,3523319,'14,June,2004','2004-06-14 00:00:00','   Any transfers you make should be in the Czechs I reckon...German/Holland will not be high scoring, nor is it safe to predict either will keep a clean sheet. It\'s kind of hard to keep track of what points you should have though because of managers/cards/goals/clean sheets...even if there is something slightly wrong with the calculation - we all \"suffer\" from it.  '),(15527,3523319,'14,June,2004','2004-06-14 00:00:00','   last night that i went on graham norton\'s show, and spent the whole time telling him what a big gay bummer he was, to the delight of the audience and much to his amusement. it all seemed fine and funny at the time, but in retrospect i must have come across a terrible homophobe.  see what i did there?  hur hur gayers.  '),(15528,3523319,'26,June,2004','2004-06-26 00:00:00','   Blimey did the French lose last night?  That\'s a turn up for the books.  I\'m having a few days off from football to get over the terrible, crushing disappointment.  I wish I didn\'t wake up at 7:50 am every day, regardless of when I need to be up.  '),(15529,3523319,'25,June,2004','2004-06-25 00:00:00','   Au revoir Froggies  I\'m in a kind of unique situation in that I really don\'t dislike any of the teams still left in this tournament ans wouldn\'t mind it if any of them won.  I can\'t remember that ever happening before  '),(15530,3523319,'24,June,2004','2004-06-24 00:00:00','   This page is really gaying me up today.  '),(15531,3523319,'24,June,2004','2004-06-24 00:00:00','   I hadn\'t heard that.  I have now.  I\'ve run out of things to do now really.  Can I be arsed to hang around for another half an hour?  There is only me and some weird, low-paid, religious nut who breathes to loudly and talks too quietly left here.  I think I\'ll go. I want to complete Splinter Cell.  '),(15532,3523319,'24,June,2004','2004-06-24 00:00:00','   \"Posh Spice was confirmed to have broken the World Gang Bang record on Thursday night, when in 120 minutes her shaved c**t f**ked the entire nation.\"  '),(15533,3523319,'24,June,2004','2004-06-24 00:00:00','   has anyone not heard the posh spice gangbang joke?  '),(15534,3523319,'24,June,2004','2004-06-24 00:00:00','   professionalism. i\'ve had about 15 football emails today. we\'re a bitter, bitter nation, aren\'t we?  '),(15535,3523319,'24,June,2004','2004-06-24 00:00:00','   It\'s always a good idea.  I\'ve been needlessly productive today.  In fact, I\'d say I\'ve done a whole week\'s worth of work in my few hours here today.  But there\'s nothing else to do...  It seems wrong, yet it sounds so right.  I have one more job to do, but I can\'t charge for it if I do it today because I\'ve already charged out my six hours of time.  So I might leave it for m\'colleague to do on Monday, and spend the next hour an a half filling out a job application form instead.  '),(15536,3523319,'24,June,2004','2004-06-24 00:00:00','   i went to the pub  and now i\'m a bit silly.  might miss yoga tonight  smell like fags and booze. again.  '),(15537,3523319,'24,June,2004','2004-06-24 00:00:00','   Oh god, I have another 4 hours of this misery to go.  '),(15538,3523319,'24,June,2004','2004-06-24 00:00:00','   Well, I don\'t think I\'m supposed to talk about it, but it probably can\'t do any harm.  However, I think I\'ll wait until I\'ve been released for good (next week) before I spill my guts.  '),(15539,3523319,'24,June,2004','2004-06-24 00:00:00','   are you allowed to tell us about your trial of the century yet?  '),(15540,3523319,'24,June,2004','2004-06-24 00:00:00','   I have been away from work for four days.  Now I\'ve come back, I have remembered two things:  1) Avril on reception is fit and I want to muck up her lens.  2) Work is incredibly boring.  '),(15541,3523319,'24,June,2004','2004-06-24 00:00:00','   Yes, I enjoyed seeing Cole piss rings round the spotty little wanker.  Also, Neville played well the whole tournament.  Gary, not Phil, obviously.  '),(15542,3523319,'24,June,2004','2004-06-24 00:00:00','   and at least ashley cole pitied that fool ronaldo all night long.  '),(15543,3523319,'24,June,2004','2004-06-24 00:00:00','   and get rid of the jump  '),(15544,3523319,'24,June,2004','2004-06-24 00:00:00','   I appologise if this has posted twice, I can assure you it\'s not my fault.  ...  message ends.  '),(15545,3523319,'24,June,2004','2004-06-24 00:00:00','   Yeah, the dutch are going out in the next match anyway, so my support for them will be short lived.  Besides, I\'ve noticed while watching them that they are very entertaining - every so often they let what appears to be a horse on to the pitch and they just let it gallop at the goal.  From six yards out.  The context banner at the top seems to be about football now.  Oh yeah, dave, I\'ve noticed that when I load this page and scroll down, it suddenly jumps back to your anchor tag at the top.  So I have to scroll down again. I thought perhaps I had a #top on the end of my bookmark, but I don\'t.  So it must be your fault that this happens.  Make it stop doing that to me.  '),(15546,3523319,'24,June,2004','2004-06-24 00:00:00','   my first reaction was to go with the dutch, as i\'ve done in every competition since 1998 after we\'ve gone out (\"dennish bergkamp! dennish bergkamp! DENNISH BERGKAMP!\" etc), but then i remembered that the world\'s only other diving horse plays for them, and i don\'t want him to win anything.  so now i\'m an honourary czech. go praha. go pavel.  '),(15547,3523319,'24,June,2004','2004-06-24 00:00:00','   I couldn\'t post a message so have sent him an email instead...told him it was 12 against 11 last night and that I had awarded him the Man of the Match award - I have never seen a player influence a match so much.  '),(15548,3523319,'24,June,2004','2004-06-24 00:00:00','   www.ursmeier.ch/referee/ref60.html  He\'s got his own webshite, leave the fucker some feedback  '),(15549,3523319,'24,June,2004','2004-06-24 00:00:00','   I\'m siding with the dutch and the Czechs, just as soon as I\'ve assassinated a certain Swiss referee.  '),(15550,3523319,'24,June,2004','2004-06-24 00:00:00','   That was a negative, and what i need is 2 positives.  Oh well, at least all that \'eng-er-lund are gonna win\' bollocks is finished.  All we need is Henman to go out of Wimbledon as soon as hunmaly possible and we can have a relaxed fucking summer.  The weight of my support is now behind the Czechs  '),(15551,3523319,'23,June,2004','2004-06-23 00:00:00','   my fantasy football team is going to need some tinkering.  '),(15552,3523319,'23,June,2004','2004-06-23 00:00:00','   I\'m sticking with my line up of Anglais, Dutchies and Czechs.  They excite me  '),(15553,3523319,'23,June,2004','2004-06-23 00:00:00','   I\'ve gone all English tonight, no portu-gays. Call me Jingoistic if you will.  I\'ve got a few frogs, dutch and a swede too.  '),(15554,3523319,'23,June,2004','2004-06-23 00:00:00','   i\'ve got loads of dam\' portugesh in my side now.  '),(15555,3523319,'23,June,2004','2004-06-23 00:00:00','   i\'m sticking firmly with sweden tonight.  i can\'t afford any decent players anyway.  '),(15556,3523319,'23,June,2004','2004-06-23 00:00:00','   Indeed.  I have put a guy away.  I have to go back on Monday and do it all again though.  And I have to go to work tomorrow I think.  Also, Malaga is in Spain, not Greece, although I suspect damo will protest that he already knew that, but not offer an explanation of his Greece - Malaga comment.  Call me Nostradamus.  I need to tinker with my team for tonight\'s match... do I put my eggs in the England basket, or do I \"bet against\" them to ensure a result either way. Hmmm.  '),(15557,3523319,'23,June,2004','2004-06-23 00:00:00','     '),(15558,3523319,'23,June,2004','2004-06-23 00:00:00','   i hear ya.  '),(15559,3523319,'23,June,2004','2004-06-23 00:00:00','   I see mediawatch has \'The Page That Would Get Excited If It Wasn\'t So Busy\' at the top the page today.  Thats me that is  '),(15560,3523319,'23,June,2004','2004-06-23 00:00:00','   i did it this morning, and that was fine. i might now skip the gym and knock off back to camden and change before the match. that way, i could eat some Original Home Cooking. which is to say, microwave nasty.  i need a bambos now. my guts are all wrong.  '),(15561,3523319,'23,June,2004','2004-06-23 00:00:00','   japs all top themselves  '),(15562,3523319,'23,June,2004','2004-06-23 00:00:00','   I suggest that we all get drunk and be merry tonight...all those in favour say aye. Dave, Yoga at 6.30 is not allowed tomorrow...I heard it on the radio  '),(15563,3523319,'23,June,2004','2004-06-23 00:00:00','   a wank i think.  where is ross? oh that\'s right, glastonbury. and fonz is putting someone away for bum-rape. oh, the irony.  i have work to do and am bored. is the rest of the country like this? is this why the japs are doing better? ARE the japs doing better?  ...  bored bored bored bored bored.  this must be how bambos feels all the time.  '),(15564,3523319,'23,June,2004','2004-06-23 00:00:00','   popbitch is also 3 steps short of what it promised.  '),(15565,3523319,'23,June,2004','2004-06-23 00:00:00','   POpbitch has lots of links in it today, none of which I can look at at work I\'m sure  '),(15566,3523319,'23,June,2004','2004-06-23 00:00:00','   Yeah, the last one was good fun...I\'ve decided I might get wasted...I\'ve invested heavily in the English as a sign of support.  Personally I think the French are where the points are at for the quarterfinals though...the Greeks are going to get hammered right in that place they enjoy so much...Malaga!  '),(15567,3523319,'23,June,2004','2004-06-23 00:00:00','   it was an offal lot better last time round. and with carwyn at our table, we can unleash him on anyone who irritates us this time.  '),(15568,3523319,'23,June,2004','2004-06-23 00:00:00','   I didn\'t really like that place.  It was a bit full of braying twats.  If I were you I would watch the match elsewhere  '),(15569,3523319,'23,June,2004','2004-06-23 00:00:00','   who\'s coming to ed\'s big gay partay at motion this evening? i\'m told that a monkey-wolf-fish-boy will be in attendance.  '),(15570,3523319,'23,June,2004','2004-06-23 00:00:00','   now i\'m nervous.  '),(15571,3523319,'23,June,2004','2004-06-23 00:00:00','   ok, i\'ve put some eggs in englands basket  '),(15572,3523319,'23,June,2004','2004-06-23 00:00:00','   I thought you had a bit of a shit day pointswise yesterday Dave.  I had four Dutchmen, including Van Nistelrooydonk.  I was bit gutted about the Czech second string, i was pretty sure Nedved would be rested but I expected Cech to play at least.   I actually like the look of my team now and am going to restrict my tinkering, i wouldn\'t mind a Portugeasy player or two but would that be disloyal?  I might transfer out Reizeger for an englishman maybe.  '),(15573,3523319,'23,June,2004','2004-06-23 00:00:00','   i can\'t believe that i forgot to tinker with my team yesterday. not a single dutchman. i\'ll never recover.  '),(15574,3523319,'23,June,2004','2004-06-23 00:00:00','   i\'m still disappointed that they played their second XI last night. i missed out on all the points i should have got.  right... portugese defenders for tonight\'s match, then?  '),(15575,3523319,'23,June,2004','2004-06-23 00:00:00','   I lurve the czechs.  I furking lurve them  '),(15576,3523319,'22,June,2004','2004-06-22 00:00:00','   That\'s weird, it looks better now.  It\'s an intermittent problem.  Also, I don\'t think the Ag monster is in that film, but one of the two main characters is 30.  '),(15577,3523319,'22,June,2004','2004-06-22 00:00:00','   I see.  Two more things:  This page is broken.  The bit on the right is all wrong in my browser here.  Secondly, does anyone know when Andy And Jenny\'s anniversary is?  '),(15578,3523319,'22,June,2004','2004-06-22 00:00:00','   although someone tole me that the aguilera monster was in it, and she\'s about 33. the main one is 17, as far as i know. which is legal in some states. anyway, you should never use that phrase, as what you need to say is:  \"...while definitely a girl film, there\'s plenty of eye candy from which the boys can form ejaculatory teenage fuck fantasies...\"  i have a new plan for the weekend. ed\'s driving up on saturday morning from epsom, so i\'m going to high-tail it down there pronto on saturday and join him in the capri-of-love all the way up the m40. assuming we go down the m40. as such, i won\'t be going up on friday anymore. in case that wasn\'t clear.  i\'m hungry.  '),(15579,3523319,'22,June,2004','2004-06-22 00:00:00','   I\'m writing a review of Mean Girls.  Do you think I\'m allowed to say the following:  \"...while definitely a girl film, there\'s plenty of eye candy for boys to feast their eyes on too...\"  Or does that just reveal to the world what a pervert I am? I think the girls in it play 15 year olds, but the actresses that play them are about 18 I think.  I\'m wrong, aren\'t I?  '),(15580,3523319,'22,June,2004','2004-06-22 00:00:00','   I didn\'t realise god had such a fondness for chips.  If I try to download that tube map thing, will it cost me Bucks, and will it be shit?  '),(15581,3523319,'22,June,2004','2004-06-22 00:00:00','   the old ones are the best ones, eh?  '),(15582,3523319,'22,June,2004','2004-06-22 00:00:00','   ha ha HA  '),(15583,3523319,'22,June,2004','2004-06-22 00:00:00','   Ronaldo, Luis Figo and Wayne Rooney are standing before God at the throne of Heaven. God looks at them and says; \"before granting you a place at my side, I must first ask you what you believe in.\" Addressing Ronaldo first he asks, \"what do you believe?\" Ronaldo looks God in the eye and states passionately, \"I believe Football to be the food of life. Nothing else brings such unbridled joy to so many people from the slums of Rio to the bright lights of Madrid. I have devoted my life to bring such joy to people who stood on the terraces supporting their club.\" God looks up and offers Ronaldo the seat to his left. He then turns to Luis Figo , \"and you, Luis , what do you believe?\" Figo stands tall and proud, \"I believe courage, honour and passion are the fundamentals to life and I\'ve spent my whole playing career providing a living embodiment of these traits.\" God, moved by the passion of the speech offers Figo the seat to his right. Finally, he turns to Wayne Rooney , \"and you, Wayne , what do you believe?\"   \"I believe\" says Rooney \"you\'re sitting in my seat.\"  '),(15584,3523319,'22,June,2004','2004-06-22 00:00:00','   just having a Shiddy Day. sorry.  and i had so much to say, too.  '),(15585,3523319,'22,June,2004','2004-06-22 00:00:00','   seriously  '),(15586,3523319,'22,June,2004','2004-06-22 00:00:00','   is anyone there  '),(15587,3523319,'22,June,2004','2004-06-22 00:00:00','   http://www.bbc.co.uk/weather/5day.shtml?id=1045  dammit  '),(15588,3523319,'22,June,2004','2004-06-22 00:00:00','   Orange users can text \'tubemap\' to 60835 to download it as a free application  '),(15589,3523319,'22,June,2004','2004-06-22 00:00:00','   i see that the italians are claiming that the scandy result was a fix. well they would, wouldn\'t they?  '),(15590,3523319,'22,June,2004','2004-06-22 00:00:00','   I\'d probably change Ibrahimovich for Larsson and possibly Zambrotta for, wait for it....Ashley Cole. I hate Ronaldo - He is a one trick wonder to me but he throws defences into disarray which is pretty damn useful so I would leave him in.  '),(15591,3523319,'22,June,2004','2004-06-22 00:00:00','   tomasson might be worth a shout up front, but only because of end result, rather than thkillth and that. and gravesen is a gee boo.  i\'d say that to his face.  i too had a message from aj, and it only upset me a little. but then he really does love too much, doesn\'t he?  that ross: hippie.  send me the pit crew thing.  i am damp.  '),(15592,3523319,'22,June,2004','2004-06-22 00:00:00','   I\'d have Van Der Saar (?) in goal and I\'d leave out Ronaldo, he\'s a one trick pony.  Overall, I\'m pretty glad the Italians are out.  They are the Peter Ebdon of European football - they grind you down.  '),(15593,3523319,'22,June,2004','2004-06-22 00:00:00','   Ok, I know the group stages aren\'t even finished so this is premature but this is my team of the tournament so far.  GK: Sorensen LB: Zambrotta RB: Neville CB: Campbell CB: Carvalho LM: Nedved RM: Ronaldo CM: Zidane CM: Gravesen F: Ibrahimovich F: Rooney  Discuss  '),(15594,3523319,'22,June,2004','2004-06-22 00:00:00','   I\'m going to Glastonbury.  '),(15595,3523319,'22,June,2004','2004-06-22 00:00:00','   I actually felt really sorry for the Italians at that moment.  Then i remembered how they\'ve bored me shitless during the rest of the tournament and began taking glee.  I\'ve seen that photo of Rooney, its quite good.  Has anyone seen the mocked up sky sports article about ferrari hiring some new pit crew, thats also quite funny.  Did anyone else receive a slightly disturbing, slightly Damo-esque message from Mr Sinnott after the England game  '),(15596,3523319,'21,June,2004','2004-06-21 00:00:00','   eeeexcellent. good old the scandies.  did anyone else take perverse glee from the look on cassano\'s face when he scored and 6 seconds later?  i\'m damaged, aren\'t i?  '),(15597,3523319,'21,June,2004','2004-06-21 00:00:00','   Have Bewilderness now so will bring copy on thursday  '),(15598,3523319,'21,June,2004','2004-06-21 00:00:00','   No.  '),(15599,3523319,'21,June,2004','2004-06-21 00:00:00','   To explain the drivel in last post...photo of rooney halfway through his celebration flip...trail of hub caps, wallets, etc. falling out of his shirt. Haven\'t seen photo of ZZ Bald Top\'s chunder...kind of glad I haven\'t as well!  '),(15600,3523319,'21,June,2004','2004-06-21 00:00:00','   Has anyone seen the footage of Zidane chucking up before he scored the penalty against us?  Shame the French bastard didn;t slip over in it like that guy in Gran Canaria.  '),(15601,3523319,'21,June,2004','2004-06-21 00:00:00','   I wouldn\'t Ross...he just wants to make a masturbate over you. Anyone seen the photo of Rooney doing halfway through his flip celebration from the Switzerland match with a continuous flow of wallets and hub caps falling out of his shirt? It\'s tres amusing  '),(15602,3523319,'21,June,2004','2004-06-21 00:00:00','   i demand that you post a picture of yourself with no hair.  a picture of your head, that is.  '),(15603,3523319,'21,June,2004','2004-06-21 00:00:00','   i thought i might have been labouring the point a little. anyway, fonz is the one with the dodgy blood system - clearly not enough blood to go around there.  '),(15604,3523319,'21,June,2004','2004-06-21 00:00:00','   you forgot the bit about big men having poor cirulation  '),(15605,3523319,'21,June,2004','2004-06-21 00:00:00','   Hats Off  To Daniel Taylor of The Guardian for this couplet: \'Owen, once the baby-faced assassin, has had to make way for Rooney, the assassin-faced baby.\'  '),(15606,3523319,'21,June,2004','2004-06-21 00:00:00','   600 Baby!  '),(15607,3523319,'21,June,2004','2004-06-21 00:00:00','   i\'m going to catch you, damo. and when i do, i\'m going to hold you down.  i have no definite plans at all for the weekend, aside from finish work at some point and get on a train. then call joe when i get to new street.  maybe i should think about it a bit more.  '),(15608,3523319,'21,June,2004','2004-06-21 00:00:00','   so, as far as this weekend goes.  has anyone got any firm plans yet?  has anyone actually spoken to Joey and found out what, if anything might be happening on the friday night.  '),(15609,3523319,'21,June,2004','2004-06-21 00:00:00','   I shot ahead on the fantasy again...only a few furlongs to go now!  '),(15610,3523319,'21,June,2004','2004-06-21 00:00:00','   i dun did yoga again this morning. i feel curiously warm now. i wonder whether the two are related.  does anyone know any english wingers?  '),(15611,3523319,'21,June,2004','2004-06-21 00:00:00','   In answer to your question fonz, it happened the minute David James became England keeper.    Trot forward 3 steps, he\'s coming to get it.  back 3 steps, oh no he isn\'t.  oops they\'ve scored  '),(15612,3523319,'20,June,2004','2004-06-20 00:00:00','   tee hee hee hee ross hee hee hee.  wee hee man tee hee.  '),(15613,3523319,'20,June,2004','2004-06-20 00:00:00','   Good game, good game.  Shame we can\'t defend set pieces really, isn\'t it? When did that start happening?  '),(15614,3523319,'20,June,2004','2004-06-20 00:00:00','   wow, that Rooney\'ll be a good player once he\'s grown into his face  But dammit, When is Garth Crooks going to confirm we\'ve qualified  '),(15615,3523319,'20,June,2004','2004-06-20 00:00:00','   Shit! He must have a foot like a traction engine!  '),(15616,3523319,'20,June,2004','2004-06-20 00:00:00','   I just listened to His \'n\' Hers. I\'d forgotten what a spanking good album it is.  Also, I\'ve had all my hair cut off, and now I\'m cold.  Message ends.  '),(15617,3523319,'20,June,2004','2004-06-20 00:00:00','   To clarify...it isn\'t the website introducing the ban on smacking children - there won\'t be bbc police running around banning people from the website with shouts of \"You were smacking your child whilst viewing our website!\"...no it\'s the government.  Stupid government  '),(15618,3523319,'20,June,2004','2004-06-20 00:00:00','   Just read an article on the bbc news website that is talking about introducing a ban on smacking children.  All fine and well...what happens when you have a deaf, dumb and blind kid raping women left right and centre...how do you let him know he\'s done something wrong - the beats of course.   I think they need to work a few clauses into this bill...clearly not a enough thought has gone into it  '),(15619,3523319,'20,June,2004','2004-06-20 00:00:00','   That game is appalling. What were you thinking, Dave?  '),(15620,3523319,'20,June,2004','2004-06-20 00:00:00','   That is the worst game I\'ve played in a long long time.  If I\'d made that, I\'d be ashamed to put my name to it.  '),(15621,3523319,'20,June,2004','2004-06-20 00:00:00','   urlLink this  is rubbish, and reminds me of a beavis and butthead minigame from 1997 which was much better. and it\'s in italian.  '),(15622,3523319,'20,June,2004','2004-06-20 00:00:00','   damn you fonz, my gun was jumped. i\'m mostly looking forward to you trying not to discuss the trial before it\'s concluded, and making one small but significant mistake which leads to death by burglary.  '),(15623,3523319,'20,June,2004','2004-06-20 00:00:00','   or something  '),(15624,3523319,'20,June,2004','2004-06-20 00:00:00','   I\'m shitting bricks about tonight, and I am not confident at all.  It\'s a potential death trap. Mine field.  Whatever the expression is.  I\'ve been juryorying.  It\'s incredibly dull.  I might tell you all about when it\'s all over.  '),(15625,3523319,'20,June,2004','2004-06-20 00:00:00','   If we look at recent facts though. 1/You were filled with impending doom for France - Proven correct. 2/Switzerland you were confident we would win - Proven correct  Result - Confidence tonight = Victory for England  Reasons why the above is balls statistically...endless  '),(15626,3523319,'20,June,2004','2004-06-20 00:00:00','   you\'re right, ross, but unfortunately i find it really, really funny. bambos is to blame, honest.  he  is .  i\'m feeling confident for the first time this tournament, which is a very bad sign indeed.  '),(15627,3523319,'20,June,2004','2004-06-20 00:00:00','   i\'m feeling extremely unconfident for tonight  '),(15628,3523319,'03,July,2004','2004-07-03 00:00:00','   Yeah, I\'ve already got almost the entire portugal squad in mine, so only a fool would do the same now.  I am cursed.  And, like most people, I have no desire to see Macy Gray in the nude.  Mmmm, nude.  '),(15629,3523319,'02,July,2004','2004-07-02 00:00:00','   I\'m gonna get as many Portugeezers in my team as possible and then greece are going to fuck me. i just know it  '),(15630,3523319,'02,July,2004','2004-07-02 00:00:00','   So Macy Gray is to perform naked for charity then? I can\'t help thinking she would raise more money asking people to pay her to keep her clothes on...there\'s just no need Macy  '),(15631,3523319,'02,July,2004','2004-07-02 00:00:00','   If I can get away with mucking around with a countess, I think you will be fine Manx!  '),(15632,3523319,'02,July,2004','2004-07-02 00:00:00','   pissed, and wondering if the posh bird is too posh for me  i\'m scum  '),(15633,3523319,'02,July,2004','2004-07-02 00:00:00','   oh shit, marlon brando\'s dead.  he looked so healthy too  '),(15634,3523319,'01,July,2004','2004-07-01 00:00:00','   pore old fat marlon. i heard he was living in poverty the other day, so it\'s probably for the best. the fat bugger.  i\'m all going on a summer holiday so i\'ll see yous all in a week or so. westside.  '),(15635,3523319,'01,July,2004','2004-07-01 00:00:00','   i was knackered when i woke up, then i had to work hard all day.  Now however i have that friday feeling  '),(15636,3523319,'01,July,2004','2004-07-01 00:00:00','   Me too.  I was tired when I woke up, then I used all my energy getting to 12:00, and then I nearly yawned my head off.  Need... sleep...  '),(15637,3523319,'01,July,2004','2004-07-01 00:00:00','   No it isn\'t, you are just being silly now.  Have a good holiday stupido Dave.  This afternoon has gone very slowly.  Tonight I intend to go to bed early and sleep for 30 or 40 hours.  It will be good  '),(15638,3523319,'01,July,2004','2004-07-01 00:00:00','   urlLink this  is damo. it is.  '),(15639,3523319,'01,July,2004','2004-07-01 00:00:00','   sometimes my rss is a little slow, but it gets there in the end. try tinkering with its properties.  '),(15640,3523319,'01,July,2004','2004-07-01 00:00:00','   bbc news website!  '),(15641,3523319,'01,July,2004','2004-07-01 00:00:00','   My RSS thing is not updating automatically.  What the deuce?  '),(15642,3523319,'01,July,2004','2004-07-01 00:00:00','   yes, but i\'m off tonight and you\'re not sitting next to me with those books in your hand, unlike my waterstones bag. although it doesn\'t have hands. you know what i mean.  '),(15643,3523319,'01,July,2004','2004-07-01 00:00:00','   I could have lent you at least two of those books  '),(15644,3523319,'01,July,2004','2004-07-01 00:00:00','   You lucky, lucky man.  I\'m updating my website.  Really I need Dave to revamp the design for me, youy know, before he pisses off on holiday.  The good news is, it looks nice on my new monitor.  It looked pap om my old one.  '),(15645,3523319,'01,July,2004','2004-07-01 00:00:00','   i\'ve fed mine the feedburner rss, then all i need to do is leave the mouse pointer over each post for a couple of seconds to get the full effort. it\'s not perfect, though - although you can jam it full of rss feeds from all over the place and it keeps you informed. sort of. the bbc ones are good. it also saves me looking out of the window by telling me what the weather is like outside.  i went shopping. i got:  empire magazine (although i had to buy a second copy cos mysteriously the pages got stuck together on the first one, right about page 100) black dahlia dave gorman\'s googlewhack adventure love all the people (bill hicks) last tango in aberystwyth (read the first one - briwwiant)  it was 3 for 2 on, um, 3 of them. now i want to sit down and read them all immediately.  '),(15646,3523319,'01,July,2004','2004-07-01 00:00:00','   I was wondering that.  Ok I\'ver got this sidebar thing going on,and I\'m on board with the idea.  But the best I can make it do is display the first few words of each message on the BOA, and if I click on it, it loads up the full boa in IE.  That\'s no good.  Tell me how to configure it Dave.  '),(15647,3523319,'01,July,2004','2004-07-01 00:00:00','   how do you  find  this crap?  '),(15648,3523319,'01,July,2004','2004-07-01 00:00:00','   Fuck me! Look at the picture of the three guys a bit further down...Instant reaction is \"Geeks\"...then look at the T-Shirt the fella on the left is wearing...he\'s a member of the Bridge Clug...Fuck me sideways  '),(15649,3523319,'01,July,2004','2004-07-01 00:00:00','   http://www.speakeasy.org/~lion/ This guy is the biggest fucking geek in the world - Yoav could end up like this chap. Do not start reading the web page...you will find yourself disgusted and cursing in disbelief that someone could be such a complete geek.  The picture that loads up at the top should be all you need...Jesus I can\'t get over what a complete geek this person is....  '),(15650,3523319,'01,July,2004','2004-07-01 00:00:00','   and damo is ted ghopal.  '),(15651,3523319,'01,July,2004','2004-07-01 00:00:00','   urlLink Marshall Boone  is manx.  As for books... Gee, I don\'t know.  Have you read The Beach?  I assume you have, but it\'s a great book to read on holiday.  Treasure Island is good also.  I like to read those Penguin Classics, cos they\'re cheap.  '),(15652,3523319,'01,July,2004','2004-07-01 00:00:00','   i use desktop sidebar, which is not the best rss reader by any means, but has a host of other handy little functions. i\'ve got 2 15\" screens, so plenty of space for it; don\'t know whether your single gorgeous 19\" lovestick can spare a couple of inches down the side.  i\'m going out at lunchtime to buy books and a copy of empire. now is the time to give me any final recommendations.  i think wobu is back from israel - either that, or there was a really clumsy burglar who didn\'t steal anything crashing around in the kitchen at 2 o\'clock this morning. i know on which i\'d bet.  '),(15653,3523319,'01,July,2004','2004-07-01 00:00:00','   Knightley: got told off by her landlord for smearing blue paint on the carpet while filming King Arthur.  That was all she revealed about her home life as far as I can tell.  I think Empire is probably worht you buying this month, not just for the Knightley, but mostly for that.  Someone tell me which RSS Reader I should be using.  I want to know which one is the bestest.  My new computer at work is now complete, and I\'m very excited.  Just look at my new 19\" TFT monitor.  LOOKATIT!  '),(15654,3523319,'01,July,2004','2004-07-01 00:00:00','   thatsa lotta crappa!  etc.  '),(15655,3523319,'01,July,2004','2004-07-01 00:00:00','   I\'ve been told to clean up disc pace at work.  apparentlyI have over 600 gig of crap lying around.  oh well  '),(15656,3523319,'01,July,2004','2004-07-01 00:00:00','   \"Football fans all across Europe will rejoice this Autumn, as Konami of Europe confirms the first details of what will rate as one of the most anticipated games of 2004: Pro Evolution Soccer 4. Due for release on PlayStation 2, PC DVD and for the first time in the seriesâ€™ history, the Xboxâ„¢ videogame system from Microsoft, Pro Evolution Soccer 4 once again sets new standards in control, aesthetics and realism.\"   full announcement and screenshots (on page 2)  urlLink http://www.pesfan.com/bulletin/showthread.php?s=&threadid=37438\" target=\"_blank\">here   spidermun 2 sounds hellagood. it\'s just a shame that ross isn\'t in it. how\'s the new house, wee man?  tell me more of the knightley interview. which parks does she walk through at night?  '),(15657,3523319,'01,July,2004','2004-07-01 00:00:00','   That is not the bright side.  There\'s quite a good interview with the Knightley in this month\'s Empire magazine.  It also has a cool Spider Man 2 cover.  I am starting to get feverishly excited about the next Spidey outing.  '),(15658,3523319,'01,July,2004','2004-07-01 00:00:00','   um... if she\'s anything like the aguilera monster, spears is probably getting banged rough up the arse on a nightly basis, so she may not be preggers...  wait a minute... up the arse... nightly... sounds a bit like   excuse me for a moment  '),(15659,3523319,'01,July,2004','2004-07-01 00:00:00','   I have two things to announce:  1). I have effectovely insured that Greece will win Euro 2004 by buying the entire Portugal team in Fantasy Football.  2). I will officially dissociate myself from anything Spears related if I find out she\'s pregnant.  '),(15660,3523319,'01,July,2004','2004-07-01 00:00:00','   Beers, cocktails, football, strip joint - Last night was a proper night out it has to be said.   Tonight - Recover  Saturday - Cocktail bar, me one other guy...14 fit girls...this week looks like being one to remember  '),(15661,3523319,'01,July,2004','2004-07-01 00:00:00','   it\'s my bloody curse again. mind you, portugal have nothing to worry about:  in their quarter final, the referee was swiss; their opponents had knocked out switzerland in the group stage, and his performance was highly questionable.  in their semi final, the referee was swedish; their opponents had knocked out sweden in the previous round, and his performance was highly questionable.  now all we need is a czech referee for the final, and we\'re all set. although he\'s german. but he probably hates the greeks anyway.  seriously, it IS a conspiracy.  i\'ve not done any packing, and that was my one aim this evening.   i  can\'t do anything.  '),(15662,3523319,'30,June,2004','2004-06-30 00:00:00','   Fuck  I can\'t believe the fucking pseudo-germans have reached the finals  '),(15663,3523319,'30,June,2004','2004-06-30 00:00:00','   i\'ve cursed poor old pavel too. greeks are doing well, though, innit?  '),(15664,3523319,'30,June,2004','2004-06-30 00:00:00','   I put in a couple of token greeks, but I\'ve got enough transfers to take all the Czechs out of my team if they don\'t make the final I think.  '),(15665,3523319,'30,June,2004','2004-06-30 00:00:00','   If the Czechs fuck up tonight then my team is a bit screwed  '),(15666,3523319,'30,June,2004','2004-06-30 00:00:00','   I\'m here, and I\'ve finally been dismissed from Ian Dury duty.  I\'ve got an evening of FFX-2, beer and football planned, and I literally can\'t wait to have it spoilt by whatever will arise.  The best laid plans etc.  And these plans aren\'t at all well laid.  '),(15667,3523319,'30,June,2004','2004-06-30 00:00:00','   nuts to this. i\'m off for a barry white.  '),(15668,3523319,'30,June,2004','2004-06-30 00:00:00','   i keep trying to do that. work all day, that is. unfortunately, it turns out that i\'m not trying very hard.  '),(15669,3523319,'30,June,2004','2004-06-30 00:00:00','   I was working for all of yesterday really without break...your copy is made and merely needs to be delivered!  '),(15670,3523319,'30,June,2004','2004-06-30 00:00:00','   where have you been? and where\'s my copy? and other demands!  i demand that ross festoon me with his gooey lifestyle wisdom.  '),(15671,3523319,'30,June,2004','2004-06-30 00:00:00','   Allo, Allo...I watched Bewilderness last night finally - Half his jokes I\'d heard at the Hammersmith...surprisingly enough the other half I hadn\'t  '),(15672,3523319,'30,June,2004','2004-06-30 00:00:00','   i\'ve had maniche for ages. he\'s helluvacheap. shame i cursed every single other player i had (and even maniche got subbed off):  valente: booked maniche: subbed overmars: booked and subbed robben: booked and subbed deco: bollocks horse: bollocks  pretty good i think you\'ll agree. transfer time...  '),(15673,3523319,'30,June,2004','2004-06-30 00:00:00','   It turns out I had a 2-1 ratio of portugays to dutch, but I didn\'t have any goal scorers by the looks of things.  Best make some transfers.  '),(15674,3523319,'29,June,2004','2004-06-29 00:00:00','   hmmmm, I had a few too many Dutchies and not enough Portugays in my team tonight I reckon.  Still at least I had Ronaldo.  I\'m quite pissed off that I flipped a coin to decide between Maniche and Seedorf for the final place in my midfield and it landed on Seedorf.  Now for wholesale changes  '),(15675,3523319,'29,June,2004','2004-06-29 00:00:00','   I accidentaly just did my impression of Madonna circa 1987.  Sorry about that.  I just finished reading  Little Women , by Louisa May Alcott, which I believe makes me Edercated.  I quite enjoyed it, really.  But it\'s not as good as the Geisha book.  Speaking of holidays, my mum has been saying I should have one, and she\'s now got to the stage where she\'s going to go in some travel agents and try and find one for me, so I\'ll get back to people about it when she comes up with the goods.  Obviously, I\'m too lazy to investigate the situation myself. I thought at one point I wasn\'t. I was wrong.  '),(15676,3523319,'29,June,2004','2004-06-29 00:00:00','   ok.  where is everyone else today? is damo dead(er)? has ross recovered from glasto? does he still have more hair on his chin than on his head?  '),(15677,3523319,'29,June,2004','2004-06-29 00:00:00','   Just to clarify, I meant you should read \'Join Me\' by Danny Wallace, not just that you should join me in reading the Black Dahlia.  although you should actually do that too  '),(15678,3523319,'29,June,2004','2004-06-29 00:00:00','   might as well. any others? has ross read any books while growing his leg hair?  '),(15679,3523319,'29,June,2004','2004-06-29 00:00:00','   Join Me  '),(15680,3523319,'29,June,2004','2004-06-29 00:00:00','   black dahlia  '),(15681,3523319,'29,June,2004','2004-06-29 00:00:00','   on saturday, for a week in france with my dad and my little sister (who just finished her gcses and reckons she\'s suddenly a growed-up). i\'m going to take a lot of books with me, i think. any suggestions? i\'m just finishing fonz\'s geisha book. it\'s good.  '),(15682,3523319,'29,June,2004','2004-06-29 00:00:00','   jamie \"fat-tongue\" oliver was telling me the virtues of sainsbury\'s sausages just last night. i didn\'t believe him, though.  i hope banksy enjoyed his fry-up leftovers.  '),(15683,3523319,'29,June,2004','2004-06-29 00:00:00','   Tesco have a good selection of sausages I\'ve discovered.  I had some traditional pork and herb ones cooked by Gibbos fair hand on sunday and some pork and chilli ones last night.  I also have some Pork and stilton in the freezer.  They\'re on offer at the moment  Great  '),(15684,3523319,'29,June,2004','2004-06-29 00:00:00','   I watched that Boosh thing once.  I didn\'t think it was very funny.  I have seen Monkey Dust, and I thought it was quite good.  '),(15685,3523319,'29,June,2004','2004-06-29 00:00:00','   at last, and watched monkey dust and the mighty boosh on bbc three. i was hella-impressed.   \"have you ever seen brian ferry? does he look like terry wogan?\"  '),(15686,3523319,'28,June,2004','2004-06-28 00:00:00','   urlLink Director news    urlLink Similar news   Hmm.  I\'m sure I remember more news than that.  The news section was called something different until a few months ago.  It looks like they only have the High Noon news available on their search engine.  '),(15687,3523319,'28,June,2004','2004-06-28 00:00:00','   I can\'t find any interesting stories, where are they.  apparently i\'m 42% black, not as high as Joe i\'m sure  '),(15688,3523319,'28,June,2004','2004-06-28 00:00:00','   Sooner or later,  urlLink this  is going to cause me to kill someone.  '),(15689,3523319,'28,June,2004','2004-06-28 00:00:00','   I\'ve done that before, by leaving a posting window open for a few minutes, it uses the time the page was loaded, not the time you pressed \"Submit\".  I remember now, Scarlet is playing the woman.  I should have mentioned that earlier.  There have been quite a few interesting stories about the Dahlia on the BBC film site over the past 6 months or so.  I\'m supposed to be writing a review of Shrek 2 (bless you) but all I really want to do is play FFX-2.  '),(15690,3523319,'28,June,2004','2004-06-28 00:00:00','   How did I magically post a reply to Manx before he said it?  '),(15691,3523319,'28,June,2004','2004-06-28 00:00:00','   Scarlett Johansson is to play the woman aparently, thats a choice i actually agree with  '),(15692,3523319,'28,June,2004','2004-06-28 00:00:00','   Oh yes, it was Scarlett, not Kiera.  '),(15693,3523319,'28,June,2004','2004-06-28 00:00:00','   knightley.  now you\'ve done it.  '),(15694,3523319,'28,June,2004','2004-06-28 00:00:00','   I don\'t think so, but I did hear about her being in something recently...  I just can\'t remember what.  Oh yeah, the new Pride and Prejudice.  Like we need another version.  '),(15695,3523319,'28,June,2004','2004-06-28 00:00:00','   Isn\'t that Keira Knightley chap pencilled in for a role? Or did I imagine that?  '),(15696,3523319,'28,June,2004','2004-06-28 00:00:00','   Its just such bad casting, both those cops are supposed to be double-hard ruthless bastard ex boxer types, I also think wahlberg and hartnett are too young.  I\'d cast Michael Chiklis, in both roles  '),(15697,3523319,'28,June,2004','2004-06-28 00:00:00','   I am a font of Knowledge, dahlia wise.  Lessee: I\'ve read it, and enjoyed it.  It\'s about a woman who gets murdered in LA in the 50s, and two cops have to investigate.  And they both fall in love with a girl who looks a bit like the girl who was murdered.  She is nicknamed the Black Dahlia, and it\'s pronounced DAY-LEE-AAHH.  Hartnett and Wahlberg do play the two cops, Fire and Ice or whatever their names are. Their names both begin with B I think.  Anyway, it\'s been in the pipeline for quite a while now, and will doubtless be cack.  No LA Confidential, that\'s fo\' sho\'.  '),(15698,3523319,'28,June,2004','2004-06-28 00:00:00','   ross, with this on-line geekery, you are spoiling us.  i\'m tempted, which shames me. but it would cut down my wanking hours, which is a problem. i\'d have to miss work for 30 days to compensate.  i wonder whether anyone would notice?  '),(15699,3523319,'28,June,2004','2004-06-28 00:00:00','   Anarchy Online recently celebrated it\'s third anniversary, and as a special pressie I have been given three keys to allow people to play the game itself, and both expansion packs, free for thirty days.  I know, you can scarcely contain your excitement.  Nonetheless, if any of you want to try it out, let me know and I\'ll emailer you one of the keys.   urlLink This  is the game\'s site, if you want to check it out.  '),(15700,3523319,'28,June,2004','2004-06-28 00:00:00','   maybe marky mark will play a bumbling, affable corpse in this one.  i\'ve never read the black dahlia, and have no idea what it is about nor how to pronounce it correctly. hell, this probably isn\'t even the right spelling.  what\'s wrong with the neoboa today?  '),(15701,3523319,'28,June,2004','2004-06-28 00:00:00','   They should call it the Black Daisy to avoid confusing the Yanks.  '),(15702,3523319,'28,June,2004','2004-06-28 00:00:00','   Hmmm, I see their making a fillum of The Black Dahlia.  That would be good, except the cast list seems to include the words Wahlberg and Hartnett  '),(15703,3523319,'28,June,2004','2004-06-28 00:00:00','   tee hee. they\'re right about the Big Man, again, too.  '),(15704,3523319,'28,June,2004','2004-06-28 00:00:00','   \'Soccer star Jonathan Woodgate was helped from a bar pouring blood from a head-wound - after glassing HIMSELF\' - The Sun. Apparently, he thought he had a blackhead.  '),(15705,3523319,'28,June,2004','2004-06-28 00:00:00','   the monkey claw is smelly.  '),(15706,3523319,'28,June,2004','2004-06-28 00:00:00','   was the doll amusingly bearded, and sitting on a rusty iron spike? that\'s attention to detail, that is.  '),(15707,3523319,'28,June,2004','2004-06-28 00:00:00','   It wasn\'t me...I was in the pub with my mates - drinking to Ross\'s Last Stand.  That\'s what they\'ve named the mud puddle that Ross passed away in at Glastonbury.  Some sweet fella from the Gnome Brigade has even left a little doll dressed in a tartan skirt next to the now dried up spot that claimed wee Ross Mctaggart.  When the wind blows the skirt lifts gently to reveal a couple of shrivelled walnuts...they knew him well.  '),(15708,3523319,'28,June,2004','2004-06-28 00:00:00','   why is this blog listed as not found? have we been shut down because i mentioned rusty iron spikes? or was it the german marching?  or was it damo?  '),(15709,3523319,'28,June,2004','2004-06-28 00:00:00','   You may also want to ensure there is a nearby hospital - This will allow you to sit on your 6 foot spike safe in the knowledge that you are only a stone\'s throw away from getting NHS treatment for the recurring internal haemorraging and gangrene outbreaks you suffer at the hands of your preferred domicile.  '),(15710,3523319,'28,June,2004','2004-06-28 00:00:00','   is it a pain in the bum because you\'re moving onto a 6-foot rusty iron spike?  remember, ross: location, location, location. ensure that your rusty iron spike is close to local schools, shops and amenities, and has a view of the canal.  '),(15711,3523319,'28,June,2004','2004-06-28 00:00:00','   I have to move house tomorrow. That\'s a bit of a pain in the bum.  '),(15712,3523319,'28,June,2004','2004-06-28 00:00:00','   your arse  '),(15713,3523319,'28,June,2004','2004-06-28 00:00:00','   I\'m back, and sporting a rather fetching beard.  It\'s coming off when I can be arsed to shave.  '),(15714,3523319,'28,June,2004','2004-06-28 00:00:00','   Interesting.  I\'ve just been watching T. Menman playing tennis.  For a moment there, it looked like he\'d learnt how to hold his serve.  '),(15715,3523319,'28,June,2004','2004-06-28 00:00:00','   I think its two days after manx\'s birthday but I\'m not certain  '),(15716,3523319,'28,June,2004','2004-06-28 00:00:00','   not until august time if I remember correctly  '),(15717,3523319,'28,June,2004','2004-06-28 00:00:00','   I can\'t make that link work.  I forgot to do my transfers before the Czech Republic game, that pisses me off.  I have been in court again today, laying down the law at some mofos.  Seems I\'ll be doing the same thing for much of this week.  Now somebody tell me when Andy\'s Wedding Anniversary is.  '),(15718,3523319,'28,June,2004','2004-06-28 00:00:00','   urlLink a touch of the damos .  '),(15719,3523319,'28,June,2004','2004-06-28 00:00:00','   i need a wee.    that\'s better.  '),(15720,3523319,'28,June,2004','2004-06-28 00:00:00','   if only he\'d taken his mud raft.  '),(15721,3523319,'28,June,2004','2004-06-28 00:00:00','   Ross fell over face first into a pool of mud and died.  His wee arms were unable to lift himself that inch or so necessary to clear the mud line and he suffocated to death.   In actual fact I doubt he has managed to get away from the site yet...  '),(15722,3523319,'28,June,2004','2004-06-28 00:00:00','   is fonz still having his day in court today? and when is the wee man back from the land of druids and mushrooms?  '),(15723,3523319,'28,June,2004','2004-06-28 00:00:00','   i\'ve made some transfers again. i think i need more czechs.  '),(15724,3523319,'28,June,2004','2004-06-28 00:00:00','   Yeah I know...I have to give up now really...have missed 3 days of transfers (dad was in hospital) as you would see if you look at who makes up my team! I would be champion still else...and you know it.  '),(15725,3523319,'28,June,2004','2004-06-28 00:00:00','   you suck Damo, you getting yo ass kicked in the fabtasy football now  '),(15726,3523319,'28,June,2004','2004-06-28 00:00:00','   after getting back to the flat in time for the second half of the czech game, i thought i\'d best get to bed early or i\'d be as sick as a dog this morning. so i went to bed after fantasy football, and read for a bit, then turned off the light, lay back, and immediately felt incredibly sick.  spent several hours going to and from the bathroom feeling like i was going to throw up, then not throwing up.  i AM bitter today. i could have spent that time playing pes with a bucket in my lap, if i\'d known.  '),(15727,3523319,'28,June,2004','2004-06-28 00:00:00','   I\'m very tired and would appreciate it if a little imp were to setup a bed by the side of my desk that I can sleep on for a while  '),(15728,3523319,'09,July,2004','2004-07-09 00:00:00','   Well, he was right.  I\'ve been massaged, and now I can\'t move properly.  Ross, truthofitismate, that is why web front ends are the way forward.  Also, no client side installation is required for uprages.  That\'s supposed to say upgrades.  Ouch.  '),(15729,3523319,'08,July,2004','2004-07-08 00:00:00','   I like a man who knows his own worth  \"Within minutes of injuring my knee I sensed it was not going to be a happy ending for the Czech Republic\" Pavel Nedved  '),(15730,3523319,'08,July,2004','2004-07-08 00:00:00','   hmmm  '),(15731,3523319,'08,July,2004','2004-07-08 00:00:00','   haven\'t done any more work yet  '),(15732,3523319,'08,July,2004','2004-07-08 00:00:00','   i\'ve done all my work for the day now  '),(15733,3523319,'08,July,2004','2004-07-08 00:00:00','   This morning I mostly split my database into a back and front end, and converted the front end so I can run it from both XP and NT using the same data. Unfortunately, that seems to have buggered up the referential integrity rules.  Look how interesting my job is.  '),(15734,3523319,'08,July,2004','2004-07-08 00:00:00','   Last night I downloaded and viewed the episode of South Park that Dave told me too.  It was quite funny , the only problem was that Dave had already told me in extremely minute detail exactly what happen  '),(15735,3523319,'08,July,2004','2004-07-08 00:00:00','   If I were to be arrested, I\'d want it to be for something like  urlLink this .  '),(15736,3523319,'08,July,2004','2004-07-08 00:00:00','   I bet it was really interesting.  Come on, I can\'t take the suspense!  I\'m running out of things to read on the interweb.  '),(15737,3523319,'08,July,2004','2004-07-08 00:00:00','   I had something to say yesterday, when the BOA wasn\'t working.  Now i can\'t remember what  '),(15738,3523319,'08,July,2004','2004-07-08 00:00:00','   It looks like it.  I am bored today, and it\'s looking like I have to stay here until 5:30 today, when I\'m entitled to leave at 12.  That\'s rubbish that is.  '),(15739,3523319,'08,July,2004','2004-07-08 00:00:00','   is this working yet  '),(15740,3523319,'08,July,2004','2004-07-08 00:00:00','   bit bored today  '),(15741,3523319,'07,July,2004','2004-07-07 00:00:00','   bit bored today  '),(15742,3523319,'07,July,2004','2004-07-07 00:00:00','   http://www.spiteyourface.com/spidey.html  Download this and watch it - its better than the film !  '),(15743,3523319,'07,July,2004','2004-07-07 00:00:00','   http://www.wagenschenke.ch/  '),(15744,3523319,'07,July,2004','2004-07-07 00:00:00','   http://www.empireonline.co.uk/site/news/newsstory.asp?news_id=15992  YES  '),(15745,3523319,'07,July,2004','2004-07-07 00:00:00','   In order of fitness it goes blonde, darkie, mutyaisevil munter.  However, I\'d only class the blonde as definitely fit.  But I think the darkie would get it.  '),(15746,3523319,'07,July,2004','2004-07-07 00:00:00','   I\'m going to a party in August...on the invitation list one of the people is called Kiera Knightley...more as a joke I asked Amelia if this was Kiera as in the film star.  She just laughed and failed to give a response - now I\'m all intrigued.  '),(15747,3523319,'07,July,2004','2004-07-07 00:00:00','   Well the very coloured one...I wouldn\'t use the term \"fit\" however more \"acceptably photographical\" - and yes...she would get it  '),(15748,3523319,'07,July,2004','2004-07-07 00:00:00','   2 fit ones, which is the other fit one besides blondie.  Please refrain from using their names as I don\'t know them.  Colours will suffice  '),(15749,3523319,'07,July,2004','2004-07-07 00:00:00','   2 fit ones, which is the other fit one besides blondie.  Please refrain from using their names as I don\'t know them.  Colours will suffice  '),(15750,3523319,'07,July,2004','2004-07-07 00:00:00','   That would be the one...I find it pretty amusing that almost all the (I refrained putting a made up statistic in there in favour of a sweeping statement) Sugababe photos you see are of the two fit ones.   I guess mutya or whatever her name is would be better suited to the \"Dirty Slut\" editorials that find haven on the top shelves of magazine stalls.  '),(15751,3523319,'07,July,2004','2004-07-07 00:00:00','   Which one\'s the chav?  The really munting one presumably. Mutya?  '),(15752,3523319,'07,July,2004','2004-07-07 00:00:00','   I quiet fancy getting trollied in a pub... I was pleased to see that the Chav from the Sugababes was voted in the top 10 celebrity mingers last night on sky.  Gaffers was there obviously as well.  '),(15753,3523319,'07,July,2004','2004-07-07 00:00:00','   What am I saying? I must have misread her age.  I almost forgot I was 19 then.  I want to go back to bed about now.  '),(15754,3523319,'07,July,2004','2004-07-07 00:00:00','   younger than you.  I\'m 19  '),(15755,3523319,'07,July,2004','2004-07-07 00:00:00','   I can\'t believe this  urlLink raddled old tart  is younger than us.  '),(15756,3523319,'06,July,2004','2004-07-06 00:00:00','   Is it \"Ku Klux Klan\"?  I can never remember, but it\'s a stupid name anyway.  I just got a free copy of the Scissor Sisters\' album.  Maybe that will help me decide if I like them or not.  Where the giddy fuck is manx? And why doesn\'t my RSS feed work properly?  My Blur-esque quote from earlier was intentional by the way.  I\'m going to try and work in an Oasis one later.  '),(15757,3523319,'06,July,2004','2004-07-06 00:00:00','   I had a look at that BBC thing about the modern art. It is indeed poo. What is more poo, though, is that the reporter made the schoolboy error of calling the KKK the \'Klu Klux Klan\'.  Now see how many far-right Google ads we get.  '),(15758,3523319,'06,July,2004','2004-07-06 00:00:00','   I just got my computer back from being fixed, and now it has XP on it. Mmm, XP.  Unfortunately, however, it doesn\'t have Access, which is a slight drawback when I\'m supposed to be doing database design. Oh well.  '),(15759,3523319,'06,July,2004','2004-07-06 00:00:00','   Ooh, you\'re like Blur. Almost.  '),(15760,3523319,'06,July,2004','2004-07-06 00:00:00','   Well, hopefully this one will be hanging around for a couple of weeks, giving me plenty of time to muck up her lens.  Modern art is rubbish.  '),(15761,3523319,'06,July,2004','2004-07-06 00:00:00','   Fonz you are so wrong in the head, you need serious help, that was a wasted opportunity if I\'ve seen one.  You should have jumped her bones... http://news.bbc.co.uk/1/hi/entertainment/arts/3871039.stm Half way down page - a donor kebab stand with references to the klu klux clan and world war 2.  It\'s fucking stupid...and I bet the national lottery funded it with the \"we are funding new artists to produce cutting edge pieces of work\" line...cutting edge my arse.  I don\'t spite the artists, only the retards that insist on spending money to sponsor this sort of crap.  Makes me want to break stuff.  '),(15762,3523319,'06,July,2004','2004-07-06 00:00:00','   You\'re heading down a dangerous road, Fonz... it\'s called Spencer Pyle Avenue.  '),(15763,3523319,'06,July,2004','2004-07-06 00:00:00','   I\'ve been quite busy today.  I\'m getting worried.  A fit girl was shown round our office today, and I definitely would have.  Turns out she\'s still at school and is here on work experience.  This means she\'s 15 or 16.  I have completely lost the ability to judge girl\'s ages.  That\'s the third time that\'s happened to me this year.  '),(15764,3523319,'06,July,2004','2004-07-06 00:00:00','   Sweet Jeebus work is boring today.  '),(15765,3523319,'06,July,2004','2004-07-06 00:00:00','   We can but hope.  '),(15766,3523319,'06,July,2004','2004-07-06 00:00:00','   I think it will be like X-2 - a defining piece of comic book film-making.  I wonder if Dunst will get her nips out again.  '),(15767,3523319,'05,July,2004','2004-07-05 00:00:00','   It will be like X-Men 2 I reckon - A roaring ride of action/special effects with a cheesy love story underneath...enjoyable to watch but not something I can see more than a couple of times.  '),(15768,3523319,'05,July,2004','2004-07-05 00:00:00','   Don\'t trust someone named after an operating system.   Spider-Man 2 will be arse-some, like 100 hot dogs.  '),(15769,3523319,'05,July,2004','2004-07-05 00:00:00','   From the Onion horoscopes:  Aquarius: (Jan. 20â€”Feb. 18) Newton\'s laws say that for every action there is an equal and opposite reaction, proving he knew nothing about women.  I don\'t believe you about Spider-man, Bambos, for the simple fact that you think Armageddon is a good film.  '),(15770,3523319,'05,July,2004','2004-07-05 00:00:00','   Saw Spider Man 2 last night...  It was utter bollox.  '),(15771,3523319,'05,July,2004','2004-07-05 00:00:00','   Yeah, he bonked her alright.  Where is everybody?  Looking back, it seems I spent most of yesterday talking to myself.  No change there really...  What do I put on the front page of my film website?  I demand the answers!  '),(15772,3523319,'05,July,2004','2004-07-05 00:00:00','   Simone made him do a sex wee...It fell off afterwards and now he believes he can only satisfy small children  '),(15773,3523319,'05,July,2004','2004-07-05 00:00:00','   Hey hey hey!  I just remembered something.  When Joe bonked Simone in Chamberlain.  Happy days...  [He bonked her.]  '),(15774,3523319,'04,July,2004','2004-07-04 00:00:00','   Right, here\'s a good question:  What the hayell do I have on the front page of my film website?  '),(15775,3523319,'04,July,2004','2004-07-04 00:00:00','   He went to France to make love to a schoolboy.  At least, that\'s what I heard.  '),(15776,3523319,'04,July,2004','2004-07-04 00:00:00','   I couldn\'t agree more.  When you pinch a girl\'s bum they should smile and possibly give you a flash of their undies, and when you pounce on them and try to snog them, they should not only let you, but be extremely grateful for the attention afterwards.  '),(15777,3523319,'04,July,2004','2004-07-04 00:00:00','   Fonz\'s message came back. What the fug is up with Blogger?  Damo, you were doing so well - but it would also be \'fewer people in jail\'...  I must take up Dave\'s crusade while he is away. Where did he go, anyway?  '),(15778,3523319,'04,July,2004','2004-07-04 00:00:00','   Interweb is important it has to be said.   Saturday night was good fun...me, 3 other guys, 15 or so fit women.  In a club being molested by 4 women at once was awesome. Which brings me around to an important point.  When I\'m in a club and grab random women, a slap can be expected or perhaps a firm kick in the junks.  Women need to be more open to the occasional grab and squeeze/forced kiss/love making...that way there would be fewer rapes and less people in jail.  '),(15779,3523319,'04,July,2004','2004-07-04 00:00:00','   I seem to have killed Fonz\'s message.  '),(15780,3523319,'04,July,2004','2004-07-04 00:00:00','   We get the keys to our new house tomorrow evening. One of the first things I\'ll be doing is phoning up Telewest to say \'make proper telly come on my telly, and make porn come on my interweb.\'  Yes.  '),(15781,3523319,'04,July,2004','2004-07-04 00:00:00','   Oh yeah, you done moved house, didn\'t you?  Let\'s all laugh at Ross and his cack TV and lack of innernet.  Where the heck is everyone today?  And when I say everyone, I mean Manx.  I went to the pub at lunch time and it was great.  There\'s a new bar maid there who is the fittest bird I have clapped eyes on for a very long time.  I\'m going again tomorrow lunch time, at which point I will be asking for a detailed copy of her work roster.  If she often works there between the hours of 11-6, I\'m going to have to find a different job I think, so I have more time to camp at her door step.  I think \"camp\" was the wrong choice of word there. Perhaps not.  '),(15782,3523319,'04,July,2004','2004-07-04 00:00:00','   why no tv and no interweb McRoss? I lost interweb for a bit yesterday...  '),(15783,3523319,'04,July,2004','2004-07-04 00:00:00','   No really, it does. I only have five TV channels. It\'s like living in the stone age.  '),(15784,3523319,'04,July,2004','2004-07-04 00:00:00','   I\'d quite like to see him direct though.  I only hope they can convince Vin Diesel or The Rock to take the part of Superman - it\'d be one hell of a film...  Lois Lane - Whoopi Goldberg?  '),(15785,3523319,'04,July,2004','2004-07-04 00:00:00','   The Man of Steel is also grabbing headlines today after the Superman logo appeared on www.michaelbay.com, which claims to be the \"official\" website of the golden-haired helmer. It sparked a wave of rumours that Bay might be about to replace Charlie\'s Angels director McG in the hot seat of the long-awaited Superman remake.   The Superman people try desperately to sign up the worst director in Hollywood.  '),(15786,3523319,'04,July,2004','2004-07-04 00:00:00','   Ahhh, another classic 1-0 victory by the Greeks.  I sure am glad I didn\'t have to watch the Italians grind their way to another final when we\'ve got the Greeks to do it instead.  '),(15787,3523319,'04,July,2004','2004-07-04 00:00:00','   The best thing about this whole championship - seeing Ronaldo cry...it made me smile and feel warm inside  '),(15788,3523319,'16,July,2004','2004-07-16 00:00:00','   We will be wedding-ing, but the invitation is still packed somewhere, and I\'ve forgotten all the details. This means I haven\'t booked as a hotel or anything yet, as well.  '),(15789,3523319,'15,July,2004','2004-07-15 00:00:00','   christ, knightley\'s looking uncannily like portman these days. or what portman ought to, had she not let herself go a bit. anyone seen the photos from the king arthur premiere?  another pot here, nurse.  '),(15790,3523319,'15,July,2004','2004-07-15 00:00:00','   nope. he said the guys he was working for were \"pretty small time\". or something like it. i think he was going for trump before.  '),(15791,3523319,'15,July,2004','2004-07-15 00:00:00','   Did he get that internship he wanted?  You know, for that guy I\'d vaguely heard of, but didn\'t know who he was? You know, the one that made Wobav cream himself when he spoke of him?  '),(15792,3523319,'15,July,2004','2004-07-15 00:00:00','   who else is going to weedavey\'s weeding? that\'s when we run around a park with some roundup and a trowel.  '),(15793,3523319,'15,July,2004','2004-07-15 00:00:00','   he should keep a diary of his internship in noo yawk. \"today, i made coffee, did some photocopying, and gave seven blow jobs. i love my job. dah.\" i\'m telling you, it\'s all true.  '),(15794,3523319,'15,July,2004','2004-07-15 00:00:00','   I wish Wobav would put something on his blog, so I could muck it up.  '),(15795,3523319,'15,July,2004','2004-07-15 00:00:00','   i\'m intrigued by...  no, nothing. nothing at all.  '),(15796,3523319,'15,July,2004','2004-07-15 00:00:00','   That bloke is clearly not me. He likes rubbish things, like Macs, iPods and shit photographs.  Plus, red is not a good choice of background colour, especially when you use mustard yellow for links. Mmm, culinary.  '),(15797,3523319,'15,July,2004','2004-07-15 00:00:00','   I\'ve found  urlLink Ross\' Blog . It took me about 30 seconds on google.  '),(15798,3523319,'15,July,2004','2004-07-15 00:00:00','   Hmm, I think I might burn a buggerload of MP3s onto a cd and then post them to manx.  And take them home for myself obviously.  '),(15799,3523319,'15,July,2004','2004-07-15 00:00:00','   Yes, I might pop for one of those pretty shortly.  I must remember to send my CV into work so i can apply for some jobs while I\'m here.  '),(15800,3523319,'15,July,2004','2004-07-15 00:00:00','   for a barry white.  '),(15801,3523319,'15,July,2004','2004-07-15 00:00:00','   dyud  '),(15802,3523319,'15,July,2004','2004-07-15 00:00:00','   dude.  '),(15803,3523319,'15,July,2004','2004-07-15 00:00:00','   apparently it takes just 6 o\'s to differentiate between \'Wake up\' and \'that is very cool\'  '),(15804,3523319,'15,July,2004','2004-07-15 00:00:00','   Good to see BASEketball getting the recognition it deserves there.  '),(15805,3523319,'15,July,2004','2004-07-15 00:00:00','   I looked up  urlLink dude  in the Wikipedia. It\'s quite entertaining. I especially like the \'Meaning\' section.  '),(15806,3523319,'15,July,2004','2004-07-15 00:00:00','   Dude, that imagery has made me laugh out loud.  Now my boss thinks I\'m weird.  Well, weirder.  '),(15807,3523319,'15,July,2004','2004-07-15 00:00:00','   see, i told you he was still moving. someone take that spade out of his head.  '),(15808,3523319,'15,July,2004','2004-07-15 00:00:00','   Is that because of the people that work in them?&nbsp; Tee hee  '),(15809,3523319,'15,July,2004','2004-07-15 00:00:00','   Dude, I\'m, like, totally getting my head around these Japanese verb conjugations. That rawks. Dude.  '),(15810,3523319,'15,July,2004','2004-07-15 00:00:00','   Eeeexcellent. My boss is out all day, and the only person who sits behind me has just left early, meaning no-one can see me surfing the interweb all afternoon.  To be specifimic, Fonz, I now live on Warwards Lane, which is the other side of Raddlebarn from Hubert and Tiverton and that.  I work almost in Aston, the other side of Aston University, about 15 mins from the city centre. It is a shitty area. If this were SimCity, it would definitely be light industrial, with quite a lot of black buildings.  '),(15811,3523319,'15,July,2004','2004-07-15 00:00:00','   Gosh darn it, I demand details.  Which street do you live in?  I recall somethhing about the Bristol Road...  And where do you work?  I mean WHERE? You know, geomagraphically.  I haven\'t done anything all day.  And I don\'t care.  '),(15812,3523319,'15,July,2004','2004-07-15 00:00:00','   look on the bright side: at least you still have a soul to extinguish. you could have become a corporate shill.  '),(15813,3523319,'15,July,2004','2004-07-15 00:00:00','   I still live in Selly. I still work at OCR.  It is slowly extinguishing my soul.   Helen was also temping, until our agency screwed things up a royal treat yesterday. Now she is waiting for a new assignment.  '),(15814,3523319,'15,July,2004','2004-07-15 00:00:00','   Well, you only have to leave them a couple of days without washing them and hey presto - Life!  It\'s easy really, I don\'t know what all the fuss is about.  Why don\'t the people who live on other planets just hurry up and invent life like we did on Earth?  '),(15815,3523319,'15,July,2004','2004-07-15 00:00:00','   judging from that article, there may well be life in my socks.  '),(15816,3523319,'15,July,2004','2004-07-15 00:00:00','   I found it interesting, and then being the sensationalist I am, I have told everyone that they\'ve found life on Mars!  When asked to expand I told them they\'d found a little man.  He was just like a man, but his eyes were in the back of his head. And he was little.  I don\'t think people are believing me.  Ross, where do you live and work now?  And what does Helen do now?  '),(15817,3523319,'15,July,2004','2004-07-15 00:00:00','   I just read  urlLink this , and I\'m a little ashamed how interesting I found it.  '),(15818,3523319,'15,July,2004','2004-07-15 00:00:00','   I\'m gonna go ahead and assume \"almost\" is the most important word there.  Now make my site prettier.  '),(15819,3523319,'15,July,2004','2004-07-15 00:00:00','   i have some snazz on a sponge somewhere.  someone\'s been flooding our network with udp packets. it\'s disrupted all the printers. we need a packet sniffer, stat.  you know, it\'s almost exciting here some days.  '),(15820,3523319,'15,July,2004','2004-07-15 00:00:00','   I assume that\'s what they\'re talking about.  After all, that\'s exactly how Arsene has been hanging on to him all these years.  Make my website look a bit snazzier dave.  '),(15821,3523319,'15,July,2004','2004-07-15 00:00:00','   From BBC Sport: \"Real Madrid could use Fernando Morientes as bait to lure Arsenal\'s Patrick Vieira.\" In what way, exactly? Morientes in hotpants, slapping his arse and looking coyly over his shoulder with fluttering eyelashes?  '),(15822,3523319,'15,July,2004','2004-07-15 00:00:00','   i\'ve got an apple, thanks. your biscuits are soggy, anyway.  '),(15823,3523319,'15,July,2004','2004-07-15 00:00:00','   Would you like a biscuit?  '),(15824,3523319,'15,July,2004','2004-07-15 00:00:00','   Oh yeah.  Well I never liked the rest of those freaking losers anyway.  '),(15825,3523319,'15,July,2004','2004-07-15 00:00:00','   i fear that my mind has wandered into dark places. again.  '),(15826,3523319,'15,July,2004','2004-07-15 00:00:00','   She\'ll probably prefer it delivered fresh, rather than potted up I\'m betting.  And you know I can produce as many pots a day as required, so I\'m ideal for light snacks between meals.  '),(15827,3523319,'15,July,2004','2004-07-15 00:00:00','   we appear to be the only two people here this morning.  let\'s get down.  '),(15828,3523319,'15,July,2004','2004-07-15 00:00:00','   you\'ll need a pot for breakfast, a pot for lunch, and a proper dinner.  i\'ve got two pots right here.  '),(15829,3523319,'15,July,2004','2004-07-15 00:00:00','   Yes, I have a plan to back over this Oirish turd she\'s dating with my car, buy her flat and rent it to her, and then feed her up on a steady diet of man-phat.  Who\'s with me?  '),(15830,3523319,'15,July,2004','2004-07-15 00:00:00','   with all this king arthur nonsense going on, my knightley shares are beginning to look more and more like the jizz-sponges i always knew they were.  has anyone else seen arena this month? she really is very, very thin.  '),(15831,3523319,'15,July,2004','2004-07-15 00:00:00','   Oooh, there\'s going to be some serious knee-capping in the \'Daq on Monday.  Dave and I need to make an extra 500 grand by Monday to double our winnings, but it seems unlikely.  '),(15832,3523319,'15,July,2004','2004-07-15 00:00:00','   That throwing and catching shit is overrated anyway.  '),(15833,3523319,'15,July,2004','2004-07-15 00:00:00','   i\'vr had beer, and softball.  Coclusions.  I can\'t field for shit but i can twat the ball pretty well  '),(15834,3523319,'14,July,2004','2004-07-14 00:00:00','   i get it  i\'ve had wine. again.  '),(15835,3523319,'14,July,2004','2004-07-14 00:00:00','   In his btoilet?  '),(15836,3523319,'14,July,2004','2004-07-14 00:00:00','   this blog is shit and for gays.  there, i\'ve said it. where\'s ross\' blog?  '),(15837,3523319,'14,July,2004','2004-07-14 00:00:00','   It seems I\'ve  urlLink started a Blog  in my sleep in the manner and style of a really geeky Tyler Durden.  '),(15838,3523319,'14,July,2004','2004-07-14 00:00:00','   sod it. i\'ll save my money and energy for tomorrow night. i\'ve got some mince that\'s about to go off anyway.  that\'s right, mince. as in \"oooooh\".  that fuggin\' lexie. i\'m going to the gym.  '),(15839,3523319,'14,July,2004','2004-07-14 00:00:00','   No-one at all stupido dave!  Up to you whatchoodo...  '),(15840,3523319,'14,July,2004','2004-07-14 00:00:00','   Right, I\'m confused.  How\'s Dave supposed to shout in Lex\'s ear if she\'s gone home?  Why is it her penultimate day at work?  What the hell is going on?  I\'ve just been filming sheets of glass exploding.  They made me jump.  '),(15841,3523319,'14,July,2004','2004-07-14 00:00:00','   oooh, that fuggin lex. i can\'t believe she\'d do that on her penultimate day at work. will anyone else be there? i might go home and drink wine after all...  '),(15842,3523319,'14,July,2004','2004-07-14 00:00:00','   Lex has done that thing where she goes home ill...Let me know if you are still up for Angel Mr Wragg...I invited a bambi along on monday but he said he thinks he may be hitting people with sticks.  '),(15843,3523319,'14,July,2004','2004-07-14 00:00:00','   i just booked me a hotel room, a day and a half off work and part of a rail ticket for the weedave wedding spectacular. i feel all like a grown-up now.  i take it that it\'s cool if i turn up in jeans and flip-flops then?  '),(15844,3523319,'14,July,2004','2004-07-14 00:00:00','   Today is going frustratingly slowly. I want to go home.  '),(15845,3523319,'14,July,2004','2004-07-14 00:00:00','   i have to stay off the messenger or the bogtrolls get me. that is:  our isa-server and firewall block many ports and protocols, and messenger is among them. however, the managing director and the chairman want to do things like use messenger and listen to hebrew radio, so there has to be a small ip range that sits outside the protected band for them to occupy.  i\'m not supposed to be on that range... but... i am. so i can use messenger, but if i\'m spotted, then it\'ll be immediately obvious that i\'m not going through the firewall, and that my web access is unrestricte and i\'m running kazaa all the time.  that would be a negative. having said that, it\'s lunch time at the moment, so you never know your luck.  '),(15846,3523319,'14,July,2004','2004-07-14 00:00:00','   Yes, if only they\'d taught us something useful, like how to spell the word practical.  '),(15847,3523319,'14,July,2004','2004-07-14 00:00:00','   Dave why are you never on messenger these days?  That Onion article got many cynical chuckles from me. I think we should send it to all our compski lecturers. I cant think of a single practicle thing they ever taught us.  '),(15848,3523319,'14,July,2004','2004-07-14 00:00:00','   i just finished reading that article, too. although i was pretty sure i knew where it was heading after reading the headline...  tee hee. no practical skills.  '),(15849,3523319,'14,July,2004','2004-07-14 00:00:00','   tomorrow night is party night. come to the griffin at clerkenwell or my flat or something at some point.  you love the griffin.  oh yes you do.  '),(15850,3523319,'14,July,2004','2004-07-14 00:00:00','   urlLink This one  hits a little close to home.  '),(15851,3523319,'14,July,2004','2004-07-14 00:00:00','   Yeah im not a big fan of punctuation. Or grammar. Or spelling. I tend to pretty much make it up as i go along.  '),(15852,3523319,'14,July,2004','2004-07-14 00:00:00','   It happens to us all occasionally.  I\'ve used and configured Apache/Tomcat, IIS and Resin, and I have to say IIS comes out as the worst of the three in my opinion.  It\'s arguably easier to configure than Tomcat, and definitely easier than Resin, but other than that it sucks ass. It\'s not very powerful and, like ASP, I find it a bit clumsy.  I like to Be In Control.  Plus IIS probably isn\'t very secure.  That Onion article from 1969 is pretty funny this week.  '),(15853,3523319,'14,July,2004','2004-07-14 00:00:00','   That\'s all useful information.  Wait, no it\'s not. You\'ve all said contradictory things. Perhaps I just won\'t bother.  People keep trying to get me to use FireFox, but it reeks of socks and sandals, so I don\'t want to.  Also, Wee Davey, has your apostrophe key broken, or are you just allergic to them?  '),(15854,3523319,'14,July,2004','2004-07-14 00:00:00','   This silly site said the URL was invalid but had apparently posted my message anyway  '),(15855,3523319,'14,July,2004','2004-07-14 00:00:00','   i\'ve edited your post, dave, like a little imp. tee hee.  '),(15856,3523319,'14,July,2004','2004-07-14 00:00:00','   Right I see we need a proper compski here and I\'m the properest I know.  Step 1: Stop using Outlook and ie cause there insecure and generally shit. Use Mozilla Firefox(browser) and Thunderbird(mail client) especially Firefox cause it has tabbed browsing, blockes all popup adverts and wont let spyware on to your machine via activex. Thunderbird doesn\'t auto open virus ridden attachments either. There both basically great.  Part the second: Compski as I am I still cant love linux. You\'ll spend a week finding the text files you need to edit to make things work then another week figuring out how to edit them. Everythings painful and even though you think your learning things youll forget it instantly and besides what youve learnt wont work on another distro which brings me to the next thing about Linux. Choose your distro wisely. I reconmend SuSE cause it has some graphical tools for configuring things. But overall you may well be better off with winxp and and windows apache client.  As for web servers Apache and Tomcat are ok but you still have the annoying impenetrable text file based config files.  Ive never used IIS I suspect its insecure but easy to config.  Dave with Yoav you should get him on one of those reality shows where they dump someone in a jungle and get them to fend for themselves. When he cant use his psychic link to Zodiacs Wobu slaves to sort out the festering snake bite on his leg he will either learn really quickly how to organize his own life or you wont have a problem with him any more! I bet we can forge a will signing his flat over to you.  '),(15857,3523319,'14,July,2004','2004-07-14 00:00:00','   What\'s the point of all this anyway?  Ross doesn\'t know any HTML.  You can\'t make a website like that.  '),(15858,3523319,'14,July,2004','2004-07-14 00:00:00','   hosting .NET is free as long as you have a windows server 2003 licence. which most people don\'t.  then there\'s PHP.  '),(15859,3523319,'14,July,2004','2004-07-14 00:00:00','   You can use Outlook to connect to hotmail.  '),(15860,3523319,'14,July,2004','2004-07-14 00:00:00','   Yeah, this free stuff is all well and good, but it\'s mainly geard towards the use of Java and JSP.  So if that\'s what you\'re after, then go for it.  If you\'re only after regular HTML, then anything\'s good for that.  Otherwise, you could get some .NET action going on, but that\'s probably far from free.  I mean, hosting .NET is free I think (?), but the tools for developing it aren\'t.  Then there\'s PHP.  '),(15861,3523319,'14,July,2004','2004-07-14 00:00:00','   resin be free for personal use and small companies, but not if you\'re the size of zodiac. that\'s why we had to switch to tomcat. that and the fact that resin can\'t handle struts.  turns out that wobu\'s insurance has also expired, and he has no vehicle registration documentation. the mind boggles...  '),(15862,3523319,'14,July,2004','2004-07-14 00:00:00','   Right, I\'ll shower tou with my opinions and gloss over the facts:  I use Outlook for my regular mail, and I think it\'s more secure than Outlook Express.  However, you can\'t use it to connect to hotmail, so I use outlook express to connect to one of my hotmail accounts. I have no idea if it\'s secure or not.  As for servers, I have no preference for operating system.  Linux is supposed to be good and free, and is probably worth trying out at least.  Is this going to be a server for serving up web pages and the like then?  Have you any idea what sort of site you want and what technology you\'re going to use?  If you\'re after free stuff, I guess Tomcat and Apache are the way to go, but in my opinion, Resin is still a better server.  Not sure if it\'s free or not for personal use.  That\'s not much help, is it?  '),(15863,3523319,'14,July,2004','2004-07-14 00:00:00','   1: murunno. outlook\'s prolly better. 2: murunno. prolly. everyone lubs linux. and apache\'s fine. we use tomcat here, which appears to work.  i am the crown prince of noncomittal compsc.  '),(15864,3523319,'14,July,2004','2004-07-14 00:00:00','   nope, i didn\'t pay for it - the company did. apparently his dad will settle up when he\'s back from his latest business trip. but the thing that gets me is:  wobu will never know.  when he gets back in september, he might just notice that one of his tyres has changed shape or that wheel is no longer yellow, but i doubt he\'ll think anything of it. cos, you know, stuff like that is always happening.  '),(15865,3523319,'14,July,2004','2004-07-14 00:00:00','   I have some right proper compsci questions, the answers to which I can\'t be arsed to find myself. So do my thinking for me.  Question the first: Is Outlook Express still riddled with security problems, or is it okay to use now? I remember Microsoft recommended you didn\'t use it to connect to Hotmail, but that was years ago now.  Question the second: I\'ve realised that I have enough old computer parts to build another machine, so I was thinking of putting together an interweb server. I remember people talking about Linux and Apache. Is that still the best (free) way?  '),(15866,3523319,'14,July,2004','2004-07-14 00:00:00','   Right, I don;t know if this\'ll work, the site is gaying me up a treat today.  Do I understand then Dave, that you paid for the work on Yoav\'s car?  Why?  Why would you do that?  Why can\'t the Mecha-Zinch stump up?  Also, I suck pretty bad at that golf game, but not as badly as Ross. I do pronounce it balls though. If that were a real golf course, I don\'t think anyone would ever get round it.  Not without golf balls that float and a decent boat anyway. That game is nowhere near as good as  urlLink this one .  '),(15867,3523319,'14,July,2004','2004-07-14 00:00:00','   Wow, I suck super hard at that golf game. I gave up before finishing the round. I didn\'t see the point as I was already 30 over par.  '),(15868,3523319,'14,July,2004','2004-07-14 00:00:00','   Sooner or later, Yoav will kill himself.  For most people this would be a comment in jest...in this case I 100% believe it.  '),(15869,3523319,'14,July,2004','2004-07-14 00:00:00','   Sooner or later, Yoav will kill himself.  For most people this would be a comment in jest...in this case I 100% believe it.  '),(15870,3523319,'14,July,2004','2004-07-14 00:00:00','   shaddap youse - you\'re older than me. well, you look like you are. it\'s the hairy godzilla face, you see.  i\'m accidentally a bit hungover today. i should probably go to angel this evening and shout in damo\'s ear for a couple of hours. that\'ll help.   a story about woab\'s car once upon a time, woab went to new york, leaving his car in the garage of our building. his dad came up to me at work on monday and asked me to bring in the car keys and the widget to open the garage doors so that they could take his car to get a service in his absence. fair enough, i suppose - i brought the keys in on tuesday.  yesterday morning, the bloke who\'s sort of the official driver for the office came over to my desk and asked me for the address and instructions of how to get there etc., as he\'d been told to spend his working day getting yove\'s car sorted out. fair enough, i muttered, and gave him directions.  he came and found me again at about 4 yesterday afternoon. it turns out that he\'d taken the yovemobile to the garage, and there they had discovered a few things:    he\'d been driving around on a flat tyre, and had destroyed it, meaning a new tyre   he\'d also been driving around with the emergency tyre on another wheel, left over from the puncture before last, for probably a month   his rear brake pads were 95% worn through   one of the rear wheel bearings was cracked and buggered - the entire hub had to be replaced   his MOT was two months out of date   he\'d had no road tax for over a month    the total bill for the work done, including road tax and vat, was Â£880-odd.  fair enough, i hissed through gritted teeth. someone has to take care of these things...  '),(15871,3523319,'14,July,2004','2004-07-14 00:00:00','   urlLink This  looks like the said golf game to me.  '),(15872,3523319,'14,July,2004','2004-07-14 00:00:00','   How can you not see the adverts THEY\'RE RIGHT THERE.  the golf is on the BBC sport site, you should be able to find it quite easily.  I\'ll attempt to improve on my score later, i\'m sure i can shave two or three strokes off  '),(15873,3523319,'14,July,2004','2004-07-14 00:00:00','   I have no adverts.  I\'d like to see The Cooler, but I can\'t seem to interest anyone else in going. William H. Macy looks like Dave did five years ago.  Ho ho ho. Dave - he\'s very old.  '),(15874,3523319,'14,July,2004','2004-07-14 00:00:00','   I\'m not quite sure how it has happened, but I don\'t see the adverts anymore, not at home or at work.  They must get blocked by some sort of ant-virus thing maybe...  What\'s the address of this golf game that you so badly want me to beat you at then Manx?  Or am I going to have to look for it myself like some kind of chump?  I watched The Cooler last night.  It was pretty good.  It reminded me of all the other Vegas movies I\'ve seen, like Swingers and Very Bad Things, but most of all it reminded me of Leaving Las Vegas.  All spam is shit and for gays.  '),(15875,3523319,'14,July,2004','2004-07-14 00:00:00','   oh, they\'ve gone now.  Nobody will believe me that they were there  '),(15876,3523319,'14,July,2004','2004-07-14 00:00:00','   Errr, why have we got adverts for \'makeup4sisters\' and afro hair clinics now.  Whose fault is that  '),(15877,3523319,'14,July,2004','2004-07-14 00:00:00','   i read an article on the register about how easy it is to send sms spam and i became concerned. they need to Do Something About That, and quick.  i\'ve had a lot of wine and some whisky too. i don\'t see myself at yoga and 7am tomorrow, somehow. tee hee good intentions melt away...  '),(15878,3523319,'13,July,2004','2004-07-13 00:00:00','   You know what I hate? Fucking SMS spam. I seem to be getting more and more of that shit.  '),(15879,3523319,'13,July,2004','2004-07-13 00:00:00','   The interweb is back at home, where it belongs. Now, where did I leave my Anarchy?  '),(15880,3523319,'13,July,2004','2004-07-13 00:00:00','   5 under par.  I fucking own that game  '),(15881,3523319,'13,July,2004','2004-07-13 00:00:00','   Ha Ha, par is mine  '),(15882,3523319,'13,July,2004','2004-07-13 00:00:00','   I just got my best score on the BBC golf game.  42, now to break par  '),(15883,3523319,'13,July,2004','2004-07-13 00:00:00','   it\'s trev\'s birthday on saturday. he will be 688 years old.  '),(15884,3523319,'13,July,2004','2004-07-13 00:00:00','   now i think about it, to see what damo has to say about his favourite pornography. and ross, for that matter; i imagine that he can get all the thrills he wants just by lifting his kilt and peering downwards any time he likes, so it may be an interesting answer.  '),(15885,3523319,'13,July,2004','2004-07-13 00:00:00','   Yes, I think all my porn has naked ladies in, so that must be my favourite sort.  '),(15886,3523319,'13,July,2004','2004-07-13 00:00:00','   Have you forgotten that Damo uses this page Dave.  Are you asking to be disgusted and revolted by raising the question you just raised.  '),(15887,3523319,'13,July,2004','2004-07-13 00:00:00','   what\'s the best kind of porn. i think it is the kind of porn that has naked ladies in it. what do you think?  '),(15888,3523319,'24,July,2004','2004-07-24 00:00:00','   i saw the headline, but couldn\'t check it from work... now i\'ve lost the link, as it\'s at work. someone post it and we can find out.  i\'m feeling veddy, veddy bad today. veddy bad indeed.  but probably not as bad as ed did last sunday.  '),(15889,3523319,'23,July,2004','2004-07-23 00:00:00','   I does not arouse me.  Dave, did you see Omar from Television X in popbitch this week?  Was that him?  '),(15890,3523319,'23,July,2004','2004-07-23 00:00:00','   I just watched a Gamer TV excerpt of Doom3. It looks like it will be awesome like about four thousand hotdogs, and by the sounds of things, you won\'t need a complete bastard of a graphics card to play it, either.  I\'m moist.  '),(15891,3523319,'22,July,2004','2004-07-22 00:00:00','   i\'m about to go to the walkabout at temple with all the work lot. damo, you\'re welcome to join me - and bring anyone else who might be around.  there will be booze.  '),(15892,3523319,'22,July,2004','2004-07-22 00:00:00','   It\'s a warehouse, where they house wares. The top floor has been converted to offices, but when they did that they made the wise move of bricking up all the windows and covering all the skylights.   The light comes as quite a shock to my eyes when I leave of an afternoon. Seriously.  '),(15893,3523319,'22,July,2004','2004-07-22 00:00:00','   ross, do you live in a cube? come on, you can tell me.  '),(15894,3523319,'22,July,2004','2004-07-22 00:00:00','   I have no windows in my office. For all I know, the sky could have fallen in.  '),(15895,3523319,'22,July,2004','2004-07-22 00:00:00','   i just read the week 1 boa archive, to see if i could. times were better then.  '),(15896,3523319,'22,July,2004','2004-07-22 00:00:00','   urlLink aaarrrrr .  and cambridge just signed a frenchman called Mongin.  '),(15897,3523319,'22,July,2004','2004-07-22 00:00:00','   i walked down the Sunny Side of the road on my way to work today. there\'s nothing quite like the feeling you get spending eight hours in an air-conditioned office with the blinds drawn during one of the nicest days for ages.  mmm, surf that wave of satisfaction. my fingers are chilly.  catwoman looks like a shit film.  '),(15898,3523319,'22,July,2004','2004-07-22 00:00:00','   Maybe he gets to bonk Halle Berry as part of the deal?  I hate you guys, you guys are assholes, especially Dave I hate him the most.... All together now!  '),(15899,3523319,'22,July,2004','2004-07-22 00:00:00','   Bah, my identity has been discovered.  Dude, the weather here is awesome. Like 10 hot dogs.  I\'m going to sit in my garden and nurse my gut rot.  '),(15900,3523319,'22,July,2004','2004-07-22 00:00:00','   I read that the first time and thought it said \'Catman agrees\'.  Maybe that\'s Fonz\'s crime-fighting identity.  '),(15901,3523319,'22,July,2004','2004-07-22 00:00:00','   damn that fonz\'s summer hours.  '),(15902,3523319,'22,July,2004','2004-07-22 00:00:00','   Screw you guys, I\'m going home.  Heh, fags!  '),(15903,3523319,'22,July,2004','2004-07-22 00:00:00','   Wasn\'t implying you were Mr Wragg...more a general comment in response to young master fonz.  I just went to  urlLink www.aquarterof.co.uk  and purchased Â£25 worth of sweets for delivery to my office in 5 days time.&nbsp; The guy opposite me did the same....we now have 6.5kg of various sweets arriving - Nice.  '),(15904,3523319,'22,July,2004','2004-07-22 00:00:00','   the sql will be gone soon, and then none of us need mention it again.  '),(15905,3523319,'22,July,2004','2004-07-22 00:00:00','   Perhaps \'amusing\' isn\'t the word I was looking for there. \'Perfunctory\', maybe.  Also, Dave, your SQL has made the BOA look all squiffy for me. Bad Dave.  '),(15906,3523319,'22,July,2004','2004-07-22 00:00:00','   I saw a telly programme a few weeks ago called \'Train Cops\', or something, following Transit Police about their business. It went between stations all over the country, and showed some amusing national sterotypes.  In Lahndahn - some blokes are having a scrap on the platform, and when the Train Cops turn up, some woman goes mental and has to be arrested.  In Brum - some thieving scally nicks a woman\'s suitcase while pretending to help her carry it.  And, of course, in Glasgow - a skaghead is caught shooting up in the station toilets.  '),(15907,3523319,'22,July,2004','2004-07-22 00:00:00','   is it ok if you have imaginary friends?  s\'orrit - wasnae drunkie, just making pes harder with a couple of drams. och.  '),(15908,3523319,'22,July,2004','2004-07-22 00:00:00','   Some \"amusing\" quips there, but joking aside..drinking on your own to get drunk is a bad thing to get in the habit of. Anyhow...I\'m off to get hammered...nice  '),(15909,3523319,'22,July,2004','2004-07-22 00:00:00','   I\'m kneedeep in Java and gut-rot and i\'m thinking of leaving work in half an hour.  Brilliant, eh?  '),(15910,3523319,'22,July,2004','2004-07-22 00:00:00','   urlLink fonz .  i\'ve got my grid in order. it is ponderous and slow, but the data\'s all there. i\'m the king or something.  today, i am mostly digging my own hole.  '),(15911,3523319,'22,July,2004','2004-07-22 00:00:00','   I\'ve spent the last few hours half-heartedly shuffling the labels around on a report I\'m supposed to be making. I\'m \'perfecting the layout\'. Tee hee.  '),(15912,3523319,'22,July,2004','2004-07-22 00:00:00','   I look forward to visiting Dave in a few years time, when he\'s the reeking, shouting, swearing, piss-soaked wino living outside a Tube station.  '),(15913,3523319,'22,July,2004','2004-07-22 00:00:00','   Don\'t listen to that dead guy, drinking alone is fine.  I tell you what\'s not fine.  Moving around after death. That\'s what.  '),(15914,3523319,'22,July,2004','2004-07-22 00:00:00','   I had a quick peruse of your SQL, and it looks quite hard, and needs more think-power than I can muster today. So up yours, Delors.  Is that how you spell Delors?  '),(15915,3523319,'22,July,2004','2004-07-22 00:00:00','   Drinking Whisky or Wine on your own such that you get drunk is pretty wrong...if you are only having one or two glasses it\'s ok though... I watched Eternal Sunshine of the Spotless Mind and Shrek2 on DVD last night...the first was fucking brilliant and Shrek2 was&nbsp; a bit pooey really.  '),(15916,3523319,'22,July,2004','2004-07-22 00:00:00','   i was going to drink wine last night, but i exercised some self control and didn\'t.  so instead i drank whisky.  i may have a problem. but it\'s cool while you\'re young, right?  ross, hold off on the kanji until you\'ve given me an elegant solution to my combined queries. i demand it.  '),(15917,3523319,'22,July,2004','2004-07-22 00:00:00','   I\'ve decided I don\'t really fancy doing any work today. I think I may spend most of the day learning my Stage 1 Jouyou Kanji. Yes.  '),(15918,3523319,'22,July,2004','2004-07-22 00:00:00','   Bleurgh.  This hang over shit is overrated.  '),(15919,3523319,'22,July,2004','2004-07-22 00:00:00','   Sorry, I went to the bank, and then I went home.  If you\'re lucky, I might look into your problem tomorrow at work. If you\'re not, I\'ll just ignore it.  '),(15920,3523319,'21,July,2004','2004-07-21 00:00:00','   i\'ve got seven pages of generated SQL and an invalid column index. i bet alanis morisette would consider that ironic.  '),(15921,3523319,'21,July,2004','2004-07-21 00:00:00','   i\'ve killed the wee man with my SQL of death.  '),(15922,3523319,'21,July,2004','2004-07-21 00:00:00','   come on, ross, stop trying to hit me and hit me.  '),(15923,3523319,'21,July,2004','2004-07-21 00:00:00','   Nice. I was too lazy tyo check the source.  '),(15924,3523319,'21,July,2004','2004-07-21 00:00:00','   &lt;PRE&gt;...&lt;/PRE&gt;, which is Preformatted text.  '),(15925,3523319,'21,July,2004','2004-07-21 00:00:00','   Yeah, that\'s too complicated for me.  How did you get the code to look like code?  What tag did you use?  '),(15926,3523319,'21,July,2004','2004-07-21 00:00:00','   [taps foot, looks expectant]  '),(15927,3523319,'21,July,2004','2004-07-21 00:00:00','   yes, it is SQL. i have a grid of results 21 columns wide, and each column has its own query, as they all pull slightly differently filtered data from the same tables, with slightly different joins. the rows are generally, but not always, the same, is it\'s to do with attendance and processing figures from the zodiac remote offices - one count query per office per column, and some offices don\'t show up in every single one.  aaaanyway, i need to combine the individual queries into a single one, while keeping the data produced separated with different column headings. i\'m thinking that i\'ll union each resultset with the next, keeping the SQL distinct, but in order to union two queries they have to have the same column headings, which means that i\'ll have to fabricate column headers on each query for all but the column that is being populated.  of course, just to make things interesting, the queries are being built out of little modular blocks of my own creation, as (because of the different ranks involved) each query is effectively run 3 times, with one line different in the Where clause but everything else the same. so the whole thing needs to generate in a loop.  here\'s an example of one of the queries - not too complicated:  /*GET NEW APPLICANT OFFICERS*/ select i.description, count(*) new_applicant_officers  from persons p, identifiers i, persons_history ph  where p.persons_rank in (\'CPT\',\'C/O\',\'2/O\',\'3/O\',\'C/E\',\'1/E\',\'2/E\',\'3/E\',\'EE1\',\'EE2\',\'EE3\') 	  and p.source_code in (\'BUL\',\'BOT\',\'OZR\',\'EST\',\'ODS\',\'TPI\',\'LAT\',\'BEO\',\'CNT\',\'DIA\',\'ESP\',\'BEG\',\'TGL\',\'DON\') 	  and p.source_code = i.code 	  and i.identifier = \'SOURC\' 	  and p.persons_code = ph.persons_code 	  and ph.status_code = 0 	  and to_char(ph.timestamp, \'YYYYMM\') = ? group by p.source_code, i.description  and i need to combine it with the next one, which looks like this:  /*GET NEW JOINER OFFICERS*/ select i.description, count(*) new_joiner_officers  from persons_voyages pv, persons p, identifiers i  where p.persons_code = pv.persons_code 	  and pv.sign_off_date is null 	  and pv.joining_rank in (\'CPT\',\'C/O\',\'2/O\',\'3/O\',\'C/E\',\'1/E\',\'2/E\',\'3/E\',\'EE1\',\'EE2\',\'EE3\') 	  and p.source_code in (\'BUL\',\'BOT\',\'OZR\',\'EST\',\'ODS\',\'TPI\',\'LAT\',\'BEO\',\'CNT\',\'DIA\',\'ESP\',\'BEG\',\'TGL\',\'DON\') 	  and p.source_code = i.code 	  and i.identifier = \'SOURC\'   	  and to_char(pv.joining_date, \'YYYYMM\') = ? 	  and not exists  	  	  (select * from persons_voyages pv2  		  where pv2.sign_off_date is not null  		  		and pv2.persons_code = p.persons_code)  group by p.source_code, i.description  and then five more like it. note how some of the tables are the same, making a standard join impossible without table aliasing, and i\'d have to do it at least seven, and possibly 21 times. it\'s also worth noting that these two queries are in fact six, because each will be run three times (with different rank codes in the persons_rank collection).  so, wee man, help me.  '),(15928,3523319,'21,July,2004','2004-07-21 00:00:00','   I was about to boast of my SQL prowess, but then I remembered I can\'t do any of the complicated stuff involving nested queries, temporary tables or stored procedures.  So I guess that makes me a simple farm boy.  '),(15929,3523319,'21,July,2004','2004-07-21 00:00:00','   What type of query? Is it SQL? \'Cos I\'m \'tha bomb\' at SQL now. Yes.  '),(15930,3523319,'21,July,2004','2004-07-21 00:00:00','   someone combine my layered query for me.  '),(15931,3523319,'21,July,2004','2004-07-21 00:00:00','   i believe it came out a couple month ago but only really floated into my awareness when i saw an ad for it in a magazine last week  '),(15932,3523319,'21,July,2004','2004-07-21 00:00:00','   when did that one come out? i had a look at the beta band web page once, and found it disappointing.  '),(15933,3523319,'21,July,2004','2004-07-21 00:00:00','   Riiight.  Maybe I\'ll look into it.  '),(15934,3523319,'21,July,2004','2004-07-21 00:00:00','   You sent me the 3 ep\'s.  most of which i already had.  I\'ve got bits and pieces of \'the beta band\' on mp3.  I\'ve had Hotshots II on CD for literally yeeeeeeaaars.  WHAT I want heroes to zeroes (i.e. the new one)  '),(15935,3523319,'21,July,2004','2004-07-21 00:00:00','   What?  Is your e-mail working yet?  '),(15936,3523319,'21,July,2004','2004-07-21 00:00:00','   don\'t forget hot shots ii.  '),(15937,3523319,'21,July,2004','2004-07-21 00:00:00','   Which Beta Band album are you after? I\'ve only got the Three EPs, which I hope I sent to you.  But I might not have.  '),(15938,3523319,'21,July,2004','2004-07-21 00:00:00','   i did get it thank you, haven\'t had much ofx a chance to listen to its yet.  Now i think you should download the Beta Band album and send me it seeing as the version i downloaded don\'t woik.  Either that or just buy the album and get it posted to my address.  '),(15939,3523319,'21,July,2004','2004-07-21 00:00:00','   Oh yeah, that reminds me - I sent you a cd the other day manx, and forgot to tell you.  Did you ever get it?  I hope I got your address right.  '),(15940,3523319,'21,July,2004','2004-07-21 00:00:00','   I almost bought about 5 albums last night.  then i just downloaded them instead.  must have saved me a shedload of money  '),(15941,3523319,'21,July,2004','2004-07-21 00:00:00','   I didn\'t get no damn stinking picture of no damn stinking golf bats. Damn.  I had a dream last night that was the sort you can get on DVD if you know the right people.  It was fun.  '),(15942,3523319,'21,July,2004','2004-07-21 00:00:00','   i got my sticks. i\'m not hungover, but i am tired. i took a photo of my sticks and tried to send it to ross and fonz last night, but i have no idea whether it worked or not. my computer has been reformatted again and everything is wrong on it. i have a curious ache in my left shoulder. it\'s not too hot but it is too humid. last night i had a dream that i went to the doctor to complain that my pants were too tight, and began describing how in graphic detail to a flustered nurse. then carwyn appeared on my shoulder and grinned at her. i had a much better dream the night before - the kind of dream that i\'d like to get on dvd were it to become available.  i think i\'m ready for some cawfee now.  '),(15943,3523319,'21,July,2004','2004-07-21 00:00:00','   Bwahahahaahahah.  Now who\'s kicking whose ass across the floor of the \'Daq, Ross?  '),(15944,3523319,'21,July,2004','2004-07-21 00:00:00','   urlLink This  is the AVI tool.  It\'s pretty good.  '),(15945,3523319,'21,July,2004','2004-07-21 00:00:00','   I realised I could just look at the BOA archive, now it has one. But I can\'t be arsed.  Is one of you trying to send me a MMS or something? My phone is struggling like it\'s trying to do a particularly large message-poo.  The declaration that Dave looks for every day in the paper is in the  urlLink bottom right-hand corner of the page .  '),(15946,3523319,'20,July,2004','2004-07-20 00:00:00','   am i an alcoholic yet?  the codec thing was fonz. i don\'t know what it is.  '),(15947,3523319,'20,July,2004','2004-07-20 00:00:00','   Can one of you porn monkeys give me a link for that codec thing you mentioned a while back? The one that lets you know which codecs you have, and lets you get the latest versions of them. Pur-lease.  '),(15948,3523319,'20,July,2004','2004-07-20 00:00:00','   mmmmgaararagh golfsticks  '),(15949,3523319,'20,July,2004','2004-07-20 00:00:00','   When I am the law, I reckon I\'ll make everything you all do or own illegal, and then get rich by turning you all in.  It\'s nothing personal, mind.  '),(15950,3523319,'20,July,2004','2004-07-20 00:00:00','   etc.  '),(15951,3523319,'20,July,2004','2004-07-20 00:00:00','   I done got it working.  '),(15952,3523319,'20,July,2004','2004-07-20 00:00:00','   The sort you have is.  I tried that download, and it all seemed to be going swimmingly, right up to the point where we rebooted the machine and the password change hadn\'t worked at all.  We ran the nt_pass tool on it and changed the admin password and the user password, and saved all the changes.  And nothing seemed to have changed at all.  Hmmm.  I wish I could read the FAQ for that utility, but it\'s blocked from here.  '),(15953,3523319,'20,July,2004','2004-07-20 00:00:00','   kalashnikov: bad nail clippers: good damo\'s nail bomb: bad ross: good modchips: bad showers: good  it\'s a good thing that porn isn\'t illegal.  '),(15954,3523319,'20,July,2004','2004-07-20 00:00:00','   Hmm, it appears I\'ve suddenly borken the law then.  So what\'s to stop people ruling that other stuff I have is now illegal to own? Seems like you could get stitched up pretty badly by this \"law\" stuff.  '),(15955,3523319,'20,July,2004','2004-07-20 00:00:00','   urlLink bad mojo   i\'ll check the link and edit the post below when i\'ve made it work. bear with me...  '),(15956,3523319,'20,July,2004','2004-07-20 00:00:00','   Dude, that link doesn\'t seem to work.  '),(15957,3523319,'20,July,2004','2004-07-20 00:00:00','   urlLink suck it down . go on.  it might even work this time.  '),(15958,3523319,'20,July,2004','2004-07-20 00:00:00','   Ahhh, I could do with you putting that file in your temp folder and giving me the link to it.  Cheers geezer.  '),(15959,3523319,'20,July,2004','2004-07-20 00:00:00','   but i\'ve been restarting my computer every five minutes so far today, and my email is all to cock. fonz, if you\'re waiting to hear from me, i can\'t run outlook at the moment. but i\'m working on it.  i hate computers.  '),(15960,3523319,'20,July,2004','2004-07-20 00:00:00','   and i\'d rather not talk about it.  '),(15961,3523319,'20,July,2004','2004-07-20 00:00:00','   Me too.  What happened to Dave?  '),(15962,3523319,'20,July,2004','2004-07-20 00:00:00','   I\'m bored. I want to go home.  '),(15963,3523319,'20,July,2004','2004-07-20 00:00:00','   I do like the look of that new Xbox. I can\'t really justify buying another one, though. Or afford to.  '),(15964,3523319,'20,July,2004','2004-07-20 00:00:00','   Wow, let\'s all get one of  urlLink these . It\'s so shiny.  '),(15965,3523319,'20,July,2004','2004-07-20 00:00:00','   Did you get my last e-mailer Dave?  About eugoogolising?  My e-mail is so gay. And I want to fix some people\'s computers.  '),(15966,3523319,'20,July,2004','2004-07-20 00:00:00','   I\'bve just deleted 618 e-mails from my inbox.  feel lighter somehow  '),(15967,3523319,'20,July,2004','2004-07-20 00:00:00','   You knows it.  I missed GoldieLookinChain at Glastonbury, which is a shame, because they were apparently very funny.  '),(15968,3523319,'20,July,2004','2004-07-20 00:00:00','   Bah, Ross is kicking my ass all over the trading floor of the Daq. All over it.  '),(15969,3523319,'20,July,2004','2004-07-20 00:00:00','   and balls.  i had wine.  '),(15970,3523319,'19,July,2004','2004-07-19 00:00:00','   Hee Hee...he said \"Joke\"  '),(15971,3523319,'19,July,2004','2004-07-19 00:00:00','   God damn it, my company has blocked redmeat again.  How am I supposed to entertain myself at work?  Cue favourite joke about wanking in the toilets at work.  '),(15972,3523319,'19,July,2004','2004-07-19 00:00:00','   That link thing happens to me sometimes.  Don\'t know why.  '),(15973,3523319,'19,July,2004','2004-07-19 00:00:00','   i remembered what i\'d written - i was horrified and self-loathing at the time, but couldn\'t be arsed to change it.  not sure why the links have moved, but i\'m pretty sure that it\'s damo\'s fault.  '),(15974,3523319,'19,July,2004','2004-07-19 00:00:00','   Also, why are the links on the right all the way down the bottom of the page?  Hur hur. I said bottom.  '),(15975,3523319,'19,July,2004','2004-07-19 00:00:00','   How did you see what he did? Did you remember, or is there some way to see changes to posts?  '),(15976,3523319,'19,July,2004','2004-07-19 00:00:00','   Anyone seen this  urlLink Star Ocean  game? I\'m looking forward to it as much as playing Kingdom Hearts. Man, chipped PS2\'s are the best thing since normal PS2\'s. Everyone should get one. Unless they have an X-box.   Dave, I changed your use of \"There\'s\" to \"There are\". Now tell me I was wrong.  '),(15977,3523319,'19,July,2004','2004-07-19 00:00:00','   jesus - i see what you did. and rightly so.  '),(15978,3523319,'19,July,2004','2004-07-19 00:00:00','   which one and how?  [furious]  '),(15979,3523319,'19,July,2004','2004-07-19 00:00:00','   I\'m not holding out much hope for Fallout 3. It\'s being done by Bathesda now. They make quite beardy games.  Holy shit I\'m bored this afternoon.  '),(15980,3523319,'19,July,2004','2004-07-19 00:00:00','   It\'s a different software house making it now, right?  I diddled Dave\'s message.  I hope that makes him angry.  '),(15981,3523319,'19,July,2004','2004-07-19 00:00:00','   there are&nbsp;strings attached. so we shall see.  '),(15982,3523319,'19,July,2004','2004-07-19 00:00:00','   Fallout 3 was officially announced last week as well...I imagine this makes certain people happy?  '),(15983,3523319,'19,July,2004','2004-07-19 00:00:00','   urlLink http://news.bbc.co.uk/1/hi/uk/3908215.stm  One dedicated pervert=DBF  '),(15984,3523319,'19,July,2004','2004-07-19 00:00:00','   how much shit is there to sort?  '),(15985,3523319,'19,July,2004','2004-07-19 00:00:00','   I went to the pub again, but my favourite bar maid was not there.  there was a different one with a CRACKING arse, but it was scant substitute.  Does that even make sense?  Anyway, I\'ll have to go again tomorrow until I get my weekly fix of ridiculously fit bird.  Am I supposed to be coming up to London on Friday then?  If that\'s the case, I might have to come back on Saturday so I can sort some shit out on Sunday.  '),(15986,3523319,'19,July,2004','2004-07-19 00:00:00','   I\'ve been playing boules and now the disc space nazis are aftre me.  must be a tuesday then  '),(15987,3523319,'19,July,2004','2004-07-19 00:00:00','   o why are the adverts offering me ally mcbeal dvds now? how many women has it detected here?  '),(15988,3523319,'31,July,2004','2004-07-31 00:00:00','   i bet you can\'t remember all the cheats.  i\'ve had a day of golf, and am pretty pleased, if physically broken.  '),(15989,3523319,'30,July,2004','2004-07-30 00:00:00','   I\'ve been playing Doom2 with a nifty little utility called jDoom. It\'s awesome like 5 hotdogs.  '),(15990,3523319,'29,July,2004','2004-07-29 00:00:00','   weekend. now.  '),(15991,3523319,'29,July,2004','2004-07-29 00:00:00','   Killing Charlie is a fundemental part of any American child\'s education.&nbsp; The first time&nbsp;an American&nbsp;child wakes up screaming \"They\'re in the trees...they\'re in the fucking trees\" it\'s cause for celebration within the family.&nbsp; It\'s kind of like the Barmitzvah (if that\'s how it\'s spelt) in Judaism,&nbsp;the Coming of age celebrations in Hinduism/Sikhism or your first car bomb in Islam. &nbsp; They celebrate with hot dogs and bud beers...it\'s true (see what I did there)...MTV told me.  '),(15992,3523319,'29,July,2004','2004-07-29 00:00:00','   I noticed that Kerry snuck in a bit about believing in science and supporting stem cell research. That surprised me, and is really going to piss off the Bible Belt. Damn God-botherers.  '),(15993,3523319,'29,July,2004','2004-07-29 00:00:00','   I mentioned in the office that I didn\'t think a presendential candidate\'s fondness of killing Charlie should really be the basis of voting for him. They looked at me like I was crazy.  '),(15994,3523319,'29,July,2004','2004-07-29 00:00:00','   let\'s go and watch a film that we know will be bad, for all the wrong reasons.  who\'s with me?  fonz did this one, didn\'t he?  '),(15995,3523319,'29,July,2004','2004-07-29 00:00:00','   That reminds me Dave, will we be able to see King Arfar on Sunday or not?  Doens\'t matter if not.  Right, I\'m sodding off up to London. Who\'s with me?  No one?  Right.  '),(15996,3523319,'29,July,2004','2004-07-29 00:00:00','   I just read John Kerry\'s nomination acceptance speech. Christ, he can talk.  God bless America.  '),(15997,3523319,'29,July,2004','2004-07-29 00:00:00','   \'Reminisce\'  Yes, I can.  '),(15998,3523319,'29,July,2004','2004-07-29 00:00:00','   everyone knows that sunday dvds are tfotr, ttt and rotk. ee\'s. back-to-back.  that\'ll learn it for being sunday.  '),(15999,3523319,'29,July,2004','2004-07-29 00:00:00','   can you spell \"reminisce\"? i\'m not sure i can.  '),(16000,3523319,'29,July,2004','2004-07-29 00:00:00','   I dun watched the Borne Supremacy yesterday. It was quite good. There were adequate car chase scenes, not (quite) enough beatings, more than enough who-am-I moments, and nowhere near enough boobs.  Worth watching, but possibly as a sunday afternoon DVD.  '),(16001,3523319,'29,July,2004','2004-07-29 00:00:00','   I\'d do all kinds of things with that sort of money.   Who am I kidding? I\'d be dead within a year.   I\'d spend all that money on irradicating people who type their whole e-mail message in the subject line and leave the e-mail blank.  '),(16002,3523319,'29,July,2004','2004-07-29 00:00:00','   i\'d hire a girl to lick my balls clean, 24/7. although lewis black had that idea first.  he\'s a genius.  '),(16003,3523319,'29,July,2004','2004-07-29 00:00:00','   *more* porn?  '),(16004,3523319,'29,July,2004','2004-07-29 00:00:00','   What would you do with 13 billion dollars Dave?  '),(16005,3523319,'29,July,2004','2004-07-29 00:00:00','   I concur with damo, ross and yoav. Which is what I was stating rather confusingly below.  So it\'s not \"Ewan\" then?  '),(16006,3523319,'29,July,2004','2004-07-29 00:00:00','   $45bn - $32bn = dave\'s christmas bonus, i fancy.  '),(16007,3523319,'29,July,2004','2004-07-29 00:00:00','   silly welsh.  '),(16008,3523319,'29,July,2004','2004-07-29 00:00:00','   Hmm, I don\'t think that\'s correct Fonz.    urlLink This  is a&nbsp;list of all the companies in the world making more than $100bn of revenues in a year. The top is Wal-Mart who make $266bn, which is around 140 billion quid. So I don\'t see anyone incurring as much cost as you have there.   Are you still working for the viagra people? They\'re raking in $45bn a year and spending around $32bn a year, which is around $60k a minute I think.  '),(16009,3523319,'29,July,2004','2004-07-29 00:00:00','   Hmm, having  looked into it (I am very bored) I couldn\'t find anywhere a rating of just how large this company is, or how many employees it has worldwide even. But I did find out that it\'s sales income is about 30 billion dollars a year, which makes me think that damo\'s theory about the million pounds a minute thing being rubbish is not rubbish.  In summary then, Chris.  '),(16010,3523319,'29,July,2004','2004-07-29 00:00:00','   I make it 577 billion pounds a year in outgoings.  Of course, this company is way bigger than Microsoft.  '),(16011,3523319,'29,July,2004','2004-07-29 00:00:00','   I keep reading about this guy  Ioan Gruffudd , but I have no idea how to pronounce his stupid name. Can someone tell me?  '),(16012,3523319,'29,July,2004','2004-07-29 00:00:00','   Maybe I done did the math wrong in my head but that\'s over Â£125billion a year before profit...Microsoft only do around $20billion...sounds fishy to me!  '),(16013,3523319,'29,July,2004','2004-07-29 00:00:00','   Yeah, that was awesome, but only like two or three hot dogs I reckon.  I hope Dave\'s right about the making profit equation.  I am freaking bored, and everyone here seems to have gone home on their summer hours scheme.  What shall I do to pass the next two hours?  '),(16014,3523319,'29,July,2004','2004-07-29 00:00:00','   urlLink The Shining in 30 seconds  is awesome like 20 hot dogs.  '),(16015,3523319,'29,July,2004','2004-07-29 00:00:00','   yeah, but in theroum they\'re making more than that. or the whole world would explode or something. so that\'s some tidy moneymaking right there.  go on, shake your moneymaker.  '),(16016,3523319,'29,July,2004','2004-07-29 00:00:00','   I don\'t read the tabloids. I look at the pictures. The only reason I picked the Hewitt was that he was top of the \'movers and losers\' list last divvy-day.  Honest.  '),(16017,3523319,'29,July,2004','2004-07-29 00:00:00','   Yeah, ed sent me some links and a squad list for the proper telegraph game.  I can\'t remember if he told me which of the leagues we were going for, but either way picking my team was more entertainig than the work I was supposed to be doing.  I found out today that it costs one million pounds per minute of each working day worldwide to run the company I work for.  That\'s an expensive operation right there.  '),(16018,3523319,'29,July,2004','2004-07-29 00:00:00','   well, it turns out that our pseudo-scottish snake-in-the-grass has been reading the tabloids the whole time. i think there should be an inquiry.  fonz, did ed mention that he and banksy are considering going for the on-line telegraph fanfoot this year? apparently you get a 18-man squad and that. it\'s supposubly better.  there was quite a good article on the bbc news web site last night about the bbfc and that. i agreed with it. although it quoted the professor of computer games at stafford university (or somesuch), which i thought devalued it a little.  '),(16019,3523319,'29,July,2004','2004-07-29 00:00:00','   Well done for having the Hewitt.  I don\'t actually have access to the tabloids, so I was wondering why the hell he was top divi this week.  I concur with your rant.  '),(16020,3523319,'29,July,2004','2004-07-29 00:00:00','   This week on the \'Daq I had James Hewitt, who did some coke or something, and the Aguilera Monster, who is a fairly consistent divvy-er. Suck my \'Daq, Fonx.  '),(16021,3523319,'29,July,2004','2004-07-29 00:00:00','   I posted a rant about Dixons withdrawing Manhunt, but then my damn interweb connection swallowed it.  Suffice it to say - Dixons bummers, mentalists bad, games not responsible, BBFC correct.  Consider yourselves ranted at.  '),(16022,3523319,'29,July,2004','2004-07-29 00:00:00','   I didn\'t know that was Stiller\'s wife.  Ross, how did you do so well in the \'Daq last week?  Which celb did you have?  Also, has anyone icked a fan foot team for this year yet?  I think I\'ve just done mine. I think Ross should join in this year.  '),(16023,3523319,'29,July,2004','2004-07-29 00:00:00','   guess which governator of california is 57 today? it\'s also ben stiller\'s wife\'s birthday. i realised yesterday that she was the bald chick in friends.  and i found  urlLink another one .  fonz: keys are in place.  '),(16024,3523319,'29,July,2004','2004-07-29 00:00:00','   damo\'s post just appeared from nowhere. it was a bit scary.  '),(16025,3523319,'29,July,2004','2004-07-29 00:00:00','   I might try and download that album when I get the chance.  I\'ve just been over to the canteen for a cocked breakfast.  Cheeky or what?  Dave, did you remember to leave you keys with the man on the front desk?  And if not, what should I do?  I\'m bringing my golf bag and clubs up with me, and I\'m gonna ninja your ass on the golf course.  '),(16026,3523319,'29,July,2004','2004-07-29 00:00:00','   why can\'t i post today? and why is ross reading the sun? or at least, looking at page 3. you enlightened renaissance man, you.  fonz and i are going to be golf-ninjas very soon. why do i keep trying to spell ninjas with a g?  maybe the ginger ones are ningas.  someone get me the new soulwax album that isn\'t out yet.  '),(16027,3523319,'29,July,2004','2004-07-29 00:00:00','   Jesus, Bambos, you have only just found that? Dude, you suck at the internet.  '),(16028,3523319,'29,July,2004','2004-07-29 00:00:00','   What was actually happening&nbsp;there Dave! - a cop in the background whispering into his walkie talkie... \" We have one, I repeat we have one...eta to cell 1230.&nbsp; Good condition...mild bruising to left leg, good runner, one careful beating.&nbsp;\" All the cops leave the streets and convene at the local nick.&nbsp; David Dickinson makes an Orange faced appearance and wanders around all the cells passing such comments as \"There\'s a little chip here which may reduce it\'s value but&nbsp;its head will certainly look good mounted on the wall...a bargain at Â£30\"...he then introduces two fresh out of training school cops.&nbsp; \"There\'s a Â£100, you have half an hour to find some bargains!\" \"How will our amateurs do? Find out at the end of the program!\" David then introduces a hardened cop...\"So how long have you been in the business?\" \"Oh about 20 years\"...\"Anything caught your eye today?\" \"There\'s one in cell 3 with a fine set of teeth...it\'s going for about Â£30 quid...but is worth well over Â£60...imagine the fun you\'d have putting your boot through those!\" Two cops return, bright eyed and bushy tailed.&nbsp; Satsuma faced David turns to them \"And what have you fellows picked up today then?\" \"Well, I\'ve bought the one in cell 6 for Â£10...a bargain...come and look\" David looks perplexed...\"Oh deary deary me...sometimes you can\'t beat experience it has to be said...that looks like a half cast to me...you\'ll be lucky to make back half of what you paid at auction!\" I can\'t be bothered to type anymore  &nbsp;  '),(16029,3523319,'29,July,2004','2004-07-29 00:00:00','   http://www.realultimatepower.net/  '),(16030,3523319,'29,July,2004','2004-07-29 00:00:00','   I\'ve discovered I have debt well over the national average. Yay me.  '),(16031,3523319,'29,July,2004','2004-07-29 00:00:00','   I\'m not worried about it either, because I\'m a selfish bastard and I have very little consumer debt.  And a savings account.  So I think she and I should get together and exchange bodily fluids.  I mean ideas.  Yes.  '),(16032,3523319,'29,July,2004','2004-07-29 00:00:00','   The girl with her tits out on Page 3 of The Sun today says she\'s not worried that consumer debt has risen over a trillion pounds, because \'we\'re a rich country\'.  Well, I\'m reassured.  '),(16033,3523319,'29,July,2004','2004-07-29 00:00:00','   You hear that?  That\'s the sound of justice.  '),(16034,3523319,'29,July,2004','2004-07-29 00:00:00','   damo, you\'ll be pleased to hear that the mugger is darrk. the policemen are not. they\'ve not kicked him yet, though.  '),(16035,3523319,'28,July,2004','2004-07-28 00:00:00','   i just saw three plain-clothes policemen take down a mugger on the opposite side of the street from me. there was lots of shouting, then they tackled him as he was running away. walkie-talkies and everything.  one of them just brought the car around. the mugger\'s sitting handcuffed on the pavement at the moment, and the bald copper keeps shouting \"shut up! shut the fuck up\" in his ear, and pushing him over.  it\'s like television. the whole neighbourhood\'s watching from its windows.  '),(16036,3523319,'28,July,2004','2004-07-28 00:00:00','   80% of the work is done in 20% of the time. The other 20% of the work requires the remaining 80% of the time.  What that really means is that the easy stuff is easy to do, and the complicated stuff is usually not necessary. It\'s why when you\'re enjoying your work, people can see results, and when you\'re hating your work they keep asking you when they\'re going to see more results.  It\'s very handy because you can talk your way out of doing anything boring and time consuming.  '),(16037,3523319,'28,July,2004','2004-07-28 00:00:00','   What\'s the 80/20 rule?  It sounds an awful lot like something I can use to get out of doing any work for a bit, and information like that should be shared.  '),(16038,3523319,'28,July,2004','2004-07-28 00:00:00','   i\'ll draw them a graph with a line on it represting my income, a line representing the pressure on my income, an umbrella, and a fat cash bath.  i reckon i could, too.  i\'ve just been emailed a request to explain everything that\'s wrong with everything by the end of tomorrow. but i want to do my buttons.  my buttons.  '),(16039,3523319,'28,July,2004','2004-07-28 00:00:00','   Debbie Does Dallas Special Edition is just like the first one except all the cocks are replaced by walkie-talkies. And Han shoots first, if you know what I mean. Hur-hur.  Those buttons do need re-doing. The travel system will take a while to re-do, make sure they know that because otherwise they\'ll say to you, look, it was 80% of the way there, why is it taking so long? You should explain the 80-20 rule with some pictures and maybe a puppet.  '),(16040,3523319,'28,July,2004','2004-07-28 00:00:00','   my guts have got some serious ache on. i think i can hear all the porcelain within a mile radius quailing.  i\'m sure i have debbie does dallas somewhere. at least, debbie does dallas: the next generation. which is like the old one, but one of the characters is wearing tyre over his eyes.  yove: i\'ve given up on my real work (re-creating the travel tables in the database) and i\'ve started making some new buttons for the menu screens. i am king.  '),(16041,3523319,'28,July,2004','2004-07-28 00:00:00','   Yes, and this new one never says anything about queer-ies. It has no sense of humour.  '),(16042,3523319,'28,July,2004','2004-07-28 00:00:00','   It can certainly be a bit temperamental. Unlike the old BOA, which was perfect.  '),(16043,3523319,'28,July,2004','2004-07-28 00:00:00','   I don\'t think I understand this new boa. I hit publish post, and it tells me that the blog doesn\'t exist anymore, then it posts my message twice.  I\'m looking for another button to click but it\'s the only one i can see.  '),(16044,3523319,'28,July,2004','2004-07-28 00:00:00','   I want to see Debbie Does Dallas, in 30 seconds, acted by bunnies.  No hold on, I just want to see Debbie Does Dallas.  '),(16045,3523319,'28,July,2004','2004-07-28 00:00:00','   I think  urlLink that  is good.  I feel sleepy and muzzy headed.  I have a sneaking suspicion it\'s from sitting in a beer garden drinking lager.  '),(16046,3523319,'28,July,2004','2004-07-28 00:00:00','   i\'m not sure whether  urlLink this  is good or not.  i feel bloaty. i bet it was that third sandwich.  '),(16047,3523319,'28,July,2004','2004-07-28 00:00:00','   Dude, she\'s so fine I\'d happily pay you a small fortune to come round and have a turn, broken legs or no.  I think this makes us both wrong in the head.  '),(16048,3523319,'28,July,2004','2004-07-28 00:00:00','   Sorry Fonz, you went on so much about her...well - She\'s tied up in my wardrobe at the moment with a rag of cloth in her mouth.&nbsp; I hope one day to be able to untie her and let her roam free about my room...if the worst comes to the worst I may well break her legs but it\'s ok&nbsp;- she loves me so she\'ll understand  '),(16049,3523319,'28,July,2004','2004-07-28 00:00:00','   Dude, that sucks.  It is fricking roasting here Mr Bigglesworth, so I\'ve been drinking in the beer garden.  The good news?  I am drunk.  Tha bad news? I don\'t think the fit bar maid works at the pub anymore.  She was too good for us anyway...  '),(16050,3523319,'28,July,2004','2004-07-28 00:00:00','   it does feel an offal lot like friday today. why is that?  i got my scare cut. there will be no ambiguity about my sexuality now.  '),(16051,3523319,'28,July,2004','2004-07-28 00:00:00','   I was sure it was Friday today. I was quite crushed when I found out I have to come to work tomorrow.  '),(16052,3523319,'28,July,2004','2004-07-28 00:00:00','   I did that once.  I couldn\'t sit down for a week afterwards.    The barber\'s chair had a big spring in it that upset my piles, you see.  '),(16053,3523319,'28,July,2004','2004-07-28 00:00:00','   would i want to rent an ally mcbeal dvd? is it the way i walk?  i\'ll get a haircut, and tell them that i want a \"hard man\". that should put an end to this.  '),(16054,3523319,'28,July,2004','2004-07-28 00:00:00','   and that\'s why damo\'s not on c  oh.  '),(16055,3523319,'28,July,2004','2004-07-28 00:00:00','   You should phone up the&nbsp; people in Shanghai and say \"Allo Mr Sars, it\'s me Mr Dave...wha\' yu wan\'?\"  '),(16056,3523319,'28,July,2004','2004-07-28 00:00:00','   I\'m here, and I have nothing to say.  '),(16057,3523319,'28,July,2004','2004-07-28 00:00:00','   That\'ll learn you.  '),(16058,3523319,'28,July,2004','2004-07-28 00:00:00','   the shanghai office had a problem with their email this morning. the great thing about shanghai is that it\'s about 7 or 8 hours ahead of us.  i\'m on call this week.  i want to die.  '),(16059,3523319,'28,July,2004','2004-07-28 00:00:00','   I don\'t know about that...Fonz\'s response had he been the interviewer would have been \"You know someone that would pay for Fonz love?\"  '),(16060,3523319,'28,July,2004','2004-07-28 00:00:00','   That was weird. At work, I couldn\'t see Yoav\'s message, hence my \'talking to himself\' post. But at home it suddenly appeared in the past. I suspect witchcraft.  I think Fonz is equipped to be a prostitute.  '),(16061,3523319,'28,July,2004','2004-07-28 00:00:00','   I\'ve done it! I\'ve taken monday off work to shop my way around London.  Now I need to go home and strip my golfing equipment down to the bare minimum and see if I can carry my shit. Oh, and practise my swing, because I haven\'t played since... well, since I played with Dave last time.  '),(16062,3523319,'27,July,2004','2004-07-27 00:00:00','   ross will help you. he has the soul of a saint and the powers of a super-hero. i know that he won\'t let us down.  right?  '),(16063,3523319,'27,July,2004','2004-07-27 00:00:00','   Who wants to help me carry my golf clubs to london?  No one?  Alright then.  '),(16064,3523319,'27,July,2004','2004-07-27 00:00:00','   he\'s very deaf sometimes. it must be his godzilla ears.  ooh, tea-time.  '),(16065,3523319,'27,July,2004','2004-07-27 00:00:00','   pay attention won\'t you?  '),(16066,3523319,'27,July,2004','2004-07-27 00:00:00','   come to london this weeked, or i\'ll kick your grass.  '),(16067,3523319,'27,July,2004','2004-07-27 00:00:00','   yes, but in lahndahn we have all things in the proper proportion (try saying that on a cold day), whereas everything in tokyo is tiny, tootiny.  '),(16068,3523319,'27,July,2004','2004-07-27 00:00:00','   I think Wobav is talking to himself.  '),(16069,3523319,'27,July,2004','2004-07-27 00:00:00','   i wanted to email you a supposed picture of your winky that i found on the innernet for your confirmation. i\'m going to remove yoab\'s extra post - just watch me!  '),(16070,3523319,'27,July,2004','2004-07-27 00:00:00','   Apparently, Lahndahn is the second most expensive city in the world, behind Tokyo. And Tokyo-ites have a much better quality of life. This must mean that anyone who lives in Lahndahn must be the very end of the bell. The donger, if you will.  '),(16071,3523319,'27,July,2004','2004-07-27 00:00:00','   Yes, they say everything twice here to hammer the point home.  Or maybe I\'m not used to this new-fangled boa.  '),(16072,3523319,'27,July,2004','2004-07-27 00:00:00','   I could have. If I remember, I\'ll set up one of our Blueyonder mailboxes with an alias. I make no promises about checking it, though. Why?  '),(16073,3523319,'27,July,2004','2004-07-27 00:00:00','   What I am learning in America:  It is a portion of National Public Radio (NPR) interview between a female broadcaster and US Marine Corps General Reinwald who was about to sponsor a Boy Scout Troop visiting his military installation  FEMALE INTERVIEWER: So, General Reinwald, what things are you going to teach these young boys when they visit your base?  GENERAL REINWALD: We\'re going to teach them climbing, canoeing, archery, and shooting.  FEMALE INTERVIEWER: Shooting! That\'s a bit irresponsible, isn\'t it?  GENERAL REINWALD: I don\'t see why, they\'ll be properly supervised on the rifle range.  FEMALE INTERVIEWER: Don\'t you admit that this is a terribly dangerous activity to be teaching children?  GENERAL REINWALD: I don\'t see how. We will be teaching them proper rifle discipline before they even touch a firearm.  FEMALE INTERVIEWER: But you\'re equipping them to become violent killers.  GENERAL REINWALD: Well, you\'re equipped to be a prostitute, but you\'re not one, are you?  The radio went silent and the interview ended.  '),(16074,3523319,'27,July,2004','2004-07-27 00:00:00','   *the* hole? where is it? where are you hiding it?  '),(16075,3523319,'27,July,2004','2004-07-27 00:00:00','   Yes, it\'s a guilty pleasure, like whacking off while looking through the hole in the wall of the ladies changing room showers.  '),(16076,3523319,'27,July,2004','2004-07-27 00:00:00','   *you* make me sick, ross. come to london and help me wipe the bile from my jaw. do you have an email address that isn\'t your hotmail one? that you read?  watching friends makes me feel sad. in both ways. why is that?  '),(16077,3523319,'27,July,2004','2004-07-27 00:00:00','   The Metro suggested that it could be Smithers and Mr Burns getting hitched. I suggest that the writers of the Metro are idiots.  The SpaceShipOne team have set the dates for their attempts at the Ansari x-Prize. It\'s so exciting, I may be sick. Well, maybe not. I want to go into space, damnit!  '),(16078,3523319,'27,July,2004','2004-07-27 00:00:00','   presumably, bets on smithers are pretty short.  '),(16079,3523319,'27,July,2004','2004-07-27 00:00:00','   I agree with Fonz...he looks like he had an injection of gay at some point - not fonz...the dad ok!  '),(16080,3523319,'27,July,2004','2004-07-27 00:00:00','   I heard that Simpson\'s rumour.  I reckon Millhouse\'s dad could be the gayer.  '),(16081,3523319,'27,July,2004','2004-07-27 00:00:00','   what\'s been wrong with the boa this morning? can i post yet?  '),(16082,3523319,'27,July,2004','2004-07-27 00:00:00','   Apparently there\'s going to be a gay wedding in the Simpsons. How frightfully modern.  '),(16083,3523319,'27,July,2004','2004-07-27 00:00:00','   Well at least you\'ve proven your potential...  '),(16084,3523319,'26,July,2004','2004-07-26 00:00:00','   The fools! NatWest have agreed to lend me ten grand, and all I needed to do was spin them some bullshit story about training to be a lawyer.  Bwahaha!  Oh.  '),(16085,3523319,'26,July,2004','2004-07-26 00:00:00','   A long story for ananova really, but all you actually need to know is in the last couple of sentences of the third last paragraph.&nbsp; It runs in the family then?  urlLink http://www.ananova.com/entertainment/story/sm_1035845.html?menu =  '),(16086,3523319,'26,July,2004','2004-07-26 00:00:00','   Dave, will I have any time to do any shopping in london this weekend?  I\'m considering taking Monday off work and doing some shopping then instead. What should I do?  '),(16087,3523319,'26,July,2004','2004-07-26 00:00:00','   oh, it\'s gone. sorry.  '),(16088,3523319,'06,August,2004','2004-08-06 00:00:00','   why would you wear red on a friday?! surely you knew the consequences of your savage actions?  how do you see at the moment, ross? are your wee eyeses all red and sore? like bloody pissholes in the snow?  '),(16089,3523319,'05,August,2004','2004-08-05 00:00:00','   You must have loved it Ross. You\'re constantly complaining of lack of mansweat in your daily life.  '),(16090,3523319,'05,August,2004','2004-08-05 00:00:00','   In an interesting development, I have become allergic to my contact lenses, apparently.  Whoops.  '),(16091,3523319,'05,August,2004','2004-08-05 00:00:00','   I\'ve been to the New York Stock Exchange. We were there for closing. A bell rang, and everyone ran around like morons, then it was all over. It smelled of mansweat.  '),(16092,3523319,'05,August,2004','2004-08-05 00:00:00','   I\'m here for another... 7 hours. Goddammit.  I went to the New York Stock Exchange yesterday with another guy from the office. We were shown around by a broker. It\'s a pretty amazing place. It\'s just this huge room (with american flags hanging on the walls) and 1366 people in it. Actually it\'s one huge room and a few smaller rooms next to it.   There are no chairs, and basically at the sides of the room stand people who wait by the phone to get buy and sell orders. They get a call to buy shares of whatever, and they relay it to the guy who then goes to do it. He walks to something called a \"post\" which is a free standing circle of computer monitors reaching a few meters high, with about 5 people inside the circle and 5 people outside. Each person in a post trades between one and ten or so companies, so there are lots of posts dotted aruond the room. So he says to the person outside the post, I want to buy these shares, and he points to someone else nearby and says he wants to sell them, let\'s negotiate a price and we\'re done.  Of course this is all very hectic and completely a waste of time. The NASDAQ does it all electronically and these guys know that they\'re doomed. Every one of them that I talked to gave me a speech on why their particular function couldn\'t properly be repliacted by a computer and why they can still \"add value to their clients\".   So it was interesting. People in the office are blaming me for the market being down today because I\'m wearing a red shirt.  '),(16093,3523319,'05,August,2004','2004-08-05 00:00:00','   i\'m going to give you a paint enema, wee man. forcibly.  '),(16094,3523319,'05,August,2004','2004-08-05 00:00:00','   I got to work at 2:15 today. And I only have 45 minutes left until home time.  It\'s like being a student.  '),(16095,3523319,'05,August,2004','2004-08-05 00:00:00','   still bored.  '),(16096,3523319,'05,August,2004','2004-08-05 00:00:00','   bored bored bored bored bored.  '),(16097,3523319,'05,August,2004','2004-08-05 00:00:00','   I\'m going home.  '),(16098,3523319,'05,August,2004','2004-08-05 00:00:00','   Is this page working yet?  I just ordered a gun for my PS2.  I have the urge to shoot the shit out of some mofos.  '),(16099,3523319,'05,August,2004','2004-08-05 00:00:00','   did anyone see a monkey come past here yesterday? small, hairy fella? likes climbing?  no?  '),(16100,3523319,'05,August,2004','2004-08-05 00:00:00','   I\'ll also be in Brum this weekend, bitch slapping pooky and drinking beers with banksy.  '),(16101,3523319,'05,August,2004','2004-08-05 00:00:00','   you know it, baby. unfortunately, i\'ll be in birmingham this weekend, attempting some key-hole surgery on ross\' ear with a paint-ball gun. and i\'m in edinburgh the weekend after. who are we playing the weekend after that?  '),(16102,3523319,'04,August,2004','2004-08-04 00:00:00','   Dave, aren\'t we\'re going to Wycombe this Saturday?     He\'ll be there, will you?  Is Herve Renard the most pornographic manager in the English League?  '),(16103,3523319,'04,August,2004','2004-08-04 00:00:00','   I\'m back on the BoA, cheers fonz. Next time i post i\'ll think of something interesting to say first. I donâ€™t really know what Iâ€™m doing with this new blog stuff.   I\'m going to the Lake district for a race weekend tomorrow. I\'ve got to run and cycle up mountains and canoe in lakes. It all has an endurance theme. It is going to be hell and i volenteered. What the fuck is wrong with me?  '),(16104,3523319,'04,August,2004','2004-08-04 00:00:00','   You want a piece of me?  Come get some.  '),(16105,3523319,'04,August,2004','2004-08-04 00:00:00','   one of these days, ross, i\'m going to catch up with you and hurt you real bad.  how\'s saturday for you?  '),(16106,3523319,'04,August,2004','2004-08-04 00:00:00','   I just spent an hour watching a gamespot panel discussion on how computer game graphics have progressed over the years.  I am Geek + 4 now.  '),(16107,3523319,'04,August,2004','2004-08-04 00:00:00','   I\'m going home  '),(16108,3523319,'04,August,2004','2004-08-04 00:00:00','   you can only do so much with your hands tied behind your back, can\'t you.  metaphorically or otherwise, eh damo?  '),(16109,3523319,'04,August,2004','2004-08-04 00:00:00','   You big Welsh bastard.  I also failed to nob the Knightley last night, but I didn\'t feel like I was putting much effort in.  '),(16110,3523319,'04,August,2004','2004-08-04 00:00:00','   i also failed to nob the knightley last night. i was too busy eating ratatouille.  ra-ta-tou-ille!  '),(16111,3523319,'04,August,2004','2004-08-04 00:00:00','   I\'d say you were funny and talented Dave.  And I am glad the damo \"failed to nob\" the Knightley last night.  I\'ve done been to the pub, and it has taken me two minutes to type this sentence.  Bah.  '),(16112,3523319,'04,August,2004','2004-08-04 00:00:00','   i just ate some ratatouille. cheers, lexie.  '),(16113,3523319,'04,August,2004','2004-08-04 00:00:00','   i\'d say that i was \"funny and talented\".  well, \"a bit funny\" certainly.  '),(16114,3523319,'04,August,2004','2004-08-04 00:00:00','   urlLink news   Dave, at least you aren\'t the only one failing to nob the Knightley - I did last night...she\'s ok I suppose.  '),(16115,3523319,'04,August,2004','2004-08-04 00:00:00','   She\'s my precious! To be honest I can\'t remember which of the cells I\'ve chained her in - When I find her I\'ll send her your way but it\'s highly likely she\'ll be pretty fucked up...it took rather a long time to convince her she loved me.  '),(16116,3523319,'04,August,2004','2004-08-04 00:00:00','   Well we all knew that already. Are you a lawyer yet?  I\'m going to the pub in a minute.  I still live in hope of the fit bar maid returning.  Come on damo, give her back.  '),(16117,3523319,'04,August,2004','2004-08-04 00:00:00','   Dorothy from the Wizzard of Oz!  '),(16118,3523319,'04,August,2004','2004-08-04 00:00:00','   Hello Lex,  I was wondering if you were still alive.  Have you found out who I look like yet?  '),(16119,3523319,'04,August,2004','2004-08-04 00:00:00','   Hello?  Is anybody there?  '),(16120,3523319,'04,August,2004','2004-08-04 00:00:00','   yes, she wandered into my office just after seven and started complaining to people that she had a metal plate in her head. they knew to send for me.  i had some interesting questions to answer this morning, i can tell you.  '),(16121,3523319,'04,August,2004','2004-08-04 00:00:00','   Having spent 5 hours in the pub last night I came home to attempt something I have seen done on TV a few times.  After 1 hour, 5 flies and plenty of frustration I succeeded in attaching a small paper aeroplane via thin line to the back of a blue bottle fly.  It was a defining moment in my life and one I shall remember for years to come. My next effort will involve 5 or 6 flies with which I hope to stage some sort of Battle of Britain re-inactment.  It\'s amazing what interests you whilst drunk! Did you manage to find lexie in the end last night big \'un?  '),(16122,3523319,'04,August,2004','2004-08-04 00:00:00','   i was as professional as i\'ve ever been in those days. the illegal spider racing was one thing, but i\'ll never, ever forget my short film about the hedgehog.  i was the master of paint shop pro 4 and gif animator, so i was. i almost miss those days.  '),(16123,3523319,'04,August,2004','2004-08-04 00:00:00','   Is that the word \'professional\' used as an adjective to \'web developer\'?  After working at CIS I didn\'t think there was such a thing!  Dave will explain, illegal spider racing and all.  '),(16124,3523319,'04,August,2004','2004-08-04 00:00:00','   Oooh, I\'ve been lent Ikiru on DVD.  I knew this day would come.  '),(16125,3523319,'04,August,2004','2004-08-04 00:00:00','   We professional web developers don\'t need lessons in that.  Teach me something useful, Nemesis.  '),(16126,3523319,'04,August,2004','2004-08-04 00:00:00','   It wasn\'t porn, it was a lesson in image sourcing and resizing in HTML.  '),(16127,3523319,'04,August,2004','2004-08-04 00:00:00','   Oh. Now it\'s gone away. I\'m not lying to impress people, honest.  '),(16128,3523319,'04,August,2004','2004-08-04 00:00:00','   Ok, I can understand the ad for Takeshi Kitano DVDs... but why are we being offered the chance to save some stupid wildlife?  '),(16129,3523319,'04,August,2004','2004-08-04 00:00:00','   Dude, I\'ll get sacked if you post stuff like that on here.  And I probably won\'t be the only one.  '),(16130,3523319,'03,August,2004','2004-08-03 00:00:00','   Here there be porn    I see those image resize tags work though. This one enlarged from 90 x 70 px.  '),(16131,3523319,'03,August,2004','2004-08-03 00:00:00','   How dare you drastically resize my image Dave.  How would you like it if I resized something of yours?  And there\'s is no point now telling me I got an invite when I quite clearly didn\'t, it\'s called lying and it\'s not cool.  Anyway for those of you wondering what other amusing place names I happened to find on my travels you can look  urlLink here.  The photos of Bummer\'s Hill are still unfortunately in my camera.  Dave did promise me some from the gay orgy he attended there last month, but has not been forthcoming with the goods.  I recommend the  urlLink original web site  also.  '),(16132,3523319,'03,August,2004','2004-08-03 00:00:00','   It\'s sunny....it\'s midweek...I just realised I\'ll be Â£800 better off than I thought this month...it can mean only one thing...drinking beers beers beers!  '),(16133,3523319,'03,August,2004','2004-08-03 00:00:00','   My onion horroscope is pretty accurate this week.  What am I still doing at work?  '),(16134,3523319,'03,August,2004','2004-08-03 00:00:00','   no you are.  '),(16135,3523319,'03,August,2004','2004-08-03 00:00:00','   Yes, I thought that about the translators earlier, until I realised that to translate Japanese to English, you\'d need to enter the Japanese in kanji, or at least hiragana.  And I wouldn\'t say that fucking the Daily Mail is out of character at all. Although fucking the Daily Male would be more in character.  Ha de ha. Ha.  '),(16136,3523319,'03,August,2004','2004-08-03 00:00:00','   wow, these on-line translators are bollocks.  i saw that news item earlier today, and i found it grimly satisfying. some people may find this a little out of character for me, but: fuck the daily mail  '),(16137,3523319,'03,August,2004','2004-08-03 00:00:00','   watashi-wa igiris-jin desu. dabido desu. nani da yo.  '),(16138,3523319,'03,August,2004','2004-08-03 00:00:00','   http://www.gamesindustry.biz/content_page.php?section_name=pub&aid=3918  '),(16139,3523319,'03,August,2004','2004-08-03 00:00:00','   Sayonara, loi.  '),(16140,3523319,'03,August,2004','2004-08-03 00:00:00','   Yes.  '),(16141,3523319,'03,August,2004','2004-08-03 00:00:00','   Another good one phonetically is \"manko\"...some japs won\'t understand \"what the hell\"...\"manko\" is universally understood - japanersally more to the point.  '),(16142,3523319,'03,August,2004','2004-08-03 00:00:00','   I found out why I couldn\'t get Zatoichi from Blockbuster - it wasn\'t out on DVD until last week. Stupido Blockbuster.  '),(16143,3523319,'03,August,2004','2004-08-03 00:00:00','   that sounds spot on. an elderly villager says it in zatoichi. i pissed myself at that point.  '),(16144,3523319,'03,August,2004','2004-08-03 00:00:00','   I\'m having a little trouble with this, as I can\'t view Japanese properly on my machine at work. I\'ll have a shufti when I get home.  However, I think the first one should be \'nan(i) da yo\'. Nan, or sometimes nani, is \'what\', \'da\' is, informally, the existance of a thing, which isn\'t a verb in Japanese, and \'yo\' adds emphasis at the end of a sentence. So it would be something like \'what (the hell) is that?\', I think.  '),(16145,3523319,'03,August,2004','2004-08-03 00:00:00','   1st one depends on where the emphasis in your accent is...gentle it means \"what the hell\"...if you emphasise (phonetically) the \"ooooo\" bit at the end it become more like \"what the fuck\"  The second one is to commit severe \"cunning linguist\" on the mother \"bitch\" of your target.  By severe I mean getting the head down there and causing nastiness to come out.   Kani-Nabu-Chao-Chi-Bei is a phonetic equivalent of that in Chinese.  All though this one is for doing bad things to the dead relatives of your target...(cue: some tiresome jokes from fonz!)  '),(16146,3523319,'03,August,2004','2004-08-03 00:00:00','   well, the first thing is \"what the hell?\" or similar, and the second, i\'m told, is the most offensive possible phrase in japanese.  start with looking up the kanji for \"your mother\'s dead ancestors\" and work backwards.  '),(16147,3523319,'03,August,2004','2004-08-03 00:00:00','   What does it mean, though? If I knew the English, I could look stuff up.  Come on, you\'ve piqued my interest now.  Also, I wish my fucking interweb connection would stop signing me out of Blogger when I try to post.  '),(16148,3523319,'03,August,2004','2004-08-03 00:00:00','   s\'welsh for ian. it is. anyway, i tole you on the tube that it was either ian or yoan, and you said you reckoned it was ewan. so i suppose we\'ll just have to kill each other.  ross, i\'ve never seen my noise written down, i\'ve only ever heard it, so i\'m afraid that\'s the closest transliteration that i can manage.  go on, find the kanji now.  '),(16149,3523319,'03,August,2004','2004-08-03 00:00:00','   Yes, but I was closer then your \"Ian\", and you had the cheek to pass yours off as fact.  More importantly, I knew you were wrong, which makes me keewwweelllll.  '),(16150,3523319,'03,August,2004','2004-08-03 00:00:00','   wHAT DOES YOUR NOISE MEAN, THEN? i MIGHT WASTE SOME TIME LOOKING UP THE KANJI.  tHIS IS HOW EVERYONE WILL TYPE IN THE FUTURE.  '),(16151,3523319,'03,August,2004','2004-08-03 00:00:00','   you said he was called \"ewan\". that\'s not \"yo-an\".  ross, if you were real, i\'d kick your ass for being so stupid.  '),(16152,3523319,'03,August,2004','2004-08-03 00:00:00','   Hmm, I think I just accidentally wrote a rant about King Arthur.  Still, it wasn\'t very good, was it?  '),(16153,3523319,'03,August,2004','2004-08-03 00:00:00','   I knew it!   Welsh Actor Ioan Gruffudd (Pronounced YO-an GRIFF-ith)...  '),(16154,3523319,'03,August,2004','2004-08-03 00:00:00','   Probably - he is a thieving barstard !  '),(16155,3523319,'03,August,2004','2004-08-03 00:00:00','   Was it Dave?  '),(16156,3523319,'03,August,2004','2004-08-03 00:00:00','   Just got a phone call from Lloyds TSB.  Someone tried to use my Credit Card to buy a Â£750 camera in Canada last week !  Luckily they caught it and didnt let it through. I lost Â£40 though as the exchange rate dropped a lot before it was refunded. I think the insurance covers it though.  Pain in the arse !  '),(16157,3523319,'03,August,2004','2004-08-03 00:00:00','   That\'s just noise, Dave.  '),(16158,3523319,'03,August,2004','2004-08-03 00:00:00','   I reckon a Monkey-Wolf-Fish-Boy has signed up to this blog and not posted anything.  I reckon he\'s furiously blogging himself in the corner.  '),(16159,3523319,'03,August,2004','2004-08-03 00:00:00','   manu ganegaushi  '),(16160,3523319,'03,August,2004','2004-08-03 00:00:00','   Ikiru is the dictionary form, or present informal positive tense, of the verb to live, or exist. It looks like a group 2 (godan) verb, but is in fact a group 1 (ichidan), and so the present formal positive is ikirimasu, and not ikimasu as you\'d expect.  Look upon my works, ye mighty, and despair!  Ozy-who-dias?  '),(16161,3523319,'03,August,2004','2004-08-03 00:00:00','   Yeah, the bond fan has loads of kurosawa, including both those, but has thus far only lent me The Seven Samurai.  I might steal Ikiru pretty shortly, though.  '),(16162,3523319,'03,August,2004','2004-08-03 00:00:00','   james has it - he brought it round. he\'s left me with ikiru, which has no samurai in it, yet is still a kurosawa film. bizarre, non?  '),(16163,3523319,'03,August,2004','2004-08-03 00:00:00','   Have you got Yojimbo on DVD? I want to see it.  '),(16164,3523319,'03,August,2004','2004-08-03 00:00:00','   he looked a bit old and buggered in zatoichi, but i put that down to him being old and buggered.  zatoichi is a good film. i watched yojimbo last night, and that is also a good film. but it made me want to watch sergio leone again.  '),(16165,3523319,'03,August,2004','2004-08-03 00:00:00','   Today I have learned that the Takeshi in Takeshi\'s Castle is actually Takeshi Kitano. I also learned that he had a motorbike accident in 1994 and broke his face. Yes.  '),(16166,3523319,'03,August,2004','2004-08-03 00:00:00','   i know. and i understand.  '),(16167,3523319,'03,August,2004','2004-08-03 00:00:00','   I might shoot you in the nuts this weekend, Dave. You know, just because.  '),(16168,3523319,'03,August,2004','2004-08-03 00:00:00','   i need a new word.  '),(16169,3523319,'03,August,2004','2004-08-03 00:00:00','   i resized it, although not very well. it\'s just a width and height property in the img tag.  today, i am rubbish. i need another weekend. although ideally not one spent being chased by bambos around some woods.  although that\'ll have to do.  '),(16170,3523319,'03,August,2004','2004-08-03 00:00:00','   If you right click on the picture and select \'properties\', you can see the location.  '),(16171,3523319,'03,August,2004','2004-08-03 00:00:00','   No, I haven\'t. I still need a computer case to actually build the machine before I can set it up. When I can be arsed, I\'ll find a cheapo case and put it together, then you can all give me useless, conflicting advice again.  '),(16172,3523319,'03,August,2004','2004-08-03 00:00:00','   Who cleverly re-sized that image then, eh? Where is it hosted?  I haven\'t managed to make struts work.  But .NET does much the same thing, and that works a treat.  Ross, did you ever set up a server and your own website?  Or was our advice as useful as a chocolate thing that\'s not supposed to melt, but does because it\'s made of chocolate?  '),(16173,3523319,'03,August,2004','2004-08-03 00:00:00','   tom, i sent you an invite ages ago to the board, but they expire after a few days i think, so if you missed it at the time then that was it. but you\'re here now, and that\'s what matters etc.  so tired, so sleepy. struts are QAF.  '),(16174,3523319,'03,August,2004','2004-08-03 00:00:00','   That is him.  I can confirm it.  '),(16175,3523319,'03,August,2004','2004-08-03 00:00:00','     '),(16176,3523319,'02,August,2004','2004-08-02 00:00:00','   OK, point taken.  I will try and use your/you\'re in the correct context, except when abusing Dave because it\'s even more offending when it sounds like it\'s from a fat, dumb, terrorised American.   Can we put HTML in here?  '),(16177,3523319,'02,August,2004','2004-08-02 00:00:00','   from now on, that\'s how you program.  '),(16178,3523319,'02,August,2004','2004-08-02 00:00:00','   I am so tired of this Robot bull-shit.  I\'m ready to go home now, Mr Hat.  '),(16179,3523319,'02,August,2004','2004-08-02 00:00:00','   That made me laugh Ross...I\'m going to admit myself tonight I think  '),(16180,3523319,'02,August,2004','2004-08-02 00:00:00','   Also, \'bollocks\' means \'postage stamp\'. Go on, Yoav, ask someone to lick your bollocks.  '),(16181,3523319,'02,August,2004','2004-08-02 00:00:00','   there\'s an episode of friends where rachel shouts \"he touched my fanny!\".  it made me laugh in a bad way.  '),(16182,3523319,'02,August,2004','2004-08-02 00:00:00','   In America, what they call a fanny is your arse, not your minge.  (Ho ho)  '),(16183,3523319,'02,August,2004','2004-08-02 00:00:00','   i knew that he wouldn\'t have the balls.  although, strangely, he does have the nutsack.  '),(16184,3523319,'02,August,2004','2004-08-02 00:00:00','   Why? I\'ll just do what I do every year - concentrate on my team for three weeks before forgetting about it, only to be reminded that I was playing when the final results come in and I\'m somewhere near the bottom.  I\'m pretty sure I\'ve never even made a transfer in fantasy football, because I never pay attention long enough to need to.  '),(16185,3523319,'02,August,2004','2004-08-02 00:00:00','   Quick, let\'s make Ross go to  urlLink www.tffo.co.uk  and sign up for a fantasy football team.  '),(16186,3523319,'02,August,2004','2004-08-02 00:00:00','   i have a mighty team of mighties. i shall win, undoubtedly.  '),(16187,3523319,'02,August,2004','2004-08-02 00:00:00','   This fantasy football lark is hard. And I\'m probably supposed to be doing some work.  '),(16188,3523319,'12,August,2004','2004-08-12 00:00:00','   6. blade. razor  '),(16189,3523319,'11,August,2004','2004-08-11 00:00:00','   There was quite a good article on the onion by a guy from Gillette that definitely pre-empted that one manx.  I think it was entitled \"Fuck it, we\'re going to five blades\"  '),(16190,3523319,'11,August,2004','2004-08-11 00:00:00','   i\'ve just had a revolutionery idea that will definately, definately make me a millionaire.  5. blade. razor  '),(16191,3523319,'11,August,2004','2004-08-11 00:00:00','   I thought of sun cream in a spray bottle first. I did, I tells ya.....  '),(16192,3523319,'11,August,2004','2004-08-11 00:00:00','   I can back you up on that manx.  You also thought of the washing machine with three drums, which is still waiting to pop I think.  '),(16193,3523319,'11,August,2004','2004-08-11 00:00:00','   I still maintain that four-blade razors were my idea  '),(16194,3523319,'11,August,2004','2004-08-11 00:00:00','   It\'s about time, Bambos.  Although Dave did mention that article yesterday, you big, hairy, razor-ruining bastard.  Incidently, I have a four-blade razor now. It\'s like living in the fucking future.  '),(16195,3523319,'11,August,2004','2004-08-11 00:00:00','   http://www.theonion.com/news/index.php?issue=4032&n=1  '),(16196,3523319,'11,August,2004','2004-08-11 00:00:00','   Opera is a good venue to experience \"culture\" certainly...I still don\'t think you can beat the ramblings of Carwyn whilst he is in the messy depths of an alcohol induced dementia though.  '),(16197,3523319,'11,August,2004','2004-08-11 00:00:00','   That\'s mental. Some people have too much time on their hands.  '),(16198,3523319,'11,August,2004','2004-08-11 00:00:00','   urlLink http://www.thebricktestament.com/  The Bible, acted out in Lego. How veddy odd. I \'recommend\' The Old Testament/Law......it\'s very Eddie.....  '),(16199,3523319,'11,August,2004','2004-08-11 00:00:00','   i\'m about to go to the opera. i\'m not kidding.  '),(16200,3523319,'11,August,2004','2004-08-11 00:00:00','   Oooh, so can I.  Nice.   I\'m here, I just can\'t be arsed to do anything.  I don\'t talk to people before lunch anymore. It\'s too early.  '),(16201,3523319,'11,August,2004','2004-08-11 00:00:00','   Ooh, I can quick-edit posts now.   Where is everyone today?  '),(16202,3523319,'11,August,2004','2004-08-11 00:00:00','   By \'the guy being sick\', you mean the guy bending over to scratch his ankle, right? Unless it\'s invisible sick.  '),(16203,3523319,'11,August,2004','2004-08-11 00:00:00','   i might start doing lynndies at the opera. or possibly at some drunken scotsmen.  i\'m not sure that i can handle the culture.  '),(16204,3523319,'11,August,2004','2004-08-11 00:00:00','   Is this  urlLink Richard Kipping  doing a Lynndie?   urlLink     I think I found one of  urlLink Taufik  as well!   urlLink     Note the guy being sick and is it me or is the bird sitting there quite hot?  '),(16205,3523319,'11,August,2004','2004-08-11 00:00:00','   it\'s true. my java is shocking.  '),(16206,3523319,'10,August,2004','2004-08-10 00:00:00','   Yer well, when he finds out about YOUR crack habit too Dave YOU\'RE really going to be in the shit, aren\'t you?  I think the grammar is correct in that one?  I\'m really shit, some have compared the poorness of my English with that of Dave\'s Java!  '),(16207,3523319,'10,August,2004','2004-08-10 00:00:00','   http://badgas.co.uk/lynndie/  '),(16208,3523319,'10,August,2004','2004-08-10 00:00:00','   uh-oh - dubya\'s bound to have someone reading this board. and now the C-A-T is out of the B-A-G. we\'ll all hang.  or fry.  '),(16209,3523319,'10,August,2004','2004-08-10 00:00:00','   The A-Team one is a classic.  Does anyone fancy terrorising the Americans?  We could set off a bomb in one of their cities or something, you know, just for kicks.  We could then look on while they invade some African nation with oil as a result.  '),(16210,3523319,'10,August,2004','2004-08-10 00:00:00','   Add me in bizatch! I want to check out the competition  '),(16211,3523319,'10,August,2004','2004-08-10 00:00:00','   You have a crack habit...I\'m going to enroll you at the priory...although you\'re paying.  '),(16212,3523319,'10,August,2004','2004-08-10 00:00:00','   you have a disorder, damo. i\'m going to edit you.  '),(16213,3523319,'10,August,2004','2004-08-10 00:00:00','   Ok have picked my \"Super\" team...address is damo_beastAThotmial.com They are called Dead Men Walking like the Euro Team  '),(16214,3523319,'10,August,2004','2004-08-10 00:00:00','   i\'ve been wondering about  urlLink this  myself for some time now.  and the a-team one is brilliant.  '),(16215,3523319,'10,August,2004','2004-08-10 00:00:00','   i think that\'s it - it\'s the LZW compression nonsense. might as well use PNGs instead, though - they\'re better, and free.  '),(16216,3523319,'10,August,2004','2004-08-10 00:00:00','   OK I\'ll think about it.  What\'s the deal with GIFs these days?  Your allowed to put them on web sites aren\'t you?  You just have to pay if you want to use the algorithm in your software.  '),(16217,3523319,'10,August,2004','2004-08-10 00:00:00','   point scoring starts about 12.30 on saturday, with the big gay spurs/liverpool game, so that would appear to be about the deadline. i think i can add teams to the mini-league at any point. i think...  '),(16218,3523319,'10,August,2004','2004-08-10 00:00:00','   I know it\'s the sort of thing that I\'ll tire of a few matches in though...(excuses out of the way) I\'ll sort it out sometime today!  '),(16219,3523319,'10,August,2004','2004-08-10 00:00:00','   As suspected the BBC have made a mistake on the  urlLink Cambridge United page  of their  urlLink web site.   Shall I email them in an irate fan style.  How late can I leave my doomed decision to enter the big gay ff league?  '),(16220,3523319,'10,August,2004','2004-08-10 00:00:00','   I was going to play but I see Sawston Rovers isn\'t among one of the teams from which you can choose players.  As a result I cannot choose myself, Spence, Saggs, Shoulders, Steve, Sam Simon or Big Dave and so can\'t really see myself picking any team that will do very well.  Although last time I did come fourth having forgotten about my team three days into the season.  '),(16221,3523319,'10,August,2004','2004-08-10 00:00:00','   i\'m pretty sure that united were at home last night - we usually lose at home to orient, so that\'s a point of profit right there.  manx and aiden are on board the fanfoot train. where\'s your entry, damo?  '),(16222,3523319,'10,August,2004','2004-08-10 00:00:00','   I watched I,Robot and Catwoman a couple of nights ago on DVD.  I,Robot was better than I expected whilst Catwoman was...well...absolutely shit.  '),(16223,3523319,'10,August,2004','2004-08-10 00:00:00','   that\'ll be the \"security\".  i\'m about to add a batch of leaguers. are you playing, tom? head for www.tffo.com if so.  '),(16224,3523319,'10,August,2004','2004-08-10 00:00:00','   Everyone knows you should wait for Service Pack 1 for Service Pack 2 to come out before installing Service Pack 2 for XP...christ Ross, what were you thinking?  '),(16225,3523319,'10,August,2004','2004-08-10 00:00:00','   I used my hotmail, (fallon_paul2@hotmail.com) mini-league me up  '),(16226,3523319,'10,August,2004','2004-08-10 00:00:00','   I installed Service Pack 2 for XP last night, like a good little Microsoft minion. Now my machine hangs a lot for no reason.  Nice one, Bill.  '),(16227,3523319,'10,August,2004','2004-08-10 00:00:00','   That\'s nothing, the mighty Gills beat Leeds last night at home. Myabe we won\'t suck so hard this season.  Manx, did you use your hotmail address to sign up to Fan Foot?  If not, you need to let Dave know which e-mail address you used, and he\'ll put you in the mini-league.  '),(16228,3523319,'10,August,2004','2004-08-10 00:00:00','   F*** Me Dave, the U\'s actually managed a home draw last night!  Although there seems to be some confusion on \'the Internet\' over where they played.  Isn\'t Brisbane Road Orient\'s home ground?  '),(16229,3523319,'10,August,2004','2004-08-10 00:00:00','   Hmm, i\'m unlikely to maintain interest all season enough to keep changing players every week.  anyway I picked all the same 11 i have in my work fantasy team then i added 7, which gives me a pretty balanced squad anyway.  do i need to be added to a mini leage then?  '),(16230,3523319,'09,August,2004','2004-08-09 00:00:00','   That was my thinking, until I realised I could afford more than 11 really good players.  The trick is, you can select players for Saturday AND Sunday games, so you can get all 18 of your players to play each week if you get it right.  But if seven of them are bollocks, you\'ll only get 11 of them playing.  '),(16231,3523319,'09,August,2004','2004-08-09 00:00:00','   Re: Fantasy League  Is there any reason not to just buy 11 really good players and make the other 7 a bunch of cheap Crystal Palace losers. Do only 11 score for you at a time?  '),(16232,3523319,'09,August,2004','2004-08-09 00:00:00','   No...who do you think I am...Mary f***ing Poppins?!  '),(16233,3523319,'09,August,2004','2004-08-09 00:00:00','   i\'ve seen you trying to work out what hitler would have done, damo, but then aren\'t you supposed to do the opposite?  '),(16234,3523319,'09,August,2004','2004-08-09 00:00:00','   In light of your \"famine\" related post...the \"reactions\" column about Sudan on the Onion is quite amusing  '),(16235,3523319,'09,August,2004','2004-08-09 00:00:00','   That reminds me, I\'ll be flapping my wings over Somalia tonight to drop off food supplies and other aid packages - has anyone got any cans of food to donate?  '),(16236,3523319,'09,August,2004','2004-08-09 00:00:00','   Give me Â£20 Dave  '),(16237,3523319,'09,August,2004','2004-08-09 00:00:00','   we couldn\'t help it - once fonz saved the first rainforest, it was like an addiction for us. now it\'s just a world of charity round these parts.  '),(16238,3523319,'09,August,2004','2004-08-09 00:00:00','   Hmm, I see the BOA now has ads at the top Iran Earthquake relief.  have you all turned into a bunch of fucking tree huggers  '),(16239,3523319,'09,August,2004','2004-08-09 00:00:00','   ross is the one on the left.  '),(16240,3523319,'09,August,2004','2004-08-09 00:00:00','   This fight looks pretty cool!   urlLink http://www.ananova.com/news/story/sm_1051853.html  '),(16241,3523319,'09,August,2004','2004-08-09 00:00:00','   i love that spud-faced pigdog.  not vin diesel.  '),(16242,3523319,'09,August,2004','2004-08-09 00:00:00','   Is it Vin Diesel shouting \"Welcome to the Xander zone?\"  '),(16243,3523319,'09,August,2004','2004-08-09 00:00:00','   \"Science has unlocked the gates to the unknown, and now only one man stands between Hell and Earth. And no, it\'s not Wayne Rooney.\"  So say Game about Doom 3. Ho ho.  '),(16244,3523319,'09,August,2004','2004-08-09 00:00:00','   he would have got the email that banksy sent, and the one i followed it up with, so it\'s up to him i think.  '),(16245,3523319,'09,August,2004','2004-08-09 00:00:00','   \"Ross wins\" is the phrase you\'re after.  I\'ve said it before and I\'ll say it again, everyone should sign up to the Premiership at  urlLink www.tffo.co.uk .  Do you think we should drop manx a note about it?  '),(16246,3523319,'09,August,2004','2004-08-09 00:00:00','   I took an umbrella. I rule.  '),(16247,3523319,'09,August,2004','2004-08-09 00:00:00','   where are the paint-tanks and paint-artillery-strikes?  i am a very wet boy this morning. the light shower when i left the flat quickly became a torrential downpour. my shoes are not, it turns out, waterproof, and nor am i.  it\'s unpleasant.  '),(16248,3523319,'09,August,2004','2004-08-09 00:00:00','   I\'d also quite like it if there were medics who ran around wiping paint off for you - If they manage it within 30 seconds you live...obviously this is a shit idea but it would allow me to shout \"Reviiive me\" as well as the statutory \"I neeeed Ammo\"  '),(16249,3523319,'09,August,2004','2004-08-09 00:00:00','   As I lack the skill and know how of those that play paint ball all the time the only real chance I had of victory against a team of profs last year was to purchase a shedload of ammo.  Fed up with losing, on the last game myself and three friends spent around Â£30 each on ammunition.  We were protecting the base...we spent the whole game firing continuously...there\'s nothing like hearing 4 or 5 of the opposition shouting \"You\'ll run out of ammo soon\" from behind a tree then 5 minutes later see them make a last ditch charge into a hail of gunfire. Next time I hope to discover the No_Clipping Mode so I can run through trees.  I don\'t think I\'ll bother with God_Mode though as it will probably make it all a bit too easy.  '),(16250,3523319,'09,August,2004','2004-08-09 00:00:00','   it\'s from spaced. i said it a lot that day.  that and \"ow\".  '),(16251,3523319,'09,August,2004','2004-08-09 00:00:00','   Is that a joke off Frasier Yoav?  '),(16252,3523319,'09,August,2004','2004-08-09 00:00:00','   I almost got shot in the nads. Luckily, my cock got in the way.  '),(16253,3523319,'09,August,2004','2004-08-09 00:00:00','   \"You shot me in the bollocks Tim.\"  \"Like I said - no hard feelings.\"  '),(16254,3523319,'09,August,2004','2004-08-09 00:00:00','   Did anyone get shot in the nads?  '),(16255,3523319,'09,August,2004','2004-08-09 00:00:00','   I\'ll have you know I was a fucking ninja in one paintball game, when I single-handedly stormed a temple and took out four bastards, and then got some stirling backup from Wee Dave and his sister, and we took out another two or three.  Of course, when I got shot and went back to the elimination zone, there were only three opposing players there. You do the math.  '),(16256,3523319,'09,August,2004','2004-08-09 00:00:00','   Tom has slipped back into his naughty your/you\'re ways. He must be punished.  '),(16257,3523319,'09,August,2004','2004-08-09 00:00:00','   they\'re sly, these dawgs. still, ross and i heroically hid in our base during one game, and were shot to shit in the bushes.  they gunned up my knees, the bastards.  '),(16258,3523319,'09,August,2004','2004-08-09 00:00:00','   Do you remember I got done for pistol whipping those guys who wiped paint off themselves and then shot me back at close range.  No point going if your going to have to play against a bunch of chavs.  '),(16259,3523319,'09,August,2004','2004-08-09 00:00:00','   i am a bad, bad soldier. i\'m too big a target, and crouching makes my knees hurt.  i got the cack shot out of me.  still, we were playing against a bunch of cheating army-boys, who HAD A FUCKING PASSWORD and wiped paint off themselves when there was no marshal around. it wasn\'t as much fun as i\'d hoped.  but dave was a winner. a WINNER.  '),(16260,3523319,'09,August,2004','2004-08-09 00:00:00','   Got it from Game...they tend to deliver between 3 and 7 days before UK release date -tis good.  Got a mail today that it was dispatched for next day delivery this morning...Think the site is  urlLink www.game.uk.com .  Only time I buy games is if I want to do the multiplayer stuff...else I would have downloaded Doom.  '),(16261,3523319,'09,August,2004','2004-08-09 00:00:00','   Did Ross get shot in the hand four seconds into a half hour game?  '),(16262,3523319,'09,August,2004','2004-08-09 00:00:00','   Did you have as much success at paint-balling this time as you had at the Dragon\'s Lair in Brentwood?  '),(16263,3523319,'09,August,2004','2004-08-09 00:00:00','   yes, it\'s going to hurt in many, many ways. but you only live once, eh? and at least i get to choose the manner of my passing.  or is that my pissing?  '),(16264,3523319,'09,August,2004','2004-08-09 00:00:00','   Dave, apparently you are an Edinburgh virgin and your going for the first time with Billy!  I must warn you when me and Patrick went up there for a few days we got pissed on the plane on the way up and didn\'t return to sobriety until at least twelve hours after we\'d crossed south of the border.  This was mainly due to Bartram showing off his enhanced student lifestyle where if he wasn\'t at the radio station he was either in an establishment that specialised in the sale of alcohol or at home where his fridge contents consisted solely of Stella, Vodka and Cranberry Juice!  Plus you are going for some Whiskey tasting do!  '),(16265,3523319,'09,August,2004','2004-08-09 00:00:00','   what aboot far cry? and when am i getting my half life 2?  '),(16266,3523319,'09,August,2004','2004-08-09 00:00:00','   I\'ve played it for about 7 hours now in an optimum environment of small dark room encrusted with speakers on my own and the more I play it the more boring it gets. There is literaly no variety - all the rooms look the same the monsters are dull and very very stupid. The weapons are crap and its not scary. Its pretty but not ridiculously so and it really is a big tech demo to make people licence the engine. As such it should be free and you shouldn\'t buy it. Just download the fucker and get bored with it and imagine what companies will do with the engine.  So there.  '),(16267,3523319,'09,August,2004','2004-08-09 00:00:00','   doom cops: the taste of mars justice.  '),(16268,3523319,'09,August,2004','2004-08-09 00:00:00','   No-one should buy doom cops its rubbish.  '),(16269,3523319,'09,August,2004','2004-08-09 00:00:00','   Damo, where are you getting Doom from? Play still say they\'re getting it on Friday.  '),(16270,3523319,'09,August,2004','2004-08-09 00:00:00','   Yeah, it\'s not the sort of game you play with other people around though to properly experience...it probably takes a reasonable amount of time to get into as well.  Either way it\'s got to be better than Soldner which sucked complete monkey balls.  '),(16271,3523319,'09,August,2004','2004-08-09 00:00:00','   No-one can come close to my 4 million in one week Keanu-a-thon. No-one. Ever.  '),(16272,3523319,'09,August,2004','2004-08-09 00:00:00','   i believe that i have made my daq-tactics clear in the past. i am still living in hope of a hostile knightley-takeover.  '),(16273,3523319,'09,August,2004','2004-08-09 00:00:00','   Did anyone notice Matt getting 2 million squid in the \'Daq in one week this week?  I think he wasn\'t far off three million too.  I reign supreme in the League Of Kings.  '),(16274,3523319,'09,August,2004','2004-08-09 00:00:00','   i played some doom 3 over the weekend at dave\'s. i was hella-disappointed.  '),(16275,3523319,'09,August,2004','2004-08-09 00:00:00','   I just had an email saying my copy of Doom3 is arriving tomorrow...cool.   Doubt I\'ll be able to play it though as turning on my computer for any length of time results in the room temperature breaking 40 degrees...heat and rotting flesh don\'t mix too well.  '),(16276,3523319,'09,August,2004','2004-08-09 00:00:00','   I think I may be allergic to mine....every now and again they hurt.  '),(16277,3523319,'09,August,2004','2004-08-09 00:00:00','   You get up and go to work on a Monday morning,  you\'re gonna have a bad time , ok?  '),(16278,3523319,'09,August,2004','2004-08-09 00:00:00','   do you concur? i do.  '),(16279,3523319,'09,August,2004','2004-08-09 00:00:00','   Fuck it up its stupid ass.  '),(16280,3970082,'29,July,2004','2004-07-29 00:00:00','   God Bless Reverend Al Sharpton. I have been avoiding the Democratic National Convention. I did not want to hear yet another sanitized speech which gave no opinions about the way this country has been run since the 2000 election. I did not want to hear once again, that the Democrats were not going to attack Bush, or take a stand on any of the numerous issues they could be talking about. Instead, the Democrats have been looking like ostriches with their heads in the sand, looking for their butts. This morning, I heard about Al Sharpton, and the fact that he did not give the cookie-cutter speech approved for by the DNC, and instead spoke from the heart. He spoke the words of a lot of Americans, words I have been hoping to hear since the primaries. Here is a quote from the speech last night. I had hoped to copy the entire speech, but discovered that because he did not give the speech the DNC had approved, no news agency so far has a copy of the speech. This is the passage that touched me the most (Quoted from  urlLink www.signonsandiego.com ):  \"Mr. President, the reason we are fighting so hard, the reason we took Florida so seriously, is our right to vote wasn\'t gained because of our age,\" Sharpton said. \"Our vote was soaked in the blood of martyrs, soaked in the blood of (civil rights activists) Goodman, Chaney and Schwerner, soaked in the blood of four little girls in Birmingham. This vote is sacred to us. This vote can\'t be bargained away. This vote can\'t be given away.  \"In all due respect, Mr. President, read my lips: Our vote is not for sale.\"   Mr. Sharpton has earned my respect, and my admiration. I did not expect much out of this election process, except to hopefully oust President Bush. A lot of you may have heard me declare that I plan on moving to Canada and pleading for political asylum on Wednesday November 3rd 2004 if we re-elect President Bush. Mr. Sharpton has reinvigorated my desire to work toward getting John Kerry elected. He may not have the ability to declare his opinions as bluntly as Reverend Sharpton can, but he can be the face that helps to make those opinions fact.   I\'m gonna end this entry with one more quote from the honorable Reverend Sharpton: \"The issue of government is not to determine who may sleep together in the bedroom, it\'s to help those that might not be eating in the kitchen.\"   I now wish I had spent last night listening to his entire speech. It definetely gives me hope for the future.  '),(16281,3970082,'26,July,2004','2004-07-26 00:00:00','   For the past few days I have been thinking of my friend N/A. Yes, I called him N/A for various reasons including the fact that he liked to be known as \"not-applicable.\" He was a tall, thin man who liked to walk, liked to drink and liked to cause mischief. Ok mischief is not the&nbsp;right word, he liked to get in trouble...liked to push boundries...and liked to do&nbsp;most of it while intoxicated.   I remember he only had one pair of shoes, a pair of brown leather mocassins that he wore without socks. He&nbsp;walked in those shoes&nbsp;across the city of Washington DC, rather than spend the $2.20 to take the&nbsp;metro. He wore holes in the soles of the shoes, yet refused to&nbsp;buy new ones until the snow got so bad he had to, or&nbsp;risk frostbite.   I remember one night in the dorm, he came over to hang&nbsp;out with me,&nbsp;some other friends.&nbsp;He had a bottle of Banana flavored liquor. It tasted like a combination banana and cough syrup. It was some of the most foul stuff I have ever consumed in my life...but we drank it. We were young, stupid, and a tad self-destructive.   We wandered down to the music room, and played chopsticks on the piano. He leaned forward, &nbsp;looking like he was going to kiss me. I was sitting on the floor, and he was sitting on&nbsp;a&nbsp;chair with an attached&nbsp;desk-arm. He nearly decapitated me. That is a memory I will never forget...nearly losing my life in the music room...by a chair decapitation.   I hope whereever he is now, and whatever he is doing, he is&nbsp;living truthfully.  &nbsp;  '),(16282,3970082,'26,July,2004','2004-07-26 00:00:00','   I own an umbrella that I have chased around the country. I have left it in truckstops, and restaurants, and have back-tracked 100 miles to retrieve it. It may seem silly, tracking down an umbrella, but this umbrella represents for me, one of the darkest times in my life.   I was 20, almost 21, and home from college because I was sick. No one was completely sure what was wrong with me, except that I was dizzy, fainting, and had massive headaches. I was scared. I felt completely alone and isolated from everyone around me. Part of it was my own fault, dealing with people made me miss the life I had been leading as a care-free college student. The other part of it was the \"fault\" of the people who were around me. They had no clue how to deal with someone who was sick, and not elderly. I was also incredibly afraid of people.  Still, I digress, this posting is about the umbrella. During this time, an exhibit on Monet came to the Allbright-Knox Art Gallery in Buffalo. It was only pictures he had done at Givenchy, and I really wanted to go. My parents, and their friends got us tickets. We went to the exhibit, and about halfway through it, I nearly passed out.   I remember my father sitting with me in a chair, and that I never finished viewing the exhibit. Monet is one of my favorite artists, and I felt really horrible, since I had wanted to view the entire showing. I wanted to cry.   My parents bought me an umbrella, with Monet\'s water lilies on it. I got better, and have gone on with my life. Still, I keep this umbrella, to remind me of a time when I was so weak, I needed my father to help carry me.  '),(16283,3970082,'26,July,2004','2004-07-26 00:00:00','   I am sitting in Starbucks, reading a magazine article about the gentrification of New Brunswick NJ written four years ago, because there is also an article about the 2000 Democratic convention in it, and wondering why? There was a time when I would have scoffed at sitting in a Starbucks, and would have chosen a non-chain alternative. I have one, 2 blocks away. I could have gone there, but I chose to stop here.   Since moving to South Florida, I feel like my level of consumption has risen. Maybe itâ€™s the fact that I am living with my parents, and being around them for long periods of time these days makes me sick. I find myself driving, thought I know it is bad for the environment. I look for jobs in the next county, or the county above that, just because I know I would not be able to live with them and work, if I got those positions. Every time I fill up my gas tank, I know I am being cruel to the environment, but I canâ€™t help myself. Its like I am saving my sanity by polluting the planet.   I eat out more than I did up north and I usually consume fast food. Burger King and Wendyâ€™s may no longer use cardboard boxes to wrap their food, but they still use so much stuff, I feel like I have killed a couple trees in the past 4 months.   When did I become this person, who does not take into account my level of consumption as I lead my life? When did the term â€œneedâ€ start working with a pair of ruby red high heels that are so insanely high I could never wear them? Last night, I said that, even though they were probably the most frivolous item I could buy. Wait, no, considering the store I was in, the pair of shoes is not the most frivolous thing I could have purchased.&nbsp; Still, realizing that I used the word, â€œneed,â€ when it came to shoes, alarms me a lot.   Just now, I witnessed a scene where a toddler was being entertained by his big sister. She was pushing him back and forth with her feet, in his stroller. Their father did not like the fact that the toddler was laughing really loud. I was smiling at the children, and when the father interceded and told them to calm down, I gave him a look. He took it to mean I did not want the children to be loud; when in fact, it was just the opposite. I wanted them to be louder. I wanted them to scream and run around this Starbucks, and threaten to tear the place apart. And yet to that father, I looked like a person who did not like rambunctious kids.   Iâ€™ve become a maâ€™am. In restaurants and at work, especially when I am wandering around the office in a business suit, advising people twice my age on financial planning and their taxes, I get called maâ€™am. Iâ€™m too young to be maâ€™am. Maâ€™am is a term for my mother. It is a term for older women with greying hair. Wait, I have at least five grey hairs at the moment. So that definition does not work for me. Women who are called maâ€™am worry about mutual funds and the stock market. Ok, that doesnâ€™t work either, since I do keep track of my measly portfolio and wish I could put money away for the future. I guess I donâ€™t have any more arguments against being maâ€™am. I have reached that age.   It makes me amused to realize that a lot of the declarations about myself I made at 16, I have reversed by the time I was 26. I told myself at 16, no matter my interest in the Middle East, I would never work for an oil company. Now I wonder if Halliburton is hiring, though I donâ€™t bother going to their website to seriously pursue it. I did look at BP and Exxon-Mobil. Just looking at the websites made me sick to my stomach. I felt sick, mainly because I realized, I could work for those companies. I could sell my soul to a corporation and be relatively happy. The thing that stops me is the move to Texas or Oklahoma and the fifty-mile a day commute out to the suburbs I see in my future. Oh, and donâ€™t forget the SUV and the 2.5 kids. I have to make sure they are in the picture.&nbsp;   &nbsp;   &nbsp;  '),(16284,3970082,'25,July,2004','2004-07-25 00:00:00','   I have had four hours sleep, and should be doing something more productive than writing yet another blog message, but the message I had written, and was going to post...an analysis of my consumptionist ways, is on a floppy disk. The computer in the library where I am sitting, a nice gateway pentium 1 by the way, has a broken floppy disk drive, and of course no USB port. Therefore I am screwed. So I have decided to rant for a few minutes about the lack of technology in today\'s libraries.   Personally, I am against computers in libraries, particularly in school libraries. I think there is something to be said about the tactile feeling of finding a book before writing a paper. Looking my first book up alone in the card catalog is a memory that will last me the rest of my life. It was a big accomplishment, since I didn\'t ask the librarian or teacher to help me, and I found the book myself. Its hard for me to picture my future children not having that same experience. Just like&nbsp;it is hard for me to know that they will probably have worse handwriting than I do, (And those of you who know me, know my handwriting sucks) because I know that their teachers will demand they type their assignments, instead of being forced to decipher primitive script.   Still, if there are going to have to be computers in every day use, I really would like them to be functioning well. I really wanted to tell the librarian at that branch, that I could streamline their access to the internet, and make the webpages load faster, if she just gave me 10 minutes. I figured that they would think I was gonna hack into their system or something, so I didn\'t. Instead I left the library feeling like I wasted my 45 minutes. I wish I could have done something for them, because I know other people don\'t have the access to the internet that I do. I am lucky, I have it at home. I just use the libraries for convenience when I am out and about. As I do my job hunt, more and more I am refered to websites, and I know that there are lots of people out there who cannot apply to those jobs, because they have no access to the sites. It makes me sad, that the technology exists, but the people do not have access to it. The lines in the libraries prove it. Down here in Miami-Dade, I have witnessed people waiting an hour for their 45 minutes to send e-mail or do a job search. That is all the library can afford to give them.  &nbsp;  '),(16285,3970082,'25,July,2004','2004-07-25 00:00:00','   Ok Folks, I am posting my first new poem in several years. I may get rid of it, since I am not sure if I like it yet...but for now...here it is:   Moanings at 3 in the Morning   I ache, deep in my soul, where does it come from? why can\'t it stop? I try so hard, so hard to feel like there is more to me than a reflection of melodrama. I hear the whine in my voice. I crave more, but I fear to give into  temptation. I fear giving into myself. Why don\'t I know what I want? Why isn\'t it tangible? I can\'t see it. I can\'t let go of the past. I can\'t picture the future. Dreams fall away, carving a hole in reality. Leaving me empty.  '),(16286,3970082,'24,July,2004','2004-07-24 00:00:00','   â€œThe Eyes of Texas are upon you, all the livelong day. The eyes of Texas are upon you. You cannot get awayâ€¦â€ is a line that conjures up memories of baseball caps and very loud marching bands. In Austin, that song is sung at every sporting event UT has. Even though, I never attended UT Austin, the songs, and the spirit are ingrained in me. I found myself recently making fun of Aggies (graduates of Texas A&M) with a friend of a friend over dinner. She had lived in Dallas. I had lived in Austin. We had nothing in common, but that was enough. We were able to converse like good friends, after a couple drinks and a round of Aggie-bashing. Iâ€™ve discovered, since leaving Texas there are two types of peopleâ€¦people who have been to Texas, and people who have not. People who have been to Texas, even the ones that hate it, never completely put down the state. The people, who have never gone to the state, are the ones who bash it.   I admit, it is the state that brought us two President Bushes, but it is also the state that brought us President Johnsonâ€¦so in a way, it will all balance out in the end. Well, maybe not, but this is not a treatise comparing the presidents who came out of Texas. What is this exactly? Iâ€™m not sure. I guess in a way it has to do with poetry, and the song, â€œYellow Rose of Texas.â€ I was recently exposed to the fact that any poem by Emily Dickinson can be sung to the music of â€œYellow Rose.â€ If Ms. Dickinson was writing her poetry today, she might choose to have it come out that wayâ€¦. just to poke fun at the state of Texas. Instead, it is just another one of those things that people with too much time on their hands discover on the Internet.  '),(16287,3970082,'21,July,2004','2004-07-21 00:00:00','   I watch the people around me searching for love. They cling to each other in some dark misshapen attempt to exorcise the demons within them. They call physical attraction and a sense of trust, love. It makes me wonder, as my friendsâ€™ marriages break up, and others find their soul mates, where does the line existâ€¦and how do we define love? Most days I take a Supreme Courtâ€™s definition of pornography approach to love, â€œIâ€™ll know it when I see it.â€ Still it makes me wonder, as the happy couple turns into the shrill singleâ€¦what are we all really searching for? In my quest to find answers, to questions yet unasked out loud, I have turned to poetry. I just finished Stephen Dunnâ€™s book  Different Hours , the book that won him the Pulitzer Prize. It was impressive, and a bit overwhelming, to read his words and connect with them so forcefullyâ€¦but this poem entitled â€œTheir Divorceâ€ really illustrates how I feel each time I watch the friends in my life separate from each other:   Their Divorce  &nbsp; Not them. Not even with the best binoculars&nbsp; on the bluest day could I have seen it coming. Not with scrutinyâ€™s microscope, or with the help of history or gossip.  Of all the people, not them.  They hadnâ€™t fallen in love with others.  Not even a night of drink or proximityâ€™s slow burn drove them  to lapse, say, with a coworker. It means no one can know what goes on in the pale trappings of bedrooms,  in anyoneâ€™s secret harrowed heart.  It makes time itself an executioner- a fact I always knew applied to couples  whose bodies contradicted their Darling this, Honey that, and even some who exhibited  true decency and respect. But this is a mockery, a defeat. My friends were perfect, perfect. â€œEvery married couple appearing together in public is comic,â€ Adorno said, and I wrote â€œStupid!â€ in the margin. Now theyâ€™re broken up, finished. Oh Adorno, you son of a bitch, you perspicacious bastard, sometimes what a cold eye sees lasts longer than any of us.    Copyright 2000 by Stephen Dunn  &nbsp; Stephen Dunn is a new discovery, only recently introduced to me by the fact that he won the Pulitzer and his book was in the Miami Dade Public Library. I should own more poetry than I do. I used to devour it like pancakes for Sunday morning breakfast. Now, I only nibble at the edges of it, like its expensive caviar, and I donâ€™t want to finish off the last cracker.   Still, back in my feasting off poetry days, I consumed a vast quantity of Anne Sextonâ€¦probably at too early an age to understand what she really meant, and definitely at a time when I could not discern all metaphors from real life. I am not ranting about my history in this message, I am making allusions to poems that make me think about the break ups of friends. I probably should just sit down and write my own poems, but since I am wary of opening that Pandoraâ€™s box, I will post an Anne Sexton poem that also makes me think about the shredding of the heart. It is called, â€œThe Wedding Ring Dance,â€ and was published in  45 Mercy Street  in 1976.   THE WEDDING RING DANCE  &nbsp; I dance in circles holding the moth of the marriage, thin, sticky, fluttering its skirts, its webs. The moth oozing a tear, Or is it a drop of urine? The moth, grinning like a pear, or is it teeth clamping the iron maiden shut?  The moth, who is my mother, who is my father, who was my lover, floats airily out of my hands and I dance slower, pulling off the fat diamond engagement ring, pulling off the elopement wedding ring, and holding them, clicking them in a thumb and forefinger, the indent of twenty-five years,  like a tiny rip leaving its mark, the tiny rip of a tiny earthquake, Underneath the soil lies the violence, the shift, the crack of continents, the anger, and above only a cut,  a half inch space to stick a pencil in.  The finger is scared but it keeps its long numb place, And I keep dancing, a sort of waltz, clicking the two rings,  all of a life at its last cough, as I swim through the air of the kitchen, and the same radio plays its songs and I make a small path through them with my bare finger and my funny feet, doing the undoing dance, on April 14th 1973, letting my history rip itself off me and stepping into something unknown and transparent,  but all ten fingers stretched outward, flesh extended as metal waiting for a magnet.   &nbsp; I read that poem the first time fifteen years ago. Then, I saw it as a piece reminiscing about the wedding. Then again, I was 11, and not nearly as cynical as I am now. Me, cynical? Nahâ€¦never!!! I think that is part of why I am writing this blogâ€¦in order to stop hiding me. I have written other things, and put them out there, but I was always afraid of what people would think, so I censored myself. I am sick of having to choose between friends as they separateâ€¦but in order to stay a good friendâ€¦I donâ€™t tell them so. I just write it in my blog so anyone in the world can read it.&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  '),(16288,3970082,'21,July,2004','2004-07-21 00:00:00','   The undertoad is strong today. That irrepressible feeling that something bad is going to happen. Nothing bad may happen. I may be fine today. It may be yet another day, full of banter and nostalgia, but it could be painful and scarring at the same time. I cannot get out of my head today. All I want to do is sleep, but I am stuck in the middle of a traffic jam on 95. I wonder, if I napped right now, would anyone notice?   I write sonnets in my head as the radio blares. Who cares about being springtime fresh? Why is the triangle considered an instrument in elementary school music classes? Isnâ€™t it just there so the kids who cannot carry a tune can participate?   The air conditioner kicks off. It keeps doing that, stopping in the middle of 100 degree daysâ€¦as if to remind me it does not want to be down in Miamiâ€¦it prefers the cold of the Buffalo winter. I remember when we bought the carâ€¦it is whiteâ€¦â€A Florida colorâ€ my father said. I had declared that the only color car I did not want was white. I drive a white car in Florida.   In Buffalo, during snowstorms, my car would get buried and completely blend in. I used to think of it as one of those monkeys that live in Nepal, near the hot springsâ€¦.you never see them until you get right on top of themâ€¦or they dive into the hot water to warm up.   Now my car is easily spotted in every parking lot. It glows like a radioactive sheep. I can hear it saying, â€œBaaaaahhhh, baaaahhhhhâ€ as I turn it on. It likes me, my little Iris does, yes I named my carâ€¦but it does not like having to run its air conditioner during the Miami afternoons.   I keep half a bicycle in the back seat. Just the wheels, for they are easily accessible for spinning at long stop signs. I see them sometimes, and tell myself I should put them in my storage locker, but I never do. Other times, I look at them, and think,&nbsp; I could make a really great diorama of&nbsp; the solar system with themâ€¦but then I rememberâ€¦no one in the real world makes dioramas.   Then I realize, that is the reason adults become elementary school teachers. They have dreams of leading children in creating dioramas, so they go back, into the cavernous halls from which they came. I need to be more specific in my reference there, but I really donâ€™t want to be, so I wonâ€™t.   I will just go back to driving along 95, and wondering why it is that everyone else is on the road with meâ€¦do they also have no where to go? Hello Mr. Undertoad, Iâ€™ve been expecting youâ€¦   For those of you who do not know, the undertoad is a reference to  The World According to Garp  by John Irving .  '),(16289,3970082,'21,July,2004','2004-07-21 00:00:00','   Several years ago, my friend Rimma and I made a pact. We would take the Foreign Service Written Exam, everytime it is offered by the State Department, until one of us passes it, or one of us moves out of the suburbs of Buffalo NY. That pact lasted through one all-day Saturday exam.&nbsp;Rimma decided not to take it again, but that did not stop me from religiously signing up the next time around and the next time. Yes, I took the exam three times. I wasted three Saturdays, trying to remember the dates of the Franco-Prussian War and why Alexander Hamilton was famous. I still don\'t know the dates of the Franco-Prussian war.&nbsp;I was such a&nbsp;mascochist, I opted to take the exam&nbsp;once I moved out of Buffalo, and down here to Miami. I gave up attending a large party for my office, in order to sit in a university auditiorium and once again&nbsp;remember what \"opportunity cost\" means in economic terms. Finally, my wasted Saturdays paid off. Today, I received a letter from the State Department, letting me know I&nbsp;passed the exam, and have been invited to interview with them. I know all that means is I made it over the first hurdle. I know there are a lot more before I can say I work for the State Department, but today, just for today, I plan on reveling in the fact that I accomplished something that&nbsp;very&nbsp;few people in the country even attempt to do. YAY ME!!!!!!!!!!!!!!!&nbsp;&nbsp;  '),(16290,3970082,'20,July,2004','2004-07-20 00:00:00','   These articles&nbsp;are actually from the AP \"News Of The Weird\" on Yahoo:   Vandal Edits Swear Words in Library Books      Tue Jul 20, 7:46 PM ET       Add  urlLink Strange News - AP  to My Yahoo!   LAYTON, Utah - A self-appointed editor of library books has given new meaning to \"purple prose.\" The do-it-yourself censor is turning swear words into \"darns\" and \"hecks\" â€” in purple ink.       &nbsp;  \"They believe it\'s within their right to deface public property and impose what they believe on others,\" library director Pete Giacoma said.   \"It\'s a crime. It would be prosecuted if we were to find who did it, by luck or accident,\" he said.   Destroying or defacing library materials is a class B misdemeanor. It\'s punishable by up to six months in jail and a $1,000 fine, but Giacoma said the library usually will settle for restitution.   \"I think the worry of the public, every once in a while, is that we\'re doing it,\" he said. \"We\'re not.\"   Giacoma said in cases like this, the vandals will usually tire of editing books and stop on their own.    And now for article number 2:    Alleged Library Book Thief Sparks Chase      Mon Jul 19, 2:01 PM ET       Add  urlLink U.S. National - AP  to My Yahoo!  By WILLIAM KATES, Associated Press Writer  SYRACUSE, N.Y. - A 36-year-old man led police on a brief car chase, driving on sidewalks, through parking lots and even against traffic on a busy boulevard, so he wouldn\'t get caught with stolen library books, according to police.       &nbsp;  \"The officers were a little taken back when they found out what the deal was. They couldn\'t believe it,\" said Syracuse police spokesman Sgt. Tom Connellan, who added police broke off pursuit because the situation Sunday became too dangerous.   \"It was stupid and reckless,\" said Connellan.   There were no injuries, and â€” even more amazingly, Connellan said â€” no property damage.   Byron Haynes of Syracuse was to be arraigned Monday in Syracuse City Court on misdemeanor charges of reckless endangerment, petit larceny and reckless driving, as well as numerous traffic infractions.   Haynes offered no explanation to police when he was arrested, Connellan said.   Connellan did not have the titles or authors of the five stolen books but said they all dealt with Jewish religion. The books were swiped from Syracuse University\'s Bird Library, which has substantial holdings of rare books, manuscripts and collections, including original writings of authors Joyce Carol Oates and Dr. Benjamin Spock, 19th-century reformer Gerritt Smith and transcontinental railway builder Collis Huntington.   \"Our special collections are kept on the sixth floor. Access to that area is very regulated. These were taken off the shelves that everyone has access to,\" said Kevin Morrow, a university spokesman.   Haynes was not a university student, Morrow said.   An officer spotted Haynes running a stop sign about 4:30 p.m. Sunday near the library. Moments earlier, a suspect had fled the building with the stolen books, setting off the building\'s alarms as he ran out the doors.   The car chase lasted about 10 minutes. A dozen police cruisers were involved, although only three or four chased Haynes. During the chase, officers saw the suspect toss a backpack out the car window, Connellan said. When they recovered it, they found the stolen books.   \"We assumed it was drugs or guns the way he was driving,\" Connellan said. \"We even brought in a canine to go through the car. Nothing.\"   Onondaga County deputies in a helicopter tracked Haynes until he tried to ditch his car at a city park. City police arrested him after a short foot chase.   Connellan said investigators were doing a background check on Haynes.    So this sparks the question....Why are libraries so contraversial all of a sudden?  '),(16291,3970082,'20,July,2004','2004-07-20 00:00:00','   I have started writing poetry, again, and am going to post some on here soon. To prepare myself for the inevitable doubts about doing that, I am posting some stuff I used to have posted on my website...that I still have posted on that site. I hope you enjoy it....and I promise new stuff soon.   Writing by Candlelight   Scraps of paper, ripped and torn; like a heart on the edge of commitment. dividing time between pleasure and procrastination. knowing that to leap the chasmwould force consciousness. not wanting to wake from this dream. Embedded in the torn up letters, words, sentences, admitting the truth and the birth of the knowledge that loneliness is fatal.     For You, Whatever Name You Use   So you know my secret, the iridescent knowledge of my heart as, if you\'d dined on it the night before Where are you thinking as I write these lines? The glow of your heartbeat next to me, bathing my senses in an acid overflow, Can it cleanse my soul? Through sleepless eyes, I stare at you, trying to bore a hole through your hormone encased shell, The soundtrack in my head has stopped. Someone pushed pause. Your dialogue starts, but the snoring in the background keeps me from hearing you clearly.  '),(16292,3970082,'20,July,2004','2004-07-20 00:00:00','   \"Someone to touch your hair, someone who cares.....Reach out and touch me!\" I am not sure if that is what the words are for that song. Still it feels right. I have come to the realization I spend an inordinate amount of time alone. As I job hunt, I am alone. As I drive, I am alone. I eat alone, watch TV alone, go to the movies alone, drink alone, read alone and feel sorry for myself for being alone...alone. Still its funny, most of the time I revel in being alone. Since moving back in with the parental units, they spend so much time wanting to know what is going on with my life, that being alone felt good. That is, until being alone became lonely. Now all I want are a group of friends I can call upon to no longer be alone.  '),(16293,3970082,'19,July,2004','2004-07-19 00:00:00','   I am a 26 single straight woman who believes in defending my right to get married. I have a right to not be asked by friends and relatives whether I am in a relationship. I have a right to tell those same friends and relatives to go to hell when they suggest I meet a \"nice guy\" and \"settle down.\" I have a right to have dreams that have nothing to do with a big white dress, seven bridesmaids and losing my born again virginity on my wedding night. I have a right to have my single male friends not ask me to be their bethrothed if they do not find the woman of their dreams by the time they are thirty, or thirty-five or forty...you get the picture. I have a right to feel insulted when I am told my biological clock is ticking. I have the right to choose who I sleep with, when I sleep with them, anot not have to commit my heart and soul to good sex. I have the right to get my heart broken at least half a dozen times before I choose the right guy for me. I have the right to choose when I want to have a family, and who I want to raise my children. I have a right to all of this, as long as my choice is a straight single man...and on days like today...that does not seem like much of a choice.  '),(16294,3970082,'19,July,2004','2004-07-19 00:00:00','   The undertoad is strong today. That irrepressible feeling that something bad is going to happen. Nothing bad may happen. I may be fine today. It may be yet another day, full of banter and nostalgia, but it could be painful and scarring at the same time. I cannot get out of my head today. All I want to do is sleep, but I am stuck in the middle of a traffic jam on 95. I wonder, if I napped right now, would anyone notice?  &nbsp;  I write sonnets in my head as the radio blares. Who cares about being springtime fresh? The rain comes pouring down, stranding me behind an 18 wheeler that takes up 4 lanes. I want to cry.  &nbsp; I want to shed tears for the woman I am, and the child I was. I want to have a home of my own, a career of my own, a life of my own, and I am sick of the protocol that is set up, that I must follow to get there. Why must I follow the rules? Why can\'t I be exempt? Why won\'t the headaches stop?  &nbsp; Why must everything be in triplicate, and typed, and double spaced? I wish just once I could stop the pain...that creeps into every decision I make...every part of my being...everything I create. Everyone judges me, and expects me to judge them in return. I am tired of judging...can\'t I just enter for the fun of it?  &nbsp; I hurt, am tired and hate my life....I thought it was supposed to get easier as we got older? Why does it just get harder? Why doesn\'t trying our best count for anything anymore?  '),(16295,3970082,'19,July,2004','2004-07-19 00:00:00','   I was just in Barnes and Noble looking for a magazine. It occured to me as I stared at the stacks, that there was not one \"womens interest\" magazine that interested me. Sure, I have read my share of Cosmo and Marie Claire...find me one PMSing woman who has not consumed the trash at some point in time...and I will show you a liar. Cosmo is like chocolate, it makes you feel good, or bad, depending on how much you consume. Still, I have no desire to learn how to give a better blowjob, get a man to be interested in me just by wearing a specific scent during the summer solstice or figure out which shade of white works best on my skintone.  &nbsp;  So I wonder, why is there no \"mens interest\" section? There are sports magazines, hobby magazines, political magazines and everything else, but men do not have their own interests. Don\'t men want to know how to give better head, or if black really will make&nbsp; them look slimmer?  &nbsp;  Wait a second, of course they don\'t, because the magazine publishing world is aware that they have a brain. The magazine world just assumes, that because I ovulate, that I am interested in magazines about weddings, there were 30 different magazines there, all about being a better bride....and even more about how to have a good pregnancy...and how to be a good parent. Why are there&nbsp;no mens magazines about good parenting? Isn\'t parenting a two person job?  &nbsp;  The older I get, the more my attitudes return to the time when I was in middle school. Back then, I could not understand the \"womens interest\" section. I also had no desire to comprehend it. Clothes were nothing more than the thing I put on in the morning. Makeup was for other people...an alien substance that I wasn\'t sure I ever wanted to experience. Somedays I miss that time in my life.  &nbsp;  &nbsp;  '),(16296,3970082,'19,July,2004','2004-07-19 00:00:00','   Today my father died. No, not today, in the year 2004, but today, in my mind 12 years ago. Twelve years ago today, my father received a heart transplant, and 12 years ago I killed him off in my mind. For one brief moment, I considered life without him in it. I was fourteen years old.  &nbsp;  I spent most of my conscious life with one idea in my head: my father is dying. I remember in kindergarten, explaining to a class of five year olds what a quadruple bypass surgery was.&nbsp; I was more than a precocious child. I was a freak. I had myself pegged that way from that point on.  &nbsp;  I didnâ€™t really think I was that strange till third grade, when I realized other children didnâ€™t come home to empty houses and messages on the fridge saying, â€œGone to the hospital, there is money for pizza on the counter. Donâ€™t watch too much TV. Get your homework done. Weâ€™ll go to the hospital after school tomorrow. Iâ€™ll pick you up early. Love, Mom.â€  &nbsp;  Other girls spent their pre-teen years fawning over crushes. I spent mine trying to sleep on orange vinyl couches and molded plastic chairs. I never learned what the latest clothing trends were, but I knew where to get the best sodas and snacks in any hospital in downtown Buffalo. Perfumes were never my thing, but I liked that they drowned out the stench of the emergency room.  &nbsp;  On the day my father got his transplant, we admitted him to Buffalo General Hospital. It was a beautiful sunny day, with temperatures about 80 degrees. I remember thinking I wanted to go swimming that afternoon. I was a bit angry that he was in the hospital, since that meant I could not go to the public pool. I didnâ€™t want to go swimming in our back yard, since the pool was so small I could stand up easily and it felt like I was in a big bathtub. I gave up on the idea of swimming, or doing anything fun, and resigned myself to a day of drawing ill-conceived faces and reading three-month-old People Magazine.  &nbsp;  Coming back from the hospital, we stopped at Burger King. We went through the drive-thru and ate the burgers in the car. We passed the airport, and were almost home. I suggested to my mom that I go for a bike ride. I needed to clear my head. I needed to get away. I needed to keep going and going and never stop. She agreed it was a good idea. We got into the driveway and she put up the garage. I grabbed my bike, rode it up to the car and gave her a kiss on the cheek as she walked to the mailbox to get the mail.  &nbsp;  I started to pedal hard. Pushing my muscles harder and faster, I careened forward not wanting to stop. I remember my right leg cramped from sudden activity after hours of sitting, but I didnâ€™t care. I kept pushing. As I rode, I cried.  &nbsp;  &nbsp;I got thirsty, and had no water bottle with me so I stopped at a diner owned by my former best friendsâ€™ family. Their father was glad to see me. He gave me a soda and let me sit for a few since I was red in the face and out of breathe. The fifties nostalgia diner was sea foam green and white, with splotches of pink and coca-cola signs everywhere. I drank down my coke, listening to the jukebox playing, â€œDanielâ€ by Elton John.  â€œDaniel is leaving tonight on a plane. I can see the red taillightsâ€¦.and I can see Daniel waving goodbye Lord it looks like Daniel. Must be the clouds in my eyes.â€ I sat there, my eyes still stinging with tears, and I let my father die.  &nbsp;  I closed my eyes and I saw the future. My fatherâ€™s funeral, the hundreds of people turning out to say what a brilliant sun of a bitch he was, my immediate family in pain, and my extended family squabbling over everything from the will to the way the priest conducted himself. My siblingsâ€™ weddings, occurring in my mind over the next five years, without my father there to make sure the functions happened according to his traditions. My mother in whiteâ€¦.lots and lots of white. White is such a hard color to clean.  &nbsp;  I understood then why Hindus wore white as the mourning color. With white, on dusty roads, you must be aware of yourself. You must always be conscious of what you are wearing. You must be careful every time you eat or drink. You never forget the person who died.  &nbsp;  I saw my high school graduation, my college graduation and every other milestone in my life. I saw the probability of moving to Texas to be closer to more extended family, or Washington DC for the same reason. I saw my mother working. I saw us losing the house I grew up in, or selling it to help my sister pay for graduate school. I saw me and my mother in a cramped apartment in downtown Buffalo. I saw all of this, and I decided it was ok. I resigned myself to the idea that my father was going to die. I gave up on him.  &nbsp;  I left the diner, after thanking Mr. Harrishburg for the soda, and headed out again. Once again pushing my legs hard, feeling the pain and pushing for more. I whipped past the houses of friends who had known me since the day I was born. I stopped and rang their doorbell, hoping to get a game of catch out of them, or at least a drink of water. No one was home.  &nbsp;  The tears stopped. I felt empty. I had accepted my fatherâ€™s death, and the inevitability that came with it. I turned my bike around and slowly started for home. I pulled into my driveway, and my mother pulled in behind me. She was honking the car horn like she was in a parade. I turned and watched her waving frantically for me. I ditched my bike on the front lawn, not even putting down the kickstand, and ran to the car. â€œWe got the heart! Weâ€™re going to Cleveland! Put your bike in the garage, and go get packed!â€  &nbsp;  I started to cry. I had thought no more tears could come, but out they came. My mother thought they were tears of joy, but they were self-pity. I felt guilty. I had given up. I had given up and there was still hope. I have felt guilty about that for the last twelve years.  &nbsp;  For twelve years I have made myself suffer for a lot of things in my life, but none more than that ten minutes. The ten minutes in which I allowed my father to die.  &nbsp;  Recently, I encountered head-on my unforgiving nature and how hard I am on my adolescent self. My friend invited me to hear a choral concert of Elton John songs. I sat there in the audience, and heard â€œDanielâ€ for the first time in twelve years. For the past twelve years, I avoided it. If it came on the radio, I changed the station. I never bought any Elton John albums and I drowned it out with conversation if it was played in restaurants.  &nbsp;  In a theater, I could not run away from it. I sat there, and allowed it to wash over me. I was fourteen again, and angry at the world. I was fourteen again, and angry with my father for being sick. I was fourteen again, and giving up. My heart seemed to stop as they sang it, the song that was the aural impression of my subconscious disassociation with my father. I sat there and cried, and wanted to run. I am not a rude person, so I suffered through it. I listened to the lyrics, and remembered. I allowed the emotions to coarse through me. I started to forgive the innocent fourteen year old. Which is why, for the first time in twelve years, I can write about this, without just recounting a narrative.  '),(16297,3970082,'19,July,2004','2004-07-19 00:00:00','   I saw the best minds of my generation consumed by madness. That line has been throbbing in my head for the past 48 hours. When does it become more than just a song in the head, since the lines from the poem are put to a pithy pop-rock beat? When does it truly become madness...madness as in the victorian age, with fainting couches and corsets and high tea? It all makes no sense, this ability to have a tune floating in our minds for hours at a time. And yet we listen to music.  '),(16298,3970082,'18,August,2004','2004-08-18 00:00:00','   Both these quotes were overheard on an ABC TV show. The show was about television shows that were only pilots and never aired. They were just too funny not to write down:  \"What could be more adorable than a boy and his troll?\"  \"Have you ever had a man stand behind you and kiss you on the back of your knees?\"  I know this is not a big post, but I will write more later. This week I have been celebrating my parents\' 39th wedding anniversary and my sister\'s 36th birthday. Plus we are moving Mehala into her apartment here in Miami. I\'m kinda tired.  I do promise more later. *SMILE*  '),(16299,3970082,'16,August,2004','2004-08-16 00:00:00','   I just lived through the second hurricane of my adult life. In the first one, I was living in DC. It was 1999, and I blacked out in my apartment, so I had to go to Georgetown Medical Center for observation. My boyfriend at the time wanted me to come out to his place to ride out the hurricane. We had a huge fight since my mother wanted me to go to my uncleâ€™s house in Maryland. I preferred staying with the boyfriend, but I did not want to worry my folks more than I already was worrying them. So off to the wilds of Northern Maryland I went. I donâ€™t remember much, except that when I returned to campus, there were tree branches everywhere.  This time around, the hurricane missed us. It went up the gulf coast, and all we got was one storm at one-thirty in the morning on Friday morning. It was an amazing sight to watch. I was completely awe-struck by the power of nature. Still, it was just a really harsh storm and not the actual hurricane.  I did not see the fall-out of the hurricane until yesterday. My father and I went to a Muslim wedding. It was my fatherâ€™s first Muslim wedding, and my second. The wedding was that of a cousin of mineâ€¦of sorts.  The true family connection is through the husbands of two of my aunts. These men are first cousins to the groomâ€™s father. The family raised their children Hindu, but the groom had become a Muslim.  On our way to the wedding, my father and I conversed about my life. We talked about relationships, my siblings and my sisterâ€™s impending arrival in South Florida. I had been expecting cautionary tales, warning me not to turn out like my siblings. I get those about once a month from him. Instead, thanks in part to a mix tape with  Shania Twain  and  10000 Maniacs  on it, we made it to the wedding with no tears, and a few laughs.  I had never been at a wedding before, where I was afraid to talk to the men. I guess afraid is the wrong word. I was uncomfortable speaking with them. I offered my hand to the groom when I first arrived, and he refused it because I was a woman. I had dealt with that before, but usually with men of Arab decent and they are a lot more gracious about it. They understand that in the US men and women shake hands upon meeting each other. This guy looked insulted that I had offered him my hand.  I spent the wedding talking to the brideâ€™s parents: a nice American couple from the Fort Myers area. They were having a hard time dealing with their daughterâ€™s newfound faith in Islam. Still they were there to support her and her decision to marry a man she had met on the Internet.  Most of Fort Myers had no electricity or running water, including the mosque where the wedding was taking place. Before the wedding, my dad and I went to  Publix  to get a soda. The supermarket was crowded with people, but we were warned as we came in that there was no frozen food, produce or refrigerated items. So we enjoyed a warm coke and drove to the mosque.  The Imam was two hours late. This was the first wedding I ever attended where the reception occurred before the wedding. Everyone was roasting in the heat, since the interior of the mosque was even hotter than outside. There I sat, in heavy emerald green silk with tons of gold embroidery, feeling like I was the wicked witch of the west. The phrase, â€œIâ€™m melting! Iâ€™m melting!â€ barely even covered the way I felt.  All around us were broken tree limbs and other debris. If I had not seen the intersections without signal lights, and the fact that even taco bell was closed, I would have thought that they had just experienced a bad thunderstorm. I saw some interesting angles for photographs, but realized the wedding party did not have a photographer. I used film on them instead.  Finally, the Imam showed up. He claimed he was late because he was stuck behind President Bushâ€™s motorcade. They held the ceremony outdoors, so we didnâ€™t die of heat stroke. If they had held it inside, I would have had to sit separately than my father. Since it was outdoors, we were allowed to sit together.  It was weird. In the other Muslim ceremony I witnessed, the bride opted to represent herself before God. In other words, she said her own vows. In this ceremony, the bride asked another gentleman from the mosque to recite her vows. So instead of the bride and groom holding hands, the groom clasped hands with this man, and the bride sat to the side looking a bit lost.  The ceremony itself was ten minutes. At the end of it, my father and I said our goodbyes and hopped into the car. I drove, and sped all the way to the other side of the state. On our way home, we stopped for barbecue. I ordered pork. I guess Iâ€™m never gonna be a good Muslim.  '),(16300,3970082,'13,August,2004','2004-08-13 00:00:00','   I have not written for several days. For the most part, I have been busy. Work has been happening on a more reasonable level, and I joined bookcrossing.com. Bookcrossing.com was going to be the subject of this blog, since it is my new passion du jour. I had the anecdotes all planned out in my head, and knew how the paragraphs were going to flow, and then this morning my father decided to take a car ride to the auto mechanic and turn it into a family history lesson.  He wanted to talk about my grandfather. For the most part, when I think the term grandfather, I mean only one man. My motherâ€™s father was never in my life. I never met him, and my motherâ€™s mother signed the lonely birthday cards I would receive every year. Still, I digress. I was writing this about my fatherâ€™s father.  In my family, everyone called my grandfather â€œAnna.â€ It is pronounced Un-Nah, and in Sanskrit means, â€œbig brother.â€ The irony did not get past me when I read 1984 by George Orwell. In so many ways, Anna was the â€œbig brother.â€ Every person who came into the house, and every phone call, had to be introduced to him, or summarized for him. He had to know everything.  He had been a government official in India. He was a highly respected man in South India, both in the British government and in the subsequent Indian government. Anna was respected and feared by people. He was forced into retirement at a relatively young age by circumstances beyond his control. He was only in his 50s.  For the first ten years of my life, he lived in our house in Buffalo New York. He moved in when I was three weeks old and did not leave for 10 years. He had the downstairs bedroom, which had originally been designed to be a den and laundry room. The room barely fit the furniture we squeezed in it: a bed, bookshelves, and easy chair were pretty much it. We recently sold the house, and walking into that room with it being empty, made it feel so much smaller than it felt when it was full. Partly I think it was full because my grandfather had such a profound effect on my life.  Most children have two parents, I had five. I really do think, I was parented in their own way, by both my siblings and my grandfather. My siblings, bossed me around like kids do, and because they did, I became an experiment in pavlovian conditioning. They didn\'t need a pet, they had me.  My grandfather bossed me around as well. We also fought, more like we were siblings than grandfather and grandchild. I remember screaming matches over the television. I wanted to watch  Scooby-Doo , and he wanted to watch  Hawaii Five-O.  He always won those arguments, but I always put up a good fight.  I remember a broken man, who really wanted to be independent, but knew he was claimed on his son\'s tax return. It hurt him a lot. He used secrets and intrigue within our family, to get some kind of power. His powermongering hurt our family in a lot of ways...ways that are still being expressed eleven years after his death.  This year, for the first time my father did not join his siblings in honoring the anniversary of his father\'s death. He did not even tell me it was the anniversary. My mother told me he fasted that day, but there was no pomp and circumstance. Its funny, considering the ceremony that the family has always had to honor him. Part of me wonders why...why we follow the rituals and why...after a while...we let them fall to the wayside and just continue to live our life.  '),(16301,3970082,'09,August,2004','2004-08-09 00:00:00','   Sometimes it is not so bad, living in the spare bedroom of the people who created me. After all, I am not paying for rent, food, utilities or any of the other things that people my age struggle to pay for month after month. Yes, I am trying to see the bright side of it all. After months of job hunting, I am still no closer to a full time position than I was in November 2003 when I was sequestered here. I try not to complain, or to feel depressed, but it is as bad as living in the parents\' basement...except this time I am trapped on the 24th floor of a highrise condo. For those of you who do not like or know me well, I don\'t like heights like that. I don\'t freak out, but I also don\'t go out on the balcony, at all...unless circumstances beyond my control force me.  Lately I have been forced out onto the balcony more and more, by life and situations involving the parental units. It is easier to go there than to run screaming from the house at 2 in the morning when my mother is online plugging away at her own blog, and all I want to do is check my e-mail for five minutes. My mother stays up later than I do, and my father wakes up early. So, basically there is no time alone, unless they are both going someplace together, or at the same time.  So I spend tons of time locked in my room, with the television on in the background, reading, or doing one of my many hobbies. Ok so this piece says nothing more than what is going on each and everyday in my life...but sometimes that is for the best.  '),(16302,3970082,'06,August,2004','2004-08-06 00:00:00','   I keep discovering strange things growing in my backseat. I guess I should back up. In the back seat of my car, I keep one half of a bicycle (the quick release wheels and seat), a bicycle rack, various computer components for work and personal use, books from my own collection and the public library, clothing for various stages of Miami weather and three or four pairs of shoes. I also keep back there papers that I need, most of which are not actually important, but I feel should not yet be shredded, various tools, for use in the car, my jewelry making tools, and generic brand breakfast bars in both strawberry and mixed berry flavors.  Even with all that stuff in the backseat of my car, it must have taken a bolt of lightning hitting it or something, because last night I discovered my library and personal books had reproduced. My silly romance novels, and books by Arthur Schlesinger and Jennifer Hochschild have had a secret liaison and produced a love child unlike any I had ever seen. It is definitely something out of Mary Shellyâ€™s  Frankenstein . I discovered in my backseat a book entitled  Thieves in High Places  by Jim Hightower.  I know I did not purchase this hardcover concoction, and it has not remnants from the library, so my only course of deduction is it was created from sexual or asexual reproduction. Since I highly doubt  The Ecology of Commerce: A Declaration of Sustainability  mutated during mitosis to create some new tuber, and such a different book from its parent. I see no choice but to believe that my books are practicing non-safe sex.  I really must talk to them about this behavior, because bastard love children take up a lot of space, and like all books, make me want to read them. I barely have time for the oodles of books I have on the public libraryâ€™s hold list. Still, I took the book in my arms and tried to read it during dinner last night. I could barely get passed the first chapter. It felt as poorly written as right-wing rhetoric, yet it was from the left.  And so I will conclude this piece of backseat melodrama. I need to go out to my car, see if there are any radioactive spiders inside of it, and have a serious conversation with my books. They need to learn that unsafe sex has consequences, since I donâ€™t want to get a call from the public library complaining that their books are diseased.  '),(16303,3970082,'05,August,2004','2004-08-05 00:00:00','   Tuesday morning I crawled out of bed about noon and was confronted with yet another test of my strength. My mother announced to me that she thinks she had a heart attack two days before. She had called the doctor, but could not get an appointment until Friday morning. It is now Thursday afternoon, and I have been feeling like an elephant has been kickboxing my chest ever since. Memories of the times I spent sitting in emergency rooms and waiting rooms of ICUs came flooding back to me. I have a big ache in the pit of my stomach. I know no matter what happens, it will be ok. I just want to know something now, not tomorrow and not next week.  Oh well, such is the life I lead. I\'m at the library, where I usually am when I decide to pound out a rant. I just tried to access nerve.com. I was not allowed access due to content. That of course makes me curious what kind of content is on it, that I am not allowed to view? I guess the sex survey done by the unemployed for this week, may be it...but is it really that different than the stuff on slate.com? Where are they drawing the line? I love lines....they lead to slippery slopes. Slippery slopes are fun...because you never know how far down people and morality will fall.  My hands hurt, from hours of typing, and my eyes ache from staring at the screen. Why do we subject ourselves to spending hours in front of a keyboard, only to communicate with people who are no where near us? Why do we race home from work, or school to check e-mail from people we see everyday? Why do we seek this validation?  If anyone has any answers to my questions...please let me know. I am sick of trying to answer them for myself.  '),(16304,3632466,'28,June,2004','2004-06-28 00:00:00','   friday night with the a-s-s crew.  wandered out to slurrey to drive through scary neighbourhoods and quench ai\'s thirst.  of course the night could not have ended without a visit to \"homecourt\" at kensington (is that where that is?).  before arriving, some drunk vina kid cuts his thumb on a empty beer can, he ends up getting lined up the most.  i can now keep the ball at an acceptable height level on the free kicks.  saturday morning poker tournament in ditchmond.  the tourney started at noon, so i figured that would give me enough time to win it all and make it to erica\'s birthday dinner at skybar at 7pm.  4:45pm, guess who\'s the chip leader going into the final table, plenty of time to make it to dinner.  get knocked out in 6th place, two spots shy of making any money.  check the watch, 7:10pm!  ai calls, \"where are you, i\'m here already!!!\"  gggggggggg gun-it home, shower, and i\'m at skybar at 8:15pm.  little boozin\', little schmoozin\'.  got drunk bien\'s shirt-off without waking him up.     now it\'s time to go home.  drive ai and her car home.  drive all the way back to downtown to pick up my car.  yesterday, watched the thoccer game and some nascar action, then had an impromptu bbq at ryan&joy\'s place with sof and ai.  good food, good company.  no weak end. '),(16305,3632466,'23,June,2004','2004-06-23 00:00:00','   why do i play when i\'m just gonna get lined up?  probably \'cuz i\'m a sucker for competition, and i\'m not one to back down from any challenge.  although i\'m usually lined up more than anyone else, i live for when it isn\'t me up against the fence.  seeing pain and anguish in my friends\' faces is priceless.  i know i could easily just punch them in the face or something, but where\'s the fun in that? '),(16306,3632466,'22,June,2004','2004-06-22 00:00:00','   i have to check the  urlLink results  every 30 minutes \'cuz i still can\'t believe it...  daca, i hate agreeing with you, but you\'re right.  just imagine what this accomplishment means to some of us that have been doing this now for sooo long!  this isn\'t just an achievement for all the hard work this years team has put in, this is for all the other ruckers that have paddled somewhere along the way with us.  i wish there was a way that i could thank everyone that ever paddled for ruckus without coming across as an ass that\'s just trying to rub it in. '),(16307,3632466,'21,June,2004','2004-06-21 00:00:00','   silver medal in the rec A championships!!!     i still can\'t believe it. '),(16308,3632466,'19,June,2004','2004-06-19 00:00:00','   guess who\'s going???  that\'s right, mother ruckus. '),(16309,3632466,'18,June,2004','2004-06-18 00:00:00','   so i\'m guessing a fat steak wouldn\'t be good for the team dinner, huh?  i was asked by kaz and ai to help think of encouraging remarks for every individual on the team, a way of thanking everyone for their own specific contributions.  i started off thinking about the members that actually contributed something and wrote uplifting remarks for them.  but then i got to indivduals that didn\'t really contribute anything to the team.  in fact, for some members i questioned why they were even on the team!  instead of positive remarks, i was serving straight disses.  which leads me to this... am i really that big of an asshole?  i\'ve always considered myself to be an ass, but damn, sometimes i can be straight-up mean.  but isn\'t it justifiable?  i\'d never be an ass to someone who didn\'t deserve it.  discuss. '),(16310,3632466,'17,June,2004','2004-06-17 00:00:00','   ballin\' at the cage, followed by some glutony in the form of richmond sushi.  that is all i will write for today. '),(16311,3632466,'16,June,2004','2004-06-16 00:00:00','   i can\'t tell if i have a cold or my allergies are acting up.  so i\'m od\'ing on both claritin and dimetapp. '),(16312,3632466,'15,June,2004','2004-06-15 00:00:00','   do you ever get the feeling that \"big brother\" is watching you at work?  every now and then i do, so i test it out just to make sure he isn\'t.  some of the tests i\'ve pulled in the past are the  sleeping at my desk  test, or the  sleeping under my desk  test.  both proved that on those particular days, \"big brother\" didn\'t have a clue!  there\'s also the  go for a quick rip up to squamish  test (only when jeanne drove the m3 or the 911), or the  shopping at park royal  test.  today i\'m gonna do a tried, tested and true version, the  surf-all-day  test, but to spice it up a little i\'m adding the  feet on the desk  test.  my job\'s okay. '),(16313,3632466,'14,June,2004','2004-06-14 00:00:00','    '),(16314,3632466,'14,June,2004','2004-06-14 00:00:00','   ever drink so much you don\'t remember certain time frames?  i remember all the crown shots.  i remember telling kaz i\'d find another ride home.  i remember opening the door at the burger king drive-thru.  then i remember kaz calling and waking me up to go play golf.  friday night was fun. '),(16315,3632466,'26,July,2004','2004-07-26 00:00:00','   i was wandering the web earlier and came across some done up vans in ha-pan.&nbsp; the vans are owned by fans of hamasaki ayumi (some singer bitch).&nbsp; click  urlLink here  to see them...   if you make your way down the page, the guy who posted the pics does a quick write-up about the singer...     &nbsp;    ahhhhhhhhahahahaha!!! and this...     &nbsp;  jap\'s amuse me. '),(16316,3632466,'25,July,2004','2004-07-25 00:00:00','   &nbsp; '),(16317,3632466,'23,July,2004','2004-07-23 00:00:00','   don\'t you love it when a joke works perfectly?&nbsp; and then, you\'re still able to carry it on for a couple of days???&nbsp; that\'s the best!!!   jeff free, stephen nguyen, clark gable, barry bunds, longranger,&nbsp;kristie c., sakura77, ginzaboy, mattrix, touchyfeely, softee, i am gaynadian., bryson, garreth, brent, kazimodo... what do they all have in common?&nbsp; all created by yours truly.   i\'m the best.   taken from  urlLink www.dacango.blogspot.com ...    &nbsp;&nbsp; '),(16318,3632466,'22,July,2004','2004-07-22 00:00:00','   i can now lift my arms way up high, but my arms still take time to bend at the elbow.&nbsp; so the shoulders and the chest are good, but the arms still burn. '),(16319,3632466,'20,July,2004','2004-07-20 00:00:00','   mi-mommy\'s funny, she asks me for my permission to play mj with her friends.  &nbsp; when i say no, she says, \"please, \'Co?\" &nbsp; keekeekee '),(16320,3632466,'19,July,2004','2004-07-19 00:00:00','   you\'ll be back to normal IN no time VAL!&nbsp; take cared kID.  &nbsp;     &nbsp;   &nbsp; '),(16321,3632466,'18,July,2004','2004-07-18 00:00:00','   what better way to spend a saturday night than a trip to night market...  &nbsp;  first of all, i\'m really trying not to be racist (seriously), but why pick an area where there is no parking except tight-ass stalls and street parallel parking?&nbsp; now we all know what the dominant race at night market is, so you think the organizers could find a spot with a giant parking lot where the bad driving wouldn\'t really matter.&nbsp; the car circus is easily seen as soon as you get off the knight street bridge.&nbsp; although it can be hella funny at times, it does get pretty stressful being in the middle of real-life bumper cars.  &nbsp;  once you\'ve found a parking spot (5km\'s away), you have to battle through a sea of people.&nbsp; as soon as we got in the market, it was like being on the shinjuku line at rush hour.&nbsp; that\'s right, private part touching included!&nbsp; no joke ask eugene, when we were leaving, i turned around to say something to euge and this little maybe 4yr old kid sticks his hand out as he walked passed me and touched my thing!&nbsp; it was so random!!!  &nbsp;  it never ceases to amaze me how people can just sell loads and loads of absolute crap.&nbsp; who thinks, \"damn, i should buy all this crap (in asia) and make money selling it at night market!\"???&nbsp; if the prices for all this shit&nbsp;are so cheap, does that mean they give this stuff away where it comes from?&nbsp; 100cd spindle for 30 bucks?&nbsp; 6 pairs of socks for 10 bucks?&nbsp; cartoon print flip-flops for 4 bucks?&nbsp; how much does this stuff actually cost?  &nbsp;  one of the&nbsp;never fail&nbsp;pluses for night market is the food section.&nbsp; where else can you get such a variety of asian \"tapas\" at such bargain basement prices?&nbsp; (shout out to the flips selling bbq and lumpia)  &nbsp;  but no matter how much trash i talk about the zoo that is night market, there are times when you just happen to find a \"diamond in the rough.\"&nbsp; 2 bones for my newest toy!!!  &nbsp;     &nbsp;  this electric fly swatter is going to be useful at home and at bbq\'s exploding unwanted bugs, but most importantly, it\'ll be the provider of hours and hours of drunken fun (in the form of electrocution).&nbsp; 2AA batteries give a pretty good jolt.  &nbsp;  time to sleep. &nbsp; '),(16322,3632466,'16,July,2004','2004-07-16 00:00:00','   this story is about&nbsp;some b at work that pissed off mandeep and i.&nbsp; we like to \"take it to the extreme\" at work when it comes to cracking jokes and making fun of each other.&nbsp; anyways, this b went overboard and sent an e-mail where man-slip and i were dissing some chick, to the chick that we were making fun of.&nbsp; the girl we were making fun of was upset, but later understood that we were just joking \'cuz she really likes&nbsp;man-creet and i&nbsp;(no joke, she wants to have sex with man-dip and i, probably at the same time too!&nbsp; but that\'s another days post).&nbsp; so man-tid and i confronted the b about her forwarding the e-mail and told her she was&nbsp;stupid, she lashed back saying that that what we write is not her responsibility or some shit (she talks so much smack about the girl we were making fun of too, just not on e-mails like man-sleep and i).&nbsp; i had to hold man-jeet back \'cuz he wanted to slap her.&nbsp; anyways, man-preet and i decided to just avoid&nbsp;b altogether from now on.  &nbsp;  so&nbsp;man-creep and i&nbsp;haven\'t talked to this b in 2 months, and it\'s pretty tough to do considering we\'re basically on a team of 6 people.&nbsp; i get to work this morning and look who\'s sitting in man-peed\'s cubicle crying... crazy b.&nbsp; she starts apologizing madly and saying&nbsp;how she can\'t stand working here anymore without man-sleep and i even acknowledging her existence.&nbsp; she showed&nbsp;us some e-mail she printed out apologizing to the girl we were making fun of in the first place.&nbsp; we told her that it didn\'t have to be like this but she brought it upon herself.&nbsp; we said we\'d be her friends again.&nbsp; she got up and went straight to the bathroom to wash her face or something.&nbsp; and man-jinder and i laughed so hard i almost passed out.  &nbsp;  yeah, i\'m an asshole at work too. '),(16323,3632466,'15,July,2004','2004-07-15 00:00:00','   can you put a value on your life?   would you play  any  game to have it back?   would you die trying?   can you see what i  urlLink saw ? '),(16324,3632466,'14,July,2004','2004-07-14 00:00:00','   i was just at the quay and it\'s boiling in there!  if my office wasn\'t a refrigerator, i\'d probably pass out, that\'s how i figure i\'ve fallen asleep the past couple nights at home, from heat stroke.  the worst is like last night when you\'re sitting around (at bien&jean\'s), not in any direct sunlight (or any light for that matter), and you\'re not doing anything, ANYTHING, and you\'re sweating like a dog!!!  feels so gross.  i wonder if there\'s a pill you can take to cool your core body temperature, that would be the shit!  where can i buy that mist contraption?  i\'ll probably need it tonight at daca\'s (fuckin\' sauna).  heat sucks, i wish it would snow tonight. '),(16325,3632466,'13,July,2004','2004-07-13 00:00:00','   there\'s nothing better than being able to accomplish multiple goals in a single time frame.  quinton \"rampage\" jackson manages to score a gazillion points, knockout ricardo arona with a dibilitating power bomb, and incurs a gash just above his right eye that easily sows up with 40 stitches.  now that\'s efficiency!!!     support the black man. '),(16326,3632466,'12,July,2004','2004-07-12 00:00:00','   i can\'t!  i can\'t! '),(16327,3632466,'09,July,2004','2004-07-09 00:00:00','   i guess that could be considered a problem... if you ever rub me the wrong way, you\'re screwed.  some of you might know what i\'m talking about, it only takes one incident and i\'m never gonna forget.  i could give a few examples, but some of you might read this and not want to be my friend anymore.  ...but then again, i probably wouldn\'t really give a shit! '),(16328,3632466,'09,July,2004','2004-07-09 00:00:00','   wake up sweetie, it\'s your birthday!!! '),(16329,3632466,'08,July,2004','2004-07-08 00:00:00','   my face has started to peel.  i figure it\'s just a matter of time before my chest, sholders, and back will follow suit.  i wonder how all the chafing from my golf bag tomorrow will aid in that process.  my skin feels like dry leather. '),(16330,3632466,'06,July,2004','2004-07-06 00:00:00','   the perfect seat for a birthday boy, an old toilet sitting in an alley.     although the long (super long) weekend ended with the need to cover my body in aloe every 2 hours, it was fun nonetheless.  playing hookie from work, poker @ lex\'s, golf on canada day, bbq @ pat\'s, playing with kyla, shopping with jeanne, lounging @ sof\'s, kitchening with kent, bbq @ eva\'s, splashdown with the triple A crewette for charlene\'s bday, bbq @ john\'s, all capped off with an alienz game of a-s-s \'til 1am.  good clean fun!!!    this, on the other hand, is straight up dirty chug styles.  (he even put the lid up!!!) '),(16331,3632466,'05,July,2004','2004-07-05 00:00:00','   ever spend 8 hours in the sun at splashdown without having any sunblock touch your skin?  i didn\'t know my skin could look as red as it currently does.  last night we were out on the deck and my legs and arms were feeling chilly, but my chest, shoulders and back were on fire!!! '),(16332,3632466,'02,July,2004','2004-07-02 00:00:00','   guess who has 2 teeth that just popped out...     hi kyla!!! '),(16333,3632466,'12,August,2004','2004-08-12 00:00:00','   today.  hien and i go walk and buy.  sorry i no want sari!  i buy hien cup, she drink coffee in cup.  what up bee-otch? '),(16334,3632466,'11,August,2004','2004-08-11 00:00:00','   i think i\'ll post as if i\'m the japanese guy that didn\'t like the singer cuz of her \"small boob\".  hi, am in toronto.  there is a lots of black, not like vancouver no black.  everywhere you take bus is jetsgo.  jetsgo on jane street.  the black on the subway sit quietly but i know they have gun.  don\'t shoot me, i can not run, hurt legs from \"what the???\" downhill-uphill  swim hike.  it\'s hot it\'s cold it\'s rain it\'s wind, toronto weather is fishy but not smell like fecal chloroform in creek false.  oh, and hien like a bug.  i take shots and put in soon. '),(16335,3929371,'22,July,2004','2004-07-22 00:00:00','   Friends, Romans, countrymen, lend me your ears;  I come to bury Caesar, not to praise him.  The evil that men do lives after them;  The good is oft interred with their bones;  So let it be with Caesar. The noble Brutus  Hath told you Caesar was ambitious:  If it were so, it was a grievous fault,  And grievously hath Caesar answer\'d it.  Here, under leave of Brutus and the rest -  For Brutus is an honourable man;  So are they all, all honourable men -  Come I to speak in Caesar\'s funeral.  He was my friend, faithful and just to me:  But Brutus says he was ambitious;  And Brutus is an honourable man.  He hath brought many captives home to Rome  Whose ransoms did the general coffers fill:  Did this in Caesar seem ambitious?  When that the poor have cried, Caesar hath wept:  Ambition should be made of sterner stuff:  Yet Brutus says he was ambitious;  And Brutus is an honourable man.  You all did see that on the Lupercal  I thrice presented him a kingly crown,  Which he did thrice refuse: was this ambition?  Yet Brutus says he was ambitious;  And, sure, he is an honourable man.  I speak not to disprove what Brutus spoke,  But here I am to speak what I do know.  You all did love him once, not without cause:  What cause withholds you then, to mourn for him?  O judgment! thou art fled to brutish beasts,  And men have lost their reason. Bear with me;  My heart is in the coffin there with Caesar,  And I must pause till it come back to me.   '),(16336,3929371,'05,August,2004','2004-08-05 00:00:00','   '),(16337,3929371,'02,August,2004','2004-08-02 00:00:00','   Ok. Maybe qualifying something from Shakespeare as an official blog is way too much to ask. But isn\'t it a nice o\'l speech, eh? And a fitting one, at that, for my hopefully rewarding dip into the world of non-private thinking.  A great orator such as Marc Antony tends to make his audience believe that what he is saying is TRUE, regardless of the righteousness of the idea. That \'Friends, Romans, countrymen\' speech was immortalized by that very nature -- its  use of rhetoric to affect emotions and spur action in the process.  By posting that speech, I had just imposed upon myself something of a precedent for how things should work around here. I\'m not saying I can persuade anyone into action, or lure someone into a pipe dream by my mere soliluquy. All I\'m saying is perhaps, my rants, raves, and eternal damnation of the organized majority can be viewed in such light as PURPOSEFUL. Non-censorship of web contents guarantees me freedom for cussing and screwing anything and anyone, but by golly -- be it for some intent other than satisfying myself.  There\'d be some sporadic trivial and inconsequential stuff here too, sure, but hey it isn\'t fun talking about serious matters all the time.  As the great Oscar Wilde oh-so-aptly puts it -- \"Seriousness is the only refuge of the shallow.\"  So, there.  By the way, it took me a while to understand what was behind Marc Antony\'s speech, so if I were you, I\'d go back on that post and read it again. It\'s beauty and classic appeal had just dawned upon me on my 10th, no make it 15th, intentful re-reading.   '),(16338,3642942,'17,June,2004','2004-06-17 00:00:00','   Que difÃ­cil se me hace  mantenerme en este viaje  sin saber a donde voy  en realidad.  Si es de ida o de vuelta,  si el furgÃ³n es la primera,  si volver es una forma  de llegar...   How difficult it is to advance not really knowing where I\'m going. if I\'m going or returning, if I\'m on first class or last class, if returning is a way to arrive.  Que difÃ­cil se me hace  cargar todo este equipaje,  se hace dura la subida  al caminar.  Esta realidad tirana  que se rÃ­e a carcajadas  porque espera que me canse  de buscar...  How difficult it is to me to keep on draggin all this bagagge, up the hill. This tyrant reality that laughing at me, wanting me to  stop looking...  Cada nota, cada idea,  cada paso en mi carrera  y la estrofa de mi ultima cancion.  Cada fecha postergada  la salida y la llegada  y el oxigeno de mi respiracion.  Y todo a pulmon,  todo a pulmon.   Each note, each idea, each step in my carreer and the lyrics of my last song. Each posponed date Leaving or arriving and the oxygen in my breath. With all my breath. All my breath.  Que difÃ­cil se me hace  mantenerme con coraje  lejos de la tranza y la prostitucion.  Defender mi ideologÃ­a  buena o mala pero mÃ­a,  tan humana como la contradicciÃ³n.   How difficult it is to have courage away from selling out and  prostitution. To defend my ideology good or bad but mine, so human like contradiction.  Que difÃ­cil se me hace  seguir pagando el peaje  de esta ruta de locura y ambiciÃ³n.  Un amigo en la carrera  una luz, una escalera  y la fuerza de hacer todo a pulmÃ³n.  Cada nota, cada idea...   How difficult it is to keep on paying tolls in this road of crazines and ambition. A friend in the race a light, a ladder and the strength to do everything with all my breath.  '),(16339,3642942,'15,June,2004','2004-06-15 00:00:00','   Over the past few weeks, ever since we came back from our honeymoon, \"how is married life so far?\" seems to be the question I get asked the most.  \"So far so good.\" comes naturally as my standard answer. Yet a bit of reflection makes me understand that this is not just a simple clichÃ© comical commentary on the current situation of marriages in the country. This answer has slowly been brewed over years by watching other couples get together, split or sometimes live happily together. The cruel reality of our time.  Yet the cold uncertainty suggested by such an answer is a constant reminder that, in the words of uncle Cesar, relationships work only as good as the work we put on them. Our marriage is â€œso far so goodâ€ not because an inevitable malady has not yet arrived but because we have worked hard, so far, to make it be â€œgoodâ€.  As we move into our second month of marriage all I can hope is that the word â€weâ€, will keep on creeping into my writing so that the \"so far so good\" answer keep on comming, naturally, making people smile as the years go by.  '),(16340,3642942,'08,July,2004','2004-07-08 00:00:00','   Isn\'t it funny? Just when you have figured out how to be a toddler you become a kid, when you know how to be a kid you become a teen and when you have the teen-thing perfected you become a young adult. Of course once you have managed to learn how to be single you marry. But the most disturbing thing is that just when you have managed to love your job you find out you need to move on.  '),(16341,4163734,'22,August,2004','2004-08-22 00:00:00','   Today my eyes pop out leh...OUCH so pain...all my sis(.audrey.) fault hhorrrr? haiz...jkjk  when i am all alone at home for the whole day so most of the time so i will hu si luan xiang(anihow think), haha...i will think of the past where my cousins(only 4 of us at that time) will gathered at my grandma house. when i was in pri 1(i very cute de :>), my parents work until very late so for the whole week, i stayed in my grandma house...sometimes with my sis. after sch...they would wait for me and we would go to 7-11 and buy lots of ice cream and stuff then we will share all the chocolate to see which is the best...haha...then the whole family(i mean everyone lor) will be eating together at least once a week. :p  years past...my grandma moved to JP there after she injured herself and for a long period of time the whole family did not eat together...now is the worst...we nv even meet once a month...my grandma has 9 children(they have thier own family liao) and 9 grandchilden, but it seems like all of us have been too busy for a dinner...we meet only during new year and christmas and when there are special occasions...haiz...but it is worserererer on my father side, we really dun meet at all...coz they live too far and my mum and me are just too lazy and we never make an effort to keep in touch with them...and i am really quiet at their house...like nothing to talk about... :c     but there are happy moments too...but mostly when i am with my 4 cousins coz we are around the same age so we get along real well...heehee...though we have different father(of coz) our surnames are LIM blah blah...that is why we are soooooo close...not now coz one is in the army, the other in collage 2nd year and another in the sport sch so there is little chance to be together but deep down in our heart, we care for one another and when we are together, we talk abt everything under the sun...moreover my mother very curious so she seems to know everything that happened...then when it comes to results, die la...me is like the worst except for chinese coz they all speak english...lolx :] i have our(me and my cousins together) photo with me all the time coz i realli treasure all the precious moments we had together! not forgetting my other 5 adorable cousins...they are still young and blur :o ...haha.   P.S: to my darling(.B.) lolx jkjk...thank for ur accompany when i am alone...though u can\'t talk :(...or else we can talk about everything hor? but i realli learn alot from u...she very \'clever\' de worx always, aniway nvm...and we oso had great time together, sorry i now can\'t go out too often coz got exams...next time when i am free, then we go out and play basketball, k? promise u...haha  Love all of ya~ do take care always! O_o  '),(16342,4163734,'21,August,2004','2004-08-21 00:00:00','   wa lau...today a math so difficult...then 1hour only how to finish? :c die liao....haha go to sch 4 1hour only...teacher very wu liao leh...haiz...:(  ------>. suddenly...i tot of a friend...that time was when i P6 then i went to malaysia with a group of ppl from church...then i met a girl and we became the best friend for 3 days. on the last day, i avoid her. for that day, we didn\'t talk and she looked very sad. i dun noe why i avoid her but i regret now...really, i am sorry.  hope GOD will give her friend...a freind that will be better than me and will treasure their friendship, take good care of her and be nice to her...sorry again...please forgive me and if i have a chance to see u again...please let me be ur friend again...hope to see u one day...take care and eat more(i remember u only eat rice and soy sauce at malaysia...)though u may not see this...sorry...hope u wun blame me...  '),(16343,4163734,'20,August,2004','2004-08-20 00:00:00','   let me see...  MuMMy: =miss HaizA= =EsTher= =meLisA= =wAn tinG= =sHi hUi= =yIn fOOng= DaDDy: {Han qiNg} {adeLinE (cum naNny)}  LaO gOng: *pOlly* *HarFizAh* *shI Yi*       lAo Po: +WeI qUin(cum penguin)+ +pOh LinG+ +bErniCe+  siStEr: ~auDrey~ ~Jolyne~ ~iVy~ ~cHai tiNg~ ~Lydia~ ~deNiSe~ ~qIu Yu~ ~SeRene~ ~chU er~ ~coNNie~ ~Pei sHi~ ~yUan XuaN~ brother: yEE hoAu . Jun minG . cHe Hao . Jia yI . jEssiCa (she is a girl la...)  dAUghTEtr: -ShErLyn- -aZiZaH- sOn: _hUda_(she oso girl hor)  not forgetting my 2 wonderful grandmother... [cAren] [sHeila] # opps...those who i forgot to write down...sorry lolx #  '),(16344,4163734,'19,August,2004','2004-08-19 00:00:00','   today i hold my lao po(.poh ling.) de hand and purposely let my sis(.audrey.) see...then she so jealous worx...haha...very funny...then must say sorry to my lao ma(.esther.) coz i nv pei her go home and forgot to pass the letter to her..pai seh...dun be angry le la abt both me and her...cool down...first time go home with my sis(.lydia.) today...haha  didn\'t know my sis(.audrey.) got such good taste...know how to watch the matching making show...the show really very de nice...lolx one day i gonna kill my lao ma(.esther.) for spreading rumour abt me and my father-in-law...wa lau...tml got a-math test...  -breaKing The hAbit- by LINKIN PARK (very nice leh)  '),(16345,4163734,'19,August,2004','2004-08-19 00:00:00','   haha the matching drama very nice and super funny worx...that show make up my 45mins of laughters...lolx :p   lesson learnt today... -people are selfish- -adults alway lie-  '),(16346,4163734,'16,August,2004','2004-08-16 00:00:00','   I pass my A math leh...wat a surprise xia...haha maybe teacher mark wrongly coz i really thought i would fail somemore when teacher collect paper that time i still check answer with my lao po(.poh ling.) then we got different answers...lucky i pass...phew...  today melisa gave me a sweet..haha so sweet and nice of her...lolx unlike someone...saw rachel after sch...she change hairstyle liao...haha esther went mad and piniched her cheek...so ke lian...haiz...nitz   =chemistry suckz=  '),(16347,4163734,'13,August,2004','2004-08-13 00:00:00','   Amongst all my sisters and brothers hor, she is the most nasty one and she is a big bully. Knew her through my lao po(.poh ling.)-dun jealous :p haha ok la...i admit i copy and paste de...aniway whose no. ar? 9008992 u anyhow write one huh?    firstly hor...i know u meh?(this is her favourite line) n hor i NOT les lor...dun get me into trouble...later everyone think i les then i die liao...i will pull u down into the water(read in chi) haiz...dun noe how come i know u...the most bhb and lame sis hor? =) lolx...jk la...this sis very gong one and very nice to bully...haha but she hit ppl...actually only me coz she biased de...very pain lor(i regret letting esther know her, now they 2 gang up and bully me then my sis here still can act victim worx...) then she oso like to act cute and act young...haha though she is a lao ah ma liao...   WARNING: [this paragraph is to praise her or else later she will kill me de so...no choice. =>] actually she not bad de la...amongst my sis, she is the smartest and the most friendly one...though sometimes she nv do homework and forgot to brings things and always bully me and always suan wo and...ENOUGH! but we joke around only la...so all this for fun, laughters and peace so dun believe wat she write...k?  to say the truth...i am glad to have u as my friend..haha though u always bully me and suan wo...sobsob but without all these my life would be so boring then there will not be much fun hor? heehee i am nice so i will forgive u de...nitz  '),(16348,4163734,'11,August,2004','2004-08-11 00:00:00','   haha yesterday saw penguin during recess...then i tickle her but she nv scold me then she laugh...coz she got A2 for chinese...glad for her coz her oral was like...erm...haha not so good la. then i went to find sec 4 de...polly got A1 for chinese and her oral not merit but DISTINCTION leh...sure or not? somemore she said english during oral lor...wa lau, the examiner must be see her cute then forgot to deduct her marks...haha but so happy they got good result. well done worx. i forgot to ask jolyne and the rest(hope they did well too)...oops...nvm then hor i lent my sis the math set then she BEAT me worx...wang en fu yi...haiz... :p   yesterday first time sit with XL in class during chinese EXTRA lesson...she ar really very funny lor...though she always act fierce but she really very scared to offend ppl...haha yoz...me going to kaakyakingg (dun noe how to spell...) on september...is red cross and scout together so esther must be very de happy...hor?   today...not so fun coz i fail chemistry again...haiz i think i wanna drop it liao...someone got B3 for chinese, u can do it de so dun be sad hor...jia you!!! on sunday is my mummy(.miss haiza.) b\'dae...haha she ar...coz i nv underline properly female or male in the consent form then she asked if i am boy or girl LOUDLY...nvm suan liao...haha  '),(16349,4163734,'11,August,2004','2004-08-11 00:00:00','   hmmm..let me see...wa lau...the outcome of not studying is FAIL..haha geog must remember so many thingy de(my brain small)...then the test gave questions thet must think out of the box so....suan liao...zhi yu say that we common test must get about 34 to pass...funny xia...so i must add a lot a lot of oil!  today got assembly, when we sat at the corner then the first thing i saw was penguin (.wei quin.) then she purposely give me that kind of \"so unlucky\" face -.-\'\'\'...wa lau she good...then miss yee talk abt our sec 3 camp...must pay $65??? crazy ar? national camp only $30 then we still go pulau ubin lor and got so many thingy...the sec 3 camp is like dun know go where...erm haha  after school, my sis(.joyne.) drink bubble tea in sch then nv buy one 4 me de...somemore set bad example 4 me...haiz but i will forgive her de coz she shared it with me =)...haha then the other sis(.audrey.) SMOKE leh...haiz...i must tell her to quit smoking, it is bad 4 ur health lor...jk la...that is another person with similar name...phew...got bullied :x by someone(.chai ting.) worx...wa lau...she ar...haiz nothing to say liao...ohhh still got my lao gong(.polly.)...she fever, u better rest more hor...then so ke lian pester by esther k.s.m-i miss esther so much...during the holidays, no one screming tsunnammi(how 2 spell ar?) and all that stuff...oops then hor denise oso not feeling well, must take care...  today so lucky leh got a chance to eat with my father...haha but we still dun talk so much bcoz either he is at work or i am at home so very little time together...becoz i dun have the camp thingy so he showed more concern but i think he is the last person to know that my camp thingy is cancelled...haha me go watch tv liao...the show is super nice and funny... :p        -accidentally in love-  '),(16350,4163734,'09,August,2004','2004-08-09 00:00:00','   HAPPY 39th BIRTHDAY SiNGAPORE!  our pledge:  I, the citizen of singapore, pledge myself to strike regardless of TOTO or 4D, based on quick pick and system 7 so as to achieve car, cash and condo.  haha...lame la~ today is the first time i sang the national anthem and said the pledge so loudly in my secondary school life..haha coz in sch we like sleep through it de lor...then in primary school we sang loudly then we got scolded for shouting...haiya the teachers all so troublesome de...today, the NDP was great! the fireworks were very beautiful and the preformance was splendid! the kallang wave oso very fun! though my ears hurt when we sang songs coz ppl were sccreeaming...i drank too much coffee so like not tired like that but some of the performances were like the SYF...heehee   at nitz, got one bhb and a little bit oso not cute and not nice de ah ma cum piggy \"pulled\" my leg worx...very pain lor! hor...u noe who u r la...nitz! zzz...  '),(16351,4163734,'08,August,2004','2004-08-08 00:00:00','   suddenly tot of fuhua...haha i oso dun why my mum choose this school so after the PSLE i had to choose btw jurong and fuhua, but jurong too old le and fuhua too young le...haha whatever la but in the end i still went to fuhua.  in fuhua, everything was so different from rulang. maybe bcoz i have different friends, different teachers, different subjects and different CCA.  sec 1...quite boring...i met sheila and caren coz they are in charge of our class during sec 1 orientation...both of them always bullied by us (1C)-me still shy so nv talk much, lolx then other than eng,chi,math and sci the other subjects were all average so my mum forced me to have tution...haiz...and my mum want me to join red cross dun noe for wat so my first choice is rc la...the training was tough lor and i want to quit many many time...the seniors all so fierce during the camp...sobsob but it was very fun and i learnt alot...i nv talk to them before..haha opps i forgot to say...mr loh is our form teacher he ar...a good teacher. at the end of the year, one of the teacher treated us pizza but i forgot who liao =>  sec 2 ar...i start to talk in class and red cross. began to talk to esther they all but not much-just hi there hi here. in class i sat with poh ling haiz...becoz of her...i start to study and pay attention in class...haha 2C very united. this is very true lor...then ar in red cross, i start to play hard and learn hard there...but i still very quiet didn\'t talk much...then start to like training and all that stuff...chee yong sir and xuan hui mdm always encourage us to endure then this year the sec 4 senior passing out liao. then, at the end of the year, we got to eat pizza again. this time miss chua(form teacher) treated us(should be bah...) when we were eating half way, miss chua announced which class we would be going...yeah...my mum want me to go 3B...luckily me squeezed in 3B(or else i die ar...)so with poh ling again haha...  sec 3 liao...wa lau dun know anyone again must restart...haiz...suan le. then sad to say, after all our sec 4 express de senior passed out, i began to talk...then i make friend with them-yvonne,lek teng,kok seng,madeline,mahadair,and the \'always so supporting\' sec 5 seniors-chai ting,big and small hafizah,dina,atiqah,hidaya and nural they all...ya hor, still got wei quin they all. ok...last time saw them just hi only now we will play a bit but usually i got bullied... :x then we have footdrill competition and fuhua speech day(before and after training sure go drink bubble tea de) haha...maybe bcoz of all the training, we become more united. then i hang out with qiu yu, denise,esther and the sec 2s and the sec 1s...back to class...everything so fake...haiz but really get to know many friends.then i know audrey, polly, jolyne...dun noe how come i know them...like we all met under funny sitution??? nvm...i blur liao...nitz if i remember somemore then i write again...  today got fireworks leh..my lao gong going to watch with her darling...suan le she so bad. i go NDP tml without her.. :p  '),(16352,4163734,'07,August,2004','2004-08-07 00:00:00','   yoz! coz my sis (audrey) want me to practise engalissh (english) so here i am lor...but hor i dun know wat to write leh..today not very good mood la, coz cannot go to the camp quite pity but at least i found out that surprisingly some people who alway bully me actually an wei me leh...can\'t believe it...haha jk la... :p  then today me n abigail met at kembegan then we go explore the shopping centre...very quiet de so eerie..haha then we went to the campsite n everything went so smoothly...want bus got bus, want train got train...then we reached the campsite le hor, we saw abt 3 ppl waiting then the gate is locked...i got a bad feeling but we still wait then more ppl came...then so suai we really came to the wrong place for briefing...wa lau!!! so we were told to fly to \"dobby got\"...  when one is unlucky ar...want train miss train then suddenly the train move so slow...then the traffic light all red de...haiz...when we reached there then they told us we are reserve and this briefing we dun need to come then she took our attendence...huh, for wat? me and abigail cursed mrs kim :x...but realli very fun lor...run there run here just bcoz we got a blur teacher in charge...damm it. :C  '),(16353,3389016,'20,May,2004','2004-05-20 00:00:00','   urlLink    This is my Boss (on the ship) Robert, but call him Bobby. (And the sheep call him Bhaaaby!, it is still funny.)When I woke up it was 8:30am (I got a whopping 4hrs of sleep) and I had to be in a meeting at 9am. I was lucky to wake up. I got ready and went to the conference room and I felt very tired. During the meeting JP shows up apparently he and Bruce and Robert (Bobby) were out till about 8am they went to three sisters and then got lost on the way back since there was such thick fog. Did I say these guys could party? I watch JP do training after the meeting from 9-12:30. Man what a trooper. He is really good and funny too. I drink a lot of coffee and orange juice.&nbsp; urlLink   '),(16354,3389016,'20,May,2004','2004-05-20 00:00:00','   urlLink    I met a lot of fun people from the IT group that night. These guys party hard!&nbsp; urlLink   '),(16355,3389016,'20,May,2004','2004-05-20 00:00:00','   urlLink    He and another guy drove us to the ship and stopped at a shell on the way for something to drink. Once we got to the ship they took pictures of us for out ID cards (god I looked awful!) and got us checked in to our rooms we were then informed that we would have to leave the ship in two hours at 6pm since there was a black out. (I still haven\'t slept at this point) I check into my room get totally unpacked, moved in, and take a shower. I am on my bed counting my new monopoly money and Mohammad (the front desk guy) & Joe knock on my door and tell me that I have to change rooms and I only have 5 min to do it, since the ship is evacuating. So we run down the hall with my clothes draped over their arms (I am in heals) to my new room and I have to leave immediately to meet everyone out side. I forget to bring my ID and my jacket (lack of sleep). Now I have to figure out how to walk across a gangway riddled with holes in heals, lets just say I was getting laughed at and I hadnï¿½t even been visible for more than 5mins. I find out that we are going go-cart racing at an indoor track down the road. It is so cool when we get there (remember I am in heals). The track is pretty complicated with tunnels and bridges. We sign up for the race and I have to wear a very smelly jump suit and drive in high heels but somehow I manage to race pretty well. I am all hopped up on RedBull and ready to go. My highest score placed was 7th. I had a lot of fun waving at everyone while I was driving. I can\'t believe that they let us drink and drive (I love this country). We also had a barbeque that we got to cook our own meat on the grill.&nbsp; urlLink   '),(16356,3389016,'20,May,2004','2004-05-20 00:00:00','   urlLink    This is Joe. So we get to the college town of Groningen and start looking for who is going to pick us up. When we finally find a phone (we figure we are going to need to call a cab since we are late from the luggage incident) and are reading it to see if we can use it to call the ship (it is not in English, but come on we are both System\'s Engineers. How hard can it be to figure out how to use a phone right?), just then a guy walks up and asks us if we are going to Royal (he is wearing a RCCL jacket). He saw us wandering tired with our luggage and figured that we were the ones he was looking for. Thank God!&nbsp; urlLink   '),(16357,3389016,'20,May,2004','2004-05-20 00:00:00','   urlLink    When we departed the plane we went to get our luggage, we realized our luggage was missing when we noticed we were the only ones standing at the turn style (I was a little out of it from lack of sleep), but I really didn\'t care cause I was finally standing on foreign soil. We finally get our luggage about a half hour later and went to buy our tickets for the three-hour train ride we were about to take up to Groningen. I am now on the train writing this and I am taking some videos here and there. The weather is nice they said it is 50F out but it feels like 70F. It was raining now it seems to be clearing up. &nbsp; urlLink   '),(16358,3389016,'20,May,2004','2004-05-20 00:00:00','   urlLink    Hello There, Welcome to my Blog. This couldn\'t come at a better time now that I have just made my first trip to Europe.   Oh the stories...and the secrets.   Anyway, I will try to take you on the trip with me as it happened. I will try to catch up to now and keep a record of my journeys. This is really for all of you out there that I have bought a postcard for and never dropped it in the mail. I have a whole stack of them piling up.   Feedback is always good to hear so let me know if you have any suggestions on how I can make this less narcissistic. I hope you enjoy! &nbsp; urlLink   '),(16359,3768627,'27,June,2004','2004-06-27 00:00:00','   My first step into the world of Blogger.  My livejournal is just to read and comment on others.  This...this is me.  My hopes...  My fears...  My dreams...  Once before I tried to trust, and found my dreams shattered.  Even as the glass rose exploded outwards, glass shards flying everywhere, glittering in the light till it seems a thousand stars fell from the sky...  I found it beautiful.  And now, once more, I set forth my hopes and dreams. Intact they are wonderous, in progress they gleam, and broken they are beautiful.  Help them break or help them grow, that is your choice.  If I could have only one thing, it would be the power to change. People will not agree with me; in fact, most people do not. But if one thing could happen, it would be everyone who reads this learns something. They may not agree with my views. But I hope they take away an increased understanding of another person\'s perspective. '),(16360,3768627,'28,July,2004','2004-07-28 00:00:00','   I just wanted to say thank you...    Took a good hard look at myself today.    Despite the fact that Iâ€™ve been with Sothis all weekend, Iâ€™ve felt horribly alone. Iâ€™ve been feeling ill. And worthless. And even as life looks better for me, even as Iâ€™m gaining these new friends and supporters, Iâ€™m being pulled down into a vortex of despair.    And itâ€™s of my own making.    Everythingâ€™s weighing down on me. Iâ€™ve taken on so much. Did I really think I could do it all? I canâ€™t seem to find the time or energy to work on anything, and thereâ€™s so much Iâ€™ve devoted time to. I seem to be lurking on messengers and chatting,a nd thatâ€™s all I do anymore.    My two Pern RPGs, EvenStar and Fallen Stars. Theyâ€™re both wonderful places, wonderful people. Last few months itâ€™s been post post post, lots of stories coming out, especially ebtween me and Jade and me and Heather and me and Sunrose. Theyâ€™re three amazing ladies, and Micheleâ€™s been awesome too. Through July, though, itâ€™s been blah, donâ€™t want to post. Iâ€™m letting them all down...    Then there are the projects Iâ€™m working on. Warped is kind of dead, and Iâ€™m not ploting to help with it, I just haveta tell iPenguin. Now Ilathid... Ilathid is an incredible game, incredible. And I have a pretty good role in it too. Important enough for me. Iâ€™m so glad Sivrehn let me in. Heâ€™s such a nice person, helping me whenever Iâ€™m down, cutting me a lot of slack. i wish I could give more to the project.    Then thereâ€™s Avriga. I love it dearly. The places and peoples and...itâ€™s becoming my world, slowly. Iâ€™ve sunk so much into this, used everything Iâ€™ve got. Am combining the mythologies of so many of my previous stories, giving it a richness and a broadness (I hope) that it deserves.    But itâ€™s also been my downfall. I threw so much into it that the fight I had with a certain person sapped me. I hated what it did to me. I can take criticism of the story for an actual problem, but not for something based soly on a biased, extremist belief. i know many people who hold the same/similar beliefs and werenâ€™t as judgemental and biased. And they didnâ€™t shove their beliefs down my throat.    I was wrecked by that incident. Iâ€™m not sure if Iâ€™m fully over it, or if Iâ€™m ever going to be fully over it. But Iâ€™m writing for Avriga again, which is a good thing. It no longer really hurts to read the writings. There are still times when I canâ€™t continue with the story, but they are few and far between. Iâ€™m determined not to let this universe die.    Then of course thereâ€™s the most important thing. All of you. The Myst community. The Lysts, DPWR, MC, and CC have so many wonderful people there. Theyâ€™re supportive. Theyâ€™re communicative. Theyâ€™re people I can laugh with, debate with, and share myself with. And I know theyâ€™ll understand.    Theyâ€™re becoming my world.    For so long Iâ€™ve been trying to keep my on and off-line worlds from converging. In real life Iâ€™d be one person, and online Iâ€™d be Capella (or Yoake, for the Pernese gang). But lately itâ€™s been so hard to keep them seperate.    The lineâ€™s blurring until I donâ€™t know who I am at any given time anymore.    At first it unnerved me, but then I realized iktâ€™s okay. Why should I be keeping my two lives seperate? Why should I act like one person in real life and be a completely different person on the net?    Okay, so maybe my deliberate harshness and coldness in real life is based on my circumstances, and itâ€™s what I have to do to make it. But what would happen if I let that Elyse go? What if I blended her with the hyper-er, looser, more outgoing and talkative and hyper and more sensitive and caring and a (kind of) better person, Capella, that I am online?    Whenever Iâ€™m chatting with people from the Myst communities, an awful lot of them compliment me. Say how much they like my works, how nice I am, how considerate, how much fun I am to talk to, etc. Itâ€™s a strange feeling, because I never get that in real life. Iâ€™m sure theyâ€™d hate the person I am in real life. But they seem to like the me I am on the internet (whoâ€™s closer to the true me than my â€˜realâ€™ me).    There have been times Iâ€™ve felt like nothingâ€™s going to go right and felt like doing something drastic. But someone always signs on or messanges me around then. and theyâ€™ll say something, and Iâ€™ll realize that hey, itâ€™s not so bad after all.    This communityâ€™s given me so much. A place to discuss, debate, and share without anyone being autiomatically prejudiced. A place where I can be accepted as who I am, instad of what I look like  and where I fit in an artificial social hierarchy. A place where I can belong when I donâ€™t have one in real life.    And there are a few special people I want to thank (not by name, but I think theyâ€™ll know who they are). A crazy-n-fun person, a musician, a dreamer, a writer, a programmer, a fellow (decent) compy geek, and a real life friend. For believing in me, for encouraging me,  for giving me a reason to go on, and for slapping me when I need it.    Thank you ALL, and I love you all. '),(16361,4256858,'14,February,2001','2001-02-14 00:00:00','   Once upon a time,There lived a king. The King had abeautiful daughter, the Princess. But the kingdom was a sad place.There was no laughter, and no joy.The problem was, that everything the Princess touched would melt.No matter what, metal, wood, plastic .. .. anything she touched would melt!! Because of this, men were afraid of her.Nobody would dare marry her.The king despaired.What could he do to help his beautiful daughter?He consulted his wizards and magicians. One wizard told the king, \"If yourdaughter touches one thing thatdoes not melt in her hands, shewill be cured.\" The King was overjoyed. The nextday, he held a competition. Any manwho could bring his daughter anobject that would not melt at hertouch, would marry her and inheritthe King\'s wealth. Three young princes took up the challenge. The first Prince brought a very hardalloy of titanium.When the Princess touched it, itmelted. The Prince went away sadly. The second Prince brought a hugediamond, thinking that diamond isthe hardest substance in the worldand will not melt. But, alas, once the Princess touchedit, it melted. He too went away disappointed. The third Prince approached. He told the Princess, \"Put your handin my pocket and feel what is in there.\"The Princess did as she was told,though she turned red. She felt something hard. She heldit in her hand .. .. and it did not melt!! The King was overjoyed!!Everybody in the kingdom wasoverjoyed!! And the third Prince married thePrincess and the both lived happily ever after.  The question is?!?!  What WAS the object in the Prince\'s pocket?!?!   They were M&amp;M\'s, of course.THEY melt in your MOUTH,NOT in your HAND!!  '),(16362,4256858,'02,April,2001','2001-04-02 00:00:00','   TRUE STORY....  AFTER A FEW OF THE USUAL SUNDAY EVENING HYMNS, THE CHURCH\'S PASTOR SLOWLY STOOD UP, WALKED OVER TO THE PULPIT AND, BEFORE HE GAVE HIS SERMON FOR THE EVENING, BRIEFLY INTRODUCED A GUEST MINISTER WHO WAS IN THE SERVICE THAT EVENING.  IN THE INTRODUCTION, THE PASTOR TOLD THE CONGREGATION THAT THE GUEST MINISTER WAS ONE OF HIS DEAREST CHILDHOOD FRIENDS AND THAT HE WANTED HIM TO HAVE A FEW MOMENTS TO GREET THE CHURCH AND SHARE WHATEVER HE FELT WOULD BE APPROPRIATE FOR THE SERVICE. WITH THAT, AN ELDERLY MAN STEPPED UP TO THE PULPIT AND BEGAN TO SPEAK.  \"A FATHER, HIS SON, AND A FRIEND OF HIS SON WERE SAILING> OFF THE PACIFIC COAST,\" HE BEGAN, \"WHEN A FAST APPROACHING STORM BLOCKED ANY ATTEMPT TO GET BACK TO THE SHORE. THE WAVES WERE SO HIGH, THAT EVEN THOUGH THE FATHER WAS AN EXPERIENCED SAILOR, HE COULD NOT KEEP THE BOAT UPRIGHT AND THE THREE WERE SWEPT INTO THE OCEAN AS THE BOAT CAPSIZED.\"  THE OLD MAN HESITATED FOR A MOMENT, MAKING EYE CONTACT WITH TWO TEENAGERS WHO WERE, FOR THE FIRST TIME SINCE THE SERVICE BEGAN, LOOKING SOMEWHAT INTERESTED IN HIS STORY.  THE AGED MINISTER CONTINUED WITH HIS STORY, \"GRABBING A RESCUE LINE, THE FATHER HAD TO MAKE THE MOST EXCRUCIATING DECISION OF HIS LIFE: TO WHICH BOY HE WOULD THROW THE OTHER END OF THE LIFE LINE. HE ONLY HAD SECONDS TO MAKE THE DECISION. THE FATHER KNEW THAT HIS SON WAS A CHRISTIAN AND HE ALSO KNEW THAT HIS SON\'S FRIEND WAS NOT.  THE AGONY OF HIS DECISION COULD NOT BE MATCHED BY THE TORRENT OF WAVES. \"AS THE FATHER YELLED OUT, \'I LOVE YOU, SON!\' HE THREW OUT THE LIFE LINE TO HIS SON\'S FRIEND.  BY THE TIME THE FATHER HAD PULLED THE FRIEND BACK TO THE CAPSIZED BOAT, HIS SON HAD DISAPPEARED BENEATH THE RAGING SWELLS INTO THE BLACK OF NIGHT. HIS BODY WAS NEVER RECOVERED.\"  BY THIS TIME, THE TWO TEENAGERS WERE SITTING UP STRAIGHT IN THE PEW, ANXIOUSLY WAITING FOR THE NEXT WORDS TO COME OUT OF THE OLD MINISTER\'S MOUTH.  \"THE FATHER,\" HE CONTINUED, \"KNEW HIS SON WOULD STEP INTO ETERNITY WITH JESUS AND HE COULD NOT BEAR THE THOUGHT OF HIS SON\'S FRIEND STEPPING INTO AN ETERNITY WITHOUT JESUS. THEREFORE, HE SACRIFICED HIS SON TO SAVE THE SON\'S FRIEND.  HOW GREAT IS THE LOVE OF GOD THAT HE SHOULD DO THE SAME FOR US.  OUR HEAVENLY FATHER SACRIFICED HIS ONLY BEGOTTEN SON THAT WE COULD BE SAVED.  I URGE YOU TO ACCEPT HIS OFFER TO RESCUE YOU AND TAKE A HOLD OF THE LIFE LINE HE IS THROWING OUT TO YOU IN THIS SERVICE.\"  WITH THAT, THE OLD MAN TURNED AND SAT BACK DOWN IN HIS CHAIR> AS SILENCE FILLED THE ROOM.  THE PASTOR AGAIN WALKED SLOWLY TO THE PULPIT AND DELIVERED A BRIEF SERMON WITH AN INVITATION AT THE END. HOWEVER, NO ONE RESPONDED TO THE APPEAL. WITHIN MINUTES AFTER THE SERVICE ENDED, THE TWO TEENAGERS WERE AT THE OLD MAN\'S SIDE.  \"THAT WAS A NICE STORY,\" POLITELY STATED ONE OF THE BOYS, \"BUT I DON\'T THINK IT WAS VERY REALISTIC FOR A FATHER TO GIVE UP HIS ONLY SON\'S LIFE IN HOPES THAT THE OTHER BOY WOULD BECOME A CHRISTIAN.\"  \"WELL, YOU\'VE GOT A POINT THERE,\" THE OLD MAN REPLIED, GLANCING DOWN AT HIS WORN BIBLE. A BIG SMILE BROADENED HIS NARROW FACE, HE ONCE AGAIN LOOKED UP AT THE BOYS AND SAID, \"IT SURE ISN\'T VERY REALISTIC, IS IT? BUT I\'M STANDING HERE TODAY TO TELL YOU THAT STORY GIVES ME A GLIMPSE OF WHAT IT MUST HAVE BEEN LIKE FOR GOD TO GIVE UP HIS SON FOR ME.  YOU SEE --- I WAS THAT FATHER AND YOUR PASTOR IS MY SON\'S FRIEND.\"  '),(16363,4256858,'24,June,2002','2002-06-24 00:00:00','   Dati mo na syang kakilala. Lumalabas kayo pero hindi date yun ha. Wala lang. kain lang sa labas, kwentuhan lang. Walang katapusang kakatext. Namumulubi ka na nga kasi magka-iba ang provider nyo. Naiirita ka kung ang tagal ng reply nya, kaya kahit alam mong wala ka pang pambili ng load tatawagan mo sya. Napipikon ka na sa provider mo kaya para masaya,nagpa-sim swap ka. Pag malungkot ka, bad trip, naiirita kahit pa ata bente pesos na lang yung natitirang load mo, tatawagan mo sya. Wala lang. Basta ang alam mo lang you want that person to be there with you. Emote! Naguguluhan ka. Hindi kasi sya katropa. Conflicting ang interest nya tsaka ng mga friends mo. Pero kapag niyaya ka nya kahit na sino pang magtampo sasama ka. Then, katampuhan kayo. Hindi mo na sya kinausap at pinanindigan naman ng loko. Hurt ka syempre. INIYAKAN MO, ITINULOG, IKINAIN.Pero NANGITIAN KA LANG OK NA. Ngayon magkasama na kayo ulit. As if naman may choice ka. Kahit ayaw mo kunyari wala ka namang magagawa. Tamang-tama broken-hearted ka. Litong-lito. Since ok na kayo,lumalabas ulit kayo. Gaya ng dati, wala lang kain lang ulit sa labas,kwentuhan. Ngayon mas kilala nyo na ang isa\'t-isa. Alam na nya halos lahat ng tungkol sa\'yo. Alam nyo kung bad trip ang isat-isa. Alam nyo kung nalulungkot o nag-eemote yung isa. You still have your misunderstanding tsaka tampuhan pero this time after the tampuhan, syemps babawi na kayo sa isat-isa. Sweet nyo noh? Grabe na ito! Takbuhan mo na sya ngayon kung may problema ka. Naikwento mo na lahat ng sama ng loob mo, mga bagay na hindi mo masabi dun sa lalaking iniyakan mo. Unang-una dahil iba ang kasama nya at hindi mo sya makausap ng matino.Bukod pa dun sa parati na lang kayong nag-aaway nung iniyakan mo pag nag-uusap kayo. Pero this one,comfortable ang feeling nyo with each other. Sweet kayo kahit para syo hindi. You actually have your breakfast together at hindi lang yun, you share the same cup of coffee pa. Pinagtitimpla mo sya ng coffee at ginagawa mo para sa kanya yung hindi mo magawa para dun sa iniiyakan mong may kasamang iba. RING!!!! Oops phone call para sa kanya. siginificant other nya. Bakit ka namumula? Magpaapekto ba. Selos ka noh? BAKIT. KAYO BA??? Hindi naman diba! Concentrate. Deep breathing..... Inhale.... Exhale..... Kaya mo yan. Focus.Try to look back. Ano ang nakikita mo? Sino ang iniisip mo? SIYA. Hindi ba dapat yung iniiyakan mo nung nakaraang buwan lang? Syet! You are starting to realize how cool this person is. Syeter! You are beginning to doubt you\'re feelings dun sa sinabihan mong mahal mo. Syetest! Pero knowing you, idedeny mo na you are actually starting to feel something for this person. What \'dja gonna do? But of course, alamin mo kung friends lang ba kayo.How the hell are you going to find out? Aha! Go to the person and ask it right to their face. \"Gusto mo ba ako?\" Grabe, as if puwede.Do you have the guts to do it? Do it. I\'m right behind you, theoretically speaking. Eniwey, you still have the option of whether you would like to know or not. Think. Think hard. Super malambing siya sa iyo pero... Alam mo naman na ganun din sya sa iba. Eh wala naman palang problema eh! Eh bat ka pa apektado???? Sabagay hindi kita masisi kung bakit. Lalo na nung he starts humming love songs (pamilyar di ba? parang yung kakilala mo.), starts asking you funny questions. Like... \"What would you do if a guy na hindi mo masyado kilala asks you out for a date?\"Ikaw pa ang tinanong. Eh di ba sumama ka nga sa kanya. Remember...nag-aalangan ka pa nga nun kze hindi mo pa sya ganun kakilala. Kung sino-sino na nga binulabog mo para pagtanungan kung papayag ka. Now he starts asking you about certain situations. \"Pano kung niyaya kang manood ng sine at hinawakan ang kamay mo. Ano daw ang reaksyon mo?\" Sabay patong ng kamay nya sa kamay mo. Dramatization kung baga. Nanlalamig ka at kinakabahan, hindi dahil sa tanong syempre noh.Yun na yun.Sinagot mo syempre. Sabi mo tatanungin mo kung para san yun syempre. \"Eh pano kung hinawakan na talaga?\" Sabay hawak sa kamay mo, holding hands na this time. Syempre nanlalamig ka na at ninenerbyos. Sabi mo, \"ok lang magkaibigan naman ata sila eh\". \"Eh pano kung I-kiss ka?\" Sabi mo, \"What for?\" Thank you daw for the wonderful evening..ek ek.Sabi mo \"Ah ok....basta sa pisngi lang ha. Kasi pag sa lips na masyado nang intimate yun. di ba?\" Eh pano daw kung hindi nagpaalam tapos kiniss ka sa lips? Stolen kiss daw. \"Aba loko noh?! As if naman merong magtatangka\".Then he answers you \"Ako, I would!\" SOOOOOOWWWWWSSSSSSS !!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Tapos nanood na nga kayo ng sine.Medyo biyolente yung eksena. Takot ka syempre. tinatakpan mo yung mata mo ng kamay mo. Inaalis naman nya. Kinakagat mo yung mga kuko mo. Inaalis pa rin nya yung kamay mo. Kinuha nya tapos pinatong nya yung braso nya para hindi mo na makuha. Ayaw mong paawat kaya yung kanan naman. Sumuko, binitiwan yung kaliwang kamay mo.Kinagat mo ulit yung kuko mo sa daliri mo kze natatakot ka. Kinuha nya ulit yung kamay mo. Hindi rin makulit no? Agaw trip ba? Akala mo hahawakan lang. Aba ang loko, nakikagat sa kuko mo. Ano ba yan????? NO MALICE???? NO MALICE.... Ang gulo noh???? Tutuksuhin ka dun sa lalaking sinabi mong mahal mo. Titingnan yung reaksyon mo.Pag ikaw naman ang kusang nagkwento, iirapan ka tapos tatalikuran na. PAMBIHIRA. Men..... Hayaan daw syang tumulong para makalimutan mo na yung lalaking iniiyakan mo. Feel mo nagbibiro lang..... hmmmm nagbibiro nga lang ba? Ayaw mong bigyan ng malisya lahat ng nagyayari sa inyo kasi magkaibigan kayo tsaka alam naman nya na inlove ka. Pa rin? Eh bat ka nalilito ngayon kung mahal mo nga yung sinabihan mong mahal mo. Mahal mo pa nga ba??? Eh ba\'t hindi mo na masagot yan ngayon? Ano akala ko ko ba dati mas pipiliin mo yung lalaking mahal mo. Eh bat nag-iba ata ang ihip ng hangin. Mas gusto mo na ngayon yung mahal ka. Hindi lang sa salita kundi sa gawa. WALA LANG TALAGA YANG LAGAY NA YAN HA??? AAWW SHUT UP!  '),(16364,4256858,'25,December,2002','2002-12-25 00:00:00','   Please don\'t drink and drive, someone cares about you!! ........the true meaning of christmas and all year around...  This has gone around before, but it is well worth sharingagain. Please read,, this hits very close to home On the last day before Christmas, I hurried to go to thesupermarket to buy the remaining of the gifts I didn\'tmanage to buy earlier. When I saw all the people there, I started to complain to myself,It is going to take forever here and I still have so many otherPlaces to go.  Christmas really is getting more and more annoying everyyear.How I wish I could just lie down, go to sleep and only wake upafter ...Nonetheless, I made my way to the toy section, and thereI started to curse the prices, wondering if after all kids really playwith such expensive toys.While looking in the toy section, I noticed a small boy of about 5 yearsOld, pressing a doll against his chest. He kept on touching the hairof the doll and looked so sad. I wondered who this doll was for. Then thelittle boy turned to the old woman next to him:\"Granny, are you sure I donthave enough money?\"The old lady replied: \"You know that you dont have enough money to buythis doll, my dear.\"Then she asked him to stay there for 5 minutes whileshe went to look around. She left quickly. The little boy was still holding the doll in is hand.Finally, I started to walk towards him and I asked him who he wanted togive this doll to.\"It is the doll that my sister loved most and wanted somuch for this Christmas. She was so sure that Santa Claus would bring it toher.\"I replied to him that maybe Santa Claus will bring it to her afterall, andnot to worry. But he replied to me sadly. \"No, Santa Claus cannot bring itto her where she is now. Ihave to give the doll to my mother so that she can give it to her when shegoes there.\"His eyes were so sad while saying this. \"My sister has gone to be with God.Daddy says that Mummy will also go to see God very soon, so I thought thatshe could take the doll with her to give it to my sister\".My heart nearly stopped. The little boy looked up at me andsaid: \"I told daddy to tell mummy not to go yet. I asked him to wait untilI came back from the supermarket\". Then he showed me a very nice photoof him where he was laughing. He then told me: \"I also want mummy to takethis photo with her so that she will not forget me. I love my mummy and Iwish she doesn\'t have to leave me but daddy says that she has to go to bewith my little sister\". Then he looked again at the doll with sad eyes,very quietly. I quickly reached for my wallet and took a few bills and saidto the boy.  \"What if we checked again, just in case if you have enough money?\" \"Ok\" he said. \"I hope that I have enough\".I added some of my money to his without him seeing and we started to countit. There was enough for the doll, and even some left over.The little boy said: \"Thank you God for giving me enough money\".Then helooked at me andadded:\"I asked yesterday before I went to sleep for God to make sure I haveenough money to buy this doll so that mummy can give it to my sister. Heheard me. I also wanted to have enough money to buy a white rose for mymummy, but didn\'t dare to ask God too much. But he gave me enough to buythe doll and the white rose.\" \"You know, my mummy loves white roses\". A fewminutes later, the old lady came back again and I left with my cart.I finished my shopping in a totally different state from when I started. Icouldnt get the little boy out of my mind. Then I remembered a localnewspaper article 2 days ago, which mentioned of a drunk man in a truck whohit a car where there was one young lady and a little girl. The little girldied right away, and the mother was left in a critical state. The familyhad to decide whether to pull the plug on the life-assistingmachine, because the young lady would not be able to get out of the coma.Was this thefamily of the little boy?Two days after this encounter with the little boy, I read in the newspaperthat the young lady had passed away. I couldn\'t stop myself and went to buya bunch of white roses and I went to the mortuary where the body of theyoung woman was exposed for people to see and make last wish before burial.She was there, in her coffin, holding abeautiful white rose in her hand with the photo of the little boy and thedoll placedover her chest. I left the place crying, feeling that my life had beenchanged forever. The love that this little boy had for his mother and hissister is still, to this day, hard to imagine. And in a fraction of asecond,a drunk man had taken all this awayfrom him.  ...  FRIENDS ARE LIKE ANGELS, WHO HELP US FLY WHENOUR WINGS HAVE FORGOTTEN HOW TO FLY.  '),(16365,4256858,'14,February,2004','2004-02-14 00:00:00','   break up lines... Ano yung mga break up lines na nagamit na sa inyo? Share ko lang ang mga break lines ng iba....meron pa kayong mas grabe pa??? ? \"I just realized that I don\'t want to be touched.\" -- Whaddafuckisdat?!  urlLink   \"I\'m confused and I need some time out to find myself!!!\" -- Eto flashlight, go look for yourself!  urlLink   \"maybe this is not the right time for us.\" --anak ng tipaklong! di na kita tatanggapin ulit no?!  urlLink   ? \"Di kita maalagaan ng tulad ng iniexpect mo. You deserve someone better. That\'s not me.\" -- eh niligawan mo pa ako at pinasagot? jerk!!!!  urlLink   \"Lasing lang ako kagabi. Sorry.\" -- Oo nga at nag enjoy ka pa! a**h***  urlLink   Ito kinwento sa akin ng friend ko, na sinabi daw sa kanya ng boyfriend niya. \"We are too different from each other.\" To which my angry friend retorted: \"Tama ka diyan! I\'m a woman, you\'re a wuss.\"  urlLink   true lies and true lines: ?\"we have to meet other people to see what it\'s like to be with someone else. i\'m giving you that choice. if you come back to me, we\'ll both be better from having chosen freely\" -- ano raw?  urlLink   \"Someday, hahanapin kita, when we\'re both ready. When we don\'t care about the odds. Kung tayo talaga, tayo rin in the end, di ba?\" -- ULUL!!! Wala ka nang babalikan!  urlLink   \"I really think that we should break up.\" \"Why?\" \"Because I don\'t know if I still love you.\" -- Ouch pare! Oks lang sana kung ganun lang eh...tapos next week may kasama na siyang iba!!!  urlLink   \"I just realized I haven\'t had time to walk my dog.. \" -- ?????????  urlLink   ? \"its not you, its me...\" -- ket?, whats wrong with you?  urlLink   ? \"sorry ha.. kaya kita niligawan kasi gusto ko lang mapalapit sa kuya mo.. \"-- This Guy is inlove with you pare... ? urlLink   \"hindi ko kayang buhayin ang buong pamilya mo.. kaya paalam na lang...\"-- sabay takbo.. hehehe  urlLink   ? \"teka, anong araw nga ba dapat ikaw nasa schedule ko...\" -- hehehe..naka-playboy...  urlLink   ?.. reality check... eto yung normal na dahilan kung bakit may break-up...ayaw lang umamin ng karamihan sa mga lalaki... ? lalaki sa babae: ... sawa na ako sa iyo at napaka-routine ng ginagawa natin lagi...  urlLink   ? ... mas maganda yung nakilala kong bebot at mas masaya kasama kesa sa iyo..ikaw kasi boring...  urlLink   ? ... ang hilig mong mag-nag... puro ka selos-dito selos doon.. nakakapika na..grrrrr..  urlLink   ? ... sorry hindi na kita feel kasama..  urlLink   ? ... wala na akong time para sa iyo kasi ang tutuo may nililigawan na akong iba..  urlLink   babae sa lalaki: ... para ka kasing bakla kaya ayaw ko na sa iyo mas maarte ka pa sa akin..  urlLink   ? ... ang dami mo namang restriction.. nasasakal na ako.. bakit kasal na ba tayo at marami akong nde puedeng gawin..  urlLink   ? ... kaya lang naman kita sinagot kasi wala akong partner sa JS prom...  urlLink   ? ... kaya lang naman kita sinagot kasi pinilit ako ng barkada ko.. pero sa tutuo lang walang akong feelings sa iyo..?? urlLink   ? ... sorry ha.. may nakita akong mas pogi at cute kesa sa iyo.. ? urlLink   ... sensya ka na ha.. pagod na akong maghintay at sumakay sa jeep o kaya sabus... feel ko na naman yung may Car- racter..  urlLink  '),(16366,4256858,'17,August,2004','2004-08-17 00:00:00','   1.) You have scars on your knees and elbows 2.) You owned a bike 3.) You had a barkada around your neighborhood (all of you had bikes) 4.) You loved climbing on your house\'s roof (and your neighbor\'s roof as well) 5.) You went inside an abandoned house in your neighborhood just to see what it looks like inside (ghost hunting) 6.) You ate all the aratilis in your neighborhood 7.) You plucked all the gumamelas in the area for soap bubbles 8.) Your parents forced you to take afternoon siestas with the threat that you will not be allowed to play outside. 9.) You are never found in your house in the afternoon. You are often found playing in the street with your neighborhood friends. 10.) You loved exploring vacant lots for hidden knick-knacks 11.) You just can\'t resist jumping in a sandpile 12.) You know all the street games (patintero, agawan base, langit-lupa, etc., etc.) 13.) You owned a family computer. 14.) Your hand-to-eye coordination is terrific due to family computer. 15.) You\'d rather go outdoors in the afternoon than play familycomputer. 16.) You gleefully boast that games today are so easy because of the character\'s life bar (remember when we used to play Mario? we died the minute a goomba hit us) 17.) You know this code by heart UP-UP-DOWN-DOWN-LEFT-RIGHT-LEFT-RIGHT-B-A-B-A-START (select-start for 2 players) 18.) You owned a superhero costume (especially a superman costume) 19.) For the girls: You dressed up like Punky Brewster, Madonna and Debbie GibsonFor the boys: You dressed up like David Hasselhoff\'s knight rider, David Bowie or had Clarke Kent\'s little bangAnd now you think that the 80\'s had the suckiest dress sense. 20.) You had a denim jacket 21.) You had a sticker book especially that Age of Dinosaurs sticker book. 22.) You were addicted to rainbow brite, carebears, my little pony, thundercats, bioman, voltes v, mazinger z, daimos, etc., etc. 23.) You played pc games like tapper, moonbugs, alley cat and prince of persia. 24.) Ms Word did not exist in your vocabulary but Wordstar did! 25.) You love 80\'s music even if you don\'t want to admit it. 26.) You\'ve climbed up mango trees to catch salagubang, tie a sting around its neck and let it fly around in frenzy 27.) You\'ve spent hours in the afternoon catching tutubi... yellow-green was the easiest to catch, blue being finicky, and red being a rare breed... 28.) You used to take Flinstones vitamins (which you didn\'t mind cause it was yummy) and Scott\'s li ver oil 29.) You know the cartoon show, Beverly Hills Teens 30.) You know Kuya Bodgie from Batibot 31.) You watch Uncle Bob\'s Lucky Seven Club 32.) You used AQUA NET to fashion ur 4-inch-high bangs. 33.) Your blouses had paddings. 34.) You owned wide studded colorful belts. 35.) You watched some of your favorite shows on betamax or even UHF 17 (the channel from Clarke Air Base - or was it Subic?) 36.) You collected and swapped; perfumed stationeries with your classmates and friends, but followed the unwritten rule that you never write on them. 37.) The only place you go to for summer vacation is BAGUIO ! 38.) On that note, camp john hay served BEST ice cream 39.) Most of us were brought to the EDSA revolution 40.) You remember what Ricky Martin used to look like back then 41.) ...when you thin k that Julie Vega is a better actress that Judy Ann 42.) You get confused playing playstation because of all the buttons on the keypad (nintendo only had the direction pad, a & b buttons and the start &amp; select keys ) 43.) You know all the Bagets and Ninja Kids 44.) You got to ride the train ride at greenhills shopping center 45.) Magic Johnson and Larry Bird were the players at the time 46.) ...when you know more 80\'s music than the song Buttercup 47.) You were allowed to bathe in the rain 48.) You were taught to comb your hair one-sided 49.) You\'ve collected matangpusa and mongo beans so you can have ammunition for sumpit 50.) You knew who madam bola and sitsiritsit and alibangbang were . 51.) Every Christmas you anticipate going to BIG Bang sa Alabang- with the giant slide. 52.) You k now the Ewoks 53.) You had Mighty Kid shoes and Greg shoes. 54.) You know what Time Space Warp means (and you know who Fuma Lae-Ar is) 55.) You and your barkada had a specific Bioman name. 56.) You sucked all the nectar from the santan plant hence your mother got really pissed at you for destroying the santan plant. 57.) You played with marbles and text (yung cards ha!!!) And you count cards like this: I-SA, DALA-WA, TAT-LO, A-PAT... walo na iyon! 58.) For girls: You wore denim miniskirts with rubber shoesFor guys: You had those bitin na pantalon which you wore with high-top rubber shoes!!! 59.) You loved cheezels and chicakdees because of the great prizes it had! (remember sticky hands, bear popups, and the stick on tattoos which were banned due to drugs daw?) 60.) Puffy cone still existed!!! 61.) Sosy ka if you bought a Ma gnolia drumstick 62.) Twin Popsies were meant to be shared with a friend. 63.) Ice Drop was the cheapest treat. 64.) You have those disney bow biters for your rubber shoes 65.) You know who Alf is. 66.) You\'re familiar with the show Perfect Strangers. 67.) Idol mo si McGyver. 68.) That\'s Entertainment ang the bomb nung mga panahon na yun. 69.) Sikat ka kapag alam mo ang wordstar at lotus 1,2,3 70.) Six digits lang ang telephone number niyo dati. 71.) Tatlong .25 cents lang eh makakatawag ka na sa pay-phone. 72.) Cute pa si Aiza non sa Eat Bulaga. 73.) Si Amado Pineda pa ang nagbabalita ng panahon 74.) You drank Chocolate milk from the Magnolia glass bottle which you kept for holding water in your ref. 75.) Brown Cow tasted better than Hersheys! 76.) Shake Rattle and Roll 1 was the most horrifying movie for you then. 77.) The most comfortable shoes for you is still Sperry Topsiders. 78.) Dress shoes mo eh loafers pa rin. 79.) The best movies of all time are pretty in pink, breakfast club, 16 candles and some kind of wonderful 80.) You show off your pencil case which have hidden compartments that pop-out at the press of a button... 81.) You have Bensia pencils which are refillable... 82.) Fiesta carnival was the place to be (kumbaga enchanted kingdom sya ng 90\'s) 83.) Takot kang mag-year 2000 ksi baka magunaw ang mundo. 84.) Masarap ang Goya and Serg\'s. 85.) Nakakasakay ka pa sa kotseng walang aircon. 86.) You know the lyrics ng Tinapang Bangus at Alagang-alaga namin si Puti ng Batibot. 87.) You know these commercials:a. Y C BIKINI BRIEF - remember this one? Yc bikini brief /yc bikini brief / yc bikini brief for the man who packs a wallop / YC had fashion / YC has style...b. RA Homevision - those guys from cash and carry makati couldn\'t have done it better. Sports. Adventure. Cartoons. Award Winners and More.Featuring the voice talent of Frankie Evangelista. I will never forget the creepy hand at the start of the ad. Parang ET na nasa spaceship!c. Arthur\'s Legaspi Towers - nuff said.d. La Germania Mama Mia commercials - dito 88.) Bumibili ka ng caramel candy, texas or bazooka bubblegum, tira-tira at tootsie roll sa tindahan 89.) Naabutan mo pa na korteng flower ang singko 90.) You\'re familiar with this song: si nena ay bata pa, kaya ang sabi niya ay um-ah-um-ah-ah. hanggang patanda siya nang patanda at pabastos ng pabastos hehehe!  _________________________  '),(16367,4256858,'17,August,2004','2004-08-17 00:00:00','   JUST FRIENDS ? By Susan Nikaido  If my sources are correct, the following scenario isplayed out frequently among the world of Christiansingles. It happens on both sides of the gender line,but allow me to talk about it from my own femaleperspective.A man meets a woman and begins to show interest inher. He asks her to spend time with him on a regularbasis : hiking, biking, watching videos. He calls herat least once a week just to talk. He begins tellingher about the deeper things in his heart and invitesher to share at this level as well.He sends her mushy \"friendship\" cards and tells herthat she means a great deal to him. He may become alittle affectionate : the hugs begin to linger.The relationship has been defined as friendship, if ithas been defined at all. But after all this specialattention, the woman is definitely feeling more. Soshe asks the guy what\'s going on.To her surprise, he does a quick about-face. Heinsists they are just friends. After that, he avoidsher, leaving her hurt and bewildered. She feelsrejected - she has lost not only a romantic interest(she thought) but worse, a close and trusted friend.And she feels stupid. Did she really misread all thosesignals?After I experienced this as a single woman, I asked afriend my friend \"Joe\" to help me understand why guysdo this. After I told him my woeful tale, he said, \"Idid that to somebody once.\"\"What? Why would you ever do such a thing?\" I asked.\"We were getting too close, and it scared me,\" hesaid.This helped me understand why the \"just friends\"syndrome is so widespread. A guy wants to get to knowa girl, but without the pressure of dating. So hespends a lot of time with her, treating her in manyways like a girlfriend but defining the relationshipas friends. This way, if it begins to look likethere\'s no future in the relationship, or he\'s notready to \"get serious,\" he can back away with no messybreakup. It sounds like a nice arrangement - for theguy.But that approach can be a problem for the woman. If aman tells a woman he just wants to \"be friends\" but hetreats her like it\'s more than friendship, she willbelieve his behavior instead of his words. It sets herup for a big disappointment. Or if he invokes the\"just friends\" mantra after being asked about thenature of the relationship, but then promptly beginsto distance himself from the friendship, again, hisactions do not match his words.He may think he\'s sparing her feelings by avoiding abreakup. But by defining the relationship as afriendship, he hurts her even more deeply when hedisappears. A dating relationship comes with certainrisks. But she expects a friendship - especially sucha close one - to continue. Think about it this way : Abroken dating relationship says only, \"I don\'t want tomarry you\"; a broken friendship says to her, \"I don\'twant/value you on any level.\"Taking a woman down this path violates two scripturalprinciples. First, it\'s dishonest. The apostle Paulsaid that it was the way of the world - not of a godlyman - to say \"yes, yes\" and \"no, no\" in the samebreath (2 Cor. 1:17).A man of integrity will call a relationship what itis. Second, it is not kind or loving. The \"justfriends\" approach may be safer for the guy, but it isharmful to the woman. In effect, he is asking her forthe rewards of a dating relationship - companionship,emotional intimacy, even affection - without theresponsibility. He is playing with her heart, and herheart will probably get broken.But what if a guy does only want to be friends - orwants to develop a friendship before he decides todate?It\'s pretty simple. He just treats the woman like allhis other friends. He doesn\'t spend more time with heror call her more often than he does his other friends.He usually invites other people along when he getstogether with her. He doesn\'t pick up the tab whenjust the two of them go out. He avoids complimentsthat might communicate she is \"special\" to him. Helets her know he spends time with other women. He\'sextremely careful about showing any physical affection* even playful shoves or hugs.If, after getting to know her from a safe emotionaldistance, he wants a deeper relationship, he tells herthat he wants to date her.What if you are not thinking about more thanfriendship, but she asks about your intentions? Tellher you appreciate her friendship, but be honest aboutwhere you are. Above all, though it may be awkward fora while, continue to be her friend.Years ago, I began to be attracted to a male friend.Though I hadn\'t really been getting any signals thathe was interested in me, I knew it would help settlemy emotions to hear it from him. I asked. He affirmedme as a person but told me gently - but clearly - thathe thought of me only as a friend. And then he did awonderful thing. He kept being my friend. Though ithurt a little to learn I wasn\'t attractive to him inthat way, it helped to know he still valued me andwanted me around. This was nearly 15 years ago, andthough we have both moved to different states andmarried, we are friends to this day.Women can be great friends. But guys, unless you areready for a dating relationship, please be careful totreat us \"as sisters, with absolute purity\" (1 Tim.5:2), not as girlfriends, nor as something in between.Your honesty - with yourself and with us- will bepleasing to the LORD, and prevent hurtful confusionfor your sisters in Christ.In the ocean of life. God is our buoyant force.No matter how strong the wind and how giagantic the waves are, we shall not sink as long as we are anchored on HIM.  '),(16368,4256858,'17,August,2004','2004-08-17 00:00:00','   NOONG IN-LOVE KA PA Ang sarap ng in-love.  Sa unang beses niyo mag-date tapos naramdaman niyo ang sinasabi nilang \"spark\" nako, magkaharap pa kayo naiisip mo na ang bukas Para matawagan mo na siya at maayang makipag-date ulit, pero sa susunod kayong dalawa na lang.  Hindi ka mapakali pag wala siya sa tabi mo o hindi mo siya nakakausap sa isang araw, sigurado at sigurado tatawagan mo siya sa opisina, sa bahay at pag wala doon maglo-load ka sa cellphone mo at uubusuin ang 300 na prepaid sa loob ng isang tawag na wala naman ibang laman kundi bungisngis, kamustahan at kuento tungkol sa lahat ng sama ng loob mo sa mga ex mo.  At ang goodbye sa telepono ay nde matapos-tapos, hindi niyo malaman kung sino ang unang magbababa at umaasa kang me \"Take care\" sa huli. Me email pa yan sa umaga para masabi mo sa kanya na masaya ka na nakapag-usap kayo. Minsang gumimik kayo at upbeat ang sounds, hala! sayaw ka kahit parehong Kaliwa nag mga paa mo, pero gusto niyang sumayaw kaya sige sasayaw ka na rin. Kahit ilang mule, arctic at sub-zero ang orderin niya wala kang pakialam, mas mabuti nga kse nga puede mo na siyang akbayan at I-hug pag medio lasing na siya, pag hindi siya pumalag nako! score! para kang nasa langit.  Malamang, matapos mo siyang ma-hug eh ayaw mo nang maligo at lagi mong naaalala ang scent ng pabango niya na me halong amoy alcohol na pero para sa yo mabango pa din siya.  At siempre tuwing matapos kayong gumimik eh ayaw mo pa rin matapos ang gabi kaya hihirit ka pa ng coffee, kahit isandaan ang isang baso, \"So what?!?\" kamo, pera lang yan ang importante kasama mo siya.  Masaya ka rin pag na-traffic kayo kse makakapagkuentuhan pa kayo pauwi. Matiyaga mong inaral ang pagda-drive ng manual gamit lang ang isang kamay kse yun isa hawak yun kamay niya o nakadantay sa hita niya habang nag mamaneho ka. Araw-araw magmamakaawa ka na ihatid mo siya at kung puede ka rin niyang Ihatid pauwi, at kung lulusot baka pati lunch eh puede na rin na kayo ang maging lunchmates.  Pag me free time ka eh nasa bookstore ka para nagtingin ng mga puedeng ibigay na greeting cards. Nakalimutan mo na ang barkada mo, para sa yo malaking abala lang sila sa napakagandang lovelife mo kesehodang magtampo pa sila sa yo at magsolian na kayo ng kandila sa inaanak mo sa kanila. At kung aalis man kayo nde ka magkasya sa pagte-text lang sa lab mo, kelangan mong pumunta sa banyo para lang magkarinigan kayo pag tinawagan mo siya at sabihin na nde ka nag-eenjoy at mas gusto mo na siya ang kasama. Pagdating ng weekend nako para kang intsik! alas dies pa lang ng umaga nasa kanila ka na at me dalang suhol na breakfast para sa nanay niyang nakasimangot dahil natutulog pa ang anak niya eh andun ka na.  Pagdating ng gabi kahit antok na antok na siya eh ayaw mo pa ring umuwi, hinihintay mong makatulog siya sa mga bisig mo para makanakaw ka ulit ng kiss.  Iniisip mo rin kung kelan ka kaya niya ipapakilala sa friends niya? Nung kayo na, lahat ng monthsary ice-celebrate niyo, me kasama pang surprise na regalo at date. At nde miminsan mong nabanggit na gusto mo na siyang Pakasalan at wala nang ibang babae sa puso at wala ka nang makikita na katulad niya. AFTER ONE YEAR Mahal ang gimik sa bar, mas maganda kung kakain na lang kayo sa Jollibee at manonood ng sine. Wag na kayo magkape, masyadong mahal with matching comment na \"Leche, me ginto ba yan?\" Mas gusto mo nang kasama barkada mo dahil \"minsan\" lang kayo magkita sa isang linggo. Pgakahatid mo sa kanya, nagmamadali kang umuwi sa gabi dahil pagod ka na sa trabaho. Pumapasok at umuuwi na siya mag-isa dahil nde mo siya masusundo dahil puyat ka. Syet! wag ka niyang pipiliting sumayaw at nakakahiya. Anong tawag? Sa load mong 300, mauubos yun at 1 beses mo lang tinext (against 245 na text niya sa yo) at 3 minutes mo siyang natawagan.  Naubos ang load mo kakatawag sa mga barkada mo at kaka forward ng joke sa kanila. Pag weekend mas gusto mong manood na lang  ng TV o matulog. Pupunta ka lang Pag tumawag na siya at nagmamakaawang dalawin mo  naman siya at me suhol na ipinagluto ka niya. Kahit automatic na kotse mo nde mo pa rin makuhang hawakan ang kamay niya habang namamaneho ka. Magastos ang mga monthsary, kung anniversary niyo nga eh wala kang regalo, monthsary pa?!? Mag-uusap kayo? Baket me problema ba? Kung wala, isang oras kang manonood ng TV habang siya eh nakatutulog na kahihintay na kausapin mo siya. Miss na niyang ini-email mo siya,sagot mo? \"Jusko naman araw-araw na Tayong nag-uusap ano pa ba naman ang sasabihin ko sa yo? Baka gusto mo pa ng card?!?\" Pag aayain ka niya para gumimik with her friends ang sagot mo? \"Utang na loob, kung gusto mong lumabas kasama friends mo ikaw na lang mag-isa at naiilang ako.\" Pero pag lalabas kayo with your friends umiinit ulo mo pag tahimik siya pag nagjo-joke sila tungkol sa inyo ng ex mo sabay tanong \"Nde ka ba nag-eenjoy? Buiset, umuwi na nga lang tayo!\" At tungkol naman sa kasal...\"Jusko naman, nde mo ba ako maintindihan?!? Wala pa akong pera saka nde ko maiisip yan ngayon! Sana wag ka namang makulit.\" Sa lahat ng ito, ngingiti na lang siya, iisipin lahat ng ginagawa mo noong nanliligaw ka pa  malamang umaasa pa yun na babalik yun dati sabay buntung-hininga at sabi ng malakas...\"Kay sarap ng in-love.\"  '),(16369,2839449,'18,June,2004','2004-06-18 00:00:00','   Join us to stop the relocation of undergraduate courses from Southampton College to the C.W. Post campus of Long Island University!  This is just the beginning of the movement to help Save Southampton College!   Please be patient as this blog is updated - send us your feedback!  '),(16370,2839449,'25,June,2004','2004-06-25 00:00:00','   This was the lead article in Dan\'s Papers this week.  A must read!!!  (it\'s long so I won\'t post it all here)   urlLink  What Happens After Southampton College Closes in 2005?    Includes comments from  John Kanas , member of the Board of Trustees,  Dr. Daniel Rodas , Provost of Southampton College, and  Congressman Tim Bishop (Dâ€“NY) . Bishop was for many years the Provost of Southampton College and is currently the Congressman in Washington for New York District-1, which includes Southampton.  '),(16371,2839449,'24,June,2004','2004-06-24 00:00:00','   Good afternoon people.  In the interest of a letter writing campaign, lets tell everyone who opens an envelope or an email what\'s happening at Southampton College.  Below you\'ll find emails and addresses to various papers so you can speak your mind.  Let\'s try to get the word out to every paper or news organization we can think of.  There\'s a story here, and they can\'t close our college quietly.   New York Times:  The New York Times accepts opinion articles on any topic. The suggested length is 650 words, but submissions of any length will be considered. Articles may be sent in any of these ways:  By e-mail to:  oped@nytimes.com   By fax to: (212) 556-4100  Or by mail to:  The Op-Ed Page 229 West 43rd Street New York, NY 10036   To submit an Op-Ed for Long Island:   opedli@nytimes.com    Newsday:  Viewpoints occasionally publishes unsolicited opinion essays. Submit your work of 700-800 words as open text to  oped@newsday.com  Do not send attachments. Please include your name, e-mail and home addresses, as well as work and home phone numbers. An editor will contact you if your article is being considered for publication.    New York Post:   urlLink http://www.nypost.com/postopinion/letters/letters_editor.htm    The Daily News:  LETTERS TO THE EDITOR (â€œVoice of The Peopleâ€)  The Daily News cares what you think! If you wish to express your opinion on any topic by submitting a letter to the editor (\"Voice of the People\") for publication, please include your full name, address and phone number. (This information will be used for verification purposes only; failure to include the information invalidates your request.) The Daily News reserves the right to edit letters. The shorter the letter, the better the chance it will be used. Then send your letter to  voicers@edit.nydailynews.com .   Southampton Press:  Got a hot news tip? Know of an interesting story? We would love to hear about it!:  skotz@southamptonpress.com  Feeling strongly about something you read in the paper this week? Let the editor know what\'s on your mind:  joeshaw@southamptonpress.com     Also, you may want to try contacting other local papers that may be interested in covering the meeting on 7/7.  I will try to post info of the following papers this weekend:  East Hampton Star Southampton Independent East Hampton Independent Dan\'s Papers     We\'ve also thought of making tee shirts to wear to the meeting.   Up soon will be word format designs that you can print onto iron on paper (available at most  urlLink Staples , etc.) to make your own shirts to support Southampton College.  Talk about grassroots!   Please email or post a comment with your own tee shirt layout or slogan idea.  We\'d love to hear/see them!   Some slogan ideas, besides \"Save Southampton College!\" so far (thanks Sam) are:   \"Save the Students Who Save the Whales!\"      \"Choose Students Over Profits!\"     \"Hell No, We Won\'t Go!\"    So let\'s get out there and let everyone hear our voices.  We love Southampton College and closing it down is taking away a fabulous resource to the community.  I\'m not ready to see it go, and I want to know  WHY , don\'t you?  There.  Not very professional, but from the heart.   We want to hear from you, and we don\'t bite...hard.  ::wink::  '),(16372,2839449,'24,June,2004','2004-06-24 00:00:00','   THEY are  urlLink the LIU Board of Trustees . Write your letters to them at the following address:   University Center Long Island University 700 Northern Boulevard Brookville, NY 11548-1327 (516) 299-1926    Here\'s a list of who THEY are:   LONG ISLAND UNIVERSITY BOARD OF TRUSTEES CHAIR   Roger Tilles , Director, The Tilles Investment Company     SECRETARY   Steven J. Kumble Hâ€™90 , Chairman, Lincolnshire Management, Ltd.     EXECUTIVE COMMITTEE   Stanley F. Barshay â€™60 , Chairman, Schering-Plough Healthcare Products  Mark A. Boyar â€™65 , President, Mark Boyar & Company  Donald H. Elliott Hâ€™85 , Partner, Hollyer, Brady, Smith & Hines LLP  George L. Engelke, Jr. , Chairman, President & CEO, Astoria Financial Corporation and Astoria Federal Savings and Loan Association  Neil R. Grabois , Vice President and Director for Strategic Planning & Program Coordination, Carnegie Corporation of New York  Msgr. Thomas J. Hartman , Director of Radio and Television, Telecare  Alfred R. Kahn â€™84 , Chairman & CEO, 4Kids Entertainment Inc.  Eric Krasnoff , Chairman & CEO, Pall Corporation  Theresa Mall Mullarkey , Former President, Mall Associates; Chancellor, C.W. Post Campus  David Pinter â€™72 , President & CEO, Zwicker Electric Company, Inc.  Thomas L. Pulling , Managing Director, The Citigroup Private Bank  Robert F.X. Sillerman , Chairman & CEO, The Sillerman Companies; Chancellor, Southampton College  Ronald J. Sylvestri â€™66 , Senior Vice President, Hudson Valley Bank; Chancellor, Regional Campuses  Edward Travaglianti , Former Chairman & CEO, EAB; Chancellor, Brooklyn Campus  Rosalind P. Walter Hâ€™83 , Board Member, Channel 13/WNET   MEMBERS   Sidney Braginsky , Former President, Olympus America Inc.  Jerry L. Cohen , Tishman Speyer Properties LP  Anne W. Ellis , Former Chairperson, Channel 21/WLIW  Michael N. Emmerman â€™67 , Managing Director, Neuberger Bergman, LLC  Les Goodstein , President and COO, New York Daily News  John A. Kanas â€™68 , Chairman, President & CEO, North Fork Bank  Steven A. Klar , President, The Klar Organization  Howard M. Lorber â€™70 , â€™91, Hâ€™01, Chairman, Hallman & Lorber Associates, Inc.  William Lynch, Jr. , President, Bill Lynch Associates  Michael Melnicke , Operator/Administrator, Rockaway Care Center  William R. Nuti â€™86 , President & COO, Symbol Technologies, Inc.  Lester J. Owens â€™79 , Managing Director, Deutsche Bankers Trust  Bruce C. Ratner , President & CEO, Forest City Ratner Companies   Warren L. Schwerin , President, Related Properties Corporation  Terry Semel â€™64 , Chairman & CEO, Yahoo! Inc.  Edward E. Shorin Hâ€™99 , Former Officer and Director, Topps Chewing Gum  Harvey Simpson , Retired Vice Chairman, NAB Construction  Marvin M. Sirota â€™62 , President & Chairman, Interactive Health Management, Inc.  Howard M. Smith , Chief Operating Officer, Bulbtronics  The Rev. Dr. Paul Smith , Senior Minister, First Presbyterian Church, Brooklyn Heights, NY  Martin L. Sperber , Retired Chairman & CEO, Schein Pharmaceutical  Dan K. Wassong Hâ€™92 , Chairman, President & CEO, Del Laboratories, Inc.  William Lie Zeckendorf , Co-Chairman, Zeckendorf Realty, L.P.    TRUSTEES EMERITI   William F. de Neergaard â€™47, Hâ€™98 Arthur I. Goldberg â€™53 Thomas Russell Jones Abraham Krasnoff Hâ€™85 Eugene H. Luntey Hâ€™98 John M. May Harry Shechtman â€™33, Hâ€™99 Robert A. Smith Hâ€™96 Richard Stark William Zeckendorf, Jr. Hâ€™92    EX OFFICIO   David J. Steinberg     H -indicates honorary doctorate from Long Island University  '),(16373,2839449,'24,June,2004','2004-06-24 00:00:00','   The following was adopted last Spring by the College as an \"identity statement.\" Thanks to  urlLink Professor Scott Carlin  for posting this.  Southampton College, Long Island University\'s small liberal arts and sciences college, is dedicated to creating an environmentally sustainable and humane world through critical thinking, creativity, and global citizenship. The curriculum emphasizes academic excellence, inter-disciplinary, and experience-based learning, and takes full advantage of Long Island\'s marine, environmental, literary, and artistic wealth. Our mission is to empower students to care for the world.    June 26th  - In response to this mission statement,  Jennifer Rogers, environmental studies major, Southampton Class of 2000 , wrote: ( changes in blue )  It should be tweaked a bit if it is a cooperative model we\'re trying to implement, no?   Southampton College,   (a unique post-secondary education center and eco-village located on Long Island\'s East End, is dedicated to creating an environmentally sustainable and human world through critical thinking {thank you environmental thought class ;) }, creativity, and global citizenship.) Chances are SC will no longer be affiliated with LIU?    The curriculum emphasizes academic excellence,  community partnership,  experience-based learning, and takes full advantage of Long Island\'s  environmental, marine,  literary, and artistic wealth. Our mission is to empower students to care for the world.  '),(16374,2839449,'24,June,2004','2004-06-24 00:00:00','   excerpt of an  urlLink Independent  article by Carey London  urlLink LIU To Close Southampton Campus; Students Rally Together     Student Activism   Many students are not going out without a fight -- a \"positive\" one, that is. On Friday afternoon, about 20 students gathered in the Avram Theatre for the first time to share ideas on how to raise money to save their school. Student organizers Joseph Brondo and Carly Kenkel stood before their peers and tried to rally the troops.  \"It\'s us who have the emotional ties to the education of our future,\" said Kenkel, who emphasized that they must run a positive campaign to show that the students are not resentful, but rather passionate about continuing their education at Southampton. The focus must be on the \"reason we need to save this college, not why it\'s falling apart,\" she said.   Organized only the night before the rally, the group has just begun to seek financial assistance from a variety of corporations. \"We\'re trying to find money wherever we can,\" said Brondo. Two companies they have recently tapped for funds are McDonald\'s and Upromise, an organization that encourages businesses to donate a portion of their funds to their customers\' Upromise college accounts. If possible, the students will embark on a nation-wide campaign and even seek out television personalities.  \"We\'re going to write to Oprah,\" said Kenkel, among chuckles, \"anything to show that we care.\"   Recommendations within the group were few and far between, but some suggested a \"grassroots\" style campaign, whereby students compose a general letter, or petition, and mail it out to parents, teachers, and community members to sign and return to LIU officials. However, Kenkel said, they have no access to home addresses and can only use email at the moment.   Brondo, a communications major with a focus on theory and film, said that one reason he is so dedicated to saving the Southampton campus is that he \"fell in love with this school in the past five months I\'ve been here.\" And the response this group has received from faculty, administration, and alumni has been overwhelmingly positive, he reported.   \"We\'ve been getting so many calls\" from alumni students who want to help save the college, said Kenkel, who majors in marine biology and just finished her freshman year at Southampton. \"Don\'t give up hope.\" But by this same token, she also advised students to prepare for the school\'s closing. \"Figure out what you\'re going to do if this doesn\'t work,\" she said. \"I\'m definitely applying to other colleges, but I\'m putting my heart into this.\"   The students\' next meeting will be Tuesday at Avram Theatre, at 7 p.m.  '),(16375,2839449,'24,June,2004','2004-06-24 00:00:00','   urlLink Newsday article:A fond farewell to Southampton College  Written by SANDY McINTOSH Sandy McIntosh, a 1970 graduate of Southampton College and a former adjunct faculty member, is a poet whose latest book is \"Between Earth and Sky.\" He is the managing editor of Marsh Hawk Press.  June 24, 2004   The founding of Southampton College by Long Island University in 1963 encouraged many expectations. Local businesses hoped for an influx of money and jobs to help in the slack winter months. Cultural promoters looked forward to a new audience of intellectually curious students and faculty. And Long Island University itself anticipated...what?  Begun in 1886 as the Brooklyn College of Pharmacy, LIU had always promoted a service-oriented curriculum. What it planned for Southampton was a college meeting the needs of local families. What it got, however, was something else.  Almost by accident, during the 1960s, Southampton began to look and act like a northern version of Black Mountain, the experimental North Carolina college that for a short time had been a hub of American cultural production, mainly because of famous artists and writers who taught there. Black Mountain was long gone by 1963, but the willingness of accomplished artists and writers in the Hamptons, happy to be asked to teach, gave its memory new life. (In fact, several artists and writers from the community had taught at Black Mountain.)  It wasn\'t entirely the case that the college sought out these prominent artists. In the early years, departments continually scrambled for instructors, and were glad to hire anyone competent. So, the Abstract Expressionist painter Willem De Kooning was often seen in the halls of the Fine Arts building. The representational painter Fairfield Porter taught an undergraduate painting class, and the Neo-Plasticist painter and filmmaker Ilya Bolotowsky, with his thick Russian accent, could be found wrestling freshman English students through James Joyce.  Prominent writers, too, led classes. David Ignatow, winner of the Bollingen Prize and of a National Institute of Arts and Letters award for a lifetime of creative effort, patiently endured my fumbling first attempts at writing a poem.  Why did these artists happily submit to the chaos of a college in its formative years? Perhaps they were lonely.  In the \'60s, the Hamptons\' tourist season ended abruptly at Labor Day, the streets were deserted, and you might find a coffee shop open on Main Street, but not after 8 p.m. If you were marooned at the college in fall or winter, you could walk the empty streets in near-darkness to the ocean, where the crashing of waves was truly primitive and frightening. It was beneficial to meet colleagues in the halls of a college, rather than always on the stools of a bar.  There was much chaos in the early years, and not only the creative kind. There were constant rumors of faculty or staff misadventures, or about money that had gone missing, such as the gift of a wealthy local resident for the installation of ornate entrance gates. Nevertheless, the college did mature, if not in the image of the other LIU campuses, then in its local strengths.  The announcement that Southampton College will transfer its undergraduate programs to the C.W. Post Campus next year-effectively closing its doors-saddens those of us who share its history. But according to David Steinberg, LIU president, the college has accumulated several million dollars of debt. In the past this debt would have been addressable, but the other branches of the university have their own negative balances-notably C.W. Post, due to last year\'s faculty strike-and Southampton\'s is the last straw.  So, sometime in the near future, the university will consider selling the valuable campus real estate to a developer, who\'ll no doubt build a new Levittown of identical mansions for multi-millionaires.  The last students will have a lonely time, if a short one. Without the noise and lights of a bustling campus they\'ll walk the empty paths, where the crashing of waves from the nearby ocean will once again be heard, primitive and frightening.  '),(16376,2839449,'24,June,2004','2004-06-24 00:00:00','   \"With the terrific infrastructure that has been created for natural and marine science programs at Southampton, it would be wasteful to relocate those divisions to CW Post, which lacks waterfront access or adequate laboratory space for such a proposal. My question is what changes would be proposed to ensure that the marine and environmental programs benefit from an inland educational facility, and what would the University propose to do with the existing lab and marine space at Southampton?\"  - Shawn Kiernan, Southampton College Class of 1999     Dear President Steinberg,   As a 1991 graduate of LIU Southampton I was shocked and disappointed to hear of the universityâ€™s decision to relocate the undergraduate programs from Southampton to CW Post. Like other alumni, Southampton College means a great deal to me. Not only did I receive a top-notch education, complimented by co-ops and internships, I also met and married my wife, Kathy (also a 1991 graduate) on campus.  While I cannot deny an obvious emotional attachment to the Southampton Campus, my dismay at the universityâ€™s decision emanates equally from the lack of detailed information coming from the administration.     Here are just a few of my questions:  1. Didnâ€™t the University just spend about $60,000,000 on renovations to the campus? Why make such a large expenditure if the plan was to significantly cull the campus population?   2. How will funds generated from a smaller number of graduate students make up for funds generated from a larger number of undergraduate students? Research grants? Significantly higher graduate tuitions?      3. Does LIU have enough graduate programs to justify keeping the Southampton College open?   4. Will graduate programs from CW Post and other campuses be transferred to Southampton?   5. Is this a precursor to the university selling the land on which the Southampton Campus is based?   6. What about WPBX? I understand that the radio station will still broadcast from the campus. How will this be accomplished when many of the staff members are undergraduate students?   7. How does the university intend on continuing the Marine Science program?    8. How will the university make up the funds for lost alumni support? After all, what is our motivation to donate money to an institution that has eliminated our attachment to it?    9. Why didnâ€™t we hear about the campuses financial difficulties earlier? Was the money mismanaged? Were there some creative accounting practices in place?   10. How will the university stem the tide of departing students? Undergraduate Southampton students clearly made a decision to attend a small semi-rural campus. In all probability many of them will transfer to other schools. Conversely, what about students at other LIU Campuses? They clearly made a decision to attend suburban campuses. If forced to travel to Southampton, they may leave in droves as well.   I, and the rest of the Southampton Campus Alumni, eagerly await your answers to these and similar questions.    - Scott Watson, BAHistory/Political Science, Southampton College Class of 1991     Email your letters, stories, photos, questions - anything and everything to me!   scogar@alumni.liu.edu   Pass this blog address onto all your alumni friends!   http://savesouthamptoncollege.blogspot.com  '),(16377,2839449,'24,June,2004','2004-06-24 00:00:00','   -----Original Message----- From: Louis Antoniello [mailto:lantoniello@licb.com] Sent: Friday, June 25, 2004 8:52 AM Subject: RE: JUNE 24th Meeting @ SC  Hello All:  Joan is correct, the best thing we can do prior to the July 7th meeting is to write, write, write.Send your e-mails/letters to David Steinberg, the press, etc.  I have been contacted by  \"CBS Investigative Reporting\" ,  they believe that there may be a story as far as the management of funds at LIU over the years . I have provided them my opinion and invited them to the meeting on 7/7 at 7, if you know anyone in the local newspapers bring them along. This is our opportunity to let the administration at the University know that we do not appreciate they way they have treated SHC for the past 30 years and we believe that they have made the wrong choice and should have never made the public statement pertaining to SHC and its undergrad classes.  I myself will have some very direct questions that I will ask asked, they need to be answered and they will be. I believe that after the Q & A the press will realize that LIU was the reason for financial problems not SHC.  I Look forward to seeing you on 7/7 at 7.  Enjoy your day.  LA  '),(16378,2839449,'24,June,2004','2004-06-24 00:00:00','   -----Original Message----- From: JoanTuttSeashell@aol.com Sent: Friday, June 25, 2004 2:50 AM Subject: JUNE 24th Meeting @ SC  Dear Alumni Friends:  The meeting today lasted over 2 hours. I am emotionally spent. It was obvious to me, that the faculty was unaware about the changes that were about to take place.  Many people spoke and there was an exchange of ideas. The meeting was respectful and never became a shouting match.  Amongst the faculty present that you may recall were Dr. Don Baker and Bob Gerbereaux from the library.  Genie Kornhaber, \'68, and I were the oldest in attendance; not by age, but by years from the beginning of SC. We both expressed the undying and unwavering love and support we are receiving from the alumni of \'67,\'68, \'69 & \'70. Those in attendance: faculty, students & staff, were very appreciative and maybe a little surprised at our fervor and commitment. I signed a list to be considered for a future committee.  There were no administration present, so this was not a Q & A session. On the contrary, it was a forum to discuss concerns and methods of making our voices heard and listened to. Suggested on several occasions was  a letter writing campaign . Many of you have written me from CA & FL & points in between wanting to help, but not knowing what to do from afar. WRITE!!!!    Write to your local papers.....write letters to the editors......write poems....write anything you choose, but just write! Those in NY & LI.....get your letters into Newsday & The NY Times, as well as your local papers. Some alums are involved with the media....go to it!!! Do your thing!!!    When you do, make a copy of your letter and where and when it was published (i.e. July 5, 2004, The Miami Herald by Jane or John Doe) and send it to Dr. Steinberg and the Board of Trustees. And if you want, send one to me and I\'ll keep a file of all of them for committee use. \"The power of the pen is mightier than the sword!\"   Below is a letter from Don Getz, \'68, Marine Science Station Staff. After the meeting was over reporters interviewed and took pictures. Newsday said there was to be two articles; one Friday and one this weekend. I\'ll be forwarding to you a Newsday piece by Mr. Sandy McIntoch, \'70.  I am too pooped to go into any more details at this time.  Mark your calendars for an ALUMNI meeting with Dr. Steinberg, on Wednesday, July 7 @ 7PM @ Chancellor\'s Hall. Remember: 7/7 @ 7 !   I also have a call into The Southampton Chamber of Commerce, of which I am a member. Wondering where they see themselves in all of this controversy. Needless to say, the local community has benefited substantially from SC\'s presence!  Hope to see you 7/7 @ 7! Maybe afterward we can all go out to  Barristers ,  The Publick House , or  Armand\'s .  ~Joan DiPaola Tutt, \'68, \'74   In a message dated 6/24/2004 8:09:37 PM Eastern Daylight Time, getzaway@optonline.net writes:  Hi Joan Today a meeting was held at the Avram theater chaired by Scott Carlin. The meeting was open to the entire college community as a forum to prevent the relocation of our college. Press members were asked to leave in that this was a private meeting. Ideas were exchanged and I believe a committee will be formed in an effort to affix blame for Southampton\'s financial problems and the university as a whole. Ideas express would involve the college and the surrounding community.  Don  '),(16379,2839449,'24,June,2004','2004-06-24 00:00:00','   I may or may not be able to attend  urlLink the meeting  on July 7th with Dr. Steinberg and Dr. Rodas, but I have thought of some questions to ask:  What will happen to the faculty at Southampton who will not be able to commute to C. W. Post on a daily basis to teach classes?   Will there be jobs at C. W. Post or other campuses for the staff? (administrative professionals, food service, public safety officers, physical plant, etc.)  What will happen to the staff at Southampton who are close to retirement? Will the University honor pensions and/or retirement plans? Or will the staff be forced to commute? Or will there not be any work for them at the other campuses?  What will be done with all the money raised at this year\'s  urlLink All For The Sea concert ? What happened to the donation I sent to the college earlier this year?  How can you justify closing a school whose marine and environmental sciences programs have produced  34 Fulbright Scholars in the past 28 years ?   If the school is closing for financial reasons that have been lingering for years, what was done to try and stop it? What failed? What was not tried? Whose fault is it that the college was so poorly managed at the financial level for so many years?  How long will the college stay open with only graduate courses available on campus?  urlLink To quote ,  \"The University has determined that the expenses of rebuilding Southampton College are cost-prohibitive and not financially sustainable.\"  If you can\'t keep the college open and making ends meet with undergraduate tuition, how do you plan on staying open with limited monies from a small number of graduate students\' tuitions?  What will happen to the campus? Will it be sold to the highest bidder? Will it be donated to some East End charity so the University gets a big tax break? Will it be turned into a hotel/resort for events like the U.S. Open and Parish Art Museum gatherings?  (sorry - my cynicism and knowledge of East End motivations got the best of me on this one)    What questions would you like to ask President Steinberg and Provost Rodas?    If you cannot attend the meeting, please email your questions and comments to Carol.gilbert@liu.edu or call the Alumni Office at 632-287-8348. Send a copy to me, too, and I\'ll post them here.  '),(16380,2839449,'23,June,2004','2004-06-23 00:00:00','   Would you like to be a team member on this blog and help post important info, stories, and photos? Blogging is easy - even if you\'ve never used  urlLink Blogger  before. As interest from alumni and others across the country grows, it would be ideal to have a group of people to post on the blog. Email me if you\'re interested (scogar@alumni.liu.edu).  '),(16381,2839449,'23,June,2004','2004-06-23 00:00:00','   From the  urlLink Southampton Press :  Southampton Collegeâ€™s most prestigious undergraduate program, the nationally recognized Marine Science program, could continue at the campus via a partnership with the State University of New York at Stony Brook, Long Island University officials said this week.  urlLink Full story...    I don\'t have an online subscription to the Press. If you do, please copy the story and email it to me so I can post it.  '),(16382,2839449,'23,June,2004','2004-06-23 00:00:00','   Hello! Sorry for the introductions to be a bit late.  My name is Samantha Cogar, an alumni of Southampton College, and I served as President for the Class of 2000.  Like many of you, I was shocked and saddened to hear of the decision to relocate undergraduate courses from Southampton College to the C. W. Post campus due to financial reasons. Like many of you, I wondered if there was anything I could do to stop it from actually happening. I do not live close to the college (currently I\'m in the Washington, DC metropolitan area), so starting this blog is my attempt to unite alumni, staff, administration and current/future students. There are some links on the side bar and I\'ve tried to post any information sent to me from the college (usually via email).  But to make this blog truly useful, I need to hear from you.     Send me an email (scogar@alumni.liu.edu) with any questions, comments, stories, information, links, photos, etc. Please include your name, email address, and role at Southampton (class/years attended and/or job title) and put Save Southampton College in the subject line (so I don\'t think it\'s spam and delete it).    Some questions you may want to address:  How do you feel about the relocation of classes? Do you have any ideas of how to stop the relocation? What made Southampton College such a great place to go to school? What are your fondest memories of Southampton College?  '),(16383,2839449,'23,June,2004','2004-06-23 00:00:00','   The meeting with President Steinberg and Provost Rodas will take place as described below. If there are alumni interested in car-pooling to this event, email me ( scogar@alumni.liu.edu ) with the following info:   1. your name  2. email address 3. where you live 4. day to depart to Southampton 5. day to depart back home   I\'ll post a list of those interested in car pooling to the meeting here on the blog so people can network to get there.  For cheap, last minute flights, I would suggest  urlLink Southwest Airlines  that flies to Islip.  urlLink JetBlue  flies to JFK and  urlLink Spirit Airlines  flies to LaGuardia.  Anyone interested in flying to NYC or Islip and car pooling from those airports and anyone interested in picking up people at those airports should also email me.  Duke Lecture Hall isn\'t that big - Let\'s make it STANDING ROOM ONLY and show our support for Southampton College!  '),(16384,2839449,'23,June,2004','2004-06-23 00:00:00','   TO: Southampton College Alumni  FROM: Greg Connly, Class of 1985   RE: MEETING WITH DR. DAVID STEINBERG President, Long Island University and DR. DANIEL RODAS, Provost, Southampton College    Wednesday, July 7, 2004 at 7 :00 p.m.  in Duke Lecture Hall, Chancellors Hall, Southampton College, Southampton, NY  This meeting will provide Southampton alumni with a forum to discuss the changes occurring at Southampton College with President David Steinberg and Provost Daniel Rodas. Members of the Southampton College Alumni Association Board will be present. All Southampton alumni are encouraged to attend. If you cannot attend the meeting, please email your questions and comments to Carol.gilbert@liu.edu or call the Alumni Office at 632-287-8348.  '),(16385,2839449,'23,June,2004','2004-06-23 00:00:00','   To: The University Community  From: David Steinberg, President  Date: June 16, 2004  Subject:  urlLink Relocation of Southampton College Undergraduate Programs   Yesterday the Universityâ€™s Board of Trustees instructed the University Officers to relocate the undergraduate liberal arts and sciences programs currently offered at Southampton College onto the C.W. Post Campus, effective September 2005. The University will continue to offer its outstanding graduate programs at Southampton. It is with a heavy heart that I write to inform you of this painful but necessary decision, one in which the University Officers unanimously concur. In response to the Boardâ€™s unequivocal instructions, the administration now must move thoughtfully to bring this decision to fruition.   In the fiscal year ending August 2004, the Southampton campus is anticipated to lose $9 million. The anticipated deficit for the 2004/05 fiscal year is projected at greater than $12 million. The cumulative deficit since inception at Southampton will have reached an unacceptable $77 million by September, 2005. Despite the extraordinary efforts of a great many people on that campus and the embrace of generous alumni, overseers, trustees and supportive citizens on the East End, hard fiscal realities make a continuation of high quality but very expensive undergraduate instruction at Southampton beyond the reach of this University.   For more than four decades, Long Island University has sought to sustain three discrete residential campuses, each functioning in its own environment and developing its own character. The effort to create at Southampton an academically excellent, financially self-sufficient, world-class, small college has been a labor of love for countless individuals over many years. Ultimately, however, the Universityâ€™s primary obligation always must be to its current students -- both those who study at Southampton College and, equally, those who study elsewhere in the University.   Several years ago, after an exhaustive review and analysis, a special committee of the Board concluded that maintaining the status quo at Southampton was no longer a viable choice. The Board instructed the administration to draft a Master Plan, to redesign a dynamic core curriculum, to begin a capital construction effort and to launch a fundraising campaign. What could not be measured then was the timing and size of future enrollment increases as the result of new curricula and facilities. Despite a modest upturn in potential freshmen for the upcoming year, the required enrollment growth necessary to reduce significantly Southamptonâ€™s operating deficits is not foreseeable. Consequently, the loss of tuition dollars has created deficits beyond the Universityâ€™s ability to stay the marathon course. The Southampton deficit must not jeopardize the health of the entire University.  The dynamic new core curriculum for freshmen as well as the existing courses required for returning students will be taught at the Southampton campus next year and at C.W. Post thereafter. The University is also in active discussions about how best to meet the needs of those students enrolled in marine science after next year. We intend to work with every student to help each one meet his or her educational goals. Every Southampton student will be given a careful explanation of his or her options, a chance to visit C.W. Post or Brooklyn, and an opportunity to receive professional help in any efforts to transfer to other colleges, if that is their choice.  Over the next days and weeks, the University will communicate with the many Southampton stakeholders through detailed memos, personal meetings and individual sessions. Faculty, staff and administrators will receive an explanation of the implications of this decision. At every appropriate juncture, timely information and a comprehensive array of services will be provided for Southampton students and staff to help them through this transition.  The Universityâ€™s fundamental commitment to provide each student a superior education and to treat every student, teacher and employee with genuine care, concern and respect will not waiver. This Board decision will affect individual human lives, each with a unique set of circumstances, cares and concerns. Many at Southampton College have labored long and hard in pursuit of a dream that, sadly, now will not happen there. Although it pains me and my colleagues deeply, the Boardâ€™s decision sets the parameters of what the University must do now to achieve its potential in the decades to come. In that spirit, I charge all of us to transform this disappointment into a renewed commitment to Long Island University.  '),(16386,2839449,'02,July,2004','2004-07-02 00:00:00','   The Meeting with President Steinberg and Provost Rodas   WEDNESDAY, JULY 7th at 7:00PM  MOVED to THE AVRAM THEATER     If you cannot attend the meeting, please email your questions and comments to Carol.gilbert@liu.edu or call the Alumni Office at 631-287-8348.    Next Save SC Meeting Date  THURSDAY, JULY 8 at 7:00PM in the Avram Theater     For more information, please contact Nora (ndetweiler2002@yahoo.com)   Everyone and Anyone is encouraged to attend BOTH meetings!  '),(16387,2839449,'01,July,2004','2004-07-01 00:00:00','   Thanks to   Catherine Baker   for sending the entire text of the article.    By Donald G. Baker  Issue Date: Southampton Press 07/01/04  The Southampton communityâ€”indeed, the entire East End of Long Islandâ€”has a legitimate right to be outraged by the Long Island University administrationâ€™s abrupt and unilateral decision to close its Southampton College undergraduate program.  For years LIU leaders acknowledged but took few actions to correct the collegeâ€™s economic problems (yearly losses of approximately $2 million have continued for 16 or more years, but only in the last few years has the LIU administration addressed the issue), but they did not discuss in advance with the community or the college community (faculty, staff, students, or even the alumni advisory board of the college) that the school would be closed. Indeed, the LIU administrationâ€™s announcements over the past year that it was engaged in a $60 million renovation and its proud announcement that Chancellor Robert F.X. Sillerman was donating millions for the new library (now being built) raises serious questions about how candid that administration has been with the collegeâ€™s constituent groups and the Southampton (and larger East End) community.  Only last week did the university announce that it was closing Southampton College (it claims it is only â€œmovingâ€ the college to its C.W. Post campus), yet in numerous recent newspaper stories it appears that even prior to that public closing announcement the administration has been involved in â€œintensiveâ€ discussions with the State University of New York at Stony Brook about moving the collegeâ€™s marine science program there. This information was never shared with the college community.  Moreover, the administration now admits that it is already exploring possibilities for selling off the college propertyâ€”without first considering and discussing with all involved and affected college and community groups the broader environmental and economic implications and impact on these parties and the East End itself.  This raises a series of legal and related issues that perhaps need to be examined:  1. Does Long Island University have the legal right to sell the Southampton College property? LIU did not buy the propertyâ€”the property was bought by an East End citizens group, desirous of establishing an undergraduate college. It explored that possibility with LIU, which said it would establish an undergraduate college only if the local community purchased the property. The committee purchased the property and granted title to LIU specifically for and contingent upon the establishment and continuation of an undergraduate college on the purchased property. Otherwise the group would have looked for alternative ways of establishing an undergraduate college.  For whatever reasons, LIU has now decided it can no longer operate an undergraduate college on that property. Hence the question: Does this action abrogate LIUâ€™s right to that property? The East End groupâ€™s decisions and motivations are clearly indicated in newspaper accounts from 1962 and 1963 (and acknowledged in various publications, including a LIU booklet celebrating the universityâ€™s 75th anniversary) that clearly state the property was, after being bought by the local committee, donated to LIU solely for the purpose of establishing and maintaining an undergraduate college.  Thus a specific question and possible legal issue can be raised: Has LIU, by its decision to terminate the undergraduate college on this property, forfeited its claim to the property, whatever the title to that property may state? The communityâ€™s intent for offering LIU the property solely for the purpose of establishing and maintaining an undergraduate college is clear.  Is there, therefore, a legal basis for questioning LIUâ€™s retention of the property, at least of the property itself and the original buildings on that property? Does the community, at a minimum, have a vested right to the property itself and the original buildings?  That local community committee, it should be noted, raised $325,000 to purchase the property, and then raised another $50,000 for needed repairs. Only after this, as newspaper reports indicate, was the property turned over to LIU, which agreed to establish and maintain an undergraduate college.  LIU, it can be argued, is now reneging on that agreement, and the basic legal question is that by doing so, has LIU forfeited its right to the original property and buildings, its rightful claim only for the other buildings erected on the property? Is there a legal basis for the town/community reclaiming the property, since LIU has stated it will no longer live up to that agreement to maintain an undergraduate college on the property?  2. There is a possible second legal issue. Chancellor Robert F.X. Sillerman generously, even after numerous earlier large gifts that kept the college afloat, subsequently contributed millions of dollars for building a new library, his donation based on the administrationâ€™s public statements that it would invest possibly $60 million for the collegeâ€™s renaissanceâ€”only to abruptly decide to close the college. Given the circumstances, and the board of trusteesâ€™ statements about the purported large infusion of funds, was the chancellor himself misled in his decision to provide grants for the new libraryâ€”and does he have legal recourse to be reimbursed by the university?  3. The designated college closing raises yet another issue. The New York State Board of Regents is involved in the establishment and evaluation of academic institutions in the state. This is evident in the earlier case of Adelphi University where the Regents investigated the mismanagement of Adelphiâ€™s president and the universityâ€™s board of trusteesâ€™ questionable or non-fulfillment of its fiduciary responsibilities in overseeing the economic and managerial practices of the institutionâ€™s president. The Regents removed both, acknowledging its oversight responsibilities in scrutinizing how institutions of higher education are operated and managed.  Thus, in such cases as the closing of colleges, or the economic hemorrhaging of $2 million yearly at Southampton College for 16-plus years, surely that Board of Regents has a responsibility to probe the management of Long Island University and the oversight responsibilities of the LIU Board of Trustees, particularly since Southamptonâ€™s ongoing economic problems appear to extend beyond the confines of Southampton College to the larger university itself.  4. A final legal issue: The LIU administration appears to be hastily seeking to sell off the Southampton College property, but does it or the community own the property? We require for all (usually new) major property developments an Environmental Impact Statement, hence a basic question here is: before the property is sold, shouldnâ€™t the proper authorities require an environmental (including an economic) impact statement that addresses the question(s) of the economic and environmental impact on eastern Long Island in general and more specifically of the collegeâ€™s termination of employment for one of the largest groups of employees in eastern Long Island?  Donald G. Baker has been at Southampton College since 1964, when he became professor of political science and history and founding director of the Social Science Division.  '),(16388,2839449,'01,July,2004','2004-07-01 00:00:00','   Commencement ... The Cruise To Nowhere ... The Beach ... SeaMester ... Freshman Orientation ... The Senior Dinner ... Internship/Co-Ops ... Club Activities ...   If you have photos from your years at Southampton College that you would like posted here on the blog, scan them and email them to me. Please include your name, years at Southampton, what the picture is of, who is in the picture, and what special memory of Southampton College the photo brings back to you.  '),(16389,2839449,'01,July,2004','2004-07-01 00:00:00','   On the side bar you\'ll find links to newspapers that will cover Southampton College news. Let me know if I forgot any. Most of the East End papers come out once a week, between Wednesday and Friday. Check out the front page news as well as the editorial and opinion pages for new information - or letters from concerned students, alumni, staff, and faculty - like you! Why don\'t you write a letter, too? Or send me your email address and I\'ll add you as a contributor to this blog. Check the side bar - We\'ve got new members already! And thanks to everyone who ordered t-shirts (yet another part of the side bar)! Wear them with pride! And send me ideas and suggestions for items you\'d like to buy!  '),(16390,2839449,'01,July,2004','2004-07-01 00:00:00','   Found this Newsday article online (from June 9th, 2004):  urlLink University agrees to $112G fine in waste case   Long Island University has agreed to pay about $112,000 in fines for violating waste management regulations by failing to properly store and keep records of hazardous materials at its C.W. Post and Southampton College campuses in a settlement reached last week with the Environmental Protection Agency.  The university also must establish 10 environmental management seminars for high school officials in New York and New Jersey about how to comply with regulations governing toxic substances.  Terry Ippolito, spokeswoman for the EPA\'s Region 2 office in Manhattan, said the seminar requirement, which will cost at least $115,000, stems from the agency\'s belief \"that some of the safety concerns that we have at colleges and universities could very well be evident if one were to do an inspection of high schools ...\" ... She said the EPA inspections dated back to November 2001 at Southampton, and a short time later at C.W. Post and at LIU\'s Brooklyn campus - where the university reached an earlier settlement with the EPA. The violations that were found included not having correct storage containers and not keeping a record of how much waste was accumulating. The hazardous materials involved, Ippolito said, included solvents used in science labs, vehicle service garages and carpentry and paint shops.  The EPA fined LIU  $39,057  for the chemical waste violations and another  $73,244  penalty for underground petroleum storage tank violations at C.W. Post.  Could there be more environmental problems on the campus that the administration does not want the EPA or anyone else to find out about? Could this fine by the EPA just be the start and the University is worried they will have to spend more and more money to fix problems on campus and to do community outreach programs? Anyone remember having to endure the stench of the septic tanks being pumped in the Montauks and Peconics? What else is out of date and under code at the college? Perhaps these code violations from the EPA is just the tip of the iceberg... Could the fines for admitting there is problems and the cost of repairing them be too much for the university to handle? Are they worried it would further tarnish their reputation? Perhaps selling the land to the highest bidder, making a profit, and washing their hands of anything dirty is the answer...   urlLink Officials mull sale of part of campus    urlLink Levy eyes campus buy   Lots of articles in this week\'s  urlLink Southampton Press  (you\'ll need an online subscription to read them)  What is the real motivation behind this sale? Why now? Why such a drastic decision when all other actions by the University were starting to point towards helping Southampton? There\'s something awry...something doesn\'t fit and something doesn\'t make sense....  What do you think? (email a post to me: scogar@alumni.liu.edu)  '),(16391,2839449,'01,July,2004','2004-07-01 00:00:00','   The opinion piece by  Sandy McIntosh, Class of 1970  and former adjunct faculty member, was reprinted in the  urlLink July 1st issue of Newsday .  Great work, Sandy!   Found in the Suffolk Life Newspaper, June 23, 2004:  urlLink S\'hampton College Will Lose Undergrad Programs In 2005   Good news for athletes:  urlLink Displaced Southampton athletes won\'t lose aid    Proof of the type of quality contributions Southampton College makes to the East End community:  urlLink Mercedes Ruehl and Alan Alda Set for Rosenblatt Reading  '),(16392,2839449,'01,July,2004','2004-07-01 00:00:00','   Dear Blogspot Members,  I had sent a letter to Assemblyman Thiele in reference to the \"relocation\" of our college and would like to share his response with you.  Nora Detweiler      Dear Ms. Detweiler,   Thank you for your recent letter regarding the decision of Long Island University to close the undergraduate programs at Southampton College.   As an alumni and elected official who has maintained a close relation with the College through the years, I was shocked to learn of LIU\'s plans for Southampton College. This decision was apparently made in a vacuum without consultation with the many stakeholders who have an interest in the future of Southampton College. I regret that I learned of the decision not from officials at LIU or Southampton College, but from a newspaper reporter.   Of course, my initial reaction was simply anger. Anger because my alma mater was to be closed, but also because of the unprofessional way in which this decision was made. If this decision and the way it was made are any reflection of the way the College has been managed, it is no wonder that the College\'s financial problems continue to worsen.   We were all aware that Southampton College has not been a moneymaker for LIU. Yet, we had hope for the future.  The College had announced a major capital campaign. A new library was under contruction, an agreement had been reached with the Parrish Art Museum for their new facility. A new curriculum was being implemented. In short, there was a vision for the future. The decision of LIU is a betrayal of that vision.   While we are all justifiably angered by the decision of LIU, we cannot dwell on that anger. Rather, we must join together to find a constructive solution to the crisis with which we are now confronted. We have a little over a year before the College closes. We have no time to waste.   I pledge to do all that I can to insure that the Southampton College campus remains devoted to the purposes of higher education.  The College has been so much more to our area than merely an educational institution. Obviously, as one of the region\'s larger employers it is a critical part of the local economy. It has also been a resource for policy makers grappling with critical regional issues like the environment and transportation. Most important it has brought life-changing opportunity to many local residents who otherwise would not have been able to take a class or earn a degree. The College has brought a multiple of tangible and intangible benefits to the people on the East End.   Further, I pledge to fight just as hard to insure that the campus is not sold off as surplus land to be converted to some speculative development scheme. The East End needs a college, not another commercial or condominium project.   I have spoken with many of my fellow elected officials including State Senator Ken LaValle,Congressman Tim Bishop, County Legislator Jay Schneiderman, and Southampton Town Supervisor Patrick Heaney.  We are all committed to keeping higher education at the Southampton College campus.   We are now at the very beginning of exploring the available options to maintaining a college at Southampton. It will take vision and creativity to construct the right solution, but it can be done. We will be exploring the participation of the State University of New York at Southampton, as well as the possible involvement of Suffolk County Community College. There may be other educational institutions who would also be interested in the joint future use of the campus. While much work needs to be done, I am encouraged by the initial reaction of officials and the public to these possibilities.   Please be assured, that I will insist that the public and all stakeholders be very much involved in decisions relating to the future of Southampton College. It is imperative that this is a public process and that all with an interest in the College are heard.   I look forward to working with you to insure that the Southampton College campus is not closed, but is transformed into an institution that will serve our region even better in the future.\"   Sincerely yours, Fred Thiele, Jr. Member of the Assembly    Assemblyman Thiele also attached a copy of the letter he has sent to David Steinberg which is posted below.    Dear President Steinberg,   I write in response to your June 16, 2004 \"Memorandum\" to the University Community announcing the closing of the undergraduate program at Southampton College.   Let me first state my deepest disappointment at this decision, suddenly made by you and the LIU Trustees, without warning to the public or Southampton College stakeholders. While I was certainly aware of the financial challenges confronting the College, I do not believe that you or the Trustees had given the public any indication that the closing of the College was under active consideration.   Quite to the contrary, the Trustees had presented a face of hope to the public about the future of Southampton College. A new library was under construction. A new capital program was being pursued. An agreement had been announced with the Parrish Art Museum for the location of its new facility at the College. A new core curriculum was being implemented. We were all left with the impression that LIU was committed to a new and better Southampton College.   The LIU Trustees have betrayed that commitment. They are now Trustees who are not worthy of our trust.   As devastating as the decision itself was to those who have a stake in Southampton College, even more troubling was the process by which this decision was announced. There was no indication that such a decision was being considered. There was no consultation with persons like myself who are alumni and elected officials. In short, it had all the elements of \"Pearl Harbor\". I was very disppointed to learn of your decision not from a representative of the University or College, but from a local newspaper reporter. If the decision to close Southampton College is any indication of how the College was managed on a day to day basis, it is no wonder the College is in trouble.   In the aftermath of this decision, there have been equally troubling comments in the media about the future of the campus from LIU officials. The focus seems to be on appraisals, property values and selling the campus to the highest bidder.   Let me state to you in the most unequivocal words I can muster. I will use all my authority and effort as a public official to insure that the campus of Southampton College is used in the future for the purposes of higher education. If that is not under the auspices of LIU, then with some other institution such as SUNY to insure that the residents of the East End have access to a four-year college.   Similarly, I will use all of the power of my office to insure that Southampton College campus is not auctioned off to the highest bidder from some speculative real estate project. The East End needs a College, not another condominium project or commercial shopping area.   This reflects the position of every elected official I have spoken with about Southampton College. I hope that LIU will pledge to work with the East End community to to insure that the campus remains devoted to higher education.   Sincerely yours, Fred Thiele, Jr. Member of Assembly    Contact Information (send him a thank you!):   Assemblymember Fred W. Thiele Jr.  2302 Main Street, Box 3062 Bridgehampton, NY 11932 Phone: (631) 537-2583 Email:  thielef@assembly.state.ny.us     And a BIG THANKS TO NORA! Thank you for sending the letter to Assemblymember Thiele and for forwarding his response for everyone to read!  '),(16393,2839449,'30,June,2004','2004-06-30 00:00:00','   Here\'s to alumnus   Genie Kornhaber Weisman  , who\'s letter to Newsday shows she has Southampton Spirit!  Bravo, Genie!    From Newsday\'s  urlLink Letters to the Editor (Long Island Edition)  June 29, 2004   DEGREES OF SEPARATION  A community hopes to save its college from obsolescence  As a graduate of Southampton College, reading \"A fond farewell to Southampton College\" [Opinion,  urlLink June 24 ] sent a chill down my spine. Having returned from a forum of stunned, concerned, but nonetheless determined and enthusiastic teachers, students and alumni, I felt compelled to respond. I believe the fond farewell is premature. We gathered to try to get a grip on the situation. Unfortunately, there were plenty of questions, but no answers. Most of us were incredulous, still shell-shocked.  The timing of the closing announcement is suspect. When, how and by whom were these decisions made? There are lots of smart lawyers, bankers, business people, scientists, teachers, artists and other professionals out there who have Southampton College to thank, not only for fond memories, but in some part, for who they are and where they are in their lives today. We need as many of them as possible to ask the appropriate questions, mull over the answers, then try to come up with some alternatives to just closing down this wonderful institution. It makes me sick to think the property will be sold off for more mansions. Time is of the essence. We certainly can\'t let this happen without a fight.  Members of the Southampton community, friends of the college, as well as alumni, need to support the students, faculty and staff. We need to get involved now. With everyone putting their heads together, we just might be able to turn things around. It\'s worth a try. We need support and positive thinking, not eulogies. It ain\'t over yet!   Genie Kornhaber Weisman  Patchogue  '),(16394,2839449,'29,June,2004','2004-06-29 00:00:00','   A few days ago, I received an email from a former professor of the college. The subject line was  \"give it up\" . The contents of the email are below:  No campaign is going to change a damned thing. This has been on the cards for over a decade and most of the questions being posed are naive in the extreme. I predicted this outcome, in writing, about 6 years ago and was just about 3 months off with regard to the actual announcement date. Faculty are tenured at LIU, not Southampton College and the University is under no obligation to offer them anything more than a job at another campus. Untenured faculty are simply going to have to look for other employment. That\'s how academia works. The College has been running at a multi-million $$ deficit since it opened and, while other schools were growing and improving all their programs, Southampton was resting on its laurels telling everybody they were great and believing their own publicity. This is a business decision and there are many other choices out there for students that are a better expenditure of the same amount of money. I too was president of my class and president of the alumni association for over a decade. I cannot believe this was an easy decision for anyone, but it has been made and the best thing to do now is move on. It\'s not going to be reversed and a lot of energy is going to be expended for nothing. Steinberg and the Board of Trustees will listen politely, answer questons  (sic)  with rhetoric and then go about their business. Why waste the time? Nothing anybody can say or do is going to change the picture. Southampton was never a real priority within the LIU system and it surely isn\'t now. I could write a series of books on the subject, I\"ve got a 34 year history with the place. The best thing people can do is work on a deal for the juniors because they will be hit the hardest. Seniors can graduate, freshmen and sophomores can transfer. Juniors can only transfer a certain number of credits outside LIU and many may not want to go to Post (or Brooklyn). I hope when all this blows over, you can look back and remember the good times you had, the friends you made and keep sight of the fact that your degree is from LIU, not Southampton College. Best of luck in your future endeavors, XXXXXXXXXXXX -------------  Is this email an indication of how this professor mentors and instructs their current students? Give up in the face of opposition? Do not question authority? Do not question the policies given to you? Do not question the choices made for you? When faced with the odds against you, do not work hard to attain your dreams and goals? Just give in. You can\'t do it. You\'ve lost. Don\'t even try.  Is there a need to post the name of this particular former professor of SC? I\'m guessing that is exactly what they would want. Having dealt with this person during my undergraduate years, watching them manipulate students and faculty alike, their motivations for sending me an email are not shrouded in mystery. (And I\'m guessing there are quite a few of you out there who can guess who wrote this email.)  With this particular former professor, it seems it is always a view of winning and losing. Life doesn\'t work that way. Just as there is not an absolute zero, there is no true winner nor a true loser. The winner has sacrificed something along the way to win and the loser is able to walk away from the situation having developed a skill or learned a lesson.   Is there really a chance that we can stop LIU from moving Southampton\'s undergraduate courses to another campus? I\'m not sure. The odds are clearly against us. But I\'m not one to fold my hands in the face of defeat. If I were that type of person, I never would have taken this particular professor on in the realm of campus politics. If I were that type of person, I would not have graduated from Southampton College. If more people at Southampton were that type of person, perhaps this former professor would still be teaching at Southampton. Instead, it is easier for this former professor to trash the college from a distance and toot their own proverbial horn about being right about EVERYTHING - yet their actions do not show this. If this former professor did know where the college was heading, why did they spend 34 years at a college that they saw was doomed to fail?   The question again:  Will we, as the collective group of alumni, students, staff and administration, be able to change the board\'s decision?  It\'s true that it\'s a long shot, but I\'ll be damned if I\'m going to sit by the sidelines and not show support for the college that taught me so much - from lessons in the classroom to life lessons I\'ll never forget.  I\'ll be damned if I\'m going to throw my hands up in the air and say there is nothing I can do.  I can show support for my alma mater. I can show support for the students who have attended and still attend. I can show support for the professors who have worked so diligently and honorably to help their students learn and achieve. I can show support to the campus staff who\'s dedication to Southampton College is deeper and stronger than staff at other colleges.    I will not give up in the face of opposition.   I will not give up on Southampton College.   Who\'s with me?  '),(16395,2839449,'29,June,2004','2004-06-29 00:00:00','   TO: Southampton College Alumni  FROM: Greg Connly, Class of 1985  RE:  MEETING WITH DR. DAVID STEINBERG President, Long Island University and DR. DANIEL RODAS, Provost, Southampton College    Wednesday, July 7, 2004 at 7 :00 p.m. MOVED to the AVRAM THEATER , Fine Arts Building, Southampton College, Southampton, NY  This meeting will provide Southampton alumni with a forum to discuss the changes occurring at Southampton College with President David Steinberg and Provost Daniel Rodas. Members of the Southampton College Alumni Association Board will be present. All Southampton alumni are encouraged to attend.  If you cannot attend the meeting, please email your questions and comments to Carol.gilbert@liu.edu or call the Alumni Office at 631-287-8348.  '),(16396,2839449,'28,June,2004','2004-06-28 00:00:00','   Or at least your opinion on a shirt...  Visit  urlLink this online store  for some t-shirts with Save Southampton College! messages. Save Southampton College t-shirts are sold at base price - no profits being made by me. I\'m looking for more t-shirt ideas, too! Send them in!   Thanks to Tracey Schock, Class of \'99 for her t-shirt idea!   See the comments under  urlLink this entry .  '),(16397,2839449,'28,June,2004','2004-06-28 00:00:00','   Entry from the  urlLink SaveSC Yahoo Group  on June 26th:  The Administration claims that is is the decrease in enrollment that necessitates the closing of Southampton college. That is a fact, but what is ignored are the factors that --in my opinion--precipitated that lower enrollment over the years, namely: (1) a series of bad and often myopic decisions concerning programs, development, ignoring and undermining specific disciplines in order to enhance the one that had \"national recognition\" (i.e., the marine science program), cutting back on faculty in other areas which meant students leaving because of insufficient course offerings in those other disciplines (this is not a criticism of marine science per se, but of the ignoring of the other disciplines that, needing less equipment, laboratories, boats, etc., in reality brought in surplus funds to support the marine science program); (2) an organizational structure, established by the LIU administration that led over the decades to constant conflicts, disagreements or confusion over who was repsonsible for what in terms of the truncated leadership of a Provost and an Academic Dean, resulting ultimately in a turnover of more than six deans in ten years, that resulting in chaos within the organization; and (3) leadership and management, the leadership of the University for decades ignoring the college, resulting in the college having to rely almost solely on student tuition for paying virtually all of its costs and expenses, then, in the last couple of years, switching to micromanagement, where first one LIU vice president and then a subsequent one willy-nilly cut courses right and left, further curtailing course offerings and driving away students because of fewer course offerings,then imposing a \"new core curriculum\" that was largely agreed to by faculty only under the threat of being closed otherwise. The reality is the Board of Trustees rather than close the college should have fired the president and vice president; for 16 years or more the Trustees allowed the college to lose two million dollars or more each year (a corporation and its board of directors long before that period of time would have fired the top management for bad management). What is interesting is that the college continued along for decades losing roughly two million a year; once the university administration switched to micromanaging, that figure raced from four to five and then a projected twelve million loss each year. Low enrollment is certainly a factor in the demise of Southampton College, but to find the real reasons --at least in my opinion-- one needs to look at the above factors. What else could be expected for Southampton College given the above factors? The question is, if that management continues in place for the university, and long after Southampton is gone, can the rest of the university, in terms of its future, be far behind? Pray for it!    - Professor Donald G. Baker  '),(16398,2839449,'28,June,2004','2004-06-28 00:00:00','   Hello all, Just a note...bit of a disclaimer if you will... We try to post anything and everything that is sent to us. We do take editorial license (no obscenities and we may fix a spelling error or two), but this blog is about having a  place for you to voice your opinion in a community forum. We don\'t agree with everything written, and we do not necessarily share the same views others express in their letters and entries. If there is a particular entry you would like to address, please leave a comment (located at the bottom of each entry) or email me a response to post.    It is the discussions and exchanges of ideas and information that will help us try to find ways to help Southampton College, its students, alumni, staff and faculty. What are you waiting for? Get typing a comment or sending an email!    Thanks!  '),(16399,2839449,'10,July,2004','2004-07-10 00:00:00','   (Newsday did not print this letter from Alan, but he was kind enough to share it with all the Save Southampton College! readers.)   Southampton College has been the home to many students and faculty over the last 40 years. Sadly the Real Estate minded board of Trustees has decided that now is the time to close the college, sell the campus, and settle old scores and debts.  The trickle down effect will be gut wrenching for thousands of people.  The East End of Long Island will suffer the lose of itâ€™s largest single employer, the baker, the gas station, the supermarket, the cleaners, the real estate agent and the East Endâ€™s economy in general will suffer greatly.  So will the displaced students who are enrolled in Southampton College. The faculty who has made homes and lives there for many years will be uprooted and the plethora of jobs that have supported so many will be gone.  The Local population will feel the sting more than they realize.  The college has become a part of the community in so many ways that most people forgot what it was like before it was here.  When a parent comes to visit his son or daughter at the college he often takes them to eat in town, he might buy a something at Hampton, stay overnight at a hotel or motel and have breakfast in town on main street.  Little gifts or souvenirs are not an uncommon item to purchase from any shop in town. This will end.  They will not come, they will not spend their money and any student who was enrolled for the fall, when the enrollment was up 13 %, will probably transfer out as fast as he or she can.  The Economy and the people here will take a big hit...  The town will be hurt by the loss of the intellectual energy coming from the various activities at the college.  The Writers conference, the Steinbeck Book Fair, The Art Shows, The Plays, the Concerts, the frenzy of the summer will die at the end of the summer and it will be very different.  There will be no fall or winter Buzz from the college and the town will revert to what it was 40 years ago when Southampton â€œclosed upâ€ after Labor Day.  I went to school at Southampton many years ago and I remember what it was like at the  End of the summer. Labor Day came and it became a very quiet place. This was a summer resort town, not a year round, not a vibrant, living, thriving community as it is today.   The Alumni are gathering for a fight, we do not want to see our history in the trash pile at LIU.  LIU has promised a college within a college.  Southampton college will cease to exist after it moves to the CW Post campus, maybe a it will last a year or so but no more, there will be  no reason for it to last longer .  They have promised to continue the scholarships for all athleteâ€™s that transfer to Post...well great, good for Post what other cherry picking ideas do they have to enliven the CW Post campus... maybe they want to take the college Windmill with them, the Radio Station and the Unfinished library.  They can not take our memories, our feelings or our desire to keep the college here and alive for generations to experience the same influences that we experienced.  We were all exposed to a love of Art, Literature, intellectual freedom and the ability to dream.  To Dream, that what you want--- can happen.  To take a course on the Bible sitting on the beach and listening to a profound explanation of some verses in Micah will never happen at CW Post campus. â€¦â€¦  To See FX Sillerman up on the stage with Paul Simon or Jimmy Buffet or Mel Brooks and listen to him introduce his â€œFriends â€œand know that he really means it when he says â€œfriends â€œ... thatâ€™s the man who put his money were his mouth was, his millions and he helped keep the dream alive.  We, the alumni, the friends of the college and the community have to fight LIUâ€™s decision now. Before it is too late.  Nothing will be the same again if we do not win this fight.   - Alan Michael Braveman  '),(16400,2839449,'09,July,2004','2004-07-09 00:00:00','   The following Alumni meeting update was emailed to Sam by JoanTuttSeashell@aol.com:  THEY CAME FROM ALMOST EVERY TOWN IN SUFFOLK COUNTY: FROM HUNTINGTON TO EAST MARION. THEY CAME FROM NASSAU, BROOKLYN & QUEENS. THEY CAME FROM NYC. THEY CAME FROM WESTCHESTER.  THEY CAME FROM NEW JERSEY, CONNECTICUT, VERMONT, MAINE, AND COLORADO!  WHY?     TO SUPPORT SC!!    The Meeting ran smoothly, thanks to  Bob Martin, \'68,  who orchestrated the format.   LIU President David Steinberg  gave a 41 year history of the financials of SC and a review of LIU, too.   That took about 45 minutes.    Mickey Director ,\'68, spoke on governance and directed questions to the LIU panel that consisted of Legal Council, Current & former Financial Officers, Dean of Academic Affairs, Chief LIU Fundraiser, The President of LIU and SC Provost.    C.L.A.S.S. provided a court stenographer. When we have that piece ready, we\'ll alert you.    The media was there also: Channel 12 Newsday Dan\'s Papers  & More   Alum Matt LoGuercio  videotaped the entire proceedings.   Alum Alan Braverman  photographed the entire evening.   Bob Martin  then directed questions concerning finance.  LIU gave out printed info pertaining to the bottom line.  In between each of the three segments, questions were taken from the floor.   Peter Philip , \'67 addressed the segment on Problem Solving.  In conclusion, Bob Martin  asked that the university to consider not making their announcement irrevocable, in order to allow the SC stakeholders, students, parents, alumni, faculty, staff, community, businesses, government and concerned citizens time to formulate options for the future of SC.  Dr. Steinberg said he will take our request to his board.    On a radio or in a paper near you:  Check out The Southampton Press 7/9/04 for SC Articles: \"Lottery Longshot Was College Official\'s Best Hope\" Page 1  \"Thiele\'s Letter Blasts LIU Board\" Page 1  Letters to the Editor Page A11 \"We Stayed\" by Alan Braverman \'68 \"Good Investment\" by Bruce C. Doscher \"No Free Lunch\" by Len Czajka  VIEWPOINT  Page 14 \"A COLLEGE\'S SUCCESS IS ROOTED IN PEOPLE, NOT MONEY\" by Mark E.Strecker  \"Like Swallows to Capistrano, Writers Return\" Page B1  A Sultry Voice Put to Good Use\" Kathleen Turner @ WLIU 88.3 FM Page B6   Speaking of WLIU,  Alan Braverman & Matt LoGuercio  were interviewed by same on 7/7/04,concerning the current situation at SC.  The program will be aired Monday, July 12 @ 10 AM.  '),(16401,2839449,'08,July,2004','2004-07-08 00:00:00','   I listened to the Fiscal History of LIU as presented by Dr. David Steinberg last night at Southampton College. Later I listened to the LIU numbers cruncher explain in painful detail that the red ink has flowed from Southampton College since its inception. Even latter we heard from the chief fund Raiser of LIU about how hard it is to raise money for Endowments from former students. We even heard how hard it is to get students to come to this campus. And how expensive it is...  But no one ever called and said â€œthe college is in severe financial shape we need your help Now â€¦.or we may have to closeâ€¦. how much can you contribute?â€  And no one ever called and said â€œwe do not have the money to repair the dorms, they need to be painted, in fact the entire campus needs a paint job will you as alum come and help... donate money, or paint, and time and energy?  And no one ever called and said â€œwe are having trouble recruiting students  ...do you know anyone that would like to come to college here?  No one, not once in all these years.  You have my assurancesâ€¦. we all would have said yes, we all would have been happy to help.  There are more then 8000 students who have attended Southampton college... there would have been eager to help.   LIU has announced that they plan to close this campus and transfer â€œThe students and faculty (who want to go) to a more complete campus, with better faculties... CW Post.â€  Southampton College is to become a campus within a campus at CW Post, we are assured.  Assurances are fine; we have had many of them from LIU over many years.  The assurances that that they would maintain an Undergraduate Facility at Southampton was the first.  More recently we had assurances that all was well and that Southampton was undergoing a great change, a new core curriculum was established, a new Provost installed and a new 100 million dollar building campaign to refurbish and rebuild the campus at Southampton College was announced.  New Faculty had been hired, a deal had been struck with the Parish Art Museum to move onto the Southampton College Campus, a new library was under construction and President Steinberg took great pride in discussing all these new and wonderful developments less then a month ago.   We had his assurances.  Last night he gave us his assurances again...  He assured us that he would carry our request for an open and frank discussion with the board about the opportunity to keep the college from closing and he further assured us that there had been no discussion with anyone and no particular plan to sell the campus had been put into play.  He assured us of this.  He assured us that there has been a continuing effort over many years to find more and better ways of funding the needs of Southampton College.  He assured us it has been an unrelenting commitment form LIU to Southampton College.  But no one ever took the time to tell us how much trouble the college was in, not once in 40 odd years did anyone ever even hint that financial doom was just around the corner..  Not once.  But we have his assurances that all will be well once Southampton College is closed and re-fashioned at CWPost campus.  I am not assured. You should not be assured. You should be greatly distressed and worried as we all are.   Alan Michael Braveman  Former Editor of the Southampton College Windmill 1967-68  '),(16402,2839449,'08,July,2004','2004-07-08 00:00:00','   The following letters have been forwarded to the site by Bob Danziger.   \"Thursday, July 8, 2004   To:   The Southampton Community From: Bob Danziger  Wednesday night I attended the Angry Alumni Meeting. It was a pleasure to see about 250 people who care so much about the College. President Steinberg and other members of the Administration argued forcefully that the financial picture at LIU (problems at Post and Brooklyn also) leave them no alternative but to close Southampton. President Steinberg gave no explanation about why he has been unable to develop any meaningful endowment for LIU. He argues that courses in the Natural Sciences lose money at Southampton, but I would imagine the same losses might apply to science courses at Post and Brooklyn, so maybe they should eliminate science there also.  The Humanities, Fine Arts and Marine and Environmental Sciences programs evolved over the last 40 years, and as we all know, there are lots of success stories from our alumni. President Steinberg argues that the new curriculum could have saved the College. One might also argue that the improbability of it being successful was actually a foregone conclusion, and realizing that, the Board will finally use the sale of the College\'s 110 acres as a means of bailing out LIU, demonstrating President Steinberg\'s fund-raising skill, and making money from Southampton.  I hope there is some kind of white knight out there.  In any case, the following letter is the one that I delivered to The Southampton Press at the beginning of the week. I trust they will print it next week (Viewpoints column). I don\'t know if it will do any good to explain where the mistakes occurred over the last 20 years, but right now, it can\'t hurt.\"   Bob Danziger  July 5, 2004                    The College Goes Belly-up  \"At the very first Southampton College faculty meeting I attended, back in September, 1968, there were \"Huck Foxy\" signs around the campus, as well as (I gather) all over the CW Post and Brooklyn campuses of LIU, calling for the dismissal of President Gordon Hoxie. At the most recent meeting on June 21st, when President David Steinberg and other members of the LIU Administration told us why the Board of Trustees was closing Southampton College, I think there would have been a lot of faculty and staff who would have held \"Huck Steinberg\" signs, if they had been given any advanced warning of this shocking decision, which flies in the face of the current campus-wide building renovations and construction.  For twenty years LIU probably epitomized administrative mismanagement while under the leadership of David Steinberg. Not only have other private colleges on Long Island (Hofstra, Adelphi, Dowling and St. Josephs) been flourishing, over this period LIU has been unable to develop any kind of an endowment. Faculty disenchantment with Steinberg, at all three main campuses, is probably at its peak right now, but it might be worth describing some of the reasons leading up to the demise of the College.  It would be a mistake to put all the blame on Steinberg. I think some of the gravest mistakes were made by Tim Bishop over the last ten years he was Provost. During that period we had seven (or was it eight?) Academic Deans, at least three Directors of Admissions, a completely incompetent Associate Provost, and a revolving array of other key administrations. The turnover in deans prevented the College from delivering any kind of coherent academic philosophy, or scholarship and financial aid policy. The faculty was invariably rebuffed when information was requested regarding how money was being spent for administration, staff, faculty and physical plant expenditures. At a time when the Marine and Environmental Science programs were winning national acclaim, (in 1997 seven Southampton students won Fulbrights, out of a graduating class of about 250 - the highest percentage of any college or university in the country, and the College should have generated some kind of a \"ripple effect\", Provost Bishop was never able to communicate the success of our programs to prospective students.  I have always believed that a college must \"recruit for excellence\" - to try and get the best students, and then to reinforce them when they get here - with outstanding faculty, with Internships and Co-ops which would enable our students to clarify their career goals and to distinguish themselves when they graduated - I don\'t think Bishop ever really appreciated that.  But it would be unfair to put all the blame on him - after all the turnover in Deans could not have happened without the President (Steinberg) and the Board of Trustees going along with it, and Bishop was always under an incredible amount of pressure. The highest level decisions are made at the University Center in Greenvale, and, of course, there was also a lot of turnover over there during the last ten years; three or four Academic Vice Presidents, and a couple of Assistants to the President. One could argue that Bishop was only \"following orders\" from the Central Administration, but I think that some of his policies - particularly with respect to scholarship policy - had not only a profound impact on Southampton College, but also on the rest of the University.  In any case, Robert S.F.X. Sillerman, whom Bishop met playing beach volleyball in Southampton, became our Chancellor about 11 years ago, and Sillerman has lots of money. I think Bishop might have been canned six or seven years ago, were it not for the fact that Sillerman was the \"money cow\" that Steinberg and the Board had been seeking for the past 40 years. Since Sillerman provided the money for the summer \"All for the Sea\" concerts (presumably supporting the Marine Science scholarships), financing Chancellor\'s Hall, and the new library, Bishop continued on as Provost. And Steinberg, rather than jeopardize the cozy relationship with Sillerman, allowed Bishop to remain as Provost until Sillerman could promote (ca $500,00+) Provost Bishop\'s successful campaign for Congress. (I must add that Tim has been a damn good Congressman - probably a better Congressman than Provost). I would encourage everyone to vote for him.  Which brings us around to Chancellor Sillerman, a member of the LIU Board of Trustees. He became interested in transforming the College shortly after Tim Bishop went off to Washington, in January, 2003. When he met with the faculty about six months ago, explaining why he was pushing the \"transformation\" of the college, he said that when he looked at the College, he saw a \"blank slate\" - it had nothing to justify its existence in its present form. He felt that we should change into a College with an interdisciplinary curriculum which would address the issues that the youth of America would face in the 21st century. At that faculty meeting I asked how could say that the College had a \"blank slate\", when there were a couple of hundred alumni who have received their doctorates, many are teaching at universities in this country and abroad, and 37 graduates have received Fulbrights - not what I would call a \"blank slate\".  But, as Sillerman said in the March 4 issue of the Southampton Press, \"I am not trained as an educator\" and \"I have devoted my life to commercial pursuits, and I\'ve been fortunate to be very successful\". In the article Sillerman described how the new curriculum would eliminate standardized test-driven approach prevalent in much of academia\". In his words; \"Our thrust was to challenge the faculty to design a program that is less focused on teaching the answers to predictable questions, and instead teaching people what those questions are\". Hmmm....what does he mean by that??? Would you cough up the highest tuition on Long Island ($25,000 +), for this kind of an education? But these changes were supposed to bring financial stability to the College, adn, as he said \"Hopefully\", within five years, the deficits would be a thing of the past\". I can\'t help thinking of the old story about the Emperor\'s (read Chancellor\'s) new clothes - this new curriculum made absolutely no academic sense, and the aforementioned youth of America (prosepective students) probably recognized it. As I said at the faculty meeting with Steinberg on June 21st (Marine Science metaphor); \"The new curriculum went over like a lead balloon; it smelled like a dead fis, looked like a dead fish, and it went belly up\".  The Marine Science Program, as it evolved in the early 1970s, consisted of what was essentially a double major; a classic major in Biology or Chemistry, with another 20 credits of Marine Science courses. The new core, according to Sillerman. as he said in the March 4th article, was a \"campus-driven intiative\", but it was actually orchestrated by Jeffrey Kane, LIU\'s most recent Academic Vice President, Chuck Hitchcock, Southampton\'s most recent Academic Dean, and Paul Forestell, head of the Faculty Council and chief cheerleader for the transformation. Faculty from the Natural Science Division were virtually excluded from the deliberations, and there was alway the \"Sword of Damocles\" over the faculty; if they didn\'t go along with the transformation, the Board would close us up. the administration dismantled the traditional academic divisions (Natural Science, Humanities, Fine Arts, etc.) and replaced them with smaller programs centered on each major. Kane, Steinberg and Hitchcock eliminated the majors in Chemistry, Cell/Molecular Biology and Sociology, and maybe a few others.  Along with two other members of the Chemistry faculty, I met with President Steinberg and Academic VP Kane at the University Center in July, 2003, shortly after they decided to eliminate the Chemistry major. While rejecting the proposition that a Chemistry major was fundamental to the Natural Science\'s, Steinberg also went on to suggest that one of the problems at the College was that too many bright-eyed and bushy-tailed prospective students, with the hopes of getting a Fulbright some day, were applying, with hopes of getting a big scholarship (i.e.; discounted tuition). The implication, I suppose, was that we need more rich dumb students.  It is presumptuos for me to suggest to Chancellor Sillerman what he should do with his money. He sold Clear Channel for about $1.5 Billion so I think he has some big bucks around the house. Rather than invest in \"bricks and mortar\", the \"Bishop-Burke Learning Center\" (i.e., library, which will probably become an empty shell) and \"All for the Sea\" concerts, it might have been wiser for him to create about 50 full tuition University Scholarships for outstanding students, some of whom migh actually be interested in the new curriculum. I would bet, however, the majority of them would be interested in our traditional disciplines in Marine and Environmental Science, the Writing Program and Fine Arts..  I really believe the members of the Board, and Sillerman, do have the resources to keep the College afloat. If only Sillerman would acknowledge some of the mistakes that he\'s made, I think he could save the College, and provide the resources to enable it to fulfill its promise. At a time when the East End\'s marine and terrestrial environment is under constant pressure from developers and commercial interests, one can only hope htat a white knight may emerge. The other great tragedy, of course, is that the East End is losing its only institution of higher learning, and it will also will lose its major employer. Approximately 300 members of the Southampton Collegs community - facutly, staff and administrators- will be laid off a year from now, and this will have a significant impact on the local economy.  Three years ago we hired six outstanding young scientists; a biochemist, a physical oceanographer, a marine ecologist, a geneticist, a marine mammalogist and a mathematician. Getting a job in academia is not so easy these days, and these men and women all had plans to make a life at the College, just as some of the older faculty did. They bought houses, put their kids in school, and participated in the design of the new core. At the June 21 meeting with Steinberg, when they learned that their efforts were in vain, not suprisingly, there was an overwhelming sense of betrayal.  In some respects though, the people who may be affected the least will be the students. They are young, bright, intelligent and resilient, and with any luck, this will be the worst thing that will ever happen to them. And like the little boy in the story about the Emperor, when they considered the transformation and the new curriculum, they collectively said \"Look!! It\'s wearing no clothes.\"   Robert M. Danziger Professor of Chemistry Southampton College  '),(16403,2839449,'08,July,2004','2004-07-08 00:00:00','   urlLink Provost of Southampton College Daniel J. Rodas      Reporter Clara Jablonski:  \"What happens next?\"   Daniel Rodas:  \"The change is to better serve the educational needs of the students and faculty but in a new location. The new undergraduate core curriculum will go forward as a separate curriculum, but at our C.W. Post campus in Brookville. We will have housing there for the first-year students. Although housing at the Brookville campus is limited, off-campus housing options are being explored. Students will still apply to Southampton College and receive their degrees from Long Island University at graduation.  The graduate programs in English and Education offered at the college will remain at the Southampton campus, as well as the writing program, which has included professors Roger Rosenblatt and Jules Feiffer. Long Island University Public Radio, WLIU, will continue to broadcast from its present location on campus. These programs do no require the same level of support services as the undergraduate programs.  There are a number of options for the future of the marine program facilities and discussions are ongoing. We continue to talk with the Parrish Art Museum who leased nine acres of land from the college to create a privately-funded, state-of-the-art facility. We are actively working with the faculty and staff who will be affected by these changes, and have invited the undergraduate faculty from Southampton to teach at C.W. Post. We are committed to achieving the best possible outcomes in a very difficult situation.  My role is to support this transition and help the students meet their needs. On June 17, we had an open meeting with the students, Charles Hitchcock, Dean of College, and Paul Forestell, the Faculty Chair. The students voiced their concerns. In the first 24 hour period, the student reaction was one of surprise and support for the college and each other. We assured them that their financial aid awards would be preserved. The overwhelming response of the students is \"how can I help?\"  There is a lot to do. We already have started construction on the $10 million Bishop-Burke Library Learning Center. The exterior will be completed, but the interior will remain incomplete for now. There were other planned projects in the building program, including a 65,000 square foot recreational center, new marine science facilities, and renovations for the dormitories and student union that will be discontinued.  And there are a lot of other things to deal with. There have been concerns raised in regard to the allocation of funds from the college\'s summer fundraiser. In the past, money raised at the All for the Sea benefits has been used to provide scholarship support for students and that will again be the case for this year\'s All for the Sea 2004 Rod Stewart concert on August 23.  The unfortunate reality is that it is expensive to run a quality institution. This is a private tuition-dependent college, not a state-run university, and like other institutions, there is an increase in cost over time. Southampton College has utilized resources wisely and prudently, and this is no fault of anyone\'s. Our ultimate goal is to continue to best serve our students.\"  '),(16404,2839449,'08,July,2004','2004-07-08 00:00:00','   I did not attend the meeting with President Steinberg and Provost Rodas last night and I have yet to receive an email from anyone who did attend.   If you did go to the meeting last night, please send me an email to post here on the blog.  Let the other blog readers know what happened, what was said, how you reacted, how the crowd reacted, who was there, was it productive, how did it end, and what happened afterwards - anything you\'d like to write about the meeting. The email address is  scogar@alumni.liu.edu .  Thanks in advance for keeping all of us informed!  '),(16405,2839449,'07,July,2004','2004-07-07 00:00:00','   urlLink The Southampton Press  didn\'t have coverage of the meeting last night in this week\'s edition, but the cover story is one to check out:  Against all odds, Library Director Robert Gerbereux is trying to save Southampton College. Unfortunately, last week the New York Lottery put those odds at 135 million to 1.   From July 1st, a  urlLink letter to the editor in Newsday :  How to save Southampton  Regarding Sandy McIntosh\'s op-ed \"A fond farewell to Southampton College\" [Opinion, June 24]: As a former Southampton College employee in the undergraduate admissions department during the 1980s, it saddens me to learn of the fate of the campus\' academic programs. And it distresses me to think that the property will be further developed.  For a college that is nationally known for marine and environmental science programs, isn\'t it ironic that this may come to pass?  A reasonable solution would be to use the property as a research center similar to the Woods Hole Oceanographic Institution in Massachusetts. This would keep hard-working local people employed, maintain the property as is, and could be funded by a combination of private funds and federal, state and local grants.  The only \"mansion\" on the property should be the one that is already there.   Edward J. Tobin  Sayville  Still looking for articles in the papers that covered last night\'s meeting. Due to the publishing schedule of East End Papers, I would expect Newsday to have a story about it tomorrow, Dan\'s Papers to cover it in the Friday edition, but it will be NEXT Thursday to hear anything about it in the Southampton Press and other local papers.  '),(16406,2839449,'07,July,2004','2004-07-07 00:00:00','   (My deepest apologies to Heather - her email was lost in my inbox. She sent this to me on 6/25.)    Everyone that went to Southampton, whether they were there for one semester  or four years, is connected to the spirit of that campus. Its alumni are a unique blend of individuals who all chose SC because they felt that connection when they first came to visit. It seems almost too easy for Steinberg and the rest to dismiss our pleas and ideas because they have never felt that spirit. To them it is a business, one in which they poorly managed. How is it that in 2003 the students were told the campus was $55 million in debt and two years later it is up to $77??? Was nothing done to correct this?? Were the admissions staff directed to continue recruiting students mainly from the Long Island area, wealthy students I might add?? This was an admissions tactic that was admitted by Tim Bishop to the students in an open forum, one that he felt was a bad idea and needed to change. Chancellor Sillerman \"donated\" a great deal of money to the campus which he announced during the 2003 commencement ceremony. Where is this money going towards? How can someone who has such personal ties to the facilities at Southampton (Sillerman\'s radio station) and who is on the Board of Trustees not see this coming and yet continue to donate his own money??? Has this plan to relocate the undergraduate programs been in the works for awhile, but kept quiet? There seems to be much more going on than the surface shows. And I also heard a member of the Board of Trustees donated $25 million to UCLA, heard anything on this? And if the renovations continue, and if the land is sold, is the university seeking to make a profit? Could this be why they were so willing to sink such much more money into the campus recently? Would it have made a difference for Southampton if the funds from each campus stayed in each campus, instead of collected and redistributed by the university? Oh and what about the infamous new cell phone tower that was built in 2003 behind what I will always continue to call Southampton Hall(I don\'t even remember what they changed the name to)? Where will the revenue from its occupiers go? Was that really necessary to put the students and employees in danger of a new cell phone tower to make some quick money?  '),(16407,2839449,'06,July,2004','2004-07-06 00:00:00','   The meeting with President Steinberg and Provost Rodas is  TODAY, WEDNESDAY, JULY 7th at 7:00PM IN THE AVRAM THEATER . If you will be attending, be sure to sign the attendance sheet. This way there would be a count of all who attend - from staff and faculty, to current students and alumni. Please include your name and class year or years taught or years worked for Southampton College. We need a real count of who showed up to support keeping Southampton College open. Let\'s not let the University fudge numbers with whom they say attended. There were discrepancies like that when comparing the counts by the organizers of The March For Women\'s Lives (who had sign up sheets for all involved) compared to the Anti-Choice movement and the local police.   There will also be t-shirts available, sponsored by  urlLink C.L.A.S.S.  The t-shirts are navy blue on white, have the Windmill logo, and say \"SAVE THE WINDMILL, SOUTHAMPTON COLLEGE, LONG ISLAND UNIVERSITY\". Sizes include medium, large and extra large and the price is a \"goodwill offering.\" They will be available at tonight\'s meeting and future C.L.A.S.S. meetings and events. You can also get your t-shirt via mail - send Joan Tutt an email (JoanTuttSeashell@aol.com) to get mailing information (I don\'t want to post her address here). You can make checks for t-shirts and general donations out to: C.L.A.S.S.  If you cannot attend tonight\'s meeting, there is still time to email your questions and comments to Carol.gilbert@liu.edu or call the Alumni Office at 631-287-8348.  '),(16408,2839449,'06,July,2004','2004-07-06 00:00:00','   Published in the July 2, 2004 edition:  urlLink  VOICES 2    You won\'t be able to access the old links since they update their website to list only the current issue - so read this one TODAY!   urlLink Last week , we published, under this same heading, VOICES, the opinions of three individuals. These were John Kanas, who is CEO and Chairman of the North Fork Bank, a Southampton College Graduate, and a member of the Board of Trustees of LIU. 2. Tim Bishop, the former Provost of Southampton College who is now our U.S. Congressman and 3. The new Provost of Southampton College, Dr. Dan Rodas.    urlLink This week , under this same heading, VOICES, we publish the opinions of three other concerned leaders. They are Shirley Strum Kenny, the President of Stony Brook University, which includes Stony Brook University Hospital. 2. Fred Thiele, a graduate of Southampton College and currently our N. Y. State Assemblyman and 3. Harry Marmion, the very respected former President of Southampton College during the years 1972 to 1979.   Next week, we hope to publish the opinions of Bob Sillerman, the Chancellor of Southampton College, and Dr. David Steinberg, the President of Long Island University.  '),(16409,2839449,'05,July,2004','2004-07-05 00:00:00','   The  urlLink meeting  with Long Island University President Dr. David Steinberg and Southampton College Provost Dr. Daniel Rodas  has been moved to the Avram Theater, Fine Arts Building , Southampton College.  The meeting is TOMORROW, July 7th at 7:00pm.   Looks like they finally realize more than a few people will be showing up tomorrow and Duke Lecture Hall would not be able to accommodate everyone. Since I will not be able to attend, would there be someone willing to pass along meeting notes, information, etc? Perhaps a photo or two to share?  Another idea -  Would someone please make a sign-up sheet for those who attend?  Include name and class year or years taught or years worked for Southampton College. Let\'s not let the University fudge numbers with who they say attend.  We need a real count of who showed up to support keeping Southampton College open.  '),(16410,2839449,'05,July,2004','2004-07-05 00:00:00','   I wanted to thank everyone who has been visiting, reading, and contributing to this blog. Take a moment to scan through earlier posts to see read some of the excellent comments that have been left by visitors. Join the discussion! Leave a question, answer a question, or leave a comment! Help Save Southampton College!  '),(16411,2839449,'05,July,2004','2004-07-05 00:00:00','   For the past three years I have had the honor to serve as an Adjunct Professor of Political Science and History at LIU Southampton.  Having previously taught at American University, Rutgers University, Dowling College, Hofstra University, and having served as a federal prosecutor for many years, I feel that I am in a position to make an informed evaluation of the quality of the students and faculty at Southampton.   Over the three years,I have taught scores of dedicated and gifted students who not only eagerly sought to acquire knowledge but, more importantly, sought to apply that knowledge beyond the walls of the academic setting.  It has also been a sincere privilege to work with some of the finest faculty members that I have ever known  Based on my experiences, I felt compelled to respond to the Administration\'s abrupt and, in my mind, arbitrary decision to eliminate the undergraduate programs at Southampton.   First, I believe that the students, staff, and faculty of LIU should feel betrayed by the way in which the decision to close was reached and conveyed to the community.  For the past months, all of us, to varying degrees, at the specific direction of the Administration, were working on a new core curriculum designed to bring a new sense of vitality and academic excellence to the college.  Now, it appears clear that the countless hours of hard work, the innumerable meetings and discussions, the seeking of input from students and other academic institutions, was done at a time when the Board of Trustees and the Administration were obviously considering and, one might believe, preparing for the closure of the school.  A professor would be sanctioned were he or she to assign a major research project, demand the students work endless hours on that project, and then arbitrarily state that the course was cancelled and all the work was for naught.  And yet, is that not exactly what the Administration has done to each of us?   Second, the duplicity was exaggerated by the fact that, thanks to the generosity of certain donors and the entire community, a new Library is under construction and well on its way to completion.  What possible logic exists for the expenditure of millions of dollars in constructing a new Library while, at the same time, the future users of that facility are being shipped off to CW Post many miles away?  Were I am alumnae or donor who contributed to the building of that library, I would demand my money be returned--otherwise one suspects that they have contributed to a building that will be part of the new condominium or commercial development--enriching the coffers of LIU at the expense of those who loyally and faithfully contributed to the college.   third, if I were an Incoming Freshman, Sophomore or Junior at LIU Southampton, I would feel an ever greater sense of betrayal.  When one applies to and opts to attend a four year college, one expects that they will receive a four year education.  But each of you have essentially had the contract you signed with LIU Southampton unilaterally breached by the Administration.  In announcing the closing, the Administration congratulated themselves for giving the community a one year reprieve.  To that I say, thanks for nothing.  Consider the \"choices\" that undergraduates face:  (1) incoming freshman can withdraw from their attendance but will find themselves well beyond the application and decision dates at other schools--so your choice is attending a school whose life span is one year or scrambling to try to salvage your first year of college; (2) sophomores will find themselves having paid full college tuition for what will become a two year/community college education in the sense that all of the riches of the upper division classes will be lost forever when the school closes; and (3) juniors will find themselves literally forced to make the switch to CA Post as I know of very few schools who will accept a transfer student to receive a degree when only thirty credits remain  to be completed.  And lest the graduate students believe that they have been \"saved,\" let me suggest that a few hundred student graduate program will not long survive the logic and calculations of this Administration.   Fourth, the fact that the Administration has apparently been actively engaged in the sale of the campus property underscores all of the concerns expressed above.  While real estate is not my field, I can only imagine that the value of this campus property in Southampton is staggering, indeed.  And one only has to think about the controversy and dismay attending the debate over a potential casino, to recognize that a new commercial or residential development is the LAST thing that the Southampton community needs or deserves.  And yet, it would appear, with no input from the faculty, staff, students, alumnae, donors, community, and elected officials, the Administration has demonstrated amazing (and unique) efficiency in moving towards the potential sale of the property.     Fifth, beyond all the substantive concerns expressed above, in my many years of professional work, I have never seen such a poorly handled communications process.  Receiving letters in the mail days AFTER the public announcement of the closing does not reflect a caring and compassionate response to the concerns and needs of the community.  If the Administration is seriously concerned about the \"family,\" then I would assume they will soon announce that a substantial part of the proceeds of the sale of the campus will be directed to the many staff at the school who, come May 2005, will find themselves unemployed.   What will happen to the secretaries, the custodial workers, the food service workers, the librarians, and others--does the Administration seriously expect them to make the move to CW Post (even if such positions were offered them).    In summary, I find myself with truly ambivalent feelings.  On the one hand I mourn for the potential destruction of a fine college and wonderful community of scholars.  On the other hand, I rejoice in the initiative and hard work demonstrated by many of you who are committed to saving the college.  Whether you ultimately succeed or not, I can promise you one thing--as the end of the day, you will feel far more pride and honor in your actions than those who took this action will feel.   To the extent that any of you wish to communicate with me- feel free to e-mail me at wps1952@optonline.net.   Good luck in all your efforts and never forget the words of Franklin Delano Roosevelt:   The only limit to our realization of tomorrow will be our doubts of today, so let us move forward with strong and active faith.   God Bless    William P. Schaefer  '),(16412,2839449,'05,July,2004','2004-07-05 00:00:00','   Are you an alumni with your own blog or website? Send me the link and I\'ll put it under \"Alumni Pages\" on the side bar. Our goal is to try to save Southampton College from closing and we also have a great opportunity to network, catch up with, and meet other alumni. The first link is  urlLink Tidewater Bound  but I know there are many more Southampton College alumni, faculty, students, and staff with blogs and websites. Send them in! And thanks in advance!  '),(16413,2839449,'05,July,2004','2004-07-05 00:00:00','   Has anyone contacted David Dubin a former adjunct that taught Environmental Law at the College to see if he would be interested in providing some pro-bono work for us?   I believe he is with Twomey, Latham, Shea and Kelly.  '),(16414,2839449,'05,July,2004','2004-07-05 00:00:00','   The LIU Administration and apparently members of the Board of Trustees have been cited in newspapers as saying they are getting property appraisals, thinking possibly of selling the campus. If they succeed, it will be too late to consider all alternative measures. There is a question of whether or not \"interested parties\" can seek a court injunction to prevent that action before it is too late. Injunctions have been used to stop developers until Environmental Impact Statements have been done; other groups use injunctions to stop questionable actions by government agencies and private groups;etc. Are faculty and staff an \"interested party\"? Alumni? The Southampton/East End townspeople who put up the original money to buy the property and present it to LIU for an undergraduate college, LIU now deciding to terminate the undergraduate college to possibly make a fast dollar from developers? All of those interested legal minds, including all the lawyers who cut their teach as undergraduates at Southampton: where can we move from here? Suggestions?  - By Donald G. Baker  '),(16415,2839449,'05,July,2004','2004-07-05 00:00:00','   I came to Southampton College 8 years ago, filled with enthusiasm and probably more optimism and ideas than were reasonable. I am a \"Let\'s put on a show and save the old homestead, ala Judy Garland and Mickey Rooney,\" kind of person. I have always bucked the odds in my professional career, getting my PhD from Virginia Tech at \"ahem\" a mature age. And I have developed a passion for teaching the students I have had the privilege of teaching over the last 8 years. I looked forward to keep on lighting the fire for critical inquiry under my students for years to come. Thus, what is happening has caused me to grieve along with many others lamenting the demise of what we had.  A few years ago, I started doing research in Mondragon, Spain. Fifty years ago, during the repressive Franco regime, a few Basque people, taught by a priest and sociologist, about working together as a community, started a cooperative experiment.. They made little camp stoves. Fifty years later, they are a corporation of cooperatives, making many products sold the world over, and providing many services with sales in the billions of dollars (or Euros now). And the Corporation is still based on the idea that one can have a successful business, with democratic decision making, and without exploiting the lowest paid employees.  Under very difficult circumstances, these people committed themselves to principles and ethics that are quite unlike the kind we see in most ventures today and what Franco had in mind for them. They followed a set of principles which included:  a) the primacy of work over capital, without diminishing the importance of money, which is an indispensable element. b) equality of all members-giving one vote to each member, while avoiding egalitarianism so that they could allow a differential in wages in a ratio of one to three;  (Which has more levels these days, but still a ratio is followed - comment mine)  c) supremacy of the General Assembly with solidarity in every thing so that paternalism is avoided; d) money invested in education brings the highest of profits . . .   **Taken from \"Prologue\" by Juan Leibar of Mondragon, page 12 in  urlLink From Mondragon To America: Experiments In Community Economic Development , written by Greg MacLeod, 1997; Sydney, Nova Scotia: University College of Cape Breton Press.   Faculty have been apprised of the financial problems at Southampton College for a good many years. We followed instructions we were told would help to keep the place going, not without question, but certainly as good employees and dedicated faculty and staff. Direct access to the Board of Trustees is not available, due to some unfortunate situations involving breeches of confidentiality on the part of a few faculty some years ago. In other words, it is not easy in a traditional institution such as this for everyone to have a say in decisions made, nor are they privy to all the information available regarding said institution.   We are led to believe this is the best and only way to ensure a successful enterprise.  I started speaking with various people about an idea I have for a college that would be remarkably different in its administration. That idea was developed in my mind during my visits to Mondragon and while researching the community during the last four years. I conducted a travel class there with a few students in 2000. We saw first hand how many of the cooperatives worked and are still working. We were especially impressed by the University and the schools that are run as cooperatives.  Being a crazy optimist, I feel it\'s time to put this idea out in a more public forum. We are facing a very tough situation. The financial problems LIU has with Southampton College are very real.  But, I would like the opportunity to explore an alternative college, one that is based upon the principles of Mondragon Cooperatives. The faculty, staff, administration and students would form the cooperative corporation which would encompass a variety of smaller cooperatives. For example, instead of the College acting as landlords for dormitories, students would \"own\" their dorm room or dorm apartment. They would sell them to the incoming students upon graduation or moving into Winter housing. They would be responsible collectively for making sure the rooms and apartments were maintained in good repair, and perhaps rented out to conferences or other events in the area to help pay for the upkeep. The upkeep, in turn, could be purchased, perhaps as a service from a Maintenance Cooperative which sells their services to the Student Dormitory Cooperative.  I won\'t go on and on with details or examples of how this could work, but I would like it if people would consider developing this idea and presenting it to the administration.   I met with the dean and faculty of the Massachusetts School of Law on June 22nd. I had written a blurb for the cover of a book,  urlLink Against The Tide  by Debbie Hagan, about the formation of that Law School from scratch, designed so working class students could obtain law degrees. It seemed to join both of my interests, working-class and poverty issues with the cooperative model of capitalism. I made the appointment to meet with them before I heard the news of the relocation of the undergraduate program and possible sale of the land at Southampton College. The luncheon and meeting with MLS certainly took on a different direction. We talked at length about what could be done to save the college without LIU having to carry the burden of the large deficit it has accumulated at SC.   I will give you one example of the suggestions these folks had. Run the college as a cooperative for five years, with option to buy at the end of perhaps five years.  So, we need to develop ideas and ways that offer alternatives to the ones in place right now. I believe it would be difficult to save a four-year undergraduate college located at the present site by protests alone. There must be a way for LIU to discontinue to burden of SC without the losing the college we love. It might mean separating from LIU so the university can concentrate on maintaining the larger, more urban campuses.  I would be happy to discuss this and other ideas people might have at the upcoming meetings. We need a practical plan.  For more information on the Mondragon Cooperativa Corporacion, use this URL.   urlLink http://www.iisd.org/50comm/commdb/desc/d13.htm   Peace,   Dr. Barbara J. Peters , Assistant Professor of Sociology and Women\'s Studies **** LIU-Southampton College   urlLink Yahoo! E-Group SaveSC  '),(16416,2839449,'05,July,2004','2004-07-05 00:00:00','   Joan DiPaola Tutt, \'68, \'74, has started a group called C.L.A.S.S. (Colonial Legacy Alumni Save Southampton) and a large mailing list. If you would like to be included on that mailing list, please email Joan at JoanTuttSeashell@aol.com. Tell her you heard about it on the Save Southampton College blog.  The C.L.A.S.S. group has also started a website,  urlLink Save Southampton College.com  - I\'ve put a link on the side bar. It\'s still under construction but you can check back to see when it has more content.  Also received a link from Kathy Kaz and her new blog,  urlLink Save Southampton Alumni .  As always, the Save Southampton College blog is looking for more contributors. Just send me an email with your interest and you\'ll be able to post directly to this blog.  '),(16417,2839449,'16,July,2004','2004-07-16 00:00:00','   Thanks to  Scott Carlin  for the following info :    Letters to the Editor Campaign    Now that the initial burst of energy has subsided, it is critical that we continue to generate letters to the editors of local papers. HereÂ’s some helpful information on how to reach the media. If you want your opinions to be read by LIU then send them a copy of your comments. If you do not want your name printed in the media (and why is that?) then send your comments to LIU Â– make sure someone is hearing your voice!   Dr. David Steinberg, President  University CenterLong Island University700 Northern BoulevardBrookville, NY 11548-1327(516) 299-1926   David.Steinberg@liu.edu    Dr. Daniel Rodas, Provost  Southampton College  239 Montauk Hwy  Southampton, NY 11968  (631) 287-8301   Daniel.Rodas@liu.edu    Long Island UniversityÂ’s Board of TrusteeÂ’s are posted at:   http://www.liu.edu/but01/factsfig/officers.html   Roger Tilles, Chair  University CenterLong Island University700 Northern BoulevardBrookville, NY 11548-1327  Letters to the Editor  1. Newsday  Letters should be fewer than 250 words. They will be edited for length, clarity and accuracy. Writers must mention relevant financial, political or other interest in a subject. All letters, including those sent via e-mail, must have the writer\'s day and night telephone numbers and mailing address. E-mailed letters should NOT be sent as attachments. We cannot acknowledge or return letters. Letters become the property of Newsday and may be published in all media. Anonymous letters will not be printed. Send e-mail to  letters@newsday.com . You may also write to Letters Editor, Newsday, 235 Pinelawn Road, NY 11747-4250 or fax to (718) 793-6422.   2. Southampton Press - Eastern Ed.  Include - If this letter is intended for publication, please include your Name, Town and Phone Number. If you go to  http://www.southamptonpress.com/sendletter.ihtml   There is an online form.  Otherwise send submissions to:  MAILBAG@SOUTHAMPTONPRESS.COM   FAX: 631-283-4927 (editorial dept);  135 Windmill Lane, Southampton, NY 11968   The following list is from the Pine Barrens Society Â– (send them a donation for their good efforts.)  http://www.pinebarrens.org/take_action/media_list.php       MEDIA OUTLET&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; FAX&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; E-MAIL&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; REGION SERVED    Newsday&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; (631)843-2986&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; urlLink letters@newsday.com &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Long Island&nbsp;  NY Times&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; (212)556-3622&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  letters@nytimes.com &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; NY Metro/National   East Hampton Star&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; (631)324-7943&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  letters@easthamptonstar.com &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; East Hampton   East Hampton Independent (631)324-2351&nbsp;&nbsp;  indepnews@aol.com &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; East Hampton   Dan\'s Papers&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; (631)537-3330&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  editorial@danspapers.com &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; South Fork   LI Business News&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; (631)737-1890&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  editor@libn.com &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Long Island   The LI Press(New Island Ear) (516)992-1801  rwoliver@longislandpress.com &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Long Island   The News-Review&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; (631)298-3287&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  mail@timesreview.com &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Riverhead   The Press of Manorville&amp; Moriches (631)878-5898  frankc@pressmm.com  Manorville/Moriches/Eastport   Medford News&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; (631)265-3504&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  messenger127E@aol.com &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Medford   Riverhead/North Fork Independent (631)324-2351  indepnews@aol.com &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Riverhead/North Fork   Sag Harbor Express&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; (631)725-1584&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  letters@sagharboronline.com &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Sag Harbor   Shelter Island Reporter&nbsp;&nbsp;&nbsp; (631)749-0144&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  sireporter.com &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Shelter Island   Smithtown Messenger&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; (631)265-3504&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  messenger127E@aol.com &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Smithtown   Smithtown News&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; (631)265-6237&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  info@smithtownnews.com &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Smithtown   Southampton Press(Eastern Edition)&nbsp; (631)283-4927  mailbag@southamptonpress.com &nbsp;&nbsp;&nbsp;&nbsp; Southampton   Southampton Press(Western Edition)&nbsp; (631)288-4965&nbsp;  mailbag@southamptonpress.com &nbsp;&nbsp; Southampton   Southampton Independent&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; (631)324-2351&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  indepnews@aol.com Attn: Editor&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Southampton   South Shore Press&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; (631)878-7805&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  sspress2000@aol.com &nbsp;&nbsp; South shore mid-Suffolk   Suffolk Life&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; (631)369-5930&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  slnewsroom@hamptons.com &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Suffolk County   Suffolk Times&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; (631)298-3287&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  mail@timesreview.com &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; North Fork   Times of Smithtown&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; (631)751-4165&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  ssn@tbrnewspapers.com &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Smithtown/St. James   The Traveler Watchman&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(631)765-1756&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  editorial@travelerwatchman.com  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Riverhead   Times Beacon Record&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; (631)751-4165&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  vbr@tbrnewspapers.com &nbsp;&nbsp;&nbsp;&nbsp; North Shore mid-Suffolk   Village Times Herald&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; (631)751-4165&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  vtimes@tbrnewspapers.com &nbsp;&nbsp; North Shore mid-Suffolk   Westhampton Independent&nbsp;&nbsp;&nbsp; (631)324-2351&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  indepnews@aol.com &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Westhampton  '),(16418,2839449,'14,July,2004','2004-07-14 00:00:00','   DEAR C.L.A.S.S.,  Below&nbsp;is information concerning The SC Orientation.&nbsp; It should prove very enlightening!  Thank you Stephanie!  Joan DiPaola Tutt \'68, \'74  Colonial Legacy Alumni / Save Southampton C.L.A.S.S.   My name is Stephanie and I came home from orientation today. The first two sessions were titled \'Welcome to Southampton College\' and \'Transition of Southampton College\'. The speakers consisted of student activities director Tara McCarthy, provost Daniel Rodas and president David Steinberg. During the session of \'Transition of Southampton College\' parents as well as students were allowed to ask any questions about recent events. Most questions were repetitive from past meetings with Steinberg, as well as the same questions that everyone has been asking. As one may have expected the tension and heat of the room increased until some people began to blow their tops. Steinberg began answering questions in a much louder voice then the tone in which the students and parents were asking questions (one mother began to express her feelings as well, and began shouting). Parents and students became extremely dissatisfied when asking Steinberg if we could raise the money, Steinberg shot down the question directed with a firm \'no\'. He then went on to say that he went up and down dune road looking for donations and received no help. He made it clear to parents and the students that our efforts would be a lost cause. Many parents, especially one who directed this next question at him, wanted to know why he took the easy way out and how he could just not try to keep this school afloat. After running his answers into circles he just continued saying that we\'ve been losing money for the past 43 years. There was much more which happened during this hour long discussion but most of it was questions which are unanswerable at this time, however I noticed that Steinberg was gradually becoming very aggravated with so many of the questions being directed at him. I felt that most parents were upset at the fact that he pretty much told us there was nothing to do to save the college and that we are definitely going to C.W. Post. He kept telling everyone that there was no money and there was no way to get the money, however when I talked to most of the students they are willing to do anything they can to keep Southampton going, and don\'t feel that Steinberg is helping us at all. If anyone has any questions regarding this meeting please contact me back and I would be happy to answer them.  Thank you- incoming freshmen,  Stephanie Beierle  &nbsp; Comments from  John Petrone : &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Sounds like Steinberg is getting very defensive. It also sounds like there is no intention to save SC from his part. Like I said earlier I believe the trustees intention all along was to sell off their cash cow (Southampton real estate), and dump the money into the rest of LIU.&nbsp; I am afraid that the only way to save Southampton College is to initiate some kind of litigation against LIU. Anybody been looking into this matter further? Dr. Baker posted a very interesting argument for a basis of a suit, or perhaps some kind of class action by the parents of incoming freshmen? Has Fred Thiel added anything since his last input on the debacle?&nbsp; Any other legal ideas out there?   Comments from  Barbara Peters : &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Thank you for such a thoughtful report on the orientation.The comment about being up and down Dune Road looking for money isinteresting, especially when considering a recent letter to the editor in Southampton Press from an Alum that stated, in all these years, no one ever asked for money to help the college survive.&nbsp; Good job. You appear to be the kind of student that makes us (faculty members) proud to teach.  '),(16419,2839449,'14,July,2004','2004-07-14 00:00:00','   I urge those of you who don\'t belong to the Save SC mailing list to join.&nbsp; Many great questions are posed there as well as possible solutions.&nbsp; Some important questions are reprinted below: &nbsp; Message from  Lisa Freese : &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Maybe I\'m stupid... but I don\'t understand something. In the LIU 2001 Anniversary Magazine on page 7 it is stated that in 1985 Steinberg was appointed President. It then goes on to say that during his tenure endowment went up, he has overseen 175 million in new construction ( on what campus?), that new doctorial programs have been added, A CUMULATIVE OPERATING DEFICIT HAD BEEN ELIMINATED... now wait a minute.. cumulative means all LIU Campuses right? So, if all CUMULATIVE OPERATING DEFICITS HAD BEEN ELIMINATED how are we 77 million in debt?&nbsp; It also says that operating reserves have been created &amp; increased.... now, to someone such as myself, fairly intelligent, 4 four year degree, etc... doesn\'t that mean that there was no operating deficit as of 2001??? And reserves were created??? well, where are they???? Maybe someone could shed some light on this because maybe I am just not reading in between the lines.... &nbsp; From  Barbara J. Peters : &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; So, how do we get this information out? I wondered about the cumulative deficit. When I was 16, I had a little account at a department store. I made angora bows to sell. So, that year, I didn\'t sell very many, and had a deficit. Now, I don\'t add that deficit to my debts now. ???? Maybe I\'m missing something, not being an accountant?? &nbsp; From  Lisa Freese : Don, &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Thanks, the actual endowment figures from 2001 are as follows: when he started endowment was 4.4 million, in 2001 it was 45million! I am told that alumni can donate to just Southampton as a restricted donation to come straight to us!!!&nbsp;&nbsp;Another question, We are holding a concert to benefit OUR marines Science division.. Well, there won\'t be anymore boats bought this year.. so who is that million going to benefit???? POSTor BROOKLYN????&nbsp; So we here at Southampton do all the work and raise the money only to give it to them????  &nbsp; From  Donald Baker : &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Good th ought. Maybe, since it is Sillerman who sponsors the concert (he usually does; I am assuming it is so for this one, as well), maybe the ALUM group should propose to Sillerman that he put the money that usually goes to SC and marine science into an escrow account to be used for a new SC not under LIU.  &nbsp;  '),(16420,2839449,'14,July,2004','2004-07-14 00:00:00','   Please save Southampton College - I was a member of the second freshman class there.&nbsp; What a shame and great loss if that wonderful place of learning were closed!  '),(16421,2839449,'14,July,2004','2004-07-14 00:00:00','   The following is a statement compiled by C.L.A.S.S.: &nbsp;   LIU Receives Failing Grade in Leadership    &nbsp; &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Southampton College alumni reject the notion that the closing of the school is attributed to its location, purported alumni and community disinterest, and the inability of faculty to develop successful programs.&nbsp; The Southampton alumni, based on the limited information provided by the University, maintains that the fundamental cause of the collegeâ€™s problems can be attributed to decades of apparent mismanagement by the Long Island University leadership.  &nbsp;  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; For the period 1962 through 1990 a cumulative loss of approximately $10 million was generated.&nbsp; From 1991 through projected 2005, the cumulative loss will be approximately $67 million.&nbsp; Our assessment is that the accelerating decline in financial performance is attributed to questionable business decisions by the LIU leadership as summarized below:  &nbsp;  1.&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Failure to invest adequately in marketing resources and programs.  2.&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Insufficient capital investment and overall maintenance support.  3.&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Maintaining an operating structure that limited the ability of local leadership to be effective.  4.&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Operating a model that could not generate the required cash flow.  5.&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Operating a curriculum that generated Fullbright scholarships at the expense of financial stability.&nbsp;  6.&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Painting a public picture of success while almost all financial indicators clearly showed the opposite.  &nbsp;  &nbsp;It is the view of the Southampton alumni that the LIU Leadership should be held accountable for the unacceptable financial performance of the College. We are exploring if the alumni should request a thorough review of the LIU leadership decisions by the NY State Board of Regents.&nbsp; This decision will be made by the end of July.  &nbsp;  It is also the view of the Southampton alumni that the LIU Leadership did not vigorously pursue alternative operating models that would maintain the undergraduate program at Southampton College.&nbsp;&nbsp; It is the alumniâ€™s belief that the current location on the eastern end of Long Island is an asset - not a barrier - as suggested by the LIU leadership at the July 7, 2004 meeting held with the alumni.&nbsp; While it is true that the sale of the Southampton campus would generate significant revenues to help LIU address its historical disappointing endowments and other funding issues, it is equally true that this site, with the appropriate undergraduate programs, staffing and leadership, would generate both healthy minds and a healthy financial position.  &nbsp;  We recommended to representatives of the leadership of LIU at the July 7, 2004 meeting that they work side by side with the alums and other stakeholders to revisit their plan and to explore alternatives.&nbsp; We also requested that they agree that their decision to relocate the undergraduate program is revocable.&nbsp; We are disappointed that given the urgency of this issue, we have not had a response.&nbsp;&nbsp; While our invitation to work together remains open, we need to pursue other alternatives now.&nbsp; Our next step is to form a team of alums, faculty, community leaders and other stakeholders to pursue privatization and/or an alignment with another University.&nbsp;  &nbsp;  &nbsp;  Southampton College Alumni Association Board  '),(16422,2839449,'14,July,2004','2004-07-14 00:00:00','   \"Welcome to the new Southampton College at C.W. Post! I will be your guide to this exciting new facility brought to you through the generosity of the Trustees of Long Island University! Before we begin our tour, does anyone have any questions?\"   \"This is a hallway.  Is Southampton College going to become a hallway?\"   \"I see that you didn\'t bring an open mind. You\'re seeing the glass as half full! Southampton College at C.W. Post isn\'t going to be this hallway! It\'s actually only going to be MOST of this hallway! And the first three classrooms on either side. At least on Tuesdays and Thursdays.\"   \"And all of the undergraduate programs are going to be taught in these three rooms?\"   \"Well, you see there was no point in moving the English, Social Science, Art or Business programs here, since they were already available at C.W. Post. Many of the Natural Science courses as well. What we are dedicating this space to, for at LEAST one semester, are those courses that were unique to the old Southampton College.\"   \"You mean like the Marine Biology and Writing programs?\"   \"Well, the Writing program was possible in Southampton because so many gifted writers lived in that area. We are currently looking for gifted writers who live on the North Shore of Nassau county, who while unwilling to teach at C.W. Post, but might be willing to teach at Southampton College at C.W. Post. If we find one, the Writing program may be offered.\"   \"C.W. Post is essentially landlocked. How are you going to teach Marine Biology here?\"   \"You don\'t have to go to France to learn French, do you? You have to be flexible. As you can see, the trustees have gone to great lengths to recreate the environment, to capture the feel, so to speak, of the original Southampton College.\"   \"How?\"   \"For example, we have moved the college dumpsters right outside these 2 classroom windows, to try to recapture the nauseating stench of the old Natural Science building. This sand all over the floor will give students the feel of the shore. And, that puddle that you all had to jump over to get in here will be maintained, rain or shine, in front of the entrance! After all, Southampton College is all about the water!\"   \"Is that all?\"   \"No. We hired an aging hippie to paint the words \"SMILE! BE FREE!\" in white paint of the roof of that tool shed over there. A permanent flow of foul grease, just like the one behind Wood Hall, will stain the road into the parking lot. And, of course, what would Southampton College be without the Windmill!\"   \"That\'s a rusty old floor fan!\"    \"Use your imagination! I mean, it\'s really old, and it\'s brown, and it has blades! In fact, this is better than the windmill out there, since it actually works! Just jiggle the plug. There! Can\'t you just feel those balmy breezes that evoke Autumn and Spring in the Hamptons... Now, let me show you the half of a shelf that has been set aside to store forever the trophies of the Southampton College Athletics Department...\"      Submitted by:    Bill Barbari  Southampton College Class of 1982  '),(16423,2839449,'13,July,2004','2004-07-13 00:00:00','   Like most alumni I was shocked and saddened by the announcement of Southampton\'s closing by LIU. My first reactions were anger and disbelief that once again Southampton had to play the Redheaded stepchild to the rest of LIU. My second reaction was the bastards were probably trying to sell the place off because of the value of it\'s real estate, and my final reaction was to try and figure out what I could do to stop it. I\'m heartened and pleased to see the reaction to try and save the college and stop this insanity. The university could save Southampton if it really wanted too, but from everything I\'ve read or been able to find on this subject it seems that the university board made up its mind and are not too willing to listen to any other suggestions. Why go through all the trouble of building a new library, planning a new marine station, renovating the natural science building? This option that they\'ve taken was not done so lightly or over a short period of time. You don\'t close down a campus without a lot of thought and this option has probably been floating around the board for longer than anyone will ever admit. To echo some of the other comments made on the BLOG, where was the call for help by the university and the college? In the last 18 years since I left Southampton I\'ve made occasional contributions as an alumni when I could, but I\'m just a marine scientist who works for a government agency, I don\'t have a bank account like Mr. Sillerman\'s. I have had the privilege to recruit a couple of promising marine science students to Southampton and though I lost touch with them I\'m glad that they\'ve already graduated because I would feel like I personally betrayed them if they were at the college when it shut down.  I have to admit I don\'t know what kind of place Southampton is anymore, I haven\'t been there in 15 years. My graduate education and work has taken me to many different places in this country, but one thing has always remained constant. Wherever there was a gathering of marine scientists (from all kinds of specialties), Southampton was almost always represented, even if it was just myself (which it usually wasn\'t). In all the places I\'ve worked and studied I\'ve never seen another institution that provided the combination of classroom training with amount of practical field training that Southampton does. Coupled with a dedicated faculty and the natural resources of the east end of Long Island, this program has turned out world class scientists for it\'s entire history. Does this program bring in lots of money for the university? Of course not, because it\'s not about the money, it\'s about the future. I certainly don\'t claim to be a brilliant scientist, but I love what I do, how could I do anything else. I think it\'s interesting that the development of the core curriculum was undertaken to provide a more interdisciplinary approach to learning, without all that stuffy academic core that made studying boring. Dr. Danziger hit the nail on the head when he wrote that it \"smelled like a dead fish and looked like a dead fish\". Marine science is an interdisciplinary field to begin with, you can\'t understand a biological or ecological process if you don\'t have a basic understanding the related chemistry and physics that allow that process to occur. Many people I\'ve worked with didn\'t get that kind of understanding until they got to graduate school, Southampton teaches this same material to undergraduates! Hence, our Fulbright scholars and many distinguished researchers all over the place.  I still don\'t exactly know what I can do to help save Southampton, I don\'t exactly live in the neighborhood anymore. But I\'m still going to try and I hope all the other Alumni out there feel and react in the same way. I would hate to see Southampton go down without a fight, as I\'m sure it won\'t!  Vita mutatur, non tullitur!   Christopher J. McDonough (Class of 1986)  Fisheries Biologist Marine Resources Research Institute South Carolina Dept. of Natural Resources  '),(16424,2839449,'13,July,2004','2004-07-13 00:00:00','   Rodger   urlLink posted this  in the comments:  Perhaps for those not in the New York area, there could be regional representatives to get information out and \"rally the troops?\" I\'d be willing to help keep contact going in Texas, if you can tell me where to get a list of Texas alumni and let me know how we can assist from afar. I also call on other former Student Government Association presidents to take up the cause as coordinators.  Anyone interested in serving as a coordinator for a regional area or a specific topic here on the blog, send me an email with your name and interest.  I also encourage everyone to get the word out! Send an email to other Southampton College alumni and see if they would like to get involved!  '),(16425,2839449,'13,July,2004','2004-07-13 00:00:00','   (From this week\'s  urlLink Independent Letters To The Editor )   Dear Editor;   Yes, to paraphrase Shakespeare, something is indeed rotten in the state of Denmark. I refer to the sudden, misguided and self-serving proposal by the Board of Trustees, headed up by President David Steinberg and his henchmen, to close the undergraduate program of Long Island Universityâ€™s Southampton College. Itâ€™s no secret now that the board, citing financial losses over â€œmany years,â€ has decided to terminate all but the minuscule graduate programs and move all undergraduate programs to the C. W. Post Campus. What is still a secret is the boardâ€™s thought process over the time leading up to the move, the real reasons for the purported $9,000,000.00 loss in the last academic year (the board has stated simply that declining enrollment is the culprit) and what actions were taken, to enhance enrollment and generate capital, before we arrived at this point.   What has become obvious is that the announced closing reeks of being a carefully calculated plan by some of the Board of Trustees to satisfy their agenda of liquefying their assets and drawing as little attention to themselves as possible. Before the ink was dry on the press release, we came to find out that the board curiously retained a group of real estate professionals to appraise the campus property to see how much it would garner on the open (and very lucrative) market. (Since this revelation, the sharks have been circling, as several different scenarios have been alluded to by various groups, all of which accomplish the boardâ€™s agenda: to fragment, spin off or sub-divide the property for purely financial gain.) Timing-wise, itâ€™s not surprising that the University waited until 95% of the students had gone home, some hundreds and thousands of miles away, where they could offer little in the way of organized rebuttal or resistance. Throw in the U. S. Open, when virtually the entire area was caught up in a mania of a different kind, and you have the perfect distraction for Steinbergâ€™s dog-and-pony show to roll into town â€” good timing indeed!   Rather than spending his days looking for ways to clean up his financial mess and keep the college open, Steinberg seems to find the need to rationalize his and the boardâ€™s misguided decision. At a recent function held at Southampton College to honor the president of the Peconic Land Trust, John Halsey, as the collegeâ€™s man of the year for his many accomplishments in land preservation, Steinberg chose to spend most of the time at the microphone defending his and the boardâ€™s decision to close the college and sell off the assets, rather than pay tribute to the guest of honor. Was this his sly way of telling Peconic Land Trust that the college property was up for grabs and they could have it if the price was right? Or was this, in light of dissenting public opinion that was growing by the minute, merely his guilty conscience showing through? Needless to say, more than a few were offended by his disrespectful monologue that took away from Mr. Halseyâ€™s evening. Curiously (and admirably), another board member, Harry Marmion, stepped to the microphone, unscripted, and made it clear to the crowd that not all was as it seemed, that he and some other board members did not feel the same as Steinberg. Stay tuned.   I happened to be with a number of Southampton students, freshmen, sophomores and juniors, the night of the devastating announcement. A sense of shock and sadness, accompanied by a fair share of tears, pervaded the room, but what was most disheartening to me was the looks of helplessness, fear and disbelief that their college â€” their college! â€” could betray them in such a way. These were kids who just 24 hours earlier knew that they had found an academic, artistic and social environment where they could thrive, develop and give back to a community that many of them fell in love with at first sight. To see this type of enthusiasm and vibrancy struck down by a few people who forsook the human element in favor of the bottom line was devastating to me, and truly unconscionable. These students made a commitment to the college and the college made a commitment to them. The Board of Trustees, in my view, has an obligation to these students. The offer to relocate them to C. W. Post or to offer academic counseling does not offset the issues of incompatibility in program offerings (last I checked, Post does not have a nationally recognized Marine Science program, probably the biggest draw at Southampton), lost athletic scholarships (what happens to the students who canâ€™t afford to lose that money when Post canâ€™t absorb them on their teams?) or lost transfer credits and money for students who donâ€™t want to move to Post, students who are better suited to a smaller, more rural setting, to name a few. The Board of Trustees, thus far, has failed miserably in addressing these issues.   The economic impact on the East End, particularly Southampton Town, will be staggering. The College is a significant source of employment, with many administrators, faculty and staff residing locally. The many local organizations who benefit from the numerous co-op placements from the college will be devastated. Local merchants, restaurants, pizzerias and such look to the college crowd for a needed financial â€œshot in the armâ€ during the long winter months that fill the gap between seasons. Were any of these points considered before the Board of Trustees lowered the boom on 1,200+ students and thousands of residents who will be directly affected? It doesnâ€™t seem likely.   Culturally, the area stands to lose a potential art gallery, a theater group that produces both Broadway and student authored plays, a location for events such as The All for the Sea Concert, Meet the Writers, Pianofest, the Master Theater Class with Alec Baldwin, lecture seriesâ€™ by area writers and scientists and much more. As for the environment, the East End will lose a great resource and a center for research that in the past has provided valuable insight into the brown tide outbreaks that plagued the local waters. College co-op students and graduates have worked with local foundations in the areas of endangered species, particularly the piping plover, and the rescue and rehabilitation of sea turtles, seals and other sea mammals. These losses, in an area like the East End that truly reflects a fine arts and liberal arts philosophy, would be most tragic.   Here are some things for the Board of Trustees of Long Island University to address, ponder and try to explain to its students and the community at large before following through with their plan (many of us in the community have already done so):   1) Southampton College lost $9 million in the last academic year and $77 million overall? As a recent graduate of Southampton College, my tuition as a full time student was around $23,000 annually, WITHOUT room and board, student activities fees and whatever miscellaneous fees they managed to add on. That would be around $31,000 for the average student staying on campus. While the campus is nice, it is modest in what it has to offer, much more modest than that of C. W. Post, yet their overall tuition is LESS than Southamptonâ€™s. We pay more money for less, yet we canâ€™t seem to balance the books. Someone at the college, purportedly in the know, related to me that it was his understanding that the tuition proceeds from LIUâ€™s campuses are pooled, averaged then distributed and the college gets back less than the original $31,000 paid by the student. This would explain a lot. Are the Post and Brooklyn campuses financially healthy? Please, Dr. Steinberg, correct me if my figures are wrong. A lot of very interested people would like to know the truth.   2) Hereâ€™s my thoughts on declining enrollment. It doesnâ€™t happen overnight. Neither does increased enrollment. The University set out a year or so ago to address the enrollment issue by revamping departments and curriculums in order to make Southampton a more appealing option to potential incoming freshmen. I must say, in my last semester, I didnâ€™t see any tangible evidence indicating that this plan was in place, but that also doesnâ€™t mean it wasnâ€™t. In any event, a year plus does not seem like enough time to wait to see the fruits of your labors. Who decided to throw in the towel?   3) Other than the aforementioned efforts, what else has been done to attract potential students? Being a married father of two, Southampton was an attractive option for me, logistically (certainly not financially) speaking, as I needed to remain close to home while I completed my degree. Gee, I wonder if there are others out there like me? Why hasnâ€™t the University aggressively targeted my demographic, the adult wishing to complete a four-year program close to home? For many of us non-traditional students on the East End, Southampton is the only viable option. How about a tuition discount for parents who wish to return to school? Relax pencil-pushers, the discounts would be made up over time with the increased enrollment. Other colleges on the Island use radio and print ad campaigns to help recruit new students, young and old. I canâ€™t remember the last ad of this nature Iâ€™ve seen that came from Southampton. Or how about an alliance with the Riverhead Campus of Suffolk County Community College, maybe develop continuing programs that compliment the Associates degree and give those students a local option? â€œBuild it, and they will come.â€   4) Has the Board of Trustees enlisted any help or insight from local, county, state or federal agencies, if not for outright financial assistance than for leadership and guidance? Itâ€™s not embarrassing to ask for help when you need it. Surely Southampton isnâ€™t the first educational institution to find themselves in dire straits, and I doubt theyâ€™ll be the last. If politicians and governing bodies have not been enlisted to this point by the trustees, I assure you they will be by the community.   I am not naive. I realize that a private university is a business, and as such, money needs to be made. Strong financial leadership is important, and many if not most of the Board of Trustees have backgrounds in the financial sector â€” investment firms, banking, etc. This fact is reassuring, and at the same time troubling to me. I fear that the art of making money has (or will soon have) transcended the art of education. We as a society need to build more institutions that provide fertile ground for learning and the opportunity for the individual to grow. Funny, but we already have one of those places right here. Right between Montauk Highway and County Road 39 in Southampton. Letâ€™s not let such a good head start go to waste.   If this letter seems written only out of bitterness, I assure you it is not. I graduated from Southampton in December 2003. While there, I had nothing but positive experiences. The faculty and staff were knowledgeable, understanding and cooperative. I met many incredible people there, faculty, staff and students alike, and I carry those friendships beyond college. It would be an incredible waste to let everything that took so long to build come crashing down, all because of a few individuals who have lost sight of what really makes a college a successful entity â€” not the bottom line, but the people.    MARK E. STRECKER  '),(16426,2839449,'13,July,2004','2004-07-13 00:00:00','   We need a court injunction, as suggested by Dr. Baker. We have to be careful the land doesn\'t get sold from under our feet. There has to be time for raising money (which no one asked for during all these years, as mentioned, I believe by Alan Braveman). So many townspeople, staff and alumni would offer financial help. Not only that, Southampton Town (businesses) is sure to be very interested in not letting go of this College that has been an integral part--financially, culturally--of the community for more than forty years.  I am a Southampton College graduate (B.A. 1968 / M.S. 1976). Unfortunately, I now live in Florida, but I will certainly make a small donation. If there is anything I can do on the internet, or mailing stuff from here I would be available for that, or whatever. Let me know.  Sincerely,  Christina Italiani  '),(16427,2839449,'12,July,2004','2004-07-12 00:00:00','   SAVE SOUTHAMPTONâ€™S COLLEGE  â€œSouthampton College has come into existence because the people of Eastern Long Island wanted a college and were willing to do something about it.â€                                           Chancellor John H.G. Pell 1963  On Friday, July 9th a broad cross-section of the Southampton College Community came together to formâ€ Save Southamptonâ€™s College.â€ We are administrators, alumni, community members, faculty, parents, students and staff. We have united to keep undergraduate education in Southampton. At that meeting, Theater Director Michael Disher urged the committee â€œto protect the future of undergraduate education. We are not here to save our own individual jobs. Our mission is much bigger than that â€“ we are here to fight for the students and to protect undergraduate education at Southampton.â€ The Childrenâ€™s School Director Kathy Bishop noted that our fight â€œhas already attracted the attention of elected officials, the media, and the business community. A groundswell of support exists for Saving Southamptonâ€™s College!â€ Professor Scott Carlin noted that â€œSouthampton deserves a 21st Century undergraduate campus that is intellectually vibrant and ecologically sound. Itâ€™s our responsibility to pursue that vision.\"  The founding vision for Southampton College has produced a campus rich in learning, culture, prestige, and economic value. We are:  â€¢	A nationally recognized Marine Science Program â€¢	A highly esteemed writers and artists Mecca â€¢	A forward thinking Environmental Program  â€¢	An extremely dedicated staff and faculty offering an integrated, experienced based curriculum â€¢	An institution of higher learning and thinking for a group of enthusiastic and  passionate future leaders of our global community. â€¢	A proud legacy of 37 Fulbright Scholars â€¢	The only year-round theater group on the East End â€¢	A tremendous resource that provides our community with extraordinary cultural, environmental and educational programs for children & adults â€¢	One of the larger employers on the East End â€¢	Substantial supporters of the local economy  â€¢	Your neighbor(s)!  Like the founders of Southampton College, we are â€œwilling to do something about it!â€ to make sure their vision continues to grow and mature. If we do not act now the East End will never be able to replace us in the future.   We must hold onto this valuable resource of ours and not let the Board of Long Island University â€œrelocateâ€ our students, faculty,staff and nationally recognized programs to Nassau County, nor can we allow our land to be sold to developers; Southampton College is so much more than a developerâ€™s piece of dream real estate in the Hamptons. As our mission statement states:  Southampton College is dedicated to creating an environmentally sustainable and humane world through critical thinking, creativity, and global citizenship. The curriculum emphasizes academic excellence, inter-disciplinary, and experience based learning, and full advantage of Long Islandâ€™s marine, environmental, literary, and artistic wealth. Our mission is to empower students to care for the world.  We must let our friends, families, neighbors, business associates and politicians know that we need their help to preserve one of the East Endâ€™s treasures that is about to be lost to major land developers. We will speak loudly and clearly: we must â€œSave Southamptonâ€™s College!â€  In coming weeks we will be announcing public events and other opportunities for the public to join us in celebrating Southampton College and building a new College for this new century!  * * *  '),(16428,2839449,'11,July,2004','2004-07-11 00:00:00','   I thought that all of us may propose another solution for LIU to raise about $60 million. I believe that the CW Post campus sits on about 300 acres of land â€“ in Greenvale, property equally as valuable as land in Shinnecock Hills. Why not suggest that they sell off about 50 acres of CW Post land, dispose of Steinberg who, according to the figures presented, has only raised $40 million in 20 years. When the Administration talks about the high discount rate, as they acknowledged, the discount rate at other colleges is not significantly different. The reason LIU and Southampton are in trouble is that thereâ€™s no endowment backing it; e.g.; compare LIU with Susquehanna College and Swarthmore. Anyhow, if there is a concerted effort by faculty, staff, the alumni and the community, we may still be able to save the College.   Bob D  '),(16429,2839449,'23,July,2004','2004-07-23 00:00:00','   Save Southamptonâ€™s College!!            A Community Forum'),(16430,2839449,'21,July,2004','2004-07-21 00:00:00','   Does anyone have any information about what will happen to some of the programs that are unique to Southampton College such as SEAmester&nbsp; and Friends World?  '),(16431,2839449,'21,July,2004','2004-07-21 00:00:00','   To all Southampton College Alumni:  Enclosed as an attachment to this communication, please find a statement of the Southampton College Alumni Association\'s position, on behalf of its membership, based on our discussions following the July 7, 2004 meeting with representative of Long Island University. This statement is being&nbsp; distributed&nbsp; electronically to&nbsp;Southampton alumni&nbsp;at the request of&nbsp; your &nbsp;Alumni Association President Greg Connly \'85. It may also be obtained by contacting the Alumni Office at 631-287-8348. &nbsp;&nbsp;     LIU Receives Failing Grade in Leadership  &nbsp;   Southampton College alumni reject the notion that the closing of the school is attributed to its location, purported alumni and community disinterest, and the inability of faculty to develop successful programs.&nbsp; The Southampton alumni, based on the limited information provided by the University, maintains that the fundamental cause of the collegeâ€™s problems can be attributed to decades of apparent mismanagement by the Long Island University leadership.&nbsp;  &nbsp;  For the period 1962 through 1990 a cumulative loss of approximately $10 million was generated.&nbsp; From 1991 through projected 2005, the cumulative loss will be approximately $67 million.&nbsp; Our assessment is that the accelerating decline in financial performance is attributed to questionable business decisions by the LIU leadership as summarized below:&nbsp;  &nbsp;  1.&nbsp; Failure to invest adequately in marketing resources and programs.  2.&nbsp; Insufficient capital investment and overall maintenance support.  3.&nbsp; Maintaining an operating structure that limited the ability of local leadership to be effective.  4.&nbsp; Operating a model that could not generate the required cash flow.  5.&nbsp; Operating a curriculum that generated Fullbright scholarships at the expense of financial stability.&nbsp;  6.&nbsp; Painting a public picture of success while almost all financial indicators clearly showed the opposite.&nbsp;  &nbsp;  It is the view of the Southampton alumni that the LIU Leadership should be held accountable for the unacceptable financial performance of the College. We are exploring if the alumni should request a thorough review of the LIU leadership decisions by the NY State Board of Regents.&nbsp; This decision will be made by the end of July.  &nbsp;  It is also the view of the Southampton alumni that the LIU Leadership did not vigorously pursue alternative operating models that would maintain the undergraduate program at Southampton College.&nbsp;&nbsp; It is the alumniâ€™s belief that the current location on the eastern end of Long Island is an asset - not a barrier - as suggested by the LIU leadership at the July 7, 2004 meeting held with the alumni.&nbsp; While it is true that the sale of the Southampton campus would generate significant revenues to help LIU address its historical disappointing endowments and other funding issues, it is equally true that this site, with the appropriate undergraduate programs, staffing and leadership, would generate both healthy minds and a healthy financial position.  &nbsp;  We recommended to representatives of the leadership of LIU at the July 7, 2004 meeting that they work side by side with the alums and other stakeholders to revisit their plan and to explore alternatives.&nbsp; We also requested that they agree that their decision to relocate the undergraduate program is revocable.&nbsp; We are disappointed that given the urgency of this issue, we have not had a response.&nbsp;&nbsp; While our invitation to work together remains open, we need to pursue other alternatives now.&nbsp; Our next step is to form a team of alums, faculty, community leaders and other stakeholders to pursue privatization and/or an alignment with another University. &nbsp;&nbsp;  &nbsp;&nbsp;  - Southampton College Alumni Association Board  '),(16432,2839449,'20,July,2004','2004-07-20 00:00:00','   This is from  Kathy Bishop , an alum that heads up The Children\'s School on campus.&nbsp; Her father was Dean Burke.&nbsp; Her husband is former Provost, Rep. Tim Bishop. &nbsp;  Joan, &nbsp; Please pass this on to others for me.&nbsp; I am very distressed by some of the negative outburst some of our faculty have had recently. They only \"cut off our nose to spite our face.\" It shows such a narrow view of the whole picture and it sends the wrong message out to those we hope will support us.&nbsp; We need to let the public know about the significant difference we have made in the lives of so many graduates and the wonderful effect we have had on members of all ages of our community from the Theater\'s Southampton Players, to The Children\'s School, the Creative Retirement and the many other programs. I think it is important to focus on the tremendous support and hard work that Laura and Bob Silllerman have given to this college; more than any other single board member has ever given to any campus of the university. I also believe that in spite of the lack of support from the majority of the LIU board Tim Bishop and many other hard working and dedicated members of the college community devoted endless hours to the survival of this college that has struggled from the start. Even now when we face our darkest times they are all still out there fighting for our continuation because they recognize our strengths and our value to the community. We as a college community need to say thank you to those who have fought for our survival and we need to stop the finger pointing in the public arena that only outlines one important reason why we will not be able to move forward. Anything positive that has ever happened here has been from hard work not negative rhetoric. &nbsp; Thank you, Kathy  '),(16433,2839449,'19,July,2004','2004-07-19 00:00:00','   From:  Don Baker  Subject: your photo-op &nbsp; Alan, Nora: Your comments brought to mind another thought: why not a petition signed by local merchants(maybe Riverhead, Hampton Bays, East Quogue, Bridgehampton, etc., as well, protesting the closing of the college, as it will have a decided impact on them.&nbsp; What about a rallying petition or something of the sort collecting signatures or comments (brief ones, but a list of them, from such as Spielberg, Alan Alda, and all the other film/theatre, etc. people out here that can be a brief statement with all their signatures, and run it in the local papers, on Channel 12, Newsday, etc.).&nbsp; And are there any theatre programs going on on the campus, where signatures to a similar petition could be acquired?&nbsp; And contact Ana Daniel, who is an adjunct history professor and one of the big-wigs in the Sag Harbor theatre group; she maybe (or others) could get similar signatures from those patrons of the arts.&nbsp; That\'s just the beginning: I\'m sure all of you can come up with additional local East End groups thatneed to get involved with their voices.  '),(16434,2839449,'19,July,2004','2004-07-19 00:00:00','   Written by  Donald Baker   &nbsp;  The Newsday article, unfortunately, simplfies the closing in terms of two things: economic problems (low enrollment) and Sillerman vs. Tilles, and doesn\'t really get at the fundamental issues, as the alumni and others have pointed out: inept management, evident for at least two decades, management that, unless one knew otherwise, could be construed as aimed at closing the institution. Clearly, at least in my opinion, the \"noble effort\" to move the college to C.W. Post as an act of compassion is, in fact, a charade, assuring the death of Southampton College in exile within two years.&nbsp;  &nbsp;  There are two issues that need to be addressed, hopefully by the Board of Overseers and all those concerned with preserving Southampton College:&nbsp;  1. Is Southampton College vital, essential, and invluable not only to Eastern Long Island but also to the kinds of students, from afar as well as local, that it could recruit? &nbsp;And how does it benefit the local area in terms of research (marine, environment, sustainable development, community problems and social change); in terms of working with local schools and programs and teachers; and culturally, with year-round programs as well as others that could be developed.&nbsp; And:&nbsp;  2. Is it financially viable as an institution. Yes, but not as part of Long Island University. And the yes would mean thinking creatively about incorporating new features: (a) a marine science program as part of SUNY but remaining here. but students would also takeliberal arts courses and live in dorms of SC as aprivate institution (see similar arrangements, where SUNY school of ceramics follows this model on the campus of private Alfred University; ditto for SUNYcollege of forestry on the Syracuse University campus;&nbsp; and SUNY agriculture for a long while as part of rivate Cornell University), with SUNY picking up costs for science laboratories, marine station, library resources, boats, etc., all part of marine science, hence not coming from the SC part of the budget; (b) an expanded liberal arts program, which costs very little in way of equipment etc. in contrast to marine science; (c) an expanded summer and year-round cultural activities program, in theater, journalism, writing programs, music, as well as lecture series (NYC\'s \"Y\" programs here) etc. especially in the summer but also year-round which could recruit funds especially from the NYC and arts/film etc groups that live part-time in eastern Long Island; and (d) as Dan Rodas had been working on, some kind of an arrangement where private entrepreneurial groups build new dormitories and lease them to the college, the dorms, including numbers of apartment style living arrangements as well as the normal dorms, all air conditioned, rental in summer to outside families as well (but first priority would go to students, possibly a summer theater and musical groups, such as at Saratoga in the summer), etc. The East End is reaching a \"critical mass\" of people of diverse backgrounds and interests, and they could be drawn upon, as well as the local community (as it begins to recognize the value of the college to the community), in terms of the recruitment of funds--that fund recruitment done here at the college, now independent(Steinberg said he \"went up and down Dune Road looking for funds, and no one would give\"--that attitude, and the indifference of the LIU fund recruiters over the years to Southampton College needs, as attested to by alumni all over the country, shows why development/fund activities for SC up to the present have been so miserly and need to be conducted at SC.&nbsp;  LIU\'s blatant closing of the college (Newsdayattests to that) reminds one of factories in the North over the decades that moved their factories to the south, away from labor unions. Or the \"outsourcing\" that is now going on by companies: move the establishment far away from those costly objects (faculty, students, etc) so they can\'t follow (how many eastern Long Island students are going to drive to Post when St. Josephs and Dowling, among others, are closer?).&nbsp;  &nbsp; Deja vu? Ask John Strong, who has written an article for Southampton Press on an earlier LIU leader by the name of R. Gordon Hoxie who tried to sell off the Brooklyn campus in 1947; his response to the Brooklyn students and faculty: \"Oh, Brooklyn\'s deficitis hurting the rest of LIU, so we must close it; sorry about that, but you students and faculty are welcome to teach or finish your program at Post (50 miles away from Brooklyn) or Southampton (100 miles away from Brooklyn)\". That dismembering of LIU was ultimately prevented, but it did lead to the firing of that leader before he could dismantle the university. The fact that Southampton College has lasted this long, given the caliber of leadership LIU has had over the past couple of decades, suggests that it can be a viable institution, and certainly viable if separated from LIU and restructured possibly along the lines suggested above--or something like this proposal. Possibly, from this&nbsp;restructuring, it might well be possible to recruit the necessary funds from the diverse groups that are in eastern Long Island. This would be \"their\" institution as they worked hand in hand with alumni and other groups to make it the kind of institution envisoned by Ed Glanz, who founded the college, and whose vision was of a college that focuses on its environment: a humanities program that incorporated the local artists and writers; a marine science program that dealt with the sea; a social science program that turned out students intent on working for social change and a better environment for all of us to live in; an education program that worked closely with the surrounding schools to assist them and their teachers in turning out students interested in making their lives and the lives of others better; and a business program that turned out people with the entrepreneurial spirit to participate in America\'s economic development.  &nbsp;  From:  John P. Petrone   Subject: Re: todays newsday  &nbsp;  This article disturbed me tremendously! LIU does not deserve to run SC anymore. Did everyone notice that Roger Tilles offerred to keep SC open if Mr. Sillerman kicked in five million dollars a year? The college was treated like a red headed stepchild from day one, and the fact that it appears that personalities and egoes played a role in this debacle is outrageous. I think Tilles is enjoying spearheading the termination of SC just to stick it to Mr. Sillerman. I am sickened!&nbsp; Has anyone approached Mr. Sillerman to let him know that there are eager interested parties out there that want to see an independent Southampton College? If the Save SC effort could gain Mr. Sillerman as a proactive ally in thwarting LIU\'s efforts to sell the college away, he would be a most powerful addition. I wonder if the Chancellor could be persuaded to join the fray on our side? How about approaching Tim Bishop to solicit our ideas to the Chancellor? Hey Mr. Sillerman, how does the title of President for Life of Southampton College sound?  &nbsp; Each time a person stands up for an ideal, or acts to improve the lot of others, or strikes out against injustice, they send forth a tiny ripple of hope, and crossing each other from a million different centers of energy and daring, those ripples build a current which can sweep down the mightiest walls of oppression and injustice.- Robert F. Kennedy---  &nbsp;  From:&nbsp;  Barbara Peters   &nbsp;  I felt sad, angry, betrayed. We spent so much time and energyworking on the new curriculum...often undoing one week what we had done the previous week and returning to what they\'d told us we had to do three weeks earlier. Everyday was filled with stress. We came to work, ready to teach and someone would come down the hall, shaking her/his head and we\'d say, \"Now what?\"  &nbsp;  I especially did not like the way Bob Sillerman was protrayed. Chancellors Hall could have been named Sillerman Hall, like the Tilles Center, but no . . . it gave credit and honor to past and future chancellors.  &nbsp;  Bob may not have been a trained educator, but he did understand that learning had more to do with learning to think than it had to do with numbers.  &nbsp;  SAD, SAD, SAD.  '),(16435,2839449,'19,July,2004','2004-07-19 00:00:00','   No matter what I do the formatting gets screwed up, so I apologize for the lack of paragraphs...  Thanks to  Rocksroy Bailey  and  Bill Barbari  for forwarding the article to us!  &nbsp;  East End  The little college that couldn\'t  &nbsp;  BY MICHAEL ROTHFELD AND OLIVIA WINSLOW  Staff Writers  &nbsp;  Robert F.X. Sillerman, a self-made billionaire who built his fortune in the entertainment industry, decided at age 45 to become chancellor of a plucky college in the Hamptons, where he hoped to inspire younger generations with the free-thinking lessons he had learned growing up in the 1960s.Chancellor Sillerman, who drew only $1 a year in salary, loved Southampton College so much he pledged $20 million toward the campus\' capital improvement project and raved to one and all about its innovative new undergraduate curriculum designed to encourage \"social responsibility.\"But in May, after serving more than a decade as the small college\'s top supporter and benefactor, Sillerman received stark news. His most personal venture was about to be decimated.Trustees and administrators at Long Island University, the college\'s parent institution, told him enrollment was down significantly and the school was hemorrhaging more money than ever. They said Southampton College would not survive without an annual infusion of $5 million -- implying that Sillerman could simply write a check each year to rescue his pet project.\"I didn\'t comment on it at the time because I thought that their behavior was actually erratic,\" Sillerman said.The trustees, headed by chairman Roger Tilles, a Great Neck real estate developer with whom Sillerman has a tense relationship, had watched in dismay for years as the little college drained revenues from their two more profitable campuses. Nonetheless, they gave Sillerman one last chance to open his wallet.\"If Bob had wanted to, there certainly was every opportunity to do that,\" Tilles said. \"And he has been very generous, and we just assumed that wasn\'t something he was willing to do.\"Sillerman went home and discussed the bailout plan with his wife. \"Laura and I made the decision that this group of people, acting and operating the way that they were, were not people who we wanted to be the sole solution for,\" he said.Shocking surpriseAnd so on June 15, with Southampton College\'s deficit for next year projected at $12 million, the highest in 40 years of losses, the trustees -- Sillerman among them -- effectively shut down the school. They voted to transfer its undergraduate programs to their C.W. Post Campus in Brookville beginning the fall semester of 2005.The decision marked the death of the intimate school founded in 1963 on the ideal that an academic community could thrive in one of Long Island\'s most stunning landscapes, buoyed by the wealth nearby. After scraping by for its entire existence, the college\'s final weeks came down to a standoff between two rich and powerful men. One of them likened it to a mortal duel.Why did Southampton College fail? The answers aren\'t so simple, according to interviews with dozens of people within the university system and the community.They point to the lack of an endowment, to mismanagement, a 40-year identity crisis, insufficient support from LIU and from the college\'s well-to-do neighbors, physical neglect of the campus, and the general complexity of operating an institution of higher learning on a budget of next to nothing. And out of that has resulted deficits projected to reach $77 million by the end of the next school year.Although the college produced 37 Fulbright scholars, most from a pioneering marine sciences program, and has won praise for its graduate writing program, the school struggled with its image and had trouble attracting undergraduates who could afford the $20,000 tuition.\"We have reached a point where the board was asked, \'Do you keep the patient going artificially or not?\'\" LIU president David Steinberg said. \"At what price to the rest of the institution, which is healthy? ... We would prefer to run a small, residential college on the East End. We don\'t have the money.\"The LIU board\'s decision to kill the college was so swift and harsh no one saw it coming, not even Sillerman. The new library was nearly complete, the new curriculum was about to debut, the students were on their way home for the summer. Sillerman said they were on course to turn the college\'s fortunes around.And yet Tilles and the 40-member LIU board of trustees had made their decision and given up on the college as constituted. Said Tilles: \"What we had planned even a year ago was not going to work.\"More than one trustee implied the board had grown weary of Sillerman and his creative ideas. \"This was a guy that just wanted to experiment with the university\'s money,\" said a trustee who asked not to be identified.Sillerman still has some fight left in him. He promises to be \"an obstacle\" if the trustees opt to sell the 110-acre campus for real estate development.\"I will be Aaron Burr to his Alexander Hamilton,\" Sillerman said, referring to Tilles.Local leaders had long dreamt of founding a high-quality liberal arts school on the East End. They convinced LIU to take on the venture and raised money toward the purchase of the property. But from the beginning, the stewards of Southampton College struggled to make ends meet and had a difficult relationship with LIU overseers.Lovely siteThe college sits on a series of gently rolling hills that are bordered on the north by the renowned Shinnecock Hills Golf Club, four times home to the U.S. Open, and on the south by the lush Native American reservation of the Shinnecock tribe. Despite its proximity to some of the region\'s most lavish estates, the school lost money right away, running an $87,000 deficit in 1962, the year before admitting its first class of 249 students. Within a few years, Southampton\'s liberal arts college began to assemble a strong contingent of scientists on its faculty and developed a marine science program with a national reputation, in which students had access to underwater laboratories, studying brown tide and seals wintering in the Long Island Sound.The school cultivated a loftier image than LIU\'s campuses in Brooklyn and Brookville; those places offered trades-oriented programs, but Southampton rejected cooking or hotel classes that might have made money.\"What we found was a commitment to a quality, liberal arts education and not just preprofessional training,\" said Jeffrey Kane, LIU\'s vice president for academic affairs.But the college could not compete for students with elite schools such as Wellesley and Haverford. Its student body actually had more in common with LIU\'s other campuses, in that many who went were the first in their families to attend college and needed financial aid. Jason Rodriguez, 20, a junior from the Bronx, said he chose the school because he has to pay only $2,000 a year. \"I wouldn\'t be able to go to college without a ridiculous amount of loans,\" he said.Without an endowment, the school relied almost entirely on tuition for sustenance. Of LIU\'s current $40-million endowment, only $1.4 million is for Southampton. And enrollment, which fluctuated between 1,000 and 1,300 over the last two decades, did not pay the bills, especially with a financial aid burden equaling about 35 percent of tuition. In the last school year, Southampton spent an estimated $31.1 million, 38 percent more than its $22.5 million in revenues.Harry Marmion, Southampton\'s president in 1972, said there never seemed to be much of a vision from LIU headquarters in Brookville. \"It was inadequate planning, abrupt planning, no planning,\" he said. \"They built a gym, a brand-new gym, gave me the keys and said, \'You pay for it,\'\" said Marmion, 73, who still lives nearby. Mandates to cut kept coming: eliminate the president\'s job, suspend the academic dean for a year, trim financial aid. They let facilities age and did not build new ones. The library and athletic facilities were high school-caliber, the dormitories were run-down and small. There were only 13 or 14 classrooms on campus. A 1993 report by a college accreditation team called Southampton a \"leader in experimental learning,\" but said inadequate facilities \"impacts quality of teaching and student learning and student life.\"Tim Bishop, the Southampton Democratic congressman who worked at the school for nearly three decades, rising from admissions counselor to provost, said he begged the university for help but was rebuffed. \"There was not apparently sufficient resources on the part of the university to invest,\" Bishop said. \"And as a result, we continued to slip.\"Steinberg agreed: \"You can\'t build with spit and Scotch tape,\" he said.Niche schoolBishop decided to cater to the school\'s niche, marine science, which he said was \"without any question the most attractive program we had.\" Nearly three-quarters of Southampton\'s students came for that program, he said, though some switched to other majors.But the approach had problems. On campus, faculty members and students in other disciplines resented the financial favoritism given marine science, which political science professor Donald Baker called \"a massive sinkhole.\" By dropping courses in other social sciences and humanities, the school lost potential students, he said.Even in marine science, there were not enough resources. Sandra Shumway, the program\'s former chairwoman, said by the 1990s, Southampton had lost its competitive edge. While administrators were \"believing their own publicity, these other programs were moving forward at a rate of knots,\" said Shumway, now a researcher at the University of Connecticut.\"The bulkhead [at the marine station] was falling in, and they didn\'t have any money to repair it.\" When prospective students came, Shumway wouldn\'t take them to the marine station because she feared \"they would be put off.\"With all aspects suffering and the post of chancellor vacant, Bishop turned in 1993 to a friend: someone he had met 15 years before through a co-worker at the college; a man with whom he had played volleyball on the beach; a flamboyant, wildly successful entrepreneur. He turned to Robert F.X. Sillerman.Sillerman, 56, has a dramatic way about him. He quotes Oscar Wilde, compares himself to mythological figures and refuses to say what the X initial stands for. In fact, he gave it to his 10-year-old daughter, Mackinley, and has yet to let her in on the family secret.He grew up in Morningside Heights in Manhattan and graduated from Brandeis University outside Boston. He quickly established himself as a shrewd and savvy businessman, buying and selling radio stations, amphitheaters -- including the one at Jones Beach -- and ticketing companies, and he also managed broadcasters and athletes.In 2000, while he was executive chairman of SFX Entertainment, Clear Channel Communications bought the publicly traded company for $4.4 billion in stock and debt. He had risen on his own ingenuity.\"It\'s funny, people ask me what I\'m trained in,\" Sillerman said. \"I guess if I\'m trained in anything, it\'s marketing, but you know, I\'m self-taught.\"Sillerman lives in Manhattan, but he and his wife, Laura Baudo Sillerman, have developed a network of friends in Southampton since the 1970s, where they own an oceanfront home. It turned out practically everyone they knew had close ties to the college. \"Our circle of friends out here had been derived exclusively from people who devoted their sweat and blood to this college,\" Sillerman said.Sillerman is a good friend to have. He told Newsday he is paying for the college educations of 11 children of his friends, including the Bishops\' two daughters. Out of devotion to the same crowd, Sillerman said, he agreed to become chancellor. \"We started off kidding around about the job,\" Bishop recalled. \"And I said, \'Would you be interested?\' And he said he needed to know more, but he said he would be interested.\"Sillerman, who took a helicopter to Southampton on his first day of work, said he had wanted to help students think creatively and ask questions rather than give rote answers. And he saw an opportunity to mold this little college that lacked history or tradition.\"It still is very young, and so there was no need to pay homage to the past,\" Sillerman said.By the time Sillerman took over, Southampton College had lost $16.5 million in its first three decades. It has lost $48 million more in his 11-year tenure as chancellor.Entertainment connectionsSillerman began by recruiting, through his entertainment connections, stars such as Paul Simon, Tina Turner and Bob Dylan, to All for the Sea, the college\'s annual summer fund-raising concert Bishop started in 1992. The event raised up to $1.2 million in a single year.Traditional fund-raising was much harder, as professional fund-raiser Tom Lloyd learned when he arrived on campus in the mid-\'90s to solicit for the capital fund. The campaign Lloyd led raised $13 million over five years, half what the university had hoped for, in part because summer residents were more interested in New York City charities than the local college. \"They wanted their children to go to Amherst and Yale and Harvard,\" he said.The alumni base -- a few thousand young graduates in nonlucrative fields -- didn\'t provide much support, either.The best source was Sillerman, but exactly how much of his own wealth he has given makes for fuzzy math. He estimates $15 to $25 million, but some trustees say it was closer to $10 million. They acknowledge he also brought in donations from other contacts and events such as the concerts.Agent of changeSillerman was not a typical donor, both in that he helped make policy and tried to leverage his gifts, requiring for instance that the university find $2 million in exchange for $1 million from him. Sillerman saw himself as an agent of change, not just a cash cow.\"Bob is arrogant, very assertive, sometimes bull-headed,\" Shumway said. \"He didn\'t get where he is by being John Q. Milquetoast, but those aren\'t necessarily bad assets.\"By 1996, the school at last could afford to build something. Chancellor\'s Hall, a modern academic building, opened in 1998.That same year, in an effort to capitalize on the abundance of authors in the area, the school started a graduate writers\' program boasting Roger Rosenblatt, a well-known essayist with a home in Quogue, as its marquee faculty member. (The college paid him $217,000 in 2001.) The program, which spun off an annual writers conference in July, has become extremely successful, with authors such as Peter Matthiessen, E.L. Doctorow, Frank McCourt and others.But then, Sillerman said, he decided to market the school nationally and raise admission standards in the hope of building a smarter and more affluent student body. The bold move backfired: enrollment declined, from 1,312 in 1999 to 1,198 in 2000.Sillerman admits now this was a mistake, that he was influenced by the buzz in the community: \"There was a lot of, \'Oh my gosh, I thought it was a community college, I thought it was part of the state university system, I had no idea it\'s as great as it was ... \' And I overestimated that.\"Tilles and Sillerman clashed often. Sillerman looked out for Southampton\'s interests and Tilles presided over the entire university, but people at Southampton always felt he favored the C.W. Post and Brooklyn campuses and some even said he stocked the board with others with similar loyalties. They are both rich, enjoy high profiles, have healthy egos and are given to philanthropy, according to those who know them.Tilles, 58, who earned his law degree in Michigan, returned to Long Island in 1983 and joined his family\'s real estate business. The Tilles Center for the Performing Arts at the C.W. Post Campus is named for his parents.Sillerman operated in flashier circles and had created his own wealth.\"I think you had two very powerful men who sort of jockeyed for position,\" said a person who has worked with both men. Sillerman was reportedly furious when Tilles, a Democrat, at first backed Republican incumbent Felix Grucci in the 2002 congressional race.And Sillerman wasn\'t happy after Tilles and the board commissioned a study the same year to ascertain how Southampton College could get out of the red.The study\'s conclusion: the college needed a $100 million endowment to generate $5 million in interest. Sillerman scoffed, calling such a solution \"fairy-tale land.\"John Kanas, an LIU trustee, bank president and Southampton graduate, said a committee discussed closing the campus and \"what changes might be made to try to stem the leakage.\" Sillerman, Kanas said, went into marketing mode and \"convinced the remaining trustees that the way to go about bringing change in Southampton was to repackage the curriculum.\"The new vision, announced last year, included a capital campaign. The school would undertake a $60-million building plan to include a library and learning center and a new gymnasium. The Sillermans agreed to contribute $20 million if LIU would put up the rest.\"His ego was driving the process,\" said a trustee who asked not to be named. \"And the board, since he threw some money out there on his own and hung out the prospect of giving them even more, they were sort of seduced by it.\"Huge capital expansionSillerman\'s salesmanship and offer of $20 million launched the huge capital expansion and set new programs in motion. He was clearly excited about the school\'s future just two months ago as he talked with members of the campus community about the building plans and dynamic new curriculum, which would \"empower students to care for the world.\"He spoke of \"the vision of the new campus, the new curriculum,\" recalled student Nancy Boland. \"How it was new and exciting and it would take, oh, three to five years\" to implement.Sillerman could already point to the nearly completed brick frame of an airy new library that was to be called the Bishop-Burke Library Learning Center for the congressman and his father-in-law, former dean Bill Burke. (In all, 10 members of Bishop\'s extended family have been employed at the college.)But the college\'s financial picture was continuing to unravel. Enrollment, which had been 1,276 in 2002, when the plan was hatched, was 1,105 in the school year just ended, and was not expected to increase next year. The annual deficit, as a result, increased to $8.6 million. And next year\'s deficit was projected to hit $12 million. Sillerman got the bad news the same day he visited campus, he said.Tilles said he and the trustees wanted the plan to work, but they could not afford to see it through. \"There were several assumptions that were made,\" he said. \"One was that as we invested more money in the campus and changed the curriculum, we would attract more students and wouldn\'t lose existing students, and that assumption was a bad one.\"According to Tilles, the goal was to keep the deficits to $5 million or reduce them. Instead, said Steinberg, the finances spun out of control: \"It\'s just that they\'ve reached a point where they have gone through a zone, like the sound barrier,\" he said.Sillerman, however, said the faculty strike at C.W. Post last fall, which hurt LIU\'s finances, made trustees less willing to fund Southampton\'s losses.For his part, Sillerman said he chose not to resuscitate the college with a $5 million check because the LIU board reneged on its commitment to the $60-million revitalization plan.\"If people were going to act that way, then count me out,\" Sillerman said Wednesday.As a trustee, he went along with the vote to transfer most of Southampton\'s programs to C.W. Post.To date, Sillerman has turned over less than $1 million of the $20 million he promised LIU for Southampton College\'s capital campaign.In Southampton, the community has been reeling since the news broke. \"We as a community weren\'t ready for it,\" Edwin Schwenk, an East End activist, said. Southampton Village Mayor Joe Romanosky Jr. has vowed to do whatever he can to get the decision reversed. The alumni say they feel robbed.The university has said its graduate programs, in education, writing, business administration and accounting, will remain in Southampton. Some faculty will be invited to C.W. Post, but others and nonacademic employees will surely lose their jobs or retire. The future of the marine science program is unclear, although there have been discussions about taking it over with the state-run Stony Brook University, which has a similar graduate program.LIU has hired a law firm to explore selling the property, and Tilles says it will probably be sold, though for what purpose has not been decided. Sillerman, meanwhile, insists it will remain the site of an educational institution. \"There will not be a Wal-Mart there, I can tell you that.\"Last week, LIU officials met with Suffolk County Executive Steve Levy to discuss a venture with Suffolk Community College. Southampton\'s Parrish Art Museum -- which had planned to build a $35-million to $40-million museum on campus -- is reportedly exploring its options.Sillerman refused to divulge his plans, but he hinted he would make them public by the next trustees meeting in October. Would he remain chancellor? Would he donate another dollar to the university? \"Ask me that in October,\" he said, explaining he does not want to disturb the \"healing environment.\"Sillerman said he was \"deeply disappointed\" by his own flawed decisions and in the university for breaking its word. Wearing shorts and a polo shirt and sporting a tan in Chancellor\'s Hall, he seemed resigned and sad.\"There really was an opportunity to do just what I had hoped ... to redefine higher education in the 21st century.\"  Copyright Â© 2004,  urlLink Newsday, Inc.  '),(16436,2839449,'19,July,2004','2004-07-19 00:00:00','   Where did you live during your freshman year?     The Montauks&nbsp;&nbsp;64%&nbsp; 47 votes    The Peconics&nbsp;&nbsp;14%&nbsp; 10 votes    The Feds&nbsp;&nbsp;12%&nbsp; 9 votes    Off-Campus Rental&nbsp;&nbsp;3%&nbsp; 2 votes &nbsp;   With the Folks&nbsp;&nbsp;1%&nbsp; 1 vote    Transferred to SC after freshman year&nbsp;&nbsp;4% 3 votes    I crashed on any available couch or floor&nbsp;&nbsp;1%&nbsp; 1 vote    73 votes total    Yay! Montauks! Okay, so I\'m a little biased... New poll will be up soon - looking for suggestions!  '),(16437,2839449,'19,July,2004','2004-07-19 00:00:00','   From:  Lisa Freese  Subject: Re: steiburg at orientation &nbsp; You know in Steinberg\'s 1st meeting with faculty &amp; staff he said... \"years ago, they had asked the community for help &amp; got little response so, now they would not ask anyone!!! they would not knock on doors\"  &nbsp; From:  John Petrone  Subject: Re: steiburg at orientation &nbsp; I would wager if we took a poll of local residents to find out to whom Dr. Steinberg was referring to, I have a sneaking suspicion we would find that he knocked on nobody\'s door.&nbsp; We are going to require the support of local residents if we are going to save this college. They should have a vested role as well. Here is an idea, perhaps if we polled them as to which they would prefer, another luxury housing development and/or summer resort, or keep it an institution of higher learning, and then have the results published. I could venture to guess the results. That would add additional pressure on the \"LIU Gang of non-trustees\" In addition whomever conducts the actual polling could also put out feelers as to whether the local residents are willing to help save the college as well. &nbsp; From:  joantuttseashell@aol.com  Subject: Re: SC POLL &nbsp; LIKE YOUR IDEA! How to implement? Any suggestions? &nbsp; From:  Barbara Peters  Subject: Re: Re: SC POLL &nbsp; As part of our efforts, we are planning a big event, no date yet set, but we\'d like it to be after the students return, so . . . the first weekend after Labor Day. &nbsp; The event will celebrate Southampton College history and effect, based on \"It\'s a Wonderful Life\" with Jimmy Stewart. One of the things that we are going to request is for people, and that means anyone, send in their \"evidence\" \"rememberances\" and so on re: SC. How would their lives have been different without the college here? How have their lives been impacted by the presence of the college. These can be big and little ways and can be sent to us as videos, audio tapes, photos, scrapbooks, writting testimonials and so forth. &nbsp; These things will be on display at the celebration. And there are plans to make a video based on \"It\'s a Wonderful Life.\" (Made byProfessionals)  &nbsp; I will talk to the committee over the weekend about sending out a press release ASAP.&nbsp;  &nbsp; This is open to everyone. It might be more effective than a poll even. We have to work fast though.  '),(16438,2839449,'19,July,2004','2004-07-19 00:00:00','   NEWSDAY SUNDAY, JULY 18, 2004 by Michael Rothfeld &amp; Olivia Winslow &nbsp; FRONT PAGE: Shutdown 101 &nbsp; How the struggle between a chancellor\'s lofty ambitions and the LIU board\'s financial worries doomed the East End campus (with photo of Phillips Gate &amp; Southampton Hall in background AKA Pratt Center) &nbsp; Page 7: The little college that couldn\'t Despite visionary ideas, regime changes and an abundance of good intentions, in the end not even a billionaire could save Southampton (arial photo of campus) Southampton College\'s 110- acre campus, where lush, rolling hills contrast with outdated buildings in need of upkeep.&nbsp;  &nbsp; \'We have reached a point where the board was asked, \"Do you keep the patient going artificially or not?\" \'~LIU president David Steinberg (photo)  &nbsp; \"There really was an opportunity to do just what I had hoped...to redefine higher education in the 21st century.\' ~Robert F.X. Sillerman(photo) &nbsp; Subtitle: Shocking surprise &nbsp; Page A 38 The fall of Southampton College &nbsp; Graphs entitled Budget Blues under college sign  subtitles:&nbsp; Lovely site &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Niche school &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Entertainment connections &nbsp; Page A 39 (photos of Southampton Hall, formerly The Tucker Mill Inn, Sillerman in front of partially completed library, Roger Tilles, Rep. Tim Bishop) subtitles:&nbsp; Agent of change &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Huge capital expansion &nbsp; Caption under photo:&nbsp; Above, the school\'s main administration building. Sillerman, right, at the new library, scoffed at the conclusions of a 2002 study that said the college needed a $100 million endowment, calling such a solution\"fairy-tale land.\" &nbsp; Alumni: there is one sentence in 4 full pages of this article about us...6 words: \"The alumni say they feel robbed.\" &nbsp;  www.NEWSDAY.COM  ~Joan &amp; C.L.A.S.S &nbsp; Note: The Southampton College Board of Overseers meets tomorrow.  '),(16439,2839449,'29,July,2004','2004-07-29 00:00:00','   Save Southamptonâ€™s College!!  A Community Forum        Sunday, August 1, 2004&nbsp;&nbsp;    5:00 to 7:00 pm'),(16440,2839449,'27,July,2004','2004-07-27 00:00:00','   Dear Bloggers,  I sent a letter to New York State Senator Caesar Trunzo and would like to share his response with you.&nbsp; Please write to your&nbsp;elected officials and ask that they support&nbsp;us in our &nbsp;efforts to \"Save Southampton\'s College!!!\"        Dear Ms. Detweiler:   Thank you for your letter regarding the decision by Long Island University to close undergraduate programs at Southampton College.   I would like to take this opportunity to state that I understand and share your concerns. I have been in contact with Senator Kenneth LaValle, Chairman of the State Senate Education Committee, whose district Southampton College falls under.&nbsp; Senator LaValle has been meeting with state and local officials to discuss the future of Southampton College.&nbsp; The agreement is that this institution should be kept as an educational facility.&nbsp; Please be assured that I will continue to work with Senator LaValle to address this issue.   Once again, thank you for taking the time to write and share your concerns with me. If I can be of further assistance, please feel free to contact me.   Sincerely,  Caesar Trunzo  Senator&nbsp;  '),(16441,2839449,'26,July,2004','2004-07-26 00:00:00','   My family hails from Shelter Island and for almost three years I served as a member of the administration of Southampton College as Director of Housing & Residence Life.  During this time I was a member of the \'President\'s Council\' which was the senior administrative group at the college under president Don Wilson. Prior to assuming the position at Southampton, I held professional staff positions at the State University of New York and at Alfred University, both in Alfred, NY, while earning my undergraduate and graduate degrees from Alfred University.  I share this as it gave me a perspective on how three different higher educational institutions operated.  I can without a doubt state that Southampton was one administrative mess.  I am truly amazed it took LIU as long as it has to try and \'escape\' from the money pit the college has been for them .... not that it should or had to be that way.   On July 13, I took a day off from my day job to visit both Shelter Island (where my family has our roots) and the college I called home (literally as I lived in Bridgehampton\'s basement apartment for two years).  That apartment was where my spouse and I called our first home too!     As I reflect back on my years at Southampton the following thoughts come to mind:    1. MIS-MANAGEMENT  - I wish I could say I was part of the solution and not the problem, but I quit in disgust.  The administration and the multiple unions are part of the reason for possible demise of the college.  Years after leaving the college, I was still fond of telling a story I called the \'Southampton Screwjob\'.  Before my arrival on campus someone selected a new wall light for the dorm rooms that utilized a theft resistant screw to lock the (Tupperware white) globe to the wall unit.  To change a light bulb required three work orders: One to the engineers to remove the screw, a second to change the bulb, and a third to replace the screw.  In truth it was three parts of one work order, but the two unions involved (engineers and custodial workers) would not work together and the engineers would not allow the custodians to use a screwdriver necessitating three visits to the dorm room to complete a light bulb change.  This is a true story, the end result was by the time I left the college all the globes were in some state of disrepair as our students unable to remove the theft resistant screws would rightfully take matters into their own hands.  This situation was true to many ways in which things worked \'differently\' at Southampton.    2. LONG DISTANCE MANAGEMENT  - Long before distance learning was in vogue, \'distance management\' was the way of the world.  Back then, the each of the college\'s had a \'president\' and the big cheese (now called the president) was the chancellor of the university system.  Don Wilson was president during my tenure at the college and tried to run the college, but the chancellor and his staff were intent on controlling the place.  They even had a \'Dean of Administration\' (Joe Carbone) who reported directly to LIU Headquarters.  Poor Don Wilson, had to ask Joe for permission to do anything that cost a penny.  Don Wilson had a vision, and whether one agreed with it or not, he was never empowered to lead Southampton.  (One shining star while I was at Southampton was Tim Bishop.  I am sure he walked this tightrope as best he could as Provost. In many respects he was a visionary even back then when he was, I believe, the director of financial aid.)    3. DIFFERENT MISSIONS  - Faculty, administration and students did not share a common understanding of a long term plan for the college.   So ... how does Southampton College continue into the future:    A.  The alumni, students, faculty, and loyal local administration should band together to continue Southampton as a private college focusing on it\'s strengths in marine science and liberal studies.  (I was attracted to work at Southampton as I felt the environment would be similar to Alfred University which has a rich tradition of liberal arts, ceramic art and ceramic sciences ... more on that to follow.)    B.  Contact every state legislator, the governor, the Chancellor and Board of the SUNY system to establish a \'Statutory College (or School) at Southampton College\' specializing in Marine Sciences.  The program has the reputation to win this status and financial support.  Such a designation will: 1) Provide capital to improve and maintain the facilities of the marine science program; 2.) Provide credibility to the college in general; 3.) Possibly provide short term capital funds to improve the campus.    C.  Eliminate Unions - if everyone truly wants to see Southampton continue it can be done without Unions.  Concessions will need to be made on compensation at all levels.  But one way to pay the greatest net earnings to the dedicated professionals who will make the college continue is to eliminate the percentage the union takes away from the college.    D.  Lease out the dormitory system to a professional management company, but only if they can guarantee the college a net operating income that exceeds the net profit (if any) the dorms generate today.  Having personal knowledge of the quality of these facilities (and they look like they haven\'t changed in twenty years) I would probably think they could be torn down and replaced with new and attractive apartment residences for students.    E.  Consider adopting a \'residential college\' concept where faculty live on campus in significant numbers and interact with the student body as in a living and learning environment.  This could be one of the single biggest attractions to the school for prospective students, i.e.: knowing they will have significant contact with their professors both inside and outside the formal classroom.    F.  Seek legislation to allow the college to run the housing facilities on a for profit basis from Memorial Day to Labor Day each year. When I worked at the college we offered what I called \'continuing education\' classes that were notoriously non-attended and we ran a Jitney (van) to the beach everyday.  Most of these \'students\' were solely there for cheap rooms ($7 per night in 1981) and a way to enjoy the beach.  Well guess what ... rather than hide behind the \'non-profit\' status, seek legal authority to run all or part of these facilities as motel rooms at a fair rate that would really contribute to keeping the college in the black.  (The rates today for August 2nd, one night in Southampton listed on Orbitz range from $190 to $260 for a one night stay.  People would pay $100 per bedroom per night for a properly outfitted suite of three bedrooms, i.e.: $300 per night!)   I live in Texas now and am too far away to help Southampton survive into the future, but the college will be in my thoughts and prayers in the year ahead.  Good Luck in all your endeavors!    Ken Thompson  '),(16442,2839449,'26,July,2004','2004-07-26 00:00:00','   I just posted a new poll for you to take. Note that it states:   What was your major when you  STARTED  at Southampton?   There will be a follow-up poll question about what your major was when you graduated from Southampton.   Still looking for more contributors to the blog - just send an email to  urlLink SaveSouthampton@excite.com  and you\'ll be invited to be a contributor through  urlLink Blogger .   So far, I\'ve only heard from Rodger Baker in Texas who would like to organize some kind of group there. And then there is me here in Maryland and Nora, Joan, and the  urlLink Yahoo group  in New York. If you\'re interested in starting&nbsp;a local group to Save Southampton College, email me with your name, contact info and location. Once we have more interest and contacts, I\'ll post it on the side bar.   urlLink Let\'s fill up the map!     Another update : 27 shirts have been purchased via the  urlLink CafePress store ! Hope y\'all are enjoying them - I love mine! If anyone has other t-shirt ideas, please email them to me and I can create more items.  '),(16443,2839449,'26,July,2004','2004-07-26 00:00:00','   This post has been removed at the request of the author.  '),(16444,2839449,'06,August,2004','2004-08-06 00:00:00','   urlLink  Group pleads on land use     By Michael Rothfeld   In a show of political force, a bipartisan band of East End elected officials Thursday declared that the Southampton College campus \"must\" be preserved for educational purposes and not sold off for residential development.  Long Island University, the college\'s parent organization, announced in June that it would move the school\'s undergraduate programs to its C.W. Post Campus in Brookville in the fall of 2005 because the Southampton campus had lost $65 million since it opened in 1963.  On July 19, a committee of university trustees began exploring the fate of the 110-acre campus, which is zoned for residential development. The university plans to announce its intentions after the trustees meet in October.  \"The elected officials are unified that the campus must, and I underline the word \'must,\' remain an educational campus,\" said State Sen. Kenneth LaValle (R-Port Jefferson), chairman of the Senate\'s committee on higher education.  The others who gathered at Southampton Town Hall included Assemb. Fred Thiele (R-Sag Harbor) and Republican Southampton Supervisor Patrick Heaney, both alumni of the college; U.S. Rep. Tim Bishop (D-Southampton), the college\'s former provost; Suffolk Legis. Jay Schneiderman (R-Montauk); and Kevin Law, chief deputy to Suffolk County Executive Steve Levy, a Democrat.  Long Island University officials have already held discussions with representatives of the state university system and Suffolk County Community College. LaValle called talks between LIU and the state \"productive, positive and ongoing.\"  The elected officials Thursday appeared optimistic about a possible deal with the state, particularly about a merger between Southampton\'s undergraduate marine sciences program and a similar graduate program at the state\'s Stony Brook University.  \"At state university tuition, Southampton College becomes an incredibly attractive educational alternative,\" Bishop said.  A university spokeswoman, Paola Curcio-Kleinman, declined to comment. In a statement released by LIU on July 29, board chairman Roger Tilles said the committee dealing with the campus \"stated its clear preference to preserve the 110-acre site, if at all possible, for educational, cultural and not-for-profit services.\"  Tilles, who did not return calls Thursday, previously said the trustees had \"a fiduciary responsibility\" to recoup some of the money the college had lost over the years through a sale of the land and they wanted to determine its \"highest and best use.\"  James C. Barnett Jr., a Huntington appraiser, said that means the most profitable viable use.  \"Highest and best use is for residential development -- that\'s the most gorgeous piece of property,\" Barnett said.  But Southampton College Chancellor Robert F.X. Sillerman suggested an educational use might be more economical because a new academic building and library wouldn\'t have to be torn down. In any case, he said, LIU is \"under no obligation to take the highest price.\"  Copyright (c) 2004, Newsday, Inc. August 6, 2004  '),(16445,2839449,'05,August,2004','2004-08-05 00:00:00','   urlLink  Suffolk politicians trying to keep Southampton College for education use       urlLink    Watch the Video     Some Suffolk politicians are urging administrators at Long Island University to keep the Southampton campus for educational use.  A supervisor, a legislator, a senator and a congressman all say they have ideas and influence to keep Southampton College going after the undergraduate program moves to Old Brookville in September 2005. Officials say the campus is zoned for residential use with a special exception for education. Southamptonâ€™s town supervisor says things could get complicated if LIU decides to try and sell the land for development.  LIU released a statement in which the chair of the board of trustees said there is a â€œclear preference to preserve the 110-acre site if at all possible for education, cultural and not-for-profit services. However, it is incumbent upon the Board to explore all options.â€  No final decision will be made about the fate of Southampton College until the board of trustees of LIU meets in October.  ************************   WHEN  and  WHERE  in October will the Board of Trustees be meeting?? I feel a protest coming on...  '),(16446,2839449,'05,August,2004','2004-08-05 00:00:00','   urlLink   A Human Face          The Personal Consequences of the Closing of Southampton College    By David Stoll   Much has been written about what will happen to the campus of Southampton College and its unique programs after it closes in the summer of 2005. These are very important issues, but they represent only part of the story. This report spotlights the human equation. The closing of Southampton College will affect individuals of all stripes in a broad variety of ways, all unhappy. And Southampton College cannot fix these problems, as the problems were engendered by the very way that the closing was handled. Here are some of the groups of people who find themselves in impossible situations. These are real people whose stories were told and confirmed during the research of this article. We believe it is accurate.  First, though, it is important to make the following point. Long Island Universityâ€™s C. W. Post campus has much to offer, both in terms of facilities and programs. (Indeed, some have questioned how Southampton College can be such a drain on the LIU system given that the C. W. Post campus receives so much more money.) When comments are made in this story about people not wanting to attend C. W. Post, it is not meant as a criticism of C. W. Post, but merely an acknowledgement that different people have different interests and desires. Someone who wants to live in Southamptonâ€”a small town during the off-season and a resort area during the summerâ€”may not wish to live in one of New York Cityâ€™s suburbs, or its boroughs. (Brookville is a mere 40 miles from NYC, and LIU has a Brooklyn campus too.) Likewise, others came to Southampton College for the Marine Sciences program or the ability to attend the Writers Conference each summer. Letâ€™s face it: whatever C. W. Post may offer, lectures by Jules Feifer and Roger Rosenblatt will not be part of the curriculum.  Class of â€™08 Scholarships. The Southampton College Class of 2008 will come in September. Many of those students have scholarships. When they agreed in April to come here, they turned down scholarships at other colleges. Now they have only two choices: to spend a year at a college as it disappears and then transfer to another college, hoping to obtain another scholarship, or to go to LIUâ€™s C. W. Post campus. These students and parents are devastated, as the students do not wish to go to C. W. Post. They enrolled at Southampton College for a reason.  Librarians. Southampton College has three tenured librarians. Tenured faculty cannot be fired, but these librarians have been told that there are no jobs for them at C. W. Post. Instead, they will be offered jobs at LIUâ€™s Brooklyn campus. With no disrespect to that institution, the Southampton College librarians did not celebrate when told they must now work in a major City at a campus with an entirely different academic program.  Scholarship Athletes. There are a number of athletes now at Southampton College, a Division 2 school, who are on scholarship. While they have been assured that they will keep their athletic scholarships at C. W. Post, they have been informed that there is likely not to be space for them on C. W. Postâ€™s sports teams. These are not people who are looking for a free ride. These are talented kids who love sports and are willing to devote substantial time to trying to achieve victories for their college. They have been benched.  Tenured Faculty. The faculty has been reminded that tenure is granted by LIU, not Southampton College. That means that, just like the librarians, the academic faculty can be sent to C. W. Post or Brooklyn, and many faculty members are hearing whispers that they will be sent to Brooklyn. Thus, their choice is (1) to commute over 65 miles to C. W. Post or over 100 miles to Brooklyn, (2) to move from the East End to New York City or its suburbs, or (3) to look for a new job. Some choice for someone who has chosen to make a life on the East End.  It has been rumored that the plan to close Southampton College has been in the works (quite secretly) for some time. LIU professors are unionized and command higher salaries than many of their counterparts at other colleges. People have remarked that the College has not replaced retiring professors in various disciplines. Indeed, the Political Science department has shrunk from a high of nine professors to a mere two today. Thus, there is a feeling that the closing of Southampton College and the whispers of sending people to Brooklyn are tactics to encourage resignations and early retirements. This article does not confirm or refute conspiracy theories, but it is a fact that many faculty are looking for jobs at colleges that offer an environment similar to that at Southampton College rather than move to Brooklyn or even C. W. Post.  Local Students. Quite a number of students at Southampton College grew up on the East End and live with their parents while in school. These students cannot afford student housingâ€”even subsidized student housing. They will have no choice but to drive over 65 miles to C. W. Post (or even nearly 100 miles to Brooklyn).  Low-wage Workers. Southampton College is the largest employer on the East End, and many of the employees are low-wage workers, without tenure, such as groundskeepers, janitors, cafeteria workers and secretaries. Once their jobs disappear, there surely will be insufficient new jobs on the East End for all these people. Thus, if they are â€œluckyâ€ enough to be offered jobs at C. W. Post or Brooklyn, they will be left with the choice to commute, or else to uproot their families and move. For the working poor, this is plainly a Hobsonâ€™s choice.  Student Housing. Those Southampton College students who elect to go to C. W. Post will be housed at an â€œadjacent campus,â€ according to official statements. What this means, it turns out, is a SUNY campus in Old Westbury, and students apparently will be transported to LIU by bus. Thus, students who came to Southampton College for a familial atmosphere and in some cases for the ability to walk from home to school now must be segregated from the rest of the student body and instead live among students at a completely different institution. These students feel as though they will have fallen from the status of favored child to that of unloved step-child. To make matters worse, Old Westbury has a reputation (whether or not deserved) for crime problems.  Recruiting. The official story is that Southampton College, with its own programs, will continue to exist at C. W. Post; only the location will have changed. However, it was announced at the end of last week that the Director of Admissions of Southampton College is being transferred immediately to C. W. Post, leaving Southampton College without an admissions director to recruit the high school class of 2004. Moreover, the recruiting department has been instructed not to print brochures or otherwise to undertake efforts to bring students to Southampton College (wherever it is located). The feeling is that the institution is being left to die.  Incoming Juniors. As the closing of Southampton College was announced only recently, many students about to enter their junior year in September will be unable to transfer to another college this year. They will study hard this year and apply to transfer for the fall of 2005. The problem is that an enormous number of colleges and universities will not accept more than two years of academic credits for its incoming transfer students. So, the juniors this year not only are unable to graduate from a school at which they will have spent three years, but they will not even be able to graduate in May 2006.  Transfer Students. There is at least one person who was to have transferred to Southampton College from another college effective this September. She has been told not to do so, but has been offered a place at C. W. Post where she has not interest in attending.  * * *  The purpose of this story is to associate human faces with the decision to close Southampton College. The closing of Southampton College would be a blow to the East End as a unique place however accomplished, but it is also important to realize that the particular problems described above are caused not just by the closing but by the way the closing was handled. Indeed, it is said that the Trustees of LIU originally voted to close Southampton College as of August 2004, but were then asked to consider the chaos that would ensue. Their response was to delay the closing for a mere year. That is simply not enough time to prevent so many sad consequences to so many innocent people. And to add insult to injury, the decision was announced in the summer during the U. S. Open, when so many students and local residents were away or in hiding. That is simply disrespectful.  '),(16447,2839449,'05,August,2004','2004-08-05 00:00:00','   Using the links and information on the side bar, I wrote to Sen. Clinton and Sen. Schumer yesterday. The response from Sen. Clinton said that she would not be able to respond to me since I do not live in New York State. Sure, I lived there for 25 of my 26 years, but that doesn\'t matter now that I live in another state. :(  I ask that alumni in New York State take a moment to write to Senators  urlLink Clinton  and  urlLink Schumer . Their online web emails are also linked on the side bar.  If you live near the college, in NY District 1, your Congressman is  urlLink Tim Bishop , former Provost of Southampton College. Email him at  urlLink timothy.bishop@mail.house.gov . If you are in another District, you can  urlLink go here  to find out who your Congressman/Congresswoman is.  Contact your representative in the  urlLink New York Senate  too! Just  urlLink enter your zip code  to find out your representative\'s contact information. For District 1 near the college, it is  urlLink Kenneth P. LaValle .  Don\'t forget the  urlLink New York Assembly .  urlLink Enter your zip code  to find your Assemblyman/Assemblywoman contact information. For the 2nd Assembly District near the college, it is  urlLink Fred W. Thiele Jr .  More contact information is located on the side bar for the NY Governor, Southampton Mayor, etc. Get writing!  '),(16448,2839449,'04,August,2004','2004-08-04 00:00:00','   To All Alumni  At the request of the Alumni Board I have been coordinating our efforts regarding the announcement by LIU to relocate the undergraduate program at Southampton College. This has not been an easy task given the sheer number of alum\'s and how dispersed we are. For this and prior documents I have tried to capture the prevailing views and to focus on critical items only. Staying focused has been the priority. This document fulfills the obligation I signed on for and it is time to pass the responsibility on to another alum. This person should have relevant leadership experience and be prepared to dedicate the time. Those interested should express their interest to me ( urlLink mrobert168@aol.com ) and Greg Connly, President of the Alumni Board ( urlLink gconnly@dolanconnly.com ).  In closing I wish to express my thanks to all the alumni for the discipline and maturity in which they have addressed this issue. I wish to thank Carol Gilbert for her support and balance during a very difficult period. I thank Mickey Director and Peter Philipps for the contributions they have made at our meeting on July 7. Likewise, thank you Matt, Alan, Louis, Pete, Larry and the many others that took the time to provide input. Finally a big hug and thank you to Joan Tutt whose heart and soul keeps us all motivated and moving forward.  Bob Martin        Southampton College Future Update July 29, 2004        Update:     Based on a three hour meeting with the LIU leadership on July 7th, 2004, the Southampton College alumni rejected the notion that the closing of the school was attributed to its location, purported alumni and community disinterest, and the inability of faculty to develop successful programs. The Southampton alumni, based on the limited information provided by the University, maintained that the fundamental cause of the collegeâ€™s problems could be attributed to decades of apparent mismanagement by the Long Island University leadership. As of now nothing has been provided that would suggest that we should reverse this position.  It was also the view of the alumni that the LIU Leadership did not vigorously pursue alternative operating models that would maintain the undergraduate program at Southampton College. It is now apparent that during the time leading up to their decision they did not pursue negotiations with various Universities to maintain the undergraduate program. However we are told that this will change going forward. They have recently stated that their primary goal now is to preserve the current site of Southampton College for educational and cultural purposes. It appears that they seriously underestimated the backlash from the various stakeholders. They now recognize that the sale of the land for residential or commercial development is not their best option. They have appointed a Special Committee, co-chaired by John Kanas (class of 68 and a member of the Board of Trustees), to identify and pursue appropriate candidates and to lead the negotiations. Discussions are in progress with at least two institutions.  We recommended to the leadership of LIU at the July 7th meeting that they work side by side with the alums and other stakeholders to revisit their plan and to explore alternatives. As of now, the LIU leadership has not accepted our offer to work together. The Board of Overseers of Southampton College has recently made a similar offer to the LIU leadership. The Board of Overseers consists of individuals whose primary interest is the welfare of Southampton College. The members include alums, significant contributors, educators and some members of the LIU Board of Trustees.  We requested at the July 7th meeting that they agree that their decision to relocate the current undergraduate program is revocable. They have confirmed that their decision is irrevocable.      What Next:     While the alumni and other stakeholders have good reason to request a thorough review by the NY State Board of Regents, it may not be the appropriate action to take at this time. Our primary objective should be to ensure that an acceptable undergraduate program is provided at this campus and that the name Southampton College is preserved. With this as an objective our efforts will be better served by supporting the Special Committeeâ€™s effort to negotiate the sale of the Southampton College campus to another institution. Given the track record of the LIU leadership and the LIU Board of Trustees, this strategy contains an element of risk that is sure to make us all uncomfortable. It is therefore imperative that we monitor their progress and be prepared with alternative positions in the event that this strategy fails to satisfy our objectives.      Supporting the Special Committee:     As a first step, and until we are better organized, we will advise the Committee chair that we will support the Special Committeeâ€™s effort. The most important role we can play is to proactively promote the Southampton location in the most favorable way - one that welcomes institutions of higher education and that provides a favorable and friendly operating environment. The tone has to be positive by all. Elected Officials, special interest groups and others should not try and capitalize on the collegeâ€™s failure. Public positioning to further and individual or groups own interests will only risk negotiations with other institutions and jeopardize our goal of maintaining a Southampton campus.            Alternative Positions:     In the event that our objective is not achieved we need to be prepared on a number of fronts:   1.  Privatization of the school has been ruled out for now because of cost and the time it would take to obtain accreditation. If that became the last option however, we would need to have an education model that was financially viable. One that would attract students and endowments.  2.  Recourse options need to be identified. This should be limited to a select few and maintained confidential. Publicity at this stage would only jeopardize the Special Committeeâ€™s effort to negotiate and put at risk our objective.  3.  Funding to explore alternative positions is required.  4.  We need to be organized.      Getting Organized:     First we need to select an individual to lead the alumni. Bob Martin agreed to lead phase one which was our meeting with the LIU leadership. We now need someone to lead the next phase as identified above. All interested candidates should forward their qualifications to Greg Connly, President of the Alumni Association.  Second, the leadership of each stakeholder group needs to meet and agree on objectives and strategies. The faculty has appointed Scott Carlin to chair their effort. Bob Sillerman chairs the Southampton College Board of Overseers. We are not aware if the business community, residential community or elected officials have organized in a similar fashion. The Board of Overseers (Bob Sillerman) should take the lead of this initiative by coordinating the stakeholders on the following:   1.  Define expectations - communicate to the Special Committee the required outcomes from the sale of the campus to an appropriate university. This should be limited to three or four points and it should not be a laundry list.  2.  Communicate to the Special Committee the support of all stakeholders subject to their agreement to our outcomes as well as their agreement to keep the Board of Overseers abreast of their effort.  3.  The Board of Overseers will keep the leader of each stakeholder group informed of progress and issues. Given the nature of negotiations, a high degree of confidentiality will need to be maintained by all.  4.  Conducting a public relations campaign that positively positions Southampton as an ideal location a college campus.  Third, the alternative positions should be organized into two groups: Privatization and Recourse.   1.  Privatization should include members from each of the stakeholder groups. The output from this initiative is alternative education models that are financially viable. Each model must be highly marketable to potential students and financial supporters from the community. This effort should be co-chaired by someone from the faculty and someone from the community. Harry Marmion is uniquely positioned to pull this together and act as an advisor going forward.  2.  Recourse should include an attorney from each of the stakeholders, with the exception of the Board of Overseers. The exclusion of the Board of Overseers is to avoid a potential conflict of interest for those members who are also on the Board of Trustees. This group should be small and Michael Director will co-chair this initiative. Attorneys and CPAâ€™s who want to help should email their interest and background to Michael Director ( urlLink directorlaw@optinonline.net ).      Funding:     We will require funding to finance:   1.  A PR campaign that continues to build widespread support for our effort.  2.  The costs incurred to prepare â€œalternative optionsâ€ as defined above.  3.  Privatization as a last resort.  Funding requirements and strategies for the first two initiatives will be different from the third initiative. The first two initiatives will require an estimated $30,000. This is a very rough estimate and could increase as we learn more. Our target audience is all stakeholders. Our strategy here is to raise small contributions via widespread support. Selling T shirts, alumni donations, raffles etc. are examples of tactics we would deploy. From the alum side, Joan Tutt will lead this effort. The other stakeholders need to appoint someone. We will need to establish ourselves as a nonprofit organization.  The third initiative is to raise at least $100 million. Before we can start soliciting candidates who are in a position to contribute at a significant level we need a product to market. Once we have a financially viable model we can start to solicit interest. In the interim we can identify prospects and develop a strategy to approach each candidate. At this point it appears that the Board of Overseers is best positioned to lead that effort.      Summary:     Since our meeting on July 7th, the LIU leadership has adjusted their position somewhat. Subject to their agreement to our expected outcomes, we will support their new position. In the event that they do not deliver on these outcomes, we will be prepared to take action with alternative positions that will achieve our objectives and hold the LIU leadership accountable. In the interim, all stakeholders need to take a public position that reflects positively on Southampton as a location for higher education.  '),(16449,2839449,'04,August,2004','2004-08-04 00:00:00','   urlLink July 23, 2004, Newsday    Can college go forward?   Is it time to write Southampton College\'s obituary? Or is the college undergoing a transformation that will allow it to go forward in a new form with a renewed spirit [\"The little college that couldn\'t,\" News, July 18]?  I\'m a Southampton College adjunct professor who teaches a course titled \"Environmental Ethics: Western & Indigenous World Views.\" In class, we\'ve learned how different cultures value land. Our culture values property as a resource from which to earn a profit. But indigenous cultures believe humans should seek union with the land\'s spirit to gain insight concerning its proper use.  As a former city planner, I\'m familiar with the planning principle that land should be put to its \"highest and best use.\" The best use of property is that which expresses the highest spiritual value of a culture. Can the human spirit and spirit of the land that this college rests on be joined so that an undergraduate college continues? I believe it\'s possible. The real question is: Are we up to the task?   Peter Maniscalco     urlLink July 29, 2004, The Southampton Press   (just links to articles - you will need an online subscription to view them)    When Southampton College holds its commencement next May, the Class of 2005 will be joined by one graduate with a unique perspective.   urlLink Go to article.    Long Island University and State University of New York officials are close to striking a deal to bring SUNY Stony Brook students to the Southampton College campus in 2005.   urlLink Go to article.     urlLink July 30, 2004, Dan\'s Papers    Correction Maybe: Did LIU Really Make A Commitment Not To Sell The College Land?   By Dan Rattiner    This is a little behind the scenes stuff. In our edition last week, I wrote an article about Southampton College, in which I said that the owner of the college, Long Island University, was pledging to never allow the 100-acre campus upon which the school sits to be used for anything else but educational purposes. And this was a very good thing. As you know, two weeks earlier, LIU announced that the college will close in September 2005.   Perhaps you noticed that no other newspaper wrote anything about this pledge. Danâ€™s Papers had the story. It wasnâ€™t in Newsday, the New York Times or any of the local weeklies. How come? Did we make it up?   We did not make it up. Nor did we get it wrong. But we do want to apologize, sort of. Or write a correction, slightly.   The sequence of events was this.   On Monday morning (11 days ago), the LIU Board met privately to discuss Southampton College. Danâ€™s Papers had no idea what was said at this meeting.   But Danâ€™s Papers does know what two members of that Board said when they met with the Southampton College Overseers later that same day. The Overseers, I hasten to say, have no power or authority whatsoever. They are a group of men and women who ought to have some control, but who are told politely by the Board that they are â€œadvisors,â€ as in, advise us on what you think. What some of the Overseers thought, in the absence of anyone telling them what they were being told was top secret, was that if Danâ€™s Papers called them they would let the paper know what the LIU Board members told them.   According to the Overseers, the two LIU Board members told the Overseers what was decided by the LIU Board earlier in the day. My information came from two different Overseers. What they both said was that the LIU Board members told them that 1. the LIU Board was making a commitment to never sell the campus for residential and commercial use, but always keep it for educational use, 2. they were irrevocably in agreement that the school would be closed in September 2005, 3. they would immediately open negotiations to sell 10 acres to the Parrish Art Museum, and 4. they would form a search committee to seek offers to continue providing an educational experience on the site, 5. They would issue a press release on Wednesday to the effect of what was said in number 1 - 4.   Based on this information, Danâ€™s Papers published a heads up on what the LIU Board members said was decided at the morning meeting. We trusted that the Board would never mislead or lie to the Overseers and we continue to believe that is true.   We went to press on Tuesday night. On Wednesday morning, Danâ€™s Papers edition was on the press being printed for Thursday and Friday distribution. We were committed.   No press release came out on Wednesday. One did come out on Thursday afternoon, and it mentioned the commitment to proceed in closing the school and it mentioned the formation of the search committee, but it only used such phrases as â€œour best effortâ€ and â€œour hopeâ€ that the land should continue to be an educational institution. There was no mention about selling anything to the Parrish.   What happened?   Well, I really donâ€™t know. I did ask several LIU Board members about the discrepancy, because I know them, but I was told they could tell me nothing as they had made a blood oath to never reveal what was said that Monday morning. Of course, the Overseers had made no similar oath.   â€œWhat if I ask you something and you just stamp once for yes and twice for no?â€ I tried.   Nothing doing.   And so, here is my speculation on the matter. I believe that the LIU Board DID make the commitment never to sell the property for commercial or residential use, but between Monday and Wednesday it had to go through the filter of $400-an-hour lawyers. And so what started out as a thunderbolt came out as a little poop, even though there was no follow-up board meeting to say it differently.   Here is what I think the lawyers said. â€œYou agreed on WHAT? You put that in writing, it will probably cost you thirty million dollars down the road. Just ten words. Thirty million dollars.â€   Their hearts are still in the right place. There are more meetings coming up this week. Time is short. Shoot the lawyers.   * * *  Hereâ€™s the latest news about the college: there isnâ€™t any. Meetings are going on all over the place. Everybodyâ€™s sworn to secrecy. My job is to let readers know what is going on. Itâ€™s an honorable profession. I wish I had something.    urlLink July 31, 2004, Newsday    Let\'s save this college   I am an adjunct professor of advertising and business at LIU Southampton and several other schools on Long Island and would like to say it is terrible what is happening to Southampton College [\"The little college that couldn\'t,\" News, July 18].  I truly hope that as a community we can stop the closing and stop all these vulture developers from closing our college and taking its beautiful land. For 43 years this college has been a source of culture, education, intellect and training for many eastern Long Islanders. It also supports the local economy. Without it, the local ecomony will be adversely affected due to higher unemployment and lack of disposable income to support local businesses.  About 400 people would lose their jobs. If our government and our local corporations sit idly by and allow this to occur, it will be a shame. If there are any kind corporations who would be willing to save this beautiful college, the college and the local community would be eternally grateful.    RÃ©agan Lorraine Lavorata     urlLink August 4, 2004 The Independents    Growing Outrage Over College Closure   By Carey London   With the impending closure of Southampton Collegeâ€™s undergraduate program looming on the horizon, an alarmed community gathered to speak on the issue with elected officials, college professors, students and alumni Sunday evening. The forum preceded a press conference to be held tomorrow at Southampton Town Hall at 10 a.m. featuring New York State Senator Kenneth LaValle, Congressman Tim Bishop, County Executive Steve Levy, and Assemblyman Fred Thiele.   A handful of panelists, along with Southampton Professor Scott Carlin as moderator, spoke of the intention not only to keep the Southampton campus an educational institution, but to maintain a four-year program as well. The land, they said collectively, will not fall prey to developers eager to conquer and divide what is considered to be some of the most valuable property in Suffolk County.   â€œI was very distressed when I first heard,â€ said Suffolk County Legislator Jay Schneiderman of the schoolâ€™s closure, adding that the 110-acre campus is a â€œprime piece of real estate.â€ Schneiderman entertained the idea of collaborating with Suffolk County Community College, in Riverhead, but â€œideally that wasnâ€™t the best solution.â€ Following Levyâ€™s vetoes to the county budget earlier this summer, Schneiderman re-established an authorization to spend $30 million on the Riverhead campus, much to the chagrin of some residents who felt the money could have been better spent elsewhere, such as at Southampton College. The SUNY system would be a better option, Schneiderman continued, because it would keep the facility a four-year educational institution. Levy could not attend the forum, but Ben Zwirn, his Assistant Deputy County Executive, did. Zwirn, whose son attends the college, said that he does not want to see the programs disbanded, especially the prestigious marine science program, whose fate has yet to be decided.   Considered to be one of the best in the country, the marine science programâ€™s success is in large part due to the collegeâ€™s surroundings. Because of Southamptonâ€™s on-campus marine station on Shinnecock Bay, students have unlimited access to bays, tidal creeks, barrier beaches, dunes, salt marshes, salt ponds and the ocean.  â€œThis school should be a star in the LIU system,â€ Zwirn proclaimed.    The Shinnecock   The forum provided a first for the Shinnecock Indian Nation, who, they said, had yet to be included in the decision-making process either publicly or privately. Tribal members Elizabeth Haile and Lance Gumbs joined the panelists. For Haile, Southampton College represents the opportunities not available to her as a young woman. In 1948, she enrolled in the Oneonta State Teachers College because there was â€œno such school on Long Island,â€ she said, citing long bus trips and a lot of time away from home. â€œIf there had been a Southampton College I could have walked to school, lived at homeâ€¦â€ And now, â€œit looks to me like weâ€™re about to go backwards in time about 50 years,â€ said Haile, who is also a Southampton College Doctor of Humane Letters. â€œThis issue is most serious. Time is of the essence.â€   And in what some believed to be an attempt to advance personal politics, Shinnecock Tribal Chairman Lance Gumbs spoke of the potential benefits to having an Indian Gaming facility on Shinnecock land, stating that some of the proceeds would go to funding the institution. The college, said Gumbs, has given a tremendous amount to the tribe and â€œitâ€™s time now that the Shinnecock Indian Nation give back.â€  He also spoke of â€œworking togetherâ€ with local politicians who have up until this point failed to include the Shinnecock in any of the brainstorming discussions about the college. The school is the largest source of employment for the tribe, said Gumbs, adding that losing this institution would be a devastating blow not only to the surrounding community, but also to the Shinnecock as well.    Irreplaceable History   A current student and a recent graduate also joined the panel. Nora Detweiler, a 30-something student, choked up at the thought of the school closing. And Sharon Wallen, an alumna and a pre-school teacher, remarked on the incredible â€œimpactâ€ four years can have on a person, adding later that, â€œthe love, dedication and history cannot just be replaced,â€ she said. â€œPlease give others the opportunity that I was so lucky and blessed to have.â€   Tim Bishop, a former Southampton College Provost, was the final panelist to speak prior to public commentary. â€œThe most important thing for us to do is focus on where we go from here,â€ he said. Although the discussions regarding the future of the school have been â€œcollegial, productive and positive,â€ there is no concrete deal in the works with SUNY Stony Brook, as erroneously reported in one of the local newspapers, said Bishop. â€œThe article was premature,â€ he continued, adding that they are â€œnot yet at that point.â€ In fact, in response to the article, Long Island University released a statement reaffirming the ambiguity of the collegeâ€™s future.  Once the roster of speakers concluded their position statements, the floor was opened to public debate.   Most audience members took liberty with their relegated three-minute time slot, opining on the state of affairs and offering suggestions to save the college. Changing the current zoning codes for the campus was one well-received idea. Currently zoned R-60, or residential, the property could be altered to a specific zone for higher education, thereby eliminating the risk of subdivision. Others agreed that politicians should include the Shinnecock in any decisions made for the school, commending Gumbs for his forthright speech.   Southampton Professor Helen Mendez said that this forum was the first time in weeks that she had felt any kind of hope of protecting the college from falling to the â€œcorporate Godzillas that are circling around usâ€¦You know what a jewel we have here,â€ she said.    Closing The Doors   Slated to close its undergraduate doors in May 2005, Southampton College will move affected students to the LIU CW Post campus in Brookville. Citing fiduciary responsibilities, the university has insisted that this is the most viable opportunity for reimbursing a $77 million debt accrued by the college, $17.2 million of which was accumulated in just the last three years.   But many are now asking for whom does the educational bell toll? Some have said it tolls for the 365 full-time employees who work at Southampton College. Others believe it tolls for the 1,196 undergraduate students currently enrolled and soon to be displaced. And still more have said the bell tolls for the residents who have relied on the school to stand fast as one of the three pillars in this bustling community. The other two, Southampton Hospital and the Parrish Art Museum, are also experiencing financial distress â€” the latter having recently finalized a deal with the school to build a facility on the campus.   Gumbs noted that regardless of having a unique campus situated in one of the countryâ€™s most affluent areas, LIU is still looking at closing this institution.  Dr. Avram Morrell, a United States Army Veteran of the Korean War and a graduate of LIU in Brooklyn in 1953, summed up the mood of the crowd. There are â€œjewels and diamonds in the rough, such as this college,â€ he said of the Southampton school. â€œThis is a shame...The university must stop cannibalizing this campus.â€     urlLink August 5, 2004, The Southampton Press   (just links to articles - you will need an online subscription to view them)    A deal between Long Island University and the State University of New York to bring SUNY Stony Brook students to the Southampton College campus is unlikely to be announced at a press conference scheduled for this morning in front of Southampton Town Hall, officials said this week.   urlLink Go to article.  '),(16450,2839449,'04,August,2004','2004-08-04 00:00:00','   SOUTHAMPTON COLLEGE FOR SALE? No way! Keep our number one center for undergraduate education, culture and jobs on the East End!  How can you help? The following leaders can protect Southampton College.  Please call, or email, at least one leader below, and thank him or her today!   And then please forward this to your friends. You can easily make a BIG difference. You should also email any evidence of fraud and mismanagement by LIU to the Attorney General and the Board of Regents (email and websites below).   Remember, it is an election year, and we should support whoever keeps the College on the East End.    NY Sen. Kenneth P. LaValle  (631) 696-6900 Email:  urlLink lavalle@senate.state.ny.us     NY Rep. Fred W. Thiele, Jr.  (631) 537-2583 Email:  urlLink thielef@assembly.state.ny.us     Steve Levy, Suffolk County Executive  (631) 853-4000 Email:  urlLink county.executive@suffolkcountyny.gov     Hon. Patrick \"Skip\" Heaney, Southampton Town Supervisor  (631) 283-6055 Web email:  urlLink http://www.town.southampton.ny.us/contact.ihtml     US Congressman Tim H. Bishop  (631) 696-6500 Email:  urlLink timothy.bishop@mail.house.gov     US Senator Hillary Clinton  (631) 249-2825 Web email:  urlLink http://clinton.senate.gov/email_form.html     US Senator Charles E. Schumer  631-753-0978 Web email:  urlLink http://schumer.senate.gov/webform.html     Governor George E. Pataki  (518) 474-8390 Web email:  urlLink http://161.11.3.75/     Attorney General Eliot Spitzer   (516) 231-2400 Web email:  urlLink http://www.oag.state.ny.us/online_forms/email_ag.jsp     Richard P. Mills, NY Education Commissioner  (518) 474-5844 Website:  urlLink http://www.nysed.gov/     New York State Board of Regents  Website:  urlLink http://www.regents.nysed.gov/regents.htm     Joseph P. Romanosky, Jr, Mayor of Southampton Village  (631)283-0247, Ext. 221 Website:  urlLink http://www.southamptonvillage.org/  '),(16451,2839449,'11,August,2004','2004-08-11 00:00:00','   urlLink   Officials Unite To Save College          By Kathryn Georgette   During a press conference last Thursday at Southampton Town Hall, attended by nearly 30 people, Senator Kenneth P. LaValle, Congressman Timothy Bishop, Assemblyman Fred Thiele, County Legislator Jay Schneiderman, Southampton Town Supervisor Patrick Heaney, and a representative from County Executive Steve Levyâ€™s office announced that they will unite under the effort to \"maintain an educational presence\" at the Southampton campus of Long Island University.  Further, they announced the goal of not only maintaining the campus as an educational site, but as a four-year undergraduate college. \"We are united behind the goal that a four-year college remains at the Southampton College,\" said Thiele. Long Island Universityâ€™s Board of Trustees announced on June 15 that Southampton Collegeâ€™s undergraduate programs would be relocated to the C.W. Post and Brooklyn branches of the University effective as of September 2005, due to the campusâ€™s projected $77 million cumulative debt.  Students and alumni of the college, who are fearful that the residentially zoned 110-acre campus will be razed and developed for real estate sale, have met the impending closure with protests. Despite fears that the collegeâ€™s Board of Trustees would lean towards a profitable real estate venture, Bishop said he is \"convinced that [keeping the campus as an educational facility] remains their interest as well. I believe the board will be determined to not subdivide, as feared.\"  Bishop acknowledged that \"technically, we have no power [over a real estate sale].\" Thiele commented that Southampton town is willing \"to undertake our own study of the uses of the property,\" and \"make recommendations on its highest and best uses.\"  On-Going Talks LIU has been in talks with officials in the state university system and Suffolk County Community College. Still, no agreements have been made, and discussions are \"on-going,\" according to Senator LaValle, who is also the Chairman of the Senate Committee on Higher Education. A hot topic at the meeting was a partnership between the Stony Brook University marine science graduate program and Southamptonâ€™s undergraduate program. \"There is a natural fit, a natural synergy between the two programs,\" commented Bishop.  The appeal for Stony Brook would come from \"the melding of a great undergraduate program with a graduate program in one of the most desirable, beautiful places in world, Southampton,\" LaValle opined.  \"At state university tuition, Southampton College becomes an incredibly attractive educational alternative,\" said Bishop.  County Legislator Schneiderman offered that although the state university appears to have \"the lead role\" in negotiations, SCCC would provide an additional route for the college to take. SCCC is in need of athletic facilities, as well as library space, according to Schneiderman, which the Southampton College campus could provide.  \"We will continue to discuss with LIU and our state partners how Suffolk County and our distinguished community college may play a role in keeping this site reserved for higher education,\" said County Executive Levy. \"There may be a role for the county to play,\" Schneiderman continued. \"But we donâ€™t want to stand in the way [of a deal with the state].\"  For many of these officials, this battle is both personal and professional. Bishop was the Provost for the college previous to his congressional campaign, and Supervisor Heaney and Assemblyman Thiele are both alumni.  \"The importance of Southampton College to eastern Long Island must not be underestimated,\" said Thiele. \"Itâ€™s not just the educational opportunities, the economic benefits, the jobs, and the cultural resources the college provides, but it is the opportunity the college has provided to so many individuals to fulfill their dreams.\"  Some in the audience wondered if only \"the dreams\" of marine science majors will be solved in this effort, due to the focus of the partnership with the Stony Brook Marine Science program during the press conference.  \"There have been informal discussions about what would happen beyond that,\" said LaValle. \"A lot of this is truly in the exploration stages, but as we explore, we will explore different synergies.\"  In regard to budget concerns, Thiele commented, \"I donâ€™t see the state budget standing in the way.\"  LIU trustees will not announce their decision regarding the campus until after their meeting in October.  '),(16452,2839449,'11,August,2004','2004-08-11 00:00:00','   urlLink Newsday - August 10, 2004      C.W. Post sees burst in enrollment     By Olivia Winslow   Long Island University\'s C.W. Post Campus announced Monday that it will have its largest freshman class in 20 years when the semester starts next month, with 950 students, including significant numbers with strong academic credentials and an increase in students from Long Island and Queens.  \"It\'s at least the largest class going back to 1984,\" when the freshman class numbered 1,034, said Gary Bergman, C.W. Post\'s associate provost for enrollment services. Bergman said freshman enrollment last year was 843 students. \"Freshman enrollment has been laboring in the 800s since around 2000,\" Bergman said.  One factor not triggering the increase this year is LIU\'s controversial decision to relocate undergraduate programs from its Southampton College campus to C.W. Post in Brookville next year, a move prompted by Southampton\'s mounting deficit, officials have said.  \"None of this increase here is from the Southampton campus,\" Bergman said. But Southampton has seen a drop in its incoming freshman class, likely caused by the relocation plan, Provost Dan Rodas said. \"We\'re expecting about 170 freshmen,\" Rodas said. \"Last year we were right around 200.\"  Bergman said the freshman enrollment surge at Post prompted the administration to cap it at 950 students and establish a waiting list. \"We\'re trying to hold it to 950,\" Bergman said, citing the college\'s interest in maintaining small class sizes, which average 18 students, that are a hallmark of Post, which has about 4,800 undergraduate students.  This year\'s freshman enrollment represents a nearly 13 percent increase over last year. \"We\'ve had incredibly strong interest from students from Long Island,\" Bergman said. He said 360 of the new freshmen come from Nassau County, 230 from Suffolk County and just under 100 from Queens. Those numbers are up between 25 and 30 percent over last year, Bergman said.  Bergman also touted the high academic caliber of many of the students. While SAT averages are not yet available, Bergman said that among the more than 45 percent of incoming freshmen receiving academic scholarships from the college, \"a lot of these students have 1200 and 1300 SATs and higher. They\'re the students that every college would love to have.\"  Bergman cited student interest in new degree programs, such as forensic science, which begins next month, popular programs such as teacher education, the recently-developed Access program that helps freshmen uncertain about a major to develop a field of study, and new facilities.  '),(16453,2839449,'24,February,2004','2004-02-24 00:00:00','   My version of the Bush gay marriage speech....  (be sure to imagine the voice and the accent)  Hello America. My name is W. Some of you may know me as your president. It has come to my attention that gays are being allowed to marry in some of our rebel states. If we continue to allow this to happen, we will allow evil to spread through our great nation. We must defy our evil enemies. The current members of the axis of evil are gays, Muslims, and environmentalists. Don\'t let them corrupt you. If asked by a gay to get married, say \"No!\" and immediately go see the The Passion of Christ. Just remember, vote for me this Novemeber - oops, I mean don\'t marry gay.   DISCLAIMER: THE VIEWS EXPRESSED HEREIN ONLY REFLECT THE VIEWS OF OUR CURRENT WHITE HOUSE ADMINISTRATION, NOT THE RESIDENTS OF LAPPLAND.   Here in Lappland, we say allow gays the right to marry. Allow them the legal rights that the rest of us have. This nation has much bigger problems than gay marriage. Maybe we should focus our time and energy on other matters such as the economy and the war in Iraq. Maybe, and I know this is a big maybe, our current administration should stop allowing their religious beliefs interfere with what\'s best for our country.  Hmmm.... That\'s a not really new concept, is it?  After all, we have separation of church and state for a reason. '),(16454,2839449,'23,February,2004','2004-02-23 00:00:00','   Welcome to my blog! '),(16455,2839449,'25,March,2004','2004-03-25 00:00:00','   Over the weekend, my martial arts school had a contest. I didn\'t win, but that\'s okay because there are so many good people in my class. I was just there for the experience. I got in a really good groin shot, and the crowd winced. The sound of my knuckles hitting my uke\'s cup vibrated over the pavement (the actual hit was planned, of course). It was cool. For a little while, I felt like a bad-ass. My friend Mary was also in the contest, and she did really well. She looks sweet, but she\'s really an ass-kicker! I also have a funny story. I have recently lost some weight, so now I\'m able to fit into clothes at the slutty teenage girls\' store (ever heard of Charlotte Russe?). Anyway, I bought a PVC S&M jacket that made me look like Trinity from the Matrix. It\'s really cool. So I\'m thinking to myself, \"I may not wear this in public much, but it sure will provide some good times at home.\" Well, when I got home, I put on the jacket for Matt and he couldn\'t stop laughing. Everytime I moved it crinkled, and with each crinkle I heard another laugh. He said that it looked like I was wearing a garbage bag. So much for sexy, right? Needless to say, I will return the jacket with my head down and my tail between my legs. I guess I\'m getting too old for PVC. '),(16456,2839449,'02,March,2004','2004-03-02 00:00:00','   The Wisdom of William Hung...  \"What everybody needs to understand is that everybody goes through struggles to succeed.\" '),(16457,2839449,'18,June,2004','2004-06-18 00:00:00','   dun dun dun&nbsp; urlLink     urlLink '),(16458,2839449,'18,June,2004','2004-06-18 00:00:00','   MY PARTNER IN CRIME&nbsp; urlLink     urlLink '),(16459,2839449,'18,June,2004','2004-06-18 00:00:00','   &nbsp; urlLink     urlLink    Rowan  Willow&nbsp; urlLink     urlLink    Willow '),(16460,2839449,'18,June,2004','2004-06-18 00:00:00','    '),(16461,2839449,'18,June,2004','2004-06-18 00:00:00','   In my attempt to use my blog more, I\'ve decided to update the template. I hope everyone likes it. Let me know if you have any ideas.  I\'m still a novice. B '),(16462,2839449,'28,July,2004','2004-07-28 00:00:00','   Today, the Mistress is having her carpet cleaned, therefore the children of evil are residing at Casa de Pablo. The stress began last night when we started moving things around in the apartment so that the cleaners would be able to get to more areas.  If you aren\'t familiar with cats, they don\'t like it when you change things. Then, they had to get their nails cut, have Advantage put on, and have their collars put on.  These were not fun times for the Mistress or her children. This morning, I had to catch them and put them in their carriers. At one point, I was on my knees holding a cat by the leg and tail as she squirmed to get under the bed. My partner in crime lunged over my head to grab her. Too funny! The worst part of all is that Rowan peed all over herself while she was inside her cat carrier. My scent for today is \"Essence of Cat Piss.\" The Mistress and her chidren are not amused. '),(16463,2839449,'26,July,2004','2004-07-26 00:00:00','   to go and see The Bourne Supremacy.&nbsp; It\'s a great flick, but beware of the bad cinematography during most of the action scenes. The Mistress loves to watch fight sequences, not get dizzy from them.&nbsp;      The Mistress loves bad boys.&nbsp; Check out this one: '),(16464,2839449,'26,July,2004','2004-07-26 00:00:00','   The Mistress loves the Crystal Method. The Mistress saw the Crystal Method in concert Friday night with her trusty sidekick Mary.&nbsp; The Mistress highly recommends this experience to anyone, no E required.&nbsp; Apparently skeevy boys who wear 6 x 2 inch belt buckles that say \"girl on top\" like the Crystal Method too.&nbsp; And, has the Mistress ever mentioned how many pretty people there are here? It\'s frightening! Surprisingly enough, the Mistress has managed to move to the scariest place on earth.... '),(16465,2839449,'06,August,2004','2004-08-06 00:00:00','   There\'s a reason why you can find the word ass in embarrassing.  Today, I was speaking with my boss about a project from a company called Domainex. I didn\'t miss a beat when I called the company Dominatrix. I turned bright red immediately.  At least she got a good laugh out of it. '),(16466,2839449,'05,August,2004','2004-08-05 00:00:00','   \"Our enemies are innovative and resourceful, and so are we. They never stop thinking about new ways to harm our country and our people, and neither do we.\"  					- George W. Bush, August 5, 2004 '),(16467,2839449,'03,August,2004','2004-08-03 00:00:00','   The mistress is learning how to ride a motorcycle. Look out world... '),(16468,2839449,'03,August,2004','2004-08-03 00:00:00','   and by new, I mean new to me!  Prince - Musicology       I miss the days of arseless pants. Prince without sexuality...what is he thinking? At least he has a name again.  Anyway, it\'s still Prince.  He\'ll always be great.     Slipknot - Vol. 3: The Subliminal Verses        \"Is this CD good?\" you ask. \"Yes.\" says the Mistress.  \"Does this CD rock as hard as Slipknot\'s other CDs?\" you ask. \"NOOOOOOOOOOOOOO!!!!\" says the Mistress.  It\'s like when Kiss took off their masks.  What were these boys thinking? Put your masks back on and kick some arse.     Black Eyed Peas - Elephunk        Makes the Mistress want to shake it like the whole earth\'s quakin\'.     Superjoint Ritual - Lethal Dose of American Hatred and Use Once and Destroy        Phil got tired of singing and decided to start screaming.  This is very different from Pantera and Down, but I like it. It\'s edgy.       Fugazi - 13 Songs        Indie at it\'s best.     Lords of Acid - Farstucker       If you\'re really sad about Prince playing it straight, look no further. It\'s raunchy, and the pictures on the inside of this disk say it all. Lords of Acid = Total Sex Kittens (Me-ow! baby)         The Crystal Method - Legion of Boom        Tweekend\'s still my favorite.  This one\'s a little more trance than their previous CD\'s, but it\'s still good. The best way to hear it is live.     Marilyn Manson - Antichrist Superstar        (This one\'s not new at all really, it\'s just been revisited.) This CD rocks so hard that it exploded in our CD player. Marilyn should have taken Superjoint Ritual\'s name: Use Once and Destroy. Note to self...Marilyn Manson disks have a shelf life. '),(16469,2839449,'02,August,2004','2004-08-02 00:00:00','   Be sure to check out the reallycrazygirl.wmv at the Kiss My Monkey Ass site.  During the middle of the video, you\'ll find that they play \"Every Song is a Death Metal Song\" in Japan as well. Rahhhhhhhhhhhhhhhhhhh!!! '),(16470,2839449,'02,August,2004','2004-08-02 00:00:00','   Yes, the Mistress attended this very scary show.  Who knew there were so many skin-heads in LA. I\'ve never seen so much \"white pride\" in my life.  And everyone thinks it\'s only in the south. Twisted! I\'m a little bummed that Slipknot played on the 2nd stage.  The sound quality was bad, and there was so much dust that you couldn\'t see the stage.  The show on the main stage was cool.  Phil Anselmo is a talker!! I think he talked through half of Superjoint Ritual\'s set. I never thought that I would see Judas Priest and Black Sabbath reunited. Rob Halford\'s costumes reminded me a little bit of Mr. Slave on South Park (\"Jesus Christ\"). Ozzy was talking crazy as usual. Anyway, it was fun. Hopefully, they\'ll be in San Diego next year so that I don\'t have to drive out to the desert. '),(16471,2972526,'30,March,2004','2004-03-30 00:00:00','    If I\'m in the business of planning, why does it seem that things are never planned & I barely get by when \"planning\" with my boss? Why is it that we organize one thing, she gets involved & all plans are changed without our knowledge. It\'s fine to change things, but let us know sooner than the day before! Deb\'s going to this meeting completely blind, and I\'m not even sure what\'s been arranged anymore even though I\'ve been in it from the beginning. It\'s all extremely frustrating... especially when you have more than one event being planned like this.   '),(16472,2972526,'28,March,2004','2004-03-28 00:00:00','    It\'s amazing how many different features humans have in common with other animals. Our powerful spring-like legs, our sense of hearing, our molar & fingernails. It\'s a wonder that we don\'t have more scales or feathers.   '),(16473,2972526,'24,March,2004','2004-03-24 00:00:00','    heh... curtail is a fun word.   '),(16474,2972526,'23,March,2004','2004-03-23 00:00:00','    Yup. I agree with the AAA report that gas prices are really high.  10.559 gallons cost $19.00 this morning. Ouch.   '),(16475,2972526,'22,March,2004','2004-03-22 00:00:00','    Hurray for the Madness! The games are good, but the brackets aren\'t always so kind.  This time of year is always exciting... spring is around the corner (although it\'s 94 in Phoenix today), and the sports are exciting. First college basketball, then college hockey, then pro hockey & basketball... all the while pro baseball is coming to life.   '),(16476,2972526,'15,March,2004','2004-03-15 00:00:00','    Who says the bad guys aren\'t winning?  Spain\'s got a new president, Putin\'s still in (democratically?) & Osama\'s still on the loose.  According to French General Henri Bentegeat,  \"Our men were not very far, on several occasions, I even think that he slipped out of a net that was well closed.\"  ...and Maryland won the ACC tournament?!?   '),(16477,2972526,'13,March,2004','2004-03-13 00:00:00','    urlLink http://www.legacy.com/LegacySubPage2.asp?Page=LifeStory&PersonId=2021215   Rest in peace, Gail.   '),(16478,2972526,'11,March,2004','2004-03-11 00:00:00','    I wonder if chipmunks\' legs or feet go numb scampering around on snow.   '),(16479,2972526,'10,March,2004','2004-03-10 00:00:00','    urlLink http://www.boston.com/business/articles/2004/03/10/house_moves_to_ban_lawsuits_blaming_obesity_crisis_on_the_fast_food_industry/   I just read that the House is considering legislation to prevent people from suing fast food restaurants for obesity. It kind of follows from my previous post.  Good notion by the lawmakers...  Even though un-nutricious food is an option for the public, doesn\'t mean they are forced into purchasing and consuming. If people choose a Big Mac over a carrot stick, that\'s not the food industry\'s fault.   '),(16480,2972526,'09,March,2004','2004-03-09 00:00:00','    urlLink http://www.usatoday.com/money/autos/2004-03-09-eurosuvs_x.htm?POE=MONISVA   Great, we\'re dragging down the standards of other countries now.   ...okay, so I like SUVs - but not the huge ones! Maybe because I \"appreciate the fashion statement SUVs make in the sea of small economy cars and hatchbacks\". Maybe because they haul a lot of stuff. Maybe because my first car was a Plymouth Grand Caravan & a SUV is the next logical step.   '),(16481,2972526,'09,March,2004','2004-03-09 00:00:00','    Iâ€™m not sure what happened. I graduated from college & started gaining weight â€“ a lot of weight. Lots of ideas have been tossed around as to the cause. Maybe I have never eaten properly, but lots of exercise and activity had offset the negative habits. Maybe I became depressed & sought comfort in food.  The result remains the same: Iâ€™m much heavier and less physically fit than Iâ€™ve ever been. For the last 2 years, Iâ€™ve been fighting & relapsing into my old ways. As Dr. Marion Nestle (former chairwoman of the Department of Food Studies & Public Health at NYU) has mentioned, â€œIf people want to be smaller, they have to eat smaller amounts.\"  I read in  USAToday  that according to a new study published in todayâ€™s  JAMA  â€œpoor diet and lack of exercise might end up killing more people than tobacco use and become the leading cause of preventable deaths in the USA by as early as next year.â€  Thereâ€™s a sobering thought.  I just thought that I was enjoying new & flavorful food that I never had been able to purchase on a high school & college studentâ€™s budget. I didnâ€™t realize how bad the situation has become across the nation.  Sounds like people may be sucked into the same wonderful pictures and aromas of new foods that swayed me to become more rotund.   No wonder the fast food companies continue to post profits; more Americans than ever are eating out at fast food restaurants. Why not? With the advent of the $1 menus everywhere, itâ€™s cheaper to eat out than fix something at home or go somewhere more nutritiously conscious â€“ not to mention quicker & more convenient.  A helpful step in the right direction is McDâ€™s discontinuing the super-size menu options. Pepsi is looking to enter a pseudo-diet drink, â€œPepsi Edgeâ€ (70 calories versus the full 150 in the regular version), into the market this summer. FritoLay has decided to take trans fats out of their products. Maybe if companies continue to realize that killing customers is bad for business, people will be forced to eat the more healthful choices that are becoming available.   '),(16482,2972526,'09,March,2004','2004-03-09 00:00:00','    March Madness is about to begin. Some may say it has already started. I was asked recently if I could think of any ways to keep more people involved for longer in tournament bracket pools. The guy at work that sets up the pool was concerned because after the first round is played a substantial number of the people no longer have a shot at winning the pool. He was looking for a way to keep more people involved for longer.  The first thing that came to my mind was the people shouldnâ€™t have picked crappy teams. The second thing that came to my mind was possibly adding a â€˜margin of victoryâ€™ (MOV) pick along with the pick for winning team. This way, even though people may not pick winning teams, maybe their MOV was close to the outcome of some other games. Even if they choose the losing team, if they had a good MOV pick, then they would still get points.  For instance: Winning team pick = 5 points MOV point pick (within 5pts ranges) = 2 points Duke vs. Michigan, you pick Michigan to win, and pick a MOV of 6 points. If Duke wins by 10 points, you get 2 points for being within the MOV 5-10 point range.  This way, people may still be competitive if they pick correct MOVs.  The guy at work agree to consider this idea, but seemed like he was leaning towards having people re-pick after the 1st or 2nd round (in case all their teams were gone). The way I see it, this doesnâ€™t reward the people that did pick the correct teams from the start.   '),(16483,2972526,'07,March,2004','2004-03-07 00:00:00','    I tend to have random thoughts throughout the day that I wish I would remember.  Maybe this will be a tool for myself and others to sift through the clutter of my mind.   '),(16484,2972526,'30,April,2004','2004-04-30 00:00:00','    One marine\'s mother tried to explain the pictures of abuse at the Abu Ghraib Iraqi prison as \"stupid, kid things - pranks.\" Really?? I don\'t remember ever pulling those kind of pranks when I was younger.  Now, I\'m not saying these soldiers are innocent or guilty, but the pictures are pretty damning. I\'m sure there\'s some scapegoating going on, but they can\'t hide from the fact that they were there & took some part (active or passive) in the abuse.   '),(16485,2972526,'19,April,2004','2004-04-19 00:00:00','    Hysterical...   urlLink NY Times Article (registration may be required)   For something as simple as an international conference on energy creation from animal manure & landfill gases (not a politically hot topic) they censored some of the translations. I can only imagine what they did to Cheney. I\'m sure the White House staff knew that there would be some censorship, but I doubt they envisioned the extent to which the Chinese would go out of their way to hide what he said.   '),(16486,2972526,'15,April,2004','2004-04-15 00:00:00','    I think I\'ve figured out what I\'ve been craving the past few days. Cheesecake. So good and so bad.  Thanks to Erin for helping me determine this.   '),(16487,2972526,'13,April,2004','2004-04-13 00:00:00','    Even after 8 hours of decent sleep for the first time in quite a while, I\'m exhausted. I think everything is catching up with me. I\'m basically exhausted. I probably wouldn\'t have come in to work today until noon (ish) except I had scheduled an 65k service for drop-off this morning. I\'d leave early, except I have no car.  I have totally moved beyond the \"so tired I\'m punchy\" stage to the \"mush for brains\" stage. I don\'t believe I\'ve had a complete, coherent thought all day.   I should go running tonight because I haven\'t moved since Thursday (and even then, it was sporadic). But I\'m exhausted. But then again, the game is on Thursday & then we\'ll be on vacation & lazy. I haven\'t shaved. ??? I don\'t know what to do.  So sleepy.   '),(16488,2972526,'07,April,2004','2004-04-07 00:00:00','    Last night it I determined the perfect tattoo for me. Ready for it?  \"Sproingy\"  Good stuff, huh? It would be green letters, possibly outlined in black. There would be tiger stripes on all of the letters, but not too many as to take away from the word.  The green and tiger stripes would symbolize frogs & tigers. Both frogs & tigers have powerful legs that are very similar to springs. Just like my legs :) They may not look like much (just like frog legs look pretty pathetic), but they contain the hidden potential similar to a compressed spring ready to release its power. My arms are like pythons, so I may include some scaley texture to the tattoo, but that would need to be worked out by the artist that draws the design.  Where would I get the tattoo? Well, I don\'t want it to be anywhere that could potentially be stretched or get horribly saggy. I also don\'t want it to be anywhere that would be extremely painful to get a tattoo (like on a bone). I\'ve determined that my right butt cheek would be the ideal setting for Sproingy. I can get my butt to go numb by sitting in a funny position so the needle wouldn\'t hurt, and I\'ve always had a fairly static butt, so it won\'t get too stretched or saggy. If it does get stretched somehow, then the stretch marks would just add to the tiger stripe effect.  I should say that this is a work in progress, but it seems like a solid idea to me.   '),(16489,2972526,'06,April,2004','2004-04-06 00:00:00','    Deb: I won\'t be in tomorrow unless Kate needs me for something. ... I\'m trying to check-in with Kate before I leave now. Have you seen her?  Me: No.  Erin: Is her computer still on? If its on, then she\'s still here. If its off, she\'s left for the day.  Me: She was just here a couple minutes ago & didn\'t mention anything, so I\'m pretty sure she\'s still here.     Kate: I just found out the proposal I thought was due Friday is really due Monday. I\'m going to have to distribute some work to squeeze out this proposal.  Me: (look of horror)  Erin: (sing-song voice) There you go, Deb. Looks like you\'re coming in tomorrow.     Deb: (as she\'s almost sprinting down the hall to leave) Guess I\'ll see everyone tomorrow.   '),(16490,2972526,'04,April,2004','2004-04-04 00:00:00','    So my dad\'s moving to Salt Lake City. I know very little about the area, so it should be neat to visit sometime. Ever since I learned of his move out west, I\'ve found myself wondering \"what\'s a Ute?\" Well, to help all my faithful fans out there, a Ute is an American Indian tribe for which the state of Utah is named. They\'ve been in the Utah/Colorado area for at least 1,000 years & they were among the 1st American Indians to use the horse as transportation.  According to my dad, who has also done some research on the area, the Utes & Mormons really didn\'t get along too well, so the Mormons sought help from the government. To show their gratitude, the Mormons have generally been supportive of the government & its programs. Which is why big chunks of Utah are owned by the government for testing & military bases/research. I must add that this paragraph is oversimplified, but it\'s the gist of what has transpired over the last couple hundred years. -- On a separate note, I must say that Duke should not have lost to UConn. UConn is an excellent team & they won fairly, but Duke allowed them to do so. Poorly timed passes, lax defense at times, and poor shot selection were all factors in Duke\'s loss. They have the knowledge and the skill level to perform at a much higher level, but almost seemed as though they chose not to. I honestly believe they were playing at a level that would have helped them to the championship. I\'m not sure what crumpled (Maybe pressure of being underdogs? Maybe too many young players looking forward? Maybe...) I suppose there\'s always next year.   '),(16491,2972526,'02,April,2004','2004-04-02 00:00:00','    If I run for just a half a mile 3 out of 7 days (but really push myself on the pace), and then go & play indoor soccer, I can play almost as aggressively as I\'d like. I still get winded & tired, but it\'s a great improvement over what I would deem to be my most horrific showing ever. Hopefully, if I keep up the run & expand its distance, I\'ll be able to go for longer indoor without getting winded. -------- In other news, despite his party membership, I\'ve never really cared for Benson. While he was running, I had an opportunity to meet him, albeit pretty briefly. He didn\'t strike me as a good politician then, and he certainly doesn\'t now. The rumor mill amongst the different campaigns at the time held that he was pretty much a jerk with the money to get where he was aiming. Now that he\'s in the top spot, I can\'t see that he\'s done much good or bad. However, what was he thinking on importing Canadian drugs? Political suicide. I\'m not sure that I mind what\'s following, but you would hope a governor would have decent advisors & really think something through without going ahead with what seems to be a cutesy exercise in civic disobedience. Nevermind that it\'s a federal crime (I believe). Perhaps he has terrible advisors because the good ones turned down the chance to work for a jerk after the campaigning ended.  He\'s losing the only support he had, for what gain? Even if the labs show identical drugs, that still doesn\'t mean it legal to import them. Without the rank & file support he used to have, there\'s no chance of passing a law in either NH house that would make the importation legal. I\'m not sure he really has any good option for a next move.   '),(16492,2972526,'27,May,2004','2004-05-27 00:00:00','    urlLink High Gas Price??   Ha!   I apologize. I\'ll be more careful about mentioning the cost of gas, or else I\'ll doom the world to prices that are significantly higher.   '),(16493,2972526,'26,May,2004','2004-05-26 00:00:00','    On the way home last night I was listening to Marketplace, All Things Considered or some such program on NPR. A commentator (Christie Mellor) came on & gave told a little story about going to a friendâ€™s party were a lot of alcohol was consumed by all. She eventually went to the bathroom, only to find the toilet seat locked â€“ much like every other hinged object in the house. She went on to describe the lengths her friends had gone in order to make their house childproof.   After describing the unattractive table bumpers and locked rubber band drawers, she put her finger on what had happened to the previously normal couple. Fear. They were bullied by safety product manufacturers into believing that serious harm and/or death would befall their child if they didnâ€™t spend $39.99 on rubber bumpers for their coffee table. The commentator went on to say that the money (and appearance of the furnishings) could have been saved by simply telling their child â€œno.â€ If a child gets a bump on the head after disobeying & running in the house, theyâ€™ll develop some character, think twice & understand the consequences of disobeying. She concluded that saying â€œnoâ€ more often to the child would also help prevent embarrassed guests from having a potential real accident in the bathroom with a locked toilet seat.  I understand some childproofing is necessary, but there should be limits on the protection, too. Iâ€™m curious to see the how far my previously normal friends will go as they start baby-proofing their house.   '),(16494,2972526,'26,May,2004','2004-05-26 00:00:00','    Every once in a while, I find myself thinking that itâ€™d be nice to lead a more exciting life. Iâ€™m not sure what I deem to be more exciting. Maybe exciting means I want to get more life goals accomplished (or even starting in the general direction of accomplishing them), travel to more exotic locals (read: not New England or the Midwest), go to more sporting or theater events, have a social life or meet more people from across the US or world.  The problem is that I lead a fairly fast paced life as it is. I travel pretty frequently both for work and for familial affairs. I see movies on a pretty regular basis. When thereâ€™s time, I watch sporting events (on TV). Iâ€™d love to fit time in to volunteer somewhere, to read interesting books, or to go to school. Iâ€™m not sure where Iâ€™d fit it inâ€¦ we donâ€™t eat dinner until 9ish as it is.   Honestly, my life is pretty cool. I do travel some neat places & see some neat things (I just wish I could bring my husband, too). Weâ€™re further ahead of the â€œaverageâ€ person in terms of financial independence. I work for a company that does is interesting governmental work. I have a wonderful husband and a crazy (but interesting) family. My car runs reliably (and has a turbo, CD changer, sunroof & heated seats). I do have good friends that Iâ€™ve known for over a decade.  Maybe Iâ€™m happiest when Iâ€™m complaining about my life: not enough time, not as much money as Iâ€™d like, a job with not much satisfaction or intellectual stimulation, only human interaction is with co-workers & husband, out of shape, lack of life direction. Maybe I just enjoy living in a state of pseudo-depression?   '),(16495,2972526,'20,May,2004','2004-05-20 00:00:00','    Me:  Allowing 5 minutes per person, we are now full for both the 1-2PM slot and the 4:30-5PM slot.   US Gov\'t employee:  How many people have signed up so far?  Anyone with basic math skills can figure out that 12 people (at 5 mins each) would fill the 1-2pm slot & 6 people would fill the 4:30-5pm slot.  Heh... and they run the country?   '),(16496,2972526,'12,May,2004','2004-05-12 00:00:00','    I started thinking about blogs today. I wondered why people start them & why people read them. I know my reasons for starting one and I\'m its similar to some others\'. I went a step further & began thinking about why I read other peoples blogs.  At first, I was ready to chalk it up to some sort of voyeuristic curiosity of how other people live. It\'s always interesting to learn new opinions, perspectives, storylines, and people. It\'s nice to see the ups & downs of daily life that happen in someone else. It helps remind me that the grass is greener on my side sometimes, too.  I also believe that I read blogs to gather information. I\'m always poking around on the internet, trying to find more info on different topics. Much of the time, the topics I research have little relation to each other (ignoring the fact that they all started as a fleeting thought through my head). Blogs are an excellent source of my fleeting whim topics. I can read opinion blogs and then try to find an opposing opinion. Or, I can read information/news related blogs and then try to find more information on the current events. I crave information & really love the search for solid, credible info.  In a somewhat related vein, the DNC is issuing press credentials to some bloggers. My guess is because they distribute information (for better or worse) to more people than other news organizations. People in Wildrose, North Dakota may not receive many news carriers, but if they have a modem, they can reach a broad range of opinions & facts.   '),(16497,2972526,'10,May,2004','2004-05-10 00:00:00','    Hurray!  Only 41 years, 4 months until I can retire with full social security benefits.   '),(16498,2972526,'07,May,2004','2004-05-07 00:00:00','    I love my husband!  Happy Wives\' Day   '),(16499,2972526,'04,May,2004','2004-05-04 00:00:00','    Mass. Governor Mitt Romney wants to bring back the death penalty to Massachusetts. He says that basing the sentences off of science (mainly DNA evidence) will corroborate guilt or innocence. He claims it would be foolproof.  The only foolproof plan, in my view, is no death penalty. Then there\'s no chance of an \"oops.\"   '),(16500,2972526,'29,June,2004','2004-06-29 00:00:00','    A comment made by a co-worker yesterday has been bothering me.  She very recently signed up for Eharmony.com and received a few matches with her personality profile. One of the matches asked her a few questions (implying heâ€™s interested). By answering the questions, she was then able to ask him some questions.  Since this all happened yesterday, she was pretty excited, but I can tell sheâ€™s not necessarily expecting the sky to fall. Today, sheâ€™s looking around match.com. I wonder if sheâ€™s going about this with the wrong attitude. One of the questions she wanted to ask was if heâ€™s comfortable being on his own. It seems she doesnâ€™t want a full-time relationship. If he comes over for dinner or something, sheâ€™d like him to leave eventually. It seems to me that if you donâ€™t want to spend time with someone, maybe you arenâ€™t ready for a relationship?  Between that comment & a few others, it just doesnâ€™t seem like sheâ€™s going into this whole match-finder thing with a very open mind.   '),(16501,2972526,'23,June,2004','2004-06-23 00:00:00','    The airshow in RI was great! The only way I knew about it was through a co-worker. She found out from a flier she had picked up at Ace Hardware. I\'m not sure what Ace has to do with the planes, but I guess it doesn\'t matter.  The weather on Sunday was beautiful... not too hot with a slight breeze. The planes were pretty amazing. The museum-type planes were sort of ho-hum, especially after a recent visit to the  urlLink Museum of Flight  near Seattle. However, the aerobatic planes did incredible maneuvers throughout the day. The USAF Thunderbirds were a great way to end the day. It\'s almost frightening how precise they are. If my life depended on that type of precision all of the time, I\'d be in trouble.  Monday, I left for NYC for business. I was to arrive around 1pm, check into the hotel & meet with our contact and my boss around 2:30. In the taxi, on the way to the hotel, my boss calls and says \"it\'s such a nice day, why don\'t we change the meeting until 3?\" I okay it - it\'s only another 1/2 hour.  Around 2:15, my boss calls again and wanted to move the meeting until 4. It was a nice day, but I wasn\'t able to go too far or accomplish anything because I needed to hang around to do actual work. (You know, what I was there to do.) In the end, we did meet around 4, but our contact had gone home for the day (as she usually does) around 3. I warned my boss about this, but I\'m guessing she thought I was joking. What the whole day boiled down to, was I only managed 2 hours of directly billable work - and I had anticipated more.  Later, we met with another co-worker & went wandering around Greenwich Village. My boss thought she knew where she was going, so it took us 1 subway stop too far & 30 mins of walking before she found the general vicinity she had thought of. It was a very nice area & I did see an entirely new area NYC.  Tuesday\'s meeting went well and I was able to catch an earlier shuttle back to Boston. Woot! That meant I would land when it was still a little light out, so I could read the construction signs to determine a way out of Logan Airport. All in all, a good trip, but very tiring. Not only due to the long hours and business attire (although I only buy what\'s comfortable), but also because it\'s draining keeping up & placating my twitchy boss.  I\'m really glad that summer-like weather seems to be in the Boston area for good. :)  side note: the spell check wanted to replace aerobatic with arbutus and billable with bilabial. ???   '),(16502,2972526,'19,June,2004','2004-06-19 00:00:00','    I\'m supposed to go see the  urlLink Thunderbirds  today.  I\'ve really been looking forward to this for a little while now. Even with my love of things that fly (especially airplanes), I\'ve never been to an airshow. This one looks pretty cool. The down side of it is the location. It\'s in Rhode Island & it\'ll take me about 2.5 hours to get there (not assuming traffic) - and I want to be there for the opening parachute jump at 9:55.  I was planning on getting to sleep at a decent time... but I was waiting for Paul\'s show to end (it wound up being a 2 hour show); meanwhile, he was waiting for me to get off the computer.  Grrrr!  We didn\'t get to bed until way late.  I don\'t know if I was too excited to sleep, or just too uncomfortable to sleep (my leg is pretty banged up from this week\'s pick-up soccer game, so I can\'t really find a position that doesn\'t hurt). Now that it\'s 1.5 hours of sleep later and I\'m wide awake, I know that later I\'ll cranky. Paul\'s the same way. He didn\'t get much sleep, so he\'ll be cranky. The cranky twins have shown their ugly heads before & we never enjoy ourselves when it happens. Since it\'s my first airshow, I want to see & do everything.   On the plus side, there\'s a chance of showers throughout today down in RI (in fact, it\'s heavily raining right now). Tomorrow looks like it\'ll be 5 or 6 degrees cooler, but sunny all day. I guess I\'ll wait another day.   '),(16503,2972526,'18,June,2004','2004-06-18 00:00:00','    I\'ve been doing a lot of thinking about short term & long term goals lately. It\'s starting to bother me that I haven\'t even started down the paths to achieve some of them. I keep thinking about all the interviews I had that asked where I saw myself in 5 years. Well... 5 years has now become next year.  Some of the people I went to high school with have achieved & accomplished what I hold to be pretty amazing things. Things I wouldn\'t mind trying. Learning new language s , advanced degrees, neat/unusual travel and jobs, etc. I\'ve done some pretty cool stuff, but I\'m not sure it\'s the kind of thing they write about in my smallish town paper. I\'ve accomplished more than some, but less than others. That should be good, but I strive for more.  My biggest problem is wanting to do, see & understand everything. However, I never make any concrete decisions in terms of direction, so I wind up doing absolutely nothing. In some ways, it\'s procrastinating. In other ways, it\'s a severe case of indecisiveness.  I\'ve got a lot to do before next year\'s 10 high school reunion. I suppose I should get a direction. Or skip the 10yr. reunion. Maybe I\'ll do something that\'ll make me proud by my 15 or 20 yr. reunion.   '),(16504,2972526,'16,June,2004','2004-06-16 00:00:00','    I have been to the following states (at least the ones I have a good recollection of & didn\'t just pass through):    My husband & I have been to:     urlLink create your own personalized map of the USA   or  urlLink write about it on the open travel guide   We have some work to do...   '),(16505,2972526,'14,June,2004','2004-06-14 00:00:00','    What a good weekend. I managed to accomplish almost everything on my list. After a very slow start on Saturday, I cleaned our porch, washed my car, went to storage to get the boxes for the Christmas tree, took down the Christmas tree, returned the full boxes to storage, vacuumed, watched the Red Sox game, purchased some flowers, planted the flowers & cleaned up afterwards.  After Paul returned and showered (he was helping friends move apartments all afternoon), we went to a bar that was having live music. We managed to last all the way from 8:15 until 1am (closing). It was a good time, although afterwards I have to admit weâ€™re getting old.  The music was loud & the smoke bothered me at times. The third band was nothing but noise. Iâ€™m not sure if there was music involved because my ears were ringing the entire time. The band took away from the fourth band, which was pretty good.  At any rate, Sunday was just as nice weather-wise & we took turns walking once we finally woke up after the late night. (We also stayed up late on Friday night watching Reaganâ€™s funeral & related coverage.) We then went out & purchased about $100 worth of clothes & discovered we left a credit card at Friday nightâ€™s restaurant.  All in all, a good weekend. Iâ€™m hoping for more sunny days & less rain (even though it rained again today) as summer is set to officially begin soon. I really hope next weekend is going to be nice â€“ we have a trip to Rhode Island for an air show in the works.   '),(16506,2972526,'09,June,2004','2004-06-09 00:00:00','    I follow 7 blogs (8 if you count my husband\'s). Everyone seems to have creative or insightful things to say on a decently regular basis.  I\'m wondering if the only way I can display creativity or insightfulness is when my brain is already working overtime. Lately, my job has been exceptionally slow. I\'ve pretty much given up on trying to find things to do. As a result of inactivity, I don\'t seem to have much to say.  Sorry.   '),(16507,2972526,'04,June,2004','2004-06-04 00:00:00','    Harry Potter starts today in theaters. How did it sneak up on me with all of the hype that surrounds the whole marketing empire? I was even in NYC the day of its US debut! I could\'ve sworn I had another month.  I usually read the book before I watch the movie - I guess it\'s the appropriate college-graduate thing to do. However, I haven\'t even purchased the book and it certainly isn\'t going to be available in my library for the next 7 months, either. I\'m not sure it\'ll still be in the theaters by the time I\'m done reading all 560 pages (yes, I know that it\'s a quick read... but come on, I can\'t finish an interesting 80 page magazine in less than a week).  Do I go ahead & watch it without reading the book? Do I delude myself into thinking I\'ll read the book after I see the movie? Do I save myself $14 & not by the hardcover? (It would have to be hardcover to match the previous 2 volumes I\'ve purchased.)  Then again, outside of Harry Potter books, I can\'t remember the last book-turned-movie I\'ve read. Maybe something like Patriot Games - and I read it after I saw the movie several times. Then again, I\'m not sure if I ever finished reading that.   '),(16508,2972526,'03,June,2004','2004-06-03 00:00:00','    Well, I managed to haul my ass around for 3.5 miles at the  urlLink JPMorgan Chase Corporate Challenge . I was 2nd place within the women from my company. However, I really think I could\'ve been 1st if I hadn\'t stopped & walked briefly a couple of times. I never really got too winded, but the humidity caused it to be very sticky early on in the race. At the same point in time, I have a long way to go to get into shape, so I could\'ve done worse, too. Last year\'s goal was to run this year. I think next year\'s goal is to run the whole time in 35 minutes, or less.   '),(16509,2972526,'01,June,2004','2004-06-01 00:00:00','    A new look for the page. I think I like it, but maybe it\'s just a way to waste time at work... I keep changing settings.  New addition - the ability to leave comments! Woot! I\'m excited already at the idea that all the millions of folks (read: my husband) that read my page will be able to leave a little note of appreciation, dissatisfaction, tidbits of inspiration or anecdotes, etc.   '),(16510,2972526,'20,July,2004','2004-07-20 00:00:00','    So much for the random thought blogging, eh?   July has been a pretty busy month so far, and August, September, and October look to be as busy, too. Actually, the weekdays are pretty standard from week to week:   Monday: Guys poker night. I either exercise or veg in front of the TV/computer  Tuesday: Exercise  Wednesday: Company softball game for me (hubbyâ€™s version of my&nbsp;Mondays)  Thursday: Indoor soccer game  Friday: Eat out after work (aka non-fast food)   You would think that these things wouldnâ€™t fill an entire night. However, it takes me about an hour to get from work to home. After completing the activity of the night &amp; driving home, itâ€™s typically 8:30 or 9 when I finally get around to walking in the door of the apartment &amp; thinking about dinner. Dinner then winds up being fast food or a couple mini York peppermint patty (and yet weâ€™re both able to keep off the weight weâ€™ve lost).   The weekends have been filled lately, too. Itâ€™s amazing our laundry is sort of up-to-date.   The weekend after the 4th of July we drove down to Queens for my friendâ€™s wedding reception. Since she got married during our trip out west in April, it was a reception primarily for his family &amp; family friends. We arrived on Saturday afternoon, checked in to the hotel &amp; hopped on the train into Manhattan to meet up with them. We wandered mostly around the Greenwich Village/Little Italy area. After wandering, drinks &amp; dinner we took the train back to the hotel. Long day, but a lot of fun. The reception in  urlLink Bayside  was really nice, although we didnâ€™t know anyone but the bride &amp; groom. They were pretty grateful to see familiar faces (especially the bride). The groomâ€™s friends made us feel welcome &amp; included us in pictures &amp; discussion. The drive back was pretty uneventful. I think traffic was light because everyone had traveled the previous long weekend.   This past weekend was nothing but down time. We thought weâ€™d clean and straighten out stuff with the apartment, and shopping. In the end, not much was accomplished. It was really nice. We managed to clean a lot of things off of the TiVo (although we still havenâ€™t watch the last 2 Friends episodes). We began the weekend watching Rally Magazine and ended the weekend with the Rally Argentina. Iâ€™ll have more to post about the rally &amp;  urlLink Subaru  later.   This coming weekend Iâ€™ll be off to NYC (again). I like my trips there, but theyâ€™re never long enough. Not nearly enough \"play\" time since itâ€™s for work. It should be interesting flying out of Boston with the  urlLink Democratic National Convention  officially starting on Monday. A lot of people&nbsp;are less than excited about it coming.   '),(16511,2972526,'07,July,2004','2004-07-07 00:00:00','    Alright, so lately I\'ve been unreliable with blogging. I figure the blog was originally supposed to be an outlet for some of my random thoughts, so I should go back to the basic randomness.  The fireworks  urlLink on the Esplanade  were great. Again. David Lee Roth was okay, but the original 1st choice of Aerosmith would\'ve been better. I\'ve probably more than my fair share of fireworks & the Boston 4th is always impressive. I think the hubby & I are making this into some sort of tradition. Each year, we perfect the whole process a little more (for example, we didn\'t sit directly next to the howitzers this year). If you ever get the chance to visit Boston around the 4th of July, I  highly  recommend finding a good viewing spot in Cambridge or Boston along the Charles.  I\'ve allowed a few more blogs into my daily reading circle. Some of them seem to do a lot of quizzes. One caught my eye, so I tried it out. It seems pretty accurate to me (and I had some pretty non-nerd responses scattered throughout):    You\'re a Speak & Spell!!  You nerd, you.  Just because you were disguised as a toy doesn\'t mean you weren\'t educational, you sneaky bastard.    urlLink   What childhood toy from the 80s are you?      brought to you by  urlLink Quizilla   '),(16512,2972526,'10,August,2004','2004-08-10 00:00:00','    Paul & I are still getting over the \"vacation\" back to Michigan for his sister\'s wedding. From the moment we touched down Thursday afternoon, until we took off again on Sunday, we were scheduled to be somewhere or do something the entire time. There was even a bathroom schedule that we had to abide by. As I wasn\'t part of the bridal party, so my \"to do\" list was actually pretty lengthy. When everyone else was off doing pictures, or ushering guests, etc, I was running around setting up hors d\'ouerves, shuttling toasting glasses, etc. Lots of little stuff. Not much down time.  Friday, I did get to drive down to visit with my mom &amp; grandma for a bit. It was really good to see both of them and wander the farm briefly. I\'m still struggling with the idea that it very likely was the last time that I would see my grandma alive. Although she was doing really well when we saw her, she had last rites (a Catholic sacrament) the day before we were there. I guess she was planning on dying Saturday. She didn\'t, but even still... Even with 3 years to prepare for this, I don\'t think I\'m very close to being ready for her to die.  Back to the wedding... It was nice to see the same people that were at our wedding, but without the hassle of being the bride & groom. Congrats to Kelly &amp; Andy. They both seemed very excited and happy to begin their married life in the Bahamas.  I\'ve always been curious about the Bahamas because it\'s tropical, but doesn\'t take a lot of time to travel to. Apparently the rainy season in the Bahamas is from May-October. After checking out weather.com, it seems the newlyweds have picked a bad week to vacation. It\'s warm, but there is a good chance of thunderstorms & rain every day this week. I guess being in the middle of both the hurricane &amp; rainy season means I shouldn\'t think of visiting the Bahamas during this time of year.  I keep thinking that I could use a short vacation to catch up on sleep (or rather, start sleeping through the night again) and just relax. Something warm, involving a pool or beach would be nice. It\'s not going to happen, though. Things are starting to pick up a bit at work, and our weekends are either rainy, or booked with other obligations.  Now we are back in New England & literally hitting the ground running. We\'ve signed up for a 5K race as part of HP\'s team. Since Paul did a race as a part of my company\'s team, I can do one as part of his company\'s team. The race is Thursday, and we ate a lot of bad foods &amp; didn\'t exercise over the extended weekend, so we are trying to prepare ourselves in the limited time before the race.   '),(16513,2972526,'04,August,2004','2004-08-04 00:00:00','    Maybe it\'s just me being overly sensitive, but I don\'t think it\'s very professional to send an email to a client saying:  \"Soooo, ummmmm, \"Show me da money!\" and later in the same paragraph... \"Where the hell\'s my money?\"  Even though we (by we I mean the sender of the email) have a good relationship with this particular client, it still seems like it\'s stepping over some sort of client-contractor professional courtesy line.   '),(16514,2972526,'02,August,2004','2004-08-02 00:00:00','    It\'s amazing how my time gets sucked into a big black hole. I think nothing much is going on & before I know what has happened, it\'s all gone &amp; it\'s a week later.  I didn\'t feel too well last week. Some weird feeling. People kept asking if I was pregnant. What kind of question is that?? If I was, would I tell people this soon? No. However, I am not pregnant, so I can fight that question off. My guess is I felt weird because I haven\'t been able to sleep very well for about a month now. No real reason why I can\'t sleep much. Just a frustrating situation.  This weekend we went to  urlLink Cirque du Soleil\'s Varekai  performance in Boston. It\'s a very good show. It\'s amazing what the human body can be told to do. I had seen the show on TV & followed the performers from the start to the launch of the tour on the Bravo series, so I was wary about paying so much for a show I\'d pretty much seen. I had nothing to fear. While I was familiar with the acts &amp; the performers names, there is a lot more that goes on during the act besides the main performance. The whole show was good & our seats were awesome. I\'m glad we didn\'t pay for the $75 tickets, since our view was just as excellent. It may be pricey, but it\'s definitely not typical theater, either.  The rest of the weekend (Saturday morning and all Sunday) fell into the black hole. We accomplished some little things, but not enough to account for the amount of time spent.  Now we\'re gearing up for Paul\'s sister\'s wedding. We leave Thursday, but literally every minute is accounted for from the time we land Thursday until we fly out Sunday. I hate using my vacation days on things like this... there\'s absolutely no time to relax &amp; we wind up exhausted going into another work week. It\'ll be nice to see people that I haven\'t seen in a while, but I\'ll only have 30 mins available to spend with them before my next appointment. Look out SE Michigan; Paul &amp; I will be tearin\' it up soon.   '),(16515,3558648,'17,June,2004','2004-06-17 00:00:00','    I work in IT support.  Sometimes users are just plain dumb.  We got this message this morning:    The information below has been verified by the police as happening in NOTTINGHAMSHIRE at the moment.     POLICE WARNING - One of the officers who works with the Dare programme has passed along the following warning and asked that it be shared with all drivers.  THIS IS AN EXTREMELY SERIOUS MATTER  If you are driving after dark and you see a car without its headlights on DO NOT FLASH YOUR LIGHTS! DO NOT BLOW YOUR HORN OR MAKE ANY SIGNALS TO THE DRIVER OF THE OTHER CAR.  There is a new common gang initiation \'game\' going on the streets. The new member being initiated, drives along without his headlights on until someone notices and flashes their headlights or makes some sort of other action to signal him. The gang member is now required to CHASE THE CAR AND SHOOT AT OR INTO THE CAR IN ORDER TO COMPLETE HIS INITIATION REQUIREMENTS.  Make sure you share this information with your family, friends and anyone else you can reach. If you have any questions or information, please call your local police department.  PLEASE TAKE THIS SERIOUSLY. THIS IS NOT A JOKE. PLEASE PASS THIS ON TO EVERYONE YOU KNOW ON E-MAIL AND IN PERSON. IT COULD SAVE SOMEONE\'S LIFE!   Now anyone with half a brain will immediately spot that this is a hoax.  In fact it\'s been going around by e-mail since 1993.  It\'s even been in the movie \'Urban Legend\' (the clue\'s in the name).  Some user I\'ve never heard off took it on himself to send it to EVERYONE in the address book.  Duh.  This wouldn\'t be so bad if, later this afternoon, yet ANOTHER copy came from another user telling us the self same thing.  Apparently it\'s OK as the information had been verified by police in Nottinghamshire.  WHAT..SO YOU RANG THEM TO CHECK DID YOU?  NO?  DIDN\'T THINK SO!!  DID YOU CHECK ON THE NET TO SEE IF IT WAS FOR REAL?  NO?  SO WHY THE *@#$ DID YOU MAIL EVERYONE REGARDLESS!!!!  muppets  Anyway, I\'ve decided to start my own chain letter as retaliation.  \"Goverment scientists have discovered that a chemical produced by the human body has been found to cause diseases such as Alzheimers, Parkinsons and cancer.  However, the chemical is only found in people with GINGER hair.  Consequently an emergency edict has been issued that makes it compulsory for all people with GINGER hair to report to special decontamination camps.  This has been verified by the second cousin of the brother-in-law of the woman who does the ironing for the Secretary of State for Health.    DO NOT IGNORE THIS.   YOUR LIFE COULD BE IN DANGER.   SEND IT TO EVERYONE IN YOU ADDRESS BOOK.   NOW!   WHAT ARE YOU WAITING FOR!!\"   '),(16516,3558648,'13,June,2004','2004-06-13 00:00:00','    Actually it\'s not chilly at all, it\'s just that I went to see \'The Day After Tomorrow\' at the flicks last night.  Now I know that some critics have been giving the film a hard time, mostly on the basis that the science is somewhat unrealistic with regard to timescales.  What critics seem to forget is that WE DON\'T CARE...IT\'S JUST A FILM!!!!  So the cataclysmic events that unfolded were sped up a wee bit (OK, probably more than a wee bit), but it wouldn\'t have been much of a disaster movie otherwise would it?  Take the \'realistic\' scenario:  Heroic scientist goes before a global warming commission with predictions of dire consequences.  Cut to end of film, 60 years later with same herioc scientist, now around 100 years old, waving his zimmer frame at the equally geriatric commission shouting \"I bloody told you so!\"  Hardly gripping stuff.  Same thing applies to the big catastrophe scenes - I wouldn\'t have been the same if all you go was a bunch of folks at the beach going \"Y\'know..I reckon this water is a couple of inches deeper than it was 20 years ago\".  Still, the film is quite critical of the US stance on global warming (good on \'em), so it was quite something for a big studio to bring it out.  It was probably given the green light on the basis that the Canadians get it first and worse ;-)  So that\'s most of the big blockbusters seen for the year so far - Harry Potter 3, Day After Tomorrow etc.  I\'m just waiting for Sky Captain and the World of Tomorrow, plus possibly Hellboy.  Then it\'s Alien vs Predator in the autumn!   '),(16517,3558648,'13,June,2004','2004-06-13 00:00:00','    In case anyone is interested, this is the type of bike I have.  It\'s a Suzuki GSF400V - the V means it has all that Variable Valve Timing Stuff.  Bottom line..pulls like a train from 8000 rpm.   '),(16518,3558648,'09,June,2004','2004-06-09 00:00:00','    It makes you wonder, there\'s not really a lot to do on a Sunday apart from watch sport on TV.  And today was a full one.  There was bike racing (Superbikes AND MotoGP), Car Racing (LeMans 24hr and Canadian GP) and footie.  The GP was as predictable as it was dull.  Dear old Schumi cleared off and \"von der race in ein most efficient manner, ja?\".  Started from 6th, finished 1st - yet overtook absolutely no-one in the race.  BOORRRRIIINNNGGG.  The entire thing is being dictated in the pits.  They ought to go back and make everyone run the whole race on a single set of tyres and fuel load.  That\'d slow the cars down too and actually make people try and overtake each other.  That and the fact that Bernie Ecclestone would rather have a season full of tracks that pay him loads of dosh rather than tracks that have good races (see under \"Belgian Grand Prix\" for details).  My advice?  Watch motorcycle racing instead.  Or maybe lawn green bowling.  Disappointed by the footy - 1-0 up with 90 minutes gone and then lose 2-1 in injury time.  Bugger.  Flippin\' French.  Apart from televised sport it\'s been the usual type of day - ironing stuff for work (slightly less boring that the GP), odd jobs (like oiling the fuel filler cap on my Suzuki because I couldn\'t get it open to put more petrol in..always a bit of a problem that) and trying to persuade the cat that the fat little grey bowling ball shape is not quite as \'de riguer\' amongst other felines as she thinks and, no, she doesn\'t need ANOTHER packet of cat food.   '),(16519,3558648,'09,June,2004','2004-06-09 00:00:00','    Jeesos it\'s been hot out.  Over 30 degrees.  I\'m going off to hide in our lovely air-conditioned data centre in a minute.  I think Russ Abbott had the name for it - Betty Swollocks.  It was too darn warm at badminton the other night too - sweating like a pig within minutes.  So dehydrated by the end of 2 hours that I could barely move my limbs.   Funniest news story of the day - Metallica\'s drummer (Lars) has been taken ill just days before the band headline the Download festival here in the UK.  Rather than pull out and disappoint their fans the band have \'borrowed\' a drummer from one of the other bands on the bill - Dave from Slayer..  That has left Slayer short one drummer, so they\'ve borrowed Joey from Slipknot and Sal from Life of Agony to fill in on their set.  Presumably these 2 guys will do their own band\'s set too.  Talk about musical drumstools!   Did I mention is was hot out?  Yuk sticky.  I\'ve got a new toy - Creative Zen Xtra MP3 jukebox with 60 (that\'s SIXTY) gigs of storage (take note Apple with your meagre 40 gig iPods).  Got it from a guy on eBay.  Not that I have an eBay habit.  Really.  Honest.   '),(16520,3558648,'07,June,2004','2004-06-07 00:00:00','    What a morning.  Monday morning at work, never at my best.  I\'m usually fully awake somewhere around Tuesday lunchtime.  Anyway, puncture duly fixed and bike working again so I stagger into work at the crack of 9.30.  Spent a chunk of the morning trying to fix an error where our users can\'t connect to one of our partner locations.  Looks like asymmetric routing loop.  Look it up.  The rest of the day is meetings.  Lots of them, all short and no free grub.  Bum.  Get to prove how unfit I am tonight by attempting to play badminton.  Should get done by 8 when it\'s the series finale of the latest Star Trek series - yay!  As long of course as those d*cks at NTL don\'t screw up the TV again.  I was watching the box last night when my entire cable TV service died at about 8.45, just before the last Hell\'s Kitchen.  So I had to run around and reconnect my terrestial TV aerial so I could get a picture on ITV1.  It wasn\'t fixed when I went to sleep at 1am but it was OK this morning.  Here\'s hoping.  What ticked me off the most is that there is NO FAULT HANDLING WHATSOEVER on a Sunday from NTL.  You pay fora 24/7 service and you get 12/6 support.  Useless.  Rant over.  SYL   '),(16521,3558648,'06,June,2004','2004-06-06 00:00:00','    so, now I\'ve got a blog.&nbsp; Now what to do with  it?      &nbsp;    Well it\'s a Sunday.&nbsp; And I\'ve been at  work.&nbsp; Flippin\' network switches that don\'t wanna play ball.&nbsp; It\'s all  overtime.&nbsp; Ho-hum.      &nbsp;    At least my dear little Suzuki hasn\'t let me down  today.&nbsp; Got a new battery during the week and it\'s starting OK now.&nbsp;  It\'s such nice weather that I took a detour back from work - about a 20 mile  one.&nbsp; Went for a quick thrash around some of the country roads around  Lincoln - you just gotta watch out for tractors.      &nbsp;    The rest of my day is gonna be fun - I need to fix  a puncture on my pushbike so I can get to work in the morning.&nbsp; Then I need  to iron my work gear (I just looooove ironing...not) and then do the  dishes.&nbsp; At least I can listen to some CDs while I wash pots - a little  Slipknot to make the time fly by methinks.   '),(16522,3558648,'06,June,2004','2004-06-06 00:00:00','    Well, I finally got my own domain so I thought I might as well get a blog too.   '),(16523,3344485,'18,May,2004','2004-05-18 00:00:00','   Wake up, french kiss her, good morning... Talk about all the things that are still longer to believe. Talk about the love, the truth, and how much she meant to me. The truth is, she\'s all that I need... All night I would sleep on the bed just to be closer to her...  Everyone is talking, talking about us, the love, the truth between us. I didn\'t care much, just need to be closer to her, so as to take a closer look at her...   Although the night is not long, I would be thinking about her. When I closed my eyes, I won\'t be alone because I\'ll be thinking about her. I\'ve got nothing to prove to her about the truth, just as to tell her, how much I love her.  She didn\'t need to hear, when I said these words, \"I can\'t sing a love song to you, I can\'t be romantic like the way it\'s meant to be. Guess I\'m not that good anymore, just the way it meant to be.\"   If she told me to lie for her, to cry for her, I could. If she told me to die for her, I would. Just take a look at my face, I would pay anything... just to say these words to her, \"I\'ll always love you... and I\'ll always be there for you... Love You, always\"   Love, Alan  '),(16524,3344485,'16,May,2004','2004-05-16 00:00:00','   Darkness falls... everything goes so quiet. I could hear the flame burning on my cigarette as I take a puff. Trying hard to capture the silent that I\'ve long desire... The silent in my heart, the silent in my mind...  Silent became my precious, once... Only then, will I be able to hear voices, voices that I had long missed. Voices that only I could hear. The voices of my own prayers...  \"beg of you, hear my prayers, let me go, help me... please help?\".   Loneliness starts taking over my life & my prayers has become an obsession which I prayed every night. Don\'t know where to go, don\'t know what to do, mainly staring blankly at the wall... \"What have i done to deserve this? Why me? Why do I have to face this alone? Why is everyone ignoring me?\" questions came to my mind soon after.   Blank mind, it\'s so quiet.. so comfortable... so dark... so peaceful... reaching out but there is nothing out there... blank!  Walking alone to the kitchen room, turn my head, and there he is again... this time, he\'s different, he\'s staring at me with no smile, no laugh in his face. I walked up and looked closely into his eyes... cold is the word to use, so bloody cold... I starts to shiver for that moment from his look in his eyes... He started to reach out for me, so am i... we could barely touch our palm from the opposite side of the world. So real, its so real, I could barely touched him.  \"help me, help me alan... alan, help me...\" voices that i hear when i lay down on the bed staring at the ceiling wall. These voices that I hear, their so familiar... heard it somewhere, but I could not remember whose it belongs to... So familiar... so...  '),(16525,3344485,'16,May,2004','2004-05-16 00:00:00','   \"Why am I doing this to myself?\" i asked myself in the mirror this morning when i woke up. There he is again, smiling at me with his wicked smile. I knew he\'s laughing at me inside. Laughing at me for my stupidity, laughing at me for not letting it go...  It\'s been awhile, since i could, hold my head up high. It\'s been awhile, since i could, feel my own heartbeat. It\'s been awhile, since I could, say I love myself. It\'s been awhile, since I could, been addicted to the man behind the mirror. \"Just make these go away!\" is the thought, when I\'m alone with no one else. Everything I can\'t remember, everything seems like a dream. It\'s been awhile, since I\'m sorry...  The concequences are deep... and I had no one else to blame but myself. I kinda fucked things up again and again. \"why can\'t I just make all these go away? Why?\"   \"I will give in anything just to hear your voice, i will give in anything just to be with you...\" these are the words that are crying in my heart. Tears starts to flow downwards onto my cheeks. I hold my head up high, praying to someone else, praying...  This is just the beginning, the beginning to my painful prayers...  '),(16526,3344485,'15,May,2004','2004-05-15 00:00:00','   Everyday, I have to face him when i look into the mirror. The man is always looking at me, smiling and laughing at me.   \"let me out...let it go...\" voices that keep whispering to my ears when i see the man behind the mirror. I always sees him as my worst enemy that I had to face everyday. The man whom I do not want to see everyday. He is the man whom is not rational. Whom he loves pain, violent and cruelty. He is evil. He is crazy. He is barbarous. \"is that who i really am? Is that really me? Who am i? Really, who am i?\" these are my words that came out of my mouth when he\'s there, in the mirror. I do not know what I have become. I\'m so tired at times... repeating myself every single day, reminding myself who i really am, fighting my will against his... But there he is! He\'s always there when I look into the mirror, staring at me, smiling at me... so evil, so merciless... the unforgettable smile, the unforgettable laughter, the unforgettable voices that i hear, that i see... Sometimes, I kinda enjoy how he feels, what he thinks and what he wants... Its beautiful! No worries, no pain, no mercy... simply evil.  '),(16527,3344485,'16,June,2004','2004-06-16 00:00:00','   Since the day I met you, I can\'t forget your look, of what you said. I\'m addicted, I\'m so addicted to you.   When I see you down the streets, I can\'t pretend I don\'t know you, I can\'t pretend I do not care about you. I try, try not to think about the pain inside of me. But it\'s just too late and I\'m ok, I just need to stay away.  Missing you, wonder why? When your near, my heart pumps harder, when your far, my heart hurts. You always seems so far away, and you don\'t seems you care anymore. I\'m sorry, I can\'t always be there for you. I\'m sorry, I can\'t be your perfect guy. I can\'t stand it anymore, and nothing seems alright anymore.   No turning back, walk away, I try, but it seems too late. Nothing going to change the things you said, nothing going to change the things you done, just walk away, no turning back. I just wanna make you proud, but there is nothing I can do anymore.  I need a hero to rescue me. Always thought you can be my hero to rescue me, although you can\'t be perfect but I always think you are perfect. But you always seems you don\'t care anymore, why?  Now, I\'ve lost it all, I can\'t go back, nothing I can do anymore. I\'m sorry, I try, I already try, I just need to talk to you but you don\'t seems you understand me anymore. I\'m sorry, I\'m just not a perfect guy. I\'m just another guy who is standing behind the mirror, pretending to be a perfect guy, but it\'s just too late...  '),(16528,3344485,'05,June,2004','2004-06-05 00:00:00','   It\'s not easy to be me... I\'m not crazy, I\'m just a little bit unwell. Just trying hard to be like everyone else around me. Had hard times when i looked into the mirror where i can\'t face myself no more...   Most times, when i looked into the mirror, i had to find the man behind it, wonder where he had gone, wonder how come he\'s not there anymore? In my place, where i couldn\'t change, i was lost... lost in my own fucking place.. I was scared, i was not under prepared, i couldn\'t find myself anymore... god damn it, i was so scared...  Saying please, saying please... screaming out loud, in my place, in my heart, in my mind, where have i gone to? where have i been to?   All day, staring at the ceiling... all night hearing voices telling me, i might want to get some sleep because tomorrow will be a better day! It\'s me, it\'s only me, talking to myself in public, people looking at me, people staring at me, wonder am i crazy or something. laughing to myself, smiling to myself, just to make myself feel easy...  Can\'t remembered what had happened. Everything seems fine before and it just happened... I\'m so addicted, addicted to the man behind the mirror, can\'t forget his face, can\'t forget his smile, can\'t forget his laughter, can\'t forget him... He\'s so strong, so free... free of troubles, free from mortal... laughing, i\'m laughing now.. don\'t know why, don\'t know what happen... just laughing...  Guess i\'m just addicted...  '),(16529,3344485,'03,June,2004','2004-06-03 00:00:00','   when sun falls... i will remember what you said, although its not in my head but still i can\'t sleep day after day... why cant we talk about it? there can be a better way... you doesn\'t make me want to stay... there is nothing left to say and i got my eyes shut, praying they will stay that way.. and when i stay strong, that\'s what makes the difference everyday...    you said that we been in love, but you always give me a reason, is that ok? may that be the last kiss, i will walk away... i will keep my eyes shut and we\'ll pretend not to know each other..   why don\'t we stay together? why can\'t we stay together? what is holding you up? always there is a reason for you not staying togther.. everytime i try to talk to you, i get so excited, everything i say to you, comes out wrong, never comes out right.. always there is a reason from you for not staying together.. always there is a reason from you for loving me..   if i could make you mine, i will go wherever you will go.. if i could, i would be with you, i will follow wherever you will go.. tat is if i could, i would do so... if i only could...  '),(16530,3344485,'07,July,2004','2004-07-07 00:00:00','   Loneliness covers me throughout my life... i wonder why? I only ask for companion and concern, nothing more...   I look deep beneath my heart and soul, looking for an answer, a reason, a way out of this... I treat people as nice as i really could... i treat people with one heart and one soul, i help people and ask for nothing in return... but y am i still so lonely? what are the reasons?  Don\'t understand, really don\'t understand... why do i always have to be so lonely? am i destine to be this way? no matter how hard to try, no matter how hard i change, it seems nothing has changed.   People always say sorry, but nothing gonna change, saying sorry and expect nothing had happened, and arent gonna change their mistakes as well.. so what is the point of saying sorry when nothing can be done, when nothing can be changed or nothing is going to changed?  I try so hard to change myself, i really tried... people are greedy and ask for more.. so i give them more and they want more... i always try to satisfy them cos i know i needed a fren, a companion, who will give me abit of concern, who will standby me and lend me their listening ear... no one, no one could do that for me! why? don\'t understand, really confused.. what have i done to deserve this kinda treatment? what have i done to deserve this kinda loneliness?   Day after day, month after month, years after years... i have been waiting, waiting for a companion... now im getting very tired of waiting already... really tired of waiting... been so lonely for so many years... why do i felt so lonely? why do i have to wait for so long, so long?   Saddness....  '),(16531,3344485,'02,July,2004','2004-07-02 00:00:00','   I never meant to be like this.. He is whom am i to be.. He is whom i had become.. He is whom suppose to be me... He is the man behind the mirror..  I lived in fear once. Fearing what will happen to me next.. fear of losing my meals, fear of being alone, fear of adults, fear of humans, questions starts to arise... the \'why\' word starts appearing in my head...   Time flies, i learn to survive, learn to enjoy pain, learn to enjoy darkness, learn to enjoy the loneliness around me, learn to lie, learn to cheat... just to survive and i did..   As i grew up, i always wanted to change, change to be a better man and i did... tats why i never give up, never give up on people, i wanted give them hope, give them support that they need.. but why do they give up hope on me? \'stubborn\' is the word they use on me because i stated my reasons and my points of view. Chance is what i need and i have been chasing after it. i change wat i can and i already did my best but people still don\'t give me chance when i fail once... they just give up on me... why?  Now, i begin to like the name \'Sin\'. I wonder why? maybe it\'s because I am one. So i will start calling myself Sin and it seems i liked it alot.   I have been living in the darkness too long... now i\'m starting to break loose.. what will i become next? why do i fear myself so much? what will happen next? i think only i can explain...  '),(16532,3148889,'18,July,2004','2004-07-18 00:00:00','   My first post on my new weblog.  I started Fat Girl Films last month....well, I\'ve had the production company for a number of months but I\'m just starting to do something with it.  It has also undergone a few name changes as of late.  I finally decided on fat girl films because it\'s funny, cute and gets attention.....hmm, sounds just like me...exactly!    I\'m working a couple of short film scripts which will go into production in the next few weeks.  I am very excited and extremely nervous about this.  I also purchased a new 12 in. PowerBook G4 with SuperDrive....You gotta mention the SuperDrive just to let folks know that you spent a little extra cash.  I would have loved to get the 15 in. but the pocketbook was already starting to feel the strain of the 12in but hell, you gotta have the SuperDrive.  I have great eye sight and I can easily hook this sucker up to a larger screen if need be.    I so love this little laptop.  I totally changed platforms.  I used to own a Dell Inspiron 2600.....really great computer but there was always a little something off about it, like it wasn\'t meant for me but this PowerBook is totally me and I plan to do great things with it.  My creative juices are really starting to flow and I\'m starting to feel so confident about this new direction my life is taking.  I\'ve always known that I was supposed to do something where I get to be creative.  I\'ve tried singing, dancing, acting and even dabbled in graphic arts which I still love but this filmmaking thing is the cat\'s meow.....totally me.    I am also a writer.  I wrote a wonderful novel a couple of years back that I have yet to publish.  I will probably wind up doing self-publishing...it\'s more my speed.  I\'m a writer from way back and I\'ve always written poetry...got into short stories then finally got the courage to do the novel.  It took me approximately 6 months(3-4 days a week in the barnes and noble cafe).  I would go into the cafe, order a frappuccino or hot tea....sit in my little corner and get to writing.  Oh, it felt so good...I miss doing that....it was a great time in my life....that\'s kind of why I know I\'m going in the right direction because I love to write, I love being bossy so I\'m perfect for directing and producing....lol!  Anyway, I\'ve yapped for long enough tonight....laters!  '),(16533,3148889,'18,July,2004','2004-07-18 00:00:00','   I was in a yahoo group for filmmakers in Oklahoma but the bastard had too many rules and I am not one for conformity so I had to jet!  Perhaps, I\'ll start my own...hell, it can\'t be any worse....  '),(16534,3148889,'18,July,2004','2004-07-18 00:00:00','   How good is that song \'The Reason\' by Hoobastank.....soooooooo awesome!  I love that guy\'s voice.  He could sing the alpahbet and make me swoon......  '),(16535,3148889,'18,July,2004','2004-07-18 00:00:00','   I was supposed to start filming on a documentary today but the subject of my documentary has decided to make other plans.  It totally sucks so I will have to reschedule her for next week but I also have another one for filming next sunday but that should be fairly easy.  It\'s just this little idea that popped into my head.  It\'s mostly improv with a few lines provided by moi.  It looks really funny on paper...hoping for that to translate onto film.  alrighty, got to get out and enjoy this day....laters!  '),(16536,3148889,'18,July,2004','2004-07-18 00:00:00','   Not much going on today but I did happen to catch that cute Citi.com commercial with the little old lady dancing.  It cracks me up so much and more so because the old girl has got a little rhythm.  It\'s 4am.....I should be in bed but I\'m going to watch \'View from the Top\' starring Gwyneth Paltrow, Mike Myers and Christina Applegate among others....btw, why is mena suvari, kevin bacon and andie macdowell in \'Beautyshop\' the new movie starring Queen Latifah?  Things that make you go hmmm?  laters!  '),(16537,3148889,'18,July,2004','2004-07-18 00:00:00','   My new favorite show is Aqua Teen Hungerforce on Cartoon Network.....Love It!            urlLink aqua teen hungerforce  '),(16538,3148889,'18,July,2004','2004-07-18 00:00:00','   We\'re having a beautiful day in Oklahoma City.  The sun is out but it\'s Oklahoma City, the weather can change any minute.  It\'s early in the day and I\'ve worked on two short film scripts.  I\'m very excited.  I have two short films ready to shoot and a third almost ready....it\'ll be ready by Sunday afternoon.  According to Tech TV, friendster.com is so 5 minutes ago.....  '),(16539,3148889,'18,July,2004','2004-07-18 00:00:00','   I just found out about the American Idol results....I can\'t believe that Jennifer Hudson got the boot!  Are people crazy?!  I\'m not saying that she had the best performance of the night but she certainly did not have the worst one.....Will people please stop feeling sorry for John Stevens!  He\'s not improving....well, at least his hair is...lol!  He\'s a one-trick pony and that trick is played out!  '),(16540,3148889,'18,July,2004','2004-07-18 00:00:00','   What a day!  I currently live in Oklahoma City and I had my first taste of tornadic weather.  I was so scared today.  We had hail ranging from your normal pea size to baseball size.  It was crazy.  A few tornadoes touched down but I wasn\'t a witness to any but I was very close to the center of the madness today....twice.   I saw collapsed carports....dented vehicles and hail that was two feet deep on some streets....so much hail that it looked like it had snowed.  The biggest news of the day....I got caught in the ice on one street and this guy helped to free me but the truck had so much momentum that it took off and I couldn\'t catch up as it jumped a curb and landed in someone\'s yard.....yes, it really happened.  Now my truck is stuck in someone\'s ice-covered front yard.  The strangest part was that no one was trying to help me.  People were just standing around watching me struggle with my truck.  Finally, one older gentleman came to help me....he recruited two other guys to push my truck as I put the sucker in reverse....All I have to say is......FREE, I\'M FREE AT LAST....  '),(16541,3148889,'18,July,2004','2004-07-18 00:00:00','   I just finished watching American Idol.  Fantasia killed it.  She was so awesome....she really took the concept of rearranging the song to suit yourself.  Everyone did a great job except for John Stevens.  He\'ll probably go tomorrow.  I know I said everyone did a great job but most of them were boring even Latoya\'s....Fantasia really woke me up and ended the night on a great note.  Paula Abdul was just as boring.  Simon was actually fairly tame.  Randy just gets on my nerves.....\"yo, dawg.\"  \"what\'s up, dawg?\".....he\'s way too old to be using that type of lingo plus it doesn\'t sound right coming out of him......No Mo Rocca dreams today....darn!  '),(16542,3148889,'18,July,2004','2004-07-18 00:00:00','   I\'m finally back.  I\'ve been without my new laptop since last thursday.  The little bugger was acting up.  The keyboard was replaced and now all is well.  I had the strangest little dream about Mo Rocca today.  I had a dream that he broke into my apartment and decided to move in with me.  I don\'t think there was anything sexual about the incident although, I am a fan.  Just strange.  I finally bought a new camcorder.  I\'m ready to get down to business.  I\'ll have my first short film done in the next few weeks.  I\'m very excited.  Gotta jet. laters....  '),(16543,3148889,'18,July,2004','2004-07-18 00:00:00','   Oh yeah!  Happy First of the Month!  Rent is Due!  '),(16544,3148889,'18,July,2004','2004-07-18 00:00:00','   I am actively pursuing this disc jockey thing....I\'m a huge music fan and I\'m a fan of everything.....I can always pick out the hottest tracks......this is totally me PLUS! you get to work at night.....free drinks....hell yeah! I think filming is going to be delayed again until around May 4th...sucks!  '),(16545,3148889,'18,July,2004','2004-07-18 00:00:00','   Don\'t chuck things at your computer after I say this but I think the Friends theme song sucks and has always sucked.....it\'s just too corny!  '),(16546,3148889,'18,July,2004','2004-07-18 00:00:00','   We have a serious little storm starting to brew in OKC.  Interesting litle tidbit.....out of all of The Apprentice alums.....Omarosa has made the most money post-Apprentice......She\'s made 1.1 million dollars......it pays to be a __________!  '),(16547,3148889,'18,July,2004','2004-07-18 00:00:00','   I\'m off today and it\'s driving me crazy.  My apartment is a mess and I just want to relax but I can\'t because of the damn mess.  Guess I could clean it....hmm?  I\'ll think it over...   The USB cable for my printer isn\'t working...grrr!  This whole Mac thing is starting to piss me off....I\'m almost ready to go back to Windows!  '),(16548,3148889,'18,July,2004','2004-07-18 00:00:00','   MUSIC TOURS    I would appreciate it if music artists wouldn\'t bypass Oklahoma City.  I don\'t want to have to travel to Dallas, TX all the time to catch my favorite artists.  OKC has a huge local music scene......Big THANKS to The Strokes and Kid Rock for coming out and rocking OKC.  The Strokes took over our local rock station  urlLink The Buzz  with Strokes Radio playing all of their favorite tunes.....awesome!      A CHANGE    I can\'t begin to tell you how much I need a change.  I need a new hairstyle.  I kind of want to cut it all off.  Right now, it is shoulder length and VERY thick.....tres annoying in this climate!  I like wild out of control hair.  I gotta do something new.....FAST!  My current hair state does not reflect me.  I\'m an artist and I need to have that artist look.....It\'s going to happen really soon, in like the next two weeks.  My hair used to be super short and bleach blonde...it was awesome and total freedom....I probably won\'t do anything that extreme but it\'s definitely going to be something awesome and totally me!  Oh yeah, I want to be a DJ.  New career, ahoy!  '),(16549,3148889,'18,July,2004','2004-07-18 00:00:00','   I bought my mother an awesome ring for Mother\'s Day.  It\'s sterling silver(hey, i\'m a poor black chick) with fresh water pearls and sapphires(all real baby!).  It\'s so gorgeous....she could just tell someone that it\'s platinum....it looks that awesome!  I\'m also going to get her one of those giant greeting cards.    I kind of had a slight panic attack about an hour ago.....well, i\'m still coming down off of it.....I NEED VALIUM!  '),(16550,3148889,'18,July,2004','2004-07-18 00:00:00','   I just finished watching the Mary J. Blige DVD that came with my CD.  It\'s pretty good.  I have a lot of respect for that woman.  I actually wrote a poem about her a few years back that\'s on my  urlLink  personal site .  Her new CD \'Love and Life\' is just awesome.  I didn\'t become a Mary fan until around 1996 when I heard her song \'Everything\'.  I\'m not a big R & B fan but I love real music and Mary makes real music.  My favorite song on her new album would have to be \'Don\'t Go\'.  '),(16551,3148889,'18,July,2004','2004-07-18 00:00:00','   General Wesley Clark   After watching the television show STAFFERS on  urlLink Discovery Times   and kind of getting to know some of the candidates about a month too late....I can honestly say that I would have voted for General Wesley Clark.  I just really liked the way he interacted with his staffers as well as the public.  I wish others had the chance to get to know this man.  I\'m not saying I\'m a democrat but General Clark really impressed me.  '),(16552,3148889,'18,July,2004','2004-07-18 00:00:00','   Had some bad chicken today.....grrr!  Toilet, ahoy!  Tough Crowd and Chappelle\'s Show was hilarious tonight.  Bought my firewire on eBay for $6....yay!  Bought my USB cable for my printer at CompUSA for $23.....grrrr!  I should have bought it on eBay but I really needed it as I need to print some scripts tomorrow for an actor.  I have a beef with Mac.....why don\'t they have all of the drivers pre-installed like Windows XP....I feel like I\'m Mac\'s mom....I have to tell it....\"Well, Mac, you must go to a website and download the driver for what you would like to use.\"  Steve Jobs sucks at this moment for me.  I\'m back in love with Bill Gates until Regis and Kelly come on...lol!  '),(16553,3148889,'18,July,2004','2004-07-18 00:00:00','   Incoherent Ramble  Sean \"P. Diddy\" Combs is doing \"A Raisin in the Sun\" on Broadway.  That\'s very cool because hopefully it will open young african-americans to the wonderful world of Broadway.  P. Diddy really surprised me with his performance in \"Made\" with Jon Favreau and Vince Vaughn as well as his bit part in \"Monster\'s Ball\".  I personally think that Halle Berry shouldn\'t have won the Oscar for that role....she sucks as an actress but for some reason people like her....go figure.  Anyway, back to the main topic...the critics are being way too hard on Mr. Combs.....this is his Broadway debut!  This guy comes from hip-hop....he auditioned twice and the producers thought that he would do an adequate job of playing the main character.....that\'s major.  My beef with critics is that they come in with preconceived notions.....I absolutely hate that.....that\'s kind of the reason why we still have racism in this world.  I\'ve whined enough...gotta jet.....laters!  '),(16554,3148889,'18,July,2004','2004-07-18 00:00:00','   Worst video ever  urlLink The Comfort Zone  by Vanessa Williams.  '),(16555,3148889,'18,July,2004','2004-07-18 00:00:00','   Learn Pizza Customer Etiquette  urlLink Andy\'s Rant  '),(16556,3148889,'18,July,2004','2004-07-18 00:00:00','   Just found this cool Daria shrine  urlLink The Daria Shrine ...enjoy!  '),(16557,3148889,'18,July,2004','2004-07-18 00:00:00','   My pre-production meeting went really well.  We\'ll start filming on May 2, 3....yay!  I got an offer from this guy in Tulsa to help me with my audio for FREE...awesome!    If you guys haven\'t seen this new search engine....you gotta check it out  urlLink booble .    I was in the Taco Bell drive through when I got an unexpected compliment \"you have a really great smile\".....seriously cute latino manager.....hey!  Favorite song of the day \'Baby I Love You\' by Jennifer Lopez.  Btw, while I\'m on the subject of music PJ Harvey is finally coming out with some new music  urlLink PJ Harvey .  She\'ll be doing a few dates on the lollapalooza tour if I remember correctly.  I\'m so addicted to Jane\'s Addiction lately.....lol....I just thought about that....hilarious!  I totally need to go out and shake my booty....this dance mix I\'m listening to is making me feel real good!  '),(16558,3148889,'18,July,2004','2004-07-18 00:00:00','   Yeah baby!  I\'m finally awake and ready to face the world....Mmmm, I love days off!  I need to start on two new script ideas and I also have a pre-production meeting with the subject of my documentary.  I\'m very excited to get things started.  '),(16559,3148889,'18,July,2004','2004-07-18 00:00:00','   I sold a camcorder on ebay today....yay!  Now, I\'m going back to bed since I\'ve only had 3 hrs of sleep....  '),(16560,3148889,'18,July,2004','2004-07-18 00:00:00','   I was just catching the tail end of Parkinson Starburst on BBC America.  Samuel L. Jackson and Robbie Coltrane(harry potter films; hagrid) were the guests along with some playwright.  The playwright made the comment that he\'s the fart between two huge buttocks.  He was, of course, referring to the guys being in two huge movies at the time Star Wars and Harry Potter.  I suck so much at trying to get these jokes across.....perhaps, i\'ll stop....  '),(16561,3148889,'18,July,2004','2004-07-18 00:00:00','   Downloading some Daria episodes since MTV is being an ass about putting out the damn DVD.  MTV, you could have made $50 -100 off of me and countless others!  Why must you make me steal?!  '),(16562,3148889,'18,July,2004','2004-07-18 00:00:00','   ITALIAN SODA   I love getting italian sodas at the Barnes and Noble cafe but in Oklahoma City, they don\'t offer irish cream....wtf?!  It\'s ludicrous that you can\'t get an irish cream italian soda!  It\'s the best most refreshing drink ever.... with or without the whipped cream.....I settled for the vanilla italian soda which isn\'t nearly as good but still good.  The italian soda at Borders is pretty good but they don\'t put milk in the drink which takes away from the kind of import look of the drink.....  '),(16563,3148889,'18,July,2004','2004-07-18 00:00:00','   I was watching the Comedy Central Bar Mitzvah earlier and Dane Cook was one of the performers.  I love that guy...he\'s awesome.  Anyway, in part of his bit he says he enjoys watching Justice Files because 1. He loves justice and 2. He likes files.  He loves it so much that he enjoys a justicey jizz....lol!  Hilarious!  I\'m sure I just fucked up his great joke but if you laughed then I\'ve done my job....  Btw, do any other black people in the world watch The Office on BBC America....oh yeah, I watched it before it became mainstream....suckers!  '),(16564,3148889,'18,July,2004','2004-07-18 00:00:00','   I wish I had known my direction when I first left high school.  I thought I wanted to be a sociologist and I also attended a black college because I wanted that \"black experience\".  It didn\'t work....I was a fish out of water around my own people.  I still feel that way because my interests lie elsewhere.  Sure, I love hip-hop music but I\'m a bigger fan of rock music.  I watched Comedy Central before it was cool for black people to watch it.  I love Frasier and I\'m totally bummed that it\'s going off the air.  I watch Friends but I\'m not really a fan....it never had an impact on me like frasier.  I love the Bernie Mac Show and Girlfriends....they are smart shows although some of Girlfriends\' characters are a little over the top and the episodes of late are getting really cheesy.....it\'s like they have run out of ideas but Bernie\'s show is soooooo good!    Back to what I was originally talking about....I studied music(trumpet) in high school...left that because of a really tyrant of a band director.  I went into creative writing....this is where I really blossomed but I didn\'t get enough out of it because it was the only class they offered.  I didn\'t have anyone to nurture me after that class.  Sometimes, it\'s about what you can do for yourself and going to a black college.....you don\'t get that support.  Black people have the same preconceived notions as any other race but they lie more on the socioeconomic level.  I tried to fit in with my classmates but it didn\'t work and they took my kindness for weakness.  There\'s so many things I wish I could do over.  I wouldn\'t change much but I probably would have stayed in Texas to attend college...a predominately white college.  I hate that I feel this way but it\'s my own people who have made me the outsider.  I\'ll become a great independent filmmaker and never look back....ha!  The best line I\'ve heard all year was on Chappelle\'s Show on Comedy Central when Wayne Brady says \"Is Wayne Brady gonna have to choke a bitch?\"  '),(16565,3148889,'18,July,2004','2004-07-18 00:00:00','   I am so PISSED OFF!  Apple iBook now comes with an optional SuperDrive on the 14 inch model. It will run about 1499 with the addition of the SuperDrive.  I just spent $2000 on a 12 inch PowerBook with SuperDrive......grrrr! If only I had waited two weeks....damn it to hell!  Btw, why is it that Apple doesn\'t hook you up with a word processor like a Windows computer.  I feel so cheated.  Paying all of this money and I don\'t even get the thing I need the most......I\'m a freaking writer!  I had to purchase a word processor.  It works great but I am just disappointed.  They give you a test-drive of Office for Mac for 30 days then you have the option of buying it for nearly $400....the reason I left a Windows based computer was to get away from any Microsoft......  '),(16566,3148889,'18,July,2004','2004-07-18 00:00:00','   Just woke up....typical Sunday.  Eat some pizza left over from last night...yum!  Wash some clothes....ewww!  Watch \'King of the Hill\' and \'The Simpsons\'.....perfect!  '),(16567,3148889,'18,July,2004','2004-07-18 00:00:00','   Have you ever found yourself not appreciating yourself whether it be by defiling your body sexually, mentally and chemically?  I don\'t know why I\'m bringing this up but I was just listening to some moody music and I guess that inspired that brand of thinking......okay, back into my shell.....  '),(16568,3148889,'18,July,2004','2004-07-18 00:00:00','   Oh my goodness.....that stormfront website is out of control....i know there are still racists out there but it\'s crazy to dislike a person because of the color of their skin, facial features and such... It made me queasy to my stomach and frankly, a little afraid......  On Tuesday, a famous soccer announcer from the UK called a famous soccer player a \"fucking lazy thick nigger\"....needless to say.....he lost two jobs in a 24 hour period..... urlLink ron atkinson racist remark  '),(16569,3148889,'18,July,2004','2004-07-18 00:00:00','   I stumbled upon this truly horrible message board about white nationalists....it\'s crazy that people still think like this in 2004!  urlLink stormfront white nationalist community  '),(16570,3148889,'18,July,2004','2004-07-18 00:00:00','   I just came up with this awesome story idea for a script.  I don\'t think anyone has ever I can\'t wait to get to work on it but I have another script to finish by tomorrow afternoon.  I also need to get up in the morning and pick a firewire cable from bestbuy.  Damn me and my lazy ass.....I should have done that like three days ago!  I still think Limp Bizkit is awesome.....they\'re just hindered by Fred Durst and his \"love life\".  Fred, leave those hoes alone!  '),(16571,3148889,'18,July,2004','2004-07-18 00:00:00','   Hello All! I\'m finally back.  The Mac is okay but safari apparently is busted.  I\'ll have to wait until safari fixes the problem.  Oh well...just wanted to say hello.  Oh yeah, filming starts on May 9th...yay!  '),(16572,3148889,'18,July,2004','2004-07-18 00:00:00','   I am all about The Crystal Method and yeah, Phil Collins with a little Tom Petty today.  Your eyes aren\'t deceiving you.....I said Phil Collins!  If you don\'t know about the musical group  urlLink Air  then you are missing out on some wonderful music.  If you\'re a newbie to Air then I suggest their album  urlLink MOON SAFARI .....you\'ll become an instant fan.....I might be away for a few days....damn Mac is acting up again...It\'s not even a month old!  '),(16573,3148889,'18,July,2004','2004-07-18 00:00:00','   urlLink David LaChapelle\'s  website is my new favorite.......love it!  '),(16574,3148889,'18,July,2004','2004-07-18 00:00:00','   I\'m such a freaking tool....I\'m having a ga-ga moment over Brad Pitt....that is so unlike me but I must admit....that man is HOT!  Forgive me, fellow geeks for indulging in a little bit of lust....  '),(16575,3148889,'18,July,2004','2004-07-18 00:00:00','   Punk\'d is so much better this year!  The only real news today is that I will be rescheduling the shooting for my first short film until next week.....it sucks but it can\'t be helped....oh well...  '),(16576,3148889,'18,July,2004','2004-07-18 00:00:00','   I\'m on eBay trying to find a cool new ring.  The shipping and handling for some of the rings is outrageous......$6.95 to ship a ring!  It\'s ludicrous.  I sell on eBay from time to time, so I know it shouldn\'t take more than $2.00 to ship a ring....that\'s including packing materials and postage.  Okay, back to my search....  '),(16577,3148889,'18,July,2004','2004-07-18 00:00:00','   Had a great moment with this super cute guy at 7-11 earlier....There\'s something about the attention of a man...especially an attractive one that makes one giddy.....I\"m going to bed.....  '),(16578,3148889,'18,July,2004','2004-07-18 00:00:00','   No one would be questioning the antics of the  urlLink Vanity Fair  editor-in-chief if he weren\'t black.  '),(16579,3148889,'18,July,2004','2004-07-18 00:00:00','   Wild James Spader movie from 1996 called  urlLink Crash ....it\'s about people exploring their sexual fetishes which basically centers around car crashes....much wilder than 9 1/2 Weeks which I always thought was kind of tame....  '),(16580,3148889,'18,July,2004','2004-07-18 00:00:00','   urlLink Frasier finale  did not disappoint.  Loved it!  '),(16581,3148889,'18,July,2004','2004-07-18 00:00:00','   FRASIER FINALE:  I\'m heartbroken to see my favorite show leave the air.  '),(16582,3148889,'18,July,2004','2004-07-18 00:00:00','   I am so irritated with the state of popular music.  The radio stations suck.  Down with Clear Channel!  '),(16583,3148889,'18,July,2004','2004-07-18 00:00:00','   Oklahoma was listed at 47 out of 51 states for being healthiest for women.  My former homestate of Texas was 46.  Minnesota was ranked #1.  No wonder, I\'ve been plagued with allergies and such.  I have some filming to do tonight...that should go well....looking forward to finishing up so I can start editing....  '),(16584,3148889,'18,July,2004','2004-07-18 00:00:00','   I just read that William Hung from American Idol is 31 years old....I totally thought that guy was like 21 or something.....Worst than that...the guy is still a virgin....I\'m not really knocking him for that but his mother still accompanies him to media events...isn\'t that a little odd?  When I first saw Mr. Hung...I thought he was \"special\" because he has that \"special\" look...and there\'s also the way he dances and sings which might lead one to think that he was \"special\".  Anyway, I gotta jet....big things starting to happen in Oklahoma.....  '),(16585,3148889,'18,July,2004','2004-07-18 00:00:00','   I just saw a really cool movie  Dancer, Texas Pop. 81 ...oh yeah, Ethan Embry was in it.....cute little movie with heart....  '),(16586,3148889,'18,July,2004','2004-07-18 00:00:00','   Today, I am the biggest Dido fan!  '),(16587,3148889,'18,July,2004','2004-07-18 00:00:00','   SHOOTING TODAY   Yay!  Today is the day I begin shooting my first short film.  So exciting.  Oh yeah, I caught this interesting documentary called Scratch the other night on IFC or Sundance....I can\'t remember...anyway, look it up...you won\'t be sorry...laters....I\'m running late....  '),(16588,3148889,'18,July,2004','2004-07-18 00:00:00','   I got a location for the film festival.  It will be at the Deli on the Labyrinth.  It\'s an awesome little place....great tea!  I\'m so excited!  '),(16589,3148889,'18,July,2004','2004-07-18 00:00:00','   For those of you who watch SNL...you will know  urlLink Fred Armisen .  Hilarious guy.  '),(16590,3148889,'18,July,2004','2004-07-18 00:00:00','   Dave Attell was hilarious on Conan tonight.  Look for his new hour-long special on Comedy Central May 31st.  Very close to getting a location for the film festival.  Crossing my fingers.  Oh yeah, check out the  urlLink Jerry Seinfeld  and Superman short films.  The second one is my favorite.  '),(16591,3148889,'18,July,2004','2004-07-18 00:00:00','   Hello All!  I\'ve got the entry forms up for the film festival.  Feeling really good about the whole thing.  Just need ten entrants and this thing will happen.  Hell, I might even do it with five.  Hope all is well.  Not much to say because I\'m stressed about finding a location for the festival by May 21st.  Laters  '),(16592,3148889,'18,July,2004','2004-07-18 00:00:00','   Cool new singer... urlLink John Legend .  Check him out if you like great soul music.  '),(16593,3148889,'18,July,2004','2004-07-18 00:00:00','   I wish people like Britney Spears and Janet Jackson would go away.  They add nothing to society as a whole.  Their music is lame for the most part.  I\'m not going to lie and say I don\'t listen to a few tracks from each artist but I can\'t take them seriously like say a  Coldplay or Dave Matthews Band.  They are cancers on our society.  Most people would say that it\'s jealousy but why would I want to be known for lip-synching and showing my ass....Nah, that\'s not me...  Janet Jackson is way too old to still be acting like she\'s still 21 and Britney is just too damn wild.  I know young people are to get out and enjoy their 20s but she\'s just annoying to me(I\'d like to read my US Weekly without reading a Britney article) and grown people act like they worship her.  Truth be known, if she didn\'t have that body she wouldn\'t be anyone....  '),(16594,3148889,'18,July,2004','2004-07-18 00:00:00','   I finally got the official film festival rules up as well as all of the release forms.  I\'m trying to locate some judges.  I already have two judges secured.  I need three more.  I need to secure the location by May 19th.  '),(16595,3148889,'18,July,2004','2004-07-18 00:00:00','   Finally starting to get some details of the film festival together.  A little disappointed in the membership at the yahoo group I started for this event.  I\'m sure things will start to pick up by May 19th.  That\'s when I will start accepting entry forms.  '),(16596,3148889,'18,July,2004','2004-07-18 00:00:00','   Is this the real  urlLink Brad Pitt ?  Scroll to the end of the page.    There was a high school graduation down the street from my place last night.  Made me reminisce about the hell of high school...kidding.....made me wish I could start all over.  I would make so many changes.  I wouldn\'t have gone to Grambling State University and wasted all that money and time.  I would have stayed in Texas, gotten a job while going to school.  Nah, now that I think about it, everything I\'ve gone through was worth it.  All of the ups and downs and feeling like a clown.  I\'m finally heading in the right direction.  Hell, who cares if it took me 8 years and loads of tries.  I\'m getting there and a lot of people will know my name by year\'s end.  '),(16597,3148889,'18,July,2004','2004-07-18 00:00:00','   For those of you who dig that Limp Bizkit sound but not the Limp Bizkit lyrics.....I highly recommend  urlLink tobyMac .  He\'s originally from the Christian rap/rock group DC Talk.  '),(16598,3148889,'18,July,2004','2004-07-18 00:00:00','   Here\'s a cool young filmmaker  urlLink Stephen .  He\'s 13 and he designed his own website.  '),(16599,3148889,'18,July,2004','2004-07-18 00:00:00','   The day is going well.  I can\'t really complain.  I\'m really very bored.  I don\'t feel like going to work but, I must!  Life isn\'t fair when Madonna gets to sleep with Guy Ritchie every night and I don\'t!  Okay, I\'m done talking nonsense.  Laters!  '),(16600,3148889,'18,July,2004','2004-07-18 00:00:00','   I don\'t know if I should be revealing this but I just got an invitation to join a couple for relations from a dating site I\'m a member of.  They\'re from Washington state and they\'re visiting Oklahoma. Btw, I\'m not going!  '),(16601,3148889,'18,July,2004','2004-07-18 00:00:00','   Don\'t say I never gave you anything...open your eyes to  urlLink zebox .  '),(16602,3148889,'18,July,2004','2004-07-18 00:00:00','   I don\'t know if I\'ve ever made a post about  urlLink David LaChapelle  but he\'s so awesome.  '),(16603,3148889,'18,July,2004','2004-07-18 00:00:00','   urlLink Ali G. , you\'re welcome....  '),(16604,3148889,'18,July,2004','2004-07-18 00:00:00','   How much do you love celebrity poker showdown on  urlLink Bravo ?  The episode with Wanda Sykes was hilarious!  '),(16605,3148889,'18,July,2004','2004-07-18 00:00:00','   If you didn\'t love me before, you\'ll definitely love me for finding this  urlLink link .  '),(16606,3148889,'18,July,2004','2004-07-18 00:00:00','   It stings so much more to get rejected by a black man.  It seems like more and more black men just don\'t want anything to do with black women.  This concerns me greatly because I always thought that I would marry a black man.  I haven\'t really dated a black man since 2001....isn\'t that sad?  I don\'t mind dating white men but when the men of your own race don\'t want you, that\'s just more than one should have to take.  I\'m not one of those ghetto girls and I don\'t have any children.  I work, I have my own vehicle, my own place and I\'m somewhat of an entrepreneur.  '),(16607,3148889,'18,July,2004','2004-07-18 00:00:00','   My new favorite  urlLink FREE  game site...enjoy!  '),(16608,3148889,'18,July,2004','2004-07-18 00:00:00','   Btw, I\'m shooting a short film this week and the deadline for the film festival entries are on Saturday...yay!  '),(16609,3148889,'18,July,2004','2004-07-18 00:00:00','   urlLink Jennifer Lopez  and Marc Anthony reportedly were married on Saturday a little after 6PM.  It\'s not really that interesting but I\'m starting to think that child is either a ho, homewrecker or is just afraid to be alone.  She needs some major counseling.  '),(16610,3148889,'18,July,2004','2004-07-18 00:00:00','   Congrats to  urlLink Phylicia Rashad  on becoming the first black woman ever to win a Tony award in a lead actress role...awesome!  Only took 58 years!  '),(16611,3148889,'18,July,2004','2004-07-18 00:00:00','   I spent half the night listening to Chris Rock and Dave Chappelle stand-up routines....one of the best nights of my life....laughter all night!  '),(16612,3148889,'18,July,2004','2004-07-18 00:00:00','   I just finished watching  urlLink Love Actually .  It was so wonderful!  I was crying throughout the whole thing, so freaking romantic.  Perfect script!  '),(16613,3148889,'18,July,2004','2004-07-18 00:00:00','   song of the day: realizer by the crystal method  '),(16614,3148889,'18,July,2004','2004-07-18 00:00:00','   I am so sick and tired of people giving  urlLink Bobby Brown  chances.  He\'s a fuck up and if it was anyone else, their asses would be in jail.  Btw, I\'ve been 26 for 6 months as of 6/15...yay!  '),(16615,3148889,'18,July,2004','2004-07-18 00:00:00','   Btw, Fantasia Barrino\'s single  urlLink I Believe  is coming out June 22!  '),(16616,3148889,'18,July,2004','2004-07-18 00:00:00','   The  urlLink Pledge of Allegiance  controversy is ignorant.  I don\'t believe children should have to recite it.  I never did as a child but that\'s only because I hate conformity.  I like that it has \"under God\" in it but I\'m not going to force some little kid to say something they don\'t believe in.  If you don\'t believe in God then that\'s something you\'ll have to deal with on judgment day.  Good luck on that to all non-believers!  I\'m a christian and proud of it!  '),(16617,3148889,'18,July,2004','2004-07-18 00:00:00','   I absolutely love being a member of  urlLink Triggerstreet  but some days, reviewing can be the pits.  It seems like some people aren\'t even trying to put up quality work anymore.  Seems like they\'re just trying to get something up as quickly as possible.  I feel like it hurts the integrity of the site.    Song of the day:  Love Profusion by Madonna  '),(16618,3148889,'18,July,2004','2004-07-18 00:00:00','   SONG OF THE DAY: Take a Picture by Filter  '),(16619,3148889,'18,July,2004','2004-07-18 00:00:00','   Serena Williams lost today at Wimbledon because I overslept and missed the whole thing.    I was so pissed when I went on ESPN\'s website and read that she had been beaten.  It\'s just a whole sports thing with me when you think if you watch your favorite team or player that they will win because you\'re there supporting them.  Yes, it\'s strange but that\'s me.  '),(16620,3148889,'18,July,2004','2004-07-18 00:00:00','   I\'ve never been the biggest Bill Cosby fan but he spoke nothing but the truth in this  urlLink article .  I totally agree with him.  Has anyone noticed that  urlLink Franz Ferdinand  and  urlLink Modest Mouse  sound just alike and their videos are fairly similar?  '),(16621,3148889,'18,July,2004','2004-07-18 00:00:00','   Origins of  urlLink chess .  '),(16622,3148889,'18,July,2004','2004-07-18 00:00:00','   Carson from \'Queer Eye\' deep throated a weiner on last night\'s show.  His response after doing so was \"I guess I do have gag reflex.\"  LOL!  '),(16623,3148889,'18,July,2004','2004-07-18 00:00:00','   urlLink Sony  is coming out with an awesome 20 gig digital walkman to rival Apple\'s iPod.  '),(16624,3148889,'18,July,2004','2004-07-18 00:00:00','   SONG OF THE DAY: Far Away by Candlebox  '),(16625,3148889,'18,July,2004','2004-07-18 00:00:00','   urlLink Pat Boone  has his own Moblog!  '),(16626,3148889,'18,July,2004','2004-07-18 00:00:00','   HAPPY NEW MONTH!  '),(16627,3148889,'18,July,2004','2004-07-18 00:00:00','   My tombstone will read:  Here lies Keela, harsh and demanding, unable to love herself and others.  Isn\'t that sad?  But, so true.  '),(16628,3148889,'18,July,2004','2004-07-18 00:00:00','   Halle Berry is going to star in the new version of  urlLink Foxy Brown .  Why?!!!!!!!!!!!!!!!!!!!!!!!!!!!!  That\'s just bad casting.  She\'s not cool enough to fill Pam Grier\'s shoes.  Outrage!  '),(16629,3148889,'18,July,2004','2004-07-18 00:00:00','   SONG OF THE DAY: You Don\'t Want Drama by 8ball & MJG  '),(16630,3148889,'18,July,2004','2004-07-18 00:00:00','   If one believes homosexuality is wrong, why are they labeled homophobic?  I feel for today\'s Christian, we are to have open hearts yet we aren\'t to accept one when different.  It\'s not too hard for me but I have noticed that some churches are open to everyone as they should be but they seem to have forgotten the Lord\'s word when it comes to homosexuality.  You must lead others out of temptation no matter the sin.  I\'ve always been taught to and just naturally feel it\'s my place to embrace everyone but not the \"lifestyle\" or whatever.  I don\'t believe in harming someone just because their beliefs, morals, interests or race is different from another.  Totally moronic to hate someone so much that you would want to hurt them.  Why are some gays total stereotypes?  Women acting like men, men acting like women.  Women looking like men, men looking like women.  Doesn\'t that seem a bit strange?  It would seem like a man would just be a man even if he enjoyed the company of another man and the same for the women.    I was reading an article about a music artist\'s concert being cancelled due to so-called homophobic overtones.  That\'s what inspired this post.  '),(16631,3148889,'18,July,2004','2004-07-18 00:00:00','   I\'m starting on a new short film.  Filming will be on July 24th.  I\'m very excited and nervous as hell!  In the meantime, I\'m working on scripts and trying to improve my directorial skills.  '),(16632,3148889,'18,July,2004','2004-07-18 00:00:00','   GUILTY PLEASURE:  urlLink Half-Baked  starring Dave Chappelle  '),(16633,3148889,'18,July,2004','2004-07-18 00:00:00','   Caught Napoleon Dynamite last night.  It was really great, made for folks who lovely quirky films.  '),(16634,3148889,'18,July,2004','2004-07-18 00:00:00','   Just caught the Bridget Jones: The Edge of Reason  urlLink trailer .                I really don\'t know what to say about it.  It wasn\'t good but it wasn\'t bad.  The trailer wasn\'t much, I\'ll definitely need to see more to make an educated decision to waste my money.  '),(16635,3148889,'18,July,2004','2004-07-18 00:00:00','   I finally got around to watching 8 MILE.  The sex scene was pretty decent.  A lot of Junior Mafia was blaring throughout this piece, I kind of liked that, reminded me of my college days.  Em, did a good job.  It\'ll be interesting to see him do more films.  '),(16636,3148889,'18,July,2004','2004-07-18 00:00:00','   SONG OF THE DAY: Hell is around the corner by Portishead featuring Tricky  '),(16637,3148889,'18,July,2004','2004-07-18 00:00:00','   I never feel like I\'m going to get car jacked in OKC.  Is that strange  or have I developed this sick sense of security?  I hope I\'m not setting myself up by saying that.  Next post will read, \"I got jacked!\"  LOL!  '),(16638,3148889,'18,July,2004','2004-07-18 00:00:00','   I am a woman but I can\'t stand dealing with women and children.  I get hives!  The whining and nagging...ugh!  I hate it when I catch myself acting like that.  '),(16639,3148889,'18,July,2004','2004-07-18 00:00:00','   New David LaChapelle  urlLink Declare Yourself  campaign ads.  They\'re wild and featuring some celebrities like Christina Aguilera and Andre 3000 of Outkast.  '),(16640,3148889,'18,July,2004','2004-07-18 00:00:00','   In my little suburb of Warr Acres, we were treated to an impromptu fireworks display at the neighborhood baseball field/lake.  It was beautiful and a lovely treat after a boring day at work.  '),(16641,3148889,'18,July,2004','2004-07-18 00:00:00','   urlLink Faithless  is your typical pop/political/hip-hop/rock kind of band.  I suggest you get to know them.  Check out their current single  Mass Destruction.  '),(16642,3148889,'18,July,2004','2004-07-18 00:00:00','   urlLink Eddie George  is going to the Dallas Cowboys!  Yay, finally a hot guy on the team again!  '),(16643,3148889,'18,July,2004','2004-07-18 00:00:00','   John Kerry and George Bush are lame.  We need someone who is going to stick it out no matter what the controversy is.  When I read that John Kerry\'s campaign decided to distance themselves from Whoopi Goldberg after her Bush remarks; I just thought that was so rude and showed me that he has no backbone.  I\'m not a republican or democrat but I will vote for the one who kisses a  FAT BLACK BABY .  There\'s no real difference between the two, so I think that\'s the only deciding factor left.  '),(16644,3148889,'18,July,2004','2004-07-18 00:00:00','   New  urlLink Dave Chappelle  stand up special on Showtime in September.    Oh yeah,  urlLink Brandy\'s  marriage was fake!  Supposedly, her mother told them to say they were married, so she could maintain her wholesome marriage.  What wholesome image?  I never considered her wholesome in the least.   I guess only a woman can scope out the ho in wholesome....lol!  '),(16645,3148889,'18,July,2004','2004-07-18 00:00:00','   SONG OF THE DAY:  Let\'s Push Things Forward by The Streets  '),(16646,3148889,'18,July,2004','2004-07-18 00:00:00','   The great Coach  urlLink Eddie Robinson  of my alma mater Grambling State University is battling alzheimer\'s disease.   Mr. Robinson is a beautiful, wonderful and gracious man.  I am saddened that this disease has gotten ahold of this inspiring man.  He always greeted you with a large smile and was super-friendly.  I\'ll always appreciate the way he respected the Grambling students.  '),(16647,3148889,'18,July,2004','2004-07-18 00:00:00','   Why are some movie critics calling Napoleon Dynamite another version of Rushmore or Election?  It\'s nothing like Rushmore unless you\'re just going by that it takes place at a high school and there\'s an election.  The setup of ND is different from the others.  You probably won\'t grow to love or care about any of the characters, but you\'ll feel like your $10 was well spent.  I urge all who love fun quirky films to go check this one out.  In theatres nationwide on Friday, July 23!  '),(16648,3148889,'18,July,2004','2004-07-18 00:00:00','   urlLink Andre 3000  of Outkast was named Sexiest Vegetarian by PETA.  PETA sucks major ass for their extreme way of getting out the word on the mistreatment of animals but they got this one right.      PETA recently went after several well-known heavyset celebrities, like  urlLink Ruben Studdard  and Michael Moore by sending them care packages on how to lose weight.  I could never support an organization that would do such things.  If these celebrities get fat, it\'s not PETA\'s problem.  So, go back into your lettuce igloo and chew on some carrot sticks.  We, at FAT GIRL FILMS will be enjoying a giant steak with a side of fried pork sticks and a chicken shake...yum!  Oh yeah, for dessert, we\'ll be having an ostrich parfait....  '),(16649,3148889,'18,July,2004','2004-07-18 00:00:00','   urlLink Adam Sandler  has the best celebrity site on the internet.  Check out his tribute to his beloved dog Meatball.  '),(16650,3148889,'18,July,2004','2004-07-18 00:00:00','   Get to know  urlLink The Streets .  Awesome UK rappers!  '),(16651,3148889,'18,July,2004','2004-07-18 00:00:00','   The  urlLink Sci Fi Channel  got me so good the other night with this show on M. Night Shyamalan \"creative process\".......Turns out that it was a bit of a hoax....great hoax.....    In other news, I almost got hit by a train last night.  No lights came on or anything.  The train didn\'t make a noise until I had crossed the tracks.  I might have been 2 or 3 yards from the train.  I have to thank God for allowing me to escape that situation without harm.  Whew!    SONG OF THE DAY: Sunday Morning by Maroon 5  '),(16652,3148889,'18,July,2004','2004-07-18 00:00:00','   The  urlLink ESPY Awards  were so great tonight.  Jamie Foxx was hilarious and Serena looked awesome.  '),(16653,3148889,'18,July,2004','2004-07-18 00:00:00','   Meet  urlLink Joe User .  '),(16654,3148889,'18,July,2004','2004-07-18 00:00:00','   urlLink Queen Aishah  is one comedienne to watch.      SONG OF THE DAY: Watch Out for the Big Girl by Baltimore Club  '),(16655,3148889,'18,July,2004','2004-07-18 00:00:00','   Today is the big day!  Filming Tit For Tat in Edmond, OK.  Wish us lots of luck!  '),(16656,3148889,'18,July,2004','2004-07-18 00:00:00','   Let me sum up Napoleon Dynamite: Art imitating life....  '),(16657,3148889,'18,July,2004','2004-07-18 00:00:00','   I sold my mother\'s old upright piano for $177 on eBay today.  She\'s very excited about the sale and happy to get rid of it.  She\'s moving at the end of next month and doesn\'t want to haul it to her next residence.  It\'s  great piano and someone in Midland, TX just got a really great deal.  '),(16658,3148889,'18,July,2004','2004-07-18 00:00:00','   I\'ve been having some tech issues with my old Lexmark printer.  I\'ve done just about everything you can do to possibly fix the issue because I like to handle my own problems.  I finally gave in and contacted the tech people at lexmark.  They didn\'t know what the hell to do.  Then, I decide to ask around in some Yahoo! tech group with basically no help.  People just telling me what I already know and tried.  I put the damn thing on eBay last night and I\'m going to make the switch to Hewlett Packard.      I\'ve been doing pre-production stuff on my Tit For Tat short film.  I\'m doing most of the work myself, but I\'ve got some great volunteers helping me.  The only thing I have to is not f*ck it up!      I\'ve got a bit of a stomach ache.  I\'ll be off today, so I can just pamper myself to no end while prepping my script for shooting.      I thought I wanted to listen to some old school Off The Wall Michael Jackson, but after listening to a few tracks, I\'ve come to realize that I\'m not really a fan of Michael.  It\'s funny what media hype can do to a person.      Also, The Streets only has a couple of hot tracks.  He\'s not really a rapper, he basically talks over the tracks.  He doesn\'t ride the rhythm of the track and his wordplay skills aren\'t on point.  He might want to take a couple of classes at the School of Eminem.      SONG OF THE DAY: You Make Me High by Toni Braxton  '),(16659,3148889,'18,July,2004','2004-07-18 00:00:00','   SONG OF THE DAY: Wind of Change by Scorpions  '),(16660,3148889,'18,July,2004','2004-07-18 00:00:00','   Meet  urlLink Beverly Cheng and Johnny O\'Neal , they married on July 24. I stumbled upon Johnny\'s website while checking out a fellow triggerstreet member\'s website. I read Johnny\'s blog and read about his engagement to Beverly. It\'s a very sweet story and they really seem to vibe well.  '),(16661,3148889,'18,July,2004','2004-07-18 00:00:00','   It\'s late and I just finished watching Startup.com about two guys trying to get an internet company off the ground. I have learned that the co-founding member  urlLink Bryan Mundy  of their rival ezgov.com is dead. Incidently, ezgov.com is still going strong and govworks.com was bought out by someone and reincarnated as  urlLink govonesolutions.com . The documentary was awesome! Tom Herman has what looks to be a bi-racial child.  urlLink Kaleil Tuzman  looked like a total meathead but still you could tell that he was a cool and very intelligent guy. I felt bad for Kaleil\'s suffering girlfriend Dora. Dora made one point that most men don\'t get. Let\'s say you\'re running multi-million dollar internet start-up or even running Pizza Hut, if you\'re going to be busy for a few days.....let the girl know. Don\'t go about your life for a week without calling and expect for the girl to be happy to hear your voice. A couple of minutes out of 24 hours is not too much to ask, but men don\'t see that way. Ugh, I think I\'m having a moment..... The guys have gone on to start another company called  urlLink Recognition Group .  Tom has a small page on his family\'s company  urlLink website  about the whole govWorks, Startup.com and other things people might be interested to know about.    Btw, Tit for tat\'s filming was postponed until July 31....this is the final postponement because I can\'t take it anymore....    I\'ve got movies to make and by golly, they\'re going to get made!(doesn\'t that sound like a good slogan? I might have to use that....)  '),(16662,3148889,'18,July,2004','2004-07-18 00:00:00','   I was totally grooving on some old school Keith Sweat.  Anyone remember the now classic    urlLink Make It Last Forever  album from the late 1980s?  Get educated on some of the grooviest R & B music and my closest friends know this is my least favorite musical genre, so this must be some good sh*t!  '),(16663,3148889,'18,July,2004','2004-07-18 00:00:00','   Why do I always expect for my soulmate to be some lonely spaz until my vibrant spirit inhabits his world?  That\'s so silly for me to think because I would never want to be with someone like that.  I love people with personality, like moi.  Someone crazy and a bit off their rocker, because they would have to be in order to want to be with me.  '),(16664,3148889,'18,July,2004','2004-07-18 00:00:00','   urlLink Dodge Magnum  is my new dream ride of the week.  Last week, I was all about the  urlLink Chrysler 300 .  '),(16665,3148889,'18,July,2004','2004-07-18 00:00:00','   My favorite show on television is American Chopper on Discovery.  Take a gander at the  urlLink OCC  website....hot sh*t!  '),(16666,3148889,'18,July,2004','2004-07-18 00:00:00','   I\'m so excited that the new  urlLink Canon XL2 \'s are coming out, so I can get a better deal on a Canon XL1.  Yeah, it\'s pretty much all about moi....lol!  '),(16667,3148889,'18,July,2004','2004-07-18 00:00:00','   SONG OF THE DAY:  Alone in Kyoto by Air  I got a marriage proposal from some Indian dude via email, so he could move to the United States....lol!    He says he\'ll take care of me for the rest of my life and never cheat on me.  Hmmm.  Nah!  '),(16668,3148889,'18,July,2004','2004-07-18 00:00:00','   A cool female filmmaker.....Meet  urlLink Dina Mande .  The  Tit For Tat  shoot went well on Saturday.  I was running late as usual and it was a 30 drive to the set.  On top of everything, I was running on about 2 hours of sleep.  I felt good about all of the performances, I just wish I had more time for extra shots(closeups and such).  We didn\'t start filming until nearly 2pm and I had to be back in OKC at 4pm.  The film will be between 5 and 10 minutes in length.  The cast was fantastic and everyone got along so well.  I had a wonderful director of photography and lighting guy.  The total cost of production was around $60, which is basically little to no budget.  We\'ll be doing the editing next week in Tulsa.  '),(16669,3797536,'21,July,2004','2004-07-21 00:00:00','   Admittedly I can\'t quite tell whether people are&nbsp;commenting or not, this is due to the situation I explained in my last post... and such a sad situation it is.  But please... if you wish to comment please continue to do so... because I have a generous, loving, perfect&nbsp;minion that is gonna email me copies of all your wonderful comments be they sparse or in abundance.  *shines at the existence of Blessi*  Actually I haven\'t asked her yet :P  so, COMMENT... please... '),(16670,3797536,'20,July,2004','2004-07-20 00:00:00','   CONTROL!!!... It is absolutely terrible, and everybody is obsessed with it... everyone wants control. &nbsp; The perfect example is my goddam boarding school... I consider myself moderately&nbsp;lucky that I can get to this page to post on my blog... but I can\'t actually get to my blog, or anyone elses. &nbsp; *overcome with profound sadness* &nbsp; Oh, plus I can\'t access hotmail from here... those damn control freaks! It is ALL their fault. &nbsp; So I am terribly sorry but I can\'t read anyones comments for a little while, or anyone elses blogs. *snif snif*... &nbsp; BUT I WILL POST... just to defy the masses.... POWER TO&nbsp;THE INDIVIDUAL! '),(16671,3797536,'17,July,2004','2004-07-17 00:00:00','   My friends are so depressed I feel the question Of your loneliness Confide... `cause Iâ€™ll be on your side You know I will, you know I will &nbsp; X girlfriend called me up Alone and desperate On the prison phone They want... to give her 7 years For being sad &nbsp; Chorus : I love all of you Hurt by the cold So hard and lonely too When you donâ€™t know yourself &nbsp; My friends are so distressed And standing on The brink of emptiness No words... I know of to express This emptiness &nbsp; Chorus : &nbsp; Imagine me taught by tragedy Release is peaceI heard a little girl And what she said Was something beautiful To give... your love No matter what '),(16672,3797536,'15,July,2004','2004-07-15 00:00:00','   hello...   hmm writing it a second time makes it seem weird and kinda pointless.    &nbsp;  I began to scratch my arms \'till they bleed again. I think it must besome kind of nervous reaction. I suppose it is because I am not ready to see some of the people that I am about to see again after the schoolholidays... boarding school sux.   I hate myself for having trusted people, and&nbsp;what\'s is worst I still trust, and will most likely continue trust forever.   I told him about me in confidence... that was silly, he has betrayed that... he told them all about how I am a \"psycho\" and how I used to hurt myself. Then he tell\'s me to go kill myself... and that he is serious... but still expects me to treat him with respect afterwards.FUCK HIM. But I can\'t, I can\'t hate him for some reason... life is shit.    &nbsp;  maybe I should die... '),(16673,3797536,'15,July,2004','2004-07-15 00:00:00','   OMG, Ð¿Ð¾Ñ‡ÐµÐ¼Ñƒ Ñ Ð¿Ð¸ÑÐ°Ñ‚ÑŒ Ð½Ð° Ñ€ÑƒÑÑÐºÐ¾Ð¼ ÑÐ·Ñ‹ÐºÐµ? Ñ Ð¿Ð¾Ð»Ð°Ð³Ð°ÑŽ Ð¾Ð½ Ð¿Ð¾Ñ‚Ð¾Ð¼Ñƒ Ñ‡Ñ‚Ð¾ Ñ Ñ…Ð¾Ñ‡Ñƒ ÑÐºÐ°Ð·Ð°Ñ‚ÑŒ Ð»ÑŽÐ´ÑÐ¼ ÐºÐ°Ðº shit Ð²ÑÐµ, Ð½Ð¾ Ñ Ñ„Ð°ÐºÑ‚Ð¸Ñ‡ÐµÑÐº Ð½Ðµ Ñ…Ð¾Ñ‡Ñƒ Ð¸Ñ… Ð·Ð½Ð°Ñ‚ÑŒ... Ñ Ð¿Ð¾ÑÑ‚Ð¾Ð¼Ñƒ messed Ð²Ð²ÐµÑ€Ñ… Ð¿Ð¾ Ð¼Ð½Ðµ Ñ‚Ðµ ÑÑ‚Ð¾ Ð¼ÐµÑÑ‚Ð¾, Ð·ÐµÐ¼Ð»ÑŽ, ÐµÐ³Ð¾ Ð±ÑƒÐ´ÐµÑ‚ Ð½Ð°ÑÑ‚Ð¾Ð»ÑŒÐºÐ¾ ÑƒÐ¶Ð°ÑÐµÐ½.  GRR, this world or the next... that\'s what I wanna know... '),(16674,3797536,'14,July,2004','2004-07-14 00:00:00','   pooq .  MOOSIC '),(16675,3797536,'13,July,2004','2004-07-13 00:00:00','   I suppose I should start using my blog for its real purpose... a weblog... not just advertising the things that I love. Where do I begin?  I suppose, as they say: there\'s no time like the present.  Now is a funny time... I guess I am going through that everpresent phaze in society known as adolescence... I HATE my parents [not only that, but I am so typical, I even think that my hatred for them is justified... haha]. My friends continuously turn their backs on me, making it ever-so-hard to consider them friends... And not only that BUT, I think I am having and identity crisis... yesturday, someone thought I was a butterfy!! and before that someone thought I died of cancer 3 years ago... ... AND I simply don\'t know who I am.  ~seulement alors volontÃ© vous~ ~comprenez votre individu vrai~ '),(16676,3797536,'13,July,2004','2004-07-13 00:00:00','   amidst rush of blood and reptile tears we lie, i with no questions      you with no answers and each day we come, trade roles      change masks become the victim of the crime tracing subconscious designs on each other\'s skin till mandalas rub raw           and you leave without trust there can be no love when morning certainties       give way to noontime doubts where filament nerves       curl underneath flame and a burgeoning bruise      stains a faded memory, we\'ll be till nightfall comes, and we rise to die and kill and wake like silk once more ever making regenerating  ...Blessi... '),(16677,3797536,'11,July,2004','2004-07-11 00:00:00','   My trust, so spun-glass thin, so fragile, hold it gently, lest it break. I realise I trust you I trust them - I donâ€™t know how I canâ€™t think when. But this feeling- trepidation yet excitement closed in but free. So simple so hard. Tread softly, donâ€™t wake MY DREAM.  ...Sarah... '),(16678,3797536,'09,July,2004','2004-07-09 00:00:00','   urlLink '),(16679,3797536,'09,July,2004','2004-07-09 00:00:00','   urlLink '),(16680,3797536,'09,July,2004','2004-07-09 00:00:00','   urlLink '),(16681,3426176,'22,May,2004','2004-05-22 00:00:00','   I woke up around 11:00 today and wandered aimlessly around the house.  Now that school is out, there is really nothing to do.  I went to practice around 13:00.  That\'s 1:00 PM for those who don\'t speak European.  The weather was annoying because it was warm and sunny, then it would get cold and rainy, then warm and humid, then cold and cloudy, then drizzle, then pour, then sunny, then pour again, then sunny......  We played a fun little game where you had to shoot the ball off the crossbar and kick it in the goal on the volley.  I did it once and seriously screwed up my hip flexor.  Then we started making bets trying to hit various objects off of various posts from various angles from various distances.  We were just looking for stuff to do.  If more people showed up we could have scrimmaged. (Practice Wednesday at 13:00 on Fedora Field)  I was driving home and more crazy weather struck.  The skies opened up and I was driving through a wall of water.  I could not see a damn thing.  The streets were flooding and the windshield wipers weren\'t wiping fast enough.  I was driving under black clouds then suddenly everything was clear.  Blue skies, white fluffy clouds, and sunshine greeted me after I drove past watery hell behind me.  Random thought:  I really like those \"Fanta\" commercials.  The song is quite catchy.  Much to my dismay Xavier Malisse lost in the fourth round 7-5, 6-2, 7-6 (8-6) to 12th ranked Lleyton Hewitt in the French Open today.  He still can win some hardware for he advanced to the fourth round of the men\'s doubles championship.  The same goes for Virginia Ruano Pascual.  She lost in the fourth round, but advanced to the quarters of the women\'s doubles championship.  Another random thought:  I was just watching \"X-Play\", a show about videogames, and I just learned that a \"weinernut\" is a hot dog in the shape of a doughnut with the hole of the doughnut as a basin to hold cheese, relish, ketchup, and mustard.  Wow.  Just when I thought they made everything...  Now I must go.  I have alot to do tomorrow morning.  Bye.  '),(16682,3426176,'22,May,2004','2004-05-22 00:00:00','   I had today set up perfectly.  So I was up past midnight last night making the preceding post and it was close to game time between the Los Angeles Galaxy and the Dallas Burn.  I decided to stay up and watch the game, which I did.  (Quick recap: L.A. destroyed Dallas 4-1.  A player I don\'t like, Steve Jolley, had a game to forget.  He blocked a cross with his jollies [pun intended], headed the ball into his own goal, and deflected a Jovan Kirovski shot into his own goal.  Kirovski received credit for the goal, but it really should have been another own goal by Steve Jolley.  Sasha Victorine, Jolley (o.g.), and Kirovski [2 goals, one from the penalty spot] scored for Los Angeles and Ronnie O\' Brien scored for Dallas.) Staying up past 02:30 reminded me of when I woke up at 01:30 to watch the 2002 World Cup everyday.  I would stay awake most of the time and watch the games while everyone else in the house was sound asleep.  I remember this one time where I was up from 01:30 and watched all three games (including the game where Italy dominated Equador 2-0) and fell back asleep around 08:30.  I woke up when my mom came home from work at 14:30.  That\'s 2:30 PM for those of you who don\'t speak European.  I know some of you are saying, \"2:30 PM?  I wake up at that time every day!\"  Well, I don\'t.  If I could I would.  It\'s just that my mum usually wakes me up around 09:30 (for no good reason) and if she doesn\'t I can\'t sleep past 10:30.  Yes, everyone laughed at me and ridiculed me for sacrificing much needed sleep for the beautiful game.  I swear, that whole month I had to have gotten a grand total of about 15 hours of sleep.  Most of it during the day.  I was going to sleep late today, but I figured that since I didn\'t get to watch the Fire game I could watch the replay online.  Well they weren\'t showing it.  They weren\'t even showing the highlights.  I was not happy.  I woke up at 08:00 for nothing then.  I decided to stay awake and watch tennis so it wasn\'t so bad.  Xavier Malisse advanced to the round of 16 while Virginia Ruano Pascual lost in the third round. (or was it the fourth round?)  So with my Fire game watching hopes dashed, I waited for the Indy 500.  It was raining over there.  Damn, now I have to wait out a super long rain delay.  They got the race started about 2 hours late and ran about 25 laps when the rain started again.  Bloody hell.  I just have no luck.  They were close to getting it back under way, but I had to go with my mom and dad to my grandmum\'s house (the good grandmum) for a while.  After that we went shopping and I got some Chinese Fooooooood (and then?) for myself and went home.  By the time we were home, the Indy 500 was well over.  Buddy Rice won, which meant that David Letterman (yes the talk show host) won his first Indy 500 as an owner.  Quick Rant:  For some reason, my mum and dad have been taking me along on their endless shopping excursions and when they go they play this  Ultimate Manilow  CD.  Now, it takes alot for music to aggravate me and I was OK with it... FOR THE FIRST 40 FRIGGIN\' TIMES!!!  Bloody Hell!  They play this every single waking moment of the day.  And it\'s all of that sappy love song crap, which I can remotely get into on very rare occasions, but not every damn day.  Jesus H. Christ.  So we got home and the Coca-Cola 600 was already past half way done.  I didn\'t miss much considering Jimmie Johnson was annihilating the whole field.  So I watched the end of the race while I munched on Moo Goo Gai Pan and my Egg Rolls. (why does that sound dirty?) Jimmie Johnson deservedly won the race.  He made this sick nasty pass where he slipped between a lapped car and Kasey Kahne in the middle of a turn.  He must drive on the expressway alot.  I think I\'m getting sick again so I better go and get some sleep.  I really need it.  I have practice tomorrow at 13:00 on Fedora Field.  That\'s 1:00 PM for those of you who don\'t speak European.  And Fedora Field is what Pat and I named the practice field.  I want to call the stadium Spellman Stadium, but Pat doesn\'t like it.  I do.  It\'s an honour to have the stadium named after you.  More honourable than a practice field.  Anyway, off I go.  Bye.  '),(16683,3426176,'22,May,2004','2004-05-22 00:00:00','   Dear Christ.  No, the title is not wrong.  I\'ve been out for 7 hours at a bunch of stores shopping with my mom and dad.  I was shopping for a multitude of necessities for my trip to Italy.  I got shorts, a few shirts, a toiletries pouch and some chips, but that was for the seemingly endless car rides to a new store.  We also searched for gifts for my exchange student and his family.  I think we did well.  The only thing is that when you are around my mum for a long time, she tends to get... how do I say it... oh yeah, annoying.  It\'s not bad, it\'s just... annoying.  It\'s hard to explain, but she just endlessly blabbers on about stuff.  Oh well, it\'s all good.  I would rather have that than if she was silent all the time.  I also missed watching the Fire game online.  Because I\'m a cheap bastard, I don\'t subscribe to get all of the MLS games on satelite so I have to watch the home Fire games on my computer that has a shaky on-line connection.  I guess I will wake up early tomorrow and watch the replay online.  I will be going to bed late because the Los Angeles Galaxy v. Dallas Burn match is on tonight at 00:30.  That\'s 12:30 AM if you don\'t speak European.  Ugh, it is going to be a late night.  One more thing.  My dad went to the currency exchange and he exchanged $300 and we got 240 Euros back.  The Euros are cool b/c they are different colours and different sized depending on the bill.  For example, the 5 Euro bills are green and roughly 6.5 x 12 cm.  The 10 Euro bills are red and roughly 7 x 12.5 cm.  And the 20 Euro bills are blue and roughly 7.5 x 13 cm.  I like how they are different colours and different sizes.  That\'s just real cool.  And I forget that it is money and treated as such because I am so used to seeing dollar bills.  These just seem like coloured play money.  I almost lost it a few times.  That certainly would suck.  OK, off I go.  Big day tomorrow... OK, today.  It just turned midnight.  I have to watch Los Angeles v. Dallas, Fire v. Colorado, Indy 500, and Coca-Cola 600 tomorrow.  Someone has to do it.  It may as well be me.  Bye.  '),(16684,3426176,'22,May,2004','2004-05-22 00:00:00','   My summer vacation has officially begun as of 9:30.  My Spanish final was difficult, but I think I did fairly well.  Some of the questions were confusing and the reading part took me a long time so I had to rush through the other questions a lot more than I had wanted to.  But all in all it was pretty good.  So that was it.  The 2003/2004 school year was over.  Arrivederci Junior year.  One more left.  Just one more year until... I don\'t know.  Summer break started the best way possible.  After I shamelessly promoted my blog, I walked out to my car to go home and do nothing.  Just then who do I see but Steve!  He needed a ride so I was more than happy to oblige.  He mentioned going to Mickey D\'s for breakfast and I figured why the hell not?  So we went, I had the Steak, Egg, and Cheese Bagel and he had the Two Breakfast Burritos.  When we left we saw Howie and Scott walk in.  We greeted them and Howie and I exchanged \"man-hugs\".  We went back to Steve\'s place where I watched him get the monkey feces beat out of him in  Ninja Gaiden  and then I joined him in playing an amusing round of  Halo.   It was hilarious when we were on the same team, but Steve kept killing me on purpose.  That\'s what friends are for.    Visit Steve\'s blog here:    urlLink Bubbadoo4you    Eventually I had to go home.  When I got home I was watching an interview of Andrew W.K. on Conan O\'Brien.  Mr. W.K. has a genius plan of gaining weight that I have to try.  When he was a wee lad, his situation was similar to mine.  While everyone is trying to lose weight, we are trying to gain weight.  I won\'t lie to y\'all.  I am freakishly skinny.  I am 5\'9\" and 113 lbs.  That\'s not too good.  So that\'s why I might do this strange diet plan.  It\'s the Sweet and Salty diet plan.  First, eat a lot of sweet stuff like candy and donuts and cake and stuff like that.  When you are full, or tired of sweets, eat salty stuff like potato chips and french fries.  When you are full of salty stuff, go back to sweets and just continue until you gain at least 20 pounds.  Just eat a helluva lot of unhealthy crap and you\'ll be set.  Well, that\'s all for now.  I\'m looking forward to this weekend.  The Fire play at home against Colorado, and the Indy 500 and Coca-Cola 600 are on Sunday.  Yes, I\'m hoping for a Fire win and two great races.  Talk to y\'all tomorrow.  Bye. urlLink  '),(16685,3426176,'22,May,2004','2004-05-22 00:00:00','   Tomorrow is it.  The last final for me.  The last official school work as a Junior.  The last test until August.  But tomorrow will not be an easy task.  The final final is Spanish.  Strangely enough I feel very confident.  Spanish has been extremely hard this semester, but in the recent weeks I have been doing very well.  That momentum will carry over into tomorrow\'s test.  I will destroy Spanish.  I fear nothing.  I have to relinquish my title of Iron Chef.  I was cooking a frozen pizza in the oven and I undercooked it.  It turned out terrible.  I hang my head in shame.  It may be a surprise to many of you, but one of my favourite sports is tennis.  I was watching The French Open coverage today and I was pleased to see some of my favourite players advance.  I like the underdogs.  Players no one knows like Xavier Malisse from the men\'s side and Virginia Ruano Pascual from the women\'s braket.  Malisse made an incredible run to the Wimbledon quarterfinals a year or two ago and Ruano Pascual plays doubles with partner Paola Suarez.  Ruano Pascual and Suarez are ranked first in the world in women\'s doubles.  Malisse won his third round match against Daniel Elsner and advanced with a 6-3, 4-0 (retired) win.  The Belgian will take on Albert Costa in the quarterfinals.  Malisse had to play double duty as he and Olivier Rochus advanced past Todd Martin and Mikhail Youzhny 6-3, 2-6, 6-3.  Ruano Pascual won her third round match against 12th ranked Ai Sugiyama 6-7 (4-7), 6-2, 6-1.  Ruano Pascual will face Francesca Schiavone in the quarterfinals.  Ruano Pascual also had to play two matches as she and Suarez easily advanced past Gulnara Fattakhetdinova and Galina Fokina 6-1, 6-3.  Now I must go to bed.  I\'m tired and I have a huge final tomorrow.  The end is nigh.  '),(16686,3426176,'22,May,2004','2004-05-22 00:00:00','   OK, before I get to the important stuff I want to recap a post I did not make yesterday.  I had the day off, did nothing, and did not study for any finals.  Today, I had a U.S. History final (it sucked massively) and a Physics final. (it wasn\'t too bad)  Tomorrow I have a Religion final and an English final.  Again, I didn\'t study.  OK, now that the boring crap is out of the way, it is time for what really matters.  CHAMPIONS LEAGUE FINAL!!!  \'Twas a great final.  It was between France\'s AS Monaco and Portugal\'s FC Porto.  These two teams defyed all odds as they are not the heavyweights of European football.  (That\'s soccer if you don\'t speak European)    AS Monaco were the lucky dogs of the competition as they beat Germany\'s Bayern Munich 2-2 on aggregate, (away goals 1-0) shocked the world with a stunning come from behind win against Spain\'s supergiants Real Madrid 5-5 on aggregate, (away goals 3-2) and eased past Chelsea 5-3 on aggregate to lead them to the final.  FC Porto reached the final when Benni McCarthy sunk England\'s supergiants Manchester United with a last minute goal.  Porto won the series 3-2 on aggregate.  Porto then got past France\'s Olympique Lyonnais 4-2 on aggregate.  Porto would play Spain\'s Deportivo La Coruna for 180 minutes and manage only one goal, (and that was from the penalty spot) but that was all they needed as they advanced to the final 1-0 on aggregate.  Then today\'s final...  Monaco ended up playing a dreadful match and gave the cup to FC Porto three goals to nil.  If you want, you can read my review here.    urlLink FC Porto : AS Monaco    Coach Didier Deschamps consoles his players. Someone has to lose.    Coach Jose Mourinho leaves FC Porto as European Champions.    2004 UEFA Champions League Winners: FC Porto    The Joy of Victory.    The Agony of Defeat.  '),(16687,3426176,'22,May,2004','2004-05-22 00:00:00','   Today was the last full day of school so that meant that today was the last lunch period.  Over the past months many good times have been shared with my friends Greg, Kim, and Gina (and Anna for the first semester) and many songs have been sung.  Here are some of the best moments of this lunch period:  I remember dubbing Gina with the name \"Tubesock\" after trying to explain to her how a tubesock was different than a normal sock.    I remember explaining to her that Holland is a country, not in Poland and their culture consists of wooden shoes, windmills, those funny windmill cookies and last names that have lots of vowels and start with \"van\" or \"van der\". (Ex: van Nistelrooy, van der Vaart, van Hooijdonk, van der Saar)    I remember singing the Zoobalie Zoo song with my mouth full of cookies and sounding like a retarded person.  I remember singing the Goldfishes song.  I remember singing Jimmy Crack Corn and I Don\'t Care at random.  I remember going with Gina to buy chips or pop or candy for no particular reason except just to go with her.  I remember Gina and Kim teaming up on me trying to kill me by tickling me to death and trying to get revenge by poking Gina\'s ticklish sides.  I remember discovering the fact that spoon backwards is \"noops\" and Zoop backwards is \"pooz\" and Zoobalie Zoo backwards is \"ooz eilabooz\".  Yes, lunch was great this year indeed.  I\'m looking forward to next year already.  I took a picture of us at our last lunch today.  Maybe I will post that eventually.  '),(16688,3426176,'22,May,2004','2004-05-22 00:00:00','   I do not want to make a post.  I am in an awful mood.  The Chicago Fire lost to the Columbus Crew 3-1 at home on national TV today.  It was the most abhorrent display of soccer I have ever seen.  I don\'t want to talk about it.  I don\'t want to talk about anything else for that matter.  Go away.  '),(16689,3426176,'22,May,2004','2004-05-22 00:00:00','   The theme of Saturday (right now it is early Sunday) was sports.  Sure there was some wicked weather and that may be a tad important considering how dangerous it was, but really Saturday\'s sports transcended all of that for me.    First I woke up at 9:00 to watch the Live Commentary of the FA Cup final match between Manchester United and Millwall.  If you don\'t follow what the English call \"football\" Manchester United is one of the best teams in the English Premiership or all of Europe for that matter.  Millwall is a midtable team in England\'s first division (one division lower and the Premiership) who was lucky to make it past the fourth round let alone all the way to the finals.  Anyway, Man U had an easy time winning 3-0.  American goalkeeper Tim Howard became the first American to play in a FA Cup final since John Harkes did for Sheffield Wednesday in 1993 and first American to win the FA Cup ever.  Congrats to Timmy Howard.  He won the FA Cup, won Premiership Goalkeeper of the year honours, had a 0.97 Goals Against Average, got a bunch of clean sheets, plays for one of the best clubs around the world... but only managed to beat my Chicago Fire once out of 10 games when he played for the Metrostars!! Take that!  Then the girls soccer playoff game.  They played the hardest I have ever seen them play.  I\'ll be honest ladies, not necessarily the best, but certainly the hardest.  The game was 0-0 for most of the match, but in the final 15 minutes Andrew managed to out score us 2-1.  It was a great season.  I enjoyed going to your home games (and some away games too) no matter if you won or lost.  Next was the Cubs 7-1 win over the Cardinals.  I didn\'t watch the game because I was at the girls soccer game.  Hey, I thought \"watch the Cubs or watch the girls soccer game.  What should I do?\"  I chose the latter and I made the right choice.  When I got home from the game, I watched the New England Revolution v. Metrostars match.  It was a well played match as New England really dominated the second half.  Pat Noonan scored for the Revolution and Cornell Glen scored for the Metrostars.  I gave my legs a wicked workout as I used this exercise machine in my house.  You sit on a seat and push it back with your legs while two elastic bands resist the seat.  I was on it for over 30 minutes and did over 1200 reps and I can still feel it.  Next was the NASCAR All-Star Race.  When the Nextel Open got the green flag, there was a pile up before the first turn of the first lap.  Hey guys, drive much?  In the main event, there was another wreck when Kurt Busch ran into Greg Biffle and after that all hell broke loose.  My favourite driver Tony Stewart won the first segment, but finished 3rd.  Matt Kenseth won.  Finally, I watched the San Jose Earthquakes v. Los Angeles Galaxy match online.  To put it lightly, these two teams HATE each other.  Last year in the playoffs, Los Angeles had a 4-0 aggregate lead with about a half hour left in the second leg and San Jose came back to score five times and advance. (It was the greatest MLS game ever) Today, San Jose got out to a 3-0 lead and when they thought it was safe, Los Angeles pulled two back to make it 3-2 courtesy of Andreas Herzog and Joseph Ngwenya. (Herzog has got to be one of the best passers I have ever seen) But a monumental comeback was not to be as Dwayne DeRosario put the game away for good with a strike of his own to seal the win for the \'Quakes 4-2.  Oh yeah, the Tampa Bay Lightning beat the Philadelphia Flyers 2-1 in Game 7 of the NHL Eastern Conference playoffs.  Tomorrow I will be glued to the tube again as my Chicago Fire play the Columbus Crew at Soldier Field on ESPN 2 at 15:00.  That\'s 3 pm if you don\'t speak European.  \'Till then, see ya and go Fire.  Oh yeah, check out my good friend Stephen\'s blog  urlLink Bubbadoo\'s Place  '),(16690,3426176,'22,May,2004','2004-05-22 00:00:00','   Ugh.  I feel like feces.  I most definitely did not want to go to school.  I was dreading the fact that I would not only go to school, but to soccer as well which meant I would probably not get home until after 18:00.  That\'s 6 pm if you don\'t speak European.  So I went to school and fortunately it is the end of the year, which means work is at a minimum.  I did have to take a Spanish test however.  It wasn\'t very hard for a change.  Math was certainly fun though.  All I did was grade Geometry papers.  When school was over I went home and took an hour and a half nap.  Yes, life is good.  One more thing about yesterday and it\'s a good thing.  I went to the girl\'s soccer game and was treated to a 5-0 Marian win over Crete-Monee.  It was a well played game all around.  I was impressed with the goals we scored.  Hard work, good teamwork, and a deft shot gave us a well deserved five-nil rout against an archrival.  It was a good result considering they bounced back from a disappointing 5-0 loss to Homewood-Flossmoor.  Well done.  Good luck in the playoffs.  '),(16691,3426176,'22,May,2004','2004-05-22 00:00:00','   Today I had to take the last half of my trigonometry final.  Usually completing a final would be something to be excited about, but I knew that when I finished the exam and the bell rang I would never see most of the seniors ever again, for today was the senior\'s last day.   It was strange to see so many empty seats in physics class.  Most of the seniors didn\'t have to take the final exam.  A few did though, and for them it was just another physics class period.  Talking about the Sox, talking about movies, talking about college, talking about nothing in general, and making jokes was all that happened.  Oh yeah, and something about a final, but no one really paid any attention to that.  Yup, same old, same old.   The rest of the day went on as planned until 7th period ended.  I was walking back to my locker where I saw my good friend and lab partner Michele.  I knew this would be the last time I would ever see her so I made sure I gave her a nice good-bye.  She summed it up really well when she said that she would most likely never see me again and she wished me a good life.  It was a hurried good-bye because she was going one way and I was headed another.  I guess that could be perceived as a nice symbolic ending to a friendship.   So there I was in trigonometry class.  I was ready to finish the final, but not the class.  Over the past months, some of the best times of the year were made in that class.  The best part about it was I contributed to the good times.  Usually I\'m just an observer who watches the good times, but I made many of them happen in this class.  I was known as \"weird kid\".  Never had I been known for anything in any other class.  It really felt good to open up and be noticed.   So there I was at my desk working hard on the final.  With about five problems to go I looked at my watch and it read 14:10.  That\'s 2:10 pm if you don\'t speak European.  So with only 10 minutes left a thought came to me.  In 10 minutes I will never see most of the people in this class ever again.  It\'s a harsh reality when you see them everyday for nine months and suddenly they are gone.  In fact, I\'ve known some of them two or even three years.  It is naturally hard to say good-bye to so many people you have come to know and grow close to for so long.  It will be even harder next year when I have to say good-bye to everyone I know, good friends and all, not just a select group of people that I knew from a few classes.  But as I like to think, the hard times in one\'s life can not be stopped.  They will come.  It is how you handle them and get past them that makes you who you are and makes you better.  When the bell eventually rang and I turned in my test and I went back to my desk and just stood there and took one last look at everyone for the last time.  I guess this really was it.  I wanted to say bye to everyone, but most of everyone was still taking the test so I didn\'t.  I wish I did.  I did give Bux a good handshake and \"man-hug\" so I felt good that I at least did that.   That was that.  Good-bye to every one who made going to class a lot more enjoyable.  Good-bye to everyone who made homework, tests, and grades a little less of a burden.  Good-bye to Kate, Michele, Jon, Kevin, Bux, Steve, Amy, Kris, Andy, Missy, Blondie, Chris, Chuckie, Rocky, Brandon, Bond, Dana, Magpie, Gina, and everyone else I shamefully forgot to mention.  However we knew each other, whether it was from soccer, to physics, to math, to other subjects or activities, thanks to you all.  I don\'t know if you feel the same about me, but you all made life a little bit better for me for that short time we knew each other.  Thanks a lot and good luck to you all.   Now I have that damn Green Day song in my head.  '),(16692,3426176,'22,May,2004','2004-05-22 00:00:00','   Hello everybody.  This is my first post in my brand new blog.  I am new to this so it will take me a while to find new cool additions to post on my blog, but for the meantime, what you see is what you get.  I will post stuff about whatever I want.  Rants, random observations, news, personal stuff, and anything else in my slightly deranged mind.  I will try and post pictures as well.  I hope you will enjoy it.  Tell your friends about it too.  '),(16693,3426176,'22,May,2004','2004-05-22 00:00:00','   Yes it is me and I am back from beautiful Italy!  It was the best time of my life and the most beautiful place I have ever seen.  I would make an epic post about how I loved Italy, but I won\'t.  I will make my daily post and at the end write about something that happened one day of the trip.  I will start tomorrow at the first day and post every day\'s events until I reach the end.  As for today\'s events, I woke up at 10:00 after sleeping 13 hours.  My room is the messiest room ever witnessed by mankind.  It\'s terrible.  Yeah, I guess I have to clean my own room again.  I wish I was in Italy.  I had practice today and there was only three of us there so there was nothing to do.  We shot at goal and stuff.  Then we ran by the cheerleaders.  That was my idea of course.  Then we worked on trapping and played some poker and went home.  I went to Chicago Dough for some pizza.  It felt great to have American pizza again.  Pizza in Italy is very plain.  The  Margherita  pizza is a thin layer of dough with crushed tomato sauce and cheese.  That\'s it.  So naturally I ordered a pizza today with everything on it.  Mmmmm... heavenly!  Also my favourite waitress took our order.  She was the waitress for my wild 16th b-day party.  That was a great time so whenever we see each other we are always nice to each other.  I also had McDonald\'s today.  It felt great to have a greasy, fatty, unhealthy double cheeseburger with super salty fries.  It was glorious.  That\'s all for now because I am so very tired.  More tomorrow.  '),(16694,3426176,'22,May,2004','2004-05-22 00:00:00','   Hello from beautiful Italy!  The weather is great over here.  So far.  The days of 100Â° weather are soon to come.  I have to post real quick b/c I have to get up at 3:30 to see the Pope tomorrow.  I wrote a journal on the plane ride.  Its moderately interesting.  When I have time, I might post something about it.  I love this country.  On newspapers, on TV, on signs on walls, everywhere is stuff about Italy in Euro 2004.  Mmmmmmm... football.  Ain\'t it great?  Yes it is.  The luggage situation is bad however.  My mum and dad\'s stuff arrived, but the suitcase full of gifts and my suitcase did not.  As long as my soccer boots are OK and I get most of my new clothes back, I should be moderately happy.  Ok, now I\'m off to go and go to bed.  In three hours I have to get up and see the Pope.  Ciao.  '),(16695,3426176,'22,May,2004','2004-05-22 00:00:00','   Today will be the longest day.  It just past midnight so it is Sunday, June 13.  In 16 hours, I will be on a school bus with a bunch of other hooligans (cough:: Pohlman ::cough) on our way to O\'Hare airport to catch a plane to London, England to catch a plane to Rome, Italy to catch a bus to San Benedetto del Tronto, Italy.  After the plane rides, layovers, and bus rides, I will be one tired American.  It all starts around 15:30.  That\'s 3:30 PM if you don\'t speak European.  Everyone who is going meets up in the Bloom parking lot to put their luggage and stuff on a van and board a bus that is going to O\'Hare airport.  Once we get to O\'Hare, we get our stuff and board the 20:20 plane.  That is 8:20 PM if you don\'t speak European.  I will arrive in London, England at 10:20 local time (that is a 7 hour plane ride) and board the 14:05 plane to Rome, Italy. (7:05 local time)  That is 2:05 PM if you don\'t speak European.  I will arrive in Rome at 17:30. (10:30 AM local time)  That is 5:30 PM if you don\'t speak European.  After I arrive in Rome, I will get on another school bus with everyone else and ride to San Benedetto del Tronto, meet up with the Quevedos, the family I will stay with for the next two weeks, and go home.  All that is today.  Yesterday, I got everything all packed up.  Clothes, gifts, reading material, snacks, music, shoes, everything.  It\'s all packed up and ready to go.  I am so excited I can\'t even type.  I was out all day yesterday looking for a Chicago Fire t-shirt and I came across a very troubling problem: NO STORES CARRY CHICAGO FIRE MERCHANDISE!!!!  Damn you all!!!  I could go on a huge rant, but I have a lot more to post and I need to get to bed soon.  As for Friday night\'s late night soccer?  Tremendous fun.  I had such a great time.  We played on a dimly lit tennis court.  We used one side of the court and set up two mini goals on either end.  Towards the end, lightning started to light the sky.  We played a bunch of 3 v. 3 games with one substitute.  I played moderately well considering my hip flexor was messed up pretty badly.  I got a goal, and a few good passes, and dug hard to win the ball and fight off a defender.  I also got kicked in the face.  Ahhhhh, the beautiful game.  How I love it.  I hope we do late night soccer again.  My Fire lost 3-2 to the Los Angeles Galaxy on Saturday afternoon.  It was good to see Ante Razov get back on the goal scoring track again.  He even scored a penalty kick on the same goal that Pat Onstad saved his penalty kick in MLS Cup 2003.  Ante is back, however, we do miss Armas and Beasley and Ralph and Williams so very desperately.  Let\'s hope Armas and Beasley make us proud and beat the living feces out of Grenada today.  You gotta love this:  today\'s strategy for the United States World Cup Qualifying match against Grenada?  \"Go out there and annihilate them.\" and \"Score as many goals as you can.\"  This should be fun.  Well, off to bed I go.  The next time you hear from me, I will be in beautiful Italy.  Ciao amici.  '),(16696,3426176,'22,May,2004','2004-05-22 00:00:00','   Tonight, something monumental, something of epic porportions, something so crazy, it could only have been thought of by me will happen. (if everyone shows up)  Midnight Soccer.  Yes, we will gather and play until twilight.  Until it is high moon (clever isn\'t it?) we will grace the pitch in the pitch black.  It actually starts around 9, but it will end at midnight.  I will write about it most likely tomorrow.  Today\'s practice was anything but epic.  Four of us showed up.  We played the crossbar game for about a half hour and played cheap baseball for the rest of the time.  My mum I got a lot of stuff done for Italy today.  I got super sexy shirts that will have the Italian ladies on me like white on rice.  We also got cameras and an Italian dictionary and other stuff like that.  Tomorrow is Italy Eve.  Man, I am just so excited.  I can\'t think right now.  The Chicago Fire return to the place where the dream season ended in a nightmare.  They go back to the Home Depot Center to face the Los Angeles Galaxy live on ESPN 2 at 15:00.  That\'s 3:00 PM if you don\'t speak European.  Last year at the Home Depot Center, the Fire lost to the San Jose Earthquakes 4-2 in MLS Cup 2003.  Let\'s hope the Fire can overcome this field of nightmares.  '),(16697,3426176,'22,May,2004','2004-05-22 00:00:00','   Finland and other Scandinavian countries have been known for their unique death metal style. One part thrashy guitar, one part harmonious guitar, one part brutal drums, one part thunderous bass, one part melodic keyboard, and one part scratchy vocals all come together to make a style that is greatly under-appreciated.    Kalmah , hailing from Finland, is no different. They follow this formula, but add their own self-dubbed \"swampy\" touch. Their first album,  Swamplord  exhibits their masterful musicianship. Antti Kokko\'s melodic guitar is first and foremost the highlight of this band. Many times, Antti\'s guitar licks are somewhat soothing while his brother Pekka\'s rhythm guitar blugeon the ears. This dynamic of grace and rage works astoundingly well.    Swamplord  starts off with \"Evil in You\", a speedy number that never lets up. The highlight of this song is towards the end where Antti\'s lead guitar repeats the most beautiful harmony on the album. At the very end, there is an interesting twist as a funeral organ closes out the song.   \"Withering Away\" is a song with a wonderful intro and a speedy tempo. This may, however, be the weakest song on the album.   Pasi Hiltula\'s keyboards and Antti Kokko\'s guitar shine on \"Heritance of Berija\". Everything on this track from the first note to the last note is simply majestic. Towards the beginning, there is a keyboard line that sounds like a violin. This touch to the song is spot on. Antti Kokko\'s leads are absolutely superb. There is another keyboard solo in the middle of the song that bridges the verse to the guitar solo. This guitar solo is the best on the whole CD. In the second half of the solo, the keyboard and the lead guitar harmonise creating a fantastic dynamic.   The atmosphere is heightened in \"Black Roija\" as the keyboard, Antti Kokko\'s lead guitar, and Pekka Kokko\'s rhythm guitar blend to add a mysterious and swampy feeling. The guitar solo has the feeling of a cold, dark alley in the summer night. This is yet another strong track.   The single to  Swamplord  is \"Dance of the Water\". This song opens with a eerie, cold keyboard intro. Ominous guitars enter and transition into another fabulous Antti Kokko guitar solo. The verse and chorus is different from the other songs as there are no harmonies or melodies. The lyrics portray an interesting picture to those who have the lyrics or those who can somehow understand the vocals. There is another excellent solo towards the end, which leads to a great harmony to close out the song.   \"Hades\" starts out with a lead that sounds like it came from a pirate movie for some reason. After the momentary break in the song, there is crushing drums and rhythm guitar, but a delicate little melody from the lead guitar and keyboard. The lead guitar offers many wonderful sharp screeches throughout the song that will kill little doggies. The overall atmosphere is somewhat laid back. This is the most evident in the instrumental break where one of the catchiest melodies lie. It\'s off to a full speed solo after that which transitions back into the catchy melody.   A keyboard intro that sounds like an instrument I have never heard opens \"Alteration\". There is another superb melody in the beginning of the song. The best part of this song is the instrumental break that takes one melody and repeats it in different ways that eventually leads into a great solo. The general atmosphere of the instrumental break is outright supernatural. Yet another wonderful song.   \"Using the Word\" starts out immediately after \"Alteration\". At first I thought they were one long song, but much to my surprise, they were two. The main lead gives a strong, triumphant feel. The keyboard adds in a harp melody that adds just the right flavour to this track. Another crafty instrumental break as both Kokko brothers offer a solo. Both are wonderfully done and add just the right atmosphere to the song. Great use of the whammy bar is also included. The track finishes strong and Pekka gives a nice ending to the album with a great raspy yell... then spits into the microphone.   However, there are some critical downsides in the production department. First of all, the bass drum can hardly be heard. The double bass drumming is great in \"Evil in You\", but it takes a lot of effort to hear it. Another problem is the keyboards are unnoticeable.  Kalmah \'s music concentrates on the atmosphere and mood and the keyboard provides most of it. However, it is missing for large portions in most songs. It is only noticeable in the solos and in parts of \"Evil in You\" and \"Using the Word\", but just barely. And finally, the bass is non-existent. It simply cannot be heard in any of the songs, except for the intro of \"Evil in You\" and the interlude in \"Hades\".    Swampsong , musically, is a spectacular mix of driving rhythm guitar with elegant lead guitar. The mood and feeling is a concentrated theme that this band has to offer. If  Kalmah  can improve their production, perfection will be within reach.    Kalmah  -  Swamplord   Spikefarm (2001)  8 tracks (36:12)      Pekka Kokko: Vocals, Rhythm Guitar  Antti Kokko: Lead Guitar  Petri Sankala: Drums  Altti Vetelainen: Bass  Pasi Hiltula: Keyboards    Immortal Songs:  \"Evil in You\", \"Heritance of Berija\", \"Alteration\"   92 out of 100  '),(16698,3426176,'22,May,2004','2004-05-22 00:00:00','   Ok, here\'s what happened:  I went to practice about 20 minutes late.  I had a few great bullet shots.  We played 5 v. 4 and I was on the shorthanded team again.  I gave an average performance, but managed to score a goal.  We played World Cup and I did OK.  I finally ate something at 16:00.  That\'s 4:00 PM if you don\'t speak European.  I had to run back out to an Italian meeting where I paid more money for the trips for my mum and dad.  I came home and ate my first meal (Big Mac meal) at 20:00.  That\'s 8:00 if you don\'t speak European.  I\'m tired.  That\'s it.  Bye.  '),(16699,3426176,'22,May,2004','2004-05-22 00:00:00','   Yesterday, (Tuesday) was a very busy day.  I was up at 08:00 when my mum came into my room and opened my windows and rearranged the stuff in my room.  That was a rather unpleasant way to get up, but I\'m used to it.  All day after that we did stuff like decide what clothes to pack, gather stuff to take and argue about pointless stuff.  Our house has no air conditioning so I had to run out and get my dad from work and we were outfrom 17:30 to 20:00.  That\'s 5:30 to 8:00 PM if you don\'t speak European.  Soon after I ran back out with my mum to look for more clothes to take with us.  I didn\'t get back until 23:30.  That\'s 11:30 PM if you don\'t speak European.  That was it.  That was my day.  If I have time, I will make my next album review.  I will review  Kalmah \'s  Swamplord .  '),(16700,3426176,'22,May,2004','2004-05-22 00:00:00','   Today at Fedora Field, we played the crossbar game.  Not a lot of people made it out today so we played this time waster game.    The rules are:  1. Shoot from outside the 18-yard box.  2. Hit the crossbar. (hence the  Crossbar  Game)  3. Shoot the ball into the net.  4. Volleys are worth 2 points, one bouncers are worth one.  5. First to three wins cash money.  This game is quite difficult.  The hardest part is hitting the crossbar and hitting in a way that it bounces back to you and not straight down or not over the goal.  I was very unsuccessful for the first hour, but after some good coaching from Pat, I got the hang of it rather quickly and almost got two points.  Blasted headers.  If I could have volleyed it I would have had a sure two points.  Oh well, I was shut out.  Can\'t wait \'till the next practice on Wednesday.  '),(16701,3426176,'22,May,2004','2004-05-22 00:00:00','   I am good.  I am damn good.  Usually I can\'t do two things at once, unless it is music related, but today, I managed to watch three entertaining sporting events at once and not miss anything.  First I woke up to watch the Men\'s French Open Final.  It was Gaston Gaudio v. Guillermo Coria for the coupe.  I started watching just after the third set.  Coria was two games away from the championship but Gaudio won three straight games to force a fourth set.  During the first game of the fourth set, Coria injured his left leg.  It was so bad he couldn\'t even run.  He just hobbled around trying to catch every ball that was hit to him, but couldn\'t catch most of them.  Coria gave away the fourth set to Gaudio.  The time it took Gaudio to win the fourth set was enough for Coria to recover from his injury.  He came back in the fifth set to reach championship point  three times .  However, Gaudio came back, won the game and tied up the fifth set 6-6.    It was time for breakfast so I had to leave this riveting match to eat an omlette.  When I got back, I turned on the telly just in time to see Gaudio smash a tremendous forehand past a frozen Coria to win the French Open.   After doing a bunch of stuff to get ready for my Italy trip, it was time to watch the NASCAR race from Dover.  My favourite driver, Tony Stewart, was ripping stuff up.  Later it was time to watch the New England Revolution v. Columbus Crew match online and the USA v. Japan women\'s international friendly.  I had the Crew v. Revs match on the computer and the race and USA v. Japan on split screen on the TV.  I was watching the race when I walked back to the computer just in time to see Frankie Hejduk score for Columbus.  One-nil Columbus.  After watching a little bit of the game, I changed the channel to the USA v. Japan game and I was just in time to watch a Japanese player score on a breakaway.  One-nil Japan.  I went back to watching more of the Crew v. Revs game when I tuned in to the race.  I was just in time to witness a big ass crash.  About 20 cars piled up in turn 4.    I went back to the Revs v. Crew match, A PENALTY!!!  Hejduk cut down Steve Ralston in the box.  Jose Cancela chipped a shot and made Crew \'keeper Jon Busch dive the wrong way.  1-1.  Back to the USA v. Japan match and someone from the USA scored.  I think it was Abby Wambach, but I cannot be sure. 1-1.  Back to the Crew v. Revs match for ANOTHER PENALTY!!!  Jeff Cunningham got tripped in the box, got up, and buried the penalty to give the Crew a 2-1 lead.  Pat told me that the Cubs were about to win.  I switched over to the Cubs v. Pirates game to watch the last out.  Cubs won 4-1.  I switched back to racing in time for the green flag.  After a few laps of racing, Casey Mears blew an engine spilling oil on the track.  No caution.   A lap later, Kasey Kahne, in first place with 20 laps to go and pulling away, slipped on the oil, smashed into the wall and took about five other cars with him.  I was helping my mum cook dinner when it was time for my break.  I was just in time to see Mark Martin win and Tony Stewart finish second.  I went back to the USA v. Japan match.  The final whistle blew and the match ended 1-1.  I walked back to the computer to see a THIRD PENALTY!!!  In the second minute of stoppage time, the Revolution had a chance to tie the game up.  Cancela chipped the ball on goal again, but Busch was too smart and tipped the shot over the crossbar to give the Columbus Crew a 2-1 win.  My timing was spot on all day.  I am a lucky bastard.  Practice tomorrow on Fedora Field.  Full report tomorrow.  See y\'all.  '),(16702,3426176,'22,May,2004','2004-05-22 00:00:00','   I hate VH1.  I hate ESPN.  Here I am at the computer listening to  Damnation and a Day  (I seriously can\'t stop listening to it) and I am watching VH1\'s countdown of the \"Most Metal Moments\".  This has to be the most abhorrent countdown of \"metal\" I have ever seen.  There has been countless mentions of bands that I would never (and will never) think of as metal.   Motley Crue  is not metal.   Poison  is not metal.   Bon Jovi  is not metal.   Def Leppard  is not metal. (Although the drummer does deserve some credit for losing his arm and coming back to play again)   Stryper  sure as Hades is not metal.  Bands that dress up like women with make-up and teased hair and wear multi-coloured spandex   ARE NOT METAL!   They sing little girly songs with flashy pyro and dancers and little girls in the crowd ogling over them.  They play music that retarded monkeys could play.  The singing sounds like a squirrel who got his testicles caught on a fence.  Yet VH1 still decides to kiss ass and mention these soulless, tasteless wannabes as metal gods.  Metal Gods?   Slayer  are metal Gods   Anthrax  are metal Gods   Pantera  are metal Gods   Exodus  are metal Gods   Megadeth  are metal Gods   Motorhead  are metal Gods  A  Slayer  fan takes a knife and carves \"SLAYER\" in both of his forearms.  That\'s metal.   Pantera \'s \"House of Shock\" is metal.   Anthrax  making an MTV appearance and trashes a fan\'s house is metal.   Gwar  getting in trouble for their bloody stage show is metal... well... sort of.  All of these are very notable metal moments.  However, VH1 had to include at least 10 entries about groupies alone.  I hate to say it, but groupies are not metal.  Any musician will at some point or another get a bunch of groupies. (the exception being  Clay Aiken )   Other entries that did not make sense at all were:   Poison  revolutionise information maitenance system with groupie database.    Ace Frehley claims he was abducted by aliens. (that is not metal, that is just pure lunacy)    Bon Jovi  marks the beginning of hair metal. (in my opinion the darkest day in metal history)   Poison \'s C.C. Deville goes throws a fit at the VMA\'s.     Poison  screen potential groupies with a video camera.     Motley Crue \'s Nikki Sixx and Tommy Lee compete in a Stink and Scorefest. (a man named Nikki is not metal)    Motley Crue  making fans pee in a kitty litter box to meet them backstage.    Nikki Sixx sets himself on fire.    Here\'s the worst one:  Motley Crue \'s Tommy Lee steers boat without using hands.  YOU CAN\'T BE SERIOUS!!   That  is the #21 top metal moment?  I drive with out using my hands when I\'m on the phone or when I\'m changing the radio station or when I am trying to reach in my bookbag for a CD to put in.  Is that metal?  No, that\'s just me stupidly endangering my life because I am trying to do too many things at once.  As a true metal fan, that countdown sickened me.  I am tired of the old people at VH1 thinking they know everything about metal, having a countdown, and loading it up with washed up hair metal bands.  That\'s why I fully endorse the upstart music channel Fuse.  They are primarily a rock channel, but when they do metal, they do it damn well.  Their metal show \"Uranium\" is hosted by the beautiful and very metal Juliya Chernetsky.  This show plays balls out metal all the time.  You will see the bands appear on \"Uranium\" nowhere else .  When was the last time you saw  Children of Bodom  on VH1?  My point exactly.  The lovely metal goddess: Juliya    So after a while of watching that disgrace to metal, I went to Mickey D\'s to get food for me, my mum, my dad, my cousin, and my aunt.  I went up to the cashier (who was  very  hot) and ordered four salads, a happy meal, a #3, and a double cheeseburger.  So I was getting my drink and when I got back, I saw a huge ass bag and the happy meal and I thought that everything was there so I took it and left.  I felt like an idiot b/c I was struggling to carry all of that out.  I got home and discovered that the only thing in the huge ass bag were two salads.  So I had to go back, get the two other salads which were in another huge ass bag, the #3, and the double cheeseburger.  The cashier must have recognised me b/c she immediately said you forgot something.  I replied bashfully, yeah, I know.  So after sorting the whole mess out with the manager, I got my food and took off, but not before I thanked the cashier again.  She smiled and replied, no problem.  Oh yeah, she wants me. ;)  So I went home and ate my first meal of the day.  It was 14:45.  That\'s 2:45 PM if you don\'t speak European.  I was in a great mood.  I got McDonalds, flirted with the hot cashier, and in 15 minutes, the Fire game was going to come on.  I was enjoying the game when around the 38th minute ESPN commentator Rob Stone stopped his colour commentary to say that ESPN 2 was moving coverage to WNBA Basketball and the Fire game can be seen in its entirety later tonight.  I was absolutely livid!  Why did we have to get bumped off of the air for WOMEN\'S BASKETBALL?!?!  And what caused this happen?  Well, President Reagan died, so I understood why ABC went to Breaking News coverage, but I was  insulted  by the fact that MLS got pre-empted by the WNBA.  I mean, I understand that the WNBA is a new league and they are trying just as hard as MLS to get attention, but what bites my bum is the fact that they get ABC airtime on a silver platter while we get cheap ESPN 2 airtime that is often taken away from us by tennis and the Little League World Series, and now women\'s basketball.  It really really sucks being a soccer fan in America.  If we were in Europe, a nuclear war wouldn\'t cut into a soccer match.  That\'s why I can\'t wait to get to Italy.  They are as passionate as I am about soccer.  I doubt they even care women play basketball.  I doubt they even  know  women play basketball.  Well I\'m off to watch the Fire game on tape delay.  As for you ESPN, I guarantee a venomous e-mail is coming your way.  Especially now that you got the Metrostars v. Dallas score on your sports ticker thing wrong.  Dallas won 2-0 not 3-0.  You are totally soccer inept.  I can\'t wait to get my degree in statistics and flood your sports ticker with soccer stats from Botswana.  Then you will pay for your injustices against the beautiful game.   Bwaaa hahahaha!!!  '),(16703,3426176,'22,May,2004','2004-05-22 00:00:00','   So I woke up today at the ass-crack of 11:00 and I had to go with my sis to get the car from the school that my mum works at and the car that my dad leaves at the train station.  By the time that was done, it was time for lunch.  So I went to Subway and got a... I forgot and went to eat it in the car while I listened to the album rated 100/100 by me,  Damnation and a Day . (read yesterday\'s review if you haven\'t already)  After I ate, I still had another hour to burn.  So I called up my buddy Steve (go visit his blog) and I went to his house.  We looked at funny stuff online and chatted about stuff.  I saw this hilarious video of this guy who seems to think he is a Swedish version of that \"Rico Suave\" guy.  The line this guy uses is something like, \"Ooooh, you touch my tralala, my ding ding dong\".  It is so cheesy and hilarious.  Here\'s the link to it right here:   urlLink Tralala, Ding Ding Dong    After my wonderful visit at Stephen\'s house, it was off to soccer practice.  At first we didn\'t do much.  We just took shots on goal.  My left footed shot has worsened dramatically.  I was almost happy with it, but after today\'s performance I realised it still needs lots of work.  My right footed shots weren\'t too great either, but I am doing better with my power shots.  We then went on a long run.  I am somewhat pleased with my performance.  My legs held up very well, thanks to my 45 minute exercise chair workouts, I wasn\'t out of breath, which is another improvement, but my heart could not take the running at all.  That was my weak point.  If you know of any cardiovascular exercises that will improve stamina, I would like to know what they are.  (I know what is the first thing you all are going to say, but that won\'t work b/c I don\'t have a girlfriend)  After the run, it was time for a 6 v. 5 match.  I was on the team with five players and I think my performance suffered on account of that.  I could not find anyone to pass to, so I was making ill-advised passes.  The rest of the match, I was an absolute non-factor.  After we lost it was time for a desperately needed break.    We played juniors v. seniors and us seniors enjoyed most of the possession.  This game was tremendously better for me.  Towards the beginning I decided to do away with the ponytail.  It may have helped.  I began to make good passes and good decisions.  Maybe it was the teammates that were making me better, but I also felt a lot more energised, a lot more confident, and a lot more in pain.  My hip flexor is a problem that is not going away.  I made a plethora of good passes, almost good passes, a few good shots, one goal, and one great run up the middle beating the entire junior team and getting my shot blocked.  After that, the confidence shot straight up.  My defending became better, my aggression became more intense, I could read the game better, I made great supporting runs, my first touch was coming with ease, and I seemed to have endless amounts of energy.  I felt like that was one of my best performances in a very long time.    Sometimes I play like absolute excrement (the first game) and sometimes I can surprise everyone and play extremely well. (the second game)  I just hope that when tryouts come in August I can play extremely well.  If I do, I just might have a slight chance of making the team.    During the long run and the first game, I realised I will most likely not make the team at all.  I barely did last year and there is so much good talent coming up from last year\'s sophomore and freshman team and even the eighth graders are going to advance up to varsity.  Only five seniors graduated so with so much talent coming in and so little going out, I fear that cuts will be made.  It will be a major disappointment if I fail to make the team since it is my last year and I have not lived up to my expectations for the last six years.  But it\'s also not fair if a freshman is better than me and he gets sent down just so I can stay on the team.  I\'ll leave that decision to my coach Alan.  But the second game gave me hope that I can overcome all odds.  If I attend the soccer camp and play a lot over the summer, maybe that will show Alan that I have the dedication to make the varsity team.  I am already starting to get ready for next year.  I just need to remember this second game and how well I played.  That will allow me to desperately cling on the hope I have for next season.  That will give me the confidence I desperately need to play at my highest level.  I will not back down.  I will not give up yet.  '),(16704,3426176,'22,May,2004','2004-05-22 00:00:00','   Usually when a band makes a record they write about 10 to 12 songs and put them on an album.  Usually the entire CD is anywhere from 45 to 60 minutes long.  Usually songs are conveniently placed in 3 minute boxes ready to be consumed by the audience.  Usually there are three or four stand out songs and the rest are album fillers.  Usually they are separate songs that don\'t correlate to each other.    Usually.     Cradle of Filth  destroys that mold with their latest album  Damnation and a Day .  First off, there are 17 tracks on this album.  True some of them are symphonic bridges between songs (more on this later) but the album would be seriously lacking if it weren\'t for them.  Secondly, these 17 songs take over 76 minutes to play out.  Third, most songs are well over five minutes long, some over six.  Next, every song is just as intense, just as brutal as the previous.  There is never a let down.  And finally and most importantly, this album tells a story.  Each song has a theme and it rolls into the next song much like an epic movie.  Story?  What the hell is this?  What kind of album has a story?  Well, be prepared for something new.   Damnation and a Day  takes on the story of the war between good and evil from the evil (obviously) perspective.  (I must warn you that you really need to know the characters and the stories in the Bible to follow this album\'s theme)  There are four main chapters:  Fantasia Down, Paradise Lost, Sewer Side Up,  and  The Scented Garden .  Each chapter begins with an orchestral track complete with a 40 piece symphony and 30 choir singers which introduce a new theme in the battle between good and evil.  In a few of these orchestral songs, there is a little narration to set up the stage for the new chapter.  The chapter opens with \"A Bruise Upon the Silent Moon\" introduces the theme of Lucifer\'s fall from Heaven\'s angel to ruler of Hell. (this is the  Fantasia Down  chapter)  \"Damned in Any Language (A Plague on Words)\" opens the theme of temptation of sin upon humans. (this is the  Paradise Lost  chapter)  \"The Mordant Liquor of Tears\" displays evils throughout mankind, (this is the  Sewer Side Up  chapter) but also foreshadows the next theme.  \"A Scarlet Witch Lit the Season\" projects - get this - a love story.  Yes, a love story. (this is  The Scented Garden  chapter)  And so it begins.   Cradle  kicks in.  And they kick hard! \"The Promise of Fever\" is an unforgiving song that assaults the ears and never lets up.  Lead singer Dani Filth\'s screams and howls highlight this ominous track.  There is a part where the orchestra builds the energy to a climax and Dani\'s howl pierces through to bring on the ending assault.  The next song is \"Hurt and Virtue\".  This track has a very catchy guitar riff that is one of the best on the entire album.  It speeds by until the end where is slows slightly and leads to a dramatic end.  \"An Enemy Led the Tempest\" has another wonderful guitar riff that blends in wonderfully with the keyboards and the opera singers.  The track changes tempo during the middle, but breaks out to full speed again before the sorrowful agonising ending.  \"Better to Reign in Hell\" is a track with more brutal guitar that switches to elegant very quickly... Then back to brutal again.  This track also features female backing vocals that adds a hint of eerieness and a 45 second instrumental that is top notch.  \"Serpent Tongue\" starts with haunting drums and keyboards which enters again in the middle and end of the song.  Dani\'s whispers are extremely chilling.  \"Carrion\" sounds alot like a  Slayer  song with an orchestra, opera singers, and piercing screams and howls in the background.  There is another part where Dani whispers scripture and when you least expect it,  Cradle   bursts out in full fury.  \"Present from the Poison Hearted\" features another excellent guitar riff.  The keyboards are especially chilling in this track.  Dani features great mouth control when he lists names the Devil corrupted rapid-fire.  I mean, any one who can accurately and clearly say Mahalaleel super fast deserves some credit.  The middle part has a relaxing orchestra part that is eventually raped by Dani\'s howls and the band\'s brutal playing.  \"Doberman (Pharaoh)\" starts out with a sitar.  A crushing guitar riff follows with a female voice harmonising with the guitar riff.  Orchestral breaks with Dani\'s low gravely chants are dispersed throughout this track.  The sitar makes another cameo near the end and the keyboard makes itself known.  \"Babalon A.D. (So Glad For the Madness)\" is the adjoining song between the  Sewer Side Up  chapter and the  Scented Garden  chapter.  This song starts out crushing, then goes off to another eerie moment before embarking on another ultra-catchy guitar riff.  Then it sounds a little hypnotic until it goes onto the eerie again.  After a while, there is a nifty little jam before the catchy riff again.  Towards the end, the song stops completely and the only thing that is heard is a heartbeat for about a second before the band kicks in again.  This part scares the feces out of me every time.  Next is \"Mannequin\".  I have heard this song a few times before, but never the whole thing so I kind of knew what to expect.  The video is... strange (of course) but that\'s not the point.  I never knew what the hell Dani was singing about until I read the lyrics in the book and thought about how it correlates to the theme of the album.  It all makes sense now.  Anyway, this song has one of my favourite guitar riffs on the album and has lovely female backing vocals.  This song has an excellent atmosphere and that is what  Cradle of Filth  is all about.  Song number 15 is \"Thank God for the Suffering\".  This is one of my favourite songs on the entire album.  Strangely enough it\'s a love song and quite possibly the closest  Cradle  will ever come to an all out ballad.  It\'s a great dynamic from the previous songs about sin, Hell, and slaughter.  It\'s cold, sorrowful, and eerie.  This is the turning point of the theme of the album.  I don\'t want to give too much away, but this is really the crowning point.  The absolute best part is at the very end where there is chanting which abruptly stops leaving Dani to whisper in the most haunting voice \"I thank God for the suffering\".  Immediately after \"The Smoke of Her Burning\" starts and it is a neck breaking change of tempo.  As the last real song on the album this is naturally the final battle between good and evil.  I don\'t understand how Paul, the guitarist, did not break his wrist off from playing so fast.  He really has to be playing about a million notes per second.  This is also another memorable guitar riff. (how many is that?)  Towards the end, the song slows from the mindblowing speed and produces one of my favourite line of lyrics ever: \"Weld the gates to Heaven shut/ The abyss leers in hissing ruts/ Unhilt the black grimoire of death/ Inscribe all names God has left.\"  The band fades out, thunder rolls, the orchestra plays with a sorrowful tone, and the choir sings ominously.  The final track \"End of Daze\" enters and it closes with thunder, wind howling, and the last narration with evil laughter that fades away and brings this epic album to an end.  The only thing I neglected to talk about is the drumming.  Adrian puts together the performance of his life.  The double bass pounding is consistent and the snare sounds like a machine gun. (listen to \"Mannequin\" for an example)  He uses the ride cymbal, high hat, and crashes appropriately so they each add different characteristics to a song.  And I will never know how Dani can change his voice from glass shattering high to ground shaking low so quickly.  You may not like the style of singing, but it is amazing how seamlessly he goes back and forth.  Some people may say this is not music.  Well they are right.  It is something so much more.  It truly is an epic.   Cradle  certainly had lots of balls to try and make an album like this.  Concept albums usually don\'t work, but they made it perfect.  The story alone is very intriguing.  Each song is excellent when listened to on their own, but they are only fully appreciated when they are listened to completely all the way through in the wee hours of the night while reading the lyrics.   Damnation and a Day  has raised the bar of heavy metal to dizzying heights.  Maybe the only way this album can be topped is by  Cradle of Filth  themselves.   Cradle of Filth  -  Damnation and a Day   Sony Music (2003) 17 tracks (76:53)    Dani \"Filth\" Davey: Vocals Paul Allender: Guitar Adrian Erlandsson: Drums Dave Pybus: Bass   Immortal songs:  \"Hurt and Virtue\", \"Better to Reign in Hell\", \"Thank God for the Suffering\", \"The Scent of Her Burning\"  100 out of 100.  ');
INSERT INTO `msgs` VALUES (16705,3426176,'22,May,2004','2004-05-22 00:00:00','   I know I was supposed to give you the  Damnation and a Day  review for you all today, but I want one more day to listen to the whole thing continuously.  That\'s the key word: continuously.  Don\'t know what I mean?  You will.  I voted for the MLS All-Star game yesterday.  This format foreshadows a match between the MLS All-Stars and the most famous soccer club in the world Real Madrid.  The sad part is many Americans know more Real Madrid players (David Beckham, Ronaldo, Roberto Carlos, Luis Figo, Raul, Zinedine Zidane) than MLS All-Stars.  However, Real Madrid scheduled a Japanese tour that will clash with the All-Star game.  What will happen?  Stay tuned to find out.  Anyway, I still voted for the MLS All-Star team and this is who I cast my vote for:  Goalkeepers (2): Henry Ring (CHI), Kevin Hartman (LA)  Both â€˜keepers have done outstanding for their team this season.  Who would have thought the Fire had an All-Star sitting on the bench for two years while Zach Thornton dominated the â€˜keeper position.  Defenders (5): Jim Curtin (CHI), Kelly Gray (CHI), Frankie Hejduk (CLB), Cory Gibbs (DAL),  Jeff Agoos (SJ)  It was hard only picking five.  I chose Jim Curtin because he is the most underrated player in MLS and has been huge in the Fireâ€™s defensive success so far this season.  I chose Kelly Gray because he is my favourite Fire player, and I chose Frankie Hejduk because he is a great defender, but also gets into the attack very well and has outstanding speed.  Gibbs deserves an All-Star spot because he has been solid all season, and Agoosâ€™s veteranship will be great against a team like Real Madrid.  Midfielders (5): Chris Armas (CHI), DaMarcus Beasley (CHI), Landon Donovan (SJ), Andreas Herzog (LA), Cobi Jones (LA)  This may not be a very good midfield.  True, it is loaded with outstanding players, but it wonâ€™t work because there are way too many attacking midfielders.  Donovan, Jones and Herzog are all mainly attacking, Beasleyâ€™s main job with the Fire is to support the attack, and Armas canâ€™t stop Real Madrid all on his own.  I would probably put Herzog as the only attacking midfielder because his play this season has been simply outstanding.  His crosses, passes, and vision are second to none.  I cannot stress enough how wonderful Herzog has been this season.  As for Donovan, I hate him, but you cannot deny his wonderful play.  Cobi Jones is getting up there in age so if I have to leave one out, I would leave him out even though he is one of my favourite players.  Forwards (4): Ante Razov (CHI), Brian Ching (SJ), Jovan Kirovski (LA), Carlos Ruiz (LA)  The only one out of these four that will definitely be on the line up is Kirovski.  Ruiz has injury problems, Ching may not get enough votes because he is another one that is underrated, and Razov has been cold most of the season, and anyone not living in Chicago hates him passionately.  Kirovski has been outstanding this year showing why he needs to be considered for the U.S. National Team.  However, four of his five goals have been scored against the Dallas Burn so he may have that going against him.  If Ruiz is healthy, there is no question he will find an All-Star spot.   You can vote for you MLS All-Stars here:   urlLink Vote for your MLS All-Stars!!!    That\'s it for now, I\'ll get you the  Damnation and a Day  album review tomorrow.  See y\'all.  '),(16706,3426176,'22,May,2004','2004-05-22 00:00:00','   Section 8.  The home of the best, most loyal, most passionate soccer fans in the United States of America.  Section 8 also gets help from fellow Fire fan groups like the Barnburners, the Arsonists, and the Ultras, but usually there are a little - no - alot more than 9,000 of us at Fire games.  C\'mon people, we symbolise what the future of Major League Soccer is about.  Thousands of fans packing a stadium supporting the best sport in the world by cheering, singing, banging drums, and cursing out the other team and officials.  All of this energy we generate goes to the team we are supporting leading them to reward us with goals, saves, and wins.  However the attendance numbers this year have been a joke.  Four games played, all of them with less than 13,000 fans.  I\'ll be straightforward here.  That is shit.  I mean no wonder why the Fire lost two of their first three games at home.  No support at all.  That 3-1 home loss to Columbus was the most disgraceful loss I have seen them play and it was also the smallest, tamest, crowd I have ever seen at home.  Other teams fear us, but if we keep getting such lowly attendance numbers the home field advantage will wear away.  This Saturday Freddy Adu will arrive into town so the numbers at the game should soar, but I would hate to think that so many people are attending this game to watch someone from the other team rather than our team.  So I say to you go out and watch the Fire game at Soldier Field.  Go out and watch Freddy.  But remember why you are actually going out in the first place: to watch the Fire win, to cheer them on, and to watch the beautiful game.  Another thing that I am pissed about.  Chivas USA.  For those of you who don\'t know, Chivas USA is close to becoming the next expansion team to enter MLS.  That\'s cool, but the problem is the owner of Chivas USA is none other than Jose Vergara, owner of Chivas CD.  Chivas CD is a Mexican soccer club that has an outstanding history behind them.  Every player to play on Chivas all time has been of Mexican descent.  This is all good, but Mr. Vergara wants MLS to lift their restriction of three foreign players per team just for him so he can field an all-Mexican team -  IN AN AMERICAN LEAGUE .  That\'s not going to work buddy.  I don\'t even know why Vergara wants a club in the United States as it is.  When it comes to soccer, America is Mexico\'s worst enemy (whether they admit it or not) so why is he trying to help MLS with his investments?  I think if his club is successful, he will make it look like MLS is a joke of a league, if Chivas is not successful, he will turn around and say the League treated him unfairly or some crap like that.    CHIVAS SHOULD NOT BE IN MLS!!   And as if that is not enough, Vergara wants his team placed in a city that 1) already has an MLS team in it and 2) already has (or has plans for) a much envied \"soccer specific stadium\".  He wants to go to Los Angeles so he can leech off of the Galaxy\'s newly constructed Home Depot Center or - get this - he wants Chivas USA to call  CHICAGO  home.   OF ALL BLOODY PLACES!!!   Do you believe this?!?!  This guy has the balls to come into a hostile country, put a team in their league, want rules to be broken to satisfy his desires, and request that his team share a world class field that he did not invest a goddamn penny in.  Buddy, if you want to stick your foot in the door of my team\'s stadium and request to share it with us, I will rip the boot off of your foot  AND SHOVE IT STRAIGHT UP YOUR ASS!!!   You are going to take the hispanic fans attending Fire games and build your club with their image so they will go to your team.  Hell no buddy.  And Peter Wilt (general manager of the Fire) what the hell is your problem?  You are encouraging the enemy into our town and taking our fans to move into our stadium that we are paying for?  I hope you are just talking out of your ass b/c if you are really serious about welcoming Chivas you are seriously an idiot.  And if you aren\'t talking out of your ass, be a man and tell Vergara to go do something unpleasant to himself.  Don\'t make your fans do it for you.  MLS, if you want a city worth expanding for, promote the Rochester Raging Rhinos from the A-League to MLS.  They have everything you want.  1) soccer specific stadium [that is expandable mind you] 2) a good team, [this team beat the Fire in an Open Cup match two years ago, they are for real] 3) a dedicated town, owners, and fans.  In fact Rochester may have the second best soccer fans in the country. (Only behind the Fire fans when they feel like showing up)  This is so obvious.  Take Rochester, screw Chivas.  Do the right thing for American soccer.  Tomorrow I will talk about my MLS All-Star team and I will give my first album review.  I bought  Cradle of Filth \'s  Damnation and a Day  and I will give a review of that tomorrow.  Until then, see ya.  '),(16707,3426176,'22,May,2004','2004-05-22 00:00:00','   Editor\'s Note:&nbsp; Obviously, this did not just happen.&nbsp; I am desperately behind schedule on my posts.&nbsp; Just be patient and I will do the best I can.&nbsp; These are immensely long posts.      &nbsp;   Today, I went with my sister, her boyfriend, his brother, and two of their friends to see NASCAR Busch Series and Nextel Cup Qualifying. We got to the parking lot, which was a giant field and walked all the way around the track to the entrance. We could hear each car go by as they did their qualifying laps. We got in and looked at the sky and knew we were screwed. Dark black clouds and green clouds hung over our heads ominously. We got seats and sat down to watch two cars qualify. Then a giant Tropicana blow-up orange broke loose and ran a qualifying lap of its own. A safety truck tried to corner it, but the runaway orange rolled right over it. When the orange crossed the finish line, the crowd cheered. The orange finally got hung up in the safety fence in turn one. It was popped, but still stuck to the fence. The safety crew finally got the orange off the fence when the first raindrops pelted the track... then poured. We ran under the grandstands and stood in a little dry rectangle. After the it rained for an hour and the crew dried the track for another hour and a half, it was time to continue qualifying! I got up close to the fence to take pictures of Kasey Kahne\'s car and Tony Stewart\'s car. I will post them later when they get developed.   Back to the Italian madness:      The alarm rang at 03:15. Or somewhere around there. It was too damn early, let me just say that. Anyway, I got dressed, put contacts in, got everything I needed for the three hour bus ride in my bag, and was ready for take off. Werner and I got to the coach bus around 3:45 so we were early. We were waiting for 10 minutes after our scheduled depart time when Eric, Alan, and Pietro arrived fashionably late. After that, everyone was aboard so it was off to the Vatican.   The five hooligans (Me, Tim, Eric, Anthony, and Robert) sat in the five-seat backseat on the bus. Occasionally, Beatrice would come by and sandwich herself between Eric and me. We talked most of the way there about random stuff while looking through Robert\'s Italian dictionary for off the wall phrases and obscene words.&nbsp; For example, we learned how to communicate with animals from that book.&nbsp; Then someone took it a step further and called me \"assholeballs\" in Italian.&nbsp; How do you mess that up?&nbsp; Then Tim remarked that the Pope\'s home should be renamed \"Pope-al Stadium\".&nbsp;   Yeah, that made no sense either. &nbsp;  &nbsp;   Eventually we got to the Vatican.&nbsp; We got out and climbed down an eternity of stairs and there we were.&nbsp; We were at Pope-al Stadium. &nbsp;    &nbsp;   We got inside and found our seats.&nbsp; Originally, this was only supposed to be a private gathering with us and about 200 other people, but somehow, over 1,000 people were in attendance.&nbsp; They were groups of people from all over the world.&nbsp;&nbsp; I believe the entire... ceremony I guess one can call it... was delivered in five languages. (English, Italian, German, Spanish, French)&nbsp; The Pope made his way out and everyone took a bunch of pictures.&nbsp; Here\'s something that bothered me:&nbsp; it was over 90 degrees, the Pope almost 90 years old, and he is in bad health so why is he in layers of heavy polyester robes and gowns?&nbsp; That\'s not smart if you ask me.&nbsp; Here\'s something else that is on my mind:&nbsp; why is it that the holier the place I go, the more morbid my inner thoughts become?&nbsp; Why was I hoping for the Pope to suddenly croak?&nbsp; Why was I hoping for some strange little catastrophe to happen?&nbsp; Why was I booing everyone else when their respective group was called out?&nbsp; Oh well, I\'m going to Hell. &nbsp;    &nbsp;   After the Pope\'s little ordeal was mercifully over, it was time to go... somewhere else.&nbsp; Oh yeah, we were supposed to go to the Sistine Chapel, but for some unknown reason we couldn\'t.&nbsp; So for the next two hours we wandered about eating our packed lunch, going out into the town square for a drink and gelato and just chillin\'.&nbsp; While we were chillin\', Tim made a bet with me that Eric was about to \"get some\" on the trip.&nbsp; I set the wager at $5 and counted down the days until payday.&nbsp; I was sure I was going to win.&nbsp; Later, I ordered myself a gelato with ease and looked at a little shop that was selling cheap knock off jerseys of famous soccer players. &nbsp;  &nbsp;   We eventually had to gather again to go.&nbsp; So we all went&nbsp;to walk to the bus.&nbsp; The walk was uphill on a narrow sidewalk.&nbsp; That\'s a stressful walk when the group is over 60 people large.&nbsp; We made it after about 20 minutes and I thought that it was finally time to escape the heat and make it back on the bus... I was incredibly wrong.&nbsp; It was time to go into... someother holy building and look around.&nbsp; We had to split up into two separate groups because our group was so large. &nbsp;    &nbsp;   We got to the end of that so I decided to get an early jump on everyone and I left with Matteo, Ulderico, and Anthony.&nbsp; We would eventually run into Alan, Eric, and Bea.&nbsp; We got to the exit, but to our surprise, no one else was there.&nbsp; We sat around at a gift shop like the tour guide said and waited for about 15 minutes when we decided we should try and find everyone else.&nbsp; Reluctantly, we got up and tried different exits of where everyone else might have been.&nbsp; Luckily, Matteo had his mobile and he called Werner who was sitting comfortably on the bus.&nbsp; Matteo decided we all should walk back to the bus to be safe.&nbsp; We were walking back when Alan ran into some of his old Italian friends.&nbsp; He stopped to talk with them for a minute or two and brought them with us to our trek back to the bus.&nbsp; We happened to run into a few of the&nbsp;chaperones on the way back and they decided to go with us back to the bus.&nbsp; We were on the same path we went on to the entrance and were almost done when we saw the whole group we lost rushing in the opposite way.&nbsp; Extremely frustrated that we made the long trip in vain, we went back to go into the Sistine Chapel, the same chapel we were denied access to in the first place.       We hauled ass and followed everyone to the entrance to the Sistine Chapel... which was a few feet from where we were in the first place!&nbsp; We rushed through security, the chapel, and all of the stuff inside.&nbsp; I touched the famous foot of St. Peter for good luck.&nbsp; I think it is supposed to bring good luck.&nbsp; After that, we left for the bus.      The walk back to the bus took an eternity.&nbsp; It was hot and we walked uphill most of the way.&nbsp; We finally got to the bus and everyone filed in.&nbsp; We were waiting for about another 20 minutes because Alan was missing.&nbsp; The tour guide got on the microphone and made a remark about us being behind schedule because we were wasting time.&nbsp; Wasting time?&nbsp; They were the ones who&nbsp;made us sit around for two hours because they wouldn\'t let us in the Sistine Chapel and they want to complain to us?&nbsp; Piss off.      Alan finally made his way to the bus and was unjustly disciplined for losing his way.&nbsp; We headed for home, but first made a quick drive by the Colisseum.&nbsp; Towards the end of the bus ride, Werner, Matteo, and Bea started a raucous rendition of the Italian National Anthem and other Italian songs.&nbsp; Frankly, I enjoyed it.&nbsp; But then, a few American girls started the worst rendition of \"Take Me Out to the Ballgame\" in the history of man.&nbsp; Even worse than Mike Ditka.&nbsp; Yeah.&nbsp; That bad.      We finally made it home and it was time for the first good news of the trip: THE LUGGAGE ARRIVED!!!!&nbsp; Most of it at least.&nbsp; There were a few pieces missing and it was reported that some of the suitcases were destroyed.&nbsp; We got to the Hotel Arlecchino was told that the luggage was at the Hotel Progresso.&nbsp; We got there and searched for our luggage.&nbsp; We found most of it.&nbsp; My mum\'s suitcase was wrapped in packaging tape.&nbsp; My dad\'s suitcase had the zipper ripped off and gouges on the outside.&nbsp; My suitcase had a huge rip in the side of it and several scratches on the surface.&nbsp; I couldn\'t wait to get home and see if anything was missing.&nbsp; As long as my soccer shoes are in there, I would be OK.      But first, I was starving so we made a quick jaunt to a nearby McDonald\'s.&nbsp; I was really interested to notice the difference between Italian McDonald\'s and American McDonald\'s.&nbsp; I ordered a Big Mac and noticed that it was not as greasy as the American version and the bread was somewhat stale.&nbsp; The meat tasted better and the secret sauce was a little bit sweeter.&nbsp; I heard from people that the fries were alot better because they had olive oil on them.&nbsp; I think I like them the same.      I opened up my suitcase to discover that half of what I packed was gone.&nbsp; Two new pairs of shorts were gone.&nbsp; My shaving kit was gone.&nbsp; A shirt was gone.&nbsp; But worst of all, those bastards at the airline STOLE MY SOCCER BOOTS!!!  (Long obscene rant deleted)        Angry, I got ready for bed and went to sleep.&nbsp; Those bastards will rue the day...     Day 5 later.&nbsp; I will also&nbsp;add pictures once I finally learn how.  '),(16708,3426176,'22,May,2004','2004-05-22 00:00:00','   On with Italy...    Day 3 was a blur. I woke up past 11:00 and just sat on the bed in a tired daze. I don\'t really remember anything. I was so tired I probably slept all day. I do remember going to bed real early. The next day was a big day. It was the trip to Rome.    The Rome trip later.  '),(16709,3426176,'22,May,2004','2004-05-22 00:00:00','   So we continue the trip to Italy...   So there we are.  Weary and restless we file out of the plane.  We board a shuttle to take us to the terminal (I think that\'s what they call it) where we rode on the left side of the road.  That was quite strange.  So we got to the terminal.  I had to piss like nothing else.  After an eternity of walking and climbing stairs, I finally take a whiz.  After that it was time that time again... AIRPORT FOOTBALL!!!!  I dribbled around a bit b/c no one wanted to play.  Their loss.  Then I stumbled across a couple of errant English newspapers.  I turned right to the sports section to read four whole pages about England\'s chances in Euro 2004.  At this point I knew England were destined for disappointment.  After more waiting and more walking and more... nothingness, it was time to board the plane to Rome.  We got on and fought with another member of our group for a seat and I prevailed.  After that, we ate and I slept for about a good two hours.  When I awoke, I was pleasantly greeted by beautiful scenery.  We flew over France and the Alps in Switzerland and northern Italy and we flew over endless green fields.  It was a wonderful sight.  We got to the airport and I was glad to see signs in both Italian and English.  I told the elder Pascarella, Stephanie, that I hoped all signs would be in both Italian and English.  I would be mighty disappointed.  It was time to go to the baggage claim.  We waited for our baggage... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... then played some airport calcio... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited... and waited until we said \"screw it, let\'s go to our homes and have someone drive them to the hotels\".  The hotels were where the chaperones were staying.  I was all for the idea of leaving a long ass time ago because I wanted to get home and watch Italy vs. Denmark.  So we got on a coach bus (last year in America all of the trips were on school busses) and began the three hour trek from Rome to San Benedetto del Tronto.  Tim was the genius who decided to bring his portable DVD player so we watched  Eurotrip.    Eurotrip  was pretty funny.  The moron guy was hilarious and it was especially funny when the dweeb guy was the only one to get some kind of action if you know what I mean.  And the song \"Scotty Doesn\'t Know\" is an instant classic.  And the Manchester United hooligans were hilarious too.  And have you ever been to Bratislava?  Good times.  We stopped at a gas station for snacks.  I walked in and stood in an aisle blocking everyone because Sweden vs. Bulgaria was on.  At the time Sweden was winning, by how much I\'m not sure, but they were winning.  I bought pizza and realised how much of a problem the language barrier is.  It was quite strange.  And very uncomfortable.  Especially when the waitress out of nowhere started to speak English.  I was totally off guard.  It was time to go back to the bus.  We rode the bus for an eternity in the dark.  It was silent.  I think every one was asleep.  I fell asleep for about 20 minutes and awoke to find myself drooling on my shoulder.  Yeah, it was a good thing everyone was asleep so they didn\'t see that.  After 24 hours of travelling, we finally... FINALLY reached the Hotel Arlecchino.  It was so great to see everyone again.  Werner, Mario, Beatrice, Ulderico, Alessio, Mattio, Alessandra (all of them), Pietro, Maria and everyone else from last year.  Werner rushed up to me and we gave each other a welcoming hug.  I tried to point Tim to Alessio and his family, but then I say Bea and we gave each other a huge giant mega hug.  It was wonderful to see her again.  And she was cute as always.  Her hair was down to the bottom of her ear which is alot longer than last year.  I really liked it.  I met Werner\'s family and introduced them to my parents.  We went to the hotel and got the key for my mum and dad.  I left with Werner\'s family to go and get some pizza.  I had the Margherita pizza which is a very thin, chewy crust with a layer of crushed tomato and cheese.  Yeah, a little bland.  So we sat and talked about the flight and the missing luggage and how Italy can\'t do a damn thing in international competitions because Giovanni Trapattoni sucks as a coach.  Italy played an uninspired match with Denmark and drew 0-0.  I was tired so it was time we went home and went to bed.  From the time I walked into the house, I felt incredibly comfortable... and tired.  I took out my contacts, got into bed, and fell asleep.   Day 3 later.  '),(16710,3426176,'22,May,2004','2004-05-22 00:00:00','   Today and yesterday I had stomach problems.  Not the I-feel-like-I-have-to-vomit feeling.  Not the I-feel-like-I-got-punched-in-the-stomach feeling.  More like a I-have-an-alien-feeding-off-of-my-insides-and-it-is-ready-to-burst-out-and-destroy-mankind-and-build-an-intergalactic-empire-and-make-Earth-a-giant-alien-wasteyard-filled-with-alien-SUVs-and-RVs-and-all-kinds-of-alien-gas-guzzling-vehicles feeling.  Y\'know that feeling?  Anyways, this is the part of the blog where I post something about my Italy trip.  We will go back to the first day: 13 June, 2004 and relive the randimosity.  Off we go:   So it begins.  All of the endless preparation for the past two weeks will finally pay off.    I got up and was prepared for the huge event... USA vs. Grenada in World Cup Qualifying of course!  This should have been a great match to watch because the Americans should dominate the Grenadians by at least 7 goals, but the Americans played like rubbish.  It was 0-0 until DaMarcus Beasley represented MLS and the Chicago Fire well with a goal in first half stoppage time.  The Beas would add another for the Americans and in second half stoppage time Greg Vanney added one more to make things a little more respectable.  During the game, I got my stuff packed and checked over for the last time.  I needed a few more things.  I went to Wal-Mart to get what I needed and was shocked by the total price: $6.66.  Not a good sign.  Especialy because I will be on a plane in a few hours.  I got that thought out of my head and it was off to Bloom High School to get picked up by a bus and go to the airport.  Then I thought and I told my sis, what if this was like  Final Destination?   She laughed at me, for it was a joke, and I said my last good-byes to my aunt and her family and my sis and her boyfriend and it was off to the bus.  It started to hit home on the bus.  Two weeks in a foreign country that doesn\'t speak my language.  At least I have a bunch of friends with me.  My good buddies Anthony, Robert, Eric, the new kid Tim, Suzie from soccer, another good friend Marissa, and the Pascarella twins who are always cool to be with.  After the bus ride, we were at O\'Hare Airport.  This is where things got interesting.  We waited in line to drop our luggage off at a... luggage dropoff place and we were on our own until we had to meet to board the plane.  That meant we had about 3 hours to ourselves.  Instead of going to McDonalds as I suggested, we went straight for the boarding gate.  I got in after fumbling through the security check-in like it was my first time (it was my second) and had three hours of nothing to do.  We sat down and I started a game of Airport Soccer with the elder Pascarella twin, Stephanie.  After a few minutes of that mindless fun, the five stooges (me, Eric, Anthony, Robert, and Tim) went off to explore the airport.  That got boring quick and we went back.  Anthony and I left everyone else and walked to the other end of the airport.  There was a huge congregation of boistrous college aged frat boys.  Their plane was delayed so they were given a pass to bypass security and go get some food.  Anthony and I heard this and decided to try our luck and blend in with them in hope of getting McDonald\'s like my original plan.  The college congregation was so immense we never would have gotten food and gotten back to the gate in time.  We only had about 45 minutes left.  We decided to go back to the gate and starve.  After a while of conversation, it was time to board the plane.  Off we go...  Our seats were scattered so I had no idea who I would sit next to.  I could be sitting with the guys or I could be between two fat dudes.  I was pleasantly surprised to find out that I would spend the next 8 hours sitting next to the Iacopelli\'s.  There were little TV screens on the back of the seats so we could watch movies during the flight.  I watched BBC World news and after that they played highlights of the European Championships.  Greece shocked Portugal with a 2-1 win and Spain won 1-0 over a flat Russian side.  Then it was movie time.  There wasn\'t much interesting on so I decided to watch  The Girl Next Door  with Kim Bauer from \"24\".  I didn\'t want to seem like a perv though.  Oh well, it wasn\'t terrible.  It was like a normal movie aimed at teens.  After a while I tried to sleep.  I didn\'t sleep much, but I did sleep some.  I woke up and my feet were numb, my eyes hurt, and the back of my head was burning up.  Later was breakfast time.  We had orange juice with muffins and yogurt.  We were over Birmingham so there wasn\'t alot of time before we got to London.  Minutes before we landed the plane was low to the ground so we could see the city from the window.  There was green pastures everywhere with town homes with red roofs.  When we were close to landing, Marissa and I took a picture and before we knew it, we were in London.   Day 2 tomorrow.  '),(16711,3705153,'28,June,2004','2004-06-28 00:00:00','   As usual, I woke up at 7am, prepare myself and go off for school.. Last night I guess I din sleep well, cause is like I have to change back my body clock u see so i wake up quite a few times in the middle of the night then went back to sleep again. Actually I woke up at about 5.30am in the morning but is too early so i went back to sleep for another 1h 30mins. While I was walking to the Jurong East MRT station, I was like thinking DIE today is culture & Communication module, will it be hard? Beckham misses 2 penalties in EURO2004, I gonna lots of nonsense from my friends. My class know I am a Beckham fan cause I wore my Manchester United NO.7 Beckham jersey to school sometimes and my wallpapers are usually Beckham But they had never seen me wear my REAL MADRID NO.23 Beckham jersey and maybe somedays I should wear. True enough, when I step INTO my class, I\'m Surrounded with Nonsense HElp!!.. LOL.. Beckham-ed it, Beckham misses penalty and stuff like that surrounded and I have nothing to say so I say \"Don Blame Him, his times is over and pioneers should take over\". I blamed on the pitch also. After that we carry out with the discussion on EURO 2004 then I was like saying \"FCUK rite ?\" when the class suddenly quiet down. LOL.. then everyone laughed at me then I looked at the facilitator but she never say a word so I hehe what to do ? HECK CARE LOR.. Unexpectedly, Culture & communication module for today is quite easy, we don have to research much and is just talking about how to be a good speaker. The whole class was slacking, talking and chatting with one another, some surfing the net and stuff. One of my classmates announced SCIENCE UT RESULTS are out CHECK IT NOW.. Everyone was like \"DIE SURE FAIL\". Suddenly YAY!! I GOT A !! some CB KNN I GOT D+.. I myself got an A and very very happy cause I din expect a A instead I thought I will get a C or something.. My predicted Grade for the module is A as well.. OMG I REALLY LOVE SCIENCE.. I only 1 B so far for my Science and the rest are A\'s.. HEHE.. I think i shall not be complacent.. Continue to work hard for my other modules and my SUCCESS will come eventually.. GOoD Luck to myself. Tml I will be having Visual Basic UT, I Hope for a B for that. TN0203A Rocks !! LIVELY CLASS, FACILITATORs Enjoy Teaching US !!   '),(16712,3705153,'27,June,2004','2004-06-27 00:00:00','   Determined Becks  Beckham is my idol since he came into Manchester United.. His style of playing and crossing really impressed me although nowadays his skills is not as impressive as before and deteriorating.. Nowadays I may not idolize him as much as before BUT after reading the news posted at Beckham Magazine makes him idolize him even more.. He is such a strong person.. People may write him off, criticise him and such but he has his own reasons which other people and fans do not understand..   Beckham said: \"If they want to write me off I\'ll just keep coming back and back until I\'ve won.\" \"The intrusion and the invasion into my life is tough because I\'ve got children and I want them to enjoy their lives but people will just say \'why\'s he complaining?\'.   \"I\'m not complaining but of course it\'s tough. It\'s tough on any family.    I guess he is still not adapting well enough to the way of life in spain, he said that It\'s full of four and five cars following me every day everywhere I go, to the park with my children\" At Least in Manchester, he is able to go to the trafford centre and walk around with his family. In Spain is different story, he cant even step out of his front door.. He is willing to take critiscism BUT at the end of the day he will fight back..  His determination impressed me dude ! He is losing, he admitted and made no excuses but in the end he will fight it back.. He playing two important roles, 1st is to maintain his reputation in soccer and 2nd take care of his family.. WELL DONE BECKHAM ! KEEP IT UP.. EVERYONE DO MAKE MISTAKES, NO ONE IS PERFECT.. IT DEPENDS ON ONESELF TO BE STRONG AND STAND UP ONCE AGAIN.. DETERMINATION SPEAKS THE LOUDEST.. =)    Stop All Becks Stuff, Verdict = Beckham Is My Idol & Forever My Idol..     NO.7 David Beckham (ENGLAND RULEZZZ)  Today I went west coast with some of my friends; Keith, Yanxiang, Yaoxing and Waijun.. Yaoxing brought his doggie there too (Mocha).. We also had a great TCS there.. TCS = Talk cock Session.. We threw the bone and mocha picked it up and brought it back.. FUN~ but he so lan lan type after a few runs.. Just lie down on the grass and rest.. Mocha saw one other dog and he ran towards the dog.. YAoxing like so stressed up type.. LOL.. the other dog named ROMEO!! Hehe.. Mocha Found a NEW friend.. After that we talk talk talk........ Keith said he HUNGRY liao then we head to the mac and have our lunch and WENT HOME.. TIRED when i reached home.. Tml school starting no more life, WORK WORK AND WORK.. STRESS, STRESS and STILL STRESS.. Best Of Luck to myself as well, must be like beckham DETERMINED TO DO ANYTHING I WANT..   '),(16713,3705153,'26,June,2004','2004-06-26 00:00:00','   Oh my god, this match is just like a repeated match of Portugal -vs- England. Both teams end up deciding who will go in the semi-finals by PENALTY.. The winning team will be facing the HOST(Portugal) in the semi-finals.. Another great match I should say, either being saved by the goalkeepers or shoot it wide.. Netherlands being the luckier ones won the penalties shootout over Sweden with the score of Sweden 4 - 5 Netherlands.. Let me review on the penalties shootouts.   Penalty takers for Netherlands - Ruud van Nistelrooy, Johnny Heitinga, Michael Reiziger, Philip Cocu, Roy Makaay, Arjen Robben. Captain of Netherlands Philip Cocu misses his penalty.  Penalty takers for Sweden -  Kim Kallstrom, Henrik Larsson, Zlatan Ibrahimovic, Fredrik Ljungberg, Christian Wilhelmsson, Olof Mellberg. Sweden\'s Forward Zlatan Ibrahimovic misses his penalty, his mistake is being recovered by Netherlands Cocu. However, Netherlands Goalkeeper STUNNED Sweden by saving the penalty taken by Olof Mellberg.  Full time Score - Sweden 0 - 0 Netherlands (Netherlands Win 5 - 4 on Penalties)  My Favourite Netherlands Players    NO.19 Arjen Robben   NO.10 Ruud Van Nistelrooy   NO.11 Rafael Van Der Vaart  **All Rights Reserved - Brooklyn** Proudly Present To You By Brooklyn-   '),(16714,3705153,'26,June,2004','2004-06-26 00:00:00','   Euro 2004 Updates..  [26/6/2004] FT France 0 - 1 Greece   I watched this match but is only the second half.. Greece being the underdogs defeat France with a assist from Theo Zagorakis and a powerful header from Angelos Charisteas which brought them into the semi-finals, leaving the french behind. If You guys watching the next Greece\'s match take note of Zagorakis, he is one who has the potential to play in premier league.. I am sure after this Euro 2004, premier league clubs will be scouting for players in Greece.  Below are some of the pictures I\'ve got..    This is the header from Angelos Charisteas which give French\'s bathez no chance to save the ball.. GREAT !!    Celebration After The GOAL..    Team Greece running towards their fans after the final whistle.. Final Score France 0 - 1 Greece. Underdogs Rules Euro 2004..  Now England accompanied by France, Back to their Homelands. \"Let\'s go\" England : \"We are resting for Premier League\"  **All Rights Reserved - Brooklyn** Proudly Present To You By Brooklyn-   '),(16715,3705153,'25,June,2004','2004-06-25 00:00:00','   Walking like a drunken out of my bedroom, scratching my head and not knowing what to do next. Dragging my restless body into the bathroom and starting brushing my teeth.. Weird rite? 5.30pm evening i brush my teeth.. After that, I took my small brother\'s WANG ZAI NIU NAI and drank half of it.. LOL.. He is kind enough to let me drink it.. Hehe.. These Pictures, I Took When I Wake up ( Hair Messy ) LOL..       Following the daily routine, I switch on my computer and see whether my friends are online.. Once I logged into MSN.. Someone msg me.. Guess who? It\'s KEITH (The Father Bear) LOL Kidding ar.. Keith.. We started discussing where we should go at night and after around 10min of talk he suggested THE NIGHT MARKET AT JURONG POINT (Pasa Malam). We jioed a few of my friends too.. Yaoxing Chunjie Waijun and Sarah HEHE..   Ok I dun wanna be long-winded.   .........................................................  Ok I meet Yaoxing and we both go together.. We reached JP.. As usual, \"We Are the Latest\".   Then WALK WALK WALK.. Waijun brought his holland jersey, we ate burgers, otah and such.. After We Went FFA (Free For ALL) I Guess That\'s About All.. Still considering whether to watch Euro 2004 France -vs- Greece.. If I am watching I will Update My Blog with the lineups and all.. If not, I will be sleeping..   Boring ya? Cause there\'s Nothing to talk about and all I do is spending most of my time SLEEPING, That\'s it. Some regarding to my personal life i dont wish to share LOL.. Nothing actually.. =)   '),(16716,3705153,'25,June,2004','2004-06-25 00:00:00','   Oh My God.. Match gone into extra time..   End of 90mins - Portugal(Postiga 82\') 1-1 England(Owen 3\')  England played very defensive but sad case, James(Clumsy Ass)conceed the goal.. Stadium came back alive once again with the portugese fans cheering on. Rooney injured during the first half and also being sub-in by Aston Villa\'s (Vassell). Very exciting match and my heart is still with England.. Sole Campbell\'s goal is being disallowed by the referee. Its kinda foul or something. Happy for second but GONE after next !! England No Luck At all ! Totally Nonsense !! Oh my godyyy...   We are enjoying at Keith\'s house, still watching the match and some of them playing those newbie & childish game (Gunbound). Ok now the match has gone to another half of 15mins. Both teams are playing very aggressively, Hard tackling and such.. I hope England is able to pull through into the semi-finals. Panic man.. My blog is almost all about soccer matches.. LOL.. Mainly because I have nothing memorable to talk about, if I have I wont hesitate to share with u guys..   Ok England Lost.. 2-1 to Portugal. JAMES You think you are real tall? 1.96m Means You No Need To JUMP? Why Are u So CLumsy !!! I can see portugal\'s fans Crying.. SADDED*** LOST MOOD..   OK DAMN LUCKY WOWOWOWOWOOWOWOWOWOWOOWOWOWOWOWO.... FRANK LAMPARD !!! It\'s 2-2 DRAW DRAW.. CHEER CHEER !!! HAIL TO ENGLAND !! Match Might Settled With Penalty Shootout.. ONE OF THE BEST MATCH IN EURO2004.. EXCITING.. VERY VERY EXCITING.. Last attack for England Its A Corner !! Blocks Out.. MAtch Has GOne Into Penalty ShootOuts.. My Heart is Pounding REal FasT, aNy Moment It Might POP OUt !! Commercial Break Now.. England Lost In Penalty Shootout.. Haiz.. No LUCK.. Bye.. SLeep !!  **All Rights Reserved - Brooklyn** Proudly Present To You By Brooklyn-   '),(16717,3705153,'24,June,2004','2004-06-24 00:00:00','   Ok I have review some information from soccernet.  LineUps for Portugal : Ricardo; Andrade; Miguel; Valente; Carvalho; Costinha; Deco; Figo; Maniche; Gomes; Ronaldo  LineUps For England : James; Campbell; Cole; Neville; Terry; Beckham; Gerrard; Lampard; Scholes; Owen; Rooney  Both teams have potential to win this match. Portugal has the famous youngster to start Cristiano Ronaldo. Ronaldo has done a great in all three matches. He scored a 90min goal against Greece. Seems Threathening ya? Miguel Nuno Gomes being the super sub for portugal during the match against spain. He was being sub-in during the 45mins and he sinks spain with a single strike during the 50min. Deco played well as well. Figo doesnt seems threathening enough as his actions clearly shows.  As for England, Beckham the instrument, Rooney the MAN.. Gerrard, Lampard, Scholes Did Great in whole the matches except I guess the first match against france they lose 2-1 maybe due to some miscommunications and mistakes made by some of the players. England just did well but not up to my expectations. Ya That\'s about all. Afterall I prefer England. =)  After the match I will continue to update so no worries about the Results..  Let\'s Watch It  All The Best !!  **All Rights Reserved - Brooklyn** Proudly Present To You By Brooklyn-   '),(16718,3705153,'24,June,2004','2004-06-24 00:00:00','   I woke up at 11am.. I slept for only 4 hrs i guess.. After that i got nothing to do or something to work on so AGAIN I switch on my computer. Clicked on \"Shortcut to Prolink 8000\" and logged into mIRC & MsN. I waited for nich to online and carry on with our \"Big Plan\". \"ohhhhh.. Big....\" Metal Slug.. Ahhh forget it, doubt you all understand. Before he came in, I created a new friendster account for #css cause nowadays getting lesser and lesser people joining the channel. I hope this will help to invite all css students and ex-css to join the channel. Nonsense ! LOL.. HOPE lor.. Stop all these boring stuffs.. Later I going to YaoXing house first cause he wanna use my Hp to take pictures of his new dog named (Mocha). After which we will be heading towards keith house for euro2004 once again. Ok I go first, before the game starts i will give one more debrief if i have the time.  WHAT A BORING DAY RITE??   '),(16719,3705153,'23,June,2004','2004-06-23 00:00:00','   Guess what time is it now ? LOL it is 6am++ and i still have the energy to write this blog.. Ok stop all my nonsense and get on with the LATEST Euro Update for the day. Firstly I gonna say that Netherlands had done a GREAT job, they beat Latvia 3-0 with Ruud Van Nistelrooy double strike and a Great Goal by Makaay. The other match, regret to announce that the Germans are being beaten by the Reserves of Czechs. It is 2-1 to the Czechs. Czech Republic is the first team achieving 9 pts for all three matches. It rather disappointing to see the Germans packing their bags now BUT I am Happy !! Contradicting.. Hehe.. I am Proud of Netherlands But I doubt they will qualify in the semi-finals because now they are facing the stronger teams. Wish them luck. So today is the 24th of June which is also the First match of Quarter Finals..  [25/6]Portugal -vs- England [02:45]  Hope England will Beat Portugal and Do Me Proud !! (England Fan) Tips For the Day = 2-1 to England..  It\'s time for me to sleep and I\'ll be back with more updates Bye bye..   **All Rights Reserved - Brooklyn** Proudly Present To You By Brooklyn-   '),(16720,3705153,'23,June,2004','2004-06-23 00:00:00','   Ok.. I will talk about Euro2004 updates first. Until now England, France, Greece, Portugal, Sweden and Denmark have qualify into the quarter finals. Today Group D is having their last game and there is a tough challenge between Germany and Netherlands. Germany is taking on the czechs who have already qualify in the quarter finals. Netherlands is taking on a weaker team (Latvia). If Germany win, they will auto be qualified in the quarter finals BUT If Netherlands win over Latvia and Germany LOSE, then the Germans can pack their stuffs and return to their homeland. LOL.. Both teams have the potential to qualify and let\'s see who are the luckier ones.   [24/6] Germany -vs- Czech Republic & Netherlands -vs- Latvia [02:45]  **Take Note** All the timings is in 24hours mode.               **All Rights Reserved - Brooklyn** Proudly Present To You By Brooklyn-   '),(16721,3705153,'22,June,2004','2004-06-22 00:00:00','   erM.. How did this open dairy came about ? My friends and I were at Keith(One of my friend too [PAMP PAMP])house watching Euro 2004 - Croatia -vs- England. It reaches half-time and I\'ve got nothing to do so I went in to keith\'s room. Yanxiang, Nich were also in his room playing PS2 and Simon was posting his Blog. I find it sort of \"Fun\" so I created one too.. Simply Legend ya ? LOL   '),(16722,3705153,'11,July,2004','2004-07-11 00:00:00','   My mood is getting better day by day so guys don\'t worry about me. No matter what I still have to live on.. Overcome everything that I faced. I just return from wedding dinner. The groom is my eRm.. my mum\'s cousin, I guess so.. We reached there at about 8+, I was talking to my aunts and my cousins.. We always joke with one another and my aunts are always acting like small kids.. HeHe.. FUn.. I saw the bride, erM.. ok lah quite chio. She came from i guess thailand? The couple was like so romantic man.. Envy them.. Having so much fun and i guess today is their moments of their lifes. Most Beautiful Moment.. I ate very little cause I have my dinner at home.. I only eat shark fin.. I saw the happiness of others Hehe really envy them a lot.. I saw one of my ex-css school mate as well, I was walking towards the toilet and I heard my name being called. I turned back guess who was it? \"Calvin\" class 5/1. I asked him \"eh.. You also attend this wedding dinner?\" Singapore is So small.. My small brother a.k.a \"Not pain\" will be future singer.. Omg.. He went up to the stage and sang duet with one singer.. hehe Song title was \"Xiao Wei\".. Everyone was like clapping thunderously for him man.. After they wedding dinner I went to my ah ma house and my parents were talking to my ah ma.. We went home at about 12+.. That\'s all for today.. Happiness !!   '),(16723,3705153,'10,July,2004','2004-07-10 00:00:00','   These few days i have not been updating my blog because I am very very very tired when I reached home.. There are things that happening in my life although I look like those happy-go-lucky person but things sometimes cant be avoided anyway. Not everyone will know that I have problems cause I wont show on my face. Sometimes I just feel that I screwed up everything in my life and I just regret doing everything. Borned-Loser I guess, No choice my fate is this and I cant change it.. My breasted-friends please dun blame me for not sharing my problems is just that I dun wish to let u guys see the other side of me.. My job is make you guys happy and i try to forgot about my problems.. Problems are really popping out 1 by 1 and I very sick of them. I need some space to breathe. I guess no one will understand what I am trying to do.. It\'s ok.. I will be in mystery..   **Sadded** Brooklyn- a.k.a Born-ed Loser..   '),(16724,3705153,'06,July,2004','2004-07-06 00:00:00','   Today I went to Spiderman 2 with my friends. It\'s the best movie ever man. If you were to ask me \"Once more?\" NO PROBLEM.. We watched it after school. We went to Golden village at Jurong Point, we got tickets for 2035 show. Before that we went to mac and relax ourselves. 2hrs before the show we got our tickets and guess what ? It\'s 3rd row from the front, selling fast dude.. I played Winning Eleven 7 at mac while waiting anxiously for the time to pass and I can watch my hero.. Sounds like a kid.. LOL..  I watched Spiderman 1 for two times. I am kinda attracted to it. \"With Great Power Comes Great Responsibility, Who Am I ?? I\'m Spiderman !\" \"Focus On What You Want!\" Hehe.. Too bad Peter Parker cant be together with M.J because if something were to happen to her, Peter wont never forgive himself. \"You are nothing but an empty sit\" LOL those watched will know what I am saying.   \"Harry, now it\'s your turn to help me.\"  \"Revenge for your dad\" \"But he is my Best Friend\" \"REVENGE FOR ME\" \"NO!!\"  Will Harry revenge for his dad? Let\'s look forward to the next movie Spiderman 3 (Green/yellow Goblin). GREECE have won the EURO2004 as I have expected. Theodoros Zagorakis & Angelos Charisteas did Greece proud. Cristiano Ronaldo is great but is depends on teamwork to win over the other. Too Bad.. I don\'t know if tml I want to go for my ALC training.. Very Tired.. Ok I will end here.. Bye Bye Good niteyyy!!   '),(16725,3705153,'04,July,2004','2004-07-04 00:00:00','   When parents & siblings are not by my side, who else can i seek fun from ? FRIENDS.. My BREASTED Best Friends; Weijun, Keith, Yaoxing, Yanxiang, Simon. They are like my brothers and we always enjoy each others\' company. My parents and my brother went to malaysia and I was the other one at home and is kinda bored so I ask my BREASTED Friends + some girls come my house. Yesterday we planned to go West Coast Park on sunday which is today and we went to giant and get some food. They came to my house and started making \"crab mayo\". At around 11+ , Simon is here GUESS WHAT? It\'s Supper Time !! Simon sent Wendy and Elaine home first and we head towards the usual place \"24 hrs West Coast Roti Prata Shop\". I had 2eggs 1cheese and 1 teh tarik which cost me $5. We enjoyed to prata and there we go again ! Talk COCK ! LOL.. Normal lor..   After that simon wanted to go to my house awhile so we head towards my home. In the car we had a stomach exercise dude talking about some funny video clips(some sort of vulgarities). Once we reached home I instantly showed Simon and Yanxiang that particular video clip and the whole house was like surrounded with our laugther. Yaoxing brought some oldies movies to my house and we started watching with my laptop. It is DAMN Funny dude and we cant stop laughing !! It was like 3+ 4 in the morning already and they haven go home yet. After finish watching the LMAO show, Yanxiang Simon and Yaoxing went back home. Keith and Weijun Stay Overnight at my house. They two were player Winning Eleven 7, One using desktop, the other using laptop and Weijun was like giving self-commentary. Hehe ! sort of like talking to himself \"Great Goal There\" \"Hit it Wide\" all kind of stuffs. Weijun was the first to sleep ! followed by me and keith was the last to sleep, he slept at 630am.. We woke up at around 815am and Guess What Weijun woke up the Latest !! LOL.. We get ourselves prepared then I did my house chores. While I am doing Yaoxing Elaine Wendy Sarah Came !   Sian.. Don\'t feel like typing liao cause I just bathe, wearing my new contact lens( DAMN CLEAR ), tired.. Haizz.. My parents are not back home yet !! Ok then we went West Coast Park today play play play.. then FFA.. Ya tt\'s All I have A GREAT NITE WITH MY FRIENDS STAYING OVER AT MY PLACE.. BREASTED FRIENDS HOPE YOU GUYS ENJOY TOO.. HeEzZ** Thanks for the company =)   '),(16726,3705153,'02,July,2004','2004-07-02 00:00:00','   I am back from the first day of school. Hehe! I neglect my blog for 2 days already and there are many things that happen during these two days, For example ; The Mass black out in singapore + My first training for ALC and also the first training after 8 - 9 months of slacking. First is the black out ; It is around 10+ in the night and I was using my laptop in my study room. I was chatting with some of the Manchester United Fans in #ManchesterUnited. Suddenly \"BLACK OUT\", I really cant see a single thing, slowly i regain my vision and I get up from my sit. My brother was like talking on the phone before the black out then he came out of our bedroom \"eh.. yi black out go get candles..\" He lighted up a few candles and I used my light stick, is green colour somemore then made my house look kinda like a ghost house. I was waiting in front of my laptop i thought i was gonna disconnect but it didnt so i continue to play and chat with my friends. I think if I carry on I guess it would be boring so i give my last sentence. My scenary was BEST man.. 21st floor facing the north man.. That was kinda kill cool u see, this whole jurong is DARK.. Once in a life time experience. Wai jun did i TEMPT u ? LOL..   Next was the ALC training yesterday, at first i was kinda like reluctant to attend the training but in the end I forced myself to go. I get to know a lot of my ALC mates and went on with our training but today is the first day of ALC MASS training so the adviser Give Chance.. It\'s kinda tough to me but i made it through. Oh my god i was happy that i made it through because after the O levels training I have been slacking for 8 - 9 months and yesterday was my first day of exercising. We went through 3 sets of 6 chin ups, short run to orchard and ran back to school THEN this is the toughest i guess, we have a sip of water and continue stairs climbing of 12 floor up and down 3 times. My whole body was numb after those training and i expected it. Hehe ! Exhausted, tired, Numb and all kinds of shit.. LOL..  Finally today I am back to update my blog. I don wanna update euro 2004 cause I find it lame !! Finals - Portugal -vs- Greece  Oh my god this is unbelievable ! Greece the underdogs made it through into the finals. That\'s why I said Underdogs Rules Euro 2004. Today my whole body is aching but unexpectedly is not that pain. Ya I guess that\'s about all.. Hehe later I going to orchard with my friends !! Keith Wendy Elaine.....and More and more !! hehe !! LEt\'s Chill Out !!   **All Rights Reserved - Brooklyn** Proudly Present To You By Brooklyn-   '),(16727,3802511,'30,June,2004','2004-06-30 00:00:00','   So here it is, my Blog!.. Thanks to Ray for introducing me to this, hopefully now I\'ll be able to streamline my BBOSA which I know has been very few and far between lately!.. I haven\'t got much time at the moment to add stuff but WATCH THIS SPACE!!!..   '),(16728,3802511,'30,July,2004','2004-07-30 00:00:00','   So last week Pauline and I decided to take Paige out for a girl\'s night out for her birthday.. Originally it was going to be a dinner at the  urlLink Rooftop Gardens  in Kensington (very posh - owned by Richard Branson) but when Kris (a girl I met here through other friends who actually lived in Innisfail for a bit and went to Good Counsel as well!) found out that there was a speed dating night being held there, we couldn\'t think of a more fitting theme for a girl\'s night out!!..   so off we went charged up on a few mojito\'s from bar cuba, we ascended to the idyllic gardens 7 floors up from the busy Kensington High St.. amazingly beautiful setting!.. Kath and I sat out for most of it, just taking in the hilarity and weird social setting of the speed dating!!.. funny funny night..   '),(16729,3802511,'27,July,2004','2004-07-27 00:00:00','   So.. the big news of the week is that I have decided to move in with Blake.. *gasp*!.. since my other flatties Pauline and Sam are wanting to find a better/cheaper place, and since a bigger room has become available at Blake\'s pub ( urlLink The Prince of Wales ), I\'ve taken this opportunity to fulfill my lifelong dream of having a pub as my lounge room!!.. and of course to spend more time with my nearest and dearest boy!... ahhhh...   '),(16730,3802511,'15,July,2004','2004-07-15 00:00:00','   After a week or so of my work server banning me from my blog, somehow I have made it through!.. hip hip hooray!!..   '),(16731,3802511,'15,July,2004','2004-07-15 00:00:00','   urlLink Click here  to go to my online photo album of Menorca!..    Cala en Porter    Calas Coves    Playa\'s    Me & Blake @ Cala Mitjana    Blake the golfer    Kill that ball!!   '),(16732,3802511,'13,July,2004','2004-07-13 00:00:00','   While you\'re waiting for me to get my ass into gear and get all my photos uploaded, thoughts gathered, etc... check out Ray\'s take on the Menorca trip on his blog -  urlLink click here   '),(16733,3802511,'02,July,2004','2004-07-02 00:00:00','   So here I am.. sitting at my desk at work in London.. I\'ve just checked on my blog page to see if anyone has commented in my absence and I\'ve just realised that the photo I posted there as an example of a Menorcan beach is actually the beach we went to yesterday before flying back to this urbanised habitation.. Cala Mitjana, I think the most beautiful of beaches we visited (though with so many and my head still clouded with brilliant sunshine and crystal turquoise clear waters it is hard to judge).. Only 16 hours ago I was jumping off that rock ledge and swimming in the caves shown in the photo in the previous post.. Oh how the memory is so fresh but oh so far away....   '),(16734,3802511,'02,July,2004','2004-07-02 00:00:00','   Only 3 days until Blake, Ray, Macca and I take off to the paradise that is Menorca... a week of lying on the beach in the glorious sunshine.. I\'ve just checked out the  urlLink weather report .. Oh I can\'t wait!!..   '),(16735,3802511,'01,July,2004','2004-07-01 00:00:00','   Okay.. so I\'ve finally got myself a working photo hosting site, so I can link some glorious photos to my blog... once again, i haven\'t got much time at the moment, but here is just a little snippet of my cinque terre experience..  Blog on!..   '),(16736,3802511,'02,August,2004','2004-08-02 00:00:00','   Last weekend was the hottest and sunniest weekend of summer yet.. and what do you think i was doing??.. playing in the park?.. swimming at a beach?.. sunbaking on my balcony?.. NO!!.. I was tucked up in bed UNDER A BLANKET delerious with fever, aches and pains with that stupid flu going around!.. oh yes so happy.. and so happy to be back at work on Monday having used my weekend so fruitfully!!..   '),(16737,4028811,'31,July,2004','2004-07-31 00:00:00','   I start saying this words two days ago..when my day begin lil bit overload. And those words really made my day, increasing my motivation facing more brighter day. As usual, works for me means doing routine but still with high volatility..(confuse?..routine means doing the same work everyday. For me it means, reporting almost the same business news in town. Volatile for me means, find more tidal circumstances in town..sometimes go slow..but sometimes go fast).  Routine always boring and volatile sounds challenging..i\'d rather face the second one actually, but i have to admitt that sometimes i do grumbling about this condition to my friends.. But two days ago, my brain recall this words from my memories..and allow me to think about this word again. I can do It ..i exclaimed it to myself.. \"just take it easy, do everything one step at the time and allow yourself become a problem solver. If everyone around u able to finish each problem well, then u can do better than them\" And now, nothing seems impossible for me..no matter how huge my problem is, i convince myself that i\'m  able to solve it..and today will be brighter than yesterday, tommorow will be brighter than today :)   '),(16738,4028811,'30,July,2004','2004-07-30 00:00:00','   I feel very lucky to have lots of friends who always wake me up, when Iâ€™m about to give up..Thanks to Irfan with this story. Pygmalion Pygmalion is a youngman who talented in art sculpture. Heâ€™s really proficient in carving sculpture Heâ€™s handcarving is really remarkable. But not by his talent, he conceded and loved by friends and neighbors. Pygmalion known as a positive thinker. He always consider everything in positive angle. When the city square become wet and muddy, and all people grumbling. Pygmalion said, â€œThank God the other field is not as wet as this fieldâ€ When a sculpter buyer looked persistent in bargaining price, and Pygmalion friends whispering,â€ how stingy that person is,â€ Pygmalion said,â€Perhaps that person still need to spend the money on something more important,â€ When kids stole apples at his yards, Pygmalion never complaining. He feel sorry to see those kids and said,â€Poor children, they donâ€™t get proper education and food at home,â€ Thatâ€™s Pygmalion  sight. He never see any circumstances in bad side, but on the good side. He never think bad about others;in the contrary, he try to imagine some good side from people\'s bad manners. One day, Pygmalion carve a women sculpture from wood. His hand carving is really smooth and soft. That sculpture size is equal to the human size. The face of the sculpture is really charming and has a magnificent smile. The sculpture body looked very impressive. Pygmalion friends said, â€œItâ€™s just a sculpture, no matter how wonderful that thing is, itâ€™s never be your wife,â€ Many times Pygmalion looked deep at the statue, and rub its surface, treat it like a real lady. God and Goddess of the Olympus watch him, and really appreciate what Pygmalion done for the statue, and they decided to grant him remarkable gift. They transform the statue into a real lady..thatâ€™s the end of the story..Pygmalion live happily ever after with the most lovely women in the whole land. Now, Pygmalion remembered to describe the effect of positive thinking pattern. Whe we think positively about one circumstances or someone, the result really positive. For instance, When we try to be friendly to someone, then that person will be friendly too. When we treat our children as smart kids, they will become smart kids at the end. If we believe weâ€™ll be succeed, bigger possibility that the effort is half way of success The effect of this positive thinking, called Pygmalion effect. Our mind sometimes have fulfilling prophecy effects, both positive or negative. If we want to consider our neighbor as shabby person, so we never want to socialize with him/her, so at the end..they will become a shabby person to you. If we suspicious to our children, at the end they will really dishonest to us. If we  gave up on something and think that we never able start an effort, bigger chance for us to face real failure. Pygmalion thinking pattern, is to think, expect and hope only good things about conditions or people. Imagine how big the effect is, if we thought in positive way like that. Weâ€™ve never come into prejudice on others. Weâ€™ll never gossiping bad stories of others..and we never expect anything bad about others. If have bad thinking on others, we always find a reason to guess something bad about them. If someone give us a present, obviously itâ€™s a good deed. Bad if we think in bad way, weâ€™ll become suspicious.â€Perhaps he/she try to bribeâ€ or may be we come to grumbling side â€œaahhh cheap onesâ€¦â€  Who get the disadvantage of this negative thoughts is ourself. We come easily suspicious, and canâ€™t be happy                                                                  In the contrary, if think in positive ways, we get easily enjoy the present and joyful with the gift, â€œohh our neighbor is a generous person, though busy as a bee, still remember giving such nice present to us,â€.. The color of life, perhaps depend on the spectacle we wear. If we wear some grey one, everything will looked grey. Life will be grey and blurred. But if we wear the bright one, everything will looked clear. Spectacle which full of prejudice or hate, will only lead is to suspicious, envy, hatred life. But the peace one, make our life bright and comfy. Life will be perfect if we see it in a good side. Think positive about ourself Think positive about others Think positive in any circumstances Think positive of God (God will grant us everything as you expect from him) The effect of positive thinkin weâ€™ll realized soon. Warm family Trusted friends Friendly Neighbor Works seems fun World seems friendly Life seems lovely Just like Pygmalionâ€¦thatâ€™s it Have a great life, everyone :) Gee...my english still bad...need lots of help   '),(16739,4028811,'28,July,2004','2004-07-28 00:00:00','   These are good advice from a friend, i like each word he expressed..Perhaps..my writing isn\'t perfect but his words are 1.To My Friends Who Are...........SINGLE Love is like a butterfly . The more you chase it , the more it eludes you. But if you just let it fly, it will come to you when you least expect it. Love can make you happy but often it hurts , but love\'s only special when you give it to someone who is really worth it. So take your time and choose the best.  2.To My Friends Who Are............NOT SO SINGLE Love isn\'t about becoming somebody else\'s \"perfect person.\" It\'s about finding someone who helps you become the best person you can be. 3.To My Friends Who Are.............PLAYBOY/GIRL TYPE Never say \"I love you\" if you don\'t care. Never talk about feelings if they aren\'t there. Never touch a life if you mean to break a heart . Never look in the eye when all you do is lie. The cruelest thing a guy can do to a girl is to let her fall in love when he doesn\'t intend to catch her fall and it works both ways...  4.To My Friends Who Are............MARRIED Love is not about \"it\'s your fault\", but \"I\'m sorry.\" Not \"where are you\", but \"I\'m right here.\" Not \"how could you\", but \"I understand.\" Not \"I wish you were\", but \"I\'m thankful you are.\"  5.To My Friends Who Are............ENGAGED The true measure of compatibility is not the years spent together but how good you are for each other.  6.To My Friends Who Are............HEARTBROKEN Heartbreaks last as long as you want and cut as deep as you allow them to go. The challenge is not how to survive heartbreaks but to learn from them .  7.To My Friends Who Are............NAIVE How to be in love: Fall but don\'t stumble, be consistent but not too persistent, share and never be unfair, understand and try not to demand, and get hurt but never keep the pain.  8.To My Friends Who Are............POSSESSIVE It breaks your heart to see the one you love happy with someone else but it\'s more painful to know that the one you love is unhappy with you.  9.To My Friends Who Are............AFRAID TO CONFESS Love hurts when you break up with someone. It hurts even more when someone breaks up with you. But love hurts the most when the person you love has no idea how you feel.  10.To My Friends Who Are............STILL HOLDING ON A sad thing about life is when you meet someone and fall in love, only to find out in the end that it was never meant to be and that you have wasted years on someone who wasn\'t worth it. If he isn\'t worth it now he\'s not going to be worth it a year or 10 years from now. Let go.....  11.TO ALL MY FRIENDS....... My wish for you is a man/women whose love is honest, strong, mature , never-changing, uplifting, protective, encouraging, rewarding and unselfish   '),(16740,4028811,'27,July,2004','2004-07-27 00:00:00','   I wanna share a story about elevator in my office. since it\'s not a big office, so there\'s only one lift here. But, this lift already took lots of victim who get trapped in that lift..from common employee, office CEO, stakeholder, former Indonesia president Gus Dur, and last night Krisdayanti.... We call that lift as a wingit lift..wingit means lil bit spooky. The funny thing is, most employee never want to use that lift, we prefer use stairs, so the exclusivism of that lift begin more and more obviuos. Only special guests use it...and more special the guest is..the lift more eager to trap those people in... :))  Perhaps, it\'s the way wingit lift want to demonstrate its loneliness..never being used properly and daily by all people..and perhaps, it gets bored to serve those special guests..:p so Wingit lift wants to scare them :D hahhahahah that\'s my story about this wingit lift...and...i never use that lift since i work here.     ind   '),(16741,4028811,'26,July,2004','2004-07-26 00:00:00','   I know, i should not get carried away again. \'coz it\'s only crippling me .. weakening and think something fool again. But i do miss u, perhaps God won\'t mind if i send this ...perhaps it will help friend who miss their mom too...(just a hope) D U Knw Y Mom Always Smile..... \'Coz she wants us to face brighter day D U Knw Y Mom Always Weep..... \'Coz she try to convince God..we deserve the best D U Knw Y Mom Always Laugh..... \'Coz she wants us to taste a piece of heaven with her D U Knw Y Mom Always Cry..... Those are tears of happines when she walked us through the door of success ... Taught us well about love Taught us well how to fight too.. Yeaa..Fight and conquer this world with this two hands Convincing we could grab those stars..with this two small hands Build our hope to c as many different skies as possible Bring those blue sky to keep within... ... Need more than pen to write Need more than tears to express Need more than strength to build all she\'s done... ... Light up again when everything goes dark Brace again when i\'m about to loose Refresh when get too exhausted Bring Fresair to heal all wounds   '),(16742,4028811,'25,July,2004','2004-07-25 00:00:00','   Why fresair, it\'s just perfectly all people need..and what i need to change something inside my head :) Freshair to breatheâ€¦.to fill this brain again with pure oxygen and make it work properly A friend of mine, told me why u donâ€™t try to write diary or something similar with it..but I always say, I never try to share my feelings to others, if they see me smile then it means Iâ€™m happy but when I begin more quite it means, I donâ€™t wanna talk about it. And it seems to be enough for me..for all this years.  But after I read her posting today, I begin to think..perhaps thereâ€™s something wrong inside of me that I should change, now here I am writing something. Pak-I, thanks for introducing me to this blogger..ïŠ   '),(16743,4028811,'14,August,2004','2004-08-14 00:00:00','   Month of August, should be have special meaning to our people. Though the respons isnâ€™t quite the same like the celebration fourth of July. But I always notice something special about this month, yeaa make my love upon this nation grow even bigger. Albeit many people mocking or insulting this country failure holding up their image, in front of its own people. Either when all nations questioning our ability to recover, my love to this nation is raise greater. Donâ€™t bother to deny that this country indeed have major problems in all cases. Merely when we found the spirit of itâ€™s people struggle for better life, my love heave up to the higher level. One thing for sure that make me love this country is the people Iâ€™ve seen and lived with. I was amazed and still, every time I saw Mrs Karta her husband and their only grandchild carry more than 50 kg bananas and cassava on their each back, few hours before dawn to the traditional market which only 8 km away from their home. The fall of wet dew and soft cold breeze become their closest friend, to begin their day. I know this family better, cause sometimes Mr Karta give a big help at my place dealing with all unwanted grass at the yard. In his nearly 80s, work isnâ€™t a burden. In fact, he thought his muscle will start aching if he stays at home n doing nothing. Same quote I found from mrs Kartaâ€™s too. Though I believe their income from selling bananas and cassava ainâ€™t enough to provide good meal for the entire family. But I hardly found them grumbling about this condition, in fact they always remind me about being â€œnrima ing pandumâ€. Which means accept each role that God given to them without complaining. They believe that God already create a big plan in their life, so they just fit in with it and do the best to adjust with it. Adjust means, stay humble and do better their role in this life .  This two old couple give me a clear vision, that our people ainâ€™t lazy or obstinate as others said. Sadly, sometimes our own people concur with this hilarious idea, and also pointing their finger at their own brothers. Sounds pathetic but it happen, while Iâ€™d rather see everything that happen on this country in a clear and bright spectacle. To watch single tiny effort from everybody around me try to fit in their role in life, indeed make me more in love with this people. I hardly found any despair in their eyes, I only found stream of wet sweat falling from their forehead and their hands. They only offer something positive, hope that theyâ€™re doing the best that they could. How can I not be fall in love to this people? I would become a foolish if I dare to underestimate my own people who encourage me everyday to play to best role in life. In this August, I try to remind you, that weâ€™ll never able to manipulate our identity as a mighty and sedulous Indonesian. I proudly said, that Iâ€™m 100% indonesian, 25% javanesse, 25% aceh+malay, 25% borneoâ€™s dayak, 25% non-native immigrant from arab. That will be a great reason why i couldn\'t get easily drawn into ethnocentrism. And my biggest intention is to meet each of my brother out there, to show how proud and lucky I am to have their blood in me. As well as I will try the best to make them proud for sharing their blood in me ïŠ   '),(16744,4028811,'10,August,2004','2004-08-10 00:00:00','   This afternoon a friend of mine remind me again about the different between rose and jasmine. She told me, rose performance always stunning, every single eyes ardent to touch when they start blooming. Thatâ€™s common, since their color is really gorgeous and sweet. But one thing, which sometime we got easily slipped while looking their exquisite color, right after we take a good look at them and sense them within the breeze. Scent of rose doesnâ€™t spread as vast as the scent of jasmine, though we find one tiny jasmine about to bloom, we easily sense the scent of it. One tiny jasmine could make roomy space specific fragrance which we could easily notice. Sometime I thought, how we could be like jasmine, flourished entire space only with our tiny appearance. Sounds nice if we could give a meaning to any environment with a delightful fragrance, and they donâ€™t have to touch us or stunned with our outside appearance. They just know our presence but donâ€™t have to say something good about anything that weâ€™ve done. See, doing something good but never care about compliments or any flattering statements from other. Sometimes, compliments only weakening and bring us nowhere than keep satisfied about what we have done. Itâ€™s hard to stood back when weâ€™re still floating on the air and feel contented with it, joyous a lot with good stuff weâ€™ve done and forget that every achievement means a challenge to double it twice, three times or more.   '),(16745,4028811,'09,August,2004','2004-08-09 00:00:00','   I remember what James Gwee said about few characteristic of a ditherer, and I couldnâ€™t deny that some of those could easily found in me. Worries a lot, unsure and afraid in all circumstances, always imagine the worst and trying the best that I could to avoid failure. Always thought that life is all about reducing uncertainty as much as I could. Hence, with this characteristic he put me on a ditherer  side. But some of his words is not quite fit in me, something like thoughts that never satisfied, seldom happy and always complaining. Those later characteristic rather hard to find in me, since Iâ€™m kind an easy going person, always put smile on my face, and hide all problems deep inside so no one can sense it. So, what am I know? Still a ditherer, or what?..actually I donâ€™t care too much about this, but there is one of his note that make me curious, he said ditherer always see the ground as the safest place..and ditherer favorite color is brown. It lil bit shocked me up, since I do think that way. I thought the ground is the safest place but I donâ€™t scare of heights either. And one thing for sure is, I really like brown color in all cases. My outfit mostly in brown, stationary, fancy stuffs, everything dominated with brown. But did the color really imply who we really are? I never expect to be a ditherer, Iâ€™d rather become someone who has positive thinkin on everything like Pygmalion. Left the words â€œwhat ifâ€¦â€ behind and drop all worries away. Would it be a mistake if we try to seek more security? I mean, that will be great if our plans work out well and get the best achievement by reducing any tiny obstacle that might occur in our way. Itâ€™s just my thought, or perhaps Iâ€™m really a ditherer inside and outside so it flows all over my bloodâ€¦:)   '),(16746,4028811,'08,August,2004','2004-08-08 00:00:00','   One day, this woman come home from her work and saw three old men with long beard were sitting in front of her nice and cozy yard. She doesnâ€™t know who they are, and asking â€œI donâ€™t know any one of you, but Iâ€™m certain that you might be sturfing right now. So, lets go inside Iâ€™ll cook something for you,â€. But one of the old men ask back,â€ is your husband at home ?â€. She replies ,â€No, not yet, he still out for work,â€.  The three old men said again,â€ Then weâ€™ll wait until you husband gets home,â€ they said that in peaceful smile. In the evening, the whole family gathering again in that house, and the lady told her husband about the three wise old men outside their yard. For a moment, the husband little bit confuse about the story, and then told his wife to call the three old men outside, to come in. Hence, the lady go ask the three old men to come in, but instantly they reply,â€ we canâ€™t go inside together, you have to choose one of us, one by one to invite us in,â€. The lady look with astonishment, then ask,â€Why..?,â€ One of three old men explain to her,â€ His name is fortune, and the man who sits beside him is success, and my name is  compassion. Thatâ€™s why you should choose which one of us, who should enter you house first. Go ask your husband, who he need to invite from us most,â€. The lady leave again to tell her husband about this uncommon circumstances, and again the husband feel amaze.â€Wow, this is unbelievable, Iâ€™d rather invite fortune, so he can fill this house full of fortune my dear..,â€ the husband said. But his wife refuse this idea, and told him that sheâ€™d rather invite Mr, Success to come in first,â€We need success in our farm harvest, why donâ€™t we invite Mr Success first?,â€.   But in coincident, their dearest children  heard about this argument and said,â€That will be great if we invite Mr Compassion to come inside. Our house will feel more cozy and warm with compassion,â€. Then the parents agreed with their beloved childrenâ€™s  idea, and decide to invite Mr Compassion in. The lady go outside again to ask a men named compassion to come in, then Mr Compassion stood up, walk to the house. Suddenly, the two other men, also rise up and walk along behind Mr Compassion. Feel curios about this action, the lady ask again,â€I only invite compassion, but why fortune and success also come along?,â€.  Those two men answering in the same time,â€ If you just invite fortune or success to go inside, you will only have one of them, and the rest of us will stay outside. But, since you invite compassion, in whatever place he goes, we will walk behind him. You should know, that both of fortune and success were blind, only compassion who can see. Only compassion who can lead us to, the right path of life, and good deeds. Therefore, we need his guidance when we walk and running this life,â€. Said Fortune and Success in the same voice.    No matter how huge our intention is, but never leave the essence of our life, compassion, love, and care for other   '),(16747,4028811,'02,August,2004','2004-08-02 00:00:00','   Love is like butterfly..the more you chase it..the more it will elude you. But if you just let it fly, it will come to you when you least expect it. Nice quote right....?? I guess so, but sometimes really confusing when we accidently feel certain chemistry reaction on someone..but the other person we least expect make an approach. That\'s different circumstances, but lil bit related too.. See..everything don\'t exactly go the way we wished to be. Perhaps, letting it fly would be better. So i can lose it forever or it will come to me when i least expect it..sounds pathetic :(   '),(16748,4198810,'12,August,2004','2004-08-12 00:00:00','   Ah, Friday.....that lovely day has yet again come along, and the week\'s drawing to a close. As always, my mind\'s not quite here at the office with the rest of me, as witnessed by the fact that I\'m blogging when I really should be finishing work on that file thats been sitting on my desk all morning.  Got to the office and got an sms from a dear colleague of mine, H. Horror of horrors, dufus is at it again. This time, in his sick little mind, he\'s decided that he, as head of department, can stake a claim to crediting himself for our work. We\'ve been involved in this major telecommunications deal for ages....H has been doing a bulk of the work and dufus came in far too late to be involved in any significant part in the project. Besides, with his kinda intellect, even if he did start it off from the beginning, its unlikely he would have contributed much anyhow. Think 1st generation ATARI PC and you\'ll get the general idea where dufus\' intellect lies. If he were reborn with his current brain, he\'d likely be an amoeba............  In any case, it transpires that H managed to take a peek at dufus\' CV and guess what? He\'s claiming he \"lead the team\" that drafted those landmark agreements! Can you beat that? What the heck does he mean by \"lead the team\"??? The only thing this man could lead is a horse drawn carriage.....with him as the horse! And even then, with hi puny little physique, I doubt he\'d get very far.......  You know, I really, really, really wish I could somehow rig his chair with some form of microwave generating device that would cook him to a crisp in under 5 minutes at a push of a button.  In any case, he\'s a nutter that deserves to be given the boot, but alas, its not the Firm\'s style to kick out unwanted baggage.......  If theres a bright side to all this is the fact that I have 2 clear days to forget about the numbskull.  Ah...the weekends..ain\'t it a lovely concept? You know, whomever thought that it\'d bea  good idea to have 5 days of work and 2 days of rest needs a slap on the behind. Why couldnt it be the other way around? 5 days rest, 2 days work...now THAT would be just SMASHING!  Its odd though - you look forward to the weekend with anticipation akin to a guy waiting for his bride-to-be to say \"yes\" to his proposal, and yet, when it does eventually swing by, you just wanna do NOTHING!  I have this friend, lets call him Mr. A. He kinda illustrates what I mean. All weekend, the boy doesnt even make it out his door! He sleeps, eats, goes back to bed, watches a bit of TV, eats again, sleeps, then wakes up the next day on Sunday to repeat this insane cycle of perpetual laziness. OK....so maybe not everyone shares my enthusiasm for doing stuff and filling my weekend schedule with stuff you want to do rather than stuff you HAVE to do to get paid, but seriously, how can ANYONE do NOTHING thru the weekend?  To be fair to Mr. A though, he\'s recently broken away from his inane weekend schedule. He\'s gotten involved with someone, you see, and when a guy starts to date, well, his life kinda changes. Now his weekend is something like this:- Sleep with girlfriend, eat with girlfriend, go back to sleep with girlfriend, watch a little TV with girlfriend, go back to sleep with girlfriend.....you get the general idea........  I thought I knew this bloke well enough, but since he got involved, he\'s become this little tyke - like my nephew - the kinda kid you feel like smacking hard with a thick hardback book cos he never listens and always has smart-ass remarks and excuses to throw back at you  when you wanna give him good advise.  Mr. A used to be known as the noble, loyal, trusted friend - a pessimist, but nonetheless, the kinda guy you\'d go out of your way to help out if he ever needed it cos he\'s always been nice to you and has never, in the history of my college year, pissed anyone off. And thats bloody rare in my college cos EVERYONE pisses of everyone else once or twice a year - its almost tradition!  But now A has taken on this arrogant air about himself. Guess scoring a gal after 10 years of dating no one except Susie Palmer and her 5 sisters changes even the most humble of people.  This is a man who can\'t drive, didnt get a driver\'s license until he was 31 yrs old, refuses to take delivery of a car he ordered cos he\'s having last minute regrets about taking a car that won\'t take 7 people at a go (he forgets cars here can only legally carry 5), has boobs bigger than some women I know and comes to the gym in his pyjamas (I kid you not!)......its hard to pity a man who has all these characteristics and then adds a large dose of new-found arrogance to the soup.  This weekend his girlfriend has booted him out of the apartment cos her mother is visiting. Call me insane, but if I dated someone who did that to me, I\'d take offence to it. She doesnt even want him to meet the mother, if you can imagine that. So after a week\'s worth of being sarcastic with the regular guys he hangs out with, me inclusive, and rubbing in the fact that he has a girl who he needs to kep company 24/7, he\'s decided he needs our company this weekend. I say Phooey.......dunno about the rest of the gang, but I don\'t see myself as standby company for someone to hook up with when one\'s girl refuses to let you see her parents......  '),(16749,4198810,'11,August,2004','2004-08-11 00:00:00','   Ever wonder why people aspire to owning german automobiles? Its odd, isnt it? These cars - BMW\'s. Mercedes-Benz\' and Audi\'s - all do pretty much the same thing your good ol\' Toyota does....except its double, if not triple the price.  This train of thought left the station cos a friend of mine just splurged 150 big ones on a 15 year old Mercedes! Why? I honestly don\'t know. Prestige, they say....the satisfaction of driving something engineered by the same people who almost took over the western world.....  Thing is, for that kinda money, there are a heck of a lot of brand new cars at his disposal. But he didnt want Japanese. Didnt want British either.......DEFINITELY didn\'t want local....OK...after having a brand new Proton act up on me the day I took delivery of the blasted car, I can perfectly understand the last write off, but Japanese?  I like JapMobiles. They\'re usually cheap (though anything above 90k can hardly be called cheap in my eyes), cheerful and go on forever and ever an ever as long as you stick to their service intervals. They\'re very well oiled machines, and that perhaps is their ultimate failing, as strange as it seems.  This is my theory......when you take something for granted - that the car will start first time, everytime, and will get you from A to B in absolute comfort and minimum of fuss, you tend to forget the machine that makes that possible. Drive any modern Japanese car and you\'ll understand. They\'re so utterly efficient that they are deviod of any characteristic that endears you to them.  Wanna test this theory? Drive a Toyota Corolla fo 30 mins. Then hop into a Honda Civic for 30 mins....then hop into a Mazda 323. They drive the same, feel almost the same and in many ways, look almost the same unless you have a keen eye for detail required to tell them apart.  On the other hand, \"continentals\" as we\'ve come to term them, are utterly unreliable, expensive to maintain, temperamental, fuel guzzlers and often have baffling interiors. Their equipment can equally be as baffling as the cars themselves. Whomever thought it was a good idea to put HEATED mirrors in a tropical - spec Citroen Xantia really, ought to be shot......Same goes for cars from Mercedes.....what kind of \"luxury\" car makes do with vinyl seats? They call it \"MB-Tex\"....fancy name for \"taxi seats\", really......  Fact is though, these quirks make us remember these cars. They leak oil, guzzle petrol, have air conditioning that doesnt condition air at all....but these are things we love. It gives the car \"character\", as car enthusiasts tell me.  And of course, living in a typically asian society, a person\'s worth is measured directly by his posessions, and having a \"continental\" car parked in your porch can sometimes mean the difference between being accepted by your future in-laws and getting the boot out the door by the very same people......drive a Mercedes, no matter how old, you\'re brilliant son-in-law material...the type mothers would pay you to marry their daughters. Drive anything Japanese and you\'re \"trying to make it\".......perfectly acceptable, but you\'ll always be subject to the \"upgrading anytime soon\" questions and looks......  For me, though, if I had 150k, I\'d be bloody happy with a Honda CR-V in Zircon Blue Metallic...........;) To hell with what others think. They don\'t like what I drive they can catch a cab - likely fitted with \"MB-Tex\" upholstery too - LOL!  '),(16750,4198810,'11,August,2004','2004-08-11 00:00:00','   Ain\'t it odd whenever you get up, its a thursday, and you wish it wasn\'t? If I were born as a day in the week, I\'d really hate being Thursday.....or Monday for that matter. Thursday\'s always that pesky little day you have to get over to arrive at Friday, which is when the REAL party starts.  Honestly, I don\'t even know why we bother showing up for work on Fridays. No one gets anything done on Fridays, for the simple fact that our minds get into warp 9 from the word go and zooms right into the weekend leaving our bodies to go mouldy sitting down in the office.  Assessment day is over, thank goodness! One thing I\'ve learnt from last night\'s assessment video.....NEVER, EVER watch yourself on playback when you record yourself. Its perhaps odd, but although I stand a pretty tall 6ft 3.5in, I\'ve never seen the world as a small place where a person my size would dwarf anything. I mean, I used to drive a Kancil, for crying out loud, and I felt the car was perfectly fine. If you don\'t know what a Kancil is, well, picture a one-seater lawn mower with a roof and air-conditioning and you\'d not be far off ;) But seriously, looking at myself on playback, the bike, itself a solid cast metal contraption standing at least 3 feet tall and weighing a good solid 75kilo\'s, looked like a kid\'s tricycle beneath me!  I was really touched though.....some members of my class who normally didn\'t attend my Wednesday classes took time off to head down to class just to give support to their gargantuan instructor for the night. Some of them I actually thought of as complete asses till last night.  Theres this one guy, Mr. X, who annoys me as much as flies do when you\'re trying to have lunch. The type thats loud, brash and smells oddly like durian even if you\'re 6ft away from him . If you wonder what Durians are, its a fruit revered here in Malaysia as the \"Kind of Fruits\", though a friend once put it quite aptly into descriptive words as \"the best textured pudding eaten in a very badly kept toilet\". He wasnt very far off, truth be told.  But anyway, Mr. X actually took time off work, drove halfway across town and fought KL\'s infamous gridlock to get to class to show his support, though I thought his whole \"thumbs up\" action at the end of the assessment in front of the video cam was a bit OTT....;) But that kinda made me realise one thing - people you want to write off for the smallest faults often can surprise you in the biggest of ways.....  The opposite, of course, can be true as well, that of people you hold in high regard only to realise that they are complete asses.  Take one example - my Head of Dept.....popularly known as \"Dufus\" in my office. We had such high hopes for this man when he was hired to take charge of our department. That illusion however, was quickly shattered when he started snoring at his cubicle within the first week at work. I kid you not. The man had, somehow, somewhere, mastered the art of sleeping in vertical alignment on his office chair without keeling over. Thing is, if he limited that whole act to the office, it\'d not be a problem, really....I mean, maybe the man has an over-active sex life (though if you looked at him you\'d never believe it - he\'s best described as looking like a confused guppy at feeding time) and can\'t catch a wink of sleep at home, but noooo......the man has to fall asleep even when we\'re at meetings with clients! Its awfully humiliating to justify to clients why your boss is snoring during a meeting with the top management staff of the company present. Amused, they\'re certainly not. Neither are we......  In any case, getting back on track, we expected a whole lot from this man, who once owned his own firm prior to joining us here at the Firm (I capitalise this name, cos its very much like the movie namesake....once you sign up, you\'ve sold your soul to the Devil). Let\'s just say that we were mildly dissapointed. OK...maybe not mildly. This is a man who claims that software can be patented, for crying out loud! In the US, maybe....but any first year law grad could tell you thats BS.....I think our clients realised that too....but despite our constant protestations to our senior partner about the state of stupidity that our head of dept is obviously a permanent resident of, he remains in the Firm.....and constantly makes life a living hell for all of us in the department.  If there\'s a bright side to all this, is the fact that we don\'t really see him much these days in the office. Marketing. What a convenient term for \"I\'m sleeping in late/going out for an extended lunch with friends\". He often disappears for days on end, but really, I\'m not one to complain. The less I see of dufus, the happier my life is in the office.......  Gloomy day again in KL.....haze is back and so is my sore throat....the day\'s only beginning. My body is here in the office, but my mind is really in bed, saturday morning, curled up in cold sheets kept icy cool by an air conditioner set at maximum chill....ahhh....now THATS life!  '),(16751,4198810,'10,August,2004','2004-08-10 00:00:00','   Day\'s finally drawing to a close.....as 5pm swings around, all I can think of is whether I\'m gonna royally screw up the assessment tonight.....  Got e-mails from old college friends, as I always do at this time of the day....seems everyone is either married, with kids or dating....suddenly I feel so bloody old!  Its a funny thing, being single. You love it because of the freedom it gives you. I mean, think about it - you can be as disgusting as you wanna be, hang out with whomever you wanna hang out with and do whatever, whenever without somebody howling down the phone line threatening to rip your CD collection apart if you\'re not with him/her within the next 20 minutes.  But on the flipside, you realise your friends whom you hang out with aint gonna be single forever, unless you do everything within your human powers to keep them single. Thoughts of telling gals that my best buddy is gay or telling your best friend the girl he\'s dating is a pshycho stalker chick may come in handy, if thats the game you wanna play....  But no, thats not for me....A lot of blokes I know get involved cos its \"natural\". When asked why they are dating, they just tell me that its the \"natural course of a man\'s life...\". I disagree.....afterall - whats \"natural\" anyway? Who decides whats \"natural\"? Your mom?  I know my mom would kill to see me with a gal right now. I\'ve been single for ages, and truth be told, I can\'t imagine myself dating anytime soon - just too much I wanna do with my life right now. Its like when you go on a short dream vacation - you pack only whats really necessary - and right now, a relationship for me is about as useful as a tank top on a skiing trip to Norway.  You know....that Carrie character in Sex & the City may be on to something afterall.......  '),(16752,4198810,'10,August,2004','2004-08-10 00:00:00','   Funny thing is, I\'ve actually been looking forward to today for ages! Well, at least since January, when some silly obsession posessed me to take on training for RPM....a programme that I was so obsessed about I would throw fits if I missed class!  Anyway, today\'s basically assessment day. I\'ve got to video myself (UGH! Hate self videos - what they say about cameras adding 10 lbs is a lie - it adds a good 20!) and submit the danged thing to the guys down under (no, not basement level of the gym, but the wonderful land of the Koalas and Kangaroos).....  Spent the whole of last night just going mad trying to perfect the choreo for the tracks I\'m doing.........:( Well, hope all goes well tonight...  Still stuck at the office now though. On hindsight, maybe I should\'ve just taken the day off.....my body is here in my cubicle, but my mind is elsewhere.............  Watched this silly marathon run of sex & the city last night on cable.......Honestly, its my first time watching the series and I already have issues about those 4 women.  Carrie is the one I really wanna hit her over the head with that book she constantly seems to be working on. I mean, c\'mon......whats her hang up about letting her friends have their own fun? She seems single mindedly bent on making their lives centred around her. This one episode was hilarious - she was trying to get the rest of them to join her and Samantha in Atlantic city for a \"girl\'s night out\".....when the other two didnt wanna go cos they had LIVES to lead, she did everything in her power to get them to go.....I dunno bout you guys, but theres something seriously loony about a woman sho doesnt understand the concept behind the saying \"..get your own life, for crying out loud!\"  And dont even get me started on Samantha....I mean, seriously, what the heck is the woman\'s problem with men and jewellery? She has this obsession with this Richard character...obviously rich, and she\'s obviously loving it not for the person.....I mean, have you seen the man? If he was a jar of jam, he\'d be well past his sell-by date (but then again, Samantha\'s seen better days too...). Anyways, still, the woman is so insecure about herself that she runs 11 floors up the hotel emergency stairway in a pearl thong just cos she\'s insanely suspicious that her man is sleeping with the room keeping lady.....if samantha was a friend of mine, I\'d do anything in my powers to get her admitted for a psychiatric evaluation.......Carrie too, if I have space in the trunk of my car......;)  Just cant wait for the movie \"Brokeback Mountain\" to swing by. An Ang Lee production, its an odd storyline to have Heath Ledger and Jake Gyllendhall (not sure of the bloke\'s last name - some names are just un-spell-able) star as .....if you can believe this one....gay cowboys.....  I think Hollywood has run out of ways to make the whole gay angle for movies interesting. Nowadays, you can barely watch a sitcom without either some gay inferences being made or an outright gay character being portrayed as Samantha with pecs.....Its sad to see Hollywood perpetuate the old-fashioned stereotype of the gay man. Lets hope Ang Lee\'s movie breaks away form that mould......  '),(16753,3377701,'18,May,2004','2004-05-18 00:00:00','   Just as the title suggests , im testing our how this thing works '),(16754,3377701,'27,July,2004','2004-07-27 00:00:00','   I\'m siccckkk. Arrghhh !!! Fever and stomach ache. Not fun at all. It\'s weird cause yesterday in Lit class , one of my friend turned to me and said \" You look pale today , i think you\'re gonna fall sick tomoro \". I was shocked. And how true it turned out to be. One hour after i came back from college , my temperature increased. My stomach growled. Ouch. Had to skip class today , thank god i only had one lecture today ( the most boring one too HOI ). I\'m actually&nbsp;feeling better already , this evening two kind friends were nice enough to drop by and visit me. They even brought me apples! ( Thank uuuu ) Tonight we already planned of going this place to eat mamak food so even if i wasn\'t feeling too good , i just had to go. Had some good old teh tarik and roti plaster. YUUUMMMMMM. Too bad my stomach wasn\'t in the condition for my 2nd roti so i had to donate it to some greedy vultures ( kafaii duh ) Oh well i still enjoyed it , i mean , its MAMAK! Now to find some place that sells mango special and garlic cheese naan... '),(16755,3377701,'25,July,2004','2004-07-25 00:00:00','   Ahh it\'s been a busy weekend. Let me tell you how it went :   SATURDAY  1. Woke up and ate breakfast ( DUH )  2. Went to the market to shop for groceries ( vege , fruits , meat )  3. Back home to unload.  4. Off to China Town for more grocery shopping ( sauces , ill tell you why later )  5. Next , to Saveway for more shopping ( snacks , drinks , not-so-necessary items )  6. Back home , AGAIN  7. Off to Nike to meet with Daniel  8. Back to China Town , AGAIN , to help Daniel buy prepaid internet card  9. Back home , AGAIN AGAIN!!!  10. Finally some rest for like half an hour  11. Off to basketball ( got trash buy a bunch of Hong Kongians )  12. Dinner at KFC ( our friend supposed to cook but he dumped us ) 13. Finally home for good   Phew , that wraps up Saturday , now its :   SUNDAY   1. Woke up and ate breakfast ( DUH DUH DUH )  2. Played some games before getting a call from Kafaii for lunch...  3. ... in China Town ( 3rd time in 2 days wow )  4. Off to Saveway AGAIN ( to buy stuff we forgot to buy yesterday )&nbsp;  5. Back home for some rest ( but daniel was noisy so i couldnt sleep )  6. Friends start gathering at our place  7. One of them is cooking for us ( we call him Uncle cause he\'s 21 )  8. Started at 6.30 ( we heped cook as well )           9. Ended at 9.00 ( HUUNNGGRRRYY )  10. Clean up the mess we made ( it was baaaaaddd)  11. Online ( or else i wont get to post right ? )   And tomorrow\'s college again. ( Notice ? No homework done...... ) '),(16756,3377701,'23,July,2004','2004-07-23 00:00:00','   Another skool week is over. Hurraayy! Starting to get really used to life over here. Everyday its wake up - college - back - nap - dinner - online/hw - sleep - repeat. The walking distance does not matter anymore and the cold weather is not felt anymore. But it is getting colder everyday though =P its reported at about 6 degrees today. Brrrrr!!!&nbsp;Classes are good so far. Coping really well and there isn\'t really&nbsp;much homework.&nbsp;Even if there was ,&nbsp;the lecturers wont bother to check it. So its pretty relaxed i\'d say.  &nbsp;Two days ago we helped our friend move his stuff. He was shifting because he didn\'t have a proper place to stay before this ( He was sleeping at the friends\' living area ) Had to carry his bags for about 20 minutes walk. Caused me a shoulder ache ( Oww ) But oh well , good deed has their rewards - he treated us dinner. How nice of him. Ohh and last night , kafaii\'s sis cooked us bak kut teh. Obviously it wasn\'t as good as back home , but who cares ? Its BAK KUT TEH !!! SYOOK!!!!   Tomoro\'s weekends again..... wonder wat to do ? '),(16757,3377701,'19,July,2004','2004-07-19 00:00:00','   Yes , i did. For the first time in my life , i cooked for dinner. It was just some simple instant pasta ( Macaroni and Cheese )&nbsp;, but it still required some skill at least =P It turned out not too bad. It wasn\'t as tasty as i expected though. I guess i added too much milk. Oh well , there\'s always a next time.&nbsp; Lip Zhun fried the fish fingers and Kafaii.....umm....... washed the dishes. Oh waiit he did cut some onions. And he cut them into real big chunks. Yummm......I HATE ONIONS !!! I also found out that my cup cracked for no obvious reason , so i had to walk all the way to Kafaii\'s place to borrow a cup later that night. How sad..... '),(16758,3377701,'18,July,2004','2004-07-18 00:00:00','   Just before i start , here\'s a group pic of the winning team during camp : &nbsp;   &nbsp; We started the weekend by going to Smith Street to do a little shopping. Daniel and a few other friends of ours came to join us. Poor&nbsp;Daniel , had to wake up early just to miss a train ; and when he finally made it here at 9.30am , we we\'re supposed to leave at 10. ( kafaii told him we\'re leaving at 9 lol ) Anyways we took a tram to our destination. Smith Street is filled with factory outlets like Nike , Adidas , Reebok and so forth therefore making it a perfect place for me to get a pair of walking shoes. The shoe i brought here was spoilt during camp time. How sad. It was my favourite shoe too , a Nike Shox Limitless. Oh well , at least i got myself a nice Converse shoe for 49.95 $ and an orange Nike T for 10 bucks. Not a bad bargain huh. Kafaii and the rest also found themselves on a spending spree, buying shirts , jackets and shoes. We bumped into a few other college friends on the way and they too joined the party and we had lunch together. &nbsp; Later that night, me and Daniel went for dinner at China Town. After that , we decided since we had nothing to do why not crash Jenshen and Kwok Liang\'s&nbsp;place. They have after all&nbsp;, a Gamecube and not to mention a&nbsp;lot of snacks.&nbsp;Keewwlllll ! So we did , and George came over too so we played multiplayer games till about midnight.&nbsp;Not fair , they have a bigger apartment than us =(  &nbsp; Sunday morning , woke up pretty late today. Once again we joined our friends to eat at an Indonesian restaurant. Then it we went grocery shopping. Didn\'t take too long , managed to stock up on food and snacks. Its amazing how small this world is , we bumped into nearly half of the&nbsp;fast track students. Things got pretty boring after that. Daniel played the comp while i took a nap. For dinner , we we\'re joined by George and a friend of his to get a nice hot bowl of Vietnamese beef noodles. MMMHHMMMMM '),(16759,3377701,'16,July,2004','2004-07-16 00:00:00','   I had drama class yesterday.&nbsp; Yes , drama. For the first time in my life i\'ve attended a class which is stress-free and best of all , no books! It was&nbsp;really fun , i mean when was the last time we actually had fun in class? But maybe it\'s also the fact that over here , we\'re not really close to our classmates yet so it feels OK if we do something stupid or make a fool of ourselves. Drama takes like 30 % of our english marks, we have to do solo and group performances. Sounds interesting , doesn\'t it. Getting marks for something other than writing on paper :/ . But either way , its still my favourite subject so far. One thing though , we need to take off our shoes during class. So better wash our feets every Thursday.....&nbsp; Anyways , for dinner , we suddenly came up with the idea to crash our friend\'s place Calvin ( not Yew thank god ) and order some pizzas. His apartment was opposite our college at Swanston\'s , and man is his house big! It\'s like umm how am i gonna describe this..... lets just say his living room is bigger than my whole apartment. Yeah it is.&nbsp; And he umm lives there ALONE. Wow! Rich kid. He\'s even got cable TV. I\'m so darned jealous. But oh well , he was kind enough to let us hang out there for dinner. We had 4 large pizzas , and it was divided among 6 of us. Mannnn it was gooood. Better than Malaysian pizza anytime. Hehe... &nbsp; Today is the shortest college day of the week , or every week. 2 classes from 11 to 3. We planned to play sports after skool but it didn\'t turn out so well for me. We took off to the uni sports center and when we arrived it started raining. Not onli that , the courts we\'re full. So we headed back. And just when we were about to arrive home , our friends called and said that they were still going to play football. So kafaii went on and play with them but i headed home. No bball kakis sniff sniff. Oh well , good time to sleep anyways. It\'s starting to get really really cold here , might be because of the rain , but still..... Brrrrrrr &nbsp; Here\'s the group pic at the airport. Yeah i know , its FREAKING SMALL but i\'m sure u\'ll find ur head there somewhere =) &nbsp;   &nbsp; &nbsp; '),(16760,3377701,'13,July,2004','2004-07-13 00:00:00','   Things are starting to become a routine. College life, it seems. Yesterday things were back to normal. However i did manage to squeeze some time to meet up with my cousin during lunch break. Unfortunately , she was stingy and she refused to treat me . Sigh. Some cousin she is ( just kidding! ) Last night, we had a mini party at our friend\'s place and the girls cooked dinner. Not for us though, we already ate. Homework is starting to pile up. Not a good sign.....  Woke up this morning at 7.30am thanks to my dearest friend Daniel who was so kind to give me a morning call ( DAMNNN YOOUUU!!!! ) I was feeling drowsy. Class starts at 11, and HOI ( History of Ideas u noobs ) was first on queue. As expected, history is never fun. The whole class nearly felt asleep , and Kafaii described her as Jigglypuff ( Those who play Smash Bros should know wat it means ) And thankfully there was no psyhics lab in the afternoon so YEAAAAYYY free time!!! Finally got some time off to relax. Had to teman kafaii to bank to settle some of his stuff. The weather is getting colder and colder.....  *** My posts are gonna be real short starting from now cause everyday life is kinda the same now so ill just put stuff thats out of the ordinary =) *** '),(16761,3377701,'11,July,2004','2004-07-11 00:00:00','   Phew i\'m exhausted. We had a lecture marathon on Thursday. Man it was sooooo tiring!! It was from 8.45am to 3.45pm with onli 45 minutes break for lunch. Not fun. Yeah yeah many of u will argue that it\'s nothing different from college back home but, i had 8 months holiday so... =P hehe. It was so damn boring i had to turn out my ipod during HOI class ( that\'s history of ideas ). I planned to stay up late to chat wif my friends back home but damn the internet had to malfunction. How unfortunate. That\'s why i didnt post anything on Thursday. The next day was camp. It was held at Rawson Village, no idea where it is myself. At first , we we\'re all really unhappy about the camp. Karen and Jenshen told us how boring the place was and how dull it would be. They really turned us off. Thank god things turn out differently ( Explained later ).  It was about 2 hr\'s drive from Melbourne to get to our destination. It wasn\'t too cold, we we\'re then allocated to our rooms. We had lunch, fish and chips. Yummmmmm.....free meals hehe. The next activity sounded boring at first, it was drama activities. But it turned out to be good fun, we had to act out 5 slides of still pictures at a certain location. And of course with a story behind it. Each group had their own ideas and all of them were unique and very hillarious. Our\'s was about me and my wife having a honeymoon at a luxury resort, and were later attacked by a shark ( How sadistic ) The next activity was bushwalking , probably the most boring activity at the camp. It was just a really really long walk through the forest with nothing too exciting. But we chatted on the way so it wasn\'t too bad. Ironicly after dinner, we had to best activity at the camp. Bush dancing. Something new. Something cool ;). It\'s hard to describe the dance, they taught us quite a number of it. Its kinda like a party dance, like dance dance dance then switch partners and it goes on and on. So basically we get to dance wif all the gurls hehe. Everyone was sporting, everyone participated. It was really great, we didn\'t want to stop, but we had to. But the fun didn\'t end there. We we\'re playing around wif the basketball a little , and then the gurls suggested to play captain ball. Not a bad idea, we thought. And it wasn\'t. We had heaps of fun , and the gurls turned out to be more aggresive than the guys. Scary. We could have went on forever , but kafaii got a cramp, and it was then only that we realised it was midnight already. I made a lot of new friends , and slowly we\'re getting closer to each other.  Next morning, woke up freezing. The wind was blowing hard, and the air was cold as ice. The schedule was like NS , meal - activity -tea - activity - meal - and so forth. We were seperated into four groups, each doing different activites at the same time. My group was really sporting. We got together really well. We did high ropes , something i missed out in NS. It was scary at first , and it really hurt my hand. It felt fiery after that. Later, we played captain ball , AGAIN. Our lecturer in charge joined us, making things a little more interesting. We did the flying fox ( its starting to really sound like NS huh ) And my favourite , basketball !!!. There\'s this Japanese girl , she\'s really good in basketball. She plays for the Osaka Sabers. Wow. Hehe. It was great to shoot some hoops again, its been like a month ? Ohhhh it felt good. Next up was the mini-olympics. It would have been fun if it wasnt for someone in my team who is a real pain in the a**. Just thinking of him pisses me off. At first , the events sounded real serious, like 100m dash , high jump and stuff like that. Turns out things were different , like the long jump, was actually the limbo. Haha. It was quite fun, and our group turned out second. We we\'re awarded a lollipop each. Hehe. During intervals we play ping pong and some football. I earned myself a new nickname , chicken wings. It\'s spreading like wildfire , and i\'m always the joke. So what else is new? Oh , and i spotted a koala hanging on a tree that night. Not something you see everyday huh.  The start of a new day,  and i was feeling like an eskimo. We had to pack our bags , it was the last day of camp already. After breakfast we were seperated into our four groups again for some kind of competition. It consists of 4 mini games, and the points were added up to judge the champion. And guess whose group won =P We had teamwork , we worked hard , and enjoyed it at the same time. Our prize was a packet of Maltese , something i wanted to try for quite some time. ( It\'s choc btw ) We had tea , and we had time for just one more activity , the low ropes. No competition , just something to test our balance and stuff. Wasn\'t too interesting , but it was a good way to kill time. After lunch , it was back to Melbourne again. We got together wif our college friends for dinner, and we ate Indonesian food. Sigh time flies so fast when you\'re having fun. Tomorrow\'s the start of a tight schedule...... '),(16762,3377701,'07,July,2004','2004-07-07 00:00:00','   My room...     My apartment... ( Lip Zhun is there to portray the size of it )     Group picture !      The view from my apartment.... and Karen of course =) '),(16763,3377701,'06,July,2004','2004-07-06 00:00:00','   Ahh , the start of another new day. Woke up early today, at 9, surprisingly. And i slept at about 1 o clock. Something bad happened last night but i\'d rather not talk about it. Just another of my stupid accidents. Warmed up the garlic bread that came with the pizza the last time for breakfast. Played my notebook for a while, waiting for kafaii to arrive. Together , we left to the bank to activate our accounts. Didn\'t take too long, just a simple procedure. Recharged my pre-paid (don\'t ask why) and then we rushed to college. They\'re holding a Book Room, where we can buy our necessary lecture notes. We then had 2 hours break before the next lecture, so we decided to get our lunch at the Union House. 2 other friends joined us, they\'re from Malaysia too and from the same NS camp as kafaii. I ate kebab, the meat was not too bad but the bread was just too think. At the end of the meal i was exhausted from all the munching and chewing =P Kafaii on the other hand, ate tomyam noodles , but nooooo it wasnt enough. He had to feed on his friend\'s chicken rice too! tsk tsk tsk. Then we went to this common area where to got some rest, actually we were just killing time before the next lecture. We started to look at our Math 1 lecture notes, figuring out how to solve some questions ( Yes it\'s freakky i know ) After a long wait, the lecture started. It was just introduce to us on how to use the Trinity Portal. Nothing i couldn\'t do myself. It was soooo damn boring. Later, we found ourselves back home again. Didn\'t do much, just hung around till it was dinner time. Lipzhun went and joined his cousin, while me and Karen joined kafaii and his sis. His sister brought us to this Malaysian Chinese restaurant. Ate noodles with char siew. Not bad not bad. At least they had a decent ammount of vege in it. Haven\'t seen greens much these days...hehe. The night was still young, so they all came over to my place to just , umm well , hang out. Kafaii brought us to this website www.albinoblacksheep.com. It\'s full of lame video clips and laughed our heads off watching some of them ( SHOW ME , SHOW YOU , KIKKOMAN ! KIKKOMAN! hahahahahah ) Can\'t believe we spent 2 hrs on that... '),(16764,3377701,'06,July,2004','2004-07-06 00:00:00','   Warmed up some overnight pizza for breakfast. Finally something not cereal. Hehe.  We were supposed to be attending some introductory lectures at college but we decided to  skip it. Our first \"ponteng\". We had more important things to do. We had to get our phones. Kafaii met up with us and off we were to the nearest Optus outlet. ( Optus is something like Maxis ) Me and kafaii got ourselves Optus pre-paid, so its free for us to call each other now. Not to mention its free to Karen and Daniel too =) Next up, we went to Exhibition Street to get our internet cable connection ( for me and lip zhun ) It was quite a long walk, and it took us a while to find the place. We got ourselves a 190k cable connection. We were hungry then, so we dropped by at the nearest Hungry Jack for lunch. Fastfood is , after all the cheapest meal around. We then rushed back to try out our internet. Worked like a charm. Super fast too. I downloaded stuff at 100kb per sec. That\'s about 2.5X faster than my streamyx back home. Coooooooollll! Couldn\'t enjoy it for long though. We needed to go college , for subject selection. We left for college, and found ourselves back in the apartment in less than an hour. The subject selection didn\'t take long. We got back just to find out that....our internet was not working! Darn!!! Had no choice but to take a nap , praying that it\'ll work when i woke up. It didn\'t. Damn. Dinner time. We had a good meal, at A1 again. Roast duck rice, yummmmmm. It\'s always good to have some Chinese food once in a while. Went back to the apartment, and Internet\'s back on. Hurraayyy. Jenshen and Kwok Liang came over for a while, just to catch up on old times. Played a few games of DOTA , dont you just miss it ? =) '),(16765,3377701,'05,July,2004','2004-07-05 00:00:00','   I couldn\'t go online yesterday to blog so here it is. It was our first day of college! Yeaaayyy!!!  Woke up early as college started at 8.45. Lip Zhun stayed up to watch the Finals of EURO2004 and while kafaii planned to watch it, he didn\'t. He slept like a rock. Greece won...NOOOOOO!!! Anyways we packed our bags, ate a simple breakfast (cereals, AGAIN) and off we were. It took about 25 minutes for us to get to the Evan Burge Building, where our registration was held. It was simple, very organized. We got to know a few new friends from Indonesia, Singapore and even China. We were then treated for lunch. It was yummy and kafaii wanted more, but he \"pai seh\" wor.... Hehe. We were seated with the Deputy Director of Trinity and she was really friendly. She\'s actually from Malaysia. She came to Melbourne more than 20 yrs ago for her studies and never turned back since. And oh yeah Lip Zhun said loudly in front of her that we\'re skipping class tomorrow! *smacks hand on face* Later, we were brough on a tour of trinity college. It was a long walk, about an hour. The tour guide was really nice, and from Malaysia too. We had to attend some lectures after that, to guide us and inform us about our acedemic stuff and about subject selections. Almost fell asleep ZZZZZZ. 4.30, finally ends or first day of so-called college. Back again to the apartment for a well deserved rest.... We got pretty bored after that so me and Lip Zhun tried out some LAN gaming. Wasn\'t too bad, better than nothing. We played Jedi Outcast and Warcraft3 haha. And to our surprise , there\'s a channel on our TV that shows a surveillance camera at out main door! How cool is that? We ordered Pizza for dinner. Australian pizza and BBQ chicken pizze. Yummmmmm...very full tho! Couldn\'t finish it all so we kept it for tomorrow breakfast. End of day 4. ****Btw Lip Zhun is starting to join the \"lame\" gang ;)***  (my posts are getting shorter and shorter, i hope its not becoming a trend) '),(16766,3377701,'04,July,2004','2004-07-04 00:00:00','   Woke up in the morning , freezing. Woke up my house-mate, who was sleeping without a top. I salute him. Met up with Kafaii and Karen, AGAIN but this time with Lip Zhun joining the party. Stopped by a few places to shop for a compact heater, i need them to warm my feet from time to time. We went to this small chinese restaurant for lunch, named A1. Karen suggested i eat rice with some egg thingy on top so i ordered them with roast pork. It costs me 7.50 bucks, but it was worth it. The portion was big and oh am i glad to be eating rice again =) It was quite a long walk back, and i was carrying my brand new heater so we decided to \"hijack\" a tram. We call it \"hijack\" cause we\'re gonna go on it without paying. Don\'t ask why , its just some lame thing we made up. It was scary at first, but it doesnt seem so bad after all. We sat the tram to my place to drop of the luggage and get some rest. Somehow we ended up taking some pictures of the apartment and even a video to send to my parents later on. We locked Karen out on the balcony =P The day was still young, so we thought we should check out the location of our college. With Karen as our tour guide, she brought us on a half an hour walk to the college. We managed to catch a glimpse of the sports facilities on the way. Since we\'re already out, we then decided to crash Karen\'s place, just to kill time. We stayed at her shack for a while, enjoying a cup of lemonade while checking out some \"interesting\" magazines *Nyek nyek* We also watched a tape of.... South Park! ( OMG they killed Kenny! You bastards!!! ) We felt tired after that so we hijacked another tram back home. Kafaii got seperated cause he geniusly ran across the road and into the tram without waiting for us. The rest of us couldn\'t cross. There were cars. Bye kafaii! Luckily he didn\'t get lost hehe. It was only our second time hijacking a tram. Had a 2 hour nap ZZZZZZ and then it was dinner time. Went out for fastfood again, this time its Nando\'s. The price was not too bad , at 5.95, but the chicken was freakinnn smalll!!! The side dishes was in a quite big portion though, but it wasn\'t enough. We went back to kafaii\'s place for a cup of instant noodles. We\'re staying overnight cause kafaii and lipzhun\'s gonna watch Euro 2004 Finals next door....while i sleep =D '),(16767,3377701,'03,July,2004','2004-07-03 00:00:00','   2nd day: Woke up freezing at 9.30pm . Brrrrrrr! Made myself a simple breakfast. Cereals and hot chocolate. Mantainance staff supposed to come to the apartment for a checkup, but turns out he couldn\'t make it. If only i was told earlier, i\'d love to sleep a few more minutes :P I went to kafaii\'s place to meet up wif him and Karen. It was Simpsons Weekend on FOX8, so we took an hour to drag kafaii out of the apartment. We went to Victoria market to pick up some groceries. Karen spent us some nice , hot spanish donuts. MMMHMMMMMMMM !!! It\'s like \"yau char kuai\" wif sugar on it. It was lovely, especially during this kind of weather. Next stop , shopping again. Haha. This time it was mostly for food, you know, instant stuff. I bought pasta, instant noodles, milk among some other stuff. Karen tempted us to get muffins even though i didn\'t want to. I wanted to save. But she talked me into it, darn another few bucks wasted. We then got back to kafaii\'s place and cooked lunch. Well actually not we, mostly Karen haha. Me and kafaii helped wif the cutting and of course, tasting. Not really hard to cook though, just dump in the ingredients and let it cook ( the pasta has sauce wif it already in powder form ) It was OK, could have been better cause the pasta was a little hard and chewy, not as soft as it should have been. It was still i good meal, we dumped chicken, onions and carrots it in, not to mention all kinds of salt and seasonings we could find in the cupboard haha. While waiting for the food to be ready , i warmed up the muffin i talked about earlier and maaaannnn it was worth it. Ohhh it was hot and oh so delicious! After that, it was back to the apartment for a little rest. Lip Zhun was still out, wif his cousin. Karen decided not to go dinner, so me and Kafaii went to the nearest fastfood outlet for our dinner. Hungry Jacks. Ate a Whopper meal , which is the cheapest around at 5.85 bucks. And after that i went home to get my laptop and brought it over to kafaii\'s place. There\'s where i am right now =) '),(16768,3377701,'03,July,2004','2004-07-03 00:00:00','   1st day: Arrived at airport at about 7 am , aussie time. Didn\'t get enough sleep, it was uncomfortable. Kafaii slept like a rock. It was havoc at the airport. There was no trolley available, had to drag my super heavy 30kg luggage. I think i built some muscles already =) The imigration department was holding the line back. Took about half an hour to get out from there and to find a nice person waiting , holding a board \"Mr. Chee Weng\". He drove me to my apartment , Arrow on Swanston. His ride was a luxurious one. Had GPRS , heck he could even charge his phone inside. Took a while to check in , and i found out that my apartment-mate, Lip Zhun was not there yet. Turns out we accidentally left him at the airport (Sorrryyyyy) The apartment looks like a mini hotel. I actually need a keycard to to access the lift , and works as the key to my shack. And guess what, my apartment , with kitchen, toilet , living room and 2 bedrooms, IS AS BIG AS MY ROOM AT HOME!!! But at least everything looks brand new, i guess it shouldnt be a problem living here for a year. After an hour, we got out of our apartment to meet up with Lip Zhun\'s cousin, this nice guy who knows almost everything about Melbourne. He brought us to the heart of the city, and we met up wif Karen. Lip Zhun and his cousin went their own way for lunch, as me and Kafaii patiently wait for Karen to arrive at Nike shop. Yup, we waited and waited and waited. FOR HALF AN HOUR!!! So sieeen summore so coollddd! We went for lunch, ate fish and chips. And in case u guys are wondering, an average meal here costs 6-8 bucks. Yes its that expensive. Enjoy food as much as u can in Msia. Went shopping later to buy some stuff. Its really convinient to shop here since we\'re already living in the city... Reached home at about 5pm, finally to get a good rest and a hot bath. I unpacked my bag ( Had to , there isnt even enough place to leave it open ) I went dinner with kafaii and lip zhun, at a Vietnamese Noodle House. Ate a delicious bowl of Beef Noodles, at a huge price though. 7 bucks (7 X 2.8 = ?) After that , boredom. There\'s absolutely nothing to do here. Hung out at kafaii\'s place for a while, watching TV. We got bored , so we decided to check out Lip Zhun\'s cousin\'s place, which is coincidentally next door !! And OMFG his place is da bomB!!!! He has all kinds of entertainment stuff , he has 1500 K broadband man!!! And he has this really choon PC made by alienware wif a cooling tank at the side wif plastic fishes swimming in it! How cool is that? Not to mention he has the best Altec Lansing speakers connected to his TV and guess wat, he even has Gamecube and PS2!!! And of top of that , FUZZZZBALLL!!! Damn he\'s got cash. After a while, i got back to my apartment and played a little game of Warcraft3 to kill some time. ZZZZZZZZZZZZZZZZZZZZZ '),(16769,3377701,'09,August,2004','2004-08-09 00:00:00','   Wtf ? Okay , its some kind of rare disaese that is spreading rampantly in Aussie or something. We got free vaccines for it , so wats to lose? We all took our jabs today though i needed to wait a FREAKIN HOUR for it. Man the doctor took like wat , 10 min per person !??? Grrr!!! And it onli took him 2 sec to jab the freakin thing!!! ( didn\'t feel a thing ) Anyways dinner wasn\'t much better either. Me and my stupid bright ideas. Tried to copy Daniel\'s mom\'s tuna pasta. Failed terribly.... nearly ate raw potato if not for my friend who was coincidentally around ( potato cant be eaten raw ? LOL ) Tasted like dog food , is was so wet and yucky. Next time i better get the proper recipe before experimenting again.... BLeh!  Still rethinking about the blog layout changing thingy , since Jolene kindly offered her assistance to me =P '),(16770,3377701,'03,August,2004','2004-08-03 00:00:00','   Some of you might realize that i haven\'t been blogging for some while. This is because i was finding some new layouts hoping to redecorate my whole blog. But after lots of thoughts , i finally decided to stick with this one. Even if i was going to change anything , it\'ll just be a small change so.....Anyways im back now and ill be starting to blog again soon =) '),(16771,3377701,'03,August,2004','2004-08-03 00:00:00','   Here\'s Lip Zhun and our friend Angga trapped at the balcony in the freezing cold :        And here\'s my poor monkey getting tortured by my evil housemate.... : '),(16772,3377701,'01,August,2004','2004-08-01 00:00:00','   Today we had physics lab and this is the first time for us doing it alone without assistant from lectures/teachers. Man was it a horrible experience. Mine was to study the coefficients of static friction of a variety of materials and some other factors. Have to put a mass on a plane then increase the \"tita\" to a point where the mass slides down and measure the \"tita\". Have to change material everytime and all those bull also. Man my results we\'re all over the place with hugh errors and inconsistency. And to top it all we didn\'t have time to finish our reports ! Arrghhh !!! And to think that its counted for our total marks in PH ! Omg...... And everyone else had problems too. Next time i\'m sure everyone will come to lab prepared for everything......   *** It seems only jolene bothers to comment on my posts...for everyone\'s info u guys can talk cock by clicking  on the COMMENTS button below every post. And if karen u\'re seeing this we can go lepak anytime , but its hard to contact u when ur phone is broken =P *** '),(16773,3377701,'01,August,2004','2004-08-01 00:00:00','   It\'s starting to look like the only thing interesting here is food. Everything else is so normal. I\'m starting to play basketball again , but not much. And it will only get less , cause the skin of my right palm is starting to peel again ( its that time of the year i guess ) and its hurting too. I also found out that it is not wise to go shopping alone. You need all those extra hands you can get when you are buying really heavy stuff. I bought a set of kitchenware and sweated my way back to the apartment ( its winter btw ). I\'ve been walking around alone a lot lately , i wonder why...  Anyways back to the food. Lip Zhun\'s family is in town , and his parents were kind enough to treat me n Kafaii to a luxurious dinner. Lobster noodles.....OOHHHH SYOOKK!!! Something we wouldn\'t enjoy for a loooooooooooong time. And the other dishes we\'re unbelivably delicious too! Yummy. And just last night , i walked to Grattan ST to tapau some nasi lemak. They tasted good , obviously not as good as back home , but good enough. That afternoon we also got the chance to try Gelati ( some italian icecream u noobs ) Omg its the best ice cream ive ever tasted! Texture smooth as silk , flavour rich as ummmm Bill Gates ? Oh god that\'s lame. Must be too much kafaiism. Hmm.... And oh yah i cooked dinner ! Without professional help ! Really ! God and it actually tastes good too. Here is the pic of my newly created dish :    * Yes , yes it doesn\'t look appetizing but it tastes good ! I promise !*  I forsee an interesting and busy week in front of me , especially with drama. We have to do a 2 min solo mime performance each ( mime = acting without props and sound ) That\'s gonna be something..... '),(16774,483266,'17,May,2004','2004-05-17 00:00:00','   Well, another year ticks on and I make it to 33! WOW - Where has the last year gone ?  I know that the last 48 hours for me has been quite full on ! Started on Friday with my birthday - It was nice, thanks Tina and co :-) A nice easy transition into 33.  Saturday was then another memorable day. Why ? Well it was my leaving do from my current work place. A nice evening with friends from the ward at Gino\'s Italian pizza restaurant.  Thanks to: Nikki, Mandy, Karen, Marie, Mary, Helen, Margs, Tracey, Lisa and Wayne all for coming - and thanks to everyone else on the ward for their time, support and encouragement - Cheers :-)  Sunday was then a lovely day out to the sun and a trip to the Torquay region of Devon. Was so nice, thought I would include some photos taken on the day:    The village of Cockington    Brixham Harbour - 1    Brixham Harbour - 2  What a lovely day out - TOP. Went with my Mum and wife :-)  Then onto today !! First day of my new job.  An induction week which kicked off with the annual drug update :-/  And that was my weekend :-)  '),(16775,483266,'13,May,2004','2004-05-13 00:00:00','   Well, my new job is looming on Monday, and far be it from me to not make an effort. As my hair was getting rather long, I decided it was time for a crop.  The last two hair dressers have closed since I started visiting them (not sure if there is a connection there) so I tracked down a new one to try.  Whilst sat in the chair, I thought - what the hell!! Lets get radical. My old friend Mr Rudd always has his cropped and chopped, so I thought I would give it a go. The hair dresser then informed me that you can now get a grade 8 trimmer which was quite popular in today\'s trends.  So I went for it: Grade 8 all over the top and grade 3-4 at the back. You judge for yourself.    Before    After  '),(16776,483266,'12,May,2004','2004-05-12 00:00:00','   Anyone that owns an XBOX and knows a little bit about the current E3 computer fair will be as excited as I am about the new Halo 2 Multiplayer video available to watch.     Not seen it? Trust me, it\'s AWESOME !  '),(16777,483266,'12,June,2004','2004-06-12 00:00:00','   Well - it\'s finally happened.  We\'re gonna have a baby :-)  Ye haww !!  Sometime of just yet. For those that know me, keep an eye on the bump blog for the latest details.  COOL !  '),(16778,483266,'01,June,2004','2004-06-01 00:00:00','   After rushing across Exeter in the mad rush hour traffic, I finally made it to the Odeon to go watch The Day After Tomorrow.  Bugger !  Sold out - So next on the list was the latest Harry Potter movie, the one where someone escapes from a prison.  I\'m not sure if it was because I was looking forward to the other movie but to be honest, I found the movie kinda boring !  It didn\'t really grab my attention, and seemed really blank and drawn out in places. There are some nice special effects, especially I thought the moving tree and the horse come chicken come bird creature.  None the less, there was nothing ground breaking just more of the same but done better.  Still, at least I have TDAT to look forward to still :-)  '),(16779,483266,'01,June,2004','2004-06-01 00:00:00','   For me, someone who can\'t help but take things apart, it was only a matter of time before my XBOX got taken apart. Now I\'ve done it I\'m even more excited than before. The potential is huge !!  That little buty could be re-built into anything - only my imagination is the limit !!     Compact little bugger ain\'t she ?      Why! It\'s a PC in a box, crying to be modded !   I\'ll post any pics of final systems.  '),(16780,3288591,'23,May,2004','2004-05-23 00:00:00','   If you follow the progress of the  urlLink MARID IETF working group  â€¦  â€¦ then youâ€™ll know that the prospect of a spam free future is within our grasp. The champions for competing standards for MTA Authorization Records in DNS (MARID) are beginning to agree on standards for the definition of email domain usage policies. Once implemented, the MARID proposal will ensure that email that purports to be from a domain (mimecast.com for example) actually does come from that domain.  Most anti-spam pseudo guru types will try to tell you that this does nothing much for the spam problem â€“ but the application of a little brain to the process reveals a massively powerful anti-spam system in the works. Most of the people who donâ€™t understand the MARID proposal are of the content filtering, artificial intelligence ilk. Weâ€™ll get to these poor folk in a moment.  Firstly, if every email from a domain actually is from that domain, then â€œspoofingâ€ (also known as â€œJoe Jobbingâ€ â€“ which describes the use of someone elseâ€™s email address without their knowledge) and â€œphishingâ€ (dolling up an email to look like itâ€™s from someone you know in order to get you to give up some info or cash) would simply cease to exist. At first glance, this seems like it has nothing to do with those hundreds of emails you receive suggesting that your sex life needs a boost, but the angel is in the detail.  Secondly, if every email can be proven to be from its purported domain, then the world can be rid of â€œbounce spamâ€. Hereâ€™s the geek description of Bounce Spam â€¦  Bounce Spam is what you get when someone Joe Jobs you to an invalid email address.  Basically, this happens when a spammer uses your email address to email someone non-existent at a valid domain. The email server (usually a pretty dumb one) accepts the email because itâ€™s destined for an email domain that the mail server deems as its responsibility. The email server then works out that the domain is valid, but the actual recipient in invalid. The mail server then dutifully constructs a bounce message â€“ and delivers it to the source of the original email (your email address).  You end up with a bounced message telling you that an email you *never* sent has *not arrived* â€“ or worse, you end up with a bounce message telling you that an email you never sent WITH A VIRUS could not be delivered. Either way, if you are not fully informed as to why this could happen, you panic.  The MARID proposal resolves this problem.  So, why would an end to spoofing, phishing and bogus bounces also be the end of spam?  The net result of the MARID change is the restoration of trust in email. After MARID, the only way to deliver an email will be through the use of a legitimate email domain. This, in turn, leads to reputation based controls for domains â€“ so good domains who act responsibly and do not send unsolicited commercial email (UCE) get to deliver their emails easily. Bad domains (those that deserve a sever spanking) find themselves locked up within black lists for abuse of their domain. These policies cannot be enforced in a pre-MARID world.  Right, back to the title of this blog. If MARID succeeds, then content filtering systems will be switched off. Most content filtering systems are based on the idea that software can accurately determine desired from undesired messages based on the contents of the message. The anti-spam industry invested heavily in trying to figure this stuff out, but failed to realise two things:  1. We havenâ€™t invented artificial intelligence yet â€“ and itâ€™s a long long long way off.  2. Had we already invented AI, it still wouldnâ€™t be able to work out what we want, or do not want, to read. This is because we havenâ€™t invented (or even proven the existence of) ESP yet.  One manâ€™s spam is another manâ€™s email newsletter (which is currently being binned by an ultra-clever, ultra-expensive false positive solution shipped by an anti-spam vendor).  If I ask for a glitzy, image saturated HTML based email newsletter from CNN, then I expect to receive it. I donâ€™t expect some brainless software to â€œfilterâ€ it out for me. Anti-spam vendors have had their day. Now itâ€™s time to get back to using our email powers for good.  Roll on MARID  '),(16781,3288591,'14,May,2004','2004-05-14 00:00:00','   The EU seems unable to stop itself from legalising software and business method patents. Iâ€™m not going to get into why software patents are bad, as there are many web sites dedicated to pointing these things out.  Have a look at  urlLink http://www.ffii.org.uk  if you need the background.  What I am going to suggest is a solution. Simple.  SUE THE PATENT OFFICE.  The idea of patents is fine, and I firmly believe in the processes by which a patent should be granted. The big problem stems from the fact (and there are many examples) that the patent examiners tend to err on the side of granting a patent if it is presented in sufficiently minced patent speak (possibly because they really donâ€™t understand what it means).  Given enough time, you can author a patent application for anything at all â€“ and â€œproveâ€ the inventive step if you use enough of this patent garbage speak. Donâ€™t believe me? â€“ have a peek at Google, and search for â€œpatently absurdâ€. Youâ€™ll find good articles on thousands of ridiculous patents that have been granted.  The patent office needs to take more responsibility for what it is doing, and it needs an incentive to NOT grant a patent. Patent examiners should be rewarded for finding prior art and breaking the gobbledygook code used by patent authors to get to the true inventive step in order to disprove it.  The patent office (here or wherever you are in the world) will not change all by itself. It must be made to change by suffering defeat after defeat in lawsuit after lawsuit challenging bogus patents. The patent office must start to feel the weight of responsibility that comes with their power to assign rights of invention, and it must change its default behaviour to that of declining to issue patents.  One other change needs to be applied, in my view. Patent authors should be charged for patent rejections. This means that all of the time (in billable hours) that is applied to investigating the patent should be charged to the author of the patent upon rejection. The patent office should carry the can for the costs if it passes a patent. This changes the dynamic for patent application, and starts to ensure that bogus applications are discouraged, thereby freeing up the resources of the patent office.  Patents are good (even very specific software patents), but the patent office is bad. It needs to get a severe spanking â€“ and we can all help by banding together to sue the pants off it (spanking generally works best with trousers down).  '),(16782,3288591,'10,May,2004','2004-05-10 00:00:00','   My father was particularly interested in writing. He spent a reasonable amount of his spare time in front of an old manual typewriter and laboriously produced sheet after ink stained sheet of text. He passed away when I was quite young, and the only thing I have left of his, is an old leather suitcase containing all of his written work.  I have no idea if his work was any good, but the fact that the work endures is possibly the most important thing about it - for me anyway. This blog is hopefully a new form of enduring written work and, no matter how bad the prose, grammar, diction (and spleling) - I hope it endures and is of value to those people who find it.  Last year I was introduced to some really smart and creative people. After realising that we had similar goals in business, and in the tech sector, we decided to co-found a small software company called Mimecast â€“ which is now based in London. The trials and tribulations of launching a startup are well known, so I wonâ€™t bore you with the details (Iâ€™ll hold that off for another blog).  Iâ€™m a bit of a technical general dogs body within the company. I try to help with a bit of technology vision, a little bit of business application, a little bit of technical pre sales, a little bit of software design and a lot of coding. I wouldnâ€™t have it any other way.  This blog will undoubtedly contain both random and considered thought. More random than considered I can assure you. To me, blogs are like steam valves for an overactive mind. Sometimes its necessary to release an idea or a thought pattern in order to make way for new thoughts and new ideas. The problem is that without a mechanism like a blog, those purged thoughts and ideas may be lost to infinity.  The reason why this blog is called â€œThe Mud Island Blogâ€ is because Iâ€™m a slightly sun scorched South African who now lives in, what most South Africans consider to be, the very cold and wet England. If you happen to love the development of software, then the wetter the climate the better. I personally donâ€™t think the weather is so bad, but I probably donâ€™t get away from my notebook long enough to notice one way or the other. I live in a small village about an hour away from London, and there is an awful lot of mud around. I can tell the season by whether itâ€™s frozen or soggy.  However you got here â€“ welcome.  '),(16783,3614613,'15,August,2004','2004-08-15 00:00:00','   I spent like, quite a long while to wait for this site to load before anything popped up. Maybe it\'s only my computer, or maybe ( ugh, I wish ) lots of people are all coming to my site at the same time because it\'s so popular and there\'s why even blogspot\'s server can\'t support it.  Ah music. I like music. Anyway, I wanted to put up music iwebmusic but I tried going to my own site and this security warning appeared. Remember kids, when ever you see that, always click no. Just in case anyone didn\'t know and tried to press yes, I removed iwebmusic altogether. Yeah, I do care about my reader\'s computer health you know?  Almost every paragraph ends in a -_- if you\'re someone particularly sensitive to bullshit. Anyway, you know the line about \"ah-choo, I\'m allergic to bullshit\"? Well I\'m quite tired of hearing that already. It\'s getting too repetitive, if you get to give me that -_- face in real life where I can see it, think of something else witty to say.  So... It seems that in today\'s newspaper there\'s an article on somebody getting an Asia\'s best blogger award. I forgot what the site is, and I don\'t think you people care much about that either, so I suppose that doesen\'t matter. Anyway in it, it\'s reportedly notorious for dissing about a lot of things.  '),(16784,3614613,'14,August,2004','2004-08-14 00:00:00','   Ah-Hem... I now begin my apology to Tan Li Lin  (her blog) in order to please her royal highness:  Dear Finest Woman On Earth (a.k.a Tan Li Lin)  I humbly ask for your forgiveness, for disregarding your feelings at that particular moment of time, whereby I have created a picture, depicting your face upon a picture of a body builder. I was not thinking clearly and was extremely bored at that point of time, therefore creating that picture. So I\'m wracked with guilt, hoping you don\'t tell your mom about this. Please, just forget about it.  It doesen\'t mean I\'m a bad person.  From Nana.  '),(16785,3614613,'13,August,2004','2004-08-13 00:00:00','   I mean like okay, this is a weekend after all... and I\'m so bored. This morning, I went to change the colour of my braces again. This time it\'s blue. Not any blue though, it\'s er... baby blue. I look ridiculous, nobody will take me seriously this way! And I didn\'t choose it, it\'s my dentist just randomly select for me.  okay chatting with jean now. She sent me a pic of Li lin at junction at, they probably went out there after going to CJC oogling at guys. (jeanie reports that everyone were nerds there, so I guess the li lin\'s hopes were dashed eh?) In any case, I opened the pic, and I began a staring competition with li lin. She won.  Oh  and the hateful tagboard is gone. Get out. Get out of my site right now.  '),(16786,3614613,'13,August,2004','2004-08-13 00:00:00','   Er right, somehow when I get a chance to post, all that I want to say is gone. Tsk. I should really stop this ridiculous habit, lol. Okay, I\'ll go through the day\'s event in school, since school is really what this blog is based on. Civics and Moral education was really really boring today, I\'d say it\'s about as useful to learn it as learning yak acupuncture. And isn\'t what we learnt today about the Total Defense a form of propaganda? The purpose being to unite the people together. It certainly isn\'t working on me. I rather watch my toenails grow... and play shooting with Li Lin. Oh and today I used one of Jeanie\'s Babies as a gun, I pointed the \"tail\" and held the legs to shoot. It even has sound effects, nations shall tremble when I point that gun and it squeaks. -_-; -_-; -_-; -_-; -_-; -_-; -_-; -_-; -_-; -_-; -_-; -_-;  Here\'s a thought for you 2hians out there. What if beginning of last year, people were not the same people as before? Like for example. Felicia was like Jeanie and vice versa, and Edward was like Joel and vice versa. In that world, Weijin had straight teeth, Li Lin was very lean and tanned, and Ying ying was like what Li Lin is now; white. Then, people like Hwee Kim and Kim N. would be exactly like Havi and Kim A.  If by now you have a clear understand and a vivid imagination, don\'t you find it weird? I have often played with this idea during extremely boring lessons.  '),(16787,3614613,'12,August,2004','2004-08-12 00:00:00','   Well okay look at the difference between this entry\'s time and post number 6. time. That\'s the amound of time I have slept the whole day. Anyway the scout thing is solved already, I have called joel john up and asked him for money. He\'s the type that would keep a huge store of money, so that it may be spent the next time people like him go out to have fun.  Er let\'s see, what else? I mean like what can I say there is nothing to say. I have been asleep after all. Oh okay, let\'s talk about this very site then. I got a new tagboard today, and I can\'t seeme to find a way so that it fits on the side bar. Looks like everything I want to add from now on will have to fit on top. Nevermind, so long as it works.  Hopefully I\'ll get my water bottle from jeanie tomorow... And eee, eeerr.... What else? Oh okay, this isn\'t some kind of court where I have to confirm the testimony of other people but yeah, I did play \"shoot shoot\" matrix game with li lin. It was like a gulp of sanity in a desolate and barren desert. I read that description somewhere. And er yeah, it WAS fun in a way. I did try doing the neo thing, and pretending to freeze all the bullets she shot and then flick them off the air.  Oh and, if you people haven\'t noticed this already, Mr. David is losing his confidence... Such a small sentance yet...  '),(16788,3614613,'11,August,2004','2004-08-11 00:00:00','   Oh geez, I really am I really am lazy at scouts... Anyway, I lost 2 scout donation draw booklets. Terrible terrible.. The only consolation is that Mr. Ong called to say that a police report is not necessary. The only difficulty now is, where do I get 40 bucks... 2 booklets is worth $20 each...  I\'m trying to borrow from my sister, but if she don\'t wish to lend me... We\'ll I\'m quite resourceful.  Anyway, today I did CIP again after school... Honestly all that thing about grooming youngsters into becoming responsible and all those caring people are just poppycock. Yep, the word CIP sounds as ridiculous as the word \"poppycock\", check the dictionary. I learnt that word in an old \"Archie\" comic. Anyway, today\'s task was to pick up litter, which I did not. Haviender found a few thrash bags full of litter already, gathered from the bins around the school by one of the janitors already. How convenient. Then after that, I waited for WeiJin to finish his detention, meanwhile talking to the councillers there, persuading them to let weijin go early.  Anyway, the councillor there makes me feel like I\'m a very lazy person, which I admit openly and freely. In terms of scouts and studies. But whatever, I still survive don\'t I?  By the way, this message has been brought to you by blogspot.com and the online writer, websurfie. It has been typed with the aim of either 1.) Making you go tut tut tut, like a priest to a sinful person or 2.) Make you deny that you are not a goody-two-shoes nor are you lazy in studies like me.  '),(16789,3614613,'11,August,2004','2004-08-11 00:00:00','   Today, there was a new teacher to take over Mr. (?? ?? Buck) of science. Name was Mr. Jude David. He\'s okay, except I notice he has a lot of policies and all those \"my standards for this class is...\" Are we going to go through that everytime we change teacher?  It\'s not ususally one of those things you notice AND remember but it\'s kind of funny to see li lin poking the air just in front of his ah-hem with her ruler. All it needs to top it off is for someone like calvin lim to fan his behinds when he is not looking.  After lots of whining from my family, I removed the pictures of my sister. My father was going through reasons like what if people copy my sis face and paste on naked bodies or something. Yes people, go ahead, give him the -_- or the \"...\". They are the same anyway, arn\'t they? What are the chances? But I suppose he can\'t help it, he\'s very insecure and we must all be very patient.  I promise this will be the last time I involve my family, EVER. Supposing my mother will bring me to suntec in future, it shall henceforth be simply phrased \"I went to suntec\", unless you people probe, but I doubt it very much.  Let\'s see, er I did CIP today... uh, at the D&amp;T workshop, cleaning up. It was one hour only, but then again, I have a huge boon of CIP hours, just ask any scout. And er, after that, walk home with Weijin. I was trying to convince him that he is not clever, and thank goodness I didn\'t suceed, which makes him a worthy friend, for I don\'t like the weak minded. Or at least I think I didn\'t suceed. Haha, well Lin, any snide remarks about how I failed miserably?  And you arh, Jean, please get my bottle tomorow afternoon okay? That ought to be it for today, and er, I decided to update old pictures that are found on my blog, unless they are really really good pictures or they are necessary for the text to make sense.  '),(16790,3614613,'10,August,2004','2004-08-10 00:00:00','   Well I don\'t live happily ever after after buying that camera you know? Daddy was complaining, (I\'d say whining but on second thought I\'d better not) over and over again to mommy for letting me get the camera. The point I see in this whole matter is that if even I don\'t complain, then he has even less right to do so, after all, it\'s my money. For now, I\'ll just play with the new toy of mine. I\'ve posted pictures, all of my sister, because she\'s the only one who is decent enough to be taken and she loves having her cutesy face taken.  '),(16791,3614613,'09,August,2004','2004-08-09 00:00:00','   Well, I got a digital camera today. At suntec. It\'s really lovely. I like it so much. But I have to stress that it belongs to me and me alone. I paid it with what I saved, so there! Please note that I\'m not particularly saying this to anyone, except my family members who may occasionaly borrow it without my permisson.  Oh and of course, this is rather predictable, my sister says she will get one that is the same except that it is smaller, slimmer and can record sound. I shall leave my readers to judge what kind of message I am trying to put across.  Tomorow school is going to reopen, finally a day when I can really write something that is related to school. I wonder how it will be like? Most likely chinese will be involved. I haven\'t passed up both my files (pure forgetfulness, not never do okay?) and I haven\'t done my story book review yet that was given in the june holidays actually. I shall do a quick ppt. slides on the book review today, hopefully I can snitch some marks for that.  To all those people reading this, work hard to get what you want, and you will get it. But first you must identify what you want, how to get it, and think about how much you want it, so as to motivate you to get what you want. I got my digital camera out of my savings, and though this may sound corny, go out there and fufill YOUR dreams (er, non-dirty dreams).  Pardon me for being nasty for a few seconds but according to Ervina\'s site, Weili = Hamtaro???? Oh dear... Better not say anything, or else get whacked by Ervina.  '),(16792,3614613,'09,August,2004','2004-08-09 00:00:00','   Today was rather boring. Really, try thinking about how boring a day can get. Go on, think.  Anyway, today my parents when to JB for shopping, what ever for? True, things are cheaper there but I don\'t really think it\'s worth it. Then again this rarely happens and it\'s their time, not mine. Speaking of which, perhaps I should have went, time today was really wasted. I should have done something.  Searched the internet for digital cameras today, I\'m hoping to get something which is quite slim, yet gives good pictures and preferably holds more than 1 picture? lol, make that 30. Most importantly I\'m hoping that is has a fast phototaking time, once during a school concert, while I was taking photos for ST club, I pressed the shutter button but when the camera finally \"clicked\" the people on stage were gone. Sheesh.. Jeanie does your camera have that problem?  Oh wait... Parents are back, I\'ll be back after greeting them and so on and so forth. *back* In any case, my sister is not to pleased about something my mother bought for her, because she wants to choose it herself. After boo-hooing and screaming and shouting, I found out from my mother that she bought... Bras. Well! No one is going to see them anyway, so why so fashion concscious unless they are uncomfortable? Then again, this is a girl\'s issue and I will desist from commenting futher for fear of insulting female readers  Momma also bought a shirt for me. I find it horrendous. Look at if yourself.  urlLink    WC (the brand) other than meaning \"wild channel\" can also mean \"water closet\" or, old-fashioned toilet bowls.  urlLink     Look and stand in horror! It\'s yellow! Anyway, I do not like bright colours. I only like black clothes, nothing else. Unless it\'s dark, then I think I will like it as well. But yellow? And I don\'t think I like that bullseye target thing, cupid aiming for my heart? Sorry, but it\'s made of stone so even if you hit accurately cupid, I can\'t promise you that even your steel arrows can penetrate. And what\'s that man with horns on the head...  Anyway according to my mother, my father chose this shirt because my surname means yellow in chinese. What kind of reasoning is that? In anycase, I\'ll wear that shirt, because I hate wasting money, but my mother never learns that she must never buy clothing for me... then I\'m forced to wear clothing like water closets or something.  Joke of today: What does LiLin do when she had a boring day? Answer: She Doesen\'t Blog! Hahaha! huh? oh, right, yep people all over the world are giving me that -_-\" face again.  '),(16793,3614613,'08,August,2004','2004-08-08 00:00:00','   (read previous entry, no. 2) Well it wasn\'t so bad after all, but it wasn\'t good. I hated the way I had to act short all the time, but at least I looked okay in the shots.  Is there a problem Li Lin? And please don\'t smirk, perhaps it has got something to do with today\'s family phototaking but beautiful people smile, the rest of them can smirk.  Hmm, it\'s getting into my brains now, I should get rid of it. One last quote on looks. People look their best when they don\'t even try. That ought to fill up philosophers appetite for today, it\'s food for thought.   urlLink    I took this on our way to the photograph studio. The houses look kinda nice to shoot, but I\'m afraid it\'s not very good photograpy.  urlLink  '),(16794,3614613,'07,August,2004','2004-08-07 00:00:00','   This is after all only my second post, and I\'m feeling the itch to blog again. For no particular reason. Strange, maybe this will wear out sooner or later and I\'ll probably abondon this site, like a site I made in primary school  urlLink http://www.maxpages.com./websurfie . It\'s completely empty so don\'t bother finding anything there, unless you want to take a test on how fast you can type over there...  Today there\'s going to be a family photo taking thing, so my family went to the early morning worship so as to come back early enough to prepare for the photo taking. I assure you this is not going to be those let-granny-take-the-picture one, it\'s one of those \"professional studio\" thing. The occasion? My father got his master degree in business, and he wants to remember this occasion. I hate having my picture taken. I hate it even more if I must smile against my will, but oh well, if it pleases my parents, whatever.  I read this comic strip \"baby blues\" (not today\'s newspaper). Anyway in it, the mother, Wanda was telling her children \"From now on, you are only allowed to say, yes mom, okay mom, sure mom, of course mom, no problem mom, right away mom and that sort of thing when I ask you to set up the table.\" The comic strip ends with hammie telling zoe, \"I still think whatever sounds better.\" To which zoe would reply \"shh... don\'t let her hear that.\"  I learnt today that I must be generous to others, all because my mother bought tissues from a woman and her blind mother selling them at the market today.  '),(16795,3614613,'07,August,2004','2004-08-07 00:00:00','   Well, it is this day that this blog is made, who knows? What will be said? Perhaps it shall stay near-annonymous like cassandra? In anycase, I think shall be asking Jeanie (probably the most visited site belonging to the queen (okay okay jeanie, queen does sound old!) princess of 2H, to advertise my blog. Then we\'ll see what happens. Question for thinkers and dreamers: What kind of people visit blogs anyway?  Oh so sad cassandra, don\'t worry I\'ll always be there to argue with you. That way you won\'t fall asleep in class hehe! And yeah, in case you haven\'t noticed, I HAVE been talking to you you know? Albeit it may be in the form of arguing. By the way, you were saying something about an expressions voucher?...  It\'s my first time (blog virginity?) making a blog so maybe there will be lots of things to say. And really I have no idea how to handle this stuff. The last time I made a site was primary school and believe me when I say I was good.. At that time. Now I forgot everything there is to know, but as long as this words get published, Honestly! Who needs music and pretty icons?  I\'ll have to be careful with what I say... Jeanie seems to handle okay though, maybe it\'s because it\'s all \"Ooooh! I got this Newee Pinky baggy! Oooh! and look, I also got this newee pinky skirt! And Look and Meeee and Hubbee!! Love him to bits!! Oooh!\"  Haha, I didn\'t say anything about her directly, so don\'t put words into my mouth. Besides, Jeanie is part of my microscopic list of true friends to, so I don\'t think she\'ll mind.  Speaking of which, I know the answer but is it possible to inflict this anti-social behaviour on Cassandra so she\'ll literally be happy that people ARE leaving her alone? Okay I promise this will be the last time I\'m talking about her; scared  kena  get whacked by her on wednesday.  To those people who are out to bring me down in the politcal race in the class of 2H, it\'s alright if you choose to find ways and means to search for things that I said wrongly and exploit it. I survived many of these sort of things. Oh and feel free to mock me if you find any bad English, I didn\'t realize that what I say have to be  scrutinized (cheem  enough?) all the time.  Right, so things end here. Meanwhile I\'ll be learning more about this blog machine thang and I\'ll just be living life. Conte Partero.  '),(16796,3582331,'29,June,2004','2004-06-29 00:00:00','   There has been a lot on my mind lately. I am just going to type and see what comes out.   I spent a day and a half with my mom and thought I was going to go insane. It\'s sad how depressed and lost people are without Jesus. It\'s hard to be around them sometimes. I love my family more than words could express, but sometimes I just want to run far far away.   I am leaving for Texas in about 4 and a half weeks and I am getting a little overwhelmed. There is so much I need to do and figure out before I go. I don\'t know where I am going to stay when I get there, where I am going to work, what classes I am going to take, anything....the reality of me going just seems to be crashing down on me all at once right now and I am trying my best to give it to the Lord, trust Him and relax. It\'s a little challenging.   I have been feeling a desire for deeper friendships. I have a few deep friendships, but I don\'t get to spend as much time talking with them as I would like and I would really like to develop some other deep friendships. I just want people around me who understand me, who I can be serious with and who I can be crazy and funny with. I want balance, honesty, love, trust. So many relationships are built or \"kept together\" with dishonesty and deceit and it\'s so sad. I just want REAL people who are willing to be REAL.   I miss so many people so much it hurts. I never knew I could feel this way about friends, about people I hardly know, but wish to get to know better. I never thought it was possible to miss someone so much, even when they weren\'t family. In reality, they could be more family to me than my physical family because they are my spiritual family. I just want to see them again, and I will soon....for a little while. I long for that family feeling, that understanding, common faith, encouragement, love. God has blessed me with so many amazing people, I am overwhelmed by His goodness and love for me.  Okay, I think this is long enough and random enough. I love you all....if anyone even reads this.... if not, that\'s okay too.   SMILE :-)  '),(16797,3582331,'27,June,2004','2004-06-27 00:00:00','   I want somone to really know me  I want to share Christ with random people  I want my family to understand me  I want all the childern in the world to have loving homes  I want every person who is new at coming to Christ to have someone to disciple them  I want families to be reunited  I want to see my dad  I want my mom to be financially okay  I want the lost to see hope in God  I want the unloved to feel loved  I want to rest in God\'s arms  I want to talk with my friends and say I am sorry if I have ever hurt them, to tell them I love them and to open up to them who I REALLY am.   I want to cry....cry for the hurting. The lonely, those without families, without hope, without love, those who are slowly perishing, those who are knowinlgly turning their backs on God.   I want to have arms big enough to hug all the children who have never been hugged  I want my life to make a difference in this world, not by my doing, but because I have died to myself and allowed the Lord to live in me.   I want a life that glorifies the Lord in everything I do  Is that all too much to ask......?  '),(16798,3582331,'25,June,2004','2004-06-25 00:00:00','   You don\'t see pictures of the hard times in your photo albums and scrapbooks, but they are what get you from one happy picture to the next.   Don\'t let go....hold on to EVERY MOMENT! No matter what it \"feels\" like or how hard it is. They are all a part of the life that God gave us. Be blessed everyday, and if you are not blessed everyday remember that it\'s your fault and not His. God WANTS to bless us every single day, but it\'s up to us to recognize and accept the blessing. Some times the blessing come in a disguise and we think it\'s a curse, but we must daily trust our Father.   I thank God for so many of my unanswered prayers. My life would be so much worse off if I would have gotten it \"my way\"  Be blessed and smile :-)  '),(16799,3582331,'22,June,2004','2004-06-22 00:00:00','   So life lately has been difficult and I am going to write about it. I am not going to worry about who reads this, I am not going to be fake or censor what I type today. I have had a hard week and I need to vent. If you don\'t want to hear real words from a real person who struggles with real things then please don\'t read on....  So, I am a sinner....oh my gosh...I admitted I sinned....everybody does. I have realized lately that I have a lot more sin in my life than I recognize and I am a lot more fleshly than I want to be. I also try to justify my actions WAY TOO much. I haven\'t spent near enough time with the Lord lately (basically none in these past few days) and it hurts my heart and I know it hurts His. I am almost ashamed to go back to Him because I feel I have so deeply neglected my One True Love. I feel emotionally drained and overwhelmed, my heart is heavy. I have been condemning myself the past week and I know most of it is a lie from the enemy and I feel ashamed that I have fallen in to his trap. I am struggle to breathe and keep from drowning in my own selfish world. I have not spent any time recently around Christians. I have been with family and friends that I love dearly and desire with all my heart for them to know the Lord, but it is a draining and spiritually drying place to be when I am with them too long.   I miss feeling the true deep sense of joy that I had at Teen Mania. I can sense and know that I have the joy of the Lord who has saved me from eternal hell, and I am thankful for that joy. But, there is a feeling of true joy, laughter, healthy happiness that I miss. I want to talk with someone...anyone who knows me....who understands me and knows and agrees with my faith in the Lord. There hasn\'t been anyone for a long time. I long to be held by my Father, to rest in His presence again, to cry to Him and allow Him to comfort me. Something has been keeping me from that.   Lord, help me. I am so sorry that I have hurt your heart by neglecting you. I am in pain in my heart because I am ashamed of the way I have been and I know it has not been pleasing to you. Help me run to you again, to feel welcome in your arms, I want you....I need you....I am DESPERATE FOR YOU! Please hold me. I am sorry, I am so sorry, Lord, forgive me, I am sorry.....I don\'t know what else to say Lord, I am so sorry, I love you so much and am sorry, I want to want you again. Help me please. I am weak, I am poor, I am so broken, but Lord I am yours....please hold me now.  '),(16800,3582331,'14,June,2004','2004-06-14 00:00:00','   I want to curl up in the fetal position and just cry for the sake of crying.  I want to throw myself on the floor and throw a tantrum.  I want to be a kid again and climb in my mom\'s lap and let her rub my forehead.  The more I pay attention to myself, the more I don\'t understand myself.  I am being emotional today. If a guy was talking to me he would probably roll his eyes and say I am emotionally needy...that\'s okay, guys just don\'t understand girls sometimes and that is okay. We don\'t understand them sometimes either.   There are a few people I would REALLY love to talk to right now, I always feel good after talking to them, whether it\'s on IM or phone or in person. There are just those people who I REALLY enjoy talking with.  '),(16801,3582331,'11,June,2004','2004-06-11 00:00:00','   I am just being random for the sake of being random. I am soaking wet, because some kid that I am watching spilt Apple juice on me. I am IMing with my friend Amelia in Canada....gosh I miss her. I miss a lot of people. Even those I have recently seen. I feel so blessed to have the people in my life that I do.    &nbsp;   I am posting this from email...it\'s really cool. I am learning lots about the computer. I hope it works.    &nbsp;   LOVE YOU ALL   SMILE  __________________________________________________ Do You Yahoo!? Tired of spam?  Yahoo! Mail has the best spam protection around  http://mail.yahoo.com  '),(16802,3582331,'10,June,2004','2004-06-10 00:00:00','   Things in my life seem to be going really well. I know what I am doing next year, I am financially okay (praise the Lord), I should be happy. For some reason I feel an overwhelming sadness right now. WHY??? I don\'t know. I mean, I do have the joy of the Lord and I am content with my circumstances, but there is the \"emotion\" of sadness lurking around me. I HATE IT! Maybe it\'s because what I see in sight is still too far away to reach. Maybe it\'s because I want to help so much, yet seem so helpless. I want the people that I love the most in my life to know the hope and joy I have in Christ. My heart breaks when I hear of hurting children, not only in the world, but right here in my home town, and there is nothing I can physically do about it. Families are broken, lives are lost....and here I sit, typing, thinking, crying....because I don\'t know what else to do. God is so much bigger than all of this and I have to pray and trust that His will will be done. If I dedicate each day to Him, lay down my agenda for His, I know He is faithful. The problem is, sometimes I don\'t lay down my agenda, sometimes I try to live this life on my own strength and still expect God\'s will to be done. That simply can\'t be. It\'s the daily taking up the cross, dying to ourselves and allowing God to live in us that will make the difference. So maybe this sadness is because I realize I have been trying too hard for too long to make it on my own and I realize that something has got to die....and that something, is me. It\'s a hard concept to think about. Dying....giving up control, being broken and weak. But it\'s the only way for God to truly live in you, and for Him to get all the glory He deserves. So now, I choose to die.....  '),(16803,3582331,'08,June,2004','2004-06-08 00:00:00','   So I got this survey thing from my brother Ashby\'s site and thought I would fill it out just for fun, since I have nothing better to do with my life. :-)   Body: ??? Height:: VERY SHORT (like 5ft 2) Natural hair color:: brown Eye color:: blue Number of siblings:: LOTS! 2 real brothers and many many bro\'s and sis\'s in Christ Glasses/contacts?:: Just got contacts, have been wearing either one or the other for about 10 years now.  Piercings:: 2 that are still there Tattoos:: not yet.... Braces?:: NEVER..Praise the Lord  FAVORITE Color:: blue and purple Band:: don\'t know. I like Third Day a lot Song:: Proud to Be an American....just kidding although I do get teary eyed whenever I hear it.  Stuffed animal:: My dog that I take on all my trips with me Video game:: classic mario bros. or tetris TV show:: Trading Spaces Movie:: I LOVE MOVIES..it\'s too hard to pick just one Book:: Bible *duh* Food:: Mexican (YUMMY homemade chicken enchiladas) Game on a cell phone:: the only one I have ever played is snake CD cover:: No Idea Flower:: Really bright tropical flowers (don\'t know names) and sunflowers Scent:: Coffee Animal:: Horse and little dogs Cereal:: Corn Pops Website:: ??? Cartoon:: Mickey Mouse  DO YOU Play an instrument?:: NOPE!! Someday I\'ll play the piano Watch TV more than 60 hours a week?:: Heck No, I do have a life Like to sing?:: yes...good at it? NO Have a job?:: I\'m a nanny (or a 23 year old babysitter) Have a cell phone?:: nope Like to play sports?:: sometimes, depends on who it\'s with Have a crush on someone?:: you\'ll never know.....(not really, crushes are from the devil Have any special talents/skills?:: listening exercise daily?:: Not daily...regularly, yes Like school?:: LOVE TO LEARN  CAN YOU Sing the alphabet backwards?:: YES I CAN....for real, thanks to my gpa! Stand on your tip toes without wearing shoes?:: OUCH!! Speak any other languages?:: A little spanish Go a day without food?:: yup....3 or 4 days Stay up for more than 24 hours?:: I can but it\'s not pretty Read music, not just tabs?:: I wish Roll your tongue?:: yes Eat a whole pizza?:: that\'s a lot of pizza...I don\'t think so  HAVE YOU EVER Snuck out of the house?:: It doesn\'t count as sneaking out if your parents really don\'t care Cried to get out of trouble?:: probably (I AM a girl) Gotten lost in your city?:: ha ha! Yes, I didn\'t know how to get to the mall Seen a shooting star?:: several Been to any other countries besides the united states?:: Mexico, Cananda, Philipines, Africa, Peru....and more to come Had a serious surgery?:: had my tonsils out and wisdom teeth pulled, don\'t think those count as serious though. Stolen something important to someone else?:: No, don\'t think so Solved a rubiks cube?:: not that smart or patient Gone out in public in your pajamas?:: YES! I used to go to Wal-Mart in the middle of the night in my PJs all the time Cried over a girl?:: yes (a friend) Cried over a boy?:: don\'t think so Kissed a random stranger?:: ??? Hugged a random stranger?:: YES. Hugging is great Been in a fist fight?:: yes, mostly with my brothers Been arrested?:: NO Done drugs?:: NO!no! Had alcohol?:: Yes Laughed and had milk come out of your nose?:: it wasn\'t milk, but yes Pushed all the buttons on an elevator?:: No, but was with someone who did Gone to school only to find you had the day off because of a holiday/etc?:: No, how stupid would that be.... Swore at your parents?:: probably...sorry Been to warped tour?:: Nope Kicked a guy where it hurts?:: unfortunatly....sorry Been in love?:: Love with the Lord, yes.... Been close to love?:: nope Been to a casino?:: Yes, my parents work in casinos, I grew up around them, have been several times.  Ran over an animal and killed it?:: Yes, I think it was an armadillo Broken a bone?:: YES, my left arm once and both my arms at the same time Gotten stitches?:: twice Had a waterballoon fight in winter?:: yep Drank a whole gallon of milk in one hour?:: YUCK Made homemade muffins?:: YES YUMMY Bitten someone?:: Yes....sorry Been to disneyland/disneyworld?:: Disneyland More than 5 times?:: probably more than 50 times. I used to work across the street and I had a season pass...I went ALL THE TIME, it\'s my favorite place! Been to niagra falls?:: No Burped in someones face?:: again...sorry... Gotten the chicken pox?:: YES..horrible, horrible experience  WHENS THE LAST TIME YOU Brushed your teeth:: Morning. Went to the bathroom:: 20 minutes ago Saw a movie in theaters:: about 3 weeks ago Read a book:: constantly reading Had a snow day:: snow....that would be nice Had a party:: a few months ago for my friends bday Had a slumber party:: Gosh I miss my friends.... Made fun of someone:: yesterday....in fun.... Tripped in front of someone:: all day today at the airport, I was wearing the wrong shoes Went to the grocery store:: last week Got sick:: don\'t remember Cursed:: depends on what you consider cursing  PICK ONE Fruit/vegetables:: fruit Black/white:: black Lights on/lights off:: off TV/movie:: movie Car/truck:: Car Body spray/lotion:: Spray Cash/check:: Cash Pillows/blankets:: blanket Paint/charcoal:: paint Chinese food/mexican food::  MEXICAN Summer/winter:: (depends on where I am) In Texas- winter, California-summer Snow/rain:: Snow!!! Fog/misty:: misty Rock/rap:: Rap Meat/vegetarian:: MEAT!!! Boy/girl:: both make great friends Chocolate/vanilla:: Chocolate all the way Sprinkles/icing:: icing  Cake/pie:: cake French toast/french fries:: French Toast. Strawberries/blueberries:: STRAWBERRIES are one of the best things God created Ocean/swimming pool:: OCEAN OCEAN OCEAN  Hugs/kisses:: HUGS...for now Cookies/muffins:: cookies Wallet/pocket:: pocket Pink/purple:: Purple Cat/dog:: Dogs  Long sleeve/short sleeve:: depends Pants/shorts:: Pants Winter break/spring break:: Winter Spring/autumn:: spring Clouds/clear sky:: Clouds Moon/mars:: Moon  FRIENDSHIP How many friends do you have?:: I like to think I have a lot What are their names?:: I don\'t want to leave anyone out Do you have a best friend?:: I have a few best friends Have you ever liked one of your friends?:: not liked liked...no Do you have more guy friends or more girl friends?:: girls, I think Have you ever lost a friend?:: oh yes Have you ever gone to an amusement park with a friend?:: Just yesterday with Megan Whats an inside joke between you and a friend?:: well, if I told you it wouldn\'t be an inside joke.....have you all seen Shrek? Have you ever gotten in a big arguement with a friend?:: Yes, not fun Whats the nicest thing youve ever done for a friend?:: ssshhh! Whats the nicest thing a friend has ever done for you?:: Invited me to church, where I got saved Do you miss any of your old friends?:: YES....A LOT!! What friend have you known the longest?:: Jeanna...my whole life Do you regret anything youve done to a friend?:: oh yes, but thank God for forgivenss If so, what is it?:: You don\'t need to know Has a friend of yours ever died?:: unfortunatly yes Whats the dumbest thing youve done with a friend?:: there are WAY too many to name What do you think your friends think of you?:: don\'t know  LOVE AND ALL THAT CRAP Have you ever been in love?:: not in love...not yet Do you believe there is someone for everyone?:: Not sure What is your idea of the best date?:: I don\'t want to date, I want to hang out with friends Do you think love is a load of crud?:: NO....Love is great Whats the best experience youve ever had with the opposite sex?:: being friends, being comfortable with them, having good conversations and hanging out  SILLY STUFF What is your favorite genre of music?:: I like oldies and Christian What time is it now?:: 10:07pm What day is it?:: Tuesday Whens the last time you called someone?:: About 4 hours ago How much money do you have right now?:: enough Are you hungry?:: nope Do you like parades?:: not really Do you like the moon?:: yes, especially in Texas What are you going to do when youre done with this?:: Go to sleep If you could have any magical power what would it be?:: I don\'t think I would want one Have you ever had a picnic?:: LOVE EM  DO YOU THINK YOU ARE funny?:: sometimes pretty?:: not really sarcastic?:: yeah, sometimes...it can get me in trouble though lazy?:: sometimes...it\'s from the devil hyper?:: after starbucks...yes friendly?:: I try my best evil?:: Heck no smart?:: not as smart as I could be strong?:: physically, not really, emotionally, most of the time talented?:: hm? dorky?:: definitaly   FOR OR AGAINST suicide:: NO...AGAINST love:: Love is all you need.... drunk drivers:: WAY AGAINST airplanes:: LOVE THEM war:: Not sure, in general war is never a good thing, but sometimes it\'s necesary canada:: LOVE IT...and I love Canadians...they are great united states:: I am proud to be an america rock music:: for gay marriage:: against. school:: Love to Learn surveys:: depends parents:: they\'re GREAT cars:: oh yes... killing:: WOAH! NO, AGAINST britney spears:: She needs the Lord coffee:: TOTALLY FOR....YES...STARBUCKS....OH YEAH pants:: Pants are good...  WOULD YOU EVER Sky dive?:: probably not Play strip poker?:: don\'t think so Run away?:: not any more Curse at a teacher?:: have done that, didn\'t get me anywhere Not take a shower for a week?:: If I was camping ask someone out?:: I don\'t think it\'s the girls job to do Lie to someone to make them think better of you?:: hope not Visit a foreign country for more than a month?:: YES! Go scuba diving?:: LOVE IT Write a book?:: I\'d probably try Become a rockstar?:: no Have casual sex?:: WHAT???HECK NO!  '),(16804,3582331,'08,June,2004','2004-06-08 00:00:00','   Okay, so I gave in and did the xanga thing, now I am giving in and doing the blog thing, it\'s the thing to do. I feel so cool! Actually I just figured this would be the place where I just vent about feelings, thoughts and \"deep\" stuff. Where I won\'t care who or if anyone reads it, I won\'t worry about grammer or spelling, I can just get on and vent. I am still going to do xanga (what can I say, I am addicted) but this will be more of a journal journal. So if you don\'t care to read it, I am fine with that. I think better when I write so this is more to help me think. We\'ll see how it works. :-)  Peace out Chels-dog!  '),(16805,3582331,'29,July,2004','2004-07-29 00:00:00','   I keep forgetting about my blogger....I probably won\'t write here anymore.  '),(16806,3582331,'13,July,2004','2004-07-13 00:00:00','   I feel very blah today. I want to cry, yet I want to laugh. I want to talk with a friend, yet I REALLY want to be silent. I don\'t know what to say. I miss people a lot. I love people a lot. There are people out there who don\'t know how much they have impacted my life. There are people out there that don\'t know I still think about them, care about them or that I ever did. That makes me sad. I want to have a good deep conversation with someone that isn\'t shallow, petty or selfish, yet I don\'t know what I would say if I had the opportunity.   I feel blah today!  '),(16807,3582331,'02,July,2004','2004-07-02 00:00:00','   So lately I have been talking to and reading a lot of emails and sites from people who are thinking about, wanting, desiring or actually getting married and having babies.   It seems to be a little overwhelming, EVERYONE lately has \"love\" on their mind. I realize as I am writing this, that possibly I, too, have it on my mind, I just don\'t seem to vocalize it like a lot of people these days.   I think this is for several reasons. #1) at this point in my life, it is not an option available to me, so why dwell on it and talk about it when it\'s not going to happen right now? That would just be feeding my emotions. #2) I have trust issues and I am afraid to let people really know how I am feeling and what is going on in my mind. #3) I assume no one cares....  I don\'t know, maybe I have issues.....  I do think about getting married, what it would be like, having kids, a family, working for God together, growing old together....I guess everyone thinks about that, but for some reason I have noticed it\'s been vocalized a lot more lately from EVERYONE....it must be that time of year.  '),(16808,3570813,'30,June,2004','2004-06-30 00:00:00','    Alton Brown will be at the orange county fair on Thurs, July 29 at 5:30PM. I will also have to find a way to be there. I think i may have a convienent 3:30 doctors appointment that day.  They couldn\'t possibly have picked a stupider time to schedule him. oh well, it\'ll be worth it.  anyone else want to come?   '),(16809,3570813,'29,June,2004','2004-06-29 00:00:00','    This weekend on \"While You Were Out,\" instead of paying the money for stained glass they decided to make something that would look similar. they took some sort of acetate, lined up gummy bears in an attractive pattern and melted them in the microwave. Now it was unclear how they enclosed these now melted gummy bears, but somehow they added another sheet of acetate, bent it into a half cylinder and made 2 wall mounted lamps out of it.  It was definitely cool, because it\'s not my house. I hope those people don\'t have ants.   '),(16810,3570813,'29,June,2004','2004-06-29 00:00:00','    I don\'t understand.   I\'ve been doing a lot of faxing lately and maybe that will be a separate post, but since I\'ve been standing by the fax machines I\'ve had lots of time to go through the box that they put resumes in. Currently there are  only 5 resumes/cover letters in the box but they all have very, very basic grammar errors. I don\'t understand why they couldn\'t just press spell check, or grammar check. Because it would have caught these problems.   Meanwhile my resume is pristine, clear and simple. I don\'t use big words for the sake of using big words (like 2:5 of the resumes). Everything is spelled correctly. When I write my cover letters I never say crap like \"Please call me so we can discuss how my capabilities can be applied to the realization of your companys goals\" [yes, company plural, not company posessive]  Am I doing something wrong? Should I make major grammar errors. Should I be wordy and vague? Should I sound like a person who has never written anything in my life and make it seem doubtful that 7:10 times I can form complete thoughts into complete sentences?  I am very glad I am going back to school. And maybe in a year I\'ll quit working and do it full time. Because one thing is obvious. I need to be in a supervising and hiring position. And if people call to inquire about  resumes they sent that weren\'t up to my standards then I will tell them that we don\'t hire people who have more than 2 errors in their resume and maybe they should think about having someone proof read it for them.  I don\'t understand why I can\'t get a better job if this is the caliber of people out there. Is it because I refuse to use buzzwords?  I\'m going to hang out by the fax and wait for more resumes, and memorize these people\'s names. If they get hired, I may just have to quit.  p.s. I\'m well aware that my blog is nowhere near perfect gramatically, and most of the time i can\'t be bothered to use capital letters. It is NOT a resume though, and the only job it will find me is when I become famous and they publish it as a book.   '),(16811,3570813,'29,June,2004','2004-06-29 00:00:00','    3 razor blades really are better than two. who knew? I\'m glad I ran out of my normal razors and found something new around the house. I\'m now a believer.   '),(16812,3570813,'29,June,2004','2004-06-29 00:00:00','    Just when I thought I was safe, and wouldn\'t have to eat at McDonalds for a while they bring back $1 chicken nugget tuesdays. I will not go today. Next week is another story though.   '),(16813,3570813,'25,June,2004','2004-06-25 00:00:00','    I finally did it. I was a little too hasty with my spring rolls and I finally got a small burn on my wrist from the 350 degree oil. Oh well, it\'s hardly noticeable and I\'m hardly dying. It was totally my fault, no fault of the deep fryer, except for having the most useless basket in the history of deep fryers.  The spring rolls were great though, and a steal at $.99.   '),(16814,3570813,'25,June,2004','2004-06-25 00:00:00','    The McDonald\'s neopets promotion is now over. I\'m sad because I still need 5 more. I\'m happy because I can stop eating the food.  I\'ve pretty much given in and decided that I\'ve been totally screwed by the lady on ebay and i\'ve lost my $28. And I will never see those neopets i ordered. So, yesterday (thurs) I made a last ditch effort to collect the neopets the old fashioned way. I went McDonald\'s yesterday. A new record for me. I went at lunch time to the wonderful one that gave me 2 unis last week. but alas, they were completely out of neopets. they did however try to offer me at least 4 different types of toys they had left over from previous promotions. I didn\'t even try going to the other one by my work, because they ran out of toys last week. I considered going to the one in the mall, which i could have reasonably made it to and back without being more than 5 minutes late. But I went 2 hrs before closing on wednesday, and they only had wockys, and lots of them.  So, after work I headed to the McDonalds that I thought would be my heaven. And it turned out it was, on wednesday. Wednesday they opened their 2 \"customer satisfaction\" boxes that held some of all 16 toys. ALL 16!!! But yesterday they only had grundos. And, this is the ugly frog thing with antennas that I\'d been avoiding the most. But, I bought it anyway.  Then I tried another McDonald\'s on my way home. This also proved useful, because I picked up a red yurble and a yellow quiggle. I also tried minute maid strawberry soda, which I highly do not recommend. uggh! I don\'t know if i didn\'t pay attention to every other mcdonalds i\'d been to the the last month, but I\'d never seen it before. And hope to never see again.  Then I got greedy, I thought I\'d go a little out of my way to try the mcdonalds by el camino community college. I figured these were people desparate to prove to someone that they were grown up and legitimate. That mcdonalds would have plenty of toys left. But no, they were completely out. They did have some kind of cool wizard lego toys that i almost bought, but with 3 happy meals in my car already i decided against it.  I tried to ask people to buy their display of neopets in the plastic case (the promotion was over after all). But the only one who had a reasonable, informed answer told me that they couldn\'t sell those toys. The give them to the workers after the promotion is over. I should have totally gotten a part time job at mcdonalds. I bet i\'d really like making egg mcmuffins.  I need 5 more neopets, and I guess I\'ll end up buying them individually on ebay. Unless anyone has them to donate to my cause. Factoring in the techo that Stephanie is going to give me these are what I still need:  Aisha Blumaroo Moehog Lupe Shoyru  I don\'t feel so bad, I made a good show of trying to get them. The only one I had an opportunity to buy, but turned down because I thought i was getting it on ebay was the moehog.  p.s. I have an extra yellow ixi if it\'ll make anyone trade me for one that i need. I\'m also considering listing my cloud scorchio on ebay. Last night one sold for $10.50 plus $3 s/h   '),(16815,3570813,'24,June,2004','2004-06-24 00:00:00','    Malibu rum has new flavors. and a funny new radio commercial featuring rastafarians giving job advice.  The point is, the classic coconut-pineapple taste that you get from ordering malibu and pineapple. can now be acheived, with much less liquid by having a stiff drink of malibu coconut and malibu pineapple. they also have malibu mango, and i\'m not quite sure the world is ready for mango rum. i know they\'ve been increasing in popularity ever since that 1996 win in the pilsbury bake-off. but, i just don\'t know.  now, a few years i swore off alcohols in the 40 proof range. this meant goodbye girly things like midori and malibu. then i realized you could get most of the taste more of a kick when you just added some vodka. silly me for taking so long to realize that. even the stupid beauty bar knew that, you can get malibu, vodka and a splash of pineapple juice served in a martini glass and a crappy manicure for $10, plus tip. for a better $10 martini and manicure i recommend the lighthouse cafe in hermosa beach on wednesdays.  so here\'s my new drink idea--not that it\'s all that new or different.  1 part malibu coconut 1 part malibu pineapple langers coconut-pineapple juice to taste  this from the person who wouldn\'t buy a bottle of tequila because i would never finish it, now is relying on owning 2 flavored rums. maybe when i open the school themed bar....   '),(16816,3570813,'23,June,2004','2004-06-23 00:00:00','    urlLink Berkeley legalizing prostitution?   Obviously they can\'t legalize it, because they\'re just some dumb city. But they can decriminalize it, i guess. Maybe those massage parlors will open again. And, any poor kid who somehow ends up on a web cam having sex with a prostitute will only have to be embarassed, and not worry about legal action against him. And that gay purple teletubby will not only be welcomed openly by the city council, he can have a prostitute while he\'s visiting too.  An open letter to the Berkeley city council.  Seriously, city council. I don\'t know why you would decline to comment on legalizing prostitution, you do so much other crap. Afraid you\'ll all finally be ousted for pissing off too many people? We know your opinions on teletubbies, and we know you hate students, renters. What about prostitutes? I need to know, so i can form the opposite opinion.  I don\'t care anymore. I\'m never going to live in Berkeley again. But really, do something useful for a change. Stop bitching about how the University takes over your precious little hippie hovel and realize that without the University you might not even have a hovel. And that thing about not being able to \"own\" pets. Pure crap. Maybe I should come steal your dog. Since you don\'t own him, it\'s not really stealing, is it? Work on apartment buildings near campus, bring in some corporate money to help. You know what also would be helpful, allowing businesses to stay open past 6pm without a special permit. Focus on things you can change, not the things you can\'t. And remember that law about sitting on the sidewalk? Don\'t do any more crap like that. You\'re as useless as the ASUC, but older, so you should know better.  If you stop debating useless and stupid things, you\'ll probably have more time on your hands than you know what to do with. Let me offer you a suggestion: Instead of debating, and passing bills and such post your dumbass ideas on a blog. It will take up your time, you\'ll still allow people to hear your useless ideas, and it will improve your typing speed. Believe me, this blog thing works to relieve your boredom.  p.s. I just heard about  urlLink w ketchup,  you may want to purchase some so you can debate it and declare it evil. If you do, can I please buy a bottle off of you? It only makes sense to buy in bulk, and I don\'t eat very much ketchup.    wow, i must be really bored today.   '),(16817,3570813,'23,June,2004','2004-06-23 00:00:00','    Now for the least surprising news ever:   urlLink Close Female Friends Help Women Through Depression   help me! i need mental stimulation. give me suggestions of text heavy web pages (so it will sort of look like work) that will amuse me.   '),(16818,3570813,'22,June,2004','2004-06-22 00:00:00','    What\'s wrong with me? I\'m posting about video games. I bet you all never thought you\'d see the day.  My sister somehow convinced my dad to fish our old atari computer out of the garage, because she thought it would be fun. I don\'t know the the exact model off hand, but this thing is a good 20 years old. And because we had this, we never got a nintendo, therefore not exposing me to loads of pop culture that would have proven useful for me in the future.  and then i read  urlLink this  this on yahoo news today.  I haven\'t spent much time with the new/old toy yet. But, so far, IT\'S B-O-R-I-N-G yet oddly time consuming. You think i\'d just stop when it\'s not fun anymore, but i just keep waiting for it to get good. For a person who spends no time playing video games, you\'d think they would be a bit harder. and maybe less boring. at least the music is good.   i have yet to play the classics like pac man, donkey kong, q-bert, dig-dug and pong. but if the games i have played, including popeye, frogger and food fight, are any indication, i should just sell the whole thing on ebay.  i guess video games are not for me. the new ones are too hard, and the old ones are too boring.  classic games convention. whatever!   '),(16819,3570813,'22,June,2004','2004-06-22 00:00:00','    Thanks to this Sunday\'s LA Times Magazine I now have a huge list of restaruants that sound good, but are too much trouble to go to, and also, are not inviting to the un-celebrity likes of me. And I like to think that being a girl, who on occasion can flaunt my sexuality to get me things would help me in these situations. But, this is LA and there are already a gazillion other girls, who are blonder and surgically enhanced trying to do the same thing.  Luckily, for me this drink is fully explained and pictured:   The libation is called You\'re a Pepper Too, and it  is  just the thing of you\'ve had the daylights slapped out of you for no reason. You get an 8-ounce bottle of the original Dublin Dr Pepper, along with two sidecars, one with Absolut Vanilla and the other with Monin brand organic vanilla syrup. And there\'s Dr Pepper Jelly Bellies to boot. Not approved for sniffling first-graders, but quite and attitude adjuster nonetheless.   In the picture it\'s served with an adorable giraffe swizzle stick. It looks fantastic. And with the non-celebrity substitutions of regular dr pepper, and torani vanilla syrup, you\'ll have quite the cocktail.  On a side note, I don\'t drink enough. It\'s not really worth it for me to buy entire bottles of alcohol, because they\'ll never get finished. I have half a bottle of absolut that i opened a year ago. I need lydia back, we encourage eachother to drink well.   '),(16820,3570813,'21,June,2004','2004-06-21 00:00:00','    i never go on vacation. yet, somehow, i\'ll be out of town on the day the cat olympics will be in santa monica. *sigh*   urlLink Meow Games   '),(16821,3570813,'18,June,2004','2004-06-18 00:00:00','    It\'s been a while since i\'ve had a frying update. Maybe next week i\'ll have something good to post, but for now:  trader joe\'s frozen crab cakes = delicious especially when served on a bed of cheddar mashed potatos (not deep fried)  also, I finally fried the last of these frozen hash brown patties that my dad accidentaly bought. they were completely raw, unlike the delicious pre-browned kind found in every supermarket and trader joes. Don\'t ever buy the completely raw kind. they\'re just not good.   '),(16822,3570813,'18,June,2004','2004-06-18 00:00:00','    This weekend, trying hard not to be \"LA,\" but failing miserably, I gussied myself up and took a trip to the Beverly Center. Now, the Beverly center, for those of you who don\'t know or have forgotten, is disappointingly far from the freeway. This is a very bad trait in locations, and usually keeps me from them. But, Beth was driving, so I was fine. We managed to get there, finally, purchased a few sale items and made our way up to the food court where we were going to the A&W restaurant to get a root beer shake. Now, the A&W store is a magical place. I\'d never seen one before the first time I went to the beverly center, and since then I\'ve seen them attached to Kentucky Fried Chicken establishments, but never stand alone. And, nowhere besides the beverly center did they have milkshakes. Root beer floats, yes. But not a milkshake in sight, not even the regular flavors. And let me say, as strange as it sounds, the root beer milkshake may be the best milkshake I\'ve ever had. And, I\'ve had a milkshake or two in my day. This is made with some kind of root beer flavored syrup.  So, we make our way up the escalator toward the food court, look around, look around again, and alas, there is no A&W in sight. Now I will never have a root beer milkshake ever again. And that makes me very sad.  Also, failing to not be LA, we stopped by Fred Segal, where someone on ebay said they carry a line called \"Barbie Vintage\" and I wanted to try on a shirt, to make sure it wasn\'t too small for me, the way this size Large Barbie  tanktop is at Hot Topic. Of course there\'s no Barbie in sight. There is however, a large pile of \"disney vintage\" shirts.  Also, in the disappointments category, no neopets yet. And I got an email from some freak who contacted me because he saw that I won an auction by the seller, and he won an auction and hasn\'t gotten his thing yet. I spent $28, and if i don\'t get my neopets, drastic measures may need to be taken. I have the mailing address, and I recieved an email confirmation that my payment was recieved, so the address is valid. Now may be the time to start plotting revenge. You all can help.  so, right now i\'m a good 700 calories lighter from not having a milkshake, and pissed off about ebay. Of course, that will all change when i find a set of high numbered baby-sitters club books selling for a low price.   '),(16823,3570813,'18,June,2004','2004-06-18 00:00:00','    i don\'t really like competiting myself. can you think of a more non-competitive profession than librarian? But competition is fantastic. Especially watching it on tv, or reading about it in the new book i\'m reading called  Cook Off- Recipe Fever in America  by Amy Sutherland. Apparently cook offs are quintessentially American. And I\'m pretty darn quintessentially American (in the good sort of way). Really, if we can think of something, we can think of a way to make it competitive.  Sit me down in front of the tv watching iron chef, junk-yard wars, animal face off, trading spaces, the pilsbury bake off, food festivals etc and i\'m set. people who are good at what they do make me happy. and watching them fight to the death (or for a new stove) is even better. And that word freak book about competitive scrabble. I love the crazies!   '),(16824,3570813,'17,June,2004','2004-06-17 00:00:00','    The McDonald\'s i went to yesterday has pictures of its charity projects on the wall and stuff. it\'s a little odd. but this little kid probably under 4 yrs old points at some old picture of someone and Sam the 1984 olympic eagle, and asks \"who\'s that\". and with only a very short pause the mom answered back correctly that it was sam the eagle. I thought for sure that she wouldn\'t know. She looked like the dumb type. Obviously she wasn\'t smart or she wouldn\'t be feeding her kids such food at such a young age. then I tried to figure out how i knew the answer was sam the eagle. i have no memories of the 1984 olympics. certainly, the olympics are boring now, they must have been horribly boring for my four year old self. and i don\'t know any olympic mascots since then. i\'m confused. was sam super popular, or is it just an la thing? i just used the phrase \"la thing\". arrgh!   '),(16825,3570813,'17,June,2004','2004-06-17 00:00:00','    So apparently torrance elementary schools got out at lunch time today. and all the children converged on the local McDonald\'s establishments. I was not aware that this would happen today.  First I went to the one on crenshaw and skypark (by the crossroads) and was jarred as I walked into a room full of children playing with their quiggles. there must have been at least 30 of them. and i went to the counter to ask what toys they had and the lady showed me a grimace driving a car and said that was the only thing they had. i\'m not blind! i saw the quiggles! not that i wanted a quiggle, but the neopet promotion doesn\'t end for an entire week. apparently all of those horrible children had got the last of the neopets. i hardly believe that, but we\'ll see. I needed a short pep talk from lydia at this point, but continuted on my way. I\'m debating writing to McDonalds about this. It got lydia a white ixi.  I finally got to the McDonald\'s on PCH just off of Hawthorne where they have a small room for kids and video games, so it kept the main part of the place less kid dense. And i saw a few kids with grundos, but I decided to wait in the relatively long line because I couldn\'t believe my luck could be that bad. And it wasn\'t. First, my order was taken by the nicest Japanese boy I\'d ever seen at a McDonald\'s. He was well groomed, friendly and generally helpful and good at  his job.  And he had unis! I got a red and a yellow. and lydia promised me her blue. that means i\'ll have all the primary colors, and I can breed them into secondary colors!! (okay, maybe it doesn\'t quite work that way)  The lovely Japanese order taker, gave me the unis. The order filler people also gave me two grundos. I was in such a good mood that I was honest and gave them back. Besides, no one likes the grundo.  Now I can finally stop going to McDonalds. I\'m going to get that set from ebay, eventually. And, I have doubles of all the ones I want doubles of, except maybe the aisha. But, it may be too late for that.  Two happy meals is more food than I realized. It\'s too bad I forgot to get apples instead of fries in one of them.   '),(16826,3570813,'16,June,2004','2004-06-16 00:00:00','    It\'s been exactly a week since I won my neopets auction. And It\'s been exactly a week since I\'ve heard anything from the girl who has the toys. And she hasn\'t left me feedback. Will I get my neopets, ever? This I can\'t blame on the former president.  Meanwhile, while on ebay I checked out the auctions for the complete set of 109 neopets toys:  the ronald mcdonald house made $1,715.00 and the neopet charity made $1,225.00  that\'s $15.73 per toy. if you go to a cheap mcdonalds you could get 7 happy meals for that price. not too shabby for some cheap marketing toys.   give me 5 years. then I should be able to competitively bid on things like this.   '),(16827,3570813,'11,June,2004','2004-06-11 00:00:00','    Seriously, I do not get paid enough, really, to make me want to answer the phone from non-mafia people. When there\'s a chance the Nigerian mafia is calling there\'s no way in hell I\'m picking up the phone. I don\'t care what my boss says. So I smile, and I nod and let her go to lunch confident in the fact that if Claudio from Chile calls or Mike from Scotland calls that I\'ll answer. They can leave a message, they\'re nice guys, and will completely understand that I\'m not answering the phone because I want to live.  I only like the Al Capone type mob, or maybe Sonny from General Hospital. But Nigeria, there\'s no glamour there!  Now to explain myself: One of our sales reps got a lead from this shady guy in Nigeria. We\'re pretty sure it\'s a money laundering scheme because we quote them absolutely ridiculous prices and they\'re willing to pay without even trying to bargain. We ended up saying that we can\'t sell to Nigeria, but this guy keeps calling.  I so do not get paid enough.   '),(16828,3570813,'11,June,2004','2004-06-11 00:00:00','    Since my internet connection is down, and posting to this blog was supposed to be my new hobby, i think it might be time to tell you all about my sunburn. WARNING: this may be more info than you need to know.   the guys at victoria\'s secret are marketing geniuses. and they have a line of bathing suits. i got the best one of the lot. just the bottom, because i\'m still not comfortable with those ridiculous pieces of triangle fabric that they consider a top. and it\'s black, with magenta trim. AND... a heart cut out on the butt. Genius! it allows you to give in to their marketing campaign and wear a  skimpy bikini with major VS sex appeal and feel like tyra banks. Yet, you can still aspire to be good and cheerful and all that\'s right with the world, like a care bear.  I think I will be victoria\'s secret bear. I laid out for an hour on sunday, completely covered except the heart on my butt. It\'s friday and my heart is still pink as ever. I really have to admit that my ass is adorable. I didn\'t have much of a strong opinion about my butt before. But now that I\'m practically a care bear, it\'s great. Also, I want to spread joy throughout the land. It\'s a good thing I don\'t know any care bears songs, because I would be singing them. It\'s also good that I\'ve been taking belly dancing, because I bet it\'ll help with my care bear stare.   '),(16829,3570813,'10,June,2004','2004-06-10 00:00:00','    Now, I like Ronald Reagan as much as the next person. Maybe more, because of my elementary school brainwashing. He was my first president and I\'ve had nothing but good feelings toward him, because I was 0-8 when he was in office and not exactly capable of analyzing his attitudes and policies. I just knew what I learned in school about civic duty and all that. The president was an good, important man. And, It\'s very sad that he had alzheimers, and it\'s sad that he died. But we don\'t need to have a federal holiday! Some things transcend people and should carry on as scheduled, like mail!!!  Mail is often the highlight of my day. I get catalogs featuring necessities like leg lamps, and postcards and sale ads from Joann\'s promising me 40% off any single regular priced item. I even get a few bills, but they don\'t mar my enjoyment of mail, especially when you\'ve ordered something online and are waiting for it... like neopets! I mailed off the money order to a girl in irvine yesterday. That means, she could have potentially recieved it today, potentially mailed off the neopets today and I could potentially have them tomorrow. But no, the ex president had to go and die, and they had to shut down the government. And I won\'t have my neopets until Monday at the very earliest, but probably tuesday or wednesday. Now, based on what I know about my first president, which is deplorably little, I\'m sure that he would want me to have neopets. He was a regular guy with a little longing for the cowboy live he loved so much in movies. And I think he\'d understand, and encourage the neopets to get to me quicker. So, when I die, I want the mail to continue. Also, it\'d be really cool if I could still have mail but close down all the schools. Because I understand, I want people to get the day off to sit by their mailboxes waiting for their packages.  In the meantime, I\'ll see you at McDonalds.   '),(16830,3570813,'10,June,2004','2004-06-10 00:00:00','    Hine says that shoppers crave unpredictability. Something on sale one day is full price in another blah, blah, blah, and that makes things interesting:  You never know what you will find in a dollar store, only what it costs. (Hine 105)   '),(16831,3570813,'09,June,2004','2004-06-09 00:00:00','    Barbie has a line of clothing for adult women that they\'re selling in Japan and could possibly sell in America.(although I\'m afraid they\'re japanese shaped adult women and not me)  anyway, thanks to lydia for finding this website. if you can read it, please tell me what it says:  http://www.sanei.net/brands/index.html  click on \"Barbie\"   '),(16832,3570813,'08,June,2004','2004-06-08 00:00:00','    from yahoo/AP  KORNEUBURG, Austria - Here comes the bride, and there goes her veil â€” all 1.7 miles of it. Eva Hofbauer went to her wedding on Wednesday claiming to be wearing the longest bridal veil in the world, the Austria Press Agency reported.       Her groom could only watch, robbed of any attention, as Hofbauer was assisted by hundreds of young people who carried the veil into the church in Korneuburg, a town on the Danube River just northwest of Vienna.    Mayor Wolfgang Peterl said he hoped the veil would earn a place in the Guinness Book of Records. The previous record-holder was said to be a Japanese bride whose veil was 1.2 miles long.   '),(16833,3570813,'08,June,2004','2004-06-08 00:00:00','    what if you took a twinkie and injected it with nutella...  and then deep fried it...  also, i went to the jelly belly website to see if they had anything to say about Ronald Reagan. They did. But this is much more interesting.  http://jellybelly.com/Cultures/en-US/NewsEvents/NewsStories/Wedding+Beans.htm  somehow i think i\'ll manage to have a jelly bean free wedding. although i\'m not promising anything to the kind people who have said i could plan their wedding on the TLC show.   '),(16834,3570813,'08,June,2004','2004-06-08 00:00:00','    is our increasingly short attention span a good thing? http://www.marinij.com/Stories/0,1413,234~24409~2197660,00.html   '),(16835,3570813,'08,June,2004','2004-06-08 00:00:00','    so, i\'m trying to stop being crazy about the whole neopet happy meal thing, but things like this make it very, very hard.   urlLink http://cgi.ebay.com/ws/eBayISAPI.dll?ViewItem&item=5902391736   my tuition for next year is just over $600 per class. I can assuredly say that i want to go to library school more than i want 109 neopets. i think.   urlLink http://www.neopetsfoundation.org/   do you think they need someone to wear the chia costume?  the 20 minutes later update:  i bought the set of 16 neopets on ebay for $28 (free shipping). That\'s $2 cheaper than the buy it now price. At the 44 second mark it was only 23.50. darn those last second ebay people! but now i can relax a little, and stop forcing myself to eat happy meals. thank goodness that next months promotion is going to be awful and i can stop buying them then.   '),(16836,3570813,'08,June,2004','2004-06-08 00:00:00','    things fried in my house in the last few weeks:  successes: cod catfish corn dogs onion rings w/beer batter onion rings w/corn batter broccoli won tons donuts bananas black berries strawberries (tastes like and inside out funnel cake) chicken croquettes  failures: fries (this is my fault for not giving the oil enough time to heat up) cookie dough strawberry hershey\'s kisses  mixed results: pickles (i don\'t like them to begin with, so i couldn\'t make the final decision)  things to try soon: twinkies whatever the american equivalent of a mars bar is oreos ice cream salmon ham and cheese sandwiches pb&j sandwiches corn bread batter more potato things   '),(16837,3570813,'08,June,2004','2004-06-08 00:00:00','    What I\'m reading now:   I want that! How we all became shoppers  by Thomas Hine  From the front cover:  Shopping has a lot in common with sex. Just about everybody does it. Some people brag about how well they do it. Some keep it a secret. Most people worry, at least a little, about whether they do it right. And both provide ample opportunities to make foolish choices _______________  You don\'t  need  Barbie. You do need water, but you don\'t need bottled water. Nobody needs a ceramic sculpture of mother and daughter hippopotamuses, but at under eight dollars, few would label it a luxury. -Page 16  ___________________  One study revealed tha tif you put mena nd women on treadmills the man will walk faster. But if you put the same couple in a mall, the woman will walk faster. Women feel that they are accomplishing something when they shop. Men don\'t. -Page 27  *****************  Books I returned to the library yesterday.  *The ultimate brownie book : thousands of ways to make America\'s  favorite treat, including blondies, frostings, and doctored brownie mixes Weinstein, Bruce, 1960-   I made cormeal brownies--surprisingly good--  and I made S\'mores brownies-- slightly disappointing   *Salt : a world history Kurlansky, Mark.   Very interesting, but i read all 400+ pages and there wasn\'t really a wrap up. I learned a bunch of things, but no point was made  *Driven : how human nature shapes our choices Lawrence, Paul R.   good-recomended by Alton Brown. Not quite a beach read though.  *Retro desserts : totally hip, updated classic desserts from the \'40s, \'50s, \'60s, and \'70s Brachman, Wayne Harley.   Very disappointing. not enough pictures, and the recipes were much more complicated than they needed to be  *My 1,000 Americans : a year-long odyssey through the personals Morton, Rochelle.  quick, fun, easy read. A british lady living in America goes on 1,000 dates with people who answer her personal ad. She used the rule of calling every third message in her voicemail box, no matter how crappy they sounded. And she went on a date with them as long as they didn\'t sound dangerous, or were pleasuring themselves on the phone with her.  *Food festivals of Southern California : traveler\'s guide and cookbook   apparently southern calif. goes all the way up to monterey. I love food festivals and want to visit them all. the book could be better though.   '),(16838,3570813,'08,June,2004','2004-06-08 00:00:00','    also on the apple cider kick:  mix 3 parts apple cider (trader joe\'s  brand of course) and 1 part butterscotch schnapps. it may be the best drink ever. Not particularly alcoholic, but delicious.  I haven\'t tried it, but i bet if you add in some vodka it will be legitimately alcoholic. or i wonder how spiced rum would work. hmmm....   '),(16839,3570813,'08,June,2004','2004-06-08 00:00:00','    Hello, welcome.  Just though I\'d let you all know the best donut recipe ever. I stole it from someplace online. I used trader joe\'s apple cider. I think that makes the difference. Don\'t be fooled by the baked apple cider donut recipe that\'s floating around. If it\'s baked it\'s a small cake, not a donut!  Vermont Apple Cider Doughnuts  1 cup apple cider 1 cup sugar 1/4 cup solid vegetable shortening 2 large eggs 1/2 cup buttermilk 3 1/2 cup all-purpose flour 2 teaspoonbaking powder 1 teaspoon baking soda 1/2 teaspoon cinnamon 1/2 teaspoon salt 1/4 teaspoon nutmeg vegetable oil or shortening for frying  2 cups confectioners\' sugar 1/4 cup apple cider  Boil apple cider in small saucepan until it is reduced to 1/4 cup, 8 to 10 minutes; cool.  Beat sugar with shortening until smooth. Add eggs and mix well, then add buttermilk and reduced cider. Stir together flour, baking powder, baking soda, cinnamon, salt and nutmeg in another bowl. Add to liquid ingredients; mix just enough to combine.  Transfer dough to lightly floured board and pat to 1/2-inch thickness. Cut with 2 1/2- to 3-inch doughnut cutter; reserve doughnut holes and reroll and cut scraps.  Add enough oil or shortening to fill a deep pan 3 inches; heat to 375\'F. Fry several doughnuts at a time, turning once or twice, until browned and cooked through, about 4 minutes. Remove to paper towels with slotted spoon.  For glaze, mix confectioners\' sugar and cider. Dip doughnuts while warm; serve warm   '),(16840,3570813,'30,July,2004','2004-07-30 00:00:00','    Oh my gosh, I can die happy now.  I took off work yesterday and headed 30 miles south to the orange county fair. I\'ve been to the fair probably over 10 times. and this was definitely the best one ever. quite possibly in the entire history of the fair.  before we even got in the gates some nice old couple gave the 4 of us 2 free tickets, so we got in for half price, a fantastic way to start a fantastic evening. i was sure it was my good karma from bringing 3 people who had never been to the fair.  we got in, took some pictures in front of retarded grape signs and headed off to eat some australian battered potatoes... the thing about the fair is that it changes enough to be interesting, but everything is always the same. and even though i hadn\'t been for an entire year i knew exactly where the potato shack was. the potatos are basically why i go every year, they\'ve been a favorite of mine since before i realized that deep frying is the best method of cooking ever. and i got to share my potatos.  on our way to the potatos we passed by a curious sign though. it was the kind they have in department stores with a metal frame that holds a small poster sized piece of cardboard advertising sales, or in the case of robinsons-may it holds the picture of the store manager. this sign however, and the 5 others we saw, informed us that if we wanted to come back friday we could meet the great Arnold, governor of this lovely state. I of course made my usual 10 second tirade about how he should be fixing the budget and letting kids go to college instead of hanging out and eating potatos. and then quickly forgot about him because there was only one man on my mind, and boy is he a man.  I am more in love with Alton Brown now than I was 24 hrs ago. If only he wasn\'t married.  We passed by the stage that Alton was going to be at when we got there at like 3:30-4:00, there was some band singing oldies songs, i thought there were suspiciously a lot of people there for that show. but i never would have guessed they were waiting for Alton. Then we ate our potatos, looked through the children\'s room and came back out to see mobs of people around the stage 50 minutes before his scheduled time of 5:30. So i let my friends go have fun, and i set out to find myself a vantage point. no amount of $1 rides (before 6 on thurs) or corndogs was going to make up for me not being able to see alton. and after unsuccessfully looking for an actual seat (you\'d think one person would be able to squeeze in somewhere) i found the best spot ever. it took a little while, and a little squeezing in where someone a few inches taller wouldn\'t have dared. but i managed to make my way to the corner of the stage behind a little old lady who was shorter than me, and some  35ish lady who was definitely legal midget status, although she was porportioned like a regular person. Then half way though the show the little old lady had enough, and i got her spot at the stage!! i was so ready to die happy.  So, after standing in my cramped bit of space for 45 mins, Alton comes out, and everyone forgets their troubles. Or at least i do. He comes out and takes a video of the whole crowd cheering with his camera phone and tells us all that in a week we should be on his website. Isn\'t he cute! oh yeah, you know how on tv he\'s always wearing a hawaiian shirt, or a bowling shirt to match his retro kitchen accessories. he wasn\'t. he was wearing a white shirt and a kilt. the boy was wearing a skirt. and i have to say that at one point he was close enough that i could have looked up his kilt, but i managed to stop myself. apparently it\'s something called a utilikilt  urlLink www.utilikilts.com . and they\'re popular with a certain crowd. When asked why he was wearing a kilt he said, \"Because I have great legs,\" and indeed he does for someone who turns 42 today.  So he starts off with like a monologue, about him coming to the fair, and last year\'s fair (apparently he was here last year and i just didn\'t know) and he\'s being cute and charming, self-depricatingly funny and generally someone i definitley want to know, and have children with... and i was in heaven. then he starts making fun of people in the audience, they ugly stuffed dog on the floor, the guy with the mohawk, the lady with the fringy parasol. i totally puffed out my chest,  because i found the perfect use for my \"talk nerdy to me shirt\" (not that i really needed help, because the t-shirt for being so dumb fits me awfully well). but alas, he didn\'t notice me. but my friends made me feel better when i told them this later by saying, it was because he thought i was cute that he didn\'t make fun of me, and you don\'t make fun of people when you\'re staring at their chest. (unless you\'re the lady we passed who was wearing the hat that said \"got tits,\" and yes she did.  then he started \"cooking\" there wasn\'t anything special. in fact, i wasn\'t really impressed at all. he had to use grapes, because the fair was grape themed. he did make a grape smoothie with frozen grapes, muscat wine and mint that i might try. but grape gazpacho, that\'s just retarded no matter if the man i love is making it or not. he got some kid volunteer to help him make the crust for his no-bake grape cobbler. it didn\'t look horrible, but it didn\'t look great. and this fat nerdy kid who wore a fanny pack with hand sanitizer dangling from the outside that alton made fun of was just in heaven. it was cute. then he picked some cute, very orange county type 35ish lady to help stir/melt some butter and brown sugar. he let her drink some of the grape smoothie, out of his glass! that\'s practically like kissing him! or so they say in carls jr commercials. anyway, she poured her mixture on top of his thing. then she held the pan while alton scraped out the last bits. and he called her a \"limber vixen.\" that\'s like my fantasy. me and alton cooking together, he scrapes out the pan, he calls me a \"limber vixen\" and then we really go test out how limber i am. if i was on stage with him, i probably would have just melted. dropped the hot pan on my foot and had to go to the hospital where alton would come visit me and then we\'d get married. his live show was even more fun than his tv show. for lots of reasons i\'m sure. it was also a bit more raw. he made a lot of drinking jokes and was just a bit less polished and family friendly, nothing actually dirty though. there were kids. in fact one group had a sign that read \"alton, you could out cook bobby flay in your sleep\" he took the sign as a souvenir, then made a joke about if anyone had a sign for him that said that about a \"mr. lagasse\"  after his demo he had a Q&A. and there were Alton freaks like there are comic book/sci fi freaks. it was awesome. 2 of the nerdiest possibly guys were there asking very technical questions regarding one aspect of one show. unfortunately they weren\'t so attractive (the way you\'d imagine), and meeting them would not have made a good alton substitute. anyway, like a good fan, i was carring his book around, hoping he would sign it. but apparently everyone else was too. the autograph line started in the back, and i was in the front. people were actually standing in the line while the show was happening. if i wanted to stand there for 3 hrs i probably could have had my book signed.  but i wasn\'t that crazy, i\'d already wif he comes back next year, i\'ll know to stand in the line first, since i\'ve already seen him up close. he comes out with a baking book in october. that is the most exciting news i\'ve heard since i learned he\'d be at the fair.  so after his thing was over i gave up on the autograph line and went to see the zebra, i didn\'t believe there was a zebra, but there was, and it wasn\'t just a painted stripey horse, it was a real zebra. the saddest zebra i\'d ever seen, stuck in a cage in the petting zoo. petting zoos are for farm animals, goats, sheep, pigs, chickens... even alpacas, but not zebras. it was cool to see a zebra, but he made me sad. then i got a fried twinkie to try to east my sorrows, because a budwiser (fair sponsor) just wasn\'t going to cut it. and it was good. the first fried twinkie i\'ve had. it was VERY lightly battered, and had a little strawberry jelly on top. it stopped tasting like a twinkie and  was more like a donut. the filling was all melty, and much better than the room temperature cream filling.  after this we went back, and the alton line hadn\'t gotten any shorter. so we got up close took a picture \"with\" him when he turned around and went on our way. i was able to drown my sorrows in a blue icee, because in southern california we\'re civilized and have lots of blue food to choose from.  Last night I dreamed of Alton. we were at the fair, at the same stage, but i was standing on the aisle in the back row. and he does that celebrity running thing where he\'s jogging by touching hands with everyone on the aisles. so he does that all the way down. and stops at me, and since i\'m at the end he just starts talking to me and he\'s smart and charming and funny.... He\'s wearing his usual hawaiian shirt and pants. and that\'s pretty much the dream. I\'m just giddy to be talking to him. ahhhhh.....  i realize now that i should have waited in the autograph line. i could have gotten him to sign my ass and then have it tattoed on. then i could be a care bear on one cheek and a cabbage patch kid on the other.  i\'m still giddy. i was *this* close to Alton. yea! it was the best thing ever. like if zack were to get a swimming pool full of cane sugar dr. pepper while everyone else in the world had a glass full of mr. pibb.   '),(16841,3570813,'29,July,2004','2004-07-29 00:00:00','    I don\'t know if anyone else has noticed, but there is a new car color on the market and it\'s metallic orange. My theory on car colors is that cars should be metallic. White, and black to a lesser extent are not good car colors because they don\'t shine in the sun. Metallic orange is really going too far though. My favorite color is dark gray. I\'ve never called cars sexy before, but nice sleek dark gray cars really are.  It took me a while to decide my opinion on this orange color. In the beginning I saw far too few of them. But now, i\'ve decided. It\'s the color of metallic vomit. There\'s a guy who works in my warehouse who drives one. How an 18 year old can afford a shiny new car I don\'t know. But maybe I should have bought him a pound of the vomit jelly bellys to match his car.   '),(16842,3570813,'27,July,2004','2004-07-27 00:00:00','    Does south-northern calif. have something against blue?  I tried to buy a blue icee at bonfante gardens, there were 2 kiosks that sold strawberry and cherry. 2 red flavors, that never happens! and the strawberry icee is definitely the worst flavor ever. much worse than candy cane.   then i tried to buy a blue snow cone at the garlic festival. silly me assuming that the 2 most likely colors are red and blue. luckily they didn\'t try to sell me a garlic snow cone, and they had bubble-gum, but it was pink.  I think by signing some pact with the devil to give up flourescent blue food gilroy got the right to have more vanity license plates than should be in a small town.   '),(16843,3570813,'27,July,2004','2004-07-27 00:00:00','    urlLink http://atxcases.com/Cases/devilcat.htm   i acted too soon and bought a computer that doesn\'t fit.   '),(16844,3570813,'27,July,2004','2004-07-27 00:00:00','    &nbsp; I didnâ€™t get to fully experience the wonder of the Gilroy outlets because we wanted to try to get home at a reasonable hour. But I will tell you about the 2 shops I bought stuff in:  Jelly belly outlet: They had 3 different bags of belly flops. 2 were the boring fruit flavors and 1 was all garlic (uggh!) so I decided against them. They were also selling all the harry potter jelly beans in the individual containers the way they do at candy stores. You could buy a full pound or more of just vomit, or dirt or earwax. I thought about asking if anyone buys them when theyâ€™re not in a cute little harry potter box, but I thought I would be happier if I didnâ€™t know the answer. I managed to escape the brightly colored wonderland only buying $3 worth of jelly beans. None of which involved harry potter or garlic.  Easy Spirit  Most of you know what ugly shoes I usually wear. Mostly sneakers every day at Berkeley. This is because I have very low tolerance for foot pain. I managed to find a new pair of sneakers I liked for only $40. I would have bought 2, but they only had one in my size. Also I bought a very comfortable pair of sandals that arenâ€™t very attractive. But theyâ€™re starting to grow on me. Theyâ€™re beige colored and look a lot better with the khakis Iâ€™m wearing today than they did at the store.  There was also an In-and-out. It may have been the fastest one Iâ€™ve ever been to.  &nbsp;Conclusion: insufficient data   '),(16845,3570813,'23,July,2004','2004-07-23 00:00:00','    It was HOT at the garlic festival. Like 90 degrees with little shade. But that wasnâ€™t really their fault. Also, there was a lot of traffic, but that wasnâ€™t their fault either. It actually was surprisingly well run. You park in a huge lot and get shuttled to the fair entrance. The line for the shuttle was long, but went quickly. And the shuttle was a full on charter bus with plush seats and A/C. Not some rinky dink school bus.&nbsp; There were plenty of lines to buy tickets to get in, and I didnâ€™t have to wait at all. The food lines looked long, but the booths were so well run that it didnâ€™t take long at all. It was a miracle that such a well run fair could spring up in the middle of nowhere. But I guess theyâ€™ve had years of practice.  And the thing about the middle of nowhere is that people are friendly. Usually Iâ€™m suspicious of friendly, but since I had driven around enough to see just how isolated Gilroy is I could understand. They were all just happy not to have to look at the same 20 faces they see day in and day out. The guy who sold me the garlic shot glass seemed genuinely interested in where I was from and my trip to the bonfante gardens. I guess they get bored. Another thing to support this theory is that most people have super nice, new houses. But we passed this one run down shacky place that had 3 satellite dishes. They knew what was important tv, a connection to the outside world.  I tried the garlic ice cream, and surprisingly, it didnâ€™t make me want to die. It wasnâ€™t the worst thing Iâ€™ve ever tasted. It tasted like vanilla with a garlic after taste. Actually it wasnâ€™t ice cream at all, it was soft serve. This seems like false advertising. Maybe Iâ€™ll get my own ice cream maker and try to make actual garlic ice cream.&nbsp; The garlic jelly bellys however, were quite awful.&nbsp; I was trying to figure out who I could buy a full bag of the garlic belly flops for. But I donâ€™t hate anyone that much.  I ate a combination plate containing a garlic sausage, garlic mushrooms, garlic chicken pasta stir fry, garlic bread and garlic shrimp. I also tasted the calamari (surprisingly unbreaded and unfried), and pesto garlic pasta.&nbsp; Later we passed by the fried garlic bologna sandwich booth, I wasnâ€™t hungry, but since I read about it in the paper I had to try it. I canâ€™t tell if it was a disappointment because I didnâ€™t like the bread, because I expected the bread to be grilled along with the bologna, because I wasnâ€™t hungry or if I just didnâ€™t like it.&nbsp; I have to say the most disappointing thing was the bun for the sandwich and my sausage. I donâ€™t know why, I just didnâ€™t like it. But thatâ€™s an awfully good sign that I didnâ€™t like the pre-packaged food, and I liked everything they cooked there.  I saw the garlic bobblehead doll. I wanted to buy it because I read about it, but it was just so ugly that I couldnâ€™t bring myself to do it. Poor ugly garlic man.  The biggest disappointment was that the garlic man was no where to be found. This was probably good for him, because he might have died from heatstroke, but sad for me. (meli if you have a picture with the garlic man, send it my way, I may just photoshop myself in)  We stayed shorter than I would have liked, but just about as long as we could because of the heat. Maybe if it was an ice festival we could have stayed longer.  Conclusion: fun, but Iâ€™m glad the whole trip wasnâ€™t just for the garlic festival, that would have just been a freaking long drive.   '),(16846,3570813,'23,July,2004','2004-07-23 00:00:00','    Much cuter than Disneyland. Much shorter lines. Most of the park was reasonably cool and shady even though it was 90 degrees on Saturday.&nbsp; Unfortunately it has many fewer princesses and much less cool souvenir stuff to buy than Disney. Iâ€™m in love. Although, I think itâ€™s definitely a place for women and children. None of the rides are scary, and itâ€™s awful cutesy. Probably, the only one who reads this blog who will love the park is Meli (if you havenâ€™t seen it, go, go now!).  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  For those of you who havenâ€™t bothered going to the website.  urlLink www.bonfantegardens.com  Bonfante gardens is a theme park with an agricultural theme. The main/ first attraction was the circus trees which are really a bunch of trees grafted together to make shapes and stuff.&nbsp; Michael Bonfante made them himself, was the only one who knew how, and took the secret with him to his grave. This makes him a big jerk! And I donâ€™t like him, I hope heâ€™s spinning in his grave about the fact that paramount is running his park and selling their crappy souvenirs.  The kiddie ride part of the park is equally great.&nbsp; Most of the rides are food themed: artichoke, strawberry, apple, banana, garlic, mushroomâ€¦ and unlike every other theme park in existence that makes all of their rides shiny like a carousel horse, many of these rides were matte. And they had very much attention to detail.&nbsp; Fine lines, shading. The artichokes even had the little pointy tip debossed and painted. It is also the first theme park that Iâ€™ve noticed artichokes being sold, granted they were deep fried, but still impossible to imagine being anywhere else. They had a pedal boat attraction, but of course it wasnâ€™t a regular pedal boat, they were duck pedal boats. There were mallards, yellow ducks and swans and they looked so ridiculously adorable floating around the pond.   The strangest part about the park was the total laid back California attitude. You could pretty much do whatever and nazi teenagers wearing leiderhosen or old fashioned dresses werenâ€™t around to stop you. They didnâ€™t offer any instructions on how to drive the pedal boat ducks, they just slapped you in a life vest and sent you on your ducky way. The park was so uncrowded that most of the rides all you had to do was wait for the current cycle to finish and youâ€™d get right on. They made sure you were safe and that your belts and bars were all correct, but somehow it seemed to be more laid back, and that made it run more smoothly. The only thing they were picky about was the 1 adult per strawberry car limit. Even though two of us could have more than safely been under the weight limit. And the words â€œplease keep you hands and arms inside the car at all timesâ€ were heard so very little, it was refreshing.   Another great thing about the park is that there is a small mid-way section with games and over half of them involve you winning a prize just for paying your $3. This is how we almost ended up with a car full of stuffed toys, but last second managed to talk ourselves out of it. We did, however, come home with velvety plush roses with a bendy stem.   Conclusion: go, go, GO!   '),(16847,3570813,'23,July,2004','2004-07-23 00:00:00','    I try not to complain, or at least be amusing when I complain on my blog so people will still read it. But, not this time:  Traffic out of LA on Friday night was surprisingly light. Why? I canâ€™t really tell. Maybe everyone already left for Gilroy on Thursday, because everyone in the whole freaking world seemed to be there.  Saturday morning it took us&nbsp; about 90 mins to get to Bonfante gardens. It took us less than 30 minutes to get back to the hotel because we skipped the freeway and took surface streets.&nbsp; We took those same streets on our way to the Garlic Festival and probably saved ourselves over an hour.&nbsp;   After leaving the festival we stopped by the Gilroy outlets, unfortunately they were on the way to the freeway and there was traffic. More unfortunately, there is a highway 152 that mapquest told us to take. Itâ€™s one lane each way. And thereâ€™s a yielding thing that caused a 15 mile backup which took us a good 90 mins to get through.&nbsp; So, we finally thought we were out of the woods, made it to the 5, made it to the grapevine, and thereâ€™s some major accident that makes traffic stop. Literally, people were getting out of their cars and hanging out on the freeway. And by this time it was very, very late, and we wanted to die. But no one could run over us and put us out of our traffic misery because no one could move. I got home at 1am. The trip that was supposed to take less than 5 hours took 8. Thank goodness that I donâ€™t have to commute to work on the freeway anymore. Because I think I would rather have quit my job than face traffic again.  &nbsp;Conclusion: I need a private jet.   '),(16848,3570813,'23,July,2004','2004-07-23 00:00:00','    urlLink Eyeglasses Double as Chopsticks?   I admire their creativity, but why?   '),(16849,3570813,'22,July,2004','2004-07-22 00:00:00','    urlLink Women Say Single Life Is Better?    Maybe they\'re right. It\'s totally not worth it to get married if you have to watch sports on tv.   '),(16850,3570813,'22,July,2004','2004-07-22 00:00:00','    Luckily I\'m not as quick to buy things on the internet as I am when it\'s sitting in front of me at the store.   urlLink Girlie Men Shirts   '),(16851,3570813,'21,July,2004','2004-07-21 00:00:00','    I was driving to work this morning and I had just passed a truck, when I ended up stuck behind another one. But this one was different. It said \"Shred-it mobile and on site document destruction\"  urlLink http://shredit.com/   Here\'s a bit from their web site:  &nbsp;We provide you with the security of knowing that your confidential materials are destroyed completely â€“ right at your doorstep. From our free locked security consoles, to our uniformed and bonded Customer Service Representatives to our mobile shredding trucks, your security is guaranteed  and i recommend this picture:  urlLink http://shredit.com/why/benefits-enviro.html   who knew that existed as a service? who knew there was a large enough market for it that it can exist as a service? and why do they need such a big truck? I guess this is why I haven\'t struck it rich yet.   '),(16852,3570813,'21,July,2004','2004-07-21 00:00:00','    I am a bad person. I may actually have a medical condition of some sort, because my mom has it too.&nbsp;Or quite possibly it\'s a gender flaw. But I can\'t myself from buying things I don\'t need, or will really ever use.  A year or so ago I bought a $4 tank top from Hot Topic that reads \"I *heart* shy boys.\" Why did i buy this? because it was cute, and cost $4. Not because I\'d actually wear it. Months and months later I finally decided it was reasonable to sleep in. So I do. and I\'m very proud that I was able to put my dumb purchase to use, even if you can\'t put it in the clothes dryer.  But yesterday, oh boy, I bought a $5.50 clearance shirt from ross. It\'s brown and in big white glittery lettering it says \"Talk nerdy to me\". This I cannot wear to sleep, because it will shed glitter in my bed, and it\'s not particularly soft. It\'s not even that attractive.  If anyone wants to match me there are at least 30 more, and they come in small, medium and large. The also have shirts that say \"too young for ashton\" \"ashton\'s future crush\" and \"you must be this tall to ride this ride.\"  Unfortunately, or maybe fortunately, for a certain person I know who has a boyfriend named Justin, they didn\'t have any Justin shirts. If I had a boyfriend named Justin I probably would go out an buy the $40 shirt from urban outfitters that says \"I did Justin, 3 times\"   '),(16853,3570813,'21,July,2004','2004-07-21 00:00:00','    Great!  The box is pink and has a squirrel with big&nbsp;pink cheeks&nbsp;wearing a chef hat hiding behind an apple. So I had to try it. And it\'s good. It tastes good, apple-y and sweet, but it smells even better. And, now that I can finally breathe through my nose again, I appreciate it even more. Maybe I\'ll try honey next time.  The entire box only has about 60 calories and 2 grams of fat.   '),(16854,3570813,'21,July,2004','2004-07-21 00:00:00','    I can\'t believe so much fun is happening in one weekend. I\'m going to the garlic festival in gilroy. But for those of you lucky enough to be in LA there is the tofu festival. Apparently there will be a guy dressed as the \"tofu ninja\" which alone seems worth the $8 admission. Huell Howser and the food network are also supposed to be there.    urlLink http://www.tofufest.org/    and lets not forget the cat olympics:    urlLink http://www.meowmix.com/newsevents/goldmedal/about.asp    i think some people i know should go to the events and give me a recap.   '),(16855,3570813,'20,July,2004','2004-07-20 00:00:00','    Yesterday I got a letter from ING Direct. that bankish thing that has a orange sphere as it\'s main image even though i think it actually has a lion as its logo. Apparently I\'ve been pre-approved for.... no, not a credit card like everyone else sends me.... a mortgage. me, pre-approved for a mortgage. I\'ve spent the last few days seriously trying to figure out how i can afford an apartment, and they want me to buy a house. Somehow they didn\'t mention how much of a mortgage i\'m qualified for.  I want a house, more than anything, really. But it\'ll be a while before I can afford one. Maybe my credit line will still be there when I\'m ready.   '),(16856,3570813,'19,July,2004','2004-07-19 00:00:00','    For a few years now, basically ever since I heard of it, I wanted&nbsp;to try botox under my arm to kill my sweat glands. I can\'t really justify the cost now,(or the fact that i\'m willingly injecting posion into my body)&nbsp;&nbsp;&nbsp;but It was on my wedding plans list. Weddings make perfect excuses for mostly everything.  &nbsp;  But, this procedure could be equally as useful.&nbsp;&nbsp; &nbsp;&nbsp;&nbsp; &nbsp;  urlLink Podiatrists can plump you up for stilettos   People fattening their feet so their shoes are more comfortable. This kicks ass! I guess we\'ve come a long way from Cinderella\'s stepsisters cutting off parts of their feet so they fit in her shoe.   '),(16857,3570813,'19,July,2004','2004-07-19 00:00:00','    at $9 a bottle I expect a longer lasting nailpolish. It\'s actually really a good consistency and makes my nails feel very strong when I\'m wearing it. It just peels off too easily. And not just chip, but peel. This I thought I was over with that bonnie bell kids stuff \"brush on, peel off.\" &nbsp; conclusion: next time I\'ll buy $3 nail polish from the drug store and buy a $13.50 lipstick to push myself past the $19.50 mark for a free gift. or maybe 2 pencil sharpeners at $1 each.   '),(16858,3570813,'19,July,2004','2004-07-19 00:00:00','    urlLink http://www.bonfantegardens.com &nbsp;21.99 discount online,&nbsp;$7 parking-- no discount,&nbsp;sat. hrs 10-8  &nbsp;   urlLink http://www.gilroygarlicfestival.com/pages/tickets.html &nbsp;free parking! yea! $1 discount coupon with northern calif. pepsi cases. possibly steal stickers off cases in grocery store? hrs 10-7  &nbsp;   urlLink http://marriott.com/property/propertyPage/SJCMH &nbsp;courtyard marriot morgan hill. indoor pool! checkout noon.  &nbsp;   urlLink http://www.gilroyvisitor.org/home.asp &nbsp;visitors bureau &nbsp;   '),(16859,3570813,'19,July,2004','2004-07-19 00:00:00','    urlLink http://home.earthlink.net/~mndctrl/Ax2004/nyanko.jpg  &nbsp; better late than never, I guess   '),(16860,3570813,'19,July,2004','2004-07-19 00:00:00','    I\'m going to try a new labeling system, and see how I like it. It will make it easier for you to sift through my blog on those days when I post 5 times. I am worried that my newly formed creative title writing skills will once again disappear, so maybe i\'ll make that the first line of the body.... anyway... &nbsp; I spent most of the weekend on the couch watching tv deciding that I was dying of tuberculosis, malaria or SARS, but somehow I managed to survive, despite all of the crap I watched. By far, the worst show I saw was MTV2\'s from Wack to Mack. Now, I should have gotten a clue by the fact that it\'s an original MTV2 series, but I didn\'t quite realize that until at the end of the makeover the kids said \"thank you MTV2\" in whatever vernacular they were using. &nbsp; Now, I\'ve seen a makeover show or two in my day. And in every makeover they start out somewhere between hideous and mediocre and they aways improve.&nbsp; Always. Leave it to MTV2 to be on the cutting edge and take a guy from looking horrible, to making him look equally horrible, yet different. Quite possible this is because my taste runs more classic or preppie than hip hop star. But, honestly, his parents couldn\'t have liked his new look much more than his old one. They were just happy he chopped of his stringy disgusting long hair. And the before picture of the girl was her wearing some kind of logo tracksuit not much different than the after look of the guy. Also, they went to Jacob the jeweler, who is apparently mentioned in some kind of popular song. I didn\'t quite understand if they got to keep the 75 carat total diamond cross on a diamond chain or the 8 carat diamond ring and matching earrings. but, wow. those were probably the most expensive things those kids had ever touched. mtv doesn\'t have that kind of budget i\'m sure. now that i think about it, it\'s pretty mean. taunting these kids with what they can\'t have... &nbsp; anyway. this is a horrible show. at least the girl got a good haircut out of the deal. I think clinton and stacy need to have the learning channel call mtv and give them some advice. &nbsp; Yes, I know I\'m old. and square. &nbsp;   '),(16861,3570813,'16,July,2004','2004-07-16 00:00:00','    an award for the best spam title today: stop pie. it almost made me click on it to see what they were trying to sell me. almost. but who would want to stop pie. i think pie is pretty swell. &nbsp; also, this morning i was pretty sure i was going to die of tuberculosis, even though i\'m not entirely sure what it is. i think it involves coughing up your lungs or something, and that\'s how i felt. this means i must be on the road to recovery. &nbsp; and, now that I\'m not dying I can look for love on my next flight.  urlLink http://story.news.yahoo.com/news?tmpl=story&amp;cid=572&amp;e=3&amp;u=/nm/airlines_love_dc   '),(16862,3570813,'15,July,2004','2004-07-15 00:00:00','    Catfight: Women and Competition by Leora Tanenbaum   I don\'t recommend it. So, far i\'m not quite 1/3 of the way through the 300 pages, i hope it will get better, but i doubt it. As you know, I judge books by their covers, and I seriously mis-judged this one.   I don\'t like being told that i\'m oppressed, or being held down by the man, or whatever. I am totally privilidged, and have never felt any real effects of sexism. I can form my own opinions, thank you. The author is a woman with absolutely no self-esteem, where every guy on the street, every magazine ad, every hollywood profile makes her feel inferior. She needs to get over herself. The book is peppered with examples of her youth where she let herself feel inferior to just about everything. Unfortunately for the american public, she got over this just enough to make herself write a book, but not enough for the book to be good, or helpful to me.   Let\'s take a look at some of her insight:   You\'d think the desire to be liked would make girls and women bloom beneath the rays of another woman\'s admiration. Yet the opposite is often true: We worry that admiration could lead to envy and resentment. When a female friend or colleague tells us how much she respects our work, we\'re inclined to say, \"Oh, it\'s not as important as what you do.\" If she admires our outfit, many of us habitually respond, \"This? Uch, it\'s just something I got on sale five years ago\" or \"really? you do?\" the words \"thank you\" don\'t occur to us. God forbid we come across as complacent or proud.   I don\'t know about this \"us\" and \"we\" thing. I personally tell people thank you every time they compliment me. IT\'S THE FREAKING POLITE THING TO DO! And, I believe that I deserve every compliment I get. But I don\'t go out of my way to get compliments, because I don\'t base my self worth on what other people say. And when someone compliments my outfit, i will usually say something like \"thanks, i got it on sale for $10 at ross\" I\'m not belittling my self by saing i got it on sale, in fact i\'m proud of it, and i\'m offering people tips so they themselves can find a great skirt for $10.    So   Dear Leora Tanenbaum, and whoever wrote that Men are from Mars... crap,   Please, next time you feel free to categorize an entire gender, leave me out of it. I will make my own rules for myself. And I don\'t want to listen to you telling me about how I am.   Do you know why I\'m not an astronaut? Because I had some really bad science teachers in High School that made me lose interest in science. Who\'s fault is this, my own, for not keeping the interest up on my own. I wasn\'t given bad teachers because i was female, both genders equally filled the class. Not because I don\'t feel like I can compete with men in a male dominated profession. And not because my lack of a penis makes me bad with figures, numbers and concrete examples.   Do you want to know why I\'m not a fireperson. Because I prefer a career path based on brains rather than brawn. Not because I couldn\'t be strong enough if I tried. Not because it\'s \"men\'s work\"   Why do I not look like a supermodel? Because I choose to spend my time and money pursuing other things, like education, and having a healthy relationship to food. If my full time job was to be beautiful, then I could certainly succeed. I could excercise, I could starve myself, I would have people paid to dress me, fix my hair, fix my makeup, and I could be up on a pedestal (a tall one to make up for my height).   In the immortal words of Mattel \"We girls can do anything. Right, Barbie?\" And if Barbie could talk the astronaut, rock star, WWF panda protector, fashion model, presidental candidate, ballerina, teacher, army recruit, equestrian, and fantastic date would say \"Abso-fucking-lutely.\"   Once again, please, in your writing, please change the word \"women\" to the phrase \"women, except Cynthia and others like her\"   thank you,  -------   Conclusion: gender studies is crap  Conclusion 2: It\'s not my fault so many women have such low self esteem. I\'m doing my part.  conclusion 3: i should write a book and show these cowering jerks how it\'s done.   '),(16863,3570813,'14,July,2004','2004-07-14 00:00:00','    Book Review: The Thoroughly Modern Married Girl: Staying Sensational After Saying \"I Do\" by Sarah Bliss  The spring/summer 2002 i read through a stack full of books with titles like bombshell manual of style, swell girls guide to the good life, fantastic flirty and fun, the single girl\'s guide to being fabulous, good fabulous pink stuff...--okay, so i\'ve made up some of the titles. but you get the picture. there are gazillions of these books on the market. none of their content is very different from eachother, so you really should pick based on the pictures.  but i was at the library last week and saw  The Thoroughly Modern Married Girl  which is exactly the same as all those books, except it\'s for young, fabulous and married women. i think it actually expects you to have read, or at least have a good knowledge that the previous books exist, or maybe the opposite. because the author takes exactly the same tone as all the rest of these books. you are already fabulous, now just make a few changes to be fabulous and married.   for some reason my mom got a subscription to ladies home journal and in the last week we got 2 issues. this is a very, very bad magazine for old boring people. one of the articles was how the author tried to go on a romantic date with her husband, they swore they wouldn\'t talk about bills, children etc. 5 minutes into dinner they started talking about the children, and then decided to go straight home after dinner to be with the kids. the concusion was not: oh my gosh, i\'m old, boring and horrible. it was: yup, that\'s the way it is, i\'ve grown out of all the fun and romanance and everything good, and i wouldn\'t have it any other way. the recipes aren\'t even any good.  now if you cross a much younger ladies home journal without all the depressing boring old people articles but still the same concerns about how to have fun, how to hang out with another couple when one is your friend and the other is the dumbest, worst person on earth etc with any one of those single girl books written by single girls who think the whole world revolves around them, you\'ll get this book.  i could have done with less condesending fake situations and a few more real life examples, recipes, and song lists.  conclusion: don\'t buy it for yourself. but it wouldn\'t make a half bad wedding shower gift. i read/skimmed through it in less than 2 hrs. the cover is cute, and there are a few cute illustrations inside. also, maybe someone who didn\'t spend months reading pastel colored self help/entertainment/reference books might find this book new and novel.  second conclusion: i should write a book. too bad i\'m lazy.   '),(16864,3570813,'14,July,2004','2004-07-14 00:00:00','    This is fantastic!   urlLink 1000 Flowers   I\'m considering making the $20 donation so I can get the nail files.   '),(16865,3570813,'13,July,2004','2004-07-13 00:00:00','    I woke up again today with a sore throat. I totally thought I was over that on sunday, but I guess not. I had 2 large cups of tea this morning, but they did nothing to soothe my throat. Then I debated if i should eat my lunch or go buy Jamba Juice and vitamin shock my system into getting well. Somehow my leftover ravioli and meatballs sounded better, even though I didn\'t have an appetite when I started, and it looked rather disgusting all cold and congeled. But as I was eating i started to feel better. It wasn\'t just the food, because I ate breakfast. It was totally the meat. Meat made me feel better. But only while i was eating. It\'s back now. I just came back from the store though, and I\'m hoping if i drink 1/4 gallon of orange juice in the next 3 hours it will help.  And speaking of oj, what happened to donald duck orange juice? It was no where to be found at the vons i went to. that stuff was good.  Now i\'m thinking i should go home and marinade a steak in orange juice for dinner. Although, I bet that works better without the pulp.   '),(16866,3570813,'12,July,2004','2004-07-12 00:00:00','    I was totally wrong. I got my two favorite atari games working. And video games are for me! at least Mr. Robot and Spare Change are. For those of you who haven\'t been blessed with these wondeful games, Mr. Robot is an obstacle course type thing that involves trampolines, conveyor belts, escalators, fire poles to slide down and alien fires to kill. It also has the best song I\'ve ever heard come out of my atari. Spare change takes place at a carnival. You have to fill your token box with 18 tokens before the zerks fill their piggy bank with 5. If you do they put on a show for you. ahhhh. good times.  p.s. does anyone know if i can get the frogger song as a cell phone ring tone?   '),(16867,3570813,'09,July,2004','2004-07-09 00:00:00','    Who excercises/dances in a thong. Not me. But, apparently the girl who stood in front of me in my belly dancing class does. it was quite visible through her white pants. if she was my friend i\'d tell her. but since she\'s a stranger who takes too small steps i\'ll let her suffer the humiliation.   '),(16868,3570813,'08,July,2004','2004-07-08 00:00:00','    I saw Supersize Me this week at the $3 theater. I was happy that i only had to pay $3 but even happier that i didn\'t have to get my ass on the freeway so i could go see it somewhere trendy like los angeles or santa monica. there were about 35 people in the theater with me. all of them were of average weight or below. not one single fat person was there.  it was exactly the movie i was expecting, no surprises. and it kept me entertained. i was very happy to see that his final conclusion was basically \"stop the supersizing already, the food is bad enough in it\'s regular portion\" this adds sense to the rumors that mc donalds is doing away with the supersizing because of the movie. no news story that covered it mentioned that was the guy\'s conclusion, so it didn\'t make any sense to me that the end of supersizing was a reaction to the movie. but apparently it does answer that. i blame shoddy reporting.  and now i want fries.   '),(16869,3570813,'08,July,2004','2004-07-08 00:00:00','    i may be getting a sore throat and i do have a headache, but i\'m certainly not at the delirium stage of my cold yet. although i must be, because i\'m sitting at work, and i swear that i\'m listening to a conversation about whores. yes, women who get money for having sex. surely, this isn\'t happening in my office.   '),(16870,3570813,'08,July,2004','2004-07-08 00:00:00','    I am not registered with a political party, I hoped that this would save me from getting all sorts of political crap in the mail the way my parents do. I am officially registered as \"Decline to state,\" and have been since I\'ve been for the last 6 years (wow, I\'m old). And as you all know, I do lean a bit left. But, God bless the republicans!  First Lydia made me aware of the  urlLink W Ketchup   then, today  urlLink yahoo news  reminded me of the ketchup and introduced me to   urlLink star spangled ice cream   You can buy 4 bottles of this ketchup for $20 including shipping. This seems a bit pricey, but the price goes down when you buy in bulk.  However, for the republican ice cream in various politically charged flavors you will be paying $76 for four quarts including shipping. That\'s $19 a pint. That\'s almost 8x than the low trader joe\'s price of $2.50 a pint for Ben and Jerrys you can. If i\'m going to pay $76 dollars for ice cream I\'m going to buy it by the bucket and fill a large bathtub and swim around. The republicans better have damn good ice cream if they expect to sell it.  And, better yet, for $5 a pint, plus $100 label design, plus shipping that includes dry ice packing that they warn you will be \"signifigant\" you can get your picture on the ice cream. Now, that would be a fantastic wedding favor. It beats my personalized water bottles idea by far.   God Bless America! The land of the people who will spend $19 on a pint of ice cream and not bat an eyelash.   '),(16871,3570813,'08,July,2004','2004-07-08 00:00:00','    So, apparently the weekend of July 24 is the cool weekend to hold events. It is the weekend of the Gilroy Garlic Festival which I\'m going to. It\'s the weekend of the cat olympics in Santa Monica. And it\'s the weekend of the LA Tofu festival, which i knew, but didn\'t know until monday that they food network is going to be there. If they\'re there, can they also be in Gilroy? or have they already been to gilroy too many times and they\'re just going to eat tofu? I will be sad if the food network doesn\'t show up in Gilroy. But not sad enough to not ride in a giant garlic, and get my picture taken with a giant garlic man. They probably wouldn\'t let me in anywhere else afterward anyway, on account of my garlic breath.   '),(16872,3570813,'07,July,2004','2004-07-07 00:00:00','    Who knew fame news would spring up in 2004   urlLink Irene Cara\'s age?   '),(16873,3570813,'02,July,2004','2004-07-02 00:00:00','    The Pillsbury bake off (oh my gosh, Microsoft word knows that Pillsbury is a word that should be capitalized and auto corrected it!) was 2 weekends ago. This threw me for a loop, because they donâ€™t usually have it in the summer. And it was held in la. I wish I could have gone.  I will when I get my makeover show. I bet I could have done my makovering there.  This is the million dollar recipe:   urlLink Oats \'n Honey Granola Pie   Iâ€™m planning on making it soon. Iâ€™ll be the judge of it being worth a million dollars.  You can go   urlLink here  to see all the entering recipes.  Iâ€™ve decided I need to enter contests, and I need to win more prizes. Unfortunately, it seems that people really need to dedicate themselves to winning prizes in order to do it. And I just canâ€™t fully dedicate myself to a hobby that involves more than sitting on the couch. Iâ€™m hoping that when I start school again my need for procrastination can be filled by contests and cooking.  And, once again Iâ€™ll recommend  Cookoff: Recipe fever in America  by Amy Sutherland. Iâ€™m almost done with it and itâ€™s very inspiring, well not so much inspiring as making me want to win contests.   '),(16874,3570813,'02,July,2004','2004-07-02 00:00:00','    it was well stated in the movie mean girls, and i\'m not sure anyone really had the guts to point it out before in such mass media, even though it was a well known phenomenon. girls are just dying for the opportunity to look like whores and still be socially accepted. not real whores, because real women who sell their services, are doing it for the cash and not as a hobby. they donâ€™t put nearly the effort into their choice of clothing and lack thereof as the subset of anime girls. Now, most normal girls can only dress up in something truly skimpy one day a year, Halloween. Everyone will tell them how cool their costume is, the boys will notice them and the girls wearing more modest outfits will covet the attention or talk behind their backs, or both. For nerdy girls, this opportunity comes twice a year, Halloween and Anime Expo. And for the really geeky ones who stay in and hand out candy in their sweats, AX is the only chance they have.  Now, I have to say that this year either people were showing a bit more restraint, or I was blissfully not paying attention to the super ugly girls who were baring way too much skin. Besides a few girls and  urlLink man faye  I was pretty comfortable with peopleâ€™s clothing choices. The super ugly girls are sad to begin with, but dressed up in little more than underwear and they become the saddest population, and makes one want to avoid conventions just to stay away from the unpleasantness. So, un-self-aware yet aware enough that they aren\'t sexy and dying to be sexy these girls equate very little clothing with being attractive, and you feel bad for them, yet not too bad because youâ€™re forced to look at the train wreck that is them, and itâ€™s not fun.  To them I say, save up some money and then get a professional makeover, or sell your soul for a tv makeover.  Everyone will be better for it, and once youâ€™re attractive you can replace your fake anime friends with real ones. Just as a general rule, if you have to buy all of your clothes from the plus-sized section, please wear that clothing. You may be a lovely person, with lots of good qualities, you may even find a boy and get married one day. but youâ€™re not going to attract anyone based on looks alone, please stop trying.  Then there is the other, super minority, the super-hot girls that you know are way too hot to like anime and be at the convention. Why are they there? Itâ€™s a completely mystery, because they could be in the Bahamas living a life of luxury being kept by some rich hunk.  Maybe theyâ€™re too young, and the day they turn 21 theyâ€™ll be on that plane. Or maybe their robot will fly them there. To the girl in the red Chinese type dress who was showing a mile of leg, and the one in the furry bikini, god bless you for giving every geeky boy hope that one day he can meet someone half of your attractiveness and she will give him the time of day. Because these boys need hope. And those  urlLink MPS  boys are going home with a burst of inspiration because of Lydiaâ€™s  urlLink dress . Theyâ€™re going to win their next competition, and sheâ€™ll probably get a robot in the mail next week. And Iâ€™m very jealous of that. Note to the second mps boy from the right: I canâ€™t help you build robots, but Iâ€™m a girl. If you want, I can paint them, or help you pick out fabric for the insides. And I will hook up with you if you let me drive the robot, you can park it though.   I should skip this whole library school thing and take up cosmotology, fashion and poise. I really need to work behind the scenes at a makeover place/show. Iâ€™m pretty sure my true calling is making people better at being themselves. Screw this whole defending free speech, free information and promoting literacy crap.   '),(16875,3570813,'02,July,2004','2004-07-02 00:00:00','    who knew i was such a trendsetter? as many times as i wore those ridiculous earrings from the late 80\'s to the early 90\'s. as often as i carried my dumb ornament sized teddy bear to school in middle school. nothing ever caught on. but now i guess i\'ve blossomed. I am popular, a trendsetter if you will. People do things that i do. I experienced this a little with neopets. but this blog thing has really caught on because of me. i have to use my power for good, not evil. what should i make catch on next? suggestions?  I guess now is the time to put my toaster idea into action.   '),(16876,3570813,'02,July,2004','2004-07-02 00:00:00','    because all of you care, i\'m letting you know that i got my neopets last night. and they\'re better than i imagined. right now in my house there are almost 30 happy meal neopets. wheeee!  i love my neopets. and i like to pretend my cat does too.   '),(16877,3570813,'02,July,2004','2004-07-02 00:00:00','    there\'s a brand of vodka they\'re advertising in this week\'s Albertson\'s ads. I\'ve never heard of it, but it\'s called UV Vodka and it comes in 3 colors, flavors actually, icee flavors. Red cherry, Blue raspberry and regular. Perfect for the 4th. except they\'re vodkas and they weigh the same so if you tried layering them to get a red white and blue drink you\'d just end up with purple.  I haven\'t decided yet, but I think i like the blue raspberry alcohol concept.   '),(16878,3570813,'02,July,2004','2004-07-02 00:00:00','    I was listening to an interview on the radio this morning with Puff Daddy, or whatever his name is. And, I came to the conclusion that I would totally like him if he spoke proper english. I already like the fact that he\'s not ashamed to be extragavant. He just needs to learn to talk.  He apparently is also a big fan of the library, even though he insists on calling it the li-berry. He said he knew he wanted to be in music, so he went to the library and read books on it and found himself an internship because of that. And the rest is history. He talked about how great the library is for at least 30 seconds, which is a lot in a short interview.  Maybe when I\'m a librarian I\'ll get invited to his parties.   '),(16879,3570813,'12,August,2004','2004-08-12 00:00:00','    The tin of altoids i bought at the 98 cent store with the abominable snowman (labeled: abominable snowmint) on it is making my desk drawer smell like mints, so i had to switch it to a draw that didn\'t contain any other food. in the process i noticed that the serving size is 3 pieces. 3 pieces has 10 calories and 2 carbs. 2 carbs for 3 mints. now why would they ever make that sugar free crap if you could get something respectable like altoids at less than 1 carb per mint?   i fished the box of tic tacs out of my purse and found that one tic tac has less than .4 grams of carb. further investigation led me to find that 2 tic tacs are about the same size as 1 altoid, and that at 3 calories for 2 almost the same as an altoid.  now i\'m going to find some buttermints. those must be worse for you, because they taste better.   '),(16880,3570813,'11,August,2004','2004-08-11 00:00:00','    urlLink What\'s in a Name?   something to think about.   '),(16881,3570813,'10,August,2004','2004-08-10 00:00:00','    I start library school at the end of the month, in preparation for that I have orientation this Saturday. And in preparation for that, I had to complete the \"new student technology course\" this week. This course consists of 6 short documents you have to read, one evaluation of fill out, one super short assignment basically there to make sure you know how to upload a word document, and free run of the new student message board, which, when you start class for real will be full of class discussion relevant to class, but is now full of pictures of people\'s cats. I totally thought librarians and cats were just a stereotype, a stereotype based in reality is more like it.  So the assignment was to write a paragraph about yourself in a word document and then send the word document via the electronic \"black board\" site. Really, really simple, I\'m not sure how to even mess that up. So I sent in my paragraph yesterday, it was about me going to the garlic festival and trying garlic ice cream. And this morning I checked and I recieved 10/10 points. Yea for me. I can attach a file. And the person checking all the files wrote me back a few sentences so I could see where instructor comments go, how that works and stuff. She also said she\'d always been curious about garlic ice cream. Nice, cordial response, the black board works fine. Great. Then I notice that the average class grade is 8.73. Not that the grade counts for anything, but really how do you do that assignment wrong? Are there just a few losers who didn\'t attach the file correctly who got a zero? and that brings everyone\'s 10 grade down, or are there people who get less than the full 10 points? or are the grades just random to show a reasonable class average? I don\'t know. But I can only hope that the rest of my assignments are 1/10 this easy.  Meanwhile, this internet thing is great an all, but I\'m meeting these people on Saturday and I don\'t know what to wear. If I wear my cardigan will they think i\'m some freakish librarian stereotype? (not like they don\'t all have their cats at home). Do I dress for school, or do I dress for business? Should I carry a backpack, messenger bag or a large purse? Some things are so complicated.   '),(16882,3570813,'10,August,2004','2004-08-10 00:00:00','    I think this link should take you past the first page where you have to enter your birthday   urlLink Ciroc   anyway, this is the greatest vodka ever. somehow, it\'s made of grapes, the special snap frost grapes, from france. or something.  anyway, i read about it last week in the la times business section. it was an article about how the california pizza kitchen guys are planning to change the mood of their late night dining. it will still be kid friendly by day, but by night they\'re trying to draw in a hipper, cocktail drinking crowd. and they\'re starting with this magical vodka. I don\'t know if it\'s available at the CPK near you, or anywhere yet. But I tried it at the LA Food Show (1 restaurant so far in manhattan beach) and it was amazing. I was a little wary about trying it in a cosmo, because i thought the other flavors might mask it, but it was the first drink on the list, and i couldn\'t have been happier with it.  then the waitress left us for freaking ever, and i couldn\'t order a second drink. I wanted to try a ciroc and tonic to truly taste it. but my pocketbook thanks her for being inattentive.  I have yet to see this in a store, but I also haven\'t been looking yet. But, when I do find a bottle, I will own one.  My friends who tasted it with me tell me it\'s flavored like Japanese grape candy. I also need to try some of this candy.   '),(16883,3570813,'06,August,2004','2004-08-06 00:00:00','    I saw the cutest dog this morning. it was a miniature collie. I\'m pretty sure those don\'t exist, but i swear i saw one. Maybe it was some furry little lap dog that had the good fortune to be marked like a collie. But, that thing was cute. I don\'t think it could pull you out of a well though, nor would it even bother trying.   '),(16884,3570813,'04,August,2004','2004-08-04 00:00:00','    urlLink Texas Sex-Ed Textbooks Face Contraceptives Battle   This is not really surprising news. but the news isn\'t the interesting part, this is:  They say the books are lacking. For example, one textbook under review advises that a good way a teen-ager can prevent a sexually transmitted disease is to get plenty of rest so he or she can have a clear head about sex and choose abstinence.   '),(16885,3570813,'04,August,2004','2004-08-04 00:00:00','    Splenda did me in again. But I\'m learning my lesson.  I went to the 98cent store today at lunch as I often do. And they got in a lot of cool new stuff since I went last week. So I was distracted and happy as I wove my way into the grocery section. And I was craving something that was ridiculously flavored, and I stopped in front of the Jones soda section in the fridge. Did i want blue, i always get blue, i should try something else. how about green, hmm maybe. Vanilla cola, could be good, but I\'m in the mood for something fruity. chocolate fudge, i remember trying a chocolate soda like 15 years ago and it was the worst thing ever. then i see it. Watermelon, that could be great. So i picked it up, avoiding the one with the kid making a funny face, opting instead for a skateboarder. Wow, this was gonna be great! A new flavor of jones soda!  In chosing Jones soda I passed by the large display of melon soda in a super cool metal bottle, imported by Sangaria, Torrance. product of japan. I pick up a bottle of that on my way to the checkout as an afterthough.  And i pay, and get to my car, very excited about trying this new watermelon flavor, i crack open the twisty cap, take a sip, taste the watermelony goodness going down my throat, and pause. \"Hmmm, something is a little strange\" i think. So i look at the front label. No, it doesn\'t say diet. So i look at the nutrition label. 90 calories, that\'s suspiciously low, but it does have 89 more calories than diet soda. then i start on the ingredient list. carbonated water, corn syrup, color... splenda. the last ingredient is splenda. I bought Jones\'s version of C2. UGGH. I was so disgusted with myself I had to stop drinking, even though it wasn\'t horrible, just a little off...  Splenda,  is slightly less  disgusting than the other fake sweeteners, but not less disgusting enough. and really, why do they think they can just stick it in stuff and not have people notice. at least with nutra sweet and the like the packages are CLEARLY labeled. grrrr.  So, I\'m learning my lesson. I will try to stop making horrible impulse purchases at the 98 cent store. I know that horrible bottle was meant for me because the fortune said \"you will find much fufilment with someone special very soon\" surely, that was for me.  Meanwhile, this melon soda is fantastic. I may have to go back and buy a case. This will not be a horrible impulse purchase if i still want it tomorrow.   '),(16886,3570813,'03,August,2004','2004-08-03 00:00:00','    urlLink http://photos.yahoo.com/bonafantegardens/   Not my pictures, but pictures of me nonetheless. The weekend of July 24-26, 2004.  Bonfante Gardens and the Gilroy Garlic Festival   '),(16887,3570813,'02,August,2004','2004-08-02 00:00:00','    urlLink I wanna live forever   '),(16888,3654666,'28,June,2004','2004-06-28 00:00:00','   Everyday is the same.  I wake up assuming that I\'m going to have a good day.  It never fails though, by 10AM, my stomach is in knots, my heart rate is throttling about 30% above average and I\'m lighting another cigarette.  I\'ve thought about this for a long time and I really don\'t think I\'m over-reacting or delusional.  I seriously think that, at minimum, five people out there exist for the sole-purpose of pissing me off.  What a conspiracy theory that is.  The other question at hand, are the motives of these people deliberate or sub-conscious?  Either way, I\'m starting to wear very thin.  I need to find some way to channel all of this negative energy into something good. Is that possible?  I really don\'t know.  What I do know is the countless nights I go home and just want to be left alone.  Now, this wouldn\'t be such a problem if I were the only one affected, but there are other people in my life that suffer from my inability to be content.  Oh, by the way, all of you self-help gurus out there, who have suggested breathing exercises to me, screw off and get a life.  Those remedies only work for the poor f*ckers who can\'t breath and hyperventilate when they get upset.  Yes, I\'m thankful that I\'m not prone to wheezing.  However, I don\'t want to live my whole life this way.  Will I ever be completely content?  Am I really asking for too much?  I don\'t think so.  '),(16889,3654666,'18,June,2004','2004-06-18 00:00:00','   Am I the only person who feels like I\'m surrounded by nothing but stupid people?  Have you ever stopped and counted the number of stupid people and then the number of intelligent people you know?  It\'s completely insane!  Now, I know that I\'m not the smartest guy in the world, but I\'m constantly being tortured by the misdeeds of the special people, as I like to call them.  They\'re everywhere you turn.  It does not matter where you are, your bound to be subjected to their unintelligible ways.   And it just so happens; the greatest concentration of these people is your place of employment.  Now, I\'m not asking for everyone to have an IQ of 145, but can Wal-Mart or Target please start selling common sense.  I don\'t know, I always thought that intelligence and common sense came hand in hand.  Maybe I\'m wrong, but is it really too much to ask for everyone to at least have one of these traits?  I\'m tired of cleaning up after the mistakes and misadventures of these cracker jacks.  What ever happen to the thought of working hard and doing your best?  Of course, maybe they are doing their best.  If that\'s the case, I pity the children of these buffoons.  Another thing, have you ever noticed that the moron or dumb ass that you went to high school with has not changed?  Is there not a rule somewhere, describing survival of the fittest?  If there were such a thing, wouldn\'t you think that all of these half-wits would be squashed out?  Or maybe we\'re spending all of our time and energy catering to these people.  What a concept that is.  '),(16890,3654666,'17,June,2004','2004-06-17 00:00:00','   What a divine pleasure it is to be escorted by a child through the grocery store, well any store for that matter.  Those of you without children have no idea the pain and frustration we go through just to pickup a few items to prepare dinner.  In fact, I envy those of you who are completely oblivious.  Before I had children, I was also unaware that any such phenomenon existed.  For those of you unaware, I have provided before you an enlightening experience.  The excitement begins right after you proceed through the doors.  Instantly, you have two bogies streaking off into different directions.  You catch back up with them and have this crazy idea that you will be able to go about your business.  This is not the case.  You get the two year old into the cart, no problem. However the four year old is a completely different story.  The imagination and level of excuses that he bestows upon you are unprecedented.  He either wants to ride on the end of the cart or walk, promising to behave and not to unleash his wrath upon you.  You have heard this before and make an attempt to levitate him into the cart, big mistake.  You have now awoken what I like to call Satan\'s minions, because nothing else could create such a horrific sound.  Oh, by the way, this is the part where everyone in the store has now paused to see who is beating there child.  Be smart, give in, and let him walk and be on your way.  First stop, the produce department, you couldn\'t forget bananas for those little monkeys, could you?  Your thinking now this isn\'t so bad, right?  Just what you thought, the infamous cereal isle is right around the corner.  This is the second worst area in the store, we\'ll get to the first later.  A kids dream, more chocolate, marshmallows, and sprinkles than you can shake a stick at.  Did I mention that all of those items are placed on the bottom shelf?  Geez, I wonder why.  Now your being bombarded by two high pitched screaming voices who feel that their world is going to end if you don\'t get fruity pebbles, cocoa puffs, and crunch berries.  Being the diplomatic guy you are, you attempt a compromise by allowing one box of the sugar filled toxins.  This idea does not go over well.  What do you know?  You now have one child imitating a screaming fire truck and one trying to show MJ a few new dance moves while spinning on the floor.  Once again, you are looked upon as the evil parent abusing his kids.  At this point you throw all three boxes in the cart along with the squirrel, which has been rolling around on the floor and make a mad dash to the check out.  Just when you thought this nightmare was almost over, you\'ve arrived at what you perceive as the light at the end of the tunnel, afraid not.  Not only are you subjected to evil marketing schemes throughout the store your are now faced with and endless array or candy bars, suckers, gum, you name it.  It not over yet.  You now have to children hurling themselves fearlessly over the sides of the cart to get their hands on anything possible.  You try as you might to contain the little monsters while giving the cashier the look.  To those of you who are not parents let me describe the look. Imagine a face filled with helplessness, anger, and anxiety all rolled into one.  You struggle to reach your wallet without releasing your new found prisoners, and hand over the cash.  You are now on your way out the door with two wiggle worms in hand, 3 boxes of cereal, and a bunch of bananas.  In to the car they go, when you know realize that you\'ll be making an emergency stop at McDonald\'s, because you somehow managed to leave the store without dinner.  '),(16891,3654666,'16,June,2004','2004-06-16 00:00:00','   Does anyone really know where they are going?  I\'d really like to think that someone does.  However, I go through this state quite frequently where I feel lost and destined for a higher purpose.  Right now, there are only a few things that I\'m sure of.  I know that I am a father of two beautiful boys that have been an exhilarating force keeping me headed in the right direction.  I don\'t know where I would be without them.  But once again, I feel like I am at a crossroads in my life.  I have been separated from my wife for a year now and the divorce has been initiated.  This last year has been such a weird experience for me.  Iâ€™ve always had this perception that I was a completely independent person and now Iâ€™m not so sure.  I really have to say that up until about two months ago I was unable to embrace this new life of freedom.  I was unable to escape this endless feeling of loneliness and boredom.  Iâ€™m finally to a point where I feel like my true self again.  Itâ€™s funny how one person can mold you into something you donâ€™t even know you are.  Actually its scary, the whole thought of our lives and existence being so strongly influenced by people who come and go.  Itâ€™s not really a new concept.  I just never thought about it.  So here I am now, feeling free, focused, and more powerful than ever, but what now?  '),(16892,3654666,'11,August,2004','2004-08-11 00:00:00','   If I could pick one movie to portray my workplace it would be Office Space. If you havenâ€™t watched it, I suggest you drop everything and go rent it. Iâ€™ve worked in a lot of different environments, but Iâ€™ve never experienced so much bureaucracy and confusion in one place. Most chaos can be attributed to our ownerâ€™s habitual nature of changing job duties and authority on an almost weekly basis. At this point in time I am the poor fool in charge. This will last until someone else has some bight idea on making the company profitable or until our fearless leader gets pissed off at me again. Either way, Iâ€™ll accept this mission and be happy that Iâ€™m not currently on his shit list.  Our office is made up of a variety of annoying personalities, one of which, a man I like to Mr. Limp Dick. He possesses the most annoying personality of all. He wanders around the office aimlessly, putting his nose into everyoneâ€™s business, and tries to convince us all that he is the office manager. In reality, he is nothing more than a brown noser and tattletale. He possesses absolutely no leadership skills what so ever and is unable to handle any forms of pressure being bestowed upon him. Iâ€™m not complaining about his inability to lead, because this company would be the next titanic if he were in charge. It just kills me that people like himself, one of the special people, have brainwashed themselves into thinking that they know shit, when they donâ€™t know jack. Iâ€™m thinking that the majority of his malformed personality stems from the fact that he inhabited his parentâ€™s house until the age of 40.  I also have the pleasure of working with Amnesia Man, another delusional personality, who makes my day almost unbearable sometimes. He was sane at one time, yet here we have another man whose mind was fucked up by a woman or should I say by her rejection of him and his mommaâ€™s boy tendencies. I would really feel sorry for this pathetic soul if he didnâ€™t piss me off all the time. I have never known anyone who could fuck everything up beyond recognition in the half-hour between him and I coming into work. I depart in the evening and everything is chronic bliss and I arrive in the morning to nuclear meltdown and upon questioning his actions, he remembers nothing of his foolhardy demise. I just really hate it when I am trying to take a proactive approach and trying to make things flow more efficiently, and I have to step back and react to an unneeded disaster and waste of time. It wouldnâ€™t bother or surprise me so much if he wasnâ€™t at one time a great guy to work with, but heâ€™s inevitably lost his mind. Picture a grown man, in just a diaper, curled up on the floor, sucking his thumb and you would understand my frustrations.  Six months ago, I would have complained about working with women. However, we are now only graced by one female presence. You can all say what you want, but when you have two or more women working together in a small office, there is bound to be trouble. Donâ€™t get me wrong, when they were friends all was well, but women are always conspiring against each other. I would get one day of peace and then act as a mediator the next. I finally got fed up with these frequent occurrences and let the claws fly. Needless to say we no longer have the afternoon catfights.  Anyway, though this place drives me crazy on a daily basis and Iâ€™m constantly scanning the web for new employment, I do realize that no workplace exhibits an aura of perfect harmony. I have accepted this hell for the time being only, because of the experience and knowledge that I might gain and who knows one day I might be the all-knowing one.  '),(16893,3654666,'06,August,2004','2004-08-06 00:00:00','   I canâ€™t wait to pick up my boys tonight. Itâ€™s only been three days but I feel empty when there not around. I canâ€™t wait to see Chrisâ€™s constant smile and Mikeâ€™s scheming face. Iâ€™ll arrive at their momâ€™s apartment. Mike will run out the door with cars and trucks in hand yelling some line from Fast &amp; the Furious. Chris does something extremely different yet interesting. Heâ€™ll run towards me doing a combination of running, jumping, and skipping. I really donâ€™t know what you would call that, but I know heâ€™s happy to see me. Their mother weâ€™ll be trying to make small talk, but Iâ€™m completely oblivious to anything she has to say. Iâ€™ll walk them down the hall and out the door to my old Chevy beater. Theyâ€™ll fight over who get to sit next to Dad in the truck, but before long weâ€™re on our way.  Itâ€™s crazy how much excitement they get from my driving. We wonâ€™t even round the first corner and mikeâ€™s screaming for me to squeal the tires. Drive faster and rev the motor up are among other requests that I receive. The funniest part however, is the fact that they think that old truck is the fastest vehicle around, believe me its not. Iâ€™d be lying if I said that I didnâ€™t enjoy it. It kind of makes me feel like God, toying with their reactions.  Iâ€™m not sure what weâ€™ll do tonight. Iâ€™m excited to see them, but tired as hell. I really hate having them one day at a time. I always want them to have the best time ever when there with me. I guess I sort of feel like if they donâ€™t leave happy that unpleasurable thoughts swirl in their minds about me. It wouldnâ€™t be so bad, but when I pick them up on a weeknight and theyâ€™re tired, cranky, or just plain evil, itâ€™s hard to have a good night and then theyâ€™re gone in the morning. Iâ€™m trying to get something arranged with the EX as far as having them for a few consecutive days during the week, but obviously have seen no success thus far.  I know that I shouldnâ€™t worry so much about it, but Iâ€™ve always wanted to be the best dad possible. I want them to know that Iâ€™m in their corner and that Iâ€™ll always be here. The divorce has really got to be confusing for them. I donâ€™t want them to feel like I did and think their dad didnâ€™t love them. It was a real eye-opener to me when Mike was born. At that point, I realized just how much of a piece of shit that my father was to me. I never will understand men that donâ€™t take and donâ€™t want any part in their childâ€™s life. Those kids are a piece of you that will be here long after your gone.  '),(16894,3654666,'03,August,2004','2004-08-03 00:00:00','   First of all, sorry to those of you who have read this blog on a regular basis and have seen absolutely no activity, but I have been busy as hell for the last month. God that was a long sentence! I just moved back in with my parents. Out of all of the things, Iâ€™ve had to do this year; it was definitely the hardest. I felt like a looser at first, knowing that I havenâ€™t had to run back to mom and dad, since I moved out, 6 years ago. I just couldnâ€™t make it anymore, financially speaking. Itâ€™s funny how you go through life, honestly believing that you can make it through any obstacle, untouched and unscaved. It just isnâ€™t so. Through the course of the last year, I have lost my house, car, financial stability, and pride. I fought for every one of those things and inevitably lost. What can I say? I have been struck by the joys of divorce, among other things. Oh well, such is life. I havenâ€™t given up though and I donâ€™t plan on it. So, to all you motherfuckers out there who are trying to break me, itâ€™s not going to happen. Iâ€™ve come to the understanding that anything that holds me back or wears me down only makes me stronger. Iâ€™ve also acquired a new outlook on life. Itâ€™s time to focus on the important things. Iâ€™m going to slow down, absorb as much as I can, and try to live everyday as if itâ€™s my last. My first two priorities in life are Chris and Mike; everything else is just detail. All I can hope for is to influence their lives like my dad influenced mine. Its kind of funny, how I realize now, how much my parents are responsible for who I am today and what I have accomplished. I wouldnâ€™t be anything without them, thanks! Thereâ€™s also another moving force in my life thatâ€™s helped me keep my head above water, she knows who she is, thanks to you also. Its hard to say how the last few months would have fell into place without your reassurance and support. Another piece of enlightenment that I have acquired, is the fact that life means nothing without your friends and family. So here I am ready for anything, refreshed, and never more confident.  '),(16895,3948533,'16,August,2004','2004-08-16 00:00:00','   I read that a year ago was the big blackout of 2003.  I remember it well.  I was totally oblivious to it, as I had daycare children so the television news wasn\'t on and I had no idea  1/4 of the country was in the dark.  I\'m in one of the areas that was dark...exactly  11 miles down the road was the cutoff.  While people 11 miles from me could not turn on lights, I was sitting in air conditioning watching movies and eating microwave popcorn.  Life\'s weird, isn\'t it?  And I know I should have felt bad, but I didn\'t.  I turned that AC to high and sat back and relaxed.  What would I have done if we lost our power for that long?  I would have lost a freezer full of food...deep freezer as well as the one in the fridge.  I would have gone kind of nuts not being able to shower.  Laundry would have piled up.  As long as I could flush the toilet and wash I\'d be ok, but having no hot water would suck.  We are so dependent of electricity, right now I have on fans, television, the fridge and freezer of course, the computer, lights...I know that I am a huge energy hog.  But what should I do, turn off one 40 watt lamp or unplug the stereo...would that help?  I\'m just thankful that I wasn\'t one of the \"blacked out\" people, that we kept our power, and hope to continue to do so.  I like candles for their smell...not as a light source.  '),(16896,3948533,'14,August,2004','2004-08-14 00:00:00','   I found something wonderful at the grocery store the other night....footlong hot dogs and buns!  All summer I had seen the hot dogs, but not the buns.  I asked at the store about the buns and they said the didn\'t have them.  I said, \"But you carry the foot long hot dogs!\"  Then last week they had several packages of footlong buns hanging by the hot dog case....but no foot long hot dogs in sight.  It makes you wonder, doesn\'t it?  Who the hell is the Einstein ordering this stuff?  So, anyhow, I now have a package of 6 footlong Koegel\'s franks and a package of 6 foot long buns and I\'m going to grill them today!  There isn\'t too many things more fun than a foot long hot dog fresh off a charcoal grill.  All you need is a squirt of ketchup (you may need mustard, however, I do not) and you have a whole meal.  If you have some potato chips, it\'s a bonus day!  Foot long hot dogs were a favorite of my kids.  We would always get them when we went camping, cooking them on a little grill over the fire.  They fall off sticks too easily, but on a grill right over the fire they have that same taste.  So, today I shall have my foot long hot dog.  And I shall be full!  '),(16897,3948533,'11,August,2004','2004-08-11 00:00:00','   Using Fritos scoops to eat your chocolate almond ice cream sounds like a good idea You go shopping and everything in your cart is chocolate flavored and/or drenched in salt You would actually murder someone over M and Ms You want to have a fondue party, make a chocolate fondue, and dip Ruffles into it You can justify eating potato sticks in a can as a \"vegetable\" You can justify eating double butter, salted popcorn as a \"vegetable\" You can justify eating candy corn and jelly beans as a \"vegetable\" The portions in a \"Hungry Man\" dinner make you giggle.  (too small) You can almost fall asleep walking around And you don\'t care....  '),(16898,3948533,'08,August,2004','2004-08-08 00:00:00','   Lately most of my online friends have abandoned the regular AOL chatrooms for private chats.  Got to keep the riff raff out, you know.  Anyhow, I\'m a fan of these rooms because quite frankly, outside of a couple I know well, I really don\'t care to be pals with teenagers or idiots.  Why is it that in every private chat, there has to be one dingleberry who insists on inviting in the very people you have left the regular chat over?  If I wanted to talk to these bozos, I\'d stay in the regular AOL chatroom.  I want to avoid their stupidity, small mindedness, attention whoring, whatever...and BLAMMO there it is, in a room that was supposed to be \"secret\".  Now they know the link and we can never use that room for privacy again.  Nothing said in these private rooms is X rated, there is no cybering or anything, just talking to friends without the drama and idiocy of these teenagers and others with the IQ of dryer lint.  I don\'t hang around with people like that offline, why should I be subjected to their drooling rants online?  I got an AOL account to surf the web and chat with like minded folks....not Special Olympians.  And I am not dissing people with real mental defects.  I have a few online friends who are \"slow\" or even borderline mentally handicapped, but they don\'t annoy me.  The ones that annoy me are the \"HEY WASSUP?\" \"Any fine hoes in diz hizzhouse?\"  \"What size bra do you wear?\" morons.  I am also so sick of the teenagers who think AOL is their domain.  Sorry, it\'s actually an adult forum.  Paid for by adults.  I get so pissed when some 15 year old types, \"You are too old to be in here.\"  Oh really?  You have your own credit card to pay your own AOL?  I know I do.  I fear for this generation of idiots.  Soon we\'ll all be knee deep in drool.  '),(16899,3948533,'07,August,2004','2004-08-07 00:00:00','   Anyone who has had to share a computer with someone can share my frustrations.  I love my boyfriend, but sharing a computer with someone, even someone that you love, is definitely an exercise in....oh, let\'s say an exercise in not smothing someone to death in their sleep with a pillow.  Point in case, I work at home, so I do try to get off the computer in the evening/night/weekend so that he can enjoy surfing, chatting, etc.  Instead he will spend most of his online time downloading something...music, a game, whatever.  Then we will be infected with five dozen different viruses and he\'ll deny downloading anything that might have caused it.  Unless the virus came from a government website, the local newspaper website, or an email from Betty Crocker, Nestle, or General Mills, I don\'t think it came from anything I\'ve done.  Then later that night I may be fortunate enough to catch some friends online.  Someone might even start a little private chatroom that we all go to.  Now, if HE was online, and they invited him in, he wouldn\'t go.  But if I\'m on and in there, he wants to read over my shoulder, comment, and if I go get drink or, god forbid, go to the bathroom he\'ll actually get on and start typing in the chatroom.  He did this last night, so I finally just signed off and let him on.  I don\'t know if he went back to that chatroom, I just covered up on the couch and went to sleep.  Now, I know he is going to read this get he\'ll probably get that all pouty voice and say \"I didn\'t mean to (insert whatever).\"  And I know that.  I know that he doesn\'t sit in a chair next to me, groping me, pulling my hair, and making all types of sexual references and gestures on purpose.  He isn\'t trying to get my attention away from the screen when he pulls on my clothes, tickles me, or tries to pull my pants down.  (\"I\'m just kidding with you.\")  I just wonder how he\'d view my \"kidding\" if I anally assaulted him with a cucumber while he was downloading yet ANOTHER Psychopathic rap artist or ANOTHER video game ROM, or whatever.  Which is why I say, bludgeoning someone to death with a blunt instrument is, indeed, an act of love.  And honey, I\'m just kidding with ya.  I love you.  '),(16900,3948533,'06,August,2004','2004-08-06 00:00:00','   Soft Drink:  diet Coke alcoholic drink:  strawberry daiquiri color:  orange and black TV Shows:  CSI, CSI Miami, Stargate SG1, The Grim Adventures of Billy and Mandy, Fairly Oddparents, Spongebob Squarepants, Sealab 2021, Strangers With Candy, Farscape, Lexx, Kolchak: The Night Stalker, Monsters, Night Gallery, Twilight Zone, Kindred The Embraced, Dark Shadows, Forever Knight, Toonsylvania, Pinky and the Brain, Animaniacs, Wild Wild West, Star Trek TOS, Petticoat Junction (s0 sue me), Bewitched, Kids in the Hall. sandwich:  Chicken salad cookie:  Nabisco Chips Ahoy fruit:  watermelon side dish:  Rice meat:  fried chicken snacks:  Ruffle\'s Potato Chips with French onion dip, Bugles, Chex Mix, mixed nuts  '),(16901,3948533,'05,August,2004','2004-08-05 00:00:00','   As a daycare provider, I get to see a side of life many people don\'t get to see or choose to overlook.  The seedy underbelly of \"the moronic state of children today.\"  And before you think \"Oh, all OLD people feel that way about kids.\" let me just say NO.  Let me cite some examples.  Last fall/winter I had five kids in my daycare home.  Their ages were 9 8, 7, and two 5 year olds.  At first I was really excited, I foresaw aftersch0ol games of Boggle and Scrabble, Monopoly tournaments, reading, etc.  Boy was I in for a reality check.  None of them knew how to play one game.  Not one.  They didn\'t even grasp the game \"Concentration.\"  You know, where you put cards face down, pick two randomly, and try to maych them.  I even had Pokemon Memory, same thing with Pokemon characters on cards, and not even the 9 year old could \"get it.\"  What the FUCK is there to GET?  You turn over two cards, remember what is on them, and try to find two the same?  The front of the box says \"Ages 4 through 8.\"  I guess I got spoiled because my own two children, now grown, were actually able to \"get\" games.  By the time they were 5, they were playing Skip Bo and Uno and Monopoly.  And not \"babified\" versions either.  My mom taught them cards, and she was not a merciful teacher.  You either learned it right or didn\'t play, period.  So I guess I just don\'t \"get\" how a kid who is going into middle school in a year and a half can\'t figure out how to play Pokemon Memory or Mickey Mouse Yahtzee.  They begged me to play Mouse Trap, remember that game?  You set up this elaborate mouse trap and try to catch you opponent\'s mice under the big net at the end?  I told them I wasn\'t going to sit there and do it all for them.  They all reassured me that they had played Mouse Trap on many occasions and knew how to do it.  Liars.  All the board games eventually went up on a really high shelf and I brought out the video games.  They really didn\'t \"get\" a lot of them either, but at least the bright lights and sounds shut them up for a while.  '),(16902,3948533,'03,August,2004','2004-08-03 00:00:00','   I have posted and deleted at least half a dozen entries so far in this blog.  This one is staying, I guess, no matter how lame or whiny or self-indulgent that seems.  Isn\'t a blog sort of self-indulgent anyway?  Why would I think anything in my life could possibly be of any interest to any other person in the entire world?  So...here I go.  Who I am, what I want, what I do...this is as hard as when my boss at work asked us to all write our out goals and how we are going to attain them.  My goal at that time was to get the Hell out of that place.  I just wrote down that I wanted to be debt free.  I didn\'t want a confrontation.  Just roll over, take it up the poopoo shoot without Vaseline, and smile.  Be nice.  Try to help others.  Seems the more I do that stuff, the more Vaseline I need.  I just want someone to be kind to me.  To talk kindly, to put me first, to care what I think or feel or want.  Not ask me, and do whatever the hell they want anyway, but to actually take into account what I say, think, or feel and make it matter.  Most of the time I feel like the only thing that matters about me to anyone is what I can do for them.  Babysit for free or cheap.  \"Loan\" them money I know I\'ll never see again.  Pay their bills.  Be their cheerleader.  I\'d like a cheerleader of my own, someone to really share things.  Someone to share the bad stuff, not just the good.  Someone who doubles the joy and halves the sorrow...gag, yeah.  You know, what you want and what you get aren\'t the same thing most of the time.  I guess, all in all, I just want to matter.  I want my existance to matter.  There was a man here in town that most everyone in town knew of.  He died.  Nobody went to his funeral.  Not one person.  Not ONE PERSON!!  I don\'t want that to be me.  '),(16903,3737568,'24,June,2004','2004-06-24 00:00:00','   urlLink    Me a couple of months ago. &nbsp; urlLink '),(16904,3737568,'24,June,2004','2004-06-24 00:00:00','   The day started off with me sleeping in (as I do when youâ€™re on holidays, partly more of that later) into Canberraâ€™s icy winter morning starts at about -10 degrees Celsius. After telling my self I should get up after two hours of lying in bed think about how cold, windy and generally dodgy the weather was outside I managed to coax myself out of bed and get ready for my fun eventful \"miniday\" at college. What is a miniday you ask? well its where you go in to your classes for about 2 minutes get your marks back see how bad you went, think â€œwell I could of done better in that but I didn\'t, oh wellâ€.... then leave. Pretty stupid invention if you ask me. Youâ€™re telling students to voluntarily turn up to school, do nothing but see how bad they went then go home. All with the weather being like its in sub-Antarcticâ€¦. in winter!    I think itâ€™s a sham. Foolish teachers trying, and succeeding, to irritate students who are basically on holidays other than these meagre couple of hours they have to wait at school for when they could be out (trying) to have fun somewhere. Anyway where was I. after that I went home trying to keep my hat from getting blown away by the furious winds that were flying by swirling man, animal and woman into a vicious windstorm of destructionâ€¦...well maybe not that insane, but it was windy ok. So I get home and sit down for about five minutes read my newly purchased monthly issue of Atomic when the phone would ring and who would it be, who else rings you just when you think its safe to have fun again?.   Work, just when you think its safe to have fun again. They were kind enough to ring me up about three hours before I was supposed to start there saying Iâ€™m not needed. Well thatâ€™s crap since Iâ€™m poor and I need money to fund the ludicrously expensive computer Iâ€™m trying to save up for but cant due to that fact that I have to live. Splurge. So I go and tell my mate, who is over as usual sinceâ€¦wellâ€¦I donâ€™t know I think he was here when we moved in actually and my sister who thinks my dog is a real person and that you should talk to him and act as though he is a person in all respects of the word (by the way my sister is 19 turning 20 this yearâ€¦yeah scary isnâ€™t it) that work rang me and said that they donâ€™t need me to work. Well. They both said that â€œyou should of got them to pay you for your shift since they didnâ€™t give u more than 24 hours notice etc.â€ Iâ€™m like shit guys you need to tell me this before the situation happens so I can get some free money and to that they responded in their usual dull sultry manner â€œwhateverâ€ so I sat down and tried to figure out what I was going to do for the rest of the nightâ€¦â€¦I still am and its not the evening anymore, itâ€™s the morning. So essentially I did nothing to of worthy note for humankinds existence on this arctic day on the 24th of July 2004.   Fare thee well. '),(16905,4176051,'20,June,2004','2004-06-20 00:00:00','   My teenaged daughter left today to spend the summer with relatives. Don\'t get me wrong, I love her dearly, but the last few months have been *challenging*, to say the least. I think we both need a break for a while, so I\'m thankful that she has the opportunity to spend the summer away - and I\'m sure she is too.  I hope she\'ll have a lot of fun.  That said, I\'ve got more than enough to keep me busy while she\'s away. It\'s time to start whittling away at my ever-growing stash. After all, I need to make room for some new yarn, right? Today I start on my summer poncho, another origional from my favorite LYS. Well, maybe it\'s not quite an \"original\". I was in the shop chatting with some ladies there when a woman came in wearing a gorgeous mauve-colored poncho (a la Sex in the City). She wore it with blue jeans, a white tee and some very sexy-looking high-heeled slides.  She looked fabulous!  She had purchased it from a trendy shop in the city and was wondering if someone could help her figure out how to make another one for herself.  Of course, we all jumped at the chance to take a closer look, after asking her to take it off! :)  After trying 3 different yarns, 3 different stitches, and at least 3 different needle sizes, I\'ve finally got the look I\'d hoped for and plan to start knitting as soon as I finish blogging.  I\'m using Maggi\'s Linen in a beautiful deep mustard color. It should be a quick knit. Since I intend to do nothing else today except knit, I should get quite a bit done.  '),(16906,4176051,'14,June,2004','2004-06-14 00:00:00','   Here it is, the one I\'ve been waiting to wear... My Rumba sweater. The pattern is a Lettuce Knit original (my favorite LYS - local yarn shop). The yarn is called Rumba. This one was truly a labor of love. It\'s knit in all one piece and although the yarn is beautiful, it\'s not the easiest to work with. And don\'t drop a stitch because it\'s impossible to pick up! I was about 4\" from the end when I noticed that a join had come undone. I had to unravel the entire sweater and start again. That\'s the bad news... the good news is that I love this one so much that I\'m going to make another one in black! This time I\'m going to be knitting it along with my buddy Nim, who\'s going to be knitting my sweater\'s twin, so we\'ll have each other along for moral support.     urlLink  '),(16907,4176051,'13,June,2004','2004-06-13 00:00:00','   A picture of me knitting...    urlLink  '),(16908,4176051,'07,June,2004','2004-06-07 00:00:00','   Here\'s sa picture of one of my prized possessions... handmade knitting needles from my husband, Jonathan. Most people would make needles from wooden dowels, but not my DH. He started with a plank of rosewood, which he cut into 2 \"strips\". Hours of sanding (by hand) later, those strips were transformed! Now I\'m on a mission to find the perfect project with which to christen my new needles.    urlLink  '),(16909,4176051,'07,June,2004','2004-06-07 00:00:00','   I\'ve tried journaling before, but never had much luck with it. Perhaps I was intimidated because I didn\'t believe I had anything profound to say. Or maybe I was afraid that someone would find my journal and read my most private thoughts. And many times, I just didn\'t *feel* like writing.  But this time it\'s different. I\'m on an incredible journey and although my insights may not be profound, they are significant (at least to me) and I want to record them in some way... and since I type much faster than I write, this seems like the perfect way to do that.  At 45, I\'ve finally found a passion in life, in a most unexpected format - knitting. It\'s become so much more than a pastime or a hobby. It\'s become a way of life. For me, knitting is relaxation, release from stress, meditation, prayer, functionality, creativity, philanthropy, gift-giving, and just plain fun! Through a variety of colors, textures and patterns, I create things that please my eyes, soothe my skin, and massage my ego - all at the same time.  It\'s about creating a vision and making it happen... it\'s about understanding that giving a gift can be as great a joy as receiving one... it\'s about practicing the principles of simble abundance... it\'s about realizing the awesom mystery of creating a piece of wearable art from a single piece of string.  I knit from my heart... I knit for my soul.  '),(16910,4176051,'30,July,2004','2004-07-30 00:00:00','   I thought that subject line would get your attention! ;)  Yes, you guessed it. I\'m talking about that dreaded \"S\" word... Swatches! I understand the need for swatches to measure the stitch gauge before starting on a new project. But I\'d always thought, and been led to believe by some pretty experienced knitters, that row gauge didn\'t matter. After all, you just keep going until the piece is as long as it needs to be.  Who care how many rows that is, right?  NOT!!!  I found a wonderful pattern with raglan sleeves that extend all the way to the neckline. Here\'s a picture.  Raglan V Sweater  urlLink     I did a swatch and the stitch gauge was fine. I did the front, back and all the way to the armhole shaping on both sleeves (knit at the same time), only to find that my sleeves were at the right length for my arms, but according to the pattern, I had many more rows to knit in order to get the correct number of increases.  Go figure!  (I wish I had).  Anyway, I\'ve shared my dilemma with some of the \"old school\" knitters at my LYS and I\'m going to take it with me to the next knit night and hope that they can help me figure out a way to save my project.  '),(16911,4176051,'07,July,2004','2004-07-07 00:00:00','   Today was a great day.  I finished the work I needed to do for the day at 10:00 and instead of going home and finding more work to do, I decided that this was going to be a mental health day. Of course, this meant going to my favority LYS to sit and knit.  I took a couple of projects, one of which just needed the dreaded finishing that I\'ve been putting of for weeks. Not only did I get the cable shell finished, but I learned a beautiful finishing technique in the process.  Here\'s a photo...   cable shell  urlLink      I made some progress on the cable/lace shell also. I can\'t seem to bring myself to spend a lot of time on the linen summer poncho. I know it\'s going to look great when I finish it, but I\'m just not \"feeling it\" right now. I\'m about a third of the way done, so I\'ve convinced myself that I\'ll have it done in time for cooler weather in the fall.  While at LYS, I met a new knitting buddy... It didn\'t take long for us to discover that we both have an adopted child, we are both expectant godmothers, and that we are both interested in starting knitting ministries at our respective churches. We took a short break for lunch, and then returned for an afternoon of knitting.  Now that\'s the life!  '),(16912,3776616,'30,June,2004','2004-06-30 00:00:00','    I really haven\'t the faintest idea on where to start this.  I don\'t even really know what I wnat to say, other than the fact that I have a lot of shit that I want to get out.  I\'m just...  I don\'t even really know what I want to say... I\'m sick of everything.  I\'m sick of having to lie for people, even when I know it\'s wrong.  I\'m sick of doing what\'s right and being reprimanded and shut out by my friends. And most of all, I\'m sick of telling the truth and hurting people.  I sick of watching people suffer, even when I know that what I\'m doing is right.  I\'m so tired of hurting and crying myself to sleep every night because I hurt someone, all the while not accepting that was I did was right in the long run, not wanting to accept that sometimes the happiest way isn\'t the best way. Life  ISN\'T  fair, there\'s no getting around that fact...   I\'m sick of fighting with my friends and hurting them, I\'m sick of backstabbing and secret telling, I\'m sick of fighting with my sister and my parents, I\'m sick of listening to my parents fight, or having to listen to them bitch about each other, and I\'m sick of having to hurt the person I love most in this fucked up world. Most of all, I\'m sick of lying to myself, sick of convincing myself that eventually everything will be ok, I just have to work out a little problem right now.  I\'m sick of not accepting the fact that before that problem is even gone another one is on the way, and tha my whole life will be like that, problem after problem piling up on each other until I can\'t take it anymore.  Of course, I have my \'Get Out of Jail Free\' card -- college -- but how much good will that do?  Sure, I\'m starting over, wiping the slate clean, and leaving behind this life, but what happens then?  I mean, I can\'t completely start over, because I\'ll always have links back to here -- a few of my true friends, and the fact that I can\'t completely abandon my family.  And what kind of disillusioned fantasy world would I be living in if I told myself that problems will never start up while I\'m at college??  It\'s like that stupid saying \"You can run, but you can\'t hide.\"  But it\'s true.  Sure, you can run from your problems for awhile, but eventually everything\'s going to catch back up to you...  And what can you do about it?    And now, with everything that\'s going on in my life... I don\'t know what to do.  I don\'t know who I can trust or confide on.  I don\'t even know if it\'s possible for me to trust anyone...  I can\'t talk to my parents...  They\'re always too busy, and they would never listen anyway.  God forbid they take a moment away from the computer or TV to listen to me...  I can\'t stand being around them anyway.  When they\'re both home they\'re usually complaining or fighting about something, and if I\'m ever alone with one of them they\'re usually bitching about the other one.  Every car ride or shopping trip is like a campaign to get me on one of their sides, and I can\'t take it anymore.  I want more than anything to be out of this house.  Away from the fighting, away from my sister, away from all the comments about my laziness and how I never do anything around the house.  I\'m sick of feeling ungratefull because I want to have a friend over instead of me going to other people\'s houses.  I\'m sick of everyone getting pissed off because I want to take the car to pick up a friend, of getting told no just because I wouldn\'t go pick up my sister from somewhere last time they asked...  and then there\'s my friends.  I\'m not sure who is siding with who, who will tell so-and-so, or who\'ll keep everything a secret.  I don\'t know who\'s going behind my back and repeating everything I say to someone else, or who secretly hates me but just puts up with me for their own benefit.  I just don\'t know anything anymore, and I\'m scared, and I don\'t know what to do...  Bethany... I didn\'t tell her  ANY  of the Ken stuff, and I don\'t really know why.  I feel like I\'m losing her somehow, like she\'s slipping away from me.  And that scares me more than you\'ll ever know, because I\'ve already lost too many friends since I started high school, and I can\'t take losing another one...  And then there\'s Ken... there\'s only so much I can talk to him about, because a lot of my current issues involve him...  Which is hard to tell Bethany, because she\'s never been in a relationship before so she doens\'t really understand everything.  Lainie\'s been pretty helpful, and Mande.  And of course Sam, because he shows me things from a guy\'s perspective.  I just... I don\'t know.  Every time I think I can confide in someone something happens to fuck it all up...  And now all I truly have is my journal and my blog, and I know from emperience that those things can fall into the wrong hands...  Written words provide tangible evidence, which can be used against you in more ways than you ever thought possible.  Everything just comes back to bite you in the ass...   I don\'t know... I don\'t have the energy to write anymore right now.  So I\'ll just call this \'Part I\' '),(16913,3776616,'29,June,2004','2004-06-29 00:00:00','    Yesterday was... In all honestly, I don\'t really know.  It started off quite boring, b/c I pretty much sat on the computer and did laundry.  I talked to Beth for a little while, but that\'s about it.  Then around 6:30 my dad dropped me off at Lainie\'s place of employment, so I hung out there w/ Lainie, Beth, Sarah, Jen, and sarah\'s lil\' sis was there for awhile too.  Sarah, Beth, and I thought it would be fun to stalk the band while they wre out marching, which was quite funny.  Sam and Alex gave us the finger, my sis gave me a dirty look, Hanlon laughed at us, Teagan, Kaylla, and Nick all gave us weird looks, and I believe the rest of the band was just scared b/c we were following them around and \'appearing out of nowhere\'  Then after we left Lainie\'s work we dropped Beth off then Sarah, Jen, and I drove around for awhile.  We went to Sam\'s to give him his Girlfriend Cookie, but he wasn\'t home, so then we went to Nick\'s.  We put a little note on it, went up to his door, rand the doorbell, then handed it in through the partially open door and ran.  As we wre pulling out of the driveway we could see Nick chasing us down the driveway w/ something in his hand, we weren\'t really sure what it was.  Then we went to give Alex his, but he was somewhere w/ Teagan, so we didn\'t.  Then we went to J-ville and wandered around Wal-Mart for a bit, then Sarah dropped me off at 10:00.  And then around 11 I texted Lainie, and we texted for over an hour... Then I went to bed b/c I felt like shit.  Ken never called me, so I guess he must\'ve had to work really late... He said he\'d probably just spend the night up at the fireworks store most of this week...  Blah... I might not get to see him again until the 5th, which kinda sucks, but at least the 5th will be a day for just me and him, no distractions.  YAY!!! '),(16914,3776616,'28,June,2004','2004-06-28 00:00:00','    Blog number 3...  Maybe I\'ll get lucky with this one and no one will find it that\'s not supposed to...  Third time\'s a charm, right?  On that note -- please don\'t link my blog to yours anymore, ok?  If it\'s not linked it\'ll be harder to find.  You can put it in your favorites or whatever, just not where everyone can see it.  And I won\'t be linking any blogs to mine.  But yeah...  I feel shitty, and I think that I might\'ve gotten sick from the heat on Sat.  Heat exhaustion or something... idk.  Anyway, I\'m gonna go back to playing on neopets (don\'t laugh).  Later. '),(16915,3776616,'31,July,2004','2004-07-31 00:00:00','    Mom\'s taking a nap, so I\'m allowed to get on the comp... w/e...  I still have to pack, and I have to take my grandma shopping at some point today, plus fill the car w/ gas again so my dad doesn\'t get mad.  I was supposed to wash it, but who gives a fuck... it\'ll just get dirty again anyway.  In all honesty I\'m too damn lazy to do much of anything... I want to call about my senior pics too, but I don\'t feel like it right now...  This sucks.  I hate being lazy. '),(16916,3776616,'30,July,2004','2004-07-30 00:00:00','    This may be my last post for a few days, I don\'t really know.  I doubt I\'ll have a chance to get on later today (currently 1am), and our flight leaves at 7 something Sunday morning so we need to leave for Milwaukee by 4, so I won\'t be getting on then...  My dad said I\'ll probably be able to use his comp while we\'re there, so I\'ll update every once in awhile.  Just to let you all know how much fun I\'m having, of course.  Trapped in our timeshare w/ my sister while my dad\'s at work...  Maybe I\'ll just walk around and look for hot or semi-hot guys to pass the time.  Make a friend or two to hang out w/ for the week...  Sit on the beach, getting some small amount of sun.  We\'re going to Universal Studios one day and Epcot another.  YAY!  And the best part of it all is -- NO MOM!!!  Of course, she\'ll probably have Don over while we\'re gone, but as long as I don\'t have to listen to them have sex I just don\'t care anymore... if my mom wants to fuck some random guy then fine.  Our family is already fucked up as it is, why not make everything worse, right?  I mean, there\'s no way in hell that they don\'t know about the drinking, yet they still let me go to Robert\'s all the time.  I don\'t really think they care all that much.  I don\'t think they know about the happy smoke, but maybe, idk.  In 4 months I\'ll be 18 and I\'ll do whatever the fuck I want and it won\'t matter.  Not that I don\'t already, but... They have nothing to threaten me with.  They aren\'t going to help me pay for college, so they can\'t take that away.  If they kick me out I\'ll just go live in a dumpster or something, visit friends every once in awhile.  I\'ll be fine.  Life sucks, oh well, you get over it.  A year from now I\'ll be in a completely different state, far enough away that I won\'t have to deal with my mom and Don and my dad, and Laina, and Kait, and everything else...    but yeah  this post started off happy then I got ranted and got off on a tangent so I\'ll stop... '),(16917,3776616,'30,July,2004','2004-07-30 00:00:00','    Idk where to start everything... yesterday was a LOOOOONG day... Went to Lainie\'s around 12:30, hung out there until 4, went home quick, went to Sarah\'s to pick up pics of Derek, picked up Beth, went over to Robert\'s, then we went to Winner\'s Deli to visit Jamie. By then it was like, 5, so we went to the fair and waited for Mande and Lainie to meet up w/ us. WE got piggy tattoos at the Mercy tent, and Robert and I went on the bungee thing. That was awesome. Then Mande and Lainie got there and we went on the Tilt-a-Whirl, then me, Robert, Beth, and Mande went on the Fireball, which was awesome. Then Robert left, and we went and got drinks and stuff, and walked around to all the tents collecting as many stickers as we could. That was cool. Then we went on some more rides and Robert came back and went on some more, then we played the coin game for a little while. We lost Bethy\'s glasses somewhere... I feel really bad, b/c it\'s kinda my fault b/c I\'m the one that adjusted the hat and knocked them off...  But anyway... we left around 10:30, and Me, Robert, and Beth went back to Robert\'s. This is where all the fun partying started..... Ordered pizza, and Richie, Andy, Jake, and Derek/Miller Time came over. Derek\'s cool, Hat Guy/Brian isn\'t. We played a game of Fuck You. Derek and Jake kept fucking me... I only had to beer bong once during the game, then I did another one later on (which, by the way, I did in like... 3 secs. Robert said stop and I didn\'t understand what he was talking about, b/c I didn\'t see how I could be finished already...). I played Duck Hunt, which was cool, b/c I haven\'t played that in awhile. I sucked at it though... oh well... We played tag in the back yard... Drunken tag is fun, b/c you can\'t run very well and everyone trips and falls all the time... Except Robert was chasing me and I fell and he tripped over me and landed right on my leg... That hurt like a bitch, then I couldn\'t walk, so Richie put ice on it, and they carried me inside. So then Richie and I layed on the couch and he put ice on it and we both fell asleep there. And... yeah, Lainie I want to talk to you again about the Richie stuff, b/c... idk, I still have to talk to Robert, but... Anyway, Beth and I left around 8 then fell asleep at her house. Around 11 something mom called me and started bitching me out. Apparently she didn\'t know I wasn\'t coming home last night. Wtf? I told her several times... Anyway, then I told her I didn\'t think I\'d be able to take grandma shopping today b/c of my leg, and she yelled at me b/c apparently I should\'ve come home when it happened so she could take me to the hospital. Never mind that it was 3 in the morning, the only thing that was open was the ER, and she probably wouldn\'t have taken me anyway... I failed to mention to her that I couldn\'t have driven home anyway b/c I was drunk, but you know... Then she gave me a list of things to do and told me to be home by 12...  Anyway... I got home, slept some, showered, did my chores, called grandma and told her I\'d take her to the store tomorrow, then I fell asleep again. Woke up at 9, made some Ramen, Ken texted me so I was texting him back and forth, then I sat in my room until mom left at 11 and now I\'m here... I don\'t really know where she\'s going, nor do I care, but at least I can get on the comp... w/e... I\'m gonna go play some games and read everyone\'s blogs... '),(16918,3776616,'29,July,2004','2004-07-29 00:00:00','    I hate my comp... oh well, it doesn\'t matter.&nbsp;   Last night was fun.&nbsp; I picked Beth up at 5:30, then we put air in my tires and went to see Lainie at work.&nbsp; Beth and I went to see Robert for a few minutes so we could talk to him about parties, then we put some stuff in his car and I got Burger King, then we went back to Lainie\'s.&nbsp; Jamie, Missy, and Jim stopped in, then Jamie came back later on and hung out w/ us until Lainie closed.&nbsp; She took us on a Happy Smoke walk, which surprised me, b/c I thought she despised me w/ every ounce of her being, and you don\'t share w/ people you hate.&nbsp; But she had a sweet ass one-hitter that looked like a cigarette.&nbsp; Beth got fucked up, b/c she\'s a lightweight.&nbsp; I just felt good, which was nice.&nbsp; And I got cotton mouth, but that\'s easy to deal with.&nbsp; Anyway, while Lainie was closing I talked to Race for awhile.&nbsp; He was listening to Jamie and Beth arguing about whether Alaska was part of the continental U.S.&nbsp; That was pretty funny.&nbsp; Anyway, then Jamie, Beth, and I sat there talking until 9:30ish, I dropped Jamie off, then we went back to Beth\'s.&nbsp; Sat in her living room until 10, then went over to Robert\'s.&nbsp; Decided to go to Perkin\'s, and that was fun (of course).&nbsp; When we got back Andy was at Robert\'s.&nbsp; I walked in the door first, and Andy thought I was Robert so he jumped out from behind the couch and screamed.&nbsp; OMFG he scared the shit out of me...&nbsp; Anyway, we sat and talked at Robert\'s for awhile.&nbsp; Richie called, and he was drunk, so Andy turned his phone on speaker so we could all hear it.&nbsp; Then around 11:45 Andy and Robert went to Kwik-Trip, and I walked Beth home then went home.&nbsp;   On another random note -- I haven\'t talked to Ken since Saturday, other than texts Sun and Mon.&nbsp; And surprisingly it hasn\'t bothered me at all.&nbsp; Like before, if I went 2 or 3 days w/o talking to him, I\'d be depressedish and miss him a lot.&nbsp; I still miss him, of course, and I want to talk to him and see him and whatnot, but I\'m just so so so happy that it\'s not bothering me anymore.&nbsp; Like, big step for me kinda thing.&nbsp; I really do think that this two weeks of not seeing each other is going to help A LOT.&nbsp;&nbsp;   But yeah, I\'m gonna go shower so I can leave soon.&nbsp; Later! '),(16919,3776616,'28,July,2004','2004-07-28 00:00:00','    I really wanted to go to the fair with Ken, and I wanted to go w/ Sam too, and neither of them are home this week, which really sucks.&nbsp; Part of me doesn\'t even really want to go, but I will anyway b/c it\'ll still be fun, and it\'ll give me something to do.&nbsp; Tomorrow, I suppose, with Lainie and Mande and Bethy.&nbsp; If anyone else shows up that\'s cool too.&nbsp; Meh. '),(16920,3776616,'27,July,2004','2004-07-27 00:00:00','    I\'ve been thinking about this a lot the past few nights...&nbsp; I mean, I know that I want to grow up more before I get back with Ken, but I\'ve never really thought about  what  growing up actually is.&nbsp; Is it something that I\'ll just know?&nbsp; Will I just be sitting somewhere one day and out of the blue realize that I\'ve \'grown up\' to the point that I want?&nbsp; HOW WILL I KNOW!?!?!?!?&nbsp; In a deperate attempt to grasp a small part of all this, I made myself a general list of some of the things I want to accomplish before I\'ll consider myself \'grown up\' in any sense.&nbsp; The&nbsp;first thing is responsibility --- I want to prove to myself that I can be responsible.&nbsp; I want to get a job and a car, those are both necessities.&nbsp;&nbsp;I want to keep up with my homework, because I slacked off on that a lot last year, and I want to&nbsp;make sure my grades stay up, because&nbsp;I &nbsp;KNOW  that I\'m capable, I\'m just lacking in motivation and self-determination.&nbsp; I consider myself a fairly reliable person, and I don\'t think I have much of a problem with reliability, but I want to work on keeping myself organized and rationing out my time so I don\'t push myself too much or let other people down.&nbsp; I don\'t know, maybe I\'m setting my goals too high here, but I suppose things can be adjusted later...&nbsp; And then there\'s the whole category of my future.&nbsp;&nbsp; I want to decide for sure what college I want to go to.&nbsp;&nbsp; I want to apply for college, financial aid, scholarships... all that good stuff.&nbsp; Most of the colleges I\'m looking at aren\'t all that cheap, and w/ mom and dad not paying for any of it, I\'m going to have to rely a lot on scholarships and financial aid.&nbsp; I don\'t want to be paying off student loans until I\'m 35 like my parents did...&nbsp; Plus I want to be financially independent.&nbsp;&nbsp; Once I graduate I\'m going to have to pay for everything by myself... cell phone, gas, car insurance, etc.&nbsp; And of course there\'s food.&nbsp; I mean, I\'m planning on living on Ramen for the most part, and that\'s less than .10 package,&nbsp; but still...&nbsp; I\'m just kinda worried about all that stuff.&nbsp; Graduating in general is just going to be big for me...&nbsp; In all honesty I\'m a little scared about what my life is going to be like a year from now.&nbsp; It\'s going to be weird...&nbsp; but yeah, on top of those two things ^ there\'s some other personal stuff that I want to work on.&nbsp; Like communication, because I suck at that, especially in relationships.&nbsp; I have a hard time talking about my feelings and expressing when something bothers me, and that\'s caused quite a few problems for me in the past few years.&nbsp; I also want to work on my self-esteem, b/c having a low opinion of myself is going to negatively affect me in a lot of aspects of my life. *sigh*&nbsp; &nbsp;  I don\'t really know, but this whole topic is putting me in a bad mood right now, and I just want to be happy, so I\'m going to end this. '),(16921,3776616,'27,July,2004','2004-07-27 00:00:00','    I forgot to post about this earlier -- when I saw Jason and Sami at Target I was telling Sami about the party where I got drunk and made out with hat guy.&nbsp; I was talking fairly loudly and said \"when&nbsp;I was drunk\" at least twice.&nbsp; About 5 mins later I realized that my mom was still about 10 feet away, and had been the whole time.&nbsp; Ooops.... oh well, she never said anything, and if she doesn\'t know about everything by now then she\'s just stupid.&nbsp; I\'ve slipped up way to much lately.&nbsp; Like the other day she told me I should drink some more of my blue uv, so we were at hte store and they didn\'t have lemonade, and I almost said \'that\'s ok, you can mix it w/ sprite and orange juice and it tastes fine\'&nbsp; Yay, go me, I almost opened my big mouth on that one...&nbsp; Oh well, I can\'t help being stupid.  So basically, there\'s no way in hell my parents don\'t have at least SOME idea of what I do.&nbsp; And as far as I\'m concerned that\'s fine as long as they don\'t keep me from going places. '),(16922,3776616,'26,July,2004','2004-07-26 00:00:00','    I hate being female ^ ...&nbsp; I mean, they aren\'t as bad as they used to be, so the birth control helped a little, but it\'s still annoying.&nbsp; I bought Target brand tampons last night... My mom&nbsp;won\'t go to the W store anymore b/c it\'s evil, so it\'s all Target and Shop-ko now.&nbsp;   Speaking of Target and shopping...&nbsp; My mom and I went on another shopping spree last night after we dropped my sister off.&nbsp; I don\'t know where this money is coming from... I  KNOW  that we don\'t have this money... My dad\'s gonna shit when he sees all the receipts... Oh well, it\'s school shopping and whatnot.&nbsp; Sorta...&nbsp; Anyway, the Dollar Store was having a 79 cent sale, so I got some cool socks and candles and shit.&nbsp; We spent $20 there... go us.&nbsp; Then we ate dinner at Panda Express.&nbsp; Muahahaha, I got my Orange Chicken!!!!&nbsp; After that we went to Target.&nbsp; Found some fun stuff at the $1 Spot (mini stapler/staple remover/tape set, magnets, mini highlighters)&nbsp; Saw Jason and Sami.&nbsp; Found out that they\'re moving into Gregg\'s old place, and Gregg\'s moving to Milwaukee.&nbsp; At least now I know where they live.&nbsp; Poor Sami, having to live w/ Jason.&nbsp; Muahahaha.&nbsp; Then I got some more clothes.&nbsp; A red shirt w/ Ralph on it picking his nose that says \"Pick Me\"&nbsp; ;)&nbsp; Plus, and here\'s where it gets good --&nbsp; yes folks it\'s true,   I   got&nbsp;a skirt.&nbsp; *wiats for the shock to pass* Yes, a skirt.&nbsp; Green.&nbsp; For Homecoming and other random events for which&nbsp;I feel like dressing up.&nbsp; B/c yes, I  do  do that every once in awhile.&nbsp; Infact, I\'m in the mood to just dress up and wander around right now...&nbsp;&nbsp;Too bad I\'m home alone and don\'t really feel like showering and calling&nbsp;to see who wants to do something.&nbsp; Maybe in a little while.&nbsp; After lunch or something.&nbsp; All because of yesterday\'s shopping excursion. Muahahahaha.&nbsp;&nbsp;Oh&nbsp;yeah, I got a hairbrush and chapstick too, but that\'s not very exciting, is it?  MUAHAHAHAHAHAHAHAHA!&nbsp; I\'m so hyper I really need to do something.&nbsp; Maybe I should go to Milton...&nbsp; but I need to go to the gas station first, b/c the pressure on my tires is really low...&nbsp; Lalalalalalalalalalalalalalalala *dances around*&nbsp; The internet gets really boring after awhile.. This isn\'t fair... '),(16923,3776616,'26,July,2004','2004-07-26 00:00:00','    Yesterday was... interesting.&nbsp; I went w/ my mom up to Fond du Lac to drop the Don guy off... Saw his two kids...&nbsp;On the way back home we stopped at Culver\'s and slept for awhile... then we went to East Town mall.&nbsp; My mom spent $50 on me at Hot Topic.&nbsp; Camo shoelaces, Happy Bunny keychain, Blink-182 t-shirt, Cold t-shirt, and another shirt about the little men behind my eyes :)&nbsp; Plus the shoes she bought me on Sat and the pants, etc. she\'s going to buy me tonight... It\'s early school shopping, but still, I\'m getting new stuff and Kait isn\'t.&nbsp; Only b/c my mom\'s mad at Kait b/c Kait\'s being a bitch, and she treats all her stuff like shit.&nbsp; Muahahaha!  Trav wants to have sex with Beth... That\'s just sick.&nbsp; I swear to god I\'ll kill myself if he somehow shows up at the party this weekend... '),(16924,3776616,'24,July,2004','2004-07-24 00:00:00','    Kait started asking me questions about Ken today...&nbsp; \"Is something wrong w/ you two, I haven\'t heard much about him lately?\" \"Did you guys break up?\" and whatnot... ARGH... I wanted to kill her.&nbsp; Then she starts telling me how she doesn\'t really like him.&nbsp; There\'s \'something about him\' that gives her a \'weird feeling\' or something like that...&nbsp; She thinks he\'s too obcessed, especially with the ring and all.&nbsp; She was kinda pissing me off with all the shit she was saying...&nbsp; I mean, I know she\'s only doing it because she doesn\'t want to see me get hurt, but still.&nbsp; Does anyone else feel that way?&nbsp; Do you all think Ken is...&nbsp; weird/creepy/whatever?&nbsp; I\'m not going to hate you if you tell me that you do, I\'d rather you all be honest with me.&nbsp; It won\'t change my opinion of him, b/c I know how I feel, I\'m just curious as to how my friends feel about him...  But anyway... he left for Pennsylvania tonight, and I didn\'t get a chance to see him.&nbsp; Part of me didn\'t want to, just b/c of all the stuff that\'s happened the last few times I\'ve been around him, but now that he\'s gone I want more than anything to see him, at least to say hi.&nbsp; And now I won\'t see him until I get back from Fla. on the 8th...&nbsp; 2 weeks from now...&nbsp; I\'ve never gone more than a few days w/o seeing him, and the last time I saw him was Thurs. morning before he left for work.&nbsp; Blah.&nbsp; I feel terrible though, b/c I have been horribly rude/bitter/cold to him since we broke up.&nbsp; I\'m using it as my defense mechanism so I don\'t cry when I can\'t see him, but it\'s not fair to him at all.&nbsp; My mind is still set on \"You\'re my boyfriend we need to see each other right now\" instead of the \"You\'re my friend hey when can we hang out next?\" that it should be on...&nbsp;&nbsp; I don\'t start crying if I can\'t see/talk to&nbsp;Bethy or Pookie every day, so why can\'t I get it in my head not to cry when I can\'t see Ken?&nbsp; This is so unfair... I can\'t handle this...&nbsp;   WHAT THE HELL IS WRONG WITH ME????  My mom keeps bringing Ken up all the time too... I can\'t stand to have them ask me questions all the time.&nbsp; \"Are you and Ken back together yet?\"&nbsp; \"How are you and Ken?\" \"So are you guys just friends now?\"&nbsp;&nbsp; Seriously, LEAVE ME THE FUCK ALONE!!!!&nbsp; They can just mind their own business and let me tell them what I want to tell them, which is pretty much nothing.&nbsp; I talk to my friends about stuff, and that\'s pretty much it.&nbsp; Mostly Pookie, when it comes to Ken stuff.&nbsp; She\'s the only one that I showed the letters too, and they only one I really feel comfortable talking to about relationship stuff.&nbsp; Bethy just has no experience w/ relationships so it\'s hard for her to relate and understand stuff, Mande compares stuff to her and Fred all the time, Crazy Katie is too... light-hearted? about everything, and Sam... I just haven\'t talked to much lately.&nbsp; Idk...&nbsp; It\'s just easier to talk to one person anyway, b/c then I don\'t have to tell the stories over and over again, which drives me absolutely insane...&nbsp;   AAAAAAAAAAAAAAAAAHHHHHHHHHHHHHHHHHHHHHHH!!!!!!! '),(16925,3776616,'23,July,2004','2004-07-23 00:00:00','    * My grandma called me a fucking bitch after dinner tonight, but only b/c everyone dared her to.&nbsp; It was pretty damn funny, actually.  My mom bought me shoes... black sneakers and blue flip flops.&nbsp; YAY!&nbsp; She\'s buying me more clothes tomorrow... I talked her into taking me to Vanity to buy some stuff for my senior pics and whatnot.&nbsp; Plus she doesn\'t mind buying stuff for me b/c she knows I won\'t lose it, wreck it, or stop wearing it w/in a week like Kait does...&nbsp; Muahahahahaha!  Went to&nbsp;lunch w/ mom and Don.&nbsp; Went to dinner at Red Robin w/ mom, Don, Kait, and grandma.&nbsp; So I haven\'t really eaten at home all day.&nbsp; This has got to be REALLY healthy for me...&nbsp; Who cares????  In a little while&nbsp;I\'m going to post about Ken...&nbsp; b/c I\'m in a really bad mood now and I feel like ranting and raving...&nbsp; That\'s the thing I hate about me and happy smoke, b/c I always get super-depressed the next day... Totally unfair... But I want to play a game first to vent some frustration/confusion/anger/resentment/etc. '),(16926,3776616,'23,July,2004','2004-07-23 00:00:00','    Muahahaha, Robert\'s was fun last night.&nbsp; Robert was  REALLY  drunk, and it was pretty damn funny.&nbsp;&nbsp; And then he got even funnier after we went outside w/ the Happy Smoke.&nbsp; We played Asshole and Fuck You,&nbsp;and&nbsp;I got pretty drunk, but no where near where I was on Robert\'s b-day.&nbsp; It was still awesome though.&nbsp; We played&nbsp; Spin-the-Bottle for awhile, but we never really did anything.&nbsp; Me and Andy sat on the couch while Beth, Robert, and Mande (she stopped by for a little while) were in Robert\'s bedroom.&nbsp; I can\'t even remember what we talked about... It made no sense whatsoever, and ever 5 mins we were like \'What were we talking about?&nbsp; Where did that come from?\'&nbsp; And every time we heard something weird from the bedroom (strange noises or comments) we\'d look at each other and be like \'we need to drink more\' and then we would, and it was fun :)&nbsp; I think I talked to Andy more last night than I\'ve ever talked to him in all other conversations combined.&nbsp; I asked him if he thought I was crazy -- he said yeah, but in the good way, not that bad way.&nbsp; Then Beth and Robert were geeking out about police and shadows and people chasing them... Robert made us turn off all the lights and hide in the dark in the \'safe corner\' of the bunk beds.&nbsp; Andy and I just laughed hysterically, b/c it was pretty&nbsp;damn funny.&nbsp; They were drunk enough that they were loud and stupid, and they were stoned enough that they were hallucinating.&nbsp; Muahahahaha :)&nbsp; But anyway, finally we sort of crashed around 2ish (I slept on the couch w/ Andy... Beth thought it was hilarious), then Beth and I got up at 7:30 so we could leave before Robert\'s dad got home.&nbsp; Sat in my car for a few mins, then drove to a dif. street and sat there for awhile.&nbsp; Then we went to Burger King, sat there for awhile, and finally came back and sat on a dif. street, where a crazy guy stared at us.&nbsp; Then a little after 9 I dropped her off and came home.&nbsp; Blah.&nbsp; When I got home my mom was in bed and someone was in the shower... and my sis isn\'t home.&nbsp; WTF?&nbsp; Turns out to be my mom\'s \'friend\' Don who is trying to get a job down in J-ville and spent the night here last night and now she\'s taking him around town and blah blah blah... w/e, I don\'t care anymore.&nbsp; If she wants to do things w/ other guys while dad\'s out of town that\'s her business, not mine.&nbsp; I\'ll just stay out of this one... '),(16927,3776616,'23,July,2004','2004-07-23 00:00:00','    I\'m finally home from house-sitting!&nbsp; Muahahahaha!&nbsp; It was fun, of course, but I\'m just glad to be home.&nbsp; In my own bed.&nbsp; In my own house that doesn\'t give me sinus attacks.&nbsp; With access to a computer (although it\'s a bit of a POS...).&nbsp; But anyway... this is just a little overview of how my week went, even though most of you know about it or where there.   Sunday:  Play practice was from 12-3:30ish, and that went ok, although Fred was there, which kinda sucked.&nbsp; Poor Hanlon...&nbsp; Anyway, then Mande and I went to Fran\'s.&nbsp; Sat in the pool for awhile, which was fun.&nbsp; Played with Goofy for awhile, which was just funny.&nbsp; Then Mande left, and I passed the time idly by unpacking my stuff, making random cards, blah blah blah...&nbsp; Spent the night alone in that creepy house... Fell asleep on the couch, but oh well.   Monday:&nbsp;  Mande came over in the morning and we hung out for awhile.&nbsp; We were supposed to go swimming w/ Lainie, but her mom freaked out (b/c she showered before she cleaned Nick\'s cage...), so she couldn\'t come w/ us.&nbsp; So Mande and I went to Taco Bell and the mall instead.&nbsp; That was fun too, b/c I haven\'t had Taco Bell in awhile,&nbsp; and we looked at all the fun Hello Kitty stuff at the mall.&nbsp; Then we went back to Fran\'s and played was (3 min game!), then Ken came over and we played cards some more...&nbsp; Then Mande went to babysit and Ken and I hung out for awhile... which didn\'t go very well, but oh well.&nbsp; Then I went to play prac., then back to Fran\'s.&nbsp; Lainie and Mande came over and we watched movies.&nbsp; Pookie spent the night.&nbsp; We stayed up until 2 talking about the&nbsp;imaginary rapists Leonard and Luke, bobbing for apples, and random other things..&nbsp;   Tuesday: (The best night of the whole week!)  Pookie and I woke up around 8:45 when Mande got there.&nbsp; Dropped off Lainie\'s van, then we went to MFR for breakfast, which was yummy :) (even though the creepy cook was flirting w/ us... EWWW!)&nbsp; Went to the park and played, then went to the mall.&nbsp; Then Lainie and Mande went home, then I went to play prac.&nbsp; Made lots of random phone calls in ref. to the bonfire, went to Subway to get dinner and saw Sarah, blah blah blah.&nbsp; Then after play prac. all the fun started. :)&nbsp;&nbsp; I went straight to Fran\'s.&nbsp; Then Mande and Max got there, and we tried to start the fire but that didn\'t work.&nbsp; Then Heathy got there, then Robert and Beth, so Robert got to fire started.&nbsp; We sat and made s\'mores and talked, then Robert tried to throw someone in the pool and it just went from there... pretty soon everyone was in the pool w/ their clothes on, wrestling and having fun.&nbsp; Sam got dropped off shortly after all the pool stuff started.&nbsp; Then Heathy left, and we hung out in the pool a little more.&nbsp; By the time we were done about half the water was in the grass, and the other half was brownish and filled w/ mud, grass, etc.&nbsp; So we drained the rest of the water, started refilling it, and sat by the fire.&nbsp; Luckily I had other clothes, but everyone else had to stay wet :)&nbsp; Then we had marshmallow fights, then played Red Rover for awhile.&nbsp; Mande and Maz left around 1, and me, Beth, Sam, and Robert sat by the fire until 2ish.&nbsp; Then we went inside and played w/ the magnets, piano, cat, food, etc. until 2:30.&nbsp; Robert burned his socks in the fire then put them on...&nbsp; Then we drove back to Robert\'s house, and Beth went home to say she was \'home\', then snuck back out.&nbsp;&nbsp; Then we stopped at Sam\'s then went to Perkins, which was really awesome.&nbsp; The sign in front said \'ALL U CAN EAT PUSSY $8.99 FRIDAY 4PM-10PM\'&nbsp; I noticed it and started laughing, then pointed it out to the rest of them, and I think we laughed at it for at least 10 mins...&nbsp; Then we went inside and ate and made disgusting things w/ our food.&nbsp; Robert hit his head on the light when we were leaving...&nbsp; We played the crane game, and Robert and I won matching hippos.&nbsp; They\'ve got some pretty damn ugly pants...&nbsp; Then we went to Woodman\'s to get a camera.&nbsp; Took pics of us in carts, then went back to Perkins to take pics of the sign and the love hippos.&nbsp; We used up all 27 pics in the 1 hour span... Muahahaha.&nbsp; By then it was about 5 and it was getting light out, so we went back to Fran\'s, talked until almost 6, then crashed until 8.&nbsp; Beth had to babysit, so Robert took her and Sam home, and I fell back asleep for awhile.   Wednesday: The first half of Weds is up there ^.&nbsp; I woke up at about 11 and just sat around idly.&nbsp; Watched a lot of MTV and Fuse.&nbsp; Cleaned up the rest of the backyard from the bonfire.&nbsp; Slept some more, then took a shower.&nbsp; Went to Dress Rehearsal, which went pretty well except for Fred.&nbsp; He had a script in front of him and he still missed some of Jessica\'s lines... idiot...&nbsp; Then I went back to Fran\'s.&nbsp; Mande and Ken came over.&nbsp; We watched TV for awhile, then Mande left, and Ken spent the night.&nbsp; That went horribly, but that\'s a whole different story and I don\'t want to go into it right now...  Thursday: Woke up at 8ish.&nbsp; Packed up all my stuff, did the dishes, cleaned things up (removed the \'unwholesome\' sayings from the fridge), blah blah blah. Watched TV, ate some Ramen noodles.&nbsp; Left a note for Fran, put all my stuff in ny car, then went to Milton.&nbsp; Picked up Beth and went looking for Robert, but we couldn\'t find him so we called his cell.&nbsp; Turns out he was going camping Thurs night, so the party wasn\'t going to happen... GRRRRR!&nbsp; Beth, Sam, and I all remembered him saying that Thurs worked better for him b/c Fri night he was going camping... Oh well, it doesn\'t matter b/c everything\'s straightened out now...&nbsp; Then I dropped Beth off and went to play.&nbsp; I was surprised, b/c I wasn\'t really nervous at all when I did my line.&nbsp; I thought I would be, b/c I was alone on the stage w/ the spotlight on me, but it wasn\'t all that bad.&nbsp; My mom and grandma came to watch, and Ken and Pookie and Jen.&nbsp; Pookie and Mel gave me flowers :)&nbsp; Thanks guys!!!&nbsp; Then me, Heathy, Mande, and Lainie went to Perkins.&nbsp; Had yummy pie, then went home b/c I was tired and I wanted to wash the stage makeup off my face.&nbsp;   So yeah, that\'s how my week has gone.&nbsp; It\'s been pretty cool :)&nbsp; I\'ve got the paly again tonight.&nbsp; Beth\'s gonna come watch.&nbsp; Then we have the party at Robert\'s.&nbsp; I might go to the cast party for a little bit, but I\'m not really sure.&nbsp; I\'m not even sure how I\'m going to make it to Milton on time tonight b/c my \'wonderful\' mother failed to leave me any gas money.&nbsp; She won\'t be home until prob almost 5:30, then I still have to go to the gas station, pick up Beth, and get to school but 5:45.&nbsp; I don\'t think I\'ll make it to Milton on the gas that I have, and I have enough $ for one gallon, which isn\'t going to help all that much.&nbsp; Grrrr.....&nbsp; Idk, I\'ll figure things out later.&nbsp; '),(16928,3776616,'18,July,2004','2004-07-18 00:00:00','    I\'m so confused right now...&nbsp; I don\'t know what to do about anything. I just want to curl up in a ball and scream...&nbsp; Ken came over last night.&nbsp; Things were fine for pretty much the whole night - we just played cards and shit - but then I walked with him out to his car.&nbsp; Oiy...&nbsp; I just...&nbsp; had a huge urge to kiss him, and he kissed me back, so it all seemed ok.&nbsp; But then afterwards it was really awkward, and I just wanted to cry...&nbsp; It\'s not fair.&nbsp; I still love him so much I can\'t stand it.&nbsp; I miss him when I\'m not around him, and even when I am with him I miss being able to be close to him w/o part of me thinking that I shouldn\'t be doing that.&nbsp; It\'s not fair...&nbsp; Why does this have to be so hard?&nbsp; After the kiss we sat in his car and talked for a few minutes... neither of us really said much... this is the&nbsp;general idea of it:  &nbsp;  *silence*  Ken:&nbsp; So what do you want to do?  Me:&nbsp; *pause* I don\'t know...&nbsp; I\'m really confused  *long silence*  Ken:&nbsp; Maybe we should try casual dating. More of the playing card games and stuff.&nbsp; We\'ll put the ring and the bracelet aside.&nbsp; And if you want to date other people, I have to let you. *pause*&nbsp; Because I know that\'s part of what this is, you want to experience other people.  Me: *pause*&nbsp; I don\'t know... maybe.  Ken:&nbsp; Will you think about about it?  Me:&nbsp; Yeah.  Ken:&nbsp; Ok. I\'ll call you tomorrow.  &nbsp;  Then I got out of the car and barely made it into the house before I started crying, which was absolutely fabulous...&nbsp; Ken forgot his phone and&nbsp;when he came back to get it I think it took me at least 2 mins to answer the door b/c I tried pretty much successfully to stop crying.&nbsp; Then I got ready for bed and stared at the wall thinking for an hour before I fell asleep...&nbsp; I don\'t really know what to do.&nbsp; I don\'t know if&nbsp;the casual dating thing will work out.&nbsp; It\'s just... Idk.&nbsp; I\'m afraid that I\'ll get too involved and it won\'t stay casual.&nbsp; And I really do think I need this break.&nbsp; As much as I love Ken and have no idea what I\'d do if I lost him, I think I need to be single for awhile at least.&nbsp; I want to stay friends with Ken though, which is the really hard part...&nbsp; Every time I\'m around him I just want to cuddle and kiss him and just... be happy, I guess.&nbsp; And I know it\'s somewhat the same for him, at least on some level.&nbsp; And all of this makes being \'just friends\' extremely difficult.&nbsp; Oiy... I think that the only way we can make this work is to try to distance ourselves in some way.&nbsp; Not by not seeing each other, b/c I don\'t think I could handle that, but... maybe just hang out in groups, at least for awhile.&nbsp; If there\'s other people there then all of my attention won\'t be focused on just him.&nbsp; I hope...&nbsp; Maybe us both going on vacation is a good thing.&nbsp; He\'s leaving for Pennsylvania on Sat and won\'t come back until the day I leave for Florida, so we\'ll have 2 weeks of not seeing each other, and probably&nbsp;not much&nbsp;talking to each other.&nbsp; Maybe that will help in some way... And then when we\'re both back we can just hang out during group things for awhile... And then school will start so my mind will be occupied a lot more so it\'ll be easier for me to think about him less often.&nbsp; And then hopefully we\'ll both get to a point where we can hang out just the two of us w/o everything being awkward.&nbsp; There\'s so many \'if\'s in that though...&nbsp; And I don\'t know how he feels about any of it b/c most of it just came to my head last night or this morning.&nbsp; *sigh* Idk what I\'m going to do... I guess I\'ll talk to him tonight or tomorrow night, see how he feels about all of this.&nbsp; We can discuss the casual dating thing too, but I just don\'t know how well that would work out...&nbsp; ARGH! '),(16929,3776616,'17,July,2004','2004-07-17 00:00:00','    Ok, this is just my rough sketch of this week\'s schedule.&nbsp; Heathy calls Tues or Thurs (if you don\'t call me I\'ll call you).&nbsp; Except for the bonfire night I should prob. only have 1 or 2 people over at a time.&nbsp; If you want to come over just call me, it\'s all good.&nbsp; I\'ll be there most of the time, except when I\'m at play prac., visiting Lainie, or doing something random.&nbsp; If you need a ride I can prob pick you up.&nbsp;  &nbsp;  Sun - Mande during the day, no one overnight yet  Mon - Lainie Tues - bonfire :)  Weds - no one yet  Thurs - no one yet  Fri - no one yet  &nbsp;  Oh yeah, one other thing -- I probably won\'t be able to update my blog this week b/c Fran\'s computer doesn\'t work.&nbsp; I won\'t be able to read anybody else\'s blog either, unless I go home for awhile or go to the internet cafe or something.&nbsp; So yeah.&nbsp; Limited internet for Brit :)&nbsp; That just means I\'ll spend all my free time reading or playing card games or something.  &nbsp;  But yeah... that\'s about it, so I\'ll see y\'all this week sometime. '),(16930,3776616,'17,July,2004','2004-07-17 00:00:00','    Muahahaha, I just woke up from my nap :)&nbsp; I got home from the car wash around 2:30ish and I fell asleep by 3 and didn\'t wake up until a little after 5.&nbsp; I\'m still kinda tired but I can sleep tonight, so it\'s all good.&nbsp; The car wash was... ok.&nbsp; I was the first person there, so I sat in the parking lot for 5ish mins w/ the seat back and listening to music before Fran got there.&nbsp; Then we went in to talk to the guy that was there and they had NO idea what was going on... So&nbsp;the wonderfully smart \'general manager\' that I talked to failed to mention the car wash to anyone else... I knew he was stupid when I talked to him, but oh well...&nbsp; Anyway, we got stuff worked out and everything set up all right.&nbsp; There were only a few people there -me, Mande, Jon, Annie, and Meghan, then Sam came at 10 and Heathy around 1:30ish.&nbsp; Some peope *coANNIEugh* were being a bit bitchy and rude the entire time.&nbsp; Mande and I were talking about it -- don\'t you think that you should prob. NOT be super rude to your captains?&nbsp; They have a little more power than you.&nbsp; Dur.&nbsp; But anyway, overall things went fairly well, despite the poor weather and lack of help.&nbsp; I\'m&nbsp;not really sure how much we made, but it wasn\'t as much as last time.&nbsp; We has mushroom pizza for lunch, that was good.&nbsp; Now I\'m hungry again though... &nbsp; My mom still isn\'t home...&nbsp; She got to Pizza Hut shortly before I left so she could eat, then she was going to run a bunch of errands and come home... She\'s getting me some snack foods and soda and shit to take with me to Fran\'s.&nbsp; Speaking of that -- I\'ll put up a little schedule later of all the days, who\'s coming when, what\'s going on what day, etc.&nbsp; All that stuff.&nbsp; For now I\'m just going to play games on MSN to amuse myself... '),(16931,3776616,'16,July,2004','2004-07-16 00:00:00','    I\'m so freaking tired...&nbsp; Like, I just want to pass out on my bed for a few hours, but I\'d rather write in here for awhile first so I can vent a little bit.&nbsp; And I wish I had my cell phone, but I left it at Robert\'s and I\'m too lazy/tired to go back and get it... So blah.&nbsp; Anyway, this is how everything has gone... b/c last night was kinda bad.&nbsp; At least one specific part of it... &nbsp; Went to visit Lainie at work around 5:30 ish, then went to play prac.&nbsp; Idk if it was hot in there or what, but I kinda took a downhill slide and felt shittier after I\'d been in there for about 15 mins.&nbsp; Then I went to do my line and I talked to fast and couldn\'t remember a little part and I wasn\'t authoritative enough and blah blah blah...&nbsp;&nbsp;Church was grumbling, and it wasn\'t even like it was a big deal or anything, but I&nbsp;was ready to snap on him.&nbsp; That or just run out of the&nbsp;room crying...&nbsp; I was so fucking emotional... grrrr..... &nbsp; But yeah, prac got done at 8:45ish, then I went over to robert\'s.&nbsp; Him,&nbsp;Beth, Sam, Andy, and&nbsp;Jake were already there playing&nbsp;badminton in the backyard.&nbsp; So we hung out there for awhile, then we went inside and ordered pizza.&nbsp;&nbsp;We sorta just watched TV and waited for the pizza to come.&nbsp;&nbsp;Jenny and AJ both showed up.&nbsp;&nbsp;Then Jake and Jenny left to go get some guy and came back with 2... hat guy and no hat guy.&nbsp; Brian and Derek.&nbsp; Minty Fresh (bad story behind that, I wish I hadn\'t heard it) and Miller Time.&nbsp; Call them whatever you want...&nbsp; hat guy was the sorce of all of my problems.&nbsp; Anyway,&nbsp; a little while after Robert\'s dad&nbsp;left we started drinking.&nbsp; We played a game called Fuck You, and that was pretty damn fun.&nbsp; Everybody fucked everybody, isn\'t that your idea of a good time?&nbsp; It got me pretty damn drunk... I was fine until I had to do the bong, and that\'s when things started to go downhill.&nbsp; But first -- how come no one went in the kitchen to watch when Robert bonged it, but when I had to the whole fucking party followed me into the kitchen?&nbsp; I hate audiences... Beth said it\'s b/c I\'m a girl... Does that make it funnier or something?&nbsp; Idk...&nbsp; But I did it no problem, and about 5 mins later I was pretty far gone...&nbsp; And this is where my details start to get vague... B/c I can\'t really remember everything that happened, the order it happened it, or who was there...&nbsp; Anyway, we went outside w/ Beth\'s pipe and passed that around for awhile.&nbsp; Then all of a sudden I was hanging all over hat guy and he was doing the same...&nbsp; Then we fell on the grass and made out a little bit... that was pretty gross, b/c he tasted like beer (although I imagine that I did too), and he was a bad kisser.&nbsp; The main point of the story is that little asshole was hanging on me all night and I was too damn drunk to care and I guess it got bad enough that Andy was making comments to Beth and him and Beth and Richie and Idk who else were all watching hat guy so he wouldn\'t do anything.&nbsp;&nbsp;&nbsp; It\'s kinda reminiscent of the Christmas party at Gregg\'s...&nbsp; There I had Amanda, Gregg, Jim, Missy, and god knows who else watching me to keep other stuff from happening...&nbsp; But yeah...&nbsp; lots of&nbsp; \"Zirkels of Love\" w/ everyone dancing in a big circle in the kitchen.&nbsp; I have crazy bruises on my feet from people stepping on me.&nbsp; Plus one on my hand that appeared out of no where, and one on my arm where Sam bit me...&nbsp; But then Beth and I left around 2:30 to go back to her house... to get away from the crazy hat guy...&nbsp;NO MORE PARTIES WITH HAT GUY!&nbsp; If I ever go to a party and hat guy is there I\'m leaving, I swear to god.&nbsp; &nbsp;Oiy... I don\'t want to go into any more details about last night, if you know about it you know about it, otherwise too bad.&nbsp; Lainie I might call to tell you some random things... Ken I might tell some stuff, just b/c I think we both need to work on the \'getting over our physical and emotional attachments\' thing...&nbsp; We both have a problem with that right now...&nbsp; Anyway, that\'s a bit off topic. Now to today... &nbsp; Beth and I got up around 9:15ish.&nbsp; I put my contacts in, we drank a glass of water, filled some water balloons (w/ intent to throw them at hat guy if he was still at Robert\'s), then walked over to Robert\'s.&nbsp; Instead we put the balloons under Robert\'s car tires (muahahahaha!!!) the peeked in the window to see if anyway was awake and walked inside... Saw Robert sleeping on the couch.&nbsp; Poked him until he woke up.&nbsp; Then we went in to see Sam, and the 4 of us sat in the living room talking.&nbsp;&nbsp; Blah blah blah, that was fun, luckily hat guy was gone before we got there though.&nbsp; And eventually everyone else that was still there woke up and we just sat in the living room and watched TV and talked.&nbsp; Robert made pancakes and french toast.&nbsp; Watched more TV, then people started to leave...&nbsp; Around 2:30 it was just me, Robert, and Andy, and I was getting tired, so I left.&nbsp; When I got home&nbsp;I realized I\'d left my cell phone at Robert\'s house... Go me...&nbsp; Oh well, I just called it and told him to take it and the disco ball over to Beth\'s, and then I can get my phone from Beth some time this weekend.&nbsp; So yeah, don\'t bother calling it unless you\'d like to hold a conversation with Robert. :)&nbsp; And now I\'m going to do a few more things then take a nap.&nbsp; Then I\'ll shower, change, make the car wash posters, and probably by then Ken will be here... It\'s already 4, and he said he got off at either 6 or 8, so...&nbsp; Idk.&nbsp; Argh, I just need to crash.&nbsp; Later. '),(16932,3776616,'15,July,2004','2004-07-15 00:00:00','    Yesterday I wa pretty much glued to the couch watching MTV and music videos all day...  I was pissed about the Fugliest Rock Star thing -- they said Chad Kroeger was fugly, and I don\'t really think he is but I guess you could see it that way, but THEN they tried to tell me that Sully Erna (Godsmack) was fugly!  He is  NOT  fugly.  He\'s pretty damn sexy if you ask me, but maybe it\'s just b/c I like that type... Tatoos, piercings, slightly demonic appearance...  But anyway.. .  I got up at 8:30 yesterday and other than getting up to get a glass of water, take some ibuprofen, get the phone, or go to the bathroom, I didn\'t get up until 6, and that was only to go plant my ass on my bed instead.  I slept for like... 5 hours during the day, then I went to bed at 9:30 and didn\'t wake up until 11 this morning... YAY!  But I\'m starting to feel better.  I was actually able to eat last night.  My dad brought me home stuff from Noodles at 8ish, and I ate almost the whole thing, and the only reason I stopped was b/c it made me full (Tues all I ate was 15 little seashell noodles and a small piece of cantelope, Weds all I\'d eaten before the Noodles was 3 olives and 2 pieces of cantelope...).  And I\'m actually able to go 5 or 6 hours w/o taking more ibuprofen, instead of wanting to take 4 every 4 hours...  And now I can drink a glass of water/iced tea in 15 mins instead of 2 hours or not at all...  So yeah, I see it as an improvement.  I\'m actually going to go to play prac tonight, both b/c I think I feel up to it and b/c I\'ve already missed three and I don\'t want to miss any more.  If people are gone I\'m not sure how long I\'ll be able to read their parts for b/c talking/laughing/etc. for extended periods of time makes my throat hurt more.  But anyway, after prac. I\'ll probably end up going to the party at Robert\'s, although I don\'t think I\'ll drink.  If Beth even brings the alcohol, b/c she told me she wasn\'t really sure she was going to.  I def. won\'t do the beer bong b/c that\'s shraing of germs and I don\'t want to make anyone sick.  No spin-the-bottle for me either...  Oh well...  Anyway, I have some more I want to write about but I\'ll do that later or tomorrow. '),(16933,3776616,'13,July,2004','2004-07-13 00:00:00','    The 4N6 meeting was short, but then I had to go over to Fran\'s so she could show me around her house and whatnot.  I start house sitting on Sunday, so if someone wants to claim Sunday night just call me.  We can eat popcorn, watch movies, play w/ the fat cat, build a fire, whatever...  She\'s got a little pool too, so we can go swimming...  It shall be fun :)  I have 5 or 6 nights, and people can come ovr during the day too, so...  I\'ll talk to everybody at the car wash on Sat. if I don\'t call you sometime this week.    I\'m so freaking hungry, but I can\'t eat anything w/o wanting to scream.  I want Taco Bell soooooooooooo fucking bad...  There\'s no way in hell I could swallow that that.  I hope I feel better by Thurs, I don\'t want to miss my pics and Robert\'s b-day thing and all that fun stuff.  I get to see Ken tonight after he gets off work... I\'m not sure how it\'s going to go, b/c we haven\'t talked since saturday.  He read my blog and posted a comment, plus we\'ve texted back and forth a little. I suppose I probably would have answered last time he called if I could actually talk.  I don\'t know how well tonight is going to go, but...  *crosses fingers*  I hope we can work everything out.    But yeah, I\'m gonna go take some more ibuprofen and see if I can find some soft food to force down my throat. '),(16934,3776616,'12,July,2004','2004-07-12 00:00:00','    When I FINALLY got home at 10:15 last night after running around to god knows how many places I just crashed in my bed...  The PA at Urgent Care said I have strep...  The rapid strep test came back neg, but she said it looked exactly like strep to her, so she gave me amoxicillin anyway.  She\'s going to call me to tell me for sure as soon as the other results get back from the lab... Blah this really really sucks.  I can\'t swallow at all (go ahead and laugh, my sis laughed at that too -- that\'s NOT how I meant it), and I still can\'t talk very much.  I\'m gonna go to the 4N6 meeting today, but I doubt I\'ll go to play practice.  If I can\'t talk I\'m not much help am I?  Plus I\'m prob contagious, so I don\'t want to make the rest of the cast sick.  Otherwise they\'d all be gone on the days of the play, so I\'d be the only healthy person, and I\'d have to run madly around the stage trying to play all of the characters...  Oiy...    but yeah, Mande\'s on her way here, so I g2g.  Poo-head I\'ll talk to you tonight, Pookie I\'ll call you if I feel better, and Bethy I\'ll call you tomorrow, k? '),(16935,3776616,'12,July,2004','2004-07-12 00:00:00','    Oiy vay, I figured out why I couldn\'t log in to anything earlier today... My mom disabled cookies last night when she was on the comp.. WTF???  Oh well, everything\'s fixed now, so it\'s all good.  Anyway...  sorry I haven\'t really been answering calls/talking long/calling when I said I would.  My voice won\'t go above a whisper at the present time.  I\'m going to Urgent Care to get a strep test and whatnot, but I have be dragged to Milton for my sis\'s 4-H picnic thing first.  So Pookie, I might be able to stop in and see you at work *crosses fingers*  umm... Bethy I might call you tomorrow otherwise I\'ll just wait until you call on Weds, and I\'ll see you on Thurs.  Sam... idk if you read this or not, but we really need to talk.  So I\'ll call you as soon as I get my voice back.  Plus we can go to the gym then too, before you go on vacay, b/c I\'ll be leaving for Fla. shortly after you get back, and by then it\'ll be almost the middle of August...  Other things...  Ken I\'ll try to call you back when I get home tonight I\'m not sure what time it will be or how much I\'ll be able to talk but I\'ll give it a try.  Mande I\'ll see you tomorrow at 9:15 are we supposed to bring anything or is she going to have all the cleaning supplies?  EEEEEK...  I need to stop spazzing and mom\'s ready to go so I\'ll talk to everyone w/in the next few days. Later. '),(16936,3776616,'11,July,2004','2004-07-11 00:00:00','    Heathy and I went to Frostee Freeze at 6ish.  That was fun, and I got my banana split that I\'ve been craving for god knows how long.  Plus I got out of the house and occupied my mind a little...  I was talking to Jorge earlier and I said something about \'my ex\' and I didn\'t even realize it until I went back and read what I\'d written.  That was kinda weird, and it kinda made me cry a little b/c the word \'ex\' has such a negative connotation and it doesn\'t feel entirely appropriate for me to use it in this situation...  Idk...  I have to clean up my room tonight, and fold my laundry...  Mom bought me some new clothes yesterday, a new purse, new flip flops, plus a charm bracelet.  One of those add-a-link things.  There\'s a Happy Bunny charm at Claire\'s that I want to get for it once I get more money.  They had these cute little stuffed green and purple frogs at Kohls too, I want to get one of those too, they were soft and cuddly.  So were the little hedgehogs.  And I saw Nicole Petroff.  I didn\'t realize that she still worked there...  But yeah, that\'s been the small amount of \'Bright Side\' to my entire weekend...  Hopefully tomorrow is decent.  Play practice is going to suck b/c I can\'t really talk above a normal tone... stupid sore throat.  Mom was going to take me to Urgent Care, but what will they do?  Tell me my throat hurts, give me some pain pills, and tell me to go home.  Come back if it\'s not gone in a week.  So why don\'t I just wait and go in a week if I still have it?  I\'ll just keep taking ibuprofen...  ARGH.  Life sucks.  Hardcore. '),(16937,3776616,'10,July,2004','2004-07-10 00:00:00','    So this... this is going to be a REALLY long post.  I can guarantee it won\'t make sense...  This is for for me, so I can organize my thoughts; Ken, so I can explain EVERYTHING; and for everyone else who cares enough to read the whole thing...  I\'m not really sure where to start this.  There isn\'t really a definite beginning..  Idk, I\'ll start w/ when Ken called me while I was at the mall...  By then I\'d already had it made up in my mind that I was going to tell him everything that night.  Then he started telling me about how shitty his past two days had been, and all the bad things that had happened, and I wasn\'t really focusing much on the convo b/c my mind was 8000 miles away from anywhere, just mumbling replies every now and then...  And then when he hung up he didn\'t say I love you, which I think is the first time he hasn\'t in a long time, and I wanted to cry b/c I knew that he knew, and then I felt REALLY bad for wanting to dump all of this on him on top of his super shitty past few days...  So anyway, then I texted Lainie and told her that I felt bad for doing it that night, and she called me back a little later and I vented on her for a few minutes but she had to get back to work so I didn\'t really accomplish much I just told her I\'d call her tomorrow...  So then I went home and I think I was home for like, 2 mins before Ken got there, and I was sitting on my bed about 2 sesconds from hysterical sobbing and I didn\'t know what to do...  I couldn\'t look at him knowing that he knew...  And all the words had totally evaporated from my mind.  I really didn\'t know why to say, I just kept my face on my pillow and tried to keep myself from crying...  Then I finally managed to semi- look at him, and he told me that he was pretty sure he knew what was coming, and he was just waiting to hear it.  The tone in his voice... he just sounded so upset, and that made me cry...  then he gave me a note that he\'d written and went to the bathroom.   No way in hell would I have been able to read it in front of him, so I went outside and sat in the rocks on the side of the house, reading and crying and ignoring the mosquitos...  The letter hurt a lot, b/c... some of it was right, but some of it was totally wrong, as far as how I felt and why I wanted to break up with him...  And I wanted him to know that, but there was no way I could tell him that night.  And that\'s what he said in the letter too, that if I broke w/ him that night he wouldn\'t be able to talk about it and he\'d want to leave right away.  Anyway... then I came back inside and leaned up against the wall next to him, and all I could say was \"I don\'t know how to phrase it,\" and he said something like \"I don\'t know why you have to find the right words to just say it\"...  And... god, the tone of his voice... he sounded so.. hurt... angry... upset... The way he said it hurt me more than anything had the whole night, and I just blurted out \"it\'s not just breaking up, there\'s so much more to it\" and then I started sobbing and couldn\'t stop...  then he started crying too and I felt so horrible and I wished more than anything everything in life could just be perfect and fair...  Then he said that he was going to leave, and he\'d call me later, and that he wanted me to write down what I wanted to say to him...  So I just wrote down my blog URL and gave that to him, told him that he could find a lot of stuff here...  I\'m not sure that was the best idea, b/c all of the older stuff is missing,  plus...  I\'m not really sure what I\'ve said about him on here.  Nothing bad, I know that, just...  stuff I might\'ve said out of frustration or something...  But I can\'t hide from shit anymore...  I\'m honest with everything else in my life, this was the last piece of the puzzle.  He has every right in the world to know everything I\'ve ever felt.  But yeah... after he walked out of my room I threw myself down on my bed and curled up under my covers and sobbed for god knows how long... Kait came in and wouldn\'t leave me alone \"is something wrong brit?  is there something wrong?  what\'s wrong?  is it something with your friends?  with ken?\" and she wouldn\'t fucking go away...  Finally I kicked her out and just sat in my room crying...  I couldn\'t find any Kleenex, so I had to use a roll of toilet paper...  I turned on music b/c music always helps me... Of course the first CD that came on was Evanesence...  That makes me cry anyway, so it just pretty much kept me crying,  but I was too stupid to change the CD... Then I texted Lainie to call me when she got home from work, then I started throwing things around my room...  so it\'s pretty much trashed right now... then I dumped all my journals all over the floor and got all my Ken stuff and put it in that box and shoved it under my bed...  The ring, the bracelet, v-day card, the pic of us w/ the tiger, the handcuffs, the letter (the only letter I\'ve gotten from him in 7 months, and it was about breaking up), other stuff... everything that reminded me of him, b/c I couldn\'t look at it without crying...  The only thing I left out was the giant monkey, b/c it was comforting sorta.  B/c like, after I\'d trown everything in the box I just wanted more than anything to hug him and cry, so I hugged the monkey instead...  Then Lainie called me a little after 10 and we talked for over an hour, and by the end of the convo I felt a bit better...  She offered to let me come over to her house, but I didn\'t have a way there, so it didn\'t really matter.  Then I tried to eat ice cream, b/c ice cream is an awesome comfort food, but I hate vanilla, and all we had was nasty tasting chocolate stuff...  Then when I hung up w/ her I was going to watch a movie, but the only movie that sounded good at the time was Sleepless in Seattle, which A) is Ken\'s; and B) last time I watched it was with him.  So I ruled that out, and figured that I\'d just force myself to sleep instead.  Which actually worked fairly well, most likely thanks to the convo w/ Lainie calming me down a lot...  Before I fell asleep I texted Ken... apologizing for the way things had gone, and asking him to read my blog when he got a chance... then he called me, and I couldn\'t answer it... I hit the silence button after the first ring, then I stared at my phone for another ring, then threw it on my bed an rolled over on my side...  waited until the phone message alert beeped then listened to my voicemail...  then fell asleep...  any now today is a new day, and I haven\'t really done much all day...  Crazy Katie and Beth called wanting me to come over to Robert\'s, but I\'m not really in the mood to hang out with anyone...  Lainie said she\'d call me tonight.  And Ken said he\'d call me today, but... idk if I can.  I\'ll probably just ignore his calls for a few days, until Tues or Weds maybe...  I need to get drunk so bad right now...  but I have neither money nor alcohol at the moment...  ARGH.... what\'s it going to do anyway?  I\'ll have a few hours of happy time, then everything just comes back anyway...  But now... the rest of this...  that ^ was the how, this is the why.  This is everything, folks...  100% why I feel the way I do and why I did what I did, so repeated some new, depending on who you are... here goes... mostly addressed directly to Ken...  Ken I love you so much I don\'t know how I can make you see that... I didn\'t do this b/c I hate you, I did this for so many reasons it\'s going to take me hours to write it all...  First off, the letter -- you were right and wrong, as far as the ring goes.  Yes, it did scare me, b/c suddenly it put ideas in my head that I never realized I\'d think about this early in my life...  Just... commitment in general, especially.  I\'m too immature to think about that right now, not on any level.  I still need to grow up and figure myself out.  I kinda described this in my \"Ken and Sam\" post...  I just feel... inferior as far as experience and maturity level go, and I don\'t think that\'s a healthy thing to have in a relationship.   I\'d be a much better person, friend, girlfriend if I could just grow up and experience the world around me.    And I\'m so sorry that I never told you these things right away, it was just that...  all of my friends seemed so excited and happy for me, and you just seemed so happy... I didn\'t want to ruin that.  I didn\'t want to burst everyone\'s happiness bubble...   And then in that first week everything hit me so hard and I overreacted to a lot of things and I just kept digging myself deeper and deeper into a hole I had no way to get out of.  And in a way breaking up kind of threw me a rope to get out of that hole, but now I\'m in a completely different hole, and this time I have no one to hold and be with...  But I suppose that\'s what I get for being selfish and wanting to \'try out\' other people before I commited myself to one person.  And I\'m terrified that when I grow up and finally prove to myself that it is you that I want to be with, you won\'t be there anymore.  And it\'ll be no one\'s fault but my own, and that\'ll only make it harder to deal with...  But I suppose there\'s nothing I can do but wait and live life the way it is right now...  Besides the ring, age difference, maturity level things...  Those were the main reasons I need the break from you.  There\'s everything with me wanting to date other guys before I commit myself to one person...  And I know that it sounds selfish and horrible, like trying on clothes at the store and just returning the ones that you don\'t want...  You find something you love, but then you want to try a few other things before you buy it...  You can put a hold on it and leave it at the layaway counter, but that\'s not fair to all the other shoppers.  But if you put it bak on the rack, there\'s always a chance someone else will come along and snatch it up...  Oiy, I know that was a horrible analogy, but I\'m not really thinking a whole lot right now, just typing.  And then there\'s other stuff.  Little stuff, pretty much stuff that\'s just in the back of my mind, but random things nonetheless.  Things that we need to addressed if we get back together sometime in the future.  And some of this is my fault, simply b/c I can\'t talk about things very well and I\'m not very good at voicing when things are wrong...  First is the emotional vs. physical thing.  I know we\'ve talked about this several times, and every time I bring it up you would get better about it for awhile, but then after a while things would go back to kinda like what they had been before.  Never as bad as it was in the beginning, but back nonetheless, and after awhile I just gave up on bringing it up.  It was extremely unfair of me to expect you to entirely change yourself for me, b/c there are some things abut me I wasn\'t and still am not willing to change about myself for you.  Like the blowjob thing, b/c I know that no matter how many times you said it didn\'t bother you that it really did.  You made enough comments that I picked up on that quite quickly...  but as far as that goes...  I know that it\'s something that I can get over eventually, but bj\'s in general are just nasty.  Maybe some people find drinking salt water to be appealing, but it\'s not very high on my list.  And I know that there\'s the whole \'reciprocity\' thing -- you do this to me, I do that to you, blah blah blah, but...  Tell me, at any time did I even ASK you to do anything to me?  Everything you\'ve done was upon your own free will.  I mean, of course I enjoy those type of things, but I\'m perfectly happy without them, as I\'ve tried to explain to you.  Especially when we\'re at my house, b/c there\'s too much of a chance of getting caught.  I really don\'t like the idea of my parents accidentaly walking in on something or coming downstairs and seeing something they don\'t need to see.  They\'d lose respect for me, and they probably would have prevented me from seeing you as much too.  And as long as I\'m living in this house there isn\'t a whole lot I can do about what they think of me.  It\'s a little late now, but I\'d be so much happier if they were pretty much clueless as to what I do with boys.  I\'d rather have them thinking all we do is play cards, watch movies, have water fights in the backyard, etc.  Which would actually be quite fine w/ me if that were pretty much all that happened anyway.  But I\'ve said that countless times before, and if you haven\'t figured it out yet I don\'t know how else to explain it to you.  If that bothers you a lot maybe you should have brought that up to me every time I told you how I felt...  As far as sex itself goes... that was never a problem.  You are always so patient, and you never tried to talk me into anything.  You have no idea how much more that made me love you.  I know how much you wanted to have sex, yet you never brought it up unless I brought it up first, and never got mad or defensive or whatever whenever I gave you a reason why we couldn\'t at that time.  Thank you so much.    So yeah... that was that.  Please don\'t think the sex and stuff were the reason I broke up with you, bc they weren\'t.  I really do think that I need to a break to let myself grow up, and I hope to god that you\'ll see that eventually and not hate me for all of this.  And hopefully in a few days I\'ll be able to talk to you, and we can talk some of this stuff out w/o either of us freaking out (most likely me).  And hopefully none of this will be a problem later on, and eventually I\'ll be better at communication.  So I guess I\'ll leave you with one final message and quit typing, b/c I\'ve been writing this for over 2 hours and I imagine it\'s getting to be a bit long...  I never meant to hurt you like this...  I didn\'t want to be the one to come out and say it directly.  I wanted you to see all of this ^ and see why I feel the way that I do.  I wanted you to understand and agree with me that breaking up is a good idea for right now.  I wanted everything to be as close to a mutual agreement as possible...  I hope you believe me when I say that I really don\'t want this break up to be forever.  Just a break, time for me to discover myself.  I don\'t know how long I need...  6 months?  a year?  two years?  All I can do is hope that we can stay friends throughout that time, and that when I am ready you\'re still available...  I love you poo-head. '),(16938,3776616,'10,July,2004','2004-07-10 00:00:00','    Last night I was forced to go to a Mary Kay party that was supposed to start at 7/7:30ish, but my  lovely  mother wasn\'t even ready to leave the house until almost 8.  Then we didn\'t leave until after 10:30, which was simply  fabulous  b/c I was tired, bored, irritable, and in pain (sore throat/earache, both on left side).  OF COURSE it was at her friend\'s house, and she had to just chatter endlessly...  Then I got home and took meds and sat in my bed in a fairly pissed off mood.  Ken had called me at 7 to see if I was back from my \'kidnapping\', but dur, I hadn\'t even left yet.  So he was going to go to Orfordville after he got off work and then when he got back he was going to call me and make me watch him light off fireworks or whatever.  I wasn\'t listening all that intently to the conversation...  Anyway, 11 rolls around and he  STILL  hasn\'t called me, so that didn\'t help my mood much, b/c there was no way in hell I was still going to go out anywhere that late feeling like I did.  Plus I was pissed off b/c after reading Lainie\'s comment on one of my posts, plus random other little things, I felt like I had my thoughts collected enough to tell Ken   EVERYTHING  .  And now, of course, I wasn\'t going to get to see him.  So I just texted him \'Im going 2 bed.  we need 2 talk 2m\' and hoped I\'d be in the mood for it tomorrow.  Then I got ready for bed, and about 20 mins later, right before I turned off my light at whatnot, I texted him again \'I meant what I wrote in the letter\', I don\'t really know why I texted him  that , I was just in a sharing mood and figured I\'d take advantage of it...    Oh yeah -- my mom talked to Fran and everything is good w/ the house sitting thing. :)  They talked for like, 20 mins, and my mom won\'t tell me what they talked about b/c she said she told Fran she wouldn\'t.  They doesn\'t really sound good, does it?  And then there were my mom\'s replies... \"Don\'t worry about that, she likes vodka\" and \"I\'ll check her luggage before she leaves\"   Hmmm.. comforting, isn\'t it?  Anyway, she told me that Fran\'s going to talk to me after the 4N6 meeting on Tues. and give me all the details and whatnot.  So yeah, sometime next week I\'ll talk to everyone about spending the night and whatnot.  Today\'s probably going to suck for most of the day...  I have an eye appt. at 1, but it\'s supposed to be really short.  Ashley\'s grad party starts at 12, so I\'m just going to go after my eye appt... I didn\'t RSVP, but I don\'t think it\'ll matter, b/c I\'m just going to stop by and say hi anyway...  I have to go shopping w/ my mother at some point today, so she can buy me a new outfit for my senior pics...  And today\'s my grandma\'s b-day, so we\'ll probably all go out for dinner...  And  then  I\'ll get to talk to Ken tonight, which will be one big fucking party...  If he doesn\'t come over tonight I think I\'m going to just snap.  Rattle everything I want to say off in one big long sentence over the phone.  I can\'t deal with this anymore, honestly...  notfairnotfairnotfair...  *runs headfirst into a cement wall* '),(16939,3776616,'09,July,2004','2004-07-09 00:00:00','    Went to Subway w/ Lainie.  Talked about boys and parents, that\'s about it.  And scaring people, b/c that\'s always fun :)  Then we walked down to Pier 1, b/c neither of us had been to this one yet, and Lainie didn\'t really want to go home.  We picked out all the stuff that we\'ll eventually have for our houses... :)  Themed rooms of course, b/c that\'s the best way :)  Then we sat in my driveway listening to music and fogging up the windows until Lainie had to go home.  Muahahahaha!    Anyway, dad\'s home and I\'m bored, so I\'m gonna go in my room.  I still need to give him an answer on whether or not I\'m going to Florida w/ him and Kait. Should I?  The idea of being alone in the hotel room/timeshare w/ Kait while he\'s working doesn\'t seem all that appealing to me, but beaches and whatnot.... Hmmmm...  Idk, he needs and answer by tonight, so I suppose I\'ll figure it out.  Anyway, later! '),(16940,3776616,'08,July,2004','2004-07-08 00:00:00','    Fran just called me...  She wants me to house sit for her while she\'s on vacation...  Like, feed her cat and stuff, plus she said I could stay at her house the whole time.  And she said she knew that it\'d probably be creepy sleeping at her house alone, especially sinc it\'s out in the countryish, so I could have a friend or two over at night if I wanted.  No big parties or large groups, but she has a fire pit in her backyard (s\'mores!) and all that fun stuff.  So idk... I mean, it\'d be cool and all, and it\'ll give me a chance to get some money.  Plus I\'ll have my dad\'s car that week, so it\'ll be easier to get back and forth from Milton and whatnot, b/c I\'ve got play stuff all that week...  But all of this is dependent on whether or not my wonderful mother agrees to it...  I talked to my dad, and he didn\'t really seem to care, but he\'s going to be out of town for the next three weeks so...  Idk, I think my mom will eventually go for it, it just might take a little convincing.  And I don\'t think she\'ll let me stay over the whole time, but w/e...  So yeah, if I get to do this, who thinks they\'ll be able to spend a night or two with me?  Bethy? Lainie? Mande? Nicole? Heathy? Sam? Mystery Person?  I think I\'ll have.. 5 nights?  I can\'t remember exactly what she said...  Anyway, I\'m just gonna sit and wait for Lainie to call to see if her mom is spazzing or not... Hopefully not, b/c I really do want to talk to her in person about all this Ken stuff... '),(16941,3776616,'08,July,2004','2004-07-08 00:00:00','    I\'m just going to rant about my two main problems at the moment...  This whole Sam thing... I just don\'t know what to do anymore.  I love him to death and I don\'t want to lose him, but...  He told Mande that I don\'t put enough effort into our friendship, and apparently Hayley was at his house the day he and I were fighting b/c she knew about everything and said something to Heathy about it at work.  Plus she thinks that I\'ve never liked her, which isn\'t true.  I just never  KNEW  her, and the only thing that I did know about her was that Sam is always talking about her, and that is seems like he\'s hanging out with her all the time.  And that just makes me feel like shit, b/c I feel like I\'m losing him, and I don\'t want that.  It just doesn\'t make sense to me how everything could go from the way it was last summer to the way it is now...  And I\'m not saying that it isn\'t partially my fault, but in no way am I willing to take all of the blame.  I mean, I know I haven\'t called him much, but then again I haven\'t really called anyone a whole lot.  Beth and I call each other back and forth, so we talk a few times a week, I\'ve talked to Nicole once or twice, Lainie a little bit...  I mean, Sam tried to tell me that he was sick of being the one that called all the time, so he was just going to quit calling.  That doesn\'t work very well, does it?  Idk...  Maybe I\'ll call him tomorrow or Sat....   And then there\'s Ken...  I love him so much, it\'s just...  I  really  do need that break.  And I don\'t think he took the letter I wrote him too seriously.  I need to sit down and talk to him, bring it all up again, this time in conversation so I can make sure he grasps the full extent of everything...  It\'s just... Idk how to tell him...  The whole \"Let\'s just be friends\" thing is too cliche, but that\'s sort of what I\'m going for.  I don\'t want us to be a couple, b/c I need time to be single and grow up and whatnot, but I don\'t want to totally lose him.  And when I say that I\'d like to consider the possibility of us being together in the future I really do mean it, it\'s not just something I\'m saying to try to comfort him or anything...  Am I being selfish throughout all of this?  Is it selfish of me to want to be single, flirt around, maybe date other guys, etc. during the rest of high school, some of college, whatever it is until I feel \'grown up\'?  I\'m trying to look out for both of us here, b/c I think it\'d be better for both of us if we waited for awhile... My parents and family always seem a little... bothered... by the fact that he\'s 21, and his parents don\'t like me b/c I\'m 17.  I\'m still in school, I don\'t have a job or car, and overall I\'m fairly immature and unprepared for commitment.  Ken, on the other hand, has job, car, etc. and is no longer living in the hell that is high school.  He\'s just got... experience - with life in general - that I don\'t have yet.  And until I reach a similar point myself I\'ll always feel slightly inferior to him, which I don\'t think is a very good quality to have in a relationship.  The thing is, I don\'t think he\'ll understand or agree with any of the stuff that I just said...  And I want this whole thing to be mutual, or as close as possible.  I want him to realize why I\'m doing this and how it\'s better for both of us...  Is this too wishful of me?  Am I asking for something I have no chance of getting?  And then there\'s a few other factors in this...  like us hanging out as friends, and all the stuff that I\'ve gotten from him.  As far as being friends goes... that\'s what I want, but I know that it\'s not going to be possible right away.  We can\'t break up on Friday and expect to hang out as friends that weekend...  That\'s kind of what happened last time...  I hung out with him WAY too soon, and the whole time we were hanging out I was just flooded with emotions... It was way too overwhelming, and I couldn\'t hold back my feelings.  I need to be more careful this time... Give myself a chance to get over the attraction and emotional attachment part... Which is going to be hard, and take quite some time, but I think that I can do it.  Maybe...  And as far as all the stuff I\'ve gotten from him... I\'m going to tell him to keep the ring in a safe place and save it for later...  I\'m giving him back the bracelet... and I\'ll let him take whatever else he wants, he can make that decision.  I\'ll probably end up throwing whatever he doesn\'t take in a box and shoving it in the closet for awhile...  too many memories and emotions at once are a bad thing for me...    So yeah...  I want to bring this all up soon, or I\'ll lose the.. courage? that I have now...  Tonight maybe, at least bringing it up initially.  We don\'t have to make our final decision tonight, but I would like to talk about it...  AHHHHHHHHHHHHHHHHHHHHHHHHHHHHH.... I\'m gonna go crazy... '),(16942,3776616,'07,July,2004','2004-07-07 00:00:00','    Deviled eggs sound really good right now... I wish my eggs would cook faster...  I talked to Beth today.  Her sister may be psycho, but Katie\'s pretty cool sometimes :)  Muahahahahaha...  Anyway...  Hopefully I get to see Ken tonight... He called me around 9:30 last night and was all like \"Will you be terribly disappointed if I don\'t come over tonight?\" then he listed his reasons why, and told me he\'d come over right after work tomorrow.  Of course I was just like \"No, that\'s fine\" and we said good-night and that was it.  I mean honestly, what was I supposed to say.  \"No, that\'s not fine, I want you to come over right now.\"  Dur...  but yeah, I\'m gonna go check my eggs..  See if I can get rid of this damn craving... '),(16943,3776616,'07,July,2004','2004-07-07 00:00:00','    I know this is a bit random,  but w/e...  I was thinking about what Mande said the other day about how it felt good to lie to her mom, and at the time it didn\'t really make a whole lot of sense to me, simply b/c I\'m used to it, and it just sort of feels normal to me.  But when it all comes down to it, that\'s part of the reason that I drink and what not - to defy my parents.  Othat than the fact that those things are fun, I also get the satisfaction of \'ha, you guys don\'t know what I\'m doing, but you wouldn\'t like it if you did\'  And maybe they do know about it, I\'m sure they do to some extent, but still...  I don\'t even think about whether or not they believe my story anymore.  That isn\'t really a big issue, the main issue is that I get to do the thing that I want to do, right?  And the weirdest thing about all of this is that eventually it gets addictive.  It gets easy to lie about little things, so you start lying about bigger and bigger things.  I think I\'ve even lied about where I was going before when I knew that they wouldn\'t care if I told them the truth.  It was completely pointless, but it was fun nonetheless.  Does anybody else get what I\'m trying to say here??  I just reread it and it doens\'t make a whole lot of sense to me, but oh well... '),(16944,3776616,'06,July,2004','2004-07-06 00:00:00','    Not really much to do...  I woke up at 10 this morning, did all the dishes (the huge pile in the sink and the stack on the counter), folded all the laundry, started a new load, put the newspapers in a bag and took them to the garage, and straightened up the living room.  It\'s past 12 and Kait isn\'t even up yet...  But of course, dad will come home and what I did won\'t be even close to enough for him, then he\'ll yell at me for something stupid, and I\'ll probably end up getting even more grounded than I am now, which is just fucking stupid as hell...  GRRRRRRRRRRRRR....  Idk...  Lainie still hasn\'t called me, even though she\'s been telling me that she will for a few days now.  And Idk when she works, so idk when I can call her...  Sam\'s pissed at me again b/c I told him all the stuff I wanted to tell him at the carnival, and he thinks we\'re all overreacting about the Hayley thing and it\'s not true at all, and...  wtf???  I\'m glad I ended up doing it in IM instead of on the phone or in person, b/c I started sobbing when he replied to every thing I said.  He didn\'t really answer any of my questions, and the ones he did were the answers that I was expecting...  I just...  It\'s not fair, and I know that\'s a stupid line to say, b/c everyone knows that life isn\'t fair, but still...  I guess as long as I have something to keep myself occupied I\'ll be fine.  Play practice has been fun so far, although my parents haven\'t been all that supportive...  On the way home from play prac. last night my mom asked me what my lines were, and I showed her.  \"Is that it?\" then she laughed.  Then she asked me if I was on stage any other time, and I told her those were my only lines, and she just gave me this look... idk really, it just pissed me off so much.  Like, what does she expect?  I got a part in a play that I didn\'t even try out for, and all she can do is laugh b/c it\'s a small part instead of some fucking lead role.  She probably won\'t even fucking want to come to watch the play...  My dad already can\'t b/c he\'s out of town... I don\'t care really, I\'d rather that they weren\'t there anyway.  I\'d prefer that my friends come, and my aunt and stuff.  w/e... '),(16945,3776616,'06,July,2004','2004-07-06 00:00:00','    The 4th was awesome!!! :)  I met Beth at the Red Rooster around 12:30 and we walked over to Sam\'s house to watch the parade.  Then I went home for a little while, and Lainie picked me up around 3:30 and we went over to Sari\'s.  The little Swanson was there, and Marie and Nick.  So we walked to Kwik-Trip to get stuff to eat/drink, then back to Sari\'s, and decided to go to the carnival.  Lainie and I drove, but everyone else walked... Oh well, I can\'t help it that I\'m lazy...  Then the two of us got sno-cones and sat on the swings having a forced convo w/ some stupid little girl named Rachel...  Argh...  Then we all sat on the playground waiting for Beth, Heathy, and Sam to get there.  Hung out for awhile... I tried to talk to Sam, but I couldn\'t get anything out... Grrr... Idk know when I\'m going to do it...  This is so not fair...  But yeah, then we went over to Beth\'s for a little bit, then back to the carnival where we met up w/ Robert, Crazy Katie, and AJ.  Robert and Beth went on the Tumbleweed together, and so did Mande and I.  That made me kinda sick...  but yeah, I got a cheeseburger then we all sat on the wet blankets and watched the fireworks.  Then me, Lainie, Beth, Marie, Nick, and Sarah went to Mande\'s and her mom lit off fireworks.  Then at 11 her mom kicked us out, and everyone except me, Mande, and Beth went home. We went back to Beth\'s quick to check in w/ her mom and stuff, then we went over to Robert\'s.  They were just starting to light off fireworks when we got there, then around 11:30 ish Robert had to take Crazy Katie home and left us outside w/ AJ (Robert\'s sis and friends were inside drinking and cooking bacon...).  Then AJ left, so we sat in the driveway playing w/ fire until Robert got back, and then everyone else left.  Robert shot some potatos out of his potato gun, then tried to play golf w/one, kept missing, so he just started beating it w/ the club, the he got a croquet mallet and started pounding it w/ that...  he\'s crazy sometimes... But that\'s ok, b/c he\'s Robert :)  Then we went inside and pulled out the beers that his sis had left for him, plus the ones that were left from Fri. night and the other stuff that was left from Fri. night.  I hate the taste of beer w/ a fiery passion, and I wasn\'t going to drink any (I called the only Triple Black that was left, and I was going to just drink the rest ofthe Blue UV).  But then Robert decided to make a beer bong, and he told me you can\'t really taste it as much that way, so I figured I\'d give it a shot (I\'d never used a beer bong up until then...)  So he got his old squirt gun tube, a milk jug, and rubber bands/duct tape and put his little thing together.  Beth went first, and her first time she got a few swallows then spit stuff out and didn\'t plug the hose, so there was a little bit of beer sprayed around the kitchen, but we got it cleaned up.  Anyway, then she went again and got it, so then I went, and then Robert.  It was actually pretty cool, and it does taste A LOT better that way.  Then I drank my Smirnoff and some of the blue UV.  Mande tried a little of both of those too.  Then Beth had another beer, and by then it was almost 1:30, so we needed to leave.  Beth was drunk (3 beers, a drink or two of triple black), Robert (3 beers, drink of two of triple black) and I (2 beers, triple black, few swallows of Blue UV) weren\'t really even buzzed, although I did just feel nice and relaxed.  Poor/lucky Beth... She\'s such a lightweight, and at times it\'s a blessing, but other times it\'s a curse...  so I walked her to her driveway then Mande dropped me off at my house.  I stayed up until 4 when Ken finally got there, then I fell asleep.  I woke up at 10:45 yesterday when Beth called me.  Did random stuff until Ken finally woke up at 2.  Yesterday was supposed to be a \'Brit & Ken\' day, but that didn\'t really happen.  After he got up we went to Autozone to get stuff for his car (he finally got his Mustang back), then we went to his house to get other stuff for his car, then we stopped at Kwik-Trip so he could show Jason... (guys and their cars...)  Finally we got back to my house and it was 4:30 ish, and I was in semi-sick, semi-depressed, semi-pissed off mood, but I agreed to go out back so we could have a water fight.  Then I came in and changed, sat on the porch w/ him until 5:45 when he left to go work on his car, then went back in my room until Mande picked me up for play practice.  That was fun even though I was in a bad mood, but then I got him and got in a fight w/ my dad about the comp, so I got grounded.  In an attemt to keep myself from totally trashing my room I went running.  Walking most of the time, running/jogging every time I felt emotion.  I just kept running until I thought I\'d pass out, and by then my thoughts had cleared again.  That worked pretty well, and I didn\'t get home until 9:30.  No one even said anything to me...  Anyway, then I ate and Ken came back over and we sat around and now it\'s today.  So yeah...  Idk if I\'m still grounded or to what extent I am, so Idk what I can and can\'t do...  I REALLY want to have another shindig at Robert\'s.  That was fun, and it\'d be a lot more fun if I didn\'t have to worry about going home :) '),(16946,3776616,'04,July,2004','2004-07-04 00:00:00','    Oiy...  So much has been going on in the past few days, shitty and good.  I don\'t really have much time because we\'re leaving for the parade shortly-ish, and I really can\'t go into detail on everything right now anyway.    Friday night...  Everything got all messed up, stuff almost didn\'t happen, Lainie couldn\'t come, but in the end it was me, Beth, Crazy Katie, Robert, Sam, and Heathy.  Heathy and I scared Robert when we were waiting in his car after he got off work...  We originally had everyone in the car w/ arms and legs hanging out so it looked like we had a giant orgy, but then Beth, Sam, and Katie left so we wouldn\'t miss the pizza guy when he came.  But yeah... the biggest part of the night was the Spin-the-Bottle game which everyone except Beth played.  At first it was: guys shake hands, otherwise same sex kiss on the cheek, opposite sex kiss on the lips.  Then after awhile it was changed to: same sex on the lips, opposite sex w/ tongues.  We started counting when two people were kissing to see how long they would go for -- Crazy Katie and Robert seemed like they\'d never end, and Sam and Heathy went for awhile a few times.  Robert and I went for like, 10/15 seconds before Katie jumped up and pulled us apart, and I\'m not really sure about me and Sam.  But then shit went downhill and lots of people got quite emotional, and... Yeah.  Then Robert and Sam went back to Robert\'s house around 3ish, and the rest of us sat up talking a little longer then we went to bed at about 4.  Poor Heathy... she had to get up at 6...  But yeah... Argh, I don\'t want to post any of the other shitin here, b/c I already vented to Heathy about it at Beth\'s, I\'m going to talk to Lainie today, and I REALLY don\'t need the entire world knowing all the details.  B/c some of the details are things that I want to keep secret...  Plus there\'s stuff that I want to talk to Sam about, but I can\'t tell him yet, and there\'s stuff I REALLY need to tell Ken and make him see and accept...  Not fair not fair not fair...  Yesterday...  it started off really shitty b/c I woke up and George thought it was cool to stop by in the middle of the night, and then I had super bad cramps.  I really wanted to talk to Lainie, and Thurs. I thought she told meshe worked 12-5 on Sat (which I must\'ve heard worng...), so Beth and I walked down to the Red Rooster.  Now, walking from Beth\'s to there w/ cramps, a headache, being tired as hell, and feeling 150% shitty makes it a damn long walk, and I wanted to just collapse halfway there...  Then we got there (1:30ish), and Lainie\'s van wasn\'t there.  So right then and there I think I lost half of my grip on everything, then I went inside and Ghandi was working and he said Lainie didn\'t come in until 4, and I started to freak out and he looked at me weird and asked me if everythign was okay, and that\'s when I just totally lost it.  I ran to the bathroom and sobbed for 5 mins, called my dad and said I wanted him to pick me up, then I sat by the b-room crying until I saw his car pull up...  Went home and sat in my room for god knows how long, and took a nap...  But then around 8:30 ish I called Robert (he and Crazy Katie were supposed to go up to R&B in Madison, and I wanted to make sure he knew it was postponed).  Anyway, he asked what I was doing and I said nothing, so he asked if I wanted to come over. So he picked me up and me, him, Katie, AJ, and Richie went to his house and lit off fireworks in his backyard.  Robert was putting stuff in potatos and blowing them up, then throwing little potato chunks at me... Grrrr...  Then at 10 Jenny came over, then everyone except Richie went to Kayla Shaapveld\'s grad party.  It was a little akward b/c I didn\'t know/get along with a lot of the people there (Scott Boie...), but it was fun anyway.  Then Katie and I left and went to get Dunkin Donuts, but there was a bunch of creepy looking people stanging outside the door so we just left and came back here.  Talked for awhile, then Ken called and I talked to him, then I came back inside, and blah.  Katie left at like, 8:30 this morning.  She wants to cancel her backpacking trip, and her mom is going to freak out...  Idk...  \"Are you going to the parade?\" ... so help me god, if either of my parents asks me this question one more time I\'ll shoot myself.  YES, I AM!  Just like I said the last 6 times you asked...    I think I\'m going to call Beth now, and Lainie.  Then get in the shower.  I\'ll see everyone at the parade or tonight at the fireworks or whenever I see you... '),(16947,3776616,'01,July,2004','2004-07-01 00:00:00','    On the phone w/ Lainie...  I\'m SOOO confused about tomorrow\'s plans.  As far as I knew before, me, Beth, Crazy Katie, and Robert were going to party at Beth\'s, and Lainie and Mande were going to hang out...  Then Mande got online a little bit ago and asked me is we were still planning on getting Lainie drunk tomorrow night (which, by the way, people seem to have misunderstood slightly...).  So I said that I thought she was hanging out w/ you, and that I wasn\'t sure how much fun she\'d have w/ us anyway b/c she might feel weird sitting and watching us, and that we hadn\'t really updated her on times or anything most of the week b/c we assumed she was hanging out w/ you.  And then... Argh, idk...  everything\'s all crazy-ish, and Lainie was talking too fast for me to comprehend at 11:30, and I don\'t think that  ANYONE  knows for sure what\'s going on tomorrow night, b/c everyone had what they thought was the plan in their head, but everyone had a diff. plan.  GRRRRRRRRRRRRRRRRRRRRRR... So as of right now no one really has any clue what\'s going on, and Lainie doesn\'t even know if she can get out of her house anyway...  So yeah, I\'ll update again on Sat, b/c by then all of this will be over and everyone will know what the plans were :) '),(16948,3776616,'01,July,2004','2004-07-01 00:00:00','    My tummy hurts and my muscles ache, I don\'t really know why.  Plus I\'ve got cramps and I shouldn\'t, I\'m thoroughly exhausted and I just got up 3 hours ago, and... BLAH.  But on the bright side of everything, I alreaady got my chores done for the day, I get to go get my contacts at 4, I got to talk to Ken last night for ore than 5 mins, he\'s going to try to get off work before 12 tonight so he can come and see me (his Mustang is supposed to be dropped off at his work today!), and... and...  oh yeah, tomorrow I\'m going to Beth\'s!  So YAY!!  The rest of this week should be pretty cool.  *does a little dance*  Hopefully everything goes well...  YAY!!!!  I just got off the phone w/ Church!  He told me to show up for play prac. on Monday night!  I get to be the judge, which is just a small part in the beginning, but really that\'s all I think I could handle...  Too much stress, especially since the play is in... 3 weeks?  But yeah, it\'s all good.  So see, today\'s getting better by the minute.  I still have to get ahold of Heather, b/c she wants me to babysit her rat for her while she\'s on vacay.  I missed her call this morning, then I left her a message, and now I\'m waiting for her to call me back. Lalalalalalalalalala...  *random jig*  Now I\'m just bored... '),(16949,3776616,'15,August,2004','2004-08-15 00:00:00','    So I just typed everything and it deleted... and I don\'t feel like typing it all again, so here\'s the abbreviated version...  yesterday I went school shopping, got some bras and underwear and notebooks and whatnot.  Then I went over to Ken\'s and talked to his parents then we went down to his part of the house and he showed me his kitchen and bathroom and bedroom and living room and shit.  That was cool, then i got home at 12:30 and the parents didn\'t say anything so now I\'m going to start coming home late every night.  YAY.  anyway, nothing\'s going on today, mom and kait went to fond du lac and dad and i are staying home.  fun fun '),(16950,3776616,'13,August,2004','2004-08-13 00:00:00','    Went to Jason\'s w/ Ken yesterday...  We watched Scary Movie 3.  These damn painkillers make me tired as hell, so I slept some...  I\'m not taking any today, b/c of the party and b/c I don\'t want to fall asleep at 7pm...  But anyway... not much is going on today.  Lainie\'s going to drop off Nick sometime, and I have to call K-dogg sometime after 5ish.  After Ken gets off work he\'s coming over here, and we\'re going to Robert\'s around 10ish, when he gets off work.   So yeah, that\'s my day.  fun fun... '),(16951,3776616,'12,August,2004','2004-08-12 00:00:00','    Got my wisdom teeth pulled yesterday... that was \'fun.\'  Getting knocked out was pretty cool though, b/c I got all numb and felt drunk-ish.  That was cool.  And then after I\'d sat in the little recovery room for awhile and they told me I could leave they tried to help me stand up and walk.  But it felt like walking drunk, so I didn\'t have much of a problem, and they told me I was pretty good.  I almost said \"I\'m used to walking drunk, so this is pretty easy\" but I caught myself.  I don\'t know what would have happened if I did... not much probably.  Anyway, then I came home and slept a lot, then Ken came over, we played Monopoly (I lost), then we went to bed.  So yeah.  Yesterday was fun.  I was feeling bored around 12ish and thinking that I hadn\'t seen Heathy in awhile, so I texted her to see if she wanted to do something.  So we went to Subway to see Scary, then we went over to Beth\'s to see her and Sam.  Then Sam left, and Heathy went to work, so Beth and I walked down to the Red Rooster.  We got there at like, 4:30, so Pookie wasn\'t there yet, but we got ice cream and talked to Becca.  Then Lainie got there, and we talked to her for awhile.  Got chinese.  Then Beth and I left around 7 something and walked to Andy\'s house, but he wasn\'t there, so we went to Robert\'s.  His car wasn\'t there, so we called him and he said he and Andy were on their way back.  So then Beth and I helped Robert and Andy clean up from the party monday night (I helped clean up after a party I wasn\'t even at... oh well).   Then Jenny came over and we played cards and ate watermelon.  Then we met AJ at Taco Bell.  Came back to Robert\'s for awhile.  Watched some movies.  Jason Perdelwitz came over... We all played Monopoly, then Jason and Jenny both left.  Then me, Beth, Robert, Andy, and AJ got the bottles out of the cornfield, then put them in Robert\'s trunk and burned the boxes.  Robert put a lot of gasoline on the fire... Then  AJ cleaned out his car and we burned a rug and shit.  That was cool.  Went inside and watched another movie, then Beth and I tried to sneak back into her house and the door was locked... So we went back to Robert\'s and spent the night there, and when we got back there was a note for Beth saying she was grounded for not coming home... So that really sucks, b/c she\'ll probably be grounded until school starts.  She can still prob. sneak out for Robert\'s parties, but I don\'t know what she\'s going to do for Jake and Millertime\'s party...  Anyway... don\'t know what I\'m doing today. Ken works until 12, then he\'s going to Tabitha\'s funeral, then he\'s going to come over.  We might go over to Jason and Sami\'s.  And I need to drop off Pookie\'s sunglasses at her house sometime before she leaves.  I need to call Robert too, to see if Ken can come to the party.  He could get Jason and Sami to come too, if we want it to be a bigger party.  God, I\'m so excited that I get to get drunk again tomorrow night... I haven\'t gotten drunk in two weeks...  I\'m just not going to take my painkillers tomorrow.  Maybe one when I wake up, but I won\'t take anything after noon.  Otherwise I\'ll get sick.  And I can deal with the pain... once I get drunk it won\'t matter anyway...  That\'s it for now I guess.  Call me if you can Bethy. '),(16952,3776616,'10,August,2004','2004-08-10 00:00:00','    I\'M FINALLY HOME!!!!!  Sunday night our flight was overbooked, so they were looking for volunteers and my dad got us on the list to take a bump.  They ended up needing our seats, so we got $300 each in airline credit, plus our hotel and shit.  The next flight they could get us on left at 8:10 in the morning, but then we would\'ve had to wait at the airport until mom could pick us up at 6.  So we took the 11:30 instead, had an hour layover in Detroit, and finally got to Milwaukee at 3:30.  Then we had to sit in the waiting area until 6, and by that point dad was pissed at mom for being late, then we all had to sit in the car together until we got to Janesville, THEN we had to go out for dinner... That sucked, b/c mom talked on the phone pretty much the whole time, including a long conversation w/ her boyfriend in fond du lac, all while my dad was sitting right next to her.  That was great... We\'d been home two hours and they were already on the verge of killing each other.  But finally around 9 we got home, and I just took my stuff in my room and shut the door.  Said hi to my kitties and puppies.  Called Ken and he came over.  He watched me unpack and I showed him stuff, then he showed me pics from PA, then we talked and blah until he left around 1, then I went to bed.  Woke up at 8 this morning.  Did some random stuff for awhile, then got on the comp...  I still have to call everyone... Bethy, Lainie, Sam I\'ll call you for sure today.  Robert...  I\'ll talk to sometime this week.  I need to call Nicole b/c I haven\'t talked to her in awhile, and I have postcards for Mande and Heathy that I need to give you guys...  My dad\'s an asshole...  He never renewed our sub. to the milton courier... which would explain why we never got the one w/ the play pic in it, which I really wanted.  And I like to read it b/c I know people in Milton and shit...  Another thing - does anyone remember Tabitha Braden?  I haven\'t really talked to her much since middle school, and then last night I found out from Ken that she died in her sleep on sat. night.  That\'s depressing...  Earlier this week I was thinking about going out w/ Ken to see her sometime soonish, but yeah...  Anyway...  I get my wisdom teeth pulled tomorrow... That\'ll be great.  Just great.  Simply fabulous.  So Weds and Thurs are gonna suck, but I\'m still going to try to find time to hang w/ Lainie and everyone.  Friday night I\'m partying at Robert\'s.  I\'m feeling party-deprived.   I really wanted to go to Robert\'s last night, too bad my dad\'s a jackass...  So yeah.  That\'s about it for now.  I think I\'m gonna go start calling people and whatnot.  Later! '),(16953,3776616,'07,August,2004','2004-08-07 00:00:00','    Forget my post from yesterday, I just go caught at a really bad time and had a wonderful little mood swing.  I\'m sorry I made you cry Bethy I didn\'t mean that and you know I love you (platonically, of course).    Anyway...  Thursday dad had to work and didn\'t get back till almost 9, so I didn\'t really do much.  I didn\'t feel like going to the pool alone and being attacked by the people down there... God, I hate feeling like the minority.  This sucks...  Then I texted with Sam and Derek for awhile.  That was cool, b/c I haven\'t really talked to Derek in a loooong time, and I haven\'t talked to Sam since he left for Fla.  Lainie didn\'t respond to my texts, and Heathy couldn\'t text me back until she got off work.  Then my dad got home and we ate dinner, then I called Sam and talked to him for about 15 mins (I was only supposed to be on for about 10, but oh well).  Then after midnight Ken finally texted me back telling me he\'d read the letter.  I was SOOOOOOOOOO freaking happy, and that lifted a lot of the tension off.  Part of me was afraid that I\'d screwed too much stuff up and that he wouldn\'t want to try to make everythig work out and start over.  I know that I screwed up really bad and hurt him a lot, and I know that I don\'t deserve to have him back, so I considerr myself extremely lucky to be in love w/ someone who loves me back as much as he does.  Yesterday we went with dad while he did some of his work.  They were stores on the east coast, and we were going to go to the beach while we were in Melbourne,  but it was raining all day,  so that didn\'t really happen...   We ate lunch at Friendly\'s.  I\'ve never eaten at one of those before...  it was pretty good though.  I had chicken strips, and I can\'t figure out what my obcession has been w/ chicken strips lately... Oh well, they\'re good, so I\'m just going to keep eating them.  We didn\'t really do much else yesterday.  I was finally able to get postcards, and I\'m going to send them out today, so you\'ll get them after I get home, but oh well.  Mande, Lainie, Bethy, and Sam, you\'ll get yours in the mail.  Heathy, Ken, and Robert... I\'ll just bring yours.   I wanted to buy everyone little souvenirs, but I only brought $20 with me, which isn\'t enough to buy much of anything in this place...  But anyway... today we\'re probably going back to Universal Studios and/or Islands of Adventure for awhile.  I\'m going to talk my father into buying me this really cool necklace I saw, and maybe talk him into giving me money to buy stuff for everyone, but I don\'t know how well that will turn out.  We\'re eating lunch at Emeril\'s restaurant, b/c my sister is a big Emeril fan...  I don\'t get to go to Epcot, but oh well I guess...  Tomorrow we\'re probably doing the same general thing as we do today.  Our flight doesn\'t leave until 6:15, and we have to check out of the condo by 10am, so we have most of the day.  Then we get home around 8:15 WI time...  Hopefully mom is on time picking us up, and I can be at my house by 9:30, no later than 10. *crosses fingers*  She\'ll probably forget about us, but oh well...  But yeah, it\'s only 9am and I\'m already up, showered, dressed, and ready to go, except for putting in my contacts.  Kait\'s fast asleep and dad\'s supposed to be back by about 10am (he had two stores he had to work at today).  Then we can leave...  So I guess I\'ll go try to wake her up and get her going so we can leave shorty after dad gets back.  I\'ll see/talk to everyone on Monday!!! '),(16954,3776616,'05,August,2004','2004-08-05 00:00:00','    Well, I\'m officially in a super-shitty mood...  I was just in a fairly bad mood... bored, missing people back home, the Ken stuff...  and then I read Beth\'s blog and I freaked out and started crying...  All the stuff about the party at Robert\'s the other night, and the party that\'s going to be tonight, and all the stuff about Sam and... I don\'t know...  Part of me just doesn\'t want to go home.  I want to leave and never come back... I don\'t know where I\'d go, I just want to go, really far away, away from everyone and everything...  And right now I\'m really desperate to party...  And I\'m missing all the stuff this weekend, and no one is going to be in the mood when I get back... Tues night I can\'t b/c I get my wisdom teeth pulled on weds so I can\'t drink past midnight... Then I won\'t really be in the mood weds night... so the soonest will be Thurs or Fri... I don\'t know if I can wait that long... and I know I sound like some freak crazed alcoholic or something stupid like that, but frankly I don\'t care what I sound like.  I\'m in a REALLY bad mood, feeling a bit of party withdrawl, which is worsened by all the talk of partying, and I don\'t know how much longer I can stand this... GRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRr  sorry... forget everything.  fuck it all '),(16955,3776616,'02,August,2004','2004-08-02 00:00:00','    Vacations are fun, but it kinda sucks being cut off from everyone else... I tried to call Beth last night but she was busy so I guess I won\'t get to talk to her until I get home... Plus I\'m missing the party tonight, and I\'m not sure what she decided on what will and won\'t be done tonight (as far as alcohol, happy smoke, other, etc.). I\'m not depressed about missing the party that much, even though it\'s different than our other ones b/c it\'s at Beth\'s and they don\'t have to hide out in the basement, I\'m just depressed if they do the shrooms w/o me, b/c I wanted to be w/ Bethy and Robert and them when we all did it together for the first time... Oh well, I guess it doesn\'t matter all that much. If they do it, they do it. If they don\'t, they don\'t, and we can do it as a group sometime later... Other things... Lainie has only responded to like... half the texts I\'ve sent her. Heathy hasn\'t replied at all, and I don\'t have her address, so I can\'t send her a postcard. The only person that I\'ve really had contact with is Ken, and there\'s so much other stuff connected with that...  oiy, forget all of that.  Anyway, I\'ve been pretty much having fun here.  Monday night we went out to Boston Lobster Feast.  All you can eat lobster, crab legs, shrimp, etc.  I ate too much... only 2 lobsters, somecrab legs, shrimp scampi, fried shrimp, cocktail shrimp, london broil, crab cakes...  Then I ate cake for dessert... Tuesday we went to Universal Studios.  I lost my sunglasses... GRRRRRRR  The MIB ride was pretty cool.  You get to shoot aliens! :)  Yesterday we went to Islands of Adventure.  That was better, just b/c there were more rides.  The water rides were awesome, especially when we first got there b/c there was almost no wait at 10 am,  even though it was already hot as hell.  It was good for us though...  Anyway... I got a cool necklace.  It\'s a silver dragon\'s claw w/ a red marble in it.  There\'s another one that I want to talk my dad into buying for me when we go back, but I\'m not sure if that will work out...  w/e...  Today... dad has to work until around 6 ish, so I don\'t really know what I\'m  going to do. It\'s nearly 11 and I haven\'t showered yet, plus Kait is still sleeping.  The crazy maintenance guys were just here, and I didn\'t understand a word of what they were talking about b/c it was either heavily accented irrecognizable english or rapid spanish... oh well, hope it wasn\'t too important.  Anyway, I\'m going to sign off and shower, then get dressed and ready and make some lunch.  We\'ll probably go somewhere when dad gets home...  I\'m going to send postcards out tonight, hopefully.  I\'ll get home before they do, but oh well, they\'re still fun to send :)  Pookie -- I love you!!! Bethy -- I love you too! (platonic).  Have fun tonight! Ken --  I miss you!  Love you, and can\'t wait until I get home. Heathy -- why aren\'t you responding ot my texts??  I need your address! Mande -- Have \'fun\' watching the brats!  See everyone later... '),(16956,3776616,'02,August,2004','2004-08-02 00:00:00','    Having fun in Florida...  I haven\'t really done much.  Kait and I met this girl named Erica on the plane on the way down.  She was like... 12?  Idk, she was cool though, and gave us someone to talk to though.  Then we got picked up (stupid rules about minors flyin alone... I hate having somone watch over me all the time.  I\'m 17, I don\'t need a babysitter to get on and off an airplane)  We went to Planet Hollywood for lunch... came back to the timeshare, sat around, then went to Super Wal-mart for groceries and whatnot.  I got elastic cord so I could finish my anklet.  It\'s really cute :)  Today Kait and I are just lounging around while dad\'s at work.  Eating, watching TV... meh.  Anyway, I\'ll update my blog as much as possible while I\'m here, two or three more times this week.   I love you all!!  I miss you Bethy and Pookie and Ken!  And Bethy, you HAVE to call me after the party and tell me how everything goes!  You should call me during the party too, if possible.  I won\'t be able to talk long,but...  Anyway, gotta log off, later!!! '),(16957,3569188,'11,June,2004','2004-06-11 00:00:00','   Well, Erika and I have just recently gotten done with the most retarded game ever...not so much retarded as gay..but whatever. Anyhoo, I think that I never want to play it again, as of right now, but- knowing me and my indecisive nature, I will be back and playing it next week :-P.  I am really tired. Not much of an other update, and if there was I couldn\'t give it to you because it would take way to fricken long. Whatever..going to bed.  -Alyssa-     urlLink      I am truly passionate.   urlLink  Find your soul type  at  urlLink kelly.moranweb.com .       That is so me ^. Nearly every detail!       urlLink        urlLink What\'s Your Love Style?            urlLink Find your inner random object!  by  urlLink Emily       I\'m...a squirrel. :-I. A squirrel???      urlLink >    Take the  urlLink >\'Which Nintendo Character are you most like?\'  quiz by  !       I\'m almost a perfect gentleman..how flattering ;)             Germany  -  Despite a controversial recent history, it has had a tough and powerful history.  A modern-day technological and cultural beacon, it is still target to stereotypes and antiquited thoughts.     Positives:   Technologically Advanced.  Culturally Admired.  Global Power.     Negatives:   Target of Historical Fervor.  Constant Struggle.  Funny-Looking Ethnic Clothing.                 The Goddess Death: Dark, morbid, appropriately represented by the color black and the element water.  Death is reserved, intellectual, introspective.  Rarely does she do anything requiring a lot of energy or display of emotion -- but when she does, anyone within a few planes would do well to duck and cover.    urlLink   Which Goddess Owns You?     brought to you by  urlLink Quizilla             If I were a Neopet... I\'d be a  Kiko !  The amphibious Kiko is just as happy on land as it is in water.  They dislike very dry places and need to be near water when young.           Which Neopet are you? Visit  http://neopets.retrograde.org/quiz/  to find out!       Heart of Crystal    urlLink   What is Your Heart REALLY Made of?     brought to you by  urlLink Quizilla        Nerdslut    urlLink   What\'s your sexual appeal?     brought to you by  urlLink Quizilla       urlLink        urlLink find your poetry style  this quiz was made by  urlLink mamaslyth            footsie - you like to goof around and laugh with the people you care about.  '),(16958,3569188,'08,June,2004','2004-06-08 00:00:00','   You are Revelation.    urlLink   Which book of the Bible are you?     brought to you by  urlLink Quizilla     No time to write today. Have to clean house before dad gets home :-\\!!!  '),(16959,3569188,'07,June,2004','2004-06-07 00:00:00','   Alirght, so today sucked. Biug deal. I have never really gotten excited about Birthdays. I am really disappointed in Erika. I figure I am just not gonna bring it up, so she will eventually realized it\'s passed, and feel awful. I don\'t know, aren\'t I mean? I don\'t care. So anyway, tonight I am going to completely organize my closet and room and get rid of everything I don\'t need. Yikes..I am such a packrat that this is gonna be hard. I\'ll get it done somebow. Once that is accomplished, I plan on designing a new blog or site layout. I don\'t know which one I want to do right now, but most likely it\'ll be the site. I am going to do a black/red/pink theme- you know, eventual color fading and stuff. It should be cool.   I need to start working on the violin. I had so many summer goals, I just need to start breaking up the time of summer and start planning everything out. I am going to bed early tonight so that I can get up early and go running. It\'s gonna be a good summer if I can keep myself busy. I admit- I am still hung up on Corey. I feel bad about Nate. I mean, I used him in a way to try to distrat myself. But I don\'t know..it\'s just that I have this wierd spot for Corey in my heart as a friend and that seems to be affecting the way I still feel about him. One of the most dragged out crushes ever. I\'ll get over it. Just need to keep distracted. Grr, I am so angry at myself for becoming attatched at all. I mean, it was one year and I\'ve already let 2 people in. THat\'s two more than I would prefer. I don\'t like it..I have good reason too. What\'s the point of letting people get close to you if you are only gonna get hurt. That\'s why it\'s casual dating and superficial teenage relationships for me next year, thanks. Whatever. I need to have more of an open heart. I mean, the problem I have is that it\'s so big and gullible. I hope nobody ever finds out how to work me. If they can get that close :-P. Not gonna happen.  So anyway, I need to drink more water, I am totally dehydrated. Still mad at Erika a little. I don\'t know why I get so annoyed with this, but I don\'t know. Wish me luck with running in the morning...it\'ll be torture. ;-)  '),(16960,3569188,'06,June,2004','2004-06-06 00:00:00','   I just found this thing that makes your text glow around the edges. Just testing it out:      I found one for images too, but I closed out of the page before copying the code. I am an idiot ;-).              Still playing with the glow filter today. Wanna play?  '),(16961,3569188,'06,June,2004','2004-06-06 00:00:00','   Alright..so it\'s not \'neglection\' parse, but it\'ll do justice to describe how I am feeling. Erika totally forgot my Birthday! I mean, she went to the beach with Stori, and then she\'s going off to celebrate Stori\'s B-Day later. Grr to her. Anyhoo..  Corey remembered. He would though..he knows what it\'s like to not have anyone remember. So far today, I have pretty much wasted the day. I watched Charlie\'s Angels; Full Throttle (which is so completely gay), then I cleaned, and then that\'s pretty much it besides the internet. Total waste. I think I might clean out my closet later...fascinating experiences for a fascinating Birthday. Whatever. Well, write later maybe.  '),(16962,3569188,'06,June,2004','2004-06-06 00:00:00','   I am really tired. Uggh. I just transferred all of my entries from ebloggy.com to here. Not many, so it wasn\'t that bad. Tomorrow is monday..I am supposed to go running and start putting my nose to the grindstone (grindstone?or is it grind your nose to the stone? or to the limestone...???) well, whatever. Anyhoo...(been saying that word a lot lately) Tomorrow I think I might call Mary. Don\'t know. I still can\'t believe that it\'s summer. It\'s not hot enough yet. Guess I just haven\'t gotten into my routine..well tomorrow\'s my chance.   Tomorrow is my birthday. In about 8 mins I will be 14. Well, not technically until like, 4 in the afternoon, but poo on technicallity. Omg, I just said the word poo...I really am losing it. Whatever...(been saying that word a lot too..hmm..). I am gonna wait and see if anybody remembers my birthday tomorrow..I doubt it. Nobody really cares. Erika might remember, but I don\'t care if she does. To be honest, I kind of want people to forget so I can blame them. That\'s so mean, but I guess it\'s just me. I don\'t know, I am trying to figure out who I  really  am, but what if I don\'t like that person? I most likely will, seeing how I surround myself with people like me..I don\'t know.   Corey\'s still on my mind..deciding whether or not it would be a good thing to see him much this summer. I don\'t know...I mean, I don\'t want to get any closer to him. I am already freaking out with the way things are. I mean, it\'s like I just break down around him and I don\'t hold anything in! Isn\'t there something completely messed up with that? I don\'t know, I just don\'t like people to know me...I feel limited and they form their own assumptions and I am no longer free to be just me, which is dually hard when I don\'t know who \'me\' is. Anyhoo..I don\'t know. It\'s like he has this wierd control over me. I can\'t put my finger on it right now..maybe later. But I mean, I dont\' really get mad at him, and if I do (which has happened like, twice) It\'s just because I want to be, and because I am really stupid. Grrr to you COrey Shotwell! I guess I will just have to deal with it. (Feel more sorry for him- he\'s got the end wehre he has to listen to me complain and whine :-P).   Well, it\'s 11:59 on the computer clock. Waiting...   12! Well, I am unofficially 14 years old. Omg..that\'s really wierd..anyway. Got to get to bed and stuff. Write later..most likely tomorrow.   -lyss-  '),(16963,3569188,'06,June,2004','2004-06-06 00:00:00','   So bored today. Went to my baptism class, came back and took quizzes. Erika and I went to the park. We climbed over the stupid fence that\'s like, 8 feet tall, and Erika was afraid to jump about 2ft from the trash can to the ground. She\'s not a very good partner in crime, but as Corey isn\'t here (and even if he was, he wouldn\'t do anything cause he\'s such a wimp about breaking rules) she\'s good enough. So I get inside and we decide it would be funny to put an orange cone on top of the shed...I don\'t know  why , but hey, come on- it\'s summer and we were super-bored. So it\'s up there and I am like, ok, that\'s waaay too noticable, so I try to step down, except Erika wasn\'t paying attention and forgot to sturdy the bin I was standing on. So I like, fall and this brick thingy lands on my leg and Erika starts laughing hysterically. So we are sitting there- cone on the roof, me being physically impared, and Erika making a ton of noise. So I am sitting there and going, \"SHHHHH! ERIKA, shut up!\" and then I was freaking out about the cone in case somebody saw it, and then Erika tries to hit it down with a mop and I am still stuck in the bricks, and then the crappy stuff is falling off the mop onto my face and we are freaking out. So eventually we fixed the situation, and we went back home.......to get tools. :-P   So we come home, get two phillips screw drivers, a monkey wrench, a hammer, and wire cutters. Instead of going in depth to all of our little \'adventures\', heres a list:   -unhinging a shed door to see what\'s inside (i got bored and only followed half way through)  -carving initials into shed  -erika carving initials of who she likes into tree (not so bad)  -me puncturing the school\'s hose (hey, it\'s our tax money! and besides, Erika told me...:-P)  -Carving my initials into a fence  -both erika and I slashing \'z\'s into the screens (zorro theme!)   Yikes! I feel like a bad-girl. I love it! Total adrenaline rush...didn\'t know I was that way...Anywhoo...   Here are my results from quizzes today:     You represent... hope.  You\'re quite a daydreamer and can be a hopeless romantic. You enjoy being creative and don\'t mind being alone at times. You have goals, and know what you want in life... even if they are a little far fetched.     urlLink   What feeling do you represent?     brought to you by  urlLink Quizilla           Your Superhero Persona  by  urlLink  couplandesque        Your Name           Superhero Name    Emo Kid      Super Power    Irresistable Sexuality      Enemy    Martha Stewart      Mode Of Transportation    Giant Hamster Named Skippy      Weapon    Attack Gerbils             Created with the ORIGINAL  urlLink  MemeGen  !               Your Superhero Persona  by  urlLink  couplandesque        Your Name           Superhero Name    The Armadillo      Super Power    Ability To Breathe Underwater      Enemy    The Disgruntled Mailman      Mode Of Transportation    Volkswagen Beetle      Weapon    Vinyl Records             Created with the ORIGINAL  urlLink  MemeGen  !               Who will give you an orgasm?  by leslie13      Name           Age           Virgin?           So, who will make you moan?    Orlando Bloom      How?    Saying \"Go, go orgasm!\"      Will it be good?    It&#39;s mediocre. -shrug-             Created with the ORIGINAL  urlLink  MemeGen  !               Who will give you an orgasm?  by leslie13      Name           Age           Virgin?           So, who will make you moan?    Yourself.      How?    Manually.      Will it be good?    It&#39;s mediocre. -shrug-             Created with the ORIGINAL  urlLink  MemeGen  !               What will your Funeral be like?  by rashock      Username           You will die by:    You died of natural causes. How sweet. Your funeral is an adverage open casket.      Death Date:    July 2, 2013      Number attending your funeral?    107      How much will you leave to friends and family?    $4,456,284             Created with the ORIGINAL  urlLink  MemeGen  !               What will your last words be?  by  urlLink  cum_on_bitch        Your LJ username           Your real name           Your sex           Your age           Your last words will be...    \"i hate you so much...\"               Created with the ORIGINAL  urlLink  MemeGen  !               Your Life: The Movie  by  urlLink  mintyduck        Who will play you:    Jack Black        Who will play your love interest:    Ashton Kutcher        Weeks you will stay in the box office:    13      Song that will play during your love scene:    Earth Wind and Fire - September        Song that will play during your death:    Frente - Air        Your name:                  Created with the ORIGINAL  urlLink  MemeGen  !            As if you were born into a world of tears, you always tend to look at the darker things in life. Inside you crave attention yet push away society, and you\'re a hopeless romantic. Drawn to things like the occult and mysteries, you spend your time daydreaming.     urlLink   What Type of Soul Do You Have ?     brought to you by  urlLink Quizilla         You\'re just the happy go-lucky type. You might have your pet peeves, but other than that, you\'re mainly calm. Blending in with your surroundings, you\'re the type of person who everyone likes. Usually it\'s you who cracks jokes at social gatherings - after all, laughter is the best medicine. Sometimes you pretend to be stupid, but in all actuality, you could be the next Einstein.     urlLink   What Type of Soul Do You Have ?     brought to you by  urlLink Quizilla         You\'re Exotic Dancer Barbie. You have some moves, and will do anything for a few bucks. Take it off girl, but keep it PG-13 please.     urlLink   If You Were A Barbie, Which Messed Up Version Would You Be?     brought to you by  urlLink Quizilla       YOU ARE MARRIED TO A WoODCHUCK!!!      urlLink what\'s YOUR deepest secret?     brought to you by  urlLink Quizilla         Your Heart is Red     urlLink   What Color is Your Heart?      brought to you by  urlLink Quizilla         Your soul is bound to the  Rose Petals : The Wronged.      \"\'ve come undone and all hopes of mending me are gone because the pain took my soul. Can\'t you see? The only one who can put me back together again is me.\"      The Rose Petals are associated with sorrow, reflection, and wisdom. They are governed by the goddess Persephone and their sign is The Teardrop, or Broken Love.     As a Rose Petal, you are always self-reflective and may be hard on yourself. You probably have been hurt in the past by other people and can sometimes distance yourself, as a result. You don\'t usually let other get too close to you, but you are very good at mending your spirits back together by yourself.     urlLink   What Rose Is Your Soul Bound To?     brought to you by  urlLink Quizilla         Nemesis     urlLink   ?? Which Of The Greek Gods Are You ??     brought to you by  urlLink Quizilla         You are Form 0,  Phoenix : The Eternal.      \"And The Phoenix\'s cycle had reached zenith, so he consumed himself in fire. He emerged from his own ashes, to be forever immortal.\"      Some examples of the Phoenix Form are Quetzalcoatl (Aztec), Shiva (Indian), and Ra-Atum (Egyptian).  The Phoenix is associated with the concept of life, the number 0, and the element of fire.  His sign is the eclipsed sun.     As a member of Form 0, you are a determined individual. You tend to keep your sense of optomism, even through tough times and have a positive outlook on most situations. You have a way of looking at going through life as a journey that you can constantly learn from. Phoenixes are the best friends to have because they cheer people up easily.     urlLink   Which Mythological Form Are You?     brought to you by  urlLink Quizilla         You\'re Most Like The Season Summer ...     Whoa.... Passionate eh ?? Typically you\'re a fiery, zesty dominant person. As the hottest season, you certainly ooze Sex appeal. You have confidence which draws people to you, and you have the makings of a good leader.  However sometimes your exterior is stronger then you are and so you scare people off before they can get close.     Well done... You\'re the most memorable of seasons :)     urlLink   ?? Which Season Are You ??     brought to you by  urlLink Quizilla         You\'re the smirk,a frown-smile hybrid that\'s a little bit cocky and usually associated with evil or arrogant,but attractive people.You probably just don\'t give a damn,but it\'s everyone else\'s fault if you don\'t because you\'re too awesome to have any real faults.     urlLink   What Kind of Smile are You?     brought to you by  urlLink Quizilla     urlLink        urlLink What kind of bagel are you?   Another pointless diversion from  urlLink Bijouriel        urlLink        urlLink Which monkey are you?   Another pointless diversion from  urlLink Bijouriel        urlLink          urlLink Which lj user are you?   quiz made by         urlLink          urlLink Which lj user are you?   quiz made by         urlLink     Take the  urlLink  What Fruit Are You?   test by  urlLink  Ellen  !        urlLink       urlLink ...and which lesser Harry Potter character are  you ?    (yeah right)       urlLink       urlLink ...and which lesser Harry Potter character are  you ?          urlLink              urlLink What Pattern Are You?           urlLink  Which HP Kid Are You?         urlLink     href=\"http://www.catwalkqueen.net\">Which Literary Heroine Are You?        urlLink     urlLink Which Teen Movie Queen Are You?       urlLink     urlLink Who\'s Your Josie Alter-ego?            because i thought this picture was pretty, laura of  urlLink mewing.net  told me this: \"you are an effervescent, happy person, but your friends and coworkers have noticed with dismay that you tend to bark like a dog and lick your knee, elbow, or wrist when left to your own devices. at the age of seven, you took a road trip with your family, and discovered that your favorite state is kentucky. at the age of twelve, you fell in love with the boy {or girl} next door, but were rejected in favor of your little sister. your career path involves animals and/or fashion design, but you will die alone and miserable at the age of 82.\"   urlLink whatever will laura tell you??          urlLink    You are  TENJOU UTENA  You prize nobility above all else, striving to be reunited with your prince. Apart from this, you treasure your friendships most of all. Unwilling to compromise your values in search of your dreams, you often seem to create more trouble than solve it. However, your persistence is what sets you apart from all others.  urlLink Take the \"What Magic Girl are you?\" Quiz  '),(16964,3569188,'06,June,2004','2004-06-06 00:00:00','   Hey!  I am at Erika\'s right now (duh) and doing some stuff on Gaia. Mostly just rating avatars, and such...Anyhoo. G2g, we are making macaroni and I might write later...  Baptism class tomorrow :-\\. Wish me luck and stuff.  '),(16965,3569188,'05,June,2004','2004-06-05 00:00:00','   I\'m sitting here, crying over the keyboard. I don\'t know what I am going to do. I just feel like- nothing. I feel empty and I feel lonely, and I feel completely unloved. I hate life. I sometimes feel like everyone else\'s life would have just been better if i hadn\'t been born, or if I died. It\'s so pathetic though...of course people would be sad. Very few though.   I just want somebody to hold on to, who can tell me everything is going to be alright, and I don\'t have that. School is out, the only friends I see are Corey and Erika...My mom hates me, my dad is completely horrible, and my sister hates me like she always has. Kevin hates me...I think that a few people in school liked me, but school is out and I don\'t know who to turn to.  Erika is really pathetic. She has to have everything her own way, and she is the center of the universe. Anyone who declares otherwise, should run for his life. I mean, I just hate her. And it\'s real. I mean, I just don\'t like her at all, and I wish she would die. Watch, tomorrow I will be all- I love Erika, but from now on...I don\'t know.  I hate me! Why the hell is everything so messed up. We were talking and Erika had brought up something about Corey and me being a pushover. She says that I am stubborn and I always have to argue and stand up to people and then I claimed that I didn\'t- for example Corey. And she said that he\'s got control over me, and I realized that she\'s right. I don\'t know why...I mean, it\'s like Corey\'s a special case...and with me, that\'s definitely something when it comes to me standing up for myself. Why do I let anybody have that control over me? I just can\'t let him find out because it would give him complete control over me and he\'s too much like me already. He would work me and get what he wanted way too much...Love is blind, and I love him as a friend...but I love Erika, and yeah, I still do, yet I argue with her so much it\'s not even funny.  Omg...I am a weak person today. Tomorrow I am going in the pool and that\'s all I want to do. I would be so happy to fricken drown while swimming. It would be perfect..except I am utterly afraid of drowning. Whatever..I babble. I don\'t know what I am going to do with myself.  Best Friends as of This Morning:  -Corey -Kellie -Erika  I miss Kellie. She had her date with Lance last night...wonder how it went. :-) She probably wasn\'t paying attention for most of the movie. Whatever. Everyone loves Kellie...  Well, I am gonna go to sleep and stop dwelling on my retarded problems for a couple of hours. I love having this, I don\'t need to complain to people as much...  '),(16966,3569188,'04,June,2004','2004-06-04 00:00:00','   Yaaay! It\'s summer, and I am really, really happy. First time in quite a while. I have so much that I need to focus on this summer. I guess it\'s dedicated to self improvement. More on that later.   I am so pissed at Erika. She is so retarded sometimes. She\'s annoyed that I always need the last word, and yeah, I guess that can get annoying but listen: I was arguing with my dad and I was disagreeing and attempting to get the last word subconsciously and Erika like, storms off because I was doing it. I don\'t even understand why she thinks that she has the right to just get all angry because I was arguing with someone else. Frankly, I don\'t give a crap, so whatever.   Went to PoA with Corey, Shelly, and Erika this morning. It was at the IMax and it was pretty cool. Erika and I got there early and we had to sit through this documentary about caribou. Then Shelly and Corey got there and the movie started after a little while. Once the movie started, i was a little annoyed because of the excess cuddling (corey-shelly). I mean, I can understand the hand holding and stuff, but it\'s just really annoying to me, and i don\'t know why. Erika was even more annoyed. Don\'t know why she even cares..anyway.   I am so bored. I can\'t wait until the pool is done tomorrow, I am going to go swimming! I am so excited. I got to jump in before Erika :-P. FIrst time ever...and it\'s my pool to begin with! Whatever. Grrr, she annoys me right now. Anyway, I guess I will stop writing...not much to do besides internet which sound completely boring. Maybe I will go for a run or something. I could play the violin...babbling now, must go.  '),(16967,3428326,'26,May,2004','2004-05-26 00:00:00','   like driving into on coming traffic...like getting a rope and finding a tall tree to clime.....like just not getting out of bed......like the world was against you...helpless....depressed.....hate.....love....like hate loves you....like love hates you...like no matter what you do or how hard you try it never gets better....like parking on the RxR tracks and taken a nap in the bed of your truck....like things have to get worse before they can get better but know that the best you have ever seen is worse than you ever wanted...like no one listens...like you listen to no one, or for some reason try to listen but cant or wont.....Have you ever  FELT ??? I HAVE. I DO. I always do and I know no matter what anyone says, its not going to help how I feel.  FEEL  lmmfhfao !!!!  LIFE IS A BOOK. IF YOURS IS GOOD KEEP READING. IF ITS A BAD BOOK, READ ON ANY WAYS BUT IM READY TO PUT MINE ON THE TABLE AND SEE WHAT\'S OUT THERE THAT\'S  NOT  IN MY BOOK  '),(16968,3428326,'25,May,2004','2004-05-25 00:00:00','   I just got home from  my fist day of work since thanksgiving. Its fuckin 330 in the morning.....Its supposed to be a second shift job. I got in some trouble in November and lost my last job, the judge gave me 6 months to pay my fines...Well 6 months is up in 3 weeks and I haven\'t paid anything on it yet. Finally I get a job $8 an hour second shift .....My shift starts at 3 in the afternoon. Its a meat packing plant I get to put hamburger patties in boxes all night, its a brisk 48 degrees in there but it didn\'t bother me as much as I thought it would. Most of the people are nice but have no since of humor and the place is cram packed with good lookin girls. At that point most guys would think \'heaven\' good lookin women walkin around with T.H.O. all day. Yea that\'s great and all but I like to know what time I\'m gettin off work. This stayin till 230 in the morning just aint gonna cut it.....If I didn\'t need this job so bad, trust me , I wouldn\'t be going back tomorrow. I\'ll tell ya this though, I sent out 3 resumes today before I left for work , let one of them places call me back and I\'m out of that hell hole and never looking back, and if they don\'t call then as soon as these fines are paid I\'m gone any ways. My girlfriend probly thinks I\'m out cheating on her. I doubt she\'s gonna believe that they don\'t tell when you are getting off work until its time to go...Well I got my gripe out for the day, yall taker easy and remember it could be worse.  '),(16969,3428326,'23,May,2004','2004-05-23 00:00:00','   I\'d like to dedicate this first blog to milk and cookies  urlLink   because if it weren\'t for that site id never know what a blog is.....Special thanks to jaxon for telling me how to get started...I guess that\'s it for now....Rock on!! There\'s supposed to be a link to M&C in this but it doesn\'t look like it in my preview so just in case there isn\'t >>>> www.milkandcookies.com '),(16970,3428326,'23,May,2004','2004-05-23 00:00:00','   urlLink    yep, thats me  Posted by  urlLink Hello  '),(16971,3428326,'15,June,2004','2004-06-15 00:00:00','   yep , I\'ve screwed myself again . I\'m currently on probation for my second D.U.I. , 16 months worth and a years suspended license . I live 15 miles from every where which means I\'m driving while I\'m on this suspention . Well I have to get a job to pay for my fines and costs . To make things worse they want me to go to alcohol classes  and  AA meetings , all well and good except the classes are 3 times a week and the AA is twice a week . That\'s 5 days a week I\'m gonna be in counseling , 5 work week days . The classes are 6 pm to 9 pm , when I start working I\'m gonna be working second shift , 3 pm till 11 pm , at least 5 days a week . The problem is this , if I don\'t get a job and pay , I go to jail , if I don\'t go to counseling , I go to jail . I have only enough time to do one or the other , not both . Well , I have to see the counselor on Wednesday at 10 am , hopefully we can work something out and make this work or I\'m just gonna have to tell my P.O. to fuck off and do my time in jail , which I believe is 6 months (not sure on that one). Well the people are home and I got to get this published and get off here.  '),(16972,3428326,'10,June,2004','2004-06-10 00:00:00','   yup I\'m still alive....not that any of you care , but here I am. My puter crashed shortly after my last blog, up an runnin again as you can see. I lost half my shit tho and dunno where the other half is . I didn\'t think I was gonna be able to find my blog again. I got 2 checks in the last 2 weeks , $75 a piece , drank them both up when I should have sent them to my probation . Have to go and see them bastards today , hope they don\'t throw me in jail . Broke up with my girlfriend last week (giver take a few days) she called beggin for me back , tellin me she was sorry for blowin up at me over hearin a female voice in the background while I was on the phone with her . If she had ask me what was goin on before she cussed me and hung up on me , she would have know that it was my buddies ole\' lady and not some chik she imagined I was seeing . That\'s ok , I think she is gonna give up on being jealous... I hope , heheheh. Never the less I gave her a second chance . Got a new job , worked 1 day was late the 2nd day by 15 minutes , the supervisor said it was kool but not to let it happen again but the temp service that sent me there called me and said that my assignment was finished and they didn\'t want me back , I\'m gonna go ahead and show up on Friday and see what happens . The worst they can do is send me home . Been washin trucks with a buddy of mine tryin to make extra money , I didn\'t think he was gonna pay me but he dropped off $50 yesterday . Just in time too cuz that gives me just enough cash to fill my gas tank and pay my P.O. for the piss test he popped on me 2 months ago ... yayayayyyayay!!!  '),(16973,1156278,'29,January,2003','2003-01-29 00:00:00','    Clear and present danger?  I watched the RTL4 and NOS news this morning (for those not familiar with these channels: they are Dutch TV-channels) to get a glimpse of the State of The Union, as presented by Pres. Bush last night.  Well, what can I say, it couldn\'t be more clear: the US is gonna atack Iraq no matter what. They don\'t need proof or a smoking gun either:  \"Some have said we must not act until the threat is imminent,\" Bush said. \"Since when have terrorists and tyrants announced their intentions, politely putting us on notice before they strike?\"  What this means (in my humble opinion) is, that if you just label a country as a terrorist state, you can just do whatever you like.  I just think Bush wants a war to distract the US citizens from all the internal (economic) problems...  Another interesting titbit is of course the the formation of a \"Terrorist Threat Integration Center\" to collect and analyze foreign and domestic intelligence information. The FBI and CIA were widely criticized for not cooperating and communicating before the September 11, 2001 attacks.   The office supposedly will be based at CIA headquarters in Langley, Virginia, with CIA Director George Tenet in charge, administration officials said.   Again, to me this clearly points to the start of a government within a government. The kind of shady powerfull organizations you usally see in movies like  urlLink Enemy of the State (1998)  or  urlLink Conspiracy Theory .  Now, I am not a US citizen but I do feel connected with the US, not in the last place because I have some family there and a close grilfriend.  All in all I do worry a lot about the results and repercussions that will follow the start of the war.  \"It is dangerous to be right when the government is wrong\" --Voltaire (1694-1778)   '),(16974,1156278,'28,January,2003','2003-01-28 00:00:00','    Blogger sucks!  Nice free service I thought. Well publishing something to my page seems hardly possible: although Blogger states \"upload succesfull\", it turns out that checking the FTP log actually states it DID NOT work.. Grrrr...   '),(16975,1156278,'26,January,2003','2003-01-26 00:00:00','    Where is the smoking gun?  Call me suspicious, but why can\'t the UN produce a smoking gun when it comes to the evil empire Iraq? Maybe Iraq isn\'t half the threat the US wants the world to believe.  On the other hand: if they (they as in \'the US\') really want a war, you would think they can \'invent\' the proof they need. Afterall: they were able to blame Lee Harvey Oswald for the murder of John F. Kennedy back in the 60\'s.  So how hard can it be to forge some documents and convert them into the smoking gun they need? Or  plant some evidence in some strategic places.  Last week they almost got there when they found some obsolete, forgotten warheads. The US claimed it to be warheads for missiles, but anyone could see by just looking at them that these \'warheads\' would only be good for the scrapheap.  I am very curious what Bush will tell the world today when he addresses the congress and the American public.  I, as a European, am very intrerested to find out how he will deal with it.   '),(16976,1156278,'26,January,2003','2003-01-26 00:00:00','    USA to act alone?  Just read this article on CNN ( urlLink Powell: U.S. prepared to act alone ).  Is it me or is this whole deal just getting more and more out of control?   A collegue of me send me a cartoon about the whole Iraq situation that is rather funny. I will try and add it here (through the magic wonders of FTP).  Anyhow: my 2 cents worth: they should have finished the job years ago when Schwarzkopf had the chance. In my opinions (but who\'s intrested anyway) this is a darn farce anyway.   '),(16977,1156278,'26,January,2003','2003-01-26 00:00:00','    Monday Alert: another week coming up!!   Mondays really fuck up my life (pardon my French). I get up in the morning, slowly forcing my body to actually move to the bathroom. Looking back at me on the other side of the mirror, I see some freak with small eyes, exploded hair, yawning like I\'m trying to get a new world record. O yeah, it’s Monday alright.   Monday tells you 2 things:   A. Your weekend is over B. Work, Work WORK!  A typical Monday for me looks a bit like this:  7:45am - Starting up movement-coordination brain cells 8:05am - Starting shaving feature (trying not to bleed myself to death) 8:15am - Dressing up (damn, my shirt is wrinkled again) 8:20am - Feed my cats auto-drive system kicks in 8:30am - Make sandwiches for work 8:45am - Starting driving ability system to commute between Almere - Amsterdam 9:20am - Boot up my Apple Powerbook G4 Titanium (first smile of the day) 9:30am - powering up coffee-mechanism control center 9:35am - Life support enabled (after 2 coffee, milk, 3 sugar) 10:00am - Main act-react system online 10:30am - Speech enabled (more or less) 11:00am - Finally I wake up mentally, I find myself sitting at my desk. I’m at work, Arrrggh...  And I just know millions of people go through this same routine like me!  I\'m *NOT* a morning person. I can stay up late (4am) and sleep all day. When I\'m behind a PC (like now, writing this) I can sit there for hours forgetting the world around me. A social life? Where can I download that? Let\'s face it, I am a sorry ass GEEK! And I need a vacation.   On the positive side: I do like my job. It\'s demanding since we have a rather small team, but it\'s challenging because of that.   '),(16978,1156278,'25,February,2003','2003-02-25 00:00:00','    \"A true Gadgeteer\"   All right folks, I have a confession to make: I’m a slave of modern technology. If something cool and new comes out, I just have to have it. Ok, I do not own a state of the art plasma TV or stuff like that (I am not a millionaire), but I do own 2 Pc’s, 3 PDA’s, 3 phones, my car is equipped with GPS, I have internet via cable (and my own network), I work with Windows, Linux and OS X, etc. etc….  So when I finally saw the brand new  urlLink Sony Ericsson P800 mobile phone this weekend, I had to have it. In the US you can’t get it yet, but you can  urlLink pre-order the P800 here.  This is the ultimate gadget: it’s a phone and PDA in one and it is not half as bulky as, lets say,  urlLink the XDA from O2  or has the problem its doesn’t work in Europe (like the Tungsten from Palm).  Among the cool things:  - pops-up a photo of the person calling - give every caller its own ringtone - can play midi, wav and MP3 (as ringtones) - polyphonic - has full PDA capabilities - built-in camera (max 640x480) - Handsfree from the handset (flip it open and it turns on the mic)  And so on…  So now I start taking little pictures of all the people in my phonebook: that’s one of the coolest things so far… For instance: when my office calls me, I see a picture of the building and I hear Beavis scream: “I am Cornholio, I need teepee for my bunghole”.  I think the expression is: a fool with a tool is still a fool…   '),(16979,1156278,'16,February,2003','2003-02-16 00:00:00','    The number of the beast  In all the commotion around the upcoming war in the Middle East, people regain new interest in all the biblical and other mystical prophecies that surrounds these global events of epic proportions.    Although I consider myself not religious in a biblical sense, I do find these prophecies - or better said: the interpretations of the prophecies – very interesting (one look at my DVD collection will reveal some movies with a very religious theme).  So I dug up some interesting things on the net. Let’s start with the bible, King James Edition in English, and look up Revelations:   \"16:13: And I saw three unclean spirits like frogs come out of the mouth of the dragon, and out of the mouth of the beast, and out of the mouth of the false prophet. 14: For they are the spirits of devils, working miracles, which go forth unto the kings of the earth and of the whole world, to gather them to the battle of that great day of God Almighty. 16: And he gathered them together into a place called in the Hebrew tongue Armageddon.\"   Now, off course we can interpret this in various ways, but some say Armageddon is placed in the Middle East and the three spirits are the armies of the US, England and Iraq. So they are there to spread the war all over our world.  Fascinating stuff.  One of the countless websites that says its on too the explanation of these things is this one:  urlLink Count the Number of the Beast: Who is the Antichrist, The Beast of Revelation 13? . The author is also a guy who developed “(…)a new holistic method of prophecy(…)  to do his own prophecies.  Off course (and this is pretty obvious) he has a page stating Osama Bin Laden is the antichrist:  urlLink Count the Number of the Beast: Who is the Antichrist, The Beast of Revelation 13? Osama the \"Destroyer\" . How lame can you get? I also think that the picture of the painting has been tampered with (I will try and verify this, since it’s a Dutch painting).   TCPA, TPM, DRM and the number of the beast?  Again a quote to start with:    “And he causeth all, both small and great, rich and poor, free and bond, to receive a mark in their right hand, or in their foreheads (Rev 13:16): And that no man might buy or sell, save he that had the mark, or the name of the beast, or the number of his name (Rev 13:17). Here is wisdom. Let him that hath understanding count the number of the beast: for it is the number of a man; and his number is Six hundred threescore and six (Rev 13:18).”    “Ok,” you ask yourself, “how the hell – pardon my French – did this guy mix computers and religion together?”. Well, hang on this is an interesting theory that actually binds Microsoft, the US government and computers together.  When I hypostasize (hmm, I had to look up that word in a dictionary) about the number of the beast, I must say  I rather believe the Mark or number of the Beast will be a (financial) identification system that the Antichrist will establish during the tribulation then a person. He will use the Mark as a tool for controlling all aspects of society.   Well as I see it, TCPA, TPM, DRM might be part of that.  First let me explain the terms TCPA, TPM, DRM;  TCPA = Trusted Computing Platform Alliance TPM = Trusted Platform Module DRM = Digital Rights Management  In short this is how it works;  TCPA will make copying software (or digital assets in general) impossible whether or not the copying itself is legal or illegal. In Holland for instance, the law says that anyone is entitled to make a copy or backup for personal use of any medium (CD, video, DVD whatever), as long as I am the owner.  TCPA will block anything the owner of the original (i.e. Microsnot software or music) doesn’t want you to do. TCPA will also prevent (the same) software from running on more than one computer.  The core of it is a chip called Trusted Platform Module or TPM. TPM is just a super-CPU-ID that will also contain - next to a unique identification number – a cryptographic key and some checksums. Its like a mega-dongle (a dongle being a hardware key. It\'s a secure, copy-protected device supplied with software that plugs into a computer port).  Just like with a steel ball on a demolition crane the TPM in itself is harmless: it’s the chain connected to it that makes it dangerous. The chain in this case is called Palladium and is propagated by … Microsnot!  By matching serial numbers of software with the PC it runs on, the same serial will never work on another PC. Music, movies and other files will be chained to your PC by use of the TPM chip. Copying or trading these files to another PC will be impossible; copying to systems without TMP won’t work either.   Did I say PC? Well TPM is not just good for PC’s. Mobile phones, PDA’s and anything else more or less electronically can have a TPM chip.    Which brings me to the part that says: “And that no man might buy or sell, save he that had the mark, or the name of the beast, or the number of his name”  And that number is not 666, it’s the unique identification number TPM will give you! And it will make it impossible for you to buy or sell something unless you have a TPM!    Other views please?  Some folks believed the Beast number was to be a certain year. Because of this, the year 666 AD was looked on as the date for the advent of the Antichrist. When that date came and went, the year 1666 was then looked on with dread. Nostradamus pointed to the year 1999 as an inverted 666 year.  Because the Beast number will have something to do with the Antichrist\'s name, people all over the world have been busy working equations to figure out if every world leader of their day has a name that can be made to equal 666 in some way. Guess what: depending on how you figure it, 666 can be made to add up to just about anyone\'s name.  I even have heard about a number of superstitions or phobias about this number (all from the Internet). One person refused to ride in a car because the license plate had 666 on it. Supposedly whenever some people go shopping, and the cash register reads triple six in anyway, they add something like a stick of gum to go past the unholy number.  When President and Mrs. Reagan moved out to California their house number was, you guessed it, Six Six Six. They had it changed to 668. When a governing body, like the UN or Congress, produces a bill numbered 666, it receives a good going over by prophecy watchers to see if it has a connection with end time events.  Now on the other hand, the next time you receive your water bill and upon opening it, you discover it comes to Euro or $ 6.66 or you realize you\'ve used 666 cubic feet or cubic meters of water - don\'t go pulling your hair out. Unless it has a direct connection to the Antichrist or is some special message from God, 666 is just another number that randomly comes up.   What can I say: I just love conspiracy theories!   Next time I have time for it I will focus on the Pale Rider.   (…)and behold a pale horse: and his name that sat on him was Death, and Hell followed with him. And power was given unto them over the fourth part of the earth, to kill with sword, and with hunger, and with death, and with the beasts of the earth(…)   Ps: if these ramblings don’t make any sense: don’t worry, I wrote them while doing lots of other things during this day, like cooking, cleaning, eating, ironing and watching some movies.   '),(16980,1156278,'06,February,2003','2003-02-06 00:00:00','    Give peace a chance - Point of no return  At the same time I’m writing this, probably a thousand other people are also filling their Blogs with the same words: Give peace a chance. Although I did not attend any of the so many peace protests that went on yesterday (see:  urlLink CNN.com - Cities jammed in worldwide protest of war in Iraq - Feb. 15, 2003  and  urlLink CNN.com - European protesters fill cities - Feb. 15, 2003 ), I do support the effort to keep peace.  Even my girlfriend in the US did attend the peace rally that went on in here city ( urlLink Tallahassee Democrat | 02/16/2003 | Hundreds rally against war ) and I bet she never ever did something like that before.  One of the reasons I did not attend the peace rally that took place here in Amsterdam, is the fact that I do not believe they make a difference: only voting does. What I mean is that they should have some referenda about these subjects. Just let people vote directly.   I mean: let’s stay realistic here: Bush already invested a couple of billion to get soldiers and equipment over there in the desert; he has reached the point of no return already: backing out now will mean he will loose face and I think that nothing else frightens Bush more than that.  And while we are on this subject: it is so blatantly obvious that Bush is measuring with two measures. Go figure: how else can it be that he is going to attack Iraq but not taking action against North Korea? According to the CIA they already have nuclear capabilities ( urlLink CNN.com - Tenet: North Korea has ballistic missile capable of hitting U.S. - Feb. 12, 2003 ). Iraq on the other hand has not: at least not proven yet. They don’t have even have a nuclear plant (like North Korea does). They might (emphasis on might) have some materials to maybe construct one. But they certainly do not have it yet.  What’s the difference one might ask? Well I can only think up one word: oil.   '),(16981,1156278,'06,February,2003','2003-02-06 00:00:00','    Unraveling Donnie\'s mystery  Before I start raving about the peace protests that went on yesterday all over the world, I will just start with a lighter subject;  Yesterday I  bought the Donnie Darko DVD. Yes, it’s true. I shelled out some 20 Euros (at this moment it’s the equivalent of about 23 dollars, since the Euro went up) and bought me the “DTS Special Edition” supposedly “loaded with extras”.  Ever since I discussed the movie on Feb. 3rd, I couldn’t help being (and staying) very mesmerized by it. There was another coincidence that I didn’t mention here yet. When I gave the movie to my colleague to watch it, the next day when he gave it back to me, he showed me something really funny;  As it turned out he had a newspaper clipping with him from that very same day, discussing the DTS Special Edition of Donnie Darko. Weird huh? Remember this movie came out in 2001. The odds of a newspaper writing about it in 2003 are very small!  Anyway, it was a very positive critique about the movie but very negative about the extras on the DVD. In fact it said it spoiled the mystery of the movie. I didn’t have time yet to watch all of it with the director’s commentary on, but I did watch the deleted scenes already. One of them is rather unveiling; it’s a short sequence that shows that the medication Donnie is taking is actually a placebo.  As soon as I get time I will get back on this issue, but it’s fascinating already to watch these deleted scenes (with the director’s commentary).   '),(16982,1156278,'06,February,2003','2003-02-06 00:00:00','    Getting there  It feels like the war can break out this week. Bush\'s war retoric isn\'t going to end soon and we just know he wants to start a war. Whats up with that and whats up with Korea? Are the US afraid to tackle the problems they have with Iraq by diplomacy? They are not doing anything war-like against Korea.   Movie freak  I watched Donnie Darko again the day before yesterday. Then I gave the film to a collegue of mine to watch it: he liked it too. I will buy this movie soon. Its about $20 but worth every penny. I am a DVD addict now. Ever since I got my widescreen TV and my DVD home theatre I can\'t control myself and I just keep buying movies like crazy. Have a look at my collection here:  urlLink DVD Profiler\'s My Profiler Collection . Is that slowy getting out of control or what?   '),(16983,1156278,'04,February,2003','2003-02-04 00:00:00','    \"In measuring a circle one begins anywhere\"  Ok, its 23:45 on a Monday night and I just saw a movie that left me puzzled way beyond usually. I pride myself in being rather a movie buff and I own a fast growing library of DVD\'s including the Sixth Sense.  Now, I have seen some Sixth Sense rip-offs like \"The Others\" but none of them came close to leaving me behind with a very very disturbed feeling like this one did: \"Donnie Darko\". There, I finally said it.  I never ever heard about it, never ever saw a preview or anything else. When i rented the movie I did it because finally it went from the A list (which costs like 4 Euro\'s to rent for just 1 day!) to the B list where you can keep the movie a week.  Already I know I will have to see it more than 1 time this week because it is so damn intriguing.  For those that have no clue whatsoever I\'m talking about, go check it out. If you only read the plot synopsis you just know how strange it’s going to be!  \"Donnie Darko needs to take his meds (he seems to be a schizophrenic). He starts seeing a tall, mutant rabbit named Frank, who leads Donnie to do bad things at night because, well, the world will end in 28 days, 6 hours, 42 minutes and 12 seconds. Frank also saves Donnie from a jet engine that falls from the sky and crashes into his bedroom, not to mention turning him on to time travel.\"  If you seen movies like Memento or Fight Club or anything that compares to those you will need to see this one.  The famous (or should I say infamous) Rotten Tomatoes has some really funny comments on this movie from some magazines:  \"reels the viewer in like a black hole, like a quicksand hourglass.\" \"Think Rushmore meets Twelve Monkeys and you\'re on the right lines.\" \"... has \'cult classic\' stamped all over it.\" \"It’s like nothing you’ve ever seen before, and it sets a great tone that takes you places you\'ve never been before.\" \"Puts the comparable American Beauty to shame.\" \"A stunning blend of sci-fi, psychological thriller, and dark comedy.\" \"There was no easy logline for Donnie Darko. That’s probably why it was doomed to become one of 2001’s best movies no one heard of.\"  When a movie is this controversial it usually means its either a great piece of art or something you will see sold for under $10 (or Euro\'s) real soon.  And what do I think? Let me sleep on this one. I just know I will be back on this one soon!   '),(16984,1156278,'02,February,2003','2003-02-02 00:00:00','    Bizarre coincidence?  Sometimes life is so weird. I was reinstalling a PC of a friend and I put XP on it. Now you might know that XP has this login screen with little icons on it. You can change them and replace them with a photo of yourself if you want. It seems that XP chooses a random icon to represent the people that can login. I created 2 users so I had 2 icons.  Now here comes the bizarre coincidence: XP choose an icon of.... the space shuttle! I could *not* believe my eyes but it is true. Go through the icons and you will see there is one that is the space shuttle at take off.  Life IS more bizarre than any story!   '),(16985,1156278,'01,February,2003','2003-02-01 00:00:00','    The Final Frontier  Wow. Thats the only word I can think up for now. This afternoon I received an e-mail alert by CNN stating that NASA had lost contact with the space shuttle while they were on their re-entry to earth. Needless to say this is again a big tragedy. I still remember when the Challenger blew up years ago. Now the Columbia is gone and with it a fine crew of 7.  Off course this will set back all the missions planned for this and coming years. I\'m sure they don\'t allow any shuttle to lift-off untill they have a clear idea what caused this. Since they broke up or exploded on re-entry my first - non expert - idea would be that the heat shield got damaged and as a result the shuttle exploded. But who knows? One thing is for sure: there won\'t be any round-trips to the ISS soon (unless they need to get the people of it or transport some food and other things.  Given the curcumstances (a war is at hand)  I wonder what will happen now: will Bush just go on or will this tragedy slow down his plans?   '),(16986,1156278,'28,March,2003','2003-03-28 00:00:00','    When looking back one is always right   Nothing is easier than looking back on things and analyzing them. Such is also the case with the war in Iraq. People that were against the war (like me) are afraid that many innocent people will get hurt. But now it seems it’s a walk-over for the US. I wonder how much longer the war will last.   '),(16987,1156278,'23,March,2003','2003-03-23 00:00:00','    “There is nothing more frightening than ignorance in action – J. Von Goethe (1749-1832)”  With the war going into another day, this is a truly interesting quote that popped into my mind. The Iraqi’s seem to be able to hold back the US and British army more than we expected and this will make it more and more difficult to tell when this war will be over.   Just a dream?   In the night the war started I woke up in the morning from a dream I had. Well, actually it was more of a fragment of a dream. The fact that I woke up from it and remembered it actually, was in itself very unique. Usually I don’t remember anything I dream. So this was pretty special and the subject was also very frightening.   My dream consisted of a vision – or image – that popped into my mind together with one thought.   The setting was very apocalyptic: the image was that of a giant mushroom cloud of an atomic bomb and then the blast impact rushing towards me. Then came what I can only describe as my final thought: “Oh my god the stupid idiots finally did it”.   Then I woke up. When I turned on my TV I discovered that the war had started and together with the image from my dream this gave me a very uneasy feeling that day.   Off course this dream and thought might come from the many movies I’ve seen. In particularly a scene in Terminator 2 pops into my mind. But all in all it was rather a weird experience.   '),(16988,1156278,'19,March,2003','2003-03-19 00:00:00','    11 hours, 11 minutes to go  According to my countdown clock we are now 11hrs, 11 minutes away from another Gulf war. At least, that is what we think: given the circumstances the US and Brittain (I keep forgetting them all the time) might attack earlier. This morning I spoke to a friend of mine who is an officer in the Dutch army. He said that it is no accident the US did choose this date to attack. According to him it is an advantage that it is full moon right now. Even though the US troops are technologically very advanced, they can use the extra light from the moon when attacking. Night vision doesn\'t work in complete darkness, contrary to popular belief. In fact they work the best with a full moon. So do all the electronics in the fighter and bomber jets. Off course these kind of considerations go way beyond normal people like me. It sounds as a pretty clinical approach. But in times of war that is the only option.   '),(16989,1156278,'18,March,2003','2003-03-18 00:00:00','    \"I would never die for my beliefs because I might be wrong.\" -- Bertrand Russell (1872-1970)  With a new Gulf war on its way in the next 24 hours, I can only think up this quote from this famous British philosopher who also was a know pascifist. Who are we to decide for another person - let alone another country - what is right and what is wrong? Now, we all know Saddan Hoessein is an evil dictator who exploits his own people. However, starting a war is not the solution to the problem, since it will hurt innocent people that have part in anything going on.  That the war was coming was obvious: and I quote myself from my February 16th blog:  \"(...)I mean: let’s stay realistic here: Bush already invested a couple of billion to get soldiers and equipment over there in the desert; he has reached the point of no return already: backing out now will mean he will loose face and I think that nothing else frightens Bush more than that.(...)\"  This is the reason he never even put up a vote in the UN and why is going to ignore the UN.   Now, another big issue in my mind is: who will have the power to stand up against the US next time? It seems the US is the only super-power left in the world that can ignore the rest of the world (UN) and do whatever it wants...   In retrospect this is all because of what happened during the first Gulf war. The Following is an excerpt from the Encarta Encyclopedia:   The Gulf War, August 1990 - February 1991    The Iraqi-Kuwaiti border had been the focus of tension in the past. Kuwait was nominally part of the Ottoman Empire from the 18th century until 1899 when it asked for, and received, British protection in return for autonomy in local affairs. In 1961 Britain granted Kuwait independence, and Iraq revived an old claim that Kuwait had been governed as part of an Ottoman province in southern Iraq and was therefore rightfully Iraq’s. Iraq’s claim had little historical basis, however, and after intense global pressure Iraq recognized Kuwait in 1963. Nonetheless, there were occasional clashes along the Iraqi-Kuwaiti border, and relations between the two countries were sometimes tense.   Relations between the two countries improved during the Iran-Iraq War (1980-1988), when Kuwait assisted Iraq with loans and diplomatic backing. After the war ended in 1988, the Iraqi government launched a costly program of reconstruction. By 1990 Iraq had fallen $80 billion in debt and demanded that Kuwait forgive its share of the debt and help with other payments.  At the same time, Iraq claimed that Kuwait was pumping oil from a field that straddled the Iraqi-Kuwaiti border and was not sharing the revenue. Iraq also accused Kuwait of producing more oil than allowed under quotas set by the Organization of Petroleum Exporting Countries (OPEC), thereby depressing the price of oil, Iraq’s main export .   Iraq’s complaints against Kuwait grew increasingly harsh, but they were mostly about money and did not suggest that Iraq was about to revive its land claim to Kuwait. When Iraqi forces began to mobilize near the Kuwaiti border in the summer of 1990, several Arab states tried to mediate the dispute. Kuwait, seeking to avoid looking like a puppet of outside powers, did not call on the United States or other non-Arab powers for support. For their part, the U.S. and other Western governments generally expected that at worst Iraq would seize some border area to intimidate Kuwait, so they avoided being pulled into the dispute. Arab mediators convinced Iraq and Kuwait to negotiate their differences in Jiddah, Saudi Arabia, on August 1, 1990, but that session resulted only in charges and countercharges. A second session was scheduled to take place in Baghd&#257;d, the Iraqi capital, but Iraq invaded Kuwait the next day, leading some observers to suspect that Iraqi president Saddam Hussein had planned the invasion all along.   The United Nations Security Council and the Arab League immediately condemned the Iraqi invasion. Four days later, the Security Council imposed an economic embargo on Iraq that prohibited nearly all trade with Iraq. Iraq responded to the sanctions by annexing Kuwait on August 8, prompting the exiled Sabah family to call for a stronger international response. In October, Kuwait’s rulers met with their democratic opponents in Jiddah, with the hope of uniting during the occupation. Fewer than half of all Kuwaitis stayed in Kuwait through the occupation; of those who stayed, some formed resistance organizations but with little effect.   Any armed attempt to roll back the Iraqi invasion depended on Saudi Arabia, which shares a border with both Iraq and Kuwait. Saudi Arabia had neither the power nor the inclination to fight Iraq alone; if the Saudi government invited foreign troops into the country to attack Iraq, however, it risked appearing to be under their influence. Saudi rulers did eventually open the country to foreign forces, in large part because they were alarmed by Iraq’s aggressive diplomacy and because U.S. intelligence reports claimed that Iraqi forces were well positioned for a strike against Saudi Arabia. Other Arab countries, such as Egypt, Syria, and the smaller states along the Persian Gulf, feared that even if Iraq’s conquests stopped at Kuwait, Iraq could still intimidate the rest of the region. Western powers supported a rollback of Iraqi forces because they were afraid Iraq could now dominate international oil supplies. Finally, other members of the United Nations (UN) did not want to allow one UN member state to eliminate another.   Beginning a week after the Iraqi takeover of Kuwait and continuing for several months, a large international force gathered in Saudi Arabia. The United States sent more than 400,000 troops, and more than 200,000 additional troops came from Saudi Arabia, the United Kingdom, France, Kuwait, Egypt, Syria, Senegal, Niger, Morocco, Bangladesh, Pakistan, the United Arab Emirates, Qatar, Oman, and Bahrain. Still other countries made other contributions: Turkey allowed air bases on its territory to be used by coalition planes, and Japan and Germany gave financial support. The initial goal of the force was to prevent further Iraqi action, but most countries were aware the force might ultimately be used to drive Iraq from Kuwait.   The Iraqis tried to deter and split the growing international coalition through several means. They made it clear that their adversaries would pay heavily if war broke out, and they hinted they would use chemical weapons and missile attacks on cities, as they had against Iran during the Iran-Iraq War. Iraq also detained citizens of coalition countries who had been in Kuwait at the time of the invasion and said they would be held in militarily sensitive areas—in effect using them as human shields to deter coalition attacks. Iraq eventually released the last of the foreigners in December 1990 under pressure from several Arab nations.    In an effort to weaken Arab support within the coalition, Iraq tried to link its occupation of Kuwait to the larger Arab-Israeli conflict in the region. The Iraqis argued that since the UN had not forced Israel to leave Arab territories it occupied during and after the Six-Day War of 1967, it should not force Iraq to leave Kuwait. The Iraqis further implied they might leave Kuwait if Israel withdrew from the Occupied Territories. Several Arab countries responded positively to Iraq’s statements; however, most of these were states such as Jordan and Yemen, which were not part of the coalition. Only in Morocco and Syria did government support for coalition involvement weaken as a result of Iraq’s initiative.    The coalition’s greatest military concern during the closing months of 1990 was that Iraqi forces would attack before coalition forces were fully in place, but no such attack took place. The coalition was also troubled that Iraq might partially withdraw from Kuwait, which could split the coalition between nations eager to avoid fighting and nations wanting to push for full withdrawal. The United States in particular feared that signs of progress might lessen the resolve of some coalition partners and so discouraged attempts to mediate the crisis. Iraq’s uncompromising stand helped build support among coalition members for the American hard line.    On November 29, with coalition forces massing in Saudi Arabia and Iraq showing no signs of retreat, the UN Security Council passed a resolution to allow member states to “use all necessary means” to force Iraq from Kuwait if Iraq remained in the country after January 15, 1991. The Iraqis rejected the ultimatum. Soon after the vote, the United States agreed to a direct meeting between Secretary of State James Baker and Iraq’s foreign minister. The two sides met on January 9. Neither offered to compromise. The United States underscored the ultimatum, and the Iraqis refused to comply with it, even threatening to attack Israel. For the United States, the meeting was its way of showing the conflict could not be resolved through negotiation.    A large minority of the U.S. population opposed military action. Opponents were concerned that the armed forces would suffer large casualties and argued that the only reason for the invasion was to guarantee a cheap supply of oil. Many such opponents thought economic sanctions would eventually force Iraq to leave Kuwait.  President George Bush maintained that larger political principles were involved and that economic sanctions would not work. He also argued that the UN resolution gave him the authority to use military force. Other Americans believed the president did not have the constitutional authority to order an attack without a congressional declaration of war. On January 12, 1991, the U.S. Congress narrowly passed a resolution authorizing the president to use force, nullifying the domestic debate.    When the UN deadline of January 15 passed without an Iraqi withdrawal, a vast majority of coalition members joined in the decision to attack Iraq. A few members, such as Morocco, elected not to take part in the military strikes. In the early morning of January 17, 1991, coalition forces began a massive air attack on Iraqi targets.   Although the coalition air forces often used very precise weapons, the attacks caused many civilian casualties and completely disrupted Iraqi civilian life. The third task, weakening Iraq’s ground forces, was larger still. The coalition used less sophisticated weaponry to strike Iraqi defensive positions in both Iraq and Kuwait, to destroy their equipment, and to undermine morale. After five and a half weeks of intense bombing and more than 100,000 flights by coalition planes, Iraq’s forces were severely damaged.   In an attempt to pry the coalition apart, Iraq fired Scud missiles at both Saudi Arabia and Israel, which especially disrupted Israeli civilian life. Iraq could thus portray its Arab adversaries as fighting on the side of Israel. The strategy failed to split the coalition, in part because the Israeli government did not retaliate. Iraq also issued thinly veiled threats that it would use chemical and biological weapons. The United States hinted in return that such an attack might provoke a massive response, possibly including the use of nuclear weapons. Iraqi ground forces also initiated a limited amount of ground fighting, occupying the Saudi border town of Khafji on January 30 before being driven back.   One month into the air war, the Iraqis began negotiating with the Union of Soviet Socialist Republics (USSR) over a plan to withdraw from Kuwait. Had this initiative come before the start of the coalition’s attack, it might have split the coalition; now it simply seemed a sign that the war was weighing heavily on Iraq. The war made diplomacy difficult for Iraq: officials had to travel overland to Iran and then fly to Moscow to ferry messages back and forth. Sensing victory, the coalition united behind a demand for Iraq’s unconditional withdrawal from Kuwait.   LAND WAR   On February 24 the coalition launched its long-anticipated land offensive. The bulk of the attack was in southwestern Iraq, where coalition forces first moved north, then turned east toward the Iraqi port of Al Ba&#351;rah. This maneuver surrounded Kuwait, encircling the Iraqi forces there and in southern Iraq, and allowed coalition forces (mainly Arab) to move up the coast and take Kuwait city. Some Iraqi units resisted, but the coalition offensive advanced more quickly than anticipated. Thousands of Iraqi troops surrendered. Others deserted. Iraq then focused its efforts on withdrawing its elite units and sabotaging Kuwaiti infrastructure and industry. Many oil wells were set on fire, creating huge oil lakes, thick black smoke, and other environmental damage. Two days after the ground war began, Iraq announced it was leaving Kuwait.   On February 28, with the collapse of Iraqi resistance and the recapture of Kuwait—thereby fulfilling the coalition’s stated goals—the coalition declared a cease-fire. The land war had lasted precisely 100 hours. The cease-fire came shortly before coalition forces would have surrounded Iraqi forces. On March 2 the UN Security Council issued a resolution laying down the conditions for the cease-fire, which were accepted by Iraq in a meeting of military commanders on March 3. More extensive aims, such as overthrowing the Iraqi government or destroying Iraqi forces, did not have the support of all coalition members. Most Arab members, for example, believed the war was fought to restore one Arab country and not to destroy another. The United States also worried that extending the goal would have involved them in endless fighting.   The Iraqis achieved none of their initial goals. Rather than enhancing their economic, military, and political position, they were economically devastated, militarily defeated, and politically isolated. Yet because the government and many of the military forces remained intact, the Iraqis could claim mere survival as a victory. The surviving military forces were used a short time later to suppress two postwar rebellions: one involving Shiite Muslims in southern Iraq and one involving Kurds in the north.   Almost all of the casualties occurred on the Iraqi side. While estimates during the war had ranged from 10,000 to 100,000 Iraqis killed, Western military experts now agree that Iraq sustained between 20,000 and 35,000 casualties. The coalition losses were extremely light by comparison: 240 were killed, 148 of whom were American. The number of wounded totaled 776, of whom 458 were American.   CONSEQUENCES OF THE WAR    The end of the fighting left some key issues unresolved, including UN sanctions against Iraq, which did not end with the war. On April 2, 1991, the Security Council laid out strict demands for ending the sanctions: Iraq would have to accept liability for damages, destroy its chemical and biological weapons and ballistic missiles, forego any nuclear weapons programs, and accept international inspection to ensure these conditions were met. If Iraq complied with these and other resolutions, the UN would discuss removing the sanctions. Iraq resisted, claiming that its withdrawal from Kuwait was sufficient compliance.   Many Western observers believed the victory was hollow because Saddam Hussein was still in power. At first, when Hussein was greatly weakened, Western powers believed a rebellion might succeed in overthrowing him. Meanwhile, potential rebels within Iraq believed they might receive international help if they rebelled. But when the Shiite population of southern Iraq rebelled shortly after the cease-fire, they were greeted not with international help but with Iraqi military forces returning from the southern front. It quickly became clear that the rebels would receive no international help, although several governments gave them verbal support. Under the terms of the cease-fire, which established “no-fly zones” in the north and south, Iraqis could not attack the Shiites with airplanes, but could use helicopters, which they did to great effect. Spontaneous and loosely organized, the rebellion was crushed almost as quickly as it arose.   The defeat of the Shiites made the debate over helping Iraqi rebels even more urgent. Ultimately, however, most Western governments decided that if Hussein collapsed, Iraq might disintegrate, ushering in a new round of regional instability. A short while later, Kurds in the north of the country rebelled, and they too received no help. The Kurds were able to withstand Hussein longer than the Shiites, in part because they had a history of organized, armed resistance. In the end, though, the Kurds achieved only a very modest success: a UN-guaranteed haven in the extreme north of the country. No permanent solution—such as Kurdish self-rule—was negotiated.   Elsewhere the effects of the war were less severe. In Kuwait the prewar regime was restored, and in 1992 the emir, Sheikh Jaber al-Ahmad al-Jaber al-Sabah, honored his pledge in exile to reconvene the country’s parliament. Palestinians in Kuwait fared poorly after the war, in large part because Yasir Arafat of the Palestine Liberation Organization (PLO) and other prominent Palestinians had endorsed Hussein and his anti-Israeli rhetoric. Blamed for collaborating with the Iraqis, most of the Palestinian population (estimated at 400,000 before the war) was expelled from Kuwait or forbidden to return.   Following the war, thousands of American soldiers developed mild to debilitating health problems, including abdominal pain, diarrhea, insomnia, short-term memory loss, rashes, headaches, blurred vision, and aching joints. The symptoms became known collectively as Gulf War Syndrome but their cause was unknown. Speculation about the cause centered on exposure to chemical and biological weapons; experimental drugs given to troops to protect against chemical weapons; vaccinations against illness and disease; insecticides sprayed over troop-populated areas; and smoke from burning oil wells ignited by retreating Iraqis. The U.S. Department of Defense originally stated it had no conclusive evidence that troops had been exposed to chemical or biological weapons. However, in 1996 the department acknowledged that more than 20,000 American troops may have been exposed to sarin, a toxic nerve gas (see Chemical and Biological Warfare). In 1997 the U.S. Central Intelligence Agency (CIA) suggested the deadly gas may have spread farther than previously thought, affecting perhaps hundreds of thousands of troops.   The UN continued to maintain most of the economic embargo on Iraq after the war, and several coalition countries enforced other sanctions, such as the no-fly zones. As hardships to the Iraqis mounted, the consensus on sanctions decayed. The United States, the United Kingdom, and Kuwait insisted on maintaining the sanctions, however, arguing that many of the hardships were the fault of the Iraqi government. For example, the sanctions allowed Iraq to sell limited amounts of oil for food and medicine if it also designated some of the revenue to pay for damages caused by the war. Until December 1996, Iraq rejected this deal as an infringement on its sovereignty. Hussein also complicated matters by mobilizing forces on the Kuwaiti border in late 1994 and by interfering with the work of UN inspectors. This interference nearly led to renewed military conflict in 1998, but a UN-brokered agreement allowing the inspectors to resume work averted the immediate crisis. Reports that Iraq was continuing to block inspections prompted the United States and Britain to launch a four-day series of air strikes on Iraqi military and industrial targets in December. In response, Iraq declared that it would no longer comply with UN inspection teams, called for an end to the sanctions, and threatened to fire on aircraft patrolling the no-fly zones. Through early 1999, Iraq continued to challenge the patrols, and British and U.S. planes struck Iraqi missile launch sites and other targets.   '),(16990,1156278,'04,March,2003','2003-03-04 00:00:00','    Tripod\'s messing up!  For those that know my webpage a bit longer and have visted my webplace more often, know that I had my share of problems with Tripod. Ok, its free and you get 20meg of diskspace (10 more than with my own ISP - UPC) and 1 gig of traffic for free a month, but they do mess up a lot. Very annoying! Well, they did it again! they deleted my vandegoor.tripod.com account! I have no idea why except when I tried to access it I got this message that the user was deleted becasuse of some violation!  What it was I will not know, but the fact of the matter is I lost about 4 months of information! Luckily I had a backup and also a \"mirror\": hvdgoor.tripod.com. You are there now, when you visit www.vandegoor.com. Anyone know a provider that can host me for $100 a year?  Anyway, this is costing me way to much time!   '),(16991,1156278,'30,April,2003','2003-04-30 00:00:00','    Queen\'s Day; the Dutch national garage sale day   Ok, for those not familiar with the ‘Dutch way of life’; today is Queens Day. We celebrate the birthday of our Queen (well, actually the birthday of our Queen’s mum) and this is a national holiday so everyone has a day off.  We celebrate this by means of the largest outdoor garage sale you can imagine. Everyone cleans there attic or garage and tries to sell all the rubbish they didn’t want to throw away with the garbage. But that’s what most of the stuff is: garbage. You’ll find the most stupid things on a day like this: old clothes, toys, books and old electronics. You name it, and someone is trying to sell it.  It’s really funny how you can sell junk to someone else that probably will try to sell it next year.  Anyway: I’m kicking back and relaxing by watching some movies and editing my webpage (like this blog) and just enjoying my free time in general.  So from this place I say to everyone: Happy Queen’s Day!   '),(16992,1156278,'31,May,2003','2003-05-31 00:00:00','    It’s all in the game   One last post on the verge of another month. Lately I have been very busy with another hobby of mine: online gaming. I play this game called Soldier of Fortune II “Double Helix” rather vigorously. That’s why I haven’t been writing or posting a lot in my Blog.  Playing games is a way of escaping reality I guess. Things are not going so well at work at the moment and releasing the stress by killing some people (virtually of course) is one way of dealing with it.  The reality is that the company I’m working for at the moment is in deep financial trouble. We work a lot for advertising agencies and it doesn’t come as a surprise that they haven’t been spending a lot of money lately. In all reality we can say that the Dutch economy (as many others) is in trouble.  Because of that my company is going to restructure and in the same process they are merging with another firm. In fact it’s the 3rd time they reorganize in one calendar year. For me personally it’s the fourth time in one year after leaving Lost Boys in January 2002. On June 26th the people that will be fired will hear so. Since I have a 6 month contract (after having a year contract) they can simply not extend my contract and that is not the same as firing me; that means I will not get any unemployment benefits. A scary thought.  But back to SOF2 as it is called. Since the end of 2002 I’m playing this game rather fanatically and I also joined what is called a ‘clan’. Basically a clan is just a group of people that play online games (as opposed to the Klu Klux Klan whore are just a bunch of raving idiots). My clan is called Cool & Temper and our ‘tag’ (yes, I admit it either sounds childish or it sounds like being in a gang in downtown South Central LA) is *[C&T]*. People are supposed to put their tag in front of their name so people can see you’re a clansman.  I must say I rather enjoy it and you also meet a lot of people (mostly in the virtual world). I even bought a PC recently (a HP Vectra VL600) and converted it into a SOF2 clanserver so now we can play on our own server. All in all a pretty cool way of spending some time. Especially since I always like to tinker around with PC’s.  Well I guess we’ll meet again in June. Stay tuned!   '),(16993,1156278,'05,May,2003','2003-05-05 00:00:00','    Freedom and justice for all  Today is Liberation day in Holland. And although it\'s not a national holiday anymore (God knows why?) a lot of people have the day off to celebrate the fact that the Allied countries (mostly the US) saved our butt in WWII. There are a few open-air concerts today and of course the TV is filled with the usual movies like \"The Longest Day\" and such.  Looking out of the window of my work (its cold, clouded and looks like rain), listening to an oldie (A Forest, The Cure) on my iTunes, makes me a bit depressed. However I am working on some documents that need to be finished soon and the fact that I\'m alone in the office will ensure I\'ll cover some ground here today!  It\'s days like this that music is my inspiration and coffee my brother in arms.  Catch you on the flip side!   '),(16994,1156278,'29,June,2003','2003-06-29 00:00:00','    I did it again. I bought myself a new toy. While my girlfriend is on the other side of the ocean buying a house, I decided I wanted a digital camcorder to record our trip to Paris were we will be as of June 9th. We had a camcorder with us when we went to New York in 2001 (just a few weeks before the September 11th attack) but that was a NTSC camcorder that Carrie got from a collegue. Well, this time I bought a camcorder myself. I bought the  urlLink Sony DCR-TRV33E , a nice 1 megapixel MiniDV camera with a 2.5\" touchscreen, DV-in and out, S-VHS in/out, USB streaming, memorystick etc.etc.. All in all a nice package witch lots of options and the camera received good reviews.  I had my choice of camera\'s narrowed down to the Sony and the  urlLink JVC GR-DX300  which is one of those really small cigarette box camera\'s. The JVC got the camera of the year award from a magazine called Media Totaal and I must say I liked it a lot. It has a 3\" screen and weighs less than the Sony. It also has a 1.3 megapixel camera mode instead of 1.03. But the design in general seems less ergonomical than the Sony. Besides that it has the microphone on top of the lens and not in front of it which makes me believe that the sound would be less than with the Sony. Also, filming with the screen open will consume lots of energy, so my choice would be to film using the viewfinder. And the JVC only has a B&W viewfinder but the Sony has a color viewfinder. Besides that: when holding the JVC I found that one is very likely to cover the microphone (on top of the lens).  Together with the cam I bought three tapes, another battery (the stand one will allow you to film about 30-40 minutes) a nice camera bag, a UV-filter and (shame on Sony) a firewire cable that wasn\'t included. Also, I tried the camera with the brand new iChat AV from Apple (cool!) which turns the cam into a webcam. And the camera works like a charm with iMovie thats included with Apple OS X.  All in all, I\'m ready to enter the movies!   '),(16995,1156278,'28,June,2003','2003-06-28 00:00:00','    Ok, the 26th passed and I’m still employed. Al lot of my colleagues got fired though. Some I know pretty well. They offered me a new job as ‘client services manager’ – whatever that may be. They said they want to extend my contract with one year, so I guess that is positive.  Most people I know that got fired were from the sales department. Another guy was from the IT department that I work with and he was a ‘programmer’ (although he wasn’t any good at his work he was still kind of a nice guy).  I didn’t mention this yet in any of my Blogs, but I bought a 15” TFT screen about 2 weeks ago. I bought it from one of the sales guys at my work (Pieter). Well, he got fired too. So now when I work behind my PC I have a constant reminder of that. The TFT screen was only 190 Euro which is ridiculously cheap. I’m very happy with it since it makes it possible for me to actually watch my screen when the sun is shining into my office in the morning. Some people that got laid-off worked for more that 25 years with the company and one guy was even fired one year before he could retire early! It’s a dog eat dog world out there.  Some other news this week was that my girlfriend (Carrie) bought a house. I have no idea if I mentioned this before but she is an American and lives in Tallahassee, Florida. That’s also were she bought the house. Apparently it’s a cinder block house (we use bricks mostly and re-in forced concrete) with three moderate bedrooms, a small kitchen, two bathrooms and a swimming pool.  Needless to say that I’m very excited for her (in fact I’m sponsoring her a bit too). I’ve been in Florida (Tallahassee) quit a few times over the last years  - ever since I met Carrie - and I really love the US and Florida. Although it can get there a bit too hot for my likings. One time I had to change a flat tire and it took me over twenty minutes or so because I had to get into the shade every 3 minutes because I got fried out there.  One of my favorite places to go (and this is a tip for those visiting Florida) is  urlLink Wakulla Springs . It’s the largest natural well in the world and it’s really beautiful. Just look at  urlLink this picture   and also  urlLink this one  . Anyway: now I haven’t lost my job I’m pretty sure I will go to the US in December and help out Carrie with her home and visit Wakulla!  Ok, that’s all for now ffffffolkssss….   '),(16996,1156278,'23,June,2003','2003-06-23 00:00:00','    Were are my @#&*&*******!!! archives!   Ok I just noticed that for some odd reason my archives link on the left side doesn\'t work. Seems some javascript error. It is always something with Blogger!. Anyways just as a service I will post some links here:     urlLink January 2003    urlLink February 2003    urlLink March 2003    urlLink May 2003    urlLink June 2003      Hope they fix that Javascript error soon....   '),(16997,1156278,'23,June,2003','2003-06-23 00:00:00','    \"Slaves can\'t be fired: they get sold\"    As I mentioned in my latest blog my company is cutting jobs. This week, thursday the 26th to be exact, it\'s \"bijltjesdag\" as they say in Dutch. Roughly translated this means it is the day when people will get \'the axe\' - a nice euphemism for saying they wil get a pink slip. As I explained before I can\'t be \'fired\' in the legal sense of the word. They can just say they will not extend my current contract. In which case I will not get any unemployment benefits and will roughly have a month to find another job.   Because of this sword of Damocles hanging above each and everyone\'s head (in some departments every 1 out 2 people will be fired) the mood at work is all but great. I have great difficulties - like my collegues -  to motivate myself to go to work every day. Off course I have been looking around for another job ever since they announced the restructuring, but so far I had only 1 job interview and the sorry bastard never ever called me back (the guy was supposedly from a professional head hunters & recruiting firm but turned out to be all BUT professional).   Today\'s newspaper alone reported of more then 9100 lay-offs at the Dutch defense department (our \'army\'). And another 6000 jobs will be cut at some IT-firm here in Holland. Let\'s face it: the Dutch economy is in bad shape and it\'s not going to improve soon. That\'s for sure! The outlook isn\'t that bright and since I\'m 36 I better find something with a future soon! I do not feel like job-hopping till I\'m ready for my retirement (\"pensioen\" in Dutch).   Oh well. Recently I got some money from my savings account from the last 6 years at Netcast and Lost Boys and I\'m afraid that I will have to use that as a buffer untill I fiend something suitable. In between my girlfriend is trying to buy a house (which I finance a little as well), I\'m trying to get set for our vacation in July (we\'re going to Paris) and I am contemplating of buying a digital videocamera (which off course I will or will not do if the 26th truns out bad for me).   But enough of the complaining: at least I will see my girlfriend again in a few weeks time! Tomorrow I have to bring my car to the garage for its first 20k tune-up. For now I figure I might watch some TV of play SOF2 for a few hours... Anything to escape the thought of losing my job.... Escapism at its best (or worst).   Keep you guys posted!   '),(16998,1156278,'17,July,2003','2003-07-17 00:00:00','    Wow, I just had a few perfect days of vacation with my sweetheart. Between July 9th and the 14th we visited Paris, the city of lights. I took the car and we arrived there in a few hours (we left at 8am and got there at 2pm) driving through Belgium, taking the A1 toll road.  I really enjoyed myself and I think my girlfriend did too, although we didn’t have enough time to do everything we wanted. For instance, before we left I wanted to visit Mont St. Michel, a small peninsula on the west coast of France. But we never went there. One thing about the French: they do not like to speak English. With the exception of the few people working in the tourist industry they do not even try to speak English at all.  I know a few words French and I can even translate a few words, but when they start talking to you it is so hard to follow because they talk real fast. Luckily we managed and were able to do most things.  I had the idea to drive around Paris to get were we wanted, but once I got there I parked my car in a garage and never tried to get it out. We just took the busses and the subway to get around (besides that we mostly walked). Man, those Parisians drive like crazy. Years ago I went there with my school and we stayed for a week. I remembered from back then that no car in Paris is without dents or scratches on them (and I mean every car; even Rolls Royce’s). Well it was still the same!  They just park everywhere they please: even on zebra’s and on sidewalks. In fact we had a hard time going through with the wheelchair. Paris is very unfriendly for people with a handicap. It is a very good thing that Carrie can still walk and only needs to use the wheelchair after she walk’s a few miles. Otherwise it would have been impossible to do some things, like visiting Versailles for instance.  The roads there are filled with holes and are made out of cobblestones and that’s almost impossible to ride on with a wheelchair. As said we managed and so we had a great visit to Versailles, the Louvre, the Eiffel tower, the Sacre Coeur, Centre Pompidou and the Notre Dame and lots of other parts of town.  Most Parisians turned out to be very nice and even helpful some times. The prices for drinks however, are outrageous! A single glass of beer would range from 4,50 to 6,50 Euro (about 5 to 7 dollars). Food is about the same price as in Holland, so essentially they only screw you on the drinks.  And boy, we did need lots of drinks there: it was about 33 degrees Celsius (about 88 Fahrenheit) while we were there and I was drying out the first few days. Luckily we picked up some bottles and I kept myself drinking a few liters a day. Off course Carrie is used to these temperatures and I kept telling myself how lucky we were for having this weather. It was so hot that most fountains were turned off. Only a few were left running and that was a shame sometimes.  We left Paris on the 14th (a national holiday) and arrived in Almere around 5pm. The next day we spend some quality time and visited my family.  Finally my girlfriend had to leave on Wednesday morning, so we got up at 3:30am and left for the airport. I sure regret she couldn’t stay a few more days and miss her already.  Oh well. I guess I have to make arrangements to visit her this winter. Florida is always a nice place to be when it’s winter.   '),(16999,1156278,'30,August,2003','2003-08-30 00:00:00','    Want to listen to streaming audio? Do you like lounge, triphop, downtempo fuzed jazz and other music? Then check out  urlLink SomaFM , it is listener supported, commercial free Internet Radio. I listen to it at work, at home and about everywhere else where I have internet.  I really love Internet radio. You can listen to it almost anywhere (if you have internet) and the cool thing is that there are so many channels out there that are \"special interest\" which means a lot of so-called \"non-mainstream\" music. One of my other favourites used to be  urlLink Massinova  but they are offline now because the guy running it did this through his work and they stopped supporting him. Massinova was even cooler because you could request songs and you could keep a list of favourites.  Anyway: enjoy the stream!   '),(17000,1156278,'24,August,2003','2003-08-24 00:00:00','    This morning I had a close encounter with a real life bloodsucker and it nearly gave me a heartattack! Want to know what happened?  Ok, I was in my study editing some video that I shot way back in \'97 when I first visited my girlfriend Carrie in Florida. I was just making some nice scene cuts and adding a few effects, when I noticed the noise coming from my guest bedroom. It seemed as if my cats were tearing up the place or - which sounded the most logically - they were fighting each other.  So I yelled a few times in the hope they would cut it out. Since they still seemed to be jumping up and down and making lost of noises, I decided to get up and walk into the room and kick them cats out so I finally could go on editing my video in peace and quiet. And then it happend.  As I strolled into the room a little fluttering animal flew towards me when I least expectated it! it was a BAT outta hell! I wasn\'t prepared for it and it almost scared me into a heartattack! You see, I live on the 4th floor (3th floor if you count the floors as we do in Europe) of an appartment building and its in the middle of town. I never have flies, bugs, mice or any kind of critters around here - except my two cats. How it got in is beyond me: it must have crawled in throught the window that was open just a bit.  Man, did it scare me! Not that I\'m afraid of them (this one was just about 2 inches): I just wasn\'t prepared of finding a animal in there! My cats were going nuts about it, trying to jump it and running all around the room. So that was the sound I heard. The bat itself didn\'t make any noise and was just fluttering around trying to avoid the cats. As it would hang on to the ceiling (in itself an accomplisment) I tried to figure out how to get it out there. Bats are known to sometimes carry rabies and other diseases so I figured I could not just grab it.  In the meantime I got on the phone and called my sister to let her know of my close encounter. While on the phone the little bat flew in between the window and the blinds and my cats were desperately trying to attack it. Once on the window ledge it was quite helpless since the can\'t walk but just crawl and slide a bit. Anyway: I figured I had to get it out quickly before my cats would hurt it.  The problem was solved quickly: when the bat was hanging from the ceiling again, I went up there, opened the window and it went out just like that!  But boy did that little sucker scare the hell outta me! I had enough exitement for this sundaymorning :-)   '),(17001,1156278,'18,August,2003','2003-08-18 00:00:00','    Mondays really fuck up my life (pardon my French). I get up in the morning,  slowly forcing my body to actually move to the bathroom. Looking back at me on  the other side of the mirror, I see some freak with small eyes, exploded hair,  yawning like I\'m trying to get a new world record. O yeah, it’s Monday alright.   Monday tells you 2 things:   A. Your weekend is over B. Work, Work WORK!  A typical Monday for me looks a bit like this:  7:45am -   Starting up movement-coordination brain cells 8:05am -   Starting shaving feature (trying not to bleed myself to death) 8:15am -   Dressing up (damn, my shirt is wrinkled again) 8:20am -   Feed my cats auto-drive system kicks in 8:30am -   Make sandwiches for work 8:45am -   Starting driving ability system to commute between Almere - Amsterdam 9:20am -   Boot up my Apple Powerbook G4 Titanium (first smile of the day) 9:30am -   powering up coffee-mechanism control center 9:35am -   Life support enabled (after 2 coffee - with milk and 3 sugar) 10:00am - Main act-react system online 10:30am - Speech enabled (more or less) 11:00am - Finally I wake up mentally, I find myself sitting at my desk. I’m at work, Arrrggh...    And I just know millions of people go through this same routine like me!    I\'m *NOT* a morning person. I can stay up late (4am) and sleep all day. When I\'m  behind a PC I can sit there for hours forgetting the  world around me. A social life? Where can I download that? Let\'s face it, I am a  sorry ass GEEK!   '),(17002,1156278,'15,August,2003','2003-08-15 00:00:00','    CNN.com reports:  urlLink Texan, 91, arrested for bank robbery - Aug. 14, 2003 . I guess you are never too old to be a fool.  Apparantly it wasn\'t the first time he did a robbery:  Rountree left a prison in Florida, where he was the oldest prisoner in the state, about a year ago after serving a three-year sentence imposed on him for a 1999 bank robbery in Pensacola.   He was caught holding up a bank in Biloxi, Mississippi, in 1998 when he was 87, and given three years\' probation.   Police said he always used the same tactic -- walking unarmed into a bank with a note saying \"robbery\", then trying to flee in a getaway car parked outside.   In a prison interview with the Orlando Sentinel in 2001, Rountree said he had been a businessman in Texas but had fallen on hard times.   He said he robbed his first bank when he was about 80 because he wanted revenge against banks.   \"A Corpus Christi [Texas] bank that I\'d done business with had forced me into bankruptcy. I have never liked banks since,\" he told the Orlando Sentinel. \"I decided I would get even. And I have. Banks are the easiest things in the world to rob.\"   Police said Rountree, who was not armed, asked a teller to stuff money into a large envelop with the word \"robbery\" written on it.   A witness took down the license number of his vehicle and he was arrested on a highway about 15 miles outside Abilene. He did not resist, police said.   If he is convicted, Rountree faces from two to 20 years in prison.   '),(17003,1156278,'15,August,2003','2003-08-15 00:00:00','    Recently I bought a T-shirt from Think Geek ( urlLink \"I see Fragged People\" ), but now they added a new new one that is kinda funny:  urlLink ThinkGeek: I\'m blogging this.   Put the world on notice: I\'m blogging this :-)   '),(17004,1156278,'15,August,2003','2003-08-15 00:00:00','    Ha, sometimes I suprise myself. I already booked my next vacation and I also got my tickets this week! This time I decided not to mess with renting a car and driving over to my girlfriend; I just transfer in Atlanta and fly to Tallahassee in a hour! I fly on December the 22nd and return on January the 4th 2004.  The suprise this time came with my tickets: a booklet with all kinds of info on what I can take with me and what not. Also they made sure that insurance doesn\'t pay more than 10-20k in case you die due to a terrorist attack..  Ever since september 11th flying has become such a hassle. First the US government requires you to agree to give all kinds of personal information and they also ask that from the airlines. With it they can see who you are en what you are doing.  The idea that they can see what kind of meal you had on the plane (for me total irrelevant) is just too much. Big brother is watching you, that\'s for sure!   '),(17005,1156278,'15,August,2003','2003-08-15 00:00:00','    urlLink CNN.com - As power slowly returns, the questions begin - Aug. 15, 2003   The above article is a nice example of how authorities always try to blame others for mistakes and problems.  Since I\'m not old enough to consiously remember the poweroutage in new York in \'77, this one is having more of an impact.  It was really weird seeing all the footage from a blocked up Times Square, people running around with flashlights and candles.  Good to see though that - once again - New Yorkers are not in a state of panic.  The power outtage in New York comes at a moment that we just had some outtages here in Holland as well. This was caused by a heatwave we had and people were using way to much electricity and the power plants had problems cooling their turbines.  However: it does match up with the scale  we had them in Holland.  I do agree though, the big question is: how did this happen?   '),(17006,1156278,'26,September,2003','2003-09-26 00:00:00','    Talk about viral marketing! Recently a friend of mine told me to check out some new software that makes it possible to make VoIP calls (VoIP = Voice over IP) to anyone on the internet. Now, this has been around for years already but usually its a big hassle to set it up  especially if you have a firewall our router.  Without getting to technical; VOIP (which is also possible with MSN Messenger, Netmeeting etc.) uses so called port numbers to communicate (a port is an endpoint to a logical connection and the way a client program specifies a specific server program on a computer in a network). Usally the ports are blocked by your firewall or router. And if you have more than 1 PC behind a router, it\'s even almost impossible to use VOIP programsbecause of problems with NAT (Network Address Translation).  Besides that: usually these programs use some central server (i.e. Netmeeting) to find other users. Now there is a new program called  urlLink Skype , from the makers of KaZaa which promises a zero-configuration VOIP, P2P telephony solution that should work with any firewall and should be completely hassle-free. Off course I was eager to find out if it works and I downloaded it immediately.  As the makers promised, the installation was completely transparant: even the usual tweaking of the microphone and headphones didn\'t occur. Just a straighforward install. Then my firewall (I installed it on 2 of my 4 PC\'s) asked if the program was allowed to connect to the Net and off course I answered yes. Ok, so fa so good. Next I called the guy that told me to get it. Now he lives in Holland so it wasn\'t a real long distance call with lots of traffic on the net.  To my suprise the quality was as Skype promised: crystal clear with no delays or echo\'s. Next I called two other guys I know: one in Sweden and and in the UK. Again: no delay\'s and a crisp, crystal clear sound. We even experimented with talking to each other while downloading some huge 40meg file. Although we experienced some little hickups (much like with a GSM phone) the quality still was excellent. Later I talked to a guy that has a 64k ISDN dial-up link and it still sounded perfect. According to the makers of the program it should work even on a 33k3 modem connection. All in all: impressive, even though this is still just beta-software.  In conclusion I would like to recommend people that want to save money on their long distance phonebill to get this program. If you have a broadband connection thats on 24/7 anyway, just download it and be amazed by it! Its worth your try.  Oh, and did I mention yet that all calls are fully encrypted against evesdropping?   '),(17007,1156278,'09,September,2003','2003-09-09 00:00:00','    CNN reports this morning that urlLink 261 music file swappers are sued  but also that an amnesty program has been unveiled. The federal lawsuits and amnesty program are the latest moves by the Recording Industry Association of America (RIAA) in its fight against illegal music file trading on the Internet. According to them filesharing is responsible for a 31 percent drop in compact disc sales since mid-2000.  So what do you do? First they tried to put the companies out of business that were making programs to share (like Napster). They did succeed in that; most filesharing sites were using a central database with users and it was easy for the RIAA to take measures against that. But then they didn\'t count on the ingenuity of programmers; they found a way that made it very difficult to track the people: peer-to-peer (read this article for an explanation of  urlLink P2P networks).   But now they also declared war on P2P and they are - for the first time - trying to sue individuals. Off course they know that this is just scare tactics (it\'s impossible to sue millions of downloaders); and it also is the kind of tactics that you\'d expect from an industry that is corrupt and rotten! If you want to know more about the RIAA and the way they conduct themselves, I suggest you visit this website:  urlLink mp3 is NOT a crime.org .  For instance, according to the article  urlLink \"CDs and the Scarcity Principal\"  the RIAA is making prices artificially high by both controlling distribution and production.   However the RIAA is really ridiculous when it comes to explaining what and how they do to track people.  For instance: the US media has bought the RIAA\'s \"ingenious\" and \"surprisingly technical\" approach to identifying how to tell the difference between mp3 files which were obtained from legally purchased CDs and those which came from \"pirated\" sources.  What they say is this: \"By comparing the fingerprints of music files on a person\'s computer against its library, the RIAA believes it can determine in some cases whether someone recorded a song from a legally purchased CD or downloaded it from someone else over the Internet.\"  Duh?  Since a digital copy of an mp3 is NOT a perfect copy of a CD track (in spite of what the industry has led so many to believe) once it\'s been created and placed on the Internet, each recurring copy of this file WILL be an EXACT duplicate of the original mp3. If this were not true, the RIAA\'s statement would fall apart at this point in the logic. If it were altered in any way, this entire wild theory about hashes and digital fingerprints would be quickly proven false.  I don\'t want to be overly technical here, but ask yourself this: If the mp3 was not created from a legally purchased CD, where the hell DID it come from?  What can the origin of an mp3 file be if no one possessed the physical CD? Magic? It just appeared on their computer one day? They hacked into the recording studio\'s computer? Broke into the studio in the dead of night and stole the master recordings?  If an mp3 was not created from a legally purchased CD, then there\'s only one other possible source - the record label itself, if not the artist.  Do the $4 billion a year in free promotional copies have different digital fingerprints that the legal CDs? Can you tell a shoplifted CD from one which was legally purchased?  Unless the recording industry itself is responsible for creating mp3s from the master recordings and distributing them over the Internet, ALL mp3s MUST, by definition have come from purchased CDs. To entertain any other definition is really dumb.  Anyway: I could go on and on, but I suggest reading this funny analogy called: urlLink The MPAA, RIAA, and Fellatio .  And to all RIAA jokers out there: GO F* YOURSELF!!   '),(17008,1156278,'08,September,2003','2003-09-08 00:00:00','    This weekend I almost bought a MP3 player: the  urlLink iRiver iHP-100 . This neat little machine at first seemed some good competition for the Apple iPod. It is very nice looking and has some interesting features. For instance it has built-in radio, 16 hours playtime (or so they claim), optical in and output and you can record with it. It also supports most current formats including WMA. They sell it here at the local Media Markt for about 475 Euro. It seemed a good deal.  But now there is the brand new  urlLink Apple iPod  in 10-20 and a whopping 40 Gig!! It offers more space for less money... Way less money...  I mean: compare it: 10gig for 475 or 40 for 499!  What the iPod lacks thought:  - built-in radio  - WMA support  - Digital i/o  - 8hr play vs 16hr?   Anyway, I\'ll have to rethink it a bit... But I do love the looks of that iRiver. Especially the cool blue display (on both the player and the remote).   '),(17009,1156278,'31,October,2003','2003-10-31 00:00:00','    I have only one thing to say: If you visit this page you know again why women are so wonderful: urlLink Gostas Pouco Gostas .  Ps: this page may load really slow, but believe me: this guy has good taste!   '),(17010,1156278,'26,October,2003','2003-10-26 00:00:00','    Last September I wrote that I almost bought the iRiver iHP-100 MP3 player. One of the problems I had with it, was the fact that was just 10 gig. Well low and behold: there now is the  urlLink iRiver ihp-120  which packs 20 gig of discspace. They just released it on oktober 14th. Here is the  urlLink CNET iRiver iHP-120 Review .  Also, the price tag was a bit steep (at least here in the Netherlands): Euro 475,-. Yesterday I walked into the Media Markt (a huge electrical appliances store here) and saw they lowered the price of the iHP-100 to Euro 399,- which still is way to much since the iHP-120 is priced at U$ 399,- (list price; so you can get it for lower prices than that). With the current Euro - Dollar rate it is just not interesting to buy me this baby in the Netherlands. Although, there is a Dutch website that claims to be able to sell the player for 449,-  Euro (see:  urlLink MMport . The lowest price I have seen on a US site is $365,95 but then they do not have it in stock.  So maybe I just have to wait till I go to the US and buy me one there ;-)   '),(17011,1156278,'20,October,2003','2003-10-20 00:00:00','    It is out now:  urlLink Max Payne 2: The Fall of Max Payne : PC game play at its best! I just bought it yesterday and I played it already for hours! This game is truly a great combination between action and cinematics. If you played the previous Max Payne, this sequel is a must buy!  If you were diasppointed in the Matrix game, you now have a chance to see what \" bullettime\"  really was meant to be! This game was the first one to really capture the action style that made the movie The Matrix so great. The storyline is that of a film Noir and just adds greatly to the gameplay! This game was one of the reasons for me to buy a faster and better videocard and PC!  If you have never played it: buy Max Payne 1, play it and then get part 2.  Payne to the Max baby!   '),(17012,1156278,'20,October,2003','2003-10-20 00:00:00','    If you read my Blog you might have noticed that I installed a pre-release of OSX Panther earlier this month. Well, it had some problems with it that I didn\'t anticipate. For instance, my browser (Safari off course) didn\'t accept any input from my keyboard anymore (which is a bit of a problem when typing an url) and it also ignored my scrollwheel on my mouse. Also it seemed that my bluetooth was completely gone. Other than it seemed to run smooth and I have to say that it also fixed some annoying problems (more on that later).  Off couse since Panther comes with Safari 1.1 it could not just reinstall Safari or bluetooth (its part of the OS). Luckily the final release (or \"Goldmaster\") came out a few days later and was spread very soon thru the new fast filesharing protocol called  urlLink \"BitTorrent\" . And low and behold: after I installed that version I had no problems with Safari anymore; is working smooth as ever. My bluetooth however is buggy as hell. I\'m aleady addicted to Expose.  A lot has changed and its not only the obvious stuff like the look and feel of windows. For instance; if you close System Preferences it really closes (it stayed open in previous versions). The control panels (in system preferences) have changed to. Also Panther has a better way to switch applications (apple-tab) and together with the new versions of DVD player, Safari and others its really an improvement! Not to mention things like File Vault (encryption of your user files), Font Book (view your installed fonts),   Most impressive however, is the fact that somehow it seems they tweaked the OS so it even runs smoother on the same hardware! Thats right: with \"that other OS\" you need new hardware whenever they change to another version. With Panther it seems that my PowerBook G4 667 mHz runs better and not slower!  And they finally fixed one big annoying thing! With previous versions of OSX my PC would lock-up if I forgot to disconnect from a network server and just put my PowerBook in its bag. Now you just get a error and my G4 doesn\'t lock up anymore! Hurray!!   '),(17013,1156278,'12,October,2003','2003-10-12 00:00:00','    Say you have shelled out some $30 for the DVD LOTR Two Towers and you want to loan out that DVD to your neighbour. Wouldn\'t you be mad when they return it with scratches on it? I rent a lot of DVD\'s and when you get them they look like crap. It is as if people use them as a frisbee or a coaster or let their dog chew on it. That\'s the reason I just make a bakcup of my DVD\'s and give the copy to whoever wants to see it. That way I make sure my DVD\'s stay in perfect condition.  Now, you ask yourself, how the heck would you make a backup of a dual layer DVD that can have about 9gig of data on it to a single layer DVD that has about 4.22gig of data? Well I discovered it is pretty easy. What you need is a program that can extract the movie and then processess it (i.e. shrinks it) so it wil fit on a single layer DVD-R/+R/-RW/+RW. For that reason I bought  urlLink DVD2one . It is a kick-ass program that will copy either the full DVD (with all menus) or just the movie to just one single layer DVD-R/+R/-RW/+RW. And the best news: this program is available for either Windows users and Mac OSX users!  The catch is that you need a pre-processing tool like  urlLink DVD Decrypter  to \"rip\" the DVD to your harddisk, since DVD2one can\'t do that, and also need a tool to burn the DVD like  urlLink Nero Burning Rom . But the combination of these tools is way better than the programs that supposedly do all-in-one.  The benefits of this wonderfull piece of software are:    World\'s fastest video compression engine. Process a movie to fit on a single recordable in a matter of 15 to 30 minutes (depending on the speed of your computer).   Makes movies-only copies. Why waste space and image quality on unnecessary intros, menus and language tracks?   Makes also full disk copies. For those who want to keep those fancy menus and intros.   Powerful join mode. Useful for putting multiple episodes of a (tv) series or joining both sides of a flipper on 1 disk.   Written from scratch, it\'s not based on any old code, libraries or existing opensource projects.   Time critical parts written in assembly, for the highest possible speed.   Burn the movie with your favorite recording software!   Both PC Windows and Mac OS-X versions available.    So, if you have the same idea about leaving your precious DVD\'s intact, I would say: shell out the money for DVD2one and never be afraid to get your DVD back as a coaster ;-)   '),(17014,1156278,'09,October,2003','2003-10-09 00:00:00','    People who have read my Blog before, know that I have a very nice Apple Macintosh Powerbook G4 from my work. I had a choice between a regular laptop or the slick and sleek G4 - running OSX which is basically a Unix flavor OS. Well, for a technology freek like me that was a clear cut choice. I got the Mac.  I have been a real fan of OSX from day 1. It is by far the best OS in terms of stability for a laptop I have seen so far. I have run many different OS systems from DOS, WIN 3.1 up to XP Professional. I also like to experiment with Linux, which I think is growing into a great alternative OS for Intel based PC\'s.  Well, Apple is steadily making OSX into a prime OS for their PC\'s. So off course I\'m very keen to get my hands on the new version of OSX: Panther. And \'ve got it! As I type my G4 is running in the installer in the background and I can\'t wait to start testing it. It\'s not the official release yet, but it is Build 85 and supposedly it\'s one of the final release candidates.  I\'ll keep you posted on my experiences!   '),(17015,1156278,'09,November,2003','2003-11-09 00:00:00','    So far this weekend seems like its been an early Xmas! It\'t 10:45 in the morning and while I\'m making this Blog - sipping some fresh java - I\'m enjoying my brand new superfast ADSL... Thats right: this weekend I already got my new fast ADSL. And when I say fast, I mean FAST. Lightning fast... Blazing saddles fast... I\'m talking warp 10!  Ok, you think: this guys is over-reacting. Well, my ADSL is 8(!)megabit downstream and 1(!)megabit upstream. And I downloaded a Linux .ISO image that was 660megabyte in just 16(!)minutes last night! When I had cable it took me at least 1hour and 20 minutes to do that!  The deal is I wanted to be able to run my own servers. I had Internet access by cable but then I was not allowed to run my own servers or even connect more than 1 computer to my cable (UPC). Then a friend of mine said he was getting rid of his cable (UPC) because he was getting ADSL from  urlLink Demon Internet . He said he was going to pay half what he paid for with UPC and in return he would be twice as fast! So natuarally I thought I\'d check how and what the deal was. So, then I saw that for just 10 Euro a month more than I pay right now for my cable, I could get the fastest ADSL possible.  I enrolled 2 1/2 weeks ago and figured it would take at least 3 weeks or so. But when I checked the track & tace pages I saw they send me my ADSL modem already and that they also approved my line. And low and behold: yesterday morning (saturday) I got my modem, splitter etc...  So last night around 6pm I figured: what the heck, let\'s put that thing together. So I managed to install the splitter (a bit of a hassle) before my ISDN line (I have ISDN) and when I was ready doing that, I figured: let\'s see if the modem works. When I proceeded to connect my modem, it worked fine. But also I noticed that suddenly the light next to the ADSL label was burning... Well, supposedly I should have to wait for my phone company to actually activate my ADSL. But as it seemed it was already activated! And when I connected the modem to my router, I did get an IP address! Wowie!  So now I have superfast Internet!!  Like I said: it\'s like an early Xmas this weekend!   '),(17016,1156278,'08,November,2003','2003-11-08 00:00:00','    Oh man! Next week it will be there! my  urlLink iRiver iHP-120 . It\'s waiting to be shipped as you can see at that link ;-) Boy, next week is gonna be a good gadget week 4 me! Also I got news that my ADSL (8megabit downstream!!/1megabit upstream!!) will be ready by the end of november!!!  Happy, happy joy!   '),(17017,1156278,'07,November,2003','2003-11-07 00:00:00','    Mac users all over the world usually hate Microsoft and vice versa. For a more indepth and - mostly - objective insight check out this website: urlLink Mac OS X vs. Windows XP . They provide a truly indepth view on this matter. No aspect of both OS\'s are missed. I would make it the \"Geek-link of the week\".   '),(17018,1156278,'03,November,2003','2003-11-03 00:00:00','    Call me a gadget freak or a technology addict, but I did it again. I ordered a iRiver iHP-120 MP3 player and in anticipation of it arriving (this or the following week), I also bought the Sennheiser PXC 250 noise cancelating headphones ;-) that I want to use with this beauty. For a review of this gadget, check out the  urlLink Sennheiser PXC 250 review by PC Magazine . As I write this (at my work - mind you) I\'m listening to a soundtrack from the movie \"Trainspotting\" (Dark Train by Underworld) and I have it cranked up with the noise cancelation turned on. It blocks out the irritating sounds of the AC at my work more than adequate! So I bet I will be enjoying it when I fly to the US later this year!  Only bad thing about it is that I probably paid a bit to much for it, since I know see all kinds of adds where it sells for about 10-15% less than what I paid :-( Anyway: back to the iHP 120 that I ordered. I found it was way too expensice until I stumbled upon a way to buy one cheaper. The makers of DAPreview did some communal purchase action and so I got the opportunity to buy it for at least 60 Euro less than the cheapest store here in Holland. I know: if I import one from the States, it might be chaper, but (import)taxes will even out the advantages.  Anyway; I keep you posted on when I get it! In the mean time enjoy  urlLink this comparison between the iHP-100 and  the Philips HDD100 . from DAPreview...   '),(17019,1156278,'06,December,2003','2003-12-06 00:00:00','    Ha, caught you off guard huh? You\'re wondering: what the hell does that mean? Well, for those nog familiar with Dutch traditions, instead of Santa Claus, we have \"Sinterklaas\". And guess what: basically they are the same! You even pronounce the names almost the same.  I don\'t know the exact origin of Santa Claus or Sinterklaas real well, but supposedly he was a figure that lived in the 4th century a.c. in Myra, Turkey as a bischop. According to the legend he saved his village from starvation, resurrected three kids that were dead, and gave gifts to the poor people. He supposedly died on December 6th 343 a.c. In 1087 his remains were moved to Bari in Italy (but for some reason in Dutch tradition he came form Spain).   The fact that he gave gifts etc. became a tradition that we Dutch celebrate in December 5th; \"pakjesavond\". Kids that behaved good all year get presents and the ones that were not so good are supposedly put into a bag and get whipped with a stick and then are taken back to Spain by the black helpers of Sinterklaas (\"Zwarte Pieten\").  Anyway: this goofy tradition has been around for many years and almost every Dutch kid grew up with this myth, only to discover that Sinterklaas doesn\'t exist and it all was just a big scam.  So how did the Americans wind up with Santa Claus? Well supposedly an American preacher called Clemet C. Moore described the red suit that \"Sinterklaas\"  was wearing, in a poem called \" \'t Was the night before Christmas\"  and thus introduced the Dutch legend in the United States of America. And in 1773 the Dutch introduced him as St. A Claus.  So, the truth of the matter is that the modern Santa Claus is a conglomeration of sources, a legendary being that has evolved over the years. Along the way, pre-Christian legends, the story of St. Nicholas, Dutch immigrants to America, Washington Irving, Clement Moore, Thomas Nast, and the Coca-Cola company, all made their contributions.  If you want to read a very good indepth piece on it, check out:  urlLink The Origin of Santa Claus and the Christian Response to Him .   '),(17020,1156278,'01,December,2003','2003-12-01 00:00:00','    Wow, it has been an extremely busy month of November here at my work. So busy I didn\'t even have time to follow up on my story from November 9th. I didn\'t even have time to post my experiences with my iRiver iHP-120 or my superfast ADSL. All because because I was so wrapped up in my work with deadlines and very demanding customers.  On the bright side: it\'s now December 1st and I will be with my honey in exactly 21 days ;-) But untill then I still have a busy schedule since I have to wrap up some stuff at my work.  But back to the iHP-120 ;-) I\'m listening to it right now; some track called Trance Nation 2002 in MP3 format. I have about 10 gig of music on it and I still have 10 left so I can really use it to the fullest. I put all my favourite music on it and I have room left for some regular files. It\'s very handy that I can use it as a portable harddisk as well. with the USB 2.0 interface transferring files is a breeze and even with USB 1.1 it\'s pretty acceptable.  Anyway: the last week was a low point as far as it goes with health etc. Even though I didn\'t call in sick I had a bad cold and I was having a slight fever. I guess that pure willpower kept me on my feet. It could not have been worse timing for me to be sick, so i decided to just take a few extra aspirins and bite the bullit. I haven\'t called in sick this year, except for 2 days in July when I came back from my Paris trip.  Anyway: it\'s 10 past 17:00 and I\'m ready to pack up my stuff. Tomorrow I have some dorky training from my work (some sales-training) and I already know I\'ll be in a coma after the first hour or so. I better go prepare for that mentally.  Blog you later!   '),(17021,1156278,'26,January,2004','2004-01-26 00:00:00','    It\'s a grey monday morning and I\'m sitting at home behind my PC. I\'m not at work because I have a rather irritating problem: I can hardly walk because I hurt my left foot. So what happened?  Well, since two weeks I\'m jogging/running again. I signed up for this as part of my news years resolutions. The goal is to run the 10km on May 2nd when we have the marathon of Almere. I haven\'t done anything fysical for the last 2 years so off course I have no stamina whatsoever. We train two times a week: wednesday evening and sunday morning. Last wednesday we did a so called \"interval training\" which is meant to enhance your stamina. Afterwards my foot was already hurting a bit, but I knew I had 3 day to recover, so yesterday morning I just started to run again.  I had a few hestitations but I started out running. We were supposed to run for 30 minutes, but after 15 minutes I just had to stop. I could not run anymore and even walking hurt like hell. Afterwards I could barely drive home cause even pushing the clutch hurt me. Luckily its only a few minutes drive.  However: I bought a brand new printer ( urlLink Lexmark i3 ) saturday, but it turned out a DOA. So I had to  go exchange it. The shop were I bought it (Media Markt) luckily has a huge parking lot next to it so I didn\'t have to walk that far. After taking a few painkillers (generic over the counter stuff) I just embarked and went to exchange it.  I always am a bit hestitant to complain about pain and such since it\'s so relative (my girlfriend has chronic pain), but let me tell you; you really appreciate the fact that you can walk and run once you feel how much such a simple task can hurt. So when I got home and installed my printer, I was happy to put my foot up and start relaxing.   '),(17022,1156278,'06,January,2004','2004-01-06 00:00:00','    Well, it seems a long time ago that I “blogged” something. And you’re right. Since December 6th I haven’t had much time to do anything an now I just returned from a 14 days vacation in Florida with my girlfriend Carrie.  Right now its 0:06 and I have a bad jetlag. This is the reason I’m writing this entry in my nice Blog. I guess I will tell you guys all about my trip later. For now I’m getting a melatonin capsule and hope that it will help to put my biological timer back on track.  Lucky for me I have the week off so it’s not that bad a start at all.  Oh, did I wish all you guys & girls out there a happy new year yet? I didn’t? Well: here you go:  Happy New Year!   '),(17023,1156278,'29,February,2004','2004-02-29 00:00:00','    Today it was February 29th; a leap day in a leap year. And for a lot of people the only time in 4 years they can celebrate their birthday. A  leap year is a year in which an extra day in added to the calendar in order to synchronize it with the seasons. Since the tropical year is 365.242190 days long, a leap year must be added roughly once every four years (four times the fractional day gives ). In a leap year, the extra day (known as a leap day) is added at the end of February, giving it 29 instead of the usual 28 days.  In the Gregorian calendar currently in use worldwide (except perhaps the Russian and Iranian calendars), there is a leap year every year divisible by four except for years which are both divisible by 100 and not divisible by 400. Therefore, the year 2000 was a leap year, but the years 1700, 1800, and 1900 were not. The complete list of leap years in the first half of the 21st century is therefore 2000, 2004, 2008, 2012, 2016, 2020, 2024, 2028, 2032, 2036, 2040, 2044, and 2048.  The extra rule involving centuries is an additional correction to make up for the fact that one extra day every four years is slightly too much correction (). This scheme results in the vernal equinox gradually shifting its date between March 19 and 21, being shifted once every leap year, and then being abruptly shifted in non-leap centuries.  Confusing? Well I know I was when I read this.  Anyway: I will not have a chance to write my Blog on Februari 29th until 4 years from now: 2008. Wow. I\'l be 41 then.  Some trivia: In any ten year period, what\'s the minimum number of leap years? What\'s the maximum?  Answer:   Minimum: 1 (2097-2107 would only have one leap year, 2104! 2100 is not a leap year (as explained before). Maximum: 3 (1996-2006 would have 1996, 2000, 2004)   Also: 2004 is a leap year, with 29 days in February. February has five Sundays - it starts and ends on a Sunday. Between 1904 and 2096, leap years repeat every 28 years which means that the last time February had 5 Sundays was in 1976 and next time will be in 2032.   Have a nice leap year!   '),(17024,1156278,'26,February,2004','2004-02-26 00:00:00','    Ok, this is a Dutch only blogpost... It\'s about the chaos on the Dutch roads following some snow. Only in februari 1999 there was more chaos: 975 kilometres of traffic jams.   Sneeuwval leidt tot extreem lange files    Uitgegeven: 26 februari 2004 07:25  Laatst gewijzigd: 26 februari 2004 11:43     DRIEBERGEN -  De hevige sneeuwval heeft donderdag tijdens de ochtendspits tot extreem lange files geleid. Omstreeks 08.00 uur stonden er 63 files met een totale lengte van 860 kilometer, aldus de meldkamer van het Verkeerscentrum Nederland. Alleen op 8 februari 1999 was de verkeersellende nog groter: 975 kilometer file.   Ook de Nederlandse Spoorwegen ondervonden veel hinder, vooral bij de hoofdstad. Tegen achten liep het treinverkeer naar en van Amsterdam Centraal volkomen vast.   Na enige tijd kwam het langzaam aan weer op gang. Een woordvoerder van ProRail sprak van \"een blizzardachtige situatie\", waardoor op veel plaatsen wisselstoringen ontstonden. \"Het is bar en boos.\"    Schiphol   De luchthaven Schiphol ondervond ook hinder. Om 07.00 uur kon slechts één baan worden gebruikt. Twee uur later waren alle drie banen weer in gebruik. Tijdens piekuren heeft Schiphol drie banen nodig. Intercontinentale vluchten hadden het minst last van de vertragingen. Wel waren er toestellen die naar andere Europese luchthavens moesten uitwijken en vielen er vluchten uit.    Files   De langste files stonden op de autoweg Amsterdam-Den Bosch. Richting hoofdstad stond er omstreeks 08.30 uur 57 km file tussen Deil en Holendrecht en in omgekeerde richting tussen Holendrecht en Culemborg 43 km. Ook op de A12 tussen Den Haag en Arnhem was het motto langzaam rijdend en stilstaand verkeer. Tussen Veenendaal en De Meern stond 39 km, tussen Gouda en Lunetten 35 km. Om half tien was de totale lengte van de files bijna gehalveerd tot 388 km.    KNMI   Het KNMI had een weerwaarschuwing uitgegeven. In heel het land zou het door sneeuw en bevriezing glad zijn. Die waarschuwing geldt ook nog na de ochtendspits. In de loop van de dag kunnen er nieuwe sneeuwbuien vanaf de Noordzee het land binnentrekken.   De weggebruikers hebben zich volgens de ANWB goed aan de omstandigheden aangepast. Tot 08.50 uur hadden er volgens de KLPD 75 aanrijdingen op de hoofdwegen voorgedaan. De ANWB meldde dat de Wegenwacht het niet drukker had dan normaal.   De KLPD gaf omstreeks half zeven het advies thuis te blijven als dat kon. Door de sneeuwval waren op dat moment de wegen praktisch onbegaanbaar. Korte tijd later werd dit advies ingetrokken.  Original message form www.nu.nl:  urlLink nu.nl/algemeen | Sneeuwval leidt tot extreem lange files   '),(17025,1156278,'14,February,2004','2004-02-14 00:00:00','    Well It\'s one of those days again: one inch of snow, slippery roads and miles and miles of traffic jams. Obviously I am not trying to get to work since even the Dutch police caution everyone \"to \"stay home if possible\". Luckily I\'m in a position that all I need for my work is an Internet connection and e-mail ;-)  Well mostly: sometimes I have to admit that being in the office is more handy, but let\'s just say that for now I\'m happy being home: I\'ve turned up the heat, have some fresh coffee and some lounge music in the background...  For those wondering how it looks: check my webcam at  urlLink Zion.demon.nl .  Blog you later!   '),(17026,1156278,'14,February,2004','2004-02-14 00:00:00','    Happy Valentine\'s Day to all you out there that want that one special person in their lives to know you love them!  For me that special person is  urlLink Carrie Tyree . Carrie, I\'m proud to know you and I love you for the person you are :-)  Now, if only someone could take away the +10.000 miles that separate us.  Peace to all out there!   '),(17027,1156278,'08,February,2004','2004-02-08 00:00:00','    Ok, I know what I want for my birthday! I want the ultimate GEEK wear: a ScotteVest, and more specific the  urlLink ScottE TEC Sport Jacket . Yes thats right, this blazer makes James Bond look like a moron.  I always have found it hard to carry all my stuff with me. At least women have a handbag they can stuff with their toys. But we men are always stuck out there with our wallets, cellphones, pda\'s, MP3 players, sunglasses and what have you. And if you stuff your regular blazer - like I do - you will find your pockets start ripping like nothing before.  So this baby is a real good invention. In fact  urlLink ScotteVest  has much more functional and innovative wear. Yep, it seems that that even President Bush has one. Normally a reason NOT to get it, but hell, for once he had a good idea!  Now I only need to figure out what size I need...   '),(17028,1156278,'05,February,2004','2004-02-05 00:00:00','    Ok, here is a quick note. If you are looking for a CMS to maintain your website, you should definately check out  urlLink opensourceCMS . It is a website that has all major opensource CMS\'s listed and you can all try them out for yourself.  Best of all: all the CMS\'s are free!   '),(17029,1156278,'30,March,2004','2004-03-30 00:00:00','    It\'s tax season and I need to finish my tax form by April 1st. Luckily in Holland our \"Belastingdienst\" (the Dutch IRS) is fully automated so I can just fill out some electronic tax form and just push a button: presto!  When I compare this with other countries (i.e. the USA) we are much more automated. For instance electronic banking: I have been doing that for 7 or 8 years now. My girlfriend can\'t do it that way yet (at least not as easy). Weird because the concept of electronic banking and using the internet for it, is a concept that came from the USA...   Movies galore   Right now I\'m in a movie copying frenzy: I have lots of DVD\'s lying around that need to be copied. Ever since I got my new PC with DVD burner it really made my collection grow. Right now I burn most DVD\'s first on DVD-RW(+/-) so I can see if it is working right, then I just burn it to a DVD-R (or +R).  Lined up are: Triggermen, Master and Commander, Kill Bill Vol. 1, Den of Lions, Blind Horizon, Crime Spree and last but not least Matrix Revolutions.  Speaking of Matrix Revolutions: it really sucked! I mean: why the hell did they kill both heroes? The whole movies series was evolving around the love story between Neo and Trnity! Ok, they saved mankind (Zion), but why not end the series with some positive message where boy gets girl? Although: the silly thing is, that they do leave open the possibility for Neo to return (afterall; it is not all clear if he died or not and even the Oracle says they \"very probabaly\" will see him again).  Anyways: the effects were awesome but the ending did kill the good feeling I had about it.   '),(17030,1156278,'26,March,2004','2004-03-26 00:00:00','    A word of advice: if you plan on using Norton Systemworks 2003, do *NOT* install Norton Goback. If you mess up your system, it will make thing WORSE than actually help you!  I just spend the last 4 hours trying to fix my sisters and brother in law\'s PC that they bought only 3 weeks ago. When they bought it, I loaded it with stuff like Norton and some DVD burning tools. I also installed Goback. I figured that it would give better protection in case my sis (or her kids) would mess up their PC.  However, this tuesday the PC crashed (XP Home edition) and after that it started to get into a vicious circle of rebooting, starting and the rebooting again. Off course they tried the restore option that Goback suggests. However that just didn\'t work and they called me.  So when I inspected the PC I saw that after Goback gave its prompt (press space for restore options), it just didn\'t start and would reboot again. First I suspected that somehow the MBR got damaged or just the NTloader wouldn\'t work.  So I booted from the XP cd, went into the recovery console and just used the FIXBOOT command and then the FIXMBR command. That didn\'t work and by doing so I also wiped the special Goback bootstrap. I didn\'t know it yet, but that just put me into more shit. Somehow the driveletter were XP was appeared as I:\\ and not C:\\. This started me to suspect it was that damn Goback because 1 time I had the same problem with a W2K machine that had Quarterdeck Bootmagic installed.  Using the recovery console I figured out that the volume (not the partition) wasn\'t accesible anymore. I then used PEbuilder to make a special XP boot CD with some diagnostic tools but again: I could not acces the volume. Even after making it the active volume.  Then I decided to just use Partion Magic to examine the harddisk. Well there it was: the harddisk wasn\'t accesible for this program as well. I decided to look up the fysical info for the harddisk: it turned out to be a type 44.  So, using google I figured out what was going on. Goback loads its own MBR version and also changes the identifier for the harddisk from 07 (NTFS) to 44 (RAW). And neither XP or most tools can access it! Or so they think because in reality its just the identifier and it\'s not really type 44 disk.  Now all I had to do is change it back to 7. But how? I looked again with google and came upon an article: it seemed that a tool called ptedit.exe could do this. And I had it since I have Partition Magic! Praise the Lord!  So about 4 hrs and a lot of sweat later, I managed to fix the problem simply by two clicks and a OK button, and it just restored XP as it used to be with all files etc. intact.  Needless to say I immediately removed Goback. Norton really fucked up there! They should not mess with the damn harddisk identifier!  You have been warned!   '),(17031,1156278,'21,March,2004','2004-03-21 00:00:00','    Ok, just a quickpost for today. As many people know, I like to experiment with my PC\'s. When reading some PC magazines I stumbled upon a article on how to set up your own Shoutcast streaming radio station. Off course I could not resist and I just set up my own station:  urlLink Radio Zion: Lounge 24x7 . Just a big mix of electronic & lounge music. Something to relax by. You will need a player capable of streaming audio like Winamp.  Just try it: you might like it.  I cannot guarrantee It will be online by the time you try this ;-)   '),(17032,1156278,'14,March,2004','2004-03-14 00:00:00','    As some might know I have been trying to make a more attractive website for some time. My HTML knowlegde is very, very basic and I really don\'t like to mess with it too much. I have been using Frontpage but it doesn\'t produce attractive looking webpages. So I am still trying out some stuff and I could use some help. I do have some Ideas but how to put them into a design and then into webpages?  Anyhow: in my search for some cool resources I stumbled upon this website:  urlLink FreeFoto.com  they provide free (for non-comemrcial use that is) photos for your website. I already have downloaded some photos since I do have this idea about a city-like (or urban) look for a website.  Now, if only I could find a webeditor that has some cool designs and is WYSIWYG in nature.. Any tips on this?  email me.   '),(17033,1156278,'14,April,2004','2004-04-14 00:00:00','    This morning, on my way to work I got caught up in slowing traffic near the point were the A6 motorway changes into the A1. While I was slowing down I noticed the red X above one of the lanes and before I even could think aloud \"probably an accident\", I got passed by an ambulance zooming by with all lights and sirens on. Obviously the accident just happened. After a minute or so I got close to the spot where it all happened. Just 100 yards before I got there, I could hear the dreaded sound of a helicopter, wich means there was a severe trauma going on and the trauma team got called in.  When I finally drove by the spot itself I saw three cars piled up but that wasn\'t the reason why the trauma team was there: about a hundred feet futher a motorcycle was piled up into a heap of scrap and near it the medics from the ambulance were holding the driver\'s head stable. Although I only saw it for a second it seemed that the guy still had all his extremities attached (once I drove by a motorcycle accident where the guy got decapitated). So it was unclear how much damage he had.  It send some chills down my spine: I never liked motorbikes just because you are so damn vunerable if anything happens. No airbags, no cage surrounding you, no rollbar etc. etc.. sure: it might be a fast way to get form A to Z and during summer it probably is one of the coolest ways to get around (besides a convertible of cabriolet as we call it), but the chances of surrviving an accident are very slim.  The problem is that the last few weeks traffic is getting slower (hollidays are over and people are returning from their sky trips etc.) and the weather better. So lot\'s of people are getting their motorbikes out and try to elude the traffic by zooming by. A lot of the drivers are unexperienced or take to many risks trying to pass the traffic jams too fast.  Well, I saw the result of that this morning. Again.   '),(17034,1156278,'08,April,2004','2004-04-08 00:00:00','    I just heared a very amusing story from a collegue that travels a lot to the US. Last summer he went to visit some relatives in Pensylvania. He met some other Dutch people on the plane that traveled with him from the airport to this little town. On their way, in a rental car, they accidently ignored a stop sign and got pulled over by a squad car.  The officer asked his traveling companion, who was driving the car at that moment, for his drivers licence. He also asked him why he thought he got  pulled over. Since in Holland you can ignore a stop sign when there is no traffic coming, he answered that he didn\'t know. When the officer explained the situation (you always stop at a stop sign even when you don\'t see anyone for miles around), the guy aplogized to the officer: \"Sorry hoor, I didn\'t knew that\".  The cop got agitated and asked him \"What did you just say?\"  so again he said: \"Well, I said: sorry hoor, I didn\'t knew that it was illegal.\" The cop got even more angry now and asked him to step out of the vehicle. By now his traveling companion started to wonder what the hell was going on and why tyhe cop was getting angry.  Then suddenly he realized it: \"sorry hoor\"  is a typical Dutch expression. When you pronounce it in a certain way (depending on the Dutch accent you have) it sounds like: \"Sorry, WHORE\". which is, off course, pretty insulting!  Luckily he could explain it to the cop who had a good sense of humor and could even laugh at the situation.   '),(17035,1156278,'04,April,2004','2004-04-04 00:00:00','    This past week Wednesday evening (March 31st) turned out to be a scary evening. After I went out for dinner (Chinese) I arrived at my home at about 8:45pm. I changed clothes and slipped into my training suit and started to zap through the channels (nothing on as usual). Then at about 9:15pm my doorbell was ringing. I could here from the sound that someone was actually at my door and not downstairs (it rings differently and my camera turns on). So I went to see who it was.  There was my next-door neighbor from number 109 (I\'m 111). She was standing in the hallway with her girlfriend who was holding her baby (about 9 months old or so I estimate) and she had this panicky look on her face when she gasped out of breath: \"Fire! I have a fire in my kitchen, what do I do?\". She was standing there holding her cell phone which still had the power plug attached to it, dialing 112 (the Dutch equivalent to 911).  When I looked behind her into her hallway I could see and smell the smoke coming out of there, but there were no flames - yet. I didn\'t panic at all but asked her what it was that was burning. She explained that they smelled something and then when they looked they saw the kitchen was filled with smoke and it came from a little pan she had left on with some oil in it.  Anyway I told her that it was best to go outside and warn the other neighbors on her way out. I went back into my apartment, think up what I could do. Grab some belongings? Get my cats? Obviously if the fire would go big I could do nothing about it. I also called 112 - just to be sure they got the call - and after that  I just unplugged all my computers, my TV stereo and other stuff and got my shoes on, grabbed my mobile and my keys and went outside.  On my way out I passed some cops (they patrol on bikes here) and told them they should go to 109. They also started to bang on the doors and evacuated everyone. Two minutes later I was outside with about half of the other people from my apartment (12 families live there).   The firemen arrived and just rolled out there hoses and went in. From the street I could not see any flames at all so in had a feeling it wouldn\'t be to bad. After all: we cook electrical here and she didn\'t have a vent above her stove. Usually those vents catch on fire and make things worse. A few minutes later I saw them open some windows and also the kitchen window. I could see the ceiling had turned all black from the fire.  They had everything under control but I could not go back yet: they examined the air first and after another 20 minutes or so I could go back in again. Boy did it STINK there!  Luckily no harm done no casualties and I could start up everything again....  Off course I am going to install some smoke alarms and check on my insurance!    Coffee Junkie  Ok, I admit it: I\'m a coffee junkie. I have been using a regular coffee machine for some time now, but if I just want 1 cup I have to use instant coffee which I dislike. So this week (thursday April 1st) I decided to treat myself and I bought the latest invention around. A  urlLink Senseo Crema machine .  Great coffee and you can get a cup within 3 minutes! Downside is I do drink more coffee now and I suspect it\'s not so healthy.   '),(17036,1156278,'30,May,2004','2004-05-30 00:00:00','    Check out this cool blog: urlLink DUTCH HOPTO .  Its about some Dutch guys that plan to travel through the US in the ultimate pimpmobile (my own personal favourite car): a Lincoln Continental Mark V.  Peace brothers!   '),(17037,1156278,'05,May,2004','2004-05-05 00:00:00','    Just some Dutch Blog I spotted and like: urlLink Zomaar uit verveling ... . Some Dutch female Blogger (\"Ladyhawk\").  I\'m thinking about linking to more blogs that I think are interesting.  Blog u later!   '),(17038,1156278,'05,May,2004','2004-05-05 00:00:00','    Ok, I just finally (after a lot of shit) got my FTP working again for my Bloggers@large. So, I decided to also put up my profile. You can see it by clicking this link: urlLink Blogger: User Profile: Henk .  Not that it is very interesting, but hey: one needs to fill ones pages ;-)  Anyway: I will probably post some more stuff today. I have a lot of stuff to write/contemplate on.   '),(17039,1156278,'05,May,2004','2004-05-05 00:00:00','    I just realized that Blogging (before it was called like that) has been around for quite a few years. In fact, they were called .plan files (\"dot plan\").   A .plan is basically a file which a user on a UNIX system (well, generally UNIX or Linux systems) can put into their \"home\" directory. This file is there so that the user can describe themselves, write a log, keep a list of things they have to do etc. Using a utility called \"finger\" it is possible to have a server return the contents of this file by passing it the user\'s email address (i.e.: \"finger john@idsoftware.com\")  During the development of the PC game Quake, programmer John Carmack of id Software started using his .plan file to keep the users updated of his progress. This idea snowballed until the whole id development team were using them, and many other developers followed.  Many important events have been first announced in .plan files, from staff changes and child births to the first mention of new games (like Quake III Arena or Doom III).  Since this so called \"finger\" protocol got out of fashion (mostly because it was targeted a lot by hackers) it isn\'t used that much anymore and steadily got replaced by .plan files that were published on the WWW and subsequently called \"blogging\". But thats just a new word for the same thing in my opinion.  A typical .plan from John Carmack looked like this: John Carmack Feb 23 2000, 18:41:51 (GMT) johnc@idsoftware.com id Software - Lead Programmer 	  Name: John Carmack Email: johnc@idsoftware.com Description: Programmer Project: Quake 3 Arena ------------------------------------------------------------------------------- 2/23/00 ------- This is a public statement that is also being sent directly to Slade at QuakeLives regarding http://www.quakelives.com/main/ql.cgi?section=dlagreement&file=qwcl-win32/  I see both sides of this. Your goals are positive, and I understand the issues and the difficulties that your project has to work under because of the GPL. I have also seen some GPL zealots acting petty and immature towards you very early on (while it is within everyone\'s rights to DEMAND code under the GPL, it isn\'t necessarily the best attitude to take), which probably colors some of your views on the subject.  We discussed several possible legal solutions to the issues.  This isn\'t one of them.  While I doubt your \"give up your rights\" click through would hold up in court, I am positive that you are required to give the source to anyone that asks for it that got a binary from someone else. This doesn\'t provide the obscurity needed for a gaming level of security.  I cut you a lot of slack because I honestly thought you intended to properly follow through with the requirements of the GPL, and you were just trying to get something fun out ASAP. It looks like I was wrong.  If you can\'t stand to work under the GPL, you should release the code to your last binary and give up your project. I would prefer that you continue your work, but abide by the GPL.  If necessary, I will pay whatever lawyer the Free Software Foundation reccomends to pursue this.   '),(17040,1156278,'30,June,2004','2004-06-30 00:00:00','    A dream has ended... Portugal beat us 2-1. We didn\'t even score ourselves (a Portugese player scored for us)... It was an exciting game, I was on the edge of my seat, but they were to strong for us. Portugal, who is hosting the European Soccer tournament really had the benefit of playing in their own country. Almost every decision the referee made was in the advantage of the Portugese :-(  Oh well, back to normal again. The Portugese will face off with winner of the match between the Tsjechs and the Greek...   '),(17041,1156278,'26,June,2004','2004-06-26 00:00:00','    Wow, another barrier taken: we beat Sweden! After a breathtaking 90 minutes we still had a Mexican stand-off. So we had 30 minutes more playtime (extension). After that it was penalties. If there is one other trauma in the Dutch soccer history, its taking penalties. The last 4 times we had to do this we lost. Basically our team sucks at penalties. So everyone in Holland was holding their breath when we had to continue.  But this time we we kept it  cool. Cooler than ice. Our goallie, Edwin van der Sar, was undoubted the hero of the match when he stopped one of penalties from the Swedes. I was watching the match with my sister and brother-in-law and it really was very exciting. The new generation of players did not suffer from the traumas from the past. Van Nistelrooy, Makaay and Robben all did their job. The only one missing a penalty for the Dutch was Cocu: he hit the goalpost.  Pff, next we have to beat Portugal in the semi finals on the 30th...   '),(17042,1156278,'23,June,2004','2004-06-23 00:00:00','    It\'s been a stormy day here in Holland. Literally: a rare summer storm sweeps across our nation today. Just as our Dutch soccer team swept the Latvians of the field: 3-0. But, as I mentioned before: this time it was not in our hands: if the Germans would win from the Tsjechs, we would loose anyway. But today the impossible became possible. Even though the Tsjechs played with, what is considered a complete B-team, they beat the Germans 2-1!  It could not be closer than this. And to tell the truth, it was a nervewrecking match. As always the Germans had a very good comeback in the second half of the match. In Holland its a well known fact that the Germans are never beat until the whistle blows and the referee says its over. I couldn\'t help myself zapping from the Dutch match to the Germans, every ten seconds or so.  It is only fair to say that, in all reality, the Latvians never should have been a problem for the Dutch. They played even poorer than the Dutch throughout the tournament. But in reality the Dutch haven\'t played that well either. Last Saturday they finally played a bit like they should, but we still lost. Today we dominated although there were a few scary moments. Particular after the break when the Latvians had three big chances.  When I zapped back and forth between the two matches I got more and more anxious - to my own surprise. Usually I\'m not a huge soccer fan, but this kind of important match is always a different experience.  Having said that: if I would have to make one prediction, I would predict that it would not be surprising at all if the Tsjechs will become the next European Soccer Champions of 2004. They do have a hell of a team and a really smart coach.   '),(17043,1156278,'18,June,2004','2004-06-18 00:00:00','    Today is \"our\" last chance as a soccer nation to go for the title of European soccer Champion of 2004. After a 1:1 stand-off against our German \"friends\", we played a good match against the Tjechs. Even though we started with a 2:0 advantage, we managed to loose the match 2-3. Mostly due to a very questionable player change from our Dutch coach \"Dick\" Advocaat.  So today we need to win big from Letland if we want to proceed to the quarter finals. However, if the Germans win from the Tsjechs, we still loose because we won\'t have enough points in our Group D (we only have 1 point, number 1 are the Tjechs with 6 points).   In other words: half of the outcome is chance, half of it is our deal.  I, for one, will not shed a tear when we loose. The way I see it we do not deserve to go on to the quarter finals since we have not displayed any outstanding or special qualities. The big suprise of this tournament are the smaller countries that normally do not score to well. I\'m talking about Letland, the Tjechs and even Portugal.  Looking at the scores so far (not that I am an expert), my prediction is that the semi-final probably will have the French, English, Sweden and the Tsjechs...  Too be continued....   '),(17044,1156278,'18,June,2004','2004-06-18 00:00:00','    Haha, a collegue just send me this great Harry Potter spoof from Saturday Night Live: urlLink \"Horny Potter\" .  You need Windows media player to see it ;-)   '),(17045,1156278,'16,June,2004','2004-06-16 00:00:00','    Ok, this is NOT a joke, I repeat: this is NOT a joke! Remember Harry Potter\'s invisibility cloak? Well, a professor at the University of Tokyo has created an optical camouflage system that makes anyone wearing a special reflective material seem to disappear. Here\'s how: a video camera records the real-life scenery behind the subject, transmits that image to a front-mounted projector, which then displays the scene on the reflective material.  It\'s the darnest thing I ever have seen: It\'s like the special effects you see in movies for real! I admit I thought it was a hoax or a joke, but as you can see for yourself at the following links it really exists! Off course the technology is still in its early stages, but anyone can see that its just a matter of time before it really will be applicable (the japanese professor thinks it will be around 2008).  Check it out:   urlLink TIME Magazine: Coolest Inventions 2003, Invisible Technology .   urlLink cool video\'s of the invisability cloak .  Again: reality seems to catch up with fantasy... What\'s next?   '),(17046,1156278,'14,June,2004','2004-06-14 00:00:00','    Let\'s face it: Microsoft Internet Explorer is the standard browser for 98% of this world. But there are a lot of issues I have with MSIE in regards to what other browsers have to offer. For instance: I\'m completely addicted to tabbed browsing (something I started to like once I tried out Opera). Also, built-in ad-blocking and pop-up blocking (also something Opera offered years ago) is a nice option.  For those looking for an alternative to MSIE, look no further: just download  urlLink Avant Browser, a tabbed browser with Flash Filter, Popup Stopper, Cleaner, Web Search and more.  It just does what I want and adds some good options. It\'s just MSIE with smart features! And best off all: it\'s free! No ad-ware no nags, nothing.   '),(17047,1156278,'06,June,2004','2004-06-06 00:00:00','    With the European soccer cup in full swing, some memories pop up about the famous – or should I say: infamous – World soccer cup final of 1974. Ever since WW II there has been some animosity between the Dutch and the Germans. It’s the same kind of animosity that exists between the British and the French (although their history goes farther back then ours).  Back in ’74 the Dutch soccer-team definitely was the underdog of the tournament. We never got far and usually we stranded way before the semi- or quarterfinals. But that year it was different. We had the best teams in years (most notably we had Johan Cruyf, the Dutch soccer player who is as famous as the Brazilian Pele was in his days). We managed to make it into the semi-finals in which we defeated the Brazilians (the number 1 favorite for the world title).  We had a complete new system/concept in playing, which was dubbed “total soccer” that also involved a new rule that we used to our benefit regarding “off site”.  Anyway: the final was a thriller of epic proportions: Netherlands vs Germany. Even though I was only 7 years old at that time, even I still remember the outcome of the match.  After just 3 minutes we got a penalty against the Germans and scored. So it was 1:0 for us. We played liked lions but in the second half the Germans managed to get a penalty as well: 1:1. And, in the last minutes of the game they got another goal and we lost 1:2.  We all felt like the Germans had stolen out title! The Dutch team played at a constant pace from the start of the tournament and had proved that the system and the players were a perfect match. Our coach, Rinus Michels aka ‘The General’ aka ‘The Sphinx’ had turned a bunch of players into a machine that ran perfectly.  All in all it was a trauma the Dutch would never loose and even 30 years later there are hordes of people that still remember the ordeal.  It wasn’t until the European Soccer Cup of ’88 that we regained a bit of our self-esteem by becoming the European Soccer Champions.  If only we could become the World Champions!   '),(17048,1156278,'06,June,2004','2004-06-06 00:00:00','    I just spend another weekend being totally absorbed by my computers. It all started when I wanted to upgrade my main game computer to Norton System Works 2003 and also installing Norton Internet Security Professional.  Now, I have had my fair share of computer problems with Norton (if you can’t remember check the March 26 blog called  urlLink Norton Goback, Go Away ) and this wasn’t going to be any different. First I just tried to upgrade. This didn’t work well. Then I tried to uninstall all Norton stuff. That didn’t work well either.  After going to the usual registry hell, I just gave up and tried to go with the free-ware versions for my firewall and virus scanner. But that gave some other troubles. I used AntiVir-personal (you should too: it’s for free and it works good), but for some damn reason it just didn’t install right.  It took me a few hours literally to hack it. Now it’s running and so is ZoneLabs ZoneAlarm.  Bah! I just hate PC’s sometimes. Word of advice here: stay away from Norton 2004. 2003 versions work better and with less hassle!   '),(17049,1156278,'27,July,2004','2004-07-27 00:00:00','    Why is it always so hard to start working again after a vacation? I came back this Saturday from Florida, and starting this Monday was pretty annoying. I kind off walked in like a zombie and even three mugs of coffee later, I didn’t feel very alert.  It also turned out that some people I work with suddenly are on a holiday too, even though they said they were not going for a while. Annoying too, because its not like I can decide on all kinds of things without conferring with them first.  Anyway, tomorrow its Wednesday and we are at the middle of the week: only two more days to the weekend. By then I should have my blank DVD’s (I ordered 100 since I ran out of them) and also my Maxtor OneTouch 250 firewire/USB 2.0 external hard disk. I ordered it today because I really need to backup my PC and want to store some files.   '),(17050,1156278,'25,July,2004','2004-07-25 00:00:00','    Jet lag is a bitch. Flying back from Carrie is always the worst; I loose a day and it feels just like that. When I arrived yesterday, I felt pretty ok. I slept for one and a half hour and stayed up till 11:30pm which is good I think. But this morning I feel like I have a bad hangover. I got up at 10am after sleeping real good (I only woke up twice) had a large coffee, but still it feels like my engine hasn\'t started.  I guess what adds to the jetlag this time, is the fact that I had a 9 hour lay-over at Atlanta. So basically I was up at 5:45am and we rushed to the airport (God, I hate saying goodbye) and I arrived at Atlanta at 8:30am and had to wait till 5:40pm. So I walked around the airport (Atlanta is huge) and shopped around a bit (bought myself yet another watch, this time a nice Nautica).  So once I got on the plane I was pretty tired. As always I don\'t sleep real well on a plane, I kind of doze off for a few mintues I guess. They showed two movies I enjoyed: \"Fifty first dates\" (so funny!) and \"Man on Fire\". Both were good entertainment.  Anyway, I just put my  urlLink vacation photo\'s online  online for you guys to enjoy. Gee, I do like New Orleans but next time we should go there when its not +98 degrees (36 celcius). Then we can go see more of the town without melting like some popsicle ;-)  Starting work tomorrow again: brrr, I don\'t feel like that at all. I better chill out some more today ;-)  Blog you later!   '),(17051,1156278,'20,July,2004','2004-07-20 00:00:00','    Car trouble (1)   &nbsp;  The last week of my vacation here started with some anxious moments. My baby was supposed to go to work at 8:30 am, but less than 2 minutes after she left her home, she pulled back in to the carport. The car was running real weird and it didn\'t even want to stop running when she turned the key off.  &nbsp;  A little panic struck me (and her too I\'m sure). One bad thing is, that I know absolutely nothing about car mechanics. I do understand how they work and stuff, but whenever I pull the hood open I can barely distinguish the battery from the other parts. Sure, I know how to fill up the screenwash or put some more oil in there, but other than that I\'m pretty lost.  &nbsp;  Now, my brother-in-law on the other hand is a bit of a car wizz. Lets put it this way: what I know about computers, he knows about cars. He is rebuilding an old BMW 3-series right now, and he really is doing a great job.  &nbsp;  So, when Carrie pulled up there I could only give her a sheepish look and try to think what the heck might be wrong with the car. I checked (so I thought) all the fluids, and they seemed ok. Now there was one little thing that should have warned us: the engine had been running pretty hot and when I pulled the hood open (before we left for New Orleans), I saw some smoke coming of the engine (oil burning). Also, there was a huge puddle of what looked like water and oil beneath the car under the carport.  &nbsp;  However: since we checked the oil before (when we left for New Orleans) it seemed like that could not&nbsp; be oil, but radiator fluid - or something. Carrie told me that the puddle started to grow when she had her oil changed at Superlube some weeks ago. We talked a couple of times about it over the phone and I suggested she would check if the bolt that is underneath the car was screwed on the right way.  &nbsp;  Anyway, we decided it was best to have the car towed to Action Auto and have them look at the problem. I feared the worst, since it sounded like a major problem (maybe a belt or something). So after waiting for more than one hour for the tow truck, the guy asked us what was wrong with the car and checked the oil again.&nbsp;And sure enough:&nbsp;there was no oil left in the engine.  &nbsp;  Now, as everyone knows, that is a sure way to kill your engine. So we could only hope that this wasn\'t the end of the car. Luckily Carrie drove it only for 2-3 minutes. Any longer and it would have died for sure. So after the tow truck left she took Zach\'s (her son) car to work.  &nbsp;   Mosquito cost   &nbsp;  While waiting for the verdict on the car, I mowed&nbsp;Carrie\'s front lawn. Since I had taken a nap for 1 hour (hey, I\'m on vacation you know) it was about 12:30 when&nbsp;I started to do that. The middle of the day and probably the hottest moment. Boy, I think I lost a gallon of sweat there. The grass was pretty high and the lawnmower - that otherwise works great - died on it about 20 times. And sure enough: the mosquito\'s really started to follow&nbsp;me around and were trying to eat me alive. They even follow me around while I\'m in the pool!  &nbsp;  Funny thing is: I never ever had a problem with mosquito\'s - in Holland that is. Ever since I visited Florida for the first time however, it seems that I am allergic for the American version of this pesky little critter. And what a drag it is: once bitten the bite mark turns into a huge purple/red spot and itches like crazy. I know a thing or two about how mosquito\'s work and I am sure I am allergic to the stuff they inject into my bloodstream in order to prevent my blood from clogging up. It works to good I guess, so that\'s why it looks so bad.&nbsp;Not to mention&nbsp;it itches like hell.  &nbsp;   Car Trouble (2)   &nbsp;  Finally Carrie called me and told me what the&nbsp;verdict was on the car. First she tricked me and said it was 4500 dollars to get it fixed, but then she explained that in actuality it was two bolts (or nuts) on the engine that weren\'t tightened up enough, so the motor had been leaking oil all the time! The big puddle under the car, after all, was oil. The charged some 80 dollars for fixing it. Not bad, considering things could have been a lot worse!  &nbsp;   Indian Dinner   &nbsp;  The day ended with us having a&nbsp;nice dinner at Curry and Wine. \"Us\" included Carrie\'s girlfriend Denise. Denise used to work at the same place Carrie is working. While Denise was having some vegetarian food, Carrie and I chowed down on some chicken curry. I had the spiciest version of the Chicken Korma, but compared with what I am used to it was actually the mildest dish I had in a while. Nice, but not as spicy I was into.  &nbsp;  Afterwards we dropped a video at Blockbuster (Mystic River - excellent movie) and proceeded&nbsp;&nbsp;home to watch the movie Monster. One of the saddest movies (next to Mystic River and Monsters Ball) I have seen lately. And so ended the day.   '),(17052,1156278,'17,July,2004','2004-07-17 00:00:00','    I just came back from New Orleans after a three day visit there with my girlfriend Carrie. Boy, it was a blast although way too hot for me - and suprisingly also too hot for Carrie (her being from Florida)..&nbsp; &nbsp;  To our suprise New Orleans looks a lot like some European cities we visited. Mostly I could see a resemblence to Paris in some ways, which off course isn\'t surprising since New Orleans exactly means wat the name says: its&nbsp; a&nbsp;new version of the French city Orleans. &nbsp; Anyway, we went&nbsp;around town, had dinner in a fancy restaurant&nbsp;(Oliviers), went to Acme Oysters, seen the Missisippi river, the Audubom Aquarium and&nbsp;last but&nbsp;not least, we gambled some money away in Harrah\'s Casino. All in all it&nbsp;was a fun time and we&nbsp;stayed in a great Hotel (Iberville Suites) for a good price. &nbsp; I will be putting up&nbsp;some pictures from the trip shortly.   '),(17053,1156278,'11,July,2004','2004-07-11 00:00:00','    Ok guys, I\'m ready to go to the good oll\' US of A! I packed my bags, have my carry-on full with gizmo\'s and I\'m ready to put on the sun tanning lotion! I\'m just wrapping up some PC stuff over here and I will be on the road (air, actually) for some 15+ hours crammed in a plane.  Off course I got to have some tunes on the go. So thats why I have my iRiver IHP-120 packed. I just love the chill-out tunes and old seventies stuff. So If you have a chance, tune in to my  urlLink Radio Zion  Internet radio station. You now can make requests as well! One of my latest additions has been this cool retro-sounding soft-porn album \"Schülmadchen Report\":   urlLink       It is so funny.   Ok, got to go! Blog you later!   '),(17054,1156278,'07,July,2004','2004-07-07 00:00:00','    I guess everyone remembers that feeling you got when you were a little kid and it was just 1 day before your birthday and you were waiting for that 1 gift you wanted. That feeling of anticipation that gave you little butterflies in your tummy. Well, even though I\'m 37, I still get that feeling from time to time. Today it\'s because I am expecting the arrival of a new toy/gadget: a brand new digital camera.  After holding back 4 years, I finally decided to give in and buy myself a piece of the new digital photography era: an  urlLink Olympus C 765  4 megapixel, 10x zoom camera. I was looking for the right balance between options and usability, meaning I want to be able to take the no-brainer snapshots, but also be able to control everything. I looked at some ultra-compact camera\'s too, like the Pentax Optio S4i and the Sony Cybershot DSC T-1 and the Canon Ixus 430 and 500.  See, my problem is that I used to my Olympus OM-2N SLR and I did my own developing and printing for a while. But going on a vacation and taking all that equipment with me, just wasn\'t practical. And ever since I saw the first digital camera\'s and how easy it is to digitally retouch your photo\'s, I knew that one day I would have to go digital too.  So I was looking for a good compromise between options and ease of use and stumbled onto this Olympus (which I think is funny, cause a few years ago I considered them to have lost the digital race).  Update: while I was writing this the postman rang twcie and delivered the little box! Time to unpack my present ;-) In a few hours I will update my blog again....  In the words of Ren & Stimpy: Happy, happy, joy, joy. Happy, happy, joy, joy.   '),(17055,1156278,'04,July,2004','2004-07-04 00:00:00','    Well, I guess I should not try to make any predictions anymore. The winner of the EU Soccer Cup is..... Greece! I would have bet on the Tjechs to win, but I would have lost (Portugal beat the Tjechs). The amazing thing is that Greece is like a soccer lilliputter. They never ever got into any final and never ever won a tournament before. They were always the first country to get out of the tournament, but somehow they stumbled through this EU cup. And now they beat the Portugese 1-0. Just one lucky moment to decide a match.  The Greek are really going out of their mind now and parties are all over the place (in Greece that is). And I just read that some Greek guy who lives in England won 400.000 Euro\'s by betting on the Greek. They were 80-1 when they started the tournament and this guy must have had psychic powers or something to pick the most unlikely winner of all!  There is no second place, winner takes it all.   '),(17056,1156278,'01,July,2004','2004-07-01 00:00:00','    Only 8 more days, then I\'m on the plane to visit my honey ;-) Two weeks vacation to try and forget all the stuff going on at my work. Last summer Carrie (thats her name) visited Europe and we went to go see Paris. This year we will go on a (short) trip to New Orleans. I am very much looking forward to this.  A lot of fun goes into looking for information on this nice town. Well, I did find a nice website (albeit a bit shallow) called: urlLink Things to do in New Orleans . Carrie lives in Florida, Tallahassee to be exact, and thats a pretty nice town too. I love the old buildings there and the scenery. This year I will make sure that I take enough photos for my website.  The first time we met we went on a trip to Pensacola beach and I think that the buildings and architecture I saw there will be pretty much the same as in New Orleans. Oh well, guess I have to go see for myself ;-)   '),(17057,1156278,'01,July,2004','2004-07-01 00:00:00','    What nobody expected, happened. The Greek beat the Tsjechs and will face the Portugese this weekend in the 2004 European Soccer cup finals. Its amazing because the Tjechs so far had won every game and the Greek never ever before even got into the quater finals, let alone the semi finals!  Its going to be an interesting match for sure: the Greek who have nothing to loose and the Portugese that are organizing the tournament and are playing for their own countrymen.  I don\'t dare to do any predictions for this match... We\'ll just have to see   '),(17058,1156278,'06,August,2004','2004-08-06 00:00:00','    Alrighty than! With the help of some new cool tools I can now turn my Blog into a pictureblog as well. I will share some of them here in the near future.  So keep a look out here and you will see some cool stuff!  In case your wondering: get you software from  urlLink Hello!  and from  urlLink Picasa . If you have a Blog its fee!   urlLink    Hi Guys! Just saying Hello! Now I can send some pictures too. If possible I will publish some cool pictures in the near future.   '),(17059,1156278,'06,August,2004','2004-08-06 00:00:00','    Yes! It\'s FRIDAY again. Thank God, another work week is over! The fact that it\'s friday always is good for lifting spirits at my work. Also, since it\'s vacation time everyone is either  gone or about to go (the running gag when people come to our department is: \"Look here, this is our tech-department. Currently everyone is on a sabattical\") it\'s a bit more relaxed here than usual.  We have had two tropical weeks so far, with temperatures in the 28-34 centigrade ranges (82-93 degrees) and this weekend will probably be like that as well. At my home its about 27/80 degrees. Enough to make me sweat when I wiggle my big toe. Funny thing is that last year around the same time, we had a heatwave as well accross Europe. That was the year that Carrie and I were in Paris. A lot of older people in France died that year.  Given the fact that we are all talking about global warming I find it strange that so few European households have airconditioning in their homes. Especialy the old folks homes. In the US you find an airco in almost every home. So if the whole theory about global warming is true (which I doubt, but I\'ll provide a link on this subject later), chances are we will have some more heatwaves in the near future. And if that is true, why not install more airco\'s?  I, for one, am very happy with the airco in my car. Really: I get in the car with the same big smile as I get out.  Oh well, enough about the heat. It\'s friday and I will talk to my sweety again tonight. Another highlight of the day. So all in all its gonna be a good start of the weekend.  Cheerio!  ps: I keep getting questions from people what a Blog is. Well, to be short, its a abbreviation for WeB LOG. I consider it to be more of a online diary, but as I mentioned in a previous blog, in my opinion it originated from people using a .plan file.   '),(17060,3490604,'29,May,2004','2004-05-29 00:00:00','    life\'s awsome when yur having fun and happy...but is it really all that worth it in the end?.... my answer to that is no.. and ill tell you why  these 2 past years ive changed alot.  I havent focused on God enuff, ive been just trying to live the life that everyone expects a teenage gurl to live.  Hanging out with different people, some good and some bad, haveing fun and joking around.  These 2 past years have been awsome, laughter and joy have been the main two things.  But.. whut about God? did i forget him along the way?.... i mean i prayed when i needed him most, to make hard desisions and stuff.. but i never really had a proper relationship with Christ. I always knew that i was a Christian and that i believed in The Lord, i knew that Jesus Christ was our savior and i aknowledged that.. but i didnt relli care.  Jesus pointed out a good thing that teaches me alot about how he wants us to fix our priroities.                                            ****Jesus****!!!!....                                            **!Others!**....                                            *n then yourself*  My prorities were a little different than that.  in my world i wanted to be with my friends and one speacil friend all the time and have fun with them. (which was relli kool) school was next, then sports and then.... God (if i had extra time).. i look at this order and i think WHY?!.. why wouldnt the Father of our world be put first!!!?? without him i made wrong desisions and did some really bad things.. my parents didnt relli help either. they wanted me more than ne thing to get straight As in school, they could have cared less if i was without Christ. But i cant blame it on them.. GOD SHOULD HAVE BEED #1 ON MY LIST!!      I think thats where i went wrong.  I should have focused all my spare time and energy towards God and not towards guys, my image, where i was going out on Friday night, or my friends.  Ive been thinking alot lately, and i think its time to change my ways, and get back on the walk with the big guy.  Hes always been there for me, and he never lets me down. I think its time that i should start thanking him for whut hes done for me.  Hes given me a second chance to walk with him, rite before i slipped away.  People and friends around me are changing and thinking more about God, and im so proud of them!!! im REALLY happy and ecxited about my fresh start to walk with Christ. I love him so much!!!the Father of our world is AWSOME! remember.... GOD LOVES YU!!!  so put him #1 on ur list!!!   '),(17061,3308717,'11,May,2004','2004-05-11 00:00:00','   I just had a thought that would make everything fit together!! I have indeed figured out where we come from, though how 42 fits into it... I am still working on that.  Heres the quick run-down untill I have time to forumlate a full theory.  Scientists say we evolved from Monkies, etc...  Creationists say we were made in Gods image...  Now, in the bible there are \'gaps of time\' that could account for long periods of \'evolution\' present in science.  Ok, taken that, and what we read and have been taught/know et al...   WHAT IF...  God was a monkey? I don\'t mean that in a sacrelgious way... not at all, but ... hmmm I think I am on to something here, more research required.    -intX    The tree that Eve snacked on was really banana!  '),(17062,3308717,'11,May,2004','2004-05-11 00:00:00','   If you are interetested in programming and you are interested in free stuff then you just might want to look at  urlLink this . Basically, just watch five movies and rate them and MS will send you a free, not-for-resale, copy of VB.NET 2003. This worked out good for me since i have VB.Net but not the 2003 version.  So whatta ya say, even if you don\'t program go do it and get a copy and learn yourself something.   -intX    a free-for-all, fuck \'em all!  '),(17063,3308717,'11,May,2004','2004-05-11 00:00:00','   So yesterday I woke up to a swolen face. Not completely swollen, but like my sinus areas and my upper lip like I had been punched. There was no real pain other than pressure. It did make me a little disortientated. I think it was just more freaky than anything. I called my doctor firt thing but either they hadn\'t got in to the office yet or they were just swamped with other calls.  Anyhoo, they called me back and told me to take some benedryl (even though I am on zyrtek...) so I did. I guess allergies was all it was afterall. I zonked out and woke up feeling a little bit better and \'The Burbs\' was on. I opened my eyes just in time to see him cut the power wire and fall into the shed. Good stuff.  I\'m feeling much better today, though, ever since yesterday my sense of smell has been noticably better or something.  I had something to post on about VB.Net but now cannot, for the life of me, remember what it was. Well, if I think of it I will post later.     -intX    Rock\'n\'roll with all my friends, hopin\' the music never ends.  '),(17064,3308717,'11,May,2004','2004-05-11 00:00:00','   I think I should do my duty as a, self proclaimed, movie buff to infrom you that  urlLink Bubba Ho-Tep  is now out on special edition DVD. Go get your copies while you can. =)    -intX    shazbot!  '),(17065,3308717,'11,May,2004','2004-05-11 00:00:00','   Ran across  urlLink this article  earlier which reports on a couple of guys that found and recorded locations of quite a few motion sensors scattered about the desert around the \"Area 51\" base location. Supposedly agents and officers of the FBI and Air Force didn\'t much care for the snooping going on.  I like to think they have a Stargate [ urlLink image ] in there! How rad would that be?    -intX    Uhhh, fire the thoton portedos and go... that-a-way!  '),(17066,3308717,'11,May,2004','2004-05-11 00:00:00','   Next on my list of \"Things I must have\":    ( urlLink link )  Who thinks this shit up? And did they get fired?    -intX    I got two turn tables and a microphone...  '),(17067,3308717,'11,May,2004','2004-05-11 00:00:00','   Well, it\'s Monday... yeah. I\'m pretty beat, I\'ve been going and going all weekend. Saturday I went out to the gun range for a while before I needed to get back into town to go to A Perfect Circle concert. The show was awsome. They were just as good as, maybe better than, when I saw them back around six months ago in Tampa. Needless to say, the concert tired me out, but I wasn\'t done yet for the night.   After the show we went downtown to say happy birthday to Bill. He got a pink fanny pack and pink party hats. Good times. Had a couple beers there and whatnot then headed home to get some sleep. The problem is I ended up plopping down on the couch and found a cheesy movie to watch that kept me up till 4am or so.  The next day I decided to sleep in a little and not go to the gun range as was planned the previous day. I was just too tired. But around 2 or so my friend insisted so we go out to do some skeet shooting and target practice so out we go. While we were out there, I somehow tore a muscle in my shoulder and it didn\'t help that I was shooting a semi-auto 12g shotgun either. Also I managed to get a shard of something in my left index finger and dehydrated myself while I was out there.  So in other words, \"ouch\". I might have more interesting stuff later.    -intX    I\'ve fallen, and I can\'t get up...  '),(17068,3308717,'11,May,2004','2004-05-11 00:00:00','   urlLink This article  talks about the chance of a Privacy Czar. This person is to oversee, as you would assume, privacy issues. It talks about not only creating this one position but several other positions under the \"czar\" which manage the different government departments. Hopefully this new position is added and the Patriot Act and, to an extent, the DMCA can be put into check. (One can only hope.)  I had a dream last night, I went to the moon, or maybe it was Mars, I know it was an extraterrestrial body of some sort. Anyhoo, for some reason we were up there tossing a football around in the lower gravity. We got some damned good distance with it too. I remember heading back to earth and remembering that I had neglected to pick up some rock samples as souvineers for my friends back home, so when I got back I called up some friends and we went zooming through space yet again.  Go Friday, Go!!     -intX    Meanwhile, back at the lab...  '),(17069,3308717,'11,May,2004','2004-05-11 00:00:00','   Most of you know my general stance on censorship, I fucking hate it. There, if you didnt know before, now you do. The reason I hate it is not that it shields people from something they would rather not be exposed to but that it forces that shielding upon everyone, no exceptions. With that in mind I totally (heart) the ratings guides.  The ratings: G, PG, PG-13, NC-17, R, et al... as well as the entire video game ratings from the ESRB and such. Great idea, I love them. The question remains tho, why do we need censorship when we have ratings? I\'ll tell you my take on this.  Parents do not want the job of parenting. There is the summed up bulk of it all right there. Noone wants responsability. It\'s always someone elses fault that their kid is screwed up, never their own. Granted, there are exceptions as there are for many rules, but to blame a generation of fuckups problems on TV, video games and movies... that is absurd.  According to the First Amendment of the United States  urlLink Bill of Rights :    Congress shall make no law respecting an establishment of religion, or prohibiting the free exercise thereof; or abridging the freedom of speech, or of the press; or the right of the people peaceably to assemble, and to petition the government for a redress of grievances.    Now according to the  urlLink FCC :   The Federal Communications Commission (FCC) is an independent United States government agency, directly responsible to Congress. The FCC was established by the Communications Act of 1934 and is charged with regulating interstate and international communications by radio, television, wire, satellite and cable. The FCC\'s jurisdiction covers the 50 states, the District of Columbia, and U.S. possessions.    OK, so we find out that the FCC is established by the  urlLink Communications Act of 1934 . You can read that yourself, it\'s in PDF format and I didn\'t feel like typing a bunch of stuff.  The point is, nowhere in any of these documents have I been able to find information granting the FCC the power to censor what is broadcasted. From my understanding, and please I want someone to prove me wrong, the FCC is only in charge of regulating fair use and charges associated with the different forms of telecomunications.  My view on censorship is basically, if you are easily offended, stay away from the stuff that offends you or learn to deal with it. The world is full of offensive stuff and having someone shield you from it does nothing to help you other than to help you hide from the world. To tie the ratings system in, if everythign was \"rated\" and everyone could find out before hand the \"rating\" on something they could choose not to watch it. If, with the rating in place and censorship out of the picture, the program/broadcast somehow exceedes its rating, at that point it should be fined based on severity alone.  Anyway, I\'m done typing now. bye.    -intX    guns don\'t kill people, they just shoot hot projectiles at them  '),(17070,3308717,'11,May,2004','2004-05-11 00:00:00','   So there I was, yesterday morning, ready to go to work when as soon as I get out of bed I notice I have a pounding headache and a sore throat. These things don\'t normally bother me since it\'s usually just one of those things that happens when one wakes up out of some deep mouth-breathing sleep session. I hopped in the shower hoping the steam would revive me. I was wrong.  I ended up calling in sick and stayed home. I poked around on my PC for about an hour laying out the finishing touches for my essay saved that and uploaded it so I could print it later. Then I made it a point to go lay down, at which time I fell back asleep untill around 4pm. I felt like such a bum for it, but oh well that\'s that.  Maybe I should work a little harder on obtaining a new girlfriend. I had pretty much forsaken relationships for a while when out of nowhere I started seeing this girl, problem was the same thing happened again that had originally forced me to swear off dating. Now I have that itch again. Something deep inside is telling me I should go for it, I should just go out there and get ground into the pavement again. I think it\'s also telling me I should run with scissors too so I don\'t really know what to make of it.  Done with that babbling, I picked up \"Aqua Teen Hunger Force\" season 1 the other day and have thoroughly enjoyed that. I also picked up an old classic,  urlLink \"Disgaea: Hour of Darkness\"  for the PS2. This game is brilliant and I would reccommend it to anyone, that is... anyone who is into brilliant games. If you aren\'t, well, this one isn\'t for you... go play Poke`mon or something.  At first this may seem like some sugar coated cutesy game, but as you play it the story really developes into some twisted shit. Let\'s just explain a few things you play as  urlLink Laharl , a psychotic and dimented kid and only son of Krichevskoy the king of the Netherworld. Basically, you are a demon prince. Your direct subordinate is a little cutesy demoness named  urlLink Etna , though she is loyal to you, she sometimes freaks out and has temper tantrums. You\'re goal in this game is to basically overthrow this guy that has taken control of your, now deceiced, father\'s kingdom and destroy everyone in your way.     -intX    do a little dance...  '),(17071,3308717,'11,May,2004','2004-05-11 00:00:00','   It appears that people who wish to have children have to perform this thing referred to as... \"sex\". I don\'t know about you but this is some earth-shattering news that has been laid upon me. Appearantly it is also news for  urlLink this couple . I\'m glad I found out now instead of when I decided to start a family. I mean, I can barely imagine the anger these people must feel having tried fo so long and in so many different positions: Sitting on wooden chairs, standing on stepping stools, riding a marry-go-round. All of these seem like wonderfull ways to conceive a child, but I guess it just doesn\'t work that way.  Oh well, live and learn I guess.   -intX    what would YOU do for a klondike bar?  '),(17072,3308717,'11,May,2004','2004-05-11 00:00:00','   urlLink heyhey16k   I hope you enjoy the above link, it really brought back some memories. Mom yelling from the living room, down the hall to my bedroom \"You need to get off of that computer and go outside. It\'s going to rot your brain...\". We all have good laughs about that now and she even comments on how she should have never done that.  Anyhoo, that\'s it for now. Later.    -intX    oompa loompa doopidy doo  '),(17073,3308717,'11,May,2004','2004-05-11 00:00:00','   In the  urlLink article  I just read is a statement by former Congressman Allan Swift on the way the DMCA SHOULD be. He has been making home copies of T.V. programs, singles from cassettes and CDs. He even gives these compilations out to his friends, but he never charges money, not even the cost of the media he puts it on.  I think a quote from him pretty much sums it all up:    When I buy a CD or a DVD, that content should be wholly mine to do with as I please as long as I am in no way selling its contents or profiting from it. ... Present law is predicated on the assumption that consumers will rip-off copyright holders. The vast majority are innocent of that assumption, but all are treated as guilty.     -intX    keepin\' it real in the \'hood  '),(17074,3308717,'11,May,2004','2004-05-11 00:00:00','   As the title may implay to some of you, this posting is more or less just a status update on things I am working on, such as essays for class and software I am writing. Not a major update I guess, but as I have yet to look around for something good to write about I figured I might as well do something.  To start with, as a few of you may know, I have started back at college. I\'m on my second week of a six week, three credit class. To my surprise I actually like being back in school. Well this class currently has me working on one essay and a longer-term essay. The essay is a process essay discussing the Rational Unified Process (RUP). Basically, this is a general process by which software is planned and designed.  The research project is an argumentive paper discussing the problems with the DMCA such as how it prevents researchers, programmers along with many others from reverse engineering, circumventing security and/or finding holes in software and reporting or publishing papers on such research if they do decide to do it anyway. Several people have so far served cease and decist orders, arrested, charged and extridited under the broad blanket statements of the DMCA.  I\'m also making some progress on a program I am working on in my free time (like I have any). I\'ve been getting a lot of it reorganized here and there and some new parts written. I\'m about to get to the point where I need to start working on the printing interface for it.  Well, that\'s about all I got for now. I\'ll probably post more later on when I get a chance.    -intX    stab stab stab!  '),(17075,3308717,'11,May,2004','2004-05-11 00:00:00','   Saw something earlier that I found quite funny, and just now got around to posting it. Hope you enjoy it as much as I did.   urlLink http://musicdish.com/mag/?id=9452     -intX    I sound like I\'m on old-time radio  '),(17076,3308717,'11,May,2004','2004-05-11 00:00:00','   I was pointed  urlLink in this direction  earlier today. I figured since I was talking about cyborgs and whatnot the other day, and then had the dream last night that I mentioned earlier, I would pass this link on.  It\'s a good read and raises some good questions. Check it out if you\'re interested.    -intX    teaching rabbits to kill since 118ad  '),(17077,3308717,'11,May,2004','2004-05-11 00:00:00','   Not to get all \"Dear Diary\" on you but I had a dream last night that was rather strange. I don\'t remember the whole thing, but I do recall parts of it.  I was back in my old neighborhood where I grew up and was at my neighbors house. Somehow I had my left forearm replaced with a synthetic/mechanical arm. It must have been a new addition because there were still stitches and it was sore. I also had a remote controled truck of some kind. It was a full sized vehicle but I could controle it with an arm mounted remote, kinda like K.I.T. needless to say it was pretty fucking rad. Pretty fucking rad that is, until my friend started messing with the controls and it crashed into another neighbors house and took out part of a wall. I back the vehicle up into a garage and closed it and we ran inside before anyone noticed.  Once inside it was like our old neighborhood get-togethers. All of our parents were there, the grill was going and it seemed like a good time. The next thing I remember was that I was lost in a strange house that had a T.V. in every room, I was trying each one, looking for anything except static, I think the static was causing my robotic arm to act up and I couldn\'t control it efficiently.  I think this is where I woke up. I know there was more to the dream but I just can\'t remember it. It\'s strange, I have been having very vivid and outright strange dreams the past week or so and I rather enjoy it.  In other news, I have decided on the topic for my research paper. I am going to write about the downfalls of the DMCA. It is an argumentive essay so I will be arguing that the DMCA does more harm than good and that it should be drasticly altered or done away with. Depending upon how it comes out I might look at publishing it. Also, if I enjoy writing this essay as much as I believe I will, I think I might write another on the problems with the Patriot Act and how it steps on the rights of individuals.  On a scientific note, I\'de like to point you to  urlLink this article  I saw on slashdot about David Deutsch\'s experiment which, as he describes it, shows proof of a parallel universe. He also gives instructions on how you can test this at home with a laser pointer, a pin and a piece of white paper.  The parallel universe he describes sounds like some evil, negative, shadow universe where an exact copy of yourself exists, except with an opposite moral allignment. If you ask me, I think that\'s pretty bad-ass. I would caution against taking this article for fact though. I might take some time and do some more research about this, though I have done a lot of reading and research on quantumn theory and what he is proposing does, actually, make some sense.    -intX    randome phrase  '),(17078,3308717,'11,May,2004','2004-05-11 00:00:00','   While it might not be news that Hitchhikers Guide to the Galaxy by Douglas Adams is being turned into a  urlLink movie . It would probably be news that an  urlLink official website  has been set up. There is even a production blog there for you to read. Though I would like to take credit for discovering this, I actually spied it over at  urlLink BoingBoing .  Enjoy.    -intX    powered by whizbang newfangled ideas  '),(17079,3308717,'11,May,2004','2004-05-11 00:00:00','   A  urlLink friend  pointed me to  urlLink this article . Now, I had heard about this supposed  urlLink DMCRA  which is meant to \"fix\" some  urlLink things that were wrong  with the  urlLink DMCA , so this doesn\'t come as total news, but I\'m glad to see people are starting to take notice of this.   Ok, I realize I have given you a lot of links to read over, but this is kind of important to me. I am glad that this issue is finally being addressed. Too many companies have brandished the DMCA at anyone they deem \"evil\" (which in most cases, translates to people like me, and probably you, that like to figure things out or possibly write our own software to emulate another program). In my opinion, and in the opinions of others, the DMCA was a knee-jerk reaction to events and concepts that were not understood by the people that made the law. The DMCA greatly cripples the ability to legitimately explore software and hardware. It gives companies the power to file legal action against individual researchers giving reports of vulnerabilities. It also gives companies the ability to extradite foreigners who reverse engineer their products or security features, even if it was done for personal \"fair-use\".  From what the article says however, it doesn\'t look like this proposal will get too far. Butch Otter (R-Idaho) looked down on it saying, \"Theft is theft and property is property\".  Hey Butch, no shit, I\'m glad you explained that in such a way that my simple mind could comprehend it. The problem is, Butch, the DMCA doesn\'t just stop thieves from thieving property, it stops legitimate researchers from doing anything more than sticking their collective thumbs up their collective asses. Without research, Butch, we cannot innovate, and without innovation we cannot progress as a society.  I think that about covers my thoughts at the moment, I might comment further in a little while.    -intX    hug the duck  '),(17080,3308717,'11,May,2004','2004-05-11 00:00:00','   I have often had dreams of the future. A place where we humans may one day be physically enhanced by machines. A place where I could rule with, literally, an iron fist. Since I have fond thoughts of this \'future\' thing, I thought it appropriate that I should mention  urlLink this article . Give it a quick scan over, maybe read some bits here and there then come back here and finish up with what I am babbling on about.  Imagine if you will, the ability to switch your vision to the infrared light spectrum at will. Viewing over-lay diagrams of something you are working on, perhaps your cars engine. Maybe you wish to have notes for your presentation lightly scrolled in your periphrial vision. This is all  urlLink a lot closer  than most people would think. In the link right back there it talks about head mounted \'glasses\' which reflect low-powered lasers onto the users retina displaying virtually any image provided to it.  With enhanced vision not too far off we are that much closer to my robotic future, but we need something more. We need strength enhancements,  urlLink no problem , this is being worked on already. I haven\'t quite decided if I would get a couple limbs lopped off yet, but I would abso-fucking-lutly go for an exo-skeleton type frame connected directly to my brain.  I can imagine, however, that with all these robotic implants it would make air travel a little more irritating. I wonder if there would be any exceptions or special planes that cyborgs would fly... Shit, i may just get rocket boosters attached to my mechanical body and fly myself wherever the hell I wanted to go.  Of course this would, more than likely, lead to a war between the humans and the cyborgs. The humans would start it by trying to revoke our freedoms and rights claiming that we were no longer self-aware or that we were a threat. In the end the cyborgs would crush the human opposition and we would expand to other planets and travel the galaxy waging wars with any who stand in our way.  I guess that\'s it for now, just something I was thinking about.   -intX    ...and your point is?  '),(17081,3308717,'11,May,2004','2004-05-11 00:00:00','   FYI, I\'m fucking addicted to these hardened pieces of tropical goodness.     In other news, I have yet to win tickets for the Perfect Circle concert that is coming to town so I might just break down and buy some. I don\'t wantto just buy one, but I don\'t also want to be in the position of people backing out and not paying me my damned money for the ticket.   -intX   Fo Shizzal...  '),(17082,3308717,'11,May,2004','2004-05-11 00:00:00','   I\'ve got such a headache today. I think I learned too much yesterday. Speaking of, I had to write a damned essay the first night. Oh well, it wasn\'t that bad. It actually seems like this class will be pretty easy. I don\'t know if it\'s just because I have had such a long break or if it\'s because I\'m just a little older and a little wiser.  I need to pick a topic for my research paper. I think I\'ll go with \"Censorship In America, More Harm Than Good\". I really dislike censorship, I think ratings are a great idea, but not manditory censorship. If you don\'t like it, DON\'T FUCKING LOOK AT IT YOU STUPID MOTHERFUCKER. ...I might include that in my research paper. ^_^  Got quite a bit done on my application last night. Reorganized the classes and did some much needed commenting. =) I still think I need to rewrite the search functionality of it. Maybe I\'ll have it return each match individually instead of as a single array... then again, it might be more memory/cpu intensive that way... I shall stew on it.   -intX    what?  '),(17083,3308717,'11,May,2004','2004-05-11 00:00:00','   Huzzah, first blog post here. Well, what can I say?  I start back at college today. in fact, in about two hours I will be on campus. I\'ve decided to finish off my degree after taking a short \"break\" 5 years ago... Oh well, it will be good to get back into it. I think I will go for my CS degree, I was originally going for an AA but quickly realized that it was not something I wanted to waste my time on when I could be working. I don\'t regret it, I got a good start on my career even without college.  On to a different subject, I am working on a .Net application. It is basically just a image viewer/thumbnail browser. The big deal with it is its built in search functionality and print queueing. I would carefully estimate that I have about 70% of the core programming done. After I finish that up I\'ll move on to some hardcore testing and kicking the crap out of it. I have one tester lined up, a photographer who has gotten into the digital imaging scene and the, what I call, photo-on-demand market. That is what this application is aimed for.  Well, I\'m off to raise some hell.    -intX   What?  '),(17084,3308717,'11,May,2004','2004-05-11 00:00:00','   \"G-Fresh in the mother-fuckin house!\"      Yeah, anyhoo...  The above image is from Orgazmo if you weren\'t already in the know. If you weren\'t, hey don\'t sweat it. Being in the know is not all it\'s cracked up to be. I spend most of my days answering great questions such as \'when will  item_x  be done?\' and \'can you put a new water jug on the cooler Mr. Man?\'   To make sure, however, that you are not completely out of the know, allow me to point you in the direction of a certain site by the name of  urlLink http://www.experiment301 .  And yes, we will be serving punch and cake later.    -intX    mastering the art of interrogating koalas since 1909.  '),(17085,3308717,'11,May,2004','2004-05-11 00:00:00','   urlLink Link   SpaceShipOne, the shuttle I constantly talk about, made it\'s flight today and surpassed the 62 mile marker adding a new page to the history books and earning Mike Melvill, the civillian pilot, his astronaut wings.  I wish I could have been there. Shit, I wish I could have ridden in it. Oh well, someday I guess.    -intX    I\'m just a cook!  '),(17086,3308717,'11,May,2004','2004-05-11 00:00:00','   This has allowed us to develop a complete manned space program from scratch for the price of one of those government paper studies, I’m so proud of that it brings tears to my eyes, it really does   (Rutan)  Just think, today could be the day that the first privately funded vehicle makes it out of our atmosphere on its own. \"June 21, 2004\" the history books will read, \"was the day Scaled Composites showed to the world that it was time for humanity to make its move into space.\" This will also signify that the governments around the world need to wake up and take things a little more seriously.   Some people would argue that we do not need to go into space, that we still have oceans to explore here. I would say they were correct; however, it is not the fact that we do not need to go into space, it is what we learn by attempting these things that gets us further. I would estimate a good 85% of what we use daily was developed durring or for missions into space.  Anyhoo, good luck Scaled Composites.    -intX    Warp Seven! ENGAGE!  '),(17087,3308717,'11,May,2004','2004-05-11 00:00:00','   I had an idea to get the way I feel across to the government. I wouldn\'t suggest this to anyone else, but I will support you if you do decide to do this. I may one day just do this myself, but I wan\'t to make sure I either do it first, at the same time, or not at all.  Here\'s my plan:  Have my mouth physically stitched shut in the morning before sunrise. Stick a set of horse blinders on my head and put on some earmuffs. Position myself in front of one of the main buildings in D.C. and hold a sign that reads \"I had a better sign, but it got censored.\" and on the back \"I\'de be more controversial, but then I wouln\'t make the news.\"  Spend a day out there, yes that means no water, no food and probably pissing yourself. It would be worth it though. I just need to go do it now. It only took a single man to stop a tank...      -intX    This phrase has been censored by the U.S. government.  '),(17088,3308717,'11,May,2004','2004-05-11 00:00:00','   urlLink fucking link!   Goddmanit, as if the DMCA wasn\'t bad enough. First they decide to throw a piece of crap like that into legislation. Now we got some jack-off senator aiming to toss another monkey-wrench in the gears of innovation and personal freedom.  A hardy \'fuck you sir\' goes out to  urlLink Sen. Orrin Hatch  (R-UT) for his wonderfull work toward a safer world full of lawsuits over meaningless crap. Way to clog the system buddy.  Well, now that I have that out of my system I think I can enjoy my weekend calculating how long we have until Skynet or Big Brother come in to the picture.  Have a good weekend you mindless dumbfucks.   -intX    Omnia mutandur nos et mutamur in illis  '),(17089,3308717,'11,May,2004','2004-05-11 00:00:00','   Well, I ended up getting a 98 on my final research paper. The 2 points I lost were from indenting the first line of my blockquotes. I indented all of the blockquote like normal, but I also indented the first line as you would for a new paragraph. Pretty lame thing to lose 2 points on. =/ I mean, at least I was consistant in it.  I believe I also did quite well on the final exam, which was 60 questions and an in-class essay about frequent testing in classes bleh.   So yeah...    -intX    blasphemy  '),(17090,3308717,'11,May,2004','2004-05-11 00:00:00','   I was minding my own business earlier when all of a sudden a  urlLink googlefight  broke out right in front of me. I thought it was kinda interesting so I figured I would let you know about it.    -intX    wakka wakka wakka  '),(17091,3308717,'11,May,2004','2004-05-11 00:00:00','   Form of... Warped Tour tickets!  Yep, I got mine. Have you got yours?  urlLink Here  is the current line-up for the show. So far I think we have 5 people going, possibly more.  Show starts at noon and goes into the night. BAM!    -intX    I\'m Spicey!  '),(17092,3308717,'11,May,2004','2004-05-11 00:00:00','   urlLink link   Above is a link to an ebay auction. This person is selling TONS of video game hardware and software. From NeoGeo to X-Box this guy is selling a huge package deal. It also comes with a lot of the development hardware and software. It\'s labled as the \"Japanese TV Game Museum Collection Set\" and spans around 20 years of video gaming history.  So if any of you feel like owning my soul, just get this for me and I\'ll be your slave for quite some time. Well, whenever I am not occupied with the merchandise. =)   -intX    Sorry, but the princess is in another castle.  '),(17093,3308717,'11,May,2004','2004-05-11 00:00:00','   It\'s way out... in the water. See it swimmin\'?  Actually my mind is at Spec\'s where I will be picking up my Pixies tickets in about an hour or so. So you\'ll have to excuse me if I don\'t just sit here and type on further about nothing.       -intX    slicin\' up eyeballs  '),(17094,3308717,'11,May,2004','2004-05-11 00:00:00','   So there I was...  (You can always tell a good story when it starts off with \"So there I was...\")  So there I was, sitting at my desk putting some finishing touches on a report I was writing in FoxPro (yeah, *cringe*), and I think to myself \"Self, this needs to be compiled and tested. - You are absolutly right self!\" So I begin to compile the program which contains the report, the one which I had mentioned earlier. Things go fine, then a couple errors crop up that keep me from proceeding. You see, this isn\'t all together strange, we work in a multi-user environment so all of our code is stored in a source control, usually when this happens it just means you maybe havent added, removed or updated some files; no problem.  (If you watch closely this is the part where my morning becomes... not so happy and fuzzy and stuff... as if it were before. Well, it was, to some extent, a bit more on the cheerfull side... but I digress.)  So I was deleting a couple files and I guess I just got on a delete spree for some reason and clicked on the report I had just written (I am currently blaming the fingers but the fingers keep pointing in other directions). As I looked this file dead in the eyes (or... whatever form of ocular organ a file may have) my finger swiftly and deliberately pounced upon the delete key. Immediately I realized what I had done when the confirmation dialog box popped up.   \"Do you want to: [Remove] [Delete] [Cancel] ?\" it enquired.  \"Interesting choices, though I really don\'t want to delete this file. This is the one I have been working on.\" I said to myself as I clicked directly on the button marked \"[Delete]\".  The dialog box left in much the fashion that it appeared. I could feel my brow do a summersault and land in a half furrow, half questioning position. My eyes blinked of their own accord for a moment and then my hand joined in the fun and smacked both of them just for the hell of it, it made sure to include my nose just to make sure my mouth would be upset. Obviously it was quite upset, as it immediately let go of my jaw and muttered a few curse words.  Needless to say I will be spending the rest of the morning rewriting that report and hoping I do not delete it a second time.     -intX    Bottles and cans, just clap your hands  '),(17095,3308717,'11,May,2004','2004-05-11 00:00:00','   Dear Nintendo,  I noticed that you have recently started your NES Classic campaign whereby you are bringing back such hits as The Legend of Zelda and Donkey Kong as well as a fanny pack full of other greats. While I applaud you for re-introducing today\'s youth to the revolutionary ideas and game concepts I cannot help but see this as nothing more than a way to rake in some fast cash by whoring out poor Link and Mario, two of your greatest champions. I certainly hope I am wrong in this, believe me, I do.      urlLink Link to IGN   Anyhoo, that is about how a letter from me to  urlLink Nintendo  would go. As you can tell, I have sorta mixed feelings on the whole situation. On one hand I am glad to see the classics getting out there on a newer system that today\'s children will possibly buy and find as much entertainment in as we did back in our youth. On the other hand, releasing each one sepperately for $20 each seems like more of a way to bring in extra cash. I think I would have been a little more happy about this had there been, I don\'t know, 3 or so games on one cart.  I read a  urlLink review  which covers Super Mario Brothers and The Legend of Zelda. The reviewer lost some credibility when he mentioned \"The game has been given the same treatment as Super Mario Bros., with a gray cartridge and all\". Now, I would like you to take a look at the last part of that sentence. The reviewer mentions earlier in his review that the SMB cart matched the old NES game, however the Zelda game was, in-fact, gold. They possibly did come out later with a grey one, but I have yet to see one.  I digress as I jst wanted to point out that the \"new\" classic NES games you may or may not choose to purchase are, in fact, out there for consumption. I MAY break down and buy one or two. If I do it will be Legend of Zelda for sure, and if I find out that there is a Kid Icarus coming I will definately have to purchase that one.  The thing is, however, I do still have a working NES attached to my living room television. I also still have Kid Icarus, Zelda (I & II), SMB 1,2, and 3 and just about any other classic NES game anyone would ever want to play (including Gauntlet, \"Blue elf needs food badly\").  Well, that about summs things up for the time being.    -intX    Justin Bailey  '),(17096,3308717,'11,May,2004','2004-05-11 00:00:00','   If you haven\'t heard of it yet, please check the  urlLink Monolith  project. The premise of this project is that it combines two files to create a third completely different file with a .mono extension.  See it works like this...  Let\'s say you have:   Metallica-Master_of_Puppets.mp3  And you create a text file with enough characters to make the files the same size. We\'ll call this file, junk.txt groovy? groovy.  Now Monolith will take the two files and \"munge\" them together. You now have a file called munge.mono or something similar. The contents of this file are neither of the original files, yet they are both of the original files. (mmm quantum states...) Anyhoo, this new file is not, nor can it be copyrighted, and therefore is free to be distributed since none of the data contained has anything to do with the copyrighted Metallica song.  The nifty part is, what a person does with this file is totally up to them. They could for instance get a copy of your junk.txt file and extract \"Master of Puppets\" from it, but ONLY with an exact copy of junk.txt in this particular case. To bring this all together junk.txt is not a copyrighted file is it? Nope, we just made it... I guess you could go TRY to copyright it but that is neither here nor there. Avoiding further digression, junk.txt is free to distribute right along side your munge.mono. When the person you send this to gets the files he/she can run their Monolith and combine the junk.txt and the munge.mono files to pull out the Metallica-Mater_of_Puppets.mp3 file. Of course their ownership of that file is subject to scrutiny but the trasferring of data cannot be considered criminal, since the data transmitted has absolutely no copyrighted content within it.    -intX    UNO!  '),(17097,3308717,'11,May,2004','2004-05-11 00:00:00','   I don\'t usually post these quizz results on my blogs but I felt I would just do a run-down of my personality.           What the photon? You\'re freakin Zaphod Beeblebrox! Two-faced, a little arrogant, and most definitely vain, you somehow have managed to be not only the life of the party but the afterlife of the party as well. People like you, but hold you at arm\'s length, mainly because you\'re so self-centered. Stop worrying so much about yourself and you might find yourself a few more true friends.              -       Disorder      Rating        urlLink Paranoid :    Low       urlLink Schizoid :    Low       urlLink Schizotypal :    Moderate       urlLink Antisocial :    Low       urlLink Borderline :    Low       urlLink Histrionic :    Low       urlLink Narcissistic :    Moderate       urlLink Avoidant :    Low       urlLink Dependent :    Low       urlLink Obsessive-Compulsive :    Low       --  urlLink Personality Disorder Test - Take It!  --       -  And according to Myers-Briggs test thing, I am  urlLink ENTP . Extroverted Intuitive Thinking Perceiving \"Clever\"  Fictional Characters: Q (Star Trek) Bugs Bunny Wile E. Coyote  Real People: Alfred Hitchcock Thomas Edison Lewis Carrol Weird Al      -intX    this will be the one and only time I do this  '),(17098,3308717,'11,May,2004','2004-05-11 00:00:00','   I thought I would share a little bit of my adventures from many years ago, so here you go.  My friend took the picture below as I was about to reunite the crystals. It was a taxing event and nearly killed me, but as you can see by my expression in the photo, I was quite determined. The world is now safe, don\'t get me wrong I had a lot of help along the way (thanks Aughra!).       -intX    sdrawkcab  '),(17099,3308717,'11,May,2004','2004-05-11 00:00:00','   Did some catching up on some of the usual reading I do and came across  urlLink this article  discussing virii and the people that write them. Basically saying how difficult it is to actually track the writers down. This article then goes into mentioning WiFi and how it opens up a whole new attack point for virus writers.   urlLink This  has been spotted in several places and has thus been reported by these places. I decided to give you the  urlLink BBC  version. It seems that companies have been selling off hard drives on eBay. Cool, companies recouperating losses to an extent... well, the not-so-cool part of it is that these drives contain residual data. Even though companies and individuals may think they have cleaned their storage devices residual data can still leave a magnetic imprint. The article gives a few suggestions to avoid this problem such as encryption and access control procedures. In the past I wrote scripts that would rewrite file contents with random characters tothe exact byte-size as the former file. Then it would rename each file to a random string, also of the same length, then delete these files (or format, fdisk, format on the particular partition). Sure it can be a lengthy process, but having a script that performs most of it helps greatly. An honestly, if the data is sensitive you should be going out of your way to clean it up whether it be personal or business.  Finally, if  urlLink this  is true you will have to excuse me while I sit here and shake my head in dissapointment. Land of the free and home of the brave my ass. I love watching our rights fade away on a daily basis.    -intX    We control the horizontal  '),(17100,3308717,'11,May,2004','2004-05-11 00:00:00','   I just got done reading an interesting  urlLink article  in which Declan McCullagh raises the question (which I have been raising for quite some time now) of \"[is it] time to abolish the Federal Communications Commission\"?  He brings up several good points. But only barely touches on the main point which I use in my arguments, censorship. He mentions Howard Stern and Jannet Jackson\'s nipple. This falls under the category of censorship. The FCC was originally a good idea, however times change and because of that fact, we need to change as well.  As you may have read in some of my other posts, I have found nowhere that the FCC has the power to censor. In fact, that is an impedement of our First Amendement: The freedom of speech and of the press.  Well, you just think about that today as you hear your lyrics bleeped off of the radio, or see a nipple blurred off of your favorite movie. Maybe you can think about it while you\'re reading the morning paper and get the feeling you are missing out on some facts.    -intX    omnia mutantur nos et mutamur in illis  '),(17101,3308717,'11,May,2004','2004-05-11 00:00:00','   urlLink Link .  Follow the link above to check out a nifty game. It\'s \'Super Mario War\', a Mario deathmatch if you will. I just saw this link floating about so I decided to check it out. Haven\'t downloaded it yet, but will plan on doing so once I get home from work tonight.   -intX    The cow goes \'moooo\'...  '),(17102,3308717,'11,May,2004','2004-05-11 00:00:00','   As the title may suggest I am currently listening to Ænema off of Tool\'s album Ænima. This has always been, by far, one of my favorite songs from them. Tool in general has always been one of my favorite bands. I believe nothing more needs to be said about this subject since it is obvious that I have covered all there is to be said. =)  If you haven\'t figured it out by now, it is in fact Friday. I will only be working a half-day today as I have a doctors appointment later on and I needed to get a few things done while I was at it. I\'ve got a real craving for sushi right now, and quite possibly a bottle of sake could accompany the meal to result in a well rounded meal and a warm fuzzy mood.  I\'ve noticed a few new people have actually stopped by. Not only have they stopped by, but they actually posted. Aren\'t I the popular one? Ok, so maybe I\'m not *that* popular, but hey, it gets me through the day.  Sorry I don\'t really have news right now. I\'ve been busy at work and just didn\'t have time this morning before I came in to the office. Speaking of the office, we are trying to decide which language we should migrate to when we do our big systems re-write. So far it looks like the majority are leaning to C#. The other main choice was VB.Net. I think either would be fine honestly. The languages in the .Net framework all compile down to basically the same thing so there really isn\'t much in the way of performance. Several quarks have been pointed out to me in VB.Net that I never noticed. I would probably ellaborate on it however, I don\'t see why most of the people reading this would even care. =p  My Comp class is almost finished. Summer A session is being a real bastard. We\'ve had 1k word essays due each week. I have one small essay due tuesday discussing opposition to my argumentive research paper. I need to follow that with a rough draft of my paper on Thursday and then turn in my final the following Tuesday. That thursday we receive the papers back and take the final exam. So I am almost done... *phew*.  I\'ll see about digging up some interesting news, in the meantime, consider joining me and many others at the  urlLink Warped Tour  in Tampa, July 30. Their list is showing 161 bands on the roster. Mmmm music.    -intX    modifying the immutable since 1978.  '),(17103,3308717,'11,May,2004','2004-05-11 00:00:00','   urlLink Space.com  is reporting on  urlLink Scaled Composite\'s  SpaceShipOne, the privately sponsored craft that is on it\'s way into history as the first non-government vehicle to reach sub orbital altitudes and earn it\'s pilot his astronaut wings.  Who knows, some day in the near future we may be taking vacation cruises into space just as we take cruises on the ocean. With talk of colonies on the moon and Mars, we may need ships like this just to get to work. =) How rad would that be?  Read more on this  urlLink here . They have a few more images, but here is one that gives you a view of the cockpit and the pilot. This was taken during their last test flight which took them up to around 40 miles. Their goal to reach is appx. 62.5 miles of altitude, which they plan on reaching on this next flight.        -intX    Some people call me the Space Cowboy...  '),(17104,3308717,'11,May,2004','2004-05-11 00:00:00','   According to  urlLink this site , Danger Mouse is due to arrive on DVD in the US in 4th quarter of this year (2004 fyi). If you get a chance stop by their sounds section and listen to some of it. Brings back the memories. =)       -intX    cor...  '),(17105,3308717,'11,May,2004','2004-05-11 00:00:00','   I have been a little slacker-ish in my posting of news. You\'ll have to forgive me, I still do not feel that great and I have been supremely busy writing essays for my class. =/ Anyhoo here are a few things I thought some of you may be interested in.  First up is  urlLink this  from BoingBoing which points to  urlLink these search reseults  for Japanese Uniforms. ( urlLink hot )  Nest, on to  urlLink The Marvel Directory . A database with information on anything from the Marvel universe. =p  Finally, I wouldn\'t call this news really since I heard about these a year or so ago. However, it looks like they are finally coming to market.  urlLink These virtual keyboards  are projected from portable devices onto any smooth flat surface and a miniature camera senses where your fingers are. Pretty nifty eh?    -intX    What? Behind the rabbit?  '),(17106,3308717,'11,May,2004','2004-05-11 00:00:00','   Looks like a  urlLink fire started  up Saturday evening in the southern portion of the county and is still burning away. The smoke is noticable pretty much everywhere and I think it\'s what is making me feel dizzy and disortiented this morning. At the bottom of the picture below (and in the news article) is the interstate and just above that is TECO arena where I saw A Perfect Circle a couple weeks ago. This is also where the Everblades (I hate that name) play, if that sort of thing interests you at all.      -intX    boom shaka-laka boom  '),(17107,3308717,'11,May,2004','2004-05-11 00:00:00','   I seem to be a hub of advice. It has been this way for some time now. I have never understood it fully. I mean, I listen to the advice I give, but I rarely follow it. Well, that is, if I have not already experienced it. I don\'t know if it is a gift or a curse. People tell me constantly that I anticipate their thoughts and I do admit to that. At the same time I seem to be, as one would describe, an absentminded professor.  I think back on my life in a sort of fast-forward review of events leading me to my present position in the space-time continuum. I am happy about one thing over everything else. I have given myself to experience. I try not to take someones word on things, be it good or bad. I\'ve hurt myself plenty of times. I\'ve even hurt others, and by hurt I mean physically and/or emptionally. I don\'t appologize for any of it, it\'s life. That has lead me to conclude that life is meant to be lived, lived with no regrets. Obviously you need to moderate your activities and minimize negative influences on others, but often that cannot be avoided.   This brings up another thought. People, in general, appologize for far too many things. At times they appologize for events that are well out of their control. We also tend to over-use other words and phrases, such as \'love\' and \'hate\'. I am not saying I am innocent of this, but I do try to watch my careless usage of such words. I admit, however, that it is hard in fits of excitement to check every word. Once I began thinking about this subject years ago I started noticing more and more people using these words in completely wrong contexts. Then again, who am I to say what is right or wrong, such things are subjective.  I recently read an inteview with Maynard from Tool, one metaphore he used to describe knowledge and information is this:   \"If you look at the cycles of the moon, it starts as a thin crescent and then gradually waxes until it becomes full; then it gradually wanes back into another crescent and then it is gone. The moon reflects sunlight like humans reflect information. We wax and wane and when we become full moons, our egos are full.   I think he hit it dead on. We should not take credit for information. The information was always there, we didn\'t create it. However, interpretation is something to be proud of. Sharing knowledge I learn with others is one thing that really makes me happy. To inform someone of something they did not know before. Sometimes I think I should be a teacher or a historian. I very much enjoy ancient history and learning of cultures that were pondering the same questions we have still today.  I\'ve recently been thinking about looking for a new female companion. The last two times I ventured this i got burned. Maybe I\'ll just poke my head out of the treeline and see how things are going out there these days. It\'s not insecurity, nor is it fear. Some might call it hesitation, but I don\'t think that would fit either, if anything I am not hesitant (for the most part) in fact most of the time I am quite the opposite - \"shoot first ask questions later\", \"going both feet in\", etc. Maybe I would call it, cautious I guess. Yes, that works.  Anyhoo, I\'m pretty much done rambling. Mynext post, I promise, will be something back in line with the standard dork-factor, which is normally present here.   -intX    what, me wory?  '),(17108,1841346,'28,May,2004','2004-05-28 00:00:00','    A couple days ago, I finished a wonderful novel.  Masterfully written, intelligent and completely absorbing in its psychological and philosophical examination of overcoming the emotionally repressed life as seen through three seemingly different individuals.  The following quote is from the second to last paragraph of the novel.  Traveling the Midtown Tunnel, a  metaphore for the more indirect path we sometimes choose in life and how it inevitably leads us where we need to be (major theme), resonated with me as much for recent events as past ones.    The Midtown Tunnel is a funny place....Approach the city by any other route-the lofty bridges, even the Holland and Lincoln Tunnels-and you will emerge knowing precisely where you are.  But this route takes you in obliquely.  You come up not into the city\'s tarnished brilliance but onto a road beneath a shabby underpass, gray shadows broken only by weedy trees of heaven and billboards promoting airline shuttles.  You emerge onto a nondescript sidestreet-one that, every time, leaves you confused.  The needle on your compass spins, because you are no longer able to see the city\'s outlines or even to see that you are on an island, a glamorous place in many eyes, but to yours...a hard yet reassuring place to live. ~ urlLink  Three Junes , Julia Glass  urlLink >>more    '),(17109,1841346,'26,May,2004','2004-05-26 00:00:00','    Last Tuesday, I get a phone call from my stepmother telling me that my grandfather has died.  I am sad, but not devastated as I didn\'t really get to spend much time with him.  Mostly, I worry about my grandmother and my father.  Christian informs me later that to fly to NY will cost upwards of $1000 then asks about driving there - it\'s six hours away according to  urlLink MapQuest .  Fine, I say as long as once we get to NY you drive.  It\'s a deal - we\'re off.  We switch after four hours so I can nap.  I wake up at 8:00pm (we left at 2:00pm) thinking we must almost be there.  How wrong I was.  The first thing I notice, other than the time, is Christian.  He staring straight ahead with the most intense look of concentration I have ever seen on his face.  Clutched in his hands laying flat against the steering wheel are the MapQuest directions.  I look around.  A road sign reads  Newark, New Jersey .  Noticing the miasma of lights ahead I think,  Wow. Newark is a really big city.   My cell phones rings.  It\'s Dad.  I tell him we are in Newark.  He says we are about an hour away and then tells me how the last time he drove to NY he missed an exit.  (This from the King of Driving - nothing scares him and he never makes a mistake.)  As we hang up, Christian says, \"I missed it!  I missed the exit, I think.\"  I grab the directions and thus begins The Debacle.  We finally figure out how to turn around (going back through a toll, again) and do our best to make it into the city.  Thirty minutes, another missed turn off, two repeated tolls and annoyed toll booth operators later, we end up turning right at 34th and Dyer only to figure out in hindsight that it should have been a left.  We are lost.  In Newark.  Ok.  We can handle this, there are worse things.  Christian pulls over, gets out of the car and asks the first bystander, \"Hi, can you help me?\"  Meanwhile, I am in the car marveling at all the lights.  \"Where are you trying to get to?\", the woman asks.  He shows her the MapQuest directions and says, \"New York City.\"  She looks at him rather sympathetically (she was a nice lady) and says, \"This  is  New York City.\" Apparently, we were in  urlLink Times Square  and  we didn\'t know we were in New York City .  Yep, we were lost in America\'s most recognizable city on it\'s most recognizable stretch of concrete. Is that worse?  I don\'t know, but it is definitely much funnier.   P.S. Eventually, we found the  urlLink Queens-Midtown Tunnel  and with my cousin\'s help (who stayed on the phone with me for 30 minutes until we pulled in front of my Aunt\'s house in  urlLink Long Island ) we made it.  P.P.S.  Plan to visit NY?  I discovered these helpful sites:   urlLink New York Convention and Visitors Bureau ,  urlLink The Paperless Guide to New York City , and  urlLink Aaccessmaps .  (I plan to be better prepared for our next trip to NY.)  urlLink >>more    '),(17110,1841346,'25,May,2004','2004-05-25 00:00:00','    Yahoo is beginning to cramp my style.  I just discovered that  urlLink Google\'s free email  gives you 1000MG free space!  Help me test this Beta version by sending me an  urlLink email .  urlLink >>more    '),(17111,1841346,'17,May,2004','2004-05-17 00:00:00','    Book: Have you read this yet?   urlLink Excellent book .  Set in 1660\'s Holland, is it about a sixteen year old girl who is forced to go to work as a maid for the Dutch painter Vermeer.  Not to give too much away, but she ends up becoming the inspiration for  urlLink one of his most famous portraits .  Sounds rather mundane, but actually it is well done.  Tracy Chevalier does a believable job of combining fiction and history. What impresses me is her ability to be so inspired by the painting that she creates an entire story around its subject.  What I wouldn\'t give to be that creative and talented.    Movie: First, let me say that I  love   urlLink Colin Firth .  He is absolutely one of my favorite Brits.   [ side note : He is also in  urlLink Love Actually . The movie is rather droll and the ending is cheesy, but there is something so wonderfully endearing about British humor.  I laughed myself silly.]  Not surprised this story  urlLink went hollywood .  Has all the delicious elements.  If nothing else, it should be a treat for the eyes as it is already nominated for three academy awards - Best Art Direction, Best Cimematography, and Best Costume Design. Can\'t wait.  urlLink >>more    '),(17112,1841346,'17,May,2004','2004-05-17 00:00:00','    I am so excited to see this movie.  I have recently started reading Greek Mythology of which the Trojan War is one of my favorites.  (Brief synopsis in previous post.) For cool movie details check out the  urlLink official site .  For years, scholars and other dedicated individuals have attempted to sort fact from fiction.  Did Troy really exist?  Homer wrote of his accounting some 500 years later so how could it be true?  If so, where was it?   urlLink This  is a cool site to help sort fact from fiction.  Don\'t feel like reading?  Catch the  urlLink History Channel  program  The True Story of Troy .  It is fascinating.  More into myth?  Go to  urlLink A&E\'s web page  and look up times for their special,  Troy: The Passion of Helen .  urlLink >>more    '),(17113,1841346,'17,May,2004','2004-05-17 00:00:00','    The Judgement of Paris:  The evil goddess of Discord, Eris, was the only deity not invited to an important marriage between King Peleus and Thetis.  So resentfully, she threw a golden apple inscribed with the words  For the Fairest  into the banquet hall.  It came down to three goddesses who appealed to Zeus to choose between them.  He refused telling them to go Mount Ida near the city of Troy and ask a young prince, Paris, who was there tending sheep for his father, King Priam.  Each goddess attempted to bribe Paris, but it was Aphrodite who promised him that the most beautiful woman in all the world, Helen of Troy, daughter of Zeus and Leda, would be his.   The Trojan War:  Helen was in an arranged marriage to Menelaus, brother of Agamemnon and now King of Sparta, who was chosen by King Tyndareus.  Aphrodite led Paris to Sparta where he was received graciously by the couple.  Trusting Paris, Menelaus left for Crete only to return to find Helen gone.  Paris had taken her to his homeland and Menelaus called on all of Greece\'s chieftains to help wage war on Troy.  One thousand ships were to set sail for Troy under the command of Agamemnon.  Before they could leave, Agamemnon had to sacrifice his daughter to appease the goddess Artemis in order to calm the north winds at Aulis and be able to sail.  The battle waged for nine years without a clear victor.  Many battles were fought but eventually the greatest heroes Hector and Achilles fought each other one last time.  Achilles defeated Hector but was soon struck down by Paris who was in turn killed by Achilles\' son Neoptolemus.  It was at this point the Greeks devised a crafty plan to trick the Trojans and get inside the city.  They retreated leaving behind a great wooden horse that contained some of the remaining chieftains.  The Trojan\'s believing the war over dragged the wooden horse through the gates of their city.  Late that same night the chieftans stole out of the horse and opened the gates to admit the entire Greek Army.  By morning the Trojans were slaughtered and Troy was deafeated.  urlLink >>more    '),(17114,1841346,'30,June,2004','2004-06-30 00:00:00','    This is what I think I have.  This is exactly my problem.  My circadian rhythms are so off.  I have battled this since early adolescence.  When a person\'s biological clock gets out of sync with accepted norms, sleep disorders arise.  In delayed sleep phase insomnia, the person falls asleep and awakens later than required for normal activities  such as work and school. Once the person falls asleep, however, his or her sleep is restful and uninterrupted.  Awakening at an acceptable time, even with an alarm clock, is extremely difficult.  Resulting symptoms include daytime sleepiness, particularly in the morning; fatigue; impaired memory and concentration; low productivity, particularly in the morning.   Relatively rare, fewer than 2 percent of those seeking help at sleep disorder clinics are diagnosed with it. It frequently begins during childhood and is most common during adolescence. These symptoms are also seen in people suffering from depression.  Therapies:  Chronotherapy  â€” This technique re-sets the person\'s biological clock by having the person stay up and awaken three hours later each day over a several-day period until their sleep and awake times are in sync with others. During that time they cannot nap, and once they\'ve changed their schedule they must maintain it even on weekends (varying it by no more than an hour) to prevent the return of the problem. [this seems promising]  Light therapy  â€” This technique uses bright lights (as bright as sunlight) for two hours in the morning and avoids bright light for several hours before bedtime to shift the sleep-wake cycle.   Source:  urlLink Discovery Health::Sleep and Dreams:Sleep Disorders   I discovered this  urlLink article  and was rather astonished.  I have been taking antihistamines since I was a small child.  I quit taking them regularly a couple years ago.  The article concludes that, \"Histamine appears to be necessary for the maintenance of the circadian rhythmicity of sleep-wakefulness cycles, food intake, motility and adrenocortical hormone release.\"  urlLink >>more    '),(17115,1841346,'28,June,2004','2004-06-28 00:00:00','    I just googled when the sixth installment of Harry Potter would be coming out.  No luck, but I did find this article,   urlLink Next Harry Potter Title Unveiled  .  The book will be called \"Harry Potter and the Half-Blood Prince.\"  I went to  urlLink J.K. Rowling\'s site  just to check it out and all she will say is that the half-blood prince is not Harry or Voldemort.  Hmmmm.  urlLink >>more    '),(17116,1841346,'28,June,2004','2004-06-28 00:00:00','    It\'s called  urlLink All Consuming  - I love that title.  This site displays the most popular books bloggers are reading by the hour.  Check out the \"Things For You\" section in the grey box towards the bottom of the page.  What I like best is that it allows you to maintain a book list (see my sidebar).  And it is all free!  If you blog and enjoy reading this is a fun way to maintain a just purchased/currently/just completed/favorite etc. reading list.  And it looks rather spiffy on your blog.  urlLink >>more    '),(17117,1841346,'28,June,2004','2004-06-28 00:00:00','    I am looking forward to this weekend.  We are driving to Buffalo to spend it with my aunt and uncle.  Fourth of July is always a special time of the year for us because it comes right after my husbands birthday (7/1 - happy birthday sweetheart).  I can\'t wait to hang out and eat BBQ sausage, hotdogs and homemade pizza.  Hopefully, it will be a nice weekend so I can take my son in the water.  Pittsburgh has turned out not to be the lonely exile we thought it would be with the Buffalo contingent only three hours away.  Truly, it has been our saving grace.  urlLink >>more    '),(17118,1841346,'28,June,2004','2004-06-28 00:00:00','    urlLink     Posted by  urlLink Hello    My previous post about pulp fiction has prompted me to mention one of my favorite movies.  The Whole Wide World(1996) is a  urlLink true story  about the relationship between Novalyn Price, an independent and intelligent school teacher, and  urlLink Robert E. Howard , a ground-breaking writer of pulp fiction (created Conan the Barbarian) in the 1930\'s.  Set in West Texas, it is based on the novel Price wrote from her memoirs,  urlLink One Who Walked Alone .  Zellweger and D\'Onofrio give amazing performances and the movie in total is done so well that I have been (reminded that I was) inspired to read the book.  urlLink >>more    '),(17119,1841346,'25,June,2004','2004-06-25 00:00:00','    Hi.  My name is Leah and I read pulp fiction sometimes.  And when I say pulp, I do mean it in the perjorative sense.  Currently reading the Stephanie Plum series by Janet Evanovich.  It\'s trash, but it light and fun.  I only have two criteria for my pulp: (1)The protagonist must be a woman who is smart (or at least lucky). (2)She must be able to kick ass (including but not limited to the use of guns, knives, martial arts and/or magical powers).  Crime solving is always a plus.     It all began with pulp magazines.  urlLink Pulp magazines , often called simply \'pulps\', were inexpensive text fiction magazines widely published in the 1930s - 1950s. Most of the few pulps still thriving today are science fiction or mystery magazines.  The name comes from the cheap woodpulp paper on which they were printed.  Pulps were the successor to the \'penny dreadfuls\' and  urlLink \'dime novels\'  of the nineteenth century.  Though many respected writers wrote for pulps, the magazines are perhaps best remembered for the  fast-paced, lurid, sensationalistic and exploitive stories  often featured in their pages.    Source:  urlLink www.wordiq.com   I suppose the more stolid readers in the family might want to do an intervention.  But I figure, hey, I also read books/mags about particle physics and cosmology so it all evens out.  [Besides, I have been sick for six days.  This pulp stuff is the best nighttime mind numbing, congested-stuffy head, sore throat, runny nose, cough, so I can get through the mountains of tissues and trips to the bathroom sink to hack mucus medicine.]  urlLink >>more    '),(17120,1841346,'24,June,2004','2004-06-24 00:00:00','    According to Andrew M. Simons, professor of biology at Carleton University in Ottawa, Ontario, \"The presence of nipples in males is probably best explained as a genetic correlation that persists through lack of selection against them, rather than selection for them.\"  Male nipples in essence are \"a genetic architectural by-product of nipples in females.  So, why do men have nipples? Because females do.\"   From  urlLink Scientific American.com , Ask the Experts  urlLink >>more    '),(17121,1841346,'23,June,2004','2004-06-23 00:00:00','    People around the world in countries that haven\'t been peaceful for decades-if ever-are connecting to the internet, writing blogs and pouring their hearts out about the struggles of their people.  You know, difficulties like having enough food and shelter.  Not getting murdered for having their own set of beliefs.  It\'s a reality I don\'t have a clue about.  I have been so lucky by comparison.   Right now I am stressing about taking the pacifier away cold turkey from my son.  He has been crying now for 40 minutes.  As I sit here listening to him fuss in his crib, I am struck by the incongruity of my problems to those of, say, a mother in Iran who lost her husband and can\'t feed her four children.  And I feel absolutely stupid for stressing over a pacifier habit.  Which I actually did for 30 minutes.  It took 45 minutes for him to fall back asleep without his pacifier.  My little world feels alright again.  It is strange that during this global age of interconnectivity and mass information exchange I feel so disconnected.  So encapsulated in a safe reality that I have never had the courage to step outside of.  Now I have a son who will grow up to be a man and it makes me wonder what his reality will be.  urlLink >>more    '),(17122,1841346,'22,June,2004','2004-06-22 00:00:00','    I have been on a mythology bent for awhile now.  Recently, I have become interested in the legends of King Arthur.  What is kind of spooky is that just when I decide to read about a legend a movie seems to be about to come out.  The movie about King Arthur seems really promising because they will attempt to tell the story of \"the real hero.\"  When I saw that on the  urlLink official site  for the movie, I wondered what real hero?  So, I did some digging and as google would have it found two interesting men who have looked into it:    urlLink  Geoffrey Ashe  , an internationally known historian, is considered an authority on British history and mythology.  He has written twenty books and  urlLink his story  is very interesting.   urlLink  Graham Phillips  , former radio journalist and broadcaster for the BBC, has written a book called  urlLink King Arthur: The True Story .  He is now a researcher/writer and has worked with an archaeologist to uncover  urlLink his version  of the truth.  Their findings/interpretations seem very similar up until you get to which historical figure they think King Arthur actually was.  They also differ in where they think \"Camelot\" is located.  Much of the discrepancy seems centered around the interpretation of dates and pinpointing exactly when King Arthur lived.  The BBC Special  urlLink Arthur, King of the Britons  seems to follow Ashe\'s theories.  I hope to watch that. I may even read Phillips book, as well.  One thing I have taken away from this little endeavor is that it makes sense that a legend of this magnitude would have some basis in real events.  Ashe even speculates that King Arthur could be a compilation of several historical figures.  Even so, could you imagine for a minute being the kind of person that would inspire legendary stories for centuries?  urlLink >>more    '),(17123,1841346,'21,June,2004','2004-06-21 00:00:00','    urlLink    Coolest Jazz by Alfred Gockel  Posted by  urlLink Hello    I love  urlLink Art.com .  urlLink >>more    '),(17124,1841346,'19,June,2004','2004-06-19 00:00:00','    Oh, the good days are back.  I can actually surf the internet in peace.  I have reformatted my computer and installed  urlLink Spybot  and  urlLink ZoneAlarm .  I feel as if I have won one for the good geeks.  Life is good.  I have many pics from the Houston trip which have yet to be developed.  I did fasttrack (one-hour photo) one roll while in Houston, so I will have to scan those in.  Now, if I could just hook up my scanner...  urlLink >>more    '),(17125,1841346,'19,June,2004','2004-06-19 00:00:00','    It\'s been a great visit.  Josh\'s house is gorgeous.  I am so homesick.  Although, I must say for the record that I do not miss the oppressive humidity of Houston.  Ick.  But, it\'s home and I can\'t wait to move back.  Hayden had so much fun with his cousins.  Pics to come of some of his cuter moments.  While I am sad to leave Houston, I am looking forward to going home because my cousin  urlLink Nicole  is coming to visit!!  And, girl, I just read that your man sent you roses!  He is definitely a keeper.  Gotta love those Texas men.  They truly are a breed above the rest.  urlLink >>more    '),(17126,1841346,'10,June,2004','2004-06-10 00:00:00','    urlLink    This is my get-down-and-funky-with-my-puzzle pose.  Posted by  urlLink Hello    Photoblogging works.  As soon as I take some updated photos I will post more.  Maybe even do a weekly feature.  I love this free stuff!  [Barb: This is the coolest free blogging tool.  Seriously, no muss and no fuss.]  urlLink >>more    '),(17127,1841346,'09,June,2004','2004-06-09 00:00:00','    Book VI of The Dark Tower series (inspired by  urlLink this poem ) by  urlLink Stephen King , Song of Susannah is out!  Want a taste -  urlLink download excerpt .    It is  urlLink on sale  everywhere for 30% of the cover price.  The saleswoman at Waldenbooks says to me, \"I just want to warn you that there is quite a cliffhanger in this one and the last book doesn\'t come out until November.  I am telling everyone so they\'re prepared.\"  (Did I ever tell you I worked for Waldenbooks once - absolutely the only job I ever had that I loved.) Little did she realize that I know exactly when the last one comes out.  I love people who read books.  They  get  it.  Oh well, at least I only have to wait four months.  I usually wait a year or more with my other authors.  Go SK!  urlLink >>more    '),(17128,1841346,'09,June,2004','2004-06-09 00:00:00','    Yep.  Going to  urlLink Houston  in T minus 2 days.  I\'ve been told it\'s  urlLink very hot .  I plan to get Hayden in a pool for the first time (which means, sadly, that I will have to get into a bathing suit).  The best part of all is that my brother lives there!!!!!!!  [I won\'t comment on the irony of him  finally  moving to Houston not two months after we are forced to leave.]  Nonetheless, it will be very groovy as his sweetie and her daughter get in the day before I do.  I can\'t wait to see their new house.  Hopefully, we will get to see Christian\'s family.  I know Hayden would love to see his cousins - who ask about him all the time and can\'t wait to see him.  I am a bit worried about the flight.  I hate flying (didn\'t used to, in fact, used to love it - not sure what happened).  Plus, I don\'t know how Hayden will tolerate having to sit on my lap for three hours.  All those times I got pissed at the parents of \"the screaming kid\" or \"the crying baby\" that sat close to me on a flight.  Oh well.  Turn about is fair play.  Little did I know, they didn\'t  want  their kid to fuss up a storm.  They were probably good parents with an intense (code word for \"difficult\") child.  Of course, now when I hear an unhappy wee one, I feel quite a large amount of empathy for the parents.  Ah, to be young again.  Remember being blissfully unaware of what you didn\'t know and wonderfully righteous about what you thought you did?  Well, it all comes full circle I suppose.  urlLink >>more    '),(17129,1841346,'09,June,2004','2004-06-09 00:00:00','    What _________ hear in response to the question, \"How\'s Hayden?\"  1. friends:   He\'s pretty good actually.   Blah blah blah blah.  Blah blah blah blah blah.  Blah blah blah blah blah blah blah blah blah blah blah blah blah blah blah.  Blah blah blah blah blah blah.   Just the other day in Ann Taylor Loft several sales girls crowded around him  blah blah blah blah blah blah blah.  Blah blah blah blah.  Blah blah blah blah in Walmart blah blah blah.   The kid  blah blah blah - blah  is such a flirt.  His favorite thing is watching Winnie The Pooh.   Blah blah blah blah blah blah blah.  Blah blah blah blah blah blah, blah blah blah blah blah blah blah blah blah blah blah blah blah blah.  Blah blah blah blah blah blah blah, blah blah blah blah blah blah blah.  Blah blah blah blah blah blah blah blah blah blah.  I\'ll send pictures.  Anyway, how are you ?   2. family:  He\'s pretty good actually.  He is so cute.   Blah blah blah blah blah.  Blah blah blah blah blah blah blah blah blah blah blah blah blah blah blah.  Blah blah blah blah blah blah.   Just the other day in Ann Taylor Loft several sales girls crowded around him and ohhhed and ahhhed all over him.   Blah blah blah blah.  Blah blah blah blah blah blah blah blah blah.   The kid cracks me up - he is such a flirt.  His favorite thing is watching Winnie The Pooh.   Blah blah blah blah blah blah blah.  Blah blah blah blah blah blah, blah blah   he hands me the remote and \"tells\" me to put on Winnie.    He is getting picky about food now,  blah blah blah blah blah blah blah.  Blah blah blah blah blah blah blah blah blah blah.  I am planning to teach him to swim   blah blah.   I\'ll send pictures.  Anyway, how are you ?  3. grandmothers:  He\'s pretty good actually.  He is so cute.  I just love that kid.  People stop me where ever I go and comment on how cute his hair is.  Or how gorgeous his eyes are.  Just the other day in Ann Taylor Loft several sales girls crowded around him and ohhhed and ahhhed all over him.  Which he totally loved.  Today several checkout ladies in Walmart fawned over him.  The kid cracks me up - he is such a flirt.  His favorite thing is watching Winnie The Pooh.  He doesn\'t like to watch  anything  else.  If we have the TV on, within minutes he hands me the remote and \"tells\" me to put on Winnie.  He is getting picky about food now, but he loves veggies so that\'s ok.  I am planning to teach him to swim this summer.  I\'ll send pictures.  Anyway, how are you?  urlLink >>more    '),(17130,1841346,'09,June,2004','2004-06-09 00:00:00','    Just Finished: Angels and Demons by Dan Brown - The DaVinci Code was better but neither was all that fabulous.  Somewhat interesting if you don\'t know much about the history of religion.  Some of the ideas brought forth in TDC were rather against accepted dogma but nothing that hasn\'t been said somewhere before.  TDC has spawned several books that explain what is truth and was is not with titles like, The Truth Behind the DaVinci Code etc.  I say, get over yourselves.  It\'s  fiction  people - as evidenced by the stupid ending of Angels and Demons.  No other word for it.  Stupid, stupid, stupid.    Currently reading: Fortune\'s Rocks by Anita Shreve Great Expectations by Charles Dickens Mythology by Edith Hamilton  Up Next (in no particular order): Song of Susannah (Dark Tower VI) by Stephen King Memoirs of a Geisha by Arthur Golden Legends of Arthur by Richard Barber The Siege of Troy by Greg Tobin  What are you reading?  urlLink >>more    '),(17131,1841346,'21,July,2004','2004-07-21 00:00:00','    Oh, Mr. Sun, Sun, Mr. Golden Sun, please shine down on me. Oh, Mr. Sun, Sun, Mr. Golden Sun, hiding behind a tree!    These little children are asking you to please come out so we can play with you!  Oh, Mr. Sun, Sun...   I have been watching way to much Barney.  Many parents have warned me in the past.  But I didn\'t listen.  Now my brain is slowly turning to mush.   Oh, Mr. Sun, Sun...   AAAAAHHHHHHH!!!!!   I have also seen way too much Winnie the Pooh.  However, he is not quite as insidious as Barney.  After all, Pooh is a bear of very little brain, so how bad can he be?  But Barney.  Too much is dangerous.  It\'s all the singing.  A new song every few minutes...   The more we get together, together, together the more we get together the happier we\'ll be.  \'Cause your friends are my friends and  my friends are your friends.  The more we get together the happier we\'ll be!    urlLink     Thought I should practice before I hunt him down.   [Disclaimer: The views expressed in this post are not the views of the sane person and wonderful Mother who gets up everyday at 6:00am to press play on the VCR.]  urlLink >>more    '),(17132,1841346,'19,July,2004','2004-07-19 00:00:00','    I\'ve sort of had an on-again off-again thing with  urlLink Alanis Morissette .  Her music is very distinctive but not always accessible.  What continues to draw me back in is her ability to communicate an acute self-awareness of both the good and the bad in herself.  Although, I don\'t think she is an amazing musical talent (although Jagged Little Pill was fantastic), I must say her gift is her ability to write captivatingly profound lyrics.  Often, I relate perfectly to what she writes.    I just purchased,  urlLink So-Called Chaos .  The salesman (overweight forty-something balding dork who wouldn\'t understand female angst given a brain transplant) tells me that, \"This one is really a departure (yes, he said departure) from her other albums, it\'s much more upbeat.\" I just turned around slowly and stared at him until he said, \"Is there anything else I can help you find?\"  That\'s right.  Remember your place, man-dork.  When purchasing Alanis music it is important to have  attitude .  In this latest CD there are only a few songs I appreciate musically, but there are several that have amazing lyrics.  I found one in particular resonated with me immediately (which is something I can  always  count on)...    Out Is Through    Every time you raise your voice I see the greener grass Every time you run for cover I see this pasture Every time we\'re in a funk I picture a different choice Anytime we\'re in a rut, this distant grandeur  my tendency to want to do away feels natural my urgency to dream of softer places feels understandable but I know  The only way out is through The faster we\'re in the better The only way out is through ultimately The only way out is through  The only way we\'ll feel better The only way out is through ultimately  Every time that I\'m confused I think there must be easier ways Every time our horns are locked I\'m towel throwing Every time we\'re at a loss we\'ve bolted from difficulty Anytime we\'re in stalemate, a final bowing  my tendency to want to hide away feels easier The immediacy of picturing another place, comforting to go but I know  The only way out is through The faster we\'re in the better The only way out is through ultimately The only way out is through The only way we\'ll feel better The only way out is through ultimately  We could just walk away and hide our heads in the sand We could just call it quits only to start all over again, with somebody else  Every time we\'re stuck in struggle I\'m down for the count that day  Every time I dream of quick fix I\'m assuaged  Now I know it\'s hard when it\'s through and I\'m damned if I don\'t, no quick fix way What formerly was treatment silent\'s now outdated  My tendency to want to run feels unnatural now The urgency to want to give to you what I want most feels good and I know  The only way out is through The faster we\'re in the better The only way out is through ultimately The only way out is through The only way we\'ll feel better The only way out is through ultimately   urlLink So-Called Chaos  Lyrics  urlLink >>more    '),(17133,1841346,'14,July,2004','2004-07-14 00:00:00','    Out with the photo album and in with the  urlLink Fotopage !  This is great because it has all the features of an online photo album plus I can post text and allow comments too!   I plan to put this link in the sidebar and delete the photo album.  All in good time, just not right now.  It\'s 2:00am and I\'ve got to go to bed!  urlLink >>more    '),(17134,1841346,'14,July,2004','2004-07-14 00:00:00','    This weekend I have the priviledge of attending the  urlLink Buffalo Philharmonic Orchestra\'s  production of Carmen at  urlLink Artpark .  My Aunt sings in the Buffalo Philharmonic Chorus and was able to snag me a ticket.  I am so excited - I can\'t wait!  I had heard of the opera before but did not know exactly what it was about.  So I googled it.  Carmen was created by  urlLink Georges Bizet (1838-1875)  and founded on the  urlLink novel  by Prosper MÃ©rimÃ©e.  The story begins in 19th Century Spain...   ACT I Soldiers and townspeople mill around in a square in Seville. A young peasant girl, MicaÃ«la, asks the soldiers if they have seen her sweetheart, Don JosÃ©. Telling her heâ€™ll be back soon, they try to persuade her to stay with them, but she declines. The relief soldiers, including Don JosÃ©, arrive. Factory bells ring, and a group of cigarette girls emerges from the factory where they work, including the popular gypsy beauty, Carmen. She focuses her attention on Don JosÃ©, who pretends not to notice. Before leaving, she seductively tosses a flower at him. Alone, Don JosÃ© recovers the flower and reflects on Carmenâ€™s charms. MicaÃ«la finds him and delivers both a letter and a chaste kiss from his mother, who asks her son to marry MicaÃ«la. Don JosÃ© promises his love and fidelity to MicaÃ«la, despite the temptations of Carmen. A ruckus erupts from the cigarette factory. Carmen has injured another woman, and the officer Zuniga commands Don JosÃ© to jail Carmen. But Don JosÃ© succumbs to her charms. He agrees to a rendezvous and lets Carmen escape.  ACT II  At Lillas Pastiaâ€™s inn, Carmen and her friends Frasquita and Mercedes consort with Zuniga and other soldiers. A group of revelers arrives, celebrating Escamillo, the illustrious bullfighter. The crowd cheers as Escamillo boasts of his victories. He notices Carmen, but she remains indifferent. Zuniga, also smitten, tells Carmen that he plans to return to the inn later to visit her. When the crowd disperses, the smugglers Remendado and DancaÃ¯ro try to enlist the aid of Carmen, Frasquita, and Mercedes. Mercedes and Frasquita agree to help them smuggle contraband, but Carmen, expecting Don JosÃ©, wants to stay at the inn. Don JosÃ© arrives, and Carmen dances for him. But distant bugles signal him to return to his quarters and he prepares to leave. Carmen mocks his obedience and encourages him to run away with her and lead the free gypsy life. Don JosÃ© remains unconvinced until Zuniga returns to the inn seeking Carmen. In a jealous rage, Don JosÃ© defies his officerâ€™s orders to leave. As the smugglers pounce on Zuniga and escort him out of the inn, Don JosÃ© has no choice but to remain with the gypsies.  ACT III  At the mountain hideout of the smugglers, Don JosÃ© longs for his mother, who still believes him an honest man. Carmen taunts him and urges him to leave, but he refuses. Frasquita and Mercedes tell their fortunes with a deck of cards. When Carmen takes her turn, the cards foretell death for her and Don JosÃ©. The gypsies set off to smuggle contraband, leaving Don JosÃ© behind to guard the camp. Micaela arrives at the mountain hideout searching for Don JosÃ© and hides among the rocks. Escamillo approaches the camp looking for Carmen. He and Don JosÃ© exchange words and begin to fight. But the smugglers return in time to stop Don JosÃ© from wounding Escamillo, who invites them all to the bullfight in Seville. Her hiding place discovered, MicaÃ«la begs Don JosÃ© to return home to his mother, who is dying. Despite his violent jealousy, Don JosÃ© leaves with MicaÃ«la.  ACT IV At the bullfight, a crowd gathers to watch the procession of toreadors. Escamillo and Carmen arrive together. Mercedes and Frasquita warn Carmen that Don JosÃ© is lurking about. Carmen, unafraid, waits alone for Don JosÃ©. He approaches and begs her to leave with him. She insists that their affair is over, that she does not love him anymore, and that she now loves Escamillo. As Don JosÃ©â€™s demands become more desperate, Carmen throws at him the ring he once gave her, with fatal results.  via  urlLink NYC Opera  urlLink >>more    '),(17135,1841346,'13,July,2004','2004-07-13 00:00:00','    urlLink     London 1, The Saatchi Gallery   From  urlLink spencertunick.com    Spencer Tunick is coming to Buffalo.  I have to tell you that when I originally started this post, I wrote, \"Would You Post Nude?\" for the title.  That\'s what happens when you blog late at night.  I think I might.   Might .  Of course, thousands (maybe only hundreds) of other people will do it with me.  You get a free print and a memory that I am sure would last a lifetime.  Check out his  urlLink other photos .  They are fantastic.    Live close?  You can  urlLink sign up to pose .  At the very least, I would like to see his collection at the  urlLink Bodily Space: New Obsessions in Figurative Sculpture  exhibit.  urlLink >>more    '),(17136,1841346,'06,July,2004','2004-07-06 00:00:00','    urlLink     From your family in Pittsburgh! :)   And thank you so much for everything.  It was so fun to hang out with you, Ron and Barb over the holiday weekend.  We all have been going through withdrawal since we got back.  urlLink >>more    '),(17137,1841346,'01,July,2004','2004-07-01 00:00:00','    Michael Moore\'s bloggin\' - read  urlLink his message .  Have you seen  urlLink his latest documentary ?  I guess the better question is do you really want to?  How much agitprop can you take (believe)?  I just read Mike\'s  urlLink Must Read  list.  Truth is I am almost scared to see this movie, but that\'s not stopping me from feeling this strange patriotic obligation.   agitprop \\AJ-it-prop\\, noun: Political propaganda, especially favoring communism and disseminated through literature, drama, art, or music: â€œIt also is a conspiracy movie, agitprop against today\'s targets, big government and big businessâ€ (George F. Will).     Agitprop  comes from Russian, from  agitatsiya , \"agitation\" +  propaganda .   urlLink Source: Dictionary.com  urlLink >>more    '),(17138,1841346,'01,July,2004','2004-07-01 00:00:00','    urlLink  urlLink >>more    '),(17139,1841346,'11,August,2004','2004-08-11 00:00:00','    urlLink     From  urlLink www.vgcats.com    Thanks bro.  This one cracked me up.  Just what I needed.   You kind of have to be a fan of the  urlLink Alien movies  or this probably wasn\'t very funny.  urlLink >>more    '),(17140,1841346,'10,August,2004','2004-08-10 00:00:00','    Found this site that seems dedicated not to a particular political ideology but to facts.  Imagine that.  Trying to cull the provable facts of what happened - not the spin usually fed us in the media. Their mission statement: We are a nonpartisan, nonprofit, \"consumer advocate\" for voters that aims to reduce the level of deception and confusion in U.S. politics. We monitor the factual accuracy of what is said by major U.S. political players in the form of TV ads, debates, speeches, interviews, and news releases. Our goal is to apply the best practices of both journalism and scholarship, and to increase public knowledge and understanding.  The Annenberg Political Fact Check is a project of the Annenberg Public Policy Center of the University of Pennsylvania. The APPC was established by publisher and philanthropist Walter Annenberg in 1994 to create a community of scholars within the University of Pennsylvania that would address public policy issues at the local, state, and federal levels.  The APPC accepts NO funding from business corporations, labor unions, political parties, lobbying organizations or individuals. It is funded primarily by an endowment from the Annenberg Foundation.  urlLink >>more    '),(17141,1841346,'09,August,2004','2004-08-09 00:00:00','    In an effort to stay a bit more current with world events I went looking for a way to manage all the news sources I want to read.  I have been trying out different free Web based aggregators.  They pull the stories/posts from whatever feeds you specify so you can view them all in one place.  The  urlLink Feedster  site is not always reliable (I found pages often loaded incorrectly and were not updated), however it does have a cool feature called  urlLink FeedFinder  which allows you to check if a specific site offers a feed.  I switched to  urlLink Bloglines  which is better organized and offers more services.  I plan to check out  urlLink more newsreaders  in the future.  So, check back for updates.   Wondering what the heck a feed is?  It\'s all about content syndication.  There are different formats but  urlLink RSS  is the most widely used at present.   urlLink Creating an RSS feed  is actually pretty easy.  I found a free software tool that will convert  urlLink RSS to JavaScript  which allows you to include feeds in your website.  Then there is  urlLink Feedroll  which is a free online service for syndicating RSS and ATOM news feeds on your website or blog (Bloglines offers this as well).  urlLink >>more    '),(17142,1841346,'04,August,2004','2004-08-04 00:00:00','    As I sit on the plane still at the gate in LA, I watch my son warily and wonder how I can calm him down.  He is bouncing off the walls.  I think this is all well and good but soon he will start giggling.  He always gets giddy before he hits overtired.  Then wham!  You\'re in parent hell.  I reflect that I could get him in the  urlLink Guinness Book of World Records  on sheer decibel alone when he throws an I-am-overtired-do-something-mom-I-want-I-don\'t-know-what-I-want-definitely-not- that -I-can\'t-stop tantrum.   We finally taxi to the runway and the captain informs us that we will arrive in Pittsburgh in about 2 hrs and 16 minutes - 1:06am Pittsburgh time.  Bad news.  No car at the airport. We missed the last bus.  No one to pick us up.  We will have to take a taxi to the bus stop where Christian\'s truck is.  My son starts laughing.  It is unclear what has amused him.  I tell Christian a tantrum is eminent.  As we pick up speed for take off, I put Hayden in my lap.  Immediately he begins fighting, kicking and screaming.  I almost feel sorry for my husband as he has not experienced the special horror of a full-blown tantrum within a confined space packed with judgement and scorn.  Thus begins the game of Hayden Hot Potato.  He literally bounces back and forth between us and we alternately ssshhh and stroke him then give that up by attempting to cover his mouth.  Everything we do is in vain.  This goes on for thirty minutes.  During which time I am simultaneously worring how miserable my son is and when Mr. JerkMan will start something all the while receiving what I call Completely Useless Helpful Hints from several people including a stewardess.  I finally turn around include all in my gaze and say very loudly, \"Thank you but that won\'t help.  He is overtired.  If I knew of something that would work, I would certainly use it.\"  Mercy desends many minutes later when he finally lets me rock and sing him to sleep.  The quiet is wonderful.  As I watch the moonlight play over the landscape I begin to notice that my tailbone has started to throb.  I am unable to sleep but I endure and we make it to Pittsburgh.  In the airport trucking to baggage claim.  The only bright spot in this whole ordeal is that Mr. JerkMan\'s foot is run over by someone else\'s stroller while we wait for the tram - take that you baby hater.  We get the luggage, hail a taxi and load up.  I start watching the meter.  We only have 35 dollars between us and we have no idea how much this will cost.  Fifteen minutes later we arrive at the bus stop.  We locate the truck, unload the luggage and pull Hayden out of the taxi.  And then Christian turns the key in the ignition and nothing happens.  He tries again.  It cranks but it won\'t turn over.  He keeps trying but the truck won\'t start.  It\'s now approaching 2:00am.  Resigned, we pile back in the taxi.  After ten more minutes and a nice conversation with our driver who is from Kenya we are at our front door.  It is finally over and we are home.  urlLink >>more    '),(17143,1841346,'03,August,2004','2004-08-03 00:00:00','    The trip to Texas was wonderful.  But I was ready to go home.  Missed my bed.  Everything started out so well.  Made sure Hayden had his nap.  We even got the bulkhead seats - perfect for Hayden to play at our feet.  Not sure why it went so horribly wrong.  It all began about 40 minutes into the flight....   The captian comes over the loud speaker and announces that there is a small oil leak.  Not a big problem, there is plenty of fuel, but procedure requires maintenance take a look.  We will be diverting to Shreveport, LA were there is a maintenance airport waiting to receive us.  It is then I notice a wait-a-minute-did-you-say-diverting-is-everything-ok-with-the-plane-god-what-if-we-crashed-geez-why-am-I-thinking-about-9/11 air about the plane.  Ok, that was just running through my head.  Twenty minutes later as we begin our descent into LA, the captain announces that we will be circling Shreveport for about an hour to burn enough fuel to make the plane regulation landing weight.  Suddenly, I realize an argument is taking place behind me.  Then Christian turns arounds and yells something to the effect of, \"So what!  He\'s a baby.  You were a baby once remember?!\"  Huh?  I look around to find a jerk of a man complaining very loudly about my son.  He is sputtering something along the lines of \'that is the worst child I have ever seen, most babies aren\'t that loud\' etc.  There is a man couple rows back who apparently was defending my son but I didn\'t hear what he said.  The stewardess shows up and tells my husband to ignore him.  That\'s what she said with some disgust I might add, \"Just ignore him.\"  Hayden is not crying.  He  is  talking rather loudly however.  Christian and I have a tense few minutes of attempting to teach our son the meaning of \"SSSHHH!\"  To no avail, of course.  Then, BAM!  What-the-heck-is-that-is-everything-ok-with-the-plane-god-what-if-we-crashed-geez-why-am-I-thinking-about-9/11!  The Captain kindly tells us that he had to put the landing gear down to create extra drag which would aid in the consumption of fuel.  Nothing to worry about, it\'s just very loud.  And so began the monotony: straight, bank right, straight, bank right, straight, bank right...six times I counted before we were able to land.  We land smoothly and I think, \"Terra firma!\"  We are allowed out of the plane but are confined to the gate as the security is closed down.  As fervent rest room and cell phone activity begins, I resolve to ignore  those looks .  Yanno, the ones you get from people who think you are a bad parent.  Fortunately, one feels so much better after peeing and calling no less than four people to grip about one\'s current misfortunes.  The announcement.  The captain informs us that, \"blah, blah, blah, blah, blah, sorry for the inconvenience.\"  I found out from another passenger that the oil leak in the wing was entirely a figment of the captain\'s imagination and is most likely condensation therefore it is time to board.  I turn to Christian and say, \"I vote for another plane.\"  Because if we are going to be cautious - you know the adage: better safe than sorry - then let\'s commit completely to it.  As luck would have it, another plane wouldn\'t have made any difference because sorry it was to be.  urlLink >>more    '),(17144,918175,'05,November,2002','2002-11-05 00:00:00','   Panaghuy   Bakit ka umiiyak? Bawat patak ng luha mo’y Nilusaw ang liwanag  hatid ng nag iisang bituin  Ng langit ba ika mo?  Marapat bang sabayan  Ng maka lumang lab song Ang bawat hagulhol ng puso?  Malungkot ka ba? Bakit me bahagyang kislap ang ... Teka, ‘kaw ba ay nangiti? Inagos na ba ng luha ang nakakamatay  Mong pighati?  uMiyak ka pa.  '),(17145,918175,'07,January,2003','2003-01-07 00:00:00','   mahirap magmahal  nang nag-iisa.  o baka siya man ay lito na rin?  '),(17146,918175,'13,February,2004','2004-02-13 00:00:00','     '),(17147,918175,'19,February,2004','2004-02-19 00:00:00','   ascend  and detach the soul from the body detach the mind from the heart  must i die  to forget?  '),(17148,918175,'30,March,2004','2004-03-30 00:00:00','   you punish me with your thoughts     the memory of your smell     stabs me like a sword would     and kill.           the pain prolongs.  '),(17149,918175,'07,May,2004','2004-05-07 00:00:00','   read somewhere...      i wish i could carry your smile in my heart   hmm..liked it a lot.  '),(17150,918175,'03,May,2004','2004-05-03 00:00:00','   so irresistible..  '),(17151,918175,'28,May,2004','2004-05-28 00:00:00','   doubtful happy doubtful jealous happy jealous jealous envious happy happy  i fear  i want you   happy  '),(17152,918175,'04,June,2004','2004-06-04 00:00:00','   pain.  '),(17153,918175,'07,June,2004','2004-06-07 00:00:00','   copy all to..[dream]  '),(17154,918175,'15,June,2004','2004-06-15 00:00:00','   TO: zork Subject:	Re: HOIST  yeah rught?? like i am hearing these from the elder sister who is suppose to take care of us in absence of the parents but chose to live far away and left us hanging, like we\'ll never kknow if we\'ll ever have a chance to see each other again?  wal man gud jud ka kabalo how i feel? because you were never there from the start anyway. didto pa sa pikas balay, you never knew everything because all you were busy about are your friends, do you even realize how lucky are you kay daghan ka frineds? ako i dont even have friends as a matter of fact.. i live alone knowing and thinking that nobody even cares if i die. ikaw lang gani akoa maka istorya, i live in distress i dont know where i am heading becuase of this misery. do you even realize what you and the hell family took away from me? happiness and love for mysel\\f. kabalo ba ka unsa ko? kabalo ba ka unsa akoang gibati? honestly i envy you, so much... ikaw ang wala dirir but eveybody was adoring and wishing for your presence!!! si papa, walay lain ginatext kundi ikaw, kamusta ka because you never took courage to even speak to him. i saw hoe you dared stoop down on him when you were here, how you dared hating him and there you are telling me these???? how could you even forget. si bob? kabalo ba ka how lonely he has been? how broken ang iyahang pagkatao but i cannot reach our to him,b ecuase from the start maski katong gagmay pa ta wala jud sha garespeto sa ako. do you even see that??? maski kay mama wala shay respeto,w ala ko kabalo unsaun sha but i have to keep quit because my voice is never heard and it has always been my life inthis bulshit family na wala koy kwenta. latagw na anak ug pirmi ginakasab an. not even our relatives respect me. akoang work. how i dreamt of leaving but i cant. i am not even performing. how dare you advise me na dili maghate kay papa when all these years ako lang ang nabilin para sa iyaha only to realise how unworthy of him of my love. kamo tanan. i hate all of you!! i wish you all dead.  all this time ikaw akoang gina praise, i look up to yout and how i believe in you. all these years sa skwela, ikaw akoang gina sundog. all this time sa imoha ko na ga duol pag lonely or even sad. but reading your message, kinsa sa atoa ang nagpabaya. si papa nihilak na sa akoa nagapangandoy sa imo kay ikaw ang paborito niya, pero unsa? mangumusta sa imo dili man gani nimo makuha i text, or kung naa pangayuon murahon mo pa. shit! shit mo tanan. sa sandawa hilas kaayo ang mga taga pikas walay lain gibuhat kundi dautan kop, dautan akoang mama ug akoang papa ug si bob. ikaw? wal ka gisautan, kay ngano wala man ka nabuhaytt. muuli ka diri dili pa para sa amoa kundi para sa iimohang mga friends. gamhanan ka! you and this bulshit family took away alot of things from me, you even took my confidence. how envoius and jealoud i have become.  tama na kay nakapahungaw nako. text me na lang kay kasab anan pako sa akoa boss. thanks.      TO: zork\'s sister from : zork  Subject:    HOIST       naunsa ka? nasuko ka kang papa?  ayaw ana bang, maski wala siya ka pa eskwela kang bob and naging irresponsible siya (and this has never been okay with us) - don\'t be judgemental.  naglisud pud gud to bah. naguol unsaon niya pagbuhi sa iya anak (mga anak). that is kung naguol man siya. cguro nangita si papa ug laing pamilya, kanang parehas sa ato sa una. mao nangipon siya ug lain. tapos ging anakan pa jud..  so, be kind. be yourself! nyahahhaa! dili bitaw oi..i pray na lang na..basta ayaw lang jud pag palabi sa imo kaluoy and  generosity because ma disappoint lang ka. dili man gud ma iwasan to expect something back di ba? lalo na kun sa parent nimo i-offer.  eh teka..busog kaayo ko..mag tsa-a sa^ ko..  mag email ko usab ha.kay haay naglagot ko ..  '),(17155,918175,'14,June,2004','2004-06-14 00:00:00','   i was about to write you about my anger and my insecurities, selfishness and jealousy. until i read this story from my friend.   this is a better thought than my angst.   \" a sweet love story\"   > From the very beginning, the girl\'s family  > objected strongly on her dating > > this guy, saying that it has got to do with  > family background, and that the > > girl will have to suffer for the rest of her  > life if she were to be with > > him. > > Due to family\'s pressure, the couple quarrelled  > very often. Though the girl > > loved the guy deeply, she always asked  > him: \"How deep is your love for me?\" > > As the guy is not good with his words, this  > often caused the girl to be very > > upset. With that and the family\'s pressure, the  > gal often vents her anger on > > him. As for him...he only endured it in silence. > > After a couple of years, the guy finally  > graduated and decided to further > > his studies overseas. Before leaving, he  > proposed to the gal: \"I\'m not very > > good with words. But all I know is that I love  > you. If you allow me, I will > > take care of you for the rest of my life. As  > for your family, I\'ll try my > > best to talk them round. Will you marry me?\" > > The girl agreed, and with the guy\'s  > determination the family finally gave in > > and agreed to let them get married. So before  > he left, they got engaged. > > The gal went out to the working society,  > whereas the guy was overseas, > > continuing his studies. They sent their love  > through emails and phone calls. > > Though it was hard, but both never thought of  > giving up. > > One day, while the gal was on her way to work,  > she was knocked down by a car > > that lost control. When she woke up, she saw  > her parents beside her bed. She > > realized that she was badly injured. Seeing her  > mum cry, she wanted to > > comfort her. But she realized that all that  > could come out of her mouth was > > just a sigh. she had lost her voice.... > > The doctor says that the impact on her brain  > has caused her to lose her > > voice. Listening to her parents\' comfort, but  > with nothing coming out from > > her, she broke down. During the stay in the  > hospital, besides silent > > cry...it\'s still just silent cry that  > accompanied her. Upon reaching home, > > everything seems to be the same. Except for the  > ringing tone of the phone > > which pierced into her heart every time it  > rang. She does not wish to let > > the guy know and, not wanting to be a burden to  > him, she wrote a letter to > > him saying that she does not wish to wait any  > longer. > > With that, she sent the ring back to him. In  > return, the guy sent countless > > replies and  phone calls.. all the gal could do  > besides crying is still > > crying...The parents decided to move away,  > hoping that she could eventually > > forget everything and be happy. With a new  > environment, the gal learnt sign > > language and started a new life. > > Telling herself everyday that she must forget  > the guy. One day, her friend > > came and told her that he\'s back. She asked her  > friend not to let him know > > what happened to her. Since then, there wasn\'t  > anymore news of him. > > A year has passed and her friend came with an  > envelope, containing an > > invitation card for the guy\'s wedding. The gal  > was shattered!  When she > > opened the letter, she saw her name on it  > instead. When she was about to ask > > her friend what was going on, she saw the guy  > standing in front of her.... > > He used sign language to tell her, \"I\'ve spent  > a year to learn sign > > language. Just to let you know that I\'ve not  > forgotten our promise. Let me > > have the chance to be your voice. I Love You.\"   > With that, he slipped the > > ring back into her finger. The gal finally  > smiled...... > > * Treat every relationship as if it\'s the last  > one, then you\'ll know how to > > give. > > * Treat every moment as if it\'s the last day,  > then you\'ll know how to > > treasure. > > * Treasure what you have right now, or else you  > may regret one day...  '),(17156,918175,'25,June,2004','2004-06-25 00:00:00','   aahh namimiss ko ang elbi. :( lagi na lang pag gus2 ko magmunimuni o kumain ng pinakamaanghang na siomai, gus2 ko dumayo ulet sa elbi. kakaiba ang tlga sa elbi.   namimiss ko yung nagmomotor kami sa loob ng UPLB campus, naglilibot lang..ninanamnam ang malinis na hangin at malawak na kapaligiran. ang paborito kong puwesto ay yung daan galing UP papuntang IRRI. tipong yung mga punong nakahelera sa daan eh nakadungaw sa inyo habang nagmomotor kayo at malakas ang dampi ng hangin sa mga mukha nyo. tapos pagdating sa IRRI yung mga palay parang naka-bow pa sila in honor of your arrival.   pinangarap ko dati ang magtrabaho sa IRRI para lang matira ako sa elbi. pero natakot akong mag apply dun nung makita kong parang walang social layp ang mga scientist dun. baka gawin akong guinea pig  mwehehhehehe..sa totoo lang nde naman tlga ako nakapasa eh bwuahahaha! blessing in disguise na rin cguro yun.   at ang pinaka mimiss ko sa lahat eh ung siomai sa maliit na tindahan sa me bungad ng UPLB. saraaappp!!!  sa lahat ng probinsya sa Pilipinas na narating ko, pinaka papangarap ko ang matira man lang ng mas matagal tagal pa sa elbi. aywan ko ba, me kakaibang magic ang elbi.   kailan kaya ako makakabalik dun?  '),(17157,918175,'24,June,2004','2004-06-24 00:00:00','   finally my PAG-IBIG loan voucher arrived today. gotta have that shoess..gotta have that shoes..  ***  and again, i\'d rather stay home and cuddle with the rain outside than play badminton. hmmm maybe next week i\'d bring him to the court with me and afterwards we\'ll do the cuddling.   ***  sweet talks.  '),(17158,918175,'24,June,2004','2004-06-24 00:00:00','   why get mad?  '),(17159,918175,'01,July,2004','2004-07-01 00:00:00','   ARTS LAB   DAVAO CITY (MindaNews) - Take nine days off in July this year to learn history, music, dances and craft of the Maranao, Maguindanao, Tausug and other peoples of the Autonomous Region in Muslim Mindanao.  KulArts, Inc. and the Mindanao State University Cultural Affairs Office through the Sining Kambayoka Ensemble are hosting a Cultural Exchange and Exposure Journey dubbed as \"Kulintang Arts Lab\" on July 1 to 9 in Marawi City, Lumba Bayabao and other venues in Lanao del Sur.  \"Kulintang Arts Lab\" is a newly-opened program that explores traditional artistic expressions centered on Kulintang music and related fields of peoples in the ARMM.  The multi-disciplinary research and development programme in theatre, traditional dance and music of Maranao, Maguindanaoan, Tausug, Samala and other Moro communities, will include components on visual arts, films and rituals curated by Moro contemporary artists.  The program will have classes, workshop explorations and nteractive lectures.  \"This laboratory, the first with such extensive scope and continuity looks at the different creative strategies of individual artists, both traditional and contemporary, thorugh the recognition of differences and similarities between the many Moro communities in Mindanao,\" the program\'s press statement said.  It added: \"the focal point is on cultural negotiation and exploring process in cultural work - the heart of managing diversities in Mindanao, aided by this clash of time/space through the dynamic interaction between the traditional performers, Moro and Non-Moro contemporary artists and the Fil-American community from San Francisco, California in view of the development of artist in a more dynamic muscularity, understanding one\'s roots and advancing the Filipino expressions towards a common ground.\"  Day 1 of the arts lab will have Sining Pananadem, an all-Maranao traditional performing group welcome the participants.  Day 2 will feature lectures on Maranao Culture and Maranao Dance and Music by Prof. Minang Shariff. Shariff will discuss Maranao historical perspective and Sultanate genealogy, traditional practices in the morning and will conduct a workshop on Maranao dance in the afternoon with the members of the Sining Pananadem or Kambayoka.  Shariff will also conduct a basic conversational Maranao lecture in the evening.  Day 3 will feature lectures on Maguindanaon historical background and genealogy of the sultanats, Maguindanaon cultural tradition, beliefs and practices and Maguindanao legends and folklores by Faisal Monal.  In the afternoon, the Maguindanaon Musical Ensemble will conduct and interactive demonstration with Amal Lomontod on the kulintang, Datuan Kalanduyan on agong and gandingan, Datuan on kudyapi, Noraisa on dabakan, Datuan Kalanduyan on tumpong (flute), Akmad Siao on babendil.  An interactive demonstration on Maguindanaon dances will also be conducted by Noraisa Kalanduyan and Akmad Siao.  Day 4 will bring participants to Sulu and Tawi-tawi, at least through these provinces dances and music.  An inter-active demonstration will be held on the dance and music of the Sulu archipelago and Tawi-tawi, as well as lectures on Samal, Tausug, Badjaw and the Jama Mapuns by Taalao Manandaw, director of the MSU Cultural Affairs in Bongao, Tawi-tawi and Saki Nur Ayin Delasas, director of the Tambuli Cultural Troupe also from MSU in Tawi-tawi.  A film showing will be held in the evening featuring \"House under the crescent moon\" by Teng Mangansakan. The film was adjudged the Best Video Documentary and won the 2001 Gawad CCP for Independent Film and Video  Mangansakan will also be the main speaker in the forum on \"The Bangsamoro struggle for self-determination\" whild Geejay Arriola of Earth Music Foundation, who conceptualized and directed \"Salima,\" a multi-media concert theatre presentation on the evacuees of Mindanao from the point of view of a child, will lecture on Peace and Development.  Day 5 will have the participants visiting Togaya for the brass casters, Maranao sculptures, (treasure) chest makers, lunch in the Marawi market, travel to Tuka to visit weavers and visit the Maranao market in Marawi City.  In the evening, a film showing will again be held, this time, featuring Mangansakan\'s \"A boat in deep waters\" on the history of the Bangsamoro women in the last 100 years through the lens of the memory of Bai Tinomimbang Pendatun-Matalam, wife of Datu Udtog Matalam, founder of the Muslim Independence Movement.  \"Daud,\" a video-essay installation portraying a child mujahideen\'s thoughts on peace and war will also be presented by Mangansakan.  Day 6 will feature lectures by Teng Mangansakan on \"The life of Tao sa Laya\" or People of the Upper Valley and Philipp P. Matalam\'s \"The living legacy of the Sultanate of Buayan.\"  An interactive demonstration and continuation of Sama-Tausug workshop will also be held.  In the afternoon, Matalam will conduct a lecture demonstration on the usage of the Maguindanaon malong, designs and patterns of Maguindanaon Inaul (malong) and Maguindanaon hairstyling and accessories.  A workshop on the Kambayoka theatre form will also be held in the evening.  Day 7 will have participants traveling to Lumba Bayabao, home of traditional Maranao artists. An interactive demonstration with Maranao traditional artists will be held prior to a Pagana Maranao at noon. A workshop on music and dance follows.  Day 8 will feature dance and music intensives through small group workshops. In the afternoon, Sister Delia Coronel will lecture on the Darangan epic followed by Alleluia Panis\' overview of post modern Filipino American expressions.  Day 9 will have participants visiting Aga Khan Museum at the Mindanao State University campus. In the afternoon, they will observe how to prepare a Pagana Maranao, facilitated by Cecil Mambuay.  A Pagana Maranao will start at 6 p.m followed by the closing ceremonies which will feature a performance showcase of \"Two Rivers: Mindeping a Pulangui\" written by Mangansakan, curated by Carlo Ebeo and Geejay Arriola with set design by Cris Rollo.  Registration Fees are P15,000 for Manila-based Filipinos and P8,000 for Mindanao-based Filipinos. The fees are inclusive of food, accommodation, translations, and inland travel in Lanao del Sur only.  Deadline for applications is June 20. Applications will be on a first come, first served basis and applicants will go through a screening process.  Filipino applicants are urged to e-mail ucarloebeo@y... while foreign participants can go to www.kularts.org. (MindaNews)  '),(17160,918175,'29,June,2004','2004-06-29 00:00:00','   finally! the sun\'s up with a smile. not too much and it still is windy but at least we got a rest from the rain.  ***  PUNDOKTINYO\'s moderator (my former editor from my college student publication ATENEWS) actually got me in the group\'s mailing list. It\'s  a writers group for atenews alumni so we can all get together and touch base at least while online.  -MiSael   so while the moderator\'s now on the hunt for other members, i\'m creating a photo album..i just wanna share to them all people what i\'ve been busy with after college.  '),(17161,918175,'29,June,2004','2004-06-29 00:00:00','   it\'s raining cats and dogs. i miss my bed. i miss you.  '),(17162,918175,'28,June,2004','2004-06-28 00:00:00','   what do you do with all the gift certificates for a first class gym?? spend 2 hours in the sauna! huh! it was so great i could feel my sweat ran from my armpit to my legs..oh la la la!  the shower room have it\'s own shampoo/conditioner and body wash. the shower was so great i felt my skin dry up for spending too much time washing.   the locker room was not crowded that  sunday afternoon so i had the mid section all to myself. what with the large mirrors and hair blowers and AC on, it was sooooOOoo tempting to walk around butt naked.   the gym itself was fantastic. first, they have all types of soda and iced drinks free bottomless! huh! then the lounge area was loittered with ahemm..handsome pretty young men. mwehehehhe. deceiving young men oggling at other young men bwuahahaha! at least they\'ve got the bod and the looks to at least feast at.   second, the machines looked monstrous. you\'d think that after using them for a bit ten minutes, you\'ll have all the muscles toned up in places you\'ve always dreamed of toning..like your ahem..chin? but then again, this person next to you catching his breath lifting a 50 lbs \'monster\' is himself breath-catching.   i\'ll be back wednesday to attend to their group classes. got to decide from yoga, pilates, body balancing, body stressing, how about body undressing? huwaw!  i will never question, never again, why people with so much money could pay P3600 a month for a fitness program. and spend hours in the gym. i mean .. and spend hours in the gym.  '),(17163,918175,'07,July,2004','2004-07-07 00:00:00','   my sister resigned from work and moved out of her house to live in with my Mom. she set a one-month period to trim down her weight. all this for her application to japan. she wants to work in japan as an entertainer. (set aside moral issues, that would be another harbored topic)  she\'s only 25 and she wants to be financially able by the age 30. and so japan, she sees is the answer to all her wants and needs.   then there is also korea. if she opts for korea, she will then have to apply as an english teacher. since korean entertainers earn as much as tutors. (whatever!)   the bad news is, if she decides to go to korea. she wants me to join her. (urgh!)   do i need this? korea? money?   i could settle for a large McDonald\'s fries. i don\'t mind earning my meager salary as long as i\'m peaceful wherever i am. but korea?   is that in mars?  '),(17164,918175,'06,July,2004','2004-07-06 00:00:00','   i cooked my own sinigang na bangus for the first time in mah layp! huwaw!  '),(17165,918175,'16,July,2004','2004-07-16 00:00:00','   finally after 4 days of rendering 3 hours of OT, the script i\'m working on passed final QA. the bad news is, this is not even my work. it\'s someone\'s else\'s issues and i get to clean it up for her. or him.     &nbsp;  demmet.     &nbsp;  the good news is, my own task\'s due date\'s moved to..tadaaan..NEXT WEEK! yippeeee!    &nbsp;   &nbsp;   &nbsp;   &nbsp;  '),(17166,918175,'13,July,2004','2004-07-13 00:00:00','   urlLink www.friendster.com\'s  block from my office. demmet!  '),(17167,918175,'12,July,2004','2004-07-12 00:00:00','   what do you do when you are scheduled for a 9 am interview and you\'re not called after a four-hour wait?     &nbsp;  simple. call it quits.     &nbsp;  if they don\'t have the time for you, then they are not worth your time and skills.     &nbsp;   &nbsp;  exactly what i did. i was hungry, cold and developing a headache. i do&nbsp; not have the time to waste on very unprofessional environment.     &nbsp;  yes, i am a quiter after all.     &nbsp; &nbsp;  '),(17168,918175,'23,July,2004','2004-07-23 00:00:00','   Fertility Tree                              The headlight blurred my vision; I raised my right arm against my eyes and watched as the approaching vehicle turned to my left. It was one of those mini-buses’ that roamed the campus. It stopped a few meters from my tree. Stopped and headed off. There was nobody. I am alone again.                             I sat on the grass under the tree. What kind of tree? This one was large, the board with it’s scientific name on it was torn years ago, only a nail that marked it that It was once named, remained, protruding inviting danger. Rusty and inviting .  This rusty nail faced against wind and storm and sun and it grew old with the tree. The tree and the nail, now nameless.                               I sat on the grass under the tree.    What kind of a tree? I do not know. The wind blew and I felt it cold.    School is about to close; the year is about to close. The store across the street is about to close. There were a few students lounging along on the grass, but no, there was no one with me under the tree. Another car is approaching my way; it’s the third in ten minutes.    The campus is about to close.                              On the grass, under my un-known tree, I watched as the sun changed the color of the skies. In a few minutes, it will be dark soon. But look, see, the sky is turning to orange, then red, then blue. Then darkness. The campus is about to close. I watched as those who lounged a few meters from me decided to go anyway.                              The year is about to close. The wind is cold. The field in front is empty.    I sat still, now on a large root under my tree.                                 My phone rang, three times I think, maybe more. I felt it. I reached and didn’t recognize the number. “What time is it?” I asked.    “What? And now you think you’re some place else, maybe the North Pole?”    The voice on the other line retorted, then laughed so hard.    I laughed along. “Damn phone call.” I murmured. The voice said goodbye finally and left me hanging, thinking what was the call for? Damn phone call.                              The wind felt cold and my heart murmured. Damn phone. I shouldn’t have brought it with me. I never wanted to own one anyway.    If they wanted to reach me, they could go ask for my home phone instead and we could all talk sensibly at the comfort and privacy of my room.    They could call me at night, early morning, any time.    There’s always a line for that. But not this phone. I don’t want a call, I don’t need a phone.    But I did buy one.    Damn phone.                             The night is closing in, should I call?  '),(17169,918175,'22,July,2004','2004-07-22 00:00:00','   not telling the truth is the ultimate sin to me. talk or never mention it at all. should you plan to deny anything at all, make sure that i have no hint of knowing nor of doubting. my questions will probe even your darkest secrets. and my questioning will hurt us both that it is too painful to handle.     therefore, hide or stay quiet.  '),(17170,918175,'21,July,2004','2004-07-21 00:00:00','   the sky seems to cry with me. my heart aches, my mind is a mess. what i know is that i do not know.  I am not at peace.     i wish to soak in the rain and wash away this burden.  '),(17171,918175,'20,July,2004','2004-07-20 00:00:00','   anong makati ang masarap kamutin?  '),(17172,918175,'06,August,2004','2004-08-06 00:00:00','   i finally found a new home ==>  urlLink http://headspot.blogdrive.com   visit me here..thanks!  '),(17173,918175,'06,August,2004','2004-08-06 00:00:00','   i am thinking of moving my blog. so anyone could post a message to every entry. with blogger, you need to sign up before you can post a message per entry.       i will let you all know when i am done with the new one. :d       thanks for hanging by  '),(17174,3464484,'29,May,2004','2004-05-29 00:00:00','   Being a bank teller is a shitty job.  Iâ€™d love to say that it just sucks, or that itâ€™s crappy, but without the actual curse word, it just doesnâ€™t do the situation the justice that it deserves.  SHITTY.  Iâ€™d get more vulgar, but then Iâ€™d almost seem as if I could advance myself no further than the likes of most of the customers that the poor bank tellers wait on, on an everyday basis.    First of all, bank tellers do not feel as if the customerâ€™s money is their own money.  We donâ€™t care if you want to take out five cents or five million dollars.  However, if there is a problem with your account, for exampleâ€¦THE MONEY ISNâ€™T THERE, then please, donâ€™t suggest that WE did something with it.  I can assure you, we did not.    Secondly, donâ€™t come and ask us for refunds on your own mistakes.  If you have a hundred dollars and you spend two hundred, YES you will receive a fee.  Think of it this way,  the bank is lending you money for being A. Stupid  B. un-mathematical  and C. irresponsible.  They are also covering your ass by paying that business what you owed them, without having the â€œreturned checkâ€ charges businesses will hit you with in addition to overdrawn account fees.  Thank the bank for doing that for youâ€¦ITâ€™S A BIG FAVOR.  Thirdly, if you open an account in which you receive special privileges for maintaining certain balances, please donâ€™t come in complaining that you received a small fee for falling below those balances. Understand that the privileges received in those â€œspecialâ€ plans are NOT given to the people who have the regular checking accounts and donâ€™t receive those fees.  OH, and if youâ€™re in one of those â€œspecialâ€ plans because loans are a part of your combined balances, do notâ€¦I REPEATâ€¦do not come in and act high and mighty to the tellers because you are in the â€œmost special of all the plansâ€ because we know your game, you own more than your life is worth.    Fourthly, donâ€™t question the tellers as to the jobs they are doing.  If they are slow, or confused, that is one thingâ€¦but if we ask you to sign the check, or that we canâ€™t cash it because the date on it is three years ahead of the present date, we are doing our jobs.  Doing anything else would result in our terminationâ€¦and if we are working as a teller, then we pretty much need our money DAMN badâ€¦so let us do what we are there for.    Fifthly, swearing at us will get you nowhere.  NO, I donâ€™t care if you â€œfucking wipe your assâ€ with your check, Iâ€™m not going to cash itâ€¦and after that Iâ€™m pretty sure no one else will either.  NO, I donâ€™t care who your mommy and daddy are, if YOUR account is overdrawn, mommy and daddy probably want nothing to do with it.  If they do, thatâ€™s their fault and youâ€™re a spoiled brat.  NO, I will not fill out all of your paperwork, add up your checks and then figure out how much money you need for the weekend.  That is your job, you should be able to figure out your money situation.  We are there to hurry through transactions and get to the next person in line.    Last...and definitely not least, PLEASE don\'t throw your college education, your masters, your doctorate, your WHATEVER in our faces, and then NOT understand why your deposit is screwed up due to YOUR failure at filling out a deposit slip.  It\'s a simple task, it shouldn\'t take more than one time after being shown how to do it to fill it out correctly.  A lot of us bank tellers are college/grad students working our way through, and thinking to ourselves \"THANK GOD I\'m getting an education, and the common sense of filling out a simple deposit slip\".  If you don\'t know how to do it, just ask, most of us love to show you how to do it, assuming that you\'ll remember the next time you come in.    I am not a bitch.  Well, I am kind of a bitch, but that\'s because of situations mentioned in the above commentary.  I believe that all people should take the time to work in some sort of service industry....bank tellers, waiters/waitresses, retail..etc, just to see how HORRIBLE people can truly be.  If ALL people were made to work, or at least try out a service position, I think that the world would be a better place because there would be a lot less assholes bringing workers to tears due to common everyday mistakes and problems.  I think that people would be a lot more understanding of the difficulty of the jobs of the people that service them on a daily basis and that hopefully those assholes would discontinue being, well, assholes.  This is in no way ignorant of the lazy bank teller that sits and picks at her nails, or checks his baseball scores on the internet, or pretends to be on the phone when there is a line out the door.  I hate those tellers too...in fact, those tellers should be made to see what it\'s like to stand in a god-awful 10 person line and watch the people behind the counters mess around.    What goes around comes around people.  Keep that in mind next time you bitch out the teller (waiter/waitress/attendant/clerk) who wants to be there as little as you do. Tip well, smile, and be kind.  It\'s not that hard...even if it\'s a stretch.   '),(17175,3464484,'26,May,2004','2004-05-26 00:00:00','   Have you ever had an argument that you believe to be very interesting, and try as you might, anytime you try to argue for that thing that you believe so greatly in, no one cares?  I\'ve decided to present that argument as my first entry into this new world of blogging.    Sad love songs.  I hate them, and I love them.  I can listen to them no matter what the mood I am in.  Does this make me sappy?  Possibly, but I\'m down with that.  One song, that will bring me to tears everytime I hear it, is \"Can\'t Make You Love Me\" by Bonnie Raitt (forgive me if this isn\'t the original singer of the song, but it is the version I\'ve always heard).  The lyrics are incredibly simple, let me break them down:    Turn down the lights, turn down the bed Turn down these voices inside my head Lay down with me, tell me no lies Just hold me close, don\'t patronize - don\'t patronize me   First time listeners can handle these lines just fine.  Sure the patronizing part stinks a little bit, but like I said, until we know what comes next, we can all stomach these lines just fine.   Cause I can\'t make you love me if you don\'t You can\'t make your heart feel something it won\'t Here in the dark, in these lonely hours I will lay down my heart and I\'ll feel the power But you won\'t, no you won\'t Cause I can\'t make you love me, if you don\'t   WOW.  I mean come on.  Complete dejection, but with dignity.  Unrequited love has a way of stinging so hard, Bonnie must know this cuz the way her voice sounds in this song makes me feel her pain. The speaker understands why his/her lover doesn\'t/can\'t/won\'t love back, but it makes it hurt nothing less.  A heart is layed down, and the power of love is present for the speaker...but not for the lover.     I\'ll close my eyes, then I won\'t see The love you don\'t feel when you\'re holding me Morning will come and I\'ll do what\'s right Just give me till then to give up this fight And I will give up this fight    PHEW.  Hiding from the situation makes it better only for the time being, only for the last time.  One last taste of something so powerful, something that can\'t be acquired, and something that is going to be given up.    The chorus is repeated, and here I go crying again.  I don\'t  believe that it takes the experience of not being loved in return to appreciate this song.  I\'ve listened to tons of sappy, sad love songs in my day, but this one takes the cake.  Keep in mind, country music is not in my regular play lists, so I could have lots of disagreement based on  the songs of broken down eighteen wheelers and lost dogs, but it is an interesting idea, none the less, to attempt to place one song as the saddest of all.  Did VH1 do this on one of those countdown shows?  Possibly, and I also understand that every song has its own way of cutting people in their own particular ways, but if I have to give an award...Bonnie, you\'ve won it, cuz that song brings me to tears every SINGLE time I hear it.  Ain\'t music wonderful??  So as I finish this posting, I realize that in the future I might change my mind.  I will mark this time period by expressing that at this time (and for as long as I\'ve known it) that this is the saddest song.  More on other \"...EST\" songs later......   '),(17176,2543400,'22,December,2003','2003-12-22 00:00:00','   Well this is my new blog since I took down herzblut. I may go back once I get a new hard drive. It\'s just a matter of time for me to wait to get one. I\'ll buy one, maybe later! As of now I have no real use! I just need an old hard drive or something. I can see if there is a spare 2 gb hard drive anywhere. However no more hosting, since I\'m gonna start using OpenBSD....  Well more later...  '),(17177,2543400,'31,May,2004','2004-05-31 00:00:00','   Warping away into nothingness Watching the rose go through it\'s time Waiting for my rose to find it\'s roots  Life is but a rose So beautiful, yet so hard to grasp It has many faces all mine are dreadful  It is not my time When will it be my time? Waiting for the rose to change Waiting for the beauty to show  All I see is the thorns pierce through me the touch of pain as it destroys me slowly.  As revelations come and go lifes true eyes show. Life is not yours to have  '),(17178,2543400,'30,May,2004','2004-05-30 00:00:00','   urlLink    Carlo the senior class Valedictorian. Hail to his brilliant smartness :p&nbsp; urlLink  '),(17179,2543400,'30,May,2004','2004-05-30 00:00:00','   Man everyone now a days are depressed. It\'s like a friggin trend. It doesn\'t do shit, it just makes you feel worst. But depression has it\'s benefits I guess.  1. When I\'m depressed I think a lot. This is a mixed blessing, as it may lead to good thoughts or just thoughts on why I shouldn\'t kill myself. It\'s just like that. When I\'m depressed I seclude myself. Doesn\'t matter what anyone else says, only thing that matters is what I\'m thinking. 2. It brings you close to people. No one likes being alone, unless you totally antisocial and wanted everyone dead, but that\'s a different story. You find people to be there that you thought didn\'t care.  3. It makes you observe your view of everything. For example, I came to the conclusion that there is no God, for if there was a God that only wanted good, then he would have not created me. Therefore, God is not an option for me, but you do become more spiritual and see more to this lifetime then afterlife. This may differ you may be closer to God, but for me it was further. 4. It made me more compassionate toward animals. Sure not all of them are great, but no one/thing\'s life should be taken from them/it. Everything has a natural right to live, and it is only up to them to make a decision if they want to kill themselves or not. No i\'m not a vegetarian, but when you do eat meat, don\'t waste it, after all an animal did give it\'s life to feed you. Enough of my antiwasting ramp. 5. Some of the best works come out during depression, so write a lot. You\'ll get some good stuff out. Or I did in my case. Just write a lot and edit it later. 6. This is not really something you get out of depression but something to help it. Especially if your depressed about a girl, just watch a comedy, like American Pie. (Just rerun the Shannon Elizabeth sequence a couple times :p) Just watch a lot of comedies, it\'ll help you out of it. Or it did for me anyways. 7. Pain does seem real at this point, more real then anything else, but don\'t kill yourself. Life the first Noble Truth of Buddhism, Life is Suffering. It\'s ok to punch walls and do damage to yourself, that\'s nonlethal, if your totally depressed and find none of my information useful, either get help, or go ahead and kill yourself (Your choice there) 8. There is no #8  Well just some shit I\'ve noticed when I was out of it. Hopefully someone will make use of it. And Connie, this is me, not you so it may differ. :p  '),(17180,2543400,'30,May,2004','2004-05-30 00:00:00','   Girls are so hard to understand, man it\'s like we males need to like act like ones to find out how they act. For girls, its easy for them to be like males, since males make everything. LOL! not trying to be sexist but it\'s true. Girls need to tell us males what they like and stuff, we can\'t figure it out. It\'s like hell trying to know a girl, another hell trying to figure out what she likes, and a third hell just to talk about the stuff they like.  :-/ Man life has so much I don\'t understand and never will.  '),(17181,2543400,'30,May,2004','2004-05-30 00:00:00','   urlLink     urlLink How evil are  you ?  '),(17182,2543400,'30,May,2004','2004-05-30 00:00:00','   Damn Kid\'s Fest, waste of money. Five bucks for a small ass candle stick. There was nothing there! Gah! I need sunglasses, and I got some pics though, I guess something got accomplished.  '),(17183,2543400,'30,May,2004','2004-05-30 00:00:00','   urlLink    fucking evil mate&nbsp; urlLink  '),(17184,2543400,'30,May,2004','2004-05-30 00:00:00','   Man I just remembered something I wanted to post. Friday when Amanda and me went out to eat we got stuck in the theater parking space, or she did anyways. Come on now anyone could have gotten stuck, how can anyone park in that angle?  Anyways, after many tries to get out and sitting there for a few minutes, we got out using mad geometry skillz. :p Maybe I shouldn\'t have laughed, but it was just so cute to see her that way, I couldn\'t help it. :(  Anyways, we parked up higher. Damn I should of taken pictures, and oh yea... not even a scratch. :p  Man I feel bad, might be why she wasn\'t talking to me. *sigh* This stuff is so complex, why can\'t I simply be a rock. Rocks have such a better time don\'t have to worry about shit... or maybe they do...but other wise they don\'t have much to worry about...  Well i\'m becoming an addict and net gone morrow, gonna cry :(  '),(17185,2543400,'30,May,2004','2004-05-30 00:00:00','   Gah! So figgin bored.... should start on my portfolio, but won\'t :(  Man I\'m gonna miss my senior friends, all of em are leaving :( I\'ll be all alone, holy crap that\'s gonna suck ass.  Here are some good songs no particular order: Blink182 - What\'s my age again? Blink182 - Dumpweed Blink182 - All the small things Blink182 - I\'m lost without you Rammstein - Tier Rammstein - Engel Rammstein - Links 2 3 4 Korn - Counting on Me Korn - Here to Stay Radiohead - You Radiohead - Planet Telex In Flames - Swim In Flames - Whoracle Linkin Park - One Step Closer Linkin Park - Crawling Linkin Park - By_myslf (Remix of By Myself in the ablum Reanimation) Rage Against the Machine - Wake Up Melissa Auf Der Maur - Followed the Waves Green Day - Good Riddance (Time of Your Life) System of a Down - Aerials Megaherz - Herzblut Tomoyasu Hotei - \"Battle without Honor or Humanity\" Strawberry Alarm Clock - Incense and Peppermints   and dan dan daaaaa an Indian song: Dil Se Re from the movie Dil Se  Well there are of course more songs but these are some of my all time favorites. Good stuff  '),(17186,2543400,'30,May,2004','2004-05-30 00:00:00','   Hmm...well I thought I type some random stuff into my blog so here it is. Lets talk about Shrek 2, I loved it! So what if it\'s a kids movie, it was great. Man the whole Ponicco wearing thongs part, that was great. Laughed my ass off! And Puss in Boots, he\'s so cute,  man I feel like a lil girl right now 8-), but it was funny.  It was a great sequel better then the Matrix sequels that made no sense. Just a lot of fucked up shit. But I recomend everyone to see this movie, so what you waiting for...go see it, now! I order you! :)  '),(17187,2543400,'30,May,2004','2004-05-30 00:00:00','   urlLink    I like myself in comic form, leave me alone!&nbsp; urlLink  '),(17188,2543400,'30,May,2004','2004-05-30 00:00:00','   Well this will probably be my last blog from here for a tad while. Dad is canceling net morrow. I\'ll add pictures and stuff. Well yeah....  '),(17189,2543400,'30,May,2004','2004-05-30 00:00:00','   Well five more days till I go to India, man I\'m gonna be bored there. This will likely be my last post as my dad\'s gonna cancel my net morrow. Good riddence, Comcast is beginning to get on my nerves not letting me have multiple computers on my network. The bastards! Any yes I did try everything, even changed my router, to no avail.  I went to a \"party\" last night with parents, man I was bored as fuck, all the dad\'s looked the friggin same, come on can\'t you come up with a new fashion trend? But my dad was not part of that so I\'m happy, were always out of the norm.  My cuz\'s dad was like we all need to move to India, hell no! I\'m not moving there now, after all that I\'ve accomplished. Screw that! Life\'s a bore!   Well this post is out of the norm as well since I usually write when I\'m fucking depressed! I should eat and do my portfolio.....ummm....shit!  Anyways will write later and add more pictures, fuck yea! Man I fucking curse a fucking lot. :p It just lets out the fucking pressure. LOL  '),(17190,2543400,'30,May,2004','2004-05-30 00:00:00','   urlLink    My ugly, but cute sister :p&nbsp; urlLink  '),(17191,2543400,'29,May,2004','2004-05-29 00:00:00','   Hmm....blogger rocks :) All I have to say! One more week of tormenting week of internet.  Well my new digital camera rocks :) Can take video Yay! :)  Anyways, a lot of stuff to finish this week and finals on Tuesday. Man I\'m gonna be bored this summer lot of hw to finish. What a bitch!   Need to take more pictures of friends and stuff this week. Gonna be fun, party on Thursday hopefully :)  Food on Weds :p  Man almost a senior, I can\'t wait....  '),(17192,2543400,'29,May,2004','2004-05-29 00:00:00','   urlLink    Outside my home to the right.&nbsp; urlLink  '),(17193,2543400,'29,May,2004','2004-05-29 00:00:00','   urlLink    Was bored give me a break, lol man i need a shave...&nbsp; urlLink  '),(17194,2543400,'29,May,2004','2004-05-29 00:00:00','   urlLink    amanda in english, and cecilia in background&nbsp; urlLink  '),(17195,2543400,'29,May,2004','2004-05-29 00:00:00','   urlLink    If you ever wondered what Thuyet looks like drunk, here it is. :p It\'s not water he\'s holding, it\'s all vodka. ;) jk&nbsp; urlLink  '),(17196,2543400,'29,May,2004','2004-05-29 00:00:00','   urlLink    Amanda, wtf are you doing? Oh yea, feeling your heart pulse, lol, better run up and down those stairs&nbsp; urlLink  '),(17197,2543400,'29,May,2004','2004-05-29 00:00:00','   urlLink    Math class, my head magically appearing :p&nbsp; urlLink  '),(17198,2543400,'29,May,2004','2004-05-29 00:00:00','   urlLink    Me looking evil, fuck I hate how I look, need to gain weight&nbsp; urlLink  '),(17199,2543400,'29,May,2004','2004-05-29 00:00:00','   urlLink    Amanda and me (man i look gay! or is it the picture? nah! it\'s me!)&nbsp; urlLink  '),(17200,2543400,'29,May,2004','2004-05-29 00:00:00','   urlLink    Rock on, the coolest teacher ever, looking evil :p my hero ;)&nbsp; urlLink  '),(17201,2543400,'29,May,2004','2004-05-29 00:00:00','   urlLink    The Physics Male Threeo, ok 2o, since I don\'t do shit in that class :p anyways, alex lookin gayer then usual jk, and jason what the hell, that doesn\'t work dude! and omar looking all depressed in the background.&nbsp; urlLink  '),(17202,2543400,'29,May,2004','2004-05-29 00:00:00','   urlLink    The Crew, looking smart, or confused....or whatever....leave me alone, i\'m not smart :(&nbsp; urlLink  '),(17203,2543400,'29,May,2004','2004-05-29 00:00:00','   urlLink    Connie and Me, good friend&nbsp; urlLink  '),(17204,2543400,'29,May,2004','2004-05-29 00:00:00','   Well new blog as this is what i\'m gonna use to rss feed my site. And plus it\'s better then xanga. Well that\'s all :p  '),(17205,2543400,'29,May,2004','2004-05-29 00:00:00','   Life is really starting to get to me, it\'s fucking hell. Every small thing doesn\'t make sense. How can i conseive of reality, everything looks so true, so real, except there is something missing.  There is something missing from me. Something trying to get in, something I always seem to block.  My mom is constantly pissed off at me, and most people are as they have someone to turn to, and I have no one.  I am alone in my mind, friends will be there but it doesn;t matter, i\'m alone in my mind, alone in the darkness, alone with the anger, alone with the impulse that drives me ever closer to self-destruction.  Pain doesn\'t seem to matter anymore to me. It;s something more real then anything else. I can count on pain to be there, when I can\'t count on others.  I no longer or ever care that I go to college or die before I go, death is more real to me, simply stop exisiting. There is no place to go just cease to be.  I have no raison d\'etre and it will be that way for the rest of my life. I see people all around me enjoying being, while the only thing I see of myself, is craving to end this enjoyment. I want to bring evil to the world, want to make everyone suffer the way I am suffering.  FUCK YOU ALL! FUCK YOU ALL TO FUCKING HELL, FUCKING ASSHOLES!  '),(17206,2543400,'05,June,2004','2004-06-05 00:00:00','   Sorry about the last post not being complete. I was in Hong Kong when I wrote it as I was in transit to Singapore.  Anyways I\'m in singapore right now and they have better keyboards at their free internet terminals then did Hong kong.  Anyways, like I was saying man there are a shit load of indians here it;s friggin annoying. They\'re everywhere! It\'s not even one or two, more like 10-20! I hate my race, damn bastards. Anyways I\'m at a free net terminal in terminal 1. LOL! like that makes a difference. Anyways the plane ride was not as crampt as i thought it would. It was ok i guess.  I need to buy some sunglasses by the time i leave. Need em for the heat in india.  It\'s so nice to have net even for few minutes. I tried to connect to the wifi here but found out it was for pay. GAY!  Anyways I\'m standing here writing this. I haven\'t gotten into a fight with parents yet and they are letting me go on my own tours. Man i guess i\'m grown in their eyes. Anyawys, i got pictures and i\'ll try to upload them as soon as possible. I\'ve been doing nothing more then shooting pics. Some are nice others are crappy oh well. Sinapore airport is nice has many of these unusual osaises, will put pics later.  I\'m already missing my friends right now. They seems so far away! Sucks! But I gotta get gifts for them all.  Well that\'s all for now, i\'ll come and write more as stuff comes to my head. i got a long wait ahead so might as well. more later  '),(17207,2543400,'03,June,2004','2004-06-03 00:00:00','   Well i\'m writing his with the view of Hong Kong behind me t a terminal. I\'m sorry if there are a lot of errors in this post as this keyboard fucking sucks. That plane ride was figgin long and with so many indians, man it was like all of india was on teh god damn plane.  Anyways, we are currently in Hong Kong waiting as our plane gets ready to go to Singapore in like an hour.  '),(17208,2543400,'03,June,2004','2004-06-03 00:00:00','   Man I\'ll be thinking a lot over this summer so expect a lot of philosophy related posts. Just an FYI. I\'ll try to edit these as much as possible as I want to put them into one, and maybe have like a small manifesto of some sort.  Anyways a list of somethings to do over at India and after I get back: Take lots of pictures Meditate a lot Gain weight Try to go to gym Practice driving Climb up to tariputi English AP Hw Biology AP Hw Develop my website Start writing prompts for Berkeley Sleep a lot Read a lot I\'m running out of stuff to do....add stuff later lol.  '),(17209,2543400,'02,June,2004','2004-06-02 00:00:00','   I\'m already missing everyone, it\'s starting to hurt. It\'s like i\'m all alone again. This bites! I feel like a whole part of my life is missing, gone from view, it\'s over. I know I\'ll see most of my junior friends again next year, but it\'s just that the end is so near, it\'s like everyone is gonna go their seperate ways and I\'ll be there alone in a whole world. It is what is scaring the shit out of me!  It\'s just being alone!  And next year, everyone will be so busy, different classes and shit, not being able to do a lot of things. Man it\'s killing me! Fuck, and I don\'t want to go to India now, last 2 summers were so fulfilling meeting so many new people, it was great.  Basically I hate saying goodbye. I don\'t like saying it, and it is just that it is so final! I didn\'t even feel like then when I left El Dorado behind. It\'s just I want to accomplish so much and it feels so far and distant now. I\'m just going to miss everyone!   All of a sudden I feel so alone....I hate ends...  '),(17210,2543400,'01,June,2004','2004-06-01 00:00:00','   urlLink    Prom picture at home, look gay, as usual, whatever leave me the fuck alone....&nbsp; urlLink  '),(17211,2543400,'01,June,2004','2004-06-01 00:00:00','   Well friends graduated! :( It\'s kind of sad to see them go! Sure I may see them in the future but won\'t be the same, not all of em together. :( I\'m so friggin sad about that. Ah well, friends are friends for life.  Noor, Cecilia, Stephanie, Thuyet, Jason, Fabian, and me went to see Shrek 2 and went out to eat at I love Teriyaki, forgot the damn pictures, and hella took out the camera too. I\'m a fucking idiot! It was fun though. The waitress asked where Amanda was, lol, we do go there often though. Laughed seeing Shrek again, Ponicco rocks :p \"I\'m a real boy!\" And Puss in Boots, man that was so cute, almost wanted to buy a cat.  Took my Physics and Math finals today, I have a feeling I did not do so well. There is no point in me going to class morrow. Nothing to do there!  Man I\'m gonna miss everyone! :( And well this means I\'m a senior now, lol, or almost.  Well all I want to write for now. Maybe more later.  '),(17212,2543400,'01,June,2004','2004-06-01 00:00:00','   urlLink    Is a good picture, the detail, and stuff, my dad. Bah! Whatever....&nbsp; urlLink  '),(17213,2543400,'01,June,2004','2004-06-01 00:00:00','   urlLink Releases: Herzwerk II - Megaherz.com   Best song ever or one I like anyways. Megaherz - Herzblut  '),(17214,2543400,'01,June,2004','2004-06-01 00:00:00','   Well I did it I stayed up for about 28 hours, not bad. Next time when I don\'t have finals to worry about I\'m gonna try to stay awake for 2 days, possibly during Relay For Life.  Man stupid ass finals. I made hella stupid mistakes in physics, which is gay, and finally , I did ok in math. Stupid parabola\'s and shitz.  Seems Amanda\'s mom gave her a scare last night, just as my dad gave me a scare, opened the door real fast to my room and made me have an adreniline rush, I was more awake then shit. Damn parents, don\'t understand, we kids don\'t do stuff in advance. :p She also stayed awake the entire night, but for work purposes. I on the otherhand being a freak of nature stayed up, doing nothing, not even really studying. But got some good pics though, guess the night was good for something.  Well times a ticking, running out, just a few more days and I\'m out of here. God save me, I don\'t want to go. :( Want to have a free summer, ah well, I guess I\'ll take a lot of pictures and meditate a lot.  Well I gotta take a dump :p well not really, just wanted to end this post funny. LOL! I\'m a freak!  '),(17215,2543400,'01,June,2004','2004-06-01 00:00:00','   Well it seems this is gonna be the last blog entry I\'m gonna post from home...that is all, LOL, and w00t, no sleep for 24 hrs.  '),(17216,2543400,'01,June,2004','2004-06-01 00:00:00','   urlLink    Good picture as the sun rises.&nbsp; urlLink  '),(17217,2543400,'01,June,2004','2004-06-01 00:00:00','   urlLink    I love the color of the sky, fo no other good, this picture has that.&nbsp; urlLink  '),(17218,2543400,'01,June,2004','2004-06-01 00:00:00','   urlLink    Another fav, gives a blacked out city look. Doom is coming, doom, this is the city of DOOM!!!! ;)&nbsp; urlLink  '),(17219,2543400,'01,June,2004','2004-06-01 00:00:00','   urlLink    Another pic, but there\'s more light, therefore more evil lookin :p&nbsp; urlLink  '),(17220,2543400,'01,June,2004','2004-06-01 00:00:00','   urlLink    Oh la la, such a pretty ass pic, the lines are so nice.&nbsp; urlLink  '),(17221,2543400,'01,June,2004','2004-06-01 00:00:00','   urlLink    Nice blurry morning shot.&nbsp; urlLink  '),(17222,2543400,'01,June,2004','2004-06-01 00:00:00','   Well that was gay you can\'t see the sunrise from where I\'m at. All the trees block it. Well maybe in india will sit on grandpa\'s rooftop and take pictures.  '),(17223,2543400,'01,June,2004','2004-06-01 00:00:00','   Woot! Five o\'clock and still no sleep. 46 mins till sunrise. Gonna go take some pics... Will post from somewhere...  '),(17224,2543400,'01,June,2004','2004-06-01 00:00:00','   Not gonna sleep tonight, I know i\'m friggin crazy, but hey whatever. I might finally get those sunrise shots I always wanted, but have been too lazy to get up for.  '),(17225,2543400,'28,July,2004','2004-07-28 00:00:00','   Well I got my net back and I promised myself I\'d be online sparingly. This means no more being on for hours on end. I\'m gonna set myself a limit. And since i found out how to manipulate myself so i can change I am gonna do this.  Anways, SBC Yahoo is a bitch for multiple computers. Can\'t have internet hosting anymore. What a bitch! But I\'m gonna install OpenBSD on my box to store my code and private stuff. you know just for safe keeps as a server. Was thinking about FreeBSD 4.10, but then thought it\'d be better if I used a secure OS just for the hell of having a secure system.  '),(17226,2543400,'15,July,2004','2004-07-15 00:00:00','   Whoo! Almost back in the US of A! Well other then the possibility of my and a bunch of other people\'s death, it wasn\'t bad i guess. I traveled a lot, saw a lot of my cousins and family. What not! Got a bunch of pics, yay! And i guess that\'s pretty much it. But I had fun and got to miss school, a big plus! &nbsp; Anyways, i\'ll post the pics as soon as i\'m back in the country and have a decent net connection. i better be off now, as this net is getting slower by the second and people are wanting of a net connection. i\'ll writer asap. Ohh... I can writer without looking at the keyboard or the screen. :p This is fun! Anyayws better be off.  '),(17227,2543400,'08,July,2004','2004-07-08 00:00:00','   well i haven\'t been writing in here fore a while so i thought that i\'d do a quick post. nothing much just some measily little post about how horrible it is here. Man the worst thing about india is not even the diarea, it\'s about having to go take a crap. it\'s horrible! man americans are so advanced they have toilet paper,  and a nice toilet, not a bloody hole in the ground.  '),(17228,2543400,'09,August,2004','2004-08-09 00:00:00','   Another kick ass song :) Ill Nino - Cleansing  From guilt to dust You think that I don\'t want to hate you I know that I never did like you This is not another passing phrase But I can see your face But when shit will finally break me Persistent in trying to get me El fuego que quema adentro Gets to my head  Stay away I never want to see your face Forget my name There\'s nothing you can do or say Stay away I never want to see your face Forget my name There\'s nothing you can do or say  You think that you\'re so much like me But you will not ever remind me Of everything that I cannot embrace But I can see your face I\'m trying not to remember I\'m trying not to point the finger El fuego que quemo adentro Gets to my head  Stay away I never want to see your face Forget my name There\'s nothing you can do or say Stay away I never want to see your face Forget my name There\'s nothing you can do or say  I think I kind of hate you I never really liked you I think I kind of hate you I never really liked you This is not a passing phrase I think I kind of hate you I never really liked you I think I kind of hate you I never really liked you This is not a passing phrase  Stay away I never want to see your face Forget my name There\'s nothing you can do or say Stay away I never want to see your face Forget my name There\'s nothing you can do or say  '),(17229,2543400,'06,August,2004','2004-08-06 00:00:00','   Ok I read in this book if you want to improve something you have to set goals and keep thinking of them and your mind will formulate for you that which you desire.  So here are my long term and short term goals, enjoy being bored :-)  Long Term: Gain Weight, work out, straight A\'s in school, SAT I & II prep. Build vocabulary Practice Guitar everyday  Short Term: Learn Calculus, German, Finish Summer HW by Aug 18, Gain weight  Well i\'m a boring guy, leave me the f alone ;) And here are the lyrics to Ill Nino - This Time\'s For Real. Kick ass song :)    Why do you tell me all these lies?  I just want to live my life  I don\'t want to leave my dreams behind  Tell me  Why am I only getting older?  My patience getting shorter  I\'m running out of time  And I hope that I get what I wanted   This time\'s for real  Is there anything I wanted?  Is there anything I needed?  Anything at all?   This time\'s for real  Is there anything I wanted?  Is there anything I needed?  Anything at all?   Why do you make me feel denied?  Always pushing me aside  You do not decide what\'s wrong or right  Tell me  Why am I feeling so much pressure?  Am I searching for the answer?  I\'m running out of time  And I hope that I get what I wanted   This time\'s for real  Is there anything I wanted?  Is there anything I needed?  Anything at all?   This time\'s for real  Is there anything I wanted?  Is there anything I needed?  Anything at all?   This time\'s for real  Is there anything I wanted?  Is there anything I needed?  Anything at all?   This time\'s for real  Is there anything I wanted?  Is there anything I needed?  Anything at all?   Is there anything I wanted?  Is there anything I needed?  Anything at all?  '),(17230,2543400,'06,August,2004','2004-08-06 00:00:00','   Ill Nino -This Time\'s For Real kick ass song. :)  Well I talked to Andrew today, friggin didn\'t regonize his voice. Fuck I guess he\'s friggin different and I\'m still the same old shitty me. I so need to start truely gaining weight and going to the gym or something. Damn! ;-  Well I keep doing push ups and sit ups at random times through out the day. And specifically before I sleep. For example, last night I did 55 push ups and 105 sit ups before sleep. Not much of an accomplishment for others but it\'s a damn big acheivement for me. :) After all i\'m just a fat person who has flabby arms, ok maybe not fat :-) but i\'m so friggin out of shape i\'d probaly pass out if I ran 10 yards. :-)  Well gonna meet Andrew on Sat or Sun. For the rest of the week my goal is to gain at least 2 pounds and to keep doing push ups and sit ups, along with meditation. :) Shit my mind is so full of thoughts if I close my eyes all I see is an endless film running in my mind without a stop. Try concentrating on your breath when that\'s going through your mind. It makes you want pop corn. Haha!  Mmm.....popcorn.....  '),(17231,2543400,'06,August,2004','2004-08-06 00:00:00','   Well I\'m eating a lot. I guess that\'s a good thing, spending less time on the net. I still need to gain weight damn it.  Need to start and finish all 5 essays for Bio morrow. Screw everything else. It has to be finished. As well as finish reading this book which is also due morrow.  Gah! Well gonna try to get up early morrow.  '),(17232,2543400,'06,August,2004','2004-08-06 00:00:00','   Today started out bad, but it has gotten better. Day started at around 8 when I set the alarm to. When I didn\'t get up my mom nagged at me until I finally gave in and got up at around 9:30.  I started getting ready and left with the gifts at around 10:55. When I got there Jason, Douglas and Mark were already there. Finally a few minutes after me Alex showed up. And a short while after Sam, Connie and Mylien. A few minutes after that Lauren showed up.  We went to eat. And after that Lauren and Sam left, as well as Jason and Doug. The rest of us went to see a Harold & Kumar go to white castle. It was a good movie funny yet stupid. It\'s a comedy why would one expect. After that we went to Fry\'s and did a couple of circles there. For an hour or so. After that we went to Petco. Then Half Priced Books. I was bored out of my mind and so took my leave then.  Then in my military fashion I left to home.  It\'s odd how aware one is when their back is straight. It\'s almost like I was a spy looking around. When I got home I played my guitar for a few minutes around 20 :p and then went to the bank to deposit some money and do the laundry.  Then I came here. Sat around for a few minutes and then picked up the laundry and then came here to write this.  Love is hard to do, I have to completely change myself. This life of me is just too mundane and crappy. Need to do something new and unusual. Just out of the ordinary.  Crossed legged doesn;t hurt no more. Just feels wierd.  '),(17233,2543400,'05,August,2004','2004-08-05 00:00:00','   w00t! New cell phone, it rocks :)  anything else? not really.... have to start writing some stuff reading some stuff and playing my guitar.  '),(17234,2543400,'03,August,2004','2004-08-03 00:00:00','   urlLink    Some Indie Rock and Emo and crap. Good stuff.... Some more ideas to steal. Lauren showed me this site.  '),(17235,2543400,'01,August,2004','2004-08-01 00:00:00','   I feel this is going to be one long ass post so bare with me....  There is something inside me that is thriving wanting to get out, something. A feeling! It has to be something good a new energy it\'s making me more useful.  I saw the Bourne Supremacy and Bourne Identity on video. Good movies now I have to read the books. These are the books I want to read in the following order: Bourne Identity Bourne Supremacy Bourne Ultimatum Crying of Lot 49 (Because of the reference Justin Frankel used for WASTE)  There are many things I need to do starting this month. 1. Start my weight gain plan. 2. Start working out 3. Start work on summer hw 4. Start work on SAT 5. Start work on crypto program 6. Start work on some research papers 7. Start work on learning effective communication (i.e. Talking clear) 8. Start learning French again from the beginning, need to be fluent 9. Start work on learning German 10. Start learning Calculus and other high end math. (Why? Well because) 11. Start working on reading a lot 12. Stop going on computer for more then 3 hours (even that is too much) 13. Start meditation again 14. Start learning about business, investing and etc. (Why? Need to make $$$) 15. Start being more effective at what I do 16. Start being more focused and concentrated 17. Start learning matial arts (Akido, Kung Fu, etc.) 18. Start learning cryptography 19. Start writing a lot 20. Start learning guitar and bass once I get it  This is all in no particular order.  In the future I would like to learn these languages fluently. Some may be ones I know somewhat:  French (fluency) German Spanish Russian Telugu Hindi Japanese Chinese (Mandarin)  Why do I want to do this stuff? Because if 96% of my mind is unused then I think that is a lot of unused potential needing to be exploited. Therefore I want to learn the damn things. And plus it\'ll help me if I ever travel to Europe which I hope to do.  '),(17236,174357,'28,June,2004','2004-06-28 00:00:00','    You Are Right Brained In Love  Bit of a drama queen  Peacemaker, first to end a fight  Good at thinking up creative dates  Tend to fall in love and get hurt easily  Going with your gut instead of your head  Emphathetic and caring, sometimes to a fault  Good at recognizing patterns in relationships  Been in love many times, perhaps too many to count  Wildly passionate and intense when falling in love  Spontaneous with relationships, going with the flow  Overly visual - can play back past dates like movies in your mind  Roses, love poems, and stuffed animals are a good start to winning your heart      urlLink  Are You Right Brained or Left Brained in Love? Take This Quiz :-)        urlLink Find the Love of Your Life  (and More Love Quizzes) at Your New Romance.'),(17237,174357,'24,June,2004','2004-06-24 00:00:00','    Im so tired. I think today is just one of those days. I\'ve been dragging all day it seems like. I\'m kind of bored. I should be at the Blue Note right now, but I really didn\'t feel like going tonight. I really dont have any money either but that\'s okay. There is a free concert saturday. My friend Ashley wants me to go with her. I\'m thinking it might be cool. Our friends Denial are opening up for Lit. I think Lit is pretty awesome anyway, so it\'s just an added bonus that our friends are the opening act. It\'s kind of exciting. I\'m really really happy for them. Blah I need to take a shower. I just feel like one. I\'ll feel refreshed or something. '),(17238,174357,'28,July,2004','2004-07-28 00:00:00','    So I started this new job this week and it seems pretty cool. It\'s a lot of training though. But it\'s okay. Hopefully i\'ll be making a lot of money soon. I have cramps so it sucks right now, because when I have cramps I cant sleep, and I have to get up early in the morning for training. So I probably wont be concentrating too much tomorrow. Well i\'m sure nobody really cared to hear that, but hey this is my blog lol so oh well. It\'s not like anybody reads it anyway.   Ugh there is this really disgusting guy that keeps telling everybody that I went out with him and i\'m like wtf? And my boyfriend is like wtf? Cause this guy is absolutely repulsive and he\'s just ick, and never in a million years would I go out with him. I\'m like....ok whatever if anybody is dumb enough to believe that I would go out with him theny so be it, cause they must not really know me as a person. Cause if they did they would know that I dont cheat on my boyfriends and that this guy is totally not my style.&nbsp; So that\'s that.   I need to study this packet that I got from my job. I have a test tomorrow. Since i\'m gonna be up anyway I might as well study right? But I just dont feel like it. I feel more like laying around on the couch curled up in a ball, cause that\'s just what I do when&nbsp;I have cramps. You know, maybe if I got up off my fat butt and exercised I wouldn\'t have cramps this bad. I\'ll try it, but it\'s highly unlikely that i\'ll really exercise for more than 30 seconds. Cause i\'ll just be like...owwww...pain. And then i\'ll stop. Cause....well...i\'m a loser in a nut shell. AAAAHHHHHHHH I have to e-mail mary. '),(17239,3524887,'26,June,2004','2004-06-26 00:00:00','   Holland vs Sweden  26/06/04 KO 0245 --->Hat-Trick Challange! Sweden are trouble with some injury to first team players. Anyway the Sweden\'s bench had lot of talent and quailty lads which will be ready to field up any hole in the squad. Holland on the other side was in better situation when going for selection of players, everyone want to play! Holland is on fire again and prove that they can be so dangerous when they can play together as a team and playing to the coach\'s tactic. Sweden is unbeaten as they are lucky not to lost against the Italy and Denmark which they managed to grap the equaliser in the last few minutes. Sweden and Holland have lot of good offensive players and they perform better when playing attacking football. Larsson and Van Nistelrooy should get their on the scoresheet. Should expect more goals in this game. As it always say, never change a winning team (Sweden\'s injury problem and Holland\'s stars selection problem). Team performance should depend on selection of players for tonight game.   PR: Expect Holland to win.  Result:  Holland vs Sweden  '),(17240,3524887,'25,June,2004','2004-06-25 00:00:00','   France vs Greece  26/06/04 KO 0245 --->Shocker! France finally played a game that had pleased everyone and advance to this quarter-final match with the surprising Greece. France was still not impressive during the the win over Swiss but they had improved and getting much better in shape. Greece play a tough group matches and have beat the host Portugal and draw with Giant Spain. They lost to Russia 1-2 but fight back to score a goal for advancement. Seem like Greece play better and have more luck to play against Giant team. France will be much happier to meet the Greece but a little mistake from the defence could cost them the game as their defence seem to be unstable. Greece need to perform another counter-attacking tactic in order to shock the holder which I believe they are capable of.  PR: Receiving +ball will be recommended.  Result:  France 0-1 Greece  '),(17241,3524887,'25,June,2004','2004-06-25 00:00:00','   Portugal vs England  25/06/04 KO 0245 --->Future stars shine! Both Portugal and England managed to win two matches in a row after losing their opening match. With so much young and exprience talents in both squad, both look equally strong and tough. Portugal may have the advantage of playing host with so many Portugese cheering than the English but most Portugal\'s stars are still not in very good form yet. England\'s form is peaking after matches and matches and seem like they are in very good state now. Once again Rooney will need his good form to find the winner for England and if he does, England wil stand a very good chance for the semi-final. Portugal need to play at the standard they play against Spain in the last game and they need to stop wasting chances which they are during the Group stage. If both countries are willing to start with attcking tactics then it will be a match not to be missed!  PR: Winner to decide in extra time.   Result:  Portugal 2-2AFT PK(8-7) England  '),(17242,3524887,'24,June,2004','2004-06-24 00:00:00','   Germany vs Czech Republic  24/06/04 KO 0245 --->So tough German! After playing so hard and without luck in the last game against the hardworking Latvia, German are in a difficult position. They need to avoid defeat to have the chance to qualify for next round. The German got to play hard in their defence in order not to conceed to the already qualify Czech. Czech might be fielding only some first team players in this game but definitely their bench players are still very good players that have the required standard in the squad. Without their first team defenders playing at the back, Czech may conceed but their attacking pressure must not be neglected. Germany need to win to qualify and all will depend on how Voller play his tactic tonight.    PR: Germany to draw. Receiving +ball will be recommended  Result:  Germany 1-2 Czech Republic    Holland vs Latvia  24/06/04 KO 0245 --->Winner takes all! Both team need to win and hope Germany\'s failure to beat Czech for a place in the last eight. With so many world class players in the Dutch squad but they still have not played together and perform as a world class team. Holland are capable to beat Latvia in any meeting but not this time round. Holland are in lot of pressure at the moment and surely they cannot play as their usual standard. Latvia are in very good form with their shinning star striker Verpakovskis on fire, they might have a chance to make history. Very fast in counter-attcking and against the full attacking Holland tonight, their tactic will be more useful than any other game. Not to forget, if Holland draw with Latvia and Czech crash Germany, Holland might advance.  PR: Latvia to win. Receiving +ball will be recommended  Result:  Holland 3-0 Latvia  '),(17243,3524887,'22,June,2004','2004-06-22 00:00:00','   Denmark vs Sweden  23/06/04 KO 0245 --->Handshake!  PR: None....  Result:  Denmark vs Sweden    Italy vs Bulgaria  23/06/04 KO 0245 --->Late!  PR: None....  Result:  Italy 2-1 Bulgaria  '),(17244,3524887,'22,June,2004','2004-06-22 00:00:00','   Croatia vs England  22/06/04 KO 0245 --->Defending Fortress! Did England really play very well against the Swiss in the last game? I personally do not think so. Maybe only Rooney did play at his very best which lead England to win. Croatia played very well against the holder France but was unlucky not to keep their lead and will have to defeat England to qualify for the next round. Croatia will have the chance to win if they can keep pressing on England and shooting the ball on target. This requried fitness and team flair which Croatia do not really do possess. England to play defence and draw goaless? I hope they do not apply to this tactic as I do not trust James. Croatia will have to play more attacking tactic which it suit the England counter-attcking style. England can win if they can score and lead in the first half.   PR: England to win.  Result:  Croatia 2-4 England    Switzerland vs France  22/06/04 KO 0245 --->Undefeated must go on! France need to stay undefeated to qualify. They have stayed 20 international game undefeated and could this be the end of the run. Switzerland can play with everything they have now as they do not have any choice. Do it or die game, just like Croatia. Must operate in attacking mode until they scored. Still can\'t find the net in this tournament yet and I believe they will in this game. France are still not finding the form which lead them in world cup 98 and euro 2000. They are slow and lack of fitness but luck is on their side which help them lead the table. Unless they find their form back in this game if not, I believe Switzerland may destroy the France\'s run. I think a small wage on Switzerland to win the over-rated France can be a option.  PR: France to draw. Receiving +ball will be recommended.  Result:  Switzerland 1-3 France  '),(17245,3524887,'20,June,2004','2004-06-20 00:00:00','   Russia vs Greece  21/06/04 KO 0245 --->Russia roulette may row the last time. Russia was out of the tournament after losing to host nation Portugal. This time round pressure is on Greece. Russia may play more comfortable in this game with any pressure and thus display better football. Greece just need to avoid defeat and may qualify for the next round. So will Greece still choose to play defensive tactic and allow the Russian to bombard them? All knows that Greece perform better when they are the underdog and not the favorite, therefore there may have a twist in scenario this time round. Russia are surely not as weak as they seem too be if they really playing with ore confident and less pressure. Tonight Russia may break the duck and hit a few balls into the back of Greece\'s net. Greece may get the job done if they could find the net and lead by half-time.  PR: Russia to win. Receiving +ball will be recommended.  Result:  Russia 2-1 Greece    Spain vs Portugal  21/06/04 KO 0245 --->Portugal on the mission Finally the much awaiting game in Group A has arrived! Whoever win will qualify for the next round. Surely result from the other game between Russia and Greece will affect the outcome on this fixture. Hard to comment what will happen for the outcome now but both teams must be preparing some winning formula for sure. Portugal look sharp again when they finally won a game and put them a step into the final eight. Spain in the other has taken a step back as they failed to hold their nerves and let Greece stole the important equalising goal. Both team suffered in the hand of Greece. Portugal surely need the win more badly than Spain and if they are confident and with home crowds fully behind them, they may emerge as the winner. Spain will have better chance if Russia can do them a favour by beating the odds in which they still could afford to lose by close margin result. Time for Real Madrid\'s galatico Luis Figo and Raul to battle it out!  PR: Portugal to qualify. Letting -ball will be recommended.  Result:  Spain 0-1 Portugal  '),(17246,3524887,'19,June,2004','2004-06-19 00:00:00','   Latvia vs Germany  19/06/04 KO 2359 --->Germany march on! After a superb performace from Germany from the last game, they had proved that Germany is still a tough competitor in this tournament. Latvia was unlucky to throw away the only goal they scored in the first half and ended up with nothing after the final whistle. Latvia do not look too good in defence and against the tough German, they have to rely on their speedy counter attack. If Latvia can sneak a couple of goals, they might be the three pointer instead but I doubt they can do it this time. Germany is on fire now. They have been criticized for some poor display in friendies and seem hopeless for this tournament earlier on. The German will only be the one who know what they are made of and I believe they will rise to ocasion and display one of their best performace tonight game against the lowly rated Latvia.  PR: Germany to take lead by half-time. Letting -ball will be recommended.  Result:  Latvia 0-0 Germany    Holland vs Czech Republic  20/06/04 KO 0245 --->Battle up once more. Holland and Czech have met twice before this tournament and ended with a draw and a win for Czech. With dismay in the Holland\'s camp which have affected the standard of their football diaplay. Czech was lucky not lose and instead gain all three points from Latvia with their determination display. If Czech got the determination and the formula that they use to beat Holland, they can do it again. Don\'t forget Czech\'s Rene Bolf, Zdenek Grygera, Jan Koller, Pavel Nedved and Tomas Rosicky are suffering illness and injury. Responsiblity may rely on Vladimir Smicer. Holland have some injured players back in contention and selection for the first eleven may prove crucial for Holland\'s qualification chance. Holland\'s coach Dick Advocaat facing huge pressure now will have to re-shuffle their playing tactic to play better football in order not to suffer any lost of points again.  PR: Czech clinching the winner. Receiving +ball will be recommended  Result:  Holland 2-3 Czech Republic  '),(17247,3524887,'18,June,2004','2004-06-18 00:00:00','   Bulgaria vs Denmark  18/06/04 KO 2359 --->Rumble match Bulgaria can\'t afford to lose this encounter with Denmark. Both team need the 3 points to have hope of going into the next round. Bulgaria attack was a positive one but lost of confident and lack of mental strength in defence during the match with Sweden has make Bulgaria look really bad now. I believed those Bulgarian will understand the situation they are in now. As for the Danes, they miss the points against Italy which they should have won. They got to play with the standard or even better against Italy if they do want get into a must win situation for the last group game against Sweden. If Bulgaria do not tighten up their defence, Denmark should have not problem finding the net a few times. I believe the Bulgarian will make life difficult for the Danes surely.  PR: Denmark to get the winner. Letting -ball will be recommended.  Result:  Bulgaria 0-2 Denmark    Italy vs Sweden  19/06/04 KO 0245 --->Heated match! Italy had lost Toti and Del Piero\'s is having all the pressure now. It may be a good sign for Italy to play with only one playermarker rather than two. They got to knock in goals tonight if they do not wish to pack their bags and ready to leave Portugal. Basically Italy did nothing wrong in defence line expect they seem too be tired running under hot sun. Maybe playing at a later time might help them a bit. Sweden, the first team to score 5 goals in this tournament so far. Very good team work and confident whenever they have procession in opposite half. Larsson is still so sharp as ever and Ibrahimovic will have to perform at his peak in order to unlock the steel lock of Italy.  PR: Basically a typical draw game! Receiving +ball will be recommended  Result:  Italy 1-1 Sweden  '),(17248,3524887,'17,June,2004','2004-06-17 00:00:00','   England vs Switzerland  17/06/04 KO 2359 --->No more mercy! Switzerland was lucky to have the one point after they were playing with just 10 men against Croatia in the last match. Swiss seem to have improve in their overall play but still lack the lethal finishing. Swiss\'s defence may seem strong but they were merely lucky not to have concede any goal yet. If they want to win this game, they have to rely on their strikers as both had not scored for a long time. England was unlucky not to take any point from the showdown last game. Defensively they are up to task but their goalkeeper was really in a bad shape. Any more mistakes from James could end England chance for the next round. If England can perform the way they did in the first half against the mighty France, they could win this game. Owen and Rooney should be fully fit after being substituted in the last game.  PR: England to win by a goal.  Result:  England 3-0 Switzerland    Croatia vs France  18/06/04 KO 0245 --->Champion to next round? Defensive style of Croatia had failed to take the advantage of 10 men Swiss and being held back for scoreless draw. France did not perform at their world class standard but was lucky enough to top the table. The defending champion will go into the next round if they manage to beat the determined Croatia\'s defence line. Croatia may surprise France if they can score the opener and defend till 90 mins which they are capable of. Match will go either way if no one take the lead by half-time. There is news from Croatia camp that they want to rest some of their important players for the last game against England. Will that mean France will have a comfortable win or just some news to disturb their opponent mentality.  PR: Croatia! Receiving +ball will be recommended.  Result:  Croatia 2-2 France  '),(17249,3524887,'16,June,2004','2004-06-16 00:00:00','   Greece vs Spain  16/06/04 KO 2359 --->Revenge matchup Spain will have all the reason to win tonight game against the improving Greece. After denying Spain victories in both qualifying games, Greece will have all the confident they wanted for so long. The Greek has proved to the world that luck from their side have made them look one of the most likely underdog to going forward into the next round. As for Spain, let hope their poor history of displaying world class football in major tournament do not haunt them again. They can win this game easily if they can manage to score a couple of goals in the first 45 minutes. Will Greece perform a mixture of luck and determination to deny Spain confident? Early goals from Spain will do the job nice n clean. Don\'t forget that whoever win will book a place in next round for sure.  PR: Spain to lead at half-time. Letting -ball will be recommended.  Result:  Greece 1-1 Spain    Russia vs Portugal  17/06/04 KO 0245 --->Fire up Portuguese! What else can you say when these Russian have to face host nation after playmarker had been send home. Russia\'s coach has lost respect from his players and many still think they can\'t do any better without their best and experienced midfielder. If Portugal do not take up the 3 points, they will almost facing exit from the tournament. Host nation still have their home fans cheering for them and hoping Portugal will fire up at this moment. They do have a squad of talent and experience players all round who can lift their play to the peak. Last word! Portugal have to win of they want to get into the next round. Russian cannot rely on the defensive tactic against the Portugese anymore. Early goals from Portugal may help them to build more confident before going into second half.  PR: Portugal to lead at half-time. Letting -ball will be recommended.  Result:  Russia 0-2 Portugal  '),(17250,3524887,'15,June,2004','2004-06-15 00:00:00','   Czech Republic vs Latvia  15/06/04 KO 2359 --->Red Knight war cried! Czech are having pressure at the moment. They not only need to win this game but also to win in great style. A team that had peak since last year, good or bad? Well at least the mighty Czech had beaten big names in europe in the qualifier and some friendly. Strong from midfield to attack, Czech will attack and score goals rather than playing defensive no matter if they are leading or not. Latvia had stunned most soccer fans after taking up a slot in this tournament. They play defensive in most of their games and had most of the result they wanted. Counter attacking style of play is the only weapon Latvia has in their coach\'s tactics book. Fitness may be the factor to decide if Latvia can overcome the huge odds on them. Czech will have to score earlier goals to deny being upset in this game.    PR: Czech will gun down the underdog. Letting -ball will be recommended.  Result:  Czech Republic 2-1 Latvia    Holland vs Germany  16/06/04 KO 0245 --->World War 2  \"Group of Death\", exit of Hollamd or Germany to be decided in this game? With all the world class strikers all available for Holland at the moment, they have the craftier attacking style in europe. As for Germany, their thrifty defence will have a huge montain to climb to prevent any shot hitting onto the back of their net. Germany will still rely on their old but reliable defenders to accomplish the task. Holland have problem in their camp and team spirit seems very low as recent international friendlies has shown. Holland fail to score any goal and lost two games with a 1-0 results. Germany can\'t do any better as they flop in recent friendlies too. But will experience of the old German deny the total football of Dutch?  PR: Draw is more appropriate for this game. Receiving +ball will be recommended.    Result:  Holland 1-1 Germany  '),(17251,3524887,'14,June,2004','2004-06-14 00:00:00','   Italy vs Denmark  14/06/04 KO 2359 --->Defend rumble Italy and Denmark are well known for their defensive tactic. Denmark seem like will be missing a few important players tonight. Just hope they don\'t lose their original tactic if they want to deny the odds on them. Italy seem like perfect in this tournament. Don\'t forget what the Italian are capable of, wait and wait before scoring a goal. If both team know they could not afford to lose their opening games, outcome will be only one. Statemate!  PR: Draw is more appropriate for this game. Receiving +ball will be recommended.  Result:  Italy 0-0 Denmark    Sweden vs Bulgaria  15/06/04 KO 0245 --->Return of the legend What Sweden has been missing for is the key man in the attack. Henrik Larsson. What more do you want from Sweden when they have proved their strength in the backline and toughness midfield players. They can play and win games with only a goal in the last few months but all are without Larrson. Bulgaria in another point of view is a lucky team to have quailfy for this tournament. Lack of fitness in their attacking and defence could cause them the failure to overcome the Sweden penetration attack. If only Bulgaria could be as lucky as Greece to score an early goal, Sweden will celebrate when referee blow his final whistle.  PR: Back the Sweden in all aspect. Letting -ball will be recommended.  Result:  Sweden 5-0 Bulgaria  '),(17252,3524887,'13,June,2004','2004-06-13 00:00:00','   Croatia vs Switzerland  13/06/04 KO 2359 --->The battle of underdog Swiss has not perform well in their last few international friendly matches which lack the type of determination that Swiss bring themselves to quailfy for this tournament. Defence had not been a problem for the Swiss expect slow passing in midfield make the Swiss look weak in holding possesion. In other words, they seldom control the game and have to rely on the back four. Croatia seem to get on form at the right moment.Croatia in all look more complete and could really go far if fitness and injury do not hit them. They do have big names playing in major european team and thus make them look more favourite in this showdown. Croatia really can perform well in tournament as record shown. If Swiss still have not tighten up the midfield line by tonight, they will be the prey for the Croatia.   PR: Croatia to go with the odds. Letting -ball will be recommended.  Result:  Croatia 0-0 Switzerland    France vs England  14/06/04 KO 0245 --->Premiership Champion against England France\'s trio Veira, Pires and Henry have know England\'s players playing style by heart. This can help them for this game. In other words, France can be the one celebrating if luck do not help England a bit. France seldom lose their matches ever sice they are crowned as world champion but of course draw may not be avoid. Rooney is the key in this game. He can be the one who can win this opening game for England if he can perform at his peak. England backline is fragile which could cause them the game if France can find the net early in this game. Don\'t forget England always perform better when they are the underdog.  PR: Draw is more appropriate for this game. Receiving +ball will be recommended.  Result:  France 2-1 England  '),(17253,3524887,'12,June,2004','2004-06-12 00:00:00','   Portugal vs Greece  12/06/04 KO 2359  --->Home team against a fast improved team in last 6 months.  Greece has shown improvement in away matches since their last failure to go into WC2002. Pratically a proven giant killing team in their home of Greek. Advantage on playing in hot weather and always perform well against the odds in the qualifying round especially against the mighty Spainards. As for Portugal, home soil advantage can be the winning factor in this opening game. With Figo, Rui Costa, Jorge Costa and various stars playing in big european team, they will perform well if they can forget what had happened two years back in Korea. Young players from the European Cup winners FC Porto will have all the confidence to perform at the peak of their talent. Don\'t forget Solari still have the winning formula that he bring back the world cup for Brazil in the last world cup.   PR: Greece to hold Portugal for a goaless result or a narrow win for the home teams. Receiving +ball will be recommended.  Result:  Portugal 1-2 Greece    Spain vs Russia  13/06/04 KO 0245 --->Neutral ground match up Don\'t even think Spain will be playing at home standard in this game against the big-sized roulette as the Spainard will most likely lose all the high ball challenge in this game. Russian will overcome the odds if the central mastermind Mostovio and side wizard Karpin could deliver world class passes into the Spainard safeguard area. Spain, a born loser in major tournament, might break the curse again. They did show the world in last world cup but without experienced backline hold by Hierro(retired) and company, this could be another upset for the Spanish supporters if the new backline failed to clear the high balls bombarded by the Russian troop. Just a trick from Morientes can bring glory to Spain if he\'s not properly man-marked.  PR: A 50-50 game in my opinion and odds for Russia has hugh return. Receiving +ball will be recommended.  Result:  Spain 1-0 Russia  '),(17254,3524887,'01,June,2004','2004-06-01 00:00:00','   A legend will rise after history is dated......  '),(17255,3524887,'08,July,2004','2004-07-08 00:00:00','   Very busy recently.  Will start the new preview and prediction of Barclaycard Premiership 2004/5 and some matches around Europe on August.  '),(17256,4177364,'28,May,2004','2004-05-28 00:00:00','   Went to Rush Copley ER on Friday at 6pm. Was diagnossed with a Cervical Strain and was given off of work till 6/2. They prescribed Vicadin (505mg generic) has made no difference and back is still in pain, also have had only 3-4 hours of sleep per night. Due to the accident, I lost the sheduled OT (2 hours) Thursday and Friday. Also, Saturday 8 hours OT was available and 8 hours Double Time (plus holiday pay, 3x) on Monday. Due to accident, I canâ€™t do any of it.  '),(17257,4177364,'27,May,2004','2004-05-27 00:00:00','   Thursday (5/27) i was in a car accident.  A car turned infront of me, crossing my lane to enter a strip mall on North Aurora Rd about 100 yards east of the intersection at Eola. The traffic light was out and the traffic to turn left (southbound Eola) was backed up almost 200 yards. The North Aurora Rd. has 2 lanes heading in each direction and where this other car came from, was a left turn lane from the opposite direction. My (right) lane was clear to the stoplight. The speed limit there is 45, but i was already slowing down, between 35-40. The car that pulled out in front of me came from that left turn lane and with the cars in my left lane, i never saw her coming till she was almost 3/4 of the way into my lane. I hit the brakes and tried to turn to miss her, after hitting her I slid up into the grass. From looking at the skid marks, i was only on the brakes for about 15 feet before i hit her right front corner of her car with the driver side corner of mine. She admitted fault to me, the police officer and to the 2 witnesses there as well as the cop gave her a ticket for failure to yield while turning left. When i hit her, i tore off the entire front end of her car. She was driving a 2002 Lexus RX 300. She wasn\'t hurt; i did jerk hard and hit my head.  I Went to Enterprise to get a rental after the Jeep went to the body shop. State Farm said that even though it was her fault, they couldnâ€™t pay till they got a hold of her. Enterprise took $100 from my debit (actually only the autorization from the CC company for the $100 for deposit if needed). They first gave me a Chevy Cavalier. I tried driving it was very cramped. After taking care of business, I went back to Enterprise and they gave me the 2004 Taurus they just got back in.  '),(17258,4177364,'11,June,2004','2004-06-11 00:00:00','   6/1  called my regular doctor; he was booked so they got me in with Dr. Michelle Bardack. She wrote me off work, 6/1 â€“6/7 and prescriptions for Valium (5mg generic) and Ibuprofen (800mg). Also prescribed massage therapy.  6/2  State Farm finally called and said they got a hold of her and that everything would be 100% paid. My insurance office called the next day to say the same thing. Don at the body shop called and told me the damage was at $4500 and not sure if repairable (they recommened to the insurance to total it).  6/4  started therapy at Athletico twice a week. Also went back to doc, put me off till the 11th.  6/7  returned to Dr. Bardack. Wrote to return to work on 6/14 on FAX that Lesley sent.  6/8  the body shop called and said the repairs were being started.  6/10  Talked to Leslie (works contact from home office), agreed to go back on the 14th as original written by doc. Dr. Bardack did send the 3M medical info to Leslie but had a return to the dr on the 14th and work on the 15th. I told Leslie I feel good enough to get back on the 14th and she agreed it was ok since I have the info doc originally wrote.  '),(17259,4177364,'19,June,2004','2004-06-19 00:00:00','   6/14  tried calling Leslie in refrence to going back to work today, she is out of the office till July 6th. Called Eileen (facility nurse) at work and the only info that she received was the return to doc today and work tomorrow the 15th. Eileen also gave me phone # for John Blatner (the company psycholigist). Dr. Bardackâ€™s office called back said if I felt good enough, no need for apointment today and I could return on the 15th with no restrictions. I also asked the drâ€™s office to fax that info to Eileen. I called the body shop, they said they were waiting on parts, paint is done. Don hoped to send it out to get the allignment done on Friday. I havenâ€™t heard anything from the State Farm since 6/2.   6/15  returned to work for unrestricted 8 hours. Was in some pain but thought it was manageable. Did have 2 hours physical therapy before work, everything felt ok. While at work, Wes talked to be about a safety violation of mine on the 26th of May. I told him I didnâ€™t remember it as all the accident stuff was hard to get past, but I said that I wouldnâ€™t deny it when he said others saw me. He told me that the saftey team decided that he should just talk to me about it and make sure I knew it was wrong. I thanked him and apologized for my actions.  6/16  woke up in pain. Went to work, left after one hour, pain was increasing. Got appointment with Dr. Bardack at 4:45. She filled out med info sheet for 3M limiting me to 4 hours.  6/17  when I got to work, Eileen said she (as in 3M) needed in print from Doctorâ€™s office \"Only 4 hours per day\". The medical sheet to 3M is read as vague because it was marked 1-4 hourâ€™s sit and 1-4 stand; they read it as 8 hours. While at Athletico, agreed to 3 times a week for therapy.  6/18  Called Doctorâ€™s office and asked for something from the office, from the doctor to be faxed to Eileen. Dr. Bardack is off till Monday, they said they would ask Dr. Paras to do it. Saturday was scheduled; I lost another 8 OT.  '),(17260,4177364,'28,June,2004','2004-06-28 00:00:00','   6/21  had therapy in the morning, but once at work, I was hurting the first hour. I talked to Wes at 5:15 telling him I needed to leave due to the pain. He suggested that if I need the cold packs to work, maybe I should be off work.  with suggestions like that, the ass should become a doctor! Called the body shop, they believed they would have it ready today.  6/22   Gonna swing by body shop to check on truck. Back hurts, will freeze before work. Need to ask for some painkillers from doc on Wednesday. Got the Jeep back. Need to go back for a couple of problems.  6/23  had appointment with doc. No new drugs; she did send me to hospital for more x-rays and to take some labs at the office too. Bardack also wants me to return next week to Paras; thinking his specialty may be of help with this. Only worked till 4:30, pain is worse. After work I went to the Rush for the x-rays. I called Bardackâ€™s nurse and asked if there was anything I could do for the pain. They called back with a prescription of Percaset, but I have to pick up the script, they canâ€™t call it in because it is considered to be a controlled substance. The office closed before I could get back down there, have to go tomorrow morning. Work had to force OT for Thursday and Friday; I would have had to work 2 hours OT each day.  6/24  Went to therapy at 10:30, Maria really worked on the right side of my back good. Left therapy at noon so I could drive to Yorkville to get my script for the medication. After I got it, I dropped it off at Walgreens to pick it up after work. Work was rough, but I did get thru 4 hours, first time this week. Have noticed some things I donâ€™t like on the truck, the new bug guard is all-wrong, I have to call the body shop and get the Jeep back in. Didnâ€™t take the new meds yet.  6/25  woke up in a lot of pain. Took 1 percaset at 10am, 1pm I could still feel its influenceâ€¦very helpful. Back does hurt a lot, work is gonna suck. They announced that Saturday would be scheduled, could even work more than 8 hours. Also said that Sunday maybe offerred. I keep losing OT and since I canâ€™t work, when I do get back to full duty, I will be on the bottom of the OT list. Called the body shop, gonna go in on Monday to get center cap fixed. I also asked them about the bug deflector.  6/26  Pain. Morning is still tuff to get moving, more so than normal. The Endocet (percaset) helps, but it only lasts soo long. I still will only take one at a time, I am afraid of how 2 will make me feel. Later in the evening I took Naproxen and Tylenol at the same time. I feel a vibration in the wheel that I canâ€™t identify. Iâ€™m gonna go by Brianâ€™s (hope he is back from vacation) before I go back to Cars on Monday.  6/27  Pain wakes me early, too hard to sleep during that day. Took Endocet and then Ibuprofen again today. Canâ€™t seem to get comfortable anywhere. Every night, after sleeping, I wake up sweating; some times on the couch, others in bed and the A/C is on, set at 80.  '),(17261,4177364,'09,July,2004','2004-07-09 00:00:00','   6/28    KILLER KARMA.   Was in a lot of pain at work today, at 5:20 I went to go tell Wes that I was going to go home due to the pain. He said that we needed to talk about my safety violation (from May 26th). He then told me that I was suspended for the next 2 days with no pay and that \"I should be real careful as my job is in Jeopardy.\"  What the hell is that suspose to mean? He also made remarks saying that the suspension was a result of my not remembering my \"violation\". He then said that because of my accident it must be due to \"post traumatic stress disorder\". When the fuck did he become a doctor.  6/29  should be a fun day! Called and left message for Blatnerâ€™s service as needing someone to talk to. Called Karen Ryder and discussed my accident. She did say, that while investigating the incident that one of the witnesses said that \'they\' offered to move the truck but i said, \"no need, i\'ll get it\" and used the truck to push it. Because of that statement, they wanted the punsihment to fit the action. She was as confused as i was as to what Wes said yestaurday, \"You need to be real careful around here, your job is in jeopardy!\" She said that she is part of the safety steering committe and she knew there was nothing like this in my safety record, so she has no clue why he said that.  6/30  did nothing, just canâ€™t think. Did go see Spider-Man 2...cool movie!  7/01  therapy and 4 hours at work.  7/02  therapy. Did 2.5 hours at work. Wes again giving me shit because for some reason he doesnâ€™t know what my restrictions are; granted they have been the same since 6/16.  7/03 â€“ 05  long weekend off, didnâ€™t do much. Did spend time priming my front axle as it looked like shit. those bastards, when they replaced my axle, they did it with one that was covered in rust.  the axles seems to work fine, just looked horrible.  Back was really hurting though, had to keep it short. Good for the day.  Gary called; he has the race tickets and said we get a bus ride to the race also, with free snacks and beer.  Free tickets to the Nextel race in Joliet...too fucking cool!  7/06  Dr. Limaâ€™s (a shrink with Dreyer) office called, best they can do is 2 weeks. After therapy I need to call Dr. Parasâ€™s office and ask for a referral. Iâ€™m just too tired of dealing.  7/08  Dr. Paras gave me a referral to a psyhologist, not a psycaitrist. Also has me on same rstrictions for next 2 weeks plus a new muscle relaxer to take at night.  '),(17262,4177364,'15,July,2004','2004-07-15 00:00:00','   7/15  Itâ€™s getting harder and harder to remember to write down the daysâ€™ activities. Gary and I went to the race on  7/11 . I was really hurting, but didnâ€™t let it bother me, as I wanted to have fun. When I got home, I iced, took muscle relaxers and was asleep before 10pm.  7/12  was horrible; I was soo sore. Besides my back and shoulders, my knees hurt badly too. I only worked 2.25 hours. Had therapy at 6:30pm, but nothing helped.  7/13  was a little better, but still hurt. Should have gone for shots, but was too lazy and too sore.  7/14  had therapy early, didnâ€™t use stim, I hurt so much I took one of the new muscle relaxers before work. By 5pm I was dying to leave. The 13th, 14th &amp; 15th I worked 3 hours. As much as I wanted to leave early those days, I pushed till 5:30. All week, everyday they have been needing 120 hours of OTâ€¦i would have gotten forced everyday. In doing so, I find myself taking muscle relaxers more and longer periods on ice. Numbing the pain is only good for so long. Taking the percaset first thing in the morning is the only thing that allows me to get past the pain first off. Itâ€™s not getting better and Iâ€™m tired of all the drugs and the ice. I made an appointment today for a shrink at Dreyer for the 30th at 11:40am. Maybe she can help, Iâ€™m just so tired of all the pain. The only people I seem to talk to anymore is people at Fridayâ€™s and Karen (she has enough of her own problems). I made an attempt to talk to dad about how I have been feeling and it was the first time I have ever seen him back away from the conversation. Things must be bad if Iâ€™m scarring him away. For some reason my left arm and shoulder have been in a lot of pain. My left knee still hurts real bad.  '),(17263,4177364,'21,July,2004','2004-07-21 00:00:00','   7/16  I worked 3.5 hours; think it was a mistake as I hurt a lot.  7/17  I was under the Jeep, painting the axle that was replaced (black flat primer). When I got done, I was using the gloss enhancer to clean the hood, get rid of dirt and overspray. As I did, I noticed problems in the paint from the body shop. I also noticed that not only was the bumper the wrong color (they gave me a flat black one over the gloss I had) but my bumper guard (on the passenger side) doesnâ€™t even fit the bumper. Also ripped 35+ CDs to the puter. Took the old (store bought) CDs to Disc Replay, they only took 5 for $10. They only take those that have no scratches. I need $ bad enough to sell a large part of my CD collection.  7/18  did nothing. Helped out Karen at Tiger, she bought me a spindle of CDRs for helping.  7/19  felt well, therapy hit me hard and I stayed at work for the 4 hours. It hurt, wonâ€™t know till tomorrow if it was that stupid to work the 4. Tomorrow is Mariaâ€™s last day at Athletico.  7/20  Momâ€™s being a real pain in the ass. She emailed on the 11th and then again on the 12th. I emailed back telling her that I was just busy and tired and would write soon. This past weekend (Friday-Sunday) she must have tried 5 or 6 times. Thanks to my answering machine message, she wonâ€™t leave a message. She just doesnâ€™t understand that I donâ€™t want to talk.  7/21  worked 3 hours yestaurday, did nothing much last night. When I got up today, I thought maybe I was feeling better. Went about 2 hours before I gave in to the pain and took the percaset. As I told Wes Tuesday, everyday I go to work, I hurtâ€¦it just depends on that day how long I can withstand the pain. And that is the problem. I canâ€™t even remember the last day that I wasnâ€™t in some kind of pain. Even on good days, I hurt. I thought this morning I was doing good, did 2 loads of laundry. But as I was on the stairs and I felt no pain, I bent down for something and took a while to get back up. I canâ€™t believe that this is just going get to better in time, the pain sucks. As I am totally broke, going to go back over to Disc Repaly tonight and try to sell some more CDs. I seperated the best condition ones from the rest. Hopefully Iâ€™ll get some more money from them  '),(17264,4177364,'11,August,2004','2004-08-11 00:00:00','   urlLink    This is just so how i feel at work sometimes&nbsp; urlLink  '),(17265,4177364,'09,August,2004','2004-08-09 00:00:00','   urlLink    Is this a double negative? &nbsp; urlLink  '),(17266,4177364,'09,August,2004','2004-08-09 00:00:00','   urlLink    I think this resort is for specail people&nbsp; urlLink  '),(17267,4177364,'09,August,2004','2004-08-09 00:00:00','   urlLink    Someone found Nemo&nbsp; urlLink  '),(17268,4177364,'09,August,2004','2004-08-09 00:00:00','   the body shop that fixed my truck has a list of things to fix. i called the insurance Tuesday about the other damaged wheel. it seems the body shop just called it in, they new about it for the last week. the insurance company, State Farm, has dragged there feet with this, but now the body shop is doing the same. this is one of State Farm\'s certified shops called Cars Collision. they charge State Farm $5300 for the repairs, took a month to get my truck back. 600 miles later, here is the laundry list of shit they have to fix: 1) paint - dirt in paint on front end and hood 2) front bumper - they got a flat black one, mine was gloss - also, the old bumper guard (original equip) won\'t fit on this replacement bumper. 3) hood - not lined up correctly 4) bug deflector - the one they put on was different, the new one moves around while driving 50mph or faster. They ordered the new one. 5) windshield washer - the washer pump is in the receiver which is in the front fender. the pump isn\'t working 6) passenger wheel - inside of rim is bent being repair, may have some damaged to the tire and allingment. *7)now this one is a good one.  about a week after i had the car back, i told the guys at the body shop that my driver\'s door isn\'t lining up right.  they told me that it is normal wear and tear due to using the door as leverage to get in and out.  there lies one problem, even as a big guy, i don\'t use the door, i use the steering wheel as leverage.  so yestaurday i went thru the car wash, first time as they said i needed to wait atleast a month due to the paint.  as the car got a wash, i received a shower.  that car has never leaked and i got soaked. fuckers will hear again from me.  i\'m sometimes afraid to drive the thing as more may break. it sucks, it\'s just not my truck yet. but that is just it, it is MY truck. i\'m keeping this beast. things that sucks with the insurance is they won\'t settle, or pay anything at all, till all the medical is done. so i\'m screwed on getting other repairs done or doing my lift or new rims and tires or a new bumper. the only extra cash coming in now is from selling shit i own. granted, i do have a lot of shit, i suppose it\'s about time i got rid of some of it. so, music CD\'s, VHS movies, some DVD movies, a couple of books, PC video games and then i can pick away at my collections of trading cards and comics and diecast cars.  it\'s hard to think about how normal my truck will be, or how fun it will be to play off road again when i hurt like hell, i can\'t enjoy.  '),(17269,4177364,'09,August,2004','2004-08-09 00:00:00','   No matter what you add to cheap pizza, it still tastes like shit. i try to save a little and i get the blandest tasting shit i have ever had. hell, this time i tried to add some pineapple and extra cheese and still shit. i figure now what they must do is just drown everything in water so that it has no taste. generic versions of other foods are usually pretty good. bread is the same. even the fake Mac & Cheese ain\'t bad, but not Generic peanut butter, that stuff sucks. food always has generic versions of the real shit. they also have cheaper versions. generic water, milk, ice-cream, butter are all fine.  Beer, has cheaper versions and some cheap beer really sucks.  but i would hate to try generic beer.  '),(17270,4177364,'09,August,2004','2004-08-09 00:00:00','   TIME   we never seem to have enough of it, time flies, where has the time gone? so many statements about time and yet it is always the same....we just never have enough of it.   TIME  ....Before the car accident, i never had enough time. I\'d work 8-10 hours a day and spend every weekend catching up on shit. well now, i\'m working 5 hours, 3 hours paid by medical; no overtime at all. so, now i\'m broke and have tons of time. Weekends, shit i don\'t do anything anymore besides relax and go to the bar on Saturday night. the local has karoke and as much as i hate hearing some people sing, it can be quite funny and besides, there are always lots of ladies. But, i\'m there to have a few beers and relax. this past Saturday this woman, a real pain in the ass to everyone (I\'ve nick named her Diane Chambers - as from the show Cheers, as she has an opinion about everyone and won\'t shut up) is pestering me about why i should go hit on these nice looking women. my thought was that i\'m relaxing, i feel good and if they want me, they know where i am. ok, well that didn\'t happen...it did Friday night, met a nice lady; she walked away from her date to come say HI. i thought it was kool. back to   TIME  . Weekends now, relax. Weekdays, i actually do a lot more than before. With only 5 hours work, i should have more time. But when i start the day on muscle relaxers and heat, get off work to come home to muscle relaxers and ice, well it sucks. Feels like an 8 hour day of work. i hate the pain, i hate even having to think ahead about doing normal shit as it may hurt like a bitch.   TIME  , weekdays last too long, the pain is bad. i sleep less these days, makes the days even longer. Any other time, i would luv to take a walk, exercise, but i can\'t as i know i need to go into work as fresh as possible to last 5 hours. To get thru the day, to not get mad about shit, to relax as much as possible, day dreaming can help. for months I\'d lost my creativity. i would day dream about things i wanted to build, things i wanted to do, things i wanted to write. \"Eating Peanuts\" refers to many things, a new story, a way of relaxing, a distraction that can suspend the pain for hours. Not to mention, it just sounds goofy. What exactly is it.....Most read on, not giving that up yet.   TIME  , during this time at home, must keep mind rolling or I\'ll wind up a couch potato for sure. And just where did \"couch potato\" come from? is a potato supposed to be lazy? Any more than a corn cob or green bean? a tit seems lazy, could be a \"couch tit\"? that sounds weird too. Speaking of weird, i\'m listening to Roger Water\'s album \'Amused to Death\', this is a weird album, much like his others. And as odd as it is, it is so easy to listen to. so, trying to stay amused, i play some dumb games, post on here, add occasionally to my web pages (i get bored with those occasionally too) and then there is the online personals. Trying to find someone that compliments, challenges and completes me. Damn, when i was a kid, I\'d never thought about any of that shit. the women did, hell i just wanted the sex. Now older, it\'s the other way around. Women have tried for the relationships, they failed and they just want sex. Not that is bad, but now i need more....that is so funny. ok, i keep getting distracted on my topic of TIME and now i actually have to do some things before work. Yuck, one of those things is call the insurance company and raise hell. Will have to update on that later and also have to tell about the fight i was almost in Friday...Quite funny actually.  '),(17271,4177364,'08,August,2004','2004-08-08 00:00:00','   The Beginning   It seems as time goes by i find one thing or another to occupy my time. and as i try more things it always seems that i forget how i got here to begin with. \"The Beginning\" is my first time, my blog, my thoughts and the world that i am floating in. Web pages, i have played with different things and ideas, but this will allow me to express who the wizard is behind the curtain.    Twisted Karma   Most people don\'t believe in karma, many don\'t believe in anything. i have my moments, my karma, is very twisted. good and bad, ying and yang, karma i didn\'t really believe in till 2 years ago. i have to say, i\'m afraid to have anything good happen, as in every turn in the last 2 years, once there was good, the bad just wasn\'t a fun trip.    2002  , won a years supply of M&Ms...that was funny as shit, a package came with 365 coupons for individual bags of M&amp;Ms. as much as i loved those things, i hardly eat them anymore. also that year i participated in an online nascar fantasy game. the game owner was dumb enough to make the entry free figuring they would have a lot of people buying merchandise from there store. there problem started when when they allowed up to 50 people to win prizes....for the first 12 races, 42 were playing. i won every week and when they decided to make a change and charge people, it was time to cash out. so for free, i got 2 hats (usually $25 ea) and a t-shirt. Hell, i was doing well. then at work i injured my back, restricted duty, physical therapy and drugs for 4 months. that sucked.    2003  , the little things, the occasional smiles from hot women, helping someone you don\'t even know, just doing a good deed, these always slip the mind. the Boy Scouts motto is \'do a good turn daily\'. easy to do as a kid, but as we grow up, things tend to slip away. me, i still try to do a good turn daily. anyway, this year was filled with new girlfriends, keeping my truck from getting stuck in the mud and just having fun. June i injured my shoulder, but not too bad. July 1, had to leave work early as i didn\'t feel good. on the 3rd, my doc checked me into the hospital for cellulitus in my left leg; i was there for 8 days. several doctors, to many nurses, pretty good food and $1200 per day room; thanks to work for awesome insurance. later, after i was back home and heading in a better direction, the doc told me, \"one more day, if you had waited one more day to come see me, you would have lost your left leg from the knee down.\" thanks doc, that is one thought that has never left me. took 6 months to rehab my leg to get back to normal. one month after getting home, my friend and i went to the local riverboat casino. we went for the food. after eating, i only had $30 in my wallet, but decided to play anyway. started with a 25 cent video poker, it\'s cheap and fun to play. lost $10 in 2 minutes. put in my last $20, deuces wild game and i was betting only 4 quarters (5 was max) and i was dealt a royal straight flush. freaked me out, i won $250. the old bag sitting next to me had to point out, \"If you had made max bet, you would have won $1000.\" thanks bitch. i played $100 and walked away with the rest. of course my best friend (John) was calling me an asshole for winning. as we were about to leave, i stopped at another dueces wild machine, had to stop. it played 3 hands at a time, so i played the $2 bet. first 3 hands, won $75, John called me a dick, so i kept playing. next hand, won $150, he called me a dick again.....it\'s always nice having your friends call you names when you win. this game deals you one hand, you select what cards you want to save, it saves them for 3 hands and then deals again. i kept playing. after 10 minutes or so, i realized something. i had been hitting the \"Max Bet\" button as i was playing, i thought it was just $6. nope, it was 15 credits x $2 = $30 a hand. Oops. holy shit, $30 at a time. i did this for a few hands, kept winning and listened to my foul mouth co-hurt next to me. the last hand, i was dealt 3 dueces...the dueces were wild. i froze, new that 3 hands with 3 wilds was gonna be real nice. after the new cards were dealt, i just stared at the machine trying to figure what i won. my 3 hands: 5 of a kind, royal straight flush and 4 dueces....that last one was the kicker. i won $2290. that was a good day. it seemed to be Karma smiling at me, giving me something nice for dealing with my leg. 2 months later, i was carted out of work in an ambulance, i couldn\'t breath or move. wound up being pleurisy (the muscle around my lungs were basically bruised).  i didn\'t need that. come November, i won a trip for 2 to Las Vegas from my works x-mas party.  took my dad out there in March of 2004 to see an old friend of his and to see a Nascar race.    2004  , this year started good.  had some hurdles to overcome at work, back to full duty, leg seemed good and the trip to Vegas was awesome.  good enough, my dad wants to go back next year and 2 of my brothers want to join us...i just hope i\'ll be able to do it.  things felt good.  couple relationships i had, they were wierd.  women thinks guys are assholes...there are a lot of twisted women out there too.  May, boss was an ass, i got pissed, that was dumb...they gave me 2 days off in June, no pay cause i work with a bunch of people that don\'t understand that people in glass houses shouldn\'t throw stones.  then, May 27th.  i was heading to work, maybe doing 5-10 of the limit, nothing i didn\'t usually do.  actually i just started to slow down, maybe 45ish.  a woman driving a 2002 Lexus RS300 pulled out right in front of me, i had 15 feet till i hit her.  i drive a \'99 Jeep Cherokee...it still isn\'t right after that.  i did succeed to tearing the front of her car right off.  she got the ticket, accepted full blame.  over 2 months later, shit is still fucked up.  my Jeep will have to go back for a day\'s worth of repairs.  as for me, well pain killers, muscle relaxers and restricted work hours, i\'m still messed up.   so, do i let karma mess with me?  things good and bad, this life is still a fun trip.  besides, i beleive that karma atleast owes me some winning lottery numbers.  the only thing i can seem to win is a couple dollars...lol. wait, gotta check tonights numbers.......damn, didn\'t win again.  well, for a first posting, i believe i have started out good.  as for what \"Eating Peanuts\" means?  that will have to be another time.  '),(17272,4177364,'08,August,2004','2004-08-08 00:00:00','   urlLink    gotta love Canada.  those little fishies are Walleyes, 3-4 usually fed the 4 of us.  i miss fresh Walleye&nbsp; urlLink  '),(17273,4177364,'08,August,2004','2004-08-08 00:00:00','   urlLink    yeah, i know, i just look like a goof sitting in front of the puter.  oh well&nbsp; urlLink  '),(17274,4177364,'08,August,2004','2004-08-08 00:00:00','   urlLink    Perfect....this is one pic that i really dig&nbsp; urlLink  '),(17275,4177364,'08,August,2004','2004-08-08 00:00:00','   urlLink    doesn\'t look like $5300 damage, does it?  my back and neck hurt just looking at this pic&nbsp; urlLink  '),(17276,4177364,'15,August,2004','2004-08-15 00:00:00','   always an interesting title, crap.  i think the weekends are harder on me then the weekdays, i\'m always too damn tired and have way too much shit to do. i need to relax  '),(17277,3683867,'19,June,2004','2004-06-19 00:00:00','   Wooohooo!!! My first blog log....! I didn\'t know it\'s THAT easy...   '),(17278,3683867,'17,July,2004','2004-07-17 00:00:00','   Let me not to the marriage of true minds Admit impediments.  Love is not love Which alters when it alteration finds, Or bends with the remover to remove: O no! it is an ever-fixed mark That looks on tempests and is never shaken; It is the star to every wandering bark, Whose worth\'s unknown, although his height be taken.   Love\'s not Time\'s fool, though rosy lips and cheeks Within his bending sickle\'s compass come: Love alters not with his brief hours and weeks, But bears it out even to the edge of doom. If this be error and upon me proved, I never writ, nor no man ever loved. (Sonnet 116)    Love looks not with the eyes, but with the mind, And therefore is winged Cupid painted blind. (A Midsummer Night\'s Dream, 1.1.231-2)    She never told her love,  But let concealment, like a worm i\'th\' bud, Feed on her damask cheek. She pined in thought,  And with a green and yellow melancholy  She sat like Patience on a monument,  Smiling at grief. Was not this love indeed? (Twelfth Night, 2.4.115-120)     I love thee so, that, maugre all thy pride, Nor wit nor reason can my passion hide. Do not extort thy reasons from this clause, For that I woo, thou therefore hast no cause But rather reason thus with reason fetter, Love sought is good, but given unsought better. (\"Twelfth Night\", 3.1.151-6)     Love comforteth like sunshine after rain... Love\'s gentle spring doth always fresh remain. (\"Venus and Adonis\")   '),(17279,3683867,'08,July,2004','2004-07-08 00:00:00','   urlLink MTV   ä½œè©ž:å¾æ²³å¨œ  ä½œæ›²:æœ´èª æ—¥(foræ–°æ‘music)  ç·¨æ›²:æœ´èª æ—¥  å°ç™½:æˆè©©äº¬    Baby don\'t you know it\'s for you   ä½ çš„è‡‰è®“æˆ‘é–‹å§‹æ„Ÿåˆ°è¿·æƒ‘ æ‰€æœ‰çš„ç–‘å•è¿·å¤±åœ¨ç©ºä¸­  åƒå¯¶è²èˆ¬å‘µè­·çš„å¸½ æ²’æœ‰å¤ªå¤§çš„è®ŠåŒ–  å¯’å†·çš„å†¬å­£æ°£å€™ å°±åƒæˆ‘å€‘åˆ†æ‰‹çš„é‚£å¤©  é€™æ˜¯ä¸€å ´é¬§åŠ‡ æ²’æœ‰ä»€éº¼æ„Ÿè¦º(ä¿é‡..çœŸçš„å¾ˆå°ä¸èµ·)  ä½ æ˜¯å¦é‚„è¡£ç„¶çœ·æˆ€è‘—ä¸çŸ¥é“ä½•æ™‚èƒ½æ’«å¹³çš„å‚·å£ å“¦~~ åˆ†æ‰‹çš„é‚£å¤©æ˜¯æˆ‘å…ˆè½‰éŽèº« ä½ æœƒä¸æœƒæ¨æˆ‘  æ˜¯ä¸æ˜¯æˆ‘çœŸçš„å¤ªå£ž ä¸è¦èµ° é€™ç¨®æ²’æœ‰éª¨æ°£çš„è©± æˆ‘è£ä½œè½ä¸è¦‹  åœ¨å¿ƒä¸­ä¸åœå–ƒMY Love  å¿ƒç—›çš„å–˜ä¸éŽæ°£ä¾† ohh   åƒæ¼«ç•«ä¸€æ¨£ æˆ‘å€‘ä¸€è¦‹é¢å°±åµä¸åœ åƒæ‹‰å¾·èˆ‡æ„›èŽ‰ä¸€æ¨£æ¯å¤©..........  åªæœ‰å¬´æ‰èƒ½è®“æˆ‘çš„å¿ƒå¾—åˆ°å¹³éœ ...ooh (å°.....æˆ‘å°ä¸èµ·ä½ )  ä½ æ˜¯å¦é‚„è¡£ç„¶çœ·æˆ€è‘—ä¸çŸ¥é“ä½•æ™‚èƒ½æ’«å¹³çš„å‚·å£ å“¦~~ åˆ†æ‰‹çš„é‚£å¤©æ˜¯æˆ‘å…ˆè½‰éŽèº« ä½ æœƒä¸æœƒæ¨æˆ‘  æ˜¯ä¸æ˜¯æˆ‘çœŸçš„å¤ªå£ž ä¸è¦èµ° é€™ç¨®æ²’æœ‰éª¨æ°£çš„è©± æˆ‘è£ä½œè½ä¸è¦‹  åœ¨å¿ƒä¸­ä¸åœå–ƒMY Love  å¿ƒç—›çš„å–˜ä¸éŽæ°£ä¾† ooh   **ååœ¨æ¨“æ¢¯å£ç­‰å¾…ä½  æˆ‘æœŸç›¼èƒ½å¤ å†æ¬¡æ“æœ‰  ä¸åœåœ°ç¥ˆæ±‚è‘—ä¸çŸ¥ç¾åœ¨ä½ é‚„æ˜¯ä¸æ˜¯ä¸€å€‹äºº  ä½ æœƒä¸æœƒæ¨æˆ‘ æˆ‘æ˜¯ä¸æ˜¯çœŸçš„å¤ªå£ž  ä¸è¦èµ° é€™ç¨®æ²’æœ‰éª¨æ°£è©± æˆ‘è£ä½œè½ä¸è¦‹  åœ¨å¿ƒä¸­ä¸åœå‘¢å–ƒMY love  å¿ƒç—›çš„å–˜ä¸éŽæ°£ä¾†~ooh no no wowo no wowowoooh  è£ä½œè½ä¸è¦‹åœ¨å¿ƒä¸­ä¸åœå‘¢å–ƒMY love  å¿ƒç—›çš„å–˜ä¸éŽæ°£ä¾†ooh    ì•„ë§ˆë„ ì‚¬ëž‘ì´ê² ì£  - ìž¥ë‚˜ë¼     Baby don\'t you know it\'s for you   ê·€ì°®ê²Œë§Œ ëŠë¼ë˜ ê·¸ëŒ€ ì–¼êµ´  ì°¨ì°½ë°–ì„ ì§€ë‚˜ ë©€ì–´ì ¸ê°€ìš”..  ë§ˆì¹˜ ë³´ë¬¼ì²˜ëŸ¼ ì•„ë¼ë˜ ëª¨ìž.  ë³€í•œê²Œ ì—†ë„¤ìš”...   ë²Œì¨ ì½”ë ì‹œë¦° ê³„ì ˆì´ ì™”ì£ ,,  ìš°ë¦¬ í—¤ì–´ì§€ë˜ ê·¸ë•Œì²˜ëŸ¼  ì¡°ê¸ˆ ìš°ìŠµì§€ë§Œ ê·¸ëŸ´ë“¯ í–ˆì—ˆì£ ...  (ìž˜ê°€.. ì •ë§ ë¯¸ì•ˆí•´..)   ê·¸ëŒ€ë„ íž˜ë“¤ì—ˆë‚˜ìš”  ì„œë¡œê°€ ì–¸ì œ ê°€ì•¼í• ì§€ë¥¼ ëª°ëžì—ˆì£ ..  ì²˜ìŒí•˜ëŠ” ì´ë³„ ë‚˜ ë¨¼ì € ëŒì•„ì„œ ë²„ë ¸ì£ ..   ë‚´ê°€ ë¯¸ì› ê² ì£ ..ë‚´ê°€ ë‚˜ë¹´ì–´ìš”  ê°€ì§€ë§ë¼ë˜ ê·¸ ìš©ê¸°ì—†ë˜ ë§ì„ ëª¨ë¥¸ì²™í–ˆì£ ..  ìž‘ì•„ì§€ë˜ ëª©ì†Œë¦¬ MY LOVE~~  ë‚´ ë§˜ ì•„í”„ê²Œ í•˜ë„¤ìš”..   ë§Œí™”ì²˜ëŸ¼ ìš°ë¦° ë§Œë‚˜ë©´ ì‹¸ì› ì£ ..  í†°ê³¼ ì œë¦¬ì²˜ëŸ¼ ë§¤ì¼ì„..  í•­ìƒ ì´ê²¨ì•¼ë§Œ íŽ¸í–ˆë˜ ë‚´ ë§ˆìŒ..  (ê·¸ëž˜.. ë‚´ê°€ ë¯¸ì•ˆí•´..)   ê·¸ëŒ€ë„ íž˜ë“¤ì—ˆë‚˜ìš”  ì„œë¡œê°€ ì–¸ì œ ê°€ì•¼í• ì§€ë¥¼ ëª°ëžì—ˆì£ ..  ì²˜ìŒí•˜ëŠ” ì´ë³„ ë‚˜ ë¨¼ì € ëŒì•„ì„œ ë²„ë ¸ì£ .   ë‚´ê°€ ë¯¸ì› ê² ì£ ..ë‚´ê°€ ë‚˜ë¹´ì–´ìš”  ê°€ì§€ë§ë¼ë˜ ê·¸ ìš©ê¸°ì—†ë˜ ë§ì„ ëª¨ë¥¸ì²™í–ˆì£ ..  ìž‘ì•„ì§€ë˜ ëª©ì†Œë¦¬ MY LOVE~~  ë‚´ ë§˜ ì•„í”„ê²Œ í•˜ë„¤ìš”..   ë‚®ì€ ê³„ë‹¨ìœ„ì— ì•‰ì•„  ë‚  ê¸°ë‹¤ë¦¬ë˜ ëª¨ìŠµì´..  ë‹¤ì‹œ ì°¾ì•„ì™”ìœ¼ë©´ í•´ìš”..ê¸°ëŒ€í•´ë³´ì§€ë§Œ  ê·¸ëŒ€ ê³ì—” ëˆ„êµ°ê°€ê°€ ìžˆë„¤ìš”..   ë‚´ê°€ ë¯¸ì› ê² ì£ .. ì´ì œ ìš©ì„œí•´ìš”..ë‚´ê°€ ë‚˜ë¹´ì–´ìš”  ê°€ì§€ë§ë¼ë˜ ê·¸ ìš©ê¸°ì—†ë˜ ë§ì„ ëª¨ë¥¸ì²™í–ˆì£ ..  ìž‘ì•„ì§€ë˜ ëª©ì†Œë¦¬ MY LOVE~~  ë‚´ ë§˜ ì•„í”„ê²Œ íž˜ë“¤ê²Œ ë§Œë“œë„¤ìš”..   ëª¨ë¥¸ì²™í–ˆì£ ..  ìž‘ì•„ì§€ë˜ ëª©ì†Œë¦¬ MY LOVE..  ë‚´ ë§˜ ì•„í”„ê²Œ í•˜ë„¤ìš”..   urlLink   '),(17280,3340813,'25,June,2004','2004-06-25 00:00:00','   In life, we are all bound to be hurt by certain memories, feeling or thoughts,even by people.ETC. Commonly, it\'s what the most common yet sacred thing that leaves a scar deep enough in us.{\"Relationship.\"}Everyone is given a special innate ability,and it\'s all real up to us on how one interpret it. Being empath at times, flashes and thoughts of emotion do seek out for me.It hurts.Really. When the first light was shone out of the dusty cloud that seems to shroud every single negativity within,True feeling\'s also being spurred out in search of those who are hurt.Deep and subconsciously. It\'s to heal them.Certain phrase of our lives do make an impact on us,especially our love\'s not being reciprocate back.But deeply, there lies a Will.A Will to both love and care, and to protect the ones from all danger, always lingers between our inner sense.  *Have you ever felt hurt when the ones you really hope for\'s hurt? *Have you ever felt pains from the ones whom you hope for? *Have you ever had a sudden urge to really hug and tell the ones you hope for, the 3 sacred words? *Have you ever had a sudden urge to \"save\" them when trouble brews? Well, I have. All this experiences are what\'s present in every single human. ~There lies a golden \'lock\' deep in one\'s heart.It is what it is to find the real \'key\' and that\'s to release the passion and desire within and send forth to the ones you really hope for. Flowing with it, comes along \'pain\' and \'tears\'.It\'s hard to avert it,but for as long as you know, there always lies a pair of hand out there, waiting to help you up and giving you strength.To move on. My life was filled with every different possible emotions and feelings for you can ever imagine.* There is a boy,a boy with extraordinary life. Yet,he has found the right key,to say. But sadly,it\'s a key that seems impossible to hold. From then on,there\'s always memories and flashes about this unique girl and the most important phrase that keeps him holding on :\"There\'s always hope as nothing\'s impossible.\" Up till this very moment,this very time,this very month, this very day,this particular boy still preserve the urge to protect,to love and care for \'her\'.And for one thing he hopes to get this to her ears:   Cry if thou will, but stand up once again from where you have fallen and lead a whole new life,with light and transformations for thou\'s greatest good and highest joy.It is what it is,and what it is to be.  '),(17281,3340813,'04,June,2004','2004-06-04 00:00:00','   Every phrase of life leaves behind a foot step for Reflection and rememberance.Dont think about the past but focus on the present as it concerns the future.It\'s ain\'t wrong to think of the olden times{past}, but remember,it should only act as a lesson learned, and not a regret. We all have a pair of eyes, it is what it is to look far beyond the superficial, and deep into the silence Heart of others.That\'s the true Way of understanding them, their beauty and bright perspective. Beauty is indeed in the eye of the beholder. A Phrase of Advice,\" don\'t ever let Regret devour you\".Sometimes, we underestimate or even overlook the importance of some people in our presence.The true values will then only surface when the one is gone...forever.And deep down, we all know that certain things when gone, is never going to be back.Even if it is, the once warm yet familiar bond will change,perhaps to certain extent that\'s impossible. Every one of us is special and unique in certain way,each of us contains innate gifts that\'s waiting to be awaken and embrace.Thus, find yourself.Your \'True Self\', don\'t let time flies anymore.Grab It. You Will. We have our own life to walk,and this once ina life time opportunity should be handle with pride,besides Dreams and ambitions are waiting for us to catch with \'them\'hence,all the more we should bring out the best in ourself. Treasure all those that\'s true and important to you. Cherish the times spend with them. Remember, Don\'t let regrets and sorrow devour you..Take care. It is what it is and what it is to be...  '),(17282,3340813,'31,July,2004','2004-07-31 00:00:00','   Dreadful things often happen during period of uncertainty.Well, just received some news that it\'s up at it again.It is often some miscommunication and misconception that actually give spark to this\" Trouble Fire\".I mean, look friends, there\'s always a solution for all these seemly unsolvable problems.Indeed indeed, gone are the days when \"FRIENDS\" got together.Perhaps some of us need a break, a break from one another?certainly. The situation around here just got better a little.but, it seems to me that they have just being their once again trouble journey.Why?. Reflect upon the great times. You know what I mean.  '),(17283,3340813,'28,July,2004','2004-07-28 00:00:00','   Have you ever had the feeling with someone who seems distant from you yet there lies a invisible but deep,strong friendship, to the extreme extent? Believe it or not, this case of relationship do exist. It\'s pretty much magical to really accept this bond as part of you.I mean,you do not seem to have anything to do with him/her,yet Your heartbeat accelerate whenever something\'s amiss, that at the same time, concerning him/her.Perhaps even to the extent of shedding tears base on the smallest mishap that could ever possibly befall on him/her.It\'s a warm and strong emotion down there, but do not let these theory lead to the misconception.[supposingly if both are of the same gender]It\'s really there, deep down that this pure friendship do lingers. Well, but as they say, not all fairy tales end with a positive ending. At certain times,things do change a little, that just because of this little minor change, it might possibly cause depression due to the fact that the invisible friendship\'s dimishing. That is why, people feel  down out of the blue. Both positive and negative influences are out there in the world awaiting for us to explore,just imagine, should him/her walk towards the negative sector, and for example take up the art of smoking,What would you feel? But of course, heartbroken!{with the heartbreaking level to the maximum}Once more, theres no reason for this.Actually there is, but it\'s base on the pureness status of the friendship between both of you. The purer it gets, the more excruciate the pain feels.  '),(17284,3340813,'26,July,2004','2004-07-26 00:00:00','   Yes, \"Procedures\" certainly are still fine this days.Well, but one particular arising issue seems to threaten the peace and serenity around here a little. Disappointments between friends? Why do such things exist.Perhaps, there in this time, still a solution?.Disagreements and disputes seems to be soaring all around that =|once familiar area of our lives|=.[You know what I mean]Maybe there still some light for all of us? ~certainly there is~. why should the same kind turn against each other in a both unfavourable manner? Please,calm down, it\'s only turning out to be a lose-lose situation. Have we all forgot those great times we had? The memorable feelings and thoughts?. I hope not so {but it may jolly well proved me a little wrong.} Let\'s get back to the golden days.There might be different sortings of \"groups\" and people just simply obtracize those who are against them or something, and honestly, this doesn\'t feels good. As a matter of fact, this case \'s still existing till today.With the only simple reason, that would be, one unable to take notice of this situation. Will there be a tomorrow for all of us?[Think for it]  '),(17285,3340813,'17,July,2004','2004-07-17 00:00:00','   Mind-Mode: currently working hard in persuit of my dream. Mood:Whitish orange. My everyday life? well, it just got better a little, bits by bits. After all those strange and absurd trails and tribulation[minor ones], well,it seems that a whole new overlook idea about my path have manifested. Working hard and thinking positively\'s difinitely crucial for me in all walks of life. My dream? Lets say, the \"main course\" would be being a designer.\"Dessert\"? lets leave it to Fate.  '),(17286,3340813,'16,July,2004','2004-07-16 00:00:00','   \"Gee,gd luck to ya n all tha best for ya chinese Os tmr..Jia you!!.\" \"Haha..U too!! Enjoy yr day.. Cya.....\" \"Oh ya hor ur o level oral tml..Jia you!\" \"Haha,Yeah..Sweet dreams...\" \"a NICE FREN is lik wiNd,u\'ll nEvEr KNoW wUt\'S iNsiDe thEir hEaRts, but u wilL alwAx FEel,theIr pReSeNsE & siNcErITy.thaNk u fOr bEiNg oNe!Nitez...\" \"Yup..Study hard..Wish ya ALL tha best.\" \"Work hard.Jia you!!\" \"Haha..Thanks..U too..Do yr very best...\" \"Oh..good luck then..\"  These are all those wonderful and encouraging messages That I\'ve received.Thanks.  '),(17287,3340813,'15,July,2004','2004-07-15 00:00:00','   \"Friend\",this word indeed cause a big impact in my life.Well,It really does.They gives you strength to move on, a shoulder to cry on.I really wish to express my deepest gratitude towards those who\'ve accompanied me during the dark path.Sometimes, the bright sunny smile and warm healing touch lightens my day, and come to think of it, it makes me feel real \'secure\'. \"Secured\",that Friends are always there for me, listening to all my worries and fear. cheers, to you. Thanks for having trust in me. You are important to me.[as in terms of friendship] This above message\'s specially dedicated to my two beloved buddies and all my girl-friends,last but not least,to those who knows,who and what am I talking about.{you\'ll understand}Well, there\'s still so much to express...  '),(17288,3340813,'14,July,2004','2004-07-14 00:00:00','   It\'s just another boring day.Things still goes the same way.Nothing much.Well, there\'s one little tinky winky thing that actually marks today.Ask me? That would the oral version of Geography memory test.Well, guess that was so \"nice\" of the teacher to actually think such method to implement.Anyway, guess what? I flaunk my math test again, as usual.Honestly,I belong to the anti-mathematic organization. Perhaps the effort wasn\'t high enough. hello? Anyone care for a counseling for me?  It\'s really terrible to received the 25/80 paper first thing in the morning when it is supposed to be wonderful for me.In the end, the result hit me like a bolt of blue.Unexpected! But it wasn\'t that bad afterall,I mean at the very least, the lowest score wasn\'t me.{ Ok, trying to optimist around here.} Minutes after that devastated period in the morning, well, my spirit seems to\'ve astral projected to the other dimension, leaving with only a depressed smile. Friend? He \'s wonderful[guess?]Not only did he not gave me a weird look due to my \"excellent\" result,instead,words of encouragement reached my ears.He\'s really one of the friend that me, really treasure and cherish.Alright, his words and warm touch gave me strength and it actually lifted my spirit! Wonderful! Cheers to him!Thanks a lot.Currently, This 25/80 boy\'s trying to get out of the A.M.O.  [well, that\'s just being optimist]  '),(17289,3340813,'10,July,2004','2004-07-10 00:00:00','   Well, actually it\'s kind of weird that certain thoughts come searching for you.Maybe it\'s regarding the future? Different people really do have different perspective point of views. Sometimes, we do have dreams to realize and that would means starting from now? \"Stories\" emerged from different phrase of our lives and some of them do cause great emotions emitting out of our mind/heart/soul.Maybe that\'s what they called \"experience\". Tears and happiness, pain and fear, Cries and laughter, all of\'them\'once came up to us , guess it\'s time to have psychological preparations from now for the \"Time\".  '),(17290,3340813,'03,July,2004','2004-07-03 00:00:00','   ShadowLand my feets are on, Boundry seek no barrier fond. Dawning Light i seek no more, Misty Dark \'s end for all? Heed no soul my heartache be, Who\'s always there for me?? Verbal for they only knew, Heart\'s content ever conseal. Trouble cauldron brews for me, Lonely path i walked through free. Questions ask of me I say, \"Why is always in this way?!\" Tears and blood and fear conseal, heed no voice my mind should feel. Orbs of path by vision\'s array, Pain no more,my fear shall sway. Golden dawn my path there is, Challenge expel through my Fist.  '),(17291,3340813,'09,August,2004','2004-08-09 00:00:00','   Time have clearly shown the truth, before my mind\'s eyes. A truth that implies simply only on one particular issue.Well, at the very least, I do know there\'s someone out there.Maybe it\'s better for me to be in this way for a period of time. =\"It\'s alright to leave me out,it\'s alright to neglect me\"= , but have you ever experience this? So what if you did?, have I not?!All \"you\" does, is come searching for me with all those inner cries, and you know that there\'s surely someone there for you.You know something, a thing which is always conceal within? I \'ve never voice out any grievance! Impressive, no one seemly even care about this.Good bye. Gone are the days.Good bye!  '),(17292,3340813,'08,August,2004','2004-08-08 00:00:00','   Are they actually even there? So, all those words and phrases are  whole lots of boof?!Yes,things changed but not till the extent of this situation!Gone are the days, at the very least, there\'s still someone for me to hang on. Perhaps that\'s just me, the way my life was meant to be.There are ways indeed to solve this matter, but are there anyone who cares? \"We\"[refering to a huge amount of people] talk about everythings under the sun often, and yet at times miseries from you[refering to individuals] could be heard.Have anyone thought about mine? Even if there is, do they even care to pour your feelings out like I do? There\'s not much time left for eveyone,so it\'s better for me to be left alone.For a while.Sorry for any inconvenience cause.Accept it, the way it works.  '),(17293,3340813,'06,August,2004','2004-08-06 00:00:00','   Under the scorching sun, walking about aimlessly, that\'s how the day passed. Throughout the time of 13 hours,my mind was constantly reflecting upon those abstract questions. =Do they really seem to be the ones that knows you inside out?= =Are they the ones who actually really go through thick and thin with you?= Sometimes, in our lives, I mean OURS, we tend to be hurt, but do they, what you termed as friends, feel your pain that\'s excruciating? even from afar, it might not work.Some of them would just reply you\" Dont\'t dwell too much on this particular issue\" or even \" things are fine\" Have they not thought about the real pain that seems to be conseal within you?. Maybe it\'s just normal to them. isn\'t it? I might have failed in some of the above, but at the very least, I thought about it, with the right state of mind. Certain issue\'s too much diffcult to express verbally,even between  bossom friends. Now, this stage requires the telepathic skills.[but looks like nothing seems to work out fine.]This is how the way it works, for me.{nothing much to express. For the simple reason, I\'ve already gotten used to it}  '),(17294,3340813,'05,August,2004','2004-08-05 00:00:00','   Again, a new discovery on me. After much observations on the surrounding,finally a almost accurate conclusion\'s out.Indeed, I am just a free flow-er.A person who \"flows around easily, with not really a group to finally ground the feets on. It\'s pretty much normal for me as I \'ve got used to it, lets say since the beginging of march? yes.Companies of friends are always within my area of vision, and honestly, this doesn\'t felt good. Perhaps these are the reason to solve the self contradic question impost on ourself,:\"Why Am I so down suddenly, or recently?\" As they say,life still flows.Of course it\'s ever flowing but with the only exception, That is being a Free flow-er in all walks of life. My everyday process was seemly quite peacful.[that\'s a good thing to celebrate.] why? Why is it happening on me?, really, is it the only reason for me being weird? as in unable to blend in? Being a optimist, I would certainly never allow myself to fall in a never ending pit of ALMOST excruciating pain.  Indeed, The days are fine  and wonderful experiences are often around, so much in being a epicurean of fine daily simple food and laughter.  '),(17295,3340813,'05,August,2004','2004-08-05 00:00:00','   Part1=The passing of time seems to inject some sense into me.For the only thing that\'s what I am talking about. Yes, recently, indeed the \"strife\" have been unknowinly resolved,but there still lies a hidden emotion in everyone.  Part2=Remember those time when some of us lost our most precious belongings? Well I did recover those things, but with the addition of a feeling, \"heartache\".=Why do you have to lied to me for not taking \'them\' with you last year?= Things are often out of our expectation.The news did really hit me like a bolt from the blue.Why is the prime suspect you?!! Please!, don\'t make me have a second view on you. Enough of this.  Part3=Bonds between friends seems to be dimishing as well.Days are no longer the same anymore, Not as close as ever and those telepathic skills are slowly grinding to a halt.Why?!  '),(17296,2413978,'29,November,2003','2003-11-29 00:00:00','   Last night was awesome in that I got to see Ricky - the coolest person in the world - again since he went to college in August. Its seemed like forever since he left even though its only been like 3 months - I guess because I\'ve gotten involved in so much in school life seems to have gone a lot slower then it actually has gone.   Since my Mom has been in the hospital I\'ve been in a really bad mood - I guess because even though I\'m consciously worried, subconsciously I\'m a lot more worried then I let on? This is my Jr. Psychology Mind speaking... Last night at XSCAPE there were a bunch of Hip-Hop Bands and they were just giving me a headache - Cara went with me. It was pretty bad with Cara because she was too shy to go with me to hang out with my friends, but she had to be really critical of everyone. Whenever someone she didn\'t approve of walked by, she whispered some comment about them in my ear and it was really annoying. Finally when she left I went over by Ricky and he got the hint that I wanted to talk and we talked awhile. It was cool talking to someone about some of the stuff I\'ve been dealing with for awhile (Jacob, Gabe, Ricky, Tommy, Becky - if you\'re reading this you know what I\'m talking about) and I have like...the ambition to overcome this junk now.   Also this one guy told me off when he heard me call my Dad an ass-hole - and even though he was a little over-board while yelling at me - he somehow made me appreciate my parents a lot more then I did - which I needed. School has just been having a bad effect on me lately in terms of cussing and I got kinda a wake up call on that by someone online and several people in person. I mean when you slip \"F\" at church, its pretty bad.   If you\'re reading this and you\'re a Christain, or practice some religion where you pray for the well-being of people, please keep my Mom in prayer, because she got really sick a week ago (Sunday) and its pretty scary. They\'re going to (as of now) remove a portion of her colon and minor intestine - and I\'m gonna have to take care of her for awhile - which is an honor since shes taken care of me my whole life - but if she has a bag or something with her waste - thats not gonna exactly be something I\'m gonna want to deal with - how gross would that be if I had to empty her bag of waste? (ick)   Tonight when I was browsing around on the internet I came accross this \"Blog This!\" button on my Google Toolbar, and I decided to check it out - so that is what you\'re reading right now.   Finally today my $100 check came from Virgin Mobile - so now I\'m rich! (actually after phone bills I only have $42 left.) I guess my money from the essay I had purchased from the text-book company and the report card money is going to really be needed when I buy Christmas Presents and CDs and stuff.    Michael R Brant One Love - One God - One Way  '),(17297,2413978,'28,November,2003','2003-11-28 00:00:00','   This is just a test to make sure this Blog Archive System works right.   Michael R Brant One Love - One God - One Way  '),(17298,2413978,'03,December,2003','2003-12-03 00:00:00','   This morning when I woke up - I knew immediatly I was like \"REALLY\" sick. I got a flu shot - so either the shot gave me the Flu, I got a different kind of flu - or I have a super bady bad cold.   My face is burning (but I have no fever), my head hurts, my ears are plugged up, my nose is running \"like a student in a tardy sweep\" and I just feel like crap.   It sucks I missed school today cuz I have a big test tomorrow and I needed to get help. Plus I had a chance to re-take my dance test and do better and I think I blew it.   I\'m still fuming about my Dad\'s evilness...he didn\'t even care when I told him I was sick today *sarcastic gasp* and I think he brought this on by being a d*ck last night - all the stress.   My Mom\'s also sick so we\'re being sick together (fun huh?). I\'ve been taking Zinc Pills cuz thats supposed to help you get un-sick but all they\'ve been doing is making me get nausious - its not fun.   Love + Cookies Peace Out  '),(17299,2413978,'02,December,2003','2003-12-02 00:00:00','   Right now I feel super sick. My head is hurting, my throat is hurting and my nose is running like a faucet. It wont stop and I\'ve gone through half a roll of toilet paper - its hoorible I says!   I dont know if I have a cold or if I\'m getting the Flu after getting the shot (maybe they gave me too much of the shot or something and I actually caught the virus it\'s self). But this hoorible. I really dont feel like going to school tomorrow because I feel so sick - but there are tests and a Mid-Term coming up in several classes so it would be stupid to miss again.   Peace Out  '),(17300,2413978,'01,December,2003','2003-12-01 00:00:00','   Today was good at school - but when I got home it just got terrible. After school I had to go with my Dad to come to the Hospital to pick up my Mom. As soon as I got in the car he was like \"You\'re going to be \'punished\' for not calling me last night when Gloria got to your house\". Since my Mom was in the hospital last night (got out today) someone was here staying with me so if the boogy man comes to masacre me, they can die along with me. My Dad is really...stupid about some things. So basically he\'s gonna take all sources of my Money because I didn\'t call him to tell him someone was here - and he was in bed - he wouldn\'t even have known until in the morning that she was there or not. Whats the point?   Right now I\'m venting on one of my friends on AIM. I guess thats better then calling him and venting...atleast he doesn\'t have to hear me all worked up...But what are friends for?   Today should have been a happy day. I got my report card - all A\'s again except one B in PE - plus my Mom got out of the Hospital - but my Dad just has to ruin everything. He\'s ruined my life. He doesn\'t even know who I am, what I stand for, what I like to do - what I dont like to do. Its really sad. Everyone tells me how lucky I am that my Dad stuck around when my parents got drivorced - cuz a lot of parents leave their kids...but how lucky was I really if I have to deal with this s**t all the time?   I guess it helps me to like get all this stuff out on the net - it lets me vent a little and then I might not start telling someone off for something simple tomorrow. Ok - now that I\'m done writing about all the crap. Heres the good stuff. Today my grades just rocked, and I did better then I expected I would do on my dance quiz. I still think I got a \"C\" or \"D\" - whatever a 2 on the Rubric is, but thats still pretty good. Cuz I really was dreading today.   We went to the store and I got some new hair gel (I was outta it) and some body wash - my Mom spent like $106 on medicine (man is that stuff expensive) and then I went home and cooked Spaghetti. It sucks that my Mom couldn\'t eat the Spaghetti - but its cool Gloria and I enjoyed it. It was pretty good if I have to say :).   Gotta stay after school tomorrow to get help (desperate help) on Geometry. So I better go and get my good sleep.   Mikie Peace Love and Crackers  '),(17301,2413978,'01,December,2003','2003-12-01 00:00:00','   Right now I am so tired. I just got back from the Animal Feed store where I bought Rabbit Food, Cat Food and Cat Litter - and then Subway and I just feel like I\'m gonna colapse.   Because my parents dont want me spending the night alone in the house - I\'m waiting for Gloria Sampson to show up. I wish she would hurry up because I really want to go to bed right now.   Today in 6th period was interesting. I\'ve been absent the whole time the class was working on Photoshop (and learning it) and today we had a test I was forced to take and I totally was able to take it. I think I probably aced the test. My little wooden people looked nice.   My Mom said she might get out of the hospital earlier then we all thought. She was able to eat today and hold it all down so she might get out tomorrow. And also she sounded super strong. I just got done telling her that I\'m gonna cook her this huge Thanksgiving Dinner when she gets home since she was unable to eat during Thanksgiving.   Peace Out - Penguins - Ninja Monkeys - and Seals!  '),(17302,2413978,'30,November,2003','2003-11-30 00:00:00','   Today so far has been ok, but I\'m really tired of Geography Class. The teacher is really awesome, but she doesn\'t teach us anything - she either talks the entire period or gives us book work - so that basically makes her an over-paid baby sitter.   Ashley is being more evil then ever. Today she said this whole line of crap and I was able to fight saying anything sarcastic to her but I couldn\'t stop myself from rolling my eyes and she saw me and made a huge scene in the classroom. I hope the teacher does something to stop Ashley from being a mean person all the time because if not there will have to be a big deal with my parents going into school and all that.   Pretty soon I need to go on to Geometry and then Multimedia - then after that Ill get sleep (finally - I\'m so tired).   Peace Out   Michael R Brant One Love - One God - One Way  '),(17303,2413978,'30,November,2003','2003-11-30 00:00:00','   Last Night I got in bed at 11:00 and probably didn\'t fall asleep until around 2:00 am. I\'m totally going to be dragging today. I don\'t even have coffee to like completely stay awake.   Today at school since I missed Wednesday I get to get a admit card for Wednesday, make up a quiz and a bunch of work *ick* and I have to find out how to get out of my truancy from PE Wednesday - cuz I was there...just not there (Misunderstood the teacher).   Time to get going - peace out  '),(17304,2413978,'30,November,2003','2003-11-30 00:00:00','   Dinner was good tonight. Consisted of Pork Roast, Cabbage, Mashed Potatos and Bread. Now I\'m like really full.   Michael R Brant One Love - One God - One Way  '),(17305,2413978,'30,November,2003','2003-11-30 00:00:00','   Today I went to church and it was cool because we were completely packed down. I talked to some friends and really entered into the worship. We broke into our cell groups - although my cell group isn\'t the greatest. It has awesome people, but I have absolutly nothing in common with my group leader person. He\'s a rapper/skater where I\'m a nerdy regular person. And since I dont feel I have anything in common with him, where I should feel I\'m able to spill my guts and talk to him and other people in my group about \"anything\" I dont feel like I can.   After church I hung around a little to say bye to Ricky before he drives back to Texas,  talked to the Pastor for a couple of minutes and then climbed down the hill to jump in the car with my Dad. We went by my Mom\'s house and I changed my Cats\' food and played with the babies a little while. Then we drove to the hospital, PVH, where my Mom is and I went and hung out with my Mom for about an hour. After that I came back to my Dad\'s Apartment and here I am now, being bored and updating.   Michael R Brant One Love - One God - One Way  '),(17306,2413978,'30,November,2003','2003-11-30 00:00:00','   I woke up today at 8:00! I know I say that like \"ITS SO EARLY!\" but it totally does seem really early today since I was up really late watching Little Nicky last night.   Starbucks \"Double Expresso\" is keeping me awake with my whole 7 hours of sleep (whoopdie doo!)  Time to be off to church now - shall be back at 1:30.  '),(17307,2413978,'13,December,2003','2003-12-13 00:00:00','   Right now I\'m doing a lot of HMWK make up because I was out three days last week. I have a load of Biology HMWK to make-up, some make-up Geometry HMWK, regular English HMWK and then some Dance HMWK. Because I did bad on a dance test I\'m doing a Dance Research Resport to make up for it and I have to type 3-5 pages which really sucks. Thats like a major deal, I cannot believe thats the teacher\'s definition of make-up work...and I have to write it on Ballet. How lame.   Today I was at school from about 8.5-3.0 editing the newspaper, but I got a lot of it done. My Mom was in some kind of strange depressive mood and tried to pull this huge guilt trip off on me - which really was irritating. But finally later in the day she cut it out. When I got back from school we went grocery shopping, to the ATM and then to check on my Dad who my Mom was all worried about since we hadn\'t heard from him - and it turned out it he wasn\'t even home.   I\'m going to continue working on my HMWK and then get to bed. Right now I\'m feeling sick, but hopefully in the morning I\'ll feel better and make it to church.  '),(17308,2413978,'10,December,2003','2003-12-10 00:00:00','   Today was an average day at school. In Multimedia I finished up the photo editing Travel Photo things and tomorrow I will probably be a roaming aide to helpout the desperate freaking out people who aren\'t finished or dont know how to do something to their project.   After school I took Adan his missing work and stuff - and then I headed home. At home I did some chores and just basically messed around until it was time to go back to school for the Dance Concert. The Concert was pretty cool. The best dances were this one by Madonna \"Hollywood\" and of course Evanesence\'s \"Bring Me To Life\". The other songs (because most of the people were beginners) wern\'t anything to write home about.   Peace Out  '),(17309,2413978,'09,December,2003','2003-12-09 00:00:00','   Today during school I was roaming around and finding out various things about the school swim team. I\'ve gathered its a Spring Sport - not a Winter Sport and that I\'ll be one of three guys *blushing*. I found someone I kinda know who is on swimming, Liliana Ramos (surprised me) and she was pretty thrilled to hear I was gonna join up on the team.   I have this plan for a crazy work out (Every Weekday from 6-7:15 am at Loma Verde Pool (and I\'m  soooooo  not a morning person - this will be interesting.) I\'ve just found out it\'s either $1.75 a pop (to swim at the pool) or $60 quarterly (thats a better deal if I\'m gonna stick with the deal). Also on top of the swimming I need to tone up (I would make people go blind if I was in a Speedo) so I\'m going to start doing sit-ups, crunches and also start running more (ick - I hate running).   If I can stay motivated and do all this stuff, I\'m going to be a great swimmer and well on my way to a letter jacket (You have to make it to Metro-Championships to earn a Letterman.)  '),(17310,2413978,'09,December,2003','2003-12-09 00:00:00','   Today was a typical school day. I walked to school because my Mom is still sick and I should have gotten up early to go run (to make up my grade) but I really didn\'t feel like getting up for some reason - so I stayed in bed.   In PE we worked on technique today rather then the dance and the girls in the line I was in are very evil. I was like up in the front of the line and they kept like screaming at me to move and they all cut me. I know this sounds lame - but it sucked. What was I gonna do? Obviously I cant push them outta the way like I would a guy, cuz they\'re girls - and I\'m not going to rat on them to the teacher; so I just had to live with being in the back of line. I guess my satisfaction can be that I\'m still a better dancer then they are.   In Geometry today we messed around again, but it wasn\'t as fun because two of the key people I goof off with were absent today - so it was one goofer-offer and then the two serious people: our \"Humble Leader\" Marcella and Rachel \"Our Secretary\". We all have titles, because we\'re doing this thing where each person in the class (in a group) is assigned a project [or section] of Chapter 5 to teach to the class. My title is the \"Substitute Person\". This is actually kinda funny - I guess reading about it rather then being there makes it less fun.   In Multimedia it was a cool day. It really sucked because Adan wasn\'t there (he must still be sick) and that class isn\'t the same without him. I never thought I would openly say I have guy-friends who I \"love\" but him and several others are even cooler then \"like\". Not, like/love in a sexual way, obviously - but they\'re like...better then friends. Almost like people I consider to be brothers. Anyways - in Multimedia today I was the photographer person for this project we\'re doing there and that was kinda fun. Then Mr. Pearson took several pictures of me for my project that I had to erase the background on, and then I\'m going to add my pictures onto a background of some place to travel to. Its going to be pretty fun I think - and they\'re going to turn out good.   After school I went home to where I finished clearing the backgrounds from the pictures of myself, and then proceeded to HMWK and eating. Dinner was good - consisted of some little mini steaks that my Mom cooked well and some of the best mashed potatos I\'ve ever tasted. My Mom said she used whipping cream rather then half & half or milk - and man we need to start using the stuff more when making mashed potatos.   Right now I\'m writing here, updating my Calendar and drinking really good Jasmine Tea. But I gotta get on to my Grography HMWK - so peace out.   Wanna check out my Calendar? Go to http://calendar.yahoo.com/mrbrant2002. Peace Out   New Pictures  '),(17311,2413978,'08,December,2003','2003-12-08 00:00:00','   Today has been a sucky day.   At school I had to get all the work I\'ve missed for like a long time (fun fun) so now I have a major load of stuff to make up - and I had to spend my lunch break in Mr. Pearson\'s class to make up my absences where I didn\'t accomplish very much (and he had to deal with stupid students who dont do their work - I feel sorry for him)   My night has consisted of very nasty food from La Bella (they messed up the whole order and gave this dressing they call \"Ceaser\" and I also think I got this kind when I asked for \"Italian\" before too.) Then the cat ate too much or something and puked in three places around the house. After cleaning up cat puke I had to clean up her litter box that had Cat Diahrea and then also clean up the bathroom where the litter box was... Also the nights activities entitled one of the kittens getting stuck in the back of my chair (no idea how) and when I got up outta the chair it almost got crushed to death...Its my favorite of the bunch of kittens we have so I was freaking out and then my Mom started screaming at me (like its my fault) and she tried to swing at my but I grabbed her hands and stopped her. I\'ve never wanted to hit her back more. She always blames every problems or issue on me - no matter how far-fetched it is.   My Mom\'s friend is over who luckily calmed her down and she has chilled out now. I wish I could just go in my room and be left alone right now. Its not my fault my Mom\'s sick and she expects too much. She thinks every waking hour I\'m not at school I should be waiting on her ass hand and foot and if I dont she calls me lazy and says I dont do \"anything she asks me to do\" which is the biggest load of BS I\'ve ever heard. Right now I\'ve just had a crappy night and I usually would sleep it off but my Mom\'s friend who is giving her a break is sleeping in my bed - but thats better then having to deal with my Mom all by myself I guess.   I gotta go - peace out.  '),(17312,2413978,'07,December,2003','2003-12-07 00:00:00','   This week/weekend since I\'ve been sick on Wednesday, Thursday and Friday have been pretty miserable. My Mom has been super moody and just been insulting me like crazy. You\'d think she hates me, but I dont know how seriously to take it since shes probably not in a right state of mind since she is so sick right now   I e-mailed some of my teachers to find out what work I\'ve missed and only two of them responded so I\'m going to have a lot of catch up to do when I get back to school on Monday (tomorrow)   Today I\'ve been lazy pretty much. I didn\'t even go to church because its been raining today, my Mom caught the Cold/Flu I had, and I had no ride other then the bus in the rain (bad idea) so I watched a movie and now I\'m on the net.   I\'ve been buying my friends Christmas Presents and stuff and I just put a check in my bank account that my Dad gave me and I have a lot more money then I thought I had - which is good because now I can pay off my book and CD bills and Ill still have a surplus to buy necessary presents.   It sucks that I missed church today because I wanted to see my friends and since there was no XSCAPE on Friday because there was a Marriage Seminar - I didn\'t get to see them till today - so now I wont get to see them till Friday.   I found out one of my friends has been in the hospital for 2 days with the flu *shiver*. In San Diego we supposedly have some strain of the flu that the vaccine does not protect against - it just helps give you a more managable strain of the flu or something like that - so its cool the Hospital people were able to take care of Adan (the friend in the hospital). There will probably be a lot of sick people tomorrow at school.   I better be off. Still got more presents to order and all that good stuff. Peace Out  '),(17313,2413978,'18,December,2003','2003-12-18 00:00:00','   This is going to be short since I have little time. Lately I\'ve been really busy with Newspaper at school but we\'ve been pretty productive. Its been fun. I\'m now officiall the \"General Editor\". Today I had to run like 1.5 miles but I think 1/2 of the extra mile was extra credit or something. Today it was cool because we had 20 minutes to run atleast 6 laps (1/4 miles each) rather then run a mile and you\'re scored by your time. I actually made the cut today, so I guess it counts as an \"A\".   My friend Adan has been out of school now for two weeks, which is a bad sign since he was in the hospital with the flu, his condition likely got worse.   I had to go to the Assistant Principal over being harassed the other day, and unfortunatly I was unable to think of direct quotes (that the harasser was saying to me) so they couldn\'t do much about it.   I\'m getting ready to get in shape for competative swimming so I\'m going to start these intense workouts starting January 5th.   Today myself and a group of my classmates had to teach a lesson in Geometry, Medians of a Triangle, and we did generally well. Part that sucked was I spent like 3 hours putting together our notes, cool down and homework assignment, but then all in all we were probably the most organized group.   My Dad and I have been getting along terribly lately, so I\'m going to generally be staying with my Mom for a while, and see how it works out with my Dad. Never know, we could start getting along again - although I\'m not really counting on it. I think I\'ll have a much more balanced life staying with my Mom. Its a pain going back and forth.   This week I\'m going to help my Mom by cooking some new recipies that she can hopefully eat. Last night we went to the Commesary and spent like $211 on all this funky stuff to go in the Asian dishes i\'m going to try and make. One was this funky-fied vegetable that we eventally put back cuz it looked pretty gross.   One more day of school until I get out for Christmas Break and then we\'re off to San Fransisco for Christmas. On Christmas day as a matter of fact.   We\'re getting ready to get rid of the Kittens which is gonna suck, but in the long run it will be nice. One adult cat is a lot easier to care for then 6 babies. I just had to scrub like 50% of my Mom\'s room and remove poop from like all over.   Michael R Brant One Love - One God - One Way  '),(17314,2413978,'22,December,2003','2003-12-22 00:00:00','   I\'ve been super busy, so right now I\'m posting just to keep this gradually updated, I feel like I\'ve been neglecting my Blog.   Today we had to take the 4 remaining kittens to the Humaine Society and put them up for adoption because my Mom doesn\'t want to keep them, and it was pretty hard. I stayed strong and didn\'t cry, but my Mom broke down like three times in the little room where they ask you a million questions. We went inside of the actual place and looked at the kennels and they were like feline condos. They were rooms about the size of my bedroom (how cool!) and the cats had like these huge fort thingies to play around in. I would want to live in there if I was a cat.   After we got done there we drove back to Chula Vista and went to this Sushi place, but we forgot its closed on Mondays, so we didn\'t get Sushi, but my Mom saw one of her friends at the nail place next door and they stood at the car and talked for literally 1.5 hours...I was bored to tears. After we left there we went for lunch at the 50\'s Diner where I got Pork Chops and then we went shopping after that. At the store we went to, I got two pairs of pants, a sweater and some long sleve\'d shirts which rock-cat-butt.   After that I went back home and waited for the skirmish with my Dad to take place. He finally showed up late \" Howwwww Disrespectfulllll \" as my dearest father would say. This morning I called him and asked to stay home just because I didn\'t feel like dealing with his crap and he tried a guilt trip that I kinda ruined. He was like \"If you don\'t want to come over, then don\'t come over for presents either.\" So I was like, \"Ok, then do you want me to mail your presents?\" and it was really funny cuz he was speechless. He thought that would make me like come running over - where in moderation I could give a crap about his presents. No present is worth putting up with his immature crap. I would rather clean an oven or something. I could deal with that brain damage at school if I wanted to.   After our little supervised immature skirmish thingie, some of my Mom\'s friends came over and like completely surprised her since shes been pretty sick. They came with some awesome Philapino food that we like pigged out on (not literally) and it was awesome. Then they made Martinis and got drunk and acted kinda wierd (but it was funny.)   Misty, my cat, has been super cuddly tonight with me since her babies are gone now. Its been a little strange because since shes been a Mom shes been kinda anti-social.   Last night we went shopping at Horton Plaza Mall, went to Seaport Village and then went to the Hyatt 40th Floor and got Drinks and Poo-Poos. It was pretty cool. After that we went to the Navy Exchange (Dept Store for narrow-minded brainwashed Navy \'folk) and I got some new clothes - and my Mom got some new shoes and clothes as well.   Over the last weekend I saw \"The Last Samurai\" with Tom Cruise and \"Lord of the Rings: The Return of the King\" and they were both pretty good. Although LOTR was very off the book - they twisted around two of the major themes from the book. On the other hand, Last Samurai was the best movie I\'ve ever seen in my life. I loved that movie a lot. I\'m really gonna buy it. One of the best things about that movie was it had romance, but no sex and I don\'t recall it having any cursing either...which are always a plus.   I opened a few Christmas presents already and I got a new cell phone that actually has a plan rather then the pre-paid one, got some clothes, got the movie \"The Ring\" and something else I kinda forgot about *gasp*.   I\'m out - Merry Christmas to all.  '),(17315,2413978,'31,December,2003','2003-12-31 00:00:00','   urlLink 7StoryMacys.jpg     urlLink KniveIBought.jpg     urlLink MikeGoldenGateWharf1.jpg     urlLink GherradeliXmasTree1.jpg     urlLink LombardStreet1.jpg     urlLink ArtMuseum.jpg  '),(17316,2413978,'31,December,2003','2003-12-31 00:00:00','   For Christmas my primary gift was a trip to San Fransisco. We flew out from San Diego to San Fransisco on Christmas Day actually and left San Diego right at the start of a big storm. We flew into a lot of turbulance and the flight seemed to go pretty fast because my Mom\'s friend who went with us had one of those portable DVD player thingies and we watched 007 the whole flight.   When we got to Oakland Airport we took the BART (Bay Area Regional Transportation) which is a high speed public transit train (goes above and below ground) to San Fransisco and then took a Taxi from the BART Station to our Hotel. We stayed at the San Fransisco Carlton and it was pretty nice. They had free Wine and Cheese and all that good stuff in the lobby (too bad I\'m not 21) free for adults - so that shows how nice it was. Anyways, after we un-packed we went to this little Resturaunt down the street that was Italian (all the people who worked there including the cook were Chinese) and we had some pretty good food. I ordered Veal Parmasagne (one of my favorite dishes) and then we went back to the room and hit the sack. After flying all day we were all pretty tired.   On the second day we went down to Lombard Street on the Cable Car and then to Fisherman\'s Wharf, and later on to Japantown. At Japantown we went to this Japanese Buffet place and got all this awesome Japanese food that would have cost a fortune at a non-buffet resturaunt. After that we went to all these cute little Japanese stores and bought some cool stuff. After that we went back to the room and crashed.   On the third day in San Fran we went to Chinatown where we looked at leather jackets, eat some dim sum and I bought this cool Dagger thing (my xmas present to myself.) We took Gail to the BART Station for her ride home and then my Mother and I got on the MUNI Metro (A Public Train that runs just in San Fransisco - above and below ground) to Haight/Ashbury, this area where all the hippys used to hang out and we went in some really curious stores. One had a bong that was larger then I was...and I\'m pretty big. After that we went to a Coffee shop and got some Tea, and then went to this Cheese store and got some Gouda and this nasty Non-Alcoholic Peach Wine Stuff. We took the MUNI back to our hotel and then watched TV for awhile and crashed. On the 4th Day we went to Chinatown again and bought leather jackets (my Mom and I both bought them.) Later that day we went to Union Square (the 7 story Macys) and later went to Japantown - back to the same Buffet we went to on day #2.   The last day we stayed in our room (it was stormin\' in San Fransisco) and we took the Bus to the BART Station and took the BART back to Oakland. We rode the AirBart to the Airport in Oakland and then departed back to San Diego. On the flight I actually fell asleep for like the 2nd time every I fell asleep on a plane to be woken up by this little kid behind me screaming about \"PIKKKKKKACHU!\". I wanted to turn around and strangle the little kid....   We got back home \"home sweet home\" and just like hung out...today is the first time since Monday Afternoon that I\'ve left home.   What I got for Christmas: Digital Camera, Phone, Sweaters, Phone, ReChargable Batteries with Charger, Cologne, Candy, Leather Jacket, Indiana Jones DVD Set, Bad Boys II DVD, Termenator III DVD and some other usuaul stuff. It was pretty good this year overall. I got my parents books and my Mom got some CDs too (Yo-Yo MA).   Yesterday and Today my friend Isaac Martinez came over to my Mom\'s house with his Laptop and I fixed it up for him and then installed a bunch of junk on it for him. He installed some cheap software on his computer that fried the whole operating system and I had to re-install windows and all the Samsung OS stuff that came with it. Then I put Microsoft Office XP, Macromedia MX, Photoshop 7.0, Imageready 7.0, and Flight Simultor 2002.   Peace Out - Happy New Year.   Michael R Brant One Love - One God - One Way  '),(17317,2413978,'17,January,2004','2004-01-17 00:00:00','   Wow...its really been awhile since I\'ve updated. But what can I say, I\'m almost always busy as a teenager in High School. But life continues to get cooler almost everyday.   The last couple of weeks have been pretty awesome. On Saturday January 10th 2004 I turned 15 years old and went up to Anaheim California to go to Disneyland with my Mom and her friend Gail. It was a really enjoyable time other then my constant fighting with Gail because she is ghetto and wierd. We got the Southern California Residents two day tripper and went to Disneyland Park on Sat where we rode every ride except the rides in Toontown. And we went to The Blue Bayou for lunch, which was a five star resturaunt inside of the Pirates of the Caribbean ride where I ordered Prime Rib and got Birthday Mouse. Gail took a used candle from my Birthday Mouse and put it in her purse. Maybe she\'ll try to sell it? Thats what I mean by Ghetto. And we still have a day in California Adventure which were going to use around the end of the month, the next weekend I\'m with my mother.   Right when I got back from Disneyland and went back to school on Monday the 12th, I was completely loaded down with work in school. The whole week I was completely stressed out with finals coming up and I was kept very busy. Most of the week I was running on caffine. Thank God for Coffee.   On Friday January 16th 2004 I ran my Dance/PE running final and made a really good time. Where I usually fail the runs, I made a C on the final at 11:30 for a mile, which was really good. Dance/PE is the class I was most stressing about. And I took a Geometry make up test on which I scored well too. I got a 79%, and even though thats a C, its still good. It didn\'t drop my grade that much, I still have a 93%. I\'m hoping to makeup for the grade a bit on the final and get it up to a 94% so I have a flat A, rather then an A-.   For my B-day, belatedly, my Dad took me shopping and I got some clothes, some new Vans and three DVDs. The DVDs were Pirates of the Caribbean, Underworld and LOTR: The Two Towers (Extended.)   Also Adan Padilla and I were working on our Multimedia Final as a team and produced a pretty cool looking, but amateur animation in Macromedia Flash. He came and hung out at my Mom\'s house and we worked on the animation for four to five hours and he ate over there. We bonded quite a bit, I learned some interesting things about him, and then he took off home at around eight pm.   I\'ve started a web design buisness as well, and I\'ve been having quite a fun time doing that. I\'m making quite a lot money and hopefully I\'ll go somewhere far with the skills I\'m developing.   Also this week, I\'ve heard back from Michael Inzunza, the school district magazine (Blurb) editor and he is going to publish the Blurb article I wrote in, came to my school and took my picture on the \'legenday soap box\' and he\'s going to start working with me more. So I\'m going to have a lot more opportunities with Blurb.   A bad part about me spending so much time with finals, my personal life and Blurb is that I\'ve been neglecting the newspaper at my school - and since I\'m editor and fought to get that role, I should be more devoted to the newspaper.   Peace Out  '),(17318,2413978,'29,January,2004','2004-01-29 00:00:00','   This week my primary focus has been newspaper. Since Monday, every night, I haven\'t been getting in bed before 1:00 am. Its been crazy, but I\'ve come up with some awesome articles. All the articles can be found on  urlLink TOG Zone  which is my website. There are some editorials and opinions and some hard news articles.   Blurb, which is the school district magazine has just published some of my works. Things To Be Thankful For, which is also on my website, was publised in the January Issue of Blurb and I made $20 off of it.   Since this is the start of the semester, I have to work back up an insurance grade. In English I had such a high average, that missing a couple of assignments would have still kept me at an A. Now I dont have that, so I have to work my butt off in that class. Also Biology, I had to arrange ahead of time with the teacher to turn in work late and still get full credit. Work that was due today, yesterday and the day before I haven\'t even touched yet. I\'m going to do it Saturday during Saturday school and turn it in Monday. If I would have done it ahead of time, and got it in early I could have scored a bunch of extra credit - which would have also helped me in that class build up an insurance grade (my phrase for it) since I usually get a bad grade on most of the quizs.   My Union Tribune subscription has begun this week and I\'ve been loving it. ITs soo cool to have a newspaper delivered to my day every morning so I dont have to go to the store and purchase one. I\'ve been getting awesome ideas from the newspaper for stuff to write about and I\'ve been working as a journalist to go out there and get people\'s opinions at my school so I can make my aritcles more interesting and related to my school.   My Dad was telling me that he can get back issues of LA Times free from where he works because when the newspapers get three days old they throw them out. That newspaper also has some interesting articles that I\'ve been using. Although I still prefer The Union Tribune - San Diego\'s Newspaper better.   I just decked out my room in a 1) calendar with big boxes to write appointments in, 2) a white board, 3) a cork board, and 4) a mini board with a cork board and a white board. I\'m getting more organized so I dont forget things I want to do.   I just read a proposition about ending world religion. This organization wants to pass an UN bill that will end all organized religion. Sounds like events from the book SOON by Jerry B Jenkins.   Its time for bed and I still need to do my chores. So I\'m off - Michael   Would you like to comment on a post? If so  urlLink Click Here  '),(17319,2413978,'25,January,2004','2004-01-25 00:00:00','   Whoa has my life been stressful the last few weeks. I did ok on my finals. Math I got a D+ which is a fair grade considering how hard the final was, and in Biology I had the second highest grade, which sadly was a C-. Whoever makes the final tests for Math and Science classes really have no idea that the people taking the tests aren\'t brilliantly smart or something, because those tests are usually an overkill.   My dance final I scored a \"C\" which is not fair in my opinion. I did better then a C...I deserved atleast a B, but an A would have been nice too. I think just because I dont kiss the teacher\'s butt she has it out for me or something. Also since its the end of the semester I\'ve been working on clearing up all my attendance and I have to go to Saturday school to clear off two truancys, and this morning I was so tired I didn\'t go. I just couldn\'t wake up - so I\'ll have to go next Saturday which just sucks.   I got a check in the mail today from my boss (work on his business\'s website) and from this Allergy Clinic for refering a patient to one of their studies. That was pretty cool. Also with part of my report card money from my Mom I bought the movie \"The League of Extrodinary Gentlemen\" which was a great discounted buy at Blockbuster.   Today my day blew, I cleaned the garage much of the day and that is about the most boring thing in the world. Plus my Mom never just gets the show on the world. She kinda just messed around for two hours out there before we actually got busy and she wouldn\'t let me do anything else, I had to stay in the garage, but then she complained that I was just standing around. She needs to make up her mind.   Jackie and her mother came over today with this super smart little kid they baby sit and we did our late xmas gift opening thing with them and did dinner. I\'d forgotten how annoying Jackie\'s mother is. Its been nice not having to deal with her since last October. We were trying to watch Underworld and every 5-10 minutes the phone would ring with her wanting something. I was ready to throw the phone out the window. She has called atleast 15 times before the night was over.   I\'m really tired and getting ready to go to bed. Hopefully I\'ll be up to updating sometime soon. I\'ve been so busy.   Also, grades came out Friday and I have a 4.0! I don\'t know if I already said this or not...      My new article that I just wrote:  urlLink Dowry Deaths in the Middle Easy and India.doc  (You must have Microsoft Word to read it.)   Would you like to comment on a post? If so  urlLink Click Here  '),(17320,2413978,'14,February,2004','2004-02-14 00:00:00','   Wow...I get less and less time to post here. Well lately I\'ve been like super busy. I\'ve been busy with newspaper, been busy with other clubs at school, been trying to apply for a job at Starbucks, trying to keep up my grades and really having no life outside of school.   My Mom was just in the hospital for surgery to remove a portion of her colon and intestine and it went well. She was only in the hospital four days, including the day they actually did the surgery which went like 4 1/2 hours. The second day she was in there, it was a little creepy. She had cords all over her and she showed me her scar from where they cut her open and it was really ick. There were like 6 staples, this little bottle to collect the icky drainage gunk and then like a thing that caught her urine and went up into her bladder (that must have hurt.)   School has been pretty busy. I\'ve been giving more to Newspaper, and the advisor doesn\'t tell me about a lot of stuff. I guess maybe she doesn\'t think I\'ve proven myself as editor or something? But it really sucks sometimes. I\'m supposed to be the Newspaper editor and she like...just expects me to read her mind and stuff - and no one tells me ahead of time about upcoming events and such. I should talk to her about it or something - maybe it will accomplish something, maybe not.   Lately I\'ve been getting ready for competitive swimming which is coming up soon. It\'s gonna rock. She also helped Rachel pay for her bathing suit and I\'ve gotten pretty close to Rachel. It\'s nice. I just wonder if its because she wants my Mom\'s help or because she really likes me. But who knows, shes getting touchy and huggy - so that would make me thinks it\'s real.   On Valentines day I bought three roses, unfortunatly they only had red ones...but oh well. I gave one to my math teacher (she was a little...surprised I guess), one to Lilianna and one to Rachel. Rachel gave me a hug, Lilianna was shy and was like \"from you!?\" but she liked it. Mrs. Hernandez was \"Thank you sweetie.\" But no hug... She probably thinks I\'m a freak or something.   Last night I was talking to one of my best friends, and I told him something like really personal and now he\'s avoiding me...so I think I might have done something stupid by telling him...so I hope I didn\'t just ruin my life at school or something.   Thursday I went for a job interview at Starbucks Coffe Co. and I was interviewed and I thought it went all well and everything and they said I would know by Friday at 5 pm if I got selected and I never got a call back from them. So that sucks. Probably they thought I was too young or something. Oh well - not I need to find a job elsewhere.   I\'m going to try and update more often but I\'m usually really busy, so who knows. It might be every two weeks or so.   Random Thought: My New Years Resolutions fell through...   Would you like to comment on a post? If so  urlLink Click Here  '),(17321,2413978,'05,March,2004','2004-03-05 00:00:00','   Again - whoa am I really busy!   When I thought I was busy before I didn\'t even know the beginning of it. Since I\'ve started swimming I\'ve hardly even had time to completely homework for school. Right now I\'ve saved most of it for this weekend. I have quite a bit of Biology HMWK, have English HMWK and I have to make up some Geography HMWK because the teacher messed up and lost my paper or something.   Competitive swimming has practices every weekday and some Saturdays from 3:30 - 5:30 pm. I\'ve been going to every practice and I\'ve went from size 42 pants to size 38 which is really an accomplishment for two weeks time.   School has been as crazy as ever. I\'m still pulling an almost 4.0 along with swimming and Newspaper. My life now officially revolves around like 90% school related stuff. But anyways, we got early grades, they dont actually come out until Tuesday and in Geography the teacher gave me a \"B!\" I got a B and I turned in all of the assignments. A lot of other people are angered and complaining as well as me. She was not the most happy person today. She even notices I was pretty mad. She was like \"I\'m sure if you knew VooDoo I would be dead right now.\" Fortunatly for her I do not know Voo Doo.   Miss Derwin is helping me step out of my comfort shroud and step up more and more as newspaper editor, so she is having me do a complete newspaper issue, which is good and bad. Bad - because sports isn\'t my thing and we\'re a little blank on it. Good because I can do things my way and not have to comply as much with how Miss Derwin wants to do things. I know thats a bad attitude to have...but I have a lot of good ideas that Miss Derwin doesn\'t always want to go along with.   I\'ve had an interview with the Principal on the whole issue at school where we get detention if we\'re late to first period and I asked her if there is any end in sight...and she said no. It was an interesting interview, but she didn\'t want the evil administrator image I wanted to portray. Boo.   Mrs Angela Hermamdez is officially the coolest CPH faculty member who is currently living or who will probably ever live. She is really sweet, seems to really love /care about me, and she really bails me out of everything in her class. I really dont know why she does it, I\'ve never met a teacher as giving as her. I mean...I so suck at math it looks like I dont even try...and you can see she thinks that, but she hasn\'t given up yet. If I had a student like me I would have given up so long ago.   We had a swim meet yesterday and I swam a 100 free, 100 breast, 400 freestyle relay, and 200 IM relay. I made 2nd place on the 100 freestyle and on the 400 freestyle IM. I got disqualified on the 200 IM relay and I scored last place on the breast. As hard as it is to admit this, Isaac Martinez, one of my best friends but also a long time rival, is officially better than me. I mean...my stroke looks so much better, but he is faster. I guess what I have going for me though, is I know how to pace myself and he either doesn\'t know how, or doesn\'t want to. On the 100 freestyle he only beat me by like a second though. My goggles came off when I dove in because there was not a seal on my face cuz it was dry, and they went down to my mouth. I was slower because I didn\'t flip-turn since my goggles were making it harder to breathe, and if I would have I could have beaten him because he doesn\'t pace himself.   I\'ve been neglecting the Christian Club at school lately but I have still been going occasionally. Last weekend I went to ATF, which is Acquire the Fire, a Christian Youth conference and I had like a weekend of hell. Most people had a lot of fun, but I guess I\'m hard to please or something like that. I had to share a room with this very ignorant guy and he made the whole room suffer with his ignorance. He was running up and down the hall like a savage and randomly knocking on people\'s doors with no regard to the fact it was late and they could have been sleeping and he kept the rest of us up with this stupidity. I would have liked to imflict physical pain to him...but then it was a Christian thing. And I\'m not an agressive person anyways. I just wish I could do that kinda stuff where in reality I\'m all mouth. (all bark no bite so to say) Anyways I couldn\'t sleep, ATF was really long, my CD got stolen, I got dirty water thrown on me, I got in the middle of a mosh-pit...and it just sucked. I felt like the people that were running the event from my church and then the ATF people couldn\'t have cared if I lived or died, and thats really sad when you think of your church like that because they are supposed to love and care about you like a lot.   Good thing I guess is I saw Ricky Valdez again but I didn\'t get to talk to him at all really. Just wassup, how ya doing, ok see ya around. Not that the conversation would have been facinatingly interesting anyways. He probably would have been preaching to me or something.   Lately I haven\'t really been wanting to go to church and my Mom makes me go...sucks. I wish I could just not go at all. I would have like more free time for other stuff. I mean, I can love God and be a Christian and all that jazz without going to waste two hours of my time. They should have drive up worship services where you get like a 10 second summary of the message and what it means to me, get hugged and take off. What difference does it make if you have to sit there for two hours?   Anyways I have a ton of HMWK, have swim practice at 9:00 am tomorrow and need to get started on the HMWK (I FORGOT MY BIOLOGY BOOK AGAIN!) Peace,   Would you like to comment on a post? If so  urlLink Click Here  '),(17322,2413978,'14,May,2004','2004-05-14 00:00:00','   Theres this one internet guy who I\'ve known since like forever. And I\'ve considered him a really good friend even though he\'s from the net, and vice-versa (or so I thought.)  We were like supposedly best friends and all this crap...and now he doesn\'t respond to my e-mails, respond to my text messages or IM me when he signs online.  Yesterday on my Cell phone I got a message \"DukeBsktball13 has logged in\" like way late at night, looking like he actually  sent me a message  after all this time. Its like \"Yay! he doesn\'t have some issue with me where he doesn\'t want to talk to me anymore\" but I guess I was totally wrong.  Now I\'ve given up on him and if he doesn\'t want to be my friend, then fine, I dont wanna be his friend either. So screw it.  Hope I\'m wrong, but I seriously doubt it.  '),(17323,2413978,'13,May,2004','2004-05-13 00:00:00','   Tonight at 4:00 pm I have to go to this allergy place for my pre-park study appointment. I have no idea what they\'re going to do, but oh well. Its a load of money.  My Mom is being totally evil and told me if I dont clean the floor that she wont take me on Saturday to the park. She can never ASK me to do something, she has to threaton me with something. I sometimes wish I had a different, more polite, kind and rational mother.  She loves me, but she\'s evil.  '),(17324,2413978,'13,May,2004','2004-05-13 00:00:00','   Today we ran the mile, like we usually do on Friday and I got a C! today...thats really bad. I was just kind of out of it today...wish I would have eaten breakfast or something.  My teacher said to start running on my own or something.  '),(17325,2413978,'13,May,2004','2004-05-13 00:00:00','   Last night I went to MOVE, which is a Dance Festival at Southwestern College with Linda Ramos. It was really cool.  They had some really good pieces, some ok pieces and two or three bad pieces. My favorite piece was this piece that was modern/hip hop dance to something by Missy Eliot and then my second favorite was this one that was kind of Arabic(ish) and had really great costumes.  The pieces I didn\'t like were a couple of really sleezy pieces where the dancers were dressed like whores (literally) and dancing to some music from Moulon Rouge, which was fitting to the costumes, but it was very untasteful and shouldn\'t have been in a dance concert.  Another thing I didn\'t like is between almost every piece they closed the curtain and turned the lights back on and had this long pause before getting started on the next piece.  My current dance teacher choreographed one of the pieces in the dance and today she was telling me that there were a lot of technical issues in her dance and she was really mad about it...but I didn\'t really notice them.  Linda enjoyed it and waited around with me after the show to see if my teacher was there...and I guess she escaped and took off.  All in all it was a really great experiance and I have some modern dance moves I want to use in my choreography project.  '),(17326,2413978,'13,May,2004','2004-05-13 00:00:00','   Dance today was really awesome. We learned a bunch of new choreographical technicques and just kinda chilled with the teacher who kept it fun today, not to mention teaching us a lot too.  Tonight at Southwestern College (The community college I will be taking Psychology and Jazz Dance at this summer) is having a Dance Festival and Mrs. Macias choreographed one of the pieces, so she offered a load of extra credit if we go tonight.  For going we\'ll get a free run, or an A on any existing run of our choice. So I\'m going to make up the run I\'m missing to guarentee myself an \"A\" in the class. And if I miss the make up run next Tues morning, no pressure because I will still have a backup free run.  And then yesterday I talked to Linda Ramos and she\'s gonna pick me up and go with me to the festival which is really awesome. She\'s an awesome person thats a youth leader at this church I used to go to. I plan on going back to the church eventually, but they kinda did me wrong, so I don\'t really have my heart set on it.  Going to have to stay after school today to catch up on some math I missed last week due to the Dentist and then around 4 I\'m going to go home, eat something really fast, do some chores, and then wait for Linda to pick me up.  ~Out  '),(17327,2413978,'12,May,2004','2004-05-12 00:00:00','   Today at the Christian Club we had 5 new people! It was really cool. Also one of the new people was really like into the whole thing. She participated in the discussion about pretty much every topic and opened up about some of the stuff she\'s dealing with.  Its probably a good thing that I forgot my actual lesson today because then people wouldn\'t have been opening up as much. I had this Mini-Lesson on Faith that went really well because people could relate to some of the questions and voice their opinions.  Also there were a lot of prayer requests given to us, and discussion on that helped a lot too.  At the end of the lunch period we prayed and then I caught the new girl and gave her my number in case she wants to talk or anything about her situation but I doubt she\'ll call me since she doesn\'t even know me. But then I guess its the Gesture that counts.  '),(17328,2413978,'12,May,2004','2004-05-12 00:00:00','   My Mom is like completely enjoying her grounding me from the computer until June. She won\'t even let me do HMWK on it. She is evil and something should be done about her.  Last night she made me wash the car by hand and she demands perfection so it took me like 1:30 hours where it could have gotten done in 5 minutes at the Car Wash for like $6.  Also this morning she rushed me out the door to school so I forgot all my Christian Club stuff and I had to come to school last minute and print out a lesson plan, which is totally not what it was supposed to be. Their online ones suck a lot...  Back to last night: My Mom\'s annoying friend Lora came over and just like spazzed out for 2 hours talking my Mom\'s ears off and she is a super-over-religious-fanatic so I love to get her started. I made a couple of subtle comments about Harry Potter and she went on a tangent! It was hilarious and fun. Then she went on a secant (like a tangent, but not) about Van Helsing and how the characters look horriically (I doubt she even knows what that word means) evil and stuff. Its like, \"Lord - its a movie. Fantasy-Fiction: That means NOOOT real!\" but she didn\'t get it.  She was like \"That movie is evil! You don\'t know what my Kid\'s are like when they watch horror movies! They have night mares.\" Van Helsing, to start is not a horror movie. Its Fantasy Fiction based on mystical horror personalities.  Then she went on to say her daughter was having nightmares and hollucionations over the Sixth Sense, of all movies. That movie was a Sci Fi thriller that  sooooooooo  wasn\'t even close to being scary *sigh*.  Off to Dance.  -- Current Music: \"Hey Ya\" by Outkast Current Mood: Pessimistic  '),(17329,2413978,'12,May,2004','2004-05-12 00:00:00','   Today my teacher in Multimedia assigned a new project. We have to Produce, Direct, Edit and find actors for a movie that is going to count as 80% of our grade. Should be interesting.  I\'m working with:  Adan Padilla  Ron Soucey  They\'re both fairly cool people, I just hope they do their part. Never worked with Ron and Adan seems to be absent the day of the main work on any project. It would be coincidence, but it\'s happened three times so far, so I doubt it.  But then lately I\'ve stopped trusting anyone becauase I\'ve been screwed over so many times. So it could just be my pathetic paranoia. Who knows?  ~Out  '),(17330,2413978,'11,May,2004','2004-05-11 00:00:00','   We now have a Quizznos in Bonita accross the street from my Dad\'s house! Today I went over there and got a Large Italian something sub and it was strange how much meat and cheese they put on the sub, but then I guess you get what you pay for, the sub did cost like $9 and it wasn\'t even a foot long. So like twice as much as Subway with 2 inches less food...but then I guess its piled down.  My Dad only gave me $10, so I couldn\'t afford Quizznos\' soup, so I made soup when I just got home now. Chicken Garlic and Wild Rice. Twas interesting. Definately much better then that strange Thai Food from last night.  Oh: just remembered. I didn\'t note what we  actually ordered  last night...so here goes.  Thai Iced Tea  Wierd Garlic Chicken Junk  Chicken Thai Fried Rice (dont know why Mom insists on Chicken in everything. Pork would have been much much better(er).  -Out  '),(17331,2413978,'11,May,2004','2004-05-11 00:00:00','   Today in 6th period which is my Multimedia class, I was an actor with a group of guys. Groups of two people are supposed to find a script and then get a group of actors to direct a play. I filmed and directed my own play on Monday and now I\'ve went with another group as an actor.  In the past Acting wasn\'t my best talent, but today I think I did pretty good. I mean, it wouldn\'t take much to do better then I did last time.  '),(17332,2413978,'11,May,2004','2004-05-11 00:00:00','   Man I\'m so sick of going to the Dentist. Yesterday they had to numb me  Again  with Novacane, and even asking the Dentist to take it easy, my mouth was still numb for like 4 hours. They had to glue another bracket onto an exposed tooth, and then apply pressure to try and get the tooth to come down.  After I was done at the Dentist my Mom had to get a bunch of stuff done and as it turns out she has to get more then $2000 worth the stuff done on herself, not to mention the $400 she has to pay for my two tooth exposures.  While she had her stuff done I was in the car reading and then napping.  After leaving the Dentist we went to the Air Port and bought some tickets for Phoenix Arizona. We\'re going sometime in June. I have no idea why as we know no one in Phoenix, but hey, it will be a weekend of relaxation that should be really fun.  Once we left the Air Port we went to this Thai food resturaunt in College Grove and it wasn\'t too great. I did try Thai Iced Tea that was interesting but all of the food was greesey and had a strange sweetish after taste. Other Thai food I\'ve had didn\'t have that. But the Tea, as I said, was good. It was regular iced tea with milk and sugar cain juice in it, which made it very unique.  Right before we went to the Thai place my Mom had to go in this damn Sorority store, because College Grove is by SDSU, hence the name, and she was looking around at all this shit she isn\'t going to possibly buy. She has to go in every fucking store!!! I\'m so sick of it.  And then after the Thai place she went to Vons, a grocery store, to buy **MORE** food. Like we need more! Our whole garage is practically a store. And then she has both freezers loaded down. Then she says I want food? I occasionally ask for stuff that **I** actually eat. GRR...parents.  We went to Cold Stone after Vons, I got an interesting Ice Crea with Pineapple, Coconut and Pecan Pralines mixed into French Vanilla (wish I would have gotten Sweet Cream instead) and then we got home **LATE**.  When we got home I still had to Choreograph a dance (32 counts) and then do HMWK, so I was up until 12:30 am.  Then in the middle of the night, around 5:00 am I got up really sick to my stomach. I felt like my Digestive Tract was going to explode, and then I couldn\'t even go to the bathroom. Constipation is the absolute worst feeling in the whole world. So I almost didn\'t go to school this morning I felt so bad. I blame it  ALL  on SODA. No more Soda for me. I\'m going to stick to Water, Iced Tea and Juice.  '),(17333,2413978,'11,May,2004','2004-05-11 00:00:00','   Today in Dance it was really cool. We were doing the Choreography project and Mrs. Macias just kinda threw us into Choreography. She put us in groups of two (I\'m a third in one group) and she told us to each make 16 counts of a dance and then we\'re going to put it all together and do it for the class tomorrow. We spent most of the period just experimenting around and don\'t actually have anything to present to the class as of now. So tomorrow it should be interesting just following the lead of each other. It will be just like follow the leader, alternate the leader and follow the next person. Or the blind leading the blind.  I want to do part of it with us on the floor, doing floor movements. Martha wants to do the beginning of the dance to a leap ending us up on the floor. And Brianna I have no idea what she wants to do. She never comes to school anyways, meaning she probably wont even be here tomorrow.  So yeah.  '),(17334,2413978,'10,May,2004','2004-05-10 00:00:00','   There are really ignorant fools in my 4th period class. They spend like the whole period screwing around, making stupid comments, interrupting the teacher, making rude sexual stuff, humping each other, etc.  I\'m super fed up of that class. And the teacher, having no control whatsoever over her class doesn\'t help the bad situation at all.  Today we were having a good class until the teacher gave the whole class Lolly Pops and they started making crude comments to each other. Then they started like getting against each other and humping for this perverted guy, Carlos, to take pictures of them with his phone. He probably goes home and masturabates to the pictures.  The only good part of the period is we had some discussions about crime and punishment and I learned some more about Interpol, which is a really fascinating organization.  Right now its lunch and I have no food, so I\'m really hungry. But then thats a good thing I guess, maybe I\'ll burn off my fat preserve.  After school today I have to go to the Dentist and get *more* surgery. I\'m so sick of getting my mouth cut open. It so sucks. But oh well...soon all the surgery will be over and then I will have all straight teeth. The only next step after that will be bleeching my teeth. Then they will be white on top of everything else.  One good thing I guess will be a weekend with little to no work. Even my stupid mother cant ruin that. I have to go to this park all day to feed my allergies and they\'re gonna test drugs on me, so I get to read for like 8-10 hours on Sat and Sun. Two full days of relaxation sound really good right now, even if I\'m being a Guinnea Pig.  '),(17335,2413978,'10,May,2004','2004-05-10 00:00:00','   Because of my Dental surgery today in 5th period I was totally behind in everything! I have to come after school on Wednesday to actually catch up on the work and then I have to go after school on Thursday to take the test its self.  Today in class I caught up on a worksheet while the rest of the class took a test, its really awesome the teacher cut me slack and didn\'t make me take the test today. So I read a lot of the period too, was fun.  Now the next thing we\'re doing in the class is going to be fairly easy, since its Special Polygons, which we\'re already covered. She just kicked them up a notch.  After school I  neeed  Tortillas! So I\'m gonna cook some from this one recipe that I got on Food Network. Should be fun. Lets hope I dont destroy the kitchen this time along.  I\'m in 6th period right now getting ready to shoot a practice movie. GRR.  '),(17336,2413978,'09,May,2004','2004-05-09 00:00:00','   Today 4th Period was completely terrible. Its usually terrible, but today it was even worse. The stupid immature people were like making hateful racist comments, hateful homo phobic comments and just being evil.  They interrupted the whole period and the teacher was like really frustrated too, but she never does  anything  about the stupid people and what they do. She lets them treat her like crap. I dont get it.  Dance was really cool today, we did choreography exercises and took notes and then I had to go to 4th period and it just ruined the entire mood.  '),(17337,2413978,'09,May,2004','2004-05-09 00:00:00','   My Mom is really starting to get on my nerves more and more. She is really evil. She\'s irrational and really seems to dislike me these days. Last night she was an all time evil. She was telling me that if I don\'t kiss her ass that she wont do anything for me.  We were at Cold Stone Cremery last night and she was like \"Enjoy this Ice Cream because this is the last thing I\'m gonna ever get you if your attitude doesn\'t change.\"  That means if you don\'t kiss my ass, then I don\'t do anything for you.  And who knows, this may even include clothes or something. I mean she is a really twisted vendictive person, so who knows.  I think once my Mom has no more power over me then we will be really good friends, but she doesn\'t seem to be able to handle power.  At this point, per advice of a friend, I\'m just thinking \"only three more years, only three more years...\" Can\'t wait until college to get away from her. I mean, I love her but I hate her too. She\'s my Mom, I couldn\'t not  love  her, but I can\'t stand her like 90% of the time.  Its sad she had to do all that stuff on mother\'s day of all days.  Also last night I had to stay up all late and do HMWK, it wasn\'t too fun. I was up till like 12:00 am and when I got up this morning, I was dragging.  I thought it was really cool on Survivor last night when the guy, Rob, proposed to the other girl on Survivor, the one who won. Only bad thing is his motives. Who knows if he really loved her, or just proposed so he got money either way. It would totally suck to marry someone for money.  Right now I\'m off to PE. Hopefully we\'ll continue to work on the Choreography project today. Good ol\' third period.  '),(17338,2413978,'09,May,2004','2004-05-09 00:00:00','   Today for mothers day my Mom was having a really hard time. She lost her mother like two years ago and has been just having a really hard time. Today she was really snappy and I thought she had something against me until I put two and two together.  I gave her a card and some lotion that I bought at K-Mart.  She was out in the front yard selling stuff today, gift baskets and such and we sat outside on the front lawn reading while my Mom just hung out there. I finally went inside of the house because there were icky bugs in the grass and becuase I was actually getting pretty cold.  Then I went in and took a nap and my Mom woke me up with a blaring stereo playing \"I\'m gonna get ya good\" by Shania Twain because she was trying to outplay the stupid stoner neighbors next door.  I\'ve made a new friend on the internet. I don\'t remember her first name, something Turkish, but she used to be an admin on Hoggies Moo and we kinda hit it off, which rocks cuz most of the people on there, excluding only a few, are evil cliquish, angry psychos that hate me.  I\'ve bene also having some interesting political debates today with people. Been fun. At my friends house right now. Not technically supposed to be on the internet at all, but what Mommy doesn\'t know cant hurt me.  Also the New Blogger really rocks! I love it soo much. Its a lot like Live Journal now and its free. Give it up for the awesome people from Google!  Out--  '),(17339,2413978,'09,May,2004','2004-05-09 00:00:00','   Busy Life lately. Iâ€™ve been doing Newspaper still, been doing competitive swimming, had several holidays, been on several vacations, had oral surgery, been at war with both parents and a lot more fun stuff.  Through swimming weâ€™ve had very few male swimmers because two quit, one transferred schools and two others werenâ€™t eligible (the un-eligible ones are our fastest and best guys) so that left three guys. We arenâ€™t bad swimmers, but weâ€™re definitely not the best either. Of those three guys, I was probably the worst of the three, maybe the second worst. But this one guy Iâ€™ve known since like 4th grade, Isaac Martinez, is much better then me, and he reminded me of that everyday of swimming (hence the season, at least for me is over.) My theory behind rubbing his speed into my face is that Iâ€™ve been better at him in pretty much every other area since like 4th grade. Smarter, got all the positions of power (Student government and club chairs), been more outgoing, etc. The only things heâ€™s better than me at are getting the girls, and swimming. But then my swimming stroke looks better than his, and Iâ€™m not into girls, if you get what I mean.  Swim meets have been interesting. Theyâ€™ve been swimming me in a large variety of events. My best events seem to be the 100 Butterfly, the 100 Breaststroke and the 200 Freestyle. I did moderately well at every event I swam, excluding the 500 freestyle. Although Iâ€™m really happy the year is over. It gets really old, really fast swimming back and forth across a pool.  Also, a lot of the people on the swim team had something against me. There is a large clique on the team that includes the majority of the girls and their humble leader didnâ€™t like me, so they didnâ€™t either. Plus blurting out my opinions didnâ€™t exactly help. Christina, Jessica, Isaac, David, Denise, Abigail and Alya were the only people I got along with really well. Without them I wouldnâ€™t have been able to bear this season, because the other people just ignored me. Jessica even at times didnâ€™t talk to me much because she was involved with the clique a little bit and understandably would rather have like 20 friends then just me. She was never mean like the cliquey people, but not super friendly either.  In school Iâ€™ve been doing a lot better in all my classes, excluding Multimedia because Iâ€™m rarely there due to swim meets and other things at school. Thatâ€™s the class thatâ€™s the least important to me, so I donâ€™t do near enough work in it. In Dance Iâ€™ve been doing much better and I canâ€™t stress how cool the teacher, Mrs. Macias is. She has become my favorite teacher, or second favorite teacher. She is awesome, she encourages us, and the other day when two girls didnâ€™t feel like running because of the heat, she started running with them. That gave me a tremendously larger amount of respect for her. And I already had a major amount of respect for her before that incident.  Lately, per advise of Miss Derwin, the Newspaper Advisor at my school, Iâ€™ve been getting involved in this organization known as book crossing. Its basically where you trade books with people through the mail, â€œforgetâ€ books places for people to pick them up, read them, recycle them and log them on the website, tracking them on their journey around the world. Iâ€™m #51 in California as of right now, out of 18,000 people in the state, which is quite an accomplishment. They measure your rank based on how many books you register and such.  During Spring Break I went to Avalon, Santa Catalina Island which was really fun. My Mom and I had a couple days of relaxation. I went kayaking, snorkeling and just hung out. I also probably gained weight from all the Ice Cream and awesome food I ate there. To get to Catalina, you take a boat from either Long Beach or Dana Point California for 30 miles to Catalina. Since we had a high-speed boat it only took around an hour, but on a regular speedboat it would take quite a long time. Perhaps 3-4 hours.  My GPA at school is still staying up. On the last report card, 12-week mark, I only had one B in Scholarship and only one B in Citizenship. So I had a 3.87 in Scholarship and Citizenship. I need to haul ass to get it to a 4.0 though.  Iâ€™ve been doing a lot of reading and been working on my website too. Also in my free time, which there is little of, Iâ€™ve been back on Hogwarts Moo but I have concluded half of the people on there are evil and deserve to have their whole bodies waxed, or other equally painful deeds.  Listening to more and more country music these days. Not say that Iâ€™m a country fan now, but Iâ€™m into Shania Twain and the Dixie Chicks now. Theyâ€™re pretty good.  This summer I got into Southwestern College. Through persistence and filling out petitions and having forms signed by my school guidance counselor, the Dean of Outreach accepted me. Only bad part about the whole SWC thing is that one of my classes, the Dance one with Macias, my current dance long term-sub starts on June 7th and my school doesnâ€™t end until June 18th, so it should be fun keeping good grades at both places.  Not much more to write, Iâ€™ll try to update more frequently.   Would you like to comment on a post? If so  urlLink Click Here  '),(17340,2413978,'22,May,2004','2004-05-22 00:00:00','   Just saw Troy and it was really interesting. And man if I can say so, Brad Pitt has a nice butt (just kidding) - although I\'m pretty shocked he was compeltely nude in the movie - that was different. Especially cuz he\'s married. I mean, he had to at least me somewhat aroused rubbing against a naked person...that would be hard (figuratively and literally).  Anyways, here is my review on Troy: Troy combined a lot of fine characteristics that a good movie needs. It had a love story, some moral lessons, history (I dont know if it was mythology or real history though) and it involved fighting.  Throughout the movie it showed Brad Pitt (Achilies) and his emotional battle through life. He tried to be tough but through several emotional losses throughout the movie, it showed he really does posses love and above all else, honor.  Orlando Bloom\'s (Paris) character was a little disappointing. He had very little honor through the movie (until the end) and just...wasn\'t the typical character that Orlando Bloom plays. Although Orlando got naked too, always a plus for the lustful girls in the audience.  All in all the movie was very entertaining and Brad Pitt will probably win an Oscar for his role as Achilies.  '),(17341,2413978,'21,May,2004','2004-05-21 00:00:00','   I\'m almost able to say that I hate or at least strongly dislike my mother. Everyday she pushes me away more and more, in every way possible - and whats really sad is she thinks we\'re really close.  I used to be able to talk to her about anything and she seemed like she cared and showed compassion toward me. Now I dont talk to her about anything because no matter how good or bad something in my life is, she blames it on \"the computer\" and then either grounds me from it long term or carrys on about it for a long time. She is either jealous because I like technology more then her (I wonder why?) or she is disillusional and having problems accepting that society is almost completely centered around technology.  She goes on and on about how I have it so much better then her. Her Mom was always out running around looking for men, sleeping around, etc. She was molested by her mother\'s many boy friends. So my Mom doesn\'t date. So she uses that to justify how she treats me.  She is also the cheapest person I\'ve ever met. She never wants to spend money unless its on food. I dont know if she thinks shes going to be buryed with her money or what, but she just hordes it up and she\'s pretty loaded.  Today she called me and almost immediatly she started nagging about something. She added two days to my \"sentence\" of no computer until June (that started because I said it was immature to say you hate something whole when you\'re only familiar with part of it) and ways...continuing from where I left off, she added two more days to my sentence for me not making my bed to \"her liking.\" And then she added two more days for me explaining to her what the movie \"The Village\" is about.  My biggest pet-peeves about how she acts is that her gods are the television and the phone. She swears shes never on the phone, but like...whenever I call shes like \"I\'ll call you back, I\'m on the phone right now\" so shes in denial about it...  Then at home whenever I go in the room where she is, shes like \"shh! I\'m watching tv! You always interrupt me when I\'m watching tv!\" so I ask her when I should talk to her since shes always watching tv or on the phone and I\'d be \"interrupting\" and she says \"during a commercial\". She records every channel from 8-10 every night (we dont have cable so thats only like 5 channels...imagine if we did have cable) so she fast forwards during commercials, so that gives me like 20 seconds tops.  Then when I\'m doing something, being on the computer, reading, relaxing in my bed listening to music, etc and she makes me come into her room and \"talk to her\" which always leads to getting me in some kind of trouble because I usually am not to happy about dropping what I was doing and having to \"talk\" to her, so I \"have a attitude\" or some shit like that.  Right now I\'m so fed up with her that I took the liberty of posting countdown timers in this blog and my personal blog (at the bottom and at the end of every post) of how long I have before I\'m out of here and at college where she will have no more control over me.  Maybe when I\'m an adult and I can tell her to go screw herself when she is being irrational we\'ll get along better. She wants me to be her friend one minute and the next she\'s my \"Mom\" and I\'m being disrespectful. She cant have it both ways. But for awhile now, since probably last December I\'ve been strongly disliking her and I wish she would just leave me alone. Given that if she did \"leave me alone\" I probably wouldn\'t be as good as I am, but still.  I haven\'t even had a childhood past the age of four or five because she has treated my like an adult my whole life and when I do something thats remotely \"teenage\" she gets all pissed and punishes me. My Dad says she\'s substituting me for her husband that she doesn\'t have anymore but where a husband is a consenting adult I\'m not and she can control me. Dont take that last comment wrong like we have an incestuous relationship or something *gag* thats just sick demented thinking. I sure hope my Dad isn\'t thinking that... cuz anymore I dont even hug her.  Another thing too...sometimes she plays around and like hits me or tickles me or something and then I cant do it back \"because I have fibermyalagia and it hurts me!\" but she can do it to me to her hearts content. What wrong with that? Then whenever she has a chance she uses that analogy \"if you cant take it dont dish it out\"!!  Right now I wish she would just...move away and leave me alone. I\'m already practically an \"old\" adult so...what do I need her for? My friends are always raging on me about how I need to act more teenage and stuff...but its hard. I\'ve seriously never been given a chance to just be a kid.  On the other hand I\'ve had a pretty spoiled life and stuff, but I want to be loved too. I dont think my Mom knows how to really love. All her anger that she just holds on to gets in the way. She is still holding on to anger from when she was a little kid and it shows occasionally when she explodes and starts screaming.  I look at my friends and how well they get along with their parents, how much they both love and like their parents. Adan totally clicks with both of his parents. Isaac and his parents like adore each other. Then this one youth leader guy that used to go to my church, Cornerstone and went away to college has literally the most perfect family.  The only comparison of friends to my family would be Lisa and Beth, and then they\'re like 200x worse off them me. But I\'m not gonna touch that today.  Heard of Goodbye Earl? Well Goodbye Mother!  --                                           	       Countdown to my Freedom (From my Mother, that is) (Days/Hours/Minutes/Seconds)  '),(17342,2413978,'21,May,2004','2004-05-21 00:00:00','   Dare You To Move by Switchfoot  Album	: A Beautiful Letdown,   Learning to Breathe Submitted by	: Laurel Corrected by	: GotGod Rated	: 9.8 (276 votes)   Welcome to the planet Welcome to existence Everyone\'s here Everyone\'s here Everybody\'s watching you now Everybody waits for you now What happens next? What happens next?  I dare you to move I dare you to move I dare you to lift yourself up off the floor I dare you to move I dare you to move Like today never happened Today never happened before  Welcome to the fallout Welcome to resistance The tension is here The tension is here Between who you are and who you could be Between how it is and how it should be  I dare you to move I dare you to move I dare you to lift yourself up off the floor I dare you to move I dare you to move Like today never happened Today never happened  Maybe redemption has stories to tell Maybe forgiveness is right where you fell Where can you run to escape from yourself? Where you gonna go? Where you gonna go? Salvation is here  I dare you to move I dare you to move I dare you to lift yourself up off the floor I dare you to move I dare you to move Like today never happened Today never happened Today never happened Today never happened before  '),(17343,2413978,'19,May,2004','2004-05-19 00:00:00','   Yesterday I got my allergy check for $440 and I dyed my hair and eyebrows (Auburn). It looks pretty cool but the dye also stained the skin under my eyebrows so they look super thick. I cant wait until that skin sheds off so the dye isn\'t there anymore because it looks pertty stupid.  Also last night Lisa came over and we picked \"Dare You to Move\" by Switchfoot for my choreography project and did a couple of ballet/modern moves, but none of them truly apply to the song. I\'m working on some lyrical moves, like //Dare you to get yourself up off of the floor// I can slowly get up off the floor and start the mid-level-change dancing (If this is Greek to you, comments on this and leave your e-mail and I\'ll give you a translation)  Today we ran bleechers at school (yuck) but I was able to get a \"B\" which I really need in PE right now.  Bad classes: Today in Biology we got a pop quiz and I totally failed it. We needed to know in what sequence blood flows through this strange picture of the heart and I wasn\'t at school the day the teacher taught that, so I died.  In 5th period it was super hard. Surface area of prisms. The teacher looked like she was ready to collapse too. Probably had an  exciting  night with her husband, if you get what I mean. Plus a long day at school and I was like \"Mrs. Hernandez! Mrs. Hernandez\" every 2 minutes because I was absent, hence lost, and needed help with my math homework.  In 6th Period we worked on the video editing and finally I have someone that understands some English. She is their VO (voice over person) so its getting better. During 6th period today I went to talk to Miss. Stratton, who is the Drama teacher, and I interrupted a rehersal and she got snappy. I went back after class though and she was much better. She was like \"since you\'re not in Drama you dont know this, but NEVER INTERRUPT A REHERSAL.\" So I learned something new today.  Went to La Bellas for dinner. Also went to the post office and spent like $20 between two Priority Mail domestic documents and one international.  Also talked to Christian from Cornerstone. He\'s a pretty cool guy and it was nice getting to talk to him again, although I still think Cornerstone went commercial and that the Pastor is now speaking more for the general TV audiance then the people who attend his church. But then I haven\'t been there since the Beginning of February, so I don\'t know if that is still accurate.  Anyways, I\'m out. Need to go burp up my Pastrammi Sandwich.  '),(17344,2413978,'19,May,2004','2004-05-19 00:00:00','   Last night they aired the allergy park study footage that they shot while I was in the park study last Saturday and Sunday. Unfortunatly they had to cut everyone\'s interviews to a few comments each and decided to not show my interview at all. All they showed was me blowing my nose and during a little teaser before the segment aired they showed me sneezing.  I called several friends and asked them to tell me what they thought and one of my aunts said \"You\'re a very good nose blower Michael. You cover your whole nose and catch it all.\" It was a very sarcastic complement which really lifted my spirits.  She was telling me that my cousins did a drum roll and then cheered exaggeratedly. So it was cool. My (literal) five seconds of fame payed off.  Tomorrow its back to school.  Today sitting at home I just read and chilled and did some sample choroegraphy to a bunch of different songs.  '),(17345,2413978,'17,May,2004','2004-05-17 00:00:00','   On Tuesday May 18th I was feeling really sick. In second period it was so bad I asked to go to the nurse and she said I have some kind of virus thats going around. So she called my Mom and my Mom said to let me walk home.  Well on the way home it was really hard to hold \"it\" and I barely made it to the bathroom. The virus really sucked and it felt like I was being continually punched in the stomach throughout the day. Hence, the pain sucked and it really hurt(ed).  Yesterday and Today I was out of school and kinda just hung out at home being lazy, reading etc. It made for two very boring days. And yesterday I couldn\'t even eat anything.  Now when I go back to school I\'ll have to play  more  catchup with all the classes I missed. I\'m still trying to catch up from the last time I was absent.  Also my Mom has been really evil lately. I dont know what her problem is but she has to get snappy and over-react and exaggerate   everything   and its getting really old really fast.  '),(17346,2413978,'17,May,2004','2004-05-17 00:00:00','   I just looked at my bank balance and noticed that I\'m $58 underdrawn. My Dad is probably going to kill me for it. Within the last few days I\'ve been buying stuff: blockbuster, 7-11, etc and I paid my Mom...so I probably drew my account a little low, and then a check for like $60 cleared for Amazon.com and it hit me today.  So I\'m hoping my Dad won\'t notice that balance that much...but knowing him and how nosy he is, he\'s bound to notice. I\'m crossing my fingers until Thursday when I get my check for $440...it should knock my account back to where it should be. Then I need to keep some money in backup incase this happens again.  I\'m also hoping he\'ll raise my allowence this weekend. We\'ll have to go over it and talk about it some more.  Gotta be off to math. Peace.  '),(17347,2413978,'17,May,2004','2004-05-17 00:00:00','   This weekend was definitely interesting. As I sit in Biology Class, supposed to be working on a Physiology Project, I write about it.   5:00 Am Saturday Morning My alarm clock goes off. I jump out of my skin because I set it very loud so I would wake up right away, and it was successful. I slowly walk to the bathroom, in my comatose state, and turn on the shower, trying not to collapse in fatigue. Once the shower water heats up, which takes about 4 minutes because it is at the end of the hot water line, I get into the shower and slowly wake up as the hot water hits my body. After I get out of the shower, I run to my room and turn on the radio to get dressed to.  Once Iâ€™m done I go into the kitchen where my Mom is getting our food for the day ready. We get in the car, go to Jack in the Box, get some breakfast sandwiches, go to 7-11 to get ice and some hot dogs and then we take off to the park, Morely Field in Balboa Park.   6:30 am We park at Morely Field and start to unload the car and setup in the dirt under some trees. We pitch out tent for the day and setup out chairs. Weâ€™re ready to relax. I go to table 6 to check in with my study advisors, along with 200 other study participants. They hand me a â€œWeekend in the Parkâ€ t-shirt, and a bottle of water and then tell me to come back to the table in 10 minutes, and to stay in the immediate area of the setup area.   8:00 am After this time we had already had about 4 meet ups during which we had to fill out diary cards depicting out allergy symptoms for the day. I thought it would be an easy day and an easy $440, but I was dead wrong. They sure made us work for that money. Next, at like 8:15 we had breakfast. They served juice, Danish, fruit, croissants and muffins. We had almost an hour to eat this time and then we went back for a bunch more diary cards to fill out and got drugged at 9:00.   Placebo Group On the second day my allergies were much worse and I have concluded that I was part of the Placebo group, which is sugar water and doesnâ€™t actually have any active ingredient in it, hence not preventing any allergies. I was dying and probably went through a grand total of 5 boxes of tissues. This worked in my favor to get me on TV though. Reporters from â€œWorld News Tonightâ€ with Peter Jennings was there recording and they got me on camera sneezing my head off, blowing my nose and they then asked me if I wanted to be interviewed on TV. I was all for the interview and now am going to be on TV either Tuesday or Wednesday, they werenâ€™t sure which day, but all the same I will air around the us for sure. My dÃ©but of fame! (Yeah right)  We packed up after the second day, Sunday, May 16th and then went to Old Town and went to Guadalajara CafÃ©, which was fairly good, then headed for home. Once I got home I had to complete like 4 hours of Biology HMWK.  All in all, even though we had frequent diary cards to fill out, I had a very relaxing day, got to finish my book and got a lot of sleep.  '),(17348,2413978,'28,May,2004','2004-05-28 00:00:00','   Awesome! I looks like we\'re going to Spirit West Coast which is this Christian thing that like...has some really awesome bands. Jeremy Camp, Sean Reuban, Grits, La Symphony, Chris Tomlin, Matt Redman, etc.  Its expensive  and Jeremy Camp is like the only good artist on Sat...but either way I hope we still go on Sat and Sun.  Out-  PS: Looks like we\'re going to Irvine instead of Laguna Beach - but maybe I\'ll get to go to Wild Rivers.  '),(17349,2413978,'27,May,2004','2004-05-27 00:00:00','   Those little \"UnCrustable\" peanut butter and jelly sandwich thingies are really really good.  The last couple of days I\'ve been going to Albertsons and buying them and yumm! Although I do not need those calorees...  Also speaking of Albertsons last night - I swear I saw the hottest girl I\'ve ever seen in my life there. I would have liked to find out how old she was, but then she could have been old and looked young and said something to the point of \"Little Boy, I\'m Married\" and then I would have like crawled under a couch or something and died. (People in question see Losing Grip for a better explaination of \"you know what\")  Out-  '),(17350,2413978,'26,May,2004','2004-05-26 00:00:00','   Today in English my teacher passed out these papers for us to like fill out and give to our teachers in appreciation. Well I got three and even though she is like one of my favorite teachers, I didn\'t get an apple for her because like that would have been wierd since I\'m gonna be right there...instead I got her a book yesterday, but I handed her an apple to distribute to other teachers and she got all happy thinking it was for her and then when she saw it wasn\'t she looked bummer, which really sucked, I hope I didn\'t offend her or something. I\'m going to have to send her an e-mail or something.  Yay! Today is a minimum day and I only have to be here until like 1:20 because of the Christian Club then I get to go bye bye and sleep or something.  Mom has decided that we\'re going to take an \"Adventure\" up the coast and just sleep wherever we can find. I think she has some fetish for beach camping, so God help us. I seriously hope we\'re not camping on the beach.  She says we\'re going up to Laguna Beach. Hopefully we\'ll be staying in a hotel there.  Also, upbeat/flipside notice, I will officially be ungrounded from the internet June 3rd. Its looking really close now! Yay!  Reminder to Self: Go to dumb Maraichi Dance thingie tonight to get an extra credit grade thingie. But then I hate maraichi, so it should be interesting.  And then next week I get to go to a cool dance show that will probably be mostly Jazz, but oh well, at least Jazz is usually up-beat.  Need to get back to class, announcements are almost over.  Out -  PS: In regards to Last Post, I can still Rant and Rave here! Yay! The AP just said to have a good attorney (joking) in case I mess up and people have an issue with it.  '),(17351,2413978,'26,May,2004','2004-05-26 00:00:00','   I saw an Assistant Principal yesterday posting in a Blog so I opened mine so he could see it and maybe comment so I could lead into the terms of it and such and I found out that they, unlike Castle Park Middle, don\'t really care if you post in a Blog because he said the school doesn\'t host it. He said just to be careful with names and information that could be considered Libelous. But I\'ve already been really careful with that, after the female Satan at the middle school tried to destroy me (exaggeration).  Umm - not much else to say, need to get home.  '),(17352,2413978,'25,May,2004','2004-05-25 00:00:00','   Today in 6th Period I\'ve like been working on this movie project for these other students and stuff and today GRRR! its taking forever. I have to render color into black and white which will probably take another half hour and has already taken another half hour. The group I\'m working with, if you can call them a group gave me totally irrelevant video feed so I cant really even work with it that well. *sigh*  So it looks like I\'m gonna be here until like 4:00 pm (and we got out at 2:00 today) to finish this damn project. what really sucks even more is that its like 80% of my grade for this semester...as well as a big chunk of this other classes\' grade and they just dont care I guess. Its unfortunate. I wish I had better stuff to work with though.  Also today in Math we took a test. It looks like I did it well, the teacher made it so the answers are even numbers, but I sill dont quite know what I scored on the test.  Rachel is still zoning out. She didn\'t even try the test today. She just sat there listening to my Cd player.  Back to school tonight to go to this Ballet Forklorico (Maraichi Dancing) thing that will help my dance grade a little more.  Back to my movie making *yawn*.  Out-  '),(17353,2413978,'25,May,2004','2004-05-25 00:00:00','   As you probably read earlier, I went to an allergy study and made $440. My Mom freaked out last night when I told her I was going to spend most of it on books and like lost it. She said she wasn\'t going to take me to any more studies and all this shit.  So this morning I ripped up the check I was sending to Amazon.com and transferred a bunch of money to my savings account and I guess I\'ll just buy the other books I want one by one or something. Its sad how much she freaked out though.  I dont know why I conformed to what she wanted me to do, and how to spend my money and all that. I mean - she shouldn\'t be doing that. She was telling me with books I\'d have nothing to show for my money, which so isn\'t true. I still would have quite a bit to show for it. Just on a smaller scale. I wouldn\'t have like a TV or something, but I\'d have a whole hell of a lotta books.  Anyways I have to be off to PE and get my Amazon.com Order straightened out.  Out-  PS: We\'re performing today in Dance. Should be interesting. Unfortunately my audiance is going to be small.  '),(17354,2413978,'25,May,2004','2004-05-25 00:00:00','   As you probably read earlier, I went to an allergy study and made $440. My Mom freaked out last night when I told her I was going to spend most of it on books and like lost it. She said she wasn\'t going to take me to any more studies and all this shit.  So this morning I ripped up the check I was sending to Amazon.com and transferred a bunch of money to my savings account and I guess I\'ll just buy the other books I want one by one or something. Its sad how much she freaked out though.  I dont know why I conformed to what she wanted me to do, and how to spend my money and all that. I mean - she shouldn\'t be doing that. She was telling me with books I\'d have nothing to show for my money, which so isn\'t true. I still would have quite a bit to show for it. Just on a smaller scale. I wouldn\'t have like a TV or something, but I\'d have a whole hell of a lotta books.  Anyways I have to be off to PE and get my Amazon.com Order straightened out.  Out-  PS: We\'re performing today in Dance. Should be interesting. Unfortunately my audiance is going to be small.  '),(17355,2413978,'25,May,2004','2004-05-25 00:00:00','   As you probably read earlier, I went to an allergy study and made $440. My Mom freaked out last night when I told her I was going to spend most of it on books and like lost it. She said she wasn\'t going to take me to any more studies and all this shit.  So this morning I ripped up the check I was sending to Amazon.com and transferred a bunch of money to my savings account and I guess I\'ll just buy the other books I want one by one or something. Its sad how much she freaked out though.  I dont know why I conformed to what she wanted me to do, and how to spend my money and all that. I mean - she shouldn\'t be doing that. She was telling me with books I\'d have nothing to show for my money, which so isn\'t true. I still would have quite a bit to show for it. Just on a smaller scale. I wouldn\'t have like a TV or something, but I\'d have a whole hell of a lotta books.  Anyways I have to be off to PE and get my Amazon.com Order straightened out.  Out-  PS: We\'re performing today in Dance. Should be interesting. Unfortunately my audiance is going to be small.  '),(17356,2413978,'25,May,2004','2004-05-25 00:00:00','   In Dance today I took my stereo to class so I could work on my dance better and stuff and it rocked. Several people borrowed my stereo and I was able to come up with a few consistant movements.  Right now I\'m in 4th Period messing around on the internet because most of the people are on a field trip so we\'re just like free today. Right now these girls next to me are looking for shirtless Brad Pitt pictures.  Anyways, the dance is going pretty well. I\'m hoping to show Mrs. Hernandez (Carmen) my dance or something if she is able to download the song from iTunes or something.  At lunch, in a couple of minutes, I have to go to the FCA room and like meet with the regional lady thats coming down to give us the play books and stuff so it should be interesting.  Then after school I have to stay all late to work on a video for this project in my computer class thats like 80% of my grade and I have to render it all into black and white and its going to take forever, but I guess a project couldn\'t be more structured and easy as this project is right now.  What sucks today is that because my Mom is on strike or whatever and I had to walk to school so I forgot to grab water and I spent all my money at Starbucks yesterday so I\'m like broke and cant buy and beverages. So I guess I\'m gonna go get like money from a teacher or something I can borrow.  PS: Partially Nude Brad Pitt takes up the whole screen on the computer next to me. Hernandez encouraged the students to put it on there, on top of merely allowing it to be there.  Out-  '),(17357,2413978,'24,May,2004','2004-05-24 00:00:00','   Today and I guess last night too my Mother started a war. She has been all over me about all this crap for awhile not and I guess now I have really set her off or something. She started refusing to give me rides to school (which I guess isn\'t too bad cuz its not that far and it will force me to /not/ be lazy - and gives her one less thing to threaton me with) and she said I cannot use the dining room table anymore because I had a book sitting on it and apairently that makes it look bad.  If she wants to be evil and be a bitch all the time, I guess I can play that game too. Example: hiding the TV remote, randomly turning off the answering machine and stuff like that. It will be subtle, but it will get her attention.  On the flipside, I was choreographing my Dance last night in the living room and my Mom seemed to like my moves and stuff, so I guess I\'m going to stick with some of the moves.  Also the Dance teacher gave us a choice of performing for this one special education PE class and them presenting back to us, or we running, so of course everyone, even the shy people picked the performance.  I have to be off to PE.  PS: I\'ve met some interesting people through this Blog. If you\'re reading this and not like someone I\'ve told to come here, be sure to leave me a comment, and include some means of replying to you.  Out-  '),(17358,2413978,'24,May,2004','2004-05-24 00:00:00','   Right now I feel really depressed and just out of it. It seems like the whole consists of stupid people. They\'re all over. They wont leave me alone. I know what people that hear voices feel like, expect I\'m sane. Stupid people everywhere messing with me, being annoying and being conformists.  I wish I had more friends, but I think I\'m scared of being stabbed in the back and hurt more. I\'ve been sad lately, I think out of loneliness - I\'m pretty anti-social, keep to myself and stay home, etc - which I dont like doing, but I dont know any other way. My Mom has kind of encouraged it over the years and its just kinda stuck with me.  So I dont really know what steps I should take to be happier, but I\'m gonna try it out.  '),(17359,2413978,'23,May,2004','2004-05-23 00:00:00','   I now have deviant art! If you wanna check out my Deviant Art page, click on the title of this post to be redirected.  '),(17360,2413978,'23,May,2004','2004-05-23 00:00:00','   Soo tired right now!  Last night I stayed up late talking to this guy, Brad, whose information I just randomly saw on Blogger and I asked for a picture just to see what he looks like, and my Dad happened to walk in to a perfectly clothed picture but somehow got the wrong idea and started screaming for me to go to bed because its late and he\'s worried to death trying to sleep in there.  You see - my Dad, along with my Mom have major issues. Neither of them know how to just leave me alone. They have to control my every move in life, yet they want me to act like I\'m an adult and never show any maturity. Its like, \"why control me if you want me to be all mature and shit?\"  Every night I\'m over here, which thank God isn\'t very often, my Dad like...gives me a bed time! Its like \"hello dude, I\'m 15 I dont need a bed time. At my primary living quarters (my Mother\'s) I haven\'t had a bed time since I was like 5 and it wasn\'t even enforced then.\" He finds it necessary to walk in the other room 4-5 times to like \"check on me\" or something. If he thinks I\'m looking at porn or some crap like that, I\'d rather just be told no internet after he goes to bed. Its really annoying to have him snooping around and like saying crap like \"when you gonna go to bed soooooooooooonnn\" (he says \"soooooooonnn\" instead of \'son\').  Anyways I need to go get ready for school *yawn*.  '),(17361,2413978,'23,May,2004','2004-05-23 00:00:00','   I just removed the password protection from this Blog. I seriously hope I dont regret it later.  Also just read the terms of use so I know where I stand legally.  In the past, the administration of my Middle School found my LiveJournal (similar to a blog) and tried to persecute me in many ways with it. On top of that they tried to scare me, theraton me and discredit me among the staff members of the school, in my opinion lead by Janet Bonney.  This time around I wont give anyone anything that would potentially be damaging.  If you find anything that offends you, please let me know and I can change it before you try to take any action.  Also if you have comments to me, about how maybe I should put the password protection back on, please let me know via commenting.  Thanks-  '),(17362,2413978,'23,May,2004','2004-05-23 00:00:00','   Pirates of the Caribbean!    urlLink   What movie Do you Belong in?(many different outcomes!)     brought to you by  urlLink Quizilla  '),(17363,2413978,'23,May,2004','2004-05-23 00:00:00','   Your Heart is Purple    urlLink   What Color is Your Heart?      brought to you by  urlLink Quizilla  '),(17364,2413978,'23,May,2004','2004-05-23 00:00:00','   Well looks like I\'m supposedly gonna marry Brad Pitt. I dont know about that. But I would take Jennifer Aniston.    You are going to marry Brad Pitt. He is always friendly to anybody he ever meets and he is very talented as an actor. He is also very sincere and friendly. He will respect you until the day he dies. Congrats!!    urlLink   Which celebrity are you going to marry? (now 12 (i just added more, and still more to come!)results that have pics!)     brought to you by  urlLink Quizilla  '),(17365,2413978,'23,May,2004','2004-05-23 00:00:00','   Love. You Truly Desire Love. You long for someone to hold you and take the pain away. You haven\'t been in much relationships or you need to work on how to handle them. You always seem lost in a daydream about the person you care about most.   PLEASE RATE     urlLink   What Do You Truly Desire? *PICS*     brought to you by  urlLink Quizilla  '),(17366,2413978,'23,May,2004','2004-05-23 00:00:00','   Apathy, well I can say your lucky, in some ways. You see Apathy is no emotion, basically you don\'t care. But that does not make you a bad person. Some of my friends are apathetic and I love them, but it wouldn\'t hurt to care a little more. Trust me life hurts, most people who are apathetic do it cause they were hurt. But don\'t worry, life is pain, its also pleasure. Good luck. (please vote)    urlLink   What Emotion Dominates you?      brought to you by  urlLink Quizilla  '),(17367,2413978,'23,May,2004','2004-05-23 00:00:00','   Reincarnation: You are nice enough to go to heaven, but Earth won\'t be as fun without you. So you shall come back as someone or something else. As a real optimist and lighthearted person, you always see the good in things. People probably respect you for your wonderful personality and love for life. People like you make the world a happier place (please rate my quiz)    urlLink   **Where will you go when you die?**(now with pics)     brought to you by  urlLink Quizilla  '),(17368,2413978,'23,May,2004','2004-05-23 00:00:00','   You are not in love but you are not not in love,you are trying to find things about the world and want to find that woderful one but you cant find what you are looking for GOOD LUCK and please rate my quiz     urlLink Are you in love?     brought to you by  urlLink Quizilla  '),(17369,2413978,'23,May,2004','2004-05-23 00:00:00','   You\'re addicted to.....  Nothing!  Your addicted to nothing at all? Well..... ok I guess thats a good thing but come on just think of the possibilities!    urlLink   What are you addicted to? (pics!)     brought to you by  urlLink Quizilla  '),(17370,2413978,'23,May,2004','2004-05-23 00:00:00','   kiss on the lips - you\'re sweet and simple but quite daring. you move for the kill confidently knowing the other person wants the same thing.    urlLink   What Sign of Affection Are You?     brought to you by  urlLink Quizilla  '),(17371,2413978,'23,May,2004','2004-05-23 00:00:00','   Youre gonna be attacked by a ridiculously small animal and bleed to death     urlLink   Choose your Dramatic Death (Now w/pics!!)     brought to you by  urlLink Quizilla  '),(17372,2413978,'23,May,2004','2004-05-23 00:00:00','   avoidant    urlLink   Which Personality Disorder Do You Have?     brought to you by  urlLink Quizilla  '),(17373,2413978,'23,May,2004','2004-05-23 00:00:00','   I dont think this quite fits...but maybe it does?    Demi-Cerberus: A demi Cerberus is very similar to Cerberus in appearence but the difference is that a Demi-Cerberus only has two heads.  You are cruel and inhumane. You lack the food intentions and sense of being that Cerberus has.  You like to play with peoples emotions and torture them to the last.  You have few friends because many are scared of you and you are also prone to violence.    urlLink   What Mythological Creature Are You (Many Results and Beautiful Pics)     brought to you by  urlLink Quizilla  '),(17374,2413978,'23,May,2004','2004-05-23 00:00:00','   you are the cute but psycho happy bunny. You adorable, but a little out there. It\'s alright, you might not have it all, but there are worse    urlLink   which happy bunny are you?     brought to you by  urlLink Quizilla  '),(17375,2413978,'23,May,2004','2004-05-23 00:00:00','   My inner child is sixteen years old!    Life\'s not fair! It\'s never been fair, but while adults might just accept that, I know something\'s gotta change. And it\'s gonna change, just as soon as I become an adult and get some power of my own.    urlLink   How Old is Your Inner Child?     brought to you by  urlLink Quizilla  '),(17376,2413978,'23,May,2004','2004-05-23 00:00:00','   I should listen to this CD more. They\'re pretty good. All acoustic and just singing, and the one guy person has a nice voice.  Listening to \"A Mark, A Mission, A Band, A Scar\" by Dashboard Confessional (I think its their most recent CD, or at least the most recent one BMG has)  '),(17377,2413978,'23,May,2004','2004-05-23 00:00:00','   Ok - today has been trying of my patience. I supposedly have an \"attitude\" and am \"talking down to my inferiors\".  My Dad brings home a laptop that he uses to dial into the building computer to like control all the stuff since he\'s an engineer.  Well anyways he has major issues with scroll bars that he swears \"aren\'t ever there\" and he like harasses me into fixing something that isn\'t a problem. He also has issues with those little buttons in the right hand side (left side of mac) pc programs that open and close the programs. He was like freaking out over nothing.  Then he was saying some toolbar that he needs to control the program or something wasn\'t there, so when I finally got it open I told him it was on authohide and thats why he couldn\'t see it, and he was like \"yeah I know, its supposed to be there\". What was the point of all that?  Oh well...life sucks then you die.  Then he went on to tell me through this lecture that was really...pointless and I tuned it out and just nodded at the right points. But I was killing his \"integrity\" (not sure he knows what that word means) by talking down to him or some shit like that *sigh*.  '),(17378,2413978,'05,June,2004','2004-06-05 00:00:00','   This is just a test to see how this thing works out.  urlLink  '),(17379,2413978,'04,June,2004','2004-06-04 00:00:00','   Did the filming for our movie today from like 8-10:30 at school and it went pretty well I guess. We got compeltely done with a bunch of different poses and shots in about two hours. It would have been faster if everyone would have showed up on time though. I had to call and wake Ron up, he had apairently went to a concert the night before, Shavelle was a little late but I\'m just happy she showed up because I didn\'t get her number and we wouldn\'t have been able to do it without her. Then Adan, of course, came, although I had to call and wake his ass up too. He showed me his arm, his chest and his leg where poison oak hit it...and ick. It was really gross. And he put some kind of lotion on it or something for the itch and it made it look like some of the boils had burst and there was puss or something...it was nasty - and now I really feel bad calling him a whimp and everything last night on AIM because wow...thats like...bad. I would be dying, figuratively, if I had all that on my arm. He must have had a serious allergic reaction or something.  During the Saturday School people\'s 10 minute break between session 1 and session 2 at 10:00 we had people all around so we had to pause for awhile before resuming filming and this one guy, who I cant say I\'m surprised had Saturday school as he is a nuisance in my 4th period class - wanted to moon the camera and I think I would have honestly been scared for life had I seen his ass... bad thought. But then since he\'s a jock, it mighta been good looking too, you never know with jocks these days.  Shavelle walked most of the way home with me and then departed to wait for her mother.  After we left the school, that is my Dad picked me up at my mother\'s house (LOTR3 came!) my Dad picked me up, we drove downtown to the building where he works, I sat in the car awhile reviewing my tape, he came out, drove clear back home to get his glasses or pager or something and then we drove back to Mission Valley and saw PoA (Harry Potter 3) which I think was really good, although I\'m a little disappointed that they didn\'t show more of Harry\'s classes, especially developing his classes with Lupin. They showed how Harry and Lupin were really close and stuff, but they didn\'t show how that relationship between them developed. Also the ending was a little spuratic...but the movie had a lot funnier stuff then the last two, which was cool. Also it was a lot darker, more for adults and teens then kids like the last two. IT wasn\'t a happy world of magic doing and stuff.  What was really wierd is that the theatre had more adults then it had kids and there was this guy sitting next to me that was a space nerd...he was talking about how he wanted to make a car that had all these things you\'re likely to hear about on Star Trek or something...I was trying not to crack up.  After that we went to DZ Akins and I had Mish Mash Soup and a Corned Beef Reuban. My french fries were cold though...yuck, I had to ask for new ones. Then on the way home my Dad was telling me about how truly evil my Mom is. If that stuff he said is true, and I know at least some of it is, shes a lot worse inside then she is externally with me...thats kinda a scary thought. Is she going to get worse with her stupid rules and punishments then she already is?  Her last punishment, thou shalt not use thy washer and thy dryer was stupid. Thats like saying I cant use the shower at home anymore, I have to go in the backyard. Is the no more shower going to be her next rule?  Right now I have some HMWK to do and I\'m putting it off...so eventually I\'m gonna have to get to it. I need to:  Write Out Story Boards for the Movie I Filmed  Do my Dance Critique for \"Last Chance to Dance\"  and I need to update my Dance Journal thing  Other Stuff: Update Check Register  E-mail Mr.P about not turning in the camera on time on Monday   Excited about starting SWC on Monday. I dont exactly know what to expect but it should be fun nonetheless. I dont think Mrs. Macias will disappoint me. Her last class was a lot of fun. The only thing I\'m worried about right now is that the permenant Dance teacher at my school is going to be super difficult on her grading like she was last semester and kill my grade, figuratively, on the final exam. Also I wonder if shes going to apply all my extra credit and stuff to my grade that Mrs Macias said that we needed.  Anyways, need to get to it. Out-  PS: Talked to Beth. Apairently her mother is suicidal since they took away her cigarettes or something like that. That family has sooo much drama.  '),(17380,2413978,'04,June,2004','2004-06-04 00:00:00','   Tonight I\'m kinda in a bad mood. I have to get up early and I just feel...exhausted and overwhelmed with life. My Mom is always on me and even thought we\'ve made up, this time, I still feel like as soon as I get back home she\'s going to start with something else.  I\'ve just kinda been sitting around here, bored. Doing nothing for most of the day and I have to get up all early because of these people at my school who procrastinate. We have to film a movie at school tomorrow and we have to be there when the administration people are there, so I have to get up at like 7:00 to get there by 8:00 am, so it should be really fun.  Then I\'m sitting here right now talking to this guy thats in my group totally dying over Poison Oak. I could be wrong, he could have had a serious nervous reaction over it, but I think he\'s making a big deal over nothing to keep out of school...he did this the last time we had finals except last time it was the Flu, which I think he exaggerated too...grr. I really dont know why I care, its not my business really what he does and what his grades are along with this other guy in my english class that I nag, but I like both of those guys, they\'re cool...and seeing people like slack off and stuff makes me get on them...but yeah. Gonna shut up now. I should go to bed since I have to get up all early.  Out.  '),(17381,2413978,'03,June,2004','2004-06-03 00:00:00','   Well I talked to my Mom today and I guess we\'ve made up now? It sounds like she\'s gonna let me use her washer and dryer although she didn\'t actually  say  that to me...  I went to the post office today to mail out some books for bookcrossing and when I was in the post office (where I was overcharged, I might add) and I saw Jenny, this Journalism instructor at SWC and then I saw her at the library where I went to get my Dad because the library is right by the post office...and she apairently lives in the same apartments as my Dad and was telling me about journalism camp this summer, as well as inviting me to journalism camp.  Right now my Dad wants to go to  Sizzler  ick ick ick! I despise sizzler but I think I talked him into going to Cafe la Maz...which is much better then Sizzler and still moderately cheap.  Well - I\'m out. Just update on the situation with my evil mother. Excited to go to SWC for my first class. We shall see how it goes.  '),(17382,2413978,'03,June,2004','2004-06-03 00:00:00','   The drama with my Mom at home continues. Last Night at about 11 pm she told me to get out of bed and go fold clothes because she wanted to do laundry and my clothes were in there so I said \"I\'ll do it in the morning.\" She insisted that I do it and I finally was like \"Mom, why cant I just do it in the morning\" (And I seriously was going to do it in the morning) so there was silence for like...10 minutes or so and then she appeared at my bedroom door and started up with her screaming.  I\'m no good, I never do anything she asks me, I\'ll never amount to anything, she\'s ashamed to call me her child and other stuff that I can\'t rember right now. Anyways, to make a long story short she said that I cant do my laundry at home anymore because I didn\'t want to get out of bed and go do it right when she said and she pointed out that I do this all the time, which is so not true.  Yesterday I had told my Dance teacher about all the drama with my Mom and today she asked me how its going, so I told her about all this and she was like \"whoa.\" But she also said \"Hey, I didn\'t pick my parents either\" so I guess she can relate to me in other ways. I still dont know if my Mom is going to pay for dance in the fall. I seriously doubt she will because she is the cheapest person on the face of the earth, but I really hope she will. She only actually has to pay  half  of $65 plus transportation...and that really isn\'t much for college classes. What would she do if I went to UCSD or something and it was outrageous.  But we had to run in PE today and I took out all my anger with my mother when I ran so I feel a lot better right now and I got a B for my grade which didn\'t hurt.  I asked my Dance teacher if she could still give my Mom a call to tell her I\'m a good dancer with potential and stuff and she said that she would but that on Saturday she is sleeping all day long and on Sunday its her anniversary, so I\'m not gonna push it condsidering that.  But someone needs this computer so I\'m out.  PS: Saw Michael Rodriguez yesterday and I had to mow the lawn and Mom said if I dont mow the lawn she was going to ground me for two weeks. She can never just...say \"Make sure you mow the lawn\" she always has to give some threat. But yeah. Bye.  '),(17383,2413978,'03,June,2004','2004-06-03 00:00:00','   I went to the dance show last night at school, \"Last Chance to Dance\" and it was really cool and entertaining. Probably the best dance show I\'ve seen anywhere, as of now. My favorites were \"Random Eye Movement\" by Lorena Macias and Glorious (Tree63) by Lupita (since she goes to my school and shes a minor I cant put a last name...)\'  I want to go back tonight, but I dont see that happening since my Mom and I have been at war literally.  Last Night I had to clear literally  everything  out of my room. Everything off of my dressor this morning, everything off my bed and all thats left in my whole room is a neatly made bed and a dressor thats empty on top...so its going to be really fun putting it all back in there. And my Mom said that shes going to  help  me put everything back in, which means she\'s really just gonna stand there in my way and supervise...which will probably lead to a fight and I\'ll end up  getting grounded longer.  Anyways, I\'m in class right now and need to get back to working on my Portfolio.  Out-  '),(17384,2413978,'02,June,2004','2004-06-02 00:00:00','   Yesterday I have to completely clear out my room so it can be painted today. It was terrible and I went to bed all early because I was so tired from it. I had to completely clear out my closet and I piled all my clothes and boxes of stuff in the guest room, took some of the movable furnature into the living room and then I moved my bed, dressed, computer desk that lacks a computer and my night stand into the middle of the room and made this little area for the painter guy to get around in. And then I also had to clean the walls, remove some adhesive starts from one wall and clean all the closet shelves.  It was quite an ordeal...but in the end I guess I\'ll have a cool painted room.  We went to Home Depot and got Licorish Stick Red and then white. One wall is going to be bright red and the other ones will be white.  But right now Dad needs the phone line and I need to be off to my Dance thing. So,  Out-  '),(17385,2413978,'01,June,2004','2004-06-01 00:00:00','   My Dance piece was ok I guess. I think I was 3rd of 4th best which isn\'t saying much since most of the people aren\'t even half way there. But my teacher gave some good advice. She said she liked the on the floor resistance part and I should make my theme \"resisting gravity\" and she also liked this turn thing I did along with a channe balance that collapses. So I think I\'m going to go with more collapsing. I\'m gonna work some more and get some advice from Linda.  Out-  PS: I want Chili Fries!!!! Really Really Really BAD! Hopefully when Dad picks me up we can go there!  And then I have to go to this other dance presentation tonight. Should be fun, and much better then the Mexican dancing accompying maraichi which I have passionate loathing for...  '),(17386,2413978,'01,June,2004','2004-06-01 00:00:00','   Been thinking lately about my whole college thing coming up. Next week starting Monday I begin my dance class at SWC and then on the 21st I will begin my Psychology class which will then be a whole college load. Should be fun.  But I\'m going to have to miss two days of school, or at least a second partial day...so thats going to suck especially in Multimedia because we\'re doing this final movie project thing and its like important and a group project and two days of missing 6th period will be bad.  One good thing though is that Mrs. Macias, the dance professor said that she is also teaching a night time Jazz Dance class so I would easily be able to make up my absences at night, which rocks.  Right now I need to find out which Psychology Book and workbook I need to get going.  Right now, as a I write this, I sit in my 6th period classroom, after hours with people on the same level as the \"space nerds\" from  Dude where\'s my Car . They\'re all sitting around talking about very wierd things, trying to sound smarter then the others, while playing MTG...I feel like a nature observer in wild territory or something...this is really strange. One guy seems to be normal along with the teacher. The rest though - who knows.  Also talked to my awesome math teacher and I have an A- in her class, but I\'m barely holding the grade...so I need to tread softly...  Well - I\'m out.  '),(17387,2413978,'31,May,2004','2004-05-31 00:00:00','   Tomorrow we have to do a dance presentation to the class and the teacher of what we have so far on the choreography project. It should be interesting. I\'m having Linda come over tonight to help me get it ironed out but I haven\'t heard from her yet so I\'m not completely positive that she will be coming. I can only hope.  Also just discovered Upoc.com and its really awesome.  '),(17388,2413978,'31,May,2004','2004-05-31 00:00:00','   Last night we were at K-Mart and I was asking my Mom for some shirts and she was losing it telling me that I take care of nothing and therefore deserve nothing. Basically that means I dont kiss her ass so I\'m evil and not worthy.  She even made a scene screaming that she\'s only buying two packs of underwear and one pack of undershirts, which I\'ve needed awhile. But at least I guess I got something out of her. I had to buy these two shirts and a third pack of underwear myself.  After K-Mart, she rushed over to Starbucks on Broadway and H Street to prep some guy on Avon selling and like get him going at Mesa college. I really feel sorry for the guy because he probably thinks shes all nice and professional and the first time she messes up she\'s gonna be all over him because since he is younger she feels that he is inferior so thus her minion.  At Starbucks she offered to buy me a Tea so I asked if I could apply the $2 a tea would cost into a Frapuchino and she freaked out, accused me of saying a Tea was not good enough and then bought me nothing, which I guess was fair enough, although that was  not  what I said to her...  Anyways, done ranting and raving.  '),(17389,2413978,'31,May,2004','2004-05-31 00:00:00','   My Mom had this brilliant idea that we were going to go to Laguna Hills California on this \"awesome, fun\" vacation experiance for Memorial Day and I would have to say that I was miserable. The whole weekend she insulted me and put me down in every way possible trying to provoke me so she could ground me more... at least that was my perception of it.  She drove all the way to Long Beach on PCM (Pacific Coast Highway) which I guess was pretty at times but she like...was a bitch all the way up there, screaming, nagging and all this other shit that got really old really fast.  She went all over the whole city of Long Beach looking for this Krispy Kream, as if she needs doughnuts. I\'m not saying that like I\'m all skinny, but I know I dont  need  more carbs or fat...but she like...doesn\'t get that yet and then says she wants to lots weight...grr!  The first night there, after she screamed at me the whole way there because she got bad directions from Map Quest and it was my fault, as everything else in life is, we checked in and then went to this Resturaunt called Monteray Bay Canners, which has some romantic significance to her because her and my father used to go there in Hawaii.  The resturaunt totally sucked, was very over priced the waitress was terrible. I got some Shrimp that cost her like a mint and some other stuff that was just whatever. I would have rather went to Mimi\'s Cafe or something. Someplace I already  know  is good, but Noooo!  Then the next day I slept in and she was just as bitchy. Thats when we took that coast drive thing. Well after that when I got on the phone with my Dad she started harassing me so I got off the phone then she hit me or something and I called her a bitch, which is is by the way, figuratively, and then she grounded me more! When will I learn? She is a matriarch, there is no freedom from here, guilty until proven innocent, etc. Sometimes I wonder if she even knows anything about America.  We drove home on Sunday, she hated me even more and let me know it the whole way home and she was like on the phone the whole way home ignoring me and then complaining about my music when she wasn\'t. I put in all these different CDs trying to find one that she likes. Ones that a month ago she likes now she doesn\'t because she just wants to disagree with me about something. She had to drive on this costal highway that was totally congested on the way home too, just to make me miserable and make it take longer.  And then when we got to San Onofre she had to pull over onto the state beacuse which is like...nuclear waste and stuff. I think she wants to kill all my little fishies (sperm) by subjecting me to all that crap (I\'m serious - its a nuclear waste area - but a nuclear reactor.) We drove around for like 30 minutes so she could find the freeway (I-5) while she kept on commenting about how \"this is really creepy\" and how she doesn\'t want to be here. It\'s like, hello dude, you\'re the one who freakin\' pulled over here!!  One good thing is on the way home we went to this one good Jewis Delicatessan place in Encinitas which totally rocked. I got a Corned Beef Reuban, Mish Mash Soup and ate tons of the pickles on the table which were totally awesome pickles. I\'m a pickle person...  Anyways on the way home she went on and on about how ashamed of me she is, how bad of a child I am because I don\'t kiss her ass and on and on and on. She said from now on I\'m to go to straight to my room when I get home at night because I\'m so bad and stuff. Then she called my aunt and put on this fake accent to pretend like she actually cares about my aunt. I think she was just bored.  Then she told me because I\'m an evil sinner we\'re going to church, which actually turned out to be somewhat amusing. I wasn\'t bored out of my mind for a change and actually got something out of it. Whats really sad, though, is that I think the only reason she even went to church was because she had some Avon business with this one lady. She thinks this lady is inferior or something to her so she like was very rude to the lady and like was in her face, right at church. I guess she wasn\'t kidding when she said shes the same at church as she is anywhere. \"She doesn\'t pretend\". I guess thats good. People can see she really  is  a filty bitch.  Well I\'m done ranting and raving now. Got it all out.  Today I was doing a lot of reading, sanded my wall, and did laundry. At Dad\'s now. He leaves alone and seems to at least like me. Although I\'m not particularly fond of either of my parents, I would have to say at the moment I like my Dad more. But I have more loyalty to my Mom, even though I sit here and slam her. Its wierd how it works out like that.  Well I\'m out-  '),(17390,2413978,'12,June,2004','2004-06-12 00:00:00','   This is Little Mary Ellen, my best friend Jackie\'s little sister.  urlLink  '),(17391,2413978,'10,June,2004','2004-06-10 00:00:00','   Heres me solo, at the prom. *turns on the \'I\'m too Sexy for my Clothes song\'*  urlLink  '),(17392,2413978,'10,June,2004','2004-06-10 00:00:00','   Me at night in my comfy clothes that I sleep in. Note the uncombed hair.  urlLink  '),(17393,2413978,'10,June,2004','2004-06-10 00:00:00','   My CD collection. I condence it all to little cases. The one on the left side is the big one with origionals, then the next one to the right is DVDs, then the rest of them are duplicate burnt copies sorted by genre.  urlLink  '),(17394,2413978,'10,June,2004','2004-06-10 00:00:00','   This was a picture of me and Jackie at my church\'s formal prom this year. This was our goofy one. I\'ll put the formal cuddly one up later.  urlLink  '),(17395,2413978,'10,June,2004','2004-06-10 00:00:00','   Last night I went to a church formal thing at this place called the Catelina Room in Marina Village over in West Mission Bay Drive.  I got my Tux (pictures to come - probably later tonight) and then I drove out there and got there late. But everyone else came even later - the place was really hard to find but I guess we got there after all. I hung out with Jackie, one of my best friends and we took some formal looking pictures and some goofy pictures and then we had our professional pictures taken, sat down and ate some fairly good Italian food. They had Chicken Parmasagne, some pasta, salad, parmasagne garlic bread and lasagne (ick) which was actually pretty good for Lasagne...cuz I dont really like Lasagne.  Well after the eating the youth people did a little skit making fun of the graduating seniors and then we had this dancing thing which was pretty fun but I couldn\'t really get into any of the music so it wasn\'t really  that  fun.  Then some chick grabbed me during the Salsa and just kinda steared me around, I had no idea what I was doing and she went to spin me around so I turned with her thinking thats what I was supposed to and we got all tangled up and she gave me a dirty look and then we went on, the spun me off and grabbed some other guy that was as bad if not worse then me at Salsa.  Hip Hop I can do, Jazz I can do, Modern I can do, and even some Ballet I can do, but Salsa is something I cant do...its almost all freestyle based so that makes it really hard.  After that they had this one band come that I guess was like Gospel / R&B that I just couldn\'t get into. Then they had this mass alter call and turned it into a Christian Factory, as I like to call it, some kind of wierd revival thing. Jackie went storming out with some of the band people following her and like, I dont know what happened, but I think she was just being dramatic.  Right now I\'m at MAry Ellen\'s house on her computer so little Mary Ellen could go see her or something cuz Grandma Mary Ellen and little Mary Ellen (5th grader) haven\'t seen each other in like forever due to family issues they have with each other and my Mom is like the mediator or something...  Anyways, gotta go. Need to go finish putting my house put back together and fixing up my room.  '),(17396,2413978,'10,June,2004','2004-06-10 00:00:00','   Last night my stuff from Amazon.com came so I opened my Pilates and tried it along with the Yoga-latties (which is a mix of pilates and yoga) and tried it and my abs were just too sore to keep it up. It was also somewhat strange that everything about both was centered on breathing...didn\'t expect that.  I also have my Charlie Bone books and my \"Black\" by Ted Dekker book.  '),(17397,2413978,'09,June,2004','2004-06-09 00:00:00','   Dance yesterday kicked my butt! The teacher increased our crunches from 80 to 140 and wow...I was still sore from the day before and now am sore on top of sore...I dont even want to think about how much she\'s going to increase the crunches today...  After school yesterday I went to the Tux place and tried on my Tux which I guess fits well? The ladies were busy and didn\'t really care or look to see if it fits...but its their tux thats gonna be messed up if the pant legs drag on the floor... After that my Dad and I went to this Filapino place and they didn\'t really have anything fresh, so he didn\'t want anything, so I just got two chicken empenadas (yumness!!!) and then we went to Weineschitlez (dont know if thats spelled right) and I got like 4 chili dogs and chili chesse fries, which was a little heavy for dance and I almost got a cramp from it.  Then we went to SWC and I went in the book store with my Dad and showed him the book that I\'ll need for Psychology and asked for this one school t-shirt and he bought it for me. After that we went to the huge library and I talked to one of the circulation guys who prepped me on the usage of all the stuff there such as computers and check outs and such.  After that I went to my dance class where I practically got killed between the dances and all the warm up...but at least I wasn\'t the only one sore.  I dont feel like going to 3rd Period today so I\'m going to get out again.   My Existing Situation:  \"Active, but feels that insufficient progress or reward is being made for the effort exerted.\"   My Stress Source:  \"Has an unsatisfied need to ally himself with others whose standards are as high as his own, and to stand out from the herd. This desire for preeminence isolates him and inhibits his readiness to give himself freely. While he wants to surrender and let himself go, he regards this as a weakness which must be resisted. This self-restraint, he feels, will lift him above the rank and file and ensure recognition as a unique and distinctive personality.\"   My Restrained Characteristics:  \"Feels that he is receiving less than his share and that there is no one on who he can rely for sympathy and understanding. Pent-up emotions make him quick to take offense, but he realizes that he has to make the best of things as they are.  Exacting in his emotional demands, especially during moments of intimacy leaving him frustrated in his desire for a perfect union.  Insists that his hopes and ideas are realistic, but need reassurance and encouragement. Egocentric and therefore quick to take offense.\"   My Desired Objective:  \"Feels he has been unjustly and undeservedly treated and betrayed in his hopes. Disgruntled and in revolt against his existing circumstance which he considers an affront.\"   My Actual Problem:  \"The unsatisfied desire to be respected, to stand out from amongst his friends, is causing some anxiety. As a result, normal gregariousness is suppressed and he refuses to allow himself to become involved, or to participate with others in their ordinary activities.\"   those random results and observations by ColorQuiz.com  '),(17398,2413978,'09,June,2004','2004-06-09 00:00:00','   Yesterday was quite Chaotic after school and at school (more about that in mikerbrant.com/personal). But anyways, after school my Mom picked me up and I went straight home, ate and then went straight to SWC to my class. We got there a little early so I walked around a little and then went to the Book Store to buy a bus pass and look at what kind of Psychology Book and such that I\'ll need for this summer within a few weeks.  I was almost late to class, but made it on time. We started out with all the stretching and warming up technique and then went on to the Jazz Dance and later we did the Hip Hop Dance that Mrs. Macias choreographed here at my school to \"Dirt off your Shoulder\" by Jay-Z and I really like the song and the dance (to watch) but actually doing it is really hard because its a super super fast dance and like...wierd - so keeping all the movements straight is really hard...  Anyways now I need to be off to class and then get ready to do this crazy routine again tonight.  Out-  PS: Also last night I was in bed at like 9:00 pm because I was all tired and them my Mom\'s dumb friend showed up with Jackie and Mary Ellen (her daughters) and Jackie\'s like my best friend so I woulda been fine in bed just in my underwear with her cuz she doesn\'t really care and I\'ve known her forever, but she brought this one guy from church with her Arcey or something like that (seems like a decent guy) so I had to get up and get dressed. And my Mom was in the bath when they came over so I had to go open the door and stuff and then later she came out of the bathroom and was walking around in a towel with this guy there... I wanted to die. But oh well, I guess everything was covered, that was just wierd...  But yeah - bell rang - really need to go.  '),(17399,2413978,'07,June,2004','2004-06-07 00:00:00','   Wow...today was pretty hard. We did pivots and jazz chasses with this sweeping leap thing that was...strange and took me awhile to get down. It looks like this is a Jazz and Hip-Hop class so we\'re going to do two hip-hop dances and two jazz dances. It looks like the teacher is going to run it much the same as she ran the High School class except this class is strictly jazz and hip hop where the High School class has pretty much every kind of dance involved.  The bus came at exactly 12:10 so I just pretty much stood around for like 45 minutes for nothing...but hey I got there early and was able to scope out the campus. I located the guy\'s locker room and it was nice and clean. It had three big group showers, 100 or so lockers, a bathroom and benches to change on. I got the prep from this really bored looking guy sitting in the office and they had towels and stuff, so I guess we dont have to bring our own, which is cool.  I\'m looking forward to tomorrow starting the Hip Hop dance but also tomorrow I have to go back to real school (High School that is) and then I\'m going to have to go to night classes to make up for missing the day class from 5:45-7:50 which is going to suck. My Mom is already bitching about driving me out there. And then next week is really gonna be insanity because I have finals at my school and want to make the daytime class so I\'m gonna have to have Lora Sisco come pick me up at school (hopefully be able to get out of class a little early) and then get to Southwestern really fast and hope to get in there on time.  Last night and today I\'ve been working on a new version of my website, TOGZone.com and its finally looking pretty good I guess?  Anyways there isn\'t much more to write, so I\'m gonna be out! Ciao  '),(17400,2413978,'06,June,2004','2004-06-06 00:00:00','   Getting ready to go to Southwestern College today for my Jazz Dance class. Dressed up all alternative for once, wish I had a picture of it, but I dont have my camera with me.  Anyways - nervous but happy. Need to go catch the bus in a couple of minutes.  Out-  '),(17401,2413978,'06,June,2004','2004-06-06 00:00:00','   urlLink'),(17402,2413978,'06,June,2004','2004-06-06 00:00:00','   Today was been pretty cool. I was on touchtone teller with my Bank and apairently they screwed up or something and I have like $50 more then I should have in my savings account so I withdrew like $40 and bought some new shoes at the mall. Then JC Penny was having a really cool sale and I got a bunch of cool clothes that were really cheap. These t-shirts were 4 for $20, $50 pants 2 for $20 and some other cool stuff.  Also we went to the iMax theatre and saw \"Forces of Nature\" which was pretty cool, but some parts of it were too loud. It was about Volcanos, Earth Quakes and Tornados.  After that we went to Jaliscos in Bonita, really good Mexican food, and I had Chile con Carne and got sick...  Interesting Post from Brad:  urlLink http://www.electricbiscuit.com/2004/01/pro-gay-pro-adoption-i-am.shtml  about his rantings and ravings + opinions on Gay Marriage. Its worth a read. A little lengthy.  Out-  '),(17403,2413978,'19,June,2004','2004-06-19 00:00:00','   Been working on my website a lot. If you\'re reading this, go check out http://togzone.com and let me know what you think of it, what it needs, your favorite parts, etc.  Also today I picked up my Jay-Z: The Black Album, Alicia Keys: The Diary of Alicia Keys and No Doubt: Rock Steady CDs that will greatly help me with my dance class. It kinda sucks that the Jay-Z CD isn\'t edited though...it has some really bad language in it that my Mom isn\'t going to approve of at home and my Dad looked terrified \"Don\'t let the neighbors hear that!\" heh. I guess I cant blame him.  Also ordered some new books. Pre-ordered the sequel to Soon by Jerry Jenkins and some others. Still cant get over how cool it is to have Gmail!  Well - off to roleplay. If you wanna read the logs, check out http://togzone.com and click on RP Logs.  '),(17404,2413978,'18,June,2004','2004-06-18 00:00:00','   Very exciting and action packed from the start, the Chronicles of Riddick kept me entertained through the entire movie but had some very unbelievable if not completely impossible scenes.   Vin Diesal did a very good job portraying his character but it would be nice to see him playing in a movie where he isnâ€™t a hardened criminal or at least a very tough person. In this movie, as in Pitch Black, Fast and the Furious, and XXX he was a criminal who turned out to be a hero in the end.   The movie shows a Sci Fi(sh) cult movement invade planets and convert people to their faith (make them followers) in more sense then one or exterminate them. I really didnâ€™t expect the movie to be about that, but it added to the twist.   Vin Diesalâ€™s character, Riddick goes through some crazy trials including running across a planet that had no atmosphere thus was 700 + degrees Celsius when the sun was shining on it and â€“200 Celsius when the sun is not shining on it. Well, when the sun was shining he and his girlfriend somehow managed to not melt and escape in a space ship that didnâ€™t melt. Riddick somehow managed to wet his body enough so the heat merely steamed off of him without inflicting tissue damage.   The end was interesting but I wish they would have closed the story. It was very obviously left open for a possible sequel.   In the end this movie was very entertaining and had a lot of interesting special effects, but it defiantly lacked depth.  '),(17405,2413978,'18,June,2004','2004-06-18 00:00:00','   Just was roleplaying basic life of two Slytherin Students at Hogwarts School of Witchcraft and Wizardry on a Harry Potter text based RPG.  A log of our RP can be found at the following:  urlLink http://www.togzone.com/nuke/modules.php?name=News&file=article&sid=36   RP is a major stress releif.  '),(17406,2413978,'17,June,2004','2004-06-17 00:00:00','   Since I saw on Blogger that I can get Gmail because I\'ve been active for sometime or something like that, I\'ve been IMing random people and telling them about it. Thats a REALLY good thing.  In the case you dont know what Gmail is, its an e-mail service through Google where you get 1000 mbs of space and ideally you save every e-mail you ever get, and right now since its still in Beta testing, you can only become a member through invite.  Its really awesome! Its made my whole day.  Also just got back from dinner. Had War Wonton Soup, Pork Lo Mein and a Po-Po Platter. Twas whatever.  '),(17407,2413978,'17,June,2004','2004-06-17 00:00:00','   Just went to Staples and spent like $20 because I\'m sick of the post office ripping me off...so I bought clear tape and then some padded packages to keep all my shipping for bookcrossing realted stuff affordable.  '),(17408,2413978,'17,June,2004','2004-06-17 00:00:00','   Today was my last day of this school year and it was kinda...boring. But it was cool that I got a 4.0 and it sucked that I had to say bye to several of my teachers that I\'m really close to: My English Teacher, my Geometry Teacher and my Biology Teacher.  All the classes were uneventful and we basically just had to sit there for 30 minutes while our teachers wrote out our grades. It sucked to say bye to the forementioned teachers though...cuz they really rock. I wish I could have them next year. My other teachers are neither really good or really bad, just whatever. My PE teacher I will have again next year though...so lets hope that goes good.  At my college yesterday, after I typed my paper for my dance class, I went to Counseling and I got some papers with all the under-grad requirements for a Psychology B.S Major and for a Biology B.S major at both UCSD and UC Davis. I\'m thinking of maybe a Psychology B.S with emphasis on Biology.  T\'would be interesting to take Psychology as a major with Dance as minor. Psycholgoy can still lead me into Medicine, give or take a few science classes, because both majors require a lot of science and a lot of math.  Time to go get dinner and then sleep sleep sleep.  Out-  PS: Looked at my Psychology Text Book - we\'re going to take two chapters a week and each chapter is like 40 pages with the Study Guide being about another 40 pages per chapter (give or take) so this should be a really  full  summer.  Reflecting: Had to say goodbye to Martha, Anabel, Tiina, Shavell and some other cool people today. Everyone minus Anabel who is a grade ahead of me is either graduating or moving to other schools, so I\'ll probably not see them again. I WANT TO TAKE MRS. HERNANDEZ AGAIN NEXT YEAR FOR MATH! ITS NOT FAIR THAT SHE CANT TEACH EXTENDED ALGEBRA!  '),(17409,2413978,'17,June,2004','2004-06-17 00:00:00','   Just got out of my dance final. The teacher, surprisingly enough, graded me an A-. She is usually a very critical grader and I\'ve come to the conclusion that no matter how good my dance or anything in her class is, I will never get a full A. It will always be an A-. Oh well...that still counts the same as an A on my transcript.  Last night I got my Psychology WorkBook and TextBook and my parents were both all wierd about paying, but my Mom gave me a check for half...so we should be all good. I also bought a packet of scantrons because the teacher told me we have two quizs per week.  After my class, with Jesette, I took the 709 H-St bound to the Trolley and took the Trolley to Palomar Station and then caught the 701 home...which took me out of my way but still got me home quicker then the 703 would have and I didn\'t have to hang out at the creepy station. All in all it took me 45 minutes to get home.  Right now its off to my Multimedia Final. That should be somewhat simple. I also found out that I got an A- on the Biology Final (actually it was a low D, but curved with the highest grade which was 38/50, I got 31/38). Also need to find out what my math grade came out to.  Peace  '),(17410,2413978,'16,June,2004','2004-06-16 00:00:00','   My Dad is over here freaking out about how he \"has nooooooo money\" and all this crap. Going on and on and asking me if I want to call the bank\'s touch tone tellar and affirm that he has no money. Its like he\'s trying to convince himself that he has no money or something. I really dont...if he says he doesn\'t have any...must not. But he\'s like trying to prove it to me, someone who already agrees with him.  I called my Mom and was asking her if she would pay half of the fees for my college text book and she said yeah and then I walked toward my Dad and so he could hear her say \"yes\" and he got up any walked into the other room *Snort* that was mature...  Then my Mom went on and on about how if she finds a cheaper one outside of the campus bookstore I\'m paying her back all the difference. She is too lazy to drive out to the college and give me a ride but I bet almost anything she is going to make a special trip and burn $5 worth the gas to affirm that there is or is not a cheaper book. She\'ll probably find a book that has a different title thats cheaper then the campus bookstore and tell me I should get (given) book because  it is, after all, a psychology book .  She is the cheapest bastard in the world. Today when I got home from school I went to my bathroom which is attached to her bedroom and I looked on her bed and she had hid the photo paper under a pile of dirty clothes so I dont \"steal\" the paper from her to print pictures for  her . Whats really sad is I dont /need/ her paper, I have my own, but I would rather use someone else\'s paper...anywyas its sad. I made sure I moved the paper and half-assed put it back under the clothes so she freaks out and accuses me, so I can tell her to count the pages to make sure I took none, which I\'m sure she will do... lol!!  Anyways, now I\'m off to school to get my text book. QPB just had an awesome special and I just got a bunch of books I\'ve been wanting super cheap.  Also forgot to say this, yesterday when I was at the bus stop going to SWC these cholos (Mexican Gang Member types) walked up to me and asked if I had any \"butt\". Because my $30 watch looks expensive and I had my book bag I guess they steryotyped me as a drug dealer and tried to buy \"butt\" from me (who the hell calls it butt? Its called dope or even better \'weed\').  Tonight I\'m supposed to take the bus home and get home all late but I\'m sure my Mom is going to pick me up and park at SWC to check for cheaper text books! Atleast I dont have to take the bus though. I guess her cheapness is going to pay off after all, atleast in a small way.  '),(17411,2413978,'15,June,2004','2004-06-15 00:00:00','   My biology final wasn\'t that bad today...but something else  is  bad. Actually two things. One, I have a wisdom tooth coming in that hurts like hell and two I have a math final next period that is said to be  very  difficult...so hence I\'m not looking forward to it.  Last night I argued with my parents over getting money for my Psych book and my Mom was being a cheap bitch as usaul and said she wasn\'t going to pay because I dont deserve it and I was the one that wanted to take the class and all this other shit. In the end she said if I bring her a receipt she would pay half, but she has ripped my Dad off in the past and will probably do it again.  Last night our computer got a virus, no doubt from one of my Mother\'s many forwards or popup mess waiting to happen flash movie sites and somehow she still managed to blame the virus on me. How she can justify that, no idea - but she still did it. I haven\'t even been on the computer for a month and right when I got ungrounded she grounded me again, so I will probably be on her computer very little from now on, but still I am blamed.  Oh well...such is life.  Also I got a couple copies of blurb today and my Atkins article was published, although they changed the title from \"Atkins, Weight Loss or Suicide\" to \"Atkins, Weight Loss or Health Risk\"...far less dramatic...but oh well, I have a by-line.  '),(17412,2413978,'15,June,2004','2004-06-15 00:00:00','   Today has been a little chaotic. I was sitting, and still am sitting, in the SWC library on the third floor where the computers are and there was an earthquake. Me and this girl next to me in my row of 6 computers got on the floor and under the desk and this slick looking jock guy just sat there and then when it was over in about 10 seconds he just gave us dirty looks.  Anyways, to this morning. I woke up and then took off to school early. In English I didn\'t have a final and for some reason the teacher wouldn\'t let me on the computer to work on my website so I just sat there for an hour until Gilbert was done with his final and then he came back and just kinda ignored me and went with his Drama people that I guess he\'s working with on his Drama final so I went to my Geography teacher\'s class for her final and got it out of the way so I could go home and then go to SWC at 10:30.  Just myself and (Geography Teacher\'s name here) I got to talk to her some and opened up with some stuff to her and she was pretty supportive about the secretive stuff. Also I explained my whole ordeal in middle school to her and she doesn\'t particularly like the culprit\'s son anymore then I do. She was telling me that he, everyday just eats his whole Geography period and like does no work. So I guess he is really the loser I thought him to be along with his mother, who is satan.  I finished my final and it was pathetic how easy the teaser questions were. I was telling the teacher some of the questions and she even got a kick out of them. Some of the questions a 1st grader could have gotten right (figuratively).  I left right as my real period started and then this guy who I never particularly liked because he is somewhat annoying and one of those attention seeker class clown types, brought me my off campus pass and seemed to be really decent. I had never really talked to him before. But I guess my opinion changed of him, although he will probably continue to be an attention seeker and even though I\'ll like him, he\'ll still be annoying to me.  I also found out that I won this raffle in my English class and that you had to be present to collect your prize and I wasn\'t...but I think it was only like a pencil or something... What sucks is I never win when I\'m present for the prize...and when I do I dont actually get my prize.  I walked home, got ready - my Mom\'s friend came over and was ringing the door bell and screaming her name - I guess she forgot my Mom works and she wouldn\'t just go away so I had to hide and be real quiet so she doesn\'t know I\'m there.  I got on the bus, the 701 and took it to where it turns onto H street and then I got on the 709 bus and came to SWC. I went to my Dance class which was really hard but somewhat shorter then the night class so better and it was neat I guess?  Which brings me back to being in the library and the earthquake happening...  Also I\'ve been working on my site, http://togzone.com and I got a Guest Book and an Image Gallery installed which work really cool...  Now I need to head for home. Out-  '),(17413,2413978,'14,June,2004','2004-06-14 00:00:00','   Today I went to my Mom\'s house after school to pick up stuff I need tonight for class and to drop off my stuff from school. Right after I got there she pulled in the drive way, in all honesty, I think she left work early (or actually later then she\'s /supposed/ to be there) to make sure I take no photo paper. I dont know why the fuck she would leave work for any other reason then that.  Then she proceeded to make me clean up the table and all this other shit she does...and then I was asking her if she was going to give me a ride home from the college and she gave me all this bull and ignored me and then answered me in little riddles.  As I was walking out the door I had two soda (one for myself, one for my Dad) and in her evil bossy voice she commanded me to put the Coke back. I argued the fact with her and said she\'s really messed up cuz my Dad gives her sodas all the time and she should return the favor and that I was just being nice and she let me take the soda but then she grounded me for a week from the computer. I really dont give a fuck about the fucking computer anymore. She\'s always grounding me from it anyways, and when I\'m ungrounded she doesn\'t let me on it because she comes up with some shitty reason or chore I need to do before I can go on it and then she\'s on it doing her Avon.  Anyways, vented that out. Feel better now. But I still hate my Mom and I guess she doesn\'t get it that she\'s pushing me further and further away. And whats really funny is I was already grounded from the computer this week because I didn\'t want to wash her car at 8:30 last night, do a crappy job since I cant see the car and then have to do it again. So I\'m doubley grounded from the computer this week...and whats really funny is I can use it at school all I want, minus certain blocked site (most of them I\'ve found a way around though) and then this weekend I\'m with my Dad...so I wonder what she thinks she\'s keeping me from in all honesty.  Anyways - talked to Dad. I think we\'re going to the fair this weekend. I got a free ticket from the newspaper advisor at my school to go and look at the newspaper judging contest. I think we\'re going Friday or Saturday. It will be a nice break from school seeing as Friday is out last day and theres no SWC classes on Fridays.  Plus I\'ll have some money too from grades and such.  Also - some shit head on Hot or Not rated me down and now I\'m at a 7. Some people gave me 1s and 2s...I would have to be pitiful to get that low...that was just wrong. I\'ve only ever rated one person a 1 and he was so ugly it made me turn my head...that is soooo wrong that they rated me that.  '),(17414,2413978,'13,June,2004','2004-06-13 00:00:00','   Vote for me on HotorNot.com!!  Follow this Link: http://www.hotornot.com/r/?eid=OLGRSUS&key=NHA  Last time I checked, about two hours ago I was an 8.3  '),(17415,2413978,'13,June,2004','2004-06-13 00:00:00','   Today all hell broke loose in my 4th period class. The teacher was not present so we had this really nice sub that I\'ve had before in another class where the kids behave and they were just cruel to her. She is a little older, probably in her late 60s and she is retired (she used to work here...for 35 years) and they stressed her out so much she looked like she wanted to pull her hair out...  And the class was lame today. Although I was able to talk to Traci about books for awhile so that was cool I guess?  Tonight I have to go to SWC early so I can get my ID made and get some extra work on my Dance. Mrs. Greenway said that I need to connect the movements more and go more with my theme of resisting gravity. Which I guess I will do. And today she was really cool, I take back not liking her - I guess she was so moody and like \"bleh\" in the first semester due to her pregnancy which would just be  wrong  to hold against her.  Mom was in rare form this weekend today and is a truly evil bitch. But today I\'m biting my tongue, being nice and kissing ass because I need a ride home from SWC tonight since I\'ll probably need Linda to come over to work on \"connecting the movements\" more on my dance.  Now tis off to Math Class, where it is the last day to cram and go nuts to get all of the final review packet finished.  '),(17416,2413978,'13,June,2004','2004-06-13 00:00:00','   Everyone say hello to my  UNDERAGE  cousin, Dan. Would you believe he is only 19 (I think...)?  urlLink  '),(17417,2413978,'13,June,2004','2004-06-13 00:00:00','   Sitting here right now, probably not supposed to be on the computer, but oh well, waiting for Linda to come  finally  so I can work on my choreography project that she said she\'d help me with.  I\'m really bored...  '),(17418,2413978,'13,June,2004','2004-06-13 00:00:00','   urlLink'),(17419,2413978,'26,June,2004','2004-06-26 00:00:00','   urlLink'),(17420,2413978,'24,June,2004','2004-06-24 00:00:00','   Theres this one guy in my Dance class...and man is he creepy. Throughout the whole period he does his own thing. At some point when he comes within a league of doing something even close to what the class is doing - he\'s delayed like 10 seconds; which leads me to wonder why he is there? And also - am I being cynical and he\'s actally retarted or something, or is he just super wierd?  Whenever we do any accross-the-floors, he does this wierd turning jazz walk thing...it never fails. And any sideways stuff...he does a forward jazz walk (and a forward jete - if he does  that  much.)  And he walks up to me and asks me really dumb questions. I think he\'s like...creepy. I try to get away from him and avoid eye contact. Whats even wierder is that he\'s like...30.  Oh well - I\'m gonna feel really bad if he\'s like retarted or something and I just wrote all this. But lets hope he\'s not...and just wierd for the sake of my conscience.  '),(17421,2413978,'23,June,2004','2004-06-23 00:00:00','   Today was hardcore in dance and in Psychology.  In psychology we had our second quiz (this one girl pointed out that theres only one day a week we dont either have a quiz or exam - scary) and since I studied to some extent this time, I think I aced the test. She came around and told me my score from the Quiz yesterday that I didn\'t think I did too hot on and I got a 12/15 (which is an 80%) which is bad, but not  too  bad.  (See post in Personal Blog on really hot \"person\" in Psych Class  urlLink Click Here )  In psychology right after the quiz she had us take another load of notes and then do some test strategizing (not looking forward to the test covering chapter 1 - I dont get chapter 1) and wow was it a lot. She let us out early again, so I guess we\'re going to get out at that time everyday.  Then in dance I think we had the hardest day so far (And I\'ve been saying that all week - so what does monday have to bring? Scary Thought) and this was easily the hardest week of my life between the dance class and the Psychology class. Today was our mid-term for dance and she gave us a hardcore warm up and ab-killers before we were video-recorded and had to perform the test (I think I got an \"A\" which is good.)  She also let us out of class early - and I was going to type up this psychology paper but I got ahold of my Mom and she came and picked me up (Saved!) so I didn\'t have to take the bus home and be against a bunch of sweaty people with no air to breathe (ick) piled in like sardines.  She drove me home - I immediatly made hamburgers and garden burgers (I was sooo starving after the dance final - I should start eating Lunch) and then loaded them with veggies and ate two and a little chili (then I was ready to explode) and then I went to take a nap and digest and I just woke up, which brings me to this post, and the really hot person in my psychology class *drools*.  Now I need to go mow the lawn and then do HMWK - and apairently we\'re going to Phoenix tomorrow (no idea why) and just chillin in our condo place (Mom has a time-share) and like relaxing I guess? Whatever - relaxing is gonna be real nice after this week. I can catch up on my reading and study for my test *shudder*.  Off to the lawn then my paper.  '),(17422,2413978,'23,June,2004','2004-06-23 00:00:00','   Lately as much as the dance class has been killing my muscles, I\'ve noticed definate results from it, in the weight loss and tone-up department.  I dont know that I\'ve lost \'pounds\' per se, but I\'m definatly noticing my fat toning up to muscle, particularly in my upper thighs, my lower calves and my stomach. When I was looking in the mirror standing up yesterday (was naked since I was about to take a shower) I noticed that my stomach (the fat anyways) is almost completely gone from the crazy crunches Mrs. Macias has been giving us - and my upper thighs are getting rock hard rather then that yucky flabby mess. The next thing I need to work on once this class is over is Cardio and upper body strength. Upper body strength is something I totally lack.  What sucks is that when I\'m sitting down and my stomach is all crammed together, there is still quite a bit of fat, but standing up the muscles stretch it out, and I look almost skinny. I still want to lose about 40 pounds though...even if I\'m toning up, I dont want to weigh 235.  Peace-  '),(17423,2413978,'22,June,2004','2004-06-22 00:00:00','   Ok - ranting and raving, not that interesting of a post, so not important to read.  Last night I was on the computer, happily, in our computer room, working on my website and my Mom called me, like she does every 5-10 seconds to do something she can easily do herself. This time she called me clear in the other room to turn on her tv that is on top of the refridgerator, where another tv was on in the next room, which she could see from the kitchen (I tell you she is obsessed) so when she told me to turn on the Tv on top of the refridgerator (which I might add she wouldn\'t be able to see while she cleans the kitchen or whatever she was doing any easier then the one in the family room) and I rolled my eyes just because it sounded stupid to me. When I rolled my eyes I happened to be twisted TURNING ON HER DAMN TV and she slapped the hell out of my face because I think her wanting to have every TV in the house on is dumb and I slapped the crap out of her too out of reflexes...if you\'re twisted around and someone slaps you, its an involuntary reflex to slap them back or just twitch...and my arm happened to whip around. I actually think it was pretty cool that I slapped her unvoluntarily and didn\'t get in trouble - which was a shock.  Well she started bitching about how it was all disrespectful to roll my eyes at her and that is an an adult and can do \"whatever the fuck she wants in her house\" and that I better never say anything about her TV habit again (so she knows she is obsessed and has a problem and gets all touchy when I hint at it!) What really sucks about all this is that she totally over-reacted. I roll my eyes all the time, not like its something directed at her, its a natural thing I just do...and if she would have been asked something that ridiculous she would have rolled her eyes too (maybe /that/ wasn\'t ridiculous to her - but something that would be.)  What really pisses me off is that she called me  clear  from the other room to do that when she could have easily done it herself. Everyday I hate TV more and more because I swear thats all she fucking does anymore. And when I want to go on the computer she wont leave me alone. She made some comment about how I\'m back there ignoring her (but what is she doing watching TV? Socializing with me?) yet whenever I\'m in the room when she\'s watching TV and I say something she huffs and puffs and get all pissed that I  dare  interrupt her precious show and she tells me to ask her during commercials, but since she records every channel that there is, there are about 10 seconds of commercials because she just fast forwards through them. GRR!  Everytime I\'m on the computer, I swear she finds a fucking reason to kick me off! I cant even do homework on the computer in peace (yet she wants peace to watch TV). It really sucks. Last night I was not on the computer 5 minutes and then she came in and kicked me off to use it (and check her e-mail) for  a second  (which lasted more like an hour) to spam up her contacts\' inboxs with her shitty virus containing forwards. Even more ironically last night she had two virus containing e-mails and somehow managed to blame our last three viruses on me (when I hadn\'t even been on the computer in like 2 months). Then right after I actually got the computer back she called me to  turn on  her  tv  (because she is an adult who can do whatever the fuck she pleases!)  I think the next time I get grounded from the computer long term (she threatoned six months at three different points last night) I\'m going to get even. I\'m going to go around and move her remotes some her various TVs and more them from where they should be, but have them where they\'re not hidden either (she doesn\'t like changed from her daily routine and doesn\'t know how to look for things, yet tells me constantly \"Its not gonna jump out and say \'here I am\'\") so she has to get up off her sorry ass and look for something herself or just not watch TV (better yet.)  In some ways my Mom drives me so crazy its not even funny.  But right now I need to get to another post and then to class.  Peace-  '),(17424,2413978,'22,June,2004','2004-06-22 00:00:00','   Today in Dance was probably the most packed and hardest day we\'ve ever had in there. The teacher killed us...  And on the hip hop dance she changed the beginning part to the break dancing so it took me awhile to get that down.  Psychology wasn\'t that bad. She even let us out 15 minutes early. If it stays at this pace it wont be near as bad as it supposedly is going to be since its only a six week course.  Now off to my psychology HMWK.  ');
INSERT INTO `msgs` VALUES (17425,2413978,'21,June,2004','2004-06-21 00:00:00','   Today I started out on Psych 101 and it was interesting. We did the whole introduction to the couse thingie and then went in to the psychological motivators behind 9/11. Then the class ended and I went on to dance where the teacher worked us like there is no tomorrow. But I can feel myself getting stronger and toning up and I went to this church\'s office (that I used to go to) to say hi to this one guy, Gabe, that I like really love (in a brotherly way) and he wasn\'t there so I had to talk to this guy where there is mutual dislike and he was acting all political like he actually liked me *gag* and then went and talked to Christian, who isn\'t on the same level as Gabe, but is still a really cool guy, gave him the update and then went and talked to the office ladies who I get the impression dont like me either, but they\'re still much more discrete about it then the other guy. If someone doesn\'t like me, it pisses me off when they pretend they do...just tell me you dont like me, dont pretend you do and be nice to me, and we\'ll be cooler with each other. I can take that - you wont get along with everyone in life.  Anyways - was giving my Dad and my Mom my little commentary on religion and our origin. They dont agree but I thought I\'d post it here just in case it changes so I can look back someday.   On Prayer/Meditation I believe that prayer, though it is talking to God, is not some magic method where God goes \"poof\" and whatever you pray for appears. The reason I feel is doesn\'t work that way is many people from many different religions pray to different deitys and they\'re prayers (in their opinion) are answered and then people who have no deity and just meditate still find solutions to their problems - which leads me to the conclusion (I heard this from someone else, so its not origional - but it is customized) that when we pray or meditate, through focused energy it brings us to a higer state of consciousness and we\'re able to think clearer and solve our own problems.   On Religion Vs. Science Thought I believe in the God of the Christian/Jews/Muslims (One in the same) I dont believe word for word about the origin of the earth and I also believe Religion and Science can work together. I think God did indeed spark off the whole big bang creation thing, and he guided it, but I think he let it run its own course. Since the earth is said to be somewhere around 7 billion years old and the universe way older then that, I think that each \"day\" of his creation cycle symbolized somewhere around a billion years. So the creation of animals, he evolved them over that billion or so years and then stopped them there and went on to the next cycle. I know that is conflicting, but it makes sense for me. I used to be brainwashed into not believing in evolution at all, but agiain, there is too much evidence for it. In the end of the cycle, I believe God stopped macro evolution and then thats when natural selection and micro-evolution (adapation) came into the picture. I believe God used Evolution to create the earth, set it up so he doesn\'t have to intervene and then just lets it run its course.    On Micro Evolution Next would be micro evolution and natural selection. I believe those are both very active today, and most narrowminded Christians that are actually informed and make some of their own opinions cant even deny that. Survival of the Fittest is in every species and ecosystem and things do adapt over time and evolve over time on a small scale. Maybe over a couple hundred years the size of a bird\'s beak changes to accomidate a new food source or the color of a community of rabbit\'s coats changes to better blend in. I do not believe that macro evolution (one species changing into another) happens anymore and if it does still happening, its very gradual and no one will ever live long enough to observe it happening - so why worry about such things?  On a more Christian scale I think that people believing God *poof* gives them what they pray for, in example of someone hanging money to someone who is in need, is purely coincidence but then yet I believe in spiritual warefare too, so some of my opinions conflict.  As for an afterlife (Heaven, Hell, Purgatory, Summerland, etc) I dont know what I believe. The bible seems to me to be a corrupted book that started out good and went bad. Anything written by man is corrupted because man is biased and they put their own opinions in there. As much as someone asys \"It was inspired by God\" that doesn\'t necessarily mean the men  listened  to God. And provided they did, the various translaters who switched the bible from Armanic to English and various other languages (even Latin) put their biases in there and changed it around. So how accurate is the bible to its origional context? But then who am I to say its corrupted too? I dont know that for sure.  Again - these are purely opinion - which is somewhat scary since I\'m the Christian Club President at my school, but maybe God is using my opinions to truly inform people? A lot of the stuff in the bible can still be accurate, but tahts also leaving room for some stuff being corrupted. What you have to do to be \"enlightened\" and go to heaven is a little farfetched and hardcore. If you really have to do all that stuff (die to yourself, sell all your stuff - give the money to the poor and follow God, etc) then very few people are actually going to heaven.  This is all for now.  '),(17426,2413978,'20,June,2004','2004-06-20 00:00:00','   Just reflecting I remember when I was a little kid and I used to buy little kiddie shaving kits and be all excited about having facial hair and having to shave...now that I have it - it really sucks. When I only had to shave like every week or so it wasn\'t bad, but pretty soon (probably within a month) my facial hair will be growing so much that I\'ll probably have to shave like every day.  On second note, start Psychology 101 today - should be interesting. Two hours of note taking then a kick ass dance class. I\'ll probably be about to drop dead when its all over - then staying up till like 1:00 doesn\'t help either...  Out-  '),(17427,2413978,'20,June,2004','2004-06-20 00:00:00','   I was just in a meeting on this Harry Potter themed text based RPG (Moo) that I play on and we were having a meeting about what will happen if Magic stops working altogether and the damn admin guy pretty much dismissed  everything  that I contributed to possible things that would happen.  It really sucked. I mean, I already dont like him, but now I dislike him even more. If my ideas sucked, he could have at least been polite about it. He was just like \"No\"...what kind of politically correct admin is like that.  Anyways, on further, got in an argument with my Dad about how Strogonoff and Mashed Potatos done go together...and he got all pissed, got all offended and said then dont eat them *rolls eyes*. He doesn\'t get the concept that primarily Carb meal isn\'t the way to go. Noodles and Mashed Potatoes...carb heaven - but still just wrong.  He was freaking out about how I need to help around the house so I vacuumed and cleaned down the bikes and watered the plants! Go me! Then we got in another argument about how he wanted me to give my bike away and I told him that Chula Vista has very few flat surfaces to ride bikes and most people I dont know wouldn\'t have ridden them anyways so he got all pissed about that and just blabbed on about it for like 30 minutes. I just tuned him out and nodded at the appropriate times.  Mom came over and brought my dance shoes, which was really decent, and then I talked to her for awhile. This is like the first time in a long I would rather be with my Mom then over here...but oh well - not gonna happen.  '),(17428,2413978,'20,June,2004','2004-06-20 00:00:00','   Tonight about 1:20 my Dad got up and went pee and then I heard him just like gaging his head off. He was spitting up this crap for like a good 15 minutes and then I asked him if he was ok and he said it was just indigestion.  Whats really scary is I think its from him being an alcoholic since he was like 16. He\'s probably eated away at most of his asophagus and stomach and intestines so they can\'t take anymore alcohol so it causes him to puke up crap. Next will probably be heart, liver and kidney issues. The body isn\'t made to take that much alcohol...but I guess its his body...he has to live with it if he messes it up.  What sucks for me is that my Dad\'s family has super addictive tendancies for a lot of stuff and I\'m gonna have to be really careful as an adult or I\'ll probably end up being an alcoholic just like him because of the addictive personality I already have with some stuff. I mean, when I find something cool, organizations, people who show me attention, etc I like obsess over whatever for up to a month and then like it climaxes out and I find a new obsession...and if I\'m doing that now what am I going to do the first time I\'m offered drugs or alcohol and are stupid enough to accept out of peer pressure? Scary thought..  Anyway - its late, I\'m off to bed.  PS: The Rock Steady and Diary of Alicia Keys CDs are both really good. The Jay-Z: Black Album one is vulgar and I wish I could have just got that one song rather then the whole CD and saved my money...oh well.  '),(17429,2413978,'03,July,2004','2004-07-03 00:00:00','   Today has been fun I guess.  My Dad and I went to the fair today and right when we went in we went over to the Student Showcase where they had the High School Newspapers and I saw the paper from my school (it was the one that had  none  of my articles in itâ€¦) and then my Dad was moping around and looking at all this crap in the wood workers area *sigh* and it took me forever to get him to leave that area and get over the rides.  He gave me $30 which as really nice â€“ it should get me outta debt and then I can pay off my books. He bought me $20 worth the ride tickets, which was 6 rides and then I went on all sorts of their crazy rides â€“ however I didnâ€™t get to go on the zipper. I wanted to go on the zipper but my Dad didnâ€™t and they donâ€™t allow single riders to go on it and I was alone.  After going on all the rides we went and got some Australian batter fried potatoes, I got ranch and my Dad got cheese â€“ and they were crazy big orders. We could have split an order and probably not eaten all of them. After we were leaving the area where we ate, we went by this area where this guy forcefully put us into seats where we got $12 â€“ 10 minute long full body massages. He did a really good job, although he was pretty rough at first on my shoulders and I had to ask him to be gentle. A couple of times he got a little too far downâ€¦  After the massage I went and did the climbing wall thing â€“ and because the 18 year oldish skater dude attendant guy was hitting on me (he wasnâ€™t bad looking himself) I got to climb the wall 4 times instead of just one for my $5 â€“ so it wasnâ€™t a bad deal all in all.  During my 4th time up there was this chick that was scared to repel, climbing down the call and causing a traffic jam, so I was just standing there holding on to the wall and when I finally got down, my hands and fingers and forearms were killing me (and still are â€“ my fingers are randomly cramping up.) Well the hot guy who was hitting on me â€“ offered for me to go up the fall a fifth time (am I that hot?) but I had to decline cuz my fingers hurt.  Then after that we went to the exhibit halls and I found an ear piercing place â€“ so I decided because they were cheap I wanted to get my left ear pierced so I did â€“ it didnâ€™t hurt at all since she used a gun thing (I did flinch though â€“ just nerves) and now I have a pierced ear that doesnâ€™t look to bad.  I wanted to get my eyebrow pierced too, but she said they only do ears â€“ so I guess that will be another time and another place.  Before we went to the fair, my Dad took me to my Momâ€™s to pick up the mail and these super lazy â€“ stupid stoner neighbors of oursâ€™ kids were actually working. Their mother has to beg and plead for them to even take out the trash and the one skinny stoner dude (he constantly wears this weird cap thingie around) acts like its super hard and kills him to move the trash cans all of 10 feet from the side of their house to the curbâ€¦ Anyways â€“ he and his stoner brother were cleaning their yard (wow! Hard work) and removing these ugly railroad ties that the last loser neighbors put in there. Who could have guessedâ€¦  Anyways â€“ now I sit here â€“ tired and yeah.  Peace out for now.  (See Biting My Nails â€œNew Postâ€ -  urlLink Click Here  '),(17430,2413978,'02,July,2004','2004-07-02 00:00:00','   Question:  Suppose that a new \"miracle pill\" is developed that will allow a person to function with only one hour of sleep per night. However, because a night\'s sleep is so short, a person who takes the pill will never dream again. Knowing what you do about the functions of sleep and dreaming, what would be some advantage and drawbacks of such a pill from a personal standpoint? Would you take such a pill?   My Responce:  REM (Random Eye Movement), a cycle during which sleep takes place, is a major portion of our nightly sleep cycle and gives our brain the ability to rest and manifest random pictures (dreams) because it becomes stimulated by random neurons.  Theoretically, the body requires a certain amount of REM sleep per night to function properly. Subjects taking part in sleep studies, when awoken during a dream, spent significantly more time in REM sleep when they fell back to sleep, making up for the time lost when they were awoken to record their dream â€“ showing that humans do in fact require a certain amount of REM sleep.  So â€“ to function properly, humans requiring a certain amount of sleep per night, is a given. If deprived of that sleep they will be able to live their lives normally, but they will be snappy and irritable because their brain isnâ€™t rested, even though they themselves may feel rested.  According to the Dreams-for-Survival Theory, â€œâ€¦dreams represent concerns about our daily lives, illustrating our uncertainties, indecisions, ideas and desires. Dreams are seen, then, as consistent with everyday living. Rather than being disguised wishes as Freud suggested, they would represent key concerns growing out of our daily experiences (â€œEssentials of Understanding Psychologyâ€ page 123)  The text goes on to say, â€œDreams permit information that is critical for our daily survival to be reconsidered and reprocessed during sleep.â€ (â€œEssentials of Understanding Psychologyâ€ page 122)  So, if a miracle pill were invented, under the influence of which we would have to only sleep one hour a night, it would cause major problems for humans. Yes â€“ they would have much more time on their hands to do the many tasks in their life which they donâ€™t have time for â€“ but they would be snappy to their colleagues, they wouldnâ€™t be able to â€œpermit information that is critical for our daily survival to be reconsidered and reprocessed during sleepâ€ and their body would be constantly asking for more REM (as seen in studies of victims awoken during REM.)  Due to the above stated issues, I would personally not take any kind of miracle pill that made it so I didnâ€™t have to sleep. Though I would have more time on my hands, I wouldnâ€™t feel as rested as I do with 6 or more hours of sleep (including REM, of course) and my life probably wouldnâ€™t be as fulfilling. My senses would probably also be somewhat numb after awhile. Plus dreams, in a lot of cases are interesting â€“ especially when they are erotic dreams.  '),(17431,2413978,'02,July,2004','2004-07-02 00:00:00','   Question:  If the eye were constructed with a second lens that \"unreversed\" the image hitting the retina, do you think there would be changes in the way people perceive the world?   My Responce:  The human eye, seeing things upside down, yet the brain reversing the image and letting us see it right side up is a very interesting human ability. According to the book we do actually see things upside down and the brain turns them right side up, however, our bodies do not work like a microscope in that there is not a second lens that turns seen images right side up â€“ we have adapted to do it ourselves.  If for some reason our bodies were to change or evolve where we acquired a second lens that turned everything we see un-reversed, there could be two outcomes. One â€“ our bodies would see everything upside down, like the researcher who wore the glasses in England where she saw everything upside down, then after a certain amount of time our bodies would get used to the upside down image and be able to function normally, seeing the world upside down, thus we would just learn to live with the world upside down, and after awhile would know no significant difference.  Or â€“ a second outcome â€“ our brain over time would turn things back reversed so our body functioned as it did before; although a much more possible outcome would be that the function of the brain associated with un-reversing all of our vision would simply become obsolete because it wouldnâ€™t be needed to reverse the images anymore â€“therefore the second lens would ultimately perform that function for the brain and the brain would no longer deal with the turning/reversal of images.  '),(17432,2413978,'02,July,2004','2004-07-02 00:00:00','   In Psychology we\'ve been studying the states of consciousness and stages of sleep and its been really interesting. I think I\'ve retained almost all of it just because its so interesting. The laws and theories of organization (Gestalt) have been throwing me off though, so I\'m going to have to get those down by Tuesday. (Monday is going to be the cram and study my ass off day for the exam Tuesday.) Thats the main thing that sucks about college - the prof\'s dont want to be grading a bunch of writing assignments you can BS your way through - so they make most of your grade tests on scan-trons to make things easy for them, and that sucks for me because I\'m not the best tester in the world...  I\'ve been dropping a lot of weight, or at least toning up the weight I do have but to get rid of the love-handles and excess fat Mrs. Macias has told me I have to start running  YUCK!  so I guess I will start running, but I swear I look like a girl when I run, so I need to find a place to do it where people wont see me - maybe the track at my High School after-hours or the track at the college after dance (provided there are no classes using it) so I can get completely in shape by the start of High School.  Unfortunatly dance is only going to continue for two more weeks and then I\'ll have to be self-motivated until school starts back in the fall (Modern Dance then and then High School PE - with timed running)  YUCK!  So I\'m going to do Pilates and then go somewhere and run on my own (Maybe do something active with my Mom - run while she bike rides so she can get some kind of health benefit from the deal) then then, like I said, continue the toning and dance supplement by doing MTV Pilates when dance is over.  I hope I come out of school with a really good grade in both classes so I can continue on with cool classes and get a lot of college out of the way before I actually enter college - so I can take a whole load of the biology classes I wanna take.  Right now I need to get on with my Psych writing assignments and then get ready for bed because we\'re going to the fair tomorrow.  Peace Love and Golden Oreos!  '),(17433,2413978,'02,July,2004','2004-07-02 00:00:00','   Lately I\'ve been considering just testing out of High School and going to college full time because it comes pretty easy to me, I like it more because we have a lot of freedoms in Collge (as adults rather then children) and its just all the more laid back. The people there want to be there, they dont have to be there, most of them care because they\'re paying to be taught where High School their parents pay, etc.  And as easy as I could just go test out of High School, get a GED and go to SWC full time (and get completely done in my early 20s) I would be ineligable for most scholarships, have to pay for super expensive student loans and there would constantly be a war with both of my parents over whose paying this and whose paying that and I would probably eventually get to the point where neither would or could pay anymore, and I would just be sitting there not going to school - working full time at a going-no-where job or be just spending like crazy and piling up my loans fees.  So I guess in the end I\'m staying in High School, or at least to my junior year. That would be at least one more year, probably two.  Whats really great is that I have an \"A\" in Psychology right now (right at 90% average) and I probably have an \"A\" or \"B\" in Dance, which is good considering I\'m barely a sophomore and I can successfully do great in high school classes.  Without the encouragement from Mrs. Hernandez, Mrs. Murphy and Mrs. Macias I totally wouldn\'t be here right now though. Without Mrs. Macias being a true inspiration in dance at CPH I would have had no reason at all to take any form of college classes, be at home being lazy the whole summer or working my ass off doing chores my Mom wants done which sucks. Thats how I spent summers before last year. Last year I had a break - I was at Cornerstone everyday of the week, but look where that got me now... (see two posts down from here for more information)  Peace-  '),(17434,2413978,'02,July,2004','2004-07-02 00:00:00','   Added my cousin as an Abstract Secant contributer because her life is as screwed up if not more screwed up then my own and we can be a pessimistic duo this way.  She\'s Beth Schardin. Say sup.  '),(17435,2413978,'29,June,2004','2004-06-29 00:00:00','   To start, its really sad to me this is how things have turned out. I used to go to Cornerstone and be really happy there - then they found out I had some issues that I was seeking help with, and I was removed from every \"ministry\" I was involved in and then the little \"sinner\" whom they \"wanted to help\" but never actually initiated that process.  Recently for some reason, several of the people that I totally love from Cornerstone haven\'t been responding to anything - email, instant messaging, they\'re \"not there\" if I call - etc, which is getting pretty obvious - so I guess since I left now I\'m the \"evil evil sinner\"?  I have really good reasons for leaving - the Pastor turned Hollywood and was there more for his TV camera then his congregation, that whole getting kicked out of everything and the youth leader guy is another reason. He screams and gets all mean looking when he preaches, his messages are very fundamentalist right-wing messages where he doesn\'t even icknoledge that maybe another side exists (IE: Brainwashing) and he told me I\'m going to hell for watching MTV (and since I watch it whenever I can, I guess going to hell really really now?)   7/2/2004		  1:14:40 PM Michael Brant: sup   1:23:04 PM Michael Brant: I have samplers of \"Something Like Silas\" from EMI Record Group - if you\'d like one for the purpose of XSCAPE along with little sticker thingies for the new Chris Rice Album coming out, it would be bad to continue ignoring me...   1:23:17 PM Michael Brant: if you want them...   1:23:25 PM Michael Brant: and assuming you do, where do I mail them to?   1:23:45 PM Gabe: yeah dude that would be sick   Notice that he doesn\'t answer me until he knows that he has something coming (A free CD sampler thing)   7/2/2004  1:44:25 Michael Brant: whose preaching (or whatever its called) tonight?   1:45:06 PM Gabe: Richard   1:45:24 PM Michael Brant: is he preaching next week too?   1:45:35 PM Gabe: no miggy is preaching next week   1:45:54 PM Michael Brant: well...I\'m certainly not gonna be here this week if Richard is preaching.   1:45:57 PM Michael Brant: maybe next week?   1:46:07 PM Michael Brant: trying to get Beth to go - cuz I dont wanna go alone   1:46:13 PM Gabe: whats up with that dude   1:46:19 PM Gabe: you dont like rich   1:46:22 PM Gabe: ?   1:46:29 PM Michael Brant: I didn\'t say that   1:46:35 PM Michael Brant: I just think he gets a little overboard   1:46:46 PM Michael Brant: and is very one-sided on all of his messages   1:46:52 PM G: he just tells it like it is   1:46:55 PM Michael Brant: he doesn\'t allow the possibility for anything else   1:47:06 PM Gabe: give him a chance   1:47:12 PM Michael Brant: actually - he tells it how someone else says it is   1:47:17 PM Michael Brant: most of those aren\'t his own opinions   1:47:24 PM Gabe: what do you mean   1:47:31 PM Michael Brant: and when he said all the people who watch MTV are going to hell - that kinda was it for me   1:47:42 PM Michael Brant: I mean - that most of his sermons he gets from somewhere else   1:47:49 PM Michael Brant: so those aren\'t his opinions   1:47:53 PM Gabe:  he never said that   1:47:57 PM Michael Brant: he\'s just re-iterating someone else\'s opinions   1:47:59 PM Gabe: or meant it the way you took it   1:48:03 PM Michael Brant: he did say that - at Sunday school at XSCAPE   1:48:09 PM Michael Brant:  err...   1:48:18 PM Michael Brant: when we were at Hilltop   1:48:20 PM Gabe: but he didnt mean it like that   1:48:21 PM Michael Brant: in that one classroom   1:48:25 PM Michael Brant: thats what he said...   1:48:29 PM Michael Brant: not even like I perceived it wrong   1:48:41 PM Michael Brant: \"If you watch MTV - you\'re not really a christian\"   1:48:44 PM Michael Brant: or something to that effect   Here\'s the whole MTV incident.   7/2/2004 		 1:49:02 PM Gabe: well hopefully ill see you tonight man i gotta get some stuff done   1:49:13 PM Michael Brant: you wont see me tonight - but bye   1:49:21 PM Gabe: alright see you later   At the end, when he knows I\'m right and theres nothing he can do to sway me, he turns it around like nothing happens, we\'re all happy Christians out of the Cornerstone \"Christian Factory\" (Inside thing - read the book The Visitation by Frank Peretti) and I\'m still going to be happy too and go tonight - to a church I dont like - to listen to a guy who preaches super right-handed messages he gets from some website - ignorant to all the things that he denies (a lot like my Mom actually - when it comes to his opinions - he seems to me, to not form his own)  '),(17436,2413978,'29,June,2004','2004-06-29 00:00:00','   7/2/2004		1:14:40 PM		Michael Brant		G		sup 7/2/2004		1:23:04 PM		Michael Brant		G		I have samplers of \"Something Like Silas\" from EMI Record Group - if you\'d like one for the purpose of XSCAPE along with little sticker thingies for the new Chris Rice Album coming out, it would be bad to continue ignoring me... 7/2/2004		1:23:17 PM		Michael Brant		G		if you want them... 7/2/2004		1:23:25 PM		Michael Brant		G		and assuming you do, where do I mail them to? 7/2/2004		1:23:45 PM		G		Michael Brant		yeah dude that would be sick 7/2/2004		1:23:50 PM		G		Michael Brant		Silas is sick 7/2/2004		1:23:56 PM		G		Michael Brant		Miggy saw them in concerct 7/2/2004		1:24:02 PM		G		Michael Brant		he always sees them at flood 7/2/2004		1:24:22 PM		Michael Brant		G		whats the address for there? 7/2/2004		1:24:39 PM		G		Michael Brant		347 1/2 Third Ave. 7/2/2004		1:24:44 PM		Michael Brant		G		ok - still same one 7/2/2004		1:24:46 PM		G		Michael Brant		Chula Vista CA 91910 7/2/2004		1:24:48 PM		G		Michael Brant		yup 7/2/2004		1:24:50 PM		Michael Brant		G		I figured you people moved already 7/2/2004		1:24:57 PM		G		Michael Brant		us people? 7/2/2004		1:25:08 PM		Michael Brant		G		the church staff people 7/2/2004		1:25:14 PM		G		Michael Brant		oh right 7/2/2004		1:25:56 PM		Michael Brant		G		k 7/2/2004		1:26:00 PM		Michael Brant		G		expect them in the near future 7/2/2004		1:26:17 PM		G		Michael Brant		thanks 7/2/2004		1:26:25 PM		G		Michael Brant		why dont you come tonight 7/2/2004		1:26:59 PM		Michael Brant		G		uhh...dont want to? 7/2/2004		1:27:12 PM		G		Michael Brant		howcome 7/2/2004		1:27:40 PM		Michael Brant		G		umm...cuz? 7/2/2004		1:28:28 PM		Michael Brant		G		today is day off from school 7/2/2004		1:28:30 PM		Michael Brant		G		I\'ve been super busy 7/2/2004		1:28:34 PM		Michael Brant		G		I slept till like 12:30 today 7/2/2004		1:28:44 PM		Michael Brant		G		so 12-13 hours of sleep 7/2/2004		1:28:51 PM		Michael Brant		G		no school till Tuesda 7/2/2004		1:28:53 PM		Michael Brant		G		Tuesday* 7/2/2004		1:28:58 PM		Michael Brant		G		I wanna take full advantage of that 7/2/2004		1:33:08 PM		Michael Brant		G		what time is Xscape at these days anyways? 7/2/2004		1:33:16 PM		Michael Brant		G		I heard something about a HS/MS split? 7/2/2004		1:33:20 PM		G		Michael Brant		yeah 7/2/2004		1:33:27 PM		Michael Brant		G		if for some reason I /did/ want to come... 7/2/2004		1:33:28 PM		G		Michael Brant		8:30 for you 7/2/2004		1:33:30 PM		Michael Brant		G		ok 7/2/2004		1:33:46 PM		Michael Brant		G		is Linda Ramos still going to the High School one since she works at night? 7/2/2004		1:33:52 PM		Michael Brant		G		or no? 7/2/2004		1:33:53 PM		G		Michael Brant		yeah 7/2/2004		1:33:56 PM		Michael Brant		G		ok 7/2/2004		1:33:58 PM		Michael Brant		G		cool 7/2/2004		1:38:41 PM		Michael Brant		G		how has been life? 7/2/2004		1:41:27 PM		Michael Brant		G		If I go tonight, can I get a ride from someone? 7/2/2004		1:41:31 PM		Michael Brant		G		since it gets out late, I assume? 7/2/2004		1:41:39 PM		Michael Brant		G		My Dad isn\'t gonna want to pick me up I dont think 7/2/2004		1:41:45 PM		G		Michael Brant		probably 7/2/2004		1:41:50 PM		G		Michael Brant		hold on real quick 7/2/2004		1:43:56 PM		G		Michael Brant		Chris probably wouldnt mind 7/2/2004		1:43:58 PM		G		Michael Brant		but hes not here 7/2/2004		1:44:03 PM		G		Michael Brant		so ... 7/2/2004		1:44:05 PM		Michael Brant		G		Chris wong? 7/2/2004		1:44:09 PM		G		Michael Brant		yeah 7/2/2004		1:44:13 PM		Michael Brant		G		at DAd\'s this weekend 7/2/2004		1:44:17 PM		Michael Brant		G		tis would be out of the way for chris 7/2/2004		1:44:25 PM		Michael Brant		G		is Richard still leading up Xscape? 7/2/2004		1:44:30 PM		G		Michael Brant		yeah 7/2/2004		1:44:31 PM		Michael Brant		G		oh... 7/2/2004		1:44:48 PM		Michael Brant		G		:S 7/2/2004		1:44:48 PM		G		Michael Brant		im sure someone could give you a ride later 7/2/2004		1:44:59 PM		Michael Brant		G		whose preaching (or whatever its called) tonight? 7/2/2004		1:45:06 PM		G		Michael Brant		Richard 7/2/2004		1:45:24 PM		Michael Brant		G		is he preaching next week too? 7/2/2004		1:45:35 PM		G		Michael Brant		no miggy is preaching next week 7/2/2004		1:45:54 PM		Michael Brant		G		well...I\'m certainly not gonna be here this week if Richard is preaching 7/2/2004		1:45:57 PM		Michael Brant		G		maybe next week? 7/2/2004		1:46:07 PM		Michael Brant		G		trying to get Beth to go - cuz I dont wanna go alone 7/2/2004		1:46:13 PM		G		Michael Brant		whats up with that dude 7/2/2004		1:46:19 PM		G		Michael Brant		you dont like rich 7/2/2004		1:46:22 PM		G		Michael Brant		? 7/2/2004		1:46:29 PM		Michael Brant		G		I didn\'t say that 7/2/2004		1:46:35 PM		Michael Brant		G		I just think he gets a little overboard 7/2/2004		1:46:46 PM		Michael Brant		G		and is very one-sided on all of his messages 7/2/2004		1:46:52 PM		G		Michael Brant		he just tells it like it is 7/2/2004		1:46:55 PM		Michael Brant		G		he doesn\'t allow the possibility for anything else 7/2/2004		1:47:06 PM		G		Michael Brant		give him a chance 7/2/2004		1:47:12 PM		Michael Brant		G		actually - he tells it how someone else says it is 7/2/2004		1:47:17 PM		Michael Brant		G		most of those aren\'t his own opinions 7/2/2004		1:47:24 PM		G		Michael Brant		what do you mean 7/2/2004		1:47:31 PM		Michael Brant		G		and when he said all the people who watch MTV are going to hell - that kinda was it for me 7/2/2004		1:47:42 PM		Michael Brant		G		I mean - that most of his sermons he gets from somewhere else 7/2/2004		1:47:49 PM		Michael Brant		G		so those aren\'t his opinions 7/2/2004		1:47:53 PM		G		Michael Brant		he never said that 7/2/2004		1:47:57 PM		Michael Brant		G		he\'s just re-iterating someone else\'s opinions 7/2/2004		1:47:59 PM		G		Michael Brant		or meant it the way you took it 7/2/2004		1:48:03 PM		Michael Brant		G		he did say that - at Sunday school at XSCAPE 7/2/2004		1:48:09 PM		Michael Brant		G		err... 7/2/2004		1:48:18 PM		Michael Brant		G		when we were at Hilltop 7/2/2004		1:48:20 PM		G		Michael Brant		but he didnt mean it like that 7/2/2004		1:48:21 PM		Michael Brant		G		in that one classroom 7/2/2004		1:48:25 PM		Michael Brant		G		thats what he said... 7/2/2004		1:48:29 PM		Michael Brant		G		not even like I perceived it wrong 7/2/2004		1:48:41 PM		Michael Brant		G		\"If you watch MTV - you\'re not really a christian\" 7/2/2004		1:48:44 PM		Michael Brant		G		or something to that effect 7/2/2004		1:49:02 PM		G		Michael Brant		well hopefully ill see you tonight man i gotta get some stuff done 7/2/2004		1:49:13 PM		Michael Brant		G		you wont see me tonight - but bye 7/2/2004		1:49:21 PM		G		Michael Brant		alright see you later  '),(17437,2413978,'29,June,2004','2004-06-29 00:00:00','   I took a nap from like 6 - 9 cuz I was super tired and since I got up my Mom has been a naggy fuck. I swear: When I first got up, after I took a nap because she was going to \"borrow\" the computer from me for a short time - three hours later, before she would give it back to me for studying (I swear she takes it away from me whenever I\'m on it for her damn Avon) she made me vacuum and \"do something around the house\" cuz I haven\'t done anything for \'months\' before she\'d let me back on!  Now shes walking around the house randomly screaming stuff at me, randomly telling me things to do and randomly accusing me of stuff. Right now she said I put two loads in the washer and if I ever do it again she wont let me use the washer anymore and I\'ll have to just wear my clothes dirty (she\'s done this in the past, so I dont think she\'s kidding) and her washer is a \"double capacity\" so I did  not  overload it.  Right now I\'m overly pissed.  And I just found out this person I have a mad crush on will never work out...which makes me sad. Cuz they dont feel the same way about me. *frown*  Talked to my dance teacher today about some stuff, she\'s apairently a more laid back super smart person too - she told me she was taking college classes at the same age as me, that she took AP classes and all that and she had some curious recomendations. She suggested that I just take AP classes to get the weighted GPA, but not worry about the tests because they\'re dumb, and you have a \"chace\" of passing or failing the test and not getting the credit. She said in college you actually have something to show for your efforts, like selling your text book back, etc - and that a semester is equivelant to what a year in high school gives you, meeting twice a week rather then everyday. So that sounds better to me.  Tonight we went to Costco where my Mom FINALLY bought some ink cartrages (shes too cheap to buy ones for the *REAL* Dell Printer - so she settled with ones for the crappy Epson printer) and I was looking at books - thinking of the ones I\'d like to buy and I found out that I\'m broke (underdrawn) and my Dad is going to (figuratively) kill me when/if he finds out, so I need to get my financial dealings under control (use self-control) so this doesn\'t keep happening.  We went to Lolitas for dinner and got the #1 (Two Tacos, Beans, Rice) and I still am really surprised how something so simple is the best thing, in my opinion, that they have there. I used to get all this wierd fancy, expensive stuff and now I found my favorite thing (Although I think I\'m getting Carne Asada Fries next time - I\'m due for Carb Overload.)  Well - off to studying.  Out-  '),(17438,2413978,'28,June,2004','2004-06-28 00:00:00','   Today has been a crazily busy and full day.  I started in Psychology with taking a bunch of notes on the eyes/brain relations and how they work together and then being confused for a long time. I still dont get it now, so tonight I\'m going to read the text, try to make sense of it and then hope to do good on my quiz tomorrow. Right now my standing grade in the class is an 89%, after my exam which I scored 89% on. The teacher said she would give us 25 points free if we go see Farenheit 9/11 and show her a ticket stub, so even though thats a dumb biased political propaganda movie, I still think I may go see it becuase 25 points is 25 points.  http://www.togzone.com/nuke/modules.php?name=News&file=article&sid=46&mode=&order=0&thold=0 is a review on the movie from TogZone.com (My website), be sure to check it out, rate it and then leave comments.  Dance we continued our chair dance to Usher (Im not too big on him) and she worked our asses off. I\'m sore all over - and will probably take a bubble bath to relax my muscles tonight. As much as I love the teacher, Mrs. Macias and the class, I honestly cant wait until its over in 2 1/2 weeks because its such a load.   urlLink Personal Blog  has been updated.  Out-  '),(17439,2413978,'28,June,2004','2004-06-28 00:00:00','   See the next post down, then vote for this picture.  urlLink  '),(17440,2413978,'28,June,2004','2004-06-28 00:00:00','   Today I posted up a new picture of my cousin that he gave me to be rated, so the race is on between us. I also posted a new pic of me - less formal - it is from D-Land.  So tell me what you think of the \"sexy\" cousin:   By going to: http://www.hotornot.com/r/?eid=OLOMS8E&key=TLR   And you can always vote for me on the right hand panel right above my profile, with my displayed picture.  '),(17441,2413978,'28,June,2004','2004-06-28 00:00:00','   This weekend I went to Phoenix Arizona, which, this being summer, was 107 or higher everyday - and amazingly I enjoyed the heat - just sun bathing in the pool area of our 4 star resort, The Legacy Golf Resort outside of Phoenix Arizona.  The first day we were there, we had gotten there around 10 (left home at 4:30 am!!) my Mom called her cousin, Cathy and later at night they came to our resort and took us to dinner. What really sucked about the dinner, is my Mom didn\'t know they were going to pay and we ordered this dish that was pretty expensive (and nasty) then we were all like \"oh shit!\" when it was time to go cuz they caught the bill, then my Mom offered to pay the tip and didn\'t put enough, realised it later, and was even more put out.  The resturaunt was a very western steak place equipped with a Bull, a slide down from the bar to the dining area (piled with little kids) and some awesome bread and salad (bread rocked, BBQ sucked) and they had this cool cinnamon butter junk to go on the indian flat bread and on the biscuits which just made the dinner complete.  During dinner I asked my Mother\'s cousin what she does for a living and she told me she used to work for Sprint as a lead over several states until she was layed off and now she is a median, she communicates with the dead, she goes into people\'s previous lives and removes negitive energy, she clears the constructed areas within people\'s energy meridians, she does accu-pressure and she does transcendual meditation. She went on to explain how she communicates with spirit guides and \"descended master\" (fancy name for a Demon...) and it really creeped me out. When she went on to say that she \"communicates wth the dead\" her voice even changed - it went down an octave. So as much as I dont really want to do the God thing anymore, I still totally believe God exists and Jesus and the whole bible thing, and definatly demons and angels - so it was really wierd. (Back to my beliefs - I just have a lot of conflicting beliefs about doctrine and stupid people within the church)  Anyways, after our dinner thingie we went back to the hotel and just chilled and then went to Bed.  Saturday for the most part was uneventful. We rented a car, got to take a really fancy car to the rental car place (free of charge) and then we got this piece-of-shit car that the Air Conditioner barely worked and then went to the grocery store and bought some junk along with somewhat good food and then went back to the room, took a nap, chilled, Mom cooked - etc. Then that night we went to the pool, chilled by the pool, got and then we went to the bar and got some Nachos and Pina Coladas (I bet they woulda sold me alcohol if I woulda tried since I look old, but I didn\'t want to chance getting arrested or my Mom finding out and me getting in big trouble with her - so I didn\'t try it - just went with virgin Pina Coladas - which are still good)  Also on Saturday we went to this mall and saw Stepford Wives and ate some Nathans Hot Dogs. Also I went to the Virgin Megacenter place and bought a book (wanted to buy a load of CDs too) and then thought I shouldn\'t have bought the book cuz I\'m broke so I had my Mom take it back and play along with me about how I shouldn\'t have bought the book and she\'s all disappointed and I\'m in big trouble and she was able to get my money back...which was good.  After that we went to a 99 cent store where I found some great books super cheap and got them for a buck each. Then we went back to the resort and did the above explained pool vs. bar thing.  The next day we went to my Mom\'s cousin\'s house in this rich-snobish part of Phoenix and went in the pool and then she cooked for us and we had a truly awesome lunch. I sat around her pool and read my book \"Watership Down\". I could have been playing with her kids, but they\'re 6 and 10 and totally not my age and that would have been strange, so I just read. At one point they did come join me in the pool and we played some in-water basketball. The 10 year old caught the rebounds for me and the 6 year old was my cherring comitte. After that we had these huge juicy pork chops, brownie w/ ice cream, twice baked potatos, etc and it rocked.  Then after that we drove 30 minutes or so back to our room and this guy Brian Agosto, an awesome guy that I grew up with way back when, came with his wife to visit us and we got to catch up with him and find out how his life is doing now and all that. Most of his conversation was about his older brother and a lot of religious stuff with he and my Mom so I didn\'t really get to talk to him much, but he is still really awesome. As kids his little brother and I were best friends and used to always be together, but I dont really remember his little brother much because they moved away when I was like...8 or 9. But we did, all-in-all have tons of fun together way back when, I just forget the details. Anyways- Brian is 21 now, getting formally wed in September with his wife (they had one of the \"you\'re legally married now says the state\" kind of weddings and then are having the wedding with all the family and stuff in September.)  I remembered Brian as a really religious dude, super involved in church and stuff when we were kids and now its really awesome to see him way more laid back and stuff. He even cussed a couple of times, talked about be careful you dont sleep around in Brasil because people have a lot of STDs...etc - so that was awesome to hear/see.  They took off around 11, we went to bed and then got up at like 7 to get all packed up and stuff and then take our flight home. At the airport there were these two super ugly dudes that seriously looked like terrorists and had me a little scared at one point but our flight was fine and then we got home and got a ride with Lora and then dropped off my stuff at school, grabbed my school stuff and jumped in the car.  I barely got to my Psych class on time, but got there, and then we took a test (dont know how hot I did on it - had to do a bit of quessing on some of the outside-of-the-box answers) and then now here I am.  Exciting weekend huh?  Out-  '),(17442,2413978,'09,July,2004','2004-07-09 00:00:00','   Well, tonight I went back to Xscape after about...6 months and it was a surprise from the beginning.  To start I was talking to Linda, who couldn\'t get me a ride, so she said she would call Gabe to get me a ride, who called Chris Wong to get me a ride who called Linda and suggested Joel to give me a ride. So in the end Joel picked me up, which was a big surprise but was cool.  Joel seems a lot less on fire then he was in the past - but you cant judge a book by its cover - so I dont know that. But he was talking about girls, political movies, etc - he didn\'t have that whole \"everything is about God and conservativeness\" attitude that I remember him having. But it was cool to get a ride from him, its been a long time since I\'ve seen him.  Right when we got there, he started talking so I just walked up the stairs. I guess he didn\'t feel compelled to hang out with me, since its not like I\'m new or something. When I got upstairs, they have those little poles with the ribbons stretched from one pole to the next, like to make a line (not sure what they\'re called, but like they have to get into a club or something - to herd the people) leading to where you get name tags at *snort*.  I saw Priscilla there, and she has really changed. I\'d hate to say that she\'s been brainwashed...but it sure looked like it. She has turned into one of their flunkies - like I used to be (again - my perception.) She gave me a name tag, acted excited to see me (dont know if it was genuine or not) and then explained their new rule that you cant go to the bathroom after it starts (what!?)  So I walked into the main room thing, saw these two girls that I went to Elementary School with that recognized me (I recognized them after they asked me if I remembered them) and then I walked in and got a seat, after hugging Linda (probably the only person there that I know for sure loves and cares about me - the rest may - but since they never kept in contact or anything - its unlikely.)  I went inside and sat alone - because I had my eyes dilated today and didn\'t really see anyone right away that I knew (and really there wern\'t that many of them there.) I\'ve concluded that they either have left (not a huge surprise) or they became middle school leaders and were there earlier and since went home. I cant picture Elias and Ishmael leaving since they have been converted from catholic to protestant and really regard Corenrstone high - so they probably became leaders or just wern\'t there tonight. Elias was one of the people I was closest to - but it was almost an unspoken closeness - we didn\'t talk that much but both seemed to enjoy each other\'s company.  Well - Xscape now has this video at the beginning that has \"the rules\", which includes a bunch of random (illegal and pirated) video clips, a bunch of songs (someone likes Twista) and it was dumb. That new rule \"You cant go to the bathroom - no matter how bad\" is the stupidist thing I\'ve ever heard. I understand the concept behind it - it is disruptive to the speaker to get up and go during them speaking - but its still really dumb. They sell all kinds of various foods and drinks. They\'re  asking  for people to have to go to the bathroom. Whats up with that? And what if someone was sick and needed to take a dump? I personally have a lot of stomach problems and never know when I\'m gonna have to just go in a seconds notice - so that just sucks.  The rules also went on to say that \"if you dont like these rules - we can escort you to the exit.\" God, if I would have been new and not known how they really are, I probably would have walked out. That is really stupid. And I bet anything I knew who came up with those rules too...  Well - the video at the end had a five second count down - then there was the dramatic entrance of the worship team - which made everyone come to the front of the room and then they have  no  worship anymore. Everything is high paced praise. They want everything to be happy and fun I guess - but thats not reality. How is there reverance in all that praise?  It was interesting that several people came over right away and said whats up to me. I mean, it didn\'t exactly surprise me, but it was just surprising...if that makes sense. I mean - it was real - that was surprising - but that they came over wasn\'t surprising. Daniel ran right over and Jennifer stopped me in the hall and said hi. She even commented on my cartilage piercing (which I think the left part is infected.)  Right after worship I glanced over and saw John, Randy\'s (this former youth leader who really reached out to me and was thus really hurt by Cornerstone) little brother. After everything that happened with Randy (the pastor telling off his Mom over the phone) I\'m really surprised to see his brother.  Michael (youth leader) spoke on the prodigal son and there were several comments that other people didn\'t really catch that I caught. There was a comment about how people dont come to Xscape because they dont agree with how its run (so I guess I\'m not the only one...or maybe that was aimed at me) and then he made another comment, toward the end, about how he\'s seen fired up youth leaders just...fade away over time and get lost because they try to deal with their issues on their own (which is totally me.)  He forgot to add on there that Cornerstone tells you to go to them, and not God to get help with stuff. But the message was indirectly for me I guess. It was about coming bitter and stuff and turning away from God and how to know if you\'re backslidden or lost or what and how you can get back.  Right now I\'d really like to get back straight with God, but I have a couple issues. The generally uneducated-in-worldly-matters Christian community has no knowledge of science and political matters - so they automatically go with what someone tells them (classical example of brainwashing) and Christianity in general doesn\'t give credit to science.  Two - I like my sin. Sad to say - but I\'m being honest. I like my sin - I would have to give it up to God and not do it anymore if I wanted to get back straight with him.  Three - I have a lot of issues with Cornerstone. I\'ve been very hurt in the past and right now my heart it too hard and I\'m to bitter to let them help me again. But tonight they did say to go to God with your problems - which I guess I can do. All they\'re gonna do is tell me to work my ass off for them and it will all go away - so I\'m not gonna bother.  Four - yeah, I haven\'t been going to Xscape because I dont agree with a lot of what goes on there. Those stupid rules aren\'t going to help that. All of their other rules were a given. Whose gonna drink alcohol, smoke cigarettes or do drugs in church (come on.) And most people aren\'t going to cuss because they\'re being someone they\'re not to impress the church people. If I wanted to cuss, I would cuss no matter where I am. God see\'s it anyways - why pretend to be someone I\'m not for church people who are themselves probably doing the same thing?  The thing that surprised me the most today was Richard. I was expecting him to come up to me, act all excited to see me, tell me how much he loves me, etc, etc because either he knows how I feel about him or just he knows my reasons for leaving.  I don\'t dislike Richard in the least. I think he\'s an awesome guy, but in a lot of ways he\'s way overboard - which I dont like. And I dont agree with a lot of how he runs things - but thats his job, not mine, and his choice to make, however I am entitled to my opinions and I decide if I go there or not. But it was cool to see that he didn\'t come over and pretend to like me. I mean, I was trying to avoid him, but when I was leaving, he grabbed me, hugged me and then I was on my way. He didn\'t try to get mushy, he didn\'t try to talk to me - and that gave me temendous respect for Richard. I didn\'t really have respect for him before now. I mean he were there, I love all three of his kids, so I couldn\'t obviously totally dislike him - but now I have that respect. He wasn\'t trying to force me to be someone I\'m not, to do something I dont want, etc.  So even though, if possible, it got worse, I\'m gonig to keep going back. I dont know if I will ever get back to where I was, or if I want to ever be a leader again - just because I have to be someone I\'m not for a cause I dont believe in and be brainwashed - but I may be able to lead a dance thing.  Also...very wierd - you have to have a leadership certification... The beginning \"leadership\" positions are vacuuming, operating a camera, giving out nametags, etc. What is there to be certified for? I\'m guessing its some kind of brainwashing session where they tell you how much God is going to bless you for working for them. You\'re going to get a special place in heave, etc, etc. I was never subjected myself to that, but what else could it be?  Well - I\'m gonna stop bitching now. I\'d like to talk to Gabe and some people there just to get an idea if they still love me or if its a game. Gabe and I used to be pretty tight and then through the time I haven\'t been there he\'s gotten pretty distant. He doesn\'t even answer e-mails and stuff anymore - but then they\'re gotten pretty more advanced then they were. They have some powerpoint like program where the background is animated (no doubt pirated - like everything else.)  On last thought - its really ironic that I\'d have to be certified if I was to get heavily involved again. I see the irony because I would be willing to pay that there is no one in that church more advanced then me when it comes to technology - especially computer related. They didn\'t even know that they could import slides from other presentations such as lyrics into one large powerpoint presentation. They had individual presentations for each song, for the message, etc - so I taught them something as basic as that. Other then graphic art and maybe video editing and possibly the hook up or sound equiptment (which is more of just the knowledge of how something is assembled rather then actual knowledge) there would be no one who knows that I know about technology, yet I would have to be oriented (certified) as well...thats pretty whacked up.  -Mb  '),(17443,2413978,'08,July,2004','2004-07-08 00:00:00','   Well - even though I\'ve been hurt by Cornerstone a lot in the past I\'m going back to their youth tonight, so face people I\'ve burnt bridges with and then friends who I haven\'t heard with - I dont know if I\'m like a dog going back to it\'s vomit or something - but maybe this will be good?  I\'ve always enjoyed Xscape in the past and I guess all I can hope for is that the people I\'ve burnt bridges with don\'t hold grudges...  I dont know if I have a way of getting a ride - so I may be taking the bus and then walking like 1/2 - 3/4 a mile...but oh well. I guess if all goes wrong I can just leave - I\'m kinda nervous about it actually.  Speaking of burnt bridges, on the way to the Optomitrist, my Dad was telling me that last night when he almost severed his finger while cutting an onion, he had to go to the ER, and then this neighbor that he\'d been at it with over parking happened to be the head ER doc...man would he have been in bad shape if that doctor would have been the one who had to stitch up his finger - so I guess that tells me to be civil to people I dont like.  This eye doctor I had today I wasn\'t too impressed with. She didn\'t seem to really care about her patients - she was just doing her job and doing the list of stuff they\'re supposed to do. I miss the other doctor I used to have, Aaron Lech. I just tried looking up an e-mail for him, but all I found was an Address and Phone number - which probably doesn\'t even exist anymore, since he\'s in the Navy and has probably been transferred to another hospital by now - plus I would seem very stalkerish if I was to call...so I guess unless he gets an e-mail thats listed, I\'ve lost contact.  What really sucks is Dr. Lech was willing to go all the way with my vision - he lied about a non-existant disease to get me free contacts and stuff from the military and I was a little kid who didn\'t want to be bothered back then. Now I have this conceited female civilian doctor who doesn\'t want to be bothered...ick.  Well - yeah, off I go.  Out-  '),(17444,2413978,'08,July,2004','2004-07-08 00:00:00','   According to the Red Book of Westmarch, In Middle-earth, Michael Brant was a Virtuous DÃºnedain   Elven Name Possibilities for Michael Brant The root name suitable for feminine and masculine is: Ascachil Another masculine version is: Ascachilion More feminine versions are: Ascachiliel Ascachilien Ascachilwen   Hobbit lad name for Michael Brant Andwise Brown from Whitfurrows Hobbit lass name for Michael Brant Melilot Brown from Whitfurrows   Dwarven Name for Michael Brant NÃ®m Coaltooth This name is for both genders.   Orkish Name for Michael Brant IshhÃ³sh the Insane This name is for both genders.   ** NEW ** AdÃ»naic name for Michael Brant ** NEW ** ZabazÃ®r     Find yours at: http://www.barrowdowns.com/middleearthname.asp?Size=  '),(17445,2413978,'07,July,2004','2004-07-07 00:00:00','   Today in Psychology it was interesting. We were doing learning and the psychological principals behind learning. We also looked into conditional and unconditional responces and stimulus.  As one of those examples, the teacher showed us a clip that had this kid just screaming and freaking out in a store because his Dad wont buy him candy. He was screaming, rolling around on the floor, throwing stuff, etc. He goes on and on and the Dad looks hopeless. Then in big letters over the screen comes up \"Use Condoms.\" My whole class almost died laughing.  The teacher was saying that if you\'re not 18 you have to get outta the room or something. And then she asked is anyony under 18 and like 4 of us said \"I\'m 15\" and the teacher was astonished. But then she concluded that we had a form signed by a parent and they know that College is for adults coming in and then we didn\'t have to leave after all.  Dance was awesome but hard today. We added a lot onto the Lyrical Britney Spears dance which is turning out to be really pretty. Its a dance that I completely have so far, compared to the Jazz Dance which is evil and I dislike a lot.  I found out Mrs. MAcias is gonna teach full time at CPH - teaching PE, which is so awesome it makes my whole day. Again I think shes like the coolest teacher in the world and one of my favorite people - and now I get to go bug her when I feel like randomly getting out of class *whee*! I\'m hoping to be able to be her aide or something I\'m going to try and do it this year but may need to wait until next year.  New post in Biting My Nails  urlLink Click Here  '),(17446,2413978,'06,July,2004','2004-07-06 00:00:00','   I seriously feel for my Dad\'s mental state or lack-there of.  Last night we were watching Cold Mountain and at the end of the movie he was like \"this was such a deep and powerful movie - I\'m surprised you have an interest in it\" *dies*. Does he not know me at all? Does he forget how smart I am (way smarter then himself)? I swear, not braging or anything - but I\'m in all honors, have been way above all my peers intellectually since I was like 5...and he doesn\'t understand how I could appreciate a \"deep\" movie. How old does he think I am? 5?  And a lot of times I tell him I saw or want to see a certain movie, his response - \"Thats an adult movie son\" just like that! But then he\'ll let me watch American Pie and dumb comedies. What are those? Wholesome family entertainment? The whole fucking movie is about sex! I\'m looking for a movie where he said that... I think he said it to War of the Roses or something like that. That movie was only like PG...and who cares what the messages are in it? If they\'re so \"adult\" since I\'m like 5, wont they just go  right over my head  or something?  Grrrrrrrrrrrrrrrrrrr...(so on and so forth)  -Mb  PS: My friend/cousin (we call each other cousins, but really are not related by blood) just flew out to Florida this morning. She has  very dysfunctional family and now that she\'s 18 she left. I guess thats admirable - but I wish she could have just got out on her own here in Cali. Why go clear to Florida where she knows no one? Oh well - I give her 6 months before she\'s back. She\'s never really had to work - although she was the slave at her house, when she wasn\'t slaving she had a really mushy life. She\'s never had to provide for herself. Just wipe her Mom\'s smelly ass (literally).  '),(17447,2413978,'05,July,2004','2004-07-05 00:00:00','   Today I went to the movies with Lee and my Mom before Lee goes to Florida tomorrow - since she is moving there cuz he life sucks.  First off we went to Lolitas and I got a Taco and then we went and saw the notebook. After that I came back to my Dad\'s.  And then I came on the net and got in a fight with one of my best friend online contact people - Michael M. over something really stupid. Apairently between my comment that he went over his limit in space on his domain (I host him free and he was 17 mbs over his limit - or 170% more space then I gave him) and him breaking up with his girlfriend - it sent him over the edge and now he\'s having some kind of fit... He just said something about how he doesn\'t want to be five different people anymore - he doesn\'t what to explain what he means by that and now he isn\'t talking to me...wierd. Maybe multiple personalities?  My REVIEW of THE NOTEBOOK:   I went to see The Notebook, a movie derived from a book written Nicholas Sparks â€“ the dramatic love story between two lovers, Allie Nelson (Rachel McAdams) and Noah Calhoun (Ryan Gosling.)  It starts out in a Nursing home with a senior version of Noah Calhoun is reading a story out of a â€œnotebookâ€ written by his now delirious wife, Allie Nelson. Through reading the story of how they met and how they fall in love, Allie has bouts where she remembers and then slips back into her delirium and freaks out, at one point physically hurting Noah.  The director of this movie had no problem showing us how hunky Ryan Gosling is, in several love scenes. The main theme of this movie was two teenagers falling in love in a small North Carolina town. Allie was a rich girl living in Charleston, North Carolina who goes to a small North Carolina town on the ocean, to live in her summer house. Noah is a poor, simple (sexy!) country boy who doesnâ€™t have two nickels to rub together but wins Allieâ€™s heart through several social events. Of course, Allieâ€™s snobbish parents do not approve of someone who only makes 40 cents an hour and pull Allie and Noah apart.  Once forced back to Charleston, Noah writes 365 letters to Allie, one per day of the year. Allieâ€™s mother, in Charleston, takes every one of her letters from Noah and hides them â€“ so because of a bitter break up, figured Noah wants nothing to do with her and moves on in life.  The movie was very sad in parts and had me crying more then once. But all in all was a very romantic chick-flickish movie. I would defiantly recommend it.  '),(17448,2413978,'04,July,2004','2004-07-04 00:00:00','   Just saw on the news, since Dad has it on about 24/7, that they are making power cells out of Spinach. They already have a spinach cell battery that will power an average laptop for three weeks - but apairently they\'re only 12% effective and they fun off of photosynthesis - so no sun - no power. They\'re gonna have to get the spinach to run the Krebs Cycle when theres no sun (which is a chemical cycle that produces energy.)  I think this is really cool - its going to make power more efficient in time, completely natural, gonna help the environment and there will be so much less spinach for me to have to eat - since they\'re gonna be making batteries out of it.  Out-  '),(17449,2413978,'04,July,2004','2004-07-04 00:00:00','   Today has been interesting. At 12 we left for a Padres game at Petco Park. I really had a bad attitude about going because I\'m not a big Baseball fan. But the park was awesome, they had a cute black lab puppy that I wanted to take home, I actually understood a little of what was going on in the game and I got a lot of food. LOL  We had a couple of wierd characters around us - this dude that\'s fat hung so far over his seat that I was crowded and had no leg room. I swear his fat hung a good foot over the chair and he was right in front of me. He moved a lot though - and finally because we were getting so sun burnt me and my Dad moved to the back section of where we were to get in the shade.  The highlight of the game was this flag that took up most of the field and then the jets flying at mach 5 over us. Was loud - but cool.  After the game we went to Applebees where I got this nasty orange chicken platter thing and my Dad got all you can eat riblets - but all he actually ate was the ribs of one helping - he didn\'t even touch his fries or cole slaw (that best part) but got all wierd and angered when I suggested that one order was \"more then you can eat\" and just a regular meal was cheaper - so I left him alone and let him waste his money.  After we were done eating we walked around the mall a little, I found the Piercing Pagoda and then got my cartilage on both ears pierced for only $13 (including an extra bottle of ear care stuff.) Was a great deal. And I think the earrings are gold plated to prevent infection - so thats even a better plus. Although I was no in extreme pain and still am not - the cartilage hurt a lot more then the lobe (yesterday) hurt me.  After the mall we came home and I replied to a load of random messages on this mailing list I\'m on (http://groups.yahoo.com/groups/Any_Discussion) which lead to the inspiration to write an editorial (I dont really like it too much - it was inconsistant as a friend put it - and I\'d have to agree) but heres the link if you wanna read it:  http://togzone.com/nuke/modules.php?name=News&file=article&sid=60&mode=&order=0&thold=0   I unfortunately had to go see fireworks, which wern\'t too impressive, before I got to write my editorial though. We went to the building my Dad works at downtown and went to the roof to watch the fireworks and he was being all paranoid that me, his \"little baby\" was gonna go walking off the side of the building or something. I swear he treats me like I\'m 5, when in all reality I\'m probably more intelligant and more mature then he is in many ways. But he and this other engineer had to raise the flag and had to keep going back to the flag to try it again because right when they\'d think its safe - the wind out pick up and they didn\'t want to get sucked up in the air. When they finally did it, I walked within 50 feet of the harmless flag pole and my Dad said if I dont go back to the way other side of the building - we were going to go home. Well - first off, since I was certainly 10 feet atleast away from any potentially dangerous part of the flag pole - he should have been all happy that I\'m interested in what he does for a living. Then I started the drama - when he was done I was like \"You know - how you\'re threatoning me that we\'re gonna go home - you say that like I actually  want  to be here, and that is some punishment to leave - when in all reality it would be a favor\" and you could see it on his face instantly that he was pissed. Whats really interesting is that he didn\'t blow - and he has major anger issues - so he excercised a lottt of self control in not doing that. I guess he already is consided a bitch by his co-workers cuz he talks loud, is wierd and thinks he\'s all that - so he didn\'t want them to think he was absuse too...so basically I knew I went over the edge, got a reaction which is what I was looking for and had to give him some BS line about not meaning it so I dont get introuble and not be able to write my edtiroial when I got home. So we made up, I was miserable through the completely over-rated display of fireworks and we came home.  New post in \"Biting My Nails\" -  urlLink Click Here  '),(17450,2413978,'04,July,2004','2004-07-04 00:00:00','   Until recently I\'ve never remembered my dreams. Then in psychology we talked about dreams and sleep patterns in general and the past few nights I\'ve remembered my dreams - so I\'ve decided when I remember the dreams long enough to enter them here and then have time to do so I\'m going to keep track of my dreams. Should be fun.  Last night I dreamt that I was in Mammoth Lakes CA on the ski resort in July and Mammoth was way smaller then it actually is, yet there were signes that said Mammoth (dont ask). I was skiing dont some last remaining snow, right along side a grassy area (I vaguely remember something about my Dad\'s explorer being up there with me) and I just remember a lot of grass. And then in the end, I skiied down the hill until I was in complete grass and then the dream just ended...  I think we were at the top of the hill or something.  I think I dreamed about this dream because I was looking at this Mammoth book my Dad had. Their monthly newspaper or something and it just caused me to reminisce about when I went skiing there. Only real definition other then it was a just a completely random dream. Only wierd thing is that the place I was in looked nothing whatsoever like Mammoth Lakes CA and there was very little actual snow + plus it wasn\'t cold.  Then the other night I had a dream that I was laying in my living room floor at my Mom\'s on my phone then I woke up at 12 years old with my best friend sitting in my chair at the dining room table assembling some sort of food while my Mom was cooking it and then I looked like I did when I was 12, except internally I was 15 and then I used my cell phone - that somehow worked when it runs on completely different signals then they even had back then (but then its a dream - they\'re not legalistic) and I called my English teacher to see if she knew me to find out how to get home. I dont remember our specific conversation though. I guess I associate my English teacher as a very smart and very informed individual who could help me - which she probably could have. So I think that is symbolic as someone I can maybe turn to for help on life matters? And then I remember just...ending the dream. I dont remember if she knew me or anything.  Well off to the Padres Game.  '),(17451,2413978,'17,July,2004','2004-07-17 00:00:00','   [The Forum] Charity: WHOA.&nbsp; The alarm of the Stadium in YuYuHakusho is the same sound that Mario makes when he jumps in Sumper Mario! :D XD &nbsp; [The Forum] Mike stares. &nbsp; [The Forum] Charity: Random bit of information for you.&nbsp; Just remember, if you ever go on a game show and are asked that, you\'ll know! XD &nbsp; [The Forum] Mike adds this to his semantic memory.  '),(17452,2413978,'15,July,2004','2004-07-15 00:00:00','   \"My dog is worried about the economy because Alpo is up to 99 cents a can. That\'s almost $7.00 in dog money.\"  '),(17453,2413978,'15,July,2004','2004-07-15 00:00:00','   Well its been interesting lately. Been as hot as hell to start. Today it was 91 degrees farenheit, 100% humidity and the heat/humidity index (according to weatherbug) was 123 degrees...no wonder I feel so ick.  &nbsp;  Right now its really nice to not have the extra weight carrying around to have to contend with and be more hot (if I was as fat as I used to be I\'d be worse off emotionally by the weather.) Both of my parents have been totally evil the last few days because of the weather and its getting annoying. Tonight my Dad got paged and was in the shower so he didn\'t hear his pager and when he did come out he was freaking out about it...*sigh* dont know whats up with him. Then he called work and told someone off or something. I think its humorous but sad.  &nbsp;  Today my Dad picked me up at my Mom\'s house and then we went to Sports Authority where I bought some resista-bands to go along with my Yoga-Pilates workout and they\'re really helpful. All the people at Sports Authority were evil though. I dont know if they just dont value their customers or if the heat was getting to them too (but then the place is air conditioned...so I dont think thats it.)  &nbsp;  I was gonna go to Xscape tonight and listen to Gabe speak and stuff but I concluded thats its too damn hot plus Xscape is upstairs, so that would be a really bad deal since heat rises and there are enough sweaty/hot people up there just because of the intense action and the friction as it is without it being 90 with no AC.  &nbsp;  Also today, on a better note, my check from Blurb came. I finally got my $20 and this is a really good time to get it.  &nbsp;  Also today I\'ve been trying to recruit some people to come to BMG Music because I\'m running low on free CDs and I need to replenish my stash.  &nbsp;  On&nbsp;a sadder note, this dude Michael M. who I\'ve been internet friends with for a long time kinda broke it off with me today cuz we were arguing and he was being stupid and threatoned to block me (wrong thing to do) so to show him that he has like no persuasion over me I said go ahead and block me...and he did. Kinda sad deal - but I\'m sure he\'s gonna unblock me in time. I\'m not about to apologize, cuz he was the one being stupid. He\'ll probably need me for something and IM me. Then I can be all hardass on him.  &nbsp;  Tonight my Dad and I went to La Bella Pizza and then got this one kinda pizza with mushrooms, onions, pepperoni and green peppers and then went to Blockbusters and he got mad in the car cuz I wanted to rent \"Mean Girls\" cuz I like Lindsay Lohan...and left blockbuster and drove home *sigh*. I guess I can blame it on this heat?     Then he came home, turned on the AC and then went to bed. So I\'ve been just up messing around. Have a&nbsp;psych quiz&nbsp;on Monday so I\'m gonna have to get studying sometime this weekend to be ready for it. &nbsp; Also talked to one of my best friends, Isaac Martinez and he\'s been telling me he went to UCSD for this upward bound program thingie to get high school credit and the whole under-under-grad experiance thing going and apairently he\'s miserable. He figured it would be all laid back and he said its constant homework. And what really sucks is that he doesn\'t even get college credit for it. I think it counts toward his High School GPA. And it sounds like the classes are totally accellerated, where they get a whole credit for the classes (a whole year) in 6 weeks. So its interesting nonetheless. &nbsp; He\'s also going out with this girl that I told him I liked - its not the first time he did that. I told him I like someone, he gets to know them to help me out and then I dont do anything and he ends up liking them himself and taking them. It sounds like they\'re pretty happy too. Movies and stuff. I guess thats cool. I\'m pretty happy for him. &nbsp; I also told him about this one guy that asked me out at SWC (he\'s very naive and homo phobic) and I just said that I told him \"thanks but no thanks\" and I told him even though it was a guy it was a total compliment to be asked out cuz someone thought I was hot (if he only knew). And he went on to say that he\'s glad he wasn\'t there cuz he would have beat the guy up *rolls eyes*. I explained that that would be 1) a hate crime and 2) assault and battery. I think its sad that in this day and time with all the gay marriage and gay rights movement that he is still so conservative in terms of his homo phobicness. But then this is the latin community of southern california living back in the 40s and stuff. \"stone the fags\" so to say. &nbsp; Oh well. &nbsp; Peace- &nbsp; PS:  urlLink Biting My Nails &nbsp;has been updated  '),(17454,2413978,'15,July,2004','2004-07-15 00:00:00','   Need to be concise here because I dont have that much time and I need to go to Admissions in a little bit and register for a course, but today we had a written dance final and then just kinda hung out and watched our dance performances on past tests and did&nbsp;a mild workout. &nbsp; In Psychology I found out that my overall grade is up to 100%, which is a great thing - and I also found out that my Mom has had someone spying on me (Denise someting) in my dance class. The whole semester she has been reporting my activity to my Mom...*grr* &nbsp; Also I found out that my Modern Dance class starts on August 16th - so I\'m going to have very little time off of school this summer. My psych class ends on the 29th, I go on vacation until like the 15th, then I start Modern Dance on the 16th and then on I think the 17th I have to go to this thing at my s chool to get trained for freshmen orientation (I\'m in Link Crew) and then on the 23rd I have to go to High School for myself and then on the 7th of September I start High School full time while maintaining a college course on Tuesdays and Thursdays. Should be an interesting year. &nbsp; Mom and I are still at each others throats I think because of the weather - and I just finished my book. Yay me! The Amulet of Samarkand was a really good book (My Review  urlLink Click Here ). My book Silenced by Jerry Jenkins came in the mail yesterday and thats next on my list to read. &nbsp; So yeah - need to be gone now and go register. &nbsp; Ciao - Love and Lobsters - Peace &nbsp; PS: In 6 weeks of dance I\'ve lost one pant size and one shirt size (down to a 38 waist and down to a Large shirt) but I haven\'t lost and pounds. I still weigh 240. So I guess I need to start running or something *sigh*.  '),(17455,2413978,'14,July,2004','2004-07-14 00:00:00','   Well today I have my dance \"dancing\" final during which we danced both dances: the prop-chair dance to Confessions by Usher and a really pretty lyrical dance to Everytime by Britney Spears.  I was talking to Mrs. Mascias today and telling her about how I dont get along with my Mom and stuff and she said basically \"dude - you\'re a teenager - you\'re not supposed to like your parents\" so I guess thats settled? She also went on to (I think this was yesterday) tell me that I should go straight on to UCSD rather then going to Community College because I\'m too smart to be with the slackers and I\'ll probably myself get lazy at Community College. But she said if finances are a problem then go to SWC for GE stuff and lower division classes because she said Student Loans suck.  Also I wrote this letter about me being her Aide and she signed that off too. Thats awesome. I can be an aide for my favorite teacher. She\'s gonna have to get back into bitchy mode though at the High School (she\'s not a bitch at all - but she has to be all tough and stuff at the High School) because theres a bunch of un-motivates little freshmen (hah! I can make fun of them now! Like I would do that... fuck conformity) and they like...will be chewing gum which they can choke on during PE and stuff. I was telling her that she should make them do some dance stuff, but apairently she\'s not supposed to do that. I\'m hoping she\'ll sneak some in there because kids should be enlightened more with the arts. They should force everyone to take an artistic class. Dance, Drama or Music. They can easily opt out with Art 100 (equivelant with college).  Well - our final was good I guess. Cant say the same about my Psych test though - I dont think I did too hot on that one - but maybe I did too? But anyways - the dance test, I was a little ahead on some of the counts of the songs, but my technique was really good for the most part. I dont think I made any mistakes, except being early on the counts.  Psychology, the hot dude that is my current eye candy, when the teacher said sex is a \"need\" and not just a \"desire\" said \"I better tell me Girlfriend that. \'I need sex Baby\' and then maybe I can get it 4 times a day.\" That dude is funny. Too bad he doesn\'t like boys...I would be on him in a second. But oh well. Whatever floats your boat I guess. It seems all gay guys are those steryotypical fashion obsessed lunaticts whose life\'s goal is to be a hair stylist. But oh well - hopefully I\'ll find my someone. I just dont want a steryotpicaly hair stylist if I can help it.  Mom is still on the war path for the most part, but she seems to have calmed down. This terrible humid weather seems to be getting to her lately, but I\'ll be at my Dad\'s this weekend (dont know if that is a completely good thing) and will have a break from her.  Today at my Dad\'s I took some food to the computer, cuz he has this thing \"when I call dinner, get your ass over to the counter.\" It has to be all formal, but we never eat together at home. Probably haven\'t eaten together more then 10 times. So I dont get it. But back to what I was saying, he said \"I dont agree with you having food at the computer\" so I was like \"that nice.\" And continued eating. So he was like \"do what I told you!\" and I was like \"You didn\'t tell me to do anything\" which is the truth, so he got all pissed, told me \"what to do\" and then I complied. End of story.  Right now my Dad is freaking out. He\'s getting old and getting a \"double chin\" and the last couple of days he\'s been changing everything. Been walking a lot, been changing his diet to include a lot of meal supplements and replacements (he doesn\'t eat that much as it is.) What he doesn\'t understand is that 8 beers a nice isn\'t going to help his weight...thats like his caloree intake for a whole day. Even light beer (which he drinks - dont know why since it tastes like crap - but his internal organs and brain...not mine.)  Should be studying, but I\'m gonna be lazy and do it in the morning.  Peace-  '),(17456,2413978,'13,July,2004','2004-07-13 00:00:00','   My Mom has been a complete bitch the last couple of days, but then its been really humid and she doesn\'t do humidity or any form of heat well - so I guess I\'m not gonna let it get personal if she calms down once the humidity goes away.  She\'s been biting my head off, not letting me on the comptuer to do hmwk, and just bitching constantly. Last night she told me I was evil and I\'m bitter and all this shit and that if she didn\'t have to put up with me, she\'d kick me out (ouch). This is what heat does to senile old bitches.  Then yesterday we went in this grocery store to buy hot cocoa, but nooo - she cant just buy one thing - that would be impossible (I swear she had issues with food - its hard to get her out of grocery stores and no matter how overflowing our garage is in food - she still buys more) but I guess she was somewhat good...she only bought a couple of things - but most of it we didn\'t need.  Finally after like...30 minutes of me waiting by the car she came out (with  another ) bag of something... But it gets better. After that, with ice cream of all things in our car, she goes for like an hour to deliver her Avon shit...so needless to say, my ice cream was all soft and icky when I got home...and is probably ruined. Re-frozen Ice Cream sucks.  While she was on the road delivering Avon books to various ghetto apartment complexes, I did recieve a call from one of my cousins, which was really awesome. Made me feel loved. I usually talk to him online, and he hadn\'t seen me on in a while (cuz of my stupid mother hogging it - or randomly not letting me on it for no appairent reason) so he called me to make sure everything is alright. Again, that rocked.  But back to ranting and raving (need to get this all out so I dont get pissed at her and mouth off and then in trouble), I was sleeping in her room, on the floor last night, because theres AC (Air Conditioning) in there, and she wouldn\'t turn the fucking TV off! It was like 11:00...she wasn\'t even  watching  it...but she insisted on it being left on. Finally at 11:30, probably 5 minutes before the sleep timer was gonna shut the TV off, she  finally  shut it off, which made me happy because I could actually go to sleep...  I think in the future I\'m just gonna sleep in my room and roast to death - better than putting up with her crap. Also I had to listen to her tell me that I was no good, and always insult her (I pointed out, in all humility, that she used to beat me as a kid - which she denied, but got awfully defensive over) which I didn\'t, cuz it was true. And then she told me I\'m cynical (which I am I guess, but she\'s way worse then me - and doesn\'t even see it about herself!) And then she went on to say that I never say anything nice about anyone, which is the biggest lie there is! I only dont say nice stuff about her or my Dad because they\'re both idiots (well - I do sometimes...but not too often - more often about my Dad then her - he at least makes an effort to be nice to me - she doesn\'t get anything for being nice to me - so why bother?)  Well - done ranting and raving - time to go bomb my psych quiz (thank God for extra credit.)  -Mb  '),(17457,2413978,'12,July,2004','2004-07-12 00:00:00','   Church yesterday, with my Mom, was boring as it always is. And youth was even more boring. These rude people called me a nerd basically when the pastor asked what makes us individuals and I gave the psychological approach on it... they\'re lucky I didn\'t turn around and cuss them out. I mean - I didn\'t want to be there anyways and being escorted out would give me a reason not to go back that even my Mom wouldn\'t argue with.  Then at the end, this dude, Lamont (I guess he\'s a pastor now or something? Who knows) came up and gave this long speech about why we should tithe. Dude - its teens. Everyone who was going to tithe already had pre-determined it, probably got the money from their Mommy and the speech was a waste of your breathe and trying to my patience. Most of it I just tuned out, but I did distinctly catch something about \"smashing the devil\'s face in the floor\" *rolls eyes*. Most of the people there wouldn\'t even take on a simple, lowly demon and live through it - I\'d like to see them smash satan\'s face in the floor. Damn naive people.  Then last night my Mom was a bitch too. I dont know what her problem was, but she was being truly evil last night and probably will continue it today.  I continued my psych behavior modification project - trying to modify my sleep patterns cuz I always go to bed too late cuz I usually take a nap or something - so its in the works. Was asleep by 12:00 am last night, which is an improvement I guess.  This morning, my bus was like 20 minutes behind and I figured I missed it, so I caught an alternate route (which is a good thing) and then I got to school like 30 minutes later then I usually do, but I thought it was an hour and 30 minutes later so I was running to my psych 101 class, and I was all winded and then I walked in and I saw all these people I\'ve never seen before and I was like \"Is this psych 101\" and the instructor was like \"that class hasn\'t started yet\" and then I realised it was 9:30 instead of 10:30... I wanted to crawl under a desk and die. That was so embaressing.  Well right now its 10:0(something) and I should get to class - so peace out.  '),(17458,2413978,'12,July,2004','2004-07-12 00:00:00','   My review:   Talking rabbits throughout the entire book, Watership down was a book that included tons of imagination on the authorâ€™s behalf, but in my opinion did not meet all the hype from various fans, proclaiming it the best book ever written. It didnâ€™t really even take off until the last 100 pages or so and at the beginning was a very slow read â€“ almost rendering me into such boredom that I put it down and went onto something else.  Once the book took off though, it was edge of your seat all the way to the end of the book. I think this book would have been much more exciting to me, being it talking Rabbits, had I not read the Redwall books in the past. But having read the Redwall books, talking animals are no new concept.  The author did a lot of research into the habits of wild rabbits, how they live, how they mate, their excretion habits, etc â€“ and the book would have been loads better had it had some action in the beginning. Also, like Lord of the Rings, this was a very deep book â€“ one that you can just sit down and read through while thoroughly enjoying because of its depth and messages to us, as humans, about destruction of woodland habitats, littering, encaging wild animals, etc â€“ although I enjoyed this book much more then Lord of the Rings.  '),(17459,2413978,'23,July,2004','2004-07-23 00:00:00','   urlLink'),(17460,2413978,'21,July,2004','2004-07-21 00:00:00','   I found a ride to Flood starting next week. Some dude in Chula Vista named Tony who I\'ve likely met before but only vaguely remember. He was at Xscape this one time when I was still a leader and this band was there that he interned with or something.  Also Something Like Silas is going to use me for promotional stuff. The band member who I sent the e-mail too said to add me as a \"key\" street team contact and to add me on the promotional team. That is so awesome. I hope they put me to work cuz they\'re a really good band for a really positive cause.  Mom doesn\'t like to let me on the net anymore and I try to be at school as little as possible - so I dont hang out there on the net all the time anymore. Plus they have set every computer where you have to login with a student id number and it wont let me. This one time I kicked the power cord of the computer I was on accident and it turned off...and ever since then whenever I try to login it tells me I\'m already logged in...so no go. They only have like 4 computers where you dont need an ID number. And thats probably not for long...so I\'m screwed.  Next week is the last week of Psychology. Should be interesting. We have at least 3 more chapters to cover, so at least 3 more tests and 3 more quizs and then the final in 4 days... can someone say lack of sleep? But then its over and I\'ll have a long vacation.  Heres some cool quotes from TogZone:  *I\'m multi-talented. I can talk and piss you off at the same time  *Laugh alone and the world thinks you\'re an idiot  *I tried sniffing coke once but the ice cubes got stuck in my nose  *I\'m not weird. I\'m gifted  *True love burns the brightest...but the brightest flame leaves the deepest scar  *Pride is tasteless, colorless, sizeless but the hardest thing to swallow  *Consciousness is the annoying period of time in between naps  *Sorry, I haven\'t been the same since that house fell on my sister.  *Sarcasm is just another free service I offer.  *When someone loves you, the way they say your name is different. You know that your name is safe in their mouth.  *The more I know, the less I understand  *As long as you have one true friend you\'re famous  *Honesty is the best policy but insanity is the better defense  *Things that piss me off:  1) People who point at their wrist while asking for the time. I know where my watch is buddy, where is yours? Do I point at my crotch when I ask where the bathroom is?  2) When people say \"Oh you just want to have your cake and eat it too.\"  Well, what good is a cake if you can\'t eat it?  3) When people say \"It\'s always in the last place you look\". Of course it is. Why would you keep looking after you\'ve found it? Do people do this? Who and where are they?  *Therapy helps but screaming obscenities is cheaper  *Never say oops always say \"ahh interesting\"  *Some mistakes are too fun to make one only  *We don\'t need torture cambers~we got them right between our ears  *True love doesn\'t have a happy ending, true love has no ending  *Once I had a dream. I was in the woods and I was running... running...  running... then I smacked into a tree  *There are no stupid questions, just stupid people  *If you never walk in the rain, you\'ll never know how good the sun feels  *You never know when you\'re making a memory.   Ill try to keep updating as much as I can though. Even though Mom thinks the net is evil and doesn\'t let me on *grr*.  -Michael  PS: Today I had to go to my High School and drop off papers for residency. I decided to take the bus cuz I have a pass and I\'m lazy even though its only like a mile, and at the bus stop there was this wierd stoned lady talking to me about college. She seemed sane enough, but was very strange. Then the school told me the bill doesn\'t say \"chula vista\" on it...so I\'m only half-registered and they need to follow up.  Tonight I\'m supposed to meet my English teacher at Starbucks and my Mom said no...so I guess we\'re disrupting her plans... Which sucks a lot.  Well better get off here before I get caught and busted.  '),(17461,2413978,'21,July,2004','2004-07-21 00:00:00','   Well I\'ve been going through this whole crisis situation today and last night about how I want to test out of High School and go on to college and I got tons of good advice and was considering doing it until both of my parents were assholes about it. So I guess I\'m staying high school now. Can continue to be anti-social and shit. Oh well.  I guess I\'ll take another college class and take High School and college concurrently or something.  Should be fun  PS: Would make this longer but I feel like crap, am tired and am in a bad mood. So I dont feel like writing.  Had Subway for dinner. Bombed test today. Put a bunch of cool stuff on TogZone (new avatars and smileys - go check them out - http://togzone.com)  '),(17462,2413978,'19,July,2004','2004-07-19 00:00:00','   Evryone send hate mail to Inpop Records (joke). I requested an interview with Superchick so I could publish a story about them in Blurb magazine and they wrote back \"Superchick is just too busy right now. Thank you for your interest\". I bet if it would have been Peter Jennings they wouldn\'t have been too busy. But my audience, which is teens would have been much more in interest of the band then the general TV audience. How many teens watch World News Tonight? My audience on the other hand would be 50,000 or more teens in the San Diego area. Probably at least 100 of those 50,000 would have bought their CD (or acquired some of their songs in some way).   So now I\'m not too thrilled. I just hope that the band will over-ride that decision if my e-mail gets through to them.  '),(17463,2413978,'18,July,2004','2004-07-18 00:00:00','   Last night I was on the Something Like Silas website and I saw that in San Diego they were playing at something called \"Flood\" a lot, so I googled \"Flood San Diego\" and I found out that its a cuhrch thing. It looks really really really awesome. I was on their website a long time last night. I definately want to go out there. &nbsp; The only really bad part about it is that its clear out at College Avenue Baptist (like 30 minutes away) and it would be on a Sunday night that we went. Their only services are at 5, 7 and 9 pm. Whats really good though is that I can get a new start there. At Cornerstone I\'m the evil sinner person, here they can see me for what I really am. Also its like a huge youth church. Most people there are between 18 and 28 according to the website. And their High School ministry looks small, so maybe I can easily become a leader again or something? I doubt any church will let you just jump up in there like Cornerstone does though. Thats unheard of. To be in leadership after like...3 weeks. Maybe it was my charm? (joke) &nbsp; Whats even better I think Something Like Silas leads the Worship at Flood. So maybe I can get to know the band people and then once I get to know them help them out and promote them and stuff. I\'ve been looking for a band to do that with more directly then I do with Superchic[k], and my heart would be way more in it if the band people were friends rather then just random people who happen to sing well. &nbsp; Also hoping since this is as CABC, and I uesd to go there for BSF, that some of the less decent BSF people (like Ryan) dont go to Flood. Cant say I miss them... &nbsp; But off to shower and then class I go. &nbsp; (and as my cousin would say)   ok well, g2g, ttyl, bye!  (I refuse to do that on a blog - yet I just did it...how wierd. I basically feel this is your thoughts...and though you\'re writing for an audience, it is not for them like a conversation...so why say bye? I could say doing like I do \"running to the shower or school\" but to say ttyl (talk to you later) bye. Who are you talking to? OH well...had to put this in there.)  '),(17464,2413978,'18,July,2004','2004-07-18 00:00:00','   Ol\' Trey (his alias) turns out to be alive after all. I completely gave up on him ever signing back on the net. I used to be really close to him but we both got lives I guess and aren\'t anymore. I have his phone number and cant call him - so I\'m really not holding my breath on ever really even talking to him again. If I do great, if I dont oh well. I\'ve learned that getting close to internet people just hurts in the end when they turns out to be 60 year old perverts or just never come back on. I do, however, hope he\'s doing good and stuff though. He was an awesome guy. Also hope he\'s making good choices. Ok - so officially I\'m letting go of him now. Used to think he was the greatest person in the world and stuff - no point in that anymore. Bye bye Tommy Miller.  urlLink  '),(17465,2413978,'18,July,2004','2004-07-18 00:00:00','   My friend Catherine sent me this e-mail. Its pretty funny   At New York\'s Kennedy airport yesterday, an individual later discovered to be a public school teacher was arrested trying to board a flight while in possession of a ruler, a protractor, a set square, a slide rule, and a calculator.   At a morning press conference, Attorney General John Ashcroft said he believes the man is a member of the notorious Al-gebra movement. He is being charged by the FBI with carrying weapons of math instruction.   \"Al-gebra is a fearsome cult,\" Ashcroft siad. \"They desire average solutions by means and extremes, and sometimes go off on tangents in search of absolute value. They use secret code names like \'x\' and \'y\' and refer to themselves as \'unknowns,\' but we have determined that they belong to a common denominator of the axis of medieval with coordinates in every country. As the Greek philanderer Isosceles used to say, \'There are three sides to every triangle.\"   When asked to comment on the arrest, President Bush said, \"If God had wanted us to have better weapons of math instruction, He would have given us more fingers and toes.\"  '),(17466,2413978,'18,July,2004','2004-07-18 00:00:00','   \"Nobody talks so constantly about God as those who insist that there is no God.\"  '),(17467,2413978,'31,July,2004','2004-07-31 00:00:00','   My views on US-Mexican border relations, as well as other illegal immigration.   For awhile now Iâ€™ve been very concerned and infuriated by illegal aliens cheating the American school system and wasting our tax dollars by coming to school, free to them, at a cost to us, in America. Illegal aliens also come to America when they are ready to give birth, hoping to have their child in the US in order to get welfare and other aid intended for American citizens. Putting a stop to this selfish unlawful occurrence would be the best choice for the American and Californian Government.  On top of all the other challenges facing the American educational system, massive illegal immigration is costing the nationâ€™s schools $7.4 billion annually and has added 1.1 million students to the countryâ€™s already overcrowded classrooms (Dan Stein, editorial: Massive Illegal Immigration is Crippling Americaâ€™s Schools, September 15, 2003.)  Living on the border, in California, a border state, I see firsthand on a daily the effects of having a school largely populated by illegal Mexican immigrants has on the California economy. We get new textbooks only every seven years because of budget cuts, are discouraged from using air conditioning unless it is unbearably hot (in schools that actually have air conditioning) because of budget cuts, have outdated technological equipment and generally go without in science labs because of budget cuts. This is but a small list of things we go without due in large part to illegal immigrants. The perks that we get in school usually come from the occasional grant and not the California school system. However, this problem would be greatly reduced if a method of preventing illegal aliens from attending American schools without paying tuition was instituted.  In California, which has the greatest number of illegal aliens in its schools and, not coincidentally, a budget deficit as large as the other 49 states combined, the $2.2 billion now devoted to educating illegal immigrants could be spent hiring an additional 41,764 teachers, or adding enough new schools so that no child in the state is sitting in an overcrowded classroom (Dan Stein, article: Massive Illegal Immigration is Crippling Americaâ€™s Schools, September 15, 2003.)  Some would say that I sound selfish, others would say heartless, and still others would say I donâ€™t know what its like to not have all the benefits that I have as an American. And theyâ€™re right, I donâ€™t know what itâ€™s like. But the opportunities in America are just that, for Americans. Not everyone else in the world who doesnâ€™t want to earn what we, as Americans, have. America didnâ€™t just appear and end up this way. It would be different if the illegals became citizens, but they donâ€™t want to give back to America or have any responsibility (like paying taxes that go toward schools) that Americans have. They just want the perks Americans have earned, without being an American.  Debating this issue with others at my school Iâ€™ve been informed that â€œeducation is not free in Mexicoâ€ and Iâ€™ve tried to explain that it isnâ€™t free in America either. Yet most Mexicans and even many misinformed Americans have the misconception that it is free here. Yes I agree its unfortunate that they do not have education provided by the state in Mexico, but why should American citizens have to go without? How is it our problem that education is not provided in Mexico?  This brings me onto another topic, illegal Mexican immigrants crossing the border shortly before they are due to give birth to their child. In many cases the pregnant woman has no idea who the father of her child even is and happens to be so poor that she doesnâ€™t feel she can support her baby in Mexico, so she comes to America to have her child and make her â€œAmericanâ€ child the responsibility of a government which she does not claim belonging. This is another way American tax dollars are wasted on illegal immigrants.  Also keep in mind, welfare breedsâ€™ welfare. Any illegal immigrant turned temporary residentâ€™s children will think the same way they do. â€œI have no allegiance to this country anyways, so why not get everything I can free from it, I can always run to Mexico if something goes wrong.â€ That is exactly the mentality they have; Iâ€™ve heard just that, repeated numerous times.  There are approximately 8 million undocumented immigrants in the United States; 2.5 million reside in California. The costs of providing social services to these individuals and incarcerating illegal immigrants is staggering - easily exceeding $3 billion a year (Arnold Schwarzeneggerâ€™s undocumented immigrant agenda, http://www.joinarnold.com/en/agenda/#E1.)  Immigration is a national responsibility and California should not be left to shoulder the massive burden of providing public services to illegal immigrants. If an immigrant does not want to be become a citizen they shouldnâ€™t be allowed here unless they have a visa granted for a certain period, to visit.  Amazingly, even with the knowledge of the above stated statistics, very little has been done in prevention. However, one of the most significant efforts within California was that in 1994 by Gene Garman. Garman proposed that California work off of a school voucher system where legal residents of California with social security numbers receive annual school vouchers usable at either a private institution or a public school. This would have made it where residents could go to either type of school and it would have prevented illegal undocumented immigrants from lying about their addresses, because regardless of their â€œcousinâ€™sâ€ address: no social security number, no American school. Itâ€™s as simple as that.  However, the teachers and public school district officials were very concerned that if children had the choice of going to private schools over public schools with school vouchers, they would be unemployed because no child in their right mind would chose a public school over a private, so the proposed bill for school vouchers was pushed out.  The school districts in California try to put a stop to illegal aliens going to our schools, but in all realization they really just make it harder for residents to go to school. Illegal immigrants simply use a relativeâ€™s address, lying about having residence and their children go to American schools, undermining the entire purpose of residency verification. I have various friends that brag about how many school documents come to their houses when school is starting and how many report cards come to their house at the end of a semester, addressed to their various distant relations. They think itâ€™s a cool thing that they are getting over on the American Government and getting something for free, at our expense.  Rumor has it that Arnold Schwarzenegger, governor of California is in the works of passing a law where there will be a certain amount of time for all illegal residents of California to apply for citizenship and then all whom are denied citizenship (or who never sought citizenship) would be deported if over the set amount of time. Though, I have not been able to confirm the validity of this rumor. However, I think this â€œrumorâ€ easily and without question would be the best-case scenario for California. To have a closed border after a set period of time, where, after the border is closed, all non US residents would have to have a visa for a temporary visit in order to cross the border.  Closing the border to illegal children going to American schools and preventing illegal adults from giving birth in the US would save California billions of dollars and it would make our schools much better in the long run. He would have supplementary current text books, less crowded classrooms, schools that speak primarily one language, newer technological equipment, more athletic resources, more of a budget for clubs and much more. Schools would be better for everyone.  Most Californian schools have bilingual classrooms, a sure sign, if someone doesnâ€™t even know the language, that they donâ€™t live here. Our taxpayersâ€™ dollars knowingly go to educating illegals. What is wrong with this picture? If an American was to go to France or Germany (or gasp Mexico) you can sure as hell bet they wouldnâ€™t be educated in English rather than the countyâ€™s native language.  On closing, Arnold Schwarzenegger is a classic example of an immigrant who became a citizen. He came here from Western Europe, knowing little to no English and probably very little about America, yet he tried and worked tirelessly to become a citizen. Look where he is now. He is a very popular actor, who has won multiple awards and has recently become governor of California. If he can do it, any other immigrant can do it; they just have to expend a little effort. If they were to become citizens, and were giving back to the California economy we would be much better off and have much more of a budget not just as a state, but also as a country.  Note: Wanted to clarify, that though I\'m still moderatly against illegal residents living in the state of California, someone living here (a resident) has every right to go to American schools. They are a resident after all. My comments on illegal immigrants not going to American schools was in terms of those who come over the border in the morning, go to American schools, use a \"cousins\'\" address and then go back over the border at night.   And on another note, although this article seemingly targets only Mexican immigrants, it applies to any other immigrant too. The problem in Southern California, obviously since we\'re on the border, is Mexican immigration.   If you would like anything clarified, have any questions or would like anything changed for political correctness, please let me know via a comment or a private message.   Comment I recieved:  Re: Open Us-Mexican Border - A Mistake? (Score: 0) by Anonymous on Saturday, July 31 @ 16:44:04 PDT (IP: 68.32.2.111)  I challenge you on this... America as a whole is a wealthier country than Mexico or most countries that bring in a lot of illegal aliens... and as a Christian, we should be willing to sacrfice things to help others. I challenge you to look at this from a Christian perspective, and remember that Christ requires us to put others before ourselves. What makes you think that just because you were born in a wealthier nation, you deserve to have a better education than the less fortunate? You being born in another country doesnt make you any better or any more deserving than anyone else. Think about this... what if you were one of them? All you want is the best for your child, a strong education and for them to live in a place where they have more opportunity than if they were in their home country. My opinion, you need to stop thinking about yourself and think about others. That\'s what Christ would have done.   So what? Some of your tax dollars go to those who are less fortunate... think of it as saving a life rather than a waste of money. You are sharing with those less fortunate. Don\'t be self-centered... Jesus had every right not to die for us, it was unfair of him. But he managed to show his love for dying for us. You should extend the same kind of love for others.   Afterall, Jesus was a Jew and could have only saved the Jews... after all they were HIS people, and he could have wanted only the best for his nation... he could have said, \"Only the Jews deserve salvation because only they pay taxes to the Jewish government.\" Thank goodness he didn\'t say that, huh?  '),(17468,2413978,'29,July,2004','2004-07-29 00:00:00','   Today was definitely interesting to say the least.  I slept in till about 12:00 pm because I was up so late the night before (and am up really late right now) and then went on the computer for about an hour and worked on my site, togzone.com â€“ I installed a bunch of new stuff, answered some of the topics on the forum and wrote an article for it.  Then I worked around the house, cleaned up cat poop, scrubbed the ceiling and cleaned my bathroom (all shiny and pretty now). When my Mom got home I was just finishing up cleaning my bathroom, jumped in the shower and then we went to see â€œThe Villageâ€.  My review of \"The Village\"   Mysterious from the beginning, the Village was a very unique and interesting movie. It showcases a valley â€œvillageâ€ surrounded by a forest and isolated from â€œthe townsâ€™ with happy residents virtually free for any sort of crime and evil.  Written by M. Night Shyamalan, The Village stars Joaquin Phoenix (Mel Gibsonâ€™s Brother in Signs), Adrien Brody, William Hurt, and Sigourney Weaver and is categorized as â€œDrama, Romance, Science Fiction/Fantasy and Thrillerâ€ by  Yahoo! Movies .  The villagers stay inside of their village and â€œthose whom we do not speak ofâ€ stay in their forest. Until something changes (or does it), the villagers and creatures kept to themselves, in peace. According to one of them main characters, â€œthose of whom we do not speak, have not been inside the village boundaries in many years.â€  From the beginning of the movie, a series of odd events happen. Two young women are playfully sweeping their porch until they see a red flower that they promptly bury. Later in the movie, Adrien Brodyâ€™s character acquired some red berries that are also promptly buried, implying that something is strange about the color red. What does it do to â€œthose whom we do not speak ofâ€?  The color yellow also seems to be conspicuous. People of the village, at times, wear druid looking robes the color yellow. Also, along the boundary of the village, the torches used to ward off â€œthe creaturesâ€ have yellow banners draped on them. What anxiety does the color yellow cause?  M. Night Shyamalan put a lot of deep thought into this movie, many elements of suspense and some very twisted thinking, however the movie gives away the twist (and the climax) way too early. The viewer knows what is going to happen and the plot gets boring for about 20 minutes. Shyamalan would have been better off to hold off the twist, as in Signs, until the end of the movie.  I wouldnâ€™t rate this movie as an excellent movie, however nor is the movie in by any means a bad movie. It is just a very strange movie, one that is not good or bad. I do, however, feel the movie is a very deep movie, with quite a few messages about the nature of man and what anxiety can cause man to do to stop it. I would recommend this movie to anyone who doesnâ€™t frighten easily, as The Village is a very suspenseful movie, however by no means a â€œhorrorâ€ movie. Just a tip: plug your ears when suspenseful scenes come on. The music gets VERY loud.  All in all I was disappointed with this movie. Iâ€™ve been anticipating seeing the movie since I saw a teaser trailer beck in December and once it was overâ€¦I was very disappointed.  Some well written movie review quotes from users of  Yahoo! Movies :  From user  drudashoe913  (their review:  urlLink Click Here )   You will get a great storyline, absolutely amazing acting and characters (probably the best part about the movie, I think some of the acting could possibly be Oscar-worthy!!), some real scares (I almost jumped out of my seat a couple of times), and an ending that will leave you feeling utterly dissapointed. Remeber the phrase, \"What seems too good to be true, probably is.\"   From user  ademptious18  (their review:  urlLink Click Here )   Further, I was shocked that Shyamalan would have so much political commentary in the movie. First of all, the leader, who is trying to freak out all the people as a power ploy, conventiently has the name \"Walker\" (Funny story, thats Bush\'s middle name and is a subtle commentary on Bush\'s policy with the war on terror). Further, the warning colors around the town are the same code yellow that we are on, and the bad color, the warning color, is of course red. I won\'t continue and ruin the story, but Shyamalan could have made the movie far better by focusing on delivering a worthy scare than on political commentary.     After seeing \"The Village\" I came to my Dads and then we ordered a pizza, I worked on TogZone a lot more and made a bunch of new changes and a bunch of additions and then talked to my friend Carrie for a long time.  Nothing that interesting today. Mom has actually been letting me on the computer though. Which is big improvement from how she usually is.  And also we\'re going on this \"drive up the coast\" vacation thing where we\'re driving along the coast all the way up to the Bay Area. And then back down. Should be interesting being in the car with my Mom 9 hours each way.  Also: this dude from Cornerstone instant messaged me because of one of my Xscape ranting and ravings where I insulted him. I was totally sorry about it...look what ranting and raving can do to you. So I edited the insult (that wasn\'t directly to him, but stll taken as an insult) and now that he loves me again, we\'re gonna go to the movies next week and see The Bourne Supremacy! Yay!  Yesterday I went to my Aunt\'s house (not really my aunt, but a really good family friend) and I hung out with Beth, best friend/cousin (again, not blood cousin) and we hung out. Then I borrowed The Bourne Identity from them so I could watch it again before I go with that dude next week. Was such a good movie. I forgot how good it was. And the ending rocked.  '),(17469,2413978,'29,July,2004','2004-07-29 00:00:00','   Well tonight I\'ve been making tons of changes to TogZone.com  One of my friends, Zach, who I haven\'t talked to in a long time joined up today and he already has 60 something posts on the forums. Thats crazy. He\'s really getting it to liven up. I\'ve also been drafting up some documents for the site, I setup a board of advisors (The Council and \"The Round Table\") and then made a lot of propositions and worked out some errors.  Then I was talking to this guy that I met the other day, Moises and he wants to meet up and go see The Village on Friday. I told him the first time we need up it has to be with a parents present so I know he\'s not gonna kidnap me or something and he agreed to that. So we shall see if he actually shows up or not. If he doesn\'t, its a sure sign he\'s a psycho stalker freak person. If he does, I might have just found a really good new friend.  Well - I have finals tomorrow, I still have some chores to do and I need my sleep - so I\'m gonna be going now.  My Mom actually let me on the net tonight. Its a miracle and she even trusted me enough to be on while she\'s in bed. I feel truly loved tonight. She probably would freak out if she knew I was still on the net...but still pretty shocked she let me on. She usually seems to have major issues with the net.  Oh well. Time to do chores then sleep.  '),(17470,2413978,'28,July,2004','2004-07-28 00:00:00','   Today was awesome at school. I got there early, was talking to Manny and this one girl about porn, alcohol and how when you leave home and go out on your own, you dont know how to cook (believe me, I didn\'t pick those topics). Then we did some last minute cramming for the quiz (which we didn\'t have - thank God) and we went in the room when the teacher got there.  She called me, Rebecca and this other girl (forgot her name) up to the front of the room and she told us that we have the highest grades in the class and it would be wrong to place us in a group because it would be unfair (we know all the answers so our groups would basically have an un-fair advantage) and then we played \"Rocha Jeopardy\" (the teacher is Ms. Rocha - hence Rocha Jeopardy.)  Well, we decided that I was going to be the caller-outter person (when they raise their numbers in the air signalling that they have an answer to the question on the screen, I write down their numbers in order) with Rebecca as the score keeper and the other girl as the right/wrong answer teller. We were a good team but I had the hardest job because they were screaming at me and harassing me if they felt I wasn\'t calling their numbers.  The last question is the one who really determined who won though. It was worth 50 points. So it killed the people who got it wrong and really helped the people who got it right (only two people got it right - one was already winning.)  What was really cool is that because I was a helper person who was running the game I automatically got 25 free points (the equivelant of what first place was). That makes it where I really dont even have to study for the final for tomorrow because it shouldn\'t hurt me even if I do terrible.  Before the class started I was looking at the grade sheet and I had a 103% and then the 25 points I got today probably put me at 107% or 109%. Maybe not that high, but it was pretty high.  So right now I think thats really awesome.  Need to go work on my site, http://togzone.com and respond to a bunch of e-mail. Peace out.  '),(17471,2413978,'27,July,2004','2004-07-27 00:00:00','   Today I got up kinda late. I felt like being lazy again, but I still got on the bus in time to make my morning class. At 9:20 or something the 701 came and then I got to school by like 9:45 and just sat outside of my psychology class reading Black.  My review:  A thriller about the end of the world, Black ties in a lot of elements for a good book. It seems also to have indirect religious messages tied into the story, that would go over many peopleâ€™s heads, but as a Christian I got them right away.  The book starts out where the main character, Thomas Hunter is being chased around an alley in Denver by loan sharks. One of the loan sharks shoots him in the head and it only makes a dent on his skull without actually hurting his brain. After taking the blow he starts to have these â€œdreamsâ€ of another world. One way in the future. And he learns that in 2010 there is a virus that wipes out everyone. Well, in his alternate reality or non-dream it is 2010, so the fight is on to stop this virus from being released.  Throughout the book, when Tom falls asleep in one reality he wakes up in the other reality, but the realities have no time correlation between each other. He could be in one reality for 15 years and then wake up in the other one and he was only asleep for 30 minutes.  I would recommend this book. Its not the best, and the end was disappointing (since it is a trilogy and the author wants to keep you reading) but all in all it was a fun book. It kept me turning the pages and I was able to finish it in two days. Thatâ€™s pretty quick for a 400 page book while also in school. That tells you how exciting it was.   Well psychology was good today. This was our last day of actual instruction today. We went over psychological diseases and disfunctions such as bi-polar, syhizophrenia and depression. It was really interesting all things considered. And I got to see the cool people in my class again. Thank God for them. I would be losing my mind if I had to sit through the section 5 Psychology 101 class...really really would. Those people in there...geez. No wonder people from that class come to my class all the time. Between the helpless blonds, the steryotypical Abecrombie-God\'s-Gift-to-Women-and-the-world jocks to the lady who is inventing symptoms of pregnancy because she wants to be pregnant (tmi man...TMI!!)  Manny was also back today. He is a really awesome guy. I wish I was gonna be at SWC longer to I could get to know him better. I\'m sure to him I\'m this little kid or something but he is a really awesome guy. Hopefully I can get his contact info and we can hook-up outside of class. But that probably wont be a reality.  Today I got my grade for psychology. I am sitting at 94% before she added the extra credit. Today I got to drop a test that I didn\'t do too hot on and then I got to also apply my 25 points for perfect attendance. So that should get me back up to like 97% or maybe higher. So if I dont do to hot on the final shouldn\'t really affect me too much. I should definately still have an A in the class.  Now Mom is kicking me off because I \"can\'t be trusted to get off\" so cya. Ill add to this post later.  '),(17472,2413978,'26,July,2004','2004-07-26 00:00:00','   Been crazy this week. I got the schedule for my Psychology Class and we\'re behind so the teacher is cramming in a lot of stuff. Final shouldn\'t be too bad though because its only 100 points and a regular test is 90 points...so if I totally bomb it (which is very unlikely - even without studying) I should still get out of the class with an A.  Today I slept in because I had some stomach bug thing and wanted to make sure I wasn\'t still sick so I attended the later class (wow are there some strange ones in that class.) Mental note to attend my regular class the rest of the week. Plus I have some friends in there (or at least acquaintences) that I want to say bye to since this is the last week and get phone numbers and stuff.  Mom has been keeping me off the net and today at school the net was down and then at my Dad\'s the power went out right when I turned the computer on...so I haven\'t been on the net in like 5 days... too long!  When I came on today I had a LOAD of email and then I have a lot of activity on TogZone to contend with too. But thats good. People are finally starting to use it without me having to IM them and harass them into it (well...at least SOME people are).  Been reading a lot. Which reminds me I need to update my site of what I\'m currently reading. And then I went to the library and got more books today. The books of magic and something else. I want to read Diane Duane\'s series of young sourcerors or something too. Reading Black by Ted Dekker right now. Its a really cool thriller type thing.  Well - gotta go. Thought I should update so Chris doesn\'t kill me for not updating. So thurr be le update.  -Mb  '),(17473,2413978,'05,August,2004','2004-08-05 00:00:00','   Shrunk this picture for a friend. Thought I would post it here too. Was from the Prom with Jackie in June.  urlLink  '),(17474,2413978,'05,August,2004','2004-08-05 00:00:00','   Met some really interesting internet people in the last couple of days. One, Tony, lives here in Chula Vista (I met his on this Church\'s website\'s forums when he offered me a ride to church) and we hit it off pretty well. He seems really awesome. We were gonna take a surfing class at SWC this fall, but the class was full...so I guess I dont get to do that.  I was bummed when I went to SWC to sign up for the class and it was booked. Then I tried to sign up for Human Sexuality, but that section (at night) that I wanted was full as well...so in the end I signed up for Social Psychology. It was $58 for the class since it was 3 units and then the text book (used) was $79. Lameness. Textbooks are so damn expensive.  Then also I met this awesome net person in the near past, Carrie. We totally connected, so much in fact she\'s already an Admin on my website (kinda wierd, but I feel like I can trust her even though I barely know her) and then recently I\'ve met her Boyfriend, who I like as much if not more then her. They\'re awesome. I hope to meet both of them the next time I go to San Fransisco (they live about an hour 1/2 away, but thats a close central public place for us to meet and hang out.)  Also been talking to Isaac. He seems totally in love with this new girl he\'s dating. And he has me (indirectly) to thank for even getting to know her. I was going on and on about how crazy I was about this girl (and she doesn\'t like me in the least...oh well) so he got to know her, trying to help me get a date with her (but when I found out she doesn\'t like me, I thought \"why bother\") and through getting to know her to help me out he started having a crush on her and then asked her out. I think he\'s her first B/F too...not sure.  Right now we\'re getting ready to go on vacation up the coast. Its gonna be pretty awesome. I\'ll make sure I have some pics to post up here of the vacation. We\'re going up all the way to Pismo Beach which is 5 hours away, but we\'re stopping at various places along the way. Gonna be fun! Then staying at Pismo Beach a couple of days and coming back. But my Mom says we\'re stopping like at two points along the way, and staying the night, so it will be little like 1:30 hour chunks of travel time. I shouldn\'t get too bored.  Also, on BookCrossing I was just talking to this dude, who is gonna send me Alice\'s Adventures in Wonderland and Through the Looking Glass (from Scotland) for free! I had no idea he was in Scotland. I feel totally bad now. Thats gonna be so expensive. I didn\'t even look at locations when I clicked on his name for the book.  And this other lady I\'m trading for \"Cause Caleb\" by Helen Fielding for \"Flowers in the Attic\" by Virginia Andrews.  Right now need to get back to my convo and then on with some chores before I head out to bed.  I may be going to see The Bourne Supremacy tomorrow with Christian Cornier, if all goes well with both of our schedules. And then will probably be at the laundry mat late tomorrow night cuz the Rabbit pied in my Mom\'s bed (and I have to wash the comforters - since as the kid everything is my fault). So it should be fun.  And Illianna, Isaac\'s G/F was supposed to take this dance class with me, the Modern Dance one in the fall...but she hasn\'t come to get the papers...so it seems like its not her priority. I hope the class doesn\'t fill up and she doesn\'t get to take it. But I\'m sure it already has...  So oh well.  Oh - also went to dinner tonight with my Mom and one of her friends from Work. It was cool, except that stayed there talking like 3 hours and I was bored to tears. But the Pizza was good. So I had something to do at least.  But yeah...need to get going. Peace.  '),(17475,3930072,'19,July,2004','2004-07-19 00:00:00','   Stupid fucker.  So I wake up this morning to go to practice and come downstairs to a lovely little message Jeff sent me on AIM. It basically tells me to stop putting things in my profile about him and if I want to say something to his face then I should call. It also says he\'s sorry things have to be this way but I\'ve obviously made it clear that they do. Of course, he was much more eloquent than that about the whole thing.  Alright. I\'m only half guilty.  Maybe the \"If you don\'t give a damn I don\'t give a fuck\" was somewhat directed at him.  But the \"I\'m sorry, I\'m allergic to bullshit\" I put as my away message was a quote from I,Robot, which I saw yesterday in the theater. I just thought it was funny.  And I\'ll probably tell him that the other message was about something else.   In fact, I have a pretty good idea of what I\'ll say. It goes something like this:  \"Look. You need to stop thinking of yourself so highly. That stuff in my profile isn\'t about you. The away message was a funny quote from the movie I,Robot and the \"I don\'t give a fuck\" was about something else entirely, more as a joke than anything. Coincidentally, it just happens to fit the way things are between us. So what. I told you I was done, and I am. It isn\'t about you anymore. Although I don\'t see why you give a damn about what I say about you or how I feel. You never did before. I\'m sorry if I make you look the ass that you are, but maybe it\'d do you some good to think about what you\'ve done to me. But of course you won\'t. Because it\'s all about you.\"  Or maybe I won\'t say any of it.   I probably won\'t even call.  I\'m just so sick of this all. '),(17476,3930072,'16,July,2004','2004-07-16 00:00:00','   You say I only hear what I want to You say I talk so all the time so.  And I thought what I felt was simple, and I thought that I don\'t belong, and now that I am leaving, now I know that I did something wrong \'cause I missed you. Yeah, I missed you.  And you say I only hear what I want to: I don\'t listen hard, I don\'t pay attention to the distance that you\'re running or to anyone, anywhere, I don\'t understand if you really care, I\'m only hearing negative: no, no, no (bad)  And so I, I turned the radio on, I turned the radio up and this woman was singin\' my song: the lover\'s in love and the other\'s run away, the lover is cryin\' \'cuz the other won\'t stay.  Some of us hover when we weep for the other who was dieing since the day they were born. Well, this is not that: I think that I\'m throwing, but I\'m thrown.  And I thought I\'d live forever, but now I\'m not so sure.  You try to tell me that I\'m clever, but that won\'t take my anyhow, or anywhere with you.  You said that I was naive, And I thought that I was strong. I thought, \"hey, I can leave, I can leave.\" Oh but now I know that I was wrong, \'cuz I missed you.  You said, \"I caught you \'cause I want you and one day I\'ll let you go.\" You try to give away a keeper, or keep me \'cuz you know you\'re just so scared to lose. And you say, \"Stay.\"  You say I only hear what I want to.   --Lisa Loeb \"Stay\"  I heard this song today while I was at Kohls and it reminded me a lot of what I\'ve been going through. It\'s pretty accurate except for the part about staying. But most of the phrases are so true. Yeah...  The trip to Kohls was a success, however. I got a cute white tank with a green flower pin, a lightweight gray cardigan, and an awesome belt that\'s bright and colorful with little nostalgia pics on it.   I tried on some bras as well but that, as usual, was a failure. I mean, I tried on that one brand with sizes like \"nearly A\" and \"nearly B.\"  I can even fit a nearly A.  How embarrassing.  On a better note, I\'ve finally decided on all the outfits for my senior pictures. I picked out six just in case. They are as follows:  1. cute blue strapless dress with white polka dots 2. plaid pants and red halter 3. jeans and maroon polo 4. abercrombie track jacket and jeans 5. new white tank with flower pin and jean skirt 6. brown top with white lacey skirt and striped scarf  I\'m pretty excited about this. A lot less stress, you know, once you have your outfits picked out.  I think I\'ll get back to watching the Olympic track and field trials now. '),(17477,3930072,'15,July,2004','2004-07-15 00:00:00','   Thunderstorm.   I started watching Fight Club (just picked up the DVD today) but the rain and thunder is so loud I can\'t concentrate on the movie.   I\'ve got to stop eating. I mean, I wasn\'t really hungry but I just went into the kitchen and opened a bag of chips to munch on.   So I munched. And munched. By the handful, at that.   Which is bad.   Food is NOT therapy.   Hah. If there was anything good about me being depressed and shit after the whole breakup thing was that I didn\'t feel like eating. Every time I felt full I wanted to puke. Every time I thought about what happened my stomach would do a backflip.   With a round-off and a half twist.   So yeah, I totally dropped a few pounds and now I\'m back to eating chips. All of a sudden, though, I realized I should be drinking fluids instead because of the bladder thing.   It was more like I realized I hadn\'t peed in a couple hours.   So the trusty water bottle is back by my side, though I\'m not quite sure I will reach the 104 ounces I consumed yesterday. Ah, well, I don\'t need to run tomorrow so fluid intake isn\'t quite as important.   Oh. I got the results yesterday from the AP biology exam I took in the spring.   I scored a three...   ...out of five, that is!   I must say I was rather excited. I didn\'t know what the fuck I was doing on the test and I wasn\'t expecting to get a passing score but I guess you never know.   Who says bullshitting won\'t get you anywhere?   Actually, no one says that. We all know it\'s how we passed (or are passing) high school.   Which isn\'t to say one must always bullshit. My theory is, put the real effort into what you like to do, when you want to do it. Bullshit when you don\'t care. Bullshit when you want to save time.   Bullshit when the task at hand is bullshit.   Example. English teachers are big on the rough draft thing. You always gotta write a rough draft and do the peer editing and self editing crap before you can turn in your paper. But nine times out of ten the teacher doesn\'t even read what the hell you wrote. They just check \'em to see if they\'re done and give you 10 points credit or whatever. But writing always has to be a fucking process.   Well, I despise rough drafts. I loathe them. They are the biggest waste of time ever. So what do I do? I bullshit them. You can, too. Here\'s how:   1. Write in cursive. Sloppily.  2. Skip lines.  3. Use wide-ruled paper.  4. Make sure you have five paragraphs no matter what.  5. If you feel ambitious, write a great intro so the teacher thinks the rest of the paper is great.  6. If it\'s a research paper, include long direct quotes.   It\'s easy. Then you usually have another week or so to write your real paper. I mean, you aren\'t going to turn in the bullshit draft as the final one. No way. You\'re going to write a fucking awesome paper and get a 97% on it, but you don\'t have to deal with the bullshit of a rough draft. Because it\'s the real paper that\'s important. You put your heart and soul into that shit.   Rant. Rant. Rant.   What was I talking about anyway?   Oh yeah, AP bio. In short, I didn\'t take time to study for the exam but bullshitted the essays and was lucky on the multiple choice and I got a three. Ta da.   This means at State I can waive three biological science entry classes--BS110, BS111, and BS111L and start at a higher level.   You know, I was thinking. Here I am and I\'m smart. I got a 29 on the ACT, not to mention a 35 on the English part which is in the 99th percentile. I got a three on the AP bio exam so I can skip frosh classes in college. Then I was thinking, not only am I smart, but I\'m athletic. I\'ve been all-city, all-conference, all-regional and all-state in both cross country and track. I can run in the low 19:00s for a 5k. And then I was thinking, not only and I smart, athletic, but I\'m artistic, too. I\'ve won first place at student expo three years in a row for drawing, ceramics, and painting, respectfully. So here I am and I got all this talent and brains and ability and yet, I lost the one thing that made me truly happy.   Sigh.   I know my life shouldn\'t revolve around something so fleeting, but I guess it was just that important to me. He was important to me. Being in love made everything different. It was new and special and I don\'t know. And I wonder if I\'m destined to have everything everyone wants but not the one the thing everyone has that I want. &nbsp; Blah blah blah. &nbsp; Excuse the self-pity. &nbsp; I just wonder if I\'ve lost it all for good. &nbsp; '),(17478,3930072,'15,July,2004','2004-07-15 00:00:00','   Well I did it.  I told Jeff I don\'t want to do this friendship  thing anymore. That I can\'t. That I am so angry, so hurt, it\'s impossible for this to work.  I told him goodbye.  But goddamn all I feel right now is like crying. I can\'t bring back all of the anger I felt before. I can\'t find that jealous little pang in my heart when I\'d see him with Kara. I just feel miserable.  It was just like a film. I half expected to turn around and see Steven Spielberg and a camera rolling over my shoulder. We had it down to an art, really: standing at the doorway, looking into each other\'s eyes for the last time, one final kiss on the forehead, the fateful \"goodbye,\" walking away, turning around for one last glance...  I guess everything still feels like the movies.  Just not the good kind anymore.  I know I had to do it or else I\'ll never heal. But every day there\'s going to be something I want to share with him, and everyday I will miss him. And I guess I wish one day he\'ll call me up and say, \"Jen, I can\'t live without you\" but the chances of that are slim. He doesn\'t need me anymore and he\'s always been so much better at this kind of thing. He\'ll move on.  I\'m scared I may never see him again.  I had to do this partly because I was sick of being the one who called, the one who said \"hey, let\'s hang out.\" It was so hard because I feel he never made the effort for this friendship that I did. I was the consolation prize when Kara was busy, the backup when he needed something to do. So I walked away.  I don\'t know if he\'ll ever really miss me. He\'s good at finding distractions to prevent thinking.  I know I\'ll live.   I made a vow to stop denying how I feel, however. Whether it\'s right or wrong there\'s just one thing I wish circumstances would have permitted me to say to Jeff yesterday:  I love you. '),(17479,3930072,'14,July,2004','2004-07-14 00:00:00','   I had just an awful run today. I was shooting for five miles but by the time I reached Minges three and a half was looking good. I forced myself to make it through four but it was the worst run I\'ve had this whole summer. No joke. It was one of those utterly-exhausted-feet-dragging-if-I-have-to-go-up-one-more-hill-I\'m-going-to-die runs where you seriously don\'t think you\'re going to make it.  I blame the antibiotics.  Which pisses me off to no end because I have to take those damn pills for ten days and the Riverfront Run is in nine. If I\'m going to feel like I did today there\'s no way in hell I\'m going to run that race. I\'d rather not collapse on the course or even worse run an embarrassingly slow time.   So it can be said I came home from practice rather grumpy. And then, while I sat at the computer, angrily sipping red Gatorade, Clarissa IMs me out of the blue.  \"Hey. I have a question.\"  I\'m like ok, go for it. And you\'ll never guess what she asks me.  She invites me to go with her to Chicago next week to the Art Institute.  Yes, the Art Institute of Chicago. That museum in Ferris Bueller\'s Day Off. The one with all the famous paintings. The one she told me about last year when she went and the one I\'ve wanted to go to forever.   Oh my God.  My mood does a complete turn around. All of a sudden I\'m smiling and my heart is light and I just can\'t stop thanking her enough for inviting me and for thinking of me and I feel great. It was a crazy coincidence because I was in such a bad mood before and this opportunity came at the perfect moment. I mean, to a lot of people this wouldn\'t be a big deal (it\'s just a museum) but to me a trip to the AIC is exactly what I need.  Times like these make me think the big guy in the sky is looking out for us after all. '),(17480,3930072,'14,July,2004','2004-07-14 00:00:00','   I went to the doctor today for what I assumed was a bladder infection or UTI or whatever they call them in the business. They gave me antibiotics and pills to numb my bladder so I can pee without pain. Hah. It\'s rather interesting because the bladder-numbing stuff makes my urine a bright orange color, like macaroni and cheese.  It\'s exactly the same color as my Nalgene water bottle.  So I slept a lot today and sulked around the house because even though my body feels fine having an infection gives one the perfect excuse to mope about. No one gets on you about it because they assume you feel like shit and if they do say anything you simply reply, \"Shut up. I have an infection.\"  It works 95% of the time.  This is all great for me because if I didn\'t have an inflamed bladder or bacteria in my urethrea there\'s no way I could go around feeling sorry for myself or being pissed off at the world.  It\'s not like I don\'t have a good reason to do so. I do.  But \"my ex-boyfriend is a stupid bastard who doesn\'t give a fucking damn about my feelings or being my friend like he said he would and while he\'s off in Candyland with his new girl after a post-break-up period of less than three weeks I am seriously considering cutting off his dick while he sleeps and tossing it out the window of a moving car\" won\'t really cut it with the female Nazis of the household who insist I just \"get over it.\"  I suppose I can handle a little burning while I piss just to get some peace when I\'m unhappy.  I didn\'t run today, but it\'s alright. I\'ll make up the five miles tomorrow when everyone else plays frisbee.  Although, such an aggressive game would provide the perfect opportunity to smash in Kara\'s pretty face or kick her in the shins.  Did I just say that?  Did I mention Jeff\'s new girl is my fucking teammate?  Anyway. Maybe tomorrow during my run I will be able to enjoy the sound of the wind in the trees or the solitary pounding of my Mizunos on pavement instead of Lindsay\'s incessant chatter in my ear. I love the girl to death but she talks so much and out of courtesy at least I feel like I should respond to what she says.  It\'s called conversation or something.  Dad\'s about to come home with burgers for dinner so at least I can indulge in some greasy cow intestine, although after watching Compassion Over Killing\'s documentary on farm animal cruelty, vegetarianism is looking more appealing than ever. '),(17481,3613005,'16,June,2004','2004-06-16 00:00:00','   After thinking for a very long time...I decide to stop myself from seeing her anymore...Cause i know i irrating..but hais...i realli nv follow her..i oni wait for her...i scared later we walk away she can\'t find us...so i wait until she can see us den i walk..hais dunno lah..i tinking stop seeing and talking in the best way...like that wun make her feel irrated and i also wun feel strange strange de...Ytd night..at prp...I been looking at the beach...Thinking of wad should i do...maybe my waiting is sth tt is useless...hais...let see wad will happen in the future bahh...it either we stop being friend or take it like we are stranger... '),(17482,3613005,'12,June,2004','2004-06-12 00:00:00','   wahaha...today is sat..slp until 10 plus...den wake up go wash up..wahahah..mum prepare lunch le..cose she going to work soon..hahaz.. so go eat lor..den 10.30am watch pokemon..=D...nice..hahaz...den use the computer lor...so bored sia...soon it was 3pm..den i rush sia...cose need to go service..but lazy go...but when i tink about all the fun..i faster go pom pom den rush to catch the bus which bring us directly to church..hahaz...meet sam,wei ming and wq..hahaz..soon it was service time...wah sianzz wor..sing song for 30min..i stand until tired...hahaz...den sit down lor..the chair nt nice to sit..veri uncomfortable...den cose pastol kong nt here..so gt the wad rev christian dunno wad lah..the way he tok quite funni lah..but i could nt laugh..dunno why...den finally sing song again..soon it was over lor..den dunno why my cell group still nv go..den we also wait lor..den gt 1 sister came..and tokt o us..i tot..\"who is she?\"...hahaz..she tok about today service lor..den suddenly she sae \"where eric?\"...den i thought wad so act lor...den everyone point toward me..hahaz...dens he sae wad..eric birthday...\"i tot bd was 2 days ago.\" -.- den they all sing song..wah lau..wad god is blessing u now..i hear le also feel pai seh sia...but nvm thx anyway..she pass mi a present..den she ask everyone to touch me and pray..i mena like put hand around shoulder...den pray lor..so i juz close my eyes..hahaz...den soon the praying is over..den she ask mi open present..i dunno how to open sia..hahaz..so i juz slowly tear open lor..den she sae is u like the present juz rip if off..so i tear hardly lor..hahaz..den everyone laugh at me.. -.-..hahaz...it was a t-shirt..hahaz..it nice..thx u frienz...hahaz..realli like it alot..hahaz..den went home lor..send gx a song..which i tink is nice..hahaz...dunno wad to write le..stoping here..bye and thx my friendz alot... '),(17483,3613005,'12,June,2004','2004-06-12 00:00:00','   Another same old morning..Today wake up pretty early..cose can\'t slp well...ytd there juz seem to be alot of star...I look at everyone of them...Each look so beautiful to me...Today nth to do...No one to tok to...guess i she hate mi more le bahh...but it\'s ok...this happen all the time..there ppl who hate me..so i use to it le..usually i offend ppl easily without knwoing bahh..hahaz..so i juz listen to music the whole day lor...den at night sam gx online..send hima  song..\"to zanarkand\" nice music it ink..so sent him..hahaz...tok abit oni lar...  den nan shun sae wad so mani thing happen when he is gone..lolx..i told him go to dd blog see...he sae tired...-.- den i also dunno how to sae..so nv tell lor..hahaz...nth to write liao..bye.. '),(17484,3613005,'12,June,2004','2004-06-12 00:00:00','   Today my birthday leh~so happy...cose once per year mah...and tis year gt go out celebrate...ok lah..meeting them around 1plus..we go sentosa..me,sam chong,sam ng,cheng hao,wen qin and wei ming went together..wahaha..when reach sentosa we all play soccer..so fun...fianlly our team score one and we win~yes..but wen qin dun seem to be happy..den we went to have our lunch...den wen qin come and shouted veri loud..saying \"Who never cut his toe nail\" -.- ..cose duirng soccer..his leg gt cut by someone..no one admit sia..den in the end..all of us had to sae sorry to him...lolx..den later on they take the cake out and celebrate with me..hahaz..so mani ppl sing song...coolz...thx friendz...after eating the cake and cleaning up..we went to the beach there and take a walk..wahaha...we run on the bridage..to the twin tower..wow..from there u can see the ocean..it veri beautiful..the breeze was coolz =/ ...soon we all ran back..  when running back..i try to shake the bridage..can sia...so fun...den shake and shake...den all of us try..wah sia..scary den i faster run down the bridage...i dun wan to fall into it..haha...den soon time to go home..we all walk to take the bus lor...wah...gt 1 auntie smart sia..sia..she nv line up.. juz sit at the pole infront..den when bus came..juz went up..hahaz..nvm gib old woman face =/..hahaz..soon we decide to went to pavilion to play wc3..hahaz..fun sia...saw vic there...hahaz..finally over le..den i go back home lor..wq take mrt wif me..cose he go aunt house stay overnight at simei...den reach home switch on the com...found out tt gx wanan leave bhs...hais..even one of mine best friend wan to leave..hais..i dunno wad to say...den dd ask mi wheather is he my friend..of coz i sae yes...den she ask wan him to leave bhs anot..hais..i sae dunno...cose it his choice...i can\'t ask him to leave..cose i dowan him to leave..but i juz dunno wad to sae..hais...juz hope tt he wun leave bahh...let him choose...because of tis...i tink dd is angry bahh..sae realli i oso gt abit angry..cose tehy always use the word wadever..it pretty rude...if thing is ae annoy u..u can tell mi..dun always use the word wadever..maybe u dun noe how it feel when u tok to someone and he/she reply u wif wadever...but nvm bahh..it ur way..think i stoping here le...slp tight my friend...nitez.. '),(17485,3601621,'14,June,2004','2004-06-14 00:00:00','    My first impression was: is it the session voting page the longest page in the world?   I\'ve just got an email message, informing me that  urlLink The Attendee Web site  is up and running. It took me nearly two hours just to prepare an  alpha  version of my schedule. For those of you, who haven\'t done that so far, I suggest you to do that as soon as possible, since it might be far more time consuming than you think. Yeap, there are zillions of different sessions you can choose from and mostly all of them extremely interesting.  Again, there is one more blog article that I would suggest you to check. It\'s  urlLink Tim Sneath\'s article  with a short list of tips for making the most out of the TechEd.  Here\'s the list of sessions, that I will probably attend at:    Visual Studio 2005 and SQL Server 2005 : Exciting New Roadmap To The Future   Visual Studio 2005 & SQL Server 2005 : Come get the details on an exciting new product announcement!   Visual Basic 2005 : RAD Development for the VB Developer   eXtreme Programming: the alternative way to building high quality systems   Visual Basic 2005 : Language Enhancements   SQL Server 2005: .NET Framework-Based Programming in the Database   SQL Server 2005: What\'s New in Replication   .NET and Diagnostics   Visual C# 2005 : IDE Enhancements for the C# Developer   Visual Basic: Tips and Tricks for Optimizing Your Applications   FrontPage 2003: Designing Customized SharePoint Services and Sites   RAD Data Access with Visual Studio 2005 â€“ the Good, the Bad, and the Ugly   ASP.NET: Best Practices for Performance   .NET Architecture Clinic   Migrating ADO to ADO.NET   SQL Server 2005: Web Services   ASP.NET: Tips and Tricks for Building Server Controls   Windows Forms: Controls Tips and Tricks    I will probably be changing this list thoughout this two weeks before the grand beginning of the TechEd.   '),(17486,3601621,'11,June,2004','2004-06-11 00:00:00','    Right now I\'m dealing with the following TechEd related dilema: to bring a laptop or not?   I\'ve found an interesting blog post by Bill Evjen who has prepared  urlLink list of things that one should bring to TechEd . I know it\'s been originaly written for the San Diego TechEd Event, but I guess it doesn\'t really matter as they\'re probably all pretty much the same (my guess though).  I\'ve never been to TechEd before, so I was just wondering about the  laptop  thing. I\'m not quite yet sure either I will get a laptop (don\'t have one of my own yet) by that time or not. So I was thinking: will you guys have your own laptops as well or not. And why? Thanks for your comments (if I get any ;)).  Other than that, I do completely suggest anyone who hasn\'t seen Bill Evjen\'s list yet, to follow  urlLink this link  and check it out - I found it extremely informative and especially loved his suggestions such as the one related with the business cards and writting messages on the backs of them.   '),(17487,3601621,'10,June,2004','2004-06-10 00:00:00','    Yupee. This is my first blog post. I must say that a few months ago I didn\'t have any idea what blogs are and would probably completely dissagree with anyone saying that I will have one of my own someday. Since the last couple of weeks I have come to the conclusion that it would be worth to try having my own blog where I would discuss things that I love and have passion for.  Therefor I have decided to discuss my work and things that are related to it one way or another. There are definately two subjects that I will be dealing with in the near future...first is  urlLink my own HTML editor  (which is now getting completely rewritten from VB6 code into the VB.NET and C# code) and the second is the  urlLink Microsoft\'s TechEd event in Amsterdam .  As it goes for the TechEd event, I hope I will be able to share my reflections from the event with the rest of you and give at least some useful information about some of the interesting stuff where I will attend at. And as for the HTML editor goes: most of all, I hope you will be able to give me some useful comments and suggestions throughout the process of developing, which will help me accomplishing the goal of creating a user-friendly and innovative HTML editor using the latest Microsoft\'s development tools.  I guess I should also tell you a bit more about myself at this point: I am a student in Ljubljana, Slovenia (which is one of the new EU member countries), but usualy spend most of my time developing all sorts of software. One of the major projects, that I have actually started years ago, is  urlLink my own HTML editor  that I have mentioned before and will be the source of many discussions here. Besides that I have one lovely 21 months old daughter with my lovely girlfriend that are together truly the greatest source of joy and fun for me right now! Photos of both of them will be posted somewhere in my blog soon...  Okay...that\'s it for now...it\'s 2a.m. here and I guess it\'s time for me to go to bed. Good night everyone, and read/write you soon!   '),(17488,2957170,'02,August,2004','2004-08-02 00:00:00','   i am eminently capable--many skills but the best is learning quickly.  And the worst is an evaluating eye.  That\'s why i don\'t write: i never feel good enough about my words, even when they are good.  The best writing advice i\'ve ever received was from a movie (first a book by Charles Willeford),  The Woman Chaser :  \"All I know is that you have to write something, get anything on the page.  Then you can read it over and rewrite.\"  Something like that.  But i\'m not trying to be a writer.  I\'m not.  i\'m trying to be me and--unfortunately--i feel as though i am succeeding.  Is this enough?  Is this all that i am?  What will i do, what will i do.  With any luck, not what i\'m doing now.  So many little failures, many my own, belying my smiles and laughs.  So many chances, some foolishly ignored and others cruelly taken.  As my life and country songs converge, i can\'t help but feel... FUCKED.  And the image i\'ve built up, the image agreed upon by so many--where did it diverge from whom i really am?  i remember: sleeping two-thirds the day away, and then sunshine and suntans and books well-read,  and then lost again.  Was it the Old Man\'s unnatural philandering?  Am i growing into Uncle Mike, the ubiquitous Crazy Uncle?  How much do i have to lose (both readings)?  How long can i let myself down.  '),(17489,2957170,'02,August,2004','2004-08-02 00:00:00','   In my short time, i have become fascinated with food.  Of the few instinctive necessities for living, the act of eating has the greatest potential for enjoying the means, with the end naturally coming about from the act; in other words, one has to eat in order to live, and one can eat poorly or richly.  As a counterexample i posit excretion: a good shit is one in which nothing goes wrong--you cannot up the enjoyment, you can only minimize discomfort (eating a lot of fiber, for example).  The chemical reactions taking place within the body upon consumption of sustenance do not care if your cereal is stale, if your olive oil is extra virgin, if your corn meal is in tortilla or enchilada format.  Your tongue and its buddies, your nose and your jaw muscles do care.   [Granted: the reactions do care if your sugar is nothing but fructose corn syrup, and if your cereal is full of dyes and diglycerides.  A whole \'nother topic.]   My question: Why eat foods that do not satisfy your aesthetic when you do not have to?  Considerations of economics and geographics aside, why eat \"bad\" food when you can eat \"good\" food?  i do not have an answer for that question posed to myself.  Many plants and animals have been domesticated in the past 10,000 years, to the betterment and detriment of humanity  [another topic] .  Cows, wheat, gourds, rice, chickens, llamas, and others that we don\'t eat or use anymore.  All were domesticated--their evolution husbanded by humans in order to maximize their benefits to and for our consumption--because they could be, not because of their aesthetic values.  From a mechanical/utilitarian perspective, i can see, all domestication is beautiful--but rare is the utterance \"Pigs are gorgeous.\"  Many people don\'t even appreciate the taste of pork--crazy, i know, but true--though it has nutrients the body wants.  But we\'ve come to a point in our scientific advancement where one does not have to eat pork if one does not want to, because the end of eating can be accomplished by eating other things, whether attractive or ugly, delicious or nasty.  But where do the twain meet, what food is easy on the eyes  and  simply good and good for you?  Sunflowers  are  gorgeous.  i find those large blossoms on their stately stalks to be the most beautiful biomass dedicated to human consumption.  The seeds are very healthy--the shells great for cleaning teeth--and the oil pressed is healthy as well.  The salt added after roasting is necessary--especially if you eschew processed foods loaded with salt--and, to my tastes, delicious.  It is for these reasons that i hope someday to harvest my own field of sunflowers, worn out after following the sun\'s path during a hot summer, and selling at reasonable cost \"Anthony\'s Organic Roasted Sunflower Seeds w/ Kosher Salt\".  It is in ways like this that i hope to continue my enjoyment of my basest needs.  '),(17490,2957170,'02,August,2004','2004-08-02 00:00:00','   And so in my 27th year i continued through the dark forest.  Caught in a loop with ends in Berkeley and Sacramento, i\'m once again unemployed (how unattractive that sounds!).  This time after one single day of floor training at Roseville\'s Old Spaghetti Factory, employment at which--for better and worse--offended too many of my sensibilities; basically, it was this: i cannot serve food that i do not enjoy, even when practically penniless.  It bothers me that i cannot sell out even when i want to.  But there\'s more: Tyler practically yelled at me for \"wasting\" my brain at such locations, before recommending an Elephant Bar at which his friends work.  And he\'s right (about the brains), but how to proceed?   urlLink McGeorge , UOP\'s respectable 3rd tier law school in Sacramento, has rolling admissions, a safe bet for a Cal alumnus with a 167 LSAT--but what do i do for the next five months?  There do not seem to be many opportunities, but i\'m looking.  But she keeps me shamed into focus: Thank you, Sarah, for stoking my ambissions and salving the pains of my failures.  i want to be better because of and for you.  Next stop:  urlLink Sacramento Natural Foods Cooperative .  i do care greatly about good food and cooperative businesses.  With luck they may have a bagging position for an ideological foodie.  '),(17491,2957170,'02,August,2004','2004-08-02 00:00:00','   US soldiers mistreating Iraqi prisoners?  Not surprisingly, it is so.  The photographs recently aired on \"60 Minutes II\" (perhaps the worst title for a news program, the suffix beating out the utter banality of \"Investigative Reports\") prove what any one of us who have been fighting cognitive dissonance should know: Our military forces are not the professional, ethical body as portrayed by, well, anyone in power.  Naturally, like the officer who documented these abuses, many uniformed representatives of American Power are fine people.  But many are not.  The fact that the abuses took place within Abu Ghraib prison, where many of the rape rooms to which President Bush refers were located, is both poignant and appropriate.  Abu Ghraib, like Evin prison outside of Tehran, was and is a symbol of former President Hussein\'s outrageous abuses--the metaphysical locus of the prior regime\'s worst atrocities against their Sunni Arab brethren as well as the subaltern Kurds and Shi\'ites.  Ostensibly a convenient (re)use of existing infrastructure, the US\'s co-opting of the physical prison is horribly foolish, literally placing the US military in the position which former President Hussein\'s security forces held: Incarcerating dissident Iraqis, torturing dissident Iraqis.*  Towards the end of \"Michael Collins\", the film about the fight for and founding of the Republic of Ireland with Liam Neeson as the eponym, the Union Jack is lowered from whatever government building (was it the Castle?) and the new Irish Republic flag is hoisted.  Nothing really changes.  When the Goths overran Rome in the Fifth Century, Rome did not fall into darkness, giving up all Europe to the illiterati.  In fact, the successive Gothic kings assumed the decadent mantle of their Roman predecessors, carrying on.  The structures of oppression are exactly that, and it is never surprising when the supplanting power inherits the ideals--wedded to the structure--of the previous power.  After revolutions, in which the supplanting power is necessarily native, the new governments are often nominally divorced from the previous government, exemplified by retaining the term \"revolution\" in the official name and retaining the image of militarism (ostensibly against the previous government), such as the Revolutionary Government of Iran and Fidel Castro\'s uniform.  But without dismantling the existing infrastructure of oppression, the tendency towards creeping remains, creeping oppression by the powers, creeping indignation by the people.  Reza Shah built--and filled--Evin prison long before the Revolution.  It is in everyone\'s best interest that the US scrap Abu Ghraib, recycling the salvagables and such.  Build a new prison somewhere else.  How much of a publicity coup would that be?  How much respect would that win the Occupation from Iraqis and Arabs worldwide?  Fuck the pragmatic course of putting Iraqis back in the prison, because the pragmatism is certainly an illusion--and, i hope to dog,  not  a red herring.  Because if the US occupied Iraq in order to colonize it, then the despicable actions of soldiers from a supposedly-civilized nation are completely appropriate to that end.  Scare the natives into submission, and what better locale than one infused with decades of the most heart-wrenching cries, the blood of thousands of innocents, the broken dignity of men and women and children, flayed flesh, scorched flesh, severed flesh, tears and teeth, the piss and shit that runs down your legs when your throat is throttled.  i do not believe in ghosts.  If i did, however, and if they could act upon the physical world, i would not be surprised if their message to the US troops who replaced the Ba\'athists would echo that Navy bomb, readied to drop on Afghanistan:  urlLink http://bushspeaks.com/img/fag_bomb.jpg   * i use \"dissident\" as a value-neutral term, denoting a (perceived) opposition to the current power.  '),(17492,2957170,'02,August,2004','2004-08-02 00:00:00','   What is the appeal of self-destruction?  The big gun, i suppose: Self-destruction is art.  It\'s beautiful--living rather quickly, dying at an early age, a comely cadaver... .  A firework flash takes the breath away; a 15-watt eco-bulb burning, at best banal.  Like driving the lane--a spurt of sport--taking the contact and trying, trying to jam it in.  And then on your back, surprised,  resigned, the release of so much steam--respite.  Unfortunately, the luck of the artist is not pegged to chemical dependence, it\'s a spike in the graph.  A burp from your tupperware, gasping, needing to be heard.  Sometimes you just toss out the funky food.  Sometimes it  becomes , a transitive from within, an intransitive from without.  Beer, wine, vinegar, fishsauce, garum, sauerkraut, kimchee; apotheosis, deification, the Infinite.  Art is not always for the patient.  Sometimes a catalyst--some yeast, some salt, some vinegar--transforms the ugly to the beautiful.  More accurately, the useful to the more useful.  The hyper of the above foodstuffs is informative--> cabbage (lively and moribund) pickled/fermented becomes sauerkraut/kimchee, adding staying power.  Grape juice?  Has nothing on the shelf-life--or mystery--of wine.  Chris Farley will always be known as the Funny Fat Man from the 90s, David Spade will fade into syndication.  And self-immolation is no longer my thing.  The special (pertaining to species) benefits no longer obtain: i need look no farther for a mate, so the advantages of self-denial (or, as appropriate, self-indulgence) are wasted.  Ideology, and its concomitant (may i coin?) eigen-negation, still can display my stubborn refusals to do what is expedient for myself instead of Universally Right, thus hinting strongly at a selflessness that i may apply beneficially to my love object-subject.  But it salves only now the Princess Butterfly, from the Valley, and not even that much.  In other words, i have grown up.  Mostly.  The smaller gun (though at no cost to efficacy): Trapped!  Without an escape route.  Without a plan, design, future.  \"Word acrobatics performed without harness or net, I\'m so full of shit.\"  Not unfair, as criticism goes.  While away the time, burn that too-short wick, hope that something--ANYTHING AT ALL!!!--happens.  Make a goal for yourself and achieve it.  Sometimes you get so alone, but don\'t hold your bottled brethren to the same standards--they\'re always hanging out, though you may not always be on the List.  Just bide your time, shuffling from hole to hole, and maybe Jack Nance will rifle through your Life to your advantage.  But i would not wait on that, Mr. Chenowski.  [And now, in chorus]: I\'ve got so much to offer.  Every moment unrecognized is a novel never written, a song not sung, a canvas untouched, a memory.  And memories are worth only so much.  And maybe less, if you can keep them.  History is written by the winners, they say, but who wins?  And who never played in the first place... ?  *   *   *  Even more personally, i\'m not looking for It.  i\'m not striving, nor starving.  i\'m living in a whence-removed world.  And who cares?  Who heeds Dunne\'s call?  Does it matter whether i have health insurance?  Well, my world is functionally rather small anyway, so i should simply keep my eyes on the horizons, steering toward the lumps as they rise.  But why ignore the peaks at hand?  To complete the lyric, \"but I will remain until this self-awareness fades, \'til I defeat the purpose of this soapbox that you\'ve made.\"  Take heart whenever you can, it\'s not always easy to find.  '),(17493,2957170,'02,August,2004','2004-08-02 00:00:00','   But first, a side note: i am now the 35,000,001st American living with tinnitus.  For a week now, there has been an incessant, unwavering ring in my left ear.  i also feel that, in negative compensation for the ring, i\'ve functionally lost a little hearing--though unexpected sounds (cars driving by, restaurants, music...) raise my hearing in that ear, and with that my ring.  i haven\'t been listening to music, i\'ve been quiet and submissive, sleep is not consistent but ganj helps.  i may have to live with it, give up the music arts for the plastic arts.  Or words.  *   *   *  And speaking of, how \'bout those shrouded coffins?  Nothing like the poignancy of the remains of men and women who died--most certainly a painful death--for some random reason or another.  The poise of the marine guard flanking the ten or so stainless steel caskets, impeccable United States of America flags on each, the poise was magnificent.  This honor--this amazingly egalitarian ceremony of honor--to each dead soldier nomatter what branch of service, this beautifully anonymous process is literally enshrined at the Tomb of the Unknown Soldier, flame burning perpetually.  These soldiers died as they worked and fought, in indistinguishable uniforms reifying their accepted placement in a grand process (in most cases  for  their country).  So, when White House spokesman Trent Duffy says, \"In all of this, we must pay attention to the privacy and to the sensitivity of the families of the fallen, and that\'s what the policy is based on and that has to be the utmost concern,\" in explaining the enforcement of the Pentagon ban on coverage of the return of war-dead from overseas, i gotta wonder.  A certain Ms. Silicio worked for a contractor in Kuwait, doing something in airplanes transporting coffins back to Dover Air Force Base, the longtime point-of-entry for dead US servicefolk.  She took a shot--camera-wise--of the flag-draped coffins and it ended up in the Seattle Times.  She was fired.  But--the point is--how was the privacy of the families of the dead compromised?  On what day was that photograph taken?  Which flight?  Among the bereaved, is there a bill of lading for that shipment circulating?  Cross-checked with departure schedules and time-changes?  Verified by concerned witnesses of their own mortality, who shepherded their brothers\' and sisters\' remains back home, thinking \"There, but for the grace of God, go I\"? NO! The privacy of the family is ensured by the anonymity of the process--no names are stenciled onto the flags in cholo-style \"Old English\" script.  It is every grieving family\'s son or daughter in one of those coffins, and the dead are our own too.  Soldiers obviously treat their own dead with sanctity and care, and that care is afforded every soldier; war is a jerk, but honor is honor.  i like the photographs--they keep it real.  *   *   *  What happened to Dennis Kucinich?  Whe did the press stop covering him six months ago?  i really like the guy.  He\'s going to be played by Steve Buscemi in  Secondary Colors  and he\'s going to be radical!  Representative Kucinich is amazing: that a man that kooky-looking could be so positive!  And he looks like the guy in  urlLink Red Meat !  And--of course--i have great respect for his socio-politico-economics.  Universal healthcare?  Check.  Increased US  participation  in the international arena?  Ditto.  Dismantiling the military-industrial complex (of which we were warned by valedicting President Eisenhower, remember)?   BAM!!!   And that\'s why i voted for him on 2 March, here in California.  Both he and the Reverend Sharpton should nail some planks for the DNC\'s fall platform, in my opinion.  Hell, they should host a talkshow together!  Al could be the smooth-talking main host quipping and jabbing, and quirky Dennis doing color-commentary (odd) and fact statements and such.  Dog, that would be great.  But what i\'m saying is, let\'s hear some more from the guy.  On the local NPR i caught him, sick in Hawai\'i and flight delayed, drop some science and economics.  Great hippie fare, i ate it right up.  Hey, i think i hear the \'phone ringing... .  '),(17494,2957170,'02,August,2004','2004-08-02 00:00:00','   urlLink    Up against the wall, Commies!  '),(17495,2957170,'02,August,2004','2004-08-02 00:00:00','   Say what you will--for i know i have--about the stultifying effects of modernization, the anaesthesia of modernity, the banality of the modern.  Say it, but praise the process for clean water.   Of dog and man  Clean, reliable and clean water is perhaps the simplest proof (still short, in my estimation) of the existence of a benevolent God.  Knowing that i--and any number of millions of people in many locales--can safely assume that the water flowing from the tap will not harm me; that is an expression of the highest goal of social organization: to provide every member of the community with the basic necessities of living.  Would that we could all realize that health care is of similar consideration.   One last round for all my friends  Water, clean and reliable and ubiquitous, though conservation is paramount.  While enjoying the benefits of constant access to clean water, society should set a task for its experts: achieve an understanding of earth\'s most precious resource and its role in the earth\'s entire ecology, so that we may best husband it.  Both eyes to the future, as once lost, it cannot be reclaimed.  As with gasoline, which fuels an infrastructure that should hang a left towards creating a new infrastructure, one not dependent on the dirty and dwindling petroleum supply, while there\'s still gas in the tank.   Down the drain  Things are likely to get uglier in certain places on earth.  Occupied Palestine is in a bad spot due to Israel making its desert bloom.  Turkey lies upstream from Jordan.  Africa.  The dwindling Aral Sea, now Stain, festering with the USSR\'s graveyard of chemical tests on a  former  island, Sakhalin, a  windwhipped former  island.  The Pentagon\'s doomsday office published a report detailing how global warming events could, within 50 years, alter world weather patterns in such a way as to necessitate massive steps towards martial law in order to protect United States borders from migrating, starving peoples of the earth.  Hm.  Could?  Owing to the gravity, we might want to keep this scenario in mind.   Resources not too wasted  But not to forget clean water\'s history to our peoples.  In the West, clean water was scarce when people became communal, poo and nascent agriculture poisoning the streams.  So the grape, and the barley and hops (and salt, dear salt!), and the fermentation: a clean source of potable water ( vodka  is the Russian diminutive of  voda , water).  Noah, according the the book, was the first vintner.  Either way, Lot was wasted when his daughters straddled him in that pimped-out cave.  The fact is, up until the 18th century, alcoholic brews were the only reliable source of clean water, keeping the West in a constant, if mild, state of inebriation until we learned microbiology, or something (funny: alcohol is an Arabic word because it was North African Moors who first distilled alcohol in the 9th century, disconcerting considering the Islamic injunction against the grape--but Muslims also wash their arms face and nostrils before praying--associate!).  In the East, some people lack a gene that produces an enzyme speeding the metabolization of alcohol, but Hey! Let\'s boil some water for tea!  Tea is great stuff.  i mean, really great.   Lack-rymose  So drink wine, shower with friends, make coffee, boil some basmati.  Be thankful for the water, reliable and clean and there for all the right reasons.  Don\'t misuse it, not for the wrong reasons.  Resist notions of giving control away to ogres who do not understand its true significance: a sacred, priceless structure--no moneychangers in dog\'s house.  And appreciate what we have, for it may not always be with us.  '),(17496,2957170,'02,August,2004','2004-08-02 00:00:00','   Yes, i still love the Word, and the Word is in lyrics.  i take serious satisfaction in well-written words.  As we all know, lyrics in popular music are not often a focus of hard work and skill, ruining a good structure by half-assing the cap.  Some musicians can compose an edifice so beautiful that it does not matter, or does flatter, the sagging lines expressed above the music.  Others make it so tall, hiding the words in the clouds.  Others do it all, as The Streets does.  *   *   * 2002\'s \"Original Pirate Material\" is--and this needs to be said outfront--England\'s \"Marshall Mathers LP\", trading Detroit and LA for Birmingham and Manchester, in both music and content.  What they have in common is an approach to flow remarkable for sensitivity, towards the aural quality and appropriateness of diction, towards the rapper\'s position  vis-a-vis  himself, friends, society (though Eminem gets to discuss media attention too).  Mark Kinner, who is The Streets, guides us with his conversational style, weaving in and out of the distinctly British rhythms.  That means: clubby, guess he listened to Massive Attack, beats with small cracks, pleasant.  And engaging, the tracks never obscuring the tone-poem, for good reason.  i\'m not going to post any as i feel that any amount would be too little.  Good descriptions of getting wasted, also Eminemish.  *   *   * Now some personal stuff.  \"Sexxy Underwear\" (the extra \"x\" because they\'re extra-sexy) is a really sweet song, i feel.  Yes, it certainly is a Dr. Masseuse rocker, with grunts and emotion, but i wanted the love to be evident.  Hence starting in media res with  \'Cause you know, i\'ve got to go--i\'m leaving California , speaking to someone who\'ll miss me dearly.   Heading out for the big show , i explain, acknowledging then that i\'ve got  a brand new cell \'phone and new haircut , which means i\'m looking good but i\'m still connected and faithful--you know how musicians can live on the road.   \'Cause you know i won\'t be gone for too long, girl, when i find my treasure true--it\'ll lead me right back to you  Yep, continuing out away from home, but aware of how important home--and she--is.  And also echoing the fist line.  Second verse, i\'m  getting high, strung up like a kite , suffering under  this West Coast cross i bear .  Alright, pot, blah-blah.  But visualize it!  A kite, traditionally with that cross bracing it, connected by a thin string to the flyer, safe on the ground.  And some internal rhyming, \"West Coast cross\" sounds cool.  And then realization,  wherever i go, i take myself with me--here i am, here i am , Where am i?  Here i am, not entirely alone, mayhaps on-stage, shouting in one breath  \"If you feel what i\'m giving it\'s a reason to believe in giving power to the people RAISE YOUR FIST UP LIKE A STEEPLE!\"   And why, why music, why steeples, why love?  Hell, why anything?   \'Cause we\'re all looking for something to believe in .   \'Cause you know i won\'t be gone for too long, girl, when i find my treasure there--you\'re sleeping in you sexxy underwear  Well, putting it a little more rootedly, to pun away.  i never said that i didn\'t enjoy a little cuddling now and then.  And then Robin explains--and rightly-- i need something write about, i need the words i can sing, i see those words wrapped around your hips and cover up your lips --to woman\'s underpants, may not be the best thing on earth, but close to it.  And then a bunch of rockin\' and chorus repetition.  Go Bears.  '),(17497,2957170,'02,August,2004','2004-08-02 00:00:00','   Through my somewhat extensive readings of the--mostly Western--canon, only one writer has altered my frame.  Friedrich Nietzsche had the incisive reasoning, granted only a few, to not only recognize the frame of culture [and split infinitives], but to pin it down and--so doing, as in sub-atomic physics--change it.  In this regard he joins the cadre of folk like Aristotle (scientific method), Martin Luther (institutionalized individualism--a nifty oxymoron), and Freud (hidden processes of thought determined by experience).  In fact, i tried to approximate Nietzsche\'s course of study after reading him early in college.  i had been studying Latin, a rather native speaker of German, then studied Old English, Middlehigh German... i aimed for philology.  The term itself, qualifies \"love\" (Gr.  philos ) with  logos , an expansive word and trove of denotations, the most important being \"word\" and \"argument\" and, in that striking Gospel of John, \"Christ\" (\"In the beginning was the  logos  and the  logos  was with God.\"  John 1:1).    In a break from the more mundane, though nonetheless brilliant, German philologists who mapped the history of language in the 19th c., Nietzsche applied the socio-philosophical connotation of  logos  to his study of language, most importantly and remarkably in  Towards a Geneology of Morals .  The first of three sections of that work describes the evolution of the competing moralities of \"good v. bad\" and \"good v. evil\", tracking these simple adjectives common throughout Indo-European languages.  In short, the culture of positive, or \"Master\" morality, start from the premise that they are good, concluding that the Others are therefore bad--common or  vulgar , as the Romans said.  Competing, the Man of  Ressentiment  decided that the Others who were above him and circumscribed his life were \"evil\", necessarily making him good.  This is the much-maligned \"Slave morality\", to Nietzsche the foundation of Judeo-Christian thought.  His argument is persuasive and his scholarship--simply impeccable.  Exposed to this text, and with the help of Judith Butler\'s excellent professing, i have tried to internalize that lesson.  It probably helps that i am a male of northern and western European descent.  So, not to engage in Christian-bashing but to post my favorite aphorism, here is this gem (translated by Walter Kauffman, whose understanding of Nietzsche\'s great style makes him by far the best English interpreter).    \" The philology of Christianity   How little Christianity   educates the sense of honesty and justice can be seen pretty well   from the writings of its scholars: they advance their conjectures as   blandly as dogmas and are hardly ever honestly perplexed by the   exegesis of a Biblical verse.  Again and again they say, \"I am right,   for it is written,\" and the interpretation that follows is of such   impudent arbitrariness that a philologist is stopped in his tracks,   torn between anger and laughter, and keeps asking himself: Is it   possible?  Is this honest?  Is it even decent?          What dishonesties of this sort are still perpetrated from   Protestant pulpits today, how crudely the preachers exploit the   advantage that nobody can interrupt them, how the Bible is pricked   and pulled and the art of reading badly formally inculcated upon the   people -- all this will be underestimated only by those who go to   church either never or always.          In the end, however, what are we to expect of the   aftereffects of a religion that enacted during the centuries of its   foundation that unheard-of philological farce about the Old   Testament?  I refer to the attempt to pull away the Old Testament   from under the feet of the Jews -- with the claim that it contains   nothing but Christian doctrines and belongs to the Christians as the   true Israel, while the Jews had merely usurped it.  And now the   Christians yielded to a rate of interpretation and interpolation,   which could not possibly have been accompanied by a good conscience.   However much the Jewish scholars protested, everywhere in the Old   Testament there were supposed to be references to Christ and only to   Christ, and particularly to his cross.  Wherever any piece of wood, a   switch, a ladder, a twig, a tree, a willow, or a staff is mentioned,   this was supposed to indicate a prophecy of the wood of the cross... .   Has anybody who claimed this ever believed it?\"  NB: The Nietzschean Scholarship Rule of Thumb (NSRoT): Whenever making an argument on the interpretation of Nietzsche, find at least three (3) passages which counter your argument.  '),(17498,2957170,'02,August,2004','2004-08-02 00:00:00','   urlLink    Holla  '),(17499,2957170,'02,August,2004','2004-08-02 00:00:00','   i think it\'s just great.&nbsp; Really, all this gay marriage--garriage--division.&nbsp; Sputtering Santorum (R - Pennsylvania) going on about national and marriage security, and he the most sculpted-yet-innocent cakeboy of the group.&nbsp; Riiiii-diculous.  But what\'s is so great is that the frame of the debate has shifted away from less enlightened moorings.&nbsp; It is no longer a debate about whether fags can buttfuck, thank goodness, thank SCOTUS, but whether they can marry!&nbsp; I mean: Garry!&nbsp; That was a quantum leap, from, say three years ago.&nbsp; i cannot say whether there will be steps taken back, as could well be the case, but coming so far so quickly is, how would one say, engruntling?  *&nbsp;&nbsp; *&nbsp;&nbsp; *  Here\'s a link to an epicurious recipe for trout.&nbsp; Read that if you like trout, read the comments in any case, \"... golfer\'s outfit... brightly colored snake...\":   http://www.epicurious.com/run/recipe/view?id=10049&kw=&amp;action=filtersearch&QueryZip=%28%28trout%29+%3CIN%3E+%28CNContent%2Ctitle%29%29&amp;Filter=recipe%2Dfilter%2Ehts&ResultTemplate=recipe%2Dresults%2Ehts&amp;QueryText=%28%28trout%29+%3CIN%3E+%28CNContent%2Ctitle%29%29&Collection=Recipes&amp;ResultStart=31&ResultCount=15&amp;BrowseLink=%3CA+HREF%3D%22%22%3EQuick+Search%3C%2FA%3E+%26gt%3B+keyword%3A+trout   Helluva link.  *&nbsp;&nbsp; *&nbsp;&nbsp; *  i sent an email to my sister Sara today.&nbsp; She lives in Munich.&nbsp; You can  urlLink translate  the German.&nbsp; i was having fun.&nbsp; The sZ is the  urlLink sueddeutsche Zeitung , the big newspaper from Munich.   So i decided to kick it up a notch today, adding another stop to my news\' site* circuit. &nbsp;This is in the sZ\'s write-up on the discovery of former engineer--present and perpetual victim--Paul Johnson\'s head:    &nbsp;  \"Nun habe Johnson etwas von dem zu SpÃ¼ren bekommen, was Muslime seitlangem von Raketen aus Apache-Kampfhubschraubern erlitten hÃ¤tten. Der49 Jahre alte Amerikaner hatte am Ziel- und Nachtsichtsystem der Kampfhubschrauber gearbeitet.\" (   http://www.sueddeutsche.de/ausland/artikel/726/35691/    )    &nbsp;  Maybe i\'m just too accustomed to anglo-journalism, but this has the ring of rhetoric, no? &nbsp;What about those subjunctives \"habe\" and \"hÃ¤tten\"? &nbsp;Almost, um, Arabesque? &nbsp;However, i had also never before come across the word \"enthauptet\". &nbsp;The elegant efficiency of thatGerman construction makes \"decapitated\" sound like a bag of shit. Somewhat.    &nbsp;  To change the subject--severely--, here\'s a picture of my Sarah. &nbsp;You can remember which is whom were... what? &nbsp;i mean: you can remember which one you are by saying, \"\'h\' is for Hebrew.\" &nbsp;She and i get along alright... LIKE A FOX!    &nbsp;  Ahhh, thanks for letting me tickle the ivories for a while. &nbsp;Strokin\' the keys, that is. &nbsp;i hope you\'re having a wonderful evening.&nbsp;&nbsp; Hellos to everyone.    &nbsp;  take\'r easy-anthony    &nbsp;  * i\'ve felt for a while that in terms such as \"news site\", the \"news\" is not an attributive adjective. &nbsp;Rather, it functions as a nominal genitive, following the constructions \"reading the news\" and \"watching the news\", where \"the news\" is the direct object. &nbsp;That\'s why i added an apostrophe. &nbsp;And then an asterisk. &nbsp;Which is why you\'re reading this.  '),(17500,2957170,'22,July,2004','2004-07-22 00:00:00','      '),(17501,1405024,'20,April,2003','2003-04-20 00:00:00','   man, i\'m really high...  '),(17502,1405024,'20,April,2003','2003-04-20 00:00:00','   so...wow, that was a good hit.  Happy 420!  :D  '),(17503,1405024,'20,April,2003','2003-04-20 00:00:00','   it\'s 4:20am on 4/20!! how  exciting  !!  '),(17504,1405024,'20,April,2003','2003-04-20 00:00:00','   This past week completely sucked, for reasons into which I will go later. But it\'s 4/20, so today is bound to be a great day. Right?  '),(17505,1405024,'12,June,2003','2003-06-12 00:00:00','   Okay, so, I really, really shouldn\'t be doing this.  Doing what?  Writing this shit right now.  Not like I ever do, but right now for some reason, right now of all times, I\'m writing.  I wish I could write for a long time, I\'m really in the mood...but I totally have to pack up my stuff.  I have sooooo much more shit to pack...I\'m sadly moving from Berkeley, my hometown of almost five years now, back to L.A.  The Valley.  Ugh.  I don\'t know if I will be able to adjust myself to the superficiality that runs through the veins of the city - after living here in Berkeley for so long, I can\'t stand anything of the sort.  Come on dude.  You\'re normal if you are a gutter punk living on the street asking for change for beer and pot.  You\'re normal if you consider Bush and Cheney the true terrorists of the world.  You\'re normal if you have like 20 piercings and pink hair and strip in Tijuana for money so you can buy pills and then get arrested and...oh wait, maybe that\'s not normal.  But...L.A.?  You\'re only normal if you look and act and talk and think just like everyone else, and wear the same clothes, and do the same things, and have money for this and that and other stupid shit.  What the fuck am I gonna do dude?  I suppose My Mother will want me to slowly assimilate - in fact, she\'s already made me a hair and nail appointment for when I get down there.  Not that that\'s particularly an L.A. thing, I mean, I know people who do that here (even though I hate her...er, them), but I think it just shows very well how the culture is down there.  Fuckin shitass.  '),(17506,1405024,'28,July,2003','2003-07-28 00:00:00','   Fuck.  I feel so...depressed right now.  Why?  I\'ve been fine all day.  Well, I did sleep all day, maybe that\'s why I was fine.  Yes, I went to sleep earlier, and wound up sleeping until about 4:45.  Fucking a!  When I said earlier that I\'d probably get in bed and sleep for 8 more hours, I thought I was joking and exaggerating.  But noooo.  I guess I have to watch what I say!  I really fucking want some ice cream right now.  I wish I had some.  Maybe I\'ll drive down the hill and go to Safeway to get some.  I think they close at midnight.  '),(17507,1405024,'27,July,2003','2003-07-27 00:00:00','   It feels really good to close my eyes.  I don\'t exactly feel tired, but I guess I am.  I\'m trying to decide if I should do some more oxy (I did some like 4 hours ago or something), and right now I\'m leaning towards no.  It\'s already late, I\'m starting to get tired...I can just save it and do it tomorrow.  However, I will smoke a bowl (of course).  Sometimes I wonder what my lungs look like, since I smoke so much weed, and I did smoke a bunch of cloves a few years ago, and now I smoke a cigarette now and again.  But mainly, the weed.  Are they black and dripping with resin?  If someone put my lungs in a pipe and smoked it, I bet they\'d get really fucking high.  Isn\'t that disgusting?  I mean about the smoking lungs part, not the part about getting high from my body parts.  That\'s actually pretty awesome.  Now I\'m trying to decide again about the oxy.  It would be pretty good right now.  Bah.  I guess I\'ll do a small line.  I just want that taste in the back of my throat.  Really, there\'s no need to be concerned.  It would be so much better with a little beer though, I wish we had some beer other than Keystone Light.  Yuck.  Give me a break!  I hate when Ed buys that shit.  I like real beer, baby.  Fuck, I just dropped the small amount of the 80mg pill of oxy, the last bit I have left.  It\'s dark in here, and the carpet is that weird offwhite-grayish color (the pill is white).  How the fuck am I gonna find it?  I already have a line sitting here, so that\'s not the problem - I just don\'t want it to get smashed into the carpet, or worse, have Jezebel eat it or something.  I better get my lighter and get on my knees.  Ooh, that kind of sounds erotic.  Now I close my eyes again, and I think, what am I doing?  Why am I doing oxy, I don\'t really need it right now.  It\'s already like 3am.  No, it\'s later.  Okay, I did it.  I did some oxy.  Now let me hit this bowl of trainwreck, yo.  Mar-i-juana is mmm mmm good.  Awesome!  I miraculously found the little chunk of oxy without too much effort.  Yay.  I feel so bloated today, it\'s kind of gross.  I don\'t really know why, but it could be because I ate, like, a massive amount of food yesterday.  Or at least I rememeber yesterday I felt like I did.  Who knows.  I also watched the movie  urlLink JFK  yesterday - it was super long, and I felt like nothing really happened in it.  But, it was still a good movie.  I got my toenails done all nice during it, a nice, shiny red.  Very pretty.  Soon I\'ll be able to take pics of these kinds of things and then post them for all to see. :)  Maybe it would come up under a foot fetish search.  Lol.  So yeah!  I didn\'t get to finish earlier.  I totally got a place.  I\'m really happy, but I keep having doubts about it since it\'s kind of small.  But, I am saving money since all utilities are paid.  I move in this week...over and over I arrange the place in my head.  I wish I could see it again, or I wish I knew the exact measurements or something.  I want to go over there again maybe tomorrow or something if they\'re okay with that, and I\'ll bring a tape measure.  Then I\'ll know exactly how much space I have to work with.  I can\'t wait to take pictures!  Seriously though, I don\'t know if I\'ll have enough space for much more than my bed, desk, and a dresser...I\'m sure I\'ll be able to figure something out though.  I hope.  I mean, I totally have enough stuff to furnish a two bedroom apartment, so I\'m really gonna have to pick and choose what I want.  Also, I haven\'t asked about pets, so Jezebel may be an issue - I want to ask about her, but I also don\'t want to risk losing the place.  Even though I know she\'s not a problem, other people may be weird about rats.  I\'m also trying to decide which philosophy of feng shui to use.  But, I still have some time.  I don\'t think I\'ll use the compass method, I don\'t like it.  I can\'t really remember the lighting situation either.  Eh well.  I\'m just so excited!  I can\'t wait to move in.  I\'ll have to decide what posters to put up too.  I wish I had money, I\'d buy so much new stuff that (I feel like) I need.  Especially stuff from IKEA, cute stuff for my new place.  Well, shit dawg...I\'m getting tired.  '),(17508,1405024,'27,July,2003','2003-07-27 00:00:00','   What the hell.  I was just writing something, then I stupidly clicked on a link and all my shit disappeared.  Good thing I had only written like 3-4 lines.  Mostly me complaining about...nothing, really.  Just complaining.  Oh well, I\'ll just start over.  Fuckin\' shitass.  Yeah, that\'s right.  Fuck-ing shit-ass.  It\'s already after 4pm on Sunday afternoon.  What am I doing with myself?  I haven\'t done anything all weekend.  I woke up at like 2:45pm or so, and I\'ve pretty much been sitting here at the computer ever since.  Well, not the whole time, but just about.  I made some tea like an hour ago, and most of it is still sitting here in the mug.  I think Jezebel might have actually drank (drunk?) more than I have so far.  No no, not really.  But she does seem to get way excited about getting to drink my tea.  She is so cute.  Hey...I got a place!  I have my own place!@#  I\'m so fucking excited, you have no idea.  It\'s the place in Montclair.  Or Piedmont, technically.  '),(17509,1405024,'26,July,2003','2003-07-26 00:00:00','   So I went to the cafe, made it there at like 7:50pm or so (he closes up at 8). He kicked some people out, then made me my drink for there. I love when he does that, it makes me feel cool. I watched him clean up, we chatted, I even did some dishes. I was really high. Then we went out to smoke our ritual cigarette and chatted some more. He asked me a little about smoking weed.   I managed to lose track of time, surprise. It was 9pm when I looked at the time, and I was still at the cafe. Fuck.   Later,I watched  urlLink Zoolander . I heard a bunch of people say it was really funny and pretty good, but I thought it sucked, honestly. A few parts made me laugh, but overall I\'d say it wasn\'t that good. Although, the more fucked up I got during the course of the movie, the funnier it was. That\'s not necessarily a good thing though.   I am so excited to be living alone...maybe I will feel all grown up. Maybe it will help me get my act together and get a fucking job. But no, really, I\'m just excited that I\'ll be more in charge of what I do with my time, and I\'ll be able to decorate however I want and arrange the furniture however I want and all that jazz. I can feng shui my heart out, with no one sitting there to tell me I\'m crazy and wasting my time.  I am waiting to hear back from these two studios I applied to. One is in the  urlLink Lake Merritt area , in this really cool art deco type building from the 1920s. You walk inside and it\'s like a hotel, not an apartment building. And the neighborhood is actually pretty nice. The apartment itself isn\'t bad either - its on the sixth floor, and has a bathroom, a full kitchen, and a good size, uh, area? The part that\'s not kitchen or bathroom. There\'s one closet, not too small, but closet space overall is pretty skimpy. The bathroom is pretty cool - there\'s this whole mirror-vanity thing with little drawers and a little counter and stuff, it\'s so cute. I still haven\'t heard back about it though. I\'d be taking over this other girl\'s lease for the next six months. I lied about a job - Veronica said I can say I work for her, so I\'d have something to put on the application. I did a credit check on myself yesterday though, and my credit was REALLY bad. It said I was high risk or something. That sucks. All the utilities in this place are paid too, except for phone and cable. Not bad. Supposedly there\'s no pets allowed, so I don\'t know what I\'d do about Jezebel - sneak her in or ask them about her. But maybe I won\'t even get it since my credit sucks so bad. Keep your philanges crossed, please.   The other place I applied for is an inlaw studio - basically, a room attached to someone\'s house with a separate entrance. It\'s in the more upscale Oakland area of  urlLink Piedmont , which is actually its own city for some reason. The place is within walking distance to  urlLink Montclair Village , the cute little outdoor shopping area with lots of rich white folk and no homeless people or panhandlers. It\'s all very nice and clean - you\'d never guess it was part of the stereotypically ghetto Oaktown. (Seriously, check out the link for Montclair Village, it totally made me laugh. It\'s a real fucking village!) The studio itself is pretty small actually - I think the ehousing listing said something like 300 sq. ft., but that didn\'t really help me when I read it, and it still doesn\'t help me now. It\'s about the size of an average room. There\'s no actual kitchen, there\'s a kitchenette - basically, kitchen stuff built in to one of the walls. There\'s no bathtub in the bathroom, but the bathroom isn\'t bad otherwise. There is a TON of closet space, which is really cool. And what\'s cool is that there\'s a window that overlooks the bay, and when I exit the studio I\'m actually outside looking at the bay on a nice deck with nice landscaping around me. This place is nicer than the Lake Merritt place, but it\'s smaller, and $45 more a month. A family lives in the house - a husband and a wife, and their two young boys (I think the older one is four years old). I\'m a little worried about smoking...but I\'m sure everything would be okay. It\'s about five to seven minutes away from A\'s house, driving. There\'s a busstop within walking distance (right in Montclair Village). Their ad also says no pets, but I\'m pretty sure they wouldn\'t mind Jezebel. They really liked me, the mom told me, so hopefully I\'ll get this place. All utilities are paid here too, including cable. I\'m wondering about internet access...I really want dsl. I can\'t stand cable modem, which I\'m using at A\'s house right now. And, since this place is pretty small, I really wouldn\'t be able to fit much of my stuff (furniture-wise) into it. That kinda sucks actually.   Eh well, we\'ll see what happens. Hopefully I\'ll get at least one of the two!  '),(17510,1405024,'25,July,2003','2003-07-25 00:00:00','   Snorting more oxy.  Smoking more weed.  Playing more minesweeper.  Fuck expert level, I suck at it.  '),(17511,1405024,'25,July,2003','2003-07-25 00:00:00','   ~just some information~   numen:  1. A presiding divinity or spirit of a place.  2. A spirit believed by animists to inhabit certain natural phenomena or objects.  3. Creative energy; genius.   (that\'s according to  urlLink this .)  '),(17512,1405024,'25,July,2003','2003-07-25 00:00:00','   LA, Part II   So I drove for a few hours, and a few hours more.  It was a full moon that night, Friday the 13th.  I kept having bad feelings, but I do tend to be a bit superstitious occasionally.  I just really felt like I was making a really bad decision.  But I just kept remembering something my mother had said to me earlier that day - think of the trip to LA as a several week long vacation, something to let me know where I really wanted to be.  I may still miss Berkeley in a few weeks, or I may be fine in LA, we could just wait and see.  I knew I wouldn\'t be fine...  I arrived in LA around 1:30am if I remember correctly.  I was super tired - I had only gotten like a total of 5 hours of sleep the previous two nights combined, added to way more physical labor than I was used to (which is...um, none), the labor that comes with moving large boxes filled with massive amounts of shit (not literal shit, of course).   I had to wake up at around 7am that day if I wanted to get to the testing site on time, which I did.  I went inside my parents\' house, found my stepfather passed out on the couch and my brother awake in his room playing computer games.  Nothing seems to have changed since I was here last, or the time before that, or the time before that....  My disassembled computer was sitting in the back seat of the rental.  Given the fact that the crime rate has been steadily increasing in the valley the past few years, and just my better judgement, I didn\'t want to leave my shit in the car.  I started to slowly bring in piece by piece of my computer, and recruited my brother to help.  Keep in mind he\'s only 12, so he isn\'t really big enough to carry lots of shit at once.  But, he helps with what he can.  As we were walking in from outside at some point, my brother accidentally let the screen door slam, which woke up my stepdad who was until that point still passed out on the couch.  He was all incoherent and angry, and started yelling at me for \"waking everyone up\" and \"what am I doing coming so late\" and \"why am I bringing all my shit in, I\'m just keeping everyone up,\" and...I could go on, as one could imagine.  I explained I didn\'t want to leave my computer in the car overnight, that I would just bring it in and get the rest of the stuff tomorrow.  He said that was stupid and I should just leave it and go to bed.  He even followed me outside and was yelling at me in the street.  It was kind of embarassing, and just reaffirmed my feelings that I was making a huge mistake.  All his yelling woke up my mom, who came down all groggy and confused from upstairs.  What\'s going on? she asked.  Wayne just started yelling, and she said of course I should get my computer, is he crazy?  She calmed him down, then bitched at me a bit for \"being incosiderate\" and whatnot, but I could tell she was just yelling at me to appease him.  She actually did a good job of mediating that night.  So this whole time Jezebel was in the car.  I cleared some space with my brother on top of his dresser for her cage, the same place we kept the cage last December when I went home for krismas.  I brought her inside - her cage was the last thing coming in for the night.  When Wayne saw me coming inside with her, he started yelling again immediately.  He was telling me some bullshit like she needs to stay outside in the garage or on the patio, and I was saying that was ridiculous.  I knew some ghetto street rat would get in her cage and rape her.  Rats don\'t belong in bedrooms, he yelled.  Wtf, I was thinking.  Why couldn\'t you tell me this before the moment I get here?  I asked. It doesn\'t matter, the rat isn\'t staying in the house. Why? Because I said so. Well, give me a reason. Because I said so. I\'m bringing her inside until you give me a reason. You better not bring her inside. I\'m putting her down so I can at least shut the door. What are you doing?  I said don\'t bring the rat inside. I\'m shutting the door, you don\'t want me to leave it open do you??  Here comes my mom down the stairs.  What\'s the problem? she asked. Wayne won\'t let me keep Jezebel inside. Wayne, why not? Rats don\'t belong inside, and there\'s no room in Michael\'s room. Oh, but I made room, I said. No, there\'s no room, you couldn\'t have. Well, I did, me and Mike.  He doesn\'t care, he said it\'s fine. Put the rat in the kitchen, he said. You want the rat in the kitchen?  Isn\'t that worse than keeping a rat in the bedroom?  That doesn\'t even make sense to me, I said. Rats are nocturnal and it will keep Michael up all night, you can\'t keep it in the bedroom. She\'s not that loud, if there\'s a problem, we can deal with it later, it\'s 2am. Wayne, just let her put the rat in the bedroom, we\'ll deal with this later.  So Jezebel got to stay in the bedroom, with me.  Yay.  I must have the most difficult, ridiculous parents ever.  Or at least it seems that way, all too often.  I got in bed, looked over my CBEST book, and set my alarm for 7am.  I actually got another 3 hours of sleep that night, I think.  And I still managed to do great on my test the next day.  Go me.  And so ends the day of arrival.  '),(17513,1405024,'25,July,2003','2003-07-25 00:00:00','   Well now. It\'s been awhile since I\'ve written anything, but, here I am in full force. I\'m really fucked up right now - is that a surprise? It shouldn\'t be.   I\'ve grown up a lot in the past month or so. As I mentioned in my last post, I packed up all my shit (well, most of it...what I didn\'t pack up and bring with me I put into a storage locker I rented) and moved down to LA. Where do I even begin with the LA experience? There\'s so much to talk about. Friday, June 13th (bad luck already - I should have realized the bad omen), I got my rental car, packed the shit I was taking with me into the car, and set off, leaving my Berkeley days behind me. I was terribly upset - I couldn\'t stop crying all day. I wouldn\'t have left that day except I had to take the CBEST the following day down in the valley. On my way out of town, I stopped for one last double vanilla iced latte at Fertile Grounds, made by the owner, Ayman. I knew I\'d really miss him - we\'d always had a fun, flirty rapport. Plus, I have the biggest crush on him. On the way there I almost got into an accident - yes, it\'s only a block and a half away from my old apartment complex, but hey now, I was dazed. Everything was fine though. Then as I parked, I witnessed an accident in the intersection directly in front of me. Freaky, I thought. But, in any case, I got my latte, and walked back to the car while wiping away tears. I really didn\'t want to leave.   When I got into the car, I proceeded to immediately spill a portion of my latte all over myself, my purse, and the rental car seats. Great. I\'d have to go back into the cafe and clean myself off and get some napkins for the car seat. So I walked back in, which I\'m sure was weird to Ayman. I told him what happened, and he gave me some paper towels. As I was trying to wipe off my purse, my glass chillum fell out onto the cement floor, shattering into a million pieces for all to see. I was so fucking embarassed. At first Ayman didn\'t notice, but when he did, he came out with a broom and asked if it was my pipe that broke. I told him yes, and he replied that it was a good thing. That\'s when I left for real, and started crying, although I can\'t say which happened first.  '),(17514,1405024,'16,August,2003','2003-08-16 00:00:00','   urlLink what if drugs were legal?  '),(17515,1405024,'01,August,2003','2003-08-01 00:00:00','   am I a lesbian?   apparently I\'ll consider anything that moves...   urlLink are you a lesbian?  '),(17516,1405024,'01,August,2003','2003-08-01 00:00:00','   Guess what!  I am a low threat to the Bush administration.  Heh heh, that\'s what they think.  No, actually, they\'re probably right...    Threat rating: extremely low. You may think you can subvert the government, but if you should try you will be smited mightily because God likes us best.    urlLink   What threat to the Bush administration are you?     brought to you by  urlLink Quizilla  '),(17517,1405024,'15,December,2003','2003-12-15 00:00:00','   Jezebel passed away on Friday, December 12. May she rest in peace, my little furry best friend. I will miss you always, and I love you dearly. You shall never be replaced.  '),(17518,1405024,'26,February,2004','2004-02-26 00:00:00','   Why do I feel so fucking tired?  I think I\'m getting sick or something.  Boy, I have so much to say.  Pass that dutch.  But will I really say anything at all?  My left hand is hurting for some reason - not really my hand, but more like the bottom of my ring and middle fingers.  I don\'t know what the fuck is going on, but it sucks.  I wonder if I will ever be censored for anything I write in here, or have my shit removed.  You know, for like, talking about drugs and whatnot.  Whoa!  Drugs are cool.  I seriously wish more than anything that I would fucking get up and clean my room.  Why the lack of motivation?  Could it be the pot?  No.  Really I think it isn\'t.  Some kind of psychological block or something preventing me from cleaning.  My room is a complete disaster though, shouldn\'t that be enough motivation?  One would think so.  Not me, though.  I\'m hella lit.  '),(17519,1405024,'25,February,2004','2004-02-25 00:00:00','   My neck hurts so much.  :(  If I was rich I would go see my chiropractor 1-2 times a week like I used to and also go see a masseuse.  I\'m so tense and tight for some reason.  It sucks ass dude, being in pain all the time.  And I\'m only 23.  Although lately everytime I look in the mirror I feel like I\'m about 34.  I\'m getting high.  Let me take another hit right now and I\'ll report the results.  '),(17520,1405024,'20,February,2004','2004-02-20 00:00:00','   More thunder and lightning!  Yay!  This is so cool.  And I still haven\'t even gotten high yet.  I should post after I take each hit.  I know I\'m posting a lot today/right now.  Maybe I\'m making up for lost time.  Maybe I\'m just in a crazy mood.  Maybe both.  Maybe I want to get high.  Well, duh, of course I do.        What am I listening to?  The Toadies.  Remember the Toadies?  Hehe.  Do you wanna be my angel?  '),(17521,1405024,'20,February,2004','2004-02-20 00:00:00','   Ok I totally thought an earthquake was about to start.  Turns out it was just really loud thunder.  How awesome!  We never get thunderstorms.  I remember when I was a little girl and I used to stay with my grandparents out on Long Island for the summer, I loved it when there were huge thunderstorms.  My Nana would always come wake me up if it happened to be the middle of the night, and the power would go out and we would watch the sky flicker and listen to the house shake.  It was so much fun.  I hope my power doesn\'t go out now, I just thought about that.  What a motherfucking crazy storm this is right now.  My windows are hella rattling.  This is so awesome.  My weather report looks like this:  showers, showers, showers, showers, showers, showers...times like 37 or some shit like that.  And I totally have to open the store tomorrow at 9, which means leave house at like 7:10am...ugh.  Plus I left my umbrella at work.  Maybe I will just call a cab tomorrow igga.  This morning walking to the bus stop in the pouring rain was fucking hell dude.  It was like walking in a river.  But it wasn\'t so bad really, I just got super stoned and like, separated my mind from my body during that time.  I must have, it\'s the only way I could have dealt with that shit.  Being rained on and all that.  Praise the almighty jesus for my marijuana.  '),(17522,1405024,'20,February,2004','2004-02-20 00:00:00','   I am so fucking cold dawg. I have so much to write about, if i so desired...I suppose I would need to use my discretion to decide what exactly I want to write, what I want \"out there.\" It\'s like once I formulate thoughts and feelings into actual words - and write them down somewhere, no less - it just makes them that much more concrete. If I wanted to say that I was wrong or mistaken and I never thought or felt a certain way, now I can\'t. Putting shit out there makes it harder to deny later. And plus there\'s less mystery. Does that even make sense? I sound like a crackhead, and I haven\'t even hit my pipe yet. No, not my crack pipe. I don\'t even have one of those. I mean my mari-ja-wana pipe. I haven\'t hit my pipe since earlier,&nbsp;with Krislyn. What a crazy cat that chick is.  '),(17523,1405024,'20,February,2004','2004-02-20 00:00:00','   the boys are waiting  '),(17524,1405024,'20,February,2004','2004-02-20 00:00:00','   watch if you\'re smart  '),(17525,1405024,'20,February,2004','2004-02-20 00:00:00','   HEY HEY HEY I\'m what\'s happNUN.  Oh yeah, you know it.  I just got home a bit ago, and I\'m eating taco bell (I\'m a health goddess) and packin a bowl.  Pass that dutch.  Pass that dutch.  I seem to have suddenly gotten WAY into Missy Elliott in the past few days.  How did this come about?  I\'m not sure really.  She is just so much woman, it\'s cool.  And she doesn\'t seem to give a fuck about what people think.  That\'s how I want to be.  A strong, independent woman who\'s cool and sexy and says what she thinks.  HEY HEY HEY.  I am going to smoke a little and write some more in a few.  '),(17526,1405024,'20,February,2004','2004-02-20 00:00:00','   I\'m a sexy bitch!  '),(17527,1405024,'15,February,2004','2004-02-15 00:00:00','   I feel like my life is a chaotic mess.  This feeling is represented well by my living quarters - random shit everywhere, not put away, making it impossible for any forward movement or progress.  Similarly, random issues and thoughts and worries are cluttering my mind.  I guess I\'ll just take another hit from my pipe and try to forget, once again.  I think I\'m going to go get some food, and when I come back I\'ll sit here and write.  I promise.  Btw, I think the whole world is nuts right now...like there is some kind of odd alignment in the stars making the beginning of this year of the monkey super freaky.  Yow.  '),(17528,1405024,'15,February,2004','2004-02-15 00:00:00','   I feel like starting anew.  '),(17529,1405024,'31,March,2004','2004-03-31 00:00:00','   Hee hee!  '),(17530,1405024,'27,March,2004','2004-03-27 00:00:00','   So, I did laundry tonight.  Finally.  I love having clean clothes!  I spent like, $16 getting most of my stuff done.  Not even all of it.  Sigh.  Some day it will all be clean!  I\'m putting my shit away right now, so I\'ll come back and make random statements every once in awhile. :)  Until then, play  urlLink this game ! I think it\'s in Dutch...just click on \'ja\' and then \'start.\'  Then try to pour a perfect glass of beer!  How entertaining.  I\'m surprisingly good at it.  When you fuck up and want to try again just press \'nog een keer.\'  '),(17531,1405024,'27,March,2004','2004-03-27 00:00:00','   Wow, it\'s been awhile again, huh.  I just woke up a little bit ago, and I\'m already pretty stoned.  It\'s going to be a GREAT FUCKING DAY.  I\'m going to see the Breeders tonight, at Slim\'s.  How fucking awesome is that shit??  I\'m almost shitting on myself!  '),(17532,1405024,'16,March,2004','2004-03-16 00:00:00','   Seriously I wish I wasn\'t so tired, I am in a cleaning mood.  But I kind of just want to go to bed.  I wish I had some yay.  '),(17533,1405024,'06,March,2004','2004-03-06 00:00:00','   Wow dude.  It\'s been awhile, huh.  As usual.  I\'m incredibly tired for some reason.  Or maybe I\'m mistaken, and it\'s just me being high from the trainwreck I just smoked.  So I\'ve started collecting homies lately.   urlLink Check it out.   These are the ones I have so far:  urlLink Babylocs   urlLink Big Dopey   urlLink Brandy   urlLink Bruja   urlLink Chepe   urlLink Freckles   urlLink Mad Bomber   urlLink Papi Chulo   urlLink Puppet   urlLink Shaneequa   urlLink Spider   urlLink Syco   A few I duplicates of.  And obviously I like Brandy the best. :)  I can\'t wait to get more!  '),(17534,1405024,'06,March,2004','2004-03-06 00:00:00','   I\'m no longer unstoned!  Yay!  '),(17535,1405024,'05,March,2004','2004-03-05 00:00:00','   It\'s early and I\'m unstoned!  I actually woke up at 6:45 this morning...for what?  Oh, nothing.  I guess I just got the right amount of sleep or something.  But it\'s oh so hard to wake up when I actually need to, like, for work or something.  What\'s up with that yo.  '),(17536,1405024,'05,March,2004','2004-03-05 00:00:00','   please buy me things.  I will love you forever and ever!  '),(17537,1405024,'05,March,2004','2004-03-05 00:00:00','   Isn\'t it weird how I don\'t post every day, but when I do it\'s in spurts as my thoughts come to me?  Whoa there, noisy kids upstairs.  Don\'t be jumping through the ceiling now.  And I am so sick and tired of my left shift key sticking.  Don\'t get me wrong, it\'s not sticky, but when I press it from a certain angle there seems to be too much friction.  Whoa!  I said friction.  '),(17538,1405024,'05,March,2004','2004-03-05 00:00:00','   some people are really creepy and just don\'t get it  '),(17539,1405024,'05,March,2004','2004-03-05 00:00:00','   Have you heard that song by  urlLink Peaches ?  Eat a cookie and a big dick every day!  What?  What?  What?  '),(17540,1405024,'05,March,2004','2004-03-05 00:00:00','   My fucking feet hurt like a motherfucker.  And I have little itchy bug bites.  Like spider bites or some shit like that.  What the fuck dude.  Hehe, enough complaining.  Well, maybe.  But I\'m done for the moment.  Is it unpatriotic for one to say she may wish the worst for \"the ailing attorney general\"?  No, I guess not...but don\'t fucking expect me to  urlLink pray for him .  I bet Bush is already coming after me just for those two sentences right there - I\'m obviously a terrorist.  It would appear I\'m anti-American and *gasp* anti-Christian too!  Oh dear.  '),(17541,1405024,'17,April,2004','2004-04-17 00:00:00','   Man.  I totally just woke up, at like, 5:30.  Yes, pm.  And my alarm was set for 11:30am.  And yes, I was indeed hitting snooze every nine minutes between 11:30am and 5:30pm.  Can you believe that?  I can, it\'s not really all that out of the ordinary.  But I really didn\'t want to wake up so late, I kind of wanted to get up at a reasonable hour and be productive.  You know, clean my room, run some errands, shit like that.  I really want to see Kill Bill Vol. 2.  I was supposed to go yesterday with some friends, but it was sold out everywhere by the time we tried to buy tickets at like, noon.  That sucked.  So I\'m thinking maybe tonight I\'ll go to Tim and Patrick\'s and watch Volume 1.  I\'d still rather go see Vol. 2, but it will be good to refresh my memory I guess, right?  Okay, shower time.  I wish I had a lot of drugs.  '),(17542,1405024,'13,April,2004','2004-04-13 00:00:00','   Guess what!  Tim came over tonight to chill with me for a bit and smoke some bowls and he brought me a homie, which was really nice of him since he\'s collecting them too.  Usually we go and get them together, but I guess today he got a bunch from the machine on Shattuck by the In&Out Deli.  He let me pick one, which was awesome.  I chose the Eminem homie, aka  urlLink Angry White Boy .  He\'s so cute, he looks just like Eminem.    It tastes like burning!  '),(17543,1405024,'13,April,2004','2004-04-13 00:00:00','   Weed is my antidrug.  '),(17544,1405024,'13,April,2004','2004-04-13 00:00:00','   I\'m eating lasagna with a spoon.  I\'m not quite sure I\'ve actually done this before, but I don\'t have any more clean forks and I\'m definitely not about to wash one.  No way dawg.  I\'m listening to the streaming video from msnbc.com - all the shit going on with 9/11 and Ashcroft and the FBI and Iraq and hostages and oh who else can we blame and drag into this and it will never end, never ever.  Congratulations, you\'re watching the beginning of WWIII.  I just finished my lasagna.  I\'m pretty high.  I\'ve been smoking pretty steadily for the past few hours.  I\'ve been cleaning, but when I clean it always looks like nothing really happened.  I concentrate too much on detail, not the big picture.  I get lost in the small things.  Even from way up here.  Now I\'m craving a can of Dr. Pepper, which I happen to have in my fridge.  Rock on!  Once again, I\'m going to say I intend to update multiple times daily, whenever I feel like expressing my random thoughts.  Hey, I\'m going to link to Kim\'s blog in hopes that it will encourage her to update it.  '),(17545,1405024,'13,April,2004','2004-04-13 00:00:00','   Sup.  I totally got some new homies today, and last night, and like last week sometime.  I\'m excited.  Here are my newest ones:  urlLink Devil Dog   urlLink Diablo   urlLink Dreamer   urlLink El Paletero   urlLink Home Lee   urlLink Ice Cream   urlLink Lizard   urlLink Mr. Clean   urlLink Pinoy   urlLink Tennishoe Pimp   Just for reference, these are the ones I already had:  urlLink Babylocs   urlLink Big Dopey   urlLink Brandy   urlLink Bruja   urlLink Chepe   urlLink Freckles   urlLink Mad Bomber   urlLink Papi Chulo   urlLink Puppet   urlLink Shaneequa   urlLink Spider   urlLink Syco   and then I also have the Homie Clown  urlLink 2Face .  Obsess much?  Not I.  '),(17546,1405024,'13,April,2004','2004-04-13 00:00:00','   Hmm...what are my plans for the evening?  I really, really should clean my room.  Right now I\'m drinking my usual of the moment - an iced venti hazelnut lowfat latte.  Mmmm.  I\'m also smoking my usual - a bowl of weed!  Imagine that.  '),(17547,1405024,'31,May,2004','2004-05-31 00:00:00','   I just took a drug knowledge purity test.  This is what it said: You answered \"yes\" to 76 of 100 questions, with a total value of 84 points out of 116, making you 27.6% drug pure (72.4% drug corrupt).  I lost points for never going to jail and never getting caught.  Oh, and for not selling my body for drugs.   urlLink Take it yourself!  '),(17548,1405024,'30,May,2004','2004-05-30 00:00:00','   find your inner PIE @ stvlive.com     Come here and eat me.  '),(17549,1405024,'30,May,2004','2004-05-30 00:00:00','   Discover your Zodiac Personality @ Quiz Me     I do drugs to escape from today.  And I\'m a cancer goddammit!  '),(17550,1405024,'30,May,2004','2004-05-30 00:00:00','   Are you Addicted to the Internet?           62%                Hardcore Junkie   (61% - 80%)  While you do get a bit of sleep every night and sometimes leave the house, you spend as much time as you can online. You usually have a browser, chat clients, server consoles, and your email on auto check open at all times. Phone? What\'s that? You plan your social events by contacting your friends online. Just be careful you don\'t get a repetitive wrist injury...        The Are you Addicted to the Internet? Quiz at Quiz Me!  '),(17551,1405024,'30,May,2004','2004-05-30 00:00:00','   So I took another quiz.  Here are my results:      Take the 100 Acre Personality Quiz!    So I\'m Winnie the Pooh, eh?  A little slow?  Hey fuck you.  And you forgot that I need good drugs to be happy too!  '),(17552,1405024,'30,May,2004','2004-05-30 00:00:00','   So...hi.  There\'s so much on my mind right now, thoughts flying everywhere.  And what am I doing?  Trying to distract myself from thinking by cleaning and doing drugs.  You know, the usual.  My head just started weirdly hurting.  I think I\'m really fucked up or something.  I\'m probably dehydrated as well - maybe I should drink some water.  Man, I\'m really tired for some reason.  Maybe I\'m just fucked up.  I did work all day today...well, not all day, but like, a regular workday.  I got up at around 7:40am (wow, I can\'t believe it) and was in town on time to catch the 8:40 bus! (That deserves some kind of recognition at least.)  That\'s not as bad as having to get on the bus at 7:26, but still, I was happy.  Work went alright I suppose.  I worked with someone who usually works at another one of our stores.  I was really high when I got there, and then I got even higher when the new girl from next door came to visit me and we smoked a couple more bowls right after I got done smoking a couple bowls.  It was pretty awesome.  I talked to her for a long time - I wish I had more time to talk to her today.  Next time, I hope.  I talked to my mom today.  I\'m flying down to LA in a couple of weeks to see my brother graduate from middle school.  While I\'m super excited to go down there and everything, this forces me to realize how fucking old I am.  What the hell am I doing with my life?  My brother is ten years younger than me, graduating junior high, and I\'m doing what?  I was supposed to be doing something by now.  Something.  Fuck.  I keep thinking about A.  So much to say on this subject...I think I\'ll go wash my face.  And then clean or something.  And then maybe, just maybe, I\'ll force myself to actually process thoughts on this subject.  But, we\'ll just have to see about that, now won\'t we.  '),(17553,1405024,'30,May,2004','2004-05-30 00:00:00','   Every time I go to write something, I get distracted in some way or another.  That only happened about 50 times tonight...and now I\'m tired and incredibly high.  I kind of just want to lay down.  Yeah yeah, I know, I said I\'d write.  Empty promises.  Forlåt!  Goodnight moon.  '),(17554,1405024,'29,May,2004','2004-05-29 00:00:00','   I have been working all day, I\'m kinda tired, and I want to get stoned.  All day yesterday I was worrying about running out of weed before I got more, and it seems like today it\'s been falling all over me.  I love when that happens.  I want to do something at least semi-productive for a little while (such as clean or something), and then I have tons to write about.  Toodles!  '),(17555,1405024,'29,May,2004','2004-05-29 00:00:00','   Wow, it\'s been so long since I wrote an actual entry.  I guess I have just been occupied with other things the past few days, so I really haven\'t had much of a chance to write.  My friend is here right now, she\'s going to spend the night.  We met in Berkeley earlier tonight and fucked around, and then we came back to my place and got high.  Then we went for a walk.  It was pretty good exercise - it\'s kinda hilly around my house.  Then we did our toenails and I cooked dinner and we watched  urlLink The Rules of Attraction .  Man I love that movie.  Now we\'re watching vh1 and she\'s lying on the floor wrapped up in a blanket.  I\'m probably keeping her awake actually, and I should be asleep already anyways.  I\'m going to set my alarm for 7am, since I have to work tomorrow.  Suck.  My nose really hurts goddammit.  I want some ice cream.  I never finished my limo driver story.  I\'ll do that tomorrow I suppose.  I\'m going to bed, maybe after a bite of ice cream...  '),(17556,1405024,'29,May,2004','2004-05-29 00:00:00','   urlLink Internet Boy \'Posed As Spy to Arrange Own Murder\'   those naughty internet boys...  '),(17557,1405024,'27,May,2004','2004-05-27 00:00:00','   urlLink Ananova - New rules on pizza to be introduced  '),(17558,1405024,'27,May,2004','2004-05-27 00:00:00','   It\'s about quarter to 1, and I\'m pretty tired.  Like, physically I feel tired, but I totally don\'t want to go to bed yet.  When I got home from work I cleaned my bathroom extremely throroughly, cleaned two of my pipes, and then washed some dishes and cooked myself some dinner.  Unfortunately I didn\'t get home from work until about 7:45pm (which actually isn\'t that bad, I just hate working), so now it\'s already pretty late.  Thank god I have tomorrow off...  Right now I feel too braindead to write.  Maybe I\'ll write more in a bit?  '),(17559,1405024,'27,May,2004','2004-05-27 00:00:00','   Have I mentioned lately that my life is crazy?  I\'m so mad right now that it\'s so fucking late...where the fuck did the time go tonight?  If it wasn\'t so late I\'d probably write a lot more than I\'m about to, so that will have to come tomorrow.  As I was waiting for the bus tonight after work, this limo pulls up and the driver appeared to be looking for something.  I was just sitting on the bench waiting for the 7:26 bus to come with this limo in front of me, fantasizing about not having to take the fucking bus and how cool it would be if someone rich and powerful was in that limo right there and took pity on me and wanted to drive me home in the limo.  I\'m looking down at my sweatshirt and the remnants of the bean burrito from Taco Bell I spilled on it last night when I hear the driver say \"Are you waiting for your limo?\"  Yeah motherfucker, I always wait for my limos sitting at the bus stop.  I looked up and replied with something like \"yeah right, I wish...I\'m waiting for the bus.\"  He asked where I lived...Montclair, I told him.  He asked if I wanted a ride home in his limo, he had an hour to kill or something.  \"How much will it cost me?\" I asked him.  It\'s free, he told me.  So I got in the back, and my adventure began...  '),(17560,1405024,'25,May,2004','2004-05-25 00:00:00','   urlLink BBC NEWS | Science/Nature | Venus clouds \'might harbour life\'   I\'m fucking tired dude.  I just feel like passing out.  '),(17561,1405024,'24,May,2004','2004-05-24 00:00:00','   Busty Wilde was   a Charming Lunatic   in a past life.   Discover your past lives @ Quiz Me  '),(17562,1405024,'24,May,2004','2004-05-24 00:00:00','   Busty Wilde spins tunes as   DJ Divine Lunatic     Get your dj name @ Quiz Me  '),(17563,1405024,'24,May,2004','2004-05-24 00:00:00','   The \'My Band\' video again???  Oh my fucking god.  I\'m still not sick of it, I downloaded the fucking video from Kazaa for god\'s sake...but come on.  Three times in about 2.5 hours?  There have got to be other videos out there.  What about that new Beastie Boys video, they haven\'t even shown that one once (but I wish they would).  Or that annoying-ass Alisha Keys video, thank god.  '),(17564,1405024,'24,May,2004','2004-05-24 00:00:00','   This one has a disclaimer at the top saying it\'s wrong.  Yeah, yeah, it\'s wrong...and so are my results...  You\'re addicted to.....  Cocaine!  I pity you, Cocaine isnt a good thing to be addicted to. One of the more expensive drugs out there and not to mention those damn nosebleeds. Get help now!   urlLink   What are you addicted to?    Hey, I have to be up at 7am, why am I still awake?  My coworker is picking me up at 8:15, bless her heart, so I don\'t have to get up at 6 instead and catch the bus at 7:26...  '),(17565,1405024,'24,May,2004','2004-05-24 00:00:00','   So I\'m curious...usually when I view a blog on blogspot I get these ads that are like personalized to the site I\'m viewing (just like the personalized ads you get when you do a google search, or if you have google mail...).  What ad do you see at the top of mine?  I just looked at my page and saw (there\'s two in the box at the top) one for an herbal vaporizer (hahaha) and another one for drug tests.  How funny is it that the main \"idea\" of my page is coming up as drugs and getting high.  How awesome.  '),(17566,1405024,'24,May,2004','2004-05-24 00:00:00','   What\'s your sexual appeal?  It\'s really funny.  I especially like the question \"What was your favorite subject in school?\"  I was all ready to click math (might I be a nerd?  of course not!) when I saw that the last option is \"I\'m a cunning linguist.\"  I totally say this in real life (yeah, I have a degree in linguistics...almost), so I had to pick that one.  And I love the fact that YOUR MOM is an answer to more than one question.  So here are my results:    Pervert   urlLink   Take it yourself!  '),(17567,1405024,'24,May,2004','2004-05-24 00:00:00','   Okay, now they\'re playing that same Outkast video  again .  Wtf?  '),(17568,1405024,'24,May,2004','2004-05-24 00:00:00','   but I did.  Now I\'m overly full and uncomfortable. :(  Maybe I\'ll smoke some to ease the pain...  '),(17569,1405024,'24,May,2004','2004-05-24 00:00:00','   How else would I see the same video twice within a half hour?  It\'s that new D12/Eminem video for \'My Band.\'  I\'m sure a lot of people find it annoying but I love it...especially the salsa part at the end.  I just love the serious way Eminem stares into the camera while holding that jar of salsa with the dog running by and singing those ridiculous lyrics.  I want to be on whatever drugs he\'s on in that scene.  Plus, I love those hot pink sock-tight things he\'s wearing.  Gotta love him.  '),(17570,1405024,'24,May,2004','2004-05-24 00:00:00','   urlLink Which slang word r u?    You\'re the slang word nigga. everyone and everything is one. please use carefully!  '),(17571,1405024,'24,May,2004','2004-05-24 00:00:00','   urlLink   Which old school Nickelodeon show are you?    Guess what I am!    You are SALUTE YOUR SHORTS.  You are a wangsta who knows business.  You have fun in the sun and know how to have a good laugh.  '),(17572,1405024,'24,May,2004','2004-05-24 00:00:00','   So I just cooked myself another meal.  This cooking thing is actually pretty fun - I actually wanted to cook tonight instead of microwaving another frozen meal.  Mushrooms, zucchini, bean sprouts, and onions over rice.    Tim came over tonight for a bit, which was cool.  He and Patrick havn\'t been able to get any weed the past few days, so I\'ve been having him come over to smoke with me (and to visit...I mean, we\'d be smoking either way so it\'s all good).  We watched the first hour and twenty minutes of The Swan pageant (gotta love Fox dude), and then we decided to go to Safeway real quick and come back in time to watch the end of the show.  I thought I was only going to pick up a few things, but, since we smoked so much before we left, I totally kept getting distracted and wound up taking way too much time.  Now I have way more than enough food for the next little while, so that\'s good.  I kind of kept getting anxious because Tim was waiting for me and I was taking so long (one of my \"things\" is having people waiting for me - for some reason that totallly makes me start to panic), but I kept trying to push that out of my head. [pause right here: I\'d like to interrupt to say they\'re playing the Outkast \'Roses\' video  yet again .  Sigh.] We didn\'t get to watch the end of the Swan, which I guess doesn\'t really matter.  I was kind of only watching it out of some sadistic pleasure I was getting.  It made me feel really gross inside though...  I have therapy tomorrow.  I\'m definitely going to talk about what happened Thursday night, maybe I can get some help in dealing with it.  Honestly I don\'t know what to make of it, but I seem to be best when I don\'t let myself think about it.  I know that\'s not healthy at all though.  We\'ll see.  I feel full, yet I totally want to eat more.  Story of my life, eh?  I think I just like to eat and have tasty things in my mouth.  '),(17573,1405024,'24,May,2004','2004-05-24 00:00:00','   Prepare for more, since I actually just searched for some.  In fact,  urlLink here\'s one right now ! (I linked to the one for women...if you\'re a guy, click on the one for men)  You are a Queen!  Beautiful, Wise, Strong Righteous, Commanding, Humble   You are the beautiful and compassionate Queen. You are the epitomy of what every woman should be. You are confident, bold, aggressive, smart, womanly and feminine. You know the right thing to do and do it. You command respect and earn praise. You are moral and loving. In times of trouble, you draw strength from within, and are a source of strength for others.  Wow, I\'m a queen.  Surprise surprise.  '),(17574,1405024,'23,May,2004','2004-05-23 00:00:00','   urlLink Ananova - Man sues guru over raw frog cure  An alternative medicine guru is being sued by a man who almost died after he was told to eat raw frog six times a day to cure a pain in his neck.  The man, identified only as Chen, went to the alternative medicine expert in China\'s Hunan province because of severe neck pains.  He was told to eat at least six raw frogs a day to get rid of the pain and had scoffed his way through 130 before he collapsed, complaining of stomach pains and headaches.  He is now suing after doctors discovered his body was riddled with parasites that had come from the frogs, the Xinhua agency reported.  ...maybe he thought he\'d shit a prince?  '),(17575,1405024,'23,May,2004','2004-05-23 00:00:00','   I fucking love this type of stupid ass shit.  The government makes up this propaganda to feed the American public like we\'re a bunch of idiots.  Sad thing is, most of the American public is...   urlLink Street Terms: Drugs and the Drug Trade - ONDCP   Let\'s go get us a sandwich bag of rompums.  '),(17576,1405024,'23,May,2004','2004-05-23 00:00:00','   And I haven\'t done that word association either.  Boy am I glad that I can keep myself entertained for hours late into the night with just some drugs and an internet connection.  Guess you could say I\'m easy.   urlLink Here we go!     Sexy:: red hot  Clique:: Mean Girls  Pledge:: lemony fresh  Carbs:: Atkins  Dream Job:: music bigwig  Sweeps:: supermarket  Soundtrack:: Boogie Nights  Hero:: He-Man (wtf?)  Shave:: Terax  Christina:: Aguilera    And here\'s  urlLink another .  (Dude I am definitely stoned, these word associations are way too fun right now.)    Elastic:: rubber  Intervention:: alcohol  Risk:: chance  Junk food:: chips  Arrogance:: fuckers  Responsibility:: run!  run as fast as you can!  X:: xxx  Marshall:: Mathers  Kill:: Bill  Brother:: big  '),(17577,1405024,'23,May,2004','2004-05-23 00:00:00','   The Dante\'s Inferno Test has banished you to  the Seventh Level of Hell!   Here is how you matched up against all the levels:     Level    Score       urlLink Purgatory   (Repenting Believers)   Very Low       urlLink Level 1 - Limbo   (Virtuous Non-Believers)   Very Low       urlLink Level 2   (Lustful)   Very High       urlLink Level 3   (Gluttonous)   High       urlLink Level 4   (Prodigal and Avaricious)   High       urlLink Level 5   (Wrathful and Gloomy)   Very High       urlLink Level 6 - The City of Dis   (Heretics)   Moderate       urlLink Level 7   (Violent)   Extreme       urlLink Level 8- the Malebolge   (Fraudulent, Malicious, Panderers)   High       urlLink Level 9 - Cocytus   (Treacherous)   Very High      Take the  urlLink Dante\'s Inferno Test    Wow, I had 3 \"very high\" ratings and 3 \"high\" ratings (after all, I am always high or very high...sorry, bad joke).  But I guess I belong in level 7, where I can be extremely high.  '),(17578,1405024,'23,May,2004','2004-05-23 00:00:00','   So I took this survey on that same site as the personality disorder quiz that asked if I recognized a series of 109 names.   urlLink Here are the results  of all the test takers.  Does anyone else find it amusing that Marilyn Manson is more recognizable than both Bill Gates and Jesus?  What the fuck.  I think I\'ll do another word association.  '),(17579,1405024,'23,May,2004','2004-05-23 00:00:00','   I just took this personality disorder test.  Here are my results:       Disorder      Rating        urlLink Paranoid :    Low       urlLink Schizoid :    Low       urlLink Schizotypal :    Moderate       urlLink Antisocial :    Low       urlLink Borderline :    Low       urlLink Histrionic :    Moderate       urlLink Narcissistic :    Moderate       urlLink Avoidant :    High       urlLink Dependent :    High       urlLink Obsessive-Compulsive :    High       --  urlLink Personality Disorder Test - Take It!  --       Look how bright and red that last one is...obsessive compulsive.  Hell, I could have told you that.  Actually I already did.  Avoidant and dependent though, that\'s interesting.  Obsessive-Compulsive Obsessive-Compulsive personality disorder is similar to obsessive-compulsive anxiety disorder. People with this disorder are overly focused on orderliness and perfection. Their need to do everything \"right\" often interferes with their productivity. They tend to get caught up in the details and miss the bigger picture. They set unreasonably high standards for themselves and others, and tend to be very critical of others when they do not live up to these high standards. They avoid working in teams, believing others to be too careless or incompetent. They avoid making decisions because they fear making mistakes and are rarely generous with their time or money. They often have difficulty expressing emotion.  Avoidant Avoidant personality disorder is characterized by extreme social anxiety. People with this disorder often feel inadequate, avoid social situations, and seek out jobs with little contact with others. They are fearful of being rejected and worry about embarassing themselves in front of others. They exaggerate the potential difficulties of new situations to rationalize avoiding them. Often, they will create fantasy worlds to substitute for the real one. Unlike schizoid personality disorder, avoidant people yearn for social relations yet feel they are unable to obtain them. They are frequently depressed and have low self-confidence.  Dependent Dependent personality disorder is characterized by a need to be taken care of. People with this disorder tend to cling to people and fear losing them. They may become suicidal when a break-up is imminent. They tend to let others make important decisions for them and often jump from relationship to relationship. They often remain in abusive relationships. They are overly sensitive to disapproval. They often feel helpless and depressed.  Okay, honestly, I agree with the obsessive-compulsive disorder part.  I might be moderatlely avoidant and dependent, but nowhere near the extremes described here.  Another interesting thing I realized while taking this test is how much I have changed over that last few months.  A lot of the questions where I would have given one answer a few months ago totally don\'t fit me anymore.  Like the ones asking if I\'m scared of being alone.  Was I at some point?  Yes, definitely.  Am I now?  Definitely not.  Shit like that.  I guess that\'s cool.  It\'s like I\'ve become more sure of myself and my ability to take care of myself.  '),(17580,1405024,'23,May,2004','2004-05-23 00:00:00','   I\'m not feeling quite as lethargic as I was a little bit ago, but I am still pretty stoned and kinda tired.  I am almost (read: almost) motivated to clean up in my room.  But, not quite.  I somehow ate half of that pint of rock and roll ice cream, and even though I\'m way full I still feel hungry.  I just want to munch I think.  I have a tiny amount of oxy left (definitely enough to get me fucked up)...I\'m wondering if I should do it.  I do have tomorrow off.  Sounds tasty if you ask me.  Let\'s get out the ol\' mirror and razor.  Too bad I don\'t have more, huh.  Do you ever read  urlLink Craigslist ?  Sometimes there\'s really funny shit if you look in the cracks and crevices.  I can\'t say I\'ve looked anytime recently, but for some reason I just thought of it.  Actually I know the exact reason...once I knew this stupid girl who bought drugs from a posting on craigslist.  What a dumb ho.  When I thought \"I wish I had more oxy,\" I reminded myself of her buying morphine.  People are fucking crazy man.  '),(17581,1405024,'23,May,2004','2004-05-23 00:00:00','   ...but I want to do  urlLink another one !    Vagina:: hahahaha, I\'m sorry, but come on, this is supposed to be a random word association?  Okay.  Um, cunt.  Racism:: vagina  Mother\'s Day:: vagina  Fire alarm:: sirens  Elvis:: pelvis  Pregnant:: vagina  Vacation:: vagina (they start with the same letter)  Waffles:: leggo my vagina  Perpendicular:: to my vagina  Hospital:: vulva  '),(17582,1405024,'23,May,2004','2004-05-23 00:00:00','   Maybe not so much bored, maybe I\'m restless?  I don\'t know.  I just broke out the Ben & Jerry\'s Cherry Garcia.  I only like this mediocre ice cream because I like rock and roll, and I\'m easily fooled by marketing techniques.  Okay actually that was a distant reference to  urlLink Mr. Show with Bob and David , the best comedy show EVER.  I decided to look back at that  urlLink word associations web site  to see if they had other lists from past weeks.  Lo and behold, there is an archives link.  I think I\'ll do another list.   urlLink Here\'s the link  if you want try for yourself.  Or you can just comment to this post and leave your own answers. ;)    Playoffs::   Morris:: the cat  Break up:: (insert name of my ex-boyfriend here)  Eggs:: chickens  Parker:: posey  Hardy Boys:: Nancy Drew, duh  Deluxe:: Yahtzee!  Protection:: condoms  Girl Scout:: cookies, mmmmm  Salsa:: verde  '),(17583,1405024,'22,May,2004','2004-05-22 00:00:00','   I just cooked.  Isn\'t that awesome?  It\'s like I\'m grown up or something.  Hah.  I sautéed some portabella mushrooms, green and red bell peppers, onions, and zucchini and made a little bit of pasta.  It was so good, I wish I had some more.  I\'m so glad I\'m able to cook again (in other words, I\'m so glad I finally cleaned my kitchen area to the point where I can cook).  I\'m suddenly paranoid that the \'Anonymous\' person who sometimes posts comments is my mom.  I wonder if I should be worried?  '),(17584,1405024,'22,May,2004','2004-05-22 00:00:00','   I wish I didn\'t have to work tomorrow.  The bus comes at 8:40am, so I should probably get up at about 7:30.  Who knows if I\'ll actually get up at that time, but it\'s not like I need sleep or anything after the 18 hours I slept last night.  Shit, I\'ve been awake for less than 10 hours and I already feel like I need to go back to bed.  I\'m working with someone I really like...and it\'s only 10-5.  And she doesn\'t mind if I\'m high.  At least, she doesn\'t act like she minds.  I hope I don\'t work Monday.  Have I mentioned how much I need a car and how much I hate the bus?  '),(17585,1405024,'20,May,2004','2004-05-20 00:00:00','   I found this website with a list of 10 words you can do word associations with.   urlLink Do it yourself.     Finale:: final  Martial arts:: karate  Flirt:: hot pink  Energy:: wild  Flavor:: sour  Guess?:: what  Accomplishment:: for some reason i really can\'t think of anything to write here, i just get this empty feeling inside.  how sad.  Prom:: dress  Diploma:: fuck  Bloody:: mary  '),(17586,1405024,'20,May,2004','2004-05-20 00:00:00','   urlLink Bush Takes Tumble During Bicycle Ride  (you might have to register to read anything on this website, so I\'ll paste the article at the end of my post)  I\'m sorry but this is extremely amusing to me.  Read the entire thing, it\'s pretty short.   Seems like Bush is a bit of a bumbling idiot, eh?  But hey, check out that part about John Kerry falling while snowboarding.  They should use that in ads to get young people to vote for him and get Bush out of the White House.    Bush Takes Tumble During Bicycle Ride   President Unhurt, Completes Course  By Dana Milbank Washington Post Staff Writer Sunday, May 23, 2004; Page A10    AUSTIN, May 22 -- President Bush, always a bit star-crossed in his leisure pursuits, suffered a new misadventure in athleticism Saturday when he fell off a bicycle.    The president was in mile 16 of a 17-mile course on his ranch in Crawford, Tex., when he fell from a mountain bike, suffering scrapes and scratches on his chin, upper lip, nose, right hand and both knees. But, true to form, Bush stayed the course: After getting cleaned up by the White House doctor, he completed his bike ride.   Trent Duffy, a White House spokesman, immediately offered excuses for his boss. \"It\'s been raining a lot, and the topsoil is loose,\" he said. In addition, Duffy reasoned, \"He likes to go all out. Suffice it to say he wasn\'t whistling show tunes.\" Fortunately, Duffy reported, Bush was wearing a helmet and mouth guard.   The spokesman warned that Bush probably would be wearing a bandage on his chin when he arrived here Saturday night for a party for his daughter Jenna, who is graduating from the University of Texas this weekend.   For the president, there is something of a history of hapless encounters with sporting activities.   Years of running gave him two bum knees that have periodically left him limping his way to Air Force One. In December, doctors at Walter Reed Army Medical Center examined him with X-rays and magnetic resonance imaging and, finding some damage to both knees, recommended \"cross training.\"   In other words, exercise such as mountain biking.   In January 2002, Bush was on the third floor of the White House residence, watching a football playoff game between the Baltimore Ravens and Miami Dolphins when he choked on a pretzel. This caused him to faint and fall, bruising and scraping his face. Bush was accompanied only by dogs Spot and Barney.   Then, in June 2003, Bush was visiting his parents at the family compound in Kennebunkport, Maine, when he fell from a motorized Segway scooter. Bush, who was holding a tennis racket that may have interfered with his coordination, did not hurt himself -- but the incident was captured by photographers and quickly beamed across the world.   Athletic miscues are a bipartisan affair, of course. Bush\'s Democratic challenger John F. Kerry has, during the current campaign, fallen off a bicycle and taken a nasty spill while snowboarding. He blamed the Secret Service for the latter.   Bush\'s close calls predate his ascension to the presidency.   During the campaign in 1999, he scraped his right leg and hip when a truck trailer overturned near his jogging path in Austin. Bush dived for cover when chunks of concrete and wood were dumped behind him.   Around the same time, Bush went swimming in the pond on his ranch with his advertising strategist when the two men spotted cottonmouth water moccasins -- the only type of poisonous water snake in North America. That time, at least, Bush was not snakebit.    © 2004 The Washington Post Company  '),(17587,1405024,'20,May,2004','2004-05-20 00:00:00','   So I met my friend in  urlLink the village  (it\'s a real village!) and we went to Safeway. Then we had to walk up the fucking hills to my house carrying our groceries. That always makes me think I\'m out of shape...but really I think maybe I smoke too much (and I\'m out of shape). Not fat, just unable to walk up hills while carrying about 10 pounds without breathing heavily. But hey, whoever said heavy breathing was bad?   I\'m stoned. We got high when we got home, which definitely eased the trauma from walking. I realized as soon as leaving Safeway that I have no oil or anything to fucking cook any of the vegetables I bought in. Damn. I guess I\'ll just get high and eat some frozen lasagna instead.   I got a friendly reminder in the mail today that it\'s time for my annual pap smear! Yay, I can\'t wait. I hope they don\'t drug test me. I mean, it\'s the same doctor that prescribes me cannabis (hah) so I don\'t think THC in my system would be a big deal, but maybe other things would...?  '),(17588,1405024,'20,May,2004','2004-05-20 00:00:00','   urlLink \"The president also revealed that U.S. forces were currently re-erecting a statue of Saddam Hussein to be re-toppled upon their July 1 return.\"  '),(17589,1405024,'20,May,2004','2004-05-20 00:00:00','   urlLink BOROWITZ report.com  '),(17590,1405024,'20,May,2004','2004-05-20 00:00:00','   Should I actually put pics of myself up here?  What do you think?  I don\'t want to incriminate myself, you know?  '),(17591,1405024,'20,May,2004','2004-05-20 00:00:00','   ...but at least I\'m stoned. Right?   So yesterday, this guy Patrick who used to work down the street from me until he got fired about a week ago came back to the store to visit me. &nbsp;He worked at Noah\'s, and I guess his boss accused him of taking $600. Did he do it? I don\'t know, he says he didn\'t but who knows. He just turned 18 a few weeks ago, and he has 420 tattooed on his left wrist, which I think is all too awesome (but what about when he \"grows up\" and wants to join the \"real world\"?). Last week, the day he got fired (maybe it was Tuesday or something, I honestly don\'t remember), he ended up getting super drunk and passing out in our backyard. (Our store shares a backyard with Ritz Cameras.) I had given the Ritz guys a disposable camera I took from their store to develop for me (for free, naturally) and Patrick took like every single picture that had me in it. Anyways, when he showed up yesterday randomly, he happened to open his wallet for something and the first picture in his little plastic picture thing that people stick in their wallets was me. ME. As in, one of the pictures of me he stole he took the time to cut to that size and put it in his fucking wallet. I totally have some weird effect on people...honestly this isn\'t even out of the ordinary for a day in my life. Like the day when the random guy on outside my work came up to me and was staring at me, so I yelled at him \"Um, do I KNOW you?\" \"No,\" he said, \"but now I have your picture!\" he said as he took my picture and ran away. He returned the picture of me to me about a month later, which his number on the back. However, it had been crumpled kind of like he was squeezing it in one hand as he jerked off with the other. See what I mean?  '),(17592,1405024,'20,May,2004','2004-05-20 00:00:00','   urlLink    This is a picture of Kim Deal I took with my cell phone last month when I saw her play with the Breeders in San Francisco.  She is so awesome!!&nbsp; urlLink  '),(17593,1405024,'19,May,2004','2004-05-19 00:00:00','   Shit dude.  Now it\'s like 12:40am and I\'m still up.  I think I\'ll go wash my face.  Right now I\'m smoking a bowl and listening to some Missy Elliott.  I\'ll be back, most likely.  '),(17594,1405024,'19,May,2004','2004-05-19 00:00:00','   Kim just left. It\'s about 11:30, and I have to get up by 6am tomorrow. Hopefully I will actually get up instead of letting my alarm go for about 2 hours before I decide to do something about it. I guess that means I shouldn\'t do any oxy before bed, that tends to make me especially comatose the morning after. Unless I get to sleep for at least 12 hours, that is. I\'m going to try to be in bed by midnight (which seems doubtful to me, but let\'s think positively) so I can be up by 6 and leave the house at around 7:15 to catch the bus at 7:26. Then I will be at work at around 8:26, which will give me time to get coffee, count the money, and maybe do a line or two in the back room. Haha, just kidding.  So I cleaned up my kitchen area a lot and finally took the double burner I bought out of the box. I cooked for the very first time since I\'ve moved in here! It was very exciting. Simple, but exciting. I made some pasta - I was just excited to be able to have clean dishes to do that, let alone a clean countertop and space to do anything. Yay.   I\'m smoking a bowl and I just did a line...I think I\'ll do something productive for about the next 5 minutes.  '),(17595,1405024,'18,May,2004','2004-05-18 00:00:00','   So I took a shower.  A really long one, it was like 45 minutes long.  I would have stayed in longer, probably, except the hot water supply was rapidly diminishing.  I shaved my legs, but only until just above the knee.  Again, I would have gone all the way up had there been more hot water.  Oh well.  What\'s up with that anyways?  I can\'t believe it\'s already after 6:30.  I feel like I wasted another whole day.  I really should clean up something in my room.  I realized last night as I was pulling pots and pans from a box that I packed all my shit up just about a year ago now.  And I\'m just unpacking pots (I still have a lot more boxes) so I can wash them (who knows about using them).  Hmm.  It makes me think maybe I don\'t really need all the stuff I have in boxes if I\'ve gone a year without it and still have no motivation to unpack anything.  I should sell everything, maybe make a little bit of cash.  That would rule.  Okay, I have to go get dressed and whatnot.  I don\'t know why I feel the need to update this everytime I do something today, but, you know, whatever.  '),(17596,1405024,'18,May,2004','2004-05-18 00:00:00','   Can I just say that I have great tits?  '),(17597,1405024,'18,May,2004','2004-05-18 00:00:00','   So, it\'s 4:30 on Wednesday afternoon.  I just got up about a half hour ago - 3:54 on my alarm clock to be exact.  Yes, I had set the alarm for noon...and I repeatedly hit snooze until I actually got up almost 4 hours later.  Nothing unusual for me there.  Thank god I have the day off - I worked six days in a row before today.  And let me tell you, that sucked.  Working retail that many days in a row, I start to hate people, hate their questions, hate their comments, their clothes, their attitudes...but I still love their money.  I wish I could just take all their money.  And oh do I try.  I\'m trying to decide what to do right now.  My choices are shower, sit here and fuck around on the computer, sit here and do something other than fuck around on the computer, or maybe not sit here but not shower and instead clean my room or something.  And my room really, really needs to be cleaned.  I don\'t know what it is with me.  I have some kind of aversion to cleaning my room.  More than most people.  It\'s not like I just don\'t like cleaning, because I do...I think there\'s some kind of psychological issue going on here.  I tried bringing it up in therapy like 2 weeks ago but I wasn\'t really sure how to explain myself.  But she definitely agreed it\'s the opposite end of the OCD (obsessive-compulsive disorder) spectrum from where I used to be.  Ah, the brain.  How much fun you are to deal with sometimes.  I think I\'m going to hit my pipe and see what happens and where I end up.  '),(17598,1405024,'18,May,2004','2004-05-18 00:00:00','   So here I am, yet again.  Up way too late, yet again.  It\'s times like these when I think to myself, what the fuck am I doing with my life?  Do I have any direction whatsoever?  Then I usually hit my pipe if it\'s around to shut those voices up.  No seriously, I am worried about myself.  How the fuck is it already almost June of 2004?  Why the fuck am I still working at a retail job I took \"temporarily\" last August?  I mean sure I\'ve been promoted a few times and now I\'m assistant manager, so I guess that\'s a reference at least.  Right?  Sigh.  Where\'s my pipe?  A friend suggested to me (a few times now) that I volunteer as a counselor and work with troubled kids or something.  That would let me work with people and help them, and plus it might help me get into grad school.  Which is totally what I want to do at this point, but I feel like I wouldn\'t have the best application.  Yeah, I know I\'d be great for whoever wants me, but how the fuck am I supposed to prove that within the narrow guidelines they use for choosing people?  Someone else might look better on paper, when I know I\'m better.  Or am I?  Of course I am.  I guess I have some kind of direction.  Grad school, straight ahead.  I feel like there\'s so many little pieces of my life that I have to take care of before I even decide on a direction.  Maybe that\'s just how I fool myself into thinking everything\'s okay and I can be immobile forever.  I don\'t want to grow up.  Maybe people feel like that their entire lives, that there are so many things that need to be taken care of before they can move forward in their lives.  Maybe not.  I guess I need to just go forward - as in, decide what I want to do with myself and then start going towards that - and take care of all the little pieces along the way, instead of just pausing everything.  But seriously, I think I needed some time to regroup and collect my life and thoughts and everything and take a break from everything.  Decide what I want to do with myself right now and in the future.  Obviously I don\'t want to do retail my entire life.  I\'m better than that, and besides that, retail sucks.  I would, however, own a business.  And I would definitely not run it the way the owners of the company I work for run theirs.  Let\'s just leave it at that.  Okay, so what do I want to do?  Go to grad school.  Pay my outstanding bills.  Maintain my intelligence and sanity.  That was a long flow of thought...one which flows through my mind every day at least a few times.  I don\'t even know if it made sense.  Wow I have so much to talk about.  '),(17599,1405024,'16,May,2004','2004-05-16 00:00:00','   Boy am I high. I had to work today, which sucked, but really it wasn\'t all that bad. It was pretty slow so I didn\'t get stressed out like I all too often do.&nbsp; Then I came home and took some oxy, and smoked some pot, and did some more oxy, and smoked some more pot, and I\'m still continuing the cycle. The weird thing about me and oxy is that instead of making me feel all lazy and tired, it makes me want to get up and do shit. I\'ve already cleaned some dishes (which seriously is a huge accomplishment for me) and started cleaning the bathroom. Let\'s talk about that for a moment. I have this problem where I get easily distracted...I always want everything clean, so I start cleaning one thing and then realize something else needs to be cleaned so I\'ll start doing that and then I wind up starting so much that nothing ever gets done. Then it looks worse than it did to begin with. Sigh. This is something I need to work on.   I went to this crazy concert on Friday night at the Fillmore.  urlLink Peaches . I honestly still don\'t know what I think about it. All week I was wondering what the fuck to expect...and when we finally got there (Tim and I), I probably couldn\'t have even imagined the diversity of the crowd. I don\'t really feel like going into it right now. But, it was interesting, to say the least. Don\'t get me wrong, her music is great. I think really I was just wishing I had some certain drugs...   I\'m all itchy.  '),(17600,1405024,'15,May,2004','2004-05-15 00:00:00','   Okay, another quiz.  I saw this one and I just had to take it - I love that little happy bunny, I keep buying more shit with him/her on it.  How funny.  I feel really fucked up right now, I think maybe I did too much oxy.  By too much I mean like more than I would usually do so I feel more fucked up than I am used to.   urlLink Here\'s the quiz.     you are the \"you\'re so dumb\" happy bunny. you are brutal in your words and enjoy putting others down.     urlLink   which happy bunny are you?     brought to you by  urlLink Quizilla    Where\'s my snare?  '),(17601,1405024,'15,May,2004','2004-05-15 00:00:00','   Okay, I just took another dumb quiz.   urlLink If You Were A Barbie, Which Messed Up Version Would You Be?     You\'re Gangsta Bitch Barbie.  You\'re tough and you like it rough, and of course you like to pop a cap in any wiggers ass.    urlLink   If You Were A Barbie, Which Messed Up Version Would You Be?     brought to you by  urlLink Quizilla    I guess that\'s cool.  What a stupid quiz though.  '),(17602,1405024,'15,May,2004','2004-05-15 00:00:00','   What childhood toy from the 80s am I?    You\'re a Speak & Spell!!  You nerd, you.  Just because you were disguised as a toy doesn\'t mean you weren\'t educational, you sneaky bastard.    urlLink   What childhood toy from the 80s are you?      brought to you by  urlLink Quizilla      urlLink Check it out, yo.  '),(17603,1405024,'15,May,2004','2004-05-15 00:00:00','   My typical day is such a rollercoaster of emotions.  What the fuck dude.  On one hand I totally want to go into it and write all about it right now, and on the other hand I just want to snort some more oxy.  Sigh.  I wish I didn\'t have to work tomorrow.  I\'ll write more in a little bit.  Maybe.  I do want to, I have a ton to say.  '),(17604,1405024,'30,June,2004','2004-06-30 00:00:00','   urlLink Army Will Recall More Than 5,000 Veterans From 48 States (washingtonpost.com)   man, everything is so fucked up, isn\'t it?  '),(17605,1405024,'29,June,2004','2004-06-29 00:00:00','   Yeah, we\'re straight, we get high on our music...  Not me though, I get high on drugs.  I worked today...it seemed to go by extremely fast, too fast almost for me to even get anything done that I wanted to.  Yet at the same time it seems like this day has been going on forever.  After work I actually went to the bookstore to buy that new  Star  magazine about M-K but they didn\'t have it.  I guess it\'s too low class or something...but they have  Us  and  People !  What the fuck.  Nevermind that my work (the salon) actually subscribes to that magazine...I wanted it today, dammit.  TODAY!  Man, now I\'ll have to wait or something.  Fuck.  Care for a line, Mary-Kate?  Something smells really bad in my room, yet I don\'t have the energy nor motivation to find the source of the stankiness.  Maybe in a little while...  Man I just realized how funny it would be if I actually documented all my terrible customer stories.  Maybe having them all compiled and together would help me realize how much stupid people suck sometimes.  Oh wait, I already knew that.  '),(17606,1405024,'29,June,2004','2004-06-29 00:00:00','   According to the internet (and what more trusted source is there?), Mary-Kate Olsen is addicted to coke, and that\'s the  real  reason she\'s in rehab.  Whatcha think \'bout that?  Check it out  urlLink here .  '),(17607,1405024,'28,June,2004','2004-06-28 00:00:00','   Looky here, I\'m awake.  Must be some kind of miracle if you ask me.  I got here at about 9:30 this morning, stayed up until, ohhh around 10:30, and then fell asleep (not even on purpose).  I woke up at like 12:50 when I got a phone call, and then looked for awhile at the bus schedules so I could go to therapy at 3:30.  All of the ways they were telling me to get there were ridiculously complicated, so I decided that I would call up my therapist and ask to reschedule for sometime when I can just leave from work.  (I had already been leaning in that direction...I hate taking the bus, and I hate taking unknown bus routes even more.)  She said that was fine, so now I\'m going on Thursday instead.  Yay!  I get to leave work in the middle of the day, but then I have to go back.  At least I get to leave work though and split the day up a little.  I\'m gonna hop in the shower.  '),(17608,1405024,'28,June,2004','2004-06-28 00:00:00','   urlLink \"California is indeed a situation of mass destruction… it\'s truly a danger to the rest of the world. But with your help we can help make it a safer place to live.\"   (click that)  '),(17609,1405024,'28,June,2004','2004-06-28 00:00:00','   urlLink hahahaha   I think this is one of my new favorite sites.  '),(17610,1405024,'28,June,2004','2004-06-28 00:00:00','   So I ended up going back to sleep earlier.  I\'ll only nap for an hour or two, I told myself.  Seven hours later I finally woke up, showered, and now here I am.  Man, I hate when I do that.  I feel like I wasted my whole day, and I had so much cleaning planned.  I should pack a bowl.  '),(17611,1405024,'27,June,2004','2004-06-27 00:00:00','   So I was looking at my blog a second ago, and I couldn\'t help but notice the banner ad at the top of my page (provided by google based on keywords in my entries, yay) and the main two links were both for alcohol and drug addiction recovery.  Um, okay.  There were also two little extra links under that...one said \'Kim Deal\' (this is awesome); the other said \'crack cocaine.\'  Google must think I\'m a crackhead obsessed with Kim Deal and I need to recover or something.  Great.  '),(17612,1405024,'27,June,2004','2004-06-27 00:00:00','   Okay, so I\'ve eaten my two eggos, I\'m drinking a Dr. Pepper, and now I want to go back to sleep.  What the hell.  I will somehow have to resist the urge.  Fuck!  '),(17613,1405024,'27,June,2004','2004-06-27 00:00:00','   Wow.  It\'s like, twenty minutes to 10 on this Monday morning, and I\'m asking myself repeatedly why I\'m awake.  I slept at A\'s last night (as has been my current trend), but we weren\'t able to fall asleep until like almost 5.  He had to get up for work this morning; I, having the day off (thank god), merely had to get up so he could take me home.  Usually in these types of situations I instantly go back to bed as soon as I walk through my door, but today I felt like I should stay up and maybe be productive or something.  But now, sigh, I am starting to get tired again.  Perhaps I should make something to eat, like an eggo, or have a can of my favorite pick-me-up elixir, Dr. Pepper.  Maybe I should just take a shower and get on with it.  We did a little bit of oxy last night too, and I remember getting paranoid that I would die in my sleep.  I promised myself I wouldn\'t mix all those drugs at once anymore.  We\'ll see what happens.  I am so excited that I have today AND tomorrow off.  Two days in a row!  Awesome.  Lots of blogging will occur.  Oh, the toaster\'s ready for me to insert my eggo.  Mmm, syrup.  '),(17614,1405024,'27,June,2004','2004-06-27 00:00:00','   Recently I was driving (when I rented a car last week to drive to LA) and I saw a car with a bumper sticker that read \"Our governor can terminate your governor!\" or something to that effect.  Funny, but sad.  Very very sad.  Didn\'t Ronald Reagan become governor of California because he was an actor too?  These people should stick to their bad movie acting, and leave the political shit to the other bad actors...er, politicians.  '),(17615,1405024,'22,June,2004','2004-06-22 00:00:00','   I haven\'t written all week!  Since Tuesday!  Sorry.  I guess I\'ve been keeping myself busy, or I\'ve just been unable to write for whatever reason.  I\'ve been spending a lot of time at A\'s apartment, so I haven\'t really been around...and even though he has a computer, I don\'t really want to sit and blog in front of him.  I worked today from 10-5.  I got home about an hour ago, and now I feel really tired.  I\'m eating a leftover vegetarian taco I got for lunch but didn\'t eat at the time - black beans, salsa fresca, guacamole, cheese...the usual.  No meat.  Yum.  Check out this pic I took with my phone on the way home from work today.  It\'s a bus that is painted to look like a giant sandwich.  Hmmm.   urlLink    eat me!  I\'m gonna eat my taco!  '),(17616,1405024,'22,June,2004','2004-06-22 00:00:00','   urlLink Defamer: Fifty Percent Of Olsen Twins Reportedly In Rehab  '),(17617,1405024,'19,June,2004','2004-06-19 00:00:00','   please?  it fits in my bowl quite nicely...  urlLink Smoking Sign  '),(17618,1405024,'19,June,2004','2004-06-19 00:00:00','   I might actually download Bam Thwok on iTunes.  Of course, I\'ll look for it on Kazaa first...   urlLink Pitchfork: We Are The World  '),(17619,1405024,'19,June,2004','2004-06-19 00:00:00','   I feel like I haven\'t written anything in, like, forever.  I mean, I have, but I haven\'t really gone deep inside my head.  It\'s 11:30 on Saturday night, I\'m watching Joe Schmo 2 on mtv (I thought this was on Spike TV?), and I\'m eating a Smart Ones Rice and Beans dinner.  Yum.  I worked today, I worked tomorrow, but I didn\'t work for the last three days.  The reason for that, my friends, is that I went down to LA for a few days to visit my family - my little brother was graduating from 8th grade.  My fucking nails are too long again for me to type comfortably - it seems like they grow way too fast sometimes.  I guess that\'s better than them not growing.  It just seems like I have to cut and file them all the time.  This show is fucking hilarious, btw.  If I remember, I\'ll write more in a little bit.  '),(17620,1405024,'18,June,2004','2004-06-18 00:00:00','   I fucking love this commercial!   urlLink Pepto Bismol - Pepto Dance , ad, advert | visit4info | TV/Cinema  '),(17621,1405024,'18,June,2004','2004-06-18 00:00:00','   urlLink WSJ.com - plug tails  '),(17622,1405024,'13,June,2004','2004-06-13 00:00:00','   It\'s only ten minutes to 7:00, but today feels like it\'s lasted forever already.  I got up at around 8 or 8:30 and got ready for work, worked 10-5 (which went by really slowly), and now here I am.  I am extremely fucked up right now.  I wish I had an ice cream cone.  I\'m ready for a nap.  I want to clean...but again, I\'m so fucked up.  My nails are getting too long again to type comfortably.  I have so much to write about, but I should clean...but I\'m so fucked up...  '),(17623,1405024,'12,June,2004','2004-06-12 00:00:00','   Hey, with Reagan dead, that means the whole \"just say no\" anti-drug campaign is dead too, right?  Man...I remember being like 5 years old and seeing all those dumb anti-drug commercials.  Obviously they did absolutely nothing to curb my drug use, but sometimes I feel bad when I snort stuff.  I\'m completely fine with smoking pot all day long.  '),(17624,1405024,'12,June,2004','2004-06-12 00:00:00','   I am really fucking high right now.  What a beautiful day boys and girls!  After getting really depressed last night, I went to bed around 11:30.  The depression mixed with the sedatives I had snorted made me pretty sleepy I suppose, and besides that I just didn\'t want to be awake any longer.  After being excited all week about A coming home, the incredible letdown that occurred just made me want to end the day as soon as possible.  I knew he was coming home last night, and I was totally planning on seeing him (of course).  He called me at around 8pm or so to say \"hey,   just left San Francisco so they shouldn\'t be here for about 30 minutes, and C [his friend who he went on the trip with] is in the shower, so I have a little time to see you.\"  I was completely insulted at this - I\'m not important enough to have my own time?  I have to be hidden from his \"real friends\" and squeezed in miniature time slots between his other plans?  Fuck that dude.  I told him no, I didn\'t want to see him like that, it made me feel really shitty and unimportant.  (This is weird btw, since I usually let him walk all over me.)  I wanted to see him when I wouldn\'t have a ten minute time limit and he wouldn\'t have to rush out and pretend he wasn\'t just with me.  \"What the fuck??  I\'m offering to come see your ass and you\'re going to bitch about it?  Fuck that.  You know what?  Fuck you!\"  I\'m sure he continued along those lines, but at that point I hung up.  Fuck him for talking to me like that - I did absolutely nothing to warrant that kind of response, and I\'m not going to sit on the phone and listen to him behave like that.  Besides that, I didn\'t even say anything wrong, it seems pretty logical to me.  He was apparently extremely offended that I didn\'t want to drop everything and see him for seven minutes and 23 seconds.  So, I did some random stuff around my place and went to bed.  I cried and cried, and eventually fell asleep.  '),(17625,1405024,'11,June,2004','2004-06-11 00:00:00','   This article is pretty interesting...   urlLink baltimoresun.com - City seeks a jolt from Starbucks  '),(17626,1405024,'08,June,2004','2004-06-08 00:00:00','   He\'s been home for less than an hour, and he\'s already calling me up bitching at me and cussing.  I feel so stupid for actually missing him and being excited that he was coming home tonight.  Blah.  '),(17627,1405024,'08,June,2004','2004-06-08 00:00:00','   wouldn\'t it be cool if that said what a long day of fucking?  I woke up at around 7:45am - much later than I had intended on getting up.  It probably had something to do with all the ativan I crushed up last night and snorted lines of until I couldn\'t stay awake any longer - come to think of it, I was still pretty groggy when I finally opened my eyes in the morning.  I didn\'t have time to shower, so I washed my face and then crushed up another ativan and snorted it.  I put on some makeup quickly and then my coworker was here to pick me up at around 8:12.  We got to work at around 8:40 - she dropped me off in front of the store to turn off the alarm and clock us in while she found a parking spot.  While she did that I crushed up a couple more pills in the breakroom and snorted those, and then I went out and tried to be coherent.  I was very relaxed and unstressed, I\'ll tell you that much.  Sedatives are great.  Actually, I think drugs in general are great.  Sedatives are great because they allow you to just calmly deal with what\'s going on.  So on with my day.  I made our daily rounds to Starbucks to get my iced hazelnut latte, stopping at Ritz on the way to say hi to whoever was working.  David gave me another 1-2 grams of weed as a gift.  At first I was like how much do you want for this, and he was like nothing it\'s a present, and I was like no really...he just wanted to give it to me.  I don\'t get it, but I do like free weed.  David\'s nice.  I made some dirty money with my coworker...we\'re so shifty.  Went in back and snorted another pill in the break room.  Borrowed a friend\'s car to go to my therapy appointment at 3:30pm and talked about a lot of issues, which was good.  Smoked a bowl in his car, which if he knew about he would kill me.  Went back to work, my boss left, and then I made some more dirty money.  Being weird, I took about $40 in $1\'s just so I can enter them all into  urlLink www.wheresgeorge.com .  I\'m strangely addicted to that site recently.  Then I got extremely stoned in the back yard with Megan from Ritz (she\'s a total hippy) and her boyfriend Rob, and bananaman Luis.  Went to Oscar\'s for a veggie burger with Megan and Rob, and then they dropped me off.  Now it\'s like 9pm, and I\'m fucking tired.  And crushing and snorting another ativan.  Details to follow!  '),(17628,1405024,'07,June,2004','2004-06-07 00:00:00','   So I have been retaking these enneagram quizzes at various times for a few reasons - it was recommended to take the tests at least several times and see what you get the most and what is most accurate, as your mood may be different at different times, thereby affecting your answers.  Makes sense, right?  The one you get the most should be the most accurate for the test taker.  Here are my results so far:   First:          Conscious self'),(17629,1405024,'06,June,2004','2004-06-06 00:00:00','   ENTP  - \"Inventor\". Enthusiastic interest in everything and always sensitive to possibilities. Non-conformist and innovative. 3.2% of the total population.'),(17630,1405024,'06,June,2004','2004-06-06 00:00:00','   INTP  - \"Architect\". Greatest precision in thought and language. Can readily discern contradictions and inconsistencies. The world exists primarily to be understood. 3.3% of total population.'),(17631,1405024,'06,June,2004','2004-06-06 00:00:00','   I do wish I had more weed, though...hopefully tomorrow...  '),(17632,1405024,'06,June,2004','2004-06-06 00:00:00','   I think I should snort ativan every time I need to do something.  That was great!  I mean I\'m definitely still feeling it...but I was totally able to go do everything I needed to do and not feel weird about it or try to convince myself to do it later like I usually do.  It\'s like some kind of underlying anxiety that prevents me from getting shit done.  Where is that coming from?  But right now that anxiety isn\'t there...I\'ve sedated it enough to where I actually feel productive.  Go me!  Plus, I feel so relaxed and peaceful...  '),(17633,1405024,'06,June,2004','2004-06-06 00:00:00','   I need to run some errands. Drop something off at the cleaners, refill some prescriptions, deposit some checks, buy an onion. Get a latte, duh. I am really fucked up right now - I decided to snort two pills of ativan (lorazepam), and now I\'m on my third. I usually take one a day to mellow me out a bit, but damn I am fucked up right now. I\'m also semi-smoking a bowl (trying to conserve, I only have about 1.5 bowls left). The ativan drip in the back of my throat is sweet tasting - not bitter like that associated with coke. I hope I\'m able to function out in public right now...  '),(17634,1405024,'06,June,2004','2004-06-06 00:00:00','   urlLink Bookies Offer 3-1 Odds on Quick End to J-Lo Marriage   She\'s already married?  Again?  Where do I place my bets on when it will end?  '),(17635,1405024,'06,June,2004','2004-06-06 00:00:00','   I\'m doing lines of ativan.  Hah.  I think I just like to get high...  '),(17636,1405024,'06,June,2004','2004-06-06 00:00:00','   ...and my results are fairly similar.  Not exactly the same, but similar.          Conscious self'),(17637,1405024,'06,June,2004','2004-06-06 00:00:00','   Conscious self'),(17638,1405024,'05,June,2004','2004-06-05 00:00:00','   urlLink SimilarMinds.com > Word Association Test Results :          I have issues with...        patience  discipline  religion  age  mother        urlLink Take Word Association Test             I have issues with...        law  discipline  selfishness  history  innocence        urlLink Take Word Association Test    So they give you a list of words, and you react as quickly as you can to the words by clicking positive, neutral, or negative.  Regardless of your choices, the test tells you that you have issues with the words/ideas/concepts that you had the longest response time for.  It\'s really interesting actually.  Odd how \'discipline\' is on there twice, eh?  Maybe I really do have issues.  '),(17639,1405024,'05,June,2004','2004-06-05 00:00:00','   Personality Disorder Test Results       urlLink Paranoid    ||||||||||||||   54%        urlLink Schizoid    ||||||||||||   50%        urlLink Schizotypal    ||||||||||||||||||   74%        urlLink Antisocial    ||||||||||   34%        urlLink Borderline    ||||||||||||   50%        urlLink Histrionic    ||||||||||||||||   70%        urlLink Narcissistic    ||||||||||||||||   62%        urlLink Avoidant    ||||||||||||   42%        urlLink Dependent    ||||||||||   38%        urlLink  Obsessive-Compulsive     ||||||||||||||||||   78%'),(17640,1405024,'05,June,2004','2004-06-05 00:00:00','   urlLink What Famous Leader Are You?    urlLink personality tests by similarminds.com'),(17641,1405024,'05,June,2004','2004-06-05 00:00:00','   So I just put on more Fake Bake.  I don\'t know why, probably because I\'m drunk, but it seemed like a good idea.  I think I put on too much , so maybe it will look way artificial when I wake up tomorrow.  Either that or streaky, since I\'m so fucked up.  I also spilled a bunch on my carpet - I tried to clean it up with resolve but I\'m not sure how well it worked.  It definitely turned the towel I was using pretty pink...  '),(17642,1405024,'05,June,2004','2004-06-05 00:00:00','   I totally have the hiccups for the second time today!!  Maybe this time it\'s because I\'m drunk, but usually it happens when I haven\'t smoked in awhile (as in all day) and my lungs need to \'be expanded.\'  Awesome.  I should smoke a bowl!  '),(17643,1405024,'05,June,2004','2004-06-05 00:00:00','   Seriously!  I mean, not to sound like a bitch or anything, but I was looking at some of the girls that were there and just thinking, wow, she\'s really ugly.  I\'m so grateful that I\'m one of the hot ones and not the fat ugly ones!  '),(17644,1405024,'05,June,2004','2004-06-05 00:00:00','   It\'s almost 4:20!  '),(17645,1405024,'04,June,2004','2004-06-04 00:00:00','   Someone told me yesterday that she loved my hair color - it\'s very exotic.  I pointed out that she\'s a dumb biotch, since I\'m white as fuck.  Lol, just kidding.  I used to be until I used that Fake Bake the other night.  I slept with it on and took a quick shower yesterday morning.  I was very concerned it would be streaky or orangey, but it actually looked really natural.  There was a part on my upper right thigh near my ass where I saw some streakiness, but no one else can see that high so it doesn\'t really matter.  I think it\'s gone now anyways.  I kind of want to do it again tonight before I go to bed - I hope it won\'t look more artificial.  I only did it on my legs, because if I fuck up I can just wear pants - if I fucked up my arms, wearing long sleeved shirts would suck in this weather.  Okay, more soon.  '),(17646,1405024,'04,June,2004','2004-06-04 00:00:00','   Reagan\'s dead, I\'m bored and horny...what to do?  Oh right, let\'s do some drugs.  I\'ll finish this bowl I\'m smoking, do some other stuff...then I\'ll blowdry my newly colored hair.  Okay, really, it\'s not a new color, it\'s the same color I\'ve had for the past few months now.  But I had my roots touched up yesterday and redid some highlights.  Awesome!  Next week I\'m going to have it cut - not short or anything, just a trim and maybe redefining my layers.  Yay.  I need to blowdry my hair, I can feel it getting dry by itself as I sit here - oh no!  Insert insufflated substance here.  '),(17647,1405024,'04,June,2004','2004-06-04 00:00:00','   It\'s been too long since I\'ve had sex.  Like two weeks already, wtf.  I really need some cock - let\'s hope I don\'t need it bad enough to just fuck anyone though.  Fuck.  '),(17648,1405024,'04,June,2004','2004-06-04 00:00:00','   urlLink Ronald Reagan: Strong, Controversial World Statesman   Okay, so I removed the post immediately preceding this one about Reagan being sick...seems like he died while I was in the shower.  This is making me feel kind of weird - it\'s not like I liked the guy or anything, but he\'s the first president who actually served while I was alive to die.  Wow, I feel old, like this is some kind of life marker or something.  Sure, he was elected when I was only four months old, but still.  Now, when a president that I actually voted for dies, that will be something else...  '),(17649,1405024,'04,June,2004','2004-06-04 00:00:00','   So I posted some pics of those dumb signs I was talking about a few days ago from Safeway.  Scroll down and check it out, yo.  '),(17650,1405024,'04,June,2004','2004-06-04 00:00:00','   urlLink Capitol Hill Blue: Bush\'s Erratic Behavior Worries White House Aides  '),(17651,1405024,'03,June,2004','2004-06-03 00:00:00','   um, apparently not...   urlLink North Korea recalls mobile phones  '),(17652,1405024,'03,June,2004','2004-06-03 00:00:00','   Have you ever heard of that website  urlLink Where\'s George?  where you enter in the serial numbers of the cash you have and it tells you where it\'s been and shit?  It\'s pretty cool, I have been hearing about it forever and just always felt like it was probably pretty lame, but now I can see myself kinda getting into it.  I just entered in like 12 bills.  Do your own!  '),(17653,1405024,'03,June,2004','2004-06-03 00:00:00','   This won\'t be a long post...tonight\'s Friday night, I have tomorrow off, and I plan on doing nothing all night.  Chilling, getting high, and possibly cleaning if I somehow find the ability to motivate myself to do said activity.  I want to post lots of short little blurbs tonight about whatever I\'m thinking at the time.  The guy from AG Ferrari\'s came to the store today to say hi to me and check shit out.  I knew he was interested in me, I just knew it.  I just get that feeling when I go in there.  But that\'s cool, he\'s pretty cute actually.  Maybe I can fuck him in the back of his store.  Or my store, like I did with the guy from Starbucks.  Hmmm.  AGF guy is studying creative writing at SF State.  Interesting.  He\'s really into writing fiction.  I wonder if he smokes pot.  Probably, if he\'s into writing fiction, right?  I hope so.  Hmmm.  '),(17654,1405024,'03,June,2004','2004-06-03 00:00:00','   Do I even have to tell you?  No, of course not.  Unfortunately a quick shower is just not Bustily possible.  Okay, actually, it probably is, but tonight I wanted to shave and exfoliate and whatnot so I could try out some  urlLink sunless tanner  I got from work.  It\'s taken me awhile to work up the nerve to try it - generally I\'m scared to use such products out of fear of turning my usual pasty white legs into more of an orangey muddy color.  Let\'s hope this doesn\'t happen this time.  I\'m also paranoid of the streakiness.  I\'m planning on wearing a skirt tomorrow - hopefully I won\'t wind up having to wear pants.  Wow I totally forgot I was writing this.  Now I\'m tired.  '),(17655,1405024,'03,June,2004','2004-06-03 00:00:00','   Will I fit my shower into the 15 minute timeframe?  Let\'s find out.  Ready, go.  Oh wait, let me do some drugs first.  '),(17656,1405024,'02,June,2004','2004-06-02 00:00:00','   urlLink D-Day 1899 and President Denzel Washington is leading liberation of New Zealand from the Nazis  '),(17657,1405024,'01,June,2004','2004-06-01 00:00:00','   So I just got home from work a little while ago.  It wasn\'t too bad...the day went by pretty fast until my boss and I had to leave to go to a class on this new line that Sebastian is releasing called Body Double.  There were about 8 or so of us there at the class, and at the end, the educator asked us 10 questions about what she had just told us.  Each right answer won us a dollar; I won $5.  I knew the other answers, I just felt bad and wanted to give other people a chance.  God, people are so dumb.  It\'s really not that hard to repeat something you were just told, is it?  But I guess this class was a little bit different than the ones I took at Cal.  Now I\'m home, doing drugs, preparing to be productive.  I want to clean...  '),(17658,1405024,'01,June,2004','2004-06-01 00:00:00','   I just got home a little while ago from the laundromat.  Now I have clean clothes.  Thanks Tim!!!  Now, I just have to put them away...  '),(17659,1405024,'01,June,2004','2004-06-01 00:00:00','   ...and ran some of my dumb little errands.  I shipped out that shit to my friend, bought tampons and a paper towel holder from Rite Aid, bought some ketchup and butter from Safeway, and got my iced venti hazelnut lowfat latte.  Yay.  I wish I could have two days off in a row, that would be awesome.   urlLink    Here\'s one of the signs I was talking about from Safeway.  Man I love stuff like this.   urlLink    It\'s that time of the month...time to get my novelties.  '),(17660,1405024,'01,June,2004','2004-06-01 00:00:00','   urlLink U.S. Backs New Iraq Leaders, Says \'Not Puppets, Like President Bush\'   [headline may have been altered somewhat by yours truly]  '),(17661,1405024,'01,June,2004','2004-06-01 00:00:00','   It\'s already fucking June.  What the fuck man.  '),(17662,1405024,'01,June,2004','2004-06-01 00:00:00','   Let\'s list my dumb little errands.  I have to buy tampons, first of all.  While that may sound like it sucks, I\'m actually very happy for this, as it means once again that I am not pregnant.  Thank god.  This reminds me of the fact that in Safeway, in the aisle with the tampons and pads and \'sanitary napkins\' and whatnot, the guide thing hanging from the ceiling that says what that aisle contains doesn\'t say \'feminine hygiene\' or \'feminine needs,\' like I\'ve so often seen before.  No, it says \'novelties.\'  Hahaha.  It\'s a novelty is it?  It\'s been around for thousands of years fuckheads.  Seriously though, I get a kick out of their little labels for random supermarket shit.  \'Can Vegetables.\'  \'Cook gadgets.\' (This is supposed to be cooking utensils and shit.)  It\'s like some little asian gnome went around and built all their dumb signs and just fucked shit up bigtime.  I also have to mail some shit to my friend.  He better be grateful.  Oh, and there\'s the necessary daily latte as well, can\'t forget that...  '),(17663,1405024,'01,June,2004','2004-06-01 00:00:00','   I finally got out of bed at around 2pm.  I had set my alarm for 11am, hoping to not sleep all day like I usually do on my days off, and besides that I knew I was supposed to be in Berkeley for therapy at 3:30.  I wanted to get up and be productive...after all, that is my \'thing\' - being productive.  Maybe that\'s because I feel in general that I am entirely too unproductive.  Anyways, I woke up at 10 and called my friend to talk about some random shit, and then wound up going back to sleep until I got up at 2.  I hit snooze about 5 billion times just like every other day - quite honestly, I\'m suprised I haven\'t smashed my alarm clock to pieces in my sleep yet.  I took a shower which ended up being about 37 minutes (haha, \"about,\" you know I totally looked at the clock) - and that was with me trying to not take too long.  Damn.  I realized there was no way I could be in Berkeley for my appointment anywhere near 3:30, so I called and left a message saying I missed my bus and I was really sorry but I couldn\'t make it.  I dried my hair, put on some moisturizer, threw on some little shorts that I wear around the house, and now here I am.  I\'m smoking my first bowl of the day, and already I am completely stoned.  When you smoke as much as I do, it\'s not really a matter of getting high anymore, it\'s more like, \"what can I do to maintain my constant high?\"  Waking up and then smoking doesn\'t just get me really high, I feel like it actually renews my level of high-ness from at least the night before.  It reactivates the high levels of THC flowing through my body already - so, instead of me feeling like I just took 3 hits from my bowl, I feel like I\'ve been smoking for awhile and the few hits I just took are just added to that.  Wow, I\'m really high.  And this Dr. Pepper isn\'t helping.  I feel like I can hardly think.  I have some dumb little errands I have to run, and then some not as dumb shit I need to take care of, but I usually put that kind of shit off.  Maybe I should get dressed and put on some makeup, yes?  Fuck.  What to do.  '),(17664,1405024,'27,July,2004','2004-07-27 00:00:00','   My fucking internet is back up.&nbsp; I\'ve been way too productive the past few days...at last, something to distract me!  '),(17665,1405024,'22,July,2004','2004-07-22 00:00:00','   So today at work this guy came in asking for some semipermanent blond toner.&nbsp; I was really the only one there who knew anything about what he was talking about, so I walked with him to show him where it is in the store.&nbsp; I got this weird feeling from him, but I&nbsp;attributed it to&nbsp;me being incredibly stoned.&nbsp; But, whatever, most guys probably feel weird coming into a beauty salon, I&nbsp;figured.&nbsp; Then he started saying how much he liked my hair, and he even touched it at one point when talking about what color he wanted his hair.&nbsp; \"You have really nice hair,\" he said.&nbsp; \"Oh, thanks.\"&nbsp; I know I have nice hair, silly, silly boy.&nbsp; I wasn\'t exactly sure about the instructions for what he was getting since his hair was already bleached, but I tried to help him the best I could.&nbsp; I hoped I wasn\'t telling him to do something that would make his hair fall out or fuck anything up, but I was pretty confident I wasn\'t.&nbsp; There were a billion people in the store, he waited in line, I rung him up, and then&nbsp;he left.   A couple hours later,&nbsp;the store phone rings.&nbsp; My coworker&nbsp;who answered the phone&nbsp;told me that someone I helped earlier was on the phone for me; I was silently hoping he didn\'t go home and follow my instructions and then all his hair fell out.&nbsp;&nbsp;Of course, it was nothing like that.&nbsp;&nbsp; \"Hi, I\'m Roberto, you helped me earlier, what\'s your name?\"&nbsp; So I told him, and then he was like \"well I really think you\'re beautiful and I\'d really like to get to know you better, maybe I can give you my number and we can go out and get a drink sometime.\"&nbsp; Alright, I said.&nbsp; It\'s not like I couldn\'t just take his number from him over the phone and then throw it away or something.&nbsp; And plus, he went through the trouble of looking up the store number and calling back and asking for me.&nbsp; On the phone he said something about really wanting to ask while he was there, but there were too many people and he was nervous or something.&nbsp; Sometimes I feel like I have some stranger power of attraction that either attracts complete stalkers/weirdos, or just causes people to behave as such.&nbsp; Or just that weird things happen to me pretty consistently.&nbsp; That\'s just how my days normally go though...  '),(17666,1405024,'20,July,2004','2004-07-20 00:00:00','   Today seemed like a long day, although I guess it wasn\'t that bad.&nbsp; I had to be at work at quarter to 9 to open the store in the morning, and since I slept at A\'s last night (yes, there is definitely a correlation between the amount of time I spend over there and how often I post), I had him drop me off at my coworker\'s since we were opening together.&nbsp; The manager and I were supposed to have an interview with some girl at 11, but she never showed up.&nbsp; Bitch.&nbsp; One of the registers broke, but then our other coworker brought In-N-Out, so that made things seem a little better.&nbsp; The manager and I hung up some random pictures in the bathroom (one of our favorite things to do is take weird pictures when we\'re supposed to be working).&nbsp; I left early to go to therapy, so that was cool, and then I took the bus home.&nbsp; No one tried to sell me any weed today!&nbsp; I guess that\'s good, I hate being bothered.&nbsp; And of course I\'m not going to buy some dumb weed off some dumb guy on the street.&nbsp; Duh.&nbsp; It\'s funny, I don\'t even look like a typical stoner.&nbsp; Most of the time at least.&nbsp;&nbsp;All this talking about weed is making me realize I haven\'t smoked all day...I should pack a bowl!&nbsp; Yay.   My room is a fucking mess. &nbsp;I need to clean it, but the idea of that is just so unappealing to me.   Okay, I\'m going to try to motivate myself.&nbsp; Perhaps I should offer myself a reward.&nbsp; Any suggestions?  '),(17667,1405024,'15,July,2004','2004-07-15 00:00:00','   Today wasn\'t a bad day.&nbsp; I was going to go to Express to get some new clothes, but instead I went to Orchard to buy some plants and soil.&nbsp; When I got home, I got really high (maybe I should just say higher) and potted my new plants.&nbsp; Yay! &nbsp; Now I\'m making rice-a-roni.&nbsp; It smells so good.&nbsp; I\'m super hungry...  '),(17668,1405024,'06,July,2004','2004-07-06 00:00:00','   I\'m the naked one, not the drugs.&nbsp; But I am doing drugs, naked.&nbsp; Yay!&nbsp; I love taking showers when I\'m really high! &nbsp; I know I\'ve been mia lately...I have SO much to write about!&nbsp; Among them, my birthday...sigh.&nbsp; Another year older.&nbsp; All of a sudden I feel so much older, being 24 instead of 23.&nbsp; I mean, of course I\'m the same from one day to the next, but really I am trying to be a little bit more responsible...a little bit.  '),(17669,1405024,'06,July,2004','2004-07-06 00:00:00','   urlLink New York Post Online Edition: gossip   Oh man, I totally shouldn\'t laugh or make fun, but this is just so easy...I hope I don\'t go to hell for this.  Cocaine hell.  Hmm, maybe that wouldn\'t be so bad actually...  \"MARY-Kate Olsen ain\'t got milk. The troubled teen superstar signed on for the iconic \'milk-mustache\' campaign with her sister Ashley in May, saying, \'We want to make sure our fans are healthy like us.\' Following news that Mary-Kate has anorexia, the ads have now been canceled \'out of sensitivity for her situation,\' according to milk marketeers the Milk Processor Education Program. Claims in the Star and National Enquirer that Mary-Kate is also battling cocaine addiction - which her reps vehemently deny - no doubt also detracted from the advisability of showing the girls with a white substance on their upper lips.\"   urlLink    mmmm, \"milk\" (I got this picture from  urlLink defamer )  '),(17670,1405024,'06,July,2004','2004-07-06 00:00:00','   urlLink Jeffrey Jones Sex Offender Bust - July 2, 2004    Sidenote (completely unrelated): next pet rat I get, her name will be Ferris. Awwwwww, how cuuuute.  '),(17671,1405024,'06,July,2004','2004-07-06 00:00:00','   urlLink Anti-Moore Film Festival Planned   An anti-Moore film festival?  Based in Texas?  No way!  Too bad we can\'t transplant the population of Berkeley for a few days to show them Texas folk what this is all about...talk about close minded.  '),(17672,1405024,'06,July,2004','2004-07-06 00:00:00','   urlLink ExtraTV.com : Moore and Stern Take On Bush   [btw, I\'m totally annoyed at blogger right now, it seems like I keep posting things, and then when it comes time to publish them, I get that dumb page that says the page can\'t be loaded, and then my whole entry is lost.  What happened to any sense of stability?  I\'m mad that my entries keep getting lost...]  He\'s likely to become one of the most influential men in American politics -- and he\'s not even a politician. Michael Moore\'s scathing anti-Bush film, \"Fahrenheit 9/11,\" has earned more than $60 million at the box office, and Time magazine made Moore this week\'s cover boy.  According to Time Senior Writer Richard Lacayo, Moore\'s movie could potentially cost Bush a second term in the White House. \"Moore\'s been able to make this film very quickly and drop it right into the most potent moment in the election season,\" Lacayo insisted.   \"Fahrenheit 9/11\" paints the president as incompetent and uncovers Bush family connections to the powerful Bin Laden family. Meanwhile, the Bush camp has called Moore\'s film a \"pack of lies.\" But in attempt to be fair, Moore set up a \"Fahrenheit war room\" to deal with the accusations of inaccuracies.   Moore has even employed former Clinton administration officials to deal with the media blitz. But it seems that the director has got a new best friend in his mission to remove Bush from office -- radio icon Howard Stern.   Stern has been on a daily radio rant against the Bush administration, saying he\'s been unfairly targeted by the FCC for indecency. So could a Stern/Moore team be enough to get the president out?   Lacayo told us it\'s amazing how much these two could sway the election by working together: \"Taken together, Michael Moore and Howard Stern are both at the top of their industries: radio and film. They both have an enormous audience, and together, they\'ll be punching away until Election Day.\"   With four months until the election, there\'s still plenty of time for campaigning and plenty of time for America\'s newest power couple to make an impact. Meanwhile, Moore is already eying his next target; he will focus on HMOs and the healthcare system in his next documentary.  '),(17673,1405024,'08,August,2004','2004-08-08 00:00:00','   urlLink \"He was walking around buck naked, masturbating and blowing kisses.\"  '),(17674,1405024,'06,August,2004','2004-08-06 00:00:00','   urlLink HoustonChronicle.com - Internet addicts excused from Finnish military  man, the Houston Chronicle has some weird stories, eh?  '),(17675,1405024,'06,August,2004','2004-08-06 00:00:00','   urlLink \"They look like the flatworms from the movie \'Evolution,\'\" said Casey Matsuo, 15. \"They mutate into gorillas and took over the world.\"   eew, gross  '),(17676,1405024,'06,August,2004','2004-08-06 00:00:00','   wait, what??   urlLink HoustonChronicle.com - Prisoner glues hand to his girlfriend  '),(17677,1405024,'06,August,2004','2004-08-06 00:00:00','   urlLink \"Insecure that his tireless campaigning on behalf of John Kerry isn\'t properly demonstrating the strength of his loyalty, Ben Affleck resorts to eating a baby as proof of his allegiance to the Democratic candidate for President.\"   whoa dude.  Definitely check out that picture though.  '),(17678,1405024,'06,August,2004','2004-08-06 00:00:00','   urlLink Cokarexic!   You know I totally have (and have already read) this issue of  Star ...  '),(17679,1405024,'04,August,2004','2004-08-04 00:00:00','   urlLink \"James was convicted in 1993 of assaulting two women. The first case occurred in 1991, when prosecutors said James and his girlfriend tied a woman to a chair, burned her with a hot crack pipe and forced her to perform sex acts during a cocaine binge at his West Hollywood home. He was free on bail when the second assault occurred in 1992 in James\' hotel room. \"  '),(17680,1405024,'04,August,2004','2004-08-04 00:00:00','   What a long day.  Had to fire someone today - never done that before.  It was interesting.  All morning while I was getting ready for work I was really nervous about it and I felt really bad; then, as soon as I got there, I remembered how mad and frustrated I was in the first place and felt a lot better about the decision.  Blah.  '),(17681,1405024,'02,August,2004','2004-08-02 00:00:00','   Somehow I just woke up.  It\'s after 6pm!  Man.  I had so much to do today too, like clean my room and go get Starbucks.  No, actually, I really wanted to be productive...I just worked the last five days in a row, and I have to work the next five days in a row, so I really just wanted some time to clean shit up and put shit away.  Blah.  I forgot to buy Beastie Boys tickets yesterday when they went on sale - now people on Craigslist are selling them for like $100/ea, when they originally cost $37.50.  Fuck.  I\'ll have to rely on my good luck and try to win them from the radio.  Okay, I\'m gonna hop in the shower...  '),(17682,1794926,'28,August,2003','2003-08-28 00:00:00','   My boss threw a nice party for me last night to celebrate my arrival in the Czech Republic.  Lord, the Czech people know how to drink!  I warn any tee-totalers out there that \"I don\'t drink\" is not understood here.  Drinking is a way of being friendly, and I want to be friendly, so I ended up drinking three glasses of wine and some vodka.  What a strange feeling--like wanting to sleep, laugh, and have sex at the same time--which is physically impossible, I think.  I was out until 3 in the morning and I had to work at 8.  Ahhh, I\'m amazed I made it out of bed....   '),(17683,1794926,'27,August,2003','2003-08-27 00:00:00','   Hey everyone,  It looks like my internet access won\'t be that great afterall.  I\'m sharing a computer with one of the employees at Hopr, the company where I work.  Every time I ask to check email, etc., I feel like I\'m imposing.  So my posts may be few and far between and lack the profundity you\'ve all come to expect by knowing me in person.  Just kidding.  Coming to the Czech Republic has clearly gone to my head :o)  Anyway, my thought for the day is that I\'ve aged half a day by coming here.  That\'s how I look at it anyway.  My watch has a digital part and an anolog part.  I keep the digital on Salt Lake time and the anolog on Czech time.  Right now it\'s 10 o\'clock in CZ and 2 in the morning in SLC.  So I\'m 8 hours older than I would be if I were in SLC.  Weird....  I had a great weekend.  I visited a medival castle here in Zabreh.  It was castle-like, everything a castle should be, motes, big halls, suits of armour, the whole she-bang. Also, the little girl I\'m teaching English to had her 5th birthday.  Let me tell you, it ain\'t easy teaching a child who doesn\'t speak your language--lots of signing and saying \"Linda, come, chocolate.  Linda sit, chocolate.\"  My one means of getting control is chocolate bribery.  Along that line, I\'m soliciting babysitting/teaching advice at suspendeddisbelief@hotmail.com I haven\'t yet added a comments section--still working on it.  Anyhow, I\'ll sign off now.  I hope you\'re all happy and healthy.  I miss everyone.   '),(17684,1794926,'21,August,2003','2003-08-21 00:00:00','   I\'ve arrived in the Czech Republic safe and sound after about 13 hours of flight time.  The whole process took about 26 hours all together.  I\'m a little rushed, but I\'ll make a few notes of my prgress:    1. No signs of jet lag so far.  I\'m keeping my fingers crossed that my body clock will sort itself out with minimal discomfort to me.  2. My boss is a nice, sophisticated lady who speaks English quite well.  We drove 3 hours from Prague to Zabreh and listened to Cher and Jennifer Lopez the whole way, then we stopped at McDonald\'s for dinner.     God save ubiquitous American pop culture.  3. I tried Kofola, Czech cola.  Tastes like a mix between root beer and Coke.   4. I\'m still adjusting to people speaking Czech all around me.  Half the time I wonder if they\'re talking about me.  That\'s probably just paranoia, but you never know...  5. I\'ve had to give up my vow to not eat beef.  When choosing between good manners and humanitarian ideals I have to go with manners.  That\'s all for now.  I\'m trying to figure out how to install something for reader comments on my blog site.  Hopefully I\'ll have that up soon.                      Cheers,                        Kim   '),(17685,1794926,'16,August,2003','2003-08-16 00:00:00','   Hi all.  This is just a test.   '),(17686,1794926,'24,September,2003','2003-09-24 00:00:00','   Yippee!!  Kendra got the comments thing to work.  You\'re the best Kendra!  You\'re one smart gal.  I might have to take you to dinner twice.  Anyway, I\'m learning to drive my boss\'s huge Mercedes station wagon through the tiny cobblestone streets here.  I hate driving in the U.S.  and here it\'s worse.  These streets were made for horses,  not cars.  Anyway, I pray everytime I get in the car that I don\'t do something stupid.  This car has about 50 buttons and I think I go into shock at the sight of them.  In other news, I ran into some Anti-American sentiment last night.  I was at dinner with my boss and her business partners from Austria.  My boss\'s sister was being friendly and asked the Austrian to talk about the differences between Austria and the U.S.  He said, \"You can\'t even talk about any differences because America has no culture.  They have only McDonald\'s and Pizza Hut and all they care about is money.\"  I had just sat down at the table and he says this.  Needless to say, I was rather surprised at this bold rudeness, but I didn\'t say much back.  It would probably be a mistake to offend my boss\'s business partners.  However, this attack has made me think about America and my own feelings about my country.  America has many flaws, but I do love it.  More specifically, I love the concept of it.  America is a beautiful social experiment-a place where people can come from across the world and become anything they want through their own talents.  I know the reality is much different (racism, sexism, etc) but this is the concept of America.  My mother grew up in a Hell\'s Kitchen (NYC) hand laundry and graduated from college at 18 through her own efforts and talents.  This is the great American story.  I don\'t know of any other country where upward mobility is so valued (Canada maybe, I\'m not sure).  When I started college, I considered every career possible--nothing was closed to me.  That is a wonderful thing.  So there is my patriotic rant about America.  I think I\'ll send it to an essay contest or something.  We can go into the phoniness of the American Dream (i.e., Great Gatsby, fabulous book by the way), that it\'s all about money, but to me it is about freedom, about making yourself the person you want to be.  I will also say that Americans must become more concerned about conditions in the rest of the world.  We must care more about the suffering in other countries and less about developing bullet-piercing armour, etc.  I think if we cared more about other countries, we wouldn\'t need the bullets.  Take Care,  Kim  P.S.  I\'m on my way to Croatia.  No internet access, I believe.  You\'re on your own for a week, kids.   '),(17687,1794926,'22,September,2003','2003-09-22 00:00:00','   Ok.  I think the comments link doesn\'t work.  I\'m enlisting the help of a computer-savvy friend.  Thanks Kendra!  I\'ll take you to dinner anywhere you want to go.  :o)  We can even go for beef and potatos if that\'s what you want!  Check out Kendra\'s awesome blog at http://crylaughsnort.blogspot.com I started my own blog at Kendra\'s suggestion, so I must give credit where credit is due.  Nice quiet weekend here in the sunny Czech Republic.  I think it was about 80 degrees F, 30 degrees C.  I still haven\'t gotten the metric system down.  Which leads me to another small soapbox next to  my cute trains soapbox.  Why doesn\'t the U.S. go metric?  It\'s easier.  The rest of the world uses it.  Write to your senator today about the elegant mathematical beauty of the metric system.  Anyway, I\'m leaving for a week in Croatia.  I can\'t wait.  I wonder if women walk around topless there.  If they do, should I do it?  Hmmm.  Maybe I should do it once to say I\'ve done it.  I\'ll tell the grandkids about it.  \"I was young once too.  Back in the eary 2000s I went topless in Croatia.\"  I\'ll say this as I wave my cane at them.  Ok, maybe I won\'t tell them that....might scare them.  Take Care Everyone,  Kim   '),(17688,1794926,'18,September,2003','2003-09-18 00:00:00','   THINGS I MISS  1.  All of my friends and family (of course).  I live in a strange isolation here, always surrounded by people I can\'t talk to.  2.  Bookstores, libraries.  I am down to my last book and I\'ve only been here a month.  I fear I will soon suffer serious book withdrawals.  I never realized how lucky I was in the States.  If I ever wanted a book, I just walked down to the library and got it.  Every Czech town I visit I look for books in English.  No luck yet.  3.  Mountains.  The Czech Republic has many beautiful wooded hills, but they\'re not my mountains--the mountains across the Great Salt Lake (I think the Wasatch).  Whenever I see those mountains, I know I\'m home.  4.  Vietnamese, Indian, Tibetan, Chinese, and Japanese food.  Ohhh, for a good Masasman curry right now.  My mouth is watering.  I think it\'s potatoes and beef for lunch again today.  The first night I\'m back in Salt Lake, I\'m going to East West Connection, Cafe Trang, or House of Tibet.  Maybe all three, depends on how hungry I am. (In a side note, Czech chocolate is fabulous, so all is not lost).   '),(17689,1794926,'15,September,2003','2003-09-15 00:00:00','   This weekend I took the train to Olomouc.  It cost about 2 dollars and took 30 minutes.  The U.S. definately needs more trains a) because they\'re cute and b) because they reduce air pollution.  Now is the time to contact your senator demanding cute trains for the U.S!  Ok, I\'ll get off my soap box now.  In Olomou, I went shopping for shirts, but didn\'t buy anything.  Here\'s another travel tip:  KIM\'S TRAVEL TIP Try not to buy clothes or shoes in the Czech Republic because the selection is rather small and the prices aren\'t great.  I can\'t speak about Prague, but that was the situation I found in Olomouc.  Brno has a better selection with many U.S. stores like Benneton and Nike Town, but I\'d still try to bring what you need from the states.    KIM\'S TRAVEL TIP Also, stand warned that there is lyme disease in the Czech Republic.  The little girl I teach went on a field trip to the mountains and came back with a tick in her leg.  Her mother tried to get it out, but left a bit of it in her.  Pack some fine-tipped tweezers in case you find a tick on yourself.  I think there is a lyme disease vaccine, but I don\'t have it and I\'m not sure about its success rate, etc.  Take Care.  That\'s all for now.  Kim   '),(17690,1794926,'12,September,2003','2003-09-12 00:00:00','   Hi All,  I\'m making attempts to add a comments section to my blog.  I hope it works.   '),(17691,1794926,'09,September,2003','2003-09-09 00:00:00','   A new look for my blogger site: Change is good.  I was getting tired of that orange.  I spent Wednesday in Vienna and all I can say is \"Oh My God.\"  Vienna is absolutely fabulous.  I think I could live there for the rest of my life (ok, maybe not my entire llife, but at least five years).  Everyone I encountered in the city center spoke perfect English.  This creates a more touristy atmosphere, but is also so much easier than pointing to things in a phrase book.  For most of the day I was in the Albertine museum looking at the Albrect Duer exhibit.  He\'s famous for the fuzzy rabbit painting.  It looks like he painted every single hair on the hare.  Ha Ha.  Hair/hare.  I know that was bad, but I couldn\'t help it.  Anyway here\'s a travel tip--  KIM\'S TRAVEL TIP: If you are traveling between the Czech Republic and Austria, buy everything in the Czech Republic.  As soon as you cross into Austria, the prices are three or four times higher.  I wrote a few days ago that musems in CZ cost about a dollar.  I paid nine dollars to get into the Albertine.  The price difference is due to Austria\'s membership in the EU.  Once CZ gets into the EU, prices will go UP, so come to CZ ASAP.  OK?  Sorry, I\'m in a weird mood.  I went out drinking last night, and I think I\'m still feeling it.  Oh, here\'s another travel tip.  KIM\'S TRAVEL TIP:  There is a dish in CZ called \"special meat.\"  This \"special meat\" is an expensive delicacy.  It\'s ground meat served on fried bread and has bits of onion and spices.  I just found out the \"special meat\" is raw beef from the cow\'s spine and a raw egg.  I am now totally paranoid that I have mad cow disease.  This meat is delicious, but don\'t have any if you don\'t want to eat raw food. Also, if anyone in CZ tells you \"we will drink together some small spirits,\" that means vodka or whiskey.  Basically, you have to drink it, but stand warned.  Finally, I\'ve been thinking about why U.S. cities are so ugly compared to European cities.  America is such a rich country, you\'d think we could do better.  My boss said that Americans are uncultured and don\'t care about beauty.  I think that\'s certainly possible.  Take Care Everyone,  Kim   '),(17692,1794926,'09,September,2003','2003-09-09 00:00:00','   A few interesting cultural differences:  *Entrance to museums and other historical sites in the Czech Republic costs about a dollar.  The average price is 25 crowns (29 crowns is a dollar).  I wish New York museums would take up this custom.  *On the other hand, good CDs of American music cost about 30 dollars.  *You also have to pay 5 crowns to use public bathrooms, shopping carts, and to get bags at grocery stores.  I think paying for bags is a great idea (a good way to reduce waste), but it\'s a bit irritating to pay when you just want to wash your hands at a mall.  *Gas stations provide free plastic gloves to use at the pump.  Interesting....  Also, my boss runs her car on diesel.  *Most shops are only open from 8 to noon on Saturday and are closed Sunday.   '),(17693,1794926,'08,September,2003','2003-09-08 00:00:00','   Hey all,        I had a busy, fun week of sightseeing.  Czech cities are so beautiful, I don\'t know how I\'ll cope when I go back to the States.  On Wednesday, I went to Olomouc, a university town of 100,000.  It\'s filled with cobblestone streets, fountains, and gothic cathedrals.  I toured an old bishops\' palace and had so Indian food, a nice change from dumplings and goulosh.  On Saturday, I hit Brno, the Czech Republic\'s second largest city.  It was remarkably clean.  I didn\'t see any litter on the ground anywhere.  My most interesting tour in Brno was of an old church crypt.  For about 50 cents, you can see thirty mummies from the 1700s in glass coffins.  Creepy, but good for meditation on the transience of life.  The worst part was that one of the mummies, an aristocratic lady, had been buried alive.  Her head was totally bent to the side and her hands were clenched.  What a nighmare.  I wonder if she thought she\'d died and gone to hell.  Anyway, I won\'t go on.  It\'s too early in the week to be morbid.  This week I\'m going to Vienna.  I\'ll keep you posted on the cool spots I find there.  Take care, everyone.   '),(17694,1794926,'01,September,2003','2003-09-01 00:00:00','   Well, I had a quiet weekend.  A little chaos this morning--I normally take a bus to work at 7:39.  Well I arrived today and found no bus and no people.  I was a bit frantic because I have no other way to get to work and I have to teach a class at 9:00.  Finally, a few other people showed up at the bus stop and stood around expectantly.  I debated trying to talk to them in English and then decided to wait a little while and see what happens.  What do you know, a bus came at 8:00, which made no sense to me because I was told the buses come at 7:39 and again at 9:00.  I got on this bus and was afraid it would take me to some faraway city because no one had told me about it--like maybe this is the bus to Brno or something.  Lucily it went right to where I work.  I asked the secretary about the bus situation, and she said all the bus times change in September.  Doh!!  Just one of the little inconveniences you face when you don\'t know the language, and therefore can\'t read the bus schedule for yourself.  Well enough of my bitching.  Everything is fine otherwise.  I\'ve walked a lot through Zabreh and am getting the lay of the land.  My boss is taking me on a day trip to a near-by city on Wednesday.  I\'m so happy I get to travel with her on business trips.  Hopefully I\'ll hit Prague and Vienna soon.  Take care everybody.    :0)  ;)   '),(17695,1794926,'31,October,2003','2003-10-31 00:00:00','   I spent a nice four days in Prague seeing the sites.  I stayed in Vinorhady, a part of town near Wenciles Square (terrible spelling).  An old man and his wife rented me a room in their home with my own bathroom and hot plate.  It was a nice setup and only five minutes from the center of town.  The toilet was one of those old-fashioned ones where you pull a chain attached to a box above you.  I think this is a superior toilet design because there is no water in the toilet bowl until you pull the chain--very environmentally conscious.  The other old-fashioned thing I love in CZ is that bookstores wrap your books in tissue paper, so you have this little parcel and you can open it up at home like a present. The highlight of my trip was walking through the gardens by Prague Castle with all of Prague spread below me.  It was like a dream--beautiful golden leaves on the trees and all the spires of the city rising up into the sun.  Tuesday was Czech Independence day, and the castle\'s main courtyard was filled with black BMWs from diplomats.  Each car had a driver waiting in the front and a little flag on the antenna showing what country they came from.  I wonder why diplomats always ride around in black cars, why not blue or red?  Maybe black is more dignified. There were two funny mishaps on my trip.  The first was that I forgot about daylight savings time, so I got on the wrong train coming back to Zabreh.  I ended up in Olomouc, 50 kilometers away and had to backtrack.  At least it only added one hour to my travel time. The other mishap happened at a liquor store.  I had gone in to buy a little bottle of absinth and have an encounter with the green fairy.  I think everyone has to drink absinth in Prague because it is the capital of Bohemia.  Anyway, I got up to the counter and asked for absinth and the lady started yelling at me in Czech.  I said \"what?\"  and she said \"you--alcohol--no.\"  I yelled back, \"I\'m almost 23-years-old for Christ\'s sake,\" or something like that.  \"No alcohol,\" she said again.  Then I pulled out my passport and proved that I was 22.  She became much more friendly and said something like \"what do you want.\"  I said, \"I\'ll buy it somewhere else.\"  By this time I was righteously pissed and walked straight out of the store.  I wasn\'t going to give money to someone who had yelled at me that way.  She didn\'t even just ask me for ID.  Anyway, it retrospect it\'s quite funny.  The legal age to buy alcohol here is 18.  She must have thought I\'m 12 or something.  That\'s the funny thing about my looks.  Some people have thought that I\'m 30 and others think I\'m 15.  Strange, huh?  I guess I have the kind of face that easily morphs from one age to another.  In the end I bought the absinth somewhere else.  I haven\'t tried any yet.  Maybe I\'ll  have a sip on my birthday. In other news, I have moved into a hotel.  I have no TV or telephone, but other than that it\'s fine.  At least it\'s private and it\'s only for 6 weeks. I\'ve also been thinking about whether I would ever settle abroad permanently.  I love Prague and feel like I could spend years there, but I think that if you can\'t speak the language, you\'re always a tourist.  You may be a tourist for a long, long time, but you\'re still a tourist.  Czech is a very hard language and I don\'t feel motivated to tackle it.  Also, I\'ve already spent almost 100 dollars on calls to the US and I\'ve only been here 2 months!  I think if I lived here for a long time I would go broke on international calls.   '),(17696,1794926,'23,October,2003','2003-10-23 00:00:00','   Hi all,       This is just a quick note to say that my email has been down for the past week.  Please direct future emails to my hotmail account.            Thank You,              Kim   '),(17697,1794926,'22,October,2003','2003-10-22 00:00:00','   I am pleased I created the impression that I am a hip nonchalant world traveler (at least according to Kendra).  I am really not hip or nonchalant and I\'m barely a world traveler.  In reality, I eat cereal with my hands, have difficulty opening strange doors, and never read Kerouac until now.  I don\'t even know what a meta-narrative is.  Still, thanks for the nice thoughts, Kendra.  Also, here\'s one of the funniest things I\'ve read in a while.  It is from a book called the Headway English course for Intermediate Students.  \"In America you should eat your hamburger with both hands and as quickly as possible.  You shouldn\'t try to have a conversation until it is eaten.\"  So remember that the next time you\'re at McDonald\'s.    I\'ve decided to start a movie review to go along with my book review of the tasty Pancake book (I agree it\'s a good name).  This week, I\'d like to discuss my favorite Woody Allen movie, Broadway Danny Rose.  This movie beats out Annie Hall by a hair, mainly because the ending is fairly happy.  It also has Mia Farrow wearing some cool 70s sunglasses.  That is enough to make any movie great, I say.  There is a funny scene where Mia and Woody are tied up by thugs, and they have to wiggle against each other to get out of the ropes.  Ha ha ha!  Go rent it today.  Here\'s my favorite Woody Allen quote from the moive Zelig, \"Zelig\'s father told him that the world was a meaningless nightmare of suffering.  His only advice was to collect string.\"   '),(17698,1794926,'20,October,2003','2003-10-20 00:00:00','   Another interesting fact about the Czech Republic.  Ahoj means hello (casual) and chow means good-bye.  All day I feel like a pirate \"Ahoj thar matey!\" crossed with a Hollywood agent \"Chow babes.\"  I do think ahoj and chow are more fun than hello and good-bye.   '),(17699,1794926,'17,October,2003','2003-10-17 00:00:00','   So it looks like my boss is moving to Olomouc and I will have to find new accomodations.  I am not sure when exactly I\'ll move and where I will end up.  Right now I am trying to be very Zen about it and live in the present moment.  I will end up where I will end up.  My current digs are quite nice, so I am sad to see them go.  There is a big garden backing onto a field and an indoor swimming pool.  Don\'t feel jealous because I\'ll be leaving pretty soon.  Life is in constant flux, eh?   '),(17700,1794926,'17,October,2003','2003-10-17 00:00:00','   So I am planning my next trip to Prague, trying to find another cheap hostel with a simpler door.  It has been very cold these last few days.  This morning there was frost on the ground.  I put the down lining into my coat for the first time this year.  I saw online that it was been 75 in SLC.  Sounds lovely. I have also developed an addiction to DVDs.  They are my only source of entertainment, and I would say I watch about three per week.  Selection is a bit limited, so I have watched things I wouldn\'t have considered in the US, such as the whole Mummy trilogy.  I quite liked the first Mummy movie.  It was campy in all the right ways with man-eating scarbs, etc.  The other two weren\'t as impressive--I can\'t really remember what happened in them. Also, I spoke to my sister, Lisa, (heyy Lisa!) last night and heard that Arnold Schwartzenager (spelling?) won the election in California.  I find this rather amusing.  I wonder if he\'ll run for president.  I am not sure he qualifies because he was born in Austria, but he would liven things up in the White House.  I don\'t think terrorists would mess with us anymore if President Arnold got on TV and said \"We\'re going to kick some fundamentalist ass!\"   '),(17701,1794926,'16,October,2003','2003-10-16 00:00:00','   Hey All,  I spent yesterday in Vienna just walking through the city center.  I think I like Prague and Vienna equally.  They have different vibes.  Vienna is bigger and very flat.  All the old buildings are white with columns and statues.  The whole area is more spread out, so you don\'t get the wall of tourists that you find in Prague.  In Prague, many buildings are light yellow or light orange.  The river runs right past the center and there is beautiful hill where the castle sits.  The view is amazing.  All in all, I would say Vienna feels more cosmopolitan, glittery and new.  Prague is more romantic and atmospheric.  I am glad I live right between two fabulous cities.   '),(17702,1794926,'14,October,2003','2003-10-14 00:00:00','   It has occured to me that trying to teach someone English is a great way to rediscover the language yourself.  People are always asking me questions like \"what does this mean,  \'raise money.\'  Raise means children growing up, right?\"  I am starting to see all these weird connections between words.  For example, lighning strikes, you strike out on your own, go on strike, see a strikingly beautiful woman, etc., etc.  Strike is a good, solid Anglo-Saxon word.  It seems to imply a jolt in many different situations.  Going out on your own is a jolt, as is seeing a beautiul person.  Interesting....  Also, I visited the Globe Bookstore in Prague.  This is a hipster, ex-pat hangout in central Prague.  I feel very hip using the word ex-pat in a sentence.  Anyway, new additions to my collection are Kerouac\'s Dharma Bums, Carson McCuller\'s Ballad of the Sad Cafe, and Marquez\'s Love in the Time of Cholera.  It says something that I spent over 30 dollars on books and 22 dollars on my hotel room.  Who needs a shower when you have a book?  Incidentally, I didn\'t shower for the three days I was in Prague.  I just trusted my cleaniness to Secret Deoderant.  It\'s strong enough for a man, but ph balanced for a women, so I didn\'t worry.  These are the times when I am glad my hair is only 2 inches long.   '),(17703,1794926,'13,October,2003','2003-10-13 00:00:00','   Ahh, Praha.  Fantastic, brilliant, wonderful weekend in Prague.  I am completely in love with this city.  The entire place is a work of art.  I walked around Prague castle, across the Charles Bridge, through Old Town Square and Wencilies Square.  I also visited the national musuem.  All of these areas were solid wall to wall tourists, even though it\'s low season.  It is hard to believe that Prague was ever a Communist area.  It is about as capitalist now as any place could be.  I even had that common American experience of seeing a KFC, then walking one block and seeing another KFC.  Reminds me of that line in Best of Show, \"We were both at Starbucks, but not the same Starbucks.  I was at one Starbucks across the street from her Starbucks.\"  Or something like that.  In connection to that, I think I ate in a restaurant alone for the first time.  I went to Pizza Hut and ate by myself.  I was not too bad, but I would have liked some company.  I think I was the only person alone in that place.  I seem to end up eating fast food whenever I go out alone.  I really don\'t like fast food, but Pizza Hut always feels familiar and I can understand the menu.  I will try to become more daring on my next Prague visit.  I hope to go back in a few weeks. I also had my first ever stay in a hostel.  I am traveling on the cheap.  I should say CHEAP.  My room was 11 dollars a night.  The place was fine--quiet, clean, etc.  My one major problem was that I never figured out how to unlock the door.  It was a very strange door with a doorknob that didn\'t turn.  Unlocking it involved some complex system of turning the key twice and pushing.  I never got it.  I am the kind of person who isn\'t sure what end of the screwdriver to use.  I think the mechanical part of my brain was dismantaled and absorbed by another part.  Anyway, it appears to be gone.  It is kind of sad, but I think I will stay somewhere else next time just because I was annoyed by the door. ;) Finally, never ever exchange money at one of the touristy Chequepoint exchange places.  I gave them 60 dollars I got the equivalent of 45 dollars back.  When I realized this, I was absolutely livid.  Highway robbery!  The bill said something about printing costs for money or some other b.s. like that.  But at least I didn\'t get pickpocketed. I would also like to thank Clint and Kendra for their witty, insightful comments.  Sorry I never get time to respond directly, but they are much appreciated.  Makes me feel like my words aren\'t falling into a void or something.   '),(17704,1794926,'09,October,2003','2003-10-09 00:00:00','   Great news!  I found a bookstore in Olomouc with one bookcase of English titles.  Olomouc is about an hour by car and I go there fairly often.  I bought two books--Truman Capote\'s In Cold Blood and Zadie Smith\'s White Teeth.  I have already finished In Cold Blood.  Fabulous book!  I had never read any Capote before, but I will definately seek out more of his stuff.  Because I am in such a bookish mood, I would like to take this opportunity to start a small book review.  So everyone gather round, pull up a chair and a plate of goulosh.  My only book review rule is that I will only review books I like.  Nowadays, book reviews have become savage excuses to shred authors and books.  \"Let\'s see some blood on the page!\"  Personally, I would rather just not hear about books the reviewer doesn\'t like than witness some poor writer torn to pieces.  Maybe because I identify with the writer.  Anyway, without further ado:  Kim\'s Cozy Book Review Corner  This week I\'d like to examine my new favorite writer, Breece D\'J Pancake.  Mr. Pancake only wrote one collection of short stories before killing himself at age 26.  The resulting collection, The Stories of Breece D\'J Pancake, is a stunning look at life in rural West Virginia.  The characters are almost invariably angst-filled young men working blue collar jobs.  Pancake renders these blue collar characters without sentimentalizing them or insulting them.  I found the characters poinent and sympathetic, mainly because they all long for better things in their life but can\'t seem to escape fate and West Virginia.  Pancake\'s writing style is spare and subtle.  Landscapes seem to mirror emotions and foreshadow later events.  Altogether an amazing performance by such a young writer.  It\'s a shame Pancake is no longer here to produce more beautiful writing.    Also in good news, I got one of my poems published!  My poem \"Extension and Withdrawal Parts I Through IV\" will appear in The Rectangle, the journal of the interrnational English Dept. Honor Society.  It is nice to a few acceptances to offset the 50 or so rejections I have gotten in my writing career.  Ah well, the road to the publishing house is paved with rejections.   '),(17705,1794926,'07,October,2003','2003-10-07 00:00:00','   For all of you stock exchange fans, Oct 7 is the anniversary of the introduction of the Dow Jones Industrial Average in 1896.  Also, this week is National Chili Week, Canadian Customer, Service Week, Math Literacy Week, Nuclear Medicine Week, Teller Appreciation Week, and German-American Day.  I feel rather festive.  Who knew there was so much drama in the second week of October?  Also, what is Teller Appreciation Week?  Is it for bank tellers?  I think they do deserve their own week because they take a lot of guff from annoyed people in banks.  Horray for Teller Appreciation Week! I got this fascinating information from the Publisher\'s Weekly newsletter.  Every week, PW lists holidays, etc., to give booksellers display ideas.  Now I want to do a display for Teller Appreciation Week.  Maybe have pens on chains dangling from the  ceiling for decoration.   '),(17706,1794926,'06,October,2003','2003-10-06 00:00:00','   The rains have set in here.  I am preparing for two months of cloudy days.  Luckily I rather like rain.  From inside, it is kind of cozy.  I will let you know if I still feel that way in two months.  I am going on a trip to Prague this weekend.  It is a three hour train ride from here.  I am going to stay in a hostel and pack in as many sites as possible.  Hopefully, I will get to go there several more times before I leave CZ.  I have been thinking about the names of cities and countires.  All the countries in Europe have different names in Czech than in English.  For example, Croatia is Crovatsko.  Also, Prague is Praha and Vienna is Wien.  I have wondered why we don\'t all a country by the same name.  I would imagine that everyone should call a country be the name they call themselves. Strange and interesting.  If my posts seem dashed off it is because my computer access has declined.  One secretary left the company where I work, so the computer I normally use has disappeared.  A little frustrating, but what can you do?   '),(17707,1794926,'06,October,2003','2003-10-06 00:00:00','   Hi All,        Croatia was great.  A little cold for swimming, so the toplessness issue never came up.   The highlight of the trip was  a boat trip to Krka waterfall.  I also think I\'ve gotten the hang of driving my boss\'s car.  I drove several times in Croatia, and everything worked smoothly.  An interesting feature of Croatia\'s streets is that many only fit one car, but people drive both directions on them.  Basically you just try to watch for any oncoming cars and then pull over into the bushes to let the other person by.  This does reduce asphalt by half but I wouldn\'t recommend it for the States.   '),(17708,1794926,'28,November,2003','2003-11-28 00:00:00','   First, I\'d like to thank Jade for commenting all over my blog.  I don\'t know how to respond to comments, but I appreciate them and always read them.  I spent Thanksgiving dinner at a party thrown by the 7th Day Advantist Church.  It was pretty fun.  I particularly enjoyed the rocking out hymns they performed with the pastor on electric guitar.  The dinner menu was a bit unusual, but a good effort.  There were several plates of cut-up turkey and, curiously, a huge number of chopped tomatos and pickles.  I did miss stuffing.  Stuffing from the turkey is the best part of Thanksgiving, I think.  So I hope all of you out there are enjoying some stuffing this holiday season. Oh, it was also a little funny when a girl at the party asked me, \"Do you really eat pumpkin pie?\"  It is a perfectly normal question, but it just struck me as the kind of thing we ask about other cultures, like \"do they really eat snakes?\"  And Clint, there is no Tofuky here.  I don\'t even think there is tofu.  CZ is the land of beef and potatoes.  And beer.  Lots and lots of beer.   '),(17709,1794926,'27,November,2003','2003-11-27 00:00:00','   Happy Thanksgiving!  I hope everyone is with family and friends enjoying a huge, long, tasty meal.  I remember past Thanksgivings sitting around all morning with the smell of turkey cooking.  I remember the waiting and the smell even more than the actual eating, (though the eating was good too!).   '),(17710,1794926,'26,November,2003','2003-11-26 00:00:00','   Egads!  There is so much to report since my list post, I\'ll have to resort to the bullet list:  1.  Turned 23 on Monday!  Yea!  Happy Birthday to me!  I feel old for the first time.  23 is like late early 20s, rather than early early 20s.  I\'m almost to early mid-20s.  Godd! ;)  2.  Went to Prague on Sunday and stayed at the most horrible hostel yet!  At 2 a.m. the upstairs bar closed and a hord of loud roudys came down.  You could smell the hormones on them.  They sat in the room outside yelling, screaming and discussing private features of their anatomy.  Privacy please!  On the plus side, I met a nice Aussie girl.  3.  I met up with my cousin Tim and moved into a 5-star hotel room he paid for as a Birthday present.  Famine to Feast.  An hour after I got there, a posh butler showed up with a bottle of champaign and called me Madam.  When I left, I took everything that wasn\'t nailed down--the shoe polish, vanity kit, shaving kit, dental kit, body wash, mouth wash, stationary.......  4.  Walked through Prague with Tim on the grand tour.  We made it to pretty much every site in 2 and a half hours.  Later stopped off for dinner.  Thank you so much, Tim!  I could never repay your generosity!  5.  Went out to the Roxy with Aussie girl and an American friend of hers.  Was offered ecstacy, but refused.  Talked to an older guy from Torronto about his travels in Asia and Europe.  Was amazed that American girl is funding 3 months of backpacking from unemployment checks.  6.  Got a star on the Hollywood Walk of Fame.  7.  Was told to get on the train to Brno at the train station, did so, and discovered 3 hours later that it doesn\'t stop in Zabreh.  Backtracked to Olomouc, then Zabreh.  The typical 3 hour trip ended up taking 6 hours.  Doh!  I did get a lot of reading done.  8.  Went to bed totally exhaused last night at 10.  9.  Please disregard item 6.  I was thinking of someone else.  Did you get all that?  Busy, busy, busy.  This was a wonderful birthday.  I just wish I could have seen more friends and family (though it was great catching up with Tim. Thank you again, Tim!).  I miss everybody, but I\'ll be home for Christmas.   '),(17711,1794926,'21,November,2003','2003-11-21 00:00:00','   I\'ve been thinking about all these concepts of not judging other cultures that I learned in sociology.  I frankly think it\'s impossible not to judge and I am not sure if it is healthy to not judge anything.  Shouldn\'t we have independent standards of right and wrong that apply to everyone? For example, yesterday I was teaching advanced students and we talked about the lack of women in management jobs.  One man in the class said that men should be managers because women create problems when they have these positions.  The reason is that \"harrassment of women is natural.\"  I just about jumped out of my seat.  Eventually I said that if men are harrassing women, the men should stop.  The solution isn\'t for no women to be managers, it is for the men to grow up.  There was some awkward silence after that.  I must say that the men who are really sexy are the ones who respect women.     In lighter news, I saw a hedgehog yesterday.  I was walking after dark and I saw something coming up the sidewalk.  Under the streetlights, it looked like a large rat.  I went up to investigate (being strangely drawn to rats since I did research on them in high school) and saw it was a hegehog going about its business, ignoring me.  I tried to feed it some bread, but it wasn\'t interested.  This is the first real life hegehog I have ever seen.  I am not sure we even have hedgehogs in the U.S.  I remember a Beatrix Potter book that I loved about a hedgehog.  It was called Mrs. Tiggie Winkle and the hedgehog was a little washer woman who lived in the mountains.  Kind of an odd book, but I loved it.  There is such ingroup/outgroup mentality around the world.   '),(17712,1794926,'20,November,2003','2003-11-20 00:00:00','   Personally, I love being bustled, if it is by the right people.  :) Well, I may go to Prague this weekend to visit my cousin, Tim, who is coming for a business meeting.  Kind of cool to hook up with my cousin in Prague.  It has been years since I\'ve seen him in the U.S.  Strange to come all this way and encounter family from back home.  I have to watch my pennies though, because all this travel is draining the finances.  I have finished my Christmas shopping, though.  I have very mixed feelings about the whole Christmas spending orgy.  This year I had planned to make donations to charity in the names of family and friends, but CZ has too much cool stuff to pass up. I remember the first time I encountered the idea of not buying gratuitous stuff for Christmas.  I was about 18 and had just started meditating.  A guy at the meditation group told everyone how he buys his friends groceries for their Birthdays as part of simple living.  I thought it was so cool and glamorous to buy soy milk and tofu for people\'s birthdays.  Perhaps I felt this because the guy himself was totally cool and glamorous.... but that was years ago.   '),(17713,1794926,'19,November,2003','2003-11-19 00:00:00','   Hi All,  I hope everyone had a good weekend.  Vienna was great.  Saturday night at a loud, crowded club; Sunday at the art museum where I spent about 7 hours and still didn\'t see everything; Monday at Schurnburn Palace walking around the garden feeding crackers to fat, red squirrles.  Perfect, huh? I stayed in a hostel dorm room with 7 other people.  All women and no snoring, but still a little hard to sleep.  I was out of bed every day by 6:30. If you plan to visit Vienna, remember that it is wicked expensive.  Cokes at the club were 5 Euros, which is maybe 7 dollars.  Also, the train ride from Brno to Vienna cost 16 Euros, but from Vienna to Brno it was 20 Euros.  I\'m still trying to figure that one out. I picked up a City Spy map with all the hipster hangouts.  The writing in this map is fabulous--like Kerouac in wonderland or something.  Listen to this description of the store Boots: \"Leather deviant\'s & all others are welcome to come by & invest into your fashion lines.  Big boots, little latex bussom halters etc......Bring your own beer!\"  Now you know where to find latex bussom halters in Vienna.  This is for hipster hangout Wirr: \"It s sort of the student ghetto\'s living room, full of \'Weltschmertz\' people with little eyeglasses & holes in their underpants getting stuck while sipping one \'Achterl\' red after the other & as all good students Wiener\'s know more about the bottom of a glass than a fly\'s ear...\"  I think there is a comma missing in there somewhere.   This is my favorite: \"Fine weather courtyard bustling with that which bustles or would be bustled.\"  Is that prose poetry or what?! You can find more at http://cityspy.info   '),(17714,1794926,'14,November,2003','2003-11-14 00:00:00','   I was on Amazon.com and I read an excerpt of the Dali Lama\'s book, The Art of Happiness.  This excerpt was from the introduction written by a pyschiatrist about meetings with the Dali Lama.  The Dali Lama told the  psychiatrist that happiness is the purpose of life, and at first the psychiatrist thought this was selfish.  Then he investigated further and decided that thinking of your own happiness is really good for the whole world.  He cited some research that people in good moods are more generous and friendly.  I spent some time thinking about this idea.  Is it really altruistic to live for your own happiness?  Are happy people always nice people?  Similarly, are all mean people not really happy?  It\'s a tangled issue.  In my own life, I feel that I became more happy when I stopped worrying so much about myself and thought more about others.  This is kind of a reverse process--be altruistic and become happy; rather than be happy and become altruistic.  I really don\'t know.  I would like to read the Dali Lama\'s book and hear his views on the matter.  I remember seeing the Dali Lama at the Huntsman Center and he seemed so laid back and friendly.  I remember he rocked back and forth while he listened to people\'s questions.  Very cool.   '),(17715,1794926,'12,November,2003','2003-11-12 00:00:00','   A foggy morning.  I am busy planning a weekend trip to Vienna.  I was checking out hostels online and one of them advertised \"free hot showers.\"  I love this.  It\'s like, \"not only can you sleep here, you can even shower.  And in hot water, too.\"  You know you\'re poor when you stay at a place that advertises showers.  Ha!  Well, I had no illusions that I was poor.  Or maybe I should say frugal.  Yes, frugal.   '),(17716,1794926,'10,November,2003','2003-11-10 00:00:00','   Czech vs. English spelling  English--telephone Czech--telefon  English--Pizzeria Czech--Pizzerie  English--television Czech--televize  English--vegetarian Czech--vegetariansky  English--homosexual Czech--homosexualni  English--injection Czech--injekce  I am sitting at my desk next to my cup of soup and two big boxes of empty wine bottles.  I haven\'t asked about them.  A misty musty morning.  I\'m wearing a black T-shirt and jeans because all my other clothes are lying damp on my hotel room bed.  This was the big weekend for washing, with me hauling a garbage bag full of dirty clothes up the hill to Mr. Hosek\'s house to use his washing machine.  All Czech washing machines are small white boxes kept in the bathroom.  There are no Czech dryers.  I\'ve noticed that clothes dry at mysteriously different rates.  For example, my black socks seem to dry much faster than my white socks.  There must be some interesting physics behind this.   I have also become addicted to cable television.  My favorite stations are MTV Europe, MTV2, MTV Hits, MTV Base, VH1, and VH1 Classics.  VH1 Classics is especially good.  I think Cher had some fabulous hair back in the 70s.   '),(17717,1794926,'07,November,2003','2003-11-07 00:00:00','   Hotel Haiku  Threadbare blanket A bottle of milk  on the windowsilll   '),(17718,1794926,'05,November,2003','2003-11-05 00:00:00','   I am trying to help a nice Czech lady find a job in the US as an au pair, and I\'ve realized how hard it is to get into America.  Basically, from all the research my Mom has sent me over email, it looks like it is nearly impossible for her to come to the States unless she has family or can perform a job no other American can do.  It is nice that most any college grad from the US can go anywhere in the world to teach English.  I\'ve looked online at other places I could go and it seems like anything is possible.  I don\'t know that I would live abroad again, but it\'s nice to know there is the option.  I would also like to wish everyone a belated Happy Halloween.  Halloween is barely noticed here, which is sad because it is my favorite holiday.  I spent Halloween playing a form of European bowling called skittles.  I wonder if this game was named after the candy or the candy was named after the game.  Anyway, in Skittles the ball is smaller and has no holes.  You roll it underhand straight down the track.  There are nine pins set up in a diamond shape and it is almost impossible to knock them all down.  Playing this game is quite hypnotic.  You get into a flow of rolling the ball down the track, watching the pins fall and picking up another ball. I also went for pizza afterward with a few people from work.  One of them was an old man who got totally drunk and wanted me to come to a brothel with him.  This guy has probably been married for 40 years and has four grandchildren.  Rather distasteful.  I have no idea what he expected me to do at a brothel.  That is one of the more surprising aspects of CZ.  It is a lot like Vegas--porn and brothels everywhere.  Utah\'s porn czar would have a ball here.  You can walk into any gas station and find a stack for porn mags right above Vogue and GQ just hanging out there for everyone to see.  No one thinks a thing of it.  I personally find opposition to porn a strange thing.  I think porn is totally harmless as long as it\'s not violent, etc.  Brothels are more complicated.  Maybe it would be better for the US to legalize prostitution and regulate it for the safety of everyone involved.  I really don\'t know.  It does seem strange that the US is so progressive about technology, film, music, yet so uptight about sex compared to a lot of the world.  Must be the remains of Puritanism.   '),(17719,1794926,'03,November,2003','2003-11-03 00:00:00','   A beautiful fall day.  Crunchy yellow leaves.  A few high clouds.  No sign of snow or rain.  A view of pine-covered hills through the window.  A good day for a long walk.  Or a game of tag.  Or a blog entry composed of sentence fragments.   '),(17720,1794926,'03,November,2003','2003-11-03 00:00:00','   I am fascinated by the little ad box at the top of my blog.  Today I believe it says something like \"Buy Wormwood Seeds\" and \"Hotels in Prague.\"  Last week it had something about \"Write a Memoir in 28 Days.\"  I keep wondering how it gets ads so closely related to my blog entires.  It must do a word search or something.  Aren\'t computers amazing!  The new Gods of the the 21st century.  For those of you who would like a precise definition of absinth, here is the entry from dictionary.com  1. A perennial aromatic European herb (Artemisia absinthium), naturalized in eastern North America and having pinnatifid, silvery silky leaves and numerous nodding flower heads. Also called common wormwood.   2. A green liqueur having a bitter anise or licorice flavor and a high alcohol content, prepared from absinthe and other herbs, and now prohibited in many countries because of its toxicity.  There you have it!  Absinth is toxic green alcohol that was consumed by the Bohemians during the great Bohemian Revolution at the turn of the century.  The green fairy is the name many people gave to Absinth, I think because you were supposed to see the green fairy after you drank Absinth.  This is something like the pink elephants Dumbo sees.  I would suggest consulting the fabulous Baz Luhurman movie Moulon Rouge for an excellent depiction of the green fairy by Aussie pop star Kylie Minogue.  Lisa, I hope this answers your question.  I purchased some of this toxic alcohol in Prague, and we can drink it at Christmas (if they let me take it back to the U.S., that is).  In book news, I finished Kerouac\'s Dharma Bums and loved it.  Jack seems like an old pal; his writing is so friendly and conversational.  I love his enthusiasm for nature and eating pea soup.  He totally lacks cynicism; I don\'t think he could have survived the 90s.  Anyway, I don\'t think many people could read this book and not want to climb a mountain and make pudding in the snow.  Well, I didn\'t want to climb a mountain because I am afraid of heights, but I was on board for the pudding in the snow.  On a side note, I am also afraid of wasps, toilet seats, and new people (old people sometimes as well).   '),(17721,1794926,'30,December,2003','2003-12-30 00:00:00','   Well, I have officially changed the description of my blog.  That means the European experience is really over.  I think it was good to have the time to think about what I really want to do with myself.  Now I feel pretty fired up about getting job, writing, exercising, eating better, etc.  Europe also made me appreciate all the nice things I have here.  Like I can walk down to the library and find pretty much any book I want.  I still remember the book starvation in the Czech Republic.  For a while there, I was praying for words in English, anything, a VCR instruction manual, anything, just let it be in English, please!  But that is behind me now.  I also appreciate hotel rooms.  When I stayed in Vegas with my parents, we went into the room and said to myself \"wow, what a big bed.  And there aren\'t any strange people here talking or throwing things.  Ahhh.\"  Now any hotel room that I don\'t have to share with strangers is pretty luxurious.  It\'s good to appreciate the small things.   '),(17722,1794926,'29,December,2003','2003-12-29 00:00:00','   Yesterday, my family had our annual Christmas dinner at the Parvazs\'.  Jasmine Parvaz has been Lisa\'s best friend since junior high, and we\'ve been doing this dinner for about 6 years I think.  I had some unexpected good news when Jas\'s little sister, Ally said \"Oh, I saw a poem of yours published.\"  I think I said, \"What?\"  She had picked up Listen magazine and seen one of my poems in there a few months back.  That was pretty surprising.  I sent them three poems last spring and didn\'t hear from them, so I just figured they weren\'t into my stuff.  I wonder now if they tried to email me in the Czech Republic.  I was having some computer problems, so it\'s possible that I just didn\'t receive it.  Anyway, Ally had saved a copy of the magazine and gave it to me.  It felt great to see something of mine in there.  I was having some downer feelings about my writing last week, like \"oh, I\'ll never get anywhere,\" so this was a nice pick-me-up. In other news, I am thinking of attending the Unitarian Church.  My life has been missing spirituality for a while now and I feel I need to get back to some sacred space.  I\'ve been curious about the Unitarians for a while now.  It seems fine to be a Unitarian Buddhist.  Anyway, I don\'t like to label myself any particular religion, so maybe the Unitarians would work for me.   '),(17723,1794926,'28,December,2003','2003-12-28 00:00:00','   Snow.  That\'s all I can say.  Snow everywhere.  There is probably about two and a half feet of snow in our driveway.  And all that snow has to be painstakingly moved one shovelful at a time.  I think the best way to approach it is as a free workout.  I was considering joining Bally Total Fitness last week and they wanted $80 to start and $30 a month after that.  Why do that when you can shovel snow for free?  And you get the benefits of fresh air.  ;) I probably will start taking ballet again at least, and maybe yoga too.  I was a complete blob in the Czech Republic.  It\'s so easy to lay in bed eating goulosh and watching German MTV.  I think my heath will improve a lot now that I don\'t have access to cable.  I also want to get out on my snowshoes in the next few weeks.  I remember I saw a moose the last time I went up Big Cottonwood.  It was already dark and the moose walked up the trail and lay down right across it.  We had to crash down through the bushes to try and get around it.  I still remember the way it smelled.  We probably smelled weird to it too.  Jade, I wish you could come out to Utah and go snowshoeing again.  You are welcome anytime!  Also, anyone else who want to go snowshoeing, just give me a call. Well, this was my first Christmas away from Salt Lake in my entire life.  That\'s a landmark.  I spent Christmas Eve in Phoenix with Lisa\'s boyfriend and his family having a big dinner.  The lack of snow was a little odd, but I enjoyed not having to shovel.  On Christmas day, we were in Vegas.  The place was packed wall to wall people, every slot machine taken, and a line for the buffet that went around a corner.  I love the tinkling sound slot machines make.  You know there is an engineer somewhere who spent weeks designing that sound.  It rained all the time in Vegas, and there was two feet of water running down the street in front of the Imperial Palace.  I also heard on the news that terrorists wanted to hijack a plane in France and fly it into a Vegas casino.  I guess you never know what can happen these days.  Well, Happy New Year everyone!  May the new year find you happy and healthy and ready for the next adventure.   '),(17724,1794926,'21,December,2003','2003-12-21 00:00:00','   Merry Christmas!  I hope everyone is enjoying the holidays with family and friends.  In a half hour, I am starting on a trip to Phoenix to visit my sister, Lisa (hey Lisa!).  This will be my first Christmas away from home.  Feels a little strange, but I\'m sure the weather down there will be great.  I am ready to soak up some sun after the Czech dark days.  When I return, it\'s back to the job search and preparations for moving.  I may look around in Phoenix a bit and see if there are any prospects.  I am feeling the combination of excitement and fear knowing that the future is wide open and will most likely involve a relocation.  I know that I am ready to move on, so the feelings lean more to excitement, luckily.  :)   '),(17725,1794926,'18,December,2003','2003-12-18 00:00:00','   I\'m home!!!  The jet lag is finally wearing off, so I decided it was time to blog.  It feels like I was in Prague 6 months ago, and I was there on Sunday.  I must say that I miss cobblestone streets and buildings that looks like art, but I am glad to be back in Salt Lake, where the sun always shines and everyone speaks English.    The flight home seemed more difficult than the first one.  I had spent Sunday night sleeping in a hostel dorm room was six strangers, and had to get up at 5 in the morning to make it to the airport.  Then, on the 10 hour London to Phoenix flight, I sat in the dreaded middle seat.  The whole process from getting up in Prague, to getting home to Salt Lake took over 24 hours.  I ended up going to sleep at 5 in the morning Salt Lake time, which is the equivalent of 2 in the afternoon  Prague time.  But at least I was in my own bed.  Ahhh.  Absinth party!   I would like to invite anyone interested to join me for an encounter with the green fairy.  I brought home a small bottle of absinth and am preparted to share it with anyone who wants to taste the official drink of the Bohemian Revolution.  I also learned the right way to mix absinth with orange juice: 1. Put sugar in a spoon. 2.  Pour absinth over the sugar. 3.  Light the sugar/absinth mixture on fire until sugar carmalizes. 4.  Put the sugar/absinth into orange juice and add additional absinth.  This is a drink that will knock your socks off.  Absinth is 75% alcohol and contains a compound similar to marijuana.  Therefore, it causes hallucinations if you drink enough.  So stand warned, but come anyway.   '),(17726,1794926,'12,December,2003','2003-12-12 00:00:00','   When I next post an entry, I will be home on my laptop in Salt Lake.  On Sunday morning, I will get a ride up to Prague, spend the afternoon there, and leave early Monday morning for the 24 hour return journey.  I must say that I actually enjoyed the cross-Atlantic flight.  For some reason it reminded me of camping.  You get all your food in these little packages and it is difficult to get to the bathroom because the seatbelt light is always on (this is one potential problem...).  I particularly like the eyemask and socks that you get.  The socks are too big for me, but I wear them anyway because they are my special British Airways travel socks.  And you can watch a movie on a tiny screen with headphones.  Ah, the tiny screen and headphones. I would probably get pretty tired of this whole process if I did it very often, but for now it is still special and interesting.  And considering how expensive it is, I will probably never take long flights enough to get bored of it. Cheers everybody.  I\'m coming home!   '),(17727,1794926,'10,December,2003','2003-12-10 00:00:00','   Does anyone remember that great 80s game Candyland?  I suddenly thought of it last night for no particular reason.  I can\'t recall much about the rules, just that it had a lot of cool places you could visit like gum drop forest or something.  Do kids still play that game?  I would like to have an 80s games party one day.  We could play UNO and battleship and Twister.  Twister was pretty cool.  Twister is like yoga for kids.  It\'s a zen thing.  Some other cool 80s stuff:  1.  Pong 2.  Smurfs 3.  Captain Crunch Cereal 4.  The Monster Mash 5.  Care Bears 6.  Pac Man 7.  Reading Rainbow with Levar (\"Take a look, it\'s in a book\") 8.  Michael Jackson before his issues 9.  High tops 10.  Jelly bracelets   '),(17728,1794926,'09,December,2003','2003-12-09 00:00:00','   Here is the question of the day: Why aren\'t people covered with hair like other animals?  I read an article about how scientists are still debating this question.  Some scientists think we have no hair so we could remove bugs from ourselves more easily.  But I would think that it would be harder for a bug to get to you if you had hair.  And if this were the case, wouldn\'t all animals be hairless?  All animals have to contend with bugs anyway.  So here\'s some interactive blogging a la Kendra: What is your theory about why humans don\'t have hair.  I personally think it might have something to do with sex.  You only need to watch MTV for ten minutes to see what a big deal sex is for humans, I think much more than for other animals.  So maybe this has something to do with the hair thing.   '),(17729,1794926,'05,December,2003','2003-12-05 00:00:00','   This past Sunday I saw the first snow of the year.  It was just a light sprinkle in the morning before the sun came out.  It was a perfect scene for that great George Harrison song, Here Comes the Sun.  After two months of almost daily clouds, I appreciate the sun more than ever.  So if it is sunny where you are right now, look outside and appreciate the beautiful light and warmth.   '),(17730,1794926,'05,December,2003','2003-12-05 00:00:00','   I remember a year and a half ago, a lady I know from meditation had a huge yard sale.  I got there at the end, but she basically sold everything she owned and took off for India.  She said to me, \"I am so glad to get rid of all this stuff.  I feel so free without it.\"  Anyone who has ever moved knows that feeling of \"my God, there is so much stuff.\"  It\'s a heavy feeling, the \"I have too much stuff feeling.\"  It\'s the stuffed with stuff feeling.  I have this feeling right now contemplating taking my stuff back to the U.S.  I don\'t think I have much more than when I came, but it was a job getting everything out here.  My one big suitcase weighed 70 pounds.  I am really hoping I don\'t have to get it on a train.   '),(17731,1794926,'04,December,2003','2003-12-04 00:00:00','   From an ad for the Museum of Communism in Prague:  \"It was a time of shiny, happy people.  The shiniest were in the Uranium mines.\"   '),(17732,1794926,'03,December,2003','2003-12-03 00:00:00','   I just read a sad report about how the Great Apes are going extinct.  Apparently, they often die from human diseases like the flu.  Kind of a strange throwback to the destruction of Native Americans by European diseases.  Also, the apes are extensively hunted for meat because the native people are so poor they need the food.  This seems to be a common theme all over the world.  People are so desparate that they don\'t consider their enviroment.  To them, a rainforest is just a bunch of plants in the way of grazing land.  It seems to me that helping the people improve their situations is a good way to protect the environment.  I think it\'s especially important for richer countries to help others with birth control.  It\'s terrible that President Bush cut funding for birth control.  This kind of squimishness about a public health issue could destroy a lot of the planet.   '),(17733,1794926,'01,December,2003','2003-12-01 00:00:00','   Here\'s something funny, Czech people have a dish called American Potatoes.  It is similar to French Fries, but with fatter pieces.  Maybe the Czechs will change the name if they get upset with the U.S.  Maybe Army Pride Potatoes, or Anger Potatoes would work.  I am coming back home in two weeks!  America seems like a dream.  Odd how you can leave a place and it\'s almost like it doesn\'t exist.  I know that a few weeks after I\'m in the U.S., CZ will seem like a distant memory.  My boss says I can come and visit anytime, which is nice.  I would like to come back, but I don\'t know when I\'ll have the cash to travel to Europe.  Who knows.   '),(17734,1794926,'28,January,2004','2004-01-28 00:00:00','   I\'m not sure if I\'ve written about this before, but it\'s pretty fascinating.  This is a list of the sounds animals make according to the Czech language.    Dog: haf                  Pig: chro  Goat: mee               Sheep: bee  Chicken: ko              Horse: hi  Cat: mnau  So when you go to the Czech Republic, you should say hi to the horses and ahoj to the people.  I still remember talking to a guy in CZ and he could not believe that a pig ever made a noise like oink.  I don\'t have enough experience with pigs to know if they say oink.  So that\'s the deep philosophical question of the day: do pigs really say oink?  Or do they say chro?  Or do they make a sound that isn\'t convertable to human writing?  Hmmm....   '),(17735,1794926,'26,January,2004','2004-01-26 00:00:00','   Let it stand that I officially became an adult on Jan. 26, 2004.  I now have a commute, a cubicle, and a water cooler to stand around.  I also have to wear grown-up clothes, no more fish net stockings or ripped jeans.  Too bad, huh.  I suppose I\'m not totally an adult because I still live with my parents.  Having dinner made for you every night isn\'t very adult.  Also having to call home if you\'re out later than midnight isn\'t too adult.  I guess I\'m about half-way adult.  Maybe three quarters adult.  I would probably put my adult percentage at .6895433.%   '),(17736,1794926,'23,January,2004','2004-01-23 00:00:00','   Yea!  I have a job, sort of.  It\'s actually an internship in marketing at Gibbs Smith, Publisher in Kaysville.  The pay is low (ok, very low), but it will be a great experience at a respected publisher.  I\'ll also get the chance to do some writing with this job related to publicity.  So the days of sleeping until 10 are officially over.  I have to be at work by 8:45 Monday.  That means leaving at 7:45 to drive through rush hour traffic.  Anyway, I\'m pretty happy that something has come along.  It has felt like a lot of things in my life haven\'t been working for the past month.  I\'m hoping this means things are looking up.  I\'m also going to work one day a week at Pince-Nez Press where I interned senior year.  By the way, Pince-Nez Press has a new book coming out--Confessions of a Slacker Mom.  It\'s about parenting without stressing out and becoming OCD.  Looks pretty good.  You can order it at pince-nez.com.  See, I\'m already getting used to marketing.   '),(17737,1794926,'22,January,2004','2004-01-22 00:00:00','   So now I regret a little bit that I wasn\'t a communications major.  I am getting much more interested in the possibility of writing for magazines, and a communications major would serve me much better for that.  Oh well.  Life direction can be hard to find when you\'re 18 years old.  Heck, life direction can be hard to find when you\'re 23, or 40.  Anyway, I enjoyed all my psych classes.  I think for sheer interest psych can\'t be matched.  The only problem is that the jobs I\'m now interested in don\'t come from a psych degree.   I also just discovered that blogger has a spell check function.  Yippee!  I write these posts pretty fast and don\'t really proof them.  At least spell check can pick up the worst offenders.  I taught creative writing to senior citizens, and one of the ladies in class said that spelling is a creative art.  I totally agree with her.   :o)   '),(17738,1794926,'20,January,2004','2004-01-20 00:00:00','   Sunday I went to church for the first time.  Thanks to Kendra for coming along because I probably wouldn\'t have been brave enough to go alone.  :)  We went to the Unitarian Church in South Salt Lake.  The service was very nice, no actual mention of God, angels, devils, fire and brimstone, just a discussion of getting out and doing good works like Martin Luther King, Jr.  Martin Luther King, Jr. was an amazing person.  Whenever I get cynical because of things happening to me personally and in the rest of the world, I think about people like Martin Luther King, Jr. and I have faith in humanity again.   '),(17739,1794926,'13,January,2004','2004-01-13 00:00:00','   I was looking through the want ads today, and I saw an opening for a hod tender.  I have no idea what that is, but it sounds kind of relaxing.  \"What do you do for a living?\"  \"Oh, I just tend the hod.  It doesn\'t need much attention.\"  Somehow it sounds to me like a kind of plant, maybe something like hay.  Maybe they feed hod to hogs and hay to horses.  I also saw an opening for a sheep herder.  I wonder if herding sheep is like tending the hod. Anyway, I have gone through many agonizing hours of deep rumination, and I have come to understand that the novel I am writing has to come first no matter what I do.  I just started when I got back home and I have about 36 rough pages finished.  Basically, it\'s now or never.  I\'ve got to give this book 100% of my brain or I\'ll never finish it.  So basically I am not in career mode, I just need a job to get by while I write.   '),(17740,1794926,'08,January,2004','2004-01-08 00:00:00','   Lisa is here! Yea!  Lisa is interviewing for two internships here in health psychology.  If she picks one of them, she would be moving back to Salt Lake for a year.  She has interviews at seven other places, so we have to see what happens.  She just came from Little Rock, Arkansas and said it was quite nice.  She has also gone to San Antonio (not Houston, Kendra) and Wisconsin.  It seems like a lot of people are preparing to dispurse across the country.  I am curious to see where everyone ends up (myself included). In other news, I have been lolling around looking at Want Ads and trying to decide what to do.  You know you have a lazy life when you get up and take off your sweat pants for sleeping to put on your sweat pants for the day.  The sleeping until 9:30 is also pretty luxurious.  Don\'t be too jealous.  It will end pretty soon ;)  The issue is that I don\'t have any savings right now and Salt Lake almost no jobs in publishing.  So basically I will have to do some other kind of work to build up savings, and I am not sure what kind to do.  If anyone has thoughts, please let me know.  Thanks.   '),(17741,1794926,'04,January,2004','2004-01-04 00:00:00','   Ok so now I have to get a job.  Like today, before the next credit card bill arrives.  I went down to the mall to buy some new sweat pants and ended up with two pairs of shoes and a new shirt.  Everything was so cheap, I couldn\'t help myself.  The shoes at Wet Seal were $10 each, down from $50 originally.  That\'s a great price, so I am justified, right?  Right?  Anyway, I wear my clothes until they disintegrate into little pieces, so I never waste anything.  But now I have to buckle down and not spend any more.  Amazing how fast money goes!  Like pay phone calls are now 50 cents.  What is up with that?  I remember back in my youth when they were 25 cents.  People had a work ethic back then.  They knew how to clear snow off driveways without these fancy smancy snowblower deals.  I don\'t know what the world is coming to :^D  I believe it has something to do with the presidency of George W.  I don\'t trust that man.  Something in his eyes.  He\'s shifty.  First there are weapons of mass destruction and then there aren\'t.  What is up with that?  And all those people who drive hummers.  That bugs me.  Do you know how much CO2 one hummer produces in a year?  24,100 pounds!  (Sierra Club Magazine, Dec. 2003)  A Honda Accord makes 10,300 pounds.  Why would anyone drive a hummer unless they are traveling across the Iraq desert looking for weapons of mass destruction (which they probably won\'t find).  *@3&!  (o+09!   I feel much better now.  I experienced a little blood pressure spike for a moment there, but everything is fine now. Happy New Year!   '),(17742,1794926,'02,January,2004','2004-01-02 00:00:00','   HAPPY NEW YEAR!  I had a quiet New Year\'s Eve at home with my parents.  I watched the New York ball drop thing on TV and toasted with some Martinelli\'s cider, but that was about it.  I considered adding some absinth to my Martinelli\'s, but I\'m a bit worried that I could learn to like that stuff too much, so I refrained.  I also watched Fear and Loathing in Las Vegas, very odd movie.  I am a huge Johnny Depp fan, and he did a great job, but that movie is just one drug-induced hallucination after another.  Hunter S. Thompson also has a pretty interesting voice.  Overall, I\'d give it 1 and a half stars. I also discovered yesterday that I have gained 11 pounds in the last four months.  I don\'t worry much about my weight, but it\'s interesting considering that I haven\'t gained a pound since I was 17.  Must be all that Czech fried cheese.   '),(17743,1794926,'29,February,2004','2004-02-29 00:00:00','   Forgive me for I have been remiss in blogging.  It\'s been a busy two weeks.  This morning I sat watching the snow fall and realized I need to take more time for quiet sitting and contemplation.  I think I got in a situation where my brain was always two steps ahead of my body, trying to figure out what to do next.  Lately I\'ve been getting back into meditation to find that calm spiritual space.  This past year has been a whirlwind when I think about it--graduating, going to Denver, going to Europe, coming back, interpersonal drama, trying to find a job, starting a job, etc., etc.  I think it\'s time to chill for a little bit.   '),(17744,1794926,'21,February,2004','2004-02-21 00:00:00','   Right now I\'m thinking about the conflict between doing work you love and making a living.  I bumped into an acquaintance who stopped working for nonprofit organizations because she made so little.  It\'s unfortunate that people doing really valuable things, like working for nonprofits or teaching, have to struggle financially.  On the other hand, people who work for tobacco companies and weapons manufacturers are rolling in cash.  There is something wrong with that picture.   '),(17745,1794926,'12,February,2004','2004-02-12 00:00:00','   I\'m reading the Tipping Point right now.  Great book.  I highly recommend it.  I was reading the section on Market Mavens and realized that my mother is the greatest example I know.  A Market Maven is someone who knows the price of everything and the best way to get a deal.  My mother actually keeps stacks of Consumer Reports dating back ten or so years and refers to them whenever she wants to buy anything.  She stockpiles toilet paper when it\'s on sale and comparison shops between supermarkets.  She goes to Dan\'s for fish and Albertsons for produce.  It\'s amazing.  I don\'t think I have the energy to become a Market Maven.  I\'ll have to depend on my mom.  I am pretty good about buying things on sale, but I have no idea what the going price is for toilet paper.  For years I\'ve tried to convince my mom to go on The Price Is Right, but she doesn\'t like the host guy.  She thinks he\'s shifty.  She\'s probably right.  I would actually like to go on Jeopardy some day to make use of my pointless trivia knowledge.  For example, did you know that the blue whale is the largest creature that has ever lived.  It\'s true.   '),(17746,1794926,'06,February,2004','2004-02-06 00:00:00','   Today at work I completed a huge blad mailing.  Blads are color reproductions of book covers with a few sample pages inside.  It\'s a pretty menial task, but there is a kind of Zen to it.  You can get into a hypnotic state when you\'re stuffing one envelop after another after another.  Also, I am fascinated by the similarities between the words blad, blog, and blob.  I would like to write a poem using those three words, maybe a sonnet.   '),(17747,1794926,'01,February,2004','2004-02-01 00:00:00','   I played chess against my friend Bob yesterday.  He won (he also won our other game).  I was just considering how chess is a very feminist game.  The queen is the most powerful peace, while the king can barely move and needs everyone else to protect him.  Maybe we\'re meant to think that the king is really old and bedridden, so the queen really has all the power.  It\'s interesting considering that chess is a very old game and wasn\'t created at a very feminist time.  Anyway, I feel like I\'m getting better at chess and want to keep playing.  It\'s pretty fun.   '),(17748,1794926,'29,March,2004','2004-03-29 00:00:00','   I\'m reading a pretty interesting book called Join Me.  It\'s a about a British guy who started his own cult (or as he calls it, collective) by putting an ad in the paper saying only, \"Join Me: Send one passport sized photo to the address below.\"  He just did it for fun to see how many people would respond.  Eventually it became a group devoted to random acts of kindness, a good direction for it to go.  It\'s an interesting thought to see how many people would join a totally undefined group like that.  I think that people actually want to be part of something.  There\'s a lot of alienation today, and maybe things like Join Me can counteract that.  Anyway, it a pretty fun, funny book.  I recommend it.   '),(17749,1794926,'27,March,2004','2004-03-27 00:00:00','   Yesterday I recieved the most backhanded compliment ever.  I had passed by a salon near Cottonwood Mall that advertised eyebrow shaping and said \"walk ins welcome.\"  I wanted to get my eyebrows done by a professional eyebrowist, so I went in and forked out $20.  The girl who did my brows was made up and dressed to the 9s, while I was in my scuffed shoes and sweater.  She said \"I can tell you\'ll want your brows to look natural.  You\'re certainly a natural girl.\"  She said all this in quite the tone of disdain.  Just as I was about to leave, she said, \"I never take walk ins, but I saw you and said that you really need your eyebrows done.\"  Hey, thanks.  I guess my eyebrows were an emergency case.  That lady need to check her salon\'s sign.  I remember reading that Kate Moss got a backhanded compliment when someone told her she was looking \"healthy.\"  In the modeling world, this means fat.  Go figure.   '),(17750,1794926,'24,March,2004','2004-03-24 00:00:00','   I\'m editing a new book about polygamy.  It\'s amazing to read some of the things the polygamous guys say in here, like \"do not voice your opinion if it is contrary to mine.\"  Men seem to always be the ones making the rules, huh?  Hopefully that will change in the future and everyone will be equal.  I still think women\'s rights are a pretty big issue in the world with genital mutilation and forced prostitution in many parts of the world.  I\'ve always felt so lucky to have been born in America where at lest things are a little better.   '),(17751,1794926,'20,March,2004','2004-03-20 00:00:00','   I spent most of today washing, waxing, and vacuuming my car.  There is something satisfying about a clean car.  My brain feels more cluttered if the environment I\'m in feels cluttered.  Next I\'m going to move on to cleaning my room.  I have little piles of papers all over the floor that I need to get out of the way.  It\'ll be nice to see the carpet again and get rid of all the dust bunnies and spider webs.  I think cleaning is actually really good exercise, too.  Why go to the gym when you can clean your house?  Well, I guess you can\'t meet potential dates while cleaning your house, but still.   '),(17752,1794926,'17,March,2004','2004-03-17 00:00:00','   Ahh, I\'m so glad the sun is out and the snow is melting.  Today I wore sandals, so it was offically the first day of spring.  I think everyone should consider spring to begin with the first day of sandals. I\'m also considering going to Wendover for the weekend with my parents.  Wendover is kind of odd.  It\'s amazing how you can cross over from East Wendover to West Wendover and watch the sin creep over everything.   '),(17753,1794926,'29,April,2004','2004-04-29 00:00:00','   Here\'s a great word-nerd joke that demonstrates the importance of proper comma use:  \"A panda walks into a cafe.  He orders a sandwich, eats it, then draws a gun and fires two shots in the air.  \'Why?\' asks the confused waiter, as the panda makes towards the exit.  The panda produces a wildlife manual and tosses it over his shoulder.  \'I\'m a panda.  Look it up.\'   The waiter turns to the relevant entry and finds the explanation.  \'Panda.  Large black-and-white bear-like mammal, native to China.  Eats, shoots and leaves.\'\"  Ha! Ha! Ha!  Doesn\'t that just crack you up?  Who ever said punctuation was boring.  That joke came from the book Eats, Shoots and Leaves, which I\'m reading to refresh my editing skills.  Cool book.   '),(17754,1794926,'24,April,2004','2004-04-24 00:00:00','   I\'m proud of myself because I had a fairly productive week.  I\'ve written a new nonfiction piece that I want to send out to magazines soon (I\'m keeping my fingers crossed).  I\'m still tweaking this piece, but I\'m pretty proud of it.  I\'ve also gotten a couple of applications done that I\'ve procrastinated on for a while.  Probably the most exciting thing to happen this week is that I was involved in a protest at Smith\'s on 9th and 9th.  The protest was organized by the group CodePink to address the fact that Smith\'s keeps its condoms and pregnancy tests in a glass case that\'s locked after 9 pm.  This restricts access to the condoms so people have to embarrass themselves asking someone to open the case.  I stood on the sidewalk outside of Smith\'s holding a pink sign that said \"Condom Liberation\" and wearing a pink beehive hat made out of a spray pained flowerpot.  We handed out a lot of condoms to passers by and gave a writ of habeas corpus to the Smith\'s manager.  It was really quite fun.  Now I want to protest something else.  Link to SL Trib article: http://www.sltrib.com/2004/Apr/04242004/utah/160197.asp   '),(17755,1794926,'17,April,2004','2004-04-17 00:00:00','   Today I went on a fun hike with Kendra (hey Kendra).  We didn\'t go too far because it was cloudy and cold, but it was still nice to get out.  I love living in a place where the mountains are ten minutes away.  Right now I\'m making a resolution to hike more 1) to get in better shape 2) to enjoy nature.  I\'m also making a resolution to eat more cheese.  Cheese is a healthy, simple snack and I\'d like to support America\'s dairy farmers.   '),(17756,1794926,'09,April,2004','2004-04-09 00:00:00','   I just went to the dentist and found out I probably need gum surgery.  For years I brushed my teeth too hard, and now my gums have receded around one tooth.  Doh!  It sounds pretty unpleasant.  The dentist will have to cut a chunk of flesh out of my upper palate for a graft.  And I thought brushing thoroughly was doing the right thing.  It just goes to show you that everything in life is a balance--if you brush too hard, your gums recede; if you brush too soft, your teeth rot.  I think that\'s my new life motto.   '),(17757,1794926,'03,April,2004','2004-04-03 00:00:00','   Yesterday I got a new toothbrush.  My old toothbrush was so worn out that bristles were coming off in my mouth.  One thing you don\'t want is to brush your teeth and choke on bristles.  A new toothbrush is rather refreshing.  Ah, the small pleasures of life.   '),(17758,1794926,'27,May,2004','2004-05-27 00:00:00','   In regards to the last post, anyone contributing to the day job book can reach me at khburgess@uofu.net  Thanks   '),(17759,1794926,'26,May,2004','2004-05-26 00:00:00','   I\'m finally getting underway with a book idea I\'ve had for a long time.  The idea is basically coverage of day jobbing writers, actors, musicians, painters, and dancers and how their day job affects their art.  This week I\'m doing my first two interviews for the book--a musician and a painter.  I\'m excited and hopeful that everything will go well.  Personal Plug: If there are any day jobbing writers, musicians, actors, painters, or dancers reading this, please get in touch with me if you\'d like to be in my book.  I can use all the help I can get!  Anyway, I\'m hoping that future generations of day jobbers will find this book helpful on the quest to living by their art.   '),(17760,1794926,'20,May,2004','2004-05-20 00:00:00','   Well I thought it was time to move into the modern world and post a picture of myself.  This pic is a bit old, but I still look pretty much the same (my hair is a lot shorter).  I\'ll try to scan in a newer picture if I can figure out where the scanner plugs into the computer.   '),(17761,1794926,'13,May,2004','2004-05-13 00:00:00','   Dots!  Dots everywhere!  I love dots.   '),(17762,1794926,'13,May,2004','2004-05-13 00:00:00','   I\'m feeling rather disoriented because the Blogger people changed the layout of the website.  Just when I was getting the hang of which buttons to push, they have to go and do that.  The post I just published has the line spacing all wrong and I blame this new fangled technology.  Now there\'s some weird thing where you keep typing along the same line forever.  Very confusing.   '),(17763,1794926,'03,May,2004','2004-05-03 00:00:00','   Somewhat related to my last posting I have now become addicted to the Friendster site (by the way, anyone is welcome to add me as a friend).  It\'s just so fascinating to click on  friends and their friends and on and on forever.  My cousin Jade introduced me to the whole thing and she has some really interesting friends, people like models and musicians.   One of Jade\'s friends has 500 contacts on Friendster.  I\'m not sure if I\'ve met 500 people in my whole life.  I feel a bit jealous, but I have three good friends on there, so I should appreciate what I have.   '),(17764,1794926,'03,May,2004','2004-05-03 00:00:00','   I\'ve been thinking about this phenomenon of people reading blogs by people they don\'t know because they\'re linked to the blogs of people they do know.  That was a very long sentence!  Anyway, I\'ve started looking at the blog of someone who comments on my sister\'s blog.  (Hi Lisa.)  It\'s like a little window into a stranger\'s life.  Maybe there is six-degrees of blog separation between most people in the world.  People who don\'t own computers can\'t blog, but they might know someone who blogs.  I think eventually the degrees of separation between people will get smaller and smaller as everyone is interlinked through blogs and IMing.  Basically what I\'m trying to say is that the blog can bring about world peace.  Thank you.   '),(17765,1794926,'01,May,2004','2004-05-01 00:00:00','   A question: Have you ever googled yourself?  I just did and I have no shame.   '),(17766,1794926,'24,June,2004','2004-06-24 00:00:00','   Well I got to Sarah Lawrence safe and sound and now the conference is almost over.  On  one hand I feel like I\'ve been here a long time and on the other it feels like I just arrived.  Bronxville is very cute and a 5 bedroom home costs 2 million.  So it falls into the category of a nice place where I don\'t have a snowball in hell\'s chance of ever living.  Unless I marry rich.  To do that, I should probably look into breast implants, but that\'s another story.  One high point so far has been seeing fireflies for the first time.  Somehow I thought they just glowed constantly like Christmas lights, but they\'re more like sparks--sudden flashes.  Pretty cool.  I have gotten good writing feedback as well and have some revision plans for an old piece.  Anyway, I should probably get back to work.  Cheers.   '),(17767,1794926,'18,June,2004','2004-06-18 00:00:00','   Well tomorrow I\'m on my way to NYC for the Sarah Lawrence Writers Conference.  As a typical procrastinator, I haven\'t packed yet.  Luckily I\'m not working tomorrow and my plane doesn\'t leave until 11 pm.  This should be an interesting trip, though I am a little bummed I don\'t have any day job artists signed up to interview.  I\'ll have to snoop around coffee shops and approach people who look creative.  I\'ll be watching for berets and black turtlenecks. (Puma sneakers also seem to be part of the standard day job artist uniform).   '),(17768,1794926,'14,June,2004','2004-06-14 00:00:00','   THINGS I\'D LIKE TO DO  1.  Write a screenplay (and a novel, and another play, etc., etc.)  2.  Open my own publishing company   3.  Travel to China and see where my grandparents grew up  4.  Go the Deer Park Monestary for the young adult retreat  5.  Paint my toenails a color I\'ve never used before  6.  Learn to make good vegi curry  7.  Get a dog (something small with hair that doesn\'t get too knotty)  8.  Get into an MFA writing program   9.  Meet a nice guy who thinks all of the above are really cool  10. Have a hot stone massage  11. Find the perfect eyebrow shape for me (ah, the never-ending search for the perfect eyebrow)   '),(17769,1794926,'10,June,2004','2004-06-10 00:00:00','   I\'ve gotten a fascinating new book--Buzz: The Intimate Bond Between Humans and Insects.  Most of the book is photos of bugs taken with a scanning electron microscope.  It makes their eyes look blue or red and their bodies look lime green.  Very interesting. This book also has facts about bugs such as, \"It\'s been estimated that 10 percent of the weight of a two-year-old pillow is composed of dead dust mites and their droppings--dung that the dust mite may reingest as many as three times.\"  I think my pillow is about 6 years old, so 30 percent of its weight must be dung and mites.  I\'ll be thinking about that when I go to sleep tonight.   '),(17770,1794926,'04,June,2004','2004-06-04 00:00:00','   Hey All,  I just discovered a cool new web site--www.digsmagazine.com  One of the musicians I\'m interviewing for my book told me about it.  Very similar to Ready Made Magazine.  I want to do a project on there making a box from old album covers.   '),(17771,1794926,'01,June,2004','2004-06-01 00:00:00','   I\'ve gotten into Bookcrossing since I heard about it from Lisa (hi Lisa). Tomorrow I will release my first book, How To Be More Creative.  I think I\'ll release it at the Community Writing Center, a good loction for book people.  The idea of becoming more creative from a book somehow seems ironic to me, but this book has some interesting ideas.  Here\'s one: \"Stand up and look between your legs for two minutes.  Is anyone watching?  If so, this exercise is doubly effective: to people are gaining a new perception of their surroundings.\"  When I first read this I thought they meant something else when they said \"look between your legs.\"  Then the thought of someone watching was a bit disturbing.  I mean, hopefully you\'d already be aware that someone was down there, right?  If not, I wonder about you.   '),(17772,1794926,'29,July,2004','2004-07-29 00:00:00','   Yesterday I helped a lady with a resume and she had worked as a blackjack dealer for five years.&nbsp; She said the money was really good, but the lifestyle was unhealthy.&nbsp; She herself started drinking and gambling too much.&nbsp; I must admit that after talking to her I had fantasies of going to Vegas and becoming a blackjack dealer.&nbsp; Seems like it could be a&nbsp;cool day job for a writer.&nbsp; You see lots of different people and it\'s not too mentally taxing.&nbsp; I love Vegas and I don\'t drink, so I felt that I wouldn\'t fall into any alcoholic problems.&nbsp; When I told my mom, she said it seemed like an immoral job.&nbsp; We\'ve had a few compulsive gamblers in our family, and my mom said that job would be supporting&nbsp;bad habits.&nbsp; I don\'t know.&nbsp; I don\'t think gambling is evil&nbsp;in itself, if you only gamble a little.&nbsp; But my mom is probably right that the casinos don\'t care if people throw away their life savings.&nbsp; Not very Buddhist.&nbsp;   '),(17773,1794926,'21,July,2004','2004-07-21 00:00:00','   Well Kendra is heading out for a new life in Austin.&nbsp; I went&nbsp;to Bombay House with her, Jon (Kendra\'s husband), and my parents.&nbsp; It was definitely bittersweet.&nbsp; Kendra and I have had some great times.&nbsp; I\'ll miss the Scrabble/cooking nights and hikes.&nbsp; I\'ve seen several waves of friends leave Salt Lake over the years.&nbsp; That seems to be the way of things.&nbsp; Life is about change, huh.&nbsp; My&nbsp;writing mentor, Steve, is also moving away to Upstate New York and Lisa is moving to Palo Alto.&nbsp; I have noticed people moving in clumps.&nbsp; Interesting.&nbsp; I wonder if that\'s an urban legend, like every ten minutes a crowd goes silent (have you heard that one?).   '),(17774,1794926,'14,July,2004','2004-07-14 00:00:00','   I was thinking yesterday about how little it takes to get me really upset.  This past winter I had a relationship that was emotionally abusive.  The guy would say I was stupid and laugh at me among other things.  In the scheme of human events, this is pretty minor.  I mean, there are people being sold into slavery and starving to death in this world, but it\'s amazing how much it hurt.  I think maybe it was more what the experience represented.  I had trusted this person and he turned on me.  I told my sister at the time that I wondered how I would ever trust anyone (or maybe any guy) again. I don\'t feel that way now but I\'m pretty cautious.  The whole thing was a learning experience but mostly I learned about the ugly side of life, things I didn\'t want to know.  The trip to NYC was great because I felt like I finally let that person go while I was there. I basically said, \"I won\'t let you make me miserable any more.\"  It was just a decision I had to make.  I remember an old friend told me that happiness is a choice.  You have to choose happiness because there are so many things you can fixate on and make yourself unhappy.  So many glasses aren\'t just half fill, they\'re smashed up and scattered across the sidewalk.  You have to see the beauty in the glittering broken glass.   '),(17775,1794926,'08,July,2004','2004-07-08 00:00:00','   As Kendra mentioned in her comments, I have gotten my ears pierced.  For a long time I subscribed to the notion that you shouldn\'t alter the body from its natural state.  I took great pride in the fact that I had no piercings, gauges, tattoos, etc.  Then a few years ago, I got my hair highlighted and I started down the slippery slope.  Now I\'m considering a tattoo, maybe my Chinese name in characters on my back.  The one thing I worry about is whether I could develop an allergy to the ink.  Hmmm....    The other cultural phenomenon I find interesting is tattooing someone\'s name on yourself.  So here\'s today\'s poll: would you ever consider tattooing someone\'s name on yourself and if so, where would you put it?   '),(17776,1794926,'06,July,2004','2004-07-06 00:00:00','   So I\'m taking a poll.  How many of you out there think I should grow my hair long and how many think I should keep it like it is.  Should I dye it red or blonde or not at all?  Please vote for a combination from the choices listed.  Thanks.   '),(17777,1794926,'02,July,2004','2004-07-02 00:00:00','   Hi All,  NYC was fabulous!  Fabulous, fabulous!  I went to the Frick Museum, the Natural History Museum, and a great club (Lotus--$20 cover).  I\'m basically ready to move there, but money is a big issue.  I also would like to get a dog and I would hate to inflict a studio apartment upon one.  I\'ll definitely work for a year and then see what my options are.  Big thanks to my cousins Alison, Jade, and Sue Ling who all showed me such a good time.  Hopefully we\'ll all meet up again soon.   '),(17778,1794926,'05,August,2004','2004-08-05 00:00:00','   So I got a call a few days ago about my 5th high school reunion. Instantly I remembered that line from Romy and Michelle\'s High School Reunion:  Romy: So are you going to the reunion?  Janeane Garofalo (can\'t remember the character): I\'d rather stick this [cigarette] up my ass.  The Janeane Garofalo girl ended up going to the reunion, but I don\'t think I will. Isn\'t five years way too soon to have a reunion? I\'ll make the declaration now that I will attend my 50th high school reunion. Then enough time will have gone by.   '),(17779,2004217,'30,September,2003','2003-09-30 00:00:00','    I have just realised the implication of publishing my personal musical tastes to the whole world. My reputation will be  ruined.   And now my ex is online! Whoopidoo. It\'s not his fault I find him immensely irritating... in fact, he\'s not even irritating, it\'s just that he\'s a little... slow. He\'s nice and all, but I don\'t think he was in the queue when they were handing out conversation skills, and right at the front when they were handing out free copies of \"How To Be Completely and Utterly Irrelevant\" by I. A. M. Pointless. For example, I\'m telling him about my weekend (He asked, before you have me down as a self obsessed little girl) and what does he say?  \"Was the film good?\" \"Ah right, cool\" \"So not such a good weekend then, eh? lol\"  Nooo! Nothing so witty! I got:  \"Spurs are wearing an all purple kit.\"  Why the  FUCK  do I need to know this? And how in hell is it in any way relevant to Calendar Girls? There\'s another puzzle for you. I say \"Erm..ok..\" and does he respond/apologise? Nooo! He starts telling me about some girl he met the other day. Well, at least there might be a link there... there are women in Calendar Girls?  Good Grief.  Desperate to get to the show Ceff x   # I go up, I come down, and I\'m emptier inside... tell me, what is this thing that I feel like I\'m missing? And why can\'t I let it go?#   '),(17780,2004217,'29,September,2003','2003-09-29 00:00:00','    Well, after a long and in-depth discussion of the merits of RX Bandits with Vez (Hey Vez!!), I have come to the conclusion that although The Resignation might be a very good album, it is not for the people who are expecting more of the same ska/reggae vibes that were on Progress. Although Sell You Beautiful is an AMAZING song, it is neither ska nor reggae, and as I bought the album expecting this, I was disappointed, as, I would imagine, all the other people buying this album expecting ska/reggae were. RX Bandits have definitely taken a new direction with this album, I\'m just not sure whether I like it or not. Fair play to them for recording it live and all, but a lot of it still sounds very self-indulgent to me. Bring back the Ska!  See you at the show Ceff x   '),(17781,2004217,'29,September,2003','2003-09-29 00:00:00','    Arrrggggghhh Blergh it won\'t let me publish new posts! I am desperately trying to think of something hilariously interesting today, but it is not occuring. I wish fun things happened to me.  How about a puzzle? That sounds mighty fun! Ha. Ha. Try and work this one out:  My current playlist on the  GENIUS  invention that is Windows Media Player 9 consists of:  1. Stacie Orrico- There\'s Gotta Be (More To Life) 2. Beyonce Knowles feat. Sean Paul- Baby Boy 3. Muse- Our Time Is Running Out 4. Muse- Apocalypse Please 5. Muse- TSP 6. Starsailor- Poor Misguided Fool 7. Starsailor- Alcoholic 8. Starsailor- Lullaby 9. Lostprophets- Burn Burn  (This song fucking RULES I command you all to download it from www.dragonninja.com IMMEDIATELY)  10. Taking Back Sunday- Ghost Man On Third 11. Taking Back Sunday- Heart On Her Sweater 12. Taking Back Sunday- You Know How I Do 13. Taking Back Sunday- Lost And Found 14. Taking Back Sunday- You\'re So Last Summer  (containing the soon-to-be-classic lyric:  And with my one last gasping breath I\'d apologise... for bleeding on your shirt )   Ha! Work that one out!  See you at the show (Next one: LOSTPROPHETS BABY! IAN WATKINS IN THE (damn hot) FLESH!) Ceff x   '),(17782,2004217,'29,September,2003','2003-09-29 00:00:00','    Musical Recommendation!!   I strongly recommend  Senses Fail . They are soooo cool!  Mp3.com defines them as Emo, but I don\'t like that word because it has no real definition. I\'m sure I can make one up though... but it\'s really hard cos Emo ranges from Dashboard Confessional to Taking Back Sunday and those bands are really different...why am I whining about this??  Anyway Senses Fail are really good, they have an EP out on Drive Thru, go buy it they rule.  Don\'t buy: The Resignation by Rx Bandits if you like Ska. Biiiig disappointment. There are some moments of sheer brilliance (Pal-Treaux) which then irritatingly descend into self indulgent guitar scraping. Boo Hiss to them.  See you at the show Ceff x   '),(17783,2004217,'29,September,2003','2003-09-29 00:00:00','    Wahey! Look at me, Blogging away.  Welcome! To the (probably less interesting than it sounds) Blog of Ceff. I\'ll be including jokes, hilarious stories along with my usual dose on cynicism on all things male. Don\'t get me wrong, men are fab (Mmmm... Men ;o)), but I also happen to believe that they are all mean, mean people who keep their brains in their trousers and never change one bit. I would go into details about how I came to this conclusion, but I don\'t feel like it today. Maybe another time when i\'m feeling much more self-pitying.  So! Onto today\'s random ramblings... 1. For the first time EVER I didn\'t want to go to English today. It was such a horrible feeling! I felt so guilty... I went, obviously, being the good, studious type, but the passion wasn\'t there. Gutted. Mind you, I didn\'t really want to be in school at all today, I would have much rather just stayed at home and thought about...  2. MEN. Well, first things first...GAZ. It\'s all very confusing. I doubt he really remembers me that well, but because I know ABSOLUTELY NO MEN, I remember the few that I have the privelege to meet in great detail. So- texted him on Sat, invited him out with us on Thurs, but alas, no reply. Will text again tomorrow, with the approval of Liz, who seems surprisingly encouraging about the whole situation- suggesting I went with him to the VALENTINE\'S DANCE! Good. Grief! It\'s all very confusing.  So, that is the current situation... as soon as more interesting things come up that are fun to read and play around with, I will add and stuff, cos this is all very exciting...  I should do something productive now... I\'m sure this counts!  See you at the show Ceff x   '),(17784,2004217,'30,October,2003','2003-10-30 00:00:00','    Well hello there! After a fairly uneventful week, last night saw me cutting some rug at Academy 2 for what seemed like my first gig in YEARS, but was only a month after the last one. Here\'s what I thought:  Despite being Welsh, first support When Reason Sleeps made a fairly good effort at getting the crowd going with their very \"emo\" scream-sing-scream-sing dual vocals and chirpy onstage banter. Unfortunately the Academy crowd are less than receptive; the attitude towards the support tonight is not an open-minded one. In the crowd\'s eyes, they are simply wasting time and energy, because they\'re only here for one band: Lostprophets. Not that this deters WRS- they produce a solid 30 minute set, and finish with on a suitably emo note: \"slit my throat and hope this funeral never ends\". Ouch.  Eden Maine, we are warned by WRS, will \"scare the shit out of some of you\". Imagine the collective \"huh?\" that goes round the room when 5 unnasuming yet fairly hairy men take to the stage. I expect great things of Eden Maine after their ominous introduction; I am sorely disappointed. Eden Main are tuneless, chaotic and sound totally unco-ordinated. After the 10 minute-long opening track, the lead vocalist (I hesitate to use the word singer)Adam Symonds mumbles, totally out of breath, incoherently into the microphone before launching into another 5 minutes of out-of-time screaming, unnecessary shouts and uninspiring instrumentals. After the previous pitiful attempt at speech, the band make no other effort to communicate with the crowd before they wander offstage, out of breath and out of fans.  After this disappointing start, the atmosphere amongst the audience is now so expectant that, should the \'prophets underperform, a riot might break out. Thankfully, the sheer showmanship and musical quality of the band is so stunning that a riot is far from everyone\'s minds. On a cramped stage, in a tiny venue. under such pressure to live up to their image, Lostprophets performed as if they were in a stadium: songs from the forthcoming album \"Start Something\" sounded polished and were well received and \"classics\" (singer Ian Watkins\' words, not mine) from debut album \"The Fake Sound Of Progress\" were played with total confidence and imagination. The crowd surge backwards and forwards and often drown out Watkins with their singing, especially on debut single \"Shinobi vs Dragonninja\". The \'prophets finish with current single \"Burn Burn\" and leave us breathless, stunned and gagging for the new album. Roll on January...  AND OH MY FUCKING GOD I TOUCHED IAN FUCKING WATKINS!! *screams girly scream and faints in girly manner*  Ceff x   Wanna do whatever, laugh until September...   '),(17785,2004217,'27,October,2003','2003-10-27 00:00:00','    I have just experienced what was possibly the weirdest weekend of my entire life. It was spent entirely in a cosy hotel on Lake Windermere with 9 other people for company, and included Em\'s 18th birthday party, flirting with a 40 year old and my dad\'s wedding.  So let\'s begin at the most obvious place: The Beginning. I was collected at 2.30pm from home by Daddy Dearest to be carted off to Worcester for final packing and panicking. We set off at the RIDICULOUS hour of 8.30am the next morning and arrived at the Langdale Chase hotel for lunchtime. We proceeded to the bar, where I discovered the joy of tabs (This weekend\'s most used sentence? \"Just put it on room 20\") and by 5pm everyone had arrived:  1. Me  2. Em  3. Dad  4. Sue  5. Sue\'s mum Marion  6. Sue\'s dad Roger  7. Sue\'s brother David  8. David\'s wife Claire  9. Sue\'s mate Debbie  10. Dad\'s mate Paul   Much complaining about traffic occurred before we were ushered into the restaurant for Em\'s 18th party, which was fun due to party bags and balloons. After a HUGE meal which included two desserts (both cake- oops), we headed into the piano room for Tequila: Cue much amusement on my (sober) part watching Dad, Sue, Em, Debbie and Paul\'s faces screw up in disgust after each slammer. Ha. One by one everyone left for bed until finally it was just, Me, Em and Paul. At about 1am we headed up to bed, where Me and She ended up (quite logically after some- ahem- mutual revelations) talking about affairs with older men. Then to sleep...   We awoke accidentally an hour early on Sunday(THE WEDDING DAY) due to the clocks deciding to move in the middle of the night- breakfast was brought to the room by a rather nice young waiter who I then spent the rest of the day telling people was \"with us for breakfast\", mwahahaha. FIT. At 1pm I retired to my room to do my makeup, which immediately prompted my face to go the colour of milk with red blotches. I went for a walk. Somehow (*innocent look*) I arrived outside Paul\'s room, but due to his in-the-bath nature, I went to the bar for a cold drink and some deep breathing. Skin soon resumed its normal complexion and I was able to proceed with the dressing process. At 2.20pm we arrived in the bar to much oohing and aahing, before heading up to the bridal suite to collect, well, the bride. We walked down the stairs into the main hall at 3pm to Puccini and by 3.20pm we were exiting the hall for champagne and photos to Bon fucking Jovi. Good Grief. In between time, Daddy Dearest married Wicked Stepmother, I read a sonnet and felt strangely apart from the whole event. It was as though I wasn\'t part of the new life, simply a reminder of the old one, dad\'s dark beginnings; a skeleton in some sort of metaphorical closet. We went outside into the cold and did photos, toasts and comparisons of shoes. The party then carried on in the function room for another HUGE meal where I, feeling much more relaxed about Em knowing about it, flirted a bit with Paul. I am fully aware that you now think I\'m a sick, twisted individual who is just asking for trouble, but he\'s just. So. FIT. And kind. And  very  complimentary towards me and Em about our looks. And it wasn\'t disgusting, I wasn\'t all over him, just smiling, playing with hair, laughing at jokes, a little bit (more) touchy-feely. I was much more concerned about Em: hanging off his arm, touching his leg, asking him to go OUTSIDE with her ALONE for a \"WALK\" because she was \"HOT\". He didn\'t go in the end; I took the opportunity to point out my sister\'s frisky behaviour at which point he abstained from opinion. Unfortunately, I was unaware at this point she was stood at the window listening in. Oops. I went outside to look for her (in the DARK- clearly an indication of how caring I am towards her; I am NOT a dark person), and on finding her was given ICY shoulder so I went back inside and sat by the fire. She appeared on the balcony a while later and relative normality was resumed, if you can call normality eating sausage sandwiches at 11.30pm while Dad sleeps, pissed, in a chair in the middle of the hall as his bride attempts to get him to go to bed. Once again, by 12.30 it was just Me, She and Paul. We ended up talking about the marriage, how I felt about it and how Sue felt about Me and Em. Em made a comment: \"She\'s better than some of his other partners, and I never met his first wife.\" Thanks, sis, that makes me feel a whole lot better. I left for bed and after a \"chat\" with Paul, Em followed.   We awoke again this morning with breakfast from FIT waiter, then we packed and headed down to meet everyone at breakfast. Millions of goodbyes occurred, and finally Sue\'s parents and I left the hotel for a boat ride and a long, sleepy journey home.   Right now I am considering texting Paul. Every time someone said something that involved the title of a song, we both said \"That could be a song\", which provided much amusement and a strange element of consistency throughout the whole tale. We said \"Cheers\" to each other more times than I think I\'ve ever said cheers to anyone in my entire life, and although he\'s twice my age I spent most of the weekend feeling as though I\'d rather be in his room cracking jokes and talking music than listening to Em drone on about Chris/Ben and whatever other boy might be in her life at the moment. He was the constant in the weekend, he didn\'t change, he didn\'t get married, he didn\'t go off in a strop. Dad got it right in his speech, he\'s \"fucking rock solid\". And this is not a random diatribe of some lovesick teenage girl with odd fantasies; I doubt I would ever end up doing anything even remotely naughty with Paul, I just felt close to him, he was my mate for the weekend, which was just what I needed at a time when everything that was ever dependable was thrown entirely out of place, out of time and, most of all, out of context. I needed a friend, and he was it. And anyway, he\'s banned me from being naughty. I now have to send him proposals 3 weeks in advance of what I plan to do, where and with whom I plan to do it. In triplicate. He\'s also decided he wants to interview any prospective boyfriends before Daddy Dearest meets them. This (clearly) means I\'m going to have to visit his house more often...   Ceff x    Lean on me, when you\'re not strong...   '),(17786,2004217,'24,October,2003','2003-10-24 00:00:00','    Right! Plan for tonight\'s blog: 1. Halloween Party 2. WS1   The halloween party was really quite amusing. I avoided Draper for the entire night and had lots of fun with Selina and Laura, which mainly involved laughing at Suki and Keith Leddington\'s dancing skills. Ha! All was going well until 10.30 when someone severed a wire and the speakers went \"Pffffft.\" Macho (well, sort of) chanting then occurred, ranging from \"Hemmell (sp?) is an arsehole\" to \"Engerland, Engerland, Engerland\" and \"You\'ll Never Walk Alone\". Everyone was removed from the building at 11pm. I went home, and lots of others went to Yates\'.   And so to Thursday...   We all headed out from Zeb\'s at 8.45 to WS1, where we stood in the FREEZING cold for ages, watching year 11 after year 11 get turned away, becoming more and more fearful that they would ID us, until we were safely inside. Phew! But alas, we were not followed by Kat n Coz n Beth n Nat, and we (sadly) assumed that either Beth or Nat didn\'t get in due to youngness. Turns out it was Kat who was denied entrance- weird, seeing as she\'s been in before. Boo hiss to bouncers. Inside we danced, ran into Carly (YAY), Nina (drunk), Helen (looking rather sexy) and Bhav (oddly enough). Also, EVERYONE who EVER went to my primary school was out last night- saw at least 10 people who I knew or knew me... strange... We headed upstairs at 11 where I had a nice conversation with Draper at the bar about the whole sordid affair and how everyone was taking the piss out of him at school- we cleared everything up and we don\'t fancy each other and it\'s all good :o) He left at about 12.30 and at 1 we were approached by Sean (FITFITFIT) who claimed \"sponsored kissathon for charity\". Needless to say, I gave generously. In fact, I was feeling so generous, I made three donations. mmmm MMMM!! Granted, he also got off with Zeb and Selina and a few other girls, but dammit I PULLED! Laura/Selina/Becky were much disgusted. Oh well! mmmmm Fit....   Ceff x    Let the music smother me, whole weekend recovery...   '),(17787,2004217,'23,October,2003','2003-10-23 00:00:00','    Wahey! Have found a way to immediate bloggage! Wheeee! Who knew that \"Blog This!\" would be so useful...  Nothing interesting happened today. Me and Liz made some poems about Girth and Romping, and I pondered about men... again. Ho hum.  Ceff x   I can see the headlights and the pavement, my legs won\'t carry me that far...   '),(17788,2004217,'22,October,2003','2003-10-22 00:00:00','    Ha! On the annoying Search bar that appeared on my browser when I downloaded MSG+ there is a link to \"Foot sex\". How intriguing...  My dad posted me some obituaries today. He got my flat number right, the flat block number right, the town right, the county right, but missed out the postcode and wrote HARBORN road instead of CAMBORNE! It\'s a little more than disconcerting when your own father doesn\'t know where you live. Granted, I can never remember whether his home phone number starts 724.. or 274... but come on! Harborn?? Good Grief.  2 irritating DRAPER-related things today: First: I got my ticket for the halloween party, and on the back is written \"Jess King (Uter)\" Apparently \"Uter\" is written to distinguish me from Jess May who is also coming tonight. Because obviously my SURNAME doesn\'t distinguish me from her...!! Second: text \"from\" Draper this afternoon: \"i love u\". Ha! Turns out Chris Devy sent it. I am less bothered by this, in fact it\'s kinda funny, but I still wish people would just leave me alone :o(  The search bar now says \"Sex girls\"... don\'t think I\'ll be clicking anywhere near that any time soon...  Ceff x   Reception less than warm...   '),(17789,2004217,'22,October,2003','2003-10-22 00:00:00','    Hummmm. I should be doing homework. English homework. I should be doing it with great enthusiasm, because I like analysis questions. But I find myself here on this blog, in the middle of a conversation with Luke on MSN. He\'s babbling on about something; I haven\'t replied in about 15 minutes but he keeps sending me messages. Ahh, he\'s signed out- \"gtg c ya l8r.x\" (that\'s right, not only is he TEXT LANGUAGE OVERLOAD MAN but he also has a chronic disability to use the space bar after a full stop). I feel almost dirty for going out with him- why did I have to choose someone who is so blatantly, BLATANTLY wrong for me? I doubt this is hindsight speaking though, I think subconsciously I wanted someone so much that when he messaged me and told me how wonderful I was, I just fell for it. Stupid girl.  Ceff x   You called me on the phone, but you don\'t even wanna talk!   '),(17790,2004217,'21,October,2003','2003-10-21 00:00:00','    Gah. Here\'s me, quite happily going about my daily business without fuss, no friend troubles, no family troubles and most of all no MEN troubles. Until! 1. Debbs isn\'t speaking to Laura because (from what I can gather) Laura doesn\'t like Debbie\'s boyf Matt and gave him a \"dirty look\" at Becky\'s party. Good Grief. How petty. It is nobody\'s fault that Laura and Matt don\'t get on, Laura seems to be just fine about Debbs not liking  her  boyf, so just get the fuck over it!! 2. Saw Loz at the bus stop today who was making disgusting comments about me and Draper. Apparently they\'ve been winding him up about me and him and making up texts \"from\" me asking if he\'d got a boner. I repeat, just get the  fuck  over it! I don\'t fancy him, he doesn\'t fancy me, end of story. Am now dreading Wednesday due to his (Draper\'s) presence. Woo.  Ceff x   Calm before the storm, set it off...   '),(17791,2004217,'19,October,2003','2003-10-19 00:00:00','    Shattered. Absolutely shattered. This has been one of the busiest weekends ever in the history of busy weekends...   Friday : Free afternoon, leave school at 12.30, go to post office, stamp E111 form for trip to Germany in 2 weeks (2 WEEKS! Argh!). Go to Birmingham to sell Hundred Reasons ticket- Academy don\'t do refunds. Luckily man behind me in the queue was wanting to buy a Hundred Reasons ticket- Sorted. Unsuccessful raid of Dorothy Perkins half price sale led to trip to Metal section of HMV and a purchase of \"Casually Dressed And Deep In Conversation\" by Funeral For A Friend. Buy this album. Arrived home 3.30pm, shower, left house for Birmingham University to meet Oli for curry and night out at the Guild. Spent hour previous to travelling to Guild watching Oli, James (cute) and Paul (FIT FIT FIT) play \"Pro Evolution Soccer 3\" on PS2. Was surprised I did not need to be woken before leaving icky flat. Spent night standing around quietly while Oli talked to other people who didn\'t talk to me, until about midnight when I engaged in conversation with Paul (FIT FIT FIT) about Ska and how it\'s great, and how garage rock is shit. Ha. Danced for a while then returned to icky flat collecting 2 road signs, an armchair and a traffic cone on the way to sleep in Oli\'s bed, while Oli went to his girlfriend\'s flat in search of sex.   Saturday : Awoke in icky flat, got up, did Oli\'s ironing for princely sum of Â£4.40 (student discount). Mother arrived with Oli\'s washing and food for next week or so, then headed home for quick bite to eat before heading to Gentleshaw to play with Coz\'s instruments. Rar. Songs were played at me (piano and guitar): Awesome. Cannot begin to comprehend how he comes up with the music he played- totally amazing. He can do \"Plug In Baby\" by Muse- totally beyond me. Sheer musical quality of songs made up for spending a considerable length of time in sub-zero temperatures in the garage. Began to learn bassline to one of his songs and also learnt bassline to \"Smells Like Teen Spirit\" (only 4 notes, but still an achievement). Left house 5 minutes before bus leaves Burntwood; missed bus. Waited 20 minutes in the cold before being able to get home to get ready to leave the house for Pelsall for Becky\'s 18th party. Was fun; Selina and I were a little over dressed but still enjoyed ourselves- stuffed face with flying saucer sweets and laughed at stoopid barmaid, who foolishly asked Selina to \"watch the bar\" while she went for more glasses. Rollo proceeded to stick his head under the Stella Artois tap for a drink and Zeb poured herself a generous shot of Vodka. Barmaid returned and remained unawares. Ha. Returned home to bed and awoke today ( Sunday ) with headache and aching muscles, for no apparent reason.  Phew! Thank God I\'ve only got one piece of homework to do!  Ceff x   In hope and desperation I\'ll find my only light...   '),(17792,2004217,'17,October,2003','2003-10-17 00:00:00','    Free afternoon! Wheeee! Went home and tidied my room. Boring I know, but it had to be done. And it was fun to boogie around my bedroom to Fall Out Boy turned up loud without being told to a) Go to school  or b) Go to bed. I am also eternally grateful to the Free Afternoon for allowing me to escape from the common room today. Much gratuitous squealing and repeated playing of George Michael\'s \"Careless Whisper\" made me want to stick the stereo up the rest of the year\'s superficial collective bottom. Ha.  Tonight I am out on the town! Well, I\'m going to the Hard Rock Cafe with Daddy Dearest. I\'m not sure if he\'s ever been to a HRC before, because I think if he had he wouldn\'t be taking me there tonight- Â£8 for a cheese & bacon burger... Mwahahaha \"Â£ Â£ Â£\". Tomorrow evening however, (I hope) will prove much more fruitful- I\'m staying over at Oliver(cousin)\'s Uni halls because we\'re off to the Union, yay! Fingers crossed he\'s chosen his friends on the basis of whether they\'re my type. Me? Self centred? Never! I\'M just horny...(ME ME ME ME ME)  Ceff x   He\'s well hung and I\'m hanging on...   '),(17793,2004217,'16,October,2003','2003-10-16 00:00:00','    This blog is immensely irritating. It refuses to publish what I write until the day after I\'ve written it- this can\'t happen to everyone cos Kat\'s doesn\'t do that... Help! Technology is evil! *dies*  Ceff x   '),(17794,2004217,'15,October,2003','2003-10-15 00:00:00','    Event of the day: Erotic fiction! The \"Better Than Cosmo\" area of the Wonderbook (notice proper nounage) somehow  morphed  into page upon page of erotic fiction, plus a rather random \"What turns you on?\" list that Lucy plans to turn into a story. Hee Hee! Much fun was experienced by all during our frees this morning as Liz and I created some naughty stories for our own sordid amusement/pleasure. Mwahahaha.  Event of the day before: Yesterday saw the final, much welcomed episode of the Ceff-Draper saga, when Coz rang him and asked him if he fancied me. Much **FEAR** occurred and I had to Calm myself by listening to the following songs while I waited for the outcome:  Hell Is For Heroes- I Can Climb Mountains Good Charlotte- Waldorfworldwide Something Corporate- I Woke Up In A Car  After YEARS of \":::calm:::**FEAR**:::calm:::*FEAR*\" the message came back...  He doesn\'t fancy me.  I expected it really, and I seem surprisingly un-bothered. He likes someone he works with. I could be seething; green with envy at this WITCH that is stealing my Drapes from me- but I\'m not. I suppose it would be more interesting that way, but I\'ve never been one to be predictable. At least, I hope not!  Ceff x   By the way, your hands were shaking...   '),(17795,2004217,'15,October,2003','2003-10-15 00:00:00','    I\'m good to go And I\'m going nowhere fast It could be worse It could be taking you there with me I\'m good to go But it looks like I\'m still on my own  I\'m good to go For something golden Though the motions I\'ve been going through have failed And I\'m closing on potential towards the wall At a 100 miles an hour  When I say Two more weeks My foot is in the door I can\'t sleep In the wake of Saturday  (Saturday) Saturday When these open doors were open-ended Saturday When these open doors were open-ended  Pete and I attacked And lost the story Oh we promised them decisions The mass of youthful innocence And I read about the afterlife But I never really lived more than an hour  (more than an hour)  When I say Two more weeks My foot is in the door I can\'t sleep In the wake of Saturday  (Saturday) Saturday When these open doors were open-ended Saturday When these open doors were open-ended  And I read about the afterlife But I never really lived And I read about the afterlife But I never really lived  Two more weeks My foot is in the door Me and Pete In the wake of Saturday Saturday When these open doors were open-ended Saturday When these open doors were open-ended Saturday Saturday   \"I read about the afterlife, but I never really lived\"- I wondered about what could happen, but never acted on my instinct- Story of my life?  Ceff x   '),(17796,2004217,'13,October,2003','2003-10-13 00:00:00','    Tuesday, Week 1: Most. Lessons.  Ever.  3-3!- lessons PLUS conversation... how I managed 5 lessons a day when I was a teenie greenie I will never know. I recently worked out that I have not had 5 lessons a day every day for nearly 4 years, due to my (cunning) choice of Private Study instead of an actual GCSE. Mwahahaha.  Today\'s interesting act of the day included sending an \"I want your body\"-type text to GAZ and Coz- No reply from Gaz and much confusion from Coz: \"what the fuk ? lol\" being the abbreviated version. Kat told Liz that I wanted GAZ- which is really quite untrue, I did not want him, I was just... considering options! And now Liz thinks I fancy GAZ, EVERYONE thinks I fancy DRAPER and I don\'t know what to think. *dies* Everyone wants to know what\'s happening in my love life and I just want it all to go away- I could hear either Loz or Chris Devy (or both) shouting \"Where\'s Uta?\" at me down the street- as IF I\'d know! And then he was at the bus stop, and I just didn\'t know WHAT to do, but Helen provided much-needed support and in the end nothing occurred. Arrrrgggghhhhhh MEN!  Ceff x   By sharing these things I rip my heart out   '),(17797,2004217,'13,October,2003','2003-10-13 00:00:00','    Weekend over, homework done and dusted and another successful gig. Went to London with Guides yesterday- had to get up at the ridiculously early hour of 7.30 in the morning so I could be at the coach for 8.45. I didn\'t even know that 7.30 existed on a Saturday! I thought it just... wasn\'t there! Got to London and went on the London Eye, which was very nice and immediately forgettable. Then headed to Wembley Arena for the \"Big Gig\"- 10,000 screaming girls(not one boy- gutted, hehehe), a bunch of fairly well known-ish pop stars and Fearne Cotton from CBBC. It was all fairly boring, the Cheeky Girls being the main highlight, until......   MARK OWEN CAME ON!!  Mark bloody Owen! I had no idea he was on and I was soooo impressed. I started going on about how he was my favourite member of Take That, but none of the people I was with had heard of them... felt so old. Boo Hiss. ButButBut! All was not lost when  LIBERTY X  arrived and were just  Great . They finished their set with Just A Little  (SEXXXYYYY)  and I then immediately went to the toilet to ring Selina and tell her allll about it, while Atomic Kitten were on. Then had to survive a 3hour journey home in between a bunch of hyperactive 12 year-olds and 3 miserable teenagers who kept telling them to shut up. Do they not understand how much I  need  my sleep?? Gah.  Another achievement of Saturday was the acquisition of DRAPER\'s number from Coz. Texted him, flirted a bit, got a reply: No flirt. Alas. Replied, flirted a bit more... No reply so far. Dammit.  Ceff x   And you\'re so innocent, please don\'t take it wrong cos it\'s a compliment...   '),(17798,2004217,'11,October,2003','2003-10-11 00:00:00','    Well, here ends another week of school, and I must say, I\'m glad to see it go- had my first lesson with Mrs. Rogers since  *the incident*  and she said all of one word to me in an entire hour, and it wasn\'t even an English word! It was the Buddhist word for no-self, anatta, and she only said it because she had to tell me what to work on. I\'m not sure whether she was actually ignoring me, or whether she just had nothing to say to me- apparently(according to Rukia, who is  GREAT ) she gave me a \"look\" as I walked out the room. Ha! And today had to do some  actual  work in my free today cos I forgot to do my English homework! Me! Forget English homework! Just  look  at what Mrs. Rogers has done to me. *dies* Last night saw my third venture to WS1, this time accompanied by Laura, Alex, Sel and Zeb+Co.- Loz was there, and apparently-  apparently! - DRAPER likes me! Good. Grief. This is highly stressful, because I have noooo idea what to do. Everyone seems to hate him, and I feel kinda sorry for him, so it might seem a little like pity-dating, but he\'s a nice guy really...arg blerg ick can\'t cope with all these  men ! This all began when Laura told me DRAPER wasn\'t coming into Walsall last night-  Gutted! - And then everyone gave me a *look* and it all got very confusing- Zeb told Draper he\'d broken my heart by not coming into town and now Loz and Cheese (???what???) have decided to make it their own personal challenge to get the two of us together. Help! But we\'ll see. If it\'s gonna happen, it will. From now on, I don\'t care what anyone thinks about me and him. I\'m just gonna keep it to myself and do what I wanna do. *dies again*  Ceff x   This is because I can spell Konfusion with a *K* and I can like it...   '),(17799,2004217,'08,October,2003','2003-10-08 00:00:00','    (Count The Stars- Pick Yourself Up)  Feeling much better now. Mummy is making dinner(chicken...cluck) and lil ol\' me is listening to Count The Stars, who are really good... Discussed GAZ with Liz today and DRAPER with Kat earlier...  1. GAZ seems to be taking on the job of rent-a-boyf (minus prostitute connotations)...\"Available for Special Occasions, comes complete with own suit and spending money\" 2. DRAPER seems to be taking on the role of the guy  E V E R Y O N E  thinks I fancy... No comment!!   '),(17800,2004217,'04,October,2003','2003-10-04 00:00:00','    Well, what a day- I\'m glad we finished early, I don\'t think I could have made it through German last period.   I got some work back off Mrs. Rogers today- 14/20, Level 4 standard, well done Ceffy. What I didn\'t realise, was that Mrs. Rogers had photocopied my essay- she then proceeded to hand it round the class and spend 15 minutes picking it to pieces in front of everyone. Pointed out allllll the bad bits... pointed out 1 good bit... ignored the fact that in her comment (only shown on my copy- everyone else\'s photocopy was covered with a copy of the mark scheme) she had used the phrase \"this is an  excellent  summary\"... Asked me what my last sentence meaned... I explained it to her... her response: \"Well, I still don\'t understand it- does anyone else understand it?\" to which no one raised their hands (Thanks Guys)... then I got a half hearted apology and an admittance that she \"probably should have asked my permission first\". No shit sherlock! Thanks for ruining my day... felt like crap, got angry, nearly  killed  Kat by throwing my keys at the wall and finally calmed down after going to English (SUCH a relief, couldn\'t wait to get there) and eating a bar of Galaxy.   My day was further worsened when my purchase of fingerless black+pink gloves from Etam went horribly wrong because they didn\'t take Solo cards... Next they\'ll tell me Less Than Jake have split up...  Ceff x   It gets worse before it\'s all over, I know I\'ll get through it somehow...   '),(17801,2004217,'04,October,2003','2003-10-04 00:00:00','    Good Grief! I\'ve just realised that \"See You At The Show\" is a Nickelback song! Eeeek! Gonna have to do something about that... I\'ll just use lyrics...  Ceff x   I\'ll sell your memories for fifteen pounds per year, but you can keep the bad days...   '),(17802,2004217,'02,October,2003','2003-10-02 00:00:00','    Aloha!  Last night was most eventful. Went into Walsall with Kat/Coz/Zeb/Laura. It was really good, way better than last time, we had a right laugh- Coz turned into a rap star (50 Pence- you\'ll find him in da pub), complete with bitches (me and Kat) and I flirted with Draper just for the fun of it. Got quite far, he seemed pretty into me by the end of the night, and I totally forgot about Gaz, lol, until Kat mentioned him this lunchtime, at which point I replied \"Who?\" hehehe. But WS1 was soooo cool, Me+Kat were dancing with each other rather (ahem) provocatively, much to Coz and Draper\'s pleasure/amusement/arousal. Every 5 minutes Zeb would drag me away from Draper (Booo, lol) to dance with me and shout \"No!!\" at me until I finally managed to convince her I wasn\'t going to pull him. Draper\'s lovely, but...I don\'t think so... but he\'s good for a laugh, and good for the practice ;o)  So that was yesterday!  Today, dragged myself into school and through my lessons, and then into B\'ham, where I didn\'t have enough time to go to Primark- Gutted! Although I did return a gig ticket and buy 2 new ones- Wahey! Lostprophets and Hundred Reasons baby! Yeeeaaaah, Ian in the flesh- Rarrr... Also bought some lightening bolt/ dice-shaped earrings from Dotty Perks, which I have wanted for a while now, so yay :o) Then had to rush off to Digbeth bus station (beautiful example of architecture that it is) and get the coach to Worcester/ Dad\'s house, which is where I am now... Jimi Hendrix is blaring out of the stereo and Dad is pretty drunk, so just another normal Friday night, hehehe. I want my friends :o(  See you at the show Ceff x   '),(17803,2004217,'02,October,2003','2003-10-02 00:00:00','    Good Evening all.   Today\'s most fulfilling day included 3 almost bearable lessons, 2 highly enjoyable frees, a trip to the passport photo booth, the acquisition of a bass tuner and a shopping trip to New Look accompanied by Coz+Kat followed by a nasty bout of loneliness. They\'re all so... coupley. Happy. Content with each other. And then there\'s me. Royal fucker-up of relationships.   They tried to make me go past Phones4U (where Gaz works), but resisted... I\'m not into all that enthusiasm, I\'m trying not to get all obsessed with him cos it won\'t help, and I\'m not gonna let a guy run my life without even realising it. So I\'m just gonna take it easy... see how it goes. And anyways I\'m going to WS1 tomorrow night so you never know who\'ll be out and about there! Rar... men ;o)  See you at the show Ceff x   # I\'ve got it all, but I feel so deprived... #   '),(17804,2004217,'01,October,2003','2003-10-01 00:00:00','    Lostprophets- Burn Burn    Where do I begin? With this life we\'re living in, Market youth their uniform,  Make us all fit in... But do you think that we\'re that blind? And we can\'t see through all your lies?  Cos it\'s no good (Burn Burn the station) We\'ll burn it down (Burn Burn the truth) Enjoy the flames (In celebration) Enjoy the sound (For us, for them, for you)  I know that I\'ve been told My integrity was sold Priced and placed upon a shelf, It\'s worth its weight in gold... But do you think we\'ve lost our minds? And we can\'t see through all your lies?  Cos it\'s no good (Burn Burn the station) We\'ll burn it down (Burn Burn the truth) Enjoy the flames (In celebration) Enjoy the sound (For us, for them, for you)  Burn Burn the truth, the lies, the news, Burn Burn the life that you can\'t choose, Burn Burn the hate that gets you through, Burn Burn for us, for them, for you...    (x9)  Cos it\'s no good (Burn Burn the station) We\'ll burn it down (Burn Burn the truth) Enjoy the flames (In celebration) Enjoy the sound (For us, for them, for you)  For us, for them, for you... (x5)   This song is so totally amazing it\'s unbelievable- I just want to listen to it all day long, and it\'s gonna be sooo cool live... it\'s totally cheered up my day from my shitty mood earlier  Can\'t wait for the show Ceff x   '),(17805,2004217,'01,October,2003','2003-10-01 00:00:00','    Does anyone have an answer to why I\'m trying to console my ex because he\'s not over his rebound girlfriend? Or why I do this to myself?  Wanting a date to the show Ceff x   '),(17806,2004217,'01,October,2003','2003-10-01 00:00:00','    Well, that\'s the German done with. That was possibly the hardest homework I\'ve ever done... How Mrs. Dalgleish expects me to get an A I do not know...   But on to the HILARIOUS stuff that happened today, namely... Kat. The lunchtime conversation began fairly innocently, as I was eating my yummy scrummy M&S Deli Filler (Prawn and Mayo)...  \"You know... you\'re not supposed to eat those out the tub, you\'re supposed to put them on some sort of carbohydrate\" \"I know that\" \"Jess... can you eat potatoes?\" \"No\" \"Can you eat bread?\" \"No\" \"What about rolls?\"  *look*  \"What about toast? Can you eat toast?\"  *sigh* \"No...\"  The lunchtime progressed, with Kat poking me every SECOND saying \"Jess, Jess, Jess, Jess... Can you eat bread/rolls/doughnuts/iced buns/toast/cake/loaves?\" And ending up being hilarious... somehow. After asking her politely to cease pokeage and askage of  stoopid  questions many times, I had to resort to threatening to eat Kat with my spoon, at which point boyf (Coz) rang, and she was therefore saved from such an edible end. You just wait til next time... no amount of \"meep\"ing can stop me and my  HUGE  spoon... Muahahahahahaaaaaaaa!!!! Ahem....  C ya @ da sho (innit) Ceff x   '),(17807,2004217,'01,October,2003','2003-10-01 00:00:00','    Well! Today has been a  p a r t i c u l a r l y   interesting day. Went for an interview with Marks and Spencer today... passed the first interview section (lots of questions about teamwork and problem solving) and passed the second interview section (practical task)... Which means I got the job! Wahey!   Then I found out I should have brought 2 forms of ID with me, a passport photo and an application form. Do I have any of these with me? Nooo, because my confirmation letter did not arrive until this morning,  after    I had left the house. Great.  Then! They checked their Christmas vacancy timetable and there were no vacancies for me! So I spent an hour in some back corridor of M&S Walsall, only to end up in exactly the same position as I was beforehand. Woo fucking hoo.  Although I did get a reply from Gaz- he says he\'ll try and come out next week, which means I have to try and persuade mother dearest to let me go into town 2 weeks on the trot. Which should be a whole barrel of laughs. Liz even suggested I took him to see Hundred Reasons with me! This is all very interesting, it\'s like she\'s trying to tell me that I  should  go out with him. Which is nice. :o)  So all in all, it\'s been a bad icky blergh day, so I\'m going to drown myself in my playlist and get on with what looks like the hardest German homework   e v e r . I\'ll post more later.  Wishes she was at a show Ceff x   '),(17808,2004217,'27,November,2003','2003-11-27 00:00:00','    So, December is next week, I have a whole Â£8 in the bank and I have not bought one single Christmas present. Tomorrow, thankfully, pocket money will arrive and I will be able to try desperately to find the time in between schoolwork, jobwork, Guides, going to my dad\'s and spending more than an hour at home at any one time to attempt to find presents for people. Thankfully, this year\'s list of big-pressie-people is not that long:  1. Mum Â£10 2. Dad Â£10 3. Em Â£7 4. Bev Â£7-10 5. Philip Â£0- Mother took out and insurance policy and got a free watch 6. Selina Â£25 7. Lucy Â£9 8. Oli Â£15 upwards- shirts cost money!  If I have the time or the inspiration, I will try and get some smaller presents for other people. And Liz\'s birthday present. Oh, and I suppose I\'ll have to get something for Wicked Stepmother aswell...  On top of alllllll this, I am absolutely shattered, therefore grouchy, and have PMT, therefore totally miserable, self-pitying and very, very horny. Grrr.   Thankfully wonderous events such as JYF, Hell Is For Heroes and Good Charlotte are approaching, all of which include seeing men who I either find strangely attractive or come under the category of FIT. But between now and then I will continue to drown in a sea of work, more work, irritability and... did I mention work?  Ceff x   Things can only get better...   '),(17809,2004217,'23,November,2003','2003-11-23 00:00:00','    I am currently experiencing another very odd moment. My darling sister has just informed me that one of the guys (this is debatable) from JYF, the one who likes to cross dress, has officially changed his name to a girl\'s name and started taking oestrogen supplements. This may not strike you as totally freaky; weird, maybe, but not scary from my perspective. But here comes the best bit. Before \"the change\" he was called Robin- after, he/she/it is called JESSICA ROSE, which, if you will remember correctly, is MY DAMN NAME!! I was sooo weirded out when Em told me, I mean, how scary is that?? What if he\'s stalking me? *checks under desk*  Ceff x   My girlfriend is a wannabe...   '),(17810,2004217,'21,November,2003','2003-11-21 00:00:00','    Today I feel... floaty. Not good floaty though; the kind of floaty when you can\'t really decide what to do and where to go- the kind of floaty when you feel you can\'t land next to anyone because everyone\'s doing something that seems far more important than talking to you. I felt very out of place wherever I went, like I wasn\'t quite... accepted isn\'t the word- It didn\'t feel like I fitted in anywhere, but like I was just going around after people, and everything I did or said was slightly innapropriate, or odd. Floaty.  Ceff x   If I was someone else, would this all fall apart?   '),(17811,2004217,'17,November,2003','2003-11-17 00:00:00','    *sigh*  That feeling has returned. That \"Oh-shit-it\'s-Christmas-again-and-I-just-can\'t-be-arsed\" feeling. The Christmas music, the \"Perfect gift for Christmas\" adverts, the generally shitty Christmas weather. Work didn\'t make matters much better either; after smashing a Â£4.99 bottle of red wine over the shop floor by nerves went to jelly and I spent the rest of the night dropping boxes of everything from biscuits to Christmas bloody puddings. And who buys Christmas puddings in the middle of November anyway?? Honestly.  In addition to my Christmas anti-spirit, the general dark cloud that seems to be hanging over school and my love life have made today one big grey blot on my self-centred teenage horizon.   In true geekish fashion, the highlight of my day was my English lesson; although I seem to be the only person EVER to be interested in creoles/code switching/accent and dialect. Oh well. It means I can answer all the questions and look even more enthusiastic than I already am, hee hee...  Ceff x   Gotta have faith...   '),(17812,2004217,'16,November,2003','2003-11-16 00:00:00','    Since Tuesday Richard\'s been texting me. Naughty texts. Rude texts. Downright \"I want you NOW\" texts. Normally, this would make me immensely happy, and for a while I was- until Friday night, when I he was on MSN, when he reverted beck to normal Rich, the one that doesn\'t ask you how you are of answer any of your questions, and generally makes out like he doesn\'t want to be talking to you. Then he goes offline without saying goodbye. Konfused? So am I. I\'ve got no idea whether the whole thing was a joke or whether he\'s found someone else who he fancies, or that girl from Wellingborough has asked him out. I already texted him twice in the last 2 days with no reply, so he probably already thinks I\'m some weird stalker girl and that he\'s changed his mind about getting together with me at JYF.   I don\'t know why I bother, he\'s just living up to his reputation. He decides he likes a girl and then 5 minutes later another one comes along and he\'s gone. I think the reason I keep falling for him is because he dumped me the first time, during the first part of the relationship when you can\'t keep your hands off each other and can\'t stop thinking about each other either. Maybe if we got back together and it lasted a bit longer or was a bit more stable, if he dumped me again it wouldn\'t be so bad... oh it\'s pointless thinking about it because it\'s just not going to happen...  Ceff x   It\'s to dying in another\'s arms... and why I had to try it...   '),(17813,2004217,'12,November,2003','2003-11-12 00:00:00','    So! I nearly got beaten up today. I laughed oddly at Ripu and she told me I sounded like a sheep, so, being me, I \"baa\"d at her. Unfortunately the man walking past at this particular point in time somehow seemed to take offence at this, turned around, spat on the floor and asked me if I wanted a \"fucking slap\". I told him I wasn\'t talking to him but Ripu, and he once again assured me that, had I been talking to him, I would have got a \"fucking slap\". How I love living in Walsall. No wonder I\'m scared about getting 6 rejections from UCAS- I can\'t wait to get out of this shitty town...  Ceff x   I hope you choke on those words that kiss that bottle   '),(17814,2004217,'11,November,2003','2003-11-11 00:00:00','    Where is your boy tonight? I hope he is a gentleman... Maybe he won\'t find out what I know; You were the last good thing about this part of town.  When I wake up,  I\'m willing to take my chances on the hope you hate  him more than you notice I wrote this for you. You need him I could be him I could be an accident but I\'m still trying; That\'s more than I can say for him.  Where is your boy tonight? I hope he is a gentleman... Maybe he won\'t find out what I know: You were the last good thing about this part of town.  Someday I\'ll appreciate in value, get off my ass and call you but for the meantime I\'ll sport my brand new fashion of waking up  with pants on at 4 in the afternoon. You need him I could be him... I could be an accident but I\'m still trying; That\'s more than I can say for him.  Where is your boy tonight? I hope he is a gentleman... Maybe he won\'t find out what I know; You were the last good thing about this part of town.  Where is your boy tonight? I hope he is a gentleman... Maybe he won\'t find out what I know; You were the last good thing about this part of town. Where is your boy tonight? I hope he is a gentleman... Maybe he won\'t find out what I know; You were the last good thing about this part of town.   Ceff x   '),(17815,2004217,'10,November,2003','2003-11-10 00:00:00','    I return to England after a week in Germany in which I barely spoke a word of German and discovered that the German people really are as ugly as everyone makes them out to be. I saw not ONE fit person for the E N T I R E week: Gutted.   But other than that, my 4 days in sunny-but-cold Trier were much better than I expected. Katie Herbert really isn\'t all that bad; the year 11s were all pretty cool and the teachers were really really safe with us. The week began with an average day of school, bar my teachers asking me why I was in school when I was supposed to be in Germany. Selina arrived at my house at around 10pm that night and we proceeded to watch The Little Mermaid and Moulin Rouge until it was time to go. As usual, we were the last ones to arrive, and then the journey began....  At Toddington the coach driver got off to be replaced by his son Jamie- but more on him later... by 7am we were at Dover and by 5pm (German time) we were in our rooms unpacking. After a meal that consisted of what seemed to be a wedge of SPAM and some mashed potato, we (German A Level group and teachers Charmaine, Mr. Birchall and Mrs. Dalgleish) went to the lobby/cafe type place and played poker with building blocks, under the supervision of Katie Herbert, the only one of us who knew how to play.  Wednesday we met the German class. There is one word to describe them: odd. Amongst my group were Benjamin, who stroked his \"beard\" every time he was thinking and had fuzzy hair, and Florian (yes you read that right), who was on medication for his hyperactivity disorder. Eek. Other Germans who will live on in our memories include:  Benedict, who was in love with Varsha and (according to her) had clammy hands, Joaquin, who thought he could rap, Red T-shirt guy, who everyone fancied but whose name Selina and I can\'t remember, and Jan, the weirdest boy I have ever met. He hung around with us for hours after the rest of his class had left, tried to make us go to his Harry Potter party and was in love with Davinder. I made everyone check under their seats when we got on the coach to make sure that he hadn\'t followed us on.  Wednesday morning was also whe Amy decided to collapse in the middle of the breakfast room. She was taken to hospital and kept overnight, and sent home Thursday morning with nothing wrong with her. By Friday, according to Mrs. Dalgleish, the chinese whispers around school included: \"Amy\'s in a coma\" \"Amy has epilepsy\" \"Amy has Toxic Shock Syndrome\"  LOL.  The rest of Wednesday and Thursday came and went in a blur of interviewing teachers for coursework, avoiding Jan and most importantly, WHOLE AFTERNOONS OF SHOPPING! Woohoo. Trier is such a pretty town with the coolest shops, including a MASSIVE H&M which we promptly raided. I purchased a \"very me\" wide necked pink and black stripy top for 13 Euros.  On Friday we went to see a film, in German with no subtitles... *sob* It was about a little boy whose father came home from a prisoner of war camp in the summer of the 1954 world cup, which Germany won, apparantly. It was fairly good but the boy just would. not. stop. wearing Lederhosen and I exited the cinema thoroughly disturbed.  Saturday was our \"day off\". Not that we\'d done anything but have days off all week, mind, but Mrs. D doesn\'t need to know that... we spent all our remaining euros on Diddle merchandise, drinks and cake and then headed to a hotel for a farewell party with the Germans, which consisted of a selection of traditional German food for us to try and some English food for the Germans to try. After that we headed back to the hostel for our final evening in the bar, which was virtually ruined by apperance of Mrs. Ds ex-husband, who was there to see their daughter. He put a dampener on our whole game of poker by complaining about it all the time and eventually we all headed our separate ways, leaving the three of them alone.   Our evenings were spent in the \"Cafeteria\" downstairs in the hostel. Mrs. D, MC Mattie B/Sea Scout/Mr. Birchall, Charmaine, Jamie, Katie, Becky, Selina and I whiled our evenings away with poker that we bet on with matches, which proved to be the highlight of each day. The atmosphere was so relaxed, it was really cool and we all enjoyed ourselves- Mr. Birchall would deal, Katie would be telling Becky and Mrs. D whether to fold or to bet, Selina would be yak-yakking away to, well, whoever would listen and Jamie and I would laugh at it all while he tried to steal everyone\'s matches. Jamie was waaaayyy the coolest person on the holiday- for a start, he was from Barnsley, so obviously he commanded immediate respect from me (who incidentally was the only person who DIDN\'T take the piss out of his accent). He was sooo sweet and cool and nice and chatty and just great. No I didn\'t fancy him. Well, maybe a little. I wish he was here cos I loved talking to him so much and I feel sad that I probably won\'t ever see him again. :o(  On Sunday we boarded the coach at the RIDICULOUS hour of 8am (bear in mind before this we had been up packing final things, clearing our rooms out, having breakfast AND making our packed lunches) and by 8.30pm we were back at school. We said our farewells, headed home, and (surprise surprise) Selina didn\'t come into school today. Hmph.   So now everything\'s back to normal. I was ready to come home yesterday morning, but after a day back in wet Walsall I wish I was still in Germany, where I can play poker, talk to Jamie, call Mr. Birchall names and gossip about teachers.   Ceff x   There\'s a light on in Chicago, and I know I should be home...   '),(17816,2004217,'03,November,2003','2003-11-03 00:00:00','    So, another half term over, another Monday done and dusted. In roughly 4 hours I will be boarding a coach to Germany. I have spent the last few hours packing and Selina has just arrived so I\'ll have to be quick- until Sunday!  Ceff x   '),(17817,2004217,'28,December,2003','2003-12-28 00:00:00','    So it\'s the 28th December and already Christmas feels over. Normally I\'d still be basking in the warm afterglow of eating too much and playing with all my fun new presents. Instead, we\'re back at home already and the house seems strangely under-decorated, despite a 6-foot sparkly Christmas tree standing less than 5 feet away from this keyboard. This year, the whole affair seemed decidedly un-festive; Yuletide sort of just sprang upon us, and although the presents were wonderful and everyone was happy and the food was great, those couple of days that should be happy and content seemed to lack that (forgive the pun) Magic and Sparkle. I was sat in the canteen at work yesterday (I\'ll come onto yesterday in a minute) and they were telling all these stories- someone\'s son had proposed to his girlfriend on Christmas day, another\'s daughter had a baby girl on Boxing Day, and I just sat there, all quiet. I felt like Stanley Bagshaw- nothing very exciting ever happens to me.  Work was hell yesterday. Not hell how you\'d imagine it though. It was much, much worse than anything you could ever imagine. Screaming babies, loud grannies and refund after refund after refund. Add to that the post-Christmas sale-goers, and the fact that I was on my feet  all day , and it really felt like someone had taken what little Christmas spirit I had left, strangled it, dropped it in a pile of dog shit and then kicked it into a busy road. But, I hear you cry, you work on foods! You sit down at your till points, and none of your products were in the sale! Well that\'s where you\'re wrong. For the whole of yesterday I was loaned out to fashions, because they needed more staff to deal with the mob. The tills on fashions don\'t have seats, and they had one mother of a sale. I\'m surprised I\'m still alive to tell you about it.  But! Back to happy Christmas things. Went to Dad\'s for Christmas No.1, and got 2 CDs (Rancid and Michael Jackson), my new coat (snuggly warm yay), big woolly jumper, book of Politically incorrect jokes from the net (GREAT) and book called The Curious Incident of the Dog in the Night Time, which is a murder mystery story written through the eyes of a 15-year-old boy with Asperger\'s Syndrome. It\'s brilliant- couldn\'t put it down and am currently fighting the urge to read it again until I\'ve finished my other Christmas books.  After 2 days of work (busy busy busy, but KERCHING! Overtime), Christmas Eve occurred and we headed up to Barnsley. Nothing much happened. Christmas Day then arrived and as usual, my sack of presents was sat under the tree (after many, many hopeless attempts from my family, I remain adamant that my mother put it there, after seeing her bring it out to the car before we set off). Got some really wicked stuff, but the top three have to be my new Paul Frank t-shirt (Julius the monkey is wearing earmuffs!), Eats, Shoots and Leaves- A Zero Tolerance Guide to Punctuation by Lynne Truss (brilliantly funny- a whole chapter dedicated to the apostrophe) and of course, my MiniDisc player. Of course, my mother being my mother, she forgot to buy any minidiscs and we had to wait until Boxing Day to go to Currys and get some, but we got there in the end and it\'s still fantastic. I got the new SoCo album aswell, it\'s brilliant. Mar Hon sounds... hot, to be quite honest. If he sang any of those songs to me, I would either a) melt or b) orgasm. They\'re just fantastic.  Tomorrow morning myself, mum, Selina, Aunty Bev, Les (mum\'s mate), Coral (Les\'s mate), Ann (Godmother) and Ellie (Ann\'s daughter) are heading off to Gay Paree for some New Year fun and frolics. It should be fun, as long as Bev doesn\'t keep going on about my dad. Here\'s the background to this problem: Wicked Stepmother has changed her last name from Maclean, her old married name, to Morris-King, a combination of her maiden name and Dad\'s last name. At the wedding, at the end of the ceremony, the registrar said \"Can we have a round of applause for Mr. and Mrs. Morris-King\" which is wrong, because Dad is just staying as plain old King. Of course, I made the mistake of telling Bev this story, and she\'s now latched on to calling him Morris King. You might laugh now, but try hearing that every ten minutes for two days straight. Not funny anymore. And of course, this week she will tell Selina this, and then she\'ll pick up on it, bring it back to school and then Laura will pick up on it, and you get the picture.  Ceff x   I could still be ruthless if you let me...   '),(17818,2004217,'19,December,2003','2003-12-19 00:00:00','    Horrid weather, horrid day, horrid person. I came to school cold but chirpy this morning, you know, last day of term, giving presents out, getting presents back, having a dance around the form room and playing silly games with my mates. I really wanted to give Sel her pressie because I knew she\'d like it and I couldn\'t wait to get my new underwear off Debbs. Problem number one: Sel decided she didn\'t want to come to school today. We rang her at home at about 9.45am and she\'d just woken up and \"wasn\'t feeling very well\". We eventually managed to persuade her to come in, and when I finally got to give her her present she came across as quite ungrateful. \"You did get me extra small, didn\'t you?\" Me: \"No I got you small to medium, sorry, do you want the receipt so you can take them back?\" \"Yeah look I need extra small, they\'re the right ones and the right colour, but they\'re too long. Give me the receipt and I\'ll take them back and change them.\"  No \"Thanks anyway\" or \"Sorry, I should have told you to get me that size\"- I found it really abrupt and it got me down a bit. She\'s supposed to be my best friend, for God\'s sake, and I felt really distant from her. She went home at about 11.30am.  After school finished at 1pm (woohoo!) Helen and I passed Owen Ingram tied to a lamppost outside school and went to the Wharf bar for a drink, then we both went to our respective workplaces. I discover that both times I\'m in work next week I will be unpaid! Not a happy bunny. I\'ve now worked out that I have only taken 4 hours over my allowed leave time, which I have already made up, plus 2 and a half hours extra. Fair enough, don\'t pay me for the time I\'m not there, but not paying me for the time I am there? Not on at all, especially when I\'m having to cut short my Christmas jaunt to Barnsley to come back and work in that hellhole for 10 hours straight. The bastards should be paying me extra, if anything.   Ceff x   Coming down, the world turned over; angels fall without you there...   '),(17819,2004217,'18,December,2003','2003-12-18 00:00:00','    So, last night came the evening of the Christmas Party. By this time it had already been confirmed that anyone who I planned to go anywhere with on that night was going to the Boy\'s School Christmas party at Revolution, until, in true bitchy girl style, \"someone\" (i.e., according to rumour, Pav) contacted said bar and informed them that we were underage. Shit. So, all manner of confusion occurs, including the following questions:  Are you going to the Boundary instead? No. Is it definitely at Revolution? No. Is there still going to be a party? Yes! Where is it? Don\'t know. What does Hemal say? Apparently it\'s at Stein Bar now. So it\'s definitely at Stein Bar? No. Shall I just meet you in Imperial? Yeah.  The common room became a hive of phone calls and text messages, and nothing became clear until Hemal finally appeared in Imperial at 8.45 (the party started, in fact, at 8pm, but I think we were being \"fashionably late\" or something) to tell us that we were going to the Stein Bar between 9 and 9.30pm, because that was when the football finished and the rest of the boys\' school would leave Bar Sport. Riiiight. Eventually, we arrived at Stein Bar and a generally good time was had by all, from what I could tell. Gaz tried to pull Bex but did not succeed as we were leaving, Rachelle Horton\'s boyfriend tried to feel me up  twice . and I met J(ay), the sweetest guy ever to exist. He\'s small but perfectly formed and generally a really nice guy- he complimented me on my dress (incidentally, he was the only person to do so), squeezed my hand and made me smile :o) Awww. Fab.   '),(17820,2004217,'16,December,2003','2003-12-16 00:00:00','    Let\'s conduct an experiment. Take a few of my favourite things, namely gigs, music and Benji Madden, mix them all together and see what happens. What actually happened was near enough 2 hours of sheer bliss. I can\'t really explain what it was like last night; there aren\'t any words to describe how totally amazing and mindblowing it felt to have one of my favourite bands standing, performing, less than ten feet away from me. Just... wow. Totally and utterly fantastic in every sense of the word- the songs were all brilliant, the set list was a perfect mix of the two albums and I knew every word to every single damn one of them. Their stage prescence, their musicianship, their relationship with the crowd, it was all perfect... just perfect. Everytime I think about it I smile, I just want to go back there and watch them play over and over again- watch Benji leap around the stage, watch Joel get beer thrown at him, watch Paul swear at the lighting guy and watch Billy act generally skeletal. They opened with A New Beginning, as predicted, and finished with Lifestyles, which kind of goes without saying. I was totally surprised when they played Waldorfworldwide, I only expected them to play the singles from the old album, but like I said, they mix of stuff they did was fab :o) because they\'re Good Charlotte, and they know how to do everything exactly right every single time. But, I hear you cry, are they as fit in real life? My answer: Alas, no. Benji, obviously, is still hot. Joel, despite his best efforts to look punk with his red mohican, still looks ugly compared to his twin- yes I do know how ridiculous I sound. Paul looks like a big puppy dog, bless him, and Billy is actually quite rank in real life, really really skinny and quite scary, to be honest... but his guitar playing made up for all his lack in flesh :o)   I\'ve had my walkman attached to my ears virtually all day just playing Young And The Hopeless over and over again, only to be interrupted by the carol service, which was, as usual, a ridiculous state of affairs due to its general pompousness- \"oooh, look at us, we got the TOWN HALL for our carol service... listen to our SENIOR VOCAL ENSEMBLE sing Away In A Manger, isn\'t it pretty? Look at all these LOVELY, INTELLIGENT girls who can READ\" Big whoopee. Get over it. And don\'t even get me started on you and your ego, Mr. M-J... whoever asked him to read the part of God made a  huge  mistake. I\'ve re-realised the beauty of Emotionless. It\'s the only song that\'s ever made me cry; it\'s so fragile- the simplicity of the guitar in the background makes you listen to the lyrics, which, and if you know any of what\'s gone/going on in my life you\'ll understand, really hit home hard for me. Hey dad...  Ceff x   When you lay your head down, how do you sleep at night? Do you even wonder for a while if we\'re alright? Well we\'re alright...    Life, hope, truth, trust, pain, pride, love, lust, on without the things we\'ve lost, the things we\'ve gained we\'ll take with us...   '),(17821,2004217,'12,December,2003','2003-12-12 00:00:00','    Wow. What a gig. As ever, Hell Is For Heroes came out onto that stage to deafening screams and played. Just played. Upfront, ballsy, fuck off rock songs that, in my humble, obsessed opinion, will live on forever. And they weren\'t even at their best; it\'s a perfect measure of their talent when a gig that could have been better was still fantastic. They began with I Can Climb Mountains- in my opinion it probably would have been more effective leaving it until last, but who gives a shit about the set list- all the songs were just... wow. Even the new stuff sounds fantastic- I feel like I know it already. There\'s this one song, Folded Paper Figures, and it\'s brutal- I wanna listen to it over and over again, it\'s just 4 minutes of anger and spite combined with sheer musical talent. Mindblowing. The support weren\'t bad either- first band Copperpot Journals were very good, very talented, but didn\'t quite grab me. Second band was The Lucky Nine- as soon as I heard the name I knew I recognised it from somewhere but I couldn\'t put my finger on it... then Colin bloody Doran from Hundred Reasons and Dan Carter from \'A\' walked onstage. My jaw dropped. It\'s this new Britrock \"supergroup\" that Kerrang! have been going on about for a few weeks- in a word, metal. Not my scene, but still talented.  And so now, the countdown begins... 3 days... until it\'s time...  Ceff x   I can build bridges from ones you burn...   '),(17822,2004217,'10,December,2003','2003-12-10 00:00:00','    I\'m wearing a new dress that I bought today; It\'s knee-length, black satin with a ribbon under the bust and little pink polka dots. I\'m also wearing an early Christmas present from daddy: Â£22-worth of pink hat from Oasis, plus my plain old black heels. In 2 hours I\'ll be going to see Hell Is For Heroes and I have no intention of taking this outfit off until the last possible opportunity. Why? Because I feel great- I looked at myself in the mirror and thought \"Damn... I look gooood!!\". So fuck you Birmingham, and fuck you, Rich, because I\'m sexy and you both just missed out...  Ceff x   She likes the way I stare...   '),(17823,2004217,'09,December,2003','2003-12-09 00:00:00','    So much has happened I don\'t know where to start... I\'ll try the beginning...  So last week was full of the build up to JYF; the nerves, the excitement, and most importantly of all, planning outfits. Until Thursday, when Birmingham sent me a letter telling me they didn\'t want me on their course. As I sit watching my life going down the drain while simultaneously holding an MSN conversation with the JYF committee, the novelty of the whole affair began to wear off. But eventually Friday came round and by 6pm I was bouncing around in the car, desperate to get to Bournville. I arrived at roughly 6.15 to discover that only Edd and Lailah had arrived, and that Rich and Em were at the supermarket. Caught Edd looking at me in my fabby outfit, but didn\'t mention it to anyone- got that little glow inside that makes me somehow taller and gives me that \"Woohoo! I\'m great\" sensation. Rich eventually arrived and immediately told me I looked wonderful and gave me a hug, kissed me, touched me... it was wonderful. I felt fantastic; this guy that I like, he liked me back... And it carried on: the whole of Friday was spent with him, culminating in Epilogue, when, snuggled up in our sleeping bags, things started to get really naughty. Of course, after a little bit of naughtiness in the dark, as soon as the lights went back on and the candles were blown out, he rolled over, stole my pillow and went to sleep. As you can imagine, I was somewhat put out, and so went over to Matt. Matt\'s wonderful; I just want to cuddle him all day long. I miss him so much, I just wanna see him all the time...  Friday night wore on, and as Matt and I snuggled and Richard slept soundly, I finally relaxed and got over the fact that, once again, I\'d been stupid enough to fall for him when there was absolutely no chance of us ever having what passes for a relationship. I had a fab time talking to Edd, Ed, Matt and Lailah about the most random things, which is one of the reasons I go to JYF. It\'s the perfect excuse to just talk shit for 2 days; no one judges you, no one calls you stupid or gives you funny looks, they just laugh along with you and give you a hug. So Friday night passed into Saturday and Richard woke up. And, as expected, he acted like nothing had happened and decided that he fancied Em. Of course, she was fucking lapping it up, hanging off him like a sick puppy, boobs in his face and skirt slowly riding up her legs. But Matt was there, and he held my hand and gave me hugs and leant on me in the cinema and kissed me and most importantly of all, he made me smile.  Saturday evening arrived, and Kat left the meeting house for a party. On the way out she got hit by a car and, unknown to us, was rushed to hospital. Of course as soon as I found out I was devastated- and Rich was the nearest person. So he hugged me, and kissed me, and I thought \"maybe I\'ve got him back\"... of course not. After 5 minutes he complained I was squashing him and went \"for a walk\". Towards Em, who was also \"devastated\" because she\'d been outside and seen the accident and not realised it was Kat. She\'s got some illusion that Kat actually likes her, just cos once she had sex with her. In actual fact, Kat finds her immensely annoying, self-centred and slutty: just like I do.  Matt and I slept most of Saturday night. My slumber was only interrupted by ridiculously loud snoring from Ex boyfriend No.1, Edd. I didn\'t know where Rich was and I didn\'t care, because Matt was there. And he was snuggling me and it was warm and fun. So I awoke on Sunday morning to the sound of Christmas songs, and he put his arm around me and we discussed the merits of each song that came on, before finally getting up for breakfast, packing away and a long goodbye before he had to leave before meeting to catch his train. As soon as he\'d left I missed him- I just wanted him there like he was, all snuggly and smily. As I wrote on his Rosy Glow, he made my weekend, just by making everything seem fun again even though Rich was being a twat as usual. And of course, the comments on my figure from everyone around me helped quite a bit. A classic came from Sam: \"You\'re looking very aerodynamic, Jess\". And this is why I love JYF! Sean was also highly cuddly during the weekend, and told me I looked good too, as well as Kris, who, since I last saw him (when I had a huge crush on him) has turned fucking  g o r g e o u s: tall, dark, glasses, spiky hair, intelligent... Mmmm! Yet, here I am, falling for the stocky, scarred blonde with a fixation with my sister. So overall, when you balance it out, although Kat had an accident and Rich screwed me around, the weekend was fabbity- I took every opportunity to take the piss out of Em\'s sexual habits, as did Matt B, another new friend :o).   As soon as I got in the car when my mum came to pick me up, I missed everyone and everything about JYF. It\'s like it\'s the real me that\'s there, totally relaxed and happy. So Sunday night was spent in a haze of memories and an aching in my belly, part for people and part due to my period, which evilly arrived on Sunday morning. Last night I went to work and barely spoke a word- I didn\'t feel like talking to anyone because I just wanted to go to JYF and give someone, anyone, a hug and a kiss and talk crap for 4 hours, rather than stacking box after box of belgian chocolates in a cold shop. Today was OK, I\'m still thinking about it but not missing everyone so much due to the realisation that e-mails exist. It\'s like a big drug trip that takes days to come down from, but somehow the comedown is almost as good as the trip itself.  I\'d e-mailed Birmingham back on Thursday, asking them if they could give me more details about why they didn\'t want me, and that I didn\'t want the alternative course they\'d offered me (Eng Lang with German, including a year abroad- FEAR). This is what I got:  English/MCS is extremely competative and we look for an interest in, and an awareness of, isues related to Media, Cultural Studies in the Personal Statement on the UCAS form.  This was not the case with your application.  I am afraid that is is not possible to study English Language as a Single Honours degree.  Thanks a fucking bunch! If you\'re going to reject me you could at least get the grammar and spelling right. And I know I\'m only picking on that in a desperate attempt to regain some dignity from this horrible affair, but for God\'s sake, it was my first choice and I\'m breaking up (or down?) inside. My first rejection. I... I don\'t know. I just don\'t know.  Ceff x   '),(17824,2004217,'02,December,2003','2003-12-02 00:00:00','    That feeling returns. That feeling I get before my period when I feel really low, really down, like I\'m all alone, no one wants to talk to me, they\'ve all got something better to do, someone better to see; something more interesting to say.  One thing that\'s keeping me going is that Rich seems to be paying me some attention again- and telling a miserable, PMT-ridden girl that she\'s beautiful and damn good-looking isn\'t exactly the wrong thing to do... He wants to take Friday night off work to come and see me, which is sweet, but it\'s made me all dreamy. This tends to happen when I think there\'s a chance of me getting with someone; this time I see myself striding confidently into the Meeting House (somehow my bag, bedding and the art box have made their own way in) wearing my fabulous new outfit and walking past Richard as he stops, turns, stares etc. and eventually I return (after doing something irrelevant to pass the time) and he hugs me... and then he kisses me... and it becomes the start of some beautiful relationship where he never notices another girl (although most of the girls at JYF are prettier than me) and we can talk and have fantastic sex (even though his Peeny (Nod to Kat) is tiny) and go to gigs and he\'d be original and agree with me that flowers are a cliche and he\'d remember that I can\'t eat chocolates and that I think jewellery is boring and buy me a CD or a poster or something brilliantly original that only this perfect boyfriend would ever think of *breathe*. See why I hate love? I let my mind wander and then I find myself disappointed when Luke buys me a rose because it\'s \"not original\". I should be pleased that he\'s at least thinking about me.   But why settle for satisfied when you could be... happy?  After such a ridiculous outburst of stupid girly feelings, I will clarify that after such mental diarrhoea (sp?) I tell myself to stop being so stupid and that that will never, ever happen to me because it\'s way too far-fetched. What will normally happen is, I\'ll walk in, lots of people who I haven\'t seen in ages will come up to me, hug me, tell me how much weight I\'ve lost, and he\'ll see me, think \"WHY??\" give me an uncomfortable hug and then follow Lucy Gemmell around for the whole weekend. I\'ll let you know.  Ceff x   So alone... Sometimes I swear that I can hear the dawning of a thoughtless world   '),(17825,2004217,'31,January,2004','2004-01-31 00:00:00','    Mother Dearest and I went to B\'ham today to the Turner Exhibition at the Art Gallery, and then went shopping and I bought my very first pair of CHUCK TAYLORS!! Eeeee they\'re soooo pretty, classic style but white with paint splodges all over them... soooo nice I want to wear them forever and ever. Which reminds me, I\'m going to have to put my jeans in the wash soon... *sniffle*  Got a letter from the CSA today saying that Dad should be paying us exactly what he has been paying us for the last year. Well Â£9 more, but who counts pennies? Dad counts pennies, that\'s who! I got an e-mail from him when I got home saying that he\'s going to cut my pocket money by Â£9  and  make me pay my own phone bill because of this. My phone bill comes to roughly Â£10 per month, so when you work it out, for the extra Â£9 he\'s got to give the CSA, he\'s taking Â£19 away from me. Mum says I should just tell him to stick it all up his arse and say goodbye; refuse to see him again. Admittedly, it\'s tempting, but for some reason I can\'t. I don\'t think I\'d want to be one of those kids with only one parent, it seems very scary. She apologised for giving me such a shit father as well. It amazes me how my parents are so different, when you think about it. She would never, ever dream of treating me like he has... when you cut to the chase, you can see that he hasn\'t thought through what he\'s doing, he\'s just penny-pinching and that no matter how many times and to how many people he says that I\'m the most important thing in his world, I\'ll always know I\'m second best to his own personal interests: What\'s best for him? How can he get as much money as possibly from his bitter ex wife and child? Fuck you. Fuck you. Even shoes can\'t help me now...  Ceff x   I often wonder how you feel when you\'re lying (next) to me...   '),(17826,2004217,'30,January,2004','2004-01-30 00:00:00','    I was waiting for Elena outside the locker rooms today at the end of school, and I realised how bored I was. Everything is still the same as it was when I came to school; people have the same arguments, the same problems, they hate the same teachers; and no matter how many new buildings they build, that place will always, always be the same. And it\'s boring! I can see cycles being repeated over and over again in all the years, and I can see people in my year who\'ve never changed. And don\'t even get me started on this town! Full of fucking nutters, teenage mothers and townies- nothing original and certainly nothing new. This school, this place will always hold happy memories and I will, of course, treasure the friends I\'ve made, but right now I can\'t wait to get to university.  Ceff x   We gotta get out of this place...   '),(17827,2004217,'29,January,2004','2004-01-29 00:00:00','    Hurrah! I have finally got rid of Kazaa. After one too many virus scares, rubbish search bars and bloody pop ups, I downloaded Win MX, which, although rather dull, comes with no extra software and does what I want it to do: download a song into My Music folder. I then deleted the whole 195 MB (biggest file EVER) of Kazaa and celebrated by downloading the new Lostprophets single successfully (on Win MX) for the first time. It\'s not a looped file, it\'s not Burn Burn under a different name, it\'s not quiet, it\'s not cut short, it\'s not live. It\'s not Kazaa!! Yippee!!  Ceff x   Love was once a part but now it\'s disappeared...   '),(17828,2004217,'25,January,2004','2004-01-25 00:00:00','    HMPH. Much to (I assume) everybody\'s delight today, school was closed due to the snow. That\'s one good thing about it I suppose. It snowed on and off all day yesterday, except for the 20 minute period when I walked from school to the bus stop, where the temperature rose and the snow turned into huge great dollops of rain water, thoroughly soaking me. Thanks a lot. Other than that, yesterday was a pretty normal day, except that some cheeky year 10 or 11 made fun of my earmuffles. Stupid girl, does she not think I can hear her when I\'ve got them on? They don\'t have music playing in them, dumbass. Although that would be cool...  HMPH. Valentine\'s Day is fast approaching and I am fast becoming more and more miserable about the whole red-coloured affair. No boyfriend, no crush, no life.   HMPH. Richard can\'t come to the JYF committee meeting. How selfish of him to deny me the perfect opportunity to decide to be mean to him but then not go through with it.  On the bright side, I have the whole of today to do lots of interesting things like my English coursework. Woo.  Ceff x   Game over, it\'s no fun...   '),(17829,2004217,'25,January,2004','2004-01-25 00:00:00','    Oli rang me on Thursday, and we had a chat and as I put the phone down, the relief that spread through my body was just unbelievable. He\'s fine. He was just stressed about finding a new flat for his 2nd year because he felt like his flatmates all had loads more money than him. Phew!  So, on to cheerier things: 2 gigs this weekend! Friday night saw the return of Hidden In Plain View and (swoon) The Early November, this time at the Little Civic in Wolverhampton, the only local gig venue I hadn\'t been to besides Flapper & Firkin. Hidden In Plain View were much improved from their performance at Academy last week, the practice must have done them some good! Yourcodenameis: Milo (yes that is their name), however, were not such a great prospect. They had decent enough songs, but they were possibly the least charismatic band I have ever seen live. Ever. They just sort of... stood there. The Early November, of course, were fantastic- witty, sociable, in time and armed with a bunch of really cool songs, they get better and better every time I see them.  The Little Civic is the oddest venue I\'ve ever been to; for a start, it has carpet, and the walls are covered, several times over, by tour posters. Someone\'s written \"STUPID TOWNIES\" on the So Solid Crew one. Ha. And someone drew glasses and a moustache on Amy Lee\'s face, which also amused me greatly.  Went shopping on Saturday for much needed jumpers, after I realised that the only ones I have are giant ones that you wear on a Sunday morning when you\'ve got your period and haven\'t washed your hair. Bought three; Stripy one from Gap for Â£6.99 (reduced from Â£36, Kerching!), and two others from H&M, one stripy and one plain blue. Also got some new Dreadys from Editions, proper ones with a fixed waistband and a fly! *silly grin* They\'re soooo nice and comfy and I will wear them all the time for ever and ever. Obviously I will have to wash them at some point, but we can cross that bridge when we come to it...  Went to the final local gig venue I\'d never been to, Flapper & Firkin, last night to see Edd\'s band Echidna play live. They were surprisingly good, quite punky and tuneful but also, unlike other local bands I\'ve seen, in time with each other. Kat has just informed me that one of Sean\'s friends fancies me and I cringed inside every time Kayla mentioned Rich, but other than that it was a good evening. Went to Hard Rock Cafe, where I ran into Draper while waiting outside for my mum. Gave him a hug for his 18th birthday and was puzzled by his purchase of a Â£9 cigar. Oh well.  So, all\'s well that ends well and everything is back to normal. Dammit.  Ceff x   Don\'t get me wrong if I\'m acting so distracted   '),(17830,2004217,'22,January,2004','2004-01-22 00:00:00','    I can\'t concentrate. On anything. Not school, or TV, or anyone\'s conversations. I check my phone every 2 minutes (shit! where is it?? Hang on...... got it) to see if I\'ve got a message, a call, anything. From Oli. I got an answering machine message from him one morning saying he needed to talk to me about something because he trusted me, and that neither his mum nor mine could know about it. I\'m so scared; what if he\'s in trouble? What if he\'s got something wrong with him? At first I thought it was a hoax, but after 2 texts and 2 voice messages I began to wonder. I rang him on Wednesday morning to tell him I was here and he could call me today, but he hasn\'t... I\'m so scared. I keep trying to tell myself it\'s probably something not so big: maybe he wants to change his course at uni, maybe he accidentally kissed Paul, maybe he and Victoria broke up. Maybe him and Victoria broke up because he accidentally kissed Paul... but naturally I keep coming back to the worst of the worst: he\'s got cancer, he\'s got AIDS, he\'s murdered someone, he\'s dealing drugs, he\'s been rap... No. NO NO NO.  Scared x   I\'m holding on for you.   '),(17831,2004217,'21,January,2004','2004-01-21 00:00:00','    Today\'s pop-ups seem to imply more failure in my life:  \"Failed UCAS? Try one of our online degree courses for people too stupid to figure out how to open the door and go to their local college! So easy to sign up for, and only $93482939958686 per year!\"  \"Hey you, fatso! Your diet clearly isn\'t working. Try ours instead! For just $29495930059393 dollars a month, you can\'t afford not to (you fat fucker)!\"  Ah, how I wish I could live the American dream...  Ceff x   Nicotine, valium, vicodin, marijuana, ecstasy and alcohol... c-c-c-c-c-cocaine...   '),(17832,2004217,'19,January,2004','2004-01-19 00:00:00','    Pop-ups keep, well, popping up at me. Ones that say \"find your perfect love match at I\'m-desperate.com\". Are they trying to tell me something?  Ceff x   Just once, I think I\'d like to see you get it on with me   '),(17833,2004217,'18,January,2004','2004-01-18 00:00:00','    The typical has happened. I\'m thinking about Sean. 2 days of over analysing his behaviour at Home Grown and telling myself I don\'t fancy him has made me start *thinking*. Waiting for texts. Going online to see if he\'s online; Going offline if he\'s not. And I don\'t fancy him! I know I don\'t, it\'s just that I don\'t fancy anyone else at the moment and he was so nice to me on Tuesday... argh I\'m so useless! All they have to do is *lean* and *hug* and tell me I look great and I turn into a big mushy pile of soppy goo on the floor. He probably doesn\'t even fancy me, he didn\'t before and I haven\'t changed, personality-wise, so it\'s not like suddenly he\'s decided to fancy me. Or have I changed? Is it something to do with me looking considerably better than I did a year ago (albeit still nothing very special)?  Ceff x   And I wonder if I\'m alone in your head...   '),(17834,2004217,'17,January,2004','2004-01-17 00:00:00','    It\'s been another odd day. Not because weird things have happened, but I\'ve just felt weird, you know? Everything seems to affect you more or seem generally more imposing than it normally is, even colours. I woke up at about 9am due to the slight problem of not being able to breathe properly because my nose felt like someone had shoved foaming plughole cleaner up there- sort of tingly but with no room for anything to move. Dad and Wicked Stepmother left the house at 11.30am and 10am respectively, leaving me and Em in the house together. I don\'t know why, but often I dread being left alone with her. I know I can cope, there are plenty of distraction techniques if she gets too much, but I just feel uncomfortable. We walked to the pub to meet Dad and Wicked Stepmother; I wanted to go on my own really, but there was no chance of doing that.   Our afternoon\'s activity was going to see Lost In Translation at Warner Village; after all the hype and stuff it wasn\'t as brilliant as I expected it to be. Sure, the acting was good, the setting was really appropriate, the script was sharp and different, but there was something missing, for me anyway. I think I\'ve finally worked out what \"type\" of films I like, and it\'s the ones with that child-like element of dream and imagination; Peter Pan was wonderful for that, everything was so innocent and at the same time your imagination ran wild without digressing from the plot. The same goes for Harry Potter and even Velvet Goldmine, although there was nothing child-like about that. But anyway, I\'m waffling: Lost In Translation was an excellent film, Bill Murray is fantastic, but I think it was a little too... human for me.  I had an odd ten minutes after that, where I didn\'t want to talk to anyone. That was when everything became imposing; I made the excuse of going to the toilet just to get away from it all, but the cubicle walls seemed to lean in on me so I got out. Em sped off because she got all moody and Dad was just too huge (physically and psychologically) an entity to comprehend and as I would never, ever go to Wicked Stepmother for any kind of mental comfort, I was a bit stimid. I came round though; I always do.   Dad and WS have gone out tonight with some friends of theirs for a balti in Worcester, and Em and I have had an OK time so far. The people Dad and WS have gone out with are the parents of 3 teenage boys: one at university, one in my year (who happens to be GORGEOUS) and one in year 10, who I\'ve offered to sort-of babysit on his first gig, if he wants me to go with him, because Jack (middle boy) refuses to- I\'m not really bothered as long as it involves swapping e-mail addresses...  So that brings us to right now. Another day where nothing much has really happened and yet my own mood has managed to turn it into a day worth writing about. Or maybe I\'ve done that so I can spend some time avoiding people by writing this Blog...  Ceff x   Love\'s the funeral of hearts...   '),(17835,2004217,'16,January,2004','2004-01-16 00:00:00','    I take it all back. Fuck father-daughter interaction. I wanna go home, now.   Once more, it is a Friday night and I am on the internet, talking to my friends at HOME, away from my REAL family, while Em, Dad and Wicked Stepmother have a gay old time in the living room listening to Bon Jovi. Bon fucking Jovi. The acoustic album, no less. Could it get any worse? Well, yes. The subject of money could come up. The fact that, despite all our best efforts to conduct legal proceedings regarding maintenance in a civil, neutral manner, he still decides to ring up the CSA (Child Something Agency) and claim that he can\'t afford to pay a poultry Â£143 per month for my keep. Thanks a fucking bunch Dad, You tell me I\'m the most important thing in the world to you, that you love me more than anyone else in your life, and then you try and make me feel as worthless as possible, financially and emotionally. On top of that, this stupid cheap internet connection keeps cutting out, leaving me marooned, sniffling and with a sore throat, with nothing but The Early November and Kerrang! for company. Rock and Roll...  But, on a lighter note, and I will try to make it light (promise), Zeb\'s party last night was very good. There weren\'t as many people there as I expected, but it was still a highly amusing evening for the following reasons:  1. The buffet consisted entirely of sweets and vodka jelly shots 2. The stripper was ugly, tubby and had a Teeny Peeny 3. The outfits were highly amusing 4. I Had drinks but didn\'t get drunk 5. I danced enough to feel I had balanced the effects of eating lots of sweeties from the buffet  Unfortunately my chirpyness this morning was destroyed by my choking on a throat lozenge in the car on the way to school, leaving me breathless and with a sharp pain in my chest; thankfully the pain subsided after a while but the tiredness remained for the rest of the day. Rock and Roll...  Ceff x   Do you regret your loneliness?   '),(17836,2004217,'14,January,2004','2004-01-14 00:00:00','    I am muchly pleased with my new Blogness, henceforth I have decided to make up some new wordage to celebremate this glorifous eventage. Ok enough of that. So! Last night... Drive Thru... 4 bands...9 quid... one great evening. I waited with great anticipation for this year\'s \"Next Big Thing\" Hidden In Plain View, and was thoroughly disappointed. They tried a little too hard to be just like Finch and yet be chirpy at the same time, without having the emotion, tuning or timing of aforementioned band.   Official opinion:  \"Developing\". But if they carry on the way they are, destined to be every emo band\'s support until emo goes out of fashion.  And of course, The Early November were a stunning example of newly created genre (made up by me) emo-pop-punk-rock, being everything from brutal to sensitive in the space of a half-hour set. Wonderful.  Official opinion:  \"Exceeding targets\". Way outperformed HIPV and really set the standard for the headliners. Very deserving of the newly-invented(by Sean) Ceff\'s Seal of Approval, cannot wait for their own tour.  As the lights went down and a large, circular space formed in the middle of the pit, my heart rate began to increase and I started to shake. Memories of the last Home Grown sprang to mind and I began to wonder whether standing by the bar would be a better option. But after two fantastically executed examples of pop-punk, I thought \"What the hell. If you can\'t beat them, join them\" and so I did. And had a fab time bouncing around with Sean, surviving the occasional scary moment and generally enjoying myself.   Official Opinion:  \"Great\". We met the lead singer and he was very appreciative of me coming to all three of Home Grown\'s Birmingham gigs. Really nice guy, great songs and despite the scary following, I still think they\'re gonna be huge.  By this time, due to technical (or, as HG put it, \"testicle\") problems, it\'s 10.15 and Allister still haven\'t arrived onstage. Not good for Sean, who has to leave at 10.30, and not good for me, who has no one to hang out with after Sean leaves. The band finally arrived at 10.20 and played a measly 3 songs before Sean had to leave, to my disappointment. Afterwards, however, they put on a good clean show and really showed off their talents and gave a really tight performance. Great stuff!  Ceff x   Somewhere down on Fullerton, there\'s a place we used to go to get away from it all...   '),(17837,2004217,'11,January,2004','2004-01-11 00:00:00','    I feel strangely calm today. Perhaps it\'s because I\'m at my dad\'s house. Stop looking confused, I know what you\'re thinking: that I hate it there; I feel trapped, marooned, distanced from my own little world. But ever since I started work at Marks & Sparks and going to dad\'s house involved me arriving, eating, sleeping and leaving in the space of 24 hours, I\'ve felt restless; I think it\'s because, as much as I don\'t like it, coming to Worcester and seeing my dad has become a part of me, and (I think) it\'s important for me to get that experience of father-daughter interaction. Or perhaps I\'m content because I\'ve found someone to go to Drive Thru with me :o) Love you Sean! Debbs is gonna come see Hundred Reasons with me as well, so all I need to find now is a gig partner for Mest and Less Than Jake... anyone?  Ceff x   \'Cos this life is too short to live it just for you, but when you feel so powerless, what are you gonna do?   '),(17838,2004217,'05,January,2004','2004-01-05 00:00:00','    Let\'s get drunk, you can drive us to the harbour, Wish upon a star, but do you know what stars are? Balls of fire, burning up the black space, Falling from the landscape, Exploding in the face of God  Let\'s get crazy, talk about our big plans, Places that you\'re going, places that I haven\'t been, Build my walls up, concrete castle, Keep this kingdom free of hassle, Yeah, yeah, yeah, yeah  But I hear sound echo in the emptiness all around  What you can\'t change is loneliness Look at what you\'ve found, I\'m falling down  Taste the saline rolling down your cheekbone, Tell me that youâ€™re alone, tell me on the telephone, Feel your heart, it breaks within your chest now, Try to get some rest now, Sleepâ€™s not coming easy for a while, child  Child, yeah But I hear sound echo in the emptiness all around  What you can\'t change is loneliness Look at what you\'ve found, I\'m falling down  Down, down, down, down  But I hear sound echo in the emptiness all around  What you can\'t change is loneliness Look at what you\'ve found, I\'m falling down Look at what you\'ve found, I\'m falling down   This song is so beautiful... I want to spread my arms and lean back and just float, carefree, in the air as I listen to it.  New Year was... interesting. The venue for the party was very \"Rugby Club\" as my mother put it, but we made the best of it and other than that the trips were fab and the weather was bright and sunny, if not cold Highlight of the few days was mine and Sel\'s foot bath at 3.30am on New Year\'s day, both of us quite drunk, both of us... talking in German. But for some reason I don\'t wanna talk about this. I don\'t want to have to explain all the little in-jokes that we made up, or what it was like being lifted over an iron fence in a ball gown and 4 inch stillettos to climb a muddy hill to watch fireworks, or how fun it was to walk around the bohemian region singing Elephan Love Medley; it was fun, and that is that.  Rich is on my mind. I don\'t feel anything for him (honest). Which is why I keep thinking about him. I think about how I felt about him a month ago, how he treated me and how he\'s going out with Kayla now, but I feel nothing. Not anger, not pain, not sadness. I don\'t understand it; I tell people how I feel used, and I do, but then they get angry and I have to pretend how I\'m still seething because he treated me exactly the same as he did the last time I fell into his arms. It\'s my own fault though! I know what he\'s like! I knew that eventually, sooner or later, he\'d bail on me, and I think somewhere deep inside me, I knew it\'d be sooner. But his arms...the way they fitted around me... Now now, let\'s not go down that road, Jess, look where you ended up last time. Jealous of your short, tubby sister because a stocky, blonde, scarred (emotionally and physically), self-centred boy liked her instead of you. My new year\'s resolutions this year are:  1. No one, especially not Richard Garvey, will take advantage of me. I will be strong and confident and I will  not  be treated like a doormat or a sex toy. 2. Let my nails grow. 3. Lose more weight. 4. Keep reminding myself that I am strong, confident, independent, pretty, intelligent, funny and all those other things that I convinced myself I wasn\'t when I was fat. 5. Smile more: genuine smiles that make me and the person I\'m smiling at feel good. 6. Write more letters and send more birthday cards. Show people that I care. 7. Go to as many 18th birthday parties as possible!  There. Now I have some healthy objectives, ones that will really improve my own personal well being- talk about detox! I\'ve also blocked Luke from my MSN list, and I am considering deleting his numbers from my phone book. Sorry babe, the friends thing? Not gonna happen. We\'re just not compatible. Oh and also I\'d like to get into Leeds University. Please?  Ceff x   You are my reason for breathing...   '),(17839,2004217,'29,February,2004','2004-02-29 00:00:00','    Boom. And suddenly we all fall down. It was all too good to be true. People were getting better, making progress, settling down. And then someone, somewhere presses their self-destruct button and everything falls apart. I know, I know, every family has their problems, their wayward members, but when it\'s your own it seems 100% worse. It hurts to see Oli suffer; I want to give him a hug and tell him everything will be OK- but will it? Besides, he\'s in Edgbaston, and my arms aren\'t that long. Dad hasn\'t contacted me since I left his house on the 19th; it\'s the 29th today. Ellie\'s dad hasn\'t spoken to her in over a month, so I shouldn\'t complain really, and Oli\'s parents haven\'t spoken to him in over two. It all started with him though. Dad. He makes a big fuss, we fight, I cry, he apologises over an E-MAIL and somehow I find it in myself to forgive him. Again. It\'s always about money, too. I hate that stuff. You either don\'t have enough, or you\'ve got too much, and when you don\'t have enough you feel bad because you can\'t help, and when you\'ve got too much you feel bad because you\'re not sure if you should help out or not, because in the past there\'s been problems. Problems you didn\'t even know about until last week, problems that have been around for years. Why does nobody ever tell me??? It\'s worse finding everything out now all at once, than finding things out bit by bit. Nobody told me I had a fucking sister until I was 11, for God\'s sake. On slightly different yet equally pitiful note, I screwed my Leeds interview up. I got facts wrong, I was nervous and I didn\'t know the answer to half the questions they asked, but eventually I relaxed a little and got a bit more confident. I doubt I\'ve got in, though. Everyone in my family (except maybe my mum) seems to have a self-destruct button, and someone\'s pressed theirs. I wonder where mine is...  OK, this paragraph is gonna be about good stuff. Promise. I went to another of Edd\'s gigs on Saturday! Ed Blann was there! I flirted with him! I wasn\'t thinking about Rich and now I don\'t care about him! OK now it\'s just over the top. So yeah, I flirted with him, I don\'t fancy him as such but it made me feel good and realise that Rich is a fool with a tiny penis and I can do better. I spent the afternoon with Sam and Kat yesterday, who threw lots of compliments my way and made me laugh and smile and forget about everything. JYF in two weeks- I can\'t wait. Jack\'s 18th birthday party was last night; it was thoroughly boring and there was no one to dance with, but lots of Jack\'s relatives who I hadn\'t seen in a while said how nice I looked, so that was nice :o)  I don\'t need any of that, though. I don\'t need to know how nice I look nowadays, even though it\'s nice. I don\'t need to go out and flirt, even though it makes me feel sexy. I need a cuddle. I need friendly arms around me, a tissue (yes, me, needing a tissue) and an understanding ear.  Ceff x   There was just one, at the top of a pile of one million...   '),(17840,2004217,'26,February,2004','2004-02-26 00:00:00','    So! This week. Much rushage has occurred because of 2 coursework pieces due in on Friday (tomorrow, the day of my Leeds interview). School photos arrived this week aswell, leading to gangs of girls sitting around making self-pitying comments in the hope someone will pay them a compliment:  \"My hair is fluffy\" \"My eyebrows are wonky\" \"My nose is shiny\" \"I\'ve got a double chin\" \"I shouldn\'t have worn that top\"  Get over it. It\'s done now, and if you want a different picture for the yearbook, shut up and find one. If you still want a decent, professional-looking photo then you\'re going to have to cough up the cash elsewhere. I can understand people being disappointed with their pictures, and that\'s fine. But PLEASE stop going on about it. Hmm, what else has happened... Nope, I think that\'s about it really. What an exciting life I lead!  I was talking to Kayla earlier. She\'s staying at Rich\'s house because she\'s feeling stressed and needs some support in order to sort herself out. Her MSN name last week was \"Namarie Fairy In Love\". In Love. In fucking Love. It\'s going to kill me, seeing them together at JYF, I just know it; I plan to snuggle up to my friends and hide away from them and their \"In Love\"-ness. And it\'s such a shame, because I really like Kayla. I am, of course, dubious about her choice of men, but that\'s nothing to hold against anyone. She\'s lovely and she deserves better. But why do I want him if she\'s too good for him? I don\'t consider myself below her... I keep telling myself, it\'s fine, it\'s fine, move on, move on, but I\'m kidding myself. It\'s such a horrible feeling, waiting for a couple to break up.   Just waiting.  Ceff x   I\'m just another score   '),(17841,2004217,'18,February,2004','2004-02-18 00:00:00','    Haven\'t blogged in a while...  Went to see Lostprophets on Thursday; totally amazing gig, Ian was way hot as usual- the support bands were shite but we had fun communicating by writing out what we wanted to say in the \"Write Message\" sections of our phones, lol... Sean was there, don\'t think he fancies me. Typical me, overanalysing someone just being nice. Valentine\'s Day came and went without excitement, except for a present from my mum of a little Jellycat with a shiny heart- ahhh. :o)  Sunday, Oli came over for dinner; I\'m really worried. I found out why he really rang me in the middle of the night. And the most annoying thing is that I can\'t tell you (whoever may read this) about it, because it\'s... it\'s difficult. I haven\'t been told not to tell anyone, but I don\'t want to, because it\'s uncomfortable; it\'d be a horrible feeling, having to tell everyone how what I thought was true, wasn\'t really at all... So for now, I will carry on like I only know what I knew on Saturday, and not what I found out on Sunday night.  Went to Dad\'s today. The sound of his voice makes me feel like someone\'s scraping their nails down a blackboard right behind me *shudder*. He\'s so loud! And he just spouts such crap all the time. He talks shit and he thinks he\'s right but in the end, he\'s just talking so he hears his own voice. Shut. the fuck. up.  Ceff x   A million miles from home...   '),(17842,2004217,'09,February,2004','2004-02-09 00:00:00','    Switchfoot: Meant To Live    Fumbling his confidence,  And wondâ€™ring why the world has passed him by;  Hoping that heâ€™s meant for more than arguments,  And failed attempts to fly, fly   We were meant to live for so much more... Have we lost ourselves?  Somewhere we live inside, Somewhere we live inside, We were meant to live for so much more... Have we lost ourselves?  Somewhere we live inside   Dreaming about Providence, And whether mice or men have second tries;  Maybe weâ€™ve been livin with our eyes half open,  Maybe weâ€™re bent and broken, broken   We were meant to live for so much more...  Have we lost ourselves?  Somewhere we live inside  Somewhere we live inside  We were meant to live for so much more...  Have we lost ourselves?  Somewhere we live inside   We want more than this worldâ€™s got to offer;  We want more than this worldâ€™s got to offer; We want more than the wars of our fathers, And everything inside screams for second life   We were meant to live for so much more...  Have we lost ourselves?  We were meant to live for so much more...  Have we lost ourselves?  We were meant to live for so much more...  Have we lost ourselves?  We were meant to live, We were meant to live.   This song is very cool. It\'s got this really dirty, sexy bassline and vocalist sounds like he really feels the song, which, of course, is the whole point of music. Feeling.  Talking of feelings, this past couple of days have brought me a mixture of them, from a numb Friday, a blissful Saturday and an empty Sunday, to a pissed-off Monday. The numbness came when I saw my dad. I felt nothing. I saw him and I just felt nothing. Which in a way could be worse than any of the other emotions, because it\'s like my brain is saying \"You\'re so low you\'re not even worth a reaction at all, let alone a bad one\". I was quiet for most of the evening, I didn\'t really feel like talking to him. I spoke to Wicked Stepmother; although I fundamentally hate her, I respect the fact that she hasn\'t tried to get actively involved in the past week. Despite his \"poverty-stricken\" status, Dad managed to cough up the cash for two new CDs and a top for me on Saturday afternoon, much to my delight (and secret anger- if he was giving me the right money I could have bought the damn top myself). Then I got on the coach back to Brum and met the JYF committee in New Street...  Saturday night was fantastic. It cheered me up no end, sitting in Lailah\'s big brother\'s bedroom (he\'s at Uni) next to Matt, laughing and smiling about everything from Little Britain to cruel psychological experiments involving teaching children that green is yellow and doors are walls. You had to be there. Edd really seemed to shine on Saturday night; I didn\'t get pissed off with him at all and he even came and slept on (I must clarify ON, and not WITH) me during the night. We listened to Konstantine and I snuggled up to Matt and it was warm, happy and generally wonderful. On Sunday morning we said our goodbyes and I was once again in New Street...  Nothing much else happened on Sunday, but Monday came along too damn quickly and I soon wished it hadn\'t come along at all when I discovered Katie \"The Herb\" Herbert had virtually taken over our supposedly joint effort on the International Week quiz. I know, I shouldn\'t make such a big deal about it because it\'s only International Week, but she is just. so. ANNOYING! I gave her my disk with my international questions on this morning, only to discover she had \"finished her Fair Trade questions last week\", put up the answers around school and put the questions in form trays. To be fair, she doesn\'t know what I\'ve been through inside my head this past week, but it just tipped me out of my tentative good mood into a thoroughly irritated one. Added to this, Frau Hitler-sorry-Mrs. Plummer, threw me out of the Library-sorry-Learning Resource Centre for having a bag, after I\'d made it quite clear to her that I was in there for 5 minutes to try and print off an English essay. Bitch.  I\'ve written quite a bit for someone who didn\'t really feel like blogging, haven\'t I? Oh, and did I mention that I got an interview at Leeds, and that I\'ve fallen for Maroon 5?  Ceff x   Things just get so crazy, living life gets hard to do   '),(17843,2004217,'05,February,2004','2004-02-05 00:00:00','    I\'ve always thought it was stupid to be afraid of a family member. Someone you\'ve known for all of your life, who you trusted, loved, even. But as this weekend draws closer my nerves are getting worse and worse. I know that he\'ll be fine; he\'ll act like nothing\'s happened, like I didn\'t put the phone down on him 3 days ago, sobbing and screaming, like I\'m just his little girl again. I want to tell him where to go; give him 5 years worth of repressed hatred and revenge. Doesn\'t he understand? Doesn\'t he know that when I look at him, the first thing I think is \"He cheated on my mum. Bastard\", or when I look at her the first thing I think is \"You stole my Dad. Bitch\"? Oh, I\'m just repeating myself now. And it would seem silly, wouldn\'t it, me coming out with \"I hate you, you left me, how could you etc.\" 5 years after I discovered what was going on. But I\'m so much more confident and articulate now; and surely he\'s waiting for the backlash... or maybe I\'ve left it so long that he thinks I\'m \"over\" it. Can you ever \"get over\" it? Can you ever \"get over\" watching your Dad through the bay window of your living room as he packed his things and drove away to his new girlfriend without so much as a goodbye? It makes me sick to think about it.  Things I need for this weekend:  1. CDs- I would die without them. Seriously, if I didn\'t have my music I\'d just stay in my room for 24 hours. Thinking about it, that sounds like a tempting option...  2. Notebook and pens- to review all the CDs I\'ve brought with me! And to write down Blogs while dad\'s on the computer, spending money he should be giving me on horses.  3. Money- In case a sudden urge to shop/see a film on my own grabs me  4. Clothes- Girly dressing up sessions are always a great way to cheer yourself up!  5. Books- A wonderful way of being sociable by being in the room, but not having to talk to anyone at the same time- and interesting, too!  6. Walkman- for those long walks I\'ll be taking around the city centre...  Ceff x   Here\'s to you for bringing me down   '),(17844,2004217,'03,February,2004','2004-02-03 00:00:00','    My mum e-mailed me today, telling me to be strong and that I should always know that he doesn\'t deserve me and that I should tell him so. She also said that we should think positively and know that we\'re in the right and to ignore all the bullshit around us. And I know that\'ll be hard, really hard this time, because I seem to be getting shat on from a variety of directions at the moment, but I\'ll try for her. The last thing I want to see is her suffering because of me. She\'s the most important person in my world and there is nothing that can stop me from not trying to make her happy.  I got an e-mail quiz thing off Rich today. One of those huge surveys about favourites and stuff, probably because Kayla does them all the time. There were lots of little references to her in it, including one about hoping he was in love. I keep telling myself he\'s a knob, and that he\'s not worth my time and effort, but I keep thinking about him. Like I said yesterday, I have no objections to the fact that they\'re happy together, but every time I see a reference to their relationship, or I see Kayla giggling at a text he\'s sent her, I feel this rage rushing up inside me and I just want to scream at her. Can\'t she see?? Can\'t she see that he\'s just a selfish, sex-mad little boy who only wants one thing?? Of course, it\'s all jealousy. If I was going out with him all that would be pushed to the back of my mind, or forgotten completely. And my trainers got dirty in the rain.  It\'s so hard; when I\'m with my friends it\'s so much easier to forget about all this shit in my life, to get on with my work, smile, laugh and generally just mess about. But as soon as I step outside of the classroom, as soon as I walk out of the 6th form block to go somewhere, it all comes rushing back. I want to talk out loud. I want to say, not type, \"I\'m going through a load of crap with my dad at the moment\", just so I can release some of this feeling that\'s all pent up inside of me. I don\'t have the energy to scream and shout and be angry. I simply have a lot of nerves, a dash of self-pity and an enormous sense of loss.  Ceff x   I often wonder how you drive when the road\'s too dark to see   '),(17845,2004217,'02,February,2004','2004-02-02 00:00:00','    Ah, Valentine\'s Day. For some, a blissful reminder of their love for one another, a chance to celebrate the fact that they have found true happiness and an excuse for great sex. For others, it\'s a sharp, painful reminder of the lack of decent, trustworthy men in the world, a chance to wallow in your own self-pity and an excuse to buy a lot of alcohol.   I have no objection to couples; when I hear about new couples I always wish them well and I do genuinely hope that each party is happy, but there\'s always that voice in the back of my mind saying \"How come all these people have partners and I don\'t? What\'s wrong with me?\". If I ever voice these concerns, the standard reply comes back: \"Don\'t worry Jess, you\'re gorgeous and it\'ll happen to you soon blah blah blah dreamy girly speak\". I\'m sure it will, I think one day I will find some wonderful man who really respects me, and it doesn\'t even have to be a boyfriend, just a man who wants to be affectionate and friendly, even fatherly. I just wish he\'d hurry up and make himself obvious, because it seems like I\'m getting left behind...  Ceff x   I\'m not your star... Isn\'t that what you said?   '),(17846,2004217,'01,February,2004','2004-02-01 00:00:00','    When you look at certain people, you automatically think certain things, which tends to affect your judgement. There\'s a girl in year 12, and every time I look at her, I think \"Her dad died\", which of course makes me want to be more \"sensitive\" around her. That happened with Helen for a while, but now not so much. I wonder what people think when they look at me. In year 8, did my classmates think \"There\'s Jess... her parents are divorced\"? Did my best friends, in fact, do my best friends think \"There\'s Jess, her dad\'s a knob\" every time they look at me? Or do they just see me, Jess, wisecrack, Less Than Jake fan, English geek? It almost makes me want to keep it all inside in case they start being more \"sensitive\". I wouldn\'t want that; well, obviously I wouldn\'t want them rubbing their \"happy families\" in my face, but I don\'t expect them to be asking me how I am or if I want to \"talk\". Hopefully they know by now that I\'m not one to go around broadcasting intimate details of my private life to anyone that will listen; if anyone wants to know that I don\'t want sticking their nose into my life, I\'ll crack a joke. But that doesn\'t mean that it gives them the right to take the piss out of my family, or my situation. That applies to anyone, from Wicked fucking Stepmother to Rachel Bradley.  I don\'t wanna talk about my Dad tonight. I don\'t wanna talk about how worthless I feel, or how right now I couldn\'t care less if he keeled over and died in his new wife\'s arms.  Ceff x   To every broken heart in here, love was once a part but now it\'s disappeared...   '),(17847,2004217,'30,March,2004','2004-03-30 00:00:00','    7 years. 7 bloody years of never, ever being late for school, only taking whole days off for illness if I physically couldn\'t get out of bed or open my eyes, and today, 4 weeks before I leave the damn place, and all the clocks and alarms in the house go CRAZY and somehow I wake up at 5 past 9 in the morning. \"Grrr\" just doesn\'t cover it.  The Leeds festival lineup was announced yesterday, prompting me to virtually explode with joy in my dad\'s car:  Confirmed: Green Day, 50 Cent, The Offspring, The Darkness, Ash, The Hives, Morrisey, The White Stripes and The Libertines.  Rumoured: Blink 182, Lostprophets and Funeral For A Friend.  *dies*  Joy also occurred yesterday at the re-discovery of not only MSG+ (doo doo doo-doo, doo doo-doo, doo doo-doo!) and a lil ole thing called Sick-Happy... watch this space ;o)  Musical update! Current interest: Hip hop. I downloaded half of the new NERD album and half of 50 Cent\'s album today and yesterday, which worries me in case I develop the urge to start a collection. But the NERD album is only Â£8.99 from CD Wow... ahh, the trials and tribulations of being a teenager.  Talking of trials and tribulations, I put my giant size 8s right in it yesterday by announcing to Coz that Kat had a spare boyfriend called Nigel who lived in Smethwick, which (somehow!!) seemed very real to him and made him get mad at Kat (boo). Tsch. I can\'t be arsed with boys right now (this is in no way influenced by yesterday\'s event, it just relates to it slightly); they just seem to get in the way. This summer, I want to hang out with my friends, drink, party, go to gigs and generally chill out before I pack my bags and go to uni; a boyfriend just doesn\'t figure in the equation. Good! :o)  Ceff x   Was a po\' nigga, now I\'m a rich nigga...   '),(17848,2004217,'22,March,2004','2004-03-22 00:00:00','    Ugh. Uggggghh. UGGGHHHHH. I feel like a big blob of goo on the floor, all icky and flattened. I got so pissed off this morning after being promptly told to \"get over\" the fact that one of my friends thought Kris was \"pervy\", \"creepy\" and \"weird\". God forbid I try and fucking defend a close friend, I\'m soooo sorry that my friends don\'t live up to your standards. Stop thinking you\'re better than everyone. Combine that with an inability to breathe and total writer\'s block, and the result is something akin to being kicked. Hard. Twice.  UGGGGHHHHHH x   All time lows   '),(17849,2004217,'20,March,2004','2004-03-20 00:00:00','    Yes, yes, I am a truly terrible person who never Blogs and I should think about how I would feel if someone else didn\'t blog when I wanted them to. Or something. I would go and sit in my room and think about what I\'ve done, but then I wouldn\'t be blogging. So! Much time has passed since the last Blog, the Day Count is now in minus numbers and I am officially ill.  1. Results Day. I was being all cool, telling myself that even if my results were worse than they were before (which they wouldn\'t be) it was fine because I can always use the ones from before. Of course, when the time came and I found that I had actually dropped a grade in one of my modules (C to a D), I got pissed off. On the other hand, my second retake module went from a D to a B so that\'s pretty good, but I\'m still stuck with an overall grade C and a shitload of work to do before the exams.  2. JYF. What a relief. It was a bit anticlimatic in some ways; I didn\'t get into my diary that much and there were very few people there, but we still had a laugh and I got lots of hugs and kisses. I think I almost pulled Kris, but I\'m not sure. We got very close one night; I dunno, maybe I\'m over-analysing again. In fact, yes, I am over-analysing because that\'s what I do. Bad squishy.  3. Hundred Reasons. I finally managed to drag myself to the Academy on Monday night, thoroughly unenthused at the prospect of having to stand and watch stuff for 4 hours when I could be catching up on some sleep. As it was, the gig was great and Hundred Reasons put on a wicked show, so it was alright in the end.  4. Maroon 5. Amazing band, amazing songs. Shame about the company. Liz, Gaz and Jon went with me, 2 of whom I count as friends and the other I think is very cool, so I was well up for an excellent night out. Unfortunately, Liz and Jon had a blazing row (including tears, shouting and walking off) and Gaz had brought a date, which sort of turned me into a 5th wheel. Still, it\'s all about the music, right?  The aftermath of all this galavanting is that I now have a horrible coldy-flu type thing, which has been blocking up my nose, ears and throat for the last 4 days. Ugh. Well, at least the weather\'s starting to improve.  Ceff x   Things just get so crazy, living life gets hard to do   '),(17850,2004217,'09,March,2004','2004-03-09 00:00:00','    Hola! Life is still unbelievably good, I just don\'t understand it... The sun is shining, I\'m going to Leeds, I\'m getting good marks, JYF is approaching fast and according to Debbs, Chris actually likes me... of course, at this point in the conversation I climbed into my jumper and hid under the table, but still, it\'s a nice feeling, isn\'t it? :o) I texted him today, just to say hey and (*gasp*) ask him if he wanted to meet up sometime... I suppose I should write something naughty about the thrill of the chase, but... ahhh. :o)  The unfortunate thing about being happy, of course, is that you have nothing to write about; nothing to over-analyse, nobody to rant about. Just sunny, happy feelings. How annoying!   DAY COUNT  Results Day: 2 JYF: 3 Hundred Reasons: 5 Maroon 5: 6  Ceff x   Kiss me, beneath the milky twilight, lead me out on the moonlit floor...   '),(17851,2004217,'07,March,2004','2004-03-07 00:00:00','    Mmmm. Parties. Mmmm. Boys. Mmmm. Boys at parties... I did it! I pulled! His name is Chris and he\'s a friend of Debbs\'s boyfriend Matt; he lives in Four Oaks and goes to Arthur Terry, and he\'s doing History, Psychology, English Lit and... wait for it... RS! He went to V last year, he likes Muse and the Chilis; he\'s tall, skinny, brown, floppy almost-emo hair with blonde highlights and a sort of goatee (but it\'s not rank, and it didn\'t get in the way ;o)), but best of all, he wears Chucks! I was in the kitchen and he came in and somehow we ended up with our arms around each other, then Laura came in and said \"kiss him\" so I gave him a kiss on the cheek, so Laura got irritated and went to get more drinks for us, leaving me and he alone in the kitchen... needless to say, by the time Laura returned we didn\'t need the drinks ;o)   The rest I can\'t be bothered to relate in much detail. We talked a lot and kissed a lot, then as he left he told me it was \"delightful\" to meet me (great!) and I got his number... :o)  So suddenly my life is good. And it\'s not just good, it\'s really good. I\'m not \"not unhappy\", I\'m actually really happy. I have purpose and direction when it comes to work and love, and I\'m calm and content. Of course, there\'s always that nagging feeling in the back of my head: \"don\'t get too comfortable or pleased with yourself; it could all go wrong as soon as it all went right\", so I\'m being careful. I\'m not imagining scenarios between me and Chris of us visiting each other at University, or having some sort of wonderful, magical relationship. But I would really like to see him again, because he was really nice- he got my sense of humour, he\'s intelligent and he wears the  best  shoes... :o)   DAY COUNT  Results Day: 4 JYF: 5 Hundred Reasons: 7 Maroon 5: 8  Ceff x   Just a dream, just an ordinary dream, as I wake in bed, and the boy, just an ordinary boy, was it all in my head?   '),(17852,2004217,'05,March,2004','2004-03-05 00:00:00','    Life is good, life is great, I am fab and everything is right again. I got into Leeds! That particular piece of news sent last night into a frenzy of hyperactive dancing, screaming and quoting Finding Nemo. Everything is fabulous! I feel happy, slim and successful. I know where I\'m going; I\'ve got a purpose. Good things are happening to me, for once - Yay!   I\'ve got this strange hyperactive energy, I haven\'t felt tired at all, in fact the entire opposite: I want to run around and shout and sing and dance because I know where my life is going! I know what I have to do to get where I want to go, and I know that I\'ll definitely be going there if I do it. Talk about motivation, huh?   DAY COUNT  Results Day: 6 JYF: 7 Hundred Reasons: 9 Maroon 5: 10  Ceff x   Just keep swimming, just keep swimming...   '),(17853,2004217,'02,March,2004','2004-03-02 00:00:00','    Today was a good day. Besides the fact that German orals are approaching at a speed faster than the time it took Elena to go from the English room to Claire Masters and bitch about me, I\'ve had a quite smiley day. I spent the lunchtime with Helen because the gang (sans Scooby, unfortunately) went for pizza, one of those things I can no longer enjoy doing. It made a really nice change; I don\'t get to talk to Helen all that often and we had a laugh and listened to *K*. I\'ve discovered its beauty (again). Mmmm.  I\'ve made a decision!  It is not important for me to take a date to the leavers ball.  It is not necessary for me to go looking for someone to take with me just so that I won\'t look like a loner. The leavers\' ball is exactly what it says on the tin: a ball for the school leavers. Not a ball for school leavers and their boyfriends, not a ball for school leavers and who they managed to drag with them in case they looked like they couldn\'t get themselves a man. So if I get a date, I do, and if I don\'t, then... I\'m with Becky!  There are two people on the car park wearing shell suits. Intriguing.  Word of the week:  Whoopah!  Don\'t ask me why, but I just keep saying it.   DAY COUNT  Results Day: 9 JYF: 10 Hundred Reasons: 13 Maroon 5: 14  Ceff x   Are you safe to leave behind every anchor in your mind?   '),(17854,2004217,'01,March,2004','2004-03-01 00:00:00','    I feel better today, stronger. I cheered up a lot after Blogging yesterday, partially due to talking to Sam and knowing that he was there for me, and partially due to listening, strangely, to a lot of hip hop. New additions to my mp3 collection include NERD and 50 Cent, who make me want to dance everything away. Sam was wonderful; I don\'t remember what he said, but he seemed to make me feel safe. Oh I remember now- we were talking about our afternoon in town on Saturday, and he said \"you guys are my happy pills\" :o)... He let me know that he was there for me if I needed him, which sometimes is better than trying to give advice, I think.  Went to the dentist\'s today, and came away with a shiny new filling! Goody. How I love free healthcare- not only do they repair your teeth for you, dentists like to numb half of your face for nearly 5 hours instead of payment! I spent lunchtime acting like a stroke victim: all numb on one side and having to eat everything in tiny little pieces. Now the numbness has worn off (thank god: I didn\'t mind that so much, but the inability to speak properly was beginning to piss me off) and I am left with a dull ache in the back of my mouth where the needle went in. Ouch.  Currently listening to the new Hundred Reasons album: it\'s not as good as their debut, but come on, it\'s not like anyone expected them to produce another classic like that without a little practice. It\'s not bad though- it\'s got oomph and emotion and it\'ll probably transfer really well into a live setting. Yay!   DAY COUNT  Results Day: 10 JYF: 11 Hundred Reasons: 14 Maroon 5: 15  Ceff x   Sing this with me now, try to harmonise this   '),(17855,2004217,'27,April,2004','2004-04-27 00:00:00','    Hehehe. I am SUCH a cheapskate, but at the same time, SUCH a genius. Have invented a new website,  urlLink These trusty chords , which has photos and my reviews on. The reviews that are on there are basically the ones that aren\'t on  urlLink SickHappy , although I may transfer those ones aswell seeing as the site is pretty dead. The photos are ones I want to add to blogs, but instead of paying for Blog*spot pro or whatever it\'s called, I\'ll put the photos I want to show you on there (for free!), and then put links to them! Genius! So that\'s the background. Click  urlLink here  to see photos from the Third Enigma show.  This has no relevance to any of the previous stuff, but I have to get it off my chest: the adverb \"basically\". STOP USING IT!! We had to present some information we\'d researched on various parts of grammar today, and bloody hell, the number of times \"basically\" was used in a descriptive statement was just ridiculous. When it comes to my turn I will make a determined effort not to use it. If I do, I will reprimand myself severely; starving seems appropriate, but only because I have eaten like a big fat carbohydrate covered pig for the last 3 days. Bad squishy.  The \"leaver\'s ball\" debate continues... which leaver will get to go? Hehehe. \"Babble babble bitch bitch\", as Mazza would say.  Ceff x   It\'s so easy not to think for yourself anymore...   '),(17856,2004217,'26,April,2004','2004-04-26 00:00:00','    Hola, yes, bad squishy no blog. Please, reader(s?), find it in your heart(s?) to forgive me. So anyway, my weekend was great! Which is a refreshing change, as usually nothing of any great significance or fun happens and I return to school on Monday with nothing to think about to get me through the week. Friday night, went to see  urlLink Coz\'s band  at a random school in Lichfield. After two support bands, of whom I will refrain from judging (KILL! KILL! KILL!), Third Enigma did their bit and I did a review. After much consultation with my in-house English consultant (Mum), it looks like this:  Third Enigmaâ€™s influences are very clear: their (annoyingly shortened) set echoes the likes of My Vitriol, Placebo, Muse and Radiohead. But donâ€™t mistake strong direction for plain copying - the songs are played with such energy and fervour that Third Enigma arenâ€™t simply imitation Brit-rock. Recent addition Alex Collins gels with the band perfectly for his first show with TE, and itâ€™s soon apparent that an extra guitarist really completes the sound.  After the feedback-fuelled beginning of RobotsMakingNoise, the troubled drawl of Self Esteem sets the tempo for a performance of high quality, driven songs. Most of them follow in the same, angst-ridden vein; the fluid guitar sound of Leave You Behind seems to swirl around the implausibly sweaty venue, while Extremes of Mild joins the ranks of punk in a spot of Dubya bashing. The set is brought to a crashing end with the same improv extravaganza as it began, rounding off a superb bunch of songs to great effect. Third Enigma has masses of enthusiasm, appeal and, most of all, potential; all they need is a little fine-tuning and a lot more of what we saw tonight. Accept no imitations.  It\'s pretty good, not my best but it could be worse. It was worse, in fact. But Coz seems pretty chuffed with it, so that\'s cool. But anyway, back to the story...  The band was wicked and then Coz, Kat, Dave, Oli, Sam, Kyle and I headed back to Coz\'s place for much  urlLink Muse  and talk of  urlLink Skittles , licking 9V batteries, disgusting pizza cake concoctions, drop kicking cats and Kyle taking a \"running thrust\" at Coz\'s TV. I didn\'t get to sleep until about 4am, at which point I was curled up on the floor with 3 other lads (Kyle was on the sofa). Ooer. I wish, hehehe. I woke up at about 8am and went and hung out with Sam and Oli outside in the sunshine for a bit (Coz\'s GIANT of a dog Murphy also joined us, who I had totally forgotten about since I last went to his house - in the words of Kyle: \"I could LITERALLY ride him!!\"), until Coz drove me home and I spent the rest of the weekend in the leisurely confines of Charter Place, outside, on a bench, the stereo at the window playing Rock Against Bush Vol. 1, watching a nice young man called Dom work up a sweat under the hot sun as he painted the front of Paul\'s house... rrrrr.   Incidentally, Rock Against Bush Vol. 1 is bloody good album. The best tracks (should you wish to download them) are:  Sum 41- Moron Alkaline Trio- Warbrain The Get Up Kids- Lion And The Lamb Descendents- Sad State Of Affairs Strung Out- No Voice Of Mine Pennywise- God Save The USA New Found Glory- No News Is Good News Less Than Jake- The Brightest Bulb Has Burned Out  It may seem like a huge selection, but 8 really great tracks out of 26 is pretty fair, right? *grin* and there\'s a DVD! Of course, to counter balance the good time I\'m having right now, I find myself wanting someone to share it with, but I know a relationship is just not practical before I go to Uni, it\'d complicate things waaaaaay too much. Honest.  Ceff x   It\'s such a sad state of affairs, and I don\'t care...   '),(17857,2004217,'20,April,2004','2004-04-20 00:00:00','    The terrible person within me who never, ever blogs because she is a lazy git has returned. Hola. Now is the time for a list:  1. Easter/The Zoo 2. Barnsley 3. Work 4. Dad 5. Comics 6. Miscellaneous  We are now in our second day of school after the poor excuse for a holiday that was Easter. I had lots of  urlLink driving lessons  and can now do left turns and... well, that\'s about it. We went to  urlLink The Zoo  on  urlLink Easter Sunday , and it was soooo cute and there were sooooo many cutesy lil  cute animals! There was a baby  urlLink elephant  and  urlLink lions  and  urlLink tigers  and best of all...  urlLink MEERKATS!  I want a pet one. Don\'t laugh! They\'re so sweet, all fluffy and squiggly and... aww. :o)  Went to Barnsley on Easter Monday. Stayed with my nan. I haven\'t done that since I was about 7. Oliver came over and we all discussed... stuff. It felt like some sort of horrible secret summit, a nasty little bitching session. Very odd. Dad found out about the whole sorry affair on Monday, and he seemed strangely accommodating about the whole thing. What is wrong with all my family members? Hee hee.  Work, to be specific, revision, is looming ever faster towards me, being all scary and huge. I keep telling myself \"It\'s OK, I have a whooole month before exams\" and then I remember my German oral which is on the 6TH OF MAY (sob), and then my heart starts to beat faster and I feel sick. Blergh.  Dad is scaring me. He\'s going all out on the \"you\'re my favourite daughter\" routine. From New York, not only did he bring me the hat I requested, but some choccie lip balm, a note book, a Hard Rock Cafe pin and two tops! Mum reckons some of it was meant for Em... hehe. Apparently she was meant to come to Dad\'s house this weekend but he said something \"wrong\" on the phone the other day and she got all stroppy. Oh well, at least now I get the best bed for a bit longer....;o) I\'m a bit apprehensive because this weekend I\'m going to have to discuss  urlLink Uni  fees; we\'ve been getting on so well since  urlLink Em  fucked off and I\'d rather not jeopardise that by bringing up the ever painful subject of money. It always turns things sour.  What\'s next on the list? Ahhh, comics! Ed Blann has made me a compulsive webcomic reader; not only am I reading all of  urlLink his  comics, but 3 others:   urlLink Questionable Content , which is about some indie kids who may or may not fancy each other   urlLink Count Your Sheep , which is about a little girl called Katie and her imaginary friend Ship, who just so happens to be a sheep   urlLink Chugworth Academy , which I started reading today. It\'s about a group of naughty teenagers supposedly at high school in the good ole US of A, and their lives in general.  Help me. I think I have a comic problem. Looking back I also think I have a hyperlink problem. Still, I need the practice, right?  In other news, I have decided I am going to actually DO something with my hair. Mum got me some volumising shampoo for me today, and she\'s gonna trim it for me (finally!) and highlight it. She also suggested putting a broad pink stripe in it somewhere... hopefully it\'ll turn out alright... meep! I\'ve also started listening to \"North\" again...it\'s such a smooth album and some of the tracks make me go all *mmm*. Ooh, and I randomly downloaded this thing from Microsoft which has made me a database of all my mp3s! Genius! If only it could do that with my CD collection... which reminds me, new  urlLink New Found Glory  album, Catalyst, in the shops, 17th May!! Bring it on! The new single, All Downhill From Here, is released the week after and I\'ve already downloaded. Sweet as.    PLUG PLUG PLUG : Buy Rock Against Bush, show your support for a)  urlLink Fat Wreck , b)punk rock and c)MVC- it\'s only Â£4.99 if you buy it from there and it\'s got about 294958939394 tracks on it from some wicked bands.  Ceff x   Catalyst, you exist to pull me down...   '),(17858,2004217,'05,April,2004','2004-04-05 00:00:00','    According to my horoscope, this is what I should be doing. And it was the Observer horoscope, so obviously it\'s true. Talking of the Observer, there was a giant article about blogging in their Review section yesterday. Although it feels good to be ever so slightly ahead of the \"trend\", I am slightly annoyed because mine and Kat\'s *thing* has now been broadcast to the entire nation, and suddenly everyone will want one. Just like punctuation and Lynne Truss. Gah. But I digress; as I said, in accordance with said horoscope, I am doing my best to remain as sunny as possible. I had a good driving lesson today, relaxed and went at (wait for it) over 30mph! Zoom zoom zoom. I am also digging out as many summery CDs and songs as I can find, including these albums:  A -  Hi Fi Serious The All-American Rejects - The All-American Rejects Less Than Jake - Hello Rockview  and these songs:  Nelly - Ride Wit Me Fenix TX - Surf Song Letters To Cleo -  I Want You To Want Me Self - Stay Home (Shrek soundtrack)  I also dressed up in my brand spanking new Sandra D outfit (minty green strapless dress, matching baby doll shoes and white cardy) and pranced around the house singing and celebrating the fact that the dress is a size 14. Yippee!  Inevitably, as the sun makes brief appearances, the number of people in the car park rises. I like the car park; it\'s like a microcosm of the world right outside my window. Everything happens there, and I see it all, hee hee hee... I\'ve noticed a strange increase in the number of punks hanging around. Not skater kids, or random people with baggy jeans and hoodies, but proper punks with tights jeans and anarchy symbols stitched, printed, carved or otherwise attached to skin, bags and clothes. I will watch this development with careful interest... talking of punk, I was reading the Bible-sorry- Kerrang last night and I think they\'ve FINALLY pinned down a definition of today\'s punk:  \"Bending the system to your advantage\"  Simple and, well, not very punk, but effective enough. It covers all manner of sins, including Good Charlotte. The only thing is, it was in an article about Mest, who, in case you were unaware (which you probably were because I haven\'t told anyone), I\'ve gone right off. Eugh- Mest.  Ceff x   I love the way you roll excuses off the tip of your tongue, as I slowly fall apart...   '),(17859,2004217,'05,April,2004','2004-04-05 00:00:00','    After the (generally) upbeat tone of the last post, I find myself 3 days into the Easter holidays with nothing to do and an enthusiasm only to sit at the computer, doing nothing, which isn\'t really bloggable material. The most interesting thing to happen today was a phone call from Dad: Wicked Stepmother has a new job with HMI (Her Majesty\'s Inspectors- to put some perspective on the magnitude of this job, Ofsted is their bitch), which according to Dad, will incur \"changes\". Ones that he refuses to talk about on the phone. Intriguing. It\'s probably not to do with money, although I won\'t rule it out; mum says HMI get moved around the country at the drop of a hat, so maybe they\'re moving away... I don\'t know how I feel about that. I seem to be getting on with Dad at the moment, so him moving away might make me unhappy. Mind you, if you\'d asked my in February I couldn\'t have cared less where he went, as long as he suffered a horrible car crash on the way there and died a horrible death in a horrible way. Ahem.  I bought the NERD album, by the way. It\'s good; very slick, very sexy, very \"Pharell\".  Ho hum.  Ceff x   Love is not a victory march...   '),(17860,2004217,'29,May,2004','2004-05-29 00:00:00','    Ah, there\'s nothing like a little Morrissey to remind you how great your life is.  My joy has faded; I have come down from my natural high. Even Morrissey can only barely help ease the immense sense of frustration, irritation and/or anger I feel every time Em speaks. Yes, she arrived, quietly, at lunchtime today and suddenly everything is back to \"normal\", right down to the return of my underlying reluctance to come to Worcester. It\'s not an optimistic visit to Dad\'s house anymore, it\'s regular punishment; imprisonment for a whole weekend with the remnants of his life, the dregs he scraped out the bottom of the barrel labelled \"people I\'ve fucked over\". At least now I know it\'s not Dad that causes me stress though: it\'s Em, with all her talk of, well,  her , and whether she can find a place in her garden where she can sunbathe in the nude (seriously), or how is was her that created the monster that is Rich (grr), or how she really needs to be at home with Lee right now (melodramatic? Em? What a surprise). The most annoying thing I find about her self-centredness is that after a while, I end up competing with her- she\'s going to see a band live, I\'ve seen two this week; she likes sunbathing, I\'ve been out in the sun constantly recently; she doesn\'t like Rich, I hate him with the passion of a thousand suns. And it pisses me off that I end up like her when she\'s around, because what she is is everything I hate. Ooh, she\'d love it if she knew I was writing about her...  Ceff x   It\'s OK to be angry and never let go   '),(17861,2004217,'28,May,2004','2004-05-28 00:00:00','    I\'m not exactly sure how I\'m going to write this blog, seeing as there are no words to describe last night. In answer to my mother\'s question \"Was it good?\" on the journey home last night, all I could say was \"Ohhhhh wowwww... best gig EVER\". Have you ever experienced a total connection, a total sense of belonging, like this is the place or time that you could stay in forever? That was last night. I knew the words to every single song except for one new one (from their new album coming out in July yaaaaaaay), and I moshed, skanked, ran around like a crazy person and then I moshed some more; the people were all there for the same thing, to see the same band, to sing the same songs; it was like going home. Bliss. I don\'t know if anything will ever live up to it... until I go and see them again, that is ;o)  Ceff x   This is a fair request, and I promise I will not judge any person only as a teenager. If you will constantly remind yourself that some of my generation judges people by their race, their belief, or the colour of their skin, and that this is no more right than saying all teenagers are drunken dope addicts or glue sniffers...   '),(17862,2004217,'25,May,2004','2004-05-25 00:00:00','    Eeeeeee giggage. I went to see  urlLink King Adora  with Coz on Friday. The rest of his band was there and I was chatting to them, and they all seem really nice guys. Eeeeee I know a band. Coz told me he\'d sent my review to the other 3 band members, until Charlie told me he hadn\'t got it, as did Jack, so in fact, in Coz\'s head, 3 must equal 1. Bless \'im. Much \"huzzah\"ing occurred, as well as much sweatiness do to my innapropriate dress (2 tops). I had a lot of fun, then got invited back to Chillie\'s house (not by Chillie- again, this is all Coz\'s fault :D) and rang mum to ask her if I could stay at Coz\'s (it\'s near enough) and received a definite no. Boo. So spent a couple of minutes moping before venting some anger on the moshpit. The band were good, too. I haven\'t heard their album in about 2 years but I remembered I liked their style; they really bring it off in a live setting as well, and I think I might try and get my hands on the new album. Or just copy it off Lucy. Mwahahaha.  Eeeeeee more giggage. Guess who I\'m going to see on Thursday *grin* go on *grin* * urlLink guess  *grin* Can\'t wait *grin* best gig ever *grin* after the last time I saw them *grin* Eeeeeee. *grin* If you think I\'m being restrained, you can\'t actually see me; whenever I think about it my body seems to spasm strangely and I have the urge to jump around the room like a crazy person, which, incidentally, I will be doing on Thursday night. EEEEEEEEEEE.  urlLink *grin*   Ceff x   Whatever turns you on...   '),(17863,2004217,'19,May,2004','2004-05-19 00:00:00','    It feels like last year all over again. Dad and I had a big fight about (guess what) money on Monday, namely how much of it he was going to give me in order for me to go to Uni. He seems to be incapable of giving money to anyone other than me. That might sound good, but when the money is your rent and fees, which could be paid straight to the university, making me a middle man seems a bit pointless. He refuses to give the money to my mum as well; it\'s like he doesn\'t trust her to pay it- he\'s still bitter about the whole CSA thing, which he keeps claiming is a pointless excercise, so why he get so annoyed about it is beyond me. Oh, I can\'t be arsed with it anymore. If I wasn\'t reliant on him paying half of everything, I\'d tell him where to put his money and get rid, because it stresses me out too much, and I don\'t need that a month before my exams. We were getting on so well, too. Remember that self-destruct button?  Last year reared it\'s ugly head again at 10.41pm last night (I was looking at the clock when it happened), when my phone rang:  Ceff: \"Hello?\" Speaker #2: \"Hi, it\'s me\" Ceff: \"Er... Hi... Who\'s this?\" Speaker #2: \"It\'s Luke\" Ceff (in shock): \"Oh, hi...\" Luke: \"...\" Ceff: \"What?\" Luke: \"Just wanted to see how you are\" Ceff (thinking quickly): \"I\'m fine, I\'m just going to bed\"  Luke: \"Oh, OK then\" Ceff (irritated): \"...er, speak to you soon then?\" Luke: \"Yeah, OK then\" Ceff: \"Bye\"  What the hell is going on?? For one, why is he ringing at half ten at night to \"see how I am\"?? And if he\'s ringing me, surely he should have something to say?? If he rings again I\'ll make it clear I don\'t want to speak to him again. There are various things I could try:  - Pretending I don\'t know who he is - Speaking in German - Not speaking at all - Telling him I\'ve moved to the Outer Hebredes - Not answer the phone - Pick up the phone, and put it down again - Get Debbs to answer the phone and speak in Spanish - Get Sel to answer the phone and speak Gujerati - Get Laura to answer the phone and speak Liverpudlian  I could tell him the truth/to piss off and leave me alone, but that\'s just not as fun, hehehe.  I have started revision! If revision means sunbathing and eating strawberries whilst half reading a German copy of Cosmo...;o)  Ceff x   '),(17864,2004217,'16,May,2004','2004-05-16 00:00:00','    It\'s all over! I am officially a school leaver, and celebrated thie momentous event with an afternoon of heavy drinking. It was such a laugh; Sel, Bex and Debbs were talking to teachers, Laura was sat with the TT et al (imagine me making a face similar to the MSN emoticon with one eyebrow raised) and I spent the afternoon with Kat, Joolz, Bhav, Beth, Nat, Liz, Hayl and Coz. the turning point (where alcohol was concerned) was most definitely the woo-woo drinking competition (which we won!), but other highlights of the afternoon included saying \"huzzah\" an unimaginably high number of times, Liz\'s craving for filthy pirate rum, writing people\'s code names (Be Careful, Gutted, Holier-than-thou etc.) next to their pictures in the yearbook and me being inadvertantly mean to Mrs. No. 4\'s boyf when Coz gave me a lift home. Handy tip: make sure that all windows in the car are closed before proclaiming loudly that the person standing outside the car is \"evil\".  Before all this alcohol washed away a considerable amount of brain cells, the Leavers\' Assembly was held in the hall. I won two awards: Most Amazing Transformation and Person Most Likely To Correct Your Grammar. Yay! As Ripu proclaimed to the entire year, yes, it is sad, but you know you love me for it. *grin* Some people, however, were less than gracious about my fabby award, claiming loudly to their form tutor that they \"should have won that award\". Well, you didn\'t; what are you gonna do with a star-shaped piece of card on a string anyway?   There were a couple of other unexpected events that afternoon as well- firstly, Danni Mynard asked me to sign her yearbook. Danni Mynard! Queen of the Barbies, winner of Dumbest Blonde and general Jess-disliker (or so I assume) since about October 1997. I signed it, obviously, because I\'m not gonna get all mean and stubborn on the last day, but it was still very odd. Why? It\'s not like I mean anything to her. The other odd event on Friday was Sanita, who also started talking to me, but for entirely different reasons. She came up to the table and said \"Hi Jess\" and shook my hand. Noting my (very) confused (and drunken) response, she said \"I heard you were pissed so I thought I\'d come over and say hi\". I replied \"Er... thanks\" and she went off, back to the confines of the AM. Sanita hasn\'t spoken to me in about 3 years, but somehow she finds it satisfying to come over and, well, not laugh, but very subtly mock me because I\'ve had a few drinks. If that was one of my friends I wouldn\'t mind, but it pisses me off that she thinks she can come up to me and take the mick because I was enjoying myself. Grr.  At about 5pm Coz drove me home and my mum fed me gammon and eggs and made me drink tea. Two mugs. Large mugs. On the drive home I was laughing at all the kiddies in uniform and how they were still at school; yes, it was naughty, but also strangely satisfying. You should try it some time. I got changed and returned to town, only to be ID\'d at Yates\'s and to sit in Imperial being bored. I went home at 9pm and fell asleep on the sofa watching Late Night With Jonathan Ross, and awoke the next morning (somehow) without a hangover. I am now at Dad\'s house (Em was supposed to come but hey, whaddya know, she had a \"crisis\" and had to stay at home), and I\'ve just spent the afternoon at the races, in the sun, in my new strapless, stripy summer dress.   Revision starts tomorrow... honest!  Ceff x   Last summer, I did nothing, just sleeping, thinking and hanging around...   '),(17865,2004217,'13,May,2004','2004-05-13 00:00:00','    Eek. Last day of school- ever- tomorrow. Suddenly exams seems so much closer...  This has not been the best week of my life. For a start, it has been full of sentiment: the release of the yearbook (and all the signing it entails) seemed to make it OK for people to burst out crying at any point during the day. Ick. I can\'t deal with people crying; it freaks me out. If they don\'t stop very fast, i.e. within 30 seconds, I get irritable. I don\'t cry myself; it shows weakness on my part and it ruins my face. Also, I\'ve managed to go 7 years without most of the year seeing me cry, so I\'m not going to break with tradition. It will be the end of an era though: never again will I get postcards from mum with my lunch, or go to M23 for German only to be told by Mr. Birchall that we\'re in H02, or have to write a radio script for Word Of Mouth, or go to assembly, or have last minute end-of-term parties in the form room, or fight over the stereo in the common room, or complain about the lack of Galaxy bars in the chocolate machine, or wonder who\'s free now, or, most scarily of all, have my best friends around me all the time. We\'ve had our ups and downs, our fights, our laughs, our cries, our practical jokes and our piss takes, but Bex, Sel, Laura and I (Debbs joined in year 11) have been a little group since year 7 and when I get to Uni they won\'t be there anymore. And that makes me sad.  Last day ever. Not gonna cry.  Ceff x   '),(17866,2004217,'10,May,2004','2004-05-10 00:00:00','    Last week was filled with many fun/scary/significant events. I finally decided that I was definitely going to go to the Leavers\' Ball, put my money where my mouth is and bought a ticket. So far, I have no pangs of regret, but I certainly have no great enthusiasm for the whole affair. Laura has lost her ticket; Pave says she can\'t give her another one, and has also told Laura that she can\'t give her name and ticket number at the door, because of \"security problems we had last year\", so Laura spent today searching between the chemistry lab and the form room for her missing ticket, before going to ask Cindi to see if she could get a more forgiving result. I\'m so glad that we get such good service for the high price of the ticket....  Talking of tickets, it says on mine (in fact, them all) that it is \"Black Tie\", so upon this discovery I dragged out my bridesmaid dress and tried it on. It\'s too big. Bugger. My mum is too scared to take it in and says it looks fine, but I feel horribly fat in it and I don\'t think it will flatter me at all. I tried Em\'s bridesmaid skirt on- a snug, but good fit, and am contemplating whether or not I should just wear that with my purple stilettos and a purple top. My first plan was to wear my strawberry dress, but it seems that this is not formal enough. Hmph. Accessorisation seems to be the way to go on this issue. Is accessorisation a word? Well, it is now.  We didn\'t go and see Win A Date With Tad Hamilton. Debbs dropped out to see Matt instead, Bex had to finish her Tech coursework and we were late for the mid-afternoon Tad Hamilton showing, so in the end me and Sel went to see 50 First Dates. Don\'t waste your money: the jokes were funny and Drew Barrymore looks gorgeous, but the plot was so flawed it was impossible to really get into it. Something I did get into, however, was Funeral For A Friend on Wednesday. After Frankie dropped out at the last minute, I had to ring Lucy in a state of panic at 10 O\' clock Tuesday night and ask her to come with me. Thankfully, she did and the gig was amazing; there were 3 (3!) support bands, 2 of whom were shite but the middle one was bearable, and Funeral were just amazing. The lead singer has a mouth like a toilet though; every other word was \"fuck\" or \"fuckin\'\", which started to get on my nerves after a while, but then they\'d play a song and I\'d fall in love with them all over again. Of course, my mother was, for the first time, highly disapproving of my going to this gig, because it was...  My German oral exam on Thursday morning! After writing, scrapping and re-writing my presentation 2 days beforehand, understandably I had a few worries about my first exam. Thankfully though, it went fine and Mrs. Dalgleish really played to my strengths, and says if she was marking it she\'d give me and A or a B, which certainly takes the pressure off for the written exam. Phew! The best thing that happened that day was after the oral, when I arrived at 12.20 to my first General Studies session, which started at 10.30, and said to Mrs. Clothier, \"Hi... Sorry I\'m late\", hehehe.  Of course, all of this happened unbenownst to my Dad, who, when he rang me last night (instead of Thursday, when he said he would) told me that he had no idea that I had a gig and an exam this week, even though they were the last things I said to him as I got out the car on Monday, when I saw him. Out of sight, out of mind, eh? Never mind.  4 days of school left now. Everyone seems to have something to complain about currently; most popular is the Leavers\' Ball, while another is the teachers and how they don\'t care, closely followed by something along the lines of \"I\'m so fucking sick of school and can\'t wait to get out of this shithole\". Am I the only person that appreciates the really good education we\'ve received while we\'ve been here? Yeah, so we have to come in for registration and give reasons why we\'re late; yeah, so some teachers make mistakes and some are worse than others, but good grief, if we\'d gone to a regular comprehensive, our teaching and treatment would have been ten times worse. All I want to do this week is chill out, not hear any complaints and enjoy the last week of the place that\'s got me to where I am now and will help me to get where I want to go in the future.  OK, so now my hands are getting tired and Kat is complaining that she has nothing to read (hehehe), so I will tell you all about Mark\'s 21st Birthday party and spending the afternoon with Sam and Kris another time.  Ceff x   Whether inflation, or globalisation, fear is a weapon of mass destruction   '),(17867,2004217,'02,May,2004','2004-05-02 00:00:00','    Holaaaaaaaa  Ahem. Bank Holidays! Whoever persuaded banks to take random Mondays off at seemingly illogical points in the year deserves a medal. I got thoroughly sozzled at Sel\'s 18th Birthday party last night, knowing full well that I could get 2 good nights of sleep before I dragged my sorry ass back to The Institution For Bright People on Mon-sorry-Tuesday! We (Me, Sel, Bex) are going to see  urlLink Win a Date With Tad Hamilton  tomorrow; the teen movie that\'s so bad you don\'t even have to be a teen to watch it (it\'s certificate PG).   urlLink Sel\'s party  was amazing. Despite the facts that we were the only people there from school, I only really knew Sel\'s immediate family, didn\'t like the food and was just *awful* at cricket, I had a really good time and everyone was really chilled out and friendly. Laura was clearly not having a good time at the start, though, which made it a bit tough because I was starting to relax around all the new company and she was really tense; the problem was that she hates cricket and really didn\'t want to play, plus the fact that she didn\'t talk to anyone probably didn\'t help matters much. Anyways, what matters is that I enjoyed myself and felt really welcomed by the whole thing, which was cool; I met loadsa new people who were all fab and generally had a great time of it.  There are only 9 days of school left. I have already had my last lesson with Mrs. Dalgleish. I have 5 days to decide whether I want to go to the Leavers\' Ball or not.   REASONS FOR:  It\'s a reason to get all dressed up and dance the night away; it\'s the last time we will be together as a whole year group; it\'s a great tradition.   REASONS AGAINST:  It\'s the same night as JYF; it\'s Â£30 for a ticket plus extras; some of the year won\'t go, which counters the \"last time together as a group\" argument; I don\'t like a lot of the year; Frankie and Hannah have been discussing choreographing a \"She\'s All That\"-style dance routine for the occasion.  I wonder if I could ask Pav if I could go for less money if I told her I was only going for an hour before I went to JYF... My head hurts...  Ceff x   Will you need me now? You\'ll find a way somehow...   '),(17868,2004217,'30,June,2004','2004-06-30 00:00:00','    I am less twitchy today. My irritation has subsided (although not entirely) and I\'m sure everything will be alright very soon.  I had my last exam today. My last ever reason to go to school. I handed back all my books and my locker key this afternoon, and I left. I left school. I am now a fully fledged member of society; I can get a job, start a career, do whatever I want. The only nagging thing now is that I have no base -  nowhere where I know I will find people who are my friends, where there\'s stability and where I\'ll always get a laugh and a joke. And a packed lunch. Now... now I have to work for friendships and relationships. And lunches. It all gets so much more complicated from here. I didn\'t want to step outside the gates this afternoon. I trudged towards them, thinking \"This is it. After this, I\'ll never have to come into school again\" Of course, I\'ll be back there for results day, but if I didn\'t want to, I could just have my results to sent to me. It was strange, leaving; everyone else had left early because all the staff were having a meeting in the library, so there was no one about. It was like the last scene of Friends, when the place was empty and they left it all sentimental and weepy. Obviously, I wasn\'t weepy, but it was still hard for me. I mean, that place  was  me for 7 years; it taught me everything I know, academically at least, and also gave me great friends. It\'s the place I was when my parents got divorced, when I found out about Em\'s existence, and various other big-at-the-time events that I\'ll always associate with school. I feel a bit lost now, actually.  I was kept in quarantine today. Like a foreign animal. I had an exam clash (AQA assumed I could cope with both English Language and RS simultaneously) in the afternoon, so my English exam was moved to the morning, thus consigning me to 2 hours\' sitting in the finance office being bored. The time came and I was led through hoards of people, as if going to be crucified, them all smiling, slightly mockingly, at me, while I was put into the RS room and had tables shifted around me. It was a lovely experience. I felt very valued.  Ceff x   Carve your heart out yourself...   '),(17869,2004217,'29,June,2004','2004-06-29 00:00:00','    Arrrgggghhhhhhh!!  Have problems. Cannot write them here. One, because writing on the computer screen is probably unadvisable, but also because people read my blog. Bugger. I will have to sit here and twitch with irritation until it all goes away.   She said I\'ve gotta be honest,  you\'re wasting your time if you\'re fishing round here.  And I said you must be mistaken,  I\'m not fooling, this feeling is real.  She said you\'ve gotta be crazy! What do you take me for?  Some kind of easy mark?  No, you\'ve got wits, you\'ve got looks, you\'ve got passion,  but I swear that you\'ve got me all wrong.  All wrong, all wrong,  but you got me...   I\'ll be true, I\'ll be useful,  I\'ll be cavalier, I\'ll be yours my dear,  and I\'ll belong to you, if you just let me through.   This is easy as lovers go,  so don\'t complicate it by hesitating, And this is wonderful as loving goes,  this is tailor-made,  what\'s the sense in waiting?   And I said I\'ve gotta be honest,  I\'ve been waiting for you all my life.  For so long I thought I was asylum bound,  but just seeing you makes me think twice.  And being with you here makes me sane,  I fear I\'ll go crazy if you leave my side.  You\'ve got wits, you\'ve got looks, you\'ve got passion, but are you brave enough to leave with me tonight?  Tonight, tonight, you\'ve got me...   I\'ll be true, I\'ll be useful,  I\'ll be cavalier, I\'ll be yours my dear.  and I\'ll belong to you,  if you\'ll just let me through.   This is easy as lovers go,  so don\'t complicate it by hesitating,  and this is wonderful as loving goes,  this is tailor-made,  what\'s the sense in waiting?   This is easy as lovers go,  so don\'t complicate it by hesitating, and this is wonderful as loving goes,  this is tailor-made, what\'s the sense in waiting?   Ceff x   So much for your \"endearing sense of charm\", it served you well, but now it\'s gone and you\'re wasted on me   '),(17870,2004217,'26,June,2004','2004-06-26 00:00:00','    \"Man, it takes a silly girl to lie about the dreams she had, but it takes a lonely one to wish that she had never dreamt at all\"   All hail the return of Dashboard Confessional and the gloriously bittersweet \"Carve Your Heart Out Yourself\". Huzzah, I say, and again, huzzah. From what I\'ve heard of \"A Mark, A Mission, A Brand, A Scar\", Chris Carraba has decided to use instruments that you can actually plug in this time around, which I\'m sure is very new and exciting for him.  I have discovered a new comic! It\'s called  urlLink Ctrl+Alt+Del  and it\'s laugh-out-loud funny. It\'s about this computer games geek Ethan, his best friend and roommate Lucas and his other roommate, whose name I can\'t remember, and the other roommate\'s pet penguin, Ted. It is based a lot around computer games so some parts of the story I don\'t get, but the jokes are still universally funny. Huzzah.  On a more ominous note, my horoscope reads: \"If some fickle personage has been meddling with your heart strings in the last month or two, the tide turns this week.\" Gulp. JYF is this weekend. Gulp. The R-word springs to mind...  Gulp x   Oh look now, there you go with hope again...   '),(17871,2004217,'26,June,2004','2004-06-26 00:00:00','    urlLink    \"Shake my body loose, shake my body, whatever turns you on... Make my body smoulder, make my body so sexational!\" A naughty title for a naughty album...&nbsp; urlLink     In the true spirit of girlyness, I have decided to completely change my outfit for the Leavers\' Ball, after getting a new skirt this afternoon in Birmingham. I will not divulge details tonight, though... Heeeeee. I spent some time posing in front of the mirror (in my newly proposed outfit) to King Adora when I got in; I love that album, it\'s so trashy and chaotic, yet it\'s got this raw, animalistic sexiness to it as well... I wish they\'d play some at the Leavers\' Ball because I could really, really work it to a tune like Smoulder. ;o)   Never knew I could feel like this; Like I\'ve never seen the sky before... Want to vanish inside your kiss; Every day I love you more and more... Listen to my heart, can you hear it sing? Telling me to give you everything... Seasons may change, winter to spring; But I love you until the end of time...  Come what may; Come what may; I will love you until my dying day...  Suddenly the world seems such a perfect place; Suddenly it moves with such a perfect grace; Suddenly my life doesn\'t seem such a waste; It all revolves around you... And there\'s no mountain too high; No river too wide; Sing out this song and I\'ll be there by your side... Storm clouds may gather; And stars may collide; But I love you (I love you), Until the end of time(until the end of time)...  Come what may; Come what may; I will love you until my dying day...  Oh, come what may, come what may, I will love you, Oh I will love you... Suddenly the world seems such a perfect place;  Come what may; Come what may; I will love you until my dying day...    Damn Moulin Rouge, restoring my belief in an all-consuming, all conquering love. It won\'t ever happen, it\'s not possible, but hey, the Leo in me enjoys the passion and the heat of the film. And Ewan McGregor ;o)  Ceff x   You\'re free to leave me, but just don\'t deceive me   '),(17872,2004217,'24,June,2004','2004-06-24 00:00:00','    So, after a not-so-lengthy discussion with Kat, I have cleared my messy mind up a little. In the relationship department anyway:  I don\'t want commitment, or anything heavy and intense, but I do want something exclusive. I don\'t want to get into any kind of heavy duty physical *ahem* activities too fast, but I don\'t mind a little messing about. I don\'t want to feel like I have to bring him (whoever it may be) to every event that I go to because he\'s my \"boyfriend\" (I refer to an instance when Laura pestered me to bring Luke to her birthday party even though I didn\'t want him to come), but I don\'t mind calling him \"my boyfriend\". I want something easy and fun, something that\'s chilled out with a lots of flirting but without having to change my life. I want good conversation as well; I don\'t want everything I say to be met with \"awwww\" or \"oh how sweet\" or \"bless\" or some other disgustingly patronising comment because he hasn\'t listened, or understood what I said. Is that so much to ask? ;o)  It\'s just that I can\'t stand all that *stuff* that\'s attached to having a boyfriend! The way people expect you to be attached to each other all the time, and have \"quality time\" together, or spend ridiculous amounts of money on each other because you\'re \"in love\". The way they act surprised when you just spend time with your friends, or chilling out on your own; I don\'t understand these girls that give up their whole lives, their friends, the way they are, just to be with one guy. How can you spend that much time with someone and not get bored anyway? OK, not get bored, but at least need a considerable break from each other afterwards.  Well, that helped. I feel much clearer in my head now, and don\'t feel the need to write about whatever I\'m reading/revising as a distraction technique.  Ceff x   There\'s nothing to lose; my notebook will explain...   '),(17873,2004217,'22,June,2004','2004-06-22 00:00:00','    urlLink    Huzzah! Fat Wreck Chords. Wreck-Chords! Geddit? No one else seems to...&nbsp; urlLink      \"What kind of a pick up line is that?\" \"What did he say?\" \"Oh my god I can\'t believe it\" \"Ew, look at his hair...\" \"Band guys are so lame\" \"Why is he staring at us?\" \"Ew, what is he wearing?\"  Some girls are crazy...  hee hee hee. I forgot how much I love this song, and remember now why I bought a t-shirt with \"Some girls are crazy\" printed on the back. Not gonna tell you what the song is though, if you\'re interested, ask, or take a guess at who it is.  I feel less bored today. RS went well; the questions, thankfully, were on the areas of Buddhism I actually knew about, rather than random concepts like Buddhology (don\'t ask, because I still don\'t know). The last General Studies paper was today too, Society, Politics and the Economy. The pre-release material, which I spent last night reading, was all about the EU. Fascinating stuff, as you can imagine. Thankfully, one of the essay questions at the end was about whether I thought the wars against Iraq and Afghanistan were justified, and whether this action has helped to combat international terrorism, so that let me have a big 4 page rant about good ole George Dubya. He\'s the 59th most important person in rock music, according to Kerrang, as he\'s accidentally managed to force the music world into \"thinking about more than girls and beer again\". Fat Mike is 25th most powerful (Huzzah!)- I heart Fat Mike, he runs a wicked business and manages to be a punk rocker and (god forbid) intelligent at the same time. My dad seems to be unable to mix those two concepts; he thinks that if you listen to heavy metal/punk/emo/hardcore/anything other than bands that formed in 1970-73, you are a complete moron with no idea what\'s going on in the world outside the moshpit. Gah. But anyway: highlights of my new favourite Top 100 list (\"Most important people in rock in 2004\") include Avril Lavigne at No. 10 for \"prodding 5-and-ups away from the pathway of pop and into the road marked \'rock\'\", Justin Hawkins at No. 4 (\"make way for the anti-Kurt\") and Satan at No. 2 for still having \"all the best tunes\". I\'m not going to give away who\'s at Number 1, because I want people to start buying Kerrang- and anyway, it\'s only Â£1 this week AND you get a free punk CD! Bargainous. They still can\'t help putting Brody Dalle at Number 22 though, can they? Hmph.  Ceff x   What\'s with you and all your talk of independence? What\'s with me and my lack of innocence?   '),(17874,2004217,'22,June,2004','2004-06-22 00:00:00','    urlLink    The effects of revision...&nbsp; urlLink     I am so bored. Soooo boooored. So bored I think I might just fall off this chair and die, quietly, right here, listening to the new Billy Talent single (River Below), surrounded by notes on Samsara and Nirvana and the 3 marks of existence and fuck knows what else I have to revise before tomorrow. It wouldn\'t be so bad if:  a) my folder was better organised b) I had made notes from the piles of photocopied sheets that were thrust my way over the last year c) Mrs. Rogers was a better teacher, and hadn\'t simply thrust piles of photocopied sheets my way but actually taught us stuff we could make good notes from, thus making my folder better organised.  You know, whenever I bitch about her, I always feel like she\'s watching me... it\'s probably the RS teacher effect... *looks over shoulder*  Besides the woes of revision, I have made a surprising discovery: Fountains Of Wayne (you know, the ones that did that song about Stacy\'s Mom... you know, they had that Rachel Hunter woman in the video... oh, never mind) are actually quite good! Dad bought their album after hearing them on some \"New Generation of Young\'uns\" CD that came with his Classic Rock magazine, and it turns out it\'s a pretty cool, summery album. What is the point of Classic Rock magazine, incidentally? It seems that every time I pick it up all I find is a 1324254303 page feature on the excessive, debauched days of \"real Rock\'n\'Roll\" back in the seventies, or a list of the top 100 frontment/guitarists/drug takers (which is always inevitably won by either Freddie Mercury or Robert Plant). Surely they must have realised that they\'ve covered all that stuff at least once by now, right?  Talking of Top 100 lists (Warning: Rant ahead), I read 2 of them this weekend, both of which were the top 100 British albums everinthehistoryoftheworld, or something. The first in Dad\'s Q magazine (yes, he does buy every music magazine he can find) named Oasis\' \"Definitely Maybe\" as the best British album. PAH! Mediocre. How anyone can speak of a band with a frontman as twattish and talentless as Liam Gallagher with such high regard is beyond me. The same goes for Guns \'n\' Roses, by the way. Axl Rose is a big ginger knob. And where is the new GnR album, by the way? Now now, that\'s another rant for another day... The second Top 100 list was in The Observer\'s Music Monthly magazine, and listed its No. 1 album as The Stone Roses\' self titled CD. I don\'t know it, so I can\'t really judge, but there seems to be a trend forming for music publications to say the mid 90s Britpop movement was more influential, more important, even, than all the real, groundbreaking stuff that happened in the 60s and 70s like The Beatles, The Stones, Zeppelin (although I hate them I have to give credit where it\'s due) or Pink Floyd. Having said that (I told you it was a rant), The Beatles took up positions 2, 3, 6 and 7 in the list, which is just a little over the top, in my opinion. The Streets were in the Observer\'s list as well, position 70-something. This doesn\'t really surprise me as the Observer has its collective head up Mike Skinner\'s townie ass, but it still bothers me. I will not carry on for much longer (promise), but I will say this: MUSIC MADE BY THE STREETS IS NOT MUSIC. IT IS A SILLY MAN TALKING OVER A RANDOM BEAT IN AN ANNOYING ACCENT. PLEASE STOP LISTENING TO IT. Thankyou.  Ceff (going for a lie down) x   They\'ll think I\'m insane, but you\'ll all know my name...   '),(17875,2004217,'20,June,2004','2004-06-20 00:00:00','    There are 3 types of driver: those who are good drivers, those who are accelerator-happy and those who are brake-happy. Unfortunately, I fall into the last category; my driving lesson on Friday was good other than the fact that I stop the car about a mile short of junction lines, meaning that I can\'t see a thing and put both mine and my driving instructor\'s lives at risk when I turn a corner. Gah.  After much quick packing (which, incidentally, was totally inappropriate for this weekend as I have nothing to wear with sleeves longer than \"short\" and the weather is horrid), mother and I headed into the depths of West Bromwich in the direction of the (innapropriately named, in my opinion) Sandwell General Hospital so that a nice man called Mr. Chester could feel my lumps and stick an pointy instrument into me. Hehehehe. No no, nothing happened that could result in some sort of legal dispute: the horrible lump on the back of my ear got some steroids injected into it in the hope that one day it might shrink. Blergh.  Dad came to collect me from the hospital and now here I am, in his house, on his computer, talking to my friends from far far away. Dad says they\'re going to reorganise the house and move the \"spare bedroom\" downstairs and the study upstairs. Funny, I was under the impression that the \"spare bedroom\" was \"my bedroom\".  I should be revising Buddhism, currently. My exam is on Wednesday and I do not know  anything . Well, I know some stuff, but...meh.  Ceff x   All they know is how to put you down...   '),(17876,2004217,'14,June,2004','2004-06-14 00:00:00','    I\'m blogging out of habit, really. I have nothing of any note to write about, but as I\'ve been blogging about something most days recently, I sort of feel like doing it. I guess it\'s my one last attempt at not revising before my exam tomorrow afternoon. I\'m starting to get a little tense now: little knots in my stomach and that nagging feeling that I haven\'t quite done enough work towards it. Oh well; too late to worry now, isn\'t it?  Blogger won\'t let me update my profile. I want to add new bands and new interests and stuff... Boo :o(  I feel less disgusting today, incidentally. I have new trousers and they cover my leg-related imperfections sufficiently. I also think my stomach has shrunk a little. I\'m probably talking crap (as usual when I talk about my body), but still, I feel better for it.  Ceff x   Make a chick wanna hate, make a boy wanna stare...   '),(17877,2004217,'13,June,2004','2004-06-13 00:00:00','    Mmmm. Pimm\'s. My mum brought a bottle home to try out for the party, and I have fallen in love with it- Pimm\'s is now my official drink of choice, especially how I had it earlier, with ice cubes, strawberries and apple. Mmmm.  I went into Birmingham with Kat yesterday. We had loads of fun; I laughed and laughed and laughed, and as ever I missed all the JYF peeps as soon as we went our separate ways, but again I was left with that feeling of total inadequacy that the JYF girls seem to give me. I think they\'re all great and lovely and stuff, but I always feel so fat and unattractive next to them, and that my personality is dull and boring in comparison to theirs. They do fun stuff, go out to cool clubs, dress in really funky clothes and generally seem to have a lot more fun than I do. It doesn\'t really help that I swear I\'ve put on weight recently- it\'s all this sitting around the house, doing nothing except writing and eating Jaffa Cakes.  I tried on some bikini shorts in Marks & Spencer yesterday and my legs looked hideous: all lumpy and blotchy and flabby and disgusting. I will be surprised if, when I get to Italy this Summer, if I don\'t clear the beach of holiday-makers at the sight of my horrible white pins, or at least make them throw up into their beach bags. Blegh blegh BLEGH.  Due to the horror experienced by the sight of my disgusting, horrible, vomit-inducing body shape, mum and I decided to go swimming this morning, with the intention of making it a regular thing. But as it always is with us, our plan was unsuccessful. The University baths were closed from midday, and the Gala baths aren\'t open yet after some silly person set them on fire (incidentally, how do you set a swimming pool on fire? It\'s made of water!), so instead we went to Ann\'s (Godmother\'s) house and sat in the garden eating pretzels. Gah.  Ceff x   Obviously, she\'s out of my league...   '),(17878,2004217,'11,June,2004','2004-06-11 00:00:00','    urlLink    Eddtrix: This photo made me laugh out loud. There are more, but they\'re sort of inside jokes from JYF, so the humour wouldn\'t really transfer. I wanted to put one up though, just as my \"funny thing of the day\".&nbsp; urlLink     Kerrang! radio launched (was \"unleashed\", in their words) yesterday. When I first heard about it I was really excited; a brand new radio station that\'s gonna play songs by all the really cool bands they write about in the magazine, but, as ever, it was a bit of an anti-climax. It\'s pretty much like other local radio stations, except a bit heavier. Their playlist seems to only include these songs:  D12- My Band (or anything by Eminem) HIM- Solitary Man Jet- Are You Gonna Be My Girl The Rasmus- In The Shadows Evanescence- Nobody\'s Fool Nickelback- Figured You Out Blink 182- Down and anything by the Red Hot Chili Peppers.  Up to 5 of those tracks you\'d find on a regular pop radio station, so why aren\'t the broadcasters of Kerrang! being a little more daring? The evening show tonight is a hip hop show, which, if I wanted to, I could listen to on Galaxy, and there\'s no ska, punk, Britrock, hardcore, emo, nothing. I heard a Jimmy Eat World track at about 1am last night, but even then the presenter called them \"Jimmy Eats World\". The songs are the most commercial ones by each artist that they can find, and it\'s really disappointing because I rely on Kerrang! for at least some insight into alternative culture and new music from smaller bands who aren\'t on major labels with lots of money. They have this thing called \"Rated\", where they promote certain songs because they\'re really good. Are the songs by new, up-and-coming bands with a real chance of a successful future, or a band with something new to add to the scene? No: so far, the two \"Rated\" songs I have heard are the new singles by Nickelback, whose lead singer was once called a c**t by a Kerrang! reporter, and Linkin Park, who need no background information to show you how commercial they are. Overall, though, it\'s better than the other radio stations, and I\'ll listen to it because I\'d rather listen to Blink 182 over and over again than Natasha Bedingfield followed by Blue followed by Natasha Bedingfield followed by Blue. Ugh, now it\'s The Rasmus. AGAIN.  Ceff x   We left the radio playing all night long, because the radio\'s playing all my favourite songs...   '),(17879,2004217,'10,June,2004','2004-06-10 00:00:00','    urlLink    Photos from JYF are finally available! Here\'s me having lots of fun in Coventry with Verity and Kris. *grin*&nbsp; urlLink     Heeee. JYF in 22 days. Of course, before then I have to sit 9 exams and go to (groan) the leavers\' ball, but still, it\'s nice to have something to take my mind off them. I wish it wasn\'t Laura\'s birthday that day, things would be so much easier. Sigh. But I am pleased with the photos - I like the JYF piccies because they give me a good way of tracking how much weight I\'ve lost; there are photos from Jan 2003 all the way up to March 2004, so it\'s a great gauge of how I\'m doing. And they make me laugh out loud, which is always good for your health.  I had a terrible driving lesson today. I started off quite well doing a three point turn, but then it all seemed to go downhill from there as I seemed to forget how to operate the car; I forgot to indicate, position the car for turning and use the handbrake. Gah. Next week we start parallel parking, which should be... interesting...  Ceff x   There\'s something obscene in a dream...   '),(17880,2004217,'09,June,2004','2004-06-09 00:00:00','    urlLink    I heart David and Goliath!&nbsp; urlLink     Hehehe. I love photoblogging.  Ceff x   And I never know where I stand...   '),(17881,2004217,'08,June,2004','2004-06-08 00:00:00','    urlLink    There he is: the source of many a discussion and debate, both personal and public. Hmm.&nbsp; urlLink     Rich just came online. Arggghh! I\'m all shaky and my heart went all funny and now I\'m thinking about him which is bad... the prospect of having to actually see and talk to him at JYF is becoming all the more real every day and now he\'s online it\'s suddenly very apparent to me how easy it could be for him to fuck me up just before my exams. Breathe. Calm. Think happy thoughts. Arrrrggggggghh!  Ceff x   My brain starts tickin\', and then I saw what you were thinkin\'...   '),(17882,2004217,'07,June,2004','2004-06-07 00:00:00','    urlLink    Me at dad\'s wedding (that\'s Em in the background). I\'ve just downloaded Hello Bloggerbot or whatever it\'s called, so I thought I\'d add a picture; I like this one because I think I look quite pretty, as opposed to other pictures where I tend to look fat, or I have a stupid grin on my face. Don\'t ask about the thing on my head. Long story.&nbsp; urlLink     The sun has returned, and I feel distinctly less down. I was up until about 2.30am last night feeling sorry for myself (and also very warm, strangely enough), and I wrote a little, which made me feel better, like I was getting it out of my system. I think my mood has also lifted partly due to talking to a guy called  urlLink John  yesterday, who made positive comments in my direction, which made me feel a bit better. I\'ve still downloaded the latest Linkin Park album though, just in case.  I have spent today sizzling (and revising) away quietly in the back garden, smiling politely and telling the lady from No. 11 for the fourth time (at least) that I\'m going to either Leeds or Sheffield in September. In a mo I\'ll shuffle into the bathroom in my fish-shaped flip flops (Heeeee) to apply copious amounts of Sainsbury\'s After Sun Tan Maintainer, in the hope that tomorrow morning (supposedly the hottest day of the year so far) I won\'t wake up equally as white as I did today.  Ceff x   I don\'t know what\'s worth fighting for, or why I have to scream, but now I have some clarity, to show you what I mean   '),(17883,2004217,'05,June,2004','2004-06-05 00:00:00','    As years go by I race the clock with you But if you died right now, you know that I\'d die too I\'d die too.  You remind me of the times when I knew who I was still the second hand will catch us Like it always does..  We make the same mistakes I take the fall for you I hope you need this now Cause I know I still do...  Until the day I die I\'ll spill my heart for you Until the day I die I\'ll spill my heart for you...  Should I bite my tongue \'till blood soaks my shirt? We never fall apart. Tell me why this hurts so much?  My hands around your throat and I think I hate you Still they say we never win Just like we always do. Just like we always do...  Until the day I die I\'ll spill my heart for you Until the day I die I\'ll spill my heart for you...  I\'d spill my heart I\'d spill my heart for you...  My hands around your throat and I think I hate you We made the same mistakes, mistakes like friends do... My hands around your throat and I think I hate you. We made the same mistakes...  Until the day I die I\'ll spill my heart for you Until the day I die I\'ll spill my heart for you...  Until the day I die Until the day I die...   What an interesting thing it must be to see someone sink lower into their own self pity, post by post. My timing is impeccable: the need for a boyfriend, or at least someone to just *be* around, comes during study leave, the period when you have the most time to just sit and think (because you\'re avoiding doing anything productive), but your friends aren\'t around to distract you because they\'re probably sitting at home, doing the same thing. Maybe the need for closeness to someone has come from not having those distractions... I really want to pull myself out of this mood, but it seems warm and comforting; I think part of me likes it, you know? Being able to sit at home, moping around like a greasy grunge maniac, listening to Linkin Park and Story of The Year and Seether, it\'s a bit like a naughty secret -  nobody else knows that I feel like this (or that I\'m listening to such obscenely commercial music), so I just want to keep it this way.  Yeah, I like that excuse.  Ceff x   I don\'t want to be the one the battles always choose, cos inside I realise that I\'m the one confused   PS. A quick note on the song: It seems to fit my own relationship with relationships/love/soppy feelings in general, rather than a specific person, whose name I will not mention. Although it does work on that level too. Except for the \"until the day I die\" part. I hope.   '),(17884,2004217,'03,June,2004','2004-06-03 00:00:00','    I wonder: Am I unattractive, or have I just not met anyone who thinks I\'m attractive?   If it\'s that I haven\'t met anyone who thinks I\'m attractive, then why do the guys I know not find me attractive?  Maybe they don\'t find me attractive because they don\'t like my personality. Or maybe they don\'t find me attractive because I\'m not.  But why? What do the guys I know see in other girls that they don\'t see in me?  Just a thought.  Ceff x   In the heat of summer sunshine, I miss you   '),(17885,2004217,'02,June,2004','2004-06-02 00:00:00','    I have not had sex in 11 and a half months. Nearly a year. Nearly a whole  year  has passed since I have been further than a fumble under a sleeping bag, or felt anything more than pure horn. Up until now I haven\'t wanted a boyfriend, in fact, I still don\'t want to get into anything more than a casual (but exclusive) relationship, but I\'m starting to miss the little things; holding hands, little kisses on the lips, even just having someone to think about when there\'s nothing else to ponder on. And to coincide with this, I find out that Rich has split up with Kayla (joy to the world!) and is coming to JYF. Great. What if he hits on me? I know myself- despite all my best efforts to hate him with all my heart when he\'s out of my eyeline, inevitably, as soon as I see him it will start to beat faster, my tummy will go all fluttery and I will think \"Well, maybe he\'s not so bad after all...\". I know what I\'ll do: I will wear my new yellow halter top, stand tall and walk straight past him. And perhaps give him a swift kick in his sleep.  Em is once again in Dad\'s bad books. After her \"triumphant\" return on Saturday all was well in Worcester (ha), until Em phoned Lee (current squeeze) and discovered it was going to take her about 4 hours to get back to Coventry. Despite my telling her it would take less time and be cheaper if she went on the coach, she refused to ask Dad for a lift to the coach station and was in a funny mood for the rest of the night. She sat up in my (our?) room for hours watching TV before I came up to bed and she had to turn it off. On Sunday morning she got up, got dressed, packed her stuff and then just *sat* in the bedroom for ages. Nuts. I got up to brush my teeth, then went downstairs to get some brekkie (mmm croissants) and found her half way out the door; \"Bye hun [I hate it when she calls me that]\" she said, \"I\'m off now.\" When I asked her whether she\'d told Dad she was leaving, she said \"No, it\'s OK\" and then she was off to the train station. \"Freak\", I said to myself as the door closed. Needless to say, Dad and Wicked Stepmother were not best pleased when I related the tale to them, and Dad said he found her behaviour insulting and told me to take whatever clothes I liked of hers out of her drawer, as he was going to throw the rest out because this \"isn\'t just some doss house that she can stay in when she likes\". So now I am the favourite again. Yay.  Ceff x   The fuse is burning somewhere, it\'s drenched in heat...   '),(17886,2004217,'31,July,2004','2004-07-31 00:00:00','    I came back from MAGIC (Midlands Area Guiding International Camp) today! Huzzah. It was my first Guide camp as a leader and I had a surprisingly good time and fitted in with all the proper grown-ups quite nicely. I\'ve missed blogging though, and talking to my mates, despite access to my mobile, a privelege not usually synonymous with camping. Here\'s my week in short:   Day 1  Busy. Tired but happy. Thinking about Draper.   Day 2  Crap weather! Knackered again but having a laugh. Rubbish opening ceremony involving a stripping fire eater. Am on the lookout for pretty boys. So far: none.   Day 3  Weather improving. I have a cold and Lorraine (other assistant Guider) is contagious. The boy band was v amusing, involving many cover songs and screaming girls. I realised Kati (guide) would probably never go to a gig due to her epilepsy. Scary. Janet (main Guider) said I was useful. Grin.   Day 4  Very relaxed today. Royal visit from Sophie Countess of Wessex. Missing male company. Still no pretty boys on camp, just some who stink of cheap aftershave that the Guides fancy and they know it. Played Twister and won! Twice! Huzzah.   Day 5  Shock horror: good weather. Campfire. Meh. Text from Sam! Yay. Selina says she will buy me a dictionary of euphemisms. Hurrah.   Day 6  Lots of sunshine followed by lots of rain. Thinking about Sean.   Day 7  Last day! Ace closing ceremony, but missing mum. Fear re: striking camp without Janet.  As it turns out, striking camp wasn\'t so bad and mum and I are sucessfully reunited. This summer\'s next Guiding task is to set up a Senior Section unit for the older guides before I leave for Leeds. But before then, I need to bathe, sleep and go to Italia! I can\'t wait for Monday; I know I\'m going to miss certain *people* while I\'m away (due to my laziness I have not activated my phone for foreign use) but I need a break from dull and dreary Walsall, and sizzling quietly on a beach in my favourite European country sounds just about right. Then I can return, become 18, a student of Leeds University (MEEP) and spend the last few weeks of the longest holiday EVER hanging out with my mates. Ace.  Incidentally, I am covered in insect bites. Definitely not ace. Notice how I have refrained from adding in a bad joke about wanting a nice young man to rub after-bite cream into me... oops! ;o)  Ceff x   We can live like Jack and Sally if you want, where you can always find me...   '),(17887,2004217,'24,July,2004','2004-07-24 00:00:00','    urlLink    So...tired...&nbsp; urlLink     I am far too sleepy considering I had a good 6 hours sleep last night. I went to Netherstowe School again to see Third Enigma with Kat, Draper and someone called Andy, then stayed over at Coz\'s. I didn\'t even have anything to drink or (ahem) over-exert myself, so God only knows why I\'m totally shattered now. The band were ace again, but I didn\'t expect any less, in fact, they were better this time around. The rest of the evening consisted of defending Hells Is For Heroes to too many people (Blink 182?? EMO??? Tis not normal...), stressing about coats and being disappointed at the non-stunningness of Draper\'s friend. When we went back to Coz\'s we watched about 3 films which all seemed to merge into one, leaving us with the feeling that we\'d watched a film where a boat with Dean Cain on it sank, before he climbed into the air vents of a house that was on fire, only to see a woman being chased by Jedis/Dan from Big Brother before she stabbed the bad guy in the eye with her nail file. All excellent films, I think you\'ll agree...  So... Draper. Think. It was a half-and-half situation for me last night: half the time he was annoying and obnoxious, and half the time he was strangely attractive (surprisingly good music taste- always does it for me). Despite what everyone says, he\'s a nice guy; he\'s chatty and can be quite gentlemanly when you say the right thing (\"I\'m cold\" = \"Do you want my coat?\", \"I need to get dressed\" = \"Do you want me to go out?\"), and although sometimes he makes comments that are waaay innapropriate and kill the conversation, when you think about it, everyone does that sometimes, it\'s just he does it a little more than other people. Give him a break! He\'s not quite ace, but he\'s alright...  Ceff x   As you sleep, when no one is listening I will lift you off your feet, and keep you from dreaming...   '),(17888,2004217,'22,July,2004','2004-07-22 00:00:00','    I spent most of the day with Sean today. Sam and his brother Joe (who is ace, incidentally) were there for a couple of hours but as Sam fucked up his back (mysteriously, no one knows  how  he did&nbsp;it), he had to leave at about half past 2 and so Sean and I just hung out for another couple of hours, wandering, chatting, browsing CDs. It was really, really nice; I felt so comfortable with him, like I was able to just be  me.  We talked about everything from my parents to his parents, from The Streets to Slipknot, and it all clicked: while we don\'t like the same bands, we agree (mainly) on the bands we don\'t like, and we seem to be on the same sort of level morally/politically, as well as having&nbsp;a similar&nbsp;sense of humour. And he lets me link arms with him! Most guys just let their arm sort of drop to their sides, leaving me sort of *clinging*, but Sean let me get comfy and fall into step. It made me feel valued, and thus warm and fuzzy. Huzzah.  How odd it is that I\'m now comfortable with writing about my feelings for other people in my blog... until recently I\'ve just talked about me for fear of offending or scaring people, or giving them the wrong idea. But hey, this is me, this is what goes on in my head, and I have enough language capacity to explain anything, should questions be raised. Oh my god: have I finally grown up?  Ceff x   I don\'t care if you don\'t love me, and I don\'t care if you don\'t change, I would live inside the shadows that I cast for you, if it meant that you would stay   '),(17889,2004217,'19,July,2004','2004-07-19 00:00:00','    Phew! What a hectic couple of days that was. I went swimming in the morning for the first time in ages, it was really refreshing and am considering making a habit of it. I returned home for a leisurely afternoon of internet, Jaffa Cakes and webcomics, before all manner of chaos occurred outside as Dad arrived to collect me just as my mum returned home with Rod and a new table top for the dining room. Dad was \"helpful\" and carried the table top up with Rod, meaning that he was in THE FLAT, with MUM and ME, and it was all very, very odd. I finally managed to extract Dad from the flat and we headed off to the speedway, and then to Bournville for Kat\'s \"my parents are in Wales\" house party. I consumed excessive amounts of alcohol and attempted to smoke but failed, before heading up to bed at about 3.30am with Kris and Vez to share the double bed in the spare room. I woke up far too early and then at midday we went into Birmingham before I had to dash home, grab Laura\'s birthday present and dash back out again to Walsall to meet aforementioned Laura, Selina Debbs and Becky for drinks and exchange and discussion of presents. Early night for Ceffy tonight, methinks!  &nbsp;  Sam and Sean were at the party last night, but they had to leave early because they have parents who are sensible. Sam and I seem to be really good friends at the moment, which is&nbsp;cool because we seem to really be just friends; there\'s no underlying sexual tension between us, we can chat and flirt a little and it\'s all in good fun, which is ace. Just what I need. Sean\'s different, he\'s still making me think, and I don\'t know if that\'s good or bad. But still, he\'s ace too. *grin*  &nbsp;  Ceff x  &nbsp;   I\'m having trouble saying what I mean, with dead poets and drum machines   '),(17890,2004217,'18,July,2004','2004-07-18 00:00:00','    I watched \"Never Been Kissed\" last night. Geeky 25-year-old career type goes \"undercover\" at a high school to see what the kids are up to nowadays. She makes friends with the \"geeks\" and then becomes \"popular\" and gets to date the hottest guy in school and becomes prom queen (ahhh). Along the way she falls for her \"teacher\", who happens to be drop dead gorgeous, intelligent&nbsp;and&nbsp;willing to confide in her when it comes to his own relationship problems, and conveniently breaks up with&nbsp;his girlfriend a week before the prom. After the \"popular\" kids try to play a dirty trick on the \"geeks\" at the prom, she blows her cover and tells everyone what she\'s up to, and how insignificant high school is in the grand scheme of life, the universe and everything. Then she writes a big sentimental piece in the paper, and waits, out on the baseball field for the teacher to come and kiss her. He arrives (in the nick of time!), they kiss, the bad guys get what they deserve, the good wins out and everyone lives happily ever after. Total shit. &nbsp; Ceff x &nbsp;  You were all the things I thought I knew, and I thought we could be   '),(17891,2004217,'16,July,2004','2004-07-16 00:00:00','    urlLink     My glorious new welly, that I painstakingly customised using only nail varnish. They need to make bigger brushes for that sort of thing...&nbsp; urlLink     &nbsp;  Wow! My guide leader just texted me! Very impressive. My dad can barely grasp the concept of predicted text, let alone the punctuation key- he texted me recently to tell me he was going to see Harry Routes at the cinema, which, if you translate it correctly, actually means Harry Potter.  &nbsp;  I think I\'ve become addicted to e-bay. I went on last night to see if I could buy some  denim   dungarees  (I lost out in the last minute when some bastard swooped in and bid about 20p more than me, just as my mouse packed in and 203493004 popups appeared), and have now bid on 2 limited edition Less Than Jake stickers (currently winning), a  bright red tutu  (also currently winning) and a limited edition New Found Glory sticker. I sort of cheated on the NFG sticker though, I didn\'t bother to bid and just clicked the \" Buy It   Now! \" link. Should be here in about 2 days! Hooray. I\'m currently perusing shoes, and have found a pair of gorge   pale blue Ugg boots   for Â£15. Mmmmm wantwantwant.  &nbsp;  I\'ve also, as you can see, been putting my  creative skills  to work again with my new   wellies  , which I plan to take to camp and Leeds. I bought them for a bargainous Â£10 from a work clothes place in Worcester (\"Jobtogs\"- how great is that name??), so they\'re made of pretty sturdy stuff, especially the toes, which are steel-capped. Great for moshing, as I can inflict pain without receiving it myself! Score. Perhaps owning a pair of boots such as these will qualify me adequately to work on a building site... Oddly enough, images of the second series of  Dirty Sanchez  spring to mind. Eek.  &nbsp;   Ceff  (using new Blogger tools successfully!)  x   &nbsp;   But who cares? No big deal... I want more...   '),(17892,2004217,'15,July,2004','2004-07-15 00:00:00','    Hola! I return from the south coast of England, sunburnt and sleepy. We (me and Dad) stayed in a hotel in Poole for a couple of nights, and we visited Southampton, Isle of Wight and, um, Poole. Southampton has the most gorgeous quay, with this big factory outlet shopping vilage Cheshire Oaks-esque thing, complete with Vans outlet. I wandered in a tortured myself for while: found some white trainers with a blue stripe. Â£30. Hubba hubba.  The Isle of Wight is the strangest place I have ever been to. Well, no, that\'s not strictly true; it has the strangest looking people on it that I have ever seen. There was an old man at the speedway that barked. Yes, he went up to some people sat in front of us (I assume he knew them) and just, well, barked. There was also an old man that had a huge boil on his nose and a whole variety of other, weird and not-so-wonderful characters. I was glad to get back on the boat, to be honest. The speedway was good, though; I never knew Premier League racing was so competitive, and the track was awesome for speed.  Poole is not so nice; the quay is half smart, new apartment blocks and yachts, and half covered in bird poo and inhabited by grubby pubs and tacky souvenir shops. A seagull pooed on Dad\'s head though, which amused me. We sailed over to Brownsea Island, which was reeeaally nice but not a place to wear heels (ouch). Lots of greenness, great National Trust tea shop (what NT tea shop isn\'t great?) and very nice views. There were no barking men (thankfully), but some suspicious peacocks that followed us around and glared at me with their evil eyes. Meep. On the boat back the sky was deceptively cloudy, as by the time I got off it I looked like someone had tattooed my face using beetroot juice as ink. The speedway, again, was ace (even though the view wasn\'t too great), and Dad and I spent an hour or so in the bar afterwards playing cards. He tried to teach me a game called Crib- I still refuse to believe this is a real game, and am convinced that he made up the rules as we went along so he would win. I whipped his ass at Rummy though, because it\'s a  real  game.  I\'m waiting for mum to come home now. It\'s been a week since I\'ve seen her and I miss her a lot, especially seeing as spending extended periods of time with Dad reminded me exactly why she\'s my favourite of the two. In my slightly drunken state on Sunday night, however, I texted her to tell her to order tickets for \"Sing Along Moulin Rouge\", which I may regret later, probably on Saturday night, when we end up going...  Ceff x   The boys are back in town, no strings to hold them down...   '),(17893,2004217,'09,July,2004','2004-07-09 00:00:00','    Sigh. It\'s sad, isn\'t it? Blogging not because you want to display your thoughts to the world, but because you don\'t feel like you have anyone else. And I\'m not saying that because it\'s late, but because I actually feel as if this is the only place that\'s available, the only place that won\'t say \"Sorry, gtg byeeeeeee\" when I need it most. Everyone else has lives; friends, girlfriends, boyfriends, exciting things to do, and they never seem to have time for me anymore. All I need is a hug and for someone to spend some time with me because they\'re my  friend , not because they\'re in my base group or aren\'t allowed to behave \"exclusively\".  Maybe I\'m listening to too much Morrissey...  Ceff x   They said they respect me, which means their judgement is crazy   '),(17894,2004217,'07,July,2004','2004-07-07 00:00:00','    Ho hum. Another late awakening, another wasted day. After voicing some concerns to Sean last night, said concerns have been playing on my mind a little. I\'ve mentioned it before on here; the total feeling of inadequacy I get when I\'m around the JYF girls. They\'re all small and pretty and size 10, they swap clothes with each other and get all the attention, and they\'re so confident and sexy. When I\'m around them I just stick out. I\'m the \"sensible\" one; I don\'t have the quirky fashion sense or the confidence, I can\'t swap clothes with them and I don\'t get the attention they do. I\'m don\'t want to be like them, as such, and I\'m not craving for attention, but I think I\'d just like to fit in with them a little more, and the fact that it\'s actually physically impossible for me to be half those things makes it even more annoying. The Rosy Glows say it all: on all of theirs, the lads say \"it was so great to see you, you\'re gorgeous/wonderful/sexy/so cool and I can\'t wait to see you again\" and on mine they say \"hey! Good to see you again, we really should meet up sometime\". Sean told me not to be silly and that I was pretty too, and that Sam could back him up on this, but I don\'t feel it. They see me in a totally different light from the other girls. They\'re potential girlfriends, or people they might pull if they get the chance, whereas I\'m just their mate, good for a laugh, nice to talk to. And when that\'s made blatantly obvious from what someone\'s written on your rosy glow, no amount of their telling you you\'re pretty is going to change that.  Ceff x   I never will be good enough for her   '),(17895,2004217,'06,July,2004','2004-07-06 00:00:00','    Woooo! Wicked Stepmother\'s parents have bought me 2 tickets to go and see New Found Glory!! Yay!! Sel and I are off to marvel at the glory of Jordan & Co. once more, and I am all the happier for it. *grin*   Dad and I had one of those wicked \"Ultimate\" conversations last night- you know, they start with \"What\'s your ultimate...\" and end with \"CD compilation/speedway team/football team/car?\". This one was \"What\'s your ultimate festival lineup?\". The rules were simple: 3 days, 2 stages, 2 headliners on the first stage, 1 headliner on the second. He chose The Who and Pink Floyd/Ash on the first day, Led Zeppelin and Rolling Stones/The Police on the second day and U2 and Bruce Springsteen/Free on the third. You can tell it\'s a fantasy lineup, as roughly half of Free, Led Zeppelin and The Who are dead. My festival was this: Muse and Lostprophets/Hells Is For Heroes on the first day, Michael Jackson and Less Than Jake/The Smiths on the second day, and Rancid and The Prodigy/Jack Off Jill on the third. Much better, I think you\'ll agree... ;o) But anyway, back to reality...  I have just had the busiest weekend. Ever. Friday night was the Leavers\' Ball; we all arrived in our finery and within 15 minutes, I had white wine down me. Well, at least it wasn\'t red, seemed to be the generally acceptable thing to say. The night improved from there, thankfully; the music wasn\'t  too  bad, the food was good (at least, the profiteroles were- that was all I ate) and I got really quite drunk. By the end, I was dancing with Draper. I don\'t know why, but I always seem to be drawn to him at things like this. And it wasn\'t as if he became some sort of gorgeous sex god through my haze of vodka, when I saw him, sharp suit, less *curly* hair, new glasses, he just made me... think. Yes, that\'s a suitably non-committal word. Think.  So, the night went on, and midnight came and Lucy dragged me out even though I didn\'t want to leave, and I went home and to bed, where I didn\'t sleep. I don\'t know why, but I just didn\'t. Not until about 5am, anyway, which was a bit pointless because I was up again at 7am to get the bus to Nottingham. I was met at the station by James, who promptly whisked me to UGC Nottingham where everyone else was to see Shrek 2. Lots of hugs and kisses and \"let me carry your bag\"s occurred, and then we went in to see the film, which was absolutely hilarious. Go and see Shrek 2, I\'m a total convert to the whole thing, after hating the first one. We went back to the meeting house, and lots of games were played and diaries kept and cuddles had. Sean and I got really close this weekend, brought together by various things I never knew we had in common, he let me share his duvet and lean into him when we sat down... he made me think, too.  So now I\'m back at home again, far, far away from everyone I want to be close to, missing them all like crazy, and none of them are online. I want a hug and a kiss and to keep my diary and play Honey If You Love Me or Ratchet Screwdriver, but I can\'t because no one else gets those games, or the way we are at JYF. 17th September seems so far away...  Ceff x   It\'s just a little crush, not like I faint every time we touch   '),(17896,2004217,'01,July,2004','2004-07-01 00:00:00','    I feel so drained today. I don\'t know why, as I\'ve done nothing that\'s really required a lot of energy. Well, I had a driving lesson; they always seem to take it out of me, seeing as they\'re so damn stressful. We did awareness, anticipation and planning today, all of which I was rubbish at. I\'m so envious of people who can drive.  The Leavers\' Ball is tomorrow. It\'ll be the last time I will see a lot of people with whom I\'ve spent the last 7 years co-existing, so it\'s going to be happy, sad and strange all at the same time. Maybe in 20 years I\'ll run into someone in the supermarket, someone I wasn\'t that close to, and we\'ll squint at each other, and say each others\' names unsurely and then have an uncomfortable conversation about our lives now, and whether we still talk to anyone from school, which we probably won\'t. Alternatively, even worse, I could run into Laura/Becky/Debbs/Selina, with whom I\'d lost contact with over the years, and we had nothing to talk about. I think I might just move to Leeds. And then JYF finally occurs on Saturday! Huzzah. My sorry ass will arrive in Nottingham at 11am, hopefully, and then for the next 36 hours much fun will be had by all. Twice.  So here ends my first day as a proper grown up. Thoroughly wasted, I think you\'ll agree. Only 3 more months of them to go...  Ceff x   If I only had a notion...   '),(17897,4238678,'20,August,2004','2004-08-20 00:00:00','   woohoo! I just finished my Chinese H.E.!... Well, ok.. I still have to color two more fruits, but that\'s nearly done, right? LolÃ¼ haha! I just heard from my friends that my section won the newspaper drive! yeahhÃ¼ I\'m so glad to hear that! lolzÃ¼ well, g2g! pizza\'s there! yumyum!Ã¼ nytie! big day tomorrow! I\'m gonna watch Alien vs. Predator tomorrow night too! Yeahh!  '),(17898,4238678,'18,August,2004','2004-08-18 00:00:00','   Damn.. I wasn\'t able to post for like what? The past two days or one day lang ba? Can\'t remember. Haha. Still, I had lotsa fun. Even today! Haha! We got a bunch of newspapers from Bianca Cabilangan aka Biba! Thanks to her, we might win the newspaper drive!  > We also had cheerdance practice during P.E.! It was sooo much fun! Hahaha! Then, I got a part in our group guidance play! I get to be Eine.. The hot guy, but he\'s sick and is always in bed. He has a pretty small role in the play, but it\'s fine with me^_____^ Hahaha! I\'m starting to really love school! Despite all the homework, I have my friends and not to mention the whole class of UNITY XDDD I love this class! It totally roxx! This is my favorite class ever since I came to ICA! I wish we will always be classmates, but I know it\'s forbidden.. hayy.. How sad=(=( I LOVE THIS DAYYY!! Hahaha! YEY! Soooooo much fun and excitement in school XDDDD Grrrr... Too bad me and my groupmates have a research for Civics that\'s worth 40 pts! Research on Manila, Quezon City and Pasay City! Gawd.. Mrs. Ranada\'s killing me. Then, I also have a chinese quiz on Monday and this chinese h.e. thing... I can\'t find some fruits that I\'m looking for on the net! Dammit. I\'ll just invent and draw those 2 stupid and unidentified fruits. Hahaha. ~~      Man! It\'s group guidance next week and I haven\'t bought the materials that I need! Darn. Where can I get that thick thick rope? SShoot. Good thing I don\'t need to buy those google or goggle eyes anymore since Arilia will provide it for me, but I have to pay 5 pesos XDDD Fair enough! Hahaha! Well, g2g! Still chatting and kinda busy.. be back latah.. ^___^  '),(17899,4238678,'18,August,2004','2004-08-18 00:00:00','   Daaaang! I got -19 in my Civics PT! The lowest grade mann!!! Damnnnnnnn!!!!!!! GRRRRR!!! Waaaaaaahhhh=(=(=( Haha Lol. At least I didn\'t fail. Anyway, g2g go! Dinner\'s ready..  '),(17900,4238678,'17,August,2004','2004-08-17 00:00:00','   Hehe wala lang... I\'m bored. Too lazy to the my other stuff. HAyy.. All I can do is daydream.. That\'s what I\'ll do then! Gotta go! Don\'t want my mom to catch me! XDDDD I wanna be free! Free my emotions tucked deep inside my heart! WOoH XDDD  '),(17901,4238678,'16,August,2004','2004-08-16 00:00:00','   Weeeeee!! I got -1 in my CLE PT! That still suxx \'cause I had a stoooopeeeed mistake. Arghh. Well, at least it\'s high na! yey!Ã¼ BUT! Before 3rd period.. I was veryyyy nervous! I thought I\'d be getting the results on my Civics PT! When it was 3rd period, I felt so nervous! *shivers* and I panicked and panicked and panicked.. It turns out that my teacher was absent or late T.T She let us do this seatwork then.. Damn. I wanna see my grade so I can stop worrying already! To continue on to my day at school.. It was okay. I had fun! Meg and I were passing notes during some classes. We had this activity in reading that could increase our PT grades.. It was hard, but I managed. Didn\'t answer 2 quessies though. Dammit. T.T During lunch period, me and my friends started talking about icky and gross stuff! We were all green-minded and horny, but it was still so much fun! I laughed a lot. Ha-ha. We even made the Hockey Pockey song very green-minded. Haha! That was soooo much fun! XDDD Then, our fun ended since classes resumed.. ~~~~ Last period. Music. A whole new group since it\'s a new quarter. They picked me as a leader of my group. I know I don\'t lead well, but I\'ll do my best for my group! ^___^ Lol! Then, dismissal.. I picked my cousin up in Xavier.. Boy, was he sooo mean! He kept on whacking me and the driver with the fan! He told me to shut up and I\'m stupid! *ouch!* He\'s only grade 5! He\'s been bullying me since we go home together.. Dammit. I fear him. EVIL has come. Someone save him. I\'m only safe when my achi and his achi are there. They stop him. He almost made me cry a while ago. Whacking me on the face! Torturing me. I don\'t want to ride with him when going hom from school anymore. It\'s just..PAINFUL. Finally, I got home..SANCTUARY! SANCTUARY! There was nothing to do since I finished my homework in the car. My bro told me both of us couldn\'t use the computer \'cause of what happened last night. Darn. Hayy.. But I\'m using it now. I\'m sneaking. Har har XDDDD I need to do 2 pads of my Kumon! Darn it again. T.T I\'m suppose to submit articles to our class mag. Haven\'t done it yet. Darn it again. T.T Haha. I need serenity! Woohoo! Wait! When I went to my parents\' room a while ago, the Tv showed Chad Michael Murray\'s face! I was like! Woo! Woo! Woo! That nasty girl he\'s with was doing some nasty stuff with him in the merry-go-round horsey thing.. She kissed his muscles! Goddammit! She\'s so unfair! URghhhh!! Well, g2g run! I have stuff to do pa.. Argghh.. Later!~~  '),(17902,4238678,'16,August,2004','2004-08-16 00:00:00','   Howdy ho!^__^ Wow.. I was pretty nervous this morning \'cause i\'ll be getting the results of my PT! I didn\'t get my Civics and CLE results yet.. Phew.. Civics is killing me! Well, for Reading I got -5, language -8, math -8, science -9, filipino -5 and chinese -6.75. Boy, my scores are pretty weird.. 3 subjects I got 8 mistakes? Bummer.. Just because there were stoopeeed corrections. My grades would be a little higher. Hahaha! XDDD Damn! I have to make a stooooopeeeeed speech for Filipino and it\'s worth 20 pts.! Doesn\'t Gng. Versoza have anything to do than bored us to death? I fell asleep during her class a while ago! Daaang. We were copying down notes that time! Good thing when I woke up, I didn\'t skip anything. That nap was toooooo short. Heehee. Homework is killing me. Damn. Gawd.. i\'m so bored. I wish I\'m living in my own fantasy which is better than reality. Lol. CHAD MICHAEL MURRAY would be in my fantasy. I sound too desperate for him. Ughh.. I\'m confused.. I don\'t feel happy all of a sudden. It\'s like my life isn\'t....right. Total weirdness I say. I\'m signing off..  '),(17903,4238678,'15,August,2004','2004-08-15 00:00:00','   what a day.. haha! My Sunday started out good.. I felt a little change in me..(not physically). Like I wanted to become a better person! I wanted to lessen my texts because my bill was really WOH. High. My dad is kinda angry at me for that. So I\'ll lessen my texts. Then. I got to talk to my speshie for a while.. hehe! I love him sooooo much! He\'s like a best friend and like an older bro to me!Ã¼ Then, after a lunch, my cousin came over and playd with my bro while me, my sis and my parents went to Podium! We bought stuff there and did something with my celly! Then, me and my sis watched A Cinderella Story! (well, it was my 2nd time..) Ahhhh.. I love that movie! CMM\'s there! He\'s soooo hot! Gawd.. t would be my wish to spend a day with him! He\'s soooo cute! The movie totally rocked. Sad,funny,perfect,fantasy and everything!Ãœ B-E-A-U-T-I-F-U-L! WOOOOO! After that, I went to the 6:30pm mass. The mass ended early.. like early by 20 mins. or less. Haha! And now, I\'m here.. Writing my journal.. what happened today.. Too lazy to do my kumon.. haha! And of course..waiting for my speshie to text! (if he will..) Well, I just hope he will. I really love him^_^ As a friend.. He\'s different from other guys..Lol!Ã¼ Well, that\'s all for tonight\'s journal! Haha bye, Chaddy.. I\'ll name you that..  '),(17904,4238678,'15,August,2004','2004-08-15 00:00:00','   CMM! hot, eh?      WOOHOO!! CHad Michael Murray mannn!!!! Later!  '),(17905,4238678,'15,August,2004','2004-08-15 00:00:00','   hey! my first post in my blog! well, i really suck in writing stuff so please excuse some..*ahem*...maybe a lot of my errors coming your way.. lol!=) i\'ll be posting more later.. gonna tell you about my day! ^__^  '),(17906,4022853,'27,July,2004','2004-07-27 00:00:00','   There is one person who reads my blogs and who can figure out who I am.  I\'ve given many clues as to who I am,  you just have to read the comments and look closely.  This may be my last blog, when you find out who I am, I have one last thing to say to you, but you have to find out who i am.  Its been fun and all, but hiding behind a name like Dark Angelique gets boring. '),(17907,4022853,'26,July,2004','2004-07-26 00:00:00','   Do you think aliens exist?    I personally do.    I am also a skeptic and a religious person.    It can get very confusing.    As for aliens, how can one deny the fact that life canâ€™t exist somewhere in this vast universe.    The proof is the fact that someone is reading this at all.    Humans exist in this universe and if humans can exist then why canâ€™t other beings exist also.    I mean, what are the chances that our planet is the only planet in this entire existence that is capable of supporting life.    UFO sightings are too common to all be fakes.    I mean, yes, there are people who claim to have seen UFOâ€™s and fake UFO sighting with a simple string, a tin object and camera or video camera.    These people are just looking for attention.    The sheer amounts of sightings though are too great too all be hoaxes.    As for whether aliens have landed, I donâ€™t believe it.    I donâ€™t think intelligent beings would risk an encounter with a species as destructive as us.    As for abductions, Iâ€™ve heard theories.    There is a particular one Iâ€™m very interested in.    Basically it states that these aliens are so advanced and evolved theyâ€™ve lost the ability to reproduce.    To put it simply, theyâ€™re dying.    They need our DNA and to re-sequence their own DNA to allow for reproduction.    It is also said that there will be a side effect to this.    These beings are so evolved they have lost the ability to love at all.    It is this lack affection that allows them to do the experiments they need on us to keep their way life alive, but as I was saying before the side effect of adding our DNA to their DNA is they may love again.    Itâ€™s all in our genes, how our hair looks in the brilliant sun, the way our eyes look when the moon hits them just right, and yes, love.    Itâ€™s all in our genes.    For the first time in a long time they will be able feel compassion for us and maybe there is hope for us yet.    There are two results of their compassion.    The first is them intervening in our busy lives showing us what we are doing wrong and helping us to become better beings.    The second is they destroy us before we make our way too far into the galaxy destroying and polluting planet by planet.    Either way, I donâ€™t think Iâ€™ll live to see the day either happens.    Who ever reads this; Iâ€™d really like to know your feedback.    What do you think about the existence of aliens?    If youâ€™re a believer, then why do you think theyâ€™re here?    Iâ€™m signing off for the night but Iâ€™ll leave you with a little quote I heard from a friend a while ago.             â€œWhen your soul is in doubt, look to the stars for answersâ€ - anonymous '),(17908,4022853,'26,July,2004','2004-07-26 00:00:00','   So here I am again; alone in my room, in the dark, staring at my ceiling and listening to music.    Iâ€™m not tired because I took a nap earlier.    The dream I had though, it was terrible.    My best friend was being buried alive; I was running from people who wanted to kill me, monsters killing other monsters.    I wonâ€™t read into it too much.    I was sick today so Iâ€™m not too worried.    Staring at my ceiling I start to thinking about a lot of things.    Really at the moment Iâ€™m broken hearted.    Iâ€™ll get over it, its just this would be the 3 rd  time Iâ€™m doing this because of the same girl.    We havenâ€™t even called it off, we just havenâ€™t talked.    Itâ€™s complicated, as things always are.    Sometimes I wish I was an animal and not a person.    Animals just live their simple lives surviving while humans do the same but by no means is it simple.    We complicate things by creating all these amusement.    We go to work, make money, and after that we come home looking for entertainment.    These industries and corporations just prey on us like we weâ€™re savage pack of dogs wanting fresh meat.    This analogy isnâ€™t too far from the truth.    We go to war and kill each other almost everyday.    Some of humanity or almost all of it just sickening. We kill each other, animals, take away life everyday, pollute this earth to the point where it become uninhabitable.    We multiply at rate unheard of, like a disease almost.    My sour view towards humanity grows everyday.    I donâ€™t like what I am.    Then there are things about being human that makes up for all of that.    Love. Unfortunately my love chose to ignore me.             Isnâ€™t life just grand? '),(17909,4022853,'25,July,2004','2004-07-25 00:00:00','   Why is it Iâ€™m more active at night?    I enjoy the night so much more than the day.    I can stand near my window at night and just listen to the sounds that come from the darkened field.    I stare at the moon and the stars hidden behind my big tree.    Itâ€™s at night when my thinking becomes clearer but itâ€™s also when everyone goes to sleep so I end up being alone.    Itâ€™s okay though because I like to be left alone.    Sometimes I crave companionship but other nights I just like it when Iâ€™m by my self.    I donâ€™t know, maybe Iâ€™m just weird. '),(17910,4022853,'24,July,2004','2004-07-24 00:00:00','   Where is the best place to start?  How about the beginning.  I am 17 years old, and ultimately this is to get things off my chest.  I fully realize that people will read this but donâ€™t be so hard to judge me based on what I write.  I write when Iâ€™m upset and that causes me to write very sour things towards humanity.  Sometimes I question existence entirely.    Why do people do the things they do or say things they donâ€™t mean.    How can people go to sleep so comfortably knowing that theyâ€™ve hurt someone.    When thereâ€™s anger between two people the outcome is usually very bad.    Perhaps ill post later when my minds more clear.    Iâ€™m afraid all the thoughts in my head are just flying everywhere right now. '),(17911,4022853,'18,August,2004','2004-08-18 00:00:00','   Went to the beach a couple days ago with some friends.  She was there.  I remember one moment were she was lying down on the blanket in the sand, and her hair just light scattered across her face.  And then the sun hit her red hair so perfectly, she was the most beautiful thing in the world.  I wish i had a camera.  But i can see it so vividly in my mind.  I wish I could tell her this and it hurts sometimes when I know I cant.  She\'s on my mind a lot, especially at night when I cant sleep.  I\'ve been hanging out with her a lot, and I talk to her every chance I get.  She still has no idea :(  I still can\'t tell her until im sure.  I just wish that day would be soon.  Dearest friend, how do i find words for what my heart desires....  Im sorry if you people dont wanna read this, thats okay, its just what im dealing with these days.  Perhaps my blogs will be better in the future.... '),(17912,4022853,'12,August,2004','2004-08-12 00:00:00','   Led Zeppelin  Trampled Underfoot   Greasy slicked down body, groovy leather trim  I like the way you hold the road, mama, it ain\'t no sin   Talkin\' \'bout love [X3]  Trouble-free transmission, helps your oil\'s flow  Mama, let me pump your gas, mama, let me do it all   Talkin\' \'bout love [X3]  Dig that heavy metal underneath your hood  Baby, I could work all night, believe I\'ve got the perfect tools   Talkin\' \'bout love [X3]  A model built for comfort, really built with style  Specialist tradition, mama, let me feast my eyes   Talkin\' \'bout love [X3]  Factory air-conditioned, heat begins to rise  Guaranteed to run for hours, mama it\'s a perfect size   Talkin\' \'bout love [X3]  Groovin\' on the freeway, gauge is on the red  Gun down on my gasoline, I believe I\'m gonna crack a head.   Talkin\' \'bout love [X3] I can\'t stop talkin\' about...   Come to me for service every hundred miles  Baby, let me check your points, fix your overdrive   Talkin\' \'bout love [X3]  Fully automatic, comes in any size  Makes me wonder what I did, before we synchronized   Talkin\' \'bout love [X3]  Feather-light suspension, coils just couldn\'t hold  I\'m so glad I took a look inside your showroom doors   Talkin\' \'bout love [X3]  Oh, I can\'t stop talkin\' about love.      Just had to share this one big giant inuendo with you.  Its awesome!  Good song too! '),(17913,4022853,'11,August,2004','2004-08-11 00:00:00','   I cant sleep.  Im trying to but i cant do it.  Everytime i close my eyes i see her.  The one my effection has been towards.  I talk to her online everyday now for hours on end.  I think im beginning to really care for this person but how to tell her?  I plan to see her friday.  If there was an opportunity to tell her it would be that day.  I still dont know how to tell her.  There\'s also, the rejection factor.  I\'ve never told someone i\'ve liked them unless i was absaloutly sure they liked me back.  This is different becuase im not sure.  I feel i should wait to tell her how i feel but if i wait too long i might miss the opportunity.  Perhaps i should wait to see if feelings grow stronger between us.  Maybe she doesnt see the connection now but she might see it later.  I dont know.  Everytime i close my eyes i see her smiling face.  I start making up scenarios in my head about friday and i know none of them will happen.  Its still nice to think about though.  I really do wish she likes me back.  Just talking to her makes me feel flushed.  Im afraid though.  If she does like me back what then?  We become boyfriend and girlfriend but still...what then?  I see her a lot...we play guitar...I dont know.  Things may get boring between us.  I dont know.  Maybe things will fall into place.  What if she doesnt like me...and i poor my heart and soul out to her.  What then?  Perhaps this is a friendship that should remain just a friendship.  Perhaps i shouldnt ruin what we\'ve got...but if there is the chance of success...should i take it?  I dont know.  I suppose i should distract my mind if i can, after all this is just another night where I cant sleep.  ...but what if?.... '),(17914,4022853,'09,August,2004','2004-08-09 00:00:00','   I will not try anymore.  I admit defeat...i kept fighting when all i really needed to do is give up.  In the end i win because then maybe my insides will fix themselves.  Just when things get better they get a whole lot worse.  I must free my mind from this burden.  I dont want it anymore.  Its over now.  I must keep reminding myself thats its over now.  I dont have to fight it anymore.  I had to lose to win.  I just hope twisting in my stomach subsides really soon.  I do not know how long I can endure this pain for.  Good riddens!  So much pain for such little happiness.  Not worth it.....women.  My insides do feel better now. '),(17915,4022853,'06,August,2004','2004-08-06 00:00:00','   I never slept tonight.  Too many things running through my head.  It was around 5 in the morning...I was sitting in bed not being able to sleep and I was thinking of \"her\".  She was everything to me.  I decided to go for a walk at that point.  So its just getting light...and im outside wandering the streets at around 5:30 in the morning.  I then get the bright idea that I may be able to see that wonderful smile of hers once more so I go to her bus stop.  Unfortunately I am unaware of the time having lost my watch at Newton Wave Pool.  So i leave and im back at my house around 6:30.  I am soaked at this point and my friend is passed out on the couch.  Im thinking to myself that if i just stayed a little bit longer i might have been able to see her once more.  Depressed i grab a nice warm bowl of soup and just sit in my back yard and get even more soaked in the cold rain while eating my warm soup.  I thought of her almost the entire time.  Writing this i am still soaked and though my mind may be weary, i still wonder if things could have been different...I always keep a faint bit of hope with me. '),(17916,2101657,'24,October,2003','2003-10-24 00:00:00','    Been a few days since my last post. It\'s been a rough week, with being sick. On top of that, I got a flu shot at work Wednesday, and naturally, it made me sick.  I got home from work, booted up, sent a quick note to my Spanish teacher that I wouldnt be in for class, and then went to bed and slept for 14hrs.  Had a fever slightly over 100, and today I still have the fever, but went into work because I had things that had to get done.  I\'m home for a half-day for not feeling well.  I\'m on-call, hopefully the CME   we\'re going to get hit with from the sun in a few hrs doesn\'t effect my cell phone too much as I\'m on-call.  The solar flare is going to cause distruption they figure for 12-18hrs.   Sounds like we\'re all headed out of town to work on our trailer that we rent out.  Oh boy oh joy.  Fixing a ceiling fan thats brand new and shouldn\'t have issues but apprently does, plumbing, racking up rocks that our EX tenant put all over the yard, that she never asked about.   So now I get to rack up and get rid of how knows how many pounds or river rock due that @#$@# we had living there.  Being a landlord sucks.   '),(17917,2101657,'19,October,2003','2003-10-19 00:00:00','    Well....the weekend has past more quickly then I wanted it to (whats new).  I\'ve spent a couple of hours, studying for the last test in my Spanish 101 class.  I\'m back in college through the tuition re-emburstment program at work, and have decided to finish my degree. What better way to do it then to start off with a language, right?  The classes are 2 nights a week, 2hrs long, and they go for 2months at a time.  It\'s called a TimeSavers program.  Not sure who it\'s saving time for, doesnt seem to be me.  Between work, school, and Scouts, doesn\'t seem like alot of time left.     Today\'s been a sort of recovery day for me, woke up in the middle of the night with a migraine.  I get one ever 2-3 weeks it seems.  Often enough that I take medication to help prevent them and medication for once I do get one.  No clue why I get them, I\'ve suffered from them since I was 5yrs old.    Went out and put in a couple holes in our local Frisbee golf course.  I had a argument with the wife, and I needed to get out of the house and away from her.  The course isn\'t the place to go when mad though.  There are a few er.... really steep areas, and it doesn\'t work to go trumping up them in a huff.. only makes you more tired.  Decided to do a couple of the holes on the back 9, and on the way back, I climbed down into the revene, and took it most of the way back up to the front of the course, keeping my eyes peeled for any discs laying around.  That\'s how I found the Frisbee golf disk I have now (www.discraft.com). It\'s different from playing with a Whamo(C) brand Frisbee. Back then I played with the big white 151gram disk.  Lots of hang time with those old ones. The discs you use now are the size of a large saucer/small dinner plate, and man, do they throw totally different.   I used to play Disc Golf alot when I was a teen...it\'s awesome to be playing it again.  My oldest is getting into it and wants his own Golf Disc now too. Already told him thats something I\'d happily take for Christmas  .   Still working on getting over this cold, had it going on 3-4 weeks now and I sure to look forward to not coughing all the time, and always feeling very run-down.  Ever dread getting up in the morning because your exhausted? Thats how I\'ve felt the last 3 weeks.  I REALLY look forward to this cold being gone.  Almost as much as being done with Spanish 101 tomorrow night  .  Star Trek:Enterpries was decent Saturday.  Hoshi, the communications officer is telepathically contacted by a funky looking huminoid with feelers or something on his head.  He can easily read Hoshi\'s mind, and at the end, tries to trick her into staying. He just makes a few blaring faults that she catches, and returns to her ship. He\'s able to hold an ojbect and get mental impressions left on it by whomever build it/last touched it or something, and is able to help send Capt Archer in the right direction, to find a colony of Xindi who are building part of a weapon to destroy the earth.  Looks to be quite the story arc, which is a nice change for Trek.  I loved how DS9 did it all the time, and it\'s nice to see ST:Enterprise doing the same thing :-)   '),(17918,2101657,'14,October,2003','2003-10-14 00:00:00','    Ok, time for a Star Trek post!  Ok, the latest and greastest in the line of Star Trek tv Shows has been \"Enterprise\".  As of just the last few weeks, it has FINALLY been changed to \"Star Trek:Enterprise\".  Like all series in the ST francise, it seems that around Season 3 the show really hits it stride and takes off.  Star Trek: The Next Generation (here after called ST:TNG) really took off with \"The Best of Both Worlds, Parts 1 & 2.   DS9  (Deep Space Nine)...well...thats sorta my favorite right now, so it hit it\'s stride early on. Voyager got better as soon as Janeway let that bun down. Any one would be crabby if your hairs wound so tight it\'s pulling it out by the roots!      Enterprise had an OK first two seaons. Several episodes really stood out..and some really stunk too.  People complain about how the Enterprise NX-01 can look so modern compared to what it looked like in the original series with Kirk in command.  My answer on it? They have to make it look primitive and crappy because the first show it\'s based on was from 25odd years ago, that it has to have crappy special effects now?  Hell  no..  why should they?  So it just doesn\'t match up...it\'s not done in an age of fantastic effects.  But we have them now, so lets use \'em!  Now with Season 3 of Star Trek:Enterprise under way, we\'re finally getting to see alien species that are totally CGI (Computer Generated Image).  They look way awesome and it\'s about time they started doing this.  The crew is totally on their own in a very hostile area of space. They have some Federation Marines   on-board for combat situations. Capt. Archer and crew are trying to find the home world of the Xindi, a \"species\" that has it out to destroy Earth.  There are 5 different kinds of Xindi and they are all completely different from one another.    Capt. Archer is very much different this season. He\'s harder, more seasoned, grown a back-bone. He\'s so far looked like he\'s willing to do whatever it takes, and on occassion crossing the line to make it happen.    Well....thats it for this entry, will follow up with more soon.   '),(17919,2101657,'14,October,2003','2003-10-14 00:00:00','    Hello! Welcome to my Web Blog!  I\'m Kromwell, a pleasure to meet you!  Please pardon the mess as I move in and get things arranged the way I want them.  I\'ll start off a little about myself.  I\'m a 35yr old Computer Tech, living in Dubuque Iowa.  I\'m married, to Zelda (in-case you want to read her blogg, she\'s here on Blogger), and have 2 awesome kids.  We\'re all very active with Cub Scouts.  I have a few hobbies, so I\'ll mention them... I\'m very into computers, hence my occupation. I seem to have a strange and twisted knack with them.  I\'ve built more pc\'s then I care to remember since I started working with PC\'s.  I got started with computers in 1980 when I got my computer, a Timex Sinclair 1000.  Through junior high and high school, I took everthing I could about computers. In 1987 I went to school for programming, and that was pretty much all she wrote. I\'ve been working with IBM Compatabiles since then, and haven\'t looked back.  I was writing about hobbies though, so let me get back to that.  My other primary hobby, would be Star Trek.  What can I say?  I love it.  I collect the Playmate brand Star Ship\'s, and also have the 2 released by Art Asylum.  I have various items around the room that would make it obvious that I\'m too much into star trek... like the Star Trek clock, the 2 pieces of art, the mouse pad on the wall, or the Klingon Action Figure sitting on the korner of the desk?   Oh, not to mention the Klingon ensignia thats attached to the front of my monitor, or my DS9 Uniform top with phaser and tricorder?  Hopeless aint I? LOL   Lets talk Cub Scouts now.  I\'ve been an active leader now, going on my 4th year. I\'ve a Pack Leader now, and have had my current den of boys since they started as Wolf scouts.   We have a pretty fun time too! Personally, I think the best pack meeting is when we do Pine Wood Derby!  That is such a fun event, and the boys love it! Wish we did it more often.    Well, thats my thoughts for the day.   '),(17920,2101657,'28,November,2003','2003-11-28 00:00:00','    I hope that everyone who views this meager little blog had a very Happy Thanksgiving, and a safe one.  We made a 2hr+ drive each way, and had a pretty fun day.  We did the olbigatory turkey and trimmings, plus lots of games, listening to the wife play her guitar (note, she\'s getting really good!).  We got to see my in-laws new dog Scooter, a stray female black lab.  Really friendly, and a ball of energy just waiting to rupture.    Got a new toy tonight, and all I can say is, let it snow!  My dad had a snowblower that he gave us. They made a spur of the moment visit out and brought it out with them.  It\'s an old John Deere.  If we have the snowfall that they claim we\'ll get this winter, it\'ll really be worth having.  The kids already want me to blow the whole yard into a big pile for them to build a huge snowfort out of lol  Friday after thanksgiving. Just mentioning that makes me think of the feeding frenze that goes on at the malls on this day.  I read that the average consumer will spend $672 on presents for Christmas.  I try to make it a point to do the opposite and sit at home.  Malls?  Treat \'em like the plague on this day. Traffic will be unreal, other shoppers will be mean, rude, nasty, and just out sub-human in general.  At least thats how it felt when I worked in retail.    I worked for about 3Â½yrs for a Software Etc.  The pay sucked, but the discount on software and books was well worth it.  At Christmas time, people tended, instead of becoming inbued with the holiday spirit, wanted to tear you a new one because you don\'t have the green Nintendo 64 controller that they wanted and its ALL YOUR FAULT.  Nothing like being in the spirit of the season and being nice to your fellow humans.  Anyway...     Picked up the Extended edition of The Two Towers on tuesday.  It has 43 minutes of extended/restored footage included directly into the movie. This makes it 243 minutes long.  Thank goodness for the pause feature. My bladders good but not that good.  They really are doing such an incredible job on these movies.  The restored  footage really adds to the movie, and makes you wonder why they were cut except to reduce the over-all length of the flick.  I watched the special feature on Gollum, and all I can say, is  wow , thats just the neatest shit on the face of the earth!  The guy who did Gollum is just incredible. He does the voice perfectly, he does almost all of the actual body movements for the CGI character to replicate, and just all the phsyical stuff he had to go through. His screentest is something else. They said he was hired almost from the get-go based on the screentest. He really got into character and had it nailed almost perfect right away.  This is a must to add to the collection if you like Tolkein.  All I know is I can\'t wait for the Return of the King to hit the theaters, just to see what happens next (and I\'ve read the books) :)   '),(17921,2101657,'19,November,2003','2003-11-19 00:00:00','    Well, we survived our campout.  We spent the night at Camp Little Cloud, the console camp for our local Girl Scouts.  I tell you what, the two cabins we stayed in, were nicer then what I live in.  The cabin we stayed in, had a wood burning stove upstairs, a spiral staircase going downstairs, 2 bedrooms, and a bathroom.  In the basement was the kitchen, another bathroom with a shower, a piano,  and another door leading outside.     The only draw back to the entire place?  Not a single curtain in the whole place.  Was very odd., but from what I could see, there was only a family camping in a larger building. We had 20 people total show up, so we had 10 to a building.  We had a great time!  We did a hike, taking the Scouts down to the lake, and walking around it, and then hiking up a steep hill through the leaves, and trees. We stopped several times on the way up as it was hot work, in the 45+ degree temps.  At the top of the hill we took another break, but this time got to sit on the edge of a platform that in the summer time would be occupied by a teepee.  There were five teepee platforms, and two buildings. The cabin itself, had a huge glass front to it, that formed a wedge facing south. Would of been awesome to see a sunrise from there.  At this point, we had gravel rode, and took it a half-mile back to our area.  It was very surine in the area were we walked through a bunch of pine trees. The sounds changed when we entered. Almost like when you realize you\'ve just entered the \"Haunted Forrest\".     We had left two or three leaders back at camp as we had started a fire and they were monitoring it.  We showed the Scouts how building a \"log cabin fire\", the wood stacked like a log cabin would be, burns fast and gives you a good bed of coals to work with.  We had the boys go up to the cabins kitchen four at a time, to make foil packs.  Basically they got a piece of aluminum foil 12x12.  They then put on some of following based on their own choice: hamburger, onions, carrots, potatoes,     salt, pepper, seasoning salt.  You just pile what you want into the middle of your aluminum foil square, and fold it up, so it\'s sealed.  Write your name on the outside with a nice magic marker, then take it to the fire, and have a leader sit it right in the coals.  After 10minutes or so, the leader will flip it over, and cook the other side.    MMMMMMMMMM MMMMMMMMMMMMMMMMMMMM, did it taste good.  I had some hamburger, onions, potatos, all the seasonings, and a little catchup.   Boy did it taste good!    My assisant leader had brought the makings for a popcorn popper, and had some of the Scouts help him put it together.  Basically, you take two bread pans, and you get a piano type hinge. You drill holes through the side lip on the bread pan, and pop rivet the hinge on.  Get a strip of aluminum, and you can easily hand bend it, shape it to one end of the breadpan, with about 4-5 inches extending past.  You drill and attach it the same way to the pan.  Then, take a fairly straight branch, or something roughly a 2x2, thats 4ft long or so.  You use some screws into the extended part of the aluminum strip, and attach it to the handle.  Then, you take and punch a hole from the inside of the pan, and run a screw into the handle, attaching it from there too.     You now have a handy dandy popcorn popper, that you can give a quick shake, to open it up and dig in!  We made some kettle corn. It\'s made the same way as regular popcorn but you add some sugar to the oil.    A good clean up tip?  Rub some dish soap on the outside of the pan, when you go to clean it up, the soot from the fire will just wipe off easily, instead of discoloring the pans.   '),(17922,2101657,'14,November,2003','2003-11-14 00:00:00','    Well, this is the weekend... I\'ll be taking our CubScout Pack out camping tomorrow. We\'ll spend roughly 24hrs together in the woods. We\'ll have buildings to stay in, as we\'re not staying at the districs normal camp, Camp Burton.  We should have about 20-25 people going.  We\'ll take the kids out on a couple of nature hikes, during the day and at night.  I\'m going to take my telescope along, so we can do alittle astronomy work.  During the night hike, I\'ll show them how to find the north star, how to locate the milky way, and various constellations.  We usually have a great time.   '),(17923,2101657,'13,November,2003','2003-11-13 00:00:00','    Well, not much to really say today.  We\'ve been busy it seems since our trip to BestBuy saturday. We picked up \"Mech Assault\" for the XBox.  This game, is done by Microsoft, with the auspices of FASA, the group that created BattleMech in the 80\'s, later to become MechWarrior.  All I can say about this game?  AWESOME!  My sons and I have wanted it for awhile, and we couldn\'t pass up the game, as it was marked down to $19.95, down from $49.99.  If you\'ve ever had the urge to run around in giant robots, armed to the teeth, this game is for you.  We\'ve probably put in 2-4hrs a day since we got it.  I\'ve made it quite far along in the campain mode, playing progressivly harder missions.  The soundtrack is awesome, the graphics are just as excellent.  I can\'t say enough good things about this game.  Both my boys, ages 10 and 7, are both enamered with the game, something I\'ve been, since I started playing the BattleMech rpg game.  I\'ve played all 4 Mech games on the PC, and some of the expansion packs.  If you want action, adventure, and just a good time blowing stuff up, this game series is for you.   Being that the Xbox, has a 733mhz Pentium III coppermine CPU, 128megs of memory, a special video chip from Nvidia, and high fidelity audio, it\'s hard to wonder why the games on it compare to the PC, as it basically is a PC, running a special version of Windows 2000.  With Mech Assault, Microsoft defintely came out with a winner and it makes me wish we had the XBox-Live service, so we could play other MechWarriors from around the world to see how we measure up.   '),(17924,2101657,'08,November,2003','2003-11-08 00:00:00','    Well, we\'ve hit the big time. As of 9am yesterday, we now have a Best Buy.  It\'s been under construction for some time now, and finally opened it\'s doors.  I decided to swing by, as they had a special on CD-R\'s, a 50pk for $4.99.  Well, go figure, by the time I got there shortly after work, they were sold out. I found out at the cash register that they had a substitute, but go figure, that one was sold out too.  It\'s an impressive store, and it was impressivly packed.  There were so many people there, that I decided to just get the movie I wanted then, as they had The Matrix Reloaded for $15.  I was going to walk down the computer goodies isles, but they were so jam packed, that I decided to not try and do my impersonation of a salmon going upstream.  It was nice having the store open, as McDonalds has been doing their Monopoly game, and they have had stickers for $1, $5, and $10 for use at Best Buy.  I had my little slip in hand with 7 of them, so my Matrix Reloaded cost $9.  Sweet deal.  I had the evening to my self since the family was out of town, so I sat down and watched it. The special effects are awesome. Just like the 1st movie, but 200% better. The only draw back was in the Agent\'s Smith fight scene, it was very obvious that it was CGI, and poor CGI at that.  Still impressive, but it didn\'t look like the actors  what-so-ever.  Well, off to start my Saturday, hope everyone has a good weekend, will post again later.   '),(17925,2101657,'03,November,2003','2003-11-03 00:00:00','    We got to go see the US Marine Corp Band yesterday. Or, as they are called, The President\'s Own.  I heard more volume and sound out of this band, then I have from college Marching bands with 200 people.  I must say, if you ever have an opportunity to see this band play, it\'s most worth while.  Every member of the band is a member of the Marine Corp\'s.  What I found a little missleading on that part, if I understood it correctly, they are members of the Marine Corp, but never have to go through Basic Training, and are generally exempt from that type of training. They are assigned TO the band and basically nothing else.  They get an automatic rank of Ssgt  , an E-6.  It takes most people on active duty 5-7yrs to get that rank.    The music itself, is incredible. They play alot of John Philip Sousa, and they played some songs from Broadway as well.  The conductors, whom I don\'t recall their names, are a Col. and  a Major.  There was also a gentleman who came out and introduced each song, and gave a little background on it. He sang the National Anthem, and also in the second half of the performance, sang three more songs. He can sing most excellently.  I was very impressed.    The concert was held at the high school.  We ended up in seats on the 2nd floor, toward the middle edge of the basketball court.  There was local law enforcement to help out, as well as some active duty troopers pulling security. The local AmVets were helping out as well.  There were probably 1500-2000 people in attendance.  I know, if they are back next year, we\'ll make sure to go see them again, they put on a fantastic show.   '),(17926,2101657,'02,November,2003','2003-11-02 00:00:00','    Well, we made it home about 11:35 from the Indian Pow wow held at the U OF W Platteville.  We didn\'t get too lost on the way up. We got on campus, found a LARGE map with one of those \"U R Here\" dots on it, figured out roughly where we wanted to go, and still got a hair lost on the way.  We back-tracked a ways and found the turn that we had missed.  We thought we knew the building to go to, but it turned out to be the Student/Technology Center. Pamela ventured a door to an Information  [I think]  area. She found a small pile of maps of the campus; we looked at it, went out front and determined we just needed to go one building over.    We went in, and started looking at the vendors booths as we had 20-25 minutes to spare. There was some really excellent examples of hand-made Indian jewelry.  Some of the outfits as usuall are stunning. The time and effort that must go into making their costumes has to be astounding.  We made our way 2/3\'rds around the vendors, when we run into a mother in Pamela\'s Wolf-Den for scouts.  We had mentioned the Pow wow to her yesterday at a field trip we had out to the E.B Lyons Nature Center. She\'s from Jamaica, and was very interested.   She had been up there since 2pm\'ish, so by this time she had been there 4hrs.  During the evening, when the audiance could join in the dancing, she hopped up 4 or 5 times to join in.    To be honest, I was tempted to get up and join in, but felt a little apprehenceive wearing a blue Hawaiin shirt, and how it\'s stick out.  There was one point that Pamela had gone off to the ladies room, that if she had been back, I was going to get up during the group dance part, and try my hand at a Friendship dance.  I tried to talk my oldest son into it but he always said no. I asked Pamela but she said the next Pow wow.  I\'m making note of it here so she doesn\'t forget :) Time to call it a night, will give a better detail of it later today its late and time to pay Winkin, Blinkin and Nod a visit.   '),(17927,2101657,'01,November,2003','2003-11-01 00:00:00','    I want to say  Thank You  to my  wife , for all the work she did on my Blog last night. She set it up so I can have comments on my blog, and a picture of me   and a mood indicator. Watching her edit the template and posting html code here, and typing some in by hand there, I can definitely say she knows her way around a web page. Shes only had a webpage since \'96, when Geocities was the only real place offering a free web page. She was once a featured website, and there was only 1 of those at a time, once a month.  She does an  awesome  job with the sites, and is  amazing , to me, who works with  pc\'s , but not  html .  We got our popcorn today, for our CubScouts to deliver.  After unloading all of it, it was nice for a change that our sons only had a little amount to deliver.  Everything went pretty smoothly for the most part and kids maybe knocking on YOUR door to deliver something you ordered.  Everything we do as a CubScout Pack during the year is funded by what we earn from our popcorn sales.  Last spring we were able to purchase a new 2-lane all-aluminum 35ft long track.  Wow, is it awesome to run the derby on that! Our race night seemed to go by fairly quickly.  My assistant leader and I even set it up at another school for their CubScout Pack\'s derby. They haven\'t had one before, and we volunteered to help out, and it gave us an excuse to show off and use our new track.   We had 24 kids racing, using a double elimination, and we were done raceing in 45 minutes. Was amazing, our own pack races don\'t go that fast.   We\'re going to make a trip up to the University of Platteville Wisc tonight. They are having an Indian Pow wow.  This will be our second one this year, and our 3rd one total, together.  Should be quite interesting.  After going to the Tama Pow wow on August 7th, this one will be nicer, as it wont be boiling hot outside, and we\'ll understand it and appreciate it a little more as well.  The costumes are incredibly detailed, and the dancers are very good at what they do.  It\'s an awesome experience, one  I think everyone should experience for themselves.   '),(17928,2101657,'26,December,2003','2003-12-26 00:00:00','    Well, we safely made our Christmas trek to my parents house.  We almost didn\'t make it back in the same shape though.  We were on Hwy 20, just going past Victory Ford heading east.  We were coming down the hill, and a white Audio pulls out to head east as well.  Unfortunetly this person decided they DIDN\'T want to pull into the in-side lane  (they hopefully still teach in drivers ED that when you make a turn you pull into the lane closest to you) and pull into our lane.  I was going 65mph, and slowed down some, but this guy, just goes in front of us. I started laying on the horn and had to jerk the car onto the shoulder to avoid rear-ending them.  THEN...I\'m trying to get back onto the road and the guy is STILL there and practically on my bumper. I had to speed up to get back onto the road.  All the while, I was still honking the horn.  PLUS, he had his high-beams on.    After I got back onto the pavement, and got back upto 65mph, the Audi was hanging back about a quarter mile.  I waited for him to speed up, as I figured this @#)$(*@ was the kind of person who loves to speed, and then I got his liscense plate # and called 911.  They pulled onto the HWY from a supper club area, and I figured they had to be drinking.  I was on the phone with 911 as I tried to help  keep an eye on the Audi, so they could get this person.  Unfortunetly they didnt.  They asked that if they did find him, if I would pull up behind the Sherrif and wait for him to approach me to discuss what happened.  My main concern was that after I got the ok from the Sherrif to get out, that I\'d run up and beat this person, as they almost caused a major auto accident and endangered the lives of my wife and my children.   I hope that the bad karma this person has building-up catches up to him..   '),(17929,2101657,'24,December,2003','2003-12-24 00:00:00','    Here\'s hoping that everyone has a Merry Christmas, and a Happy New Year!  I\'ll be watching my new set of Firefly on DVD, that my family so nicely got me for Christmas! Nothing like spending the holidays with Mal, Jayne, Zoe, Book, and the gang :-)  Oh, cant forget Serinity :-)   '),(17930,2101657,'22,December,2003','2003-12-22 00:00:00','    Just not having the grandest time of late.  At least I\'m 95-98% over this flu. All I seem to have left is some coughing, and getting tired easier then usual.    But, the work day is done, on to home.    Been playing Unreal Tournament 2003 since last Wednesday.  I had issues trying to get it to run on my system, a 1.5ghz+, 256megs of memory, and a GeForce2 GTS video card with 32megs of ddr ram. The darn thing should of run.  Well, I\'ve put in a Radeon 9600 Pro.  It handles UT2k3 bigtime now. The video quality, the frame rates, the speed.  Wow... I don\'t miss the GeForce2 card at all.  At least I got this right.  Well.....Christmas is just around the corner.  Three days to go now and counting.  It\'s a Happy, Cheerful time of year.  At least I think it is......   '),(17931,2101657,'14,December,2003','2003-12-14 00:00:00','    Nothing like having caught the flu, when you had a flu shot 2 months ago.  I was in Cedar Falls removing some things from the shed at our trailer, before we sold it. It was cold, and raining, and I was standing in it, moving ladders, and a Weider weight bench into the back of a friends pickup truck.  Spent probably 30-45 minutes doing this.  Well....I ended up staying home wednesday from work not feeling well.  Thrusday wasn\'t bad. Friday was not fun, I felt run-down, coughing, sneezing, aching, I sound like a Ny-Quil ad yet?  So the flu really kicked in on me Friday night.  I spent most of yesterday, w/o a shred of energy, but I still managed to pull up the box with our Christmas Tree, and 3 boxs of ornamints from our basement. That about ran my batteries totally down.  About 8:30, my head was throbbing and pounding from a migraine the flu was giving me.  I think that a can of soup that I had earlier in the evening was a bit off, and I got a little food poisoning from it, as it wasn\'t too long before I was in the bathroom praying to the toilet diety. It\'s amazing how sometimes just throwing up can make you feel better.    Today has been alot better.  I have some more energy, though my body still aches from the flu.  Well, this along is running me down, so I\'m going to stop for now, will post again in a day or two.   '),(17932,2101657,'09,December,2003','2003-12-09 00:00:00','    Looks like we get to see winter show up finally. We\'re suppose to expect between 4 to 7 inches of snow by 6am. Should be interesting.  Wonder how the new snow blower will work on it.  Todays been a Monday in hiding. Raining all day, cloudy, the day ends in Y again..  Nothing else to report.   '),(17933,2101657,'02,December,2003','2003-12-02 00:00:00','    Not much to say today. I\'m trying to not be jealous I guess (apparently Jealousy is good in life, I personally think it sucks. Should\'t have to have that emotion in your life period)  My left leg has just been hurting horribly the last few days.  For the last three weeks its ached and hurt in general. Its been between about 5inches down from my hip, to a few inches above my knee.  It doesnt feel like the muscle, feels like it hurts in the bone.  I haven\'t hit it against anything, falling, lifted anything too heavy.. it just hurts.  And when it really starts to hurt it gets there.    It made me think of the chart they gave my wife Z*lda when she had some outpatient surgery a few months ago. They rate pain based on a zero to 10 scale, 10 being the worst.  Last night I would have rated the pain around a 6-7.  It was enough I was almost crying from it.  I know I know... I must not be a real man..but it really hurt!  I\'ve tried several different pain killers, and even gave Naproxin a shot (known over the counter as Aleve).  Well...it didn\'t Aleve anything. I took some about 11:45, and it\'s not stopped hurting one bit in the last 5hrs.  Living with pain just sucks in general.   '),(17934,2101657,'01,December,2003','2003-12-01 00:00:00','    What can I say, it\'s another day that ends in Y.  Y am I here, Y do I think, Y do I try, Y do I care....nothing but Y\'s.  Course, it\'s a Monday so that says 90% of it.  I figure it\'s just part of the roller coaster of life that we\'re on.     Weather is colder today then it was yesterday. We hit 52/53 and went out for a hike on one of the trails at the Mines of Spain State Park here in Dubuque. Parts of the trail, that were on the side of the bluff, were fairly muddy and slick. When you consider at some points how steep the trail is, you don\'t really want mud under your feet.  Other then that, it\'s a really fun hike.   When we got to the top we were at the Julian Dubuque Monument.  My wife Z*lda was telling me, that the monument originally built by the Native Americans in this area, was made out of wood, with holes  so you could see Julian Dubuque\'s body. They have since redone the monument, so that it looks like a giant rook from Chess, made out of huge blocks of limestone (I believe anyway).  We spotted an eagles nest on the Illinois side of the Mississippi. It wasn\'t hard to find, those things aren\'t exactly little.  Watched some more of the 1st CD from \"The Two Towers\" Extended Edition DVD set.  It\'s just such an incredible set of movies.  The 4 disk set is the only way to go. I\'m still enjoying the extra scenes they put back in, from the book, that just give the movie that extra edge/detail it needs. Granted, I still cant sit through the whole thing w/o a trip to the bathroom. 3hrs 43minutes is just too long for 1 sitting and a large pop.   It\'s now officially December, and that means only 23 more shopping days till Christmas and Santa are here! (oh, wait, I\'ve seen santa probably 3 different places already, scratch his part LOL)   '),(17935,2101657,'30,January,2004','2004-01-30 00:00:00','    Can it get any colder?  Wait....forget it, it can.  It seems like it\'s getting colder every day.   Today when I left to go in to work at 7:20, it was -11f.  Thats entirely too cold in my opinion.  With the windchill we got down to -25 to 35 below.  Our old crappy furnace has been working over-time just trying to keep it 62 in the house, when the thermostat\'s set to 70.    When I got up I went outside and started the car up, so it would be warmed up by the time I had to leave.  I had a wal-mart bag of snacks to take to work with me, along with a 6pack of 24oz Diet Pepsi\'s.  By the time I got into the car, the pop had gone from room temperature, to fridged.  Was good and cold to drink that\'s for sure.    Short entry here, it\'s FRIDAY  (WOOHOOOOO) , and I\'m just thinking that it\'s finally nap time.  Hope everyone has a safe, and WARM weekend!  We\'re going to hit a balmy 15 above tomorrow.   '),(17936,2101657,'20,January,2004','2004-01-20 00:00:00','    Ok....this is getting old.  I spent alot of Saturday with the feeling in the right side of my head, that I had  a migraine  just waiting to jump out of the closet onto my back.  I went to bed early Saturday, around 9:30pm. Getting old sucks.   Sunday after getting a decent nights sleep, I still had that tell-tale pressure in my right temple.  Well, about 4pm, the oldest and I went on a run to the store as we needed cat litter, chocolate moo, and ice cream.  Stupid me didn\'t wear sunglasses, so the late afternoon sun was beaming me right in the eyes most of the way home.  That was enough to push  my migraine button  the rest of the way.  About 5, I took 1000mg of ibuprofen (sorry if it\'s misspelled, hate the word anyway lol), and one of my migraine meds.  About 5:30 I decided it was time to lay down. My wife Z*lda asked if there was anything she could do, and I asked her to call the ER. One of the guys at work gets migraines and he has his wife take him to the ER, where they give him a shot of demeral, and 30minutes later, he\'s out and not having to experience the pain.  Well, she called the Ask-oh-nurse, and was connected to a hospital in Des Moines.  The lady sounds like she went down her little chart, sounding like she was agreeing with Z*lda on everything, and I ended up staying home, and spent another 45minutes feeling like the vessels in my right temple were trying to escape from my body (felt like they almost made it too).    The nurse did say \"if he throws up bring him in right away\", which didn\'t make sense to me, as I almost ALWAYS throw up when I get a migraine.  Hell, it feels bad, but, it does seem to make it go away faster.  Why would they want me to come in only IF I puke?  Go Figure.    I\'d of rather got the demeral shot.  I would of slept through the night. As it was, I woke up about 9pm, thought I felt ok, then about 30 seconds later, Mr. Migraine said FU and looked me in the eye again.  So I rolled over and went back to sleep.  Then I woke up about 11:45pm. Lights on in the living room, and the house was very quiet.  So I got up, since Mr Migraine had finally given up and me and went to beat mercelessly on someone else.  Z*lda was on the couch snoozing, but woke up right away.  Oldest boy was in his bed reading some of the AWESOME book Harry Potter 5.  Littlest boy was playing with transformers (I think) on the floor.   Got him to brush his teeth finally and crawl into bed (12:30AM. for being 7 he sure is a night owl).  I finally went back to bed, and go figure, it took me until 4:15am to get back to sleep.  Then to top it off, having some of the most odd-ball dreams ever. (can anyone say astronaught?)  Well, the old alarm clock went off at 6:45, and my \"no freaking way\" meter pegged out.  I got up, and got stuff organized as I had to take our pc into work, were I was doing some digitization of video. Must say, that is some way fun stuff!  Put in the full/busy day at work, and then at 6:20, went to school for my 1:15 of learning to program in Visual Basic.net.  This is going to be a very kewl programming language to learn.  My copy of Visual Studio (academic copy) has arrived at the book store, now to figure out how to pay for it, as I just learned I\'m not taking enough credits for my student loans Tuesday to be accepted.  I\'ll get it figured out somehow.  Oh, to top it off on Monday? Mr. Migraine was back, trying to invade the left side of my head this time.  Sometime, I\'m going to be a REALLY big stick and give him some of his own back.  At least I fended him off, and he didn\'t get the better of me.    So, after the long day at work, the long evening, I get home from school, and what do I do? Throw in the Hulk.  Z*lda took the boys down and rented video games, and they had The Hulk previously viewed for $10.  I\'ll write up a review to post another day. I have mixed feelings about it, both good and bad (coming from an old comic book collector)  So as Stan Lee would say,   Excelsior!   '),(17937,2101657,'12,January,2004','2004-01-12 00:00:00','    Tonight as the start of class again for me. I\'m a student at  Clarke College , in their TimeSavers program. Basically it\'s an accelerated class for working people, to be able to take classes at night. Usually the classes are 2hrs, 2x a week, for 8 weeks.   The class I\'m in, Introduction to Programming, is still 2x a week but it\'s only 1:15 a night, and runs through the 1st week of May.  I think having the full four months will be helpful.  I\'ve done programming, for fun, and for a living.  Last time I programmed, was as a student learning C++.  After taking C++, I decided I actually liked COBOL better.   This class should be fun, as we\'ll be learning to program in Visual Basic.  This is something I\'ve never done. I\'m enthused about the class, and looking forward to a positive experience this semester! :-)   '),(17938,2101657,'11,January,2004','2004-01-11 00:00:00','    It was a very short lived series shown on FOX.  For some reason these twits pulled it after showing only 10 episodes.  Here\'s the write-up on the back of the boxed set out now on DVD:  \"Five hundred years in the future there\'s a whole new frontier, and the crew of the Firely-class spaceship  Serenity  is eager to stake a claim on the action. They\'ll take any job, legal or illegal, to keep fuel in the tanks and food on the table. But things get a bit more complicated after tthey take on a passenger wanted by the new totalitarian Alliance regime. Now they find themselves on the run, desperate to steer clear of Alliance ships and the flesh-eatting Reavers who live on the fringes of space.\"  Some of the ways they did things on this show REALLY made me wonder why Star Trek:Enterprise can\'t be more like this show?  You had a cast of 9 characters, who almost always interacted together for every episode they aired.  Mal is the Captain, Zoe is the XO, Wash (Zoe\'s hubby) is the pilot, Jayne the mentally-off musle of the group, Simon (doctor), and his sister River (hunted both by the Alliance), Kailee the ships mechanic, Book, a preacher (who has a much hinted at past that he was a high member of the Alliance), and Inara, who is a licensed Companion (space age high class call girl).  This show gelled from the get-go.  They just mesh perfectly and it\'s a joy to watch if you like SCI-FI.  They most high tech thing about the time period is they have space travel, advanced medical technology, but for weapson, hand-held anyway, they are stuck with projectile weapons (bulllets). It has a high \"old west\" feel to it.  The intro song is pretty catchy too, and is actually written by Joss Whedon the creater of \"Buffy the Vampire Slayer\".  On the 4th dvd they include his playing/singing of it to the studio to give them an idea of what he thought it should be like.  They left it intack, but spiced up the tempo.   The set is going for $39.99. If your a fan on the show, thats very reasonable, as you get the entire series on 1 set of DVDs.  Sounds like the sales of the set are doing very well, so if we get lucky, maybe the studio will bring the show back due to sales.  There is also rumor going around that Firefly is already making the transistion to the BIG screen already. If that\'s the case and is true, I can\'t wait to see Serenity flying across the screen.   '),(17939,2101657,'06,January,2004','2004-01-06 00:00:00','    Migraines. It\'s amazing how they can control your life.  On the drop of a dime you could have to stop, take medicine, crawl into bed, and hope to get to sleep before it really starts hitting you.  My doctor is nice in the fact that he gives me the free migraine medication samples that they are given, so I don\'t have to pay for a prescription. I looked up some of the prices, and all I can say, is wow doc! Thank you!   For me, migraines are dehabilitating.  I get a running fever that varies from feeling like I\'m freezing, or burning up, with drenching sweating.  Bright lights, and sounds make my head feel like it wants to explode.  I can easily tell anytime that I\'ve gotten food poisioning from fast food, because after eatting, 20 minutes later my stomach starts to rumming, then the head starts to throb, then the sweating starts.  The best thing I can hope for then, is to have to throw up as it seems to really help out.   Once, I had a migraine in the afternoon, and as I laid down, I was   ok. Then I had a cool breeze blow throw the window above our bed, and I had to throw up.  Never had that happen before.  I\'m really thinking about taking the docs offer of giving me a referral to a neurologist, to see if we can came up with a reason as to why I get migraines. It varies..sometimes I won\'t have one for 3-4 weeks, and then I\'ll have two reasonably close together.  I can do w/o the pain.  You learn to identify what signs your experiencing to tell you have a migraine coming on.  For me, usually, it starts as a throbbing in my right temple, sometimes left. It gradually gets really bad, where I feel that either a blood vessel is going to blow up, or it\'s trying to pull itself free from my body.  Either way, it\'s excrucitatingly painful.  If I haven\'t head migraine medication by then, it wont happen period.   I get the aura\'s on occassion too, and suffer from the \"after-period\" that can range from 1 day, to 3 or 4.  I would REALLY enjoy not getting these things.  I have no known trigger, except bright lights  . Thats the only one I can think of.  It\'s been suggested that long sessions of driving cause them. I make trips back to where I grew up, ever couple weeks, usually to work on pc\'s for friends, or just visit family. Z*lda thinks it\'s the long drive that does it.  It\'s 1:20 1way.  When I helped drive on our summer vacation in \'99, in my parents 27ft long travel trailer, I didn\'t have a problem at all, and I was driving for 3-4hrs straight.   At one point I had a gray tint to my glasses to help with headlights at night, as they can be really really killer.  I\'ve not had the that for a few years, but I did get antiglare coating on them, and that seems to help alot as well.     Well, enough of this for the night, time to go tend to my post-migraine, and hope it stays just a post and does become \"a\" again.   '),(17940,2101657,'04,January,2004','2004-01-04 00:00:00','    Well, we were going to go to the 11:45am showing of Return of the King (#1 son and myself).  Well, off to a semi-slow start to the morning, we went to the 2:45 showing instead.  We took christmas candy in a ziplock baggy, got our 2 matinee tickets, and then spent $9.50 on pop and popcorn.  The theatre was surprisingly empty.  It got to maybe half-full if that by the time the movie started.  We sat through the previews for Spider-Man 2, and Harry Potter 3 (both look like they are going to be really good).  #1 son had to get up and use the bathroom 1x during the movie. I was able to make it the whole way through w/o an interruption.  Just as a side note, it started to snow as we left. We had a quarter inch of snow when we left. When we got outside and back to the car, we had 2inches or so.  The roads were slick, but not too bad a driving experience.    The movie.  What can I say?  I did like it but did get bored several times.  There were a few things that again, were blatantly ADDED to the book.  Just like in The Two Towers when Frodo and Sam are taken by Faramir to Osgilliath, the oliphants in ROTK just didn\'t exist. It was dissapointing, but still an interesting scene.  There were parts of the movie that were spot-on, and others, that were totally ass-backwards. Some reason, I thought Aragorn would of looked alot more kingly at the end ceremony, but he looked like Stridder cleaned up. In the book he LOOKS like the King who has come back.  I guess I cant be completely happy with what they did, who could if they have read the books.   But as a movie, if you never ever read the real thing, it\'ll be a great series.  Even if you\'ve read the books, they are still an awesome series...their just not the same though..if you know what I mean.  If your a Steven King fan, you\'ll know what I mean lol  It\'ll be interesting to see how it goes if they can manage to get the rights to do The Hobbit.  Since they did LOTR, The Hobbit should be spectacular!   '),(17941,2101657,'03,January,2004','2004-01-03 00:00:00','    In less then 12hrs, I\'m going to take our oldest son, and we\'re going to go see Return of the King.  I\'m eagerly looking forward to seeing this.  Peter Jackson has done an awesome job with the Lord of the Rings series.  Granted, they\'ve taken GREAT leeway with the story, putting in stuff here, taking out stuff there, totally screwing with whats really in the book here and there, but who\'s complaining? :-)  I\'m really looking forward to seeing Shelob on the big screen. Plus the army of the dead, that\'ll be interesting to see as well.  3:20 is going to be alonnnnnnnnnggggg time sitting in that seat.  Going to have to make that soda last awhile.  Will post my opinons of the movie soon!   '),(17942,2101657,'03,January,2004','2004-01-03 00:00:00','    I\'m eagerly awaiting spring now.  I got 3 frisbee golf disks for Christmas. That brings my total up to 4 now. I now have 3 long range drivers and 1 medium range driver.  I took my cubscouts out this fall to play, and really enjoyed it. I used to play heavily when I was a teen, and am eager to play again.  We have a course here in Dubuque thats not very old, and Waterloo/Cedar Falls, where we moved from 5yrs ago has 3 courses, with 1 added in the last year.  I\'m really looking forward to nice weather and then hitting the \"links\" so to speak. Will be a great way to get some exercise in as well.  Also for Christmas I got X-Men2, and Pirates of the Carribiean(sp).  Both really good movies. I was surprised at how good POTC turned out to be. Johnny Depp was actually GOOD in it!  A most enjoyable movie. X2 was good, in that they seemed to really get NightCrawlers character down perfectly.  I couldn\'t of asked for a better Kurt then that.  Plus it\'s obvious that X-3 is going to deal with Phoneix and her dark hidden bad self waiting to emerge.    Happy New Year everyone!   '),(17943,2101657,'29,February,2004','2004-02-29 00:00:00','    Was working on the computer Wednesday night, when the PC locked up on me. I rebooted, and just after I put in my password and hit enter, the hard drive made a nice loud ka-chunk noise and the system froze again. I hit reset, and boom, bye bye registry. I spent a few hrs working on it to try to get it going again. no luck.  Thursday night after class, I started reloading it, and had it pretty much done by the time I crawled off to bed around 12:40.  Was using the PC after work Friday, around 5:30..KA-CHUNK it goes again, and this time when I reboot, I get drive failure warnings.  Great....just dandy....and I hadn\'t made a ghost image of what I had done yet, as a backup.  So I started searching pricewatch.com for HD\'s. I thought something in the 120gig range would be nice. I found a Maxtor 120gig 7200 RPM hard drive, with 8megs of cache on it, for $87. I called up our local Staples, to see if they would still match the internet, plus shipping/sales tax. Nope, they stopped doing that.  So, I hop onto another PC and order it on-line, as I can limp along for the week on a 4gig drive that I had sitting around. The drive will ship monday and take 6-8 business days to get here.  I can\'t wait, as the hd I have in now is only a UDMA 2, not a 5 like the 40 was.  I want my speed back! But...I can wait, it\'ll be worth the wait.   '),(17944,2101657,'22,February,2004','2004-02-22 00:00:00','    Well, most of the weekend went pretty well.  We went out to the River Walk by the Grand Harbor resort Saturday.  What a windy/chilly day to go for a stroll along the river!  And we saw 13 people who just had to be out of their minds!  We walked past the restort, and over the rail-road tracks, by the ShotTower.  We stood at the top of the north end of the sidewalk, and saw 13 people standing on the MIDDLE of the frozen Mississippi river ice fishing!     This is a ways down-stream from the lock and dam.  We\'ve heard the river moaning/cracking before, but to see *13* people ice fishing ON it [shakes head] that just seems nuts. Especially since the temp has been getting into the 40\'s the last week. We had snow and rain Friday, and more rain today.  The forcast calls for temps in the 40\'s all week [spring is springing!! WOOHOO!!].  Well, after the walk along the river, we went to BestBuy. I want to get a copy of DVD X Copy before they are forced to take it off the shelf. It\'s a really easy to use program, to use with your DVD burner. You can then make a back-up copy of your DVD.  The court says it\'s illegal now, but why not? Why should you not be allowed to make a backup copy, so incase your original gets scratched beyond all recognition and wont read anymore, should you not be allowed to backup your DVD?  Pure stupidity on the courts side..  Then we went over to the mall, to look around in the book store   and then wonder around in the video store and the game store.  Then it was off to Bishops to do as much as we could stand to eat from their buffett  .   It was in Best Buy that my head started to throb. Bright sunlight from out side, and while I was looking at software, some guy is on a PC right behind me, with MP3\'s he wants to buy BLASTING behind my head.  We got home about 4:45-5pm, and the first thing I did was to crawl in bed.  Where I thankfully fell asleep. I did wake up about 8-8:30 feeling totally ready to puke, but didn\'t, and fell back asleep.  I got out of bed at 8am... so that put me in bed sleeping for roughly 15hrs.  Yup, thats a very productive time..  Considering I had 50 pages to read for class, and 40 excersises to do, plus a program....all for tomorrow night.  Well, I\'ve got the first 2 sections of exercices done, still have 10 pages to read, and a program to write.  I have some after work, that I\'ll work on my program. I\'ll get the reading and problems done yet tonight.  Oh the joy of being a college student!   '),(17945,2101657,'16,February,2004','2004-02-16 00:00:00','    I\'m taking Intro to Programming, as part of my college course right now. I\'m learning to write in Visual Basic.net and it\'s quite interesting. It has a user interface that is incredibly easy to use.  I have my first program due tonight, it\'s really basic and simple. It just asks for you to input your age and resting heart rate, and it returns what your training heart rate should be.  Very simple.     But I was having a mental block, trying to remember how to get my data from the user.  Then I was like, DUH, smacked my forehead, and then put it in.  For my first program, I\'ve probably got 2-3hrs of playing around trying to remember how to get the input. After I remembered, it took 60 seconds if that, for me to have it working.  I\'m so proud!  Can you say WOOHOO!  So I have my program all printed out, and on a diskette to turn in tonight, along with my 20 or so exercises that were assigned.  I\'m quite happy that I have my homework 100% and ready to turn in.  It\'s a relief to just sit down for 5 minuets, and know, that you have nothing due right THIS instant, and can just relax.   Life is always about being busy busy busy busy busy.. It\'s about cramming as much into your waking hours as you can, before it\'s time for you to die.  You don\'t want to have accomplished nothing.  But....sometimes that nothing is just as important, so you can de-stress for a few minutes, before getting your nose ground against the grinders wheel.  Relax! Monday\'s almost done!   '),(17946,2101657,'08,February,2004','2004-02-08 00:00:00','    We\'ve got a neighbor thats 100% inconsiderate.  Yesterday while I was toying around with the snow blower and did our spot, and then the rest of the blow down to the korner.  A neighbor had her car burried, so I did the back end of her car, almost to the bumper. Figure she didn\'t have that much to shovel then, and I was drenched in sweat from the 30ft of street to the korner that I had done already.   So what does this neighbor do? She parks in front of OUR house, and leaves the area around her car untouched.  She backed up and left, and when she came back, parked in our clean spot.  Granted, streets are considered 1st come 1st serve for parking, even if you cleaned the spot out.  BUT, when a neighbor, w/o asking is nice enough to clean 80% of the area around your car, you think they could do the rest themselves.    So I went out and did most of what was left, before getting fed up with the situation. Yesterday, it was being nice and courteous, but today after they cant do their own space, she wants it done, it\'s $ time.   IIt\'s kinda fun to do, to be honest, but it\'s still hot work.  Z*lda and the lady up-stairs both left notes on her windshield complaining.  Thats my 2cents worth for the day.   '),(17947,2101657,'07,February,2004','2004-02-07 00:00:00','    Finally got to use the snow blower today.  I tried it Thursday night, and discovered, that it was out of gas.  So today, while talking with my Dad, I asked what the ratio should be.  You have to add 2.5oz of 2cycle oil to 1gal of gas for this little monster.  Well, I went out, started up the car, cleared off the snow, went down to my local Big A   and paid $10 for a 2.5gal gas can, and 1 quart of oil.    I made it easy on myself and just put in 5oz of oil into the gas can while I was still in the store. Made sense, I was nice and warm and it easy with it on the counter.  I figured that would work out great anyway, as when I then went to the gas station and put in 2gal of gas, it mixed it up nicely for me.  Get home, fill that bad boy up, take it out front, and then, pull my brains out to get it to start. It didnt like being cold either.     Finally got it going, and then proceeded to do the snow around our old blue beater mobile. Took care of that really fast, and I was beginning to think I might like this.  So I went past our neighbors drive way, to where I had parked our nice car yesterday, and started to clean that out.  When I left, so did our neighbor upstairs, and I beat her back, so I put our car in the clean spot in front of the house.  Z*lda and the boys were leaving shortly for a swimming party anyway.  So I proceeded to just do a 30ft stretch behind a burried in car.  I even helped by clearning the street almost up to her bumper.  It was VERY hot work though. I even did some of the street, connecting to the neighbors driveway, as I figure if it was clean, it would make it easier for us to park in front of our house.   I then went and did the side walk where I got snow on it from the street. Then for the heck of it, I even did the little sidewalk that goes between the two houses to the south of us.  I figure, it\'s helping out w/o being asked, and lots of neighbor kids cut through on that sidewalk on their way home from school.     Just like the BoyScout motto says \"Do a good turn daily\".   '),(17948,2101657,'06,February,2004','2004-02-06 00:00:00','    Ok, so it\'s only the Friday after this happened   but oh well. lol.  I have to admit, that I tuned in and watched the 1st half of the Super Bowl.  The commercials are what really draw me in. Plus the idea of what should be a really kickin half-time show.   Well.. for some reason the phrase \"purple horse shoes, green clovers, silver sun nipple rings\" popps into my head.    I\'m already being told that my finding what Jan*t Jacks*n and Justln Tim*rlake did as inappropriate, was the wrong way to feel.  But, JJ and JT took the right of parents away from all around the world, who were watching the show, that just MIGHT not have wanted their children to see that.    I wouldn\'t have an issue with it if it would of been on Pay-Per-View, or regular cable type channel.  By doing it that way, they would of been required to put a TV Rating code on it.  It was inappropriate, pure and simple, just as thats simply my opinion of it.  Right or wrong, my view is that it\'s wrong.  Plus the dancers sort of  , whom if you paid attenion, were wearing mini-skirts, garters, stockings, small types, that kind of stuff, if you were in the stadium in person, it might of made you not notice how JJ and JT were quite touchy/feely in a very intimate way if you get the drift.  Oh well.... My take is, that best part of it was the Budwiser commercial with the romantic date in the horse drawn carrage.  Ah, serious drama!   '),(17949,2101657,'02,February,2004','2004-02-02 00:00:00','    Make your own map  here .   I had to snag this neat little idea from Joy over at  Witches Moon .  I put down states that I have physically been in, I didn\'t include states that I\'ve flown over.  Missing from the map though, I see I forgot Mississippi and Arkansaw (or is that Alabama?)  lol   We used to do alot of traveling when I was growing up, during our summer vacation.  Have drove to Florida twice, went to South Dakota, with a brief foray into Wyoming to visit the Tetons and Devils Tower (no mashed potatoe jokes please).    The only reason I\'ve been to the west coast, was when I went as part of the National Guard. The one time I went to Camp Hunter Ligget(sp), we flew into VanDenberg AFB, and bused to the base.   The other time we flew from Waterloo, Iowa, to Las Vegas (didnt get to have any fun) and then were hauled in charter buses for 4hrs out into the Mohave Desert to NTC, also known as the National Training Center.  That was a fun time! Really! It was!  Very kewl way to see the desert.  Well, I\'m off to shovel snow, it looks like we\'re finally going to get some snow! woohoo!! so naturally what happens? My snowblower is out of gas (how ironic!)  Time to shovel!  And always remember, play it safe and be careful!   '),(17950,2101657,'21,March,2004','2004-03-21 00:00:00','    Well, Mr Migraine paid me a visit again yesterday.  We had a wonderful time, we all went out to the Swiss Valley Park Nature center area for a hike.  It was a  VERY  windy day, so on occasion, the 52deg weather felt  REALLY  cold.  We probably walked a mile or two, on very squishy ground. Was a very beautiful day and we had an excellent time.  Afterwards, we made our way over to Culvers. That was a good experience even.  Then we came home.  Then that tell-tale sign of the right temple started to throb.  I got up and took 1000mg of tylenol, hoping it might help some. Well, wonder of wonders, it didn\'t.  So, at 7:20 I climbed in bed. But...being me, an Episode of  STAR TREK: Enterprise  on that I hadn\'t seen, so I stupidly continued to watch it.  I made it to the end, and happy that it turned out to be a really decent and thought provoking episode.   Then I tried to get to sleep, and that was fit-full. Z*lda was working on quilt-squares or something, as I was laying in bed.  She stopped, and grabbed the ice-pack that I had handy, and I put that on my temple to try to help (think it did help some).  Then, I had that rumbling in the tummy that told me I\'d best move toward the bathroom.  Huh, what do you know, my tummy was right for a change.   After the massive bought of vomiting I had, I went back to bed, and felt better for about 10 minutes. Then the hot/cold flashes started. I think I fell asleep toward the start of them for a change, and went into dreamy non-migraine pained bliss.  Think it\'s time to check in with the doc to see if he has any migraine meds in-stock yet, or if I should just break down and have him give me a prescription.    If I get a prescription, I think I\'ll see if I can get the inhaler and pill form of Imitrex.  I\'ve found that my best luck is to take a tablet when I can start to feel the signs of a migraine, and then an hr later use the inhaler. That seems to give me the best protection.  I\'ve had that fight off a migraine on a few occasions that way.  The doctor has the injector version, but I just cant get into the idea of giving myself a shot (shudders). I hate needles as it is, and just cant get past the idea of using one myself.  I know...I\'m a wimp.  I\'ll stick with the pill and inhaler form of it.  So I put in another Saturday night in bed with a migraine. Got up about 8:40, so only about 13:30 hrs in bed again.  Migraines, what a wonderfully relaxing way to spend your evening.   '),(17951,2101657,'20,March,2004','2004-03-20 00:00:00','    Well, we had our Pinewood Derby for our CubScout Pack Thursday night.  We had a fantastic evening!  We had our weigh-in, and measuring (car has to fit inside a  box  thats 7inches long, and no wider the the car with it\'s wheels on).  We got all our cars checked in by the 6pm dead-line. We handed out the participation patches to all the kids as they were certified, just to make it easier.  I arrived at 5pm, to start unloading the track from the car, and getting stuff carried in.  The track is extreamly easy to setup. We purchased a 35ft 2 lane aluiminum track from  http://www.besttrack.com .  All I had to do was to put the pieces of track together, pushing in the ends of the metal dowell pins into the other piece of track, and clip it together with some binder clips.  We had the track together, with our electronic time keeper hooked up, was maybe 20 minutes.  I gave it a quick wipe down, and then we were ready!  We started out with our Flag Ceremony, and some opening announcements.  I got to hang our 2004 National Quality Unit Award streamer on our Pack Flag, and we handed out the patches to all the scouts to put on their uniforms.  We also handed out some patches for our Webelos Overnighter last fall (we ordered the new patches, which came out in 2004, and was our first chance to give them out).    I also had some achievements for my Webelos II Scouts as well. Then it was time to start the races!  While we were doing our awards, we also had our 2 judges looking at the cars, to pick a  Best Design  and  Most Unusual  from each den, and those cars would get trophy\'s.  We had 25 Scouts racing this year! Always a good turn out for Pinewood derby!  We had no issues, no squables, bickering, or arguing! It was awesome. No cars malfunctioned.  It was probably the best Pinewood Derby we\'ve had while I\'ve been a leader.  The cars were great, there were some excellent races (Including * 2 * re-races, as we had to have the cars flip sides and go again, because they tied!).   All-in-all, it was just an awesome evening.  Z*lda made up 2 kerchiefs for the winners, out of a checkered flag material. These went to our 1st and 2nd place winners, and they get to wear those kerchiefs to districts where they represent our pack, and to our last 2 pack meetings.  Our oldest who is a Webelos II won a trophy for Best Design for his den.  Both boys want to take their cars to district to show, to have a chance for a trophy there.  Our youngest got one last year for Most Unusual Design.      All I can say, is that I really get into doing the Pinewood Derby races, and it just gets funner and funner every year.  I\'m going to help out the local Boys and Girls Club\'s Pack a hand, and take our track down Tuesday to help out with their races as well.   Might be helping out another pack as well, they have to contact me yet.    Cant tell I had a great time can you?  It\'s a very energetic evening, and it tends to last for a little while :-) Can\'t wait for next year, to do it again! :-)   '),(17952,2101657,'16,March,2004','2004-03-16 00:00:00','    Well, I must say, that the world looks really different for the moment.  I had an eye exam after work, and my pupils are still extremely dilated, so everything is a little fuzzy/blurry.  Plus, I\'m not wearing my glasses, and from what the eye doctor said, I don\'t need them.    Talk about a shock!   I\'ve worn glasses for the last oh.....28yrs? and suddenly I don\'t have to wear them? Wow! I need to get some sunglasses! :-)  It sounds like after a few weeks, my eyes will have adjusted and I won\'t have any of the weirdness of looking at the world w/o glasses any more.  I\'m no longer visually impaired! WOOHOO!   '),(17953,2101657,'14,March,2004','2004-03-14 00:00:00','    Got both boys Pinewood Derby cars done today. We got the decals all on, got the clear coat of varnish on, wheels and axles mounted, and weights put on. Cars weigh in @ 5.0 oz\'s, just what we want to have them at.   now all we have to do is wait until it\'s Thursday, then it\'ll be time to try them out and see how they do!     We have a half-hr weigh in period to get all our Scouts through. They have to put their car in a wood \"box\", to make sure it\'s not over 7inches long, and no wider then 1 lane, so it doesn\'t interfere with the other car in it\'s own lane.  Then we\'ll put the car on our new digital scale, to make sure it weighs no more then 5.0 ounces.  Then the car can get it\'s official race day number, and then be put on the table, until it\'s time for that scout to race.  We\'ll start off with all our awards to give out, and we have alot to give a way.  After about 10-15 minutes we should be done, and then it\'ll be time to start racing!  We have about 24 scouts, and we\'ll be doing a double elimination.  Last year our oldest son went undefeated and took 1st place for his first time.  He went to district but didn\'t end up winning. Great experience though!  Also, tomorrow, I start my 2nd class, which will be Philosophy and the family. So that\'ll be 2hrs more of classes on Monday/Thursdays..those days are going to be long ones, I can see that already.  Still, I\'m pumped, I love Pinewood derby, it\'s our funnest meeting of the whole year in my opinion.  Will post how the boys do! Wish them luck!   '),(17954,2101657,'22,April,2004','2004-04-22 00:00:00','    and it feels like I\'m still recovering from being sick over the weekend.  Our oldest was sick last Thursday, and the rest of us came down with it during the night late Friday night.  I was supposed to go help assist with some training at the local district Scout Camp for OWL (Outdoor Webelos Leader) and Baloo (Basic Adult Leader Outdoor Orientation). Well, that didn\'t happen. I felt bad as there was no way for me to contact them to let them know, since the leader I was going help was there Friday evening already.  I woke up about 4am, and already didn\'t start feeling very well. I realized later, that I had already started feeling bad the evening before when we went out to eat, and I was starting to feel extremely tired.  Well, I had my clock set to go off at 7:15, which would have given me time to get my clothes together, cleaned up, and on the road. I had already packed my tent, sleeping back, a chair, and a few other things in the trunk of the car the night before.  By 7:30, I had started to throw up.  Oh thatâ€™s such a joyful experience. I wonâ€™t go into details about what else was happening, but needless to say I spent a lot of time in the bathroom.  That continued during the whole day.  I napped for about a half-hr or so, and just couldn\'t seem to quite get any actual rest. So about 4, laying on the couch, hips starting to ache like I had a case of the flu, I had the boys throw \"The League of Extraordinary Gentlemen\" in the DVD player. We have it on loan from my parents.  It\'s an interesting movie. It\'s not the greatest movie in the world, but it was a fun watch.  Some of the plot lines are more then paper thin, and I was able to anticipate what was going to happen, quite easily before it did.  The boys seemed to enjoy it though, and they seemed to like the special effects. Personally, I while I thought the Sub of Capt. Nemo\'s was very impressive; it was completely wrong/inaccurate to what is described in the classic \"20,000 leagues under the Sea\".  Plus, I never realized that Capt. Nemo was from India. Go figure.  By this time, I had started to sip on some Kool-Aid, as I was dehydrating very fast.  I also had some chicken noodle soup, sans noodles. That was all I had to eat.  Sunday I felt better for the most part, some more energy, hips still aching though.  I had some actual food type food that day.  I\'ve been slowly getting myself to eating more daily, but it\'s still taking some time to get my appetite back.    I\'ve spent some of our tax return myself finally, and can\'t wait for the items to get here.  I ordered the 19 inch monitor that I\'ve been wanting for ages. It should arrive in a day or two.  That will make it easier on Z*lda, since the 17 is hard to see text on for her.  It shouldn\'t be much different in size than our 17 inch monitor now, as it\'s a Nokia from 2000, so it\'s on the big side.  That will go down to our youngest son so he\'ll have something better then a 15 to work on.    I\'ve also made it official and have become a card carrying member of the Star Trek Fan club! This is the part that cool to me. I\'ve wanted to join the fan club for years, and finally have.  It\'s $30 to join, and for joining up, if you pay the $6.50 shipping, you get a $20 lithograph for free (is anything truly free?).  So I ordered another lithograph to go with it, that shows the Enterprise NX-01 in space next to a planet.  Just wondering where I can put that up, w/o being too obnoxious about Trek :-)  I almost got the Captain Proton one, which is from Star Trek: Voyager. It\'s based on a \"serial\" type story line in the holodeck that Lt. Paris and Ensign Kim use. Think of the old Tom Swift/Flash Gordon matinee\'s from the 30\'s/40\'s and you\'ll have it.    School is getting close to being done, and I can\'t wait.  Taking two classes, while being so active in scouts, and work, and trying to have a life just runs you down.  I have some research to do, as I\'m going to be writing a paper on the topic, that if it should be required to have a license to have children or not.  I thought it would be an interesting topic, as I can easily and readily see how the argument could go both ways.  You have to have a license to drive a car, so why not to have kids? Seems to make sense.   '),(17955,2101657,'09,April,2004','2004-04-09 00:00:00','    Have gone to the course 4 times in the last week, and am loving it.  We haven\'t played a complete 18 hole session yet. We\'ve done 15 holes. It\'s a very challenging course to walk, due to all the steep inclines, but it\'s a fun work out.  We\'ve thinking of going again today, due to the nice weather, and having the day off for Good Friday.    While playing Wednesday, I almost lost my Innova \"StingRay\".  Was the 7th hole, and I had what I thought was an excellent throw.  I thought I saw where it landed, and then the boys threw their discs.  We moved down the fareway, and I started going to where I thought my disc had landed.  Whoops! Where\'d it go!  We spent probably 20-25 minutes looking around for it. Then, for the first time, I decided to just play on and call it a loss.  We played the 8th, and then the 9th/10th play back along the way we came.  So I kept my eyes peeled and looking, just in-case.  Woohoo!! There it is, it had landed, and at an odd angle to spot from up where we had been.  So I\'ve yet to lose a disc, but came very close.  I had just got done telling the oldest son, that it would be a beautiful day Thursday, and with a nice afternoon sunshine coming down, that it would be real obvious to anyone to find.  Luckly we found it first.  Can\'t wait, it\'s 52, should be a great day at the course! :-)  Will post some pics of our trip, so show what the course looks like.   '),(17956,2101657,'06,April,2004','2004-04-06 00:00:00','    Was my last post really on the 21st? I\'m behind. Course, it seems like I\'ve been very busy lately, what with work, school, and scouts.   Plus, Sunday was my Birthday! I turned 36.   We had our CubScouts out camping this weekend. It was a huge success I feel. Course, it could of been warmer over-night (no issues, the kids had bedding, and a double-barrel wood stove to keep them warm), and I will next time make sure to NOT plan the camp-out on the weekend of Day Lights Savings time. The oldest son had to get up at 6:45 adjusted time Sunday morning.  One thing led to another, and 98% of the scouters decided to get up at that time too.  I decided to \"lay-in\" and got up at 7:15, with the last 3 other people.  It was still too early, as in my opinion, it was only 6:15.  Made for a long tiring morning. We got the boys all fed, got the cabin cleaned up, so it was nicer looking then when we arrived. We got the outsides picked up, took 2 pics of the all the scouts who attended, and then headed out the gate.  At least this year I didn\'t have a cow act like it wanted to slam it\'s front hooves on my hood.  In the afternoon, we went down (just our family) to the new Dubuque County Jail, for a public tour. It\'s taken, like forever, to get this jail finished, but I must say, it is pretty impressive.  It\'s only the 2nd jail I\'ve ever had a tour off, and it was #1  lol  Very interesting designs.  I was really surprised to notice, that there were cameras  EVERYWHERE , but in the stair wells. That seems like an obvious place, since it takes a key to operate an elevator.  It\'s actually 71degrees here!  Can\'t wait, means every day will soon be like this, and it\'ll be time to hit the frisbee golf course over by Veterans Memorial Park off Kane St.  It\'s a really awesome course, but it  WILL  give you a work-out, as just like everywhere else in Dubuque, it\'s full of hills.   '),(17957,2101657,'25,May,2004','2004-05-25 00:00:00','    Went to the Doctor today with the oldest so we could get a physical, to attend Boy Scout camp this summer.  While there I had the Doc look at a spot on the back of my lower left calf. It was decided, to make an appointment for next week to have it removed, based on his impression, it looks like it\'s skin cancer.  We\'ll know for sure after they check it out. It\'s a small spot about a quarter inch round, that on occassion hurts into the muscle. There is also another spot about 4 inches above it thats probably in my opinion going to be the same way.  Plus there are 2 other spots just like that one on the back of my left shoulder.  Hopefully I wont turn into a walking tumor.  Time will tell, I get to have it removed on the 2nd of June.  Then 2 days later I will be camping with our Boy Scout Troop over-night, to help setup for a fund-raiser the next day.  Too much to do, too little time to do it in..   '),(17958,2101657,'24,May,2004','2004-05-24 00:00:00','    Well, we went out and tried our hand at the Disc Golf League.  Oldest son and I went.  I played in the \"B\" group, Oldest son went with Disc Golf DudeÂ© and was on in the \"A\" group.  I was naturally the worst of the group of 4 I was in, but I did have a few decent shots.  We started playing shortly after 5:30, and it took about 2hrs to play all 18 holes.   It was different, as we were playing from \"alternate\" tee\'s, which I think helped really throw me off my game.  It made all the wholes 20-50ft shorter.  My best shot of the night came at the very end, when we had a CTP (closest to the Pin) throw.  We went outside the entrance of the park, and stood by the parks sign, and threw 2 discs at the practice basket. The closest to the basket got a free \"mini\".  I was about the 3rd or 4th from last to throw (about 18 of us).  I threw my Super Roc tie-dye, and I about fell over stunned, as from the 300ft shot, I landed it about 6ft from the basket.  WOOHOO!! So I got a few mini-marker, so when I throw, I can drop my mini to mark my spot, and pick my disk up to throw.  Oldest son got a free disk as well, for playing as the younest player. He got a free FireBird, a disc I\'ve been wanting to pick up for a specialty throw.  I was trying to talk him into giving it to me, but he wanted it as it\'s an Ultra-Long Range drive.     He just offered to give it to me, but I told him to keep it, as he wants a ULDÂ©. He\'ll have to work to throw it as far as ULDÂ© should be, but as he\'s so young, he\'s not throwing his Mid-Range driver as far as he should.  We\'ll keep practicing, and we\'ll get to the point where we can throw these chunks of plastic as far as we should, with control. It\'ll just take some time, but we\'ll get there   :-)   '),(17959,2101657,'11,May,2004','2004-05-11 00:00:00','    Well, school is over for the semester finally. WOOHOO!!  I had 2 classes this term, (a B- and a B) which I\'ve decided is almost too much to do at the same time as working a full-time job, being active in CubScouts as a Pack Leader/Den   Leader, and trying to help people with their home PC\'s. I\'ve got to find out how to get my grades yet, as I\'m eager to see what I ended up with. Plus, I\'d like to get submitted for tuition reimbursement (money NEVER hurts).   It\'ll be nice to have my Monday/Thursday evenings open for the next couple of months for a change. The boys and I have really gotten into playing Disc GolfÂ©. We have a really awesome course here, just a hard one for beginners due to the EXTREAMLY number of trees lining the course. It\'s a very beautiful course, located by Veteran\'s Memorial Park on the west side of the park.  The oldest son and I went down Sunday at 1, as they had mentioned on the local Cable access channel that they would be having 2hrs of lessons, for $5/person with a disc included.  We were there, and the only other person to show up was the instructor.  They mentioned about pre-registering but I hadn\'t seen anything on the TV about that. They said we didn\'t have to pay but I still did, just to cover their time.  The guy who was teaching us REALLY knows his stuff, and is going to be playing at the World\'s in Des Moines, Iowa this August.  He\'s an Amateur in the PDGA, and said next year he\'ll probably go Pro.  That guy can really heave a disc.  On the 10th hole, which is 400ft, he JUST missed the basket, and that was with just 1 throw. He was able to tell after seeing me throw 2x, that I used to play tons of Frisbee when I was a kid/teen.  I have to remember to follow through with my throw, and also have my arm higher up when I throw, as I hold it too low.    He was saying that I should also use a power grip when I throw from the tee too, as the more power I throw it with, the more the disc will spin, and the further it will go.  After seeing the way this dude threw, I want to throw like that too.  He also showed me a handy throw, where I throw the disc side armed, and instead of pulling my arm all the way back like I\'m going to throw a baseball, I just take it about half-way back, and snap my elbow forward, and give my hand a flick, and it will send the disc out straight, and then cause it to curve to the right. This kind of throw will be extremely handy on the 8th hole, where they turned it into a dog-leg.  Only thing is I need a new disc for it, as the only one I have thatâ€™s good for that throw I let my youngest son have.   We\'ve just gotten into the sport really, this spring, and we already have 11 discs between us.  Course, this is nothing, as our instructor Sunday says he has about 200.  This guy knew all the stats on the discs that we had even, being able to tell us which ones were the better ones, their speed and turning ratios, everything. He was very impressive and I hope he does really well at the Worlds.   It goes on for 5 days, and there are 5 courses in  Des Moines with each player playing 1 course a day.    It was also very hot Sunday while we played.  It was 83Â° and humid, so when we got back to the car after playing 18 holes, I was soaked.  I got the boy and I some PowerAid to carry on the course, and we had that gone by the time were on the 15th.  We got back to the car, we had some water bottles that we had with ice in them, and they   were totally melted due to how hot the car got.  All I cared about was the fact that it felt ubber-cool as I poured half of it over my head.  Boy did that feel good!  After watching the Disc GolfÂ© Dude throwing like he did, it makes me want to practice that much more so I can be a 10th of how good he is.   '),(17960,2101657,'10,June,2004','2004-06-10 00:00:00','    Well, the good news is, that the 2 pieces of me that were carved out by my Doctor last week as possible skin cancer came back negative.  Neither one was at that stage.  The way I\'m thinking, is it\'s good it\'s gone, so it doesn\'t have a chance to go that way.     The other good, is that the stiches are out.  My back is fine, Doc says that it\'s healing nicely.  The leg..well...thats a different story.  The stiches hurt alot more when they removed them from my leg then my back.  I asked Doc about how the skin was really pulled tight and causing the skin at the edge to bunch up in a line. He said they called those Dog Ears.  Woof woof I guess.  So, he said just take it easy, I\'m pretty much good to go, and sent me out the door. My appointment was at 4:15, I was leaving the parking ramp at 4:24.  The body parts smarted a hair, from having the stiches pulled out.  It\'s pouring practically by the time I\'m ready to step out of the car at home.  I open the door, put out my left leg (can you see this coming yet?) and I must of twisted my leg just enough as I stood up out of the car, that my leg popped open. Oh, talk about stinging.   I came into the house, went right into the kitchen, grabbing a phone book so I could call the doc back.  I\'m sitting at the kitchen table, dabbing the back of my calf with a paper towel, to see that yup, it did pop.  I have a live person from Doc\'s office at 4:41. They said how soon can I be there, I said 9 minutes, they said come back in.  Doc comes in, cleans the leg up, said Yup! it\'s popped. I\'m thinking SOB, I have to get one of those railroad spikes jammed into my leg again to numb it out.   He said Nope! Just pack it with antibiotic ointment, and cover it with a band-aid.  So, it gets to heal from the inside out.  Hope it\'s done in time for BS Camp, really looking forward to that.  I ran to Wal-Mart to cash a check at the bank, and decided to look at their multi-purpose tools, as I\'ve been w/o one for awhile now (if you know a techie geek and they don\'t have their multi-purpose tool, you know their mentally lost).  I picked up a nice little Gerber for $32w/tax.   My leatherman\'s small tip screw driver BROKE, which really pissed me off to be honest.  I decided since a new Leatherman would run me $29.99, and it was IDENTICAL to the one I have, which happens to be about 7yrs old, that I wanted to look at the Gerber Multi-Pliers 400.  Thats the one I ended up getting, I liked the pliers design, and the locks on the accessories that flip out, so they wont close on you while pushing with all your might (you know what I mean, and man is that a blood blister!).     Will see how good the Gerber does. I had a Winchester MultiPurpose too that I got last fall for $15, and it wasn\'t worth it.  Was $15 wasted. Looked nice, functioned not so nice..   '),(17961,2101657,'06,June,2004','2004-06-06 00:00:00','    Well, the Oldest SonÂ© and I went to the 7PM showing of the new Harry Potter movie. We are both huge fans of the series and the movies.  The books are really something else, and the 1st two movies are excellent.  I wish I could say the same for the 3rd movies adaption of the book.  I was dissapointed to the EXTREME to say the least.  They left an incredible amount of information out that is vital to the story, changed various things around location wise, made changes to BIG things in the story like, having so-n-so say them instead of who should have.  They added stuff into the story that wasn\'t there in the first place, and didn\'t put in stuff that in my opinion is crucial to the plot and over-all story.  I\'m amazed that J.K. Rolling let them do this to the story and I hope to Gaea that she doesn\'t let them do the same thing to \"The Goblet of Fire\".  Some of the filming styles were quite well, I\'ll admit that some of the scenes were nice with the way they were filmed, some of the HUGE changes to the grounds of Hogwarts was unacceptable.  I won\'t say anything for awhile, as I don\'t want to spoil the movie for anyone.    I just came out feeling very let down, like an old-friend had just cheated me.  Such is life I guess...   '),(17962,2101657,'05,June,2004','2004-06-05 00:00:00','    We had a fun raiser that we did last night for the Boy Scout troop that Oldest SonÂ© and I attended. We started out with a hike over to Wendy\'s for a quick supper, then it was a hike back to the track, for a tour of the track it self.  It was actually quite interesting.    We then went back to the camp area, and got organized. Then some of the boys wanted to go back and sit out in the bleachers (free pop!) and see what it was like to watch the dogs go zipping by.  After sitting for 2 races, we decided to head back out to our camp site, setup some tents, and then the boys procedded to play Capture the Flag with glow sticks.  I think that was Oldest Son\'sÂ© favorite part of the camp out.  He got to help setup a Eureka Timerline two man tent and spent the night in it, so he got one of his requirements met.   They did a fund raiser in association with the Walleye Club, and Oldest SonÂ© earned some $ for his individual account for Scouts, to use towards Scouting equipment/camp.  We got up about 5:45am, and started getting into gear.  Things went very smooth, and I was able to leave at 7:30 to get to the house and clean up to get ready to help out at a Cub Scout Day Camp at a local school.  That ran from 8:30 until 3pm. Was a great time!  They had some people dressed up in 1200-1500BC type clothing to reflect King Arthur.  Was a perfect day(little warm tho) for the Day Camp. The kids got a really cool bag, some treats in it to have during the day, a really nice T-Shirt, and an awesome looking patch to wear on their uniform.  It\'s been a loooooooooong day, and we might still go see Harry Potter 3 tonight, otherwise it\'s tomorrow to see it.   '),(17963,2101657,'02,June,2004','2004-06-02 00:00:00','    Well, the doctors appointment went pretty well.  I had them look at the two spots on my left shoulder, and they decided those are ok right now, but he then went on to look at 3-4 different spots on my back, and decided 1 had to come off today.  So I have 2 stiches in the middle right side of my back, 3-4 inches over from my spinal column.  The other is from the lower part of my left calf.  That one took 3 stiches.  I get to go back a week from tomorrow to have them taken out.     You would think, since novocaine would hurt so much going in.  The stick of the needle sucked, but to have that liquid fire going into your leg...OUCH!!!  I got to look at the spots the doc took out, and was surprised at how big they seemed.  The one on the leg was 8mm and they used a \"punch\" to take it out.  I should know by next week when I have the stiches taken out if they are actually skin cancer or not. Based on how they looked (EW!!) I\'d say the qualify.  Hopefully they don\'t though.   I took 1500mg of Tylenol (generic) to make sure I was ready for when the pain starts to kick-in.  Hopefully it wont be too bad, but judging by how deep he went, I\'m \"definetly going to feel that in the morning\"   '),(17964,2101657,'16,July,2004','2004-07-16 00:00:00','    I was home from work about 30 minutes late, to make up for doing to the doc during the afternoon.&nbsp; When I got home, the family was gone. They were about 10 minutes behind me..&nbsp;  &nbsp; Z*lda said she was going to go to a garage sale out by John Deere Road. I was like, ok, sure, go ahead.&nbsp; She comes back awhile later, and as I come into the living room says \"I picked this up, is it worth anything?\"&nbsp; Just from the shape of the package I knew it was a Palm Pilot, but not what kind. It could of been anything really old as an original Palm One or 1000, or something newer. &nbsp; Boy was I in for a surprise! I take it out of the bubble wrap, and it\'s a Palm M515! in perfect/brand new condition!!!!&nbsp; I was like OMFG!!! What did you pay for this!&nbsp; She then said he had $35 on it, but she only had $26 on her, so he sold it to her for that, plus the 3-4 other things she had in her arms!!! &nbsp; I checked it out on Pricewatch, and it still sells for $241 shipped! OMFG! This is just incredible!! &nbsp; And it works! Beautifully too!&nbsp; It\'s got 16 megs of memory, and runs 33mhz, and is color!&nbsp;&nbsp; Now, in a PDA, thats awesome!&nbsp; Plus I can buy one of those new style memory cards to put into it and increase it\'s ram. I\'ve got tons of software installed on it, and files put onto it already (with other stuff left to transfer yet), and it STILL has 10megs of space left (has 4 megs of memory thats used I\'m assuming by the Palm OS 4.01. &nbsp; This is so frellin kewl!! &nbsp; Now the youngest is eager as I said if I ever got a new PDA he could have my HandSpring Visor Deluxe. Now he cant wait to get it! :-)   '),(17965,2101657,'11,July,2004','2004-07-11 00:00:00','    Wow! What can I say? We went and see Spider Man 2 at the theatre last night, and it\'s just awesome.  The special effects for Spidey this time are incredible and very fluid. I couldn\'t spot any obvious jumps between live action and CGI.  The story was awesome, got a bit boring at times, and didn\'t screw with the ideas set with the comic book as badly as the 1st one did.  They have all the cast from the 1st movie back, with a short appearance of William DeFoe at the end.   The actor they picked to play Dr. Otto Octavious is perfect. The guy couldn\'t of been any better a choice if they tried, he was out and out perfect for the roll.    I must say that I\'m very impressed with it, dont want to go into details and ruin it for anyone reading (there are people reading this site still I hope?). All I can say is that when this bad boy comes out on DVD, there will be a place on our shelf waiting for it.  On a side note, picked up a fun old movie that just came out on DVD last month. Wizards.  Who remembers that one?  Different type movie, but it\'s fun as well. Was interesting seeing Avatar the hero in action again. It\'s a movie that tells the tale of good vs evil (go figure) but this time, good is 3000+yrs old, is kinda horney, and gets the girl in the end (not literally, who knows might happen off screen tho).   '),(17966,2101657,'06,July,2004','2004-07-06 00:00:00','    The oldest son and I got back from spending a week out at Boy Scout camp. We left on June 27th, and got home around 10:45am on July 3rd.  Boy was I tired!  We left the 27th at 11:45, and after picking up a few items we needed, we headed out to Camp C.S Klaus, located just outside of Coalsburg, Ia.      We arrived at camp, for our check-in time at 1:45. The Scouts all took their back-packs out of our Troops trailer, and we then proceeded to hike the 3 quarters of a mile to our campsite.  We were staying in Owl Hoot.  Got camp setup and everything organized. That evening we had our Council Camp Fire with all 7 troops present and the Camp Staff as well.  We got about 3/4\'s of the way through, before the sprinkles we\'d had off/on all afternoon, turned into a steady rain. We stopped the fire, and then started heading back to our camp-site.  By the time we got there the rain was coming down heavily, and we were soaked from the knees down (we had poncho\'s on).   On Monday the boys started working on their merit badge courses, and our 4 1st year scouts started on their \"Adventure Trail\", where they were working on 3 different merit badges (Swimming, Basketry, and Indian Lore).   On Sunday all the scouts who wanted to use the water, had to get pass their swim test. There are 3 levels, White, Red, and Blue (being the highest and needed to work on the Swimming Merit Badge).  My son just got his white on Sunday, and then Monday took the test and passed for the Red Swimmer.  Tuesday he took his Blue and passed. That took swimming 4 25m laps, with the last lap a backstroke of some sort, and then a 5-10 second float on his back.  He had a cheering section going by the time he finished.  It\'s a big accomplishment when your 11, and you take a swim test that has you swim the length of a foot ball field. Way to go!!  We had some activities that night, on a troop level, that started off with a tug-of-war.  It  was quite hilarious actually. When we walked up, 2 groups were already going, and the 1st boy in the water (they dug a pit and filled it with water, making it a very muddy mess), did a dive, just like you\'d see on Most Extreme Challenge off Spike TV. If we had had a camcorder, we could of won money on those home video shows, it was that funny.  By the time it was done, the same boy had ended up in the water 3 times.  They had a few other actives, like a team getting their members over a rope 4ft off the  ground, w/o touching the rope.  I was busy with the camera trying to get some decent pics of it as it was going on.  There was some other actives involving boards and ropes and carrying team-members around.  When we were done, we then made the almost 1mile hike back to our site.  The oldest son also had his first attempt at shooting a 22cal rifle and had a blast literally.  I\'ll have to take him out so he can practice for next summer as he now wants to take the Rifle Merit badge (along with Archery).  Wednesday was a bit of a challenge for him, as I think the long days were starting to get to him. The 1st years had to go on a 3 mile hike (1.5miles one way). He got back and was almost wiped from that, then had to go down and work as a server at our table.  I helped the 1st years with their waiter duties, as there was nobody there to tell them all what they had to do.  Then the 1st years headed down to the parade grounds for an evening of games and activities.  Oldest son decided he did NOT want to stay down there for the \"sleep under the stars\" camp out, and came back to the camp-site about 9:45.   Thursday was supposed to have the 1st years start out with a \"Polar Bear\" swim, which he didn\'t want to do either.  He was getting a little homesick by this point as he had cried some on Wednesday night before bed, and Thursday morning.  Thursday he had a better time of it, as he had another chance in Field sports to shoot the rifles and the Archery range again.   We then spent 2hrs down at Aquatics (the beach) for the Water Follies, a bunch of fun events that involved water.  One of the best practical jokes of the week went on at this event, and I\'m sure will go down in the annals of Klaus folk lore.  I wonâ€™t describe it here, in an attempt to protect the not-so-innocent lol  We also had family night that night, so about 5pm families started to show up, and the scouts could take them around camp to show it off and some of the activities they had done.  Since it was Family Night, I got to walk around with my Son, just the two of us.  We went over to the Water Falls, which were just perfectly lit, and looked awesome.  We went over and watched the demo from the fire truck, and had something to eat.  Then it was time to head up for the Council Fire, which was pretty neat. It got just AWESOME when they started the tap out ceremony for the Order of the Arrow.  I won\'t even try to describe how incredible it was, just imagine a Pow-Wow drum going, torches, and people dancing and throwing said fire torches.  It had to be the kewlest ceremony I\'ve seen in Scouting yet. After that was done, we headed back to our camp-site to put an end to our night.  The only mishap of the week happened about 3:36am that night, when my son decided to roll a little too far to his right on his bunk, and rolled out of the tent.  The tents, are Boy Scout type tents, which are on raised wooden platforms, 2 poles holding the center up, with a beam between them.  My sonâ€™s side of the tent was about 2ft off the ground, and his bunk added another foot and a half to that. I heard him fall, and jumped out of bed, and with flash-light in hand, leaned over his bunk and looked out the side of the tent. Sure enough, there he was on the ground in his sleeping back, sitting up wondering what had just happened. He only really remembers the â€œwhooshâ€ of the air going by him as he fell.  I told him to climb back up into the tent, and he looked like an inch-worm in his sleeping bag as he did it.  He fell asleep almost as fast as he had fallen out of bed.  It took me another 45 minutes to fall back asleep after that.  I did listen to the Owls hooting away though as I waited.  Friday the Scouts got to put finishing touches on any Merit Badges they hadn\'t finished yet. My son got his Indian Lore and Basketry badges finished up, and some of his Swimming done as well.  2 out of three ain\'t bad!  It was also a chance for me to perform the practical joke that I\'d had planned for the week.  We had 3 boys who pulled a joke on a leader and shouldn\'t have gone into his tent. So....I enlisted some help, and I put their cots on the roof of the latrine. It worked out very nicely. Even have pictures of the not-so-innocent helping me do it.  All 3 Scouts took it extremely well, and we helped them pull their stuff down after it was done. By evening, we got ready for our last Council fire, were each troop had to put on a Skit and do a cheer.  Our Skit was so long we didn\'t do a cheer.  It was BY FAR the best skit of the evening. We did \"The Wizard of Klaus\", and it was hilarious.  I got to be the Wizard due to being able to do a deep voice that projected so the audience could hear.  We headed back to our camp site to finish up our packing (we pull all other stuff that we didn\'t want to haul-out on our hike out, back in the troop trailer), and then hit the sack. We got up at 6:10, so we could get ready, and have our Camp Site cleared out and signed off by 6:45am. We then hiked down the hill to the dinning hall for a quick cinnamon roll and milk, before going down to the parade grounds, were we threw our packs into the trailer, divvied up the Scouts between the cars, and then made our 40minute drive back to Dubuque.   We then proceeded to divvy our gear back up into our respective vehicles and headed home.  It was an incredible week, very tiring, but we had an awesome time.   The leaders who stayed all week, at the last Council fire, were even presented with a Mini-Multi tool that say \"stayed all week leader, Camp C.S.Klaus\" which I thought was nice.  But what was nicer?  Getting to spend the week there with my Son, having a great time, and getting to sleep again in my own bed.  We can\'t wait until next year! Is it 2005 yet?   '),(17967,2101657,'03,August,2004','2004-08-03 00:00:00','    We had the opportunity to go to a political rally tonight, and we got to see John Kerry speak in person. I was impressed, even though I stood most of the time we were there. We got in line at 4:10, and the speakers started at 6:15. It was almost 7:45 before we saw Sen.. Kerry speak. I thought it was a good speech! It was an excellent chance to see THE future President of the USA in person.  In 1991 we got to see Al Gore talk at the Waterloo, IA airport, along with Sen.. Tom Harkin. Tonight we got to see Sen. Harkin again, introducing John Kerry. We also got to hear a few other local politicians, and we also  got to have 3 songs sung to the crowd by David Grohl from FooFighters, and one time drummer for Nirvana. I\'ve heard the songs that he sang on the radio, with the full band, and they sounded way different, but enjoyable ( urlLink Z*lda  didn\'t like them).  We just got home about 30 minutes ago, so we were there almost 5 hours, but it didn\'t seem like that long. We had to wait in line about a half-hour, and then go through a metal detector system. This meant that as we went by we had 2 sets of keys on the table, 3 PDAs (mine and the boys), a digital camera, and a purse. It was a very interesting experience, but it was worthwhile!  There was a quick little  jab  that they mentioned that I did like too. They had a little quip, referring to the current President Bush, and his father... it was  \"Like father like son, one term and he\'s done.\"  That stuck in my mind, really liked it for some reason :-)   '),(17968,3551081,'19,June,2004','2004-06-19 00:00:00','   A broken bridge   Times like these wish I were dead, Drowned in gloom, in hatred red Locked in chains tied to my bed Oh to what misery, have I been pledged?  The gloom and doom of painful nights Have locked me in this dreary plight My wings are broken, can make no flight Blackness covering, and blinding my sight  Wicked words and evil deeds Poisoning within, this mead Planted carefully, of blackness this seed Oh to where will this dark road lead?  Someone help me! Please assist! This anger growing like a cyst! Light a fire, clear this mist, Clenched hard like rocks, are my fists  For love and hatred do not mix One or the other, you must pick This broken bridge, this heart must fix, Like holy water no wolf can lick  by norain isa   A new beginning  Sneak a peak through morbid eyes See the fire that rages inside To act ungodly is a sin What devil, what satan lies within  I wish someone would take me high Upon the clouds, see heavens bright Release me from this deadly lock And dress me in a pure white frock  by norain isa   Me  If I could swim away, I would To explore the hidden sea Secrets written across the sand, Unknown to you or me  If I could run away, I would And caress the sandy peaks A world of laughter, happiness and joy Hidden away, for no one to see  If I could fly away, I would And join the birds so free To learn the myths, stories untold And find the inner me  by norain isa  '),(17969,3551081,'18,June,2004','2004-06-18 00:00:00','   urlLink    meeeeeeeeee&nbsp; urlLink  '),(17970,3551081,'18,June,2004','2004-06-18 00:00:00','   to the beings that seem worthy and true can only be met with despair the hard hand of tyranny and injustice backlogs years of care and repair  the ambitions and dreams of one soul determined prayers and hopes are uttered unheard a long winding road,destroyed and deserted makes each moment, a harder lesson to learn  my attempts at peace,in vain and futile faliure caused only by my own contemt,my own actions i cannot surpress,emotions conflicting a little black man forcing every reaction  a series of cries,of screams and of moans reflect on a story,a story untold locked away by the sinews within harder than diamond,reflecting my sins  red bubbling poison,rises within a coarse,fiery demon fueling it further wildly taking over,my soul and my spirit something keeps it inside,cold this benefit  a glimpse of my anger which simmers inside hells raging fires seem so cool and mild a black heart,for one black soul could this mean the end?there\'s more in store?  a white streak of lightning across the skies can seem so beautiful in my very eyes a rumble of thunder,not unlike my chest a scream in the distance,it comes from the west  today norain holland  '),(17971,3551081,'16,June,2004','2004-06-16 00:00:00','   urlLink    my dream car&nbsp; urlLink  '),(17972,3551081,'16,June,2004','2004-06-16 00:00:00','   urlLink    my best fren rachel june sim&nbsp; urlLink  '),(17973,3551081,'16,June,2004','2004-06-16 00:00:00','   urlLink    graduation in tchsd 2004 me and tatty&nbsp; urlLink  '),(17974,3551081,'16,June,2004','2004-06-16 00:00:00','   hello every one i noe i have been boring you with poetry and all that...its just my way of expressing myself....my favourites are the romantics like william blake and lord byron......im kinda new to this stuuf...i have been writing 4 about 2 years.....ill post some of my better stuff soon...as soon as i can locate them...ill be starting photography soon.i saw a blog recently and it was some really nice pictures....he was pretty good...i would love to catch the essence of the world around me....it would be nice to keep track of this rapidly changing world we live in....i dunno why i have suddenly become so passionate about everything nowadays....its so surreal suddenly...like im up in stage..i noe it sound weird.. but its true.....its like...i suddenly have an ambition in my life...i have something to dooo....being cooped up at home doesent mean i have to waste my life away...there are so may thing s for me to doscover and rediscover...my soul...my Religon..i bought these tapes recently..it was about the lives of the prophets(peace and blessings upon them all) and it is soo cool !!! Islam is not  about what newspapers say...it is not about terrorism or killing jews...its about patience and submission to God...it is a simple religion really.... u get freaks in every religion.....so u cant judge the religion from what you see on televison...it is beautiful really...every law has a good reason.....we may not know it but God knows whats best..... i thank God that i was born a muslim....i wouldnt wanna convert to anything even if you put a gun to my head..i put my trust in God anyway i gtg now.. ill keep u posted  '),(17975,3551081,'14,June,2004','2004-06-14 00:00:00','   she walks in beauty,like the night   of clodless climes and starry skies; and all that\'s best of dark and bright   meet in her aspect and her eyes: thus mellowed to that tender light   which heaven to gaudy day denies.  one shade the more,one ray the less,   had half impaired the nameless grace which waves in every raven trees,   or softly lightens o\'er her face; where thought serenely sweet express,   how pure,how dear their dwelling-place  and on that cheek,and o\'er that brow   so soft,so calm,yet eloquent, the smiles that win,the tints that glow\'   but tell the days in goodness spent, a mind at peace with all below,   a heart whose love is innocent!  '),(17976,3551081,'14,June,2004','2004-06-14 00:00:00','   the day i die dont be by my side for i will forever love you till my last breath  the day that i leave you dont let me go for deep inside me i know you are the on efor me  the day i fall in love with you please fall in love with me too for i know you need me untill your dying day  '),(17977,3551081,'13,June,2004','2004-06-13 00:00:00','   of all the souls i have come across yours is the most pure and clean untarnished by the oxides of life still new and still so shiny  slowly u discovered the world taking sstep by timid step seeing thing that you should not scarring your memory forever  i watched you turn yourself around staining your pure white mind if a drop of ink can ruin the milk a drop of this world will ruin your heart  the most beautiful are watched and admired a mistake as small as ash is seen as big as the earth a slip like and ants step,is felt like a giants stride  '),(17978,3551081,'13,June,2004','2004-06-13 00:00:00','   The reason why remains unknown why God created this emotion this despair and hurt is solely my own what good does it bring,this poison  this feeling inside,useless and futile it only brings hurt and despair a look on the bright side,how i have tried life among mortals will never be fair  take one last look within my soul the door lies though my morbid eyes i have fallen down this balck hole my lies hold no emotion  no good has come,ive been led astray my heart,my soul has fallen i never even felt anyway true love its called,will be lost forever  '),(17979,3551081,'13,June,2004','2004-06-13 00:00:00','   it was rachels bd yesterday.she was really depressed...im broke so i couldnt buy her a gift just yet....i sent her 2 e cards though...very cool...very nice.....happy birthday rachel!!!!this blogging thing is getting quite cool.although the only ones i read are mine and rachels...sometimes i probe into the lives of others but i really dont give a damn....i like reading the political ones..i like reading other pplz opinons about contraversial topics..it gives me ample warning about what ppl are gonna say next so i can think up a retaliation...i hate being wrong...i am very self opinionated...i like confrontation....especially when im right.well i gtg now...i willl be posting some poems i wrote up here tell me wat u think  '),(17980,3551081,'11,June,2004','2004-06-11 00:00:00','   today was an ok day....nothin new really..every day is kinda boring...same shit ..my life is so monotonous..i can barely wait till the day i get home. batoool called me today...man she has changed..ssshe seems so smug about her being in the uk and dissing the A level system thinkin that IB is soo damn cool.well sorry to say but A level is more recognised andd more cool.huh. shamsan had a football match today..with satwa..his team won of course...i spoke with ana on the phone..she told me never to leave shamsan.she said it is so hard to find somebody who loves you i am so scared to lose him..i love him so much...he is amazing...kind gentle generous understanding(most of the time)......words cannot express how i feel about him...i can only pray that both him and i get accepted into nus or ntu...i wannago home and be with rach..i have frens here...but its not the same..there is ana but she understands how important rachel is...she evem asks after her sometimes..... i was supposed to watch harry potter today but i came home at half past one last nite after an all girls party that i kinda passed.it was nooras graduation party..it was really fun...it was right on the boat at al boom and we danced on deck while the dhow sailed on the creek...there were all these hot guys on shore so we couldnt resist taking the oppurtunity to..ehm ehm..lool it was really cool..in both senses of the word.surprisingly it was cool up oin deck.considering it is stifling on land.i guess it is the night breeze. amazing...how people can not believe in GOD amazes me,.....how can they be so stupid?? i mean His creations surround us..what further proof do we need...how the sun floats in the sky?how mountains and icebergs have roots like teeth.....omg... anyway. am tired now rach is not online at the mo...she is basically the only reason why i come online...noone else is really worth it..i guess i miss talking in singlish to everyone else..i guess im totally homesick...  OMG...guess who added me to his msn...ali cobra..it has been a year i ignored at 1st but he added me at another addie...desperado lol...we chatted and he seems ok...not so much the jerk i knew 1 yr ago...he seems to have changed but past experiences tell me otherwise...ill keep my guard up. gtg am tired  '),(17981,3551081,'08,June,2004','2004-06-08 00:00:00','   http://www.hi5.com/i?l=NCO1T2FQ  thats my hi5 link.do wateva u want with it.kool\'  '),(17982,3551081,'07,June,2004','2004-06-07 00:00:00','   today was as uneventful as ever. last nite shamsan and i had a fite which lasted till like 3.30 in the morning.i cant recall what it was about exactly.but it was dumb....i feel so bad coz he has to work from 3 pm to 11 pm today and he is gonna be in a room controlling all police activity....he is a policeman..inspector...insspector sexy i call him lol welllllllll i gtgnow everyone is pesterin me keep ya posted  '),(17983,3551081,'06,June,2004','2004-06-06 00:00:00','   i just have so many emotions bursting inside me.i never knew that many emotions could exist at the same time. im happy sad amazed exited reluctant expectant all at the same time. i just cant wait to go home.i cant wait for uni. i cant wait to pick my outfit every day and have an excuse to buy clothes and makeup. i cant wait to make up for lost time with rach and my other frens. i cant wait to go down to malaysia n e time i want. i cant wait till i can afford a vespa.its gonne be pink..with a picture of the grim on it.Inshallah. that means god willing.Inshallah for everything. i hope everything will be ok for me.i just keep praying.it only just hit me this year when i started my a level course this year.....my life is just beginning. i can only pray for the best. even thogh i have been away from home for so long i have learned so much. it is satisfying to know that although unwillingly, i have learned alot the past 2.5 years.i met different types of people.dubai is a cultural melting pot.so many nationalities though mainly arab or indian subcontinent. its nice to know that i have one experience noone else will have.well tired now. doday is the start of the rest of my life  '),(17984,3551081,'05,June,2004','2004-06-05 00:00:00','   i must be so boring. all i ever talk about is going home.i miss it.its boring me to death here in dubai.thepeople are fake,lazy,boring the weather sux the shopping sux.the stuff they serve in food courts are a miserable excuse for food.........ugh the movies come out weeks too late........terrible i hate it here some things are ok.the guys are really hot but they are sooooooo narrow minded and they only think of one thing...***..u noe wat i mean welll enuf complaining for now  keep posted  '),(17985,3551081,'05,June,2004','2004-06-05 00:00:00','   well hello i decided to do this blog stuff coz ma best fren rachel in sg gave me the idea. she was online yesterday afta so long.man i miss her. i cant wait to go home. i miss her so much.i miss my home.but i love my bf though.i hope he can come home with me and further his education with me. i really want him to meet rach. im so homesick.i hate it here in dxb.its ok but i just miss home.i missed my teenhood.i should have been in sg not in dxb. i haf so many regrets keep ya posted  '),(17986,3551081,'04,August,2004','2004-08-04 00:00:00','   i have had this new blog for sometime now.....come visit me..... www.tabulas.com/~ainjiwa  '),(17987,3551081,'04,August,2004','2004-08-04 00:00:00','   i have had this blog for quite some time now.....feel free to visit me at my new adress www.tabulas.com/~ainjiwa  '),(17988,3719046,'17,August,2004','2004-08-17 00:00:00','   So what\'s up.  This is the first posting in the first blog created by the First Lady of Lame, me!  So listen up.  I hate all of you.  Except for maybe a few people.  I am so mad at everything.  Die, penguin, die.  Bloody fruitcake black suit-and-tie carrot fool.  Jenny\'s the name, nonsense is my game.  Go ahead, ask me a question.  I\'ll answer it with the lamest answer possible.  I hope you hate me now.  Everybody else does!  I don\'t know.  I need to sleep.  Goodnight.   '),(17989,3719046,'17,August,2004','2004-08-17 00:00:00','   So hey everybody, it\'s your favorite freak, me! Well, maybe I\'m not your favorite but we\'ll just pretend for now. As long as your in my friggin blog you\'re going to be my best friend. Ugh, I dunno, I\'m tired and also very bored. I need some friends, dammit!   Well, in my continuing research of my stupid, pointless love life, I have realized the following things:   1. I am incredibly&nbsp;repulsive and hideous.&nbsp;  2. Guys are afraid of me and don\'t like me, so they wouldnt even think about dating me   3. No girl would EVER have a reason to be jealous of me   4. Nobody wants to go out with a freaky, pale, extremely demented weirdo girl   5. Every guy I have ever liked never liked me back...at all   6. My sister is 12 and has had more boyfriends than I, a 15-year-old, have had   Well, that\'s all I can think of for now, maybe some more later if I\'m feeling especially depressed. Goodbye, dipshits.   '),(17990,3719046,'17,August,2004','2004-08-17 00:00:00','   I just want everybody who reads this (who is nobody so whatever) to know that I am going to take this lame ass blog down if people don\'t start commenting!  Now, I realize that this is a very non-threatening threat, however, I have nothing to back that up so I\'m just going to sit here on eating fruit pies and wallowing in my own self-pity while you all enjoy being the twig-like, bitchy I-can-get-any-guy-I-want bitches that you all are.  I hate you.  Get away from me.   '),(17991,3719046,'17,August,2004','2004-08-17 00:00:00','   Well you knew it would happen eventually...I\'ve got my very own club!  So far I\'ve only got 2 members including myself, but there\'s always room for more people.  There are only a few qualities a member of my club must have:  1.  You must at least hate your father, but both parents is also acceptable 2.  You must have an idea for what the club should officially be called 3.  You must be at least this tall to ride BATMAN 4.  You must be on my list of people I don\'t hate; I haven\'t yet made one in writing  Okay, so if you want to join my club, send me an email telling me.    Cfreak123@aol.com is my email address.  You blow.  Don\'t think being in my club makes you cool.  It totally does not.  But at least now I might wave to you every once in a while because I hate you less than I hate most people.  Good day to you.  No, screw that.  BAD day to you.   '),(17992,3719046,'17,August,2004','2004-08-17 00:00:00','   Hello people.&nbsp; I haven\'t posted in a little while so I\'m going to now.&nbsp; I mean, it\'s not like anybody even reads this, but I still like to pretend sometimes that people care.&nbsp; Anyway, I don\'t have anything to say because nothing really happened.&nbsp; Oh, and by the way--Saint Edward\'s Central&nbsp;Catholic High School blows ass.&nbsp; That is all.   '),(17993,3719046,'17,August,2004','2004-08-17 00:00:00','   Hey dorkwads, it\'s me.&nbsp; Anyway, goodbye.   '),(17994,3719046,'17,August,2004','2004-08-17 00:00:00','   Hey, so, yeah I have a new member in my club!&nbsp; But&nbsp;that\'s&nbsp;not really important&nbsp;because I almost forced her to join.&nbsp; Anyway, if you want to join, it\'s very easy to be in my club, all you gotta do is tell me you want to be in and you\'re in.&nbsp;   Send me a damn email:&nbsp;  urlLink Cfreak123@aol.com   If you\'re still skeptical about joining my FREE free club, here are some reasons why you should join.    It\'s free  It might be fun  ME!&nbsp; C\'mon, you like me, don\'t you?&nbsp; DON\'T YOU?.....  It\'s totally free  The club will get better with time and more people in it  You will be able to say you joined a club and didn\'t even have to do anything except email somebody   It\'s fuckin free!  You will break my fragile heart if you don\'t join...   So there you go.&nbsp; I know you want to join.&nbsp; C\'mon, maybe we\'ll all get free candy and little flags and a parade!&nbsp; Or maybe not.&nbsp; But so what.&nbsp; Well, later.  &nbsp;   '),(17995,3719046,'17,August,2004','2004-08-17 00:00:00','   Hi people. It\'s been a little while since I last posted. Like anyone cares. Like anybody is even reading this. Well what if I told you that I have a boyfriend? Ooh, what would you say? Nothing, because you don\'t know anything. Idiot. What a lame remark. Oh well, I am lame.   '),(17996,3719046,'17,August,2004','2004-08-17 00:00:00','   DUDE.  I have a boyfriend!  No lie, no joke, I am totally fuckin telling the truth.  And it rocks!  Finally, after all this shitty time of being alone...I am unalone!  Happy, happy, happy me.  But I won\'t tell you his name, because then the magic would be spoiled.  So you will never know!  Ha ha ha ha.  Just know that this dorkwad is UNAVAILABLE.  Sorry you idiots, you weren\'t fast enough.  AH HA HA HA!  I\'m nuts.  Deal with it.   '),(17997,3719046,'17,August,2004','2004-08-17 00:00:00','   Yeah so I\'m going to California on Saturday.  I won\'t be here for a WHOLE WEEK.  Whee, going to see grandma.  How great.  I know you\'ll miss me.  But it\'s so close to being school time again, that going is really gonna be shitful.  I mean I want to go, but I also want to just stay home and not do anything until school starts.  Oh yeah, school blows.  Anyway, I\'m going to miss you all.  WHAT THE FUCK????  I did not just say that.  I won\'t miss you at all, except for like 2 people.  Goodbye, assbags.  Ha ha, Ashley I stole your word.  BITCH.   '),(17998,3719046,'17,August,2004','2004-08-17 00:00:00','   So I\'m in California right now.  It\'s great.  Well, bye.   '),(17999,3389942,'30,May,2004','2004-05-30 00:00:00','   Alright.  29th, Morning, I got up at 6:30 to get ready for the programming compitition. I got to algonquin at 7:30; Registration was between 8:15 and 9:00. At about 8:00 Kevin appeared at the window of the Optra (I was in the back with my laptop out and my music pumped up). I got out, and hung around with Kevin until Matt and Roland, the rest of our team, showed up. I also got to meet / hang out with John for a bit.  We eventually got in, registered, went to our room and started programming. the first 2 hours went well, we were laughing, having a good time. We even had the Carebears music pumped up [we were in a room with 2 other teams, too ^^]. After the first 2 hours we hit a little snag, and I totally just drowned out. I couldn\'t figure out what the problem was, and i just kind of shut down; didn\'t care anymore. After enough kicking I eventually took another look at it, then it was the lunch break.  I thought about it during lunch, and fixed the problem as soon as we got back into the lab. blah, blah, blah, stuff happens, and our entry is judged.  --> It didn\'t run the judges Map properly, the units were fucking up, everything was going wrong.  after the judge had marked our program [we didn\'t do well...], I was sent back to the main room to wait for judging to finish.  I looked at the schedule to see when judging was going to start, it was in 45 minutes, so I grabbed my keys, hopped in my car and went to see if Sarah wanted to come to algonquin for the judging.   When I pulled up to where her house is, she was in the car with her mom and her sister, and we both got out of our respective cars and had a little chat. turns out she was going to the mall, so I hopped back into my vehicle and drove off into the sunset- er, I mean, back to algonquin. wow, I\'m tired.  alright, so I get back to algonquin, wait abit, final judging happens, Neither my team or Nick/John\'s team placed, but my team got Most Improved [A well deserved most improved, might I add].  After that I drove Kevin and Roland to Sarah\'s, got Sarah, and brought the whole crew to Compusmart to pick up some bawls. They had no _actual_ 24\'s left, so they made one with a 12 pack, 12 individuals and a big, brown cardboard box.  We went back to my house, and watched spiderman until everyone arrived.  the party was pretty sweet, I had a really good time. We watched spiderman, played music & socialised, watched flash movies, played videogames, and everyone seemed to be having a really good time.   Sarah went and sat in the corner at one point, but later on she ended up talking to Patrick for awhile, which was pretty cool. I don\'t think she would have enjoyed herself as much if he wasn\'t there.   Patrick is really cool, I mean, he\'s in grade 8, but I respect him more than most of my friends that are my age. I don\'t even see him as younger then me, it just doesn\'t register that way in my head. Quite the awesome guy.   I spent most of the time running around doing everything and talking to everyone... I had _lots_ of bawls.   Brent and Sam went on a walk at some point, and Sarah kept suggesting the idea that they were in my room. :s. that wouldn\'t be cool. I would have to sleep on the floor if that happend.  Roland and Kevin were the two huge nerds that they are; Roland said nothing and Kevin spent the whole time on the computer. Alan just sat there and played his guitar, but I\'m sure he has other things on his mind anyway [He just started going out with Rachel ^^, thats soo cool, i\'m happy for him, Rachels cool (alot cooler then his Grade 10 female friend {Who was apparently mad when he got a gf, even though she said he should date other people   } ) ].  and... that was that.  then, the next day, my birthday, I got Kill Bill, The Billy Talent CD, and a trampoline.  And worked on my anthro summative all day.  fun stuff,,, guh.  i\'m going to die tomorrow from sleep deprivation, I\'m still not done my anthro summative, I want to see Sarah as soon as possible (i always want to see Sarah as soon as possible, though, so thats not really somthing new), and I probably have to work on tuesday (just to spite me, so I can\'t invite Sarah over and I have to wait until wednesday).  Night, whoever reads my blog, --Grant   '),(18000,3389942,'28,May,2004','2004-05-28 00:00:00','   Anyway,  Alright. I have my Anthro Summative, with a drop dead due date of Tuesday [I\'m certain I can get it done by then], My physics summative is due at some point next week, the programming compition is... tomorrow, along with my party.  Fun stuff.  I feel bad in antrho, because I didn\'t do my homework for chapters 3,4,5,6, or 7... so I need to get those done. I figure I\'ll work on them after the summative is over with.  I\'m in computer Engineering now... I drove to school today, in the Optra, so I have my incubus mix CD, I brought it in, and its playing over the speakers in the computer engineering room right now :).  \"Whatever Tomorrow Brings, i\'ll be there\",  --Drive, by Incubus  Just heard that. I guess it applies right now, I can get all this stuff done, I can win that compition, I can get through the rest of this school year.  Out of everything going on tomorrow, I\'m really just looking forward to seeing Sarah the most. I love being around her, being with her, randomly talking about absolutly nothing, or having a conversation about somthing serious, its awesome.  \"Its all been said, with the exeption for the right part... when will we be new skin?\", --New Skin, by Incubus  What else is going on... Oh, Wilfrid Laurier University offers a 5 year program in which you get both a Bachelors in Computer Science AND a Bachelors in Buisness, which would be _exactly_ what i\'m lookinf for.  I\'m still going with the idea that I want to take over the dealership, and in order to do so, I would have to get my degree (4 or 5 years), then spend 5 to 7 years working at the dealership before I could apply to become an official Licensed GM dealer.  Anyway, --Grant.   '),(18001,3389942,'27,May,2004','2004-05-27 00:00:00','   Alright.  Yesterday: I had Sarah over. That was fun. My dad still can\'t figure out what the scratches on the wall are from... *cough*.  Today: Nothing. Atleast, nothing yet.   '),(18002,3389942,'26,May,2004','2004-05-26 00:00:00','   ---------------------------------------------------------------- I had this dream, where I relished the fray, and the screaming filled my head all day...   It was as though I\'d been spit here,  settled in,  into the pocket,  of a lighthouse on some rocky socket,  off the coast of France,  dear.  One afternoon,  four thousand men,  died in the water,  here..   and five hundred more were thrashing madly,  as parasites might,  in your blood.   Now I was in a lifeboat,  designed for ten,  and ten only, settled in.  anything that systematic  would get you hated.  It\'s not a deal nor a test nor a love of something fated...   The selection was quick,  the crew picked and those left in the water, got kicked off our pantleg...  and we headed for home.  Then the dream ends,  when the phone rings, you doing alright?  he said it\'s out there,  most days and nights,  but only a fool would complain.  Anyway,  Susan,  if you like,  our conversation, is as faint as a sound in my  memory,  as those fingernails scratching on my hull. ------------------------------------------------------------------   '),(18003,3389942,'25,May,2004','2004-05-25 00:00:00','   Guh,  My job isn\'t that bad, but sometimes it can suck.  It all depends.  Anyway, I have Sarah coming over tomorrow, so its all good.  --Grant.   '),(18004,3389942,'19,May,2004','2004-05-19 00:00:00','   Alright,  I was in Physics class, and after the teacher finished the lesson, I sat back and started thinking.  I got this... somthing hit me. Not physically, no, an idea, or rahter, a path...  I don\'t know where it came from. I\'ve wanted to be a programmer since I was 8 years old; i\'ve been programming since I was 8 years old... but... While thinking about what I like to do, the aspects of my Job I enjoy, what I like to do while programming... I think I was _made_ to sell cars.  My brother Rob went through 4 years of Electrical Engineering at Queens University, and he now works at Ottawa Hydro; sure, he does random electrical engineering things, but want to know what the favourite part of his job is? Organizing meetings, networking with people, organzing and assigning people to do stuff; and he\'s good at it. enough so, thats pretty much what he does now... Thats what he would be doing as a car salesmen.  I really like working with computers, but all the mystery and appeal of working with them is gone; coding seems dry to me... but at my Job at the independant, i\'m constantly talking to people; people who come through my cash usually come through again. I know a university professer, a women who owns her own construction company, a women who wants me to tutor her son programming... all these people who know me; even though I\'m just a cashier at a grocery store.  I think I would really enjoy being a car salesmen. I think its what i\'m built to do. And on top of that, the pay would be amazing, and I\'d get tons of time off to persue other intrests [such as programming].  I scheduled a guidence appointment, and i\'m going to talk to my dad about it, but yeah... nuts.   '),(18005,3389942,'19,May,2004','2004-05-19 00:00:00','   Alright,  So I\'ve got the programming competition coming up on the 29th. That should be intresting.  But what i\'m more looking forward to right now is the party i\'m having after the compition. I\'m having about 18 people over, its going to be pretty awesome.  anyway, --Grant   '),(18006,3389942,'19,May,2004','2004-05-19 00:00:00','   Sweet.  New Blog.  Cool, --Grant   '),(18007,3389942,'24,June,2004','2004-06-24 00:00:00','   http://frog102.system-7.net   '),(18008,3389942,'01,June,2004','2004-06-01 00:00:00','   Alright,  Yesterday was monday. The drop dead due date for my anthro sumamtive was today. I didn\'t quite finish it... I just handed in what I finished on sunday.  On a topic totally unrelated to me not finishing my anthro summative, I had Sarah over yesterday. ^^\'. It was fun, we watched tons of south park [all season 8].  ummm... My physics summative is due Friday, my Programming summative is due Friday, my Computer Engineering summative is... tomorrow, its a test.  umm...  I work tonight, for 3 hours [yay for $21!], I shouldn\'t complain, I need the money.  anyway, --Grant   '),(18009,3956613,'28,July,2004','2004-07-28 00:00:00','   Round seven is in the books:  Houston - 2b Frank White New York (PL) - P Randy Jones   Philadelphia - P Ron Reed Chicago (PL) - OF Oscar Gamble Pittsburgh - 1b Richie Hebner Kansas City - P Bill Bonham Atlanta - 2b Manny Trillo Texas - OF Mitchell Page Hollywood - C Johnny Bench   Cleveland - P Geoff Zahn Chicago (NL) - C Bob Boone Anaheim - 2b Bump Wills Montreal - 2b Julio Cruz Boston - P Moose Haas Cincinnati - P Bob Shirley Baltimore - P Jim Colborn Wisconsin - P Don Aase Toronto - C Gene Tenace Queens (NL) - 3B Sal Bando Michigan - P Jim Clancy   San Francisco - C Thurmon Munson Minneapolis - SS Ivan DeJesus St. Louis - OF Bruce Bochte Seattle - OF Bobby Bonds '),(18010,3956613,'27,July,2004','2004-07-27 00:00:00','   Round six of the APBL draft is finished:  Seattle - P Paul Hartzell St. Louis - C Brian Downing   Minneapolis - P Jim Slaton San Francisco - P John Denny Michigan - OF - Carl Yastrezmski  Queens (NL) - SS Rick Burleson Toronto - P Glenn Abbott Wisconsin - P Francisco Barrios Baltimore - OF Jeff Burroughs   Cincinnati - 1b Dan Driessen Boston - OF Cesar Cedeno Montreal - 3b Doug DeCinces Anaheim - P Tug McGraw Chicago (NL) - P Bob Stanley Cleveland - 1b Jason Thompson Hollywood - P Gary Lavelle Texas - OF Ben Oglivie Atlanta - OF Sixto Lezcano Kansas City - OF Larry Hisle   Pittsburgh - P Gene Garber Chicago (PL) - 1b Bob Watson Philadelphia - P Rollie Fingers New York (PL) - C Jim Sundberg Houston - P Mike Norris '),(18011,3956613,'26,July,2004','2004-07-26 00:00:00','   Round five is finished:  Houston - OF George Hendrick New York (PL) - P Rudy May Philadelphia - OF Gary Matthews Chicago (PL) - P Ed Figureoa  Pittsburgh - 3b Bill Madlock  .  Kansas City - P Mike Krukow Atlanta - P Lary Sorensen Texas - P Eric Rasmussen Hollywood - P Jim Bibby Cleveland - P Jon Matlack  Chicago (NL) - P Rick Honeycutt  .  Anaheim - OF Lyman Bostock Montreal - 1b Andy Thornton Boston - OF Don Baylor Cincinnati - P Willie Hernandez Baltimore - P Nino Espinoza Wisconsin - P Mike Torrez Toronto - P Milt Wilcox Queens (NL) - P Ken Kravec Michigan - 1b Chris Chambliss  .  San Francisco - P Steve Stone Minneapolis - P Floyd Bannister St. Louis - OF Steve Kemp Seattle - 3b Graig Nettles '),(18012,3956613,'25,July,2004','2004-07-25 00:00:00','   The fourth round of the APBL inaugural draft is complete:  Seattle - OF Dusty Baker St. Louis - P Rick Rhoden Minneapolis - 2B Davey Lopes    San Francisco - OF Ron LeFlore  Michigan - P Paul Splittorff Queens (NL) - OF Amos Otis Toronto - 1B Mike Hargrove Wisconsin - SS Dave Concepcion Baltimore - P Doyle Alexander Cincinnati - P Kent Tekulve Boston - P Larry Christenson Montreal - OF Jose Cruz    Anaheim - P Dave Rozema Chicago (NL) - SS Roy Smalley Cleveland - OF Al Oliver Hollywood - 3b Ron Cey Texas - P Joe Sambito Atlanta - P Ken Forsch Kansas City - 3b Toby Harrah Pittsburgh - OF Ken Griffey  Chicago (PL) - OF Reggie Smith Philadelphia - C Darrell Porter New York (PL) - P Rick Langford Houston - 1b Bill Buckner '),(18013,3956613,'25,July,2004','2004-07-25 00:00:00','   Round three went off succesfully yesterday.  Here are the picks:  Houston - P Jerry Reuss New York (PL) - 1b Cecil Cooper Philadelphia - P Mike Flanagan   Chicago (PL) - P Scott McGregor Pittsburgh - P Bob Knepper Kansas City - P Joaquin Andujar Atlanta - P Larry Gura Texas - P Gaylord Perry Hollywood - OF Jack Clark   Cleveland - P Mike Caldwell Chicago (NL) - P Charlie Hough Anaheim - P Dave Goltz Montreal - P Jerry Koosman Boston - P Ed Halicki Cincinnati - OF Fred Lynn Baltimore - P Wayne Galrand Wisconsin - 1b Steve Garvey Toronto - OF Greg Luzinski   Queens (NL) - P Craig Swan Michigan - SS Garry Templeton San Francisco - OF Reggie Jackson Minneapolis - OF Chet Lemon St. Louis - OF Darrell Evans Seattle - P Pete Vuckovich '),(18014,3956613,'24,July,2004','2004-07-24 00:00:00','   http://www.PhiladelphiaFreedom.blogspot.com '),(18015,3956613,'23,July,2004','2004-07-23 00:00:00','   Round two of the Centennial League draft is complete, here\'s how things went.   Round 2  Seattle - P Burt Hooton St. Louis - P Joe Niekro Minneapolis - P Dennis Leonard  .  San Francisco - P Jim Palmer Michigan - P Jack Morris Queens (NL) - P Tom Seaver Toronto - P John Candelaria Wisconsin - P Phil Niekro  .  Baltimore - C Carlton Fisk Cincinnati - P Frank Tanana Boston - P Bruce Sutter Montreal - OF Andre Dawson Anaheim - P Fergie Jenkins Chicago (NL) - OF Dwight Evans Cleveland - C Dale Murphy   Hollywood - P Tommy John Texas - P Rick Reuschel Atlanta - 3b Buddy Bell Kansas City - SS Bobby Grich Pittsburgh - 2b Joe Morgan  Chicago (PL)- C Ted Simmons Philadelphia - P Bob Forsch New York (PL) - P Vida Blue Houston - P Mario Soto '),(18016,3956613,'22,July,2004','2004-07-22 00:00:00','   urlLink http://anaheimevolution.blogspot.com/ '),(18017,3956613,'22,July,2004','2004-07-22 00:00:00','   Day two of the APBL draft began with the Seattle Seamen wrapping up Round 1 by selecting 1B Rod Carew with the #24 pick of the draft.  They follwed that by selecting P Burt Hooton with the first pick of the second round.  From here on in, the  A-P Bulletin  will be posting round-by-round updates as opposed to pick-by-pick.  However, owners are encouraged to fill it up with anything else they think will be of interest. '),(18018,3956613,'22,July,2004','2004-07-22 00:00:00','   The Toronto Towers have started a team blog for announcements and press releases at   urlLink http://www.torontotowers.blogspot.com/ '),(18019,3956613,'22,July,2004','2004-07-22 00:00:00','   Draft day is finally here, making picks 1-9 official.  The first announced pick of the day came from the Cleveland Rock, who used their #10 pick for young starter Ron Guidry.     The Chicago NL team selected 2B Willie Randolph     The Anaheim Evolution used the #12 pick for 3B Pete Rose     At #13, Montreal selects P Don Sutton     At #14, Boston selects 1B Eddie Murray     Cincinnati selects Dennis Martinez at #15.     Baltimore took OF Dave Winfield as the #16 pick.     The Wisconsin Badgers select OF Dave Parker at #17.     The Toronto Towers are using the #18 pick on P Bert Blyleven.     At #19, the Queens Narks took P J.R. Richard.     With the 20th pick, Michigan selected Rich \"Goose\" Gossage.     San Francisco selected 1B Keith Hernandez at 21.     At #22, the Minneapolis Millers picked up C Gary Carter     With the 23rd pick, St. Louis picked P Steve Rogers '),(18020,3956613,'21,July,2004','2004-07-21 00:00:00','   Still a full day ahead of the APBA inaugural draft\'s official start on Thursday, the early picks continued as teams announced their selections.  Philadelphia selected 3B Mike Schmidt with the #3 pick.     The Chicago Fire selected P Steve \"Lefty\" Carlton at #4.     Pittsburgh used their #5 pick for P Dennis Eckersley     At #6, the Kansas City Bison chose OF Jim Rice.     The #7 pick was SS Robin Yount, selected by Atlanta.     Texas selected OF George Foster at #8     And the last pick of the day was announced by the Hollywood Wassailers, who took OF Ken Singleton with the #9 pick. '),(18021,3956613,'20,July,2004','2004-07-20 00:00:00','   With the first pick of the first ever APBL inaugural draft, the Houston Horses have selected Texas\' own Nolan Ryan!       Later in the day, the New York Knights announced that they would be using their second pick on third baseman George Brett. '),(18022,3956613,'16,July,2004','2004-07-16 00:00:00','   As announced today by the commissioners office:   1. Houston  2. New York (PL)  3. Philadelphia  4. Chicago (PL)  5. Pittsburgh  6. Kansas City  7. Atlanta  8. Texas   9. Hollywood 10. Cleveland 11. Chicago (NL) 12. Anaheim 13. Montreal 14. Boston 15. Cincinnati 16. Baltimore 17. Wisconsin 18. Toronto 19. New York (NL) 20. Michigan 21. San Francisco 22. Minneapolis 23. St. Louis 24. Seattle '),(18023,3956613,'16,July,2004','2004-07-16 00:00:00','   This will be the official newsletter of the Atlantic-Pacific Baseball League (APBL) and will be set up so that any and all owners can get a ( urlLink free )&nbsp;blogger account and post on to it themselves.&nbsp; We\'ll use it for news, notes,&nbsp; rumors, highlights, trade talk, draft updates, pennant race watching, whatever.&nbsp; Use it to talk amongst ourselves as well as to whomever might stumble upon it.&nbsp; Everything will be archived, so two, three, even ten or more seasons in we can go back to see what was going on in 1977.&nbsp; Use the comments section below each post to discuss it with other owners.  Have fun! '),(18024,3956613,'07,August,2004','2004-08-07 00:00:00','   Round 12   Seattle (jmbreslin) - SS - Bert Campaneris St. Louis (deanod) - P - Bob Lacey Minneapolis (peajay) - P - Jesse Jefferson San Francisco (Iceman67) - P - Doug Bird Michigan (occsid) - OF - Lou Brock Queens (NL) (shaybee) - OF - Lou Pinella Toronto (ldmfan) - OF - Gary Thomasson Wisconsin (newarkwilder) - P - Bob Owchinko Baltimore (Ginandjuice) - 2B - Art Howe Cincinnati (smeric) - P - Catfish Hunter Boston (seaphil) - C - Charlie Moore Montreal (Oli35) - OF - Rick Manning Anaheim (scthallfan) - P - Jerry Augustine Chicago (NL) (discoalex) -  1B - Willie Aikens Cleveland (goldman23) - 3B - Butch Hobson Hollywood (richabbs) - P - Bill Lee Texas (brucel) - SS - Tim Foli Atlanta (eddiedzen) - P - Tom Burgmeier Kansas City (coldfeet) - P - Tom Hume Pittsburgh (jeff4noles) - 1B - Rico Carty Chicago (PL) (jmarino2) - 2B - Tito Fuentes Philadelphia (eagles2231) - P - Eduardo Rodriguez New York (PL) (Robin4) - SS Bucky Dent Houston (gonoles777) - SS Rance Mulliniks  Round 13   Houston (gonoles777) - P Jack Billingham New York (PL) (Robin4) - P Bill Castro Philadelphia (eagles2231) - 1b George Scott Chicago (PL) (jmarino2) - P Pedro Borbon Pittsburgh (jeff4noles) - P Paul Moskau Kansas City (coldfeet) - 2b Ted Sizemore Atlanta (eddiedzen) - P Jim Kaat Texas (brucel) - SS Frank Taveras (we have proxy for rest of draft) Hollywood (richabbs) - 2b Duane Kuiper Cleveland (goldman23) - 2b Rodney Scott Chicago (NL) (discoalex) - P - Don Gullett Anaheim (scthallfan) - OF Willie Horton Montreal (Oli35) - C Biff Pocoroba Boston (seaphil) - RP Warren Brusstar  Cincinnati (smeric) - P Pat Zachry Baltimore (Ginandjuice) - OF Jim Norris Wisconsin (newarkwilder) - P Marty Pattin  Toronto (ldmfan) - 2B Rob Andrews  Queens (NL) (shaybee) - C Jim Essian Michigan (occsid) - OF - Cesar Geronimo San Francisco (Iceman67) - OF - Greg Gross Minneapolis (peajay) - OF - Ken Landreaux St. Louis (deanod) - 1B - Lamar Johnson Seattle (jmbreslin) - P - Mark Fidrych  Round 14   Seattle (jmbreslin) - OF - Barry Bonnell St. Louis (deanod) - P - John Urrea Minneapolis (peajay) - OF - Otto Velez San Francisco (Iceman67) - P - Dave Heaverlo Michigan (occsid) - P - Randy Lerch Queens (NL) (shaybee) - P - Wayne Twitchell Toronto (ldmfan) - P - Jim Lonborg Wisconsin (newarkwilder) - 2B - Mike Tyson Baltimore (Ginandjuice) - P - Clay Carroll Cincinnati (smeric) - C - Duffy Dyer Boston (seaphil) - SS - Craig Reynolds Montreal (Oli35) - P - Grant Jackson Anaheim (scthallfan) - C - Ed Ott Chicago (NL) (discoalex) - 3B - Mike Cubbage Cleveland (goldman23) - 3B - Jerry Royster Hollywood (richabbs) - 1B - John Milner Texas (brucel) - P - Stan Bahnsen Atlanta (eddiedzen) - OF - Lee Lacy Kansas City (coldfeet) - OF - Darrel Thomas Pittsburgh (jeff4noles) - C - Steve Yeager Chicago (PL) (jmarino2) - 3B - Bill Stein Philadelphia (eagles2231) - P - Joe Coleman New York (PL) (Robin4) - OF - Jerry Morales Houston (gonoles777) - P - Rick Camp  Round 15  Houston (gonoles777) - 1B - Tom Paciorek New York (PL) (Robin4) - 1B - Dan Meyer Philadelphia (eagles2231) - OF - Leroy Stanton Chicago (PL) (jmarino2) - P - Jackie Brown Pittsburgh (jeff4noles) - SS - Bill Almon Kansas City (coldfeet) - P - Don Stanhouse Atlanta (eddiedzen) - OF - Billy North Texas (brucel) - OF - Terry Whitfield Hollywood (richabbs) - P - Bob McClure Cleveland (goldman23) - C - Rick Dempsey Chicago (NL) (discoalex) - P - John Montague Anaheim (scthallfan) - P - Lerrin LaGrow Montreal (Oli35) - OF - Rick Miller Boston (seaphil) - 2B - Rod Gilbreath Cincinnati (smeric) - P - Al Hrabosky Baltimore (Ginandjuice) - P - Tom Johnson (HSP) Wisconsin (newarkwilder) - C - Bob Stinson Toronto (ldmfan) - OF - Al Woods Queens (NL) (shaybee) - P - John Hiller Michigan (occsid) - OF - Von Joshua San Francisco (Iceman67) - P - Steve Mingori Minneapolis (peajay) - 1B - Willie Montenez St. Louis (deanod) - 1B - Ron Fairly Seattle (jmbreslin) - OF - Joe Rudi  Round 16   Seattle (jmbreslin) - P - Andy Hassler St. Louis (deanod) - 3B - Ray Knight Minneapolis (peajay) - P - Mike Marshall San Francisco (Iceman67) - OF - Tony Scott Michigan (occsid) - C - Rick Cerone Queens (NL) (shaybee) - P - Sid Monge Toronto (ldmfan) - P - Don Hood Wisconsin (newarkwilder) - P - Dyar Miller Baltimore (Ginandjuice) - SS - Mark Belanger Cincinnati (smeric) - C - Tim McCarver Boston (seaphil) - P Joe Kerrigan  Montreal (Oli35) - SS - Larvell Blanks Anaheim (scthallfan) - OF - Bernie Carbo Chicago (NL) (discoalex) - P - Ken Brett Cleveland (goldman23) - P Dan Spillner  Hollywood (richabbs) - OF Johnny Grubb Texas (brucel) - 1B - Mike Ivie Atlanta (eddiedzen) - P Tom Griffin Kansas City (coldfeet) - C - John Wockenfuss Pittsburgh (jeff4noles) - P - Pablo Torrealba Chicago (PL) (jmarino2) - OF - Cliff Johnson Philadelphia (eagles2231) - P - Adrian Devine New York (PL) (Robin4) - P - Eddie Solomon Houston (gonoles777) - P - Woodie Fryman '),(18025,3956613,'02,August,2004','2004-08-02 00:00:00','   Summaries for Rounds 8-11:  Round 8   Seattle - 2b Don Money St. Louis - P Dennis Lamp Minneapolis - P Fred Norman San Francisco - SS Larry Bowa Michigan - P Ross Grimsley Queens (NL) - OF Gary Maddox Toronto - SS Chris Speier  Wisconsin - 3b Phil Garner Baltimore - P Jerry Garvin  Cincinnati - OF Lee Mazzilli Boston - OF Hal McRae Montreal - P Luis Tiant Anaheim - SS Bill Russell Chicago (NL) - OF Dave Kingman Cleveland - OF Gene Richards Hollywood - P Dick Ruthven Texas - C John Stearns Atlanta - OF Terry Puhl Kansas City - OF Ruppert Jones Pittsburgh - P Enrique Romo Chicago (PL) - P Elias Sosa Philadelphia - OF Al Bumbry New York (PL) - OF Dave Collins Houston - P Len Barker  Round 9  Houston (gonoles777) - C Butch Wynegar New York (PL) (Robin4) - 2B Rich Dauer Philadelphia (eagles2231) - P Mark Lemongello Chicago (PL) (jmarino2) - OF Bobby Murcer  Pittsburgh (jeff4noles) - P Jim Barr Kansas City (coldfeet) - P Rick Waits Atlanta (eddiedzen) - P Bruce Kison Texas (brucel) - 3b Steve Ontiveros Hollywood (richabbs) - OF Richie Zisk Cleveland (goldman23) - SP John Montefusco  Chicago (NL) (discoalex) - P Ray Burris  Anaheim (scthallfan) - 1b Tony Perez Montreal (Oli35) - P Dick Tidrow Boston (seaphil) - P Doug Rau Cincinnati (smeric) - 3b Len Randle Baltimore (Ginandjuice) - P Dave Lemanczyk (HSP) Wisconsin (newarkwilder) - OF Al Cowens Toronto (ldmfan) - 3B Wayne Gross  Queens (NL) (shaybee) - OF Mickey Rivers Michigan (occsid) - 3B Enos Cabell San Francisco (Iceman67) - P - Jim Rooker Minneapolis (peajay) - 3B - Larry Parrish St. Louis (deanod) - 2B - Jerry Remy Seattle (jmbreslin) - C - Milt May  Round 10  Seattle (jmbreslin) - P - Reggie Cleveland St. Louis (deanod) - P - Doc Medich Minneapolis (peajay) - OF - Claudell Washington San Francisco (Iceman67) - 3b Ken Reitz  Michigan (occsid) - 2B - Jorge Orta Queens (NL) (shaybee) - 3B - Roy Howell Toronto (ldmfan) - OF Roy White  Wisconsin (newarkwilder) - 1B Rusty Staub  Baltimore (Ginandjuice) - 3b Eric Soderholm Cincinnati (smeric) - OF Ellis Valentine Boston (seaphil) - 3B Dave Chalk Montreal (Oli35) - P Tippy Martinez Anaheim (scthallfan) - OF - Ralph Garr Chicago (NL) (discoalex) -  P - Paul Thormodsgard Cleveland (goldman23) - OF - Warren Cromartie  Hollywood (richabbs) - 1B - Willie Stargell Texas (brucel) - P - Doc Ellis Atlanta (eddiedzen) - OF - Jerry Mumphrey Kansas City (coldfeet) - P - Mark Littell Pittsburgh (jeff4noles) - P - Fernando Arroyo Chicago (PL) (jmarino2) - OF - Bob Bailor Philadelphia (eagles2231) - 1B - Willie McCovey New York (PL) (Robin4) - OF - Omar Moreno Houston (gonoles777) - P - Tom Underwood  Round 11   Houston (gonoles777) - C Alan Ashby New York (PL) (Robin4) - P - Vern Ruhle Philadelphia (eagles2231) -  SS - Alan Bannister Chicago (PL) (jmarino2) -  P - Sparky Lyle Pittsburgh (jeff4noles) - SS - Freddie Patek Kansas City (coldfeet) - 1B - John Mayberry Atlanta (eddiedzen) - P - Matt Keough Texas (brucel) - P - Skip Lockwood Hollywood (richabbs) - P - Rick Wise Cleveland (goldman23) - P - Dave LaRoche Chicago (NL) (discoalex) -  OF Dan Ford Anaheim (scthallfan) - 2B - Rennie Stennett Montreal (Oli35) - OF - Steve Henderson Boston (seaphil) - OF - Juan Beniquez Cincinnati (smeric) - 2B - Dave Cash Baltimore (Ginandjuice) - 1B - Bill Robinson Wisconsin (newarkwilder) - OF - Bake McBride Toronto (ldmfan) - P - Steve Foucault Queens (NL) (shaybee) - P - Jim Kern Michigan (occsid) - C - Manny Sanguillen San Francisco (Iceman67) - OF - Tony Armas Minneapolis (peajay) - P - Bill Campbell St. Louis (deanod) - C - Joe Ferguson Seattle (jmbreslin) - P - Buddy Schultz '),(18026,3956613,'31,July,2004','2004-07-31 00:00:00','    The Toronto Towers have selected outfielder Roy White in the tenth round of the inaugural APBL draft.   The 33-year-old switch-hitter from Los Angeles, has played for the New York Yankees for the last twelve years.   In his major league career to date Speier has 1527 hits, 135 home runs, 636 RBI and a career AVG of .274.   White will wear uniform #6 with the Towers.   '),(18027,3956613,'29,July,2004','2004-07-29 00:00:00','    The Toronto Towers have selected third baseman Wayne Gross in the ninth round of the inaugural APBL draft.   The 25-year-old left-hander from Riverside, California, made his major league debut in 1976 with the Oakland Athletics.   He only played in 10 games for the A\'s in 1976.   Gross will wear uniform #10 with the Towers.   '),(18028,3956613,'29,July,2004','2004-07-29 00:00:00','    The Toronto Towers have selected shortstop Chris Speier in the eighth round of the inaugural APBL draft.   The 26-year-old right-hander from Alameda, California, has played for the San Francisco Giants for the last six years.   In his major league career to date Speier has 796 hits, 56 home runs, 350 RBI and a career AVG of .250.   Speier will wear uniform #35 with the Towers.   '),(18029,3956613,'27,July,2004','2004-07-27 00:00:00','    The Toronto Towers have selected catcher Gene Tenace in the seventh round of the inaugural APBL draft.   The 30-year-old right-hander from Russellton, Pennsylvania, has played for the Oakland Athletics for the last eight years.   In his major league career to date Tenace has 603 hits, 121 home runs, 389 RBI and a career AVG of .245.   Tenace will wear uniform #18 with the Towers.   '),(18030,3956613,'27,July,2004','2004-07-27 00:00:00','    The Toronto Towers have selected&nbsp;pitcher&nbsp;Glenn Abbott&nbsp;in the&nbsp;sixth round of the inaugural APBL draft.  The 26-year-old&nbsp;right-hander from Little Rock, has played for the&nbsp;Oakland Athletics for the last four years.  In his&nbsp;major league career to date,&nbsp;Abbott has&nbsp;pitched in&nbsp;73 games, starting&nbsp;45, has a record of 13-16 and a career ERA of 4.08.  Abbott&nbsp;will wear uniform #37 with the Towers.   '),(18031,3956613,'25,July,2004','2004-07-25 00:00:00','    The Toronto Towers have selected&nbsp;pitcher&nbsp;Milt Wilcox&nbsp;in the&nbsp;fifth round of the inaugural APBL draft.  The 26-year-old&nbsp;right-hander from Honolulu, has played for the&nbsp;Cincinnati Reds, Cleveland Indians and Chicago Cubs since 1970. Wilcox did not played major league ball in 1976 and welcomes a second chance North of the border.  In his&nbsp;major league career to date,&nbsp;Wilcox has&nbsp;pitched in 147 games, starting&nbsp;53, has a record of 22-30 and a career ERA of 4.43.  Wilcox&nbsp;will wear uniform #43 with the Towers.   '),(18032,3956613,'25,July,2004','2004-07-25 00:00:00','    The Toronto Towers have selected&nbsp;first baseman&nbsp;Mike Hargrove&nbsp;in the&nbsp;fourth round of the inaugural APBL draft.   The 27-year-old&nbsp;left-hander from Perryton, Texas, has played for the&nbsp;Texas Rangers&nbsp;for the last three years.   In his&nbsp;major league career to date,&nbsp;Hargrove has&nbsp;446 hits, 22 home&nbsp;runs,&nbsp;186 RBI&nbsp;and a career&nbsp;AVG of .302.  Hargrove&nbsp;will wear uniform #21 with the Towers.   '),(18033,3956613,'23,July,2004','2004-07-23 00:00:00','    The Toronto Towers have selected&nbsp;outfielder&nbsp;Greg&nbsp;Luzinski&nbsp;in the&nbsp;third round of the inaugural APBL draft.   The 26-year-old&nbsp;left-fielder from Chicago, has played for the Philadelphia Phillies for the last&nbsp;seven years.   In his&nbsp;major league career to date,&nbsp;Luzinski has&nbsp;787 hits, 112 home&nbsp;runs, 443 RBI&nbsp;and a career&nbsp;AVG of .290.  Luzinski&nbsp;will wear uniform #19 with the Towers.   '),(18034,3956613,'23,July,2004','2004-07-23 00:00:00','    The Toronto Towers have selected pitcher&nbsp;John Candelaria&nbsp;in the&nbsp;second round of the inaugural APBL draft.  The 23-year-old&nbsp;lefty from New York City, has played for the Pittsburgh Pirates for the last two years.  In his short&nbsp;major league career to date,&nbsp;Candelaria has 49 starts with a record of 24-13 and a career ERA of 3.01.  Candelaria&nbsp;will wear uniform #45 with the Towers.   '),(18035,3956613,'19,July,2004','2004-07-19 00:00:00','    The Toronto Towers are pleased to announce that they have selected pitcher Bert Blyleven in the first round of the inaugural APBL draft.   The Dutch-born right-hander, who will turn 26 just before the season starts, comes to Toronto from the Texas Rangers, to whom he was traded in June 1976 after spending 6 1/2 years with the Minnesota Twins.   In his major league career to date, Blyleven has 249 starts with a record of 108-101 and a career ERA of 2.80.   Blyleven will wear uniform #28 with the Towers.   '),(18036,3956613,'19,July,2004','2004-07-19 00:00:00','       '),(18037,3956613,'19,July,2004','2004-07-19 00:00:00','    The Toronto Towers announced today that the main colour scheme for all their logos, unforms and stadium decorations will be red and white (the national colours of Canada) and silver (to commemorate the silver jubilee of Queen Elizabeth II).   '),(18038,3956613,'11,August,2004','2004-08-11 00:00:00','    The Toronto Towers have selected pitcher Mike Garman in the twentieth round of the inaugural APBL draft.  The 27-year-old right-hander from Caldwell, Idaho has played for the Boston Red Sox, St. Louis Cardinals and Chicago Cubs for the last seven years.  In his major league career to date Knapp has pitched in 197 games, starting 8, has a record of 14-16 with 17 saves and a career ERA of 3.62.  Garman will wear uniform #44 with the Towers.   '),(18039,3956613,'09,August,2004','2004-08-09 00:00:00','    The Toronto Towers have selected shortstop Pat Rockett in the nineteenth round of the inaugural APBL draft.   The 22-year-old right-hander from San Antonio played briefly for the Atlanta Braves last year.   Rockett will wear uniform #34 with the Towers.   '),(18040,3956613,'09,August,2004','2004-08-09 00:00:00','    The Toronto Towers have selected third baseman Wayne Garrett in the eighteenth round of the inaugural APBL draft.   The 29-year-old left-hander from Brooksville, Florida has played for the New York Mets and briefly for the Montreal Expos in the last eight years.   In his major league career to date Garrett has 710 hits, 57 home runs, 306 RBI and a career AVG of .237.   Garrett\'s preferred uniform #11 is taken so he will wear uniform #9 with the Towers.   '),(18041,3956613,'06,August,2004','2004-08-06 00:00:00','    The Toronto Towers have selected pitcher Chris Knapp in the seventeenth round of the inaugural APBL draft.  The 23-year-old right-hander from Cherry Point, North Carolina has played for the Chicago White Sox for the last two years.  In his major league career to date Knapp has pitched in 13 games, starting 6, has a record of 3-1 and a career ERA of 4.80.  Knapp\'s preferred uniform of #45 is taken so he will wear uniform #42 with the Towers.   '),(18042,3956613,'06,August,2004','2004-08-06 00:00:00','    The Toronto Towers have selected pitcher Don Hood in the sixteenth round of the inaugural APBL draft.  The 27-year-old lefty from Florence, South Carolina has played for the Baltimore Orioles and Cleveland Indians for the last four years.  In his major league career to date Hood has pitched in 90 games, starting 31, has a record of 13-18 and a career ERA of 4.28.  Hood will wear uniform #32 with the Towers.   '),(18043,3956613,'04,August,2004','2004-08-04 00:00:00','    The Toronto Towers have selected outfielder Al Woods in the fifteenth round of the inaugural APBL draft.   The 23-year-old left-hander from Oakland will make his major league debut.   Woods will wear uniform #20 with the Towers.   '),(18044,3956613,'04,August,2004','2004-08-04 00:00:00','    The Toronto Towers have selected pitcher Jim Lonborg in the fourteenth round of the inaugural APBL draft.  The 34-year-old right-hander from Santa Maria, California has played for the Boston Red Sox, Milwaukee Brewers and Philadelphia Phillies for the last twelve years.  In his major league career to date Lonborg has started 320 games, has a record of 138-122 and a career ERA of 3.74.  Lonborg will wear uniform #41 with the Towers.   '),(18045,3956613,'02,August,2004','2004-08-02 00:00:00','    The Toronto Towers have selected second baseman Rob Andrews in the thirteenth round of the inaugural APBL draft.   The 24-year-old right-hander from Santa Monica, California has played for the Houston Astros for the last two years.   In his major league career to date Andrews has 171 hits, 42 RBI and a career AVG of .249.   Andrews will wear uniform #11 with the Towers.   '),(18046,3956613,'02,August,2004','2004-08-02 00:00:00','    The Toronto Towers have selected outfielder Gary Thomasson in the twelfth round of the inaugural APBL draft.   The 25-year-old left-hander from San Diego, has played for the San Francisco Giants for the last five years.   In his major league career to date Thomasson has 312 hits, 21 home runs, 130 RBI and a career AVG of .253.   Thomasson will wear uniform #12 with the Towers.   '),(18047,3956613,'02,August,2004','2004-08-02 00:00:00','    The Toronto Towers have selected pitcher Steve Foucault in the eleventh round of the inaugural APBL draft.  The 27-year-old left-hander from Duluth, Minnesota has played for the Texas Rangers for the last four years.  In his major league career to date Foucault has pitched in 206 games, has a record of 26-25 with 35 saves and a career ERA of 3.22.  Foucault will wear uniform #29 with the Towers.   '),(18048,4285920,'21,August,2004','2004-08-21 00:00:00','   Last night i had the \"pleasure\" of chaperauing(spelling?) a Highschool dance. The school i teach at is a All boys prep school so needless to say it was a eruption of hormones on a Mt. Vesuvius sort of level. My job technically was to make sure nobody got to intimate with their partner on the dance floor. But c\'mon... it\'s not really in my make up to be like \"hey get your hand off her ass\". Shit Go for it Lil Playa. I mean if she reaches up and smacks him across the face, that lesson will stay with the kid alot longer than if i was to scold him.  Also the young ladies at this dance were kill\'n me. You would have thought it was ladies night at the mercury room by what they had on. Needless to say some of my buddies were fired up to go out afterwards, I on the other hand hit the rack because this morning we had a scrimmage down in LaMarque, Texas which is an hour away. So i figured I shouldn\'t expect my players to stay home and rest if im not so i packed it in much to my buddies chagrin. Im now back from the scrimmage and in search of a beer and a nap.   A burly Irishman is drinking in a bar. A tiny gay fellow sits beside him. After a few beers, the gay guy whispers, â€œDo you want a blow job?â€The gigantic man flips out, roars in anger, and tosses the little guy out of the bar, then returns to his stool.The shocked bartender says, â€œIâ€™ve never seen you react like that. What did that guy say?â€â€œDunno. Something about a job.â€  '),(18049,4285920,'20,August,2004','2004-08-20 00:00:00','   \"See what happens when you drink kids\" As i got in the shower the this morning I had that blurry slightly amneisic feeling of what happend last night? Aggie Ring? Check, Wallet? Check, Dignity?, Up for debate. Then it hits me \"CRAP\" you started an Internet Blog. I almost wanted the \"crap the stripper stole my wallet\" feeling instead of the \"damn is email not enough, i must post random drivle for the internet masses\".I tried this once before but failed miserably. Now that I have something some what interesting to write about and a little more free time, i\'ll give it a go. Well i now have to get back to the exciting world of American History and how Columbus enslaved the indians.  There\'s a doctoer, a lawyer, a priest and a boy scout on a airplane. The plane is about to crash and there is only one parachute. The doctor says \"lets give it to the child, it\'s the humanitarian thing to do.\" The lawyer replies \"f*ck the kid I want...\" the priest interupts \"do you think we\'ll have time?\"  '),(18050,4285920,'19,August,2004','2004-08-19 00:00:00','   I have always thought that patience was a virtue. Now I\'m starting to think it\'s a curse. at the moment I\'m a HighSchool football coach and my patience is wearing thin. Each day i tell these young men what to do and how to do it, but you would think i was i giving them advice on a horse at a derby race. So far I love my job. Maybe for the shear idea that from time to time some one listens to me and thinks I\'m right and they recognize that. In my history class they take my word as gospell. Thats probably since i make the test. Out on the practice field the other team dictates if i did a good job or not. \"Better to write for yourself and have no public, than to write for the public and have no self. \"  urlLink Cyril Connolly  (1903 - 1974)  '),(18051,3479211,'28,May,2004','2004-05-28 00:00:00','    No e-mail from hubby today so far. It\'s over half way through the working day and not a single e-mail.   Given that he normally sends missives five screens long, this is unusual. It means either   a) he is ridiculously busy at work b) he is still smarting over our argument on Wednesday night, after which we are still not speaking properly.  It\'s probably (a) and (b).   We argued because I had Wednesday off work without him. I fancied some time to myself. To hubby, this is a dreadful insult, that I could want time away from the wit & wonder that is his company.  I wonder if I should fire off a quick e-mail to him? I don\'t want the bank holiday weekend to be spent in silence.   Fired off a quick one saying \"how are you today?\"  No answer. Still, at least  I\'ve  made the effort.   Ten mins later. \"I am busy. Will be back later again tonight. Will get my own tea\".   Well, at least we\'re talking now!   '),(18052,3479211,'03,June,2004','2004-06-03 00:00:00','    Hubby has less money left than he thought, for the month. About Â£100 less. We share our money, although not in a joint account.   It\'s normally me who has less money than I thought, so this makes a nice change.  It\'ll be a flexible plastic friend month-end then - I think we\'ll have to do a medium to big shop tonight in Asda, then get some booze in Morrisons later on. I have bought two bags of frozen chicken fillets from cool-trader, so we can have stir-fry chicken with brown rice for tea in the evening. Cheap and cheerful. Well, cheap, at any rate. A bit of cheap booze from Morrisons and we\'ll make meals of it.   '),(18053,3479211,'01,June,2004','2004-06-01 00:00:00','    On Friday night I had the shock of my life. Hubby went out in - of all places - Wigan, with some of his ex-collegues. He had e-mailed me back last thing before I went home, saying that he wouldn\'t be out long.   Thankfully, Big Brother was on, and I have E4, so I could watch wannabe TV presenters make fools of themselves all night. It got to two o\'clock in the morning and I got very worried. Hubby doesn\'t have a mobile, so I couldn\'t call him. The last train back from wigan left at eleven, and I knew he wouldn\'t get a taxi, because money was very tight.   Anyway, he rolled in pissed as Boris Yeltsin at half past two. I was a little upset; well, more than a little upset. I had had a few drinks myself, and was \"tired and emotional\" (as they used to say about aforementioned ex-president). I asked him where he\'d been.   \"The train was cancelled, so I had to get a taxi.\" \"So the eleven o\'clock train was cancelled, and it took you three hours to get back in a taxi. Where were you? Birmingham? Wales? Leeds?\" \"Wigan as you well know. Anyway, I don\'t give a shit. You\'re a horrible selfish woman with really odd teeth and I hate you, trying to stop my fun.\" \"So, I was sitting here trying to work out why you hadn\'t called, because you always do. I was worried that something awful had happened.\" \"Well... fuck off. Fuck you. I\'m not fucking watching this fucking shit anyway I\'m putting a DVD on so fuck yourself.\"  There is no point in arguing with a man when he is blind drunk, and you\'re tipsy yourself. Things get out of control. Horrible insults are thrown around.   I wonder if partners who go out know what the other person goes through? I\'ll tell you. It happens in stages.   1) Happens at the time they say they were going to be back. So in hubby\'s case, about half ten / eleven. Stage one involves anticipation. You wait downstairs so you are near both the front door (to greet them) and the phone (in case they call to say they\'ll be late).   2) Happens about an hour later, so in this example, about half eleven / midnight. Mild concern, and thinking up reasons why they might not have arrived or phoned. At this stage the reasons are along the lines of delayed trains, going for takeaways, popping into the 24 hour Asda on the way back to grab a Goodfella\'s pepperoni pizza.   3) Happens an hour later again. We\'re talking one in the morning here. Strong concern. The reason they haven\'t arrived or phoned gets less logical. The train has broken down, at best. At worst, a terrible fear starts to form in your mind. You try to shut it out. He must be okay. You wonder if his work salary package includes life insurance... 4) This is another hour later. He is definitely dead, or injured, lying in a road somewhere calling out for help.   This is normally the point at which he walks in, ten-pints-of-stella pissed. Your initial reaction is relief; he\'s alive! Hurrah! but after that, you\'re angry. Really angry. Seething with uncontrollable anger for the sick feeling in your stomach. Boiling over with rage for the hot tears in the corners of your eyes. Knuckles clenched and white, you shout, and shout, and shout until your face is purple.   Ever tried shouting at someone who is so pissed they might as well have had a labotomy? It achieves nothing! It makes matters far worse. You are a controlling evil witch bitch harpy tart slag who is trying to ruin his life.   Woke up Saturday with a very sorry hubby. I managed to get him to make me a coffee out of guilt, but I didn\'t want to milk it. I used to do that if he\'d upset me; milk it for all it was worth; get a couple of days dishwashing out of it. Not now. I\'d rather it was forgotten.   Anyway, we went out that evening to a Stone Roses tribute band and had an absolutely fantastic night!  On Sunday, hubby remembered something that had happened on his Friday night trip out... and started acting shifty and guilty. I knew something was wrong \'cos he was tidying the kitchen, and getting all huggie and kissy with me in a completely non-sexual way, which is normally a sign that something is wrong!  It transpired he\'d made a pass (a quaint expression meaning tried to feel-up / snog) at an ex-colleague. A pretty 20 year old ex-colleague who, on being ogled by what to her was a dirty older man (hubby is only 26 but I remember what I thought of 26 year olds when I was 20) had, with her friend, immediately left the club and gone home leaving him on his own in Wigan in the middle of the night.   The guilt continued through Sunday. I wasn\'t at all arsed; he won\'t be the first or last bloke to make a pass at a younger woman when drunk and pissed off with the wife.   Why couldn\'t he feel guilty about the right thing? Why wasn\'t he overcome with shame about not calling me, instead of about something trivial like this?   Sometimes I don\'t get hubby one bit. Still, at least we\'re talking now.   '),(18054,3961392,'26,July,2004','2004-07-26 00:00:00','    wHy?? juz b\'coz i did a mistake .. that was done months ago .. do you hav to keep doubting mi?dun you even trust mi now?not even a single bit.. dun i mean anything to you at all now? i said i was sorry.. and i mean it.. i really am.. can\'t you juz forgive mi? not even if i begged you? everything i say seems like a lie to you ... why? i love you ..i realy love you wif all i am now.. why? do you have to hurt mi.. wouldn\'t you 4give mi.. are you going to keep that anger or hatred within you juz b\'coz of&nbsp; a small thing..can\'t you even believe mi?haiz.. i\'m really such a failure.. i dunno wad to do.. tell mi wad to do.. honey..i really dun... pls.. i love you honey.. i really really love you .. wif all i am .. all i have.. i\'m sorrie... '),(18055,3961392,'25,July,2004','2004-07-25 00:00:00','    mIss mi? guess not..*sigh* yesterday i didn\'t come online.. so i didn\'t write anything in my blog lorz.. anywayz was out wif wj at nite. we watched \"i, RobOt\" it\'s scary and kinda at least to mi.. haiz.. anywayz ..i\'m bored..  today i went to church.. it\'s a miricle i didn\'t slp thru service.. neither did dear and jess.. so.. gd lorz.. after church we went to calvary.. saw my old frenz.. miss them dearly .. saw praise.. she reconised mi.. but didn\'t really talk o her much.. coz she was eating.. as for josh(my ex-).. he juz went at the sight of mi and wj.. afraid maybe.. like who cares? speak of the devil.. he juz came online.. hahaz.. anywayz we( mum,kor kor joshua,wj,jess,mi)&nbsp;went to old airport rd there to eat.. wah sehx.. they noe mi&nbsp;on a permernant diet they purposely order all the mouth watery stuff.. i manage to keep myself off the tempting food.. but did take one or two bites.. delicious.. jess.. being&nbsp;so skinny doesn\'t even need to&nbsp;care abt the number of oil and fats that is in the fat oily food..&nbsp;even wj gave up.. she still goobled down till there\'s no more food left..haiz.. ppl like them.. so luckie.. mi only can watch and drool.. haiz.. ok.. so kor kor josh drove us back (to my house)&nbsp;..dad bought durianz.. oh manz.. hahaz.. finally sth that i&nbsp;dun mind eating.. hahaz.. jess didn\'t come out.. coz she dislike durians.. hehez.. so more for mi:P Jess went home at around 430.. (bye jess) ermm..wj.. found my diary.. somehow.. den ah ya..&nbsp;i took out my older diary&nbsp;.. abt the one when i was wif josh.. but ..&nbsp;we.. no more le.. so no need to think le.. now he is reading my&nbsp;newer&nbsp;one.. when i was wif him lor.. but i nv update le.. so long kekex.. hehez... tat\'s all ba.. &nbsp; '),(18056,3961392,'24,July,2004','2004-07-24 00:00:00','    urlLink    Me aNd jEsSie i\"m tHe One Wif tHe sUnGlAsSes.. She chIo bA?&nbsp; urlLink '),(18057,3961392,'22,July,2004','2004-07-22 00:00:00','    oh manz .. my tummy hurts.. ouch! haiz..pAiN pAin!! where\'s the doc when i need them.. oh manz.. haiz.. 2day sure another one of those boring day le baz.. oh ya..   Yesterday, after skool..i went out wif wj.. den we met annie(wj ex) hehez.. her current bf is like so err.. dunno wad to say...a bit like iz lorz.. den we when to watch show... \"ella enchanted\"... is a MUST WATCH show.. damn funny. but also damn touching...kOoL bA.. hehezoh ya after the show i saw.. xiang,deser,rachelle,abel n another person 4get his name le.. my old church frenz.. miss them sia.. haiz.. i dunno lah i think i not onli miss them.. i also miss MKC..priase... i miss her lotz..   haiz.. that\'s all ba.. today sure anoter one of those boring days de baz... dear dear today got tuition mahz.. dats whys..sianz.. '),(18058,3961392,'22,July,2004','2004-07-22 00:00:00','    urlLink    mY fAmIlY&nbsp; urlLink '),(18059,3961392,'21,July,2004','2004-07-21 00:00:00','    urlLink    mY dEaR dEar..sHuAi mA?hEhEz..&nbsp; urlLink '),(18060,3961392,'21,July,2004','2004-07-21 00:00:00','    Today was another one of those normal dayz.. so sianz.. practically no one was teaching excepy hist. and science.. oh manz.. haiz.. i dun wan to say le .. oh ya.. XT joanne they all dun wan to be friend wif amelia le.. reasoN?coz they dislike her lorz... so easy... of coz i\'m at my dear dear\'s(xueting\'s) side lahz.. haiz.. after&nbsp; skool.. mi wj.. yuxin and iz.. went to KFC together.. den.. they ..haiz.. dun say le.. say le wait scared kana scolding.. den mi waiting for my real dear dear (wenjie) to finish PW.. so sianz.. like that lorz..except that i miss my dear...'),(18061,3961392,'21,July,2004','2004-07-21 00:00:00','    today.. nuttin much.. except tat i was all drenched in the morning.. lucky for some ppl .. whose parents own car(s)*aImInG aT Wj..*haiz.. fine lor this ppl.. no need get up so early... haiz..nuttin much ba.. juz that i heard that derek and ying chiu broke up...wad sia.. haiz.. why like that de..so sad..hope mi and dear dear wouldn\'t end up like that .. i will sad sad de.. i wouldn\'t cry lorz..i\'ll die ... haiz.. dear dear would you eva like that ? hope not.. after skool got PW..sianz.. after PW lehz.. dear dear pick mi up.. but he was late..sad case..hehx.. den after that we go home le.. oh ya.. my HP confiscated by my mum.. haiz.. coz of bills .. sianz.. now i feel so \'kan cor\'..juz came back from bedok interchange haiz.. no more fri for mi.. have to be stuck at home.. waiting for the damn fridge..spoil le mahz.. *~hAiz~*      haiz.*~sIa'),(18062,3961392,'20,July,2004','2004-07-20 00:00:00','    FAT FAT FAT!!! i need to lose weight seriously.. i can\'t stand it no more.. haiz.. juz see those blubbly thingys on my tummy... which is so damn fleshy ..hey..blubbly,tummy,fleshy..kOoL... wait!!nOnOnO ... no good!!i\'m going on a diet... i know i said it abt a thousand times already.. but i really mean it this time.. ! juz bread for mi.. 2 meals a day.. breakfast and late lunch ..no dinner for mi.. no fastfood.. no recess.. oh manz.. this is going to be hard...only plain water.. and hot wadeva in the morning.. haiz.. weekends.. depends.. if i\'m wif wj.. den i\'ve definately no choice...but to eat when he tells mi to.. but on other days.. i juz hav to control myself...haiz... pls lorz.. no one wans to end themsleve up in TAF club rightz? so the embarressing..especially when ppl mock you say you fat.. haiz.. juz got to accept facts.. but strive towards the just-nice standard.. wish mi luck... oh manz.. hungry le... haiz.. wHeRe\'s mY bReAd..? '),(18063,3961392,'19,July,2004','2004-07-19 00:00:00','    Today was wad i call bOrInG.. there is really nuttin to do ..why? coz it\'s a skool day.. and skool sux..BIG TIME.. haiz.. now i\'m at wj\'s place .. they(wj, victor ,kai liang) are having their dinner while i\'m stuck in his room with an empty stomach and nuttin to eat.. haiz.. can\'t wait to go home.. so hungry... i dunno why .. but sometimes i really feel myself as an inferior to him .. well.. wj is one suave , kind, sweet talker, easy going guy.. who wouldn\'t wan him? well.. my frenz are like practically drooling over him.. haiz.. which is bAd... repeat \"BAD\"... \"she\"in which i dun wan to reveal her name is practically talking abt how handsome he is.. how gd if he was her bf..etc. and i\'m sick and tired of it.. YES.. so wad she\'s my gd fren? i hate ppl who compare bfz.. well.. sometimes i cannot help it if her bf is such a jerk.. but den.. really i dun even go for the so-called wad \'looks\'.. okok..i\'ve got to admit that he is not that bad looking.. and he is nice and everything.. but that\'s juz not wad i\'m attracted to.. den wad am i attracted to? hmm.. i also dunno.. but i juz noe i really love him alotz.. okok.. the guys are in the room le..g.t.g.. talk to you later...BYE! '),(18064,3961392,'19,July,2004','2004-07-19 00:00:00','    Well..seems like in every corner..area..place..school...practically everyone has this ..\"tHiNg\"?!? So.. i decided to try out lorz...i hOpE i noe wad i\'m doing ..hehez..  &nbsp; gReAt..2dae was rather..ermm..short?!? ok.. i slept in church today(again..) wad can i do manz..i\'m so damn tired..had to get up so early...haiz.. althought the service like starts only like 2 hrs later.. haiz.. ok .. i was NOT the only one kkz..? Jess and Wenjie also lorz..well..i tugged into wenjie and slept...hahaz..well.. by the time i woke up .. the sermon was over.. althought i managed to catch the first few mins and the last few mins of service..of which i can\'t remember...after that we three decided to&nbsp;catch a movie..well wj had 2 free tickets of which mi and jess took advantage of *eViL gRiN* hahaz.. we watched \"king arthur\" of which was decided by wj..which i was not very keened on..but i juz watched...for fun. &nbsp; tHe show ended abt 1 pm..we..ok they were hungry.. we walked one round and decided to turn to eat SAKE SUSHI instead.. which costed us a bomb (abt $72?)i at the least and had to pay $15 bucks... well wj had to pay $50 so.. hahaz..after that we juz walked our walk to city hall MRT where we stop here and there.. there was a heritage thingy where they played games like \"5 stones \" \"hop scotch\" etc... ok we played awhile den \"zhao\" home le.. haiz..jess actualy say wan to  pei wo  go swimming de.. but in the end she say she got hw.. so i say nvm lor...haiz &nbsp; dat\'s my day... sianz rightz.. tmr sure more sianz de.. coz.. skool day mahz...haiz.... &nbsp;'),(18065,3961392,'12,August,2004','2004-08-12 00:00:00','    after all that happen 2day .. i suddenly tot of this song.. ya.. you guys may be thinking \" this is such a lame song\" but to mi it is a song which has meaning to me...  Lyrics: I\'m not a girl, Not yet a woman I used to think I had the answers to everything But now I know Life doesn\'t always Go my way, yeah...  Feels like I\'m caught in the middle That\'s when I realize...  (Chorus) I\'m not a girl Not yet a woman All I need is time A moment that is mine While I\'m in between  (Verse 2) I\'m not a girl There is no need to protect me Its time that I Learn to face up to this on my own I\'ve seen so much more than u know now So tell me to shut my eyes  I\'m not a girl Not yet a woman All I need is time A moment that is mine While I\'m in between  I\'m not a girl But if u look at me closely You will see it my eyes This girl will always find Her way  I\'m not a girl (I\'m not a girl don\'t tell me what to believe) Not Yet a woman (I\'m just tryin to find the woman in me, yeah) All I need is time (All I need) A moment that is mine (That is mine) While I\'m in between  I\'m not a girl Not yet a woman All I need is time (is All I need) A moment that is mine While I\'m in between  I\'m not a girl Not yet a woman   ..hehez.. i\'m not abt to tell you wad happen 2dae.. dun ask.. coz i wouldn\'t tell:p.. lOvE you guyz.. *muackz* '),(18066,3961392,'11,August,2004','2004-08-11 00:00:00','    i dunno.. but have you ever done sth that you might not think you did but you did for someone else?Well. i have.. for the one i love.. i did many things that i neer thought i would do.. some of which i think i never did wan to let go.. coz it was very important to me .. but i did let go.. for him.. i did.. coz i thought it was worth it.. but now.. i regretted till deep down to my very soul.. i let one of the most precious thing in my life go.. that was having a chance to serve God.. he was my \"god\".. i listened to every word he said.. i cried when he hurt mi.. but i still held on.. hoping tat nuttin will go our way.. yes.. i admit along the way i met this guy.. who i knew was willing to give mi happiness.. but .. i couldn\'t accept his kind offer.. b\'coz i was in love.. wif this guy who doesn\'t even deserve it.. but i still held on.. believing his every words.. till den .. 25 dec 2003.. he had enough of me.. i guess.. for no particular reason.. he juz left.. left mi... walked out of my life.. it was the saddest day of my life... i did so much.. for him.. i gave up.. almost everything...that\'s not all.. we met again this year.. i still loved him den.. and he claimed that he still loved mi.. foohishly i believed him again.. he even kissed mi and i clinged on to that hope.. but ... my hopes were dashed..i tried to let go.. but.. i was hard..    Now.. i love another .. and yes.. i\'m still as devoted to him as i am devoted to my past love.. but i love him more.. more den anything this world can provide..With him i find joy and happiness.. everything that i was looking for.. everything that i ever wanted.. sometimes i scold myself.. \"why didn\'t i meet him earlier?\" Yes i agree.. tat we still go thru our ups and downs.. but than he was very supportive.. supporting in everyway and i truely appreciate it..wad more could anyone ask for?..and now.. i hope tat we are able to go to the next stage..together.. yes.. i noe we are still youn.. but anything is possible right?  To Wenjie :   thank you for being so supportive in my life.. And though many times i became the \"morning glory\" who always need something to cling on .. but still you\'ve helped mi.. even though we juz met.. but you\'ve give to mi happiness.. much much more than wad he gave to mi.. he nv really gave mi wad i wanted.. but than i still held on.. but that was then.. now is now.. i never regretted knowing you.. althought we\'ve juz met.. but you already made and impact in my life.. and even if we are not able to make it to the next stage ..together.. i juz wanna let you noe that i love you.. it doesn\'t matter how you look like.. i will still love you .. because ..wad really matters is the inside not out.. i love you wif all my heart honey..i love you..                                                                      Steph '),(18067,3961392,'09,August,2004','2004-08-09 00:00:00','    yesterday.. durng the party i played marjong.. wif jess.. jay.. shang... my mum .. and auntie ing.. of coz took turns lahz... anywayz.. jess wasn\'t very fond of auntie ing.. which was very obvious.. when she made remarks abt here during the game... of coz mi .. being the so-called good one( like real!) ask her to stop.. bu i agree she\'s being pretty much of a nuisence.. we dislike her very much.. but despite that we still carried on wif the game.. i won..once onli lahz.. hahaz..  anywayz.. i\'m so happy.. coz dear dear is coming back today.. but dunno whether can see him a notz lahz.. i hope so.. but i think he\'ll be tired.. so.. nvm lorz.. hehez.. juz happie that he is back..kekez...miss him so much siAZ.. '),(18068,3961392,'08,August,2004','2004-08-08 00:00:00','    haiz.. now in jess house.. her sis come back from america.. for holiday ba.. haiz.. the \"guest star\" not even here yet.. funnie.. hehez... sianz sia.. haiz.. wah sehz.. her com best sia.. so kOoL.. nv lack. so gd de her com.. if only my com like that gd le.. my com so lousy .. i wanna uy new com.. mum promise mi a new co.. but its not going to be anytime soon.. everyoe like changing com.. to premtium 4.. mi? still xiao baby.. premtium 3 nia.. haiz.. somemore so laggy.. sianz.. dear also changing.. he come back le than change.. which is in .. 2 days ba..he coming back tml.. 4 or 5 like that ba.. haiz.. sianz.. mi crazy le.. mum also say i crazy.. this few days laugh for no reason.. i uess so ba.. coz miss dear so much.. juz now during dinner( ermm.. jess\'s dinner) i was there lahz.. my aunty.. aunty florance( i think that\'s how to spell her name).. ask mi where\'s my bf.. strange rightz? hahaz..see lahz.. stead wif rich kia.. which is always the case.. they always leave mi alone in sg.. haiz.. xiang lian dear dear.. *mUaCkz* '),(18069,3961392,'07,August,2004','2004-08-07 00:00:00','    Juz updated some of the pictures i took when i was in america..new york.. mahatten to be exact... for some of you who wans to go to new york .. this is part of it.. but you hav to go there to see the clearer picture..hehez.. anywayz this was like 2 months ago?? in june ba.. hehez.. you all should go see the 9/11 site.. down there got some hanging stuff for you to read ..tell you abt how the thingy happen lorz.. the weather? hmm it\'s suppose to be their summer.. but.. it\'s cold and hot.. some times can go up 27 degress.. sometimes can go down to 15 degrees? hahaz.. oh ya.. you can see alot of babes in bikini everywhere..( i mean those places where can sun tan de lahz..like park.. beaches..) you guys muz be drooling le.. hehez.. but they only sun tan during summer lahz.. abt june..july? so.. better luck next time..i definately did not spent my 10 day stay down there looking at babes..oh yaz.. it\'s a 22 hr flight juz to tell you.. it\'s damn sianz.. unless you sit on SIA de lahz.. if you got money.. mi lucky lucky.. can sit in SIA de.. but still quite sianz lahz.. hehez.. but i miss dear alot lorz..didn\'t see him for 2 weeks lehz.. sianz.. haiz..   ok for recent news.. juz watched \"3 of a kind\" wif jess yesterday.. hehez..we \"pon\" some recording for this.. (sHhH..) dun let my mum noe.. or sure die.. hehez.. anywayz.. if you guys are intending to watch \"the village\" for its scariness.. i guess you better dun watch.. dun get decive by it\'s advertisments.. it\'s only beginning scary nia.. hahaz.. can\'t wait until \"cat woman\" is out.. oh manz.. but of coz.. i really can\'t wait until dear is back.. miss him so much.. till next time.. bye! '),(18070,3961392,'07,August,2004','2004-08-07 00:00:00','    urlLink    cOnEy IsLaNd..hAnD-mAdE sAnD cAsTlE ..bElIeVe It oR nOt.. &nbsp; urlLink '),(18071,3961392,'07,August,2004','2004-08-07 00:00:00','    urlLink    iSn\'T iT lOveLy? rAiNbOw... *~wOw~*&nbsp; urlLink '),(18072,3961392,'07,August,2004','2004-08-07 00:00:00','    urlLink    mE aNd mArCuS... iN cOnEy IsLaNd.. hEy.. duN gEt iT wRonG hErE.. wE aRe jUz fRenZ k? aNy waYz hE is OnLy 13.. i OnLy LovE dEar DeAr..tHis Is jUz A pIctUrE...&nbsp; urlLink '),(18073,3961392,'07,August,2004','2004-08-07 00:00:00','    urlLink    mE, mUmMy \'n\' kOr...sAme PlaCe.. jUz tHat diFf vIeW... cAn yOu sEe ThE rAiNboW??hEhEz..&nbsp; urlLink '),(18074,3961392,'07,August,2004','2004-08-07 00:00:00','    urlLink    mE aNd MuM.. iN aMeRiCa..dUn We loOk aLiKE.. hehez.. up loaded it abit late. but nvm... hehez.. *wInK*&nbsp; urlLink '),(18075,3961392,'07,August,2004','2004-08-07 00:00:00','    oH mAnz.. this is going to be a long holiday.. althought i\'m happy that it\'s a holiday i dun get to spent it wif my dear.. 3/4 of it.. dear is in bangkok..enjoying himself.. while mi? stuck at home. sianz..i could ask jess out.. but i think she would rather spent time wif her sis.. who juz came back from God knows where.. going to hav a gathering 2molo at jess place.. haiz.. sianz.. mi miss dear like anything.. why you muz go.. i cried last nite.. my eyes still a little swollen ba..lAo GoN..Ai Si Ni Le.. wIsh ThAt yOU aRe hErE rIgHt Now.. mIsS yOu '),(18076,3961392,'02,August,2004','2004-08-02 00:00:00','    oh manz.. so fast lehz.. 8 months past liaoz.. i wish time will juz stop .. too fast le.. haiz.. pai sehz.. mi long time nv write in blog le.. this morning raining.. but i didn\'t get wet.. duH i brought my small little BLUE umbrella.. hahaz..but very cold.. but thanks to dear.. he bring jacket for mi.. hehez.. so gd rightz?*muaCkZ* Haiz..assigned to home-econ project.. first CME den science.. noe this.. haix..wHeN will this eva stop???oh manx..  Particularly for \"e\":      a particular fren of mine \'e\'( not her real name)very idiot lehz.. always say wad .. wj so good ahz.. so shuai.. so caring etc. nv hange de.. especially during MT period always wil bring up this topic.. comparing bfs..pls lorz.. if you not satisfied wif your bf den why did you even agree to be wif him in the first place? if she not my fren...i already tell her off le.. i had enuff le lorz.. pls lorz..you wan shuai ge.. den go for shuai ge.. stop bugging mi..!In the first place if you only go for shuai ge.. den i dun suggest you stead at all.. coz looks doesn\'t matter..! '),(18077,3959121,'18,July,2004','2004-07-18 00:00:00','   Hello again! I\'m hyper right now! I had a red marker and I drew all over my face and now it won\'t come off! It was really idiotic. I do stuff like that when I get hyper. I don\'t get hyper a lot so I just do weird and crazy stuff!   Hmmmm... lets see today I stayed home the whole day. I watched a lot of tv and ate a lot. I had some of my grandma\'s tamales and ate a lot of junk food.&nbsp;Wow! I never really realized that I had such a boring and normal like.   I am tres dull! If you don\'t know what that means it is that i am very dull in french.   At least my life is very dull. I\'m watching   The 4400.  &nbsp;It is a very interesting show. OMG!   My mom, dad, and brother are like watching the game of Mexico vs. Brazil. I don\'t like soccer so I don\'t bother watching it with them. I\'m hungry! I want some blueberry pie. Ok I guess I\'ll go look for something to eat.&nbsp;  I will leave you with a quote though: \" Go confidently in the direction of your dreams. Live the life you have imagined.\"  '),(18078,3959121,'18,July,2004','2004-07-18 00:00:00','   It is another boring Saturday. Other than the fact that we went to go see Spiderman 2 today nothing else happened. Spiderman 2 was really good. It made me cry as people who know me would figure I would do. I thought that it was very dramatic and it had a great story line. I\'m the kind of person who thinks of these things. I am a perfectionist and I tend to over think things. Nothing else happened today. I did have McDonald\'s though. I really wanted to eat some of their fries and my mom went to go buy us some. I LOVE their fries! They are America\'s Favorite Fries if you had no idea. Ok, before I stop writing I will leave you with a quote: \"There are no mistakes, no coincidences. All events are blessings given to us to learn from.\" I guess that\'s it for now. I\'ll write more next time. I\'m just tired right now, it\'s late!  '),(18079,3959121,'07,August,2004','2004-08-07 00:00:00','   It has been a while since I have written but we were gone for a week in Mexico. We had fun! We went to the fair there and to the \"centro\". I bought a lot of stuff and wasted a lot of the money I had. After we left Mexico we went to South Padre Island. I had gone there before but it was like 3 years ago. It was really fun. We made a sand castle and played in the ocean. I got a really bad sunburn though. I got it on my legs, arms, face, and a really bad one on my shoulders and back. I put a lot of sunscreen on my shoulders too and I still got a really bad sunburn. It hurt really bad whenever I would touch it or lift my arms up. It was like that for about 2 days but then it turned into a dark tan. Now it is pealing. It looks really nasty! But we have been back from out of town for about 1 week now. This past weekend I have just gone to gymnastics practice and then come home. I can\'t believe school is about to start. We only have a week left for school. GRRRRR!! I hate that I really don\'t want to go back. Ok I have to go eat my burger. ADIOS  '),(18080,4154216,'20,August,2004','2004-08-20 00:00:00','   I feel rather ancient today, if truth be told... Yesterday I finally organised insurance cover for Luci to drive my car. I\'ve never been a great one for white-knuckle rides, but have to say that her driving was really quite reasonable. After all, an unexpected traffic bollard after a mini- roundabout might surprise anyone,- and there is no doubt that her braking is excellent! However, nothing had prepared me for the realisation that this is another milestone on the journey into an independent life in which I\'ll have minimal part.Of course I\'d known all along that parenthood is the gradual process of working yourself out of a job, but yesterday I realised something of how it will  feel . This morning we went in to school to collect A/S results..again there is nothing I can do to fix these if they don\'t work for her (thankfully she managed A\'s in English and Philosophy, with B\'s for History,Music and Gen. Studies...she\'s cross that History and Philos. werent reversed, but otherwise quite content) Lunch time she boarded a train for a 4 day trawl around assorted northern universities...and there she was, gone! I know she\'ll be back on Tuesday (albeit briefly) but this is a foretaste at least....methinks I have some preparation to do before I\'m ready for autumn 05,not least because I\'m not that sure I\'ve done all my own growing up yet. '),(18081,4154216,'18,August,2004','2004-08-18 00:00:00','   Not quite sure where this week has gone: 3 funerals in 2 days probably has alot to do with the speed at which time hurtles past...Yesterday\'s was special. No. Let me rephrase that. I know they are  ALL  special to the families involved, and believe me we do our utmost to ensure that they feel this. I rather think God does too.Inevitably, though, you engage differently in different contexts, and some touch you personally more than others. Yesterday\'s was for a lady whom I\'d been privileged to spend a long afternoon with, only a few days before she died. We had taken her Communion, stayed on for a cup of tea, and then been stranded thanks to prolonged and violent thunderstorms..She and her husband were clearly devoted, and as they shared stories and memories, knowing that their time was likely to be short, I really did feel that we were on holy ground. When we had visited to discuss the funeral, he had worried that maybe not many people would attend, though the forest of cards all over the house suggested this might be unlikely. In the event, the church was full and the little house afterwards overflowed with love and concern, under which the family blossomed. I honestly think they enjoyed the experience...realising that the wife and mother who was so precious to them was loved and valued in so many different contexts. It made me realise just how valuable those letters and cards you struggle to send in the aftermath of a death can be....and how much more so the hugs and memories. '),(18082,4154216,'14,August,2004','2004-08-14 00:00:00','   Thank you to everyone who prayed, supported and responded to yesterday\'s blog....It would be too much to say that things are brighter today, but the weight of misery is maybe slightly less overwhelming. Professionally, the day is one of weddings, which makes a nice change from the recent run of funerals: I can\'t actually conduct the things yet, of course, but I did have a jolly time acting as unofficial creche for assorted infants and toddlers who clearly felt that all that was necessary had been said within the first two minutes of the service. So  we  played in the sunshine and \"they\" did the grown up bits inside the church....and a good time was had by all.:-) '),(18083,4154216,'13,August,2004','2004-08-13 00:00:00','   Since my beloved youngest child was born on this day 12 years ago, I\'ve always claimed a particular fondness for 13ths in general, and Friday 13th in particular. This year I\'m rather uncertain, though. The day has featured one  deeply  unhappy parishioner, two funerals plus a funeral visit and most seriously the awareness of the huge pain that one of my dearest friends is going through, and which I can do nothing to alleviate. I had never seen myself as a Mrs Fixit but clearly I have these tendencies...to no avail. All I can do is stand at the foot of the cross and weep......for which, of course, there is rather a precedent. I know I\'m not alone there, though, and this has helped me to get through the day and even manage to play birthday jollities with the boy himself. Thanks to those whose companionship in the sadness made this more possible. '),(18084,4154216,'12,August,2004','2004-08-12 00:00:00','   Maggi\'s blog,  here , carries some interesting thoughts about the role of those passages of Scripture, liturgy et al that we actually know by heart.....which sent me off on a roundabout track to consider how much our ideas (and specificially our theological ideas) are actually   created   by the words with which we express them. That may sound bizarre at first, and I do know the dangers of locating God within a particular form of service of type of language.After all, I\'m a curate in the sort of traditional Anglican parish where, when we introduce the Lord\'s prayer with the words \"As our Saviour taught us, so we pray...\" a high proportion of the congregation assume that this means that Jesus did indeed instruct his disciples to pray in the English of 1611 ;-). Nonetheless, I\'m sure I\'m not unique in finding that certain patterns of language effect me in a way that others, carrying the same meaning, do not...The Common Worship Evening Prayer includes an updated version of one of the familiar collects, retains the meaning exactly, but totally loses the effect...Why is this? Is it purely the resonance of countless Evensongs of my chorister past...or is there something inherently special about the words selected and arranged thus...Some words, I\'m sure, are themselves Sacraments...they achieve what they symbolise or describe. For me, the final stanza of John Donne\'s \"Hymne to God the Father\",  urlLink Here , does just that, and has done so ever since I read it in a train between Eastbourne and St Leonards on Sea, going home from school.to confront a new world, the day my father died. Whenever I read it, as I have so many times through the years, the same tangible certainty that all shall be well surrounds me once again, and I do indeed \"feare no more\". I suspect this may be a glimpse of the blindingly obvious, and something others have considered long since....but, hey, I had fun cogitating anyway, and nobody forced you to read to the bitter end, did they?? PS If anyone felt moved to teach me how to do those links that would take you direct to a specific section of Maggi\'s blog (and indeed anyone else\'s)..., I\'d be one up on the superteen Giles, and seriously, if not eternally, grateful! '),(18085,4154216,'11,August,2004','2004-08-11 00:00:00','   Today, my day off, I decided to redeem myself with the children by a Proper Day Out. After an entertaining visit to the orthodontist, who had moved practice to the other side of Cheltenham without warning, so that we found her only after frantic drive across town....for an appointment that must have taken a good 45 seconds..we headed south to Somerset and Kilve. Amazing beach, with huge stones, and terraced rocks which make most convincing sleeping dragons...plus handy cave for lunch in the rain. All very happy, then as rain intensified and swimming looked increasingly unlikely, we went in search of Exmoor ponies and Lorna Doone. En route through some of the most beautifully wild country I\'ve enjoyed for a while, we saw a sign to 12th century church....Children agreed we could divert, as reward for good behaviour....but have to say I wish we hadn\'t. The church itself was tiny and quite appealing...on the edge of a tiny hamlet of 10 houses. It houses the smallest and least threatening pulpit I\'ve ever encountered:not so much \"6 feet above contradiction\" as \"a small step above uncertainty\". Amazingly it claimed to have both a vicar and a curate (the former being known as The Rev and the latter as Fr....which intrigued me mildly) and was actually planning to hold a service this coming w\'end, Evensong taking place every 3rd Sunday...But what reduced me to silent misery was the fact that it was stuffed...and I mean STUFFED...with dead flowers. Three arrangements in the sanctuary, one by the font....they simply shouted of despair and desolation. Here, in the middle of all the tourist routes, signposted from 3 directions, was a church saying \"We are dead. We are irrelevant. Even we don\'t care any more....why should you?\" G remarked in some disgust that only I could find myself in the middle of Exmoor doing theological reflections about dead flowers....sadly, I fear he might not be right...and that others might also see it this way. '),(18086,4154216,'09,August,2004','2004-08-09 00:00:00','   why, when I am endlessly patient with all sorts of unreasonable old ladies, disaffected youths and wobbly clerics, I am totally incapable of remembering for half an hour what it felt like to be 17.... Feeling very ashamed of myself, as I had a Major Row with beloved daughter, who is, not unreasonably, feeling a bit deprived of social life, since I\'ve dragged her 2o miles in the wake of my ordination,  failed so far to insure my car for her to practice driving on, and vanished cheerfully out the door calling \"Mustn\'t be late for Evening Prayer\" just when she was about to talk to me.... I guess that blogs are not intended as alternative confessionals, and there aint alot of absolution to be had anyway, but I do hope I\'m not developing into Rev Jekyll/Mummy Hyde...only worth knowing when I\'m in \"uniform\". The thing is, it\'s not just professional niceness...I really do feel tremendously fond of and concerned for all these people....Of course, I feel considerably fonder and more concerned for my nearest and dearests, but showing it, that\'s another matter :-( '),(18087,4154216,'07,August,2004','2004-08-07 00:00:00','   This morning I visited a family who come to our \"Prams and Pushchairs\" toddler church...Unusually, they had ASKED for a service of Thanksgiving rather than Baptism for the younger children, and when I met their dad it all made sense. He was a really thoughtful guy, with no churchy pretensions whatsoever, who nonetheless said that he \"respected church too much to make empty promises there\". Impressive, eh? Wish he could chat to some of the other families we encounter.We had a satisfying conversation, and I was just beginning to tear myself away from the delightful baby when her big brother asked me rather sheepishly if I could do a \" sort of prayer thingy for if you\'re being adopted\". It turned out that this is a second relationship for his mother, and that the father of the girls, is going to adopt J too....I was of course more than happy to agree to include prayers to reflect this in the service, and the grin that spread over J\'s  face when he realised that not only could he be included in his sisters\' special day, but that God might actually want to be involved in his new relationship, was worth almost anything and then a bit more. The family had already asked for the inevitable \"Let the children come to me\" reading, and it was good to be able to tell them that when Jesus placed his hands on the heads of those children he was proclaiming himself their father and protector...adopting them, in fact. Hard to tell which of us was more thrilled, really.....And they call this sort of thing \"work\". I left there with a beautiful drawing of a smiling flower in a pot and a pipecleaner heart to grace the study....and a longing for there to be more moments of revelation, when people grasp that God really does care about them, and the church manages not to put its foot in it and crush their faltering moves towards him. '),(18088,4154216,'06,August,2004','2004-08-06 00:00:00','   Well, it\'s just over a month since I dropped off the clerical production line and landed as a shiney new curate in the parish...A month of new experiences, new people by the hundred (all of whom know ALL about me, including probably where I\'ve put the keys to the vestry) and more joy than I would ever have believed a \"job\" could entail. But today is different. My boss, who is truly wonderful (gentle reader, I refer to the vicar, my more immediate boss, rather than the Boss whom you might presume to be the sole employer of even infant clergy) phoned to say that Sheila died yesterday. She was a wonderful lady. A former PCC secretary who had huge reservations about the ministry of women, she had battled with a particularly unpleasant form of cancer for  nearly 10years, suriving on liquid meal replacements ever since diagnosis. When I first went with Michael to visit her, he warned me that she might be a little cool, given her strong views on ordination. Instead she was both gracious and welcoming, and over 4 weeks we became very good friends as we explored life, death, ministry...and the mysterious attractions of Big Brother. A quick visit last Friday morning turned into nearly 2 hours, we had so much to say to each other. A return trip was in the diary for today...but it\'s too late. She waited until her daughter and family returned from holiday in the States on Tuesday...fell into a coma on Wednesday and moved on yesterday afternoon. She\'d allowed us to come so far along the road with her, it hurts that neither of us were there at the end. Go well, Sheila, straight into the loving arms of God. Thank you for your friendship and your inspiration. I\'m so glad I knew you. '),(18089,3818197,'03,July,2004','2004-07-03 00:00:00','   Funny thing.  I\'m glad someone posted a comment actually asking on what my research is about.  I guess I have failed to talk a bout that since I started my blog.  My bad.  Let me tell ya.  The idea for my research came to me as I was leaving the airport on my way back from Seattle last April.  I decided I wanted to do some work based on the experiences of refugees based on the concept of diaspora.  At first, I was ready to go out there and interview everyone who has come to this country as a refugee in order to preserve their oral histories.  Then I realize how absurd that goal was considering that I only have an X amount of time to actually complete this research and have a thorough analysis of the cases studied.  So I had to narrow down and that was really hard.  I trully wanted to go out there and spend the rest of my time interviewing families and getting to hear all their stories.  So I decided to go by gender.  Females.  Then I decided Afghan women would be an excellent group to work with.  Then I spent a week in Indiana at a McNair Senior Conferences and ended up coming home with yet another change of heart.  Since I am a double major and I have an equal amount of love for both of the degrees I have been persuing (Spanish and Multi Ethnic Studies), I decided that it would fit both categories if I worked on the experiences of Cuban refugees in Idaho, more specifically in Boise.  For example, I\'m curious to know why they have resettle in this area of the country as oppose to an are such as Florida or New York where there are thousands of \'compatriotas\'.  Now, I have been refining that even further into what exaclty is that i want to get from this.  This is what I got so far: As a two part project I have set up to go on with the recording of the oral histories, I am getting ready to lear how to transcribe them so that I have a solid data base to base my reasearch on.  I want to explore the causes and consequences of resettlement and the effects that this has in these people\'s lifes.  I also want to go into the politic aspect of it and find out more about the role that both the government of the U.S and Cuba play in these people\'s life\'s decitions.  Or something like that.  So, at a huge glance, there you have it.   '),(18090,3818197,'02,July,2004','2004-07-02 00:00:00','   Last night a couple of amazing things happened to me.  I was sitting at home, actually working on yesterday\'s post when my hubby showed up.  He said we should go pick up a video we had lend to our friend Rifka so as we walked out he suggest that I drove.  I went back inside the house to get my keys and wallet and as I turned around the corner to get into our bit up 1991 Geo Storm when... BEWARE AND BEHOLD!!!!! He had gone out behind my back and bought a PT Cruiser for my birthday!!!  It is sooooo beautiful!! It is a 2001 Limited Edition with niiiice leather interiors and keyless entry and it has a cd changer in it... I feel like a spoiled brat (I know I am), but this is one of the nicest things anyone has ever done for me.  So, needless to say my feet have been several inches off the ground due to my absolute disbelief that this is our new car!  Ok, enough about that, what it\'s really touching and really relevant to everyone else, is that last night we went and saw Farenheit 9/11 at the Flicks.  I really wasn\'t prepared by any means to come face to face with the reality of war as described by Michael Moore.  And even less by the reality of what this government is doing behind everyone\'s back.  I felt good and bad because even though as I left the theater I had learned so much, I can\'t forgive myself for having lived in such obliviousness through the 4 years of the reign of terror that has been Bush\'s presidency.  There is only one thing to say to all of this: this man (if you can call it that) cannot, be re-elected.  I feel I need to go bang on the INS\'s door and demand a citizenship just so I can go out and vote in November.  This country can\'t afford another year with this government. Is just not right. It hasn\'t been from day one, but it is still salvagable.  I guess I\'m just in awe of how ignorant the general public truly is about what goes on at the higher levels of the government.  It makes sense that they keep us so busy looking at add and getting us to buy everything under the sun or keep us petrified with their ups and downs in the country\'s security levels.  By the way, yes, I did feel like an ass getting into our new car and driving home after such a compelling documentary.   I\'m at a loss of words.   '),(18091,3818197,'01,July,2004','2004-07-01 00:00:00','   So... here I am.  After reading most of Marcy\'s and Echo\'s postings of their intrepid adventures in Egypt I couldn\'t help but to want my very own blog.  Now, I am not off into any adventure this summer, so it took me about three weeks to figure out why exactly I needed to have this blog.  I have been, however, quite busy trying to put together a research project that has taken a huge amount of my energy in the past 5 weeks and it is a totally new experience for me.  So I decided, why not a blog where I can rant on about my frustrations and gains through the process of researching, or better yet, learning to reseach? And so, I figure I can justify my very own blog in the fact that I am going to drive myself crazy unless I have an outlet to vent out while working on this project.  Plus, I figure if I rant long enough about something, maybe some really break through ideas for the project may come to mind.  That would be awesome. As of yesterday, I felt as if I had already lost the battle with research before it had even begun.  I had been fairly warned that while doing research some of the most common feelings would be dispair and frustration.  I guess people can pre warn you all they want, but it really doesn\'t mean anything until you truly experience it on your own shoes.  So it hit m about a week ago and I was ready to quit when David Hall called up a \"cohort meeting\" of sorts to have everyone who was interested in sharing their own experiences come up and speak up their minds.  A total of four of us showed up and as soon as everybody else started speaking I realize that I was by no means the only person feeling at an absolute loss.  In fact (not that this is a good thing), there were a couple people that were almost worst off than I was!! So, I left the meeting feeling MUCH better and that was followed up by yet another great meeting with my mentors today. This is why I decided today would be a good day to start up a blog.   '),(18092,3818197,'08,July,2004','2004-07-08 00:00:00','   What is it about people that they always think to fix your life before they can start contemplaiting their own?  I am writing while I\'m still very upset over an email from a long time \"friend\" of my husband\'s who I inherited through him as well.  This \"friend\" of ours has been sending us over the summer series of emails on various topics including how Zack and mine married relationship sucks (this is possibly why we are so unhappy we\'ve been together very happily for almost 9 years!), how we are shallow, unimotional and unspiritual people or how my husband is heavily influenced by me (which is funny considering that one of my biggest prides is that I rejoyce in the influence that my husband has in my life), and that this is why he will always fail at what he\'s [my husband] doing.  In the latest email he calls Zack\'s father and myself overbearing.  Zack\'s father doesn\'t even live in the same continent as us and this \"friend\" of ours hasn\'t seen him in the last 5 years or so. And yes, I can be overbearing, specially when I\'m being told how to live and how to think and told even how to cook in my own house by someone who has no right to do so.  I can get overbearing and defensive just as anybody else would under those circumstances. He suggests that Zack needs to get away from us and leave with him and go to the mountains just like they used to.  I detect a bit of insanity here.  Most people don\'t understand the meaning of growing up and growing appart.  I think it has happened to all of us that while in this transition of \'growing up\' (if there\'s such a thing) we loose people that we considered to be the greatest of friends.  I lost several \"best friends\" from high school just because we didn\'t see each other eye to eye in many, many issues, but we didn\'t send each other nasty emails, we didn\'t even get mad.  We simply acknowledged the fact that our lifes had become very different and moved on.  I still keep in touch with the majority of them, but no longer at a deep level like we used to.  And that is perfectly ok.  I feel angry, frustrated and very dissapointed with this person.  I can\'t speak on behalf of my husband, but I know he\'s at the end of his rope too.  We have choosen to not even respond to our \"friend\'s\" allegations because every time we see his name on our email all we can do is hold on tight as we read the series of negative, borderline insanely hurtful comments that follow, and the worst part is that, no matter how hard I\'ve tried to look at this \'negative reinforcement\' under a positive light, I can\'t find reasons to justify continuing a friendship with someone who uses you (us) as his scapegoat for all his frustrations, failures and sources of anger.  The part that always gets me is how the emails end in a: You guys know I love you like family...   '),(18093,3818197,'06,July,2004','2004-07-06 00:00:00','   In order to get my research on its way I need to go through an IRB evaluation. What is IRB you may wonder? Well, IRB is the Institutional Review Board, the people in charged of making sure you are not going to harm amyone while carrying on your research. I have been working along my mentor(s) for the last few weeks getting this IRB package put together and I think I finally got it to a point to where it looks good enough for them (IRB) to eat.  So, now comes the fun part. Well almost. I still need to get through the lit review and then I get to start doing interviews. I\'m really excited about that. I\'m not exactly sure how it all will pan out, but I\'m sure it will be an amazing experience anyway.   '),(18094,3818197,'17,July,2004','2004-07-17 00:00:00','   Today, I am writting from Coeur d\' Alene.  I am up here for a  friend\'s wedding with my friend Christy.  Last night while talking  to Taylor who is the son of the lady that very gracefuly offered to  host us while in town, the topic of white supremacy came up.   Today, the Arian Nations were holding their annual White Power parade  in downtown.  Needless to say I headed that way this morning to  participate in history in the making.  I had mental images of KKK  hodded people and imagined hundreds of hateful people parading through  the streets of this beautiful city.  I feared for my safety, but  my curiosity was bigger than my fears.  As we arrived there, there  were dozens of police officers and both of the main streets were  blocked off to trafic.  We stood a block away from where they were  marching and saw all 20 members of the Arian Nations congregation, most  of which are well beyond 65 years of age ridicule themselves in front  of dozens of very angry citizens protesting their mere presence.   Up on an old bit up white truck, was Richard Buttler himself, equiped  with wheel chair and oxygen tank.   Every time they chanted  \"White Power!\" the general public response was, \"Shut up and GO  HOOOME!!\"  This is one of the most incredible displays I\'ve seen  in my life.  Across the street from where we were standing, there  where at least three dozen riot police officers in full gear awaiting  further instructions.  They took over the whole street and  proceded to load up into unmarked white cars that where lined up, four  policemen and policewomen per car plus two suburbans with officers  hanging in both sides of the cars.  There were also an Explosive  Disarming Unit and an Emergency Response vehicle along side fire trucks  with their hosses set to disperse any riot that may break out.  I  was a witness to all this on July 17, 2004.   '),(18095,3818197,'13,July,2004','2004-07-13 00:00:00','   Sunday my husband and I went to see Farenheit 9/11 for the second time with 4 family members who had not seen it yet.  I thought for sure that the second time around this would be an easier film to watch, but boy, was I wrong! Afterwards we went down to the Flying M and sipped on cold coffee and discussed pur feelings about it for a couple of hours.  Now, for those of you who know Zack, you know him as a very quiet, borderline shy person who only speaks when something is truly important is to be said.  Well, he has started speaking and nodoby is going to be able to stop him now.  Zack works in a place where everyone around him owns at least a gun, very conservative, Bush supporter, Christians.  He is a sort of odd-duck but has always been treated with lots of respect and his views have always been inf anything, tolerated.  On Sunday, he brought up the fact that he really wanted to put up money for people where he works to pay for the ticket to go see the film.  Today this is the email he sent out:     \"To each person at -------, I am personally putting up $20 of my own money for you to go see Fahrenheit 9/11 - no strings attached.  All I ask is that you watch it with an open heart and mind, preferably with your families, from beginning to end.  I strongly believe that this is one of the most important films of our lifetime, and each of us as American citizens owes it to ourselves and our families to at least watch it all the way through.    If you think that you already know what this film is about (2 hours of Bush-bashing propaganda and Michael Moore preaching lies to you about how awful America is) then that\'s all the more reason you should go watch it.    I\'m the type of person that is not easily fooled or swayed by clever editing, especially knowing how easy it is to skew perception in one\'s favor by splicing speeches and only portraying one side of an issue.  With such close scrutiny on Michael Moore these days, I walked into Fahrenheit 9/11 with a lot of skepticism, and as I was watching it, tried to separate each video clip from the film and place it back into the context from which it was taken, hoping to get the most honest, unbiased view of each person represented.  Even despite the parts I disagreed with, this film left me shocked, saddened, and outraged at at what is going on in our country right under our noses.  There are numerous honest, real, uncut interviews and sections of press footage that simply cannot be ignored.    Please see this film in the theaters.  Don\'t wait until it comes to home video.  It\'s playing at The Flicks downtown or Edward\'s at these show-times:[attachment here].   Matinee prices are about $5.50 if you go before 5:15pm (weekdays or weekends).  Even if you get nothing else from it, at the very least I\'ll make it free for you, plus $12-$15 in your pocket for at least giving it a try.  Zachary Black  The only thing I edited out of this email was the location of his workplace for privacy.  So far, his boss compromised to go watch it and will also take his family along...   '),(18096,4164945,'15,August,2004','2004-08-15 00:00:00','    ***whoops!  forgot to click on the \"publish\" button.***  now playing: urlLink taking back sunday - you\'re so last summer   i was sleeping for about the 5,800th time, when i was rudely awoken by the sounds of loud boisterious laughter.  like any sleep-deprived teenager would do, i buried myself in my nice comfy spider-man-covered pillow (hey! it\'s comfy, goddamit!), in hopes of muffling the audibility. it, of course, did not work.  after rubbing my bloodshot eyes, combing my unkempt hair with my hand, i walked(still half-asleep) to source of all this don\'t-let-eugene-sleep-since-he-stayed-up-all-night mayhem.  still yawning and rubbing my eyes to better adjust for discernment, i was suprisingly greeted by...my cousins.  so i thought to myself.   wait a minute, i distinctly remember barricading the house and installing the moat.  how did they...?  oh well.  so, cousins, eh?  that doesn\'t sound so bad. at least you\'re with family and it isn\'t some sick petaphile that broke into your house looking to get some.     yep.  shawn and allison are here.  my mom pointed out that shawn bears an uncanny resemblence to elijah woods.  now that i think of it, he does look like that weak pansy hobbit.   '),(18097,4164945,'15,August,2004','2004-08-15 00:00:00','    yeah, i decided to blog-hop. so what, you commie?       i just realized classic xanga is garbage, and that blogger is a lot better. calm down, ryan, owen, and the rest of the elite blogging community. simma\' da\' na\'. yes, i know all of you were on the verge of one your snoopy dances. i have a couple valid reasons for the blog-hopping madness.         yesterday, i finally got around to tinkering with my xanga\'s skin feature. so, i did some clickety-clicks here and there, and viola, nothing. you would think it was because of some minor mistake on my part, but for once in a lifetime IT WASN\'T! apparently, i had to get premium to enjoy the xanga skin goodness, the sheek uploading capablities, et cetera. then what\'s the problem, you ask? premium is not free. (haha, you know where this is heading, doncha?) the price for premium ranges from $4-$100, depending on the subscription. just hearing the words \"not\" and \"free\" in the same sentence set off some sort of alarm or spider-sense, something in my very asian consciousness. man, that was almost as bad as hearing the words \"lets get married.\" i obviously didn\'t want to pay to upgrade, so i decided to \"jump ship\" so to speak. as much as i love xanga, it just isn\'t worth paying for something that everyone should enjoy. i mean, c\'mon, if your mom is being passed around so easily, why shouldn\'t xanga? damn, for a $100 i better damn well be immortalized. so, now we\'re back to my point: why should i? why should i pay for something as trivial as typing a few measely words onto a homepage, and only at a fraction of blogger\'s capabalities. a fraction of blogger\'s capabilities! oh gee, that sounds  so  much more appealing.   premium, pssh. i suggest you take that $100 and think about that, xanga.      so enjoy my new blog, \'cause if you don\'t...a good day to you, sir. oh yeah, please leave tips. i\'m still learning html as i go.   '),(18098,3425069,'29,May,2004','2004-05-29 00:00:00','   Sorry it\'s taken so long to get another post up, it\'s been so busy at work that I haven\'t had time to post....  We\'ve reached the end of the first week, and Wow, What a Week!  Polls that show the conservatives at 35% in the Polls vs Libs at 40%, Nationally 30 for the Conservatives vs 38 for the Libs. That seems like an upset to me.   In LNC, things are flying along smoothly. We\'re getting great reactions from people at the door, and we are getting lots of \"I was a life-long Liberal, but I don\'t like that Martin/Fontana guy\". To me, that smells like a riding that will be shifting to Blue!  Personally, it\'s incredible to see the change as people are now relying on me as a bigger partner in the campaign team. I guess offering 50 hours a week to the campaign can do that :) I only wish I could bump it up to 100, but I guess it\'s a good idea to be paid, so I\'ll keep the job here for now :)  Today was a great canvassing blitz that was extremely positive & helped us to via for a dominant position in the riding. Joe Fontana\'s boys/girls are running scared & we are on a clear path with a strong strategy. Here\'s to it\'s continued success!  '),(18099,3425069,'23,May,2004','2004-05-23 00:00:00','   I\'m not sure whether you got a chance to watch Paul Martin walking from 24 Sussex to the Govenor General\'s residence. If you missed it, good for you. If, like me, you had nothing better to do at 1pm than watch it, I feel great pity for you (and for me?).  The throngs of reporters around him as he walked up with his wife was ridiculous. Why not have one or two cameras, feed it into a pool, and give the man his last dignity. After all, he won\'t be making a walk like it ever again (a la Winds of Conservative Change).   The idea that this is such a huge media event is just silly. It doesn\'t even deserve to be labeled with the much hipper word \'ridiculous\'. Let\'s call it what it is, silliness. Do any of the reporters there really expect Paul Martin to drop some type of bombshell announcement as he walks up to have the writ dropped? And if he did drop some tid-bit anyways, wouldn\'t a single feed for all the networks offer access to the same information?  Ok, so that got more attention, even in this post, then it deserved.  This morning I was out at 7 am putting signs on bridges in the city. There\'s something about putting up a sign in the pouring rain with lightening starting overhead to make getting involved in a campaign seem fun. If you doubt whether it can really be fun, give it a try (seriously good times)  Tomorrow should bring with it some light door-knocking (after all, it\'s a holiday & time to relax!)....  '),(18100,3425069,'22,May,2004','2004-05-22 00:00:00','   As pretty much everyone knows, the writ gets dropped tomorrow.  In those immoral words from 9/11 \"Let\'s Roll!\"  (disclaimer: this is not to indicate that the Liberal hold on Canada is in any way related to the grip of terror that the 9/11 terrorists held over the U.S, but instead to applaud the call to action of those brave heroes ~ if you thought this, then really, what is wrong with you ?!?)  More large signs go up tomorrow morning at 7 am for a copy of hours. After a 17 hour day today, that\'s pretty brutal, but that\'s life on the campaign trail. Good thing tomorrow is a Sunday, and therefore should mean a bit of a respite.  As the last little bit for today, I just have to say that there are some real sharp people working on the Gatten campaign. It should be quite a ride...  '),(18101,3336745,'28,May,2004','2004-05-28 00:00:00','   I guess its been about 10 days since I\'ve updated.  But it doesn\'t really matter cuz not a whole lot has been going on I guess.  It\'s Friday, again.  And again, I\'m stuck at home doing nothing.  I\'m so pissed today and I don\'t know why.  Zak\'s been acting really weird.  All of a sudden he talks to me and has been giving me high fives... ever since I told him I listened to 10 minutes of this Slipknot special thing on the radio.  but i just found out now that corey taylor, the singer for stone sour, was in slipknot.  and jim root was the guitarist for slipknot and stone sour too.  but i gotta get offline soon...  i\'ve been on forever and i have to record the simpsons.  its gonna be the episode about bart writing to his teacher calling himself woodrow (named after woodrow wilson the president)   '),(18102,3336745,'18,May,2004','2004-05-18 00:00:00','   Today my team had a science field trip to devils lake.  thursday is the other teams trip there.  we went to devils lake.  the hike was really tough up the steep rocks.  it took forever and was really exhausting.  at least the guys butts were there to motivate me and cassie.  james is really sexy;)  haha on the bus adam (he\'s really dumb) poured water on randys pants and on his white shirt.  so he got nip rock and it looked like he wet his pants.  randy laughs really weird.  he laughs with his mouth open wide and makes a funny noise.  zak was ok too.  hes usually really annoying though.  but he does know a lot about music, so its fun talking to him.  james knows about music too, but hes just better to talk to.  Glass Jaw by Systematic is a good song that I\'m listening to right now.  My sister had a 30 second clip of it on the computer and i finally found out what it was cuz it was just titled a number.  Saturday night i did go to brittanys dance recital.  they played sweet child o mine but sheryl crow sang it.  they also had some really shitty songs like backstreet boys and other crap.  it was 3 hours long..  but chris k. made it all worth it.  me and cassie were staring at him for the longest time, but he actually saw us looking so we couldnt look anymore..  and alex and matt were in the row behind us headbanging.  it was hilarious though.  cassie spent the night and then sunday we went to church becuz thomas (tom) is usually there.  he wasn\'t though.  then me and lisa (my sister...) went shopping.  we didn\'t even get to do what we went there for which was returning shoes for my concert outfit.  but i did get about 6 shirts and one pair of kakis (however you spell it).  they\'re not the ones that go tight to your legs, so i like them.  we got lisa another cell phone from cingular.  so she has one from at&t and cingular.  she was going on the family plan so she could share it with her boyfriend since she calls him the most.  once we got the 2 cute little flip phones, it was time to leave so i get back in time for religion (thank god it was my last night)  we were crossing the bridge to get back into sauk and then the phone didn\'t get service anymore.  so lisa is gonna take the phones back.  but she was gonna let me use one for the week since she decided trevor didn\'t need his for a little bit.  lol.  then i dont remember what i did on monday really... but it was a boring day.  it stormed a lot at night though.  i was interviewing my grandma on the great depression and my dad and i were gonna leave but then we stayed 2 extra minutes and we got outside just when it was pouring down rain.  the lightning was really bad too.  but we got out alive.  then i had to wait an extra hour to take my shower.  it was alright though.  today was the field trip, really boring.  james found some marijuana leaves. and tomorrow i dont have class again because i have 4 band concerts all day long and we all get to go out to eat for lunch.  thats gonna be the only fun part really.  and missing class.  but thats about it MJ   '),(18103,3336745,'14,May,2004','2004-05-14 00:00:00','   I went to bed at 2 last night and i woke up at noon.  :(  In 3 hours I have to go to my cousins graduation party and then we eat at 5, and it ends at 7.  I dont wanna go at all.  It\'s gonna suck.  But I hope i get back before 7 so i can go to brittany\'s dance recital.  The party is going to be so boring.  I should just stay home, but I have to go.  I have to take a shower, but I can\'t until my clothes are done washing and put into the dryer so i can change into something when i get out. Then tomorrow I have to go to church at 10:30 in the morning, then for my english homework i have to interview my grandma about the depression.  & at 6:30 i have to go to religion..which is really really boring.  but it\'s the last one.  there\'s also a high school band concert at 7 too.  i wanna go to that to see the major hot drummers..and a tuba player.. but at least i\'ll see the tuba player at church.  This weekend just sucks.  our washer sucks.  it keeps stopping so i have to keep restarting it.  i guess i dont really have anything else to say..  MJ   '),(18104,3336745,'13,May,2004','2004-05-13 00:00:00','   When did life get so boring?  It\'s a Friday night and I\'m home yet again.  Nothing much has happened lately.  Except I did my normal Friday night cleaning.  It\'s freaky.  Every Friday night I just happen to be in the mood to clean my room.  It\'s a surprise how much crap a person can use during a week.  I\'m gonna have a band concert next week Wednesday and Thursday night.  During the day Wednesday I have to perform 4 concerts for grades younger than me.  The part when I don\'t have to play is going to be fun, and the part when we get to go out to eat.  I really like my concert outfit.  It\'s one of those white poofy skirts and a black tank topish shirt.  Lisa just went shopping to buy me some black shoes tonight.  She picked up 3 pairs for me to choose from.  But I won\'t get to see them until tomorrow night.  I hope she found some good ones.  But in the mean time, I\'m doing nothing.   '),(18105,3336745,'13,May,2004','2004-05-13 00:00:00','   Im bored again!!  Now I\'m listening to Little Know It All by Iggy Pop..  its a good song.  There\'s hardly ever gonna be a time when Im not bored.  brittany (a friend) is in a dance recital tonight and im not going to it.  i couldn\'t find anyone to go with.  so im gonna go tomorrow night at 7 instead.  she\'s gonna dance to livin on a prayer by bon jovi.  i forgot which other 2 songs she\'s dancing to.  probly cuz bon jovi is the best one.  i went to 2 of his concerts already.  he\'s awesome live.  he actually sounds like he does on the cd.  he just gets the crowd going.  at the first concert of his i saw the goo goo dolls.  they were alright.  i never really listened to their cd so i didn\'t know many of the songs.  the lead singer was an ass.  he kept complaining about being on tour and how much he just wanted to go home.  then at the 2nd concert i saw sheryl crow.  she was actually nice.  but i dont really like her music.  i also saw carolines spine in concert.  they were at the taste of madison.  so they\'re concert was free.  i only knew 2 songs by them, but i wanted to go to a concert, so i went with my 2 sisters.  it was kinda funny because a guy who was 22 tried hitting on lisa.  if i went to the taste of madison one day earlier i could have seen v-shaped mind.  2 guys from my school saw them.  i like them a whole lot better than carolines spine.  nick, the guy i like who went to v-shaped mind thinks carolines spine is \"an aquired taste\".  i had him burn me the v-shaped mind cd cuz the only song i heard was \"monsters\"  which is a pretty good song.  but they have some other good tracks too.   MJ   '),(18106,3336745,'13,May,2004','2004-05-13 00:00:00','   Well this is my first time actually making a blog, so I\'m new to this.  My teachers are so dumb sometimes.  Each teacher for the main subjects gave us a bunch of homework all due tomorrow.  It\'s like they\'re all trying to plot against us.  Right now I\'m listening to a song called California Songs by Local H.  It\'s a pretty good song, so check it out.  Lately I\'ve been wanting to go to a concert.  So I asked Cassie, one of my friends if they could go to Projekt Revolution Aug. 27.  Her mom actually said yes.  So we were planning on seeing KoRn (one of my favorite bands) Linkin Park, Snoop Dogg, and The Used.  The only problem was getting a ride there.  I asked my oldest sister Jenny who\'s 23 to take me.  It turns out that she\'s going on a trip with her ex-boyfriend the next day, so she doesn\'t wanna go.  Then I asked my other sister Lisa who\'s 18 to take me.  She just told me she doesn\'t like KoRn.  She won\'t go either.  I know my parents won\'t take me because it\'s about a 2 1/2 hour drive.  I guess I\'ll just have wait until they come out with another cd and come closer to me.  But for now I have to finish my homework.  I couldn\'t figure out how to get back to this draft, so i couldn\'t publish it yesterday, the 13, so its actually from yesterday, not today (the 14)   '),(18107,3336745,'30,June,2004','2004-06-30 00:00:00','   Hey anyone who\'s reading my blog, which might be noone.  I haven\'t updated in awhile.. again.  I always forget to or i have nothing to update.  But Monday I had a marching band party that started at 3-6:30 when we had to go to marching band.  Me and Cassie had my mom drop us off at the party.  We had some food and then played volley ball with some really hot seniors.  I just watched though.. even if i was playing, i wouldn\'t have helped much.  they put in some music and at first we listened to limp bizkit.  some of the songs were ok, but otherwise i dont really like limp.  then Tom, the senior that i guess you could say i like, but i dont know if i do or not, (who still has his girlfriend named alex who is a grade older than him) put in bob marley.  jammin is my favorite song on that cd.  we listened to practically the whole cd, and then we listened to tool, i think it was the lateralus cd, or whichever one has the person on the front with their legs behind their head.... haha head.  because tom said he wished he could do that and then he wouldn\'t be as lonely anymore.  this was when he was playing euquer with all senior guys, only 4 total.  on toms team was chris, another hottie, then on the other team was a guy named derrick with t.j. who is decent..  then tom said something about how chris was masturbating with his leg, so then of course, with toms personality, he showed us.  haha.  then chris put in the black eyed peas and we listened to \'lets get retarded\' and tom was making fun of chris and the song.  after their game was done, t.j. got up and then tom asked if any of us wanted to start another game.. nobody really moved.  it was me, stacy (blllllahh), cassie, and some other freshmen there too.  so after about 30 seconds, chris said that they actually wouldn\'t have time for another game since we\'d have to leave to go to band soon.  some other girl sat down just to sit, she\'s a little odd i guess you could say, and tom told her that they wouldn\'t have time for another game actually.  i think he said that so he wouldn\'t have to sit by her.... i dont know.  possibly.  Tom was asking people if they needed a ride over to the high school and i was standing there so i got a ride in his ford taurus:D:D:D   Which just happens to be the same kind of car that my dad has because i picked it out for him to buy:D  hehehe.  he had mudvayne blaring in the car.  it was great.  it was the l.d. 50 cd.  he looked so good singing the 1, 2, 3, part in internal primates forever.  i gotta go now though, its \"just shoot me\" time. MJ   '),(18108,3336745,'18,June,2004','2004-06-18 00:00:00','   Well it\'s about 4:50 right now.  I\'m waiting for the 5 o\'clock friday song to come on the radio.  I didn\'t go to bed until 3 this morning, and then i woke up at 10.  i watched the price is right until 11.  somebody could\'ve won a mustang, but they didn\'t.  I went out to eat with my mom and sister in madison for chinese.  the food was alright, but the drive made me sick.  i\'m taking a survey on music for a radio station that is copying my favorite station.  it pissed me off so much.  so i\'m saying the opposite of what i like so they play crap and people won\'t listen to them anymore and then they\'re be gone.  I got to talk to Cassie in texas today too.  it made my day.  i haven\'t talked to her since sunday morning.  last night at kayser, they had a premiere night for the new dodges that they got in.  they had a door prize.  my dads name was called and we didn\'t even know it because a little kid was announcing them.  there was four things left, a weather radio, a cd player/radio, a headphones cd player, and then something else too.  my dad wanted the weather thing and i wanted the cd player even though we already have like 6 of them.  we got the cd player.  he likes it becuz they\'re digital sound.  there were two girls at the chinese place that kept looking at me weird.  and then bon jovi \"livin on a prayer\" came on the radio & me and lisa started singing.  it was lots of fun.  WAHOO!! the song\'s on:D pass it to me baby, pass it to me slow.... blah blah..... the song\'s about doing drugs.  haha.  it\'s on every friday at 5, i think i already said that...:S   MJ   '),(18109,3336745,'17,June,2004','2004-06-17 00:00:00','   Well the last day I could post was Saturdsay the 12.  It was the only real day I had access to a computer.  On Sunday after church where I saw Tom and Luke play the guitar, I went to Lisa\'s house.  I stayed there until 7:00 this morning.  Jenny came and picked me up since she was going back into town to work anyways.  I don\'t remember much of what I actually did at Lisa\'s.  But I went to work with her two times and met a lot of people there.  Lisa got paid to go get food and take cars from one car dealership, to adealership on the other side of town.  I got to go with her too.  We also watched a couple of movies at her house.  We went shopping two times, the first time Jenny came with, the second she didn\'t.  We ate breakfast at IHOP-International House of Pancakes, or as Lisa and I said, International House of PANTIES!!  Monday\'s my first day of Marching Band, Cassie\'s still in Texas, so I won\'t have any friends that I know of there with me.  So that\'s gonna suck.   MJ   '),(18110,3336745,'12,June,2004','2004-06-12 00:00:00','   well tuesday was my last day and a half day.  we had yearbook signing and graduation ceremonys.  the ceremony part was soooo boring.  after school brittany was having her party.  well.... it was boring at first until me cassie and ally left to go to tyler and joes house.  then for awhile i forgot about brittany\'s house and the other people.  we played twister at tylers house.  it was funny.  then the rest of brittanys people came over because they were done watching their movie & being boring.  once they all came it started to get boring.  before they came, tyler made us play truth or dare.  we made him lick the tile floor and write his name on it.  that was funny.  we all left joes and went on the trampoline... i don\'t remember much else, but it wasn\'t all that great.  the mosquitoes were a bitch though.  at about 9 10 some of us went to toilet paper and silly string luke\'s house.  at about 12 the next day britts mom took the 3 of us home becuz everybody else left at 10 the night before.  but today, saturday, my dad and i took lisa back home.  i\'m gonna spend a couple of days at her house after tomorrow.  today i mowed the lawn and was gonna go outside to tan.  but by the time i got out there, the sun wasn\'t shining much anymore.  so i went to take a shower.  its been a really boring summer so far.  i hope i can get a hold of my references soon for my job so i can start working.  then in a week i start marching band.. so there i can see tom and chris.  but tom got a new girlfriend i think.  and she\'s in his grade.  he did have a senior girlfriend, but she graduated & \"long distance relationships don\'t last\" thats all for now i guess mj   '),(18111,3336745,'07,June,2004','2004-06-07 00:00:00','   Well today was Monday and my school went to a waterpark.  It was about 80 degrees at 10 a.m. so I\'m not sure on the temp. the rest of the day.  There was this one guy that Cassie me and Britt kept running into.  It was kinda funny.  So after the 7th time seeing him, and him probably think we\'re stalking him, I decided that if we saw him 2 more times, we would ask him and his friends to go down a 5 person ride.  We saw him 3 more times and actually did ask.  They actually said yes.. He had 2 friends so we were a total of 6, but Brittany left us for some other people.  So then there was 5.  It was kinda fun.  Then we had to go because our buses were waiting for us.  It wasn\'t that sunny out even, so I got a little tan on my face and shoulders, not even a burn.  But tomorrow\'s graduation for 8th grade and I\'m gong to wear a white poofy skirt.  I like it.  It\'s a half day and then Brittany\'s having an end of school party so Cassie, me, and some other people are going to spend the night.  Joe, next door is also having a party.  He\'s inviting Marcus, who\'s going out with Cassie, so that\'ll be interesting.  I have to remember to bring some cd\'s to Britts house otherwise jumping on the trampoline won\'t be as fun.  I have to take a shower though, because I have chlorine in my hair still.  MJ   '),(18112,3336745,'05,June,2004','2004-06-05 00:00:00','   Well its Saturday.  I slept in until noon again.  Then I went online until 12:45 and my Dad came home from work and asked me if I wanted to go to a graduation party.  I only went because there was going to be food.  So I got in the shower and started getting ready.  Then Cassie called.  I finished painting my toenails and once my parents were ready I got off the phone and we left.  When we first got there, we couldn\'t find a parking spot, so we had to back up in a driveway.  In the driveway, there was 2 hot brothers with their shirts off.  They were so sexy;)  Then we were at the party for about 1 1/2 hours and we left at 3:30.  Once I got home I went back online.  I blasted the radio because my parents left again to go to church and a 20 year thing of being a priest party.  They\'re still there now.  But I cleaned my room too.  It wasn\'t that messy to begin with though.  I\'m also scanning my computer for viruses because I get so many from WinMx I think.  I\'m not even half way through the scan and I already have 38.  Jeez.. but I already got rid of two.  It\'s going to take awhile too.  It sucks.  So now I\'m going to clean off the viruses and get some food. MJ   '),(18113,3336745,'04,June,2004','2004-06-04 00:00:00','   Wow, I haven\'t posted in awhile again.  I guess sometimes I think of it, but then when I can I just forget to post.  I can\'t even view it though, I think it might be something wrong with my computer... not sure though.  But last weekend, me and Lisa (sister) went out to eat 3 times and spent a total of 75 bucks.  We had soo much fun.  We laughed a lot.  I even spit out my shake I got because I was laughing so much.  But then Monday I had no school.  Tuesday through Thursday was just boring.  But today, Friday, was kinda fun.  Well, yesterday we got our yearbooks since Tuesday is our last day of school.  Monday the school is going to a waterpark and it\'s supposed to be 81 degrees outside.  Anyways, today some people were signing yearbooks and I went up to Nick (remember thats who I like) and I signed his yearbook... But I\'m not a person who\'s that outgoing to tell him I like him in there.  And since Becky told me I should, I had her write it in there by hers so she told him whatever she did and then said P.S. MJ thinks your hot.  Then I guess in art class, which is the class Cassie, Becky, and Nick have together, Nick said I was cute and we like the same music.  I\'m not exactly sure if he was just saying that to be nice though.. and what he meant by the music part.  But we do like the same music.  But it made me very happy to find out that he thought I was cute.  I\'m not sure if he likes me though.  I hope so.  In most of my classes we watched movies and then the last 10-20 minutes of class we signed yearbooks.  In social studies we got to sign the whole hour because Mr. T decided that we worked so hard throughout the year.  I can\'t wait until Monday and the waterpark.  Last year it was really fun, but cold.  Since its gonna be warm it should be a really nice day.  Then Tuesday my grade is having 2 hours for yearbook signing.  But I already had everybody who I wanted sign my yearbook.  Me and Cassie filled 2 pages of eachothers books already.  It\'s really funny though.  I guess thats about it.. MJ   '),(18114,3336745,'25,July,2004','2004-07-25 00:00:00','   Well, I haven\'t posted since July 5.&nbsp; So it\'s been awhile, but a fun while it\'s been.&nbsp; Yesterday, Saturday July 24, was awesome.&nbsp; Probably the best day of my life so far.&nbsp; I was at JJO Band Camp.&nbsp; The radio station 94.1 JJO had 8 national bands and 5 local bands to come and play on willow island.&nbsp; I saw Dropbox, Skillet, Flaw, Earshot, Dope, Drowning Pool, Damageplan, and Saliva.&nbsp; At first, I had only heard one song from Dropbox, Wishbone.&nbsp; Its a pretty good song.&nbsp; After seeing them live, I found out that they have some other good songs too.&nbsp; So I\'ve been thinking about buying their cd.&nbsp; Next up was Skillet.&nbsp; The singer guy was kind of hot.&nbsp; But his wife was one of the guitar players...&nbsp; The drummer was also a girl.&nbsp; She wore a Bon&nbsp;Jovi t-shirt, which was my first 2 concerts, so that was kinda cool.&nbsp; They\'re best song was \"Savior.\"&nbsp; They have a couple of other good songs like \"My Obsession\" and \"Forsaken.\"&nbsp; They are more of a Christian Rock band though.&nbsp; I had only heard one song from Flaw too, and that was \"Recognize\" which is a slower song.&nbsp; So I thought that Flaw would be a real boring band.&nbsp; But I was wrong.&nbsp; Most of their songs were a lot heavier I thought.&nbsp; So then the Moshpits started forming more than they had been for Dropbox.&nbsp; Earshot was the next band I saw.&nbsp; The only good songs they had were \"Wait\" and \"Get Away\" that they played, and those were the last two.&nbsp; So Earshot wasn\'t as good as I thought they might have been.&nbsp; A lot of people didn\'t leave their spots to go see the local bands because Dope was on next.&nbsp; The crowd went absolutely crazy for them.&nbsp; I was getting slammed into people all over.&nbsp; I could barely breathe.&nbsp; But this one guy about 15 or 16 maybe, helped me and Cassie get out.&nbsp; It was kind of awkward...&nbsp;&nbsp;Cassie got out almost right away, but I was in the crowd for about 2 minutes longer because she could body surf out,&nbsp;and once she was out, the crowd got worse&nbsp;and the person who lifted her out, couldn\'t move to lift me out either.&nbsp; But once they were done, Drowning Pool went on stage.&nbsp; They were frickin\' awesome.&nbsp; They had a great starter, \"Sinner.\"&nbsp;&nbsp;They played all of the good songs that they had too.&nbsp; Then Damageplan went on.&nbsp; I didn\'t go watch them that much because&nbsp;I didn\'t&nbsp;like&nbsp;many of their songs.&nbsp; They had a couple of good ones&nbsp;like \"Pride,\" \"Breathing&nbsp;New Life\" and \"Fuck You\" I think it\'s&nbsp;called.&nbsp; They&nbsp;also played \"Walk\"&nbsp;from Pantera becuz they have two members from Pantera in their band.&nbsp; The last&nbsp;band that went on was Saliva.&nbsp;&nbsp;They were my favorite&nbsp;live.&nbsp; Then Drowning Pool.&nbsp; They started with \"Raise Up\" and came back with&nbsp;\"Always\" as their&nbsp;encore.&nbsp; Saliva was just so awesome.&nbsp; I\'d pay 27 bucks to see them&nbsp;by themselves.&nbsp; I lost my voice probably during Drowning Pool and even more during Saliva.&nbsp; It was so hard to talk and hear other people.&nbsp; There was a bunch more stuff that happened, but it would take forever to type it all.     MJ   '),(18115,3336745,'05,July,2004','2004-07-05 00:00:00','   Well, its been about 5 or 6 days or something like that since I last posted.  It\'s Monday night so I have marching band tonight at 6:30-8:45.  Yesterday we had a parade so I was marching in that.  I couldn\'t keep the right foot going.  So when I was concentrating on the feet, I forgot the song.  So it kinda sucked and I probably looked like a dipshit.  But my favorite radio station is having something they call a band camp.  I really wanna go to that.  It\'s going to have 13 bands, 2 stages, 26 bucks, which is 2 bucks a band.  They\'re going to have Drowning Pool, Damageplan, Skillet, Flaw, DropBox, Dope, and some other bands there.  They have times when you can call in, and if you\'re the 9th caller, you get t-shirts and hats and other cool crap and 2 tickets to Band Camp.  So far I\'ve called and been caller 5 and then it\'s also been busy.  So lots of people wanna go..  Jenny is thinking about taking me, but she\'s not positive.  She\'ll take me for sure if I win tickets, but not sure about buying them.  If I got to go, it would be really awesome.  On Friday at the fireworks I went to, they had some music in the background.  And after the guy said the finale was going to be next, Bodies by Drowning Pool (who are going to be at Band Camp) came on and right when they started shouting FLOOOOOOOOR in the song, the finale started.  I kept telling Jenny it was a sign.  So I think I might have a chance at going, hopefully.  MJ   '),(18116,3336745,'01,August,2004','2004-08-01 00:00:00','   Man. I\'m soo tired right now. Blah blah blah. I joined   urlLink  www.hi5.com    and recently I\'ve been filling in the journals on there. But lately I can\'t even post because of some error or what not. So I\'m just going to copy what I wrote there and paste it in here....   well hey.  it\'s sunday night, 11:00 p.m. on the dot. i tried making an entry this morning at 1 or something, but each time i tried posting, it would redirect me to the login page and i lost my entry.  but i figured that would happen so i got it copied and i\'ll paste it at the end of this entry if this gets posted.  maybe you can only have 1 post a day or something?  so today i listened to rockline and i went to church right after hearing \"Survival of the Sickest\" (saliva), which put that song in my head.  Basically the whole time i wasn\'t paying attention, which i never do much anyways, i was just singing that song in my head cuz i know every word..  then i came home and nobody was here.  i knew my dad was gonna be helping my sister move, but i didn\'t know my mom was gonna be mowing my g-pa\'s lawn.  there were 3 messages 1: lisa (middle sister who i look most like) she wanted to know if anybody was gonna be home to see if she should come home from madison.  2: jenny-wanted to know about something or other 3: this one lady that my mom doesn\'t like anymore.   so i called lisa and jenny back, then i went online for a little bit.  then i played sims for about 30 minutes and then lisa came home.  we called jenny and went to the gas station and the carwash.  I was gonna help my dad and jenny move, but they both get cranky after awhile and it was really warm out too.  so i decided to bail.  me and lisa watched \"Don Juan De Marco\" that we borrowed from jenny.  this was the first time i actually saw the whole thing.  its a pretty good movie.  Then we both took naps.  i ended up sleeping longer just becuz i didn\'t wanna move and she got a phone call.  Lacuna Coil just came on the radio with \"Swamped.\"  lisa\'s starting to like them.  it really pisses me off after i type a big journal thing and i accidentally close out the window.. grr.  hasn\'t happened to me yet tonight, and i hope it doesn\'t.  there\'s this new show on abc called \"The Days.\"  I\'ve been watching it lately.  Call me weird, but the guy on there is kinda cute.  The song of today is gonna be whatever song that comes on the radio that i like best.. maybe..  after this i\'m probably gonna play the sims while music vids download.  Well, it\'s still commercial on the radio so I\'ve decided to just put the lyrics for the first few tracks on the new Drowning Pool cd on there.  But I left Step Up out of there..  still a good song though.  Oh, commercial break over, but it\'s Seether & Amy Lee-Broken.  They completely overplay this song and ruined it.  It was ruined once amy lee started singing in it though.  Even though that was supposed to be the original but they never got around recording it until the PUNISHER album came along.  The one with just seether is the best version.  so i guess i\'m gonna play the sims now after i check my group things to see any new messages anybody left if they left any. \"Think\"And just before you fade out to I will hold a place for you Do you remember when we used to Do you remember when you used to Do you remember life Well, as you\'re led out to the slaughter You ought to Think about your life you wouldn\'t give it up Before you\'re led off to the slaughter You ought to And just as soon as you will find That all of this became your life Do you remember when we used to Do you remember when you used to Do you remember life Before you\'re led off to the slaughter You ought to And just as soon as all of this so All of this so pure Do you ever stop Fucking think\"Numb\"Numb To the man looking back at me Said why don\'t you go away We\'ve all heard your misery You\'ve been bleeding it for days end And it\'s plain when I look into your eyes I see All you want is everybody\'s sympathy So why don\'t you take all of this and go away So why do you always look to me this way When it seems like tomorrow might not come And it seems like the sorrow has left you numb And to the world I see Said we\'ve sat and watched you shine Pack up all your guilt and please Leave it all behind And it\'s plain when I look into your eyes I see This could be the way you want your life to be Numb Now your gonna bleed So why don\'t you just tell me\"This Life\"And when it\'s cold out I will wait for her And I will roam out To the edges of the earth And when it feels just like the rain Has soaked through my bones None of this will matter None of this will save you from yourself This life It\'s bringing on the death of me Swear that I can barely see That I\'m the one to blame This life I swear to God it\'s killin\' me Seems some times I can\'t believe That I\'m the one to blame And when it\'s cold out In the sleet and snow I will hold out \'Till the answers all unfold And when it feels just like the rain You will see through it all None of this will matter \'Cause none of this will save you from yourself None of this will matter None of this will save your from yourself\"Nothingness\"And this blood on my hands Now it won\'t go away I\'ve been out on this road for days And the fear on the wall Watch it slowly cascade As the door holds the wolves at bay The emptiness is building from The nothingness that we\'ve all become The emptiness is building from The nothingness that we\'ve all become And this road that I\'ve chose Seems it\'s taken its\' toll Leading far far away from the soul And the fear on the wall Watch it slowly cascade As the door holds the wolves at bay The emptiness is building from The nothingness that we\'ve all become The emptiness is building from The nothingness that we\'ve all become This thing I\'ve become seems its\' taken control Of my mind and my body my will and my soul Driving further and further away from my self Until there\'s none of me left I\'ve become something else The emptiness is building from The nothingness that we\'ve all become   '),(18117,3851875,'08,July,2004','2004-07-08 00:00:00','   I feel kind of guilty after my last posting. It\'s funny, you know, because I didn\'t write about anything that had not been on my mind for months. But I guess that, somehow, put feelings in writing equals to fully acknowledging them. There\'s no way to deny that you felt like this or thought about that if you have written about it; contrary to thoughts and ideas, the written word stays there, showing what you are.  I had the impulse of deleting the whole post, the whole blog. I thought that maybe I should write only about everyday things and keep my feelings to myself. But then thought that that was ridiculous. I mean, one can be pretty anonymous around here, if that\'s what one wants. And the chances of someone from my real world stumbling here and identifying me are nil. And for all I knew, maybe no one would ever read this blog. But I felt guilty about having written about the ex, and about my feelings for my husband, for anyone anywhere to read it. Maybe I felt as if I was betraying him.  I suppose this is the same feeling that has kept me from seeing a shrink. I have always been a very discreet person, who would only confide in a few friends of many many years. The idea of talking about my most inner secrets to someone unknown is still a shock for me. Now that I think about it, I even wonder how did I utlimately make the decision of starting this blog. I guess I was pretty sure that no one would read it!   '),(18118,3851875,'07,July,2004','2004-07-07 00:00:00','   So, an ex-boyfriend called today. Not any ex, mind you, but  The Ex . I suppose I shouldn\'t make a fuss over it, and in some way, I don\'t. He just called to say hi (FYI, we broke up back in 1993, and after months of (me) being heartbroken, we managed to remain best friends. Then I got married, he got married and moved out of town, and we sort of lost contact). He\'s been married for 2 years. I\'ve been married for almost 5. But to be perfectly honest, I asked him if he was very happy, and was kind of hoping that he would give me the great news that they had split or something. No such luck...  By now, one would be wondering what on earth do I think I\'m doing. Well, to be honest, I\'m not happily married. Not that I\'m in a living hell or anything, but I\'m just not happy, plain and simple (or rocky and complicated, it depends on your take!) I\'m not unhappy enough to end my marriage  (sometimes I believe that later I\'ll regret my lack of decision, though) , but I\'m sure that if I had a chance with someone else, I would give serious thought to that option. What a bitch.  Well, this guy was the love of my life, you see. I never ever loved anyone the way I loved him. On top of that, we had the most amazing chemistry I\'ve ever seen. I\'d never felt like that before, and I have not felt like that since.  Just for the record, I did love my husband when we got married. But too much has happened, so for now I only feel affection for him, that\'s all. No love, no sexual attraction. Actually, when you think about it, it\'s quite depressing...!  I guess I\'ll be writing more on this later. Who knows, maybe putting these things on \"paper\" will help me. My oh my, what a first posting!   '),(18119,3851875,'06,July,2004','2004-07-06 00:00:00','   My first posting, just for testing purposes...   '),(18120,1826527,'27,March,2003','2003-03-27 00:00:00','    Much respect to  brocore818  from the  urlLink No Motiv  message board for finding this fun and addictive  urlLink game .  Joining the  urlLink Hey Mercedes mailing list  has its perks.  What cool guys.  They sent those on the list a secret link to an mp3.  Word!  I\'m sensing this summer will hopefully be filled with good musical tours for my ears.  Hey Mercedes, No Motiv, Ozma, and The Living End are all putting out new CDs and touring!  Let\'s just hope they all don\'t come to the area and play at different venues on the same night.  I don\'t believe it is almost April.  My birthday\'s in April, but I really am looking forward to nicer weather rather than a birthday that will be peppered with a cake and having to work.  Bring on the allergies.   '),(18121,1826527,'24,March,2003','2003-03-24 00:00:00','    Exit 135.  I\'m almost there, then  BRAKE .  I left my house this morning for an interview.  I was on the parkway and coasting with ease southbound down the never-ending road.  I hear on my radio, dial set at  urlLink 880 am , that there is a 2-car accident around exit 135 in Clark, backed up for a mile and a half, with only one lane open.  Well, it sounds not good, but it\'s only backed up a little.  Well, when I got to milepost 137, a sea of cars formed a wall to my interview, which was set to begin in Brick Township at 10:45 am.  It is 9:10 am.  Ok, there\'s still plenty of time.  As I sit motionless in my car for a half hour, I try tuning in my radio to the traffic report, only there\'s a war briefing on every station, including sports stations.  So I try my hand at the FM stations, who only mention that there are 40 minute delays going into the Lincoln Tunnel and seem to think this Garden State Parkway does not exist.  I keep the radio on, hoping to hear something soon.  At about 9:40, they break away from the war briefing to tell me, there\'s a 10 mile backup on the Garden State Parkway South from 131 to 141.  Also, there\'s another accident at milepost 142, and another at the Raritan toll plaza.  Arrrrrrrrrrrrrrrrrrrrrgh!  Approximately 18 minutes later, the parking lot begins to see movement and we\'re off, but cautiously.  As we trudge along, I see lines of cars pulled over from overheating, hoping mine will not follow suit.  Then, my lane decides to stop.  Only my lane.  I notice a cop car stopped in my lane up ahead, so the folk in front of me pull into the right or left lane to pass.  As I finally get into the left lane, the cop floors it and off he goes.  Why was he stopped?  Did he drop his donut?  Does he have narcolepsy?  Was he trying to make another accident?  We may never know.         Well, I get to exit 112 and notice it\'s 10:20.  I will probably be late.  I call and tell them the story and they are quite understanding.  I ended up getting there at 10:55, only 10 minutes late after all that.  The interview consisted of the assistant superintendent and 2 principals.  I felt like a monkey in the zoo.  Three people staring intently at me.  When do I get to throw my dookie at them?  After deflecting questions, the assistant superintendent tells me I did well, and wants me to fill out paperwork to sub, so if I were to get hired, I could start immediately.  I fill out papers and this takes another 25 minutes.  It is now 12:15.  The assistant superintendent asks if I have to be anywhere.  Other than Aruba, I say no.  She wants me to visit a school to talk to a teacher.  I say ok, because that is a good sign.  I get directions to the school and end up waiting for 10 minutes.  The principal comes out and hey, she decides to fire 20 questions at me.  After that fiasco, I ask for direction to GSP North.  Get me outta here.  It\'s after 1 and I am famished.  The principal doesn\'t know how to get there, neither does the secretary.  Another teacher tells me, but gives me the wrong directions, and I notice I am headed south.  I pull a U-turn, and finally see signs for the parkway.  I get home at 2:30.  Monday, great day!      Oh yes, I posted my  urlLink philosophy final .  If you feel like thinking read it.  I actually think this was one of my finals I actually deserved the \"A\".  Peace out.   '),(18122,1826527,'22,March,2003','2003-03-22 00:00:00','    Oh yeah.  One last thing.  I was searching for some college papers and stumbled across my philosophy final.  I was impressed by how clever and philosophical I was.  I will post my work in the near future when my brain is less fuzzy and my eyes are fully open.   '),(18123,1826527,'19,March,2003','2003-03-19 00:00:00','    Things are finally looking up for now!  After sending resumes to  71  districts, yes that is  71  districts, I am finally starting to hear from a few.  After the interview I had on the 19th, I got a call yesterday for a job in  urlLink this place .  It would also be a permanent sub position from now until the end of june, but it\'s another possibility.  My tournament bracket is a joke as usual.  So far, I have 4 sweet 16 teams already out, including missouri and illinois.  So now my focus will shift fully to the cuse as a fan, and not as some bracket-obsessed drunk.  I talked to  urlLink brandon  tonight, and he has finally left the black hole of music stores.  I\'m happy for him.  It\'s officially spring, and that means my nose begins to close up.  Thank you allergies.  Beautiful days and breezy nights are hindered by my inability to perform necessary respiration.   '),(18124,1826527,'19,March,2003','2003-03-19 00:00:00','    Oh yeah!  I finally had another interview today.  It was  urlLink here .  It would only be for two months, but I don\'t care.  It is a semi-permanent job in that it is going to be the same class every day.  It would be in a 4th grade, which would be cool as well.  I\'m keeping my fingers crossed.  I just got the new  urlLink snoop  cd.  He is great!  Also,  urlLink tyronn lue  is the dumbest man on the planet, comparing the washington wizards\' playoff run to the war in iraq.  Get a clue buddy.  Meanwhile, the sixers are playing real good basketball.  They actually have a chance at the #1 seed in the east.  As for the nets, well I realized I really don\'t like them.  Especially Kenyon Marin and Richard Jefferson.  Richard Jefferson whines about every call and everyone raves about his defense, but he pushes off and hacks every time.  Whatever.  In baseball, the Jays locked up two young all-stars in Vernon Wells and Erick Hinske.  I hope they wipe the floor with the yankees on opening day.  I believe south park\'s new season starts tonight at 10.  sweet.   '),(18125,1826527,'14,March,2003','2003-03-14 00:00:00','    I am not engaged!  Was  urlLink sunset/sunrise  that convincing?  Wow!  I am flattered.  Good news for me=there is a job fair for Highlands at Matawan Regional High School on Thursday, April 3rd.  No Motiv may be touring around May/June.  Details when I find out.  I considered my  urlLink future in teaching  when I cannot seem to find a job in NJ.  Speaking of South Carolina, I bought tickets to a  urlLink minor league team  for my much anticipated trip to myrtle beach in July.  Only $14 for 2 tickets is my idea of baseball.  Who needs yankee stadium and the fun thrills of the bronx when you can have  urlLink this place .  Also, since we\'re flying down, we would have to rent a car.  Well, it is rather unfair that they insist on charging you up the ass for being one year under the 25 age restriction.  Anywhere from $10 a day to $25 a day.  Not to toot my own horn, but I may be only 23 right now, but I\'m probably a safer driver than many many people over the age of 25, but I have to pay extra.  People who have totalled 3 cars but are 26 don\'t have to pay these fees.  Now that\'s fair.   '),(18126,1826527,'05,March,2003','2003-03-05 00:00:00','    Okay, so i have written a new song entitled  urlLink the perfect storm .  I have done some things with this webpage.  I\'ve combined some pages, deleted some, and changed the layout a bit.  If you are reading this, then you know where my updates will be from now on.  mess around with my page a little.  let me know if you like it or not.  give me some  urlLink feedback .  it would be appreciated, whether positive or negative.  i\'m not a technical genius, and this isn\'t my full-time job, but i want the site to be somewhat fun for those few who visit.   '),(18127,1826527,'04,March,2003','2003-03-04 00:00:00','    i wrote a song the other day, but i\'m too tired from messing with my webpage, so i will post it tomorrow.  more to come...   '),(18128,1826527,'04,March,2003','2003-03-04 00:00:00','    testing a new toy for my morphing webpage...   '),(18129,1826527,'24,April,2003','2003-04-24 00:00:00','    So I\'ve taken a slight vacation from my blog, but it\'s for good reason.  Much work and planning.  Much pain.  Not much free time.  The job is coming along nicely.  It\'s difficult taking over a class 7 months after it has started, but I\'m slowly making progress.  The students are starting to show more respect for my authoritay, and I\'m just happy because I didn\'t have to do much today when I got home, probably because I was working straight through last night until 9:30.  It feels nice to not do stuff on a computer that\'s work-related...My birthday is tomorrow, and my wish is not coming true.  It\'s supposed to rain at night, and on Saturday, and be in the 40s.  Happy f*!@ing spring!  Along with this, there was supposed to be a  GHT  game on Saturday, which may get cancelled for rain again.  I\'ll probably be content for the next two months because I\'m working and earning a paycheck, unfortunately, I will join the millions of unemployed come June 20.  Bye.   '),(18130,1826527,'17,April,2003','2003-04-17 00:00:00','    I now know that I am employed because I have all of the symptoms:  dizziness, vomiting, migraines, insomnia.  Wait, I think I just described the symptoms of ebola.  No, in all seriousness, I am really happy to have some sort of steady job, even though it is only for 2 months.  It feels nice.  The people are all nice, and the students seem really fun. I\'m actually proud of myself for the first time in a while.  My first 3 days are over, with an extended weekend coming up.  The true test will be to see if I survive a full week.  Ask me on my birthday.  (nudge, nudge, wink)   So far this season, the Jays are 1-6 against the yankees.  Not the kind of good start they needed.  Mark it down now-season over.  No playoffs.  I got the new  urlLink Pete Yorn  cd yesterday.  For some reason, I like his music, and if you liked his first cd, you\'ll probably like this one too.  Same formula-songwriter who plays guitar with mostly good songs, and 1 or 2 that aren\'t very good.   '),(18131,1826527,'14,April,2003','2003-04-14 00:00:00','    I officially found out today that I have a job!  I start tomorrow, and I cannot wait.  It\'s only for 2 months, but it\'s a good start.  I\'m still hoping I can find a nice place to start in September as well.  Who knows, maybe I can stick around in this district if something else opens up anytime soon.  My resumes have been mailed to  82  different districts.  Are there any other towns in New Jersey?  The weekend is over, and it is really nice out today.  I\'ll probably play some basketball in my yard.  I got my hoop about 8 years ago, with my first paycheck from the deli.  It\'s a nice hoop-one that can adjust the height, and it has wheels if you want to move it.  The only problem with my yard is that it is on the grass because I have no driveway, and there are tree roots everywhere.  It\'s kind of like extreme basketball because it\'s like dribbling through an obstacle course, but shooting is still fun.  You\'d think with all the stuff in my yard, I would be a good ball-handler because it would make me more aware and consistent, but that\'s not true.  I have the body of a point guard with hands of steel.  Oh well, I\'m not going to the  urlLink NBA  or anything, so I don\'t care.    I\'ve needed gas in my car since Friday morning, but I still haven\'t gotten any.  I find it very annoying getting gas.  I don\'t know what it is.  Maybe it\'s because it costs $1.69/gallon.  Whatever, I just walked a lot more lately.   '),(18132,1826527,'11,April,2003','2003-04-11 00:00:00','    Normally, I would be complaining about this rain, but I\'ve decided against this and embraced the rain.  I would much rather have rain in April than snow.  I had to go fill out some paperwork for my job possibility today, and it was no fun driving on 280 with trucks flying along side you and splattering more water on your windshield, when visibility is poor already.  Alas, I am home for now, until I have to go pick up my very awesome girlfriend from school and bring her to work.  It\'s the weekend, and I am happy, because it is supposed to be in the 60s on Sunday, and warmer to start next week.  I watched the movie  Traffic  last night, but I wasn\'t too impressed by it.  It seemed very un-original and too long.  Benecio del Toro\'s part would have been much better in my eyes if I understood Spanish better, so I wouldn\'t have to read what he was saying.  Call me crazy, but every Michael Douglas movie I see has the same plot...or he just plays the same character over and over.   '),(18133,1826527,'10,April,2003','2003-04-10 00:00:00','    I\'m home again today.  It\'s for a sad note-there\'s a memorial service for my grandfather today...I also found out I will not be working tomorrow either.  My  urlLink honey\'s  car is having problems, so the weezer mobile gets a call from the bullpen.  That\'s my car in case you haven\'t figured it out.  I will be escorting her to the mechanic, school, and then to work.  So I\'ll be on call.  It will make me feel special, like a doctor or something.  So this puts my days working in April to 2, and that\'s with 2 different jobs!  I\'m slyly smiling, however, because I absolutely despise one of these jobs.  Still haven\'t heard from my possible replacement job...  Jays are in 2nd place!  Woo-hoo!   urlLink Tanyon Sturtze  is 2-0.  He only needs 2 more wins to match last year\'s totals, and 18 more losses too!  I subscribed to  urlLink SI  in order to receive my Cuse package.  I  do  like the magazine, but I guess I never got it because I thought they devoted too much space to golf and nascar.  If I lived in South Carolina, I\'d probably not mind as much, but we don\'t have many race tracks around these here parts.   '),(18134,1826527,'09,April,2003','2003-04-09 00:00:00','    Hey.  I\'m still happy.  Orange you?  dar-har-har!  I\'ve been somewhat lucky with my sports teams winning championships, or at least being there.  The Rangers won a cup in 94, the Jays won back-to-back World Series in 92 & 93, the Bengals were in the SuperBowl in 89, and the Sixers were in the NBA Finals only 2 years ago, and the Orange made it to the championship game in 96.  I\'ve realized that I feel much more happiness for college kids winning.  It\'s much more special to them, because they\'re not getting millions of dollars(yet), and this team that won wasn\'t a bunch of cocky jocks.  They are all good kids and team players.  So here\'s to the Cuse...  In addition, the Jays recently swept the Twins and beat the Sox last nite.  Good comeback after starting the season 0-3.  The Sixers are still in the running for first in the east as well.  I still haven\'t heard from my job possibility.  I really hope this doesn\'t fold.  I don\'t think I can take rejection...I also think I am sick of this cold weather.  It\'s now April 9th and in the forties, with snow still on the ground...The  urlLink GHT  Game of the Week for Saturday has already been cancelled because of expected rain Friday night, and the cold weather all week, which will not allow the fields to dry out...  I started out my morning by turning on the TV at 10:30. ( I didn\'t get a call to sub again)  I turned on  urlLink  CBS news  and was drawn in to the coverage of the war.  I saw US Marines tearing down a statue of the madman with Iraqi people stomping on its remains and dragging the head down the streets.  Reporter Byron Pitts was with Marines fighting street battles in Baghdad and his report was outstanding.  He was giving us coverage as the fighting was taking place, and it was even better than a Hollywood war movie.   '),(18135,1826527,'08,April,2003','2003-04-08 00:00:00','    Woo!  The  urlLink Orange  are the NCAA champs!  A happy moment for all that are devoted to the cuse.  I\'m happy for Jim Boeheim and the team.  I also respect Roy Williams, and hope he\'s the next to win a title.  By the way, he had the best comment of the night.  After the game, Bonnie Bernstein asked him about taking the North Carolina job, and he said \"...I don\'t give a shit about North Carolina right now...\" on live TV.  That\'s priceless.  She got what she deserved.  Damn media.  Will write more tomorrow.  It\'s late.   '),(18136,1826527,'04,April,2003','2003-04-04 00:00:00','    Well, there\'s good news for me.  The principal has recommended me for the job, but I still have to wait on the superintendent and board of ed to approve me.  If all goes well, I start on April 15 and end on June 20 or so, their last day of school.  The job fair yesterday was a joke.  Advertised online were 9 schools, with all needing elementary teachers for september.  I get there and pay my $7 and they give me this packet.  The packet lists the schools that are there and the jobs they will be offering.  All 9 schools are listed as advertised.  OK.  Then as I read through the jobs, only one of these schools is looking for elementary teachers, and one of the schools listed actually is not there because they apparently do not have any open positions.  Strange, considering they were one of the 2 schools sponsoring the job fair.    Anyway, I proceded to the one listed school and the guy who talked to me seemed very uninterested in being there, and avoided eye contact with  everyone  he talked to.  Joke #1.  I was tempted to leave, but I drove 40 minutes and paid money for this thing, so I decided to go to some other schools just to give them my resume and stuff.  At the second line I waited for, the guy said wow you had a far drive(looking at my resume), and I said, \"no it wasn\'t too bad, about a half hour or so.\"  He then gives me an application and says \"have this filled out so when we start interviews, you\'ll have it ready.\"  Joke #2.  I then went over to my third attempt.   urlLink This place  seemed great.  The principal I talked to was excellent.  He was friendly and informed me that while there were not any postings for elementary teachers, there were indeed other elementary schools in the district that were looking.  He and the other principal were focusing on art and librarians, but he said the other elementary schools need teachers, and he would pass my resume to them.  Thank you sir for being a genuine human being.  It didn\'t seem like there were many there.  After that, I went to my fourth and final line.  It was for  urlLink this place .  Again, the gentleman that talked with me was pleasant and encouraging, telling me that I would be likely to have an interview at least in the near future in his district.  He told me I was at an advantage because they need male teachers.  So these two folks gave me at least a bit of hope from the experience, despite the rest of it being rather crappy.  So I\'m still searching for a job for September, but I\'m told it\'s still early for most districts to interview for next year.  I\'m also looking for summer employment, because I refuse to work in a camp, and will never go back to retail.   EVER!  It\'s the weekend, but of course it\'s raining.   '),(18137,1826527,'02,April,2003','2003-04-02 00:00:00','    The lesson went well today.  I\'m supposed to find out by tomorrow afternoon already whether I\'ve got the job or not.  Unfortunately, I won\'t be home tomorrow because I\'m going to a job fair.  Hmm...I\'ve got crazy links listed over there, so use them if you\'d like.  Some are fun, and some are sites you probably won\'t even like, but you never know until you try.  So far, the Jays have been outscored 18-5 in two games.  I think that pitching is not their strong point.  Maybe I shouldn\'t watch the game tonight, because I\'ve watched the other 2 unfortunately.  Yesterday afternoon at about 2:30, it began snowing in a way that I thought we were going to have another blizzard, but it stopped.  Today, it\'s about sixty out, and tomorrow and Friday may reach 70.  Huh?  The  urlLink GHT  baseball season starts on Saturday, and it will be welcomed with 50 degrees and hurricane conditions.  Lots of wind and rain.  Hey, what\'s baseball season at  GHT  without this weather?  I think I may be adventurous and do something to my  urlLink web page , like add a new section or give it another facelift.  Maybe.   '),(18138,1826527,'01,April,2003','2003-04-01 00:00:00','    Happy April Fools Day Fool!  It was officially spring about 10 days or so, but to me, spring officially started yesterday, because it was opening day for baseball.  Unfortunately, my  urlLink Blue Jays  came up short against the much-hated yankees.  The season is young.  Six more months of this.  Is that good or bad?  Tomorrow is follow-up interview time, and the teacher who is supposed to help me out isn\'t very good with her communication skills.  I\'ve called her 3 times, and the second time she called me back.  I\'m still waiting for her to return my other calls...I\'m going to watch the game now.   '),(18139,1826527,'23,May,2003','2003-05-23 00:00:00','    The week is over and the testing is over, and surprisingly it is  raining !  Memorial Day weekend is upon is, yet it\'s in the 50s.  Hmm.  Interesting.   Yeah, Nets.  Up 3-0 over the Pistons, whom I now despise because I thought they sucked and did not really deserve to win the Sixers series, though neither did the Sixers I guess.  And can someone tell me why the Draft Lottery was on during primetime?  Hey, if I want to watch a bunch of ping pong balls, I can watch the lottery every damn night at 7:59 or whatever it is, and with that lottery, I have a chance to win money.  What do I get with the other?  Oh, I get to see the horrible teams that will ruin the careers of several young athletes.  In other sports news, I was happy the Jays beat the Yankees last night, and my boy Shannon Stewart had a nice offensive game.  I hate the Yankees, but I like Bernie Williams, and that stinks that he got hurt.  Oh well.    Bring on Sunday and Monday and some good BBQ, because Saturday is going to be a washout.  Tonight calls for pizza and either watching the Family Guy DVD or the Jays game-the usual motionless, blimp-like Friday night.  What a way to kick off the weekend!   '),(18140,1826527,'21,May,2003','2003-05-21 00:00:00','    Today was a poop day overall because the testing consisted of two 25-minute sections and a 50 minute section.  It was long and boring.  Way too much for kids to do, especially when it\'s reading and writing. I was reading this book that I am supposed to read to my class every day called  The Cay  while they were testing. It\'s about a boy who gets shipwrecked on an island in the Caribbean during World War 2 with a man who worked on a boat.  They ended up becoming friends despite cultural differences and overcome lots of challenges and stuff.  It was a really good book.  I was surprised by it.  Anyway, math tomorrow and Friday and then it\'s done.  Just in time for the long weekend, in which it\'s supposed to rain again.  So the GHT games may be pushed back or cancelled.  It\'s par for the course for this baseball season.    Speaking of baseball,  MLB 2004  is a very fun and addictive game.  I guess in part because it\'s challenging and I haven\'t had a baseball video since the untouchable by comparison  Baseball Stars  for Nintendo.  What a classic that was!  Anyway, my record in a regular season schedule so far is 8-18.  It\'s pretty terrible, but I actually feel relieved when I can actually win a game, while in Baseball Stars, everyone on my team was powered up to their max in every category and I would 10-run rule the opposing team just about every game.    I had a really dirty dinner at  Olive Garden  tonight.  The dish was called the Mixed Grill and consisted of grilled chicken and Italian sausage with roasted potatoes and vegetables in a rosemary seasoning.  That was really good.  Then I came home and watched  the Simpsons  reruns.  As usual, it was the same episodes that I had seen no more than a month ago.  That really bothers me.  They\'ve made over 300 episodes of the show, but it seems they show the same episodes over and over.  Believe me, nothing puts a damper on my day more than turning on channel 5 at 6:30 or 7 pm, and seeing \"Treehouse of Horror,\" especially when it\'s May.   '),(18141,1826527,'20,May,2003','2003-05-20 00:00:00','    I saw the new  Matrix  movie, but thought it was rather dumb.  I loved the original movie, but I thought this one was corny and it took an hour and a half for anything to happen.  When something finally happened, it was a really good action scene, but otherwise nothing great.  The ending was also not so great.  I know the other movie is coming out in November, but the way it ended was abrupt.  I was disappointed, especially since I wanted to see it so badly.  Oh well.  I also decided to make a rather impulsive purchase on Saturday.  I now own Playstation 2.  The price dropped again, so I jumped on it.  I\'ve wanted it for a while, but never had the money for it.  I\'m pretty sure that some new, advanced video game system will be out soon, because every time I buy a new system, it\'s obsolete.  For instance, we got Nintendo just before Genesis came out.  We got Genesis just before Playstation.  I got Playstation and within a year or so, PS2 and Xbox came out.  We\'ll see what happens next.    I was looking through a Brookstone magazine the other day and saw that they have arcade game machines like Ms. Pac-Man and Centipede.  I\'ve always wanted to own one of those things when I have a house, but was saddened when the price was $3,000.  Yeesh.    The new testing has begun today in school.  It is very boring for me, and very much to ask 4th graders to sit still and not talk for nearly 2 hours for 4 straight days.  What can you do?   '),(18142,1826527,'17,May,2003','2003-05-17 00:00:00','    Hello people.  If anyone has read the blog from May 17th before today, I want to say I was stupid for posting something personal that I shouldn\'t have.  I was angry and was blinded by how this blog should have nothing to do with some things in my life.  So I wanted to say sorry mainly to my girlfriend.  This was a mistake.  I did not think of how it could affect the situation before I posted.   '),(18143,1826527,'17,May,2003','2003-05-17 00:00:00','    My first official  GHT  game occurred today, and it was a pretty good game.  Depaul beat Wayne Valley 5-1.  It didn\'t feel like the end of May because it was in the 50s, but baseball nonetheless.  So it\'s nice to get some extra income at least for this one game.  I\'m gonna go watch game 7 of the Kings/Mavs series now.   '),(18144,1826527,'14,May,2003','2003-05-14 00:00:00','    Our weekend starts on Wednesday .  It really does.  It\'s May 14th and I have yet to do a game with  GHT .  You see, this past Saturday would have been my first this season, but on my way up there, my car overheated two times, so I had to get towed.  Yippee.  It\'s back and functioning for now, but who knows what will strike next?  Well, there is a game this Saturday, so we\'ll see what happens.  If it doesn\'t rain, then I\'m sure something will prevent me from doing the game.  We shall see.    Only 24 more days of teaching left for me.  That\'s sounds like very little, especially since there are 4 days of standardized testing next week, there\'s a field trip in June, and the last 3 days are half days.  So it\'s actually a lot less I guess.    I really hope the  Sixers  can pull out a victory tonite to jump a game up on the Pistons.  I am very happy the Lakers are down, but I\'m sure the Spurs will find a way to lose the series. I ordered Sports Illustrated a while ago, but I\'m still waiting for my Syracuse basketball gift package.  My brain automatically shuts down after 4 pm, so I cannot write anything else intelligent. bye.   '),(18145,1826527,'05,May,2003','2003-05-05 00:00:00','    My car is green.  No, I did not get it painted from it\'s ultra-mellow gray, but it is covered in pollen or whatever you call that stuff that is just about everywhere now.  I  think  I\'m starting to get a little better, but not too much because my sinuses are clogged from allergies and this cold that I can\'t seem to shake off.  I stayed home all weekend literally, and only went out of the house yesterday to pick up my dinner from the pizza place, but I\'m still not 100 percent...I got a new air conditioner in my room, which is nice because it has a digital thermometer instead of \"hi and low.\"  I actually used it last night on a moderate setting, not because it was warm out, but for the past two weeks, I\'ve been getting a natural alarm clock courtesy of the annoying birds ouside my window at 5 am.  I didn\'t get that this morning; instead, I kept hearing the ac click on and off because it was getting too cold in my room.  I got home rather early from school today, basically because one of the other 4th grade teachers wasn\'t in this afternoon and the other was working with the 5th grades on their play.  So I did a few things and left at 3:10.  The earliest I\'ve ever left, and school ends at 2:25.  So I got home and did my  urlLink bowflex .  That monstrosity actually works, I will admit.  I\'m not going to look like those guys on the commercial, but that\'s because I don\'t take steroids and also because I don\'t eat as well as I should, but it definitely tones you.  I guess I was already skinny, so it\'s not like I want to lose weight.  I actually have been trying to gain weight, but it\'s damn hard.  Believe me, I eat plenty.  I\'ve had the thing for I guess a year now, and I\'ve gained maybe 10 pounds.  I\'m sure when I\'m older and fat I won\'t be complaining about not being able to gain weight.  I purchased tickets to see  urlLink No Motiv  over the weekend.  I\'ve been a fan of theirs for 5 years or so, and have never been able to see them.  Of course their only opening for another band, but it\'s better than nothing.   '),(18146,1826527,'03,May,2003','2003-05-03 00:00:00','    I have to say, I do not enjoy being sick, especially when I am sick at least 3 times a year.  But this really is no fun.  It\'s going around to every teacher in the building, so it was only a matter of time before I got it.  The sixers are moving on to round 2 of the playoffs.  It seems like round 1 took about 2 months.  Maybe it did.  I am sitting home resting on a perfectly good Saturday.  I should be working for GHT, but I don\'t want to pass this cold on to anyone else.  I just finished watching Memento, and I can\'t say I enjoyed it the second time around.  I guess because I knew what the outcome was.    On Thursday, I went to this multicultural event at my school.  It was a pretty cool idea, but it was really hot in the classrooms, and I was up there since 7:30 in the morning.  It was a 12 hour day, and I was already fighting the cold, so I checked out earlier than expected.  One cool thing I was able to do before nearly passing out was get my name written in 9 different languages.  It looks pretty nifty.  And the food was good.  I ate lots of stuff that I didn\'t know what it was, but it tasted good.  I\'m not really in a good mood.  Probably the cold and probably another reason involving my girlfriend not seeing me today.   '),(18147,1826527,'26,June,2003','2003-06-26 00:00:00','    It\'s officially a heatwave up in here.  I welcomed the heatwave by getting a really bad haircut.  I look like a Marine now, and I\'m not happy.  I guess the Spanish lady at Supercuts didn\'t understand when I said \"not too short on top please!\"  I have an interview Tuesday, so I hope it grows in a bit.  Although it is only 5 days away.  Since it was unbearable outside, I spent time making some changes/additions to my  urlLink website .  I added some new  urlLink pictures  and also some new  urlLink quotes .  Let me know if you like it or not.  Jogging again tomorrow.  Bright and early at 6:30 am.  Then it\'s off to the shore on Saturday.  Woo!   '),(18148,1826527,'24,June,2003','2003-06-24 00:00:00','    Vacation has officially started, and it is definitely summer.  It was somewhere in the mid 90s the last 2 days, and it is supposed to hit 95 tomorrow again.  Since I\'m unemployed and not doing much else this summer, I\'ve decided to take up jogging again, this time with the help of my lovely girlfriend.  Tomorrow the trek begins at 6:30 am.  We\'ll see how it goes.  We\'ll start at 3x a week, but maybe it will progress into more.  It\'s good I guess.  It keeps your heart healthy, and burns away some of that weight you want to get rid of or whatever.  So I will be going to bed shortly in order to get enough rest for my early rise tomorrow.  I also plan on selling back some old cds to my friend  CD World , and hopefully get enough credit to buy some new cds.  Here\'s what my list\'s possibilities include:  the new Cursive, Texas is the Reason, Miracle of 86, Nada Surf, OK Go, Mock Orange, White Stripes, and old Sugar Ray.  I\'m a sucker for those groovin\' sounds of Sugar Ray, so sue me. (except their new cd is mung).  Other than that, I\'ll probably do some Bowflex tomorrow sometime, and the rest of the week consists of getting a haircut, cleaning my room, playing some basketball, and making a mix tape.  Yeah, that\'s right, a mix  tape .  I don\'t have a cd burner like all of you technologically advanced folks.  It also sucks because if I make a mix tape, I don\'t have a tape player in my car.  Oh the headache!  In other news, my car is doing it\'s thing again.  The check engine light is randomly coming on and then disappearing again, and my thermostat gets really high, but doesn\'t go into the red area, and thus does not overheat, but it seems too high to me.  The last 2 times I\'ve brought it to the mechanic, they\'ve said, \"I don\'t see anything wrong with it.\"  Now if I had a light that said something like \"check heartbeat,\" I don\'t think my doctor would be like \"well, I looked but I don\'t  see  anything wrong.  Those guys should examine my car if they\'re going to have it all day and find  something !  Make it up at least.  I won\'t know.  Anyway, let\'s see if I feel like taking it in again tomorrow.   '),(18149,1826527,'19,June,2003','2003-06-19 00:00:00','    School\'s out!  Well, I still have to go back tomorrow to finish a few things, but no more students or teaching.  I still don\'t have a job for September yet and that sucks.  So I guess my summer will technically start Saturday.  Next weekend I\'m going down the shore.  I can\'t wait.  I\'m so tired of here.  Even if it rains, which I\'m sure it will, at least it\'s not here.    I\'m going to karaoke night tonite.  Should be interesting.  I probably won\'t be able to drink much because a) I have to drive, and b)I don\'t want to add to my already tired-ness.  But it will be nice to \"unwind\" nonetheless.    So much has happened since my last post.  I saw No Motiv finally last Weds nite.  They were good, but they didn\'t headline, so they only played for like 40 minutes.  Then we left before the crappy Rx Bandits started their set.  This show compelled me to finally purchase one of NM\'s original 2 CDs, and it\'s a good cd.  It\'s a lot more edgy than their last 2 cds.    Let\'s see what else...I went on an interview in the school I was working for a first grade opening, which did not go well because I have no idea how to teach first grade.  On the upside, however, there may now be a 4th grade opening, so I do have a slim chance at staying at the school, but most likely not.    I guess not much else really has happened.  I changed a little bit of my  urlLink site , but not a whole lot.  I guess I have to find some more pictures and fun stuff to put there soon.  Is anyone else tired of hearing about Ashton Kutcher?  I don\'t care if he\'s screwing 40 year old Demi Moore, who\'s infested with venereal disease.  I just don\'t find her attractive.  Maybe it\'s because I fear she would try to gnaw on me with her rather large jaws.    About a month until vacation.  Now that\'s what I\'m dreaming of day in and day out.   '),(18150,1826527,'08,June,2003','2003-06-08 00:00:00','    It\'s one week later and things still aren\'t right...I just want to go away.   '),(18151,1826527,'01,June,2003','2003-06-01 00:00:00','    Welcome, June!  It\'s still raining though.    It\'s a new month, but something in my life still does not feel right.  And I don\'t know what the outcome is going to be.  Just when you think you have things figured out, it occurs to you that your problem is not solved, and you don\'t really know what to do about it, except maybe put an end to it and try something new.   '),(18152,1826527,'31,July,2003','2003-07-31 00:00:00','    Hmmm.. what can go wrong with Day 4?  How about EVERYTHING?!?!   '),(18153,1826527,'30,July,2003','2003-07-30 00:00:00','    Day 3  07.22.2003  Tuesday  We arise from bed to eat our Cheerios and decide to head over to  urlLink Barefoot Landing , which is a big slew of stores and restaurants on this swamp filled with alligators and dragonflies. We walk around some shops at first, and we purchase a pair of sunglasses.  I\'m not really a fan of wearing them, but you really needed them down there.  It was awfully hazy.    We continue in search of  urlLink Alligator Adventure , which is a zoo-type place filled with bears.  No seriously, it\'s filled with alligators, but they kept advertising this bear show.  Huh?  We pay for tix, which were an economical $12 each even after my trusty coupon.  The place is very uneventful, as every alligator, crocodile, and turtle were dead to the world, because it was hot as balls out.  We took pictures of a bunch of immobile reptiles.  We saw 2 albino alligators, which were outstandingly creepy.  We then went over to the bleachers at 12:45 because they were supposed to feed the reptiles at 1 pm in a show.  So we wait     and      wait.  Then this employee struts over and says \"uh, the feeding is taking place at the main boardwalk today.\"  Well, that\'s nice to know now.  We unemphatically trudge to the main area and immediately turn around for the exit.  I believe everyone in that place was huddled over the fence to watch the feeding, so we had no chance of seeing a glimpse of anything.  We flee.    Hungry.  We are relatively close to the  urlLink House of Blues .  I wasn\'t even aware they served lunch, but Kim guessed they did.  Luckily, she was right.  It\'s pretty cool inside.  I had never been to one before.  I had this pulled pork bbq sandwich with sweet potato fries.  Good stuff.  In the process of eating, I pull out the map of the area, and we decide we should give it another chance and hit the shops across the lagoon thing.  After lunch, we drove around to the other side and walked around a bit, and realized that indeed, this place did suck.  We left.  I have no recollection of the late afternoon.  I remember we tried to go to the Outback Steakhouse for dinner, but found they had a 55 minute wait, and there was a line of people wrapped around the perimeter.  Oh crap, let\'s go elsewhere.  We get in the car and fade into the horizon down King\'s Highway as the rain begins pelting our windshield.  Everywhere had a line, so we hit good ole\' Pizza Hut.  No wait.  We\'re in.  Everyone stares at us.  I stare back and give some really evil looks.  We eat half sausage/half plain stuffed crust pizza.  I have to relieve myself of the 3 iced teas I have.  On my way to the restroom, a weasel-looking lady in line seems reluctant to move and stares at me.  I pass through and on my way out I encounter the rodent woman again.  \"Excuse me please.\"  No movement.  I burst through her and give her a glare.  \"I said excuse me.\"    It\'s still raining, so we can\'t do anything outdoors, like swim in the pool or play some mini golf.  I say, hey let\'s try the mall by our hotel.  It\'s Tuesday, and the new  urlLink 311  cd came out today.  We enter and see a FYE store-\"For Your Excrement.\"  This store sucks.  I hate the ones up here, so I know I\'ll hate this one.  I ask the plump clerk at the counter if they have the new 311, since it was nowhere to be found on their new release shelf or in the section.  \"Three-leven?\"  \"Nah, the only new releases we gawt are Now thirrrtain and (some other crap I don\'t remember).\"  Wow, you\'ve been a help.  Bye.  We continued through the mall to find  NOTHING  of interest.  We leave.  Back to the hotel room.  We drink some JD and eat pretzels and peanuts and watch Road Trip on TV.  Tomorrow we\'ll try the Outlet stores.  There\'s a whole mess of them, and at night I have tickets to see the  urlLink Myrtle Beach Pelicans , a minor league baseball team.  See if you can find out what will go wrong.   '),(18154,1826527,'30,July,2003','2003-07-30 00:00:00','    Greetings!  I am Kim, Brian\'s girlfriend.  Bri was nice enough to let me join his blogger and I\'m stoked because I\'ve never done this kind of thing before.  So everything Brian has said about vacation is true.  I truly believe that the generator incident was an omen to not go on this trip to Myrtle Beach.  But all I can say is thank God the generator didn\'t shut down while we were in the air.  That could have been very tragic and very messy.  The mold in the room thing is true as well but Brian forgot to mention one thing - I am allergic to mold.   I did not learn of this allergy until this trip so I guess I learned something useful.    Bri also failed to mention the nasty things we came upon in the room besides the pool in front of the AC and the mold.  Now this may be a chick thing but I was grossed out by the towels and my pillow.  The towels had black and yellow stains all over them and i was not about to use them on my ass.  One of the washcloths I saw had hair on it and I\'m not talking about the ones on top of your hair.  YUCK!!!!  My pillow had some kind of red stain on it(blood?).  So me and Bri wound up sharing a pillow and bumped heads twice during the night.    Much more horror to come!   '),(18155,1826527,'29,July,2003','2003-07-29 00:00:00','    Day 2 of our trip... 07.21.2003  Monday  We had planned to do all of the Ripley\'s stuff today.(Like the Believe it or Not! Museum, etc.)  I had purchased a combo pass to four of its attractions a while back on the Internet.  The included things to do were:   urlLink Ripley\'s Aquarium , Believe It or Not Museum, Haunted Adventure, and Moving 4-D Theater.  I was informed to pick up the passes for all events at the aquarium, so that led us logically to the aquarium first.  We started with a breakfast at the hotel eatery, which was ok, but I noticed their OJ sucked.  Anyway, we made our way to the aquarium, where we found proof that the people down there were rude, as they cut in front of us with families of 8, bumped into us without saying \"excuse me,\" and stared and stared at us like we were living in the 1800s because of our mixed cultures.  Ignoring these issues, the aquarium was a pretty cool place.  It had many fish, rays, turtles, and sharks.  Its main attraction was a tunnel with a motorized walkway which led you through a sea of sharks and fish.  The sharks were swimming all around you, and you could see their teeth up close.  Needless to say we used our first roll of film at this place.  Of course, you had to exit through the gift shop, so we walked through and glanced around, and then the power failed.  It gave me an image of a Jurassic Park-esque adventure.  The sharks had been scheming to take over this place and escape into town.  They were tired of tourists gawking at them.  But the backup generator took over and lights on.  We left.  On to attraction #2.  The museum.  The next 3 places were all located together, but at a different location, so we had to drive a bit to get there.  We hit the museum, and it was what I expected.  It was kind of hokey-pokey in that it was cheesy like that tv show, but it did have some cool things in it.  But there\'s only so much I could take of seeing the bearded woman or the cow with an arm growing out of it\'s back.  Freak show.  We walked slowly down the road to our #3 stop, the Haunted Adventure.  This place was just a haunted house that you may see around Halloween, but it was pretty neat.  It started in an old elevator stuffed with 8 people.  You are then instructed by the guy in makeup to form a chain and walk through the house, which we did.  Much of it was total darkness, so it was difficult.  A few little scares, but nothing major.    A few door down was the last place, the 4-D Moving Theater.  We went in to this waiting room with about 20 other people with nothing to do but wait for the next showing.  They had this tv that kept spooling the safety rules over and over and over again.  It was poison for your eyes.  Finally, we were let in after about 20 minutes of waiting.  We were definitely looking forward to this part because they have one of these in Universal Studios, and it\'s like the Terminator coming at you and there\'s smoke and mist.  It ended up being lame.  We were a log that moved through a mine on one of those mining carts, and it looked like the tracks were a roller coaster, and we, as the log occasionally fell into water and got mist-ed.  Why a log?  Couldn\'t they come up with a more interesting idea?    We left the final attraction and saw a gift shop across the street known only as \"The Gay Dolphin.\"  I wondered what definition of the word gay was implied.  I still don\'t know, but the shop was garbage.    We left for the hotel, but decided we needed some snacks and drinks for our room, so we cruised down King\'s Highway and found a CVS.  We snagged Cheerios, bowls, spoons, milk, OJ, iced tea, a big-ass jug of water, sweet roasted nuts, and a bag of pretzels.  We headed back to our room with a plentiful booty.  Yarr.  As usual, we did not know where to eat lunch, so we utilized the in-room guest book and found a tiny Italian place.  We jumped there and ate salad, raviolis, and a broccoli roll.  Pretty good food.  Since it was still a nice day, although a bit late, we decided to head to the beach.  It was about 4, and I figured I wouldn\'t get burned by the sun at this time as badly as noon.  We got set up on our spot around 4:45, and hit the water, which was warm and comfortable.  We stayed in wave after wave until almost 8, when we decided we should probably eat some dinner.  We got back to our room, showered, and found out it was 8:30 and we didn\'t want to eat anything big, so we\'d get BK.  I first hit Green\'s Discount Beverages for some JD coolers-Lynchburg Lemonade and Hurricane Punch...Well, we waited in the drive thru for about 20 minutes and then when we got our food, my chicken sandwich was blackened, and we were missing a fry.  Bloody savages.  This resulted in my kicking and cussing back at the room.  Thus concludes day #2 with a pissy outlook for the rest of the week based on rain in the forecast and the ending of day 2.   '),(18156,1826527,'29,July,2003','2003-07-29 00:00:00','    My, it\'s been a while since blogging, partly due to vacation which occurred last week.  There\'s plenty of stories to tell, so I will do this as a recap-by-day blog.  For example, I will blog our first day today, our second day tomorrow, and so on.  Here goes.  Enjoy...  07.20.2003  Sunday  Wake up at 5:30 am.  Our flight departs at 9:30 for Myrtle Beach, and I\'m told to get to the airport at least 2 hours before departure.  I get ready, finish packing, and my ride gets here at 6:45 am.  We get to  urlLink Newark Liberty International Airport  at approximately 7:05 am, after passing briefly through the lovely town of Irvington.  For those of you unfamiliar with the area, Irvington is like New Jersey\'s Compton.  We jump in line outside by the skycaps for urlLink Continental Airlines .  The line moved rather quickly actually.  The dude checked our bag and gave us our boarding passes and inside we went.        There\'s a side note to this story.  I had made our flight reservation in February.  Since that time, I had received 3 different notices in the mail stating that our flight time had changed, with the latest notice changing my seats as well.  I called their \"customer care\" number and they said I would have to tell them at the ticket counter inside the airport.   So we make our way to our terminal and make it through security.  I had to take my shoes off, but no big deal otherwise.  We get through and make it to our gate rather early, and notice our flight is not on the computer screen yet.  We head over to the book shop to peruse their selection of reading and get back to find a place to sit for a few hours.  When we came back, I was able to change our seats so we were together.  Well, the flight finally shows up and we board pretty much on time.  We find our seats and the plane begins moving towards the runway, and suddenly the lights and engines stop.  Hmm.  \"This is your captain.  We are having some problems with our generator.  Sit tight briefly while maintenance takes a look at it.  Meanwhile, the air is not working, and we are sitting in the plane oh, for about 2 hours before we find out that we have to switch flights.    The flight we are switching to is across the airport, at gate 124 or something.  So we rush across, and have no idea when the flight is leaving, so we eat some Teriyaki chicken at the food court, which was excellent.  The flight begins boarding around 12, and we finally take off soon after.  The flight was pretty smooth, and we landed after 2 sometime.  Now it\'s time to get the luggage and get my rental car.  I head over to the  urlLink Enterprise  kiosk, and they tell me they don\'t have any cars left for my reservation, so the guy at least hooks me up with Thrifty for the same price, and probably a better car.    We get to our hotel, the lovely  urlLink Patricia Grand .  We check in at 3, and find our room.  We go in the room.  We realize this vacation is going to suck-diddly-uck.  There\'s a huge puddle on the rug by the ac and balcony, probably from the ac leaking.  It\'s not just a little damp, but it\'s like when it downpours and you go play mini-golf afterwards.  It sloshes like that.  I call the desk and they say they\'ll send a guy up.  Meanwhile, we examine the room to find mold on the walls, probably from the lake in our room, a smaller room than advertised, and an ash tray, even though I reserved a non-smoking room.  Knock.  Hey guy, come look at this oasis in our room.  Dag, yall, that\'s about 5 gallons of water, probably cuz the ac froze up.  I\'ll be back with a vac.  This is all said in the South Carolina accent.  He comes back up very quickly and sucks up the water with a monster shop vac.  He was a real nice guy at least, and quick.    We decide to go out because we\'re getting hungry and don\'t want to stay in a moldy room.  We head for  urlLink Broadway at the Beach , which is a big cluster of restaurants, shops, bars, etc., all around this lake thing.  We stroll around to see a long line at the Hard Rock Cafe, which is fine with me, because I didn\'t want to spend $10 for a burger anyway.  We find a keen place called the Key West Grill.  There\'s a line, but we decide to see what the wait will be.  The dude says, well there\'s no wait if you want to eat outside.  Of course!  We\'re hungry.  We sit right on the dock of this lake, eat some dirty salmon and potato skins, and enjoy a nice breeze.  All right, maybe things won\'t be so bad after all.  We walk around a bit and decided to hit the  urlLink Imax theater .  We saw some adrenaline video with dudes who jump off planes and fly.  We head back to the hotel and decide to take a night swim in the pool, which was freezing it seemed, but there was no way I was going in the heated indoor pool.  That\'s where pee lives.  An interesting day one with a not-bad ending after a rather horrible beginning.  We\'ll see what Monday brings.   '),(18157,1826527,'15,July,2003','2003-07-15 00:00:00','    Five more days till happiness.  This past weekend was spent down the shizzore, but none of which was at the beach or boardwalk.  Saturday night, we attended a  urlLink minor league baseball game  for the sake of my uncle\'s club.  The game was ok and the stadium was beautiful, but the food was atrocious.  How can you screw up a hot dog and Italian ice?  Day 2 was intended for the boardwalk/beach, but we could not get to that area because there was a line of cars backed up to the parkway practically.  So we did an about face and headed back to our retreat area, where we ended up hitching a ride with my bro and sis-in-law to a crafty place we call  Blackbeard\'s Cove .  Yarr.  It had mini golf, archery, batting cages, etc.   It had a pirate\'s theme, keeping up with our viewing of the  Pirates of the Caribbean  movie on Friday nite.  Yesterday was pretty much be a slug day, but I did bowflex, run, and make a cd.  The night was capped off by visiting former co-workers and friends John, Brandon, and Kate.  It was nice talking to them, and I\'m glad to see they\'re having a blast since leaving a certain working hell.  (It sounds like \"Damn Hoody\" and it\'s a place where cds are sold).  I think I handled that pretty well, and escaped any copyright infringement.   '),(18158,1826527,'09,July,2003','2003-07-09 00:00:00','    Is it going to rain or what?  I started out by morning by running, and noticed the sky was not looking very inviting.  It\'s been grey-ish for the past few days but no rain really, yet the folks at the weather channel have been \"predicting\" thunderstorms for the past 2 days and thru Friday.    Today should be an enthralling day of taking my grandmother to the doctor and waiting.  I guess while I wait I\'ll finish reading my book.  I\'m a little more than halfway thru it.  It\'s  Summer of Storms  by Judith Kelman.  It\'s really drawn me in so far.  I\'m on page 200, and the only reason I\'m not further is that I read it at night, and my eyes start getting lazy on me.  I\'m going to try and finish it for tomorrow...The rest of my day will consist of sending out more resumes, doing bowflex, eating lunch sometime, and playing some  MLB 2004  for PS2.  It\'s one of two games I bought, but it\'s fun.  I\'m playing a full season with my Jays, and am at the end of July, and just recently pulled into first place ahead of the Yankees and Sox, because of my brilliant pitching performances as of late.  The other game I bought was Metal Gear2, which I was able to complete at the very easy level.  I loved the first one, and this one was pretty fun, too.  One day, I\'ll try it on a harder level.  Eleven days until Myrtle Beach.  Eleven days until fun, and seventeen days until I come home broke, burnt, and depressed for being back home!   '),(18159,1826527,'07,July,2003','2003-07-07 00:00:00','    Today was a miserable experience at dmv.  I was there for nearly 2 hours just to get a duplicate license.  My license did not have a photo on it, since I renewed it via mail 3 years ago, but since I\'m flying to South Carolina, I need one with a photo, so I had to purchase a duplicate with a photo.  Jovial employees only made the experience better.  I looked through possible stuff to do down in Myrtle Beach today, and there\'s a lot, and it\'s going to make me broke.  You know what bothers me?  People who are dishonest.  I don\'t know where that came from, but it annoys me to no end.  So it\'s supposed to feel like 110 tomorrow, and of course it\'s supposed to be 78 and rainy on Thursday, when I wanted to go to the shore.  Oh well.  I hope I find a job soon.  Oh just a side note.  In the sports section today, there was an article on how Roger Clemens got snubbed from the All-Star team.  That is absolutely ridiculous.  Sure he won his 300th game this year, but not all of them occurred this year.  What\'s his record, like 8-6?  And he deserved to make the team?  Come on.  I hate the Yankees.   '),(18160,1826527,'01,July,2003','2003-07-01 00:00:00','    Hello July.  Happy birthday Den and Rob!(from June 29 & 30, respectively)  So it is absolutely bee-yootiful today.  I had an interview today for a 5th grade position.  The principal and people who interviewed me were really pleasant, and the school was relatively small, which I like.  The only problem is that if I actually get the job, it might only last to February because the teacher is on military leave.  So if I got the job, I would be back to square 1 again in February or March.  Hmm, I guess I\'ll just have to wait and see...  Today I washed my car, so I\'m sure it will rain sometime in the next 24 hours.  The only thing I don\'t like is all the gooey crap that\'s stuck at the bottom of my doors from my tires kicking it up.  I can\'t get it completely off.  I use soap, bug & tar stuff; it just doesn\'t budge.  It looks shiny and clean nonetheless.  only 19 days till vacation.  enjoy the weather.   '),(18161,1826527,'31,August,2003','2003-08-31 00:00:00','    Glad to hear from Matt!  Now everyone\'s official.  First of all, getting back to the Olive Garden.  The Olive Garden sucks.  Saturday was the first time I\'ve gotten vocal with a waitress, because a table that ordered 10 minutes after us got their food before us, and two of them ordered what I had.  I said, \"Bitch, where the f--- is my food?  I will kill you if it\'s not here in 25 seconds.\"  No I didn\'t really say that, but I was pissed and she knew it.  They suck.  Party at brother\'s house was interesting.  Supposed to be a \"pool party,\" but it was 70 and breezy, so if you went in the water, you would come out missing your reproductive glands.  Too cold to swim today.  But some nice food and drink.  The kind of drink that makes the day seem better.  Kim, would you like to tell about our ice fiasco before the party?    Well, here\'s part 3 of my story:  That night spooked me.  The murder scared me, and I could tell Corey was affected, too.  The tension between us was thicker than a two-day-old McDonald’s milkshake.  We didn’t talk much.  We just watched some Scooby Doo.  It entertained his feeble mind, and kept me from worrying a bit.  At about six o’clock, just before the news, the doorbell rang.  I jumped and Corey dashed to answer it.  Andy and Josie walked in. Neither goes anywhere without the other half.   I had known Andy from elementary school, and he had always been a real person.  For instance, he had to wear glasses as a child, and constantly was chastised for it.  In his junior year of high school, his doctor told him he no longer needed to wear them.  He decided that his glasses had been through so much with him, that they were indeed a part of him, just like wearing his underwear everyday.  Andy still wears his glasses all the time.  Josie was a friend of mine who I had set up with Andy.  She is a sweet, attractive girl.  She’s the type of girl that you can be friends with, but deep down you’re so damn jealous of her.  She was tall, maybe about five ten, and had beautiful brown hair and full lips.  Her body was unfair as well.  Her height presented her with long legs, and she also had killer curves, from her beautiful breasts to her sexy hips.  She meshed well with Andy, too.  They’ve been going out for half a year now, and are getting closer every day.   I looked at the visitors.  Nothing ever seemed to phase Andy, but I noticed Josie was zoned.    “’Sup guys?  I found out some more about the murder in my last class today.  I heard they brought in Erika’s ex for questioning.”  Andy was directing this mainly at Corey, because I was concentrating on Josie.  “Who?  Sammy?” asked Corey.  “Yeah.  It seems logical.  He was such a pervert.  He was into voyeurism and all that shit.  I heard one time he got caught watching Miss Burns in the bathroom.”  “Oh that’s gross Andy!” I barked.  “You know that’s not true!”  I noticed Josie’s eyes.  They opened wide when Andy mentioned Sammy, like she knew, or had experienced his perversions before.  Well, now that I think of it, she used to go out with Sammy, so maybe…  Corey signaled to Andy to chill with the murder talk for now.  Andy looked humbled, and we all silently sat down to watch some TV.  We channel surfed through the time slot of the news, and ended up watching reruns of The Cosby Show for a while.  There’s one channel that shows reruns of the show for three consecutive hours.  I think we caught five episodes, and then Cheers came on for a string of reruns.  I guess we all fell asleep.  The last thing I remembered was Cliff and Norm poking fun at Woody.  Corey nudged me at about one in the morning.  “Christy, you were supposed to be home by now.  Let me take you home.”  “Can’t I just stay?  I’m tired.”  “I wish you could, but your parents thought you were going home tonight.  What if they try to call you in the morning and you’re not there?”  I got up and stretched.  My arms extended above my head with a yawn, and my shirt lifted above my belt and exposed my bare stomach.  I noticed Corey catching a glimpse.    I was reluctant to go to a home where my parents no longer lived.  I stayed in the old house, which was small, while they moved into a larger pad.  Sure they were close, but I still wasn’t used to living on my own.  Andy and Josie must have opened their eyes, too, because they quietly shuffled out the door.  I noticed Andy giving a weary wave as he exited.   '),(18162,1826527,'30,August,2003','2003-08-30 00:00:00','    Hello all, Even with the convenience of communication supplied by the blog, I still take forever to write. My apologies. I just finished my first week of classes since coming back from my first full-time rotation, and I have realized that I don\'t want to be in classes any more. I just want to finish. Only nine more months to go.  Dave Attell was pretty funny, not a whole lot of new stuff, but still worth the trip to see him. The last 20 minutes of the act were kinda rough. It was a midnight show, and Dave had already done 2 shows earlier and went pretty heavy on the jagr shots during his set, combine this with the annoying drunks who felt like they had to yell things like: \"when you went to Amsterdam, that rocked!\" (it reminded me of \"The Chris Farley Show\" skit from SNL).  In order to waste more of my time, I picked up a Nintendo Gamecube from ebay last week, I\'m eagerly anticipating it\'s delivery so I can flush my future down the toilet and play Soul Calibur 2. That\'s the big news in my life, I shall blog again later.   '),(18163,1826527,'30,August,2003','2003-08-30 00:00:00','    Hey guys!  Sorry for the Sparta confusion.  From now on I will no longer mention the city of Sparta.  Rob, you are too funny!  A frisbee.  Hehe..  Going to stuff my face with lots of garlic today at the Olive Garden.  I keep hearing stories about people getting sick from their food, but from what I hear its their meat sauce which I think sucks anyway.  However, their appetizers are delectable as well as their breadsticks and salad, which by the way has a lack of olives.  Is it ironic?   '),(18164,1826527,'29,August,2003','2003-08-29 00:00:00','    Saturday and college football is in the air.  But I just remembered I don\'t care.  Watch out for William Paterson!  They might win 3 games this year.  As for the pros, it\'s the same old Bengals.  They lose because of 2 turnovers and they can\'t convert an extra point.  Same story as last year, and the year before that, and the 11 years before that...Here\'s part 2 of my story.   Late on a Thursday morning, the student center usually becomes a jubilant place to be, since the weekend would be creeping up soon.  Not today, though.  The murder of a fellow peer, Erika Harmon, stirred up the entire student body.  Erika was a plain girl-not very smart or athletic, just your everyday Jane Doe.  “I heard she was found with her eyes missing,” said Corey, my boyfriend.  “I guess she didn’t see it coming,” said a giggling Andy.  You would think they would be mature enough to respect the dead since they’re in college.  I won’t call them college students because I can’t say if they actually study, but they are in college.  I have been going out with Corey for only about two months now, so I don’t know where our relationship is going, but I do know I need him to grow up a little bit.  “That’s so rude.  You’re joking about a girl who was just murdered,” I snapped.  “Sorry, Christy.”  “I’ll be right back.  Please stop joking about this.  It’s not funny. This is serious.”  	I made my way to the restroom before Corey and I left for his place.  The boys continued to talk despite the plea, but kept it at a whisper at least.  “Dude, I heard they found a note that said ‘love is blind’ or something,” said Corey.  “Well, she didn’t have any eyes, so…”  Andy giggled again.  “Yeah, because the killer probably did something freaky with them, like put ‘em in his pants or something.”  “Ooooh.  Well, I gotta go see if I have my class, so I’ll see you later,” Andy uttered.  Andy was even slightly disturbed by my boyfriend’s lack of taste.  He made a grimace that Corey probably interpreted as a smile, but most would recognize as a “you went a bit too far with that one” look.  He turned towards the glaring sun and made his way to the Comm Building for class.   Andy was majoring in communication.  He wasn’t a brilliant human, but he knew how to talk and listen well.  I returned, worried.  It seemed surreal that murder could get this close.  Corey knew I was upset.  He kissed me lightly on the cheek to let me know he was there, and led me out to his car so we could go to his house.  Our classes were done for the day.  It was weird.  I didn’t really want to go.    “I almost got into an accident on my way here,” Corey blurted.  “Oh, sorry,” I said, not really aware of what he said.  “Are you gonna be OK?” he asked.  “Yeah.”  I shook my head no to contradict my words.  Corey pulled me close to hug me, but he knew I wasn’t hugging back.    He shut the car door after I was seated and went around to get in his side.  There was a ticket lodged underneath his wiper blade.  He was familiar with finding the bills on his windshield.  He opened the glove box, folded the ticket quickly, and stuffed it among the crinkled repair slips.  Careless.  The car started.  Silence.  The radio wasn’t…Well, it wasn’t on.  The horrible sounds of manufactured modern rock sounded abruptly in the speakers and Corey continued to switch between commercials and Linkin Park.  I couldn’t decide which was worse.  Although annoying, I realized that I had momentarily forgotten about the horrifying events.    Corey recklessly plops into a spot not far from his place, between an oversized SUV and dumpy pick-up.  The Subaru’s front end is brushing the tail of the SUV, and his rear is jutting out past the width of even the SUV.  Instead of straightening the car, he stops the engine and slams the door.  I get out and grasp his hand lightly.  We walk towards his apartment.  Number 7 is nearly invisible from the street.  It hides at the junction of the L-shaped complex, behind bristly shrubs.  There’s always bees hovering around from the vegetation.  We enter and he throws his keys on the old end table.  We plop ourselves on his couch and welcome the safety of a home.   '),(18165,1826527,'29,August,2003','2003-08-29 00:00:00','    Hey all.  I finished my story for now.  That\'s not to say that I won\'t change things or add things, but I\'ve made it have a beginning, middle, and end.  So here\'s what I\'m going to do.  I\'m going to post it piece by piece right here on blogger.  A little bit every day.  After posting the first part, I may try to add it to my  urlLink site , but I don\'t want to post it as a doc file in word, so I don\'t know yet.  It\'s too much work to convert it to html on my site, too.  I\'ll let you know.  Here\'s the beginning.  As usual, I appreciate  urlLink feedback .  I apologize for the lack of indentations as well.  The formatting of blogger is not being cooperative for me.   \"An Eye For Murder\" by Brian Burkhardt   \"She was found dead early this morning in her apartment two miles north of Vernon College. She was a sophomore there, where she majored in philosophy. Her devastated family would not comment.\"   \"In Harrison last night, a two-story [CLICK].  The picture faded from the TV screen after Corey hit the power button on the remote. He picked up his notebook and pen and looked at the clock.  8:38. I’m already 38 minutes late, he thought. He spun around for his forgotten keys and slammed the door shut. He paced briskly to his red Subaru. Corey was practically driving as the engine turned. A sound of screeching burned from the tires as he headed to school. His brain was working: I’m never going to find a spot. The radio was annoying him, so he channel-surfed. Nothing but commercials.  The seek button searched the stations as Corey drove recklessly through cars. It stopped scanning on the final station on the dial.   \"In Harrison, a Vernon College student’s body was found…\"  \"Shit!\"  The red Subaru was inches from striking a BMW fresh from the showroom floor. As his head falls into place on his neck, Corey utters profanity at the inattentive businessman chatting on his cellphone. Corey pounds on the off button on his stereo. The cherry on wheels splits through traffic and Corey hangs a dangerous left into lot #5, slightly beating out a charging pick-up from the opposite direction. The car ends up still in a criss-crossed patch of blacktop next to the real parking spaces.  It’s 9:01. Corey performs a Carl Lewis to the Meyer Building for his 8:00. Climbing the stairs in leaps, he stops at room 310. There’s a note. ‘Class Cancelled Due to University’s Loss.’ It appears that the entire floor is empty. Doors are locked and lights are missing. Corey heads across campus for Rudd Hall. Halfway across campus, Corey sees Andy, who states gingerly, \"Christy and Josie are on their way over.\"   '),(18166,1826527,'28,August,2003','2003-08-28 00:00:00','    I also think Dave Attel is the shiznit.  Have fun Dennis and Matt.  Kim\'s post confused me momentarily, for I thought she went to Sparta, as in Dennis Sparta.  I thought that was quite a ride to pick up stuff.  Besides, what could Kim have left with Dennis?  a frisbee?  a ballcap?  maybe a lamp?  I don\'t know, but she got it now, whatever it was.    So I didn\'t feel like cooking today at around 7pm, my meal time.  I decided to get take-out food of some kind.  However, I decided against Chinese food since I think I order from the local Chinese eatery at least once every 2 weeks, and the last time I ordered I think he (guy always there) recognized my voice.  I always order the same thing, so I got self-conscious about it.  I didn\'t order there.  I bought some cajun-salsa-wrap instead.  It was fine and I microwaved some peas and corn, as to have vegetable in my meal as well.  I didn\'t fix school yet, but I did get a \'Lost in Translation\' poster with Bill Murray on it, signed by Sofia Coppola.  Good luck with the foot, Bri.  I had something weird like that happen to me, and it took 4 days to get better.  I did it in my sleep somehow, quite painful.    Down with UNC.  (I really have nothing against UNC, but I like to antagonize.)   '),(18167,1826527,'28,August,2003','2003-08-28 00:00:00','    I am extremely jealous!  Dave Attel\'s my boy!  Enjoy the show fellas.  I hate you.  No, not really.   '),(18168,1826527,'27,August,2003','2003-08-27 00:00:00','    HI just finished my 1st week of classes including the dreaded Neurobiology.  I now know the difference between oligodendrocytes and Schwann cells.  Bring on G coupled proteins!  Books were $270, which is a bargain.  In other news, I purchased The Lord of the Rings: The Two Towers.  Great movie (Im sorry George Lucas but LOTR is better than Star Wars).  The DVD had a making of The Return of the King movie.  They were showing some cool scenes from the movie and then abruptly cut away before all the really good stuff happened.  Kinda like in The Simpsons when Itchy and Scratchy see a fireworks factory and then pass it by because Poochie the Dog joins the hillarity.  Anyway, I think everyone should read these books before they see any LOTR movie because they are great reading.  Finally Matt and I are going to see Dave Attell on Fri nite in Raleigh.  Talk to everyone soon.  On Sat, watch ABC to see the amazing North Carolina Tar Heels battle the criminlas at Florida State.  Go Heels!!   '),(18169,1826527,'27,August,2003','2003-08-27 00:00:00','    I\'M FINISHED!!!  I have successfully completed my intro to computers class.  Woo hoo!!  The final took me about 5 minutes to take and me and this girl Julie flew out of there.  We rejoiced as we walked 10 miles back to our cars at lovely  urlLink Montclair State University .  Go Red Hawks!!  I\'m glad its over with.  After class i treked up to Sparta with my brother to pick up some things.  What a house!  His roommate, the guy who owns the place, said he only plans to stay in his mansion for 2 years.  Wha?!?  Must be nice to move from one house to another every 2 years.  Lucky.   '),(18170,1826527,'27,August,2003','2003-08-27 00:00:00','    Today I step aside and dedicate this page to  urlLink Brandon .  He\'s always been supporting my page and songs, and that\'s something not many people do.  It\'s nice to hear things about what you like doing a lot(i.e. writing) whether good or bad.  I really respect Brandon\'s opinions as well because not only is he a writer, but he\'s always been genuine with his feelings and never hides anything.  So here\'s to  urlLink Brandon .  I\'ve enjoyed reading his stuff since we began working together at a really bad record store.  Now go check out his awesome works.  Thanks dude.  In other news, my foot cannot be walked on.  I don\'t know what I did to it, but yesterday after running, eating breakfast and showering, it didn\'t hurt until I tried walking down the stairs and I became a cripple.  I can\'t put any weight on it because it hurts so damn much.  I\'m not going to the doctor either.  \"Well, let\'s take an x-ray and see...  Hmm, I don\'t see anything, just give me $200 and take some motrin for any pain.\"  I may be going to my brother\'s house on Sunday for a bbq/pool grand opening.  They got their pool installed on Monday, and it should be ready by then, so they\'re gonna have some people over.  I keep getting visions of a rap video with all these guys in the pool with gold teeth and chains with some dirty girls dancing around the outside.  Bling bling y\'all.  No, God I hope not.   '),(18171,1826527,'27,August,2003','2003-08-27 00:00:00','    Hey, it\'s Rob!  Sorry, but I didn\'t see your blog until after mine showed up today.  Welcome aboard, but why the hell are you up at 3 am?  Blogging isn\'t really like dancing to techno music.  It\'s more like dancing to gospel music.  It\'s inspiring bee-otch!  What?  I don\'t know what I\'m talking about.  You don\'t dance to gospel music.  I\'m glad to read your words.  Peace out.   '),(18172,1826527,'26,August,2003','2003-08-26 00:00:00','    I ran today all by myself.  It was different, but I ran my old running route by my house, and realized there are a lot of hills to run up.  It wasn\'t very fun.  I\'m really tired now.  I don\'t think I slept well, but I don\'t know why.  I\'m probably gonna look for some crappy part-time work today, so we\'ll see how that goes.  I wrote two new things, found  urlLink here  and  urlLink here .  They\'re both pretty short, and I don\'t know why they\'re both somewhat mean and spiteful, because I\'m not in a bad mood or anything, but whatever.   '),(18173,1826527,'26,August,2003','2003-08-26 00:00:00','    Hello.  I think I figured out how to blog.  I hope blogging isn\'t like dancing to techno music.   I don\'t know how to do that very well.  Plus I always end up smelling like smoke and have some super-jacked guido dance into me.  It\'s all very bad.  I don\'t go to clubs.  Is blogging like \'gellin\'?  If it is, I want to be blogging like Magellan.   I just got home from work an hour ago.  I am nearly nocturnal now.  I should write email to my professors trying to set up meetings for school, as in getting me in it.  However, I think i\'ll listen to music and write tommorrow (maybe).  I\'ve used those mirror kits before.  They are 2 for 2 in my experience.  I hope yours makes it 3 for 3.  Seriously, I haven\'t listened to any of my classical music for like a year.  It sounds good right now.  Ok.  That\'s my blog for now.  Lata dawg.   '),(18174,1826527,'26,August,2003','2003-08-26 00:00:00','    Hey!  My rearview mirror keeps falling off my windshield.  I went to the store on the weekend, but they didn\'t have the adhesive in stock.  \"Check back Monday or Wednesday,\" guy says.  I decided to try some cement stuff we had in the house that said it worked on glass and metal, so I glued away and let it dry overnight.  I put my mirror on gently yesterday without tightening it, and this morning, it was laying on the floor, mocking me.  It said, \"Ha.  You wretched human. You cannot contain me.\"   I said, \"You\'re right.  You can just stay on the floor then.\"   I drove up to the mall without a rearview mirror because I was lazy.  Probably not the smartest idea, but I had to get my dad a birthday gift.  On the way home I stopped at said auto store above, and they had the adhesive kit.  I purchased it and attempted to use it upon my home arrival.  After scraping off the old glue and thoroughly cleansing the button and windshield, I was ready.   Place one drop of adhesive to center of button and hold firmly on windshield for 15 seconds.   I oblige, but upon removing my finger, the button came along with it.  Blast!  So I quickly dripped one more dab of adhesive and placed the button back up quickly.  This time, it stayed.  For now.  I\'ll let you know what happens when I actually put the mirror on.  I got an email from one of my friends from school who also is trying to be a teacher, and she made me feel better, because she can\'t find a job either.  She\'s going to be a sick leave replacement for a month for her cooperating student teacher, but that\'s over in October.  So I guess everyone I graduated with still doesn\'t have a permanent job either.  It\'s about 2:30, so I\'m going to watch Family Guy now.  Late.   '),(18175,1826527,'25,August,2003','2003-08-25 00:00:00','    Blogging again in the middle of my computer class.  Why am I here?  I don\'t know really.  I can\'t understand a word this guy says.  I have a really bad attitude about this class.  I\'m not going to use Access in my lifetime.  I\'ll use Powerpoint And Excel but I knew this shizzy before I registered for this class.  oh well. Only 2 days left.   '),(18176,1826527,'25,August,2003','2003-08-25 00:00:00','    Hey, sorry, but geocities sucks.  For some reason, you can\'t externally link to a picture on my site, so click  urlLink here  to see the 2 pics I linked to below, but don\'t work.   '),(18177,1826527,'23,August,2003','2003-08-23 00:00:00','    Yesterday was some good fun.  Went to  urlLink Turtle Back Zoo !  I haven\'t been there in years, but it was a beautiful day.  The sun was shining, and it was in the lower 80s.  The first thing we visited were my favorites, the penguins.  Coolest animals ever.  They were just chillin\', but then they started jumping into their pool and swimming around.  We proceeded around the zoo and met some cattle and elk.  I met a new  urlLink friend .    We saw a wolf in the distance later on, and a bobcat and cougar.  My dad  urlLink fed the goats .  The goats and horses smelled like ass.  What else did we see?  Some happy chickens, outbreak monkeys, and otters.  There\'s probably a few I missed...It only took about an hour to go through the zoo, and we weren\'t really rushing either.  It was something different, so it was nice.  I\'ve been itching to go to a zoo for a while now, so I was happy.    I saw a few job postings in the paper yesterday.  One was for a 4th grade teacher in Tenafly.  I really don\'t know where that is, but I\'m trying for it.  The ad was like send in cover letter, resume, copy of certification, and transcripts now!  No, but it said by like Tuesday or something.  Ridiculous.  Hey, I didn\'t tell you to wait till a week before school starts to post a job.  I woke up to run this morning and it was quite brisk for an August day, but it was good because I didn\'t sweat too much.  Well, I\'m going to look for some crappy part-time work this week, in case I don\'t have anything solid for the start of September.   '),(18178,1826527,'23,August,2003','2003-08-23 00:00:00','    Sparta!  Welcome aboard Den.  I\'m glad to read about your mishaps/doings.  I\'m sorry your Saturday was wasted, but when you\'re finally done with grad work, you\'ll be one happy mofo.  As for madden, that fat bastard...Well, let\'s just say I agree with you whole-heartedly because I haven\'t purchased one of his games since I had Genesis, and Madden 96 ended up flying out the window towards my neighbor\'s house and my control exploded into many pieces.  When you\'re up on the Chargers 26-0 going into the 2nd half, you would expect to be able to win the game, at least preventing them from scoring 26 unanswered points.  Not the case.  Jeff Blake throws 3 interceptions on 3 pass attempts.  The only reason I was passing was because I couldn\'t run either, so I tried passing on 3rd and longs.  After having a 12 point lead, I decide to try to run out the clock.  Well, the running back fumbles once and of course the Chargers convert a touchdown in a matter of seconds.  On the next posession, I get a good return on the kickoff, but oh, guess what, I fumbled.  They return it for a TD without getting touched by anyone on my team.  Game ripped out of system while power is still on and thrown out window at my neighbor\'s house.  Game ends up on my side of the property, without visible physical damage.  I jump on it, try to snap it in half, but it\'s indestructible.  I go on to puncture it with sharp objects, and place it back into its case.  Thus ending me ever purchasing a madden product again.  And by the way Den, cursing is acceptable.  Just try not to overdo it.  Saying madden is a fat fuck is AOK in my book, and so on.  Thanks for joining dude.   '),(18179,1826527,'22,August,2003','2003-08-22 00:00:00','    Hello all, I got an email from Brian to contribute to the blog, which I\'ll ty to do for the amusement of others of course.  It\'s Saturday in beautiful Chapel Hill, NC, but I wouldn\'t know.  I\'m just about to finish my 11th hour of work today.  Yes that\'s right, 11 hours.   This is the horror of graduate school.  Classes resume on Tues and this is supposedly my tough year with Neurobiology, master\'s thesis and comps (basically like your med school boards).  So much to look forward to!  Anyway, I just got Madden for the Xbox.  I have a love/hate relationship with all Madden games.  There is a lot of good to it, bt there is also bad stuff, which I\'ll tell you about.  When you are 14-0, the computer will make you fumble or throw many interceptions.   they will also have the most unstoppable offense this.  Apparently Trung Candidate is the 2nd coming of Jim  Brown.  My usual experience will be to enjoy the 1st quarter, get pissed by the 2nd quarter, play well in the third, then quit the gam with disgust by the fourth after Jay Fiedler throws his 6th pick of the game and Madden berates me for making a poor pass.  F*** you Madden! I\'m about to finsih up at the lab and go home to enjoy my weekend or what\'s left of it.  Also, I just finished Jon Krakeur\'s \"Under the Banner of Heaven.  Great book about the Mormons (seeing how f**** up they are) Can I curse on this page Bri.  BTW Krakeur wrote \"Into Thin Air\" about a failed Everest campaign, another good book!   '),(18180,1826527,'22,August,2003','2003-08-22 00:00:00','    I\'m driving myself nuts trying to figure out my computer homework.  There is no possible way that I know of to calculate the mean when using excel.  I\'ve checked every possible fuction there is.  Why should I have to do it by hand?  It\'s fuckin\' excel!!!    Anyways, I\'m going to see my big brother Joey in Sparta today which is about a half an hour away down the lovely route 80.  Maybe I\'ll go swimming with the bears.  Maybe I\'ll meet Fuzzy Wuzzy and see if he really is fuzzy.   '),(18181,1826527,'22,August,2003','2003-08-22 00:00:00','    Friday is pizza night.  It\'s been that way forever pretty much.  I just found out that the new  urlLink No Motiv  cd has been pushed back to 2004.  Why?  So Vagrant can promote dashboard and autumn to asses even more.  Garbage.  Last night, I applied for a job to  urlLink this place .  I saw a posting in my local paper for a 4th grade maternity leave.  It would start in mid-November and go till the end of the year.  I really don\'t want to be a sub again, but if it\'s all I can find, what can I do right?    On a sad note, I found that the man who did the voice of Monterey Jack in Chip N Dale\'s Rescue Rangers passed away on August 11th.  I was saddened by this, because he was indeed my favorite character, and it came as quite a shock.  So I wish his family well.   '),(18182,1826527,'21,August,2003','2003-08-21 00:00:00','    Welcome to a heatwave.  It\'s officially gross for the second consecutive day.  The air is stagnant.  Running this morning made my lungs work extra hard to breathe in more polluted Passaic/Clifton air.  I don\'t plan on doing much today.  I\'ll probably Bowflex, practice guitar a bit more, and work on my story more.  My love read what I had so far yesterday and she seemed to like it.  I have to expand more, but she liked the idea.  I\'m listening to Nada Surf\'s  Proximity Effect .  Remember that song \"Popular\" they had?  Well, it\'s not on this CD.  I\'m still working on my  urlLink bands  page.  I finished up four more pages today, but have quite a lot more to do.  I\'ve gotten to most of my hardcore-liked bands so far, but I still have lyrics to add, which will take a looooooooooong time.  I\'m keeping busy at least.   I started reading a new book last night.  It\'s called  All The Dead Were Strangers .  I am only up to page 30, which is the start of Chapter 2.  Here\'s what I know so far...There\'s 2 guys chatting in a bar in NYC.  A is a detective on hiatus for some unknown reason and B is some guy that was in the military, but liked it so much that he left?  So they feel they are being watched/followed.  They decide to leave.  They hail a cab and get dropped off a block from a subway station.  Guy B gets out and says \"I\'m going to Queens, Fred.\"  He tried to give false info in case the cab driver was part of the following.  Anyway, they wait for the cab to drive away, then head in the opposite direction to another bar.  Here, Guy B whips out a napkin with 5 names on it.  He asks Guy A to research the names, but doesn\'t say why.  Guy A obliges.  They go separate ways.  Guy A to Long Island to meet some character named Mickie.  Guy B goes to a hotel, avoids being seen, and sneaks up to a room on the 3rd floor.  It\'s apparently his room.  He left 3 hairs in the door latch and a trail of baby powder on the floor, in case he\'s being followed, he would know.  He carefully opens the room door, sees the powder trail untouched, and suddenly gets knocked out.  He wakes up naked and tied to the bed.  There\'s 2 guys in the room.  One wasn\'t wearing pants.  They were beating Guy B up to get info about who the guy he was with was, and what the napkin was about.  He gets punched, etc. and tells the guys.  Then pant-less man gets into bed with Guy B and thus ends the chapter.    I have no idea what happened.  What the hell kind of book did I buy?  Good lord!   '),(18183,1826527,'20,August,2003','2003-08-20 00:00:00','    Sitting in my computer class bored as hell.  I\'m a bad girl for blogging in class.  Shame on me!  It\'s only 15 minutes into class and I\'m ready to fall asleep.  Save me!!!!   '),(18184,1826527,'20,August,2003','2003-08-20 00:00:00','    It\'s August 20th, and I am still unemployed, but I\'m glad there\'s over 300 new teachers at an orientation in New York today(as heard on the news).  With my free time, I\'m trying to play guitar again.  My fingers hurt.  I\'m just starting with the basics-like learning some chords first, and playing some easy parts of weezer and hey mercedes songs so at least i play something that i know how it sounds.  I\'m happy, because so far, I\'ve been doing ok.  I have a bad memory, so I\'m not rushing, but I practice a few new chords every day, and while I try to remember them, I throw in the ones that I\'m already supposed to know.  I\'m teaching myself.  It\'s kind of cool.  I guess I do have a teaching job.    In addition to trying that, I\'ve been messing with my  urlLink webpage  quite a bit.  I don\'t really change the main page format, but I\'m working on the inner pages.  In my  urlLink bands  page, if you click on some of the names, I made my own little page for my favorite bands so far, instead of a lame link to their website.  It\'s still a work in progress, but it\'s coming along.    My brother is getting a pool installed on Monday, and having a get-together the Sunday of Labor Day weekend.  So I can drink my face off if I still don\'t have a job and blame depression, or I can drink my face off if I get a job and blame it on celebration.  Speaking of drinking, the Blue Jays suck.  Right now they\'re the biggest bunch of sucks that ever sucked.  Hey, can you pitch?  Because I\'m sure they\'d throw you into their rotation as the #3 starter.  Right behind Troy McClure and Mike Chokesondick.    I\'m listening to No Motiv\'s  Diagram For Healing  right now.  It\'s one of my favorite CDs that I own.  I haven\'t heard it in a while, either, so it sounds good.  On a side note,  urlLink Hey Mercedes  are not only a good musical band, but they are quite humorous, as they had a bowling competition while on tour and posted their scores online.  Now who, I ask you would think of doing that?  I\'m writing a short story which I will probably post when I\'m finished.  It\'s something that I had a rough idea for in my creative writing class in college, but I\'m expanding on it, and trying to detail the characters more, and polish the ending.  As usual, I appreciate feedback, whether positive or negative.   '),(18185,1826527,'18,August,2003','2003-08-18 00:00:00','    I know Brian was expecting to hear more from me about Hey Mercedes so here goes....  They played about an even amount of old and new songs.  Since I don\'t really know the names of the new songs I shall tell you what old songs they played (not in order, my memory isn\'t that good):   A-List Actress, 11 to Your 7(one of my faves),  Our Weekend Starts on Wednesday(of course), St. James St., and Frowning of a Lifetime.  Of course they played flawlessly.  They\'re my boys yo!  The new songs sounded most excellent.  Because they are the bestest band around everyone should pick up their new album    Loses Control    on October 7.  In the middle of their set, they got the whole crowd to do a \"seventh inning strength\" if that\'s what you would like to call it.  The crowd was totally into hm, especially when they stopped playing and we all sang part of 11 to Your 7.  It sounds cheesy like the 14 year old girls that sing along to the pooping Chris Carabba(sp?).  I think the singing along thing really makes the band feel proud. They realize that they have fans that care enough to memorize lyrics, pay money to see their shows, and then sing along.  If I were on stage I know I would get a complete rush if I saw a bunch of people singing my songs.  Now the people understanding the lyrics is a whole other thing but that\'s not what music is really about.  Let me stop before I bore everyone with my feelings on music.  Anyways, Hey Mercedes rocks and that\'s all you need to know.  Now go buy a T-shirt! Linking verbs, linking verbs.   '),(18186,1826527,'15,August,2003','2003-08-15 00:00:00','    Ok, well Kim had beaten me to the punch about hm, so ignore what I said about her saying more later...   '),(18187,1826527,'15,August,2003','2003-08-15 00:00:00','    Well, the preceding weekend was an unusual sunny day on an otherwise overcast two months of summer.  Friday I was able to spend some time with my dad because he wasn\'t able to work from this blackout I\'ve been hearing about.  On Saturday, I went to see my grandmother and my parents, her, my aunt, and I went to dinner for her 91st birthday, which is actually today.  When I got home, there was a long box on my front steps from  urlLink Vagrant Records .  I opened it to find an autographed tour poster from my boys  urlLink Face to Face .  I guess I won a contest that Vagrant had.  I hadn\'t heard from them, except for the poster.  Nice.  Now I have to get a frame for it!  Then at night, I saw my honey and we watched the finale of  Shark Week .  Yesterday was fun.  We went to see our heroes known as  urlLink Hey Mercedes  at Maxwell\'s.  The show was awesome.  They played many a new song, which sounded smooth, and make me anxious for the new album.  October!  I also enjoyed the sounds of one of the opening bands called  urlLink  Damone .  They even threw in a Weezer cover into their set, and they were rockin\' out.  As for HM, tight as always and nice guys, too.  I\'ll let Kim tell more about their side when she gets home later.   So that was my weekend.  It was a nice little relief from the boring days of summer I had been experiencing.   '),(18188,1826527,'15,August,2003','2003-08-15 00:00:00','    Hey Mercedes Rocks!!!!  Went to see the boys last night at Maxwell\'s in Hoboken.  The opening bands were pretty good for once.  I especially liked Damone.  As Brian said they all look like they should be in Slayer.  They had girl singer and that rocks.  Especially since she was good.  Hey Mercedes kicked ass as always.  I talked to Bob and Damon.  Bob is such a sweetheart!    I got out of class today after taking a 10 minute test.  Now I have plenty of time to do nothing before I head off to work.  Yuck!   '),(18189,1826527,'15,August,2003','2003-08-15 00:00:00','    You should hear the next and last homework assignment of 2 for the whole course.  Here goes...  Type a cover letter, resume, and print an envelope.  Easy right?  But here\'s the catch he wants you to use like 5 different fonts, do all sorts of crazy things to the margins, create our own bullet points, and all this by Wednesday.  Oh no!!!  Anyone have a sarcasm machine?   '),(18190,1826527,'14,August,2003','2003-08-14 00:00:00','    it\'s an easy A unless he tells you to type  3  letters.  ooooo.   '),(18191,1826527,'14,August,2003','2003-08-14 00:00:00','    I need to vent a little so here goes....  I\'m taking this horrible summer class at the wonderful Montclair State University.  It\'s an intro to computers class and it\'s cake.  You would think I would be jumping for joy but I\'m not.   The class is waaaayyyy too easy.  The professor is from Ghana and he has the worst accent ever and the plate in his lip doesn\'t make it any easier to understand him.  Just kidding!  There\'s one girl in the class who constantly yells at him.  She says that he ignores her and he\'s not a real teacher etc.  It\'s so funny because during the labs everyone is talking and joking around but when she starts yelling at him everyone gets really quiet and stares at the monitor.  Too funny!  But the class ticks me off because i have to be there everyday watching half of the class fall asleep on their keyboards.  And today the fire alrms went off around 8:55, 5 minutes before class started.  So the class treked outside still waiting for our professor.  I see him walk by around 9:20 and he goes to the student center.  Wha?  His class is sitting outside and he doesn\'t even ask what\'s going on?  How considerate of him.  So we go back in 20 minutes later and we have class.  Why?  We already missed nearly an hour of class time?  Why freakin bother?  Becasue he\'s a poopface.  So he lectures for 35 minutes and tells us to do 2 labs which is quite stupid because all we had to do was type 2 letters.  Here\'s the catch he already had them typed!  All you had to do was copy and paste.  How stupid is this guy?!?  Oh well.  I shouldn\'t complain.  It\'s an easy A, right?   '),(18192,1826527,'07,August,2003','2003-08-07 00:00:00','    Well, my day-by-day trip blog is complete.  You can only experience our trip if you were there, but I\'m sure my sarcasm may have given you some hints as to our fun.  I wrote two new songs.   urlLink This one  is dark and evil, while  urlLink the other one  is about a harsh break up.  The first song is called \"Herself\" and the second is titled \"Disdain.\"  I hope you like them, and if you don\'t,  urlLink tell me .  Actually, I would appreciate it if you told me if they were good or bad.    I attempted to fix up my homepage, but geocities yelled at me.  I had copied and pasted my page onto wordpad while I attempted to fix up my page, but somehow they knew I did this, and it\'s apparently a violation of their terms of service.  So I stopped, closed wordpad, and re-signed in, and all was well, except my page is not too different, except for my entrance page.  I put up a new picture.    Today I grabbed lunch with my friend  Matt .  I haven\'t seen him in about a year maybe.  He\'s going to Duke right now to complet his grad school work.  He\'ll be finished with everything in May.  Damn right.    That\'s pretty much it.  Just felt I had to blog since I\'ve been doing it on a daily basis for the past week.  Late.   '),(18193,1826527,'06,August,2003','2003-08-06 00:00:00','    Day 7  07.26.2003  Saturday  Today is our final day here.  Our flight checks out at 12:48.  We wake up semi-early, because we want to get to the airport early.  We pack up, and I check out of the hotel.  I find that a local call was priced at $4.  Wow, that\'s economical.  My phone bill was $40.  We leave and go to Dunkin\' Donuts for a breakfast sandwich, and a muffin for later.    We finish and begin following signs for Myrtle Beach International Airport.  It\'s actually quite easy to get to, at least compared to Newark.  We follow signs for Car Rental Returns and drop off the car in the lot.  The guy initials my paperwork, and I go inside to get my receipt.  Wow.  That\'s a long line, and I just have to drop off.  This sucks.  The line moves very slowly because everyone in front of us is picking up a car.  We move about halfway thru the line, when one of the girls at the desk says, \"Is anyone returning?\"  Yes.  That takes care of that wait.  It\'s now 10:50 or so.    We zip down to  urlLink Continental  to check in our luggage and get our boarding passes.  This part was cool because they have a touch-screen computer to check in.  And there was no line here.  The guy checks our bag and the computer spits out our boarding passes.  Now we find the security check-in.  There is only one line, and it wraps around quite a while.  So more waiting, and I get antsy because this could take a while.  Luckily, the line moves at a decent pace.  Same routine as Newark-shoes off, etc.  We\'re through and have to find Gate 2.  That\'s funny because there are only 5 gates at this airport, as opposed to the hundreds at Newark.  Easy.  We spot it, and sit down.  It starts raining harder.  We do not want to be stranded here.  Just get our plane here and let us leave.  While we wait, I head to the gift shop to kill some time.  Our plane arrived a few minutes late, but not much.  We just had to wait a little for them to refuel and check everything.  We began boarding at about the scheduled time, and take off on time as well.    Once we got to a higher altitude, the flight was pretty bumpy.  There was a lot of wind.  The plane shook one time that made me stop breathing for a moment.  We were watching a reel of George Lopez and news on their monitors, and after this one pocket of turbulence, the plane shook pretty roughly, and the lights dimmed and the monitors went back up and shut off.  Then everything was ok again.  We continued to hit a few bumps.  When we were nearing our home, the plane insisted on dipping up and down, and turning on its side a lot.  This was making me anxious, because I just wanted to land already.  I would look out the window, and he\'d be ascending again.  Dammit man, just land already.  The pilot\'s interest in roller coaster caused a passenger behind us to vomit on herself.  We finally landed very roughly and got the hell off the plane.    We strolled through the airport to find our ride and our baggage.  We got home around 3 or so.  Happy for now, but knowing I\'ll soon be bored of being home.   '),(18194,1826527,'05,August,2003','2003-08-05 00:00:00','    I noticed something very odd as I read Bri\'s blog - there is an ad for the Radisson Myrtle Beach on top of the page.  Very weird.... Soooooo, Bri reminded me of the sludge in the parking lot at the water park and it made me think.  The first thing that came to mind was X-Files.  If anyone is a fan of the show you will remember various people oozing black liquid from their eyes which was some sort of alien thingy.  Well that\'s what the sludge was like and it splashed up on my legs.  Uh oh.  Will my nose start bleeding like Scully\'s?  Will I become pregnant with an alien?  Was the sex good?  I guess not cuz I don\'t remember.  Just kidding!!!   '),(18195,1826527,'04,August,2003','2003-08-04 00:00:00','    Day 6  07.25.2003  Friday  Today is our last full day in Myrtle Beach, so is that bad or good?  It sucks balls down here, but I don\'t want to go home and be bored again and alone while my girlfriend works during the day.  Such a conundrum...  Since today is supposed to be partly sunny in the am, with thunderstorms later, we want to hit the water park, called Myrtle Waves.  So we prepare by lotioning and gather our towels, sandals, and money.  We drive there and get into line to pay.  The prices are rather hefty it seems.  Suddenly my love says she doesn\'t really want to be here because she got one of those feelings.  So I don\'t mind.  It\'s more money in my pocket, and we can still go to the beach, since we\'re already lotioned.  As we trek back to our car, we step in a puddle of sludge.  It wasn\'t just a puddle from the rainstorm the night before, because it felt slimy.  Disgusting.  We make it back to the hotel and go to the beach, which was very crowded, probably because it was Friday and almost noon.  We stake out a spot and sit there for a while and read our books.  We try the water since I begin to sweat from the humidity.  The surf is still a bit rough, but not as bad as yesterday.  We don\'t go in too far, just up to our waists, and still get pounded by waves.  After a while, I get tired of holding my suit on because the waves were ripping it off every time I got hit.  The thing was tied in a knot, but the water was strong.  We go back to our spot for a while, and decide to leave to get some lunch.  Besides, it was cloudy and looked like rain.  We cleaned off and tried this place I had spotted on one of our journeys called the River City Grill.  We walked in and there were peanut shells all over the floor and license plates hanging from the wall.  It felt like some weird prison movie.  But alas, we seat ourselves and order.  The food is actually really tasty, and the peanuts were pretty good too.  After lunch, the rain was still holding off, so we want to try some mini golf again.   We drive down King\'s Highway and find Hawaiian Cavern golf.  It was pretty empty.  We paid and played.  The course was really nice.  There was a sign that said the course is the site of the US Open of mini golf annually.  There was one hole that led out into their man-made lake, and a fountain sprayed mist onto you.  Refreshing.  The path of the course took you up a mountain, which allowed you to see over the horizon, which was actually blocked by all of the hotels anyway, but nice nontheless.  We went back to our room to dry our towels because they probably would not dry by tomorrow morning.  Unfortunately, we did not have enough quarters for the dryer.  I asked the front desk for change of a dollar and they acted like I was asking them to build the pyramids of Egypt.  While we waited for the cycle to finish, we had a drink, which ended up in part on the awning of the gift shop below.  When the drying was finished, it was time to leave for the conclusion of our trip.    I had made reservations the night before thanks to Kim for an 8:30 show at  urlLink Medieval Times .  We were told the doors open at 7.  We arrived a little before 7:30 with a little trouble finding the place, but we did manage.  We get the black & white knight.  All right.  Now we have to wait for an hour to get in, and I\'m already getting hungry.  I had never been to one of these places, so I was a bit pessimistic and not expecting it to be much.  We finally begin entering the arena at 8:45.  We find our seats and sit next to 2 southern girls who warn us they are going to be obnoxious and scream a lot.  Actually, it sounded like home, so we were happy, and actually joined them.  Anyway, the food was incredible and abundant, and the show was really fun.  Our knight actually won the competitions, but this one knight who has kind of like an assistant to the King betrayed the king and slayed our knight instead of honoring him.  Bastard.  He died anyway, so just as well.    So our night ended happily, as did the week, but we still have to leave tomorrow, so I\'m sure something will go wrong.  Our flight leaves at 12:48 for now, but there\'s a tropical depression brewing right to the south of us, making it\'s way from Georgia.  Joy.   '),(18196,1826527,'04,August,2003','2003-08-04 00:00:00','    Day 5  07.24.2003  Thursday  I am eager to hit the beach early today because it\'s supposed to rain later today and tomorrow.  We leave Saturday.  We wake up relatively early and get ready for the beach(i.e. lotion).  We actually find a nice spot, and there is some sun today, but there are dark clouds looming.  We sit around for a bit, but don\'t actually get in the water because the surf is rough from the storm.  We hear rolls of thunder, but decide not to pack it in just yet, because it\'s off in the distance.  We leave a while later, clean off, and I go over to Subway for some lunch and bring it back to the room.  It rains.  We let the rain wear out for a while and decide to finally try some mini golf.  We check out Captain Hook\'s.  It is themed from the Peter Pan movie.  There aren\'t too many people playing, which is good, because it looks like it\'s going to rain again.  We had a pretty good time playing, and decide to try the Outback Steakhouse again for dinner.  We pull in and see a line, but we said we would wait tonight.  I go around back, but there is nowhere to park, so we leave and travel a bit further down the road to a place called the Atlantic Grille.    I run in through the rain to see if there\'s a dress code.  They say no, so we\'re in.  The place is not crowded, and it looks very nice.  We ended up having a really nice dinner, so we didn\'t wait in the rain for the Outback anyway.  We couldn\'t play any more mini golf since it continued raining, so we went back to the hotel to try and find its advertised \"lazy river.\"    We get changed and find it rather easily, and it contains only a boy floating around with his mother sitting in a chair alongside.  We try the pool outside instead, but it\'s too cold.  We go back to the river.  Great, we each grab a tube, and sit our lazy asses in and proceed around the river without having to strain ourselves.  Minutes later, another girl and guy enter.  Fine.  They were nice and minded their own business.  About 20 minutes later, a father and his 2 young boys run in all loud and jump into the river, without a tube.  The boys decide to swim between us as we float, and the stupid father insists on floating up our asses.  How very rude and inconsiderate we felt.  We stayed only a few more minutes and then left in a bad mood.  We went back up to our room and changed again and settled in for the night.  Tomorrow is our last day.  We plan on trying the water park, because it\'s supposed to rain later in the day, but not early on.  We\'ll see what happens.   '),(18197,1826527,'03,August,2003','2003-08-03 00:00:00','    Sorry for the delay.  I was at my brother\'s house helping install a new fence in their backyard for the last 2 days...  Day 4 07.23.2003  We start our day the same as yesterday, that is, eating cheerios.  We decide to peruse the hundred or so factory outlet stores a little bit up Highway 501.  We walk around a bit, disinterested in most of the stores.  I happened to see a music store as we were searching for a Tommy store for women.  I went in to see if (ha ha) they had the new 311 cd.  Wow!  They actually did, and I picked it up and realized their tax is less than New Jersey\'s on cds.  We continue walking and stop in our original destination to find nothing.  We continue on to the Nike store, which was probably the busiest store there.  We proceeded to the snacky stand to buy a pretzel to hold our appetites until lunch, which we were looking forward to.  We consumed our treats and entered the Big Dog store, where I purchased my dad a T-shirt.  We began making our exit walk from the outlets, but stop in an Aeropostale, which had select good prices on mostly stuff people wouldn\'t buy.  We also stopped in Nautica and Docker\'s to find ugly clothing, and crappy prices, respectively.  We leave and want to eat lunch.  We want to hit up Captain George\'s, which is a seafood buffet.  We drive over and pull in the back parking lot, noticing it\'s rather empty.  We walk around looking for a back entrance, only to find a few stairs that lead to a wall.  We turn and walk around to the front, where we discover that they open at 3.  It\'s 2 now.   Motherf...  Since we are right across the street from  Broadway at the Beach ,  and we\'re hungry, we go there to find food.    We park and find the info map.  I see two places, a Japanese steak house, and Joe\'s Crab Shack, which both appear to be behind the Hard Rock Cafe, just over a bridge.  Well, there was no bridge behind the Hard Rock Cafe, so we walked very briskly to the other side of this place.  We go past Joe\'s, because it looked like a place where only white trash is welcomed.  We walk a bit further and find out, hey, this steak house doesn\'t open until 4.  OK, let\'s try Joe\'s since we\'re hungry.  We go back and enter to find a line with a wait, and no ac.  Motherf...We continue back to the other side of the park, where we started and become very sticky in the process because the humidity is near 100%.  Disgusted, we get in our car to drive away.  We end up going back down Highway 17, but I forgot that there\'s nowhere to eat really on 17, so we drive a while before I can find a u-turn.  When we do, we turn around and I drive to the first place I could think of.  Fuddrucker\'s is always open during the day for lunch.  We arrive and there is room to sit, and we eat lunch around 3:30 or so.  We drive back to the hotel a while later and notice the sky is black behind us.  Click.  I take a picture.  We go to our room to chill and digest our disgusting lunch.    Tonight we are going to see some baseball.  Oh wait, no we\'re not.  It starts raining and getting very windy.  We decide we should cleanse our nastiness because we got really sweaty from our walk around Broadway.  When I got out of the shower, I turned on the weather forecast to find that we had a Tornado Warning.  Oh great.  It was around 6.  A tornado had hit maybe 10 miles to the north of us, and the cell extended to the south of us.  Great.  A tornado and we\'re in a high-rise with windows everywhere.  Luckily, the tornado warning expires without damage to us, but it continues to downpour.  I call the Pelicans ticket office to see if the game was still on, because my tix were purchased online.  The guy said, well, it may not start on time, but we\'re still going to play.  Huh?  Anyway, it continued to rain thru 8:30, so I don\'t think they got the game in at all.  When we were finally ready, we started getting hungry, so we tried to go to Captain George\'s again.   When we pulled in, there were not many spots, and there was a cluster of people outside waiting.  I ran in to see what the wait was, but the hostess said none for 2 people.  Nice.  I ran back to the car, got Kim and we were going to eat.  This place was awesome inside.  It was decorated with models of ships and fish, and the ceiling had a deep blue color with stars on it.   We made our way to the buffet to eat crab legs, some seafood medley, deviled crab, and something else.  Devoured.  Back for round 2, where we had some scallops, shrimp, some fish fillet, potatoes, rice, and a corn muffin.  Also devoured, but with fullness setting in.  We go back to browse the dessert table and pick up some fruit, pudding, cookies, and apple cobbler.  This was a good meal.  Expensive, but worthy.  I just wish I could have eaten more.  It was late, so we just went back to the room afterwards.  I wanted to not sleep in tomorrow to get to the beach before it rained, since we only had 2 more days left and it was supposed to rain both days.   '),(18198,1826527,'01,August,2003','2003-08-01 00:00:00','    Since Brian is away for the day, I would attempt to give you the details of our fourth day of vacation but my memory is shot.  Probably because I don\'t want to remember a lot of what happened.  Wow.. I just made vacation sound really bad which it wasn\'t.  It was good, bad, and full of ugly weather and people.  I guess they can\'t help it that they\'re so beastly looking.  I blame the parents who beat them with the ugly stick.  My theory is as follows:  The parents themselves are hideous beings.  When they learn that they are having a child they are afraid the child will turn out to beautiful (in the case of reproduction 2 uglies make a pretty).  So then the parents then make the decision to purchase an ugly stick so the child will live a hard knock life.  What?  I\'m so mean.  I\'ll stop now....  I\'m listening to the new Cursive cd which is quite good.  I have never heard anyone use a cello as a key instrument in an \"emo\" song.  It makes for some unusual yet pleasurable listening.  And the best part is that it sounds different from the other stuff I\'ve been listening to.  I\'ve actually been listening to music I never thought I would lately, straying away from my usual Hey Mercedes, Jimmy Eat World, Saves the Day, etc.  Change is good I guess.  Among these different cds are The New Pornographers, The Notwist, and Via Tania.  Very interesting stuff that Notwist.   '),(18199,1826527,'25,September,2003','2003-09-25 00:00:00','    I had my Neurobiology presentation on Wednesday and rocked it.  So I\'m very happy it is over with.  I booked my flight to New Orleans for Neuroscienec conferenc.  I have a 2 hr layover in Houston which sucks, but I\'ll be on Bourbon Street by 10 on Friday.  Our hotel used to be a \"historic bordello\", which should make things interesting.  This weekend we will be celebrating my friend Alex\'s birthday somewhere in Chapel Hill.  I noticed that the new Limp Bizkit CD is out and I was wondering to myself why these guys still have a record contract.  They are the suckiest sucks who ever did suck.  In other sports news, the Dolphins have a bye this week, but I will be at the Giants-Dolphin game next weekend in NJ. Hopefully they can carry the momentum till then. Rutgers is looking to go 3-1 this weekend if they beat Navy and equal their win total from the past two years.  On Sat, Carolina is going up to NC State and will probably get their asses kicked. At least basketball season will be here shortly and Carolina should be a top ten team.  Congrats to the New York Yankees- 6 straight division championships!!! and Brian for finishing the GRE\'s.  Also, happy belated birthday to Kim. Have a good weekend everyone!!   '),(18200,1826527,'23,September,2003','2003-09-23 00:00:00','    Well, the GRE is done.  Who knows what will happen next?  The only thing I know is that, based on my practice and on the real thing, I probably should have gone into the math field.  Not gonna happen now.  I\'m glad it\'s over.  I\'m going to lunch with  Kim  as per our usual Thursday event.  I\'m starving, and tonite I\'ll probably have a drink or two to fry the brain cells that are storing any analogy information.  I\'m just about finished with my  urlLink bands  page, in which you\'ll be able to click on any band name listed there and find a bio and discography.  It may not sound like a lot, but it was a lot of work.  So feel free to check it out if you want to know about some of the music I like.    Hey, how about those  urlLink Jays  finishing up on a strong note?  They will finish above .500 for the season, which is something to start with for next season, after adding some pitching.  As for the Bengals, I predict they will win outright this Sunday.  No point spread, just win.  As for the Cuse, well they\'ll lose to Toledo by 14.  I may post more tonite since I\'ll probably be bored.  Maybe I\'ll be inebriated when I blog for the first time.   '),(18201,1826527,'23,September,2003','2003-09-23 00:00:00','    Listening to the  Blue Album .  Brings back memories of, well, I don\'t know, but it\'s one of my favorite cds.  I just submitted my app to grad school online, only to find out I still have to write yet another essay for a scholarship.  That will not be completed until after I find out whether I\'m accepted into the school or not.  It\'s due by May 1st, so I have time.    Last night, I took  Kim   urlLink here  for her bday dinner.  It was pretty good, but there was hardly anyone there besides us.    SATURNINE:  A) magnanimous B) ebullient C) finicky D) unnatural E) impoverished  Sorry, but all this studying is making me crazy.  On my final practice test, I did not do very well on the verbal, but quite well on quantitative.  Pretty strange for one who enjoys writing.  The part I thought would be the easiest turned out to be the worst.  There\'s a part where you fill in two blanks in a sentence using context, and just about every one has two close answers according to the book, and I picked the one that wasn\'t the best fit for every answer.  I did well on the antonyms and reading, with some trouble on analogies.  The math surprised me.  I was happy I did well, though.  I took my time and it paid off, making sure I checked all possibilities, like the book said.  So today, I will study some more-brush up on the math formulas once more, and look over some word roots.  Tomorrow equals rest and no studying.  Here\'s to the  urlLink Seattle Seahawks  for being 3-0, and so far, living up to my  surprise team  label in my scholarly NFL predictions.   '),(18202,1826527,'22,September,2003','2003-09-22 00:00:00','    Thanks for the song!  And I\'m glad you didn\'t sing it.  Just kidding!!  My birthday is okay so far for all of you who were wondering.  I can\'t wait to eat my fancy dinner tonight!!   '),(18203,1826527,'21,September,2003','2003-09-21 00:00:00','    Happy Birthday,  Kim !  I\'ve written you a bunch of songs in the past, but the following song seems as personal as a song I could have written for you today.  So here\'s the lyrics, since I don\'t think you\'d want me to sing them:   The Luckiest   lyrics by Ben Folds   i don\'t get many things right the first time in fact, i am told that a lot now i know all the wrong turns, the stumbles and falls brought me here and where was i before the day that i first saw your lovely face now i see it everyday and i know that i am, i am i am the luckiest what if i\'d been born fifty years before you in a house on the street where you live maybe i\'d be outside as you passed on your bike would i know? in a wide sea of eyes see one pair that i recognize and i know that i am, i am i am the luckiest love you more than i have ever found a way to say to you next door there\'s an old man who lived into his nineties and one day passed away in his sleep and his wife, she stayed for a couple of days and passed away i\'m sorry i know that\'s a  strange way to tell you that i know  we belong that i know that i am, i am i am the luckiest   '),(18204,1826527,'21,September,2003','2003-09-21 00:00:00','    For those who were worried, I did not get wiped away with the hurricane, the only problem we had was a loss of power for a couple hours Thursday afternoon.  I would have to say it was a good weekend, won $15 playing poker with some friends/classmates Friday night (sorry I didn\'t see your blog earlier in the week Dennis), Ohio State won, and more importantly, Michigan lost.  Oh yeah, I also went to see two pillars of college gridiron tradition, Duke and Northwestern, battle it out to see just who sucks more, and it turns out, Duke does. I\'ve seen them a couple times and they\'ve never looked good, but they looked exceptionally bad on Saturday. I was able to catch American Splendor last week and enjoyed it a lot, and while there saw the trailer for Lost in Translation. I\'m automatically sold on any Bill Murray project, so they can count on my movie dollars when it eventually gets released in NC, probably sometime in Fall 2004. I\'ve decided to stop doing very much work for school, I\'ve done nothing but PT related stuff for two years, so it\'s time to collect my debt. I\'ve read a lot more, and just finished Requiem for A Dream by Hubert Selby, Jr (thanks to Isabel), as per usual, as good as the movie is, the book is much better. Well, there is some reading to do for class tomorrow, so I should get on that. Ending blog transmission.   '),(18205,1826527,'20,September,2003','2003-09-20 00:00:00','    Hello everyone!  Last night me and Brian went to the  urlLink Stress Factory  in New Brunswick.  We saw the very funny  urlLink Ted Alexandro .   We\'ve seen him on Comedy Central a few times and he was great so when Brian read that he was going to be nearby he decided to take me for my birthday.  The guy hosting the show was really funny as was the girl that went on before Ted.  All in all it was a great night.  Well except for when we kinda got lost.  That was the first time AAA directions let us down.  Thank god we had directions off the Stress Factory\'s site.  I also didn\'t like the 2 drink minimum rule.  Since I\'m still not 21 I had to order 2 waters.  I understand that they need to make money but I don\'t think 2 waters will net them much profit.  Oh well.  I fun anyways and that\'s all that counts.  Hooray for the Puke Fairy!!!   '),(18206,1826527,'20,September,2003','2003-09-20 00:00:00','    Afternoon all.  Time for me to review the movie I saw last nite.  The  urlLink girlfriend  and I went to see  urlLink Thirteen  last nite.  Well, the evening started with the nu-metal ticket kid telling Meredith that  urlLink American Splendor  was a much better movie.  He gave this recommendation because he saw that she was wearing her NIN shirt...  We decided to see the movie anyway.  The movie was co-written by one of the 17 year old stars and the movie\'s director.  It was the writing, acting and directing debut for both.  And I was thoroughly impressed.  Yes, the movie has some raw spots, but for the most part it seems very true to life (or at least thats what I was told by my girlfriend, since I was never a teenage female, its tough for me to relate).  My only complaint is that the movie should be called  Sixteen  or  Seventeen .  While the characters of the movie are supposed to be in the 7th grade, I highly doubt that the situations that occur in the movie would ever happen to a 13 year old.  It seems more realistic for the girls to be a few years older.  But if you go into it with that \"willful suspension of disbelief\" thing, then you should enjoy this movie.  I also recommend going to see it with a member of the female persuasion.  This is a definitely \"chick flick\", but not the true sense of the genre; but will earn you some points with said female.  And you should enjoy the movie to boot...  Well, its time to watch the Yankees on YES and wait for tomorrow\'s football games...   '),(18207,1826527,'19,September,2003','2003-09-19 00:00:00','    Hello everyone, I\'m stuck at work until midnite running an experiment so I \'m bored and decided to write another blog, so soon after my last post.  After this night I\'ll feel like Lemmywinks the Gerbil King (guess the reference and win!!!).  Anyway, my Neurobiology class sucks big time.  There\'s about 80 pages of reading per day, plus weekly presentation where we get our asses reamed by professors assuming we know as much as them even though they\'ve been in the profession for 30 years or so.  Carolina is going to get killed by Wisconsin this weekend in football, but at least the Yankees are going to win the AL East and squash the pathetic Red Sox.  In other news, I\'m about to finish another one of Stpehen King\'s Dark Tower books. (Book III The Wastelands)  The books are pretty good although a lot of weird stuff happens.  Tomorrow night, I\'m hitting up the Chapel Hill bar scene (Matt, call me if you want to join).  Hope everyone survived the hurricane.  Kev- maybe you need some more pieces of flair to help your review.   Later, Dennis   '),(18208,1826527,'18,September,2003','2003-09-18 00:00:00','    What you need is motivation.  About those TPS reports, did ya get the memo? Sorry, those were moments from  urlLink Office Space .   I\'m sorry about your shitty company, Kev.  Have any job openings?  I added some more links on the left.  Nothing you probably haven\'t seen before, just stuff that\'s been blogged about before.  I was bored and needed a break from studying.    I\'m listening to this band  Camden  right now.  They\'re very mellow, almost like my boys,  urlLink The Appleseed Cast , but not quite as good.    I strolled around my hood to investigate any damage from that whirling bitch, and not much.  But that hurricane didn\'t do much damage either.  The only thing I noticed was someone\'s gutter bent into a \"u\" shape around their front porch post.  Lots of leaves and twigs, but nothing to get worked up about.    Well, folks, it\'s pizza nite.  Enjoy the weekend.   '),(18209,1826527,'18,September,2003','2003-09-18 00:00:00','    OK, I need to vent a little, so feel free to bypass the next paragraph...  I have zero repsect for my bosses.  They were doing our \"reviews\" this week and I just got out of mine.  What a friggin\' crock of sh!t.  First of all, this wasnt even our offical review, all we did was review our \"personal objectives\", our real reviews arent until December.  They did tell me what percentage raise I am getting this year, which was more horsesh!t.  I am getting a 3% raise, which is average for my entire group.  What pisses me off more is that these raise figures were set back in February, when our program didnt kick into high gear until April.  So, that means that all the work I did this summer will be forgotten by the time they set the next budget for raises.  They told me that my hard work didnt go unnoticed, but of course I wont be compensated for it.  This company is a friggin\' joke.  I also have a strong feeling that the other two people who do the same job as me got higher percentage raises; and I know that I do much higher quality work than they do.  F this company...  All done.  Now back to your regularly scheduled blog...   '),(18210,1826527,'18,September,2003','2003-09-18 00:00:00','    Thank you Isabel for knocking out power at MSU.  Although I was a little upset that I went to campus at 6:30 this morning only to be told that classes are canceled.  So I came home and went back to sleep.  Good times doods.   '),(18211,1826527,'18,September,2003','2003-09-18 00:00:00','    Evening all.  I just got back from \"Wing Nite\", so I am full of spicy bits of chicken and cheap beer.  We decided to sit outside in bask in the pre-hurricane festivities.  It was a rather enjoyable nite, albeit a tad windy...  I guess I am the only person who wasnt overly thrilled with  Pirates of the Caribbean .  I thought it was \"ok\", but nothing special.  I would like to see  Lost in Translation , but it is only playing at two theatres in CT, one that is 30 mins from me and another that is 40 mins.  Maybe I\'ll go see it tomorrow nite anyways.  The  urlLink girlfriend  has wanted to see  Thirteen  for some time now,  so I think I may be seeing that tomorrow nite instead.    The last book I read was  Lullaby  by  urlLink Chuck Palahniuk .  I thought it was really good (but not as good as anyway of his other books).  It took me less than two days to read, which is saying a lot coming from me.  I really couldnt put it down.  And for those of you dont know who Chuck is, he wrote the novel  Fight Club .  Well, hopefully Dennis and Matt stay safe and Isabel leaves the rest of us alone.   '),(18212,1826527,'17,September,2003','2003-09-17 00:00:00','    Hello all!  I just got home not too long ago from class and it sucks outside.  It\'s way windy and pouring on some streets and not raining on others.  I went through five periods of rain/no rain during a 10-15 minute drive.  My car actually got blown around a little on the way home too...  The wind is blowing my blinds all around.  I must agree with Bri about the eerie feeling.  Very weird.  Isabel sucks, but not for the obvious reason of rain and wind.  I still have school!!  We are in a state of emergency and the stupid university decides to keep classes in session.  MSU never closes.  There could be 2 feet of snow on the ground covered by 6 inches of ice and they\'ll expect us to go to class.  Absolutely insane..  As far as movies are concerned, I saw the pirate movie with Bri not only because Johnny Depp was in it but because it was a fun movie.  I thoroughly enjoyed it.  I don\'t believe we\'ve gone to see anything else because all the movies that have been released hork major ass!    I hope our fellow bloggers and friends are doing okay in Carolina.   '),(18213,1826527,'17,September,2003','2003-09-17 00:00:00','    Rob I second Brian\'s satement on your wallet.  250 is like a month\'s pay for me here.  As I\'m writing this, I hear 50 MPH winds gusting outside.  Luckily the hurricane has passed through Chapel Hill.  No damage to my place, but we did get a day off from work.  The coast did get hit pretty bad though.  Brian-I used the Princeton Review GRE book and it helped me somewhat.  Also, register like 2 months before, if you plan on going to the Verona testing center.  When I went on interviews, they really didn\'t dwell on the GREs too, so I don\'t think its critical you get a great score.  There seems to be movie theme here so I\'ll throw in my 2 cents.  Pirates-very good.  Lost in Translation-good.  Once upon a time in Mexico-haven\'t seen it, but I heard Johnny Depp is good in it.  BTW October 21-Indiana Jones on DVD- finally myortie movies on DVD.  I gotta read for my class tomorrow.  Talk to you later  Dennis   '),(18214,1826527,'17,September,2003','2003-09-17 00:00:00','    Just got back from lunch with  Kim .  She had a break after her morning classes and before her night class.  Rob, I\'m sorry to hear about your wallet.  That sucks horribly.  Thanks for the advice on the practice tests.  I am taking some practice tests in the book I purchased, and they are difficult, but hopefully, my brain will expand the more I study, and take in what I\'m getting wrong.  As far as the movies, I\'ve only seen  Pirates  thus far, but didn\'t plan on seeing  Jeepers  or  Dickie  anytime soon.  I would definitely recommend my first mentioned as well.  Yarr.    Good news is that the advisor from MSU mailed me back and gave me news I wanted to hear.  I also got a response from my fave professor from WP, and he is willing to give me a recommendation.  Nice.  I\'m thinking about Den & Matt right now in NC and hoping things are OK...  I finished a very good  urlLink book  last night.  I highly recommend it if you want to read a good thriller.  I don\'t read many horror works, but this one gave me the creeps at times.  Probably the fastest I\'ve ever read a book.  I think I read it in 3 days.  I had mentioned a book before in this  urlLink blog  about a month ago, which was the book I read before the above mentioned.  It royally sucked, and I still don\'t understand the beginning that I blogged about.    I put a new welcome picture on my site.  The link\'s on the left.  Other than that, not much else is new.  I haven\'t written any new songs, but I\'m working on an idea for a short story.  As I type, the wind is whirling thru the vent in my ac.  I don\'t know, when I was outside, it just feels erie.  Yeah, it\'s windy, but it feels like there\'s some supernatural power.  There\'s a tingling on my skin and my hands felt sweaty.  Weird.  I\'m going to  urlLink eat  now.  Bye.   '),(18215,1826527,'17,September,2003','2003-09-17 00:00:00','    Hola everyone.    What can I tell you all?   Well, I went to CT on Friday to play poker etc.  On the way back to Boston on Saturday, I left my wallet in the bus station in Springfield.  That was great.  I also happened to have $250 in my wallet for no real reason other than it was my \'gambling money\'.  Not suprisingly, my credit card company was alot more accomodating in sending me a new card than my bank was.  However, I did have no cash or ID until Tuesday either way.  My large advice in life is don\'t lose your wallet, cause its a pain.     I figure I can give you all movie news as well.  My theater happened to have  urlLink Lost in Translation  starting last Firday as part of its limited release.  I saw it and thought it was quite good.  There\'s a lot of critical hype about the movie, but I think its somewhat deserved.  Bill Murray is simultaneously quite funny and believable in a role that must have been created for him.  That and Scarlett Johansson is hot.  These two insomniac Americans meet in Japan in a hotel and become friendly.  Anyhow, I recommend it fairly highly a its much better than anything that came out this summer.  Other movies that are perhaps worth seeing include  urlLink Matchstick Men  (it\'s a con movie, if you like con movies, its worth seeing),  urlLink American Splendor    (it\'s a non-super hero comic book movie, if you like non-super hero comic book movies, its worth seeing), and  urlLink Pirates of the Caribbean  (it\'s a pirate movie, if you like pirate movies, its worth seeing).  However, I was forced to \'hide\' last week when the regional manager was in the building as an effort to make it look like we need more budgeted payroll (absurd), and I happened to try to watch Jeepers Creepers 2 and Dickie Roberts, but I wanted to vomit immediately upon seeing scenes from both.  So bad, words can\'t describe.  Word on the street is that Once Upon a Time in Mexico also blew, though I haven\'t seen it.   Ummm..... yeah that\'s my movie story.    Oh, my story wasn\'t published in that paper of course, it was the last thing cut adn they wanted me to write something this week, but I didn\'t have time and my power went out on Sunday to boot.  Speaking of booting, Jeremy Shockey makes me want to boot.  What an idiot piece of crap.  There\'s no particular reason for that comment other than I\'m quite sick of all his media attention for such a non-great player.  BBBBBBBBBBBBBBBBoooooooooooooooorrrrrrrrrrrrrrrrriiiiiiiiiiiinnnnnnnnnnngggggggggg.  I recommend taking practice tests a few times before the real GRE.  They helped me.  Alright, I\'m off to play Madden now, Blog out.   '),(18216,1826527,'17,September,2003','2003-09-17 00:00:00','    yeah, its funny how times have changed.  now, i am more about the grain alcohol than the band...   '),(18217,1826527,'16,September,2003','2003-09-16 00:00:00','    Dude, the GRE is making me crazy.  I got one of those practice books yesterday to aid in my studies/practice, but man, it just made me more worried.  Some of the words are absolutely ridiculous.  So I\'ve got a lot of work to do before next Thursday.  I read and studied a lot last night, but my brain is fried so far today, and I can\'t really concentrate well.    I emailed the advisor at  urlLink Montclair State  and asked him some questions I had about grad school.  Hopefully he\'ll reply with things I want to hear.  My plan if I\'m accepted into the school is to pursue  Communication Arts, which could entail either Public Relations or Speech Communication.  I\'m hoping that upon completion, I may be able to teach either high school communication classes or college courses.  If not, then I may continue with this education fiasco and pursue a doctorate.  Yeah.   You  can call me doctor, bitch.  Well, first I have to actually get into the school.  So I\'m just hoping to pass this test first.   Everyone say hi to  Kev .  I was just asking about him not too long ago, when I was listening to an Everclear cd.  He was my everclear buddy in high school.  The band, not the alcohol.  Welcome!   '),(18218,1826527,'16,September,2003','2003-09-16 00:00:00','    Greetings from Connecticut, otherwise known as the waste of space between Boston and New York.  I dont have much to say as of right now, since I am rotting away here at work.    The only interesting thing going on with me is this:  urlLink www.rep-am.com/top/6l-f.htmle   I work(ed) with this guy.  I mean we\'ve worked on the same projects and his desk was 50ft. from mine.    Talk to ya later...   '),(18219,1826527,'16,September,2003','2003-09-16 00:00:00','    Hey everyone, Back from Columbus, Ohio.  Man, what a crazy weekend.  After a 9 hr drive (damm Virginia cops!), we made it to Ohio.  Went to downtown Columbus, right by where the Blue Jackets play.  After some good old Irish ales, we made fun fun of some NC State fans and went home to rest for the big game.  Approximately 8:45 in the morning, we wake up and begin our nutritious breakfast of Bud Light and Bratwurst.  After tailgating for a few hours, we enter the game along with some airplane bottles of vodka.  Fast forward, OSU is winning 24-7, when all hell breaks loose.  NC State comes back and ties the game.  After 3 OT, OSU prevails 44-38.  What a game.  We then precede to drink more after the game until my liver starts to hurt and we go home.  Sunday we drive back to Chapel Hill.  All in all a fun weekend.  Next travelogue will be from New Orleans in November for the Neuroscience Conference.  Free trip, no girlfrined to bring along (I told Khris she can\'t come),  hotel on Bourbon St, no open container laws, what a city!!! Well I\'m about to brace for the hurricane.  Talk to y\'all later Dennis   BTW I like the Family Guy quite so here\'s one of my faves \"Guns don\'t kill people, minorities with guns do\" from the NRA episode   ');
INSERT INTO `msgs` VALUES (18220,1826527,'15,September,2003','2003-09-15 00:00:00','    I think I\'m starting to get sick.  My throat is bothering me and that\'s usually my first symptom of illness.  Everyone at work is sick.  I think it all started downstairs on the teller line and it spread from one girl to 5 and then the girl in my department is sick.  I think it\'s like the movie Outbreak where the mutaba virus spread thorugh the hospital via the vents.  I seriously doubt that there\'s a cut little monkey, or an evil one for that matter, crawling around the bank.  \"Who\'s going to protect me from the evil monkey in my closet?\" (all you family guy freaks will get this quote) I spray Lysol everyday but I think I\'m getting it anyways.  Now isn\'t really the best time for me to get sick with school and work and all that good stuff.  Oh well.  Hey who\'s ready for the hurricane?  I hope it isn\'t a hurricane - just a tropical storm strong enough that Montclair State floods and they cancel classes.  I could dream right?   '),(18221,1826527,'12,September,2003','2003-09-12 00:00:00','    OK, so I\'m really bored today.  In the last issue of SI, there was an article on anagrams for sports teams and players, etc.  So I went to the  urlLink site  that the writer used and found some anagrams for my name, such as  Arab birth drunk  and  Rabbi drank thru .  So feel free to goof around.  It gives you a mess of possibilities.   '),(18222,1826527,'12,September,2003','2003-09-12 00:00:00','    The weekend is over.  Today it\'s gloomy, and we might get hit by a  urlLink hurricane  on Friday...  Happy Birthday, Mom!  Sorry I never posted, but it was Saturday, and I didn\'t blog then or yesterday.  The Bengals lost again, but this one was a heartbreaker.  The game was theirs, but like I said before, they don\'t know how to win a game.  At least Pitt got fucking dismantled!    I\'m applying for probably my last attempt at a teaching job in West Caldwell for a permanent sub position in 4th grade.  I\'ve got experience with this, plus they advertise a benefits package.  It would be from Nov-June.  I\'m keeping my fingers crossed.  I must purchase Family Guy Vol. 2 on DVD.  Anyone have 40 bucks I can borrow?    I intend to clean my basement today.  I don\'t know why, but I have nothing better to do.  Throw away old junk, give away useful games and/or toys, etc.  I don\'t mind the getting rid of stuff part, but I\'m afraid of monstrous spiders or something attacking me when I remove our copy of   Clue  from the old bookshelf, which has accumulated 11 years of dust.   '),(18223,1826527,'12,September,2003','2003-09-12 00:00:00','    I rose bright and early at 6:30 to start my day, unfortunately, it was at the dentist\'s office.  One of the places I have learned to hate as I have gotten older(or wiser).  They started by putting me through a series of about 16 X-rays.  Normally, there\'s 4 that they take, but these kept coming.  I thought they were taking an X-ray of each tooth, but they finally stopped.  Then the hygienist began scraping and poking at my teeth and gums.  I know everyone knows this pain, but have you ever seriously  looked  at the instruments they use?  Those things are evil.  Every tool had a sharp point except for the angled mirror, which instead had several sharp points.  Really, this one used a jagged piece of glass for its mirror.  Anyway, at least she was gentle and I didn\'t have any cavities.  On my way out, I stop to see the receptionist.  \"How\'s March 16th sound?\"  Great, but I\'m sure something will come up between now and the six months that come between then.    Has anyone else seen this Hi-C commercial with the dancing penis?  Oh, wait that\'s supposed to be a tongue?   '),(18224,1826527,'11,September,2003','2003-09-11 00:00:00','    It does sound nice indeed.   '),(18225,1826527,'11,September,2003','2003-09-11 00:00:00','    just a ponderance...where do they come up with these ads atop the page.  i just saw  further seems forever  and  stain remover .  What do the 2 have in common?   '),(18226,1826527,'10,September,2003','2003-09-10 00:00:00','    You\'re damn right it\'s nice of me!  Fuckin\' A!   '),(18227,1826527,'10,September,2003','2003-09-10 00:00:00','    Buona sera fellow bloggers!  All is well with my classes except I was hammered with tons of homework in the first week.  Welcome to college is what they tell me.  And I tell them F*** YOU!!  Not really.  I can\'t be that mean.  Classes actually haven\'t been that bad thus far.  The only downside is Italian.  I think I\'ll do okay in the class but we have to do a lab which is  at least  an hour a chapter.  I actually did one today for the preliminary chapter and it was the most boring thing I have ever been through.  All you do is read along with the dude on the tape and then fill in the blanks on the conversation.  What\'s the point?  And they actually want you to say it along with them.  Right, I\'m going to sit in an empty room and repeat after the tape while people walk by and think I\'m crazy for talking to my imaginary friend...  And just so everyone knows I\'m turing the big 2-0 on the 22nd of this month.  I was considering doing a countdown but I thought it might be too much.  Brian said he\'s taking me to some fancy schmancy restaurant to celebrate which is very nice of him, right?  I can\'t wait til my 21st though.  Everyone must leave school/work to help me celebrate the next joyous occasion.   '),(18228,1826527,'10,September,2003','2003-09-10 00:00:00','    Holy crap!  I\'m officially applying to grad school.  I\'m scheduled to take my GRE on Thursday, September 25.  [Gulp]  I began filling out my app, and filled out my FAFSA in hopes of getting this shizzy going.  I still have to get references, transcripts, pass the GRE, and send everything in, but I\'m definitely motivated.  I hope I can do it, and everything works out well.    I thoroughly cleaned my car today inside and out.  I washed it, scrubbed the tires, vacuumed, cleaned the windows, and polished up the vinyl.  I could not, however get all of the tar off the bottom of the car by the wheels.  I used that bug & tar cleaner, but it\'s embedded in the paint job I guess.  My trunk has a tendency of leaking, so I also vacuumed out a puddle of water beneath my spare tire.  It\'s pretty gross actually.  The water gets all slimy and my jack and tire iron rusted pretty fiercely.    I\'m gonna go now, because I\'m meeting Kim for some lunch/dinner before her night class.   '),(18229,1826527,'10,September,2003','2003-09-10 00:00:00','    Sorry for the delay.  Things have been hectic since classes have started.  Yes, Brian, I did see the game.  I am cheering for UNC to lose all their games so we get a new coach.  As for the Miami \"We choke in December \" Dolphins (Hi Aquaman!).  I am now embarrassed to say they are my team after that craptacular performance against the Texans.  I am actually going to see them play against the Giants in October.  BTW, I will be home the 1st weekend in October if anyone wants to get drinks, food, etc.  Fortunately, my roomate and I are going on a road trip this weekend to Columbus Ohio to see Ohio State vs. NC State, which should be a good game. (He went to OSU as an undergrad).  In other news I saw Dashboard Confessional and MXPX at the Ritz in Raleigh this week.  Next show will be Built to Spill in October-very good indie rock band.  Bri, how is the job hunt going along?  Everyone can look forward to a recap of the Columbus trip in the next blog post.    Later y\'all   '),(18230,1826527,'09,September,2003','2003-09-09 00:00:00','    I attended a meeting at  urlLink GHT  last night because our season starts this Saturday, and came back with an abundance of mosquito bites and scratches on my arm.  I guess someone decided to play a little  Wiffle ball  before our meeting started, and the ball went into the shrubbery about 100 times, and each time I reached in, I came out with battlescars on my arms.  So I\'m quite itchy this morning, but trying not to scratch.    I want to go back to school to work on my master\'s, but I don\'t know what to master.  I\'m actually contemplating nixing the whole teaching area and going into public relations/speech/linguistics.  Not all 3, but one of those.  My major was in interpersonal communication.  I have lots of forms to fill out, and still have to take my GRE.  I\'m going to post the very last part of my story, but I also think it\'s still weak.  I like the ending, but as far as how to write it better and explain things more clearly, I\'m still working on it.  So hope you liked it if you read it.      Corey’s court date was a month later.  I didn’t go to the hearing.  Corey was found guilty.  There were traces of blood found at the murder scenes that matched his.  Now he’s in jail, probably waiting for the death penalty down the road, and I’m living happily with Sammy.  Isn’t love blind?  Now I’m living with Sammy and my ex-boyfriend is in jail for my crimes.  Corey didn’t really kill anyone.  I did.  Erika, Ashley, and Josie hurt my Sammy, and Andy stayed with that bitch Josie.  I had been with Sammy since Josie broke it off with him.  I knew him well now, and knew he was hurt.  They deserved what they got.  Like I said, Corey was careless.  He drove recklessly, and was always late for school.  He was always injuring himself-cutting himself.  How could you not find his blood?  I found plenty of it, and found good uses for it as well.  Now tell me, is love blind or is justice blind?   '),(18231,1826527,'09,September,2003','2003-09-09 00:00:00','    It\'s 9:03am.  Wow, that feels early.  This is the result of working in a movie theater until 2am regularly.  If you are looking for action and excitement in the life of Rob, well.....  too bad.  Anyway, I am going to school again today to discuss my master\'s project with a professor.  He sent me an email with a few papers last week seeing if I was interested.  The stuff seemed right up my alley, so I told him so and promptly arranged an appointment to hammer out details.  When I show up at this time, He talks to me for literally 45 seconds, handing me 3 pieces of paper, one of which is almost a doodle.  We do not discuss what his actual interest is in this material sample he wants me to study, a timetable for doing anything, when I can use the lab, who might help me learn some of the lab equipment, etc.  Nice meeting.  Anyway, I go back today, so that could help.  As for the New England Sucktriots... they were the suckiest bunch of sucks who ever sucked.  I actually wrote an article about them on Sunday for some free sports newspaper in Boston.  The paper has been out for 2 weeks, and its one of the free things people read on the train.  I felt better when they wrote back to me on Sunday saying they got lots of articles and the only one to be printed was mine.  However, I don\'t get paid since the paper is in infancy stages.  Coincidentally, thats why I can write for them.  Also the paper comes out tommorrow, so I have no idea if it will really be there and if its all edited up.  This is the closest thing to news in my life.  Other activites of the past week include: making smoothies, buying a shot of wild turkey for a 21-year old photographer friend on his birthday, watching some street performer fit through a squash raquet with some confusing girl, leaving the last train of the night with the photographer so he could throw up some wild turkey among other things, listening to an entire Bruce Springsteen concert on my friend\'s roof and outside Fenway Park (very good indeed), sending most of the day employees home at the movie theater since we have no money since there are no customers, and of course making sure the photographer got home alive via buying a cab as to leave the remaining wild turkey on the sidewalk.  Blog out yo.   '),(18232,1826527,'08,September,2003','2003-09-08 00:00:00','    The bloggers\' football teams are a combined 1-3 after the opening weekend.  Is this a sign of things to come?  Yeah, probably.  Den, did you happen to see the Cuse/UNC game on Saturday?  It wasn\'t televised up here, but it seemed like a pretty awesome game from the score and stats.  I know the outcome wasn\'t what you would have liked, but exciting nonetheless.  As for fans in Bengaldom, their sheer stupidity goes along with the ineptness of the team the past decade.  I\'ve been hearing some fans wanted coach Lewis to put in Palmer because Kitna was awful, but Palmer wasn\'t even active.  He\'s not going to play until their season is officially over, like after 8 games or so maybe, when they\'ll be 2-6 or 1-7.  The team itself is the same old story.  They get down by 20 points in the first half, and the game is over, because their offense doesn\'t know how to win a game.  Pathetic.  I got a Hey Mercedes poster in the mail today for their new album, but I don\'t know why.  I\'m not complaining though...  Where\'s Kevin when you need him?  I\'m actually listening to an old Everclear cd right now.  Well, here\'s the next part of my story:    I quietly scaled the stairs and slipped into my room.  I grasped a business card and dialed Andy’s uncle.  I whispered to him that I suspected the killer to be in my house.  Corey must have heard the phone click down because he came running into my room to see whom I was talking to.  I said nobody.  I lied.  He knew I lied.  He looked at me funny.  I asked him how the chicken was coming along.  He said good, but he cut his finger again.  He always cut his finger.  He was careless.  He just stood there, holding the knife tightly, looking at me in puzzlement.    Boom!  The front door collapsed.  The police barreled into my room and whisked Corey away in cuffs.  They had saved me for now.  Andy’s uncle comforted me and told me everything would be OK now.   I know.  Two of my best friends were murdered, and I stopped him from killing again.   '),(18233,1826527,'05,September,2003','2003-09-05 00:00:00','    Today I went job hunting yet again, but realized once again that the mall is a horrible place.  Not much accomplished again.  I ate tuna for lunch.  I\'ve really grown to like tuna over the years, but still can\'t get over the fact that it comes in a can.  I\'m finally able to just pretend that it\'s not cat food.  I remember the first time I tried the tuna that\'s vacuum sealed in the pouches, but that made it seem even worse-tuna in a bag?  Anyway, it\'s eaten.    While on my journeys, I picked up a frame from Linens and Things for my autographed Face to Face poster.  Well, when I tried putting the poster in, I noticed the interior part of the glass was gritty and dirty somehow, so I took it out to clean it off with some glass cleaner.  Upon cleaning the surface, I found that the cleaner only removed the grime, but there was still streaks and blotches on the glass I could not remove.  So I tried to put the glass back into the frame, but somehow it did not fit.  I was able to remove it, but it seemed impossible to get back.  So I tried pushing the little metal clasps(the ones that keep the cardboard and picture in the frame) as flush against the wood as possible, since I realized the glass was hitting them.  After completing this task and brutalizing my thumb with the pushing, the glass fits, only to shatter seconds later.  I don\'t know how it broke, but it did.  So now there\'s a corner of the frame that just cracked right away from the rest of the glass.  I cut three of my fingers during this fiasco, and decided that i was not going throw out the frame.  I took some household cement(the same stuff that failed to hold my rearview mirror) and applied it to the shards of glass.  I set them in place, so hopefully it will hold.  The glass will still have two cracks that will be obvious, but right now, if it holds, I\'ll be happy.  I have to wait till tomorrow to put the poster in, though.  Well, it\'s the weekend, and it\'s actually really nice outside.  Woo!  Pizza tonite.  Here\'s the next part of the story:    I started seeing things in my mind.   Erika and Ashley.  Andy and Josie.  Then I saw Corey.  What did Corey have to do with them?  He knew Andy and Josie, but  what did that mean, and what about Erika and Ashley?  The six o’clock hour rolled around and I was actually getting hungry.  My grumbling stomach told Corey.  He said he’d help me put something together for dinner.  We had to eat something, otherwise we’d feel worse.  I looked in the freezer to find ice cream, waffles, and a package of chicken cutlets.  We decided to have some chicken.    I went to the basement to find something to eat with the chicken—creamed corn or some slop like that in a can.  I never wanted Corey down there because it was a disaster.  My parents left it in shambles when they moved into a new house.  I was embarrassed by the mess.  He started preparing the chicken.  After selecting my canned vegetable of choice, I made my way back up the stairs.  I caught a glimpse of Corey in the kitchen holding a knife.  The thinking resumed.  I saw things in my mind going in reverse order, like when rewinding a video tape while watching it.  The tape froze.  Play.  Before Corey and I got together, Corey tried to win over Erika and Ashley.  He was desperate before I met him.  He wanted a girlfriend so badly, and I gave him a chance.  Was the killer in my house?  Was he making chicken in my house with a very sharp knife?   '),(18234,1826527,'04,September,2003','2003-09-04 00:00:00','    Today was my first day back at school.  I wasn\'t very excited about this because I took 2 summer classes and I felt like I never the left the place.  Especially since I finished my last summer class exactly a week ago.  Oh what I won\'t do to graduate in 4 years.  Today was Accounting and Religions of the World.  Me? In a religion class?  FUNNY!!!  I think they\'re going to be fairly easy classes.  Tonight I have Business Math: Calculus.  All you math nerds can appreciate how excited I truly am to be taking a calculus class.  Next semester- Statistics!!!  When I got home from class, I took a glance at all the classes I still need to take and I was saddened at what I saw.  I still have a butt load of general requirements to take, among these being 3 contemporary issues classes.  Why do I have to endure this pain?  To be a well rounded person?  Who wants to be round?  Not me!!  Anyways, tomorrow I\'m up again at 5:30am to my Business Law and Italian classes.  I hope they\'re both somewhat easy especially Italian.  I took Spanish for 5 years and I still can\'t carry a conversation so maybe if I pay attention this time it will make it easier... Ciao!   '),(18235,1826527,'03,September,2003','2003-09-03 00:00:00','    I almost forgot a very important thing to talk about today.  I know the other bloggers here either don\'t care for hockey, or are devils fans, but I am a Rangers fan, and not the best hockey fan, but I was a huge fan of  urlLink this man .  And if you believe otherwise, look at the poster on my wall of the man from his first NHL season, where he didn\'t have the modern mask with all the painted designs on it.  It will be a sad time for me at 3:00 pm today, when  Mike Richter  announces his retirement.  I may shed a tear because I\'m a big puss, but the man  was  the New York Rangers for the past decade.  Sure, you can argue with Messier, but Richter was huge in their 94 playoff run, and may not have been one of the best goalies in the league over the years, but he was one of the best in Rangers history, in terms of victories.  On top of all this, he was born in the U.S., so that makes it even better that he had a good career and wasn\'t born in Slovakia or some other European country.  Now that I\'ve babbled for quite some time, I will wrap this up.  Here\'s to you, Mr. Richter.   '),(18236,1826527,'03,September,2003','2003-09-03 00:00:00','    It\'s officially September.  You know why?  When I drive down Broad Street, there\'s dozens of hoodlums standing outside the deli by the high school.  I never quite understood the phenomena.  Let\'s stand clustered really close together and smoke cigarettes in the rain ouside of a deli.  Yeah, that\'s hardcore.  A deli that\'s right next door to a florist.  I don\'t know.  Tonight kicks off the  urlLink NFL  season.  I don\'t care about the Redskins/Jets, however.  Here\'s my predictions for the NFL season:  Bengals finish 6-10 with wins over denver, baltimore, arizona, houston, pitt, and st. louis.  i\'m hoping for better, but don\'t expect it.  i have every right not to.  it\'s been 12 years without playoffs.  NFC North: Packers, Vikings, Lions, Bears NFC East: Eagles, Giants, Cowboys, Redskins NFC South: Bucs, Falcons, Panthers, Saints NFC West: Rams, Seahawks(my surprise team), 49ers, Cards Wild Card=Falcons, Giants NFC Champ Game:Eagles over Falcons  AFC North: Browns, Steelers, Ravens, Bengals AFC East: Patriots, Bills, Dolphins, Jets AFC South: Titans, Colts, Jags, Texans AFC West: Chiefs, Raiders, Broncos, Chargers Wild Card: Bills, Raiders AFC Champ Game: Titans over Patriots Super Bowl: Titans over Eagles *disclaimer:  Do not use these as your own picks, or you will burn in hell and be ridiculed enormously.  The author has no clue about picking football standings.  This is all in good fun.  OK, now that\'s over.  Here\'s the next part of my story:  I woke up at about ten on Saturday morning.  I probably would have slept through Saturday completely, but the doorbell woke me up.  It was the police.   “Hello, Miss, I’m Officer Taylor, Andy’s uncle.  I was informed that you were a friend of Josie and Andy.”  “Yes, wh--what’s happened?”  “Well, Miss,” he paused and I noticed his adam’s apple bob in his throat, as if he was swallowing back something.  “They were found dead this morning.”  His eyes watered.  “I’m very sorry.  If there’s anything I can do…Here’s my card.  Please take care, and we promise we will end this case soon.  I will make sure of that. I’m so sorry.”  I cried and closed the door lightly.  I slid to the floor with my back to the door.  I picked up the phone to call Corey.  I couldn’t compose myself enough to be comprehendible, however.  I finally was able to spit out in blurts, “Corey…need…you…here.”  He was at the door when the I hung up the phone.  He came over and tried to comfort me.  I weeped in his arms for what seemed to be hours.  He seemed very uneasy about the situation as well.  He shook at times, and choked back tears often, as he and Andy were becoming closer as friends.  “I’ve got a bad feeling about this,” he said.  “Me too, hon.”  “I mean, I’m getting a bad vibe right now.”  “What do you mean?”  “I think I should go see if I can help Andy’s uncle.  I want this creep put away now.”  “No, you can’t, please.  I need you here.  Don’t leave me.”  “All right, but I just wish this never happened.  I wish I could go back in time or something to stop it.  I don’t know what to do.  I feel helpless.”  He was right.  We quivered nervously for the rest of the afternoon.  Every noise we heard caused a rush to see if the door was still locked, and with each click, chirp, or thud, I began to think more.   '),(18237,1826527,'03,September,2003','2003-09-03 00:00:00','    Hmm.  I went job-hunting today.  The fruits of my labor were not rewarded, however.  The closest I got was an application for Modell\'s, and Kim hooked me up with the person in charge of personnel at her bank.  The lady seemed nice, and told me to send in my resume, but she\'ll be away next week and probably won\'t get back to me until the following week.  I went to Target, because they are opening a new store in nearby Clifton soon.  I asked the lady at Guest Services, and she had no idea what I was talking about.  She said you can apply on the computers.  So I attempt to do the computer thing.  First off, the computers were slow, and after I was halfway through, this error message came up, and I couldn\'t get out of it, so I just started pounding the keyboard.  It finally took me back, and I was able to continue.  When I was on the last part, it said these questions were optional, so I decided just to hit enter and move on.  Suddenly, the screen pops up saying \"Thank You\" with that stupid Target dog.  So I left.  I was there for an hour with that stupid computer.  I say I was done.  Here\'s an update on things I said I would bring up again:  1)my rearview mirror is indeed staying put on my windshield.  I guess that kit worked for now.  2)my foot is still painful when I walk on it.  Don\'t know what to do about it.  Here\'s the next piece of my story.  It\'s very short:    Andy and Josie were snuggling on her couch watching television.  Even after the tragedies, they still managed to find comfort in each other.  Friday night was a horrible night to be home.  There was never anything on, except sometimes there were some good stand-up comics on cable.  The couple managed to forget about the world and laughed at the comedy routines they witnessed.  There was a pleasant variety of performers that gave the audience a chuckle.  A short, plump male managed to coax the most laughter from the two.  As the emcee was introducing the next comic, and a loud noise rumbled in Josie’s kitchen.  The laughter had stopped.   '),(18238,1826527,'02,September,2003','2003-09-02 00:00:00','    If anyone knows of a place I can work full time, let me know please.  God, I hate this.  I spent several more hours on the Internet researching jobs I would be qualified for with a BA in Interpersonal Communication, and there\'s some things that don\'t sound too bad, but I can\'t suddenly become a copy writer or advertising specialist without experience.  This is too depressing, and the 60 degrees and rain aren\'t helping either.  I was actually contemplating going to one of those computer training schools for computer graphics and design stuff, but I would need money to pay for it, and a job to supply the money to pay for it, and if I had a job, I probably wouldn\'t want to do that anyway.    In other news, Roy Halladay owned the Yankees yesterday, and it was enjoyable watching the Jays win a game, since I hadn\'t seen them win on tv since the end of May probably.  In a way, I kind of feel bad for Jeff Weaver.  I don\'t like the Yankees, but if he was on another team that was mediocre, he wouldn\'t be scrutinized nearly as much.  Yeah, he isn\'t very good, but when Josh Phelps blasted that 3-run homer in the 8th yesterday, Weaver\'s head just fell, and I actually felt sorry for him.  The game was over anyway, because fat Wells allowed 5 runs before Weaver entered.  Oh well.  Today doesn\'t feel like Tuesday to me.  And it doesn\'t feel like September.  It feels more like the middle of October.  Dreary and chilly.  Well, here\'s another part of my story:    Corey brought me to his house.  I thought he was hiding something from me.  He didn’t say much since Andy showed him the newspaper in the morning.  I told him to take me to my home, since his absence of mind wasn’t contributing to any comfort for me.  I would have been better off alone.  Maybe I could clear my head and forget about things for a while.  Corey walked my to my door and made sure I was in before retreating back to his place.  He ate some pizza and watched the game all night.  He knew not to call when I ask him to be alone for a while.  I always called him when I was ready.  It was six o’clock, so I watched the news when I got home.  They reported that Sammy had apparently left town and the police couldn’t trace his whereabouts.  He was gone.  They fucked up.  Commercial.  I was too worked up to eat dinner that night, so I continued to watch the science channel until about ten.  I decided my eyes were tired at that point, so I went to sleep.   '),(18239,1826527,'01,September,2003','2003-09-01 00:00:00','    Happy Bday, Ord!  My dad\'s the best.  I had a new revelation for a job...I want to work in a hotel, and work my way up, so one day, I can travel to fun places and stay for free in part of the hotel chain I work for.  That would be sweeeet.  But in reality, I don\'t see it happening.  I\'m going to try my damnedist to make it happen though, since this teaching thing isn\'t calling me.  I need a job with money and benefits right now, and subbing\'s not gonna do it for me.  Well, I have the next part of my story:   On the ride home, I was dozing, but I do remember one thing-the radio.  Corey always played around with the radio in the car, and rarely settled on a station for more than a song.  This night was strange because he didn’t fuss around.  He hit the preset channel 6 and news blared from his speakers all around my skull.  Through all the commotion and vibrations of bass in his woofers, I do remember being awake when a reporter stated that Sammy was released because they didn’t have any evidence to contain him.  Corey’s car stopped.  I was home.  I hugged Corey and sleep-walked to my bed.  I made sure to lock the door and secure the deadbolt.  I slipped off my shirt and jeans and dove into a big t-shirt.  I fell to my bed like and was nestled among my pillow friends.   The next afternoon, we all met at the usual spot in the student center.  Andy came running in with that morning’s paper.  “Someone else was murdered last night.  It was Ashley Myers.  She was found with a note.  The same one that was with Erika.”  Josie cried out nervously, almost like a yelp.  I put my hand over my mouth, trying to hold back my almost-vomit.  This time, Andy wasn’t joking about the murder.  He was shaken.  He was almost in tears.  Corey stood there, blankly staring at the paper.  “That Sammy’s one sick fucker,” screamed Andy.  “It said they just released him, and he kills someone else.”  Josie’s mouth opened ever so slowly.  It took a few tries of her quaking lips, but she finally spoke.  “Ashley Myers was one of Sammy’s ex-girlfriends.  I’m the last one he had.  I’m next.”  She took a jump step to run away, but Andy grabbed her.  “No you’re not,” he said.  “I’m here and you’ll be safe.  I promise.” 	 We adjourned our daily meeting earlier than normal on Friday.  I don’t think anyone went to class that day.  We all left.  Andy and Josie went to her place.  	 “Andy, what’s happening?” Josie asked. 	 “Some guy named Sammy murdered two girls.  That’s it.” 	 “And I’m next.  What am I supposed to do?  I can’t hide forever.  He knows where I live.  He knows too much about me.” 	 “Hon, I’m going to stay with you tonight, and until they catch him.  I promise you will be OK.  My uncle tells me the second case will only prove it was Sammy, so when they get a hold of him this time, he won’t be able to get away through some loophole.  If there’s one person I can trust on the police force, it’s definitely my uncle.  You know Uncle Trever is not going to bullshit me.” 	   “I know he wouldn’t b.s. you, but they still haven’t been able to find Sammy yet.” 	 “He’s as good as caught Jo.”   '),(18240,1826527,'01,September,2003','2003-09-01 00:00:00','    I would love to share the story of our ice fiasco...  We were sent to get ice by Brian\'s sister-in-law Rachel after we finished helping her with the food.  So we treked down to the 7-11 and bought 6 bags of ice.  We had 2 coolers so we figured it would be enough.  WRONG!!  As we were filling up the coolers Bri said, \"Hey I think we need more ice.\"  I said, \"OK.  Go get some more.\"  So I gave him the keys to my beautiful Altima and he came back with 4 more bags(i think it was 4).  So we finished loading the sodas and ice into the coolers when Brian says, \"Oh shizzy.  What about the beer?\"  I said \"what beer?  there is no beer here.\"  Brian replied that people were bringing beers.  At this moment Rachel pops out of the door and we ask her about the beer.  She said oh yeah. So we were shipped off again to Rite Aid to see if they had any coolers and some more ice.  But do they have it? Nooooo.  Why would they?  So we walked over to the liquor store in the same lot and they have ice and those crappy little coolers that couldn\'t even hold a 6 pack.  So then we walk over to the Super Stop and Shop.  They call it super for no apparent reason because they were sold out of coolers.  If you were so super you would have more coolers in the back room!  So at this point me and Brian were we even considering a garbage can to hold the beverages.  So we walk over to the dollar store in hopes of finding some sort of container but to no avail.  So we go back to the house empty handed.  We tell Rachel and Kevin that we couldn\'t find anything.  So she says, \"Well can\'t you go to the hardware store?\"  Two replies to this: 1. Why can\'t  you  go? 2.  I\'ve wasted enough gas(which is quite expensive nowadays) and time on your stupid party.  Of course I\'m not mean enough to say this to her face so we get suckered into going to the hardware store.  We get to the store and we don\'t see any freakin coolers.  I remembered there were some containers outside so we check them out.  They were 44 gallon rubber things so we called Rachel and she said get it.  But we didn\'t know if we\'re supossed to take it inside or if they get it from the stock room or some crap.  So Bri goes inside to ask and comes back outside to get me.  I guess we were blinded by our rage but the coolers were sitting on top of one of the aisles.  So we get the kid to take it down and we pay for this crate of evil by the drunken fat guy.  We go back to the house and everyone is happy that we found a cooler and brought 6 more bags of ice.  But don\'t you know, their friend Danny brings a huge cooler later on filled with ice.  So we went nuts for nothing.    So lets tally up some things here: Total trips for ice/cooler- 4 Total amount spent on ice/cooler- about $60 Total amount of gas wasted- too much and i never got refunded for this. Sorry but a hot dog doesn\'t cut it. Total amout of drinks guzzled down to get rid of anger/headache- not enough   '),(18241,1826527,'31,October,2003','2003-10-31 00:00:00','    Happy Halloween.  Spending the weekend in Vermont, should be fun...  cya   '),(18242,1826527,'30,October,2003','2003-10-30 00:00:00','    Does anyone have a remedy for a super sore neck? My neck frickin hurts!  Halloween is tomorrow and the little kids will stay from my house because there is a cemetary right across the street. Maybe they\'re afraid the zombies will eat their brains. I know I\'m afraid.  Brian and I are going to celebrate El Dia de la Muerte with my good friend Liz at her house.  PARTY TIME!   '),(18243,1826527,'27,October,2003','2003-10-27 00:00:00','    Thursday.  My day off.  Oops, scratch that.  I was asked yesterday to work today, so I have to go in at 1, and I don\'t know when I can leave.  Apparently, lots of things going down with employees or something.  I wasn\'t too happy yesterday when I was asked, basically because it was 5:25 and I was supposed to leave at 5:30.  Not the greatest way to end my shift.  Anyway, I think the manager thought we\'d have a big rush on our Halloween candy tonite.  Enough about that.  NBA season is here and I\'m not really into it yet.  Keith VanHorn had 29 points for the Knicks, but they lost.  I\'m sick of hearing about Lebron James already.  The Sixers are 1-0 and are on ESPN tomorrow nite at Sacramento.  Other than that, I don\'t know what else is happening yet.  I had a bunch of things I wanted to get done today, but they\'ll have to wait.  One thing was to get my car checked, but I can\'t because I\'ll be using it to get to work.  I hope it holds out until next Thursday.  On a positive note, I do get my first paycheck today, but I\'m sure it will be a great disappointment after I take into account my bills.  Well, if I don\'t blog tomorrow, which I probably won\'t, then have a fun-filled weekend!   '),(18244,1826527,'27,October,2003','2003-10-27 00:00:00','    Hey Mercedes rocked NYC!  The first band seemed okay at first but the singers voice sucked.  The second band was a Thursday rip off - a screamer and a singer.  This was the crap band Bri spoke of before.  The worst part was that they were trying to get the crowd involved. The singer said, \"Even if you don\'t like us, we\'re all here because we love music so clap your hands.\"  I\'m sorry but the shit they were playing was NOT music.  Moving on, Damone was ok but the singers voice wasn\'t 100%.  However, she shreds a guitar solo.  I thought Damone was the last band to go on before HM but I was told otherwise.  Then Whitelight Motorcade came on.  They tried very hard to look rock and roll but they were really a bunch of no talent ass clowns.  Everything they played has been done before.  They were completely unoriginal.  What a waste of time.  Then HM came on and all was well.  I once again sang every single word to every single song.  I\'m truly pathetic.  They played a lot of new stuff (Quality Revenge at Last, It\'s Been A Blast, Knowing When to Stop, Lashing Out, Go On Drone) and some of the more popular old songs (Eleven to Your Seven, Stay Six, Our Weekend Starts on Wednesday, Go Blue).  They played Go Blue as an encore but damn Irving Plaza has a curfew so we couldnt hear anymore of our favorites.  I think it\'s very ridiculous that a club has a curfew.  Oh well.  I think the show was worth the 12 bucks just for Hey Mercedes alone.  Everyone in the band was totally into it and you can tell that they\'re having fun up there.  Although this trip to the city wasn\'t so bad, I let it be known to a member of hm that they have to play at Maxwells every time they come to NJ.  Anyone who has been to Maxwells knows why.    Enough about hm!  I\'m doing just fine in case anyone was wondering.  Let me sum everything up for ya: School\'s a bitch.  Work sucks.  I\'m no longer sick.  I need a vacation but definitely not to Myrtle Beach.  Hope everyone is doing well.  Bye!!! :)   '),(18245,1826527,'26,October,2003','2003-10-26 00:00:00','    Damn, it\'s been a while since my last blog.  Times have been busy for the new worker, so I\'ve been a slackin\'.  The new job is pretty good so far.  Still don\'t know what my daily duties will be, but the people are nice and it\'s a job.  This weekend was full of many exciting events.  I\'ll go in reverse order and work backwards from today.  The  urlLink Bengals  are 3-4.  They are in second place in their division, and already have one more win than they had all of last season, and without their \"star\" running back, Corey Dillon.  More on him in a moment.  Anyway, they defeated the previously 5-1 Seahawks 27-24.  Their defense allowed only 7 points in the second half, and the TD came withing the first 6 minutes of the 3rd quarter.  Two in a row, and next week they travel to Arizona.  As for Corey Dillon, he got into a car accident before the game, and hopefully he\'s ok, which I think he is, but prior to that, he was being a total ass and complaining about not getting the ball enough.  First of all, you have to play to get the ball.  He\'s been injured every game!  And second of all, the team has now won 3 games without you.  They\'re playing really well and keeping games close, so shut your mouth and play like you\'re paid to.  If not, then they have a perfectly capable backup in Rudi Johnson, who broke 100 yards today.  Last night was the  urlLink Hey Mercedes  show at Irving Plaza in NY.  Just seeing the band was nice.  I\'ll have more on that in a moment as well, because there were many ups and downs involved in the night.  Upon arriving home, I was informed that the Marlins defeated the Yankees in the World Series.  Wow.  I\'m not going to rub it in, because everyone who blogs is a Yankee fan, but that put a smile on my face before bed.  As for the HM show, here\'s the deal thru a list of pros and cons:               PRO:  HM played a show      CON:  It was in NY  PRO:  HM was headlining      CON:  There were 4 opening bands  PRO:  Going with 2 cool people(Kim & Liz)      CON:  It was in NY  PRO:  Not driving into city      CON:  Finding parking in Hoboken  PRO:  Found parking in lot next to PATH      CON:  Parking cost $20  PRO:  PATH station close to Irving Plaza      CON:  Forgot how to get to Irving Plaza from PATH  PRO:  Found place pretty easily      CON:  It was still earlier than anticipated  PRO:  Doors open at 7 pm      CON:  It was 7:10 and I really had to pee  PRO:  Doors opened at 7:15      CON:  First band started at 7:50  PRO:  The bar had the Captain      CON:  Captain & Coke cost $7  PRO:  Talked to Bix, the HM tech      CON:  He had no new t-shirts for sale  PRO:  1st band usually plays short set      CON:  1st band sucked  PRO:  2nd band usually plays short set      CON:  2nd band was Canadian and sucked  PRO:  Saw people we knew(Kate, Dave)      CON:  Saw plenty of scenesters  PRO:  3rd band was Damone      CON:  There was another band after them  PRO:  Talked to people we knew      CON:  Had to leave due to mild case of food poisoning  PRO:  Returned feeling better      CON:  There was still another band to play  PRO:  4th band looked rock & roll      CON:  4th band sucked  PRO:  HM went on at 10:45      CON:  Irving Plaza has a curfew  PRO:  Found PATH station easily      CON:  It was 110 degrees in the station  PRO: Hit very little traffic on way home       CON:  My car sounded like it was going to fall apart  PRO: Turn back clocks, gain hour of sleep      CON:  It was still 1:30   So that\'s a short list of the show\'s experience.  In other news, I\'m working 2-11 tomorrow night.  I\'m wondering what I\'ll be doing since the managers usually leave at 5, and I have no idea how to do anything yet.  Hmm.  I\'m going to see Dave\'s band on Wednesday at Just Jake\'s in Montclair.  He plays the keyboard.  Should be interesting.   '),(18246,1826527,'25,October,2003','2003-10-25 00:00:00','    Ok, I had to post this link:   urlLink http://grouphug.us/   This site is hilarious.   '),(18247,1826527,'20,October,2003','2003-10-20 00:00:00','    Day one of my new job.  I worked from 9-5:30 and will be tomorrow, Weds, and Friday.  I still don\'t know what my job description is, or what I will be doing.  My supposed job title is SIMS coordinator.  It\'s like I\'m going to be in charge of inventory-keeping supplies on hand, knowing when we have too much, etc.  I\'m supposed to find out more on Weds.  Tomorrow is truck day, so I\'ll be busy with that all day.  Today consisted of me moving boxes around in the stockroom in preparation for truck tomorrow and attempting to complete a display for Christmas, although it did not turn out well, since we didn\' t have enough shelves to complete the task.  The people I\'ve me so far have seemed really nice, which is a plus, but I\'m not used to working, so my stomach and rest of body are out of whack.  I was hungry at 12, but didn\'t want to eat since I still had 5 and a half hours left.  I\'m also tired, because I had to do a lot of stupid errands when I got home, so I only got to sit down at 7:30.  Anyway, I\'m happy I\'m working.  The place seems nice, and I\'m hoping once I find out what my job will be, I can learn that before worrying about every other possible job in the store.  Rob, your story was touching, but for some reason, I knew something was going to happen to ruin it.  I\'m sorry about that.  Keep your head up.  I ordered the book the other day, because I don\'t have anything else to read right now, so I\'m supposed to receive it this week sometime.  Kev,  The Matrix Reloaded  was indeed a piece of crap.  I liked the first movie, and thought the second one made no sense and had really only one scene I enjoyed, which was the last big action scene.  I have no urge to see the new one either, but if I do, it will be with the gift certificate I have, so I technically won\'t be paying for it.  Tomorrow\'s day 2 of work.  I gotta get to sleep early tonite, because I\'m feeling a little weary and my nose is running.  Hope to sleep it off and fight off any cold or such symptoms.   '),(18248,1826527,'19,October,2003','2003-10-19 00:00:00','    It\'s Monday, and I dont want to be at work.  I am dead tired.  This past weekend was not overly relaxing.  I stayed in Friday nite and rented  The Matrix Reloaded .  What a piece of crap.  The effects were realistic, but the story was so fricken complicated that it ruined the entire movie.  I have very little desire to see the third movie when it comes out now.    Saturday nite was fun though.   urlLink Meredith  and I went to Toad\'s Place to see Stephen Lynch.  He was amazing.  I really needed to laugh like that.  He didnt have an opening act, so he went on at 8:30 and was done by 10:15, which was nice for an old guy like me.  He played most of his \"classics\" and a few new ones.  Damn I miss O&A....  Meredith liked everything he played, except for part of his encore, where he performed  Kill a Kitten .  She loves cats, and found the song offensive, but she laughed anyway.  He didnt perform  FBA , but he did play the two songs that he considers his worst songs:  The Abortion Song  and  Special Olympics .  Also, he made it a point to let everyone know that he got married recently, so I guess he wasnt going to any of the local bars after the show and try to pick up annoying Asian chicks (sorry, that was a reference to something only Rob will get).  We got up early Sunday morning to drive all the way back Bloomfield because the 9am mass at Sacred Heart was being dedicated to my grandfather who passed away 20 years ago this past weekend.  All of his children and grandchildren were there.  It was nice to see my family.  We then had lunch at Charlie Brown\'s in Upper Montclair.  I think that half of us did not enjoy our meals.  I think that place sucks.  It used to be nice, but not anymore.    Oh well, back to work.   '),(18249,1826527,'19,October,2003','2003-10-19 00:00:00','    I\'m still not feeling so well.  I was actually feeling much better for the past few days thanks to some over the counter cold medication.  It cleared up my sinuses but it made my heart race.  Despite that, I continued to take it so I could make it through my hectic school/work schedule.  I decided on Friday night that I wasn\'t going to take it anymore, but I woke up on Saturday feeling worse than ever.  Of course I had to work, so I decided to take the evil cold meds again.  When I got to work my friend/co-worker Michele treated to me to a nice cup of coffee.  Bad move.  I was so shaky through the day.  I was seriously worried about driving home.  The ride from Wayne to Clifton, although not very long, is very unfun.  So, I swore to myself to NEVER take that shizzle again, but I again woke up today with a stuffy nose.  Speaking of stuffy noses, why is it that only one nostril gets clogged?  Example, at the moment my right nostril is completely stuffed but my left one isn\'t.  What\'s the point? I can\'t breathe too well with just one so clog up both so I can be completely miserable! Anyways, I hope I feel better before Saturday when Bri, my friend Liz, and me go to see our boys Hey Mercedes.   Speaking of Hey Mercedes, I\'ve been talking online with Ryan whose a good friend of HM.  He\'s gone on several interviews with this video game company in Chicago. I hope he gets the job. Free video games for all!! SWEET!!!!   '),(18250,1826527,'17,October,2003','2003-10-17 00:00:00','    So the ball....   Let me tell you about last night.   It was like a huge funeral here last night.  On my train home last night, the conductor was sobbing.  I saw a guy lying face down on a park bench, but he just might have been homeless.  Boston just became really quiet.  I truly feel quite bad for the fans up here even though there are a ton of assholes.  Seeing people just resign themselves to believing this curse crap is real is depressing.  Anyway, that being said, it was a great game to finish one of the best series I\'ve ever seen.  Unfortunately, this is the World Series that nobody wanted to see, but I think it\'ll be somewhat close.    Sunday Sunday Sunday!  Pats- Dolphins.  Pats break the jinx 28-20.  I think I\'ll predict a Pats win by the score of 28-20 every week.  Its fun.  I also read the Corrections, right about a year ago and I agree with Dennis.  It\'s good.  I have a copy if you wanna wait until Thanksgivingish or I could just mail it to you.  It was Eric\'s book and I was supposed to mail it to him in like January, but I forgot so he bought another copy.    Other stuff:  My mom and aunt have been up here since Tuesday visiting.  They went to Cape Ann (another quaint area with an Ocean view and little shops like they enjoy) on Tuesday and I went with them for one night.  It was good just to spend time with them and whatever, but there was a funny incident.  We went to dinner at seafood place called McT\'s.  The food was good, but that\'s not where I\'m going.  The waitress was kinda cute and my mother and aunt were making offhand comments that she was \'making eyes\' at me.  I disagreed, but I did think she was cute.  But it was strange because after we left I started thinking about the waitress.   I was wondering who ends up as a waitress in Gloucester in October when there are hardly any tourists left.  I also was playing with the idea in my head of what would happen if I had actually tried to pick up a waitress while she was working.  Not a pretty thought, but anyway.  The next day I took the train from Gloucester to Boston in the afternoon since I had to work at night.  I caught the 3:48 pm train.  So, of course you see where this is going.  She stepped right past me as I sat down on the train.  I wouldnt think it was weird if there were like tons of people getting on the train, but there were only like 6 or 7 people total.  Anyway, she said hi and walked to a different part of the train.  I picked up my stuff from the seat and walked to where she was and sat across the aisle.  We ended up talking for a good 40 minutes of the hour ride.  She\'s actually really cool,  She\'s 24, and works as a waitress and a landscaper.  She was going into the city sice she takes a German course at a language institute.  She has moved 3 times in the last 6 months but is from Gloucester, and this is her last visit home before moving west, likely permanently.  She played rugby.  She also worked in the past as a photo developer and baker and is going to finish her double major in nutrition and German.  Just a set of really of really odd details worth noting only because we had a similar feeling in common.  Oldness.   We both decided that 24 is when we started feeling old and pressure to do something productive with life.  We parted and that was that since she\'s moving away in 2 months anyhow and lives in Gloucester which is 40 miles away.  Coincidence makes me laugh sometimes.    Blog!!!!   urlLink http://www.homestarrunner.com/sbemail87.html   '),(18251,1826527,'16,October,2003','2003-10-16 00:00:00','    Holy Crap.  As much as you all know I hate the Yankees, that game was, in the words of  Dennis , Incredible.  I\'m not going to analyze it or complain about anything the Sox did.  You\'ll hear plenty of that on Sportscenter, etc.  I\'m going to step aside and actually congratulate the Yankees for proving who the better team was.  I\'m also applauding MLB for an absolutely amazing post-season so far.  Hopefully the World Series will go 7 games as well and continue the trend.  The one thing I\'ll say is that I liked Aaron Boone when he was a Red(ask  Kim  for the validity of this), and was upset when he came to NY because I knew he was a damn good player.  Now all the dumb Yankees fans around my area suddenly love the man.  Patience, y\'all.  Oh by the way, Den\'s prediction was eerily close.  It\'s Friday!  They\'re actually calling for a nice weekend after today for once.  Sounds good to me.  Yesterday something strange happened.  Here\'s a little background:  My parents are planning on getting our front stairs replaced, as well as the small path to our side door.  They are going down the shore for a few days for their anniversary(Sun-Tue).  They called the mason to tell him to start the job when they got back from the shore, so he could do it all at once, instead of starting this week and waiting a few days, etc.  Kim had been over yesterday afternoon for lunch before her night class.  The doorbell rang about 3 pm, but I was in the restroom, so I didn\'t hear it.  She looked, but didn\'t know who it was.  When I returned, I went downstairs to see who it was.  I hear loud banging and open my front door to find my front steps being demolished and the railing dismantled.  The mason decided to start breaking down the stairs without telling us.  So I called my mother to see if I missed something, but she didn\'t know either.  She told them to go ahead and continue, so they could finish the stairs today.  So today I heard jackhammering at 8 am, and they\'re working right now.  Should be done with the stairs today, and will probably come back late next week to finish the side path.  I guess it wasn\'t a big deal, but I was thinking about what would happen if we were getting our porch rebuilt.  Would they just bring in a wrecking ball without telling us?    In other news, I\'ve had a person-less house next door to me for about 8 months now.  The previous neighbors just bolted one day.  Finally, someone bought it late summer, but he\'s rebuilding the entire inside-walls, floors, electric-just about everything except the shell of the house.  On the other side, we have a strange house.  One guy\'s an Elvis impersonator, and the other guy is a fat, drunk, something.  I don\'t know what he does for a living.  But the fat guy\'s an asshole who parks his big red van on the street and insists on hitting every car while parallel parking.  Needless to say, we hate him, and found out the house was recently sold, and I\'m told they\'ll be gone by the end of the year.  Woo-hoo!  Remember that show Home Improvement?  That Wilson guy was the neighbor.  Where the hell is our Wilson?  Have a good weekend folks.  Listen to  urlLink Knapsack .   '),(18252,1826527,'16,October,2003','2003-10-16 00:00:00','    Incredible!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  Yankees 6 Red Sox 5   What a game!!!!!!!!!!!!!!!!!!!   '),(18253,1826527,'16,October,2003','2003-10-16 00:00:00','    Well, it\'s officially worry time.  Game 7 Yanks vs Red Sox.  I\'ll be so nervous that I won\'t be able to enjoy the game.  Tomorrow is my Neurobiology midterm, so I have another thing to worry about.  In class, we watched a video where they gave monkeys Parkinson\'s by lesioning parts of the brain.  Obviously I\'m pro research and I think PETA is a bunch of whiny rich kids with nothing better to do, I still felt pretty ill when watching the monkeys tremble.  Bri-for books, you should read   The Corrections   by Jonathon Franzen.  It\'s one of the best books I\'ve read in a long time.  After my exam, I was going to give my book rant, much like my Rolling Stone review:), so wait for that.  For movies, get Indiana Jones on Tues.  While at the bookstore, you should check out the new    Esquire   magazine.  There is some good material in it.  \"Mmmm. 20 questions with Lorne Michaels\"  Well I gotta get ready for the game  My prediction tonite Yankees 5 Red Sox 4.   '),(18254,1826527,'15,October,2003','2003-10-15 00:00:00','    Congratulations to the Chicago Cubs for sucking.  As long as their world series drought continues, I can no longer accept Bengals ridicule.  HAHA!  No, I was kinda rooting for them, but I\'m totally fine with the Marlins, it just doesn\'t sound right.  I also didn\'t want Kerry Wood to take the loss, but the whole team just collapsed after a 3-1 series lead.  I\'ll wait for the Sox-Yanks tonite...  I\'m officially starting my job Monday.  I\'m working Mon, Tue, Wed, and Fri from 9-5 for a total of 32 hours.  It\'s my first week, so Monday I\'ll probably set up a fixed schedule for the weeks to follow and increase the hours up to 40.  I\'m technically going to be in a management position, so that\'s gonna be nice.  I have to wear a shirt & tie, but as Kim said, I would have had to if I was teaching too.  I\'m happy because I\'ll be getting some money and benefits.  Whew!  I was starting to get nervous about the lack of insurance.  Well, it\'s Thursday, so I\'m gonna meet  Kim  for lunch/dinner. She\'s feeling a bit better, so I\'m glad.  I feel like celebrating my job, but I don\'t have money from the job to celebrate the job.   Huh?  I decorated my house for Halloween yesterday.  I still enjoy this holiday for some reason, even though I don\'t dress up or anything.  I guess I\'m still 10 in my mind.    Anyone recommend any good books, movies, or cds?  I plan on buying one or more of each with my first paycheck.  Especially movies and books.   urlLink Let me know .  Thanks.   '),(18255,1826527,'15,October,2003','2003-10-15 00:00:00','    Wow, it\'s a good thing I didn\'t jinx the Cubs last nite.  What the hell was that all about?!  Today\'s my parents\' anniversary, so Happy Anniversary!  Not like they\'re going to read this, but maybe they\'ll get the message telepathically.  I watched Conan last nite, and Jim Carrey was his first guest.  It was a rerun, but Carrey did  In The Year 2000  with him and was frickin\' crazy.  Later, during his interview, he did this bit with Stephen Hawking.  Very funny stuff, and apparently, I heard that Carrey actually went to visit Hawking for dinner a few weeks after the original show was recorded.  I believe it was in June of this year.  Next Saturday is HM at Irving Plaza.  All right!  I hate the city.   '),(18256,1826527,'14,October,2003','2003-10-14 00:00:00','    My baby is sick.  This sucks.  I did my best to take care of her today.  I\'m sure I\'ll be getting sick soon.  Probably the kid with the outbreak cough from Friday night passed it along...  Congratulations to  Rob  for guessing the lyric.  It was indeed Skee-Lo from his song  I Wish .  You win!  I got my oil changed in my car today, and the guy tells me that my radiator is corroded a bit inside, and my coolant is dirty.  That sucks, because I got a new radiator put in by my awful mechanic last November, and it had a year warranty.  How does it wear down so quickly?  Anyway, I now have to get it flushed out and get my tires rotated for the winter, so my car doesn\'t ice-skate, and my engine doesn\'t get frostbite.  The Sox are pretty much cooked.  I don\'t see the Yanks losing to Burkett, especially with Pettite pitching...Cubbies look like they\'re wrapping it up.  Six more outs for the Fish.  So I will be joining the workforce again on Thursday-filling out W-4s and I-9s and all that jazz.  I am ecstatic to be employed again, well at least more than just every Saturday.  Good times.  Oh yeah, yesterday was Columbus Day(the day it\'s observed).  I remember always having off from school, at least until I reached college.  Yesterday I noticed that my old elementary school was in session.  What is this world coming to?  First they condensed Washington & Lincoln\'s birthdays into one day, then they stopped observing Veteran\'s Day in school, and now Columbus day isn\'t a day off anymore.  What\'s next?  No more Christmas break!   '),(18257,1826527,'12,October,2003','2003-10-12 00:00:00','    Hello all.  The crazy New Jersey weather of 40 degrees one day and 70 the next has finally gotten to me..  I\'m sick.  I\'m not sure but maybe it\'s a punishment from the party the other night.  Probably not.  I only had three drinks...  Anyhoo, I\'m on day 2 of my cold which basically means I must suffer for the next week and a half.  I\'m no good at being sick.  I turn into a whiney little baby.  And guess what?  Mommy isn\'t home to take care of me.  What fun is being sick if there\'s no one to wait on you?   '),(18258,1826527,'12,October,2003','2003-10-12 00:00:00','    Skee-Lo  \"I wish\"      -     There\'s pretty much no way I could ever forget that.   I\'m kinda psyched for the Strokes as well.     -Rob   '),(18259,1826527,'10,October,2003','2003-10-10 00:00:00','    Jeez, Den, you sound like a professional writer for Rolling Stone.  I\'ll take it!  There\'s this house around the corner from mine which stations a small dog that I\'d like to kick.  It tends to poop on the sidewalk, and the owners don\'t ever clean it up.  Today when I was walking by, I noticed, ironically, that their house had been tp\'d.  Well done!  Man am I bored.  I\'m literally counting down the hours until this party.  I\'m about to start drinking now.  Whoops, too late.   '),(18260,1826527,'09,October,2003','2003-10-09 00:00:00','    Hello all,  During work, I listen Virgin radio UK on the Internet while I\'m at my computer.  Normally it is a good station, lots of great Brit rock, except for a band called The Darkness.  You gotta hear them.  They are like an 80\'s hair band and their video is pretty fucking hysterical, which you can see for free at Rollingstone.com.  There\'s a guitar solo in it, where the guitarist is surrounded by a wall of amps.  Pure 80\'s metal.  Surprising to me they are #1 in the UK for the past month.  Anyway, on to good music.  The Strokes new album comes out in 2 weeks.  So far, it has gotten pretty good reviews and I\'m looking forward to it.  This year has seen some killer releases, new Radiohead, White Stripes, Stephen Malkmus (of Pavement fame), Doug Martsch (of Built to Spill), and the Strokes.  The Strokes and White Stripes may be a little mainstream for some, but they both play great music and they get kids into such bands like the Velvet Underground and Televison, which is always a plus.  Radiohead is simply one of the best bands out today.  Amazing live performances, intelligent lyrics, and arguably the best musicians in any band.  There\'s my little rant on music for today.    Yankees 6 Red Sox 2, Series tied 1-1, Clemens vs. Martinez at Fenway on Saturday.  As a baseball fan, it doesn\'t get any better than this!!   Have a good weekend all!   '),(18261,1826527,'09,October,2003','2003-10-09 00:00:00','    Hey I remember  urlLink  Disorderlies  .  The Fat Boys were funny back in da day...   '),(18262,1826527,'09,October,2003','2003-10-09 00:00:00','    \"Every day would be a Friday, and we could even speed down the highway.\"   If you know what song or artist this quote is from, then you win.  So I\'m gonna speed down the highway today because it\'s Friday, and tonight I\'m gonna exit that highway to go to Mr. Sides\' house.  Once there, we will be partaking in the imbibing of alcohol.  Woo!  And on top of that, it\'s pizza nite beforehand!  Tomorrow I work for GHT, so what usually happens when I work for them?  Well, apparently it\'s not supposed to rain tomorrow.  Ho-ly Crap.  That\'s incredible.  Unfortunately, the game doesn\'t begin until 2pm because of the kids taking SATs tomorrow.  I won\'t be home until 7 probably.  So no Saturday, but that\'s where tonite is important!  Sunday, Sunday, Sunday!  I will be seeing about this job opportunity.  Hopefully all goes well and I\'d be able to start soon after.  I have to fill out when of those apps on the computer, and then I gotta pee in a cup so they know I\'m not addicted to meth.  Here\'s a story I\'d like to share which highlights my job search experience.  Last Friday I received a phone call from  urlLink Belleville .  The secretary says they\'re going to be interviewing for two 5th grade positions.  I say I am interested and she says the interviews would be later next week, but she would call early in the week to schedule.  So I wait, and she doesn\'t call.  So yesterday(Thursday!) I call them.  She says they didn\'t interview yet, and she\'s not sure when they will be.  So I say, well will it be within the next few months?  She says she doesn\'t know, but probably.  So I ask will I still be on the interview list if they don\'t interview for another 3 months, and she says, \"probably.\"  OK, then why did you call me last week?  This is just one of the many examples of ineptness in New Jersey\'s School Districts.  Absolutely disorganized.  Remember the movie  Disorderlies ?   No one does.  Never mind.  So this  urlLink AIM  thing is addictive.  I don\'t even have many buddy names, and I\'m thinking I\'m on way too much.  But it\'s quite fun, especially when you sign on at midnight last nite to check your email, and end up chatting with  Liz  until 1.  Don\'t worry.  Liz is Kim\'s friend, not some girl I met in a chatroom.  Anyway, I\'m currently talking to  Kev  while he\'s working.  I feel kinda bad, but he says it\'s all good and he won\'t get in trouble.  Well folks, enjoy your weekend.  Listen to  urlLink The Appleseed Cast .   '),(18263,1826527,'09,October,2003','2003-10-09 00:00:00','    It\'s official.  I am addicted to Tiger Woods PGA Tour 2004.  This game is amazing.  If it is out for the system you own, buy it.  Don\'t finish reading my post, go to the store and get it.  Why are you still reading this?!?!  Go, now!!  And get a network adapter if you have the PS2, because I plan on getting one and I can then kick your arse from across a few states.  I cant wait until 3pm so I can leave work and go play Tiger Woods...   '),(18264,1826527,'09,October,2003','2003-10-09 00:00:00','    Whoa, what\'s going on here.  You two can\'t fight on the blog.  Consider yourselves banned until you both clean up your act.  What the hell is wrong with these people?   '),(18265,1826527,'08,October,2003','2003-10-08 00:00:00','    Who let Morgan on here anyway?   '),(18266,1826527,'08,October,2003','2003-10-08 00:00:00','    Yarr!   '),(18267,1826527,'08,October,2003','2003-10-08 00:00:00','    That lootin\' rummy thinks he won.  Oh.   Me takes his peg leg and shove it up hes ass.  Bring it on ye mangy scoundrel.   '),(18268,1826527,'08,October,2003','2003-10-08 00:00:00','    That bloody fool McCallister will pay.  Stupid sonofabitch screwed up me opportunity to find me a lass.  I busted his nose good.  There I was, drinking me spiced rum when I spotted a fine young lass who sat next to me.  McCallister comes back from the head and stumbles upon me peg leg and knocks her down.  Me lady left in a hurry and a slobberknocker broke out.  McCallister got what he deserved, but needs more a hurtin\'.  He\'ll pay.   '),(18269,1826527,'08,October,2003','2003-10-08 00:00:00','    Dude, you\'ll never guess what happened last night.  I was at the Town Pub watching the Cubs-Marlins game and noticed these two odd characters.  One guy was dressed in this long coat with boots and had a fisherman-type hat on and the guy he was sitting next to had this pirate theme.  He had this long mustache and goatee.  Anyway, The pirate guy was mackin\' on some chick.  The fisherman guy comes back from the bathroom to sit down and trips on the pirates pegleg.  The fisherman falls into the chick and she falls down with him on top of her.  She hits him and pushes him away and stormed out onto Broad Street.  The pirate guy is pissed and starts whaling on the fisherman and swearing about how he always prevents him from getting any box.  More punches were exchanged before a big dude from the back came out to separate them.  The bartender called for a cab to bring the fisherman home, and the pirate guy had to drive home.  Chaos.  That was something you don\'t see every day, unless it was Halloween or something, but that\'s a few weeks away still.  Anyway, the Cubs destroyed the fish.  And the Sox were victorious as well.  Their bullpen looked pretty sharp, at least Timlin and Williamson.  Hope both series continue to be close, though.  You know what I had for breakfast this morning?  Cinnamon Toast Crunch.  I never had the stuff when I was a kid, but I\'m eating it when I\'m 24.  I ate after mowing the lawn, which is especially fun when there\'s pieces of metal hidden in the yard from the guy next door\'s construction.  It was like mowing over a mine field.  Try not to get electrocuted or get a contusion on your shin.  So on Sunday, I may have a job.  It\'s not a teaching job, but where has that profession gotten me, honestly?   Kim \'s mom is hooking me up with a job full-time.  She said it would be like an inventory dude and i\'d be on the floor as well, but it\'s a job and I couldn\'t be happier.  I just hope something stupid doesn\'t happen between now and Sunday.  So I\'m happy, and I\'m sorry about all the nagging and depression I\'ve emitted the last few months on the blog.  I hope you understood a little of my feelings, but I think I was a little overboard, so things are looking up!  Oh, remember when I mentioned something positive happening with my brother?  Well, he\'s gonna be working at WFAN, probably the biggest sports radio station in the US, on a patchy basis.  He\'s happy as hell, and he\'s still gonna be at WCBS 880 as well.  Good stuff.  May blog again later.  Peace out.   '),(18270,1826527,'08,October,2003','2003-10-08 00:00:00','    Hello all!  I just got home from class and I\'m saving Loses Control to my computer.  The weird thing is that my Real One Player actually \"knew\" who Hey Mercedes was and had all the tracks listed for me so I don\'t have to type everything in.. Nice!  I think the boys have finaly made the big time.  This album is awesome and EVERYONE should own it.  Bob actually says fuck several times in one song.  I\'ve neverheard fuck sound so pretty.  It made me wonder if fuck is a really a bad word.    I haven\'t seen any butterflies either Rob, but I have seen some nice sunrises sitting in my car on campus at 6:30 am...  Wow. That\'s really early.   '),(18271,1826527,'08,October,2003','2003-10-08 00:00:00','    Me thinks me going to see me ol\' friend Captain Morgan tonite.  Drinking rum, singing pirate songs, lootin\', and lookin\' for some ass.  I hope he can pick me up cuz I don\'t have a carrrrrr.   '),(18272,1826527,'07,October,2003','2003-10-07 00:00:00','    Well, politics is officially a joke in this country.  Arnold Shwarzenegger is going to be governor of California, but strangely, I\'m jealous.  I think it\'s because he couldn\'t do any worse of a job than Gov. McGreevey has done for this state.  Yeah that guy sucks.  Don\'t pass it to that guy.  Did anyone see the game last night?  Holy crap was that exciting.  Cubbies have a must-win tonite...  Yeah, so I just downloaded aim, so if you have it, my name is  urlLink stringmule .  I have no job, folks, give me something to do.  Not like you care, but I\'m going to write a brief synopsis of the new  urlLink Hey Mercedes  album entitled  Loses Control .  These are my interpretations of the songs so far, so they may not even be close to the real meanings!  The artwork and photography are nice as well.  01   Quality Revenge At Last       Lyric:  There\'s a phone call coming from a Romeo in Joliet        Catchy song with some \"Yeah\'s\" in the chorus-probably most radio friendly I guess    02   It\'s Been A Blast       Lyric:  All the arguments are not the same without you       Most uptempo on CD in my opinion.  About a past fling he is looking back on. 03   Playing Your Song       Lyric:  The lies are piling up so high that he can\'t sleep at night       Third person story about a guy who is with a girl but loves another. 04   Knowing When To Stop       Lyric:  There\'s nothing inspirational like watching angels drop       The downfall of frienship through lies and fights. 05   The Boy Destroyers       Lyric:  All I know is I\'m never gonna go the safe way home       Loneliness. 06   Unorchestrated       Lyric:  The point in time when everything dies.  It\'s over.       Sad song.  Depression to realizing things are not so bad. 07   Lashing Out       Lyric:  Smile and sing along to every single song       Singing along to some band at a show. 08   Police Police Me       Lyric:  The stars are all smiling.  It\'s disarming.       Media, cops, and criminals are all the same. 09   Oh Penny       Lyric:  And my dear old sharona keeps em on the floor wanting more       Happiness with someone elsewhere, but Penny is here now. 10   The Switch       Lyric:  I\'m a nervous wreck. Hey I\'m a tightrope.       Sex compared to a video game?  Huh?  High score...if it takes all night. 11   Absolute Zero Drive       Lyric:  So wake up soldiers and take all your arms and leave       My favorite song.  No fear, man. 12   Go On Drone       Lyric:  Stop the press. Who is that mess?       Miserable person continues to mope.   '),(18273,1826527,'07,October,2003','2003-10-07 00:00:00','    Never fear Kevin.  You can always just root against my teams as usual.   '),(18274,1826527,'06,October,2003','2003-10-06 00:00:00','    Well, its 9am and I\'ve been at work for 2 and a half hours now, and I don\'t feel like doing any work.  So, here\'s what I\'ve been up to for the past week:  I took the  urlLink girlfriend  to see Howie Day at Irving Plaza last Thursday.  Jump, Little Children (now going by the moniker \"Jump\") opened and were rather good.  They played a good mix of old and new songs.  Then Howie went on.  He was good, but nothing special.  I really didn\'t see anything that would separate him from the other thousand singer/songwriters out right now.  The gf loved it, so I guess the nite was a success.  At least I got to see my friend Pete at the show.  I took Friday off from work, since we got back from the show at 3:30am.  I spent the day dicking with my car.  And then I completely wasted Saturday.  On Sunday, myself and 3 of my friends took two cars up to Vermont and we spent the day rallying our cars on the unpaved mountain roads of VT and Upstate NY.  It was a blast.  My car is completely filthy.  The only negative about the day is that we left at 9am and didn\'t get back until 9:30pm, so I spent 12 hours driving on a Sunday.  Needless to say, I was absolutely worthless at work on Monday; but it\'s not like I care.  It\'s sad, but I think I\'ve become disenchanted with sports.  I really don\'t care if the Yankees win or lose, I\'ve given up on the New Orleans Saints, I gave up on the New York Rangers years ago, and I never gave a shit about basketball.  I still love watching sports, but I can\'t rout for anyone like I used to.  Oh well  Later   '),(18275,1826527,'06,October,2003','2003-10-06 00:00:00','    Who the hell is Captain McCallister?  Is that the pirate on the Simpsons?  Anyway, I worked tonight during the Red Sox game.  Not suprisingly, there were no movie-goers in Boston.  We were getting updates continuously.  This guy Justin jumped on top of me when they won.  The streets filled with people.  This city loves its teams, its scary and admirable.  I\'ve been on record since July saying the Sox would win it this year, as there\'s quite a bit of something to this team.  That said, as a yanks fan, I predict Yanks in 4 for the pennant.  I know it doesn\'t make sense.  Regarding the Pats, they\'re gritty.  They only play well when they\'re dogs.  I have this sneaking suspicion that Belichick cut Milloy to motivate the team.  With 5 of their top 8 players on the defensive front out, the defense has been superb this year.  It doesn\'t show in the numbers necessarily, but the defense is as good as its been in 4 years.  The red zone D is solid.  All the young kids on defense, Wilson (now a safety), Samuel, Klecko, and Warren have all been productive.  The offense is its non-flashy \'try not to turn the ball over\' self.  I expect many close games this year again.  I think they beat the Gaints this week on a very good week from Brady and the receivers.  Truthfully, Shockey matches up well against the Pats D as all TEs do, but I think they\'ll do something to compensate in turnovers and special teams.  NE over NYG 28-20.  Non-sports, I\'ve been working alot.  I havent seen many pretty butterflies or sunsets.  My project is moving slowly due to my lack of time.  However, I\'m still paying my rent, barely.  I think I\'m going to listen to the Strokes and sleep tonight.    Blog it!   '),(18276,1826527,'06,October,2003','2003-10-06 00:00:00','    Y\'arr!   '),(18277,1826527,'03,October,2003','2003-10-03 00:00:00','    Hi everybody again.  So I am bored and decided to blog again.  I wrote three new songs today, so visit my  urlLink site .   I received my copy of the new Hey Mercedes CD, and it is quite enjoyable, as expected.  More on that perhaps tomorrow after a few more listens.  The first song is about the daily grind, and can be found  urlLink here , while the second song, about the upcoming season, can be found  urlLink here , and the final one, written mainly for  Kim  can be found  urlLink here .   '),(18278,1826527,'03,October,2003','2003-10-03 00:00:00','    Hello all!  Long time, no blog.  I\'ve been busy with school, work and Brian which leaves me no time to blog.   Football this.  Baseball that.  As the only female on this blog I say screw that shizzle!  Talk about more important things like butterflies and beautiful sunsets.  Wha?  As proven by my prior statement, I am not really into sports but I did catch some of the Giants game with Brian on Sunday and they made some pretty miserable calls.  Shockey was not taunting.. He was simply trying to get that guy off of him and the ball slipped.  See what having a reputation can get you?     Where has everyone been?  I guess I\'m not the only one not blogging..   '),(18279,1826527,'03,October,2003','2003-10-03 00:00:00','    I think I\'m going to stop making predictions before I realize how utterly ridiculous I feel.  Does that even make sense?  Anyway, I said the Bengals would win outright, but they lost in OT.  I still don\'t think Buffalo is all that good though.  All of Bledsoe\'s passing yards came after CB Jeff Burris left the field on a stretcher, and Eric Moulds was covered by a scrub.  Yeah, it was the key time to start getting yards, but still, they were lucky because he had 102 yards midway thru the 4th quarter...  I said last week the Red Sox weak link in the lineup was the leadoff spot, which is Johnny Damon.  I like Damon, but I just felt he\'d been so inconsistent during the year, and yesterday he hits a 2 run homer.  I really hope the Sox win because I want them to beat the Yankees.  I know just about everyone here likes the Yankees, but I don\'t, and frankly I\'m sick of them.  After the game, they decide to interview Jeter surprisingly.  Why?  He hit a meaningless solo HR?  How about naming Nick Johnson or Matsui player of the game when their 2-run hits mattered in the game, or maybe even fat David Wells for pitching [sigh] another important game in the postseason.  Anyway, it\'s Pedro vs. Zito tonite, and the Sox need a big game from everyone, but it\'s time for Kevin Millar to step it up.   urlLink Tomorrow is the day!   Oh yeah, how about those Cubbies?  Giants: check; Braves: check; Yankees: not yet.  Hey 2 out of 3 gone from the playoffs is a  great  start.  I think I\'m going to root for the Cubs, but I will not be saddened if the Marlins win.    Hey Rob-what\'s the deal with New England?  Their a strange team.  Blown out by Buffalo.  Blow out Philly.  Eek by the Jets.  Get eeked out by Washington.  Out-gun Tennessee.  What\'s the pattern here?  Do they get out-gunned by the Giants next week?  God I hope not.  The Giants are so boring.  Big changes may be in the works for my brother.  Good changes.  Hopefully after tomorrow things will be fit to brag about.  I have an interview later this week  urlLink here .  They have two 5th grade openings, so we\'ll see what happens there.   '),(18280,1826527,'03,October,2003','2003-10-03 00:00:00','    Welcome, weekend.  I\'m still wondering where the end of summer went, but anyway, it\'s still nice out today.  I got a haircut this morning from the lovely amateurs at  Supercuts .  The girl I had today is usually pretty good, but she always cuts the back crooked-like, so I have to fix it.  It\'s fine if it\'s a little bit off, then I wouldn\'t care, but when it visibly slopes downward to the left, it looks sorta funny.    I went to the bank to attempt to cash my check from GHT, but there was a line out the door full of old people.  It must be social security check day.  Yup, it\'s the 3rd of the month, so I left because I probably would still be there since 10:30.    I\'m listening to  Cursive  right now.  This is a very strange but good cd.  It has organs and cellos in abundance.  Very eerie.    My Red Sox prediction was wrong.  Boy do they suck.  How can they not score with that lineup.  Their lineup reminds me of an even better than the Jays of 93.  They have really no weak link, except for maybe in leadoff.  I guess pitching does win games, and their bullpen bites it big time.  Pizza night.  Tomorrow it\'s supposed to rain because we have a game to do.    I realized even more that I hate the South.  Folks in Richmond refused to name a street after tennis legend Arthur Ashe because of his skin color.  Fucking muffs.  Go buy a clue instead of a confederate flag.   '),(18281,1826527,'02,October,2003','2003-10-02 00:00:00','    Hey folks, sorry for the long absence on my part.  My computer was diseased and being repaired for a few days, so I was without this piece of technology.  Let\'s take a look back to my weekend predictions from last Thursday.  The Bengals won, so that was nice, and Syracuse kicked Toledo\'s ass, which surprised me.  Either the Cuse is much better this year, or Toledo got lucky beating Pitt.  Anyway, Walter Reyes is one of the top backs in the nation, so I\'m happy with their 3-1 start.   Did ya see my boy Shannon Stewart on Tuesday?  I know you probably did, but I had to rub it in a bit to the Yankee fans.  Also, I hope Arod doesn\'t win MVP because the (Texas) Rangers suck.  I also don\'t want Tejada to even get mentioned because he can suck my ass too.  If it\'s going to be a Yankee, then I hope it\'s Posada.  As long as Halladay gets his hardware.    Let\'s look ahead to this weekend.  Starting today, I hope they don\'t but I think the Yankees will win game 2, as well as the Sox.  On Sunday, the Bengals are at Buffalo, which is not good.  Takeo Spikes may decapitate anyone he tackles, and Corey Dillon is not playing.  However, Travis Henry is questionable, and may not play.  I\'m gonna be bold and say the Bengals win again.  Call me crazy, but I have renewed faith in Jon Kitna and his receivers.  And I also don\'t think Buffalo is as good as I thought when the season started, especially after watching them the last 2 games.  I went on an interview yesterday for a teaching job, but couldn\'t tell you how I did, because the principal showed no emotion during the time I was there.  I\'m supposed to find out by Weds if I made the final 4.  I saw my professor yesterday for a reference for grad school, so now I am done with my application process.  So now I wait to see if I get in. I got a letter yesterday telling me I was denied for financial aid, but I don\'t know why there was no mention of loans.  I have to call the financial aid office to figure out what the deal is.  On a delayed realization, I noticed it\'s October.  It didn\'t gradually get cooler, like ease into the low 70s or hi 60s, but it just became 60 and dropped into the 40s overnight.  What happened to the light jacket at night weather?  It\'s already feeling like hoodie weather.  Hey Den, what gives the hotel the right to call itself a \"historic bordello?\"  Was Lincoln a customer?  How about those Scarlet Knights, huh?  Putting up 30 a game.  Oh by the way William Paterson lost last weekend 67-0.   '),(18282,1826527,'29,November,2003','2003-11-29 00:00:00','    It was nice seeing everyone last night, but we really missed Garet.  Anyway, it was cool catching up at dinner and going to the fabled  Town Pub  for their good grog.  We didn\'t see the folks we wanted to, but instead saw a couple of handfuls of high school classmates, most of which we didn\'t say a word to.  No real loss there.  One interesting note, however, was that the Crow walked into the bar at about 10.  He emitted a cold wall around him and sat down at the bar.  There was eerily a seat on either side of him unoccupied.  If everyone comes home again for Christmas, we should try to get out again, hopefully to a better eatery, and plan ahead a bit so my work schedule doesn\'t screw up the meeting time.  In other news, I realized I didn\'t ask anyone how their Thanksgiving was yesterday, so hopefully all went well, and hopefully Matt\'s trip to FL went smoothly...The Cuse plays RU today.  They\'ll probably lose, but even if they win, they probably won\'t get into a bowl game, which sucks because even though their record is not very good, they have a great runner and a pretty good qb that the nation should see.  Anyway, the Sixers remain in 1st despite their paltry 8-9 record.    It\'s windy today and I arrived at the conclusion that if it\'s not raining, then it\'s windy.  One thing that made me feel better about myself:  I read in the paper that people were waiting in parking lots of stores yesterday at 1 am for the stores to open at 5 am.  I wasn\'t one of those people, and never will be.   '),(18283,1826527,'25,November,2003','2003-11-25 00:00:00','    Hello again, May I begin by telling Kim that your advisor appears to be an idiot. It\'s unfortunate when someone that is supposed to be helping you with your scholastic career never really takes the time to get to know you or at least look at your records and see what kind of student you are. Today was probably my most frustrating day at work yet. My supervisor has taken half days all this week, which isn\'t bad, except we\'ve gotten hit with all sorts of new admissions, and while working with brain injured folks sounds like the most fun you could have ever, there are times when it can be a little bit tiresome. Just have to get through tomorrow though. I hope everyone\'s travels go smoothly and safely. Oh yeah, say hello to Bill at the Town Pub for me.   '),(18284,1826527,'24,November,2003','2003-11-24 00:00:00','    Hello everyone,  I\'ll be flying home tomorrow for Thanksgiving.  Friday sounds like a good day to meet up.  I\'ll talk to the other half tonight and see what she says.  We can do the Alexus thing again like last winter break or just hang with Jarrett and Bill Rupp at the Town Pub.   '),(18285,1826527,'24,November,2003','2003-11-24 00:00:00','    That stuff is messed up, Kim.  I hope the Dean gets things straightened out.  I think everyone should be given an opportunity to do what they want, regardless of what people think you can do.  This is ignoring the obvious fact that he\'s a dope who should see you\'re well prepared and have a better chance of success than the vast majority of people.  So why he thinks you can\'t succeed is pretty much a mystery to me, and somewhat mindblowing.    Anyhow, I\'m comng to NJ for Thanksgiving.  I\'m taking a greyhound in the middle of the night from Boston tonight.  I\'ll be in tommorrow morning.  It\'ll be good.  I\'ll call you Bri to see if we can meet sometime.  Blog out.    Happy Thanksgiving.   '),(18286,1826527,'24,November,2003','2003-11-24 00:00:00','    So here\'s the update on my \"What Should My Major Be?!?\" saga...  I went to see my advisor today and he told me I was stupid.  Well not in those words really.  I asked him if the math class I was taking now could possibly count as a precal class and he said no way.  Ok.  Fine.  Whatever.  He then told me that I was not smart enough to take a Calculus class.  What the fuck?!? I don\'t mean to sound conceited, but does he have any clue who he is talking to?  I graduated high school with advanced calc.  I have pulled nothing but A\'s in all of my math classes to date.  I have a 3.8 GPA for crying out loud!!  He has some fucking nerve to say I\'m not smart enough.  However, I did not say this to him to his face because I like to avoid confrontation at any cost.  Anyways, I sent an e-mail to the dean and told him what happened.  I wasn\'t mean to him at all but I did get my point across.  Let\'s see if this guy replies.  I know nothing will probably happen.  He\'ll tell me that he thinks I should talk to my advisor about what happened and work things out amongst ourselves.  It\'s also my word against the advisor\'s.  Guess who\'s side they\'ll take?  (Hint: It won\'t be mine)  This event has given me yet another reason to complain about Montclair.  Don\'t get me wrong, it\'s not a bad school but Stevens looks more and more tempting as the asses at Montclair continue to be.. well.. asses!   '),(18287,1826527,'23,November,2003','2003-11-23 00:00:00','    Well, nothing new from the void known as Connecticut.  The gf and I saw  Lost in Translation  on Friday.  I loved it.  I had lunch in Boston on Saturday and did absolutely nothing but watch sports on Sunday.  I actually got to see the Rangers win a game...  This short week can\'t go fast enough.  I need to get out of here for awhile.  I will be in Bloomfield until Saturday morning, so I am game for dinner/drinks/whatever on Friday nite.   '),(18288,1826527,'23,November,2003','2003-11-23 00:00:00','    I\'m counting down the days until Thanksgiving.  I\'m off on turkey day, work Friday, and then I\'m off Sat & Sunday.  That\'s nice, except for Friday, but anyway.  If you actually are still coming home this weekend, perhaps we can do something Friday after I work(7 pm).  I have one of those throw-away cameras left over from vacation to Myrtle Beach.  Maybe I should finish it this weekend, since it\'s been sitting here on the desk for 4 months.  I also want to use my digital camera a bit, since I haven\'t been using that much either lately.  Hopefully I\'ll be hearing good news either today or tomorrow.  Montclair said I should know within a week, and today is a week.  In something related to school, my chiropractor told me I should sub again.  He says I should pick somewhere I want to work, get on the sub list and hopefully something will open up.  He says I\'m young and healthy, so I should wing it for a while without insurance.  It sounds like a great possibility, however, I did this already earlier this year at Clifton and Livingston, and they didn\'t hire me back.  I also don\'t particularly like subbing, since it\'s like starting a new job every day.  I\'ll probably see after the new year, if I can last at my job until then.  I think I may be developing a cold(surprise), so I\'m trying to drink lots and take extra vitamins and such.  Who knows if anything really helps.  I tried getting some sleep, but I can\'t seem to sleep late anymore.  Gotta be at work in about 3 hours, so I\'m not too thrilled about that, except I won\'t have to watch the Giants-Bucs game tonite.  Gross.   '),(18289,1826527,'23,November,2003','2003-11-23 00:00:00','    Elf was a pretty good movie.  Perfect for the holidays and Will Ferrell was perfect for that part.  I couldn\'t really think of anyone who could have done a better job than him.  I would recommend it to all.  So after some thinking, I think I\'m going to stick to being a math major.  I had this epiphany in the middle of my Accounting class on Thursday morning.  As the professor was explaining how to journalize bed debt expenses, I thought to myself, \"Wow.  This shit really sucks.  I have no desire to do this for the rest of my life.  It\'s just so freakin\' boring.\"  So, after I had this change of heart, I made an appointment with my advisor for Monday in hopes he can answer some questions about scheduling for me.    I have some good news!  I spoke to my Calc professor on Thursday night and we were talking about my change to a math major.  I was planning on asking him about a possible internship of sorts at his company.  The guy beats me to it and comes right out and says, \"I would love to give you a job.\"  Well thats fine by me!  I would start in January or February as a financial planner in training.  If that\'s what you want to call it.  He said the job is pretty fast paced and intense.  This is ok with me since I do absolutely nothing where I work now.  The only downside is that the job is in NYC.  Oh well.  I\'ll get over it.    Now it\'s time for me to get ready to go to Red Lobster.  Mmmm.. Cheddar bay biscuits, cesaer salad, and lots of garlic-y shrimp.   '),(18290,1826527,'20,November,2003','2003-11-20 00:00:00','    It\'s a rare weekend blog.  I couldn\'t sleep as late as I had planned, so I went surfing this morning.  I found the Sixers lost to the Knicks last nite.  Wha?  I deleted my 60 junk mails.  I decided to blog because I\'m bored.    Last nite, Kim & I saw  Elf .  It was pretty funny.  It wasn\'t exactly what I was expecting, though.  I didn\'t know what to expect, actually, but I figured it would be more of a Will Ferrell-type comedy, and it was more of a cutesy-type of comedy.  It had some extremely funny parts, all while staying with relatively clean humor.  To make it better, the Allwood theater dropped their ticket prices back down to $6 from $8.  That was nice.  Today I\'m going to lunch with Kim and her mother.  Red Lobster is our destination, and we haven\'t been there in a while.  I\'m sure we\'ll have to wait 3 hours from all the damn religious people going there after chuch.  Since my fiasco on Thursday nite, the Vandals have gone 2-1.  I replayed my game with Philly and lost a heartbreaker 28-21.  I then rattled off 2 in a row with wins over the first place Jets(Wha?) and Patriots.  I smushed the Jets\' rushing attack, and had a nice game with newly acquired starting QB Jeff Blake.  He was acquired for Chris Chandler and Larry Centers.  I then had a good rushing game from Brian Westbrook against New England.  His first carry went 79 yards for a TD and tied the game after the Pats opening game drive, which resulted in a David Patten TD reception.  I ended up actually holding NE on a few drives, which is the first time I was able to play good defense(or maybe they just played poor offense, but whatever).  Anywho, I\'m 2-2 now and sitting in a tie for second in the AFC East.   '),(18291,1826527,'20,November,2003','2003-11-20 00:00:00','    I\'ve got to tackle a few things before the week ends.  First, my brother was on tv, as Den mentioned, because he has been a fill-in for WFAN doing the 20/20 sports flash.  He was on tv because the YES network airs Mike & the Maddog, and he was doing the flash during that time slot.  It\'s pretty cool hearing him...  Next, Montclair found my scores.  They actually found them the same day I blogged, but I hadn\'t had time to blog.  More incompetence in this world...  Third, I purchased a new game for my PS2(I know have 3 games!).  I got NFL Gameday 2004 just yesterday.  It\'s pretty hard so far.  I played one game with the Bengals yesterday and got obliterated, but I didn\'t get mad because I knew I would suck.  I decided to play another game to see if maybe I got the hang of it, but ended up turning it off in the first quarter.  I read the instruction booklet fully and found the game has a cool feature in which you can create your own expansion team-design the uniform, logo, etc.  Pretty cool!  So I created the NJ Vandals with some bad-ass brown uni\'s and a pirate-type logo on the helmet.  Anyway, there\'s an expansion draft where you select the rubbish left unprotected by the current NFL teams.  It was pretty realistic because the people available were horrible.  I can list my \"stars\" of the team: Brian Westbrook, Chad Morton, Freddie Mitchell, Willie McGinest, and Darrien Gordon.  Ugh.  My QBs are Brooks Bollinger, Chris Chandler, and Scott Covington.  Who?  Yeah.  So I did my draft and played a game and got beat by the Texans, but hung tough.  I think I lost 31-21.  It\'s very difficult to pass, because the controls are sensitive, plus Brooks Bollinger is my QB, remember?  I decide to play game 2 against the Eagles.  They score on the first play-a 80 some odd yard pass to James Thrash.  Once they get past your LBs, they\'re gone.  Slowness.  I get the ball and start running the shit out of it.  Westbrook had 170 yards, Lamont Warren had 130, and Dee Brown had 40.  I couldn\'t stop them, plus they had a KR, but when I kept running, the game stayed close.  I\'m down 21-10 at half.  Get the ball to start the 3rd, and continue running and hold the ball for 4 of the 5 minutes in the quarter before hitting Freddie Mitchell with a 5 yard slant for a TD.  Lamont Warren up the middle gets me 2.  21-18.  They come back and score in less than 2 minutes because Jon Ritchie is faster than my team.  Anyway, roughly 4 minutes left in the game.  I know I can win.  I\'ve been running all day, so if I eat up the clock and get down there, I can at least tie it and not let them get much time.  I run and run and run and punch it in with 15 seconds left.  You\'re thinking, \"He\'s still gonna lose somehow.\"  Wrong, I hold on to win!  Woo!  First win for the Vandals.  Post-game, I\'m checking out my roster to see if there are any free agents, and the game freezes.  My victory is gone, my 400 rush yards gone.  Man was I pissed.  So that long ass blog you just read technically does not exist.  Sorry.  So the game is pretty fun at least when you can eat the clock.  Otherwise, I would be destroyed.  Totally bummed about the lost info, man.  I may try to play again before bed if I don\'t fall asleep.  Peace out.   '),(18292,1826527,'19,November,2003','2003-11-19 00:00:00','    Hi to everyone, even Dennis.  I\'m in lab right now, basically struggling to stop myself from running far far away.  Optical alignment I don\'t like.  Honestly, there isn\'t much lab work I actually do like.  Anyhow, I used to do it pretty well, so I\'ll get used to it again.  Anyhow, my project is starting to take shape, though finishing it in time might mean not coming home for Thanksgiving.  I have quite a bit to do, and not alot of time left, as my productivity most of this semester has been...  not ideal.  I\'ll keep you posted about next week.  Go Bengals!  Jon Kitna and Johnson and Johnson and Johnson are good stuff.  They\'ve got a good shot at winning the division if they can get through this tough road stretch with 2 victories.  The division is theirs if they can.  I\'m just happy seeing them play competitive ball again.  I\'m also happy that the Steelers and Raiders blow.  I\'ve also never been much of a Miami fan, not terribly suprising.  I can see the Pats making the playoffs as possible this year, which is really all I can ask as a fan.  Since they\'ve actually won a superbowl, i\'d like the franchise to stay solid for a little while to make the championship look more valid.  Anyway, I think Eugene Wilson might be my new favorite player.  I know probably nobody else gets to watch him, but as a corner turned safety, he\'s done everything pretty well, including tackling and especially hitting.  Those are only noteworthy when considering that he can cover as well as a cornerback too.  Enough football.  Basketball, I think I\'m the only person not super surprised that the Bucks are actually ok.  The talent level isn\'t the greatest, but at least they have some youth and no super-egos destroying any sense of team.  See ya later Mason (aka the worst thing to happen to professional basketball ever)  Ummm.....  gonna go do some work now or something.  Blog out   '),(18293,1826527,'17,November,2003','2003-11-17 00:00:00','    Hey Brian.  Congrats to your brother and the Bengals!  My parents saw him on TV yesterday and the Bengals are in 1st place at the end of November (crazy).  At least they are more exciting to watch then the Dolphins.  As for the Nets, the slow start is bothersome, but Boston is only 5-5 in 1st place and as far as I am concerned the Celtic can\'t compete with the Nets.  The Western Conference is once again loaded, so I wsih the Nets would start playing to thier expectations.  it\'s funny how two straight trips to the Finals makes someone want more.  5 years ago, I was content with a playoff berth.  Well, I\'m off to watch The Towers: Extended Edition See you in about 4 hours!   '),(18294,1826527,'17,November,2003','2003-11-17 00:00:00','    Just stopping by to pass on a tidbit of information.  If any of you miss Opie & Anthony, go to  urlLink www.live365.com  and search for a station called WHIV.  They play nothing but whole O&A shows.  You will have to register for the site, but it\'s free and worth it to hear O&A again, even if they are repeats.   '),(18295,1826527,'16,November,2003','2003-11-16 00:00:00','    Well, I\'m back earlier than expected to complain.  I called  urlLink Montclair State  this morning to find out my status of my application.  They tell me they never received my GRE scores.  Wow, that\'s great!  I emailed them 2 weeks ago to make sure they had all my info, and they never emailed me back.  So I call ETS to make sure they sent the scores to the right address and confirmed the scores were sent on October 3rd to the exact address the guy gave me at Montclair.  \"They could have been lost in the mail,\" he says.  Well that\'s great, but maybe you could have told me about this so I haven\'t been waiting for 2 months for nothing.  So know they\'re \"checking\" the office and are supposed to call me back, yeah right.  UN-FUCKING-BELIEVABLE!  Forget about the Red Sox and the curse of the bambino.  My name is cursed.  I\'ll probably find out that the 100 resumes I sent out have been \"lost in the mail\" as well.  Oh yeah, I wanted to sleep a bit today, but was scared out of my sleep when two cars crashed directly outside my window at 6:30 this morning.  Well, I got my haircut, so that\'s something I wanted to do.   '),(18296,1826527,'16,November,2003','2003-11-16 00:00:00','    Wow, it was like a friggin\' reunion here on the blogger.  Everyone checked in.  What a nice surprise.  Den-thank you for the in-depth retrospective on your journey to N\'awleans.  Bam!  Kev-thank you for your input on the Matrix, however, I still don\'t plan on seeing it.  Matt-thank you for checking in and giving me my vacation destination for 2004.  Warm Springs, here we come!  Rob-thank you for your predictions.(Cincy 28, KC 20=really damn close)  As for those Bengals, how about the team actually being .500 in November, and to better that feat, they are in first friggin\' place!  Rudi Johnson again comes up big with 160 yards rushing.  Jon Kitna threw zero INTs, and Peter Warrick is developing into the player he was at Florida State.  Good things come to those who wait, and I\'ve waited 14 years.  Suddenly, the playoffs are not far off from reality, but I will not be disappointed if they finish with a .500 record and fail to make the playoffs.  They are playing close and their offense is scoring.  They\'d probably be fun to watch, but lord knows we can\'t see the Bengals on television, only Jets and Giants.  No bathroom, only Klau Kulash.  Tomorrow I return to working my Monday nite shift.  Woo boy am I excited.  The only positive thing that may come out of that is if I am able to get my haircut tomorrow morning before work.  I have a shaggy dog atop my cranium, and it\'s no fun.  I gotta call Montclair to see what\'s up with grad school as well.  I haven\'t heard a peep, and they never responded to my email from 2 weeks ago.  If time remains, I have to clean out the trunk of my car.  It tends to leak.  Not like \"Oh, it\'s a bit damp from the monsoon we just encountered.\"  More like \"Oh, there\'s a swamp forming under my spare tire and there\'s some wildlife inhabiting the pond.\"  My tire jack is rusted, as is my tire iron.  The rim of my spare is rusted a bit, and there\'s a slime on the tire itself.  Not a pretty sight or smell.  I discarded the carpet at the end of the summer because it was ruined, and now I don\'t want the entire trunk to rust through.  In other news, I realized I have an abundance of cds I never listen to, and probably never will because I don\'t like them.  I have an urge to buy some new ones, but I haven\'t had much luck in buying stuff that I\'m unfamiliar with.  I may go browse around right now actually, since I\'m not too tired.  I\'ll be back probably mid-week to whine about work I\'m sure and maybe other things that are pissing me off.  Here\'s a possible preview: traffic, weather, the Rangers, my paycheck.   '),(18297,1826527,'14,November,2003','2003-11-14 00:00:00','    Hello all, Back from New Orleans.  Great food; oysters, shrimp, catfish, po boys, bread pudding, all spicy!!  The conference was good too.  Over 30,000 neuroscientists.  I actually had a Nobel laureate stop by my poster and tell me I was a doing a good job.  FYI, the Nobel Prize is equivalent to making the Hall of Fame for science.  That was pretty exciting.  Now, time to explain the nightlife.  Day 1-Head to Bourbon St.  Immediately stop by Patty O\'Brien\'s for some hand grenades (from Kev\'s recommendation).  After 4 hand grenades, I realize I am extremely drunk.  Walk down Bourbon St and see some great sights from the ladies (hint hint) at the balcony of the Sonesta hotel.  Have 2 more hurricanes and proceed to dry heave in my hotel.  Pass out on the bathroom and wake up at 12.  Day 2-Wake up at noon and get lunch.  After the conference, we decide to go to Harrah\'s for drinks and gambling.  I played blackjack for a while and win $140.  Of course we have to celebrate.  Enter Bourbon St.  After another blur of a night, we get back to our hotel at 4:30.  Sleep for a few hours then go to the conference  Day 3-After the conference, we head to Cafe Du Monde for some chickory coffee and beignets.  After a tasty meal, I go back for an early sleep since I have to present my data the next day.  BTW, we stop by Harrah\'s and I lose $20  Day 4-Poster session went awesome.  I had a lot of important people in the field check out my data and I got a lot of positive feedback.  My old boss was there and she took me out to Nola\'s (Emeril\'s restaurant) for lunch.  I had some awesome gumbo and jambayla.  We then go to the Acme Oyster Bar for dinner.  After shucking lots of oysters, we hang out on Bourbon St once again.  I stick to Abita beer and have a good time.  BTW, we stop by Harrah\'s and I lose $20 (notice a pattern here)  Day 5-Grad students from the Univ of Colorado-Boulder have a huge house party in the Garden District.  Big mansion and lots of free beer  Stay there until 4 and come back.  On1y 1 day left in the Big Easy :(  BTW, we stop by Harrah\'s and I lose $20.  Day 6-Last day in New Orleans.  We have a lesiurely day and night, since we are all flying out early the next day.  All the guys in my program decide to stop by the Barely Legal club for some entertainment, which was very interesting.  BTW, we stop by Harrahs, and I come out even!!!!  Day 7-Say goodbye to New Orleans.  All in all, a great trip!! Thanks to UNC for picking up the tab!!    Anyway, I will be back for Thanksgiving from Wed-Mon, if anyone wants to hang out (even Rob)    Kevin-thanks for telling me what to drink in N.O. (my stomach does not thank you though)  Finally, as for school.  In my situation, I started off in science and wanted to go to med school.  I realized that I didn\'t want to do it and ended up in psychology.  I then decided psych was not enough since everyone has psych degrees.  I then worked in a lab and sort of went back to biology.  Now I enjoy what I do.  I look at behavior and what goes on in the brain.  So I basically combined psych and bio together.  Basically, maybe combine math and business and become a consultant.  That way, you get money and enjoy what you do.  Wow,  that was a big blog for me!  Talk to everyone latter, have a good weekend!!   '),(18298,1826527,'13,November,2003','2003-11-13 00:00:00','    Minor addendum:  The best part about  Revolutions  was the trailer for  The Punisher .   '),(18299,1826527,'13,November,2003','2003-11-13 00:00:00','    Good bloggy morning to you all.  As promised, I am back with my review of  Matrix: Revolutions .  Here goes:  It wasnt all that bad.  I would say it was better than  Reloaded .  They left out a major portion of the philosophy that killed  Reloaded .  This was more of a straight forward action movie.  Per usual, the movie was visually stunning and the acting was atrocious.  It gives good closure to the trilogy.  My only major complaint, other than the acting or dialogue, was the fact that in order to understand the movie completely you have to see the first and second movies, plus the  Animatrix  and the end of the Matrix video game.  All in all, I would not recommend you run out and see it immediately; but I would not try and dissuade you from seeing it if you told me you were planning to do so.   And I guess I will add my $0.02 to the whole Stevens/major/life thing.  As for Rob\'s comments on Stevens, I agree with all of them except for the part about the established programs versus the small programs.  I was in an established program (mechanical engineering) and there was a lot more bullshit and too much structure in that program.  Since it was \"established\", they basically told you that had to take course A,B, and C and that you could choose one or two \"electives\" from a list of courses X,Y,and Z.  There was basically zero leeway in making your schedule.  Where as the smaller majors had a lot more flexibility in their scheduling, but that puts more pressure on the student to make good decisions.  Wherever you were to go school, just make sure that you major in something you truly enjoy.  I majored in something that I was proficient in, not necessarily something I really like.  That was probably my biggest mistake.  I thought about job opportunities and compared that to what I was good at, and that\'s how I came up with my major.  Not a good way to do it.  My best suggestion to any one still in school is to find an internship at a place that would be similar to the work you would be doing after you graduate; and make sure you enjoy it.  You do not want to be stuck the rest of your working life in a career you dont like.  That\'s enough out of me.  I hope this day goes by quickly, because Meredith and I are going to see Boston\'s own, Dropkick Murphys at the Webster Theatre in Hartford.  I cant wait.  I am almost as giddy as a school girl...   '),(18300,1826527,'13,November,2003','2003-11-13 00:00:00','    Greetings and Salutations, I am currently even further south, residing in Warm Springs, Georgia, and yes, as small as you can imagine the town to be, it\'s even smaller. I have to drive about 15 minutes to get cell phone reception and to do any grocery shopping. The rehab hospital I\'m interning at is pretty cool though, and there are plenty of other interns around so it\'s not terribly boring all the time. What I do fear is that I will undoubtedly be an alcoholic by the time this internship is over. The last two Fridays, the drinking has started at around 6 pm, only because there is nothing else to do on Friday night, except maybe go to the one of two places in town that sell alcohol and drunkenly sing karaoke along with the locals and their three good teeth and immaculately maintained mullets.  Unfortunately, I will not be home for Thanksgiving, I only have Thursday off, so I\'m taking a personal day Friday and going down to Florida to see dear old dad. So, no NJ for me until Christmas, I booked my flight yesterday actually. One nice thing about the location, I\'m only about an hour and 20 from Atlanta, so I at least can enjoy some good concerts. Went up on Monday and saw Thursday, Thrice, and Coheed and Cambria. It made me happy, even if I had to drive extra slow once I got on the backwoods roads to get back to Warm Springs, first to avoid running over multiple deer and second, I don\'t want Sheriff Roscoe pulling over my out of state vehicle and have me wind up rotting in some 10\'x10\' local jail cell for the next 25 years. Big ups to Bri for the Big League Chew reference, it was nice!   '),(18301,1826527,'13,November,2003','2003-11-13 00:00:00','    Thanks for the input on Stevens.  I really do appreciate it.  I e-mailed the advisor in the math department at Montclair today to set up an appointment to ask him a few questions.  Now all I have to do is sit back and wait for a reply or figure what his office hours are and harass him then.  I know I probably won\'t be 100 percent about what major to take until I\'m already working and that is unfortunate but that\'s the way cookie crumbles.  As a matter of fact my current math professor made a very strong suggestion about me working for him in the future.  It would defintely be nice to work with him as a financial analyst/advisor.  He\'s an awesome professor (this is my second time taking one of his classes) and he seems to be an all around nice guy.  Even if it\'s just to get some experience I think it would be worthwhile.   Well I have to go study some more for a test tomorrow.  Thank god there\'s only 5 weeks left in the semester.  But then there\'s 2 more years to go.  Sigh.   '),(18302,1826527,'12,November,2003','2003-11-12 00:00:00','    I, three will be in Jersey for Thanksgiving, as always.  I\'m going to be at my brother\'s house Thurs, and rumor has it my boss wants me in at 6:30 am on Friday, so I shouldn\'t be working that late on Friday.  I don\'t have a game that Saturday, either, so I\'m game for a shindig.  I will know more as the time approaches.  I also appreciate the input you guys gave to Kim (as I\'m sure she appreciates it too).  I want her to be happy, but like Rob said, I told her to find out as much as possible about the program, wherever it may be.  Kev, the news was on the F2F site, and they emailed me, so I think it\'s for real.  It is really sad.  When I got home from work tonite, I wanted to get a haircut.  I wanted to call to make sure they were open and there wasn\'t a wait, and sure enough when I called, I found out they were closed.  Supercuts isn\'t so super are they?  I\'m trying to figure out when I can go, because I won\'t have time Saturday morning.  I think they open at 9, and I have to leave for work at 9:30.  I don\'t want to work with stray hairs itching my neck all day.  As of now, my only option is to go Sunday morning, that is, if they are open.  I am so relieved tomorrow is Friday.  I just want to chill and not have to wear a tie again until Monday.  If I don\'t be bloggin\', then hope y\'all have a fun-filled weekend.   '),(18303,1826527,'12,November,2003','2003-11-12 00:00:00','    Smell ya later?  awesome.  I feel compelled to comment on Stevens.  It is wicked expensive.  But mortaging your life away is sometimes feasible with government loans and maybe in state grants.  I remember that Kevin and I both ended up with some weird state grant in our financial aid packages.  Also, I\'m like seriously in debt still, but government loans are more flexible and have lower rates I think.  All I know is that i can keep deferring them until I finish school.  So, I don\'t know how practical it is, but still possible.    As for the school itself, its a pretty good school, but its still kinda small.  The established programs or ones with lots of people (i.e. not physics) are easier to deal with the red tape since there is a better support system.  This is something I would have changed if i could go back and do it again.    Also, if you really enjoy math, I say go for it wherever you are.  Pretty much all technical careers (science, engineering, business, law, accounting, etc.) are going to have boing aspects to them sometimes, if due to nothing else other than the fact that you must deal with practically the issues in front of you, not necessarily the other aspects.  It\'s just that some fields have less of that than others and it really depends on your own personal tastes and interests.  Just remember that when dealing with something like your own interests, it never hurts to get as much information as possible.  You\'ll never be 100% certain probably, but you\'ll get an idea or feeling that might be useful.   '),(18304,1826527,'12,November,2003','2003-11-12 00:00:00','    Afternoon, ladies and gents, I\'ll keep this brief...  - I hope you\'re joking about Face to Face, Bri.  If they broke up, I am offically saddened.  At least I got to see them for the first (and I guess only) time at the Warped Tour this past summer.    - Work sucks (but at least I get a paycheck twice a month), and I miss Stevens.  Ok, I actually miss Hoboken, but nonetheless...  - I love the fact that all of the sports teams I like are mediocre this winter.  Is it Spring Training yet?  - I will be seeing Matrix:Revolutions tonite with some friends.  I am only going because they all want to go and I dont feel like sitting on my futon alone tonite.  I\'ll give a review tomorrow.  - I too will be in Jersey for Thanksgiving (probably Wednesday nite til Saturday morning).  Give my cellphone or parents\' house a ring if any of you would like to get together.  Smell ya later...  ::End Transmission::   '),(18305,1826527,'12,November,2003','2003-11-12 00:00:00','    Hello to everyone in the merry land of bloggers!  New Jersey is being hit with some brutal weather today, or at least it\'s being predicted.  The wind is ridiculous.  As I was walking to my class this morning, my glasses almost flew off of my face.  That would have looked pretty funny if it happened. hehe.  Anyway, it was like being in Oz.  All I was missing was Toto and some sidekicks missing various body parts.  As I was taking my accounting test, the windows were shaking and several people got up in the middle of the test to sit away from the windows.  Crazy huh?  I hear the winds are only supposed to get stronger and we might get some rain.  Hopefully, the winds will knock down all power lines on campus so I won\'t have to go to my night class.  Gee, can you tell I\'m with Rob in my \"I hate school\" attitude?   Since we\'re on the topic of school, I am considering ditching the idea of a business degree and sticking with my math major.  I think that I could have more job possiblilties with a math major.  As Brian told me, \"Business is boring.\"  I think he may be right.  Working for the NSA or NASA would be much more interesting.  I have also considered applying to Stevens again, but I don\'t think it would be possible for me to go there for several reasons: 1) the commute would suck, 2) I need to work a lot to pay bills, 3) living on campus would be nice but it\'s expensive (as if the tuition wasn\'t enough to scare you), 4) i would have to quit my current job(where I have job security and decent pay) and find a new one closer to campus.  As much as I would love to go to Stevens, I don\'t think it would be the best thing for me to do at this point.  Maybe if I win the lottery I can go.  Besides, they require you to take 6 gym classes.  I\'ve been bitching about taking one gym class at Montclair!  I couldn\'t imagine taking 6.    Peace out fellow bloggers!   '),(18306,1826527,'12,November,2003','2003-11-12 00:00:00','    Hey y\'all.  I just found out another one of my favorite bands is over.   urlLink Face to Face  has officially called it quits after something like 13 years or so.  I\'m extremely sullen, especially since I really liked their last cd.  Bummer...  Had a poop day today at work.  We\'ll leave it at that.  In reply to Rob-I am glad the Bengals have someone who talks smack, because they need an attitude to portray their team.  I don\'t know whether it will fire up the Chiefs, but I don\'t think the Chiefs really need to get fired up since they\'re undefeated.  I actually believe the Bengals have a real good chance of knocking them off, especially at Paul Brown.  And so what if Chad comes out wrong after the game;  they\'ve got nothing to lose.  The rest of the AFC North is also not very good or consistent.  At least the Bengals have been playing hard all season.  As for the Pats-I truly feel they will dismantle the Cowboys.  I\'m sorry, but I think their start was an illusion only made more appealing because of their fat coach.  Quincy Carter is shit, and yes, their defense is good, but the Pats seem to be playing with a mission in mind.  I haven\'t heard anything recent by the Strokes, and the only thing I know about them is that I was sick of hearing about them when their last cd was out and how they were saving rock and roll.  But I have heard many positive things about their new stuff, so I\'m glad you\'re enjoying it.  I haven\'t heard much new music, I guess because I listen to Q104.  Is Sparta back from seeing boobies in NO yet?  Waiting patiently for his review.  And also his take on the Nets slow start.  What is the deal?(In an annoying Seinfeld voice)  That\'s all for now.  See ya.   '),(18307,1826527,'10,November,2003','2003-11-10 00:00:00','    urlLink I still love the teen girl squad.   Hi Folks  I haven\'t blogged in quite some time.  I decided to get you up to date on important things.  Lets see...  I still have a poster of Gigli up on my wall.  I don\'t think its quite as funny anymore.  In fact, I\'m weirding myself out now.  My project has gone nowhere in about a month due to my not doing much with it.  I am off tommorrow and intend to spend the day there.  I pretty much hate my school, so my aversion to being there is my main obstacle to tackle.  However, I can spend all day tommorrow there.  I\'m thinking about looking into working at the patent office or something.  I get to use my science and engineering background, but not have to work in a lab.  I\'ll look into it.    I think the third Matrix movie probably sucks ass, but I won\'t see it since I thought the first was ok at best and the second blew.  I did catch Elf sorta randomly and thought it has some funny parts for sure.  Overall, thumbs up for a Christmas movie.  I haven\'t left my house for days due to the overbearing fear of being permeated my immense amounts of christmas music.  Does not that assault on our ears start soon?  Ok, maybe later.  I was joking about not leaving the house though.  Or was I?  Yes, I was.    I like the new Strokes CD.  It reminds me of the Cars vaguely with the almost keyboard-guitars and sort of backbeat on some of the songs.  I was kinda afraid it would blow, but I think lyrically its still pretty solid.  Musically, its similar to the first obviously, but there is enough variety to suit me in the song to song variations.  In particular, I enjoy \'Under Control\' for its smoothness, \'What ever Happened?\' and \'Reptilia\' for weird chord progressions on the chorus ( I think thats what I\'m describing. ), \'I can\'t win\' overall, and \'12:51\' since it mentions 40\'s.    Bri, What do you think of Mr. Johnson\'s bold prediction?  I hear various media pundits saying how it was such a bad thing and that it will motivate the Chiefs, but I don\'t think that will necessarily be the case.  I think the Orange and Black need to stay fired up and bold, at least I hope so.  Seriously, the Chiefs can\'t be the best team in football anyway, can they?  Maybe they are, but I don\'t trust their defense yet.  We\'ll see in the playoffs.  Anyway, my Pats look good through their bye.  This week the Tuna will come back and people will make another big deal out of it.  Its been long enough that there are hardly any players left here from his tenure (Law, atta-boy Troy, Bruschi, McGinest, Johnson) and none of them are exactly preoccupied with their first years in the league.  Two over-acheiving teams playing on Sunday night should be entertaining even if the media plays up the same boring angles all over again. (Belichick-Parcells; Parcells-Kraft; Glenn sucking for everyone but Parcells; Parcells\' daughter being married to Scott Pioli, the Pats co-GM essentially;and Parcells being God)  I think they should play up Eugene Wilson playing solid as a rookie at a position never played before this year, or Dat Nguyen being great all of a sudden, or the fact that Quincy Carter still isn\'t very great.  Anyhow, I\'ll watch it on Sunday night.  My predition: Pats 28- \'Boys 20.    As for the Kansas Cityans, Cincy 28- Chefs 20.  I\'ll be in NJ for Thanksgiving.  If anyone is around, I\'d like to get together with anyone, except that dirty dirty Sparta and his perverse ways.  Ok, maybe even Sparta.    Bloggedy Blog Blog.   '),(18308,1826527,'10,November,2003','2003-11-10 00:00:00','    Just a p.s., I didn\'t know Kim already blogged about the weather.  Didn\'t mean to be redundant and sound like an elderly person.   '),(18309,1826527,'05,November,2003','2003-11-05 00:00:00','    Usually, I\'m working at this time on Monday nights, but today my schedule was different for some reason, so I worked my normal 9-5:30 shift.  I don\'t think I have off this Thursday because my boss already hinted at this.  I believe last week it was 80.  Today it reached 45, and Wednesday, it\'s back up to 65.  That\'s some nice consistency there boys.  Hopefully tomorrow Roy Halladay will be named the AL Cy Young winner.  He definitely deserves it after the trooper he was for the Jays this past season.  We shall see.  Bengals 4-5!  Next week they have the Chiefs-the undefeated Chiefs.  Yikes!  I\'m scared.  I believe they can win, but I\'m afraid of the score if they have an off game.    Still haven\'t heard from Rutherford for that teacher assistant position, but today was the deadline to send in your info, so hopefully they\'ll call me.  I plan on spending Thanksgiving with my family at my brother\'s house, even if I\'m scheduled to work.  I already tried to tell my one boss about my plans, but she never really gave me any answer or response, so if I don\'t get this teaching thing and I\'m still at my current place of employment, I\'m at least going to see my girlfriend, my parents, my grandmother, and my bro and sis-in-law.  It\'s the middle of November, so I\'m pondering Christmas and its awful shopping necessities.  I\'m not looking forward to spending the little money I have, but I\'ve done it every other year, right?  I\'m somewhat tired, so I\'m probably gonna go watch some tv and go to bed early tonite to be rested for the remainder of the week.  Later.   '),(18310,1826527,'05,November,2003','2003-11-05 00:00:00','    College sucks!  I\'ve been having doubts lately about school.  I\'m at a point where  i just don\'t care anymore.  I don\'t know if it\'s because i get up really early for classes and I\'m just too damn tired to care or if I am truly fed up with my institution of higher learning.  Oh well.  I\'ll just have to deal with it, right?  In other news, it\'s cold in NJ again!  It was 70 degrees last week and then 3 days later - BOOM -back down to 40 degrees.  The trend of everyone getting sick is slowly returning and the germs will soon be traveling through the vents at work.  It kinda reminds of that movie Outbreak, but there\'s no monkey.    Brian said he\'s going to take me and my mom to Red Lobster next week.  What a guy!  I can\'t wait to eat lots of shrimp and those delectable cheddar bay biscuits.    Adios fellow bloggers!   '),(18311,1826527,'05,November,2003','2003-11-05 00:00:00','    Because I didn\'t mention it when I posted the other night...Tom Brady is one sick QB.  If there was a way to bet on the Pats when down in the 4th quarter, I would do it hands down.  Tomorrow is my day off.  Woo!  I feel like an elderly person on Wednesday nights because I\'m always fatigued and my back hurts.  I\'ll explain why my back hurts, so if you don\'t want to know then skip to the next paragraph.  My boss climbs a large ladder to reach the stock of Christmas candy that we are putting out for sale.  He proceeds to drop/throw these boxes full of candy from 2 stories up into my paws.  No you may think,\"God you pussy, it\'s only candy.\"  Here\'s my argument:  the boxes of candy contain twenty-four 10 oz bags(24 x 10=240 oz; 240/16=15  pounds).  It\'s like catching a bowling ball over and over again.  So my upper back is strained a bit, despite using my legs to support the weight like you\'re supposed to...  As per the recommendation of Dennis and Rob, I am currently reading  The Corrections .  I\'m going at snail\'s pace, because I only read it when I\'m doing my traction at night(that\'s the technical term for hanging my head and neck in a sling to correct my horrible posture and keep my spine from becoming igor the hunchback).  It\'s quite humorous, but not in a funny ha-ha way, rather a funny intellectual way.  I am enjoying it so far, but I have much to go.  I\'m on page 70, so have about 500 more pages to go.    As for Kevin\'s comment, I thought otherwise of your hand grenade comment at first, but then realized you were talking about New Orleans.  Duh!  How about those Rangers, though?  Huh? [nudge] Huh?  They are actually enjoyable to watch again, Kev!  As for other news, I miss Matt\'s presence around here.  I hope he\'s spreading his healing touch to the folks down south...You know what I haven\'t had in a long time?  Big League Chew.  If that makes no sense to you, then just let it go...  I\'ll blog to all of you blogs tomorrow.  Have a good blog.  Blog out, yo.   '),(18312,1826527,'04,November,2003','2003-11-04 00:00:00','    Dennis - Skip the Hurricane and go straight for the Hand Grenade.    Damn, I really need to get back to the French Quarter...   '),(18313,1826527,'03,November,2003','2003-11-03 00:00:00','    Home from my late night of work:a 2-11 shift.  I cannot even tell you how many people came in between 9 and 11 and bought cigarettes.  Don\'t these people know that shit kills you?  And they willingly fork over $5-$6 per pack.  Anyway, it was madd hot out yo.  It felt like it reached 80, but I don\'t know what it actually hit.  Hot nonetheless, especially for November.    Now I must go to bed in order to be at work for 9 tomorrow.  Truck day, so lots of lifting and sweating.  Woo!  In case I forget, Den, have a blast in the Big Easy, and say hi to George Lynch for me.   '),(18314,1826527,'03,November,2003','2003-11-03 00:00:00','    Hello all, Halloween was great down here.  They close off UNC and everybody is out in costume.  There was 75,000 people here this year.  First we went to my friend Garret\'s house where we had a cool Halloween party.  He has a bunch a DJ equipment.  Yours truly was spinning Outkast followed by Radiohead and other cool tunes.  After that, we went downtown with the masses.  All in all pretty cool.  This Wenesday, a bunch of us are going to see Matrix Revolutions on the IMAX.  I know most of you think Reloaded was crap, but on 2nd viewing, it wasn\'t as bad as I once thought.  If anyone has looked at the college basketball polls, UNC is ranked #10 with 1 first place vote.  Should be a pretty good year after our football disaster.  The Strokes new album is pretty good.  Definetly an indie vibe even though there are on a major label  Only 4 days till New Orleans!!! Hurricanes (the drink), voodoo, and the Hornets all in one city!!!  Talk to everyone later after my trip to the Big Easy!!!   '),(18315,1826527,'03,November,2003','2003-11-03 00:00:00','    Weekend over.  Well, almost.  Monday begins the first full week of November, and insists it will be a warm, mid-70s day.  I will spend this day working until 11 pm.  My my how this weekend flew by.  Friday nite\'s party at  Liz \'s house was a nice way to begin the weekend.  Saturday\'s late football start allowed me to get an extra hour of sleep before working on a very pleasant day, and Sunday allowed some quality time with  Kim .  I\'m applying for a teacher assistant position in Rutherford, which is full time and offers benefits.  If all goes well, then I will probably jump ship from my current position.  I will feel bad for doing such a thing, because Kim\'s mom was kind enough to get me into a working environment after months of unemployment.  However, Liz told me I have to do what\'s best for me, and that makes sense, and believe it or not, it\'s not something I hear often.  I will probably go more into this topic if I hear about this position, but for right now, I\'m gonna shut up and try not to jinx myself.  I searched for a couple hours online for teaching postitions and found a few more long-term sub positions that start in January, so there are still some possibilities.  I thought I\'d given up hope on teaching, but I\'m finally seeing some ads.  Hopefully, I will get some responses from this batch.  In sports-Bengals lost an important game, but I had a bad feeling because they were the favorites and were due for a letdown.  The Sixers probably need Big Dogg because their offense is horrible-scratch that-Aaron McKie is horrible.  I used to love the guy, but he\'s not the same player he was when they made it to the finals a few years ago.    I\'m getting tired.  Keep your fingers crossed for this teaching thing.   '),(18316,1826527,'29,December,2003','2003-12-29 00:00:00','    It\'s the end of another year.  There were many ups and downs, as usual with any year, in 2003.  I\'m about to recap the year that will soon pass.  Shows attended in 2003:  4 Shows that weren\'t Hey Mercedes:  1  Resumes sent out in 2003:  91 Interviews:  10 Teaching jobs:  2 Teaching jobs that were permanent:  0  # of times the word \"fuck\" appears in a blog in some form(not counting this one):  13 # of times the word \"unemployed\" appears in a blog:  3 # of times the word \"splendid\" appears in a blog:  0  high point of 2003:  getting accepted into grad school low point of 2003:  losing my grandfather.  I still miss him dearly biggest investment of 2003:  vacation favorite tv shows of 2003:  Scrubs, Family Guy least favorite tv show of 2003(that was actually watched):  Married to the Kellys favorite movies of 2003:  Pirates of the Caribbean, Finding Nemo, Italian Job least favorite movie of 2003:  The Matrix Reloaded favorite book of 2003:  A Winter Haunting least favorite book of 2003:  All the Dead Were Strangers favorite cds of 2003:  Loses Control(Hey Mercedes), Two Conversations(Appleseed Cast), Blink 182 least favorite cd of 2003(that was purchased):  Welcome to Poppys(Fun Lovin\' Criminals) goals for 2004:  gain weight, get strait A\'s in my grad courses, get a Jeep   '),(18317,1826527,'28,December,2003','2003-12-28 00:00:00','    Ok, I\'m back.  After our disastrous trip to SC this past summer, Kim and I decided not to go on vacation this year because of our lack of funds.  However, we also realized that we would be working our tails off in school this coming semester and working.  Yeah, we won\'t have a ton of money, but as we get older, we still won\'t have a lot of money to go away with, because we\'ll take on more responsibilities as we earn more money.  My brother had said this to me as well.  If we don\'t go and just say frig it, then we\'ll never go.  Yeah, we may not be swimming in our coin like Scrooge McDuck, but if we can\'t have fun when we\'re young, then when will we?  So we have a bunch of ideas brewing, but do understand we can\'t go overboard.  Sure we\'d love to go to Italy or Hawaii, but in reality, that\'s a bit too much.  We could probably do Arizona or possibly even Florida, depending on the Disney packages available.  Heck, we briefly discussed going to the shore, but we\'d end up spending just as much down there as we would elsewhere, and it\'s already been done.  I did some preliminary research on flights, and found Phoenix to be not so affordable at $350 round trip, and Orlando at $225 round trip.  Disney packages are quite snooty, but I ordered info so I can decide which aspects we can do without.  I don\'t mind either trip, but we would probably go in May, so we could avoid the dead of summer in both instances and not die in the desert or the Magic Kingdom.  More details to come as they come to me.   '),(18318,1826527,'25,December,2003','2003-12-25 00:00:00','    With one meaningless game in progress, the NFL playoff brackets are official.  I looked back on my preseason predictions and strengthened my beliefs in my not betting on sports.  Here\'s what I got right... Surprise team: Seattle(had them finishing second behind St. Louis, but missing playoffs) AFC Playoff teams: NE, Tenn, KC NFC Playoff teams: GB, Philly, St. Louis AFC Champ: Tenn over NE Super Bowl: Tenn vs. Philly  OK, so I didn\'t do very well, but that\'s fine.  The Bengals put a serious damper on my outlook by losing a poorly-played, must-win game.  I digress, however, because they weren\'t even supposed to compete this year, and Marvin Lewis gave them 6 more wins than last year.  8-8 is not a bad starting point.  In other playoff related news, Minnesota choked again, which is pretty sad after the start they had.  Here\'s what next week looks like...  Tenn @ Baltimore Den @ Indy Sea @ GB Dal @ Car  I\'m going against my word and picking.  I\'m going with Baltimore(unless McNair plays), Denver, GB, and Carolina.  If this holds true, the matchups would be...  Den @ NE Balt @ KC GB @ Phi Car @ St.Louis  I would take NE, KC, Philly, and Carolina in those games.  Then Philly and NE in the Big game, with NE winning because of Philly\'s injury problems and NE being a complete team.   '),(18319,1826527,'25,December,2003','2003-12-25 00:00:00','    Merry Christmas boys and girls!  I hope Santa was good to all y\'all.  I got many nice and needed things: *a new coat! *belt *wallet *two dress shirts with ties *3 shirts *one bear *Alias Season 2 DVD *Simpsons Hit & Run for PS2 *a cold *my car leaking even more than before  I think that\'s everything so far, but I feel like I left something off...Hmm.  Anyway, plans for today include Kim\'s house for dinner with her family.  To my Grandmother\'s house(Nanna) for dessert and more gift exchanging, then home to get depressed about another Christmas going by too quickly and having to work tomorrow.  Enjoy the rest of your day.  Hopefully I\'ll be talking to all of you.   '),(18320,1826527,'22,December,2003','2003-12-22 00:00:00','    I said I was going to tell you my grades so here goes (NOTE: if you don\'t care, don\'t read the next few lines):  Religions of the World: A- Italian I: A- Accounting I: A Business Law: A Business Calculus:A, BOOYAH Mr. Advisor who said I was dumb!!!  So there they are.  I\'m pretty excited that I did well but I am more excited that I have a month off from school :)  Merry Christmas everyone!  See you soon!!   '),(18321,1826527,'20,December,2003','2003-12-20 00:00:00','    So much to cover on today\'s version of the blog.  Let\'s start with the past first.  This weekend was a pretty enjoyable one.  No work, ate good, and had fun.  Saturday involved me, Kim, Liz, and Kim\'s mom going to Red Lobster to celebrate the birthday of Kim\'s mom(which is actually today).  Planned time to depart was 1pm.  Skepticism arose on my part because it was a Saturday, and the last Saturday before Christmas, and Red Lobster is on Route 46, just before the mall.  I feared sitting in a parking lot of traffic and waiting for hours to get a table.  Surprise, surprise!  Not much in terms of traffic, and when we got there, it was pretty empty.  We ate some dirty seafood and dessert and left for home around 3.  We dropped of Kim\'s mom and then took Liz home because we had to prepare for day 2 of the weekend.  P.S.-Sorry I missed your call Den, but as you can see I wasn\'t really home.  I don\'t believe the game went to 3 OTs...Anyway, I needed some warm, wool socks, which I\'ll explain in a minute why.  Consensus said to go to Kmart.  We went and they didn\'t have any wool socks, so I just got a pack of worker\'s socks with extra padding.  We return to Kim\'s place so she can pack up some stuff and back to my house so I can get my previously packed things and we\'re off to Point Pleasant.  (Tomorrow we are going to the Eagles-49ers game at the Linc.)  We arrive at about 6:30 and talk for a bit before going to eat.  Dinner was so-so, and we went to the video store to get a movie since it was still early.  As usual, there was nothing good available, so we rented  The League of Extraordinary Gentlemen , because Rachel had said she wanted to see it, and because there was nothing else.  The movie was pretty horrible.  Not much sense was made of it by me or anyone else in the room, but at least it made us tired.  Sleep.    We woke up on Sunday morning and got ready for the game.  Forecast was upper 30s, so we dressed in many layers.  It took about an hour and 15 minutes to get there, and it was early still.  We ate our sandwiches and walked around the stadium, which is really nice.  As we were walking, we noticed all the tvs had the Bengals-Rams game on, because I was the only one in the stadium wearing a Bengals coat and they knew I was cool.  Actually, the game was important to the Eagles because it represents home-field advantage throughout the playoffs.  Bengals are down 17-10.  We see bits as we walk from tv to tv.  Kitna has no time at all.  Rams drive and kick a FG, up 20-10.  More walking, Bengals get ball back and start running.  Nice drive going.  Roughing the passer on the Rams, move the ball.  Tony Stewart with a nice reception.  Ball down to the 2.  Bengals try corner route to Kelley Washington.  Intercepted in the endzone.  Old Kitna must have tied up the new Kitna in the locker room.  That pass was terrible.  Behind the receiver.  Ugh.  Rams know this is the end.  They drive right up field and score a TD.  27-10.  4th quarter, at St. Louis.  Game over.  Shit.  Kev and I grab some hot chocolate for the 4 of us and I grab 2 \"hot\" pretzels for me & Kim.  They were hot when I got them, but were cold by the time we got to the seats.  Pre-game lineups are pretty nice.  After every 49er\'s name is announced, the crowd collectively adds \"sucks\" to the name.  (i.e.-\"at cornerback, #31, Mike Rumph\"...\"sucks\"  \"safety, #33 Tony Parish\"...\"sucks\")  Funny.  Game starts with a deaf child singing the National Anthem(?).  Is that possible?  So that\'s what it seemed.  Pretty incredible.  49ers lose toss.  Eagles receive.  First play, McNabb looks down field and is picked off.  Wow, not the best way to start a game.  49ers drive, but the holder muffs the snap on a FG attempt.  Game remains 0-0.  I won\'t continue with the drive-by-drive coverage because it\'s boring.  Just know that Kevan Barlow could not be stopped and Brian Westbrook is awesome(returned a punt for a TD).  49ers scored a go-ahead TD with 1:58 remaining.  Eagles drive and Todd Pinkston makes a nice catch in the endzone.  Extra point good. 28-28.  On the kickoff, 49ers get caught at their own 10.  Eagles hold them on runs and stop clock, but with :01 left, the 49ers call timeout.  It\'s 4th down, and Garcia takes a knee.  End of regulation.  OT toss-49ers lose and Eagles get kickoff.  Crowd erupts.  Decent return.  First play of OT-McNabb intercepted.  A collective \"Fuck\" arises from our section.  Game over.  Niners kick a field goal to win 31-28 in OT.    Past is covered.  Now for the future...The 26th would be a nice night to get together.  I have to work until 5, but anytime after that would be fine.  If anyone has any fancy ideas, feel free to share, otherwise, we\'ll spend the night deciding where to go.  Hope to see everyone.  And in case I don\'t talk to you beforehand, enjoy Christmas!  I registered for my classes today.  I have class on Monday and Tuesday nights from 5:30-8.  I don\'t know what to do about work.  I\'ll be working Wed-Fri 9-5, but I would like to have off either Monday or Tuesday, and I don\'t want to work Sundays.  I have to figure it out so I can get 32 hours to keep benefits and stuff.   '),(18322,1826527,'20,December,2003','2003-12-20 00:00:00','    Horse.  Hi people.  It\'s good to hear all the holiday news from everyone.  I will be in NJ from Monday to Sunday (22-28).  I agree that we should get together indeed.  Umm....  I\'ll talk to you all in person.  I gotta go.  -BLOG OUT!!!!!!!!!!!!!!!!!!   '),(18323,1826527,'19,December,2003','2003-12-19 00:00:00','    Hello again, I\'m heading home Wednesday afternoon/evening and it can\'t come soon enough in my opinion. Work has been going well, but I got my first bit of criticism today and of course, it has been irritating me more than it should. I was told that my patient treatments are excellent, but I need to improve my ability to manage a full case load. Never mind the fact that my supervisor has been taking all sorts of days off, and somehow, the patients all seem to get treated. Anyway, enough of that. I would really like to get together with everyone. I\'ll be in NJ from Wednesday until Saturday, so feel free to give me a call at home, email me, or use this truly convenient blog and let me know what works for everyone else. Merry Christmas and Happy Holidays to all.   '),(18324,1826527,'18,December,2003','2003-12-18 00:00:00','    Well, Wednesday was my last day of work for the rest of the year.  My company gives all employees the week between Xmas and New Years off, plus I had 5 vacation days that I didnt use.  So, I am on vacation starting today, Dec. 18th, and I dont go back to work until Jan. 2nd.  I am coming back to Jersey for Xmas, but I dont know how long I am staying.  The best nite for me to go out with you all would be the 26th.  But, my plans can change, depending on everyone else\'s availability.  Have a bloggy Xmas if I dont talk to any of you before then...   '),(18325,1826527,'17,December,2003','2003-12-17 00:00:00','    Hello all...  Today is my last day of finals and I\'m stoked.  No more waking up at the ass crack of dawn.  No more studying.  No more dumb jocks and frat boys.  A whole month of freedom.  Scratch that last one.  I\'m working full time at the bank during the bank which isn\'t so bad because I get to make some extra cash.  Hopefully the extra cash will help me pay off my credit spending from this lovely Christmas season.  Anyways, I saw 2 of my grades for this semester and they were good, but I plan on revealing my grades to you all in one shot.  I doubt anyone would enjoy a day by day reveal of my grades.  Does anyone care about my grades anyway?  You doods better be coming home for Christmas!  We should go back to the Town Pub to get some grog and make fun of people because we really didn\'t do much of that last time.    Gonna go do some last minute studying.  See ya!!  P.S.-  In the little mini-mart on campus, I saw a box Big League Chew.   '),(18326,1826527,'17,December,2003','2003-12-17 00:00:00','    Greetings and salutations.  I realize Christmas is a mere 8 days away, and still don\'t know if I have to work on Christmas day or not.  I guess I will be a little bummed, but I don\'t think it will be all that bad, as long as I\'m not working all day.  I\'m also still waiting for an \"official\" letter from  urlLink Montclair State  about my acceptance.  The semester starts in a month, so I hope to get things done asap...Yesterday at work, I had a  Clerks  moment.  Some old lady asked me where a certain candy was, and like the smart-ass I can be when it\'s 5 minutes before I leave, I say \"probably over by the rest of the candy.\"  Then I left.  Oh well.  Kim and I are going to the  urlLink Eagles -49ers game on Sunday with the brother and sister-in-law.  I haven\'t been to a pro football game in quite some time.  I think I went with my brother a few years ago to an Eagle game, but I\'m not sure.  Otherwise, it was many years ago when Boomer was on his last stint with the Bengals at the Vet.  My brother\'s been hooking me up with some sports festivities lately.  I like it a lot.  Maybe next I\'ll be able to see Monster\'s Inc on Ice or something.  Hooray for Mike doing a triple axle.  I have a plan for the future.  When I talked to the comm professor at Montclair, he said NJ doesn\'t really have a good school to pursue my doctorate in comm.  I\'ve always wanted to go west(i.e. Arizona or even Nevada), so I may just move out there while finishing my education just to get away for a while.  I did a little research, and found that Arizona State has a vast variety of programs that suit my interests.  I intend on looking for UNLV, Nevada, and Arizona as well.  I think it will be an awesome experience, plus I won\'t be breathing in pollution as much as I am here.  The only thing would be that moving out there would require me to have a place to live.  Lots more research to come...  Is anyone coming back to NJ for the holidays?  Kindly share your ideas.  Sorry I don\'t AIM anymore.  I\'m not home during the day, and usually do stuff at nite.  This blog is already staying past its welcome.  Good day!   '),(18327,1826527,'14,December,2003','2003-12-14 00:00:00','    Hey Bri, your Christmas wish of no snow until February has been flushed down the toilet...  Once again, we are snowed in here in NJ and it is no fun.  I was supposed to get breakfast and go Christmas shopping with my mom today, but I dont think I want to drive in this weather.  Me and Brian were watching the weather channel last night and they something about the wind gusts being so high that they might actually turn your car over if you drive over bridges and all that good stuff.  So I think it would be best for me to keep my ass in the house.  Besides today is the perfect opportunity for me to get some studying done for my finals.  Then again, it would also be a good day to catch up on some sleep.  I shall contribute to the holiday spirit by posting my own Christmas wish list:  *I also want Rachel\'s mom to be 100% better.  It is a shame that such a sweet woman has to go through all the crap she does. *$10,000. That\'s all I need to pay off my car and other miscellaneous debt. *an Aston Martin Vanquish sitting next to Brian\'s Jeep in our imaginary driveway in our imaginary house in Arizona *Since, I intend to move where it never snows, I wish to never go anywhere snows.  **Exception, I would love to go on a cruise to Alaska. *A very large collection of my favorite sneaker - Pumas.  I already have 4 pairs.  However, I think 10 more pairs, all in various colors, should do the trick. *I also need a CD collection overhaul, but in my case I just need more different CD\'s. *A trip to Italy.  I\'m learning to speak Italian now and I\'m fascinated with the language.  I would love to mingle with the Italians and eat lots of pasta and pizza.   *Ted Alexandro and Brian Regan DVD\'s.  Two very funny doods.  Everyone should catch one of their shows. *A teaching job for Brian. *For all of friends and family to be happy.  That is my list.  Now get to work!!! Just kidding!   '),(18328,1826527,'13,December,2003','2003-12-13 00:00:00','    With the holidays approaching rapidly, I figured I needed to bring some Christmas spirit to this blog.  So I\'m gonna post my Christmas list.  Here\'s what I want for Christmas:  *Rachel\'s mom to be fully cured for good.  No more worries.  These are the things that would be nice as a bonus, but don\'t come close in comparison to priority #1:  *a Bengals playoff berth *a new Jeep sitting in my imaginary driveway in my imaginary house in Arizona *stop hearing about Paris Hilton *no more snow until maybe February *an overhaul of my music collection(I have too many cds I don\'t ever listen to) *to pay off my credit card(actually I\'m almost there) *Alias Season 2 DVD *a winter coat *spend time with Kim, my family, and friends, and not have to worry about crap for the time that I\'m with them.  So that\'s it.  Feel free to share your lists amigos.   '),(18329,1826527,'11,December,2003','2003-12-11 00:00:00','    Forgive me father, for I have sinned.  It\'s been six days since my last blog.  So friggin\' sue me.  Not too much has occurred in my life this stretch of a hiatus from the blog.  The weekend was rather uneventful because I was snowed in on Saturday.  Sunday, I was able to see Kim at least, and escape my house for a few hours.  I couldn\'t tell you what I did on Monday other than work since I can\'t remember.  Tuesday I saw Kim again and we watched the previously mentioned  Alias .  Last night I went to see the  urlLink Rangers  game with my brother.  Enter new paragraph.  Despite being a fan for my life, I had never been to the garden for a Ranger game.  Sad, I know.  Anyway, my brother, who is the best brother in the universe, got tickets from a co-worker to see them against Montreal.  My brother had to work last night, as did I.  I worked until 5, and his last report was at 6:45.  The game started at 7.  The positive was that he works in Rutherford, and I in Carlstadt.  The drive was not taxing to say the least.  He decided it would be best if we drove into the city since we were close to route 3, and we figured if we took the PATH, we\'d probably get there even later.  His friend gave him a shortcut thru Hoboken to bypass the winding road leading to the tunnel, which probably knocked a good 15 minutes off our wait, but we still had to go thru the tunnel.  The tunnel was a very slow crawl.  We ended up missing the entire first period, with the Canadiens leading 1-0, but listening on the radio en route to the arena.  When we got into Manhattan, there was surprisingly a crapload of traffic and stupid people crossing on red lights, etc.  We parked not far away, and walked to the garden. (Side note:  the only Rangers jersey I have is an old one of Niklas Sundstrom.  Ironically, he\'s on the Canadiens now)  Upon entering, we walk towards tower \"D\" which is the gate listed on our tickets.  Then some guy tells us everyone has to enter thru tower \"B.\"  Unfortunately, we circumnavigate the arena, passing long-ass lines at the snacky stands before arriving at our section, 72.  Usher guy tells us to go around to the other side to our seats, and when we do, we find people sitting in them.  We proceed back to where we first came in and ignore the usher and sit in close by but empty seats.  We sit down and Chris Simon and Darren  Langdon begin beating the bloody hell out of each other.  Nice way to start the second period.  A few minutes later, people show up who have tix for the seats we\'re in.  We go back to our real seats and make the people there move.  Now we\'re stuck in the middle of the row.  Neither of us had eaten dinner and we don\'t feel like going thru ten people to get to the stairs or waiting in line for food.  Chris Simon scores the tying goal in the second, and the Rangers fall apart in the third to lose 2-1.  After the game, we drive back into Weehawken and hit up Houlihan\'s for some post-game dinner.  Good nachos and a chicken sandwich do me well, and then it\'s off to pick up my car and drive home.  End night.  I really had an awesome time.  The atmosphere at the garden is filled with energy.  It gave me hope that I can get to more games from now on.  May blog this weekend.  If not, enjoy the snow on Sunday again.  Godammit!   '),(18330,1826527,'10,December,2003','2003-12-10 00:00:00','    Finals are here and I haven\'t really done any studying.  Studying isn\'t my bag baby yeah.  Sorry. Bad Austin Powers reference.  I have a religion final on thursday, business law on friday, italian on tuesday, and accounting and calc on thursday.  The only class I am studying for is Italian.  So my week shouldn\'t be so stressful.  However, I had/have to wake up at 5:30 to go to class tuesday thru friday.  It\'s a good thing I\'m not working saturday.  Maybe I can get some decent sleep then.  I still haven\'t done any Christmas shopping.  I only picked up 2 things so far and one of them is on backorder.  I hope it gets here before Christmas.  In other news, I am now addicted to Alias.  I never really wanted to watch the show because I didn\'t see what was so great about Jennifer Garner.  Ok she\'s pretty, I\'ll give her that, but she\'s not to die for gorgeous.  Anyway, Brian bought the first season on dvd and we can\'t stop watching it and now we\'re tuning in to watch the new episodes on sunday night.  I can\'t wait to see the second season and the beginning of the third season.  Damn you Brian for getting me addicted to Alias!!!   '),(18331,1826527,'06,December,2003','2003-12-06 00:00:00','    Guess what\'s more boring than Connecticut?  ******************************************  Connecticut buried in a crap load of snow...   '),(18332,1826527,'05,December,2003','2003-12-05 00:00:00','    Pizza nite!  And a blizzard ensues.  This crazy weather.  It\'s probably snowed about 5 or 6 inches I\'ll say, and it\'s supposed to continue through tomorrow night.  The good thing about it was that I went to work early with Kim\'s mom, and was able to leave early, at 3:30.  It had started to accumulate, but not as bad as if I left my normal time(5 or 5:30).  I woke up this morning at 6, because I had to go in early, and turned on the tv.  I noticed that route 3 was shut down at Passaic Avenue, just the place I had to drive to.  So I find out that someone was speeding at 3 am and hit a utility pole.  The road was closed until about 9 I think.  Traffic was diverted thru Clifton and Nutley for 6 hours.  That\'s a pretty poor job considering it\'s a major highway leading to the city.  Anyway, that\'s the story of the morning.    Since my next two days are shot, I\'m probably gonna shovel tonite and clean my room.  I helped my dad put up Christmas lights tonight already, but it\'s still only 8!  Man I\'m bored.  I\'m chatting with Kim and Liz right now, so I\'m gonna go.  Enjoy the snow.   '),(18333,1826527,'04,December,2003','2003-12-04 00:00:00','    Awesome!   I\'m very very happy to hear both sets of good news.   Congratulations all around.   '),(18334,1826527,'04,December,2003','2003-12-04 00:00:00','    If you already read Kim\'s blog from earlier today, or if you happen to read the blog backwards, then you know I have good news.  I met with the professor from  urlLink Montclair State , and he told me he\'s going to have the school write me an official letter of acceptance into grad school!  Dude, I can\'t even tell you how happy I am after hearing that.  He was really nice and helpful.  He explained the three paths in communication I can take, and recommended the first two classes to take.  He also said that I may have a good chance of gaining a teaching assistantship since I have some teaching background, and a communication background.  I am so psyched.  So that\'s the good news of the day.    My appointment was at 12:30, and I was done by 1, so I decided to do some Christmas shopping since I took the day off from work.  I finished shopping for Kim, mom, dad, my brother, my uncle and his girlfriend.  All that in an hour and 15 minutes.  I got to the mall at 1:30 and left at 2:45.  I beat traffic home as well.  It was a pretty sweet few hours.  Unfortunately, it\'s back to work tomorrow in a blizzard, or so I hear.  I\'m not ready for snow yet.  If it\'s going to be a big storm, then this is one of the earliest big snowfalls I can remember.  Usually they hit in Feb or March.  Anyway, that\'s winter for you.  I think if I feel adventurous, I may tinker with my  urlLink webpage  today.  Kim\'s at school, and I don\'t have anything else to do.   '),(18335,1826527,'01,December,2003','2003-12-01 00:00:00','    I had an appointment today with the chair of the math department and she was very nice.  We talked about what jobs I can do with a math degree and she helped me compare the business and finance majors.  She told me that a lot of jobs dont require math majors but they love the way math nerds think.  People employ math majors for their problem solving abilities, not necesarily the degree.  As far as the incident with my advisor is concerned, she said it came to her as a shock that he would say such a thing.  She said she talked to him about it and he supposedly feels bad that he upset me.  At this point I don\'t really care.  If he feels bad, fine.  If he doesn\'t, oh well.  To me it\'s over and there\'s no reason for me to dwell on a negative comment from someone who doesn\'t even know me.  She actually offred to arrange for me to take the calculus readiness test which my advisor never even told me existed.  I told her I would think about it but that I was ok with taking the precalculus class anyway.  All in all, it was a good meeting.  I walked away happy and relieved.  Anyhoo, we\'re supposed to be getting some sort of snow storm and I\'m not looking forward to it at all.  I don\'t mind snow when it\'s in small amounts and it makes everything look pretty.  But, when there\'s more than 3 inches of snow on the ground, I hate it!  Snow is the inconvenience of all inconveniences.    Gotta finish up some homework before heading back to class.  Peace out homeboys!  P.S.  Brian has some good news he might share with us later.   '),(18336,1826527,'01,December,2003','2003-12-01 00:00:00','    Welcome, December.  Man am I happy this year is almost over.  Hopefully it will mean I am thisclose to beginning grad school and even closer to leaving job x.  It\'s Monday, and that means the late shift at my crappy job.  Gotta be there at 2:30, so I still have some time.  I haven\'t mentioned this on here yet, mainly because I don\'t know what to do about it-my sister in law\'s father found a job for me.  It\'s in a private school in Brick as a computer teacher, which will allow me to work 3-4 days a week, with the option to pay into benefits.  The job is mine if I want it, since they really need someone.  My brother and sister-in-law have made their home available to me the days I would work there, so that\'s great and I wouldn\'t have to commute over an hour every morning.  Obviously I also want to start grad school, and I don\'t know the final decision because I have to interview hopefully this week at the school.  Anyway, I wanted to find out the decision before making a decision about the job, because it would be difficult to do both, unless I knew my schedule for work and classes.  I mention the job yesterday to one of my family members, who later in the day tells another family member, and the latter family member decides to attack me and basically question why I would even consider the job.  That really pissed me off because it has nothing to do with this person.  It\'s my choice to make for something I need to do-work in my field.  I still haven\'t made up my mind about it and say this, but I\'m still questioned because it \"doesn\'t make sense.\"  I guess I know why I wonder about my confidence so much.  People shoot down my outlooks by overemphasizing the negatives and not even seeing the positives for me.  Anyway, fuck that shit.  Bengals won another awesome game yesterday, shoving it right up Pittburgh\'s ass.  Next week is the battle for the AFC North.  On a related note, Kim asked me if I thought the Bengals could get to the Super Bowl yesterday.  After initially smiling, I realized that lately in sports, anything is possible.  Remember the year the Pats won the SB, Rob?  Who was in the SB last year?  Oh right, they\'re both awful this year?  Cuse won the NCAA, the Marlins won the series, the Spurs won the NBA.  Stranger things have happened, but I\'m not going to get ahead of myself.  If they made the playoffs, that would be simply amazing in itself.   '),(18337,1826527,'30,January,2004','2004-01-30 00:00:00','    It\'s time once again for my random blog. I\'m still in Middle of Nowhere, Georgia and it\'s going pretty well. I switched over to an Outpatient Clinic and so far, so good. It\'s interesting to see such a wide variety of patients (pretty much all ages and various conditions), but it\'s also been tough just coming from a setting where I would see one type of patient all the time. It may take some time for my brain to adjust. Anywho, I\'m glad to hear that Brian\'s accident has not left any residual physical damage, although it does pain me to hear he went to a chiropractor. In other random news, I was happy to see Bill Murray take the Golden Globe and get an Oscar nomination, but Johnny Depp was nominated in the same category for Pirates of the Carribean, what the hell?! I guess Disney paid some people off.  I\'m off to Florida this weekend to see dear ol\' Dad. He had some surgery a couple weeks ago, but has come through like a champ. Enjoy the game on Sunday, even if CBS did ban that great anti-Bush commercial. Back into the shadows of Warm Springs I go, Matt   '),(18338,1826527,'28,January,2004','2004-01-28 00:00:00','    I decided to play hooky.  I woke up at 7 because I was supposed to be at work for 8:30.  I looked out the window and saw that the snow wasn\'t what I expected.  I turned on the tele to find we\'ve received somewhere between 7 and 10 inches of snow.  I saw school closings scroll across the screen and noted colleges were closed, as were several government buildings.  I decided that I wasn\'t going in.  I wasn\'t going to risk driving in icy conditions to a job where I make next to nothing, especially since I was just in an accident.  So my morning consisted of digging out 3 cars from mounds of snow, some soft, and some quite heavy due to the plows kicking it all up against the sides of the vehicles.  I came in and ate some  Fruity Pebbles  and watched some Sportscenter.  I decided to play with my  urlLink website , and updated my  urlLink photos  section by adding some links with the pics.  There are also some new shows listed for the NY/NJ area.  No major format changes due to the fact that I can\'t find the installation disk for my scanner, so it just sits there and never gets any action.    I just got my mail and received my much anticipated  urlLink No Motiv  cd.   Daylight Breaking  is a damn good cd.  It\'s a bit heavier than their last album, but the ep that came out a few months ago didn\'t let me down in terms of holding me over until this came out.  They also had a song that was featured in the X-games, which just ended, and they played out there at the games, too.  I\'ll probably be back outside shoveling a bit more before doing more homework today.   Class was cancelled last night, but that didn\'t prevent him from piling on 6 more journals to read for next tuesday.  I haven\'t looked at all of them, but the last one is 35 pages alone.  Bloody hell.   '),(18339,1826527,'26,January,2004','2004-01-26 00:00:00','    First off, I want to thank Rob for calling me.  I got your message, but will try to call you tomorrow during the day when I am not half asleep and while I\'m home.  Secondly, I will also try to call Den, since I owe him a call from about 2 weeks ago.  And fyi, I am doing ok in terms of recovery from the accident.  My only ailments now are going to be my funds due to having to purchase a new vehicle and increasing auto insurance.  I\'m still on a 3x a week schedule at the good ol\' chiropractor to heal my whiplash and soreness in my neck.  Other than that, I think everything is ok, and hopefully will stay that way...  I had my first meeting of my other class tonite.  Borrrrrrrrrrrrring.  We had a 4 minute break, but he kept us for 5 minutes after the scheduled meeting time.  He seems like a nice guy, but he lectured at us the entire night.  No interaction as far as discussions go.  There are only 5 people in the class too.  We don\'t meet next week because he\'s going to Michigan to apply for a position over there.  So I guess I won\'t be working with him next semester.  In other news, I\'m unwillingly embracing the next storm heading our way.  The only reason being that I don\'t want to drive my dad\'s car in the snow/ice to school for one class.  It\'s not worth it, especially since my driving wreck a week ago...My job is fucking horrible.  My boss is annoying and the customers are obviously annoying, but the routine is unbearable.  I graduated school with a degree in communication and certification to teach grades K-8.  The only thing remotely concerned with either of those is talking at work.  I don\'t listen to most of my coworkers because they are ignorant(except for a few, one being Kim\'s mother, of course.)  What made it worse was that last Tuesday, I talked with the 7 other people in my class, and they all have at least a decent job, and it has something to do with comm.  The only thing I have applied for is a job with the nba, but I have no PR experience, so the odds of me actually being hired are slim.  I\'m gonna see if my classmates can get me an in or something or give me advice.    Due to the whole accident, our vacation plans are probably nixed.  I need to buy a new vehicle, and I mean new, because I don\'t want a car that will simply hold me over to my next one.  I\'m just going to invest in something I like and work simply to pay for it and insurance.  I\'m aiming for a Jeep Wrangler, and hope to bring one home within the next two weeks.   '),(18340,1826527,'23,January,2004','2004-01-23 00:00:00','    It\'s been a while since I\'ve blogged so I figured I should check in...  Nothing new has been going on.  Classes started again :( but thats ok because I was tired of being at the bank all the time...  I bought a new car!!!  That\'s right folks.  I traded in my prized 98 Altima for a brand spanking new Nissan Sentra.  I test drove the SE-R which is the super sporty model but it didn\'t have ABS so I said \"Funk Dat!  Give me ABS biznatch!\"  I really didn\'t say that, but I got my ABS along with power locks, power windows, keyless entry, power sunroof, and killer speakers to play my music even louder that I already do.  Me so happy!!!  I also got cheaper insurance which is awesome.  I was paying 4G\'s a year for the other car and now I\'m saving $2500 with a new company.  Can you believe that?!?  I got a brand new car and my insurance cut in half.  Although I love my new car I felt really bad about getting it since Brian got into the accident and totalled his car just 2 days before I went to the dealer.  However, I\'m glad he\'s not driving because I get to show off my car.  Just kidding.    I am looking forward to Brian\'s list of exciting dates.  Especially.... you guessed it... HEY MERCEDES!!!!  Actually I\'m more excited about the comedy shows.  Anyhoo, I hope everyone is doing well.  Peace out yo!!   '),(18341,1826527,'23,January,2004','2004-01-23 00:00:00','    Some dates to look forward to, in no particular order, except for chronological order:  Tuesday, January 27th   No Motiv \'s new CD,  Daylight Breaking  comes out Sunday, February 15th   Juliana Theory  @ Maxwell\'s  8pm Friday, February 20th   Brian Regan  @ Banana\'s Saturday, February 21st   Hey Mercedes  @ Mercury Lounge Friday, February 27th   Ted Alexandro  @ Stress Factory Tuesday, March 2nd   Living End  new CD, Modern Artillery comes out Tuesday, March 23rd   Weezer \'s DVD comes out, as well as the Blue Album Deluxe Edition   '),(18342,1826527,'19,January,2004','2004-01-19 00:00:00','    Just when I thought my luck was even-ing out, guess what happens?  Hey, car accident.  Car totalled.  Laceration to left knee and bruise to forehead.  Whiplash and ligament damage in neck.  Great times.  I now have no car and have missed work for two days.  Luckily, I have off from work tomorrow, and Kim can drive me to my first class at night.  This sucks.  I\'m guessing I may be at fault since I was turning and the other car was going strait, but I wasn\'t the one going 60 down the street.  Fucking legal shit.  Wonderful timing, considering my health insurance hasn\'t kicked in yet, and I was in the process of shopping for auto insurance and looking for a new car down the road.  I also don\'t have an enormous amount of money to fund a new car and insurance increases.  In addition to this incident on Saturday, my brother and sister-in-law\'s kitchen nearly burned down on Friday nite, my mom slightly dislocated her shoulder, and my mom/dad\'s car needed a new axle on their car.  Oh yeah, and my brother was at the Eagles game last nite.  So, I infer from Meredith\'s journal that it was Kev\'s birthday, so Happy Birthday, Kev.  Sorry I missed the date...No offense to Rob, but this Super Bowl is bizarre.  Panthers/Pats?  Huh?  Well, it can\'t be worse than Tampa/Oakland.  Ewww.  That\'s all for now.  Misery for a while to come, only to increase with the stress of school and work.  Happy fucking 2004.   '),(18343,1826527,'14,January,2004','2004-01-14 00:00:00','    Listening to my newly-received  urlLink Dandy Warhols  cd.  It\'s an awesome listening experience.  It blends 80s sounds(i.e. synths, drum machines) with a club groove.  I guess you have to hear it to know what I mean.  Anyway, I finally received some letters from  urlLink Montclair , although not the one I was waiting for.  I of course viewed my financial aid status and look to be receiving money as a full-time student.  I had to email them to notify them of my part-time status(6 credits).  I also had to fill out some interview online to qualify for my loans.  I\'m finished, but now have to wait for my \"adjusted\" loan info.  Whatever.  I only know that class starts Tuesday.    I heard from one of my professors via email as well.  He told us about the class a bit and explained we wouldn\'t have to buy any books!  Gotta live the technology we deal with in school now.  On a related note, Kim accompanied me to the MSU bookstore to purchase the book for my other class, which ran me $45.  Not bad for textbook prices, but it wasn\'t very thick.  Anyhow, that\'s the school news for now.  More to come when I receive more info.    Other news???  Hmm, I got my cds in the mail today if you haven\'t taken note by my first few lines.  I only listened to Caviar before, and enjoyed it thoroughly.  It made me miss  Fig Dish  unfortunately.   '),(18344,1826527,'12,January,2004','2004-01-12 00:00:00','    I have to agree with Rob.  That\'s pretty sad.  He was only 32.  Moment of silence...................................................   '),(18345,1826527,'11,January,2004','2004-01-11 00:00:00','    Yinka Dare!!!!!!!!!!!!!!!!!!!!  Noooooooooooooooooooooooooooooooooooooooooo!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!   '),(18346,1826527,'10,January,2004','2004-01-10 00:00:00','    I\'m bored.  Just waiting to go to my orientation at Montclair.  I\'ve finally received some recognition of my lack of acceptance letter.  I got an email from someone who says I\'ll receive a letter this week.  He also acknowledged the fact that a letter was sent on Dec 17th and doesn\'t know why I never got it.  Anyway, I hope it doesn\'t drag on too long.  It\'s supposed to run from 1-3:30, but I think time was included to buy textbooks, which I\'m not doing because I\'m pretty sure everyone else will be and I don\'t want to wait.  Right now, the weather bug says its 7 out.  Should be a fun walk to the student center.  Anyway, I also checked my account status online, since I haven\'t gotten a bill yet or loan info yet, but under the list of charges it has some loans listed, so I\'m hoping that\'s real, and my semester(and full year) is paid for!  Now, Disney World may just be possible.  Shweet...Strange dates, but is anyone gonna be home Feb 20th?  Brian Regan is appearing around here and Kim and I are probably going...   '),(18347,1826527,'07,January,2004','2004-01-07 00:00:00','    This entry may take a long time for two reasons:  1)there\'s a lot of useless information & 2)I have a horrible stomach ache, which I\'ve had pretty much all nite.  Anyway, if anyone may remember, I created a team in NFL 2004 called the NJ Vandals.  There was this whole expansion draft where I drafted crappy players, etc.  Anyway, I finally finished my season and did it in a good way.  I won the friggin\' super bowl!  You probably don\'t care, but I have always been a fan of good video games, and this game in particular was actually enjoyable, even though I almost lost a few times.  Anyway, I\'m going to give you the summary of the big game and some stats because I\'ve always been a stat person(hey, it\'s one of my jobs).    Philly(13-3) vs. NJ(9-7) in the Super Bowl.  1st Quarter-I win toss and receive ball.  I get a couple of nice runs and pass plays going and dig into Philly territory.  My QB(Jeff Blake) throws an int at the 20.  Oh no!  Bad way to swing momentum.  Eagles get ball and drive right downfield, capped off by a Duce Staley 33 yd TD run(7-0 Phi).    2nd Quarter-I had gotten the ball back after the Eagles drive and began driving before the end of the 1st.  I\'m running well and passing well again, and find Jake Reed for a 17 yd strike(7-7).  Eagles get ball back and go right thru my defense.  Thrash with a 4 yd rec(14-7 Phi).  I get ball back with 2:00 left.  Another good drive with variety and punch it in with a Dee Brown 8 yd run(14-14).  3rd Quarter-Eagles drive again right thru my defense, but I recovered a fumble and controlled much of the clock in the 3rd, before scoring on a Brian Westbrook 1 yd run(21-14 NJ).    4th Quarter-Unfortunately, after scoring in the 3rd, I had to kickoff.  Eagles drive thru me again and score with lots of time remaining.(roughly 3 minutes).  With the game tied, I try to be conservative at first, but suddenly could not run.  I tried to pass and they sacked me.  I go for it on 4th because they would probably score on me anyway.  Pass dropped.  Turnover.  Back to Philly.  They fumble.  I get the ball and can\'t move again.  Eagles get ball on downs and kneel with :02 left in regulation, tied at 21.   OT-Eagles win toss and receive.  I think it\'s over.  They drive as usual thru my defense and get to about my 35.  On 3rd down, they pass, but I pick it off!  Woo!  I feel the momentum and begin running them down.  I punch it in with Westbrook from the 6 and win 27-21 in OT!  Damn good game for a video game.  Blake 14-28  223 yds 1TD/1INT Dee Brown 11 rush 79 yds TD Lamont Warren 11 rush 68 yds Westbrook 7 rush 26 yds 2TD Freddie Mitchell 7 rec 114 yds Reed 3 rec 52 yds TD Pete Mitchell 3 rec 51 yds   So I had an enjoyable experience with an expansion team.  Pretty realistic because the players I got were woeful.  I traded for a QB early in the season, and created a safety because I needed someone who could tackle one on one.  I ended up winning my last 4 games to win my division, the only way I could have made the playoffs.  Despite lots of swearing and anxiety, I was glad it was challenging, and surprised I was able to find my defense in the playoffs.  Here\'s some of my final stats:  Blake 159-277   2,482 yds  18TD/8INT  96.9 Rating Warren 143 rush 1,014 yds  14TD Westbrook 159 rush 984 yds 13 TD Brown 47 rush 265 yds 7TD F. Mitchell 91 rec 1,430 yds 12TD Reed 33 rec 657 yds 6TD P. Mitchell 33 rec 432 yds 3 TD Houshmandzadeh 20 rec 327 yds Damon Moore 84 tckls 3 INT Lenoy Jones 42 tckls 3 sacks INT Darrien Gordon 36 tckls 3 INT  Sorry if you read this, but I forewarned you about the length and subject.  Anyway, my stomach still hurts in waves.  I\'m gonna go to bed soon and hopefully feel better.  PS-Kev I tried your link and it didn\'t work for me.  Den I owe you a phone call.  Sorry I missed your call the other day.   '),(18348,1826527,'05,January,2004','2004-01-05 00:00:00','    I am not making any resolutions this year, or ever again for that matter.  They are destined to fail.  I am going to make a conscious effort to get myself into better shape (eventhough round, my current condition, is a shape...).  I also want to get back to school, but I am doubting that this will happen this semester.  Oh well, enough out of me...  Bri, you can get your cd here:  urlLink http://interpunk.com/item.cfm?Item=97090&   Addendum:  Here is a link to some really good IQ tests, they are rather challenging:   urlLink http://www.highiqsociety.org/flash/nonmembers/iqtests.htm   '),(18349,1826527,'04,January,2004','2004-01-04 00:00:00','    Hey.  Yup a new year.  Anyway, I went shopping online tonight for some cds, since I realized I\'m sick of my cds that I have.  I went to amazon, buy and fye, but wasn\'t able to find much.  I am still unable to find  Mare Vitalis  by my boys in the Appleseed Cast, but found a few things I liked.  Here\'s my order:  *Jealous Sound~Kill Them With Kindness *Joe Strummer and the Mescaleros~Streetcore *Caviar *Dandy Warhols~Welcome to the Monkey House *Dane Cook DVD/CD  Explanation-Jealous sound includes the singer from Knapsack, another band who I liked but broke up.  Joe Strummer was the guy from the Clash who passed away last year.  Caviar is half of the great band Fig Dish.  Remember them, Matt?  Dane Cook is a funny guy.  That is all.  I almost got A Tribe Called Quest\'s Anthology, but changed my mind at the last minute to save a few bucks.  I also agree with Kim and want to start eating better again.  I was doing well, but started eating a bunch of snacky cakes and other crap the last few months.  I will eat more, but more of the real foods and less of the crap.  I also want to start working out more regularly.  Work has massively interfered with my bowflexin\', and it\'s pissing me off because I\'m not getting much exercise on a regular basis.   I registered for an orientation at Montclair next Saturday.  I still haven\'t received my official acceptance letter, so I have to call them again tomorrow before work to find out what the hell they\'re doing.    We decided(well, actually I just kind of said we\'re going) to go to Disney World.  Yeah it\'s a lot of money, but it\'s going to be money well spent.  I have a job and a credit card to put the hotel and car reservation on until I can pay it off.  Word.  Once again, I\'ve been stupid in predicting sports.  I picked Denver.  What the hell was I thinking?  Denver?  Well I got the other 3 right, thanks to McNair playing for Tenn.  It was funny watching the Cowboys get dismantled as well.  Go Jake Delhomme.   '),(18350,1826527,'04,January,2004','2004-01-04 00:00:00','    I know this is a little late, but HAPPY NEW YEAR!!  Big deal right?  Like most people, I made some new years resolutions and they are as follows:  1.  Stay away from red meat.  This doesn\'t mean I\'m never going to eat it again, I\'m just going to avoid it as best as I can.   2.  No fried foods.  I was doing pretty good with this last year so let\'s see if I can keep it up. 3.  When I\'m mad at someone, I usually take it out on an innocent person.  For example, I\'m mad at my brother (which happens often) and I take it out on Brian.  Is this fair? No, it is not.  So I hereby intend to yell at my brother more.. err.. express to people that they piss me off. 4.  Exercise more.  This is a very popular resolution that many make and, of course, break.  I hope to start running again once it gets warmer, but until then I\'ll do stairmaster.  However, I recognize now that I may never do either so I really won\'t be breaking this resolution.  5.  No more road rage.  When I drive, I have tourettes (sp?).  Therefore, I intend to avoid using my potty mouth when driving.  6.  Save more money.  This isn\'t a very easy thing to do, but I want to put some money away every week, even if its only $5. 7.  Continue to  not  eat after 8pm.  However, there are exceptions to the rule.  If I\'m out with friends and they want to eat after the previously mentioned time, it is acceptable as long as its not an everyday thing (which it isn\'t). 8.  Stay away from ice cream, especially my favorite - vanilla with peanut butter syrup.  I shall limit my ice cream consumption to once a month.  No more and maybe less.  I think this is enough.  If anyone catches me breaking any of these resolutions, smack me.  Please feel free to share your resolutions on this wonderful thing we call a blogger.   '),(18351,1826527,'24,February,2004','2004-02-24 00:00:00','    Greetings.  Today I had a great feeling of accomplishment.  Nothing major happened, but last night I made a list of a shitload of things I had to get done today, since I was home during the day before going to class at night.  I got all of them done, well, I was supposed to read 4 articles for class tonight.  That didn\'t exactly happen.  I skimmed 3 of them very briefly, and one was 20 pages, so I didn\'t have time to read it, but everything else was completed.  I started by getting the registration and license plates for my new car from the dealer.  On the way home, I registered my car at Monkey State so I can park there without paying the parking garage.  I then proceeded to deposit my two checks from work into the bank and mailed my first car payment.  When I got home, I registered for a new ez pass tag, filed my state and federal taxes, filed my renewal fafsa, and finished my paper which was due tonite.  I was satisfied.  It\'s been a while since I felt like I accomplished a lot.  Anyway, that feeling will subside tomorrow when I realize I have to do a 3 page paper for next Tuesday, along with 4 more readings, read a chapter for my Monday class, and begin my research project for my Monday class as well.  But I\'ll sleep well tonight.  My brother\'s birthday is next week, and as usual, I have no idea what to get him.  There\'s only so many times I can get him an Eagles or Sixers jersey.  I guess I can get him a Mike Cameron jersey.(that was sarcasm, since he\'s a Mets fan, and that\'s their big free-agent signing I guess).    I\'m now going to rant a bit...I apologize in advance if you enjoy coffee, but it\'s something that bothers me to an extent...Why is it that coffee is acceptable to drink in the workplace?  In my case, it\'s severe, and my manager(s) are never seen without a coffee in hand.  I don\'t like the smell or taste of it, and I really am not a fan of folks who always smell like it, either.  In addition, I really get peeved at the people who are shocked and put me down when I say I don\'t like coffee.  You don\'t like coffee?  Oh my!  Isn\'t it a case of a mood-altering drug?  I know caffeine is not a controlled substance, but it is obviously addictive.  People use it as a staple in their diets in order to keep things going during the day.  I guess my gripe really lies in the reason that this drink was selected to be  the     drink that\'s allowable in the workplace.  I don\'t know, I guess one day I want to go to my awful job and see a big ass jug of Hawaiian Punch sitting in the break room with two workers eating cookies and sipping their cups of the red juice while giving a hateful glare in the direction of the lowly employee with brown teeth brewing a pot of taster\'s choice.   '),(18352,1826527,'18,February,2004','2004-02-18 00:00:00','    Forgive me father for I have sinned.  It\'s been months since my last blog.  Haha.  I always wanted to say that!  Anyway since I last communicated with everyone, I\'ve been to Italy for vacation over winter break, which was why I missed the Outback dinner.  Italy was great, Khristine and I went to Rome, Venice, Florence, Verona, Pisa, And the area of Tuscany.  We spent New Year\'s Eve in St Mark\'s square in Venice, which was really cool.  Food was great and so was the wine!  Anyway, so, I\'m back at school.  This is basically my last semester of classes (Thank God!!).  But I have two pharmacology classes which meets every day and consume much of my time.  I did however manage to go to the Duke-UNC game which was an experience as Matt can probably attest to.  This past Valentines Day, Khris and went out to a restaurant in Raleigh, The Angus Barn which was really good.  The rest of the weekend I spent studying for my exam on Monday (the professors really like to fuck us over in terms of scheduling exams!) on the entire autonomic nervous system and every fucking drug invented to alter it.  (I just needed to vent over the last exam).  In other good news, I never thought I\'d be saying, but A-Rod is a Yankee!!!  I love how all the people in Boston (aka the losers of America) are already conceding the division to the Yankees.  Can\'t wait till the season starts!  In other good news, I submitted my abstract to our next conference in Vancouver, and it was accepted as a talk, which is very good, but scary since I will be talking to over 500 people in the audience.  The most people I\'ve given a talk to is around 20, but my advisor said its good for my CV, so I\'ll see if I can manage it  Bri-good luck with the new wheels!  I might steal your rims when I get back home, which will be on March 26-29 (I got tix for the NCAA Eastern Regionals-which is looking more and more like St Joe\'s instead of Duke as the #1 seed)  Anyway, I\'ll be more frequent in my blogs  Talk to you later   '),(18353,1826527,'17,February,2004','2004-02-17 00:00:00','    I got a few things accomplished today.  First and most importantly, I received insurance for my automobile.  I don\'t want to pay for it, but I need it unfortunately, and now the car doesn\'t have to sit at the dealer lot while my insurance is processed!  On my way home from Randolph, I conveniently stopped at my old stomping grounds at Willy P.  I needed to pick up a copy of my immunization records because Montclair insists on sending me an abundance of emails telling me about my registration hold because of my missing measles booster shot.  Now I have the evidence and they can shove it up their ass.    This weekend was rather anti-climactic after anticipating a nice one.  On Saturday, Kim and I went to a lovely place for dinner, and upon completing the dinner, I realized something wasn\'t right.  I mentioned we should probably leave because I felt an uneasiness in my stomach.  As soon as we made the U-turn to come home from Pine Brook, which is up on 46 West, past Fairfield, I came to the conclusion that I might not make it home if I don\'t find a restroom.  So I drive rapidly towards home, only to find(as usual) much traffic by Willowbrook Mall.  I now knew I definitely wouldn\'t make it home.  We finally made it to a service station and I ran.  I was relieved, but at the same time knew it wasn\'t over.  Luckily we made it back to my house so I could at least be sick the rest of the night in comfort.  I\'ve had a few diagnoses from different folks:  tainted food, allergic reaction, irritable bowel attack, stomach virus.  None of them sound pleasant, but I\'m hoping that whatever it was, it\'s over now.  I didn\'t eat much on Sunday due to my uneasiness, but was feeling ok otherwise.  The lovely Kim made me some pasta for dinner, which was good to eat instead of soup.  Yesterday I ate safely and was ok, and today I ate breakfast so far, and think I\'m ok.  Rob, I don\'t think we can post pictures on the blog since I don\'t own the site.  If I had my own website where the blog was, then it could work, otherwise it cannot...Any input on the Van Horn trade?  I was pissed.  I thought he was playing well for the Knicks.  I\'m off to do homework.   '),(18354,1826527,'16,February,2004','2004-02-16 00:00:00','    I feel empty inside with my dearth of recent blogs.  Hence I am posting.  However, It\'s got to be quick.  Life has been ok for me lately.  I still dislike school, like an 8 year old.  (Though I never disliked it when I was 8...)  I am going back to work after a week off from the theater.  I\'m not really looking forward to it, so maybe I\'ll like school better, motivated by my dislike of work.  hehe.  The Pats won the Superbowl.  Sorry to all you Panthers supporters out there, since they really were a fun team on a few levels.  I hope the Pats keep a solid franchise with their new structure which I\'m sure this time around will be more scrutinized and copied.  I can understand most people not being interested since they\'re not real flashy (even that is an overstatement).  But I find them likable and successful simultaneously.  I\'m biased and I know it.  The celebratons were ok in Boston.  Good mood, more subdued since they were favorites.  few injuries, one drunk guy ran over somebody.  I went to the parade and took a few pictures.  I\'ll try to post one here later if someone will tell me if it\'s possible.  I\'ve been reading a really great book right now as well.  It\'s called \'The Buying of the President 2004\'.  It\'s a book by a guy from the  urlLink Center for Public Inegrity , one of those non-partisan, non-profit groups.  The whole idea is to give actual information about the political candidates, their careers, their backgrounds, and their financial backing as opposed to all the stuff in the media which is watered down, filtered, spun, and selectively chosen.  From my reading thus far, it\'s unbiased toward either party and gives good information on who has funded the parties campaigns, and the expected return for those funds in the form of legislation favors.  From what I can see, the whole system of lobbyists and contributions almost directly leads to results in legislation, like money for laws.  Anyway, it\'s saddening but useful to know.  So I wanted to know what to actually expect from a candidate in form of action if elected, as opposed to fancy speeches and empty promises.  The list of contributors will almost certainly tell what issues wll really be tackled if a guy is the next president.  Anyway, I think I\'ll have a more informed vote.  That\'s my 2 cents for today.  Have a good one all.   '),(18355,1826527,'12,February,2004','2004-02-12 00:00:00','    Howdy.  It\'s been some time since the last blog of mine.  Unfortunately, there hasn\'t been much in terms of daily news.  However, I did have the luck of purchasing a new  urlLink car  on Tuesday.  I am extremely happy, as it feels good to finally have a car to call my own, and a new one at that.  It\'s also good because my credit was good, and now it will be even better for my future.  It\'s a four door with automatic transmission.  I didn\'t get any major options or packages besides a/c.  The only option I wanted was ABS, but it was only offered on the upper level models, so I decided it wasn\'t necessary for $4,000.  The car is a sand color, has a cd player(with controls for the radio right on the steering wheel!).  It is a compact car, but has lots of room in the interior and trunk.  It comes standard with front air bags, and has a feature in which the car can only be started with the key because of a chip that only reads the key\'s code.  The only negative, if there is one, is that I am stressing about insurance.  The dealer gave me spot insurance, which is good for 14 days.  I don\'t officially have insurance yet, so these quotes are important because I have to officially apply asap, since it takes 7 business days for most insurance policies to process, which would take me to exactly the Tuesday when the spot coverage expires.  If the spot coverage expires before my insurance kicks in, then I have to leave the car at the dealer until it takes effect.  That would suck.  It\'s not the end of the world, because it would be a couple days max, but it\'s something I don\'t want to deal with.  So that\'s my big story of good news.  The last month has been pretty miserable for me, so this was a nice lift.   Still struggling at school.  Excessive reading is taking its toll on me, and I haven\'t had much of a social life since classes began.  Work\'s the same story:  trite and annoying.  Routine hasn\'t changed, and my tiredness from reading carries over into my work, causing extreme crankiness...Valentine\'s Day is Saturday.  Does anyone care?  I sure don\'t.    In my final bit of information that you probably don\'t care about...I haven\'t been eating breakfast on a normal basis the last 2 weeks, and couldn\'t figure out why.  I always eat breakfast, unless of course I\'m sick.  Well, I\'m not sick, but I think it\'s because I\'m not hungry in the morning.  I have class Mon and Tues nite, and eat when I get home, and work late Thurs, and eat when I get home, so I think it\'s because I eat right before I go to bed.  Anyway, maybe I\'ll start eating breakfast at night.   '),(18356,1826527,'08,February,2004','2004-02-08 00:00:00','    As Brian and I have discussed recently, Sunday has to be the worse day of the week.  Brian says its a bad day because you know you have to go to work the next day and all you can think about is how bad work sucks.  I totally agree with him but I have another arguement as to why Sunday sucks - church.  It\'s all well and good that people go to church but being that I live 30 seconds away from a church I hate the traffic that is causes.  Church traffic sucks by itself but being that the Clifton Commons, Costco, Target, and Route 3 are also nearby, it makes getting out of my driveway a living hell.  I despise trying to go out on Sunday because I know it takes me 20 minutes to get out of my damn driveway.    I dragged Bri with me to see the Westfield Symphony Orchestra last nite.  My professor is the conductor of the orchestra and he requires us to go to 2 concerts and write reviews for each one.  However, this is the only actual work we have to do all semester.  That\'s fair, right?  Anyway, as we were going to cross the street to get to the parking lot, some guy in a Lexus truck says to the cop that\'s stopping traffic for us pedestrians ,\"Hey!  I\'ve got a dog that has to get to the hospital!\"  The cop had this bewildered look on his face which I thought was totally priceless.  So he tells us not to cross and lets this guy with the sick dog go.  As he drove by, I saw a german shepard in the back seat wagging his tail.  Now I\'m not a dog expert or anything but I think that if a dog is sick he wouldnt be wagging his tail.  I thought tail wagging was a sign of happiness.  I could be worng.  So next time I see a dog wagging his tail, I am to assume he is choking on his milkbone and should get him to a vet immediately.  Anyways, the concert was somewhat enjoyable... until we got lost.  When we were leaving the parking lot, we went the wrong way and continued to do so for about 30 minutes.  Once we finally got on the parkway, I was relieved.  However, this relief went away very quickly as we encountered many assholes who love to go 90 mph until they get right on your bumber and then pass you on the right.  If anyone has forgotten, you\'re supposed to pass on the left and the speed limit, contrary to popular belief, is not 90.  Actually there was some poopface in a jetta that entered onto the parkway and proceeded to cut across 3 lanes, nearly hitting some guy in a minivan which made that guy swerve into the left lane and almost hit some other guy in the the left lane.  The funny thing was the guy in the jetta didnt slow down or stay where he was.  He insisted on being where the minivan was.  Weird.  Then i got behind some drunk guy swerving all over the road and doing 35 mph. Driving in NJ is like that extreme elimination show that I love so much on Spike TV.  Yeah so thats my boring blog for today.  Hope you enjoyed!   '),(18357,1826527,'03,February,2004','2004-02-03 00:00:00','    Ahh, glorious Tuesdays.  I get to hopefully catch up on my sleep for the week by sleeping in till about 9:30.  I then continued doing homework, which is a perpetual event for me.  It only goes on hiatus while I\'m working, sleeping, and eating.  It\'s now just about noon, so I cannot function intellectually anymore, so this blog may be branded with dozens of swear words due to my lack of vocabulary.  Yeah, so homework is killing me.  The readings are just overwhelming, and I\'m hoping it gets better very soon.  I also noticed that I can barely read the notes I took, so I have to remember how to write well enough to comprehend it at a later date.  Since I woke up and did homework all morning, I have yet to take a shower or do anything else.  Luckily, the only thing on my agenda is my night class at 5:30 tonite.  No work today, thank god.  I currently have an academic hold at Montclair due to immunizations.  Apparently, I never received a second measles vaccination, which makes no sense because William Paterson has the same requirements as Montclair, and I graduated from WP without any word of this missing vaccination.  So I have to go up to WP and get my records from the health center to see what the deal is.  The only problem is that the only time I can go is Tuesday during the day, and my mom has the car today, and last Tuesday it snowed.    I will now go into my Super Bowl retrospective...I hitched a ride with my brother\'s friend  Moose  to my brother\'s house.  We arrived at 4 and ate immediately.  We had some doritos, veggies and dip, and m&ms to start, then the good food arrived: *4 pound shrimp ring *3 pounds of fried and coconut shrimp *75 wings from hooters *100 hal-dogs *chicken fingers and jalapeno poppers *nachos *Rachel\'s home-made lasagna *dirt *my jello cake thing *jello football helmets  I ate one or more of everything listed, and we played catch with the famous  Illuminator  football pre-game, and at the half.  As for the game itself, well I was bored to tears for the first 35 minutes of gametime.  Finally, there was some action before the half to stir some anticipation for the second half, which was outstanding.  I have to admit, I was pulling for the Panthers, especially when they actually kept it close, but I knew they weren\'t going to hold the lead.  The night ended on a good note for me as I won $65 for getting the final score in a grid pool.  Only $3,435 more needed for a down payment for a car!  Hey hey!  I\'m off to the showers and probably read more before class.   '),(18358,1826527,'01,February,2004','2004-02-01 00:00:00','    Greetings.  Just a short blog to wish Rob and his Pats luck today, although I can\'t really say I\'m rooting for either team, just hoping for a good game.  Enjoy the festivities all, and I\'ll talk to you soon.   '),(18359,1826527,'30,March,2004','2004-03-30 00:00:00','    I went to class last night feeling even worse than I felt last Monday, but I didn\'t want to miss 2 in a row.  I woke up today with my entire left side of my nose stuffed completely.  At least my throat doesn\'t hurt as much.  I think I actually slept pretty well, however, so that\'s a positive.  I\'ve already done some homework this morning, but still have a little more to do.  I hope I can get rid of this cold or whatever soon.  I found out last night that if I choose not to enroll in classes in the fall, I have until October to start paying back my loans from this semester.  That gives me some time over the summer to save up some money hopefully.  That reminds me, good luck to  Kevin   in your MBA classes starting tonite, and good luck to  Matt  in your exit exam and finding a job!  I don\'t believe you\'re done already.  Wow.  If I feel well enough,  Kim  and I are probably gonna hang out with  urlLink Katie  and  Chris .  Then on Sunday, I have to go to my uncle\'s house for some reason to have pre-Easter dinner.  Easter is in 2 weeks.  This means 2 Sundays in a row are going to be wasted sitting inside at a relative\'s house.  This is bad for several reasons, but mainly because I have to finish my 20-page paper, and two 10-page papers before the semester ends around May 8th or something.  Sunday is one of my days I do a lot of work, but I guess not those two weeks.   '),(18360,1826527,'27,March,2004','2004-03-27 00:00:00','    I am back in North Carolina and as scared as I am to admit it, it feels pretty good. I suppose anything would be an improvement over Warm Springs. While it was truly a great learning experience, I was definitely ready to get out of town and re-join civilization. The next (and probably biggest) roadblock is our comprehensive exit exam on Friday. Needless to say there\'s lots of studying to be done, but just not that much getting done. My motivation is very low. I really just want to concentrate on finding a job. On that note, I submitted my first application to a Rehab hospital in DC. I should hear from them by the first of May. We shall see. Kudos to Brian for having the foresight to pick Xavier to go this far. I would not be surprised to see them beat Duke tomorrow. My bracket turned to crap relatively quickly thanks to early exits by two of my final four teams, thanks Providence and Wisconsin! Anywho, I have to open up more books and try to convince myself I\'m actually being productive in my studies.   '),(18361,1826527,'25,March,2004','2004-03-25 00:00:00','    I\'m wearing my freshly-purchased  urlLink Vernon Wells  jersey with the new Jays logo.  I know every team has black as a color, but man, it looks really good with the blue and silver logo.  It\'s Saturday and I am happy it\'s the weekend.  I\'m still not feeling spectacular, but it\'s nice out today, so I have to get out and get some air.  I\'m excited about baseball starting soon and noticed that the most expensive seats to see a Jays game is $54 in Canadian funds.  That\'s pretty damn good I think to sit on the field in a nice stadium.  Anyway, while Kevin may be starting a Master\'s program, I\'m putting mine on hiatus.  By the way, Kev, best of luck.  I just wasn\'t into the program at all, and I don\'t see the point in continuing to get into further debt.  I may get back to it a few years down the road.  Who knows?  Well, I\'m gonna go check my mail and watch the Mets game before hanging out with Kim.   '),(18362,1826527,'25,March,2004','2004-03-25 00:00:00','    Long time, no blog (discounting my St. Patty\'s Day blessing...).  Well, I went and did it.  I joined an MBA program.  Tonite was our orientation and the first official class, Applied Communication in Organizations, starts Tuesday.  This is an accelerated program through the University of New Haven.  Each class is 3.5 hours long and meets twice a week.  Each class lasts for 5 weeks.  So, 22 months from now (Feb. 2, 2006) I will graduate (if I make it that far...) with an MBA.  They give us mid-July till the beginning of September off at least.  Woohoo (sarcastically...)!!!  On top of that, I\'ve been house-hunting for the past month and a half; with no luck.  Connecticut is too damn expensive for a single income household, i.e. Me.  At least the areas within a normal commute to my job are too expensive, unless I want to live in the ghettos of Bridgeport or New Haven.  Which I don\'t.  So, I have until the end of April to find something, or I\'ll just rent for another year.  And on top of all of that, I\'ve just been given double the responsibility at work.  I was pretty close to capacity before; but now I am just completely overloaded.  It\'s insane.  Needless to say, I hate my job even more than before.  It\'s got to get better, doesn\'t it?!?!?  Oh, and they just laid off 92 people today.  I feel fairly secure in my position.  I think I\'ve made my self valuable enough that they can\'t fire me... yet...  So, I think there is a pool forming for when I am going to have a nervous breakdown...  At least the house hunting is only going to last for another month; but if I do find something, then I am going to have to deal with trying to pay for it.  And if they MBA gets to be too much, I will just drop out of the accelerated program and go into a regular part time program.  Oh well, that\'s it for me.  I\'m gonna go back to sucking my thumb while laying in the fetal position in the corner, waiting for the next 22 months to pass...   '),(18363,1826527,'23,March,2004','2004-03-23 00:00:00','    Hellooooo.  I spent last night not going to class.  I felt like crap.  I don\'t know if it\'s a cold or allergies, but I just felt terrible overall.  Tired especially.  I\'m stuffy today and still tired, even though I slept 800 hours.  Probably going to class tonite, unless I suddenly feel worse, but hopefully that won\'t happen.  I got a haircut this morning, so the girl who cuts my hair and I took turns sniffling, since she is also sick.  I did some work last night for some of my papers that are due at the end of the semester, but my eyelids were too heavy to do much.  I did a little more this morning.  Don\'t feel like doing much, though.  I\'m still feeling very blah.  I have further reason to hate Warren Sapp now.  I hope the Raiders go 0-16 this year.  Oh yeah, the Bengals still don\'t have a starting cornerback.  Hey Garet, how\'s your cover skills?  As for the NCAAs, I\'m glad I didn\'t enter a pool, but I filled out a bracket for shits & giggles.  I usually pick a bunch of wacky teams, and Maryland has been a pick for me to go to the finals for like 6 consecutive years.  My final four is a joke, but I still have some teams alive...I had GT, Ok State, Maryland, and Xavier(sorry Matt, but I fantasized about Illinois beating Duke in my bracket...it could happen. hahahahahahahahahahaha).  Anyway, I still did well considering all the upsets that happened.  I actually picked quite a few, but it doesn\'t matter anyway, since I\'m not in a pool.  Somehow, the Cuse is still alive, which I didn\'t think would happen either.    In the possible-good news front, my excellent sister in law Rachel is putting in a good word for me as a teacher in Howell.  Here\'s to hoping she\'s a good persuasionist.  There\'s only so much more I can take of this walgreens nonsense.  Hey Rob, we may come up to visit you in Boston for a weekend this summer, so get the pinata and clown ready for a kick-ass party.   '),(18364,1826527,'17,March,2004','2004-03-17 00:00:00','    Happy St. Patrick\'s Day to one and all.  Here is an old irish blessing for you:  When we drink, we get drunk. When we get drunk, we fall asleep. When we fall asleep, we commit no sin. When we commit no sin, we go to heaven. So, let\'s all get drunk, and go to heaven!   '),(18365,1826527,'15,March,2004','2004-03-15 00:00:00','    Home again from work due to illness:( So since I am doing nothing of importance,  I\'ve decided to take a look at an old blog about new year\'s resolutions.  Let\'s see how I\'m doing shall we?  1. Stay away from red meat...  Not doing too badly with this one.  I\'m actually at the point where eating a burger makes me feel ill.  Is this bad? 2. No fried foods....  Doing well with this too, but when I do eat fried foods its usually in that yummy salad at bennigans or fried fish. 3. When I\'m mad at someone, I usually take it out on an innocent person...  I haven\'t really gotten too upset with anyone yet, but I\'m sure it\'s bound to happen. 4. Exercise more...  Yeah right.  There have been talks about running again but the stairmaster thing has yet to happen. 5. No more road rage....  I\'ve calmed down slightly, but there are some really bad drivers out there that require me to curse at them even though they can\'t hear me. 6. Save more money...  I haven\'t been spending a lot of my money on stupid stuff, but I need new clothes for the summer for both work and casual time.  I\'m pretty sure my savings will dwindle then. 7. Continue to not eat after 8pm....  This has never really been a problem for me.  I think I picked this resolution because I know it\'s the only one I wouldn\'t mess up. 8.  Stay away from ice cream...  Haven\'t been doing so well with this one.  It\'s always when we go to Friendly\'s.  So I hereby declare I shall  not go to Friendly\'s more than once a month.    So I guess I\'m not doing so badly.  It\'s just that damn exercise thing.  Walking 10 miles around campus with a 50 pound bag has to count as exercise right?   '),(18366,1826527,'14,March,2004','2004-03-14 00:00:00','    It\'s Sunday, and as usual, I already don\'t want Monday to come.  This week it\'s not as bad as usual because I know I don\'t have school this week.  I still have to try to do a lot of homework, however, because I will always be behind once classes resume next week.  Too much reading and writing.  Anyway, yesterday was fun.  Despite Kim still not feeling well, she was a champ, and still had the energy to hang out.  We went to Houlihan\'s in Weehawken, which is right along the piers.  It would have been even more beautiful if it wasn\'t so cold and windy outside.  We ate their dirty nachos and had a damn good meal afterwards.  We decided to catch a movie at the venerable  Allwood  theater.  It\'s 6 bucks to get in as opposed to 9 or 10 elsewhere.  We decided on  Hidalgo  because it was playing sooner than the Johnny Depp movie, which I didn\'t really want to see anyway.  So it started at 4:30 and ended just before 7.  It was a little long, but enjoyable.  It was nice not hearing a movie spattered with \"fucks,\" \"shits,\" and \"assholes.\"  I don\'t think there were any of those;in fact, I think the only word I heard only a few times was \"hell.\"  And there wasn\'t any gratuitous sex or any at all for that matter, which surprisingly was refreshing.  It was just a good ol\' adventure along the lines of a hero like Indiana Jones, but with not nearly as much action sequences.  We went home afterwards to eat our dinner of hot pretzels and watched South Park and Chappelle\'s Show until Kim had to go home.    I got a jury summons the other day.  This is gonna be the first time I\'ll have to go, and I\'m not real excited about it.  It\'s April 20th, which is a few days before my bday, and it\'s on my only day off(Tuesday).  I am not a fan of driving into Newark in the first place, so I hope I only have to report one day and get dismissed.  If I\'m real lucky, I\'ll get called off the night before.  I got summoned while at WP, but I got excused because I was working to pay off my school, and had classes that conflicted.  Also, I have to renew my license before the end of April, and there\'s all this new crap that you need in NJ now.  I need a civil birth certificate issued by the state(not the one from the hospital), as well as two other forms of ID and proof of address.  And I\'ll still have to wait in line for 2 hours.  I hope these digital licenses are worth it, because no one believes my age, so there\'s really no way I could fake this ID.   '),(18367,1826527,'11,March,2004','2004-03-11 00:00:00','    I finally got it...  The evil cold that has been going around has finally reached me.  Although I told the cold I didn\'t have time for it in my schedule, that dirty bastard insisted on making me sleepy and stuffy, as well as give me a sore throat.  I actually stayed home from work today to rest so my cold wouldn\'t get any worse and hopefully it helped a little.  I slept most of the day which is very shocking.  I never sleep, especially when I\'m sick.  I usually stay up to watch those dirty talk shows I would never otherwise be able to see.  But I woke up today at 7, ate breakfast and didn\'t move from the kitchen table.  All I thought was, \"ok kim.  you should go take a shower and get ready for work.  you shall be productive!\"  I did not listen to myself though.  Aroung ten to eight I left a message on my bosses phone telling him I wasn\'t coming in.  I\'m sure he was excited to know I wasn\'t going to be there hounding him every ten minutes for something to do.    Anyhoo, I talked to my friend Ryan from Chicago today and we talked food.  It started with me saying that I wanted egg drop soup and it continued from there.  I had mentioned something about the new Thai restaurant by Bri\'s house and he said we should go.  Thai is one of his favorites so he recommended some dishes to try.  Although they sounded yummy, he said thai food was on the spicy side and I\'m not too much of a fan of spicy food, but I\'ll try it.  I\'m kinda tired of eating at the same places every week (sorry Bennigans!).    Spring break is upon all at Montclair State University and of course I\'ll be working and probably still feeling sick.  I\'m not really looking forward to it because I\'m not going away to a tropical place, but not for nothing I don\'t really want to go at this time.  The prices are really high now because it is spring break time and you won\'t be able to escape the drunk morons who will probably be in the gutter, face down in their own vomit.  Not that I have anything against drinking.  I\'m against the people who do it all the time and have no clue how to handle themselves.  Just say no boys and girls!  A girl in one of my classes asked if I was going away for spring break and i said, \"no.  i dont do body shots.\"  She thought it was hilarious.  Going to stop blogging now because I probably don\'t make any sense.  Bye!   '),(18368,1826527,'10,March,2004','2004-03-10 00:00:00','    Next week is spring break!  Woo!  Not real sure what that means, but no class at least.  I guess it\'s like a catch-up week for my work, but I just don\'t want to do it at all...I talked to my friend Evan, who I met while subbing in Livingston, and he just got a job in Newark.  His briefcase has his planbook, some worksheets, and a gat.  Protect ya neck.  No seriously, how he ended up there is a funny story I would have to tell you in person.  I\'m addicted to Planter\'s Honey Roasted Peanuts right now.  I bought a can from work, and haven\'t been able to stop eating them all night, except while eating dinner.  Looking more and more forward to the warmer weather.  We have plans on purchasing bicycles this summer to get around as a backup to resuming our jogging schedule.  I can\'t wait to get back into the air.  I need to get my blood pumping and stuff.  I also really want to play some ball, but they took away my court in Nutley.  I think they\'re putting up a monument for Martha Stewart in its place.  Ba-zing.  No, they just tore it up for no reason.  Looks like I\'ll be playing in my crooked-ass yard full of tree roots...  NFL free agency is always an adventure.  You can\'t buy a jersey of your favorite player because more than likely, he won\'t be on the team for more than 3 years.  Over the past two seasons, however, I have begun to like it because the Bengals actually have a chance at signing some decent players.  This year they brought in 2 new defensive starters, as well as re-signed their awesome kicker and center.  They did decide to cut both of their starting cornerbacks from last season without signing a replacement for either yet, although they are supposed to sign either Bobby Taylor or Troy Vincent.  While I\'m on the subject of sports, I am really glad the Rangers cleaned house and brought in some fresh faces.  They got rid of a lot of waste, but still have an abundance of it(Tom Poti, Darius Kasparaitis).  They got rid of the only player who represented the team for the past 15 years, which was upsetting, but his play isn\'t nearly as good as it once was.  They\'ll finally have a chance to give their younger guys(young in terms of Rangers years) more of a chance(LaCouture, Ortmeyer, Purinton, Lundmark).  They\'re not the best, but it\'s better than watching a bunch of underachieving 40-year-olds skate around aimlessly.  Hmm, what else, the NCAAs are around the corner, and my 6 year tradition of picking Maryland won\'t continue this year.  Actually, I think I picked Kansas last year.  Luckily I was wrong as usual.  I miss the bloggers.  Looks like I\'ll be doing some phone usage to get your asses back up here.  There\'s no humor here without you guys!   '),(18369,1826527,'08,March,2004','2004-03-08 00:00:00','    Ok, so I just got home from class, and I am bored, so I got this from Kevin\'s girlfriend,  urlLink Meredith \'s site.  Hope she doesn\'t mind.   [[PAST]]] [1] First grade teacher\'s name: Mrs. Mossage [2] Last word you said: truck [3] Last song you sang: Why Would I Lie?-face to face [4] Last person you hugged: Kim [5] Last thing you laughed at: that st. patrick\'s day guinness commercial [6] Last time you said I don\'t remember: today at work, probably to get away from a customer or something [7] Last time you cried: I don\'t remember [[[PRESENT]]] [8] What\'s in your CD player: How To Wrong Everything-face to face [9] What color socks are you wearing: white [10] What\'s under your bed: monopoly, life, trivial pursuit, careers(board games) and the clash on broadway and doors cd box sets [11] What time did you wake up today: 6:15am  [12] Current taste: frosted flakes [13] Current hair: wet and frizzy(just got home from class and it\'s snowing) [14] Current clothes: Longsleeve vans shirt; blue jeans, blue sneakers [15] Current annoyance: school [16] Current longing: summer [17] Current desktop picture: the ocean and sunset [18] Current worry: money to pay for insurance [19] Current hate: work [20] Current favorite article of clothing:  o\'neill hoodie [22] Last CD that you listened to: see #8 [23] Favorite place to be: with kim, driving down the shore on a nice day, or just at the shore [24] Least favorite place: the patricia grand hotel in myrtle beach, sc [25] Time you wake up in the morning: Sometime between 6:30-8 [26] If you could play an instrument, what would you play: keyboard or stand-up bass [27] Favorite color: green [28] Do you believe in an afterlife: i think so [29] How tall are you: 6\' [30] Current favorite word/saying: blast!  [31] Favorite book: the red scream [32] Favorite season: late spring [33] One person from your past you wish you could go back and talk to: zbucki [34] Favorite day: thursday [[[FUTURE]]] [35] Where do you want to go: aruba [36] What is your career going to be like: fun [38] What kind of car will you have: mazda 3 [[[HAVE YOU EVER...]]] [ Said \"I love you\" and meant it?: ] yes [ Gotten in a fight w/your dog/cat/bird/fish, etc: ] No. [ Been to New York?: ] yes [ Been to Florida?: ] Yes [ San Diego, Cali?: ] no [ Hawaii?: ] No [ Mexico?: ] no [ China?: ] No [ Canada?: ] Yes  [ Danced naked?: ] no  [ Dreamed something really crazy and then it happened the next day: ] Yeah [ Wanted to be the opposite sex: ] no [ Had an imaginary friend?: ] Nope  [[[RANDOM]]] [ Do you have a crush on someone?: ] sorta.  it\'s a little deeper than a crush [ What book are you reading now?: ] text books [ Worst feeling in the world: ] sadness from losing someone close to you [ What is the first thing you think when you wake up?:] do i really have to work [ How many rings before you answer?: ] usually two-three [ Future daughter\'s name: ] Brielle or Kayla [ Future son\'s name: ] not sure [ Do you sleep with a stuffed animal?:] yes, the bear from kim [ If you could have any job you wanted, what would it be?: ] kyle korver [ Wish you were here: ] incubus song [ College plans: ] graduated willy p; currently at monkey state for grad school [ Piercings: ] None [[[THE EXTRA STUFF]]] [ Do you do drugs?:] Nope [ Do you drink: ] Yup [ Who are your best friends?] kim and my brother [ What kind of Shampoo and Conditioner do you use?: ] head & shoulders [ What are you most scared of?] cobras!  argh! night terrors(it\'s a simpsons reference) [ What clothes do you sleep in?: ] those comfy flannel pants and a tshirt [ Who is the last person that called you?: ] Kim [ Where do you want to get married?: ] in a nice outdoor park or on that cliff in rio de janeiro at sunset [ If you could change anything about yourself what would that be?] my teeth are crooked and too bunched together [ Who do you really hate?: ] kevin costner [ Favorite number: ] 37 [ Been In Love?: ] I am in love [ Are You Timely Or Always Late: ] too timely-always early [ Do You Have A Job: ] yes, i have 2 [ Do You Like Being Around People: ] most of the time [ Best feeling in the world: ] getting something done [ Are you for world peace: ] is that a question? [ Are you a health freak: ] no [ Do You Have A \"Type\" Of Person You Always Go After:] no [ Want Someone You Don\'t Have Right Now: ] I\'m happy with the love of my life, thank you. [ Are You Lonely Right Now: ] Nah. [ Ever Afraid You\'ll Never Get Married: ] Nope [ Do You Want To Get Married: ] Yes [ Do You Want Kids: ] yes, Two.  [[[IN THE LAST 48 HOURS, HAVE YOU...]]] [ Cried: ] No [ Bought Something: ] dinner, lunch, gummy bears, gum [ Gotten Sick: ] Nope [ Sang: ] Yes [ Said I Love You: ] Yes [ Wanted To Tell Someone You Loved them: ] I already told everyone who needed to be reminded. [ Met Someone New: ] Nope [ Moved On: ] from what to what? [ Talked To Someone: ] yes [ Had A Serious Talk: ] no [ Missed Someone: ] Yes [ Hugged Someone: ] Yes [ Kissed Someone: ] Yes [ Fought With Your Parents: ] No [ Dreamed About Someone You Can\'t Be With: ] No [ Had a lot of sleep: ] no  why did i just do that?  i have no idea, but what\'s done is done, eh?   '),(18370,1826527,'02,March,2004','2004-03-02 00:00:00','    With all this hype about the Mel Gibson film,  The Passion of The Christ  or something like that, I was thinking that the actor who plays Jesus in the movie can\'t really get another role.  It\'s like, \"Hey man, I just got the part of Stiffler in American Pie 4.\"  \"Oh yeah, well I was  Jesus .\"  I mean once you\'re Jesus, you can\'t really get a bigger role, so I guess his acting days are over.  Anyway, did anyone actually see the movie?  I\'m curious about it, but am afraid I may not get the full effect of it since I\'m not really religious...  Today\'s my brother\'s birthday.  Happy birthday, even though I don\'t think you read this.  I got my first paper back from my theories class.  He gave me an A and wrote many compliments on it.  I was happy because I don\'t think I did A work, and even if I did, it was largely due to the fact that everyone shared their ideas with the class and was given feedback before we worked on the paper alone.  So I handed in my second paper today, and am glad I have that A, because this will surely even out the grade to a solid B.  I honestly had no idea what I was talking about, but only time will tell what el profesor feels.  I\'m relieved that school is over for the week, unfortunately, I have to be at work at 8 tomorrow.  Zoinks!  I\'m contemplating my future at grad school because I feel like I\'m struggling with the material, and feel as if I\'m not really ready to commit myself to it at this time.  I\'m hoping things get a bit easier as I get further into different classes, but right now I\'m completely overwhelmed and pressed because whether I go back or not, I don\'t want to do poorly.    As for Kim\'s previous post, I think it would be nice to spend a week at my brother\'s house alone.  We care for their fish and maintain their pool, etc, but we get to lounge around, stay away from work, hit the beach, hit the pool, and possibly have a get-together (hint).  Nothing outrageous, but a little fun nonetheless.  So I look forward to that, and the possibility of finding a teaching job.  Yes, I\'m gonna try again after taking a reprieve, since I need a real job first, and then I can always go back to grad school.  It should be another interesting few months ahead.   '),(18371,1826527,'02,March,2004','2004-03-02 00:00:00','    So Wow.. It\'s been about a month since my last blog and with good reason.  Nothing exciting is happening.  Just school, work, and very little down time.  The only day that I do absolutely nothing is... umm... there isn\'t a day!!  I guess I can\'t say I\'m bored right?  So although Brian and I have no money to go on vacation, we\'re probably going down to Brian\'s brothers house in Point Pleasant for a week while he\'s in Peru.  There are some bonuses: BONUS 1- Rachel\'s leaving the keys to the convertible.   BONUS 2- They have a pool and a decent backyard to chill out in.  BONUS 3-  I believe they have bikes so we won\'t have to drive to the beach, which leads to BONUS 3-  They\'re only five minutes away from the beach!  I think this would be an awesome vacation.  It would be very inexpensive.  The only thing we would have to pay for is food.  Now we\'re just waiting to see when Kevin and Rachel book the trip.  I believe Rach said late June, early July which is prime beach season.  I\'m excited but I dont wanna get too excited if something doesn\'t work out.  NOT A BONUS: We\'re still in New Jersey.  So that\'s about the most exciting thing that has happened in a month for me.  On the other hand, I have been getting nothing but A\'s in all my classes and I\'m sure you\'re all so very proud... hehe.    Hope everyone is doing well!   '),(18372,1826527,'30,April,2004','2004-04-30 00:00:00','    Belated Birthday wishes to Brian. In scary and disturbing news, I am done with classes and just have to present my case report to the faculty on Tuesday, with graduation to follow next Sunday. The job hunt has been going pretty well. I\'ve been looking to head back to the DC area and have been offered one job so far, with one phone interview later today and a four interview in three day sprint coming up in mid-May. Tough decisions, but the good kind of tough decisions, as opposed to not having any job at all. This weekend (and all of next week most likely) will be a horrible sloppy mess, so Sparta, if you\'re interested in seeing some drunk PT students, give me a call. As a bonus we\'ll probably end up at Player\'s (also known as the worst possible club ever).  In continuing with the anti-Eli sentiment, I agree that he is a punk-ass. I hope Phillip Rivers becomes Drew Bledsoe to Eli\'s Ryan Leaf. One can only dream. Good luck to all in their finals weeks of classes, projects, and all that.   '),(18373,1826527,'28,April,2004','2004-04-28 00:00:00','    Yeah, so I\'m blogging again.  First I just want to say I\'m sorry to anyone who has been getting email from  freeflixtix.com .  For some reason, I thought it would be beneficial to sign up and get movie tickets.  I didn\'t realize they would bombard my friends with emails pestering them to sign up.  So I cancelled my account today when I finally found out what the actual website was, and anyone who didn\'t sign up and was getting emails should stop receiving them.   urlLink Let me know  if you get any more...  I got to work at 5:30 this morning to unload our shipment, and just happened to look at the schedule for today and found that I was mysteriously scheduled to stay an hour later than normal.  Good thing I looked.  So I stayed until 2:30, came home, washed my car and cleaned the interior.  I took a shower and now I\'m doing this.  I\'m tired as hell, and may be falling asleep around 8 or so tonite.  I listened to the Guns N Roses Greatest Hits CD yesterday, and that was fun.  I got it for $9.99 at Best Buy on Sunday.  Good deal...You know I never thought I\'d hear myself say this, but I wish the Jays had Billy Koch as their closer still.  Their bullpen has been bizarre to say the least the past 2 years.    I found out the 311 tix are gonna be $32.50 each.  That\'s a pretty tough pill to swallow, because \"convenience\" charges and taxes and such will probably bring them to $40 each.  That\'s an awful lot for a show, but they do put on a good show.  The Roots open, so that should be interesting.  Do I go or pass?  Too tired to do any homework tonite, but have this urge to play basketball strangely enough, even though I can barely keep my eyes open.  Suffice it to say I will not be basketball jones tonite, rather, I\'ll be lazy jones.  That\'s all for now.  Peace out.   '),(18374,1826527,'27,April,2004','2004-04-27 00:00:00','    Holy crap I\'m old!  Thanks for the wishes kids.  gracias.  I had my Tuesday class from 4-5:30 tonite.  He had some banquet to attend tonite, so we got to leave early, and the classtime was cut in half to boot!  Unfortunately I have to wake up at 4 to work tomorrow.  Ugh.  2 more weeks of school, and 2 more weeks to finish up my 3 papers.  Can\'t wait.  I cut my lip today, I guess while shaving, and it didn\'t stop bleeding until about 3:50, when I was walking to class...Yes, Eli Manning will get his ass kicked every game next year because the Giants have no line.  Rock!  My birthday details...hmmm.  Dinner was good I guess-Ribeye with sweet potato was tasty. I got some nice gifts-2 shirts, a new basketball, MLB 2005 for PS2, Trivial Pursuit Pop Culture game with dvd, some money.  All stuff I needed or will definitely use.    We\'re going to see Braid on June 7th, and 311 tix go on sale Saturday for a show July 24 at the PNC center.  Nice!   '),(18375,1826527,'25,April,2004','2004-04-25 00:00:00','    Happy B-day Brian!!  I\'m finally done with classes as of today!!  Only 2 more classes to take for the rest of my life next semester and I\'m finally free.  This semester has benn totally hellish in terms of classes and work, so I\'m so happy it is finally over.  In June, I get my master\'s degree, which signifies the halfway point of graduate school.  This past week was pretty enjoyable, I saw Death Cab for Cutie and Stereolab at The Cat\'s Cradle in Carrboro.  Both were pretty good shows.  Next month, I get to go to Vancouver for a conference.  hopefully Todd Bertuzzi won\'t be there to beat me up.  Anyway, the blogs should be more frequent now that I have some spare time. Hope everyone is doing well.  Talk to you later!   BTW, Eli Manning is punk and I hope he gets his ass kicked every game next year.   '),(18376,1826527,'25,April,2004','2004-04-25 00:00:00','    HAPPY BIRTHDAY BRIAN!!!!  That\'s right everyone, today is Bri\'s 25th and last nite we went out to celebrate this joyous event by going out to eat with his parents, Nana, Kevin and Rachel to the Outback.  I\'ll let Bri give the details.. not like there a lot to tell.  Sorry for the short blog but I must do some homework :(    Later!   '),(18377,1826527,'19,April,2004','2004-04-19 00:00:00','    Nilidigiphaguf Day goes back to a country in Africa once known as Botswanaland, but has since been renamed Mozambique.  On April 13 it was said that the indigenous people of the country set fire to their weapons in hopes of releasing the evil blood that may be left on them [the wooden handles are turned to ashes, which are then sprinkled on the first born child to protect him/her from evil spirits.  The metal is melted down and reformed into statues of their \"good\" gods.  So that\'s the tale of Nilidigiphaguf Day.  Hey Rob, I just read that the Pats acquired Corey Dillon for a second rounder.  I\'m happy for you, but scared at the same time, because as much as I hate the guy, he\'s a damn good runner.  Pats repeat!?    I am glad Kim got the classes she wanted because I feel awful about something that happened in her current music class.  We were supposed to attend a classical concert on Saturday[long story] but ended up not attending for various reasons when we arrived.  She was obviously upset because the show was in Rahway, and she drove all the way down there and couldn\'t go.  Anyway, since I\'m such an honest person, I tell her to email her professor and tell him exactly what happened and that you made the effort to go, but circumstances prevented you from going.  Anyway, he emailed her back and told her she has to attend another concert instead and report on it.  If she hadn\'t emailed her professor, she probably could have copied someone\'s program and b.s\'d the report.  Sorry.  Oh, that Nilidigiphaguf story was made up.  Rob made up the word when we were playing Scrabble a long time ago on April 13th.   '),(18378,1826527,'17,April,2004','2004-04-17 00:00:00','    Just finished registering for the summer and fall at MSU.  I\'m pretty happy about my fall schedule even though some of the classes are crappy.  The summer is a different story.  I\'ll be taking 2 classes (yuck!) but I\'ll be finished with both by July 8.  So if anyone cares here are my classes:  Summer Classes: -  Intro to Jazz.. the Intro to Music class I\'m taking now does not satisfy the fine and performing arts class for me.  I sure am glad someone told me this before the deadline to withdraw expired. -  Earth & the Environment...  Isn\'t it exciting that I get to look at rocks 4 days a week for 4 weeks at 4 hours at a time.. lots of 4\'s there.  Fall Schedule: -  Intro to Chinese Civilization.. ok, so it may not be interesting but maybe the professor will bring in some good food for us. -  Contemporary Issues: National Issues...  stupid requirement. just in time for elections -  Intro to Marketing.. Probably not that interesting but I heard the professor is easy though. -  Statistics for Business..  Whenever I can take a math class, I\'m happy.  Plus, I heard this professor is easy too. -  Intro to International Business..  Another requirement but this time it\'s a general requirement for business majors.  I don\'t particularly care for this topic but I gotta take it.  So that\'s that.  I actually got my classes at the time I wanted so that was a plus.  I guess it pays to be an upperclassman.  WooHoo!  Finally a junior!  That means only 2 years to go.. It seems so close yet so far.  Am I the only one who has no clue what Nilidigiphaguf Day is?   '),(18379,1826527,'17,April,2004','2004-04-17 00:00:00','    The NFL schedules came out the other day, and I\'m going to list the  urlLink Bengals  schedule, not because anyone wants to see it, but more for my own reference so I can look back and see who they play...  09.12  @ Jets 09.19  MIAMI [Sunday nite-ESPN] 09.26  BALT 10.03  @ Pitt 10.10  BYE WEEK 10.17  @ Cle 10.25  DENVER [Monday nite!-ABC] 10.31  @ Tenn 11.07  DALL 11.14  @ Wash 11.21  PITT 11.28  CLE 12.05  @ Bal 12.12  @ New Eng 12.19  BUFF 12.26  GIANTS 01.02  @ Phi  First of all, holy crap, they\'re on national tv 2 times, and they haven\'t been on Monday nite in like 15 years.  Secondly, holy crap, that schedule is no fun.  They play the Jets and Giants, so I get to see them 2 more times, and they end against the Eagles-brotherly rivalry.  Anyway, I was excited to see Monday nite, but wish it was a game they could win.   '),(18380,1826527,'15,April,2004','2004-04-15 00:00:00','    Tax Day for about another hour.  Hope all you procrastinators got that shizzy out in time...I got home from work tonight and saw a duck in my neighbor\'s driveway.  Normally I would say \"Hey duck\" and continue walking to my door, but something didn\'t seem right.  He was wobbling quite a bit and when I stopped walking to watch him, I noticed he was trying to quack but no sound was coming out.  I felt so bad, so I went inside and told my dad and mom.  They came out to have a looksy and said to try and call the animal shelter.  No one answered but a robot voice.  So I called the police non-emergency line and the guy said that animal control isn\'t there after 5(it was 7:30 or so), but he said he\'d try to call some wildlife guy.  A few minutes pass and the animal control officer called me.  He asked me some questions and told me to see if the duck moved when I chased him.  He moved...from my neigbor\'s driveway to my front lawn.  If you don\'t know my front lawn, then I\'ll tell you it\'s all on a hill.  The duck struggled to get up the hill on a gimpy leg, but somehow did it.  He nested there for a while, and then suddenly fell down the hill and onto his back.  He took a while to turn right side up, but finally did, and chilled on my sidewalk.  I felt bad for him.  Finally the animal control officer called me back and asked me some more questions before I said to come get him because he was obviously injured and would have been eaten by a cat or something if left alone.  The guy was there in about 10 minutes and took the duck with him.  I don\'t know the duck\'s fate, but at least I felt good getting him some help to maybe fix his injured leg.  I worked at 5:30 yesterday morning, and it was frankly a little creepy driving to work at that time.  No one was on the road really and it was dark, yet someone still managed to almost hit me in Wallington-the town with the worst drivers in America-on my way to work.  It went pretty quickly as I left at 1:30.  I couldn\'t really sleep when I got home, so I ate lunch and showered.  I fell asleep for about an hour before dinner, and went to bed after watching another pretty funny  South Park .  Today I went in at 10, and my mind was all flustered.  I felt like I should have been going home at like 4, even though I had to stay until 7.  Anyway, back to normal 8-5 tomorrow before the weekend starts.  Warmer weather in the outlook before  jury duty  on Tuesday.  I got some good directions and advice, so hopefully my nervousness won\'t be at its height on Tuesday morning.  Gotta go to bed soon.  Enjoy the weekend!   '),(18381,1826527,'14,April,2004','2004-04-14 00:00:00','    A very happy, though slightly belated, Happy Nilidigiphaguf Day to you all indeed.   '),(18382,1826527,'13,April,2004','2004-04-13 00:00:00','    Happy Nilidigiphaguf Day!  I just found out I have to be at work for 5:30 am tomorrow.  I think I\'ll be going to sleep as soon as I get home from class...   '),(18383,1826527,'12,April,2004','2004-04-12 00:00:00','    I\'m going to sound mushy for the first part of this blog, so don\'t read it if you don\'t want to...I felt extremely lucky today.  After all the times that I complain on this blog about how much shit luck I have and how much I hate my job and school, and how broke I am, I was thinking how lucky I am to have such a tremendous girlfriend.  I guess maybe I\'m getting over the fact that lots of people have things much worse than I do, but I believe it\'s just another sign of reasons why I want to spend my life with her.  When muddling thru a crappy Monday at only 10:30 in the morning on a day that consists of working to 3, going directly to the doctor, home to change and up to class, I found myself smiling after thinking about her.  And the sign of how true that one moment was this morning, she ended up calling me at work no less than an hour after that smile came across my face.  I\'m lucky to have such a wonderful partner.  I hope me saying this doesn\'t come off as bragging or I hope it doesn\'t make anyone feel bad, but it\'s actually just me hoping that my awesome friends have found or will find someone that will make them feel as lucky as I do.   If you are still reading, I\'ll breakdown my weekend.  Friday I worked, went to the doctor, and went to China Garden for dinner with Kim.  Good food and ate some good pistachio ice cream for dessert.  Saturday we grabbed lunch, bought some cds and the Weezer dvd, and dyed some eggs for Easter.  As Kim mentioned, we made an appearance at our friend Brandon\'s house, where we also chatted with John and Kate.  It was nice to see old friends doing well.  On Sunday, I ate Easter dinner with my family and picked up Kim afterwards to come back to my grandmother\'s house where we watched the Mets game and Arena football.  Watched Alias at night and did my homework after.  Thus ended the weekend.    Did anyone see South Park last week?  I missed it, but heard it was humorous.  So, I heard you got served.  Looking forward to seeing where the Sixers trade Iverson.  Actually, as much as I\'m sure most people hate him, I still like him.  They\'ve got some good players on the team, but they just need a few more pieces.  Willie Green, Korver, and Salmons are solid to decent fill-ins, with Green having talent to be a starter.  Dalembert is a monster in the making, and Kenny Thomas and Snow are still real good players.  Big Dog was a big waste, and DC\'s career is over, but if Marc Jackson recovers from his injury for next year, he will play as well as he did before the injury.  Anyone got Atari?  I\'ve got an itch to play some Dig Dug.   '),(18384,1826527,'11,April,2004','2004-04-11 00:00:00','    Today is Easter and I\'ve noticed it is probably the worse holiday of the year.  No one even knows what we\'re celebrating, including me.  The upside is eating good food and the downside is sitting around the house with relatives you really don\'t want to see for the most part.  I\'ve also noticed the next big holiday is Thanksgiving.  Now you might be saying, \"But Kim, what about Memorial Day and Labor Day?\"  Those holidays are kinda cool because it\'s all about bbq-ing and warm weather, but these typically aren\'t family holidays, which I think are desperately needed by those like me who despise sitting with relatives talking about the same stuff you talked about the last time you saw each other...  I guess you can say I\'m not a big family get-together person.  I don\'t think I\'ll mind it as much once I get my own place and start my own holiday traditions.    Anyhoo, I would talk about how interesting my life is but it really isn\'t.  Just work and school.  Nothing new.  Oh, but there was some fun last nite.  Me and Brian went by Brandon\'s house.  It was a spur of the moment thing and I was glad I went.  It was nice to see Brandon as well as Katie and John.  Itwas also cool to meet some new people.  I think there were about 10 people there at one time and 5 of us are/were Sam Goody employees.  Crazy.  We talked about all the silly stuff we did.  Good time had by all.  Hope to see them all again soon.  Time to eat the traditonal ham and kielbasa.  Later.   '),(18385,1826527,'09,April,2004','2004-04-09 00:00:00','    Hello blog.  I will apologize for being out of contact once again for a while.  I\'ve been fairly busy with random things, but that is no excuse damnit!  Anyhow.  I\'ve been working at the theater a decent amount still.  A couple weekends ago, I went to Atlantic City for a bachelor party of my college roommate.  That town is rotting, but there are casinos.  Anyway, his wedding is April 17th, so I\'m going to that event as well.  I think he\'s getting married in Florham Park.  My mother is convinced it is in a castle for some reason.  I think not.  I\'ve been spending most of my free time hanging around with a variety of people with whom I work.  I\'ve been doing this for a long time, but it seems this past week has been especially so.  Some people major in all sorts of things, like dancing and choreography, for example.  Other choices include radio broadcasting, saxophone, marine biology, french literature, but mostly psychology and creative writing.  I don\'t really understand some of them, but it seems nobody understands what I studied.  Anyhow, I feel pretty confident in figuring out whatever I need to, whereas I think none will ever figure out physics.  Speaking of physics, I\'m really horrible at pool.  I have always been so, but I guess I forgot.  Also speaking of physics, I went to a foot doctor and got orthotics since he told me my \'bio-mechanics\' of walking were harmful.  Anyhow, I\'ve had these things 2 days and today when I woke up my big toe on my left foot is in excrutiating pain, which is odd since my feet felt fine when I was actually wearing the orthotics.  I think they shall be adjusted.    I think I don\'t have any more \'speaking of physics\'s\'.  I hope everything is well with all of you.  I will post again soon.  -R   '),(18386,1826527,'06,April,2004','2004-04-06 00:00:00','    The fun at my uncle\'s house this past Sunday was beyond words, but I cannot post pictures on this site, so you cannot see what a shit time I had.  Needless to say, I\'m just as excited about the most boring holiday coming up this Sunday.  I finished my homework for tonite\'s class and I\'m waiting for Kim to get home from seeing her advisor before we grab a bite and head back up to school.  The game last night was pretty ugly, and it only finished as close as it did because UConn slowed down their pace and lost their momentum for the last 10 minutes of game time.  The final four games made up for the lackluster final anyhow.  Good start for the Jays, but I can\'t say I wasn\'t expecting the opening day loss.  They didn\'t start well last season either, and Halladay started 0-5 or something.  Besides, everyone kept saying how Halladay is dominant when facing Detroit.  Jinx.  Also, Lawrence Rocca, baseball columnist for the Star Ledger picked Toronto to win the World Series.  I can\'t say I wasn\'t slightly excited by this, but is he serious.  Their GM doesn\'t even think they\'ll be in the playoffs until next season.    I\'m going to renew my license this Thursday, so we\'ll see what all the fuss is about for these new digital licenses.  I\'m gonna go twiddle my thumbs now.   '),(18387,1826527,'31,May,2004','2004-05-31 00:00:00','    As Brian said, this weekend was great but today just sucks.  It\'s raining right now and I\'m stuck home writing a paper for my Intro to Jazz class.  I should have done it a few days ago but who does what they should?  The paper will give me something to do besides sit here and play free cell or lemonade game urlLink  .  Maybe after the paper, I\'ll read a book or something else nerdy.   '),(18388,1826527,'27,May,2004','2004-05-27 00:00:00','    Hope everyone had an enjoyable weekend.  Unfortunately, it\'s gloomy today.  I had a pretty nice weekend up until today, when I have to work 3-11.  Saturday,  Kim  & I went to  urlLink my brother  & sister-in-law\'s house down the shore.  We helped them finish putting rocks down in their yard and then had a big bbq to reward ourselves.  Had a burger, 2 dogs, beans, corn, fritos with hummus, tostitos, pickles, tomato & fresh mozzarella, malibu & pineapple juice, fruit salad, some cake, and I think that\'s it.  Pretty good variety of foods.  We watched some  Chappelle\'s Show  and then  Dane Cook \'s comedy special before turning in for the night.    Woke up Sunday with a very tight back and quite tired.   Rachel  made us all some good waffles with some fruit salad and sausage for breakfast.  We then helped to clean off their pool cover and put up a gazebo in their yard.  Their yard looks amazing with the rocks and the pool opened.  There\'s still a lot of room to play with a tiki bar or  urlLink wiffle ball  field.  After showering, my bro & I played a good game of wiffle ball homerun derby for old time\'s sake.  He won in 10 innings 6-5.  Obviously we felt the rust since we hadn\'t played in some time, plus our swings hurt our backs more than the shoveling from Saturday.  When we returned to the house, we were hungry, and decided to get lunch from a friendly place known as  urlLink Surf Taco .  Mmm, that was good, but left us all feeling like blimps.  So we went to sit our lazy asses in the sun by the pool for a bit and just talked and listened to some tunes.  After a game of HORSE with my brother, we got some ice cream and felt fat again.  We proceeded to chill and then it suddenly became cold and cloudy.  We stayed out for a bit, but started cleaning up and went inside.  We watched the dvd  Evolution , which I think did poorly in theaters, but was supposed to do well.  Anyway, we all enjoyed it, and I thought it was better than the first two times I saw it.  After dinner, we ate some dirty shrimp and waited for our filet minons to grill.  Good dinner.  After cleaning up, Kim & I had to drive home.  Overall, a relaxing weekend with lots of good food.    The ride home was somewhat crappy due to a huge backup at the Driscoll Bridge (somewhere around exit 129).  Mostly before and after that, we were at least driving at a tolerable pace.  Now it\'s time to mope until 2:30 when I leave for work.  Can\'t wait until our week in July.  Oh crap, it\'s not even June yet.   '),(18389,1826527,'27,May,2004','2004-05-27 00:00:00','    Jogging was fun today.  It\'s been a while.  It felt good, but at the same time, my legs burned.  Also, at the end, my stomach decided to convulse and finally digest my food from yesterday, causing me some discomfort(i.e. a stomach ache).  I\'ll probably be really tired at work, especially since I have to be there until 11.  Yikes!    The jazz place was pretty cool.  It was something different that\'s for sure.  I definitely would have enjoyed it more if I had not had to wake up at 4:30 in the morning for work.  I don\'t know what I\'m going to do today before work.  I\'ll probably clean my room or something.  I\'d clean my car, but it looks like rain again.  I\'m sure I\'ll end up blogging again later.   '),(18390,1826527,'26,May,2004','2004-05-26 00:00:00','    Last nite Brian and I went to  urlLink  Trumpets in Montclair.  I had to go see a live jazz performance for my jazz class and instead of going to a smokey nite club in NYC (which I would like to do someday) we went there and it was a cool experience.  There were only 10 people in the place: 2 performers, 4 employees, and 4 peeps in the audience.  The 2 guys were really good. The bass player was extremely good and totally into it.  I was shocked to hear Bri say that he would go back to Trumpets.  I know I would.  The only down side was the high prices for food and drink.  Last nite left me with $3 in my wallet.  Oh well...  Me and Brian are running together tomorrow morning for old time\'s sake.  Hopefully, I won\'t pass out after the first block.  Continuing with the exercise theme, I am planning on joining a gym soon.  I don\'t necessarily need to lose weight but if I happen to shed some pounds I wouldn\'t complain.  My main goal is to tone up a bit. Hopefully my laziness will not prevent me from reaching my goal.  Time to iron clothes for tomorrow.. Later.   '),(18391,1826527,'25,May,2004','2004-05-25 00:00:00','    It was warehouse day at work, which meant getting up at 4:30.  The day began on a great note, because at 6 am, a new assistant manager was scheduled(started last week), and me & one of the other assistants were working on unloading the truck and wondered where the other guy was.  So the one assistant paged the office to ask where he was.  About 5 minutes later, the new assistant comes thru the back door with this attitude about how he didn\'t like the way the other assistant paged him.  The quarrel turned to much louder voices and some other non-relevant personal info was thrown into the assistant\'s face.  The new guy had no right to do this because it\'s personal business, and he\'s new, so he doesn\'t know crap about anyone yet.  Anyway, the argument died when the assistant apologized(which he shouldn\'t have) and said he\'s sorry the new guy took it the wrong way.  About 10 minutes later the new guy starts again by saying how he interpreted the page.  I just said it\'s too early in the morning for this nonsense.  After the truck was unloaded, the two assistants talked in the office and settled the matter I guess, but throughout the rest of the day, I could still tell the new guy had a chip on his shoulder.  Unbelievable.  I wish people would act their age sometimes.  Memorial Day weekend is coming up, and I for one am going to Wildwood with Klein to drink skunky Molson Ice and smoke weed with a girl who\'s allergic to smoke.  (This is a reference that only Dennis would understand, sorry!)  Seriously, I think we\'re heading to my brother\'s on Saturday to help him finish laying out rocks in his yard and opening his pool.  We\'ll stay over and go in the pool Sunday, unless the predictions are right, and it\'s only 70.  Gotta work on Monday, though, but not sure what time yet.  Anybody doing anything special?   '),(18392,1826527,'25,May,2004','2004-05-25 00:00:00','    I got home from work at about 11:45, so I\'m not really tired yet.  Plus, I had to get the garbage ready to take out for tomorrow.  I have to be at work for 9 in the morning, which isn\'t horrible, but still not the greatest.  Can you tell I\'ve got nothing better to do with my time than blog since I\'ve had about 6 posts in 2 days?  Some lady today spent over $300 on band-aids and water.  I couldn\'t quite make out the connection between the two however.  Good night.   '),(18393,1826527,'24,May,2004','2004-05-24 00:00:00','    Ooooooo!  I like the new blog.  The green is a nice change from the blue.  I also like the profile thingy.  I made one for myself and found that the music section is almost an exact replica of Brian\'s.    Anyhoo, I am half way through with my intro to jazz class.  The class is surprisingly fun and I like a lot of the music the professor is presenting to us.  The only thing that sucks is that I have to go to a live jazz performance.  Normally, this wouldn\'t be a bad thing but the class is only 3 weeks long so the pressure is on!  I\'m sure it will be a good time.  I got a free ticket to Comic Strip Live in NYC, which kicks ass because it gets 2 people in.  The only thing I have to pay for is the two drink minimum.  Hopefully, I\'ll get to someone good like Dane Cook, Brian Regan, or Ted Alexandro. I heard someone got to see Dave Chapell there.  That would kick ass.    I\'m Rick James bitch!   '),(18394,1826527,'23,May,2004','2004-05-23 00:00:00','    I just checked, and it\'s not *\'d out...yet.  Also, I changed something from their template.  Where I have links, they originally had \"previous posts\" which gave the first sentence from the last 7 days worth of blogs so you can click on it and go directly to it.  I thought this was sort of dumb since you can simply scroll down and view the last 7 days worth of blogs.  Maybe I\'m not seeing the big picture.   '),(18395,1826527,'23,May,2004','2004-05-23 00:00:00','    I was getting tired of the blue.  Sorry.  I lost all my links when I changed the template, so if there\'s some links you want me to put up, just let me know, because I don\'t remember all of them.  Anyway, gotta be at work at 3, and I\'m going to see my grandmother beforehand.  I hope it\'s not 100 in her house.  I noticed while scrolling thru some of the archives that many bad words were *\'d out, so I\'m going to test it.  Fuck!  If it\'s bleeped out, then I\'m gonna be pretty pissed.     Alias  last night was pretty good.  Hope it comes back next season and they don\'t replace it with that dreadful Nick & Jessica variety hour of crap.   '),(18396,1826527,'23,May,2004','2004-05-23 00:00:00','    Never mind, problem fixed.  It just looked large in the preview mode for some reason.   '),(18397,1826527,'22,May,2004','2004-05-22 00:00:00','    It\'s late and I just changed the look of this blog.  I don\'t know if I like it or not, and I can\'t figure out how to change the size of the font.  Right now, it\'s made for folks with really bad vision...   '),(18398,1826527,'22,May,2004','2004-05-22 00:00:00','    I just talked to Rob!  I\'m glad he seems to be doing well.  He has been missed on our blog here, just as everyone else has.  It\'s very warm out today, but I am actually enjoying the warmth thus far.  It feels nice hitting my skin after that god-awful winter.  I still don\'t like sweating, but maybe I\'ll start taking ice baths to cool myself down before going out.  My mom\'s listening to the radio in the background, and that horrendous song by the fine young cannibals is playing-you know the one where the guy sounds like he\'s mumbling in a high pitched voice...she drives me crazy.  uh uh.  Dammit!  How come there\'s never a baseball game on when I\'m waiting for Kim to get home from work on Saturday?  Shouldn\'t they be outside in the sun scratching their nuts and chewing Big League Chew on this fine day?  No Mets, not even the Yankees, who I would allow myself to watch just to pass time, and not even rugby or something wacky.  Just infomercials, crappy vh1 programming, or terrible cartoons.  Ah well.  Adios amigos.   '),(18399,1826527,'19,May,2004','2004-05-19 00:00:00','    Happy Birthday Matt!  Next time I see you, I buy you a beverage of your choice.  Enjoy the day senor!   '),(18400,1826527,'19,May,2004','2004-05-19 00:00:00','    I knew the Blue Jays were on ESPN\'s daygame again today, and since I\'d be getting home around 3, I figured I\'d be able to catch a bit of the end of the game.  When I left work, at about 2:40, I turned on the sports scoreboard on the radio, and it was the 7th inning, and the Jays had just taken a 5-2 lead.  I got home about 20 minutes later, and the score on the tv screen reads 6-5 Twins, and then I hear something about grand slam and Matt LeCroy.  I come to find out that Terry Adams, who blew the game last Weds for the birds gave up a grand slam in the top of the 9th inning, and the Jays were then trailing 6-5.  Needless to say, they didn\'t come back and win in the bottom of the 9th.  Again I witness(sort of) the bullpen collapse.  I found out my grades last night and did ok.  I got a B and an A-  It\'s funny because I thought my grades would be switched, or at least thought I would have done worse in the other class, but I\'ll take it, especially since I\'m probably not going back anytime soon.  I\'m really tired, so I\'m gonna go lie down.   '),(18401,1826527,'17,May,2004','2004-05-17 00:00:00','    I\'m supposed to be at work in about 45 minutes, yet I really don\'t feel like moving.  I\'m not even ready yet, either.  It\'s gloomy out, so I guess I\'m just in one of those moods where I don\'t feel like doing a damn thing.  Well, I guess I shall be getting ready now.  Time for work routine: drive the same way to work, park in the same choice of 3 spots, get my water and walk to the door.  Walk in and say hi to the cashier, although I can give a rat\'s ass how her weekend was, walk to photo and say hi to the photo guy.  Punch in the combo to the back area.  Glance at the time clock.  Notice I\'m early as usual.  Stare aimlessly at the 2-year-old notices on the bulletin board, and wait until 3 minutes before my shift begins to punch in.  As I walk out the door, the manager will call my name, even though he knows I was out there for 10 minutes already and could have talked to me then instead of waiting until my right leg is on the sales floor.  Asshole.  Let the day begin...   '),(18402,1826527,'16,May,2004','2004-05-16 00:00:00','    I enjoyed this weekend a lot.  I spent most of the time with  Kim , and relaxed as well.  Today, we had to bring her brother to his home in Sparta, no not that Sparta.  We ended up staying there for a while chatting and then we ended up getting free burgers and snacks for dinner which was a bonus.  Nice house, nice day, free grub!    Upon our arrival home, I was able to use Kim\'s laptop to burn a mix cd for my car.  I tried going with a summery/upbeat mix.  Most of the stuff is pretty uninteresting to most folks, but here\'s how my setlist goes:  01. Feeling This: Blink 182 02. Into the Darkness: No Motiv 03. Sunny Side of the Street: Face to Face 04. Amber: 311 05. Guerilla Radio: Rage Against The Machine 06. It\'s Been A Blast: Hey Mercedes 07. Cool For One Night: Hot Rod Circuit 08. Blue Groove: Kid With Man Head 09. Does That Make Sense?: The Jealous Sound 10. You\'ve Got So Far To Go: Alkaline Trio 11. A Praise Chorus: Jimmy Eat World 12. Fat Lip: Sum 41 13. Photograph: Weezer 14. Big Brat: Phantom Planet 15. Warning: Incubus 16. Sell My Old Clothes, I\'m Off To Heaven: Saves The Day 17. Blue Sky: Dynamite Hack 18. Who\'s Gonna Save Us?: The Living End 19. J.A.R.: Green Day 20. The Page: Appleseed Cast 21. PS We\'ll Call You When We Get There: The Juliana Theory 22. Break It Off: Koufax 23. Intergalactic: Beastie Boys  Much of the stuff is old or whatever, but I don\'t care because no matter how many new cds I buy, I still get sick of all my cds.  I don\'t know what to do with myself since I don\'t have to work until 3 tomorrow.  I think I\'ll try to finish reading  The Corrections  finally!   '),(18403,1826527,'15,May,2004','2004-05-15 00:00:00','    I\'m home and not tired enough to go to bed or in the mood to play MLB 2005 or watch a dvd, so I\'m blogging again.  Went to lunch at Houlihan\'s and ate almost an entire plate of chicken nachos by myself before surrendering to my dinner salad.  A salad!  I was defeated by a salad.  It was good, but I couldn\'t eat anymore.  It consisted of garden greens, grilled chicken, tortilla strips, cherry tomatoes, shredded jack cheese, toasted pecans, and ranch dressing.  Unfortunately, I wouldn\'t be able to salvage it after having it wrapped up since we were going out.  Not about to leave it in my car, so I left it on the streets of Weehawken for a bum to eat at a later time.  Purchased new sneakers, undershirts, and 2 shirts.  Nothing major, but stuff that was needed to some extent.  Just enjoyed the day with Kim and not worrying about work or school.  Tomorrow I may wash my car, bowflex, and make a mix cd with the help of Kim\'s laptop.  Time for my crest whitestrips.   '),(18404,1826527,'13,May,2004','2004-05-13 00:00:00','    It\'s currently 86 outside and it\'s still sunny!  On a Saturday!  Welcome weekend...My weekends feel more complete now that school is over and I\'m not writing papers.  In addition, my work schedule has changed for the better (slightly).    Monday: 3-11 pm Tuesday: 9am-5pm Weds: 530am-230pm Thur: 9am-5pm Fri: 8am-5pm  So I got more hours, meaning a bit more money, although not at the rate I\'m getting paid.  But going in late on Monday is spectacular since I don\'t have to deal with the photolab on Monday morning from all the weekend photographers, and I get to extend my weekend by a bit.    Hope to do some shopping today since I struck out last night and came home with a belt of all things.  Need basics and new sneaks.  How about Orlando Hudson?  Man thank god for him because he\'s carried the team for the past 2 series.  Well, I\'ll finish getting ready to go out.  Enjoy the weekend.  P.S.-Congrats to Den\'s sister on graduating college!   '),(18405,1826527,'13,May,2004','2004-05-13 00:00:00','    Howdy all!!  I am being a very bad employee and blogging at work.  I have nothing else to do so I might as well take advantage of the cable connection I have here as opposed to my crappy dial up at home.  So I\'m thinking I may want to start looking for a new job.  I may be going back to work full time sometime soon and, unfortunately, I have not learned anything new in the mortgage department since I started back in august.  I like working here, but it is very frustrating because everyone but me is busy and I cannot stand sitting here and waiting for someone to come to me and say, \"Hey, can you copy this for me?\"  Plus, everyone is so busy and they wish they had more help, so why not just teach me how to do it?  Oh well.  Those are the breaks, right?  Hope everyone is doing ok.  Where the hell are you guys??   '),(18406,1826527,'12,May,2004','2004-05-12 00:00:00','    I have a few things to look forward to this summer, now that school is over.  Besides our \"vacation\" to my brother\'s house, I have the honor of attending the Face To Face farewell concert in Irving Plaza.  It\'s Sunday, August 22nd, and I am psyched because I never got the chance to see them live, and was really bummed when they called it quits last year, but they\'re getting together to do a final run.  Should be sweet.  Don\'t know who they\'re playing with, but I don\'t care.    After getting home from work today, I watched the Blue Jays game on ESPN...unfortunately.  They were down 1-0 in the 3rd when I got home, and proceeded to tie the game and take the lead 2-1 in the 5th I think.  The Royals tied it at 2, but the Jays took the lead back in the 7th.  Kerry Ligtenberg came in for the 8th and left with 2 outs and runners on the corners.  Terry Adams, their supposed closer of the moment, came in to get the 3rd out in the 8th.  The Jays couldn\'t do anything in the 8th as Delgado struck out with bases loaded, and then faltered in the 9th again.  Adams came back in to close out the game, but didn\'t have good control.  He fell behind 3-1 to Desi Relaford, who\'s batting .210 this year, and then surrendered a single to him.  Aaron Guiel grounded to first, and Relaford moved to 2nd.  Angel Berroa grounded to second and Relaford moved to 3rd with 2 outs.  Adams then walked Beltran to put runners on the corners and Mike Sweeney lined a ball past Eric Hinske to score both runs and win the game.  I watched the game for 3 hours and they lost like that!  It sucked.  The game must have been 4 hours long because first pitch was 2 pm.  Anyway, I was pissed and threw my name tag from work across the room, but soon got over it since I had a feeling Adams didn\'t have closer stuff...   '),(18407,1826527,'10,May,2004','2004-05-10 00:00:00','    Our blog had plastic surgery over the weekend apparently, at least the sign-in page.  It looks streamlined...So class 1 of 2 is over.  Presented my proposal/paper in class and the professor was very nice in not tearing it apart.  He\'s a good teacher, but I still don\'t think I\'ll be continuing in the program come September.  Now I\'ve got to do a final re-read of my paper for tomorrow\'s class, read an article, and set up an outline for my presentation of that paper for a 5 minute summary of my stuff.  At 8 PM tomorrow I\'ll be ready to drink my face off; however, I have to wake up at 4:30 for work on Wednesday morning.  Blast!  I\'m now going to complain about work, since it pisses me off more and more every Monday.  We just got a new photo lab in the store, and I\'m not trained on the new machines.  Somehow, I got stuck in the photo lab today for my last hour when I had lots of other work that I was supposed to do as part of my job position.  Also, last week, my manager tells me I won\'t be in the lab anymore because my job was to focus on inventory and orders.  So starting next week, I\'ll be working Monday nights, when there is someone else working to handle the photo lab.  I had this vision at work today telling me to leave my job and work down the shore this summer, while taking surfing lessons.  Probably won\'t happen, but very tempting.   '),(18408,1826527,'09,May,2004','2004-05-09 00:00:00','    So today was mother\'s day and I was supposed to spend some quality time with my mom but she was sleeping (she works the nite shift from 10-7) and I\'ve been studying accounting for about 5 hours.  Well we did get to eat lunch together... Anyway, I wouldn\'t have to study so much if I didn\'t miss 2 classes due to illness.  I\'m very sleepy right now and its not just from the studying.  My rude ass neighbors were fighting at 4am.  I don\'t really know what happened but from what I heard some kid was arguing with some other dude about his ex girlfriend in the other neighbors driveway.  So the other guy came out with his german shepard and basically told him to shut the f**k up and that the cops were coming.  I didn\'t hear much else but the dog barking until 4:45.  That\'s right.. I listened to a dog bark for 45 minutes in the middle of what was a great slumber until then.  Oh and by the way, I think the neighbors were fighting in my driveway which doesn\'t make sense since the guy was pissed the other neighbor was in his driveway.  Maybe he wanted to have the whole neighborhood involved.  Enough of my confused rambling.  I apologize if I made no sense what so ever.  I\'m getting ready for bed.   '),(18409,1826527,'08,May,2004','2004-05-08 00:00:00','    Yeah so I think I have officially gone insane.  I registered for another summer class which brings the total to two.  Two doesn\'t seem like a lot but it kills my summer for the most part... and my savings.  I have intro to jazz from 5/17 to 6/3, which is for 3 weeks, almost four hours a day.  The second summer class is urban anthropology which takes up 3 weeks in august and has the same days and hours as the other class.  After the class in august ends, I start the fall semester the following week.  Oh and I registered for 18 credits in the fall along with working 30+ hours a week.  It\'s poop but someone has to do it.  I\'m still unsure about my major but at this moment I am half way done with a finance major and I\'m beginning a math minor.  I figure the following things can be good about this:  1. I can still work at the bank but in a better paying position with a finance degree. 2. If I leave the bank for another finance job, I will have an edge over other people because of my math minor. 3. If I hate the finance thing, I can go the alternate route and be a math teacher.  Be very afraid for the kids if I go through with it. 4. There is a high demand for people who can do the finance thing and a huge demand for math teachers.  Yay for me!  Anyhoo, nothing really exciting is going on in jersey.  Then again, when is there something exciting happening here?   '),(18410,1826527,'05,May,2004','2004-05-05 00:00:00','    Dennis just informed me that today is cinco de mayo.  I hadn\'t noticed.  Been busy trying to finish my papers.  One\'s done(in my opinion)-that was the 20 pager.  Another needs some touch-ups-that\'s the research proposal, and the last one needs about 8 more pages-that\'s a 10 pager.  I have to say, I like writing, but not research papers.  It always seems like I put stuff into the paper just to make the limit for resources and page length, and the quality never seems that good.  I hope to finish all paper related stuff tomorrow night after work.  Also, both professors just notified me that I have to present the projects in a 20 minute session for my Monday class(the proposal and 10 pager) and a 5 minute blurb for my Tuesday class(the 20 pager).  I wasn\'t really expecting this, since neither of them mentioned this would be necessary the entire semester, so I\'m not about to put forth a great effort.  My efforts are and will be wasted entirely on writing.  Anything else they get from me is a privilege, biatch.    We saw Kill Bill Vol. 2 over the weekend, and I enjoyed it, not as much as the first, but still pretty captivating.  Quentin Tarantino\'s such a strange man, but I enjoyed most of his movies.  Most.  Glad to hear Matt\'s having good luck thus far with his school after-life.   '),(18411,1826527,'30,June,2004','2004-06-30 00:00:00','    Happy Birthday to Rob!  One right after another.  Who\'s next?  Anyway, hope you had fun.  I owe everyone a drink or something.   '),(18412,1826527,'29,June,2004','2004-06-29 00:00:00','    Happy Birthday Dennis!  Yup, June 29th it is, albeit almost over.  Hope you enjoyed your day.   '),(18413,1826527,'28,June,2004','2004-06-28 00:00:00','    Exactly one week after my interview there, I today received a letter from Bloomfield saying I didn\'t get the job.  I still don\'t even know what job it was.  They never told me any grade or school.  Fuck that shit.  It\'s exactly what happened last year in my interview.  They blew a bunch of smoke up my ass and gave me positive feedback and a day later they\'re sending me a letter.  Fuck all that.  I had an interview somewhere else today, and will be going back for a second interview in mid-July.  The interviewer was pretty nice and gave me some pointers for my call-back interview.  I really hope I get a job there.  It\'s a nice district and it\'s not here.  Afterwards, I got to spend some time with my sister-in-law, who is as cool as a sister-in-law could get.  I don\'t see her as much as I\'d like, so it was nice.  We talked and got some tacos for lunch and hung outside for a while before I left.  Of course when I come home, there\'s the letter waiting for me and it gets cloudy.  Anyway, that\'s my day.  I\'m tired from getting up early and driving and getting lost among farms.   '),(18414,1826527,'27,June,2004','2004-06-27 00:00:00','    Brian and I went to see  Dodgeball  yesterday at the Allwood for a very cool $4.  I think we were the only people in theater laughing at all.  We thought the movie was great but I guess the other movie goers did not share the same sense of humor.  At the start of the movie, Vince Vaughn\'s dog was licking him in his \"special area\" and then he said \"No grundle.\"  How great is that?!?  It reminded me of Brandon.  Speaking of Brandon, I wonder how he\'s doing....  Anyway, I highly recommend  Dodgeball .    Oh and I almost forgot, there was some old guy there that was taking the grandkids to see Shrek 2 or something and was arguing with the girl at the refreshment stand over a dollar.  One lousy, stinkin dollar.  I\'m sorry but he could have went to another theater and paid ten bucks a ticket in addition to highly priced snacks so he shouldn\'t have complained.  The girl was cool about it though.  She was like \"I don\'t know.  There\'s nothing I can do about it.\"  And he kept going on and on and she just shrugged her shoulders and moved on to the next customer which just happened to be me and Brian.  I gladly paid my $8 for a medium popcorn and soda.  I know it seems like a lot but it was really good popcorn and admisssion was cheap so it worked out to a very nice $17 for us.   '),(18415,1826527,'26,June,2004','2004-06-26 00:00:00','    Welcome weekend.  Please do not flee so fast this time...I ran for the first time in about a month on Thursday morning.  It was one of those things that I thought about doing on Wednesday night, and planned on it, but since I woke up at 5 for some reason, I decided to definitely do it when I finally rolled out of bed an hour later.  I didn\'t quite know where to go since I still had to get ready for work and all, but I did do it, and was happy with my efforts, but my knees were sore the next day and still a bit today.  That\'s sort of disheartening since I always stretch and warm up.  I\'m anticipating my interview Monday, and hoping things go well.  I even \"studied\" for my last interview and this one by brushing up on some concepts and things that weren\'t so fresh in my mind.  I have off from work after the interview, so while I\'m near the shore, I\'m contemplating hitting up a surf shop to hook myself up with a board and some accessories, but we\'ll see where the day takes us first.  Don\'t know what\'s going on this weekend, but don\'t want to stay home.    Watched  Simpsons  reruns the other day, and the one where Sideshow Bob runs for mayor was on.  For some reason the funniest part in the episode to me was when Bart is looking thru the phone book and sees the name \"A Gorilla\" listed...I\'m satisfied with the Sixers draft pick, but unimpressed with the trades that occurred.  Boring.  I guess the Lakers turned down the Shaq for Greg Buckner and Glenn Robinson proposal.   '),(18416,1826527,'22,June,2004','2004-06-22 00:00:00','    I almost vomited after seeing some rumors about the  urlLink sixers  and the upcoming nba draft.  The first one I saw was Allen Iverson to the Raptors for Vince Carter and the teams would swap first round picks.  That sickened me enough, but then I read one worse:  Sixers trade Samuel Dalembert, one of the few bright spots on the team to Chicago for Tyson Chandler and swap first round picks.  I think the Sixers would get the #3 pick, but I don\'t care.  According to Billy King, they are just rumors and hold no truth, but I\'ve heard that before.  I swear to god if they do any of those deals, I will drive to Philly and destroy Billy King\'s office.  Take that, rolodex and fax machine.  I\'m highly anticipating our week off.  Did I mention that in my previous blogs?  Well, it may be seen in just about every one from now until July 18th or whenever our fun begins.    One girl at work asked me today if I  really  want to get out of there, and I responded \"why, can you tell?\"  She sort of smiled and said \"yeah.\"  I guess maybe because I don\'t talk to anyone at work really and other little things.  I started wearing sneakers instead of shoes, since I ruined 2 pairs of shoes from getting scuffed and torn.  After punching in, I avoid the cashier and go directly to the stock room to waste some time(i.e. drinking my water, eating a granola bar, scratching my balls, \"looking\" for products to put on the shelf).  I found myself playing with a sponge ball while I was on register last night and bowling for employees.  I got one employee.  I\'ve been wearing a blank label over my name tag since I don\'t like it when people I don\'t know call me by my first name, or when my boss isn\'t working, I take off my name tag.  I can go on, but if I soon start wearing sandals and a hawaiian shirt to work, and gut a fish on the tps report, then you can say I have a problem.   '),(18417,1826527,'21,June,2004','2004-06-21 00:00:00','    Interview #1 completed.  Things seemed pretty good, but we\'ll find out within 3 weeks.  My other is next Monday at 9:30 am in  urlLink Howell .  No, not Mr. & Mrs. Thurston Howell III, but the town in Monmouth County.  Please tell me I can get one of these jobs.  I can\'t take much more of retail.  Got about another hour or so before I have to leave for work, and as usual, I do not want to go.  Something about stewing over the fact all morning makes things worse.    My brother was covering the U.S. Open all weekend out on Long Island.  It must have been riveting devoting four whole days to golf, especially when it\'s extremely windy.   '),(18418,1826527,'19,June,2004','2004-06-19 00:00:00','    Going to visit my grandmother today.  Also going to hit a ton of traffic on the way.  Tomorrow is father\'s day, so happy father\'s day to any father out there... Have an interview Monday and got a call for another one somewhere else either sometime next week or next Monday.  Wish me luck.  Gotta get out of retail.  For shizzle.   Can\'t wait till the Blue Jays play with all their starters.  That could happen late July.   '),(18419,1826527,'16,June,2004','2004-06-16 00:00:00','    Lakers: terminated.  mission: success.  Yes.    So I visited my old student teaching school and my teacher seems to be doing well.  Mostly everything\'s the same at the school.  She hooked me up with the personnel director\'s extension.  She was the principal of the school when I was student teaching, so hopefully she\'ll remember me and remember to give me a job.  It was nice seeing some of the teachers I got along with.  Then I was off to work.  Same old story.  Got home.  Bowflex.  Going to eat dinner.  Peace out.   '),(18420,1826527,'15,June,2004','2004-06-15 00:00:00','    My grandmother(my dad\'s mom) had been in the hospital for the past few weeks undergoing testing, etc.  They believed she had cancer while doing an examination.  The biopsy was \"suspicious,\" whatever that means, but didn\'t prove anything.  They made her undergo another exam and the biopsy was inconclusive.  They had gone over the treatment options which consisted of radiation 5 days a week and chemo one day a week.  My grandmother is 92 years old.  If you\'ve never met her, you would be amazed at how healthy and smart she is...While there, she also went thru 3 catscans for some reason, and the other day, they tried a different method of exam and did a biopsy by using a needle to get the cells.  This was inconclusive.  So they tell her she has cancer, but yesterday they sent her home because they don\'t know if she does or not.  They told her the only sure way to tell is surgically, but her doctor stepped in and halted that thought.  They won\'t be giving her the treatments, at least not yet, so she\'s home now.  I was going to visit her tomorrow and leave work early, but I\'ll see what my dad recommends...  I was also supposed to visit the teacher I did my student teaching with today, but my work schedule kind of interferes with that possibility, so I told her tomorrow morning I\'ll try to catch her before her school day starts.  Got my interview on Monday.  I\'m hopeful, but nervous because if I don\'t get this, then I don\'t have anyone else lined up with interviews.  At least last year I had a few interviews.  So since Kim\'s at a gym now, I\'m trying to keep up by exercising when she\'s there.  I\'ve been doing bowflex, but yesterday my arms were dead and I couldn\'t raise them for a half hour after I stopped.  It wasn\'t easy taking a shower and not being able to reach above your crotch.  Well, some would say that\'s a dream come true, but my goal was to cleanse myself and get ready for work.  Our vacation\'s about a month away!  I\'m excited just to do something because I have no money right now.  I just paid my credit card, car payment, doctor bill, bought my bro & sis-in-law an anniversary gift, and my dad a father\'s day gift.  And my insurance payment gets deducted on Thursday.   '),(18421,1826527,'14,June,2004','2004-06-14 00:00:00','    Day 2 of going to the gym...  I did pretty well for the first time on my own.  I pushed myself to keep going even though I was totally tired and kinda in pain.  I actually like the cardio part of the workout but the weights are frickin killer.  My arms are pretty sore.  Maybe if I stretch every five minutes they won\'t feel so bad :(   Anyway, as I was leaving the gym, there was a fire a block away.  Is that a sign?  I just hope the fire doesn\'t spread because I (gulp) like the gym.  Did I just say that?  Anyhoo, nothing new is going on... as usual.  Me and Bri went mini golfing yesterday and I won despite a terrible start.  I was really happy about that.  I guess I get excited over the little things.  We also exercised together on Saturday.  We went for a walk around the hood and then we took turns on his BowFlex.  That thing is pretty cool.  It actually helped get rid of the soreness I felt in my arms from my first day at the gym.    Gotta run.  Dinner\'s here...   '),(18422,1826527,'11,June,2004','2004-06-11 00:00:00','    I\'ve been in dormant blog mode this week for some reason.  I guess there hasn\'t been much to talk about besides work, and Kim blogged about the one event of the week, which was  Braid .  It\'s Friday, and I\'m home from work, although tired.  Got to shower in a minute and try to re-wake myself up after sleepwalking thru work since 5:30 am.  I feel like doing something this weekend, but I\'m not sure exactly what.  Just feel the need to get out for once...I don\'t know who won the game last night nor do I care.  One person at work insists on asking me every time he comes in whether I saw the Lakers game.  I respond every time by saying \"no, and I don\'t plan on watching the other games.\"  He then continues by telling me what happened like he\'s the orator for the associated press.  Lots of drama at work lately probably leading to someone I get along with being transferred out.  The mangagers just don\'t know how to fix problems among employees, but insist on transferring someone out rather than working at a solution.  Anyway, stop letting me ramble.  Enjoy the weekend biznatches!   '),(18423,1826527,'10,June,2004','2004-06-10 00:00:00','    Today was my first day at the gym and as if anyone cares I shall discuss what I did...  I had an appointment at 6:30 with Joyce, who was very nice.  She measured and weighed me (this was the worse part) and then she started me on this machine that is kinda like that gazelle thing you see on infomercials.  I did that for 15 minutes and wanted to drop.  After, she showed me what machines I should use and how to use them.  I won\'t get into each and every machine, but I think there were 6 total.  After the weights, I did 20 minutes on a treadmill which wasn\'t so bad but once I stopped I could have sworn I was walking really funny.  Anyhoo, it was quite the experience and since it wasn\'t so painful I might go back on Saturday.  If not Saturday, then definitely Monday.    Work is going okay.  I\'m learning more stuff now since I am officially a full time employee.  As a part of my new duties, I get to accept application packages from customers and go over everything with them.  I especially love seeing people who make $10k a month and buy brand new mansions all over Jersey.  I wouldn\'t know what to do with $10k a month.  We actually had one customer that made $70k a month.  I\'m not kidding.  We thought that was his salary for the year, but when we saw the W2\'s, we basically all crapped our pants.  Yeah and his house was worth 1.9 million.  Ridiculous. I love looking at the appraisals for the houses though.  Actually, I love imagining the house is mine and figuring out how to redecorate it.  There are lots of new houses being built in jersey.  It\'s amazing that nj is the most densely populated state and yet somehow we find room to keep building houses.  And as I look through appraisals, I find it cool that houses are so much cheaper/expensive in certain areas.  For instance, there was a beautiful brand new house (5 bedrooms, 4.5 baths, dining room, study, kitchen with dining area, finished basement on almost 2 acres of property, etc.) in Sparta.. the town.. not Dennis... that would be scary...  Anyway, it was worth $700,000 but I\'m pretty sure if you moved it to bergen county or around the clifton/montclair area, it would be worth twice as much easy.  I\'m going to stop rambling about houses now.   Hope everyone is doing well and eating lots of fried oreos.   '),(18424,1826527,'08,June,2004','2004-06-08 00:00:00','    Me and Bri went to see Braid last nite at Starland Ballroom in Sayreville.  The place was in the middle of nowhere.  Actually it was in the middle of this road with nothing but trees.  The place was smaller than everyone made it out to be.  They had some silly yet understandable rules about the 21 and under kids not being allowed near the bar.  Anyway, Braid played an hour and it kicked a$$!  I believe me and Brian are going to 2 more shows this summer: 311 and Face to Face.  Both should be fun, especially 311 since that show happens during our vacation.  Mmmm... vacation.  I wish I was on vacation right now, but I\'m stuck here at work until 5 and then heading to the gym for the first time today.  I\'m dreading it but I paid the money so I might as well go.   '),(18425,1826527,'07,June,2004','2004-06-07 00:00:00','    Greetings from Asbury Park!!  Wait, that\'s a Springsteen song.  Greetings from Chapel Hill.  That\'s better.  So I\'ve been running around here with a chicken without its head trying to get stuff ready for my conference.  This is supposed to be summer, but so far I\'ve had to prepare a poster, a talk, and my masters thesis all in the span of two weeks.  Crazy times. This weekend was indeed crazy as my friends and I decided to go on Absinthe binge this weekend.  For those of you who don\'t know Absinthe, it is a green liquor made in the Czech Republic.  It very potent and contains wormwood which is a hallucegenic, which is why it is not sold here in the US of A.  However, my friends brought some back from Europe and man I got pretty lit.  It didn\'t help that we were at a party and I kinda of tweeked, but it was fun.  Something I would recommend to try once. This week I\'m going to see The Shins in concert.  Pretty awesome band.  I advise the bloggers to pick up \"Chutes Too Narrow\" by them.  You\'ll be in for a treat. Bri-Memorial Day Tip (when relaying the ball from home, make sure you hit the cutoff!) Kev-at the NC state fair, we have fried oreos along with fried snickers.  Great for the Atkins diet Peace out and may the force be with you!!   PS: did anyone see the most metal moments on VH1.  Very funny.   '),(18426,1826527,'06,June,2004','2004-06-06 00:00:00','    Got my oil changed in my car today, but the big problem was that my car hasn\'t been starting correctly.  Anyway, I have to bring it back next Monday because it has a faulty fuel pump.  Then I have to get a loaner car for the day and go back again on Tuesday to pick my car up when it is ready(hopefully).  The service guys said Mazda was aware of the faulty fuel pump, so my question is: why wasn\'t I informed?  You know, because I drive the car.  So now I have nothing to do for 7 hours before the  Braid  show tonite.  I\'ll probably vacuum and clean my car since I didn\'t have the chance to do such things Friday.    Did I ever mention on the ole blog that I have an interview on June 21st?  I don\'t recall, but I do have an interview with Bloomfield for a leave replacement for next school year.  Hey, an interview is a huge step!  I haven\'t had one in about 9 months, so maybe my bad luck has worn off.    I\'m actually a little upset that I\'m going to miss game 7 of the Stanley Cup tonite.  Yeah, me upset over missing a hockey game?!  Off to do something non-constructive with my time.   '),(18427,1826527,'06,June,2004','2004-06-06 00:00:00','    Well, that clears up the oreo mystery.  Thanks.  I still don\'t know if I could eat one though.  I did just finish eating some mixed nuts, and while I found that I consumed 24% of my fat content for the day, I also took in 20% of my copper allowance.  Sweet, now I can self-manufacture pennies(i.e. when I make a deposit into the toilet, it can be taken literally now!).  Sorry for that gross pictorial.  I got home about an hour ago from the  urlLink Mets  game with my brother.  They actually won, 5-2.  It was a pretty good game if you take away the fact that it was 50 and windy.  We got to Shea in about 40 minutes, which is spectacular, especially when the Yanks are also playing a 1 o\'clock game at home.  My brother had hoped Smarty Jones would throw out the first pitch, but we got a worker from Gold\'s horseradish & mustard company instead.  Anyway, the game started with Juan Pierre bunting his way on and then getting thrown out at 2nd as Luis Castillo K\'d.  In the bottom of the 1st, Animal Piazza roped a 2-run double to get the Mets on the board.  The inning ended with the Mets up 3-0.  No runs scored for either team until the bottom of the 5th, when Piazza lined a homer 410 feet against a stiff inbound wind.  He drove in 4 of the 5 runs, and the game was quiet until the top of the 9th, when Braden Looper came in for the Mets.  He looked pretty awful and allowed 2 runs.  With 1 out and the bases loaded, he struck out Miguel Cabrera and got Conine to fly out.  Game over.  Mets win.    The ride home sent us through Korea, or at least a very big Korean population in Queens, before we realized we were stuck on the wrong side of the highway.  Had to u-turn to get back on track and hit the LIE.  Hit some slow traffic in midtown before trudging thru the tunnel.  All in all, took an hour and a half to get home.  Now I have to look forward to taking my car in for repairs tomorrow morning and not watching the NBA finals.  P.S.-Al Leiter\'s batting average is higher than Mike Cameron\'s.   '),(18428,1826527,'05,June,2004','2004-06-05 00:00:00','    Picture an oreo, that was dipped in the batter that is used to make funnel cakes, and deep fried.  It looks like a zeppole, but there is an oreo-y surprise on the inside... mmmmm...   '),(18429,1826527,'04,June,2004','2004-06-04 00:00:00','    Fried Oreos?!  I can\'t picture someone putting an Oreo in a pan with some butter and cooking it.  I don\'t know, tell me more...I was a little disappointed because I missed the end of the Tampa/Calgary game last night.  I saw up until 2:00 into OT, but fell asleep before anything happened.  A pretty good game based on what I saw.  Ended the week by working from 5:30-2:30 for our truck.  Actually anticipated cleaning my car afterwards, but there was nowhere to park my car when I got home in order to vacuum or wash it.  Speaking of my car-my  new  car-I have to bring it to the dealership on Monday morning because it isn\'t starting right.  The engine isn\'t turning over properly, so it\'s been taking 2 tries to start it.  Not bad considering I have just passed the 3,000 mile mark.  Are you serious? Yes, yes I am.  I got a bit of good news today when I checked my phone messages.  Bloomfield called me up about an interview for a teaching job for next year.  It\'s a leave replacement, so I have no idea how long it lasts, it probably doesn\'t offer benenfits, and if it\'s for the whole school year I won\'t have a job the next year again, etc.  I go at 11 am on Monday, June 21st, so hopefully I\'ll find out more then.    Shopping & lunch tomorrow with  Kim  & her mom.  Mets game with my bro Sunday, and Braid on Monday(after waiting for my car to get fixed).  Enjoy the weekend.   '),(18430,1826527,'03,June,2004','2004-06-03 00:00:00','    Greetings from the wonderful state of Connecticut.  Last nite marked the end of my second class on the way to my MBA.  I had my Economics (Micro & Macro) final, turned in my group paper, and presented our findings to the class.  I had to study my ass off Wednesday and and Thursday afternoon, before the exam, but it was worth it.  He somewhat graded our exams during our presentations, and since our presentations were just summaries of our papers, he was able to give us our grades last nite.  So, I got another \'A\'.  Three cheers for me.  And there is no rest for the wicked... I start my Marketing class on Tuesday.  That\'s 2 classes down, 15 more to go.  In other news, I spent Memorial Day Weekend in East Durham, NY (it\'s in the Catskills).  Myself and two friends of mine spent two nites camping and we spent the days at the 27th annual  urlLink East Durham Irish Festival .  It was a lot of fun.  There a lot of really good bands playing, but two really stood out:  urlLink Hair of the Dog  and  urlLink The Screaming Orphans .  Hair of the Dog is a really fun, upbeat Irish Folk-ish band.  And I was really impressed by The Screaming Orphans.  Ignoring the fact that they are rather attractive Irish girls with killer accents, they can really play.  They are a meld of many styles, you can hear some Irish folk, country, singer/songwriter, and even some Cranberries in their songs.  They were very enjoyable.  The rest of the time at the festival we spent drinking Guinness and wandering around.  Oh, and the food there was good too.  There was only one Irish food stand and the rest was your normal festival-type fare, but it was all really good.  Especially the fried oreos and the apple dumplings with ice cream...  Mmmmmm... fried oreo... mmmmmmm...  Ok, now that I\'ve wiped the drool from my chin...  I finally got to see Life of Agony a few weeks ago, and they were amazing.  Candiria openned for them and blew my mind.  I would not suggest Candiria to just anyone, they are definietly an aquired taste; with their fusion of grindcore, metal, rap, and jazz.  They are extremely unique, and thats what I love about them.  I picked up the new Fear Factory album (Archetype) the other day, and I haven\'t taken it out of my car\'s CD player since.  A few other things I am anticipating this summer are: the new Bad Religion album, the new Less Than Jake DVD, the new Chuck Palahniuk book, and my month and a half off from school starting July 9th.  Oh well, that\'s enough out of me, I really should be doing some work while I am here...   '),(18431,1826527,'03,June,2004','2004-06-03 00:00:00','    Today is the last day of my Intro to Jazz class.  I am somewhat saddened by this because I liked the class and the professor.  The best part was beating rush hour traffic coming home.  Now I officially start work again full time. Bleh.  I signed up at a gym yesterday.  I know what you\'re thinking, \"Kim? At a gym? Hahahahahaha!\"  But seriously I did join and it seems like a cool place.  It\'s a women\'s only club and I signed up with a friend from work which should give me the incentive to actually go.  On Tuesday, we\'re going to meet with a personal trainer to take our weight and measurements and then she\'ll tell us how to decrease those dreadful numbers.  I think we start on a weight training program that day.  Weights are fine as long as I don\'t look like one of the chicks from American Gladiators.    Braid show on Monday.  I\'m really excited for this show because I never got to see Braid play together (and I get to leave work at 3:30).  Actually, I\'ve seen 3/4 of Braid play at a Hey Mercedes show, but I guess Chris Broach makes all the difference with his screaming/singing.  Should be fun.  I heard from Katie that Brandon and John will be there.  Yet another Sam Goody reunion, but they\'re always fun so no complaints here.  Must go.  Boss may be getting suspicious of my excessive internet usage.   '),(18432,1826527,'01,June,2004','2004-06-01 00:00:00','    I currently hate the NBA.  Could there possibly be a worse finals matchup in my mind?  Well, maybe if it were the Bulls from Jordan\'s era vs. the Lakers of today.  I refuse to watch the finals.  And to top it off, the draft is garbage.  Who the hell is available for the draft?  Everyone already came out of college or high school.  The only thing I\'m interested in is seeing who\'s left unprotected in the expansion draft...  Keeping on the sports topic, the  urlLink Jays  put their 7th player on the disabled list.  Here\'s the list of people who are currently on the list: *Justin Miller-starting pitcher *Justin Speier-closing pitcher *Kevin Cash-starting catcher *Greg Myers-catcher/DH *Orlando Hudson-starting 2B *Chris Woodward-starting SS *Frank Catalanotto-starting LF  In addition, Roy Halladay, the Cy Young winner from last season was scratched from his start tonite due to shoulder stiffness.  Jason Kershner, a long reliever who I like a lot, started in his place.  He\'s never started a game before.  Also, Carlos Delgado, their starting 1st baseman, isn\'t playing because of a sore knee.  He may end up on the DL.  So that\'s my rant about sports.  Other than that, there was a tornado that touched down about 30 minutes away from me today.  That\'s interesting.   Kim  and I watched a show on Animal Planet called  Venom ER .  It\'s a pretty cool show about snakebite victims I\'d recommend if you\'ve got nothing better to do on a Tuesday night.  I\'m trying to stay up until 12:30 in order to watch Aqua Teen Hunger Force on Cartoon Network.  Michael, the photo guy at work, insists I watch the show, especially the episode at 12:15 tonite.  I usually don\'t watch solely because I am sleeping, but he urges me not to miss tonite\'s episode.  Watch it with me if you will.  Going to the Mets/Marlins game on Sunday.  Please pray for no rain!   '),(18433,1826527,'29,July,2004','2004-07-29 00:00:00','    Inventory is over, and apparently it went well.&nbsp; That means my manager gets a really nice bonus and I & my fellow workers get absolute shit.&nbsp; I\'m glad it\'s over, but I\'m tired as hell.&nbsp; I\'m about to go to bed because I worked 10 hours yesterday and had to be in at 7 and work until 4 today.&nbsp; I need some sleep.&nbsp; Anyway, there\'s about 70 movies I want to see that are all new.&nbsp; Let\'s see, there\'s:  The Bourne Supremecy ,  Garden State ,  The Village , and  Catwoman .&nbsp; I\'m only kidding about the last one, and that\'s only 3 movies I want to see, not 70.&nbsp; The only reason I would watch the last one is because  Kim Smith  is in the movie, and she\'s hot(sorry babe!).&nbsp;   I keep getting these credit card offers in the mail from Fleet and it\'s getting really annoying.&nbsp; I don\'t need any more debt, thanks.&nbsp; Oh, the Nets traded Kerry Kittles for a big ol\' sack of nothing.&nbsp; That\'s kind of funny.&nbsp; So who\'s their 2-guard?&nbsp; Rodney Rogers?&nbsp; Lucious Harris?&nbsp; I should talk, because the Sixers new point guard is Kevin Ollie.&nbsp; And they still have Big Dog and DC on the team.&nbsp; Why?!&nbsp; I\'m obviously crabby.&nbsp; It\'s off to bed I go.   '),(18434,1826527,'26,July,2004','2004-07-26 00:00:00','    In keeping with the annual tradition(?) of summarizing my vacation on the blog, I will soon detail my week off.&nbsp; It was a good week overall.&nbsp; I spent the week with  Kim , so that was the best part.&nbsp; We left last Tuesday and came home yesterday[Sunday].         &nbsp; 07.20 Tue&nbsp;&nbsp;&nbsp; I arrived at Kim\'s house at about 9 am.&nbsp; We headed to Friendly\'s to catch some breakfast, which was delectable-pancakes with apples on top.&nbsp; We hit the atm and then the parkway en route to  urlLink Delicious Orchards , which is in Colts Neck, and relatively close to my bro & sis-in-law\'s house in Point Pleasant.&nbsp; The place is basically a farmer\'s market, but they have the best breads, cookies, and donuts as well.&nbsp; The donuts are made with apple cider and they don\'t taste like grease.&nbsp; So we made a rather quick stop there because I hate grocery stores.&nbsp; We stocked up on fruits & veggies for the week and continued south on Route 34 to our destination.&nbsp; We arrived at about 12 and unpacked a bit.&nbsp; We decided to eat Mexican food, so we ordered some  Burro Burrito , which has damn good tacos.&nbsp; After devouring our lunch and Goya soda, we changed and went into the pool.&nbsp; The rest of the afternoon consisted of lounging on inflatable rafts in the pool while the sun dashed in and out of the clouds.&nbsp; For dinner, I tried grilling some burgers on the Fiesta grill, but 2 of the burgers decided to catch fire and the flames kept shooting sky high.&nbsp; The food turned out ok, but that ended the grilling right then and there for the week.&nbsp; Afterwards, I think we watched&nbsp;tv before turning in.  07.21 Wed&nbsp;&nbsp;&nbsp; We woke up without a clue what to do because Kim\'s friends,  Liz  and  Christine ,&nbsp;were supposed to drop in around 3.&nbsp;&nbsp;I wanted to go to the beach since it was still early, so we went and it was blazing.&nbsp; I ended up sweating out all of the lotion&nbsp;I applied and we didn\'t go in the water because the waves were rough.&nbsp; Some old guy&nbsp;next to us kept saying how cold the water was and everyone&nbsp;was&nbsp;mumbling about jellyfish.&nbsp; We thought the water was beautiful, but didn\'t go in because of the high waves.&nbsp; We&nbsp;went back to the house after about an hour to go in the pool.&nbsp; Liz &&nbsp;Christine arrived around 3:30 or so and we stayed in the pool for a while.&nbsp; We went to&nbsp;dinner at Friday\'s and came home to watch Pirates of the Caribbean on dvd.  07.22 Thu&nbsp;&nbsp;&nbsp; The four of us went to the beach on a cooler and breezier day.&nbsp; We ate at the snack bar by the beach and enjoyed getting obliterated by the waves.&nbsp; We all took turns&nbsp;falling on our asses and getting sucked back into the sea.&nbsp; All in all, a very fun time.&nbsp; We were all tired when we got home, however, and Liz & Christine had to drive home, so we all cleaned off and they went on their way.&nbsp; Kim & I decided to use our year-old gift certificates to the Loews theater&nbsp;&nbsp;and catch  I, Robot .&nbsp;&nbsp;We both enjoyed it to my brother\'s&nbsp;recommendation, and&nbsp;brought home a box of KFC chicken fingers and mashed potatoes to eat with our fresh corn on the&nbsp;cob from the orchard.  07.23 Fri&nbsp;&nbsp;&nbsp; I intended on visiting the&nbsp;Barnegat&nbsp;lighthouse to catch a glimpse of some good scenery and finally take some pictures, but it was rainy and overcast during the day.&nbsp; I really wanted to go because I have this bad habit of never taking pictures,&nbsp;but so the&nbsp;trend continued.&nbsp; We went to the boardwalk since we hadn\'t gone yet and because I had a bag of quarters to burn at the arcades, but it was&nbsp;terribly crowded with trash and thuggery.&nbsp; We&nbsp;got in a few minutes of the original Simpsons arcade game and some pinball, but otherwise&nbsp;were afraid of contracting an STD&nbsp;from the crowds.&nbsp;&nbsp;We played&nbsp;some mini golf&nbsp;and got stuck behind&nbsp;Orca\'s family of 7 children and 11 total&nbsp;members.&nbsp; They took their sweet-(fat)ass time even though there were two pairs of people behind them who they could have easily let pass.&nbsp; Anyway, it began raining, so we finished and I wanted to find Manasquan inlet since I heard it was great for surfing, but the directions I remember hearing said go past Jenkinson\'s aquarium and you\'ll run into the inlet.&nbsp; Unfortunately, I didn\'t,&nbsp;and&nbsp;didn\'t feel like looking for it&nbsp;as the rain fell harder.&nbsp; I think we went to go buy chicken for dinner and watched the Usual Suspects dvd.&nbsp; Kim made an excellent&nbsp;rosemary chicken dish with the rest of the corn on the cob and rice-a-roni!&nbsp; Yum.&nbsp; We attempted to watch Meatballs and Kingpin on dvd, but didn\'t get thru either of them&nbsp;due to our lack&nbsp;of laughter.  07.24 Sat&nbsp;&nbsp;&nbsp; Today was really our last day because we planned on leaving early Sunday to beat traffic home.&nbsp; I wanted to jump in the pool one last time, but it was in the upper 60s and cloudy until late afternoon, so that didn\'t happen.&nbsp; We ended up doing some laundry and cleaning up.&nbsp; We got ready for the  urlLink 311  show at the  urlLink PNC Arts Center &nbsp;&nbsp;that nite.&nbsp; We left around 5:30 or so to get some dinner beforehand, but just weren\'t hungry, so we decided to get to the place and eat there.&nbsp; We arrived around 6 and parked light years away and walked to the arts center thru whores, drunks, and guys who thought they looked like Spicoli from Fast Times.&nbsp; The rain was going to hold off at least, and it wasn\'t too hot.&nbsp; The show started at 6:35 with  Medeski, Martin, & Wood , who were pretty good I thought, and the limited crowd that was seated and not tailgating thought so too.&nbsp;  The Roots  came on next, and more people filtered in to see them.&nbsp; They were awesome musicians, and played for a good hour.&nbsp; 311 came on about 9:10 and played for 2 hours!&nbsp; They were as good as the last time we saw them.&nbsp; They sound good live and are pretty personable with the crowd.&nbsp; The only negative was the abundance of pot smoke in our faces and the couple seated in front of us who insisted on turning around and staring every minute at us or whoever moved in the row we were in.&nbsp; There was also a couple behind us who were making out or \"fucking snuggling\" as Kim put it so amusingly the entire show.&nbsp; It was pretty sickening.&nbsp; After the long walk back to the car, we got out relatively smoothly and brought massive headaches home with us.&nbsp;   07.25 Sun&nbsp;&nbsp;&nbsp; We finished packing and fed the fish and drove to get gas and breakfast sandwiches at Dunkin Donuts before driving for the parkway north.&nbsp; We left around 12:15 and got back to the home about an hour later.&nbsp; Disappointed to be home, but happy we missed traffic that would have built up as the day moved on.&nbsp;   Today sucks as it\'s back to work.&nbsp; Our inventory is this week and I could care less, even though it\'s supposed to be my responsibility.&nbsp; I\'m looking for a new job, since I realize the teaching thing is just going to kill my slowly.&nbsp; I need to find myself and figure out what to do.&nbsp; The only thing I know now is that the current place of employment is not where I want to be at the end of August.&nbsp; I applied online last nite to a few hotels, which I thought would be more interesting and I might have a chance to work my way up.&nbsp; In addition, I also investigated&nbsp;the postal service since my grandfather was a postman forever, but they don\'t have any current openings anywhere near here.&nbsp; If you have any ideas for me,  urlLink let me know , please.   '),(18435,1826527,'25,July,2004','2004-07-25 00:00:00','    Hello everyone,  It is 9:00 on Sunday and I\'m at the lab procrastinating finsihing up one of my questions on my doctoral qualifying exam.&nbsp; This exam sucks ass since it consists of 3 reserach questions which need to be completed by this Friday.&nbsp; So I\'m getting pretty stressed.  I came back from Vancouver a couple of weeks ago from a conference and I had a blast.&nbsp; The West Coast definetly kicks ass and hopefully, if I pass these exams, I can enjoy either Portland (Oregon Health Sciences U) or Seatlle (Wash. U.) as my permament address.  I found out yesterday that Franz Ferdinand is playing at the Cat\'s Cradle in Chapel Hill.&nbsp; I am so psyched!!&nbsp; This club only holds around 500 and has good acoustics.&nbsp; Hopefully, tix will not sell out.&nbsp; My friends and I saw Bourne Supremacy, which was really good (1st one was better though).&nbsp; Thankfully, we were able to use our veto power over the girls\' wish to see Catwoman.&nbsp;  In other news, we got a beach house for our conference in San Diego in October.&nbsp; Pretty sweet, right on Mission Beach .&nbsp; The best thing about San Diego, you\'ll only 20 minutes from Tijuana.&nbsp; I already told my boss, my attendance at the conference will probably minimal at best.&nbsp; More time to the enjoy the local culture (tequilla, beach, tequilla, zoo, tequilla, etc).&nbsp; Anyway, back to work.&nbsp; Hope everyone is faring well.   Later,  Dennis   '),(18436,1826527,'19,July,2004','2004-07-19 00:00:00','    If you haven\'t read my last post, then please read the final paragraph I wrote, and then I\'ll tell you I called Howell again today.&nbsp; Let\'s just leave it at that.&nbsp; Leave for \"vacation\" tomorrow, and that\'s about all I care about right now.&nbsp; When I get back, I will contemplate leaving this area and finding a new job, because obviously teaching is a big fat joke right now.&nbsp; I can deal with getting rejected for a job because I\'ve been thru it for 2 years now, but when someone promises a real chance at getting a job, and doesn\'t have the balls to call you, and you have to find out for yourself that the jobs suddenly aren\'t available, it\'s absolute bullshit.   '),(18437,1826527,'17,July,2004','2004-07-17 00:00:00','    Salutations to you, blog.&nbsp; It\'s been a while, and yet I still have nothing to write about.&nbsp; Well, ok, I do I guess.&nbsp; It\'s officially our vacation.&nbsp; That\'s right, no work until next Monday.&nbsp; Going to Point Pleasant from Tuesday thru Sunday.&nbsp; Should be fun, even though the weather forecast is not so good right now.&nbsp; There\'s plenty of stuff to do-beach, pool, boardwalk, movies, arcade, six flags, aquarium, barnegat lighthouse, driving down route 35.&nbsp; If it rains a lot, then it stinks, but at least there\'s some stuff to do down there.&nbsp; Next Saturday is the 311 concert at the arts center, and I\'m psyched for that.&nbsp; I think that might be the first concert I\'m seeing there since third eye blind.&nbsp; Remember that, Kev?&nbsp; How many drinks did we have before ordering tickets?&nbsp; Anyway, don\'t know what to do  this  weekend, however.&nbsp; Nice today, but not tomorrow.&nbsp; Maybe it\'s a good weekend for a movie, walk in the park, packing I guess would be helpful, too.&nbsp;  &nbsp; Just finished mowing the lawn, and boy, was that a mistake!&nbsp; The ground was still wet from showers yesterday, and all the grass that was cut ended up sticking to my legs.&nbsp; I felt like I had the legs of the incredible hulk, except they weren\'t muscular, and my pants weren\'t tattered, and I was wearing shoes.&nbsp; Apparently, my picture can no longer be used over on the right side of the blog for some reason.&nbsp; I guess freeservers wants me to pay for their website tools just to post my pictures on it.&nbsp; That\'s not gonna happen. &nbsp; Still haven\'t heard from Howell about my second interview, and that isn\'t cool.&nbsp; I\'m gonna call again on Monday, but I hope they don\'t pull some crap about there being no jobs suddenly.&nbsp; Hate retail.&nbsp; Cannot go on anymore.   '),(18438,1826527,'12,July,2004','2004-07-12 00:00:00','    This weekend was probably the best one I have had in a long time.  It was very relaxing, fun, and as always, waaaaay to fast.  Brian and I went down to Point Pleasant to spend the weekend with Kevin and Rachel.  On Saturday, we went out to eat at this bar restaurant place called Frankie\'s.  We had awesome nachos and, I think, an okay dinner.  After we ate, we went back to their place and we played this game about movie trivia which I really sucked at.  Oh well.  On Sunday, we all went down to Brick beach.  The water was beautiful... warm and clean.  I know what you\'re thinking.  New Jersey shore has clean water?  Yes, but it\'s not like Carribbean clear or anything.  After, we went to the butcher to pick up chicken and ribs for dinner.  When we got back to their house, we jumped in the pool for a while and then we ate.  We also walked to Dairy Queen for some ice cream.  Soon after, we left to face the wonderful Parkway.  Even though we left around 9:30 pm, we still hit a lot of traffic.  It was mostly stop and go, but annoying nevertheless.    I can\'t wait until next week when we are down in Point Pleasant for the week.  I need to get away from North Jersey and work.  Boo to work :(   '),(18439,1826527,'04,July,2004','2004-07-04 00:00:00','    Happy July 4th, or Independence Day as it\'s been formerly known.  I\'m not one of those history appreciating folks, but it seems every year, the term Independence Day is used less, except when they air the movie of the same name on Fox tonite.  But that doesn\'t count because Bill Pullman makes an awful president anyway.  So the day goes, where everyone barbeques and drinks.  Nighttime comes and some jackass neighbor insists on shooting fireworks onto your house/car/hair because he\'s too drunk to properly aim them  away .  As I was discussing with Kim, there\'s something that kills me about fireworks.  I don\'t hate them, but I don\'t love them.  I don\'t enjoy getting bitten by mosquitos while watching probably the worst display in America(Bloomfield), and I can\'t tolerate watching them on tv[honestly, I\'d rather watch reruns of Match Game on the gameshow network].  I guess it would be fascinating if you owned a boat and were on the ocean or the Hudson while they were shooting.  Anyway, enjoy your Independence Day for whatever reason.  I\'m going to enjoy mine since I\'m independent from work today.   '),(18440,1826527,'03,July,2004','2004-07-03 00:00:00','    I have yet to welcome July with a blog, so here\'s one.  Tomorrow\'s the 4th, and not much planned for the weekend.  Tonight I\'m actually taking my mom out for her mother\'s day dinner.  What, you ask...Mother\'s day in July?  Well, I had promised her a dinner as her mother\'s day gift, but we never had a chance to go out until now I guess.  I\'m currently watching tennis.  Maria Sharapova vs. Serena Williams.  Sharapova is serving for the set, and I really hope she wins.   {Pauses to watch}    Wow!  She won.  I\'ve never seen a 17-year old woman hit a tennis ball so hard.  That\'s cool...Tomorrow\'s probably a bbq of some sort at my house, but nothing major.  Just the small family I believe.   '),(18441,1826527,'11,August,2004','2004-08-11 00:00:00','    Went to the mall with  Kim  for the first time in a while tonite.  Ate \"dinner\" at Nathan\'s, and it was deplorable.  She got some clothes for work and I ended up getting 2 baseball caps-a  Jays  & a  urlLink Mets  cap-for $15.  Not bad considering the price of most caps today.  I\'m tired because I went to another store to help unload their truck at 7 am, and had to go to my store afterwards to work.  My manager made me do pointless work from 3:30 to 5, even though I started work early.  He was complaining yesterday that he has no overtime hours, yet he makes me stay and get 2 hours of overtime for something he could have done himself since he\'s there until 10 tonite.  I was really close to pulling a  Milton  today, but didn\'t.  There\'s always tomorrow.  Probably gonna go sleep soon.   '),(18442,1826527,'08,August,2004','2004-08-08 00:00:00','    Yesterday was a long day.  I had to go to work at 5:30 to help unload truck.  I stayed until 9:30 and came home.  I then tried to nap, but maybe mustered an hour of sleep out of 2 hours in bed.  I showered and changed and  urlLink Kim  called.  She had to work as well from 8-1.  Boo to working on Saturday.  Anyway, we were to go down to my brother\'s to go to a  urlLink Blueclaws  game with them and meet up with my uncle.  We ate some  Boston Market  for lunch and she drove, thank god, because of my lack of sleep.  We didn\'t hit much traffic until we got to Asbury Park, but that didn\'t matter much since our exit was a mile ahead.  We arrived at my brother\'s house to find him working on a stoop by the back door with his father-in-law.  They had been working since 11:30 and it was now 3:30 or so.  They continued working for another 2 hours roughly, and we all sat down to some grilled burgers and corn before scurrying over to the field.  The game was pretty terrible as the Blueclaws committed 5 errors.  The Bombers were leading 6-1 going into the bottom of the 9th with only 5 hits.  The Claws scored 3 runs in the bottom of the inning and lost 6-4.  Afterwards, there was a rather intriguing fireworks display at the field.  Some stuff we hadn\'t seen before.  We headed back to my brother\'s house and got back home a little after midnight.  I am glad today didn\'t start out with work at 5:30, so I\'m pretty happy today.   '),(18443,1826527,'06,August,2004','2004-08-06 00:00:00','    Hello everyone, I\'ve finished my qualifying exams.  If everything goes well, it should be the last exam I ever have to take.  This past week, I\'ve been trying to catch up on work/sleep/etc.  I\'ve also been listening to a lot of  The Smiths  and  Joy Division  and realizing how good both these bands were.  Morrissey is touring in NC this summer so I\'m definetly going to try to catch him out.   Saw  The Village  the other day and t was not as bad as I thought it was going to be.  So far the best summer movies I have seen are  Spiderman 2,  Farenheit 911,  and  The Bourne Supremacy .   Also, bought  Franz Ferdinand  tix.  Going to see thme 9/11 at the Cats Cradle.  Should be a pretty kick ass show.  BTW,I was at Best Buy and I saw that  Knight Rider  season 1 is out.  Ahh, Hasselehoff at his absolute acting prime.  later   '),(18444,1826527,'04,August,2004','2004-08-04 00:00:00','    I\'m actually impressed.  The  urlLink Sixers  made a trade and got someone who wasn\'t washed up.  They got  Corliss Williamson  from Detroit for  DC .  I thought Coleman had a few good seasons with the sixers, but last season was just a wash because of his injury.  Now they have three good, strong forwards in Williamson,  Skinner  and  Dalembert .  Unfortunately, they still have  Big Dog , who will hopefully be traded soon as well.  In related Detroit Pistons news, they signed  Horace Jenkins  today.  Who?  You may ask...Well, he was a stellar guard who played for  urlLink William Paterson University  and led them to the Final 4, but never got a break.  Since then, he\'s been tearing up the Italian League and leading there in scoring.  Good luck, Horace.  Hope you make the cut!  Kim and I saw the  Bourne Supremacy  tonite, and would strongly recommend seeing it.  We both thought it was better than the first.  Good chase scenes and good flow overall.  Still must see Garden State.   '),(18445,1826527,'03,August,2004','2004-08-03 00:00:00','    Brian and I went to see  The Bourne Supremacy  tonight and I must urge everyone to go see it. As in the first movie, there is a lot of shaky camera action but it didn\'t really bother me. Everything about the movie was incredible: Matt Damon, the chase scenes, the fights, etc. The sequel was actually better than the original which happens very rarely. Matt Damon never really struck me as a spy/action hero guy, but I must say that after watching the first movie again and this one, I think he was perfect for the part and he seems tons more comfortable with the role than he was in  The Bourne Identity .  Anyhoo, classes start again in a month and I am not looking forward to going back. I checked up the prices of some of my textbooks and one book costs $200. And I\'m pretty sure that when I sell it back one of two things will happen: 1. they will give me five bucks for it, or 2. not take it back at all. My luck, it\'s going to be the later.  One two years of this crap left.   '),(18446,1826527,'03,August,2004','2004-08-03 00:00:00','    Kim just informed me that  Ozma  broke up.  Their last cd was pretty terrible, but their first two were fun.  Plus, they gave us a good show at Krome a few years back.  Add to the list of many bands that have broken up.  I\'ve been working on my  urlLink webpage  a bit since yesterday.  I haven\'t changed much, but I\'m just fixing up a few things and making it lot a little fresher.  May go see the  Bourne Supremecy  tomorrow with Kim.  I\'ve only been home for about 2 weeks, and I\'m already sick of my house.  I don\'t want to think of what it\'ll be like in the winter.  Ugh.  winter.  Anyway, my manager keeps talking promotion to me, which would make me an assistant manager, but I really don\'t want it.  I\'m already not enjoying my job, so why would I want to spend more time there, and have a crappy schedule.  Yeah I get more money, but I don\'t know if it\'s worth it.  I\'m not making a decision until the summer\'s almost over.   '),(18447,1826527,'02,August,2004','2004-08-02 00:00:00','    Welcome to August, the month where I still whine about summer ending and going back to school in  September , even though I\'m not in school anymore, and it\'s only August 2nd.  I guess old habits are tough to break.  Anyway, this is it.  It\'s the final chance for finding a teaching job before another new school year begins.  Chances are, it won\'t happen, but I\'m way past feeling sorry for myself.  In some ways I find it to be some sort of sign telling me not to do it, which would have been more helpful had I known 3 years ago so I could have graduated without going another year for teaching.  So life continues mocking me and everyone else.  I hope to visit Point Pleasant at least 2 more times before summer\'s over, as well as getting up to Boston if all works out ok.  Should make for an enjoyable month at least.  Kim & I already want to go to Curacao for our next vacation, whenever the hell that may be.  First comes job, then money, then paying off debt, then saving money, then vacation.  In my world, conversely, the pattern goes: vacation, hopefully job, never having money, you only live once, why not enjoy yourself and be broke before you have major responsibilities.   Ramble on.  I finally used my aim account again, but it made me upgrade before I did anything, and I haven\'t noticed much difference.  Saw  The Village  on Saturday evening, but wasn\'t impressed.  I thought it was slow and I knew what the \"twist\" was before it was revealed at the end.  I\'m not a film critic either, but usually I\'m easy to appease.  I\'ve watched some truly awful movies and didn\'t complain too much, so compared to the likes of  Going Overboard  or, say,  Star Wars Episode I , it was pretty good.  The idea was interesting, but the movie did not do it much justice.  Oh well.  Two more movies to catch before the end of next weekend that shouldn\'t disappoint.   I purchased the new Hives cd and the My Chemical Romance cd the other day and like both of them, but I also purchased the Roots new cd and was disappointed because the show was so good compared to the album.  I must say, I still enjoy music, but my knowledge and purchasing has dropped off since my days at the music store ended 2 years ago.  I don\'t think there will be a correlation with me buying drugs at my current place of employment and a decline after I leave however.  I\'m in the mood for Doritos.   '),(18448,4059083,'31,July,2004','2004-07-31 00:00:00','   As the early-80\'s raced into the mid-80\'s word began to spread far and wide about the madcap exploits of  THE VEGA$ CLUB . Well, not too far and wide, but we attracted the attention of a good friend and co-worker named Leung (pronounced Lang, rhymes with Bang!). Before we could make him a full-fledged member, we had to bring him along on a trip on a probationary basis. On this trip Roger had arranged accommodations at the Hotel Continental. I understand in it\'s later years that it turned into quite a dump, but at the time (1985ish) it was a perfectly nice place to stay.  We arrived at the hotel, got checked in, relaxed a bit in the room, and gave Leung a brief lesson on what to expect in the casino, and a few pointers in blackjack etiquette so as to not to embarrass himself and us. (don\'t hand your cash directly to the dealer, one hand only on the cards, pointers on how to slip your cards under your chips without scooting your bet out of the betting circle, tipping, etc. He\'d memorized the \"basic strategy\" card, so we were ready to go!  As luck would have it, there was a dead $2 blackjack table near the entrance to the casino, just beckoning to us! Leung, Roger, Charley, and me took our seats and each placed our $20s in front of us for change. As the first hand was dealt, the waitress appeared and took our orders. At this point I was back to beer. (As of this writing, July 31, 2004, I have yet to have another  Greyhound ). Leung ordered a  Screwdriver , Roger was either in his  Fuzzy Navel  or  Grasshopper  phase at this point, and Charley was still trying to get a PLAIN coke. A few hands were dealt, things were progressing fine. Leung was getting the hang of things. The cocktails arrived, and as Leung reached forward to pick up his cards, his arm hit his full  Screwdriver  spilling it all over his cards, chips, and dripping into the dealers chip rack! \"Oh Shit!\" said the dealer, yelling over her shoulder for the floor supervisor to bring over a towel. They got things pretty well cleaned up, we got a new deck of cards in play, and the rest of the trip went well. It\'s just lucky that we\'re all nice guys, otherwise that little stunt would have forever banished Leung from entering the club.  That may have been the same trip where we discovered that Bob Stupak was a big fat liar. Bob Stupak owned \"Vegas World\" a hotel/casino  barely    on the strip, just north of the  urlLink Sahara , where the  urlLink Stratosphere  now stands. For years he\'d been cluttering up all the mailboxes in Southern California (hell, probably the whole U.S.) with vacation offers that seemed too good to be true. The deal was something like this: For only $398 up front, you\'d get 3 nights hotel accommodations, plus $400 in casino play, free Keno tickets, free cocktails, and a fabulous gift! For awhile it was billed as \"The Free Vacation\" then the courts intervened, lawsuits were filed, and it became \"The Virtually Free Vacation\". Miraculously  THE VEGA$ CLUB  never took him up on his vacation offer, but we did find ourselves in his outer space themed casino one afternoon. For the walk up customers, he had a lower priced miracle offer. For $80 you\'d get $120 in casino play, plus $20 in slot machine tokens, a free deck of cards, and Zsa Zsa Gabor\'s cocktail ring. \"How could that be???\" we wondered. We pooled our money and got one entry.  Catch #1-- You weren\'t given your chips right away, you got $10 per hour over the course of 12 hours!!! Catch #2-- The slot tokens only worked on \"Very Special\" jackpot machines, and Catch #3-- you only got your cards and Zsa Zsa\'s ring after the 12th hour of being on the property.  To add insult to injury, even the casino play chips were a rip-off! Say you put your $10 worth of chips on a blackjack table, if you lost, they took your chips. If you WON, they took your \"special\" chips and replaced them with $10 in real chips. \"Hey, If I win a $10 hand of Blackjack, shouldn\'t I end up with $20???\" \"Sorry kid, not at Vegas World. I\'ll see you next hour!\"  Needless to say, we didn\'t even make it to hour #2 and Zsa Zsa got to keep her ring!  In the next installment... Hank discovers Craps!  '),(18449,4059083,'30,July,2004','2004-07-30 00:00:00','   As I explained in my last entry, due to the passage of time, and the effects of alcohol on the brain, I don\'t have real clear memories of those first individual trips to VEGA$, just a conglomeration of individual exploits, high jinks, monkeyshines, etc. In those early years we tended to stay in modest, but perfectly acceptable accommodations like  urlLink Circus Circus Manor  (a group of 3 story buildings adjacent to the massive CC towers, featuring large, quiet rooms), the Sundance Hotel (now  urlLink Fitzgeralds , downtown) the The Hotel Continental (now  urlLink Terrible\'s , off the strip).  Finding the perfect casino isn\'t as easy as it sounds. You want to feel at home... Wander into  urlLink Caesars Palace , and you feel like you should be wearing a suit. Spend too much time at  urlLink The Gold Spike  and you need to take a shower. And it\'s not just finding a good middle-class joint either. It shouldn\'t be too big (like  urlLink Bally\'s ). It must have low-limit table games, and most importantly it needs to be lucky! I have entered many casinos that seem to be hospitable, but for whatever reason the luck isn\'t there.  urlLink  The Orleans   is just such a casino. In recent years, my wife and I have stayed there a few times. The rooms are beautiful, very reasonably priced, the casino is friendly, but for the life of me, I can\'t win a god-damned dime there!  Anyway, back to  THE VEGA$ CLUB  in the early 80\'s. One afternoon we happened upon a small casino on the heart of the strip called  The Castaways . It was a great location (so great, that a few years later Steve Wynn bought it, leveled it, and built  urlLink  The Mirage   in it\'s place!), good cocktail service, and two $1 blackjack tables open \'round the clock! I think originally the theme of the place was supposed to convey a feeling of being on a lush island in the tropics, but by 1984ish, the only hint of tropicality was that the dealers wore Hawaiian shirts. The rest of the place was pure, no-nonsense casino. We returned to The Castaways many times during those early years. The dealers were amazingly friendly, even at the $1 tables! Cocktail service was prompt (each BJ table had a little \"cricket clicker\" that would summon a waitress in short order!). Of course, we were looking to break the bank, each of us fancying ourselves as young  urlLink Edward Thorpes . Unfortunately, that never happened, but at $1 a hand, it was a nice way to enjoy the company of friends, and get ridiculously inebriated. One memorable night, Roger and I decided to switch from boring old beer to something more mature. Martinis sounded pretty classy, but sadly the well Martinis served to the $1 players at The Castaways tasted like kerosene. After trying a few different things, Roger settled on the Tequila Sunrise, and my poison became The Greyhound (vodka and grapefruit juice). Charley was our designated driver and stuck with Coke. Whenever he ordered, the waitresses would act like no one in the history of Vegas had ever ordered a Coke before, \"PLAIN Coke???\" they\'d ask, dumbfounded. From then on, whenever any of us wanted a Coke, we would order it as a PLAIN Coke. I think that Roger and I managed to polish off about 15 Greyhounds/Sunrises apiece. I do vaguely remember asking the dealer, \"SO, DO PEOPLE EVER GET KICKED OUTTA HERE FOR BEING TOO DRUNK?\" She gave me a hard stare and icily said, \"Yep. All the Time.\" Roger suggested that maybe it was time we call it a night. That trip we were staying at the Sundance downtown, and I remember that being one of the longest rides of my life. In fact, when faced with the prospect of going round and round and round up into the parking structure, Roger and I opted to get out on the street. I still remember the elevator ride back up to our room. We both occupied opposite corners of the elevator car, and as we went up, Roger slowly slid down the mirrored wall onto the floor! The next day, we had to go home, and BOY what a miserable drive across the desert that was. Every time I saw a  urlLink  Greyhound   bus, I felt like puking!   The Castaways  closed its doors on June 20, 1987. There are still a few small places left, off the strip, or downtown. But I\'ve yet to find one that had the unpretentiousness, friendliness, prompt cocktail service, and that elusive good luck. It\'s not like any of us ever broke the bank, it was just a nice place to  slowly  lose a few dollars, and have a good time along the way.  In the next installment... THE VEGA$ CLUB expands its ranks!  '),(18450,4059083,'29,July,2004','2004-07-29 00:00:00','   A year or so goes by, and things are changing at the credit card authorization center. Charley has resigned, although we remain good friends. When his supervisor asked him why he was quitting, Charley said that he needed more time to relax. Keep in mind that this was a 4 hour per day data-entry job. This is a trend that he would continue for some time: get a job for a few months, amass a couple thousand dollars, then quit in order to relax.     Eventually I get a small promotion to the \"Security and Fraud\" department of the call center. What this means is that if someone attempts to use a card that has been reported as stolen (or if they\'re a real deadbeat) then the call gets transferred to me. I advise the merchant to confiscate the card, cut in half, and send it in for a $50 reward. I worked the swing shift and most of the calls were from restaurants who didn\'t want to be so rude as to confiscate their valued customer\'s cards. What did I care? I was making a lousy $6 an hour! As the evening wore into night, we started getting calls from the telephone sex operators. Some of my less couth co-workers would try to get these ladies to give them a freebie, but to my knowledge, none were forthcoming.     ANYWAY, shortly after my promotion, the company was sold to a huge corporation called First Data Resources ( urlLink FDR ) and the security/fraud department was promptly laid off! Meanwhile the rest of the call center was moved to a smaller (cheaper) location in nearby Santa Ana. I took Charley\'s lead, signed up for unemployment, and relaxed! A few months later they called me back, and asked if I was interested in being a lowly operator again. It wasn\'t a horrible job, so I said, \"Sure!\" When I got back to work, I was advised by my supervisor that, \"These new FDR folks are a little tight with the dollar!\" I found that to be true when the schedule they offered me added up to 39.5 hours per week... a mere 30 minutes away from the coveted benefits package! I didn\'t care. I was still living at home, not paying rent, and half-assing my way through community college. Shortly after the re-hire, I was \"Promoted\" again to something called \"Lead Operator\". It was explained to me, that while I would have many more responsibilities, there would be no commensurate raise in pay. What the hell, anything to get off the phones 39.5 hours per week! One of my new duties was to create the ID badges for new employees. I got to take their pictures with a cool  urlLink Polaroid  camera, then laminate the card, etc. This is where I first met Roger, he was a year or two younger than me, built like me (XXL), and, as we soon found out, had many similar interests, including music, beer, gambling, as well as very similar senses of humor; i.e. quick, dry... well, you know, we were  funny !     It didn\'t take long in our friendship for the topic of conversation to get to Las Vegas. I told him about Charley, and how we ventured out there the previous year. Well, from that point on, he was IN! He was introduced to Charley, and luckily, they hit it off.     Over the next two years there were several trips to the magical city in the desert, resulting in some of the best times I\'d ever had. It turns out that Roger had an interest in the travel & tourism industry, and he proved to be a fine travel agent for the young  VEGA$ CLUB . Instead of merely driving out there without a clue as to where to stay, Roger would actually make hotel reservations! The trips had individual themes, including \"Non Stop Fun!\", \" The VEGA$ CLUB  Salutes Red Meat!\", and my favorite, \"We Want Our Money Back!\". In addition to sleeping and gambling, Roger made sure that there were other activities to occupy our time when the cards and dice weren\'t going our way. These activities included the addition of fine dining (no more visits to that garish McDonald\'s), \"Fun Book\" collecting/redeeming, attending the IMAX movie at Caesars, and an obligatory trip out to Red Rock canyon, for an enjoyable drive through the desert. We were on a exhaustive quest for the perfect casino... one that had low minimum games, yet high quality cocktail service. One evening while traversing The Strip, we found it!      In the next installment... The Perfect Casino!        urlLink  '),(18451,4059083,'28,July,2004','2004-07-28 00:00:00','   Keep in mind that my first visit to Las Vegas was in 1982 ( A LONG time ago) and memories are very hazy. As the years progressed, there would be a group of 3-8 friends that would be the basis of \" THE VEGA$ CLUB \". Back in 1982, however, it was just me and Charley. We met earlier that year working the swing shift at a telephone call center that authorized the use of MasterCards and Visas for merchants, long before the \"just swipe the card\" machines that are everywhere today. We became fast friends, hanging out after work at 11 p.m. seeking out the local convenience stores that would sell beer to 19 year olds. In June of \'82, just a year out of high school in Irvine CA, we decided to take the 5 hour drive to Las Vegas. We both were living at home, and we were both under 21. My parents were on an Alaskan Cruise, and Charley told his parents that he was going to spend the night at my house.  We piled in my 1962 Volkswagen Bug early in the morning, and headed out across the desert. We\'d gotten about 20 miles from home, when I saw the sickening sight of a CHP in my rear view mirror with his lights on. The old bug really couldn\'t do much over 65... 70 on a downgrade, but these were the days of \"55 Saves Lives\" and the CHP were taking it seriously. I pulled over only to have the cop pull up next to me, and announce something unintelligible over his PA. Then he drove off! \"I guess he didn\'t mean us!\" I jubilantly announced to Charley. I carefully pulled back onto the freeway and headed back down the road. My joy was short lived. About 100 yards up the road the officer had pulled someone else over, and was now standing in my lane, pointing me to park behind his cruiser. Yep, he got two speeders in one stop! \"I hope this isn\'t an omen\" said Charley. He\'s always been into omens as you\'ll find out if you decide to keep reading.  We made it to VEGA$ without further incident. Of course, neither one of us had even considered a room reservation. Certainly with all of those hotels, there\'d be a room! We cruised down the strip in awe. This was long before the mega-resorts of today, no Mirage, Treasure Island, Mandalay Bay, Bellagio, or Luxor. Just the classy old joints like The Sands, The Dunes, The Tropicana, and of course Caesars! Those places scared us, so we pulled into Circus Circus where the marquee out front (held by an enormous clown) screamed \" ROOMS AVAILABLE  * \" I\'d seen their newspaper ads proclaiming ridiculously cheap rooms, and to our 19 year sensibilities, it didn\'t seem like a bad place. We waited in the lengthy check in line only to find out there was no vacancy! \"But the big sign says  ROOMS AVAILABLE  * \" I complained The desk clerk pointed out that this was one of those  *  occasions. It turns out that next to the  *  in tiny letters it said, \" if not, we\'ll place you \" We let them place us next door at the TraveLodge. We got unpacked and decided to get out there and try our luck! We walked back to the CC and Charley approached a change lady with a $10 bill, looking to buy a roll of quarters. \"Sorry\" she told him, \"You need to be 21 to get change from me. Try going up into the children\'s arcade\". Needless to say we were discouraged. I played the majority of my pocket change in various enticing machines (except for the dimes and pennies, the two most useless coins in Vegas), all the time worrying that if I actually won something, that a crew of security personnel would come rushing out of the garish pink and white decor to check my ID, and relieve me of my winnings. We decided to try someplace  LESS  classy than Circus Circus! We walked next door to the Slots of Fun!, where there was a man standing on the hot asphalt handing out Slots of Fun coupon books! We began to feel better already! Our first stop was at the gift shop to redeem our coupon for a free gift! We took our Slots of Fun keyrings, and decided to try our luck at Black Jack!  To say we were a tad nervous would be a gross understatement. To this day, my hands tend to shake when I\'m in new, or uncomfortable situations. Back then sitting at my first real blackjack table, it must have appeared that I was having some sort of seizure! In my years of going to Vegas, I\'ve made a study of casino personnel, especially dealers, and I\'ve learned (not surprisingly) that the good ones work in the good casinos, and the Slots of Funs of the world are left with 1) very young dealers that are \"breaking into\" the business or 2) very old dealers that don\'t quite have the speed and skills required to toss cards at the fancy joints. Our dealer fit into category 2. He gave us an annoyed look, then growled, \"Everyone here 21?\" Charley and I looked at each other and said, \"Yes!\" in unison. He took our $20 bills and exchanged them for chips. Again, I was concerned about having my winnings confiscated at the door, but luckily that didn\'t become a factor, as the casino confiscated my money in the conventional manner.  I don\'t remember any of the meals on this trip, except that we ate at the McDonald\'s on the strip, near Circus Circus, and we were mightily impressed that they were open 24 hours a day, AND the golden arches sign out front paid proud tribute to the city, in glorious neon and flashing lights.  That was about the extent of the excitement for the trip. But for two sheltered kids from Irvine, it was a great adventure, and was the beginning of a lifelong friendship, and a prelude to many more trips across the desert.  In the next installment... Hank and Charley meet Roger!  '),(18452,4059083,'28,July,2004','2004-07-28 00:00:00','   I think the word BLOG sounds stupid, so I won\'t be using it here. This is my Personal Online Journal, or  POJ  if you will.&nbsp;&nbsp; I\'ve been reading about these for some time now, and up to now I haven\'t had much to say.&nbsp;&nbsp; But tonight I can\'t sleep.&nbsp;&nbsp; It\'s 2 a.m. on California\'s central coast and while&nbsp;visiting some of my favorite websites (I don\'t like the term \"surfing\" either) I came across the \" urlLink How Stuff Works \" website which had a whole article on how POJs work with links and everything.&nbsp; So here I am.&nbsp; The title of this POJ is \"Hank\'s (for the) Memories\". &nbsp;I\'m Hank. I\'m&nbsp;guessing that when viewed in the light of day I will absolutely hate that title... but at 2 a.m. it seems kind of clever. &nbsp;My initial topic I think will be Las Vegas.&nbsp; I have a special fascination with that town, and lots of fond (and a few not-so-fond) memories.&nbsp; If I get further into this writing thing, I may expand onto other topics. But Vegas is my starting point.&nbsp;&nbsp; I\'m a little sleepy now, so I will start tomorrow.  '),(18453,4059083,'09,August,2004','2004-08-09 00:00:00','   In 1985 I left Orange County (aka The O.C.) to go to school up in Stockton, California at the  urlLink University of the Pacific .   This put a bit of a damper on VEGA$ CLUB activities, but Roger, Charley, and Leung were able to make several road trips up to see me, which was good.   Inevitably the topic of gambling came up and led to a few trips to nearby Reno and Tahoe.   Tahoe was beautiful and all, but in the winter time it was pretty treacherous to get to, and Reno offered more gambling opportunities, and really cheap rooms!  One such trip to Reno had been planned for a couple weeks, we had reservations at  urlLink Circus Circus Reno  (ugh), and unfortunately there was snow in the forecast.   A little bad weather wasnâ€™t about to stop us from a fun weekend in â€œ urlLink The Biggest Little City in the World â€.   We stopped at a local Kragen, and got a set of chains for Rogerâ€™s VW Jetta, and we were off!  Somewhere this side of Truckee it started snowing, and shortly we were directed to the side of the road to install our chains.   None of us had ever put them on before.   It was really cold, so I stayed in the car while Roger and Charley put the chains on, and we were back on the road.   It seemed very noisy, and we werenâ€™t really able to go any faster than 45 or so, but what the hell did we know, maybe this was normal?   As we began to descend from the Sierra Nevadas the snow turned into torrential rain, and we were informed that we were leaving the â€œChains Requiredâ€ area.   This time it was REALLY wet and muddy, so I again stayed in the car while Roger and Charley removed the chains.   It wasnâ€™t until we parked the car in the Circus Circus parking garage that we noticed that a strand of the right front chain had come loose during the trek, scraping a semi-circle of paint from Rogerâ€™s fender!   The storm increased in ferocity through the night, forcing us to spend the evening at the â€œ urlLink Pinkest Casino in Northern Nevada â€, only adding fuel to my coulrophobia.   When we got bored with the gambling, we took a walk through the circus midway, gawking at the carnies etc.   We played a few games, including the one where you shoot water into the clownâ€™s mouth, inflating a balloon attached to the clownâ€™s head.   At the end of the night, we had each won at least one  urlLink troll doll .   Speaking for myself, I continue to treasure my little guy (I call him Keith).  The next morning we turned on the TV news, only to find that there was no end in sight for the storm.   Interstate 80 had been closed overnight until further notice, and the Truckee River was in danger of overflowing its banks.   This was a Saturday.   I had to be back at school on Monday, and the others had to be back at work.   With I-80 out of commission, getting back to Stockton by car was out of the question.   Roger, Charley, and Leung got out the map and plotted a course back to Orange County (via Las Vegas, of course) and   I got on the phone to  urlLink American Airlines  and reserved a seat on the next dayâ€™s Reno to Stockton flight.   Things seemed more or less under control, so we bundled up and hit the town!   The breeching of the Truckee River seemed to be the hot topic of conversation around town.   Being accustomed to gambling in the blistering heat of VEGA$, it was a nice change to frolic in the fiercely cold freezing rain.   I specifically remember sitting at a 50Â¢ roulette wheel at the Cal/Neva club that day, and watching workers seal up the entrances with sand bags.   Thatâ€™s when I began to be concerned with this whole Truckee River thing.  Reno reminds me of a slightly older, somewhat seedier, downtown Las Vegas.   It obviously doesnâ€™t have as much money, although Harrahâ€™s and the Hilton are pretty nice.   Itâ€™s telling that the huge expansion of Las Vegas in the 90â€™s didnâ€™t make up to Reno, and Reno never experienced the Steve Wynn-ification and maintains an old western charm.  The next day, we packed up the damaged Jetta and said goodbye to the soggy Circus Circus.   It was still raining hard, but apparently the Truckee River was no longer in danger of flooding the town.   The boys dropped me off at the airport, and were off.   I waited my turn at the check in counter.   When my turn came the clerk looked up my reservation number, and said, â€œOh, it looks like you cancelled this!â€   I most certainly had NOT cancelled my only way out of town!   We went back and forth for a while, and finally determined that I was supposed to have come to the airport the day before to purchase the ticket.   They had neglected to tell me that.   I said, â€œThey told me on the phone that I could buy the ticket when I got to the airport!â€   Of course the flight was full, but not having anywhere else to go, I stood my ground and kept repeating, â€œThey told me on the phone that I could buy the ticket when I got to the airport!â€.   Finally, she took pity, booked me on a flight to San Francisco, with a connecting flight back to Stockton.   I was elated!   The night after I got back to my dorm room, I got a call from the boys who had managed to get a huge suite at the Holiday Casino (now Harrahâ€™s)   Charley was talking to me from the phone by the toilet!  Looking back, it was a memorable trip, but I sure didnâ€™t like the feeling of being stranded in Reno with no money, no airplane ticket, and no vacancies!  '),(18454,4059083,'06,August,2004','2004-08-06 00:00:00','   Downtown has always felt a little more familiar to me than the Strip. Downtown VEGA$ people are my people. Don\'t get me wrong, I enjoy classy things; I\'m in awe every time I enter Caesars Palace, but I\'m just a little too... informal for most of the fancy strip places. I\'m most comfortable wearing a Hawaiian shirt and no pants. Because of certain societal conventions, I\'m usually forced to at least wear short pants when I\'m gambling. I also hate the whole distance thing on the strip. Say you want to go from  urlLink New York, New York  to  urlLink The Mirage , first you have to locate the valet (I don\'t even consider self parking!) wait for your car, battle the impossible strip traffic for 40 minutes to go the Â½ mile to the Mirage, then do battle with ITS\' parking lot! You\'re liable to waste the better part of the day just dealing with your car!!! When staying downtown, you park your car upon arrival and that\'s it! Want someplace seedy?  urlLink El Cortez ,  urlLink The Plaza ,  urlLink The Golden Gate  and many others are right there waiting for you. Want a step up? Try  urlLink Fitzgeralds ,  urlLink Fremont , or the  urlLink Four Queens ! Hankering for a classy joint?  urlLink The Nugget  is smack in the middle of everything.  I am heartsick over the recent events related to my all-time favorite place to shoot craps,  urlLink Binionâ€™s Horseshoe . Binionâ€™s is easily worthy of an entire posting all to itself, so stay tuned.  One trip we were lured into the horrible  urlLink Coin Castle Casino . This was one of those slot joints between The Golden Gate and The Pioneer Club, in the same vein as Sassy Sallyâ€™s with all sorts of gimmicks to get you inside. We were lured into the bowels of the Coin Castle with the promise of a free key chain or some other doodad, and Charley decided to waste a few quarters at an ancient slot machine. Amazingly enough he quickly lined up 3 plumbs and the quarters just started pouring out of that thing like there was no tomorrow! Charley quickly filled up one bucket, and was reaching for another one, when the security guard / slot mechanic / manager on duty came up and told Charley to get out of the way. He explained that Charley was the victim of a â€œrunaway jackpotâ€ and that he was only entitled to 25 quarters! Charley was crestfallen! He was of a mind to dump the bucket of quarters onto the geezerâ€™s head, but he took his meager winnings and we left, never to return for our key chains.  On another trip we were staying downtown at Fitzgeraldâ€™s. It had been a hectic day and we were up in our rooms relaxing in the late afternoon. There was an advertising card in the room talking up the Fitzgeraldâ€™s Personal Pizza room-service special. For only $9.99 you could have an 8-inch, â€œpizza for oneâ€ delivered to the room. We were hungry, and contemplating making a call to room service, when I remembered seeing an ad across the street at Sassy Sallyâ€™s for a 12 inch â€œFamily Sizedâ€ pizza for $3.99! â€œWhy donâ€™t I just walk across the street and get 4 big pizzaâ€™s for $16?â€ The boys laughed at me, which only made me want to do it more! I picked up the tray that was underneath our ice bucket and water glasses, tucked it under my arm and headed out! I waited my turn deep inside the nightmarish Sassy Sallyâ€™s (Where itâ€™s Double Jackpot Time every 10 minutes, and theyâ€™re not afraid to let you know it!) When I got to the front of the line I ordered my four pepperoni pizzas. The surly order taker looked around, â€œWho are all those pizzas for?â€ I told her they were for me and my friends. She explained that the pizzas were for people who were there gambling, that they didnâ€™t do take-out! We went back and forth, and finally I told her that my buddies were on their way, that Iâ€™d just been sent ahead to order the pizza. She grudgingly took my money and placed the order. When they called my number, she looked around the casino searching for the 3 additional gamblers that I had promised her. I quickly balanced the four pies onto my tray and got the hell outta there! I did receive a lot of strange looks walking down Fremont Street with a tray full of pizzas, and one man offered to buy one from me. I was met with more dirty looks as I traversed through the busy Fitzgeraldâ€™s casino floor towards the elevator. It was all worthwhile when I got back to the room and received a standing O from my fellow VEGA$ CLUB members. We had a cooler full of cold beer to accompany the pizza, and it was one great meal! The pizza itself wasnâ€™t the best, as you can imagine, but that didnâ€™t matter. We had said â€œSCREW YOUâ€ to both the price gouging room-service people, AND to greedy Sassy Sally who was attempting to lure in hungry slot players!  '),(18455,4059083,'05,August,2004','2004-08-05 00:00:00','   Sometime in the heyday of THE VEGA$ CLUB (mid 80â€™s) an organization called  Players Club International  burst on the scene with a group of enticing late night TV commercials starring Telly Savalas.   For a small up-front fee, you could swing like Kojak!   I forget the details, but for the price of membership you were entitled to substantial discounts on hotel reservations, plane tickets, shows and meals.     Roger signed up in a heartbeat!   The rest of us were envious of his shiny gold Players Club official membership card, and luckily he readily shared the benefits bestowed upon members.     One of the hotels that offered benefits to PC members (possibly the only one in Vegas) was the  urlLink Riviera , so we stayed there several times, enjoying the deeply discounted room rates, free buffet coupons, discounts to shows, etc.   We went to â€œan evening at the improvâ€ there and the headliner was KNBC Los Angelesâ€™ TV weatherman Fritz Coleman.   He wasnâ€™t very funny.     The Riviera was okay, itâ€™s kind of got that â€œold Vegasâ€ feel that is getting harder to come by as the years go by, but we didnâ€™t spend much time gambling there.   After a few late-night gaming sessions along the strip, Roger and I would be walking back to the Riv, only to be drawn in by the midnight Steak & Eggs special at the nearby  urlLink Westward Ho!      On another trip we found ourselves trying out late night breakfast specials at  Foxyâ€™s Firehouse .   This place had the atmosphere and clientele of Little Caesarâ€™s, but no craps table, so I had no interest in it.   I guess it was a small step up from LCâ€™s since it had a restaurant and a security guard.   On this particular visit, the lone security guard was posted near one of the entrancesâ€¦ playing video poker!   The joint was located just north of The Sahara on the strip, across the street from The Worldâ€™s Largest Gift Shop, where the Holy Cow Brewing Co. is now located.   The thing I remember most about the $1.99 breakfast special is that it was way yucky.   It was served on Styrofoam plates with plastic knives and forks, and the â€œchefâ€ working the graveyard shift at Foxyâ€™s almost seemed like he didnâ€™t want to be there!   When breakfast was over, Roger eyed the single blackjack table and decided that he wanted to win the price of his breakfast back!   I was too tired to gamble, plus after a day of cheap beer, free hotdogs, the worldâ€™s largest taco, a buffet lunch, free margaritas, and a $1.99 midnight breakfast I was a little green around the gills.   Roger plopped down 2 bucks on the blackjack table and was dealt two face cards.   I was relieved.   â€œLet him win the price of his damned breakfast back so we can get some sleep!â€ I was thinking.   The dealer was showing a 3, and turned over a 2, hit with a 3, then a 4, another 4, then a 5.      Roger and I stared at the spread of cards trying to add them all up.      The dealer was only marginally faster and announced â€œ21!â€ while picking up Rogerâ€™s $2.      Would you believe it, the same thing happened for the next three hands?    By the fifth hand, Roger was getting pissed.      He plopped down a $10 spot for one last hand.      He got a respectable 19.     The dealer was showing a 5, but at this point we werenâ€™t breathing easy.      She turned over a 6, (shit!) then pulled out a 4 (yay!) then an Ace (double yay!) thenâ€¦ another 5 (Oh, Jesus!)   Roger lost it.      â€œI canâ€™t believe it!   This is bullshit!!!â€ he screamed!      The security guard glanced over his shoulder from the poker machine, then went back to his game.      Roger stood up, lifted the blackjack stool over his head, and before he had a chance to crash it down onto the table, a short, squat woman came running over, â€œSIR, PUT THE STOOL DOWN RIGHT NOW!â€   He gave her a menacing look, â€œWho the hell are you?â€ he snarled.      â€œIâ€™M THE PITBOSS AND I WANT YOU TO PUT THE STOOL DOWN AND LEAVE THE CASINO , NOW!â€ He looked her up and down with a sneer, â€œPit boss?   You look more like a Pit  Bull !â€   He finally put the stool down, and we beat a hasty retreat.      When Foxyâ€™s Firehouse closed down, nobody shed a tear!     In the next installmentâ€¦ Downtown Fun!     P.S.   The above anecdote has possibly been embellished by many retellings over the years.     In truth, he lifted the stool about 2 inches off the ground, and had no intention of crashing it down onto the table.      He also never uttered the â€œPit Bullâ€ remark, but he wanted to, and the security guard never glanced over his shoulder.      That incident also added a new phrase to our vocabulary.      From then on, anytime a dealer got a total of 21 with 5 or 6 cards, it became a â€œFirehouse 21â€   P.P.S I just â€œgoogledâ€ Telly Savalas to see if I could find out more information on the Players Club, and I found a  urlLink website  dedicated to celebrities that are missing parts of their fingers!   Telly is apparently missing part of his left index finger!      Other digitally challenged people include James â€œScottyâ€ Doohan (right middle finger), and Daryl Hannah (the tip of her left index finger)   Isnâ€™t the internet incredible?  '),(18456,4059083,'03,August,2004','2004-08-03 00:00:00','   On one of those mid-80\'s summer trips, the boys of  THE VEGA$ CLUB  were wandering around mid-strip after attending another breathtaking presentation at  urlLink Caesars â€™ IMAX theater. We were walking around in the sweltering heat redeeming coupons at most of the little souvenir shops that used to be abundant in that neighborhood.   Those little shops and strip malls have all been torn down now to make room for places like The Monte Carlo, The New Aladdin, and Paris.   I understand that the stubborn little Tam â€˜o Shanter Motel will soon be a thing of the past as well.   Oh well.   We were walking through an especially seedy looking strip mall next to Ballyâ€™s (Where  urlLink Paris  is now) when we came across  Little Caesarâ€™s Casino .   You entered through a sliding glass door where someone had wrapped the metal door handle with about 25 layers of duct tape, in an attempt to keep the handle from inflicting 3 rd  degree burns on the casino guests.   A hand-written sign was taped to the door sternly reminding everyone to close the door behind them!   The hard-working air conditioning system succeeded in lowering the indoor temperature by about 10 degrees from the outdoor temperature.   Unfortunately on that day it was about 105 outside!   I have since learned that Little Caesarâ€™s was nicknamed â€œThe Toiletâ€ (according to legend, during a particularly hot roll at the craps table, the shooter didnâ€™tâ€™ want the dice to cool down if he paused to use the restroom, so he just peed at the table!)   On that hot afternoon, with the exception of a bustling craps game, the place was pretty much deserted.   Charley had wandered off towards the dingy gift shop, and he called us over, â€œHey, look at this!   Penny slots!!!â€   Sure enough, off in the corner were 4 ancient penny slot machines with a top jackpot of $50.   I walked over towards the craps game, where there were many unsavory characters 2 deep, and the layout was COVERED with beige chips, which I later discovered were worth 25Â¢ each!   Yes, we had found a quarter craps game!   There was also a â€œdouble exposure 21â€ table with no players, and one open blackjack table with no players.   Standing behind the other blackjack table was an old man in a $60 rumpled suit that had to be the pit boss.   Roger nudged me, â€œCheck out what that old guyâ€™s doing!â€   We looked closer and discovered that he had a deck of cards spread out across the table and he was wiping each card down with a damp towel!   This was astounding to us!   Weâ€™d all played enough blackjack to know that most places replace the cards with brand new decks every few hours, but at Little Caesarâ€™s, why waste money on new cards, when you can just wash off the old ones!   We got the impression that at Little Caesarâ€™s, every dollar counted!   On another visit, I saw the pit boss groan in agony, yell the â€œFâ€ word, and kick his little desk when a guy won a $40 pass line bet.        For reasons unknown to most of my friends, I grew attached to â€œThe Toiletâ€.   Nobody ever wanted to go with me, so I spent a lot of time by myself in there.   First off, I liked the idea of 25Â¢ craps.   I felt like a high-roller betting $1 chips, amongst all the 25Â¢ bets of the degenerate/homeless men gambling around me.   One morning I happened to notice that there were no cocktail waitresses, nor was there a bar.   I asked the dealer if they served drinks.   â€œCocktails!â€ he shouted, raising my hopes.   A short heavyset bald man who was also the change person came waddling over to me.   â€œYouâ€™re the cocktail waitress?â€ I asked. He rolled his eyes at me and asked me what Iâ€™d like.   It was about 9 a.m. so I asked for a Bloody Mary.   He actually scoffed at me, then said, â€œCoffee or beer?â€.   Iâ€™ve never been much of a coffee fan, and I can just imagine how horrible a Little Caesarâ€™s cup of java would be, so I ordered a beer.   He went over to a little area next to the cashiers station where apparently they had a keg set up.   He brought back an 8 ounce paper Dixie Cup of what I swear had to have been Hammâ€™s beer.   I tipped him the usual $1 chip and got a very polite, â€œThank you, sir!â€ in return.   Iâ€™m guessing that that was the biggest tip heâ€™d received all week, because the Dixie Cups just kept cominâ€™!     Another reason that I kept going back to LCâ€™s was that I almost always won a few bucks there.   The craps game there was something called â€œCrapless Crapsâ€ which I believe he got from Bob Stupakâ€™s Vegas World, which could only mean that it was a total rip-off game.   I havenâ€™t studied the statistics of this game, but if it was offered at Vegas World, then it had to have a bigger house percentage than standard craps.   The game, as the name implies, has no craps rolls, everything except seven can be a point.   Roll snake eyes, and 2 is the point.   Roll an eleven, then thatâ€™s the point.   It was kind of weird to begin with, but you got used to it.        On another night when the craps table was totally full, I decided to try out some $1 blackjack to kill some time until a spot opened up at the craps table.   The whole table was losing, but at a buck a hand, it wasnâ€™t too bad.   One guy at the table was playing with a stack of $20 bills instead of chips, betting $20 and $40 per hand and losing consistently.   After quickly losing his entire stack, he cursed and pulled out two $100 bills, and slapped them in his betting circle.   As you can imagine, this kind of action tends to draw the attention of the Pit Boss at a place like LCâ€™s!   The very young, non-English speaking break-in dealer nervously glanced over to the supervisor, â€œMoney prays?â€ and got the nod from the boss.   True to form, we were all dealt 15s and 16s, with the dealer showing an Ace.   Mr. Big-spender took a hit, busted, cursed and quickly grabbed up his two C notes and started to get up.   All hell broke loose!   The dealer screamed, â€œHEY!â€ the Pit boss clamped his left hand down on the players arm, while pulling out a can of mace with his right hand.   The player lamely tried to explain that he wanted to stand, not take a hit, and that the dealer misunderstood him.   He looked to the fellow players, â€œYou saw me wave my hand back and forth, right?   I wanted to stay!!!â€   In return, he got four icy stares.   The boss increased his grip on the guyâ€™s hand, â€œEither you put that money back on the table, or youâ€™ll get a face full of pepper spray, and a trip to jail!â€   It was then that I realized that in addition to not having cocktail waitresses, they also didnâ€™t have any security guards!   I also realized that I was probably taking my life into my hands coming into a place like this after dark.        A few years later, I was in town by myself on one of those spur-of-the-moment weekend getaways.   I was actually staying 20 miles south of town in Jean, NV.   Iâ€™d arrived in Jean late Friday night, getting a room at the Gold Strike.   Then the next morning I drove into town, deciding to start the day with some quarter craps.   I pulled into the strip mall, and there was a white paper placemat scotch taped to the inside of the sliding glass door.   With a black Keno crayon, somebody had scrawled â€œCLOSEDâ€.   I canâ€™t say that I was totally surprised, but it was still the end of an era.   I stared at the placemat for a few minutes and possibly shed a tear, but at the same time smiling about how fitting the crayon and placemat announcement was for such a place as  Little Caesarâ€™s .   In the next intallment... Foxy\'s Firehouse, midnight steak &amp; eggs, and The Players Club International!  '),(18457,4059083,'02,August,2004','2004-08-02 00:00:00','   I can\'t believe that I\'ve gotten this far without touching upon the incredibly important topic of Las Vegas food! From the 99Â¢ shrimp cocktail, and the $3.99 breakfast buffet to Spago\'s... Las Vegas is nearly as famous for its food as it is for its gambling! In this posting I will describe as many of the restaurants as I can remember, and rate them using the tried and true one to five star system. Ratings are subjective based on value, atmosphere, food quality, service, and overall fun. I am purposely skipping chain restaurants, Vegas certainly has its share of Ruth\'s Chris Steakhouses, Tony Roma\'s, California Pizza Kitchens, Carrows, Denny\'s etc. This is not the place to talk about those.  We\'ll Start with Dinner:   Ricardo\'s  2380 East Tropicana Avenue As I mentioned earlier, as far as  THE VEGA$ CLUB  goes, Roger was in charge of dining and hospitality. One of his early recommendations was a place that was owned by his uncle Richard, called  Ricardo\'s . Ricardo = Richard, get it? It turns out that  Ricardo\'s  is one of the best Mexican restaurants that I\'d ever been to. We always made a point of going there at least once per trip. Roger always attempted to get us a free meal by dropping his uncle\'s name. It actually worked a few times! It\'s the first place that I ever saw Margaritas coming out of one of those Slurpee machines! The child-like fun of drinking a Slurpee, plus the adult fun of drinking tequila!!! Rating  ****    The Pink Pony   urlLink Circus Circus  Main Tower I have grown to loathe all things Circus Circus over the years. For one thing, I suffer from coulrophobia, which is a fear of clowns. Circus Circus is no place for coulrophobes! Plus it\'s big, old, and slightly seedy. I have nothing against being old and seedy, but I\'m uncomfortable in HUGE casinos. The Pink Pony, however, is located in a small side-casino away from most of the riff-raff. The only item that  THE VEGA$ CLUB  members ever ordered was the Cattleman\'s Cut Prime Rib special. An average quality slap of Prime Rib served with generous portions of potato, vegetable, salad, and dinner rolls. Rating  **    William B\'s  Inside  urlLink The Stardust  main casino This is a great place when you are in town for a convention and the boss is paying! I\'m in the broadcast industry, and once every few years I get to go to the National Assn. of Broadcasters convention with my boss. William B\'s is a classy steakhouse with great steaks, and a great old-time steakhouse atmosphere. The Maitre\'d wears a tuxedo! I say again, pricey! Rating  ****    Guadalajara  Inside the  urlLink Palace Station  To me, all of The Station properties feel like  The Orleans.  They are nice and friendly, but I can\'t win for losing there! I\'ve stayed at The Palace Station several times, and really enjoy their Guadalajara restaurant. In fact, one thanksgiving several years ago my wife and I tried to go to their buffet,  The Feast  for a traditional Thanksgiving day dinner, and the line out front was about a mile long! We walked down to Guadalajara and had the joint to ourselves! They have a great salsa bar, good Margaritas, and very good food. Rating  ****    Battista\'s Hole in the Wall  4041 Audrie Street My All-Time Favorite! This gem was discovered by Charley\'s wife, Pam while she was in town on a business trip, I believe. Some places are good if you\'re with your buddies, other places are good if you\'re with your wife. This place fits any bill! It\'s dark and romantic, yet fun. It\'s not the cheapest place in town (Meals start at about $20) and it\'s absolutely no good if you don\'t like cheap wine. But, if you can afford it, and you like cheap wine... This is it! There\'s even a strolling accordion player named Gordy to serenade your meal. You get your choice of standard Italian entrees (My favorites are the Lasagna and the Veal Parmigiana) and meals include, salad, great garlic bread, your entree, hot chocolate (they call it a cappuccino, but it\'s hot chocolate) AND, oh my god, are you sitting down for this? ALL THE CHEAP WINE YOU CAN DRINK!!! They\'ll ask you if you like white or red. If you\'re smart, one person at the table will say \"red\" and someone else will say \"white\", that way they\'ll bring two carafes right off the bat... then they just keep bringing them! Oh, you\'ll feel  so  sick in the morning! Rating  ***** +   The Palatium Buffet  Near the race/sports book  urlLink Caesars Palace  Vegas is known for its buffets, and I\'ve tried a lot of them!  Most of them are perfectly adequate, a few are sub-standard ( Circus Circus  comes to mind), and a very few are quite amazing!  The Palatium at Caesars is near the top of the heap! (The buffet at  The Mirage  also gets high marks from me, although the line is usually unacceptably long there). In addition to very high-quality food, Caesars has a variety of specialty stations where men and women in chef hats give personal service. My favorite stations are the prime rib carving station, and for brunch the omelet station. Last time I was there the dinner buffet was $15.95, so it\'s not exactly a bargain, but in this case you definitely get what you pay for. Rating  ****    The Triple 7 Restaurant and Brewery   urlLink The Main Street Station  As a lover of beer, I\'d like to give special mention to The Triple 7, located downtown, next to the Union Plaza. The beer is very good (free if you\'re playing video poker at the bar!) and the hamburgers and sandwiches are very tasty. A great stop for a late afternoon snack, and the casino itself is pretty swanky, given the questionable neighborhood in which it\'s located. Rating  ***   For Breakfast, I\'m usually happiest with a traditional coffee shop, preferably located in the hotel in which I\'m staying. Some of my favorites include  The Carson Street Cafe  inside  urlLink The Golden Nugget ,  The Courtyard Cafe  inside  urlLink The Orleans , and if you\'re on your way back to California and need a place to stretch your legs, use the restroom, and get a good breakfast, try  The Wagon Master Coffee Shop  inside of  urlLink Buffalo Bills , at the state line.  In the next installment... Little Caesars... Oh My God!  '),(18458,4059083,'02,August,2004','2004-08-02 00:00:00','   For the first few years, my game was Blackjack. When the cards weren\'t going in my favor, I\'d take a break in front of a video poker machine (preferably at the bar). Meanwhile, Charley and Roger had been dabbling in craps, and were throwing around a lot of terminology, and tales of exciting runs at the table. Anyone who\'s ever been in a casino when there\'s a hot craps game going on knows how explosive that game can get. They had tried to explain the basic rules to me several times, but I just couldn\'t seem to get the hang of it, and observing a game in action just made matters worse.  One day I had just finished up a losing session at blackjack at a downtown casino, and was walking around looking for Roger and Charley. I found them at the casino\'s only craps table so I stood back and watched for awhile. I had a single $5 chip left from my blackjack play, and I said to them, \"If I wanted to make a $5 bet, without getting too complicated, where would I put it?\" They both suggested putting in  The Field . It\'s real simple they explained, it\'s just a one-roll bet on the next throw of the dice. You either win or lose! So, I carefully put my chip in the field just before the next throw. The dice flew, the stickman continued his ongoing patter, \"12 CRAPS 12 TAKE THE COMES AND TRIPLE THE FIELD\" I didn\'t know much about the game, but I knew that 12 was one of the \"craps\" numbers, and that was bad. Meanwhile, Roger and Charley were whooping it up on my behalf as the dealer slid three red chips next to mine. \"Pick those up!\" Charley said, \"12 pays three to one in the field!\" I picked up my $20, cashed them out, and from that point on I\'ve been hooked on craps!  I found out later, that betting \"The Field\" is kind of a sucker bet, but on that occasion, it was all I needed to get the fever. I bought books, computer games, I frequented some truly scary places in search of low-limit games, and I\'ve been a part of some very memorable hands (in the Craps vernacular, they are referred to as  monster hands  or  Dukes ). Generally speaking, I was always too scared to substantially increase my bets, so I seldom walked away with more than a couple of hundred dollars, while the rest of the table cashed out thousands in black and purple chips. The best I ever did was at the Desert Inn, shortly before it closed. This was in September of 1997 so I will get to it in a later entry, but it\'s the only time that I was ever in possession of a $500 chip. I love casino chips, and have a small collection, and I remember staring and staring at the pastel yellow D.I. chip in awe. Later that week, I met the woman who was to become my wife... so needless to say, I was on quite a hot streak!  In the next installment... The importance of food!  P.S. I\'ve been published!!! Big thanks to the folks at  urlLink LVconfidential.com . They like my stories, and have reprinted one of them! Check them out at  urlLink http://www.lvconfidential.com/ !  '),(18459,3894428,'16,July,2004','2004-07-16 00:00:00','   So um yea..my night    sux   .&nbsp;  I was   suposed to   go to Krome w/ Kelly and Cassie..But my mother screwed that all up. So I\'m spending my night   home alone  ..with my dog Ivo..  Cutting myself  . This just royally sux. My mom had said she thought that cassie seemed kind of   selfish  ..and the more I think about it..the more true it is. Like the only time she wants to hang out w/ me..or even talk to me is when I\'m with Kelly..or when they\'re hanging out and Kelly wants me to come. We used to be   BEST  &nbsp;friends..and now we\'re   barely even   friends. She moved closer, and instead of helping our relationship..it just recked it. I was rite wen I said Kelly was taking   my   life over..and it\'s gets truer and truer each day.   '),(18460,3894428,'12,July,2004','2004-07-12 00:00:00','   Yest. I went over to Kelly\'s for the night.   [-] I only got 5.5 hours of sleep last night..  [-] Micheal will NOT stop making fun of me for being short..okay I can tolerate it the first, i dunno, 100 times from the same people, but after that it gets really annoying...really.  [+] We went swimming last night.  [-] I can\'t hang out w/ Kelly with out Mike being there. It\'s really pissing me off. She always invites him along EVERYTIME we hang out. Personally, I find him really annoying and immature. Plus, he likes me, and I swear to God if he doesn\'t stop touching me I\'m going to deck him.  [+] As of last night I quit smoking..it\'s disgusting.  [-] I gained 5 more pounds..I\'m now 122.  [-] I have an hour long violin lesson..  [-] It\'s raining.   '),(18461,3894428,'10,July,2004','2004-07-10 00:00:00','   urlLink    jess, me, bethann, carlo, kerrin, and tif at krome..&nbsp; urlLink   '),(18462,3894428,'10,July,2004','2004-07-10 00:00:00','   ::Fierce anger is derived by Passion::  I\'ve been thinking about this a lot. The first time I heard it I thought it was crap..but now that I think of it..it\'s true.   My friends have been bumming me out a lot lately. Just the stupid things set me off. My two best friends (both of whom hated each other before I finally got them to talk..and I was friends w. both long before they even knew each other existed) hang out all the time w/o me. This includes my birthday that was like two weeks ago. It makes me so mad, and it\'s the stupidest thing in the world. But I guess if they didnt mean so much to me it really wouldnt matter to me..would it?   I\'m also a very jealous person. At least over some things. Like if I got hurt by a guy I was with, and he started going out w/ someone else, I\'ll hate that person for pretty much no reason - for indirectly pissing me off. It really wouldn\'t matter if that guy hadn\'t meant anything to me. I dunno..a bit of thought for my night.   '),(18463,3894428,'07,August,2004','2004-08-07 00:00:00','   urlLink    Pic 1 . Ross looking \"sophisticated\" at prom.&nbsp; urlLink   '),(18464,3894428,'03,August,2004','2004-08-03 00:00:00','   urlLink    Pic 1 . Bryan Glenn, Matt Casey, Andrew Mathe, and Someone else?&nbsp; urlLink   '),(18465,3894428,'03,August,2004','2004-08-03 00:00:00','   I\'m not using Nel\'s place anymore\'&nbsp; urlLink   '),(18466,3902340,'30,April,2004','2004-04-30 00:00:00','   4-30-04  20H 07  &nbsp;  Dinner at Le Cockney Tavern again! My feet are too tired to travel downtown and Iâ€™ve ruled out a fine restaurant again. The good ones are full on a Friday night anyway, and Iâ€™m not going alone. I just had another street skirmish with the same overly aggressive strip club doorman. He tried to grab my coat again but I was ready for him. He mumbled something about â€œHey Bullshitâ€ and I answered with an authoritative â€œFuck youâ€ and he disappeared pretty quickly. The French are not a tall people. Itâ€™s so easy to pick out the North Americans on the Metro. Iâ€™m about 5â€™8â€ and I tower over the French. Wasnâ€™t Napoleon knee-high to a grasshopper himself?  &nbsp;  I finished the MacDonald book. Whew. It was heavy. Iâ€™ve thought about suggesting it to Val, but I think it would be too heavy. Sheâ€™s probably already read it, anyhow. Iâ€™ve thought a lot about her this week, realizing that April 26th marked two years since the sale of the house, and our official separation. (The divorce was legal on October 1st - a year and a half ago.) I think about how much fun we might have had in Paris, at least in the early days, before whatever went wrong did.  &nbsp;  Meanwhile, I got an email from T. (from I.) today. I think sheâ€™s anxious about me, testing the water to see if Iâ€™d react. Iâ€™m not sure what her deal is, but I think she might have been freaked out initially by what Jeanine calls my â€œfreakyâ€ email behavior. It was a tactical error, OK? Nervous pussy. Thatâ€™s me! Truth is I didnâ€™t have to be nervous when I was married!  &nbsp;  I had a weird experience in the Juliet Metro station. I was&nbsp;ambling along&nbsp;like a tourist and this African woman ran right into the back of me. I couldnâ€™t see who it was, so I recoiled, then spun around thinking it might be a pickpocket. I think she was mortified when she saw my reaction. She started to squawk away in French to me. Probably â€œOh so you thought I was trying to steal from you!â€ She might have even suggested it was racially motivated, Iâ€™m not sure. Truth is she was talking away on her cell phone and wasnâ€™t watching where in the frig she was going.&nbsp; All the literature warns ad nauseum about pickpockets. Just when I was starting to feel a little more comfortable in this metropolis of 10 million!  &nbsp;  Oh well, I feel my time is winding down here. Iâ€™m really glad for the side trip to Scotland. And itâ€™s a quick social visit, no laborious treks, just friendship and good company. The barkeep has arrived, the guy from my first night here. He reminds me of Dale. Heâ€™d fit right in here. All his people were French, easy to tell.  &nbsp;  In the middle of the night, Iâ€™m positive I heard our Italian friend from the CafÃ© Lux Bar stomp drunkenly down Rue Lepic screaming foreign curses at her boyfriend. Some of it was in English. â€œFuck you, mister. Youâ€™re through! Thatâ€™s it. Youâ€™re through!â€ And he was yelling â€œMalena. Thatâ€™s enough!â€ I heard a car start and drive off. Drunk driver, I thought. Those Italian girls are fiery little trollops!  &nbsp;  Random Note: One this is for sure. Wherever the bill says â€œservice et compris,â€ the service tends to suck.  &nbsp;  Today, I was befriended by the proprietor of a cafÃ©. His waitress was quite aggressive and in my face with unexpected French phrases. I panicked. â€œMon Francais et terrible,â€ I offered humbly. She hollered for, letâ€™s call him Maurice. He was a cross between Mel (Vic Tayback) the diner cook from TVâ€™s â€œAliceâ€ and porn star Ron Jeremy with a bit of our Toronto baker-friend Pasquale thrown in for good measure. He was super friendly and asked in French if I was from Toronto, when I answered that I was from Canada. Then he said good naturedly, also in French, â€œHow can you not speak French if you are from Canada?â€ I have no problem understanding his French. Anyway, he explained every detail of the menu proudly in broken English, then processed my order, turned out the lights in the cafe, and he and his waitress sat down for their lunch. Then loud talking and animated discourse with his unfriendly waitress, and the remaining patrons followed. Sheâ€™d get up for the few regulars that would straggle in. Everything stops between 2 and 3pm in Paris. Most shops close. Imagine closing a shop in the middle of the work day! Maurice was a very warm guy and thanked me for stopping by.  &nbsp;  Ohh. Thereâ€™s instant drama here at Le Cockney Tavern! An English speaking woman has just had her purse or wallet stolen in front of her eyes. Sheâ€™s freaking. Whatâ€™s weird is that the staff doesnâ€™t seem to care in the least. The woman has gone off into the street now to look for the perp, I guess. Whoa, that was eerie. Back to the apartment now. '),(18467,3902340,'29,April,2004','2004-04-29 00:00:00','   4-30-04  15H 36&nbsp;  &nbsp;  Today is the last full day in Paris. Loud construction next door and a noisy maid upstairs pried me from my bed this morning at the ungodly hour of 9:30am! My sleep was not the best last night. It was full of unpleasant dreams Iâ€™m sure due to my poor, alcohol-laden holiday diet. Too much bread and dairy, not enough fruits and vegetables!   Iâ€™m just about finished the Anne Marie MacDonald book. Itâ€™s an epic, but also deals with some heavy sexual abuse suffered by the main character when she was nine years old. As the story concludes, it becomes quite heavy and introspective â€“ yet I canâ€™t seem to put it down. 700+ pages. I checked my email again this morning and had one from Mike. I just canâ€™t tell how heâ€™s reacting to this&nbsp;new business&nbsp;deal It brings up a lot of issues, like who owns the deal and how the money is split. I donâ€™t think he likes the idea of having a partner and Iâ€™ll be damned if I am going to become an employee again. He knows that. We just need to talk it through. Good business and relationships is hard work, man. Rain has been threatening all day. In fact, I guarantee that Iâ€™ll see some drops before I get home.  &nbsp;  Heidi is going to meet me at Prestwick Airport south of Glasgow in the morning. What a doll! It will be good to see her, and itâ€™ll be nice to have familiar company. Thereâ€™s still no sign of S, my potential English girl back in Vancouver. I just donâ€™t know what to think of that. I wish people could just be honest. A doode can take it! Know what Iâ€™m sayinâ€™? Dammit, Iâ€™m tired and I didnâ€™t really get  that  drunk last night. I just completed 8 postcards to send home. I donâ€™t think Iâ€™ll take gifts back, other than something for the girls in the office. Chocolates maybe. Itâ€™s just too much to think about. By the way, I found two Hohner Blues Band harmonicas, one in D and one in F for OB. Iâ€™ll send them along with OEâ€™s prayer card when I get back. Did I already mention this? Iâ€™m looking forward to getting a great bottle of Scotch from Cadenheadâ€™s Whiskey Shop on The Royal Mile in Edinburgh. I know a little bit more about&nbsp;Scotch now, compared with last summer.&nbsp;My collection is coming along nicely!  &nbsp;  Hey, 10 Eastern European countries will join the EU tomorrow, May 1st. It becomes the worldâ€™s largest trading block with a combined population of 450 million. Iâ€™m not sure if thatâ€™s good or bad for Europe. It could be the beginning of the homogenization of it all. It does mean, however, that a new marketplace has opened up for underperforming, restricted economies. It might give Global Americanization a run for its moneyâ€¦maybe not. What the fuck do I know? '),(18468,3902340,'29,April,2004','2004-04-29 00:00:00','   4-29-04  12H 50&nbsp; &nbsp;  Jâ€™ai arrive en Versailles au Rive Gauche. The Palace is around the corner and Iâ€™ve taken refuge in a pizza joint with the motto, â€œWeâ€™ve Got a Passion for Pizza.â€ Itâ€™s obviously designed for travelers, and is kitty corner to junk shops and the local McDonalds. I just canâ€™t bring myself to go to McDonalds in France! Iâ€™ll take the Anglo pizza place instead.  &nbsp;  The train is very efficient. But I almost got off on the wrong track, literally missing the first train. It looked like Iâ€™d have to wait up to 25 minutes for the next one. But just ten minutes later, voilaâ€¦the second one appeared. It was very un-crowded and scenic as we rolled through the hilly suburbs of Paris.  &nbsp;  The pizza was less than spectacular though, and Iâ€™ve just noticed all the American license plates on the wall. Oh how gross! Itâ€™s a nice place. It just seems a little over-the-top with all the Americana. Maybe itâ€™s just me, â€œoversensitive Canadianâ€ all of a sudden.  &nbsp;  Did I mention meeting â€œBob and Sueâ€ from Toronto on the Metro the other day? They reminded me of Graham and Dorothy Sullivan, except a little more gung-ho. Bob was lugging a great big suitcase with a Canadian flag on it down the Metro steps. So I says â€œHowâ€™s it going eh?â€ and Bob says â€œWeâ€™re Bob and Sue from Torontoâ€ and I says â€œShawn from Vancouver/Nova Scotia.â€ Theyâ€™d been traveling all over Europe since September. Jeez the lifestyle, eh? Random Note: I must order a new credit card. My magnetic strip ist kaput! '),(18469,3902340,'28,April,2004','2004-04-28 00:00:00','   4-29-04  9H 47  &nbsp;  The radio station Iâ€™ve been listening to is â€œRadio Matisseâ€¦eclectique et authentique.â€ '),(18470,3902340,'28,April,2004','2004-04-28 00:00:00','   4-28-04  20H 26  &nbsp;  What a nice quiet day today. After walking for five days straight, I gave my feet some rest. The day started with a leisurely breakfast of croissants, berries, and Special K. My sleep was good, although I stayed up until at least 1:30am reading the Anne-Marie MacDonald epic, â€œAs the Crow Flies.â€ Iâ€™ve had it since Christmas and havenâ€™t even had a chance to open it. It isnâ€™t disappointing, thatâ€™s for sure! In fact, I didnâ€™t leave the apartment today until 2pm because I was glued to the pages. I finally got out to scope out the Porte Maillot Metro station in advance of my Beauvais trip on Saturday morning. Good thing I did! Itâ€™s smack dab in the middle of a roundabout amidst freeway exits and on-ramps. I reckon 40-45 minutes should give me twice the time I need to get there on Saturday morning. I talked with Rob on the phone briefly last night. His big Paris suggestion is to introduce myself to as many girls as possible with â€œBonjour. Je mâ€™appelles Claude!â€ LOL  &nbsp;  Iâ€™m eating tonight at another bar. The food is good, and itâ€™s just so much more conducive to dining single. I passed some great looking restaurants, but Iâ€™ve no desire to go in and eat alone. Dining and sleeping single are the two biggest drawbacks of traveling solo. I suppose if I was really motivated, I could fix both of things by flashing a few Euros around downtown and using the name Claude! I just realized why thereâ€™s no place on the credit card receipt to sign off on a tip. It appears that â€œservice en comprisâ€ means the tip is included. Fuck! Now I figure it out.  &nbsp;  Anyway, I decided to set out of town tomorrow to Versailles to see the Palace, plus to get out of Paris for the day. Iâ€™m getting cabin fever. Iâ€™ll take the train which means I should get to see some countryside, and my feet are feeling pretty good after an easy day today. The weather has turned cloudy, and even if it doesnâ€™t improve, I had a great run, and have a good tan early on. Iâ€™m looking forward to seeing Heidi. This is normally the point in the trip where I would get in a rental car and bolt. But, Iâ€™m really freaked out about driving in and around Paris. Maybe if I had a co-pilot to read the map. I know I would not drive in London. Not even New York. Itâ€™s too crazy and parking is a bitch. '),(18471,3902340,'27,April,2004','2004-04-27 00:00:00','   4-27-04  20H 00&nbsp; &nbsp;  Iâ€™m back at Le Cockney Tavern for a beer and a burger trying to take it a little easier tonight, after the drunkardâ€™s dream last night.  &nbsp;  It was a busy day today, with a walk across the Seine in front of the Assemble Nationale. I stopped at the Musee Rodin where I saw some of his most famous sculptures, the highlight of course being the â€œThinking Man.â€ From there, I walked throughout the neighbourhood, past the Prime Ministerâ€™s house. I ended up at the Musee Dâ€™Orsay, which was very good. They had great paintings by Degas, Manet, Monet, etc. My toe was bothering me again (blisters), so I sat my arse down and applied by bandage right there in the museum. Relief! The pharmacist could not present me with anything resembling a band-aid. Apparently there is no translation for Band-Aid. She did fix me up with a small bandage after much discourse, me trying in simple English to explain the concept of the fucking band aid. Jesus Christ, Almighty.  &nbsp;  After a slow start (nursing hangover) this morning, I went to the internet cafÃ© to check my email. I realize how few people know that Iâ€™m away. I guess I kinda kept this one under the radar. I also sneaked a peak at my work/home email, delighted to find that itâ€™s quiet work-wise. Some of my friends had emailed the home account, like Gord and Rob. I guess I didnâ€™t even tell them exactly when I was going, and that Iâ€™d be using my hotmail. Anyway, I was glad to see the S. contract waiting for me. I should print it off, sign it, and fax it from Paris, just to get it happening. Then again, Iâ€™m on vacation. I will print it off and fax it tomorrow. Money is money, for fuckâ€™s sakes. Maybe itâ€™s something that Cecile can help me with!?  &nbsp;  Tonight is going to be a quiet one after last nightâ€™s fun at the Lux Bar. But if I canâ€™t sleep, Iâ€™m going across the street.  &nbsp;  Iâ€™m done with museums. My three-day pass expires today, and my feet are killing me (toe)! I may think about renting a car and getting into the country-side. But, Cecile has scared me from attempting to drive in the city. When I get back, Iâ€™ll look up the rental places in the phone book.  &nbsp;  An autre raison pour telephoner Cecile! Random note: The song playing in the pub is â€œHow Beautiful is the Body of Christ?â€ I think itâ€™s Amy Grant. Itâ€™s a little weird for a pub but nobody notices. Iâ€™m making a list of gifts that I need to get for friends. Itâ€™s the red light district, so Iâ€™m thinking about something from the Erotica Museum for Paula (an excuse to call and say hi) and maybe a French dildo for J. Sheâ€™s probably got more than one! She takes her pleasure seriouslyâ€¦ If I buy porn for Rob, thereâ€™s Theresaâ€™s reaction to consider, not to mention Canadian customs. â€œYes Sir, Iâ€™ve go booze, porn, and a big dildo to declareâ€¦ '),(18472,3902340,'26,April,2004','2004-04-26 00:00:00','   4-27-04  11H10&nbsp;   Holy Smokes, did I ever get some drunk last night. Just look at the drunken handwriting! '),(18473,3902340,'26,April,2004','2004-04-26 00:00:00','   2-26-04  23H 00   Thereâ€™s a secondary crowd cycling in now. The say Parisians donâ€™t get going until 11pm. So here it is! Only two tables left. Je suis lâ€™anglais seulement. I feel like such an observer, taking notes on society. Itâ€™s already later than Iâ€™ve been up all week. Câ€™est bon! This whole experience is cathartic, in a way. The guy at the next table is testing his cell phone rings. First a bird chirping and now bag pipe tunes. Hilarious. I think the couple next to him is German. The waiter is using English as a medium language and weâ€™re all just getting by.  &nbsp;  Did I mention that the Lux Bar is mentioned in my book on Paris? Another passerby is subjected to cat calls, en Francais.  &nbsp;  Those are definitely German cigarettes. I could start so easily on those sons oâ€™whores. But, I wonâ€™t. Remember Chernobyl!?  &nbsp;  Editorâ€™s Note: Heâ€™s really drunk now.  &nbsp;  Another couple lured from the sidewalk with cat calls from the gallery. Wish I had my cameraâ€¦ '),(18474,3902340,'25,April,2004','2004-04-25 00:00:00','   2-26-04  21H 55  &nbsp;  I spent some time with CNN International before I came out this evening. Daljeet Dhaliwal was great tonight. The entire deal was anchored from London. I can tell. The show that followed also featured two Brits, but I think they were out of Atlanta. Anyway, Iâ€™m not so negative on CNN International. Itâ€™s so much less American tonight.  &nbsp;  The entrecote is delicious, even though itâ€™s cooked very medium. Very Le Gavroche! Parts are much more medium than my taste. Ca câ€™est OK. Câ€™est Paris! Cecile laughed when I proclaimed this the other day. Cecileâ€¦  &nbsp;  Editorâ€™s note: Getting drunker now. Handwriting is getting sloppy.  &nbsp;  Itâ€™s so true. 10 million people. Flying into Paris, it looks like farmland. Nothing. Nada. Zip. Then you get in here and there isnâ€™t an inch of free space. Itâ€™s just like London, with five storey buildings with not a millimeter between them. The difference is itâ€™s more Mediterranean looking, with shutters on every window, everything white-washed. Itâ€™s much like Sydney, Australia - my all-time favorite city. Iâ€™d like to live and work there still.  &nbsp;  I canâ€™t believe Iâ€™ve let four nights go without getting out with the locals. Ca câ€™est terrifique! Across the street: Pharmacie Lepic and Tele France Video Club, where you can order a movie from an ATM-style machine. A neighbourhood dog has just arrived on the scene. A light brown lab. Sheâ€™s sweet and I bet she doesnâ€™t understand a word of English. Sheâ€™s looking for some love. Jâ€™aime le chien! Up the street Lintox Clubwear and Streetwear, Autour de Midi et Minuit, Nicholas-Wines.com, et Saint Dreux (Boulanger). A young teen couple just walked by, the girl wishing every diner â€œbon appetite.â€ I think she wishes she was 25 and sitting here with her beau. Audi Guy pulled up 20 minutes ago, loud music blaring. Heâ€™s totally euro trash. He parked illegally. Oppsâ€¦ Thought he was leaving. I guess he went back to get his Robbie Williams CD.  &nbsp;  Robbie Williams is everywhere! At Pont de Neuf, I saw an open air karaoke session with a poor French guy with horrible tone attempting to sing â€œReal Love,â€ half in English, half in French. It was horrible, just horrible. He is the William Hung of France. (Iâ€™m referring to the worst American Idol audition of this past season.) Then tonight, I saw an ad on telly featuring a guy who pulls up to a parking meter, sees the meter maid coming, and has no change so he starts to karaoke in order to raise meter fare. As he croons â€œReal Love,â€ the real Robbie walks by, stops, corrects the words, and contributes some money to the meter fund. Both doode and meter maid are stunned. I didnâ€™t catch the company name. Quelle Domage!  &nbsp;  I got a return email from Shango today. This is such a Shango city. He would die here. Câ€™est tres exotique. The sense of artistry is out of this world. Fuck, the world is a big place, and our last two generations are really the ones who have had unfettered access. First came the ability to travel long distances in a short amount of time, and now cost is not so much an issue as it once was. I am fortunate as I have nothing more to spend money on than lifestyle. Iâ€™m single, no kids. This is the time I will remember when Iâ€™m old. New York, London, Paris, Toronto, Sydney, Los Angeles, Chicagoâ€¦Iâ€™ve done it all and Iâ€™m not even 35. Itâ€™s a good basis for reference and thought.  &nbsp;  Editors Note: Even drunker now.  &nbsp;  Last night I watched the concert for Nelson Mandela and Aids relief in Africa, 46664.com. Dedicate a minute of your life to Aids. It was on MTV France (Europe). U2, Eurythmics, Beyonce, Youssou Nâ€™Dour, etc. And there he was: Bill Clinton, god dammit. America would be so much better off if he were still President. He is so educated, so worldly, so wise, and so fallible.  &nbsp;  I found my way to the Eiffel Tower today. Itâ€™s striking, but not as I had imagined. Thereâ€™s a great view of the city, but itâ€™s not viewed easily from everywhere. Thatâ€™s because all streets are hidden in 5-storey troughs, Iâ€™ve only seen the tower twice, both times on the hills of Montmartre. It was touristy, but a must nonetheless. Itâ€™s got security extraordinaire with Army guys with AK47â€™s. I imagine if youâ€™re going to go for Paris, itâ€™s the Metro or the Eiffel Tower â€“ but not with these guys on guard. Although, I donâ€™t imagine France is that high on the hit list due to Jacques Chiracâ€™s stance on Iraq.  &nbsp;  I have no croissants for the morning. Woe is the pity. More is the pity? Whatever, Iâ€™m drunk. What is that expression anyway?  &nbsp;  By the way, the Italian girl ditched a long time ago. And, there is a steady stream of renters at the video rental machine. Fuck, this weather is perfect. OB said they had snow on the weekend. Sacre blue! '),(18475,3902340,'25,April,2004','2004-04-25 00:00:00','   2-26-04  21:12H&nbsp;  &nbsp;  Well you can always depend on Pops for a little support, at least almost always. I just gave him a call with the update. Itâ€™s so obvious that he would probably like to travel, but heâ€™s got issues. Money, extravagance, or fear of it, and Eâ€™s health. Although I think the latter is more of an excuse than anything. Could just be another symptom of \"Scotia Disease.\" Colin and I are so lucky to have â€œgotten outâ€ of Nova Scotia.&nbsp;I remember how terribly bored I was there. Hell, Iâ€™m bored now and I fly so often I make birds jealous! I must say that Iâ€™m more fulfilled now with all the physical activity and mental extra-curriculars such as fiddle lessons. Balance is elusive. One could spend most of oneâ€™s time trying to achieve it, and miss the joys of living in the here and now. We just get glimpses of the sacred balance. Itâ€™s like a brief window opens up, the curtain parts, and there it is. Then itâ€™s gone again. I think how deeply you can connect with your passion on a daily basis determines how many glimpses you get behind the curtain. Insert another 0.40 Euro for another peek! (Euro bathroom reference)  &nbsp;  Show me a person who dares to be themselves to a fault, to allow their true thoughts, feelings, and raison dâ€™etre to be present at all times and Iâ€™ll show you a powerful person. Most of us never dare to scratch the surface of ourselves too deep. We shouldnâ€™t over analyze these things â€“ only when weâ€™re consuming good drink in Montmarte â€“ yet live to the max every day. Weed out those that wonâ€™t or canâ€™t live to be real. Iâ€™m not sure that any of these thoughts are new, just remembered anew.  &nbsp;  Iâ€™m presently sitting at the CafÃ© Lux Bar, across the street from Rue Lepic Studios, deciding whether or not to have something to eat. No one is eating, just drinking, and smoking. Carrie Bradshaw from Sex and the City was right! Paris equals cigarettes and dog poo, everywhere! Another random thought: Maybe my answer to sleepless nights is to stay up later!? J  &nbsp;  I called S. tonight, from Paris. Thereâ€™s something about her that I donâ€™t want to give up on. I left the number of the apartment. I really donâ€™t expect a call back in France. But, Iâ€™ll know like a two-by-four over the head, that the certain feeling is mutual if she does call. If I read this many years from now, and nothing ever happened, S. is the girl from Manchester who is a workaholic geologist. If this connection doesnâ€™t bear fruit, I shall move on poste-haste. Iâ€™m bringing a woman back to Paris next time, goddammit.  &nbsp;  Editor\'s Note: Nothing ever happened with S.  &nbsp;  Oh, Iâ€™ve figured out how to get to Paris Beauvais Airport. Thereâ€™s a shuttle that leaves at 6:55am from Charles DeGaulle Etoille Metro. Iâ€™ll scope it out tomorrow. Itâ€™s just 10 euro each way (one hour NW of Paris).  &nbsp;  One thing I still canâ€™t get over is the price of a drink in Paris. On average, itâ€™s 7 euro 50 for a pint of frigginâ€™ beer. Câ€™est La Vie. OK. The crowd at the cafÃ© just got really interesting. A woman just started yelling â€œKareem. Ciao Bella! Donnez moi ton numero de telephone.â€ Sheâ€™s very Italian. Heâ€™s very gay. He was just walking down the street. Sheâ€™s sitting with two other gay dudes. Câ€™est Europe, itâ€™s hard to tell. It seems as though Iâ€™ve stumbled on the local hangout. Iâ€™ve watched from my troisieme etage as this place rocks late into the night, well after 4am. Now Iâ€™m afraid I may succumb to its charms! The Italian girl has disappeared. Her drink remains. Perhaps she has cornered Kareem? I ordered the Entrecote, medium. I guess Iâ€™ve made the commitment. J Sheâ€™s back! She was in the toilette. I wonder if she had her 0.40 euro? '),(18476,3902340,'25,April,2004','2004-04-25 00:00:00','   4-26-04  14:25H&nbsp;  &nbsp;  Another night of restless sleep. But what began well ended well, and I got up again today at 10H! After getting my first French telephone call from Cecile letting me know that the maid is coming Thursday, I headed out on the #2 to Pere Lachaise cemetery to see dead people. It wasnâ€™t any more striking than the Cemetiere de Montmartre â€“ very beautiful and peaceful. The grave of Jim Morrison was uneventful. The usual crowd of celebrity grave watchers just like me. But, you canâ€™t visit Paris and not visit Jim! It brings to mind the classic visit that Mike and I made to Jimi Hendrixâ€™s grave in Seattle. We both had a Budweiser and Mike got his photo taken next to the headstone, looking contemplative and solemn. Reverence, man. Even though Jimiâ€™s gravestone was tiny, at least there was an image of his guitar etched in the stone. With Jim, there were just bolts sticking out of the cement where a bust once stood, before being vandalized. I just donâ€™t know what motivates people to vandalize a grave. You can go into the Louvre and walk right up and put your hands on a Rembrandt if you must, but nobody does!  &nbsp;  From Pere Lachaise, I rode all the way back across the blue line to Charles de Gaulle Etoille (Arc de Triomphe &amp; Champs Elysees). I was stopped by a guy with a New York accent asking for someone who spoke English. He was looking for a way to get to CDG airport. I pointed him toward the train, and then he tells me a sob story of being pick-pocketed and needing to find a cheaper way. He was calm and I didnâ€™t get a scam vibe, so I gave him 5 Euros for his troubles. Good Kharma. And judging from his reaction, I think I made his day.  &nbsp;  Je suis enroute de la Tour Eiffel. Je reste ici dan la rue Kleberâ€¦sitting in the shade on a sunny patio, streetside at the Fleurus CafÃ©. Iâ€™ve been watching a motorcade that is about to depart from a government building down the street. It has something to do with Burkina Faso. Itâ€™s too big an affair to be anything less than a head of state. The tip off is the van load of African media sitting in wait. La Circulation has just finished redirecting the traffic and shutting the street for a short time. The party must have moved on.  &nbsp;  The weather has been fantastic since I arrived. Blue skies and temperature right at 20 degrees Celsius each day. I think Iâ€˜m getting a fucking tan! It just warm enough that I can see the heat rising off the sidewalks, maybe 24? I havenâ€™t been anticipating the Eiffel Tower, simply because it seems so touristy. But the Musee Dâ€™Orsay is closed today, so what the fuck? Eh! Iâ€™m not even sure what Iâ€™ll do after that. '),(18477,3902340,'25,April,2004','2004-04-25 00:00:00','   4-25-04  18:00H  &nbsp;  Itâ€™s Sunday today, which turned out to be a great day to do the Louvre. It was a lot to take in, and surprisingly moving. In fact, one could spend three days in there. Da Vinciâ€™s Monna Lisa is of course the highlight, but hardly the lone star. Thereâ€™s Rembrandts, Rubens, Michelangelos, etc, etc, etc. I was awestruck to think of all that art, all that history in one place. And thereâ€™s still so much of Paris to see! The city reveals itself with every hour spent: little moments, a street performer, a coupleâ€™s embrace, a street scene.  &nbsp;  On the Metro on the way to the Louvre, a young woman raced to get on the train ahead of two boys with accordions. I watched curiously as she slung a drape between two vertical poles and began to put on a puppet show to a soundtrack on a boom box. It lasted two stops, and several people in the car rewarded her with change. One African woman flashed me that â€œCa Câ€™est Parisâ€ smile. Yesterday at the Metro some guy got on and did a little announcement to the whole car, a soliloquy, something about bringing peace to the world and wishes for a nice day.  &nbsp;  The easiest thing about Paris is the Metro. There are two stops right by Rue Lepic, one half a block away and the other about four. The latter being just down the street from the internet cafÃ©, where I learned the difference between an English and a French computer keyboard. My feet are quite tired, but Iâ€™m gonna make it! That reminds me. I must pick up some band aids!  &nbsp;  Iâ€™m having some vivid dreams, having to do with fulfillment. I seem to be questioning still who I am and where I am going. Paris forces you to ask the hard questions, or should I say she reminds you to ask yourself. Or, maybe Iâ€™m just full of shit! I do realize that in North America, weâ€™re living in an over-commercialized Wal*Mart lifestyle, where we live apart from our neighbours and are driven to acquire more and more stuff. I wonder if the majority of Parisianâ€™s can imagine what itâ€™s like to live in 2000+ square feet of living space, on an acre of land, with two SUVâ€™s in the driveway? Theyâ€™d say weâ€™re fucking crazy to live a mile from the baker, the cheese store, or the corner wine seller. The emphasis here is so much more social. Itâ€™s about the daily interaction&nbsp;between people. I understand the French now more than I ever have. Itâ€™s a culture built on verbal and social interaction, and romance&nbsp;avec passion â€“ which of course can be fiery! i.e. the rise and fall of the French Empire, French Revolution, and the legendary, historical Government about-faces. It all contributes to a sort of laissez-faire attitude. i.e. occupation of WWII. I think the attitude was pretty much, â€œOh fuck it. Will the boite be open as usual?â€  &nbsp;  Now, am I prepared to sell my car, my belongings, and move to Paris? No. But it reminds me that we might all give over more often to the romantic side, take flight in fancy, dare to dream, and say what must be said, to live with reckless abandon, with passion!  &nbsp;  One of the issues thatâ€™s sticking in my craw is Iraq. The further I get from home, and American media, the more I can see what is going on, or shall I say what is currently wrong with America. The more I read and see of the world media, what I see on CNN International becomes more irrelevant. Why are the Bush contradictions so evident to the world but not to Americans? Why do they allow their sons and daughters to die by the hundreds â€“ soon to be thousands, while they give Osama a free pass? Are they really sheep? What about their economy? Are they blind? How big a hole have they dug with regard to the national debt? Theyâ€™ve said â€œfuck the worldâ€ out of one side of their mouth and â€œbuy our goodsâ€ out of the other side. The time is growing near when the world will be sick and tired watching American TV teens dating each other on MTV Europe, etc. It\'s so insipid. Hell, China already makes all their clothes, shoes, household goods and Japan and Korea make the automobiles and electronics. Anyway, I see few Americans in Paris right now. And think France was always on their side. Think: Statue of Liberty. Nuff said. Bring Clinton back. He studied at Oxford! Decision made: Not going back to America unless thereâ€™s a regime change in November! End of political diatribe.  &nbsp;  I love watching the old French guys grooving down the road, cigarette in mouth. Sometimes a big belly, other times not. Some with chapeau, others not. Iâ€™m also struck by how multicultural it is here. Asians. Africans. Middle Easterners.  &nbsp;  I havenâ€™t had a vacation quite like this in a long time. Itâ€™s nice, slow, and unhurried. Iâ€™m even beginning to think in French phrases. The more comfortable I get, the better the response from the locals. I do miss the company of a woman. This is the ideal couple city. Iâ€™m not coming back without a girlfriend or a wife!  &nbsp;  Yesterday was another gorgeous day. A good stretch of â€œIl fait beau!â€ I walked around Ille de la Cite, visited Notre Dame de Paris, bought a prayer card for Elizabeth, the Rosary. Hail Maryâ€¦ It was a huge church built @ 1200 but was totally eclipsed by the Sainte Chapelle beside the Palais de Justice. That also dates back to about 1247 and was absolutely magnificent. Itâ€™s much smaller, two storey, with stained glass from floor to ceiling on all sides. The glass tells the story of Christ up to through the building of Ste. Chapelle and on through to the apocalypse. If you follow pane-by-pane from bottom to top, right to left, the story unfolds chronologically. Magnifiique! Right up there with the Louvre.  &nbsp;  I also walked through what I thought was the Right, but in fact was the Left Bank. It was everything I had heard about and more: streets and alleys of restaurants and shops for miles. Oh the humanity!  &nbsp;  The night before last I had a radio dream. Iâ€™ve had lots of those but this one was different. It seemed to suggest to me that I should find a way to create radio again, maybe via the internet. I saw a home studio and everything. Oh the time!! What is my head trying to tell me? This I do not need an answer for on day three!  &nbsp;  Oh I booked a quick trip to see Heidi in Edinburgh Saturday and Sunday next weekend for just 94 euros return on Ryan Air. Iâ€™ll get a chance to poke around the Royal Mile once more and an opportunity to get a great bottle of Scotch and hang out with the Heidster. Canâ€™t beat the price! Hope it works well. Well my dinner here at the Cockney Tavern was good once again. Itâ€™s a great place to have a pint and write up a storm. Oh boy. What will tomorrow bring? '),(18478,3902340,'24,April,2004','2004-04-24 00:00:00','   4-23-04  14:30H  &nbsp;  What a difference a day makes! Sleep, sleep, sleep. Iâ€™ve adjusted to the time change and started my day at the crack of nine to the sound of the accordion. Itâ€™s an old Frenchman sitting on the street outside the apartment. I opened the shutters to let in the sound and light. Leftover baguette (boul compagne), croissant, berries, and gruyere cheese. I decided to explore the neighbourhood of Montmartre today, consulting â€œMemorable Walks in Paris.â€  &nbsp;  After showering to the accordion music wafting up from outside, I gathered my things and headed uphill to some stunning views of Paris down cobblestoned streets and terraces. The sky was a beautiful California blue. The leaves are about 3-4 weeks behind Vancouver, but the birds are singing and the blossoms are on the trees.  &nbsp;  The sun is warm, but not hot, and thereâ€™s a nice breeze. Montmartre is the neighbourhood of Latrec, Picasso, Van Gogh, Salvador Dali, etcâ€¦all having spent time here and created pivotal works, inspired by the city and her people. The Sacre Coeur was busy but impressive, with the nuns choir singing at a noon mass. The much older St. Pierre de Montmartre was consecrated in 1147 and is much less of a tourist hangout. It seems that people make a pilgrimage to Sacre Coeur, and lounge about its steps most unceremoniously. I was met with my first view of la Tour Eiffel from the steps of Sacre Coeur. It was surreal to look over the city and see the tower dark against the bright hazy sky. Anti-climatic, or poignant? These are the thoughts that begin to invade the mind where the monotony of work once resided.  &nbsp;  I have stopped at â€œLe Cockney Tavernâ€ for a good ole Kilkenny, 7â‚¬50 per pint. (Thatâ€™s a good $11 for a beer!) The only thing English about this place is the dÃ©cor. Now I know how Susan must feel when she goes to the Outback Steakhouse. Gâ€™day Mate! I found a good little internet cafÃ©, already emailed OB, Colin, etc. '),(18479,3902340,'23,April,2004','2004-04-23 00:00:00','   4-22-04  20:08 H  &nbsp;  There is really only one way to spend oneâ€™s first evening in Paris: sipping red wine and eating cheese and raspberries in the apartment, at a table overlooking a lively street in Montmartre.  &nbsp;  I flew all day and all night from Vancouver, arriving at Charles de Gaulle airport at 8:20am. After a one-hour wait in a holding area, customs officials finally arrived to hustle the crowd through to baggage. No questions, no notice, no stamp for the passport. Itâ€™s certainly not like trying to enter the USA, our neighbours. I almost left the airport without my luggage due to poor signage and the crush of humanity trying to get somewhere, anywhere! The cab ride was uneventful. Another Mercedes, and speeds averaging 140km. I love Europe!  &nbsp;  I arrived at 9 Rue Lepic to find nobody waiting for me. Calm prevailed, and Cecile arrived within about ten minutes to thoroughly review the details. The â€œmaid,â€ an older French man, was cleaning. Cecile was the first French woman I encountered so far on my trip. She just had that â€œthingâ€ â€“ a very sexy presence that I imagine all European women must share. (With my track record, sheâ€™s probably a lesbian. Ha!)  &nbsp;  Even though many of the shops are closing, the street below continues to bustle. Frenchmen hustling home after working late. Couples strolling. Older women with dogs collecting ingredients for a late supper. Thereâ€™s an old lady across the street who is trying not to linger in the window, but sheâ€™s curious about me, her new neighbour. Itâ€™s her street, after all. 9 Rue Lepic must provide her with lots of fresh excitement.  &nbsp;  Daljeet Dahliwal is on the TV, on CNN International. She must have been offered big dough to leave BBC World. I saw her on Letterman once. He was smitten. And so was I. She is very enticing. Iâ€™m sure this channel is designed for traveling Americans. Do we really care what the consumer price index is doing in the US? Maybe Daljeet is in London, not Atlanta? I hope for her sake itâ€™s London. Now a story about reptiles reproducing without sex. Reptile sex!?  &nbsp;  Speaking of which, I was accosted by two different doormen at strip clubs down the street. One guy grabbed the arm of my coat. My â€œback offâ€ was met with a â€œFuck You!â€ I was nervous there for a minute.  &nbsp;  After the maid was finished, I crashed for about five hours to catch up on my sleep. Probably should have gotten up sooner. I may have trouble sleeping tonight. Had a great hot shower and went out onto the street to get provisions.  &nbsp;  I have no idea what Iâ€™m going to do next, let alone over the next ten days. I donâ€™t even know what has happened to me. I used to pre-plan everything, was never comfortable just traveling on the fly. Iâ€™m not suggesting that Iâ€™m â€œcomfortable,â€ just less anal retentive. Wonder who I will meet up with on this trip? '),(18480,3902340,'03,May,2004','2004-05-03 00:00:00','   5-3-04  17H 40&nbsp;   Weâ€™re about six hours into the 7Â½ hour transatlantic flight back to Canada. We flew over St. Johnâ€™s, Newfoundland about 45 minutes ago. My neck is hurting from all the strange pillows. So far, though, Iâ€™m feeling pretty good. Iâ€™ll call little buddy when we land in Montreal. I found a station on Enroute FM that sounds a little like JACK fm. The thought crossed my mind to pitch them on branding it \"SkyJack,\" but thatâ€™s probably not a very good idea!  &nbsp;  Editor\'s Note: This is&nbsp;the final entry from the Grande Paris Journal. To read from the beginning, click on April 2004 at the right to start at the beginning of the odyssey... '),(18481,3902340,'02,May,2004','2004-05-02 00:00:00','   5-3-04  10H 00&nbsp; &nbsp;&nbsp; I found the Maple Leaf lounge! There is one. Too bad I put up with â€œthe hummerâ€ for as long as I did. Itâ€™s champagne and barbeques every day, now. Did I mention seeing the new Snoop Dogg show on MTV France this week? â€œDoggie Fizzle Televizzleâ€ was a scream! Itâ€™s a fantastic sketch comedy. Iâ€™ve got to look that up when I get home. '),(18482,3902340,'02,May,2004','2004-05-02 00:00:00','   5-3-04  9H 00&nbsp;   Damned croissants! You get the flakes all over you. Iâ€™m sitting here at the gate at CDG airport. I checked in way too early as usual! Iâ€™ve got a chronic â€œhummerâ€ sitting behind me. Heâ€™s driving me crazy! He was behind me in the cafÃ© line-up and now heâ€™s followed me to this side of the cafÃ©, trying to make me crazy with his humming!  &nbsp;  Cecile ordered a taxi for me this morning which was right outside at 7:40am. Check-in and security was super fast. The security people are so lax. The woman didnâ€™t even watch her screen as my bag went through! I must say US and Canadian airport security is far superior. Heidi says that Europeans do it all behind the scenes, the only evidence being the Army guys with machine guns.  &nbsp;  Iâ€™m obsessing about gifts again. Iâ€™ve got something for Colin, Sweetie, OB, OE, Jeanine, and the girls at work. Do I have something for everyone? Who am I missing? Weâ€™ll begin boarding in another 55 minutes. I think Iâ€™ll scout out some reading material for the flight. Itâ€™s about 7Â½ hours to Montreal and then another 5 to Vancouver, but somehow the Paris leg is worth over triple the air miles. Que â€˜est que fuck!? '),(18483,3902340,'02,May,2004','2004-05-02 00:00:00','   5-2-04  20:55&nbsp;  I got back from Scotland about an hour ago. Iâ€™m sitting, waiting, for le derniere meal in Paris. What a whirlwind trip to see Heidi! The bus trip from Beauvais is more like an hour an forty minutes with traffic, and it was not without drama. A woman who spoke English, French, and Spanish got pissed off at the driver when she found out that the bus depot is at Palais de Congres, which is not exactly â€œdowntownâ€ as advertised. She spouted for about five minutes from the front seat, then got on the cell phone and started to yell at the person who was coming to pick her up. She got more and more agitated with every passing moment. The French guy behind her finally asked her to pipe down. â€œTu parle forte,â€ he said. She proceeded to let him have it in Spanish and capped it off with â€œVoila!â€ Itâ€™s a nice punctuation that I imagine must loosely be translated as, â€œSo there!â€ He flung a few words back and she hollered at the top of her lungs as she turned around, â€œI donâ€™t give a fuck what you think!â€ Well the bus was shocked. Everyone looked at each other and shook their heads. Then the driver intervened with a loud missive which must have been something like â€œThatâ€™s enough. Both of you pipe down. Iâ€™m the driver. Iâ€™ll pull over and put somebody off!â€ She simmered down then. I was shocked later in the ride when her little three year old daughter started to chirp up from the seat beside her. Some mother. What a role model! It turns out she had a huge amount of luggage, and judging by the state of the guy that showed up to pack her and her things into a Mercedes, she must be moving in with him, under some duress. Maybe she was pissed that he didnâ€™t drive out to Beauvais to pick her up?  &nbsp;  Ryan Air is a great deal. Itâ€™s â€œno frillsâ€ all the way: no seat assignments and no freebies. But they operate from low cost facilities, therefore thereâ€™s no big airport to navigate or long taxies before take-off. Both flights arrived 10-20 minutes ahead of schedule. It was well worth it, although it was a bit of a mad dash to get a good seat. I figured out the trick: enter the rear door of the aircraft and take the window seat in the last row. Then someone takes the isle and the rest of the plane fills up from the middle to the back, and you usually end up with nobody taking the middle next to you. I lucked out this way on both flights.  &nbsp;  Heidi rented a car for the weekend and drove out to Prestwick to pick me up. Itâ€™s about 25-30 minutes south of Glasgow, about an hour and 20 minutes from Edinburgh. I paid for the gas and bought the lunch and dinner. It was the least I could do. She gave me her bed again! Iâ€™m always tempted to ask her to join me, but I think it would be awkward for the friendship. Too bad because sheâ€™s looking great and seems really at peace with herself lately.  &nbsp;  We had lunch at the Hard Rock CafÃ© in Edinburgh and then stopped for not one but two bottles of Scotch at Cadenheadâ€™s. I had only planned on getting one, but then got a little carried away. This is going to be an expensive trip! Oh well. Last night Heidi and I had dinner at the Bouzie Rouge, It has a new name now, but thankfully the haggis was still on the menu.  &nbsp;  This morning, we got up, checked email, had a coffee and headed back to Prestwick. Heidi insisting on driving. What a doll!  &nbsp;  I do feel really de-stressed and have worked through so many feelings since Iâ€™ve been here. I feel like Iâ€™m going home to Vancouver with a mission to really kick ass in the business and live, work-out, date, and be social. In other words, work hard and play hard!  &nbsp;  Random Note: Iâ€™m enjoying a crÃ¨me brulee as I write this. Iâ€™ve got to kick the dairy when I get home. A woman has just stopped in front of the cafÃ© window in which Iâ€™m sitting, writing. She pulled out an instamatic camera and snapped a picture. I guess the scene was worth capturing. I guess I have fallen in-tune with the Paris vibe. Now my passport will be stamped with one entry and two exits for France in the same week. Iâ€™m hooked on stamps!  &nbsp;  Itâ€™s so much easier at the borders here, which makes me regret the state of things in North America all the more. I also regret the pictures of torture and humiliation of Iraqi prisoners on the part of the US military which are currently splashed across the front pages of the world papers. Talk about a fucked up situation. The reasons for being there are so murky and so poorly communicated to the troops on the ground. Even they donâ€™t have any perspective. I expect some of them have lost their minds in the midst of the chaos. It is SO Vietnam. Canâ€™t anybody see that? Doesnâ€™t anyone understand human nature?  &nbsp;  Iâ€™ve always felt I should have learned more about sociology and psychology. I am a student of human behavior, though, which has served me well in marketing. Anyway, Iâ€™m starting to ramble and Iâ€™m not even drunk.  &nbsp;  When I got off the bus today and transferred to the Metro, I was smacked in the face by Paris: couples smooching and people engaged in civil conversation.  &nbsp;  A Middle Eastern guy has just left the cafÃ© with his woman. They are carrying all kinds of household items they obviously bought today in the shops. He was trying to carry it all, refusing her help. He said out loud, â€œIn my Country, the woman cannot carry anything. Can you explain this to her? Thatâ€™s the problem with this Country. No respect!â€ Then he left her with the biggest piece - a painting. Not only was she left to carry the painting, but he left her to pay the bill also. Very funny stuff! Itâ€™s Sunday night and Paris sure is dead! Just couples out tonight. It was a big weekend with the 10 countries joining the EU yesterday. I think tomorrow is a holiday, as is May 6th, which is VE Day.  &nbsp;  I got an email from T. (from I.) She wants to see my pictures when I get back. Weâ€™ll see how things go. I had asked her out and thought I might have mismanaged it. A second chance maybe?  &nbsp;  This may be my last journal entry from Paris. Iâ€™ve got a long travel day tomorrow â€“ 15 hours to Vancouver. I checked my Aeroplan points and Iâ€™m up to about 65,000 already since last August. Thatâ€™s what I like to see! Thatâ€™s another business class ticket just about anywhere in the world. Where to next? '),(18484,3902340,'28,July,2004','2004-07-28 00:00:00','   My Dad is an old skool&nbsp;storyteller. You know, the kind of stories about having to walk twelve miles through the snow to school. I\'ve seen the old school house.&nbsp;It\'s right next door to where he grew up for gods sake!   Anyway, one of his yarns involves the family being so poor when he was a child that he slept on two kitchen chairs&nbsp;facing each other&nbsp;rather than in a bed. So when he was sick, we was \"sick in bed on two chairs.\"   I\'m sick in bed on two chairs this week. Or at least on a overstuffed leather couch! After the Mariner\'s game on Sunday, I retreated to the hotel where after about a half hour of air-conditioning I started to come down with a sore throat. There goes this week!   So I\'m convalescing at home, watching the democratic national convention on TV, and catching up on my expense reports (back to November \'03). Exciting, huh?   Too bad I had to cancel a date with T. from I.   Oh yeah...I just remembered it\'s fireworks night tonight. Bah humbug! '),(18485,3902340,'24,July,2004','2004-07-24 00:00:00','   Maybe it\'s a metaphor for my life and career lately, but it\'s been hot as hell here in Vancouver over the past week or so. I don\'t know why I think of 29-32 degrees Celsius as a heatwave. I guess it comes with the thinning blood of 9 years on the west coast. Good Lord. What if it ever actually  gets   humid  in the \"couv?\" I\'ve become a wuss!  What a great week! It\'s been fast, and furious.  On Monday night, I picked up a beautiful painting I bought at Lisa\'s boyfriend\'s art showing last Friday night. I own several of his pieces and didn\'t expect that I\'d walk in and fall in love. Fall, I did. However, the joy was temporarily tempered when Rob emailed four words, \"What, are you GAY?\" after seeing a j-peg I\'d sent. Anal retentive as I am, I polled some other art appreciators in the office and elsewhere and quickly confirmed that it\'s not even something someone could remotely mistake as a \"gay\" piece...whatever that is. It does have two naked women in it, though!    I got a last minute call from Muffer (Peter) on Tuesday afternoon letting me know he was on his way over from Victoria to conduct some work seminars out on Kingsway. We hung out a couple times during the week, the most fun being Thursday night. After dinner at Havana on commercial Drive, Jeanine, Muffer and I stopped by the The Eldorado beer store for a six of Canterbury and set out for the Cypress mountain lookout. We flew across the second narrows in the Acura just as the sun was setting on another super hot Vancouver summer evening. N.E.R.D on the stereo. Jeanine reclined in the back, dressed in pink - complete with the legendary pink Prada knock-off bag from New York - with an open can of beer, skirt blowing up, her batting it down. We sat on the mountain on a park bench overlooking the city lights, imbibing, enjoying laughs and the warm wind. Damn mosquito bites are driving me crazy!!! It was great to just hang out like teenagers. Jeanine had never met Muffer and I\'m glad to report that he lived up to his billing as a ladies man extraordinaire!  The other main event this week, is the emergence of an amazing business opportunity. It seems I\'ve reached another milestone in the journey from employee to entrepreneur. Just as I grow accustomed to the level of risk and comfort with the niche I\'m carving out...the ante gets upped. Oh man! Who the hell invented the concept of human development, anyway? Growth hurts. Why can\'t I just be content hanging out on Mazlow\'s first rung of the hierarchy pyramid? Yeah. Food and shelter, that would keep me occupied for a while!  Here\'s a little secret: I overthink everything.  There\'s a huge danger in taking too long to think, especially in business. I know this new opportunity is something I need to pounce on. My head tells me. My heart tells me. Emotions be damned. They only get in the way of a good thing.  I\'m on my way to Seattle tomorrow to see a Mariners game with a friend/business associate. The three hour drive will probably clear my head a little. I\'ll take the computer and get a little work done Monday before I come back.  I\'ve made Monday night dinner plans with the Ex. She\'s seeing another doode, that is, the lastest in a short series of suitors since the divorce. I\'m trying not to let it bother me. But what can you do? I guess I\'m still a little disappointed that we could not find a way to make it work. But, it\'s been two years...so I better let that go.  There I go obsessing again...  \"I want my pain. I need my pain!\" - Captain James T. Kirk, Star Trek, NBC, 1967 '),(18486,3902340,'15,July,2004','2004-07-15 00:00:00','   I don\'t have any pictures of Russ and me getting drunk and playing music the other night, but I do know what my living room looked like the next day! '),(18487,3902340,'13,July,2004','2004-07-13 00:00:00','   The planes are lined up out in the west, coming into YVR, the sun is sinking behind Bowen Island into the Strait of Georgia, and Elton John\'s \"Songs from the West Coast\" is chugging along on the hifi.     It\'s July in Vancouver. This is technically my first blog, although I\'m preparing behind the scenes to launch my grande Paris journal, which chronicles my every thought during my spring trip to France. I think I drew more words than breaths while there!  Reading my notes from the trip has been interesting. It\'s mostly forgotten thoughts and entertaining ramblings (at least mostly entertaining). But, I\'ll be glad to share my European state of mind in the next day or so. '),(18488,3902340,'14,August,2004','2004-08-14 00:00:00','   Some things are just good.  This past Thursday evening, \"my posse\" gathered at Spanish Banks beach, about ten minutes from home. We unwound from a long day, laughed, barbecued, drank beer and wine, and some of us even OD\'d on Nutella. Needless to say, those of us who arrived in a bad mood, got over it pretty quickly. No names needed. LOL       \"Canuck Place\"'),(18489,3902340,'07,August,2004','2004-08-07 00:00:00','   We had a great time last night at the Morrissey! Here are some photos from the birthday \"do.\"       Jeanine, Heike and me.        The girls, plus Rod.'),(18490,3902340,'05,August,2004','2004-08-05 00:00:00','   You know, birthdays are awesome when you\'re surrounded by good friends and some family! It makes turning 35 not such a horrible ordeal. And today was a really good day that ended with a dinner at my brother and sister-in-law\'s home. Their in-laws are in town from Melbourne. So we were all treated to a wonderful Australian-style home cooked meal of lamb, potatoes, and pumpkin. (Because, of course, pumpkin is a vegetable in Oz and  should never  be served as a fruit in a pie or as any dessert!)     The company was great, the food was wonderful, and it felt good to hoe into a birthday cake baked from scratch.    Inevitably the drinking stories started and of course I told the famous story of a 19-year old Little Buddy (Colin) showing up on my doorstep in Halifax at three in the morning loaded to the gills on beer and vodka after a Friday night of carousing DT (downtown). Halifax is a legendary port party city and the drunken shenanigans are legendary. Because I lived right downtown at the time, it was probably a more appetizing idea for His Drunkenness to turn up at my digs than pass out on Province House lawn which is the usual rite of passage for liquored teens.  After ignoring the ringing for about 10 minutes, I finally answered the phone and buzzed the little bugger in. As I stumbled sleepily to the door I could hear him bumping into things as he navigated the hallway. As I opened the door, he steadied himself and boozily, triumphantly announced, \"What do you think of your drunk brother?\"  What do you think of your drunk brother.  I could have rung his fucking neck. I was working weekend mornings on the air in Halifax radio, with a 5am wake-up and a 6am start. I might have even had a few myself earlier in the night. \"Get in here you little Bastard,\" I ordered. It took about eight steps and then it collapsed on my couch, with one foot under a blanket and one still on the floor, snoring almost instantly.  That\'s exactly where he was when I left for work just after 5am, and that\'s exactly where I found him at 12:30 in the afternoon when I got home. Moved not an inch.  \"There\'s only one thing to do,\" I steamed to myself.  With one quick call, my old roommate and neighbour in the building, Gary, arrived on the scene with his hockey stick in hand. He started poking at Little Buddy with \"the stick\". \"Wake up Champion,\" he bellowed, \"or you\'re getting the stick. C\'mon ya bastard! Git up.\" Gary is an ex-pat Cape Bretoner who enjoys any opportunity to take the piss out of anyone, especially when the targeted party is at a humourous disadvantage.  The little fella stirred, obviously hung over and hurting from the alcohol antics of the night before. \"I got some pea soup for ya buddy,\" Gary prodded in his best \"Bubbles\" voice, ten years before Bubbles from Trailer Park Boys was even dreamed up. \"I brung a ham bone for you to chew on too.\"  \"Fuck off,\" Colin moaned, choking back bile.  \"C\'mon, get up, get up, get up,\" Gary poked.  And he did. He got up. Sonofabitch got his just dessert for showing up at 3:30 in the morning and waking my ass up out of bed on a work night!       The Culprit (#2) - 12 years later!'),(18491,3902340,'02,August,2004','2004-08-02 00:00:00','   Yesterday, I skinny-dipped in a mountain stream with another guy, then picked wild blackberries and made a pie with him. Oh my gawd...I\'m turning gay! Not really...  I set out for G & K\'s place on the North Shore at about 3pm, determined to get some exercise after a week of sickness and inactivity. G and I set out along Lynn Creek in short order and after about 20 minutes we scrambled down over the bank to the best natural swimming pools ever, right in the creek. I took off my socks and hiking boots and waded in up to my knees, trying to keep my balance on the slippery rocks on the stream bottom. After about three steps my feet went right out from under me. Those goddamn rocks were slippery! I clamoured back up out of the drink, bruised legs and all.  G said, \"You won\'t be shocked if I get naked and jump in will you.\" \"Christ no,\" I said. No sooner had the words left my lips, there was a flurry of lilly white tuckus and a big splash. \"Oh hell,\" I muttered, and followed suit. OK, I\'m not usually the type to just drop trou in public even if it is the middle of the wilderness....let alone in front of another guy (call me a prude)...but I figured you only live once. It figures that I would pick this moment to unbriddle my inhibitions because -and wouldn\'t you know- I\'m up to my goodies in icy cold water and I hear a distant noise. It might have been a gasp. I turned, and standing at an elbow in the river about 400 feet downstream was a group of hikers staring at me - in my full glory. \"Screw it,\" I moaned and took the plunge. The frigid water closed in on my flu-weary chest and I instantly remembered how sick I\'d been just a few days earlier. G was like a friggin\' seal, already up the creek around the bend and out of sight. I flitted about for 4 or 5 minutes and then dragged my bits and pieces back up on the bank and hussled back into my clothes. The hikers were within earshot now and kept coming. Two dads with babies in backpacks, and a young boy and a pre-teen girl in tow. They were so consumed with navigating the river bed, they paid no attention to me, or G who was now dragging his shivering arse out of the water, aware now of the approaching, inattentive brood. What an invigorating experience!  On the way back, we stopped and picked wild blackberries - enough for a pie. When we got back to G\'s place he set to work at simultaneously preparing halibut on the barbeque and baking a pie. Needless to say we overindulged in both, including beer, blackberry vodka martinis and red wine. We were in a right state when K came home. Of course she was sober as a judge and G and I were feeling no pain. G stumbled over their antique clock while trying to reach for a light switch, sending it shattering into pieces on the floor. I may have lost my cookies in the middle of the night...it\'s a long weekend in British Columbia for gawds sakes - nekkid and baking a pie! '),(18492,3544112,'27,June,2004','2004-06-27 00:00:00','   urlLink    Jade : One day, when I was young, my Dad brought home a guitar.  I was quite intrigued and went to pick it up but he said,  \"Son, unless you can wail on that thing like Nuno Bettencourt  from Extreme or maybe Steve Vai when he was playing with David  Lee Roth and he had the sweet double-necked guitar that was like two legs coming out of a heart, keep your damn hands off it!\"  And I never touched a guitar again. So to answer your question, yes, I answer the phone whether it rings or not. &nbsp; urlLink  '),(18493,3544112,'25,June,2004','2004-06-25 00:00:00','   \" It\'s nothing .\" you say and pull your arm away from me. I can see the blood beginning to seep through the fresh bandages that cover a large part of your arm. and I know I don\'t really need to ask but I do because it\'s easier to pretend .  \" You\'re bleeding. What happened ?\"  You sigh and roll your eyes because we\'ve gone through this so many times its becoming monotonous.  \" It\'s just a scratch, nothing to worry about.\"  That\'s what you say anytime you want me to drop the subject . We both know that it doesn\'t really matter because I won\'t and you\'ll keep trying to evade the question. trying to feed me half-truths. you tell me it\'s for my own good. it won\'t do me any good to worry about you. finally after a couple of minutes of being questioned you let me look at your arm. and it\'s covered with red angry cuts that criss-cross and seem to go on forever. some have stopped bleeding but others are still seeping. for a minute I say nothing. then I meet your eyes, and you\'re smiling so I smile back. \" I\'ll get some fresh bandages.\" and i\'m out of the room before you can say anything.  '),(18494,3544112,'25,June,2004','2004-06-25 00:00:00','   I might drown in your eyes. In the light that they hold. get caught in the sweetness of the words that you speak. your lips touching mine until  your breath is all I can taste. I might die of thirst . If I can\'t have your kiss. though your words could be poison. yet taste sweet through your breath. that it would not matter if you brought on my death .  '),(18495,3544112,'23,June,2004','2004-06-23 00:00:00','   Behind closed lids explosions of colors. blue , pink , green . lips that sometimes are sweet. sometimes  taste of nothing. dry, bleeding , tasting me. cold.  \' What are you thinking ? \' whispered close to my ear. breath so warm. close. raising  tiny goosebumps along my arms.  Behind closed lids you are a million different sensations. sometimes soft  and warm , feather-soft touches. other times pain. I hold my hands up to your face. hold my arms. can you  see them ? can you ? feel them ? some are you\'rs ... and I really don\'t remember which ones anymore.  and they never really hurt ...I wanted them to hurt. look closer. yes, that\'s it. you see them. look. do you want  to know.  \' How did ...\' you begin but never really finish the question . I didn\'t expect you to. and so instead  I let you bury your head in my arms. stroke your hair. hear you whispering something. soft, under your breath.  can\'t make out the words. close my eyes. wonder if I can make the tears come. maybe ? maybe not. but it  doesn\'t really matter. you\'re looking at me now. I can feel your eyes on me. open my eyes. stare. smile.  \' What are you thinking \' you ask . eyes so dark they threaten to drown me. \' Nothing \' words forming on my lips. coming out of my mouth. even before I can stop myself. \' Nothing \'  '),(18496,3544112,'22,June,2004','2004-06-22 00:00:00','   Your lips are soft against mine. they move slowly and leave a trail of wetness behind. and even after you pull away i can still taste you. a mix of cigarette smoke and beer. you lean back in your chair and watch me with an amused look in your eyes. like you\'re waiting to see what i\'ll do. and seeing how i\'m not one to dissapoint -- i lean into you until your lips are touching mine .it\'s always soft at first. lips barely moving against each other. almost like neither one of us really wants it to end too quickly. i feel your tongue brush my lower lip . sigh. but not sure if it came from me or you. and it\'s not like it really matters right now.  Feel your fingers against the side of my face. moving slowly until they are tangled in my hair . and it\'s beautiful. what you make me feel. I melt beneath your touch. and your kiss is fire and something else that i can\'t really describe. soon your tongue is inside my mouth. but all i can really think about is how incredibly soft your lips are.  '),(18497,3544112,'20,June,2004','2004-06-20 00:00:00','   your hands are always so cold . like you\'ve been out in the middle of winter without your gloves on. that\'s why I hold onto them so tightly . or at least that\'s what I tell myself . you barely notice all the time that has gone by since we last saw each other . you smile and wrap your arms around me . and I\'m surprised to discover that you still smell like peppermint and smoke . I smile too . maybe I\'m more surprised by the realization that I still remember what you smell like . smile . your lips feel so warm against my cheek . you press harder into me and I realize you\'ve been saying something this whole time . \"come with me\" I hear you say close to my ear . and my stomach wants to drop down to my knees . I hate that after 3 years you can still affect me so very much . I hate myself more for letting you . or maybe I hate myself for letting you go in the first place . your hands are always so cold . I rub them between my own trying to warm them for you. I tell myself that\'s the only reason I still haven\'t let go . that\'s the only reason I seem unable to let go .  '),(18498,3544112,'18,June,2004','2004-06-18 00:00:00','   HE\'S MARRIED . and I\'m the last to find out . I can\'t believe none of my \' friends \' told me anything because I am sure they had to know something about this. now , I realize how idiotic it is of me to get pissed about this . I mean it\'s not like I expected him never to get married but ...still ..I feel like picking up the phone and calling him up and telling him to come over . like that would happen. should this hurt this much ?  '),(18499,3544112,'18,June,2004','2004-06-18 00:00:00','   You\'ll never reach me. I\'ll fall before you touch me . feel your fingers against my wrist . pick me up from the floor. I\'ve left my blood on your bathroom tiles.  '),(18500,3544112,'17,June,2004','2004-06-17 00:00:00','   It burns against my skin. sometimes it slices through with such ease that i donâ€™t even feel the bite of the blade.  cut. spill against my skin.  '),(18501,3544112,'04,June,2004','2004-06-04 00:00:00','   We were laughing about something silly I said. thatâ€™s how it started. you teased me about my laugh. pointed out that when I laughed I always wrinkled my nose . felt you move closer and didnâ€™t think anything about it. or maybe I did and didnâ€™t care about anything other then how much fun it was being with you.  '),(18502,3544112,'16,July,2004','2004-07-16 00:00:00','   \' don\'t go\' he says and i can feel his fingers tighten around my wrist. we stay like that for a few seconds. me staring down at his hand trying desperately to remember why i want to leave. he watches me - fingers slowly releasing their grip . i look at him now.\' don\'t go \' whispered so softly it might have been lost if it wasn\'t so unbelievably quiet . and his voice seems to have replaced his hand on my wrist because i\'m still standing there. unable to move away. feeling his eyes bore into mine.  '),(18503,3544112,'15,July,2004','2004-07-15 00:00:00','   Hold your hands in the sink. try and make sure the blood won\'t get on the floor. because I really hate cleaning blood. or maybe it\'s just the cleaning that I hate. I watch it slide down your fingers and splatter in the sink. red against white. and all I can think about is the mirror and how I\'m going to have to explain it to them.  catch you staring at me. and you\'re smiling at me. turn the water on . to try and wash the blood off your hands. hear you wince from the pain but you say nothing and so I continue with my task. after a few minutes I finally turn the water off and wrap your hands in a towel. all most immediately the blood begans to soak through. I mention we should go to the hospital. you\'ll need stitches but you shake your head and tell me not to worry.  \" Can I have an aspiring, though ?\" . Place them inside your mouth and hold the glass of water against your lips. find myself facinated by the water that\'s dripping down your chin. sit down on the couch and stare . I can tell you want to say something but you seem to be struggling with how to put it. finally your eyes rest on my face and you say in that soft voice of your\'s that I\'ve come to love. \" I promise to replace the mirror \"  '),(18504,3544112,'09,July,2004','2004-07-09 00:00:00','   there are always words sparkling just above the surface . spreading above me like stars. burning bright in my mind and disappearing before I can string them together for you .there are always words I want to say to you. thoughts and ideas I don\'t seem to be able to express properly . there is always a lack of words when you are around -- never quite right for what I\'m trying to say . and you are always there . sometimes barely brushing the edges of my thoughts . other times consuming me . and you are always there . making me wonder why I ever believed I could capture with words what you mean to me . there are always words .  '),(18505,2863105,'07,June,2004','2004-06-07 00:00:00','    Ito ang panimula sa aking pagsusulat sa aking sariling wika upang mas lalong maintindihan ng ibang mambabasa lalo sa mga Pilipino na nakatira sa ibang bansa o di kaya\'y mga banyaga na gustong matuto ng Wikang Filipino.  Higit sa lahat ang mga kuwentong aking isasalaysay ay naging bahagi na rin ng aking karanasan at nais ko ring ibahagi ang mga ito sa aking mga anak, mga apo at ang mga apo ng mga anak ng aking mga apo hanggang sa kadulo-duluhan ng aking buong angkan.  Kung anumang kamalian sa pagsusulat ay sa akin po iyon at walang iba na puwedeng sisisihin.   '),(18506,2863105,'24,March,2004','2004-03-24 00:00:00','   A thing of the past is also a thing in the future though the dimension of its occurrence is different from the timeline of its event.  Anger is a sign of immaturity and the fall of prematured wisdom.  To work with honesty is a blessing to the Lord.  The first work of dishonesty is the beginning of all corruptions.  There is only one way to win every heart and mind of the people: to work with an honest labor.  '),(18507,2863105,'20,March,2004','2004-03-20 00:00:00','   There is one thing that we need to think about...our thoughts.  The first pain is lasting, and pains that come last are very disturbing.  Laws are made for you to see the reflection of your bad images.  Nothing works when nothing is done for something that never have been done for nothing.  We seemed to be amazed of something that we sometimes believed to be worthy of wondering.  '),(18508,2863105,'17,March,2004','2004-03-17 00:00:00','   Creativity begins from the space of imagination drawn by the results of work.  When a man stands eloquently for a righteous order, his line can be seen upturned by the people who knew him well.  Wisdom is the master of manâ€™s destiny drawn by the portals of knowledge making the man different of its own.  An expounded opinion derived from the synthesis of a given order can forestall the hypothesis of undiscovered theory.  Deviation from the common view is like extracting gold from the mouth of an elephant or silver from the claws of the tiger.  Politics is the collective bargain of what can be promised and what the people can gain.  In every game youâ€™ve played, donâ€™t count how many times youâ€™ve lost.  Instead, count the number of times youâ€™ve given your opponent the hardest encounter ever had from your very best efforts.  In every success, we donâ€™t have to draw a line that demarcates ourselves from the failures of others.  In trimming the edge of arguments, skillfully know the things that you would like to gain out of it.  There is only one man who will never learn --- the man who forget about himself at all.  Impatience is an engrafted view from hatred and an overflowing disaffection of the human spirit.  The way to a womanâ€™s heart is a truthful talk generously expressed from the eyes of a loving heart.  The law concealing the intention of its maker belies the veracity of its contents.  Tolerance in friendship is unlimited, but sometimes friends outdone the limits of tolerance.  In all of our experiences that we have, we can always consider that pains in the heart are the best teachers of reasons in wisdom.  When the world will filled with honorable men, you could never imagine what will happen to the future.  We donâ€™t have to compare a pretty face to an ugly one for both have the same attributes of beauty that can conquer a heart.  No matter what have you gone through with your studies,  always remember that the basic foundation of life which you could prove its worth is Patience.  A piece of advice:  Do not spoil friends who will give you the unexpected time of humiliation, rather choose a friend who can give you lessons to live by and to be with.  Nepotism in the government thrives when the functional elements of its instrumentalities wear the crown of their alterego by resentment.  We virtually give ideas but not the usual conclusions.  When we try to conclude, it will be very amazing that we could not even believe it to ourselves.  The house with silent atmosphere can be differed to a silent river flowing deep into the ocean of uncertainty.  Sometimes, let us not adapt the way we thought it is, but the way we think that it is and will be.  Conclusions differ greatly with judgment only when we put ourselves to the thought of what we should be concluding.  The decisions that we are going to make are sometimes as fatal as the storm, but gentle as the rain drops.  Actions are governed by the rules of emotions so that the reaction of our mind depends upon the acceptance of the stress.  Sometimes, the subconscious mind denounces any illicit acts of the body but unknowingly it did iniquity.  The most difficult thing when people decide is the way and means on how they would make up certain decisions.  When people use peace as an instrument of war, the factor to decide usually comes from the footsteps of their ancestors having no traces but only dreams to follow.  A pauper who knows to handle the ways of life is worthy than a king who awaits the advices of his Council.  Big wars come from smaller ones depicting one great cycle of silent metamorphosis of peopleâ€™s awareness.  A man who is seeking something new may forget the ladder to a camelâ€™s back and even censure wisdom to a lionâ€™s portal.  A cup of sweetened lemonade can be resembled to a cup of sweet milk containing insatiable taste and vexation of the human spirit.  Manâ€™s creativity comes from the oblivious space of his own wardly spirit.  Patience is one of the unique virtues of common sense and one of the probable areas of ingenuity.  The hyperbolic view of a certain synthesis comes from the unfounded baseline of extended facts.  A reason that divides can be one of the facts that connect the roads and bridges that multiply the wisdom of the human spirit.  To trim the edge of certain argument, we must be sure enough that we knew certain facts whence it began.  The law that conceals the veracity of its content can deceive the intentions of its own maker.  One win is the greatest contribution to a competition between players and never a hundred times or more.  It takes a million pains and only one great sigh needed in order to know your real friends.  '),(18509,2863105,'17,March,2004','2004-03-17 00:00:00','   Every written words when sought from God is a living testimony of those whom He manifest as his ever great wisdom and blessings to humanity.   The true measure of a righteous man is the truthful intent and righteous goodness of his heart.  When he is not, he will surely falter. (George McKinsley)  '),(18510,2863105,'17,March,2004','2004-03-17 00:00:00','   When a man speaks in front of many people and could not look at the eyes of those who are listening to his delivered words, there is that great tendency he\'s been speaking of his own folly.  '),(18511,2863105,'17,March,2004','2004-03-17 00:00:00','   Obedience to God\'s Commandments is the highest form of worship.  '),(18512,2863105,'16,April,2004','2004-04-16 00:00:00','   When a man looks his eyes forward, without feeling his feet touching the ground, will surely miss the mark he left behind.  There are two reasons why every person is interested to enter a race:  First, it\'s about the prize and fame; and the second is that it will never come again this far.  Election is a time for winding up the rusted pen for voting; and even exalts the rugged terrain of dreaming.  When one decides, everybody is the spectator of it.  Not everyone yet realized that he is one of those persons who really is meant for something he dislikes much and of everything he would not expect.  Politics is a mingling of ideas with a lot of promises, dreams, visions, and inspirations.  Most of all, it is a vineyard for the slaggard and a paradise for the witty.  '),(18513,2863105,'09,April,2004','2004-04-09 00:00:00','   A real leader pays attention to words of experiences that draws the path of righteous endeavor.  A good leader is a good listener.  Constructive views sometimes dwells only among those who are willing to share it; while destructive views ruins the origin of it.  A clever follower always dare the master by measuring his patience.  Words of mouth delivers the piercest battle.  A real pain pierces through a hardbound rock of afflictions.  Count no hours in your work and the hours will never be counting what you have done for the day.  Impatience to every little difference results to more vaguenes of reasons.  '),(18514,2863105,'14,May,2004','2004-05-14 00:00:00','   A man who thought of winning an election means to pull down his opponents and bring himself out to be the best among them, is a man who\'s been molded with the culture of corruption.  A boastful heart wins a lot of friends, but at the back of his head, an enemy is waiting to strike him down.  We gain a lot of surprises when you least expected it.  People\'s choice becomes sacred when out of the dungeon a hero eloquently stand by their midst; but with a corrupt leader, people becomes radical.  No one could win an election with a finger pointing at the heavens; while his heart is wearied by the burden of his foes.  '),(18515,2863105,'03,May,2004','2004-05-03 00:00:00','   Spiritual pain kills the heart.  A loving friend always keep a promise and never falter nor find ways of evading.  The sincerest heart we could ever find is the heart of truthful brethrenhood.  No matter what it takes to succeed, never try to forget that your feet is still touching the ground.  One thing that can make your brethren happy or sad is your heart\'s content.  A soothing pain beats the heart.  '),(18516,2863105,'16,June,2004','2004-06-16 00:00:00','   Cunny people with wise mind sometimes overshoot their talents of expressing their thoughts to the doubting suppositions of the simple people.  It is better to listen than to insist to be listened to.  A closed heart means a closed mind.  When simple thoughts come in a way to enter into the closed heart, it will begin to sprout like weeds to corrupt a closed mind much in going into the most darkest part of it.  When you doubt, you will surely mean no reason to listen.  Fraud is what corrupt minds insist.  Doubt is the way to a prideful circumstance of your soul.  When you see the darkest part of you, you will know that the lighter part of it is the place you first notice when you never doubted anything.  To convince a naked woman to wear her clothing is the worst part of seeing yourself in companionship with her.  Who wins an argument?...No,not one. Even those who won it fails.  The only guidance against doubt is when you see others are trusting.  Call it fraud...call it anything you want. Still you are one part of it.  When people start to doubt the government, it is as if they are doubting their own.  '),(18517,2863105,'14,May,2004','2004-05-14 00:00:00','   May 8.  Two days more to go. Leaders of the candidates are drawing the winning streak. Strategies formulated from the past election experience was soon to be obsolete. Voters stigma seems a burgeoning malady on the side of those who wills to win a seat for the local elective posts.  This day seems to be an never-ending analysis and decision-making process for all the candidates. Telephones and cellphones rang intermittently or even continuedly without let up as the field agents of each candidates are trying to envelope upon their own efforts the burden of convincing the voters to their side.  The voters are adamant and would not let go of their enticing.  They have decided already. Others are a little bit confused of what\'s going on, expecting that \"something\" is going to be cooked up by the aspirants. They get ready into it. They waited along the streets and highways were luxurious cars and wabbled-buggy trucks and jeepneys are driving fast beyond its natural pace. These vehicles were clogged with streamers all around equipped with microphone and amplifiers so to speak. Tricycles on the other hand race their tracks with almost the same equipment of microphone and amplifier.  Moreso, some of the barangay officials and barangay employees were having meetings with the candidates surreptitiously with unknown envelopes where people have been thinking about either money is there or a couple of instructions from their favored candidates was mailed by hand. Nobody knows.  The political stalwarts in the province of Camarines Sur, on the other hand, are trying to analyze their winning stance.  They called a meeting with their henchmen and asked them what\'s going on at their respective places of coverage. Monitoring was on the go! Not one can miss the opportunities of knowing whatâ€™s going on in the field.  People need to be contacted for certain conferences, either overtly or covertly. Political financiers are also tagging their means to assure their supported candidate that the needed funds is already available in the bank for disposition.  Some even have been withdrawn to finance the miscellaneous undertakings at hand to the joy of the candidates and their minions.  Some candidates for senators have already conntacted their people in the province assuring them of support when they get elected.   The Political spectrum now on the verge of either disunity or division.  But the candidates are trying to amass a lot of would-be supporters an their backyard.  Keeping them, by providing foods and snacks, or even liquor drinks and others.  People are getting used to coming the house and enjoy the prepared refreshers and hospitality shown to them by the ushers and usherettes who are either hired or volunteers.  May 9.  The expected rest for all the campaign sortie, but the campaign is going to be formally ended at 12:00 midnight, so, thereâ€™s still a lot of time for visiting some friends who were not reached during the campaign. Relatives were coming in to visit their candidate even those who are not within the jurisdiction of the political division of each towns and cities. They come like a swarming bees equipped with fallacious greetings or even bad news from their supposed supporters. Everytime, they have with them solicitation envelopes related to their fiesta celebrations or inter-barangay tournaments and many other insurmountable reasons that they could get just to obtain the needed funds for such an activity. (Sssshhhh...this is secret:  Some of the money raised were kept inside their pockets specially those envelopex kept opened.  The same envelope is being used to another candidate...ha ha ha...wise enough...yes, itâ€™s really wise enough to raise funds easily with the lofty spell!)  May 9, 12:00 midnight. Vehicles were running around the highways and barangay roads like a racing event.  People are picked up by certain political leaders for an interview with the candidates; while on its way, the fetcher negotiates with a big smile.  He proposes something much promising than the others did. Thence, the envelope which contains the amount as proposed is being handed down to the expectant recipient.  Sometimes, the travel would not last up to the supposed destination, it stops in a certain place then letting go of the picked-up people.  (This is another technique of vote-buying.)  For all we know, inside a car or any vehicle  for that matter is a nice place of vending votes!)  Meanwhile, candidates with insufficient funds could not expect a high trend of voting opportunity.  Why? People would go for a price! Never mind the sacredness of the ballot, itâ€™s the time to collect from the candidates, they said.  Nighttime, sacks of rice on board a truck was delivered everywhere.  The destination is sporadic.  People are amazed of such mobility and the swiftness of delivery. Thence, leaders are calling up the number of people who will help them re-pack the sacks of rice by kilos.  Everybody participated with amusing smiles and surprising astonishments.  May 10. D-Day. Everyone is busy preparing the breakfast, snacks, and lunch.  The candidates have given the fund or some of their friends have delivered a number of kilos of pancit and meat and other ingredients purposely for the refreshment of the voters.  Stories ran out. News are everywhere that certain number of sacks of rice were delivered to such and such.  Others have already collected their share. They encourage others to do the same. They followed with blind obedience.  Voters are getting started to swarn the voting places; looking up their names on the list while other find their names, a few murmured finding themselves out of the list.  They reported the matter to the local COMELEC office.  They were denied to vote for nobody have known them in the place.  Others would not want to support because they are identified with the opposing group.  Not one lift their hands when the COMELEC officers asked a confirmation inquiry about his/her personal circumstance. Very intriguing indeed.  3:00 P.M. Voting is stopped, counting of ballots commenced.  The watchers are now languishing their stomach aches.  They needed to get food and some rest.  It take a matter of time to count the casted ballots.  In their absence, manipulations are going on; while other remained blind, others are openly doing the risk of their lives!  May 11.  A lot of scenarios were published and even broadcasted in all radio stations.  The local newspapers and tabloids are getting ahead of hand for people are amused with the nude pictures that attracts them.  Never mind the news headlines, just get on with it later.  The national televisions are having their own hay; interviews are everywhere. Radio stations are playing wild, banterings and bickerings are everywhere; no one knows what to believe and what is not believable. Take your pick....  These four days observations by this writer is a clogged-up dues and I could not keep it inside my heart and mind.  I need an outlet.  Indeed, Iâ€™m lost. Confused. I needed to shout at the hilt! I needed to relax and let my fingers roll in front of the computer.  This is my only consolation as I become witness of these tremendous burden that my own people themselves are hauling for their own load.  I can only say that the future of my country may look bleak but there is still faith that embraces all in necessity!  '),(18518,2863105,'13,May,2004','2004-05-13 00:00:00','   One of the typical example of the ill-effects of mudslinging during a political fora or political exercise is the silent turn-around of silent voters to sympathize the mudslinged candidate.   Peaceloving people would not want to propagate the culture of corruptive black propaganda during election campaign.  They would rather prefer a weak candidate for him to be developed while in office than a claimed-to-be expert who knows no bounds in slinging black propaganda against his/her fellow candidates.  People would rather vote for a neophyte candidate than to cast their sacred vote to a wanton oldies who disregard the feelings of the relatives of their opponents in a mouth-battle-tested, arrogant and pernicious talkative.  They would prefer to a sympathized-vote for a silent but a surely potential young candidate who has the dynamism to improve while in office than ferocious warrior-like candidate who promises the heavens and the stars to fall down their opponents.  Silence is dangerously deep in mudslinging...this is the lesson I\'ve learned in this May 10, 2004 local and national elections here in the Philippines.  But, these are not all.  Nevertheless, other cultures and traditions thrive in some other places where people\'s upbringing is more of \"pride and humility\" fighting each other to win the hearts and mind of the people.  I have learned that people really mean business when business is brought to their lot in either \"political ways\".  And there are many political ways available for the choosing.  We can classify them as autocratic, democratic, egalitarian, Maoist and Leninist, Communistic, Socialistic, and all those that contains the \"ian\", the \"ist\" and the \"tic\" words combined; and the grievous fault is that when we choose one, the consequence of having surely an opponent, in such a certain political perspective and dynamic culturation, is not farfetched.  Today, the aftermath of the last election is quite intriguing, why?  Because a lot of people who are now swarming the winning candidates and some if not all of the winning candidates are hiding because of so much expenditures they\'ve incurred.  Others are now even having fun with their immediate families accepting greetings and houlahoou from immediate friends and critiques everywhere.  Scenarios are drawn every second making choked-up stories to read or even heard over radio broadcasting systems all over the archipelago.  Well, there\'s the opposite of this hooked-up news and caravan cuddling among candidates who won.  They start to learn how to keep silent everytime.  Maybe they\'ve learned the past...too much talk, too much mistakes; less talk, less mistakes it\'s because they\'ve done nothing for the people. (Handsclapping.....Yeheyyyyyy...)  Hypocrisy really cripples the bone to the last bit of strength it has. May be this is one of the essential elements of \"silence\". (Watch out for the next elections....)  '),(18519,2863105,'12,May,2004','2004-05-12 00:00:00','   Ordinary people today in my place are thinking about the aftermath of the elections just undergone.    These ordinary people (including me, of course) seemed to ignore the early warnings of Parish Pastoral Council for Responsible Voting (PPCRV), a national crusade for an honest and orderly elections; as well as the National Movement for Free Elections (NAMFREL), a private conduit of the Commission Election to ensure the sacredness of the ballot, not to sell their sacred vote to candidates who are attempting to entice them to do so just to win a seat in the local or even national government elective positions.  \"Money talks everytime...\" said a teacher. \"People are really aware that corruption breeds another kind of corruption.  The Media has been vigilant enought to remind the voting populace the importance of casting their sacred ballots. They generally tried to condition the minds of the people about responsible voting and even portrayed certain situations about the effects of corrupt voting; but, they refuse to heed...\" she continued.  Since the first elections under the 1935 Constitution, breed of corruptions has already been in place.  The potential leaders are fighting themselves to gain the upperhand of holding the national government. Maneuvers were undertaken, manipulation of the national consensus about decisions and national policies were railroaded to fit their own whims and caprices in running the government.  One of the prominent statesmen, Manuel L. Quezon, President of the Commonwealth during the American occupation of the Philippines, elusively shouted infront of many spectators: \"It is better to see the Philippines run like hell by the Filipinos rather than like heaven by the Americans!!!\" (Well, I\'m not sure if these were really the words he uttered, I just remembered it now...never mind, just get on with it.  It\'s my history thoughts that is failing.)  Well said Mr. Quezon...it is really going bad now.  Your prophesy seems becoming true today. Even ordinary people is radically high enough to contest the government of such wrongdoings. Ha ha ha...wise enough to learn more about his miscontrued facts. No respect of whoever is holding the position of power.  They greatly believed about People Power. What the heck!  But examining their utterance, it seems though, money flows within their ranks coming from corrupt government officials or corrupt private moneyed-persons who paid their lot for 3-meals food for the day.  I have some information that rallying in the urban centers is now becoming a prosperous business among well-educated or ill-educated people in my country.  Money talks, they said.  And when money speaks, everybody listen, thence, act.  But, of course, there are others who played truthfully the dimensions of their struggle for the sake of the poor.  While this could be so, we could not truthfully assure the general public or the potential foreign investors, that the economy is running well. (Well, what\'s for the economy...my country is going dry...and there\'s is a great drought going on among truth-seekers, who else could be thinking that this country is getting the greatness that it lost sometime?)  As history will teach us the wisdom of learning more to gain more...the aftermath is worst enough to convince me that the decisions of the first statesmen of my country before it gained independence from the Americans WAS RIGHTFULLY WRONG!!!...but what they\'ve done was awfully RIGHT!!! It\'s a matter of choice between good and evil, and it seems that the choice was between the two warring sides.  It means that my country is living within the clutches of both good and evil. (Sigh!)  Anyway, going back to the price of selling my sacred vote....I should say that even 1/3 of my countrymen is doing this; another 1/3 is keeping this, and the last 1/3 is highly paying this...the summation would still reflect the rottenness of the hearts of those unscrupulous people who taught the Filipinos to see precious price of gold and silver; who taught the Filipinos to value more of their properties; who taught the Filipinos to study government from other countries and employ the same here in the Philippines; who taught the Filipinos that working abroad is much better than sacrificing a few peso to give it to the paupers; who taught the Filipinos to regain its greatness by empty promises so that others will follow suit; who taught the Filipinos to business as usual that everything they gained from their own practice of profession is worth than to stay stagnant among the poor; who taught the Filipinos that beauty is in the face of everyone that by grabbing the property of the other controls everyone who are living in it; who taught the Filipinos that we do not have our own identify that we belong to the culture of the world and every nations owns us!!!! (What the heck, no one might listen this poor shout...)  Okay...I thirst for the truth, really. But what can I do. I don\'t have to isolate myself from my countrymen; and I hope by doing so, I could learn a lot of wisdom by the way they treated their ownselves.  The price of selling my sacred vote?...it\'s the price of my whole country!  '),(18520,2863105,'11,May,2004','2004-05-11 00:00:00','   Today, the Philippines is given to deciding who will be their political leaders. People are rallying in the streets with a shout of either joy or madness.  They contest the result of the last May 10, 2004 elections, posting their own desire that FPJ, short for Fernando Poe Jr, won the election and that the present administration of Pres. Arroyo is amending the results of the people\'s voice.  The media is playing the lead role of informing the public.  But, you have to choose among them: which is telling the truth and which is hyping.  Such a malady; a gross supplantation of the truth. Everybody wills to win and only a few cast their lots.  The hiatus of Philippine politics is much to the grievance of all that even those from the farflung areas of the Philippines, while listening to the radio reports (because no TV is available yet), shouts their grievances to naught.  They\'ve pestered their own selves with much afflictions and hatred.  They blamed the government for its apparent apathy and ineptitude; they blamed the politicians who were buying the votes in order to win the seat of government; they blamed everyone except their ownselves.  Truly, the political bickering is most serious today that even the small unit of governance forget the truth of true public service.  \"I can see that my country is now on the verge of collapse...\" says a fisherman. \"...and we can only blame ourselves.\" He added.  \"But no!...it is not us who should be blamed...\" says the vendor, \"...it\'s a matter of education...a matter of governance...and many other factors that molded the Filipinos of long ago and now inherited by us. We need to look back the past to see the present...\'What our ancestors have done before?\' this is the great question that we need to answer and understand today.\"  \"Yes, it could be. But there is more than simple things done halfway.  We need to realize the things that we have done first; the decisions we\'ve made; the actions that we have undertaken, and to see the result. And look what happened...it\'s there...the seed of corruption still lingers to every bit of us. We could not deny it,\" said the old man who\'s been observing us chatting.  We can blame the Spaniards with their bigoted teachings of the Bible; we can blame the Japanese who taught us cruelty of human race by cutting off heads or stabbing babies while up in the air; we can blame the Americans who taught us education as a way of life; we can blame the first politicians who sold us to invaders; and we can blame everybody who hearkened to the enticing of much evil deeds!  This is it!...the seed of corruption is nowhere to be found now but everybody is afflicted with it; and it\'s no longer just a seed; it\'s become already the blood of corruption streaming to every single vein of a Filipino citizen. (I\'m sorry, but this is what I see.  It includes me, of course!)  Now, I could only realize that I am one of the products of this corruptive deeds because I did a single corruption of my own just as everybody did in the last May 10, 2004 election. What I did?...I have accepted the 2 kilos of rice; 2 canned goods; and 2 noodles for a vote to the candidate.  This is the first time that I did accept it. This is madness!!!! Anyway, the candidate that I voted won to the grief of others who lost because they could not do otherwise.  Thus, the hiatus of Philippine politics even to the last ditch of clear mind and spirit is marred with a number of kilos of rice, canned goods and noodles.  This is the mark of so much poverty...the result of corruption which comes from the poor man himself.  Who to blame???...nah...everybody does it...why can\'t I or anybody else? United we stand, united we fall... ha ha ha ha ha ha ha ha ha ha ha ha ha ha ha ....madness...purely madness...  '),(18521,2863105,'22,May,2004','2004-05-22 00:00:00','   One of the most terrible mistake that every single, ignorant and distasteful person do is to kill somebody because of somebodyâ€™s misdeeds...  The political euphoria that brought the whole country into a mudslinging arena is now practically over with the local winning candidates already waiting for June 30 for their oathtaking ceremony, while the losing candidates still counting, analyzing, and even trying to recall what is the big mistakes that his/her campaign buff out with no winning results!    They tried to call up their coordinators and political leaders in the barangays, but they found out later that they were all gone and went to anywhere else this world can bring them.  A few who are truthful with their commitments went to the house of their political bets and said sorrowful render their sorry words in tears.  They oftentimes advised that there will be another election, and maybe, just maybe, they could win next time!...to the comfort of the loser.  Funny thing,  is it? The aftermath of a national election usually or even frequently result to a remorseful supplication and remorseful vindication of the things they have done during the campaign and election period.  For forty five days, the whole bunch of candidates for the local and national posts are tagging themselves with distasteful smiles everytime they met each other; and now, you could see some of these candidates are in their secret hangouts having fun like drinking beers, talking, shagging,huggling, teasing and all that! Whew!!! Is this politics? No...not all that, of course!  They admitted with each other that they have played a good domino game where the ordinary people (with lesser education, as they said) do not even know how.  Another story went like this.  There were three (3) official candidates running for Congressman in a District in Camarines Sur province.  Iâ€™ve learned that one of them put up another candidate, a woman who is practically a relative of his opponent, so that the surname of his opponent will not be ultimately used during the ballot counting and it would certainly create confusion from among the voters whom to vote... Nice political strategy which is already aged-old cunny political fuse but still effective.  He spent something about fifty to one hundred thousand pesos for her to run the election, anyway, they thought that the qualification profile needed to being a candidate for such an elite political position matters not, even an actor without so much experience in the political systemic arena of the country is qualified by the COMELEC! Just imagine, becoming a President?!  Anyway, later I learned that the woman was apparently disqualified by the COMELEC, but the damage has already been done to the regret of both party: to the one who spent the money, and to the other who hired a lawyer just to disqualify the woman.  Both are spending a lot of money for their own motives...cunning motives. (Whoooosshh...)  Sometimes, people have thought this way:  How about when the candidates have delightful thoughts that instead of spending the money to pay off accounts for the political campaign materials to be given to the voters during the campaign period, be given to them in cash?!...The burgeoning debt had already received moratorium and thereâ€™s an ongoing negotiations to restructure it.  Maybe, just maybe, part of this debt were spent during the election. (Oh yeah?!) Well, Iâ€™m not in the treasury, but I believe this small manâ€™s thought. How about it?    No!...political luminaries eventually said at their secret meeting place.  If people would have all this money, who will work in the factory, in the groceries, in the manufacturing companies, and all other blue collar jobs?!  If all will be given the chance to be millionaires, everybody will enjoy a sumptuous meals and snacks from highly for-the- elite five star restaurants, and if at all their money were gone already from their own pockets, the same situation will happen...Still poor Philippines, as they said.  Itâ€™s really a bad idea to give out money to the poor people who are clamoring political changes and at eventide languished themselves to jail.  Last night, the same people that I see last campaign period were seating over a bottle of gin, a famous poor manâ€™s hard drink in the countryside.  Theyâ€™ve talked about the events of last election.  They counted their shares in terms of money taken and money given to the supporters of their political bets which were provided to them for distribution.  Some even enjoyed telling fancy lies about their own ventures.  Laughing and jamming themselves as if a big hay was laid on their highways of smokestacks and frenzies.  They were all buggling about something foolish that they did during the election, buying time and selling votes to the candidates who are willing to pay the price.  (Well, for all you know, they sold out their straight vote as much as five hundred pesos...yeah, five hundred pesos...and for those who could not give out straight vote for something reasonable, they are paid three hundred pesos.  Least that you can receive from small time candidates is twenty to fifty pesos, or even one hundred pesos if you have a good sense of humor and fancy stories to tell.  Hah!)  Now, the election is almost over.  There were rumors running around the corners of barbershops and beauty salons, and even at the cockfight centers, about a certain candidate who spread it that he actually gave out fifty to one hundred thousand pesos to all local political candidates of their party, to the angry and dismayed partymates.  While on the other hand, rumors were also spread about the long-time political stalwarts in a certain district of Camarines Sur has railroaded the COMELEC with money for him to be able to clinch the winning streak for the congressional seat which people said have been his direly means of getting back to THE HOUSE OF CONGRESS and be the HOUSE SPEAKER again...(Sigh...) I believe that there are certain truth behind the rumors, even a bit...just a bit. You know why?...Thereâ€™s that old man saying: \"When thereâ€™s smoke, thereâ€™s fire!\"...   Certainly the rats racing out their lounge and parties and are going to feast a certain question: â€œWho is the most deserving candidate who won the last election?â€  Certainly, they deserved to be recognized. However, who will try to do that? Me?....Oh no....not me, certainly not me...but other can. (Of course, donâ€™t say this to others:  I have a bunch of local small people investigators who will pin me the best news theyâ€™ve got around their poor corners. I know they will...he he he...)  Anyway, I am certain that the scenarios that I have been drawing speaks about our country\'s trend in this best place of worldâ€™s famous blogs which can draw crowds in the future.  And as long as my fingers wonâ€™t stop typing the facts and trends of our own political crags and breachy ways...the truth will always prevail through it!    Certainly, with this page, the political hyenaâ€™s turf will be revealed.  '),(18522,2863105,'24,May,2004','2004-05-24 00:00:00','   In a country where more people are ignorant enough, no I would say, less knowledgeable about the real score of high politics, while those who have the knowledge are keeping the engine on with cunningly crafty imaginations to put the boiling water into an outrageous debacle,  is not astonishing;  when one day, you could see a lot of people prefer to take the streets filled than the courts of law.  Most of the times, as many stories being told to so many magazines and books, itâ€™s the elite who controls the crowd.  With the money on their hand, they can make the silence into an uproar when all the fundamental foundations of governance is lying as their prey.  People, rich people, or average-rich people, with the power to undertake the latest adventure within the political spectrum, not because they loved the country, but they would love to see the whole country break down their own bones and their own people starting to shot a star to heaven knows!  The holy scripture evidently taught us:  Money is the root of all evil.    Yeahâ€¦right.  One of my friends, Perry as we call him, texted me last night when I was still on my deep contemplation.  It seems that his words what he was blaming the conservative Christianity that influenced the whole people and learned a lot of education from its own experience, the cause of division, ineptitude in government service, among the people.  Itâ€™s a matter of choice.  I remember world history.  American people who suffered the same unfaithfulness was wrought of Civil war and later realized that they themselves are the ones bringing down the hook to peddle their own mistakes.  Today, look at this noble countryâ€¦.Uncle Sam is still on the go.  Though sometimes it failed the expectations of the lot of their people, when they committed the heinous crimes against the Iraquis by abusing prisoners of war, which they themselves are one of the strong authors of  the Geneva Convention. They failed because they forget the whole mission.  Soldiers are forced to the limit of their own frailtiesâ€¦tempted to undertake the extraordinaries though they know the consequence. (Sighhhâ€¦)  Today, the Bicamiral Congress of my country is on the verge of disruptions from the multisectoral groups who spark the troubles and drawn the scenarios that they themselves are taking part.  The errors of intelligence analysis is great on the part of both the AFP and PNP can be fatal when not immediately evaluated.  I believe, even, within the auspices of the law enforcers, people with their limitless minds, can draw the crowd into a raging storm.  Everybody is on their own scenarios, except a few who stand as martyrs for peace and tranquility who stood up their ground to highlight their own motive.  Some people do not know really what their motives are.  But, I could see between the lines, they are buffers.  Just like Eddie Gil, whom the Malacanan palace said: â€œâ€¦provided a comic reliefâ€¦â€ (Hahâ€¦hahâ€¦hahâ€¦)  Really cool!  FPJ one of the closes rival of the Pres. Arroyo do not have the intent of conceding.  I donâ€™t know of some people who stand just for their own pride, look at people running the streets, shouting their own names and urging the whole country to run with them, to highlight their protests!â€¦Maybe, you could only see it hereâ€¦.except for one, Joey Marquez, an actor-turned politician who conceded for his defeat in the congressioal seat at his district in Paranaque.  I donâ€™t know of others who did the same.  Itâ€™s not the call for the day!  In the US, when the people have already spoken, the opposing candidate, usually concede and express congratulations to the winning candidate, thence, pledge support to unite the whole country.  Here?â€¦I think FPJ will not yet into that humility ground.  He doesnâ€™t know that yet.  But I believe he will.  You knowâ€¦I just canâ€™t help not to write these things.  I was dismayed over reports that certain peopleâ€¦politicians and moneyed people are turning their hues to bring this country into its greatest turmoilâ€¦no, not turmoilâ€¦.a pitâ€¦.itâ€™s deepest pit--- civil war.  Anyway, if things will turn out right, these people have seen the message when they observed that a lot of people, poor, average and rich people alike, are really heavy-burdened with the kind of government we have today.  Politicians have their own agenda, with a lot of reserved political wordsâ€¦sweet wordsâ€¦to keep the whole country on the ground.  But, as I heard them speakâ€¦their hearts are away with the real motives of why they are there.  I could smell bad with the present scenario everyday.  Thereâ€™s something fishy about many things. Maybeâ€¦just maybe, the treasury is clear open and many pockets need to take the share. (Sssshhhâ€¦.silenceâ€¦somebodyâ€™s talkinâ€™â€¦) Look!â€¦These people are pumping again the whole public with sweety wordsâ€¦.oh noâ€¦sweaty wordsâ€¦itâ€™s hot hereâ€¦ summer is yet to end.  Where are the patriots go?â€¦This is one question that my friend wants to ask me.  No to election fraud! Yeahâ€¦No. No. Noâ€¦this is what I texted back to him. But what can we do? He suggested to meâ€¦weâ€™d better get some sleep, and I did.  I slept the whole night with deepest thought what will happen tomorrow.  Will the government yield to the demand of the people to change?â€¦ or should the people demanding change have to change themselves first?â€¦I know, you are making my whole write up a laughingstackâ€¦.because, my country as you can see in the news broadcast internationally, in on the loose.  But my dreamâ€¦my only dreamâ€¦.and hope, my only hopeâ€¦let heavens come to town to punish the wicked who are turning this country into a desolate waste; and my wish, my only wishâ€¦LET ME KICK THEIR BUTTS!!!  '),(18523,2863105,'01,June,2004','2004-06-01 00:00:00','   There had been a lot of madly/mudly issues shaking the foundation of our political system and as I have observed it, it is always the President and her Cabinet who suffers most of the consequences.   The opposition on the other hand, find time of slugging every bit of wrongdoings or rather corruptions that allegedly rocked the treasury or unbelievable transactions entered into by either member of the Arroyo Administration. All they see is corruptions and all manners of transactions!!!! Such a negative attitude demeans the whole political system (or should we say the Democratic System?).  Every wisdom is available to those who see them fit at a right time, right place and in right forum...but this right forum, usually coupled with mudslinging attitude to those who tries to campaign to deminish credibility. People are now confused to whom to believe. Everybody speaks of madness...ineptitude with one another...WE EVERYBODY, but we forget even to evaluate our own SELVES!!! We see others doing the bad things, but we forget the rightful way to prevent such things to happen, because, WE ARE MADLY EAGER TO WIN THE ATTENTION OF EVERYBODY...no matter what is the consequence.  Anyway, no one is to be blamed with this kind of attitude... Everybody fails and everybody would want to win even a bit...just a bit...  First Family Corruption?...What is it?...  What represents the First Family?...You might have forgotten the general rule that whoever chooses the leader is the leaders themselves.  Although Pres. Arroyo was elected as the Vice President of the Republic but since she is the rightful person who will replace the busted ESTRADA, still she represents every Filipino voters even those who did not vote for her during the election. MAJORITY RULES, ISN\'T IT?  Many people thought that since they did not vote the person who is now holding the powers of governance, they are not liable to abide by such decision of the majority.  That\'s why, not even those who were truly defeated during the elections accept the fact that they\'ve lost the election. Pity...pity to these selfish people. They do not understand the wisdom of elections.  They have been graduates of bigtime universities, but they failed the passing grades of the simple people who composed the majority because of the pride of their own selfishness.  Since from the very first time that I have casted my vote to every elections, I always respect the result...the clean result of elections.  But, since then, I don\'t know when was the last time that clean and honest elections took place.  First Family Corruption?...everybody is corrupted with the pride and selfish ambitions of those people who do not love their own country. Even then, if they loved the country, why is it that it remains peacefully malnourished? First Family Corruption?...What is it?...Is it not the reflection of our own folly? Well, I don\'t know much about it.  But I believe somehow I was even part of it as I casted my vote to the winner or the loser of any elections before!...and my simple mind only remind me of such folly committed by everyone, including me.  I\'m sorry...I hope everybody do the same.  But it is even more folly when no one seats the leadership of the nation. It is better to have one corrupt official who could easily be checked thru the legal system than no one at all qualified to hold the power of governance. I\'m sorry for those \"perfectionists\" or critics of the government, you too is the very fiber of such corruption.  '),(18524,2863105,'16,June,2004','2004-06-16 00:00:00','   In military tactical offensive, there is a common jargon which says: â€œThe more that you delay the offensive stance of the enemy, the more that you can maneuver the striking force.â€   This is what I thought the KNP is doing while their other cohorts are trying to paint and repaint a scenario in both the national and provincial scenes that FPJ actually won the election.  Besides the KNPâ€™s valiant senators and congressmen are maneuvering the â€œNO PROCLAMATIONâ€ scenario in Congress, all other petty officers and gentlemen is trying to enflame the mass of people to go out and cry their sentiments in favor of FPJ.  Wow!  This is like a movie in the making!  Well, in my own simple mind, I could say that these things happened because a lot of the people would really like it.  These scenarios, painted or unpainted, draw a lot of eyes on to the canvassing of the Congress as a National Canvassing Board, including even the whole community of nations.  The Philippines, I believe is now in the spotlight again!  After the 1986 People Power Revolution which dragged the whole nation into a squandary, and even to a near civil war scenario, and then later, the unifying charisma of Cory Aquino has brought a newborn nation!  It was as if a painful birth has gone awry.  After Cory Aquino went through and survived many attempts of coup dâ€™etat, of course, with the help of her valiant Secretary of Defense, Fidel V. Ramos, a new nation was brought to a new beginning, a new hope for the best!   I thought, it was a new beginning for us in a new Philippines!... But, regret it to say, it was not!  FVR was made an awesome president with merely 23% win of the total voting citizens. A minority President as he was oftenly referred to.  During his tenure of office, a lot of changes was made in the bureaucracy.  Old crocodiles were shook off from their posted haven, and eventually were driven out of office.  The accomplishments of the Ramos administration of cleansing the core of government bureaucracy was never a publicized accomplishment but a silent bean of hope being planted to make the government run like the best of the other nationâ€™s been doing.  I have this faith in FVR who once again magnetized the whole quarreling nation into one unified people with one spirit and one soul.  ERAP ESTRADA was the peopleâ€™s choice.  The masses were convinced that he is for the poor masses.  Everybody was made to believe that he really is... even today, you could feel that stigmatic influence among the poor people as indicated by the mass of support on FPJ, a close friend of ERAP.  Only few, knew the dark clouds of the past about these two man of the people.   Hey, I have this buzz...Who knew about the BIG 4 syndicate?  Well, I donâ€™t have to elaborate more about these, because when I do, I have to scour the whole pile of newspapers in the national library to find one which contains the crimes theyâ€™ve done against society; I remember my old friend who died many years ago who told me about these stories. He used to used to refer ERAP and FPJ as the alleged masters of this syndicate.  I donâ€™t really know.  Anyway, when ERAP held the reigns of government, some of my friends who have friends who were former PSGs said that Pres. Estrada is somewhat an arrogant but delightful leader who usually escaped from the eyes of his security escorts just to visit certain secret friends.  He went out off MalacaÃ±ang with no escorts. While all of the PSGs were turning the whole city upside down in looking for him, some of his â€œfriendsâ€ who served as his advisers were enjoying the â€œreignâ€ with so much privilege and favor plate at their breasts. My friend even said: \"Itâ€™d cost only a call to find one stupid guy who believed that he is still a young philandering man...\" my friend usually chuckled when we had a short talk.  \"But he is good\", he added.  With ERAPâ€™s who was very busy visiting some friends in the casino or secret hangouts, some of his generals were taking the hue!  They observed that the President is not capable of running the government as it used to be with FVR.  Well, they know that FVR was an excellent strategist.  He made up things before it really happens.  â€œWhat EDSA DOS?  Really?...â€, these were the words I uttered when one of my neighbors told me the events happening in Manila.  The â€œlast envelopeâ€ was not allowed by the Senate to be opened; the people, in an uproar, to hear the truth went to the streets shouting their sentiments against the government of Estrada. They want him booted out of MalacaÃ±ang!..which they did! A soft surrender before bloodshed was opted by Estrada. Off he went out of the backdoor of MalacaÃ±ang on to a navy ship to a place nobody knew but his escorts...GLORIA was proclaimed the new President by the Peopleâ€™s Power authority!  2004 Election was reported to be full of election fraud.  The opposition kept on insisting to the general public that there was fraud!...fraud!....fraud!... But when asked to show evidence, they were not able to show one.    Today, the GLORIA administration is being rocked with so much criticisms roundabout. Even the \"little\" people would ask: â€œWhat kind of government is this?â€  Well, they donâ€™t know the workings within the spectrum of government bureaucracy, but they know the basic principle that when leaders are chosen by the majority of people, they would likely respect such a noble decision.  But, the KNP would not heed.  They have their own interest.  They have their own agenda.    In some quarters of the elite haven, they would not want FPJ as the President; but, in many farflung quarters of the common people, they dreamed of FPJ as president.  What a tremendous contrast, it would be.  While these poor people are dreaming of FPJ to be president of the nation, even if they heard a lot of bad things about him, they are firm in deciding that FPJ deserved to be president.  Why?...I asked them and they vehemently answered: â€œ...because when wiser people are reigning, a lot of poor and very poor people are suffering; and we believe that whoever is holding the reign of government, they would certainly be corrupt still.  Actually, not one of those running this government is qualified to meet our meager needs. So that, when FPJ fails to attain our dream of a good nation with him as the president, so be it! We don\'t care, we all suffer, including the rich people...\"   Hey, is he telling the truth?! I was amazed about the thought of this simple people.  Thatâ€™s what they thought it would be? Everybody suffers the consequences of a bad decision; and it is going on now.  We don\'t know when it will eventually end.  As I can observe, the GLORIA government today is very lenient as to the madic attacks of the opposition; I know in my level of understanding, that GLORIA would not want to go down to the level of angered enticings of the opposition.  Well, thatâ€™s right.  A good leader with a good heart for the whole nation should not and must not succumb to the demands of her unworthy enemies. If she would do it, the whole mad mad world will turn down to ashes and the fear of civil war scenario is not remote to happen in a few days.  Well, am I dreaming?...No, I hope not. I wish it not...Oh no, I\'m now confused. Sigh... the Filipino people is charting again a new ship... a navigation to nowhere... but with an eloquent mother like President Gloria Macapagal-Arroyo, the children will play on in the playground and will enjoy the future as others wanted it not to be.  Sigh...after this bad bad dream, I know, a new birth of a new-old nation will come to rise again for a new sunshine of hope.  '),(18525,2863105,'21,June,2004','2004-06-21 00:00:00','   It has been a gruelling event for everyone in Congress, in COMELEC, the Supreme Court that the recently concluded national canvassing for Presidential and Vice Presidential Race.  It has brought the whole nation into somewhat quadratical equation at which the people begin to submerge into an impeccable uproar or a split-second revealing tendencies to go either against or in favor.  The characters of the whole event profusely charted every move either in upper or downtrend hopes and all.  As I could described it to be a dramatic trauma for everyone when the teeth of the law has been moved to act in itself and plot the certainty of its result and the aftermath could have been a big turmoil.  Election canvassing was tauted to be the slowest when compared to other nations election.  So far, I believe that the Philippines have charted again another â€œfirstâ€ in the history of all nations.  During the canvassing, you could see the workloads and how these workloads have traumatically effected within the realm of its conceptual profanity or heaven-like divinity.  Everybody played certain roles within their own realm where events have plotted up to the very end of the last certificate of canvass.  Now, I have understood very well that the chartered course of exercising the very event as was anticipated during the flexing of the election laws within the aupices of congress before it has been approved; I have seen the wisdom of all the people who get involved in one way or another, how  this law has made its own markings.  I could even say, that the events that transpired during the canvassing of certificate of canvass, was really working well, except for some lopeholes which gave each party a gruelling or even a delicate move that would have caused chaos or even civil war instead.  As everybody sighed in joy for the last COC, it was surprising that no other than the pinch-heater Sen. Aquilino Pimentel who moved to the Chairman of National Canvassing Board to canvass the COC from Cagayan de Oro City with an earlier negotiation that no more objections that would eventually delay the canvassing.  I was surprised by such move of Sen. Pimentel, while everybody in the gallery stood up and clapping for such a surprising move.  The TV announcers was even astonished to the action taken by Sen. Pimentel; they believed that it was a historic move for the whole watching nation that the prime mover of the opposition wisely yield to the exigency of the canvass bringing the whole into a new perspective.   It was indeed a wise yielding that the opposition gained the approving smile of the whole nation.  It was so simple to heal the already acute atmosphere of dissension and chaotic maneuver from some quarters; and the whole nation eventually see the yielding gesture why thereâ€™s a need of opposition.    I have gained a lot of wisdom in the whole event that happened in the Philippinesâ€™ political exercise.  The law was placed as a machinetic grawl at which the oil and lubricants on how it will run when put into execution as the people carpentered it were all aware of things that actually will take place.  The law in action is a law for good indeed when the whole people playing the main characters and villains have done their parts assidiously as expected from them.  I could feel the tension that threatened every quarter of the whole nation.  The opposition eventually yielded the sustained and critical fight of their own lives; the administration have done the best of their wisdom to maintain sobriety and calmness as they dealt with every accusations and mudslinged rusts coming from every side of the playground.  Wow!...it was a wonderful play-act.  The canvassing was timed to be finished during the nationwide celebration of the Fathersâ€™ Day.  It was critical and it has put the whole nation into a new plot of chartered course in the navigational realm of the whole wide world indeed.  But still, questions may arise:  What will happen next?    As I can see it, the election canvassing can be a chartered course of practicing the workings of the law of land.  Just as long as everybody did their part with earnestness, the success of one is the eventful success of the whole.  '),(18526,2863105,'02,July,2004','2004-07-02 00:00:00','   As I have noticed these past few days, before the oathtaking ceremony of the 14th President of the Republic of the Philippines, President Gloria Macapagal-Arroyo, the supporters of FPJ have shown their wares in terms of civil disobedience.    They showed no respect to the duly constituted laws of the land, because, if they really are respecting it, they wouldnâ€™t have gone to the streets and aired their grievances.  I felt a bitter feeling about these my fellowmen.  They have gone awry and disrespect was the order of their day.  What if...they have won?  What will be the inspirit they have in running the government? Will there be hatred on the other side, also thinking and re-thinking the pain of losing the elections?  Would there be any chances of reconciliation between parties who have drifted beyond the unexpected result of blunders and pride?  Will there be peace?  There are a lot of questions that have come through my mind and nevertheless the answeres remained in oblivion.  As I can see the action taken by the protesters, brewing with a lot pain and grief in their hearts, they become rascals in portraying their own pitfalls.  The leadership of their group is becoming weak and incontrollable as the people are becoming unruly and hear no words from their opposite side.  As I can see it, hatred has become their bread and their pain has become their water supplier in an outrage of getting the reign of governance that they themselves have gone out but displaying their own selfish ends.  The news revealed a lot of hidden agenda that sometimes the broadcaster-reporter himself observed in the way of covering the whole news report.  Based on their own experience of covering these unruly events of people who aims is to collect P100.00 bills from their unscupulous leaders who financed such mass actions against the duly constituted government included therein also was reportedly kilos of rice for their own family good for the day.  Where is the real grievance that theyâ€™ve been trying to air?  Is there truth in their campaign? Well, I donâ€™t see any reason at all for that fact.  Because I truly believed that people who are fighting for certain ideological beliefs or political goals have to sacrifice something of their own for the benefit of the many.  Here, in the Philippines, I can see that the old paths of glory is gone awry.  The truest motive of heart from the people who conduct mass actions is not really felt within the hearts of those who are governing.    Why?   Because the words that they have spoken remains to be within the realm of their expected selfish ends.  The thunder of words that you may have heard from their ranks were words reflective of their own pitfalls, hence, their wrong perceptions has been an advanced dungeon of their own views of the realities that they see on government.  Even then, they knew that they will fail, but incessant fire is fanned by those people whose plan is for the continued suffering of the people of their own feathers.  '),(18527,2863105,'22,July,2004','2004-07-22 00:00:00','   There has been so much hype on the case of Angelo dela Cruz, a cargo truck driver kidnapped by Iraqi militants and was about to loss his head if the Philippine government have had refused to withdraw its miniscule humanitarian troops deployed in war-torn Iraq.&nbsp;&nbsp; Were it not for Angelo dela Cruz, our troops would have staying in Iraq up to August 2004.  &nbsp;  In anyway, the implication of such withdrawal of troops signifies many negative impression on the part of the international community whoâ€™d been observing the actions of the Philippines towards resolving this issue.&nbsp; Unlike Bulgaria and Korea whoâ€™d been supportive to US action against the regime of Saddam Hussein, now ousted President of this middleast country, it nearly buckledown the already spent effort of the US in terms of stabilizing the situation in Iraq.&nbsp; The withdrawal of the staunch supporter of the US, the Philippines, brought a wasted risk of lives among those who were willing to gamble their own resources just to let go and free the chained people of Iraq.  &nbsp;  The action of the Philippine government have caused collapsed international diplomatic relation that would, in the long run, affect the countryâ€™s relation with other countries.&nbsp; Dealing with terrorists, indeed, is not an easy task; and the Philippines closely know why the US refuse to negotiate with terrorists.&nbsp; They have their own experiences and wisdom why they wouldnâ€™t tolerate such demands and yield to the negotiations.  &nbsp;  I can see the wisdom also on the part of my government.&nbsp; But in any angle of perspective and dynamic realization, such negotiation with the terrorist is a cuddling action that we, as Filipinos, have that intimate views about one person in relation to the bigger proportions of the population whoâ€™d been in the quagmire of poverty as caused by decisions that have no strategic value at all in terms of the global situation.  &nbsp;  For last two months at most, the country has been within the spotlight of the world: observing what actions and alternative would the Philippine government do to resolve this impasse.  &nbsp;  Now that the crisis is over; still the issue did not die down that slow.&nbsp; It even increased into a tremendous magnitude as if the whole of the Philippines is on for a festivity; and the media is feasting for Angelo dela Cruz and other issues of higher priority is now laid down to give way of welcoming the man whoâ€™ve awakened the eyes of the world and see the â€œcowardiceâ€ decision of my own government.&nbsp; I donâ€™t know how will I call it.&nbsp; But it hurts and caused so much pain for the whole wide world that my country have this â€œentangled tendencyâ€ to yield to the enemies of the world --- the terrorists.  &nbsp;  I could not blame my government of such a hard decision.&nbsp; But still, I just hope that this would not cause so much trouble when other times come.&nbsp; History repeats itself.  '),(18528,3590069,'26,June,2004','2004-06-26 00:00:00','    So I finally went out to have some fun on Thursday to a place called Cafe Iguana...It has both a regular club music dance area and a salsa room.  I had a great time except for the expected nasty men who tried groping and sticking their hands up my skirt...I met one pretty nice guy who I might go out with that apparently used to play for the Atlanta Falcons...riiiight...we\'ll see if he\'s telling the truth.  He seems okay but a little desperate.  He comes on a little too strong sometimes which gives me the impression that he\'s desperate.  Maybe I\'m too skeptical...maybe he\'s actually nice and he just says what he thinks and stuff.  He asked me to come down and see him tonight...pretty sure I\'m not going to some guy\'s house who I don\'t even know when we haven\'t even been out on a date yet.  He said if I come down I won\'t want to leave...want to leave or won\'t be able to leave?  Misery anyone?  I\'ll keep you all updated on news there as it unfolds...  Not much else new...I\'m hoping to go to the keys for the 4th but I\'m not sure I can afford it just yet...  H\'ok...off to the office to finish my paperwork for the night...   '),(18529,3590069,'22,June,2004','2004-06-22 00:00:00','    Ever wonder what people say about you behind your back?  Ever wonder if people are really honest with you?  Not that I\'m perfect by any means but I try not to say anything behind anyone\'s back that I wouldn\'t say to his or her face...I don\'t think most people share my philosophy.  Maybe I come off like a bitch being honest but I think it\'s better to be honest with people than to lie to their face, pretending to think some way when you don\'t, then turn around and talk shit about them because you weren\'t honest enough to tell them what you really thought in the first place.  What is funny (well, not really funny actually but you know what I mean), is when people don\'t know you at all and blow shit our their asses, talking about things they know nothing about...Guess I\'m a little angry huh?  I guess I\'m a little more tolerant of people than they are of me...sad...  Don\'t get me wrong, it can be really good to find out how other people see you and hear what they say behind your back since we don\'t usually see ourselves as having faults but, trust me, sometimes it turns out to be less educational and more hurtful than you\'d expect.  Maybe I give off a different vibe than I think, or maybe some people just don\'t know who the fuck I am, nor have they taken the time to get to know me in the first place before judging me...It\'s very interesting to hear what they have to say or read it about it anyway...Funny how often we demand people be tolerant and not pass judgment on us when we are very judgmental and intolerant of them...   '),(18530,3590069,'19,June,2004','2004-06-19 00:00:00','    I\'m so happy it\'s Saturday!  I am soooo tired and my feet are so sore...Six days a week is so rough...Going to see Dodgeball tonight...that was the reward for making our office goal this week.  Should be funny...  I\'m hoping to go to the beach tomorrow for the day but there are no definite plans.  Sleep is first on the agenda if all the church goers can be quite in the morning.  I guess I could get up and go since they always wake me up anyway but I feel it\'s important to stay in your pjs until 10 at least once a week.    Today was a pretty good day...I had four sales, the second time this week and only the second time I\'ve gotten that many in a day.  I didn\'t, however, meet my personal goal for the week so I don\'t get my $20 bonus...that sucks but I\'ll definitely meet it next week \'cause I already have 7 going out next week...I serviced 6 this week, 15 last, and should service at least 15 next week...I set my goal at about 10 to 12 a week to be serviced \'cause I\'m not consistent yet...At least I made $400 today...or I will have in the end...the job sucks but you can\'t complain about the money...  H\'ok...off to the movies...   '),(18531,3590069,'16,June,2004','2004-06-16 00:00:00','    H\'ok so I\'m really messing things up here...I couldn\'t manage to publish yesterday and then when I made a comment to CJ\'s blog, I accidently erased it somehow after I published it and couldn\'t make it post again...I\'m apparently an idiot...Anyway, I didn\'t mean to erase it...this is what I had to contribute...  I CANNOT believe KFC...How white trash is that? Speaking of Nascar and white trash...I forgot to tell you all about this...We have certain incentives for selling some days and they are so funny sometimes. We have this thing called ecash where you can earn pretend money that you get to use to buy really cool merchandise like digital cameras and stuff...Anyway, one incentive to earn ecash one day was to sell to someone with a mullet...if you made the sale you somehow had to convey to the office, over the phone, that you were selling to a person with a mullet. If they figured out you were selling to a mullet, you got ecash. The trick was they had to know you were selling to a mullet and they had to confirm that over the phone with you. One way to let them know might be to say, \"yeah, Nate, we have some business in the front of the house and a party in the back\" maybe covering it up with ant problems or something...a party of ants or some serious business in the front. I so hoped I\'d get that one...I almost sold a mullet once but it wasn\'t on that day. Another incentive we\'re supposed to have in the future is to make a sale using an accent through the whole thing, knock to finish, like pretending you\'re from Australia or something...I can\'t wait to try that. You gotta do something to make this job fun...Anyway, just thought I\'d share that since you brought up Nascar and it reminded me of the mullet people...hahahahahaha...To use a familar phrase, I don\'t care who you are, that\'s funny...   '),(18532,3590069,'10,June,2004','2004-06-10 00:00:00','    So I spent half an hour updating my blog for you all yesterday...too bad it didn\'t post...I was soooooo pissed!  All that great work for nothing and I\'m not sure I can remember everything I said.  I don\'t know what happened when I went to publish it but it didn\'t show up afterward...lost in cyberspace I presume.   Anyway, we\'ll try this again...  Everyone has been asking me how Florida is, well, everyone except Miss Vaughn since she doesn\'t call, answer her phone or return calls...anyone know if she fell off a cliff or moved to Tahiti?  I haven\'t talked to her in over three weeks...Anyway, the money has gotten better \'cause I\'m actually selling well now and it will start coming in slowly but the job still sucks!  I have sooo many f\'ing bug bites it\'s not even funny!  I look like a leper!  My legs itch so bad and I can\'t stop scratching...I know, I know, if I don\'t scratch it won\'t itch but that\'s easier said than done (I should know since I have at least 25 bites on my legs alone)!  You\'d think I\'d be settled into the job now and content with selling pest control since that\'s my life now but I just can\'t quite get there...you\'d think I\'d be pumped when I sell well (today was my best selling day of the summer so far BTW) but really, I still find myself saying, great, I GET to do this all over again tomorrow...so fun!  I guess the experience is supposed to pretty much suck but the fact that I\'m making great money should make it okay...I\'ll try not to complain too much but I\'m sure you all can fell my pain just a little...I almost don\'t want to write in my blog \'cause I\'m afraid everything I have to say is negative...There are good things about being down here...I just have to keep them in mind.  It is BEAUTIFUL for one, and it\'s nice to have sunshine all the time...     Enough of that, back to some bitching...I just started selling in Miami yesterday which is great \'cause I have new area to sell in but I\'m pretty sure my Spanish sucks!  You\'d never know I took a couple years in high school or a year in college.  I\'m trying to learn some helpful phrases but all I really know that can help me for work is, \"Are you the homeowner?\" how to tell them I\'m treating for bugs, how to say ants and spiders and how to tell them the price.  That\'s really only a couple words and then they talk and I\'m lost.  I have no idea how to tell them they\'re getting a discount...I\'m at a loss there and I can\'t carry on a conversation to sell them to save my life...Hopefully it will get better...I\'ll get by in the meantime and luckily I get several people who speak enough English to communicate.  It\'s funny how many people tell you they don\'t speak English, probably to get you off the door, and then start speaking English almost perfectly when you tell them what you\'re doing and that you\'re there because of the bugs...Bastard liars!  I guess that\'s one way to get rid of me...pretend to not speak English...ruuuude!!!  The people I work with, as I think I\'ve told you all, are okay but I really need to meet people outside of work to hang out with...There is WAY too much drama!  There are a couple of really immature people that I can\'t handle but I get by for the most part.  This one girl I had to work with for the past week is so freaking crazy I can\'t handle it.  We all hate living with her and no one besides me has been able to work with her for more than 3 days.  I made it 7 or 8 before I told my boss he had to switch me or I\'d kill her.  I seriously think she has schizophrenia or MPD...she goes from normal to psycho in 2.9 seconds sometimes.  One of the girls in the apartment finally let her have it this morning and yelled at her because she was so sick of her crap!  I love living alone at home.    I did meet a girl on the doors who is around my age last Saturday.  She\'s also here for the summer...her name is Lauren and she\'s really great...we went out last Saturday night....yeeeeaaaaahhhh...but I was so tired I could hardly enjoy myself.  I think we\'re going to go salsa dancing tomorrow if I can stay awake...these long days sure get to you sometimes.  I finally saw an attractive guy here...too bad he\'s married.  He was hot though...Hopefully that\'s only a preview of what Miami has to offer.  I gotta be able to have some fun while I\'m down here, right?  Lauren knows a couple of guys and they invited us to a BBQ on an island in Miami last Saturday but I couldn\'t get off work early enough...oh well...maybe another time...maybe some of the guys will be good looking or at least have good looking friends.  Gotta have something to look at since the guys I work with are young and LDS...not a chance of any action there...    What else?  I don\'t have any fun plans for this weekend (well the one day off I get anyway).  Lauren is gonna be in NY visiting family but hopefully I\'ll get to spend some time at the beach...I really think we should get Mondays off also, at least once a month or every other week...it would be so nice to have a day for fun and a full day for rest...Half the time on Sunday I\'m too tired to do anything fun.  I guess I\'m here to work my butt off and make some money but I\'d really like to take advantage of being in Florida.    Okay, I think that\'s enough rambling for tonight...I\'ll write more tomorrow if this damn thing posts this time...Hope you all are doing well...I miss you all a lot (a great deal, Emily, not a piece of land) and can\'t wait to see you again.   '),(18533,3590069,'10,June,2004','2004-06-10 00:00:00','    H\'ok so it\'s pouring down rain...not so fun when you\'re supposed to be knocking doors!  No big deal, it gives me time to check my email at the library and update the blog.  Thanks for commenting so soon; I appreciate the interest and input.  How many fucking uuuuus CJ?  Five...There\'s this new invention called a bookmark, however, if that\'s too many for you to type in the future...just kidding...  So about life here: I mentioned the mos...did I mention I haven\'t had a drink since I left?  In fact I haven\'t had a drink since I think some of us were at the Garden that last day of finals...wasn\'t that the last day?  It\'s all a blur now...All I know is there have been no drinks for me in over a month...that\'s not gonna work...I\'m gonna have to make some normal friends soon...I have been offered a couple beers on the doors but I\'ve declined for some reason...WHAT?  Turn down beer?  Yeah, I did.  Bad Tessa.    And on another note...what have I done since I\'ve been here?  Not much but work six days a week, 10 hours a day but I did get to go to South Beach one Sunday...soooooo gorgeous!  I got to see Versace\'s home where he was killed on the front steps.  I didn\'t realize the house was literally right in town on the beach - interesting to see it in person after all the media hype.  South Beach is a topless beach (for a few anyway)...strange...I sat up to adjust my top for a minute and some guy walking by said, \"You need to take that thing off girl, for real...\"  Riiiiight!  For Real!   Oh yeah, then when we were driving back from Miami we saw two yellow ferraris that had wrecked on the opposite side of the bridge we were on...it totally looked like a Jay-Z video.  There was a Bentley stopped in front of them, a bunch of cops, the two smashed Ferraris and about 6 or 8 bottles of alcohol on top of one of the Ferraris...Crystal and Moet I\'m sure...hahahaha...Not a fun vacation for those idiots.  I bet they rented those cars, got drunk and were racing when they smashed into the guard rail and each other...lucky they didn\'t drive off...  What else?  I got to go on an airboat tour of the Everglades with some white trash mullet men...Not many gators...the trailer parkers were pretty much the entertainment for the day!  That was good times...  H\'ok so...I better get back to work...Can get rich hanging out in the library...   '),(18534,3590069,'09,June,2004','2004-06-09 00:00:00','    Most of you who will read this know I got off to a rough start in Florida...car wreck in the first three days...strange Mos....lack of money in bank...no money being made nor prospect of being made in the future...Alas, the light has dawned and I\'m finally pulling in some dough and things are turning around!  I was mooching off someone\'s wireless in my apartment building but that was nipped in the bud...RUUUUUUDE!!!  I\'m at the library now taking a lunch break and starting a blog...this could be fun...   '),(18535,3590069,'31,July,2004','2004-07-31 00:00:00','    Yeah, it\'s the end of the week!  I am so excited!  15 more days of work and then I\'m home free!!!!!   '),(18536,3590069,'28,July,2004','2004-07-28 00:00:00','    It\'s been raining here nonstop for a few days...not so fun when you&nbsp;knock doors for a living.&nbsp; Sales suck right now and I\'m sure part of that is due to the rain eating up many of the hours I could sell during but it\'s all coming&nbsp;to an end...18 work days left I believe...yeeeaaahhhh!!!!&nbsp; I can\'t wait to come home!&nbsp; I am so excited to get out of here....I\'m actually just about to call to check on changing my flight.&nbsp; I get to leave a day early so I\'m actually gonna change my flight to two days earlier and leave my last day of work if I can.&nbsp; It means I\'ll get in pretty late at home but I don\'t care as long as I\'m home.&nbsp;   What else is new?&nbsp; I got my nails done today.&nbsp; I am probably more particular about them than anything else (possibly&nbsp;more than I am about my hair for those of you who know about that).&nbsp;&nbsp;As if the rain wasn\'t enough to kill my day, I have&nbsp;the&nbsp;WORST set of nails I\'ve gotten in 8 years!&nbsp; Stupid bitch made me bleed several times too.&nbsp; They look soooooo bad but I ended up not having to pay, thank the Lord, since I have to practically redo them tonight.&nbsp; It looks like&nbsp;a 10 year old did them...I\'m embarassed to look at my own hands....I don\'t think it would have been so bad if I could have communicated with the chick...again, don\'t I live in the United States?&nbsp; Apparently not....  I went to South Beach on Saturday night...I got two HUGE blisters on my pinky toes filled with blood that ruined any chance of wearing any shoes besides flip flops for three&nbsp;days but the night was not a loss because I did meet well, see/say hi to,&nbsp;Usher.&nbsp; It was crazy!&nbsp; I was with some friends outside a club and we were deciding if we wanted to go there or somewhere else.&nbsp; I had just been talking shit about Usher a few minutes prior, believe it or not, saying how I had a confession that I\'d like to kill him for writing that Confessions song that is over-played 50 times a day.&nbsp; Out of nowhere I saw some flashing cameras and turned around to see Usher walking out of the club.&nbsp; Random!&nbsp; A chick I was with wanted to get a picture of him so we went to the car he was getting into...one of those new Chrysler\'s that look sorta like Bentley\'s....It was nice enough....He just kinda stood there I think \'cause he wasn\'t getting that much attention....I was surprised....then he just got in the driver\'s seat and&nbsp;sat there.&nbsp; We were only 5 feet away so I decided I\'d just say hi, why not?&nbsp; We stepped up to the car and I touched it but, of course, out of nowhere some security dude...a midget security dude rather, popped up and said \"you can\'t do that!\"&nbsp; I was like, \"Uh, okay....we were just saying hi, not like we were gonna maul the guy.&nbsp; He\'s just sitting there anyway.\"&nbsp; He was like, \"yeah, but he\'s already in the car and we\'re about to leave.\"&nbsp; Please...he was just sitting there for like 10 minutes...with the window down...I\'m sure just because he wasn\'t getting enough attention.&nbsp; Whatever...and there was some random ho in the front seat with him.&nbsp; Sure he just picked her up in the club and took her around the corner.&nbsp; Hahahahahaha.....  I guess it was some&nbsp;charity event at that club unbenknownst to me (although I\'d heard something about N\'Sync being in Miami that weekend) and N\'Sync was there too....Most of them had already left, however.&nbsp; Some guy gave me and one of the girls I was with&nbsp;VIP passes to get into the club (you had to have tickets) so we went in to check it out.&nbsp; The club itself wasn\'t that cool but the outside was great...It was on the beach (not that you could see it at night but you could feel the air) and there were beds everywhere and TeePees....it was crazy.&nbsp; People just hanging out on beds and some other stuff I\'m sure.&nbsp; I bet that place gets pretty scandalous!&nbsp; The blisters were worth it I guess \'cause how often can you say you got VIP tickets into club and saw/said hi to Usher in one night?&nbsp; It was kinda cool!&nbsp; I\'m not really one to freak out about famous people but I really like Usher\'s music and I think he is an amazing dancer so I thought it was neat to see him.&nbsp; He looks exactly the same as he does on t.v. He either has&nbsp;really nice skin or they must&nbsp;put&nbsp;a lot of&nbsp;makeup on him all the time.&nbsp;   I think we\'re gonna go up to Tampa this weekend to go to Busch Gardens.&nbsp; I\'m excited \'cause I love rollercoasters...theme parks really, as long as they have good rides....and Busch Gardens is a Six Flags park and they always have good rides.&nbsp;   I think that\'s it for the night...I\'m off to watch a movie and fix my hideous nails...  &nbsp;  &nbsp;   '),(18537,3590069,'16,July,2004','2004-07-16 00:00:00','    I tried to promise myself that I wouldn\'t bitch today but you all know me too well and after I read CJ\'s blog and felt her anger, I knew I was really pissed and was gonna have&nbsp;to get some things off my&nbsp;chest....err...huge boobs....  &nbsp;  First, CJ, I feel really bad for you because I know EXACTLY how you feel about the money situation...I severely feel that pain and I have yet to catch up on bills, let alone have any money for myself.&nbsp; I have bought five books ($5 a piece), a $5 jacket and a $8 shirt since I have been here...that is honestly it in over&nbsp;2 months....It sucks being poor.&nbsp; I borrowed the money from my mom for the hotel and scooter in&nbsp;Key West and have to pay her back VERY soon...nice mom, I love her!!!&nbsp; Anyway, like I said, I feel your pain with the money...Here comes my bad day and anger...  &nbsp;  H\'ok so, as you all know, this has not been the most pleasant experience to say the least.&nbsp; I am so angry with the people here and the company.&nbsp; Everyone is so incredibly selfish and two-faced it\'s not even funny.&nbsp; Everyone is out only for themselves and it\'s rare to feel like I can trust anyone.&nbsp; I flew out here because I was told I only had three days to get here for the job and had three days to get home before school starts.&nbsp; I was under the impression that that was the only option I had if I wanted the job.&nbsp; I reluctantly agreed, thinking everyone would be in the same boat.&nbsp; Not so much.&nbsp; I came out here having been told&nbsp;there would be other people without cars and that everyone would be switched up with partners all the time to make sure no one had to drive all the time.&nbsp; Not the case...I usually get sent with one person for about a week and it feels like I am the biggest burden on everyone because I don\'t have a car.&nbsp; It\'s ridiculous to feel like I have to ask for a ride to the office in the morning meetings and to beg&nbsp;someone&nbsp;to take me to the store to buy groceries.&nbsp; People bitch about having to&nbsp;drive every day when they sell with me, saying&nbsp;it\'s not personal to me but it&nbsp;always feels personal.&nbsp; The car thing is so bad....I would never have come out here without a car if I knew it was going to be like this; not in a million years.&nbsp; I was assured&nbsp;it would be fine but it has been anything but fine.&nbsp;  &nbsp;  Aside from the car thing and&nbsp;not&nbsp;feeling like&nbsp;I connect with anyone here as I have mentioned on numerous occassions, I am PISSED today&nbsp;to find out I can\'t go home early.&nbsp;&nbsp;As I said, I was told I had three days to get here and three days to get back, the reason I didn\'t drive.&nbsp; However, I was told when I was recruited that if I worked hard and got burned out that I would be allowed to leave a little early so I accepted the three day thing.&nbsp; Apparently the going home early thing is not true.&nbsp; Come to find out, another girl who starts school the same day as I&nbsp;do gets to leave 10 days before me.&nbsp; WTF mate?&nbsp; Seriously, and she has no reason other than she wanted a break before she starts school.&nbsp; I talked to&nbsp;my boss&nbsp;about it and he said I had to talk to someone higher up than him to make the decision.&nbsp; Not only did this girl get to leave&nbsp;more than 3 days&nbsp;before school starts, so do many other people (two plus weeks in one case) AND two guys left for 4 days to go on a family&nbsp;vacation.&nbsp; How is&nbsp;it policy to say&nbsp;you only have 3 days on each&nbsp;end to get here and&nbsp;leave and that that policy is non-negotiable when it obviously&nbsp;isn\'t policy in the first place?&nbsp; So I called the higher up today and got some bullshit blow off.&nbsp; He told me that he didn\'t hire that girl I was talking about (even though he\'s in charge of all the hiring for my region) and he didn\'t know why&nbsp;she got to leave&nbsp;earlier than me and that if he let me leave early, he\'d have to let everyone.&nbsp; BS!!! Obviously everyone else&nbsp;chose their end date and wouldn\'t need to leave early because they already were and I was given one as a&nbsp;condition of accepting the job.&nbsp; Also, he&nbsp;wouldn\'t have to change anyone\'s date if he\'s the boss....all he has to say is he changed my and NUNYA...he doesn\'t have to answer to anyone or say why he changed it.&nbsp; Can\'t he do whatever he wants if he\'s the boss?&nbsp; I was like, \"it\'s funny how it\'s policy but it\'s not really policy for everyone.\"&nbsp; And, \"why did two guys get to leave to go on vacation?&nbsp; None of the rest of us are allowed to take even a day off without paying rent and I can assure you they did not have to pay rent when they missed those days...why?&nbsp; Because their family is friends with my boss\'s and their mom said the only way they could come sell was if they could leave for a few days to go on a family vacation.&nbsp; So if my mom had called and said the only way I could have the job was to be able to leave a few days early to buy my books, do my reading and sleep, then would I have been able to negotiate my end date?&nbsp; Or maybe if I was a&nbsp;member of their church?&nbsp; I normally wouldn\'t pull that card because I think com,ments like that are usually illfounded&nbsp;and rude but&nbsp;I feel that has something to do with the way things are run here unfortunately.&nbsp; I only agreed to the rule because I thought everyone was subject to the rule.&nbsp; Fuck that!!!!&nbsp; To top it off, the guy told me that even if he wanted to change people\'s end dates, that mine wasn\'t even really a good enough reason to change my date...that other people had much better reasons than me.&nbsp; I guess I could have lied and made up a reason but I\'m not like that.&nbsp; He said someone had a family member close to death and that they weren\'t leaving...I told him if one of my family members was sick I\'d just quit.&nbsp; Isn\'t this company run by a family-oriented group of people?&nbsp; Obviously money is way more important than making sure your staff is happy.&nbsp; I told that guy that at this rate I had no intentions of coming back to this company in the future because I feel like I have been lied to from day one....many other things go into that but those are for another day....I told him I had wanted to come back next year but that the experience has been so negative because I was not given honest information.&nbsp; I could have brought my car and I could have taken more time to get here and home.&nbsp; I\'ve surely lost money not having a car because my selling time, my bathroom breaks, and my lunch hours are all dictated by whoever is driving.&nbsp; That sucks!&nbsp; I could sell longer and earlier if I had my car and, in turn, make more money.&nbsp; The guy just tried to blow shit up my skirt and tell me that I\'m a really strong person, much stronger and smarter than the average person as a law student, and that I will be just find working until the last minute.&nbsp; Bullshit but nice try!&nbsp; He said the only thing he was willing to do was maybe let me leave a day earlier if I changed my attitude because he didn\'t want me to go back home and talk negatively about the company.&nbsp; Fuck you and your fucking one day!!!&nbsp; What the fuck is one day gonna get me?&nbsp; It\'ll cost me $100 minimum to change my flight...not worth it for one day.&nbsp; Fuck this place.&nbsp; I\'m so glad it\'s over in a month.&nbsp; If I could, I\'d quit now and come home and sleep and get a job but, if I did, I\'d lose all the rest of my commission and any money I haven\'t been paid yet for this pay period.&nbsp; The bulk of my money doesn\'t come \'til September so that\'s out of the question \'cause I\'d essentially have worked for free if I quit before my end date.&nbsp; Fuck these people.&nbsp; I wouldn\'t bad mouth the company if they were honest and actually upheld their \"policies\" across the board to all people, friends, members of the church, etc.&nbsp; I\'m so tired and stressed and I\'m gonna come back that way if I don\'t get a couple days off to sleep and get treated with some common courtesy and honesty.&nbsp; I can\'t believe they won\'t give me an extra few days...Fuck \'em!  &nbsp;  Do you feel sorry for me yet?&nbsp;   '),(18538,3590069,'15,July,2004','2004-07-15 00:00:00','    I love reading everyone\'s blogs...I really really do but they kinda make me sad...well Emily\'s do anyway...how much fun are you having?&nbsp; I\'m so jealous...but I\'m glad you are having such a wonderful time.&nbsp; CJ\'s blogs make me laugh all the time and miss her rampages....I hear her voice when I read them and it makes me smile...  &nbsp;  I can\'t wait to see everyone again...when is everyone getting back?&nbsp; I think I\'m getting back around the 20th...we all need to get together that Saturday if everyone is back and catch up and drink some beer....I really need something anyway.&nbsp; I did get to go have a beer and play pool for the first time since I\'ve been here and that was very much a past due&nbsp;evening.&nbsp; Too bad the \"friend\" I was with tried to get me to go home with him AFTER he\'s been hooking up with a friend of mine.&nbsp; Ruuuude...why can\'t men just be friends, normal people, sometimes?&nbsp; V....if you\'re reading this....why have I never met a man who didn\'t want to sleep with me?&nbsp; hahahahhaha......  &nbsp;  Hope everyone is well and happy...I don\'t have much else to say tonight...good night all!   '),(18539,3590069,'13,July,2004','2004-07-13 00:00:00','    H\'ok so I got soaked in a rain storm yesterday and woke up feeling like I\'d been hit by a Mack Truck...Everything was cloudy and I couldn\'t talk for most of the day without sounding like a 75 year old woman who\'s smoked for 60 years....Didn\'t help that I stayed up most of the night on the phone....CB...I\'m paying for that today but I\'m not complaining \'cause it was great....Let\'s do it again sometime, K?    About 5 weeks left....Do you know how exciting that is?  I\'m actually pumped to go back to law school; well, pumped to leave this job and see everyone again anyway.  I\'m hoping to leave a little earlier than planned so I can rest and see people before school starts.  I\'d like to go camping if at all possible...anyone interested?  I haven\'t really done anything fun or interesting lately to detail...I guess I went to a huge flea market in Fort Lauderdale that is open daily...It was interesting...Some neat stuff but also a LOT of junk...The outside market was like people with stuff off the back of their trucks...I\'m sure some of it was hot and a lot of it was crap you couldn\'t sell at a garage sale...  I read The Stepford Wives and then went to see the movie...It was cute but not as good as I\'d hoped.  The movie was much better than the book but I still think it could have been done better.    H\'ok that has to be enough for today...I need to go to bed....   '),(18540,3590069,'07,July,2004','2004-07-07 00:00:00','    urlLink    Lisa, Shay and I gator hunting...&nbsp; urlLink   '),(18541,3590069,'05,July,2004','2004-07-05 00:00:00','    urlLink    Birthday girl...This one is obviously from home...&nbsp; urlLink   '),(18542,3590069,'05,July,2004','2004-07-05 00:00:00','    urlLink    I got this one on my own...&nbsp; urlLink   '),(18543,3590069,'05,July,2004','2004-07-05 00:00:00','    urlLink    Lisa and I driving...&nbsp; urlLink   '),(18544,3590069,'05,July,2004','2004-07-05 00:00:00','    urlLink    Me and a baby gator...&nbsp; urlLink   '),(18545,3590069,'05,July,2004','2004-07-05 00:00:00','    urlLink    Gator on the rampage...&nbsp; urlLink   '),(18546,3590069,'05,July,2004','2004-07-05 00:00:00','    urlLink    Gator hunting...&nbsp; urlLink   '),(18547,3590069,'05,July,2004','2004-07-05 00:00:00','    urlLink    Lisa and I before snorkeling...&nbsp; urlLink   '),(18548,3590069,'05,July,2004','2004-07-05 00:00:00','    urlLink    Lisa and I on the mopeds we rented in Key West...&nbsp; urlLink   '),(18549,3590069,'05,July,2004','2004-07-05 00:00:00','    urlLink    This is supposed to be a picture of a nurse shark...&nbsp; urlLink   '),(18550,3590069,'05,July,2004','2004-07-05 00:00:00','    urlLink    Don\'t know if you can see it but that\'s a Sting Ray...&nbsp; urlLink   '),(18551,3590069,'05,July,2004','2004-07-05 00:00:00','    urlLink    This fish was an electric blue color...It was so cool up close but I couldn\'t stay close enough and \"still\" enough to take a good picture...&nbsp; urlLink   '),(18552,3590069,'05,July,2004','2004-07-05 00:00:00','    urlLink    This thing was HUGE!!!!  It came at me out of nowhere...&nbsp; urlLink   '),(18553,3590069,'05,July,2004','2004-07-05 00:00:00','    urlLink    Flounders? No idea....&nbsp; urlLink   '),(18554,3590069,'05,July,2004','2004-07-05 00:00:00','    urlLink    Too bad I don\'t know what these are called...&nbsp; urlLink   '),(18555,3590069,'05,July,2004','2004-07-05 00:00:00','    urlLink    This is supposed to be a picture of a nurse shark...&nbsp; urlLink   '),(18556,3590069,'05,July,2004','2004-07-05 00:00:00','    urlLink    Scary...&nbsp; urlLink   '),(18557,3590069,'05,July,2004','2004-07-05 00:00:00','    The day wasn\'t going so well for sales but I pulled out two in the end and the pictures I had developed today didn\'t really turn out...that always sucks...but as I was about to post them for you all, my day just got better because one of my roommates walked in and told me she bought me some icecream to brighten my day...what a nice girl and you all know how I feel about icecream...It always makes the day better...  Here are some pics for you all to enjoy, some from home, some from here...   '),(18558,3590069,'05,July,2004','2004-07-05 00:00:00','    I finally got two days off in a row...too bad they were the only two days off in a row I get all summer...This money better be worth it in the end \'cause I can\'t wait to come home...Only a month and a half left...yeeeeeaaaahhhhhh!!!!  H\'ok so, for my two days off I went to Key West with some people I work with...Did I mention that I don\'t like the people I work with?  They are immature...self-centered...ruuuuude...immature...the guys deserve to be called immature twice because nothing else describes them.  Lisa and I were the only girls who went...You\'d think it would be a girl\'s dream to go away with 7 guys and only one other girl but you\'d be wrong in this situation.  They were so ridiculous and I was embarrassed to be with them half the time.  They don\'t think before they speak and if I heard one more remark about all the gay people down there I think I might have killed one of them.  I had certainly had enough when I was playing some music while driving home and they all started making remarks about my \"lesbian\" music...I coolly mentioned it was funny how they\'d all been talking shit about gay people the whole time and how since they knew nothing about me they might want to keep that stuff to themselves because they didn\'t know who they were offending.  I hate when people make bold assertions and don\'t bother to care if they\'re offending people.  I definitely could have spent my only free two days with better company which would have made the \"vacation\" much better but I tried to be as positive as possible...I\'m trying so hard here; it\'s more difficult than I ever expected.    I knew the trip was gonna be bad news from the start when we reserved a spot to go snorkeling Sunday at 1 and then left at 9 for a 4.5 hour drive.  No one would get out of bed and we missed our boat.  Luckily, the people were nice enough to let us go Monday morning at 9 but refused to let us go in the afternoon because they didn\'t want to risk us not showing up again, \'causing them to hold seats AGAIN for a popular time and trip.  Since we couldn\'t go snorkeling yesterday we checked into our hotels (I won\'t even get into that drama) and rented mopeds to drive around the island.  That was a ton of fun!  I was scared at first because I didn\'t want to crash but you get used to them right away.  I caught some sun and enjoyed touring Key West and the ability to weave in and our of crappy tourist traffic.  No one wanted to do anything I wanted to do of course so I spent a little time by myself (not that I minded), and then got ready to go watch the fireworks.  Not only was I late getting to Key West because of these idiots and missed my snorkeling trip, I missed the fireworks, or rather I had to watch them driving to the beach on my moped, because I agreed to wait for one of the guys who took FOREVER to get ready...worse than a woman...  No one goes out here of course so I pretty much went to bed a 12 so I could get up early and go snorkeling.  I was hoping to go Sunday so I could sleep in for once this morning but everyone ruined that for me and I had to get up at 7:30 to get to the boat on time.  Everyone, as expected, was taking their sweet time so I left without them and went down by myself so I would be sure to make the boat.  I made it of course and snorkeling was amazing...A bit intimidating at first but really fun...I saw a Man-O-Ray and a Nurse Shark...it was so cool!  I also saw tons of super bright fish...They looked like neon signs!  I\'ll try to get the pictures online for everyone to see.    The snorkeling and moped riding were by FAR the best parts of the trip.  The rest I could have done without, especially the drama of returning the mopeds.  A few people had to pay for damages and scratches to their mopeds.  They were pissed and threw embarrassing fits.  Hello, read the K you\'re signing...The guy went over all the existing damages and scratches and we all agreed to pay for any additional damage caused during our use; that\'s why he went over everything and marked all the damage down with us in the first place, telling each and everyone of us to check for damage so we wouldn\'t be charged.  They were all shocked when they had to pay.  Maybe it was expensive but they should have paid attention in the first place.    Man I feel like all I do is bitch when I write in my blog.  I gotta be more positive...oh wait, I am being positive!  The snorkeling and moped riding were really great though!  I had a good time despite the company and drama.  I think I\'ll have to rent a car and go by myself or with my friend Lauren some other weekend if I can.  I\'d like to do some more stuff I didn\'t have time to do this weekend.   '),(18559,3590069,'12,August,2004','2004-08-12 00:00:00','    Yeah, only five days left of work well, one or two really, maybe, since there\'s a hurricane coming.  Funny, it\'s called Hurricane Charley and there\'s one called Hurricane Bonnie up north that\'s caused a mandatory evacutation of Tampa.  It would have been even funnier if the one down here was Hurricane Clyde....hahaha...Anyway, yeah so I will be home in less than a week...I\'m so excited!!!!  I don\'t think we\'ll be working tomorrow since it\'s supposed to be really rainy and windy from the hurricane, Saturday too.  Monday might actually be my last day \'cause I don\'t plan to work on Tuesday and Wednesday is my last day...I guess I could work then...Whatev...I\'m kinda excited there\'s a hurricane...could be interesting...  I can\'t believe how fast the summer actually went, as bad as it was at times...it went pretty fast.  I can\'t believe we start school so soon!  Scary but good.   What esle?  I sold an NBA player\'s wife pest control today...I guess that was kinda cool.....Carlos Arroyo from the Utah Jazz if anyone follows basketball or likes him...I guess he\'s really good....I don\'t follow basketball so I wouldn\'t know anything about him but people in my office were really excited.   Just winding down the day and burning some people\'s cds before I head out of town so I guess I\'ll get back to that...Can\'t wait to see everyone....   '),(18560,3590069,'05,August,2004','2004-08-05 00:00:00','    H\'ok so I have 11 days left of work!  I am so excited!!!!  I managed to get my end date changed from the 19th to the 18th so I\'m flying out late the 18th instead of the 20th just so I can get home earlier even though it won\'t be until 2:30 in the morning on Thursday...I can\'t leave soon enough and CJ has graciously agreed to either take my car to the airport for me or pick me up at that awful hour because she loves me.  Thank you so much CJ...you are the best!  I promise to repay you somehow for being so good to me.   We have a sales incentive this week that might land us tickets to a Marlins game on Saturday. ..could be fun!  On Tuesday we had a sales incentive for the day in which we could earn chances to bat in a wiffle ball (however you spell that) game....Depending on how you did, you could win a pretty hefty sum of money.  I randomly but luckily had my best day selling of the summer that day and beat everyone out for chances to bat...I earned 4 swings...no one else got more than one chance to bat and there were only three other people that even got to bat.  Take that!  I joked about striking out \'cause I\'m not that great at baseball so my boss gave me a practice swing.  Too bad I\'m apparently the Babe Ruth of wiffle ball \'cause it was a $10 hit.  We played in a parking lot and since I hit the ball all the way to the wall, I earned $10.  Alas, it was a practice swing.  I nailed the first real swing, not too hard with a plastic bat and a wiffle ball...and earned my first $10.  The second and fourth hits were fielders so I only got $5 for each.  After I\'d hit the first two into the wall my boss said he didn\'t want to just give his money away and that if anyone caught the ball off the wall it\'d be an out and I\'d lose the $10 I earned.  Some schmuck caught the ball off the wall from my third swing and I lost that $10 but I  didn\'t care \'cause I came out one top with $20 earned and the satisfaction that I kicked everyone else\'s butts!  It was great!!!!  I hope everyone is doing well and enjoying what\'s left of their summers!  It\'s coming quickly to an end and I can\'t wait to catch up and exchange stories....  Take luck!  hahahaha I\'ll have to explain that to you all later.....   '),(18561,3323485,'26,May,2004','2004-05-26 00:00:00','    What do other people see me as? A Statistic? An automated prgramming/webmastering/making an ass of himself machine? Why not? thats really all I am. When you look through the blog listiogs, and you see Peter Ooms\' Weblog, did you click on me because you liekd the name?   ive just noticed something...ive had no effect on anything. if i was never born, the only differences would be a couple of schoolbooka and this blog wouldnt be here. holy s**t now I really am depressed. im just off to do some homework. might aswell kill myself for all the dofference it would make to everyone elses life.   '),(18562,3323485,'24,May,2004','2004-05-24 00:00:00','    urlLink    Hey! Check out this cool image I made!&nbsp; urlLink   '),(18563,3323485,'24,May,2004','2004-05-24 00:00:00','    urlLink idleworm: tech - microsoft advert Look at this - its very funny because its true!   '),(18564,3323485,'21,May,2004','2004-05-21 00:00:00','    In case you havent guessed, i hate year eihts. they think theyre so smart with their \'no longer the smallest guy\' attitude. it just gets me pissed off. I was walking around, minding my own buisiness, and a bunch of 4+ year 8s appeared from nowhere and started pissing about with me. All I did was call one of them a \'stupid git\' and it got dirty. I wont go into the details, bot there was 4+ of them, 1 of me and they were bigger, so you can guess the rest. Just before it got bad, two teachers appeared from nowhere.   I never thought I would  ever  hear myself say this, but thank god for teachers!   '),(18565,3323485,'20,May,2004','2004-05-20 00:00:00','    Hello World Today we did the 1500m boys trials. I think i did $h17. My time, if anhybody wants it, is 7 mins 49 secs. I had another of those thoughts today. If I were to die, suddenly now, would people remember me, and if yes, what as. Would they remember me as \"that little ba$+ard who gave me the finger\" ot \"a cool goy who f****d it up\". -Well, partof me today wanted to commit suicide.   Holy s**t! Ive just realised the implications of what I just said. HJeres me, a 12-year-old, sitting here, chewing some plastic, seriously concidering  suicide !  Well, just if I do snuff it today, can you please remember one thing. Please...   For centuries, the world has been, and will continue to be rigged with people both willing and able to exploit the ignorant. Someday, nobody knows when, all the ignorant who have been scammed willl rise up and defeat theese vast bodies. The rebels will suceed and life will be a fair thing. Until...    Ignorance Is Weakness  (c)2004 Peter W. Ooms   '),(18566,3323485,'19,May,2004','2004-05-19 00:00:00','    Dont you just  hate  it when you know you are right, and someone persistant keeps on contradicting you! I will say it now and say it clear. In the Matrix, Neo takes the RED pill to exit the matrix. I was arguing about this with Paul one day, and, along came Dominic and, because he fancies Paul, agreed with him. HE\'S NEVER EVEN SEEN THE F**KING MATRIX!!!!!!!! This really makes my blood boil. If I were prime Minister, I would outlaw homosexuality, outlaw this qstupid theory everyones got. I will make my views clear...  A GOOD WINNER IS BULLS**T. AS IS A GOOD LOSER. PEOPLE WIN THINGS SO THAT THEY CAN TAKE THE P**S FROMN THE OTHER TEAM. WE WON WE WON WE WON. BAD LOSERS!   '),(18567,3323485,'18,May,2004','2004-05-18 00:00:00','    Im sure you will all agree with me that exactists (perfectionists) are a f*****g nucianse! I say i wish I was dead, then someone comes over 2 me and says \"how can you say that? its i wish i WERE dead! speak properly\"   I ask u...this world is WiErD   '),(18568,3323485,'16,May,2004','2004-05-16 00:00:00','    Well Ive just finished my homework and it has taken me just under four hours. I woke up at 11:00 AM GMT and have just finished it now...at 3:19 GMT. I hate homework. Why? It reminds me of school. Also, it is a tedious wase of time on pointless busy work.  If there are any teachers reading this, DOWN WITH HOMEWORK!!! I think,on a none-uniform day, I will come to school in a t-shirt that says DOWN WITH HOMEWORK! That should get \'em!! Well, Ive got to write my PowerPoint Presentation to a CD-R so if you dont mind?...   '),(18569,3323485,'15,May,2004','2004-05-15 00:00:00','    umm...Im hard pushed to find anything interesting to say about today. Mums still doing her stuff fow work, im still left tidying my room and all that crap.   No new E-Mail Messages. As Ever, no normal humans need/want/can be arsed to communicate with me. Cant say I blame them...im always a bore and a sadass. Nothing new or interesting. Always going on about space and computers and all that f**king s**t. I wonder...if I were to drop dead right now, would people remember me...if yes, what as... I\'ll think about that one and reply as soon as another insignificant thought enters my head.   Its one of \'em days, i\'m afraid.   '),(18570,3323485,'14,May,2004','2004-05-14 00:00:00','    Its not often that I will devote a whole blog to a peice of software, but I must recommend the MSVDM. You can have up to four virtual desktops on any one machine...and best of all...ITS FREE!!!!!  For more info, see http://www.microsoft.com/windowsxp/pro/downloads/powertoys.asp .   It only works with Windows XP   '),(18571,3323485,'13,May,2004','2004-05-13 00:00:00','    The title says it all.   Apparently, today is \'national s*x day\'. Apparently: my true love has discovered she loves me yesterday at midnight and something good will hapen to me between 2.00 and 4.00 today. Well, IT DID  Though its got nothing to do with relationships  shall I tell you anyway? OK,. then    SCHOOL ENDED!!!    and its friday today! That means my first weekend blogging.  Isnt it funny that today was hilariously boring and yet I have a lot to say?  Life is Wierd.  t . h . e . . . t . r . t . h . . . i s . . . o . u . t . . . t . h . e . r . e . . . . . .   '),(18572,3323485,'13,May,2004','2004-05-13 00:00:00','    Yo (again) My mate, paul, has gone and written some script 4 a \'movie\". the movie is called war is a video game. he employs (?im only f**king 12) me to make him a website. if u r wierd enough, go 2 www.wiavg.tk - u will need Macromedia flash player 6.   Whats more is that theres about 15  people in this \"movie\", half of them dont know it and NONE of us have seen the bloody script. wierd or what? Its this kind of thing that makes me question the worlds sanity.   ALL I WANNA BE IS A F**KING REGULAR GUY WIV A DECENT JOB AND A COOL HOUSE WEN I GRO UP. NOT ONE OF THEESE COLLYWOOD FAKFACES HOO IS IN EVERY SINGLE EPISODE OF MISSION MARS BUT COULDNT TELL THE RED PLANET FROM HE MOON IF IT WAS SHOVED UP HIS A*S!!  Im so bored, someone plz contact me it wud be so nice to talk to a human rather then a deserted weblog and a couple of searchbots and a computer  it gets really boring after a while!   '),(18573,3323485,'12,May,2004','2004-05-12 00:00:00','    I am just here to live out my life, and enjoy as much of it as possible. If that means influencing other peoples, then so be it. What\'s more, if anyone has a problem with that, they can stick it right up their ****.   '),(18574,3323485,'12,May,2004','2004-05-12 00:00:00','    For centuries, the world has been, and will continue to be rigged with people both willing and able to exploit the ignorant. Someday, nobody knows when, all the ignorant who have been scammed willl rise up and defeat theese vast bodies. The rebels will suceed and life will be a fair thing. Until...    Ignorance Is Weakness   the truth is out there...   All Three comments were made by Me, Peter Ooms. Feel free to do whatever you like with them, but please, name me as the source and if possible my web adress (www.peterooms.tk)   '),(18575,3323485,'12,May,2004','2004-05-12 00:00:00','    \"Sometimes I wonder if I am the only sane person in the world. Then I look at my friends and realise that I must be or the world is doomed.\"   ( me, this morning )  Today, I will make my views clear about Britain jouning the EU:  NOOOOOOOOOOOO!!!!!  It will cost us Â£20,000,000 a  YEAR  travelling betweel brusells and Hamberg alone! Then theres all the jobs that will be lost, and most strongly,   the  POSSIBLE  increace in crime . Notice the word possible. I am not condemming anybody, just saying that the possibility is there. With any group of any size, there will always be a few \'bad apples\'.  and i bought a keyring yoday for fifty pence   '),(18576,3323485,'12,May,2004','2004-05-12 00:00:00','    hey. u may or may not know that i am into web design...anyway, ive just gone and made myself a new web site. go to http://www.peterooms.tk to see it (low-bandwith users may want to skip the intro)  dont 4get to tell me what you think!   '),(18577,3323485,'12,May,2004','2004-05-12 00:00:00','    hi.   i am peter, your average 12-year-old, bored stiff, hates hwk, all that s**t. by the way, i will try not to sware here, but this is where i will let off my raw thoughts, so i may let it slip a few times. i apologise in advance for any \'slips of the toung\'.  well, today is the 12th of may. 2004. done homework, etc, listening 2 music, generally p**sing about  theres some crap in the news about some bloke who got kidnapped in iraq and beheaded on a live webcam. idiot. anyone knows that with terrorism nowhere is safe, so he only goes to the source and gets killed. dont get me wrong, im not saying it was a good thing, i even feel a bit sorry for him, but i mean he was asking for trouble going to iraq at a time like this.  well, gotta go for now. bye   '),(18578,3323485,'18,July,2004','2004-07-18 00:00:00','    Well, i\'ve just come back from a 5-day school trip to france. Ben thought there was a ghost in our room. There was something wierd in our room-maybe a ghost but we shouldn\'t tamper with what we dont understand. The food was also a bit $h&#161;+ but apart from that it was OK. Gotta go because im txting this blog in. See you!   '),(18579,3323485,'08,July,2004','2004-07-08 00:00:00','    Hi, me again  (well, its not exactly likely to be anybody else, is it?)  I know and apologise for not having posted in a while. But. Today, I feel I have to.  Well, ive been questioning my will to live oday. I mean, it could be a lot worse...but also a lot better. But, then, ive got stuff to look foreawd to...drinking...driving...girls...  Or, perhaps, make somebody  like  me for a change.  I mean, theres Paul, but he is always talking about \'world domination\' schemes that never get off the ground. He thinks like a buisinessman, but this isnt neccessarily a good thing when trying to have a casual conversation :-D(joke, paul if you read this dont take it the wrong way):-D. Then theres Ben and Adam. Okay, but very very very annoying. Dominic and Jhonny really secerately hate me, they just pretend to like me because paul likes me and they want paul to like them and...you can guess the rest.  On a different subject, I got really close to having my first real girlfriend. Not a prank, a joke or a ten-second stitch-up as I have had in the past, but a real girlfriend. Well, that went down the plughole like most other things in my life which is, in the words of CM Burns (off the Simpsons) \"Doing that thing again\". She was acting wierd for a while...and so was I for that matter. I knew that Chalkstone was going to tell her I fancied her (just for the record, I DO but I like to keep it a secret because I am absolutely terrified of peoples reaction) so I told Paul I \'had to do domething\' and just went all break and lunch...wandering...like a lost soul...thinking...getting shouted at and things thrown at me...  (apologies if I am making anybody feel depressed, but it is your life, your responsibility, no body forced you to read this and so its all your fault. Sorry, but it is.  and dont come crying to me, okay buster?)  So, thats how the day went. In summary: -one millimeter away from my first kiss -one millimeter away from my first (i think) suicidal death -one millimeter away from getting a detention -one millimeter away from my first girlfriend Its actually been quite a millimeter sorta day. 5 millimeters away from...what? a totally different day? a totally different life? a totally different death? Who knows.   I hate days like theese. 0.5 cm away from...................what? (isnt it wierd that thursdays are the ONE days in my timetable where I dont get a maths lesson and I think about millimeters more today (a thursday) then than any other day?  I wonder if I were to drop dead now, would people remembver me? If yes, what as? an idiot? a wierdo? a sadass? a pessimist?   Well, until next time,  bye...   P.S. If you think that is depressing, you should see  urlLink THIS POST !  PPS: Change of plan: I am no longer comitting suicide.   '),(18580,3323485,'01,August,2004','2004-08-01 00:00:00','    Animal rights campaignes have it all wrong. First off, if one of their relatives gets cancer or something, they will change their their tune. Next, they fail to realise that animal testing is highly regulated and the government will stop any abusers. Lastly, just think. Only 86% of animals actually die. And, just think how many animals are saved by theese medicnes after they are releaced to the hospitals and vets. Common sense, if you ask me.   '),(18581,4127797,'03,August,2004','2004-08-03 00:00:00','   Once upon a time there was a toad named Skip.  Skip liked to hop, not skip.  '),(18582,4127797,'16,August,2004','2004-08-16 00:00:00','   I won\'t mention Iraq and Afghanistan (nor any other foriegn policy) because we all know how Bush screwed that up.  OK, you all know about Bush\'s economic policy failing to create jobs( urlLink http://www.nytimes.com/2004/08/06/business/06CND-ECON.html?ex=1092888000&amp;en=5336cf8a4f6027a7&ei=5070&amp;hp ).  We all know that Bush\'s tax policy is shifting the tax burden on themiddle class ( urlLink http://www.msnbc.msn.com/id/5689001/ ) and that the gapbetween the haves and have nots has increased more under Bush( urlLink http://www.msnbc.msn.com/id/5726153/ ).  Oh yeah, Bush has been rolling back consumer, worker, environmental, andsafety regulations (say good by to all you worked for Mr. Nader):  urlLink http://www.nytimes.com/2004/08/14/politics/14bush.html?pagewanted=print&amp;position=   What else can Bush do wrong?  Oh yeah, education.  NCLB encourages charterschools.  But surprise, surprise, charter schools are performing at lowerlevels than traditional public schools: urlLink http://www.nytimes.com/2004/08/17/education/17charter.html?hp   '),(18583,4127797,'09,August,2004','2004-08-09 00:00:00','   Funny how the GOP spinsters were eager to say that Kerry would receive a \"bounce\" of up to 12 points in the poll.  This despite the fact that most polls indicate that far fewer than 12% were \"persuadable\".  True to form, the Republican crew was quick to point out that Kerry did not receive the large bounce that they predicted.  And, of course, the pundits were quick to echo this sentiment throughout the media.  The \"Echo Chamber\" at it\'s best.  However, the point of the convention was not necessarily to deliver a knock-out blow as did Clinton in \'92 and \'96, but to give the country their first taste of John Kerry the man.  Going into the convention, Kerry was badly losing the \"nice guy\" contest as well as several other key indicators, including who would best handle Iraq and the war on terror.   OK, OK.  Rather than go through my analysis, I\'ll point you all to two other sources that are quite detailed:   urlLink http://www.democracycorps.com/reports/index.html   urlLink http://www.americanprogress.org/site/pp.asp?c=biJRJ8OVF&amp;b=134332   So, despite the negative comments by all the pundits, it appears as though the convention was much more positive than it is given credit for.  Kerry has made up some ground in several key areas.  ---  Analysis aside, there were several moments during the convention that stood out the most.  Yes, being in the Fleet Center during John Kerry\'s acceptance speech was a definite highlight.  But Johnk Kerry\'s leadership was most highlighted by speeches by Gen. Wesley Clark and former Senator Max Cleland as well as his swift boat crewmates.  Despite the constant claims to the contrary, Kerry described his main objectives and pointed everyone to his website for the more detailed descriptions.  All in all, I\'d say Thursday culminated a week that demonstrated Johnk Kerry\'s commitment to his country.  I with the paid pundits paid a little more attention to the actual content of the convention instead of editorializing throughout every speech.  Other moments of note include watching John Edwards speech at a bar with a bunch of strangers and having a good time.  Barack Obama\'s speech definitely lived up to the hype I had given it.  He spoke with such eloquence and passion about the Democrats\' desire for unity and moving forward.  His comments reiterated the positive message of opportunity and unity by Bill Clinton on the conventions opening night.   What wasn\'t on TV or anywhere else was a speech by Senator Joe Biden to the college Democrats, perhaps the best I\'ve ever heard.  He called for mutual sacrifice and mutual benefits in a time of increased security risks.  His message of unity included the fact that \"rich people are patriotic too\" -- something that too many people on the left fail to recognize and appreciate.  There are plenty of wealthy liberals out there, reason to take a second look at the traditional class warfare that so many on the left try to fight.  Progress, in my opinion, will be more quickly realized when we work together instead of dividing ourselves on class lines.  The conservatives try to divide the country by race, religion, etc.  Progressives should not try to resort to similar strategies.  For me, the choice in November is now clear: unity, hope and opportunity or division, fear, and the status quo.  My vote is for the former.   '),(18584,4127797,'06,August,2004','2004-08-06 00:00:00','   Surprise, surprise.  Job growth is stagnating:   urlLink http://www.nytimes.com/2004/08/06/business/06CND-ECON.html?hp   Two months in a row with less job creation than expected.  Only yesterday, the media and economists were predicting over 220,000 new jobs:   urlLink http://www.msnbc.msn.com/id/5612752/   I\'m no economist, but that\'s not good.  OK, no more \'pessimism\'.  On the plus side, the Bush Administration is trying to establish new \"Flex time\" legislation:    urlLink http://www.latimes.com/news/politics/la-na-flextime6aug06,1,3141357.story?coll=la-home-headlines   For those not familiar with the latest  urlLink Newspeak , \"flex time\" is a euphemism for removing overtime benefits.  The Economic Policy Institute has a few publications:   urlLink http://www.epinet.org/content.cfm/briefingpapers_comp   urlLink http://www.epinet.org/briefingpapers/familyf_bp_1997.pdf   I do applaud the Bush Administration working on legislation other than faith-based constitutional amendments, but I still disagree with his economic policies.   By the way, I just finished reading Robert Rubin\'s (Treasury Secretary during the Clinton Presidency) new book,   urlLink In An Uncertain World  , and would recommend it.   '),(18585,4127797,'03,August,2004','2004-08-03 00:00:00','   While setting up a blog for the  urlLink Stanford Democrats , I set up my own personal blog as well.  You can visit my website,  urlLink Gilbert Martinez  (threw that in to bump me up on Google), for more personal info and to get an idea what I\'ll most likely be posting about.  If you\'re too lazy to go to my  urlLink website , you can expect posts on my science interests, political interests, and general philosophical rantings.  I\'ll post sometime soon on my impressions from the recent Democratic National Convention.   '),(18586,3441713,'29,May,2004','2004-05-29 00:00:00','   Yesterday about 20:00 o\'clock I went to Iduna, cause there were playing some hardcore bands, including For the glory (A portugese HC-band). For the glory came with us; (Miranda, Jelte, Kelvin and me) to Miranda\'s house (They were going to crash the night there). It was waaay cool! and I had a great time, although sometimes I couldn\'t understand a word from that Portugese language, I had fun! :)  Today at three o\'clock they went to Germany for another show, too bad :( I had 26 kisses from 8 portugese For the glory band members! Haha. Yaaay, I\'m not so tired anymore.. Tonight, that will be round 8 o\'clock I\'m again going to Miranda\'s place, to watch a movie :) Well, that\'s all I needed you guys to know.  Byeee  '),(18587,3441713,'25,May,2004','2004-05-25 00:00:00','   Went to work today! it is okay :) and I miss Jelte :( I didn\'t see him for only a day.. but still, I miss him terribly! I have to stay at home tonight.. stupid mommy. Tomorrow I will see him again, finally! :D Hmf, I\'m tired and I care for a smoke.. but my mommy doesn\'t know that I began smoking again.. so.. blah.  Jelte is just like.. SO sweet! I love him with whole my heart.. and I\'m not planning to lose him in what kind of way.  Byeee  '),(18588,3441713,'24,May,2004','2004-05-24 00:00:00','   Parents.. buh, Sometimes I just want to strangle them with my own two hands. But still, they are my parents, so I love them and blah blah bla. But by mom is so damn stubborn about stupid things.. She treats me like I\'m an five year old.. And I hate it.   And Woohoo, I finally got me some work! starting tomorrow.. It doesn\'t earn much (2,47 in an hour) but I\'m glad that I can do something in the afternoon and earn money for it! :)  By the way: Hello :)  '),(18589,3441713,'08,June,2004','2004-06-08 00:00:00','   My neighbour died sunday.. that\'s like a big shock for me.. cause I\'ve not even get over the death of Jink yet.. I was really confused and didn\'t know what to do.. My parents said I needed to talk more but I just can\'t, it\'s too hard for me.. My dad asked me if I wanted to come with them to see her one more time.. but I can\'t and won\'t.. I just can\'t believe that someone with that age has to pass away. I\'ts not fair, it really isn\'t. Damn it. Jelte cheared me up that night, he was all conserend about me.. Really sweet :). The funeral is friday, in the same room where the funeral for Jink was, I think that\'s going to bring up all of the memory\'s.. Hmf. I really can\'t let this go :( but I have to, even if it\'s too hard.   I\'ll have to miss Jelte today, cause I have to stay at home.. That\'s what my parents said.. And oh yeah, I got fired last thursday.. buh. stupid people and stuff.  Byee  '),(18590,3309541,'26,May,2004','2004-05-26 00:00:00','   Ok, Now we\'ve done it.  NE Florida has finally followed much of the nation and gone past $2.00(US) per gallon.  Someone told me it\'s expected to go to $3.00 by the end of the summer (unless the gas companies stop their extortion).  My theory is:  Per year, bring the price up to a new high, bring it down (but not nearly as low as before) when everyone finally screams loud enough, then the following year go to a newer high.  And so on....  this way they get the pricing/profits up incrementally.  All while blaming OPEC.  '),(18591,3309541,'11,May,2004','2004-05-11 00:00:00','   I just heard on a news story that documentary evidence has been obtained that oil refineries are artificially inflating the prices at the pump.  The oil barons  claim  the war and OPEC are to blaim but as the report I saw showed, they are merely hiking up the profit margins.  While am an advocate of the free market system that\'s supposed to balance supply with demand and price, this is obviously a case of abuse of the system.  The oil companies have us over a barrel (pun intended) as the ability for americans to conserve gas by using less pretty much only works for those chosing to drive during their vacation (OK so the American penchant for bigger vehicles could be curbed, but for those of us with a large family, that\'s not an option).  The real problem is that as gas prices rise, all prices rise to pass on the cost of transportation to market. I think it\'s pretty sick that oil companies are willing to sacrifice a long term economic recovery for short term gains in their profits.  Eventually, the higher prices at the pump will even affect the oil barons for every thing they purchase will cost more.  Eating away at any gains their bank accounts might\'ve had.  My $.02 (or drop of gas).  '),(18592,3309541,'11,May,2004','2004-05-11 00:00:00','   Greetings to the world! As this is the first time I\'ve \"Blogged\" (is that a word?), I\'m gonna keep it simple. I\'ll probably post something about myself as that\'s kinda the point, right?  Ok, um... c\'ya.  '),(18593,3309541,'11,June,2004','2004-06-11 00:00:00','   Ok, so I\'m on vacation with my family and kids in the mega-suv and was passing through southern Tennessee today (Did I mention this was a major road trip?), and we are suddenly stopped on the freeway (as in parking lot).  We\'re about 4 miles from the last exit and can\'t turn around to go another way if we wanted to.  Anyway, we start to look around at the others in our unfortunate position and besides the truckers, we seem to be in the middle of  a college bash.  I mean people from all different states and walks of life.  So we\'re thinking \"What\'s up with this?\", are we near a college town or what?  It\'s not long before we notice a car with \"Bonnoroo or Bust\".  So what\'s Bonnoroo?  There\'s several thousand cars in this line with us and it\'s starting to give us the impression of some sort of WoodStock revival.  Well, come to find out  , sheesh that\'s what it is.  We\'ve been so out of touch I guess that we\'ve never heard of this.    Oh well, had to go online at the hotel to find out what it was that kept us over an hour and a half off our schedule.  Sounded pretty interesting if I were 10 yrs younger or maybe more plugged in to the latest music stuff.  But I\'ve got to say that one of the reasons I\'m not plugged in to the latest would be that I can\'t find anything I like.  It seems to be either too commercial (aka Brittany/Boy Band), or too bizzare/angry/bitter/sappy.  But I guess most of it isn\'t even my interests.    Most of what I listen to is either Christian based, or instrumental.  Not what you\'d call \"main-stream\".  Oh well.  I think it\'d be interesting to attend Bonnoroo or something like it, or at least to hear some of the stuff played there just to get a clue of what\'s out there to form an honest opinion.  I imagine that some of those bands are truly trying to make a statement or difference and are not tied in to a corporate entity.  By the way, did I mention that a Yukon costs over $55 to fill at todays prices?  I was the celebrity at the pump.  The folks around me were amazed.  If you see us on the road, wave.  Oh well, Ciao for now.  '),(18594,3309541,'12,July,2004','2004-07-12 00:00:00','   Wow, I just read an interesting submission to Slashdot today  urlLink Here , about the fallout patterns from the nuke tests (both above and below ground).  Look like the state of Delaware (I was born there) was covered several times with the stuff from multiple tests.  I wonder if that explains why my grandfather (paternal) had skin cancer, my grandmother (also paternal) had thyroid cancer as did my father, and why recently my uncle passed away from pancreatic cancer?  I know it\'s a long shot, but it wouldn\'t surprise me.  '),(18595,3767000,'31,July,2004','2004-07-31 00:00:00','   another week gone.. 44 more days to prelims.. 28 more days to dnt portfolio Os submission.. less than a month! think i\'m not gonna get any sleep when the date gets nearer..   so many datelines to meet.. especially for syf!! sort of regretted joining it.. but it can be fun at times.. the designing part is killer! soOoOo many things to design.. should be called the design com and not the publicity com..    &nbsp;  God i pray that i\'ll be able to get enough sleep and You\'ll bless me with strength and joy for the new day ahead!    &nbsp;  my heart almost stopped when i saw the crowd at the notice board.. eng prelims oral results is out.. i was on the verge of tears! (well thats a bit of exaggeration..) but i was soOoOo scared!! i do not dare to look at my results of course.. jocelyn had to do it for me... ...     32/40!!!! first time in my entire life i\'ve got an A for english!!! haha.. ya my english is that bad.. i couldn\'t believe it!!! i was just so excited that i couldn\'t stop laughing.. i couldn\'t stop jumping.. i couldn\'t stop hitting annabel!!! haha... (too bad she was beside me)    at least something good happened this week..! =p Praise God!!  '),(18596,3767000,'26,July,2004','2004-07-26 00:00:00','   i love.. i love.. i love chocolates.. i love grey colour.. i love to take photos.. i love meself.. i love God.. i love being loved.. &nbsp;  i hate.. i hate.. hmm.. what do i hate.. i hate being hated.. i hate hatred.. &nbsp;  why must there be hatred and dislikes in this world? why can\'t there just be love and peace?  '),(18597,3767000,'23,July,2004','2004-07-23 00:00:00','   1st blog.. hope its not gonna be the last..    &nbsp;  the past week is terrible.. actually, every week is terrible in term 3.. teachers are giving tests and homework like nobody\'s business.. well, these tests are to prepare for us for prelims and Os.. but we don\'t even have the time to study! so what\'s the point? these teachers think that we ONLY have their subject.. wow.. piles and piles of worksheets.. what&nbsp;s**t is this? damn..   most of my friends are getting stressed up because of this.. hope i won\'t be influenced by them.. all of us gonna fall into the \'depression mode\' soon.. u teachers want us to work hard for our own good.. but is it really the best for us? what will happen if we reached the peak too soon? what will happen if we fall and can\'t find the energy to carry on?   the final words..  its for your own good..     &nbsp;  52 more days to prelims..   12 more hrs to eng prelims oral..    &nbsp;   but those who hope in the Lord will renew their strength.     They will soar on wings like eagles:     they will run and not grow weary,     they will walk and not be faint...     (Isaiah 40:31)  '),(18598,3767000,'14,August,2004','2004-08-14 00:00:00','   i hate myself.. how could anyone be perfect? i tried.. but i couldn\'t.. can anyone be? how?       i screwed up my eng O\'s oral.. it sucks.. it sucks..   i screwed up my friendships too.. it sucks too..       i didn\'t mean to hurt you.. i\'m sorry.. i really didn\'t mean to.. do you believe me? its just the way i am.. i couldn\'t help it.. i\'ll change.. i\'ll try.. yes i will..      friends are gifts from God.. i want these gifts.. i treasure them.. i really do.. all my friends out there..  i love you..  i really do..   who\'s willing to take that 99 steps for me? when will i ever have the chance to take only one step?        31 days to prelims..   15 days to dnt submission..  '),(18599,3767000,'07,August,2004','2004-08-07 00:00:00','   Would you know my name if I saw you in heaven?   Would it be the same if I saw you in heaven?   I must be strong and carry on,   Cause I know I don\'t belong here in heaven.      Would you hold my hand if I saw you in heaven?   Would you help me stand if I saw you in heaven?   I\'ll find my way through night and day,   Cause I know I just can\'t stay here in heaven.      Time can bring you down,    time can bend your knees.   Time can break your heart,    have you begging please, begging please.      Beyond the door there\'s peace I\'m sure,   And I know there\'ll be no more tears in heaven.      Would you know my name if I saw you in heaven?   Would it be the same if I saw you in heaven?   I must be strong and carry on, Cause I know I don\'t belong here in heaven.  '),(18600,3413512,'21,May,2004','2004-05-21 00:00:00','   T he wind and the waves are always at the side of the ablest navigators. Life is a journey with many routes to take. With so many people youâ€™ll be able to meet, for sure decisions will be hard to make. My life has been through rolling waters, deafening thunders and storm-tossing conditions. But my ship is strong and enduring and with the wind blowing on my direction, I can carry through this journey. 	 A t age 10, I was already leading my own ship. From the fast paced cities of Manila, I found myself living a simple and carefree life in the province of Pangasinan. For less than a year, I had to live with my auntie, and study in the public schools of the town for at that time my family endured a great deal of financial shortcoming. I had to do it for my family, but I realized what it had done for me. Doors opened up for me leading me to a vast horizon of my unexplored talents awaiting to be discovered, and waiting to be showed off.  	 M usic runs in our blood. My brother and sister play the guitar so well and my aunts and uncles have the musical fingers to fondle the ivories of the piano.  When I spent a year at the townâ€™s school, I was given the chance to be part of artistic students gifted with talent to harmoniously work on music. The star of the so-called â€œPangkat Kawayanâ€, an ensemble of bamboo pipes and shakers, I played the bamboo xylophone-the marimba. After tedious practices and laborious recitals, I was able to perfect a repertoire of sweet melodious love songs and heritage aria of the Filipiniana. And soon, I found myself, applauded by many people delighted after a free concert. Aside from being a musician, I also became laureate of the Filipino rhetoric following a second victory in an interscholastic competition. And still with luck by my side, I became the no. 1 in our batch. 	 B efore my journey to the province, I never realized my abilities of being a good student. With my confidence at that time emanating, as soon as my family resiliently overcame the problems, I prepared myself to set sail back to the place where I truly belong. Where sailors are more viscous and where competitions begin. After the stay in the province, I became determined to keep up with their standards and before long I was striding side by side with the best of the best students. Proof to this were the meritorious certificates, medals and incentives from my hard work. From a no place student, to an ace student. 	 M y intellectual development was alongside the cultivation and the maturation of my individuality. And from my daily experiences, and reflections upon them, I have formulated several strategies to deal with the stresses of daily life. Moreover, in every instance, I am required to alter my being, my physical features and my total self to carry through. But more often, these self-alternation, turn to be mere impersonations and costume changes that eventually lead to entertainment. For in my life I have become a little drummer boy, a mouse on suspenders, a German army man, and even a sheik from the land of the deserts.  	 S ails on a ship control the west wind, so let it not be drifted away, for dark waters may come its way. Everyday, I continue my journey, battling the course of the tides. Pursuing my dreams, conquering islands and meeting the people in it. Still and all, moments will come that we get tired and give up when all our aching muscles pin us down. We leave it all to the wind and let our ship drift away. Itâ€™s a good thing that we do not sail away without our companions. When itâ€™s too much for us to bear, they take control of our ships and lead us to wherever  we have dreamt of going to. They fulfill our dreams for us. My family sails and navigates me, and sets me free , without binds, without bounds.       L ife is like the ocean, you never know how deep it is. You cannot always dive to the heart of it, but at least we are given that single chance to sail on it and enjoy it .  '),(18601,3376420,'19,May,2004','2004-05-19 00:00:00','   here are some jokes, not really that funny, but i like em.  1.Yo mama\'s so fat; when she stepped on the dog\'s tail we had to change his name to Beaver. 2.How are the New York Jets defense and yo mama alike? You give them a quarter and they\'ll let you score!  3.Yo mama\'s so fat, when the cops see her on a street corner they yell, \"Hey you guys, break it up!\" 4.Yo mama\'s so fat, even Richard Simmons laughs at her! 5.Yo daddy is so bald, when he wears a turtleneck he looks like a broken condom. 6.Yo mama\'s so fat she makes Godzilla look like an action figure. 7.Yo mama is so bald I can see what she\'s thinking. 8.Yo mama is so dirty she has to creep up on the bathwater. 9.Yo\' mama like a Big Mac -- full of fat and only worth a buck! 10.The only difference between your mama and a washing machine is, after you drop a load in the washing machine, it doesn\'t follow you around! 11.Yo\' mama\'s lips so big, she can whisper in her own ear. 12.Yo\' mama so dumb, she waited at a STOP sign until it said GO! 13.Yo mama\'s so big, fat and clumsy, when she tried to get to Wal-Mart, she stumbled over K-Mart and landed right on Target. 14.Your mama cooks so bad, your family prays after they eat! 15.What is the difference between yo\' mama and a winding road? The winding road has curves you can get used to! 16.Yo\' mama a saint...a St. Bernard! 17.Yo mama is so ugly that when she worked at the bakery they dipped her face in the batter to make animal cookies. 18.Yo\' mama\'s breath is so nasty, when she burps her teeth have to duck. 19.Yo\' mama\'s breath stinks so bad, people look forward to her farts! 20.Yo mama\'s so fat, she uses a semi-trailer as a couch. 21.Yo mama\'s feet are so scaly that they filmed Crocodile Dundee in her footbath. 22.Yo mama\'s so dirty, plants grow off her ass. 23.Yo mama\'s glasses are so thick that when she looks on a map, she sees people waving 24.Yo mama\'s so fat, when she farted pluto\'s ice caps melted. 25.Yo Mama\'s so ugly, she had to trick-or-treat over the phone. 26.Yo mama\'s house is so small, I threw a rock in the window and hit everyone inside. 27.Yo mama\'s house is so dusty the roaches ride around in dune buggies. 28.Yo mama is like a brick -- she is always getting laid. 29.Yo Mama\'s like the Pilsbury Doughboy. Everybody wants to poke her. 30.Yo\' mama\'s lips so big, she doesn\'t use chapstick -- she uses Mop \'n\' Glo! 31.Yo Mama\'s breath is so bad, she could clear a chat room. 32.Yo Mama is so poor, she told your little siser that Santa Claus was dead. 33.Yo\' mama so dumb, that the Psychic Friends only charge her half price to read her mind! 34.Yo\' mama got such bad dandruff, the principal declared a snow day! 35.Yo mama is so bald that she has to use rice for rollers. 36.Yo mama\'s so bald that when you rub her head you can see into the future. 37.Yo mama is so bald that when she takes a shower she gets brainwashed! 38.Yo\' mama so country, she thought an elevator was a mobile home! 39.Yo mama\'s so dirty, the U.S. Army wants to use her bath water as a biological weapon. 40.Yo mama is so disgusting that every time she bends over to tie her shoes, she\'s gotta fart! 41.Yo\' mama so dumb, she failed a pregnancy test! 42.Yo mama\'s so dumb, she thought masturbation was a karate teacher. 43.Your mama\'s so dumb, she had to have a cook help her make ramen noodles. 44.Yo mama\'s so dumb she thought Subway made trains 45.Yo mama\'s so dumb she thought Master P was a special restroom. 46.Yo mama\'s so dumb she sat on the TV and watched the couch 47.Yo mama\'s so dumb, when god was giving out brains she thought they were milkshakes and asked for extra thick. 48.Yo\' mama so dumb, she bought a solar powered flash light! 49.Yo mama\'s so dumb, she thought that the the International Dateline was a global dating service. 50.Yo mama\'s so dumb, she stuck the phone up her ass and thought she was makin\' a booty call. 51.Yo Mama\'s so dumb, she went to a movie that said \'\'under 18 not admitted,\'\' so she left to go get 17 of her friends. 52.Yo mama is so dumb that she was on her way to the airport and saw a sign that said â€œairport left.â€ So she turned around and went home. 53.Yo mama is so dumb she stuck a battery up her butt and said, \"I got energy!\" 54.Yo\' mama so dumb that if brains were dynamite she couldn\'t blow her nose! 55.Your mama is so dumb, she sold her car for gas money! 56.Your mama is so dumb when she was pulled over for drunk driving and they asked her to walked the line she said, \'\'Which one?\'\' 57.Yo mama\'s so dumb, she tried to drown her pet fish. 58.Yo mama is so dumb, it took her an hour to cook minute rice. 59.Yo\' mama so dumb, when she filled out her job application and it said â€˜sex,â€™ she wrote â€œnot lately.â€ 60.Yo mama is so dumb that she got locked in the bathroom and peed her pants. 61.Yo mama\'s so dumb, she thought menopause was a button on the VCR. 62.Yo mama is so dumb, she thought a ribbed condom was soul food. 63.Yo mama so stupid, when you were born she saw the umbilical cord and said, â€œHey it comes with cable!â€ 64.Yo\' mama so fat, the tag on her dress reads: \"Made In Hungary, Turkey, China, U.S.A., Algeria, Japan, Indonesia...\" 65.Yo\' mama so fat, she falls out of bed on both sides! 66.Yo\' mama so fat, she has to make a long distance call to talk to herself! 67.Yo\' mama so fat, her doctor said she had a flesh-eating disease and told her she had 13 years to live! 68.Yo\' mama so fat, I take a lap around her for exercise! 69.Yo\' mama so fat, when she needs to wipe her butt, she sends out a search party! 70.Yo\' mama so fat, she got more rolls than a bakery! 71.Yo\' mama so fat, when she goes to a stadium, she sits next to everybody! 72.Yo\' mama so fat, she has to sleep in shifts! 73.Yo\' mama so fat, all the restaurants have signs that say \"Max. Occupancy 240 OR Yo\' Mama.\" 74.Yo\' mama so fat, she uses transport trucks as rollerskates! 75.Yo\' mama so fat, they showed her a picture of her feet and she couldn\'t identify them! 76.Yo\' mama so fat, people exercise by doing laps \'round her! 77.Yo\' mama so fat, when fighter pilots see her they say, \"Bogie at 11, 12, 1, 2, and 3 o\'clock.\" 78.Yo\' mama so fat, when she dances, the band skips! 79.Yo mama is so fat, when you slap her leg, you can ride the waves. 80.Yo\' mama so fat, she fell in love and broke it! 81.Yo\' mama so fat, when she inspected the Sears Tower they had to rebuild it! 82.Yo\' mama so fat, she fell asleep at a new gas station and when she woke up she was painted orange with a 76 on her and was stuck at the top of a pole! 83.Yo\' mama so fat, she can stand over the homeless and give them a home! 84.Yo\' mama so fat, she\'s the reason there\'s night and day! 85.Yo\' mama so fat, she has an echo in her belly button! 86.Yo\' mama so fat, fat is a compliment! 87.Yo\' mama so fat, she has to play pool with the planets! 88.Yo\' mama so fat, she has to use a telephone pole as a tampon! 89.Yo\' mama so fat, the president declared her the 51st state! 90.Yo\' mama so fat, that in order to kiss her, your dad has to hit her in the stomach and ride the third wave in! 91.Yo\' mama so fat, she fell through a black hole and got stuck! 92.Yo\' mama so fat, they had to take aerial shots for her senior pictures! 93.Yo mama is so fat, she\'s on both sides of the family. 94.Yo\' mama so fat, if she were one of the three little pigs, she\'d say \'Not by the hair on my 26 chins!\' 95.Yo\' mama so fat, the photographer charged her for a family photo! 96.Yo\' mama so fat, when she goes to the movies, she sits next to everyone. 97.Yo\' mama so fat she has to walk on the street instead of hogging the sidewalk 98.Yo\' mama so fat she went on an airplane and turned it into a boat. 99.Yo\' mama so fat, her beeper goes off and people think she\'s backing up! 100.Yo\' mama so fat, that when she wears a yellow rain jacket, people think she\'s a taxi cab. 101.Yo\' mama so fat, when she jumps off the high dive, she shows up on radar! 102.Yo\' mama so fat, I had to take two trains to get on her good side! 103.Yo\' mama so fat, when did the splits, she gave the road a hickey 104.Yo\' mama so fat, she was alone on a corner and the cop arrested her for loitering 105.Yo\' mama so fat, that when you tell her to haul ass, she has to make 2 trips! 106.Yo\' mama so fat she has to call Sherwin-Williams to paint her toenails 107.Yo\' mama so fat, that when she sits around the house, she really sits around the house! 108.Yo\' mama so fat, you use her thong as a hammock! 109.Yo mama\'s so fat, when she ran away they had to use all four sides of the milk carton 110.Yo\' mama\'s so fat, she broke the stairway to heaven! 111.Yo mama\'s so fat that she went into 7-11 and didn\'t come out until 11:07 112.Yo mama is so fat, the last time she saw 90210 was on the bathroom scale 113.Yo mama\'s so fat a parachute wouldn\'t cover her. 114.Yo mama\'s so fat, she got busted in the airport for having 200 pounds of crack. 115.Yo\' mama is so fat, when she joined the army and the soldier said head for the trenches, they all jumped in her butt crack 116.Yo mama is so fat that she keeps a pack of bacon for dental floss. 117.Yo Mama is so fat that she got baptized at Sea World! 118.Yo mama\'s so fat, first she filled the bathtub, then she turned on the water. 119.Yo Mama\'s so fat, when she goes to get out of the bath you have to call the Coast Guard. 120.Yo\' mama so fat, when she goes to the beach, the tide comes in! 121.Yo mama\'s so fat, she goes to the beach to sell shade 122.Yo\' mama so fat, she goes to the beach and she\'s the only one who gets a tan! 123.Yo mama\'s so fat, she has to put her belt on with a boomerang 124.Yo mama\'s so fat that when she bends over, we go into daylight savings time 125.Yo mama is so fat that when she took a vacation to New York City, she wore a black dress and everyone thought there was a power outage 126.Yo mama\'s so fat, her blood type is Ragu 127.The difference between yo mama and a blue whale? About ten lbs 128.Yo\' mama so fat, when she was born her mom said, \'\'Great, triplets.â€ 129.Yo\' mama so fat, she broke the family tree 130.Yo mama is so fat that when I took her to a buffet, she pulled up a chair 131.Yo mama is so fat that, when she was bungee jumping, she took the whole bridge with her 132.Your mama is so fat, a bus drove by and she said, \'\'Stop that Twinkie!\'\' 133.Yo Mama is so fat, her ass has it\'s own congressman 134.Your mama is so fat, people use her butt cheeks as a ski slope. 135.Yo\' mama so fat, when she puts on BVD\'s it spells out boulevard  '),(18602,3376420,'18,May,2004','2004-05-18 00:00:00','   he\'s guatemalan, a little bit on the short side, but he\'s beautiful inside and out.  i met him at walmart, weird place to meet someon right? i was with my friend (now ex-friend)  keyona at the time shopping with her mom for bras.  i saw him with his friend checking out a shirt, and i knew i liked his style at that moment.  he was lovely, divine... god i loved that secound.  i smiled at him when he passed the sunglasses stand... he smiled back, and came to me.  he asked me my name, and i told him, then he told me his... we exchanged phone #s and it was over.  that night he called me and we talked.  he had a gorgeous voice, which i was instantly in love with, and he told me he wanted to go to the movies with me (cinemax was actually the word he used, since his english was really limited, but i understood after a few minutes)  this was about 7 months ago and now i\'d die for him.  doesnt matter how physically attractive someone is... its the love  the love that i feel at this moment  '),(18603,3376420,'18,May,2004','2004-05-18 00:00:00','   k, so ill say something to all you ppl who think your life sux: its NOTHING compared to mine, ok so maybe im exaggerating a bit, but my life is pretty gay. read on:  i live with a paranoid bitch, my mother.  she believes that im plotting against her no matter what i do.  if im nice, she freaks, if im bad, she screams... *sighs*  i go to Classical high.  basically the worst mother fuckin school you can imagine.  the kids are afraid of me mostly.  they spread rumors, and call me, \"goth\"  i hate labeling so you can imagine how much that pisses me off  i have a bf, well more like a ball-and-chain. his names julio, doesnt matter what happens he\'ll always say im cheating on him.    i have the worst little bratty bitch-of-a-sister in the universe.  she snitches, she kicks, and oh yeah... she bites.  well, i got a load off my mind now... think ill play online a little  '),(18604,3376420,'18,May,2004','2004-05-18 00:00:00','   w00t!!! so now i have a blog... crazy man, just fuckin crazy... im not really sure what to say, so i think ill stop talking, er typing now.  '),(18605,3308014,'12,May,2004','2004-05-12 00:00:00','    I hope the lackluster buildup doesn\'t cause this post to be a disappointment.  Many years ago the 3 surgical hospitals/campuses in my city were overhauled by some team of somebodies.  The somebodies decided there was no need for services to be duplicated at sites as this caused a need for 2 (or 3) sets of staff and equipment to be supplied.  They came up with a game plan of which site would absorb some services, and give up other services, and then they came up with a time line (milestones) of when all the changes would take place.  Since then, building and renovation have been taking place, and they are finally beginning to switch around services.  My hospital is going to become an ambulatory centre, focusing on cataract/eye surgery, urology surgery, and hand and upper limb surgery, with small contingents of general, gyn, ent and ortho thrown in for good measure.  When the teams/powers that be looked at what was transferring, they took into account the amount of work that was being sent, the number of people hours it would take to make that work amount function, and then they broke it down into specific job lines (ie. 8 full time RNs, 2 part time RNs, 1 Attendant, etc).  When a program transfers, those job lines have the \"opportunity\" to transfer with the program, since many nurses and other health care workers have specialized in that program over the years, and may not wish to stay in a hospital which will no longer be providing care in their area of expertise.  In November I was told that an attendant with the ophthalmology program wanted to transfer to my site, and he had more seniority than I.  Despite working in the OR for 17 years, I was the least senior in my department, so I was becoming \"redundant\".  They outlined all my options, but I wasn\'t given official notice because the transfer date had been pushed back.  According to Union contract, they had to give me 4 months notice (and only 4 months) and they were not going to give me one day more.  I looked at my options.  1. Accept a buy out package, return to school, and get reimbursed up to $5000 for any tuition paid out in the first year.  This left me with a lump sum (which would prevent me from collecting unemployment until it ran out), and the knowledge that unemployment would only let me attend school for 1 year, if it lead to a job within one year.  Therefore, university was out, and any college program more than 1 year long was out.  2. Exercise my bumping rights and move to the job of the least senior full time person in my job category.  I spoke to several people familiar with work in the unit that person was in, and knew I would hate switching to that job.  3. Apply for any job postings which came up in the meantime, and get one of those.  I went to a few unemployment workshops, to learn about the aide available from them.  I spoke to college and university counselors, and I seriously considered becoming a pharmacy technician, while attending a private medical school and retaining part-time employment, therefore keeping my seniority.  I would have to pay for the schooling myself, but the job would be in my current union, so I would have 17 years seniority in applying for any pharm tech jobs in the hospital, and therefore job security.  I went to the private school, and they had me take a multiple choice aptitude test before the interview.  The interviewer then did his job too well.  He told me I had the best score he had ever seen, and he raved over my transcript marks.  Shortly thereafter I was conversing with a co-worker who was just about finished her BA of Ed, after returning to school in her early 40\'s.  When I was younger I wanted to be a teacher, but I got sidetracked.  My conversation with her, my circumstances, and the interviewers remarks all led me to decide I was going to go to university and become a teacher.  Surprisingly, my husband agreed.  He was more supportive of my longer and more expensive schooling than he was of my short-term pharm tech program.  I applied for a part-time job in the Central Processing department, and 1 week before the deadline the hospital had in which to give me notice, they called and said I was the successful applicant.  I accepted, and so my other options never came to pass.  Now I am set on a very exciting course, but I\'m very frustrated that the university has not yet accepted me!  Grrrrrrrr  And that was my restructuring story  Cheers Jay-A   '),(18606,3308014,'11,May,2004','2004-05-11 00:00:00','    You\'re still not getting the \"restructuring\" story, because I need to vent.  I applied for University in March.  I can check the website with my own \"special and private\" id to see the status of my application.  The site says they received it at the end of March, and that it should take 2-4 weeks before they offer me admission (or say neener neener, NO!)  I just spent over 20 minutes waiting for \"the next available operator\", and He Said:  \"We received it in the beginning of April, it was processed by April 14th, in 2 days it will be 4 weeks\" (No, in 2 days it will be 1 month, today is 4 weeks, kid), and \"it\'s a really busy time of year.\"  They may have it processed within another 2 weeks.  GRRRRRRRRR  I changed my life for this opportunity, and the University counseller that I went to see (twice) told me  \"Historically, we have never turned down a qualified mature applicant for part-time studies, and Jay-A\",  , \"you are a qualified mature applicant for part-time studies.\"  So basically, she gave me the best assurance she could without getting her butt into a legal bind if by some chance I didn\'t make it.  I know that MOST LIKELY I\'ll be accepted, but I don\'t like most likely, I want I AM ACCEPTED.  Grrrrrrrr, Arrrrrrrrg  Out of all the choices I was given regarding my restructuring out of my job, I took the reins of control and MADE this choice an option, and threw all my other options out the window when I did so.  If I don\'t get accepted to University  a)  I\'ll be hugely embarressed, because everyone thinks I\'m going  (over 100 nurses, surgeons and anaesthetists in my current department, plus other hospital employees, my husband\'s family, my family, etc, etc)  b)  I\'m stuck taking the part-time job I begin at the end of this month.  Sure I get full-time training for the summer, and sure they are short staffed and I can probably work full-time hours even past that, but I don\'t care!  I want to be a student.  Grrrrrrrrr, Arrrrrrrrrg, Growl!  I\'m still not completely vented, but I don\'t know what else to say, and it is time to head out of town to my second session of the course I need to take for my new job.  sigh  Jay-A   '),(18607,3308014,'11,May,2004','2004-05-11 00:00:00','    It\'s evening, the kids are in bed, the hubby is (finally) on his way home from work, and I was clicking links in my profile to see who is out there in Blog Land.  On one lady\'s blog (she wasn\'t memorable) I found JEREMY  http://www.whoisthatwithjeremy.com/  At the moment, only one page of photos is accessible, but I\'ve emailed Jeremy\'s dad about the problem, so I\'m sure that will be fixed soon.  If you are a parent, and you like silly harmless things that take a bit of imagination, insanity, and hutzpah (dunno if I spelled that right), then you will like the site that Jeremy had his dad create.  I\'m off to do stuff, but I\'m sure I\'ll get to that \"restructuring/career change\" story that I promised ... tomorrow.  Not that anyone is actually reading this.  Perhaps I\'ll tell my friends and family about it, but I think not.  They hear me yap all the time, so why inflict them with the pressure to read my less than stellar thoughts?  Cheers! Jay-A   '),(18608,3308014,'11,May,2004','2004-05-11 00:00:00','    I don\'t have much time for this initial post, but I feel a need to write something in order to complete the creation process.  I\'m thrilled about heading to University in the fall, and in order to implement this opportunity, I have applied for and taken a new job within the hospital that I work at, and I begin this new job on the last day of May.  I\'ve spent the last 17 years working as an OR Attendant (muscular nurse\'s aide with only on the job training - no courses), and I shall now be seeing things a bit differently as I wash, package and sterilize the surgical and respiratory equipment used throughout the hospital.  Fortunately, the OR and the Central Processing departments work very closely together, so I know the majority of my new co-workers, and I\'ll still see my old co-workers daily.  My next post shall concern the reasons (city-wide hospital restructuring) that have led to my life-altering choices.  Cheers! Jay-A   '),(18609,3878084,'31,July,2004','2004-07-31 00:00:00','    Hey today was fun. I went to the Rapidz with Toniann, Eileen, Kimmy, Andrew, and Jessika. The only bad part is that we ran into juss gues???Â¿Â¿Â¿ jordan OG. Well the day went like this.   Andrew, Jessika, Kimmy got dropped off at 9 by kimmys mom. Then we went and picked up Eileen, and Toniann in Nautica. Then we all went to the rapidz. We got their around 10:30 and we went on all the slides. In the lazy river we all wrestled. LOl andrew kept getting hurt by me toniann and eileen. Yeh then we came home around 4 and swam in my pool for like and hour. Then kimmy and jessika came over. And andrew went home. Then we all played basketball. Then every1 had to go.  But i stayed home and went to run at about 6:30. Then came home and got ready to go to Red Lobster and then we came home around 10:30. Then i got ready for bed and was pissed off at jordan and went to sleep. And now jordan is lying... Long story tell ya later. Peace Out Christian   '),(18610,3878084,'21,July,2004','2004-07-21 00:00:00','    Hey today wasnt so good of a day.&nbsp; First i woke up as usual rite at 8.&nbsp; Same as alwayz i felt like crap.&nbsp; I was still pissed at andrew from the nite bfor.&nbsp; Then Andrew got me even more pissed after he told Eileen who has been cuming longer than him to go to a different room when it was only his 3rd day.&nbsp; So I ignored him for the rest of the day.&nbsp; Then i got picked up at 4 came home and their was this huge thunderstorm rite over my house.&nbsp; I called Jordan to cum get me which he did.&nbsp; Then we went bac to his house and did absoloutly nuthin. Then finally we went to the mall IT WAS SO BORING!!!&nbsp; We didnt do anything!&nbsp; Well Jordan got new school shoes.&nbsp; And in FootAction i got a livestrong bracelet from Lance Armstrong the Tour De France person which itz a dollar and it all goes to the Youth Cancer Society.&nbsp;&nbsp; Then i came home had a Manwhich.&nbsp;Then rode my scooter a lil bit.&nbsp; Came in same as usual dad alwayz in a crappy mood MAN I AM SO SICK OF HIM!!!&nbsp; Then to make it even better Im pissed at my mom to cuz she wont let me go to lunch tomorrow with Toniann and Eileen tomorrow but i have to go eat stupid Thai with my mom and her friendz.&nbsp; And Shes UHHH im SO SICK of BOTH OF THEM.&nbsp; If any1 needz a new family member give me a call.&nbsp; Then i watched Newlywedz and the Ashlee Simpson show.&nbsp; To finish off my LOVELY day I spilt the Fish bata food EVERYWHERE!!! It took me 20 minutes to clean up everything.&nbsp; Then I went to the computer to type this.&nbsp; Im talking to Ashlea Jackson online.&nbsp; G2G Sick of life.  &nbsp;&nbsp;&nbsp; PeacE OuT***Â¿Â¿Â¿ &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; christian   '),(18611,3878084,'20,July,2004','2004-07-20 00:00:00','    I woke up at 8 like i alwayz do.  Went to jcc til 3:30.  Once again it was fun hanging out with toni and peepz.  We ate at renzoz\'.  I hate leounard shez a female dog SOOO Much.  One of these dayz i am going to tell her off so bad.  MG andrew is suppose to cum tomorrow that sux.  I hung out with jamal and rode my scooter which i havent rode in ForEver.  After that while jamal ditched me for travis wedgle i went over to andrewz and had a boring time.  Then i came home and chilled and talked online.  I am so pissed at Andrew.  talk to ya later     peace out   christian   '),(18612,3878084,'18,July,2004','2004-07-18 00:00:00','    Hey itz me after a while.&nbsp; I didnt go to church today which thatz a shock cuz normally i go every sunday.&nbsp; Well my momz been gone all weekend at sum woman of faith conferance then sawgrass mills.&nbsp;&nbsp;She got home today.&nbsp; My dadz been pissed cuz no matter wat he wont go grocery shopping so hez mad that my mom didnt go grocery shopping rite b4&nbsp; he left.&nbsp; IDk stupid i guess.&nbsp; Well i went over to jordanz was gonna cross the pole but decided not to.&nbsp; So ed came and picked me up at the end of hypoluxo rd.&nbsp;&nbsp;Then we went to his house hung out for a while.&nbsp; Then got in the car went to quizznos subs then to justinz house then to sportz authority and after that the italian bakery and then to giannis the salon that his aunt owns.&nbsp; Me and jordan got hilites or blonde put in the front of our hair.&nbsp; Then we went bac to jordanz hung out for a while.&nbsp; Then went over to his grandmas at nite and ate and went on the internet and stuff.&nbsp; Got home around 9.&nbsp; After that all ive done is talked online and stuff.&nbsp; Talk later &nbsp;&nbsp; Peace &nbsp; Christian   '),(18613,3878084,'16,July,2004','2004-07-16 00:00:00','    Hey i havent put anything in here lately.&nbsp; Ive been going to Jcc everyday.&nbsp; I went to lunch with toniann eileen jessika and kimmy.&nbsp; It was tite.&nbsp; I was suppose to go to the Rapids with alot of peeps but Stetson was being a beotch so i didnt want to go. Talk later &nbsp; peace  &nbsp;christian   '),(18614,3878084,'12,July,2004','2004-07-12 00:00:00','    I woke up at 8. With a sore throat.  But i still went to jcc. My throat got better.  The day was pretty boring.   I dont knoe where i am going with the jcc.  Then my mom picked me up at 12:45 then i went to Dr.Darling.  He did X-rays and moldings and stuff.  I ran into Robert Aron at the orthodontist.  Then i came home and went over to Jaime and Jakes.  Came home then went to go get Spiderman 2 (the titest game ever.)  Been playing it forever and been talking to toniann jessicka and sum other peeps online. Im suppose to go eat lunch with Kimmy,Toniann,and Jessika tomorrow after Jcc since they all go there.   Peace out,     chrisb   '),(18615,3878084,'11,July,2004','2004-07-11 00:00:00','    Got up at 9 cuz we had to go to trinity EARLY serive.  It ended at round 11. Then we came home and set up the house for Chris and Joesaphine our family friends.  Then of course Jordan came and picked me up and we went and chilled at his house.  Thats where I was the rest of the day.  He had alot of family over.  1 cool thing is I got 2 new cds.  Yellowcard Ocean Avenue, and Linkin Park Meteora.  Got home from Jordans around 9.  Basically after that all Ive been doing is talking online and listening to my new cds.  Got jcc tomorrow and have to go to the fricken orthodontist for all u peeps who want to knoe dr.darling.   '),(18616,3878084,'10,July,2004','2004-07-10 00:00:00','    In case u didnt knoe my cell # is                312-9014    give me a call anytime.          chrisb   '),(18617,3878084,'09,July,2004','2004-07-09 00:00:00','    I woke up at 11:30 today which is early for me besides when i go to jcc.  I been cleaning my room for awile.  My room was like bing in he|| b4 i cleaned it up.  Then i went and jogged to the recreation center and bac.  I had the worst cramp.  Then i came home and cleaned my room summore.  I mite go to Palm Springs Pizza with Jordan later.  Guess wat i ended up not going with Jordan.  He forgot to call me.   But i ate steak it was good.  I thought that u would like to knoe.  After that i hung out with Jake and we helped my mom hang bird feeders and windchimes FUN,FUN,FUN!!!  After that Allie called.  Jake went home around 9.  Tomorrow im going to church. If u go to Trinity Church look 4 me.  Itz weird I havent talked to many peeps today. Ok Jordan called me at 9.  Hes with his cuzins Cameron and sum other kid.  Now i have been talking online for hourz to toni and andrew and stetson and eileen.  Doesnt sound like fun.  Peace Out    chrisb        nib=Nothing Im Bored   '),(18618,3878084,'09,July,2004','2004-07-09 00:00:00','    I HATE U!!!!   '),(18619,3878084,'09,July,2004','2004-07-09 00:00:00','    Andrew one of my best friends. For sum reason peple call us twins have no clue why. He has red hair and is 23hours and 54minutes older than me. We ran into a weird ladie at hallmark last night juss letting u knoe.&nbsp; We hang out alot.   '),(18620,3878084,'08,July,2004','2004-07-08 00:00:00','    Hey.  2day was not my average day.  U will probaly think i have mental issues or sumthing bcuz i woke up at 8:00 on my own free will.  U wanna knoe why keep reading if u do.  The reason is, bcuz Andrew told me about wat he was doing so i went to the jcc to b a chaperone for the lil kidz.  The thing is these kids R 2,3,4 years old.  So u can already guess that they r annoying and i thought that it was probaly goin to b He||.  But it wasnt. But the thing is when i left alot of them started to cry so i didnt knoe wat to do so i told that i would b bac on monday so they all quit crying thats why i am gonna do it for the rest of da summer. Plus their not that bad i knoe wat am i saying, and no i did not convert over to jewish to go there.  I juss had to tell u that.  Ceasar got a bath today.  Like u care. Well we went to Harry potter 3 then we went to eat at Brews Room.  Then we went to walk on the beach which i dont knoe why.  Last I am talking online later.                 Peace ChrisB   '),(18621,3878084,'08,July,2004','2004-07-08 00:00:00','    Hey its Christian.  This is pretty tite. This website thing.       Me Andrew and his big bro Bobby went to da mall 2day.   '),(18622,3878084,'08,August,2004','2004-08-08 00:00:00','    I                Toniz da beSt.         christian Oh yeh today was ok.  we went to church.  and me and johnny walked around the church the whole time.  Then our parentz got mad at us.  But it was fun.  After church  we went to the mall.  Then i went over to the melearz house til 7.  Then i came home and ate pizza and went runnin and to jamalz unclez house.  Now im here talking on the computer.   oh yeh i found sum guyz wallet @ the mall and he was 17 and he had like 5 condoms in his wallet.  I was like he must stay busy. Lol.                           Peace Out                            christian   '),(18623,4072259,'30,July,2004','2004-07-30 00:00:00','   Why does it seem like a good idea when you are at the bar with your friends to keep saying \"yes\" to \"do you want another beer?\" when you know that you have to be&nbsp;at work at 8am and if you stay out late 8 am comes even quicker than it normally does. Didn\'t they ingrain it into us as children to \"Just Say No\"? Nancy Reagen would NOT be proud of me. I did not D.A.R.E. to keep myself off drugs last night.   INSANITY: preforming the same actions over and over and expecting different results.  yup that is my life. And the bar wasn\'t even worth it because this person that&nbsp;I have been doing a proverbial dance with for the past year decided it would be a good idea to try to come talk to me after standing me up the last time we were supposed to have a big serious talk about our relationship, or lack there of one, not to mention it was said persons idea to have this talk and I was the only one that showed up for it. yeah whatever. I was going to comment on Kerry\'s speech last night but I feel I need some more time to collect my thoughts, so for now Salute and just say NO!!  '),(18624,4072259,'29,July,2004','2004-07-29 00:00:00','   Sooo I am just now jumping on this whole blogging band wagon. It seems pretty cool and seeing as though I am perpetually bored at work, hopefully this will prove to&nbsp;be a source of&nbsp;entertainment to pass the hours untill I am bored and not at work. &nbsp;I just recently graduated from college and I find myself in a job that has absolutley NOTHING to do with my degree (which I am sure I am the 1st person this has ever happened to) but the thing is that there is not even an opportunity to get into my field.&nbsp; All my plans got messed up. I was supposed to be going to law school in the fall but I didn\'t get in, it was really sad there were 250 seats for the one school I applied to and there were 1900 people applying for those seats (damn economy), yeah I wasn\'t one of them, so now I gotta wait another year and try to be a productive, contributing member of society&nbsp;and I failed to realize what I was getting myself into when I signed up for that label.&nbsp; And now here I am working in an office doing accounts payable, which is a far cry from my degree (political science with a public service specialization). But ya know whatever I guess.&nbsp;and so concludes my first blog, I hope it was as good for you as it was for me...right.  '),(18625,4072259,'12,August,2004','2004-08-12 00:00:00','   Do you ever just want to like go somewhere that is not where you are and at really desperate times could be anywhere and re-create yourself? Not so much re-create yourself I guess but re-create your life. New friends, new social circles, whatever. I\'ve done it once but it was a long time ago and as much as it was a disaster (I came back home in 6 months; the grass was way greener on the other side) I think it has been so long that it is starting to sound like a good idea again. I have this problem with staying in one place for too long. I believe that this is my mother\'s fault. (I don\'t mean that in the sense of I blame my mother for how my life turned out and I don\'t mean it negatively, I like the fact that I am so restless, I think it makes me strive for more so that I can move on) Anyways, when I was little we used to move all the time and I don\'t remember it ever being sad I was always excited at the possibilitiesI guess. I got really good (and still am) at adjusting quickly but I guess it had some negetive effects in that I am extremely restless. And  Icould leave a place with out looking back pretty easy. I mean don\'t get me wrong there are people that I would still talk to. I think I missed my calling and I should have joined the circus or worked for some traveling festival. I have no musical talent so maybe I could be a roadie. Do they let girls do stuff like that? hmmm.       It\'s really funny the closer you get to people the more they think they own you, like by them knowing you better than most gives them some claim on you, not so much posession in the psychotic way but more in like a jealous way. And then you do something and some how your action that had nothing to do with them turns into a personal attack on them and it is all completely your fault? yeah I guess this makes no sense from the outside and it is completely spurratic but it is working for me so i\'m going with it. Why is that that we think we can claim possession over people? that\'s a really strange concept. Like is it human nature ? Or are we programmed to \"obtain\" things and so it just floods over to our personal lives along with our material objects. Like it\'s a little different my husband/my wife whatever but like just your friends even. \"Oh so and so is MY best friend\". \"Oh so and so is MY really good friend\".  What about \"WE are really good friends\". \"WE are best friends\". that\'s rather interesting to me i suppose.       Human nature, what a bitch.       So stay stuck in this town for another year and then move somewhere else and start the process all over again...I often wonder when I get the bug to just up and go if I will be like this for the rest of my life. that kinda scares me. Maybe I need to be challenged to plant my roots. but that\'s way too permanent.    I\'m commitment phobic and I am proud!  '),(18626,4072259,'09,August,2004','2004-08-09 00:00:00','   You have no Idea how I wish it was Sundy. I really think the weekends wear me out more than the week days. Saturday night I attempted to watch a movie at my friends house and I became painfully aware of the fact that I have ADD. I cannot sit and watch a movie to save my life. I can however look at everything in the room except for the TV,  go to the bathroom 600 times, smoke like 4 cigarettes, play with matches, get everyone a beer when they need it, light candles, and play solitare. I apologized to everyone after the movie for being such a distraction. I can only really watch movies if I am exausted, stoned, or in the confines of a freezing over priced movie theater where it is dark, you are stuck in an uncomfortable seat, and there are other people there who are not your friends and not so forgiving when you do everything but watch the movie.  At least do to my aprox. 6.7 second attention span I can do many projects at once and do them well. I just have to have like 6 things going on all simultaneously so that there is no chance of me getting bored. It\'s funny at work I will start a project do half of it start another project, do half and keep doing this until I am bored and start back at the beginning project. I started studying for the LSAT again and I think this time around it is harder because I am kinda mad that I have decided to take it again and therefore have an aversion to studying for it. After all the stress of taking it the first time my consolation prize was that I would not have to ever do that again...and here I am getting ready to go through all that stress agian. I can\'t wait for the bar (NOT looking forward to that test) . I have researched the test and it drives me crazy that this \"aptitude\" test is only like aprox. 50% accurate in determining whether you will be successful in law school, they just continue to use it out of tradition or sadistic torture. My friend asked me the other day why I wanted to be a lawyer and I quickly answered because I am a massochist. anyone who has gone through the long hard process of either professional school or graduate school knows exactly what I am talking about. I just keep reminding myself that it is worth it. And yeah I do know it is. so I will just keep studying it. blah back to work.  '),(18627,4072259,'04,August,2004','2004-08-04 00:00:00','   Me and a friend of mine were having a conversation the other day about drinking and motives and such and I came up with a conclusion about it. I think that \"smart\" people drink to become dumb and \"dumb\" people drink to appear smart. Now not that I consider myself to be exceptionally intelligent but my mind is always going, always thinking, always contemplating and when I drink it shuts it up for a short while. And I have noticed that some people when they drink attempt to exert their \"mental dominance\" upon other people. It\'s quite interesting to go to the bar and just observe (well you know drink too, but observe in the process). It\'s quite amazing to me this world we have here. We have enough resources so that every person on this earth could live comfortably, but we are obsessed with accumulating as much wealth and/or resources as humanly possible and then sitting on it so that it benefits no one but ourselves. For as civilized as we are supposed to be (which when looking at technical and medical advances we really are) we are still barbarians. We trample other countries (not specifically the US, any 1st world nation), steal or take or declare their resources as our own, and give them nothing in return. We are like little children at recess and then the littler (read weaker) children either make friends (read become the lackeys of) with the bigger stronger children or they mock the behavior on children smaller than themselves. For as civilized as we are we completely lack compassion. We try to keep two people who love each other from becoming \"married\" because it will ruin the sanctity of marriage, tell me where the sanctity is when people can get married for 18 hours/be married 5+ times/ commit adultry/beat their spouse etc. thats sacred and upkeeps the sanctity of the union? And how many people don\'t see the absurdity of that argument? \"God\" said marriage is between a man and a woman well \"God\" also said a whole lot of other crap that no one pays attention to. I am striaght and am completely against the idea of marriage because I think it\'s a joke and means no more of a commitment to me than just being in a relationship with somebody.  A woman can be threatened with laws that will directly affect her body and life because some religious right wingers want to make personal decisions political?? YOU ARE A MAN YOU HAVE NO SAY!! when you get pregnant then we\'ll talk!! If you are that against abortions then DO NOT HAVE ONE!!! But do not decide that your choice or your values should be imposed on every other human being just becuase you THINK you are right. We come up with wonderful life changing medical advances (e.g. drugs, precedures etc.) that will save lives or improve the quality of some peoples lives etc. and then we allow pharmacutical companies and insurance companies to charge outrageous prices so that only like the top 10% of the population can afford them (now I am exaggerating slightly but you get the jist)? we are making money off of disease and diseases that are occuring at record numbers because we pollute and have no respect for the environment. \"Criminalize the symptons while we spread the disease\" thank you Ani D. UGGGHHHHH!!!!!!!!!  yeah that\'s been building up for a while. perhaps I should stop talking now and that would be a wise idea. See there\'s my point about my head always going. and that my friends is why I drink  '),(18628,4072259,'02,August,2004','2004-08-02 00:00:00','   So I think that I am convinced that there is some strange time thing that exists and it makes Saturday and Sunday go by in double time. I swear it was just Friday but here I am monday morning back at work. Oh well. It just almost became worth it though because someone just came and dropped us off bagels and although I am much more an english muffin type girl, Einstein Brothers has the best cranberry bagels.    I had this small epiphany this weekend. I live in very very Northeastern Ohio, and  I had to attend this wedding is very very Southern Ohio.  And I now know why most people think of backwoods, farms and corn when they think of Ohio. Wow there was like nothing and the wedding reception was at this \"Country Club\" yeah it was like a barn, and at the reception there were actually people wearing jeans and cammo and little boys taking their shirts off and running around. it was only like a few hours away from where I live but wow it was totally like another world. Amazing to me. I am so a city girl, like even Cleveland isn\'t really big enough for me but did I appreciate it soo much more after this weekend. I am also estatic because they are announcing today at 1:00pm where, and when Farm Aid will take place and also who is playing. If you don\'t know what Farm Aid is, it is a concert that is put on every year and all the proceeds go to independent farmers throughout the country. Willie Nelson, Dave Matthews, Neil Young and John Mellencamp are all like \"board of directors\" or whatever and orchestrate the whol;e thing. Willie nelson started it I think back in like 1989. It is always a great time and is for a great cause. Ok well there is my randon tangent filled blog for Monday.  '),(18629,3146933,'14,April,2004','2004-04-14 00:00:00','   Well, I decided to try this thing instead of Diaryland. Why? Because I\'m bored. I wish I could see Zach right now! Yeah, ZachWillis and I have been going out for a couple days now..he\'s great! He came over last night and watched a couple movies with me because I\'m still grounded. Today makes it exactly a month since I got grounded :/ It sucks, but he understands. Monica\'s in New York right now :( I\'m  supposed  to be there!! and yesterday omg she got to see  Good Charlotte ! Oh I love them! I was going to go to Zach\'s game today but I wasn\'t ready in time so I decided not to go late. :( So far..this Spring Break has sucked  so  bad! To tell the truth, I\'m about ready to go back to school! Pretty sad huh? Anyways, so the insurance ppl call about my car and they just decide my car is totaled after all...which really makes no sence..b/c it wasn\'t before..! lol whatever but I\'m gettin a new car soon I think..not sure what it\'s going to be...but hey I just want something to drive b/c I\'m goin  CRAZY  not having a car! Stuck at home..bumming rides off ppl when I can get them...But I think I\'m about to wait on my guy to call me and see if we can do something today... I '),(18630,3146933,'30,May,2004','2004-05-30 00:00:00','   Went to watch \"the day after 2moro\" w/ monica last nite and it was pretty good. it\'s scary but pretty good. my birthday\'s 2moro. i haven\'t talked to matt since graduation night. i didnt get to talk to him in person but since i didnt then i called him and told him i\'d give him his graduation money later if he wanted to come by and get it. so i guess i\'ll be seeing him again sometime soon. i hope. i want to. it doesnt really matter though i guess. i love him but gah this is drivin me crazy. im not sure what to do. omg and lemme tell you. *some1* was msgin me last nite and it was CRACKIN me strait up. \"i\'m seein little green goblins. is that healthy?\" whoa that scared me a little but after a while he jus started crackin me up so bad i didnt care. he was outta it really. but anyways...i dont think i need to say anything about that. he knows who he is..and dont you try flying off 2 story houses either!! ...if he reads this. haha. Well, I\'m going to go watch a movie. My bday\'s 2moro and I better have something to do. If I don\'t then I\'ll take u up on that offer Brad..if I can anyways. movie. that\'s where I was headed. . . . . .I\'ll write 2moro maybe. :) *Please God, let 2moro be flat out GREAT!!!*  '),(18631,3146933,'26,May,2004','2004-05-26 00:00:00','   hmm...what\'s happened? A LOT. Lemme tell you. I was prayin for who knows how long that Matt would hug me memory day and after that, I\'d get to hang out with him at least. It happened. I was so scared to go up to him at the memory day thing and so I was like right behind him and I looked at Adrienne and I was like \"i can\'t\" and she tapped him on the back and pointed at me! lmao! omg! and he hugged me...i cried so much that day. I started cryin  SO  bad when Matt\'s pictures came up on the screen..that was the first time I cried...then I finally got calmed down after a few minutes and then Monica\'s pictures came up and I didn\'t expect to see my pawpaw but she had a pic of her and him on there and omg i broke down again..ugh so sad. then a pic of me and her when we were little :D we were so cute. and holdin hands in our lil dresses. but anyways, ok..yeah it was terrible. It was too sad..I wouldn\'t go thru it again. But at least Matt hugged me, it made my day. Anyways so I get home and I talk to Matt and we\'re not doin anything so we do somethin together. It\'s like we were never even apart...well ok kinda. We talked a lot. He\'s crazy. I love him with all of my heart, still. I\'m out of school. Today was my last day. English exam? freakin hard. Math exam? freakin hard. I hate exams. But anyways, so I\'m a senior now. One more year. It\'s sure gonna fly. My birthday\'s Monday *excited* I hope I see *matt*  '),(18632,3146933,'21,May,2004','2004-05-21 00:00:00','   \"I\'m twisted \'cause one side of me is tellin\' me that I need to move on On the other side I wanna break down and cry\" \"Sendin\' pages I ain\'t supposed to Got somebody here but I want you \'Cause the feelin ain\'t the same Find myself callin\' *him* your name\" \"why are we Strangers when Our love is strong Why carry on without me?\" \"And everytime I see you in my dreams I see your face, it\'s haunting me I guess I need you baby\" I have *2* days left of school. Memory day\'s comin up. Gah I\'m so scared...I\'m debating on whether or not I\'m even going to go to it. I almost cry thinking about it. Ok and if I haven\'t run into Matt...like 500 times a freakin day...here lately anyways. It makes me SO mad. I see him...every freakin where I turn. But I\'m scared about givin him his graduation gift. Man, I don\'t know what to think...I swear. I have *someone* tellin me one thing and other people telling me different. I have mixed feelings right now. You know though..I want Matt to talk to me memory day...because...he spent 1/2 of his time in high school with me. I don\'t know what I\'m going to do if he doesn\'t...I won\'t be able to take it. I\'ll try though..talkin to him. I guess I dont want to get back with him...since he\'ll be goin to Alabama and I\'d never see him anyways, but I really wouldn\'t mind being friends with him. I\'ll always have feelings for him though. any guy i\'ve talked with/dated/gone out with since me and him broke up..they haven\'t made me as happy as what matt did. Oh well I don\'t feel like talking about him. Memory day. Yes...OK when my class, the juniors get up to take our place in the senior section..dear God, I\'m going to bawl. I remember wishing every day away and wishing I was out of school and that I was a senior...well...it went by too fast. It\'s here now. My last year getting things \"handed\" to me. It\'s a scary thought.. Happy 18th Birthday Monica!  '),(18633,3146933,'16,May,2004','2004-05-16 00:00:00','   Well, it\'s Sunday. I think I start off every post I do on Sunday\'s like that though. I got my car! My new celica...Friday. It\'s just like my last one...good thing b/c I missed my old one :( ! But yeah and it feels right b/c I was drivin a jeep and i felt so high up and now im back down on the road like i should be. :D! The most AMAZING thing happened to me the other day. I talked to Matt. No, I mean conversation. He talked to me too. Like...we\'re friends. Do you have...ANY idea how great that felt? It felt great. I miss him. I got a graduation invitation from him in the mail and..I cried for an hour after i got it :/ . B/c I was already missin him before i got it and then after i got it, i was just..so confused..? It\'s weird. But I wanna do somethin for him for graduation. I hope he\'ll let me. Or at least get him something. I don\'t want him to move off without seeing me at least one more time..you know? I\'ll always love him! I know he\'s got to always love me too... Anyways, but yeah and these other guys that I was thinkin about \'talkin\' with or whatever..no..they either try too hard to get me to like them, or they\'re immature, or they\'re boring, or ... there\'s just always something. I just don\'t want them I want Matt. Ok. If I do not sound pathetic. I know I do, and I\'m sure you\'re thinkin the same exact thing...but I really don\'t care what anyone thinks. I can\'t help it! for some reason i can\'t let go of him, even after what happened when we broke up. but I\'ll try... I have 7 or 8 days left of school. then I\'ll be a senior...last year...it kinda makes me wanna shed a tear. but what they hey! Time sure goes by fast..  '),(18634,3146933,'08,May,2004','2004-05-08 00:00:00','   Well, it\'s Saturday. I\'m sick..I have been since yesterday morning! It sucks. I\'m coughin my head off and sore throat. Yeah..it\'s suckin. But I\'m really excited. My dad got me another celica. same as my other as far as I know. I think I have to get a job though. :( Not that it bothers me too bad...considering I\'m not ever doing anything anyways. Ok let\'s see...we have 13 days left of school I think. GOOD THING! I miss Matt. Yes, Matt Harding. You know? It\'s weird. Everyone was telling me \"hey you need to date other people, you dont need anything serious while in high school\"...well i don\'t care. I miss him so much, we were together too long for me just forget him..I can\'t and I won\'t. Couples have their ups and downs...and their break ups and stuff...PLEASE let this be a bump in the road! I\'m not going to try to talk to him and get him back, I\'ve done that before. I\'ll let him go. I know that\'s what he wants...I think he likes the whole dating thing better anyways. I love him with all of my heart still though. If there was...ANYTHING in the world I could do to get him back, I\'d do it. I\'d die for Matt. I don\'t want to be with anyone except him. I\'ll love him more than anyone, always. and if he wanted me back, I\'d go with him quicker than anything. I don\'t care that he\'s going to Alabama, I\'d stay with him, hoping to see him any time that I could. I want to be his so bad. omg and that song BURN-USHER..yeah I cry everytime I hear that song. I\'ll pray everyday that something happens and I get to be with him. I\'d be the luckiest girl in the world. He means more than life to me. I know I love him, if he loves me back and it\'s TRUE love, I know I\'ll end up with him one day...no matter how long. But if I don\'t end up with him, i know it\'s not what\'s meant to be.  Ok, I just felt like writing about Matt. I love him with all of my heart.  -alicia  '),(18635,3146933,'25,June,2004','2004-06-25 00:00:00','   I\'m sittin here. Last night was GREAT although I was stuck at home! Ok, so...this dude has been harassing me and Monica the past couple days...so Brad says somethin to him about it (I won\'t tell all about that) but and it got bad and Forrest...got mad about that dude bein mean and so he says somethin too! AH! I just wanna say I love them! I heart them! They are...the best guys! I\'ll hopefully be hangin with them a lot more soon. I\'m glad Brad isn\'t moving to Hville in a way. He lives down my road. I would miss him :( ! Zach called me last nite b/c he hadnt talked to me in a while lol...well..since my bday. I thought that was nice. I think he still likes me ;) ! I saw *him* the other nite again.  hmmm \"do you miss this?\" HEEEEEEEEEECK YEA I DO BUDDY! lmao! anyways ok...so my dad may be movin to Pensacola, and I think I\'d like that b/c I\'d like goin down there! woo hoo! maybe Panama City but doubt it. Pensacola has a better chance I guess.  So..I got my senior pix and most of them turned out SO great. *GET THEM DONE AT CLASSIC PORTRAITS!!!* Mom\'s havin a time tryin to find 24 for my portfolio...she\'s goin crazy I think. But it\'s ok. Right Brad? lol. I like Brad. But I\'m goin to bed...  '),(18636,3146933,'21,June,2004','2004-06-21 00:00:00','   Last night was great. Anyways, I hope I can hang out with Brad Wednesday because I feel bad for letting him down all those other times. And Forrest. Poor Forrest, I want to see him too. I hate how much he\'s workin now! grr! :) Hopefully we\'ll still get the chance to hang out though. They\'re so great. Let\'s see....did I mention anything in my last entry about *him~? Probably not. Because I hadn\'t talked to *him~ in quite a while. But it was still great. Well, dad\'s just about done with the tanning bed. I\'ll be glad when I can use it. I\'m so light and I wanna go fishin with Brad and whoever sometime but I refuse to do it unless I have a LITTLE tan at least. I\'ve gotta go walking too. I\'m gonna gain weight sittin around here on the computer all the time and eatin every few minutes. lol. So, I guess about almost half the summers gone by and I\'ve done nothin. Things better start lookin\' up. alicia  '),(18637,3146933,'18,June,2004','2004-06-18 00:00:00','   If my life isn\'t just stupid sometimes. I cannot wait til I\'m 18. Seriously. I\'m talking to Monica, Adrienne and Brad right now. I love them. They are so my best friends in the world. I\'m so bored though...but at least I\'m at mom\'s right now. Went to the drive-in with Monica, Mom, Kenneth, Kristi and Tori. Rode with Monica and got in trouble b/c I wasn\'t supposed to be with her that night. Long story but oh well. Maybe I\'m not grounded. Who knows? I stay grounded. I wish I was perfect so I wouldn\'t :/ *sigh* but oh man. I had some fun today at least. I got to have my senior pix taken and I got them done at Classic Portraits. I say..everyone should get theirs done there. The woman that takes the pics is so nice...and her sets are so pretty. She has like, everything. I hope they turned out as good as Monica\'s did. I got new glasses too tho..they\'re cute. They make me look smart..lol..ok I so don\'t have any exciting news..or well any news at that..so I\'m off I guess...to talk to monica and brad...I love Monica and I heart Brad. alicia  '),(18638,3146933,'10,June,2004','2004-06-10 00:00:00','   \"You don\'t remember me, but I remember you. I lie awake and try so hard not to think of you. But who can decide what they dream? And dream I do... I believe in you. I\'ll give up everything just to find you. I have to be with you to live to breathe, you\'re taking over me. Have you forgotten all I know and all we had? You saw me mourning my love for you and touched my hand. I knew you loved me then. I look in the mirror and see your face if i look deep enough So many things inside that are just like you taking over me.\"  Nothing\'s happened the past couple days. Brad came back today. I missed him actually...we have these late night conversations about these random things and I missed that...having someone to talk to and get their opinion on things.  I\'ve been trying to get my dad to let me go to Orlando with Monica and Jessica and Chasity or Kellie...but he\'s so not budging. I don\'t think he trusts me. But oh well. I\'ve never been drunk and I\'ve never done drugs/whatever and I sure don\'t plan on it. I just wanna head out and have a couple girl days...dude that\'d be so much fun. If I can\'t go this year babe, you better believe next year it\'ll so happen. Ok now, hookin up with some guys and walkin around/riding with them/dancin at this club with them..heck yeah you better believe I\'d be doin that! But that\'s not bad at all. I could possibly meet some guys that live there and get to know some people you know. But I dunno. I mean I wouldn\'t do that if I was with Matt. Speakin of Matt...I didn\'t talk to him for about a week/week n a half...and he msgs me 2nite. dude im so freakin happy you dont even know...he asked me if I wanted to see him 2moro nite. well i dont know lemme see what i got planned ok yea it\'s def. on! :D I love him yall, I really do. Only God knows how much I really love him. I guess it\'s how this first love thing goes though. We weren\'t meant to be and I\'m gettin that...I do like hangin out with him sometimes though. It\'s good to keep in touch and talk n stuff. He\'s a great guy and whoever ends up with him\'s gonna be lucky ;) But I\'m tired. ZzZZzzzzzZzzzZzZZZZ  '),(18639,3146933,'05,June,2004','2004-06-05 00:00:00','   Ah. well, it\'s been pretty good I guess. Lately. The past couple nights I\'ve jus been hangin out with Monica. Man, Friday we went to hville and it was pretty fun, but when we got back..yes..to sboro..we had more fun. lemme explain. well we end up on the phone with bradsmith and hes like \'yea andrew\'s girl\'s up here talkin to some dude n has been for the past hour\' so me and monica and amber ride up there and sure enuf...possum is talkin away to some dude. Ok so Derek, I love him, derekpage forrestwilliams and coreywilson all ride up to where we are and freakin video tape it. hilarious. so anyways and then wrylandreed robbievalentine and i think it was botaylor come up and we\'re all followin her to the square b/c of course the dude she\'s talkin to leaves...so yeah 4 cars followin her. but we end up losin then and i had to be home at 11:30 neways. sucked. but it was fun while it lasted. she was mad.  2nite was ok though, me and monica met bradsmith at the park and we talked a while and walked his dog Booger around. Monica, it\'s not BUGER lol. She kills me. but then he left us :( and we went to watch Shrek 2. that was funny. then we rode around and there were cops up at the square lookin around and stuff again so we were like forget that and i came home. here i am on the computer. im bored and tired so im goin to bed.  oh, btw, i dont know what the deal is with matt. i love him though. gnite.  '),(18640,3146933,'02,June,2004','2004-06-02 00:00:00','   Oh my GOSH! I cannot stand this one guy. It seems like he does everything he can to make me mad about things. ESP Matt. Well you know what? I really don\'t care right now. If MATT is \"talking\" with Leslie then whatever...let them talk...let them watch movies...I talk and watch movies with other guys too...what\'s the big deal? Talkin and watchin movies doesn\'t mean you\'re in a serious relationship. Oh he has me shakin and cryin right now. I cant stand guys sometimes. bleh i feel sick. I\'m talkin to matt...and I jus found out a bunch of stuff that has been makin me mad isn\'t even true...go figure. \"hey! lets be 18 and act 10!\" *rolls eyes* Man, I\'m startin to hate most of the male gender...lol...that doesn\'t sound good. I\'m seriously..debating on moving to Florida. Seriously. One condition being we move close to my cousins...I\'m sure they know a lot of good people and they could help me get comfy down there. They\'re like 18..or 19..around the same age as me anyways. But yeah livin in FL would be pretty cool...I\'ve always wanted to live there again..kinda..at least i\'d be further from matt....i love him. God, I love him. remember to dl Ares  '),(18641,3146933,'31,July,2004','2004-07-31 00:00:00','   Me and Monica ended up goin to that Christian concert up on Skyline mountain(?)!! When we got to Dustin and Patrick\'s church to meet up with everyone...we were gonna ride with Patrick. Brandon was in the front with Patrick and then in the middle was Monica, me and then Dustin...and then this guy named Greg was just gonna sit in the back...where there\'s like..no seats but one of the women wanted him to get out b/c they thought Patrick would get in trouble...SOOO me and Dustin ended up ridin w/ Dustin\'s mom on the way down there...and that Greg dude took our seat!! :( When we got there...oh Lord...it was HOT...me and Monica only brought a blanket to sit on...BIG mistake...the ground was still a little wet from the morning rain and our butts got wet...haha..but it wasn\'t too bad b/c other ppl\'s butts were wet too from sittin on blankets too :D hehe! Anywho...this girl and her bf wanna sit on the ground for some unknown reason and they gave me and Dustin their chairs...*thank them!* and so that was better. Once it got darker...it was much much better...and cooler...but yeah Dustin was all holdin my hand again..*sigh*...it was sweet...SOOO on the way back...we rode with Patrick...and that Greg dude rode with someone else that lives close to him. Aw and Dustin was all holdin me and stuff...and Brandon was on the other side of me...lol..oh my..he kills me..and Monica sat up front with Patrick. So anyways, it was a good time... :) MUAH!! yeah baby! I talked to Brett when I got home...I guess he was a little tired though b/c when my dad called...he got off the phone with me :(.... ugh. dad. yeah. he calls FREAKIN out and yeah he sure did ground me....I won\'t EVEN get started on that. Dad\'s ghey. I\'m tired. Goodnight.  '),(18642,3146933,'29,July,2004','2004-07-29 00:00:00','   Ok so I have to write about tonight. It was a lot of fun ya\'ll!! I was ready a little earlier than I was supposed to be so I got Patrick to just come pick me up here first before meeting Amber and Dustin at the bottom of the mountain. Ok so I do like Dustin. He\'s soooo sweet and he\'s really funny. Anywho, we went to Applebee\'s to eat and I got cheese sticks lol. Didn\'t eat all of them, but close enuf...Then we went to watch Spiderman 2...aww and Dustin was holdin my hand...that movie was pretty good...had some \"Whoa!\" moments...can\'t wait til the 3rd one comes out! *smiles*...Kisses!!! :D Monica had&nbsp; a bad day...and I feel really really bad for her...I hope she gets better...:( I heart her! I have a class with Adrienne so far! Like... I have...1st block with her...should be a lot of fun...we didn\'t have a class together last semester... :( But I\'m thinkin bout changin my schedule. I talked to Brett a little while today...on the computer...only&nbsp;a&nbsp; few minutes though..and I read his blog...aww he mentioned our dog dying..:( it was really sad...I was on the phone with him when I found out and started cryin...he doesn\'t care though... :) aw...well I heart you Brett, you\'re a great friend!  I\'m totally going to bed after this long but great great great&nbsp;night...!!  '),(18643,3146933,'27,July,2004','2004-07-27 00:00:00','   :D !! The last two days have been so great! Aw ya\'ll I really think Dustin is sweet. The night before last, Patrick and Dustin called me and wanted to come over but mom was goin\' to bed soon so I told them I\'d have to come outside to talk to them...so they asked if I wanted to jus go with them and get some food. Sooo I did. It cracked me up so bad bc when they got here, I got in the back and Dustin was like \"Do you want me to drive and you sit up here and make Patrick sit in the back?\" I looked at Patrick with an evil smile and said \"yeah sure!\" lol he was like *hmph* :D Dustin\'s so great! I haven\'t known him long at all either, but I know I\'ll like him more and more if I keep hangin\' out with them! Well, when they dropped me off, they wanted to know if I wanted to go to church with them the next day..Wednesday...and I said yes...and I told them if Monica couldn\'t go, thennnnn I\'d have to have one of them pick me up. They said that was perfectly fine.  Anyways, Monica comes and pix me up for church Wednesday night and we sit with Dustin *smiles*. Patrick was mostly on stage playin\' guitar. They usually have some youth service but their teacher wasn\'t there so we went upstairs anways and they talked about some trips they\'re takin\' soon. Saturday they\'re goin\' to some thing. I don\'t even know what it IS and Dustin\'s mom (the one talkin\' to us about the trips) asked who all was goin\' and Dustin had her put me and Monica on the list when we don\'t even know if we can go yet! I\'ll have to ask mom and Monica\'s gotta see if she can switch hours. *sigh* lol he\'s so crazy! He\'s like \"if ya\'ll can\'t go we can always scratch ya\'ll\'s names out\"! *smiles again* He really wants us to go though. And they\'re goin\' to Six Flags, August 5th! He wants us to go with them there too! lol.. Six Flags WOULD be fun though. I\'ll see if I can go...but I\'m not sure if I can or not..depends on if we have enough money for me to go. I hope I end up goin\'....I\'ll be sad if I can\'t..*frowns*! Anyways, after church, me and Monica watched their band practice...they are so good...I love the way they sing and some of the songs they sing crack me up ;) !! Well, after they practice...we go outside and stand at Monica\'s car and they all start writin\' on her car hahah!! It was funny!! BUUUT it upset me! B/c Dustin wrote \"Alicia likes Brandon\" I was like \"noooo Alicia likes Dustin!!\" oh, aww! I was standing on Monica\'s car..like on the side where the door is opened and he was standin\' behind/beside&nbsp;me and he had his hand on my back and I just thought that was so cute b/c I was about to fall offa there :0 !! Patrick finally went over to Monica\'s side and wrote \"Dustin likes Alicia\" on there and I was thinkin\' \"ahh! thank you!\" ;)&nbsp; Ohhh no. And the other night...I don\'t remember where this was but I wrote \"Dustin is hot\" on Patrick\'s window and it was still on there last night and Patrick showed it to him! haha I said...\"Patrick! now you KNOW you wrote that\" He knew I was jokin\' and that it was me *smiles*... Well, we\'re going on our double date tonight. Hope all goes well. For me and Dustin.  '),(18644,3146933,'27,July,2004','2004-07-27 00:00:00','   I just got off the phone with Dustin! Well, here\'s what I know. Me, Dustin, Patrick and one of Dustin\'s best girl friends are goin to the movies Thursday. We were gonna go in Fort Payne, buuut I think he said Patrick wants to play putt putt and that\'s fine with me b/c I love love love playin! I used&nbsp; to go all the time and I haven\'t been since my family + matt went...so obviously a long time. Anywho, Dustin\'s settin Patrick and this girl up on a blind date. Oh my.  Well, I\'m starving to DEATH! I haven\'t had like..any food all day! :( Buuut moms gone to Krystals now. I will totally devour that food once it gets here.  I was gonna try to go hang out with Brad today but when I got up I had to help mom with somethin and I think he had some other stuff he had to do. *shrugs* It\'s never gonna happen. Made me mad/sad at first but now it kinda cracks me up... I talked to Adrienne and Brett a little while today! Adrienne is hatin bandcamp more than Brett is...Brett\'s just like \"whatever\" and Adrienne\'s like \"AHH I can\'t stand it!\" lol I love her. She kills me.  My food\'s here, so on that note...I\'m gone. Anything that needed to be said, will have to wait :D  '),(18645,3146933,'26,July,2004','2004-07-26 00:00:00','   Sittin here lissenin to Yellowcard\'s \"Ocean Avenue\" and Three Days Grace \"Just Like You\"...I love those songs. Wow...last night was great ya\'ll. Patrick Hall asked me and Monica to come see their band play at their church...Section Church of God...they did SO great! And...their dramas...oh my gah...they were great also. Patrick, for real. They were awesome! So...Patrick wanted me to come so I could meet Dustin Bailey! He\'s from Section of course...and he\'s in the band too...and so was this guy named Brandon and Brad Smith (he\'s from here..the one in my grade) and some Chris guy. They\'re all cool. But yeah, back to the Dustin deal. I know a lot more than I\'ll say buuuuuut Dustin was cool. and hot. very hot. and he\'s TALL!! lol! He\'s like 6\'7\'\'! and I\'m 5\'3\'\' so there\'s a big difference in height but...since when does height matter? :D&nbsp; Monica said she saw Dustin mouth \"Friday\" to Patrick....I won\'t assume anything but I hope he meant he wanted to try to see us Friday again! wow, you guys don\'t know how happy I am! I\'ll try to keep updates on what happens with that. Brett called me last night and he\'s not happy about goin to bandcamp AT ALL...which reminds me, I need to start chargin my phone now&nbsp;so he can get ahold of me on bandcamp break. :D He\'s supposed to let me know how that\'s goin. Ok well he called. haha he said it\'s not TOO bad. He\'s about to eat...of course I would be eating too if I had bc all day. He crax me up! haha... Nothin goin on with Brad. *he* misses me and wants to come see me soon. Dunno what\'s happenin there...it\'s weird.  I\'m just happy. All\'s going well I think. On that note, I\'m going to get food.  '),(18646,3146933,'23,July,2004','2004-07-23 00:00:00','   Most of today was slept away. Seriously. I got up at 2:45 and then took a nap later...at like 5:30 and slept til 8. Lazy day eh? Not much happened today...been on the computer a lot...I think I missed the computer just a little too much.  Jessica B\'s gettin her hair cut Tuesday and so...she\'s wantin it kinda like mine so I\'m goin with her to help her out! She\'s goin to the same woman I\'m goin to too! Should be fun! Talkin to Patrick Hall right now. Hopin\' Monica and I can go watch his church do some dramas and sing some! He\'s wantin me to meet this Dustin guy that\'s in his band. Said he\'s freakin 6\'7!! Whoa! Tall guy! But they\'re best friends so he\'s gotta be cool! Soooo I agreed!  Brad is so crazy! I was watchin him on his webcam tonight and he crax me up! He\'d like start bustin out laughin at these random times...I love it. I hope we can hang out soon. Brad is just...great. No...I didn\'t forget. I talked to Brett a lot today...since I\'m home, he can call my home phone and I don\'t have to wait until freakin 9 to talk on the phone to people! lol! grr!! He\'s upset that band camp starts Monday. I would be too...band camp sucks..and I\'m not even in band. *sigh* I\'m ready to get schedules!! I hope I can change accounting...I heard it sucks. I\'m a senior. I want easy classes!! :D hehe! And I\'m ready to get the last year outta the way! Buuuuuuuuut anywho...I\'m off to bed!...more sleeping...like I need it. *rolls eyes*  '),(18647,3146933,'21,July,2004','2004-07-21 00:00:00','   Been in Pensacola.&nbsp; I drove my car down there which wasn\'t too bad...I followed dad and Josh Killian followed me. He\'s gonna be workin with dad down there. Pretty cool...he\'s a nice guy. Keeps rubbin in my face that he\'ll be 21 August 30 though and so I\'d like to punch him in the face for that! lmao! but nahh really. he\'s cool...I think Dylan messed around with him a LITTLE too much...but Dylan\'s like that. Pretty annoying most of the time. I had fun while we were out...didn\'t get to go out much though b/c dad works from like 6:30 in the morning til from 4-8 every nite. But he does have Sunday\'s off! So...Sunday...we all loaded up and went to the beach...was lots of fun. Got a little burnt...but that\'s what happens when you stay in the house all the time..never get outside..then all of a sudden head to the beach in a bikini and stay there for hours. But it\'s ok. As long as I get a tad bit darker from it, it\'s worth it.  Oh   Oh ...mad is what I am   right   now. Far beyond MAD...let me tell you why. My dog, Furball..yeah..the people that were lettin us keep him...came and got him AND his dad, Choco, while we were gone on our little vaycay. I will get him back, if I have to get a job and pay her for the thing, I\'ll do it. I love him so much...I walked in the house today and I\'m&nbsp; so used to him runnin up to me and waggin his little tail and try to jump up on me...he\'s so cute...and he sleeps with me and follows me around everywhere...but not anymore...and it makes me really sad...I just cried about it.  Anyways. Brad took Laura Page to H\'ville this week while I was gone. It upset me when he first told me but whatever...I mean we AREN\'T anything at all. There is no \"us\". I can\'t get mad over stuff like that...can I? I shouldn\'t. Considering some things that I won\'t mention. So all I have to say to that is....Whatever. *he* wants to see me...what am I going to do? See him probably.  I talked to Brett a lot while I was down in Pensacola..he\'s the only one that cared enuf to call and check on me! lol! We talked for like 4 hours one night...long time huh?  Anywho, I\'m off to bed...pretty tired...ridin\' in a car then van from Pensacola to Scottsboro with 3 kids is...horrible.  '),(18648,3146933,'13,July,2004','2004-07-13 00:00:00','   So...I know for a fact that Brad is mad at me about something. Well, maybe not mad but upset with me or whatever. He is not being nice to me...at all. I just wish I knew what I did to him...to make him act this way towards me. And...I wish...that we could hang out more...I\'m so tired of wishing that it\'ll happen though. If he still wants to hang out more when I get back from Pensacola, hang out more is what we will totally do. I like Brad. He\'s a ... pretty cool guy. ;) Brad Brad Brad... I saw *him* tonight! Aw, I was so excited. He\'ll miss me...he said so himself. :D Wow, I feel loved. Really. *HUG* I\'m going to Pensacola here in a little while today...pretty excited. I hope I get to go shoppin\' and meet people (mostly guys!) while I\'m down there...be nice...but whoa..I really need a tan. Very bad. Anyway, we get schedules soon...I hope all goes well with that. Oh and we need to get my senior pix portfolio! Ahh! *excited*...well I\'ll write more later maybe..I\'m pretty tired. That is..if I can get to a computer down in Pensacola! :O :D  '),(18649,3146933,'09,July,2004','2004-07-09 00:00:00','   Gah, what\'ve I done the past couple days? haha. Oh gah Well I gotta say...the 4th of July was so great! Me and Monica followed Brad to Corey and Ashley\'s and they were shootin fireworks..uhh well when we pulled up in the drive-way..there were TONS of cars there...I was like..wtc..whoa..haha b/c it was supposed to be me and brad and monica and forrest and corey n ashley as far as I knew...I guess..but yeah Derek was there and so was Ashleigh Venable and her b/f Jason Hadden..so me and Monica kinda talked to them for a while .. and Derek gives us a nice \"welcome\" he walks up to us and he\'s like \"hey\" and turns around to walk off and he throws a bottle rocket back over his shoulder at us! lmao! oh God...I ran to his car .. that way he wouldn\'t be throwin them at me! :D I\'m smart like that. So after a few min..most ppl leave..the ones we didnt know neway and Forrest, Brad, Corey, Brandon Wilhelm and Derek all went in the back yard and had this freakin bottle rocket fight. Was very funny. Dereks pants were freakin saggin cuz he had so many stashed in all 4 pockets lmao! We got front row seats toO! they have a screened in back porch so we could sit there n watch the whole thing. Gah it was crazy!! Well whatever. I\'ve had fun at least! The other nite I went to Adrienne\'s with Adrienne, Jermiah, Brett and Jessica. Then Erik came over at the end and then Monica came over lmao! Whole buncha people up in Adrienne\'s lil computer room. Was pretty fun. Jus sat around and talked and Adrienne got on webcam and showed everyone her friends! haha she\'s so crazy! Then...uh...Erik and Monica came over to my house last nite and we got on my computer and we played pool and talked to Adrienne..then..later that nite Monica called dad and asked if I could stay the nite w/ her and so I did...and while we were at Wal-Mart earlier she got a webcam so when we got to her house we hooked it up and was like...invitin everyone lmao. I\'m not explanin but Monica, our plans...never go thru. *GRRRR* Oh and we got this new Smores typa icecream thing from Sonic and it was gooood!! mmmm Marshmellows! i love them! I saw Brad. and Forrest. I saw them both. I saw Derek too..the other day. I was drivin by and I seen him outside and I honked at him lol...he didnt know it was me :( but he DID say that if he knew it was me he woulda waved :) I never catch him outside so it was funny. Anywho. *HE* asked me if I was busy. I said yes. I really wish I could hang out with Brad more. :@:@:@:@:@ Monica better not be sleepin right now. gah...she better be gettin ready. I want to do somethin with her tonight. oh and Erik\'s gone to FL for a couple days...gah! lol Everyone\'s leavin!! :( I\'m prolly goin to Pensacola Sunday to hang out down there and see dad too...we\'ll see how that goes...  '),(18650,3146933,'03,July,2004','2004-07-03 00:00:00','   4th of July...ok so really the 3rd but we had the fireworks at Goosepond Saturday....the 3rd. Whatever so I went with Adrienne and Kristi and JessicaB. We met up with Brett Harden (aw i love him yall) and Jermiah (i love you too) and Mackenzie...great I dont know how to spell his name. But it was pretty darn cool. The fireworks....were so pretty. I had to get Adrienne to still go b/c she was gonna go with Tristan but...he was too sick to go out so she was upset..but I told her she\'d still have a blast with us and I think she did :)..I\'m glad she still went. Aw, I love her and I hope things work out for her and Tristan. I do feel bad b/c Erik Burkett wanted to go with us and he couldn\'t..but that\'s a long story.. :). We all laid on Adrienne\'s cover she brought...lol..was pretty big..and just looked up at the sky...I love the end of the fireworks..the last ones .. you can always FEEL them...it\'s so great. And we stole Bretts coke..which wasn\'t nice of us b/c he paid 2 bucks for a coke... :( he\'s so sweet yall! I really hope this group can hang out more.  K so I come home and Monica\'s here...been here for hours..lol readin magazines...we get on the computer and matt valentine msgs me and he wants to stop by and say hey to us so we talked to him a few min outside and then they left and I\'m here and I\'m talkin to monica adrienne and brad. I loved today pretty much.  '),(18651,3146933,'01,July,2004','2004-07-01 00:00:00','   Sittin here bored...bout to head to bed...cuz I gotta help dad move more 2moro. I cried when I found out he was movin to Pensacola..it\'s pretty sad...but I\'ll still see him so I guess it\'s all good. And at least I\'ll get to vaycay down there this summer!! lol...anyways. I really like Brad..I wish I woulda answered his question different the other day..and that\'s all I\'ll say about that haha. But yeah I do like him. He\'s so great. Havnet talked to him much lately though...kinda upsets me...  Haven\'t talked to Matt in a while..not that I\'m complainin..I msged him one day and he said he was at some orientation thing at alabama so i was like ok or not. Made a new friend too..Brett Harden..he\'s pretty cool. The last couple days I\'ve talked to him on the phone like average 2 hours...haha...wow we\'re blabbermouths or maybe I am...and he listens...it\'s possible! My room at moms..is so pretty..if i know you well, you should ring my cell and stop by n see!  I still need to pick out what senior pix I want for my portfolio...hmm...K well that summed up my week .. not much greatness but that\'s ok. Maybe this week will be better.  '),(18652,3146933,'06,August,2004','2004-08-06 00:00:00','   I have no clue what I\'ve been doin the past couple days. Oh well. I do remember...dad came down Monday and he was here for a while...I helped him some...we switched cars so I\'ve got my celica back! *Smiles major big!* That\'s a great thing. Being stuck at home without an option sucks. Anyways, aw, I love my dad. He let me go to Six Flags with Dustin and Patrick\'s youth group Thursday! I spent the night with Monica Wednesday night and I could NOT sleep. I was tossing and turning all night long! Every time I\'d get comfy, I\'d wake right back up. I don\'t know what exactly the deal was there. Sooo I got an hour or so of sleep. Got up at 4:30 and started gettin ready b/c we had to meet them at their church at 6:30. Wow, and I\'m glad I woke up at 4:30 b/c Monica set her alarm for 4 and I guess she turned it off...Just...woke up..*shrugs* lol. We wore jeans and a sleeveless shirt under a t-shirt. Good thing. Oh boy. Our mornin started off....crazy. Ok so we leave and had some odd minutes to get there...we get down the road and she got gas...then..she realized she forgot her pillow...so she had to back to her house and get it...omg so scary...we were tryin to hurry. anyways..on the way back to her house...she looks over at me and she says \"Oh God, Alicia, PLEASE dont move, PLEASE\" and I was like \"monica..what\'s wrong. what is it?\" and she said \"oh please dont move, i\'ll tell you but please dont move whatever you do\" and I wa slike MONICA JUST TELL ME...IS THERE SOMETHIN ON ME?\" she says \"nooo please dont move please\" and i was like \"Monica, is there somethin NEAR me?\" and she was like \"oh GoD, Oh God, i\'ll tell you but stay where u are..\" so it turns out to be a grasshopper and we pull over at the next store we come to..she freakin FLIES outta her car .. so do i tho and i go to her side and hit it out with her shoe...and some guy that was there came out to see what the deal was b/c we were so loud and he was crackin UP at us...LOL I would too tho so it\'s ok. Soo..on the way down there...I sat with Dustin..pretty cool.. :) I really got to know more of the people on that trip! Like...Sammy Aplin...he\'s so great yall...and Matt Brown...he\'s pretty cool too :) and Cody Elrod I think? and let\'s see..hmm..Josh Newman...Ashley Newman...and a bunch of other people but I\'m thinkin that\'s the only ones that I hung out with the whole time..:D  *marquita says hey* So anyways...we get there and it was   SOOO   much fun but   SOOO   hot! Me and Dustin...we held hands the whole time...aw, so cute!  lol...the Batman and Superman rides were GREAT...when we rode this water ride..oh GOD...gross..there were these 2 girls..they were like...14 or 15 and they were freakin makin out. and I was like...NASTY. When we got off...we had to walk by them and Monica was like...\"..and they stand there makin out and it\'s SOO nasty, that ruins my daY\"  and I was like..yeah mine too. Anyways, I wont\' say anything else about that but yeah the guys made me mad. *sigh* Guys. Dustin was def. sorry :D aw haha. But anywho, the day was great...rides and all...then at the very end...it started POURING down with rain :( *tear* I mean, we had like 45 min to an hour left and it rained that bad..GRRRR my hair got soaked and it was wavy and doin all this weird stuff and i just got mad and stuck it up in a ball and i was like..\"if dustin doesnt like it he can get over it\" buuuut when we got to the bus finally....and i sat with him he was like \"you\'re so beautiful\" and i was like \"uhhh even when i look like this\" and pointed at my hair...and he was like \"I wish you looked like that more often\" ;) how sweet. I look cute even when I look crappy. Aw, and when we were walkin back  to the bus in the rain...Dustin and Patrick ran off and left us...and...Sammy came to us and said he wouldn\'t let us walk back by ourselves! Sweet huh? He\'s really nice. Me and Monica played cards a couple times on the bus b4 we got off 30 min later and ate at Micky\'s! It was sooo funny..they were like orderin from the drive thru and they weren\'t even in a car! Oh gah..Dustin kills me.  I fell asleep on him a couple times but someone always woke me up :( but it\'s ok. I think me and Monica went...right to bed when we got to her house. :D...long day! More later?  '),(18653,3810203,'30,June,2004','2004-06-30 00:00:00','   Oh god, this is strange. I just got back from the Enchanted Castle (oh yay, tokens for everyone!) with my church and I am as exhausted as hell. Since this is my first post, I\'m not sure what to write. Guess you should get to know me a little bit better. Umm, here are some things that I love: Elijah Wood, music, mocha javas, reading, relaxing, Elijah Wood, movies, vampires (of course who doesn\'t?), Asian stuff, Elijah Wood, speaking in different languages to infuriate people, going crazy go nuts, and Elijah Wood. I think that\'s all the time we have for today, folks. It\'s 11pm and I\'m tired, so g\'night.  TAKE CARE OF YOURSELVES!!  Rachel/Mouse Wood  '),(18654,3810203,'02,July,2004','2004-07-02 00:00:00','   well, at the moment i\'m trying to create an alanis morissette fan website, and i must say i\'m not having much luck. i have everything planned out - i even have a logo! but i need to find a server to post it on. i was thinking about using blogger as the base for the website, but that\'s just an idea. i don\'t know how it would work to do that, because i don\'t think you can post any pics on here or anything, and that kind of sucks. if anyone can help me find a good server (besides freewebs - i\'ve got like 4 websites on that server), i\'d be eternally grateful.  thank you so much to everyone for your love and support!!!!  mouse  '),(18655,3810203,'01,July,2004','2004-07-01 00:00:00','   \"anyone plain can be lovely anyone loved can be lost what if i lost my direction? what if our love is the cost? anyone perfect must be lying anything easy has its cost anyone plain can be lovely maybe our love is the cost\"  -barenaked ladies, \"falling for the first time\"  isn\'t that a wonderful song? *sigh* despite the fact that i\'m listening to it right now, i\'m in a terrible mood. of course i was woken up today at 6:30 after about 4 or 5 hours of sleep, what else do you expect? right now i\'m doing another \"babysitting\" job for my little brother. he\'s watching scooby doo in the same room as me, so i\'m doing  some  of the job. ^_^  well, matt\'s coming over today and will undoubtedly be caffeine-high. (not that that\'s a bad thing, but for  this  little guy, it sort of is.) i\'m feeling a little more awake now, but my stomach\'s starting to hurt and i think i might throw up on the keyboard or something. (lovely sentiment, really it is.) so i think i\'m going to say good-bye for today and post sometime in the near future.  until then, adios!  mouse  p.s. \"mouse\" is my nickname, by the way. most people call me rachel, but i am known as mouse to those who know me well enough. if you want an explanation on that nickname, email me.  '),(18656,3810203,'01,July,2004','2004-07-01 00:00:00','   ...so freaking tired. krystal is at my house right now and she\'s being all weird-like, like even moreso than she usually is. she said she took that personally, but, whatever. elmo threw up on the couch today (or yesterday) and i\'m SO FREAKING AMAZINGLY MAD AT HIM! but i still love him.   well, tonight we plan on watching tons upon tons upon tons of movies, and hopefully we\'ll get to see 50 first dates because of the lovely sean astin. he is a great actor, i\'ll give that to him. and if you guys reading this ever saw the lord of the rings, then you\'d be surprised to see him flexing his \"glutes\" in that other movie. he also has a lisp, which is pretty damn funny.  okay well that\'s all. i\'ve posted enough for today.  au revoir. and toodles! no matter how gay that sounds.  mouse wood  '),(18657,3810203,'01,July,2004','2004-07-01 00:00:00','   \"there was nothing in the world that i ever wanted more than to feel you deep in my heart there was nothing in the world that i ever wanted more than to never feel the breaking apart all my pictures of you\"  -the cure, \"pictures of you\"  jeez, i hate mornings. my mother woke me up today at 9:00 from a lovely dream that involved elijah wood and a harley-davidson  (oww-oww~!), only to tell me that i had to baby-sit my little brother nathan for an hour and a half while she took kyle to the optometrist. that probably sounds lame, but i think that during summer vacation, people need their  rest  for crying out loud. ...well, whatever. i\'m up now, but i don\'t want to be.  hmm, i\'m going to have to think of stuff to write about. it  was  raining when i got up this morning but now it\'s sunny. *hisses angrily* i\'m not too fond of the sun. that\'s okay, though, coz i\'m going to see shrek 2 again today and then come home to watch various brandon lee movies. god, i love brandon lee. R.I.P.  since not much has happened since i last checked up here (which was two seconds before i went to bed), i might have to sign off. one last thing, though - i had a craving for a smoothie last night so i decided to mix fruit juices, and i ended up blending orange juice and white grape juice. it\'s DELICIOUS. seriously. when you drink it you feel like you\'re on some tropical island. and for a tired little girl who got back from an evening at enchanted castle, that is pretty damn good.  well, tally-ho!  rachel  '),(18658,317581,'31,August,2001','2001-08-31 00:00:00','   Dude, CAhlen, when I just now read the post about the snow it reminded me of when the dude in American Beauty was like, it was just me and the bag dancing, and that just cracked me up, you complete me:)  '),(18659,317581,'30,August,2001','2001-08-30 00:00:00','   And now for greatest Anime... Blood, the last Vampire Ghost in the Shell Record of Lodoss War Akira Cowboy Bebop Series Ninja Scroll  I think that\'s it for best one\'s so far  '),(18660,317581,'30,August,2001','2001-08-30 00:00:00','   Well, it\'s been about a week since Cliff left for spain:( I want him to come home and play with us, but I hope he\'s having fun. And now Marc leaves in like 3 weeks for Australia:(  On the upside, I have music to listen to today. I got a bunch of MP3\'s copied onto a CD before I left this morning, so I have the new Cake, Alien Ant Farm, some Green Day and Dynamite Hack. It\'s so nice to not have to listen to this game:)  Also, the more I read of \"Youth in Revolt\" the more I like it.  Dej- do you have this alleged \"Last Unicorn\"? I would like to read it. :)  '),(18661,317581,'30,August,2001','2001-08-30 00:00:00','   What do you think about as you drift off to sleep?   Usually, I\'m just like going over stuff in my mind and then I can\'t even remember it the next day. Sometime\'s I contemplate life and other times I just appreciate the people in my life. But then I just lie awake for hours:)  You know, I think the best thing in life is friends and family(aka relationships):) Check you guys later, and cahlen, enjoy your sleep...  '),(18662,317581,'30,August,2001','2001-08-30 00:00:00','   timmah  '),(18663,317581,'30,August,2001','2001-08-30 00:00:00','   Blog fog dog smog hog bog  Ti tay ta to we way wi wo  i so tired, help... me........  fe fi fo fum I smell the blood of an englishman  '),(18664,317581,'30,August,2001','2001-08-30 00:00:00','   \"TIMMAY!!!\" - Timmy  '),(18665,317581,'30,August,2001','2001-08-30 00:00:00','   \"SHHH!!!\" - Stan  '),(18666,317581,'30,August,2001','2001-08-30 00:00:00','   \"timmah....\" - Timmy  '),(18667,317581,'30,August,2001','2001-08-30 00:00:00','   \"To travel is to discover that everyone is wrong about other countries.\" - Alduous Huxley  '),(18668,317581,'30,August,2001','2001-08-30 00:00:00','   \"Love is being stupid together.\" - Paul Valery (1871 - 1945)  '),(18669,317581,'30,August,2001','2001-08-30 00:00:00','   \"The fool doth think he is wise, while the wise man knows himself a fool.\" - William Shakespeare  '),(18670,317581,'30,August,2001','2001-08-30 00:00:00','   \"If it\'s tourist season, why can\'t we shoot them?\" - Seen on A Bumper Sticker  '),(18671,317581,'30,August,2001','2001-08-30 00:00:00','   \"Ask yourself whether you are happy, and you cease to be so.\" - John Stuart Mill (1806 - 1873)  '),(18672,317581,'30,August,2001','2001-08-30 00:00:00','   \"There are no facts, only interpretations.\" - Friedrich Nietzsche  '),(18673,317581,'30,August,2001','2001-08-30 00:00:00','   I suppose it\'s all just different ways of looking at the same thing.  I suppose I\'m a romantic.  All I really think we can do is keep our minds open.  What would it be like to be deaf, blind, and couldn\'t taste, smell, or feel?    Hey Mark, I waiting to get blown away (at how easy you are to beat at Rune... MWA hAHAHAHAHAH)  '),(18674,317581,'30,August,2001','2001-08-30 00:00:00','   So, I guess my point was, I can talk until I\'m blue in the face about nothing, I just don\'t see the point, but I\'ll indulge you...  So, we can\'t really tell if we\'re real or not, Suppose we\'re a computer program, God know\'s we\'re running simulations of our earth on some super computer right now. It has these scripts(people) and they come together and make a blank script, then they add there parts to the new script(babies) then these scripts grow and learn and question, then commune with other scripts. The program runs, the scripts don\'t know any different, they\'re all living in their world. Perhaps there used to be users that interacted with the program(the greek gods?) Perhaps the last user to interact with our particular program(They have lots going of course) was Jesus Christ of Nazareth. Suppose there\'s other users that decided to stay in the program but just be watchers that jump in occasionally(angels and demons if you will). It\'s an idea.   But the point is, we can never know. It\'s kinda what the matrix was getting at, but I don\'t think that we exist outside the computer. God could be the programmer. I don\'t know, and Cahlen doesn\'t either. But the thing is Cahlen will talk about it, and I just keep it on the inside and ponder it while driving places.   Dejah, I do want to congratulate you on speaking your mind:) I hope you know Cahlen and I laugh at half the shit he says on here. In case you were wondering, Cahlen and I are still on speaking terms.  '),(18675,317581,'29,August,2001','2001-08-29 00:00:00','   I know I\'m full of shit, but do we really have to dwell on what we don\'t know?  '),(18676,317581,'29,August,2001','2001-08-29 00:00:00','   I\'ve never seen the world more beautiful  than when it was covered in snow,  glowing like the heavens  beneath the shining moon.  When the snow comes down it makes no sound but the sound  of an angel\'s sighing.  Can you say this isn\'t real?  It\'s not a dream that I\'m seeing here.  It would be a dream if I could live  inside that moment,  and run along side the shining unicorn.  But instead I watch as the sun  melts her color into green.  -thoris  '),(18677,317581,'29,August,2001','2001-08-29 00:00:00','   Can you guess the time of dawning strike cover your ears  or you might miss  your last chance to forget  '),(18678,317581,'29,August,2001','2001-08-29 00:00:00','   Cahlen, can I say that you amaze me? What you said wasn\'t a \"long psychotic rant.\" At least, I understood what you were trying to say. Maybe that\'s because we\'re related....   Tim on the other hand is so concerned with \"bullshit\" that he can\'t see past the end of his nose. All I can say to you right now buddy is... GET A LIFE! You\'re the one who\'s full of shit. Just because you\'re pissed that your life isn\'t the way that you once thought it should be.  '),(18679,317581,'29,August,2001','2001-08-29 00:00:00','   Yeah, that all sounds a lot worse than I meant it too sound  (not the ariel thing, the long psychotic rant.  I would never give a disclaimer about ariel.  She\'s hot.)  '),(18680,317581,'29,August,2001','2001-08-29 00:00:00','   Favorite Movie of all time:  Little Mermaid:  Ariel is hot  '),(18681,317581,'29,August,2001','2001-08-29 00:00:00','   No deep throught right now.  When I write them it will rock your world.  Fav movies of all time: Simon Birch: makes me cry like a little girl Die Hards (yipee cayay mother fucker) Tommy Boy: Fat guy in a little coat and a skinny dude saving the factory Lethal Weapons: Rene Russo is a little hotty Matrix: Keannu is really emerging as a great actor....wow Meet the Parents: Pamela Martha Focker  '),(18682,317581,'29,August,2001','2001-08-29 00:00:00','   Dude, it\'s best movies: best anime comes later this week, cause I couldn\'t get it all on there if I had movies and anime mixed. and dude, you\'re freakin:)  '),(18683,317581,'29,August,2001','2001-08-29 00:00:00','   I know I contradicted myself , but I meant something different.  You had said that we will make our own problems in the sense that we consiously go out and find them, like drugs.  But now I don\'t really think you were saying that, but I just wanted to make the point that it\'s an unconcious thing.  Which is what you meant.  So I was being redundant.  Hey give me a break.  It was 4:30 in the freakin morning.  Dej, when you mentioned that moment at Friday Harbor that is what I think it is all about.  I had a similar moment last winter.  Everyone was asleep in the house and I went outside and the moon was out and it lit up the snow.  It was quiet and untouched and I thought I wish I was apart of that.  I would want nothing else just to be a part of it.  I thought, if there is a God, THIS is God.  And if this is God then everything is God.   It was just in this one moment that I was able to see that.  But that doesn\'t means that\'s the way all the time.  i just don\'t realize because I\'m caught up in everything else which distracts me.  It\'s like we\'re in a dream all the time and when we look back we feel like we\'ve missed something, and I think we have.  There is beauty in everything and we take it all for granted.  And that\'s just the way our brain works.  It gathers information and catalogues it away for quick reference to quicken the analyzing process.  For example, a tree.  You see a tree and you think tree.  Really not a whole lot else unless you need more info.  Yet everything we see is infinitly complicated and we miss it.  That rock and that harbor will be there when your off running around, it\'ll be there when you are dead.  It\'s beauty will remain no matter what happens in your life.  Whether you lead a \"good\" or \"bad\" life.  Do you think it really matters?  That beauty is always there in everything, and the only thing I want out of life is to be apart of that.  I want to wake up from this dream live in that moment all the time.  I believe that its possible.  Do you know what sin really means?  It means to miss the point, that\'s what it originally meant.  When Jesus says he\'s son of God, I think he means we all are.  I don;t think he was any different than us other than he saw the world the way it is.    Tim, I think that in society there is a point worrying.  It gets things done and it\'s a trait that just got us this high up in evolution.  If you have one person who sits back and gets only the necessities accomplished and someone who goes for the gusto all the time well he\'s gonna be the one who\'s gonna pass on his genes.  We find a point to something and are motivated to get there.  But I think that\'s just a tool and we can\'t really use it to answer the meaning of life.  Why does there need to be a meaning?  And what if it is someone\'s dream?  We can\'t and won\'t know any better, and even if we did what could we do about it?  Would it really changed how you live your life?    Dude, you forgot Blood the Last Vampire on the best movies of all time list...  '),(18684,317581,'29,August,2001','2001-08-29 00:00:00','   All in all though, I\'d prefer to just live in the moment.  '),(18685,317581,'29,August,2001','2001-08-29 00:00:00','   Best Movies of all time(in no particular order): Fight Club Snatch American Beauty Moulin Rouge American History X Usual Suspects Momento  '),(18686,317581,'29,August,2001','2001-08-29 00:00:00','   Bullshit, the thing is we never know shit, we don\'t know if we really exist so contemplating death doesn\'t even make sense. This is the thing I love, you can\'t know anything for certain. For all we know, we could be in someone\'s(God\'s) dream. He/She wakes up some day, our existence is totally annihilated, the point is you don\'t know. You can\'t know whether or not you\'re doing anything, a butterfly in brazil triggers a rainstorm in Norway, it\'s all bullshit. Rock\'s shit, death is the end? Nihlist? We believe in nothing. Cahlen, you\'re so full of shit, I can\'t even look at you.  This\'ll be a point where we have to just respect each other\'s opinion and not force our beliefs on one another:)  As far as making your own problems, you said yourself you do, in the sentence after you said you didn\'t. Worrying about what other people think is bullshit, yet, if you worry about that, you are creating that problem. It has nothing to do with the outside world. It\'s us, fucking up our own lives.  ARGH! Oh well, check out these blogs I found today, I think they kick ass!   Sunday Hero    Prostitute Laundry  '),(18687,317581,'29,August,2001','2001-08-29 00:00:00','   Everyone should read Peter Beagle\'s \"The Last Unicorn.\"   Everyone. Do you hear me? It is the greatest fairy tale ever written.... yes, even more than LoR.   LoR was long. And dark... they were walking and they were tired, afraid, thirsty, depressed. Did I say that it was dark? Oh, and they were thirsty too, and tired of eating Lembas. but it never ran out so they had to keep eating it. That sucks. They could have started eating the grass. Grass is could for you. Just like sourcraut. Is that how you spell it? Well... ITS GOOD FOR YOU! Don\'t ever let your mom feed it to you for breakfast.   It\'s gold today shiver me timbers throw out the red garnett and wash your hands of everything you ever believed in  -thoris  '),(18688,317581,'29,August,2001','2001-08-29 00:00:00','   Cahlen, what are you doing up so early? Or should I say.... late?   I agree with you on the whole rock falling off a cliff thing. Sometimes I wonder what the point of everything is and if going to college and getting good grades really will matter in the long run. Well, I don\'t know whether that kind of stuff is important or not. But I do know that despite all the dumb things we do to make ourselves feel like we\'re doing something special- getting a job, going to school, etc- there are moments in between that make it all worth it. Yesterday I was sitting on a rock on the shore at Friday harbor labs watching the boats go in and out... watching the waves and the sunlight as it glittered and sparkled across the tremulous blue. There was the wind, the sun, the rock, and myself. In that moment, the only thing that mattered was that I was there. And all the work that I\'ve done to get to that spot made it all worth it.  '),(18689,317581,'29,August,2001','2001-08-29 00:00:00','   mmmm... I\'ll write more later....  '),(18690,317581,'29,August,2001','2001-08-29 00:00:00','   Please note the time of these entry\'s... Wow  '),(18691,317581,'29,August,2001','2001-08-29 00:00:00','   I just remembered a great line from Fight Club.    \"Fuck Martha Stewart.  She\'s polishing the brass on the Titanic.  It\'s all going down man.\"  (sorry if the quote isn\'t completely acurate)  '),(18692,317581,'29,August,2001','2001-08-29 00:00:00','   Brown cow ice Cream... never had any of that stuff.  Probably pretty good.  It\'s those weird named things that always taste good.    I don\'t think you can really say that someone \"makes\" their own problems.  I think that whatever situation you are in your mind tries to create an equalibriem.  So if you are very well off then things that I would never cross the mind of someone less well-off as being a problem become more prominent.  An example:  Someone simply disagreeing with anything I said would really ruin my day.  I would blow it up into something huge.  While people who might be going through a bankruptcy wouldn\'t really worry about that kind of thing.  they\'ve got other problems.  It\'s almost like their \"worry quota\" was filled.  So I think my point is if one minute you worry about something then something bigger comes up you forget about the previous.  You don\'t even give a second thought to it anymore.  And when you\'re not worrying about these things they still work themselves out without a second thought on your part.  There can always be something bigger to worry about so why worry about anything if it can be deemed inconsequential by something else.  Anyway, though worrying is pointless i think the mind needs something negative to meditate on.  Even in the best situation it\'s gonna find something.    And now for something completely different:  I really hate writing about serious stuff.  We take ourselves too seriously.  If you take a step back and watch we\'re really pretty hilarious.  An observation a co-worker made, \"We\'re just a bunch of monkeys wearing clothes.\"  He was on shrooms at the time...  We just run around thinking we\'re accomplishing something.  Is there really anything we can accomplish?  It\'s a road going nowhere.  We think, oh it\'s coming.  If I just work a little harder.  Get through grade school.  It\'s coming!  What do i do now?  Well High School.  Ok, and now?  College.  Got it, now what?  Get a job.  Sweet.  What next?  Get married, have a family.  One day you\'re gonna realize \"Wow, I\'m there...\" and you\'re gonna feel the same way you always have.  I think it\'s quite obvious there isn\'t a point to anything we do.  So we try to make a point and maybe it\'ll get us somewhere.  Really it gets us nowhere, there\'s nowhere to go.  Religion is the funniest thing when you think about it.  It\'s just us trying convince ourselves there\'s a purpose to us being here.  Why does there need to be a purpose?  We think we\'re different than everything else in the world.  So we\'re farther up the evolutionary ladder, so what?  Face it, you\'re gonna die why is that so scary?  I was listening to an Alan Watts lecture a few days ago and he said something that I think really sums it up.  He said it\'s like the moment we were born we were pushed off a cliff.  So you\'re falling and it\'s obvious you\'re not gonna make it.  There\'s a rock falling right next to you and you cling to it for dear life.  Well the rocks not gonna do you any good, so just sit back, relax, and enjoy the ride.    So maybe that wasn\'t something completely different, here ya go:  I\'m a lumberjack and I\'m ok...  '),(18693,317581,'29,August,2001','2001-08-29 00:00:00','   Had some Brown Cow Ice cream tonight, from the Tillamook Cheese Factory.... mmmmmMMMMmmmmm, that\'s stuff\'s tasty:)  Also, I think I\'d sell some of the cheese for to make some money, but then I\'d eat alot of it too, Cahlen.  I think you have a point, problems are usually replaced by other problems, like if someone has a perfect life, they make their own problems(\"drugs, sex, & rock&Roll)  I dunno, too much deep thought:) I gotta go to bed... Catch you on the flip side  '),(18694,317581,'28,August,2001','2001-08-28 00:00:00','   Enjoy the silence?  I got money to burn, I want to burn it on you, we got one short life, I want to spend it with you... (\"Richard Ashcroft\")  In the immortal words of Bob or Doug McKenzie(\"I can\'t remember which is which, but it\'s the one played by Dave Thomas\")... \"I\'m taking you to the looney bin, then I\'m going to the brewery, Eh?\"  All in all, I think this life we live, isn\'t really predestined... Anyway, I found this poem today over at  urlLink Hangingon.org . and I thought it was really cool, so I thought I\'d share it:  Sex Without Love by Sharon Olds   How do they do it, the ones who make love without love? Beautiful as dancers, gliding over each other like ice-skaters over the ice, fingers hooked inside each other\'s bodies, faces red as steak, wine, wet as the children at birth whose mothers are going to give them away. How do they come to the  come to the come to the God come to the still waters, and not love the one who came there with them, light rising slowly as steam off their joined skin? These are the true religious, the purists, the pros, the ones who will not  accept a false Messiah, love the priest instead of the God. They do not mistake the lover for their own pleasure, they are like great runners: they know they are alone with the road surface, the cold, the wind, the fit of their shoes, their over-all cardio- vascular health -- just factors, like the partner in the bed, and not the truth, which is the single body alone in the universe against its own best time.   Cahlen, that one\'s for you big guy, I hope you\'re having fun at work:)  '),(18695,317581,'28,August,2001','2001-08-28 00:00:00','   So yeah, sup doggs... I finally got the LP\'s.  Freakin A.  Yeah Dej, whenever you come home you can head over here and listen to them before I wear them out.  Anime is cool.  That vampire one was especially cool.  Those blood-sucking-demon-red-eye-people-who-can-fly-only-after-a-brief-transformation-which-looks-really-painful-only-to-get-chopped-by-a-scary/cool-vampire-girl are awesome.  I wish I had seen the whole thing.  Anime women are the best.  Anyway...  So here is my deep thought of the day:  If you could own a cheese factory would you keep it all for yourself?  Hmmm...    And now the real deep thought of the day:  Seems to me that no matter what there are times when we are happy and times when we aren\'t.  No matter what is going on outside my head all emotions really seem to balance themselves out meaning I think I get a pretty good balance of \"good\" and \"bad\".  So if I was rich and had everything I wanted I don\'t think it\'d be all that different.  When people are sad all the time we call it depression, an unhealthy state.  Would it be true to say that constant happiness is also an unhealthy state?  Besides, would we really know we were happy if we were never sad?  Ok, enough with the seriousness.  WHOO HOO YIPPEE AAAAAAAAAHHHHHHH HAHAHA WHOO WHOO WHOO YE HA HA!!!!!!!!!!  '),(18696,317581,'28,August,2001','2001-08-28 00:00:00','   Life is just like the beach: lottsa hot girls that you\'re too nervous to talk to:)  '),(18697,317581,'28,August,2001','2001-08-28 00:00:00','   My Ode to Pepsi: Oh Pepsi, My brown sugar, My sweet love, you are my upper when I am sleepy, you are the candy of my body, you are my sugar high  '),(18698,317581,'28,August,2001','2001-08-28 00:00:00','   GooooooooooOOOoooood Morning:)  I\'m feeling really good today:) I LOVE free pop. That\'ll be the problem with my weight, all this free pop:) After winning 2 games of racquetball last night, I\'m feeling great.  Once, there was a guy with a bucket on his head, He\'d walk around and walk into everything, and then one day he got a car and drove around with a bucket on his head. The end  (I\'ve always had problems ending stories:)  Time to play video games  '),(18699,317581,'28,August,2001','2001-08-28 00:00:00','   So, concerning the whole frick frock issue, no comment.  I just saw this movie called \"Blood The last Vampire\", which was soooooooo AWESOME. I loved it:) I highly recommend everyone see this anime. two thumbs up:)   I LOVE anime. Also, I ordered some Jewel Vinyl today. I think I figured out my birthday present request. I\'m gonna ask for a record player:)  I hope you are all having a wonderful day, and I\'m going to bed.  '),(18700,317581,'27,August,2001','2001-08-27 00:00:00','   Nevermind.  I don\'t have a car.  '),(18701,317581,'27,August,2001','2001-08-27 00:00:00','   Dude.   Where\'s my car?  '),(18702,317581,'27,August,2001','2001-08-27 00:00:00','   Wow. Four in a row. The boss is on a role.   Cahlen was Captain once upon a time. And Zorf too. He was a great Zorf. I was once his Crewmemeber. But most of all, I was Zot. The wicked, the lovely, the master of sneaky-ness. It was fun to be vaporized, but only by Zorf, and never by myself. Could never vaporize myself. So I revaporized the Captain. Because I didn\'t want to vaporize myself. And I hate big butterflies with fangs because I\'m too short to reach the button.   This all makes sense in some other life....   At least one person understands.  '),(18703,317581,'27,August,2001','2001-08-27 00:00:00','   Concerning the 8/25 entry \"I think Cahlen should have said Fuck when he said Frick in the last rant.\"  WHY!!!!??? Just because you feel comfortable saying it and think everyone else should just say it too instead of saying things like  \"Frick and Frock.\" Maybe that\'s what Cahlen really meant. Frick. Frock. Not Fuck.  '),(18704,317581,'27,August,2001','2001-08-27 00:00:00','   Yeah Tim, you complain too much. It\'s okay to get upset about things, but hey, when it\'s all you write about, it gets old after a while.   Chill out. ;)  '),(18705,317581,'27,August,2001','2001-08-27 00:00:00','   timmay timmay timmay!  There once was an old lady who swallowed a fly. I don\'t know why she swallowed a fly. Perhaps she\'ll die. Or have an upset stomach.  At the very least, she\'ll feel a little queasy.  '),(18706,317581,'27,August,2001','2001-08-27 00:00:00','   Hey Cahlen! That\'s a pretty cool dream you had there. Too bad you had to wake up. Guess I\'ll have to come over to your place and listen to your turntable Gordon Lightfoot LP\'s.   What is it about Gordy? He\'s the man. My roommate doesn\'t like him. I can forgive her for that.  Maybe.   So once I had a dream that I was this dragon. The king and his daughter were riding out from his castle into the snowy forest, but he didn\'t know that he was in danger.  I knew, but I was trapped in the dungeon with the king\'s son. We had been captured and chained to the walls. The king and the princess were captured by the evil men, and the princess was dropped into the lake with a ball and chain. Somehow, I broke my chains, and crashed through the glass wall of the dungeon that overlooked the lake. I dove in after the princess.... I can\'t remember if I ever came out.  '),(18707,317581,'27,August,2001','2001-08-27 00:00:00','   I just realized that it sounds like I\'m complaining a lot on here. Which really isn\'t my point. Weird, well, anyway, I had some free time today so I bought some movies and a couple of records. I\'m soooooooooooo excited. I got some Anime and Forest Gump.  So, I\'ve been having Anime cravings recently. Some people don\'t understand Anime I guess, but somehow it just makes sense to me. I LOVE Anime. Also, I think I have one of the best jobs in the world. Play video games then tell the dev\'s how to make it more fun:)   Well, lunch is almost over, I hope you are having a great day:)  '),(18708,317581,'27,August,2001','2001-08-27 00:00:00','   And another thing, it seems no matter who\'s week it is to take out the garbage, no matter how many times I remind them, it\'s always up to me. Not that I expect anything much, I just wish it wasn\'t always my turn:) It\'s especially funny cause the guy who\'s turn it was to do dishes this week, waited until last night to start them up. (*The thing you have to realize for our house is like 2 people eat 3 meals there a day. Thus causing about 10 dirty dishes a day, minimum.which makes doind dishes a daily need.*) I just laugh cause of how no one talked to him about it:) Also, I don\'t eat food prepared at home, so I never have any dishes so it wasn\'t a big deal for me, besides it looking like shit.  anyway, hope everyone is well, and I\'ll check you later...  '),(18709,317581,'27,August,2001','2001-08-27 00:00:00','   So, one thing some of my friends don\'t understand is how I can read like on vacation or something.   Like take for instance when I was in Greece by myself for about a week. Basically, I read the whole week long, got through almost all the way through \"Lord of the Rings\". But I think it\'s weird that I read a lot and don\'t have a HUGE sex drive. Although, I am 23 and I guess I have passed my sexual peak, biologically speaking. Don\'t get me wrong, I want to have sex. I want someone to return the ball.  Love: physical or mental? This is the problem, usually people who are physically attractive aren\'t the brightest. Although, one of the things I love about being a guy, is we usually get better with age. Not that I don\'t think girls get worse, it\'s just that women usually find older men attractive. It\'s weird, I think I\'d always want to be mentally attracted to a person. Although, the physical attraction, usually overpowers any mental attraction (at least as far as I can tell).   Isn\'t it weird how we all find the same people attractive? It\'s the whole, diamond in the rough thing. Finding the one that others have been looking over, cause they don\'t find them so attractive as say, ricky Martin or Elizabeth Shue. Finding the diamond, that\'s what it\'s about  '),(18710,317581,'26,August,2001','2001-08-26 00:00:00','   So, we just got finished watching one of the greatest films ever made... \"Empire Records\"  It\'s one of my favorite movies since I first watched it back in like 95. Anyway, I highly suggest seeing it, if you haven\'t. It was Cahlen\'s first time, kinda. He\'d seen the end b4.  I so want money to spend, but I don\'t have any cause of bills and the shitty economy:( C\'est la vie...  You ever notice how we are never quiet anymore? Like, I can\'t remember the time I went outside and looked at the night sky. Also, I\'m used to way more personal time, and I never get it (*one of the funny things living with 4 other guys*). But besides that, it\'s weird with the whole, personal time vs. lonely time, cause when you get personal time one can sometimes feel lonely. But other times, you\'re like, I just want to be alone. 1 of the things I\'ve learned to do, is always have stuff for personal stuff time. Like I now have about 5 books that I have lined up to read. Now, I just need the vacation time to read them all.  '),(18711,317581,'26,August,2001','2001-08-26 00:00:00','   Hip Hip... On an Island in the Sun, having some fun, I feel fine, I can\'t control my brain. Hip Hip....  '),(18712,317581,'26,August,2001','2001-08-26 00:00:00','   So it turns out I\'ve been combing my hair backwards for as long as I can remember.  See the dealio is this, I was combing my hair in the mirror, and in the mirror everything is backwards. Thus when I was combing my hair to look normal, I was actually combing it backwards. This is really hard to deal with, considering I\'ve been doing it since I was like 8 or whatever age it is one starts combing one\'s hair.  hmm... Cahlen was dropped on his head.... that explains a lot of things:)  well, now it\'s time to play video games, I mean work:)  '),(18713,317581,'26,August,2001','2001-08-26 00:00:00','   So a little more dissemination of information...  here\'s the dealio, I think. First you\'re born and everyone loves you, then you learn that everyone doesn\'t love you later. Then you deal with that, and how you deal with that is how you turn out. But really that\'s just bullshit too, cause you react to everything in a certain way, and that\'s just one thing you\'re reacting to.   If I had to choose between knowing everything and knowing nothing, I would most undoubtedly choose to know nothing.   A common dillemma has hit our house, we\'re out of TP. It\'s not a big problem, but it means I have to go to Costco, and I can\'t go to costco until people pay me for bills, which is basically like pulling teeth. Not you Cahlen, the other people in the house:) (and by other people I don\'t mean kevin or cliff)  We watched \"Cowboy Bebop\" Disk 1 tonight, which was way fun. I highly recommend the entire 6 disk series of which I only have disk 1 and 5.  Sorry for the depressing poem earlier, I just had to get it out.  The truth of the matter is this, I really do know nothing, but at least I have a good idea:)  I love that I have no idea whether or not anyone is actually reading this thing. This way I can write whatever is on my mind and it\'d just there in black and white for all to see, if they wanna look.  g\'night  '),(18714,317581,'26,August,2001','2001-08-26 00:00:00','   I was dropped on my head?   hehe......cool......  Hey Dej, I bought a turntable.  I also bought 13 Gordon Lightfoot LP\'s (records).  Sweet eh?  So I had a dream.  It was pretty cool.  I got this box in the mail.  It was a long box.  I thought, \"Hmmm, maybe there is some Gordon Lightfoot LP\'s in here.\"  I thought that mostly because I had bought some.  I opened the box and there were some Gordon Lightfoot LP\'s.  I was happy.  So I played them with the syringe needle.  Hey, I\'m sure it works as good as any other needle plus come one, it\'s a freakin dream.  Yeah and they played music.  I woke up and I was happy.  Then when I actually came to, you know it takes about 30 min before you say hey that wasn\'t real, then I wan\'t happy anymore.  I wrote an email.  I was mad.  They said they sent them though.  Then I thought, I had them when I was sleeping, and now I have to wait.  Frick.  I\'m going back to bed.  '),(18715,317581,'25,August,2001','2001-08-25 00:00:00','   and now a much lighter poem:  An Ode to Brak: I love you baby- duh duh duh- So I bit him- I think I just went to sleep and woke up stupid- Minkey boodle- pupupu platter- I told you never trust a monkey- Hi everyone I\'m blip, eep eep eep eep do you wanna blow up? Better watch a cartoon Who let a fluffy? Whoa, what died in here? The end  '),(18716,317581,'25,August,2001','2001-08-25 00:00:00','   And now a poem by yours truly...  At night, can you hear me cry? Can you see the pain?  God weeps, for the hell we make and the hell we live, This silence overwhelms me, coming from my soul, A cry, dear God, can you hear me?  The servant that cannot live without you, I feel so alone, will you not comfort me, like you did so long ago?  GOD!!! I\'m here, what is it that you want from me? All the bullshit I see, this can\'t be what was intended....  can you hear my silent screams, though they never come?  '),(18717,317581,'25,August,2001','2001-08-25 00:00:00','   Is Cahlen really a crack baby? I don\'t know, I think he\'s just creative. One time we sang this song about his headgear:  I have to wear my stupid headgear    he has to wear his stupid headgear I have to wear my stupid headgear    he has to wear his stupid headgear     all day headgear   Maybe it\'s because I grew up with him, but I appreciate his weird sense of humor.   Oh yeah, and he was dropped on his head when he was a baby. He fell off the changing table when mom wasn\'t looking.  '),(18718,317581,'25,August,2001','2001-08-25 00:00:00','   So, my love life...  I can recall at one point in my life, when I liked a girl and she liked me back, maybe 2, but 1 is confirmed. Now the problem is this, as soon as I think a girl could possibly like me, I just freak out. I\'ll be like, \"no, you can\'t like me\", which is just stupid.  I really have no idea why I force people that might be attracted to me away. It\'s really the stupedest thing I do. The other thing I do is tell the truth, pretty much unconditionally, which causes people to get hurt. It\'s just all the Bullshit, it\'s a drag, you know?  Then these girls that I could like, are like totally scared that I might like them. Which I don\'t understand. It\'s not like I\'m really a scary guy. At least I don\'t think I am. Maybe, it\'s that I don\'t stand out, but then I do have blue hair, maybe it\'s cause I\'m more plump than the next guy.  Really though, are we supposed to change for the ladies? or are we just supposed to find one that we\'re compatible with?  I know this for certain, \"All women are crazy\", and  \"All men are crazy\", it\'s just finding the craziness that fits together is key. It\'s finding the someone with the same kind or balancing craziness that\'s the trouble.  '),(18719,317581,'25,August,2001','2001-08-25 00:00:00','   So, I went to  urlLink Michael\'s Toyota  this morning to see if I could maybe get a corolla with less payments. That way I\'d still have a nice car but my payments would be less a month.  Actually, I was trying to decide between a Corolla and an Echo,  but after test driving them, I decided on the Corolla. Anyway, so we sit down to go over financials and the dude is like, I\'l be right back. And I\'m like, cool, cause the corollas are having a special 4.9% APR for a 60 mo loan. So, I\'m like \"Kick Ass\"....  About 20 minutes later (mind you all I had to keep me entertained is my cell phone with snake on it.), he comes back with his manager. The manager goes into this little speel... \"Today, we can get you in a 2001 Tundra, and at 60 mos, you\'ll pay 505/month, plus we need a down payment of $1200\"... or translated... (*We don\'t think you\'re serious so please leave our store*) Assholes.   Anyway, I hope everyone is enjoying my postings, I know I am:)  '),(18720,317581,'25,August,2001','2001-08-25 00:00:00','   OK, can I just say right now that Cahlen is a bit of a crack baby?  '),(18721,317581,'25,August,2001','2001-08-25 00:00:00','   Whoa whoa whoa I\'m sitting in a chair  Doo Doo Doo I\'ll never lose my hair  Yay Yay Yay I like the smell of pear  Yip Yip yip Man, I\'m bored...  '),(18722,317581,'25,August,2001','2001-08-25 00:00:00','   So, maybe a little disclaimer...  These rants and raves by the peoples that I know...  They are opinions and are not to be taken as the whole Truth and nothing but the Truth...  So perhaps a little \"Beliefs\" section by your\'s truly...  So here are my beliefs, feel free to write, I don\'t bite (*too hard*)  I believe myself to be christian, probably in the most liberal sense of the word. I do miss ignorance, but am happily living my hetero single lifestyle. I love in your face statements that\'ll probably piss people off, cause no matter how true they might be, our society has told us another way. (*example: evolution, we see it everywhere, yet most christians still believe in creation.*) I do dislike some christian pastors, I feel sorry for them really, cause if you want a big congregation you have to speak to the lowest denominator. The problem with this is you don\'t have a chance to learn much from each sermon and then they ask you for money.  I guess it\'s just this, I like my friends. I don\'t think people should judge me as a sinner for saying fuck, or shit, or asshole, or flipping someone off. None of us is better than the other. Isn\'t that what Christ told us? But then, I feel like I\'m a reverse person where I judge people for not saying fuck or shit or asshole. When really, I\'m fine if people don\'t want to say those words, but I don\'t think they should take offense to shit coming out of my mouth.  I think I\'m very open minded. (*too open minded?*) Anyway, I love meeting new people, even if my Schizotypal self isn\'t:) Well, I gotta go...  '),(18723,317581,'25,August,2001','2001-08-25 00:00:00','   I think Cahlen should have said Fuck when he said Frick in the last rant  '),(18724,317581,'24,August,2001','2001-08-24 00:00:00','   I like El Camino\'s.  Cliff is gone to Spain.  Dejah is my sister.  I want to go to Wyoming.  Tim has blue hair.  I think I\'m starting to like it.  I used to be the Captain.  I used to be Zorf.  But I can\'t do it anymore.  My voice changed.  My youth is lost.  I have a 929.  Where are my Gordon Lightfoot LP\'s?  Frick!  I payed for them last week.  Argh!  What\'s the use of having a turntable.  I have tickets.  American Pie 2 is funny.  I\'m going back to SPU.  I think.  Jebe is far away.  I wear pajamas to work.  Jeff is a cop.  I used to play the oboe.  Why do I have to look at my life as a timeline I need to plan to fill?  Why should I care what happens tomorrow?  Frick.  Frock.  Uncle Toad said you better straighten up right.  It\'d be nice to have money.  I think I\'m gonna become a nomad.  Don\'t beat me down.  '),(18725,317581,'24,August,2001','2001-08-24 00:00:00','   How\'s it going.  Tim made me do this.  Excuse me, are we having a party!!!  '),(18726,317581,'24,August,2001','2001-08-24 00:00:00','   Sup  '),(18727,317581,'24,August,2001','2001-08-24 00:00:00','   I just want to give a shout out to Dejah and Cliff for signing up to post on here. I\'d also like to suggest that Cahlen get his ass in gear  '),(18728,317581,'24,August,2001','2001-08-24 00:00:00','   I guess if there\'s anything that best discribes oneself, it would be your friends...  This is what one friend wrote about me:  Tim - a guy with too much free time who writes descriptions of his friends that are so honest he might lose some of them after his memoirs get published.  he gets pissed when you spell it \"dood.\"  he likes to talk, and his form of humor is to repeat the same jokes over and over.  he\'s anal about being on time, spelling, and about maintaining friendships. he\'s always up for drinking and having a good time.  '),(18729,317581,'24,August,2001','2001-08-24 00:00:00','   So, it turns out I might be a little  urlLink Schizotypal  which pretty much describes me to a T:)  '),(18730,317581,'24,August,2001','2001-08-24 00:00:00','   So, how about a little about your\'s truly?  I\'m a 23 yr old virgin that plays video games for money and does web design on the side.... (now that I got your attention, maybe you would like an explanation)  So, I graduated from Seattle Pacific University last year and went to Europe for 6 weeks. (A Tim\'s adventure which will someday be packaged into a little (*big*) website for all to see and love) Oh wait, the whole virgin thing... Umm... basically I haven\'t gotten laid yet, that\'s how I\'m a virgin. Don\'t get me wrong, I\'d like to have sex, just haven\'t gotten around to it yet. For right now I\'m content with  urlLink Self love    Back to the whole game worker thing.... Wehn I get back from Europe I start working for a dot com, and get hired on as the highest paid webdeveloper there. (Granted I don\'t think I\'m the most highly skilled web dev, but I asked for a lot since it was the internet BOOM and all. Well, come January, said Dot com runs out of money. (*This is after I buy a 2001 Toyota Celica in November*) They decide to drop to 50% employment. Well, I was one of the \"lucky\" few they asked to stay on at a major salary cut. I stayed with them until the money dried up for real in March. At which point I went on uneployment.  Then the tricky stuff starts happening. Since I\'m bored most of the time, I start writing my memoirs, apply for undergraduate work at SPU for a degree in Christian Theology, and basically sit and play video games all day. I contact about 7 \"recruiters\" or placement agencies, but nothing comes through.  Well, I get accepted to SPU and go in to meet with my financial Aid officer, who is great, very helpful. While I\'m there, I check the job boards. I see a web dev job opening for a small company (7 peeps). So I figure what the heck, unemployment is almost run out so I send them my resume. (This is on Monday) Wednesday night they call me for an interview on thursday. Thursday i go in and get the job. I start friday.  Now, my buddy, who is a game tester calls me up and says, \"Dude, they\'re looking for peeps to hire and send me your resume\". So I do. I get called from Volt during the week after my interview with said small company. \"So, MS is hiring a game tester to play new games and tell you what you think of em.\" says Volt person and can I come interview. I say \"Hell yeah\" then go to job interview. Well, I got the job, which is fine considering, said small company only wanted to hire a part time person.  This is my hectic life in the year 2001. I am now not going to SPU, but am working 2 jobs at aroun 60 - 70 hours a week. Well, back to work...  '),(18731,317581,'24,August,2001','2001-08-24 00:00:00','   Second Rant. When I\'m on a roll what can I say?  So women. Can\'t live with them, can\'t understand them, but I\'m still in love with them. One thing you ladies NEED to know. Men masturbate, we do, it\'s not like you\'re going to find one that doesn\'t. I think it\'s bullshit that we can\'t come to grips with this as a society.   The perfect guy is bullshit. I think that women are looking for this guy they think they want, but they never really want it when they get it. Women want a strong man that isn\'t a bully and that is handsome and never screws up. However, by definition, Man is going to screw up, you women just need to learn how to live with it. We\'re going to piss you off occasionally.   I\'m still trying to figure out what women want. I tried to be the sensitive guy, but then they were just like \"I like you but not like that\", of course I was going after christian girls and my buddy says that they\'re just cockteases. Not my words, his. Unfortunately, I don\'t get out much, and when I do I don\'t drink enough to lose all inhibitions. But really, what\'s a guy to do?  Well, I should go to bed, I can\'t think clearly anyway, check you latah-  '),(18732,317581,'24,August,2001','2001-08-24 00:00:00','   So, my first rant. Selling God.  What is it with Christians trying to sell God? It\'s not like you can package him in a little box and candy coat him and say here\'s God. Using God to get money really pisses me off. Take for instances when I was in church on sunday. This dude was asking for money so that he could finance these people to go to his seminar. It wouldn\'t have been so bad but they asked for people to give 5000 dollars and basically said if you didn\'t that you wouldn\'t get as big a reward in Heaven.   This is what church has become. Basically, show off and look better than the next guy. Well, I call Bullshit. I mean you\'re going to tell me I\'m going to Hell if I say Fucking Shit motherfuckers? I don\'t think so. Maybe if I said, \"You motherfucking cocksucking sons of Bitches\"? Well, I don\'t remember God saying well, if this dude is cussing he\'s going to HELL.  Here was the dilemma I dealt with. I was \"set apart\" as the church people say. But then I had trouble relating to non christians and that\'s not what Jesus was about y\'all. Anyway, I just figured that I should immerse myself and try to save them on the inside:)  Anyway, I\'m just ranting. I hope I may have enlarged someone\'s brain just a little bit or at least pissed someone off.  Check you guys later.  '),(18733,317581,'24,August,2001','2001-08-24 00:00:00','   So, Kevin and I just got back from \"Jay and Silent Bob Strike Back\". That was some funny shit. Plus you gotta love the hotties in that movie, I mean, yowzer. Anyway, I\'m gonna rant on my next post which is coming along shortly.   The links I added are to my friends\' bloggers. They don\'t know I have one yet.  '),(18734,317581,'24,August,2001','2001-08-24 00:00:00','   Sup peeps, I think I finally got this mofo up and running. It\'s not like I have a whole lotta shit to say, I just wanted to be a bandwagonner and all my friends were doing it. So, I was like WTF.   Anyway, I hope you are all having a good one. I\'d like to Recommend the book \"Youth in Revolt\" which is fuckin awesome. It\'s basically a look at the thoughts of a 14 yr old male. I found it true to what I was feeling at that age, and if you\'re a lady I think it\'s a great case study at the male psyche. Well, check you latah  '),(18735,317581,'30,September,2001','2001-09-30 00:00:00','   Hi, this is Cliff and I am writing my first blogger input.  I talked to a manager at Taco Bell the other night and found that the Grilled Stuffed Burrito is the hardest thing that they make.  Also, the quesadilla is one of the easiest things to make.  I hope that clears up some confusion.  '),(18736,317581,'29,September,2001','2001-09-29 00:00:00','   in case you were wondering it\'s 10pm  '),(18737,317581,'29,September,2001','2001-09-29 00:00:00','   So, CAhlen has been trying to break into my account so that he can \"talk\" for me....  I wonder why he would want to do that:)  Fuck off CAhlen  '),(18738,317581,'29,September,2001','2001-09-29 00:00:00','   Saw zoolander tonight... It\'s pretty funny, in a dumb sort of way:)  '),(18739,317581,'29,September,2001','2001-09-29 00:00:00','   We went and saw \"Don\'t say a word\" last night (we got there too late to see zoolander)   I felt this movie was really good, and I liked it, 1.5 thumbs up by Tim  '),(18740,317581,'28,September,2001','2001-09-28 00:00:00','   YOU TRY AND INVENT SOMETHING LIKE INWARD SINGING!!?!?!!  well, I got to sleep in today, my only day off in the past 2 weeks:) So, I got my car in for it\'s 15000 mile checkup at 12:30 today:) Woohoo!!! Maybe I\'ll be having subway for lunch... that\'d be delish... Also, I think I\'m gonna pick up silent hill 2 today. Supposed to be really scary... AAAAAHHHHHH!!!!:)  well, it\'s almost time for lunch... Peace out:)  '),(18741,317581,'28,September,2001','2001-09-28 00:00:00','   Tim why do you insist on making these things up?  Give it up dude...  '),(18742,317581,'28,September,2001','2001-09-28 00:00:00','   in other \"other\" news, acording to the manager at Tacobell, the hardest thing to make is the Grilled Stuffed burrito, followed closely by the supreme pizza...  The quesadilla, he said, is rather easy to make by comparison  '),(18743,317581,'28,September,2001','2001-09-28 00:00:00','   In other news, my sister is getting ready to ship out at the end of october:(  '),(18744,317581,'27,September,2001','2001-09-27 00:00:00','   I was visiting some sites tonight and thought the guys out there might like to visit  urlLink this website . Which is the site of the girl that plays the main vulcan on the new start trek show \"Enterprise\"  please wipe drool up when you are done...  '),(18745,317581,'27,September,2001','2001-09-27 00:00:00','   Kevin\'s been doing some desktop customization and last night I found this website which is pretty cool   urlLink Wincustomize.com  '),(18746,317581,'27,September,2001','2001-09-27 00:00:00','   knowing that I\'ll be hitting the snooze button at least one time tomorrow morning reminds me of this morning...  My alarm clock (\"my cell phone\") has a 6 minute snooze button, which is the stupidest thing I\'ve heard of yet.... Take this morning for instance, I wake up at 7:20 to my alarm and hit snooze(\"I don\'t actually have to be on the road until about 8:15 or so and it only takes me about 15 minutes to get ready in the morning\"), well at 7:26 when I\'m about to fall asleep again, right at the point of losing consciousness, it goes off again.... I mean how annoying is that? So I hit the snooze again and history repeats itself at 7:32... I hate this snooze button soooooo bad.... Maybe it\'s time I took my regular alarm clock back from Cahlen  '),(18747,317581,'27,September,2001','2001-09-27 00:00:00','   In other news, I\'ve decided to Add Psycho to our list on the left side, cause I hate trying to go down half a page to find out the URL for her site  '),(18748,317581,'27,September,2001','2001-09-27 00:00:00','   hmm... Tenacious D was on \"The Daily Show\" tonight:) Those 2 guys crack me up:)  Also, Can anyone tell I\'m prone to mood swings? hmm... interesante... random thoughts de Tim:)  '),(18749,317581,'27,September,2001','2001-09-27 00:00:00','   So, now a lesson on cussing...  If you don\'t make a big deal out of it, it won\'t be a big deal, although sometimes I think maybe I should curve my cussing skills...  But then sometimes people really piss me off, and my feeling can\'t really be better expressed than, \"You shit ass motherfucker!\"  Hmm... are there any words that maybe I could make cuss words? Like \"hankering\"... hmm...  Speaking of hankering... I\'ve had a hankering for Thai food a lot recently... and it\'s been like 15 days since my last thai fix, and I want to have like 3 different thai dishes which means I have to go 3 times to the thai restaurant... I tell you my palate is insatiable:)  People don\'t say \"hankering\" any more, how weird is that? also, why do people no longer say, \"2 legit, 2 legit 2 quit\"?  mmm... Hammer:)  '),(18750,317581,'26,September,2001','2001-09-26 00:00:00','   Whoa, stupid dryer...  Well Big A, I\'m glad you\'re still with us:) In other news, Amidy sent me this  urlLink link   I think that I like saying the word \"Fuck\" a lot... cause I find that I can offend some conservative types, and I like to offend conservative types...  maybe secretly I\'m an asshole;-P  '),(18751,317581,'26,September,2001','2001-09-26 00:00:00','   So I almost killed myself today. I was messing around with a plug for the dryer (after moving it), and so I plugged it in (the chord wasn\'t connected to anything). Needless to say, there was one big zap sound, and some burned piece of metal fell on my hand. But I\'m still alive, and a little wiser for the wear.   Anyways... back to watching tough enough.  '),(18752,317581,'26,September,2001','2001-09-26 00:00:00','   subconsciously, I think the thing I fear the most is that a girl might actually fall for me... It seems as soon as I think someone might like me, I do everything I can think of to get them to not like me like that(aka love?)  How fucked up is that? I mean when you combine it with the fact that I long for intimate contact... It just doesn\'t make sense. My subconscious only foils me...  '),(18753,317581,'26,September,2001','2001-09-26 00:00:00','   I said T.E. yo, but that\'s besides the point...  '),(18754,317581,'26,September,2001','2001-09-26 00:00:00','   So, I woke up this morning.     That\'s all.  '),(18755,317581,'26,September,2001','2001-09-26 00:00:00','   TR\'s? is that like a TE?  I woke up with a blue fro this morning... For some reason my hair has gotten to a length where it feels like it wants to sit up and say hello to everything... who am I to argue with my hair?  In other news- I am delighted with the new star trek series \"enterprise\". I feel it had a good opener and enjoyed it very much. (Especially the oiling part:))  Well, off to work...  '),(18756,317581,'26,September,2001','2001-09-26 00:00:00','   So I was reading this book at work.  It\'s a pretty good one about some dude who gets T.E.\'s.  Tim recommended it to me.  I just thought it would be interesting to note that this guy has a split personality and the bad-ass is called Francois.  Francois is a French name, as is another name we\'re all familiar with: Pierre.  Did I mention Tim recommended this book to me?  '),(18757,317581,'26,September,2001','2001-09-26 00:00:00','   And now for something completely different... Sasquatch is real....  Now, can I tell you guys that no matter how many people I meet, the human condition always amazes me.... Artists consistently impress me, stupid people consistently stupify me, and bloggers always interest me... (well usually anyway)  This one time I read this one blog(I don\'t really remember it) but then I was like \"who are these people and what are they talking about\" and then I was like I dunno and now I\'ve forgotten about em except how I wrote this blog about them.  '),(18758,317581,'26,September,2001','2001-09-26 00:00:00','   In other news- Yes Travis I think that Tenacious D did have a sitcom, I think... but I never saw it:(  '),(18759,317581,'25,September,2001','2001-09-25 00:00:00','   In other news- I hate my CD burner. 30% of the time it ends up as a coaster:P  This is unacceptable, Cest la vie....  '),(18760,317581,'25,September,2001','2001-09-25 00:00:00','   Well, I got my first email from someone who actually reads our shiznit!!!  Woohoo, our first  urlLink fan :)   Random emails are the bestest:) Especially when you randomly find out that your personal thoughts may be of someone else\'s interests:)  '),(18761,317581,'25,September,2001','2001-09-25 00:00:00','   So what books are for Tim, DVD\'s are for me. I was at Border\'s with Tim on Monday, and spent far too little time browsing the DVD section, but I did find a lot of stuff I wanted. Then I had to go to Freddies tonight, and found even more. And then I checked the net to find out that Moulin Rouge comes out Dec 18, in a super special edition, and it\'s in a box set with Strictly Ballroom and a new special edition of Romeo + Juliet. Woo hoo! Too many dvd\'s, too little time (and money). Why I didn\'t just major in film studies and become a critic instead of getting a stupid computer science degree, I\'ll never know :)  Lessee... it\'s almost moving day, just a bit more cleaning to do and I\'m out of this crazy temporary living situation. Woo hoo! A new apartment (still in Lovely, Downtown Monroe *roll eyes*). At least it\'ll be our place instead of someone else\'s house.   On a sports note... how about them San Diego Chargers! Woo hoo! 2-0! (for those who don\'t know, but are actually wondering why the SD Chargers, I\'ve been a fan for 20+ years, ever since I saw them beat the Seahawks by like 40 points or something way back when).    And for those who don\'t regularly contribute to this site (or are not listed in the links) who actually read this stuff and write to us, what the hell is wrong with you? :) I mean, um, thanks.  '),(18762,317581,'25,September,2001','2001-09-25 00:00:00','   To go along with that post about books earlier... Monday, I went into a \"Borders\" and was like, \"I want this book and this book and this book, and dammit they\'re out of this book that I also want to get\". Then yesterday 2 books I order from Amazon finally arrived to be added to the pile of books to read. And then today, I got my advertisement from  urlLink SFBC  and was like, I want this book and that book and....  You get the idea, I need to stop reading so damn much:)  '),(18763,317581,'25,September,2001','2001-09-25 00:00:00','   Amidy: No. Steve made \"Rye and Ginger\" last night. Basically it\'s 50% rye whiskey(Wild Turkey) and 50% ginger ale. Then top with a twist of lime:)  It was a good drink, and only took me 3 tall glasses to create the wonderful posts you saw below... but it might have been 4, I lost count after 2:)  '),(18764,317581,'25,September,2001','2001-09-25 00:00:00','   NOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO!!!!!!!!!!!!!!!!  That\'s why I loved twilight zone, they were always about the \"be careful what you wish for\" things:)  '),(18765,317581,'25,September,2001','2001-09-25 00:00:00','   Cahlen, you SO have to get your own blog. This shit is too good, it needs its own home.  '),(18766,317581,'25,September,2001','2001-09-25 00:00:00','   Tim you remind me of a twilight zone episode where there\'s this dude who really likes to read.  He is sad, though, because he is so busy he never can.  But one time when he was in the bank vault a bunch of bombs go off and kill EVERYONE.  He comes out and is like, \"What they hey?\"  Then he walks around and sees all these books laying on the street and he realizes this is where the library was.  And then he\'s like, \"Cool dude!\", because now he can read forever and no one can bother him.  You know what happens next?  The stupid idiot breaks his glasses and he\'s like, \"NOOOOOOOOOOOOOOOOOOO!!!\"  Sad story eh?  So be careful Tim...  '),(18767,317581,'25,September,2001','2001-09-25 00:00:00','   Also, for those concerned, I found the CD  '),(18768,317581,'25,September,2001','2001-09-25 00:00:00','   December 19th.... if only it were tomorrow...  In other movie news... Zoolander comes out on friday... I want to see that one good:)  \"What is this? A school for ants? How are the children supposed to learn how to read if they can\'t even fit through the doors?\"  '),(18769,317581,'25,September,2001','2001-09-25 00:00:00','   ZANZIBAR!!!  I think I\'m addicted to buying/reading books... If I had my life to live as I please, I may very well have chosen to just read. If I retire, I imagine I\'ll mostly be reading... well back to reading:)  '),(18770,317581,'25,September,2001','2001-09-25 00:00:00','   hmm... I\'m feeling better now:)   I still love Tenacious D:) Well, work is going well... There was some stuff I wanted to talk about but now my lunch is over:( Amd i don\'t think I could remember anyway:(  Check you guys latah  '),(18771,317581,'25,September,2001','2001-09-25 00:00:00','   And now I have to live with my CD on the florr, cause it just fgell there, and now I can\'t find it cause it\'s dark  '),(18772,317581,'25,September,2001','2001-09-25 00:00:00','   don\'t worry, this is still drunk tim,   I just realized how random my last blog was, but then I realized I was drunk, so you all we\'re expecting much...  BTW- Jack Black is the bestest in forever, in other news, I don\'t think I can spell when I\'m drunk... these may seem  like totally tangible paragraphs, but also, I love Tenacious D:)  '),(18773,317581,'25,September,2001','2001-09-25 00:00:00','   Hello- This is drunk tim:)  I am listening to Tancaious D  It rocks:)  Also, Steve got me drunk tonight:) He is bad:) And then kevin put a donut down my pants. I did not think that was appropriate. But maybe tomorrow I will forget everything about it, except how I have this blog now to remind me. Also, my teeth are numb, this is my best:) And I think my lips are numb too, but I can\'t tell to sure. Well, I guess I need sleep, after all I am lopsided as I post this post.  Good night  '),(18774,317581,'25,September,2001','2001-09-25 00:00:00','   Hey everybuddy  So work has been weird. I guess we\'re not working OT again for a while, which I just don\'t get. First they say we\'re working all the OT we can, then they say no more OT. I just wish they\'d make up their minds. It\'s tough to get into a normal routine when your schedule seems to change everyday. On a bit of a side note, my lead can be kinda weird sometimes. Like Saturday morning, I forogot to shut my web browser after I checked my hotmail account before work, and he\'s all giving me some inquisition about having a web browser open while I am working. Then yesterday my co-worker had a web site open while working and nary a word said by the lead. Of course, this is the co-worker who the lead always goes to lunch with and carpools with. Draw your own conclusions. It\'s just wierd to have a lead who acts all cool one day towards you, and then acts all superboss the next. Argh. At least I\'m enjoying the game I\'m working on a bit more.   Anyhowie.... this weekend is moving time, so the temporary living situation is almost over! Finally!   Time for sleep.  '),(18775,317581,'24,September,2001','2001-09-24 00:00:00','   The other night when we got Top Foods chinese food. The asked for my \"secret code\" at the ATM pay thing. I liked how they said \"secret code\" cause it made me feel like I had a secret that no one else could know. I felt like a spy:) Anyway, I think they should all say \"secret code\" instead of PIN number, that way I can feel like a spy all the time:)  '),(18776,317581,'24,September,2001','2001-09-24 00:00:00','   In other news, I foiled Kevin.... Midi sound level is now set to 0:)  and now to the great news, I got my starbucks this morning. Can I just sing the praises of this wonderful system? I get there and there\'s my coffee:) I hear people whisper amongst themselves (\"How did he do that?\") It\'s like I\'m superman and I have kevin to thank for that. You and your company are da ant:)  PS - Steve is in town tonight....  '),(18777,317581,'24,September,2001','2001-09-24 00:00:00','   on a side note: On my way home tonight there was a light fog. And it\'s not like the heavy fog, I could see just fine.   Anyway, I love the fog like that. It adds an air of mystery to the world. I think we need more mystery in the world. It was like i could see the edges of the world, but I could never get there.  goodnight  '),(18778,317581,'23,September,2001','2001-09-23 00:00:00','   My Bisnitch is the Shiznit, by Tenacious D is the bomb:)  hmm... I wonder why my blog isn\'t blocked by SPU? I mean really, I have said Fuck and shit on here before.... Maybe it will be blocked now... I guess if it is, someone could email me about it But then how would they know to email me?  Anyway, I gotta shower cause I\'m sweaty from racquetball, check you latah....  '),(18779,317581,'23,September,2001','2001-09-23 00:00:00','   hmm... maybe CAhlen and I are thinking of different people, cause evidentally there are more than 2 people named pierre in this world....  hmm... weird  BTW, tell your dog hi for me:)  '),(18780,317581,'23,September,2001','2001-09-23 00:00:00','   All this talk about a dog... wow. Who knew you guys could actually speak to the animals. Maybe you should become vetranarians or something.   I actually didn\'t have to work OT today, so I\'m just chillin waitin for Tim to go hang out for a bit. Woo hoo! Unfortunately, I\'ll be doing tons of OT for the next couple of weeks, so my one whole break will probably be moving on Saturday. Fun times.   What the hell is SFS? ohhhhhhhhhhhhhhh, wait, I get it. nevermind.   Noich noich noich.   Snoochie Boochies. Who the hell says Snoochie boochies? That\'s like baby talk or something.  '),(18781,317581,'23,September,2001','2001-09-23 00:00:00','   BTW - I spoke with Pierre as well...  I asked him what he thought about Timmay and him saying that he was French and had all sorts of women after him and all, and you know what he refused to talk and in fact looked a little confused.  I think he was interested in the bacon I had in my hands.  It\'s beyond me how Tim got all that stuff out of him when all I could get was a \"ruff\".  Pierre is so cute...  '),(18782,317581,'23,September,2001','2001-09-23 00:00:00','   umm... not me... also... Cahlen is the one who is the ball licker  '),(18783,317581,'23,September,2001','2001-09-23 00:00:00','   So, I finally got to register for classes. It\'s been close before, but I don\'t think I\'ve ever done the day of classes. After I paid the outrageous amount, including the extra account interest charge, I came back to my apartment and tried to register, only the hold wasn\'t removed. How annoying. I was about to condemn SFS to some extra dimensional void, but then I called the girl who was helping me and she fixed it.   Raise your hand if you like SFS..... Anybody? Anybody?  '),(18784,317581,'23,September,2001','2001-09-23 00:00:00','   And shaggy.  '),(18785,317581,'23,September,2001','2001-09-23 00:00:00','   That\'s Jules Pierre to you buddy! And he\'s only dirty most of the time.  '),(18786,317581,'23,September,2001','2001-09-23 00:00:00','   Pierre = Dirty Cockerspaniel  '),(18787,317581,'23,September,2001','2001-09-23 00:00:00','   BTW- I spoke with Pierre...  1st he\'s mad at Cahlen for inferring he is a dog. But when I asked him about his name vs. upbringing he explained it as so...  \"My dad is a french international banker. He fell in love with my mother on a vacation to Sicily.  They had a passionate love affair and I am named after him. They had to remain apart mostly until I was 5 when he brought us to live with him in the US.\"  So that\'s Pierre\'s story...  I guess that\'s good cause I hear french people are stuck up, but in my opinion it was the Spaniards that were the one\'s that were stuck up....  '),(18788,317581,'23,September,2001','2001-09-23 00:00:00','   I just remembered that my family\'s dog\'s name is Pierre.  '),(18789,317581,'23,September,2001','2001-09-23 00:00:00','   And now for something completely different- I woke up at 6:45 today, which means I got about 9 hours sleep. I have a feeling this might throw off my entire day:( Oh, but I do get to play racquetball tonight with Kevin:)  '),(18790,317581,'23,September,2001','2001-09-23 00:00:00','   So as you may or may not have heard/understood from kevin, I went to bed at 9 something last night. Needless to say, I needed that. I found that I was starting to do things that I usually only do when I am drunk....  Anyway, to answer kevin\'s bitching question(\"Don\'t worry Kev, I didn\'t take it personally\"). I think we tend to complain when... a) we\'re bored b) we don\'t know what we\'re doing c)have really stupid management d) have a job. If anyone of these things is true, you\'re most likely not doing the thing you\'d be choosing to do for 8 hours on a given day. Like take sleep for instance, I know an individual (perhaps 2) that loves to sleep almost all weekend:) Now preferably this individual would stay up all night surfing the internet and working on projects, unfortunately said individual has a job and therefore can\'t stay up till all hours of the night(although he does)  In other words, Bitching comes with a job:)  In other news, Travis is a bitch:)  '),(18791,317581,'23,September,2001','2001-09-23 00:00:00','   I think Tim\'s been watching Fried Green Tomatoes or something again, with all this wrapped in Saran Wrap stuff. Anyways...  so I haven\'t written anything in here for a while, but I don\'t really have anything to write. I must lead one of the most boring lives of anyone on the planet or something. It\'s scary sometimes. What can I say though, when you only have 1 day off a week, and you just spend your day resting. Maybe I should just move to Japan for a while or something :)  Oh, and it\'s nice to see you posting again, Dejah.  '),(18792,317581,'22,September,2001','2001-09-22 00:00:00','   hmm... work:(  So today was another boring day in the whole game playing world. I played for about 6 hours today. I\'m ready for a day off. (5 days and counting down) I\'m just tired all the time. I did get my CCR the other day. They\'re on Vinyl and I\'m loving them, when I get to Cahlen\'s room to listen to them. Maybe I should sleep more, I think I\'m falling asleep right now.  hmm... SPU starts tomorrow which means I have had some friends move back to seattle:) It\'s good to see friends.  Speaking of which, Steve\'s back in town which is always a whirlwind of drinking and partying:) Also, there\'s another guy I know coming up from portland this week...  well, good night, and Trav, I\'m serious dude, don\'t you be fallin for my girl...  '),(18793,317581,'22,September,2001','2001-09-22 00:00:00','   Yeah, I  saw a guy today. I said hi to him because I knew who he was. But he already has a girlfriend. Too bad.   What\'s this about saran wrap?  '),(18794,317581,'22,September,2001','2001-09-22 00:00:00','   In other news, I saw a girl today, but I didn\'t know her and I didn\'t say hi to her. In fact there\'s really no reason to mention it, I mean besides the fact that I still haven\'t gotten laid:)  Speaking of which, I talked to Cahlen and he said that Kelley did not come over here today in saran wrap. I mean really, what\'s up with that kelley? Well, it\'s way past my bed time.  '),(18795,317581,'22,September,2001','2001-09-22 00:00:00','   Also, you can now email pierre personally and he\'ll get back to you as soon as he\'s available at  urlLink Email Pierre  '),(18796,317581,'21,September,2001','2001-09-21 00:00:00','   k now I\'m getting jealous of Travis and how he\'s quoting my girl. If he wants to throw down, I\'m here:)  JK, I did feel a pang of jealousy to think that someone might be falling for my girl....  '),(18797,317581,'21,September,2001','2001-09-21 00:00:00','   Also, it turns out I registered www.unwelcomehero.com for my friend pierre, he\'s a little pissed about that, but I don\'t know what to tell him, except maybe, unwelcome hero is almost maybe even better than unwanted hero, and who knows pierre someday one of our friends may get that site:)  '),(18798,317581,'21,September,2001','2001-09-21 00:00:00','   Umm... yeah... I call bullshit  Also, did anyone else notice how cahlen\'s low on avoidant is red? Maybe someone doctored the text so that it read low, but could not change the color of the text.... hmmm...  '),(18799,317581,'21,September,2001','2001-09-21 00:00:00','   Hey Tim, I figured I should go ahead and post my results because I have nothing to hide...        Disorder      Rating        urlLink Paranoid :    Low       urlLink Schizoid :    Low       urlLink Schizotypal :    Low       urlLink Antisocial :    Low       urlLink Borderline :    Low       urlLink Histrionic :    Low       urlLink Narcissistic :    Low       urlLink Avoidant :    Low       urlLink Dependent :    Low       urlLink Obsessive-Compulsive :    Low       --  urlLink Click Here To Take The Test  --        Mwa ha ha, I\'m perfectly normal what\'s wrong with you guys?  (How does this HTML stuff work again?)  lB^)  '),(18800,317581,'21,September,2001','2001-09-21 00:00:00','   I am afraid....  '),(18801,317581,'21,September,2001','2001-09-21 00:00:00','   Also, Kevin and I went to dinner(Cafe Veloce, wonderful food, 4 stars by Tim) and a movie(the musketeer, do not see this movie unless you are looking for a laugh) for a belated bday thing. I had fun:)  In other news, I\'m still a virgin and Jebe I don\'t care if it lives up to the hype, I just want to get it out of the way:)  PS - I don\'t really just want to get it out of the way. But I do want to get it:)  '),(18802,317581,'21,September,2001','2001-09-21 00:00:00','   hmm... Just got done listening to some Creedance Clearwater Revival. It conjurs images of weekends with dad(the smell of his house, the kitchen on saturday morning when the light would come in through the window, driving with my step mom)  It reminds me of little tim., aka pierre. Happily living oblivious to all that goes on in the adult world. No bills, no worris, just normal schizotypal Tim:)  PS- Cahlen is avoidant, which is probably why he didn\'t tell you himself... (True story, he told me that he took the test and is avoidant, when asked to post it, he said, \"no way\")  Also, it\'s interesting that he has a job as a security guard where he has little interaction with people. (it\'s a night time gig, in case you didn\'t know)  '),(18803,317581,'21,September,2001','2001-09-21 00:00:00','   Umm... Mary, I think Pierre really wants to kick this guys ass. In fact, I know he does  '),(18804,317581,'20,September,2001','2001-09-20 00:00:00','   last time I took this, I was only moderate on schizoid and schizotypal:)  I guess it\'s confirmed I\'m schizotypal:)  '),(18805,317581,'20,September,2001','2001-09-20 00:00:00','   Disorder      Rating        urlLink Paranoid :    Moderate       urlLink Schizoid :    Moderate       urlLink Schizotypal :    High       urlLink Antisocial :    Low       urlLink Borderline :    Low       urlLink Histrionic :    Moderate       urlLink Narcissistic :    Low       urlLink Avoidant :    Low       urlLink Dependent :    Low       urlLink Obsessive-Compulsive :    Low       --  urlLink Click Here To Take The Test  --  '),(18806,317581,'20,September,2001','2001-09-20 00:00:00','   dammit, I can\'t even spell saran right... Argh, and after you all know how anal I am about spelling. Sometimes, I just could kick myself, but then I realize I\'m not that flexible:) (ok I am that flexible)  Also, Pierre thinks he would like to kick someones ass, and in the middle of the ass kicking say, \"And call me pierre dammit\"....  '),(18807,317581,'20,September,2001','2001-09-20 00:00:00','   dammit, I can\'t change other peeps posts. Oh well, sorry Tim.  '),(18808,317581,'20,September,2001','2001-09-20 00:00:00','   I know, I\'ll fix Tim\'s spelling of \"Saran\" wrap.  '),(18809,317581,'20,September,2001','2001-09-20 00:00:00','   Kevin here, posting to Tim\'s blog for the first time. Okay, apparently that\'s all I have to say.  '),(18810,317581,'20,September,2001','2001-09-20 00:00:00','   So, I was just sitting here reading kevin\'s post thingy, and I tried to hit F5 for refresh, but I accidentally hit one of the many keys above my keyboard(web/home) and all of a sudden I was at my home page. I was like umm, what happened?!?!?. Well, it turns out all those keys actually work. I had no idea. It\'s like I just got a new toy:) No money down, no payments ever:)  '),(18811,317581,'20,September,2001','2001-09-20 00:00:00','   pogo pogo pogo pogo pogo pogo pogo pogo pogo pogo pogo pogo  '),(18812,317581,'20,September,2001','2001-09-20 00:00:00','   also, I\'m still addicted to reading, here\'s a list of things to come(in order, I think): The Last Unicorn(Now) Lolita Soul Survivor: How my faith survived the church The Boy Star Wars: Darth Maul Star Wars: Cloak of Deception Year\'s Best Sci-fi No.6 His dark materials  Plus I have one more Martian Tales that I still am trying to read... and a book about vikings that aaron has  '),(18813,317581,'20,September,2001','2001-09-20 00:00:00','   Umm... also, I am LOVING \"system of a down\". I don\'t know but I think they have a lot of good points, plus it gets my blood moving:)  '),(18814,317581,'18,September,2001','2001-09-18 00:00:00','   on a side note(to Jebe, Kelley, Amidy, Mary) it appears that  urlLink sunday hero  went to middle and high school with Deanna Wright who plays \"Kay\" on passions...  '),(18815,317581,'18,September,2001','2001-09-18 00:00:00','   hmm... on second thought, if you are one of those girls talked about below, I\'m sure cahlen would be happy to entertain you:)  '),(18816,317581,'18,September,2001','2001-09-18 00:00:00','   So now Kevin and Craig are in my room playing GT3... Craig had not seen it and did not know what the fuss was about.  Damn, damn, I really want to watch the daily show from last night, but no one has it:( Soak Kelley, I didn\'t think you were really mad at me, I just wanted everyone who reads this to think you were:) (oops, now will you really be mad at me?)  Also, I am working this weekend, so if any girls who read this site that I don\'t know about were planning on showing up at my house sometime saturday or sunday afternoon dressed only in syran wrap, you\'ll have to wait until evening or do it next saturday. (is that a run on sentence?) hmm...  '),(18817,317581,'18,September,2001','2001-09-18 00:00:00','   Just remember, as one of our friends so eloquently put it: \"you don\'t have to be dating to be sexually unfrustrated.\" Now, \"go be productive in bed.\" :)  '),(18818,317581,'18,September,2001','2001-09-18 00:00:00','   so, my meeting went well tonight, no new work for tim:)  Also, to the amazement of all that know him,  urlLink Kevin  (You know my roommate, who\'s birthday it is today) has started his own Blog:) It kinda brings a tear to my eye:)  In other news, I still haven\'t gotten laid:)  '),(18819,317581,'18,September,2001','2001-09-18 00:00:00','   IT\'S KEVIN\'S BIRTHDAY!!!! YAY!!!  '),(18820,317581,'18,September,2001','2001-09-18 00:00:00','   So evidentally Jon Stewart and the daily show were hilarious tonight, but my mom called right at the beginning and wouldn\'t let me watch it. So now I\'ll never know and now I think  urlLink Kelley  might be mad at me:(  Also, I was thinking, I am really sexually frustrated almost to a point of actually pursuing sex. Not quite to the point of actually pursuing sex, but it seems that I cannot live in this state of sexual frustration forever. I mean, I want someone to return the ball:)  Another thing I find interesting is that the peeps that read this thing. Like there\'s people I talk to who\'ve I\'ve mentioned the site to and then they tell me later that they find it interesting:) (or at least that they read it, they don\'t really say it\'s interesting)  Also important news tonight. I\'ve registered www.timsadventures.com and www.unwantedhero.com  The timsadventures will be the site for normal tim, and the unwantedhero will be for alter ego tim(or as I like to call him, John)  well, check you later  '),(18821,317581,'18,September,2001','2001-09-18 00:00:00','   Dude.... I\'d say let\'s go do something Saturday, but then I\'m probably going to be working that day anyway (just til 5 though).   Work, it can be fun, and suck at the same time. It\'s great to have a job and all, especially one where I\'m just working on testing games, but when it sucks away all your life, it can be kind of annoying. oh well, I\'m just glad I have a job.   So I don\'t have anything else to say, I can\'t help it, my brain is broken.  '),(18822,317581,'17,September,2001','2001-09-17 00:00:00','   so, I don\'t think I\'m a very good teacher... I was trying to get jebe\'s supergreg picture to link by telling her about the onclick=location.href thingy, but then it didn\'t work and she had to call her parents.  So now I feel bad cause I don\'t know what\'s wrong with it and I may have had her break it accidentally:(  Oh my goodness: it works now:) I guess she figured it out:) YAY! I love when links work:)  Sorry for cunfusing you jebe!  '),(18823,317581,'17,September,2001','2001-09-17 00:00:00','   So I was reading Jebe\'s posts right now, and then I clicked refresh and like a book appeared. I couldn\'t understand most of the words in the \"post\" but it seemed to be something to do with medical stuffs.   Well, now I\'m waiting for cliff to get home so we can play video games or something.  I started reading \"The last unicorn\" today. The jury is still out on this one. It gets right to the subject right away, which is good, but.... I don\'t know.  Anyway, I got asked to work saturday today, so I have to decide if I\'m gonna do that. (Notice: if you want to do something on saturday, tell me tonight, otherwise I\'ll probably work) I said I\'d find out tonight and get back to my boss tomorrow. It\'s a fun job that I\'m enjoying, I just don\'t know if I can work 13 days straight.  It\'s especially difficult when I don\'t get to bed before midnight....  '),(18824,317581,'17,September,2001','2001-09-17 00:00:00','   In other news- Cliff is home now... He and hannah pulled up as we were going out to get TacoBell. So now we\'re listening to his (\"How I heard about the WTC terrorist attack\") well, time for bed...  '),(18825,317581,'16,September,2001','2001-09-16 00:00:00','   So cliff and hannah are back in the us-  They\'re supposed to get in at 10:30am tomorrow. So I think I\'ll not be having a meeting with mean company:P They only want to harrass me anyway.  well, I gotta eat and sleep.... mmmm sleep good  '),(18826,317581,'16,September,2001','2001-09-16 00:00:00','   So I finally finished my book, \"Youth in revolt\" and decided to give you a couple of quotes that I enjoyed from the book.  \"THURSDAY, December 31, New Year\'s Eve - Consider, if you will, the morning boner. What a metaphor of hope and renewal! How can anyone give way to despair when one\'s groin greets each new day with such a gala spectacle of physiological optimism? And what an anotommical prod to philosophical contemplation on the final day of the year\" pg. 493 - Youth in Revolt, C.D. Payne   And now I can\'t remember the other quotes I wanted to share:( Well, Cliff is supposed to come back tonight(cross my fingers)  Check you guys latah  '),(18827,317581,'16,September,2001','2001-09-16 00:00:00','   Argh in my eye!!!!  Today, I reformatted(after spending 7 hours at work) and then realized I failed to back up my messages:( I have it backed up from 2 weeks ago, but I archived almost all the messages from the past 2 weeks and then forgot to save the archive:(  So, now I have no written record of my emails....  Hold on, Thank God for \"Deleted items\" folder on hotmail. All of my stuff I \"moved\" to archives is still waiting to be deleted. I copied it back to my archives folder:)  It\'s 1:15 I need to go to bed:)  '),(18828,317581,'16,September,2001','2001-09-16 00:00:00','   Yes, CAhlen, remember all those girls you talked to during school? Or lack there of....  In other news- Kevin got done talking to Austin who said that his \"friend\" has another apartment and that he hasn\'t been staying here for 2 weeks, but that he would introduce him and that he would do the dishes today(sunday).  I haven\'t heard from cliff or hannah since thursday. :( Hopefully they\'re on their way home...  '),(18829,317581,'15,September,2001','2001-09-15 00:00:00','   I have a problem talking to girls?  Where did that come from?  Oh and you forgot, it\'s CAhlen to you Tim.  '),(18830,317581,'15,September,2001','2001-09-15 00:00:00','   argh, confrontation!!!!  CAhlen and kevin and I finally confronted austin about his turn to do the dishes, but now it\'s our fault for ganging up on him. Also, Kevin is now confronting him about his roommate.  more later...  '),(18831,317581,'14,September,2001','2001-09-14 00:00:00','   I don\'t know...  So CAhlen has this issue with talking to girls that I think we should confront for him. An intervention if you will:)   Although, I\'m not one to talk. I mean I talk to females fine, and all that, I just can\'t start a \"relationship\" with one.  Anyway, it\'s time to go out to see rockstar, sorry for giving you a bad time CAhlen:)  '),(18832,317581,'14,September,2001','2001-09-14 00:00:00','   Ok, now for life on the lighter side, just for a moment, as I\'m getting sick of getting these stupid hints from this dumb crush thing on email (see the archives if you missed the earlier rant). I\'m enlisting all of you to help me out, as I don\'t know who to guess, and I only know of 3-4 females who actually know my email address (2 of whom are related to me, and the other 2 didn\'t do this crush thing). Here are the hints:  1. the first name is 5 characters long. 2. the last name is 4 or fewer long. 3. the first name begins with a letter between a and e 4. the email username is 8 or more characters long 5. the email account is not with hotmail.com There. Any ideas? You\'ll get a special bonus if you can id them.  '),(18833,317581,'12,September,2001','2001-09-12 00:00:00','   I saw a dude at the entrance to the freeway waving and holding a sign saying \"I don\'t hate Arabs\"  Yup  '),(18834,317581,'12,September,2001','2001-09-12 00:00:00','   You gotta love the crazy people who think they need to attack anyone who looks like they might be islamic. I\'m speaking mostly of the dumb idiot who attacked a mosque in Seattle yesterday. Why can\'t stupid white americans realize that just because someone is a muslim it doesn\'t mean that they\'re a  terrorist? Just like back in the 40\'s when just because someone was a japanese-american didn\'t mean they were all spies for Japan? Hello? Are we all still that stupid?  '),(18835,317581,'31,October,2001','2001-10-31 00:00:00','   so trav, I understand all the girls love you, but you\'re blind to it...  lucky bastard:)  '),(18836,317581,'30,October,2001','2001-10-30 00:00:00','   umm... home for a while. Austin is moving out. Need to find another job...  Life is a dream, and then you wake up, and you\'re dead  '),(18837,317581,'30,October,2001','2001-10-30 00:00:00','   I think John Denver said it best when he said...  \"Friends I will remember you, think of you Pray for you And when another day is through I\'ll still be friends with you\"  I just want to give a shout out to all my friends and tell you that you\'re awesome...  '),(18838,317581,'29,October,2001','2001-10-29 00:00:00','   waiting... Now I\'m going to a thing with Cliff tonight with his church, but it doesn\'t start till 7:30. So now, I\'m waiting on the westside, and all my peeps are not home. Maybe if I drive to SPU I will find em, but Dej didn\'t answer her phone, and she\'s the only one I know left on the campus...  hmm...  '),(18839,317581,'29,October,2001','2001-10-29 00:00:00','   So Mary- Yes, I\'ve read LOTR and the Hobbit. (Nice reference. Although, precious was more about the good person you want to be, not a ring so much) I dunno who MzJaxon is, but they are a mary on the internet that is stalking me.  Everyone Else- Isn\'t it weird how we all put up a front? I mean, we can\'t walk around being sensitive all the time. Like I really enjoy museums and especially paintings, but if I did what I wanted to, sit and stare at a painting all day, whoever I was with would get bored and leave. And so I move on in a somewhat quickened pace and don\'t allow myself to want to look at the painting. I think I\'m using this as a metaphor, U long for a deep trusting relationship, but you don\'t really have the time, and you don\'t know if the other person wants the same thing, so you skirt the issue... You play that you\'re not that interested then, if it comes up and they don\'t like you, then you can say, \"well, I didn\'t feel much for you anyway\", when in fact, you wanted to badly, but disallowed yourself to feel.  Everyday, I feel as though I could live like the kid in American Beauty, finding just the most mundane things beautiful, and it\'s cause I can. However, I find living in such a state of awareness, exhausting. The emotional taxing would be so great, and I\'d probably cry all the time.  Well, gotta run, keep it real  '),(18840,317581,'29,October,2001','2001-10-29 00:00:00','   So evidentally I\'m sweet now.  At least according to the last 2 girls I asked out. Well, I guess I can\'t blame them, they are not in my time zone. I guess it would be hard to take me seriously.  '),(18841,317581,'28,October,2001','2001-10-28 00:00:00','   So I\'m still recovering from this weekend. I haven\'t stayed up so late on consecutive days for a couple of years, basically, so it\'s taking me a while to get back to normal. But I\'ll survive... and it was such a good weekend of hanging out with everyone, and doing a little reminiscing. Maybe we should do this more often.   Anyhowie... 2 more days of work. I really need to find a job soon.  '),(18842,317581,'28,October,2001','2001-10-28 00:00:00','   Busy weekend... Steve left for Belize. SAw  urlLink the carolines  on yesterday. These guys rock and the lead singer lived on my floor in college:) HE\'s a fun guy.  OK, well, hopefully I\'ll be more writingful next time.  Sorry for missing your shindigs trav. Hope you had fun.  '),(18843,317581,'27,October,2001','2001-10-27 00:00:00','   so I found the real  urlLink Cowboy Bebop site   and then Manga has a link to  urlLink this site  which shows manga shit on the net:)  '),(18844,317581,'26,October,2001','2001-10-26 00:00:00','   so, in case you were wondering, I didn\'t go to said party last night.  I chilled with Big A at the house instead, had to run to the bank too. Sorry Trav, we\'ll have to meet some other time. Hope your 21 is all that you could ever want:) Especially if all you want is to legally buy alky:) Cause then you got your dream:)  also, I was thinking about something to write on blogger the other day, but then I forget what it was. Maybe I\'ll remember it.   There was something about masturbation, but I didn\'t know if it were appropriate to talk about that here;) Actually I just figured no one would want to talk about it.  So, as we move on... The question of the day is, \"who is mzJaxon?\" if you can tell me the answer, I\'ll write \"her\" back. If no one can, she\'ll just have to keep emailing me with no replies:(  '),(18845,317581,'26,October,2001','2001-10-26 00:00:00','   Also, Mary, why (a.k.a. \"Gollum\")?   How did I get this nick?  It\'s the first time I\'ve seen it. Also, are you mzJaxson? Ewww.. I am for real. Never meant to make your daughter cry. I apologize a million times.   lalala  '),(18846,317581,'25,October,2001','2001-10-25 00:00:00','   so, I\'m supposed to go to Trav\'s party tonight, but I don\'t have anyone to go wif. C\'est la vie. I know if marc were here he\'d go with me:( Maybe I\'ll get my ass in gear, but something tells me I\'ll prolly just chill with the roommates like I do everynight. Not that that\'s bad, but they don\'t put out;)  Actually, I love you guys, ttyl  '),(18847,317581,'25,October,2001','2001-10-25 00:00:00','   so, I\'m super productive Tim today, I think I got almost all my stuff blasted through at work today!!!  '),(18848,317581,'24,October,2001','2001-10-24 00:00:00','   ok, they XPised,  urlLink msn.com   and now, (french?) urlLink bebop   I think it\'s the apocalypse  '),(18849,317581,'24,October,2001','2001-10-24 00:00:00','   YAY!!! Internet speeds are back up at the house:)  Also, I am feeling better today, maybe I just needed an extra 5 hours of sleep.  Also, kevin and I started watching \"The Sopranos\" last night, so far I give it 2 thumbs up.  BTW kevin, Don\'t forget that you want to buy Momento:)  '),(18850,317581,'24,October,2001','2001-10-24 00:00:00','   I feel I should say something, but words are eluding me.  I cannot articulate myself at all.  My misunderstandings cannot be perceived.  chaos ensues.  '),(18851,317581,'23,October,2001','2001-10-23 00:00:00','   Nice poem Tim.  '),(18852,317581,'23,October,2001','2001-10-23 00:00:00','   What a lazy day off. I didn\'t really do much today. Played some  urlLink Red Faction , messed around on the internet a bit, cleaned up my room a little, and stuff like that. Nothing big, that\'s for sure. I did get  urlLink Musicmatch Jukebox  installed on my machine, so now I can make mp3\'s and burn mix cd\'s. Woo Hoo! I also picked up the new Incubus cd, but haven\'t had a chance to listen to much of it yet. What I have heard sounds pretty good, but I\'m not sure what songs I like on it yet.   I just got done watching High Fidelity. That\'s a pretty good movie. I would say more, but then I\'d have to dissect my thoughts too much, and I\'d write a few pages instead of a few sentences.  '),(18853,317581,'23,October,2001','2001-10-23 00:00:00','   I came home sick today, slept from 2:30 - 7:30, sorry kelley I won\'t pass the quiz on \"passions\", I was asleep, doing better now, ttyl  '),(18854,317581,'23,October,2001','2001-10-23 00:00:00','   Random fun quote of the day, courtesy of Roger Ebert:  \"Being eaten by a raptor ranks high on my list of inappropriate touching.\"  '),(18855,317581,'23,October,2001','2001-10-23 00:00:00','   It\'s sad when the house you have high speed internet in is slower than your shared 768 connection at work.  '),(18856,317581,'23,October,2001','2001-10-23 00:00:00','   also, I don\'t think you could be addicted, but curiousity did kill the cat, I remember when I was like 20 and I would have cravings for beer, and the funny thing is, I hadn\'t had a beer since I was like 4. My mom always used to tell me to try the \"brussel sprouts\" cause I might have grown to like them since last time, and although I\'ll never \"grow to like\" brussel sprouts, I think I did grow to like beer. weird moms  '),(18857,317581,'22,October,2001','2001-10-22 00:00:00','   I\'m in a funk, can anyone set me free?  '),(18858,317581,'22,October,2001','2001-10-22 00:00:00','   A loss of innocence  I sit in my corner waiting, I will protect my precious, you cannot see him, he is precious, If you approach, I will surely kill you, Do not get too close, my claws are sharp, my mouth is hungry, You would destroy him, but you cannot, I will not let you, he is my precious and you cannot hurt him,  In the coming darkness, the voices cry out, the hands reach, you want my precious, you will take him for your own, your claws reach and pull, I struggle, I lash out, I hit, I bite, Yet precious is drawn to you, you are no good, but I struggle,  My precious, he is delirious, He wants to go, you  will sure destroy him, you cannot have him,  It is a struggle I cannot win, you will feast on my precious, he will be dead  '),(18859,317581,'22,October,2001','2001-10-22 00:00:00','   hey guys-  urlLink check this out  '),(18860,317581,'22,October,2001','2001-10-22 00:00:00','   So, I\'m contemplating teaching english in Japan, yet again. I don\'t know if I really want to do it, or am just looking for something completely different to do with my life. At any rate, I do know that one group will be interviewing in Seattle in December, and so I have til mid November to make up my mind.   I get to sleep in tomorrow, as I don\'t have to work. I\'m down to 5 days of employment. I wish there were other projects hiring right now, but everything pretty much shuts down in the winter. That\'s life in the game industry.   Why does everything have to be on sale when the economy sucks? It\'s like, hey, everyone\'s out of work, let\'s have sales on things they can\'t afford to get anyways. Like people are going to be able to buy something when it\'s 10% off as opposed to regular price. Don\'t stores know that they don\'t have the money anyways? I guess it\'s great for those with job security.  '),(18861,317581,'22,October,2001','2001-10-22 00:00:00','   argh, another day another drachma...  got the call from mom tonight, no turntable for bday:(  '),(18862,317581,'22,October,2001','2001-10-22 00:00:00','   No matter what time I leave my house, it seems like I can\'t get to work late. I even stopped at 7-11 this morning and still made it here on time. Oh well....  Yesterday went by without a hitch and I even have more stuff to do today. I brought in my music this morning, so I\'m listening to a little D right now:)  Also, I got to eat at Pete\'s Pizza last night which was fun, good company.  and another thing, do you think you could get cravings or be addicted to drugs you\'ve never had? just a contemplation...  '),(18863,317581,'22,October,2001','2001-10-22 00:00:00','   In the immortal words of the wallflowers, \"God don\'t make lonely girls\"  '),(18864,317581,'22,October,2001','2001-10-22 00:00:00','   So the Seattle Mariners showed what they are made of in the play offs yet again. I\'m kinda bummed, but I also kinda expected this. The torture continues for the lifelong fan.   In other news... I\'m still employed for about a week. I am trying to find something to do that is remotely computer/software testing related, but there aren\'t a whole lot of jobs out there right now.  It\'s nice to get off work after only 8 hours though. wow. Who knew what the world looked like when there was still a little sunlight out?  '),(18865,317581,'21,October,2001','2001-10-21 00:00:00','   working working working at least at my other job the day flew by... oh well, back to the normal grind, at least I have work to do:)  '),(18866,317581,'21,October,2001','2001-10-21 00:00:00','   maybe as a drunk I am not very becoming... In other news, I\'ve worked on  urlLink unwelcome hero  and it\'s now up for public preview. hope you all enjoy it:)  '),(18867,317581,'20,October,2001','2001-10-20 00:00:00','   I am going to have to agree with Deja, Tim:  Don\'t post when you are drunk, it is not very flattering to your personality.  '),(18868,317581,'20,October,2001','2001-10-20 00:00:00','   I toyed with the idea of a comments section, but it slowed the whole site down... sorry for all you mystery readers, if you want to get in touch, you\'re just gonna hafta email us...  '),(18869,317581,'19,October,2001','2001-10-19 00:00:00','   random quotes I\'ve heard in my life: \"Either it\'s hot in here, or I\'m going through the pause again\" \"I like to sit on cracks\" \"This one time at band camp\"  Oh wait that last one was from a movie, not really my life...  Hmm... how to get drunk tim not to post... this will be the trick I think... Personally I think he likes to post, he feels that he\'ll say what I won\'t say... weird drunk tim...  In other news, I think we could say we have different selves for different people we\'re around, but also reoccuring themes running throughout each self. IE, in all my selves I believe in God and respect my elders. But in some selves I\'m more respectful of women and strangers than I am as other selves. When I write on blog, I think that\'s one self, but it\'s also whichever self I am at the time. Does this make any sense? Does anyone understand me?  '),(18870,317581,'19,October,2001','2001-10-19 00:00:00','   Psycho time. I\'m away from bloggs for one day, and come back to this..... Tim, I REALLY recommend you NOT make posts when you\'re drunk. It\'s not very flattering to say the least.   And now it\'s my turn to complain about something. I really hate it when I write a huge long email to one of my family members and they reply, three days later, with an email one sentence long. So much for attempting to facilitate a flourishing correspondence....  '),(18871,317581,'18,October,2001','2001-10-18 00:00:00','   Also, there\'s nothing to do right now, I just called amidy but she is driving her mom to the airport, but not to see Jebe. Anyway, maybe I won\'t be bored later. Also, i have added  urlLink DAMN YOU MARIA  to my vocabulary. Psycho Tim...  '),(18872,317581,'18,October,2001','2001-10-18 00:00:00','   So i think i may have figured out what to do with unwelcomehero.com  Make it a tim\'s poetry corner/other writings and thoughts opinions and shit... Now I just need to start programming it:)  Well, check you guys later... ps, my hair is back to brown/black.  '),(18873,317581,'18,October,2001','2001-10-18 00:00:00','   ARGH!!! My head hurts, damn headaches!!!  '),(18874,317581,'18,October,2001','2001-10-18 00:00:00','   Also, drunk Tim can\'t fuckin type, and he rambles on incoherently(sp?)  I don\'t know why I even let him on here. Or how he can get the fuckin password right? sorry if I offended anyone last night, but michelle from american pie 1 and 2 is a hottie. She ALMOST makes me want to watch buffy... almost....  '),(18875,317581,'17,October,2001','2001-10-17 00:00:00','   hmm... I was trying to think of a way to get a  urlLink live journal  account and then realized I don\'t have any other friends that post to a live journal. Just Sunday hero, and I don\'t think she wants to talk to me anymore... :( Well, if anyone who reads this site, has an access code, I\'d love to have it, so I could join and maybe \"publish\" some of my written words on it. (of course when I say \"written words\", I mean my poor excuse for poetry:)  check you guys latah  '),(18876,317581,'17,October,2001','2001-10-17 00:00:00','   I just wanna know how many people actually click on the \"I wanna cum on your face\" emails that they get on hotmail? Cause evidentally someone is, otherwise why would they continue to send the emails? Fortunately I set up a filter so I don\'t actually have to see the shit, I can just delete it:)  '),(18877,317581,'17,October,2001','2001-10-17 00:00:00','   :(  '),(18878,317581,'17,October,2001','2001-10-17 00:00:00','   Also, I\'d have to agree with Trav on this one, CAhlen, WHAT THE FUCK?!?!?!  '),(18879,317581,'17,October,2001','2001-10-17 00:00:00','   So, I\'m in Oregon right now. And Kevin\'s in NY. Talk about a house divided:) Although I don\'t think that\'s what they were talking about...  In other news, everytime I go use the urinal at the Woodinville theater I notice the cake holders have a website on them. And I always mean to mention it to you guys, cause I thought it was really funny. Like, hey, the thing I\'m pissing on has a website. WHAT THE FUCK!!!   Anyway, I\'m gonna kill off unwelcome hero as pierre\'s site, but I will hopefully do something useful with it. Do you guys have any suggestions?  Also, do you think you can see your parents as other people see them? or is there a parent filter there... weird...  prolly won\'t be able to write again till sunday cause I\'m at home so I\'ll talk to you guys latah  '),(18880,317581,'17,October,2001','2001-10-17 00:00:00','   I want to be a  urlLink 1337 h4X0r  like that guy.  '),(18881,317581,'17,October,2001','2001-10-17 00:00:00','   sometimes, I get the impression that Cahlen likes Henry David Thoreau....  weird how I would get that impression...  '),(18882,317581,'17,October,2001','2001-10-17 00:00:00','   \"What men call social virtue, good fellowship, is commonly but the virtue of pigs in a litter, which lie close together to keep each other warm.\" - Henry David Thoreau (1817 - 1862)  '),(18883,317581,'17,October,2001','2001-10-17 00:00:00','   \"The man I meet is seldom so instructive as the silence which he breaks.\" - Henry David Thoreau (1817 - 1862)  '),(18884,317581,'17,October,2001','2001-10-17 00:00:00','   \"That man is the richest whose pleasure are the cheapest.\" - Henry David Thoreau (1817 - 1862)  '),(18885,317581,'17,October,2001','2001-10-17 00:00:00','   \"No way of thinking, however ancient, can be trusted without proof.\" - Henry David Thoreau (1817 - 1862)  '),(18886,317581,'17,October,2001','2001-10-17 00:00:00','   \"It takes two to speak the truth--one to speak, and another to hear.\" - Henry David Thoreau (1817 - 1862)  '),(18887,317581,'17,October,2001','2001-10-17 00:00:00','   \"I never found the companion that was so companionable as solitude.\" - Henry David Thoreau (1817 - 1862)  '),(18888,317581,'17,October,2001','2001-10-17 00:00:00','   \"City Life.  Millions of people being lonesome together.\" - Henry David Thoreau (1817 - 1862)  '),(18889,317581,'17,October,2001','2001-10-17 00:00:00','   it\'s 12:50 am , and yet I\'m still drunk:)  This one time at band camp, I stuck a flute in my pussy:)  Say my name bitch:)  '),(18890,317581,'17,October,2001','2001-10-17 00:00:00','   JEBE- you must - soon or we\'ll think you meant 3.14 :)  Also, the drunk tim is crazy:)   sometime he\'s just like I wanna fuck you:) but then there\' s this girl Michelle:) She\'s so fucking hot!!!  maybe tomorrow I\'ll be okay, I doubt it:)  also, SUCK MY COCK!!! (tomcats reference:))  '),(18891,317581,'17,October,2001','2001-10-17 00:00:00','   sometimes I get the feeling cahlen is trying to get my password again.but sometimes when cahlen get\'s nervous he sometimes wets his pants. OK that\'s not true, but at least I get to watch american pie:)    BTW- I am drunk, in case you were wondering:) Also, at least I didn\'t open a roommate\'s mail:)  '),(18892,317581,'17,October,2001','2001-10-17 00:00:00','   mmm.... I think the numb self is the bestest.  We watched \"American Beauty\" and now we\'re watching \"american pie\" which is always fun:) We all get laid before we graduate:)  mmmm.... laid:) BTW- this movie kicks ass:)  \"no longer will our penises remain flacid and unused\" :)  drunk tim still is here:)  '),(18893,317581,'16,October,2001','2001-10-16 00:00:00','   hello this is drunk tim again  I\'ll talk more again latah, hopefully rachael will be on, cross my fingers:)  '),(18894,317581,'16,October,2001','2001-10-16 00:00:00','   To answer Tim\'s perception question, yes, I think people can change their perceptions of others or the world. As you have experiences and gain knowledge throughout life, your views/perceptions of things will change. Now, changing how others perceive you can be kind of difficult, especially people who have known you for a while. They\'ll pretty much always perceive you as the way they\'ve known you for however long.   What/which of your perceptions were you planning on changing?  '),(18895,317581,'16,October,2001','2001-10-16 00:00:00','   You\'re an i-i-island.  '),(18896,317581,'16,October,2001','2001-10-16 00:00:00','   Hey kev- I guess, I prolly use movies, video games, and tv shows for my basis of argument because I somewhat believe that they are our modern day story tellers. I mean, Jesus/Mohammed/homer, didn\'t have the written word, and so they told parables/stories. I think that the forms of media that I use to convey a point have taken the place of story tellers. True, I prolly give em more credit than they are due, but I think one could make some assumptions about civilization based on the top grossing films. I also believe that music is a great persuader, maybe I should start using music to discuss my points:)   All-in-all I\'m not trying to argue this with you, as I didn\'t take offense, just thought you might wonder why I argue the way I do sometimes.   Although, I\'m just a rock, or was it an island?  '),(18897,317581,'16,October,2001','2001-10-16 00:00:00','   also, jenny, when I read your post, I was like, \"How on earth is she ever going to = pi, I mean how many years have people been working on pi, and still not been able to figure it out? And what the hell does this damn software have to do with the circumference of some damn circle?\"   But then I asked kevin and he was like, she means private investigator.  '),(18898,317581,'15,October,2001','2001-10-15 00:00:00','   moving on to perception...  I was just wondering what you guys thought about, whether one could change their perception. I know that people can get some pretty messed up perceptions on life, and I was wondering if one might change it.   I was thinking of changing mine the other day, and was wondering if I should/could. reinvent oneself? change of perception? mpd?   I dunno... what do you guys think?  '),(18899,317581,'15,October,2001','2001-10-15 00:00:00','   So we\'re all just a little crazy. Or maybe Tim\'s just had experiences with total psychos :)  Anyhowie... such a fun day of work. Lack of communication/mixed signals or something just left me feeling weird about when I left work today, so hopefully I don\'t get in any trouble tomorrow. But if I do, it\'s not like I\'m around much longer, as my contract is up in 2 weeks. I hope I can find some employment normalcy soon, so that area of my life would be stable so that pretty much everything else stays stable. Ah, the joys of debt.   And Cliffy, that was an interesting anti-Clinton post you had there. Since you introduced politics into the blog, just remember to vote for me, Big A, for President of the United States in 2008. I can\'t be any worse than the others available to vote for :)  '),(18900,317581,'15,October,2001','2001-10-15 00:00:00','   ain\'t that the truth cahlen, sometimes I\'ve been a fuckin idiot too  '),(18901,317581,'15,October,2001','2001-10-15 00:00:00','   I think when Tim says crazy he means, like you said, we think differently and that\'s cool.  Once upon a time I was a stupid little kid and I fell in love with this girl.  Then I dumped her because she was \"weird\", but now I realize that\'s what I loved about her...  I\'m a fuckin idiot....  '),(18902,317581,'15,October,2001','2001-10-15 00:00:00','   hahaha, so I had this whole thing written out, then blogger puked on me:P  It\'s not that I find women are crazy all of the time. In fact most of the time they are perfectly rational. and I don\'t think this is limited to females. Men can be just as crazy sometimes too. Like in this past week\'s episode of \"enterprise\", when the only rational person on the planet after the pollen got stirred up, was a girl. I mean given, she was a vulcan, but the men were just totally crazy.  Like Cahlen said, \"you would not believe how little it takes for a persons mind to snap\" or something along those lines:)  Well, that\'s pretty much it. I can\'t get on MSN today:(  '),(18903,317581,'15,October,2001','2001-10-15 00:00:00','   I found this quote mildly entertaining...   \"...men have a reduced mental capacity that warps their perception of reality and ability to reason. They have a terribly deluded sense of self-importance....\"   -Bradley Trevor Greive  '),(18904,317581,'15,October,2001','2001-10-15 00:00:00','   So, you think women lack the gift of rational reasoning. And men are sooo good at it. Well, all I can say is that men and women percieve things differently, so naturally, they disagree on what is an overreaction and what is not. And I know that some girls are worse at \"overreacting\" than others, just as some are more emotional and flighty than others, and some are more emotionally balanced. But please don\'t say that women lack rational reasoning. It\'s just that sometimes their rational reasoning is overrided by their emotions.... like Captain Kirk.....  '),(18905,317581,'15,October,2001','2001-10-15 00:00:00','   hmm... crazy... how can I think of crazy... I guess my feelings on crazy are lacking of rational reasoning. not to say that one\'s feelings aren\'t rationale, sometimes they aren\'t though. And usually I\'ve found it to be true mostly with girls. but the fact that he asks if it\'s that time of the month is bad. The fact that he doesn\'t argue with you means he probably doesn\'t think it\'s that big of a deal to argue about, but he shouldn\'t blow you off by asking if it\'s \"that time of the month\". So, maybe what I\'m refering to as \"craziness\" is overreacting to the smallest of things.  I dunno.  Also, it\'s 11pm and I\'m the last one up. Weird.  PS - Weezer, Tenacious D, and Jimmy Eat World on November 15th. I\'ll be there, will you?  '),(18906,317581,'14,October,2001','2001-10-14 00:00:00','   So, what do you mean by saying that girls are \"crazy?\"  I know that girls in general have this reputation of getting all emotional and think that their boyfriends \"should be able to tell/know what is bothering them\" or whatever. Is that what you mean by crazy? I think it\'s just that you just don\'t understand how we think. Or maybe you do understand how we think and you think it is \"crazy.\" I know a guy (I won\'t mention any names, and no, it\'s not you Tim or Cahlen) that discounts anything I say when I\'m a little upset because he thinks I\'m making too big a deal out of such an unimportant issue... most often he asks if it\'s \"that time of the month,\" as if that could discount the validity of what I\'m saying to him. I have no idea whether or not any of this is making any sense to you who happen to be reading this blogg. Maybe I am just crazy. Or maybe it\'s just time to go to bed.  '),(18907,317581,'14,October,2001','2001-10-14 00:00:00','   THINK ABOUT IT!  After the 1993 World Trade Center bombing, which killed six and injured 1,000; President Clinton promised that those responsible would be hunted down and punished.   After the 1995 bombing in Saudi Arabia, which killed five U.S. military personnel; Clinton promised that those responsible would be hunted down and punished.   After the 1996 Khobar Towers bombing in Saudi Arabia, which killed 19 and injured 200 U.S. military personnel; Clinton promised that those responsible would be hunted down and punished.   After the 1998 bombing of U.S. embassies in Africa, which killed 224 and injured 5,000; Clinton promised that those responsible would be hunted down and punished.   After the 2000 bombing of the USS Cole, which killed 17 and injured 39 U.S. sailors; Clinton promised that those responsible would be hunted down and punished.   Maybe if Clinton had kept his promise, an estimated 7,000 people in New York and Washington, D.C. that are now dead would be Alive today.   AN INTERESTING QUESTION:  This question was raised on a Philly radio call-in show.  Without casting stones, it is a legitimate question.  There are two men, both extremely wealthy.  One develops relatively cheap software and gives billions of dollars to charity.  The other sponsors  terrorism.   That being the case, why is it that the Clinton Administration spent more money chasing down Bill Gates over the past eight years than Osama bin Laden?    THINK ABOUT IT!  It is a strange turn of events. Hillary gets $8 Million for her forthcoming memoir.   Bill gets about $12 Million for his memoir yet to be written.  This from two people who have spent the past 8 years being unable to recall anything about past events while under oath!  Incredible!  THINK ABOUT IT!  '),(18908,317581,'14,October,2001','2001-10-14 00:00:00','   We\'re going to Ibiza:) Not really, just listening to muzak.  in other news, running was ok this morning. Gotta call the parentals today and see about $$:(  Trav: tru dat, tru dat, I don\'t know why the older ladies like you, maybe they\'ve realized how a good guy could be, and they got stuck with jerks? :) just a thought.  also, I just got this link from JEBE about  urlLink free mammograms  and you should all click on this. You know I\'m always in favor of saving breasts... umm... women.  '),(18909,317581,'14,October,2001','2001-10-14 00:00:00','   for those ladies in my life, and those reading this blog... I\'ve discovered something in my 23 some odd years of life...  Women are crazy. This is not one of those sometimes, some of em, it\'s pretty much all of you:)  Not to say this is bad. In fact how could it be? I mean if it were all bad no one would hook up. The thing I\'ve learned is that if you\'re a guy, you just need to find the girl who\'s level and type of craziness is compatible with yours. I mean, there\'s gonna be times, guys, when you\'ll be like, umm... what are you talking about, and she\'ll be like, you should know. And sometimes you gotta remember that no matter what she does, this might be the least crazy girl you\'ll ever meet.   I remember some times when different guys in my past have been like, \"Tim, I dunno if I can put up with this craziness\" and to be honest, I talk bad about girls sometimes, expecially when they interfere on guy time. (IE when I want to hang out with my guy friends, but they have to go spend time with their girlfriends) But I just tell em, \"Look at this wonderful girl you got. Sure she\'s crazy sometimes, but they ALL are. You got a good one my friend\" OK, so that happened like once, but I can pretend it happened more than that:) And no, don\'t worry ladies, it\'s no one that reads this blog;)  Well, check you latah, I\'m getting up in 6.5 hours for running:) YAY  '),(18910,317581,'13,October,2001','2001-10-13 00:00:00','   Yay junk email. I just love getting spammed in my old email account. Of course, everything sent to that account is spam, for the most part, since I don\'t really use it anymore. At least I have a constant means of email though. Kinda like getting junk mail in the actual postal mail. It makes you feel almost alive, or at least that some junk mail list cares that you still exist.   Anyhowie... yes good old slow weekends. I got to work yesterday, then I just kinda vegged the rest of the weekend. I did change some of my room around, and hook up the neighbor-piss-off-anator (aka the home theater subwoofer, at least I turned it way down).   I need to win a million dollars. Or 100,000. That\'d be nice too.  '),(18911,317581,'13,October,2001','2001-10-13 00:00:00','   hmmm... guess it\'s just back to me.Now for a story about Tim and \"Melissa\"...  A long time ago (aka a few years ago) Tim was just doing the school thing and what not, when he got this random email from \"melissa\". She said she had found my site on the school\'s server and wandered if we could be friends. Tim was like \"sure\" but then he asked \"melissa\" if she had a webpage(after a few emails back and forth) and she was like sure... and then she sent me this page about this person that had a sex change operation(not that those are bad, it just really shocked me) And then there was a delay in writing and she got all like, \"did I scare you?\" and I was all like \"no\". But then later my friend sent me a link with our little conversation written out on the internet for everyone to see. Up until that point, I had no reason to believe \"melissa\" was not who she said she was. But then my \"friend\" had just been playing me. So now if I get contacted from people on the internet, I get all paranoid.    Besides that, some of my real life \"friends\" do not like to contact me personally, they prefer to hide behind an AIM alias which has no real world connection. Unfortunately since they won\'t tell me who they are, I am forced to block them:(  C\'est la vie....  '),(18912,317581,'13,October,2001','2001-10-13 00:00:00','   mmm... kinda a slow weekend on the blogger anyway. Spent saturday cleaning the house(much needed).  Oh and I drank a little. But nothing I couldn\'t handle:( Steve\'s gonna help me drink next week I think. Watched \"The way of the gun\" and \"O brother where art thou?\". I still love O brother, and the way was pretty good.  Definately worth the 9 bucks;)  I NEED CAFFEINE. ALSO, just got a phone call from someone cool;)  '),(18913,317581,'13,October,2001','2001-10-13 00:00:00','   I guess what I\'d really like, is someone to take care of me, so I don\'t have to have a job:)  (\"Check out \'Why do I need a job\' by Bare Jr. if you get a chance:)\")  '),(18914,317581,'12,October,2001','2001-10-12 00:00:00','   Yeah, Jack Kerouac was a beatnic...  '),(18915,317581,'11,October,2001','2001-10-11 00:00:00','   Jack Kerouac is dead.  Do you even know what a  beatnic was?  '),(18916,317581,'11,October,2001','2001-10-11 00:00:00','   dude CAhlen, what\'s up with that? there wasn\'t any earthquake. What the fuck was it?  '),(18917,317581,'11,October,2001','2001-10-11 00:00:00','   Jack Kerouac is a beatnic...  I wanna be a beatnic....  Beatnics are cool....  '),(18918,317581,'11,October,2001','2001-10-11 00:00:00','   So I just had the best time of my life ever... yes, EVER!!!    Imagine this, I\'m driving onto the 520 bridge in horrible traffic with cops and ambulances passing us and I\'m like frick, this is gonna freakin suck.  And all of a sudden my car starts shaking back and forth like freakin crazy, and I\'m like, \"Eh?\"  Then I look at the other side of the bridge and it is swinging all over the place, and then these huge ass waves shower over my car and I\'m like \"Eh?\"  So I sit there going 2 miles per hour and I\'m like, \"This freakin rocks!\"  So I turn on some Coldplay and sit there swaying back and forth like I\'m in a 8.0 earthquake for atleast 30 minutes, and my car is getting drenched, and I\'m just chillin\' thinking, \"This is the coolest moment of my life...\"  And then I got sea sick and almost threw up and I was like, \"Freakin\' A, I gotta get off this bridge...\"  '),(18919,317581,'11,October,2001','2001-10-11 00:00:00','   I want to go to Tahiti:)  '),(18920,317581,'11,October,2001','2001-10-11 00:00:00','   2:30 is waaaaaaay to late, when you have to wake up by 8:(  '),(18921,317581,'11,October,2001','2001-10-11 00:00:00','   The one thing we yearn for in our living days, that makes us sigh and groan and undergo sweet nauseas of all kinds, is the rememberance of some lost bliss that was probably experienced in the womb and can only be reproduced at death.  -Jack Keouac  I think that\'s why I like to take baths....  '),(18922,317581,'11,October,2001','2001-10-11 00:00:00','   and now for another quick one, (still to be refined)  Tears of Heaven- sweet irritation, cleansing rawness, chapped skin,  fun splashing in puddles, rainbows, depressing gray, one moment, sad, depressed, irritable,  the next, joy, gaiety, exuberant splendor, renewal, life for all,  power, destruction, death  '),(18923,317581,'11,October,2001','2001-10-11 00:00:00','   Note: that was a first run through. So if it sucks, it\'s cause it hasn\'t been revised yet:)  '),(18924,317581,'11,October,2001','2001-10-11 00:00:00','   and now for a poem by Tim:  Conversation with my best friend- How to explain this? I think about you everyday. Itï¿½s been a year since you were taken from our lives. We were supposed to go out with Michelle and Tracy that night. We were going to see a movie. I remember how you were going to talk to Tracy about how you liked her. But that never happened. I remember walking to school and thinking how that accident seemed really bad. I remember wondering where you were. Then the principal came over the intercom and everything stopped.   From that moment there has not been a moment that I donï¿½t miss you. I remember how we used to just chill and play videogames. How you would smile all the time. I remember how good you were at sports. You wouldï¿½ve made varsity this year for sure.  Itï¿½s been one year since you were on your way to school and that lady hit you coming around a corner too fast. Sheï¿½s off paying her debt.   But what about you? There is no coming back. No homework to distract you. No person to call when I have a bad day. No one to talk with about her and what she might mean. I can feel Iï¿½m losing you. Iï¿½m sorry.  '),(18925,317581,'11,October,2001','2001-10-11 00:00:00','   Yeah, it\'s kinda quiet around here now, isn\'t it?  Lame.  '),(18926,317581,'10,October,2001','2001-10-10 00:00:00','   Well... Tim\'s not the only one losing a job soon. The project I\'m working on ends on halloween. Woo hoo! oh wait, that means I\'m unemployed then too. Now if I could just find something a little more permanent to do, I\'d be pretty happy. That or winning a million dollars.  '),(18927,317581,'10,October,2001','2001-10-10 00:00:00','   hmm... blog is boring w/o people posting:)  BTW, I\'m in one of those \"wish my job wasn\'t ending in a week/need money/cool new friend/tired/want home cooked food\" moods  '),(18928,317581,'10,October,2001','2001-10-10 00:00:00','   For those who haven\'t heard... I\'m oughta a job on wednesday. I still have second job, but I prefer to have some sort dev job that\'s fun:)   Well, I\'m working on that poem... At least trying to:) Also, if you know of any job opportunities coming up, send me an email and I\'ll send you a resume:)  '),(18929,317581,'10,October,2001','2001-10-10 00:00:00','   \"The only really happy folk are married women and single men.\"  - Mencken  '),(18930,317581,'10,October,2001','2001-10-10 00:00:00','   So, I wrote this whole blogg one time already, and the dumb thing \"couldn\'t establish a connection.\" And then it was all gone. So, I spouted obscenities to the computer and I\'m typing it again.   I found this quote in the margins of my psychology book.   \"I contend that nature has given males the heaviest burden of all: the burden of always having to Make the First Move, and thereby risk getting Shot Down. I don\'t know WHY males get stuck with this burden, but it\'s true throughout the animal kingdom.... It\'s always the male bird who does the courting dance, making a total moron of himself, while the female bird just stands there, looking aloof, thinking about what she\'s going to tell her girlfriends. (\'And then he hopped around on one foot! Like I\'m supposed to be impressed by THAT!\')\"  Acutally, in some species, it is NOT the males birds that have to do the song and dance. But in biology, there are always exceptions.  '),(18931,317581,'09,October,2001','2001-10-09 00:00:00','   I finally checked out Manuel\'s blog today (kelley -> bro -> Man) How come he watches that WB shiznit? Oh well, he can\'t be all bad, he knows the JOYS of PEPSI:) PEPSI kicks ass from here to the moon:)  Also, blogger is run by about 20 monkeys I think. (thanks for the scale kevin)   Also, I stole  urlLink this  from Trav, sorry, at least I give you credit:) Also, Check out  urlLink trav\'s site . He find\'s the craziest shiznit and posts it there  '),(18932,317581,'09,October,2001','2001-10-09 00:00:00','   In other other news... Everytime I use the urinal at the movie theatre I\'m surprised when I see a website on the cake holder thing in there... (for those ladies who don\'t know, they put \"cakes\" of pine scent in the urinals to cover up the smell of urine) I just think it\'s funny that someone would be a webaddress on something I piss on... weird.  I hope rachael is doing better:) Thanks for sharing  '),(18933,317581,'09,October,2001','2001-10-09 00:00:00','   hmmm.... well it seems we\'ll be without cliff\'s and kevin\'s input for awhile. I just want to apologize to cahlen for seeming to gang up on him, and ask him to come home sometime. I miss him:(  Also, I got inspired for poetry today, so maybe later I can write it down and let you see it:)  gnight  '),(18934,317581,'08,October,2001','2001-10-08 00:00:00','   Rocks can\'t think. They don\'t have any neural tissue.  '),(18935,317581,'28,November,2001','2001-11-28 00:00:00','   So, here I are, midway through the week and I am exhausted. well, I got home not before midnight on monday, and not before 10:45 and I won\'t get to bed tonight before 11:30, so anyway, I gotta go...  TTYL  '),(18936,317581,'27,November,2001','2001-11-27 00:00:00','   added a poem on  urlLink unwelcome hero.com   So anyway, I was going to talk about compliments for girls and how I don\'t want to seem like a puts and tell them that they look beautiful everytime I see them, cause i think they\'re beautiful, but how they don\'t want that, so I tell them never and they don\'t like that either, but I don\'t have time, cause of how it\'s 1am.  goodnight  '),(18937,317581,'26,November,2001','2001-11-26 00:00:00','   working working working...  hehehe... blogging:)  So they offered me a fulltime pos. here, so I think I\'m goonna take it. I figure at least this way I can be all like, steady income man. And they can be like, we\'ll pay you regularly people. Plus I get benefits. Now If i could just get my paycheck from last week:P  Nyway, still figuring out the roommate shit.and have the SPU concert tonight at Benaroya(sp?) Hall. Should be fun, also have to figure out the \"tradition\" thing on friday.  Well, back to the trenches-  '),(18938,317581,'25,November,2001','2001-11-25 00:00:00','   Ok, back in WA. I love my family, but it\'s wonderful to be home;)  I miss you all, I\'ll write more later....  Marc is here and Jebe and KElley, these peeps rock my world. too bad we don\'t get to see eachother much:(  '),(18939,317581,'21,November,2001','2001-11-21 00:00:00','   KAW KAW  ok, I\'m in OR. I left at 10am this morning and arrived at 1:20 all in all a very quick trip for the day before thanksgiving. OK really though, later that same afternoon, it took me 1 hour to drive 20 miles:P  anyway, hope you are all having fun... TTYL  '),(18940,317581,'19,November,2001','2001-11-19 00:00:00','   Hey, I\'ll play...  The basics:  Full name: Aaron James Schmitt Name of school: Hard Knocks... :) How often do you check your email: Too many times Color of eyes: Blue-ish Hair: Light Brown-ish Height: 6\' 7\" Location: Monroe, WA, or as redneck hicks like to say, MawnRow Warshington Zodiac sign: Taurus   Have you ever...   Gotten drunk: No  Gone out in public wearing your PJ\'s: No Missed school because of rain: No, well, I might have used that as an excuse to skip a class or 2 in college Set any body part on fire: No Kept a secret from everyone: Yes Had a crush on a teacher: No Had a New Kids on the Block tape: Hell no.  Been on stage: Yes (crazy church productions). Gotten in a car accident: with me driving, 1 (someone else\'s fault). As a passenger many times as a kid.  Favorites:   Shampoo: Pert Plus  Conditioner: see above Town to chill: Seattle Commercial: Sportscenter and Seattle Mariners commercials Cereal: Rice Krispies, Honey Combs, Lucky Charms TV shows: King of the Hill, Malcolm in the Middle, Cowboy Bebop Movies: Almost Famous, Field of Dreams, Moulin Rouge, Star Wars Trilogy, Karate Kid, Mononoke Hime, Tonari no Totoro, Princess Bride, and many many more.  Scent: I can\'t think of one...  Music group: U2, Poor Old Lu Favorite person to talk to online: Pretty much everyone on my IM lists First crush: Don\'t remember (yes I\'m that old) What is on your walls: Movie Posters (Star Wars Trilogy, Princess Mononoke, Batman Forever Teaser, Saving Private Ryan), 2 advertisement type fans I got in Japan.  Ever been skinny dipping: No Worst feeling on earth: Pissing someone off so bad that they stop talking to you Best feeling in the world:Hanging out with good friends Word or phrase overuse: Dude, yo, like,  Get along with parents: Usually.  Bedtime: 10:30pm-11:30pm What are you wearing right now: Jeans, columbia sportsware shirt, socks, undearwear Are you lonely: Not really Are you talking to someone online: Nope (they\'re all away)  Last 24 Hours...   Cut your hair: No  Worn a skirt: No Been mean: Hopefully not  Been sarcastic: Probably Met someone new: Yeah (in the last 36 hours that is) Fought with your parents: no  First thing you thought this morning: It\'s 9:30 already?  In the future...   Where do you see yourself in 10 years: Somewhere else :)  Future son\'s name: Mike (Michael) Future daughter\'s name: Xenia (it sounds cool, and it starts with X) Where do you want to live when you get married: Hopefully in a house :)  When was the last time you...   Wished upon a star: You can do that? wow... ;) Went to the beach at night: Summer of 1998, when I was in Japan Spent some quality time with yourself: Today Read a book for fun: September-ish Believed in yourself: I believe I believe in myself currently   How long did this survey take you: a whopping 16 minutes.  '),(18941,317581,'18,November,2001','2001-11-18 00:00:00','   So \"someone\" made me promise to fill this thing out, so here it goes...  Other people can do it too, not just me:)  The basics:  Full name: Timothy Brian Williams Name of school: umm... does this even apply anymore? (SPU) How often do you check your email: about 100 times a day, someone set up a counter 1 time and I think it was right around 100. Color of eyes: brown Hair: umm... brown Height: 5\'10\" last time I looked anyway Location: Bothell or as my gay roommate liked to call it, BAWWWWWTHHHUUULLL Zodiac sign: scorpio   Have you ever...    Gotten drunk: nope, never;) Gone out in public wearing your PJ\'s:all the time Missed school because of rain: yeah Set any body part on fire: not that I can recall, but maybe before I can recall Kept a secret from everyone: yeah, at least twice Had a crush on a teacher: hell, no Had a New Kids on the Block tape: nope.  Been on stage: kinda, see there was this time when I was \"expressive\" and they had a play for the \"special\" kids (as far as I can tell) and I got to play a lead:) Gotten in a car accident: 3 times, 2 times my fault   Favorites:   Shampoo: Thermasilk Conditioner: umm... Herbal essences Town to chill: To date, I\'d have to say Bath. Commercial: discovery channel\'s \"atmosphere\" Cereal: umm... Frosted Flakes, TTTTtthhhhhheeeeeyyyyy\'rrrreeeee GREAT!! TV shows: Enterprise/The Tick/Older X-Files/Sunday Night Fox Movies: Moulin Rouge, Cowboy Bebop, Lord of the Rings(hopefully) ... Scent:victoria secret Pear Music group: Jewel? (she\'s not really a group more like a Goddess) Favorite person to talk to online: can\'t really say I have a favorite, although Kevin is fun cause he and I can both stop writing for like 3 hours and then come back to the convo:) First crush: Marion Vote What is on your walls: Jewel, Yoda, Tomb Raider Ever been skinny dipping: I don\'t recollect Worst feeling on earth: Having to shoot your only friend in the back of the head cause he\'s retarded and keeps killing people Best feeling in the world:I\'m guessing sex, but I\'m knowing spending time with someone you like Word or phrase overuse: Fuck shit ass muthafucka Get along with parents: yeah...all four of them.  Bedtime: I try and get to sleep by midnight, but I can\'t usually What are you wearing right now: boxers, shorts, socks, \"freak\" T-shirt Are you lonely: no Are you talking to someone online: not right now:)   Last 24 Hours...    Cut your hair: no Worn a skirt: maybe:) Been mean: no Been sarcastic: nope :) Met someone new: uh uh Fought with your parents: no First thing you thought this morning:aww shit, another day   In the future...    Where do you see yourself in 10 years: hehehe, world traveller Future son\'s name: John Future daughter\'s name: Amber Where do you want to live when you get married: hmm... don\'t care   When was the last time you...    Wished upon a star: can\'t remember, I think I was 7 Went to the beach at night: umm... long time ago Spent some quality time with yourself: 4 years ago? maybe 5:) oh wait, I did live with Steve for a while;) Read a book for fun: this morning  Believed in yourself: now   How long did this survey take you: 21 mins Do you want your friends to send this back to you with their info? nah, they could just blog it...  '),(18942,317581,'18,November,2001','2001-11-18 00:00:00','   How come no one ever tells my house that there\'s a fucking meteor shower? I mean really people you could give us some warning  '),(18943,317581,'18,November,2001','2001-11-18 00:00:00','   PS- Paul and tarah\'s wedding was cool, I drew on the program and got to see lots o\'peeps that I hadn\'t seen in a long time. Weirdly enough, alot of them were looking for work:(  '),(18944,317581,'17,November,2001','2001-11-17 00:00:00','   note to selves: Listen to Mudhoney - Acetone  '),(18945,317581,'17,November,2001','2001-11-17 00:00:00','   Someone recently asked me if I believe that there is 1 person out there for us.  I had to respond like this...  Yes and No, I did believe in the 1 true love thing, a while ago. However, I no longer believe in it. I believe that we make the love. We can create and destroy the love. There is no 1 person you must find. No predestination, Bill Gates was not predestined for the power he has. He took it.  hmm... sorry, got a little off track. I think that you can love multiple people. There are people that I will forever love, they have taken part of my heart with them. (I think there are 5 girls, but maybe only 4, i forget the exact count and I try not to think of them) I think no matter who you love, there will always be work involved in the relationship sometime or another. I mean you can\'t be in \"love\" forever, but you can love someone forever. Do you even understand what I\'m saying?  Anyway, Harry potter was awesome, followed the book extremely well. \"But I\'ve never even been to Mars\"  '),(18946,317581,'17,November,2001','2001-11-17 00:00:00','   EEk EEk, I\'m a monkey  '),(18947,317581,'17,November,2001','2001-11-17 00:00:00','   Demotivators kick ass! check this one out:  urlLink ignorance  or this one:  urlLink loneliness . For more, go to  urlLink www.despair.com  '),(18948,317581,'16,November,2001','2001-11-16 00:00:00','   DUDE!!!! That show WAS the shiznit... Those fuckers rocked my socks off...   But now my feet are bloody hurting. OUch, well time to watch movie with kevin... Night  '),(18949,317581,'15,November,2001','2001-11-15 00:00:00','   Just wanted to give a shout out to everyone. I\'m chillin\' at kevin\'s computer at his work, waiting for him to get out of a meeting. Cause we\'re going to go see the D tonight, along with Weezer and Jimmy Eat world.   I\'m so excited!!!! (*dork*)  Anyway, I\'ll check you guys latah  '),(18950,317581,'14,November,2001','2001-11-14 00:00:00','   I\'d post more, but what am I going to blog about? It\'s not like there\'s a whole lot happening with me these days... let\'s see, get up, look for jobs a few times during the day (online), play some games and whatnot, and go to bed.   I lead a thrilling life.   Oh, and Tim, don\'t forget the 1st step of being an alky, denial. :)  '),(18951,317581,'14,November,2001','2001-11-14 00:00:00','   A couple more things I guess...  So I won racquetball tonight. And while I was doing laps on the track I got started wondering why people would pick me as a friend. I didn\'t come up with any particular reason, except I must\'ve tricked em good. I mean God knows I don\'t open up to people.   Jameson makes me warm.  Also, Mary, why is it Larry? LcCorkle? I mean, I\'m just curious:) Did you have an operation? Like a name transplant or something?  '),(18952,317581,'14,November,2001','2001-11-14 00:00:00','   so, someone has been bugging me about blogging, so I figured I\'d waited long enough:)  Well, actually, I\'m supposed to go watch \"Enterprise\" now, but now I\'m back from that:) weird how I can leave and come back and just keep typing like no time past and now you\'re reading this, like, \"Tim you\'re on crack, it\'s the same damn message.\"  So, when I got the mail yesterday, Marc had sent us a postcard of a topless girl on a beach and was like, \"I read a lot at the beach here\", and then today when I got the mail there was a postcard for Austin. This one had a dude on the front and said, \"Hey Austin, this beach has lots of topless men on it, it\'s pretty cool, you should come visit it\" I almost pissed my pants I was laughing so hard.  N - E - way, I\'m drinking a little Jameson tonight, just to keep me warm of course;) I\'d say I was turning into an alky, but this is the first drink I\'ve had in like a week and a half, so I wouldn\'t say that I am, but I\'m doing like 2 of the 4 signs. Drinking by myself, and only drinking hard shit.  So, I\'ve been really busy at work recently, my site went live. So I\'ve been maintaining it. Trying to keep it running. Had to come home early today, because my brain broke down. (You know that time when you are overwhelmed with work to the point of not being able to function, that\'s the point I was at)  So I came home and bought the new Jewel album, much kicking of the ass:) Also got to play with the game cube, which I want more time with, cause I think it\'ll be much kick ass. (Played Rogue Squadron, very fun)  So, I\'m really looking forward to the Tenacious D/Weezer/Jimmy Eat world concert tomorrow night. IT should kick some major bootï¿½.  Well, I think I\'ve blogged enough for right now, I\'ll check you dogs latah...  PS - I just checked Sunday hero and was shocked to see she\'d changed her layout once again. I guess I\'m just a guy that likes my layout, I think the blue helps to neutralize any angry feelings:)  '),(18953,317581,'12,November,2001','2001-11-12 00:00:00','   no reading:( I have no time:( Could I also tell you how I\'d like to be independently wealthy? Then I could stay home and play with my roommates all the time. oh well, on the plus side I got my car refinanced.  '),(18954,317581,'09,November,2001','2001-11-09 00:00:00','   So now that I\'m 24 and people like me again... I figured I should start blogging again...  Went up to the Lee\'s yesterday and celebrated the bday with them. Got, Grave of the Fireflies, for my bday, still need to watch it.  It\'s about midnight on sunday night/monday morning, and I just got done watching \"snatch\" in my room with big A. Now kevin is playing GT3. Over the past couple of days I\'ve gotten, SSX Tricky, THPS3, Devil May Cry, and Spyhunter. Talk about being busy. I think we need more PS2\'s in the house:)  Also, I haven\'t heard from Brad, so I need to call him. Also, I finally got the games from the amazon. Those bitches. I\'m watching some ifilms, right now, they\'re way funny.  Also, I just wanted to talk about how I love women. and how i do not understand them. But I think I\'ve already talked about those two before, so I won\'t bore you. And another thing, Weezer is gonna rock....  '),(18955,317581,'09,November,2001','2001-11-09 00:00:00','   oh yeah, and I\'m loving XP so far, it is very fun  '),(18956,317581,'08,November,2001','2001-11-08 00:00:00','   urlLink FUNNIEST SHIT EVER!!!!  '),(18957,317581,'07,November,2001','2001-11-07 00:00:00','   So another postless 24 hours... amazing. It must be just a dead time for writing for people.   I had a pretty awesome evening last night, hanging out with friends I haven\'t seen for a while. Another one of my friends is engaged... I can\'t believe it. I guess that\'s what happens when you don\'t get a chance to see someone for 6 months or so... wow. It was fun to talk finally though, after such a long time. It\'s amazing what can happen when you just decide to hang out and spontaneously call people :)  '),(18958,317581,'06,November,2001','2001-11-06 00:00:00','   Wow... no posts in 24 hours.   So anyways... I was going through my files and stuff on my computer, and stumbled across a bunch of old sent mail from like 4 years ago. So I started reading everything, and it was pretty crazy, just seeing what I was thinking/going through back then, and how much I\'ve changed (some for the good, some for the bad). It\'s pretty cool to have this sort of record of thoughts from back then. I\'m glad I saved everything.   On the job front, it\'s oh so encouraging to see about 15 whole new jobs posted this weekend. Wow... yeah, we\'re not in a recession or anything right now. riiiighhhtt.  '),(18959,317581,'05,November,2001','2001-11-05 00:00:00','   DAMN YOU GUYS!!!  Best movie month this year = December... I know a lot of good movies have come out recently, but just look at the dates...  Dec. 18 = Moulin Rouge hits DVD, You know I HAVE to own this movie. Especially since I DO consider it one of the greatest films ever. Dec. 19 = LORD OF THE RINGS hits theatres nationwide.  These will be my 2 favorite movie days I believe. Also, I think maybe they\'ll do a midnight showing of LOTR at Cinerama and you know I\'m all over that, if I can get tickets. I mean it\'s fuckin\' LOTR.   Oh so my yesterday... I went to church in the morning with Big A and Cliff and Hannah. Then we went to Thai food afterward. mmmm... Thai food. (items ordered: vegi Red Curry***, something fire**, cashew chicken***, and Pud Thai Chicken**:)) Then we went and saw Monsters, Inc at Oaktree. It was fun, and we got to see the Episode 2 teaser. Then we came back here and I got a headache, so I was in bed from 6pm till 7:15 this morning.  hmm.... my birthday is on saturday, in case anyone was curious:)  Have a good day everyone...  '),(18960,317581,'04,November,2001','2001-11-04 00:00:00','   umm... Amidy, it\'s \"Darby O\'Gill and the little people\" not Derby  '),(18961,317581,'03,November,2001','2001-11-03 00:00:00','   in the dream I play an almost 24 yr old male looking for a job... Actually in all accounts, the dream me would seem real in this dream world...  \"Today\'s story is called, \'Albert the clown\', written by Andy, age 29. Wonder who that could be? There was once this funny clown and his name was Albert. Sometimes people tried to be CUTE and pronounce his name in french, which was \'Al-bear\'. People did this so much that Albert started believing he WAS bear. He moved to the forest of the Northwestern United States, to live on nuts and berries. Soon, \'Al-bear\', became a tall hairy man with long hair and fingernails and bare feet. Because he was an ex-clown, he had BIG, HUGE FEET. Today \'Albert the clown\' is better known as big foot. So the next time you\'re camping in the great Northwest, remember to raise a big \'Hey-diddly-o\' to \'Al-bear\'.\" -  urlLink Brak  '),(18962,317581,'03,November,2001','2001-11-03 00:00:00','   I\'m Mark, I\'m an eskimo...  '),(18963,317581,'03,November,2001','2001-11-03 00:00:00','   in some weird twist of fate, what is real?  I\'m pretty much sure I\'m a figment of someone\'s imagination... But I couldn\'t be forsure. This dream sure is interesting.  Maybe cahlen is on to something. (his loving sleep, kevin too)... What if we\'re living a dream, and at night we go and lie in our beds and wake up. When in this dream we think we\'re awake...  Oh what an interesting thought....  '),(18964,317581,'01,November,2001','2001-11-01 00:00:00','   They can\'t even ship 3 silly things on time. NOoooooo, They have to wait 3 days after something comes out and then not ship it till like the next day and you really wanted it for the weekend but do you get it for the weekend? NooooOOOooooo...  FUCK YOU AMAZON!!!  '),(18965,317581,'01,November,2001','2001-11-01 00:00:00','   FUCK AMAZON!!!  '),(18966,317581,'01,November,2001','2001-11-01 00:00:00','   another day, another drachma...  well, not today, actually, today was no drachma:(  Oh well, told Austin that we don\'t really like his choice of replacement and that we\'re looking for our own guy. C\'est la vie. oh well, I\'m really hungry.  '),(18967,317581,'01,November,2001','2001-11-01 00:00:00','   So it\'s my first day of unemployment again, too bad I couldn\'t sleep in. Argh! stupid body clock waking me up at 6:30. The fun of looking for a job begins again, at least I\'ve already submitted a couple of resumes.  '),(18968,317581,'31,December,2001','2001-12-31 00:00:00','   Happy New Year  Well, almost anyway. It\'s work day for Tim!!! WOoohooo!!!! :P  Anyway, I went to the sonics game on saturday night. It was sweeeet. 13th row. Free Beer and food. fun time by all (cliff was there). Then we went to the music store to wait for traffic to die down, and Gary Payton came in and was looking through the DVDs. It was soooo cool:)   Yesterday, I played games with cliff till around 6:30 when Marc came over and we went and ate mongolian food. Then Cliff went to Hannah\'s and Marc and I played Galactic Battlegrounds. Then we watched Malcolm in the middle, which was funny, but I think it was a rerun. Afterwards we watched the last half of Half baked. and then marc went home and I finished up my second year of football.  Now it\'s new years eve and I\'m at work. Check you latah- T  '),(18969,317581,'29,December,2001','2001-12-29 00:00:00','   mmmm.... tits:)   hehehe.... well I went to some bars tonight with cliff and Corey. And there was this girl there. She was pretty muscular and had on this tiny little t-shirt that said hustler. Anyway, she did this little dance thing with her ass and tits and then occasionally she would lift up her shirt to reveal, no bra:) I was the only one watching (of the 3 of us) so cliff and Corey were both disappointed:) oh, maybe next time:)  '),(18970,317581,'27,December,2001','2001-12-27 00:00:00','   Thanks for taking care of that crazy code thing.  Oh, and I agree about talking people at movie theaters. Or Cell Phone users, little kids running around/talking/kicking seats/crying, laser pointer users, etc. They should all be summarily smacked by everyone else in the audience on they\'re way out. Kinda like that movie commercial that shows the person getting pegged by stuff while talking on a cell phone.  '),(18971,317581,'22,December,2001','2001-12-22 00:00:00','   movie annoyance  Tonight, as I watched fellowship of the ring again, I had the privilage of sitting next to a very pleasant chap. That is to say, after talking on his cell phone during the previews, quite loudly, to billy bob, he proceeded to talk to his girl throughtout the movie. What I especially enjoyed, was when he pulled out his chewing tabacco. Fortunately he had snuck in a can of pop so he had a place to spit, otherwise who knows where it might have ended up. Needless to say, I didn\'t thoroughly enjoy sitting next to mr. \"breathes extremely loud\". Maybe next time I\'ll not see it in Woodinville, perhaps that would help.   christmas/workmas  I\'m back from christmas at home. I went home friday evening around 9 pm.WEnt to a kick ass party at Amidys on saturday night. Unfortunately, i lost my voice that night, plus I had to get up at 5:30 to drive down to fuckin corvallis. Oh well, I got to see my grammie. She has alzheimers a little. So she can\'t really remember people. On monday my work called, even though I was in OR, to tell me the network was down, even though there was someone in seattle that was on call. So I worked for 30 mins on monday. Joy.   For christmas I got, Shrek, Episode 1, Bill and Ted\'s Excellent Adventure, and final fantasy. Oh and I got some clothes and a gift card for target and some cashola. Not a bad christmas, but I really need to be at work tomorrow for a full 10 hours. So I\'m going now.  TTYL- T  '),(18972,317581,'22,December,2001','2001-12-22 00:00:00','   alright, I know you\'re sick and all, but take the damn onmouseover open window command out of the code. It\'s pretty annoying, in case you couldn\'t tell.  '),(18973,317581,'20,December,2001','2001-12-20 00:00:00','     '),(18974,317581,'20,December,2001','2001-12-20 00:00:00','   crowds  Did I ever tell you how much I enjoy my solitude. or maybe how much I hate the crowds at stores and malls? Sometimes I think that children should have to remain at home, then they wouldn\'t be running into me at the store.  Thank God I got my christmas shopping done. And the presents are wrapped.  Tomorrow the neices and nephews show up.... *sigh*  I\'m still sick but at least I got drugs. Looking forward to the party tonight. I think my mom wanted me to be all excited about the presents for me under the tree. It was just difficult getting excited after driving down, being sick, and the shit hanging over my head right now. you know?  Sometimes it feels like all I do is complain. I\'m sorry. I wish I could be more jovial:( maybe next year. T  '),(18975,317581,'20,December,2001','2001-12-20 00:00:00','   Drugged and Awestruck  so I saw Lord of the Rings, Fellowship of the ring, twice last night. It was so awesome. It makes me want to see the next one. AWESOME is probably the only way I can describe it.  Unfortunately, all week I\'ve been staying up late and getting up early and now I\'m sick. Fortunately, I have dayquil to give me a buzz:)  '),(18976,317581,'17,December,2001','2001-12-17 00:00:00','   Ahh... How come, as a society, have we replaced \'why\' with \'how come\'? I mean I was thinking about this last night as I drifted off to bed and I was thinking how little sense \'how come\' means. I mean, really, where did it come from?   How come we can use \'how come\'? or why don\'t we just use \'why\'? or why do we use \'how come\'? How come we don\'t use \'why\'?  weird, 4 sentances with basically the same meaning...  HOW COME?!?!?!  '),(18977,317581,'16,December,2001','2001-12-16 00:00:00','   How come women can be like, \'You don\'t pay me enough attention?\' I mean, after all you\'ve done for her?  You tell her you love her, you spend free time with her. You buy her flowers. What do you want? How can we make you know that we love you, yet not have to spend every ounce of energy trying to prove it. Can you not just be delighted in knowing we love you?  Why is it any time I want to spend any time with my friends you say, \"You don\'t care about me, just go do whatever you want to do...\" and then turn a cold shoulder. Do you honestly believe that that makes us want to spend more time with you?  Grow up! Don\'t make us always decide what we\'re going to do. If there\'s something you\'d like to do, then TELL US.  we aren\'t living in the 20\'s, you can tell us what you\'d like to do, you don\'t have to be content with whatever we plan to do.  God knows it would help us alot.  Thanks- Guys  '),(18978,317581,'14,December,2001','2001-12-14 00:00:00','   I\'d like a vacation... I think a few days on a beach drinking rum with my friends would be excellent right now....  Also, Mary, you should move to New Zealand, it\'d give me another reason to visit:)  '),(18979,317581,'13,December,2001','2001-12-13 00:00:00','   well, I really want to watch sopranoes, but kevin is talking on the phone to jenny and building his computer. :( oh well, I guess I\'ll go play games  '),(18980,317581,'12,December,2001','2001-12-12 00:00:00','   Another day, Another Drachma... The boss is finally out of town. My coworker isn\'t in yet, cause it\'s 11 and the boss is out of town.  Anyway, My diet is going ok. I\'m caffiene free. I resisted pizza last night, for a salad. and some jello. mmm... jello.  Anyway, gotta run, at least look busy for the coworkers.  '),(18981,317581,'12,December,2001','2001-12-12 00:00:00','   hmm... have I kept you waiting long enough?  where are people? Austin moved out. Dejah is in Belize. I\'m coming out of my \"no caffiene\" coma. All in all a pretty normal week:)  I haven\'t turned on my computer at home since monday morning. so hopefully the electric will be less next month. :) Anyway, 1 week till LOTR. I\'m pretty damn stoked.  Saw, Ocean\'s 11 last weekend, twice. Do you think I liked it? I thought it was good, but there were some times when I knew what was happening. Well, Jarred is trying to convince me to become a screenwriter, so maybe I\'ll do that. It sounds like a fun job.  Peace  '),(18982,317581,'06,December,2001','2001-12-06 00:00:00','   It\'s true, I\'m an anime whore. Plus I\'ve put in 40 hours this week so far:) I need to go to bed.  Also, I updated my amazon wish list... a little  and then there\'s the new time waster:  urlLink laser game  '),(18983,317581,'05,December,2001','2001-12-05 00:00:00','   So the job market is kinda interesting. The same day I get an email about a position being filled, I hear from a game company that wants me to register with their career center (unfortunately they\'re located in NY if I remember correctly). I don\'t know how much good it will do me, but what the heck, considering I\'m running outta cash.  '),(18984,317581,'04,December,2001','2001-12-04 00:00:00','   ...and thus, Kevin returned to the blog.  (smiley face) :)  '),(18985,317581,'02,December,2001','2001-12-02 00:00:00','   well, I had a really good topic to write about, but I forget what it was...  I only worked 10 hours today, hahaha, only:)  I need to go to bed, so I\'ll check you latah- T  '),(18986,317581,'02,December,2001','2001-12-02 00:00:00','   So it\'s sunday night now...   Yesterday my internet went down until an undisclosed time, but it turns out, it was just until today.  It\'s been a super busy weekend. Brad bailed, so we\'re looking for a roommate. Marc might be thinking about it, but I doubt he\'ll accept. Even for a few months:( Cliff has someone interested, so we\'ll see about him.   Friday night, I went to tradition, but it was mostly boring, but I got to see lots of cool people. Then on saturday I cleaned up the downstairs. and dejah came over and we watched, X, Emporer\'s new groove, and 2 episodes of Neon Genesis Evangelion. It was fun. After I dropped of Dej, I came home and watched, Zero Effect with Kevin and Cliff. It was fun, but I stayed up way too late (3:30 am)  Got up for church at 9:30 (actually got up at 9:50) and then we got to church at 11am(it only starts at 10:30:) Oh well, it was fun. Then I had lunch at Rosita\'s and then went to REI with Dej. Worked from 3 - 4 then went home and saw \"Spy game\" with Kevin and Craig. It was a good movie, but I can only handle Craig in very small doses.   Ugh, anyway, I watched Malcolm and Simpson\'s with Cahlen then we came down here and watched the last 2 episodes of Evangelion. I wan\'t to watch them all, but unfortunately dejah doesn\'t have them all for me to borrow:(  Well, I\'ve now rambled and made a huge post, when I can\'t even read huge posts:( And now, I\'m watching Gunbuster episode 3 on the internet, so what can I say? I\'m an anime slut:)  Catch you on the flip side:) T  '),(18987,317581,'31,January,2002','2002-01-31 00:00:00','   Another day...   Woohoo!!! Can I just tell you how excited I am for Microsoft!!!!!   Anyway, I sent an email to Houck to see how he was doing, and it turns out he\'s still working for  urlLink Dark Horse Comics  and he\'s coming up next week sometime. Well, as you may or may not know, I love comic books (even though i was never allowed to have them when I was younger cause of the devil), and anyway, he\'s gonna bring me some:) YAY!!!  Also, I wanted to point out that my friend  urlLink Nate  got a blog going.   So I just want to give a shout out to all my peeps and tell ya to go give Nate a visit:) Peace- T  '),(18988,317581,'30,January,2002','2002-01-30 00:00:00','   alot, a lot, there, their, they\'re, your, you\'re    these are the words that I cannot use in correct context for the life of me. (I know that alot is not a word, but I still use it, in fact I use a lot very often) :)  Does it make me a hypocrite? I mean, should I have to live in a world where my boss cannot spell routine, and then he misspells it like 420 times? yet, since I can\'t spell \"a lot\",  I am a hypocrite? (PS- I can\'t spell hypocrite either)   Well, I\'m done on my little spelling defense, peace out- T  '),(18989,317581,'30,January,2002','2002-01-30 00:00:00','   HAPPY DAYS ARE HERE AGAIN   Got the call from MS, I\'m going back on monday the 11th. Woohoo!!!  I told this company I was leaving and they\'re like, ok, we need to get our budget figured out anyway. Now, I\'m just like, \"Woohoo!!!\"  Anyway, I still haven\'t found \"Intro: a 23 yr old virgin\" anywhere:(  '),(18990,317581,'30,January,2002','2002-01-30 00:00:00','   APB   This is an all points bulletin....  If you have a copy of the first draft of my introduction sitting on your computer somewhere, please send it to me ASAP. Mine seems to have disappeared.  Thank you- T  '),(18991,317581,'29,January,2002','2002-01-29 00:00:00','   Also, I was just talking with Nate (whose wife thinks I\'m addicted to computers and not books so I thought I\'d let you all in on my response to that)  Basically, it\'s the \"If I had unlimited time and resources question\" or better known as \"If I won the lottery question\" If I won the lottery...  I would buy 3 houses... In one house I would have my friends, my TV\'s, my video games, my Computers, all my technical stuffs. (1st house)  In one other house, I would have a library, and a fireplace. (2nd house)  In my 3 house, I would have my pool, my pool table and my other fun shiznit. (3rd house)  So I would live in the first house, invite my friends to my 3rd house for fun times(also have wild parties there) and then retreat to my 2nd house whenever I was feeling like I wanted to get away.  So my explanation for not reading as much as i can right now, is that I am too distracted and do not have enough time right now to both be sociable and read alot...  Ahh, it\'s good to dream.  '),(18992,317581,'28,January,2002','2002-01-28 00:00:00','   pics  So I put up some more pictures, these are of my 22nd bday, ahh only 2 years ago:) :)   urlLink Enjoy these pics   Peace out  '),(18993,317581,'28,January,2002','2002-01-28 00:00:00','   mouth noises   I have recently relearned that making farting noises with your mouth can be really fun. :)  pbbttthh!!  '),(18994,317581,'28,January,2002','2002-01-28 00:00:00','   fun sites jarred sent me that you should all go to:)    urlLink site 1    urlLink site 2    ENJOY  '),(18995,317581,'28,January,2002','2002-01-28 00:00:00','   let it snow   The weather outside is frightful... you know the rest.  Anyway, I\'m scanning some pictures cause I want to update my  urlLink Unwelcome hero  as I have not done that in quite some time. For those of you just turning in. It\'s snowing outside, and I\'m hoping to work from home at least a little tomorrow. It\'s finally sticking to the roads, but not our driveway. Weird.  So anyway, I got an email from  urlLink my friend , and figured that was a good reason enough to send you all(you know who you are, all my invisi-friends:)) to her. So there.  Also, you\'ll notice my shameless plug earlier about my unwelcome hero. I\'m thinking of putting part of my intro for my book up... we\'ll see how drunk I get tonight;) (that\'s where cliff and his friends went, to buy alky:) $125 worth :P WooHOO!!!  Well, I\'ll check you latah- T  '),(18996,317581,'28,January,2002','2002-01-28 00:00:00','   Addicted   It\'s true, I am addicted to books. They are my crack. I need my fix.  It seems to me that there would be some sort of help group for this. but I did a search on google and was unable to find anything....  Please help!!! (BTW, I just ordered \"American Gods\" from Amazon) help, please...  '),(18997,317581,'26,January,2002','2002-01-26 00:00:00','   Misspelling:)  I would like to take this moment to thank Mary for her honest and helpful reply to my misspelling:) Thank you mary:)  also, I\'d like to give a shout out to all the imaginary people reading this blog. Without you, I don\'t know if I could go on. Heck, I\'ll even give a shout out to the real people that read this blog:)  Peace- T  '),(18998,317581,'26,January,2002','2002-01-26 00:00:00','   Disdain  - The feeling Tim has for his boss.  here it is again, I almost called in sick today. I wish I had. Oh well, the other guy called in sick today, so I guess it\'s good I showed up.  For some Odd reason, I feel that setting up my voicemail isn\'t that important... Of course it might have to do with the fact that I know I\'m leaving, and my boss doesn\'t. Damn 2 week notifications.  Anyway, I am back on my \"wanting to write\" stick, but am yet somehow distracted by other things(final fantasy).  I hope Amidy had a nice time on her birthday weekend. I didn\'t get out much cause of the snow.  check you guys later- T  '),(18999,317581,'26,January,2002','2002-01-26 00:00:00','   Mmm.... The Weekend:)    Had another \"all nighter\" last night, that is, I went to sleep around 6am after playing Final Fantasy all night long. Then got up at 12:30 and played until like 2 when I ate. It started snowing around 3:10 and kevin and I went and saw Orange County. Not as funny as I was hoping, but it was kinda a feel good movie I guess.  So I don\'t know the point where I became cynical of movies.... (saw a preview for some shitty looking movie about a kid that gets a watch that speeds up the user) I mean, I\'m thinking that I wouldn\'t even go to that movie if someone paid me, well, maybe if they were paying me to write a review of it, so I could stop other people from seeing it.  *sigh* oh well,  On the other hand, we(kevin, Jarred, and I) watched  urlLink Buckaroo Banzai  last night. That was a funny ass movie. Probably would have enjoyed it better drunk, but you can only be drunk so much;)  Hmm... speaking of which, I haven\'t spoken to my girls in awhile. (probably cause they found out that I call them my girls:)) Plus, I\'ve been working nonstop and all that shit:P  Gotta bug the Volt lady about videogames again this week. Gotta work on my resume tomorrow.  Peace- T  '),(19000,317581,'25,January,2002','2002-01-25 00:00:00','   So, when I got home last night, I go out to get the mail and I hear this \"hissing\" coming from my car. So it turns out that it\'s my left rear tire. I moved it down to the bottom of the driveway(flat part in the road). So this morning I get to change the tire and take it to  urlLink discount tires  again. (note: see wednesday for the last time I had to go to discount tires to fix a flat)  Why can\'t seattle have streets that don\'t eat my tires? I mean it, it\'s like every God Damn place I go in this city, there\'s fuckin potholes everywhere.  in logical progression(in tim thought anyway) I AM anal about spelling. I keep on bugging the receptionist about the copier tracking sheet that read \"Lazer printer\". I mean really, fucking Z\'s in Laser? It\'s that kind of shit that pisses me off:) Also, have i told you like how I work with these people that couldn\'t spell their way out of a bag? Like my boss, he\'s always mispelling shit, and then he can\'t write anything that anyone could read. People are all asking each other, \"WHAT THE FUCK is THAT word?\"  this concludes my rantings for right now:) I gotta go change a tire...  '),(19001,317581,'24,January,2002','2002-01-24 00:00:00','   and now we\'re down to 100 KB left... :) You check the time stamps:)  '),(19002,317581,'23,January,2002','2002-01-23 00:00:00','   So I quit typing and the damn thing stopped, but now it\'s going again. Stupid computer. Just about to hit the half way mark.  There, did you feel it?  I did, and boy, did it feel good:)  (Never mind the dork behind the controls folks, he\'s just a bit sleepy)  '),(19003,317581,'23,January,2002','2002-01-23 00:00:00','   So today, it\'s Win2k Server, but it\'s still having network issues that seem to work when I\'m connected to Blogger, but not when I\'m not:P  Oh well, more fun for my peeps that read this shit I guess:) Fuckin, DSL/41minutes to install IE6, my ass  And the sad thing is you know, is that it SHOULD be faster. I mean it wouldn\'t be so bad if you thought it was supposed to take this long. I mean this shit on any other computer in the office only took like 5-10 minutes. But here, on this POS, I\'m stuck with 37 minutes left:P  Oh well, at least it means I\'ll type more;) I\'m going to Squish and Dan\'s on the 12th if anyone wants to come. They said to invite you:) (ok, they only named 2, but I\'m sure they\'re cool with meeting new peeps;)) Also, they got a house, so there should be enough room for us all.  Well, it\'s true, the more I type the faster it goes. So, I\'m now also downloading,  urlLink this piece of spam shit  that my boss paid for. That\'s actually been my entire job over the past 2 days. Get a computer that he can Terminal Service into and set up that program on....  The sad thing, the program is downloading at 83 KB/sec, the fuckin IE6 more like 10KB/sec:P  You\'d think MS, being across the lake as it is would be faster, but NOoooOOOOoooo, poop on them:)  Jeopardy theme song... \"Do do do do Do Do Doooo, Do do do do DO, do do do do do....\"  Would somebody please STOP THE INSANITY????!?!?!?!  AAAAAAAAAAAAAAAAAAAaaaaaaaaaaaaaaaaaaaaahhhhhhhhhhhhhhhhh!!!!!!!!!!!!!!!!!  :)  '),(19004,317581,'23,January,2002','2002-01-23 00:00:00','   I would like to take this moment and let  urlLink Amidy  know that  urlLink her blogger  takes WAY WAY WAY too fucking long to load.  '),(19005,317581,'22,January,2002','2002-01-22 00:00:00','   also, the update seems to be going faster when I\'m logged into blogger, so I\'m gonna continure to type till it gets done... :)  basically, now I\'m the guy at work everyone comes to to get shit done. A computer breaks, ask Tim, the internet goes down, ask Tim. There\'s a pregnant lady giving birth in the lobby, ask tim. :)  hehehe, OK so the damn thing stopped again. Windows update sucks, when it gets stuck. And another thing, we lose internet connectivity here, way too damn much. Like at least once a week now.  FUCKER, it got stuck at 964, AGAIN!!!!  Now, I\'m a grumpy grape. :(  '),(19006,317581,'22,January,2002','2002-01-22 00:00:00','   ok, so this computer is dumb ass slow. and I can\'t fiugre out why.  It takes an hour to download, a 2 meg file over a dsl line? Stupid ass computer:P  anyway, as long as this update doesn\'t crash again, I\'ll be happy.  Check ya- T  '),(19007,317581,'22,January,2002','2002-01-22 00:00:00','   Another DAy, another 9000 drachma... Well, I\'m losing air in my right front tire, so I have to go to discount tires today. it\'ll be a nice break from the monotony of rebuilding this damn computer:)  catch you latah- T  '),(19008,317581,'22,January,2002','2002-01-22 00:00:00','   daria is dead:(  I forgot to mention it last night after I watched the end of daria. This is sad.  Check out  urlLink this article  on MSN  '),(19009,317581,'21,January,2002','2002-01-21 00:00:00','   weird ass dreams   So it all starts out that I\'m going to a movie(weird;)) Anyway, we get to the place the movie was playing(it was an old movie, so it wasn\'t playing at a theater. Turns out, it\'s playing at the charismatic, non-denominational church that I went to during my fomative years. (aka 6 - 13) Anyway, I get there and there\'s some typical seattle people hanging out for when the shows gonna start. (shows gonna start at like 7pm and it\'s only like 4) (I figure the movie is pretty popular with people getting there so early)   Anyway, I go talk to some girls and take a drag off their roach, then I\'m like I\'ll be back. So I head for the church and I see this girl that I used to like back in HS. And I\'m like, what\'re you doing here, what\'s going on.... blah blah blah, and she\'s heading into the church so I follow her. Well, it turns out that there\'s a service going on right now, and I\'m like, what the fuck, might as well. So I follow her in and sit in the back, half facing away from the preacher.  Well, basically, I sit through the service and am about to leave and go chill withh those girls before the show. But as I\'m leaving, I pass my mom (in my old kitchen, which of course was never in that church building, but somehow ended up there in a dream?!?!) And she\'s like, the movie\'s not playing tonight. And I\'m like, \"C\'mon mom, everyone is here to see it\", and she\'s like \"No, I\'m making diner and Fred is coming over to catch up on old times\" (now the thing is, Fred was like this guy that my mom found solace with right after my dad left, in like 1980, so it\'s not like I even remember the guy, and what is he doing in my dream?)   So I go outside and tell all these peeps that the movie\'s not going to be shown, and they\'re all sad. But then later I gave oral pleasures to that girl who I smoked with... and then my alarm went off.  Quite a night for the old Tim:)  '),(19010,317581,'21,January,2002','2002-01-21 00:00:00','   hmm... well, I\'m a  4.7 on hot or not .com :( C\'est la vie.  I\'m drunk again. After \"BRotherhood of the wolf\" today at 5:30 Corry(our new roommate) and me went out. We went to Hooters and McCormicks and Paragon and Pesos. Needless to say I had a lot to drink. Basically, I told Corry to take off saturdays so we could go to bars on hopping nights, but that\'s not going to happen.  Also, Cahlen is a mother fucker. He gets FFX yesterday, then he takes it home with him today. Basically, I hate him:P  I\'m tired and my teeth feel funny:) Check you latah- T  '),(19011,317581,'17,January,2002','2002-01-17 00:00:00','   And the whole world loves it when you sing the blues, bah, bop bop, ba da  Sorry, just been listening to a little Outkast recently;)  So yeah, I\'ve been working my ass off, I guess it has to do with the fact that I\'m leaving, but also, ALL the shit I\'ve gotten assigned to me. My coworker is complaing cause he doesn\'t want me leaving him with the boss from hell, but I\'m moving to greener pastures. I got to work at 7 yesterday, and I\'m shooting for 8 today:P UGh, I\'m gonna go in late on friday I think.  Anyway, Hope you are all having a good time. Check you latah- T  '),(19012,317581,'16,January,2002','2002-01-16 00:00:00','   Sorry for being silent the last week n a half or so. I\'ve been working my ass off!!!  '),(19013,317581,'05,January,2002','2002-01-05 00:00:00','   Another weekend....  Well, I\'m working on this beautiful saturday afternoon. We got a new roommate, Cory. He\'s cool. It\'s all mundane work, but I like the fact that I don\'t have to think, I can just print and stack. Print and stack.  Watched Twin peaks first 4 episodes last night. Very interesting show.  Well, back to the daily grind:) T  '),(19014,317581,'01,January,2002','2002-01-01 00:00:00','   I am the candy man:)  Well, really happy new years now. :) So, I partied at eric\'s last night. That means that we drank lots, but I think this cold is fucking with me or something because I swear I can\'t get really drunk anymore. It\'s kinda sad for me really:( I mean if I get drunk it\'s for like 5 minutes.  Anyway, I have to drive cliff to issaquah in like 20 to get his car. We decided to car pool back to our house after cliff was drunk. It was cool, we went the speed limit the whole way and didn\'t have any problems.  I got an email from MS yesterday and they want me to play videogames for them again, to which I\'m pretty sure that I\'m going to say yes to. I mean it\'s videogames and room for promotion, whereas this job that I have now is like, no room for growth, but you can put down a kickass job title if you want:P  Well, after drinks, cards and xbox at Erics we came back here and talked with the Kelley, Justin, Amidy, Jebe and Kevin. Which was fun, but we stayed up till 4:30 and now it\'s 12:12 and I have to drive cliff to his car to get to work on time. He\'s all hung over:(  But I\'m listening to Aqua, so everything is dum de da de da:)  Roses are red, violets are blue, honey is sweet, but not as sweet as you:) -Aqua  '),(19015,317581,'25,February,2002','2002-02-25 00:00:00','   Dads   Went to mine this weekend. Going to go visit Big Steve in San Diego in April. That\'ll be fun. Finished American Gods and the star wars book I was reading. Now I\'m reading the princess bride. It\'s pretty good. Reminds me a lot of the movie:) Funny and all:)  Anyway, I got better games to test this week. So that\'s really happy.  Catch y\'all latah  '),(19016,317581,'21,February,2002','2002-02-21 00:00:00','   ahh 4 page reports:)   So, we\'ve been playing this game this week that isn\'t stable. That means it\'ll crash for no apparent reason. Which is fun:) Plus they only have the 1st mission complete so once you\'re done with it, blah blah blah, you get the point.  Needless to say, I ran out of work on tuesday, and mostly am just flying around hoping it doesn\'t crash.  So, our fridge started leaking at my house. That was pretty exciting. The landlord is coming on saturday to fix it. At least this way he\'ll think we we\'re doing dishes cause of the water being out. Instead of whoever that doesn\'t do their dishes.  So, I\'ve been getting home at 11 and then I get up at 6:30. It\'s pretty damn tiresome.  Hope you all are doing good.  PS - \"American Gods\" rocks PPS - it\'s a book  '),(19017,317581,'18,February,2002','2002-02-18 00:00:00','   Home  went home this weekend. Mom\'s bday. it was nice, but I\'m tired.  Went and saw super troopers last night. It was super funny. laughed my ass off the entire time throughout the movie. My ass shrunk 3 sizes:)  well, time for videogames- T  '),(19018,317581,'15,February,2002','2002-02-15 00:00:00','   and now, a look at the upbringing of me  (aka how come you can\'t meet any nice guys)    growing up in our christian households we\'re told that as soon as you have sex you\'re going to get AIDS. or at least some sort of STD. Then you\'ll have to live with that for the rest of your days. And then, even if you find someone that you won\'t get an STD from, then you\'ll get her pregnant and you don\'t want a kid now do you?   Then there\'s the 80\'s when it seemed like everytime I saw a movie the women didn\'t want to have sex with the man and then he got thrown in jail. and so it seemed like there was no woman that wanted to have sex, so if you ever tried anything with a girl, she was going to be like, Date Rape. So don\'t even try it.   Then when you get out in the real, real world. (aka after college, or in college if you don\'t go to a private christian university, or aren\'t me) then you realize that girls wanna have sex also, but you\'ve had it drilled into you since you were just a little kid, and you can\'t figure out what to do. I mean you\'re as horny as hell, but you\'re so scared of gettin some disease, you\'re left an outcast, by your own choosing.  '),(19019,317581,'11,February,2002','2002-02-11 00:00:00','   RECON!!!!   Ahh... games:)  '),(19020,317581,'10,February,2002','2002-02-10 00:00:00','   A very Anime day.   Well, I got my manga today. (Comics for the illinformed)   Back-up, wake at 8:30, couldn\'t sleep any longer.   Go get oil changed at 10 in order to be home for Houck\'s visit at 11:45ish.   Oil change gets done at 12:30, Houck\'s waiting when I get home. Get Manga:)   Eat lunch with Kevin and Houck at Olive Garden(Cafe Veloce was closed)   Back to the house for a little Simpsons road rage.  Next stop, Jarred\'s for reading Manga/comics/books/slam dunk contest.  7:45 - dinner at piecola\'s(sp?)  9:10 - Metropolis (Excellent movie)  11:30 - drive home:) (Cliff got yelled at by the crazy lady 2 doors down, \"FUCK YOU CRAZY LADY\")  Thumbs hurt from too much comic book reading. OUch.  '),(19021,317581,'06,February,2002','2002-02-06 00:00:00','   heh      ok, so I\'ve talked to lots of people online. (well, my buddy list is like 46 strong, given I don\'t talk to all of them, but some) Anyway, I was just wondering what the FUCK \"heh\" means. I mean, I\'ve talked to about 10 people that use it. Are you trying to say Hey? or are you almost laughing? like, \"heh, that was almost funny enough for a hehe, but you\'re a fuckin tard, so I\'ll only give you a heh\". So if someone could please explain this mysterious word for me, I would be most appreciative.  '),(19022,317581,'04,February,2002','2002-02-04 00:00:00','   In other news   I really need to sit down with Nate and teach him some HTML. I get all whiggidy whacked when I go to his page:)  When I came home tonight, someone had cleaned the entryway, and also the yard abit. weird.  Blackalicious is kickin my ass:)  '),(19023,317581,'03,February,2002','2002-02-03 00:00:00','   Pete\'s   So, I had an ok day at work. Basically since I don\'t care if anything gets done, I\'ve become a lot less stressed.  Had dinner at Pete\'s Calzone and pizza place. I love that shit. It\'s the best. It was good to see Dej again. Been like forever since I saw her last. Also, Megan was there. In case you can\'t remember she was the waitress there way back when I started going. It was weird cause we had the Pete\'s combo pizza with breadsticks and pepsi, and she\'s all like, you\'re still getting that. and I\'m like yep, haven\'t changed. But really, I usually order like calzone with feta cheese and sundried tomatoes, it\'s just tonight I felt like pizza.  Sometimes I wish I were a performing artist. Singer/Guitarist/Speaker  Sometimes  '),(19024,317581,'03,February,2002','2002-02-03 00:00:00','   superbowl?/Whatever   So the superbowl was \"interesting\". Ate lotsa food and drank lotsa pop.  In case Trav still reads my blog occasionally, I believe \"brim full of Asha\" by \"cornershop\" is the song you\'re looking for. at least that\'s what a  urlLink friend  told me.  Also, my boss is a tard. I don\'t remember how many times I\'ve told him to run a program from the server, but it has to have been a lot!!! But today, I get like a 20 min Voicemail about how he\'s going to blow his top if he can\'t run it, and he\'s tried everyway imaginable to run it, so Tim you must unfuck it up.  Ass  '),(19025,317581,'03,February,2002','2002-02-03 00:00:00','   1 down, 1 to go   Had a party last night. Well, it started at 2, and I was drunk by 3. But I didn\'t stop drinking till 10. So, it was a fun party:)  Also, we cleaned the garage(at least figured out what was trash, and what wasn\'t) yesterday morning.  And now, we\'re hanving a super bowl party today. So that should be fun. WE\'re having Pizza, pop, and chips. Maybe even a bit of a pool. (we\'ll see)  Well, I\'m gonna try and get over this \"headache\" Peace- T  PS- So far we\'ve figured my team consists of all the cool guys from my team, + 2 guys. (Cool guys from my team = Nate, Brad, Brothers Green, Jeremy and me)  '),(19026,317581,'01,February,2002','2002-02-01 00:00:00','   ps, if I spelled tenant wrong, it\'s my boss\' fault:) Also, it\'s his fault if I misused the apostrophe:)  '),(19027,317581,'01,February,2002','2002-02-01 00:00:00','   snip, snip, I just took down the internet    clip, clip, now it\'s back up   So, I was trying to get this one tenant\'s connection working, since when he switched to XP something got fucked up and he has to use this little ass cord. So, I cut this cord that we think isn\'t connected to anything, but then Matt comes and says, \"The internet is down and we can\'t connect to the server.\" anyway, I\'m the first one to be like, umm... maybe it\'s that cord I just cut, even though it\'s blue and the one in the server room is white. Well, we trace the cord and low and behold it connects to a blue one.  Well, the guy who had to use a shorter cord, has been using 2 hooked together this week, so I moved him to a short one and a different office for right then, and took his linker guy. then I clipped ends on the the cable I just cut and put them together while Matt went and got more patcher guys.  That is my story for the day.... The end  '),(19028,317581,'30,March,2002','2002-03-30 00:00:00','   inspiration   Since we last talked... Amorah got married, I got to dance with D at the wedding, Cory didn\'t get the job yet, I had an interview with another MS group, I finished \"choke\", I watched Blade 2, twice, and tonight I watched the Lord of the Rings again:) This time with the Two Towers Preview on the end:)  The LOTR:FOTR DVD info was released...  urlLink here   It\'s almost easter time, and I need to figure out what\'s going on. I think I might be going with Kevin, but he may have already left, dunno. I guess I\'ll find out in the morning.  The Two Towers looks INCREDIBLE. I am mostly looking forward to the riders of Rohan, as well as the Ents. Man oh man, I can hardly wait...  '),(19029,317581,'23,March,2002','2002-03-23 00:00:00','   future  Cahlen\'s leaving, Cory might be leaving...  does anyone care? I need to figure out my future. Amorah is getting married tomorrow. (today)  Nostalgia, I read a little of my travel journal last night, it\'s amazing how I can go back there. Just a word, a smell, a picture, I can almost remember it like it\'s today. Yet I can\'t remember what I did 2 days ago. I can\'t remember the movies I saw last week, let alone last year. I can tell you the main characters major malfunction in every Chuck Palahniuk book, yet can\'t really remember the characters from the Lord of the rings(at least the ones not in theaters right now)  Is my life passing me one day at a time? or is it a year at a time?  I think I\'ve lost all sense of time  '),(19030,317581,'14,March,2002','2002-03-14 00:00:00','   busy busy busy  Well, it\'s finally over, over time that is. I have had every night off this week. It\'s been pretty nice.  Got Zoolander on tuesday and have watched it twice now. That shit makes me laugh. Looking forward to sleeping in on saturday. Only one more morning of waking up before that:)  Dyslexic, I am not. Speak correctly, I can. Think slowly, I do.  Buh bye  '),(19031,317581,'10,March,2002','2002-03-10 00:00:00','   OT   So, I worked 25 hours this weekend. This seems to be a great amount of work. I mean, if you only count the weekend days, (sat. & sun) you get 48 hours, of which I spent more than half in a building at MS.  Anyway, it was nice to have Jebe and Kelley in the house again. Also got to see Dejah and Amorah this week. So that was nice.  what else...  Saw Time machine on friday night. Not a bad movie, could have had more of a more completeness. It left you wanting at least another 30 minutes or something.   Worked till midnight last night, worked till 9 tonight. Last day off was February 24th. Hmm.. how time flies...  Thanks for being in my life- T  '),(19032,317581,'05,March,2002','2002-03-05 00:00:00','   Beautiful Mind   Went and saw \"Beautiful Mind\" tonight.It was really good. There were 7 people at the movie. 3 girls, 4 boys, can you guess who didn\'t have a date:)  I don\'t know, I think I should move out of the boonies. It\'s hard enough to ask if a girl would like to do something, w/o having to drive 30 minutes to get her. it makes it difficult to be spontaneous.  Also, can I tell you that my girl contact has about reached 0. (Not really true, I saw Jebe and Kelley tonight, but they don\'t count cause they\'re here for kevin, I just get the pleasure of their company when Jebe is in town, (also Jebe and Kelley, thanks for hanging out and not dissing me, sorry I couldn\'t let you drive my car, but I think Amidy would have killed me)) However, back to the original point. I can\'t remember the last time I met a new girl. Especially, the last time I met a girl and didn\'t make a total ass out of myself.  2 of My flaws: 1) Not thinking before I speak, and closely related, 2) Speaking whatever\'s on my mind.  (Usually this involves sex or trying to find out intimate information about the person I\'m sitting across from. which in turn makes them embarassed)  Then later I become annoying, and also, I don\'t pay enough attention to them. (Major flaw on my part, but I have trouble focusing)  The alarm clock comes soon, so I\'ll be going to bed. God Bless- T  '),(19033,317581,'04,March,2002','2002-03-04 00:00:00','   grief   It seems that grief is a very lonely feeling. No one knows how you feel and they try to empathize with you, but they don\'t know the hurt or the pain. So there is no comfort in friends. I guess having someone to talk with it about is good.      Can you see that? The emotional exhaustion? Can you see the dried tears? Can you tell when I don\'t feel up to snuff? Is it invisible? Does it appear as a hang over? Or something more believable?  '),(19034,317581,'01,March,2002','2002-03-01 00:00:00','   weekends?   So there was a time that I had weekends and there will be a time when I have them again. But until that day comes, I will just remember them as a good time in my life.  In other news, The D is coming through town in April, that\'ll be a rockin show!!  Sleep is a good thing- T  '),(19035,317581,'01,March,2002','2002-03-01 00:00:00','   in other news  First I\'d just like to say that when you all were talking about \"Lex\", all I could think of was the Sci-fi channel show \"Lexx\". but then I investigated and you all were talking about the show \"Smallville\".  Also, my coworker is having  urlLink dreams  about me.   HAHAHAHA  '),(19036,317581,'01,March,2002','2002-03-01 00:00:00','   Silence   I\'d like to think that when I don\'t write on here that I\'m spending hours at a cabin out in the woods somewhere. Unfortunately, what it usually means is that I\'m working my ass off and that I don\'t have much to write about.  Anyway, I\'m working through the weekend. Smiley face. I get overtime :) frowny face, I don\'t get to spend all the time this weekend with my friends:(  N-E-way, hope everyone is doing good, I\'ll check you latah- T  '),(19037,317581,'28,April,2002','2002-04-28 00:00:00','   The ladies that aren\'t   well, I\'m back at it. Actually it\'s not soooo bad. I\'ve been talking to girls at church, which is a positive direction for my life.  Anyway, been busy with work and what now. Just going with the flow if you know what I mean. I volunteered to write stuff for the church newsletter, but now I think what I\'ve written is crap, but they\'ll probably publish it anyway.  Well, gotta run- T  '),(19038,317581,'21,April,2002','2002-04-21 00:00:00','   lesse...  I saw \"Scorpion King\" on friday, which is a horrible movie, but I got to see a preview for \"undercover brother\", which just looks awesome. This week has been good. I think I figured out living arrangements for after june. Work is going good. Light house, my church thing on tuesdays is great.  Umm... I saw Adam on wednesday. That was fun. He lives in Tenino right now and so he\'s a lot closer than home.  Houck is up again this weekend. So he\'ll be stopping by around dinner time. YAY. The comic book thing last weekend was really good. If you go to my links for my other sites, you\'ll notice I discontinued service due to my lack of time to work on them.  Anyway, hope you all are doing good and I\'ll check yous latah- T  '),(19039,317581,'15,April,2002','2002-04-15 00:00:00','   ok, I think he probably thinks that he was disappointed that Safeway doesn\'t make sandwiches on sunday:( Not that you couldn\'t find them.   Anyway, saw FOTR for the last time in theatres. Well, at least last time for this quarter. Got back into FF X this weekend. Watched Zoolander and Training day with a bunch of friends.  Okay time to duct tape.  '),(19040,317581,'10,April,2002','2002-04-10 00:00:00','   I went to buy a sandwich.  There were no sandwiches.  Someone said there would be.  But there weren\'t.  He probably thinks, \"There were sandwiches.\"  But he\'s wrong.  There weren\'t any.  '),(19041,317581,'10,April,2002','2002-04-10 00:00:00','   Yo yo, sup doggs!  Big C is comin back in a big way!  OK, maybe it\'s not so big.    So I\'m quitting my job!  YIPPEE HURRAY!  Who needs money anyway right?  So does anyone read this blog anymore?  I know Tim still writes in it, so probably eh?  Obey my dog! Obay?  No that looks weird.  I think the first one was right but it just doesn\'t look like it.  Maybe it\'s because of the capital O.  obey   No that\'s pretty much just a weird looking word.  HAPPY, happy, HAPPY  Heh he he heh heh heh....  '),(19042,317581,'10,April,2002','2002-04-10 00:00:00','   no wonder   So it\'s april 10th. No significance in that. Brad\'s replacement started yesterday. Dash, I forget his real name. Funny how I can just refer to my coworkers by their game names. It works, but it\'s kinda weird. Especially at Red Robin, it seemed a bit wierd.  Anyway, Church group was good last night. Met some people finally. As you may or may not know, I\'m only an outgoing person when I want to be. And if I\'m comfortable, I may not be as friendly as I should be. So anyway, it was good to be more friendly and get to know people-ish.  I\'m going to see \"FOTR\" again on saturday. Did I mention that I LOVE that movie? :)  Also, Houck is coming up this weekend, and we\'re going to go to a comic book exchange on sunday. YAY:)  Anyway, for more excitement check  urlLink this  out.  Ciao- T  '),(19043,317581,'06,April,2002','2002-04-06 00:00:00','   the partah   Had the boys over again this weekend. Cleaned up last night. which didn\'t take long. Then went and bought pizzas today.1pm guys started showing up and we started eating and playing games till around 6 when we started Leprechaun 4: Leprechaun in Space. This was THE cheesiest movie I\'d seen in a while.  Anyway, it was fun, I drank some, and now I\'m going to watch Iron monkey, ciao  '),(19044,317581,'04,April,2002','2002-04-04 00:00:00','   The Thursday   Well, I found out last night that I would not be leaving Recon, but it\'s all good. They basically felt I could do the job, but that some other people were more qualified.So that\'s good. I\'m going to study more, before the next opportunity comes up.  Check you latah- T  '),(19045,317581,'02,April,2002','2002-04-02 00:00:00','   The interview   had my interview this morning. As I keep thinking about it, I am getting more doubts. I wish I had answered a question more organizationally. Anyway, hindsight is 20/20 and I\'m living the hindsight life. Last night the upstairs bathroom leaked water into my befroom until about 12:30am. I didn\'t find out about it till around 11:30 and I don\'t know how long it was leaking before that.  The waiting is the hard part.I can\'t help wondering, hoping, worrying... maybe I\'ll go to bed early and I\'ll have a nice surprise... (Christmas memories come back like a flash) No, no surprises, I just want the job.   *sigh*  '),(19046,317581,'18,May,2002','2002-05-18 00:00:00','   WORK   ugh, yesterday was \"chicken running around with it\'s head cut off\" day at work.At least it was for Tim. E3 is next week and everyone is fighting for my team this week. So somehow we got double booked, ok it was mostly my fault. Well, so next week the head honcho is out of town, so I\'ve got to be \"super on the ball tim\", which I\'m really good at:)  Anyway, it\'s saturday morning, and kevin bought an xbox last night:)  Gotta go- T  '),(19047,317581,'14,May,2002','2002-05-14 00:00:00','   Kids in the hall ROCKS   That was an awesome show. I highly recommend you get out and see these twisted guys before they\'re gone forever.   TRUE Ticket story: call ticketmaster... please hit 42543256739439-x971 to talk to a ticket person that can help you... Teller: Hello, can I get your license number, address, CC number, dob, urine sample... Teller: hi, mr. williams, I have you sitting in... blah blah blah Teller: how may I help you? Tim: Umm... yeah, I think I may have lost my tickets, or maybe my roommates threw it out Teller: Are you sure you received the tickets? Tim: (I know I received them) Well, I\'m pretty sure I did... Teller: Well, ticketmaster has a strict no replacement for lost or stolen tickets, but if you didn\'t receive them, I could have replacements at will call. Teller: are you sure you received them? Tim: No, not sure Teller: well, since you\'re not sure you got them, I\'ll have your tickets waiting at will call. Tim Thank you and may god bless your children  End true ticket story  '),(19048,317581,'14,May,2002','2002-05-14 00:00:00','   new jobby job   Well, I\'ve started working as team lead:) Yay.   I lost the tix for \"kids in the hall\", anyway, trying to get those back by the end of today, cause that\'s when the show is. Anyway, gotta run- T  '),(19049,317581,'10,May,2002','2002-05-10 00:00:00','   Went to Nate\'s last night. That was fun:)  Watched Survivor then CSI, then had fun talks about politics and religion.  The nice thing is Nate is still working in my building:)  '),(19050,317581,'10,May,2002','2002-05-10 00:00:00','   long time, no write   Well, it\'s nate\'s last day as recon lead. He\'s moving to a core team, and Jason hasn\'t announced who\'ll be replacing him (*fingers crossed*)  I\'ve started looking for a new place to live, we\'ve found a couple places westside to live. So that\'s fun.   Umm... I\'m going home for mother\'s day this weekend. well, gotta go to work...  '),(19051,317581,'23,June,2002','2002-06-23 00:00:00','   all good things   It\'s time to pack and say goodbye... Fuck the neighbors with their hateful stares...  and their damn suggestions... how bout this, You can go take care of your yard and live your perfect lives... But not ME... I will not mow the lawn, I WILL let the forest grow, and take my garden... Let nature do it\'s thing, I care not...  Yesterday I had my last party, much to my neighbors delight I suppose... too much fucking shit for them to deal with...  Is it too hard to live this life without boundries? Can you find a boundry that you have and won\'t cross that seems infintile? You can\'t go to a nude beach because why? You can\'t say the word \"fuck\" because why?  Are you listening to what others say? Can you not challenge your ideas? Will you not challenge your ideas?  Who says you\'re going to hell? I\'m pretty sure, last time I checked, that was up to God.  peace  '),(19052,317581,'09,June,2002','2002-06-09 00:00:00','   HOUSE FOUND!!!!   We found a house today. We\'ll be living in Ballard about 4 blocks from the main drag. We ended up eating lunch at the vegetarian Australian food place. It was so delicious and they had this dessert that was just SOOOOOOO good. I can see that it may be a very tempting place to eat once we move to the new place.  I HATE cutting it to 20 days before we move. I wished we could have found a new place sooner, but I\'m really happy with the place we\'ve found.  Peace, and hope you\'re having a good life- T  '),(19053,317581,'03,June,2002','2002-06-03 00:00:00','   long time no typey   Hello Everyone- well, I\'m back. I spent the weekend at dad\'s hittin golfballs and playin golf. Anyway, workin hard.   It\'s all good. I had a dream last night that I forgot to write a paper for one my history classes and I wasn\'t going to be able to graduate. But then I woke up and realized I\'d graduated like 2 years ago:) Whew.  Also, kevin and john and I are looking for a place to live and Jebe is back in town so all these things are going on. It\'s actually a great time to be alive:)  Catch you on the flip side- T  '),(19054,317581,'11,July,2002','2002-07-11 00:00:00','   moving again   So I\'ll be moving again. There are many reasons for this and I can\'t really go into it right now. Anyway, I stopped unpacking, as I\'ll just be packing again in a few weeks or so.  Anyway, I had a good talk to the man in charge today, sounds like he might be doing some work on finding his replacement tomorrow, maybe I should get my resume in order... hmm...   anyway, check yas latah- T  '),(19055,317581,'03,July,2002','2002-07-03 00:00:00','   MOVED!!!!   Well, we got moved this weekend. That was about the hardest thing I\'ve done in my life. Well, maybe not in my life, but it involved a total of like 17 people and 10 cars and 1 moving truck.   Anyway, now I\'m working hard, playin games. Did I ever tell you that I love my job? Anyway, I do.   Well, it\'s time to Rock, I\'ll catch you on the flip side...  '),(19056,317581,'27,August,2002','2002-08-27 00:00:00','   update   So cliff and hannah are married. And they\'re back from their honeymoon. It was a very emotional wedding, at least for us guys. We were all in tears, while the ladies kept a firm upper lip:) Hmmm... role reversal?  Anyway, I\'m moving to my new house next weekend. So I need to repack all the shit that I\'ve unpacked, but at least it isn\'t as much as last time.  So, I\'m still going to be lead for my group. I was working on something, but I was overshooting my experience. It was a GREAT experience and I\'ve come away a stronger person. That being said, I\'d like to get job offers for my 100 day break. I\'ll be accepting emails at my  urlLink email   If you\'re veisiting from malaysia, please have sarahkate email me cause I lost her link and stuff. Thanks  Peace- -T   PS- \"hi\"  '),(19057,317581,'01,August,2002','2002-08-01 00:00:00','   wedding   So cliff\'s getting married on friday. It\'s kinda hard to believe. I\'ve been working major overtime. So it\'s nice to have the time off.  Anyway, I\'m working right now, so I better go. Don\'t let the man get you down. -T  '),(19058,317581,'15,October,2002','2002-10-15 00:00:00','   travelling   I was going to come home and work on my Neverwinternights, but then I was like, actually I\'d like to watch some \"Bring it on\", so I did. Then I was like, ok, now I\'ll go upstairs and work on my NWN game, but now I\'m listening to some classic 80\'s rock and I don\'t want to turn it off. So, I\'m writing on my BLOG and thinking about travelling.  Today at lunch, one of the peeps that I work with started talking about NZ and europe and travelling in general and it stired my blood. There\'s just something about getting away for a few relaxing weeks and not having anything to think about. *sighs*  Anyway, I\'m skipping church tonight cause I\'m gonna go to bed earlier than it\'s going to get over with:( I\'m gonna go when I don\'t have to wake up so early. AKA december  gnight  '),(19059,317581,'14,October,2002','2002-10-14 00:00:00','   back up and kickin   Well, I\'m moved. I\'m still living in Seattle, but now I\'m living by the U district instead of ballard. It\'s a shorter commute and I\'ve been working for the last month. When I say I\'ve been working for the last month, I mean, 30 days straight no breaks. 2 weekends I worked 24 out of the 48 hours. It was hard, but the paychecks are nice.    N-E-way, I\'ll be taking november off to go and visit peeps, like steve and my pops. So, I hope to be inspired a little more often now that I have some free time.   Catch you latah- T  '),(19060,317581,'26,November,2002','2002-11-26 00:00:00','   update   I recently read that to keep a good *cartoon* site is to make regular updates. So I\'m gonna try and do that...  On that note, I tried to go to  urlLink justbecause.com  tonight and was disappointed to see that it was either a) not functioning or b) my silly internet provider was acting up again. Either way, I\'d like to see something there. I think it\'s supposed to be a flower shop, but I\'m not sure.  Also, I\'ve been out of work now for about 3 or 4 weeks. I had/have a \"on call\" job, but they aren\'t calling. So I\'m thinking of giving plasma. They\'ll give me 125 for 4 visits. that\'s 2 weeks worth of plasma giving and I get 125 bucks.  That would decrease the need for me to call my parents every few weeks to bail me out. Which I HATE doing. It\'s not that they don\'t help me, it just seems that at this point in my life I should have a stable enough income that I shouldn\'t have to ask my parents for anything..  Anyway, I\'m going home for thanksgiving. If you\'re reading this and would like me to maybe update a little more regularly, you can send me an email  urlLink here   Peace- T  '),(19061,317581,'22,November,2002','2002-11-22 00:00:00','   to all the peeps in recon land  Get back to work:) Bye  '),(19062,317581,'28,December,2002','2002-12-28 00:00:00','   Christmas   Well, I got through the 2 towers. I\'ve seen it 4 times now. I kinda liked it:)  Anyway, Christmas with the fam was fun. There were lots of little ones running around. Which gets us to the point that some people think I hate kids. Which isn\'t true. It\'s just when they\'re in large groups that I get overwhelmed and must seek refuge in a secluded room. (1 annoying thing was they kept going into my room which was off limits:P )  Anyway, I hope you all had a nice christmas and I\'ll talk to you latah... -T  '),(19063,317581,'16,December,2002','2002-12-16 00:00:00','   Cahlen\'s back?!?   Well, it\'s officially 47 hours from the start of the two towers at  urlLink The Cinerama  I\'m a little stoked.  So anyway, Cal was here last week and I\'m doing good. We played lots of games and watched lots of movies. Unlike most unemploywed bums:)  Well, I decided against giving plasma for the time being, mostly because I couldn\'t find anyone to go with me, but also because the state decided to give me a few extra dollars.   Well, I\'m going to run. Have a great week! -T  '),(19064,317581,'12,December,2002','2002-12-12 00:00:00','   Ei-Ei-Poo!!!  '),(19065,317581,'20,January,2003','2003-01-20 00:00:00','   Half way there   Although I did get 1 request to make this more than just a once a month blog, I have found that I\'m not that motivated.   Mostly because of games and movies: Tony Hawk Pro Skater 4 Warcraft 3 Neverwinter Nights Sid Meier\'s Sim Golf and I\'m watching all sorts of movies that my roommate Dan owns.  Tonight I watched... Undercover Brother, Pitch Black, and Life as a House. It\'s because of \"Life\" that I think I\'m writing here, now.  My roommate presented me with a term about someone they knew. They called them \"Asexual\" which I realized is how some of my friends and I live our lives. I have this huge fear of rejection ingrained in my being. I think it\'s made me rather prude when I think that there\'s lots of girls who wouldn\'t have rejected a kiss by yours truly.  I\'m 25 now. Still a virgin, although I think my mom doubts it. And I\'d love to get a girlfriend and settle down. It\'s just I don\'t know how... T  '),(19066,317581,'27,February,2003','2003-02-27 00:00:00','   Timmay!!!  '),(19067,317581,'14,July,2003','2003-07-14 00:00:00','   And now some blogger luvin   So, I\'ve been having computer issues recently. Seems my HD is going bad:(  Anyway, the other thing it\'s been doing is signing me in and out of messenger service like a cop in a donut shop :) Wait... am I mixing metaphors? or just not using the right one?  Anyway, life is good. I found salmon burgers and have fallen madly in love with their low carbs, omega 3 fatty acids, and all around tasty-ness. Now if I just had unlimited amounts of money I could just live on brocolli, salmon, asparagus, cheese, crystal light, and diet mt. dew for the rest of my days. I think I could handle that.  *sigh*  back to work- Tim  '),(19068,317581,'11,July,2003','2003-07-11 00:00:00','   All things hotmail   What is UP with the damn hotmail. I\'m like tryint to write an email to a friend and have to go look up an address. When I try and come back to hotmail, I usually hit a damn sidebar by accident, then the whole damn email disappears. It\'s like, \"look, FUCKERS, if you\'re gonna have advertising all over my email at least you could have it open a new window, so I don\'t have to remember what I just typed\"  BASTARDS  So I went and got a grill from the house last night. Ended up getting the charcoal grill as it was 50 bucks and I had just spent a load at costco on junk.   I\'ve been playing Star Wars galaxies fairly often. So that\'s something. Now if my room weren\'t like 5 billion degrees I could actually spend some of my time at home developing my character.  OK, it\'s work time. Peace out- Tim  '),(19069,317581,'09,July,2003','2003-07-09 00:00:00','   So long, so long, well, let\'s see, I\'m working at MS again. Playing videogames for the man:) I\'m still living in the University district here in Seattle washington. We\'re coming up on our year lease and I\'ve decided to live here another year.   We\'re having to replace two roommates... Cindy :( and Sean :)  So, I\'m kinda sad cindy is leaving. I\'m pretty sure we won\'t get a girl to replace her, and there\'s just something nice about living with a girl, you know? Plus I feel like I\'ve just gotten to know cindy and now she\'s moving to idaho:( Also, girls are clean, which is to say, much different than boys:)  and now on to sean...  OK, so if you\'re going to be an alcoholic, don\'t be an ass about it. And if you\'re going to fall off the wagon, don\'t do it in my living room. Got home thursday night and Sean was drunk. He then continued to be beligerent/drunk all weekend long. It was fuckin AWESOME. oh wait, was that sarcasm ;-)  Well, I should get back to work, if anyone reads this please email me:) I\'ve missed you all so much ;)  '),(19070,317581,'11,September,2003','2003-09-11 00:00:00','   The idea of human nature   So the question is this: Are people naturally good, or naturally evil, or a product of their environment?  By and large, I would have to say, I have felt that people are naturally good. I think I grew to believe this due to the fact that I grew up in the church and am naturally a kind and caring person.  This actually helped me have blind faith in my government and my church.    I have come to change these beliefs recently....  People are still dying in Iraq. People are still dying in Afghanistan. Not to change questions, but why? Why are we still there? Why is there death in the world? Why don\'t trickle down economics work? Why doesn\'t communism work? Are republicans the cynics? or the democrats? or is it the growing population of independents?  These are the questions I\'m dealing with right now...  Just thought you could use a little update Peace out- Tim  '),(19071,317581,'03,September,2003','2003-09-03 00:00:00','   Working   80 + hour weeks... Brain is fried, see you on the other side... Peace- T-dog  '),(19072,317581,'31,October,2003','2003-10-31 00:00:00','   Happy Halloween   So cahlen and peter went and got a new router last night. So that\'s cool, real internet for the whole house. YAY.  DAMN LINKSYS!!! NEVER buy one of their damn routers. Ever since we got that shit home it was cutting out every couple of days, until yesterday when it cut out entirely:P  So here we are. It\'s nate\'s birthday. I have 2 parties to go to and no costume. I\'ll feel like a chump, but what can you do? I think I\'ll be wearing my white and red \"Freak\" hat, and maybe a shit that says freak somewhere on it. Where as my roommate karla has a full get up.  Today is also, pay bills and get oil changed day. So that\'s exciting:)  It\'s been an excellent movie week. In the past 2 days I\'ve watched, Beautiful Girls, Amelie, and As good as it gets. I love the 2 compliments that he gives her in As good as it gets.   I\'m driving to pocatello next week to drive cindy back. ROAD TRIP!!! :) YAY!!! Now I just need to make some good mix CDs:)  OK, I\'ll catch you on the flip side-  '),(19073,317581,'30,October,2003','2003-10-30 00:00:00','   damn internet   The router and/or internet stopped working today:(  I have my computer directly connected to the dsl modem now. Talk about a pain in the ass.  Anyway, looks like I\'ll be home for my birthday after all.   Frustratingly yours- T  '),(19074,317581,'28,October,2003','2003-10-28 00:00:00','   Just a man   Before you judge me... Before you call me a slut... Before you hate me... remember that I am just a man.   I don\'t know who I am, but I know what I enjoy. time with friends, time reading my books, time spent self improving, these are some of the things I enjoy most. Travel, meeting new friends, watching a well made film, talking, kissing.  Thinking about the un-important things in life. Why are we here? What happens when we die? Is God real? Is there life after death? Is there a soul? Things that can really only be carefully considered after the basic needs are taken care of.  Lively debate with friends in a trusted environment.  These are the things I cherish, these are what make me a man...  '),(19075,317581,'27,October,2003','2003-10-27 00:00:00','   life and all those other things   It was a very good weekend. I think that has been one of my favorite weekends in a while. time is filling up rather quickly and I should have practiced the piano more. C\'est la vie.  I\'ll write more later today. I promise peace-  '),(19076,317581,'26,October,2003','2003-10-26 00:00:00','   weddings and what not   The wedding last night was nice. It was good to see a bunch of the old gang. I know that I have changed in more ways than they\'d like to think about. Sometimes I wonder if I\'d rather have a return to innocence.   Anyway, it turns out that my good friend Brian married Amy last month. I was really looking forward to going to their wedding. It makes me sad that we weren\'t able to get the contact information together so that my invite could show up and I could go to the wedding.   On the plus side, we got to set our clocks back an hour last night. So I got an extra hour of sleep. Kinda:)  Got the new Barenaked Ladies album, and so far my favorite song is 05 - for you. Peace-  '),(19077,317581,'25,October,2003','2003-10-25 00:00:00','   still horny   I got new tires today. And then played hella SSX 3. Anyway, I\'m still horny, so I haven\'t been put into the \"asexual\" category quite yet.  Just applied to my 3 jobs for the week. Anyway, I\'m going to bed early tonight. g\'night-  '),(19078,317581,'24,October,2003','2003-10-24 00:00:00','   When   I\'ve been trying to think of the last time I really fell in love was. I think it was sophomore year of college, to a cute freshman named julie. We read each others poetry. Then I think she kinda fell for me, whcih scared the living piss out of me. So I told her that she should date the boy that had been interested in her when I showed up. Then when he broke her heart, I was the one that took her out so she could grieve. Unfortunately for me, it was too late to have her fall for me. She married this past summer. I need to call her and see about a dinner.  Since that time sophomore year, I think I\'ve hidden my feeling better, made sure I didn\'t fall for them any more. I think for better or worse, I\'ve had to lose some compassion along the way. It\'s also caused me to have respect for women in general. Not to say that I should put every woman I meet on a pedistal, but  perhaps I shouldn\'t be thinking, \"How can I get in her pants?\"  Although, before I have always not thought about getting into her pants, and that hasn\'t gotten me anything but a plethura of female friends. Which is good, I like female friends, but I think I\'d prefer female lovers:)  All this to say that if you\'re my friend and you\'re female, there\'s a chance I want to get into your pants:) But not because I am in love with you :(  Catch you on the flip side-  '),(19079,317581,'23,October,2003','2003-10-23 00:00:00','   It\'s 5:30 am   I had a bachelor party that I went to tonight. It was very SPU-ish. i.e. we went to dinner at TGI Fridays cause hooters was too much of a sin for some of the guys. Then we went to a comedy club. That was actually pretty fun.  urlLink Comedy Underground  Then we went to Goldie\'s, which is a \"casino\" here in washington. All-in-all it was a VERY mild bachelor party. Cliff drove me around for it, so I drank a little.  Funny thing is, I got home and went to bed 4 hours ago. I just woke up extremely horny like 20 minutes ago and couldn\'t get back to sleep. I know I need to change something in my lifestyle as I don\'t want to wait until I\'ve lost my sexual drive before I hook up with a beautiful girl. I\'m concerned that I may wake up one morning and not realize that I\'m not hoping to have sex someday.  My thing is that there are so many diseases out there now that I\'m scared to go to a bar and try to pick up a woman. For one thing, what if it turns out she\'s a dude? Next there\'s the whole disease thing to worry about. My main problem I think would be my job. I\'ve heard jobs can be a great place to meet women. That\'s great, if there\'s more than 1 woman per 50 men.   So my friends are getting married on saturday. I\'m nervous for this marriage. Anyway, maybe I can see some old friends at the wedding and try and get some dates out of it:)  Hope everyone is doing good and much less horny than I am:) Peace out-  '),(19080,317581,'23,October,2003','2003-10-23 00:00:00','   war   A friend sent me this link today:   urlLink http://www.truthout.org/docs_03/102303C.shtml   As part of becoming more of who I am, I share it with you. In and of the fact that I think \"war is dumb\". And the fact that my friend adam will be in \"war torn\" iraq in less than 2 weeks.   Sometimes I wonder if I should think at all. It would be so much easier to follow blindly. To stick my head in the ground and say, \"I trust our leaders. They know what they\'re doing\". But I can\'t. I\'ve been in this world long enough to know the bullshit. To see the promise and then have it removed from before me.   Just thought I\'d share-  '),(19081,317581,'22,October,2003','2003-10-22 00:00:00','   men and women... again   I would have to argue that men need women more than women need men. I\'ve heard that with long term marriages, when a wife dies the husband will surely die within a year. It\'s true for johnny cash.  It puzzles me, partly that women do not know the power they hold over us, partly that when they do, they become someone that we do not want to be with. Or is it that some women can use sex as a powerful weapon. Where a man can assault with sex, he can hardly ever manipulate with it. I\'ve known men that would do just about anything to get into a womans pants.   Now I\'m babbling, it\'s 3:30 am and I\'m going to sleep.  Peace and love-  '),(19082,317581,'21,October,2003','2003-10-21 00:00:00','   Friends   I just wanted to thank you my friends. You are my chosen family. You are the people I love, even when I hate you for what you say or do. You are the people that define who I am and who I am to become.  You are there when I have my crisis of faith and understand me when I ask you what it would mean if there was no God.  Thank you for not letting me get lonely. And allowing me to be part of your life. Peace out-  '),(19083,317581,'21,October,2003','2003-10-21 00:00:00','   endings and beginnings   Here it is 3am on wednesday morning. Why am I still awake? I should have been asleep hours ago. I wish I could say I was reading or playing the piano. But I\'ve been playing video games. Distractions from the real. From the people we know and love.  Music that is on: Guster - Either way  Sleep? who needs it? I have another post to write after this one. Will I be asleep before 4 tonight? Does it even matter?  Parts of my life are coming to an end, others are beginning to blossom. Relationships change, evolve.  I AM the hurricane that is my life.   Peace-  '),(19084,317581,'21,October,2003','2003-10-21 00:00:00','   understanding   Is it that I am so strange? Are these the thoughts of every man, woman and child? Am I alone in my wonderment? I don\'t think so...  I think we often all deal with these thoughts. How is that we\'re so scared of each other? Trust is not easily earned today.   I often ponder if we men have less or more feelings than women. I don\'t think we can generalize like that. It is more of a personal thing. I know people that are lonely on both sides of the fence. I also know of people that are not feeling much but just living.  I think that in my past I have turned my feelings on and off. I know when I am scared of rejection I can try and turn my feelings off until trust can be earned. At other times I find it hard to breath for as little as pondering holding a girl\'s hand.  Well I\'m outa here-  '),(19085,317581,'20,October,2003','2003-10-20 00:00:00','   trust   Here I am, set almost naked before you. Trusting you. Wanting to be understood.  Quietly fearing that I\'ll die alone. Unwanted. Untrusted. Knowing in my heart to be a good man. As best a man I can be.  I just got done reading \"Speaker of the dead\", Orson Scott Card\'s second Ender Wiggin novel. I say just, but it\'s been 5 hours since I went to bed at 5 and it is now 10.  I think that maybe I should have more discretion on this little blogger thing. Be more secretive. Not tell you anything. But you\'re my journal, at least of my thoughts and ideas. A way to quickly go back and remember the days gone by and see where my friends were and to find new friends. I think I partly publiish these knowing that no one will ever see them, but knowing someone will.   This is myself laid bare and I hope you enjoy reading it. If you can understand everything I say in these few pages you\'ll probably understand me as close as anyone has ever tried. Peace out-  '),(19086,317581,'20,October,2003','2003-10-20 00:00:00','   updates   I realized that I haven\'t updated you on some rather important things on my life.  Last december I started getting thirsty and would aften end up peeing about 8 times during the night. Cindy made me promise to go see a doctor as I was also losing weight w/o even trying. I promised her that if I ever hit 200 lbs that I would get insurance and go see the doctor. In may I hit 200 lbs (a whole 60 lbs less than I had been in november, and about the weight I was when I entered college) and I went and got health insurance. June 2nd I went to see the doctor. He told me that I had probably diabetes, but he wouldn\'t rule out cancer. So I pee\'d in a cup and came back tuesday morning for my blood work to be drawn.  Friday it was confirmed that I had diabetes. About that time I went on a low/no carb diet. That lasted about 2 weeks. Now I\'m on a limited carb diet. A diebetic nurse thinks I\'m in my honeymoon period. No one wants to say it, but I\'ll probably be insulin dependent in the year. That is where you give yourself a shot of insulin before every meal.  I started working out there for a while. That\'s one way I can control my sugar. My buddy, Cindy, moved to idaho and I lost interest in working out. I forced my bosses to find low carb supplements for when they ordered dinner for our team while I was working. I have come to deal diabetes. When the time comes for insulin shots, I will be ready for that as well.  Now I\'m on my break. Cahlen and Karla are moved into our free rooms. (see cindy and sean moving out) Peter and Dan are in love with my friends that I\'ve asked to move in. We had a house party last friday night. It wasn\'t as rowdy as some of the previous parties, but it was good.  I think you\'re now mostly all up to date. Signing off as your diabetic friend-  '),(19087,317581,'17,October,2003','2003-10-17 00:00:00','   understanding   Why is it that I have trouble relating to people sometimes? Or maybe not relating, but why are people scared of me? Perhaps it is misconceptions, or maybe I am just a scary person.  What does it mean to be christian? What does it mean to be set apart? Is there something that says that boys and girls can\'t be friends? Even if there is sexual tension, what about the possibility of giving up on sex for the better good of keeping a friend?  I wish to be a salve that can take pain away and allow wounds to heal, but I cannot, if I am not allowed. Perhaps there is more than just that. I find great comfort in comforting others.  I know that looks matter. And I wish they didn\'t. I need help in the dressing department, as I don\'t think I dress myself well. And the female friends I have tend to agree with that.  Well, I gotta run, catch you on the flip side-  '),(19088,317581,'16,October,2003','2003-10-16 00:00:00','   life   What is love? Baby, don\'t hurt me...  Sometimes my ideas of love swing wildly about and I\'m left wondering what I know, or what I think I know.   I asked a girl to be my friend yesterday, but she hasn\'t responded. I don\'t know how often she checks her mail, or if it went into a trash bin automatically, or if she\'s scared I\'ll turn into a lunatic....  Can men and women just be friends? I enjoyed the ideas presented in \"When harry met sally\", I think they mostly are true.   I signed up for friendster the other day, but I still need to get pictures up. It seems to me, that no picture = no contacts. After all you could be some hideously ugly freak:)  OK, I\'m going to go look for a plug in for a friends website... Peace out-  '),(19089,317581,'15,October,2003','2003-10-15 00:00:00','   depression?   They laid off the rest of my team today. 2:20pm, you\'re out of here. It makes me sad. Especially for cliff. I know he was looking forward to working till June and they couldn\'t find work for him through the winter.  Matrix 2 showed up today. So I\'m going to watch that with Cliff tomorrow. I can already tell that I\'m going to have trouble staying focused on music and cleaning my house and getting a job. Especially since cliff is off:)  C\'est la vie... See you around, space cowboy-  '),(19090,317581,'14,October,2003','2003-10-14 00:00:00','   Things   It\'s 1:37 in the morning. All day I\'ve been reading ender\'s game by Orson Scott Card. I\'m 229 pages into it. I should have it finished by morning.  Of the things I know in this life, which aren\'t many, I know this for sure... I want to find someone to Love and to be fully Loved in return.  That\'s what I comes down to. I often wonder what I should be doing to find that special someone... What I might do to make someone like me... What happens if I find her and I\'ve already screwed up my life so bad that she doesn\'t want me?  Why do we have a desire to be accepted? Am I a bad man? Am I a good man?  Adam is going to Iraq on October 31st. This reason enough gives me enough pain/emotion to hate my president. To realize that one of my best friends may die, and there is nothing I can do about it. Who is Saddam Hussein to me? What is Iraq? Why are we there?   I started taking piano lessons. Self improvement or masturbation?  That is all- Tim  '),(19091,317581,'25,November,2003','2003-11-25 00:00:00','   Dreams, unemployment, car accidents and DND   I tried to start to be part of a D&D game last night. We made characters and are meeting next monday to start the quest.  Then...  I had a dream last night. I was having dinner with a bunch of friends. Karla and a boyfriend of hers that I had never seen were also eating with us. Karla was on the table next to us and he was sitting across from me. I was on the bench side and we had two 4-person tables tables pulled together for the gathering. Over the course of the meal, whenever I turned to talk to the table where karla and the friends were sitting, her boyfriend would put stuff in my drink. One time it was a lemon, One time it was fish bones. The second time, I yelled at him to stop messing with my drink. Karla asked me why I was freaking out, and I just yelled that she was trying to poison me. She tried to explain that it was a joke, but I just kept yelling \"You\'re trying to poison me\". She started crying and I woke up. After waking up, I recalled every time I have ever let my mother down and then I felt horrible inside.  I\'ve decided that my dream and the D&D are somehow connected and I will not be playing anymore.  I woke up at 7am this morning, for I had a meeting with the unemployment peeps. Just a \"Getting a job\" class, but I read, so that was fun. I\'m finishing up the \"Children of the Mind\" by Orson Scott Card, book 4 of the ender\'s series.  Driving back from hanging out with Cliff today, there was a car accident in front of my house. I don\'t know how it happened. I was thinking about blogging, then I look and see 2 cars spinning in unison in front of me. It was so weird.  Tonight we celebrate cahlens birthday. YAY!!! I got a 160GB HD for my birthday, so that\'s fun:)  Well, I\'m going to go practice the piano now.   Peace-  '),(19092,317581,'24,November,2003','2003-11-24 00:00:00','   piano lessons   today I have my wonderful piano lessons. Thanks for listening to me rant and rave last night. Sometimes I just need to go off on my achilles heel. Subway for lunch maybe? :) Peace-  '),(19093,317581,'17,November,2003','2003-11-17 00:00:00','   Achilles Heel   I\'d have to say that my one social problem (achilles heel) would have to be my voice. In everything I attempt to do, my voice is there. My lisp is there. My \"funny talking\" is there.  I think it makes me come across as stupid. I think it causes some people to take pitty on me, while on others it makes them pissed off if I\'m put into a leadership role ahead of them. Other people tend to think that I not know what I\'m getting myself into, as if I couldn\'t understand it.   Sometimes I wonder if I sound like a down syndrome person. I know that women find me completely harmless, not that I mind. I would like a mutual attraction.   No matter where I go I take my achilles heel with me.  It is a constant reminder that no matter how hard I try at something, certain people will never take me seriously, and certain people will hate me. Certain people will try with all their might to protect me, as if I can\'t deal with something myself.  Do I look in the mirror and see a down syndrome baby?   I hate my speech impediment. Of course it helps when I look up speech impediment on google and it come up with:   urlLink \'Speech impediments may manifest as stutters or lisps or may be more severe. Severe speech impediments are relatively more common among people with cerebral palsy and can be associated with other physical disabilities or deafness. Relatively minor speech impediments may have a basis in a psychological disability or mental illness, but this should not be assumed.\'   I guess this should mean that I\'m more employable?  Anyway, I should go. I have the piano to practice. Peace-  '),(19094,317581,'17,November,2003','2003-11-17 00:00:00','   diary   Sometimes I wonder if this online diary is doing more damage than any good that may come from it. Maybe I should give it up for a while. I don\'t know. Am I in love? Could I be?  What is relationship? Trust? commitment? I don\'t know... I know I appreciate all my friendships and will continue them till my friends have an intervention....  Well, my alarm is going off, I should practice piano for my last 2 hours before piano lessons.  Peace-  '),(19095,317581,'16,November,2003','2003-11-16 00:00:00','   mmm... fish   Sometimes I ask myself \"What should I be doing?\" Most of the time I just go with the flow though. So, when I do ponder the \"what\" question, it usually has something to be doing with my real life at the moment.  I\'ve been thinking about the situation that I\'m in and about hurting people.  I found out that I\'m not hurting someone I thought I was and so that gives me some relief. I know that through no conscience act I\'ve gotten to the point that I can hurt people by not doing anything wrong. Some say that I am the victim and that I\'m being preyed upon. That is bullshit. If I\'m hurting anyone, or am going to be hurt, I have done so with both eyes open, neither fearing nor blindly following. I have seen what I want. I know what I think I want.  Blogs...  Have I been to honest with you? Will my friends hate me? Will I destroy friendships? Will I leave and not look back? Will I find a happy medium of telling my blog only certain things and then keeping myself hidden?    I once told a girl that I would not ask her out for a number of reasons and she asked me why, and I told her. This of course would be after she had just told me that she was not interested in a dating relationship. After telling her my reasons for doubts, she would not tell me why she would not date me.  She told me it would hurt me too much. I felt betrayed and pissed off. I wondered how I could possibly work on traits that she found to be undate-able, if she would not tell me. I often wonder, if she knew that she pissed me off and that I would rather know, than not know. Or maybe it was that she was shallow? Could it be that she didn\'t find me attractive? in a physical sense?  Enough philosophy... Updates: Party went well yesterday. i think everyone had a good time, except maybe Dan who was trying to sleep. I went and found Cindy an apartment today. She\'s moving to Ballard in December. Dinner was at anthonys. I had some salmon and halibut. It was SO delicious.  Anyway, I need to practice my piano before I screw up royally tomorrow.  I\'ll catch you later-  '),(19096,317581,'15,November,2003','2003-11-15 00:00:00','   Humoungously busy   So...  I hate the feeling that you\'re letting EVERYONE down. Today, I was too impatient with Dan and I didn\'t set aside enough time for Peter.  These two things were at perpendicular to each other, and were unavoidable. I was supposed to play football with Peter, Dan said that Peter wasn\'t going to be home till 4, so I should go to lunch with him. Well, since Peter would be gone till 4pm I did go to lunch. But then he wanted to go to Target for a quick shop. Long story short, we get home at 3pm. Nate is coming over for beer at Teddys and to go to best buy. Lunch was at 11:30, I should have been fine, even when peter came home at 1, I should have been fine, but that DAMN target trip. Dan\'s pissed at me for being impatient, I\'m pissed at Dan for making me miss time with Peter, and Peter\'s pissed because I didn\'t make him the priority.  This IS a GREAT party atmosphere. I think I\'ve fucked up another friendship and I feel like I\'m about as comforting as a hurricane. I try and be the best man I can be, and I think people are still getting hurt.   And now I think I may have found some work. We\'ll see. Anyway, Master and Commander was great, and City of Lost children was awesome in that way that Amelie was awesome. Hopefully, I\'ll see you all tomorrow.  Peace-  '),(19097,317581,'12,November,2003','2003-11-12 00:00:00','   ugh   Life is beautiful. It\'s pay day today. So that\'s exciting. I came to a realization last night. I\'ve been creating a computer for about the last 6 months. I\'ve gotten hard drives, a sound card, a DVD+-RW, and a case. All that\'s left is Mobo, CPU, RAM and a video card. I can\'t pull the trigger, and I don\'t really care right now. I\'d much rather spend the 500 bucks on friends and good times than on a computer I don\'t need. I was playing games last night and realized that I was enjoying myself. If I can enjoy my games without upgrading and spending the extra dough, then why should I?  So I\'ll be upgrading my fashion sense instead. I have never spent much money on the packaging of Tim. I will now be buying a new wardrobe to go with the new slender Tim. Well, I need to go deposit the paycheck.   Catch you later-  '),(19098,317581,'11,November,2003','2003-11-11 00:00:00','   happy day after my birthday   I don\'t know why I get so moody sometimes. I think it must be that time of the month for me.  I\'m sleepy. Say hello to moody tim. Seriously, I\'m not usually this crabby. I don\'t know whats up. I don\'t have the words right now. I need to go hide.  Check you later-  '),(19099,317581,'10,November,2003','2003-11-10 00:00:00','   damn blogger and office   So I had this blog all written out and went to check on something in office. Anyway, office sent me to a website and killed my blog.  I was just thinking about how we are unable to go through life without hurting someone. I have never purposefully tried to hurt someone and I think I may have gotten into a situation where someone is going to get hurt. I wish I could find a way out where everyone involved could happy and not hurt. But I can\'t see the way. I\'m trying to get to a much more flexible life, where I\'m not planning things out 2 - 3 weeks in advance, but things keep on filling up.  It\'s my birthday.  Today we\'re going to ballard bar and grill. The matrix. and Tony Romas. Tomorrow is dinner and anime with Jarred. Thursday is dinner with Cindy. Saturday is party at my hizouse for Cahlen, karla and my birthdays. Should be a good time.  OK, I\'m going to burgers now. Check yous later-  '),(19100,317581,'09,November,2003','2003-11-09 00:00:00','   Life and something like it   And I\'m back... with a vengence:)  How is everyone doing? Did you miss me? Were there any emergencies you needed me to take care of? No? OK  So I updated my friendster profile. I don\'t know how to show you, so you\'ll just have to sign up to friendster and read about it.  Anyway, I arrived with my MS software having shown up. So that was rather AWESOME. OK, I\'m going to go for now.  I\'ll check you later-  '),(19101,317581,'04,November,2003','2003-11-04 00:00:00','   women 2.0   as you all know, generalizations do not hold true to every person. If you happen to be a woman reading this blog and it doesn\'t pertain to you. Then know that you are unique, at least in what I\'ve come to understand women. And if we are friends then it is probably true that you do not date mostly jerks as I don\'t like to associate myself with such women, nor do I have friends that are jerks, so they tend not to attract said women.   I\'m leaving for idaho in 7.5 hours. I\'ll catch you guys later. Peace-  '),(19102,317581,'03,November,2003','2003-11-03 00:00:00','   what we really want   I was thinking about wants today. Thinking about the feeling of wanting to be accepted and how different people play that out as different roles.  Sometimes I wish I had it all figured out. What I really want, what would fulfill my life and make me the man I want to be. But then I realize that\'s what everyone wants. A clear path where all the choice were already made and you just had to follow the steps correctly.  But then there would be no fun. Well, I need to pack. I\'m headed to idaho to bring Cindy back to Seattle tomorrow and am waking at 5am so I can get there in 1 day.  Peace-  '),(19103,317581,'03,November,2003','2003-11-03 00:00:00','   freezing   Here\'s what an interesting little quiz told me I was-    Protector    urlLink   The ULTIMATE personality test     brought to you by  urlLink Quizilla     Also it is freezing in seattle. Literally, and my single-pane windows aren\'t cutting the chill at all.   But it\'s nothing a warm body couldn\'t cure:) Peace-  '),(19104,317581,'02,November,2003','2003-11-02 00:00:00','   commenting   So I found a free commenter thingy, in case you didn\'t notice. Piano lessons were wonderful today. Unfortunately Jewel\'s book of music is a little too advanced for me right now. So I have Bach, minuet in G that I\'m practicing for my piece right now.  Tonight I\'m going to costco to find a new piano with weighted keys. Hopefully it won\'t be too expensive. Well, I need to get back to Xenocide.  Check ya later-  '),(19105,317581,'02,November,2003','2003-11-02 00:00:00','   what I\'ve learned about women   This is something I\'ve been thinking about, and which I may be totally wrong, but I wanted to share it with you.  From what I can tell, women want power and comfort. They often tell me that they want a man that will hold them and cuddle, but often end up with jerks. I\'ve often considered this very ironic and have been perplexed by my other sensitive brothers. We have often commented that the girls always go for the jerks and I think I have a theory why. I think that they see these men as the future leaders of our society at least subconsciencely. Women are drawn to ambitious men. I\'ve often heard them say that they would be with a man who knows what he wants.   I tell you that any man worth reproducing with, will not be so closed minded as only seeing a single future within his life. He will not expect to be the next president of the united states. He will not put down his fellow man in order for his own gain.  The women see that the jerk is a jerk from the get out. But they don\'t see a jerk. They see a man with ambition. A man that won\'t let another man get in his way, and he wants to be with her. Well why shouldn\'t she be with him. If he becomes rich and powerful, why not be there with him.  I\'ve often been perplexed by young girls dreams of being a princess or a long lost relative of some rich and powerful person. I\'ve never understood why they would want to give up their family for the power of a title. I think it has to do with society placing beauty before wisdom. Women see the princess as always getting her way, their way.  Sometimes I wish this were a discussion instead of a \"post\", or that I\'d at least be able to communicate with you. I\'m going to try and get a comments board up, but am having trouble finding one. Peace out-  '),(19106,317581,'02,November,2003','2003-11-02 00:00:00','   Mortality   I\'ve been reading Xenocide, by Orson Scott Card, tonight. I was planning on going to bed around 10ish. Well, that didn\'t seem to work:)  Some of the ideas and characters presented in these books are very emotionally straining on me for some reason. I often wonder how I can tear up at a stupid story. I\'ve always been impressed by peoples abilities to pull on heart strings through meer words and stories.  I am continually being challenged by my own mortality. I was thinking about how I\'m turning 26 in one week. I often ponder who I may have affected in the first 26 years of my life that would care if I were here or not. Of course, then I start to think of the people that I love.  The people that have affected me in such a way that I would truly mourn their passing into whatever may be after death. (Don\'t even get me started on life after death as this has become a serious question of my faith recently)  How is it that we get to be where we are and not continually try to be with the ones we love. I have such emotional controll over my life that I can sit here now and talk about these things with you, that I may never have to bring them up with the people that I trully care about. Sometimes I wish I were more in touch with my feelings that I could just do stuff without thinking of consequences. Or maybe I don\'t think of consequences correctly and so I don\'t allow my emotional to help me decide things. Overall I think I have a rather boring life. Won\'t someone help me?   Sometimes I wish I could not have self doubt and/or fear of rejection. I often wonder if I could live by some of the advice I give or if it\'s all just a lie.  A man described me as a punk thinker last night. i.e. punk because of my blue hair, and thinker because of my glasses. I think that maybe I\'m just trying to get new experiences. Do I not shave my beard because I think it might make me more attractive to the women I want to meet? or do I not shave my beard because I\'m lazy?  What do I want? I know that I need to find someone who won\'t judge me for my online journal. I am looking for what everyone is looking for. Someone who can accept me for who I am, even when I\'m trying to be cute and humorous and it comes out as disgusting and inappropriate. Someone who I don\'t have sexually repulsive:) All other points can be discussed-  '),(19107,317581,'02,November,2003','2003-11-02 00:00:00','   Life or something like it... and part of me   I have fully regressed to quiet introspective Tim. At times I will be outrageous party Tim, and at other times I\'ll be quiet introspective Tim. It\'s a cycle. Ups and downs. I first heard the idea of our selfs as cycles in Highschool, from a man I trully respect and value the opinion of.  So I\'m introspective Tim today. Unfortunately I have the little motivation to leave the hizouse these days. Which means my friends are missing me:(  Piano lessons are going great. I\'m going to get my book of jewels songs on piano. So that\'s exciting. I\'m going to idaho on wednesday to drive Cindy and bubba fett back to seattle.  So that\'s something. I got the new outkast, ludacris, and Meat loaf on CD from bestbuy today. Went to the JCPenny and got flannel sheets.  It is officially freezing in washington. It snowed this morning around 8am for about half an hour. I was asleep, of course, but I heard about it.   There was a birthday party last night. I didn\'t bring a present:( and I didn\'t give a present.  I did promote our party on the 15th, but that\'s not saying much.  I need a piano or at least a weighted keyboard to practice on:( so much to spend my money on. So little money.   Part of me wants to be a contributing part of society, part of me just wants to sit back and make snide remarks. Part of me wants a girlfriend so bad. Part of me wants someone to fuck. Part of me cares so much. Part of me doesn\'t care at all. Part of me wants a \"nice\" girl. Part of me wants a girl to tell me how to please her. Part of me wants to be good with the ladies. Part of me wants the ladies to see what\'s good in me and want to have that.  Part of me feels like Charlie Kaufman in Adaptation. Part of me feels like Barry Egan in Punch-drunk love.  That is enough for now... Back to Xenocide by Orson Scott Card. Peace-  '),(19108,317581,'23,December,2003','2003-12-23 00:00:00','   Revolutions   Can it be that I am not fired up or passionate about any discrepancies that I see? Should I drop everything and change my life? Move away and become a hard worker that has a passion for life? Am I merely asleep at the wheel? Am I dead?  Sometimes I wish I had a revolution I could be taking part in. Maybe there is an evil tyrannical man oppressing people somewhere that I could be fighting against. Is there a way I can divert my energy to help people? Instead of sleeping my days away in tired America?  Am I too fearful to be passionate? Am I too drowned in debt to try?  '),(19109,317581,'22,December,2003','2003-12-22 00:00:00','   New Years Resolution 3.0 Rev 1   Save 1 night a month for alone time  Time will be spent alone in room w/o computer   OR  Time will be spent alone outside of the house  It\'s almost christmas time and I\'m heading to Oregon. I hope everyone has a nice christmas and I\'ll catch you later. Peace-  '),(19110,317581,'22,December,2003','2003-12-22 00:00:00','   what\'s going on mikon?   So another day, another drachma, Christmas shopping is done, it\'s so nice to be able to buy presents again this year.  New years resolution 1.0: Pay off 1 credit card  New years resolution 2.0: Go on at least 1 date every month  New years resolution 3.0: Save 1 night a month for housemates/alone time  New years resolution 4.0: Read 1 book a month  I think that\'s good for now. I\'m going to go check out what\'s going on downstairs now. Check you later. Peace-  '),(19111,317581,'19,December,2003','2003-12-19 00:00:00','   Party 2nite   sorry for freaking out all over you like that. I am a somabitch:)  Trogdor was a man, err... maybe he was a dragon man, well, actually he was a DRAGON!!!  But he was still TROGDOR!!!!   urlLink TROGDOR   I hope you all have a nice friday, I really need to go back to work:) Peace- Tim  '),(19112,317581,'18,December,2003','2003-12-18 00:00:00','   true obsession   do you know that feeling?...  Should I call? Does she like me? Do I like her? Am I playing the game forrectly? Am I suffocating her? Am I enough for her? Am I doing the right thing? Am I just fooling myself until she finds out she doesn\'t want me? Do I want her? Should I ask her out? Will she say yes? Do I want her to say yes? Should I continue looking? Do I want/need to continue looking? Is there such a thing as a good ending? Who am I? Where am I? What should I be doing? What do I want to do?   Inner strugglings... peace- Tim  '),(19113,317581,'18,December,2003','2003-12-18 00:00:00','   Life is beautiful and so are you   Well, it\'s been a while. I decided against remaining in the D&D game. Life is good. Christmas is almost here. I\'ve decided I\'m going back to Oregon on tuesday and will be back up here on friday.  I just finished \"The Tao of Pooh\" which talks about Tao-ism and how the characters in Winnie the pooh relate to Taoism. It was very enlightening. Tao-ism seems to be a very go with the flow of life idea, which is what I like about it.  Anyway, I went to Trillogy tuesday (all the Lord of the rings shows on tuesday december 16th.) It was pretty good, and I feel like i got my nerd-ness fill for the rest of the year:)  Well, gotta run, Peace- Tim  '),(19114,317581,'30,January,2004','2004-01-30 00:00:00','   Nothing to say?   So I have 3 days of vacation left. I\'m falling behind on my piano work. I want to be an excellent piano player, but it\'s hard to find the time sometimes. (also hard to make it a priority when you\'re trying to get through a difficult piece)  Anyway, the problem with match.com is that you have to pay to use it. You don\'t have to pay to put up a profile you just have to pay to email. Of course I\'m not sure I\'m in a get to know new people part of my life yet. But there\'s never time like the present.  While I\'m writing this, I\'m listening to  urlLink Brak tell it like it is .  Work starts in 3 days. So much to do, so little time. should I shave before I go back? Should I pay for match.com or move on to a different service? Should I go into hibernation while I work for a while?  Maybe I should just listen to some nice tunes and smoke a bowl:) Peace-  '),(19115,317581,'29,January,2004','2004-01-29 00:00:00','   The \"am I a good person?\" question   I had a friend ask me if I thought they were a good person. Of course I said yes, but couldn\'t offer any substantial proof.  I think the question is more relative than anything. Am I good compared to the next person? Or am I good compared to Charles Manson? Or am I good compared to Jesus Christ of Nazareth?  I think before you ask if you\'re a good person, you need to decide what you\'re shooting for. Am I good compared to ghandi? Probably not?  Being a good person is a daily struggle. You take 1 thing that makes you a good person and I could maybe point out 1 thing that makes you a bad person. There are no good/bad people. There are just people.  If you aren\'t a child molester/rapist/alcoholic/non-motivated/drug addict/constant complainer/murderer/stealing candy from babies person then you\'re pretty damn good in my book. If people aren\'t constantly leaving the room when you enter, then you probably have friends somewhere:)  Be content to know that you are loved beyond your ability to comprehend. And love your neighbor as yourself.  '),(19116,317581,'27,January,2004','2004-01-27 00:00:00','   What\'s up?   So I have 5 days of freedom left. Not that I\'m going to mind work at all. It\'ll be nice to have something take up my day time hours. Although I may miss the freedom to fo whatever I want I thrive in a work environment, and I really like to thrive.  Well, I need to go shower. Peace-  '),(19117,317581,'26,January,2004','2004-01-26 00:00:00','   piano & competition   I think that on some level all people want to have a boyfriend/girlfriend that is sought after by others. At least I think attraction partially works due to outside influence. I saw a movie, Iris, where the heroine chose the nerdy guy, I think as a safe bet, partially because she had already had the wild lifestyle she wanted. So that movie actually goes against what I just said, but I think that it is due to the fact that having a greatly desired mate by others is not the only reason people hook up. I think sometimes people get together because they are lonely awnd find someone they can connect with on a deeply spiritual and emotional level, and mutual attraction helps.  I blog and match.com, and take my female friends out to coffee. Not such a bad way to live.   Well, I need to get ready for my piano lesson. I\'ll check you guys on the flip side, Peace-  '),(19118,317581,'22,January,2004','2004-01-22 00:00:00','   I slip on it everytime   Here I am. I have 10 days of freedom before going back to \"the daily grind\" which isn\'t really a grind at all, but it\'s still work.  So I signed up on match.com the other day. My profile is shyguy1177. In case anyone was interested in looking at it. When I was testing out some of the features I accidentally sent a \"wink\" which says that you\'re interested in meeting someone. It was just the person at the top of my list. Turns out she has kids living at home. It was a little awkward, me telling some total stranger mom that I liked her. I think if I were testing match.com, I\'d have them put in a confirmation or something.  Well, I\'m off to see the dr. today. Evidentally diabetics\' eyes have a tendency to go bad, so I need to go see an eye dr. I hope my new glucometer comes soon.  Anyway, enough of my randomness. I\'m going to go shower now. Peace-  '),(19119,317581,'17,January,2004','2004-01-17 00:00:00','   personalities   I will never know some people. I will never understand other people. I had a friend, Justin. His personality was semi-cruel. I think that\'s part of it. I don\'t understand cruel people. But the real question, I guess, is \"Do I understand myself?\"  I have been given a disease that will eventually destroy my body. I\'ve inherited it from my father. I will pass it on to my children. My base root self needs to procreate and I would like to have an heir. I\'ve watched my father disintegrate before my eyes. He will die like Johnny Cash, but hopefully not to soon.  I wish I had a love of my life. I am a HOPELESS romantic. I know I could be a good lover. Now all I need to do, is invest a little more time in a lady\'s life.  Well, I\'m off to Oregon, Peace-  '),(19120,317581,'05,January,2004','2004-01-05 00:00:00','   vivacity   Am I living in teenage angst? Who am I? Have I managed to live myself into a hole? Am I the only one who can find my way out?  I met this girl. She\'s really cute, and awesome to be around. Unfortunately, she has to go away. She inspires me to be more than I am. When I think about her, it makes me want to be a better man. It makes me feel like I\'ve been floating with the current and nothing has challenged me in a long time, nor have I challenged myself.  What do all people want? To fall in love with someone, and have that person fall in love with them. Unfortunately then the inevitable happens. Life doesn\'t end or stand still, it just gets annoying. You never see the \"13 years later\" in the movies. He\'s devoted himself to a job that he hates because they have a house they can\'t afford and she\'s banging the milk man, since he\'s at the house more often than her husband. I believe the reason that Romeo & Juliet is such a great love story is because they don\'t live long enough to fall out of love.  My ideal candidate for girlfriend: able to travel up for anything cute not hampered by prejudice not hampered by fear  Am I cynical, a realist, or a dreamer? I think a bit of all of these things rolled into a complex little mammal that wishes to procreate:)  Today I write to the blackhole of the internet, not knowing if anyone is reading...  If you\'re out there and you\'re reading this, please email me and tell me hi... peace-  '),(19121,317581,'05,January,2004','2004-01-05 00:00:00','   How YOU doin?   I am officially invited back to work:) YAY!!! (*sense of relief washes over Tim*)  How about them mariners anyway!!! Details to follow. Peace-  '),(19122,317581,'29,February,2004','2004-02-29 00:00:00','   Scorpios and the like   I wish I didn\'t have to turn off my heart. I wish I could be loving and supportive and not jealous. I wish I was on a similar life journey. I wish I weren\'t just starting out. I wish I could trust myself to not have an irrational jealousy. I wish I could say, goodluck, I want to hear all the details. I want you to be happy no matter what. I wish I could say, Oh he sounds like quite a catch and not hate him with every fiber of my being. I wish I could stay close intimate friends. I know that my love lasts forever. At least I know it\'s lasted since 13, and I am prepared to carry love in my heart for another 100 + years of my life.  I\'d be lying if I said I didn\'t miss my friends deeply. We are spread out by jealousy and a few oceans. I was asked what I would do if I can do anything I was tonight, no penalties, no reprisals, and the cost is unimportant. What would I do? My answer was the I would call ALL my friends (past & present) and schedule a party at a local restaurant. We would have dinner and chat. I would make sure they knew they could bring their friends and I would make new friends and be able to catch up with all my old friends.  Trully, that is what I dream that heaven would be like. Hanging with all the people I\'ve been friends with and just talking and telling stories. Course my life goal is to have people say that I was a good man and a good friend, a good father and an excellent husband. But maybe I\'ll never know about that. All I can do is be the best man I know how to be, and that means hurting sometimes.  Well, I should go to bed and/or practice the piano. Oh, and I got new glasses. Peace-  '),(19123,317581,'27,February,2004','2004-02-27 00:00:00','   all about me   So I\'ve been reading through my posts and I began to think what a pathetic guy I must seem like. I can\'t even handle being alone with myself. It seems to me, I need to spend some time alone. Or at most, just hanging out with guys not even thinking about girls. That would seem like the only logical thing to do, especially if I don\'t want a rebound relationship. I think it would be best to heal. I mean honestly, I feel as though I should be alone for a while. I don\'t want to be one of those people that NEEDS to be in a relationship. I want to be the kind of person that if a relationship comes along I can be like, sure, but not strangle it.   It\'s like I would really like to be in love with somebody, but I don\'t want to just be in love with anybody. Does that make any sense? AKA, I would like to have someone to love, but it DOES make a difference on who that person is. I\'m not just going to want to be in a relationship with any ol\' girl. I think the healing process requires lots of piano and video games. Maybe even throwing myself into my work for a while.  Well, I\'m off to practice the piano.  Peace-  '),(19124,317581,'26,February,2004','2004-02-26 00:00:00','   good morning   I woke up on my own this morning. 7:20am without an alarm and after staying up till 1am.... Well, my tester is officially lost, but I called and ordered a replacement this morning. YAY. Practiced piano last night. I love the piano, it\'s so fun to just sit there and create music. Well, it\'s time to work.  Peace-  '),(19125,317581,'26,February,2004','2004-02-26 00:00:00','   overall   I\'m glad to have been in the relationship. She is someone I will forever cherish. Now I need to figure out a way to move it to a friendship. I need to not freak out and not be in love and not rebound.Is there a genie out there that can give me a \"no feeling\" pill? or a \"stop liking that way\" pill? Well, it looks like I need to order another test kit from the insurance company. I\'m going to bed. Peace-  '),(19126,317581,'26,February,2004','2004-02-26 00:00:00','   meanings   I never meant to hurt anyone. I never meant to be hurt. I don\'t want anyone to be sad, and I don\'t want to be sad. I want to fulfill and be fulfilled.I want to love and be loved. I want to heal and be healed. I want to take the pain of the world on my shoulders and have no one hurt anymore.   I wish I didn\'t hurt you...  OK, enough of the dark ranting.  The next time you see me, I\'ll be a happy camper. well, maybe.  peace-  '),(19127,317581,'25,February,2004','2004-02-25 00:00:00','   My last sisterly update unless something terrible happens   My step sister seems to be doing much better. Her short term memory is having problems, but she  is making great progress considering how close to death she was last week. I\'ll update you if she worsens :(   PEace-  '),(19128,317581,'23,February,2004','2004-02-23 00:00:00','   *sigh*   I need to stop. I need to slow down, enjoy the flowers, take moonlit walks, go see a play. I need to get out of the house. I need to move on. I need to not care. I need to tear a little bit. I need to live. I need to breathe. I need to move on. I need to not obsess. I need to let go. I need to go away. I need to cry. I need to hold out. I need to not rebound. I need to sleep. I need to interact. I need to read. I need to engross myself. I need to find passion. I need to be cool. I need to be strong. I need to not cry. I need to be me. I need to grow. I need to be me. I need to find me. I need to not sigh.  Torn-  '),(19129,317581,'23,February,2004','2004-02-23 00:00:00','   updates   My step-sister is doing better. She is answering questions and can hold conversations ok. When she\'s not drifting in and out of sleep.  I\'m doing better. I learning to let go and that\'s something that is good for me to learn. I hope you all have a beautiful day.   Peace-  '),(19130,317581,'23,February,2004','2004-02-23 00:00:00','   on relationships and women   So here I am basically coming out of my first serious relationship ever. Overall, I\'m glad I had it and am still great friends with the girl. Turning off love is hard though. As a Scorpio, I often get jealous over future boyfriends which is ridiculous. I think mostly I am sad that I cannot be with her anymore. I won\'t be able to enjoy the journey anymore. I feel like I\'m getting left behind. But even as I write that, I think of how I have my own journey to take up.  I\'m a bit torn. I don\'t want to give her up, but I also want to move on and start up the next relationship. Mainly I want to fill up the hole in my heart. But I don\'t want to just have a girlfriend that is a patch. No rebounds. I don\'t want a rebound. I just want an open honest relationship with no baggage.  Then there\'s the point of, is she easily replaceable? Hell NO! But is there a salve that won\'t hurt anyone? No. There\'s the life of the monk, but then she\'ll feel guilty. There\'s the life of the party-er, but then I and she will only despise me as a slut. (*would it be easier if I were secretly a jerk that only cared about 1 thing?*)  Anyway, I wish God hadn\'t made women so irresistible :) But I\'m glad he did. I\'m going to leave you on an up note today, while I try and sleep earlier tonight.  So check out this link which is a hilarious account of the  urlLink End of the world . (*warning will take a long time to open*)  Music: Guster Peace-  '),(19131,317581,'21,February,2004','2004-02-21 00:00:00','   updates   My Step sister is doing better. She\'s been doing some talking and can remember all her children except the one that was born 4 days before her first stroke. She remembers her husband but not the fact that he has been to visit her.  Did I mention that I\'m back on the market? Well, I guess I am. So, lookout!  Peace-  '),(19132,317581,'21,February,2004','2004-02-21 00:00:00','   Life   Well the \'rents are in town. So today was VERY filling. Claim Jumper, Sopranoes. It was all yummy. Even made it to archie Mcphees.  Anyway, no updates on the step sister as my updates come from my parents and they\'re up here. So I\'m going to go to bed and wake up soon, so I\'ll catch you later.  Peace-  '),(19133,317581,'20,February,2004','2004-02-20 00:00:00','   Sisterly update   My mom is on her way up here, but the update goes like this:  \"Heather responded to the ICU nurse this morning. They did a CAT scan early this morning and found that the original bleeding point in Heather\'s brain was shrinking down to normal size.  The latest bleeding point is still large but the doctor is truly amazed that Heather has responded to the heprin so quickly and  the results are better than he could have hoped.  Heather remains in ICU and treatment continues.\"  Thank you for your continued prayers and support.  Peace-  '),(19134,317581,'20,February,2004','2004-02-20 00:00:00','   what a week   Well, it\'s friday. I left my phone at home. So if you need to get ahold of me today, you need to email. It\'s just 1 more thing on a list of things that have happened on an ultimately frustrating week.  Now I find out I get to have the house to myself pretty much this weekend. Not really looking forward to that. C\'est la vie.  If you\'re reading this, email me and say \"hi\". I can always use a little mid day pick me up.  urlLink EMAIL   Peace-  '),(19135,317581,'19,February,2004','2004-02-19 00:00:00','   Emotional Rollercoaster of nothingness   Well, the latest news being that...   \"The doctors ordered an angiogram of my stepsister\'s brain and it revealed several clots. They are starting a regime of heprin (an anticoagulant) and will continue for the next 6 weeks if she continues to live.  If things are still working out at the end of that time, they will continue the anticoagulant for another 6 months and, hopefully, the clots will dissolve.\"  That being said, I\'m on a bit of an emotional roller coaster right now. I don\'t know how I feel about my step-sister, except I know I don\'t want her to die. It\'s been an emotional week besides that, I\'ve had to give up one of my closest friends and that hurt.    Work is work, and it does a good job of distracting me during the day.  I hope all is well in your life-  '),(19136,317581,'19,February,2004','2004-02-19 00:00:00','   mortality   My mom called to fill me in on the newest member of our family today. It seems my married step sister had her baby last thursday. But that wasn\'t the news. The news is that my step sister is trying to survive the night. It seems she keeps having strokes (bleeding in the brain).  Needless to say, this has put a damper on my day.  Maybe I\'ll throw myself into my work for a while.  Peace-  '),(19137,317581,'16,February,2004','2004-02-16 00:00:00','   Personalities, one more time   So it seems to me that I appear to be an  urlLink INFJ or Introverted INtuitive Feeling Judging personality .  It seems to be that I am that person. I feel and judge and seem cold and uncaring.  \"Indeed, because of their strong ability to take into themselves the feelings of others, Counselors can be hurt rather easily by those around them, which, perhaps, is one reason why they tend to be private people, mutely withdrawing from human contact\"  Sometimes, I get overwhelmed by the beauty of the world and I retreat. Please forgive me.  In other news, I love my job. Peace-  '),(19138,317581,'15,February,2004','2004-02-15 00:00:00','   relationships that last forever   So, I have some friend leaving for Vancouver in August. I was recently asked if I\'d miss them, and at the moment I was a bit indifferent. Not that I won\'t miss them horribly, but we\'re only a phone call or an instant message or email away. In a society of instant messages, cell phones, and email, it\'s difficult to feel that someone has moved away.  I will never turn down a friend. NEVER. I\'m not the type of person that would let down a friend. If something is within my ability to do for a friend I will do it.  On another note, There\'s a FUCKIN FLY IN MY ROOM!!! It\'s like 50 degrees out, and there\'s a fly in my room! WTF!!!!  Anyway, my heart is big enough for all of you. Let me in, let me care and you will never need to wonder if I care. I do, I will, I am. I cannot change who I am, or can I?  Time for bed, Time for work, Peace-  '),(19139,317581,'12,February,2004','2004-02-12 00:00:00','   A brief stop in...   and then I\'m off again. I hope everyone is having a good life. I think I am. I\'m getting away for the weekend which will be nice.  My job is going wonderfully. I am going in early tomorrow so I can take off early for my weekend away.  I wonder if it\'s time for something new. I\'m a wee bit tired, so I\'m going to go to bed soon. Thus my scatter brain post. Anyway, I just wanted to make a note that we all wanted to see janet jacksons boobs and fuck those prude assholes for suing everyone. It\'s a fuckin boob. Half the population has em.  Anyway, we\'re looking for a roommate. Karla has to move out. If you\'re reading this and want to be an active participant you can email me and we\'ll see about getting you a room here :P  Peace out, my ever loving public-  '),(19140,317581,'01,February,2004','2004-02-01 00:00:00','   You can\'t handle the truth   Here it is, approx. 11 hours before I start my new job and I\'m kinda nervous, kinda excited, kinda confused, kinda feel like I should have practiced my piano more this week.  Here\'s a question... how long before you know someone, do you tell them about your blog? I mean obviously there are people you can never tell. (*mothers/pastors/children*)  Should you let your friends know your personal thoughts? You online diary?   Should you hide it and not let anyone else know? Should you be up front about it, and let them decide if you are worthy of being a good friend by giving them a full picture? Or do you let them only see who you are now?   People change. Obviously. But I AM still the man who thinks that cahlen should have said  urlLink fuck   Wow, my history in a nutshell... put on display for the whole world to read. My friends don\'t write here anymore. But, I\'m not really surprised. We\'ve all moved on, basically.  I\'m living in an entirely new house, but Cahlen is living downstairs now. Some people that have blogged here, I don\'t even know anymore.  I\'m sure as my life moves on... this, too, shall pass.  Perhaps I will fall madly in love with someone. Maybe falling madly in love only happens in the movies. :)  We\'ll see, I\'ll try and keep you posted as the adventure continues... Peace-  '),(19141,317581,'31,March,2004','2004-03-31 00:00:00','   \'in love\' vs \'love\'   Trying to move from \'in love\' to \'love\' is always an interesting experience. Trying to kill the jealous scorpio is tough too. A friend described someone as an emotional basketcase, was/is that me? Is that a valid description of me.  Growing and learning to move from \'in love\' to \'love\'. I need to be able to love a person that I was in love with, without being hurt. I need to not harden my heart. I need to remain emotionally vulnerable and allow myself to grow.  A work in progress- Tim  '),(19142,317581,'30,March,2004','2004-03-30 00:00:00','   changing   Looking at the last weeks worth of entries, it\'s very interesting to see who I am now, to the person I was 1 week ago tonight. I definitely seem more emotionally stable right now. Meeting new people helps. Going out at night helps too. Throwing myself into my work helps. Dealing with my feelings helps too.  Well, I need to sleep... Peace- Tim  '),(19143,317581,'30,March,2004','2004-03-30 00:00:00','   lookin to chill   Hey all- Is it so hard to just chill? I\'m looking for someone to just chill and relax with. It seems that most people I know are just looking for the next thing, they don\'t want to slow down and enjoy the ride. I\'m looking to get out of the house, but I want some people that want to get out of the house too. Life\'s too short to watch TV:) If you\'re free give me a call and we\'ll chill...  Peace- Tim  '),(19144,317581,'28,March,2004','2004-03-28 00:00:00','   Life as an MMORPG   The idea being that you are who you represent you are. You can represent yourself as anything. You can fool anyone into believing whatever you want them to believe, because it\'s all not real. You are the only one who knows the real you, and you can hide that information forever. You can bluff anything. You can\'t know anything. We play videogames, but we do seem to be just our brains, and they\'re just playing who we are as people./  It\'s not a world I want to live in. I want to be who I am and present myself as me. I don\'t want to have to fake who I am. I don\'t want to appear to change me to better represent a compatible person. I do want to be the best me I can be and present that in the most honest person that I am.  Just some thoughts on me- Tim  '),(19145,317581,'26,March,2004','2004-03-26 00:00:00','   who would have thought      Heart of Gold    urlLink   What is Your Heart REALLY Made of?     brought to you by  urlLink Quizilla  '),(19146,317581,'25,March,2004','2004-03-25 00:00:00','   growing pains   So I am standing on my feet now. Walking away from the relationship. \"Growing up\", if you will. I don\'t know why I became so needy at the end and I would like to think that I\'ve grown.   It seems right now that I will always love you, but I\'m sure that is changing right now, even as I write this.   I am not looking to get back into a relationship right now, but I am looking for some new people to hang out with.   I hope that we can remain friends after this, but I know right now that I don\'t want to be without you, so that is what I must do.  Peace, Tim  '),(19147,317581,'24,March,2004','2004-03-24 00:00:00','   no seriously   Moving on sucks. It\'s like I want to be able to say, you can\'t hurt me any more. But I also want to try to stay caring and there for when she might need you.  Anyway, time to get a jump start on moving on...  '),(19148,317581,'23,March,2004','2004-03-23 00:00:00','   the walled up side   There is me as my asshole self. The one that wants to be in this state of uncaring. The one that just wants to say, \"FUCK YOU\" and run away. Both of me exist in this one body.  Now to move on to not caring. Fuck allowing myself to be constantly hurt. Time to move on. Time to hide the caring sensitive side and present the uncaring asshole that protects me.  '),(19149,317581,'23,March,2004','2004-03-23 00:00:00','   Who am I, Who I am...   I am an INFJ to the deepest meaning of the word it seems. I will keep you all at arms length. Will you know who I am, besides who I show you? Will I let you see the little boy that will do anything to protect himself from being hurt? No, probably not. There are a million walls to my heart and I can only show you part of it at any one time.  Love to me is letting her in. Caring to the point of being utterly destroyed when she dies. Finding a life worth loving and doing everything within my power to love that woman. I will let her see who I am and hope that she can love me for that and will trust her completely to be gentle. Finding a life worth loving and doing everything imaginable to maintain that love.  Moving back from love is where I exist in daily life. There are the walls and the disappointment when the person dies, but no personal suffering. Protection from the caring. Protection from the guilt of not being able to give people money to get back on their feet. Protection from the guilt of not being able to care for the most basic of human needs of those in need. Numbness in all that is my life. Approaching on the ability to feel nothing, thus destroying any happiness and any pain.  Removing people from my heart is ALWAYS painful. I have found only 1 way. Remove them from my life. Move the walls back in place in my heart. It doesn\'t work though. I can think of no girl that actually found my love that I have truly been able to remove. My being longs for them to say they love me. To come to me and see the man I will be and love me for that. I long for acceptance with all my being and have never removed a girl from my heart that I have loved.   Unfortunately, to date, that means that every girl I have ever loved has been removed from my life, indefinitely. If we hang out, it is never in the capacity that I will allow myself to be seen, not in the capacity that they have known me.  I hide behind the walls, because I am truly a little boy that cares beyond my ability and am hurt immeasurably by the smallest pin prick. I\'m sensitive, be careful with me, because I\'d like to stay that way. I know it makes me a pussy and undesirable by any woman looking for strong offspring, but it is who I am with all my flaws, and it may be the only Tim that I will ever be.  Understandably heartbroken- Tim  Currently listening to \"Mad World\" by Gary Jules Currently discussing political issues on \" urlLink An  Idea Exchange \"  '),(19150,317581,'23,March,2004','2004-03-23 00:00:00','   Unsatisfied   Good afternoon...  I am feeling a little underused, underappreciated, and undersatisfied right now. I should request a meeting to make sure that I\'m doing what I should be doing and making sure I am doing an ok job at work. I should have a lobotomy so I can quit desiring, so I can be fully satisfied all the time.   Anyway, gotta run, Tim  '),(19151,317581,'19,March,2004','2004-03-19 00:00:00','   In case you were wondering   I did get paid. I am going to have buco parties this weekend and I\'ll see you on the other side.  Peace- T-dog  '),(19152,317581,'18,March,2004','2004-03-18 00:00:00','   another day, another drachma   Good morning everyone- Well, the countdown continues... will I get paid today? We didn\'t receive our mail saying our time was approved, but the website says we are. If I don\'t have my money by noon today, I\'ll be making some phone calls ;)  Anyway, I hope the world is a better place today. I went out and shot pool with Joe last night and today I\'m doing a weird work schedule so I can get our of here early. I almost made a live internet video at this  urlLink link here . In one scene that looks like it is at night, that is the one in the room I work in.  Which brings up the point of, How much light does a human being need?  Well, need to look busy, Peace- Tim  '),(19153,317581,'17,March,2004','2004-03-17 00:00:00','   New discussion blog up and running   You can get to the new blog  urlLink here   Hopefully everyone can join and have the fun discussions of today... *hint hint*  Peace- Tim  '),(19154,317581,'14,March,2004','2004-03-14 00:00:00','   life and something like it   I\'m thinking I may starting a new group blog. I would love some help if there\'s anyone that reads this and would consider doing a group blog with me, let me know and we can start some discussions. No subject will ever be taboo, except of course the issue that was just under discussion. The price you pay for bringing up either my chinese or american heritage as a negative is, I collect your f*ckin head. Just like this f*cker here.  Now if any of you sons of bitches got anything else to say, NOW\'S THE F*CKIN TIME!!!  ok, what? Oh sorry a little Oren Ishii quote for you, but please, let\'s discuss, I have a variety of friends, I\'m sure we got some discussing to do.  Let me know- Tim  '),(19155,317581,'14,March,2004','2004-03-14 00:00:00','   much better   Here I am sunday night. It surprises me the affect that a good conversation can have. I am chipper and am looking forward to the week almost. I have an upbeatness that can last for a while longer as far as I am concerned. I enjoy being in a good to complacent mood so much more than I like being in a depressed state. Well, gotta run.  I hope you all are doing better tonight too...  '),(19156,317581,'13,March,2004','2004-03-13 00:00:00','   a bit better   I was just wondering how you could know for certain... I don\'t know if we\'re supposed to be together, but I do know I want to be together I enjoy your company, you make my life fun, even when I have to work 50+ hours a week   Maybe I should move on, get a life, do whatever...  I don\'t know about that stuff, I know that I love you- me  '),(19157,317581,'13,March,2004','2004-03-13 00:00:00','   another funny    urlLink Kikkoman cartoons  '),(19158,317581,'12,March,2004','2004-03-12 00:00:00','   my wants...   I want to try and work on this I want you to want to work on this I want you to want me I want you to want me, for who I am I want you to see the potential here, I want to be the man you want I want to be someone you can be proud of I want to not be a paranoid freak I want to be totally normal and not care about anything I want to remain sensitive, please be careful with me I want you to know that I love you and I want that to be enough  '),(19159,317581,'12,March,2004','2004-03-12 00:00:00','   I\'m hurting    I hurt... I want to stop feeling... I scream and yell and try with all that is within me for you to hear, but it comes out as a little whisper... I want to be held I want someone to say that it will be ok I want someone to call me and cry with me I want someone to say yeah lets go out, lets get you a life Have you thought of trying... would you like to try it with me? Let\'s have an adventure...  I want someone to be here and hold me I want someone to say that it will be ok I don\'t want to hurt anymore I don\'t want to hurt anyone anymore I don\'t want this feeling in me I don\'t want to feel and all I can do is feel  I want to talk about it, but I can\'t I want someone to love me for all that I am and for all that I will become, I want someone to say, you don\'t fit my mold right now, but we\'ll work on you getting closer together I want someone to want me  rejection sucks...  '),(19160,317581,'12,March,2004','2004-03-12 00:00:00','   Some Funny Stuff   urlLink Big Store cartoon   urlLink Weapon of Choice cartoon  '),(19161,317581,'12,March,2004','2004-03-12 00:00:00','   what is love?   It is a feeling. It is a touch. It is trust. It is knowing someone well enough that you should walk away in disgust, but don\'t. Is it something you grow into? Or something that has to just happen... Can you turn love on and off? Is there a switch that says I could love this person or no?  What about growing pains? Are there things to learn from the growing pains? Is there a magic key for becoming the right person? Or are all relationships a work in progress, a commitment to find equal ground? When love fails what is left? A friend, a past love? Do you press on or call it quits? Is there a silver lining?  I\'m going to bed, I have work tomorrow, Peace- Tim  '),(19162,317581,'12,March,2004','2004-03-12 00:00:00','   OT and the life that\'s passing by   Here I am. Working 11 hour days, well at least one. I want no personal debt to speak of, but I have it in large quantities. So now I must work the mad hours and pay it down. Goal: remove credit cards from my life, be able to go where I want, when I want.  So how are you today? I am a bit sleepy, but woke up at 7am almost on the dot. Of course I woke up because I was horny as all hell, but what can you do? Good dreams though.  Here I am again, blogging, writing my diary for the world to see. Peter and I, and whoever else at the house that wants to, are going to the new restaurant/bar 1.5 blocks from the house tonight. So that should be fun.  Well, off to play some video games... err... TEST some video games... yeah, honestly I can\'t complain, Peace out- Tim  '),(19163,317581,'11,March,2004','2004-03-11 00:00:00','   jealousy   I am so jealous of one of my friends. She has no personal debt to speak of and has a job that she can take vacation days from. I work in an industry that promotes personal debt and punishes for sick days.   Things Tim needs to do: 1) Expand my circle of friends. I need a support network in place for life\'s little tragedys. I have a good circle right now, I just need to diversify 2) Get out of personal debt. From now till my credit cards are paid off, I will buy no more video games, TVs, DVD players, sound systems. I will spend all my extra money trying to pay off my credit cards. I can go see 1 movie in the theater a week, but that is it. 3) give my old clothes to good will or another thrift shop. Time for some spring cleaning.  Question: Is it childish to love someone that doesn\'t love you back? Do you just delay the pain and/or prolong it?  Peace-  '),(19164,317581,'09,March,2004','2004-03-09 00:00:00','   up too early? or late?   I woke up at 1:30 and haven\'t been able to get back to sleep. I hope the rest of the world has been sleeping good.  Peace-  '),(19165,317581,'08,March,2004','2004-03-08 00:00:00','   frustrated   I wish I had someone else to talk to.... I wish I had someone like you that loved me.  '),(19166,317581,'05,March,2004','2004-03-05 00:00:00','   doing better   Note to self: research movies before watching trailers. Especially at work.  How is everyone doing? Good I hope. So I was reading Craig\'s list this morning. That is a weird site.  Woke up at the crack of 6:48 with my biological clock a ringin\'. Silly body, waking up early is for people that work early. Not the rest of us who have to be at work by 9:30.  Anyway, it\'s 9:30 so I\'m werkin- Peace  '),(19167,317581,'04,March,2004','2004-03-04 00:00:00','   missing her   inside there is a hole where you used to rest, I would long for you and you would come and fill the hole, but now I long and you do not come, you can\'t come close so the hole can close, and my mind tell\'s me it\'s for the best,  but my heart still hurts and it wants you, it says she\'ll fit perfectly, she\'ll take you away, she\'ll stop the pain, but it can\'t be, and i can\'t tell my heart that it\'ll be better, it won\'t hear it, my heart wants you and hurts me that I\'m letting go, It says remember her and don\'t let her go, fight, tell her how much it hurts, but i can\'t convey that, i don\'t want to hurt you,  i just want you back, but since i can\'t have that, i want you to be happy,  in my cowardly way, I post this to the internet, and those who know can read it, and hopefully it can heal, but sometimes I wonder if it won\'t just hurt people, I am so sorry, my heart hurts and I need to write it down before i can let it go, I am sorry  tearfully-  '),(19168,317581,'04,March,2004','2004-03-04 00:00:00','   So what now?   Hey Everybody, ok, I only have 1 confirmed reader, but still I like to include everyone, just in case someone stumbles upon my site by accident. (It happened 1 time, it could happen again) Props to Racheal for finding it:)  My life has been changing and staying the same all together. My Piano lessons are going great. I am proving to be a record breaking student. Everything except my project piece is coming along. The project piece should be done by next week. So that\'s cool.  I\'m still torn by my love for my ex girlfriend and wanting to move on. Maybe this is one of those growing pains they were talking about. Maybe this is throwing yourself at something so you don\'t have to deal with the emotional side of things? Maybe this is I have extra money now, so I should buy myself something nice that will help make me feel something?  Well, thus ends another uplifting post to the blog... Peace-  '),(19169,317581,'19,April,2004','2004-04-19 00:00:00','   Mama said there\'d be days like this   So it hits me like a ton of bricks. I am single again. I need to get out of this house and meet some people. I need more friends that I can call up after 9:45 and say, hey, let\'s go down to the bar and play some pool. I need some buddies that have the same schedule as I are not at home 24/7. I need more single friends:) Well, I need to go to sleep, cause I don\'t have anyone to party with ;)  Peace- Tim  '),(19170,317581,'12,April,2004','2004-04-12 00:00:00','   Sleepy   Hey, I can\'t communicate with you, and so I\'m cheating. I\'m using a technology that I\'ve known that you check to say that I miss you and hope things are going great for you. I am keeping busy and trying not to thing of how I wish things were different than they are, but I am growing, and that\'s something that I get to keep from this relationship. My search to be a better man has become partially lit by an amazing woman. Thank you for that. Love- Me  '),(19171,317581,'12,April,2004','2004-04-12 00:00:00','   The stuff that sucks   Losing a loved one, via relocation or death. Losing a friend over a disagreement. Losing a friend due to personal feelings. Losing a love due to circumstances beyond your control.  Moving to a new city. Removing yourself from a situation that makes you hurt because you care about someone. Becoming strangers. The new make over show, the swan.   Stuff that is good   Rekindling a friendship lost in time. Moving to a new city. Rekindling a love that was lost, but is now found. Meeting new friends. Dining out. Walks around greenlake. Having my health and beauty. Singing in the shower. Whistling a happy tune. Learning to play the piano. Going out with friends.  kissing. cuddling.  peace out- Timatao  '),(19172,317581,'08,April,2004','2004-04-08 00:00:00','   The long and short of it   Here it is on thursday. There are world events that include hostages, and deaths, and all sorts of bullshit.  It seems like I should take an interest in the world around me. For some odd reason I can\'t see past my circle of friends. I can\'t help pondering what kind of man I will be.  I am back together with my ex *smiles broadly*  I need to go. Peace- Tim  '),(19173,317581,'05,April,2004','2004-04-05 00:00:00','   satisfaction   So here it is monday morning. I don\'t know if I\'m doing the right thing. But right now I don\'t care if I\'m doing the right thing. I\'m just one man doing what feels right. I love and live and learn.  God bless- A sinner  '),(19174,317581,'03,April,2004','2004-04-03 00:00:00','   Passion   I wish I were more passionate about things in my life. It seems that people are often upset by something they are passionate about and can influence others to become passionate about it as well. I have some empathy, but I think that when it comes to ideas that don\'t really affect me, I can\'t really become passionate about it. I think I am an apathetic person living in an apethetic society.  I want to feel, to hunger, to thirst, to become enraged, to boil over.   I made it to Oregon and I even remembered to blog :) Tim  '),(19175,317581,'02,April,2004','2004-04-02 00:00:00','   Seattle and Oregon   I hate Oregon. I hate that it takes me 3 hours to drive there. I hate that I don\'t have a carpool buddy to get down there with. I hate that I don\'t have someone to stay home and cuddle with. I hate that Seattle traffic SUCKS. I hate that fandango just broke.  Well at least I\'m going to go see Hellboy tonight:)  Well enough of my ranting... Time for work ;) Tim  '),(19176,317581,'27,May,2004','2004-05-27 00:00:00','   You are my sunshine, my only sunshine, you make me happy, when skies are grey, you\'ll never know dear, how much I love you, oh please don\'t take my sunshine away  '),(19177,317581,'26,May,2004','2004-05-26 00:00:00','   Hey I cleaned my room last night. If you\'d like to see it, ping me and I can arrange for a viewing :)  Peace- T  '),(19178,317581,'26,May,2004','2004-05-26 00:00:00','   I am a  urlLink Scorpio Man   I wish I were as smooth james bond, but you know:)  OMG, where were classes like these when I was younger?  urlLink Oral Sex Lessons   Damn that could have saved me a lot of misunderstandings ;)  Peace- T  '),(19179,317581,'26,May,2004','2004-05-26 00:00:00','   I got a phone call from the east coast today. I think it may be for a job I applied for over there. I have a date with a Taurus tomorrow night. From what I\'ve read, she\'ll be very rooted. This is about as polar opposite from my ex as you can get.  I\'m not sure where this is going, but I\'m reading the Te of Piglet and trying to live each day as it comes.   Terminally lost in confusion- Tim  '),(19180,317581,'25,May,2004','2004-05-25 00:00:00','   I just wanted to say thanks to all the beautiful people in my life.  I don\'t often recognize my friends as how cool and special they are. I am truly blessed to know you. Thanks- Tim  '),(19181,317581,'25,May,2004','2004-05-25 00:00:00','   I know I tend to overthink almost everything in my life.  Today I was thinking about which is better, to underthink or to overthink everything. As an overthinker, I tend to think that is good, but there\'s also a point when you need to stop thinking and start doing.  However, I do like to live spontaneously, whenever possible, and try not o read too much into anything. Also, sometimes I prefer NOT to overthink something. Thinking can become tiresome.  Anyway, here\'s a cute rant about  urlLink 10 aphrodesiacs   Peace- Tim  '),(19182,317581,'25,May,2004','2004-05-25 00:00:00','   Someone showed me this last year on Craigslist. It\'s one of the nerdiest, funniest things ever. You gotta love cyber-sex :P   urlLink Tasty  '),(19183,317581,'25,May,2004','2004-05-25 00:00:00','   I try to understand, to contemplate, to see through the darkness, to see the light, to see the fright, I want to see what is to come, to know who to be, to know where to go, to live life like there\'s no tomorrow, to see what cannot be seen, to touch that which is withheld, to see potential in everyone, to know the people of my life.  I am here, lost in my sea of uncertainty, no knowing left from right or up from down. Everywhere I look I see gray where there was once black, I see gray where there was once white. I know there are things to do for free, but this is where my muse leaves and takes the ideas with her...   So I ask my friends, there are some good ideas, the problem mostly being that everything costs money. There is ALWAYS gas involved, and gas costs money. Sometimes there is beer involved, and beer costs money.  How do you start seeing someone else when you still have feelings for another? I\'ve heard that feelings for the other will never really leave though...  Well, I\'m off- Tim  '),(19184,317581,'24,May,2004','2004-05-24 00:00:00','   So I\'ve been following the rants of raves section today. So far the main subjects have been, grocery cards, God and natives vs. transplants.  urlLink rants & raves in seattle / tacoma   It\'s just funny to me, how much of my day that I spend at craigslist, and how sad it is that everyone gets so worked up over these things.   urlLink Best of craigslist  is pretty good!  '),(19185,317581,'24,May,2004','2004-05-24 00:00:00','   I\'ve often wondered why more people don\'t blog. and have had this answered in many a different way. I think my job/personality are in a unique place where I have time to blog and it fulfills part of my need to socialize. Blogging gives me a way to get my random thoughts of Tim out of my head and onto the internet. I use my blog as a way to socialize with my friends, especially when the job takes up more than it\'s fair share of my time. I think if I went out more and wasn\'t sitting in front of a computer for 8+ hours a day, then I probably would not blog as much.  I don\'t know. Sometimes I feel like a blogger, sometimes I don\'t. Maybe this was that muse I was talking about?  Peace- T  '),(19186,317581,'24,May,2004','2004-05-24 00:00:00','   don\'t see \"Head of state\"  Seriously, I was trying to think of a worse movie, and I couldn\'t.  Anyway, I\'m exhausted, so I\'ll talk to you later today Peace- T  '),(19187,317581,'24,May,2004','2004-05-24 00:00:00','   It\'s too bad we only have but one life to live. It seems that it would be nice to experience life at all the different \"beautiful\" levels.  I mean they all suck in their own little ways: Beautiful people are always being hit upon by everyone. So they become more and more jaded by the people they know, especially if they date other very attractive people. It seems that most un-attractive people would date attractive people and love them, but an attractive person would become attracted to someone that is not attracted to them, thus creating an interest in why that person isn\'t interested in them. The attractive people will often find themselves in unhealthy relationships due to their constantly falling in love with people that aren\'t attracted to them.   For un-attractive people, or even socially challenged people there\'s can be a more worthwhile struggle: Instead of having to fend off suitors, they must deal with the constant barrage of rejection. They must constantly bring themselves back from the rejection and try and try again. They more easily learn that beauty is in the eye of the beholder and that they ust be willing to compromise for love. They struggle everyday with either trying to find love, or maybe giving up and trying to find other things to fill that void. (See video games)  The cute people I think have the best and worst of both worlds. The arrogant beautiful people not giving them the time of day, and rejecting them. The non-attractive psychos stalking them. And their want to find someone attracted to them, that they too are attracted to.  In the Te of Piglet, I read that a man had 2 wives, a plain one and a beautiful one. The story goes that a traveller came and stayed with the man and by the next day came to realize that the beautiful wife was left ignored, while the plain one was treated with honor and affection. When the traveller asked the boy of the household why. He replied \"The beautiful one knows that she is beautiful; the plain one knows that she is plain.\" The moral being to put away your pride and act upon your Virtue.  I don\'t know if I\'m explaining this very well as I\'m trying to form my opinions based off what I\'m reading right now, but I think the gist is explained. And in case you were wondering, yes, I am cute :)  Lost but in peace- Tim  '),(19188,317581,'23,May,2004','2004-05-23 00:00:00','   Enjoy\'d this article, thought I\'d share it:   urlLink Men online: why we are the way we are  '),(19189,317581,'23,May,2004','2004-05-23 00:00:00','   I went out to an puget sound waterfront house on saturday night. We drank and ate and played scattergories. Sunday was spent lounging at Seattle Center watching the kids run around in the fountain. It was a really nice and relaxing time. I wrote some in my journal but basically chilled with new and old friends all weekend. It was suprisingly relaxing and easy going. I was supposed to see \"Super size me\" yesterday, but I wanted to hang at the seattle center instead. I\'ll see it someday, in the mean time, I\'ll keep eating at McDonalds :)  Peace- T  '),(19190,317581,'23,May,2004','2004-05-23 00:00:00','   I\'m at work this week. I don\'t know what I feel anymore. I can\'t tell who I am, or who I want to be. I know that I am ok with that. I know that I will be fine and don\'t need to know anything.  I met some really cool people this weekend. So that\'s nice.  floating- Tim  '),(19191,317581,'23,May,2004','2004-05-23 00:00:00','   I think that I would not like to be a woman. I mean, it seems to me, that women, at least the ones I know, always have ex-boyfriends that won\'t fall out of love with them. I\'ve always thought that women would be the ones that would fall in love more often, but it seems that my circle of friends is the exactly opposite. I\'ve found that it\'s mostly girls that would rather not be under the commitments of a relationship.=, where most men I know would rather be in a committed relationship than a string of one night stands.   It seems that I read a book recently, Tao of Pooh or Te of piglet? Where the author was explaining how women are no longer treated like ladies because the womens rights movement sought to do away with any mistreatment of women. I don\'t know why I mention it, but it seems that women are more like the men of yesteryear, and men are the sensitive guys that women always said they wanted, but didn\'t really.  I\'m sitting here lost in a sea of uncertainty- Tim  '),(19192,317581,'21,May,2004','2004-05-21 00:00:00','   I HAVE been trying to not have these feelings for my ex. I know that I need to move on. It\'s not like I can choose to fall out of love, just as much as she can choose to fall in love.  But I AM trying. I know sometimes it seems like we\'re back at square 1. Sometimes I think I have left my feelings for her in the past.  I do want to find someone who will fall in love with me, but I\'m having trouble. I think I don\'t have time for a relationship right now. I just need to make my priorities.  Anyway, I wanted to let everyone know that I AM trying.  Peace- T  '),(19193,317581,'20,May,2004','2004-05-20 00:00:00','   urlLink    this is me&nbsp; urlLink  '),(19194,317581,'20,May,2004','2004-05-20 00:00:00','   I just found out that I\'m a snake. Well that I was born in the year of the snake. So that\'s fun:) Check it out:   urlLink The Year Of The Snake   Peace- Tim  '),(19195,317581,'18,May,2004','2004-05-18 00:00:00','   So I\'m really enjoying reading documentation today. It\'s good to stay caught up on what all is going on in the product I\'m testing. Things are slow... things are good... the beer fridge is moved back by me:)  Peace- Tim  '),(19196,317581,'18,May,2004','2004-05-18 00:00:00','   So I just had this thought. Maybe the Puritans were all scorpios that were all just jealous of being scorned, so they decided sex was bad. I think I lucked out by being more liberal even though I AM a scorpio.  I dunno, I could see how people that were really jealous, could just make a blanket statement that nudity is bad. I have no idea what I\'m talking about.  Peace- T  '),(19197,317581,'18,May,2004','2004-05-18 00:00:00','   maybe I was wrong   Maybe I don\'t know anything about love. Maybe it is all an illusion. No hay banda, there is NO band.  Maybe all anyone can ever know is the feelings they have. Maybe they\'re right, you can only find love once you stop looking.   Maybe they\'re the people that think I\'m a \"nice\" person, but not the \"date\"-able type.  About as exciting as a wet sponge? At least I try. I have THAT going for me. I know a lot of people that don\'t even do that.  My pomade smells good. Peace- T  '),(19198,317581,'18,May,2004','2004-05-18 00:00:00','   a bit of history   I found  urlLink this rant  by accident today. It\'s an interesting look at my sexual frustration from 2002 and where I was at, as a human being.  Peace- T  '),(19199,317581,'15,May,2004','2004-05-15 00:00:00','   scorpio love patterns   I found this to be interesting. It\'s basically what signs the scorpio find compatible and what not:  \"You have an all-or-nothing approach to relationships, therefore you do not match up to all the signs. Here\'s a look at how you do interact with others.   SCORPIO & ARIES: - You make a hot combination in bed. Highly aggressive sex that could be described as implied violence. Lustful but not lasting due to your jealous nature.   SCORPIO & TAURUS: - Polar opposites that actually do quite well together. The Bull is stable and loyal enough to satisfy your intense nature. You both value long-term unions and are willing to work at them together.   SCORPIO & GEMINI: - This is not a likely match, the Illusive, fickle Gemini will drive you to distraction, which in turn could bring out your vengeful side.   SCORPIO & CANCER: - This is one of your best connections. The Crab is a loyal mate who is usually willing to stand behind you. Cancers are attracted to your strong, demanding ways.   SCORPIO & LEO: - This connection can lead to crimes of passion. Although you are strongly attracted to the Lion, Leo\'s flamboyant, flirtatious nature will cause anger that could erupt into murder-suicide.   SCORPIO & VIRGO: - The Virgoan has difficulties with your sexual leanings. You will bore easily and move on to more passionate partners.   SCORPIO & LIBRA: - This is not the best match. Libra is far too good looking and easy going for your jealous nature. You would likely end up spying on this social butterfly.   SCORPIO & SCORPIO: - Sexually you are truly compatible, however out of the bedroom you are far too much alike -- creating problems and mind games. You will always think that the other is having an affair.   SCORPIO & SAGITTARIUS: - This alliance is a waste of time for both parties. You are a homebody and the Archer is a travel nut. This is not a likely attraction.   SCORPIO & CAPRICORN: - This is a wonderful partnership. You both have the same values and cherish your home environment. Mentally and physically you match up well, easily satisfying one anothers needs.   SCORPIO & AQUARIUS: - Why bother. This will never work due to the Aquarian\'s need for freedom and your need to possess. Your values, ethics and approach to life are just too different.   SCORPIO & PISCES: - This is not bad. Probably the most erotic sexual combination. The Fish is willing to be a part of your whims and desires, and that makes for a lasting union.  \"  '),(19200,317581,'15,May,2004','2004-05-15 00:00:00','   Looking to be bad   I feel like being bad. Drinkin\' and smokin\'. Who wants to come?  '),(19201,317581,'15,May,2004','2004-05-15 00:00:00','   The Honeymoon period   I am confused by this statement. What does it mean to be in a honeymoon period? I understand the whole, in love with someone that you don\'t really get along with. But I have seen people that I thought would never be together in the long run, continue to make choices to stay together. They\'re obviously no longer in the honeymoon stage. It\'s just weird, maybe I haven\'t had a honeymoon stage, or maybe it\'s that I self protect. I dunno if any of this makes any sense. I need to work. Peace- T  '),(19202,317581,'15,May,2004','2004-05-15 00:00:00','   Frustration   So it turns out that I hadn\'t actually applied for the job at bioware, Alberta, BC. It seems as though my outlook was waiting for me to sign in again so it could send the application. I signed in last night, 3 days after the deadline for the job.  Anyway, I\'m really frustrated by that, as I think it would at least have been a good interview opportunity. C\'est la vie.  Another day, another 42.50.  Peace- T  '),(19203,317581,'15,May,2004','2004-05-15 00:00:00','   On Oregon, homes, and Desire   I often find that when I am at home, I have too much time to think. It is actually that I have time to reflect on what is important in my life as all of my distractions/problems are in Seattle.   My home  My parents want me in Oregon. They say that the job market is better, that rent is cheaper. They say that even if I move home I won\'t have to see them very much. They don\'t understand that it\'s not that I don\'t want to see them. It\'s that I don\'t want to live in Oregon! It\'s something of a love/hate relationship I have with the place. I sometimes love to visit there, but for the most part I LOVE living in seattle, or maybe it\'s just that most of my friends are up there? Maybe there is some sort of defeatist attitude I have about moving back home to the parents. *sigh*   My muse  My muse is back. I haven\'t missed her one bit. It seems that she only comes to me when I am heartbroken or in some sort of emotional dilemma. At least this muse is like that. I had another muse that taught me cool things like expanding my mind and learning to knit, but this muse is the old faithful \"bad poetry\" muse! This muse likes to torment me into writing from the longings of my soul and betrays my trust at every step of the way. She\'ll have me show these works of \"poetry\" to people I barely know and they may, for a moment or two, understand me.    Desire  I\'ve come to the point in my life where I can be honest with myself and realize that I have desires. I want to be respected. I want to be loved. I want to love unabashedly. I want to quench the thirst of love in another, and have my thirst quenched in return. I desire the fruit of a woman. I long for sleeping in. I wish to be held. I long for cuddling. I desire to be known. I seek the human condition in all it\'s forms. I want the human experience. I long for understanding. I seek the treasure of our human experience. I hope to reprioritize my life, to make important all that is important and to disregard that which is not.   Priorities, lies, and the news  I have been reading the Te of Piglet and have at least mildly learned some of it\'s truths, which I have lived in understanding with for a long time, but have not practiced! The real hypocracy is that my senior year of college I wrote about the importance of non materialistic endevours, but knew full well I was pursuing a career that celebrate commercialism! Anyway, I know that I cannot buy happiness, no matter what the media tells you. In fact that whole idea of marketing is a lie! The idea that you NEED something, simply by running an ad on TV you can almost guarantee, no matter how worthless your product, someone will buy it!  I want to give up TV. I know it wouldn\'t be hard. I hardly ever watch it as it is. I noticed on the news tonight, (Staying with Parents) that the news is actually just a stressor that we\'ve added to our lives. It has no GOOD quality what-so-ever! It mostly just brought news of murder, rape and theivery, that I personally could do nothing about! How is that an important thing to me?  I missed Seattle\'s volunteer day. I had to go to oregon. I want to give back to society to feel like I am an active member of our society and I\'m having trouble finding ways in.  Well, it\'s 12:30, I need to sleep, Peace- Tim  '),(19204,317581,'14,May,2004','2004-05-14 00:00:00','   Of Mice and men   Did you ever read the book or see the movie? Sometimes I feel like the retarded guy. Lenny? It seems that whenever I find something soft and beautiful, through no fault of my own, I strangle it. Most of my relationships often end because I don\'t give enough space.  I am reading the Te of Piglet and see a bit of overanalysis in my life. Can you see it too?  Peace- T  '),(19205,317581,'14,May,2004','2004-05-14 00:00:00','   Luv   I just got finished watching \"Something\'s gotta give\" for the second time. I thought it made an interesting point about love. The point being that without opening yourself up to being hurt, you never experience love in it\'s most pure/best form.  Looking back on my relationships I wonder how much I may have held back or had held back from me. Good times can be fun, but I\'m looking for love. I want the emotional side. I want a girl that I can say, \"Have you met my girlfriend?\" about. I want to be able to trust someone and be open with them. I want to be her desire.  I don\'t want to be scared of opening up and having it thrown back in my face.  I know that I can be anyone and do anything with my life. I know that I AM comfortable with who I am. I also know that I get bored at night when I don\'t have anyone to do anything with. I\'m damn responsible, and I\'m fun to hang with.  One thing I\'ve always felt about love is, that you can only give love and hope that it is returned. You cannot buy love or force someone to love you.  Part of who I am is based on this notion of love. When I love someone, I try to make myself available to them whenever they may need me. Which often means that I fall for people who are a little sad, because I can help them be happy. I think that maybe I see myself as only useful for when people are going through tough times, so when times are flying high, I feel awkward. I\'m not sure how much of this last paragragh is true, but they are thoughts that I\'m having, so I\'ll leave them for the people of the web to read.  I\'m in a blogging mood, thanks be to Oregon, Peace- Tim  '),(19206,317581,'12,May,2004','2004-05-12 00:00:00','   Too hot to handle   Here I am in sunny Oregon, and we\'re going to be watching movies. I don\'t want this to happen to me. When does the relationship change from active to passive? Was there ever any active relationship? How does one get to the point where the TV and movies have replaced the \"go out and do things together\"-ness of america?  I recall at one point being the slave to the TV. I don\'t want to be there again. I want to live through my active relationships. I want to be known as the guy that didn\'t just sit down and watch life go by. He was the guy who was always out doing stuff.  Hello, future life, can I have you today? Peace- T  '),(19207,317581,'12,May,2004','2004-05-12 00:00:00','   Oregon blogging   So I\'m in Oregon, and since I have the time to think while I drive down and up, that usually means it\'s time to blog!  I\'m thinking of selling my car and surviving on the kindness of strangers and our metro system.  When will my mere words become action? When will piglet come forth and grow? Is there a God?   This is the Tim of Oregon, signing off. Peace.  '),(19208,317581,'12,May,2004','2004-05-12 00:00:00','   Thank God for moms   So I\'m at this weird \'what to do with my life?\' point of my life. I applied for a job at  urlLink Bioware  in alberta canada.   I think I would like the change of pace that alberta would give me, although I would miss my friends.   I woke at 4:45 this morning with a panic attack and called my mom who talked to me till about 6:30. I\'m sleepy now.  I have tomorrow and friday off and am looking forward to a much needed vacation. I hope that everyone is doing well. Catch you on the flip side- T  '),(19209,317581,'07,May,2004','2004-05-07 00:00:00','   one pint, two pint, all my problems gone   0 pints: worries about: girlfriend job car father  1 pints: worries about: father car  2 pints: worries about: will I win this game of pool?  it\'s 1:42 signing off after 2 pints, Peace- T  '),(19210,317581,'30,June,2004','2004-06-30 00:00:00','   Hey Everyone-  I\'m working on my C# skills now. I have hope again of becoming an SDET and maybe even a full on SDE someday, but for now, I\'m working on my C# skills. I\'m kind of stoked about it.  I\'m still running every other day and am losing weight, so that is good. I\'m still looking for a place to live, but my schedule at work has been throwing wrenches into the housing search. I wish I had time to program at work, but for now, I\'m going to run some more automation stuff.  Peace- Tim  '),(19211,317581,'23,June,2004','2004-06-23 00:00:00','   As I walk up to you, I start to think of rivers, I think of waterfalls and places with brooks and forests, I look down to make sure everything is pointed in the right direction, I avoid eye contact to the man on my right, I\'m nervous, you are reluctant, Then relief, my body sighs, I am empty I retreat to wash my hands.  Peace- T  '),(19212,317581,'21,June,2004','2004-06-21 00:00:00','   OH sun, how do I miss thee, let me count the ways, I miss thee in the morning, while I\'m at my desk regressing, I miss thee at the lunchtime, when the cafeteria is loud, I miss thee mostly in the evenings, when I\'m back in the room of no windows, I miss thee because my job is to stay indoows while you are out.  Missing the sun, and rain, and all things that are natural. From my dark manufactured office where I spend 14 hours a day. Peace- T  '),(19213,317581,'21,June,2004','2004-06-21 00:00:00','   So I was thinking about how our parents love us. We can never fully reciprocate that love as we will never feel the same hope for a future as our parents see in us.  Your parents put all their love and hope in you, which becomes suffocating. You have to flee the nest and they hope that you make them proud.  When love is not a 2-way street it becomes burdensome and irritating.  Well, time to work. Peace- T  '),(19214,317581,'20,June,2004','2004-06-20 00:00:00','   I\'m going to pick up Cliff from the airport tonight. I don\'t know if I should tell him that I went running this morning. I feel like I cheated on him a little ;)  Well, I woke at 6:00am this morning, as well as numerous other times of the night. My alarm was set for 6:45. So I ran around greenlake and got home by 7:30ish. Not a bad way to start the day :)  I spent 9 hours here yesterday. That was \"nice\" and I may be doing OT again next weekend. We\'ll have to see.  Peace- T  '),(19215,317581,'20,June,2004','2004-06-20 00:00:00','   It\'s 1am, and evidentally I don\'t need sleep. Although I am exhausted, I can\'t seem to fall asleep. It could be do to the sweltering heat, or the fact that I just got home from a rowdy night of board games. I need to figure out what I\'m doing in my life. I know I am a hard worker, and can excell at any given task, it\'s just I need to decide what that task will be. Also, I need to decide if I want to sell my car.  Well, now is not the time to decide these things. I\'m going to bed. Peace- T  '),(19216,317581,'18,June,2004','2004-06-18 00:00:00','   Here I sit in my safe little house. Far from Terrorist threat. Far from any danger really. I mean, I\'m a 26 yr old man that lives in a fairly safe residential area of Seattle. There aren\'t many people that would come attack me or go on a killing spree in my neighborhood. At least I don\'t think so.  Anyway, this world is pretty messed up. I mean, I live in a world of car accidents and dentistry, when half way around the world, they\'re beheading people to try and get other murderers out of prison. How is it that you think kidnapping people will get your leaders to be released? I mean, I guess terrorist is a good name. You think that using the terror of what you\'ll do will cause people to do what you want? Or are you even trying to get people to convert to your religion? Or is it merely a reason to kill?  I realize that I don\'t understand you. Sometimes I wish I could talk to you and make you understand love.  Peace, from a man in turmoil- T  '),(19217,317581,'17,June,2004','2004-06-17 00:00:00','   So it was the same temperature in my room this morning as when I went to bed last night. Which means I slept with no covers on all night and didn\'t get cold at all. I am SO looking forward to moving to a cooler room.  OK, well I need to go work now. Peace- T  '),(19218,317581,'17,June,2004','2004-06-17 00:00:00','   My friend recently told me about a friend of hers who is getting married. It seems that the guy wanted to marry her after the 3rd date, but that freaked her out. So he cooled things off and she eventually came around.  Things like that make me wonder if I should give up or go down fighting, especially if a girl is like, \"I don\'t know\". I often will disappear for a time and then come back and see what\'s up, but that NEVER works out. When you cut off communication it\'s a lot easier to get over someone. You quit thinking about them and caring as much.  Anyway, I think it\'s funny that he was ready to get married way before she was, and that she was able to be won over. Or maybe she just got freaked because she thought it should take longer. Or maybe because she wasn\'t ready to settle down yet. Or maybe because she thought he wasn\'t the ONE. Or maybe because she felt something better might come along.  I have no idea why she didn\'t want to get married, but now she does. She may be tohe only person that knows the true feelings. But it\'s kind of romantic.  Anyway, I need to work on writing my short story/autobiography of the early years. Peace- T  '),(19219,317581,'16,June,2004','2004-06-16 00:00:00','   If there is a God, then I think the religions may have it all wrong. I think that God must have created woman first. I think they(God, genderless plural) must have taken the bottle of beauty and the bottle of Grace and poured almost the entire contents into woman. Then when they saw what a great job they\'d done, they decided to make man. But they didn\'t have as much beauty or grace to use on the second model. So they made up for it with fat and muscle and clumsiness. This ended up with the male model so prevalent in today\'s society.  Anyway, I was thinking about that as I was walking home from the park tonight and thinking about out more fair, pleasant looking sex.  I think that\'s all I have to say about that. Peace- T  '),(19220,317581,'15,June,2004','2004-06-15 00:00:00','   So I didn\'t go running this morning. I\'m moving xboxes around all day and will need my strength for lifting and walking.   I just found out my step sister is pregnant again. This\'ll be her 5th, but now my other stepsister won\'t be able to catch up (since she had the whole blood clot thing last birth).  DRAMA!!!  Well, nothin much new in my life. Just busy with work and reading and writing.  Peace- T  '),(19221,317581,'15,June,2004','2004-06-15 00:00:00','   OH my GOD!!!  It hurts to walk, and it hurts to breathe. So I\'m just going to sit here and not breathe. If that doesn\'t work, I\'ll just \'walk it off\' as coach used to say.  Day 2 of running around greenlake has proven to me that I should have never stopped running as I am completely sore after today\'s 2.8 jog. Well, 3 more days and then I\'ll take the weekend off.  I started eating salads for lunch. Yesterday\'s rang up at 1.95 and it was totally yummy. We\'ll see how today\'s goes.  Peace- T  '),(19222,317581,'13,June,2004','2004-06-13 00:00:00','   So my whole buy shoes plan failed miserably today. But I have made some plans. I\'m going to start running with Cliff in the mornings and try to stick to a salad lunch plan. I bought atkins bars for my snacks today. So I should be able to survive on the salad lunch, until dinner. But I need to go to sleep now as I have to get up at 6:45 :P  I hope everyone is doing good and I\'ll try and write soon. Peace- T  '),(19223,317581,'12,June,2004','2004-06-12 00:00:00','   So I\'m hopefully going to be getting new shoes tomorrow. I would very much like to get new shoes as the ones I have right now have holes in the soles.  I\'ve had some insights recently that I\'ve been wanting to share with my general public but I\'ve been busy, so here it goes.  Insight #1: Most jerk men aren\'t jerks to the women they\'re courting. In fact, some jerks will not be jerks to the women in their lives, just to other men. However, for some it will eventually shine through to all their relationships, no matter how much they care about others.  Insight #2: Some people don\'t care about anyone outside of themselves. This goes for most people, I presume, but it is still a bit surprising for me.  Insight #3: Birth and Death are all relative (and death is much more lonely than birth). Today I had dinner with 2 pregnant women and 1 woman knocking on death\'s door. The woman that is close to death was scary to look at and see what the future held for us all. I don\'t know what I\'m trying to say, except that it was a very interesting lunch with an interesting crowd.  Insight #4: to get rid of skid marks, a man must shave his ass. (I actually found this out by reading craigslist a long while ago, but have yet to test it)   OK, I\'m going to go to sleep- Peace- Tim  '),(19224,317581,'10,June,2004','2004-06-10 00:00:00','   I just wanted to give a shout out to all my adoring fans. I\'ve been busy working this week, so my updates have been few and far between. For that, I apologize....  I have been taking the bus home, while I still have my bus pass. The powers that be have decided that contractors who have always gotten the bus pass privilege will now be denied. Another way to keep the separation of FTEs and CS. Sometimes I just want to run through the halls yelling, \"CS are people too!!!\" But I\'m sure that would get me fired.  We\'re hoping that Mike\'s kid doesn\'t come tonight, we\'re getting him the hunger team aqua force season 1 DVD.  Well, I\'m sleepy, catch ya later- T  '),(19225,317581,'08,June,2004','2004-06-08 00:00:00','   I keep getting spammed with all these \"Find your true love here\" emails in my junk mail folder. I think that there\'s probably a lot of money out there to be made off singles that are so socially awkward they\'re paying a website to help them find a match. (OK, OK, I have done the eharmony stuff in the past)  I don\'t know what I\'m saying, maybe that there\'s lots of single folks out there and if they were out having fun, they might meet each other, instead of sitting at home in front of their computers... But really I\'m just full of shit... Peace- T  '),(19226,317581,'08,June,2004','2004-06-08 00:00:00','   For some reason I am not tired this morning. I had some weird dreams last night. I dreamt that I emailed Cassie and they were like, Tim I\'m sorry, but I\'ve been dating someone for the past month or so. It was so weird cause the email was just like, hey how\'s it going?  Anyway, I went to bed at 11:45 last night and today I\'m surprising refreshed. Although I did wake up several times last night from bad dreams. The one above and another one.  Anyway, I need to get to work. Peace- T  '),(19227,317581,'08,June,2004','2004-06-08 00:00:00','   I\'ve killed the ponderingpool.blogspot.com  I think everyone was tired of writing, and it didn\'t blossom into anything beautiful, so I had to put it down.  I don\'t think I\'m getting enough sleep at night. I know I\'m exhausted right now, and should sleep, but my room is SO hot!!! :P  Anyway, I had some more to rant about, but I forget. Too hot, too tired.  Peace- T  '),(19228,317581,'07,June,2004','2004-06-07 00:00:00','   Well here it is, monday morning. I\'m sleepy.  My weekend was very busy. The Harry Potter movie was good. I\'m starting to read my next book, \"Lightning\" by Dean Koontz. Interesting quote in the book, \"Being deeply loved by someone gives you strength; wile loving someone deeply gives you courage.\" - Lao Tzu  Anyway, I should get back to work, sleepily- tim  '),(19229,317581,'05,June,2004','2004-06-05 00:00:00','   Love can give you the ability to do the things you never thought you could. Love is able to have you take chances and give flight to new ideas more readily than any other form of hypnotics or drug. Love makes you want to be someone that is an image of what the person you love wants. Love can inspire you to learn a foreign language or take up scuba diving. Love can take your fears and throw them out the window. Love is such an awesome thing.  '),(19230,317581,'03,June,2004','2004-06-03 00:00:00','   I met my friends nate and kara and Dave G for dinner tonight. I haven\'t seen Nate & Kara for about half a year. This is something I am working on reversing.  Anyway, I started putting my \"new\" computer together last night. So that was exciting. My parents are coming into town tomorrow and then I\'m going paintballing on saturday with Jarred. I\'ve been able to keep my life pretty busy.  I think part of it may have to do with the fact that I\'m trying not to slow down and feel too much. A friend recently told me that scorpios are moody and/or have mood swings. I could see that.  Anyway, I\'m sleeping- T  '),(19231,317581,'02,June,2004','2004-06-02 00:00:00','   So here I am writing again.  I don\'t know what\'s going on in me. I have been here before, or have I? I have a strange sense of wandering again, but also of soul searching and longing. I recognize myself and think of who I want to be and who I want to help. I am re-arranging myself to be financially more stable.   All this thinking is fun, but I think I should start reading again. I\'ve started walking home from my boss\' house everynight, so that is good exercise. (30+ blocks or so)  I want to get in better shape, but I think I need to find someone to inspire me to get out a little more often, and maybe take up jogging with me. We\'ll see.  Anyway, I\'m sleepy, catch you later- T  '),(19232,317581,'01,June,2004','2004-06-01 00:00:00','   I am currently looking for a playmate to help get me out of the house. I would like to find a friend that is willing to go out, even late at night. Someone that I could call whenever I felt like it and she wouldn\'t feel perturbed. Preferably someone that is cute. Someone that can keep up an intelligent conversation and will call me on my bullshit. Help me expand my horizons and teach me a thing or two.  I\'m audi 5000- T  '),(19233,317581,'01,June,2004','2004-06-01 00:00:00','   Another Craigslist best of. I really enjoyed the Truths this article revels in.   urlLink The Truth about Blind Dates  '),(19234,317581,'01,June,2004','2004-06-01 00:00:00','   I had a spectacular weekend. I think I did everything new under the sun. I went to folklife and watched belly dancing. I went to a concert and had some black russians. I went to Golden Gardens and walked around. I went to a dog park and watched the dogs play. I saw a movie, Supersize Me. I ate some pho and some Mr. Lee\'s Asian Restaurant. I went to Alki Beach and walked to boardwalk. I took multiple photos. I talked to Jay about a short film entitled \"rejected\".  My friend left for Hawaii today. She is going to see her sister and was going for the birth of her neice which happened a month early, yesterday. I think that would be frustrating.  Well, I\'m sure she\'ll have fun and now I have to work again.  Lost in my sea of uncertainty- Tim  '),(19235,317581,'31,July,2004','2004-07-31 00:00:00','   I\'m talking to a friend about types. She dated someone who prefered a \'type\' that is complacent and would believe everything he said.  I can\'t figure any man that is threatened by a woman who thinks for herself?!?! Why would you want that? I don\'t understand wanting to mind fuck someone or just not be challenged when you\'re making a dumbass move. Why would you not want to consistently improve yourself?  Anyway, I think I need to go catch up on some sleep. (Work: 10 hours today, 9 hours tomorrow)  Goodnight- T  '),(19236,317581,'29,July,2004','2004-07-29 00:00:00','   I just wanted to say hi. I am going to sleep as I have to wake up at 6:30 for running and I want to get at least 1 more hour of sleep than last night.   But I wanted to share  urlLink this link  to a pretty awesome website, so you can all enjoy.   Catch you on the flip side of the work monster...  T  '),(19237,317581,'27,July,2004','2004-07-27 00:00:00','   I\'m back at work, only 8.5 hours after I left!!! I did get 5.5 hours of sleep last night, so that is good. I actually AM feeling refreshed! How weird is that?  Anyway, I haven\'t been good with the emails recently, so if you\'ve written and I haven\'t written back, please be patient. (I swear this work thing won\'t last forever!)  Peace out- T  '),(19238,317581,'27,July,2004','2004-07-27 00:00:00','   I love that I get to meet people from all over the world on here. I just wanted to let you all know that.  Also, what\'s up with Uberskinny peope being hot. I\'m sorry, but if you\'re all bones, that is no good. There should be some meat on the package. Also, I\'m exhausted and need to go to bed. Also, I\'m still horny :(  Peace- T  '),(19239,317581,'26,July,2004','2004-07-26 00:00:00','   I think it\'s humorous the cycles people can get in. But the question isn\'t whether people get in cycles, the question is, Can people change the cycles they\'re in, once they recognize them?   I know my step sister has a tendency to get pregnant everyother year, and that after she becomes pregnant, she ditches the father and goes back to church. Once the baby comes, she wants hardly anything to do with her other children... and starts sleeping with men again to become pregnant. Repeat   My question is, Do I sabotage my own relationships by opening up too little/opening up too much/opening up too quickly? Do I sabotage my relationships by wanting too much? Trusting too much? Trusting too little? Do I over analyze everything? I know I should just enjoy life and not think anything about my relationship, go with the flow (go with aunt flo?)    online peeps   I think it would be silly for me to try and hide who I am on my blog.&nbsp; I&nbsp;try to be as open an honest here as I should be in the real world. I often wonder if I am too blunt, or too raw. Part of the problem with knowing people read the blog is I may want to cover some feelings up, or censor my blog. I can\'t do that. I don\'t have the energy to hide myself from the blogging community, so please be patient with me, as I do have some prickly edges...  Time for sleep, Peace- T  '),(19240,317581,'26,July,2004','2004-07-26 00:00:00','   Hey all- I am exhausted and in a place that I don\'t sleep very well (my bedroom == HOT). Anyway, I just heard from my friend Adam who just got back from Iraq and may be making his way back up to Seattle this weekend to visit!!! YAY for alive friends!!!  I hope this message finds you all well and having too much fun...  Peace- T  The \"Tired and going running in 7 hours\" Man  '),(19241,317581,'25,July,2004','2004-07-25 00:00:00','   Well, my body decided to wake up at 6:00 this morning :P For those of you in the know, that means I got almost 6 hours of sleep last night.  I\'ve been thinking about relationships recently. It seems like I\'m fairly good at destroying them, even when I think that it\'s where I want to be. (see pushing away) Along with relationships comes the idea of how people meet. In college I fell in love with 1 girl because she shared her poetry with me, when she would show it to no one else. I totally fucked up that relationship :P What about the blogs? Are they a good way to meet people? You deal with peoples inner most thoughts on here, maybe something they never tell anyone in RL. They could be completely socially dipolar, but you may never know. They could be the best thing since sliced bread. Trust is a funny thing.  Am I being cynical? I don\'t know. I should go back to bed... Peace- T  '),(19242,317581,'25,July,2004','2004-07-25 00:00:00','   So a bit of randomness...  Cable internet got installed yesterday. It kicks DSL\'s ass up and down the street.  I worked 13 hours today. So I\'m a big loopy. I wish there was another reason I was loopy, but it just seems to be sleep related :)  I find it interesting that it\'s about 70 degrees outside and can remain 80+ degrees in my room.  I want to save money, but I get this overwhelming sense that I should spend it too. No shopping therapy for this boy. I am going to save it for my rainy days ahead.  I want to run tomorrow, but my running buddy is away, and I need to sleep after today\'s long hours.  Meeting new peeps is fun. Being a dick at work is not.  Exhaustedly yours- Tim  '),(19243,317581,'24,July,2004','2004-07-24 00:00:00','   It seems to me that people either blog, or they don\'t blog. Those that don\'t blog often have a tendency to see blogging, not as a social forum or peoples random opinions,&nbsp;as something that is time consuming that has little to no reward in the here and now world of the day to day.  I have often become a non blogger for months at a time, but I have always kept this site up. Occasionally I update it, but only when I have no/little social opportunities with my friends. I reach out to my blog and post in the realization that no one is reading it. Until they got the profiles where you can search for people with blogs with similar interests. I\'ve now had 3 people leave messages in the past 4 weeks.  It\'s an odd feeling to be found on the internet. Besides my friends, I\'m sure no one had really read it until a few months ago.  It seems maybe that I should filter it and look through and delete any posts that I have grown through or out of, but that would be faking my past. I would hate that.  Anyway, here\'s to random people stopping by my website and new friends. I hope you like what you see! Peace- T  '),(19244,317581,'24,July,2004','2004-07-24 00:00:00','   I love free soda. I get the hankering and I just have to walk down the hall to my friendly kitchen and grab me some nutrasweet loving.  Time for work, peace- T  '),(19245,317581,'22,July,2004','2004-07-22 00:00:00','   It\'s amazing how we can get caught up with who we want to be with, before deciding who we are. I think that you can not change who you are so much than to grow to be someone else. The point being that being comfortable with who you are, and respecting yourself will make you attractive to others(love wise anyway)  Sometimes I think I am a total asshole that nobody would like (work has a way of turning you into the enemy when you\'d just as happily surf the internet like your compatriots) or a \"nice\" guy that nobody wants to sleep with.   I know that I\'m a nice guy, but I have a tendency to at least try to hit rock bottom as Chuck would say. I\'m torn between the slacker that I could be and the responsible man/boy that I am.  I know that I\'m cute, but I never come off as overly confident when talking to women, which I KNOW is not a turn on. It\'s the great mystery. Wanting sex, but playing at that you don\'t want it.  I do miss sex. Masturbation just isn\'t the same as getting it on. I know that not all of you might realize this, but I have lost that sacred virginity thing a while ago :O   Anyway, it\'s not like I\'m looking at just wanting sex. It\'s the whole relationship that would be nice. I want to be desired and fulfill that desire in someone else.  Please note: In case you can\'t tell I\'m still working 70/hr work weeks and have NO time to pursue a full on relationship. As I thought of it tonight, my life is a big canvas of grey, where I have no need to leave work, but I see the rainbow at the end of the tunnel and long for that day when I\'ll have my accrued earnings and no job:)  Here\'s to new friends and old enemies- Peace- The T  '),(19246,317581,'22,July,2004','2004-07-22 00:00:00','   Is it ok to NOT use your talents? Should you always use the talent that you have in for a job if you can? I think the subway guy might want to hire me, and I think I would have a lot less stress in that sort of job, but I wouldn\'t be using any of my 1337 h4X0r skills like I use here? I think part of the dilemma would be that I\'m not full time here, but subway isn\'t quite a career type workplace.  I don\'t know, I\'m going back to work. Peace- T  '),(19247,317581,'21,July,2004','2004-07-21 00:00:00','   I\'m going running and it seems like just 6 hours ago I was crawling into bed all sleepy like. Damn you HOT room and damn you lack of sleep. I\'m going running!  Ha body, that\'ll show you!  Peace- T  '),(19248,317581,'21,July,2004','2004-07-21 00:00:00','   So I just showered and&nbsp;my ass is  urlLink clean   Just thought you all might want to know ;)  G\'night- T  '),(19249,317581,'21,July,2004','2004-07-21 00:00:00','   When you have a subject for blogging and then you forget it right when you\'re about to start. I think I was going to mention about how I found this website:   urlLink Being friends   And how it has a fairly easy (step by step) from going from lover to friend. So I was thinking of trying to follow those hints...  There was also this website:   urlLink Love handles   Which has to do with strengthening the love handles. So I was thinking of trying out that as well. I think it\'s interesting that both articles are mostly talking to women. Is it that I\'m so feminine or that I have my sensitive male side thing going on. (metrosexual? Am I a crab person?)  Anyway, I\'m running at 7 am tomorrow, I need to sleep.  Peace- T  '),(19250,317581,'21,July,2004','2004-07-21 00:00:00','   urlLink    I should go blue again&nbsp; urlLink  '),(19251,317581,'20,July,2004','2004-07-20 00:00:00','   urlLink    At work! I\'m SUCH a nerd :P&nbsp; urlLink  '),(19252,317581,'20,July,2004','2004-07-20 00:00:00','   I was eavesdropping on my housemates friend tonight. She\'s going through a breakup (went through?) and it was sad to hear my words coming out of her mouth. The \"What do you mean right now? Does that mean later?\" questions, and the \"overall acceptance that nothing may ever change so I should move on\" talk.  It made me sad to know that\'s what I sound like when I talk to my friends. And it made me sad to know that I need to give up loving \"her\".  I\'ve been thinking about the roles \"loved\" vs \"lover\". I think it is harder to be the lover than it is to be the loved. As a friend was saying, The father, no matter how traumatized his psyche was by the child birth, will never feel the pain that the mother felt. So it is for the loved and the lover. The lover will always feel more pain from rejected love, than the loved will feel from rejecting the love. At least that\'s the way it seems to work.  I got off early tonight, but I had nothing planned, so I watched my roommates play video games, and tried to fix my old computer. (Bad Hard drive, won\'t reformat :P ) Now it\'s 11 and I NEED to sleep.  Talk to you all soon, Peace- T  '),(19253,317581,'19,July,2004','2004-07-19 00:00:00','   I would like a tattoo because I would like to have some art to look at while I\'m in the shower. Or have a piece of art that I can have always to have some fun with. I think that a tattoo would be nice for conversation starting and allow me to have a unique way of expressing myself.   At this point, I am in decent enough shape that my tattoo shouldn\'t have too much misshaping over the next few years. I may go with something like a musical note or maybe a dog. I don\'t know, I need to think about this... :S   Peace-  T  '),(19254,317581,'19,July,2004','2004-07-19 00:00:00','   Also, I want to get a tattoo or piercing, but I need ideas. Leave a comment and let me know what you think I should get:) &nbsp; I\'m thinking either  urlLink Trogdor  or a gold star which I could push to make myself invinsible for 6 seconds like Mario.  Peace- T  '),(19255,317581,'19,July,2004','2004-07-19 00:00:00','   I skipped running today:( C\'est la vie, I\'m doing it tomorrow morning. 6:30 am is only 6 hours away :( &nbsp; I\'m going to sleep, Peace world- T  '),(19256,317581,'19,July,2004','2004-07-19 00:00:00','   Here I am. My last blog hasn\'t even dried and I am posting again :P &nbsp; Anyway, I saw Farenheit 911 today and wanted to recommend it. This movie is a thinker that I think everyone should see. It REALLY brought to the forefront what war is like. &nbsp; Really going to sleep this time- T  '),(19257,317581,'18,July,2004','2004-07-18 00:00:00','   I\'m thinking of doing the monk thing for a while. I kind of became frustrated with my lack of sex today and had to have another DTR with the ex.&nbsp; &nbsp; So, here I am, if I give up on the fact of having the sex again within the next 10+ weeks, I think I can become a game testing GOD! Or at least secure me a blue badge position? I don\'t know but I think I SHOULD use my amazing computer abilities for good, but sometimes it\'s so easy to be evil :) &nbsp; So, I ended up only working 15.8 hours this weekend:) Gotta love the OT. I hate that I cancelled my 7am run, but am glad that I\'ll still have a run tomorrow night. &nbsp; I\'m lost in a sea of wondering and hate the fact that I have heartache. I want to have it removed, but I don\'t know how. &nbsp; Peace from the sleepy- Tim  '),(19258,317581,'18,July,2004','2004-07-18 00:00:00','   OH MY GOD!!! &nbsp; I went to bed like 2 hours ago. I am about to go take a shower to cool off. I HATE this HEAT!!! &nbsp; Damn you heat!!!! &nbsp; Peace- T  '),(19259,317581,'17,July,2004','2004-07-17 00:00:00','   OK, so it\'s like a BILLION degrees in my room right now. I could sleep until Sunday, but I\'m at work from 10 - 8&nbsp; tomorrow :P    &nbsp;  Anyway, I was thinking about drugs the other day. I think sometimes there is a time and a place for drugs. and then there is also a time and a place to be drug free. For me, being drug free is key right now. I\'m trying to stay at the top of my game, and I\'m a bit tired now, w/o any drugs. I don\'t know where I\'m going with this, so I\'m just going to stop.    &nbsp;  Well, I have too many things to do. Including pay my bills and sleep, so I\'ll try and remember what I wanted to talk to my bloggy neighbors about, and will get back to you.    &nbsp;  Peace-   T  '),(19260,317581,'14,July,2004','2004-07-14 00:00:00','   My life for the past few years has been the life of the contractor. I thought I might share some tidbits into this most bizarre existence that we have created.  The goal of a contractor is to become a real employee (goal #1). Everything we do is a hope to someday attain the reliable check of a full time employee. At least that is the way it is for at least 95% of us.  Every 9 - 12 months I am forced to take a 100-day non-paid leave. This can put the hurt on, as most of the time this happens from November - February. One goal many of us have is to be able to contract again after this 100-day break (goal #2). If no employer is willing to rehire you, then you will have a pain in the ass trying to get back in.  Key points to maintain while you ARE a contractor. 1) ALWAYS be available for overtime - this is key, especially when you have low finances, missing 1 OT session could mean your slot is no longer available for a few weeks. (also, OT is considered a perk that may not last much longer, if you are given the opportunity you should seize it, as full-time employees don\'t get time and a half)  2) Don\'t hound about become a full time employee - first impress your boss by ALWAYS being available for OT, then be ready to answer off hand remarks about future job positions  3) Know how to be a hard worker (This is key because you\'re trying to shine above the rest of your cohorts, some of which the full times have pet names for, i.e. Crazy legs)  4) Know how to get shit done (When your boss comes and asks you to do something, the first time you may ask questions, after that you should KNOW how to do it. If given a task by your boss, you should complete your tasks and have a short 1 - 2 sentence response with your results. Having written results in an email can also be a great way for your boss to know the info, without you having to interrupt his work day)  5) If your boss enters the room, be the first person to ask if they need any help. They will begin to look to you as a person that can get shit done. This will help in both goal #1 and goal #2.   That\'s really about all you can do while you\'re a contractor. If you impress your boss enough, he may decide you are a valuable asset which should be hired full-time, but it takes a LOT of work to get to this point.   Hang in there and you may make it. But usually people will either not be hired back (black listed) or get \"real\" jobs elsewhere.  That\'s about all I have to say about that. It\'s interesting to think about the parallels that appear in my life to this particular segment. Peace- Tim  '),(19261,317581,'12,July,2004','2004-07-12 00:00:00','   I was going to run with Cliff last night. But he bailed. So I ran around greenlake by myself this morning. It was refreshing, but I think I push myself harder when Cliff isn\'t around.  Well, it\'s time for working hard, so I should go. Peace- T  '),(19262,317581,'11,July,2004','2004-07-11 00:00:00','   I AM working today. I lied about not working, although technically it wasn\'t a lie, because at the time, I thought I wasn\'t. C\'est la vie.  I dreamt last night. I dreamt I woke to check my email and had 2. One was from my ex and another from work. Before I could answer either, a coworker came into my bedroom and started asking me about a bug they had found while I was sleeping and whether or not it was a known issue. I followed him over to where Peter\'s room is and our computer lab was all setup and running with my boss asking me if it was a known issue. I told him I didn\'t know and would have to see the bug to know if others had seen it. I woke shortly after looking at the problem and telling my coworker that we didn\'t have it recorded and he should write up the bug.  The past couple of nights I\'ve been thinking about Faith and religion. I KNOW I could never vote for Bush, but have I strayed so far from the Christian faith? Do I believe that all religions are a crutch? Do I think that faith in a religion is only for those who have given up thinking for themselves or do I still believe? Do I want there to be someone/something/(an all loving God) out there that wants to see me succeed? Is there a plan for my life and am I on the right path? Should I worship at the church of Microsoft and bow to the almighty dollar?  Of religion, I can say for certain, they do re-emphasis the spirit as the most important thing. People are the reason I was put on this earth. Without friends and family, without new acquaintenances, without people, life would be ever so dull. I\'m just not sure \"saving\" the soul is what is needed.   Well, I need to get ready for work. Peace- T  '),(19263,317581,'09,July,2004','2004-07-09 00:00:00','   Holy Sh*t!!! I gots the weekend off!!! Woohoo!!! Now... what to do? Peace- T  '),(19264,317581,'04,July,2004','2004-07-04 00:00:00','   So I had random post at 5:30 this morning, which meant I woke up at 5:30 and wrote a post, then tried to sleep until 7am when I got up to go running with Cliff. That didn\'t really happen.  Fortunately, I was able to sap enough restfulness out of that hour and a half to be affective today :) Running is good. It makes me feel awake and healthy. Anyway, it seems like I may get off this weekend so I\'ll have time to run errands!!! YAY!!!  I\'ve also been remembering this movie I saw like 2 years ago with Jarred. It\'s called \"Intacto\" and is a spanish film about luck. It was really great and if you love foreign films I highly recommend this one.  Anyway, I\'ve got some Localization peeps to manage. Peace- T  '),(19265,317581,'04,July,2004','2004-07-04 00:00:00','   I often wonder what makes me want to be so responsible. It seems that if I wanted I COULD get away with not working for long periods of time and maybe just live off the state for a while (see step-sister)  But instead, I want to succeed in a career. I want to be praised by my coworkers and stand out in the minds of my bosses. I want to be desired for full time employment at the risk of sabotaging personal friendships at work.  I don\'t understand who I am, nor why I have an overwhelming sense of responsibility for a job that I could just as easily lose next week.  '),(19266,317581,'04,July,2004','2004-07-04 00:00:00','   It\'s true. I miss having even a pseudo girlfriend. Now it\'s the 4th of July and I have to be at work in 11.5 hours and the F*CKING neighbors won\'t stop letting off fireworks!!!  I have my run in the morning, but my friend Cliff (see running buddy) is moving to Lake Samamish(sp?) plateau. That means my current running buddy (see best friend) will be 45 minutes away on a good commute day :(  I am trying to get C# under my belt like no tomorrow. I have gotten my program to a point where I can download a webpage and extract all the img file locations and links. Next I need to work on erraticating duplicate links and links to other websites. Then I need to download each new webpage and rinse lather and repeat. It\'s really fun to problem solve and get this all figured out, but I\'m scared that it won\'t be enough to land me a job :(  Anyway,I\'m exhausted and horny and lonely, so I\'m going to sleep. To all those out there celebrating the 4th. Happy independence day and I\'ll see you on the 5th.  Peace- T  '),(19267,317581,'02,July,2004','2004-07-02 00:00:00','   It\'s true. I found out today that I am officially exhausted. I imagine I\'ll have 1 more day off this summer. Sunday, and then it\'s working everyday until we ship.  Props to heather for finding me and leaving the comments. I think a love of Guster is good, but finding a needle(Tim) in a haystack(The internet) is even more amazing :)  I\'ll check you guys later. A total exhausted Tim.  '),(19268,317581,'01,July,2004','2004-07-01 00:00:00','   Hey Everyone-  I just found out that July 2nd is the world wide (national?) \'I forgot\' day. That means that today you can make up for anything you\'ve forgotten about in the last year. W00T.  Anyway, yesterday was \'Canada\' day. So for all you Canadians out there. Happy Canada day! (It\'s I forgot day, and since I forgot Canada day, does that make it ok to mention it?)  and in case I don\'t get out of this working coma, Happy Independence Day!!! Let the fireworks begin!!! W00T  Peace out- T  '),(19269,317581,'01,July,2004','2004-07-01 00:00:00','   He follows you around, He thinks you\'re so cool, He wants to be just like you, You friends are getting annoyed by your little tagalong, it\'s becoming quite the nuisance  He needs to get his own life, quit sucking his life from you,  He used to be cute and endearing, but now he\'s just a pain.  '),(19270,317581,'01,July,2004','2004-07-01 00:00:00','   So I think the loneliness is beginning to get to me. I haven\'t had a good talk with anyone for a long time. I would count the days, but they all start to jumble in my mind. I think I\'ve been burning the candle at both ends for too long and it\'s finally catching up with me. I need some relax and chill time with people, if only just to talk about my feelings. OR maybe I just need a good nights sleep and then I won\'t have these lonely feelings anymore.  Who knows? Peace out- Tim  '),(19271,317581,'05,August,2004','2004-08-05 00:00:00','   It\'s crunch mode in t-minus , oh wait, we\'re in the middle of it. Another day and we\'ve got Strep throat in the office. I am praying that I don\'t get it, and neither do my testers. I guess time will tell.  I got a package from a friend today. I need to start reading again, and filling my head with fun ideas.  Anyway, I need to sleep. Running in t-minus 7.3 hours. Peace- T  '),(19272,317581,'04,August,2004','2004-08-04 00:00:00','   So, I\'m having trouble sleeping. It\'s one of those toss and turn nights.... So I popped in a little Fight club:) I love that movie. It got me thinking about some things... If I wake up too exhausted too many days in a row, Do I wake up as another person? Am I actually 2 people living my life? I feel as though my life could be pulling in 2 different directions. My reponsible side wants to get full-time employment with MS, while my playful side wants to say F*ck it all, I\'m going to work at subway!  Anyway,I used to think I had a lot of posts on here \"566\"... but now I realize that all those months that I took off from blogging were months where other people were at least blogging once or twice a week :) Good thing this isn\'t a competition, otherwise I would have lost:P  Note to self: Must update blogger profile to include mulholland drive... also, must sleep, we wouldn\'t want to wake up tyler:)  '),(19273,317581,'03,August,2004','2004-08-03 00:00:00','   So I am basically managing 14 people now. I am Mr. \"go-to-guy\" in the department. I know that some of my coworkers despise me for this, but I won\'t apologize for it. I do make myself known to be available for the developers and I do make myself available to the higher ups. If you want to despise me for doing a good job, then so be it, but don\'t downright undermine me.  I don\'t know, it seems like people would rather argue with me, since I sincerely have no power over them, than to try to help me with something that a developer might need.  Anyway, I digress... I wanted to write about the things one deals with when \"managing\" 14 people, or something like that. I forget.  Anyway, I should start programming again and maybe I\'ll be able to get a really real job!! Now if I only knew where that quote came from... dammit!  Goodnight- T  '),(19274,317581,'01,August,2004','2004-08-01 00:00:00','   I\'m going to bed, but I wanted to write you all and wish you a happy sleep and peace in this world.  Remember...  You are beautiful :)  Peace- T  ');
INSERT INTO `msgs` VALUES (19275,317581,'01,August,2004','2004-08-01 00:00:00','   So my friend just sent me a link to  urlLink Team America   Which looks hella good and I wanted to share it with you all.  So THERE! Peace out- T  '),(19276,317581,'01,August,2004','2004-08-01 00:00:00','   I\'m listening to Guster right now. It\'s good therapy music. Some of the songs are good for being alone. Some are good for heartbreak. Some are good for starting over. Some are good for finding someone new.  I\'m confused as to who I am right now, or what music I need. I know that I\'m a good person that will continually try to do the right thing. If that\'s not a desirable attribute, then I guess I won\'t be desirable. Everything in me wants to help others succeed. But even as I write that, I know it\'s not true. I know that I don\'t want people that talk down to me to succeed faster than me. I know that there are people that drive me nuts and whom I probably dislike enough to not want them working where I work.  I\'m always analyzing peoples skills and where they might fit best (at work anyway) I categorize my friends. I hide my feelings, in order to self protect, as almost everyone I know will leave sooner or later.  I think I talk better in person or on the phone than I do in email, but I don\'t know if that\'s true. I know I can charm, but I don\'t know how I do it. Someone once told me to stop manipulating them, but I didn\'t know how or if I was actually doing it. I don\'t have a conscience malisciousness, but maybe there is something darker inside?  I want to trust new friends, but I also don\'t want to get burned. It\'s a fine line. Trust is a good thing.  I\'m running in 8.5 hours, I\'m going to bed. Peace- T  '),(19277,3677344,'25,June,2004','2004-06-25 00:00:00','    Thank you to my friends and family for being an unbelievable, dependable, source of support.  Thank you for allowing me to take as liitle or as much of all that you offer as I need or can cope with, often without asking and mostly without saying \"Thank you\" in the end.  Thank you for giving me the space I needed to spread my wings and find my strength.  Thank you for being the invisible, yet dependable safety net in case I fell.  Thank you for celebrating my victories with me no matter how monumental or seemingly insignificant.  I truly believe that every victory won, no matter how small is never inconsequential.  Thank you for drying all my tears, no matter how silly or serious they were. Thank you for not asking questions, just accepting my many moods.  Thank you for offering all that you could, for giving mostly of yourselves.  Thank you for your support, encouragement, your cheerleading, your acceptance and your guidence.  Thank you for teaching me that I have worth, I matter, I am beautiful inside and out, and that my past has made me who I am today, but holding on to it was no longer necessary.  Thank you for loving me in the only way you knew how.  Thank you for being there and being someone I could always depend on.  Thank you for holding my hand when I needed added strength.  Even if you could not be there physically, I knew that you were there in spirt, in my heart and in my head.  Thank you for being the little voice in my head, saying \"Don\'t give up, this too shall pass.\"  Thank you for all the big things and the little things that you all have done and given me in the past 6 months.  Without you, I would not know my strength, my emotional strength.  Thank you for all that you do and all that you are to me.  I love you all. ~B '),(19278,3677344,'18,June,2004','2004-06-18 00:00:00','    In the power of forgivness People only have as much power in your life as you chose to give them Tomorrow is another day Kisses are not contracts, and presents are not promises Company is not security I am enough as I am Nobody completes me, I am whole on my own I can accept \'nothing more\' but will settle for nothing less. In the power of the word NO A hug from your children can have incredible healing powers. No always means No, and it\'s never too late to say No You happiness is your responsibility Love is a choice you make, not a place you fall Marriage takes work One partner can not love enough for both Marriage will only work with two participants '),(19279,3423418,'23,May,2004','2004-05-23 00:00:00','   I want Bush to be reelected this year. I do. Don\'t get me wrong. I don\'t like him, nor do i like any of his policies. I support him for \"all the wrong reasons\", according to my friends and colleagues. But I do think my support of his reelection is justified. Here goes my argument. One reason I want Bush to be reelected because I think America needs to learn a harder lesson.  For the past four years Bush etc. have done many bad things: cutting funding for education, screwing up medicare,  polluting the environment, handing money to the rich,  raiding wars on other countries, and getting many many foreigners and Americans killed. But obviously not all people have got enough from them. Actually more than half of the country agree with him and try to get him reelected, including many who were actually harmed by his policies. So my conclusion is that people haven\'t learned enough about the bad time yet. Therefore Bush should be given another four years so that he can do more damage to the country and the world. Maybe in 2008 people will finally realize how awful the country is under the management of the republicans, and then they will make better decisions for the next several terms. Another reason I want Bush to be elected is that there is really no real difference between republicans and democrats. On surface, it seems they are debating on every front. However when I look deeper into the issues, I am scared because of the similarities I see. For example, both are claiming that the US should stay longer in Iraq. Even the democrats are claiming that they will be more responsible and get the UN back to the mess, it is not clear to me that they are actually admitting that the fundamental cause of the war evil. Another example will be the attitude towards gay marriage. Kerry has already announced that he will not support gay marriage. There are also the issues of protectionism, isn\'t it ironic that after Bush imposed the tariff on steels now it is the democrats who are yelling about protecting American manufacturing industries? Some people may say all these are merely campaign tactics and the democrats are indeed much more liberal. But that will eventually put them into a situation where they have to break many of their promises and get people angry with them. The fundamental problem lies in the American public sentiment and the political system.  After all it is the people who are electing government officials. It is the people who decide to give the government power of doing all kind of things and implementing all kind of policies. So can I say the American people deserve Bush for four more years?  '),(19280,3423418,'22,May,2004','2004-05-22 00:00:00','   This is my first blog.  I don\'t know how it will go but I\'d like to give it a try.  There are so many random thoughts I have everyday, most of which get lost because of the magic of time.  So I would like to keep them somewhere from today.  '),(19281,3471740,'27,May,2004','2004-05-27 00:00:00','   Wow, my first blog. I feel so invigorated by all the possibilities! I\'m embarking on a new adventure! I can create and mould this blog into whatever I want it to be! And soon, I shall become...a blogger...  '),(19282,3471740,'14,June,2004','2004-06-14 00:00:00','   I\'m not sure why I have such a hard time getting into the swing of things on a Monday morning. I find it difficult to bring myself to do any actual work before lunchtime and feel like I\'m operating in a dazed state. Okay, well, that\'s not entirely true because I do work and accomplish things in the a.m., but I just don\'t like to launch into any involved projects before lunch. So, my morning consists of responding to emails, planning and activities for the week ahead, returning phone calls and reading any materials circulated round the office. Ususally we have a staff meeting which will take up 1 or 2 hours which works out perfectly because by the time the meeting is out it\'s usually around noon or 1 pm. On the upside, at least my coworkers know me well enough to avoid engaging me in any lengthy conversations on a Monday morning. If they do, the sound of my \'morning voice\' will likely make them think twice before doing it often. I have nothing against mindless chatter; it\'s just that I tend not to be chatty in the mornings. Of course, anyone reading this probably knows that about me already.    Today I just got back some pictures from my trip. I have yet to join the those who partake in the instant gratification of digital photography so I had to wait for my film to be developed before I could see that half of the photos I took are in fact pure crap. I love taking pictures, documenting friends and events so I can look back and fondly remember that yes, I did enjoy the time I spent (wherever, whenever) and could probably tell some crazy fun stories to go along with the pictures. So, it\'s quite a conundrum really as to why I haven\'t jumped on the digital photography bandwagon. This most recent vacation has propelled me to change that. The morning that I was going to leave for Germany, I discovered that my trusty camera, the very same one responsible for all college photo albums wrought with debauchery and embarrassing moments for all was broken. I then had no camera available to me! How could it have happened that I had been living my life, in this modern world, in a hip city, with no photographic equipment what-so-ever! You see, I come form a fairly (camera) trigger-happy upbringing. Every member of my family has at least 2, more likely 3 or 4 different cameras, yet in some state of obvious irrationality, I convinced myself that the extras were clutter and I didn\'t really need them.   In any case, I shall now embark on the search for the perfect camera. Never again shall I allow myself to be the victim of a disposable camera and it\'s featureless offerings!  '),(19283,3471740,'09,June,2004','2004-06-09 00:00:00','   Believe it or not, every once in a while I like to leave the Bay Area. As you all know, I love San Francisco, Berkeley the Bay Area (which, in case you hadn\'t heard, does not include San Jose). I also love traveling and every time I return I appreciate what a beautiful and unique place San Francisco is.   Germany was a lot of fun. Kim, Marty, Oliver and I went to quite a few places around southern Germany in the sweet Mercedes Benz we rented. I\'ll have to get some pictures and post them. Unfortunately, it rained on and off for most of the time, but I think I still got some good photos.  Some of the highlights were the beer of course!; seeing the many castles scattered throughout the hillsides; Rhine River boat cruise, visit to the Porche factory (where I tried to get a job as a test driver- unfortunately, that didn\'t work out); Baden-Baden (spa town with mineral baths and swanky shops and a casino); driving through the Black Forrest; visit to Frieburg, which is actually in the Black Forrest, has a really cool brewery, the \'old town\' has intricate cobblestone-mosaic sidewalks and streets with open gutters- which I should mention, you must be careful not to step in while walking back to your hotel drunk from the brewery; we stopped in Fussen to see the fairy-tale castle; and then onto Munich! Munich is the type of place that I need to go back to. It\'s a great city with lots to do and when the weather is nice you can fully appreciate the parks and the beer gardens.   I also picked up a little German. Here are a few phrases I found helpful while there (spelled  sort of  phonetically for ease of pronunciation):  \"Bitte ein bier?\" (say with a cute smile to the bartender as an apology for poor pronunciation)   Please, a beer?     \"Bitte ein bild?\" (say with a cute smile to a boy as an apology for poor pronunciation)   Please, take a picture?   \"shIza!\"   Shit!   \"Fair-root!\"  Cool!   \"Das iz fair-root shIza!\" That is cool shit!  \"Das iz koo shIza!\"  That is cow shit!   \"Haben zie baden shaum?\"  Do you have any bubble bath?   As you can clearly see- all very useful phrases.  So, for now, I\'m just battling some minor jet lag and wondering when and where my next big out-of-country trip will be. I\'ll try to get some Germany pictures up on this blog soon too!  '),(19284,3471740,'02,July,2004','2004-07-02 00:00:00','   My co-worker Rob (who is my bosses boss and practically the head of my entire company) asked me to house-sit while he and his family went on vacation for 3 weeks to France and Italy. So I agreed to stay there (not because I\'m a kiss-up, but because they have a cute house with a great remodeled kitchen in relaxing Menlo Park).   I was inexplicably drawn to the Viking stove and oven and All-Clad cookware. So, one evening I decided to get crafty and cook myself dinner. I had things going on the stove and something in the oven as well when I got a phone call. As I\'m catching up with a friend, I suddenly realized that I need to check the oven, so I looked- and good thing I checked because it was on the verge of burning. I quickly grabbed the towel hanging on the oven handle, take out the pan and see that the towel now has a hole burnt in it approximately 5 inches in diameter.   I immediately ran it under water to stop the burning. Made sure that I was not on fire and that the house was not on fire. I told myself that accidents happen and a towel shouldn\'t be too big a deal, right? Immediately following that, reality and my usual level-headed logic set in. The truth is, this was not a \'normal\' kitchen towel. It was a fine cotton towel with embroidery on it with flowers and writing which said \'a gift from France\' or \'a touch of France\' or something along those lines. This towel obviously came from France and had some sentimental value that I would never be able to quantify. Of course, now I\'m back to freaking out because I think they must have got it on their honeymoon or it was passed down as an heirloom in their family or someone brought it for them as a special handmade gift- the possibilities were endless!  I knew I wouldn\'t just be able to walk into Williams-Sonoma and get a replacement so I consulted the wisdom of a select group of trusted friends. Some of the great advice I got was to just throw it away, never bring it up and hope they think they lost it themselves. My other option was to just tell them the truth and offer to replace it. This seemed like an empty offer to me because this towel was obviously something imported and they had had it for a while. But in the end, that\'s what I went with and hoped that they wouldn\'t take me up on it because then I\'d probably either have to go to France and find someone to recreate the embroidered design.   How does this all end you ask? Well, I too have been full of anticipation and today, Rob is back from his trip. He hadn\'t said anything about the note I left or the towel when he came in, but I knew it was only a matter of time.   I had to go down to his office to get some things signed so I told him in person that IÂ’m sorry I ruined it. And he tells me he was thinking he would give me a hard time about it and tell me that his wife\'s grandmother gave to her on her deathbed. I think all the blood and color must have left my face at that point because he followed up with a \"no, really, it\'s zero big deal and donÂ’t even think about replacing it.\" Whew! He also mentioned that he\'s glad I didn\'t burn the house down. And of course, now they want to hear the story- which really isn\'t all that interesting. I\'m just glad it\'s finally all over with.   I wonder if I\'ll ever get asked to house-sit again...  '),(19285,518116,'28,February,2002','2002-02-28 00:00:00','   I almost forgot...he said happy birthday...  '),(19286,518116,'26,February,2002','2002-02-26 00:00:00','   Mr X called...so unexpected!  I thought I was in for a looong wait...but he had left a message on my voicemail that I got this morning.  So, of course, I held off until mid-morning and then returned the phone call--can\'t seem to excited...he said that I took him off guard with the last conversation and he hadn\'t really known what to say.  He said it kinda freaked him out...okay, I couldn\'t bring myself to ask what it was that freaked him out--not really sure, but I think he read way too much into my thoughts and questions...  He feels that I need to realize that I am remembering the past and that so much has changed since then...but no matter what, I need to let go of the past because that will never be.  He also said that he wished he could help me figure out why I wasn\'t happy because that is what he wished for me...that was sweet.  He said that he will support me no matter what I do but that I really need to take some time to think everything through...he wants me to be sure that I do it for me and not out of spite or emotions.  Ah, what age has done for Mr Xs maturity!  Of course, I had to call my voice of reality...she laughed and said it really did sound like we had role-reversed...haha, what a crack-up, my dear!    I think I need some time to digest everything...afterall, the big move is coming up and that is sure to bring so more emotions to the surface...  '),(19287,518116,'26,February,2002','2002-02-26 00:00:00','   The comments didn\'t work...oh well...  Awesome weekend...I went to see my roommate in Chi-town...she shipped off her hubby and bubs so we had the night to ourselves...IKEA here we come!!! We tried to remember that our husbands would kill us if we brought home one of everything:)  We had a rip-roaring time at Houlihan\'s and then went to find the nearest tattoo parlor...the funny thing was, there was some SHADY business next door...hmm....not drawing any conclusions but what do you think with lots of doors and an ATM??? Of course, I had to use the bathroom as soon as we get there...the place smelled HIDEOUS!!! But, using their restroom was worse then squatting against a tree in the African Rainforest...I felt seriously violated and thank god I had spare kleenex--hello, toilet paper!  It was just like old times...we even watched our party videos...boy, were we off the hook back then!!! A little embarassing but mostly just HALARIOUS!  It was great re-visiting Tow-truck Mike, Joe, John...George--I mean Kevin...sigma-kappa...and Kelly...  Hubby was a downer when I got home...he was CRABBY...so we had a huge blowout...ah, wonderful way to cap off the weekend!  I talked to Mr X again Sunday night...it was funny calling because I had wanted to suggest we see each other at some point but didn\'t have the balls.  So I was just like \"um, yeah, hi, uh-hu\"--DORK!!!  I think it made him crack up though:)  Just talked about what\'s new and then got into a deep discussion on happiness and to what extent someone is free to chase it down...but his phone kept cutting out and I was sick of calling back...so I didn\'t.  But I KNOW he has my phone number...he said he had lost it before--uh-huh, sure--so I gave it to him AGAIN...let\'s see if he ever actually calls...still holding out for my bd.  '),(19288,518116,'21,February,2002','2002-02-21 00:00:00','   Okay...I\'m trying to add a comments section, so this is just a test...  '),(19289,518116,'20,February,2002','2002-02-20 00:00:00','   Sorry I didn\'t have time to actually write something of substance yesterday...however, I thought I should spend a little time working since that is what they pay me for... although, today, I get to be here for longer...so I can goof for longer!!!  Okay...what has been making my brain do crunches...I talked to Mr X about three-four weeks ago...for a LONG time.  It was very nice actually...however, since then, I have an undying urge to call him again.  I did call back once to let him know the (fake) reason for my original call was now void due to an unforeseen (also, fake!) emergency...ah, revisiting the jr high days can be so freeing!... alas, X was not home and I was only able to communicate with the answering machine.  I was hoping I would receive a return call just to say--hey, I really enjoyed talking with you...I\'ve missed you and I wish you well...  Now, I am holding onto the hope that I will receive a message for the dreaded \"B\" that is rapidly approaching...  Time to psychoanalyze....I have come to the conclusion that I am mortified of commitment...rather humorous looking the commitment level in my life I know...  But I have a tendency to get myself WAY over my head (with commitments) and then start pushing them all away...so, is this another passive aggressive way to kill-off my current commitments?  I mean, come on!!! I used to think the separation from Mr X was the one SANE thing I did in my life...and now, I keep wondering how to wiggle back into his...  '),(19290,518116,'20,February,2002','2002-02-20 00:00:00','   Okay...I am a little nervous about the information that will be divulged...therefore, we need to have some ground rules...  #1 NO QUESTIONS...take this blog for what it\'s worth...a peek into my psyche...if you have a burning need to ask a question, you may forward to my people and they will get back with your people if I choose to respond...  #2 DO NOT DIVULGE THIS INFORMATION...this will result in termination of the blog and you will have to spend the rest of your life wondering the end of this story...  #3 Okay, I don\'t have a number three but I remember learning all effective speakers have three main points...  Finally, for a lot of you this will not make ANY sense...for some, it will make just a little sense...for others, it will make a WHOLE lot of sense...and the rest will be scared.  I am going to start where I think the story needs to start...however, I will jump backward and forward, and repeat as well...but I promise, we can get through this and if nothing else, you will leave entertained and thankful that the edge I write from is not yours...  '),(19291,518116,'27,March,2002','2002-03-27 00:00:00','   Okay...a guy I used to work with sent me this...I thought it was pretty cool--thanks, Dirk!   BITCHDOM   When I stand up for myself and my beliefs, they call me a bitch.  When I stand up for those I love, they call me a bitch.  When I speak my mind, think my own thoughts or do things my own way,    they call me a bitch.   Being a bitch means I won\'t compromise what\'s in my heart.  It means I live my life MY way.  It means I won\'t allow anyone to step on me.  When I refuse to tolerate injustice and speak against it, I    .  am defined as a bitch.   The same thing happens when I take time for myself instead of being  everyone\'s maid, or when I act a little selfish. It means I have the courage and  strength to allow myself to be who I truly am and won\'t become anyone else\'s idea   of what they think I \"should\" be.   I am outspoken, opinionated and determined.  I want what I want and there is nothing wrong with that!  So try to stomp on me, try to douse my inner flame, try to squash every  ounce of  beauty I hold within me.  You won\'t succeed.  And if that makes me a   bitch, so be it. I embrace the title and am proud to bear it.        B - Babe       I - In      T - Total       C - Control of       H - Herself        B = Beautiful       I = Intelligent       T = Talented       C = Charming       H = Hell of a Woman        B = Beautiful       I = Individual       T = That       C = Can       H = Handle anything  '),(19292,518116,'25,March,2002','2002-03-25 00:00:00','   Whew! what a weekend...and not in a good way...  My \"voice of reason\" called...I told her my decision about not calling Mr. X again...I told her I came to realize it would be the same old thing and no point...and she says \"so basically you just wanted to make sure he still wants you and then ditch him...\" ah...no!  I needed to finally feel closure on a very tumoltuous time in my life and talking about it with him was the last resort...I have been through YEARS of therapy and nothing ever closed that door and I finally feel that it is slowly starting to swing shut...ugh! whatever!  Scott and I actually had a pretty good weekend...we worked together all day Saturday to get the rest of our stuff organized and did a HUMONGO cleaning...I cleaned our little bathroom for an hour and a half!!! I found some nasty scum that I will spare you the details about...but it made me want to heave!  Ah...then the infamous Sunday...thinking I was a good dooby going to dinner at the parents to spend some quality time with my sis that never keeps in touch...and my matriarch decides this would be a good day for an all-out attack...apparently, my sis mirrors the other picture of perfection in the fam and spilled some beans found out to my parents about the other sis that mirrors my picture of total and utter imperfection...so, of course, my mom thinks that I am to blame...I am the original black sheep, you know...and heaven forbid, this sis could actually be making decisions on her own...NO...it is MY fault because I tried to look out for them and teach them things about the world that my parents would like to naively believe would never be experienced--except that now that I have talked with them about it, they are destined to experiment--because they would have never come across it otherwise!  Ahhh...then, mommy-dearest decides to start in about our past...to which I must have started to flip out because Scott was trying to corral me into a different room...not that I would have attacked her, but it did cross my mind...  Okay...first of all, my parents can never quite make up their minds what they think should be done...it is basically up to the whim of my mother\'s emotions, which change constantly...  Secondly, let\'s move on!  I didn\'t go through all of this therapy to be pushed right back where I was...I am an ADULT...and, quite frankly, I can talk to my sister\'s about anything I DAMN PLEASE!!!  I don\'t want them driving drunk, or being taken advantage of, and never having taken the opportunity to discuss these things with them...that would be guilt...not my mum trying to blame what she sees as evil on me...always looking for someone else, never looking within...  Thirdly, my sis does need to realize things about life...but I am living proof that having someone breathing fire down your back is swimming against the current...it can\'t happen until she is allowed to explore herself.  I feel bad for her because she was caught in the crossfire...and part of it was about her and alot of it wasn\'t...  Mum did come back and apologize--and after they revived me, her and my dad said how much they wanted to talk with me about what I \"went through\" so that they could understand...and to that I said, like hell!!!  Please...really, guys, I love you, I forgave you, but you can\'t slap me and call me stupid anymore!!!  For YEARS, I wanted to explain this to them and make them see...if for no other reason than to prevent it from re-occurring with my sisters, which it has...but the don\'t want to know...they want to believe they have had no part.  And that\'s their choice, but I don\'t need to put myself back in the place of being rejected and ridiculed...  So, for now, i will keep my sis in my thoughts and prayers...I know that I was delivered and I hope that she escapes with fewer scrapes and bruises than I did...  '),(19293,518116,'21,March,2002','2002-03-21 00:00:00','   Wow...I can\'t believe how long it has been...I guess that is what happens when your life isn\'t all that exciting so you don\'t feel a burning need to blog...however, after reading a few other blogs (the sis\') I decided I should update mine too:)  Not much new has happened...I called Mr. X last week(?) nothing exciting...just to say hey and to see if he was done being freaked out.  I must have been in one of my \"bitch can\'t stop me\" moods because I asked him what it was that freaked him out.  He said that I had been talking about all these thoughts and that he just didn\'t want me to get over my head without being realistic...i laughed--hehe--how\'s that for a blow to the ego--and told him that I think he misunderstood what I had said and that I never meant to propose anything because that was FAR from where I was at...had I thought about it?  yeah...sadly, so...however, i may be f#@*ed up but I am not that far gone...well, after I said that he wanted to get off the phone and told me to call him any time...I said \"please feel free to do the same\" and left it at that...so, basically, I am going to leave it there...I came, I heard, and I walked away...I know what is there and I have to believe there was a reason I had walked away before and I need to release myself from that...whew...what a revelation!  Too bad I couldn\'t have come to that conclusion before I started down such an idiotic path!  I am just very thankful that in all of that I chose not to go the step further to see him...that would have been jumping right into the fire...  Hubby is still as clueless as ever...but I think that he is going to start trying with more vigor...my resolve is back to be the best little wifey and mommy I can be...I just need some excitement in my life...so maybe I will get that tatoo afterall:)  I did get my belly button pierced awhile back...I thought my bod was made for piercings because I have never had any problems with getting pierced but now my belly button is a little sore...so I\'ve started back with the bactine...ouch!  it stings a little but I\'ve seen the alternative...a friend with a piercing gone straight to hell and it looked as though her belly button and the rest of her stomach were going to fall off or be taken over by gangrene...  One of bf was supposed to come in from Kansas in February...never heard from him...I tried to call but his phone number is no longer valid....ehhhh...I keep wanting to call his bro and ask what\'s going on, but I chicken out...I don\'t want to be told that he forgot about me or that he didn\'t want to see me...so I guess I\'ll wait until I run smack dab into one of the fam...  '),(19294,518116,'11,March,2002','2002-03-11 00:00:00','   Okay..I forgot to copy and of course I timed out...so here\'s the recap!  It\'s been a while since I have been able to blog...we had the big move that took four days instead of the anticipated 2...and of course, there was the pre-move preparation--packing--YUCK!  But of course, my husband, who is oh so understanding, helped me with the packing...oh, wait, not MY husband...he would just keep looking around and say, what is it that needs to be packed...uh, HELLO!!! see this *@#! all over...all THAT needs to be packed...he just doesn\'t get it and continues to choose to be ignorant...and it is getting VERY old!  Which leads to his next blunder...remember the bday that was coming up that I was dreading so...thinking I had made this clear to him, made me believe that he wouldn\'t have the AUDACITY to forget...well, that part was true...however, he didn\'t make time to get me anything...yes, I know, you are thinking, what a jerk...but, please, don\'t be so kind!  Ass...big, fat, hairy, horse\'s ass, would be the correct way of stating it!  He said that he didn\'t have time...because, oh yeah i forgot, he worked SOOO hard on packing...whatever!!!  So, basically, not a whole lot of people remembered....our parents remembered...my older sister called and sang with the bro in law, but then said she forgot to mail my present...my other sister called the cell, but I didn\'t get it until this weekend...other sister is still mia...my best friend called that day, but forgot to say happy birthday!!!...my gram called to say she forgot to mail the card...the other grandparents I still haven\'t heard from...  Exactly why i HATE my birthday...  '),(19295,518116,'22,April,2002','2002-04-22 00:00:00','   I forgot a great part...he told me that he offered up a testicle for research...what a whacko!  He heard the company I work for was advertising the desire of testicles....for $35,000.  Yeah, right!!!  He called and they told him it wasn\'t true, duh!  He told me it was like me donating an egg...not!!!  I can re-generate an egg, last time I checked, you can\'t regenerate a testicle!!!!!!!!!!!  '),(19296,518116,'18,April,2002','2002-04-18 00:00:00','   lol...really hard!!!  Okay...I was wrong, he did call back, but since I leave early on Wednesdays, I missed his call again...so I called him and lm, and he called back...what do you say to a person you don\'t really know anymore and don\'t really care if you talk to?!  We talked for 20minutes about nothing...I have nothing to say!  What do I say...um, I was really screwed up and needed some closure so I could finally stop thinking about you...well, I got it and now I wish you would stop calling me?!  I don\'t know if I can tell him that.  And not because I want to stay in contact, but because I don\'t want to hurt his feelings AND I don\'t want to promote any psychoticness...he is all about that he has changed, and part of me wants to believe that, but the other part sure doesn\'t want to pay the penalty if I\'m wrong!   Anyway...he went into how he kicked his girlfriend out of his apartment a few weeks ago...hmmm...no straight invitations, but it seemed like he was fishing!  No, no, no, and NO!!!  Wouldn\'t have done that when I was insane, definitely not now either!  He finally let me get off the phone, and said...I\'ll call you soon...and then said, wait maybe you should call me...I just said I\'ll talk to you later...  Can you tell I have absolutely nothing to do at work???  Ugh...it gets so boring to sit here for 9 hours and dink around.  There are only so many things I can look up on the internet!  Hey...I just got a work-related email...yeah--something to do!  '),(19297,518116,'17,April,2002','2002-04-17 00:00:00','   Wow...talk about closure complete!  So I haven\'t talked to or had the urge to talk to Mr X in almost a month...I actually was wondering yesterday what he thought of all of it... if he thought, damn, she did it to me again...roped me in to cut me loose.  Kinda felt bad...half-bad...for half- a second.    So how ironic that this morning I would come in to a message from Mr X asking me to call...so I did, no hard feelings, just no smooshy feelings either...he obviously wanted to talk but I had absolutely nothing to say...he told me he passed his test to be a truck driver so now he is looking for a new job...I almost busted out laughing...not that there is anything wrong with truck drivers...they deliver the goods that I buy...but, to have thought he was what I truly wanted and screwed up...whew, what a momentary lapse of sanity!  So we only talked for three minutes before he had to go...said he would call back later, won\'t be surprised if he doesn\'t...I didn\'t mean to be rude, but I am sure he can tell that I am not dazed by him any longer.  I love him for what he was and the lessons I learned because of \"us\" but that is about it...I would be willing to remain friends, but I really don\'t know if we would even be able to be friends...what do we have left in common?    Been trying to get the house deco done...my lovely hubby is probably ready to kill me...he always says non-chalantly, now, we probably should sit down to write out a list of priorities for the house so that we don\'t spend all our money on things that come lower on the list.  Translation:stop bringing this shit home that we don\'t really need...ah, but until he says just that, I spend away:)  This weekend I\'m going to IKEA with my mom and sis...so I am dreading that pre-shopping lecture.  But, I am just going to have to forewarn him, the bill has the potential to be high, I\'m sorry, but there isn\'t much I can do about it....why?  Because I can\'t just hop in my car and go to IKEA....and two, it is the cheapest place to get stuff!!!  '),(19298,518116,'15,April,2002','2002-04-15 00:00:00','   Never know quite how to start...so I will just jump right in:)  Been on the slim-fast diet for about 2weeks.  Now, you need to understand that I am really on the Sara-slim fast diet.  This means, I try like hell to stick to the regimen but sometimes eat when I\'m not supposed to and sometimes binge.  Great idea, huh?!  I\'m not doing that bad.  It just sucks on the weekend because hubby loves to eat, and then on the days we have lunch together, because I am supposed to have a shake for dinner, but who can do that after tasting real food???  But I am so happy, I have lost some weight already...I officially can fit into the next size down.  Occassionally I could pour myself into the smaller size, but I was in the middle.  Smaller size a little snug, bigger size a little big.  But I went clothes shopping and can totally fit into the small size--yeah!  On my weigh to my old self!  If I get to a 10 I will be happy, though I would love to get back to 8 since I have so many cute clothes from that era:)  I gave up on the 6...selling all that in my garage sale.  That wasn\'t natural anyway, and the others are much more attainable, in a healthy way!  We went to see the gram...she had to get hearing aids.  She was quite put-out that her \"family\" hadn\'t told her she wasn\'t hearing.  Her stepsons wife did.  Well, gram, to be honest, I just thought you were getting Quookie (okay, I had no idea how to spell that one)...and she is still considering moving here which would be great!  She gave us a generous gift and told me to spend it wisely--hmm...could she be reminiscing about the past???  I told her we would spend it on doing some home projects--yeah, my favorite!  So of course, we had to run out and buy a couch!!!  Now, you might think, hmmm...she bought a couch?  But you must realize.  I am the queen of thrift...I can\'t bring myself to pay alot of money for furniture...especially with a toddler--boy!  We demolished our last couch in under three years, so until I can get the B-meister to sit instead of climb (maybe by 20!) it will be thrifty furniture for me!  Anyway...we bought it at big lots--odd lots...wonderful name!  Anyway, I like it and it works with what I want to do my living room decor in:)   That\'s it for right now...I;m tired and have to start working at some point:)  '),(19299,518116,'04,April,2002','2002-04-04 00:00:00','   Wow...what a week and a half!!!  I got a call from an old friend...the type of friend that will always hold a special place in your heart, but isn\'t really a true friend.  I have known her for as long as I can remember...we basically grew up together.  She just doesn\'t know what it takes to be a good friend.  She will be there for you when it is convenient and when she is not busy with a boy.  It has caused a lot of pain and frustration but I know that people like that are truly the ones that lose.     Anyway, she called and left a message that she just HAD to talk to me.  So, of course, I assumed she was engaged (again) since this is really the only time she wants to talk to me...so I called back.  Boy, was I blown away!  No, she did not get engaged, she got married!!!  Okay, to say that I was shocked would be a major understatement!  All I could say was \"wow, congratulations\"...hmm... She told me that no she wasn\'t pregnant...she assumed that this is what I thought--not so!  I just didn\'t know what to say...I don\'t really even know him so I can be happy she is happy and pray that it lasts...  Then, I have been trying to get ahold of one of my best friends...his number was disconnected, so I called his brother.  Now, he was dating a girl who just had a baby--not his--and his brother thought it would be funny to tell me he had gotten married...please, my heart can\'t take it right now!  But no...he just up and went to TX.  Not sure why and not sure for how long.  His brother said he doesn\'t have a number to reach him...he didn\'t come back for his sister\'s wedding...kinda concerns me, that doesn\'t sound like him.  He was so excited to come for a visit...but hopefully he has email for when I get off my lazy bum and write:)  That\'s enough drama for now:)  '),(19300,518116,'14,May,2002','2002-05-14 00:00:00','   hmmm...so much for being back in a few!  I am beginning to think I have an anti-social personality disorder.  I don\'t like being around people!  I would much rather prefer to stay at home, relaxing by myself than to be around \"friends\"...and this is another issue, why is it that I always push away those that reciprocate friendship and draw near to those who suck the life out of me?!  I think that I don\'t want people to know \"me\" and so I only keep those around who minimally acknowledge my existence...why?  I hate to say this, but that is how my family has always been.  Not that I am jumping into the \"I blame mom and dad for all my short-comings/problems\" but I really don\'t think that they know who I am...quite honestly, they choose not to.  I can\'t tell you how many times I tried to let them in...desperately, but they can\'t handle it.  Ultimately, I feel like there is something morbidly wrong that causes people a) to not want to get to know me or b)to run as soon as they get a glimpse.  Therefore, I don\'t enjoy being hurt, so I keep those who would be kind at an arm\'s length distance and pull closer the ones that just use me anyway...  Case in point...my maid of honor.  Why?  That is all I can say!  She has never been and never will be a true friend, yet I continue to knock myself out for her.  After this shower, I\'m done...not that I won\'t talk or do lunch with her, but I am done beyond that...  It just is kind of a bummer when I sit back and think of all the true friends I\'ve shut out and the fact that I don\'t really have any in town.  One of my bf called to wish me a happy mother\'s day, I just started balling...she moved to IN and I miss her alot!  I never fully explained everything to her, but I never felt that it caused a growth defect in our friendship either.  When she left, I was VERY removed...in fact, I still haven\'t returned her call...I\'m a bum...my other bf lives in TX...yeah, that\'s a hike!  But we have the best time on the phone, I just wish I could see her every once in a while!  anyway...enough wallowing in self-pity...  Not much else is new...the hubby about killed me because I spent $300-400 this weekend...I decided I was going to finish buying the \"big\" items for the house...and, oops, I forgot a couple things.  They will HAVE TO wait...I seriously think he will suffocate me in my sleep if I spend any more!  No Mr X...hmmm, wish I would have been the last to call, now I feel like an ass...  '),(19301,518116,'02,May,2002','2002-05-02 00:00:00','   Yeah...i\'ve been procrastinating...  Okay...yesterday, I had to go to the \"Project Engineering Admin\" meeting...one word...BORING!!! Can we all feed more into the admin-bitch persona????  I just sat and looked around and said...uh, hell, no!!!  I will NOT do this for much longer.  Once the B-meister is in school, I will be doing something that feeds my soul, not something that is rapidly sucking my soul out of me!  Ugh...admins suck major dookie!  Stupid job, stupid power trips...personally, I hate admins...they always think higher of themselves than they are...like, come on, it doesn\'t take a genious to do what we do...we are basically white trash slaves...overworked, underpaid, and too stupid to do anything about it!  Not sure what I am going to do...I have been majorly fantasizing...I would love to do something artsy...but I am afraid it wouldn\'t be fun anymore...kinda like when I made all the Christmas presents and wanted to puke for the next 4months whenever I thought of doing anything crafty...  ahh...now that I have that off my chest...i did break down and call Mr X.  It was kinda weird...i used to think i had this premonition thing...whenever something huge was about to happen (usually bad) that I wouldn\'t have any way of knowing about...I would get these mini panic attacks...it was weird and freaked me out so I eventually stopped paying attention to those feelings...anyway, all that to say...B-meister and I were on our way to MI City to hit the Pfaltzgraff outlet when I had this insane feeling that I was about to run into Mike.  Which then got me thinking about what I would do...say...die of embarassment from the way I looked...you know, normal obsessions...after about five minutes, I was approaching the overpass by Benton Harbor and low and behold the Mr B\'s truck passed over it as I was going under...I about crapped my pants!!!  Needless to say, I had to call him the next day to see if it was him...he said nope, he was on his way to MI City!!!  WEIRD!!!  I still am kinda freaked out about it...  gotta take a break...be back in a few...  '),(19302,518116,'26,June,2002','2002-06-26 00:00:00','   I was having some very vivid, very real dreams...so I have been researching online the meaning of dreams...here is some of what I found:  Unconscious material attempting to make itself known, especially knowledge that you\'ve feared. Look to the qualities of the intruder for more awareness about your fears. It may also represent aspects of yourself that you\'ve denied  If the separation/divorce is an old one, the presence of the ex, especially in a positive way, may remind you that this is a quality you need to incorporate into your present life/relationship. The dream could be suggesting that what you learned from a previous relationship may need to be applied in a present one; that you\'re repeating the same mistake; that feelings/attitudes you have about an ex (especially as they apply to men/women in general) are being transferred onto a present relationship. What part of you is an \"ex,\" or is some part of you being \"x\'d out\"?  '),(19303,518116,'25,June,2002','2002-06-25 00:00:00','   Okay...back to epiphanes...and, no, i still don\'t know if i am spelling it correctly...  The other major shortage in my life was/is physical affection.  There were some things that happened in my mother\'s life, which she says totally effected the way she is able to show affection.  She said that she also went through a period when my older sis and i were young that she did not feel comfortable with my dad giving us an abundance of affection.  They were both so hands off I just have always figured i wasn\'t huggable turning into i wasn\'t loveable.  I guess that I have always had all the pieces but i have never been able to step back and put the puzzle together.  Meanwhile, as I was growing up, I was trying to get affirmation and affection from anywhere i could.  This leads me to another revelation.  At some point in our relationship, mr x and i began to have sex...and my understanding of when this happened and his understanding were two very different timetables.  This really freaked me out when he told me this...and i never really understood what it meant.  At one point, through all of my counseling, i wondered if i was raped...but i didn\'t really understand how i could have been and not known.  at this time, i really don\'t think that i was.  for one, mr x was not like that.  i mean he has his faults, and temper/rage was one of them but never did he force himself on me and it wasn\'t like we were drinking together--i never drank until i was almost 21--long after he was ditched!!!  i still don\'t think i completely understand what this means.  but i really wonder if i blocked out when we actually had sex.  My reasoning?  I truly, truly, truly wanted to be a virgin when i got married.  not because my parents said i had to be, but because for some reason i really believe this is the best way...and, trust me, now i know why it is--unfortunately, i learned the hard way...but anyway...i think that i was so starving for affection that this beautiful guy who was the only one who listened to me and wanted to be with me all the time was interested in ME that i wanted to be so close to the one person in my life who wanted to be that to me...i ate it up and let it go as far as it did.  i think eventually i came to terms with the fact i was 1)having sex and 2)liking the feeling of being close (and the horomones) that i was okay with the fact i just threw my morals out the window.  then i was having sex all the time.  it was hard though...having to lie.  i think that is the first time in my life that i blatantly was living a double life.  but i knew the consequences if my parents found out and i wasn\'t willing to deal...my god--they would have thrown me out, i\'m sure!  my parents were always extremely hard on me...i don\'t know why me, but it makes me laugh, hurt, and get pissed all at the same time when i see how laxidazical (again...sp???) they are with my sisters...who of course, think they are harsh...but if they could have walked a day in my shoes!!!  i also think that when things did go sour that in my mind i blew them way out of proportion...now, there were some shady incidents with bodily force being used...and there was the stalking thing...but i don\'t really think that i was beat.  just misused.  again, not making excuses, there is no reason too...but i think mr x and i had an extremely volatile relationship...alot of which stemmed from the fact we were both so needy.  when it was good, it was awesome...when it was bad, it was hell...  i don\'t know why it has been so important for me to come to grips with these incidents, but they have been playing in my mind alot lately...i think that i just have a desire to come to terms with my past so that i can go on in the present.  i don\'t want to be a \"victim\"...so far, two HUGE pieces of the puzzle put together...i really don\'t think i will ever reconcile the school thing...that is one that is still painful and i think this causes me to put blinders on...i can\'t get past it...i can\'t remember what happened and that is so frustrating!  and i don\'t want to talk with my parents about it...  anyway...this really sounds a lot more depressing than it has felt.  it is such a huge relief to finally feel like i have some answers...it gives me back so many pieces of myself that i never really understood and didn\'t know where they were hiding...  '),(19304,518116,'24,June,2002','2002-06-24 00:00:00','   wow...just was sent a really cool site... check it out!  Interviewwithgod.net  I liked a lot of the phrases:  \"You cannot make anyone love you~you can only learn to let yourself be loved...\" \"It is not enough to forgive one another but to forgive yourself\" \"There are many people out there who love you but have not learned how to show or express their feelings\"  '),(19305,518116,'20,June,2002','2002-06-20 00:00:00','   Wow...what a rollercoaster week...  I have been reading an anger management book at the request of our counselor.  And reading that, coupled with some soul searching, I have had many epiphanes (sp?) this week...i have really begun to see my adolescent to adult years in a new light.  I never have been able to explain to people why I hate myself...what happened that made me feel so unloveable...why i am into self-destruction...and why i am so angry at my parents.  I just \"knew\" that i have never felt accepted or loved by them.  Even though I knew that they always told me that they loved me...its the emotional unavailability.  My parents don\'t want to go to a deep level when it is personal.  Sure, we talk about concepts, ideas, etc...which usually lead into arguments but we never talked seriously, cautiously about something that is effecting us on a personal level.  I was never able to open up to my parents for many reasons: judgement, punishment, and total avoidance.  Yep...that\'s my parents.  They can\'t accept that their kids might have faults without selfishishly turning the focus on themselves.  It always has to turn into...so you\'re saying i did something to cause this, etc...you know, sometimes you just want to be able to actually talk about what you are feeling without having to defend something that you may not even completely understand yourself!!! I think this is a huge reason why it has taken me so long to figure this out.  They have me feeling so guilty about being angry at them that I have never had the opportunity to figure out what i am/was angry about!  I have been too busy trying to forgive and forget, yet not completely understanding what it is that i am busting my ass to forgive and forget!    Whew...which led me to my next epiphane...i am so ultra sensitive to rejection, criticism, etc that i am really fucking up my marriage.  Not that I am taking complete blame for our problems...but if scott does not do everything perfectly i get either depressed or upset...or if he voices an opinion i take it as him telling me i\'m not good enough and that i need to change.  umm...okay, in some circumstances that could be true, but freedom of speech was lost somewhere...so i wrote scott a long letter explaining what i was working through and apologizing.  i told him i am seriously committing to changing this about me and that i wanted to accept his faults and ultimately, accept myself.  he wanted to talk about it, but i\'m really not ready to.  it still is pretty open and raw and dangerous.  i think that ever since i can remember people have had this preconceived idea of what they needed me to be for them and i have busted my ass to comply.  My parents needed me to get perfect grades and be the perfect christian and be the right amount of popular, participate in extra curricular, etc...and i did this for a large majority of my life.  my friends are the same way...even my mr x was like this...and whenever i would fulfill this need i have always felt like i am dropped.  it\'s not that anyone needs me as an individual, they need something from me, a shoulder to cry on, someone to fuck, a companion so that they aren\'t alone..etc.  i think that scott was the first person who saw all of my faults and actually let me be me...not that he didn\'t voice opinions but he has never tried to control me.  and to me, that is so abnormal that it has felt more like he didn\'t care then he is actually the one loving me unconditionally.  so as i am realizing this, i am realizing what a disaster my life is because of me trying to sabotage everything.  i really think that i was trying to get him to leave because i wouldn\'t do it...and now, to have a clean conscience and start anew...i am seriously considering telling him about mr x.  why? hmm..to make myself feel beter.  i feel guilty and it would help me \"come clean\"...however, is that the best thing to do?  maybe i should talk about it with my counselor.  egh!  talk about putting myself in front of the firing squad!  but, i don\'t want to hurt scott and i don\'t want to lie either.  who knows...i\'ll think about it but it makes me want to barf.  more later...  '),(19306,518116,'20,June,2002','2002-06-20 00:00:00','   If love is blind, why is lingerie so popular?  just thought this was funny...  '),(19307,518116,'12,June,2002','2002-06-12 00:00:00','   ok...see 6/3/02...started to think I had made a complete ass of myself and i was determined NOT to call Mr X--yet still, I desperately wanted to the ball to stay in my court.  hahahaha...he called yesterday and now the ball is in my court:)  We talked for about 30minutes...not really about anything, I told him that we had gotten away for the weekend, etc and that was about it...nothing earth shattering, but definitely only a \"friend\" basis.  and that is truly all that I feel.  We talked about the \"old\" days...i told him how i had found a couple boxes of jr high/high school stuff at my parents and how halarious it was.  And it truly felt like I was talking to an old friend from the past and once we hung up it was over...no pining, no wondering...yippee!!!  And the best part is I will never call him again and I won\'t have been the last to call:)  anyway...b still has a temp but all else is fine...so I will take him to the dr tomorrow if he still has a temp but i really have no idea what it is from...  tt my bf...it was so awkward.  I don\'t really know why but i didn\'t feel comfortable talking to him and there were some definite barriers...could it be the lectures i was getting?  okay...when i said i had to watch my mouth because b mimics EVERYTHING  i meant i had to stop saying things like \"crap\" etc...the expletives have left my vocabulary a while ago, i\'m not that stupid...ahh, another word that i need to stop using!  Anyway, got a two minute lecture about how if i can\'t say it in front of b, should i be saying it at all???  hmmm...its not that i can\'t say these things in front of him, however, i would prefer my son not to go around saying crap and stupid...thanks!  that was just one of several lectures and i sound more pissed (another one!) than i really am...just annoyed.  not at him, but at how the message was delivered...also sad that it seems we have gotten to a place we probably won\'t come back from and i\'m not sure how or why we are there.  then my other bf...seems like she\'s avoiding...don\'t know if she is but everytime i talk to her she has to get off the phone suddenly but promises to call back...she usually does a few days later but still...i don\'t think we\'ve had a good conversation in weeks...i think it is her whole mr x thing...she says there are absolutely no feelings toward him, only anger...but anger comes from hurt and hurt from openness...so it started from somewhere and maybe she has never had closure...her mr x got engaged and that has to be rough...i don\'t know how i would feel.  but she told me before that she had successfully stuffed those thoughts until we started talking again, so maybe i\'m the link and she is trying to forget...i haven\'t even told her that i think i can come out this fall!!!  lots more to tell, but i have to work since i am leaving early today...  '),(19308,518116,'11,June,2002','2002-06-11 00:00:00','   ahhh....what a relaxing weekend.  We got up North and dropped b off at hubby\'s parents...then 2 whole days to ourselves.  I had reservations for a room with a two person jacuzzi...that was the first thing we did...we probably took 6 total...but it was so relaxing!  We did some crazy stuff, that I will spare you the details on...but it was like we were in college again:) Crazy!  We saw Spiderman and Star Wars...and went out to eat...we went to the bay for about 10minutes because we felt it was a must considering we were there!  We had wanted to go Saturday but it rained...  b got so sick Sunday night...he had a 104 degree temp and then later he started puking   gotta go Mr X just called!!!!!!!!!!  '),(19309,518116,'05,June,2002','2002-06-05 00:00:00','   wow...just when I was getting bummed out this week...i heard from my bf...(see 4/4/02 log).  wow, even though it was only two sentences, it made my day.  he is going to try to call tonight so that will probably make my week.  I have really missed talking to him, though i am very worried about the whole mr x thing...i don\'t know yet if i will tell him because he will be far from thrilled...he\'ll probably be mad, so i\'ll wait to see what is going on in his life to see if it is even important enough to discuss.  it\'s kinda funny how most of my friends are guys...i always thought that it was because i couldn\'t make girlfriends but guys are a lot less maintenance, plus they don\'t usually act catty like girls do.  i just don\'t think that i like to disclose much to others, so i think i only let in a few at a time.  the only problem with that is that when one relationship is on the rocks it knocks out about half of my support system...  okay...so a new admin started yesterday and i had the opportunity to help train.  she was asking me about different people (admins) and i went into this whole spiel about how i don\'t like to get caught up in the politics of offices...blah, blah, bla...so she says \"i\'m the same way!\" and then rants for 10-15 min about who she does and doesn\'t like...how do you know after 1 day?!  ugh, admins are an incorrigible bunch of weirdos!    got some cute lingerie for the get-away...now i\'m debating a bikini wax...ouch!  but it beats shaving...i wonder if a home waxing system would work.  i also picked up some comfy pj\'s and am considering any other fun gadgets...hmmm...not quite decided.  since the hubby isn\'t too into funky i figured the lingerie will be enough for him:)  that\'s all for now...toodles!  '),(19310,518116,'03,June,2002','2002-06-03 00:00:00','   whoa...didn\'t realize it had been so long!  still no Mr X...but now my stubborn side has kicked in and I am determined not to call again...bummer...not that I am still all \"what if\" about him...but sometimes it is nice to know you still appeal to other people, you know?  Especially when your hubby sucks at the affirmation part of life.  He is awesome with Brennan...he always affirms him... however, he seems to forget about me.  Now, understand, I don\'t ask for a lot...I have never been one to demand my guy to give me presents, have a fortune, extravagant anything...and, why???  Because I have always KNOWN that I am high maintenance as far as the affirmation stands.  I seriously go nuts if I don\'t get it.  I know that I am like a good 30lbs heavier...but I don\'t think I have completely gone to the dogs.  I have really been trying to keep better care of myself.  I had kinda forgotten (not really) about that...I think I was depressed and let myself go a little too far.  So, I have been trying to get back on track.  But I have realized a few things...1) if the person you knock yourself out for doesn\'t notice, what\'s the point?  2) because you knock yourself out and that person doesn\'t notice, it becomes more depressing and food looks more appealing.  which brings me to my other revelation...i\'m an addict...not sure where it comes from, but I definitely am an addict...i am addicted to food and other things, but right now lets just talk about food.  i gorge myself all the time.  It is a serious struggle not to, which i don\'t know why i try to stop myself, because when i finally give in, an hour, a day, a week later it is a thousand times worse than if I just gave in originally.  i don\'t really think about anything while i do it...but i feel like shit afterwords.  i think i do it to deal with stress and to continue feeling like shit...if i didn\'t hate myself, where would i be?  it\'s kind of scary...like, everything else has miraculously come to some sort of order, but if i let go of this last demented thing, what will happen?  it\'s hard to explain...i don\'t think i even really understand it...i have such a distorted view of myself, the opinion changes constantly...  the thing that is humorous to me is that i used to be the opposite, i used to starve myself out of disgust...what the hell???  how on earth did i do that???  so i basically have only a few \"hopes\" as far as the weight loss goes...a) spastic colon kicks back into high gear...actually i need to go on a low fat diet to help that out...hmmm  b) i learn how to vomit more effectively...okay, i know that is disgusting.  however, it helped me get through some stressful times...i don\'t know why, and it grosses me out and my hubby (who unfortunately walked in on me doing it).  I was just aching so badly that wretching over the toilet took my mind off it.  ughh...now you know why i gross myself out.  whenever my mom talks about how she doesn\'t \"get\" why we do this to ourselves and how i \"got over it\" i just cringe...i really thought I was beyond it...i still do, that part of me does not reign, however, isn\'t bingeing just as bad???  i do choose to be happy but it doesn\'t mean i know how to cope any better.  so i still have my dark moments.  '),(19311,518116,'29,July,2002','2002-07-29 00:00:00','   -------------------------------------------------------------------------------- Your Celebrity Match is: Dave Matthews   --------------------------------------------------------------------------------  Dave Matthews  -------------------------------------------------------------------------------- Born: January 9, 1967 Dave Matthews is the leader of the \"Dave Matthews Band.\" In 2001 they released a new album, \"Everday\" which went straight to number one on the billboard charts. The Dave Matthews Band also won a VH1 video music award for Best Live Performance in 2000.  Your next closest matches were:   Sean Penn  Brett Favre  Kevin Spacey     **Eat your heart out lis~  '),(19312,518116,'18,July,2002','2002-07-18 00:00:00','   Oh, I am so excited!  I just got done having a discussion with one of my friends from india...he recently was married and his wife is now here.  well, she likes art and wants to take some art classes, which is what i have been contemplating for a while now...so I was talking to him about how I want to take a painting class in the fall and if he thought his wifey might want to go too.  Anyway, he told me our art institute offers classes...duh!!! I never even thought of that!  So, I went on their website, and they even have a certificate program...general or specialized.  Wow!  Sign me UP!!!  I have to start with a basic/intermediate drawing class..good.  I need some help on that.  Then I think I will start with a painting class to see if it is what I have been dreaming about...if not, there is pottery (maybe), photography (YEAH!), and some other stuff...I think, including graphic design (hmmm...) so I think I have found my sould food!  Good god that took long enough!  Now, I just have to get the hubby to agree...they are kind of expensive.  But, hey, Christmas, Birthday, etc...right!  '),(19313,518116,'15,July,2002','2002-07-15 00:00:00','   I really should have written this a long time ago, but it has been a crazy week...  So, we went to the 4th of July parade like hubby insisted...I really didn\'t know what to expect.  I assume he remembers that is where I saw Mr X but neither of us have said anything about it, being the good little communicators that we are!  So, we parked at the High School parking lot and walked...hubby said that he was going to find us a great spot...in my head played the theme from Jaws...I tried not to be obvious when I let out a humongo sigh of relief when we passed Mr X\'s Parents house...but I hadn\'t released the whole sigh yet when hubby said \"how \'bout here?\"  Okay...less than one house down from Mr X\'s...ew.  And, of course, the direction the parade comes is the direction I have to look to see the house.  At first, I thought I was clear...I saw sister, grandpa, stepmom, and dadl but no Mr X..then it hits me...the guy in the hugely baggy shorts, ugly blue tshirt, and BALD man is Mr X!   Wow...what a year can do to a guy!  I tried to stifle my laugh...he is BALDING...and he looks like he has some disease because he is MUCH thinner than when we were dating.  Now, when we were dating, I was less than a 6 with pelvic bone and ribs sticking out and we would compete to see who had more pelvige...but he is way beyond that point...serioiusly, sickly.  Then I saw the new thang...old...bad perm up in a barrette and banana clip, bad bangs, worse glasses....short shorts--too much cellulite for me! and a halter top...now, I do not practice ageism...but unless you have a seriously rockin\' bod...you should not be in a halter top!  Those were made for skinny girls with moderate bosoms.  No big hooters...no barely there...there is a specific body type that can get away with a halter and she was FAR BEYOND it!!!  So, that made me want to burst out even louder!  Ah...it really was good for my ego...  Alas, I settled down and enjoyed the parade with my balding hubby and son...so thankful I wasn\'t about all that.  Then, I realize Mr X and sister are standing exactly opposite of me!  I really tried not to look up...very difficult, but I didn\'t want to let on that I knew they were right there...I really don\'t know why they were, unless they realized or thought it was me and was just verifying that fact.  But, I haven\'t heard from him, so I don\'t know if he saw me or not...Now, I dressed kinda cute, but not too obvious...my flare jeans and a pink tank top...the pink accentuates my tan:)  No, really I hate tank tops, but I have this evil multi-faceted tan going on and I knew it would be a primo opportunity to try to get it evened out...boy, would I like to know what the reaction was:)  But...I probably never will.  Hubby is really having a hard time with this whole thing...and he has every right to.  Counselor said a) I should\'ve never done it (i would agree in theory) and b) I owed it to hubby to be honest a LOT sooner...again, I agree in theory.  Am I making excuses?  no...I 100% agree that I crossed the line.  However, the results were satisfactory for me and will help me in the long run...it is so hard to explain.  But...if I hadn\'t called him, I would still be fantasizing about our \"life\" together and wishing it was over between my hubby and I ...plus, I have been able to get closure on a lot of questions and have been able to look at Mr X with the light of reality, not romanticizing it... as for point b...i think i would have done more damage to tell him in the midst of it that i was 2 minutes from walking out.  I would have had to deal with the counselor and hubby\'s emotions/reactions and not been able to fully complete this.  Selfish? yeah, probably...but again, it allowed closure to an area I really didn\'t think would ever see the grips of death, and now i can happily say its done.  hubby doesn\'t believe that, and I really can\'t blame him...its like a bad re-run.  but i have never had so much peace about this area...and there really are no more questions in my mind about what happened between mr x and i...  my bf is pissed i even said anything and said i was selfish to tell him.  then she said i should have told him after the first time if i was going to tell...hmm, she\'s fucked up.  she continues to lie to her hubby and i think she just wants drama in my life to vicariously live through...sorry chicky!  I hate drama...i apologize this whole fiasco caused drama, but it was not my intent...i\'ve been over that since college...maybe that\'s her deal, she never went to college...now, I\'m just being mean.  it just really disappointed me that she was so harsh and that it was so obvious that she wasn\'t giving sound advice.  that really leaves me back at square one with no one to give me sound advice.  i guess that means i am stuck with therapy for the rest of my life?!  NO THANK YOU!  I am scared poopyless for the day I have to go to counseling to talk about this whole thing.  We went to ss yesterday and I SWEAR he changed the lesson once he saw me...that\'s what we call Conviction (notice the capital C!!!)...  i\'ll fill in the rest later...gotta work...yuck!  '),(19314,518116,'03,July,2002','2002-07-03 00:00:00','   Talk about hostility...  Yesterday, b had four of his friends come over to play outside...it was pretty funny.  at first, b was trying to run around to get his toys that they were playing with.  after the 100th time of telling him that if he didn\'t share he would have to have a time out, he resolved that he would just have to make sure they new the toys were his...so he went around \"mommy that\'s my truck\"  \"mommy that\'s my car\"...to which i kept saying \"yes that\'s your XXX, isn\'t it fun sharing with XXX?\"  i don\'t think he was convinced this was fun!  But it was so cute to watch a bunch of 21/2 year olds who are only children play but never really play together...they all had their own thing going on and occassionally they would acknowledge each other...like by splashing each other or trying to swipe a toy.  Hubby and I made nice I guess...I decided, be upset or forgive...ughhh...i chose to forgive.  So, I tried to stay awake to see him but wasn\'t successful, so I just snuggled with him when he came to bed...hopefully, I didn\'t drool considering I was still pretty much asleep.  Nervous about possibility of seeing mr x tomorrow.  Hubby decided he wanted to go to the richland parade...he knows i saw Mr X last year and i am hoping the reason he wants to go isn\'t because he wants to hunt him down!  i don\'t think he would do that...especially because of his career choice, here\'s hoping!  '),(19315,518116,'02,July,2002','2002-07-02 00:00:00','   whoops...and if I would have just told him what was written down below, I would have been in a lot better place than last night!  I decided to wait up for hubby...I honestly wanted to smooth things over...it is so hard to be without your husband during the week and only get to spend quality time together on the weekend...obviously I am the only one feeling this loneliness because he just doesn\'t put forth the effort...okay, that\'s not exactly fair, but the guy is so fucking melancholy about everything--it drives me crazy!!!  I did ask if that is what the deal was...is it because we got into a groove and he figured he could lax on his responsibility and we would be okay???  He had nothing...nothing to say except I am ornary...what did he expect?!!!  I get so tired of his overgeneralizations...all of sudden I am ALWAYS ornary and there has been NO change...hmmm, I said so basically you are calling yourself a liar...he didn\'t like that too much but I told him that is so contrary to what he has said in counseling and on our date that one of them had to be a lie!  Anyway...it was bad, got worse, and still isn\'t resolved...so we\'ll see if he actually calls today...bastard.  '),(19316,518116,'01,July,2002','2002-07-01 00:00:00','   ahh...what a weekend!  Hubby actually asked me out for a date on Friday...so after a mad rush (he asked on Wednesday night) to find a babysitter, we were able to go out!  We decided to go walk around the Celery Flats and then out for dessert.  We bit the bullet and talked alot about the letter...I was able to explain alot and he told me how he could tell that my stress level had changed significantly from the time I put the pieces together.      I just talked about how I could finally understand where my parents came from...and that it isn\'t that they don\'t love me it is just that they are incapable of loving me the way I need to be loved.  How sad...I mean, I could be angry at them, but understanding the difficulty I have gone through, I just feel sad for them.  Especially to know that I will never have the relationship with them that I want because they cannot handle it.  It leaves a hole in my heart too but I think it would be worse to be them.  To have a child and yet not be able to have closeness with that child.  It just makes me determined not to repeat this with b.  I also told him about Mr X.  He actually laughed...though I don\'t think it was a haha laugh...I think it was more...I can\'t believe this shit isn\'t over...which is pretty much what he said.  He also gave me a compliment--which I don\'t really think he meant as one, but I took it as one:)  He said \"At least you\'re honest, you have always been that too a fault.\"  Exactly!  I have always wanted to stay honest, even when it hurts, I hate fake people!!  Yet, I digress...he said he doesn\'t know what he thinks right now, except that he wishes I would cut it out.  He wanted to know every gory detail...which, of course, hurt.  I mean...I had to tell him that I had seriously considered walking away from our family....he asked if I was going to go to him, and I told him I had considered it.  But when it was there, I knew there was no way I wanted to be without hubby...I sat back and pictured my life without him and it made me cry...I want to be with hubby for the rest of my life.  I tried to tell him that but I think he thought it was bunk to smooth things over.  I told him that he was like my parents, emotionally and physically unavailable and Mr X was really the only person that met those two needs.  I told him about having to find some closure and remembering why it is I walked away.  I don\'t want Mr X...I don\'t think I did in the first place, I think I wanted the things Mr X gave me...  It was exactly this reason why I never saw Mr X...I knew when I had to have this conversation with hubby, that if I had to tell him I SAW Mr X it would be over.  I knew that if I disrespected our marriage to that degree I was saying that I didn\'t want \"us\"...whew!  Thank goodness some of my reasoning skills still are intact!  Anyway...he has been very distant from me since this conversation, which is understandable.  But I get so tired of not being able to TALK things through...  Case in point...I go home for our Monday lunch...am greeted by a HUGE box that he has piled with my papers...Now, this is the guy that has boxes of junk in the attic, basement, etc that I have BEGGED him to go through, and all of a sudden he has time to pile MY junk into a box!!!  Half of it was HIS!!!  Can you tell I\'m mad?!!!  I about blew a gasket...okay, I actually did.  This whole weekend I watched and took care of B while he putzed around...I actually think he realizes and feels guilty about this because he has basically been avoiding me...which is what he does when he feels guilty.  Anyway...I started screaming about how I couldn\'t believe he did that.  That my few papers on the counter weren\'t bugging him...to which he said he was trying to cook and they were in his way...some were on top of the microwave (which is all that will fit on it) and then the others were a) on a stool or b) on TOP of the buffet...none of these places could be used to cook...which of course I screamed at him...he actually laughed and then got pissed because he knew I was right!  then he just started pissing and moaning about everything and then I laid into him about how he never even bothered to say thank you that I watched b all freaking weekend long without a break...and made dinner before the party yesterday AND then he didn\'t wash the dishes so i had to come home to do them!!!  Man, I am getting mad just  typing this!  I just get SOOO sick of him acting like he is better than I am and that by my few papers around the house, I was being a substandard house-wife...which is totally not true considering I do all of the cleaning WITHOUT his help...and he STILL has not given me the list he was supposed to FOUR MONTHS AGO!!!!!!!!!!!!! AGHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHH!  and if he has so much energy to pick up my fucking papers, maybe he should pick up his own...and tell me thank you once in a while, and call me when he is supposed to, and get up on Friday mornings!  The fact is, lately, he has gotten to the same mother-fucking point as before in which I am the one that does EVERY SINGLE FUCKING piece of relationship work and he doesn\'t do SHIT!!!!!!!!!!!!!  '),(19317,518116,'28,August,2002','2002-08-28 00:00:00','   ugh...I get so tired of people projecting their feelings and emotions onto me...  bf called...told her I had two numbers for her x.  Asked her to really consider if she wants them.  I offered myself up as an example of how a spouse can take it...she said a) i would never tell my hubby, i think that was so wrong...and b) i am not looking to \"hook-up\"...excuse me?!  All I responded was that I believe honesty is the best policy even when it hurts...she said, if you were honest with yourself you would admit that you wanted to start something with mr x.  um, no...i have been honest with myself.  do people think that it is easy for me to admit that when i did call to get CLOSURE i was so messed up that i started re-playing a tainted view of reality in my head???  I mean, come on!  The guy was psycho...why was i such a lame-o that i considered it for a few days???  that\'s honesty.  i am kinda upset but i don\'t really know why i care.  i think because it just keeps tearing away the phisod of what i thought was a mature, adult, WOMAN friendship...it is a woman friendship...never able to quite get past the cattiness of nature to be real.  and to that i say UGH.  she also said something to the effect that she thinks her x still wants her...and that i will never really know if mr x still wants me because he never saw me now, just remembering what i was in high school.  ouch!  what the fuck does that mean?  and how the fuck was i supposed to react to that?  now i\'m getting more mad.  i just am tired of people draining me to make themselves feel better and i truly need to stop giving a flying fuck.    on a happier note, since i am throwing around the f word...a guy i work with, who NEVER swears, said clusterfuck a few weeks ago.  I don\'t think i had ever heard that one and it made me laugh SO HARD!  so now, whenever it is visible that he is frustrated, i ask him what word he would use to describe how he feels...to which he replies \"FUBAR, CF\"  interpretation: FUCKED UP BEYOND ALL RECOGNITION, CLUSTER FUCK  good day!  '),(19318,518116,'26,August,2002','2002-08-26 00:00:00','   ahhh...still basking in the glow of home sweet home...nothing quite like the comfort and security of not feeling like you have to bust your ass to live up to a standard that is unattainable.  just being me...  So, this past week was when bf was in town.  He wasn\'t available until Wednesday night...so I invited a few people over for dessert...not many could make it, only K and Ang.  So weird that Ang came, she NEVER comes especially if I am there...anyhoo...Tom was being a dork.  I had to call him to make sure we were still on, this is at 5pm...people are going to be showing at 6:30pm...he says he has to run a load of laundry through the dryer and then he will be over.  I make a run to WalMart...get back at 6:20...Ang shows up 5minutes later.  No Tom, No Tom, no tom...K shows up at about 7...no tom, no tom, no tom...no offense to them, but they were not very entertaining and I didn\'t really want to hang with just them.  I finally call and leave a message for tom asking if he flaked...he shows about 8...whatever!  He looks good, really good, in a Tom way.  Someone told me he had started to look like his dad weight wise...don\'t know what they were smoking!  He is balding a little but is slim and trim and actually dresses this era!  It was so strange to be in the same room, we hadn\'t seen each other since a short visit in \'97 or \'98...I was so happy to see him...and I had absolutely nothing to say.  ugh...hence, why I hate socializing...what part of my repetative schedule would you like to hear about?  the part where I get up and go to work, get off work and pick up b, go home and make a pb&j, put B to bed, go to sleep...yeah, that\'s the routine...not much conversation.  I didn\'t really want to talk about the vacation...I was having too much fun listening to K spill the gossip on everything under the sun.  She was SOOOO shocked I didn\'t know some of this stuff...and I finally said, I don\'t really have the space to remember this stuff that has absolutely no effect in my life.  HA!  She just said, yeah, I never remember this stuff either....um, HELLO?!  You have been blabbering for 30+ minutes about EVERYONE...where does the not remembering part come into play...can we skip to that part???  Always hard to admit when you are the town crier...she actually tried to rope my mom into being the one to blab about something pretty confidential...yeah, right!  To which I said, \"K, if my mom knew, I would have been told before you were and I never knew...\"  THat shut her up for all of two seconds.  A) I don\'t know if my mom would have told me, but I would hope she wouldn\'t have told HER! and B) My mom could have told me and it could fall into the \"i forgot\" clause...but it was fun to watch her squirm:)  Talked to bf...she wants her x\'s phone number...I did better, got 2!  Not sure if she really wants it so I want to have that discussion first...good lord, look at the mess it has gotten me into!  Later!  '),(19319,518116,'21,August,2002','2002-08-21 00:00:00','   Ahhh...the sweet comfort of home!  Went on a week long road trip with the parents...hubby had training and races, so B and I bummed a ride to the vineyard and brooklyn...it was so nice to take a week off.  The vineyard was SO different than what I was expecting.  When I hear of the vineyard, I had always thought, beautiful green everywhere, rich, upscale...mmm, not exactly.  It was very pretty, I love water so I thought it was great, cars were everywhere, people everywhere.  It was neat...lots of cute little shops...lots of people with cute little accents.  My family made ourselves right at home in an italian restaurant...kudos to fernando\'s!  He was so cute...he looked as though he was tearing up the last night when we told him we were leaving the next day...but he maintained composure long enough to tell me he has never met a person who didn\'t like his tiramisu.  hmmm...mark it down.  Fernando, dahling, I don\'t like it...it satisfied the unrelenting chocolate urge, but as a texture freak, it kinda scared me.  But, I am sure it wasn\'t YOURS I am sure I would feel the same about everyone\'s.  It was good to see my sis...although, between my parents wanting to escape B and I and her working, I really didn\'t spend a lot of time with her...  Then it was off to Brooklyn.  Ahh, the big apple.  It was hotter than a mother there!  and of course, my sis\'s air only works in the bedroom, so I felt like I was receiving a steam treatment the moment I walked in the door.  We walked around Manhattan and SoHo...went to the WTC site.  When I was there last year, right after it happened, they weren\'t yet allowing anyone close, so it was really moving to see the huge void in the ground.  Guys were working...got to see the memorial everyone has been contributing to.  Also got to shop at the street vendors.  I felt it would be sacriligious to dicker with the ones selling wtc/nypd stuff.  So I just found what I wanted and bought.  BUmmed I didn\'t get to go to China Town and dicker with those vendors...that is the best part of NY!  Also, spent a few hours in H&M...the parents and B collapsed in some easy chairs and were almost passed out by the time I was done.  Went to D&DL...all out of chocolate covered expresso beans!!!!  What an outrage...no fix for me!  B was so good in the car!  The grandparent factor was working miracles...  It is so nice to be back home...i have come to so many revelations with my family, and made such great strides...I was feeling myself slip when being with them for so long.  It is so hard to still be the black sheep and the one who can never do anything right.  It gets old to hear them sing the praises of every other sister but never have anything nice to say to me.  It hurt.  It got old.  Glad to be back in the safety of my predictable life.  Glad to have a hubby who missed the shit out of me and couldn\'t wait until we were back as a family.  I think that is what hurts.  I don\'t think I will ever feel safe with my family...always waiting for the punch to come flying out of nowhere...you know it always does.  I love my family to death and would do just about anything for them, but at the same time, they drain and stress me out.  So, I was glad to get away but came to the realization extended family vacations are a thing of the past...  More later...I have to do some work-ew...  '),(19320,518116,'02,August,2002','2002-08-02 00:00:00','   cool...my bf called last night and it was actually a NORMAL conversation...we goofed around, gossiped about the people we knew, you know, bf crap...near the end he got very preachy, but it was more understood from the other things he had told me.  it was good to know that our friendship hadn\'t become all weirded out...  he is coming to visit! of course, he is arriving the week that I am going to visit my sis in Martha\'s vineyard and my other sis in NY...but he will be here for a partial week once I get back.  He wanted hubby and I to go to Cedar point with some others...but a) hubby will be out of town, b) b and I will just be getting back, and c) some others are ones that i can\'t stand for 2 hours...let alone, in a car, for HOURS and standing in line together for HOURS!  He thought it would be all cool because that is what we did on our senior skip day in which only he and I skipped!  yEAH, that would be cool from that standpoint but too many other eghs.  now, my other bf is upset because i told her i might not be able to come this year to TX...Pharmacia was taken over by Phizer so I don\'t know if I will have a job by the end of the year...also, with this NY trip, it is costing what I would have spent on TX.  I just don\'t know if we would feel comfortable spending that money so closely together...I used the cop out \"don\'t know if I\'ll have a job\"...but I don\'t think she\'s buying it...I told her I would keep my eye out for a cheap ticket but I can\'t promise anything.  that\'s all for now...I have to go work:)  '),(19321,518116,'27,September,2002','2002-09-27 00:00:00','   you know...sometimes you just have to laugh to keep from crying...  my boss\' boss was supposed to get back with me on Monday about a raise I asked for at the beginning of August...when I called on Tuesday, he said that he needed 1-2 more weeks...whatever!  I am pretty stressed out...especially when I looked on line and found out I am making the lowest amount an admin is paid in the area...and that person enter with 0-1 years of experience!  I guess I need to get out of my comfort zone and start looking...yuck  To top it all off...Scott has cut me off...no talking, nothing.  The most contact we have had with each other is when I kiss him goodbye in the morning...We had an argument on Monday (?) night about his attitude during our phone call that morning...I had asked for some money for clothes...he said no...I WANTED to say a few words, but tried to take the high road and told him I should get off the phone.  He got all snotty and threw in my face \"Now you\'re going to tell me that I treat you like...yadda, yadda, yadda\" Stuff that I said in our last \"heart to heart\"  Then I did get mad and said goodbye.  I tried to calmly talk to him about how that is not mature and we need to get beyond this so that we can have ADULT conflict...again, bearing my hurt, and my soul...he INTERRUPTS me in mid-sentence to ask me where the 6 dollars is from our date.  WHAT?!  I asked him if he was serious...he was.  Boy, was I mad.  Anyway...I guess this is all just proof for me that I am not psycho...if I bring up conflict he totally retaliates later...right after, actually.  So this morning, after the stress of work, the stress of my husband cutting me off, I don\'t give him a kiss goodbye.  He gets pissed.  I said, you haven\'t talked to me all week, yet you expect me to be lovey-dovey...I\'m sorry, I can\'t.  Then I walked out...then I walked back in and said, \"You know, the funny thing is that you can cut me off and treat me like this yet think that I should still be affectionate with you.  Which is ultimately, why I will never understand you\"  And walked out.  Looks like another great day...I really did not want to come to work...but staying home didn\'t sound very fun either.  I just want to cry.  The girls asked me out to lunch so hopefully that will cheer me up a little!  Wish me luck  '),(19322,518116,'23,September,2002','2002-09-23 00:00:00','   wow...who knew?  Hubby and I actually had a good weekend...I thought for sure it would turn out shitty after our argument...he always tries to pay me back for confrontation.  My parents watched b for us on Saturday...we went to a restaurant and shops in IN...it was actually fun.  Then, we came home and watched a movie, drank some wine HE had picked up...and other stuff that I will spare the details.  Parents decided not to go to church so we got to lounge in bed all morning...wow, amazing what some much-needed, uninterrupted sleep will do for a person!  We went to my parents for dinner...had a convo with youngest about school...she wants to switch her majors.  Tried to give her some friendly advice and offer suggestions of people she might want to talk to...she FLIPPED out.  started yelling about how she didn\'t need me ganging up on her...ugh...i felt bad, she obviously thought i was trying to tell her what to do, which was NOT my intention.  sometimes i really wish that we lived thousands of miles away.  my other sis is completely ignoring/avoiding me...i think that i will no longer invest.  i can\'t stand hyper-emotional drama...over that a long time ago...so whatev with a capital W...if they want to talk to me, they know how to get ahold of me.  bf still strange...again W!  '),(19323,518116,'21,September,2002','2002-09-21 00:00:00','   Okay...I know that the time always shows up funky...but it is 3:45 am...yeah...never happy times when I see this time of day!  Hubby and I had it out tonight...it is rather ironic that it is always his damn counseling appointment that starts it.  Today was a new one...I am childish.  Isn\'t the counselor supposed to help???  No, not ours...ours likes to give hubby fuel.  the great news is that he (counselor) wants me to come in next time.  Yippee!  I told hubby that i would but that i refuse to defend myself.  i mean, they already have it figured out that i am the root problem, so what can i possibly say to change that?  anyway...i need to calm myself down, so unfortunately, i will be major hyperv...i mean, ventilating... okay...1) knowing i was not going to see hubby for week when i went on vaca...asked numerous times for intimacy...rejected every time...lovely 2) when got back from vaca hubby NEVER and, yes, i repeat NEVER asked to hear the gory details of our trip.  I know his, including how many times he took a dump, but he doesn\'t know shit about ours (sorry, bad pun) 3) roof leaks...hubby suggests i talk to the guys at work about what to do...guys at work immediately offer to help fix and tell me what weekend they are available...i am to blame because i took over... 4) hubby refuses to call dad and friends to ask for assistance...i get frustrated 5) hubby is supposed to call to rent a roofing nailer...waits until TWO DAYS BEFORE to do so...yes, i get very frustrated 6) oh yeah, hubby tells parents at races that we will be attending cousin\'s wedding the weekend before the roof...then blames me and says that I am the one that told his mother...uh, NOT!  Then, gives no explanation to them as to why we won\'t be attending...which, as covered previously in counseling, leaves them to believe it is MY fault 7) hubby tells parents that he is bringing b up with him while he will be chopping wood...no he NEVER asks me if this is okay.  he mentions that this was an idea of his...again, tells his parents that won\'t happen and who do you think they feel is the problem??? 8) hubby calls everyday at work...almost everyday (I\'ll give him 85% because I feel generous) he starts nit-picking and an argument with me WHILE I AM AT WORK and can\'t imagine why i get upset... 9) cops an attitude constantly 10) still shows no sign of affection or intimacy for FIVE FUCKING WEEKS and yet, i am the one being childish because i suggest that this behavior might have something to do with why i am upset at him...go figure!  The saddest part is that hubby told me he just isn\'t comfortable with close relationships.  um, duh?!  then you should not have dated me for four years...and married me!  this is what really kills me and really is my grand point.  he knew...HE KNEW...i had no idea.  i told him UP FRONT all of my faults...my temper, my lack of drive for cooking, i am madly high maintenance with sex and affection...HE KNEW.  on the other hand, he played a good game getting me to think that he was very affectionate and loving.  for god\'s sake, he knows more about me than just about everyone...there is one person who may know more...but that is questionable.  that\'s what hurts.  my best friend was all an act.  and for what?  a good screw?  please, he could have gotten it elsewhere.  i am not really sure what his motivation has been for these years.  but it hurts.  i mean, isn\'t that was is \"so great\" about marriage...having your best friend live with you?  well, i thought that is what was happening...then reality set in and i find out i am living with a man who acts like he despises me...yeah, that\'s great for the self-esteem.  i really wish i could get beyond the point of caring.  but if i go there, what do i do?  walk away?  not really the life i have envisioned for myself.  divorced, single mom, child in daycare 40 hours...yeah, that\'s really fair to b.  what my roommate said is really starting to make sense...suffer through this stage...having kids and raising them...then divorce and find the real love of your life.  However, one flaw...after that many years, do you really believe it can happen?  damn.  i really wish i was not so interdependent.  i would love to have stayed single...  '),(19324,518116,'19,September,2002','2002-09-19 00:00:00','   A friend from work sent this to me and I thought it was cool (thanks Rebecca!):  A WOMAN SHOULD HAVE    ...one old love she can imagine going back to...   and one who reminds her how far she has come...    ...enough money within her control to move out and rent a place of her own even if she never wants to or needs to...    …something perfect to wear if the employer or date of her dreams wants to see her in an hour...    ...a youth she\'s content to leave behind...    ...a past juicy enough that she\'s looking forward to retelling it in her old age...    ...a set of screwdrivers, a cordless drill, and a black lace bra...    …one friend who always makes her laugh... and one who lets her cry...   ...a good piece of furniture not previously owned by anyone else in her family...   ...eight matching plates, wine glasses with stems, and a recipe for a meal that will make her guests feel honored..    ...a feeling of control over her destiny...    EVERY WOMAN SHOULD KNOW...    ...how to fall in love without losing herself...    ...how to quit a job, break up with a lover, and confront a friend without ruining the friendship...    ...when to try harder... and when to walk away...   ...that she can\'t change the length of her calves, the width of her hips, or the nature of her parents...    ...that her childhood may not have been perfect...but its over...    ...what she would and wouldn\'t do for love or more...   ...how to live alone... even if she doesn\'t like it...    ...whom she can trust, whom she can\'t, and why she shouldn\'t take it personally...    ...where to go... be it to her best friend\'s kitchen table... or a charming inn in the woods... when her soul needs soothing...    ...what she can and can\'t accomplish in a day...a month...and a year...  '),(19325,518116,'17,September,2002','2002-09-17 00:00:00','   talked to bf last night...still strange.  She is talking about coming up here with her cousin this fall.  I don\'t really know how I feel.  It will be for just a weekend so I am not sure if I would even see her.  She\'s coming to see family...I didn\'t know she even liked her family.  I know she is tight with her gma but not sure beyond that.  Anyway...I had some time to think this morning, I was more coherent than normal.  Thought about what I would say to her if I thought it would be received the right way...I have had plenty of friendships in the past that have sucked me dry.  I am very beyond the point of \"needing\" friends.  I don\'t bust my ass for my friends to turn around and have them dump on me.  Occasional dumping, I\'m cool with.  I understand everyone has to vent to someone...and I feel honored if I am that person, however, dumping and personally attacking are two very different things.  I don\'t do the personal attacks to make you feel better.  It\'s about basic human respect...which took me a LONG time to think I deserved,  and now that I KNOW I do...I don\'t deal with people who don\'t give it.    Now..why don\'t I actually say this.  I want to believe that this was a freak occurence, but I don\'t know...it has happened twice.  The other part of me thinks she does this because she feels bad about herself.  WHich, isn\'t that why people always do it?  So, yeah, basically it comes down to I am a hater of conflict, I\'m scared...I have lost many friends and I am not ready to say adios to her.  '),(19326,518116,'16,September,2002','2002-09-16 00:00:00','   Not much new...RR marathon this weekend, so that kept me happy:)  Got to catch up on the episodes I missed when we didn\'t have MTV.  Hubby went to help chop some wood for his gma up in TC...so b and I bummed it this weekend.  Hubby had told his fam he would bring b up for them to watch while he did this...of course, he never consulted with me.  I was NOT pleased.  He said \"I just figured you wouldn\'t want him here\" uh, hello???  He is MY son!  So, I left it up to him to decide what he was going to do but let him know I was VERY displeased with his action...not consulting me.  He has a shitfit if I do ANYTHING without running it by him.  Can you tell I am very unhappy right now?  I just get very frustrated, which leads to anger, and I am so fucking tired.  I don\'t have the energy for my hubby.  Like today, I asked if instead of hiking it to my house for lunch, b and I could bring some pizza.  Which, we had talked about us doing.  No, I spend too much money this weekend.  Okay, I spent $32 for Christmas shit...$15 on dinner (which was pre-approved by the $nazi)...will not be getting a puppy...and $17 at WalMart on stuff I needed.  Ah, hem...he spent a butload on gas to get to and from TC.  SO don\'t act like I was the only one who spent $...but, after I am done typing, I am going to Lerner because now I\'m depressed and I am sick of his shit.  ANyway, he has been picking fights with me a lot lately and I don\'t know what is sparking it but I am done...he can have his shitfits bite my big ass.  So anyway, he told me no, don\'t come to work.  Fine, I\'m not coming for lunch either.  bf broke down first:)  She called last night, but b and I just got back from WalMart and it was 9, so I told her I would have to call back...I\'ll probably do that tonight:)  Yeah...yesterday was cool...it was kind of rainy.  B took a 3 1/2 hour nap!!!  So I got to watch lots of RR...I don\'t like to watch with him there...he is too young for the adult content.  I don\'t really think I am ready to explain Danny and Paul making out with each other.  And I got to take an hour nap...  that\'s all.  '),(19327,518116,'12,September,2002','2002-09-12 00:00:00','   my goodness...it is a good thing that I am so determined not to get into my depressed stages anymore!  This week has been ROUGH!  see previous entry for one explanation...work has been so hectic.  I just celebrated my year anniversary--woohoo--and I\'ve realized there are dry spells, really dry spells, minor work, and overload.  Unfortunately there is no happy medium.  This is the point I came into last year...with one team, it was a nightmare.  This year, thanks to my manager, I support 2 1/2 teams.  This is my lunch break for the day because I have so much #!@$ to get done.  Anyway...background info...refer to the 8/28/02 entry as to why I was already annoyed with bf.  I haven\'t called her since then...I think I talked to her one time because she called.  Don\'t get me wrong, I still love her...I still want to be her friend...BUT I am highly aggravated.  So, knowing how I am, I thought I would not talk to her until I can think through it and basically get over it.  And, for some reason, i haven\'t gotten over it yet.  I have been crazy hectic lately, so I\'m sure that has a lot to do with it.  Anyway, she called last night, after a terrible day at work.  I was bumming with Brennan...we were both reading, so I pulled myself off the couch to get the phone...she started off with some snide comment about how she hates it when I screen my calls...whatever, told her I had to pull myself off the couch.  She immediately jumps into how she can\'t stop thinking about ex and should she or shouldn\'t she call...but that she doesn\'t want to wait until october 4th...the day before his wedding.  Then she says \"weren\'t you going to call hiM?\"  Yeah...for about two seconds I contemplated that!  I just do not like him...period, period, period.  I told her I never did and didn\'t want to.  She started to lay a guilt trip on me that she called Mr X for me and I owed her big.  Whatever, I guess I thought that is what friends are for.  I just said that she didn\'t hate Mr X like i hated her ex.  I don\'t do guilt trips.  So I tried to change the subject, she just laid into me for that too.  So I just shut my mouth...she went on blabbing and then asked what was wrong.  I told her I had a bad day at work...that I was overloaded at work.  THEN she starts \"Oh, poor you, you actually have to do work when you go in, instead of just getting paid to sit on your butt\"  MY HOLY GOD!  WHERE IS THIS COMING FROM???  I said ABSOLUTELY nothing.  Deafening silence.  A, look, you just really fucked up, silence.  Then she says, Who am I to talk? haha...no, no haha\'s.  But, yeah, that brings up a good point.  You wanted a live in nanny/housekeeper and you don\'t work outside!  I do...45-50 hour/week, have a house to keep up, a husband, etc!  So, yeah, I\'m real fucking lazy!  She tried to recover but I was done by then and just was saying , mm-hmm, yeah, yeah.  Then she said, well I can tell you are in a pissy mood (YEAH...THANKS TO YOU!) call me when you feel better.  Okay, bye.  UGH!  Could I have just quietly mourned the 9/11 anniversary without going through petty shit?  A girl at work talked about how it was the firefighters\' fault that they should never have gone into the towers...whatever, they were following commands.  And, as the wife of someone who is in that line of work, that is their conscientious obligation.  I don\'t really think that anyone believed those towers were going to collapse.  On a special I watched last night, it said the first tower lasted one hour forty-three minutes.  The second lasted fifty-six minutes.  That\'s a long time.  Long enough that you think you have time to save the thousands of people stuck in the buildings...personally, I am in high appreciation and awe at their efforts...I don\'t think I could convince myself to go into that to help save complete strangers.  So, for all the non-believers and critics, there are many others who hold you in your much earned HERO status.  Ah...explosion at the plant on Tuesday.  A runaway reaction.  It basically exploded in a woman\'s face and I just heard they had to spend 5 hours on her face.  She hasn\'t been released because she is having difficulty with her sight.  Her lab partner had his major arterie nicked.  47 stitches for him.  He was just released today.  Scary.  On a happier note...I always watch our animal control PBS show.  Saw the cutest puppy so b and I went done to check it out.  It is a little girl, about a year old.  She looks like a mix of Jack Russell/chiuhauha.  There is one person ahead of us for her.  So I have to call tomorrow afternoon to see if she is available still.  Then I will pick her up on Monday.  Not sure if Scott will go for it, maybe it will be a SURPRISE!  '),(19328,518116,'10,September,2002','2002-09-10 00:00:00','   ahhh...family time....  That should bring a warm fuzzy feeling, right?  Not so in the barto home.  It only brings anxiety, stress, and upset tummies...  My big sis has been in town for a week, so I have been seeing a lot of family.  Last night she wanted to cook dinner for everyone.  Couldn\'t think of a good reason to bail.  So I went.  It was a hideous day at work, which stresses me out to begin with...so I had no tolerance for family issues.    I asked my sis about her birth control because my gyno wanted to put me on high dose to get me to have a period which would hopefully force me to ovulate...my mom started saying \"Oh, no...don\'t do that!\"  Back off, mom.  She has absolutely no tenderness.  Thankfully, I was beyond stressed so I just told her calmly why and that I\'m not in a back-off-bitch tone.  Then, as if I was asking for more abuse, I decided to tell them the great news the dr told me...that with a little pill, he thinks I will be able to get preggo.  This pill will make my body ovulate...which is what the dr thinks is my problemo.  I don\'t really know why I thought that my family could be happy about that.  Because I\'m a fucking moron.  I should know that my parents are so self-absorbed that they don\'t give a shit.  A response of, \"yeah, but doesn\'t that make you OVER ovulate...releasing more than one egg???\"  As if this was some mortal sin.  Yes, if I take it, we have to also realize the risk of multiple births.  But since I am not the perfect woman that you obviously were, my body won\'t get pregnant without help.  And since we don\'t believe in more invasive forms of help, this is our last chance.  And that comment was absolutely it...oh, wait, my older sister said she would be jealous if we had twins...so i have a mother who thinks i am interfering in something i shouldn\'t and a sister whose reaction will be jealousy.  Wonderful...just another fucked up day in the world of barto\'s...  Had to get out of there...ate too much dessert, trying to round up b so we had to take a potty break, because we are training (YEAH!).  Went into the upstairs bathroom...b put up the lid and said \"poopy\"...I quickly shut it again and tried not to hurl.  told b that we needed to go to a different bathroom because this was broke.  tried not to make a big deal out of it, but told lil sis she needed to take care of her mess and went downstairs.  it wasn\'t \"poopy\" it was vomit...just like what we found in the hotel at the island...just like i\'m sure was in the toilet of bennigan\'s the night before...i hurt.  i know you read this, and i don\'t really know what to say to you that i haven\'t already said.  i fight the impulse to tell mom and dad...the only reason i don\'t is because i can\'t believe they don\'t already know.  and i\'m not convinced they have coping skills...they didn\'t when i went through my dark time, and i haven\'t seen a whole lot of change.  i don\'t want to make things worse...but i don\'t want you to think i am naive or turning a blind eye.  you need help...you need help...you need help.  i don\'t know who from, i don\'t know what will make it click for you.  but get it...embrace it...apply it...choose to use it.  this is not healthy, it\'s scary.  if i knew of someone to \"tattle\" to, trust me, i would be on that phone.  but i don\'t...i don\'t even know the goddamn name of your counselor.  you have to work through this and get beyond it.  if nothing else, choose to stop doing it.  it\'s hard, i know...it hasn\'t been until the last six months that i have been able to throw open the closet doors to empty out the skeletons that haunted me.  trust me, i still have dark times...still have self doubt, self-inflicting pain--convincing myself i don\'t deserve things, but it is a choice...it is a choice to entertain these thoughts, it was a choice to stop eating, it was a choice to have a psychotic boyfriend...you are choosing to vomit...to take diet pills...etc.  it must take an extreme amount of energy to continue this.  apply that energy into happiness.  I have told you before, only you can choose to be happy...it isn\'t an event, it isn\'t something that hits you, it isn\'t based on life circumstances...it is the way you choose to view things.  it is hard at first, but it gets easier.  trust me, it is worth it.  it takes too much out of me to be depressed about the little things...and i was someone that loved to be depressed...i loved to complain...i loved the drama.  but being happy is so much more.  you have a wonderful life in front of you...think of all the time you have already wasted by making your focus this.  i hope that you tap into the strength i know you have to turn away from this.  i hope that you decide to wake up and grab life by the hand and take it for all it\'s worth...and it is worth much more than hunching over a toilet for the rest of your life.  i love you more than you will ever know and i want nothing but peace, love, and joy for you.  '),(19329,518116,'29,October,2002','2002-10-29 00:00:00','   Wow...I can\'t believe so much time has gone by with nothing to say!  Okay...not really, but I have been trying to do some soul searching and changing.  I have come to the conclusion that my hubby is my hubby and I love him in spite of it.  He is who he is and may never change.  I used to think that I would divorce and leave...but I don\'t want that.  I truly cannot imagine my life without him and do not want to have a family in which he isn\'t there.  So, I have given up being angry at him...I am still hurt, I probably will always wonder what it was that caused the dramatic shift in behavior, but I don\'t think that I will ever have that answer.  But I do not want to add him to one of the long list of people that have exited my life.  So, we have had a few heart wrenching conversations...and in the past 2-3 weeks, we have been trying.  It is not always easy, and for the first time in my life, I can honestly say I have successfully biten my tongue and OUCH did it hurt.  But, it is better to ward off the nasty comments then to ensue a WWIII.  My mother-in-law is a certified nut bag and has shown her true colors.  She called me and without going into gory details, tried to get my goat.  Thankfully, the soul change had begun beforehand and I did not let her steal my goat.  So, she was unsuccessful and then tried to back pedal.  Too late...you have already shown that you are trying to get me upset and when it didn\'t work she came kissing my butt.  Well, two can play but I prefer not to.  I will be appropriate and, when need arises, remind her that she must be also.  Hubby agreed that he needs to support me and that I need to feel free to stand up for myself.  My friends and pseudo-friends excluded me...it hurt, but it shows you who your real friends are.  And, now, I just ask myself, why do I continue into the same trap of caring so much what others think of me?  I need to accept myself...my hubby needs to like me...and B needs to like me.  That\'s it.  I really need to get beyond these juvenile things and find friends who will uplift me also.  My family...haven\'t talked to the sis in AGES.  She is the one relationship that I absolutely regret...I cannot tell you how awesome it was to finally have someone who comes from the same place that can understand you more than anyone else who has come along.  I\'m so worried about her...and part of it is my fault, because I backed off and maybe I shouldn\'t have.  I didn\'t know what else to do.  I miss her like crazy but don\'t know how to extend the arm...i\'m a chicken, what can I say.  We used to email and talk...now nothing...nothing since I left the island.  anyway...before I cry...  So, life has taken its unexpected turn of events...and like usual, there are times I totally mishandled it and times where I can be proud of myself.  Damn...  '),(19330,518116,'07,October,2002','2002-10-07 00:00:00','   Well...I survived, if that\'s what you can call it.  It always amazes me how hubby escapes under the wire and i get all tangled up in it.  We didn\'t talk about mr x which HIGHLY surprised me...we talked mostly about my temper.  I love how our counselor tries to put words in my mouth...it is so evident that if I agree to what he says I will get punted across the room.  Yes, this occured again Friday.  He tried to get me to say that if hubby showed affection and intimacy that I would not have a temper problem.  Uh, no!  NOT what I was saying....I said, I have had mad stress at work, family life, and with hubby...my hubby has no idea the other stuff that I have been stressing about...so that makes it worse.  However, even if he did his part, the other stress would be there...it is just that with the other stress his not doing his part makes it worse...I am beyond the whole \"i am an island\" bull shit...i want someone there beside me.  Boy, it was funny to see him back peddle on that one.  We talked about our lack of church/christian friends...I finally pointed out after two years that this was not my choice that this was because of hubby.  To which he told me that I could be friends independently of hubby.  Yeah, I do know that...but it makes it difficult when you never see anyone.  It isn\'t like I am going to call up X and say...hey, I haven\'t seen you for a year, you wanna be friends?  Please!  Counselor told me that half of everyone does what we do schedule wise and still finds time for friends...I have friends, just not the friends that I suppose he thinks I should have.  Quite a few of my friends are religious, I don\'t know if they are Christians...heck, I don\'t know if the people that go to my church are Christians!  But I am also beyond the \"you have to go to xyz church to be saved\"...yes, I believe there will be catholics in heaven.  Anyway...I disagree with the half thing...It is very different to work full-time but have both parents home at night...I am a single parent throughout the week (as is hubby)...we have family time on the weekends...It is alot of work, not being able to SHARE the workload.  It just aggravates me that he would say he was an authority on the matter...I don\'t do babysitting much during the week...that is my time with B...and by the time we get home, he has had a full day and doesn\'t want to go out.  I am not complaining...we chose this, we deal with it...but I don\'t need anyone, who doesn\'t have ANY idea what we are going through, to pass judgement on why I am to blame I don\'t have the \"appropriate\" friends...hubby really didn\'t have much to say.    The other fun one is accountability...every week we are supposed to basically sit down and bitch at each other about the week...what you could have done differently, what I should have done differently...does it sound like the breeding ground for war to anyone else???  Healthy conflict is what I have been striving for...not what I think hubby wants.  I basically said I didn\'t like it and didn\'t want to do it...got blasted for that one!  So, we are going to do it...lovely.  Ahhh...I hate griping so much so I will make this the last...church.  Why would you have all couples who are \"committed to each other\" come forward...yeah, i\'m pretty sure it never crossed my hubby\'s mind to go up...we didn\'t and I think we were the only couple left in the pew...I was hurt at first, but now, I just am looking at it from hubby\'s view...says alot to me about where he is at.  Was at grams everyday last week...whew!  I need a break!  It has to be hard to get to the end of the road and have to start divying up your stuff...I am so glad she is in town now...I just need to be able to keep sanity.  It is really amusing to see my parents.  They have no patience with her...I\'m not sure why, but it makes me wonder if that is how I am with hubby\'s mom...  '),(19331,518116,'03,October,2002','2002-10-03 00:00:00','   yeah...i\'m getting nervous...tomorrow is my big \"summons\".  I printed off my blog because, as usual, the week of my husband starts acting appropriately and I need to remember exactly what has gone on.  she did it...she called her ex today.  i think it is bogus that she waited until 2days before the wedding, but whatever.  She of course was crying...she thinks he was sending mixed signals.  THe guy told her when they were getting off the phone that they would probably never talk to each other again...hmm, not sure what signal that was sending.  I shouldn\'t be such a bitch.  I don\'t really know why I am...I think because I am annoyed.  She is so whatever the current conversation is.  Like before, when she told me she didn\'t have feelings for him...yeah, I pegged that one...then today, she said that it was different for me and mr x because i never had feelings for him like she does for hers.  I said, well, i did, in a weird way, but yeah, i knew i am in a lot better place...yeah, i need to get hit upside the head with a sympathy stick.  I mean, what if mr x was doing well for himself?  Then, it totally would have been harder to cut myself off AGAIN...and if, on top of it, he was getting married...wow...yeah that would suck...i\'m a jerk.  I got my raise finally today...and now i have to go...  '),(19332,518116,'25,November,2002','2002-11-25 00:00:00','   Wow...I didn\'t realize how long it had been...and so much crap has happened.  Let\'s start with the not-happy #$@! that way I can end on a good note...b\'s birthday was last week, so we had his birthday party on the 16th.  my m-i-law had said that hubby\'s dad wouldn\'t be coming because it was on the opening \"Saturday\" of deer hunting season...but, low and behold, the bitch showed she is just out to manipulate my hubby because his dad did come!  Ugh!  Anyway...I had two crafts planned and then we walked to the firestation a few blocks away...now, when I had called, I forewarned them that the party was for a 3year old.  So the fireman that set it up said we would just have a two minute talk and then the kids would be allowed to play on the firetruck.  Uh, whoever that was forgot to mention that to our tour-guide.   He was an older gentleman...very nice, but VERY talkative.  He must have talked for 40minutes...at LEAST!  I guess another fireman rescued the parents and allowed the kids to start playing while the other jabbered away.  Thank yoU!  Anyway...as we were leaving, one of my friends pulled me aside and asked who my m-i-law was.  I said \"the short one\"  she, of course, had figured that out...based on past incidents I have vented about...anyway, my friend was standing with her son who had been misbehaving and had a time out from the firetruck and she said my m-i-law started saying how she couldn\'t believe that I would do this for a 3year old party...that this was inappropriate and I should have waited until at least 5 or 6.  Thankfully, this is my friend who is NOT shy in the least...turned and told her how much fun her son was having and she thought it was a wonderful idea...that her son loved firemen and they had already been to his school.  What the #$%^ was my m-i-law thinking?  I just can\'t believe she didnt\' know that comment would come straight back to me...so did my husband...so the only thing left is that she is 1)a moron (already proven) and 2) out to hurt me (already proven)...so really, nothing new came out of it.  I just need to realize this woman is fucked up and that she will never truly accept me or welcome me into her family.  This weekend hubby got angry at me...basically it came out that I am supposed to be cooking, cleaning, and watching b...NOT doing \"fix-it\" work on the house...that\'s a \"masculine\" job.  And he told me to turn in my two weeks.  I was devastated and don\'t know what to think...I told him he knew that cooking and cleaning was never going to make me happy...and that WE bought the house together and that I enjoyed helping fix up OUR house!  What a crazy fool!  Ahhh...now that I have that off my chest...my bf called last night and asked if I would do a girl\'s trip to Vegas in march...At first I was very hesitant about it...I seriously think that I am agoraphobic to a certain degree...I told her I would talk to hubby and then get back with her...well, I thought a lot about it last night and this morning and I am TOTALLY in!  Now, I am praying that it actually happens!  Call it kismet...I had two emails this morning for discounted rates to VEGAS!!!!  I am going to check them out but they seemed legit...one from Ramada...that would be so awesome and much needed!  So, I will try to get ahold of her today and tell her I am there...that also means I have to start penny-pinching!  '),(19333,518116,'07,November,2002','2002-11-07 00:00:00','   Somehow...I don\'t think this is a good thing!     CONGRATS! You\'re a Vampire! Hisssss! You\'re the oddball who sits quietly at parties. You rarely speak, but when you do, its usually so crazy that you probably shouldnt have said it at all. You dumb vampire.  '),(19334,518116,'06,November,2002','2002-11-06 00:00:00','   hmm..what to tell, what to tell.  Hubby skipped his appointment last week because b was not feeling good.  I gave him a hard time and told him it was just because I have been \"behaving\" and he would not have anything to talk about.  In a way it is true...we both have been on our best behavior and it feels so nice to have my best friend back.  We actually can kid around and laugh together...amazing!  His parents came into town on Saturday and spent the night...boy, I have never been so happy to see hubby go to work!  No offense to my hubby, of course.  But, his parents can\'t communicate with me so about 5 minutes after he left, they left...thank god!  It wasn\'t as bad as I thought it would be, but it still wasn\'t good!!!  Hubby brought up the fact that I have given up meat...you know, vegetarian...no meat...I\'m sure you know what I mean...so, on Sunday, hubby\'s mom is making dinner...a pizza thing that has pepperoni in it...so I ask her to please leave a portion without meat.  Seriously, she gives me a look like I have a gun to her head...and asks why...\"because I gave up meat\"...\"what do you mean you gave up meat?!\"  (Thinking, remember our 10minute conversation last night moron?!) \"I don\'t eat meat...but if it is too much of a hassle, I will just pick it back out\"  \"so, does that mean you won\'t let anyone else eat meat?\" \"No, it isn\'t about a social consciousness\" \"what about thanksgiving?  are you going to make my son eat vegetarian for thanksgiving?\"  (now understand, she now has a tone that she is talking to a nazi) \"no, mom.  I will make a turkey but won\'t be eating any myself.\" mom exits in a huff...what a freak!  I don\'t know what her deal is other than she is a lunatic.  She talks about hubby\'s sister-in-law like she created the moon...how very annoying!  I guess I need to be thankful that I don\'t see them very often!  B\'s birthday party is in a couple weeks...I pray that they are only coming for the day!  I am dreading the possibility that his parents and bro and wife will be staying over night.  If that is the case, I am praying for a mammoth migraine!!!  Other than that...my bf called.  B was being naughty so I had to let him go...still weird from his visit.  Actually, that was the first time since his visit that I have talked to him.  Other bf called...said she was in the hospital.  I never know whether or not to believe her...that is so terrible!  She was really shook up about her hubby and her x.  I feel her pain since I went through it...it just would have been a lot easier to support her if she would have been upfront about it from the beginning and not trying to blame me or rip on me about it...  '),(19335,518116,'23,December,2002','2002-12-23 00:00:00','   Okay...so I haven\'t been the most dedicated blogger:)  But, it is the holiday season...the time for madness and mayhem!  I am so exhausted and so not wanting to work...half of the plant has off this week...but I have $$$ in my eyes, so I am hoping to force myself to work at least three days this week.    The fam went to Chicago this weekend....it was fun!  Hubby, b, and I got up bright and early to go to IKEA...can I just say, I LOVE that store!!!  There were so many things I coulda/shoulda bought..but hubby reigns me in...plus, we had to meet the rest of my family downtown so we were doing some MAJOR power-shopping.  But I am one of those people that if I had a five minute time limit could come out with a hundred bags.  I know what I want...don\'t get in my way!  We all ate at Maggiano\'s...yummy!  Had some sinful desserts...then it was to the hotel.  Very nice...  I must give kudos to myself for not being tragically lost in the city.  I actually stayed calm and applied what I knew...my issue is that I know how to get to certain locations...I don\'t remember names of streets.  But we did awesome!  I am so proud:)  Then we went shopping...I much prefer fighting crowds in Chi-town than traffic/loonies in Kzoo.  I just felt like I was home.  I can never explain why...but from the moment I and all of my belongings hit that city, I knew that I had come home.  It is a place filled with so many different people...all shapes, sizes, colors, styles...and no judging.  It truly feels warm and cozy...like I have come running into arms of acceptance.  And as I ramble, I think I am realizing that this is what I miss...here in Kzoo there is nothing to do.  So everyone judges and criticizes.  Does it really matter that I don\'t have on designer jeans...that I bought mine for $3 at target?  But idle hands make idle minds and I think this is the disease of Kzoo life.  So, this weekend was bittersweet.  I hadn\'t been back since February...too long.  It refreshes me so much!  Yet, I realize as I talk to hubby, that we will probably never go back.  He does not miss it...he hates the traffic, people, cost...he says its too much.  So, I am not holding my breath to move back...just looking with anticipation for any amount of time I can be there to rejuvinate.  And, yet, I terribly digress...we were shopping, and shopping, and shopping...finally, hubby took b back to the hotel and my little sis and I went around.  It was alot of fun...she is the one that out of all, I have spent the least amount of time.  We shopped until our feet almost fell off--literally!  And then decided we should go rest...hubby got some krispy kremes...sinful!  And we watched a movie....b decided to get up at the butt-crack of dawn!  Too exhausting...we ate breakfast and left....we HAD to stop by Michigan City to hit the outlet:)  A little disappointing, but that could be due to how hideous b was behaving...  The fam was coming over Sunday night to celebrate christmas...hubby and I decided to have a relaxing Christmas day...so that meant no mad dash to get to fam\'s or have them over....so we did it last night instead!  I baked ALL afternoon long...man, my body HURTS!  I didn\'t get to sit down to about 7:30 and I had forgotten to eat.  It went well...I was more stressed out than I wish I had been.  All in all...a great weekend.  It\'s just a matter of deciding that someone else is not going to dictate how you enjoy your own life...and being able to apply that to family too!  Now...I guess I should talk about what has been going on in the past month...but I have to get planning the rest of my shopping (AHHH!) and Christmas lunch---why didn\'t I plan better?!!!  '),(19336,518116,'23,January,2003','2003-01-23 00:00:00','   Okay...so this is interesting, I can blog, but I can\'t see the results of my blog...so I don\'t know if it went through or not...damn the man!  '),(19337,518116,'08,January,2003','2003-01-08 00:00:00','   I am highly disturbed...my internet won\'t let me onto purplepanda or rw/rr blog...what the heck?!!!  It says it is against the usage policy...after 1+years, they changed it?!  Whatever...I will find a way!  Vegas is for sure...I think...end of March, bf, her friend, and I.  So...I am committing to less eating, and excercise...I don\'t want to be the fattest one of us!  IT is hard enough to be the fat one of the family...but I will NOT be the fat friend too!!!  It sucks...when I focus on eating less, obviously I become obsessed with FOOD which I then want to EAT...ugh.  I am quite happy though...I just ordered a new Cindy Crawford excercise video and the Romy and Michelle DVD...awesome!!!  I cannot wait until they come in!  '),(19338,518116,'08,January,2003','2003-01-08 00:00:00','   I\'m getting better...it hasn\'t been quite as long:)  I am supposed to be training someone but she got caught up so now I have some free time:)  So much to discuss, so little time...  Okay...my bf thinks she might have lupus...I don\'t know if I believe her or not, how hideous is that?  She has an appointment this Friday, so I hope she doesn\'t...we are supposed to be going to Vegas at some point...march or april...she wants to buy tickets and make all the reservations, but knows that if she is too sick she won\'t be able to go...my issue is $$$ I can\'t fork out cash for a trip that isn\'t going to happen.  So I have been trying to slow her down.  Also, I am extremely late on my period...and I am hoping for the best but expecting the worse...I don\'t think I am but I know my hopes are already high that I am...so I will basically be devastated if my damn period starts.  But, if I am, I don\'t know what that would do for the trip...I am at a higher risk of miscarriage and would need to consult with my ob before I did anything...plus, hubby may want us to save $$ so I would not have to work if I am...  Along that line...it is so hard not to have anyone to confide in about the possibility.  I don\'t really have a friend that I could tell...well, one.  But I don\'t want to tell someone, find out I\'m not, and then have that person feeling sorry for me.  I just feel so isolated with this.  No one understands how worried/deformed I feel.  No one gets why hubby and I don\'t want to do anything but the natural way...I just look at b and think that I do not want him being all by himself.  I know that our siblings will eventually have children but how old will b be at that point?  And who is to say they would be close?    oops...gotta go!  '),(19339,518116,'23,March,2003','2003-03-23 00:00:00','   Ahhh...isn\'t it lovely to be working on a beautiful Sunday afternoon???  I can hardly take it anymore!!!  I am trying to make up some hours that I will miss...even though hubby said it wasn\'t necessary...I feel guilty, so here I am!  All that is getting me through is knowing in only 2 1/2 more days I will be on a flight to Vegas...yeah, baby!  I can hardly wait!  I will be driving to meet my flight in O\'hare and stop to shop along the way:)  I am bummed only by the fact that due to my rental car, I will not be able to visit IKEA...and I will be arriving to early on Monday to stop...bummer!  But, I will just use it as an excuse to go later:)  Vegas...here I come!  All day THursday, Friday, Saturday and Sunday...four whole days of glorious freedom!!!  I pray that bf and I get along the whole time...right now it seems as though our eating habits will be what makes the difference...she sounds like she wants to eat the entire time...I gracefully tried to tell her that I can\'t do that unless I am tied to the crapper all day, lovely, eh???  Especially now that the curse has come to visit.  I am praying very HARD that it disappears by Wednesday, it could happen!  Not sure what we will do, Celine is not looking real...no response from the damn dj yet...I\'ll email Sabrina again tomorrow...maybe she can bug her again...she really thought they would be pulling through.  I told bf if we hit big while we are there, I would be willing to go...otherwise, I am not paying $150+ for anyone!!!  I am hoping two things...one, that I have a relaxing, stress-free time...and two, that I hit big enough to pay off a major bill...other than that, no agenda but fun!  Now, mind you, to succeed on number two, it must happen on the $50 limit I have placed on gambling...so, basically, if I succeed with number one, I will be happy!  '),(19340,518116,'10,March,2003','2003-03-10 00:00:00','   okay...so yes, I still am slightly weirded...and, as any good psychology major would, I have tried to psychoanalyze myself and come up with the following (please realize feelings have absolutely nothing to do with what is reality):  I was nuts because mr x was my back-up plan...yes, yes, very sad to admit...but in my head (not the rational part I might add) he was who would be there when no one else would be...being someone with low self-esteem/self-worth and having had many many many people walk right out of my life, here was a guy who thought I was god\'s gift to human kind.  Yes, I do realize he was also psychotic, but that really has nothing to do with my point.  So, when I find out he has now replaced that sound philosophy with someone else, it was disturbing...what about our hypothetical children???  What about our rendezvous in 10 more years???  Yes, I know this was never going to happen, but what if???  Now, I have no one as back-up...which, realistically, is good.  Still a little weirded out...tempted to call and say \"congrats\"...weirded out by the possibility there will be children not involving me...but overall, happy...why?  Because in the real world he was a crazy m\'er f\'er and this means he won\'t be darkening my doorstep...no, that is just a defense mechanism...glad because it sounds as though he has moved on and has, hopefully, found someone that will make him happy and is right for him...just like I found the one that is right for me...  '),(19341,518116,'07,March,2003','2003-03-07 00:00:00','   Wow...it\'s been a while...so much has happened...including celebrating another anniversary of my birth...  But, that is not what I want to talk about...this is all very strange, so please bear with me...  Let\'s start with some background...starting in high school, early high school...I would get these premonitions.  It was weird at first, I would brush it off as coincidence...then, it just got freaky.  It was like a pit of your stomach feeling...something was about to happen, I would never know exactly what--not that detailed, but I would know something.  It became too often to be coincidence.  So, for high school and most of college, I paid attention to this and it was getting more precise...being able to tell who was on the phone, thinking about someone I hadn\'t talked to in a while and getting a call...knowing something was about to happen...freaky.  So, when I had a meltdown in Chicago and purged crap out of my life, this was something I began to ignore...I didn\'t have the energy it took.  So, a few months ago, some of my work friends went to see a psychic...the psychic told one of them she got vibes she was a wicca (SP?)...well, sure enough, this girl had some freaky occurences with a ouiji board...so, my birthday...one of the girls had debated between a psychic reading for me or some quarters for Vegas...she chose the quarters--Yeah!!!  Anyway...so, yesterday I was thinking about if I did go see a psychic...would she recognize if I had any sort of funky capabilities...which got me thinking about the fact that I really hadn\'t paid attention/noticed anything for a long time...  B and I get home...I check the mail and grab the newspaper...same as I do every single day...sorted the mail--mine, hubby\'s, crap...but instead of setting aside the paper, like I do EVERY SINGLE DAY, for hubby, I decided to open it up...once it was opened, I thought, there is nothing I really want to look at...so I just flipped through each section...looked at a couple sales fliers...and on the way to putting it on hubby\'s pile...decide to check out the last page...you know, deaths, births, marriages...WHAT THE FUCK DO YOU THINK I SEE!!!!! Mr X is getting married!!!!!  I am like speachless...I don\'t know which I am more aghast about...the previous shit or the fact he is getting married...I am just so totally weirded out...I talked to bf yesterday about it...she really is the only one who would understand and she vaguely remembers this shit from high school...but I just feel so weirded out.  I told hubby last night about the impeding marriage but had to laugh it off like it was no biggie...which it really isn\'t, I just couldn\'t tell him I was weirded out because he would want to know why, and, quite frankly, I don\'t even know why...  basically, I just needed to discuss some more and I don\'t really know what to say and needed to relay it out again...hopefully, it will help me be non-weirded...  '),(19342,518116,'15,April,2003','2003-04-15 00:00:00','   ahhh...the ho-hum of life settles back in.  I have been relaxing, trying to have fun, and get all my crap done at the same time...not working so well!  I read \"What Dreams May Come\"...one of my fav movies, the book was pretty cool too.  It explained some of what the movie wasn\'t able to cover.  I also read \"Riding in Cars with Boys\"...again, love the movie.  The movie was actually just about a mirror of the book.  I have really missed reading...my next is \"Tuesdays with Morrie\"...I have heard it is a good one, we\'ll see!  For non-fiction, I am reading \"Exciting Careers with Sociology\" or something to that extent...before I commit to going back to school, I thought I would research what my possibilities are with my current degrees.  I have the Psychology one as well but haven\'t had a chance to start.  Planted a dozen rose hedges this weekend...what fun!  I can\'t wait to see them grow/blossom.  6-8 weeks!  I am now waiting for my miniature rose plants for mother\'s day.  I am hoping they get here this week yet...I am going to my in-laws for the first time in a year and would like to bring them now.  My grandma bought an african violet for me...I am hoping I don\'t kill it...my clover is not doing so hot...it keeps dying, coming back, etc!  I don\'t know why...I think it is the feng shui bug biting...but I have a thing for plants lately.  Went to MI City on Saturday with my friend and both our kids.  It was kind of chilly but we had a good time.  Hubby didn\'t scream at me about my purchases either...amazing!  We really have been doing so much better...I think that my resolve to behave mixed with my vacation has mellowed me...things aren\'t as serious as I used to crack them up to be.  It is so much more fun to enjoy life than have your panties up in a bunch.  Tomorrow is  DAY ONE  of the infamous Pfizer take over...we shall see.  No news in the big K-zoo on what\'s staying or what\'s going...  Before I go...I LOVE PUNK\'D!  It cracks me up!  This is so my sick humor...why didn\'t I market that idea first?!  Halarious!  '),(19343,518116,'10,April,2003','2003-04-10 00:00:00','   wow...I really love calling in sick.  Hence, why I believe that I would be okay to not work...  Yesterday, we had to get the furnace maintenanced.  So, I decided the remote chance that the guy would have to come between the time hubby was gone and we weren\'t back from preschool yet, I should just take the day off!  Because of the time change, B slept until 10am...wow, that felt nice!  I decided to start painting the basement floor...hubby and I ate lunch, picked up B from preschool, took a nap, planted some flowers, tried out my rice cooker, read...what a productful day!  My heart is so not into work anymore...all I can think about is figuring out a cool job to have!  Hubby is talking about going to state trooper school...how cool is that?!  I am excited for him and am hoping to do the appropriate amount of nagging/encouraging to get him to apply.  One hurdle at a time...he is concerned about quitting work and not making it through.  I suggested that first he needs to apply and be accepted...he needs to get into shape so he can pass the physical test...then we can discuss the possibilities.  I have a big beef with a company that would not support a colleague to further advance.  In my mind, I would find it an honor to have someone advance to state trooper...a selling point.  Ahhh, but that is rational thought...not hubby\'s department!  Quite frankly, I would be okay if they won\'t let him take his legally approved personal leave, him turning in resignation.  Then, if he doesn\'t make it, he will be forced to find a different job.  Scary, yes...potentially financially straining, yes!  But worth it in the long run, hell yes!  I just want him to be happy and I know this is a dream of his...enter the nagging/encouraging...  '),(19344,518116,'08,April,2003','2003-04-08 00:00:00','   ahhh...still basking in the halarious memories of my trip...I keep remembering new things that crack me up...  okay, have you ever felt fake breasts???  While, I did!  Our friend Sandy at Studio54 had some amazing breasts, which bf was commenting on, she invited us to feel. BF wouldn\'t but I figured what the heck?!  My, fake breasts feel cool...they are so firm with a little squishy.  I am seriously considering having implants just so I can feel them all day..okay, not really!  BFbugging our bell-guy, Pierre.  He had to be my dad\'s age...and she just kept bugging...he told her he had kids, didn\'t phase her... BF walking around without shoes because she wore stiletto\'s in the airport...duh?!  Her feet were so hideous, I made her get a pedicure...understand, I hate feet to begin with...but these feet needed a makeover in the worst way...the pedicure helped a little...but they still need help!  Stan...this kid was halarious...think of Cody from that crazy friday night show that I can\'t remember the name of...oh yeah, Step by Step...yet not so annoying.  At one point, he was yelling at all of us because we were degrading his IQ...he said that his IQ was \"Way Higher\" than all of us combined...so I said...you have a 50...he said \"at least!\" 50 is right above the mentally challenged level.  Chad...he was too cute...he told me that if he never talked to me again, to remember there was someone in CA that would drop everything in a heartbeat for me.  It doesn\'t do me an ounce of good since I don\'t know more than his first name...and I don\'t remember where he said they lived in CA!  BF still thinks I will hear from him...I say, no way.  He totally would have called by now.  But he was so sweet, he definitely was on top of his game!  I just thought he was really cool...and not just because he liked me over BF...I am hoping that by the end of the month I will have just a vague memory of him...  Everything else is going well...B is misbehaving like mad, but hubby and I are doing so much better.  I think overall the vaca was good for my soul.  Seriously looking at graphic arts, with maybe some more english...I could do advertising or such and then start working on books...we\'ll see.  '),(19345,518116,'02,April,2003','2003-04-02 00:00:00','   okay...I am still trying to recover...but ready to spill the good dish...  Understand, my friend is very insecure...this is the only reason I can understand that she constantly asks guys if she is hot and flirts incessantly.  After about, two hours this gets annoying!!!  I am so the opposite.  Quite frankly, I don\'t care what people think...I mean, yes, I want to look good, but who are these complete strangers to me???  So, I am very used to these type of friends getting all the attention...it is the people that look beyond that behavior that I usually mesh with...there were a couple people like that.  It is actually funny, I met more people that were severely annoyed by her than enamored by her...kind of karma for me...I don\'t mind playing the back burner...I am very used to it!  Anyway, you get the idea...we can\'t step outside of our hotel room without her trying to flirt with everyone with a penis.  After a couple days, I was tired of it...I told her I was on penis overload...there is nothing fun about having old, horny men trying to get all over you.  The first club we went to, there were these dorky hispanic men that were too old to be there who kept trying to grind on me...boners gross me out, especially when they are so obviously desperate!  I was getting so pissed at her..I kept trying to ditch them and she kept reeling them back in...she kept saying, At least they are buying our drinks...point 2-I don\'t give a crap...I will buy my own drink, I don\'t expect someone else to pay for it...especially when I know what they want in return and there is no way in heck they are getting what they want!  So I kept hiding out in the bathroom...and on one trip, she busts out and does a line!!! OHMIGOD!!!  I was not expecting drugs to come into the picture!  The freaks that were giving it to her tried to pressure me into it...hello?!  I am 26 years old...a bit beyond peer pressure, please!  If I was going to give in, I would have done that a long time ago...I don\'t believe in using those types of drugs...alcohol effects your brain, I know..but that is as far as I go with an alternative state of euphoria...I\'m a natural girl:)  So that kind of put me off too...  Also, we are in Vegas...did your hubby really believe that we would not be going to clubs dancing and drinking???  About every 3 hours, her hubby would call and need the blow by blow detail of what we were doing and what we had done since the last time we talked...drove me nuts!  Now, I understand that hubby and I went through some extreme situations, he has learned, I don\'t like to answer to anyone...so he doesn\'t feel the need to keep tabs...he trusts that I will do the right thing, and if I don\'t I will get myself back into line...unfortunately, I broke that trust and I don\'t really know that I should tell him...do I want to hurt him again?  At this point, I am trying to pawn it off as not a big deal...a little of it is working...a little isn\'t...  So...it was Friday night...we had met Sara and just partied WAY too hard at Studio 54...we were going to catch a cab...my friend needed to go to the bathroom, so I waited...we were walking out when this guy stops her for a cigarette...of course, she starts flirting with him...by this point, I had gotten so sick of it that I would keep walking and eventually she would catch up...well, she didn\'t, so I had to go back...by this time, she and Stan were sitting on a sofa type thing yacking it up...Stan had a friend...I was nice and said hi, but I think I made it clear I was not interested...but, watching her and Stan were like watching morons unite...halarious!  Two people that should never breed together, too close to brother and sister!  So, Chad and I started laughing about it...he was on the same page...so I figured I could just talk to him...he is from CA, cute, kinda short (just a little taller than I am) and exactly like Dave Matthews...he SAID he had a fiance...so I grilled him, of course she was tall blonde and beautiful...we just talked like old friends, my word, the majority of my friends until I got married were guys, so it was no big deal...I\'m married, he\'s engaged...we decide to go get a drink and so Stan and Sam followed...we were laughing and talking and forgot about Stan and Sam who kept making out...so someone mentions going to their room...I think, no big deal, Sam can get her groove on and I can keep talking to the most normal guy I have met out in Vegas...  We get to their room...not before making fun of Stanley Leroy Jackson, III who seems to be too drunk to remember where the room is!  and has the same birthday as I do...just four years earlier...there are two other friends in there...who obviously passed out after having sex...they were cool, they liked to bash Stan too...it was halarious... They pull out some weed...again, don\'t do drugs, but I\'m used to people smoking pot in front of me, no big deal...Chad opts not to as well...he thought it was \"cute\" that I had morals...yeah.  So, he grabs me and kisses me...I was completely taken off guard...i truly swear, I am the most naive person in the world...I had no fucking clue he liked me...I knew we liked talking to each other...well, low and behold, he is not engaged...it was a line...and, quite frankly, I feel like I am the only one who would fall for a bull shit line like that...I mean, come on!  So, basically, what it boils down to...I spent six hours with this guy, some kissing, some talking, a lot of making fun of Stan and Sam...he was so cool, so cute, so me.  I really felt like I connected with him...don\'t ask me why or how I rationalized all of it, but I did.  Sam and Stan were in each other\'s pants, though,  now she swears they didn\'t have sex...not quite sure I believe that.  Chad was so nice about not pushing himself on me...he kept telling me to relax, I kept telling him I have a son...and that I knew I was being played...which I did and I didn\'t...now I do, yet am still holding out hope he will call me...why?  Because, I feel like if we had met other than that way we would have been really good friends.  He left me with two monster hickies...oh so cool.  I flipped out...how would I explain those???  Luckily, they are gone, but I do feel like a jackass...and I feel like more of a jackass because I don\'t feel more guilty.  My hubby funds this trip and takes care of B and I behave like a college kid...I swear to you, I am not meant to be married.  I feel bad for my hubby...I really do.  I feel like he deserves to know that he can\'t trust me to stick around.  The only reason I didn\'t agree to go to CA with Chad is because of B...that\'s a little fucked up.  Anyway, I told Chad the only way I would believe that the entire night wasn\'t about his penis would be if he called me...so here I wait at work...and, surprise, no call...yet, I am still holding my breath...how pathetic...  '),(19346,518116,'01,April,2003','2003-04-01 00:00:00','   #41    urlLink   What Dave Matthews Song Are You?     brought to you by  urlLink Quizilla    This is crazy!  a) my sis got the same one...but most importantly... b) EXACTLY LIKE CHAD...more on THAT later!!!  '),(19347,518116,'01,April,2003','2003-04-01 00:00:00','   This is an email I sent to tell everyone the \"G\" rated memories...I will fill more in later!   Oh, boy!  I thought I would write one long email...I am so exhausted, yet excited to fill everyone in...it was definitely the trip of a lifetime  Sam and I met in Vegas...we both took red-eye flights out on the 26th...we have been planning this trip for four months, so it seemed surreal that the time was actually upon us.  Both being moms, our gut reaction was to go home to bed, but we knew how lame that would be if that was our way of greeting Vegas...mind you, Vegas is three hours behind ET time...so, our bell boy recommended \"The Beach\"... unfortunately, it was college night.  We had fun though, kind of set the mood for the rest of the trip...if you could get past the girls in their thongs (the cocktail waitresses).  Thursday...Friday...Saturday...Sunday...I won\'t recount the whole trip...we basically did not sleep until Saturday  night...we would take a couple hour nap and go back out.  It was great!  I really don\'t know how we survived...but it was a blast!    The best parts...winning $25 in a slot machine made my week...made it not hurt so bad when I lost at roulette.  Gambling is not my thing...I am too \"frugal\" to lose lots of money.  I ended up playing nickel slots...does not induce much pain when you lose all of $2!!!  We tried blackjack..everyone at the table was so sweet and helped us out...  We LOVED Studio 54!!!  It was so awesome!  I am seriously thinking of marketing this here in Kalamazoo, there is nothing like it where I have been...even Chicago pales to this club!  We met a great girl...named Sara no less!!! Who lives in Vegas.  She was so sweet and told us she would get us in and up to VIP because she is friends with everyone that works there.  When we got there, we were able to bypass the entire line, walk right in...perfect timing...their show was just starting.  There were ladies hanging from swings, doing acrobatics to the music, amazing! We felt the music getting louder and faster...Sara told us to watch up...all of a sudden, two people jump from the rafters, with bungee like cords, bouncing up and down doing all kinds of flips, etc...crazy fun!!!  To top it off...Jennifer Lopez was there...and so was Pauly Shore!  Like I said, one of the hippest places to be in Vegas...I would fly out there just to go there!  We rented this adorable car...it was a little larger than a go-cart.  It goes up to 30mph and you drive it up and down the strip...it was halarious!  We gave a couple people a lift...it was so much fun!  We took it into parking garages....quite a reaction!    The Venetian...beautiful, beautiful hotel!  I spoke with several Italian people...Sam kept telling everyone I was Italian...very embarassing when they thought that meant I could speak italian!  Our gondola ride was fun!  I swear this boy needs to marry my sister...He had an amazing voice and was serenading us down the river (mind you, this is on the second floor INSIDE the hotel)...well, we were all laughing and having fun and he said that he had a special song for us...he pulled under the bridge and said he would have to sing VERY quietly...then he busted out with \"I don\'t think you\'re ready for this jelly...\" etc...yes, yes, he was singing BOOTYLICIOUS...halarious!  We were crying we were laughing so hard!  Anyway...you get the picture...we had the most awesome time, met a lot of cool people, came home exhausted, but hearts warm with many many memories!  It was wonderful to see Scott and Brennan, and I am sure the memories will brighten my boring days at the grind with the engineers who have absolutely no idea what Studio 54 is...(hint~a recreation of a club named Studio 54 in New York, which was big in the \'70s)...  Love you all~ Sar  '),(19348,518116,'23,May,2003','2003-05-23 00:00:00','   This is so halarious...a friend emailed this to me:   MasterCard Wedding!   This is a true story about a recent wedding that took place  at Clemson University. It was in the local newspaper and even  Jay Leno mentioned it.   It was a huge wedding with about 300 guests. After the  Wedding at the reception, the groom got up on stage with a  microphone to  talk to the crowd. He said he wanted to thank everyone for coming,  many from long distances, to support them at their wedding. He especially  wanted to thank the bride\'s and his family and to thank his new  father-in-law for providing such a lavish reception.   As a token of his deep appreciation he said he wanted to  give everyone a special gift just from him. So taped to the bottom  of everyone\'s chair, including the wedding party, was a manila  envelope.  He said this was his gift to everyone, and asked them to open their  envelope.  Inside each manila envelope was an 8x10 glossy of his bride having  sex with the best man.   The groom had gotten suspicious of them weeks earlier and  had hired a private detective to tail them. After just standing  there, just watching the guests\' reactions for a couple of minutes, he  turned to the best man and said,\"F--- you!\". Then he turned to his bride  and said,\"F--- you!\" Then he turned to the dumbfounded crowd and  said, \"I\'m outta here.\"   He had the marriage annulled first thing in the morning. While  most people would have canceled the wedding immediately after  finding out about the affair, this guy goes through with the charade, as if  nothing were wrong His revenge...making the bride\'s parents pay over  $32,000 for a 300 guest wedding and reception, and best of  all, trashing the bride\'s and best man\'s reputations in front of 300  friends and family members.   MasterCard wedding for 300 guests ...........$32,000.  Wedding photographs commemorating the occasion............$3,000.  Deluxe two week honeymoon accommodations in Maui.......$8,500.  The look on everyone\'s face when they see the  8x10 glossy of the bride humping the best man... Priceless.   There are some things money can\'t buy, for everything else  there\'s MASTERCARD.  '),(19349,518116,'21,May,2003','2003-05-21 00:00:00','   It\'s so strange...I tried to pull up my archives, but for some reason, they cut out a large majority of my blogs...oh, to be internet literate!  Anyway...I always get a kick out of where I have been and where I am now...boy, was I a good case for some major Prozac!  I truly take myself a lot less seriously.  The scary thing is, I know that when I was starting to blog I had already come a long way from where I was...wow, I pity my friends and hubby from then!!!  What\'s new...hubby and I are so much better...I am not sure what happened...I know I just got tired of fighting the same fight.  Are we perfect? Hell, no!  I know he will always have his issues, I will always have mine...the challenge is finding a place to meet.  Sometimes closer to you, sometimes closer to me.  We have been pretty successful with that dance as of late.  Friends...wow, I forgot about the \"falling out\" with my mommy friends...unfortunately, that did not change.  The main culprit emailed me a few weeks or so ago.  I never responded.  I am still hurt but realize people have different ideas of friendship.  I am just beyond the point I need to try to force someone to be my friend.  I want a friendship where someone is as invested as I am...and her friendship was not that way.  I did make another mommy friendship...but she is very, very clingy.  She wants to talk every other day...I HATE to talk on the phone...I swear it is from talking too much in jr high and high school.  She will leave messages \"I haven\'t heard from you in a while...um, two days ago!!  It just bothers me...I like my space and independence.  BF...never heard from him again...and I don\'t know how to get ahold of him...I love him, and pray he is doing well...someday...  Other BF...well, I just talked to her for about 20minutes after a little over a month.  Ugh...somethings don\'t change and some days I have more tolerance than others.  I have been having some health issues and was trying to explain and voice my frustration...hubby and I have no give to our schedule...I have work all day, b all night...nothing gives.  \"Oh, that is the same way for me...\"  Umm, you just got done telling me you take them to the pool everyday...and your hubby is always taking time off to take care of them...or they are at the daycare so you can be home by yourself.  How is that the same???  I don\'t know why this surprises me...nothing ever changes.  I just need someone to vent to about it...I am trying to stay upbeat about it, but it is hard to feel handicapped and see so many changes to your body and feel like you can\'t do a bloody thing about it!  T...I feel terrible...this month has been horrible with Scott\'s training schedule...so I have been acting like a hermit.  I haven\'t called her in forever!  I got to talk to her for about ten minutes today...but that\'s it.  I can be such a bitch...I really don\'t know why I get that way...I think I would be quite content on a desert island with my family and that\'s it.  Can\'t wait for the weekend!!! Going to Riverdance with some friends...other than that, hopefully relaxing!!!  '),(19350,518116,'21,May,2003','2003-05-21 00:00:00','   it\'s over...Buffy is all done.  I liked the finale, and I didn\'t...she turned down Angel..yeah!  But they killed off Spike...booo!  And, on his deathbed, Buffy tells Spike she loves him...\"no you don\'t, but thanks for saying it anyway\"...what?!  It was kind of a weird ending...but it was cool...  This is the last week of crazy scheduling...I realized this morning that every weekend in June is booked...that really sucks!  Someday, hopefully soon, our lives will slow down!  '),(19351,518116,'20,May,2003','2003-05-20 00:00:00','   I finally got to watch BOTs reunion...what a disappointment.  As everyone had reviewed on their site...nothing exciting.  How much more interesting if Emily had come and been her \"Emily\" self...oh well...  Bachelor finale...how predictable.  At least I can say I give him credit for choosing brains and beauty over just beauty.  Kirsten should grow up and realize, if you are so busy pointing out everyone else\'s faults, you make people wonder what  you are trying to divert attention from (Like the fact you have a boyfriend waiting for you at home!!!!)  Man...I am so glad I did not waste my time on Mr Personality...now, I understand that I just saw 10 minutes of it...but was that girl a total bimbo???  She seemed like such a flake...I was watching the unmasking (everyone: ahhhhh) hoping a total dog-face would be chosen...is this \"reality\" show real???  She chose a gorgeous guy who just happens to be a millionaire...hmmm, how coincidental!  American Idol: Sorry guys, I don\'t do the reminiscing...I\'ll wait til Wednesday to find out who wins...Reuben?  I hope so...  Not much else is new...went to the Dr.  I have carpal tunnel and possibly arthritis.  My parents both flipped about that...it doesn\'t run in our family so they are convinced I can\'t have it.  Well, I don\'t particularly care what it is at this point, just that someone figures it out and gets me better!  Four more days until Riverdance...two more weeks until anniversary...  I almost forgot!!!  SERIES Finale of Buffy...choose Spike, choose Spike, choose Spike!!!  '),(19352,518116,'16,May,2003','2003-05-16 00:00:00','   I am an overly happy A.D.D kitten    urlLink   Which cute or possibly strange kitten are you?      brought to you by  urlLink Quizilla    How funny!  I think it is very strange that my sis and I usually rate the same on these things!!  '),(19353,518116,'13,May,2003','2003-05-13 00:00:00','   okay...either I have a lot to write about or nothing else to do...suspiciously, I think it is the latter.  And it isn\'t that I don\'t have anything to do...what I have to do just doesn\'t seem appealing.  Anyway...should I stay or should I go?  Still nothing on the job front...will the Pfizer honchos be the end of my reign?  Hard to tell...rumors saying everything under the sun...I\'ll wait it out though...hate change, hate new jobs, hate job hunting.  So, I may be living off unemployment for a while.  90210 reunion...I am kinda upset.  A big storm knocked out our electricity so I missed the last 15 minutes.  But what I was seeing wasn\'t that great.  It kind of reminded me of when they tried to get our senior class to have a jam session and talk about the school.  What do people who despise each other say?  We basically bullshitted...no one liked each other and no one liked our school...what a win-win situation!  Anyway...then I fell asleep so I missed BOTs reunion..bummer!  Catch ya later!  '),(19354,518116,'08,May,2003','2003-05-08 00:00:00','   okay...I don\'t usually comment on tv...how lame-0, I know!  But I have a few things I must get off my chest...  First...the Alias season finale...what was that about???!!! Do they think that I can wait until next season to find out why Vaughn has a wedding finger back on his hand...Sidney has been out of it for two years...and why they allowed Will to survive?!  I can\'t...of course, thinking back...I also was the one who obsessed for months that they could not kill off Vaughn from the season 1 finale...if you don\'t know what happened, a) you obviously have a life on Sunday evenings and b) are nuts for missing this show!  Anyway...very good cliff-hanger, but I don\'t know if I can wait...  Secondly, do you ever watch the Bachelor???  This is one of the shows I watch because I love to laugh.  Just like sorority life...but that is a different topic.  Anyway, last night...Andrew wants Tina to open up...Tina opens up and tells the guy she is majorly falling for him...first of all, coming from a family of people who do not open up, that statement to him, and it was more like a long conversation, was huge for her, and is he too dumb to realize that?!  Just because she is not throwing herself at him or bad-mouthing her competition...then, after she did not receive a rose, Andrew walks her out...he is trying to explain they did not have enough time together, yadayadayada...  then, on his follow-up interview, he says, \"Even in the end, I was hoping to get her to open up emotionally, but she just never would\" (NOT an exact quote)...umm, hello??? You just dumped the girl on national television and you want her to sob to you about how she was falling in love with you...excuse me while I exit the area, there isn\'t enough room with your head in the way!  I didn\'t even like Tina much until last night...So we are left with Kirsten, who I am getting sick of her nit-picking on everyone...\"she isn\'t ready for a relationship\"...this coming from the girl who had her boyfriend drop her at the airport on her way out!  and Jen, she seems nice...very real...as real as reality tv can be, i guess...  and while I am slumming it talking about the scary shows I watch...sorority life...like, if she, like, becomes, like, a sister, like, i don\'t, like, want to, like, be one, like, too...okay maggie!  This show cracks me up...it is the epitome of why I have issues with girls paying for friends.  One of my non-payroll friends and I were talking (okay, really laughing) about it...this is what we summed up: a) sororities, not okay...all they do is cat fight and bitch.  There really is no point except to out-mean each other. b) fraternities...okay.  Reason being, they have two goals, get laid and get drunk (and not necessarily in that order).  No harm done there...no one will be forever scarred by having an angry mob beat up on them verbally because they didn\'t curl their hair the right way. One of her \"sorority\" friends is so outraged by this show because it is so not like the way like real like sororities like behave...excuse me???  This is a mirror image of every sorority chick I know!  I think that deep down, sorority chicks are the most insecure girls...they are the ones that can\'t make friends on their own because they are too scared to put themselves out there.  So they pay for these friends and they all empower each other and they become the bitchy group of friends they are...so, in a way, I guess sororities are good.  They give girls self-esteem...but at what price does the end justify the means?  And, yes, I do realize I am now psychoanalyzing an MTV show!  BOTS...I cannot wait for the reunion...Melissa looks hot!  And somehow, I am disappointed that Genesis and her wife are no longer together...Jamie looks cute with the long hair and minus the beard...  anyway...enough about tv shows...  '),(19355,518116,'26,June,2003','2003-06-26 00:00:00','   well...I\'ve been avoiding.  Not been a good week.  Hubby and I have been going toe to toe...is it coincidence that I have been feeling like shit?  I get tired of whine, whine, whine...so I decided to stay away from the blogger until the whine has left the building.  Anyway...today was actually a good morning...after yesterday, I thought I was going to have to commit myself.  I had been up until 3:30am in hideous pain...ready to cut off what was the root, but I couldn\'t find the root.  Anyway, this morning I woke up at 5am feeling great...except I had a kink in my back.  I tried to go back to sleep but the kink hurt, so, I listened to my body and got my butt out of bed to work out.  Wow...it is so lame-o, but I really feel great now.  I haven\'t been able to work out since Vegas...months ago!  So, to feel good, and have the desire on the same day was a mini miracle.  So, I did the A workout on my new Cindy tape and the first set of abs and arms/chest on the original Cindy video.  I didn\'t want to push too much, knowing I have to survive a day at work and this evening.  But, I did 25 minutes...yippee!  Just a side note...watch Paradise Hotel?  I didn\'t...but I was flipping through and saw this girl that I \"recognized\"...so I watched until I could place her...she was one of the girls from Temptation Island!!! Whatever!  '),(19356,518116,'19,June,2003','2003-06-19 00:00:00','   God has a sense of humor, right?  I hope so, \'cause otherwise this is going to sound very sacreligious...  I want to give a shout out to the Chief (God)...why?  Because I finally found a support group online...so I thank You for empowering someone (I haven\'t a clue who) to create the worldwide web...then to Maddy who created this support group...and finally, for my new \"buddy\" Jenn who is in the US!!!!  It really has helped to have a whopping two messages from this girl who is similar in age and experiencing what I am...so thanks!  '),(19357,518116,'17,June,2003','2003-06-17 00:00:00','   Wow...where to begin...so, my so called friend, the on again, off again...emailed me \"just in case I didn\'t know yet\" that mr x got married.  The thing that makes me laugh the most is that she truly expects me to have an emotional breakdown in front of her...this is the same girl, upon arriving home from the hospital after having b, had left a message for hubby \"in case I hadn\'t let him know yet\" and that I was in the hospital in labor.  Did she truly think I would tell her before I told him?!  Anyway, she would be the last person I would share my emotions with...unless it was a well thought out manipulation to spread the word around kzoo and surrounding areas...  I sound a little bitter, eh?  well, the other friend from work is pissed at me... why?  because I asked to change the dates of a camping trip...my bf\'s daughter\'s one year birthday party was scheduled for that day...well, now she is mad...can we say jr high?  I might even give her high school... this is the same girl that ditched us when we made plans to go to the bar, at the last minute, and never let me know.  this is the same girl that was mad at me when I told her I was diagnosed with PCOS (see a few entries previous)...why do I continue to make friendships with girls?  This is just not a good time for me...my bf and I haven\'t talked in well over a month...the other bf blew me off and has yet to call me back...I get so sick of this and I feel like this is the last type of stress I need right now.    I have been in so much pain lately...when I wake up...pain...when I go to bed pain...and most of the time, pain all between.  I have been trying to keep such a positive outlook on this, but I really don\'t know how much longer I can pretend it isn\'t there.  There are days I think I am going to fall flat on my face because my legs are so numb with pain...but I feel like I don\'t have a choice in my schedule.  I feel the worst for b.  He asked me the other day why I don\'t play with him...that breaks my heart.  I really need to focus on continuing to be a mom...i am so completely exhausted when I get home from work, but I should keep pushing, it isn\'t fair to him.  I try so hard to walk normally and pretend everything is okay...it is worse to try to explain to people what is going on.    I finally had a breakdown last week.  I think it was Thursday...my blood results came back and the dr wasn\'t there to talk to...the nurse couldn\'t tell me what they meant...I was exhausted, I was in pain, my friend had just gotten pissy with me...i just couldn\'t take it.  So, my other friend called for something and i didn\'t have it...and I just broke down.  How embarassing!  I am still trying to pull myself out of the funk.    It is so hard to allow myself to talk to someone about what is going on...I feel like they will think I am whining.  I am just at a loss as to what to do...my mom suggested walking on the treadmill.  I tried to explain how much I hurt, \"oh, my legs hurt too, but walking helps\"...okay...that sounds like you have muscle pain...not the same thing, but of course, she is convinced it is exactly like her, because everything I do is exactly like her...frustration.  I can\'t admit I need help, but I can\'t continue to do everything myself either.  I just feel like crawling into a hole and hiding out for a long time...  '),(19358,518116,'10,June,2003','2003-06-10 00:00:00','   Friend Bear     You are everyone\'s ideal friend because you are sincere and genuinely kind. Sometimes you worry about your friends\' problems so much, you forget about your own responsibilities, which can get you into trouble. For you, it\'s the little things that really count. You also happen to be the main driver of the Cloud Car. No speeding!           urlLink  Take the  Which Care Bear Are You?  Quiz!'),(19359,518116,'09,June,2003','2003-06-09 00:00:00','   ew...I just joined Weight Watcher\'s...I don\'t know about you, but I really hate pushy people caked in perkiness.  That is exactly what this leader was...thankfully, I will not be going to these meetings regularly, I will be going to a different one.  I had the times mixed up, so I went this morning.  We\'ll see...  I forgot to mention...Mr X\'s wedding announcement was in the paper this past week.  He looks like he has shaved his head...she looked pretty...it was funny.  March 15th was the date...ironically, I don\'t really have any deep thoughts on this...I was just glad I got to see it:)  She is listed as a manager of a Max 10 in Coldwater (?) and he is an independent semi-truck driver...  '),(19360,518116,'09,June,2003','2003-06-09 00:00:00','   back from a wonderful weekend away...hubby and I went to Mackinaw City and Mackinac Island...to celebrate four years...wow!  It really seems so long yet only a few moments...so cliche, I know! but it is true...  Tomorrow is my big dive into the Weight Watchers club...fun, fun!  I was talking to my friends at lunch today...one of my friends, who I have suspected has a e.d., blew up...it was kind of comical, and insulting, all at the same time!  She said, workout...I said, I do...she said, watch what you eat, I said, I do!  People like her don\'t understand that not everyone who is fat is lazy...I guess I have been blessed with a life of diversity.  I have met so many people from so many different walks of life, I understand that everyone leads very separate, individual lives.  An event will effect me differently than it will you...unfortunately, not everyone is open minded or embraces diversity.  My other friend was really interested though...she has suffered alot of the similar things I have...she said her dr had mentioned PCOS but she never dug deeper.  I encouraged her to find out more...not for the soul reason of getting pregnant but just because of the health issues later on.  Just her reaction made it worth the other\'s.  I know I will face people like that for an eternity...and that is okay.  I feel bad more for them, than the discomfort they cause me.  I know what it is like to be closed minded, I used to be like that.  You miss out on life...you hold everyone else up to a certain standard, and yourself...no one can live up to that...and you spend your whole life frustrated and mad at yourself and those around you for failing.  Life passes by.  I am not sure why the mind change, but I can whole heartedly say, life is so much more enjoyable now that I don\'t take everything as a life or death circumstance.  There are serious issues, but not everything that arises is serious.  It is good to be able to sit back and laugh...I think I laugh the hardest at myself...  Anyway...sorry about the digression.  '),(19361,518116,'02,June,2003','2003-06-02 00:00:00','   what a long week...  ho-hum...dull drum of life...hubby\'s bro, dad, and uncle came down this weekend to help with the house...I keep saying we are almost there...could it be?  supposed to go to the Wayside on Friday but friends cancelled...gave me a good excuse to guilt my husband over him being an ass....anyway, ended up going to a going away party for a friend with my sis...the same sis this guy was in love with...it was nerve wracking.  High chance of running into christian school bastards...fortunately, none recent.  Just one that has not changed much...pretty funny!  Won\'t the girls be jealous!  It was so good to see P...I haven\'t really seen him since he left the christian school bastardization.  Things were weird between us since Mr X and I broke up...he and Mr X were tight for a while...but it was so nice, nothing about him came up and it was like he and I were friends all over again...however, I kept thinking, I am too old...it felt like your first alcohol party...it was funny...I had a good time though, paid for it on Saturday when I had to get up and be social...can\'t I just sleep a while?  Went to the dr again...she said she can firmly diagnose me with PCOS...is that good or bad?  A little of both...from what I read online, I should be (and am) very thankful we got pregnant when we did...there is no guarantee I will ever be able to get preggo again.  That is hard...but, I was being told that and trying to deal with that without a diagnosis.  Hopefully, it will mean some help with the weight issue...it is hard being fat.  I have been on strike since I got back from Vegas...so now, I need to convince myself to get back into gear, eat right, excercise, etc.  Supposedly, there is a medicine they can give to me if they find out my body isn\'t dealing with insulin appropriately.  This medicine will force my body to deal with it and, thus, allowing weight loss...finally, maybe!  Trying not to hold my breath since I have been fighting a long and painful battle with this for three years.  But a potential end is better than convincing myself I am okay with this.  Frustration...why didn\'t the dr\'s recognize this before?  Why did  I  have to bring it up?  She said the symptoms have been there since puberty...it is just a matter of no one having gone through my entire file...they treat for what you are presently there for...whatever...I am just trying to find someone to be mad at.  Until my bloodwork comes back, she suggested I check into Weight Watchers...she said it will be that much easier when I do see a specialist and I try to get them to prescribe me Metaformin...so, I will dutifully check into it...she actually did not recommend Atkins...whew!  That diet sounds nuts!  She said she has been on and off it three times and, yes, she loses weight but then will gain it right back.  Weight Watchers describes better the correct portions and formulas for healthy eating...here I go!  '),(19362,518116,'28,July,2003','2003-07-28 00:00:00','   A recent posting to my PCOS board...   Just feel the need to take a load off...I\'ve really been bumming it the past few weeks.  I have been trying to get myself out of it...but it\'s not working very well (lol).  I\'m so tired of constantly fighting...I\'m still fighting to get an appointment with the endicrinologist.  This past week they came up with a new excuse...it\'s a doozy.  The dr\'s partners are leaving, so he must now go through all of THOSE files to see who he will be taking on...so no appointment for me and no promise of a date either!  Ugh...frustration.  My hubby gets aggravated because I don\'t throw a bigger fit about it...he doesn\'t get that I can\'t handle the stress that would involve...it is just easier to pretend it isn\'t happening and plug along.  Then...I decide to call for my yearly.  They want a nurse to see me instead of my gyn.  I told the woman I really need to see the gyn as I haven\'t had a period in a few months and am starting to be concerned.  She told me no appointments available through September and she can\'t schedule into October yet...ugh!  And the worst, I finally took the pregnancy test...I didn\'t want to continue taking the meds if I was preggo...and, surprise, I\'m not.  Rationally, I knew I wouldn\'t be...but why do I give myself the glimmer of hope?  I knew it would be negative...yet I allowed myself to hope it would be positive...  Anyway...sorry for the rant.  I hope you all are well and happy...  '),(19363,518116,'25,July,2003','2003-07-25 00:00:00','   Wow...do you watch Fame?  It was so AWESOME!  Now, I get tired of American Idol...can anyone say follow the beat?  Those kids couldn\'t dance to save their lives!  I only would watch for Simon\'s nasty comments so I could yell at him...yeah, exciting life, I know! Anyway...Fame...these kids could sing AND dance...and quite well.  I mean, even the kids that got the boot right away tromped all over the American Idol kids.  Amazing...well, this week was the finale...and, I must red-faced admit...I actually voted!  I wanted Harlemm all the way.  Don\'t hate, Shannon, cause you have it going on too...but, please, who remembers the last ASIAN short, OLDER (30-something), male that was a POP icon?!  He seemed to have such a sweet heart...Shannon was beautiful, could sing, and dance like a mother, but she is the white female icon we see everywhere.  So, I was allowed five calls, I placed five calls, for Harlemm...and he won!  My faith in voting has been restored!  Mom saw nose ring this week...can we say maniacal?  I could not believe how explosive she was...then she started in on me going to the club...whew, get a life!  I hated to remind her I\'m an adult and her manipulation techniques no longer work on me...oh, well, good to see I can still get her blood boiling!  Look out, the number one black sheep of the family is taking back the throne!  '),(19364,518116,'22,July,2003','2003-07-22 00:00:00','   don\'t you wish everyone took the time to update their blog?  I really wish people wouldn\'t wait almost a whole MONTH to post...especially, since I know they read mine...oh where, oh where is my purple panda?  '),(19365,518116,'21,July,2003','2003-07-21 00:00:00','   ahhh...back to the life of honesty.  I was really starting to sweat all these lies I had to tell people.  But, it was worth it...one of the black sheep ditched us, so then there were three...and only one didn\'t know about the sex toy party.  So, first, one of my other friends showed up...we started talking and she was like \"Who is that?\" oh, that\'s my friend so and so... \"Okaaaay\"...no clue about what was about to happen.  Then SHE shows up...so black sheep is like \"Who is that?\"  Oh, that\'s Brenda... \"How many people are coming?  I guess I didn\'t realize friends...(her eyes see the \"Love Swing\" Brenda is carrying) OH MY GOD!  I almost felt sorry for her...she broke out into hives...halarious!  She stayed for most of it...pay backs are hell, I\'m not scared!  Found out JP and hubby are separated...hmmm...I could be mean, but I\'ll refrain...  I got my nose done!  I can\'t believe it, I got so nervous!  I tried to call hubby half a dozen times to see how pissed he would be if I had it done...but I couldn\'t get ahold of him...so I did it anyway...I\'m thinking it\'s cute but it takes getting used to!  And, I constantly forget it is there, so I go to itch my nose and grab it...ouch!  Other than that though, it really doesn\'t hurt...and surprisingly it did hurt less than my belly button!  '),(19366,518116,'17,July,2003','2003-07-17 00:00:00','   whoops!  I guess it has been a little while...I have been expending all of my energy telling half-truths to half the people in my life... My friend is moving so we are having a black sheep party on friday.  She and I thought it would be halarious to have a sex toy party...the other two did not want to, so we are keeping this a secret from them...lie 1.  Then, my hubby wasn\'t real thrilled so I waited to tell him...lie 2.  I am having hats made up for us...lie 3.  I am in charge of getting her cake at work and getting her to the conference room...lie 3.  The issue is, I talk to all these people constantly throughout the day...do you know how hard it is to remember who knows how much?!  I\'m getting my nose pierced on Friday...lie 4.  Then, this girl\'s neighbor calls and says she has a surprise block party planned on Friday.  She basically wants me to cancel our plans so hers can go through...ummm, no.  So I couldn\'t handle the stress anymore and just told my friend...she laughed at what a dumbass her neighbor is for planning a surprise going away party...did she not realize she would be making plans and have lots of things to accomplish?!  Now, the other two are dueling what we are going to be doing...I want to go to the piano bar...no I want to go dancing...well if we go there, I\'m not going...can anyone say jr high?  So, to keep the peace (and knowing if they are acting like this they are REALLY going to be pissed about the sex toy party!!!) I offered to give up the tattoo/piercing parlor...no one will hear of it.  Shit.  I\'m starting to freak out...do I really want a nose ring...how bad will this thing hurt?  you know, normal piercing reservations...oh yeah, will hubby kick me out?  Anyway...watch out KZOO...the sheep are on the loose!  '),(19367,518116,'02,July,2003','2003-07-02 00:00:00','   ahhh...so little to say, so much room to say it in.  where to begin...my dvd player is broken...which sucks.  I have worked very hard at buying dvd\'s without hubby finding out, and now I can\'t watch any!  I really need to go buy a new one, but, that would require me to go and pick one out...I haven\'t decided what features I would like.  Right now, we have a disk carousel...which I was originally very excited about...other than my movie obsession, I have a music obsession...which reminds, me...I need to turn on some music, hold on.  Okay, I\'m back.  Anyway, the disk carousel doesn\'t automatically bounce to the next disc once one is done.  That sucks.  So, I am not sure I really need that feature since it doesn\'t do what I want anyway...then, after I do buy one, I have to try to figure out the wiring again...now, hopefully, this will be very easy.  Hopefully, the same plugs that I used on the broken one will just need to be put into a new one.  My hubby fakes ignorance on this crap, so I get to figure it out...boy is he in trouble when I croak!  Well...I made it to my friends party on the 21st.  It was too funny.  And, it proved to them that I seriously always run into someone I know when I go out.  At her party was this kid I went to middle school with...and we remembered each other.  I knew I went to middle school with him (4th or 5th grade)...he remembered where I live!  I guess we rode the same bus, which I didn\'t recall.  Once again proving this town is way too small!  Last weekend was another wedding.  This time, one of hubby\'s roommates from college.  It was pretty fun.  Met some of the rednecks from up north who cracked me up...  Well...the layoffs have started.  My boss still thinks I will be okay through at least the summer.  It is going to be so hard.  The goodbyes have already started...I just get too personally attached.  Some of these guys I have worked with for almost two years, I truly care about some of them.  A lot of my friends are aliens, so if they lose their jobs, they have 30 days to find a new job or else they have to leave the country...the man at work.  It\'s just sad to watch, and I worry about these guys.  Some of them are holding that rod awfully tight...I pray they can deal when it happens...  Boy, that ended on a depressing note, but it\'ll be okay.  '),(19368,518116,'25,August,2003','2003-08-25 00:00:00','   well...she came.  It wasn\'t too weird.  We basically chased toddlers around the whole time...I\'m sure the teeny bopper at Formula K was glad to see us go.  She explained her marital status...they separated.  The next day, he was having sex with a chick from work.  A month later he wanted her back...they started talking about things, then found out he got this girl pregnant.  What a load of crap!  She is having a hard time deciding what to do...I don\'t envy her!  I told her to follow her heart/head about what was the best for her and her daughter.  How tough...split up your household or deal with his stupidity for the rest of your life...what a choice.  My older sis was back in town...I was weird...what else is new?  I don\'t know what it is about all of us being together, I can\'t deal.  It\'s like I\'m back in middle school...I have good relationships, for the most part, with each on an individual level...but when we are together I feel like a deer in headlights.  They started in on me about my memory...mind you, this is an EXTREMELY touchy subject for me.  Since B, it has gotten progressively worse...this year being an all-time low...dr says it is in relation to the PCOS...I tried to explain to them, get them to back off, but, as is typical, this made the vultures more aggressive.  Thankfully, I stopped...which leads them to stop.  Why would it be fun if the injured animal doesn\'t fight back?  Right?   But Sunday, what a doozy of a day...no good reason, I went after hubby about his family...I mean, I will never get mine to change, so why not attack his?  Very reasonable if you ask me...(Please note the sarcasm)...  I basically stayed in bed all day.  I got up to call my friend, make some lunch, and relocate to the couch so B could watch a movie...it was about mid-afternoon before I could think about what I was so gloomy for...  Okay...so why am I the only abnormal one out of my sisters?  Why was I blessed with the scoliosis, bursitis, PCOS?  Does that make sense?  Not to me.  Not that I would wish any of them on my sisters, because I would never in a million years.  But why do I have to have it?  I guess this is what my online friends refer to as the solidarity of the disease...no one can understand it unless they have it.  I would give a lot if one morning, I could wake up from a normal nights sleep and feel awesome...okay, not even awesome, great.  It is so hard to have little to no sleep, get up, be in pain, go throughout your day trying to act normal, come home and take care of B, try to relax, go to bed and start the cycle all over.    Meanwhile, I have been instructed to lose weight...ummm, how would you like me to do that?  Okay...realistically, I mess up.  I overindulge, I will admit it.  But, I really do not eat much the rest of the time...it\'s almost funny.  I tend to eat the most when I am in front of people...I figure, they think I\'m fat because I eat too much, so, I\'ll prove them right and overeat!  Like at my parents, I had steak (I\'m usually a vegetarian!), a ton of veggies, a piece of bread, a stuffed tomato with rice, and followed it with a huge piece of cake and ice cream...oh yeah, I had chips and dip beforehand.  Please note, like I said, I don\'t eat meat...I stay away from carbs, I don\'t eat tortilla chips...and I don\'t particularly like ice cream.  But, I gave everyone else peace of mind by overeating and letting them continue in their thoughts that if I didn\'t eat so much I would actually lose weight.  Yesterday, it struck me.  Why do I try?  Why am I at Weight Watchers?  Everyone who knows about PCOS says the same thing,  it isn\'t your fault, you can\'t help the extra weight, nothing you do will help ...but what am I supposed to do?  Give up?  The GP said I should be thankful I\'ve been trying...who knows what size I would be if I hadn\'t...um,  THANKFUL ?!  Yeah, I\'m really thankful I\'m only 50+ instead of 100+ lbs overweight.  But, it does scare me...what if I do stop trying so hard?  Will I gain another ton?  Yesterday was a really bad low, I have to admit.  I was two seconds away from running to the store to buy out all the crap food I haven\'t been eating...the only thing that stopped me was my own stench from not showering.  But, seriously, do I just give up the hope that I will ever be a normal size?  Now that hubby\'s dad and brother both have boats, do I feign fear of water?  Yeah...life is good.  I won\'t renew WW after I\'m done.  It\'s a waste of money.  And as much as I hate to admit defeat, I hate wasting money more.  One of my friends wants to do yoga, so I\'ll do that...better use of money.  I guess I should just resolve to be as healthy as possible.  This isn\'t really in my control...the getting pregnant, losing my memory, weight gain, etc.  But I can help myself...I could really help myself by not being obsessed.  But it\'s hard not to...  '),(19369,518116,'13,August,2003','2003-08-13 00:00:00','   oh, oh...I forgot!  B and I had to run around to purchase patio furniture...so, we are at one store, and we run smack dab into this guy I worked with...a guy that used to tell our friends that if my hubby (then bfriend) ever broke up with me he would be all over me...mind you, he roomed with my hubby for a while as well!  It was a tangled web...he was also friends with one of hubby\'s friends...anyway!  SO, he and I were talking about life, we had seen his pic in the paper for getting married, etc.  So, B, not getting ample attention, decides to stick his hand down my shirt!  I tried pulling it out, he grabbed onto my shirt and started pulling!!!  Mind you, my shirt was a snap one all the way down!  This guy didn\'t know quite what to do, and, of course, B wouldn\'t stop!  Needless to say, 1) I\'m sure he got quite the eyeful and 2) I was thoroughly embarassed!  '),(19370,518116,'07,August,2003','2003-08-07 00:00:00','   ahhh...nothing much new in my world...cuts as Pfizer are still going on...it is so hard to say goodbye.  You spend most of your day with people and then, they\'re gone.  My roommate from college is coming.  Hubby is going to the races this weekend, so she and her daughter will be coming for a visit.  It should be interesting.  I think it has been about a year and a half since we\'ve seen each other...and, we don\'t really talk on the phone, just email occasionally.  I don\'t know.  It\'s weird.  So, hopefully, it will be just like old times plus the kids...  Sadly, this is all I have to talk about...  '),(19371,518116,'07,August,2003','2003-08-07 00:00:00','   Are You A Sex Goddess?        A brilliant bolt of lightning descends! SHAZAAM! The oracle has spoken!   The smoke clears to reveal that inside you is a divine being,   DEMETER, Goddess of the Fruitful Earth.   She is a deity who embraces all that man has to offer.  Sex for you is probably just as much of a spiritual experience as it is a physical one. You are an unconventional lover who doesn\'t feel the need to obey traditional gender roles or follow the norms set by society. Although you are a firm believer in experimentation, you never rush your sexual encounters. It\'s important for you to experience sex in a deep and profound way. When in the bedroom, you not only surrender your body to the experience, but you allow your mind and your spirit to join the action. Men find you to be a phenomenal lover, and they never fail to be completely elated and fulfilled when the night is over. Sex for you is much more than the sum of its parts. It\'s an experience that provides you and your partner with an unforgettable moment. For the men who are fortunate enough to share your bed, it\'s like a light streaming down from the heavens. Behold, the skies proclaim, here lies a goddess!    That\'s right, boys...you wish you could get a piece of this!  My hubby will be so proud...  '),(19372,518116,'29,September,2003','2003-09-29 00:00:00','   Ahhh....life....what a bitch.  Just kidding, just seems to be an extremely eventful time right now...so, where to begin?  I don\'t really have a long term memory for a lot of stuff, so I will probably forget something I wanted to blog about, oh well...I guess the stuff I remember is the most important, right?  First off, the best part...I finally got my hubby to break and we bought a baker\'s rack...yippee!  Do you remember when I mentioned I\'m a bit ocd?  Well, this has been constantly on my mind for the past few months...I\'ve looked at tons!  I wanted one for the kitchen to hold my microwave.  We had been using a microwave stand I have had since Chicago!  It looks awesome!  It makes the kitchen feel so open and big...and me so happy!  It\'s the little things in life:)  I saw one on sale at Meijer\'s, so I convinced him this was the best we would get...which is probably true.  So double whammy, I got the baker\'s rack AND it was on sale!  Now, for the other crap...Hubby and B went to his family\'s this weekend...his grandma burns wood for heat, so he went to help the family chop her wood.  So, I stayed home to work on the house.  My big project was refinishing our stairs.  These ugly things have been like this since February...yuck!  So, all week, I have been working on sanding so I could stain and varnish...so, I took half a day on Friday to finish up the sanding and get started.  My friend and I were supposed to go outlet shopping on Saturday, so I stayed up basically all night to get it all done.  Saturday morning...can\'t sleep, friend hasn\'t called.  So, I decide to be proactive and get painting.  I needed to prime and paint the underside of the stairs and paint my trim.  So, I pull out the primer and paint.  I start to prime.  Well, there are some spots on the stairs where the varnish and/or stain hadn\'t completely dried.  So, the primer was simply mixing...so, I decide to stop that and work on the trim.  I am painting away...at the top of the stairs is a window I cannot totally reach, so I bring up a stool.  I\'m painting away when I shift my weight and a leg of the stool slips off the step.  I catch myself and the stool by grabbing onto the window sill...however, I also was successful in knocking the FULL can of primer down the stairs...from the top it goes clunk, splash, thud!  I was in absolute shock...I stood there just looking down the staircase, unable to fathom what just happened!  A minute later, it dawns on me, this is real, this is bad...so, I run into our bedroom and call mommy.  Trying not to have a nervous breakdown, I explain what I just did...to which she asks, \"Is it oil or water based?\"  To which I reply, \"I don\'t know\"...well, if it was oil based, you would know because it really stinks.  So, you just have to get a couple buckets of water and wipe it up, keep mopping until you get it all.  I thought, I can do this...so I tredge down the stairs (reminiscent of Titanic when they are wading down the stairs being flooded with water) trying not to make the mess worse than it is...I start cleaning it up, it\'s a bitch, so I decide I better look at the can...you guessed it, it\'s OIL based!!!  I quickly call my mom, on the verge of hysterics...\"Oh my word...how did you not smell that?  Okay, get as much mopped up as possible as soon as possible and I\'ll be over\"  Umm, have I ever mentioned, I don\'t have a sense of smell?!  So, here I am grabbing all the old kitchen towels, trying to figure out how to get a two inch puddle up...it is going everywhere!  I get as much as possible mopped up and sit down in it all...to start balling.  My newly finished steps are now stark white again...and it had been all I could do to motivate myself to stain and not paint in the first place, and now here I was back at the option of painting...ugh!  My mom comes in, and I tell her, nevermind, this is never going to get cleaned up...well, she has come armed with a dozen chemicals, and we are going to find one that works...thank god and the inventor of mineral spirits!!!  We worked for about three hours scrubbing mineral spirits with steel wool onto those stairs...and got almost all of it up!  Unbelievable!!!  I so wish I had taken a before picture, no words could ever describe how bad it was...even my mom admitted yesterday that she had not thought there was hope!  Halarious?  Yes!  But also scary that my husband still allows me to do most of the improvements!!!  So, hubby and b went to see his family this weekend...and, when he got home, he handed me this card from his brother and sister-in-law addressed to he and I...when I opened and read it, it said \"Aunt S and Uncle S\"...oh, god...she\'s pregnant.  Of course, hubby was eagle eyed to see how I would react...and, I think I earned an Oscar for my performance...I really want to be happy for them, and I think I am, but I also am dreading the next nine months.  She is three weeks along...how am I going to deal with this when all I want to do is cry?  I want to be sincere, not fake...what a jerk I am!  I feel like a real bitch, and, I probably am.  I was just going to mention to hubby about my wanting to discuss the possibility of taking chlomid...talking to the dr next time I go.  But I know he is now going to think it is because she is pregnant...and, it brings up the whole thing about the fact that all of my maternity clothes, baby clothes, and baby equipment is up north...do I have to give her what is left of our baby and maternity stuff?  I don\'t know if I could deal with seeing my stuff being used other than by us...I really don\'t think I can handle watching them use all of my stuff, am I a bitch or what?!  But, now, i know my mother in laws question will be, why don\'t we just give it to M?  I knew I should have beat them up for this stuff the past two years!  I don\'t know...I did the good \"sister in law\" thing and got a card and gift to send up there...and, again, I really am happy for them, it just makes it that much more painful for me and it makes it difficult for me to do my usual \"coping\" and just ignore everything that isn\'t happening...and on top of it, I just feel extremely guilty because I know this all has nothing to do with me but I automatically am thinking of how this is really fucking with my head!    So, the long and short of it?  My life seems never to dull of drama...what I wouldn\'t give for a quiet, peaceful life...I\'m not as depressed as I would have been if this all happened a year or so ago...but, it is getting harder to find the humorous side of life...  '),(19373,518116,'24,September,2003','2003-09-24 00:00:00','   Okay...I have like five minutes before a staff meeting...but I have to piggy back off the last entry.  Newlyweds...last night...they are at a restaurant with Nick\'s brother and his wife.  They ask Jessica if she would like any of the buffalo wings that they are all eating.  She says, \"No, I don\'t eat buffalo\"...her sil says, \"what?\" Jessica says \"I don\'t like buffalo\"  Of course, they all CRACK up!  SIL \"They aren\'t actually made out of buffalo, they\'re chicken\" Of course, Jessica is royally confused so Nick\'s brother is trying to explain...finally he says, \"didn\'t you ever wonder how they were buffalo wings but buffalos don\'t actually have wings?!\" She says, \"I guess I never thought about that\"...halarious stuff! I am hoping she is cool and can watch this stuff and laugh hysterically at herself...otherwise, she is probably humiliated...I say, hey, we are all freaks, you are just getting paid bucco bucks to air yours on national television!  Anyway...not much else is new.  Work is super frustrating, our computer systems are supposed to be merging but no one can make up their minds to press the on switch...meanwhile, the rest of us are going nuts trying to get everything straightened out...  I get to refinish our stairwell this weekend...whohoo!  Hubby and b are going to his parents to help chop wood for grandma.  So, it will give me the time alone I need to do this project...and a weekend alone!  '),(19374,518116,'10,September,2003','2003-09-10 00:00:00','   Not much new...surprise, surprise, the specialist does want to see me...but doesn\'t have an appointment until January!  I guess I will take that as a good sign...nothing life threatening?!  I must address a show...NEWLYWEDS.  Can you say frickin\' halarious?  Ummm...Jessica thought chicken of the sea tuna might mean it is actually chicken, not tuna.  She makes a mess, which she is utterly clueless how to clean up, while she is making her half ass attempt, she asks Nick is she is doing a good job...then asks if they have a mop.  Nick looks at her like she is nuts and says \"How can you ask me if you are doing a good job but you don\'t know if we own a mop?\"  This girl has obviously been extremely catered to..its kind of sad because she has no idea how to take care of herself, let alone her house and husband.  Makes you wonder what she will do when her parents keel over...but, on a non-deep level, it makes me laugh so hard!  I don\'t know if anyone else is watching it...I looked it up on MTV and it was a little buried in their site, but, I hope people are...it is so funny!  Kind of like what the first season of the Osbourne\'s was...  Other than that...my bf stole my sweats and shirt she borrowed when she was at my house...I should have known better than to give her name brand shit.  My CK sweats and a JCrew shirt...I\'m trying not to be pissed, which would be a lot easier if she wasn\'t denying that she had them!  She says she left them in my closet, well, I scoured that puppy and was not shocked not to find them...ugh!  I keep trying to tell myself they are just clothes, they don\'t mean anything, but at the same time, do I need to lock everything up when she comes over?!  I think I\'m a little ocd...let this serve as an example.  '),(19375,518116,'03,September,2003','2003-09-03 00:00:00','   ahhh...I\'m exhausted...but I have visited all my frequent internet hang outs...so I\'m left with my blog.  My \"bf\" showed up on my doorstep on Sunday morning.  Her grandma passed away unexpectedly and she had come to see her before she died.  She did not want to be there when hospice took the body out.  Her family life is messed up to say the least.  Her grandma had taken care of her when she was really little and they have been close.  She hung out all day Sunday, spent the night, and was there most of the day Monday.  The funeral is today but I\'m not sure if I will hear from her again before she leaves tomorrow.  I always forget how high maintenance she is until she is around...she really is like having another kid in the house to take care of...I feel guilty saying that, but it is true!  I think that is why I am so exhausted...  B starts preschool next week...hubby and I had an orientation meeting last night.  As an example of how tired and mentally foggy I am, a lady introduced herself and I offered my hand to shake...I had a small cup of juice in the other hand and dumped it on myself.  It was like I could not focus on more than one thing at a time...I was about to leave because the first part was socializing (which I did not realize until it was almost done, I thought the meeting was just starting extremely late!) and it was so incredibly loud...I wanted to shut my eyes and lay down, I could not deal with it.  I still feel like I am in a fog...my allergies are acting up as well, so maybe the combo of everything is making me this way...I hope.  Anyway, we have an open house tomorrow so B can meet his teacher.  She did her intro and said this is her first \"real job\" since graduating from college...oh, boy.  I have mixed feelings about that, a little apprehensive.  We also realized none of the children from last year are in his class this year...they are in a different 4 year old class.  I think I am going to call to make sure he is in the right class!  I finally had my scoliosis xrays done...according to the report my curve has grown again.  Ugh...how much more can be wrong with my body?  I dropped the xrays and report off to my specialist yesterday, so I\'m anxious to hear what he has to say about it all.  '),(19376,518116,'02,September,2003','2003-09-02 00:00:00','   Okay, girly...I hope you read this!  I had the right email address, but it says your box is too full!  L~  I got to read your blog this morning...I don\'t know if this is your email or not!  I\'m hoping it is, otherwise, whoever is about to receive this will know I am psychotic...  Anyway...blog...no, I really don\'t think you are screwed up for leaving town and wanting to find yourself...I think that is awesome that you do that!  Besides the crap I was running from, my psycho-ex, my eating disorder, mom and dad, Heritage freaks, etc...that was one of the reasons I left dodge.  Although I didn\'t have the grand plan to find myself and figure some crazy shit out, I knew I was not going to survive going the way I was going.  And Scott and I had no idea what would happen between us.  I knew that I loved him, but I also broke up with him for a while...I couldn\'t deal with everything I had to deal with and be concerned for him at the same time...selfish?  Yeah, probably.  But at the same time, I knew in order to ever find out what made me tick, I had to get rid of the static that was diverting my attention from the work.    So, if you and Tony are \"meant to be\", I think you guys will come back together and  be better for the time spent apart.  This is the best time of your life to do this...once you get older and married and kids, you don\'t get to think about just you anymore.    I think that this is part of what happened to mom when we were little.  I think about what I went through by myself in Chicago...I don\'t think mom has ever had the liberty to do that.  She dealt with taking care of her mom and then her mom dying...then I think pretty quickly after that she found out she was preggo and her and dad got married.  Where was the time for her to find out what makes her happy?  Maybe that\'s why she was so angry all the time as we were growing up...she didn\'t have many choices in what she was doing...she went to a church that didn\'t condone any type of personal fulfillment ( I swear that looks wrong!) and she was a slave in her home...she had four girls, a hubby who expected her to do just about everything in the house...I couldn\'t deal!    SO, all that ramble to say...I hope that you find what you are looking for...I hope that you realize the beauty and potential that you posess.  No person is going to complete you, you do that for yourself.  Find what inspires you and grab on...choose to be happy, not tormented.    I love you~Sar  '),(19377,518116,'30,October,2003','2003-10-30 00:00:00','   did I mention I got new furniture?  Ahhh...so nice for everything to match, it makes me feel so adult!  Our couch reclines and our oversized chair is a sleeper...it is wonderful.  We have yet to receive the oversized ottoman, but i can be patient!  Bob is so picking Kelly Jo!  Meredith got booted last night and that is who everyone THOUGHT was going to win...I think Mary and Estella are BEAUTIFUL...please do NOT get me wrong.  They seem oh so nice too, but he and Kelly Jo just have this compatibility that I think they should change it to the Bob and Kelly show.  Too cute!  Ugh...so i missed RW again!  I was up until 9:50!!! And, I thought, I\'ll just close my eyes for a moment...when you are exhausted, you can\'t rationalize, because even as I was thinking that, I knew I was going to sleep right through, but I couldn\'t stop!  Of course, I can\'t find a re-run either...I still haven\'t seen the season finale of Newlyweds!!!  MTV kills me...now they have the rich girl show on...just to watch these two try to explain on how their self-absorption is not really self-absorption is ridiculous.  But, who am I to judge?  I was definitely NOT raised as a rich girl...my little sisters lived in a little more luxury than I did, but, not I...so, we\'ll see.  They keep showing this clip of Ali Hilfiger (sp???) saying to this \"modern day Ryan Philippe in Cruel Intentions\" (Oh, so philosophical, Ali!!!) that three very important people told her she is too giving...who are the three?  Well, two psychics and her friend Liz.  Hmmm...why are you seeing two psychics in one day?  What else is new?  No call from S yet...I think it is safe to say, I shouldn\'t be calling her my bf.  It seems so safe to have someone to hand the title to, and yet, I always hand it to people that are so NOT my bf...truth be told, I have a lot of great friends...it\'s unfair to label any ONE as my bf as they all act as a great friend should...not S.  In a way I feel guilty, I explained it to one of my other friends, and told her I feel like I am living high school all over...she said, a)I\'m nuts to allow someone be my friend who steals from me and then lies about it and b) I gave her a fair shot and she still obviously disrespects me so I should let her go...I guess if she never calls again that will be the easiest way to deal with this...but it is never that easy.  '),(19378,518116,'27,October,2003','2003-10-27 00:00:00','   ahhh...Mondays truly suck dookie.  I hate coming back to work...I must say the past month or so, I have had no motivation to work...fire me...please!  I have gotten to a point where I am tired of dealing with all the bullshit...either I have a job, or I don\'t...just let me know and put me out of my misery...latest rumor?  We\'ll get word this week that we are all done and have two weeks to clear out...or that we will all be put through one company and our paychecks will get axed.  Now, I\'m not that stingy, but...cutting my paycheck would be like bleeding a dead horse, you won\'t get much!  Can you leave my measly pittance alone?!  Okay...so I am sure some previously departed colleagues may gladly take my check...but I\'m thinking, if they cut my pay, I would prefer to get laid off and get unemployment for doing nothing rather than actually working the 40 hours and getting the same amount...ya know?  So...my frugal side has taken over...I\'m all eBay\'d out for the time...it was pretty funny to watch S bite his tongue when THREE boxes were delivered on Saturday...man, i wish the postman would not ring the doorbell...of course, I was on the phone with my mil so I couldn\'t run and grab them...oh, well...he seemed to buy the \"I\'ll have the $ when the bill comes\" theory...and it was also pretty funny to see his reaction as he is looking at all the knick-knacks in the living room...HA!  I made him put up a shelf in there so I could put pics back out...he just rolled his eyes...but I have to admit they are cute...now if my Pucca stuff could just arrive!  '),(19379,518116,'22,October,2003','2003-10-22 00:00:00','   Okay...so, I had to do one of those jr high things with your answers in it the other day...I was going to cheat and use that as a blog, but I forgot to copy it!  Plus, the opening paragraph was a tad hostile...calling people lame if they didn\'t do it correctly...whatever...so, of course, everyone did it the \"right way\" and I never got my answers back!  UGH!  People have to do a lot more than that to be lame in my book...plus, I was funny, so I can\'t \"recreate\" it...too bad.  Another cool story about my roommate...she met the Bachelor Andrew you know, the one that chose the cute girl Jen...she and our college buddy stood in line for four hours to meet him...and I guess Jen is even cuter and more petite in person!  Plus, I guess Kelly Jo from KALAMAZOO is hanging out with her mom in Wheaton right now...I know my sis was wondering where she was hiding out at...there you go!  And now, to let you know why I am about to be murdered...I have an obsession...eBay.  EVIL!  I can\'t stop!  First, I wanted to add a layer to my living room...so I decided to go with a collection I\'ve been wanting to start but know my hubby will not like once he is on to me...kissing.  HA!  So, I am scooping up these figurines and such of people kissing...it is so cute!  Well...then, I was looking under \"kissing\" on Ebay when I came across the most adorable animation... PUCCA!  She is a Korean cartoon and is ADORABLE!  So, I just won a bag of her and am hoping to have the faceplate in a day...I\'m in deep shit.  But I can\'t help it...you go find her and tell me if she isn\'t the cutest thing!  '),(19380,518116,'20,October,2003','2003-10-20 00:00:00','   Hey! did you know Matrix will be out in about 2 weeks!  I hope they show his naked ass again...yummy!  Halarious, yet true story...my roomie\'s hubby has put a personal ad in the yahoo pages...I died of laughter when she told me and it was much better when I saw it for myself!  What a freak!!!  He put this picture of himself in a lake...perhaps to hide the \"few extra pounds\" as he describes it.  I asked her if the fact he lives in his parents basement is a selling point!  Don\'t think he isn\'t creative...nope, he took his position as a manager of aquatics at the Y and twisted that to mean Executive Director or some sort of bullshit...hysterical!  And, no, he doesn\'t sugar coat the fact he had a wife and child yet got another woman pregnant...he admits it...yet another winning facet.  So, my roomie finally admitted she is dating one of our college friends.  She had casually mentioned him when she was here and I asked her about him...she thought I was psychic...lol.  I just told her I was good at reading into people and the fact she had mentioned him told me there was something to that...so, she is secretly dating him, it is complicated.  But he is a sweetie and I\'m hoping it works out for her...she deserves a break!  The twitch is gone...for now!  I can\'t tell you how awesome it is not to have it twitching!  That\'s all I have time for now...  '),(19381,518116,'14,October,2003','2003-10-14 00:00:00','   wow...a whole hour without an eye twitch!  A miracle!!!  I swear it\'s from work...I don\'t know why...perhaps it is my peers bugging me to help make a quilt for the devil\'s baby...and being the only one with balls and integrity to say no.  I mean, I hate hypocrites...and wouldn\'t I be one if I did this?  I think so!  But, my posse has a good idea, tell them I\'ve changed my mind and make one...with a BIG BLACK SHEEP!  HA!  Oh, I had it out with bf...well, if you could call it that.  I called her when I got home.  She was in the car with her hubby and kids, so I politely asked if now was a good time.  Well, we sat in silence until she asked why I was mad.  I told her I didn\'t understand her point to calling and bitching at me at work.  SHe flipped saying \"I was so kidding!\" I said, well, you said you were partly serious.  To which she went off about how I never call anymore, blah, blah blah.  I said it was hard to call and talk to someone who had no interest in what was going on with me and didn\'t have time to listen, just wanted to talk about what was going on with her.  She was offended, didn\'t expect her not to be.  She wanted me to explain my statements...so I recounted the million times I called and she would unleash her life story then turn around and say she had to go if I tried to talk.  She denies...big surprise.  So, I told her, you don\'t have a clue what is going on with my life, you don\'t ask, you aren\'t there when I try to talk...she thought I meant literally, so she told me to leave messages (umm, you don\'t have a machine!)...why must I deal with morons?!    Basically, it boils down to...she doesn\'t think I ever tried to talk, she isn\'t here so I can\'t expect her to be, and she needs me to blatantly ask for support...few issues, a)her hubby was in the background telling her what to say...you know what, dude? YOu don\'t have a clue what I\'M saying so shut the fuck up...b)no shit you aren\'t here, I am not asking you to physically be here, and if you don\'t get that, well why are we having this discussion?!, c)halariously ironic...the next day, I got a cosmo girl magazine (don\'t ask!) and they had this article/blurb about friends...something to the effect, true friends are there for each other without being asked...okay, if teenagers get that, how did you lose that vital piece of 411!  To top it off, at the end of that, to which I just said, you obviously don\'t get where I am coming from, and I\'m tired of talking about it...she asked, so what\'s been going on?  I laughed so hard I about wet myself...are you kidding me?!  I just told you, I don\'t tell you shit anymore because you aren\'t there...I have other people in my life who provide me with that...and yes, that means I am questioning why you are in my life, because, so far, all you do is use and steal... but I just found this funny...so, will I call her?  Why?  I don\'t call the people I do love and need...those who are there to support.  So before I call her lame ass, I think I will put back to those who have given to me...not stolen (sorry, I had to add that in there!)  Well, it was the day of sisters yesterday...big sis called...my bil is going to be a pastor!  Halarious!  I just find it so ironic...I don\'t know why, because, to those HCA jacks I wouldn\'t have been good enough to be related to a pastor!  Well, booyah!  Then, little sis called...she is in a black hole...and I feel really bad for her, not in a pity way, but because I really can identify with her situation...she misses her friends and wants to come home...I don\'t think I really helped, I tried to lead her through the logical reasons why she should NOT come...when you\'re emotional, you want to kill people like me...but, she has just seemed so happy...and who wouldn\'t be?  She got away, she is on the road to success.  I\'m so skeptical of friends...so I don\'t think anyone should relocate just for friends.  I know the friends that I have are my friends no matter where I live and offer that support... I don\'t really know what to say to her, I just don\'t think what I said should have been it.  I hope she gets past the initial hurt and takes a step back to analyze.  Kzoo is not right for right now...and friends that would pressure you to change your life for them, aren\'t real friends.  Friends are supposed to support 110%...unless life threatening, or illegal (sometimes)!  So, not the type of friends I would relocate for.  But there lies a major difference between sis and I...at the end of the day, I truly don\'t give much of a rip what people think of me or what I do, she does.  Of course, I\'m nuts anyway, so take my opinion with a grain of salt!  '),(19382,518116,'08,October,2003','2003-10-08 00:00:00','   Why, you ask, am I posting twice in the same week?  Because I am so not in the mood to work and I am stuck here until 2pm...ugh.  SO, I have read my email, my community board, a couple other sites, and now I\'m here...I really don\'t even feel like talking, so I haven\'t replied yet to those emails (I WILL LISA!!!) nor did I post to the boards...I\'m tired of talk.  I had the strangest, freakiest, nasty dream last night...I was a lesbian, which isn\'t exactly the nasty thing, freaky most definitely...but I was a lesbo with my bf, who is in questionable status.  And, we were trying to hide it from all of these Heritage people...for, duh, obvious reasons.  But then mr x came and I was fighting with bf so I wanted to be with him...I think something I ate right before bed does not agree with me!!!  So, to my dream last night, ew, don\'t come back!  Did I mention my roommate decided to divorce her hubby?  Good...he\'s a loser... Did I mention I have missed the past TWO weeks of Real World?  AND Newlyweds?    Thankfully, I have caught Newlyweds on re-runs...but where are the RW re-runs?  I really needed to see Leah get yelled at and CT go postal.  Don\'t get me wrong, I think Leah is gorgeous...I even forgive her that she is blonde.  I wonder what size she is...she looks larger than Mallory and Christina, but those two look like twigs.  Anyway...I went to Leah\'s site...wow...so hot!  And, she doesn\'t appear to goup herself up...I think she makes me laugh because she is so like my bf who thinks, I can walk into Gucci and give them shit because if I wanted that bag I would have it but choosing not to because I still like my model from last year better than this crappy version...watching Leah makes me appreciate myself...I\'m sorry, it is hard to deal with people who like to be loud and obnoxious 24/7 and I DO think I should get a medal for hanging in there!  And CT?  He is so weird it is halarious!  Seriously, he is what I envision most caring guys to be like...trying to figure out the female population turns them neurotic...halarious!  But, we\'ll see if he gets kicked off because of his actions...they looked borderline on the preview and since no one believes in stepping in and stopping this shit before it goes down, so these people can be humiliated on public tv...whatever.  I mean, there have been very few times, but they have been there, where I want to take someone out back and beat them, to death...so, it happens to the best of us.  And, have you seen the crap he has put up with?  I would have killed someone, or at least verbally attacked them, a LONG time ago!  I give him credit for hanging in there...his roommates have been really mean and nasty to him...poor CT...  Wow...I really need to get a handle on watching reality tv...next thing you know Charla and I will be having coffee!  LOL...did I mention THAT...oh, Lord, that girl should feel like crap.  Paradise hotel...Keith decides to keep Charla and Dave.  All four (Tara is the other) must swap roommates for the vote.   So Charla and Keith win...and are given 250K and told they can either walk away with the check or split it with their \"original\" partner...they have 30min to decide.  Toni (UGH) hunts down Charla and convinces her NOT to give Dave the money...he doesn\'t deserve it, no one out there thinks he does, blah, blah, blah...didn\'t she watch the episodes?  Doesn\'t she get she would be doing herself a favor by SHUTTING UP?!  Anyway, Keith decides to give Tara half (I love him so much!) and Charla KEEPS IT ALL!  Last time I checked, you needed him and vice versa not to get kicked off!  Greedy, greedy!  At least she made up my mind for me...Tara I like better.  When Keith was telling her he was giving half of it, she was trying to convince him to keep it all for himself...what a friend.  Oh yeah, then Tara and Keith told everyone they were funding a trip for everyone to Vegas.  Now, I know this crap is pieced together...but there is no way I would have done that after the way they were treated...  Anyway...glad my life is so exciting I need to talk about TV!  There\'s junk to talk about, but I just don\'t wanna...maybe later...  '),(19383,518116,'06,October,2003','2003-10-06 00:00:00','   Wow...i love how everytime I log on, blogger looks different.  Makes me feel like I have been neglecting my blog!  What a funk I was in!  Not so bad anymore.  I think part of it has to do with shitty friends that won\'t leave me alone...and I must say, I am using the word friend very loosely!  My \"friend\" emailed me to see if I was having a Pampered Chef party anytime soon...I told her no, my mil can order if you want...well, lo and behold, she has started selling Mary Kay.  Why don\'t you just ball up and ask if I will have a party?  That\'s what kills me...you only want me around when you can use me, but you don\'t even have the guts to admit it.  Why don\'t I block her email???  Because I\'m too nice and am a glutton for punishment!  Anyway...this weekend was good.  Skipped work Friday (seeing a theme?)...had a gyn appointment, though...so it wasn\'t all fun and games.  Anyway, talked to him to get his take on clomid.  He thinks we should give the metformin 6months to kick in.  He seems to think it will make everything return to normal, including ovulation.  Then, during the exam, he seemed to think I was ovulating or was primed to ovulate!  Hope restored!  Which, scares me, quite frankly.  I already have my hopes back up and what happens when it doesn\'t happen?  Anyway, he reassured me we most likely would NOT have multiples...so, that relaxed hubby.  Actually, hubby has been really cool.  He is very open to actively trying at the beginning of the year...or at least talking about it.  Which is a start!  I thought I would have to fight him tooth and nail to get him to that point!  But, I think he knows how much I want this, and that ultimately, we all want it.  And, we both agree, we are in such a better place than when we were pregnant the first time.  And, he knew it wasn\'t all coming up just because his brother and his wife are pregnant!  Got a couple closets cleaned out...the house if finally semi-clean again...and we can park in the garage finally!  Such accomplishments!  '),(19384,518116,'30,November,2003','2003-11-30 00:00:00','   wow...I love having a long weekend.  I basically left at noon on Wednesday, with nothing to do, there is only so much surfing one can take...so the rest of the day Wednesday and Thursday, Friday, Saturday, and Sunday...wow...going to work tomorrow is really going to suck...lol.  So, I decided NOT to knit...to give myself a real break.  It has been so nice.  I have read two books...watched some movies...taken naps...gone shopping...went to see Matrix on the IMAX...took B to see Brother Bear...  What a wonderful break.  It reminds me of the good old days when I was depressed and could just wrap up in covers all day long and never get up.  I miss those days.  I couldn\'t sleep at night, so I would sleep for about four hours and then get up and watch movies or read.  Okay, I did start knitting today, but other than that I have been good!  Do you ever have a friendship you think is really great and then you find out that it isn\'t?  Well, that happened to me.  My boss and I, I thought, were good friends.  I mean, he is really cool and really flexible with me and we joke around with each other, the guy gripes to me about his problems all the time and asks for advice...normal friend stuff.  Well, he has been working on making wood working stuff to sell at a craft show.  He had reserved a booth a while ago.  I started knitting these bags/purses and he asked me if I wanted to put them in his booth and sell them.  I didn\'t take him serious at first but then everyone started to tell me I should sell them...I figured, what do I have to lose?  So, I started making these bags to sell...last week, he says to me,  so what do I get out of all of this?   Understand, I thought he was kidding!  So I said, a thank you and a smile.  He said,  How about 10%?   I must have given him a Barto look because then he said,  okay, 5%?   I told him I would have to do the figures...but I don\'t get it...now you are trying to make money off of me???  Whatever!  I was so shocked at first...then hurt...now pissed.  What a jerk.  See, the thing is, he was supposed to sell his stuff at a store, well, he and I went out there and this guy that co-owned was really rude to him and basically told him no thanks.  Bill was so bummed about it, so on the way back to work, I told him to stop at this other store because I thought they sold wood working crap...and, sure enough, this guy is now interested in selling his stuff!  Ummm...you wouldn\'t have gotten that contact if it wasn\'t for me!  So what is MY kickback?  Then, in this same conversation, he wants to know if my friends that have the writing/graphic arts business will put together a business card for him...yeah right!  I told my friend and she cracked up!  I told her she should charge him at least double...ha!  So, it will be interesting to see if he brings it up tomorrow.  jackass.  Not much else is new...this week should be non-eventful...just trying to get things done in case I decide to do the craftshow...  '),(19385,518116,'25,November,2003','2003-11-25 00:00:00','   Okay...so can we say the holidays are a tad depressing?  My LORD!  I\'m okay...really...just in a funk.   Hey, ladies...I\'m just bummed so I need to spout off...don\'t mind me, LOL!   I think it is a yearly tradition for me to go into depression around B\'s birthday...what a great mom I am.  I just can\'t get past the desperate desire to have another baby...last week!  So, when Brennan\'s birthday comes, I feel like I am getting further away from the \"ideal\" timeline for two kids.  If it doesn\'t happen next year, I seriously don\'t know if we will still try.  And then, there is B.  At his party, my friends one year old was there tormenting him, so I though \"whew, maybe he will stop asking for a sibling because of this\"...ummm, wrong.  Last night we were on the couch reading a story (\"Polar Express\"), this child gets to choose the first present of the season...B says, I would say I want a baby sister.  I\'m crying just thinking about it...he has gone from saying he wants one, to asking why we don\'t...and the best I can come up with is \"because this is how God made our family\".  I\'m sure I must be so convincing.  My heart is breaking and I feel like his is too.   To top it off, my sister in law was at the party...she is 14 weeks along and of course that is all everyone wanted to talk about...I thought I had prepared myself for it...I didn\'t do a good job.  I couldn\'t even look at her when she came into the house.  Then my father in law said it was me and hubby\'s turn for another one, I just looked at him.  He said \"I don\'t even get a response\"...rather than burst into tears or rip his head off (or both! LOL) I just shook my head.  I think they forget they we, well I, have issues...   I would be angry...but I am too tired...I hate this, I hate PCOS, I hate it, hate it, hate it...I just want to curl up under the covers and hide.  I\'m tired of being tired, I\'m tired of not being able to sleep, I\'m tired of not having a normal body, I\'m tired of the aches and pains, I\'m tired of the hair, I\'m tired of the memory loss, I\'m tired of the acne, I\'m tired of the weight, I\'m tired of the meds, I\'m tired of people in my life forgetting...I\'m just really really exhausted from all of this...and not convinced it will get better.    to top it all off I am freezing my butt off...well I guess that could be a good thing, lol...it is snowing here and I am cold...has anyone else noticed an unusual sensitivity to cold?  Literally, I used to be able to go throughout winter without wearing a coat...now I am sitting at my desk and I have yet to take my coat off...I\'m cold!!!   I\'m sorry if this was depressing...I\'m fine, I\'ll be fine...I just have to try harder to find the silver lining and right now it is covered in snow...ha! Take care, dears~S  '),(19386,518116,'17,November,2003','2003-11-17 00:00:00','   ha!  Another!!!    IF YOU HAVE TIME.......... Subject: This one is way different...really!   Here we go again .... another one of these get to know someone things!  For those of you who don\'t remember the drill ... copy this to a new forward and change the answers to be about you and then send it on to all of your friends and family.  And don\'t forget to send it back to the person who sent it to you!  1.) Name: Sara 2.) Birth date: March 3 3.) Zodiac: Pisces 4.) Nationality: Italian & Polish 5.) What time is it?  7:53AM. 6) Are you sensitive:  Yes, sometimes too sensitive ------------WHO DID THESE THINGS LAST?------- 7.) Called you:  crap...my mother in law 8.) Saw you cry:  Hubby 9.) Made you cry:  a commercial on tv 10.) You went to the mall with:  Kate 11.) Yelled at you:  Samantha ----------Have you ever?--------- 12.) Taken a picture of yourself with a milk mustache and sent it to the milk company? Nope 13.) Said \"I love u\" and meant it: Yes 14.) Gotten into a fight with your Dog/Cat?: what??? 15.) Been to New York?: Yes 16.) Been to Florida?:  YES 17.) California? Not yet 18.) Hawaii?: Not yet 19.) Mexico?:  Not yet 20.) Canada?: Yes 21.) Danced Naked: can I plead the 5th??? 22.) Had a dream about something really crazy,then it happens the next day? yes 23.) Stalked someone?: hmm...I don\'t know if I would call it stalking...scoping out someone maybe! 24.) Had a mud bath?: No 25.) Wished you were the opposite sex? Not really  26.) What time is it now? 7:55AM ----------WhIcH wOuLd yOu pReFeR---------- 27.) apples or bananas?:  bananas 28.) Blue or red:  Red 29.) Backstreet boys or nsync?: I\'ll take just Justin 30.) Wal-Mart or Target? WalMart 31.) Spring or Fall?  Spring 32.) Santa or Rudolph?:  Santa 33.) what r u going to do after this?  I guess I\'ll have to work! 34.) was the last meal you ate?: lunch 35.) high school or college:  College 36.) Are you bored?: Sometimes...when i\'m at work, ALL the time 37.) How many of your buddies are online?: can\'t tell 38.) Last movie you saw?:  I just watched White Christmas on video! 39.) Last noise you heard?: my vent just went on above my head 40.) Last smell you smelled: Sweet Pea Lotion ------------ABOUT YOUR FRIENDS-------------- 41.) Laughs the weirdest? I wouldn\'t bust them out like that! 42.) Going to have the most kids: Rebecca 43.) Who have you known the longest: Jenny 44.) Loudest?:  Do I have to claim that right???  46.) Will fill this out and send it back the soonest? Lisa? 47)  Who is the moodiest? Samantha 48.) biggest player?: good question, I don\'t know 49.) Who do you usually go to about all of your problems: Cura 50.) Things you like in a (Guy)/Girl?: Honestly, I have to think he is cute...and has to have a good sense of humor...feet can\'t be nasty either! 51.) Do you have a bf/gf?: yup...me and Eminem are tight! 52.) Do Dreams tell the future? no 53.) Has a dream ever told your future? no 54.) What book are you reading now?:  One of the Left Behind books...I think I already read it but forgot it...so I\'m refreshing my memory so I can read the latest one! 55.) Nicknames: Mommy, barto, sarbear 56.) Hair Color: black 57.) Height: 5\'6\" ???? 58.) Pets: a dog and a fish 59.) Siblings: 3 sisters 60.) Been so drunk you didn\'t remember that you were drunk:Nope... 61.) Taken any illegal substances: Nope 62.) Gone out in public in your pajamas?: Yes...gotta love college! 63.) Missed school b/c it was raining? yes...gotta love college! 64.) Set any body part on fire for amusement: ummm...NO! 65.) Kept a secret from everyone: yes 66.) Wanted to hook up with a friend? yes 67.) Cried during a movie?: Yes 68.) Had a crush on a teacher:  yes 69.) Ever thought an animated character was cute? well...yes!  Pucca rocks!  Look her up!!! 70.) Ever at anytime owned New Kids On The Block stuff? My parents wouldn\'t let me! 71.) Ever planned your day around TV Guide?: Never had a TV Guide! 72.) Ever prank called someone?: yes ---------FAVORITES-------------------- 73.) Shampoo: Aussie 74.) Day or night: night 75.) Summer/Winter: summer 76.) Lace or satin: Satin 77.) Cartoon Character: Pucca ------------IN THE LAST 2 WEEKS -------------- 78.) Cried: Yes 79.) Cut your hair: nope 80.) Worn a skirt: yes 81.) Been mean?: okay, I admit it! 82.) Been sarcastic: yes 83.) Talked to someone you have a crush on: does my hubby count? 84.) Hugged someone: Of course 85.) Wished on a star: on a star? no...but made some wishes 86.) Laughed until you cried: yes 87.) Played Truth or Dare: Nope 88.) Watched a sunrise/sunset?: does it count if you are on your way into work??? 89.) Spent time alone: yes --------RANDOM THINGS---------- 90.) Are you lonely?: sometimes 91.) Are you happy: Yes 92.) Are you talking to someone online: Not right now 93.) Superstitious:  nope 94.) What is your full name?: (Okay...I haven\'t officially changed it yet!  I LIKE my old name!!!) 95.) Who named you?: I would assume it would be my parents... 96.) When was the last time you showered? we don\'t have a shower...the last time I took a bath was last night... 97.) my pj\'s consist of: I don\'t think you want to know... 98.) what song are you listening to right now?  damn it...I forgot  my cd\'s again! 99.) If you could change your name: Goddess...ha! 100.) Sleep with a stuffed animal? nope 101.) what\'s the stupidest thing you have ever done?: called my ex 102.) What will your first son\'s name be?:  umm...here, I\'m psychic...Brennan Scott 103.) Favorite drink: Water 104.) scary or happy movies: Happy 105.) On the phone or in person?: In person 106.) Lust or Love?: Love  107.) Do you want your friends to write back: yes 08.) Where do you want to go on your honeymoon?:  ummm, we went to Chicago. 109.) How\'s the weather right now?:  foggy 110.) What did you do last night? well...I painted, I knitted, watched some tv, and NONE OF YOUR BUSINESS...ha! 111.) How are you today? tired 112.) How do you eat an Oreo?: with my mouth 113.) What makes you happy?:  my family and friends 114.) What makes you mad?:  friends that suck...hypocrites...people that are full of themselves...and really, these don\'t make me mad, just aggravate me! 15.) What do you want to be when you grow up?:  good question! 116.) What are your future goals? have a baby, get a Masters, travel 117.) My fave music: just about anything 18.) Favorite Food? Pasta 119.) Fast or slow?: Fast 120.) Are you too shy to ask someone out?: Depends on the mood 121) Fave clothes store/brand?:  Anything CHEAP 122.) what time is it?  8:13AM...okay!  Someone came to ask me for something!!! 123.) Favorite board game? Candyland? 124.) Favorite smells? My hubby with Tommy on...and Brennan after a bath 25.) Comfort foods? pasta and chocolate 126,) Favorite sounds: laughter 127.) Worst feeling in the world?: not knowing 128.) What is the first thing you think of when you wake up: crap...what am I going to wear? 129.) Favorite place to eat? I don\'t really have one 130.) Future child\'s name?  I\'m not telling...you might use it!  HA! 131.) \"If I had a lot of money, I would\": pay off debt, buy a house, and travel 132.) Do you drive fast? sometimes 133.) Storms-Cool or Scary? depends...if I\'m inside, cool 134.) What was your first car? \'87 Ford Tempo 136.) Do you eat the stems of Broccoli? Yummy! 137.) If you could dye your hair any color: I am happy with my color 138.) How many different cities/towns have you lived in? Deerborn, Portage, Mattawan, Elmhurst, Kalamazoo 139.) Glass half empty or half full? half full 140.) Favorite sport to watch? tennis, hockey, racing 141.)Tell me about the person who sent this to you: Paula is a friend that I don\'t get to see anymore but enjoy her friendship via the world wide web! 142.) What is under your bed? dog 143.) What time is it now: 8:18AM  '),(19387,518116,'04,November,2003','2003-11-04 00:00:00','   Aha!  I got another one...so here they are, not as humorous, but, oh well!   Date: 11/4/03  Get to know your friends better. Make sure you read the instructions at the bottom and have fun.   1. What time is it? 8:34  2. Name: Sara  3. On birth certificate:   4. Nickname: Sar; Sar-bear; Barto  5. Number of candles on your last birthday cake should have been: 26  6. Tattoos: Nope...can\'t convince myself I won\'t faint from the needle!  LOL...Lots of piercings though  7: Pets: We have a dog named Spike and a fish named Joey  8. Natural Hair color: Black  9. What friend will have the most entertaining answers? No, Me!  10. Eye color? Brown  11. How much do you love your job on a Scale of 1 to 10? Right now?  I\'ll be generous and give a 3...before the stress of the buyout?  I LOVED my job...a 8 or 9!  12. Birth Place: Deerborn, Michigan  13. Current Residence: ...if you watch \"The Bachelor\", the place Kelly Jo calls home!  14. Favorite food: chocolate and pasta   15. Been to Africa? No  16. Been toilet papering? A very long time ago  17. Love someone so much it made you cry? Yes   18. Been in a car accident? No, but I saw my boyfriend at the time go airborn and roll three times.  I still have nightmares about it...I really thought he was dead.   19. Croutons or bacon bits? I don\'t eat meat....and I try not to eat too many carbs (hard when pasta is listed as my favorite!)...but I do like croutons, I just don\'t eat them...  20. Favorite day of the week: Friday and Saturday...  21. Favorite word or phrase: I love you.  22. Favorite Restaurant: Tough one...right now, I would say a bar by our house called \"McGonigles\".  If I was back in Chicago, it would be \"Portillo\'s\", \"Maggiano\'s\", or \"Giordano\'s\"...  23. Favorite flower: I love sunflowers and roses  24. Favorite sport to play: Hmmm...my hand to eye coordination is not so hot.  So, to play a sport is mostly just an embarassment.  I would have to say, anything I can play with Brennan because then I don\'t feel so ridiculous!  25. Favorite drink: I don\'t like to drink...and I mostly stick to water...  26. Favorite Ice cream: Hudsonville\'s French Silk  27. Favorite fast food restaurant: at the moment, Wendy\'s  28. What color is your bedroom carpet: we don\'t have any right now!  29. How many times did you fail your drivers test: NONE, I swear!   30. Besides this one, from whom did get your last email? lovely spam  31. Which store would you choose to max out your credit card? Well, right now, I\'m doing so pretty good damage with eBay...  32. What do you do most often when you are bored: knit, scrapbook, watch movies, read, sleep,   33. Most annoying thing people ask/tell me: If you ate less...   34. Bedtime: well, it depends on what is on tv...how sad!  Usually around 11-11:30pm  35. Who will respond to this email the quickest?   36. Who is the person that is least likely to respond?   37. Favorite TV shows: Alias, Newlyweds, Friends  39. Ford or Chevy: Neither   40. Time you finished this e-mail: 8:43am...I\'m slow!  Return Directions: Now, here\'s what you\'re supposed to do... Copy (not forward) this entire email and paste it into a new email that you will send. Change all of the answers so that they apply to you.   Then send this to a whole bunch of people you know INCLUDING the person who sent it to you. The theory is that you will learn a lot of little known facts about those who know you. Remember to send it back to the person who sent it to you.    Okay...now the funny part...I\'ll see if any of my friends notice...I accidentally put on the \"name on birth certificate\" my married name!  Whoops!!!  BTW...sister is not longer MIA...she actually called shortly after I posted the last one...lol.  '),(19388,518116,'03,November,2003','2003-11-03 00:00:00','   ahhh...what a lovely weekend.  Halloween was nothing special.  B was Woody and we dressed Spike up as Buzz Lightyear...it was pretty funny.  Spike did not like the costume and came and hid his head in my arm, too cute!    We watched my parents dog...he pooped and pee\'d throughout the main floor, I was ready to kill him by the time it was time to send him home...  My in laws came...boy was my mil in an exceptionally bitchy mood.  Oh, this was funny too...they are all telling me (Hubby\'s aunt, uncle, and grandma also came) how his sil has all these cravings and is oh so moody because of the pregnancy...ummm, hello?! She is only seven weeks along...HA!  Most people do not even KNOW they are pregnant at that time...except for those who are majorly trying...she also has gained four pounds.  Boy, I sure hope she continues to pack them on...to that, I would say KARMA.  But, anyways, I was so trying to keep my mouth shut, but I must have gotten a look.  My fil says, I keep trying to tell them this is all in their head, that it is too soon.  I just said, EXACTLY!  I can\'t wait to see if they come to B\'s party and if she is bitching about being pregnant then...ugh.  annoying...  anyway...my sis has gone mia again...oh, where are you?  I need to know about your couch so we can get it out of the garage and park inside again...BEFORE the snow!  '),(19389,518116,'28,December,2003','2003-12-28 00:00:00','   I love long weekends...especially, when Christmas is stuck in the middle!  I feel pretty spoiled this Christmas...some earrings, a ring, comfy pjs...a computer...and other stuff I don\'t remember at the moment...oh, yeah, the new Dido and Britney cds!!!  yippee!  anyway...the psychic...when I first entered and sat down, she said...wow, your aura is beautiful...what do you say to that?  She says...I can see that you have a very complete life...everything is balanced...you have such love in you that your heart is bursting...you love to laugh and love living...but, your angels are so sad...they are saying that you used to talk to them and consult with them and you have shut the door on them...they are so sad and wish that you would recognize that you are older and more able to handle the communication...they wish that you would trust them again and let them help you along...  Then she did like a generic reading...I should have written this earlier, but I didn\'t, so I\'m just going to put what i remember and if it comes out, it comes out, otherwise I\'ve forgotten a lot...  I will be traveling a lot this next year...I don\'t need to worry, finances will not be an issue...I will meet a tall man with dark hair and green eyes that will be my soulmate.  I will be having a change of job in the next 2-3 weeks and the people in my present position will offer glowing references...I will have a change of residence in 2-3 weeks but not to worry, it is for the best, and, again, finances will not be an issue, I will be better off...someone will try to take advantage of me financially, so beware!  I also need to be aware for the next 3-4 days as someone will try to steal from me (She said this is not uncommon to get that in the reading around the holiday)...travelling kept coming up...financial security kept coming up...the man I am supposed to meet kept coming up...so I finally asked if he was different than my husband...she said, yes, he appears as a stranger but someone you have a deep connection to, so he could be someone you don\'t know but connect with or someone you used to know or don\'t know very well...but you and your husband will not be together in 6-8weeks...there is a lack of communication and there will be lies and betrayal.  he loves you very much but you are independent and strong and the relationship will be ended...I asked about kids...I will not have any more, there will be a possibility in 10years, but by then I will feel it is too late...she says I need to remain friends with samantha...that I am very loving and giving to her, that I tolerate her personality when most people do not...it is not readily seen to me what benefit I receive from our friendship, but that I need to continue with it...  And, that is pretty much the gist of what was said...if I remember more, i will write it down...no, I didn\'t get stolen from, and so far, I still have the same job...but we will see...  '),(19390,518116,'22,December,2003','2003-12-22 00:00:00','   ahhh....what a weekend...  Just the highlight...  I went to see a psychic for the first time on Saturday...I was out with the black sheep and we wanted to go see Lord of the Rings.  However, since all the tickets were sold out, we had to come up with a backup plan...well, our backup plan was a psychic.  We went to a place downtown that my friend had called about...they had been there before and they were told that there would be a person reading, but there wasn\'t.  SO, we headed out to a place in Richland...I wasn\'t sure if I was going to go through with it...as I\'ve said in previous blogs, past experiences have left me very freaked out about these kinds of things...I half believe they are true and I half believe they are a load of crap....but I agreed I would go along while the others had their reading and see what they thought and make up my mind at the last minute.  So, we go to this place and they have a big store as well...we arrive a half an hour early...and I\'m getting ill from the incense...why does incense do that to me?  Instantly sick...it\'s weird.  So, we are walking around, I\'m trying to mentally talk through the urge to vomit...and it is Cura\'s turn to go first.  Now, Cura and Shelly went to have a reading before...Kate had one a long time ago...so Cura comes down and it is Shelly\'s turn...I can tell Cura is emotionally moved.  So, I ask her what she was told...she was told by February she will be pregnant without having any assistance...Cura has tried so many fertility treatments...so, of course she was thrilled...she was told it would be a girl...Shelly comes down, she is not as moved...but she was told she was going to have either a new relationship or a new baby...to Shelly, that is bad news...but she also said she felt this woman was legit.  So, it\'s Kate\'s turn.  I decide, okay, what the heck, I\'ll do it...all the while, knowing what I would be told.  Mind you, we each are supposed to be getting 1/2hour tarot card readings...Kate is down in no more than 20 minutes.  Immediately, I could tell she was not happy, she breaks out in hives when she gets upset and she had hives and would not make eye contact with me...uh-oh...but it was my turn to go so I went...I will try to remember what she said, but it is in pieces, and not necessarily in the correct order...  So, after I was done, we left...I asked Kate what happened...she was also told she would be pregnant soon (she does NOT want to have babies) and that she and her husband would be splitting up...she said she got all bad news and was so pissed, she didn\'t care what the woman told her she just wanted it to be over...the weird thing, though, is that the woman said her aura was incomplete and asked her if she had a head injury or surgery on the back of her head...well, Kate was abused as a baby and the back of her head was fractured...weird...  Shelly was told she and gordon would stay together...if Kate and Steve and Scott and I are all breaking up how the hell is Shelly staying with her boyfriend???!!!  She was also told her and Cura were sisters in a previous life (I think she made that one up) and that she should go to law school and she would have her own firm...ummm, yeah...  Cura...was told Jeff needs to ask for more at work (he was just told he needs to be more assertive by his boss!!!)...Jeff is her perfect match....I can\'t remember what else...I think she was told her job would be changing soon...  '),(19391,518116,'17,December,2003','2003-12-17 00:00:00','   No cheating. Pick your dessert, then look to see what psychiatrists  think about you! After taking this dessert personality test, send this  e-mail on to others, but when you do, be sure to put YOUR choice of  dessert in the Subject box above.  ALSO, SEND IT TO THE PERSON WHO SENT IT TO YOU!   If you had your choice of those below, which would you choose?   1.. Angel food   2. Brownies   3. Lemon Meringue   4. Vanilla Cake with Chocolate Icing   5. Strawberry Short Cake   6. Chocolate on Chocolate   7. Ice Cream   8. Carrot Cake   NO ... You can\'t change your mind once you scroll down!  So think carefully, what your choice will be!   OK - Now that you\'ve made your choice, this is what research says  about you!   Angel food ... Sweet, loving, cuddly. You love all warm and fuzzy items.  A little nutty at times. Sometimes you need an ice cream cone at the  end of the day. Others perceive you as being childlike and immature at times.   Brownies ... You are adventurous, love new ideas, are a champion of  underdogs and a slayer of dragons. When tempers flare up, you whip  out your saber. You are always the oddball with a unique sense of  humor and direction. You tend to be very loyal.   Lemon Meringue ... Smooth, sexy, & articulate with your hands, you are  an excellent after-dinner speaker and a good teacher. But don\'t try to  walk and chew gum at the same time. A bit of a diva at times, but you  have many friends.   Vanilla with Chocolate Icing ... Fun-loving, sassy, humorous. Not very  grounded in life; very indecisive and lack motivation. Everyone enjoys  being around you, but you are a practical joker. Others should be  cautious in making you mad. However, you are a friend for life.   Strawberry Short Cake ... Romantic, warm, loving. You care about other  people and can be counted on in a pinch. You tend to melt. You can be  overly emotional and annoying at times.   Chocolate on Chocolate ... Sexy, always ready to give and receive.  Very creative, adventurous, ambitious, and passionate. You have a cold  exterior but are warm on the inside. Not afraid to take chances.  Will  not settle for anything average in life. Love to laugh.   Ice Cream... You like sports, whether it be baseball, football, basketball, or soccer. If you could, you would like to participate, but you enjoy watching sports. You don\'t like to give up the remote control.  You tend to be self-centered and high maintenance.   Carrot Cake ... You are a very fun loving person, who likes to laugh.  You  are fun to be with. People like to hang out with you. You are a very warm hearted person and a little quirky at times. You have many loyal friends.    Guess what I am???  Chocolate on Chocolate...very fitting if you ask me!  '),(19392,518116,'17,December,2003','2003-12-17 00:00:00','   Jingle bells, jingle bells, jingle all the way....   it\'s Christmas time!!!  I\'m so excited...well kind of...hubby does not have Christmas eve or Christmas off...BUT I get the 24th through 26th off, yippee!!!  Then, the following week, I have the 1st and 2nd off!!!  Plus, I have a couple half days in there because the sitter isn\'t working...woohoo!  I have just about all of my shopping done...just a few last minute details, which of course, include hubby.  He finally sat down last night to do a list for me, but it is all out of catalogs!!!  I told him I didn\'t know if that would work since it is so close, but we\'ll see...  Still no word from Sam...it is kind of funny actually...do I think she got the hint?  Maybe...I\'m the queen of no confrontation...so I\'m hoping I don\'t have to spell out the fact that I don\'t see how she can consider herself a friend, let a lone a best friend considering she STEALS and LIES...ugh.  This so takes me back to high school...it is annoying...  Speaking of which...B and I went to H\'s house this Sunday...this girl is nuts!  Her ex is getting married this coming year, so of course she is trying to get married BEFORE him...psycho.  To top it off, she is telling me all this drama she has in her life...like it is perfectly normal for an almost 30 year old to have so much drama, like she is back in jr high...she kept saying \"now it isn\'t any of my business, but I said...\"  Hello?!  Did you hear what you just said???  It ISN\'T any of your business, so stop getting in the middle of every blessed thing...it is so strange that adults still feed off of drama...maybe it was the amount of drama I had in my \"younger\" days (ha!)...I had more than my share and now, I just try to stay out of everything...if someone asks my opinion, well, of course, I have one and I will give it to you, but I try to bite my tongue until someone asks...  anyway...better get back to work...not really...a friend called...lol!  '),(19393,518116,'26,January,2004','2004-01-26 00:00:00','   ahhh...another monday morning...I hate Mondays...especially when woken up at 4:30am because your hubby is yelling at your toddler and trying to put your toddler back to sleep in a wet bed.  Don\'t get me wrong, I love my hubby, but does he ever get it that he is the adult and he allows a four year old to get his goat?  It astounds me...  Anyway...alot going on...not sure where to start...oh, yeah...that computer I got...sucked major dookie...so I\'m back with my old one...and ever more grateful for it...ha!    House projects: Hubby ripped down ceiling...I decided I wanted authentic tins.  Now, understand, spatial relations is not my thing.  So on eBay (I\'m addicted, I will admit!) someone was selling tins...32 (12\" each) for $105...cool!  That is cheap!  So I bought them and told the guy I would need enough for a 10x15 room.  Umm...that comes to $600 something!  And, it is closer to $700!!! HOLY SHIT!!!  I am so dead when hubby finds this out...so this month, overtime will be my middle name!!!  Britney:  Got tickets to see the concert...let\'s see if everyone follows through with paying me!  This is why I hate to be the one buying tickets...I can\'t stop thinking about it until everyone pays me...ugh...ocd at its finest!  Friend in TX: called me at work...after MONTHS...why, oh why did I have to be at my desk?!  She said she thought I was mad...I said I wasn\'t, I mean, that isn\'t a lie...I\'m not mad, just done.  So she talked at me for a while...as usual.  Supposedly she was in the hospital for 10 days...I just don\'t believe her anymore...she has presents for me she is sending.  I have to laugh now...I must have sounded like a bitch...I didn\'t get her anything, and I have no intention to...so I just said, you shouldn\'t have...oh that is nice...etc...whatever.  So, I finally told her my boss needed to talk to me and hung up...no promise to call but I have a feeling she will never get it, you know?  At least that gives me time to decide if I am just being psycho or if I really should forget about her...  Hubby:  Inmate hung himself in his cell...god, it is awful...b and I had brought him dinner and that is when it happened.  He feels responsible and guilty but he is so confused...I\'m praying he goes to see a counselor because there is only so much I can say...I don\'t know what he is going through and he is having a hard time opening up.  I don\'t think he really believes anything I say anyway...I just feel really bad for him...I hate to see him hurt like this...  Other than that...nothing...ha!  Our house is a disaster with the construction...I am praying by March it will all be done and be a faint nightmare...though, we may end up having to wait on the carpet if I keep spending all this money...I need a self-help group!!!  '),(19394,518116,'22,January,2004','2004-01-22 00:00:00','   Your IQ score is 133   This number is based on a scientific formula that compares how many questions you answered correctly on the Classic IQ Test relative to others.   Your Intellectual Type is Facts Curator. This means you are highly intelligent and have picked up an impressive and unique collection of facts and figures over the years. You\'ve got a remarkable vocabulary and exceptional math skills — which puts you in the same class as brainiacs like Bill Gates  '),(19395,518116,'26,February,2004','2004-02-26 00:00:00','   I so hate stereotypes...really...do people realize how much they can hurt?  I have a couple friends that have \"revealed\" to me lately that when they first met me they thought I was a snob and conceeded...so, I asked why.  I know enough from my past that I need answers otherwise I will drive myself crazy as to why...well, supposedly it is because of how quiet I was and the fact that I dress nice...what?!  First off, being quiet is a curse...I\'m shy.  People don\'t usually believe me, but it really is true.  Now, once I know you, no, I would not say that I am shy.  Reserved in what I reveal, yes, but shy, no.    Secondly, I dress nice?  What is that?!  At first glance, that may seem like a compliment and maybe it is in an off-color kind of way, but it is not really.  See, why should I be judged on what I wear?  I don\'t even think I particularly dress nicely.  Obviously, I suffer from self-consciousness so I wouldn\'t, but really...I know many other people who dress a LOT nicer than I do.  Anyway...plus, if they knew how I bought my clothes...you know those $3 shirts at Target that no one else wants...that\'s me!  I buy EVERYTHING on sale...literally, the most I have EVER spent on jeans was $30something dollars at the Gap.  The reason I spent that much was because a friend had sent me this awful pair as a present and I was exchanging them...so I could rationalize that it was not really my money I was wasting anyway.  Plus, name brands?  I have a couple...maybe two.  And I think both are name brand shoes!  One pair I just bought a couple months ago with my Christmas money (again, not really my money I\'m wasting)...I bought some DKNY slip-ons at Marshall\'s (the TJ Maxx type store) and I think they were $30-35 dollars...so, not a whole lot was spent!  It took me two weeks to finally pull them out of the bag and wear them because I was so horrified I had spent so much on shoes.  Obviously, I have spent money on real tennis shoes but that is because of my bursitis.  The last time I had new tennis shoes was in college so my Nikes have lasted 5 years at least!  My shoe closet is filled with PayLess Shoes...so, yes, I have a ton of shoes but they were bought at PayLess and the majority of THOSE were clearanced!!!  I don\'t know why this hurts my feelings...I guess because last time I checked I didn\'t realize it wasn\'t \"cool\" to look nice and I thought I am supposed to take care of myself.  Oh, yeah...this girl also mentioned that I am always getting my hair done...to which I replied, I have NO idea what you are talking about because the last haircut I got was LAST summer...I have dyed my OWN hair with $7 dye that you too can go to Rite Aid and buy!  Oh, yeah, plus I cut myself some nasty-ass bangs this fall...whatever!  I guess I don\'t see myself as materialistic...my parents, yeah...that was how they loved us...you are upset about something, here go buy a shirt...but both of these people are saying that this was their first impression of me...again, if materialistic means you want things to look nice, than, yeah, I\'ll admit to it...but I want things to look nice at a fraction of the cost...what a commercial!    Anyway...lol, my non-materialistic self got my bday present from hubby...it is a .50cttw princess cut anniversary band...but, in my defense, it was bought from eBay so we got a super great price...plus, I think it is every present for this entire year...lol...  '),(19396,518116,'25,February,2004','2004-02-25 00:00:00','   This is so cool...my friend sent this to me, and I would say it is pretty on for me...but then I started reading some others and not so much...but here it is (I\'m a Weeping Willow):   WHAT TREE DID YOU FALL FROM?  Find your birthday and then find your tree. This is really cool and  somewhat accurate, also in line with Celtic astrology.   Dec 23 to Jan 01 - Apple Tree  Jan 01 to Jan 11 - Fir Tree  Jan 12 to Jan 24 - Elm Tree  Jan 25 to Feb 03 - Cypress Tree  Feb 04 to Feb 08 - Poplar Tree  Feb 09 to Feb 18 - Cedar Tree  Feb 19 to Feb 28 - Pine Tree  Mar 01 to Mar 10 - Weeping Willow Tree  Mar 11 to Mar 20 - Lime Tree  Mar 21 (only) -Oak Tree  Mar 22 to Mar 31 - Hazelnut Tree  Apr 01 to Apr 10 - Rowan Tree  Apr 11 to Apr 20 - Maple Tree  Apr 21 to Apr 30 - Walnut Tree  May 01 to May 14 - Poplar Tree  May 15 to May 24 - Chestnut Tree  May 25 to Jun 03 - Ash Tree  Jun 04 to Jun 13 - Hornbeam Tree  Jun 14 to Jun 23 - Fig Tree  Jun 24 (only) - Birch Tree  Jun 25 to Jul 04 - Apple Tree  Jul 05 to Jul 14 - Fir Tree  Jul 15 to Jul 25 -Elm Tree  Jul 26 to Aug 04 - Cypress Tree  Aug 05 to Aug 13 - Poplar Tree  Aug 14 to Aug 23 - Cedar Tree  Aug 24 to Sep 02 - Pine Tree  Sep 03 to Sep 12 - Weeping Willow Tree  Sep 13 to Sep 22 - Lime Tree  Sep 23 (only) - Olive Tree  Sep 24 to Oct 03 - Hazelnut Tree  Oct 04 to Oct 13 - Rowan Tree  Oct 14 to Oct 23 - Maple Tree  Oct 24 to Nov 11 - Walnut Tree  Nov 12 to Nov 21 - Chestnut Tree  Nov 22 to Dec 01 - Ash Tree  Dec 02 to Dec 11 - Hornbeam Tree  Dec 12 to Dec 21 - Fig Tree  Dec 22 (only) - Beech Tree   YOUR TREE  ( in alphabetical order)  Apple Tree (the Love) -- quiet and shy at times, lots of charm, appeal,  and attraction, pleasant attitude, flirtatious smile, adventurous,  sensitive, loyal in love, wants to love and be loved, faithful and tender partner,  very generous, many talents, loves children, needs affectionate  partner.   Ash Tree (the Ambition) -- extremely attractive, vivacious, impulsive,  demanding, does not care for criticism, ambitious, intelligent,  talented, likes to play with fate, can be very egotistic, reliable, restless  lover, sometimes money rules over the heart, demands attention, needs love and  much emotional support.   Beech Tree (the Creative) -- has good taste, concerned about its looks,  materialistic, good organization of life and career, economical, good  leader, takes no unnecessary risks, reasonable, splendid lifetime  companion, keen on keeping fit (diets, sports, etc.).   Birch Tree (the inspiration) -- vivacious, attractive, elegant,  friendly, unpretentious, modest, does not like anything in excess, abhors the  vulgar, loves life in nature and in calm, not very passionate, full of  imagination, little ambition, creates a calm and content atmosphere.   Cedar Tree (the Confidence) -- of rare strength, knows how to adapt,  likes unexpected presents, of good health, not in the least shy, tends to  look down on others, self-confident, a great speaker, determined, often  impatient, likes to impress others, has many talents, industrious,  healthy optimism, waits for the one true love, able to make quick decisions.   Chestnut Tree (the Honesty) -- of unusual stature, impressive,  well-developed sense of justice, fun to be around, a planner, born  diplomat, can be irritated easily, sensitive of others feelings, hard  worker, sometimes acts superior, feels not understood at times,  fiercely family oriented, very loyal in love, physically fit.   Cypress Tree (the Faithfulness) -- strong, muscular, adaptable, takes  what life has to give but doesn\'t necessarily like it, strives to be  content, optimistic, wants to be financially independent, wants love and  affection, hates loneliness, passionate lover which cannot be satisfied, faithful,  quick-tempered at times, can be unruly and careless, loves to gain  knowledge, needs to be needed.   Elm Tree (the Noble-mindedness) -- pleasant shape, tasteful clothes,  modest demands, tends not to forgive mistakes, cheerful, likes to lead  but not to obey, honest and faithful partner, likes making decisions for  others, noble-minded, generous, good sense of humor, practical.   Fig Tree (the Sensibility) -- very strong minded, a bit self-willed,  honest, loyal, independent, hates contradiction or arguments, loves life and  friends, enjoys children and animals, a social butterfly, great sense  of humor, likes idleness and laziness after long demanding hours at work,  has artistic talent and great intelligence.   Fir tree (the Mysterious) -- extraordinary taste, handles stress poorly,  loves anything beautiful, can become depressed at times, stubborn,  tends to care for those close to them as well as helping strangers, rather  modest, hard worker, talented, unselfish, few sexual relationships, many  friends, doesn\'t want foes, very reliable.   Hazelnut Tree (the Extraordinary) -- charming, sense of humor, very  demanding but can also be very understanding, knows how to make a lasting  impression, active fighter for social causes and politics, popular,  quite moody, sexually oriented, honest, a perfectionist, has a precise sense  of judgment and expects complete fairness.   Hornbeam Tree (the Good Taste) -- of cool beauty, cares for its looks  and condition, good taste, is not egoistic, makes life as comfortable as  possible, leads a reasonable and disciplined life, looks for kindness  and acknowledgment in an emotional partner, dreams of unusual lovers, is  seldom happy with its feelings, mistrusts most people, is never sure of its  decisions, very conscientious.   Lime Tree (the Doubt) - intelligent, hard working, accepts what life  dishes out, but not before trying to change bad circumstances into good ones,  hates fighting and stress, enjoys getaway vacations, may appear tough, but  is actually soft and relenting, always willing to make sacrifices for  family and friends, has many talents but not always enough time to  use them, can become a complainer, great leadership qualities, is jealous  at times but extremely loyal.   Maple Tree (Independence of Mind) -- no ordinary person, full of  imagination and originality, shy and reserved, ambitious, proud, self-confident,  hungers for new experiences, sometimes nervous, has many complexities,  good memory, learns easily, complicated love life, wants to impress.   Oak Tree (the Brave) -- robust nature, courageous, strong, unrelenting,  independent, sensible, does not like change, keeps its feet on the  ground, person of action.   Olive Tree (the Wisdom) -- loves sun, warmth and kind feelings,  reasonable, balanced, avoids aggression and violence, tolerant, cheerful, calm,  well-developed sense of justice, sensitive, empathetic, free of jealousy,  loves to read and the company of sophisticated people.   Pine Tree (the Peacemaker) -- loves agreeable company, craves peace and  harmony, loves to help others, active imagination, likes to write  poetry, not fashion conscious, great compassion, friendly to all, falls  strongly in love but will leave if betrayed or lied to, emotionally soft, low self  esteem, needs affection and reassurance.   Poplar Tree (the Uncertainty) -- looks very decorative, talented, not  very self-confident, extremely courageous if necessary, needs goodwill and  pleasant surroundings, very choosy, often lonely, great animosity,  great artistic nature, good organizer, tends to lean toward philosophy,  reliable in any situation, takes partnership seriously.   Rowan Tree (the Sensitivity) -- full of charm, cheerful, gifted without  egoism, likes to draw attention, loves life, motion, unrest, and even  complications, is both dependent and independent, good taste, artistic,  passionate, emotional, good company, does not forgive.   Walnut Tree (the Passion) -- unrelenting, strange and full of contrasts,  often egotistic, aggressive, noble, broad horizon, unexpected reactions,  spontaneous, unlimited ambition, no flexibility, difficult and uncommon  partner, not always liked but often admired, ingenious strategist, very  jealous and passionate, no compromise.   Weeping Willow (the Melancholy) - likes to be stress free, loves family  life, full of hopes and dreams, attractive, very empathetic, loves  anything beautiful, musically inclined, loves to travel to exotic places,  restless, capricious, honest, can be influenced but is not easy to live with when  pressured, sometimes demanding, good intuition, suffers in love until  they find that one loyal, steadfast partner; loves to make others laugh.  '),(19397,518116,'12,February,2004','2004-02-12 00:00:00','   wow...no wonder it seems like a long time, it has been a long time!  I am officially fighting depression again...I have been so happy on met, it seems to help with a lot of things.  The hair has been falling out more slowly, the extra hair has been growing not so fast, and the mood swings seem to have evened out.  I should be thankful to be where I am at...it used to be about once a week I would wake up and not be able to motivate myself to get out of bed...literally.  I would lie there absolutely incapable of moving.  It was not good.  There was no rhyme or reason and really nothing that I could pinpoint triggering it to come upon me or to lift...now, of course, I realize it was my hormones and body being so absolutely out of order but back then, I thought it was me and that I was the biggest of all losers...the weight thing hasn\'t really been helped with the met, but I haven\'t applied myself either.  The memory thing is still shot to hell but the results I have gotten, I couldn\'t be happier with...well, here I am again, fighting depression.    Don\'t get me wrong, I know it could still be the hormones...but it could also be my life.  I mean, after what happened to hubby at work, I was so freaked out and concerned for him...my body can handle like zero stress...my brain literally gets fried, I get exhausted, and I just can\'t handle it.  It is so strange because I used to be such a stressed out junkie...I lived on stress...now, the new reality is that I can\'t give a shit about a lot of stuff that I probably should because I will get a headache, get exhausted, or just get physically ill.  Anyway, between hubby\'s job, my work, etc...things have been stressful for us.    So, yesterday at work, I had a meeting that did not go so well...no particular reason, just one of my peers was laid off and yet we are reassured not to be concerned for our jobs...yeah...also, I have had mad reports to get done that get confusing and so indepth it is hard for me to focus...so after a day of that, I am greeted with my hubby and son, son is sick, hubby is pissed at son, so that sets me off...I try to gently remind hubby he is the adult and he is letting a FOUR year old get the best of him...of course, he gets pissed at me then...so all this is adding up, I get home and have a terrible headache, which turns into a migraine...well, b has awana so I can\'t take the meds that I can\'t find anyway...so I take him to church, can\'t bring myself to go up to the bathroom to search for the meds, so I lay on the couch and konk out...well, I never heard the doorbell and pounding when my dad brought b home, so who knows how long they were out there...got him to bed and then I am wide awake, with a migraine, and now the nausea has set in...so, I stay up for another two hours wishing I could make myself upchuck...I wanted to die, I felt so miserable...finally, I find some flat coke to drink (pop, not drug) force some down my throat and truck it up to bed...thankfully, the headache is GONE today, just the nausea remains.  All this to say...I can\'t deal with stress...I\'m just tired of being physically exhausted.  Thankfully, hubby says a lot less when I tell him I must lay down...but it is still hard.  Until you have dealt with exhaustion, please don\'t tell me that you know how I feel...I\'m wishing for the days where I was just tired...I\'m talking about fatigue, exhaustion...as in, I have to motivate myself to move every muscle in my body...even typing this shit is wearing me out...not just mind tired, soul, body, spirit tired...  '),(19398,518116,'30,March,2004','2004-03-30 00:00:00','   Okay...I\'m having blog issues!  I re-wrote (which I hate!) an entry because it didn\'t show as published.  When I look at it through blogger, both showed up, so I deleted the one that was a re-write.  But when I type in my actually address neither show up!  Then I click on current posts and they both show up!  This is the point at which I wish I knew what I was doing...I\'m so lost!  '),(19399,518116,'29,March,2004','2004-03-29 00:00:00','   I\'m so bored...I hate being bored...but at least I\'m getting paid to be bored...right???  I don\'t know...BUT since I have nothing better to do and I\'ve visited most of my regular online haunts, I decided to start reading the \"most recently updated\" blogs...some of that is SCARY but some of it is really cool!  The funny thing is, I hate responding to people\'s blogs...I feel like I\'m infringing, even though I have been wanting to get comments up so I can know if someone is reading mine.  Psycho.  But someone had a halarious posting about people who talk during movies/tv...that is SO me!  Hubby and I fight all the time about this because he will ask me what just happened, then again, and again because he keeps asking me to explain, therefore missing even more...just SHUT UP until commercial!  Plus, keep in mind that because I think I\'m going deaf I am usually watching (movies) with the subtitles at the bottom...I can\'t read the subtitles and explain to you at the same time!!!  Anyway, I responded and told her I was so with her...  Blogger...why can\'t I ever be on \"most recently updated\" blog list?  I think this is soon going to become an obsession of mine...I like to see my name in lights, I must admit...ohhh, I\'m mad now...I forgot to tell Lis how dorky her older sis really is!  See, I am a rw/rr junky...and I had to report an Abe sighting to my rr/rw site...http://www.therealworldroad.cjb.net/  Well, they thanked me on their website but now I can\'t get past their most recent bloggings to find what the date was.  It must have been sometime this month, last month at the latest.  If you find out the date, let me know...  So, I mentioned Britney to K.  She laughed and told me she would check her calendar.  Which basically means, do I have to?  I\'m not letting her off that easy!  I\'ll bring it up to her later this week and see what she says...I just want to get rid of my ticket!  '),(19400,518116,'29,March,2004','2004-03-29 00:00:00','   Heart of Crystal    urlLink   What is Your Heart REALLY Made of?     brought to you by  urlLink Quizilla  '),(19401,518116,'29,March,2004','2004-03-29 00:00:00','   I\'m so glad that worked!  Anyway...so sis can\'t make it to the Britney Spears concert...my other friend who is going with us had bugged and begged for her ticket, so when Lis couldn\'t make it, I told her she could buy it...well, now she is backtracking and saying she \"doesn\'t know\" if she wants it or not...ummm, the concert is in two weeks!  I think I\'m just being bitchy but I kinda need to know so I can figure out who else to sell it to!  Thing is, until I have it unloaded on someone else, I am going to sit here and worry about it...I\'m so weird when it comes to this kind of stuff...ocd.  So, I called her today to \"talk\" hoping she would mention if she wanted it or not...she didn\'t.  Well, i\'m going out to lunch with a potential buyer, so I am just going to ask her if she wants to go with me...if she does, awesome!  If not, keeping my fingers crossed Red will...but if she wants to go and Red decides she wants the ticket...fuck me...but I\'m not going to get screwed with an extra ticket because she can\'t make up her mind...see???  Obsessive...I\'m trying to rationalize and figure this all out and it is driving me nuts!  The bigger question is...why does a 27year old feel the need to go to this concert anyways?  LOL...  '),(19402,518116,'29,March,2004','2004-03-29 00:00:00','   Okay...here I am again, trying to figure stuff out...I found these really cute emoticons that I want to use...so this is my choice for today...  '),(19403,518116,'29,March,2004','2004-03-29 00:00:00','   HA!  I love it when can actually get some stuff to work!  Of course, it isn\'t really my mind figuring stuff out, just my fingers typing in the right questions and finding it...I FINALLY have comments...I have been blogging for 2+ years and JUST found it!  Yippee!  So, leave a comment...or two!  Hubby thought he was in hell this weekend...I made him help me clean out all the closets.  It feels like a huge burden has been lifted off my shoulders because they are all (okay, one is left) organized.  I know, I need help.  I don\'t quite know why I have this insane need for complete organization in hidden places...most people think if there is a door to close on it, who cares?  Not me...it is like this unrelenting thought that I have constantly telling me that my closets are in disarray...no more!  Well, it will last at least a month...  Other than that...it was 70 outside yesterday!  Unbelievable...so B and I went outside for a few hours.  I got to clean out the gardens and revel in the fact that the plants survived and are starting to come out.  Friday B and I went outside and had a mini celebration over our first flower of the season, a crocus.  It was so small but so proud, the overachiever of the bunch.  Of course, I got B to get ecstatic with me...hubby thinks I am nuts and passing it down to B...ha!  Purplepanda is gone...it is so weird, we are so off and on with contact...but of all my sisters, she is the one most like me...I miss her, just knowing she isn\'t 20minutes away...but, alas, she will be traveling the country getting paid for it!  I envy her and I hope she is able to enjoy it...she deserves to have a fulfilling life and I hope this gives her it...  '),(19404,518116,'26,March,2004','2004-03-26 00:00:00','   man...they are always changing things up around here...you know what I wish they would do?  Get better blog templates...dude, I don\'t know how to do this stuff, that\'s why I come to you!  You are cramping my style but forcing me to use these boring templates...oh, well, you can\'t complain about free, right?  Anyway...still no change here...haven\'t heard from the inlaws, supposedly hubby intends to call them this weekend...I feel myself getting sick!  We will see how it goes but he keeps telling me I need to prepare myself that nothing is going to be resolved.  Sam\'s not pregnant, supposedly she is having a miscarriage...  i\'m being too bitchy...I need to stop blogging...  '),(19405,518116,'23,March,2004','2004-03-23 00:00:00','   lordy...when it rains it pours...I have not slept through the night since Thursday and my colon is kicking into high gear.  It is really quite frustrating and I am trying to keep light hearted about all of this, but I just seem to be getting more and more stressed.  To top it off, my internet is not working quite right today, it is taking FOREVER for pages to load and sometimes they just don\'t...ugh.  Anyway...let\'s see if my tired mind can recount the important highlights of the reception...first, we arrive 20minutes late thinking my parents will be there by then and I won\'t be forced to mingle...wrong-o.  Not there, hubby has to use the bathroom, I\'m there trying not to vomit so I run into the bathroom.  I\'m half tempted to call Sam but realize that is nuts so I am trying to use positive self-talk as I\'m staring at myself in the mirror while washing my hands.  Ummm, wouldn\'t you know I start breaking into hives around the base of my neck!  Unbelievable.  Old \"friend\" Keri...got engaged, her fiance is almost 10years older than her and works at HCA.  Doesn\'t make enough money to cover bills so he has to work a second job on the weekend...her stone is so tiny it isn\'t worth having and she is visibly uncomfortable because she keeps folding her hand under her arms...she is only having a cake reception.  As Sam said, if she is having so many people, the least she could do for all the presents she will be getting is throw people a pig-in-a-blanket...ha!  BTW...I\'m tired of using initials...  Abe...he and I make eye contact...he smiles, I look away...  My parents get there, they immediately start mingling with the enemy...hubby and I go check out the wedding pictures.  My parents pick a table right next to the enemy!  My parents have to leave because my mom has work to do...hubby and I relocate to the other side of the table because I have been facing the wall the whole time...Jim comes up to talk to us (goes on and on about how many girls he has swarming to date him...yeah, whatever).  Gives me a hard time about the nose ring and church...both of which I make obvious are not up for discussion with him and I couldn\'t give a rat\'s ass what he thinks...meanwhile, Abe comes up to sit down with us (oh, I forgot, Jim told me Missy has a baby) and Jim immediately has to get up to go attend to something else...I would have laughed except now I\'m stuck with him!  Hubby goes up to get cake...Abe starts small talking and mentions how much he likes the nose ring and thinks it\'s cool...I roll my eyes as I am used to people just saying that...he says, no really...I think it is sexy when girls have their noses pierced...actually, you should watch out because I have a thing for girls with their noses pierced.  To which, I look dead in his eyes with a straight face, you do realize that was my husband that just went up to get cake.  he says, oh I\'m just kidding...yeah, whatever, don\'t flirt with me.  So, he and I are talking about his house and his wife comes up...Scott comes back with cake and Abe\'s wife makes a big deal about it being chocolate (oh, yeah, both she and Abe got TWO pieces, for themselves!  WT) and my hubby, being gracious, offers to switch with her...after this whole fiasco, Abe says, have you been introduced? He introduces me as Sara Barto, to which I say, actually Sara H and then Scott.  Never tells us her name (of course, I know who she is!) so I say, I\'m sorry, I didn\'t catch your name...ha!  I\'m just being bitchy...Corey Nichols.  Anyway, hubby tells me after we leave that I must intimidate Abe.  All during our short little conversation he was totally downplaying everything.  His house (it\'s just a house) and his job (it pays the bills, it isn\'t anything big)...etc...never mentions they are pregnant and of course I don\'t let on that I know...I can\'t say that I didn\'t enjoy this...like Sam said, it was very therapeutic to know that even if I\'m living in a cardboard box I have more class than all of them put together.  Corey...so NOT attractive...as hubby put it...from far away she looks attractive...up close, bad news!  She has facial scarring (acne?) and has this dorky haircut that is a  bad  old school Deborah Barrone.  Bad fashion sense, bad makeup application skills...the whole package is just wrong.  Again, her rock was too small to be worth having.  Plus, her and Abe have these really ugly sterling silver engraved rings.  The only thing I could think of was that they are the Hebrew engraved \"I am my beloved\'s and my beloved is mine\" rings...  All in all, it was therapeutic to go...why should these people be able to make me feel small again?  I have gone beyond the tangled in-bred web they all spin together...I escaped years ago...it will be interesting to go to Keri\'s wedding (if we are invited)...  Didn\'t get to say more than hi and congratulations to Tom...oh well, I tried...  Sam called today and told me she is pregnant...this is weird.  She told me at the end of the summer that she and Mustafah would start trying later this year...but today she told me they never have used bc.  Whatever, I don\'t give a shit anymore, lie, don\'t lie...I\'m too stressed out to even have a reaction, which is probably a good thing!  oh, yeah...Britney got cancelled...  '),(19406,518116,'22,March,2004','2004-03-22 00:00:00','   what a fucking weekend...I\'m so tired of talking about it but feel I like I should rehash this to have it for memory sake...it\'s beautiful, who would want to forget?  RIIIIIGHT....  Anyway...the inlaws refused to return hubby\'s calls on Friday, so basically he and I got no sleep and I was sick all night.  Saturday morning both of us were looking and feeling like HELL...so I asked him to please just call again and try to get it over with...so, as he is trying to dial through our calling card, the phone beeps and it is his mom.  Let me preface this by saying, for quite a few months, hubby has wanted me to seriously go get my hearing checked...so we feel like I could be starting to lose my hearing...I was across the kitchen and could hear her sobbing and yelling at Scott...I really can\'t remember how it exactly went but the gist was: you should have called Sara was out of line to do so AND if I said such a comment, you should have known that I didn\'t say it to hurt you...so, is that supposed to release you from the responsibility of your words?  I DON\"T THINK SO!  Hubby explained to her we had decided as a couple to have me call and we do not think there is anything wrong.  Blah, blah, blah...she wasn\'t listening and she didn\'t really care.  All she wanted was for him to concede with her opinions, but he didn\'t so she hung up on him...yeah, I know, but she really is older than jr high!  He and I talked for about 3hours...it basically boils down to, I have been right for the last five years, they don\'t want a relationship with me, I\'m merely unfortunate baggage that comes with my hubby and son.  I\'m done, no more mentioning we should go up there, no more bringing up that they should be invited to come down...I will put out as much effort as you do, which is none.  All of sudden my calendar is wide open!  What a weight gone from my shoulders...  Anyway...he talked to bil on Saturday night (no, parents never called back!) and told him...the thing that makes me most upset is that people are so concerned with getting right with mom and checking on her hurt feelings (for something she has done) that they haven\'t stopped to think about how my wife is feeling...she hasn\'t slept all weekend, she is physically ill, and she is crushed that they are basically telling her they don\'t want a relationship with her...  No resolution...my bil went and talked to them yesterday about it, they stand firm on their ground I shouldn\'t have called it should have come from hubby...hubby will call some time this week but he isn\'t in a rush to do it...they will have to agree to disagree there is not other solution.  I will be very proud if he actually tells him what he told his brother...I\'m so proud he has taken his balls back from his mom...  more to tell...I have the whole reception story to recount...but I need to go pick up my car...$530 for brakes and a rotor...ugh!  '),(19407,518116,'20,March,2004','2004-03-20 00:00:00','   rats...I was trying to take a break from this to read email...I hate spam!  So, basically, I relayed this all to hubby.  He got home extremely late and I was starting to panic that she had called him first...you see, there were times in the past where, when faced with choosing sides, I was on the losing end.  Basically, four years of our marriage was that way...so, I\'m a little paranoid when it comes to that.  No, she hadn\'t called, he was just busy at work.  I relayed to him how I felt.  My feelings are really hurt that she refuses to talk to me about this and that she basically told me I am blowing this out of proportion.  I also told him that she tried to manipulate me with tears.  He completely disagrees.  I tried to explain that women use this technique all the time.  If someone can turn them on and then turn them right back off, that is a pretty clear indication they are trying to manipulate you.  I think we agreed to disagree because I think he thinks I\'m nuts.  Oh, yeah...I forgot the other part.  The day I had this conversation?  THURSDAY.  Which means hubby was going to be at home the next night...duh!  I\'m  so tired of parts of my brain not participating in group activities!  Therefore, because I left a lot of room open to confuse my mil, hubby would now have to be the one to call...doh!  Anyway...everything peachy?  Not really...my spastic colon is in full force..I have spent the better part of the last 48 hours in a bathroom...Thursday night I think I got a total of 1 1/2 hours of sleep...all I can think about it the fact this is all going to be blamed on me...  So, enter Friday morning...hubby gets a call from his brother.  Wondering what happened and what the conversation was...I think.  See, by this time, hubby has become very elusive with me.  He won\'t really answer questions and won\'t give a rundown of what is being said.  The basics he told me...brother says this is not the comment that was made the comment was more \"They were in a different financial situation than you are, if you would prefer to have something else, let us know.\"  Which, at first, holy crap!  That changes things...but as my good friend Kate pointed out, it is still the same thing.  Yup, you\'re right it is...it sounds more pc but its the same...just wrapped prettier...bil also said that fil had called him in the morning to ask him what scott\'s mental and emotional well-being had been last time they talked and that I had called and chewed my mil\'s ass.  He also said that he had had a long chat with M and that we should not read much into what she says because she is pregnant.    There are SOOOO many issues I have with this...did I miss the part where being pregnant effects your brain and mouth?  Don\'t give me that bs...I was once pregnant too!  To which hubby says, I don\'t remember you being like that before you were pregnant but you were like that after...yup, you are right...and if you wouldn\'t have interrupted me, that is EXACTLY what I was going to say...but through it all?  I knew what I was saying and I didn\'t blow things out of proportion and blame it on pregnancy!  (Side note...we now know the reason I was so emotionally unstable after b was due to my PCOS...which they are JUST NOW taking care of...so back off...)  Also, and the biggest one...I CHEWED HER ASS????!!!!  Trust me, if you are about to get your ass chewed by me, you will know it when I am done that this is what just happened...NEVER did I raise my voice.  But, when I mentioned this to my mom, she had a good point...mil lies and manipulates, we know this just from the conversation the previous night...so what would stop her from lying and manipulating the ears she is bending?  And, since B was in bed and I was alone, it is my word against hers and we all know who\'s the loser in that situation.  Hubby claims bil never shed his opinion on the whole matter...he didn\'t say who\'s side he is taking...okay, but at the same time, he is sayiing it is my fault that I took to heart what M said, I should have known better...so isn\'t that saying he thinks I\'m at fault for this?  Add to that his comment to hubby that he thinks mil is abusing her drugs and that plays a part...okay, so everyone else has an excuse and the one left standing without the excuse is the loser?  Wow...everything really can be pointed back to me.  I tried to explain to hubby...this woman had a chance to explain.  I point blank asked her twice what exactly the comment was but she chose not to answer.  RIGHT THERE she shows that she does not deem me good enough to carry on conversation or relationship...obviously, she knew then this was not the comment she made, why didn\'t she tell me what she had said?  IT WOULD HAVE ENDED THE CONVERSATION.  You know why I think?  Because I think SHE knows that what she meant is what I was told she said...I think she knows that the real comment she made IS the same thing and it is the same thing because she MEANT it as such.    I am not going to try to make excuses...you are an adult, own up to your comment.  Do I need to make an excuse for being hurt by you?  Last time I checked, it is the person\'s responsibility who does the hurting, not vice versa.  If the comment wasn\'t made, sure, that changes the hurt associated with the comment...but after our conversation, there is new hurt...hurt that you have confirmed what the last five years have been about...you don\'t like me, you don\'t want a relationship with me, and you wish I would just go away.  I\'m so sick...I\'m trying to figure out what went wrong...what did I do wrong that caused this woman to hate me so much?  If I can accept you and try to love you because of who you are to the two most important people in my life...why can\'t you?  I really feel like running away and not leaving a forwarding address.  I asked my hubby what his stand is on this and what Jason\'s stand is...he said it doesn\'t really matter what either of their opinions are...like hell!  I have put up with five years of your family thinking the worst of me...five years of their manipulations, hurtful digs, and their obvious dislike of me...WHY would I care what they think?  I know what they think...you are the only person out of this whole mess that I care about...it is ONLY important to me that you believe what I am saying.  \"Well, you\'re the one I have to live with, of course, I\'m going to believe you...\"  Well, that doesn\'t sound good to me...believe me because you believe me, not because you HAVE to live with me.  If you don\'t believe me, I would rather know that...I don\'t want you throwing misguided belief.  YOU KNOW ME...are you telling me you really think after five years of hell with your mother I took this opportunity to give her the smack down?  I am so different than I used to be, maybe a year ago, I would have...but not now.  What the hell would that solve?  So, apparently, you don\'t know me as well as you should either...you too think the worst of me but aren\'t going to admit that to your family because that is your duty as my husband...  So, all of yesterday, he tried to get ahold of them and couldn\'t...do I think this was on purpose?  Probably.  I want to talk to dad, I want to talk to bil...I want to simply tell them, look, Scott and I reacted to what we were told...I was trying to clear up what happened.  I would hope you would realize that I tried to use the utmost respect and never was there a point where I yelled at her.    It probably won\'t happen...hubby expects fil to call and yell...why?  I was an adult...I didn\'t yell...you are an adult, you deserve the respect of not being yelled at...  gotta go...about to be sick...  '),(19408,518116,'17,March,2004','2004-03-17 00:00:00','   okay...I\'m back...  I call sil back.  Hubby and i both agree, there is a possibility that mil never made this comment.  Let\'s give sil one more chance to back away from what she told me before I call mil.  Nope, she said she understood why we wanted to talk to her and it was okay with her.  I apologized if this would get her in trouble but M NEVER changed what the comment was or put a new spin on it.  so, I prepare to make the call...  my fil answers the phone...I small talk with him for no more than five minutes and ask for mom.  She gets on the phone and I ask her if she has a few minutes.  She says sure.  I say \"I\'m calling because we were informed of a comment you made.  We were told you are telling people the reason you buy us gifts is because we cannot afford them.\"  MIL\"I don\'t know what you are talking about.\"  Okay, so my tactic of being elusive is not going to work.  I did not want to come right out and say who said what but it was appearing as though I had no choice.    \"Mom, I was told that you are telling people the reason you and dad bought us the crib was because we couldn\'t afford to buy it.\" \"I don\'t think I said that.\" \"So, you don\'t think you told people you bought us the crib because we couldn\'t afford it?\" \"I don\'t think I said it like THAT.\" \"How might it have been said then?\" LOOOOONG PAUSE \"Mom?  What do you think was said then?\" \"Did Gary tell you this?\" \"No, Gary did not tell us this...\" \"It must have been Jason then\" \"Mom, it doesn\'t matter who told us...\" \"Even if I did say it like that, what does it matter?  That was four years ago.\" \"Mom, do you understand why we would be hurt by a comment like this?\" LOOONG PAUSE \"Mom, do you understand that it is Scott\'s privilige and responsibility to take care of this family?...and you are basically telling people that he cannot fulfill this responsibility.\" LOOONG PAUSE \"I think this is being blown out of proportion.\" \"So, you are saying you can\'t understand why we might be upset about this?\" nothing...long pause....lasting probably 5minutes...the pause had gotten so long, I turned around to look at the clock and it was an additional 3 minutes... \"Mom?\" \"What?\" \"You can\'t understand why Scott would be upset about this?\" Enter sniveling...\"so you don\'t want me to buy brennan anything anymore?\"   Okay...side note...I am trying to recreate but cannot EXACTLY remember the sequence of everything being said...this is a RECREATION.  Please note, already, I have explained to her we appreciate everything she and dad have done for the family.   \"No, mom, that isn\'t what I\'m saying.  We understand buying things is a normal \"grandparent\" behavior.  But when you tell people you are buying us things because we cannot afford it, that changes the motivation behind giving and, yes, that would change our view.  We feel like you are telling people we are taking handouts which is not, nor has it ever been the case.  We thought you bought us the crib because you wanted to.  And if you felt like you had to, let me reassure you, we did not then and we don\'t now, need handouts.  We are okay...we were okay then, we are okay now, and we will continue being okay.\" \"You guys didn\'t have a lot of money back then...we wanted to buy that for B because he was our first grandson.\" \"I admit things were tight then, mom, but it never came to a point where we needed to ask anybody to hand us anything.  We have always thought the reason you bought us the crib was because you wanted to.\"   LOOOOOOONG pauses taking place...really, it wasn\'t very often that she would actually answer my question, she usually made me repeat the question before she would acknowledge it....   \"I don\'t understand what the big deal is...I don\'t thin this is coming from Scott.\" \"Mom, do you think I am blowing this out of proportion?\" \"I didn\'t raise my son to think like this.\" \"Mom, do you think I am making this up?  The reason I am calling is because Scott is very busy at work.  If you think that he doesn\'t feel this way as well, please feel free to call him.  Tonight is probably not a good time but I am sure he is more than willing to discuss this with you.\" \"I did not raise my son to think of his mother like this.\" \"Mom, this does not change our opinion of you.  We still love you.  Scott and I had a long conversation before I called you.  Let me assure you, we are a unit, we both feel the same way.\" \"This is just being blown way out of proportion.\" \"I don\'t know how to explain it to you any differently.  Maybe ask dad how he would feel.  You have to understand that a man\'s pride is wrapped up in how well he provides for his family.  I think we can all agree, Scott does a wonderful job doing this and busts his ass to do so.\"   okay...I had made this comment previously...I also made mention of the fact that both he and I had made a lot of sacrifices to be where we are at...and it was hurtful for someone to belittle that by making such a comment.  I also, at some point, mentioned that just because they handle their finances differently than we handle ours does not make their way right and our way wrong.  It makes their way right for their family and our way right for ours.     \"We HAD to accept things from dad\'s parents.  We were broke and there were many times they gave us money or groceries to help us survive.  There is nothing wrong with that!  We were very thankful for that and knew we wouldn\'t have gotten by without them!  There were a lot of times they would hand us a $100 bill so we could buy food or gifts for Christmas.\" \"Okay, you\'re right, but that is different.  There has never been a point where we needed that.\" \"this is being blown out of proportion.\" \"Again, mom.  Are you saying you think I am blowing this out of proportion?\"   To which she goes off on a tangent...for the life of me, I wish I could remember what it was.  All I could think about was the fact she hadn\'t answered my question and this was probably a pretty important question to have answered.   \"Mom, you still didn\'t answer my question.  Do you think I am the one who is blowing this out of proportion?\" \"I don\'t know.\"   Well, I think I do know...anyone I know would read this as a YES.   \"I am just going to wait and talk to Scott about this.\" \"Okay mom, that is fine.  But is there a reason you can\'t finish talking to me about it?\" \"I just want to talk to Scott.\" \"I hope you know you are free to say anything to me that you need to say to him.\" \"I\'ll just talk to Scott, tell him to call me tomorrow after I get out of work.\" \"Scott knows what I was going to be saying, mom.  If you need to talk to him, why don\'t you give him a call at work tomorrow.\" \"Fine.\" \"I\'ll tell him to expect your call.\" LOOOONG Pause \"I really wasn\'t expecting there not to be resolution to this tonight, mom.\" Nothing \"You are the mother of my husband, the grandmother of my son.  We are going to be in each other\'s lives a very long time.  We need to learn how to resolve conflict with each other.  ~pause~  That\'s what people in relationships do.\" \"Just have Scott call me.\" \"I think you should call him.  He has been pretty busy at work and it would probably be better if you called him.\" \"Fine.\" \"I hope in the future you will be willing to talk and resolve conflict with me.\" \"Okay, bye\" \"Bye\"   I don\'t really remember if this is truly the point at which I made the comment about mother of husband, etc...also, throughout the whole conversation, I thought I handled myself spectacularly!  No yelling, no name calling.  I had successfully channeled the calm cool lawyer act.  However, the one flaw I find is that I allowed her to get my goat about having hubby call her.  Why?  I\'m not sure.  Probably because I was so extremely hurt and upset that she refused to have a real conversation with me.  It was unnerving with all of her dramatic pauses and offensive that she wouldn\'t answer me.  We spent probably 30-45 minutes on the phone and maybe 10-15 of it actually talking (maybe 20).  Add on top of that, she basically accused me of blowing this out of proportion and making this up...how hurtful is that?  For five years you have given me hell...my hubby has been defending YOU telling me it is all in MY head.  For five years, I have been trying to convince myself of that and you just confirmed every demon in my head....you don\'t like me, you don\'t want to like me.  So, of the whole conversation, one bad choice.  I wasn\'t disrespectful, but it was definitely a pissing match.  '),(19409,518116,'17,March,2004','2004-03-17 00:00:00','   never has the title of my blog been more appropriate...life has gone from boring, monotonous, and relatively stress-free to full blown crazy nightmare...oh what a difference 24hours makes!  Before I delve into this, please note the time...why is it on the weekdays I cannot pull myself out of bed by this time but on the weekend I can\'t sleep late???  Crazy hormones!  Okay...I have been trying to vent to everyone else and it hasn\'t helped...so, as I was laying in bed trying to sleep, I realized I had not written this down...why does writing help clear the mind so well?  I\'m not sure, but I know I need some clarity because right now I just feel like I am in a psycho thriller...  Let\'s give some backup....of course, no thriller would be complete without the overbearing inlaw. MIL to be exact.  Hubby and I going on cruise in July.  What to do with B?  There really was no consideration for his parents to take him, neither of us felt comfortable with them having him for a whole week...besides, that would mean we are asking them to take a week off work...my mom is a teacher, we trust my parents with B wholeheartedly, logic says, let them watch him...exactly the reason why we are waiting until July instead of going on our anniversary in June, my mom will be out of school and available... Last time il\'s were here, mil asked what we were doing with B.  I told her my parents would have him, I had told hubby after they left, this is not the last time we hear about this...I could tell by mil reaction she was not happy.  Enter this past Sunday...mil COINCIDENTALLY has the exact same week of our cruise off of work.  WHAT?!  Do you truly expect me to believe that?  I told hubby the only way I would believe that is if that was the week of the Cherry Festival.  That is the week she ALWAYS takes off!  Nope, Cherry Festival is the previous week...so she mentions to hubby, well if things don\'t work out, we will be available...or at least we could split the week, half with us, half with them.  Hubby tells her he will mention it to me but that are plans have been set for QUITE some time.    Hubby turns 30 this year, mil always has issue with us because we do not feel the need to make a trip to TC for his bday...if you want to see him on his bday, get your ass down here!  Enter pregnant sil...she has to have a shower thrown by mil side of the family, enter mil plot to get hubby up there for his bday...LOW AND BEHOLD she tells hubby 4/24 is the ONLY day that works for sil to have a shower...uhhhh, right!    Please understand...if you want me to do something, I approve the direct approach...ask me!  Manipulation is a pet peeve of mine...I tend to (read ALMOST ALWAYS) flee to the other extreme of what you are asking...I HATE and DESPISE manipulation and I do the opposite to prove a point to you that I will not be manipulated!  SO...knowing how my mil is, I decide, okay, so the shower is on the 24th...this entire past year has been spent being left out of occassions like this because my mil is mad at me and doesn\'t \"allow\" invites to be sent to me...why do I need to go up there for the shower?  Hubby and I have tenative plans, a party or going away, for his birthday...I am not going to choose a baby shower over that.  I decide the best person to deal with is sil...it is her shower, it is her feelings that are potentially going to be hurt, I need to take it up with her.  I call her to tell her that I potentially will not make this shower but is there an alternative one I can attend?  Start talking to M.  Tell her, mil told hubby this is the date of your shower, blah, blah, blah...well, humorously, M had told mil on Monday that this would not work.  Apparently, I had mentioned to bil that we were either going away or there was going to be a party so he knew we already had plans.  In addition, one of M\'s other friends is having a shower that day.  Whew!  M and I started talking about how we were both well aware why \"coincidentally\" she wanted the shower that day.  M mentioned that she is hurt and frustrated because she is being used as a tool of manipulation when it, rightfully so, should be all about her.  That gets her talking about other crap that has been going on...mil is getting new vehicle, she bitched to hubby that she was mad at fil because he won\'t let her get the car she wants.  Why would you tell your son this?  Take it up with your hubby, not your son!  I tell M how weird I thought that was and how I had told hubby that it is really out of line...M tells me the reason mil wants that car is because her twin has the same model and this one was a step up from hers...whatever!  M goes on to say that mil is also mad because fil won\'t \"allow\" her to give hubby and I her blazer.  I guess I never realized I wanted it!  I guess mil has been telling people that she would feel \"so much better\" if hubby and I had a reliable and safe vehicle to be driving...our vehicles are just fine bitch.  Enter comment two...I am upset about the previous and tell M this.  M says, oh you don\'t know the half of it.  Proceeds to tell me that mil came to them and said, and I\'m quoting what I was told, which I was told was a quote... \"We bought s and s a crib for b because they couldn\'t afford it, BUT if you would like us to do the same for you, we would go ahead and do that\"   EXCUSE ME???  That got my redhot Italian temper BOILING!  Yes, things were tight when hubby and I first got married, but there was never a time where we couldn\'t buy what we needed...ESPECIALLY when it came to b!  Our whole life together has been about sacrificing ourselves for the best of B...that includes sacrificing the better of our marriage!  So, don\'t go telling people that we could not live up to our responsibility and you were the white knight that had to save us!  I called hubby at work to tell him...I mean, I was beyond angry.  He got MAD.  He said as far as he was concerned, if that is why they buy us things, they can come and take everything back...anyway, we have a lengthy conversation, in which I offer to talk to her, thinking that she is not going to try to manipulate me because she knows better and maybe I can get through to her LEAVE US OUT OF YOUR CONVERSATION!  So, I get B to bed and make the call.  Please understand...my mom is Italian, my dad is a lawyer.  I admit 99% of the time, I act more Italian than anything...but I have had an extreme amount of conflict in my life.  Very early on, my daddy taught me a wonderful lesson...when in conflict, your words are much more effective in a quiet, calm voice.  If there are two people yelling, no one will hear anyone.  Besides (and the real reason I love this), people become very unnerved when they lose their emotional \"cool\" and you are seeming like nothing is affecting you.  Inside, be burning up or torn down, but do not let the other person see your cards...   SO, I went into this conversation knowing I needed to channel lawyer-me not Italian-me.  hold on...the other fun part of all of this...the spastic colon is back...ewww  '),(19410,518116,'17,March,2004','2004-03-17 00:00:00','   I am  Mr Do .   I am sedentary by nature, enjoying passive entertainment, eating when the mood takes me, and playing with my food. I try to avoid conflict, but when I\'m angered, I can be a devil - if you force me to fight, I will crush you. With apples.  What Video Game Character Are You?       Not quite sure what this means...but, okay!  '),(19411,518116,'16,March,2004','2004-03-16 00:00:00','   ahhh...a new week.  Not much happening.  We watched a movie on friday, \"The Missing\"...wow...now, I would have never gone to blockbuster and picked that out.  however, hubby\'s friend loaned it to him and I told him I would watch.  It was suspenseful and pretty good...i probably wouldn\'t watch it again because it isn\'t my kind of movie, but it was definitely worth seeing!  Saturday...decided to have another go at the red paint...so glad I did!  This last coat really made a difference and I love it!  I also painted the trim and did the touch up this weekend.  All that remains to be painted is the door jam and the inside of the closet...next week is carpet installation!  YIPPEE!!!  It will be so nice to have soft squishy carpet instead of insanely ugly 50s vinyl:)  Got my hair cut...got taxes done...  So, another boring week in Pfizer-land.  Still no word on if I have a job after the end of the month...not much going on...  '),(19412,518116,'11,March,2004','2004-03-11 00:00:00','   I just finished reading through my past blogs...obviously, I have not a whole lot to do today...anyway...it is funny how I have the astounding ability to embarass myself.  The difference is, as of late, I can be the one laughing the loudest.  The sad thing is that I still do not have a great sense of self-esteem...but oddly enough, it is like I have accepted who I am and am at peace with it...don\'t get me wrong, I still have moments where I try to be someone for other people but I\'m trying to get better with that...I need to stop trying to justify my actions and just respond \"take it or leave it, it is who I am and if you don\'t like it, there is the door\"...  As for S...well, we are talking again.  It is weird...she has the same sick sense of humor and as a person who has had a past with ed, seems to be with me eye to eye on many levels...I don\'t think I can call her my bf...i think she just understands me more than anyone else in my life (girl friend wise).  I have gotten over the stealing thing...I forgive her but it makes me understand I need to be smarter than that and take it for what it is.  My other bf is getting married...I am bummed.  I was really hoping to see tickets go done so I could fly out there to be with him but it isn\'t going to happen.  So, I will see him for a minute or two at his reception here and go from there.  It will be interesting to see what his wife is like...supposedly, when S was talking to him, his fiance kepts making statements and finally t told her he would have to call her back...which makes me wonder if that is why he wanted to get off the phone as soon as she got home when we talked...S is dying for the reception...her mr x will probably be there and his wife is now pregnant so she wants me to try to get a picture...lmao!  I continually make changes to myself...and, hopefully, they are for the better.  It is truly amazing to see where I have been.  I think a lot of the growth with my emotional health can be accredited to the met.  I really feel like my hormones have finally leveled out and am able to stay even keel most of the time.  There was a time where I really thought I must be crazy...I couldn\'t understand what was going on or why I was so depressed...I\'m starting to believe most of it was from the hormones going crazy...I feel like a new person and it is good to be here!  '),(19413,518116,'08,March,2004','2004-03-08 00:00:00','   oh boy...can souls haunt you even if they are still trapped in a live body?  And by saying this body is live, I\'m making an assumption because I would  assume  that if this body was no longer alive, I would have heard about it...  It is so very strange...hubby and I are really doing well...obviously, we have a few spats here and there but nothing out of the ordinary...it is like we have finally gotten to the place where we can understand and appreciate each other.  It used to be where we would try to point out what we were doing for the other, now we both can see it, openly recognize it, and bestow gratitude for the actions the other is doing.  I finally feel like my best friend is back...  enter weird dreams...I have been dreaming about Mr X...not ocassionally (is that spelled right?!) but almost every night.  It is really kind of disturbing.  Sometimes it is me running away from him and sometimes it is him running away from me.  Either way, in the middle of a dream we sit down together and talk about how much we still love each other but that it isn\'t the right time or that it will never be.  So what the fuck is that supposed to mean?  I mean, I haven\'t talked, seen, or thought much about this guy for quite some time...and now this?  I don\'t know, it just seems to prove how messed up in the head I am, unless, which I\'m hoping is the case, a living soul can haunt you...  '),(19414,518116,'08,March,2004','2004-03-08 00:00:00','   birthdays...they don\'t have to suck now, do they?  I guess I have really never enjoyed my birthday...and although I always try to convince myself that I am not expecting anything, I think in the back of my mind I do so I\'m disappointed when it doesn\'t live up to my expectations.  Well, NOT anymore!  I have successfully convinced myself not to expect anything and, guess what?! It works!  So, hubby decides to let me get the diamond band I have been wanting...yippee!!!  We order it off of eBay (SO addicted!) and wait...it comes a week early!  So, I wear it around the house for the night and then tease hubby that he came home to take it away...he tells me I can wear it that I don\'t have to wait...boy, does he understand!  LOL...  Day of birthday...no biggie...go to work, one of my friends brings me chocolate chip cookies...some others take me out to lunch...go get my license plate renewed (Thanks, Granholm!) and go to a dr\'s appointment for B...we ship him off to church and then have chinese takeout...exciting???  No!  But very nice...  a lot of people surprised me and remembered it was my birthday...but I wasn\'t expecting it so I wasn\'t upset when people didn\'t remember...just the way it should be!  Now, I am dreading the reception of my friend from Texas...he is getting married on the 13th and they are coming here for a reception on the 21st (BRITNEY SPEARS!!!).  It will basically be a HCA reunion...and I am NOT looking forward!  I feel like I have successfully excommunicated myself from them and feel great about that and now I am being thrown back into the lion\'s den...ugh...it was rather humorous, he calls me to let me know he is getting married (I already knew, his mom called my mom for our addresses) after a year+ of not talking to each other.  So, he asks me if I will be coming to Dallas for it...I just laughed, assuming he was kidding, which he wasn\'t so then he was offended and told me that it would mean a lot to him if I could make it.  I told him I didn\'t think it was possible but that I would see him the following week at the reception here in Michigan.  Well, BF in austin called to ask why he was calling her...she then said, \"oh, you must be dreading seeing all of those people in one place\"...okay, so why didn\'t I think of that???  Until she mentioned it, I never gave it a second thought.  It was like I was able to successfully separate the two so I didn\'t have to deal with the panic...now I am!  I don\'t want to see these morons!  Anyway...tin tile ceiling is installed and painted...awesome!  I tried painting the room red this weekend and it didn\'t work so well.  So, now I have to touch up the red and touch up the tile and paint the trim....  more later...  '),(19415,518116,'27,April,2004','2004-04-27 00:00:00','   sad, lonely, depressed, etc, etc, etc...I lost a friend and have realized the old adage \"you don\'t know what you have until it is gone\" is painfully true here.  Yes, I realize it was our choice but considering the alternatives we did not really feel that we had a choice.  I guess I should be consoled that he is in heaven happily running around but I\'m really not...I miss him.  I miss the way he followed my every footstep...I missed the way he would look at me with those innocent eyes...I miss the unconditional love...I miss the bond he and I had.  curse houses, curse material posessions...who cares about those things when I don\'t have my friend to share it all with...never again...  '),(19416,518116,'26,April,2004','2004-04-26 00:00:00','   I am haunted...the strange part is, as far as I know, the soul that haunts me is still alive.  Does he realize he haunts me?  Is he aware of his starring role in my dreams?  Does he wake up and wish he was back asleep to spend another fleeting moment together?  I don\'t understand it and perhaps I never will...is this the evidence I wanted for years?  Did I excommunicate my soul mate?  I always questioned why we were so volatile yet I could never get enough.  Such extreme highs and lows, emotions I have never felt again since my departure.    Lord help me.  I love my husband.  He is the best friend I have...I would do anything for him.  I have tried to cut out this piece of my heart because I know if he had any sense of this he would be crushed...but can I control my dreams?  I sit here and think and rationalize about it all...I have been with my husband longer than I was with him...it has been so long, he surely is not the same person, I know I am not the same...it is a dream, only a dream...but why does it keep playing?  In the waking hours, I have made my choice, I am happy with my choice, I would never change my choice if it was ever an option...it is just the night that needs convincing...  '),(19417,518116,'23,April,2004','2004-04-23 00:00:00','   Sar, your subconscious mind is driven most by Kindness   This means you have a deep desire to be kind and fair to others. You may even be preoccupied with finding kindness in the world around you, far more than you realize on a conscious level.   It is possible that the underlying reason you seek kindness in the world around you, is that you fear cruelty, the opposite of kindness. That could drive you to unconsciously project kindness wherever possible into your world. Regardless of its origin, your steadfast adherence to being kind to others is felt by people you are close to.   You are probably more susceptible than others to being overwhelmed by emotions — both yours and others\'. It is possible that your unusually empathic nature is a result of your natural sensitivity to others\' pain, and your desire to help them avoid it. For this reason, things might affect you more than they affect your friends and family. To protect yourself from too much emotional intensity, you might want to keep an eye out so you can recognize it when it starts. That will allow you to slow things down until you feel grounded again.   Overall, your strong orientation towards kindness gives you an optimistic nature, which translates into you seeing the best in the people around you. Because you\'re not one to be overly judgmental, others may seek out your company when they need a friend to talk to. People close to you likely know that you care deeply about the inner lives of others and can listen to what they have to say without imposing your views on them.   Though your unconscious mind is driven most strongly by Kindness, there is much more to who you are at your core.  '),(19418,518116,'22,April,2004','2004-04-22 00:00:00','   just in case you were wondering what my problem is...   urlLink chocolate addict      thanks, Anne!  I knew it was true!  '),(19419,518116,'22,April,2004','2004-04-22 00:00:00','   okay, kids...have you checked out my links yet???  If not, please, please, please check out  Ironbound .  Hugo has been rocking my world as of late...if you have read my blog since its origination, you would know why.  In short, he has just summed up my lamentations for the past few years in one simplistic yet beautiful sentence...   \"i can\'t love you because i don\'t know you anymore...\"   it isn\'t right being reduced to tears at work...bravo, Hugo!  '),(19420,518116,'22,April,2004','2004-04-22 00:00:00','   Can you tell I don\'t want to get back to work???  I took this from someone else\'s blog...I apologize, I shut down the internet before catching your blog title...  This week a \"getting to know you\" theme with an insane twist.  First, as a warm up, using the letters I-N-S-A-N-E reveal six things about yourself.  Introspective, Neat freak (not really, organization freak), Sarcastic, Affectionate, uNconventional (yeah, I cheated), Ever evolving.  1. What\'s the best pickup line you ever heard and did it work? this is so retarded...and it wasn\'t really a pickup line as much as my naivete...a guy just started talking to me and told me he had a fiance.  Well, he was super cool and we were a lot alike...I fell for it and when he told me he didn\'t have a fiance...well, you know... 2. If you were a character in The Wizard of Oz, who or what would you be? Glenda, the good witch 3. If you were an element of weather, what would you be? rain 4. If you were a breed of dog, what would you be and why. a lab, give me a little bit of love and I\'m yours 5. If you were a toy or a game, what would you be and why? Rubiks cube...you can\'t really ever quite figure me out  Well, kids, I hope this has been educational...and, my boss is circling my cubicle so I guess I will get back to work!  '),(19421,518116,'21,April,2004','2004-04-21 00:00:00','   okay...I signed up for gmail.  It is sar.bare@gmail.com.  I can\'t say that I will remember to use it since I\'m so engrained into Yahoo, but if you drop me a line, I\'ll try to read it by the end of the year!  '),(19422,518116,'21,April,2004','2004-04-21 00:00:00','   I\'m so frustrated...how can you judge someone when you have NO idea what they are trying to explain???  One of my \"friends\" came over this morning and started talking about her husband.  She recently told him she was going to file for a divorce so he has been trying to manipulate her and, basically, acting like a child.  Anyway, the discussion somehow led into my dr\'s appointment.  She asked if it went well.  I told her it was fine, my meds were changed, he added the fish oil...blah, blah, blah.  I guess I had never told her specifically I have been diagnosed PCOS.  I guess it just isn\'t something I throw out to people.  I don\'t want anyone to feel sorry for me, I don\'t really want to have to talk about it.  I don\'t want to get mad when you, my best friend, are supposed to feign concern for me and forget I might have issues...it just saves me a lot of headache and heartache to just not go into detail.  Anyway...she wanted an explanation of what it is...which led into, how were you diagnosed?  Well, how was I going to explain this???  So, I started at the beginning, when I first was diagnosed with hypothyroidism...well, how did you know something was wrong?  How do you know you were eating right and you should have been losing weight?  At some point, it just got easier to say, \"Look, I was anorexic, weight is a huge thing.  I was busting my ass to lose and I wasn\'t so I knew something was up and I was concerned I would start practicing again.\"  Lordy...   note to self: a)don\'t tell people you aren\'t convinced give a shit what you are actually saying and aren\'t just waiting for the opportunity to judge b) probably not a good idea to get into this type of conversation at work.   So, I guess by telling her I\'m anorexic that meant I was really dying to know what she thought about eating disorders.  Ummm...have you ever even KNOWN someone with one?  Judgement, judgement, and, yes, more judgement.  After she was done with that rant, she went into \"You really should try NOT taking the meds to see what will happen.\"  Oh, yes, that is a brilliant idea!  I have been on these fucking meds for 9months and my hormones are still not under control...lets dump the meds and send them raging again to realize I really MUST take these meds...that makes complete sense to me!  How \'bout this?  You go fucking read some literature on a) thyroid disorders and b) PCOS and then come back and we can have an intelligent conversation.  At some point, she told me I had no business being on met because it is a drug for diabetics and I\'m not diabetic.  (Mind you, I was the one who told her it is for diabetics).  To which, I graciously replied, you really should read some lit on PCOS and then you would understand the reasons MET is working for us.  Okay...I know, I\'m being oversensitive but I am just so frustrated!  Does she think I don\'t know that she and everyone else judges me on my weight?  I know, I need to lose weight...I know, in a perfect world it would fly off.  But I have tried over and over...come up with new ways to attack.  Besides the depression I sink into when the new method sucks as much as the others and no weight is lost, I have to be careful.  I am obsessed with the way I look, obsessed with the weight.  Everytime I find out the number I am fighting those demons.  I don\'t want to go back to where I was in high school but I often contemplate it.  I don\'t even want to admit this because I am riding the fence right now.  Surely, it can\'t be bad to cut out all meals but one? blah, blah, blah...yeah, they are right over my shoulder tormenting me.  Then, we have the ever present question of \"so when are you having another?\" \"why are you waiting so long?\" \"do you just want one?\"  JESUS CHRIST!  I really thought I was doing well but last night I dropped B off at his church program and saw one of our \"friends\".  They had their first when B was two and now they were toting around another...it\'s tough...I\'m not going to lie, I want that to be me.  I think I am just getting apprehensive about M\'s shower...plus, I don\'t know if I have mentioned it, my sister is pregnant.  I am thrilled for her, I was worried that she too may be cursed with this so I am so grateful she is not (at this point).  At the same time, there is a small little piece that wants it to be me...  I guess it is just the ignorance...especially the blissfully ignorant.  They don\'t give a shit what is really going on, they just want to scream at the top of their lungs and have people listen to them.  Guess what, genius?  Not everything that comes out of your mouth is true...and who are you to judge?  I guess I was just taught that you do not judge anyone else until you can be found blameless.    Oh, if everyone was as diversified as me...ha!  '),(19423,518116,'21,April,2004','2004-04-21 00:00:00','   okay...so, are all blogger users being asked if they want to try the new \"Gmail\" thing?  I didn\'t get a chance to read it...did anyone else?  Any thoughts?  Not a whole lot to post...as expected, I am sick and very tired...I don\'t know...has anyone else taken fish oil?  I don\'t really have an explanation of why I should be sick...the met didn\'t change and I don\'t think my thyroid meds do that to me...the change in thyroid meds would be what is making me tired and hopefully that will be done with before the inlaws come this weekend...  AI...well, I still love Fantasia, I still love George.  Simon changes his mind every week on who he thinks will take the competition...do you think he does that to swing votes???  Don\'t get me wrong Jennifer and the other girl do have pretty voices, I just think George and Fantasia have the personalities to go along with the voice...how cute was her daughter???  Other than that, I passed out on the couch after AI so I didn\'t get to catch up on \"Real World\"...I have been hearing rumors that Frankie leaves.  I have mixed emotions about Frankie...such as, is she for real?  I have known cutters before and it really isn\'t something they let \"accidentally\" slip out...plus, I know the editing must have been supreme but cutters usually aren\'t healed in a 10minute consultation with a complete stranger whose best advice is, distract yourself to stop the behavior...wow!  If only my first therapist had told me that when I was in high school...think of the money and time that would have been saved!  So, I guess what I\'m saying is, I feel like we are seeing these snippets into who Frankie really is...if she really has these issues, which must be big, I feel sorry for her...I\'m sure people are going to be judging her and that type of scrutiny is not what she needs if she is trying to make progression.  Other than that...still trying to figure out the menu for this weekend...  '),(19424,518116,'20,April,2004','2004-04-20 00:00:00','   okay, so I have a confession...I used to live in a glass bubble.  Seriously...I didn\'t know what a swearword was until 4th grade...I barely knew NKOTB...I didn\'t drink until I was 20...AND I never watched a complete Molly Ringwald movie until last night!  It is hard to explain to people how conservative my parents were because they end up sounding like Nazis.  I hated it at the time but I look back at what my friends started doing in middle school and high school and am glad to have been saved.  Not that I think I would have given in to peer pressure, I was always the person marching to her own beat but I would have become aware of my individuality a lot sooner...  So \"Pretty in Pink\" was the movie I watched last night...it was good!  Sometimes when you watch old movies they drive you nuts because they are the old school bad effects, etc. but this was not like that.  I think my favorite part was when Duckie was lip synching...halarious!    I must say, I so identified with Andie in high school...I was the girl hanging out with all the preps spending all their money on clothes.  I just wasn\'t that way.  It wasn\'t that my parents didn\'t have money but I was raised to be frugal so for me, a complete outfit for $15 was much more of a high than getting designer jeans for $90.  I probably committed many fashion faux pas while bargain shopping but it didn\'t matter to me...I was different and I was made well aware of that by everyone.    At church I didn\'t fit in because that is where all my preppy friends went so I wasn\'t in designer duds like they were, I didn\'t party and get trashed like they did, I chose my friends over my boyfriend, and the list could go on and on.  At school I didn\'t fit in because I was seen as the \"elite\" crowd.  My mom was a teacher, my dad was president of the school board, I was the top of my class, and involved in any extra-curricular activity available.  It was kind of humorous because at church I was a prude and weird because of the way I dressed...at school I was the trend setter.  Talk about schizophrenia!  Then there was the situation with the boys.  I felt bad for Duckie last night...and it made me realize what my guy friends must have felt like in high school.  It wasn\'t until a few years after high school that I realized the guys that I considered my best friends had, at some point in high school, wanted to date me.  And I\'m sure they felt I was so self-absorbed in my life that I was just being callous...but that really wasn\'t it.  I didn\'t have any idea in high school these guys liked me like that!  I had gotten so used to being the eccentric girl that everyone is friends with but no one remembers to include, I just figured these guys were the same...they provided much less drama than the girls...they were like brothers to me.  I found a lot of solace in these guys and I would have done anything for them.    So the lesson, kids, is this...if you like someone or think you are in love with them...tell them.  Maybe she isn\'t being cold, maybe she isn\'t just using you...maybe, just maybe, she doesn\'t like herself so she would never dream someone else could either...  To be honest, as I\'m sitting here thinking and remembering back, I don\'t know what I would have done...I don\'t know what would have happened if one of them asked me out on an actual date...and maybe they did and maybe I missed it...maybe I just thought we were hanging out like we always did...I had a boyfriend off and on in high school so I don\'t know...I just know life is too short to ask the what if\'s...you have to just grab it by the horns and hold on...   in pink, of course!  '),(19425,518116,'19,April,2004','2004-04-19 00:00:00','   hey, hugo and anne...I hope you guys don\'t mind I\'ve added you as links...if you do, just let me know and I will take them down...  '),(19426,518116,'16,April,2004','2004-04-16 00:00:00','   ouch!  My body is killing me...ironically, it feels good.  It is nice when you can get normal responses from the vessel you occupy.  Plus, it reminds me of how much I accomplished this weekend...wow!  I impress myself...haha!  Friday...endo appointment.  Not as bad or as good as I would have expected.  I am basically where I was weight wise last fall...bummer, but at the same time I know how much I was gorging on when I was depressed this winter.  Plus, it is down from my gp appointment so I know it is at least budging.  CV told me I need to stick to the PCOS diet thing but start counting calories.  He said fruit is full of them and I need to be aware so I am not counteracting my proactive approach.  He also agreed that perhaps the fish oil is a good idea...though it counts toward my calories!  He said it will hopefully help kick my metabolism into gear, help with the IR, acne, etc and help me start losing weight.  On the other hand, my thyroid is uncontrollable...he said he is not sure why the increased dose did not work but he is increasing it again and re-checking in six weeks.  If it is still not at the correct level he will have to change it again...ugh...it sucks when it is the thyroid, that\'s the one that makes me want to sleep all day when he changes it.  On the plus side, he does not want to increase the met.  He is hoping the change in thyroid and the fish oil will help and my met can stay the same.  So now I am up to 6 pills a day...7 if I would start taking a multi-vitamin.  Lordy...I hate the responsibility!  On Friday night, hubby and I watched \"Radio\" with Cuba...good movie!  Not what I was expecting (I just knew it was a movie about football) but definitely a good watch...  Saturday we went up to B & J\'s house...we haven\'t seen them in a year so it was great to finally see them and catch up.  Both hubby and I love J...we were apprehensive as B never seems to pick out someone that is not self-absorbed but we could both tell that she loves him so much and they seem to take care of each other.  They are getting married on June 5th...I am so happy for him, he seems so happy, so content.  I pray this is the one and they live a long life together...  When we got home, we saw the mess Spike made.  Hubby is so ready to get rid of him...he tore a hole in our brand new carpet...I am sick everytime I see it, it hasn\'t even been a month since we had it installed!  He was trying to get into our bedroom and gauged the door and tore a huge hole in the carpet.  He got down to the furring strips and cut himself so in addition to the huge hole, we had blood spots randomly on the carpet.  Thankfully, he is fine (it must have been a small cut) and the spots came right up...but our poor carpet.  I called the carpet place and they put our name on a board for one of their installers to come do it on the side.  I\'m sure it won\'t happen before our families came over...damn...  Sunday...woohoo!  The weather was awesome!  B woke up at 7am telling us how much he likes to hear the chirp chirp...I\'m not sure that he realized it was the birds so I explained that to him and then he was going on and on about how beautiful it was...too funny!  So, hubby was supremely irritated that B was awake so early...I got him into a sweatshirt and we went out to start working.  I cleaned out the garden and weeded...it took about 1 1/2 hours and then we went in for breakfast.  Then, I decided to clean our throw rug and brought that outside to scrub...which led to the scrubbing of the other one...then I took comet to all the dirt on the walls and scrubbed down everything...then I went to the basement and worked some more on cleaning/organizing that...when hubby left for work, I took a rest (read nap!) while B watch \"Lady and the Tramp\".  When that was done, we went back outside and I did some transplanting...then, I worked some more in the basement.  After B went to bed, I cleared out the dining room and living room so I could do a complete vacuum and waxing of the floors.  I didn\'t want to wait until this weekend because they get so slick and I really don\'t want anyone wiping out when they come over...it is spring!  I have my mother to thank for this disease...but I feel so much better when everything is in its place and clean...I told hubby I was coming to a fast realization why my mom had four daughters...she could direct while we cleaned!  I need to train hubby and B how to do this stuff...  It\'s Monday...I don\'t want to work...  '),(19427,518116,'16,April,2004','2004-04-16 00:00:00','   You Are a Peppermint Cappuccino  You\'re fun, outgoing, and you love to try anything new.  However, you tend to have strong opinions on what you like.  You are a total girly girly at heart - and prefer your coffee with good conversation.  You\'re the type that seems complex to outsiders, but in reality, you are easy to please      urlLink  What Kind Of Coffee Are You? Take This Quiz :-)          oh, oh!  I\'m so excited...Lis got online!  You should go to purplepanda to read what this chicky has been up to...I admit, I\'m envious!  It sounds like she is having a fab time in CA!  Lis~ yeah, don\'t do the boy thing...also, try to put comments on your journal...you know more about computers than me so it shouldn\'t be difficult!  I miss you!  So...I took the detour for her little coffee test...wouldn\'t you know, I got the same as her...gotta love my sis!  Yeah, I agree with what it says...except, it has only been within the last few years I have gotten in touch with my girly-girl inside!  Don\'t worry, I\'m making up for lost time...haha!  this one\'s for you Lis...  '),(19428,518116,'15,April,2004','2004-04-15 00:00:00','   Hey...I think I figured it out!  Well, not really...but I read a discussion on unkymoods.com about it and was able to follow the directions...yippee!!!  So, yeah...one blaring realization...I\'m old...now, mind you, I should have known going into a teeny-bopper concert.  However, I really don\'t view myself as old, just a young spirit trapped in an old body.  So, we get to the concert and I immediately have to go to the bathroom.  Well, there is a poor soul whose job is to direct all the crazy women trying to use the bathroom.  \"You go here, I need four more ladies to come over here...\" Basically, directing traffic.  So, I\'m next and she says, \"Mom, you just go right over here...\"  CRAP!  I mean, yeah, I\'m a mom but not of someone old enough to come to this concert!!!  She was probably assuming that someone of my age would not be coming unless they were with child.  So, there were two opening acts...bummer, the first guy was really good but I couldn\'t understand what he was saying his name was...he had kind of a caribbean feel but was alternative.  Well, we missed the second opening, Skye Sweetnam...can\'t say I know who she is or cared, I was too busy hitting the souvenir stands with the girls...that was a trip.  Of course, being the uber-dork, I bought a pink newsboy...will I ever wear it???  Doubtful.  I don\'t foresee a situation where that would be perfectly appropriate.  Well, R\'s daughter wanted same hat but in black.  The first stand sold out by the time we got to the front...so we go to another.  Last one is on display and I tell the guy that is what we want.  He informs me he is not allowed to sell the display.  So, I\'m trying to convince him it is okay because they sold the display at the last stand...while we are having this discussion, another worker reaches behind him to take it and sold it!  At that point, I was pissed!  He couldn\'t tell me if they would get more and wanted me to go to yet another stand and wait in line for 20 MORE minutes.  I know it sounds stupid...but I felt bad, it is A\'s birthday and the poor girl just wanted a hat!  (No need to worry, the first stand got some more and she got it...whew!  LOL)  Anyway...the concert starts...please, do not buy tickets to the balcony or upper tiers...mind you, we were at the front of the first balcony but we did not get the full effect of people dancing and going nuts.  I had to sit there like a sane person...bummer.  It was a good concert.  Would I go see her again?  Probably not unless I got the tickets for free or I had awesome seats.  I just didn\'t get to see the dancing like I wanted to because we were too high up.    It surprised me that the pop princess was so sexual.  Mind you, I was raised in an extremely conservative home but I have been hearing that many parents get up and take home their young children so as not to taint their impressionable minds.  I\'m talking whips, dry humping, threesomes, guys on guys, girls on girls...a little over the top.  At the same time, did these parents not see ANY coverage on the tour???  I had heard about this right after it opened...granted, you may have already bought your tickets but you can\'t tell me you couldn\'t have sold them!  However, all this sexuality was only overpowering for 2-3 songs...it was kind of cool because she had a costume like the one in the toxic video where she is only wearing jewels...if I had a kicking body like that, I would be tempted to dance around in one of those too!  She explained her voice was out of whack because she has been ill but really only the song \"Everytime\" suffered...of course, that is my favorite other than \"Slave to you\" but it was cool...I didn\'t realize she could play piano either...  I called L while I was there...I know she answered but I couldn\'t hear anything she was saying...I tried calling her again after we were out but she was probably out partying by then.  I hope she heard some of the concert, I wish she would have been there too.  All in all, a pretty cool time...it was fun to get back in touch with the teen inside...I\'m just paying like an old lady today!  '),(19429,518116,'14,April,2004','2004-04-14 00:00:00','   okay...do any of you literate computer language guys know what is going on???  When I see my blog, it replaced all moods with today\'s mood...why???  If I go into manage my posts it shows the same thing BUT if I pull up the individual posts to edit it will pull up the right mood.  Then when I republish the blog it will put the mood in all posts from whatever post I was just re-editing...ugh!!!  This is driving me crazy!!!  '),(19430,518116,'13,April,2004','2004-04-13 00:00:00','   Oh, oh, oh!  I\'m so excited...Britney rocks tonight!!!  haha!  I wasn\'t getting excited until I heard one of her songs on the way into work...yippee!!!  Me, two girls from work, and their teeny-bopper daughters are going...of course, the moms are NOT excited...me, well, I\'m ecstatic!    Last night...not much...hubby and I decided...he called his mom and told her they were invited to come down for his birthday.  He also mentioned that he was tired of trying to figure out what everyone else wanted to do for his birthday so she could figure it out with his brother and to call us when they knew what time we should be ready for them to come.  Poor hubby...we won\'t be going to a baseball game, he just wants to hang out at home so that is what we will do  '),(19431,518116,'13,April,2004','2004-04-13 00:00:00','   lordy...when does it end?  weekend...pretty non-eventful...I don\'t even remember what happened Friday, I think I passed out on the couch.  Saturday...hubby has been wanting another tattoo...so, we woke up and I told him to go forth and get ye tattoo...not as a birthday present, per say, but for his birthday he could get the tattoo...so, he has his superman emblem.  It looks pretty cool.  He \"said\" he wanted to get my initials and b\'s initials somewhere but he didn\'t know where to do it.  He had thought about it as an addition to the tribal band on his arm but he didn\'t know if that would look right.  I told him I would much prefer it on his penis...\"if found please return to slh..\" haha.  I told him it would keep him honest...somehow, he didn\'t find the humor in that!  So, we are still trying to figure out what we are going to DO for his actual birthday.  His family drives me CRAZY...now, his brother and his wife won\'t come to spend the night because they think it will hurt his parents\' feelings.  They \"offer\" to come down for lunch and then they will go back to TC...gee, thanks, you\'ll spend a whole hour with us???  So, now, hubby feels like he has to invite his parents, brother, sister in law, and my family to the house for dinner, cake, and ice cream.  Actually, a late lunch...but whatever.  So, I had thought I would take him to a Tigers game on Sunday...well, he doesn\'t want to do that either.  I\'m not trying to be all emotional or bitchy but JESUS CHRIST!  Can\'t I come up with SOMETHING to do for your birthday that you will actually want to do?  I was going to have B go to my parent\'s house to spend the night on Saturday, have us get up for breakfast and go to the game but he doesn\'t want B to be gone from us...ummm, he ALWAYS wants a break...okay...he suggests B go to my parents on Friday night instead...no good, I\'m going to be cleaning and cooking for your goddamn family all night, I\'m not going to be able to go out that night.  It is just extremely frustrating...and he doesn\'t get why my feelings would be hurt that the last 50+ ideas I have come up with have been shot down and massacred...  I know...it isn\'t about me, it is about him...but how can it be about him when he won\'t let me do anything?  I tried explaining it to him but he didn\'t get it...so supposedly he is calling me at work today so we can try to figure this clusterfuck out...  '),(19432,518116,'09,April,2004','2004-04-09 00:00:00','   I forgot to discuss something very important...haha...The Apprentice.  Please, oh please, tell me this is a set up for Kwame.  Please tell me that Omarosa is not that delusional, hypocritical, liar that she appeared to be last night!  Please believe me, this girl has bugged the bejeebees out of me since first watching...HOWEVER, is she for real?  I mean, she has to know that the cameras are there and even if she is trying to lie in the moment she would be busted in the long run.  I\'m trying to give her the benefit of the doubt until next week...because all I see right now is if Erika decides to sue over the \"n\" thing, she just won...  '),(19433,518116,'08,April,2004','2004-04-08 00:00:00','   How could I not be???  There I was, watching Survivor...watching Amber and Rob make promises they couldn\'t keep...when low and behold my mother in law called!  Now, hubby and I almost always screen calls so the answering machine picks up and someone starts leaving a message and we decide if we want to talk or not (you know, the reason NORMAL people have caller ID...I\'m too cheap!)  So, for about 5seconds I seriously contemplated just letting her leave a message...but I wanted to hear what she had to say and get all of this shit over with if that was her intention...  So, I answered, said hello, nothing else...not rude but not warm and inviting like usual either.  She asked if B was in bed because she needed to talk to me and didn\'t want me to be distracted or not able to talk.  B was in bed, I have no idea if he was actually sleeping but he wasn\'t hollering for me.  She says she just wanted to call and apologize...apologize for hurting my feelings, apologize for running her mouth...I didn\'t know what to say.  Obviously, I figured she was somewhat calling to make things right but I guess I had bought into hubby\'s \"she\'ll never say she\'s sorry\" and was completely taken off guard...I told her how much I appreciated her calling to talk to me about it.  I didn\'t let it just go though...I figured I had a right to let her hear it from me how I felt...so I told her.  I said I was very hurt by the fact she refused to talk to me that night, that dad turned around and told j that I had chewed her ass, and by them not trying to make it right with me I felt like I was not welcome as part of the family.  She apologized again and told me it was never her intention to hurt me...she was taken aback when I called, dad jumped to conclusion (they never discussed what actually happened until THIS week!!!), and I\'m one of two of the only \"daughters\" she will have.  I told her I forgave her...told her again that it meant a lot to me that she called...and that was that...  Told hubby this morning (he got home late and I was passed out on the couch)...in fine optimistic spirit, he says \"So, did you ask her why she fucking waited until THURSDAY night to call?\"  I just laughed and laughed...he\'s right but considering I feel like a piece of hell froze over, I\'m not going to sweat the small stuff...haha  '),(19434,518116,'08,April,2004','2004-04-08 00:00:00','   Stolen from  urlLink   who stole it from someone else...haha!  1. Grab the book nearest to you, turn to page 18, find line 4. Write down what it says: \"She led her department in production.  She did this in...\" 2. Stretch your left arm out as far as you can. What do you touch first? my calendar 3. What is the last thing you watched on TV? I fell asleep during American Idol...who got kicked out??? 4. WITHOUT LOOKING, guess what time it is: 3:15pm 5. Now look at the clock, what is the actual time? 3:10pm...damn! 6. With the exception of the computer, what can you hear? My sixpence none the richer cd and the white noise system. 7. When did you last step outside? what were you doing? 12-1pm...lunch 8. Before you came to this web site, what did you look at? company global phone directory 9. What are you wearing? Earrings (three hoops, three diamonds), hair back in band, white shirt with flower rhinestone, pink coat, jeans, white socks (wrong, I know!) and black boots, white bra and pink undies, plus my nose and navel rings...oh, and my wedding ring and anniversary band... 10. Did you dream last night? I don\'t know! 11. When did you last laugh? not too long ago... 12. What is on the walls of the room you are in? pictures of hubby and b...pictures b colored for me...picture of roommate with her bfriend and Andrew Firestone (Bachelor)...and a clock and calendar... 13. Seen anything weird lately? I saw my old neighbors last night and that was pretty weird! 14. What is the last film you saw? I think it was \"50 First Dates\"  Halarious! 16. If you became a multi-millionaire overnight, what would you buy first? a house on the beach 17. Tell me something about you that I don\'t know. I was asked to be an extra in a Mel Gibson movie...I couldn\'t because the hours were too long and I was in college... 18. If you could change one thing about the world, regardless of guilt or politics, what would you do? I would feed, clothe, and put roofs over the homeless 19. Do you like to dance? yes... 20. George Bush: is he really doing Dick Cheney? ewww... 21a. Imagine your first child is a girl, what do you call her? I have always thought Isabella (call her Bella)...however, I like Lucia too... 21b. Imagine your first child is a boy, what do you call him? well, his name is Brennan Scott 22. Would you ever consider living abroad? hmmm...I don\'t think so...if I move out of the country, I think it would be somewhere tropical  '),(19435,518116,'06,April,2004','2004-04-06 00:00:00','   I forgot...    				 haha...since I knew I was packing, I am wearing my little DKNY sneaker slide thingys...my boss commented that it may be the first time he has seen me without a heel...haha!  '),(19436,518116,'05,April,2004','2004-04-05 00:00:00','   okay...I liked this, and I need a break from moving offices...so, I\'m copying...   I am not: packing my crap...I\'m tired of it...packing an office you have been in for 3 years is NO FUN! I hurt: my thigh...big nasty bruise from walking into an end table...doh! I love: my boys...if it wasn\'t for them, I am not sure any of this would be worth it... I hate: liars... I fear: cancer... I forget: everything...but let\'s not dwell on it! I remember: when I was so stressed out in high school and college...what was I doing???  I miss college and the stupid fun Jen and I used to have...we were so dorky and immature but we laughed and that\'s all that mattered. I imagine: being on permanent vacation...soaking up rays on Lake Michigan in my cute little bungalow. I hope: to live a long healthy life. I crave: chocolate of course! I regret: the times when I wasn\'t a good friend... I care: too much about what other people think...I try so hard to disregard but I know those suckers sneak in... I always: psychoanalyze myself.  I want: to pay off all debt and just live to live... I feel alone: after B goes to bed and before hubby gets home... I listen: too much...I need to talk more I hide: my emotions and fears I pretend: that I am happy all the time...I know, you would never guess it reading my blog...but seriously, I don\'t know how to act sad, mad, irritated...I jump right back into happy. I drive: myself (and probably my hubby too!) crazy! I sing: ALL the time...the worst part about being at work is not being able to sing with the tunes... I cry: mostly when I\'m pissed... I destroy: myself...not as much anymore, but I\'m still prone to mentally abusing myself...at least the physical abuse is done...progress I dance: not enough!  And pretty badly!  LOL I write: a lot...other than music it is my favorite release. I wake: hubby when he sleeps too long!  haha I breathe: these damn chemicals at Pfizer...yuck! I play: music constantly... I venture: whenever I get the courage...I love to go explore new places but I think the paranoia from my mom has been passed down...I\'m always looking for the scary guy about to kidnap me... I find: laughter really is the best medicine...I love to laugh! I pray: God will keep my family safe and healthy I miss: playing the piano I kiss: my boys... I succeed: ??? I learn: more about myself everyday...okay, I really don\'t listen well enough for everyday, but every week.  I used to avoid myself and now I wonder why I didn\'t listen sooner... I feel: annoyed...I have water in my ear...I can\'t get it out and I\'m scared my ear drum will burst again... I know: that I am strong enough I joke: terribly!  I am the WORST person to tell a joke...a) I can\'t remember how the story gets to the punchline b)I\'m laughing too hard because I know the joke and it makes me giggle...I try not to tell jokes! I say: feed the American children...then you can worry about the ones that don\'t live in our own country I change: my home decor a lot I fail: at being the perfect person people want me to be. I dream: of being healthy, fit, and content with myself. I believe: I am a better person than I used to be I wonder: about many strange things...and I like to have answers so I usually make myself look foolish because of this weird curiosity of mine... I worry: that my hubby will die and I will be without my best friend...I know, I shouldn\'t even think about it I wish: for lovely weather again. I fight: only when necessary...I hate conflict I need: much less than I am blessed with  I am: who I am because of where I have been I: choose to be filled with joy! Current music: Evanescence   Feeling MUCH better than yesterday...I just have these moments of dark where I allow all the demons knocking on my door to come into the light...bad idea.   Poor, poor B...he took a nosedive right out of his loft last night...scared the bajeebees right out of me!  It was after midnight and I must have fallen asleep while waiting for hubby to come to bed...all of a sudden, BAM! and then he started to cry a little, I\'m pretty sure he was still sleeping until I ran in there and grabbed him.  Thankfully, his blanket came down with him and probably protected him from being beat up by anything in his way...tonight, I will be reorganizing his room to make a clear path if that should happen again...it is so weird because he has had this bed for over a year and has NEVER fallen out...  '),(19437,518116,'05,April,2004','2004-04-05 00:00:00','   My post on my PCOS board...   Mandi~  Well, I can tell you why you wrote...it is exactly what I have been feeling...I was so sick this weekend...surprise, a period started on Thursday...but no period I have ever experienced before...I hate them.  I used to be physically ill when I was a teenager with cramps, so I know what hideous cramps are.  But this...not even my meds could subside them in the least.  And, I would be feeling better, than get worse again.  I basically slept all weekend and am sitting here at work wondering how I\'m going to make it through tonight with B...  I look at my weight...and, yes, since having B I have gone down one size.  ONE SIZE...I have been watching what I eat, started excercising and NOTHING is happening.  I\'m so disgusted.  I hate looking the way I look, I hate feeling the way I feel and sometimes I wonder if I\'m not losing weight because I\'m too scared.  I can\'t get myself to accept this is it and I can\'t take not seeing anything move.  I started crying to hubby today because I feel like shit...it makes me wonder if I wanted the Met to work so I made myself think I was feeling better.  Right now, I feel so close to what I did last year this time.  My body has started aching and stiffening and it scares me.  The only good thing I have is that my appointment with the endo is next Friday.  I will be able to unload on him and see what he says.  Plus, I haven\'t had my blood taken yet (doing it tomorrow) so maybe it will show if something is really going on to make my muscles so tight...  anyway...thank you for your post...it really did help...and I\'m trying not to lose hope.  Most days, hope is all that keeps me looking to a better tomorrow.  Take care~S   god...I wish I could stop being a downer...I\'m really tired of being \"depressed\".  I know I haven\'t hit bottom yet, but do I have to???  '),(19438,518116,'04,April,2004','2004-04-04 00:00:00','   ugh...I hate periods.  I go through loving it when I have a semi-normal cycle that comes right on que...to this, hating it.  Why does my body have to be so messed up?!  I think this is what I really hate...my cramps have been on and off all weekend and it really is driving me nuts.  I don\'t understand it and it worries me...one more thing to take up with the endo in a couple weeks...  So, hubby and I had a knock out fight on Friday (???)...I can\'t remember but I\'m almost positive...yeah, it was Friday.  Anyway, Lis called from LA to talk to me about her new job...I miss her so much!  She sounds like she is having an awesome time...anyway, she had missed my blogs about all the mil drama before she left...so now she won\'t be able to read them until she comes back home...I tried to sum it up for her and, at some point, something was said which hubby believes Lis called his mommy psycho...I don\'t remember it...maybe she said it, more likely I said it...but, yes, we are all agreeing, mom=psycho.  He was NOT happy!  He thinks that is like the pot calling the kettle black (he thinks my family has our own issues...which we do!) and he doesn\'t appreciate me talking to my sis about his mom being crazy...  Then...I burned the steaks because I wasn\'t paying attention while I was talking on the phone...B REFUSED to eat the mashed sweet potatoes...basically, we all had melt downs...B was sent to bed and had all of his toys taken away (yeah, crazy, I know...)...then hubby and I came down here to start screaming at each other...ugh.  He basically told me that never in our years together has he heard me admit I might be wrong so he should just shut the fuck up and not say anything...I told him to go live a psycho life with his family in TC...and asked him to leave...(I also threatened divorce at some point...mature, don\'t you think?).  Well, he left...for an hour...but it was a LONG hour...to which I did actually psycho analyze myself...and, I need to go back and ask him if he really thinks that or if he only said it because he was mad.  He came back...we made up...everything\'s hunky-dorey...I feel really bad.  I think I have come so far with the temper/swearing/name calling/threatening bs and then I just allow myself to go over the edge...I apologized to which he accepted...  Okay....but do I really think I am always right?  Ummm...I don\'t know.  I am a highly opinionated person, yes, but I don\'t  think  I am always right...do I?  See, I know I USED to do that but I didn\'t think I was that way anymore...I actually think that every person is a set of circumstances that make them who they are and influence the choices they make and the way they live their lives...who am I to question/judge someone?  I have not lived in their shoes, nor have they lived in mine...my mommy would be so disappointed.  She is of the opinion that almost, if not all, things are black and white...I think we live in a mostly fuzzy gray world.  The point is, if I do behave like this, I feel really bad...I mentioned to hubby that there aren\'t a lot people who I really care to know what they think of me...he is the number one person I care about...I don\'t want him to think that I have this over the top opinion of myself in which I am right and he is wrong all the time...obviously, it happens naturally a lot of the time...haha!  As for the birthday...as hubby said, we are on plan D.  The master plan: VEGAS...shot down because of money needed for cruise...plan A: Big birthday party...gone, hubby doesn\'t want a huge crowd just for him...plan B: A family party...no good, I told hubby I am willing to invite his family but I think he is thinking it won\'t work...could be true but I feel bad, it is HIS birthday...plan C: His brother and sil coming down for the weekend...now, this one is not completely out but gets a little tricky...I guess I will call the sil and tell her party is out and dance around the issue if they still want to come down.  Hubby would love to spend time with his bro but don\'t want his parents upset they weren\'t invited...makes sense, but I don\'t really care how they feel right now...SO, if that doesn\'t work (or maybe I will go in correlation with plan C) is plan D: a baseball game...hubby has been wanting to go see the Tigers play for quite some time and ironically, they do have a game on his birthday...I don\'t know if it is a \"real\" game or a pregame, but does it matter that much???  Oh, yeah...mil is a fruitcake...she actually asked hubby if we would come up for Easter when they were on the phone...I TOLD him that she would assume I was just going to  get over  what happened!  '),(19439,518116,'02,April,2004','2004-04-02 00:00:00','   I LOVE being able to call in sick...now, was I truly sick?  It all is a matter of perspective...honestly though, it is that time of month...and since I haven\'t had a time since early February, I guess this one decided to be a doozy.  I am thankful for it though, I was getting really worried about what my hormones were doing and dreading going to the endo in a couple weeks...I still think he will up my medications because I missed again but I feel a false sense of accomplishment since there was no induction of the visit...anyway, I got up this morning and started getting ready and was overtaken by cramps.  The dog ate my bed buddy a month ago, so I have no source of heat to aid in the cramps...I took some Pamprin but it never kicked in...so, finally, at 8, I called my boss and told him I would not be showing up...oh, rats!  I went back to bed and slept...woke up about 11am to hear hubby on the phone...  one guess who it was...did you guess?  Yes, the beloved mil...I went downstairs to see what B was doing (watching cartoons) and went into the kitchen where hubby was...I wanted to find something hot to drink but we have no cafe latte...only coffee, hot apple cider, or hot chocolate...nothing sounded good...so I sat there while hubby wrapped up the conversation with his mother...ugh...  he gets off the phone and just stares at me...well?  We didn\'t really have much of a conversation about it because I\'m hunched over trying not to fall over from the cramps...anyway, the gist of what is said...she still claims she em>may  or  may not  have made the comment about our financial situation...she has the utmost respect for me because I handled myself so well and was never disrespectful to her...my fil told J I chewed her ass based on her hideous reaction to our conversation (her going postal while we were on the phone and then going to her room and sobbing after she hung up)...I took her off guard because she couldn\'t figure out who I had heard all of this from and that is why she wouldn\'t talk to me...she did not intend to send the message that I couldn\'t call her, she now claims that she feels I was strictly calling because hubby was upset...uh-huh...she also is so distraught because she knows how much I must be hurt by everything that has gone on...meanwhile, she wants hubby to tell her what needs to be done to make it right with me...to which hubby says, you need to talk about it with sar, I can\'t be in the middle of this, I support her and you need to make it right with her.  I am still trying to figure out how I feel about all of this...on one hand, yippee, they realize that they hurt me and I feel like it has been a huge milestone to get them to see I am a real person with real feelings that can be and are hurt by them...on the other...are we in jr high?  How can you not know how to make something right if you hurt another person?  To which hubby says they are really bad communicators...I\'m pretty sure even the worst communicator knows how to say \"I\'m sorry\"...even B, who is four, knows when he needs to give someone an apology.  I just find it really hard to believe that she is up there so concerned that I am down her having been hurt and she can\'t come up with a solution to try to make it better.  Meanwhile, hubby\'s 30th is only three weeks away...he wants his family and my family invited over to celebrate.  I asked him today how he felt about just his brother and sil being invited.  I don\'t mean to be a bitch...but how can I have these people in my house and celebrate with them when they can\'t extend any type of an apology...even if they don\'t apologize, could they at least try to talk to me about what happened?  I told hubby I thought it was strange in a way that she had not thought about the long term ramifications of what has happened.  I.E...I have her only grandson (for at least a few months), and does she expect me to invite them down, go there for visits, and allow B to go there without us when none of this has been resolved.  To which I followed up, but how is this any different than the past five years...I have put up with this woman\'s shit since hubby and I got married and that hasn\'t changed my behavior, it has resolved me to work harder to be accepted...damn...maybe I am nuts...  Anyway...I read the book \"Holes\" today...it is a children\'s book my sis loaned me along with \"Bridget Jones\' Diary\"...it was so good...I really like books that have twists or correlations in them...it was an easy read, it took 2 1/2 hours to read...lol...it makes me want to rent the movie that has been made but I always hate it when I end up watching the movie and not liking it...if they ruin the story it makes me mad...haha...  '),(19440,518116,'01,April,2004','2004-04-01 00:00:00','   I forgot to cover American Idol!  My lord were the judges rude!  I know it is their job to critique, and I agreed that most of them had some issues on Tuesday...but why comment about someone\'s weave!  My LORD!  It was like she sang well but we don\'t like her so let\'s find something else to rip on...oh, her hair!  The kid who sings like Frank did horrible, and I have heard that the judges had him in tears after the show...my god!  This kid is what, 16?! Leave him alone!!!  I agree, his voice is geared for one genre...but he does a mean impersonation...he is so dorky, he is cute...the thing is...even if these people have \"bad nights\" they are tons better than the majority of people who get up to do karaoke...  I didn\'t vote...haha!  Still only the one time...my favorites are still George and Fantasia...those two rock!  They had best make it into the top three...I\'m not sure who else I would choose...the thing is, the rest are all kind of on the same level...they are good but not consistent.  I must say, I was glad Amy went home.  Don\'t get me wrong, she has a good voice...but those times where she just went all airy/breathy and couldn\'t really hit the notes just happened too often for me.  You can\'t hit those, stop trying and stop picking songs that make you do that...  '),(19441,518116,'01,April,2004','2004-04-01 00:00:00','   Lordy...I hate having to do actual work when I am at work...I\'m not really sure WHAT I did, but yesterday, I never stopped...ugh.  I had some shitty training that I have had to attend four different versions of already.  I sit in the room and see people who I KNOW have already been told this crap and they have questions that have ALREADY been answered previously...people are morons!  To add to my joy of cubicle living, I have been informed I am being moved into building 306...that\'s right, the den of evil.  It is going to be miserable...no more anonymity, plus, I am sitting right by one of the entrances AND the only bathrooms in the building...ew!  My boss couldn\'t care less...he is on the opposite side of the building from management and is happy...meanwhile, I am in the center of traffic and will be visible to all...bastard...  On a halarious note, hubby called last night from work...he told me he had \"ordered\" us a present.  I was highly suspicious because if my hubby sees something he wants, it is usually in the back of a magazine or catalog but he was telling me that the surprise is free.  I asked him if he would give me a \"hint\" to which he just told me what he ordered...a better sex video and booklet and some other surprises are in it!  LMAO...I was laughing so hard!  See, I\'m considered the \"freaky\" one and he is always so resistant to new things...he liked his kama sutra book I got him for Christmas a couple years ago, but he only reads up on a new position every three-four months...so at this rate, we won\'t be through it before the next life.  I guess this radio station he listens to had a \"sex expert\" on and they were giving away these better sex kits.  My hubby decided it sounded cool and called in to order us one...haha!  '),(19442,518116,'27,May,2004','2004-05-27 00:00:00','   okay...I\'ll admit it...I\'m officially an emotional wreck at the moment...Lord, I really do love these hormonal rollercoasters I get to ride on!  This one I jumped on at the moment is called \"Time of the month Hell\"  Of course, if you don\'t choose to ride this ride as often as you should, say once a month, it tends to be that much worse when you do get on...yesterday, I was a WRECK!  I was tearing up and occassionally breaking down in tears at random moments...beautiful!  So, before my hormones went to hell yesterday, I called my sister and brother in law...aka the whackjobs in TC that just gave birth to my first niece.  I called because I wanted to see how the new mom was doing, how baby was doing, and how they were all adjusting after coming home from the hospital.  Bad move...she called me back here at work and went on and on about how she gave birth naturally and how my bil has such HUGE respect for women who choose to do it completely naturally.  Hey, dumbass, just because I had drugs doesn\'t make it any less of a challenge to push a baby out of that teeny-tiny hole.  Ahhh, I hope she achieved whatever it was she attempting to do...make herself feel good?  I don\'t know.  I have never regretted my decision to get drugs...I guess it is just annoying that she should think her birthing experience was more miraculous than anyone who had drugs.  Guess what, brainiac?  It\'s God\'s miracle in the first place...he just chose to use you as a tool...needless to say, I won\'t be calling them for quite some time...  Luckily, hubby completely understood.  See, I called from work and had to leave a message.  I didn\'t leave my work number, I just figured I would talk to them later in the afternoon.  So, they called our house to get my work number from hubby.  Hubby and his brother talked for a while about how they were doing...hubby had told his brother that I had been anxious to hear how everyone was doing and that I must have finally broke down and done it myself since he hadn\'t.  Brother asks him if we had thought about having more children...they were so excited we had dropped everything to come see them and to watch us with the baby, they were impressed and wondered why we hadn\'t had more yet.  So, hubby says, I could have sworn we have had this conversation...and goes into the whole, she has been told it won\'t happen naturally...bs.  Wow...I don\'t even really know what I think.  I think I don\'t want to think about it...anyway, brother kept going on and on about how he was so thrilled we came up right away and hubby had called ME before they called back to let me know the conversation and \"forewarn\" me she would probably go on and on about how much our visit meant...ummm, nope!  Nothing like that was even alluded to...so hubby is thoroughly disgusted, especially considering what he and his brother had talked about.  Like I told him, I am a sensitive enough person NOT to say such bs to someone but the downside is that I am sensitive to then get my feelings hurt when someone says that crap to me...I need to learn to let this stuff run off my back and not effect me...  On a much, much happier note...I was able to stay awake and watch AI.  GO FANTASIA!  LOL!  I never got through the night before so I was a little stressed she might not win.  Oh yeah, did you see the point JASMINE went to sing and Jennifer Love Hewitt called her Jennifer???  DUH!  Read the cue cards girl!  Anyway, I was so happy for Fantasia...and, of course, it made me cry!  '),(19443,518116,'24,May,2004','2004-05-24 00:00:00','   wow...I just feel so, strange...I don\'t know, it is weird.  I\'m like in between melancholy and happy I guess.  I am not depressed but I\'m not my usual self either.  I don\'t really know how to describe it...I feel so much more introverted than I usually am...more self-reflective...  So, sister in law had the baby...we got a call on Friday night to say she had been having contractions.  Well, every first time mom thinks they are about to have a baby so both hubby and I kind of poo-poo\'d it.  Well, low and behold, Saturday morning I got a call at 7:30am to say they were in the hospital.  Hubby and I had a long talk/fight about what to do.  I was so opposed to going up there...not because I didn\'t want to see the baby, because I did.  Finally, I was able to verbalize, I didn\'t want \"us\" to be in turmoil again like what happened after the shower.  I didn\'t want to put myself in such a vulnerable place to be hurt and then have him take their side.  I was scared.  But after being able to say all of that to him, I saw in his eyes how important it was to him to go.  We were supposed to have our friends from IN come...well, hubby said to have them come and he would go by himself.  I called T and explained...told her I just felt I needed to be there with him, she completely agreed and wished me luck.  I know she doesn\'t understand me 99% of the time, but I love her...she is such a good person and is such a great friend to me.    Anyway, I cried to my mom about how scared I was.  Please understand, I am not upset they were having a baby...it just brings to the forefront something I try so hard to ignore so I can go about my life...it may never happen again for us.  Add to that vulnerability and hurt the fact that his family always makes cruel comments regarding this...I just was dreading the inevitable abuse I was about to put myself in front of.  Mom said she had no idea what I was feeling like...but encouraged me to realize I was doing the right thing and I was being the bigger person...then my preggo sister called.  Talked to her for a while...kinda get the picture I was trying my best to delay??? I was!  So, we finally get ready and leave at 11:45am...car trips with toddlers!  Literally, ten minutes after we leave...\"I gotta go potty.\"  YOU JUST WENT!  Lordy...what is normally a 2 1/2 hour trip took almost 5 HOURS!!!  Three stops for the bathroom, one for gas (and the bathroom) and one for lunch.  We finally got there at 4:30pm...man!  And, of course, it was freezing and we had dressed for the hot and humid weather we had.  I actually had commented to hubby that we should pack coats...but we both kinda laughed it off and said there was no way it was going to be cold there...wrong-o!  I can\'t believe how good the baby is...she barely cried while we were there.  We stayed for about an hour and a half.  Hubby immediately wanted to hold her, which he usually is very apprehensive about babies.  At first, I didn\'t know if I could...but hubby kind of shoved her in my arms.  He always knows exactly what I need...it immediately took my mind off the bigger picture and I got to hold my new niece, Leslie Maxine...I\'m not going to say she was cute, I\'m not a fan of newborns...wait a few months and I\'m sure I will find her adorable but babies are just not my thing...anyway!  Yes, I got all kinds of, \"when are you going to have another?  don\'t you think B wants a sibling?\"  \"B, you should tell mommy you want a baby too\"  all the inappropriate bullshit people impose...and, yeah, it hurts...alot.  But, we left to go meet hubby\'s parents for dinner and stopped at Old Navy and I was able to nurse my wounds with retail therapy.  Yesterday, I slept basically all day.  Hubby woke me up for some breakfast and I took a bath...then he woke me back up at 2pm when he left for work.  I can\'t explain why I was so tired...but I was.  Hubby says I just seemed so sad...I guess I kind of am.  I want to have another baby but neither of us want to take the next step of intervention.  Part of me thinks, do it, try it, get what you want.  The other part of me says, if it is meant to be it will happen.  That part always wins out...but what happens if I regret that decision when I get older?  And part of me wonders, why isn\'t it meant to happen?  Do I suck so bad at being a parent God doesn\'t want to entrust me with another?  To top it off...my breasts are hurting...damn.  Of course, my heart says, oh, I\'m pregnant!  But my head says, you are about to have a period.  So, I look at my calendar and it has been two months since the last one.  Ugh...I\'m really off track again.  I swear I don\'t want to spend another cent on a hpt...those damn things have gotten enough money from me!  But on the other hand, do I sit here and constantly question until a week or so when it shows up?  Better to put myself out of my misery I think...  So, no, I\'m not depressed...I don\'t know what I am...self-reflective...  '),(19444,518116,'19,May,2004','2004-05-19 00:00:00','   oh lordy...Fantasia I am SO sorry...I fell asleep before I was able to get through to vote for you...damn!  If you are booted off, I am boycotting...  '),(19445,518116,'18,May,2004','2004-05-18 00:00:00','   ok...so I better write about this dream before I forget everything.  Mr X was there...he had been searching for me for a long time and finally found me.  He told me he wanted to talk to me \"alone\".  So, we went to this place, I assume it was his house.  He started asking me many random questions about my life and what I had done, etc.  I told him I was happy now and that I never knew I could be happy without him but that I was.  He kept pounding me about, well, when we were together, was this real, was that real...etc.  He wanted me to assure him that I had truly loved him.  I told him, of course, I was in love with you, I never would have gone through all the bs if I hadn\'t but it isn\'t the same anymore...and it never will be the same.  I told him I was such a different person without him and that I was with someone who inspired me to continue growing.  Enter hubby...hubby came to find me because he had heard I was with Mr X.  Hubby was crushed that I was with him and assumed that I was there because I wanted to leave him and go to Mr X.  I tried explaining to hubby that I was simply trying to explain to Mr X how much happier I am and how he and I were never going to be together again.  Well, in the background Mr X is getting more and more upset.  Finally, I convince hubby to leave and let me finish the conversation with Mr X and that I would be coming to him shortly.  Well, Mr X told me, no, I\'m not going to let you leave me again.  And then I was dead.  I didn\'t see myself die so I have no idea what happened...weird, huh?  The rest of my dream was me, dead, telling hubby how much I love him and how much I appreciate everything he has ever done for me...  Now, that I relay it, it is kind of a sad dream I guess.  But when I first woke up, I was so relieved by it.  Finally...I realize in my subconscious that I have made the right choices for me and I\'m exactly where I need to be.  I have been so haunted by Mr X for years...and here, after nine years together, I finally get it right...too long in coming.  I am not going to read too much into the rest of it...just take the peace that Mr X should not be haunting me anymore.  '),(19446,518116,'17,May,2004','2004-05-17 00:00:00','   Our weekend was good.  We got to go see Lis in Chicago.  It wasn\'t warm but it wasn\'t raining either.  We were at the McCormick Center which is right on the water, it is just south of the aquarium and museums at Soldier Field, so it was pretty windy.  It was so much fun!  They have these huge tents set up where you get to drive all of their cars.  Hubby\'s favorite was the little two seater Crossfire and realistically, I really liked the Pacifica.  I did get to drive the Volkswagon Bug.  Uh-oh, not good...hubby could barely fit in the backseat!  The guy who rode with us pushed his seat all the way forward and Scott still was scrunched back there...lol!  Plus, it was a convertible so I doubt he could have made it with the top up!  I guess I will be waiting a LONG time for my bug...haha!  Anyway, they had all these courses set up for you to drive and see the different features of their cars...of course, B and I both loved the part where we got to hit the gas and slam on the breaks at the last second...lol!  We were there for about four hours, driving the different cars, watching a cigar rolling demonstration, just hanging out with Lis and her friends.  Then we drove out to Oak Brook for our favorite pizza...yummy!   Sunday we just picked up the house, did a little yard work, and hung out.  My friend and her son came over and we took the kids to the carnival that was in the parking lot of Wings.  We were there for several hours as the kids rode most of the rides.  B has about 6 inches to go before he can ride the adult rides...boy was he upset!  He wanted to ride the Ferris Wheel so badly and all the other \"adult\" rides.  It was so funny!  They had one of those huge slides where you walk up this incredibly high staircase and slide down on something that looks like a potato sack.  Well, he was tall enough to go alone so I told him he could do it, thinking he would never make it up the stairs.  It took him about three minutes to get up the stairs.  He kept stopping and just standing there and then he would start back up.  I was just laughing because it looked like he was trying to convince himself to go up...after he went down once, he was hooked.  His friend wanted to go after watching B do it but backed out once he was on the top...B just said, \"I\'m not scared of anything!\"  But when he was relaying this all to hubby he tells him, \"The stairs made me a little worried but then I remembered I\'m not scared...\"  too funny!  The two kids did not get along and I was getting pretty disgusted.  My friend was NOT policing her child.  I understand B was misbehaving as well, but I was all over him when he was naughty.  She was just letting it continue...her kid was calling B names and being really mean.  Finally, I had enough.  I know it isn\'t appropriate to parent someone else\'s child but my child was being hurt in the process.  So, I started telling him to cut it out.  The final ride of the evening, this kid hauls off and punches B.  I yell at him to leave his hands to himself.  Not two seconds later, this kid is slapping him.  I yell out, \"C! Get your hands off B NOW!\"  At that point, my friend says, I think it is time to go home.  YOU THINK???  She had driven us so they had to drop us back at our house.  While we are enroute, C says, can we go and play at their house?  She tells him, we\'ll see.  Hell no!  So I pipe in, B and I are going to be taking naps when we get home.  I was not about to invite this kid into my house to beat on B some more.  I know, I know, I sound like one of those overprotective moms...but really, this shit had gone on for a really long time and I didn\'t feel I was protecting B by allowing this kid to act like a maniac and yelling at B when he would start doing the same thing.  Needless to say, it will be a very long time before we get together with them again!  anyway...I finally had a dream with some closure...I will relay it later.  I\'ve been avoiding work for almost two hours now!  '),(19447,518116,'12,May,2004','2004-05-12 00:00:00','   I don\'t have anything important to say, but I am desperately trying to avoid going to work...  Well, it seems to be working for the moment.  I have been taking a fish oil pill before I go to bed and I seem to be okay.  It is really amazing how much better I feel.  I was so sick for the past few weeks and was really getting down in the dumps about it.  I feel so much better!  I think I will do this for a week or two and then try two pills again.  I\'ve been feeling so good that I decided I need to get on the ball with excercise.  I hate saying this because I always flake out but I really am trying.  I keep envisioning myself on that cruise ship fat and presto! there\'s some motivation!  I\'m trying to build up a tolerance to running.  I know that walking is really good for you but the results come so much slower.  I would much rather just beat my body into compliance and then walk to maintain.  Not asking too much, eh?  I can run an embarassing amount of time before I a)can\'t breathe and b) my back feels like it is going to break.  Well, I guess I need to go back to the dr for my inhaler...ugh.  Anyway...hubby is in training this week so B is in daycare all day and then we are all home together at night.  Wow...it is so nice.  I feel guilty about B but spending time as a family at night is really cool.  It makes me hopeful for when he is done with school and hopefully can bid for first shift...I didn\'t realize normalcy felt so good...  Thursday is the award ceremony at work.  I\'m so glad I\'m feeling better...this gives me a better chance of not crying when hubby receives his award.  I guess this award symbolizes all the sacrifices we have made as a family, as a couple, and on individual levels.  It also allows him to close some doors on the trials he faced last year.  I\'m just so proud of him...  See?  I warned you I had nothing important to say!  I fell asleep on the couch watching \"America\'s Next Top Model\" update show...damn!  I really wanted to see it!  Maybe they will replay it someday...  '),(19448,518116,'10,May,2004','2004-05-10 00:00:00','   Wow...Blogger has really changed this up...it looks so cool and, hopefully, will be easy to use!  What a good weekend!  Nothing spectacular but nice...Friday, we went to SofS to get a new title and transfer my plates...then I went home and watched Oprah.  We didn\'t do a whole lot...actually, I\'m pretty sure I zonked out on the couch by 7-7:30pm!  Thankfully, I woke up around midnight to wake up hubby and get us up to bed.  Saturday, hubby had training all day so it was just B and I.  I was able to get a lot of the basement cleaned out, yippee!  We also stopped by my parents house for some hostas my mom was breaking up.  Lis was there...thank god they didn\'t notice her tongue ring while we were there!!!  That would have been pure hell...and, thankfully, they are going to see her on Sunday so we won\'t be in Chicago at the same time either...I\'m dreading the phone call when they do find out!  It didn\'t look gross at all, but for some reason I was completely freaked out by it...  Sunday...B and hubby gave me my presents...too spoiled!  I got the cd player for my car I\'ve been bugging him about...mind you, I didn\'t expect it for mother\'s day!  I just wanted to use the extra cash we had burning up for it...I also got a window box and another big planter...that means flowers!  Hubby was funny...in the morning before I opened them, I showed him the cd player at Circuit City for sale.  He says, I guess we better save our pennies.  HA!  I never saw it coming!!!  Then, when I opened the planters, he says, yes, I am well aware this means you need to buy MORE flowers...lol, he knows me so well.  Oh, yeah...they also gave me this beautiful orchid corsage.  Now, I am not the corsage type, so when we went out to eat, I put it in my hair...B loved it.  He kept asking if boys can wear flowers in their hair because \"it\'s so cool\"...so, I promised him I would wear it to work today so everyone could see it too...I love my boys!  Back to the grindstone...I decided to go drug free this weekend.  I just mentally could not take another day of being exhausted and nauseous all day.  Surprisingly (PLEASE NOTE SARCASM), I felt fabulous all weekend.  Hubby was concerned because he knows I will pay for it once I start back up(today!) but I really needed the mental reprieve.  I felt like a new person this weekend...gone were the mood swings, gone was the exhaustion, gone was the nausea...it was like the old me was back...I miss her...  '),(19449,518116,'07,May,2004','2004-05-07 00:00:00','   People annoy me...the older I get, the more introverted I become.  I am really starting to believe it is because I am so easily annoyed with people...  case in point:  I have a friend that is getting a divorce.  She decided she didn\'t want to be married anymore, she wants to live the single life.  Mind you, she just moved back in 9months ago (she had lived a year of separation with her hubby).  I\'m not trying to judge you but what do you mean you want to live the single life?  She has fooled around on her hubby, she had a guy waiting in the wings for her...just say, I want to explore these other options.    Anyway...this new guy tells her he will call, tells her they will do lunch, etc and just doesn\'t follow through.  9 times out of 10, he\'s not following through.  Now, she wants to piss and moan that he isn\'t following through (oh yeah, he has a wife and a baby too--though, I think they are separated, but not divorcing).  Finally, I casually suggested that she should just write him off.  I asked her why was she so drawn to him, they have only gone out half a dozen times.  She asked me what I meant...to which I responded, \"My little sis has taught me a lot about dating, should I ever be out there again.  If the guy doesn\'t treat you right, if he doesn\'t act like he should, or if the chemistry isn\'t there, cut him loose.  Don\'t beat yourself up about it, it is simply two people that are not compatible with each other.\"  Which should be common sense, but I know for a lot of girls (myself being one of them) it really isn\'t...our emotions take over and the rational side of our brain goes on vacation.  It\'s not that easy...is what she responds with.  I agree, if you aren\'t used to that, it isn\'t easy but it is the right thing.  So she says \"why do you think I\'m so drawn to him?\"  Okay...lesson, if you ask me my opinion, I will be 150% honest with you....I try very hard to deliver it tactfully but I can\'t make any promises, therefore, if you don\'t want complete honesty, DON\'T ASK!  To which I responded, I don\'t know if this is true, but from what I have seen, you seem to like the falling in love part and the new infatuation...but once that part of a relationship is over you seem to check out too.  Yeah, I\'m sure you can imagine, she didn\'t like that so much but she said, You\'re right.  I never thought of it like that before but it is true.  Dude, I\'m really not out to be right...I just don\'t understand why you must bitch about something you are completely bringing on yourself.  SOOOOO...she goes into how she and one of our friend\'s ex-husbands are talking and she thinks he likes her...this is the SAME friend who dumped her best friend because she started dating her ex-husband.  She is actually contemplating what she should do!  She was the one who brings up, I don\'t want to be like C and steal someone\'s ex but what should I do?  Is this really a question???  You just answered yourself!  DON\'T!  You couldn\'t handle it so why would you risk a friendship expecting someone else to???  Lordy...stupidity bugs me...  Other than that...CP and I are planning a trip to Chi-town for the end of the month.  Nothing major, just some shopping, some clubbing, and then home again.  We\'ll stay over one night but I am so excited!  K may come with us, though she gets very intimidated by large cities...so I don\'t know if she will chicken out or not.  I so need a day away from everything...only a few more weeks!!!  '),(19450,518116,'04,May,2004','2004-05-04 00:00:00','   Isn\'t life strange?  The past week has been a continuous cycle of stress...and, now, that it is all over, I just want to forget all the stress, find the good, and move on!  We put Spike down.  Please, be kind with the judgements.  He was destroying the house and we just couldn\'t figure out a way to solve it.  I miss him, he was my puppy and I love him so much.  At the same time, I couldn\'t argue with hubby about how much stress this was causing and how angry it was making him.  Namely, our brand new carpet is now destroyed.  I guess carpet is much more satisfactory to tear up than hardwood floors?  He has not torn at flooring before.  Okay, correction, he ripped up some vinyl when he was new but nothing since.  So, he suffers from extreme nervousness.  When someone other than hubby, B, and I come into the house he freaks.  He will run and hide under the bed if our room is open and bark his head off.  Seriously.  For three hours one time because I couldn\'t pull him out.  Or, if I anticipate someone coming and close off the doors, I lead him back downstairs and he crawls into my lap while people try to let him sniff their hands.  Anyway, I am trying to justify what we did, I guess.  We knew that because of his fear he would be so scared and people would have little patience so we decided to put him down than put him through that.  I think it was the right thing to do although a few of my friends have voiced their adament disapproval.  Ironically, anyone that knows Spike agreed with us, he wouldn\'t have made it if we sent him away.  Regardless, I miss him very much.  It is better this week...part of me has acknowledged he is gone so the constant thoughts are winding down.  My inlaws...I don\'t even want to go into this.  I have spent the whole weekend arguing with hubby about it all and I\'m tired of the control they have when we are trying not to let them have control.  As I told hubby, I am done seeking them out, I am done begging for a relationship.  They can live in TC with each other and live their twisted lives together.  I will continue to follow through with my obligations as their daughter in law but no more than is necessary.  The silver lining?  The shower is over...now, I know I will probably have to go up there once the baby is born but after that...I don\'t see any reason to interact until we go to the races...potentially two and a half months of freedom!!!  THAT is something to look forward to!  A plus side to this whole nightmare of a past week, I got a new car.  The Barto-mobile has been retired and I miss it.  However, there is nothing like the comfort of knowing when you turn that key, your car will respond appropriately without any screeching or squealing.  We were looking at a new truck for hubby since I cannot drive his stick and that way, when my car did break down, we would be able to share a vehicle until it was fixed.  Well, hubby decides since he is getting a new vehicle it must have all bells and whistles.  BUT it did not bother him that he wanted to buy a \'97 Expedition...maybe I\'m wrong but wouldn\'t it be beneficial to buy a younger truck with less miles and less features???  It won\'t mean a thing that your vehicle has heated seats if the stupid thing won\'t run!  So, everytime we would go look, we would get mad at each other.  Finally, I told him I was done looking.  We could continue driving the same vehicles until one died and that would be the one we replaced.  Now, don\'t get me wrong...hubby has been nagging me for two years to get a new car.  My car was paid for and I liked not having a car payment so I held off.  He won.  He suggested I look online for a car instead of a truck.  I found a car in a different state that was extremely low mileage and a great deal.  Plus, he took my car for a trade in.  He is also a cop so, of course, for hubby that sealed the deal.  It has been strange getting used to the new car but I really like it!  Other than that (!), this week looks promising!  B has an appointment with the etn dr today.  I am very hopeful this is finally the right specialist to tell us what is going on and how we can help him...my health is haywire as expected.  I am praying this is the right dose because I\'m not exactly sure I can take going through all of this again in six weeks.  Did I mention my oldest sis is pregnant???  YIPPEE!!!  I\'m so excited for her...and, now, I am trying to figure out what to do...go to NY before baby is born, after, next year, both???  Decision!  The baby is due at the end of November so maybe I should go at the end of the summer and then the beginning of next year.  Plus, I am wondering if she will be coming home before then so we can throw her a shower!  '),(19451,518116,'30,June,2004','2004-06-30 00:00:00','   man!  Do you ever get nauseous?  I swear, it is terrible!  I don\'t know what the deal is but it has been happening...it hits mid-afternoon and I just feel exhausted.  I feel so sick to my stomach and eating doesn\'t help and not eating doesn\'t help...yuck!  Monday night I put in a frozen pizza (very gourmet!) for B and I...well, he was watching \"Brother Bear\" and I passed out on the couch.  I never heard the buzzer so I woke up to smelly burnt pizza.  Thankfully, it wasn\'t burned that bad and B ate it but still!  I don\'t know how I slept through it...when B asked what happened I said, I must have not heard it while I was sleeping...he says, you must have been snoring so loud that I didn\'t hear it either.  HAHA!  So who knows what the deal was...but same thing last night.  Thankfully, I thought before we watched \"Brother Bear\" and didn\'t try to pre-make our gourmet meal.  Anyway...I am not so melancholy today.  I forgot something funny my stylist said yesterday.  She asked my if hubby knew I was coming and I said yes...is he excited?  I think so.  She says, \"Oh, you know he so is...it is every man\'s porno fantasy.\"  I was laughing so hard...and yes, uptowngirl, I COMPLETELY agree...next time it is all gone.  She was just telling me she has not done much of that and the salon is having her attend more classes as it is becoming more common...so I figured it could wait until next time...lol!  So, perhaps in a couple months?    and yes, I think he was excited...he came home from work, skipping his running/workout, and tried to wake me up...I guess I was out cold so he just stripped me and I woke up to him on top of me...ahhh, what a beautiful way for a nymph to be woken...  '),(19452,518116,'29,June,2004','2004-06-29 00:00:00','   I\'m in a melancholy mode...who knows why...just am.  Work is weird.  And to top it off, I have yet to hear if I was extended for July...okay, we are talking the day after tomorrow!  Ugh.  I\'m not stressed about it, just annoyed.  I think if my PM\'s were fretting I would get worried...I\'m sure they would drop massive hints if something was up.  I went and got a wax today...I thought it would be a \"full removal\" but I called in and asked for the Brazilian...I thought that is what my older sis said that meant...guess not.  It means one finger width is still left...damn!  I was really looking forward to being hairless.  Next time.  Oh, yeah...I also had to buy some cream because I am VERY prone to ingrown hairs...$34 DOLLARS!  15ml for THIRTY FOUR DOLLARS!  I am shocked.  My waxer offered me the small packet but I took the big tub, I mean, who knew?  So this venture cost me $80 after the tip...that will thrill the hubby!  Tomorrow I am going to get a pedicure and then next week I will get a fill in my fingernails...who knew being a girl was so much work?!  I have been going tanning almost everyday...oddly enough, I feel tan today...  Will my period come?  I need to do the monthly dance...don\'t get me wrong, skipping is cool...but I know I won\'t be that lucky.  If I don\'t have it before we go, guaranteed it will start while we are gone.  Yucky!  I\'m going to go listen to Karen Carpenter...I\'m in a mood...  '),(19453,518116,'28,June,2004','2004-06-28 00:00:00','   ahhhh...what weekends...I love them...but it makes Monday mornings especially brutal...  Friday...B was at VBS all week and Friday night was parents night.  Both hubby and I went.  First a bbq and then watching them.  They sang and then there were \"Olympic\" games...what a cluster!  The leaders didn\'t know what they were doing and the games were not entertaining even the youngest of kids!  It was an hour and a half of torture...we were SO glad when the games were over...but it was fun to see B so excited...he was so happy we were there and got to \"see\" him do his VBS thing.  After that, we ran to Sears to do some shopping.  Got B some clothes for soccer, hubby a swimsuit and some sandals.  Saturday morning we got up and went to Bob Evans for breakfast...yummy!  Then we continued shopping...yuck!  We went BACK to Sears...hubby didn\'t like his swimsuit and I exchanged some flip-flops at Ann Taylor.  Then we went on to Dunham\'s to buy B some cleats, shin guards, and socks for soccer camp.  Hubby got a swimsuit there...then on to Old Navy.  Hubby got another swimsuit and a sleeveless shirt...I found a tshirt.  Finally, we went home.  I told hubby, shopping with he and B is like shopping with two ADHD kids...they were constantly, \"Sar look over here, mommy look what I found, mommy/Sar, mommy/Sar\"  It was driving me mad!!!  I basically crashed when we got home...seriously...I was exhausted...got up, went tanning.  Why aren\'t I getting tanner than I am?  Do you stop having the ability to get tan the older you get?  It is very strange...after that, we got ready and went to M & J\'s housewarming/annual bbq party.    Their new house is beautiful...and a beautiful backyard and gardens as well.  M hired a woman to do facepainting and balloons for the kids...B had his face painted like a rainbow tiger, it was beautiful!  He looked like he was performing in CATS...seriously, it was awesome!  One of my friends, JL, was there...I haven\'t seen her since they moved to Ann Arbor.  She and M kept going on and on about how beautiful I looked...it was so nice to hear the first time but then I was just embarassed.  M says, you look like a movie star!  You look so awesome!  Then she would say to whoever was around her, isn\'t she just beautiful?  Her hair?  Her face?  Her whole package?  Just beautiful!  Seriously...I was mortified...  C was TERRIBLE!  He kept screaming because the kids were playing with his toys...duh...and, of course, M did nothing...J did nothing.  It was embarassing for them...and maddening for every other parent.  Hubby was so pissed by the time we left...I guess C had hit and pushed B a couple times...I only saw one push and I was all over him.  Hubby just could not believe M&J let him behave like that...it changed his opinion of them...I had told him this is how it was but he had never seen it before.  I doubt he will want to do things with them anymore...  Sunday...more sleeping.  I got up around 6:30am to make a list of what we need to do to sell the house...holy crap!  Not a bunch of \"big\" stuff but a full page, nonetheless!  We talked it over and decided to wait at least another year, probably two.  We have been looking at houses on a lake or have deeded lake access.  Hubby has always told me there was no way we could afford it...well, we can.  Not that it will be big and immaculate but the ones I found are a little larger than we have right now.  Definitely a realistic dream...I mean, if we don\'t have anymore children, we don\'t really need a larger house.  I would like a little bigger and would like a different floor plan but I am sure we will sacrifice some of our \"wants\" to be on a lake...both of us have ALWAYS wanted to be on a lake.  Man!  I can\'t wait.  But this will give me something to strive for...and hopefully help get my spending under wraps again...lol!  Made a big breakfast for all of us...I don\'t know what my deal is...I just want to eat pancakes...lol!  After that, I went to take a nap...then we did some quick grocery shopping and I bbq\'d, or tried to.  Something is up with our grill because it was NOT cooking the meat.  I ended up having to use the George Foreman and hubby left late for work...ugh.  B and I watched \"Brother Bear\", I rearranged the kitchen counters, and that was the extent of my fun for the day...  '),(19454,518116,'23,June,2004','2004-06-23 00:00:00','   yippee!!!  I learned how to inhale!  Today was my first day back to running after getting an inhaler.  The dr said I need to have two puffs and wait twenty minutes...then I should be good to go.  Well, last time I had an inhaler, it did not work.  I could NOT inhale...sounds stupid, but it is true...same reason I couldn\'t learn to smoke in college, can\'t inhale.  Ironically, same reason Clinton explains he couldn\'t do pot or whatever drug, he could not inhale...anyways.  So, I wake up, psych myself up and did it PERFECTLY the first puff...it was so weird because I felt it.  The second puff didn\'t go quite as well, about halfway right...but still, quite progress!  Now, after taking a week off, I need to work up to where I was again, but I made it pretty far considering...probably tomorrow I will be able to pass this level and go to the next step.  Mary Kate...wow...can\'t say that I am surprised.  That girl was looking bad, worse than Lara Flynn...I don\'t really have much to say on the matter, just that I hope she is successful with the treatment and that she gets better...sending her all of my warm thoughts, along with many others, I\'m sure...she is such a beautiful girl...I hope she learns to see that too...  '),(19455,518116,'22,June,2004','2004-06-22 00:00:00','   do you like to look back at how you have changed over the years?  It is pretty funny...I wouldn\'t say I was ever a \"tomboy\" per say...but when I was in high school I so did not care about \"girly\" things.  Sure, I liked to be all glammed up but I wasn\'t going to commit the hours that my \"girlfriends\" did to do so...I also didn\'t care about appearances...I wouldn\'t not play ball with my guy friends because I might get sweaty or messed up or dirty...  Now, I would say in the last couple years, I am turning into a complete girly-girl.  Not that I don\'t do things because I might get dirty but I LOVE pink (AHHHH!), I don\'t mind spending money on girly things anymore...actually, as of late, I have been in complete girly mode.  It started when C and I went to Chicago.  I got to go into Sephora for the first time.  Other than the madness of all the people, I LOVED it...but I couldn\'t decide what to buy after about 35minutes so I stopped torturing C and just left.  Now, I wish I had gotten their las vegas makeup bag.  Too cute!  I could order it online but I think hubby is about to kill me with all the purchases as of late...anyway...so I have been obsessing about girly things...add in my extra hormones the past week...  I was in Target yesterday filling my inhaler prescription...I got some cool \"tools\".  I got cuticle clippers (I have been DYING for some for over a year!), eyelash curlers (I decided about 3months ago I needed some), and, finally, my new love, a callous shaver...I don\'t know if that is the technical name but when I got the pedicure in Chicago, this woman used one to scrape off the nasty skin on the bottom of my feet...I was instantly in love!  I have been keeping an eye out for one but sephora.com was out and I hadn\'t found any in the stores.  I can\'t wait to use it!  Also, I have been wanting to stock up on more makeup.  Sadly, I have used the same types of makeup since college!  Appalling...I know.  SO, my friend started selling AVON and I decided to buy some stuff from her.  I know, not the highest quality but I need to figure out what I can deal with before I drop a buttload.  Also, I went to Marshall\'s and got a bigger makeup bag that will hopefully hold me over until I find the perfect big one.  THEN I went on eBay and bought this CD...it is put out by MAC and is supposed to show different ways to apply makeup.  This is what I have REALLY needed...the nice girls at Merle Norman tried to help me when I worked across the hall years ago but I don\'t really remember everything...plus, I would like some bold new ways to wear makeup...I cannot wait until it gets here!  Why am I feeling so girly?  In anticipation of my brazilian wax in a couple weeks?  Oh yeah, I also went and got an overlay on my nails last night...I\'m a walking Hello Kitty...  '),(19456,518116,'21,June,2004','2004-06-21 00:00:00','   do you ever feel like you are two steps from being committed???  that\'s me...I always over-analyze myself, and, of course, it all takes place in my head...so, yeah, I think I\'m nuts...  anyway...do you know how many pregnant people I know???  I know a bunch...I also know a load of people buying new houses...but that is not what I\'m talking about right now...my older sis, her sister in law, her friend, my friend T, my PCOS friend A, and now S...not to mention all of those I know that JUST had babies.  It is weird because I would have thought that I would have issues with it...not really...I don\'t want to say that I am above that because that is NOT what I mean...I just mean that I haven\'t allowed myself to think there is even a remote possibility so I have moved on to find other meaning in my life beyond having babies.  I\'m glad I feel this way...it isn\'t like me to be jealous of what other people have so it was a double whammy when I got all depressed AND was jealous...I just didn\'t like myself so much then because I felt I was just being bitch-y.  This weekend was nuts...seems to be the story of our life.  I feel like we are trying to cram in everything before we go on our cruise!  Friday, my parents had B overnight...hubby and I met his coworkers at Don Pablos for dinner and then went to Gary Fields Comedy Club.  We had never been there before and it was HALARIOUS!  I have not laughed that hard in a very long time...hubby was so nervous going into it because he thought I would not like his coworkers...I had met them at R\'s graduation party and they were so nice...but I think he was worried they would talk about work the entire time and I would feel left out.  They did talk about work A LOT but it doesn\'t bother me...I am so used to listening to hubby\'s stories and not having a clear understanding of everything he is telling me.  Plus, R\'s wife, J, is so sweet...they both are just very nice people.  We had a great time with them which was wonderful because since B and H\'s divorce we haven\'t really had a perfect match as couple friends... I take that back, we have T & J but they live in Indiana and we don\'t do things without the kids.  Anyway, J said she will find the schedule and we can plan a next time...I think each side passed the \"friends\" interview...lol!  We went to the early show with R&J so we got home a little after 10pm.  My sis has loaned us the Harry Potter movies so we decided to watch the first one.  I fell asleep for about 1/2 an hour but we both really like it.    Saturday we both slept in, got ready, and went tanning.  The owner was there so we talked to him about the best way to get hubby tan in three weeks, not going everyday.  He was really nice and gave us a great deal on a package and some new tan lotion.  He talked to me about how it was going for me since I come everyday...I told him, for some reason, I still wasn\'t getting much color!  That\'s when he sold me a different lotion that has accelerator in it and upgraded my bed.  He says, you are going to feel it burn but you aren\'t REALLY burning...  Well, he put me in the ultra-bed for 15 minutes...I was still getting pink at 10minutes in the regular bed!  And, boy, he wasn\'t kidding!  That shit made me hurt!!!  It burned and was on the verge of hurting..it was definitely NOT comfortable!  Then, I got out and everywhere I had put the lotion on was bright cherry red...it looked like I had fried!  The good news is, I did indeed get a lot of color...the bad news is, there are places that really did burn.  Like the back of my arms, the top of my thighs, the sides of my calves...very weird but whatever...I didn\'t go back yesterday but I think I might be okay to go today...  After tanning, we picked up B.  Then we headed to Cabela\'s.  It took a little over two hours but it was such a beautiful afternoon.  We spent three plus hours there and both hubby and B had a fabulous time...it was not as painful as I thought it would be but I definitely could have left at least an hour earlier!    Sunday was just lazy...I got up and went grocery shopping...it took forever because I hadn\'t gone for a couple weeks.  I have decided, I need to find a farmer\'s market.  I spent the majority of the grocery money on produce, including the $6.50 I spent for a bag of cherries!  Unbelievable!  So, I\'m thinking I need to find somewhere to get some of the fresh produce cheaper.  Plus, it would probably taste better because it would be extra fresh!  Anyway, we had a bbq for hubby for Father\'s day and that was it...I crashed on the couch after lunch.  I was so tired...it is funny because I get irritated when I have to remind hubby that I can\'t go balls to the wall all the time and sometimes my body just crashes...but I forget too.  I slept for about three hours!  Thankfully B wanted to watch a movie so he was content just laying on the other couch while I was dead to the world...  Well...my boss is ooto this week so I guess that is why I felt the need to ramble so long.  I don\'t have much to do and am DREADING this week...I want to work extra so I have overtime for extra cash for vacation but I don\'t know if I can convince myself to stay!  This afternoon I have to leave a little early, I have my hearing test...the results should be interesting...  '),(19457,518116,'17,June,2004','2004-06-17 00:00:00','   okay...so I decided to switch it up a bit...I\'m sad that all my comments surely got lost...so get commenting again!  '),(19458,518116,'16,June,2004','2004-06-16 00:00:00','   lordy!  Please know, I do not purposefully put myself in the place of surrogate mother to my younger sisters...and sometimes my older one as well.  I have always thought that everyone thought I was crazy...I\'m not convinced that my brother in law and older sis don\'t but I finally feel like maybe my sisters actually appreciate my brutal honesty...   when I had the infamous \"talk\" I definitely felt ostracized...I mean, how DARE I suggest that they may engage in the demons of alcohol or premarital sex...but at the same time, I didn\'t care...I NEEDED to say what I said for my own peace of mind.  Please know, I do not blame my parents for hubby and I getting pregnant before marriage, obviously, I made the decision to have sex.  BUT I think if I had not been as \"responsible\" and \"mature\" as I was, in some ways, I would have been a basketcase.  I think this is my issue with right-wingers...just because you teach someone safety doesn\'t mean they ever use it...have YOU ever used a fire extinguisher?  Do we NOT teach how to use one because that would just be inviting people to start fires?  OF COURSE NOT!  In my mind, birth control and alcohol awareness is the same thing...please teach kids the benefits of using a condom IN ADDITION to using birth control...please teach them about AIDs and other not so fun STDs that they are risk of getting...please teach them about alcohol...etc...it is important shit that if they aren\'t taught they will learn from experience, and perhaps, after it is too late...you can\'t just return AIDs to the store from whence it came...my point, I guess, is that if someone had been more honest with me about sex and alcohol and drugs, maybe I wouldn\'t have been so naive...maybe I would have...  Well, lil sis came to watch B for hubby and I...my mom usually has him in the summer on Wednesdays but since my uncle died, they were in Manistee.  Well, I get home from work and am talking to her as she is leaving...all of a sudden I see it, a HICKIE!  I was so shocked!  Now, I am even more shocked that I ever saw it...like I told hubby, it wasn\'t like I was scoping out her neck, it just isn\'t something I would have expected...plus, it really was not that big and she had makeup over it...well too bad for her!!!  It was pretty funny because I then went into my 50questions...who? what? where? when?  Of course, La was having none of it.  She just turned beat red...it was pretty halarious because it was like she saw my face the second I noticed the hickie and she turned beat red...so, I went to give her a hug goodbye and put her in a headlock...that\'s right...I need to patent this move!  She literally was stuck smashed in my face for about 15minutes until I was laughing so hard I let her wriggle free...she tried tickling, which I\'m not ticklish...she tried yanking and pulling but it was a HOT move...still didn\'t get info out of her but, as I told her today, it is good to know I can kick her ass!  Now, I just need to defeat Lis and I\'m back in the number one spot...trust me, I\'ll use my new signature move, it will be the only way!  Anyway...I called her this morning to talk to her.  I wanted to make clear I am not disappointed in her...I mean, in the grand scheme of things, it is a few broken vessels...plus, in the back of my head, I know I have gotten them from NOT so innocent things...I\'m not disgusted...I\'m not really anything.  I told her again...please use a condom...please use birth control...but especially the condom...told her I just worry about her and want her to be happy and with someone that deserves her and treats her right.  Asked her if he spent the night (come on, this is ME we are talking about!)and asked about her birthday.  SUPPOSEDLY, she told him no to the ecstacy...I pray she is being honest, I don\'t know that she would lie to me but at the same time, I don\'t know that she would want to admit that one either.  Told her I love her...told her I was available for a beating anytime she needed it!  HAHA...yeah, she loved that one!  Almost as much as when I ask her if this car is a keeper!  (Note...this girl has totalled TWO cars...within 2months of each other--and 2months is a generous figure)  Did you know the Simple Life started again?  Lordy!  Those girls crack me up...hubby wanted to get very philosophical last week about it...they are so rude, they are disrespectful, blah blah blah.  I apologized but said I did not want to get into a debate...I was trying to watch OPRAH!  Anyway, yeah, I know what he is saying and I agree last season was a little much...but at the same time, these girls have never wanted a thing their whole life plus, I am SURE they are used to being catered to by everyone...I think this explains some of it...tangent!    Sorry...so, I watched both episodes last night...halarious!  I felt bad for Paris...who wants to forever watch on tv falling off a horse?!  But I love Nicole.  I love her new word \"hot\"...ie, no swearing in front of Kayla.  Paris says, \"nicole?\"  Nicole says, \"What? Oh yeah, that\'s hot\"  HAHA!  Cut to her trying to get this little girl to say ass or bitch...too funny!  I have to give Nicole credit...everyone was SO rude about her looks last year...I think she is pretty plus she lets Paris have the spotlight and just rolls with it...it so reminds me of my friendship with S...I told her we need to pitch a spinoff where we roam around the states as two 20-something mothers on the hunt for a good time...I\'m sure it would be just as amusing...  '),(19459,518116,'16,June,2004','2004-06-16 00:00:00','   ugh...why does it seem this week is just dragging?  can it be Friday yet???  My Uncle Fran died this week...well, he was really my great uncle.  I knew him but I didn\'t know him well.  He lives in Manistee and, well, we hardly ever went up there.  He had cancer so I am glad the pain has stopped.  I\'m not going to pretend I am devastated because I really don\'t know him...but I always feel sad when another family member dies.  It reminds me how little I know about my extended family.  It started a conversation with my mom about my dad\'s side.  Who was related how...it also brought up that my dad\'s side has the possibility of the cancer gene as well.  Well, that is great to know...my dad won\'t go get tested even though his dad, his grandfather, and now his uncle have all died from cancer.  His grandfather and dad had prostate cancer!  How I never knew this, I don\'t know...but I think it is VERY important to know for B\'s sake.    It just has me thinking about all the probability that I have the mutated gene.  J escaped but I have always been the mutant health-wise in the family.  So why should this be any different?!  I know I haven\'t been prepared thus far to find out...but maybe next year.  I am at the beginning of a road...I feel good.  Yeah, I know my body is still messed up but I feel that I have finally adapted a lifestyle in which I am doing all that I can to help myself.  I am taking my meds almost religiously (I forget once in a blue moon), I am watching my diet and eating mostly fresh veggies and fruit, and, the kicker, I am excercising regularly.  Other than adding in a multivitamin, I\'ve done my part...now the meds need to start doing their part and my body needs to wake up...  I also am back at a place where I am not obsessing about baby...do I still want another one?  Most definitely!  But, I\'m not beating myself up nor am I mourning something I may never have...I\'m back where I can say, if it is meant to be, it will happen.  I\'m back to the place where I can see how much I love hubby and B and know that our family is such a blessing and more than some people have in a lifetime.  I\'m content with the way things are...    Sometimes, I think by not taking care of my body, I was hoping a baby would come and then it wouldn\'t be that big of a deal..does that make sense?  I just feel that if I whip myself back into shape and THEN get pregnant, I will end up totally out of whack again.  Having B sent my body over the edge and I don\'t want that to happen again...obviously, if I did get pregnant, they know what I have now and I am on meds for it so it may not cause havoc...but it may.  I guess that is selfish of me but if I\'m doing all this work, I don\'t want to end up back at square one or at square negative fifty!  Anyway, my whole ramble was started because, I think, if I continue on the right path, I will just get stronger and stronger.  It has been a year since I found out I have PCOS, so next year will be two.  If I am still mentally healthy in dealing with it and not just shoving it on the back burner, I think I will be mentally prepared to find out about the gene.  It is weird...I have always assumed I have it so I don\'t really know why I want to get tested.  Except, that is one more thing I will need to see specialists about and have tests, etc...I will need to stay on top of it.  My mom read me an article with the statistics...basically, by 60 years old, 87% of the women with the gene have produced cancer.  That is frightening!  I need to have the test done so, if I have the gene, I can have the tests to help keep an eye on my body.  Right now, my gyn told me he can\'t get insurance to pay for anything because they will see me as too young...bastards.  So, now that I am feeling pretty good about the PCOS, I need to start preparing myself to take the test.  I can\'t imagine how weird it would feel if I found out I don\'t have the gene...I have never really given much thought to that possibility...  Enough on the philosophical...B has to have his adenoids removed.  The Dr explained it is relatively easy surgery and does not have the pain associated with tonsil removal.  Both hubby and I agree we need to do this...we do NOT want him to get an infection in the sphenoid again.  I know B will be a trooper about it but it doesn\'t lower the worry factor!  We are looking at the end of July...hopefully it will work out because that will be after we get back from our cruise and before the NASCAR races.  I do NOT want to miss the races...not for me, but B has been told for about a year he could go with his dad and gramps and I know he is excited...I just don\'t think that would be fair to do to him.  '),(19460,518116,'11,June,2004','2004-06-11 00:00:00','   I feel so icky...do veggie-dogs go bad???  I have had some in the freezer for FOREVER...well, last night, I needed something hot to eat.  I don\'t know how people do the microbiotic diet.  I just get to a point where, even if I am really not, I need to FEEL like I am eating a real meal.  It HAS to be hot...or at least have been cooked and been hot at some point before I eat it...  Anyway, a couple nights ago, I decided to have a veggie-dog.  So, I took them out of the freezer to let them thaw from the freezer burn...well, I ended up running and it was so dang hot that the last thing I wanted to do was eat.  So, I just threw them in the fridge.  Last night, I wanted a hot meal so bad...add on top of that, I was starving because I had only had a muffin and a bowl of cereal...so, I am scrounging looking for something to be hot and quick but not blow my caloric intake...enter veggie-dogs.  So, I ate BOTH of them on 100 calorie buns.  Not bad in the diet department.  However, this morning I feel like crap!  I got up with a stomach ache which is now just nausea...I am SUCH a baby when it comes to that...I hate the thought of regurgitating...I will hold off as long as possible.  Right now, it is coming in waves...lordy!  Please let this day go by fast!!!  On an up-note...we meet with the ENT today.  I will be interested to hear why the PA thought B needs his adenoids removed.  I am praying overtime there is another approach...  '),(19461,518116,'10,June,2004','2004-06-10 00:00:00','   OHMIGOSH!!!  We have just gotten the best news in a LOOOOONG time!  Hubby has been planning to go to school in the fall to become recertified.  The police academy lasts two semesters so we have been AGONIZING over what to do...do I keep working?  does B go to daycare full time?  do I quit and stay home?  We had yet to come to a decision because we have been going round and round...  Well, yesterday, hubby had an appointment with the director of the program.  She mentioned he needed to call the MCOLES department.  She thought the certification requirements had been changed and since he was already certified once, he possibly would not have to go through the academy again.  Well, needless to say, he was so excited and hopeful but I tried to stay calm because we both figured, with our luck, it wouldn\'t go that way...  So, he tried to get ahold of someone at MCOLES but the woman he needed to talk to was not available.  He spoke to a guy there who agreed that he did not think he would need to go back through the academy...  So, today, the woman calls and tells him he is good to go!  He needs to pass the physical test and then he will need to attend an 8 day training class.  After that, he will need to take the state certification test and then he will be recertified!  I CANNOT BELIEVE IT!!!  We are both so ecstatic!  I guess after 9/11, the state had to take a look at the requirements and decided if someone had gone through an academy and been certified previously, they should not have to go through an academy again, just the \"refresher\".  Words cannot even describe what a huge relief this is and what a burden of stress has been lifted off both our shoulders...  I am so happy and excited for him...he wants to try to go to the training in August.  It will require some serious dedication and hammering done on his part but I think it is doable.  If he doesn\'t get his stuff done by August than there will be another one in October.  I told him I\'m nervous because that means he could be on the road by the end of the year but I know how happy it will make him...keep your fingers crossed!  '),(19462,518116,'09,June,2004','2004-06-09 00:00:00','   yes, I am done freaking out...the PA called back today...I told her we would prefer to see the dr, to which she said, fine but she isn\'t available until middle to late July.  Fine, no rush.  Well, when I called to make an appointment with the dr, we got an appointment for this FRIDAY!  I am so not impressed with this chickie...but we do like the dr so hopefully she got a clue and won\'t be calling back!  Other than that...not much new.  It was so freakin\' hot last night!  I have this weird thing about trying not to turn on the air until July but I broke.  I was laying on the couch resting and trying to convince myself to go downstairs and run...well, just laying there, I was dripping sweat, I HATE to sweat!  So, I went down for a run and was filling a small pool with my sweat...so, when i came back upstairs I turned on the air.  I am so bad about remembering the cover!  It took me an hour to remember...I realized the air was not pumping in cold air and then it dawned on me...duh!    So, when I put B to bed I told him to just sleep in his undies since it was still so hot upstairs (over 90 degrees!!!)...he says, can I just sleep in my butt...I burst out laughing!  It was so halariously cute!  So the air is on and I am B-less tonight...  '),(19463,518116,'08,June,2004','2004-06-08 00:00:00','   okay...I\'m officially freaking!  B had his allergist appointment last Thursday, it turned up negative for any allergies.  So, yesterday, I called the ENT doctor to see if she had seen his CAT scan and what conclusions, if any, she had come to since he has no allergies.  Well, the bastards called back at work when I had left our home phone (I had to leave early yesterday, which I knew!) and left me a message that we need to schedule a PRE-OP appointment...as in PRE-OPERATION!!!  So nice of this frickin\' PA to leave a message telling me my son\'s adenoids need to come out!  So, now, I have to wait until later this morning to call them back and try to talk to them about it...quite frankly, I want a second opinion...I am not inclined to take your word for it when I have only met you once and this is my only child we are talking about!  And, actually, the PA wants to meet with us, not the Dr...oh, I\'m sorry, the dr is too busy to meet with us unless she is OPERATING on my four year old?!  Okay...I know, I\'m being a bit irrational but I hate dr\'s and I just think this situation could have been handled a lot better...ie, \"Hi, returning your call, we have some conclusions we would like to discuss, please call us back at your earliest convenience!\"  '),(19464,518116,'07,June,2004','2004-06-07 00:00:00','   man, I hate Monday mornings...I don\'t know why because there is no one here to bug me...I think it is just the thought that, had it still been the weekend, I would not be here...but since it is Monday morning, I am sitting at my desk procrastinating at 7am!  Well, our weekend was good...the Greek Festival on Friday and T&J on Saturday.  I forget how much I miss T until I spend time with her...she has got to be the one girl that I could guarantee I will always be friends with.  She knows me so well and she is not into the petty/catty bs other girlfriends are into...needless to say, I hated it when they had to leave.  And K is such a little gem!  She was cracking me up even though she barely said anything!  She is a strange mixture of complete girly and tomboy...it is funny to watch...Sunday we went to see grandma.  Well, I hate to say it but she drives me nutso sometimes.  I know it is terrible...I know...but she lays on these massive guilt trips.  I just want to scream, do you understand I work???  It is major frustration and add to that the fact she is losing her mind and she was telling these crazy stories about B and others about my dad.  I feel bad but I just didn\'t have patience for it yesterday...before everyone here slays me, I am consciously going to try to see her more often and I do love her very much...  On a more not upbeat note...I have had two very deep conversations with both of my younger sisters.  The first was on Thursday night...B and I went to my parents to drop off some baby paraphenalia for my older sis.  They were going to see her NY and I wanted to send some bigger stuff with them so I wouldn\'t have to ship it.  Well, L and I got talking...her opening line???  Have you ever tried ecstacy?  After I lifted my jaw off the table, I said, no...the one thing I have never tried would be drugs.  Well, this loser of a guy she is \"dating\" (mind you, she says HE told her they were dating) wants to give her some for her birthday.  Supposedly, she told him no but he is insisting.  She said she doesn\'t know what to do because she doesn\'t want him to get mad because he spent so much money on it... EXCUSE ME?!?!  So, I just busted out the, have you had sex before?  To which she said, yes...are you currently?  No...I couldn\'t get too much out of her...no condom, no birth control...but I couldn\'t get much else out, she said it really wasn\'t her choice that they didn\'t use a condom...again, what?!?!  I asked her if she was telling me she was raped...she said no...were you conscious? Yes...were you trashed?  I don\'t remember what she said...she said it was a one time deal and she has been beating herself up about it.  Lordy...there was so much I needed to say to this girl...considering my parents were lurking, I knew my time was limited...so I started...you need to wear a condom, not only will it protect you from pregnancy but it will also protect you from most other std\'s that you SURELY do not want, not excluding AIDs...drugs, I understand people do them, I understand people think they are okay...knowing MANY people who have done a variety of things, I don\'t buy it...God only gives you so many brain cells and why waste them...not to mention, you don\'t know who got what where and what they put into it...unless you are growing it yourself, it isn\'t 100% assured it is what you think it is...not that I am advocating growth, because that will get you locked up...as for beating yourself up, if you think you made a mistake, step back, analyze what happened and why you did it, and learn from it...life is too short to live with regrets and all we can do is live life the best way we can...  Well...obviously, she is beating herself up because of what my parents would think.  I told her, there has to come a point in your life where you look at what mom and dad taught us and accept and reject the principles to adapt your own...you cannot live your entire life for mom and dad because, no matter what, they will find something that could be better.  She completely agreed but said that it was a lot easier coming from me considering I never cared what they thought of me...I kind of laughed and said, well I don\'t now but I did...in high school I drove myself nuts trying to please them but I was coming behind the girl they deem \"perfect\" so, like I said, nothing was going to make them happy...when I graduated high school, I gave them the proverbial finger and went about my own business.  It wasn\'t until I was out on my own in Chicago that I completely made all of my decisions.  On top of that, I basically showed through my actions that if mom wanted to lay guilt trips, I would have nothing to do with her.  I think she finally got the picture that I was never going to be her idea of what I should be and gave up...to move on to the next sister, I\'m sure.  I also told L that I would be lying if I said I still don\'t get hurt by it...I told her, it is hard to listen to everyone else\'s shining moments and then, here\'s B.  I don\'t even get introduced, B does.  It is like I don\'t exist except to be B\'s mom...it hurts because everyone else\'s parents thought I was the berries...I got good grades, I kept clean, for the most part...but my parents thought I was the spawn of Satan.  Then my dad came home with the dinner...I need to get together with her to finish it up...  Then, Friday, I went out to lunch with Lis.  She started talking about L and how she thought this supposed boyfriend of hers was a real something or other...I don\'t remember what she called him but it cracked me up.  I guess L went to see him ALL day and used some phony friend as an alibi...I mean, the friend does exist but she hasn\'t seen her for YEARS.  Well, Lis started talking about all of her issues...I guess I got this look on my face and she asked me what I was thinking.  I just told her I was so worried about her and L.  I told her I didn\'t understand where they were coming from because I never cared what my friends were doing, I just didn\'t feel the need to do something I didn\'t want to do to fit in.  I have never completely fit in anywhere so I got used to it very early and just decided what I wanted to do and what I didn\'t want to do...well, I basically busted out, are you still smoking pot?  No...I don\'t think I asked the sex question...unusual, I know but I didn\'t.  I told her, just as I told L, I am NOT judging you...it is your life and you have to make your own decisions and your own mistakes...I just wish that my mistakes could have been lessons for you guys so you don\'t have to go through the same heartache that I did...I am trying to take the same approach with you that I take with B...yes, I would prefer you guys didn\'t do this stuff, but I would much rather KNOW and be able to be there for you guys then never know and be blindsided by it someday.  She also talked about being on antidepressants...I just hope I got through to one of them...I pray I got through to both...I worry about them so much.  And I kind of feel like I am the only one that can talk openly to them and yet try to be there...I know they aren\'t telling J this stuff...she would probably pass out from it all...I know she means well but I also understand when they say they can\'t talk to her because she doesn\'t get it.  Her biggest \"vice\" was being taught how to play poker in college!  Sinner!  Anyway...I try not to worry, try just to be concerned...and I need to make a mental list, I need to be there and need to reach out to them more often...I\'m very very bad at that but it is not an excuse...I need to let them know I love them and I\'m here for them, no matter what...  '),(19465,518116,'03,June,2004','2004-06-03 00:00:00','   what a weekend...Friday, hubby and B went to TC so I was home alone.  I decided to get a little bit of cleaning/organizing done.  Well, I have not been satisfied with the way our living room is situated...so I tried moving the couches until I found a better way...none to be found.  But we now have a new gauge in the floor...yippee!!!   Please note sarcasm    Saturday, C and I headed to Chicago for the day...they had a memorial day parade right in front of our hotel so we had to find our way around it...lol!  Then we went to our room...the world\'s tiniest hotel room!  They did not have a room with two twin beds yet since we were checking in so early...yuck! I was not looking forward to sleeping in the same bed (I\'m weird about that) but I figured it was one night, I could deal.  We shopped ALL day.  We walked all of North Michigan and went in all malls and many shops...it is kind of funny actually since neither of us particularly like to shop!  I hit up H&M...LOVE that store!  I got a ton of cute stuff for the cruise and with the gift certificate from my older sis, hardly spent any $$$!  We walked down State but didn\'t go down far enough for the cute little shops.  But we did run into a nail store so we stopped and I got a MUCH needed pedicure and she got her nails done.  I have never had such a good pedicure.  She took off the yucky skin and massaged my feet and calves...it felt AWESOME!!!  That night, we went to dinner and tried to go to this bar \"Leg Room\"...very cool inside but we were just to early so we beat the crowd.  Then we walked down to Excalibur...same thing, too early.  So we ended up at Hard Rock and had a drink and went to the hotel to bed...haha!  We were both laughing about what old women we are... next time, I would probably do TWO nights so if we have to spend all day shopping we don\'t have to party the same night...  Anyway...Monday, B and I tried to go to Shrek II but it was all sold out...we went on Tuesday, it was cute but I\'m still a fan of Disney movies...lol!  Weird that I feel I have to state my loyalties...it is JUST a cartoon!  I started this running program.  According to the schedule, I will be running 3miles in 30days...well, I\'ll believe it when I do it.  I was going to do it outside but that requires me to buy a pedometer so I know how many times around the park is a mile...too much effort.  So, I am doing it on the treadmill...so far so good.  I mean, my bursitis and my knee are killing me, of course, but I am DETERMINED to do this.  I told hubby, it is my sincerest wish that on our cruise we can run together.  Doesn\'t that seem romantic?  Running on the beaches of Miami?  I just see it as something we can do together and is good for us.  Hubby is slimming down rapidly.  I\'m envious...proud, but envious.  I keep trying to convince myself, I am doing this for my health and I probably won\'t lose any weight...but I know deep down, I see this as my last ditch effort to rid myself of these pounds.  Hopefully, by the time I admit to myself this weight isn\'t going away, the program will be a habit so I will keep with it...I have to admit, it is kind of cool to use my body...  '),(19466,518116,'01,June,2004','2004-06-01 00:00:00','   Hmmm...I have been pretty hormonal this past few days...mind you, I did actually start a cycle on the 26th but I don\'t think all of my emotional turmoil can be chalked up to that...  I will admit it...I think I am more upset and pissed by my inlaws than I am trying to pass off.  Why won\'t I just admit this?  Because if I do, hubby will ask me what I want him to do about it...which makes me feel, if there is nothing to do about it, what is the point in being upset?  Which I agree with but I don\'t...am I making any sense??? Yeah, I didn\'t think so.  Okay...so, what I\'m upset about...WHY do they treat me like crap?  I mean, I know I\'m not blood but I am still technically \"family\"...I am married to their son and mother of their grandson...I just don\'t get it.  Is it me?  Is it because they are just rude and ignorant?  I think if I knew the WHY of it all it would be easier for me to deal with...would it make it any less hurtful?  No...but I would be able to say, okay, they are doing this because they are XXXX.  But I don\'t have that...right now, all it feels like is that they mercilessly poke and prode to see how far they can push me.  And what do I do?  They are FAMILY...I can\'t cut them out of my life...they are too important to my hubby and to my son.  Say something?  I\'ve already tried that route...that ended up with everyone pointing their fingers at me...everytime they do or say something, it takes me a little longer to get over it.  I hate this feeling...  Then the baby thing...I want one.  I am once again at the place where I am contemplating clomid.  Ugh...I make a decision I think I am at peace with and then I step back to re-make the decision.  Hubby says he doesn\'t think he is ready to try it...well, that makes me mad too.  He would not have to DO anything...except things he already does.  I also want to ask him to go get tested.  Part of me feels like I wouldn\'t obsess about this if I knew he was partly to blame too...part of me also feels like if he did have some issues, he could make some easy adjustments to increase our chances.  Ultimately, though...I KNOW that if it was meant to be, it will happen.  I won\'t HAVE to do anything else...if it was meant to be...so, I think I am ultimately struggling with the WHY isn\'t it meant to be...  I think if I was not so nosey about life in general...I would not have any of these questions and the voices in my head would not be leading me to a migraine...  '),(19467,518116,'29,July,2004','2004-07-29 00:00:00','   yes, I realize...this is the 3rd post of the day.&nbsp; I AM BORED!&nbsp; I wouldn\'t be bored if the \"help\" desk would get a techy over here to fix this computer.&nbsp; Ironically, I can get on to blogger without issue...if I attempted to get onto a site that actually was to do with my work, it would lock up.&nbsp; I have shut-down and re-booted this computer at least 6 times today...hence, three postings in the matter of hours.  My new topic.&nbsp; Have&nbsp;I mentioned I know someone on the verge of being famous?&nbsp; Not my older sis who was an attorney at the law firm that represents Monica Lewinsky and Koch (former People\'s Court judge) works at...and, yes, she has met Monica.&nbsp; Now, she works for the firm handling the WorldCom case.&nbsp; Nope, my beautiful friend Kate has signed a contract to publish one of her books!&nbsp; I am so excited for her!&nbsp; It won\'t actually be published for a year...it takes that long for revising, editing, typesetting, etc.&nbsp; However, once it does get published, it will be available in local bookstores (in the US and Canada) and on Amazon.com!!!&nbsp; So, folks, mark your calendars for a year and look up \"Kate Rizor\" on Amazon.&nbsp; Now that I say that, search Katherine as well...I don\'t know which she will publish under.&nbsp; Heck, look her up now because you probably will get her business site.&nbsp; Just know, you heard about her here first...one of her biggest fans!&nbsp; And, if you come to one of her book signings, you will most likely find me there.&nbsp; No, really.&nbsp; I told her she needs to put her photo on the back of her book...she just laughed.&nbsp; I also told her to promise I get to be in the movie version...she laughed so hard I think frozen yogurt came out her nose.&nbsp; Anyway, I have read a lot of books...and I read them to the end because I can\'t stand not knowing that the ending didn\'t justify the shitty writing to get there.&nbsp; This one will be good...I haven\'t read the final cut, yet, but the beginning is great!  Anyway...Kate (Katherine) Rizor...Amazon...one year...  '),(19468,518116,'28,July,2004','2004-07-28 00:00:00','   wow...did you watch Amish in the City?&nbsp; Unbelievable.&nbsp; Did they purposefully find the rudest, most self-absorbed, immature \"city\" kids they could find?!&nbsp; I was SO disgusted by all of them...I mean, we are in 2004, right?!&nbsp; Imagine with me, we replace the \"Amish\" with inner-city black kids who also have never experienced this because they live in the projects...can you imagine the outrage that would be screamed from the rooftops.&nbsp; What an unbelievable showcase of prejudice!  I just found it so sad.&nbsp; I mean, what an awesome opportunity that would have been to show these kids the BEST parts of our culture.&nbsp; It was so cute to watch them...I felt so bad for how terribly they were treated.&nbsp; They were just good, honest kids that were curious about a different way of life.&nbsp; It shows how open they truly are and how close-minded Americans can still be.&nbsp; However, I would dare say, the producers did this on purpose.&nbsp; I mean, was there a kind hearted person in the bunch?&nbsp; NO ONE stood up and say, hey, get a grip, these kids don\'t know what\'s going on...instead of mocking them, let\'s help them.&nbsp; And Mose...my heart was truly bleeding for him.&nbsp; He was trying so hard to get these kids to accept him.&nbsp; It pained me but I couldn\'t turn it off.&nbsp;&nbsp; It so reminded me of junior high all over again.&nbsp; I wanted to go smack those kids upside the head!&nbsp; His toys were very cool...and I felt so bad when he thought he was drowning (which I think, he really could have been).&nbsp; I have had the experience of taking someone to Lake Michigan for the first time...they had only ever been in small lakes and pools...it is VERY intimidating to someone who did not grow up in that kind of water to see these huge waves coming at you and over you.&nbsp; And, let\'s face it, people DO drown in oceans and lakes like that.&nbsp; UGH!&nbsp; Can you tell how outraged I was???&nbsp; Mose is the one that really got to me...especially when they were showing him reading his Bible and praying.&nbsp; I didn\'t realize the Amish believe if you are out of the Amish and die you go to hell...what a burden.&nbsp; And pretty amazing to me that he left if he really believes that.&nbsp;   Anyway, the end result...the Amish kids are really cool and people I would actually be friends with.&nbsp; The city kids...desperately need to grow up and experience the outside world.&nbsp; Pretty weird considering the premise, eh?  '),(19469,518116,'28,July,2004','2004-07-28 00:00:00','   My dear friend, Anne, sent this to me.&nbsp; It is so completely true.&nbsp; Hubby and I go round and round because he sees all the things we DON\'T have and I focus on all the things we do have.&nbsp; Gratitude and thankfulness can change your outlook on life drastically...I know it helped me change mine.&nbsp; The health comment touched me tremendously.&nbsp; This is one huge area of my life where I need to spend more time being thankful...    A newly arrived soul in Heaven was met by St. Peter. &nbsp;St. Peter and the soul took the a tour around Heaven. They walked&nbsp;side-by-side inside a large workroom filled with angels.     &nbsp;  St. &nbsp;&nbsp;&nbsp; Peter stopped in front of the first section and said, \"This is the Receiving Section. Here, all petitions to God&nbsp;said in prayer are received.\"&nbsp;&nbsp;The soul looked at the section, and it was terribly busy with so&nbsp;many angels sorting out petitions written on voluminous paper&nbsp;sheets and scraps from people all over the world.     &nbsp;  They continued&nbsp;on until they reached the second section. St. Peter told the soul,&nbsp;\"This is the Packaging and Delivery Section. Here, the graces and&nbsp;blessings the people asked for are processed and delivered&nbsp;to the living persons who asked for them.\"&nbsp;&nbsp;&nbsp;The soul noted again how busy it was. There were many angels&nbsp;working hard at that station, since so many blessings had been&nbsp;requested and were being packaged for delivery to Earth.&nbsp;&nbsp;     &nbsp;  Finally at the farthest corner of the room, the soul stopped at a very&nbsp;small station. To the surprise of the soul, only one angel was seated&nbsp;there, idly doing nothing. \"This is the Acknowledgment Section,\"&nbsp; St. Peter told the soul.&nbsp;&nbsp;&nbsp;\"How is it that? There\'s no work here?\"&nbsp;&nbsp;\"So sad,\" St. Peter sighed. \"After people receive the blessings&nbsp;they asked for, very few send acknowledgments.\"&nbsp;&nbsp;&nbsp;    &nbsp;  \"How does one acknowledge God\'s blessings?&nbsp;&nbsp;&nbsp;&nbsp;\"Simple,\" St. Peter answered. \"Just say, \"\'Thank you, Lord.\'\"&nbsp;&nbsp;&nbsp;&nbsp;     &nbsp;  \"What blessings should they aknowledge?\"&nbsp;&nbsp;&nbsp;&nbsp;     &nbsp;  \"If you have food in the refrigerator, clothes on your back, a roof &nbsp;&nbsp;&nbsp; overhead and a place to sleep ... you are richer than 75% of this world.&nbsp;&nbsp;&nbsp;&nbsp;     &nbsp;  \"If you have money in the bank, in your wallet, and spare change&nbsp; in a dish ... you are among the top 8% of the world\'s wealthy.&nbsp;&nbsp;&nbsp;&nbsp;     &nbsp;  \"And if you get this on your own computer, you are part of&nbsp;the 1% in the world who has that opportunity.\"&nbsp;&nbsp;&nbsp;&nbsp;     &nbsp;  Also . . .\"If you woke up this morning with more health than illness ..... you are more blessed than the millions who will not even survive this week.&nbsp;&nbsp;&nbsp;&nbsp;     &nbsp;  \"If you have never experienced the fear in battle, the loneliness of imprisonment, the agony of torture, or the pangs of&nbsp;starvation ..... you are ahead of 700 million people in the world.&nbsp;&nbsp;&nbsp;&nbsp;    &nbsp;  \"If you can attend a church meeting without fear of harassment,&nbsp;arrest, torture, or death ... you are envied by, and more&nbsp;blessed than, three billion people in the world.&nbsp;&nbsp;&nbsp;&nbsp;     &nbsp;  \"If your parents are still alive and still married ... you are very rare.&nbsp;&nbsp;&nbsp;&nbsp;     &nbsp;  \"If you can hold your head up and smile, you are not the norm, you are unique to all those in doubt and despair.\"&nbsp;&nbsp;&nbsp;&nbsp;    &nbsp;  \"OK, what now? How can I start?\"&nbsp;&nbsp;&nbsp;&nbsp;     &nbsp;  \"If you can read this message, you just received a double blessing in that someone was thinking of you as very special, and, you are more blessed than over two billion people in the world who cannot read at all.\"&nbsp;&nbsp;&nbsp;&nbsp;    &nbsp;  Have a good day, count your blessings, and if you want, pass this&nbsp;along to remind everyone else how blessed we all are.&nbsp;&nbsp;&nbsp;&nbsp;     &nbsp;  Attn: Acknowledge Dept.: Thank You for&nbsp;giving me the ability to share this message and for giving&nbsp;me so many wonderful people to share it with.\"&nbsp;  '),(19470,518116,'28,July,2004','2004-07-28 00:00:00','   hmmm...I forgot to mention.&nbsp; I got a new cd.&nbsp; Kevin Lyttle.&nbsp; He sings \"Turn Me On\"...very island-ish.&nbsp; I LOVE that song.&nbsp; I have listened to the whole cd once, it has potential...but it is all very island beat-ish.&nbsp; I know that isn\'t for everyone but I like it.  oh yeah, I did exchange my Gavin DeGraw (have I mentioned Gavin was a serious contendor for B\'s name?).&nbsp; I\'m glad I waited.&nbsp; It has a live cd in it as the added bonus...I really love live cd\'s...his also has potential.&nbsp; I\'ve listened to it a couple times...I think I\'m just still hung up on George (they re-ran his Oprah yesterday...again, yummy!)...haha!  '),(19471,518116,'27,July,2004','2004-07-27 00:00:00','   ohmigosh!!!&nbsp; I just got one of the best calls of my life and I\'m about to cry!&nbsp; One of my best friends, C, is PREGNANT!!!&nbsp; Please keep her in your thoughts and prayers that she will have no problems...see, C has been trying to get pregnant for as long as I have known her, three years.&nbsp; Seriously, this girl has done everything, chlomid, ovarian drilling.&nbsp; She was trying a new round of chlomid mixed with met (I had told her I read there was higher success with the two) and then, if that didn\'t work, she was going to try invitro.&nbsp; The real kicker was the dr\'s could not find out what was wrong...so she didn\'t have a face to what she was fighting against.  Anyway...she said she started cramping and her hubby said she should take a hpt.&nbsp; Well, she did (Sunday morning) and she said it was positive...she started balling.&nbsp; Of course, she didn\'t believe it because she had a shot of hcg (?) and figured it was that.&nbsp; So, she didn\'t tell her hubby and scheduled a blood test with her fertility dr.&nbsp; The blood test came back positive!&nbsp; The dr told her it wouldn\'t be the hormone shot because that was out of her system a few weeks ago.  So, she went and bought \"What to expect when you\'re expecting\" and gave it to her hubby...he, of course, was ecstatic.&nbsp; She is still stressed out.&nbsp; I think when you go through what she has to get pregnant, you know worst case scenarios for everything and expect them since, up to this point, you have experienced all worst cases.&nbsp; I tried to calm her and tell her it is okay she isn\'t feeling sick yet, she is figuring she is six weeks.&nbsp; I know I didn\'t find out I was pregnant until 8weeks and that was because I woke up with sore breasts...it wasn\'t until after that the morning sickness hit.&nbsp; I think it comforted her a bit...I was more concerned when she said she was cramping but didn\'t say anything.&nbsp; We talked about what the book said, it says the cramping is caused by the uterus enlargening and her nurse confirmed that.&nbsp; Whew...  She doesn\'t want anyone to know until after her first trimester...very understandable.&nbsp; So, only her drs, hubby, and I know...I feel so extremely honored...being the wonderful friend she is, she asked me how I was doing with the news, hoping she didn\'t depress me.&nbsp; Oh, of course not...I am so so very happy for her...I think she is going to be one of the best moms I know.&nbsp; I know what heartache she has gone through, just wanting a baby so much...I could not be any happier.&nbsp; I tried to tell her this, of course, I don\'t think she believed me.&nbsp; I seriously have had to deal with this so much this year, it isn\'t as painful as it used to be.&nbsp; Plus, like I said, I have been praying for her so much, I never gave myself a thought during the conversation...she said, well, if I end up having just one, maybe that will encourage S to let you try chlomid.&nbsp; Haha...I told her, no way, we are shooting for 6 here.&nbsp; (Her test for the hcg shot had shown 2 perfect eggs, 1 other that looked okay).&nbsp; She told me that her eggs could actually split into 3\'s!!!&nbsp; I said, well, than I\'m praying for 9!!!  Anyway...I had to share somewhere, just don\'t tell anyone else!  And, remember, pray your little hearts out...I have never known someone more deserving...I promise.  '),(19472,518116,'27,July,2004','2004-07-27 00:00:00','   lordy...I so do not feel like working.&nbsp; Fortunately, I get to leave at 2pm today.&nbsp; Our sitter is on vacation still (she will be gone a total of 2 1/2 BUSINESS weeks!!!), which is a pain since we don\'t really have a back-up.&nbsp; La watched&nbsp; B all week last week for us.&nbsp; My mom watched him yesterday and will watch him tomorrow and Thursday, if she remembers, but wanted today off.&nbsp; So, I\'m taking off...plus I will have Friday off for his surgery so I will be fortunate if I don\'t get axed for excessive time off!!!&nbsp; And, no, I\'m not really joking...  Anyway, I have to go back to Target.&nbsp; I swear that place is a racket.&nbsp; I was there yesterday to fill my 2nd asthma prescription.&nbsp; I have been waiting for their $10 gift card coupon...I found it on Sunday...yippee!&nbsp; I also wanted to pick up toys for B for his surgery.&nbsp; I got a cool Spiderman \"doll\" that is super-poseable...plus, it was on clearance for half off!&nbsp; He is majorly into Spidey right now, though he hasn\'t seen the movies or cartoons since there is violence we don\'t condone for a four year old.&nbsp; Anyway...got a spirograph,&nbsp;a viewmaster and TONS of extra disks, and a Spiderman tshirt for him to wear to surgery.&nbsp; LONG ramble...so, they had Gavin Degraw advertised.&nbsp; I picked up the cd and was charged full price.&nbsp; So I walked my hiney to customer service...well, I guess there is a 2nd cd from Gavin, also named Chariot, that does not come out until TODAY.&nbsp; So, I have to take the cd I bought yesterday and return it for that one...I could have just kept the original but the new one has 2 cds so I would MUCH rather just wait to open it up...keep your fingers&nbsp;crossed that it is in stock.  Why am I in such a cd mode?&nbsp; Not sure...but I haven\'t been keeping up with my cd list either.&nbsp; I will have to visit that and see if I can fill in the blanks...  '),(19473,518116,'26,July,2004','2004-07-26 00:00:00','   ok...so, I am driving myself crazy...it takes quite a lot of talent to do that!&nbsp;   The issue...my hair.&nbsp; It is driving me NUTS!&nbsp; I have really long hair, past the middle back.&nbsp; My original goal was to grow it to my waistline but this is the problem, I have natural wave.&nbsp; So, basically, I wash, blowdry to straighten, and then hot roll it.&nbsp; Do this every morning and you understand why I am going nuts.&nbsp; Then add the fact that the humidity cancels out all the work and I\'m back to funkedified wave...so, like today, by the time I walk into work, it looks funky.&nbsp; ugh.&nbsp; So, I am considering getting my haircut...off.  A) I think hubby will be happy because he likes it shorter but not too short B) I don\'t know what kind of hairstyle to get because I have mostly had long hair my whole life...except when I was actually skinny and could get away with short hair C) my face is round so I never know what kind of style looks good  So, I am back to looking up Tiffany Amber Thiessen hairstyles...her face is the most like mine shape-wise...I think I will walk in (I made an appointment for Saturday) and tell my stylist I need a change, have a few pictures, and see what she thinks.&nbsp; That is how I usually do it and there have only been a couple disasters...  great, my sister\'s shower is in two weeks...I know I\'ll hate my hair and see all kinds of people I never see and look crazy...  I need to get back to running...I got the second asthma med yesterday, started today...so by the time I\'m back up in the running I should be okay...today: walking one mile...god give me some inspiration!  oh, and I need to figure out how to get my nose ring out so I can change it and potentially leave it out sometimes...  '),(19474,518116,'26,July,2004','2004-07-26 00:00:00','   so, we went to the inlaws this weekend...it was interesting.&nbsp; Supposedly, my bil and sil were NOT talking to my mil.&nbsp; Well, hubby and I volunteered to babysit so they could have some \"alone\" time before sil goes back to work.&nbsp; So, we get up there, call them and they don\'t want to go out...they would rather spend the time with us.&nbsp; Well, they did not get a \"formal\" invite to dinner so hubby acted as middle man to arrange it all.&nbsp; They come over, we all have dinner, everything seems hunky-dorey.  I notice that evening, sil seems inattentive when it comes to L.&nbsp; I didn\'t want to pass judgement so I thought maybe it was just the crap with my mil.&nbsp; who knows?   Saturday we get up and get ready to go out on the boat.&nbsp; Sil says she will catch up with us later.&nbsp; We go out.&nbsp; Why do we always have to bring both boats???&nbsp; It makes for a very awkward time.&nbsp; Do we ride with dad or brother?&nbsp; So, hubby and I ride with my bil and send B with mil and fil.&nbsp; Mind you, I am still not feeling quite right...hubby tells everyone I have been having issues with my stomach.&nbsp; Umm...great!&nbsp; So now, everyone looks at me like I\'m about to crap my pants...how halarious...  We stay out for quite some time...sil calls, still not on her way.&nbsp; By the time she got there, everyone else is ready to go but we stay out for a while longer.&nbsp; I start feeling awful so sil, mil, L, and I pack up while the guys get the boats in.&nbsp; We head out to the pizza place to order pizza.&nbsp; L starts screaming at the top of her lungs...not a whimper, full-fledge screaming.&nbsp; Sil just ignores it.&nbsp; She goes in to order the pizza and L is SHRIEKING...I mean, the baby is 9weeks old but I think she is about to pass out from how hard she is crying.&nbsp; Sil gets back in the car and continues to ignore.&nbsp;   So, we get dropped off at inlaws and we will wait for hubby and fil to pick us back up to go over to bil and sil...I rest (fall asleep).&nbsp; We go over there.&nbsp; Eat pizza.&nbsp; Get going.&nbsp; The dog is annoying, barking at everyone which makes L cry...sil still ignoring her.&nbsp; On the way home, mil goes off about how she thinks L\'s problem is the breast milk because sil has not stopped to think about the way her eating habits affect L.&nbsp; True but I don\'t know if this is it.&nbsp; Regardless, it showed both hubby and I a little insight into where mil is coming from.  Anyway...I remember a friend I have, h, that also knows my sil, saying, I don\'t know if she will be able to handle a baby...it won\'t be all about her anymore.&nbsp; I had forgotten that comment until this weekend as I am watching her...I\'m still trying not to pass judgement but she was not like any new mother I have ever seen...and I\'ve seen a lot!&nbsp; Usually, when your newborn cries your top priority is to figure out what is wrong and try to get the baby to stop crying...she could not have cared less.&nbsp; And, as I was watching her this weekend, I thought, h was right...she can\'t handle her show being cancelled.&nbsp; Sad...  But, all in all, we had a great time which surprised us both.&nbsp; I had nightmares about babies both nights we were there...which was sad.&nbsp; It was a bittersweet feeling spending so much time with L.&nbsp; I loved holding her and watching her but it was tough because it brought my desire to have a baby right to the surface...  oh yeah...bil asked if they could come the WHOLE week with us to FL next year...umm, we offered two nights!&nbsp; Hubby said not a chance in hell...haha.  '),(19475,518116,'23,July,2004','2004-07-23 00:00:00','   S, in your Cinderella story, you\'d get to Kiss the Prince       You don\'t need the pouffy dress, the fancy hair, or the 3-inch glass heels. A romantic soul like you just wants a guy who\'s a true companion, a good friend, a real prince. And if he\'s hot? All the better. But ultimately, you know that home is where the heart is. Which is probably why you surround yourself with good friends you can take care of and laugh with — friends who will do the same for you.But that doesn\'t keep you from fantasizing about the perfect relationship. It\'s not that you don\'t have your feet firmly planted on the ground. It\'s just that you\'re not afraid to shoot high. Or wish upon a shooting star. So keep reaching for your goals, Cinderella. If you do, your happily ever after can\'t be far away. And you can seal that with a kiss.  '),(19476,518116,'23,July,2004','2004-07-23 00:00:00','   damn!&nbsp; I hate when I do that!&nbsp; So, as I mentioned, Target was out of Ashlee Simpson\'s cd.&nbsp; So, I ran to Best Buy yesterday to get it.&nbsp; I was obsessing, so I figured I would treat myself.&nbsp; I don\'t remember the last \"new\" cd I bought.&nbsp; Usually, I buy them off of eBay...actually, I do remember, Alicia Keyes.&nbsp; So, it hasn\'t been terribly long, but still...so,&nbsp;I picked up the cd and was scoping out the others...I LOVE music and listen to it basically 24/7.&nbsp; So, I saw a Saliva cd...there is a song of theirs I LOVE so I looked at the back to see if I could figure out if this was the right one.&nbsp; Well, Best Buy has some other music blaring and I can\'t think of anything but what I\'m hearing!&nbsp; So, I decide just to buy it...dumb, dumb, dumb.&nbsp; NO, it isn\'t the cd I want and I had to turn it off because it is too angry rock to play at work.&nbsp; Damn!&nbsp; Oh, well...one more cd I own that I don\'t listen to...so, I got on Amazon, researched what cd of Saliva\'s I want and ordered it...  oh, yeah, one of my manager\'s looks like he could be George\'s brother...no wonder I get all funny around him...haha  '),(19477,518116,'22,July,2004','2004-07-22 00:00:00','   I forgot...I couldn\'t get Ashlee Simpson\'s cd yesterday because Target was sold out.&nbsp; I\'m a dork, I know...but I did pick up George Michael\'s Patience.&nbsp; Man!&nbsp; That man is so extremely hot!&nbsp; I love the song \"Amazing\" and have listened to it close to 100 times...which means, next week, I won\'t be able to stand it...but I don\'t know.&nbsp; Just picturing George gives me the chills.&nbsp; Yummy!  '),(19478,518116,'21,July,2004','2004-07-21 00:00:00','   lordy!&nbsp; This week is absolutely never-ending...yesterday, I was seriously terrified that if I blinked too long I would fall asleep at my desk.&nbsp; It isn\'t like I don\'t have tons of stuff to do...tons may be an exaggeration but I am still trying to sort through crap from being off...but that is the problem, trying to make heads or tails of someone else\'s work is just about impossible.&nbsp; I get annoyed because I\'m looking at the same po for two days and can\'t figure out where to find the backup info.  Got my rings yesterday...yippee!!!&nbsp; I am in LOVE with my mystic topaz.&nbsp; It takes on a pretty green color when under fluorescent light and more of a purple color in natural...plus, its huge!&nbsp; Haha...I don\'t know why, but I love big rocks on my fingers!&nbsp; My blue topaz is beautiful as well but, as always, I don\'t wear much blue so I don\'t really know when I think I\'m going to actually wear it!&nbsp; Though, I did get a cute tank top with blue flowers on it so I\'m sure I\'ll wear it then.  Okay...so question...how long after a cyst bursts before you feel \"normal\"?&nbsp; Anyone know?&nbsp; I don\'t feel sick just nauseous occassionally...also lightheaded.&nbsp; I attribute it to the cyst but it may not be...barely any cramping left, so that is good...  hearing test a week from Monday...B\'s surgery a week from tomorrow...got my teeth cleaned this past Tuesday!&nbsp; Man, that felt awesome.&nbsp; I had not gone in an embarassingly long time...but the funny thing is she told me I had hardly any plaque build-up and my teeth looked great!&nbsp; Whew!&nbsp; I was so terrified that I had cavities.&nbsp; So, now, I am taking an interest in my teeth...I\'m starting a whitening system.&nbsp; My teeth aren\'t bad but I want them to&nbsp;be white.&nbsp;   Talked to my sister yesterday...explained the blow-up with mommy-dearest.&nbsp; She re-iterated...it isn\'t that mom despises me, she just doesn\'t like it when people hand her the truth and disregard the version of truth she is trying to pass off, which is what I do.&nbsp; Actually saw her yesterday, we said hi but didn\'t talk.  Can you tell I\'m bored out of my goard???&nbsp; Ugh.  '),(19479,518116,'21,July,2004','2004-07-21 00:00:00','   it\'s a boy!&nbsp; Woo-hoo!&nbsp; They are going to name him William and call him Liam.&nbsp; No middle name yet...  I forgot to mention a couple B\'isms from last night:  1.&nbsp; Mom, know what is better than smelling my feet?&nbsp; No, B, what?&nbsp; Smelling my armpits!&nbsp; B, that is disgusting!&nbsp; No, it\'s not, it\'s cool!  2.&nbsp; Mom, can you come wipe my butt?&nbsp; (he is four and hasn\'t completely taken over this responsibility...gross, I know...but you do what you gotta do) Did you go poop?&nbsp; (I don\'t know why I ask this...habit) No...&nbsp; Then why do I have to wipe your butt?&nbsp; Because I tooted and it was funky so I want to make sure there is nothing on my butt.&nbsp; (and, yes, I did go up and wipe it)...  Okay, so speaking with our son like he is an adult does not always go our way...yes, I know he has a wide vocabulary but he also understands things that four year olds probably should not.&nbsp; And, on a particularly humorous day, delivers doozies like the above two.&nbsp; It is so hard not to wet myself as I am laughing!  '),(19480,518116,'20,July,2004','2004-07-20 00:00:00','   I think J&R are having a boy!&nbsp; Of course, we played phone tag all day yesterday, so I never heard it from her mouth but La said she \"thinks\" she heard it was a boy...La is young so this is not that exciting to her...but I\'m really excited!&nbsp; Now, if I can just get her to tell me what the name is going to be!!!&nbsp;   We are going up to the inlaws this weekend.&nbsp; M&J had a rough weekend this past one so when I called to see if they were going to be around, I offered us up to watch L.&nbsp; M felt bad because she said they want to see us too, not just use us as babysitters...I told her it was up to them but not to feel bad and just let us know.&nbsp; Haha!&nbsp; I\'m back as being the bigger person in all situations and I must say, it feels good!&nbsp; M and my mil are at each others throats...well, M is really pissed at her and no one can tell if my mil even&nbsp; knows.&nbsp; BUT it takes me out of the hotseat for a little while!&nbsp; So, we will head up there Friday and probably come back early Sunday.&nbsp; We will probably go out on the boat for a while and I\'m not sure what else.  My rings still aren\'t done being sized...I\'m keeping my fingers crossed it is today...I can\'t wait anymore!&nbsp; haha  '),(19481,518116,'20,July,2004','2004-07-20 00:00:00','   It is SO hard to believe that a week ago we were on vacation...we were shopping in Nassau to be exact.&nbsp; So, here is a mini run-down of our vacation: &nbsp; Saturday: Drop B off at my parents, drive to Elmhurst for Portillo\'s (YUM!), stop at Kohl\'s so hubby can buy a watch (his broke Friday night), somehow found ourselves to the Park-n-Jet, fly from Chicago to Miami.&nbsp; Arrive around 8pm (I think) and wait for hotel shuttle to take us to the hotel.&nbsp; Get to the room, relax, go to hotel restaurant for dinner and saw the fireworks from the restaurant. &nbsp; Sunday: Wake up, go downstairs to find some info on what to do in Miami.&nbsp; Concierge hooks us up with a tour guide.&nbsp; Rush back upstairs to get breakfast and get ready.&nbsp; Meet Oscar for our 3 1/2 hour tour.&nbsp; Oscar, who is Cuban, takes us to Coral Gables, South Beach, and several other places...very nice guy...very cute accent...we got to see Julio Iglesias, Rosie O\'Donnell, and Gloria Estefan\'s houses.&nbsp; Hubby bought cigars from a Cuban guy...very cool tour... &nbsp; Sunday night we are hooked up with a FREE limo ride down to South Beach to go to dinner at a fancy Italian restaurant.&nbsp; First time hubby or I remember riding in a limo...fun!&nbsp; After a nice dinner, we walk down to South Beach but it starts raining so we go back and take the limo back to the hotel. &nbsp; Monday:&nbsp; Cruise day!&nbsp; We get up, get breakfast, get ready and go meet the shuttle to take us to the cruise.&nbsp; Getting on the boat did not take as long as we anticipated.&nbsp; We scoped out the boat a bit and had lunch.&nbsp; We go to a mini-seminar about shopping in Nassau and get some free stuff (jewelry, mug, perfume)...go to the casino and get lessons on blackjack and some poker game but I don\'t remember which one.&nbsp; Have dinner...meet a VERY nice couple from Pennsylvania, not too impressed with the other couples (one couple can\'t stop complaining about their food).&nbsp; Later on we stop in the casino where I lose $10 in the matter of minutes and I walk away.&nbsp; Stopped to watch the blackjack table and see&nbsp;a guy lose $500 in less time than I lost the $10...all of a sudden, I don\'t feel so bad.&nbsp; Saw the show... &nbsp; Tuesday: Nassau!&nbsp; We get up and get out shopping.&nbsp; Promised hubby it would be 2hours, it was 3 1/2.&nbsp; Still not bad.&nbsp; Got a bunch of free stuff and some souvenir stuff.&nbsp; Afterwards, we go back to the boat for lunch and a quick rest.&nbsp; Then we head back out to catch a taxi to take us to Paradise Island.&nbsp; We go up to see the Atlantis and walk around for a couple hours...we aren\'t allowed to go outside to the cool places which suck but serves us right for being too cheap for the tour!&nbsp; We go to their casino where I won $13.25 in the first push of the button (nickle slots)...I feel redeemed from the night before.&nbsp; Take a water taxi back to Nassau and stop by Senor Frog\'s...two children ask us to pay them $1 for a song.&nbsp; Stop at the straw market...very hot...very pushy people.&nbsp; Go back to the boat.&nbsp; Attend an art auction, get free art.&nbsp; Formal night, we have a bottle of champagne and I feel drunk...haha!&nbsp; Go to the captain\'s welcome for more champagne...is gross but free!&nbsp; Listen to the band. &nbsp; Wednesday: CocoCay!&nbsp; Get up, have breakfast, take the boat over to the island.&nbsp; Walk around the island and go get snorkeling gear and lesson.&nbsp; Begin snorkeling.&nbsp; See the shipwreck and airplane crash.&nbsp; Lose my conch shell bracelet...find it...lose it again.&nbsp; Get bit by a fish.&nbsp; Go have lunch and go to Aqua park.&nbsp; Realize the park is really for young kids and teens...decide we are too old and fat to get on a giant inflated see-saw and giant inflated \"rock\" climb (couldn\'t get our fat butts up and out of the water!) Decide to stop wasting time and go back to snorkeling...go to the last lifeguard station and then head to the deep sea.&nbsp; On the way out, people are coming in saying there has been a shark siting but we see lots of people still out and figure it is a rumor.&nbsp; We see a lot of stingrays, beautiful colored fish, and, yes, a shark!&nbsp; Small one though...I had the camera so I had to try to swim closer to get a picture, don\'t know if it will turn out or not.&nbsp; After we come in, we are told it is very possible large sharks were in the deep sea section as it is MATING season and they come into more shallow water to mate...beautiful!&nbsp; Go back to boat...see a show.&nbsp; Play the \"scavenger\" hunt game. &nbsp; Thursday: Key West!&nbsp; We have to wait for everyone to be cleared by customs...get pissed at same idiots whose names are being called.&nbsp; Get cleared.&nbsp; Get breakfast, head out.&nbsp; Call B, he tells me that he is being made to work on his vacation and asks if I have ever heard of working on a vacation...very hostile.&nbsp; Ask if he is having a good time and he says no because he has to work on his vacation...duh, mom!&nbsp; We walk around aimlessly...find the aloe store to get aloe for our sunburn.&nbsp; Find LOTs of beautiful jewelry but nothing quite my price.&nbsp; Keep walking around.&nbsp; Stop for a piece of Key Lime pie.&nbsp; Eat lunch at Hogs Breath.&nbsp; Walk some more...and some more...find a blue topaz I break down and buy.&nbsp; Start heading back to the boat, pass a market to shop at.&nbsp; BOUGHT some shells and a stingray doll for B...stop in one more jewelry store before getting back on the boat...bought a gorgeous mystic topaz in white gold.&nbsp; I feel guilty but hubby assures me I just get it.&nbsp; (He has bought quite a few expensive things for himself by this point so I\'m sure my getting the ring makes HIM feel less guilt).&nbsp; Go back to the boat and swim.&nbsp; Take a nap and go to dinner.&nbsp; Pack and fall asleep. &nbsp; Friday:&nbsp; Get up and ready immediately, we have to be off the boat quite early.&nbsp; Get to the hotel around 9 or 10am.&nbsp; I order breakfast and we both take a nap.&nbsp; Get up, watch tv, watch Jersey Girl, take a nap.&nbsp; Get up, go down to gift shop for pop, take another nap and watch tv.&nbsp; Get up, go to hotel restaurant for dinner, go back and lay down.&nbsp;  &nbsp; Saturday:&nbsp; Get up EARLY to take shuttle to airport.&nbsp; Plane loaded and then find out valve is broken...delayed 45 minutes.&nbsp; Get to Chicago, get luggage, meet shuttle to go get car...head back to Michigan...stop at parents to give souvenirs and get Brennan...GO HOME!!! &nbsp; Okay...so this is the short version and there are so many memories in my head.&nbsp; If I figure out the picture stuff I will post some once they are returned.&nbsp; It was definitely the best vacation I have ever been on...even more special because hubby and I were ALONE for a whole WEEEK!&nbsp; We were nervous about it but it was just like old times, bumming around with my best friend.&nbsp; Made this week more lonely, but that is okay...it gives me something to look forward to when I have more time with him someday!  As for salmon...yes, it was my first time!&nbsp; There isn\'t a good seafood/fish place in Kzoo...and when I lived in Chicago, I would always get shrimp at those places...I like the salmon when it was warm, when it was cold they had it too salty...  '),(19482,518116,'19,July,2004','2004-07-19 00:00:00','   oh lordy!&nbsp; My day went so badly yesterday...so the cyst burst and it was bad, bad pain...we went out grocery shopping, which I barely made it through and then I came home and slept...and slept...and slept some more.&nbsp; Well, we had invited the parents over for Key Lime Pie as an additional \"thank you\" for watching B for the week...so, they arrive.&nbsp; My mom asks how I was as I wasn\'t looking to keen...I briefly explain what happened and we move on with conversation. &nbsp; Fast forward an hour...maybe two.&nbsp; Dessert has been eaten.&nbsp; Hubby, B, and my dad go out to play baseball.&nbsp; My mom decides this is the time to start planning my older sis\' baby shower.&nbsp;  &nbsp; Background info: she had already sent out invites WITHOUT consulting with either La or I...SHE created the guest list, again, without input...and invites HER friends whom my sister has met once or twice...READ: this is really a party for HER, a social event for her to pull off.&nbsp; Now, mind you, I have lived with her ways for many years and KNEW this is what is was going to become... &nbsp; Back to last night...so, she had previously asked what I thought of having this restaurant cater.&nbsp; Stuipidly, I said I didn\'t think it was a good idea because it would cost too much.&nbsp; So, as \"we\" are planning this, the first thing she writes is Restaurant X for sandwiches.&nbsp; ummm...okay.&nbsp; Then she asks, what would go with the sandwiches...to which I say, fruit and vegetables.&nbsp; What should we drink? Tea and lemonade.&nbsp; Should we have a cake? yes...and on with the questions...which I answered.&nbsp; All of a sudden, she puts the pen down and says to La and I, I don\'t appreciate your lack of enthusiasm.&nbsp; WHAT?!&nbsp; (yes, that is what I actually responded with)...I just don\'t think you are very excited about this and willing to help.&nbsp; To which I said, I guess I\'m confused, you were asking questions, I was answering...besides, this is your party and I am assuming you are going to tell me what you want me to do.&nbsp; (Not exact words but meaning)...well, it went downhill from there.&nbsp; So, basically, there was my mom and I yelling at each other...me trying to remind her the ordeal I had just gone through....I have been trying all night and all morning to understand where she is coming from and I just can\'t...she stomped out of the house and told my dad to take her home... &nbsp; once again, I\'m the black sheep of the family...it is usually more fun than last night.  '),(19483,518116,'18,July,2004','2004-07-18 00:00:00','   Email I sent from the cruise on Wednesday, July 14th: &nbsp;  Good morning!&nbsp; We are about to arrive in CocoCay this morning.&nbsp; We will be going snorkeling and to an \"Aqua Park\" which, basically, sounds like a water amusement park--we are thinking along the lines of Michigan Adventure???&nbsp; Of course, last night, I had a nightmare about being bitten by a shark and being stuck on coral so we will see how long I last!&nbsp; &nbsp;&nbsp;&nbsp;  We have been having a wonderful time.&nbsp; We got to relax in Miami and took a tour around the city.&nbsp; Our guide was Cuban and very friendly and very knowledgeable.&nbsp; It was neat to walk around with him as he showed us all kinds of places that aren\'t necessarily tourist attractions.&nbsp; One of his friends gave us some Cuban coffee.&nbsp; I actually liked it!&nbsp; It was VERY strong and made your tongue tingle.&nbsp;&nbsp;Hubby was not a big fan.&nbsp; Later that night, a limo picked us up at the hotel to go down to South Beach for dinner.&nbsp; We ate a lovely Italian restaurant and then walked down to South Beach.&nbsp; &nbsp;&nbsp;&nbsp;  The first few hours of the cruise we were a little nervous about being bored but it is NOT even possible!&nbsp; There are a million things going on the boat and we are somewhere different everyday.&nbsp; Yesterday, we were at Nassau and took a taxi over to Paradise Island.&nbsp; I dragged&nbsp;hubby to a special seminar on the boat about the shopping in Nassau...you can imagine how thrilled he was!&nbsp; But we got some \"insider tips\" and had a successful (though I\'m sure&nbsp;hubby would say painful) shopping trip.&nbsp; You will have to ask him about the \"Straw Market\" sometime...Paradise Island was beautiful and we went up to the Atlantis Resort...amazing.&nbsp; Being the seasoned gambler, I sat down at the nickle slots and won the first push!&nbsp; I was so excited!&nbsp; It was the best $13.25 I have ever had!!!&nbsp; And then, the frugal side of me came out and we walked away...lol!!!  &nbsp;  Last night was our formal night.&nbsp; It was very nice.&nbsp; I tried salmon for the first time!&nbsp; I tried to get&nbsp;hubby to taste escargout with me, but we chickened out&nbsp; After dinner, the captain\'s reception took place.&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;  Well, we are off...we miss you!Love~S (&amp; S)&nbsp;   Well...we are back.&nbsp; We got back into Miami on Friday.&nbsp; I wasn\'t expecting us to get off the ship so early but we were off and back at the hotel by 10am.&nbsp; We basically napped and watched tv ALL day...I could have used another day of that!&nbsp; Saturday we flew back into Chicago and then drove&nbsp;the rest of the way&nbsp;home.&nbsp; It was so wonderful to see B again.&nbsp;&nbsp;There was only one time, while we were watching Jersey Girl in the hotel Friday, that it hit me hard how much I missed him.&nbsp; Other than that, I did a&nbsp;good job of compartmentalizing.  We had an amazing time and I will blog more when I have the energy...I am SO thankful, period never showed up!&nbsp; I was so concerned it would happen the second we&nbsp;left.&nbsp;&nbsp;However, this morning I woke up in hideous pain.&nbsp; What I thought was my period but am now beginning to believe was&nbsp;a cyst bursting.&nbsp; My dr\'s told me it would feel like a period but, potentially, with no blood.&nbsp; Well, it was the MOST hideous cramping I have&nbsp;had.&nbsp; I was so nauseous and it hurt to even go to the bathroom! Trust me when I tell you, I have had&nbsp;painful, wrenching cramps...I used to be knocked out for a couple days...but this, this was terrible!&nbsp; I couldn\'t cry because it hurt too much.&nbsp; I seriously thought I was going to pass out...hubby wanted to take me to the emergency room but I just wanted to be crawled up with the heating&nbsp;pad.&nbsp; Lo and behold, this afternoon, I am fine.&nbsp; I am exhausted from this morning and still nauseous but not in hideous&nbsp;pain so I\'m convinced I\'m healed...and oh so grateful it happened today and not a day earlier!  &nbsp;&nbsp; &nbsp;&nbsp; &nbsp;  '),(19484,518116,'09,July,2004','2004-07-09 00:00:00','   Have I ever mentioned, you always get fucked at the drive through...those were words of wisdom I learned when working with AOC in Chicago...whoever went for lunch got screwed...either out of money or out of food.  But, somehow, I was always the one to go get the food...  Once again, I find myself in this type of situation but there is not much I can do about it.  See, although I work for the largest pharmaceutical company in the US, I am not technically their \"employee\"...I work through a company...for FOUR years I have done this, originally I was given the title \"contract worker\" now I have been demoted to \"temporary worker\"...when I went to get my badge renewed the security guy says, you are the longest working temp I know!  Attempting a joke, I\'m sure, as many people here are in the same boat...and yet, I digress.  So, 5 weeks ago, my records indicate I am eligible for my shitty one week paid vacation.  I put it on my timesheet, expect to get paid the following week.  Not so, senora.  Rather than my company call me and let me know what is going on, they wait until I see the correct amount was not dd and I call them...have I mentioned they are a bunch of lazy losers?  So, it was then that Ms. Rodriguez tells me that their \"NEW\" policy says I have to have 1800 hours (instead of 1500)and those hours had to be accumulated AFTER the last vacation pay.  Well, well...the slimey bastards have come up with yet another way to screw me out of money.  Now, I can ONLY earn one week a year AND you just ripped away 200+ hours from last year when this policy WAS NOT in place.    Fast forward to last week.  Once again, I have satisfied all NEW policies communicated to me, so I send in vacation request.  Call Ms R to verify this is true, she confirms.  Call again to bring attention to the vacation request.  Talk to her on TUESDAY and she confirms they are going to pay me.  Last night, hubby comes home and tells me when he called the bank only one week of pay had been deposited.  My temper immediately flairs and now I can\'t sleep.  I arrive at work and, lo and behold, Ms R has left a message on my voice mail for me to call her at 8am.  I call at 8:02am...Mr R answers.  She tells me she is calling about my vacation pay...to which I say, yes, I figured as I know it was not deposited yesterday.  Ms R: Please let me finish what I have to say before responding.  Well, these rocket scientists did not understand their own frickin\' policy.  SUPPOSEDLY, my vacation cannot be paid out until September (after my anniversary date)...BUT since it was their error, they will be paying out next week.  Very lucky...she tacks on, but next year we won\'t be able to pay out until after your anniversary date.  To which I say, I pray to god I won\'t be working for any of you next year.  For some reason, she seems shocked I said that...  While I have the brain-child on the phone I say, so what\'s going on with my raise (thinking to self, you know, the one you all have screwed me out of for two years)...Ms R: I haven\'t heard, I will call today.  Great let me know.  Side note:  I requested this raise the same day I talked to her about being screwed out of my vacation time...something in effect was said, you haven\'t given me a raise for two years and now you are trying to steal my vacation too?!  Why haven\'t you received a raise?  Because YOU told me not to request one two years ago.  Oh, well that hold was taken off a year ago...  Then WHY did you not tell me or why didn\'t YOU initiate the raises I missed.  I decided to call someone here to see if they have even gotten the paperwork as my manager has been telling me he has yet to see anything.  She tells me she wouldn\'t necessarily see it but it would go through Manpower and to call them...yes, I forgot to mention, I have recently been SUBCONTRACTED...so I call her, she tells me she received the request a week ago (What was the hold-up with PeopleLink that it took a month to get to them???) and that she cannot request until I have been approved to work the month of August...oh yeah, I am also now on a month to month...so that would be the end of July at the earliest!!!  I was SOOOOO irate by this point...I basically told her I would be hitting my knees more often and praying a lot harder that I would get out of this insane triangle.  Yeah, she didn\'t like that too much either.  Then I called the area manager of my company.  She was not in so I asked for the name of the branch manager of the home office.  I called and talked to her.  She was well aware of the whole vacation screw-up and apologized for that...I proceeded to tell her the cluster about my raise.  She was surprised and had not heard anything about it and promised to contact Manpower and try to get the paperwork moved.  The ironic part?  I don\'t even KNOW what they are asking...so I could be getting my panties in a bunch about .25!    the test was negative...not much emotion there...  '),(19485,518116,'07,July,2004','2004-07-07 00:00:00','   well, I have the test.  I don\'t know if I will do it tonight or try to hold out until Friday...although, I am pretty busy tonight so maybe that will hold my attention.  \"Working Mother\" magazine has an article this month about 2nd time around infertility.  I didn\'t realize until I read the article how common this problem has become.  It really was a well written article, unlike others I have read about PCOS.  It talked about how heartbreaking it is and it said something to the effect, these women know exactly what they are missing out on, unlike women who don\'t have any children.  Well, yes, exactly.  Not that infertility is any less of a heartache because they don\'t know exactly what they are missing but women who don\'t have any will look at someone like me and say, you should be happy you have at least one.  True, and I am very grateful.  However, you don\'t have a small child begging for a brother or sister either.  Of all the cruel, insane comments I have had to deal with, B\'s hit me right in the heart.  I can\'t explain to a four year old why when I don\'t know why.  I tell him this is how God wanted our family.  To which he responds, then I will pray for him to give me a sister/brother.  Anyway, I just greatly appreciated the article and wrote the editor to tell her so.  On a funny note, we went to Perkins (yuck!) for breakfast with my parents and younger sister.  This family walked in and was about ten people.  B\'s eyes got big and said, their family goes on forever!  It was cute...  Countdown: 3 more days until departure.  Still have to clean the house.  Shop.  Pack.  Leave sister information.  '),(19486,518116,'06,July,2004','2004-07-06 00:00:00','   hmmm...still no period.  I could have SWORN I would start this weekend...I had such hideous pain in the netherregion.  I should get a hpt but then, I hate the idea of spending the money when I know I\'m not.  I have decided to wait until Friday.  I have to know for the cruise since I am sure I will consume a few adult beverages.  We went suit shopping for hubby...his last suit was bought for his college graduation and was a bit snug.  hey, at least he could get the pants on and fastened...I couldn\'t even get my left leg in my college clothes!  Anyway, we went to JCPenney.  He went to the salesman and said, I need your cheapest suit.  Dude said, $200 is the cheapest.  Neither of us was thrilled...my main beef?  I see suits right next to you that say $169.99.  So, basically, I was onto the fact he was only about the sale.  So, finally, after looking around and being bothered by him, I said, let\'s go...I\'ll figure something else out.  As we are walking out, my eagle eye caught a clearance rack.  So, yeah, we found a suit and paid $65 after final markdowns...unbelievable!  Of course, hubby was just laughing.  He says, that guy should have KNOWN better than to mess with you and your nose for a sale...haha!  It wouldn\'t have been either of our first choices but it is for one night and it will look great...plus, he now has a suit that fits properly...  Other than that...trying to build up energy so I can clean the house before we leave.  Who likes to come home to a messy house?    btw...check out the helpmebubby blog...very cool  '),(19487,518116,'05,August,2004','2004-08-05 00:00:00','   hey, fellow blogging friends!  This guy wants to know why we blog:  urlLink http://ydoublog.blogspot.com/2004/08/why-do-people-blog.html  so go forth and answer his question:)  I am in a giddy type mood...I think it is the sugar from the Gatorade I have been sipping on all day.  I decided since I can\'t find an awesome gift for my sis, she is stuck with maternity clothes.  Lame-o, I know.  We talked today...she is stressing out because of her size.  I told her, please, do NOT obsess!  Every woman is so different...every pregnancy is so different.  She is so concerned because she knows I didn\'t start showing until 6months...umm, yeah, but remember me at the end???  Plus, she is significantly shorter than I am, that makes a big difference too!!!  Not much else.  I decided I will be knitting her a baby blanket for the shower in October.  I found a pattern that is pretty cute and looks easy...we\'ll see.  I haven\'t picked up the needles for quite some time!  I went to a yarn store yesterday and was asking an opinion of one of the ladies that works there...big mistake.  This store has always had a stuck-up attitude...I have knit for so long and I will grace you with my knowledge but there is no other way than mine.  So, basically, she told me idea was stupid and she didn\'t know (and I quote) \"what is in young mother\'s heads these days...\"  Wow...let me put back the $40+ worth of yarn I have and relieve you from having such an ignoramous in your blessed presence.  So, I will go to the other yarn store that I really do like much better...I just was in the neighborhood of the other.  I will laugh if the lady ends up right...haha!  That would teach me!   Basically, I just want to add a \"fancy\" yarn in with the plain yarn on the smaller sections...oh, it is too difficult to explain...if there are knitters out there, let me know, and I will attempt to explain.  Her contention was that the baby would chew on the fancy yarn...okay, one of the yarns I picked out was like the whisper yarn, the kind with the longer \"hair\" strands...but the other was eros or something like that...it just has color blocks in it, nothing would actually stick out.  Anyway, my mom, who does not knit, says she thinks I should do it.  She said everyone always loves the \"creations\" I do so why should this be any different?  I\'ll see what yarn I find today.  '),(19488,518116,'04,August,2004','2004-08-04 00:00:00','   lol...poor Anne, I always \"borrow\" her stuff...just like a sister, eh?     What is something you dislike about yourself? the way I look...most of the time What is something you do well? People say I\'m a good person to tell problems to...I\'ve been told I am very non-judgemental and supportive What is your favorite room in your home and why? bedroom...I get to sleep and participate in extra-curricular activities here...plus, I have a really cool tin-tile ceiling What is a good neighbor? Someone you can be friends or friendly with but you don\'t have to spend 24/7 with What is your favorite time of day? after work...I get to leave work and GO HOME What is your idea of a dull evening? feeling like you want to go do something and not being able to figure out what to do What is the best way to treat meddlesome people? LOL...I learned this a very long time ago...when someone comes up to you and thinks they are going to tell you something upsetting or something you didn\'t know, show absolutely NO reaction and tell them you have already heard that...it is halarious to watch them be disappointed...you can have your reaction five minutes later in private! What is something you are optimistic about? life What is something you are pessimistic about? health What is something that makes you feel happy? family and friends...especially hubby and B What is your most indispensable possession and why? I would be really sad to lose my wedding ring...and my scrapbooks What is your favorite song and why? I don\'t really have one...though B is singing a Spiderman song lately that is pretty funny What is the best birthday present you ever received? I don\'t know...maybe concert tickets...or maybe my day at the spa What is the best birthday present you could receive? a trip What is something that makes you feel sad? sad children What is your favorite book and why? The Sabine and Griffin series What is something that really bugs you? lying What is something that really makes you angry? malicious rumors What is the best advice you ever received? just be the best person you can be What is your favorite holiday? Christmas...and Valentine\'s What is your favorite day of the week? Friday and Saturday What is your favorite month and why? June...it is our anniversary which means we get to go away for a weekend.  '),(19489,518116,'03,August,2004','2004-08-03 00:00:00','   okay...this weekend, hubby and I watched \"Bowling for Columbine.\"  If you haven\'t seen it, watch it.  It doesn\'t matter if you are democrat, republican, or otherwise.  It was extremely interesting and shed light on many things...whether you believe them or not.  It raised very pointed questions and didn\'t necessarily provide answers.  Thought provoking.  Very glad I watched it, a little scary since these nuts showcased had some tie to Michigan, but very good.  Now, I don\'t know if I have mentioned, but I have never voted.  Wait, I did vote, the first time I would have been eligible.  Not memorable as I didn\'t know who I was voting for, I voted Republican since that was who my parents voted for.  Scary, I know.  Anyway, this year, I really feel I need to vote.  The problem?  I don\'t know who to vote for.  I pretty much steer clear of politics.  Why you ask?  Because, I feel, all politicians lie.  They say what we want to hear and then do whatever they want once voted in.  I think Clinton did a lot to disenchant me...regardless, I think the US has some pretty serious issues...and the first place to fix them would be with our leaders.  So, basically, since spring, I have tried to find more information.  Do I want to vote to re-elect Bush?  Do I want to vote to kick him out?  See, I don\'t see it as voting to put Kerry in.  As I mentioned, I think they all lie, so why would I WANT Kerry.  Right now, he is lying...I can\'t tell you if he is good until he has served 4 years.  So, right now, I need to decide if I think Bush did a good job...please, feel free to comment away.  I\'m very open minded.   People never believe that when someone says that.  Don\'t get me wrong...on actual ISSUES, I have mucho opinions.  On actual people...not so much.  At the same time, I can\'t broaden my horizons without being challenged.  I realize that my opinion today may not be the same as it is tomorrow.  IE, I used to think I should vote based on my moral beliefs (abortion).  Then I realized, this is so not a current issue for our government.  By voting for someone who is against abortion, I may be voting for someone who is ripping education and food out of inner-city children\'s reach...and which is more important???  I don\'t see Roe vs. Wade being overturned anytime soon, so I need to be more concerned with the children that are here and who is watching over them...  See?  I told you...I have many opinions but enjoy hearing others.  Diversity is to be embraced... sharing ideas is the only way to come up with one that might actually work...  '),(19490,518116,'02,August,2004','2004-08-02 00:00:00','   So, Friday was B\'s surgery.  We had to check in at 6:30am!  We could tell B was a little nervous but he held up like a trooper.  The pre-op nurse was so cute...Ruth.  She had B help her with the machine to take his blood pressure and pulse readings.  She had an adorable accent and I don\'t think B could understand what she was saying most of the time.  They gave him a medication that reminded me of Staydol...I don\'t know what it was but it took his apprehensions away.  It was funny because he was trying to tell us stories and would forget what he was saying.  The anesthesiologist came in and explained what he would be doing.  He told us B would experience discomfort/pain from the tube put down his throat.  The dr came in and we again asked what type of pain B would have...oh, there is no pain associated!  I asked her how she was going to remove the adenoids (through the mouth) because B had wanted to know.  They came and took him.  Hubby and I were only in the waiting room 30-40 minutes before the dr came back out and said she was done.  B\'s adenoids were significantly larger than in the CAT scan from March so the dr is convinced this was the problem and the removal will solve his issues.  Again, we asked what pain we could expect...oh, none!  The nurse came to get us a few minutes later and showed us to the recovery room.  B was sitting in a chair whimpering.  I asked the nurse what was wrong (naively believing he would not be in pain).  She explained boys\' testosterone and the anesthesia don\'t mix and they wake up upset.  Hubby picked B up so I could put him on my lap and rock.  The moment he realized we were there he started shrieking...and shrieking...and sobbing...and shrieking.  It was awful.  I was not prepared for that (did anyone mention this would happen after asking a million times???)!  Meanwhile, the nurses are telling me I have to watch the IV flow...the faster he gets the fluid in him, the faster we go home.  Finally, after what seemed like a lifetime, they came in to give him a sedative in his IV.  Well, the first one did nothing.  So a few minutes later, they were back with another one.  He stopped shrieking and just cried.  He said his throat hurt really bad.  So, we started giving him water.  He downed it like he was parched and ate a popsicle.  A nurse came in and warned us, the more water you give him, the more likely he is to vomit before you get home.  I guess the anesthesia throws off the equilibrium and having so much water in the stomach adds to it.  So what do we do?  She advised us if it keeps him calm, just give it to him.  I thanked her for the advance warning...I would MUCH rather know what I potentially face and prepare for it than be caught off guard (ie, not being forewarned how he would potentially be when he woke up!!!).  Thankfully, no vomiting...we got home, put him on the sofa bed and turned on cartoons.  Seriously, within an hour he was up and moving around.  By the end of the morning, he was bouncing off the walls and you would never know the child had surgery.  Great, excellent, blessing...yes.  BUT hubby and I had been told he was going to be groggy and tired so he would probably sleep all day.  So WE were expecting to be able to sleep all day...man!  I am so thankful he came through okay...of course, the night before, I had hideous nightmares and got virtually no sleep.  And, yes, I really am grateful that he is back to my rambunctious B.  B and I went to Kohl\'s on Sunday...they had a Superman tshirt on sale that hubby has been looking for to match his tattoo...who do we see?  Mr X\'s mom and stepdad...lordy.  I got my hair chopped off on Saturday morning and have yet to figure out how to style it so I am praying they did not see me!  Does anyone know of any awesome hair straightening products?  My haircut has reminded me why I don\'t do short...my hair is naturally funky wave and without the weight of long hair it goes crazy...ugh.  For example, I straightened and flipped this morning.  It is already wavy again!!!  I know my hair will take a while to re-train to flip but the wave is nasty...any advice is appreciated!  '),(19491,518116,'02,August,2004','2004-08-02 00:00:00','   (27-34 points) You are Tweety:You are cute, and everyone loves you. You are a best friendthat no onetakes the chance of losing. You never hurt feelings and seldomhave yourown feelings hurt. Life is a breeze. You are witty, and calmmost of thetime. Just keep clear of back stabbers, and you are worry-free.  '),(19492,518116,'02,August,2004','2004-08-02 00:00:00','   Okay...so I \"stole\" this from Anne...I love you too, babe!  I am SO thankful for the friendship and support you have provided...and, I love this survey so I borrowed it:)     A Different SurveyName three bad habits you have: 1) Picking 2) overanalyzing 3) obsessing Name three scents you love: 1) coconut on the beach 2) family sauce cooking 3) B and hubby Name something you\'d never wear: 1) my birthday suit in public Name 3 animals you like: 1) dogs 2) monkeys 3) apes Name four television shows you love: 1) Buffy (it\'s over) 2) Newlyweds 3) Extreme Makeover: Home Edition 4) Alias Name 2 bands/genres most people don\'t know you like: 1) oldies 2) jazz Name 3 drinks you regularly drink: 1) water 2) Folger\'s Cafe Latte 3) Pepsi One Name Three random facts about yourself: 1) love to laugh...even at the most inappropriate times 2) I constantly find new moles on my body 3) I have scoliosis and my spine is an \"S\"...ironic? Name 2 random facts about your family: 1) My family may or may not have been in the organization 2) damn!  I can\'t tell the other one...but it has to do with my dad and a man featured in Farenheit 9/11 Have you ever... 1) Fallen for your Best Friend? yes 2) Made out with JUST a friend? nope 3) Been rejected or heartbroken? Yes 4) Been in love? yes 5) Been in lust? definitely 6) Used someone? I hope not...not that I can remember 7) Been used? I think so 8) Cheated on someone? Unfortunately 9) Been cheated on? no 10) Done something you regret? Definitely...though I try not to dwell on it because life is too short and I need to just learn from my mistakes Who was the last person… 11) You touched? hubby 12) You talked to? My boss 13) You hugged? B 14) You instant messaged? lordy...it has been FOREVER...probably Jen 15) You yelled at? B 16) You laughed with? B or hubby 17) You had a crush on? Man!  I just watched Paycheck with my hottie Ben Affleck 18) Who broke your heart? Mr X???  Do you.. 19) Colour your hair? sometimes...not presently 20) Have tattoos? No...if I could figure out what I want for THE REST OF MY LIFE, I would 21) Piercings? LOTS...nose, belly button, six in my ears 22) Own a web cam? yup 23) Have AOL? im 24) What should you be doing right now? working 25) What are you listening to? HVAC above my head 27) Chicken or fish? Chicken 28) Do you have a favourite animal? love dogs 29) Is ice cream the best thing in the world? NO CHOCOLATE is silly… I SO agree!!! 1. What would your dream date be? relaxing day on our own private island...no, my perfect date was going snorkeling together for the day 2. Single flower or a dozen? single 3. Silver, gold or platinum? Silver or white gold 4. Candle lit dinner in a restaurant or at home? At home 5. Roses or wild flowers? Wildflowers 6. Silly or serious romance? we are in a silly phase right now 7. Marvin Gaye or Barry White? Marvin Gaye 8. Do you consider yourself a romantic? yes 9. Who else would you like to see fill out this survey? I\'m not sending it out... Shy? Unfortunately Quiet? yes, reserved until I know you well enough Boring? I don\'t think so:) Funny? I think so...so does B!  haha The Lovey Dovey type? alot of the time Dissected something? ewww, yes, in biology Drank? yup Cut your hair? yup Kissed someone? yup Missed someone? definitely Told someone you love them? Yes Met someone new? Yes What CD is in your CD player right now? Gavin DeGraw Q. Who is your favourite Spice Girl? Posh Q. Favourite Disney Characters? Lilo and Stitch Q. Favourite fast food? Wendy\'s Q. Favourite book? Sidney Shelton Q. Favourite Sports athlete? my hubby just met someone at B\'s soccer camp...I can\'t remember his name:) Q. Favourite song? I have too many...I concur Q. What is your shoe size? 9 Q. What will you be when you grow up? You tell ME  '),(19493,2899366,'22,June,2004','2004-06-22 00:00:00','   Thank you to a Steven Jalil, I\'m at least intrigued for a night.  Anyway, folks, what should we use this team blog as?  '),(19494,2899366,'21,June,2004','2004-06-21 00:00:00','   lol, look at us!&nbsp; urlLink     urlLink  '),(19495,2899366,'17,June,2004','2004-06-17 00:00:00','   i updated some shit for the blog, and everyone should update this!  and while i m here, who\'s chillen on friday,sat, so forth?  '),(19496,2899366,'10,June,2004','2004-06-10 00:00:00','   Should we keep the colors? Looked the most original to me at the time  '),(19497,2899366,'10,June,2004','2004-06-10 00:00:00','   just trying it  '),(19498,2899366,'30,July,2004','2004-07-30 00:00:00','   I changed the name of Summer Romp to Jones\' 18th B-day, and the second summer romp has become the first, and now  only,  summer romp. Not too confusing.    Okay, that\'s all for the photo gallery updates for a while. Oh yeah, and can whoever started this thing put a link to  urlLink  my  blog  on here? It just bothers me, is all.  '),(19499,2899366,'26,July,2004','2004-07-26 00:00:00','   Yes, Jorge, that would be horrible.   As for the chinchilla thing...well, I\'d probably try dancing an irish jig.  '),(19500,2899366,'20,July,2004','2004-07-20 00:00:00','   NO NEED TO FEAR JORGE IS HERE... &nbsp; &nbsp;Sorry I haven\'t updated I\'ve been to busy having fun ith your moms box dave...lol j/k that\'d be horrible. Anywho...I have ntoing to say but I would say this... What would you do if you woke up with a chincilla in your ass?  '),(19501,2899366,'19,July,2004','2004-07-19 00:00:00','   Come on Jorge...you gots to post...  Not like there\'s a significant  reason  to, but...  Nevertheless.  '),(19502,2899366,'18,July,2004','2004-07-18 00:00:00','   Summer romp was acutally Jones\' B-day....could be wrong though  '),(19503,2899366,'13,July,2004','2004-07-13 00:00:00','   Ok, quick little update on galleries  Firstly, can you have the names pop up for each gallery, either old gallery style or stupidink style, so people know which is which  and secondly, post up a list of all the missing pics so I can helloerize them for ya  '),(19504,2899366,'12,July,2004','2004-07-12 00:00:00','   If Jorge goes back out with judith, every single member of the original 5 krew member will be going out with someone, and not only that, but we all would have g/f\'s from antioch too. And, and, and...thats it. O, and you have to count rob out i just realized...just noticing  '),(19505,2899366,'05,July,2004','2004-07-05 00:00:00','   Check this out:  '),(19506,2899366,'01,July,2004','2004-07-01 00:00:00','   Those who wish to obtain the picture code would be wise to check out the comments of the last post.   As for this blog, I suggest that we use it for trivia and other random crap like that. Since no one will protest, I will start it off with some Simpsons trivia.   I\'ll start with an easy one.  Who shot Mr. Burns?  '),(19507,2899366,'01,July,2004','2004-07-01 00:00:00','   i guess this serves... some  inane purpose to someone...   I can\'t offer much in terms of MEAT to this group blog dealy, but I can offer the code I use to put that pic on my posts to people for their own use (and obviously, their own pics)...is anyone interested?  '),(19508,2899366,'08,August,2004','2004-08-08 00:00:00','   Okay, so maybe Rob and Andy were bad ideas (they\'d never do it), but other people, IE Czank, Benny, etc. etc. The thinking behind this is that maybe with more people, we can actually find a real  purpose  to this thing.  Oh, and as for the Photo Gallery, well...too bad.  '),(19509,2899366,'06,August,2004','2004-08-06 00:00:00','   Unhappy was an understatement. Anyway, moving on  Gotta remember to link, and if you can figure out Andreas and Rob\'s emails where they won\'t delete things, by all means  '),(19510,2899366,'04,August,2004','2004-08-04 00:00:00','   We need to get a lot more people in this...Rob, Andy, everybody. I can\'t be the only one who finds this thing kinda boring...  '),(19511,2899366,'04,August,2004','2004-08-04 00:00:00','   Oh yeah...I forgot all about that dude.  I wonder how he\'ll react to the news that the Photo Gallery is gonna be shut down by next Friday...  '),(19512,2899366,'04,August,2004','2004-08-04 00:00:00','   ya, jones started it..he comes back today though, so if he ever gets online i guess,  juss ask  '),(19513,3327596,'25,May,2004','2004-05-25 00:00:00','   Tomorrow is the LAST day of school. I am REALLY excited.  Hopefully I am going to get to spend the night with Jayme tomorrow.  We might go to Austin\'s house to watch the fire.  Some guys at our school got all of the Spanish workbooks and they are going to BURN THEM ALL!!!! It is going to be fun.  Well I have some pic.s on my profile now I put a lot on there tomorrow I have to take my English II test. It is going to be REALLY hard.  well I\'ve got to go I have to go study. BYE  '),(19514,3327596,'16,May,2004','2004-05-16 00:00:00','   This weekend has been boring!  I haven\'t got to do ANYTHING!  So it sucks.  Um yeah this is going to be like the shortest entry in the world, b/c I didn\'t do anything.  Well that\'s all bye.  '),(19515,3327596,'15,May,2004','2004-05-15 00:00:00','   O.k. both of my brothers are in baseball so they have games like EVERYDAY after school so of course I have to tag-along and go to ALL of the games.  It\'s o.k. though b/c all of those little kids are cute and all of the kids on Destin\'s team have a crush on me and it\'s REALLY funny.  Well I  am about to start all my my driver\'s ed piss.  I really want to do it all and get it over with so I can get my driver\'s linces but it is going to take along time to do all of that stupid book work PISS!  Well I guess that\'s all t2yl.   Bye  '),(19516,3327596,'13,May,2004','2004-05-13 00:00:00','   Today I was still sick so I stayed home from school once again.  It was REALLY boring.  But oh well I was sick so it doesn\'t matter.  I just watched t.v. and that\'s ALL it was so lame but what else are you suppose to do when your grounded?  Well this whole blogger thing is pretty cool. I mean I have a dairy that I write in sometimes but this is better b/c now I feel like if people read this then they will feel my pain in some sort of strange way.  Oh well.  Oh yeah yesterday Jessica called me and she was like yeah Lance is always talking about you at school (meaning the school I went to last year) and hes always like \"yeah I\'m gonna take Porsche\' to prom next year I\'m gonna send a lemo to her house to pick her up with two dozen roses inside, the limo will take her to pick me up and we will go to prom\"  Isn\'t that sweet?  Oh well um I guess that\'s all for today I\'ll t2yl.  Love ya,        Porsche\'  '),(19517,3327596,'12,May,2004','2004-05-12 00:00:00','   Well today was sort of bad b/c I was sick this morning when I woke up.  I tried to go to school but I got sick during first period and called my mom to come and get me.  So I sat in the office until she got there.  I was really mad to b/c it was a B-Day and I have ag on B-Days and there are alot of hot guys in my class.  Well ne wayz I got home and slept until like 12:15 then I got up and ate some lunch, watched t.v. then went back to sleep for a few hours then woke up again like 2 hrs. before my family got home.  And that\'s how my day went.  Well I\'ve g2g I\'ll t2yl.  Bye  '),(19518,3327596,'04,June,2004','2004-06-04 00:00:00','   today has been relly boring. I cant talk to jayme b/c she is at her aunt\'s babysitting so the only thing i can do is...............................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................well you guessed it BE BOARD!!!!!! well i have to go bye  '),(19519,3327596,'03,June,2004','2004-06-03 00:00:00','   So far summer has been boring. I don\'t have much to write in you.  But oh well Jayme is always like \"you never write in your blogger any more\" so there you go jayme. Um Jayme you know that you can\'t talk about any thing that you read in here so just deal with it.  Ok dear diary, I think that Jayme gets mad when Robert sits by me and stuff, b/c she still likes him.  She tells me that she doesn\'t like him any more but I know that she does. She is going to get mad at me if he sits by me at the movies.  I told her that he should just sit in the middle, so that way he will be sitting by both of us. But she said that she wants to sit in the middle. But then if he sits by me jayme will be on the outside and I will be in the middle.  I personally don\'t care.  If he sits by me good. I would me happy. But if he sits by Jayme good.  I really don\'t care. Oh well I am about to go b/c I have to get off the computer b/4 my dad gets home.  Well ill talk to you later. BYE!!  '),(19520,3447369,'27,May,2004','2004-05-27 00:00:00','   i\'m so tired. yesterday left me feeling dead. he (david) broke up with me yesterday. i thought i would be ok, because this relationship didn\'t mean that much to me. however, it obviously did.   '),(19521,3447369,'26,May,2004','2004-05-26 00:00:00','   i wish that i could go back, if you know what i mean, to when everything was just straioght foward. life in general. i don\'t know how i\'m keeping myself under control, even though somtimes i wish i could just let everything go. it would help, i think. just to live in peace, only a day. my mind is full of things that pass my imagination. all of these things that i have let happen. all i can think about is how wrong it is. i can\'t think of a way to fix anything. it\'s driving me crazy living in silence. that\'s why i emailed him today. but he\'s nto going to email me back. i know. he\'s just going to ignore everything.  i don\'t understand why everything feels so wrong or why i just don\'t want to be anywhere anymore. i want forget myself let alone forget all the bad things i let happen.  i wish i could make someone understand me. yes, i do talk about things but there\'s other things that are untold and i wish to keep them like that but it\'s them that are my main problem. it\'s my fault, i know that. i just can\'t help thinking that there is no way out of this stupid web of confusion i let myself get caught in. someone just save me from myself.   '),(19522,3447369,'25,May,2004','2004-05-25 00:00:00','   okay it\'s getting worse but that does not stop me from trying to make it better we are uncommonly wrong as i let the thoughts flow right through me... shivering inside as though the cold has caught me my eyes are so sore - i can hardly see you swore that you would be there you promised me that you did care but you let me slip right through your fingers think about it, what you did your lies amount to nothing, awell as everything else you said.   '),(19523,3447369,'25,May,2004','2004-05-25 00:00:00','   one and two,  i\'m nearly out of the door three or four then i\'m lieing on the floor i\'d askyou a question because i\'m wondering why i am so cold.  you came to me with malicious intent you roled into my head and messed me up instead i forgot to tell you that i hated you i\'m at a loss but at least i know what was true...  and i\'m sad, i know but it\'s not about you   i took so long to firgure myself out it\'s hard to know how feel when you\'re knocked out unconcious and cold you\'re lie are now old you intended to make me die   '),(19524,3447369,'25,May,2004','2004-05-25 00:00:00','   one of the set of poems was brand new. their lyrics are clever, they\'re the words i wish i could use. oh well.  i\'m school and i think i\'ve od\'d on cough medicine. this is my method of escape. i haven\'t taken cough medicine like this for ages. i didn\'t need it. i just wanted to make myself relaise what feeling weak really is like. i\'ve done this times and times before. i don\'t mean to harm myself, it\'s to keep myself under control. though you\'re probably thinking that i\'m not after what i did.  who cares? not me. neither do you. so i\'m just going to leave this for now. maybe i\'ll have more to say later. maybe it will be less depressing.   '),(19525,3447369,'25,May,2004','2004-05-25 00:00:00','   i never thought it would be this hard to start right from the beginning it\'s the first time i\'ve actually looked at my progress of becoming out of control...  this is what you deserve when your life is on the line it\'s not from ignoring your mind it\'s from ignoring the signs  you take from me because i give you push me over my own edge you sucked the life out of me  stick to me like cellotape don\'t let me go until i\'m all taken i\'ll drink the ink that creates this scrawl - i\'ll keep the image in my head the sweat, stale sense of what i need to forget.   '),(19526,3447369,'25,May,2004','2004-05-25 00:00:00','   And we learn As we age We\'ve learned nothing And my body still aches And we take Cause they give Though I love you And my body it leaks like a sieve     i am heaven sent,  dont you dare forget. i am all you\'ve ever wanted, what all the other boys all promised. sorry i told. i just needed you to know. i think in decimals and dollars. i am the cause to all your problems, shelter from cold. we are never alone. coordinate brain and mouth. then ask me whats it like to have  myself so figured out. i wish i knew..  i hope this song starts a craze. the kind of song that ignites the airwaves. the kind of song that makes people glad  to be where they are, with whoever they\'re there with. this is war. every line is about, who i don\'t wanna write about anymore. hope you come down with something they can\'t diagnose, don\'t have the cure for. holding on to your grudge. oh its so hard to have someone to love. and keeping quiet is hard. cuz you cant keep a secret  if it never was a secret to start. at least pretend you didn\'t wanna get caught..  we\'re concentrating on falling apart. we were contenders, we\'re throwing the fight i was just wanna believe, i just wanna believe, i just wanna believe, in us.  Oh, we\'re so c-c-c-c-c-controversial. we are entirely smooth. we admit to the truth, we are the best at what we do. and these are the words you wish you wrote down. this is the way you wish your voice sounds, hansome and smart. oh my tongue\'s the only muscle on my body that works harder than my heart. and its all from watching tv, and from speeding up my breathing. wouldnt stop if i could. oh it hurts to be this good. you\'re holding on to your grudge. oh it hurts to always have to be honest with the one that you love. oh, so let it go..  we\'re concentrating on falling apart. we were contenders, we\'re throwing the fight i was just wanna believe, i just wanna believe, i just wanna believe. we\'re concentrating on falling apart. we were contenders, we\'re throwing the fight i was just wanna believe, i just wanna believe, i just wanna believe, in us.  this is the grace that only we can bestow. this is the price you pay for loss of control. this is the break in the bend, this is the closest of calls. this is the reason your alone, this is the rise and the fall.   '),(19527,3447369,'25,May,2004','2004-05-25 00:00:00','   From the time I started to believe that everything would be ok, I started to hurt more You left me outside of my hopes but inside of reality Thereâ€™s no escape from everyone elseâ€™s eyes You can always forgive yourself for what you have done  You lead me astray, I lost my clarity I let you lead me through the rain and to false shelter You spun me around in the disaster; you intended to ruin my dignity Could I set the record straight and ask you to give me the answers  If only I had a notion to notice your intention I could have questioned why I thought you cared Why I thought you loved me you could have given me hope to rise again.    this is the first and last time i\'m going to do this to you. i never thought i could ever feel like this again - i thought everything would be ok. you promised me. oh and one at loss. i bleed. tchau to you.  why can\'t you appreciate that you\'re the one you sucked up to me until i made the decision? you lead me to the point where people think they would rather die. it was when you didn\'t react to my decision that got to me. you acted as though i meant nothing to you. and oblivious to you - you\'re the one that should be feeling like shit. yesterday you said i deserved to be smacked, well, it beats me why i derserve it. you deserve everything you get!  \'\'whatever poisions in this bottle will leave me broken sore and stiff -\'\' yeah. that\'s brand new.  everytime you said that i meant everything to you, that you couldn\'t do without me, that you would try anything to be with me, i made myself believe because i wanted to know and be happy. i\'ve been hurting ever since that day we got back from paris. you asshole. you made me feel like killing myself. you had no sense of what i was feeling. you left me in the cold. and it got colder when you left. actually, when i left.  i\'m going to write more later.  i bleed.   '),(19528,3903607,'12,July,2004','2004-07-12 00:00:00','   We have a few days left in Japan before flying to Germany. To round off our trip here we are taking the train to Hakone tomorrow to catch a glimpse (hopefully) of Fuji San.  Jess\'s boyfriend, Ben, returned from his work trip the other day and since then we have had much more of the local beer and seen some interesting nightscapes. Ben is fluent in English, Japanese and American.  Prior to Ben\'s return, Louise and I were using \"pigeon-nippongo\" to get by, which apart from being frightfully funny, is completely unhelpful. A request for a ticket from Yogi Bear station (Yoyogi) to Dog-on-a-tuckerbox (Takadabobaba) would be met with a long-winded, rehearsed and passionate account of Japan\'s imperial history, with particular emphasis on the nasty invading hordes, who they called \"gaijin\" and who they clearly did not care for much.  Apart from the ticket guy, everyone here has been incredibly friendly and considerate. I personally love this place, life here is so different to Australia - however - from what I have seen, Japanese people tend to spend much more of their time existing than living. There are so many places where it is not socially acceptable to talk (like trains, small retaurants or any confined spaces), and their culture and code of conduct seems so strict that quirky features have popped up (like all-age comic book reading, rebellious fashion, and popular fetishism).  Having said that, we are yet to meet a young Japanese person who is not cool. They all seem to spend so much money on their clothes and haircuts, it has put Louise and me to shame.  After I finish this blog, we are going to the world\'s largest fish market near Ginza. We have been warned to wear old shoes. The Tokyo Lonely Planet reports that some tuna fish fetch 2 million yen each (nearly 13 thousand AUD)! I will settle for just a slice at the local Sushi Train, thank you.  Until next time, please feel free to post your comments! Love to all.  '),(19529,3903607,'12,July,2004','2004-07-12 00:00:00','   We are in Tokyo (Iogi actually) staying with our great friend Jessica Smith.  It was blisteringly hot yesterday. We only managed to walk for 15 minutes to the train station before suffering from heat exhaustion and deciding it was too much effort to actually board a train. We took refuge in a supermarket instead where it was cool and browsed the isles looking at all the different types of local food. It was great fun and we spent nearly 2 hours at it!  We have spent some considerable time sampling the local beer from the comfort of Jess`s massage chair and hope to get out and about in the next couple of days.  We have no access to the mobile phone network here. Sorry if you were awaiting a response to a message you have left us...  '),(19530,3903607,'12,July,2004','2004-07-12 00:00:00','   urlLink    Ben and Louise atop the Grand Canyon, USA, January 2004.  '),(19531,3903607,'12,July,2004','2004-07-12 00:00:00','   There must be a beginning - and there must be an end. This is the order of things.  Let me start at the start, skip ten billion years, and take you to the present day...  Melbourne CBD, morning. 13 degrees and overcast. Level 10, recently renovated residential block annexed to the soaring green-topped IOOF Building. Apartment 1007 â€“ our home for the past 18 months. Sitting at the dining room table, facing north.  I have been \"released from my retainer\". Current position: man of leisure (and mystery).  Weisie has charged me with the responsibility of undertaking some important administrative tasks necessary for our Big Trip. She is half way through her last day of work, and thinks I am conscientiously completing the list of tasks she has left behind for me. Silly of her really.  [Weisie, if you are reading this, the previous paragraph was inserted to give our opening Blog some pizzazz, or as they say in Sydney, \"zhuz\"].  Welcome to Weisie\'s World. Enjoy the trip.  '),(19532,4303113,'22,August,2004','2004-08-22 00:00:00','    urlLink    On top of the Arc de Triomphe&nbsp; urlLink '),(19533,4303113,'21,August,2004','2004-08-21 00:00:00','    This is, of course, my first posting.  I went to go see the new Exorcist movie tonight.  I don\'t know why I do this to myself.  I won\'t be sleeping until daylight.  I\'m not particularly afraid at the moment.  The fear won\'t really set in until the tv is off and the lights are out.  There seems to be something about unnatural body movement and the forces of evil that just doesn\'t sit well with me.       In other news, I\'ve really done nothing significant with my weekend.  This is, afterall, the part of the week that I live for, so, normally, I like to have plans.  I thought that I would sort of let things fall into place this weekend, instead of making plans.  The end result was that both Friday and Saturday night, I didn\'t really getting ready until around 11:00 pm.  I didn\'t really give myself enough time to have a productive night.  I did however, spend the entire morning with the cable guy who came to install my high speed internet.  He kept talking to me, he swore that he knew me from somewhere, maybe the mall.  I explained to him that I haven\'t lived here for the better part of the past five years.  I thought maybe he recognized me from high school, but he graduated seven years ahead of me, in that case, I really hope he doesn\'t know me from high school.  I don\'t have much patience for prowlers.       Anyway, having spent so much time with him, I really felt like I got to know him.  Our casual conversation led to a very awkward goodbye.  I didn\'t know whether or not to lean in for a kiss.  In the end, I just gave him a check for my first month\'s service.  I think we did the right thing...No, just kidding, we never got that close at all...he\'s just the cable guy, a stranger in my home.       After the internet was up and running, I frittered my valuable day away on various web sites, trying to find a different job, trying to relocate, trying to find a way out of Bay City.  I don\'t know how successful I was.       In any case, I should go to bed soon so that I can get all my good nightmares in before morning. '),(19534,2787343,'31,January,2004','2004-01-31 00:00:00','    This is typed from another 10-minute break. We still have 2 more rounds to go... Man, I\'m getting tired. Some of the original players had went to play xBox instead. So, the 2 of them are replaced by another 2 cousins from far far away. They are sort of having holiday back in California, so they just flew here, since they hadn\'t done so for nearly 10 years. The first thing I found weird about them is that, while the local people are finding it sort of a freezing night, they insist that we should on the air-con. Brr, we are sure in a freezer now, and they still complaint that it\'s warm! (it\'s 24 C)   What is most surprising is, they have fluent Chinese and many Chinese cultures!!! What a shame to Singaporeans! They are living in an \"ang moh\" country, and their Chinese can still be so good! Sigh, better whip out my Chinese now, it\'s sort of getting low-standard...   '),(19535,2787343,'30,January,2004','2004-01-30 00:00:00','    I\'m playing mahjong now with my cousins. We are taking a 10-minnute break now, giving me time to update this blog. I\'ve have won most of the rounds!!! Although we didn\'t really gamble real money today, I\'m still very happy. Anyway, we are going to play 12 rounds since we started at 6pm. We have just finished our 7th round, so a few more to go. Mahjong is one of my favourite games, since I always win.  It has stopped raining today, what a disappoinment!!! It\'s starting to get real hot. Haix... I had hoped that this rain will go on. Okay, this break is nearly over, better get going. Got to finish 5 rounds by 12 midnight.  Oh! I nearly forgot to talk about PE today. The PE teachers wanted us to run 3km, 80 push-ups and 40 sit-ups! It\'s killing us!!!!!!!!!!!!   '),(19536,2787343,'30,January,2004','2004-01-30 00:00:00','    Okay, this is my first blog and supposedly my first diary, unless you want to count the dream diary my teacher told me to keep. She said it can be good for literature purpose, and she got the idea from the Harry Potter book (supposed to be a divination class).   Anyway, it\'s dear old ks who told me to blog. He actually BEGGED me (no offence meant). I don\'t understand how he can be that desparate to read some other people\'s life! (Maybe he want to see how i die in school everyday.) As usual, lessons are boring today, especially maths (the prime, rational, complex, numbers are driving me nuts). And again, there is a lot of homework. Some pupils told the teacher that they are Muslims and celebrating Hari Raya and the teacher actually believed it! (Mind you, mine\'s a Chinese school!)  Also, nothing special actually happened today, unless you want to count a classmate being late for lessons and the teacher locked him outside the classroom. Another thing which is worrying me, the chosing of EP3s! I\'m still pondering over what to do.   Today, we discussed about the IP programme too. I gave in my opinion that parents are given more competition as they are usually \"hao lian\" when their kids are in the IP. That\'s more stress for the pupils. The most scary thing today is still the Chinese test! It\'s so difficult! I\'m sure to fail...   '),(19537,2787343,'28,February,2004','2004-02-28 00:00:00','    There\'s nothing much to write this time, just one word to cap it all, \"busy\"! Business (bu-si-ness) has crept all over, with a hetic lifestyle all the time. Teachers won\'t care about you at all, all they have to say is just 2 words, \"Do Project!\". I haven\'t scored much for ACE in Mathematics, only a pathetic 1 point out of 20 for the whole year, I HAVE TO WORK ON THAT!  In my school, we have scored exceptionally well for O\' Level again. We managed to push the MSG (not the seasoning for food) to 1.09 for Mathematics, that was incredible!   According the ks, SP2 for Windows XP (\"Do what you cannot do in the past, now with Windows XP.\") is coming out soon. It has a new build-in security centre, checking about firewalls etc. Even if you go to the  urlLink Mircrosoft  website, there will be the three ways to protect ur PC, so there actually isn\'t a need for it.  Today\'s reports (in fact the last few days) have come to an end once again. Everything that has a beginning, has an end... (Copyright \"The Matrix\" movie, Warner Bros. Company)   '),(19538,2787343,'25,February,2004','2004-02-25 00:00:00','    Sigh... I\'m still doing my project work now. Feeling bored? Of course, but I have no choice. Really, blogging has sort of become one of my hobbies.  When I look back at my old posts, I actually wondered how I had typed all those in a short time without writing a draft. I just login, and start typing all the words out. This blog is also typed with no preparation at all.  Once again, I must say that time is flying real fast. It looks like just a day ago when we celebrated New Year\'s Day,  but let me remind you, it\'s already the END of FEBRUARY already, and we will be into MARCH soon. Term 1 has already passed (primary schools starting to have exams already) and it will be the March Holidays soon (although only a week, but...).  Today, my computer teacher wants us to do an online digital portfolio. We have to do some profiles etc., and put all our project works and assignments into it. Sigh, I must really get into using Dreamweaver and CSS. Take a look at KS\'s blog, it\'s cleverly designed using tables and of cause CSS. (I think I will copy something from there! But he keeps putting the copyright thingy, so i have to ask permission first.)  My History test and Geography test results has came out, while others are preparing to take the test. I have managed to scrape A1s in both subjects (heng ah). Tomorrow shall the the results for the Literature test, now now, I\'m not looking forward to it...   I\'m writing a journal now (for teacher of course). It\'s a good thing that I blogged, so I can just \"copy and paste\". My next blog update???   '),(19539,2787343,'23,February,2004','2004-02-23 00:00:00','    KS has a new look in his blog, spare me for this lousyish design of my blog, but there\'s really no time to redesign. Okay, i admit it, i don\'t know how to, because the templates are already provided, i\'m wondering how can we edit it, because i don\'t see the button \"Design your own blog!!!\". HTML coding the new blog is really too tedious...  I know, i have not updated this blog for nearly a week. But there was simply no time. Look at what time this blog is posted. Although i\'m still online on MSN, but I\'m actually doing powerpoint presentations etc. for ACE (activities for curriculum extras, i\'m just joking, don\'t believe it) points, because term progress report coming soon, if i don\'t get the credit points, i\'m not surprised if i get C6 (wah lao, C6!!!). I can say that my Maths is a complete failure, there isn\'t much u can do on!  Sec 1 camp is coming soon, and i can\'t say i\'m enthusaistic about it, because camping is not my cup of tea. My real cup of tea is travelling (yes my american cousins, British spelling for \"travelling\" is different from American spelling). I also don\'t know why I love travelling so much, especially on planes. People are scared of long-term flights (long-term means 10-16 hrs, not 6-8 hrs), but I don\'t find them tedious. Yes it can be pretty boring (my family\'s a cheapskate (in fact we claim air tickets using points, becuase my Dad travel for business a lot) and always ride on planes without personal TVs and other luxurious comfort, so i actually never sit before Singapore Airlines, which is a great way to fly). Back to the topic again, sec 1 camp i think it\'s all a bluff... I will comment on what it\'s like when i come back from it 2 weeks later, i promise a longer blog.  School updates have come to an end now, so there\'s nothing else much to talk about actually as I always stay at home and powerpoint here and point there. Goodbye ;)   '),(19540,2787343,'18,February,2004','2004-02-18 00:00:00','    *Ok, Jensen, my total project count is 6. How many do you have?*  Mine can be called either a lot or a little, ks, depeding on the \"doer\" of the project. We are supposed to write a 2000-word Major Research Paper, I have a partner, and have 8 months complete that, so I don\'t worry about that too much. Then we have Project\'s Day, doing a major project on any topic you want as long as it\'s under the appropriate categories. I have another 3 members in the team, but in this the deadline is within 3 months, so actually I\'m quite worried about it.  Now for the \"minor\" projects. I have done 2 Geography projects, 2 History projects, a Maths project, and 2 Chinese projects to be completed. Oh wait a minute, I still have a Geography project to be completed by next Monday!  And... *computer studies! I can\'t believe it. They are learning more and more complicated things, and soon, they would be learning the crazy language called Javascript.  I won\'t be surprised if alot of people don\'t understand Javascript. It\'s very odd language. Maybe Jensen and their school would find it easy, as there are \"professional\" trainers to aid them. Currently, they are learning Dreamweaver. And their version? MX 2004. The latest and the greatest.  To be frank, I don\'t find great difference between the Dreamweaver versions of MX and MX 2004. Can anyone tell me the difference?(  Now now, ks, don\'t you get jealous. You are born a com expert, so you just have to glanced through books from the library and you will be Professional already. But we have to learn the hard way, because we have no base knowledge. And pls ks, CHS ain\'t that rich as you think. We won\'t have the money to hire professionals! Javascript we do know all those simple simple type one, so you actaully don\'t need to get jealous at all!   '),(19541,2787343,'17,February,2004','2004-02-17 00:00:00','    ks, so good! You all still running 1.6km, we are running 2.4km already!!!!!!!! Our last common test is on Friday, Literature. Computer studies today was toughy, with all those questions only ks the computer-maniac can answer. History was difficult too, because of the inferring question. Secondnary school life is tough, especially in CHS.  Same as ks, the blogs in future will be completely different, what i mean is shorter. Projects are now bombing us everyday, and the bombers? The teachers of cause... Now I don\'t believe in the crap that the Integrated Programme is supposed to be fun and joyful... IP means more stress because of projects!   '),(19542,2787343,'15,February,2004','2004-02-15 00:00:00','    What do you mean, ks, that I\'m not as bad as you think??? Now now, I\'m not really \"alien\" to computers, or I wouldn\'t be sitting here typing this blog now. (We are really starting to get into an arguement.)  Anyway, I will sit for my Geography test tml, so naturally I\'m studying now. Life in CHS isn\'t really that bad, there\'s very few homework, but many research work and projects and reports to do. And yes, CHS is really one of the leading schools in Singapore but that doens\'t mean there ain\'t any lousy students here, and I\'m a really good example.   Soon, March holidays willl be coming, only 4 more weeks to go! Again, I must say time fly past very fast, before we know it, it will be 2005. See, we are finishing Term 1 already. And very soon I will go back to my primary school to visit my dear old teachers. Poor old Mrs Lee teaching P6 again, so she has to work hard to produce star students.   PS. I must say Mrs Lee was very fierce to us last year, but now now, we do understand that. Now in CHS, we can score if we speak up in class! I must remember to tell her that.   '),(19543,2787343,'14,February,2004','2004-02-14 00:00:00','    *** Then I changed it to computers. Dear Jensen has changed it to entertainment. And he spelled \'favourite\' wrongly. And he\'s from one of the best schools in Singapore. Believe it or not.  Ok, i\'m not killing you. I\'m really surprised that you can still update your blog sorta regularly. Good for you. What are MQs anyway? Erm, what I said just now DOES not MEAN that I don\'t want to stop you from blogging. Keep the posts coming! ***  ks starting to personal attack me! Look at what he said in the last post. Anyway, so what if I\'m from one of the top schools? There are lousy students in a top school, and i\'m one of them. I \'ve not been performing well in tests (MQs, stands for Multiple Questions) at all. The most I every got was A2, and ks got many A1s.  By the way, I have changed the template of this blog. The reason is really simple, I like this template better (boo to those who can\'t guess it).   PS. I actually never heard of the movie MouseTrap or whatever. I just thought of entertainment because ks wrote \"a world without coms is chaos!\" (fully copyrighted by ks, and for ks, I\'m just stating an example, don\'t prosecute me). But it\'s quite true, a world without entertainment is even more chaotic!   '),(19544,2787343,'12,February,2004','2004-02-12 00:00:00','    It\'s Saturday today again, time flew by indeed fast. And it\'s Valentine\'s Day once again, but I rather recognise it as Friendship Day. This shall be a relatively short blog. Notice that I usually update the blog every alternate days, because MQs after MQs are coming. This week we have MQs nearly every day. (ks, pls dn\'t kill me.)  I\'ve been talking about ks for quite a few times (in fact many). He\'s a friend who is also a old classmate. He can\'t live without computers, and the docs still can\'t diagnose his conditions (haha, not meant to be a offence). His favourtie tagline? \"A world without computers is chaos!\" I didn\'t much agree on that, they are no computers during my parents\' generation, but I\'m still here typing this blog. But I really do agree that we are getting more and more dependant of computers. (Are computers smarter than humans???)  Presidential Election is coming soon. I am unfortunate enough not to spot the Senators giving pep-talks. But my mind is set, it shall be the one from Aphelion. I\'m looking forward to free movie previews, ice-cream etc. A world without entertainment is chaos... (Copyrighted by [Me!!!] 2004)   '),(19545,2787343,'12,February,2004','2004-02-12 00:00:00','    Didn\'t update the blog again yesterday. I will sure get a scolding from ks soon... When? I don\'t know. But I will know that I soon will be in hell! He can be very fierce! I notice that many blogs have music now. I can hate that sometimes (if i say hate it very much, i will get double scolding!), because the songs I listening will become pop-corns popping!   I had test after test yesterday. First 2 hrs to write two Chinese essays, followed by 40-min break, then English essay, then 10-min break, then Maths test. I can\'t tell the details of the tests here, if i\'m found out, I will be cruxified! School privacy statement and copyright order... But i can say that I will fail steadily in Chinese essay!  Tomorrow will be celebration of Memorial Day (in loving memory of ã—ã‚‡ãªã‚“ãƒ¼ã¨, or Light of the South, which was once Singapore, you know what i meanï¼‰. We are supposed to watch the movie \"The Pianist\", don\'t know what\'s it do with any of the 5 defences. I was hand-picked today (so honoured) by my English teacher to be part of the panellist tomorrow for a terrorist debate. How unlucky can i get!!!   '),(19546,2787343,'10,February,2004','2004-02-10 00:00:00','    Tomorrow we shall have Chinese, English and Maths test. We are informed last minute once again! Chinese and English are testing compostiotion, so don\'t need to revise also can, just remember the format. I then tried to take out my Maths textbook, but found that we transfer notes online instead. But there ISN\'T  a thing about the laws of Mathematics!!! OMG, how am I going to study without understanding them???   This is another problem with having no textbooks! We can\'t refer to the, when we don\'t understand some of the things!!!   '),(19547,2787343,'09,February,2004','2004-02-09 00:00:00','    Just recieved a geog ws from the school\'s online system. I wonder (this should be in a cold manner), what on earth is good about the school online system when teachers are allowed to set homework lsat minute and to be handed up by tml?!?!?! Maybe I will write to the Senators to tell them this problem. The school online system should meant only for lesson notes and assignments (actually I would prefer school without big assignments at all) in advance. Improve on that, Students\' Union!!! Anyway, the union should do what the name tells them to. They should ask teachers to give less worksheets and strive for no tests. (\"What on earth is a school without worksheets???\" famous quote from parents) But the stress is really kicking us down. Will parents rather have more tests so that they can \"hao lian\", or rather have children committing suicide?  This is the problem of parents. They want their children to STUDY, STUDY, STUDY. And then children will study until they DIE, DIE, DIE!!! I must still stress that school should be a fun place for us all to learn new things in a fun way, not worksheets and worksheets and more worksheets!   '),(19548,2787343,'08,February,2004','2004-02-08 00:00:00','    Today is another boring day. The Assembly was full of those election stuff (i\'m absolutely not interested in politics, even school politics). There will be multiple questions (MQ, or more well known as tests) coming soon, i hate tests!!! Why can\'t schoo be more interesting and do away with those examintations, tests and spellings etc. I hope school will be a fun place for us to learn, where we learn stuff in a fun way such as games etc. I will certainly enforce this if i\'m the Minister For Education, but again i\'m absolutely not interested in politics.   Today\'s Assembly focussed on the Senators of the coming election from four different parties, mainly Aphelion, ProEd, Ortus and iSpark. They gave us rallies today and i have a good mind who to choose already. Many of my classmates agree too, and I really hope he will win, because we will sure have a lot of advantages! This particular Senator has lead his opponents by such a big margin, but i assure you he ain\'t John Kerry or whatever.  Met Norvin today. He insists that there isn\'t an ez-link card inspector now, and he said that there isn\'t a ticket for you when you pay the bus fare by cash. When I retorted him on that, he said that he has won me in PSLE by a large margin. And he insist that it\'s Hua Zhong Di Zi and not Hua Zhong Zi Di. He also told me that he finally got an email account, I cogratulated him on that. What a blur one he is now...   '),(19549,2787343,'08,February,2004','2004-02-08 00:00:00','    sry for not updating the blog, but i was really busy, busy, busy. Busy as a bee! I took out some of my PRECIOUS time because i got a painful scolding from ks. (wah wah wah!!!) As usual, the PE teacher yesterday made us ran 2 miles, did push-ups and sit-ups (so many that in fact i lost count). Homework was quite okay this week, not as many as last week, thankfully i did most of them in school, so practically there was no homework.  You may now ask me, what was I so busy with if I have no homework? The answer is not playing (boo to people who think that!), but finding something for my research. It\' stupid really, the system in my school, damn confusing. Also, i\'m revising for my Maths test, it\'s gonna be tough i suppose, after doing the mock test. (what is the distributive law? what is the associative law? what is the commutative law?) Maths is really much much toughy compared to dear old primary school days i\'m missing. You will really lose out during the test if you don\'t use calculators, a delightful machine invented by some godly person. Gone were the days when people do simple maths for trading... Comprehension was also toughy in sec school. You must write a summary etc etc. Sigh, how i wish i will be Peter Pan and never grow up... There\'s another reason behind not growing up too, i dun have to worry that people will call be an old bastard who won\'t die when i\'m old.   '),(19550,2787343,'05,February,2004','2004-02-05 00:00:00','    Well, it\'s of course another boring day at school, unless you count a student late for every lesson for the whole week, all the teachers were so frustrated with him that he was kicked out of class during lesson. Got a scolding from my English teacher (AS A WHOLE CLASS). Finally found the Adrian Mole book, it was damn disgusting, read it and you will find out what i mean.  Today\'s also the last day of Chinese New Year, we ate some fantastic dinner at home. ks told me he got abalone today!!!!!!!!!!!! Anyway, I noticed that he started liking to use \"man\" in his blog. Visit it at  urlLink http://kschua.blogspot.com .  My American cousins going back tomorrow. They will be taking Singapore Airlines SQ20, which means they will be one of the first ones to take the first non-stop flight to Los Angeles. AND, I would like to mention something prior to their stay. They actually said that my spelling is horrible!!! My American friends, there are do people, in fact much more, who uses the British spelling instead of the American spelling, this often sparks up some confusion, as MS Word also uses American spelling, it can sometimes be so sickening. Another thing, they saw this date, 05/02/04, and they actually thought it was 2nd of May!!! (Americans put their month in front of the day.)   '),(19551,2787343,'04,February,2004','2004-02-04 00:00:00','    The sickening bird flu is spreading like mad. A new rule was introduced at my house starting from today, no brining of parts fo chicken into the house, whether be it drumstick, chicken wing or chicken faeces.   Again as usual, school is damn boring today. We got a talk on project work... And my Chinese test results had came out, I mangaged to scrape an A2, not bad... The problem is that we have to do 20 corrections for each incorrect question!  Homework? We are supposed to find some sort of boring words...   '),(19552,2787343,'03,February,2004','2004-02-03 00:00:00','    Some of my friends have told me that they are real bored! I felt really surprised, we are in SECONDNARY school, and my amount of homework daily has reached an alarming rate, yet they are still so free! Well, how I envy them...   Hey, to those who are so bored out there, I think you better check your planner carefully! If there\'s really nothing to do, may I introduce a truly fantastic website,  urlLink http://www.sleepinginairports.com . It\'s certainly a site to kill time, and you will wish you have more time instead.   '),(19553,2787343,'01,February,2004','2004-02-01 00:00:00','    It\'s another boring day at school, even with Literature and Computer Studies today. These are supposed to be my few favourite subjects. Literature was boring enough, with the completion of a worksheet on \" Lamb to the Slaughter \". Computer was okay, except for the extremely difficult task of forms and CGI scripts! (bet ks will love this)  The homework is also as usual... Sigh, how I hope the December holidays will come soon, and we can say we are moving quite fast. It\'s already February, and Chinese New Year going to be over... Why can\'t we have summer holidays in Singapore? (since it\'s summer the whole year, we can don\'t go to school!!!)  Another thing. We are talking about that  sickening  bird flu on it, and we are supposed to write 1 foot about that! I\'m still pondering over what to do... What to do, what to do? I think i can relax myself with a bit of comedies, especially like Mr Bean, but i dun have the disc... Maybe I can laugh a bit with Home Alone too, but the problem is time!!!   '),(19554,2787343,'01,February,2004','2004-02-01 00:00:00','    Stress, stress, stress!!! The homework is driving me crazy!!! I have to write a 1 1/2 foot long essay about Chinese New Year!!! After that, I have to write a proposal on my research (you may be thinking i\'m in a university, but i\'m not)! Luckily it\'s Hari Raya tomorrow, so i still have time to complete.   It\'s really sickening! They are going to Sentosa today, but because of the alarming amount of homework I have, they won\'t let me go!!!!!!!!! God knows what will happen next... Better complete my homework soon, or they won\'t let me go to the zoo tomorrow too!   '),(19555,2787343,'24,March,2004','2004-03-24 00:00:00','    Nah, didn\'t update this blog for quite a while eh? Well, busy busy busy... Even in the holidays, I got myself engaged in so many projects! I\'m currently pressurized by a MAJOR Chinese Website Designing Project. It\'s a National competiotion, so you can expect the standards. Computer Studies is getting quite sucky too, i think except for ks. Now we are learning Flash, and I\'m supposed to draw a car entirely using Flash and animate it...  Besides these project (many many more), I have another thing that is stressing me out because of concern and anxiety. Now now, ks, tests in Chinese High doesn\'t start the test on WEEK 5, but on NEXT TUESDAY. Plus, we already got a full load of common and class tests on all subjects last term, so of course it will be quite sometime before another round of studying and tests starts. Chinese class test is on Week 5, I suppose...  Progress Report is out already. I never thought I was going to make it through Mum, but surprisingly, she wasn\'t too upset by my atrocious results. I got 2 A1s, 2 A2s, 1 B3 and disappointing enough 1 B4. How I hope Singapore would use the American 10-point grading system...  Now comments on the holiday...  I really love this quote, \"Everything that has a beginning, has an end.\" Last was the beginning of the holidays, we have already reached the end. Luckily I still had time to enjoy myself a bit. Besides doing projects, I was able to spend some time to play Final Fantasy VIII. Yes, I admit it, I\'m quite a \"fan\" of Final Fantasy, but I still like 8 the most.   I broke my record and completed the whole thing in 43 hrs, without even losing to any of the monsters, lose to any bosses, or fail my mission etc. etc. I also broke the record by deafeating Ultimicia (plus her GF and her final form) in just 18 minutes, but this may be luck, as I got the all the 3 characters I want in my initial fighting team. I also broke record of deafeating Rajin in just two moves (using Aura magic and Squall\'s Limit Break). The 19 soceresses at the end of the game proved to be tougher this time, as their HP all increased to an average of 10, 000, I don\'t know why this happened too, on avearge it should be abput 4, 000.  I managed to attain all possible GFs, without even using cheat codes.   Okay enough with the final fantasy thing. Now is the end of this entry again...   '),(19556,2787343,'14,March,2004','2004-03-14 00:00:00','    ***The new website should be coming just after 7 days (which means, after holidays). The new URL is : http://kahswee.netfirms.com . Ok? Please do update your bookmarks and Favourites list. Thanks a million!   Hint #2: Fade In Fade Out.... Ah, so smoothing.. and what\'s this? Oh, I can click this?  Hopefully it gives you alot of idea and inspiration on how the site looks like. Jensen, please do shut your mouth up. Thanks!  Ok... T-Boz is off to camp. Jensen just came back from camp? Me? Camp? Sure! At home, with homework as food. Supposedly, the paper tastes bitter and sour. Bleh! Homework I\'m doing it of course, and I should have finished my homework by Monday, and I will have to study all my subjects soon after. The 2nd class tests are coming, people! It\'s held during Week 2 of Term 2. Main Subjects to focus on: Science, Science and Science. Geography and History too. Oh yeah, speaking of History, our actual History teacher is here and her name is Mrs Raja... she\'s said to be on materity leave, who knows? She\'s quite fierce and when she said \"I don\'t want History to be a boring subject... I want the whole class to participate...\" the whole class was quiet. Normally, 2 periods of History on Friday would mean Chaos. Now? Even an ant marching could be heard(literally). And she drones and drones and drones off. How can History not be boring?  AH, I\'m feeling restless. Here\'s Hint #3:  The 10 laws of life.   Ok... enough talking.  It\'s raining plitter platter for the past few days and some areas of Singapore got flooded. Poor thing. Luckily we would not have any floods as our house is on high land... so if our house is flooded, the rest of Singapore would be, too.   We are lucky for the time-being.  Please do note the words \"time-being\".***   Right now, I am going to give a lot more hints. Yes, in fact i have already a sneak preview of his website. Now I\'m going to reveal it.  1) It\'s great. 2) Nearly flawless. 3) ks a lot of \"mo shui\", a lot of da dao li one 4) ks as dashing and handsome as ever  Now, if ks house is flooded, mine may not be flooded. **Hint 1: Yes, I do live quite near ks, but mine is even more high-rise than his house (he lives in a, well...). Our area is supposed to be the highest in Singapore already. If even good ole B*k*t T*m*h (sry, but no revealing of place) is flooded, is quite true that other parts of Singapore must as well be flooded.  This update is mainly for ks, so... Everything that has a beginning, MUST has an end. But this shall not be the end, this shall be the beginning of everything...   '),(19557,2787343,'11,March,2004','2004-03-11 00:00:00','    Okay, here\'s a full (so full that maybe you want to stop reading half-way) update of the camp. Alright, I admit it, I was too tired yesterday after coming home from the camp, to update this blog. Maybe I will just use a word to describe it, sucks.  1st Day (Monday, March 8, 2004)  This is probably the worst day of the whole 3-day camp. The idiotic rain kept on pouring and pouring, without stopping at all. Luckily we got a big tentage (the one used for last year\'s Mooncake Festival at Chinese Garden), so we all had out activities inside. But what on earth is camping when you stay the whole day in the tentage, without even going out of it enjoying some fresh air? Even at night, we can\'t set up our tents at the campsite as planned earlier. We had to sleep on grounsheets, and it was freezing that night. When we had our song session that evening, the rain got so bad, that our very own principal had to teach us to do the \"anti-rain\" dance. In the end, we can\'t even have a bath, after participating in all the activities and feeling sweaty and sticky. What a nite...  2nd Day (Tuesday, March 9, 2004)  It finally stop raining this day. We were so happy that our anti-rain dance worked. There are more lookouts for activities today, but we missed our dragon-boating, awww... But there was still water activities, which were quite fun. First, we have to cross a muddy and dirty canal, walking just across a rope. I unfortunately was one of those who fell in, and when I came up, my white school shoes becomes a black sports shoes. Next was building a raft out of some pipes and some floats. Of course we got wet, we were sitting on the floats.  Showers were horrible. There were 4 big buckets in the area where the wash-basins were. We were supposed to stripd naked in front of everyone, scoop up water from the bucket and take a so-called bath   At night was the campfire, all classes must put up a peformance. I think my class\' peformance sucks a lot. The 3 judges (made up by student leaders) were American Idol style, with Randy, Paula and of course the \"infamous\" Simon Cowell. \"Simon\'s\" remarks were so damn painful, \"You dare to call that a performance? Somehow I feel more entertained looking at the trees.\". The campfire ended at around 12 am, we were so tired that we fell asleep. We were allowed to pitcht tents this night, but it was still an awful squeeze.  3rd and last day (Wednesday, March 10, 2004)  Today there\'s the bidding game. Each class was given $300 (faKE OnE), and we were supposed to bid for some REAL foodstuff and things like that. My class was stupid enough to bid $150 for some plastic cups, some broken spoons and some rubber bands. It was classified under mystery items, so we didn\'t know that it was that.  We were finally allowed to go home, the first thing I did once I got home was soak myself in the bathtub and took a luxurious bath.  The camp altogether finally ended, so called everything that has a beginning has an end. But this will not be the end of the good and awful memories of this camp.   '),(19558,2787343,'06,March,2004','2004-03-06 00:00:00','    It\'s a week to a one-week holiday soon. I\'m actually not excited at all. One week? What can you do with only one week? More tuitions? One week... Can\'t they extend this let\'s say 2 weeks? One week shall be spent in doing all the homework and few periods for playing...  I\'m sad to announce that holidays in other countries is often much much more than Singapore. In Western countries, they have 3 months summer holidays (ends after Labor Day, but it depends in which state) and 1 month Christmas holidays. That is not the end. There\'s still nearly 1 month of Easter holidays, if I\'m not wrong. Plus, there are many public holidays in these countries, unlike Singapore, although it\'s multi-racial.  Camp is on Monday, drawing closer and closer. We will be camping at Chinese Garden, Jurong East. I don\'t know why they chose a tourist attraction spot for camping. And it\'s near the lake, means a lot of mosquitoes. I\'m really really NOT looking forward to it. I rather spend a couple of hundereds to go overseas for traveling and stay in a hotel. Camping, in my opinion, is absolutely not fun. Traveling is the real fun thing, wish I could take part in The Amazing Race...  This is just another report in my life. Life is just an illusion, nothing is actually real. Even the chair you are sitting on right now isn\'t real. There\'s no such thing as the real world. Everything that has a beginning has an end. The start of this virtual world shall come to an end one day, which day? No ones knows. End of this virtual world shall be the beginning of the real world.   '),(19559,2787343,'03,March,2004','2004-03-03 00:00:00','    If you have read KS\'s blog, well.. He got 4 A1s and 1 C6 (ty). Hey, ks!!! You have one of the best results! Want to know my results? All the grades are so different that I\'m too lazy to total them up. Here are the results (boo hoo):  English - B3 Higher Chinese - A2 Maths - A2 Science (Combined) - B4 (!!! Neary last in class) History - A1 (yeh! but i think it\'s by luck) Geography - A1 (yeh again! but again I think it\'s by luck) Litearature - B3 Computer Studies - A1 (CURRENTLY) Music - Don\'t know yet but I stronly suspect a D.   As you can see, there are 3 A1s, 2 A2s, 2 B3s and even 1 B4! So ks, I\'m sure you feel even more satisfied with your results now eh? You even reached 80 for some subjects! Gone were the days in Primary School where we can score 90+ all the time. :\'(  Hehe, so ks, you all have Home Economics eh? Make sure next time you cook something for me, I\'m really enthusiastic to taste your \"Pizza Bread\"! (I\'m not being sacarstic.)  Holidays is really coming soon, even ks is going to install Sims Online. However, one thing to worry before the holidays, the progress report! What is my Mum going say? She already scolded me for my poor results when the test papers came back! Will she agree to sign the report? I also got some zeros in some of my assignments (quite a common sight in my class, especially for Oral Participation), especially Maths (for me)!  Here is the end of my update once again. Everything that has a beginning, has an end. But this will not be the end, there will be some more updates in future. (too lazy to put copyright, but you still can\'t copy, unless permission is obtained, for ks, you are at liberty)   '),(19560,2787343,'01,March,2004','2004-03-01 00:00:00','    Ha, we have half-day at school today because of the good O\' Level results! I went back to primary school again to see Mrs Lee, and the first thing we heard her is scolding \"Sickening!\" to the 6C this year. Mrs Lee told us (me and someone anonymous) that the 6C this year is more tamed (not as wild as us last year, at least), but very dumb (dumb means cannot talk, not stupid)! According to her students, \"dumb\" has became one of her favourite words, besides \"sickening\". (haha!) I really hoped to hear the 6C students this year complaining about Mrs Lee\'s bragging of 6C last year!  As school was only half-day, so naturally there wasn\'t any homework. However, I did get back my Science test paper back today. i ended up with a pathetic B4. I was really distracted and tired on that test day, so this was expected long long ago already. Not surprisingly, I forgot to put the units for my mathematical workings and put the wrong instrument for the \"match this measurement with a instrument\" questions. Instead of putting a burette, I put a beaker. I actually wanted to put a burette, but without knowing it now then I realized that I have put a beaker! Another question was instead of thistle funnel, I put a pipette! Something must be wrong with my eyes that day, that picture was so clear that it was definitely a thistle funnel!  Sec 1 Camp will be next week, I really dread this day that is coming, SOON... I don\'t know why but I just hate camping. The \"things-to-bring\" list is out. We can only bring at most $10. (wah!)   This is the end of my update, once again... Everything that has a beginning, has an end. But this is not the end, this is only the beginning! (Copyrighted by [ME]!!!)   '),(19561,2787343,'29,April,2004','2004-04-29 00:00:00','    Happiness Depends On Fate After All  After the sad momments that Jennifer Hudson was voted out, I can\'t really express my delight that John Stevens is \"kena\" voted out too!! Look at the time of this post, which is still much too early from the announcement of the results. But believe me, I do have first-hand information! Bet Simon was cooed and pissed when he hears this! Dunno why Simon likes him so much, to me he isn\'t a very good singer. He keeps singing all those dreamy classics, boring... I think Stevens is more suited to sing some Christmas carols, maybe \"Oh come, all ye faithful\"?   '),(19562,2787343,'28,April,2004','2004-04-28 00:00:00','    Surprise, Surprise  Looking at the list of the Project\'s Day PreliminarJudging results, i was filled with anticipation and anxiety... Will our team make it? It all depends a lot on this list. This list, is our lease of life... We are now one step closer to the $8000 prize money! :) You get what I mean? Yes, OUR TEAM MADE IT TO THE SEMI-FINALS!!!!!!!!!! Boy, I can\'t express my happiness when we saw that we have advanced to the semi-finals! Oh God, it was just like so unreal... We faced many obstacles upon completing the presentation, the lousy ppt we came up with, the presentation techniques... Mrs Nancy Tay, who was the 2nd judge, gave quite negative comments for our teams. I was really surprised that we were able to make it! Oh Praise The Lord! I will now attend even more Sunday services! Although I now have a sense of achievement, it\'s not really an achievement... Most of the teams did make it. But still, we must really thank God!    History Test Tml  Die liao, history test tml. This time, we were told that there won\'t be MCQs anymore, but some source-based questions and essay writing. This is the subject i have been fearing for... \"It is very common to fail in my class,\" said our History teacher, Miss Teo. Okay, i agree that her teachings are informative and you won\'t survive her class without taking down notes. She often gave out valuable information all the time without knowing it herself, so note-taking is important. And, she always go outside the textbook, giving us notes which are more informative. But more informative, means more things to memorise... Must study now...   ks is absolutely right...  i certainly agree with ks over the junk mails thingy... my junk mail has reached its new record, 100% all used up, in just 3 hours. Clearing the junk mail folder has become part of my daily routine. Between revisions and homework, before dinner, after dinner, during TV time and even before i sleep. Surprisingly, when it was nearly 12am at night, i opened and cleared the junk mail folder. After that, i refreshed, 10 came in! It must be a record, 10 came in in just about 10 seconds, and so late at night! And interesting discover, by looking at the email addresses, i found that these kinds of mails all from Singapore or Malaysia etc. this means that the person who is behind all these must be someone around this region too! I also starting to wonder whether Hotmail even realise the problem or not. Singapore and this region only made up of a small percentage of its suscribers! So, whether Hotmail will investigate till the end? i really dunno... Hopefully the police is investigating this too!   New Quote  I really love this quote, \"Life is meaningful only due to the fear of death.\" Indeed to me, life is actually quite meaningless. It\'s the same routine for everyone all the time, work, home, work, home. However,  God gave us the chance to live in this world becasue He believes in us. We are all in this world for some meaning beneath. Whether we will find the true meaning of life, it\'s up to us.  God put us through pain to let us realise meaning of life.  Really, life is where you choose your own path. We control our own lives, while God gives us chances. (This is sorta like a sermon, but i find this meaningful...)  Alright, to History textbook now. Life is meaningful only due to the fear of death. And remember, everything that has a beginning, has an end. It\'s how u control the end...   '),(19563,2787343,'28,April,2004','2004-04-28 00:00:00','    She Bangs!  Talk to me... Okay, that\'s William Hung, not Ricky Martin. Channel 5 just broadcasted the William Hung MTV thingy... My comments? It completely sux! It\'s sorta William\'s making a show, then the director keep yelling \"Cut!\" And the singing and dancing was as awful as before. I think William\'s really related to animals (not music), his face look like LuoHan fish, and he dances like an elephant. While some others say he\'s really \"All-American\", my mum says his singing is like talking. Come on will, just because of 1 lousy audition, resulted in MTV?   All Was Not That Well  Today\'s the preliminary judging of the Project\'s thingy, and i must say my team dint presented quite well in the long run. The judges\' comments were \"not really related to topic\" and \"objectives are not right...\"! Boy, i really feel like banging myself on the head after hearing that. I can\'t afford to fail... I really hope we can at least reach the semis. Oh God! May God bless us!    Again Another New Fashion Sense  Here it is, blog version 1.3 of my blog. ks says he prefer the previous one, i also dunnoe y... But in my personal opinion, i tink this is better lah. Because the previous one is well, too dull-looking, and it really gives the person a exhausted look. it\'s just like book cover, most people DO judge a book by it\'s cover. So, i dunno... But ks is qutie an expert in web designing, i should trust his eye-sight, shouldn\'t I?  These few days are sorta boring, so nothing much to write about... Well, so long!   '),(19564,2787343,'27,April,2004','2004-04-27 00:00:00','    Junk Mail Craze  The junk mail situation is getting completely out of hand... Just today, i found my school principal sending me a message \"Hi!\". I did opened the email, thinking that he mass emailed everyone about something. Luckily when i saw it was the same attachment again, i quickly deleted it. Wonder who started all these stupid automatic junk mails, hope he\'s convicted of capital offense... Really, the person who started all these damn \"bo liao\". Like nothing to do like that. Now, i find myself in the prospect of clearing my junk mail folder every 1-2 hrs, wasting time and money (time is money). Hope Hotmail get this fixed soon. I found my Yahoo account to be junk-mail free....   Project\'s Day  Feeling nervous now, it\'s preliminary rounds for Project\'s Day tml... Hope we can get into semi-finals. To be frank, we completed our presentation in just 5 days, with around 12 slides. I\'m presenting around 3 slides tml, in front of many other teams and 2 judges. Just hope that the judge is not Mr Jeffery Lim or Mrs Janey Sim. One judge is for sure, the head judge Mr Charles Low. I don\'t know his usual stadards, but he should be qutie \"relaxed\" lah, from my deduction :).   Flash rox!  In computer studies today, we learnt ActionScript in Flash and how to do a movie clip. It\'s quite complicated, and took ages to understand. Nevertheless, it was still fun. We learnt how to do buttons etc etc., call my crazy if u want, but i really luv the flashy lessons! Do amazing things with Macromedia Flash MX 2004!  I\'m suffering from writer\'s block (not sure how u spell it), so nothing to write liao lah. okay, bye   '),(19565,2787343,'26,April,2004','2004-04-26 00:00:00','    Bom Bee Bee  I\'m so touched... ***sobsobsob*** Yifeng finally won in todae\'s bom bee bee... She has been losing since the 2nd episode, and todae she finally won... maybe because ann poh\'s in her team, i dunno... Anyway, what matters is she finally won!!! Whoo!!! Wang Lu Jiang deserved all those punishments todae! the red team (wang lu jiang) was given make-up, made to eat bananas with wasabi and later made to turn a few rounds! Ha! I hope the green team will win in the many episodes yet to come! Well, the red team... Who cares about them anyway, they won so many times already. Anyway, i USED to support wang lu jiang and the RED TEAM. but seeing yifeng to \"ke lian\", lose all the time, start to pity her loh...   aiyah, can\'t think of anything to talk about already... This post came as a sudden inspiration, after yifeng won, bye...   '),(19566,2787343,'25,April,2004','2004-04-25 00:00:00','    Idiotic Junk Mails!!!  I don\'t know whether all of you are experiencing it or not, but lately, there\'s a lot of junk mails. For me, if i don\'t clear my junk mail folder every 2 hrs, it would have been flooded with more than 100 this kind of mails. I don\'t know which idiot started all these, but all the subjects are always, \"hi\", \"let\'s be friends\" and \"information\". Sometimes, there will be \"message failure\". I wonder if Hotmail knows about this, and i hope whoever started all these shall be put to jail! (yes, i\'m feeling that violent now) Ironically, i found that i seem to be sending these kinds of mail too, but be assured that i didn\'t send them MANUALLY!!! I don\'t know why also, maybe my account contracted a virus that starts seding these stupid messages automatically. i even received one from my friend lablled \"important!\" and he sent me an attachment too. so i tot it\'s really important and in the end contracted i-worm/netsky.z. later when i asked him, he swore and crossed his heart that he dint do it... oh well...   Parasites...  The parasite problem is getting out of hand. everyday, i will unknowingly install a parasite program. so i have no choice but to install SpyBot. Then in the end i found out that \"hey\"!it\'s not a bad program after all, it removes many of my parasites already! those who are interested to download, u can do it from  urlLink download.com . Many common parasites are, Comet Cursors, My Web Search, N-Case (the most irritating that i find so far) and Gator.   SARS!!!  SARS broke out again in China. Singapore\'s public is reacting to it fast, the recent newspapers is all about the SARS outbreak! Personally, i\'m quite scared of another outbreak, because we can\'t go out and no travelling and all those stuff. Come on SARS, why must break out again?! It\'s close to summer now, not winter! ANTI-SARS!!!!!!!!!!!!!!!!!!!!!!! SARS is the virus, that i just want to minus... Anyway, why has SARS broke out again? i myself cannot find any apparent reason at all. it\'s just like a new virus formed, unless SARS go thru\' biological changes once more?   Gas Prices up another 3 cents?  While the Iraqis are burning their oil wells, US gas prices leap at another 3 United States cents, making it USD 2.20 9/10 per gallon. The situation got so worse that gas stations have to make the drivers pay before they get their gas, so that the drivers cannot run off without paying. Don\'t think this will not affect Singapore! Last time when it went up to USD 2.00 9/10 per gallon, Singapore\'s petrol went up by 3 Singapore cents also. This time 20 US cents more, well... Not long after, Singapore\'s one also up by another 3 Singapore cents... Oh well, Iraqia and Arabians, why waste that oil???  All right, can\'t think of anything to mention liao. so must well end it. everything that has a beginning has an end... Sayonara!   '),(19567,2787343,'25,April,2004','2004-04-25 00:00:00','    ks...  If u have noticed again, ks have come out with his 3rd design of his blog! Oh boy, i must really hand it up to him! Only with good time management like him, then u will be able to achieve, Everyone hats up for ks! His new design is even much better than mine, i must admit, with the theme from The Sims 2! everyone, tekan at the link below and take a look at ks\'s new blog!   Abortion Rights?  Thousands of demonstrators marched up Washington D.C, somewhat for abortion rights... It was somehow led by Senator Hiliary Clinton. Some call it the \"worst constitutional rights ever\", while others back against the Republicans for illegalising abortion.  \"Stop the Violence!\"  cried some human rights associations. This debate is much more complicated than just any Republican or Democrat Campaign, but a  \"serious decision influential towards the whole globe\" . My opinion is that, whatsoever woman rights, duh... I think abortions should be stopped as much as possible, it\'s pure cruelty and defies the right to live. It\'s murder of the life of an unborn foetus! Come on, america, think before u do anything...   My Vision!  Before i end this post like any others, i must first tell everyone that this year: 1) i must \"pia\" 2) i must get mean grade of 1.8 3) i must \"jiayou\" 4) i must attend more Sunday services 5) i must put my faith in God 6) i must achieve enough to fly back to New York (mean grade 1.5, my parents\' target...)  Yah, i noe it\'s a bit too late to do some new year resolutions, let\'s make it Mid-Year Resolutions then! gtg, bye   '),(19568,2787343,'24,April,2004','2004-04-24 00:00:00','    Surpirsed?  My new outlook of blogging has came out 2 months before expected completion date! I\'m surprised myself, that I completed all these in just 2 days! (but look at the time now...) Okay, i noe this is not a very beautiful blog and it may look even worse that last time. But at least, it\'s full of originality and all created by myself! Don\'t worry, more is yet to come... Maybe during the June Holidays, I will design even a better one! Comments? You can put it into the tagboard i will create later!   '),(19569,2787343,'24,April,2004','2004-04-24 00:00:00','    Jealousy?  If you have noticed, ks have come out with a 2nd design of his blog! I wonder how he does that, with that horrendous amount of homework thrown at him plus projects... Alright let\'s not beat about the bush, i admit that sometimes i can be jealous of him (positive side of jealousy, not negative side) and will treat him as my \"mo2 fan4\", or model. Everyone who cannot manage time well enough, maybe you can consult him. Boy, i\'m really amazed by how he manages his time, while everyone in Singapore often complaint with the modal answer, \"no time!\".   Father who was jailed  Yesterday\'s news, Father you-noe-who was jailed for 71/2 years for embezzlement of mllion of dollars in church fund. If I\'m the judge (from what i understand, the judge is female), I would have sentence him to 20 years. Singapore laws doesn\'t really reflect much on embezzlement of religious groups. If it\'s embezzlement of money from a bank or something like that, it would have been much more than just 71/2 years. Still remember the guy from the beer company sentenced to life imprisonment? Stealing from a church should have been more severe than banks! Father Kang has betrayed the trust of the church-goers!    Moussaoui? Death Penalty?  US Supreme Court ruled that Moussaoui is most probably subjected to the death penalty. Moussaoui is suspected to be involved in the 911 attacks. Surprisingly, defense lawyers rebutted the court\'s decision as no fundamental rights and that he is not connected to Al-Qaeda in any way. Anyway, it\'s good enough that the death penalty talk in back in the table. However, certain district judges still rebutted of Moussaoui\'s constitutional right without a fair trial including video questioning with other county judges, besides the panal of 3 judges from the Supreme Court. The Grand Jury was also claimed \"inadaptable\" for a serious case such as the Al-Qaeda. Meanwhile, Moussaoui confessed of being in league with Osama Bin Laden but not involved in 911. Still, i think the whatsoever constitutional rights should change in leauge that the court case can have a bigger scope of focus and also charge Moussaoui with being in league with Osama even if not involved in 911. If Moussaoui is released and reunited with Al-Qaeda, wouldn\'t that be a great disaster? My final \"decisions\"? The Supreme Court is right for the death penalty and i\'m fully behind it. Who knows how many other lifes will they take the next time? Electrocution is not enough for such people, their cruelty has caused hundereds and thousands of lifes! They should be sentenced straight to poison gas, or even maybe drowning! Moussaoui\'s case is scheduled to be continued on April 28 (day of preliminary of project\'s day).   '),(19570,2787343,'23,April,2004','2004-04-23 00:00:00','    MSN Messenger 6.2!!!  I have downloaded the latest MSN Messenger 6.2. Wanna noe how is it like? It\'s quite a disappointment for me actually! There\'s not much of a difference prior to 6.1. Yes, it offers better security that solves netowrking problems automatically. Yes, if offers more games like Wheel of Fortune (only 14-day trial period, after that have to buy). Now, before u can type finish the coding for emoticons (for example: :)), they turn it into an emoticon itself already, in the past, u will still see the code happily showing itself there. And instead of calling it launch site programs, they now call it fun and games. More practical eh? ;) Thinking of download now? Go to this website then:  urlLink http://messenger.msn-int.com/Download/ . However for MSG Plus users out there, beware... You will need to download a new MSG Plus Version to ensure compatibility with 6.2. Pls also be reminded that even with this new version, it will still have some problems when using 6.2! MSG Plus 3, total compability with 6.2, shall only be release somewhere in the last period of spring (late May). MSN Messenger Version 7.0 shall only be released in winter this year.   Gmail  Most probably to rival with Hotmail, Google introduced a new email server Gmail. I\'m have  \"kiasuly\" signed myself up. This new email system offers 1 GB of space to each member (yes you read it right, 1 GB!!!) Such a comprehensive service... To round it up, there are no ADs (only the Google types, if u get what i mean) here and the server is so much faster! (my hotmail takes 30s to load finish, and my connection speed is 1500kb). With that 1 GB, you will never have to worry about mails flooding ur good ole Hotmail account anymore. However, it\'s meant for Blogger users only, so... Anyway, out of that gigantic size of 1 GB, i have only 1 mail. anyone feeling \"bo liao\" can email me at jensentan1991@gmail.com...   Chinese Lately  Today\'s the first Chinese Test we sat for that was set by our new Chinese teacher-in-training (last year in primary school prefects-in-training are called PITs, or rather prefect-in-toilet........), Mr Sunny Chong. Just like his name, he can be quite cheerful and qute fun. Speaking of Chinese test, well... I\'m mentally prepared to get at least a F9, so don\'t worry about me...  Just out of pure and innocent curiosity, we asked him yesterday what EP3 was he in. He just shook his head and told us to find out ourselves... This is the first time our class felt so \"cold\" in his class after so many weeks...  Aiya, \"wu2 hua4 ke3 shuo1\" already lah. Now u must agree with me, everything that has a beginning will have an end. beginning of a new idea also come to a stop...   '),(19571,2787343,'22,April,2004','2004-04-22 00:00:00','    New Fashion Sense...  Okay, i have now decided to have a new look (\"websites should be very neat\" computer teacher) of my blog. I already entered the 1st code to make the website. Completion shall be estimated to by the end of June holidays (really no time, so pissed with the Project\'s Day). Hopefully, there won\'t be collapse of DreamWeaver during the digging... (do u get what i trying to say?)   Nicoll Highway  Alright i admit, i use this road as often as the President. i was totally shocked and devasted upon hearing the news from none other than ks at around 5pm that day. it was almost like seeing 911 again with all the rubble. and now the ***dunnoe stupid or clever*** decide to cut up Merdeka Bridge. Have they ever well estimated the cost of joining it again? Come on people, don\'t get digging underground all the time. when there\'s no more support down there (like a tension formation of rift valley, i just learnt todae :)), the whole Singapore\'s gonna collpase, not only Nicoll Highway! Come on, Gahmen (Singlish for Government), build some of the things above ground lah! It\'s so much effective in cost. Lately, we must say that the Gahmen is having a lot of projects underground, underpasses, circle line, northeast line and whatever line and so on. Maybe the Gahmen should consider having underground houeses too...    Blackout?!  It was late at night, at around 11.45am. I got up slowly from my chair, and went to the kitchen to get a drink. And suddenly, \" pop \", i sorta turned blind. i was scared out of my soul, and went to get a torch. it was the most devasting, groping around in darkness. dont\' expect street lamps to help u get around, it\'s aboslutely nothing. there were also no corridor ligh, no handphone light, only torch ligh, \"car-light\" from the street, and computer light. after that i went to slp, and by next morning... Hey presto! the light is on again... **lame**   Joyful, Joyful (but it\'s not Christmas)  Lately, I\'m having a lot of freedom, i\'m sorta so joyful that as if Christmas is only weeks away. Both my parents are overseas (with occasional irritating international calls), one enjoying himself somewhere extremely far away from the red dot near the Equator, while the other one is still in the Asian boundary. It\'s computer games, more American Idol, and more MSN Chat! Too bad, there\'s someone else in our house who keeps nagging away, my maid...   Cut here and cut there?  In correspondence to a short movie clip i watched, i now know that there are 6 cuts in The Talented Mr Ripley, 2 cuts in Titanic, 2 cuts in Chicago, 4 cuts in dunnoe-what-else. I also reflected that horror movies are often cut in many gorry and bloody scence (no wonder lately horror movies not scary 1), what\'s the point of watching a horror movie then? And the highest record ever cut by the Board of Censors? 27 cuts, made in the Singaporean film \"15\". Lately, a rumour spread that 5 cuts were made in Kill Bill Vol.2, and it\'s still rated M18!!! Come on lah, Singapore Gahmen, other countries also not like this what. All films (except NC17) in America are available to people from all ages (for toddlers even when the movie is rated R (Restricted, Children under 17 must be accompanied by adult)). Can\'t the Gahmen be more open  (I really want to watch Kill Bill!)? Now then did i know that the expert chopping on the film strip by the censors helped the Pirated DVD Association in JB for a 65% rise in net profit! Gahmen, do you still want Golden Village and Eng Wah to continue losing money?    Elections coming...  Okay, the Republicans and the democrats are at absolute loggerheads with each other. Personally, i prefer Republicans (dunnoe y also) which means i sorta support current US President George Dubdew Bushit. Well, voting is in November and Senator Kerry + Bushit (this is not a chemical equation) are now producing ADs in the American TV. Bushit sucks a lot, still dare to mention 911 and Iraq War. Speaking of Bushit, he and his Dick Chenney give tax refunds but  raise gas prices (gas prices now cost USD 2.10 9/10 per gallon, EXTORTION!) it\'s twice as much as the prices in Singapore... So i tink Kerry will still win lah...   Hoot-U?  This seem to be favourite quote recently. I realised that i keep scolding people hoot-u. and also, i like to play about with the word hoot-u. Reflections: What do this word really means? y is singlish so \"entertaining\"? Hoot-u, i don\'t want to think also... suffering from stresssssssss, brain power is extremely low...  Aiya, can\'t think of anything else to say already lah, bye bye. Everthing that has a begnning, shall ALWAYS have an end...   '),(19572,2787343,'31,May,2004','2004-05-31 00:00:00','    Bom Bee Bee  I\'m a great fan of this show and i\'m sad to know that it\'s the last 2 episodes already. So... Haiz... Although this show going to end liao, the green team still lost today. It\'s just so annoying lah... Bei Chek!   Singapore Idol judges... Nvr even heard of them be4  To tell you honestly, i clean forgotten their names liao. It\'s just like... nvr even heard of them be4... There are basically 4 judges rather than the usual 3. MediaCorp acting smart again and try to be more \"original\". My sister said, \"For goodness sake, if they want to be original, change the \"Singapore Idol\" icon to red colour lah, y blue colour again?\" Hey, that\'s sounds like a good idea! Red may look nice and it reflects part of the national flag. Red background and white text! Hey, doesn\'t look too bad too!   '),(19573,2787343,'29,May,2004','2004-05-29 00:00:00','    Hi there. Just returned from school!   Website Maintenance Done  For those of you who cried, screamed or criticised my blog like Simon Cow, here, you got it. I just maintained the website, add some links, change the outlook etc. Now, any other comments, either positive or negative, of course you may email me or chat me up in MSN. Thanks for your valuable feedback!   Test schedule out, sianz...  Haiya... It\'s still the beginning of June holidays and the school is bothering and trying to worry us. They have given out the test schedule out for Term 3. Now i can\'t even play peacefully with the worry for the test in my mind.. Haiz... There\'s still MRP and project work to worry about liao... Now another test schedule happily come along! For those curious ones or those un-alert CHS students out there, u can visit this website to check it out.  urlLink Website    Going ice-skating  Tml is my last day in Singapore before the trip and I\'m indeed happy that I\'m going ice-skating tomorrow. It\'s fun, enjoyable but painful with blisters as well. And of course, it\'s as good as making a hole in your purse but i think that\'s worth it. Come on, it\'s holidays!   '),(19574,2787343,'29,May,2004','2004-05-29 00:00:00','    As most of you may already know, I have this habit of leaving my computer on whenever i do my hw, revise or when bathing etc. Now my mum is getting all jittery due to my \"horrendous\" results (i feel she\'s over-reacting a little, come on, my mean grade is still better than the class, there are bound to be subjects where i\'m weaker in compared to the class average). Whenever she sees me over-looking the com, she would like scold me and say that i\'m chatting all the time using MSN!   '),(19575,2787343,'28,May,2004','2004-05-28 00:00:00','    Well, if you have read the last post of ks today (May 29), he said I have a lot of broken English in my quotes etc. And I like to use Mother Nature... Now, i shall step forward for my defence! First of all, everyone should know that I love Nature more than anything else in the world (except my family). For proof, check out the theme of this blog, it\'s based on Nature too... Secondly, those quotes are mostly my own quotes and i can assure you that they are not meant to be in Singlish or whatsoever. Most of the time my fingers are tap-dancing on the keyboard while writing every post, so a few mistakes are unavoidable. To err is human  what ?  ps. and btw ah, ks, it\'s  proper english , not  correct english ... =P   '),(19576,2787343,'28,May,2004','2004-05-28 00:00:00','    It\'s holidays, holidays of our lives, our lives...   Curfew launched...  \"Listen son, you are not going to carry on with your horrendous results!\" my mum screamed last night when she saw my progress report... sigh, now curfew is launched. i\'m not allowed to on the com nor even touch it unless i complete all my hw! What the h**k! Aiya, nvm lah, striving for a better tml. After this, i will prevail with much better results! Still, i\'m allowed 1 hour per day, not bad lah. still unlimited tv...   Wow, so many ex-classmates!  Recently, i have rejoined and rebuilt the old friendship with my ex-ex-primary school classmates. Suddenly, they have all appeared everywhere. I have met up with quite a few at Chinese High alone, then i continued to met with some from other schools. BTPS ex-classmates also met some of my ex-ex classmates. Just today, another ex-ex-classmates added me to his contact list! Wah, Singapore really that small leh!   Now enjoying the last few moments of liberty and freedom before the curfew... haiz... Humans have really gotten too smart even for Mother Nature... Learn from yesterday, live for today, hope for tml... - Jensen Tan   '),(19577,2787343,'28,May,2004','2004-05-28 00:00:00','    Wah! Today i got a lot of things to write leh!   Ice-Skating was thus erm...  I broke an all-time record! Yes u read it right! I broke an all-time record! I fell down an achievable 8 times today on the ice rink while practicing ice-skating, breaking the record of the orginal 7 made by another classmate! (oh btw, the whole ProEd sec 1 went ice-skating) But then, at least i\'m of the the \"braver\" ones who really tried to skate, while others held on to the railings while skating along. Alright, to be honest, i know how to roller-blade, which means performing so badly in ice-skating is \"an insult to my mentality\" (haha, lame...)even when you have the experience. Ice-skating was indeed fun, but the rental charges very expensive, haiz...   No this will not do!  Received my term progress report today. Took a little peek before completely opening the booklet, can break into tears i tell u! I got a mean grade of about 2.66 (or at least i think...), dropped from 2.13 last term! Arghh! 2 of my subjects, English and Integrated Humanities, dint even meet average mark, i was like so pissed that i can tear the report slip... Arghh, so frustrated with myself! But the other subjects still not bad lah, homework deadline and quality most of them also achieved an Excellent grade. Subject teachers also gave quite positive comments! I particularly like this comment from the Chinese teacher, \"å®ˆè§„å‹¤å­¦â€œ. It\'s amazing how he used just merely 4 words, while other teachers broke into hysterical moments and wrote many lines when inspiration struck them. oh btw, the chinese teacher gave the whole class the same comment, \"å®ˆè§„å‹¤å­¦â€œï¼   Thus far we had came  The path of life is always filled with uncertainty with obstacles you can expect at all times. There are always at least 2 paths to take, one which u will suffer first but reach luxury later, while the other will lead you to luxury and make u suffer even more later. Which path would you take? Is it the red path or the blue path? Well, i\'m pleased to announce here the my Chinese website designing team took the red path (or rather suffer first), and we are now enjoying bliss after the hardship. Indeed, we have spent sleepless nights trying to complete the website, endured the test put within by the Mother of life by the criticisms made by our mentor. But, still we can see, as time passes by, there\'s dream in our minds, and we are going to achieve it! and we did! Okay, to sum it all up, our team got into the finals of National Chinese Web-designing competition! Everybody wish me gd luck and get first okay?   Holidays, holidays, holidays  Our vacation has finally come. The time for me to prove myself has came! The real obstalces are in front of us and Mother of life is watching over us! This vacation is not meant for only vacation only. It\'s the time to prove our capabilities and our critical-thinking minds, to create a new vision. The vision of doing well in our MRP and project\'s day, to enhance our test results! I can see the vision just in front of me, and i\'m going nearer and nearer to achiving it! The June holidays have provided me a short-cut to reach this vision faster, and i\'m going to treasure it! I have started a new vision myself, and it starts right NOW! i\'m going to achieve it! My mind is now focussed on academic and other school stuff, less focussed on other activities such as playing. Everybody be my witnesses! Strive I must and achieve I will! Strive for the best! =)   Oh Mickey!  We used this song for line-dancing last year. And by pure accident, i found the lyrics on the web:  (Oh Mickey, you\'re so fine you\'re so fine you blow my mind, hey Mickey, hey Mickey) (Oh Mickey, you\'re so fine you\'re so fine you blow my mind, hey Mickey, hey Mickey) (Oh Mickey, you\'re so fine you\'re so fine you blow my mind, hey Mickey)  Hey Mickey!You\'ve been around all night and now that\'s a little long You think you\'ve got the right but I think you\'ve got it wrong Why can\'t you say goodnight so you can take me home, Mickey  \'Cause when you say you will, it always means you won\'t You\'re givin\' me the chills, baby, please baby don\'t Every night you still leave me all alone, Mickey  Oh Mickey, what a pity you don\'t understand You take me by the heart when you take me by the hand Oh Mickey, you\'re so pretty, can\'t you understand It\'s guys like you Mickey Oh, what you do Mickey, do Mickey Don\'t break my heart, Mickey Hey Mickey!  Now when you take me by the, who\'s ever gonna know Everytime you move I let a little more show There\'s somethin\' we can use, so don\'t say no, Mickey  So come on and give it to me any way you can Any way you wanna do it, I\'ll treat you like a man Oh please, baby, please, don\'t leave me in a jam, Mickey  Oh Mickey, what a pity you don\'t understand You take me by the heart when you take me by the hand Oh Mickey, you\'re so pretty, can\'t you understand It\'s guys like you Mickey Oh, what you do Mickey, do Mickey Don\'t break my heart, Mickey (Oh Mickey, you\'re so fine you\'re so fine you blow my mind, hey Mickey, hey Mickey) (Oh Mickey, you\'re so fine you\'re so fine you blow my mind, hey Mickey, hey Mickey) (Oh Mickey, you\'re so fine you\'re so fine you blow my mind, hey Mickey) Oh Mickey, what a pity you don\'t understand  You take me by the heart when you take me by the hand Oh Mickey, you\'re so pretty, can\'t you understand It\'s guys like you Mickey Oh, what you do Mickey, do Mickey Don\'t break my heart, Mickey Oh Mickey, what a pity you don\'t understand You take me by the heart when you take me by the hand Oh Mickey, you\'re so pretty, can\'t you understand It\'s guys like you Mickey Oh, what you do Mickey, do Mickey Don\'t break my heart, Mickey Oh Mickey, what a pity you don\'t understand You take me by the heart when you take me by the hand Oh Mickey, you\'re so pretty, can\'t you understand It\'s guys like you Mickey Oh, what you do Mickey, do Mickey Don\'t break my heart, Mickey Oh Mickey, what a pity you don\'t understand You take me by the heart when you take me by the hand Oh Mickey, you\'re so pretty, can\'t you understand It\'s guys like you Mickey Oh, what you do Mickey, do Mickey Don\'t break my heart, Mickey {fade}  Anyone wants the song can email me! i got the song with me! Hey Mickey!   I have now chosen a path of obstacles and the price hidden among the midst of obstacles... Everything depend on one self\'s being.   '),(19578,2787343,'27,May,2004','2004-05-27 00:00:00','    Oh say, can you see? (the American Idol)  Fantasia won the American Idol! Basically, there wasn\'t much singing tonight, haiz... but i do like watching the 1-hour special on 10-11pm. The americans do have manners, sang national anthem and saluted before the competition started. Then the contestants who were voted out came back to sing a few songs. simon was in black again, or rather he looks \"cool\" in executive wear.;   '),(19579,2787343,'25,May,2004','2004-05-25 00:00:00','    For simply no reason, i have this sudden urge to update my blog (yeah!!) although there\'s nothing much to write.   Harry Potter... So much enthusiasm...  Don\'t get mistaken, i\'m not a potter fan. somehow, i got this sickening feeling when i saw today\'s zaobao, a pic of a girl holding up a sign, \"Marry Me, Daniel!\". Nope, i can assure u it\'s not jealousy (i rather pray that i have excellent results), but others in class have the same feeling too, it\'s downright disgusted... Although not a fan, but i\'m somehow overwhelmed by the enthusiasm (or to put it crudely, kiasuism) of singaporeans, i definitely won\'t feel comfortable when others have watched the movie and i haven\'t. still, if lord of the rings or pyscho were to come out at the same time, i would rather watch them first. oh btw, somehow i\'m love the old movie pyscho a lot, maybe because of the scene of the bates motel.    Boring lessons  Since after the tests, we seem to get even more boring lessons all the time. Geogrpahy was some lesson interesting and enlightening, but now we are just watching some videos of earthquakes and study totally crap stuff like batholith (some big piece of formation of solidified magma...), laccolith (something like a batholith, but smaller in size and mushroom shaped) and lapolith (saucer-shaped, usually found in between admist the layers of rocks). Okay, it\'s not that crap after all and still a little of interest hidden in there, but it was just 40 minutes with everybody \"huhing\" and it\'s downright irritating. (ps. SICKENING! lol, ex-6c members should know what this means!)Also, we also learn terms like sill, dike and pipes. Forgot what they are liao, so many things to memorise... Haiz, i\'m getting on with my age...   American Idol down to finak 2  American finally have the sense of hearing and the \"class\". They finally voted jasmine trias out, she should have gotten out longer in my opinion. yes, she can still sing quite well, but still lose to jennifer hudson (cries...) and la toya london. it\'s suppposed to be a tough fight now, with the best singers of america, looking forward to it! :D btw, hawaii has declared may 13 (can\'t remember the date well) as jasmine trias day... lame rite? she\'s just a contestant of American Idol and she dint even win the title!   Sleeping is essential as part of the life process... ZZZ...   '),(19580,2787343,'24,May,2004','2004-05-24 00:00:00','    Don\'t talk so much lah  After Quan Yifeng won last week, she began her habit of talking a lot in this week\'s episode of bom bee bee. She could have won again, but she spent so much time talking to the guy who was going to gelyang bahru, the final destination. Arghhh!!! It\'s just so close!! Next week is the last time i\'m going to watch bom bee bee be4 going for my trip, she must win!!! (oh btw, the idiotic airliner somehow mixed up the departure dates wrongly, got a shock when we went to receive our tickets...)   Going to buy electronic games soon ^-^!  Soon, i will have another whole box of video games! electronic games are cheap in new york (especially walmart, around USD 16 for a REAL version! only abt S$28-30, singapore cost $40 at least!!! xbox...) Anyway, i dun tink i will buy the real version. pirated version in the US are as good as the real version, 110% much better than the Johor pirated CDs. But still, since real version so cheap, why not buy some for real exp. instead?   PS3  Been eyeing the PS 3 in the US for quite some time, since my last visit to Ontario Mills, CA. last year. It\'s quite \"x\", about USD 168 (but compared to singapore, still cheaper leh...). Quite determined to buy it, but the plug is different and it\'s very difficult to use voltage adaptors all the time. Also, it\'s not easy to bring the whole machine back to signapore by luggage... So, aiyah dint buy lah! sorta regret it later... maybe i will be a bag specialised for this next time, hope it won\'t get checked in into the aircraft due to restrictions! or i must well wait for it to come out in singapore...  It\'s getting late, got to do what everybody got to do at nite! (åˆ«æƒ³æ­ªï¼ï¼‰   '),(19581,2787343,'22,May,2004','2004-05-22 00:00:00','    A quick update for the first time in 5 days...   ProEd Nite... It\'s lame...  In case u r wondering, ProEd is my consortium (something like the House system used in other schools) and the council organized a ProEd Nite last night, which was Friday. It basically sux, only everyone singing some lame songs on stage. The overall theme? Hawaiian... But u dun feel Hawaiian at all, even though quite a few no. of ppl (including teachers) wore Hawaiian shirts. The \"dinner\" or rather supper we got was a bar of Milo, something like a melting chocolate bar. It wasn\'t even fun at all, unlike the Fun Day last year in good ole BTPS. Basically, treasure hunt was in the program sheet and i\'m looking forward to it. But somehow, it\'s cancelled... So, it\'s just a nite of American Idol with William Hung... And there was this Mr Macho competition which was even lamer. Some guys are qutie macho i agree (i\'m NOT gay), while some classes stupidly (including my class) sent up some fatso (i\'m not trying to tease those ppl, just being frustrated at how some classes can sabo their classmates very well...) like Ruben Studdard. It was so damn retarded, with just some fatso who was there unwillingly posing some stupid pose. Some fat but quite muscular ones were indeed sporting and still managed to suppress a smile... What\'s really the points of this competition? when certain competitors were half-naked and strated showing their butt-clad underwear, in front of parents, teachers and all... Of course that guy dint win the title and maybe he would have won it if he wore a g-string... =P   Results revealed...  ks said he did badly when he scored 4 As... but what abt me???  English - B4 Maths - A2 Chinese - A1 Science - B3 History - C5 Geography - A2 Literature - B3 Computer - A1  ks, be contented with ur results... dun be a loser like me... =X   It\'s now another end.   '),(19582,2787343,'17,May,2004','2004-05-17 00:00:00','    I lost my bet  If u look at an even earlier post, u realised that i have given a topic for betting. I would have lost my bet if someone really put down some odds! Haha, luckily no one wanted to bet or i would have lost at least a precious $10 (i\'m very poor). So everyone, the bet\'s now officially closed!   '),(19583,2787343,'16,May,2004','2004-05-16 00:00:00','    Bom Bee Bee  Wahhhhhhhhhhhh.......... i\'m so touched once again! The green team finally won again in todae\'s bom bee bee, after losing another 2 episodes. They could have reach the destination, Sembewang, even faster loh. There\'s a boy who was going to Sembewang but answered the questions wrongly... Sigh... But nvm, what matter most is the green wins again! but looking at the short film of next week, i tink they lose again lah... wah piam!   '),(19584,2787343,'16,May,2004','2004-05-16 00:00:00','    Really nothing much seem to happen nowadays, so talk abt some very wu liao things lah!   MRP (Major Research Paper)  The 3rd submission is starting. This submission is the draft of the final paper itself. ARGHHH!!! I think i\'m goin crazy, i dun even understand the format of the final submission... How do you write a research paper? Those who know and feeling wu liao can email me and tell me. Not only those wu liao, those not wu liao got free time also can email me! OH, i\'m doing the topic on Hitler, so got time do some mini research for me! ;) =P Haha, thanx!   Jue Shi Hao Bra?  Yes, u guess it. I watched (though i\'m not a pervert) jue shi hao bra yesterday. It was a damn pervertic show only perverts would want to own it in dvd... I\'m telling ya that it\'s 101% worse than Legally Blonde (at least it\'s still a movie from hollywood...) i dunnoe.....    Troy  Judging from the gibberish (there is really such word in the English dictionary) reviews, i have decided not to watch Troy. A summary of the reviews given, \"This is a big epic battle scene. But the fight is sorta so unreal... There is nothing much else to see other than fighting, unless you want to count seeing Brad Pitt\'s naked butt.\" (yikes...) Girls, if it\'s a dream come true for u, go for that butt... But dun say i dint warn u, many critics and viewers have deemed it a show of only fighting. Oh, btw, it\'s given the R rating by MPAA for violence, language and sexuality/nudity.   It\'s Monday...  Time passes indeed swiftly and it\'s Monday once again. Monday means another day of bom bee bee... Aiya, i tink i can bet that quan yifeng will lose again today lah, with dick su... Bets start at $10! Buy with me thru\' email!   Comments  A lot of people after reading the posts of this blog, email me to give comments, as a results flooding my email. In case u dint see it (sry i dun mean to be sacarstic), there\'s a comment hyperlink at the bottom of every post. For more personal comments, of course u r allowed to email me (by Gmail pls). But for others that are more bo liao one (means post for fun one lah) can post it thru\' the comment lah. so far ks is the only one commenting that way... Btw, a big THANK YOU for those who have visited and gave positive comments of this blog. PS. in case if u r wondering where the counter is, well, i admit that i\'m too lazy to put it in (and to add in the links, but dun worry, i will add the links soon, but maybe i shouldn\'t say soon, haha..). but judging by the number of emails (sry to say, but by junk mails as well =P), still another THANK YOu for visitng my blog!  Aiya, must do hw liao... sianz...   '),(19585,2787343,'16,May,2004','2004-05-16 00:00:00','    June Vacation  After many questions from everyone, i decided to make the trip public...  Departure: To New York JFK via Tokyo-Narita United Airlines, UA 800 May 30, 7.20am 21 hrs, 55 min  Arrival: From New York JFK via Tokyo-Narita United Airlines, UA 801 June 5, 23.50pm (dun need to go to airport to see me lah... =P) 23 hrs, 5 min  Some have asked me abt place of residence, well... i only noe it\'s abt 10 miles north-east from central park...  Well, it\'s indeed \"exciting\" that the June holidays are coming. But with the amount hw given, i can\'t say it\'s that \"exciting\" any more! Stay tuned for my next update!   '),(19586,2787343,'13,May,2004','2004-05-13 00:00:00','    About a test I took today...       Disorder      Rating        urlLink Paranoid :    Moderate       urlLink Schizoid :    Low       urlLink Schizotypal :    Low       urlLink Antisocial :    Moderate       urlLink Borderline :    Low       urlLink Histrionic :    Moderate       urlLink Narcissistic :    High       urlLink Avoidant :    Low       urlLink Dependent :    Low       urlLink Obsessive-Compulsive :    Moderate       --  urlLink Personality Disorder Test - Take It!  --       Are the results correct? Kindly email me on this! Thanks!   '),(19587,2787343,'13,May,2004','2004-05-13 00:00:00','    Haiz...  A short post today. My trip is now put forward to first week of June. Just yesterday it was around 2nd week, and now it\'s first week. It\'s not our fault that we keep changing the date, but the airline company (incidentally United Airlines) put forward us, as there \"were no enough tickets for all of us\". Luckily we will be staying over at our relatives\' house. Or we would have been surcharged for the change of our dates at the hotel...    Sabbaticals...  The Hwa Chong Sabbticals, well... I can\'t say that it has been fun... For the first few days, we were just lamely sitting down there and listen to the ppt prepared by the trainer (anonymous). LOL, but there were candies to lighten us up! Whenever we felt like sleeping and slacking, the trainer would very kindly sponser us with candies. So fellow \"Sabbaticalns\", feel sleepy more! So we can get more candies! Today is the highlight of the whole Sabbaticals, we were allowed to debate (aiya, more crude-sounding means quarrel lah)! But... my team still lost...   '),(19588,2787343,'10,May,2004','2004-05-10 00:00:00','    Haiz... Lose again ah?  Quan Yifeng and Dick Su lost again in today\'s Bom Bee Bee, as usual and as expected. It was damn pitiful for them loh. The punishment today, is for them to dress up as Indian ladies (yes including Dick Su!!!) and perform some sort of Indian Dance. They dance like hell, much worse than Bollywood concert! Aiya, i lose faith in them already, now i\'m the neutral side, although my sis now all the time shout at them to move faster, as they always talk rubbish with the passers-by...   June Vacation  Since my last post in the afternoon, a lot of ppl been asking me for my whereabouts (yes recently i publicise my blog, mostly to cousins, but they dunno where i going too). So i have decided to give some clues...  Hint #1: It\'s somewhere far far away. It will take about 21 hours flight (including transit) to reach there.  Hint #2: I can now watch American Idol live from Fox Channel. Yes, live! Which means i will noe the results sooner than all of u. But by the time, the new American Idol should be out already lah... :(  Hint #3: This city i\'m going has rude service and the airports are one of the worst, but be sure that it isn\'t China or Africa etc. Oh, a major tragedy also once happened there.  Okay, u should be able to guess now. If u still cannot, wait till i change my profile\'s location lah... So remember, also dun call my hse phone too, from 3rd week to 4th week, or rather the last 2 week of June holidays. Pass this message to as many ppl possible.  PS Anyway, just crossing my fingers that my parents won\'t cancel the trip last minute. Nowadays a lot of things happening to the country i\'m going... Just hope that lately nothing else will happen.   '),(19589,2787343,'07,May,2004','2004-05-07 00:00:00','    Pls take note  Hi everyone reading this! Pls take note that my overseas trip is confirmed liao. So i\'m leaving Singapore on 3rd week of June and is expected to return on middle of 4th week. So during this period, for those of u who knows or dunno my hp no., pls dun call. Even if it\'s something really important and u call, don\'t be surprised if u get extra high phone bills. No, it\'s not just a few cents per minute, but if i\'m did remmeber correctly, it\'s around $2 per minute (AT&T Wireless charge until like siao one). Also, the time difference is damn big so even if u call in the afternoon, you may be well calling me at midnight. Of course you are welcomed to email me, but pls email to my Gmail account, as i dun want my Hotmail to be flooded. For those of u out there who knows my whereabouts, pls keep it a secret as i really want a peaceful vacation.   PS I can\'t really write a long entry to to time constrictions nowadays... so long!   '),(19590,2787343,'07,May,2004','2004-05-07 00:00:00','    Dint update my blog, for let\'s see, 3 days? KS is fuming away, \"update ur blog !!!\" The schedule for these few days are murderous! Btw, ks lost all his files and settings because his computer something wrong lah. So he asked me to sympathize with him, so \"ke lian\" him loh.   Idiotic SP2  Thinking of downloading Windows XP SP2? Well, i would advise you not to. Because of downloading this, my whole com corrupted and also lose all my files and settings etc. etc. Luckily my father was able to recover them, but ks just isn\'t that fortunate...   Exams over!  Exams (or rather tests) are finally over! I have suffered weeks of clogging for the tests and the emotional attacks of getting low grades all the time! Really, my MSG must have dropped by dunnoe how much (means a lot lah) compared to last term. But i again will not take failure in my stride. \"You can cry now, but you must get up later...\" This is Mrs Lee\'s \"motto\" for us last year. So i will take her advice and strive even for better results! From Level 3 of studying, i will now switch to Level 5 (highest level)! Or maybe becasue i\'m sitting at the back of the classroom, so distracted most of the time.  Okay, that\'s all i have to say... Watching \"Good Gracious Me!\" now...   '),(19591,2787343,'04,May,2004','2004-05-04 00:00:00','    SQ Hijack?  LAX was reported to have received a signal from SQ 20 (the long-haul flight to Los Angeles). Federal Aviation Authority immediately responded by sending a F-16 (whoa!!) to escort the landing of the light at Tom Bradley Int. Terminal. SWAT (wah!!!**Special Weapons and Tatics police=D**), FBI, and some negotiation team went up, only to find that it was a false alarm. Wow, i sorta hope that i am in this flight...   Flunk My History Test  Today, received back my History test... As expected, i nearly flunk it, with just a passing mark of 10/20! But rest assured that i won\'t take failure in my stride! \"Use this as a wake up call!\" and \"You can cry now, but you must get up later!\" were what Mrs Lee always said to us last year. This is precisely what i am going to do next! I have devoted 3/4 of my time all to studying and revision right now! I\'m going to prevail as a succeedor the next time! \"And this be my motto, In God is Our Trust!\" This may be useful now, so may God bless me!    Holidays coming!!! (more of an account)  Holidays is only a few weeks away! Boy, i\'m so happy as if it\'s December holidays and Christmas is here! The first semester for my secondnary school life has indeed been rocky, very rocky. There are new challenges set before and ahead from now. There\'s Project\'s Day and MRP. I can\'t say that i\'ve been coping with my studies very well. From the \"after-effects\" of the more than 2 months \"holiday\' after PSLE, i found that i myself is currently more prone to using the computer, watching tv etc. etc. I believe if i can overcome these bad habits, i can achieve even better! From the guidiance of Mrs Lee and Mdm Wang last year, i can say that actually they had develop a sort of \"beginnings\" in us. The hetic lifestyle last year by Mrs Lee has prepared us well for this year! I still remember that she always made us talk, make sure that we always have the initiative etc. This is similar to that of my teachers this year, especially our History teacher Miss Teo. (\"Yes? No? Answer!!!\") I think i will go back to BTPS definitely on Teacher\'s Day!  Okay, that\'s all I have to say. Because there\'s nothing more for me to say, as it\'s the end after the beginning... Next update? Who knows...   '),(19592,2787343,'03,May,2004','2004-05-03 00:00:00','    Dick SU!!!  i\'m so sad today... ***sobsobsob*** Quan Yifeng lost again in bom bee bee... All because of DICK SU!!! Last week, wang lu jiang lost because of dick su!!! and this week, well... The punishment for this week is really damn \"cham\"! Each member is to find some coins in a box of either worms, cockroaches or frogs. Yeah, it\'s something much like Fear Factor... However, it\'s even worse!!! The \"victims\" are blindfolded, so they have no idea at all at where they are touching inside the box. Dick Su got the box of cockroaches, he has no idea at all that he\'s actually touching some cockroaches! Well, i don\'t think i want to watch bom bee bee anymore... yifeng keeps losing, very \"bei chek\"...   '),(19593,2787343,'01,May,2004','2004-05-01 00:00:00','    Gmail rox!  Wow! I kinda like Gmail a lot. Not only does it offers 1 GB of storage and no banners etc., it also groups all emails with the same subject together (they call it conversation...). In goold ole Hotmail, all mails come as separate, whether they are under the same subject or not. AND, it offers a search engine in your veey own email account to search for whatever mail you wish to find (this is going to be helpful while flashing back at the archives). Yes, with 1 GB of storage, you will never have to delete any other email again! Whoa, now that i\'m typing \"1 GB\" all the time, it seems so unbelivable. 1 GB... = 1000MB... - 2MB... = 998MB Yes! 998 MB more than your Hotmail! or 994 MB more than your Yahoo! Great job Google!   Michael Jackson the Star  Yes indeed, Michael Jackson appeared at Santa Maria courthouse wearing specs for the first time as long as i noe. When the judge read out the charges, he was seen as being a silent figure and still denied the charges. Michael Jackson really made the headlines, now being judged under the Grand Jury!! If found guilty to all charges, Michael can be jailed for 74 years (not sure whether true or not, but 74 years is something like given the 99-year life sentence), this means he\'s gotta spend the next part of his life in jail! From what i noe, this ain\'t the first time Michael was charged with child \"abuse\", but he was charged with the same charges years ago, my parents told me. This time Michael arrived 45min before the session, rather than the 1 hour late the last time. Also, disappointingly Michael dint jump up his car this time and \"performed\". I really hope he would do that, but at least now he proves that he has a sense of shame...   Murder at West Mall!!!  Located only about 5 miles away from my house, there\'s a murder at West Mall! The alleged suspect just stabbed the victim, 5 stabs at the chest and 1 stab on the neck! It\'s sure horrifying enough... My family that day was still planning to go to the library at West Mall, but in the end we dint go. The murderer was charged today in court. He could face the death penalty, although he\'s only 18.   Hostage from Macon, Missisipi  Amercian hostage from Macon escaped from his captors yesterday! It was really a sigh of relief! However, those \"bu shuang\" Iraqis all killed the US soldiers being held! It was reported that soldiers were abusing the Iraqi prisoners as well, ahlthough i\'m sure not all of them do this. Come on, not all soldiers are \"wicked\". Sigh, when will this war end? After looking at some articles today, i\'m now as equally positive that Bush may win this election as well... So maybe the war may continue... Sigh... Peace  Well, that\'s all for these 2 days. Lost time is never found again, the time when i end my post now is 6:38:51, Monday, May 3, 2004. It\'s sure that this specific time will not ever appear again... Time is just an illusion, time is just a man-made system, which reflects nothing on Nature at all! Day and night is then the real time given by Nature...   '),(19594,2787343,'01,May,2004','2004-05-01 00:00:00','    Beautiful Las Vegas  I have changed the picture from a highway to my hometown, Las Vegas. Las Vegas is beautiful indeed, isn\'t it? I\'m still trying to make the picture bigger, so that it\'s clearer. Please do note that the picture is copyrighted by Las Vegas Chamber of Conference. If i would to recommend a place to visit for vacation, it will definitely be Las Vegas. Not only you can gamble (children under 21 not allowed, haha, sry!), but it\'s realyl an Entertainment City. If you can\'t get enough of Genting (Genting follows the theme of Las Vegas), go on to Las Vegas. Brief introduction of Las Vegas: a city, twice the size of Singapore, with a population of 400,000, located in South-East of State of Nevada. A place of family entertainmenet, 24 hours of day and night life, but it\'s NOT Geylang. Welcome to Fabulous Las Vegas, Nevada!   History Test Yesterday  As predicted, History was a killer-paper. It\'s much like the English test set by Mrs Lee last year, I\'m not even sure now that whether i can pass. overall, die liao lah!  Nothing seem to happen much nowadays. So nothing much to write. So long!   '),(19595,2787343,'30,June,2004','2004-06-30 00:00:00','    Just wanna say that i added 2 new links under bloggers. welcome aboard! ks, i have found another Sims 2 fan. visit victor gan\'s blog.   Finally...  Finally, i finished reading Village By The Sea. It consists of about 260 pages. The first 100 pages may bore you to death. i took about 3 days to read the first 100 pages. but amazingly, i took 2 hrs to finished the remainding 160 pages. Amazing eh?   Bala Krishnan  Bala Krishnan (the indian audio clip below) has brought me trouble. my family\'s got so tired of it that i was chided. everytime i see it, i would be tempted to listen to it. it go so intense that it\'s sorta of an addiction. so, the audio clip here must go also. nvm, find some enw music loh. haiz... anyway, the name bala krishnan would remind me of kala, the ortus english teacher. she\'s a white-indian who looks so much like a chinese that u won\'t believe she\'s an indian at first.   '),(19596,2787343,'28,June,2004','2004-06-28 00:00:00','    Another thing, Jensie\'s school(CHS) teaches C++ which is so much better! C++ is quite hard to learn by reading a book, you need some real instructors to teach you and make it some much easier. His first program, unfortunately, is not the customary \'Hello World\'. Never mind, maybe he did that in school.   well, so far my C++ teacher has been quite nice and explained in quite a clear way. able to understand him at the very least. my first program, well... it\'s modified from a certain website. don\'t fret, ks, it\'s also because of the guidance from my dearest brother and my father, that i succeeded. so really UNFORTUNATELY, it\'s not fully done by myself. yes, i wrote and hand-conded the code but it was them 2 who rectified the mistakes.   '),(19597,2787343,'28,June,2004','2004-06-28 00:00:00','    How\'s first day of school?!?!?! Okay loh...   Quah\'s lesson  Quah (english and literature) was supposed to have 2 periods (80min) today. After we greeted her with that usual unenthusiastic voice, we were told to open the top windows. Just after that, our Chinese teacher thought it was his period and came in to interrupt for 5 minutes. quah first asked us abt our holidays, then talked abt her own experiences, mostly about SLC (Student\'s Leader Convention, held annually in chs). she talked abt certain foreign delegates (during their tour around sg) like wanted to stop at zig-zag traffic lines (means cannot stp at all), just to take a picture of the Singapore River and them wanting to shop at Watsons. Although she did specified the country, i would strongly suspect either India or Iraq. Not because i really look down on them, but common sense will tell u that their country would never have Watsons. And most of the time these ppl are poor and seldom have the chance. I dint choose china because chinese ppl some are even richer than us. most of them would have seen much  bigger  rivers than us... Oh well... After that, we proceeded to do some reflections and goal-setting. 2 periods passed juz like that...   Assembly  There was a part of assembly quite ah...  restricted . It\'s talking about those teenager\'s things, the u-noe-what lah. so if u want more details, catch me on msn.   History with teocm  Teo did not start off asking how are we and all these kinds of stuff. she went striaght on to the lesson, and doing revision on the topics of the last semester. today, she gave out one flying kiss (she does that everytime when difficult questions are answered perfectly) and said \"i luv u!\" to that guy. to round it all up, i juz have to say that she\'s perfectly emotional as usual. try not to spoil her good mood, or something will sure happen...   Chinese with tangc  tangc did not do any lesson or particularly  academic  at all. he first rant off about reading  xi you ji , then he spent the remaining half an hr talking about the website competition. as he made classmates to stand up and talk about their reflections and conclusion, this rather gave me an insight to their thoughts. juz todae, surrendered the trophy over to him for him to display on his table. but nvm, the tweeny weeny little trophy proved to be rather an eyesore and heart-breakening scene. Oh well...    just a note here to all readers, can you at least just post a tweeny weeny little comment so that i noe u r there? Thanks!   '),(19598,2787343,'27,June,2004','2004-06-27 00:00:00','    Was browsing thru\' blogs just now.   Firstly, i must comment that i have accidentally found the blog of my classmate. I nvr knew he got a blog!! going to confront him tml. =D Will add a link soon... too lazy to add now.  Secondly, ks used to (take note of the phrase \"used to\") promise that he will update his blog EVERYDAY. Well, from the situation i\'m looking at right now, it\'s weekly, not everyday. Oh well...  Thirdly, chee yew (chainerboi a.k.a. cy) erm... has been under construction for his new blog for a long long time. it always say that it will end at 4.30pm, but... Oh well... click on the link below (under bloggers) to see what i mean.   Fourthly, zhong han\'s (a.k.a. zonny) rarely update his blog. so muz applaud for him since he updated the saturday before last (June 21). his latest blog update is abt, oh well... check it out for yourself. yoyoyo... tml muz confront him also.   '),(19599,2787343,'26,June,2004','2004-06-26 00:00:00','    Well, i\'m feeling damn bored now (what do you expect i can do at 12.30AM in the morning?!), so update the blog to pass my time. Nope, it\'s not because i want to catch soccer match later, but it\'s my last day that i can slp late, so must treasure it wisely! The \"art\" of sleeping late... I\'m one of the owls that many ppl call, it\'s extremely usual for me (during holidays, weekends etc.) to sleep at 4am and wake up at 12pm (8 hrs of slp fufilled). **grins**   Euro 2004  England out, France also out, Spain out too, Germany also out. All the stronger soccer teams out, now the whole competition seem so empty. Judging by the situation now, i think the finals would be Portugal VS Czech Republic. but then, Czech has never been to World Cup before (not that i have known of), so it can be Greece as well. To round it all up, i tink Portugal would be definitely be in the finals. NBA is now over (at least i tink it is), but then u can\'t catch it live on Singapore TV as well. Super Bowl (NFL) is also over, so, well........   School starting soon  Yep, it\'s the day after tml. Nope, it\'s tml, because it\'s Sunday already (it\'s way psat 12am). I tried to end my \"art\" of sleeping late days ago, but to no avail. So tml nite will be a terrible nite for me, won\'t be able to sleep well. Our form teacher recently put up an announcement, she seems quite excited that school is starting soon. Having regrets are part of being human, well, i now really regret that i haven\'t put a productive use to the holidays nor have a really fruitful one. Reflections are always made to make one regret. Oh well...    Strive I must and achieve I will  It has indeed been very rocky for me the first part of this year. It has not been an eventful and pleasant one too. It has been scary to see how i have wasted my holidays juz like that, i have neither playstation nor xBox at home currently. So, i tink i muz write a notice reminding me to strive always and paste it on my computer screen. computer has been my greatest addiction. i now assure u that u will not see me online so often now. What\'s so gd abt coms? A question pondering me while i did my reflections. I juz wasted so much time on it. It is now hereby necessary to curb this unlawful activity and habit. Sigh, today has been also a very meaningful day. i juz handed over my computer games to my dad for the locking-up ceremony. maybe i will consider locking up the computer too. (password lock pls, not purchasing a giant lock as if the whole machine\'s gonna be purloined anytime) I also hereby promise that I will be much more serious this semester and not fool about being a joker all the time.   NDP tickets, any luck?  My family has applied for the NDP tickets. We got 2 applications, one from each of my parents. Until today, we are (take note of the tense) still frantically hoping for the best. maybe tml in sunday service, i will include it in my prayers! =D sunday services have been a great help to free me from boredom, although listening to the sermons at times can be so............. at least, go to the church, sing some Christian songs, pray to God, and clear your mind. You must broaden your mind!  1:1,  Professor Trelawney  Alrite, back to NDP. So we got 2 applications out of a few hundered thousand, all vying for the actual day parade. In my opinion, many ppl attend NDP so that they can get the free gifts. It has happened, juz last year, a few Chinese nationals (PRs of Singapore) went there, took the bags, and left. How worse can they get? Tis\' a tribute to the nation, how disrespectful!    Conclusion  It\'s great and good as well to sleep late. Today, for my first time ever, i can still see quite a few ppl online at this time. Oh btw, another thing, it\'s also very warm at nite nowadays, so basically i\'m now sharing some \"leng\" jokes with my siblings to cool off. Parents are ranting away, go to slp etc. This must be the first time they ever slept so early, btw. They also usually sleep at 2am. Having really late supper (2am) is also becoming popular.  Juz not too long ago, i had a really late visit to 7-11, with parents of course. Still quite a FEW ppl on the sidewalk, and some cars (mostly taxis), no trucks or pick-ups. As the place i live a lot of condominiums with security guards, so every now and then u can see guards having supper and taking a night stroll as well. At 7-11, u may be surprised to c it packed with ppl, sitting on stools eating instant noodles. Same situation for petrol stations (the marts). As it\'s Euro 2004 now, u can c ppl streaming in and out of a nearby pub. Even i was quite attracted by it and wanted to go in. There u go now, just received a call from my aunt, wanting a  little  chat with my mum (she dint comment, bet she\'s not really sleeping).    Okay gtg now, not to slp, but to play some card games with siblings.   '),(19600,2787343,'26,June,2004','2004-06-26 00:00:00','    Added a new clip below. Check it out! You will need Apple Quicktime player.  urlLink Download QuickTime   PS. I\'m not trying to be racist!   '),(19601,2787343,'25,June,2004','2004-06-25 00:00:00','    Nope, i dint get first, nor second, but third!!! argh!!!!!!! can\'t get over the shock!!!!!! i (being maybe a bit greedy) wanted 2nd at least!!!! haiz... haiz... haiz...   Fated to get 3rd  Last year, i oso took part in National Script Wrting competition (with ks, very memorable again) we got 3rd. But that was my first National Competiotion so contented. But todae, 2nd time, wanted to get at least a 2nd, got 3rd again!!! Wah lao!!! The prize money also same leh!!!   '),(19602,2787343,'24,June,2004','2004-06-24 00:00:00','    Well, it\'s abt 12 more hours from now to my competition. Allow me to reiterate a while, the competition is all about making a Chinese (yeah, chinese) website on Lee Kong Chian (æŽå…‰å‰). Tml, we (2 of us) will have to stand in front of more than 100 ppl and present our website. Although far, but it seems very near to me, because you are going to waste abt 8 hrs later for sleeping, making time pass very fast indeed. A ex-6c of mine (anonymous), also took part in a english speaking competition and had his taste today. It\'s my turn tml, so ppl... wish me gd luck and cross ur fingers okay?   '),(19603,2787343,'19,June,2004','2004-06-19 00:00:00','    Back to Singapore. Went to Bangkok the last few days (back for abt 2 days liao). This proved to be another interesting trip. Firstly, our flight (Air Asia, 12am flight) was delayed by 4 hrs. As it was the wee hrs, so we juz slept in the airport loh... Halfway thru\'s our slp, the security guards woke us up and questioned us. Funny... There have got to more than 10 other ppl slp around the same area, y they only questioned us? Have a 2-hr flight with very little stretching space, the seat cannot recline much, no breakfast etc...  This trip was planned in the first place was because my father needed to go there for some seminar. Now got Air Asia, so the rest of us juz went there and free & easy. Nothing much to c actually, all temples and there\'s a major jewellery store. The only thing i found very interesting was the \"tutu\" cars, three wheeled car-like scooters and opened-air. Can enjoy the smoke while riding along. Although it\'s cheap, but it always brought us to somewhere else first before bringing us to our destination. Had a hard time speaking english to them...  Another thing rather \"interesting\" happened while back at Changi airport. Now i\'m questioning myself whether i looked too much like a kid? Now, after immigration clearance u must go thru\' another security check. (customs) ppl were queueing up, and the metal detector had this sort of \"traffic light\" which tells u when to go thru\' it. So it\'s my turn, then the police officer went, \"Boy, u wait for the green man ah. Ah, like that hor, the red hand, don\'t go in. Wait for the green man to appear then u go in ah. Nah, go in go in, green man appear or-ready.\" That\'s not all. Unfortunatly, the metal detector beeped. So another police officer took the garnet thingy, and said, \"Boy, keep still ah. don\'t move ah. okay very good, don\'t move ah.\" Come on, ppl, i\'m big enough to noe these things already!!!   '),(19604,2787343,'19,June,2004','2004-06-19 00:00:00','    This is my last post for the next 2 days. Life\'s busy...   '),(19605,2787343,'17,June,2004','2004-06-17 00:00:00','    Noticed something interesting on Yahoo!\'s website today. Yahoo! now provide 100MB of storage. Also, with a stronger anti-spam system a fresh new look. (i\'m a yahoo user) now i tink Hotmail is probably the worst email provider already. with only a stingy 2MB of storage. Now probably anyone would wanna join Hotmail is because of MSN Messenger. These 2 other choices would be much better than Hotmail itself, because of their email services: Gmail (Google, still under beta test, 1GB storage) and Yahoo! Mail. (100MB of storage now, more user-friendly)   '),(19606,2787343,'15,June,2004','2004-06-15 00:00:00','    Rmb last time i opened the floor for people to suggest whichever songs they wanna put up? i just got a suggestion, but i would greatly consider \"veto-ing\" it. The song is none other than å¤œä¸Šæµ·... å†·æŽ‰ã€‚ã€‚ã€‚I dun tink that guy\'s joking too, er... well...   '),(19607,2787343,'14,June,2004','2004-06-14 00:00:00','    This must be the household joke of the year! Somehow, the plastic cover of our toilet, er... bowl had broken. So we took it out to repair it. Just before we were trying to screw it back, my father felt rather erm eccentric and took it to the kitchen basin to wash. Not only that, he was using the scrub we used for washing our dishes. My mum was really furious and hope that doesn\'t turn into domestic violence. (touchwood) My father was like, \"what\'s wrong lah?\"... Imagine having to eat using plates that are scrubbed with the brush once used for scrubing toilet bowls... Yikes!   '),(19608,2787343,'13,June,2004','2004-06-13 00:00:00','    Yes i c the \"art\" of \"å†·\" jokes now... Was browsing thru\' the web when i notice this.   1) a cow goes to the south pole... (guess a sports brand) 2) which country has the hottest milk? (guess a country duh) 3) what do you call a blind deer? (guess a phrase)  here are the answers!  1) new balance (niu bei leng si) 2) nigeria (nai ji re) 3) no idea (no eye deer)    Can you endure this round of \"leng-ness\"? i guess i can be that \"å†·\" after all...   '),(19609,2787343,'13,June,2004','2004-06-13 00:00:00','    Erm... Even i am now starting to get bored of é‚“å¦™åŽ. Now i open the floor for suggestions. but here some disclaimer first, no é‚“ä¸½å› or whatsoever. But, modern pop songs are quite difficult to find, unless u want \"she bangs\".   '),(19610,2787343,'11,June,2004','2004-06-11 00:00:00','    today met one of my schoolmate. we began telling \"å†·\" jokes.  him: hey jensenï¼u here also ah? want to hear some jokes?  me (feeling damn sian, waiting for mum at cashier): okay loh...  him: there are ten birds. the hunter opened fire and hit one of them, y dint the other 9 birds fly away?  me: becoz their wings are injured?  him: wrong. becoz they are ostrich so they can\'t fly.  me: å†·...  him: ä»€ä¹ˆæœ‹å‹ä¸èƒ½äº¤ï¼Ÿ  me: åæœ‹å‹ï¼Ÿ  him: é”™ï¼çŒªæœ‹ç‹—å‹... u want to make friends with a pig and a dog meh?  me: ** shut up, and quietly left, feeling real cold...**  him: oi! å†·ç¬‘è¯ä¹Ÿæ˜¯æœ‰\"art\"çš„ï¼Œokay?!   The art of â€œå†·\" jokes?! Well, norvin sure has some. (that guy is not Norvin lah, but norvin damn pro in telling lame things too...)   just days ago at the camp, one of the seniors told a damn lame joke.   There are once 2 bananas which had a race. After the first banana ran for a while, he felt hot and took out his outer covering, and then threw it in front of him. He then ran over his outer covering, slipped and fell to his death.   End of part I...  Years later, the son of the dead banana wanted to have a race again with another banana. The son learnt from his father\'s lesson, so this time he threw his outer covering behind him. Then after one lap, the son later then stepped on the outer covering and fall to his death too...  End...    What\'s so funny about this \"joke\"? å†·...   '),(19611,2787343,'11,June,2004','2004-06-11 00:00:00','    A New Song Added  If you look carefully around this website, you may already found that i have added a embed for songs. For those of you who haven\'t found it, it\'s just below the disclaimer and copyright. The song i have chosen this time, is ç‰µå¼•ï¼Œfrom é‚“å¦™åŽ. Yes yes, i noe, old song it is. It\'s meant for enthusiasts for the good ole golden songs of the 80s like me =P. (yes i noe kit, u somehow have a strong liking for these kinds of songs) alrite, even if u dun appreciate it, just let ur dad or mum listen lah...   The Resolution thingy  Nope, it\'s not about new year resolutions or whatever. But some feedbacks to me have complained about the text being too large and thus making the blog uglier. well, is it? later then i found out that they are having 800x600 resolution. well, this blog is designed (or at least i think) on 1024x768 resolution. So, sorry, i really can\'t do much about it. Either bear with the ugly design of the blog, or change ur resolution. Anyway, thanks for your valuable feedbacks!   Nothing happened much over the weekend, so no topic to discuss. But be rest assured that i will find a topic to discuss next week!   '),(19612,2787343,'11,June,2004','2004-06-11 00:00:00','    For those who have cried, begged or whatsoever for a new update, here it is! And now for those who have been wondering, i have been to a camp these few days, so no update.    Harry Potter and the Prisoner of Azkaban  I have to quite agree with the critics, this film has not met its full potential. I don\'t think it\'s very good. Firstly, the new director Alfosno Cuaron has missed out quite a few main and important parts of the book. And also, the sequence of events are somewhat slightly changed and some facts of the book also changed. The plot is just sort of different from the book. Also, has Hogwarts a clock tower and the beautiful yard? Does it have the bridge or something? When does it have such beautiful scenery? Well, this is something for you to wonder. Now for the positive points. The visual effects are nice, and i especially like the Dementors! =)   The Amazing Race  During the camp, I played in the amazing race. Basically, we must run around the southern districts of singapore and complete lame tasks. During the trip, quite a few interesting things happened to share about.  1) If you are a regular passenger of the MRT, you may know this. Currently, every single second in every single MRT station owned by SMRT, you will hear this familiar song. \"æ‹œå…­ç¤¼æ‹œï¼Œä¼šä¸ä¼šå¼€ï¼Ÿâ€Familiar enough? Good.   2) In Orchard road, we were cornered by some ang moh entertainers, in conjunction with the Singapore Arts Festival. We were practically doing lame stuff with them, like smiling in front of a fake camera which often breaks down. I got a photo as souvenir, so for whoever who wants to see it, maybe next time when we have the chance.  3) In War Memorial Park, we had a task which requires us to tell a few lame jokes. It\'s qutie er... inconvinient to share them here, so maybe again next time when you meet me, or thru\' MSN Messenger =D.   Alrite, i tink that\'s all that i have to say, or rather can say. Until next time, cheers.   '),(19613,2787343,'06,June,2004','2004-06-06 00:00:00','    Sunday, a day of brightness and happiness. However, today is sure a dark and sad day.   President Reagan passed away  Ex-President Ronald Reagan, a \"true American hero\", passed away at 93. He passed away at 93 in California. President Reagan stopped the Cold War years ago. From a Democrat, he somehow changed into a Republican during his political days. Somehow, Reagan\'s life is partly alike to Arnold\'s (can\'t spell his surname) life. He was firstly an actor, then he entered politics and became governor of State of California.    Back to Singapore, Back to The World of Worries  I have basically no time at all to share my experiences in my trip. Maybe next time. Since the few hours i\'m back in Singapore, new work and news from school were thrown at me. I have now bad news, and have extra work to do. Haiz, it\'s just a few days overseas to relax and i come back with everything thrown at me and new worries to come.   '),(19614,2787343,'04,June,2004','2004-06-04 00:00:00','    Well, it has been days in New York and i\'m fortunate enough now to get a chance to update a little.   So much significance...  This trip really had a lot of significance. Firstly, on arrival in New York JFK airport, something MAJOR happened. Somehow, United gave the wrong immigration form. They gave the white one, but Singapore is under visa-waiver so we should get the green form instead, We queued up for nearly 20 mintues before we reached the counter. There, the immigration officer went, \"Hey pal! I think ya got the form all mixed up! Ya should take the green one instead over there.\" My father was already half-way scanning his forefinger (we thought we also needed to scan)... Well, we went back to the form-filling counter, only to realise that they have ran out of english forms. So we took a Chinese one while the rest are a mix of Japanese and Korean. Still, we passed that round. The officers were very strict, like so pissed with them. They like compare very carefully the photo in your passport and you yourself. They took out so much time that a Chinese tourist in front banged on the counter. Haiz... My mum was also so pissed that she waved her hand in front of the officer\'s face when he was checking her passport.   That\'s all for now. Will update more when i come back!   '),(19615,2787343,'01,June,2004','2004-06-01 00:00:00','    Good bye everyone. Remember, don\'t call my handphone or even my house. I\'m leaving the perfect Singapore for the tornados, hurricanes, floods, storms or even the latest cicada attack.  [bbb]   '),(19616,2787343,'01,June,2004','2004-06-01 00:00:00','    This house is getting totally crazy! Not only there are mickey mouses everywhere. But my neighbour\'s dog Mickey also likes running about in our house now. (My neighbour and us have very good relationship, 100 times better than the Joo Chiat one) Not only that, my brother and my sister have suddenly caught on the \"Mickey\" trend and started the Mickey song and sing. My maid seemed to join in the trend also, singing along while washing the dishes in the kitchen. What the Mickey?!   '),(19617,2787343,'28,July,2004','2004-07-28 00:00:00','    Today is again another lame day. (in fact, un-lame days at school are actually more or less numbered.) we went to the national stadium after our lessons, to cheer for our track & field team. i realy felt like sleeping, it\'s so boring and we were squeezed like lemons, the cool weather also added to the impressive conditions for napping. however, the weather got so cool that it actually poured, we had too much people to move into the shelter and of course i got drenched. it actually rained at 2 separate times, so this is the procedure: in, out, in and out again from the shelter. when the drizzle started to get heavier, i proceeded to get my umbrella. but the eccentric (and senile, no offense!) told us not to open umbrellas. what if i catch a cold, sue them and claim compensation? my whole bag was also dreched and my school materials were wet, and i\'m really and seriously planning to ask for compensation as well. some of my exercise books are so wet that they can\'t be used any more. mum was of course really furious, said she would call the school authorities SOON. (how soon is soon?)  i stayed till the end of the prize-giving, throughout this time, i really envied those \"escapists\". and when we really started to leave, i followed the crowd and initially thought that i had gone the wrong way. so i went to geylang interchange and took another slooooowwwwwww bus, 985. it took me around 45 minutes, so i reached home at 9pm, wet and hungry...  oh, i forgot to mention the most important part. our school got double champ again this year!!! (1st for both c boys and b boys) however, hcjc\'s performance was quite abysmal. let\'s hope that tml we can get half-day...   '),(19618,2787343,'27,July,2004','2004-07-27 00:00:00','    something lame i noticed today. bus no. 74 is not the bus with the highest frequency. today, i was sorta quite attracted to the bus-stop out of my bus window. guess what\'s the most frequent bus in singapore now? it\'s none other than 61, of our very own Toh Yi Estate and Hoover Park. How frequent is frequent? for 74, u r likely to get around 3 buses in 5 minutes. for 61, it\'s ONE BUS NEARLY EVERY MINUTE!!! and the bus are those SMRT Bendy buses, the long type. of course, this scenario can only be seen during peak hours, from around 5pm to 8pm. and now, guess the bus with the least frequency in singapore... that\'s right, it\'s again from our very own Toh Yi Estate and Hoover Park, 66... 66 still currently hold the record safely for the bus with the lowest frequency in singapore. so how slow is slow? 1 bus in 45 MINUTES. i really pity those from our school who lives in jurong east and can only take 66 back. wait for so long everyday...   '),(19619,2787343,'26,July,2004','2004-07-26 00:00:00','    Yep, today\'s my birthday... i counted the no. of birthday gifts, fewer than last year. every year, birthday gifts and practical cards get lesser and lesser. This is so disappointing, as more and more are sending e-cards. nvm, make my \"thank you\"s eaiser to send. i just send a reply message or catch them on MSN. In the past, for not so close friends and relatives, had to call and thank them. for the rest, just say \"thanks\" the next time you meet. or rather, when that person gives u the gift in person, u thank him personally. with email and online techonology, it\'s so much easier. also, i feel that each birthday is becoming more and more unsignificant. i nearly dint get a birthday cake today...   LongHorn  I\'m so frustrated with BitTorrent, i\'m still downloading longhorn despite the fact that i started downloading on saturday night actually. it\'s still stuck at 10%, and my family\'s really too poor to let it run overnight. just wait loh... u must have the patience to succeed, in which this case it refers to the benefits (and the non-benefits) the LongHorn may bring to me. For those who don\'t know what\'s LongHorn in the first place, go to http://www.microsoft.com. aniwae, for others who are downloading, trust me, u can........ die from it. what is most fortunate is that i have strong pyschological stand in me, so i\'m unaffected from it. do the breathing exercise when you are flared up. btw, sylvanus must have finished downloading by now...   Of MSN Messenger  A classmate of mine today (anonymous) had unleashed claims that MSN Messenger has more than 150 allocation places. He had 156 contacts. I don\'t understand whether i\'m surprised or exhilarated upon hearing this piece of news in the first place. yep, i\'m one of those \"siao\" ones who have reached the max. no. of MSN contacts. Often, when i want to add in new contacts, i would either have to delete someone else or just allow that person to be on my allow list. Just today, i have checked with several other fellow-chatters and they all confirmed that Hotmail had NOT recently reviewed its policy on this matter. sry to disappoint you all guys, but the max. no. of MSN contacts you can have REMAINS at 150. Maybe with the coming major changes and upgrading Hotmail is going to take soon, they should also review the allocation places for the no. of contacts in MSN. Since they can add more storage space to their emails, y can\'t they allow more contacts in the MSN Messenger? Or maybe Google implement GMessenger, with more contacts and faster and more stable connectivity.   '),(19620,2787343,'23,July,2004','2004-07-23 00:00:00','    I really have to agree that IE (Internet Explorer) is the worst among the number of browsers available. Even with the new pop-up blocker feature, it basically still sux a lot.  1) Opera - The design of this browser is less than sleek (it\'s quite square-ish)and it carries ad banenr as well. good thing is it provides tabs which mean u can open many sites at one time without having many new windows appearing.  basically alrite, shows in you detail the downloading of the site (the progress, such as the current downloading speed etc.) basically, i would love to use this if some improvements are made. be careful, embed media players etc. will not work... with pop-up blocker and password remembring tool. downloading is faster than other browsers too. certain things may not be viewed too.  2) Mozilla - Nothing special. Looks like IE but it\'s a bit more stable and faster than IE. Includes pop-up blocker and password remembring tool. some problem may appear because of tabs.  3) Safari - I like this browser the most actually. it\'s the most reliable and the fastest. Too bad, it only works with Macintosh. It\'s fast in downloading applets too. anyway, don\'t try to download it to windows. it won\'t work. i liked the interface quite a lot, sleek design.   4) Netscape - Sleek design. love the interface. but nothing special. i dun tink it has pop-up blocker too. but there\'s still a password manager.   '),(19621,2787343,'22,July,2004','2004-07-22 00:00:00','    Instead of today, let\'s talk about tomorrow now. So, tomorrow is basically a very slack day for me. I will only be taking the first 2 periods of the day and it\'s off to make the CNA video film. Glad to miss the other subjects, they are all the subjects i hate most. Tml just put up a good show and maybe we can get dismissed even earlier.    \"Dumb\" Quote of the Day  Quote of the day, \"critcising ks can be really fun. you will never get the scandalous sensation of him cringing bitterly and revolting on your remarks, no matter how painful it may be. this is one of the most impressive features of ks, as he\'s one of the minority who doesn\'t revolt when someone\'s actually having on antagonistic remarks at you. everyone, treasure this priceless chance, so let\'s all start making fun of ks.\" well, here you go. i quite agree with this statement! ks is of highly venerated because of his impressive attitude towards people. keep up the work, ks!   Of New Computers  Yes, we are finally going to get a new computer. and yesterday, i found out how SOON is soon. it\'s about 1 year more. my father said \"he needed to save money for this\". haiz, waiting slowly... anyway, i can\'t download the new build of LongHorn because my com really cannot take it. was quite frustrated and i ended up having only installed the \"LongHorn transformation tool\" which can only change the APPEARANCE to LongHorn, but not the features...   '),(19622,2787343,'21,July,2004','2004-07-21 00:00:00','    I hereby promise that i will TRY TO use proper english from now on so that i can improve my linguistic prowess or whatever you would call it. Really, i realized my language skills have deproved quite a lot lately, and i\'m determined to revive it. i find myself speaking \"unacceptable\" english nowadays, really frustrated at it. somehow, i can speak certain mandarin or malay at the same time while speaking english halfway.    Chinese &amp; Such   Today met tangc. He seemed delighted as he was smiling the whole time while he was chatting with me. maybe and hopefully that it\'s something to do with our test results. we have done well in our tests, that\'s why he\'s so happy. still, i\'m prepared for the announcement of my abysmal chinese results. anyway, today met his wife and 2 daughters in the auditorium. they were taking photos... getting apprehensive of the release of chinese results. i really want exemption from the chinese subject, as chinese usually requires a powerful memory, and it can be considered as quite a dead subject. as long as you study and memorize, u can pass with flying colors. but the problem with me is that i have a really bad memory and linguistic aspects of studying is not really my cup of tea.    Racial Harmony   Racial Harmony today was basically a farce. pupils from west spring sec came, and they were just another couple of darn brants. i heard certain chs guys are so despo that they are actualyl requesting for the tel. no. of the girls. today, i was also noticing the costumes of the teachers. miss wong (math) was wearing something that looked like a night-gown, tangc was wearing black all over, black shirt and pants, gohsj (music) was wearing something arabic, limcc (geog) and quah (eng.) nothing special. dint see teocm and kala today. so basically it was a boring day, we just lamed around and looking at exhibits. we were so desparate to be lame that we came out with lame quotes, like \"it\'s lame to be lame\" and \"it\'s even lamer to be lamely lamer.\" our world today revolves around the word lame.    Currently...   it\'s even lamer now to be doing hw at this time. i still have a stack of math ws to complete. it\'s most annoying. rest of my family clapping alont with \"the price is right\" on cbs channel with bob barker. i tink this is a lame show and wondered how it has survived and persevere till today. yaya, the price money are enticing, but it\'s still basically a lame show. i tink it\'s something to do with the personal character and attitude of bob barker, the show was preserved till today. again now, listening to william hung\'s \"inspiration\" while delightfully and exhilaraingly completing my math ws. it\'s a invigorating source of energy, william hung\'s music. i especially like to listen to him sing YMCA, very forceful and energetic, with vigor and all... oh, i can feel my nose growing longer now...    Before ending my post, i would like to say that my blog has become a medium for me to shed any lameness in me. basically, i found some of the posts i have created rather lame. for my readers, just treat them as being \"hilarious\"... hahaha... now i c y ppl don\'t comment, i can\'t comment on my own posts!! OMG!!!   '),(19623,2787343,'20,July,2004','2004-07-20 00:00:00','    Test  Today was Chinese test. the first few sections were alrite, i was getting confident enough, until i came to the last part, comprehension. i was really pissed with the passage, don\'t know half the time what it\'s talking aobut, it\'s about the story of some clone, about how humans are smarter, hehe... the questions basically sucked, and i have a feeling that it\'s the same answer for every answer the whole time. tangc basically was sorta surfing the net (can\'t see, i was sititng at the corner) for 2 whole periods we were doing the test. tangc don\'t look too happy recently...   Racial Harmony tml  tml is racial harmony day, so basically no lessons. the vip for tml is dr vivian bala krishnan. got the joke? no? nvm... aniwae, still dunno what to wear for tml...   '),(19624,2787343,'19,July,2004','2004-07-19 00:00:00','    Blocks of Life  i do notice that ks have always positive results for his tests or exams or wadever. he nearly got all A1s... well, in my class, if u got A1, the teacher would be congratulating u quite a lot and people will be so awed with your results. one word to say it all, \"incredible\" to get all A1. My tests all nearly scrap either A2s or B3s. i admit that i wasn\'t really that conscientious in my studies, but at least i do revise before tests (few days beforehand) and pay my fullest attetion in class. even with this, scraping only A2s most of the time. and usually, i\'m just that close, 1/2 mark or 1 mark, something like that. oh btw, getting nearly full marks (such as 26/30 OR 17/20) is really really really uncommon. for maths, we have highest marks like 35/40 (by a real maths pro) but the rest are averaging on 24/40. i got 28/40. haiz... it\'s just so that hard, the criterias make it difficult.   Chinese Test tml  yes yes, chinese test tml, and i\'m studying \"like siao\" back home here. there will be as usual 10 questions on word application (without hanyu pinyin pls) and another 10 on \"ci yu\" application. then we will have 5 sentences to make and another 10 questions on chinese culture, like the story of ã€Šåˆ—å­ã€‹and must memorise how to write the names of characters like \"ä¹æ–¹çš‹\" and \"ç§¦ç©†å…¬\". also, must remember \"å¤©å¹²åœ°æ”¯\". let\'s see if i can get it right now. \"jia yi bing ding wu ji geng xin ren gui, zi chou yin mao chen si wu wei shen you xu hai\". sounds confusing? not really, this are just some of the requirements. and then there will be something from our ke ben as well, just read them hard enough.  gtg, must really drill from now on, or else won\'t pass tml. serious, a lot of people fail tests.   '),(19625,2787343,'18,July,2004','2004-07-18 00:00:00','    Played SimCity 4 today and yesterdy. quite fed up with it actually, it\'s all so slow and laggy. i\'m extremely certain that it\'s due to my slow com (501mhz). still can\'t blame it, we are really really poor, just wait until my com gets upgraded. just wait patiently, one must show patience and determination to achieve...   Horn It  ks has got LongHorn. he\'s downloading it now and he\'s blocking me as \"i\'m distrupting the downloading of the software\". horn it! i\'m so jealous, my computer is too slow for longhorn and my virtual pc basically can\'t run well.   Free Tuition Class  Actually, composing every post in blogs are not easy and they are sure good for your development and polishing of writing skills. i recommend keeping blogs for those with low comprehension and composition skills. you may just improve, serious...   '),(19626,2787343,'18,July,2004','2004-07-18 00:00:00','    Today, somehow i got wind of something of grave importance. someone in our school is gonna face grave punishment because he [or she =)] is defaming a certain teacher in our school. next time, i will&nbsp;try not, oh not try, never&nbsp;to talk about teachers in my blog anymore. too risky, really. wonder how the teachers found the blogs of students. for students, i strongly recommend that you don\'t tell the address of your blog to any of your teachers. (ex-teachers (primary school) maybe alrite, they can\'t punish you and they can\'t expel you you too). i dunno whether it\'s true, but it\'s said that this guy is gonna face expulsion... be careful and warned, this is a very realistic world. even though you may not be from chinese high, but... this is still a realistic singapore. it\'s just defaming...  &nbsp;  oh btw, notice that certain elements in this blog has been either edited or removed so that i won\'t touch on sensitive issues about teachers. really freaked out now.  &nbsp;  PS. Think you know me well? Take my quiz!  urlLink http://someone1991.friendtest.com   '),(19627,2787343,'17,July,2004','2004-07-17 00:00:00','    Today, basically we went to mass jog. since i\'m on 1-week pe excuse, i dunnit run. so the whole thing is quite \"shuang\".   Of Songs  My blog has indirectly become a medium for lame songs and music. This week, i looked into \"my music\" and only found william hung. so there you go! enjoy! this song is taken from william hung\'s latest album, \"inspiration\".   '),(19628,2787343,'16,July,2004','2004-07-16 00:00:00','    Saw this interesting article in ks\'s blog...     Casuarine and I were in the same room - room 1. Room 1 is also known as the most dreaded place any interviewee ever wanted to go. It consisted of 3 EXCO (Executive Commitee) members - with the president Kegan!!! The questions they asked are all crap - and tough. Really tough. So tough that I didn\'t even expect them to come up with some \"jian\" questions like that. One example: If you had to organize a meeting for a project (and you have 9 members). 4 suddenly cannot come. Would you still continue with the meeting or would you cancel it? Provide reasons.  If i were you, ks, i would simply reply this... \"Casuarine and I were in the same room - room 1. Room 1 is also known as the most dreaded place any interviewee ever wanted to go. It consisted of 3 EXCO (Executive Commitee) members - with the president Kegan!!! The questions they asked are all crap - and tough. Really tough. So tough that I didn\'t even expect them to come up with some \"jian\" questions like that. One example: If you had to organize a meeting for a project (and you have 9 members). 4 suddenly cannot come. Would you still continue with the meeting or would you cancel it? Provide reasons.\"  &nbsp;  i was in worse sitaution before, once i was sabotaged and nominated for some rank (i dint even wanna do it).&nbsp;they asked, \"do u noe what\'s the whole purpose of this job?\", \"if i were to call u to initiate something from the student\'s welfare, what would u initiate?\", \"can you&nbsp;think of&nbsp;activities for the healthy growth and positive learning at the same time for the students?\", \"what other commitments you have and how will you ensure u have sufficient time for this job?\", \"if you have test,&nbsp; competition, mentor wants to organize meeting with you for mrp and project, and u have stuff which requires days to do for this job as well, what will be your first priority and how will you plan your time?\" i was completely stunned, and just mumbled... i wasn\'t even prepared for that job, =P. who cares, i\'m now a carefree man, who has completed most of mrp, no more project work and no more scripts to write...&nbsp; &nbsp;&nbsp;  Of Studies  &nbsp; My results have been in the worst state for this year. I have&nbsp;performed badly for certain assignments, and i have a feeling that i won\'t do that well for tests this term. this weekend, luckily no much commitments so i have time to study well for chinese test. this term got a segement for appreciation of chinese culture, they will ask something like \"tian xia zhi ma\", \"tian gan di zhi\" and something else from our \"ke ben\",&nbsp; like who is the writer for this particular \"story\" etc. omg, have to memorise the names of the writers and learn how to write them as well. i must really do well for this chinese test, or my mean grade this term will swoop and plunge down.   '),(19629,2787343,'15,July,2004','2004-07-15 00:00:00','    Today\'s fever reached 38.5 degrees, but i muz still admit that i\'m still perfectly in a good mood and energetic. There are 2 possibilities for today\'s fever:  1) Today we played basketball in the rain. and i\'m not perfectly well from the last fever. so there comes another round of fever. strange, because my fever sorta started after the pe lesson.  2) the chocolates ms teo gave me were poisoned. and she finally rmb my name, \"jensen, come over pls\". anyway, she\'s not realy that evil and she\'s really really really nice when she\'s in a good mood. i\'m not biased towards her, but somehow i don\'t like her a lot. i must try to adapt... she\'s at least better than kala... today at ms teo\'s table, saw a photo of his son and husband. she\'s married alrite...   currently, talking to yinghan (my ep3 senior) rite now. he\'s asking me questions about groups now. suspect that he\'s trying to kick me out of his group...   '),(19630,2787343,'14,July,2004','2004-07-14 00:00:00','    Life feels so empty now. my team dint qualify for finals... haiz... nvm lah, at least we are now free from the worries and woes of project.   '),(19631,2787343,'13,July,2004','2004-07-13 00:00:00','    Ms Teo  Today, i officially announce that ms teo is out of my hate list of teachers. she\'s quite a really really nice teacher really when she\'s in good mood actually. juz today, she played a game in class and promised to buy us chocolates. keep up the good mood, ms teo! oh, and she is married after all. just recently, she let slip in class, \"u noe, my son...\" basically the whole class just went \"orh...\" but she dint respond.    Project Semis  The results from certain categories are out. too bad, my category is not out yet, so just wait patiently loh... those who wanna check, just logon to project\'s day website.   '),(19632,2787343,'13,July,2004','2004-07-13 00:00:00','    Today is the semi-finals. I must say that we did unmistakably quite well. If we are accompanied by some strokes of luck, we may just get into the finals! =D Wow, dream along, dream along... Aniwae, i just tink that we muz really change our project title. we are getting the analysis for the whole subject but not the particular aspect and angle we are supposed to be focused on.    Of buses   Nowadays, the frequency of 66 has increased, good job, sbs. y? becoz 66 is always the first bus that comes when i step out of the school gate. During this period from 3pm to 7.30pm, that damn path  inside  the bus would clog up. These people are unmistakably from Little India. I don\'t mean to be racist, but, boy, u can really get the aroma floating around inside the bus. Also, with so many people, i have to stand up all the time. This is the part i hate most, the MAD DRIVERS of 66. There was one time i nearly fall on the driver himself. luckily i held on to a metal bar on time. or some drastic accident may happen.    Of teachers   Quah is feeling \"lovely\" lately. she wanted us to write some love story, and mark her words, she said she will read all 35 stories of everyone to the whole class. imagine her reading some mushy stuff to the whole class and mr. hon walks by, good contrast... What\'s worse? she created the whole setting for us that it sounds more like some love stories in ancient china. the male counterpart is called \"fu gui\", and the female one is called \"ru hui\". \"fu gui ru hua\", appearing in some place like Chinese Garden, drinking chinese tea in one of the pavillions by the lake, OMG!!! how romantic... alrite, back to the topic, teocm is in EXTRA good mood lately, she\'s all chirpy in class and as emotional as ever. Oh, did i mention be4, she\'s the discipline mistress of sec 2. oh well...   '),(19633,2787343,'12,July,2004','2004-07-12 00:00:00','    Project\'s Day and MRP  MRP and Project\'s Day are a complete waste time. They keep us really busy, but do the students really put good productive use to it? Why don\'t they just organize overseas field trips? Like last year December, they go to Havard in Boston. Along the way, they stopped in New York and tour, have a good look at the famous New Yorkish fire-engines. Also went to Ground Zero, Time Square, Grand Central and Empire State. Imagine myself in the Western US pulling slot machines while my teacher-to-be at that time is (December 2003) at the East side. Wow... Alrite, back to the topic. Project\'s Day is tml, so i\'m getting the jitters right now. So May God Bless Me. Stay tuned to my blog for the results.   Hotmail is degrading  Hotmail recently is very problematic. Firstly, the whole inbox website is full of pics. lately, download is extremely slow and i\'m very tired all the time waiting for the pics to load finish. Next, there are more advertisements as usual, maybe it\'s just after Independence Day or something like that. These ads have none of my concern at all. Some are about \"mortgage rates are the lowest in 40 years\", some about some dingy \"state farm car insurance\" and \"all state insurance\", some about \"love.com, meet ur new love\" and some about \"Nevadan state reserve has increased loan rates blah blah blah\"...These advertisements are very tiring when the inbox is loading so slowly. damn u, hotmail, improve ur services. or i will switch to gmail, forever...   Comments  Lately, there are no comments for my posts. wondering y, maybe because so busy for everyone these few weeks, and i also dint advertise much lately. dunno lah, tests are starting next week liao, even more busy. after term 3, it\'s back to skool at term 4 and it\'s final exams straight away liao. after that, real peace and enjoyment is found... aww...   '),(19634,2787343,'11,July,2004','2004-07-11 00:00:00','    Oh gosh, alvin tan criticised the project\'s website today!!!!! Firstly, \"the color does not make a good enough contrast\". Then he suggested having the customary black and white again. That is just so uncreative and unoriginal. Secondly, \"the words are too small\". It should be the customary font size 24 again. Ours was much smaller. He was very worried that the judges won\'t be able to see our words, so he told us to do a ppt instead. Imagine tml, we are the only team presenting in ppt. OMG!!! Then the audience will ask, why are you using a ppt when you should be introducing your website? die die. aniwae, ks, don\'t despair. i understand that it\'s not your fault.   '),(19635,2787343,'11,July,2004','2004-07-11 00:00:00','    Was browsing thur\' blogs just now. Here are few of my findings: 1) Chee Yew whom i tink, has completely given up his blog. aww... that is juz so extremely sad 2) Zhong Han added a new music to his blog. it\'s extremely irritating and it\'s none other the fur elise. omg, how classical. 3) Yi kun\'s blog is still as cute as ever!!! awww.... 4) jacky has finally link me in his blog. good for him! victor gan muz keep this out.   '),(19636,2787343,'07,July,2004','2004-07-07 00:00:00','    Fun Day  Yesterday was a tough day. In the morning, we went to BTPS Fun Day, which really wasn\'t very fun at all. It\'s meant for those P1 rascals, for them to jumped on the inflated balloons, for them to play some stupid games. Met Mdm Wang (tour guide for the day) and Mrs Lee. Poor Mrs Lee, she\'s so thin now. Mdm Wang looks the same. Aniwae, Mrs Lee got some heart problems, and she juz took 3 days MC recently. She dint took a day of MC when she taught us...   Project\'s Day  Yesterday, Ivan, Reuben and ks came to help in the Project\'s Day. ks designed the website nicely, with a very smooth feeling etc. That was one of the best website ever created, i feel that it\'s even better than our \"Lee Kong Chian\" website as it\'s more user-friendly. I\'m certainly proud of that website. So, we worked till about 8.30pm, then ks went home. btw, ivan and ks had a sumptous dinner at my hse, i\'m a very good host. they also had air-conditioner to ensure a soothing environment for them to work in. So we completed most of the website by 10 pm. So, for relaxation, watched \"Good Morning America\" on ABC channel. After that at 11pm, watched \"Price is right\" on CBS channel. 12am, watched Judge Judy. At 1am, the \"OC\". After that, it\'s sleep.   '),(19637,2787343,'07,July,2004','2004-07-07 00:00:00','    Found this statement from the blog of my senior (translated from mandarin): \"The moon has been especially round and bright lately. On the lunar calendar 16th, it\'s also July 4th. That is America\'s Independence Day. Long ago, the ppl made a mistake. That is to let the americans form america.\"  awww, really hurts my national pride. there\'s nationalism and americalism in me. sigh, if america was nvr formed. maybe i\'m still somewhere in china, living in a village somewhere in teochew.   '),(19638,2787343,'04,July,2004','2004-07-04 00:00:00','    Life is getting busy with Moron\'s Retarded Paper (MRP) and Project Work. So here\'s a quick and short update finally.   The Crazy Bus Service  Lately, i often take service 66 back home from school. service 66 has this bunch of MAD DRIVERS. most of the time unfortunately enough i\'m standing. service 66 then keep accerlerating and slowing down. then, when it finally S.T.O.P., there are 2 possibilities. a) u fall down on the ground and break your bone. b) u find yourself performing some aerobic stunt like stretching to the back touching the ground, and a sudden excruciating pain in your stomach. with ur heavy bag on your back at all times.  juz today, the driver was tapping his fingers on the horn (in the steering wheel) while driving with 1 hand. i dunno whether it\'s the same driver all the time, yesterday i took the bus at a different time, so... i dunno lah.   Of Songs  I\'m going to change the song weekly in my blog, or whenever i like it. sry to chinese song fans out there, but i HATE chinese songs actually. so only english songs, or occasionally tamil lah. reason y i hate chinese song, and they seem or rather \"hear\" so \"nice\" all the time? if i\'m not wrong, i have about 1/4 ang moh blood in me and this is the part of my boiling blood that triggers off the hatred. really, i\'m not lying, i just checked out my mother\'s family tree to make sure.  That\'s all now. have something to blame ks, he dint say thank you nor even mention me in his latest entry. one part of the entry actually concerns me, i made him found his happiness.   '),(19639,2787343,'04,July,2004','2004-07-04 00:00:00','    There seems to be something wrong with every blogger\'s blog. How annoying and frustrating, when when \"cannot find server\" page keeps coming out. blogger is starting to get unreliable too.   It\'s just too bad  i hate to say that i did not get thee ndp tickets. how annoying. juz give me the tickets what. oh well... it\'s impossible to negotiate now. they are already giving out the tickets.   Homework  juz recovered from fever and my brain is still feeble. how annoying, to see a 6 feet stack of homework lying cheerfully on my desk. luckily tml no school, or i won\'t be able to complete them. argggggggh. have to work my feeble mind on these sheets of paper. purely paper. oh well.  anyway, happy youth\'s day and happy fourth july to all.   '),(19640,2787343,'03,July,2004','2004-07-03 00:00:00','    high fever today. luckily and fortunately monday is school holiday.   '),(19641,2787343,'01,July,2004','2004-07-01 00:00:00','    guys admit it! i\'m a prophet. i predicted portugal would surely get into the finals. and my deduction came thru\'! tonight, it\'s greece and czech. most probably czech will win, but don\'t be too sure about that. greece is not bad this time too. great... oh well... can\'t watch match liao, because parents don\'t allow me to slp late...   '),(19642,2787343,'01,July,2004','2004-07-01 00:00:00','    I can finally see the sea (alliteration here, eh?) of bloggers. Added another link in just 1 day. seem like all my classmates finally started to create blogs. so, check out the \"bloggers\" section of the menu on your right.  (basically, qi han, if u r reading this, i\'m actually making it easier for you to check out blogs of your classmates.)   '),(19643,2787343,'08,August,2004','2004-08-08 00:00:00','    Today went to WestMere (a condo which i suspect where 1 of my EP3 seniors live) as usual on a weekly basis. However, today was especially memorable. i got chided by the idiotic security guard for no reason. i was wearing sandals and running (more of brisk walking) about the pool area, not even stepping on the water. then this security guard came and chided for \"running around the pool area with shoes on\". Hey, ur condo is so small that u got to pass by the pool whenever u want to cross from block to block! If u want to scold me for running around the pool and it\'s very dangerous, it\'s more acceptable. but i wasn\'t even running, i was only brisk walking...   tot i saw my ep3 senior the last time in westmere (he did not see me). so today went looking for him. dun tink he\'s home though. oh well, nvm... sorta shadowed him the last time, so now know where he lives.   '),(19644,2787343,'06,August,2004','2004-08-06 00:00:00','    WE FOUND MS TEO\'S PASSWORD!!!  Don\'t think i should really post such stuff on the web, but... The whole class actually found ms teo\'s password to her CHS website. There\'s this section \"History\" which requires password for entry. Incidentally, we found her password. not only her, but the passwords of several other teachers as well. but aniwae, i tink that \"History\" section is meant for students as all the notes are inside. so maybe, she used to tell her students the password last time...   National Day Celebrations  Today is a slack day for us. The celebrations ended at around 9.30am. Ex-classmates of last year were initiating to go back to primary school. but since i went back during fun day 2 weeks ago, don\'t feel like it... In the end, none of us really went. So i rushed home, can start playing my Final Fantasy. However, there are 3 term tests next week. So really must buck up, study well during the 4-day holidays. yes, i got some stupid hw (algebra and geometry, the most sianz topic of all). but don\'t feel like doing it today, only wanna play... Really must study hard for history, my worst subject. i barely passed last term.    Gifts & Such  ks wanted Sims2, so planning to get for him (pirated?). in exchange, i want (take note of the tense) Harry Potter 3 (PC game) and Final Fantasy XII, which is due next year. I don\'t mind pirated stuff honestly, pirated final fantasy can be really nice also. but for harry potter, REALLY REALLY have to buy original version, graphics of pirated version really suck.   '),(19645,2787343,'04,August,2004','2004-08-04 00:00:00','    It\'s so annoying yesterday. Apparently, we were made to give a speech during our Chinese lesson. So yesterday, it was my turn to give the speech. I basically screwed up everything and it\'s yet another farce. Just when i was speaking halfway, certain people started laughing and making fun of me, especially the guy in front (anonymous). I tried to evade his \"making fun\" of me and looked at the back of the class, and unfortunately they were laughing too. I\'m the type who can laugh easily, so i started giggling and chuckling halfway throughout my speech. In the end, i got 6+7+9=22 marks, which would have been 8+8+9=24 if i did not laugh... feel so betrayed and \"sabo-ed\".   Sabbaticals  This morning, we thought the registration for the sab. was already opened, so the whole class was fighting over the com. we all tot we had already registered, when we found that the registration site was only a demo. felt so pissed. at around 1pm, the registration really opened. by the time i got the chance to register, only the below-par courses are left. ended up joining some chinese thing, and had to pay $55 for it. later found out that my ep3 (or rather CCA) actually urged us to join that particular course.   Teachers... again...  Quah is recently in extra good mood. just today, she gave me (and some few others) a box of Oreo cookies. then, we later played a prank during class on En Hao, the all-time notorious trouble-maker of our class hated by all teachers. En Hao basically loves coming to lessons late. Like on Wednesday, he\'s always late for English lessons as he can\'t be bothered to run from the com lab (be4 eng, we were having com). So today, we secretly scheduled our class in the library, en hao was the only one who doesn\'t know about it. then after com lesson, we all urged en hao to run back quickly for class, and HE REALLY DID, and he ran back to the classroom instead! quah was like laughing aloud with us throughout the lesson. of course, later he was sent to the library as well. oh btw, quah is the one who set the lit term test. and, if it\'s ms kala, i bet en hao won\'t be strolling back to class so frequently.  teocm is not in a good mood nowadays. she\'s been quite \"un-nice\" this week. just today, she instructed 2 guys in our class (those who did not finish hw) to go with her to her afternoon lessons, in 3K. However, she\'s still quite fun in class. I really hope she will teach us again next year.   '),(19646,2787343,'02,August,2004','2004-08-02 00:00:00','    Seniors are beginning to know my blog too. At first, i only allowed close friends and relatvies to know about it. Now, it\'s sorta all screwed up. have to be really careful about what i blog now, or some adverse side-effects may happen. yi kun can be really a good specimen.    National Day  Singapore\'s National Day (August 9th) is approaching. Quite unsignificant to me, but i\'m looking forward to the holiday. This time, including weekends, it\'s a total of 4 days. So, sit back, relax and enjoy the holidays. Malaysian scholars can use this opportunity to go back to Malaysia. IF only we are rich enough, we can use this opportunity to fly back to LA too. We can spend about 48 hrs to get the USA summer feeling. Since Singaporeans are busy singing national songs (such as Count On Me Singapore) this also reminds me of a song, Back In The USA by Chuck Berry. Liked the song quite a lot.   Tests  Once again, tests are starting. So it\'s time to study again. Hate tests, and we still have final exams at the end of the year. but luckily, next week we are taking Year 1 Geography, History and English Literature for the last time. With the MRP, there are no final exams for these subjects. On the whole is quite \"shuang\", dunnit to study, haha. But I\'m worried about History and Geography. History we haven\'t really done revision, we still can\'t even finish the textbook. For geography, our past class tests have really been too easy and \"low-standard\", when i took a peek at the geog. test papers of other consortiums, got a big shock. This type of questions really need the application, not only memorizing.   '),(19647,2787343,'02,August,2004','2004-08-02 00:00:00','    Not in the mood to blog lately. if you would like to find out why, look at the index-es below. (all are out of 10, the bigger the number the more of that feeling)  anger: 9.5 busi-ness: 8 \"can\'t be bothered\" rating: 8.5 frustration that may lead to violence: 9 hatred currently in me: 8 bothered-ness: 8.5 slacking level: 3 [can\'t slack :\'(]  had ice-cream with maple syrup just now, able to cool off a bit... oh well...   '),(19648,3875375,'13,July,2004','2004-07-13 00:00:00','   Sorry for lovin you the way that i do. Sorry for caring so much. Sorry for givin u my trust. Sorry for wantin u more then ever. Sorry for w anting to have you forever. Sorry for all i have done. And now i know that love isnt fun. You think u know how a person feels. Then you figure out your wrong. And it hurts a ton. Thats why i figuered out love isnt fun.   '),(19649,3875375,'12,July,2004','2004-07-12 00:00:00','   I sit here in bed late at night with thoughts of you I wonder when I\'ll be able to look in your eyes To look deep inside you and see your soul The hours pass, then days and weeks go by Are your thoughts of me as fond as mine of you  I sit here in bed and smile with my thoughts of you Everyday my aspirations center on finding the one The beautiful love who brings out the best in me The girl who enlightens me with her existence I wonder if you might be the end to my search  I still sit here in bed late at night with thoughts of you Has fate brought us together or kept us apart Because I\'m sitting here writing these words for you to read I\'m taking a chance that I feel I must take For two people in love is a goal worthy of kings  Maybe I\'ll lay here, sleep, and dream thoughts of you  -troy this is for a special someone   '),(19650,3875375,'11,July,2004','2004-07-11 00:00:00','   Suddenly you turn around and you\'re all alone No one\'s as close to you as they used to be You never thought it could happen to you Until you realized it already had You never thought that you\'d lose all the people so dear to you You never thought they\'d all is gone After everything that you\'d done with them You thought that they didn\'t care Now that it\'s too late you realize they did You wish that you had been nicer Every day that you spent with them Now that theory not with you any more You wish you had done a lot of things differently   '),(19651,3875375,'11,July,2004','2004-07-11 00:00:00','   You never miss a good thing till it leaves you. I was heart broken when you left my world. man you dont know how much i miss you gurl. I need you now more then ever. I wish you could be mine forever. We used to break up and it was over just like that. Now its harder then ever to say goodbye  considering im closer to you now then ever before. Im sorry for hurting you. Your so amazing you suprise me all the time. Just promise me this isnt gunna be the last time your gunna be mine..   -troy   '),(19652,3875375,'10,July,2004','2004-07-10 00:00:00','   A hard task to do   turning away and never looking back losing all that build up trust   letting go of that love or lust debating to move on or not   regretting all the times you fought remembering all the good and bad times that you\'ve had   wondering how or when your love came to a end. its missing you dearly   Its missing you were near me  Its longing to hold you tight   Its wanting to tell you itd going to be allright Its loving you with all my might  its hoping to get a kiss goodnight letting go is to hard to do    since im 2 attached to you       -troy   '),(19653,3875375,'10,July,2004','2004-07-10 00:00:00','   This is driving me crazy I cant belive i lost you baby Thought we were gunna last forever But now i know you need me i\'m sorry that im never there But I just want you to know that i care. No matter what you say or do I\'ll always love you You mean the world to me And after all you are my gurl I cant picture going on without you. cause truthfully im lost if i dont got you Just never leave my side And ill try to make things allright between us.   '),(19654,3875375,'10,July,2004','2004-07-10 00:00:00','   I see you most different, than some people do.  I stare at the stars, then I look at you.  I see a twinkle, in you beautiful eyes. I sit there staring, I\'m hypnotized.  I then start dreaming, of what could be.  I see true love, I see you and me.  Our love is strong, can\'t you see.  I was meant for you, you were meant for me.  -troy   '),(19655,3875375,'10,July,2004','2004-07-10 00:00:00','   Today sucks..July 11th 2004 will always be the day that i remember i lost my first nephew..i was looking forward to see this lil guy grow up..wanted to be there for him like my daddy should of done. But i didnt even get a chance to hold him in my arms for now hes gone. Didnt even get to look at his delicate hands or his delicate toes. I was looking forward to being a big brother to this little lad. Things happen for a reason..or so i was once told by a wise woman. But this how could a live of a baby be ok to just suddenly go? I dont think its right..and i cant handle this anymore..I cant lose another.  I just want to curl up in my bed hide all day never show my face durring the day.. god you better be taking good care of.. Jack Daniel Anderson..my buddy..               -troy   '),(19656,3875375,'10,July,2004','2004-07-10 00:00:00','   Remember me when I am gone away, Gone far away into the silent land; When you can no more hold me by the hand, Nor I have turn to go, yet turning to stay. Remember me when no day by day You tell me of our future that you planned: Only remember me; you understand It will be late to counsel then or pray. Yes, if you should forget me for a while And afterwards remember, do not grieves: For if the darkness and corruption leave A vestige of the thoughts that once I had, Better by far you should forget and smile Than that you should remember and be sad.   -troy   '),(19657,3875375,'10,July,2004','2004-07-10 00:00:00','   I\'m just an outsider looking in Seeing pass the obvious Knowing what could begin Seeking the love and closeness Looking for the right friend I\'m just an outsider looking in Seeking safety and comfort With someone who won\'t pretend I\'m just an outsider looking in Hoping for happiness With a long lasting friend Looking for a relationship Without any possible end Searching for my soul mate Who will protect from the hate I\'m just an outsider looking in Seeing what the future holds With that long lasting friend. I\'m just an outsider hoping to come in   '),(19658,3875375,'10,July,2004','2004-07-10 00:00:00','   This pain is staying with me I can\'t get rid of it I can\'t get it to leave I can\'t put up with all of this All this pain is blinding me, i cannot see I\'ve lost my way on the path of life I\'m not sure if this is right I have to think about what you put me through I really want to spend time with you All these thoughts are making me lose sight On all the things we used to do We used to laugh and hang out all the time But you are no longer mine I want to tell you how I feel But i don\'t know how you would deal With all the things I\'d be telling you But now we\'re here this is the end I\'m so glad that we\'re still friends   '),(19659,3875375,'10,July,2004','2004-07-10 00:00:00','   Cold I\'M cold So very, very cold. Heart and soul frozen,  and afraid. What once was warm,  lie after lie has turned cold. What once was, was more than enough to thaw any man\'s heart. I await the sun. The warmth. The feeling still lives, in My heart and soul. I am afraid, I long for warmth, but  I am afraid if the chill comes again. I will not survive.   '),(19660,3875375,'10,July,2004','2004-07-10 00:00:00','   To say you are all that counts in this world of all the towns, in such a city that is so pretty, would not be saying nearly enough. That which seems so far, yet you\'re so near, to lose this love I have no fear. There is nowhere, not near or far,  that can make a brighter star,  You shine above all others, that sets you apart. I look in your eyes and they melt my heart. There is no light that shines stronger,  with all that can last longer with all the need, want, and hunger that your beautiful love brings. Star bright your eyes with magic they shine, with in-depth love I want for mine, deep, so deep, hidden from day,  yet the love you\'ve shone me so divine. Beautiful you are inside and out,  your love is all I need; of that there is no doubt.   -troy   '),(19661,3875375,'10,July,2004','2004-07-10 00:00:00','   Can I Borrow Some Hope I \'m having trouble looking ahead because I can\'t stop seeing the past It\'s hard to not fall hanging by a thread How am I supposed to stand fast  Can I borrow some of your hope I\'ve seemed to lost all of mine Can I have some of your trust  Offer me Your presence, your heart and your love they shine  Recovery seems so far off in the distance The road of healing is a long and lonely one Through the hard times I feel Your existence  And the promises you made shines like a midnight sun  Can I borrow some of your hope I\'ve seemed to lost all of mine Can I have some of your trust  Offer me Your presence, your heart and your love they shine  Pain and hurt are no longer my only friends You lent me your hope and I felt healing and freedom And now through me the victory extends  Your always there strong for me whatever may come   '),(19662,3875375,'10,July,2004','2004-07-10 00:00:00','   You are my true love the only one I\'m thinking of  and I just lay down and pray that we will be together someday  I keep thinking about you all the time and I can\'t get you out of my mind  they said fantasies don\'t come true I believed them, until I met you  you\'re the only one I\'m thinking of  so you must be my true love    -troy   '),(19663,3875375,'10,July,2004','2004-07-10 00:00:00','   I\'m glad that we were once togethor. I honestly i thought you were gunna be mine forever. But i was  mistaken for what i said was wrong. Every time i try to figure out what happend to our love and how it ended so suddenly. And all i have come up with is the one to blame is me. I\'ll move on sometime not real quick but slowly keeping every little memory in my head till the when i wake till the time i go to bed. Help me forgive my self for the mess i put us in. Cause hurting you is a real big sin.  NOT FINISHED YET   '),(19664,3875375,'09,July,2004','2004-07-09 00:00:00','   When I met you I was Hypnotized because I saw the beauty inside  I didn\'t know what to do I just knew I had to be with you  so I wrote you a letter asking if we could ever be together  you said ok I\'ll give it a try and I said I\'ll be there to dry your eyes  and all these years you are still in my heart it would just tear me up if we were apart  so let it always be me and you I promise I\'ll always stay true  by troy   '),(19665,3875375,'09,July,2004','2004-07-09 00:00:00','   Im sorry how this all went down. I didnt expect this to happen no way no how. But since you\'ve left ive had some time to figure some stuff out. I now realized that what we had ended awhile bak. I just was so attached i couldnt think about my life without you having my back. But things are different now i finally figured out was missing in my life and thats just no pain no agony. WE sure have been through a lot over these couple of months. I apoligize for the way im doing this and they this is going down. You dont deserve this u need a betta man around. Sorry bout all of this hope you can forgive me adn we can go bak to just bein friends..                  sorry                      troy   '),(19666,3875375,'08,July,2004','2004-07-08 00:00:00','   You are the nicest caring person i have ever met not to mention the FUNNIEST! I have known you for the past 5 years of my life. You have never left my side. When i was sad you somehow made me laugh. When i felt like nobody in the world cared you always showld me you cared. When i lost people with whom i loved you out of all the people who i knew always came and never left my side if it was in the early morning again right there anytime i needed you you always were right by my side. I love you more than any other person and i value our friendship more than anything in the world. WHen ever you need me call me at any time and i will be there adn i will visit you at any given moment and go to COLOMBIA and learn how to do the salsa dance thing and speak spanish!!! TE AMO MUCHO DIANA!! yes im learning! love you and forever your friend                          troy   '),(19667,3875375,'08,July,2004','2004-07-08 00:00:00','   WEll this is how my day went so far i woke up took a shower got dressed when to see how everything was was it still the same or did it all change. My thoughts were correct when all this mess was scatterd here and there. No longer was this room mine. I cant seem to figure out waht to do with my life but to sit and wait till this pain someday ends. I am the person who tries to be nice to evveryone polite and kind and helpful and over all i want to be a nice guy to everyone but for some reason pay back for being so nice is me getting hurt. I lose my mom how hard is that SUPER hard im still not fully over the fact that shes gone and never will be. I lose my family im alone im forced to lose the gurl that i like. Forced to say goodbye and hope that shes gunna be allright. life so complicated the hardships just keep comming faster then ever i sure hope this will someday end and i will be happy once again.                              LATA -troy   '),(19668,3875375,'08,July,2004','2004-07-08 00:00:00','   I see a reflection in the clear Of a soul that matches the one of myself But has trouble expressing the minds\' near And keeps thoughts deep in stealth Of how she feels that words can\'t explain Which makes my mind wander and often complain I care so much and I just want to help But the only way I can is to let her help herself   '),(19669,3875375,'08,July,2004','2004-07-08 00:00:00','   \"Goodbye\" It\'s time to say goodbye and please don\'t cry  I don\'t mean to cause you pain  but together we no longer remain  I\'m sorry to brake your heart  we are ment to be apart  this is so hard to say I really wanted you to stay  but still keep me in your mind like I keep you all the time  SHHHH, no more words to say but goodbye i\'ll see you again another day   '),(19670,3875375,'08,July,2004','2004-07-08 00:00:00','   The Best Grandfather There was a man I use to know From my younger days. He loved to laugh, joke and tease  In his special ways. Although he\'s gone I\'ll not forget The man he use to be  Because he was the best grandfather  That there could ever be.   I miss you love you and will never forget you. You helped me be the man i am today and for that i thank you and will remember you for ever..   '),(19671,3875375,'08,July,2004','2004-07-08 00:00:00','   Leaves and Lives A lonely leaf makes its way across the concrete driveway. Leftover from fall, it was not carried by the winter winds. Now it makes a journey with the fresh, new spring. Who knows where it will be led to with these winds: If it will be dumped somewhere to disintegrate, or If it will continue to be blown until The Day.  We are each a leaf during our life. We bud out, and eventually break away from the foundation Which formed us. From that point on, we are our own. We can fall to the ground and decay, or We can go with the wind and never stop.   '),(19672,3875375,'08,July,2004','2004-07-08 00:00:00','   Infatable Truth Searing black flames licking at the toes of my past. Suckling tasting teasing, it infects my most intimate thoughts. My eyes fill with dry painful tears for the lost unfortunate believers. Memories gently beat me into submission as I curse reality. I deny the facts and preach the only significant truth. But does it fall on deaf uncaring ears? Take off your mask. Let me see your naked inner child. Could you survive without your lies? Twist and contort the pain into a comfortable sedated numb with the doctors love. We must all realize and endure the rancid stench of mortal truths. As my pen confides in the paper I still refuse the terms of reality. Revealing my raw emotions and weaknesses to myself the mirror becomes a window of turbulent embrace. Can the injustices be exposed in time or are we destined to live in societies forgotten cage. Rotting in our own innocence. A killer walks among us while you cheer the victory of the mislead masses.  Where do you fit in?   '),(19673,3875375,'08,July,2004','2004-07-08 00:00:00','   Return Forgotten chapters which left me confused as a mixed up rubix  cube certain pictures take me back in time since I\'m running out I\'ve abandoned the true path that does not stray  everynight sorrow harasses my heart I came from where I long to return  I stand before the Lord bleeding with sin  I no longer want missery to touch me filled with cowardice but I must dispose of my heart and return on the journey that I was so prepared for  -troy   '),(19674,3875375,'08,July,2004','2004-07-08 00:00:00','   NIGHT AND DAY I was a lost soul Who found you in love, Not by chance we met But by he that\'s up above. I had asked him for guidence And he led you straight to me, He said, \"my lonely child I have brought unto you thee. Take and love her Cherish her in every way, For she shall be your night And you shall be her day. Be in love forever And be of one heart, And what god has joined together Let no one take apart...\"   '),(19675,3875375,'08,July,2004','2004-07-08 00:00:00','   This is about missing sombody.. I miss you already. The flame which burnt for us has about come  to it\'s end,you can\'t even be a friend. Where did everything go wrong? Hours spent together,are now only minutes. Soon,will be only seconds. Time has drifted,with it, so have we.  The pain of separation to be. What\'s going to hurt the most,  the thought of you in someone elses arms. The intimacy that was once ours, now belongs to another. I never gave you the things that you wanted or needed. I dragged you through a fantasy dream,only to fail in that too.  Things won\'t be the same, but I hope our paths cross again, I do.  Maybe our friendship will last longer, because, I miss you already.   '),(19676,3875375,'08,July,2004','2004-07-08 00:00:00','   Candle Light The illumination of life that dwells within. Slowly dims. The shell of protection withers. Outer beauty that once attracted. Now repels. The gifts of past memories, slowly dulls. A creative imagination that once came with grace, becomes exert. The inure signs of ripeness, pries it self into existence. The will to live slowly subsides, until the light that once shined so bright, finally flickers out.  This was my first thing ive ever written yes indeed lata   '),(19677,3875375,'08,July,2004','2004-07-08 00:00:00','   I Give I give you my heart, I give you my soul, I give you my love, I am here in whole. I give you my time, I give you my thoughts, I give you my friendship, The best of all sorts. I give you my caring, I give you my lust, I give you my strength, In you I trust. I give you my support, I give you my laughter, I give you my enchantment, Yours forever after.   '),(19678,3875375,'08,July,2004','2004-07-08 00:00:00','   Thoughtless nights tears I shed Concentration of deep thoughts While a blanket covers my head Countless questions are calling But no answers to grab To stop me from falling I think back to a day not long ago A star had fallen in my hand And there was no letting go She was more beautiful than the universe  The power of affection trailed with a curse So many times I think back If I could have changed a minute in time Or switched from this to that But as everyone must find out Love is but a word  That must have feeling behind each mouth I tell this out of courtesy just because For everyone will go through this For this is My First Love  This is a poem of wut i thought bout my first love ya fun huh sweet lata   '),(19679,3875375,'08,July,2004','2004-07-08 00:00:00','   Hey everybody! Thanks a lot for looking here.. This is my life.. Im 6\'1 blonde hair blue eyes tan im 14 years old and my name is troy haha! I have 4 sibblings 2 older brothers and 1 older sister then i have 2 half brothers! I love to box and i love to play my guitar. Here are some other things about myself that you may not of known.. I was born in Kansas city on october 3rd 1989 at 3:00am. When i was 3 years old i moved to omaha nebraska living with my late mother victoria anderson who passed away a 2 months ago..and my brother derek and sister victoria and brother josh. I learned how to play guitar whne i was 7 years old and have been playing since. I took up boxing when i was 10 years old and have one matches all over the u.s My future goals are to be a Doctor and for a part time job record my songs.  STUFF I LIKE TO DO.... I love to play football  i love basketball and baseball is my ultimate fav sport. I play pitcher and 1st base. I can pitch 80 miles a hour and im gunna be playing on a very good team in new york city for highschool hopefully i get to the pros someday!  GURLS!!!!!!!!!!!!!!!!!!! ha i love gurls! there awesome! I like a gurl who is shy yet can be very outgoing and surpising and i have to have a gurl that can make me laugh if you dont make me laugh then your out! j/p What i look for most in a gurl is all the ways that they are unique and make her special.. Things i dont like in a gurl is when they assume to much and they talk back all the time and judge me and if you have mood swings all the time thats not cool i dont like insults..no jokes here. Well thats a bit about my life thanks for reading it lata -troy   '),(19680,3875375,'02,August,2004','2004-08-02 00:00:00','   I saw you again at our place holding some other guys hand now i know that all the rumors are true for what i saw in you is no more   lookin into your eyes   knowin wut u say is all a lie   your lips arent familar   as they once were   our love without a doubt has ended im filled with the taste of revenage its obvious you wanted out but i hate how i figured you out and how i knew all along your dark secret.   lookin into your eyes   and knowing that all youve said is a lie   your lips arent familar as they once were   our love without a doubt has ended            troy   '),(19681,3923707,'14,July,2004','2004-07-14 00:00:00','    Spoken:  Look, why don\'t you play that song you played for me yesterday? Oh ok yeah, that song because of you, cool This is a song called because of you...written because of you...here we go   Verse 1: ...oh yah, oh yah... If ever you wondered if you touched my soul yes you do Since I met you I\'m not the same  You bring life to everything I do Just the way you say hello  With one touch I can\'t let go Never thought I\'d fall in love with you...   Chorus: Because of you, my life has changed, thank you for the love and joy you bring Because of you, I feel no shame, I\'ll tell the world it\'s because of you  Verse 2: Sometimes I get lonely and all I gotta do is think of you You captured something inside of me  You make all of my dreams come true It\'s not enough that you love me for me  You reached inside and touched me eternally I love you best explains how I feel for you...   Chorus:  Because of you, my life has changed, thank you for the love and the joy you bring Because of you, I feel no shame, I\'ll tell the world it\'s because of you   Bridge: The magic in your eyes  True love I can\'t deny When you hold me I just lose control I want you to know that I\'m never letting go You mean so much to me I want the world to see, It\'s because of you   Chorus: Because of you, my life has changed, thank you for the love and the joy you bring Because of you, I feel no shame, I\'ll tell the world it\'s because of you My life has changed thank you for the love and the joy you bring  Because of you, I feel no shame, I\'ll tell the world it\'s because of you...   Spoken:   Because of you   '),(19682,3923707,'14,July,2004','2004-07-14 00:00:00','    oh my.... i feel so guilty today... we had a quiz today... nd during the quiz, i am the known \"Doctor\"... i was forced to be a doctor by my seatmate... and worse of it... i did it during the time when we have visitors from other schools... ei it\'s only between you and me okei... i did it because i was afraid that my seatmate might make things that i don\'t want to happen... he\'ll make \"gulo\" on me.... huhuh... poor li\'l girl.. neweiz... the nice thing is i don\'t copy my seatmates answer, i juz \"doctor\" their answer to make it correct... hehehe.... i\'m really bothered on what i did... oh my.... i don\'t wanna do it again....   '),(19683,3923707,'14,July,2004','2004-07-14 00:00:00','    ie... dis is my first day in blog... i hope to enhance this more so that you can visit it more often... dats it na lang muna... byez!!!!!   '),(19684,3431135,'31,May,2004','2004-05-31 00:00:00','    THE REASON  i\'m not a perfect person, there are many things i wish i didn\'t do  but i continue learning. i never meant to do those things to you  and so i have to say before i go, that i just want you to know   i\'ve found a reason for me, to change who i used to be  a reason to start over new, and the reason is you   i\'m sorry that i hurt, its something i must live with everyday  and all the pain i put you through, i wish that i could take it all away  and be the one who catches your tears, thats why i need you to hear   i\'m not a perfect person, i never meant to do those things to you  and so i hate to say before i go, that i just want you to know   i\'ve found a reason for me, to change who i used to be  a reason to start over new, and the reason is you  i\'ve found a reason to show a side of me you didn\'t know  a reason for all that i do, and that reason is you   '),(19685,3431135,'23,May,2004','2004-05-23 00:00:00','    \"The best thing about loving and being hurt is that you get to know what true love really is. For as gold is tested in fire, and so will love be perfected in pain.\"  \"You attract people by the qualities you display. You keep them by the qualities you possess.\"  \"Dans l\'amour vrai c\'est l\'ame qui embrasse le corps.\" Translation: In true love it is the soul that embraces the body.  \"There is one pain I often feel, which you will never know. It\'s caused by the absence of you.\"                                                               P.S.                                                             sa su2nod nman ung iba...   '),(19686,3431135,'23,May,2004','2004-05-23 00:00:00','    magpopost na lang ako ng mga love quotes...;) ok ba?   '),(19687,3431135,'23,May,2004','2004-05-23 00:00:00','    Hi to all of you out there! I don\'t think i have d reason for making this... but i just don\'t wanna lie to you 100 percently (haha) therefore now i\'m saying dat i\'m kinda inspired and i kinda know what\'s my true objective on involving to such a thing like this and i think it\'s kinda awkward doing this in an awkward time of an awkward day... It\'s just bcoz of my self-made criteria: 50% for fun and 50% for some sort of things that pushes me to finally show my feelings, experiences, views, etc...         Suddenly I think there should be a side of me who can at last reveal my long time (not so long..hehe) suppressed self... I realized dat i\'m already racing near to the next stage of life and that made me think that i should already begin adapting myself into the grown-up ones and avoiding my childish ones... I think this may seem too serious but i realized it\'s not! It\'s just the reality... and nobody could escape from this reality. Instead, we should stand firm and face this \"reality\"...   '),(19688,3431135,'01,June,2004','2004-06-01 00:00:00','    \"The greatest love story that was ever told, could be me and you, we may just never know.\"  \"Real love isn\'t blind, It sees, sees everything but loves anyways.\"  \"Sometimes I wonder why I fell in love with you in the first place, and continue to love you after all the pain. And then you smile at me, and I realize why; without you, I can\'t breath.\"   '),(19689,3638066,'28,June,2004','2004-06-28 00:00:00','   Oh man am I exhausted. For the past week, everything was crazy man hehe. On Wednesday, slept over Tahir\'s place with Anthony and we stayed up \'till 3 watching How High. We basically played Super Smash Bros. Melee and Soul Caliber 2 the whole day. Thursday, we went to Six Flags from 9AM to 10PM. We met up with the other people from Grant and it was fun. Seeing everybody, well not everybody, was fun but kind of weird because they have changed dramatically from the last time I saw them. Friday was a chill day and I\'m happy I had that because if I didn\'t, I would be dead right now. Saturday, at 9AM, I had to go to Birmingham for the Cancer Walk thing. I feel bad that my mom had to donate 80 bucks just so I can get the 100 dollars for donations.. I hope I can pay that back to her somehow.. So then Saturday was crazy man. We had water fights, constantly really, walking, talking, beachball, volleyball, dancing on Jessica and Raul\'s part haha! Jeez how can they stand listening to (Jook Gal Remix and Culo!) for basically the whole night?! It was fun though.. Jessica is so pretty =)     SUNDAY   - We had kareoke and I was  so  ready to sing (Usher - U Got it Bad) OH MAN FUKIN MC man.. Right when I was supposed to go up, we were out of time. Bastards. Then we finally slept!!! OMG FINALLY SLEPT. It was around 2. BUT! Around 4:30, I woke back up because of people talking outside. And I looked around.. and nobody was around me. I was like WTF! Went back outside and found out everybody didn\'t sleep. I had 2:30 of sleep woot! Hehe. We went walking and talking and stuff till 4AM I think where me, Carli, and Jessika slept on the tables. Like when you sleep in class? We were sleeping like that since JESSICA had to go in her humongous tent with that guy she barely met lol. OMG I was so ready to sleep. Then I slept for 2 hours on the table then went into the tent with Raul and Stephanie. I wonder how we fit 4 people in that tiny tent!! Me, Raul, Jessika, and Stephanie.. oh man that was crazy  LOL!  Woke up at about 8:45 and we went to the ending ceremony. Waste of time but we got Icee\'s so that was okay. We took our last lap and we took pictures. I took a picture with the humongous panda! LOL! Retarded panda lmao. Then we all left and Shirley drove me home. Thank you Shirley!! Took a shower.. then rested for a bit. At 12:30, Yulan picked me up and we went to KFC and went to Northridge. OMG no time to rest! We spent a lot of time at the Sam Goody playing video games and the Apple Store going online for free LOL. I bought 2 pretzels form Auntie Anne\'s and that was  goooooooooooooooooood..  I want more! Then went over to Yulan\'s house and we watch the Kenshin movie and another Jackie Chan movie. And Orianna taught me how to make a turtle! It was cool hehe. Then went back home and went online for a bit and downloaded CULO!  LOL!  Then sleep comforted me.. And now concludes my humongous blog. I left out so many details. I\'ll show pictures later.. peace  '),(19690,3638066,'25,June,2004','2004-06-25 00:00:00','   So I\'ve succumbed to Blogspot.com. Oh how horrible --;; I was trying to edit this layout so I don\'t need this banner up there. I got so far.. that I was tired of doing it. So laziness took over. So the days.. the days.. the days have passed by quickly and slowly. Yesterday, me and my friends from Grant went to Six Flags. T\'was a fun day and night. Andrea didn\'t buy me funnel cake --;; Damn you! I had no more money.. Enough money to be in debt with Anthony lol. Thanks a lot for the 2 bucks Anthony, you kept someone from starving to death. WoOT! Hehe. So anyways, we\'re seniors everybody!! (Those of my grade lol) And we have one more year to go. I went to graduation because of Knights and Ladies. Oh yeah, and tomorrow I have a Cancer Walk to go to because of Knights & Ladies too. Great huh? I\'ve only gathered up 50 dollars when I\'m supposed to have at least 100. Just great everybody! I love it. Oh yeah, you can\'t see my sarcasm there lol --; So the whole day tomorrow.. from 9am to 9pm.. I\'ll be at school. Walking and running and playing around with my friends from Knights and Ladies. And for that 100 dollars? I  will  get it. I swear to you! I need it desperately. Anybody want to donate money for a desperate person trying to donate to the Cancer American Society?  LOL!  Ahh.. how I miss blogging. And in another 3 weeks, I\'ll get tired of it. Great lifespan isn\'t it? Okay, time to plan out how I\'m gonna get that extra 50 bucks.. Out.  '),(19691,3600967,'21,June,2004','2004-06-21 00:00:00','   Sunday was Father\'s Day and we geared up for a fantabulous day outside.  It was perfect weather, 80 degrees, bright blue sky and we had a terrific day planned.  Just before we left I read my mother\'s Father\'s Day sermon which was eloquent and sweet and moved me to tears (check it out at:http://reverendpat.blogspot.com ).  She wrote of her Uncle Ernie and revealed more about him to her congregation than she\'s ever shared with me personally.  My mother\'s family all died before my brother and I ever came on the scene (except for Ernie who we only met once) and they all carry this mythical mystique.  I only know them in stories that are pieced together that don\'t form cohesive wholes.  And since I only know them through my mother\'s eyes I always wonder what I would have seen or which of their personalities I would have encountered.  For as with any mysterious character, they all have glamorous qualities, scary dark ones, and lots of shadows.  Uncle Ernie used to send us cards on our birthdays.  Cards with crisp new bills in them.  Usually a dollar but I remember getting $5.  He never wrote notes and there was never news.  Just the new money and \"Love, Uncle Ernie\".  It was like magic for me, getting those cards on my birthday.  Like getting a gift from the beyond almost.  He was such a ghost to me in many ways.  When I met him I was 13 and there\'s really no way for me to describe what it was like.  I suppose I had all these dreams and fantasies in my heart about what he meant to me and how much I thought he was this magical leprechaun kind of creature.  And since I was extremely shy and extraordinarily self-conscious at 13, I had no way to really express any of that to him.  So I mumbled and looked at my feet and ran off to play as soon as I could and felt entirely trapped by my age and awkwardness.  And it was strange to meet him face to face, like seeing a unicorn almost.  In a way, it wasn\'t supposed to happen.  I can only hope he enjoyed it.  I know I did but I\'ll always feel as though I wasn\'t quite good enough.  And that he would see I didn\'t deserve all those nice cards he sent.  My Mom\'s sermon is about Uncle Ernie being an extraordinarily kind man.  And let me share with you this quote:   \"Spirituality is often about doing the tiny work of God, little acts, small responses to God\'s presence in our lives.  Spiritual people are about tiny things, which is the fruit of their spirituality.  The spiritual life is not a life of success; it is a life of faithfulness.  God does big things once in a while, but there is no question that the primary work of God in the world is tiny.  God knew we would naturally be dazzled by big.  That\'s why Jesus told the parables of the lost sheep, the lost coin, the lost son, the mustard seed.  The small, seemingly insignificant acts we do in the name of Jesus Christ don\'t seem very impressive to us, but they can make a big difference in people\'s lives.  Acts of habitual kindness can generate stability, unity and wholeness in a community.\"  And to my mother, Uncle Ernie, her father figure, embodied this spirituality.  This got me thinking of course.  And since I pride myself on being the kind of person who is kind to strangers I start to feel pretty darn good about myself and about how indeed I am a spiritual person living out my faith in small measure.  Yes sirree I\'m just small enough to be relevant to God.  So Mikie has a soccer game on Father\'s Day and since he loves to play soccer Collin and I happily join him.  Luckily he was playing at a field that has a playground attached.  So Collin and I set up camp at the playground while Mikie plays soccer.  When I say we happily join him....let me clarify that this is a choice I make for Father\'s Day.  Honestly the idea of keeping Collin entertained at the park for 2 hours during his nap time is not that exciting of a prospect.  And had it been an ordinary Sunday I would have sent Mikie off to play while Collin napped at home.  But since it was Father\'s Day, and luckily a gorgeous day, we joined Mikie.  And quite frankly, in spite of it all, I\'m glad we did and I would do it again.  But it\'s important to note that I am just a little frazzled and anxious at how it\'s all going to go.  So no sooner than we arrive and Collin heads up the stairs to the \"Slide! Slide! Slide!\" do we meet Roman.  Roman is 7, well, he\'ll be 7 in just 3 months.  3 months is his math, he\'ll actually be 7 in 2 months, August 31 to be precise.  Roman is a talker.  At first we merely exchange pleasantries, he actually says, \"Good afternoon\" which is charming.  We go over all our names.  He has me guess both his middle name and his last name.  He then wants to guess mine and Collin\'s names.  Then he wants me to guess his nickname, then he wants to guess our nicknames.  Then he just wants to chat.  Well, it\'s a nice day and he seems a little like a lost soul (they always know where to find me) so we strike up a chat.  Except also remember I have a responsibility to Collin who is busy on new playground equipment so I\'m sort of torn between the two of them.  I\'m getting a glimpse at what it\'s like to have two kids and it scares the hell out of me.   And then Roman just keeps talking.  And talking and talking and talking.  I\'m wondering to myself where this kid\'s parents are and I start looking around.  He eventually tells me he\'s there with his grandma who is busily chatting away with one of her friends on a park bench.  And I am torn between the desire to be God\'s servant and the desire to tell this kid to shut up so I can enjoy my own child.  I take deep breaths, focus in on Collin and remember a challenge from my mother\'s sermon, \"Listen to the neighbor kid\'s troubles instead of saying, \'Get lost.\'...Usually giving our life to Christ isn\'t glorious. It\'s done in all those little acts of love, 25 cents at at time.\"  I resolve myself to be good and ask more probing questions.  I come to discover that Roman has 1 sister and 2 brothers, all of whom live in Ukraine.    \"So who do you live with here?\" I ask.   \"THAT is not a good question...\" he replies with all the 7 year old seriousness he can muster.  So I learn even more about Roman while Collin is looking around wondering where his mother is and why she\'s talking to this kid instead of sliding.  Roman\'s had it kind of rough.  He\'s \"trying out\" his third Dad.  The first one was greedy and the second one was mean and bad.  This one is OK but he\'s the laziest according to Roman.  There are stories of abuse and stories of neglect.  It\'s obvious the kid carries a painful load.  And to help him cope I guess he talks.  And talks and talks.  And I spend a good deal of time really investing in Roman, encouraging him, asking for details, telling him he\'s great and smart and a very mature kid.  He apologizes for having to say a bad word.  But he just has to use the word \"stupid\".  Again he apologizes after he says it.  But it  is  just stupid that the community center wants to charge $72 for only 15 basketball games when swimming is free.  I told him it was OK to use stupid in that context and that I agreed with him.  I also told him I appreciated that he\'s very careful about that bad word.  And then I had several ooey gooey moments where I looked lovingly out at the soccer field and at my sweet Collin and said Amens and Hallelujahs that my life is so much more simple than poor dear Roman\'s.  I started to feel as though Roman was there for me, as a little kick start of gratitude.  Stop your whining and anxiousness and give thanks that Collin only has one Dad who is neither greedy, mean, bad or lazy.  And if Collin wants to play basketball, stupid or not, we\'d easily pay the money.  So I had nothing but misty eyes for Roman for a while.  And Collin was enjoying his company.  And really, it was so nice out.  It was turning out to be quite a good day.  And then Roman just kept talking and Collin was getting more demanding.  Collin needed to eat but he was all dirty so trying to clean him and offer him food that he would eat and all that jazz while Roman was busy chatting away was distracting and nerve producing.  It was hard to interrupt Roman and speak to Collin, I mean, I wanted to be present to this poor kid and make him feel special.  But at the expense of my own sanity and my own kid?  It started to get a little tricky.  Where was this grandmother anyway?  And there were plenty of other people at the park, couldn\'t someone else entertain Roman for a while?  Once Collin downed his sand covered banana and his energy was restored for more exploring, we headed back to the jungle gym and Roman began to recount to me all the details of the SERIES of books his teacher read to them this year.  And you know 7 year olds are not the best story tellers.    \"So Annie, she\'s King Arthur\'s sister.  You know King Arthur?  Well, she\'s King Arthur\'s sister.  Well King Arthur\'s sister\'s sister.  And she, um, Annie, the sister, she has a friend Jack.  And Annie, her sister, King Arthur\'s sister\'s sister, she made this house.  This magic treehouse.  And when they want to come back they just get in and it spins and spins and then there they are.  And this one time, Annie, she went to the dinosaur times.\"  And on and on.  My head is now officially aching and I\'m getting angry at this kid for wasting my time and distracting me with this  drivel !  Then I remembered something I read recently from Pamela Anderson in Jane magazine.  She talks about positive energy and surrounding yourself with positive energy.  And paying attention to the people you\'re with, do they give you positive energy or are they \"energy vampires\".  And it hits me like a ton of bricks that Roman is an energy vampire.  I\'m giving all my energy away to this little blood sucker and I\'m so stressed and frustrated about it that I\'m cranky and miserable.  And totally unavailable for my child and husband.  I\'m giving all of myself away to this stranger who is draining me so I\'m now empty for Mike and Collin.  It seems like the only thing to do is to just distance from Roman, to start ignoring him and paying more attention to Collin.  So Collin and I go down the slide together which delights him so we do it \"again\" and \"again\" and \"again\".  Roman, sensing my separation, now asks for me to play tennis with him.  I explain that I need to keep a closer eye on Collin and he still persists, \"we can play in the grass\" he says pleadingly.  Those innocent eyes looking so hopeful.  I apologize and keep sliding with Collin.  The guilt is nagging at me.  I\'m having internal wars about whose religion I\'m going to follow my mother\'s or Pamela Anderson\'s!  The choice seems strangely ironic to me as they both seem utterly opposite and utterly attractive at the same time.    Where the hell is his grandmother?  Occasionally she comes over and speaks to him in  hushed, not so friendly Ukrainian.  But for the most part, she has left him to me.  I start to wonder if she always does this.  He did say he\'s at this park every day.  He explains to me that he forgot the popsicles and that he\'s really upset because they are going to be here all day and it\'s so hot out.  This is when I realize, Roman is not going anywhere soon.  I\'m still dancing between ignoring him and playing with him when he starts to ask me to reach a tree branch for him so he can climb the tree.  I tell him I can\'t and suggest he ask his grandma, to do it.  He won\'t do that.  I realize of course he wants me to do something for him that he shouldn\'t do so I say I won\'t and sorry and all that jazz.  There is a tornado of guilt and obligation and woe and empathy and frustration and hatred inside me.  I keep trying to breathe deeply and \"appreciate\" the blue sky and the nice shade of the park but it\'s no use.  My anxious insides are wearing me out.  What would Uncle Ernie do I wonder?  I know the answer of course and I feel I\'ve failed him once again.  The whistle finally blows to stop the game and I scoop Collin up, announce that we\'re leaving, say a quick and kind good bye to Roman and march off to the field for some relief in Mikie.  And wouldn\'t you know, as we\'re leaving the field heading for the car, who is heading home from the park?  Roman with Grandma in tow.  He wants to know which way we are walking, maybe we could join them.  \"We\'re going in the car\" I say, \"we live far from here.\"  \"Oh.\" he says with obvious disappointment.  And while I said pleasant good byes and was as nice as I was able, I can\'t help but feel I let him down.  People are always letting Roman down I guess.  In the end I didn\'t really choose either religion.  Just my often practiced and ever faithful tension and anxiety.  And the hope that I\'ll be out of this situation as soon as possible.  '),(19692,3600967,'17,June,2004','2004-06-17 00:00:00','   We had the pleasure this week of experiencing our neighborhood pool.  We are fortunate enough to live in the Chicago area where people take summer seriously.  We\'ve been indoors all winter and we get outdoors all summer.  The city is bursting with fun and exciting things to do.  Our \"neighborhood pool\" is basically a water park.  Water playground is its official name and man they ain\'t kidding.  This place is awesome!  They have a super huge pool with lots of shallow water, 2 HUGE inner tube water slides for the big kids, and a groovy little play area for the small ones.  We naturally were in the groovy little area for the small ones.  Imagine a huge climbing structure in the middle of a shallow pool with a zillion little fountains and things.  At the top are two different water slides which we can go on with Collin.  At the very top of the structure is one of those giant buckets that fills with water and periodically splashes down on the entire area with tremendous force.  OK, so that\'s what it is.  Now let me tell you that last night we went to the pool after Collin\'s nap, we arrived about 5:00pm.  Let me also tell you that it is mid June.  Being from Texas, mid June is HOT.  Chicago has slightly cooler weather and we were a little nervous to arrive at the pool and find it not quite warm out, but not quite cold either.  We try to divert Collin\'s attention with hanging out in the warmish shallow end of the giant beach like pool.  Neither of us was really into getting truly wet.  But he wanted to head for the slides so we followed him over there.  Let me tell you, that water is COLD.  Like teeth chattering cold. This did not deter our son however and he started for the slides right away.  In order to do that he of course had to climb up through all the cold spraying water.  And since he\'s 20 months old we decided parental supervision was in order.  OK, that\'s the setup, here\'s the pitch.  You know how when you\'re faced with something cold splashing on you, you brace yourself, duck for cover and do whatever you can to cover yourself from the water?  Well, my instincts were right on the money and I did just that.  Except, what I used to cover myself from all this splashing was Collin!  I didn\'t really realize it until we got to the top of the stairs and he is sputtering and shaking and trying to get the water off of his face.  It was one of those brilliant parenting moments let me tell you.  Fortunately, it didn\'t seem to bother him much and I made adjustments.  We went down the slide \"again\" and \"again\".  Poor Collin was shaking and his little teeth were chattering but his smile never faded.  And Mikie and I freezing and shivering ourselves had one of those united front parenting moments I just love.  I turned to him and said, \"You know this is one of those times when you know just how much you love your kid.\"  His chipper response, \"That\'s right, we can\'t ruin his fun.\"  And I just smiled at this amazing man, so appreciative of the tremendous effort he and I were happily making for our boy.  '),(19693,3600967,'10,June,2004','2004-06-10 00:00:00','   urlLink    My son with his father\'s toy...&nbsp; urlLink  '),(19694,3600967,'10,June,2004','2004-06-10 00:00:00','   So my husband is a bit of a Star Wars freak.  Not the dress up as a character and go to conventions kind (although he has been to a convention or two) but worse than that - he is the collecting kind.  Oh yes, you read that right.  He collects Star Wars stuff.  He has a room, a Star Wars room, of all things.  And it\'s filled with toys and books and comics and posters and various other Star Wars merchandise.  An entire wall filled with toys in pristine condition from all the new movies.  Oh sure he has stuff from the olden days, stuff his parents weren\'t smart enough to throw away (he would say they were smart enough NOT to throw it away), but mostly he has the new stuff.    I\'m not quite sure where this collecting addiction stems from but he\'s got it bad.  And of course he\'s kind of a snob about it so you can\'t really get him anything.  He\'s got this bizarre and mysterious criteria for what\'s cool that only he knows.  He started collecting when the re-releases happened.  He was fresh out of college and flush with Wall Street cash and before you could say \"May the Force be with you\" his passion for the movies became a passion for collecting.    Now where do I fit in to all of this?  I began to see early on that this childish joy was so pure that I really didn\'t want to be the mean old witch getting in the way of things so I decided to accept it lovingly.  Wasn\'t that nice of me?  I assumed it wouldn\'t last forever.  Well, it\'s gone on way longer than expected and quite frankly he should be lucky the house is big enough and we don\'t need that room.  Of course I have days when I wonder, \"Hey, what about my room?\"  I suggested painting the the laundry room pink and making it all girly just for me but then that depressed me.  I\'d really rather not hang out there thank you very much.    Anyway, my husband has this R2D2 robot that he got from his parents last Christmas.  (Yes, they still purchase toys for their son even though he\'s in his 30\'s.  Trust me, they too think it is weird but they are those wonderful super love kind of parents who would do nothing to stand in the way of their son\'s happiness.  That and really, what the hell else is there to get?)  And now he is using this Star Wars character robot with our son (20 month old Collin).  Well not surprisingly, Collin thinks it is really cool.  And I think they are adorable, the three of them playing together; Mikie giving commands, R2 making these weird electronic noises, and Collin doing his happy dance.    But when Mikie isn\'t at home Collin points at the thing and says, \"R2! R2!\" and looks at me expectantly wanting me to give it commands to make it dance and sing those mechanical \"songs\".  Well I don\'t know how the thing works and to be frank, I don\'t really want to know.  Knowing sends me much farther down the Star Wars path than I am willing to go.  And really now, don\'t you think I\'m far enough already?  (Remember the room people.)    But when that sweet boy starts to get frustrated and throw a terrible two tantrum over it (yes, he is ahead of his age thank you for noticing) I wonder if it\'s worth it just to know a few things.  But then I remember \"diversion\" and Collin and I play with good old fashioned wooden puzzles instead.  And then we\'ll read some books or sing some songs.  I\'ll let Daddy be the one to teach him all about Star Wars.  I\'ll let it be their little bonding thing.  Just look for me in the rosy pink laundry room.  '),(19695,3600967,'14,July,2004','2004-07-14 00:00:00','   We recently enjoyed a visit from the MIL and she requested we go to Ravinia one night.  Since Mikie and I both LOVE Ravinia we were happy to oblige.  Ravinia, for those of you who don\'t know, is one of Chicago\'s most impressive treasures.  Ravinia is a place, a state of mind, an event, a world all into itself.  Basically, from Memorial Day to Labor Day Ravinia, an outdoor music arena, hosts a variety of concerts from jazz to pop to classical.  As it is the summer home of the Chicago Symphony Orchestra, classical is the predominant musical medium.  We decided to attend Saturday night\'s concert.  This season the symphony is featuring the works of Rachmaninoff and we were thrilled to learn that Saturday night, the symphony would play Rhapsody on a Theme by Paganini.  I\'m no classical snob.  I basically like the stuff I\'ve heard before.  I enjoy all types of classical, but the songs I love are the ones that are familiar tunes.  Usually made familiar by commercials and elevator music.    So I was thrilled when I learned that I would actually know some of the music.  Not only that, I do love that Rhapsody on a Theme.  You know the one.  It goes like this:  Da da da da dum Da da da Dum Da da da da dum La da da dum  And on and on.  Typically an evening at Ravinia is one of elegant sophistication and romance.  It\'s a giant picnic under the stars where people bring fancy little tables and place settings, everyone sets up candles, has wine and cheese and gourmet snacks.  Presentation matters.  The woman at the table next to us brought a hurricane lamp and surrounded it with fresh flowers.  This gives you an idea of the lengths people go to to make this a special occasion.  And it is special, magical even.  There\'s nothing quite like live classical music played by a world class symphony.  But instead of being crammed in an uncomfortable seat wearing shoes that are too tight, you are in your most comfortable (yet stylish) outfit, lying on your back looking at the stars.    OK, so imagine a 21 month old thrown into the mix whose bedtime is right around the start of the show.  Wondering how this is going to go?  So were we....  On the drive up we had lots of conversations about setting expectations.  I made it clear that we shouldn\'t force Collin to do anything.  That if we let him drive the bus in terms of walking, standing, sitting, strolling, whatever it would keep him quiet and happy and we would all enjoy the evening much better.  Quiet is important at Ravinia.  There\'s no talking during the performance.  They even have noise police in khaki pants and Ravinia golf shirts walking around with signs asking people to be quiet.  So we were a little on edge about keeping Collin quiet.  And of course we can feel the tension of the people around us wondering when this toddler was going to explode and ruin their Ravinia experience.  Well he did make a noise.  But he only endeared himself to his fellow music lovers when he did it.  You know how in a big symphonic piece there is a slight pause between movements?  A hush so you can linger in the last moment of the first section before they thrill you with the next?  Well it was during this hush that Collin broke into applause and shouted \"Yay!\"  We were the proud recipients of many warm giggles and stares after that.  For the most part, Collin decided that Ravinia is best enjoyed on foot.  He circled the entire complex several times and we spent countless minutes going over the same steps and ramps and trips to see the train and the potty.  As there were three adults there it was just perfect.  One adult was always up and completely focused on Collin, his ideal.  And then two were left to enjoy each other\'s company back at our base camp of our blanket, chairs, snacks, and bottles of wine.    I don\'t mind admitting that sometimes when I\'m on Collin duty it feels much more like a chore than play.  Especially when it\'s the weekend and I feel stuck doing the job I do all week.  It\'s just like if you had to go into the office on a Saturday.  I can\'t explain why I have that feeling and I wish I didn\'t but I do.  But not at Ravinia.  I\'m not sure if it was the music or the wine (probably a little of both) but we both had the most amazing time.  Collin was running around in his pajamas (we had brief delusions that he would go night night under the stars) looking like the happiest kid on the planet and he easily switched partners when one of us needed a break and never lost his happy grin.    I was off duty when the Rhapsody came on and I thought we made a mistake.  I didn\'t see how this song would eventually become the famous tune I was hoping to hear.  And then gradually I began to hear that it might and I was switching over to \"Collin duty\" when it started.  Well, I was in heaven!  Collin and I danced in circles to the music.  I couldn\'t contain myself.  It was the most exciting and glorious sound I had ever heard.  My feet HAD to dance.  And I was glad to have Collin in my arms both because I love him so much and because he gave me full permission to be as bold of a dancer as possible.  (Had I not been with Collin I might have danced anyway and that would be the thing that inspired Mr. D. to start his own blog....)  When certain music plays that leads me to dance, it\'s as though it is the natural extension of the swells in my heart.  Classical music always makes me yearn for the stage.  And Collin giggled and twirled right along with me.  Still too young to roll his eyes and be embarrassed by his mother.  Rather he was looking at me with pure adoration.  After our big number we headed over to the stairs where we played with Isaac.  He had one of those \"be careful\" Moms and I felt sorry for her.  Collin and I had a ball playing with her child and she was so anxious and self conscious that she couldn\'t really let go and have fun.  \"He just gloms on to people\" she said in an exasperated and apologetic tone.  \"That\'s fine! How cute!\" I yelled back as I chased both boys back up the ramp.  I could have played with them all night but she was getting uncomfortable (to her credit she was pregnant and I forgive a pregnant lady for anything) and Collin was ready to move to the next big thing.  The train!  Collin has a passion these days for trains.  And the Metra, Chicago\'s commuter railway has a stop right at the Ravinia gates.  So wouldn\'t you know it, just after we left Isaac and his Mom and we were near the front gate I saw that the Ravinia staff was preparing for the arrival of the train.  So Collin and I parked ourselves behind a fence and waited for the train to arrive.  \"Here it comes, here it comes!\" he kept saying.  \"Oooohhhh, a train! A train!\" he shouted over and over as it pulled into the station, looking at me as if he couldn\'t believe his luck.  \"Bye bye train!  Bye bye train!\" he said and waved as it left.  With a twinkle in his eye he shared with me a conspiratorial smile as if just the two of us had seen a pot of gold at the end of a rainbow.  Next we were off to the potty, one of Collin\'s other new passions.  We\'re potty training in full force now.  But this trip Collin didn\'t want to go potty himself \"No\" he just hung around with me in the stall occupying himself with all the delights of the toilet paper and flushing and the little trash can.  I started our adventure in the stall by getting out a liner for the bowl.  Just as I was about to put it on the seat I thought better of it and decided to give that one to Collin to play with, undress a little bit, then put mine on the seat and pull-down-sit-down in one swift motion so he wouldn\'t steal my liner.  Well wouldn\'t you know he took the liner and placed it perfectly on the seat just like my own personal potty valet.  And then he quietly amused himself with toilet paper which he then handed me when I was ready.  I was beside myself with the bounty of my good fortune.  It\'s not that Collin isn\'t a good kid, he\'s actually a really good kid.  But keep in mind it is about 9:30pm and he is still in excellent spirits and blissfully happy and well behaved.    His Nannie, as he calls his grandmother, was on Collin duty when Mikie and I decided it was time to go.  We packed up our little site and looked out over the crowd searching for Collin and Nannie.  It was such a beautiful sight.  Silent reverence for the beautiful music, candle light, the twinkling of the stage lights in the distance, 16 stars in the darkening sky.  We didn\'t spot Collin but sure enough, when there was a pause from one movement to the next we heard his familiar \"Yay!\" and those little hands clapping and we knew just where to go.  '),(19696,3556371,'27,June,2004','2004-06-27 00:00:00','   Why does anyone do a blog, if no one\'s gonna read it. No one\'s gonna read this, so I might as well not write it. Maybe we do it to escape. To think out loud. To let some shit out. I don\'t know. Over and Out   '),(19697,3556371,'26,June,2004','2004-06-26 00:00:00','   I have nothing to do while my cd\'s burning, so I\'m going to write my second blog today. Let\'s talk about theme park rides. I like coasters. They\'re all like \'whooooooooooo!\' I went on my first loop today. It was scary, and I\'ve got fucking bad whiplash, but I\'m Ok. I think. Hell\'s Gate is so scary, you get in this big tube thing and you spin and turn and oh shit, it\'s scary! Oh shit, that\'s the name of one of my 3 cactai, Ouch, Arrrgh! and Oh Shit! I named them after what I said after touching their spikes for the first time. They like to soak up the sun in my room. Ah bless...........Ouch! Arrgh! Oh Shit! It was dark once and I went through my room, almost tripped on the crap on the floor and landed on them. I mainly got Arrrgh!ed, as he (or she)\'s the tallest. Ouch is a normal castus, tall and straight (Like me :) ) Arrgh is taller than Ouch and his (or her) spikes are surrounded with soft fluff. Oh Shit is a big yellow flower with killer inch-long spikes hidden underneath (you can imagine why I said \'Oh Shit\'). I don\'t favor any of them anymore than any other flower or plant, but I love them, awwww......Ouch! Arrgh! Oh, Shit!   '),(19698,3556371,'26,June,2004','2004-06-26 00:00:00','   I got an MP3 player a few days ago! It\'s a tiny RCA Lyra one with 190megs of space with my card in. I have 101 tracks of pure \'Tones, Blink, Lostprophets, and other good shit. It\'s so small! Anyway, school\'s out and I\'m now officially a grade 8. For our last day of school we went to Playland, the theme park place in Vancouver. I got whiplash, sick and bruised, but it was fun. Over and out.   '),(19699,3556371,'12,June,2004','2004-06-12 00:00:00','   How about nothing!?   '),(19700,3556371,'06,June,2004','2004-06-06 00:00:00','   It\'s 3am Sunday morning. I hate Sundays. It\'s like a weekend with all the crappy parts of school (apart from school). And the atmosphere\'s always cold and dark, like someone\'s just died. It\'s weird; apart from Tuesday (The day when it suddenly dawns on you that you\'ve got the whole week to go: Monday\'s just \'Ug\' day.), Sunday\'s my least favorite day, yet I was born on a Sunday. Ironic. I was actually due on Friday the 13th, but I was 2 days late. I\'ve only had 2hrs sleep in over 48 hours, and I\'m slipping off now. Must........finish.......blog.............. My hair\'s a lot longer now, or it seems longer now it\'s all fluffed up. My profile pic was taken yesterday, and now my hair seems twice as long. Strange. I\'m trying to grow it to my shoulders or longer, but my brother, who I was racing with, gave up and had it cut. Loser. Anyway, I\'m sleepy. Gonna go in my room and listen to some \'Tones. Night.   '),(19701,3556371,'06,June,2004','2004-06-06 00:00:00','   Welcome to my blog. I will tell you about the teeny-eeny parts of my life that I want to tell you. Most of the stuff I don\'t want to tell you, but I don\'t care. Over and out.   '),(19702,3556371,'22,July,2004','2004-07-22 00:00:00','   Sorry I havn\'t been posting. Blame my crappy PC and I\'ll be posting again soon.   '),(19703,3556371,'05,July,2004','2004-07-05 00:00:00','   As you should know, I support filesharing and http://www.suprnova.org is the king of all things good and shared. I also spend a lot of my time on the forums: http://www.suprnovaforums.com where I can be found as kier. Anyway, I have recently been looking around and playing with Photoshop and have made some quite nice banners (in my opinion) here are some of them:  urlLink     urlLink    And my current Suprnovaforum Sig:   If you want a free sig, feel free to email me (djcrazymonkey@gmail.com)or leave a comment. I like to leave with a question, it\'s like my thing (from now on anyway). So, what do you think of filesharing? Is it good, bad or don\'t you really care? Something for you to think about. Leave comments on your thoughts.   '),(19704,3556371,'12,August,2004','2004-08-12 00:00:00','   That\'s Right, I\'m back and better than ever. I\'m making a new website and I\'ve created some killer banners. Feel free to email me (djcrazymonkey *at* gmail.com). Or you can visit my site at http://www.djcrazymonkey.fartit.com.  Over and out.   '),(19705,1599685,'24,September,2003','2003-09-24 00:00:00','   WE BRING THIS SPECIAL ANNOUCEMENT TO YOU LIVE    from the valley of hte half slit wrist and clenched fist broken bits of narcotics rocked in pockets got us locked from knowledge but we exist between heart beats and hard streets those that struggle to breath can\'t seem to find peace in the belly of beasts i pray more than ever now  when sleep finds me..  '),(19706,1599685,'13,September,2003','2003-09-13 00:00:00','   imma take this opportunity to finally release any of the scattered thought patterns lying dormant in my mind. so much seems to change or happen in such a short amount of time, that i find it hard to keep up with it sometimes.  it\'s hard knowing where to begin or express how you really feel about certain things.  so this will be a place where i come to meditate on any and everything in an attempt to break free from my writers block.    â€œInkâ€ In the heart of darkness  and the cool of night My ink takes flight  Black like night upon white Write to release light  Shine bright  In the heart of darkness And cool of night     Till next time. Stay real and righteous with the written word.  Knowledge Reborn  '),(19707,1599685,'23,June,2004','2004-06-23 00:00:00','   Still impressed to see the light  for the life of me... Though my mind star see far in the darkest of nights we bleed hard times got my mind  scattered and scared but i stil see the art in a beating heart the part that pumps  the positive patterns we are    First and foremost...   -- In my opinion  urlLink Saul Williams  if anyone fights to see the light...for 3rd eyesight to be free, it would be him..So i sugesst you got get this book  SHE    Here is an excerpt:   i presented my feminine side with flowers  she cut the stems and placed them  gently down my throat  and these tu lips might soon eclipse your brightest hopes...      urlLink African Links   Next i would suggest you get that  urlLink Common  \"One day it\'ll all make  sense\" if you havn\'t already...The track  \"Restrospection\"  hits home when i take a look at my daughter (born on May 14, 2004) Nijah     ...more pics in visuals.    I just discovered the  urlLink Hieroglyphics  i suggest you do the same...its so many positive artists out there that are unknown to me...that a simple search will finds one amazed at whats\' being put on the table to be consumed by eager ears.    Other links in to particular order:   urlLink The Roots    urlLink Black Lily    urlLink Dilated Peoples        urlLink Kanye West   urlLink Halftooth   urlLink Now Hear This  '),(19708,1599685,'23,June,2004','2004-06-23 00:00:00','   -escape without a trace to the place of infinite signs and space as time flies-watch the skies pay attention to what the sun do battle ignorant minds cuz they blind...redefine this art of war... all honors to   urlLink sun tzu      People who know me Would say that I am SHY- somewhat quiet person at times. So I don`t even have to say that i`m not attracted to loud women who talk to much or for that matter anyone who talks to much and never has anything to say of substance.  When I say that I`m a real LAID BACK type of person; that just means that it does not take much to keep me entertained and i rarely get stressed out. Everythang is everythang. I wonder if you can keep me entertained --mentally? Hope so.  As far as being STRONG WILLED, when i see something that I want and I have set my mind on obtaining it I will not give up until its mine. So that takes alot of self-confidence. Not bad for a shy person. So please believe me when i say that just because i am shy does not mean i am a push over; i never let people walk over me....That wouldn`t match my character at all...because i like to be in control of any and all situations i find myself involved.  DETERMINED (excerpt from a poem i wrote a few years ago) ...`Though your words stir in me restless winds/ I still remain emotionless and will not bend/ You can beat me but not brake me/ Look at me with scorn/ You can tear this flesh of mine/ But my mind is never torn...` I`ll let you interpret that one on your own.  As far as HUMOR I believe that you have to have a sense of humor because life has a way of doing things that you have no choice but to laugh at it or it might just drive you crazy. People who take the time to get to know me will find out i`m real easy to get along with. Even though im a serious person i make alot of jokes to keep the people around me smiling.  In my opinion i`m DEEP. Or better put self-conscious. If there is something about me that can be improved upon and you have that knowledge...By all means speak it..b/c I will listen. I`m a firm believer in change. Anything that continues to stand still is not making progress. You may think otherwise but I suppose the only way to find out is to try and get to know me. To find out why I hold certain opinions about life, love, religion and the human soul, poetry, and utimately the artistic poetry in loving life through God (or gods-whatever you believe; i do not judge) and your own soul. To me the meaning of life is coming to the realization that you are alive and that is profound in itself. LIVE LIFE.    -cemeteries are our only tended gardens we tendto die the moon is full. the tide is risen. the future is blossoming around us. yet, the fields of our destiny have been gardens untended. over-grown, and malnourished. we have come to cultivate this garden. we are the cultivators of a culture. (excerpt from Saul Williams linear notes.)  '),(19709,2190377,'29,October,2003','2003-10-29 00:00:00','   It\'s eleven, I have a fever, and have studied a grand total of one hour for three tests.  Oh Boy. Why can\'t Thursday be over? I think I\'m going to head home early because I don\'t feel like taking my German exam. Lalala. My pain medicine will have worn out by then.  Moo  That\'s all I have to say, and if I said anything weird to you today, it\'s \'cause I do REALLY weird things when I\'m feverish... Seriously.  I\'ve got strange numbers running around in my head *sigh*.  We\'re talking: 6.63 * 10^-34 J sec ... -2.18 *10^-18J (z^2/n^2)  -313.6/n^2 109678/cm (1/n1^2-1/n2^2) Lyman Balmer Paschen Bracket Pfund 1s2 s2 2p6 3s2 3p6 4s2  3d10  That\'s just chem! weeeee  Then there\'s AP bio, which I know nothing about.  I know like two words that pertain to crap on the test.  Oxaloaceate and Ribulose bisphosphate carboxylase. lalalala That\'s the extent of my knowledge. Ok. Maybe not.  I also know crap about carotenoids, which, coincidentally make me think of Carrot Top. I\'m going to explode  bye now  '),(19710,2190377,'28,October,2003','2003-10-28 00:00:00','   I need go-go boots very badly.  Give me go-go boots. now  As of now my halloween costume consists of: white mini-dress with bold neon orange and pink stripes, chunky white gaudy plastic necklace and clip-on earrings (sad, I know, one of my many goals in life is to get at least four piercings as soon as I turn eighteen), opaque tights of some kind or another, and fake eyelashes. All I need is a huge headband and go-go boots, *sigh*  In other news, my grades continue to make no sense, now that I\'ve aced the last four things I got back in chem, and got Cs on everything in Bio.  I\'m hoping that the ADD test dealie says there IS something wrong with me, just so I know it can be fixed......... lalala   Anyways, I cut my hair, and everyone says it makes me look like an eighth grader, even the supposedly devoted boyfriend is less than thrilled.  All this animosity towards my hair is urging me to screw it up even moreso, just as a protest of sorts you know? So this means, dye it snot green and get a pseudo mohawk. I\'m sure the guys would just flock to me.  I\'ve noticed that guys have no taste in girls, what\'s up with that? They pick these really ugly ones, and also decide to like me, so by process of association, I am offended.  Especially when people have really weird old crushes or exes... it makes me wonder if I fit into the category of bad taste girls.    I say that I\'m smart but people just think I\'m a moron, but judging from what I just wrote, maybe they\'re right.  lalala  This was a stupid post. yaaaaaaaaay  '),(19711,2190377,'27,October,2003','2003-10-27 00:00:00','   I\'m not exactly sure what to do with this thing now that I\'ve got it.  I think I\'ll just complain :) lalala So recently, I\'ve noticed that there are five recurring jokes in my WHOLE life.  Five. They are: 1. Ashleigh is a big slut. 2. Ashleigh is a lesbian. 3. Ashleigh has no friends. 4. Ashleigh is a dumb blonde.  Unfortunately, I managed to combine the last two (see how there are only four? ohhh ahahhahaha yea now you get it) while on the phone with Carisa and Matt (a conversation mostly about them calling me a lesbian slut... :P). Of course this warrants a \"Wow Ashleighs is such a dumb blonde joke.\"   I think that instead of actually saying it, people should just come up to me and say the number corresponding to the joke or, if they\'re really in the mood to slam me \"One hundred twenty-four!!! ohhhhhhhh haahahahaha\".  It would save us all a lot of time.    Now my quest in life is to find more jokes, so maybe I could get ten or so, just to spice things up.  '),(19712,2190377,'27,November,2003','2003-11-27 00:00:00','   After much boredom and isolation, I\'ve decided to run through the numerous exciting (read: fucking messed up) events of the past few weeks for my ever faithful readers.  Who the hell actually reads this? REALLY?  Oh well.  So far, I\'ve: lost best friend of five years, screwed relationship with boyfriend of six months to point of no return--via aforementioned lost friend (messy situation, I\'ve probably already told you anyways), single-handedly destroyed my academic career, been told my recent ADD test scores were so bad I scored lower than 99.5% of people that took it, been told I\'m not going to be medicated yet because my brain waves are abnormal and medication may f*** me up even more, gotten seriously ill, played basketball with 100 fever, lost voice to the point where my sexuality would be questioned by strangers, blown off at least half a dozen friends, lost seven pounds, and had three close friends tell me they would jump my bones were retarded boyfriend not to exist.    You try and swallow all that when you\'re so weak you can barely walk up the stairs.     Not that I\'m bitter or resentful of anything.  Heavens forbid.  So, tell me about your life. Oh, you feel as if I\'ve been ignoring you? I\'m so sorry, I\'m such a selfish person, how could I have put my needs before yours? Next time my life falls to pieces I\'ll drop everything so I can support you.  I realize what a sensitive person you are.  And your life is just so tough. You poor thing...  Wow.  I\'m a regular basketcase.  Just so you know, that wasn\'t directed at anyone that actually reads this, except maybe Jeremy. haha. hahahahahaha.  I hate my life.  Die bitches.   Happy Thanksgiving!!  -_-  '),(19713,2190377,'23,November,2003','2003-11-23 00:00:00','   Turkey day is a good excuse to get out of school.  My grades suck.  Apparently I have really bad ADHD.  We\'re talking three standard deviations away from the norm.  I have too many clothes.  Karin makes good cheesecake.  Elf is a funny movie.  Andrew looks a whole lot like Ed, stretched out.  Andrew now weighs 141 pounds.  I\'m very proud.  Writing disjointed sentences is entertaining.  My life sucks, but it\'s getting better.  I\'m terrible at basketball.  I\'ve grown half an inch since last time I was measured.    WAhoo  My butt, however, is still dead.  It needs to be resurrected.  Goodbye  '),(19714,2190377,'23,November,2003','2003-11-23 00:00:00','   I haven\'t wrote in this for awhile... but then again I\'ve also had one of the worst weeks of my life! :)  I\'ve been sick since Wednesday night.  Too bad I went to school both Thursday AND Friday.  Never play in a basketball scrimmage when you have a 100 degree fever, ok?  Anyways, being sick has made me realize the best skeezy quick-fix diet plan ever.  You know how they have those things you drink that basically make you throw up all weekend? Well, I figure, we can send these people food that I\'ve coughed all over or something lol. That way, they\'ll get the flu, their tummyswill fill up with mucus and their throat will hurt too to swallow anything besides jello! End result: knock off a few pounds. I\'m surprised someone hasn\'t thought of this sooner.  My butt is officially deceased.   I got two catalogs in the mail yesterday. One, American Girl Doll, the other Victoria\'s Secret. hahahahahaha   anyways. I\'m still feverish, so don\'t pay attention to anything I wrote.  '),(19715,2190377,'18,November,2003','2003-11-18 00:00:00','   k me and matt we were going to walmart it was about 11 o clock at night cuz we wanted some doritos and tootsie rolls, i dont know my matt wanted tootsie rolls but he did.  so we go in grab our doritos and tootsie rolls. and then this guy whose about 6 foot tall and probally weighing about 300 starts to follow us. so we started walking all around to try to lose him but he keeps following us. matt turns around and matt said \"WHAT THE HELL DO YOU WANT STOP FOLLOWING US!!!!\" and the man said \"im soo sorry, its.. ( he begins to cry) you look like my son who died yesterday in a car wreck.\" and matt being the nice guy he is was like im so sorry i wish i could make it up to you its just that you kinda freaked us out.  so me and matt walk around alittle bit to make sure he doesnt keep following us. we get to the register and hes right in front of us and we try not to notice him but he starts talking to matt how right before his son drove off he got in a fight with his son and he never got a chance to say bye. can you say bye dad its my fault and ill say bye son no im sorry about it. and matt still feeling bad said sure. so matt says bye dad its my fault and the guy says bye son im sorry about it. we check out and the cashier says that will be $334.43!!! we were like no your wrong and she said no im right his dad said youse would pay.. and im like quick call the cops.  and i run out there and start yelling at the guy why you do that that not right and hes like man i gotta feed my family and hes throwing in like snickers and milky ways and junk food and i try to grab him to stop him but hes bigger than me so he throws me on the ground and starts kicking me in the legs and the ribs matt came cuz he saw this out the front window of walmart he comes knocks the guy off me and has him on the ground and i start pulling his leg..... jsut like im pulling yours          (stole this from my buddy eric lol)  '),(19716,2190377,'10,November,2003','2003-11-10 00:00:00','   Weighed self again. Five pounds miraculously gone. weiiiird  Surprisingly, basketball rocked  I won\'t be thinking the same thing tomorrow, though.  I can feel the pain in my butt already.... yowsa  Too.. tired... can\'t ... write anymore.  Talk more about the weekend tomorrow :) lala  '),(19717,2190377,'06,November,2003','2003-11-06 00:00:00','   That last entry was a cop-out lol.. oh well, I   Man.  Halloween may be my favorite holiday... but it made me gain five pounds.  Haaaaaaaah.  That\'s too bad. It\'s okay, though, because basketball starts MONDAY.  The end of my freedom. Forever........  blar  The bus rides have been fun while they lasted :D Today we played telephone.  It was made all the more interesting by the fact that Matt is almost deaf in one ear, as proven by those tests that make you take in driver\'s ed.  hehe. SOOOO! Chris Peterson starts \"Ashleigh likes pink elephant nuts\",  later turns into \"Ashleigh licks pink elephant nuts\" and by the time it gets to me it\'s turned into \"Ashleigh charges nine cents a blow.\"  Obviously I was GREATLY offended by Chris :) So I changed it to his name, thinking that he was a dirty perv who started a naughty line.  He practically died when he heard the end result.    Ashleigh likes pink elephant nuts = Chris charges nine cents a blow  oh boy!!!  Time to start math project.  bye now  '),(19718,2190377,'05,November,2003','2003-11-05 00:00:00','   PERSON I GOT THIS FROM: Cathy  :x: name = Ashleigh Anne Allen :x: piercings = zero... but i want some REALLY bad, both ears, two in one.. cartilage, eyebrow, tummy.. all hot, but i cant get any :x: tattoos = Only a henna once and it wasnt as fun as I\'d wanted so I glad it wasnt permanent  :x: height = 5\'5\" .. im doomed to be a midgit forever :x: shoe size = 8ish :x: hair color = depends who u ask, dirty blonde pretty much though :x: siblings = older bro, and two super old halfsibs  DO... :x: you wish you could live somewhere else = not now.. but eventually i wanna live somewhere with a cool urban scene and decent weather :x: you think about suicide = never considered it even half seriously :x: you believe in online dating = no. how dumb.. well at least if youve never met \'em :x: others find you attractive = definitely. oh baby. *cough* :x: you want more piercings = already said so :x: you drink = nope..   :x: you do drugs = nope.. :x: you smoke = no way .. someday ill drink and do drugs prolly, but NEVER smoke, how pointless :x: you like cleaning = if no one forces me too.. and i can dance and listen to music in my underwear while cleaning, then sure lol :x: you like roller coasters = :D :D :D :D :D  :x: you write in cursive or print = chicken scratch.. kenny says i write like a 2 yr old lol   Are you for or against... :x: long distance relationships = depends how long u go without seeing the person :x: using someone = as long as no one gets hurt and its not sexual its not that bad.. like a lil white lie :x: suicide = STUPIDEST THING EVER :0  roar.. people who kill themselves are SO selfish, if they really hate themselves so much just become a monk or something, thats selfless.. and u can help people :x: killing people = bad.  :x: teenage smoking = retarded..... RETARDED almost even more retarded then suicide :x: doing drugs = meh, as long as u arent a regular user..  :x: driving drunk = BAD BAD BAD :x: gay/lesbian relationships = gay people fascinate me lol so.. yea i have no problem  FAVORITE... :x: food = so many choices, how about candy... lala  :x: song = oh man, dont get me started... thats too hard of a question :x: thing to do = dance run sing skip sleep kiss hug :x: thing to talk about = how stupid i am .. how im really not that dumb and other people are idiots.. and boys.. and perveted shtuff... and the funny way that the sexes misunderstand each otha :x: sports = SPORTS ARE YUMMY ... track bball and field hockey are my top three but i wish i was coordinated enough for ice hockey, and SOME DAY IM GONNA PLAY RUGBY BABEE :x: drinks = coffee, root beer :x: clothes = skirts sneaks and a funny tee :x: movies = scary movies and funny movies are the best.. yay :x: band/singer = too hard :x: holiday = halloween.. cause its the only thing untainted by family problems.. i hate thanksgiving, christmas is alright.. but man i HATE valentines day :x: cars = white convertible vw bug. ohh baby :x: disney movie - Lion King :x: word ? happy. definitely :x: nickname emmmmm ash, ash-e-leigh, ashee, ashwee, bobsleigh, turtle, raphael, silly lil girl :x: guy names ? something really traditional :x: girl names ? something really screwy and preferably japanese  :x: eye color- green. oh how i wish i had green eyes, but no, i get stuck with blue grey. how boring :x: flower ? silly flowers.. or really purdy roses :D :x: piercing ? eyebrows are my fave.. and lip piercings are hot too, as are tongue.. but too dangerous :x: actor ? O:-) heath ledger lol, but come on australian accents... and I  :x: actress ? Julia stiles or reese witherspoon  HAVE YOU... :x: ever cried over a girl = more than i have over guys -_-:x: ever cried over a boy= yessir :x: ever lied to someone = maybe a few times but im REALLY bad at it :x: ever been in a fist fight = nah ive beat jeremy up tho lol :x: ever been arrested = nope    WHAT... :x: shampoo do you use = herbal essssssssences :x: shoes do you wear = chucks or some other weird shoes :x: are you scared of = going blind, being alone, being disliked, failing high school :P NUMBER... :x: of times I have been in love? = i guess twice :x: of times I have had my heart broken? = its never broken, just occassionally fractured :) :x: of hearts I have broken? = millions. im just that hot :x: of girls/guys I have kissed? = 2 guys. how boring -_- :x: of drugs taken illegally? = none :x: of people I would classify as true, could trust with my life type friends? = dunno half a dozen to a dozen :x: of times my name has appeared in the newspaper? dunno  :x: of scars on my body? = not alot cept on my feet, i have like 12 ther -_-.. im just clumsy :x: of things in my past that I regret? = a lot DO YOU THINK YOU ARE... :x: attractive ? im not ugly. i think :x: funny - funny in a laughing at you not necessarily with you sense lol :x: friendly - definitely  :x: amusing ? :D hehehe i amuse myself :x: loveable - ehhh :x: caring ? usually :x: sweet - no :P lol  :x: dorky - hehehehehehe. im cool.  '),(19719,2190377,'03,November,2003','2003-11-03 00:00:00','   I\'m seriously wishing my dogs would stop having sex. RIGHT NOW.   In other news...  Thursday didn\'t suck too hard, Halloween was cool, Texas Chainsaw Massacre made me nearly pee my pants, and having bleach blonde crazy women place a couple dozen electrodes on your head while making small talk about your \"long eyelashes\" is more than a little disgruntling.  I should be reading 40 pages in my history book, but because I don\'t actually NEED to read it in order to get an A, I\'ll compile a list of desired video games games. Can you tell I like lists? I\'m not sure why, especially because I\'m not organized. at all.    p.s. Barney, my non-horny dog, likes carrots :)  lalala...oh yea  list  (all for ps2)  DDR 2, Karaoke Revolution, Kingdom of Hearts, FFX-2, and whatever the newest Spyro game is  I\'m bored. Maybe I should read about the Foundations of Indian Society, \'cause hell, that\'ll wake me up  '),(19720,2190377,'30,December,2003','2003-12-30 00:00:00','   morning :   so today andrew and i were in the middle of a very serious discussion on the phone  suddenly, he says to me in an urgent tone   \" That  was a good sandwich.\"  okay its just not as funny as it was in reality  but deal.  maria wasnt at practice today so my days been pretty boring :(  im still on the phone with him.  *gagging noises ensue* he\'s been choking on a popsicle the entire time i wrote this.  just now he overcome the popsicle ...and exclaimed in a disturbingly serious manner, \"Boy am I glad I\'m not gay.\"  thats all for now.   except  i hope you die  later today:  apexfb54 (10:18:18 PM): why do you hate ashleigh? MeNtaL AbSeNcE (10:18:32 PM): she wants to hurt me  if i wasn\'t me, that would be amusing  spaceypiXiechick: well so could johnny depp but no ones threatened by him lol EvilKennyOO7: lol well put EvilKennyOO7: i saw pirates of the carribean for the first time a couple days ago in NC spaceypiXiechick: niiiice spaceypiXiechick: didnt you want to do him to spaceypiXiechick: with the black eyeliner spaceypiXiechick: hahaha spaceypiXiechick: even though hes FORTY EvilKennyOO7: haha EvilKennyOO7: i was like \'why the hell is this guy the \"sexiest man alive\" spaceypiXiechick: haaaa EvilKennyOO7: now that orlando bloom, mmmmmm :-P spaceypiXiechick: :-D spaceypiXiechick: i love you kiddo  tuoriginal4u: you\'re not the bitch woman though  '),(19721,2190377,'27,December,2003','2003-12-27 00:00:00','   I feel as if a certain raccoon has been raiding my garbage whilst I sleep.  I\'m tired of having my crummy leftovers pawed through.  I think she needs to stop.  Otherwise... next time there will be a giant mousetrap waiting for her.    There are no raccoons.  Go to hell.    im soooo tired  ran 30 sprints, killer eights, 2 suicides, and 10 laps around the gym  *collapses*  because nothing has happened yet today, i shall instead copy and paste and be boooring wow i chose REALLY weird word order for that sentence  spaceypiXiechick: im such a seventh grade boy spaceypiXiechick: ahhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhh hhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhh hhhhhhhhhhhhhhhhhhhhhhhhhhhhh GoCombatChuckGo: ??? GoCombatChuckGo: horny?:-P spaceypiXiechick: lmao spaceypiXiechick: that too   spaceypiXiechick: cause he is such a hawtttttiee etiquette9 23: haha spaceypiXiechick: and so is that girl maria spaceypiXiechick: dang. etiquette9 23: haha etiquette9 23: maybe i should wear that jersey every single day! spaceypiXiechick: hahahaha spaceypiXiechick: u might start to smell spaceypiXiechick: but itd be worth it  i think my life has turned into roleplay  Edguy10: >=I i am not a whale penish   jocolono57: i want to hitchhike in an ambulance on day  *felt like freeing up the profile*  '),(19722,2190377,'25,December,2003','2003-12-25 00:00:00','   Man, I was seriously convinced today was Thursday.  WHAT A JIP!!! I dont know how to spell that word.    These past two days were interesting... yay  I was gonna go shopping with Karin today but her mom took us to the TORPEDO FACTORY instead hahahah IIII had to help giving directions. hot.  heh.. Torpedoes.. Doritos.. I wish I was a freshmen again.  ANYWAYS!!! At this place, they don\'t make weapons, they make art! Karin and I talked to this cute Chinese man who painted and babbled about the \"THE POWA IS IN THE PAPUH!!!\" power in the paper? strange strange, then he asked me to try with his watercolor brush.  Too bad I couldn\'t even make a straight line.    Have you ever noticed that the only people who paint or photograph female nudes are men, or lesbians? Maybe I should be the first straight (  in other news  ANDREWS REALLY WEIRD :-P spaceypiXiechick: ive started saying yea in that ghetto kinda way spaceypiXiechick: ewwww GoCombatChuckGo: haha GoCombatChuckGo: just say yee-ah GoCombatChuckGo: instead of yee GoCombatChuckGo: cuz yeeeeeeee kinda sounds... i dunno GoCombatChuckGo: vulgar in a really weird sort of way spaceypiXiechick: hahahaha GoCombatChuckGo: *does the donkey kong ?!? thing* spaceypiXiechick: i have no idea what youre talking about spaceypiXiechick: freak spaceypiXiechick: ;-) GoCombatChuckGo: kiss me, ima freak and im on the edge GoCombatChuckGo: im sure thats a line from some obscure B-movie GoCombatChuckGo: im just not sure which spaceypiXiechick: LOL  all ive eaten in the past week is fudge, cookies, chinese food, and ice cream...somehow i lost four pounds  *laughs maniacally*  im not sure if i ever thanked maria and vidya for coming over on tuesday for rocking my world w/  ice cream and JOHNNY DEPP  oh boy.. johnny depp porn :P :)  lalaaaaaaaaaaaaaaaaaaaaaaa ok im hyper  Last night was so weird, I was  alternating between being in a chat room discussing pokemon with two cool people that i barely know, and writing extremely literate dear to my heart shit that you arent privy enough to read NEENER NEENER.  unless you are a huge bitch who wants to come clean.  but i dont think you would consider yourself a bitch so...  *dances*  GUESS WHAT WORLD FRANK IS MY NEW BURNIN HUNK OF CYBER LOVE   i have more to say but its time to shut up  '),(19723,2190377,'25,December,2003','2003-12-25 00:00:00','   Merry Christmas and all that jazzzzzzzzzzzzzzzzzz  IM FINALLY FREE!!!  yay.  its nice to be able to jump and run and ...yeah haha ;)  anyways  **things charlie has eaten (or nearly eaten) in the past two days:** two ornaments sam\'s ear squeaky toy squeaking device his own excrements a paper my brother used for loogie (sp?) disposal half of the contents of our trash  anyone looking for a large male six-year old english setter? lemme know  '),(19724,2190377,'22,December,2003','2003-12-22 00:00:00','   weeee.... plugged into the wall *dances* soooo funny  so jeremy and i had this big argument over whether the plural of hippopotamus was hippopotamuses of hippopotami.... and he looked it up on dictionary.com  TURNS OUT BOTH OF US WERE RIGHT OMG hhaahaahahhaha  im so high *seconds after i typed this*  JerMan144: you sound high spaceypiXiechick: aww thanks  i got lotsa good quotes today too  GoCombatChuckGo: i think maybe masturbation stunts growth spaceypiXiechick: u should start a scientific study  imnochewtoy: ashleigh I am terrified of you in general  virgogurl4: who was ur tecaher last yar? spaceypiXiechick: charnitski!!! spaceypiXiechick: hes awesome AND cute :-) virgogurl4: ohh yeah virgogurl4: hahaha virgogurl4: i want ihm virgogurl4: ihm virgogurl4: hahahahaha virgogurl4: gemran spaceypiXiechick: LOL  marias so funny imma have to paste like half of our convo etiquette9 23: ASH etiquette9 23: how etiquette9 23: s it going conehead spaceypiXiechick: :-D spaceypiXiechick: its... going etiquette9 23: :-) etiquette9 23: i love you spaceypiXiechick: aww etiquette9 23: do you feel the electrodes in your brain moving around? spaceypiXiechick: lollll spaceypiXiechick: no... etiquette9 23: do you feel super energized cuz they aren\'t ESCAPING!? etiquette9 23: haha etiquette9 23: i would spaceypiXiechick: :-Pt spaceypiXiechick: those are elecTRONS etiquette9 23: no way etiquette9 23: like NEGATRONS????? etiquette9 23: hahahahhah!!!!! spaceypiXiechick: hahahahaha spaceypiXiechick: megatron spaceypiXiechick: TRANSFORMERS etiquette9 23: YAY spaceypiXiechick: :-D spaceypiXiechick: i love that show spaceypiXiechick: but only the old ones etiquette9 23: haha spaceypiXiechick: my tummy hurts spaceypiXiechick: and i lost my fortune :-( etiquette9 23: oh etiquette9 23: i\'m a gonna make you a fortune etiquette9 23: you will find an electron on your butt sometime tuesday. it escaped. spaceypiXiechick: HAHA spaceypiXiechick: ohhh man spaceypiXiechick: ur awesome etiquette9 23: i know spaceypiXiechick: i hope it happens for rea spaceypiXiechick: *real spaceypiXiechick: then theyll be monitoring my butt waves... ewww etiquette9 23: HAHA etiquette9 23: that would be great etiquette9 23: you could have your butt all taped up   GoCombatChuckGo: well i gotta hand it to you... youre unique. GoCombatChuckGo: and awesome. GoCombatChuckGo: and weird. spaceypiXiechick: hahahaha GoCombatChuckGo: and sometimes a little psychotic. spaceypiXiechick: what makes you say psychotic ? :-) GoCombatChuckGo: oh i dunno GoCombatChuckGo: you have a list of violent crimes to commit GoCombatChuckGo: not that i havent fantasized about PAINDEATHKILLING and  thats it  for real   wait no just kidding  hahaha... like in lotr.. and finally they actually have to say THE END so you dont sit there waiting for a sixtieth epilogue  THE END  '),(19725,2190377,'19,December,2003','2003-12-19 00:00:00','   WOW I haven\'t slept ALL week!!!! And I\'m spectacularly hyper. I can\'t even write down all the things I\'d wanna say.... weeee I went to bed at two yesterday, and woke up at seven by accident -_- But... that\'s okay! You know what\'s fun? Car crashes!! weeee. The other day on popeshead this car flipped over, and three busses got stuck on my two lane country road. HAWT. Cause Cause Cause.... I missed my connecter bus to TEe JaY and didn\'t have to take a sucky ass chem test!!! *dances*  Um.  Yea.  That and I have a 96 in AP Bio, a B+ in Chem... and other good-ish grades! German gingerbread cookies = yummy.  Kelly\'s \"christmas present\" was a lifesaver.  She is my angel.  HAha.  Assistive technology club is awesome but Luke needs to go to hell cause hes so good at drawing on the \'puter.  Can you tell I like writing down every thought that pops into my brain?   Now I\'m having illicit thoughts.  I\'m going to omit those.  I\'m starting to love my hum (if only I didn\'t have a C+ for not turning in some stuff... *angel face*) Mrs. Robbins rocks, and even though Mrs. Martin is off her rocker I\'m warming up to her too.  All we did today was discuss traditions... and I made myself look like SUCH a dork in front of fifty people.  SO much fun.  Then Jordan short boy took my balloon for maddie and let it float up to a place where noone could reach it.  At this point I went beserk and starting screaming obscenities about how i was going to eat the midget fetus freak for breakfast.  Turns out he was right around the corner :)  Apparently the kid took off as fast as he could.  Then Boris threw a shoe at it so I could grab it from the overhang.  Unfortunately, it also smacked me in the head.  Either way, my aggression carried over to my basketball game, where I tackled several people and pushed around a bunch of fat girls.  I totally owned all of them.  That and I actually got playing time.  What a great Christmas present Webber!! :D haha.  Oh man, Barbie girl was on our warmup tape.  That must be the magic thing that makes me play like someone 5\'10\" and like 160 pounds.  hahahahaha .... I played post.  Doesn\'t that make you laugh? Speaking of basketball, Maria B. is a hottie and I\'m going to marry her.  She got me NASCAR nailclippers Christmas.... haaaaaaaaaaaaaaaaaaaaaaaah oh man.  :)    btw.  JOSE R0x0rZ!!! He\'s my best kept secret.  that, and Rachel Miller is the most confusing girl on the planet.  And Matt deserves another kick in the shins...  heh...  that\'s all folks  virgogurl4: I HAVE A BIG CRUSH ON HIM virgogurl4: i mean virgogurl4: i dont virgogurl4: my fingers slipped virgogurl4: i meant to say virgogurl4: I LIKE TO CRUSH ORANGES ON HIM  '),(19726,2190377,'14,December,2003','2003-12-14 00:00:00','   lalala  Basketball sucks.  Andrew Yang rocks for giving me the Offspring CD, useless weekends are boring but nice... i hope it snows, monday i get electrodes to play with  -_-  weeee  i realize i make no sense and don\'t particularly care  screw capitalization  i have like fifty christmas presents for ppl and dont know who to give any of them, maybe ill make like a list and you can choose what u want haha  '),(19727,2190377,'09,December,2003','2003-12-09 00:00:00','   Today I wasted three fucking hours for three minutes of playing time.  Heads will roll.  '),(19728,2190377,'05,December,2003','2003-12-05 00:00:00','   In the past two weeks, I have done NO homework at home.  What\'s funny, is that my grades haven\'t really suffered at all.  >=) I\'ve done all my work in other classes... and I\'m slipping by juuuust fine.  Maybe I should stop. Or not.  Yesterday I walk into driver\'s ed and Mr. Potoker says VERY loudly, \"Even Ashleigh Allen does!\"  Which prompts a \"What do I do??!?!?!?!\" him: \"Oh. I just said that even you thrive under pressure on tests.\"   bwahahahaha.  definitely gotten B+s/As on his tests even though I sleep allllll the time :)  It was so great...  In math the other day I was pulling what I thought was a loose string off my red shirt... And Andrew Yang looks at it and goes, \"But your shirt isn\'t white!!!! I don\'t get it!!!\" in his silly Andrew Yang kinda voice right?  hehehe. Think about it.  Then he got all grossed out and talked about how I needed to duct-tape my bra together so it didn\'t come off.   Oh boy... Time to redo my chipped nails  '),(19729,2190377,'04,December,2003','2003-12-04 00:00:00','   okay. I\'m not as psycho now... haha... umm.. yeah.  SCHNEE SCHNEE SCHneeeeeeeeeee  I can\'t spell in German. Too bad.  also  SNEG = SNOW in RUSSIAN  Did you know that german cats go \"Miau!!!\" ???    like, OMG  M.R.I.s are fun, it\'s like being in an arcade game with all the friggin\' bleepin and whatnot.  Spaaace Invaders   Today we ran seventeen sprints holdin bballs over our heads.  :D oh boy! :P afterwards Vidya pushed buttons on the soda machine, and a diet coke came out!!!! yay!!! now I\'ve had two.  I\'m so high :) its like the Tal Bachman song from the nineties baby... shes soo hiiiiiiiiiiiiieeeieegggggggggh high above me, so lovely, she so hiiiieeeeeeeeeeigghhhh like cleopatra joan of arc or apppphrodite doo doo dooooo do doooooooo   im out  sorry about the lack of grammar  '),(19730,2190377,'31,January,2004','2004-01-31 00:00:00','   i dunno why this thing posted twice. grrrr  '),(19731,2190377,'28,January,2004','2004-01-28 00:00:00','   Woowwwwww.  Best song ever =  I Believe in a Thing Called Love  yeeaaaah baby. hahahaha  you gotta love feminine lookin\' wanna-be 80\'s rockers.  AND  you gotta love our three game winning streak!!! and the fact that i started (haha so what if it was punishment to the others for not wearing their sweats... i still got to play and didn\'t do THAT badly.)  Actually, I kinda rocked, in the ... scaring the crap outta their point guard way of rocking.  yay yay yay.  Besides that, I\'m a very confused person, and NEXT WEEK IS GONNA BE ICKYYYYY  I wish I could start this year over.  Actually, no.  Because it would probably end up bad anyways.  I just wish my mental state could start off with a clean state.  im soooo jaaaded.  *aerosmith plays in heaD*  '),(19732,2190377,'28,January,2004','2004-01-28 00:00:00','   *stolen from Kaitlin\'s livejournal* http://www.livejournal.com/users/futbolistautd/  you bold the things ya got in common with the person and make up your own for everything else  01.  I have a cell phone.  02.  I\'m a lazy butt.  03.  I\'m the youngest child.  04.  I can\'t concentrate on homework.  05. I don\'t have my ears pierced. 06. I am a Gemini. 07.  I can be very drunk for no reason sometimes.  08. I think of every day as one less day until the weekend. 09.  I can\'t live without my contacts.  10.  I can\'t live without music.  11. Iâ€™ve been to Germany. 12. I  would save most of my money, but my relatives borrow it all. 13.  I want to go to a good college.   14.  I love Reese Witherspoon.  15. I\'m surprisingly tolerant of annoying people. 16.  I eventually want kids.  17. I know nothing about soccer, but Beckham is hot. :P 18.  I have few horrible memories.  19. I\'m addicted to I Love the 80s 20.  I am a person.  21. I had my first kiss at age 13. 22.  Iâ€™m not sure about what to do during the summer.  23. I suck at taking pictures. 24.  I hate girls who are fake.  25. I\'m completely incapable of being mean... in person. 26.  My dreams are bizarre.  27.  Iâ€™ve been to/had a bunch of sleepovers..  28.  I painted my nails a lot when I was younger.  29. My parents are awesome...sometimes 30.  My parents are not very Asian.   31.  I love Finding Nemo.  32. I cry at the weirdest times. 33. I\'m always late. 34.  I procrastinate.  35.  I love spring.  36.  I hate mosquitoes.  37.  I love to sleep.  38. I wish I were less obnoxious. 39. I\'m pretty unhappy. 40.  Iâ€™m afraid of hairy spiders.  41.  I hate crying.  42. I hate my hair. a lot. 43.  Iâ€™m pretty close to my brother.  44. I don\'t really like the beach... or water in general. 45.  I have had the chicken pox.  46.  I\'m excited for the future.  47.  I can\'t control my emotions.   48.  I don\'t know why I take these surveys and stuff.  49.  Blackboard needs to die.  50.  I love my friends.  51. Christmas is such a dysfunctional holiday in my family. 52.  I have my own computer.  53. I\'ve only broken one bone. 54.  I can\'t live without sports.  55.  I love reading.  when I have time.. -_- 56.  I love the movie A Knight\'s Tale.  57. I\'m really bad about holding grudges. 58. I wish I were a happy person. 59.  I love to dance.  60.  I love to sing.  61. I hate organizing my room. 62.  I love cute pajamas . 63.  I love acting crazy with my friends and seeing people stare.  64.  Iâ€™m addicted to chocolate.  65.  I hate seeing animals/people getting hurt/abused.  66.  I want to go to Florida with a group of friends.  67.  I don\'t like to study for tests.  68.  I am Christian.  69. I try to be forgiving. 70.  I have a horrible sense of direction.  71. I hated middle school. 72. I miss elementary school. 73. I\'m a momma\'s girl. 74.  I hate when girls wear really short shorts to play sports.  75.  I love color.  76. I have been called dense by most of my teachers and coaches.  -_- 77.  I have no idea what color my eyes are.   78.  I hate that people are still bringing things up about Princess Diana. Let her rest in peace!  79.  I have glasses but wear contacts.  80.  I become stressed easily.  81.  I love Starbucks.  82.  I like comfy sweatpants.  83. High school is a melding of all the friends I\'ve ever had.  People from elementary school, middle school, and new kids.. pretty nifty. 84.  I hate the smell of fresh mulch.  85.  I love my family.  86.  I hate needles.  87. I am the least perfectionistic person you will EVER meet. 88.  I always wanted to learn to play the guitar.  89.  I hate the feeling of failure.  90.  I am still a virgin. Yeah, i\'d kind of hope so... haha  91.  I would love to have my own fashion line . 92.  I can be quite selfish.  93.  I still act like a little kid.  94.  I love Delia\'s.  95.  I love pictures.  96.  I love music.  97.  I wish I were more motivated when it comes to school.  98.  I love getting stuff in the mail.  99. I have problems forgetting people. 100.  I hate the feeling of being alone.  '),(19733,2190377,'25,January,2004','2004-01-25 00:00:00','   *felt like clearing out my profile*  it\'s nice having friends.  :)  That, and our basketball team is on a two game winning streak!!! w00t!.  You\'d have to be on our team to understand how amazing that is.  I wish I was good at basketball, but no.  No more basketball for me after Feb. 13, kinda sad.  Maybe if I MAGICALLY become good over the summer I\'ll play next year but I seriously doubt it.  My brother thinks I should do winter track next year.  My brother is the coolest person ever.  However, I don\'t think I\'ll heed his advice.  If I\'m this bonkers during sophomore year... next year could result in institutionalization.    Today I bought a CD at Mckay\'s, along with 4 or so gruesome vampire and/or  rampant infectious disease books.  CD = The Police - Every Breath you Take- \"The Singles\" *sigh* so good.  I find myself listening to a lot of singer song writer hooha lately, as well as waaay too much of The Cure, this band called Spoon that was on Rachel\'s mix for me, and random Christian music as introduced to me by Kelly.  The maturity level of my conversations recently seems to have increased ten-fold.  Change is good.  I thought I was one of those people that never changed but... thinking about it I\'m a little bit different every year.  You can tell a lot about the way I am at a given time (haha sounds like the existential question) by a) what clothes I wear, and b) what is written in my yearbook  7th grade: lots of icky random preppy t-shirts with big blaring TOMMY or DKNY on the front, comments about being nice but pale 8th grade: juxtaposition of cutesy floof and skankier fare leaning more towards the alternative spectrum of stereotypes, probably in hopes of fitting in with Ed\'s friends.  yearbook comments revolved around this unsaid conviction of everyone i knew that i was indeed on drugs 9th grade: lots and lots of cartoony shirts cute, nauseating, brightly colored (oh how i love it), everything in my yrbook was written by a boy and about me being a lesbian slut 10th grade: weird matchy matchy shtuff, with a penchant for anything off beat and/or striped, yr book comments will probably mention seizures and speshul pills  anyways, the point of all this is, I really want things to change.  a lot.  and i think they will.  will post more later  bye... heres my old profile    etiquette9 23: D is for the dagger i am going to stick in your heart etiquette9 23: I is for ill people, and you are going to be one if you piss me off some more etiquette9 23: and E is for every single fucking thing you have done that has sent my life downhill etiquette9 23: i hate you etiquette9 23: so just die  LedZeppelin1292: It just occured to me that almost all of the sports I\'ve done involve spandex  imnochewtoy: like on a scale of 1 to cute youre ashleigh   evilkennyoo 7: ive never had someone hop up on my shoulders and start rubbing her tounge in my hair  JerMan144: who else are you talking to? JerMan144: oh ... your project  '),(19734,2190377,'19,January,2004','2004-01-19 00:00:00','   IM SO HAPPY!!! NO SEIZURES FOR ME!!!!   This is the happiest I\'ve been in awhile.  For no reason other than DRUGS  omg  i hope its like this all the time.  '),(19735,2190377,'17,January,2004','2004-01-17 00:00:00','   spaceypiXiechick: heeey QueenOfHrts856: hey shleeeeigh QueenOfHrts856: haha spaceypiXiechick: hehe spaceypiXiechick: shleigh spaceypiXiechick: ahahaha :-) QueenOfHrts856: hah QueenOfHrts856: like sleigh :-) spaceypiXiechick: awesome spaceypiXiechick: =-O QueenOfHrts856: yeh QueenOfHrts856: like QueenOfHrts856: santa! spaceypiXiechick: im santa!!! spaceypiXiechick: people always sit on my lap and ask for presents QueenOfHrts856: hehe no santa has sex with you  Kelly rocks.  So does Maria.  Maria i love you so much.  Seriously.  You rock so hard.  So does your brother\'s band.  Too bad there\'s a ringing in my left ear now lmao.  I haven\'t decided if Friday\'s game has pulled us together as a team, or pushed us further apart.  Honestly I don\'t care that much, I just wish people would stop getting hurt (mentally physically etc)  Today I went nuts and kicked a hole in our kitchen wall.  Long story, basically my dad is an ass and made me shovel ALL of the dog crap in our front yard because he got pissed when he stepped in some.  We have FOUR dogs.  I almost puked.  What a jerk.  Then I came in and he made me move all these paint cans into the basement and I was just tired of doing crap for him.  So he went out into the garage and I got mad and kicked the wall.  REALLY hard apparently.  A little bit harder than I kicked Matt\'s shin, now there\'s a hole the size of my fist in the kitchen wall.  I guess Matt\'s shin musta hurt a bit. :)  Also basketball was terrible.   4 suicides 20 laps around the gym 16 sprints (8 with balls overhead) not to mention  stupid ass defensive positioning drills  the employment of BUBBA in our outlet-layup drill five-man from hell 2 minute layup drill...olympic... the whole nine yards. all in an hour and a half  why do i suck at life?  and why DONT hate Jeremy?   good question.  :)  Hating people is fun until you realize there is nothing to look forward to anymore  '),(19736,2190377,'16,January,2004','2004-01-16 00:00:00','   spaceypiXiechick (1:28:58 AM): im only pushy to males spaceypiXiechick (1:29:00 AM): because spaceypiXiechick (1:29:04 AM): they are the devi spaceypiXiechick (1:29:05 AM): l spaceypiXiechick (1:29:07 AM): and my grandmother JerMan144 (1:29:08 AM): uh huh spaceypiXiechick (1:29:12 AM): because i know i can get away with it JerMan144 (1:29:15 AM): yeah ... she\'s our leader I think JerMan144 (1:29:19 AM): oh ... hehehe JerMan144 (1:29:22 AM): that too spaceypiXiechick (1:29:25 AM): what?!!??!! spaceypiXiechick (1:29:27 AM): loollll JerMan144 (1:29:48 AM): yeah ... definetly wayyy too tired JerMan144 (1:30:02 AM): dude ... I squeeked JerMan144 (1:30:07 AM): =-O JerMan144 (1:30:08 AM): :-D spaceypiXiechick (1:30:13 AM): did u just agree to  being the devil and then declare my grandmother the leader of all men-devils? JerMan144 (1:30:20 AM): uh huh JerMan144 (1:30:22 AM): I think so JerMan144 (1:30:25 AM): ... JerMan144 (1:30:29 AM): oh! JerMan144 (1:30:31 AM): dude! spaceypiXiechick (1:30:36 AM): yuush? JerMan144 (1:30:46 AM): I\'ve got to show you all the quotes I\'ve gotten from english JerMan144 (1:30:50 AM): they\'re hymilarious spaceypiXiechick (1:30:51 AM): lol ok JerMan144 (1:31:01 AM): Mr. Ro is the funniest dude ever JerMan144 (1:31:05 AM): specially if you saw him spaceypiXiechick (1:31:39 AM): fuuuuuuuunny loooooooookin/ JerMan144 (1:32:02 AM): he\'s like a mix between a valley girl and a beatnick ... he\'s about 30 ... 6\'7ish ... lanky as woah ... and has the movements of a toy story character JerMan144 (1:32:47 AM): well ... maybe 6\'4 JerMan144 (1:32:54 AM): but yeah ... that\'s him spaceypiXiechick (1:33:01 AM): basically andrew bowman in 15 years?  '),(19737,2190377,'15,January,2004','2004-01-15 00:00:00','   Today was a terrible day.  Somehow, I don\'t care though. :) I talked from the beginning of lunch, to right before basketball practice.... That\'s three and a half hours.  It gave me a headache, though.  I don\'t feel like talking anymore.  So, be happy with what Rachel sent me while I was sleeping last night  imnochewtoy: ashleigh for you if I could I would be snow  Auto response from imnochewtoy: please snow  imnochewtoy: haha but you make me melt anyways imnochewtoy: omg I cant beleive I just said that  Rachels so cute :)  '),(19738,2190377,'14,January,2004','2004-01-14 00:00:00','   I have several amusing stories to tell.  Unfortunately I can\'t really remember all of them.    I DO remember me and Vidya being psychotic children babbling in fake swedish about swedish fish and practicing our retarded scooby doo laugh before mondays game... haaaaaaaah.  what a weird week.  OOH GOOD STORY  So I run into Octavian and Annie in the hall after Chem or something right?  And she waves and says \"Hi Ashleigh!!!\" and  THEN she waves to the passing Mr. Majeskie, who promptly gives her a really sour look. I start laughing hysterically.  Annie doesn\'t understand why.   You see, on her hand in huge bold letters is written, \"FUCK\" Her other hand says \"me.\" Apparently, it was Octavian\'s doing, but the fact that she didn\'t remember it and was unknowingly cussing out everyone she waved to that day makes me smile :) ok bad story.  it was so funny. you had to be there  evilkennyoo 7: aw man... most embarrassing moment of my life today evilkennyoo 7: at least between me and my mom spaceypiXiechick: uh oh spaceypiXiechick: how so evilkennyoo 7: i was eatin one of those yo crunch dealys (white yogurt) at the computer and i accidentally got some around the keyboard and i didnt realize it then my mom went down and was like \"kenny... i think you should clean up after yourself\" evilkennyoo 7: i thot i was gonna die   ALSO! Jose made my day.  That rhymed. :)  He gave me a mountain dew.  Now I am happy.  oh also  I give Rachel a ride home right? and she says to me as she leaves, \"Ashleigh you know how much I like you?\" me,\"how much?\"  \"a whole poptart much!\" and leaves a strawberry poptart on my dashboard. haaaaaah   -_-  I am obsessed with poptarts.  and carrots.  and trailmix. omg i could live on them forever.    do you like how i alternate between big words and unintelligible internet lingo?  especially when i randomly decide to capitalize/punctuate sentences  I have many many Jose quotes, but John Jarvis wrote em all down for me during HUM.  Don\'t worry, I\'ll get them  ASAP!!!!!  '),(19739,2190377,'11,January,2004','2004-01-11 00:00:00','   stolen from marias lj  1. Done any drugs? nope. ha, bet you didnt see that coming 2. Drank alcohol? yea but ive never been drunk :P 3. Had sex? with your mom? actually not 4. Fallen in love? yeah, sure 5. Fallen in love online? i wouldnt say IN love but i\'d say that ive grown to love people 6. Met someone from online? not in person, actually, i lied 7. Tried to set a bug on fire with a magnifying glass? naw.. i did other things though >=) like trying to run them over with remote controlled cars 8. Torn off a lightening bugs light and smeared it anywhere on your body? ick 9. Made a snow angel? yessirr.. last year :) 10. Dreamed you could fly? ALL THE TIME!!! well its not really flying.. its flying like an airplane, like i\'ll run super fast and then suddenly be in the air weee 11. Ever had a job, if so what? yessss.. just at my moms office and pet/baby/house-sitting 12. Bungee jumped? i wish, thatd be so hot 13. Felt betrayed? Were you able to forgive them? yes... and.. not usually. 14. Seriously contemplated suicide? unfortunatly yes 15. Attempted suicide? no because i know its dumb 16. Been in a car accident? nothing big 17. Had a close friend to die? my first real crush died. last christmas. 18. Had a close family member die? not a close one, most of my old family members died before i was born 19. Seen the eiffel tower? nope. 20. Seen the statue of liberty? who hasn\'t?  21. Told a secret you swore you wouldn\'t tell? yes. it\'s one of my biggest flaws, most of the time its not on purpose though 22. Stolen anything? naw 23. Gone out with someone for their looks? ehhhh.... kinda 24. Their reputation? no 25. Met anyone famous? maybe? i dont remember 26. Been on radio or tv? tv 27. Been on stage? of course 28. Intentionally hurt someone? yes 29. Dumped someone? yes 30. Been dumped? no 31. Been to Europe? yeah 32. Been convicted of a crime? no? 33. Been in a mosh pit? yeah but not purposefully.. i had bruises to prove it but i told my parents it was from brian lol 34. Been asked for an autograph? no... 35a. Been toilet papering? no b. Been toilet papered? yes. lol ... like ME not my house.. ive got pics to prove it c. been spud gunning? what the fuck is spud gunning? d. been cow tipping? nope 36. Loved someone so much it made you cry? yes 37. Cried over someone of the opposite sex? yes 38. Same sex? YES 39. Had toilet paper stuck to your shoe? i dont think so.. surprisingly enough considering my rep for idiocy 40. Been caught with your fly unzipped?ALL THE TIME.  im so bad about that hahha 41. Had your pants fall down in public? no, theyve been pulled down, though 42. Been in a fight? yes. 43. Cheated on someone? no 44. Hitchhiked? WITH THE PIZZA DELIVERY GUY BABY 45. Picked up a hitchhiker? no 46. Had your appendix and/or tonsils removed? dont think so 47. Gone skinny dipping? no and i need to. thatd be SO much fun hahahaaaaaaaaaaaaaaaaaaaaaaaaaa.  -_-    JerMan144: maybe I am just being a jerk spaceypiXiechick: dingdingding   >=0   nicofiened: im a rebel spaceypiXiechick: okee spaceypiXiechick: punk nicofiened: hodlum nicofiened: i prefer misdirected youth with anger thank you   nicofiened: you dont know what spud gunning is spaceypiXiechick: nope spaceypiXiechick: gtg nicofiened: you take pcv pipe put a spud in it on the other end you close it except for a little hole nicofiened: put hair spray in nicofiened: and light nicofiened: it launches it nicofiened: peace nicofiened signed off at 3:52:37 PM.   eric made TWO funnies today   heheeeeeeeeeeeeeeeeeeeeee *dies*  '),(19740,2190377,'10,January,2004','2004-01-10 00:00:00','   GoCombatChuckGo:  oh god  GoCombatChuckGo:  ew.  GoCombatChuckGo:  just. ew.  GoCombatChuckGo:  my brother left a booger on the mouse for me   funny. aaaaaahahha... :) there. a funny happening. wee  happy?  '),(19741,2190377,'04,January,2004','2004-01-04 00:00:00','   I haven\'t written for awhile... mostly because nothing notably good or funny has happened.  That I would want to share anyways... That, and people are tired of hearing me complain.  I get comments like, \"You\'re so fun to talk to when you\'re happy, cheer up.\"  This  could  be taken as a compliment, but instead I take it as me being nothing but a source of entertainment for other people.  It gets boring after awile, and annoying, because people think I\'m being weird when I\'m not bouncing off the walls and cracking 7th grade humor type jokes every 30 seconds.  Actually, if I don\'t do that it means I\'m super comfortable with you and you should take it as a compliment... but whatever.  I\'m so tired of miscommunications lately.  You\'d think that for someone who talks as much as I do, I would be correctly understood, but no.  Same for other people.  Anyways, sorry for being a bitch to a lot of people lately.  I\'m so apprehensive about my EEG results, I get them back next Monday.  This could change my life guys.  Hopefully he doesn\'t want to run any more tests.  I\'m so tired of tests, in all facets of my life.  Someone told me they thought I\'m stuck up.  That makes me sad.  Especially because I really have no self-esteem at all.  Because, just like I told Jeremy once, I\'m not a pessmist or an optomist, just a realist.  I know what I am and am not capable of.  Or at least I thought I did, until recently when I began questioning the little intelligence/athletic ability/people skills/physical attractiveness I  thought  I had.    Okay I\'m tired of writing and this probably isn\'t too much fun for you either.  '),(19742,2190377,'04,January,2004','2004-01-04 00:00:00','   100 degree fever. AGAIN. WTF  SCREW YOU IMMUNE SYSTEM *pouts* did this NOT just happen a month ago? go to hell!!! roooaaar  '),(19743,2190377,'03,January,2004','2004-01-03 00:00:00','   more shameless quoting.  but boy is my brother weird :) lalala. hes the bestest.  VTHokie0X: my new years sucked ass spaceypiXiechick: yea :-P VTHokie0X: i had to file a missing persons report on jeff spaceypiXiechick: dude VTHokie0X: on new years VTHokie0X: i thought he was dead VTHokie0X: but no VTHokie0X: he is a drunk ass  *later*  spaceypiXiechick: be good!! VTHokie0X: haha VTHokie0X: 3 people have used that VTHokie0X: as their last IM to me VTHokie0X: that\'s funny    ..... i have no life  and no hair  '),(19744,2190377,'02,January,2004','2004-01-02 00:00:00','   imnochewtoy: I just think Im clever spaceypiXiechick: maybe ur psychic too spaceypiXiechick: we should start a club imnochewtoy: Im still mad that you allready knew what I was going to say imnochewtoy: am I really that predictible spaceypiXiechick: actually ur very unpredictable spaceypiXiechick: but have certain catch phrase type thingees :-) imnochewtoy: aww imnochewtoy: ashleigh you are my catch phrase :-) spaceypiXiechick: imma take that as a compliment :-D   too lazy for anything else.  '),(19745,2190377,'23,February,2004','2004-02-23 00:00:00','   Today, Kelly, Andrew, Kenny, Matt Mahoney, and I went to see \"The Passion of Jesus Christ.\" heh. Zeke is cute, \"oh you\'re going to see Jesus??? *goofy look*\"  kids are funny.  I felt strangely anti-social today.  What do people talk about? I can\'t remember.  Weeeiiiiirrrrrrrrrrrrd.  Although last night I was all hallucinate-y and crazy.  Hearing noises that were\'nt there and whatnot.  I think something is wrong with me. Maybe I\'m just becoming a hypochondriac.  Anyhoo! Yeah.  What a good movie.  I cried for about two-thirds of it.  Which is ... a lot.  for me anyways.  I don\'t cry a lot... especially in public, except for that one day when i went kinda psycho.  :/  I\'m just kinda in awe.  wowowowow.  I watched EVERYTHING I\'m proud to say.  Horror movies are good training I guess.  ummm yeah.  Matthew Mahoney is definitely the most friendly person I have ever met.  ever. :) I wasn\'t my normal eeee look at me I\'m an effervescent ball of fluff today and he was still SO nice.  Not so much today but just my general impression of him.  Have you ever met one of those people that always seems happy and just.. nice?  Like they\'ll smile at you even if they don\'t know you.  Or if he doesn\'t look HAPPY he at least looks pensive.  I wanna be one of those people.    I have a few funny stories.  They aren\'t really postable, though.  Besides my psychotic behavior of last night anyway.  I\'ll write that later but... it\'s kinda embarassing.  By the way, Maria is black.  That girl knew all of those BSU questions.  Insane.  The only one I knew was the one about grammy awards winners or something (halle berry and denzel washington hehe)   So I was reading about the people in the movie I just watched, and was shocked to find out that the guy who played Jesus is also the guy who played the younger son firefighter man in Frequency.  Weird huh.  This is such a sedate... fragmented entry.  It\'s pretty much how I feel though.  Time to study chemistry, but I\'m soooo distracted.  '),(19746,2190377,'17,February,2004','2004-02-17 00:00:00','   HAha.  urlLink Random Slogan Generator   I Think, Therefore Ashleigh.  Ashleigh Unscripted.  ^^ story of my life. anyhoo!   Because Majid decided to be Mr. Ironic and point out the irony in the fact that I rant at people about their spelling but am unable to capitalize or punctuate, this entry shall contain actual English.  Vidya managed to convince me to work out with the soccer girls today. Unfortunately, now their coach thinks I\'m coming out.  If you had gym with me, you would know just how incredibly amusing this is.  We ran two laps around the school for a warmup and then over to the friendly neighborhood 7-11.  It was about three miles.  For someone who has been playing basketball since early November, three miles is pretty long.  However, because we rock so hard, we finished at about the same time as Abby Schuhart (probably one of the most athletic people I know.)  Then we went and stuffed our shirts with hot dogs at the varsity game. It was quite sexy, but because I am sporting the \"I-hate-men\" haircut no one seemed to notice my huge hot-dog shaped chest.    ok enough of this punctuation business it makes me wanna tear someones head off lol. weeeeeeee. ok thats out of my system.  now i have to make up some shit for english. we\'ll call ^that paragraph a warmup. yesh. warmup.   i forgot to take my pills today.  i fell asleep in like every class and had to drink two cokes just so i didnt pass out.  then my hands started shaking really bad in german. and my pants fell down a lot.  when we had to sing head shoulders knees and toes... wow.  bending over is bad.  doodododoooooooooo.    the darkness rocks.  there was something else i wanted to say. oh well.  doodleeedoooooooooooo  '),(19747,2190377,'15,February,2004','2004-02-15 00:00:00','   that list was for people who rocked friday. oof. remind me not to make lists like those again lol. karin and carisa you really do rock. for real.  much love  '),(19748,2190377,'03,February,2004','2004-02-03 00:00:00','   Long time, no posts.  Maybe because Carisa made me a a  urlLink xanga  or maybe because I have nothing good to say.   This v-day wasn\'t that bad.  Actually I think I got more stuff than last yr... lol :P and the dance was alright, also we kicked SO MUCH ASS YESTERDAY!! i almost had a heart attack that game but we pulled through.w00000t.  I have a blow pop. mmmm. and reeses, and chocolate brownies and mocha ice cream mmmmmmmmmm. Not that I\'m being indulgent or anything.  Nope.  My weight fluctuates so much it doesn\'t matter though. la. Um, so yeah. I bought a hair straightener and dyed it brown, even tho it came out red.  I think i have hair ADHD.  For reeeeeeeeeal.  Cause even though I dyed it like 3 days ago im already like.. umm.. blonde crop! weeeee. WHY AM I SO DUMB?!?! i tried to get my permit but forgot to bring my social security number thinger. grrrr. oh well.  blarrrrr.   i wish i could grow, so i could buy an entirely new wardrobe because right now i have so many clothes i cant buy anything.  i was going to write a deep entry and then decided against it because honestly who cares.  our freshmen are so hot. SO HOT. please someone marry me.  maybe i should actually talk to them. maybe. what am i even saying right now. ill actually say something real later.  except maybe... lets make a list of people who rock really hard recently:  rachel jose andrew... even tho hes a big dork joceline ella VIDYA and erin because she actually has boobs lol and maddie because she gave me food  not that everyone else doesnt rock.    because everyone does.  really really  i wrote a story about a girl named minxie. who the hell is named minxie? i dont know where i got that from... but either way if i have two girls, one is gonna be kiki and the other is gonna be minxie, and the boy will still be something completely normal.  omg maybe ill name him harry. Then I wrote this.  \"Art imitates life.  Or so they say (I think.)  But in some cases life imitates art.  And in some cases the art isn\'t necessarily good, clean, or even remotely pleasing.  You could say that television is a form of art.  For some people, their life\'s art is more of the basic cable variety.  Take Jerry Springer for example.\"  That\'s about all that came out of an hour and a half of hum class, well that and minxie.  and me coming writing really weird metaphors involving dogs, raccoons, and garbage cans.  I can COMPLETELY incapable of writing fiction. what a shame.  and deep breathing. man mrs robbins and her yoga really rock.  rachel miller is insane. :) but she should take me home with her one day cause i feel bad about never seeing her.      im going to stop writing for fear of never shutting up  '),(19749,2190377,'03,February,2004','2004-02-03 00:00:00','   More of this bolding business.  This time.. it\'s movies.   1. X-Men    2. The Craft    3. X2 : xmen united   4. Swimfan   5. Fellowship of the Ring  6.  Finding Nemo   7.  Peter Pan   8.  Home Alone   9.  Aladdin   10.  The Ring   11. 10 Things I Hate About You   12. Not Another Teen Movie  13. Spice World  14. 8 Mile  15.  Bambi   16.  Pirates of the Carribean   17. Edward Scissorhands  18. Stepmom  19.  My Best Friends Wedding   20.  101 Dalmatians   21. Scream  22.  Scream 2   23. Scream 3  24.  Big Daddy   25.  Billy Madison   26.  Chamber of Secrets   27.  Sorcerer Stone   28. Heartbreakers  29.  Dumb & Dumber   30. Cruel Intentions 31. Scary Movie  32.  Look Who\'s Talking   33.  Blade    34.  Blade II  35. O  36.  Titanic   37. Carrie  38. Carrie 2: The Rage  39. Daddy Day Care  40.  Legally Blonde   41.  Austin Powers   42. Storm of the Century  43.  Oliver and Company   44.  The Two Towers   45.  The Return of the King   46.  Mighty Ducks   47.  Fast and the Furious   48.  2Fast, 2 Furious   49.  A Walk To Remember   50. TripleX  51.  Beauty and the Beast   52. I Know What You Did Last Summer  53. I Still Know What You Did Last Summer  54.  Sound of Music   55.  Mary Poppins   56. Tuck Everlasting  57.  The Patriot   58.  The Wizard of Oz   59. Teaching Mrs. Tingle  60. Crossroads  61. Now and Then  62. Pearl Harbor  63.  Just Married   64.  Cast Away   65.  Radio Flyer   66.  Final Destination   67.  Lady and the Tramp   68.  Shallow Hal   69. 40 Days and 40 Nights  70.  Bring It On   71.  Willy Wonka and the Chocolate Factory   72.  The Outsiders   73.  The Matrix   74.  Perfect Storm   75.  Austin Powers 2   76.  Never Been Kissed   77.  Clueless   78.  Bruce Almighty   79.  Remember the Titans   81.  Girl, Interrupted   82. SWAT  83.  Sixth Sense   84.  PhoneBooth   85.  The Lion King   86. Urban Legend  87. Nightflier  88.  Lion King 2   89.  Little Mermaid   90.  American Pie   91. Center Stage  92.  Scooby Doo   93.  Bedazzled   94.  Mrs. Doubtfire   95. Save the Last Dance  96.  My Girl   97. American Beauty  98.  Romeo & Juliet (modern n old school versions)   99.  Lost World   100.  Casper    I just counted.  Seems to be I\'ve seen 69 of em. HA.  I wonder how many wasted hours of my life that is.  '),(19750,2190377,'03,February,2004','2004-02-03 00:00:00','   QueenOfHrts856: byyyyye ashleigh cake QueenOfHrts856: thats your new name spaceypiXiechick: hehe spaceypiXiechick: awesome! QueenOfHrts856: and no, no significance QueenOfHrts856: i guess it could be lke QueenOfHrts856: patty cake QueenOfHrts856: only a QueenOfHrts856: ashleigh QueenOfHrts856: cake spaceypiXiechick: :-) QueenOfHrts856: sorry im gonna have to eat you.  '),(19751,2190377,'21,March,2004','2004-03-21 00:00:00','   weeeeeeeeeeeeeeeeeeee.  Well. Yesterday I slept over at Karin\'s and stayed up till 5:30 AM talking to her and jackie about.. well thats a secret.  weeeeeeeee.  Her birthday is tomorrow!! And it\'s a short day! And my gramma and i are gonna pick her up from robinson! I am haaaappy.  And tired. Very tired.   \"Did you drink your Budapest?\" \"Joey needs a volunteer\" \"I need to adjust the bean bag.\" \"Gangrene!!!!\" \"It\'s really more like a grey area...\" And then there was Karin\'s impression of a sheep, which really was just her trying rreally hard to touch her toes... i was like \"yoga???\" holy crap. you woulda had to have been there.  there was a lot of other funny stuff too but i have no friggin clue, i mean... i had ... a toffee nut frappuccino, five mountain dews .... er \"mountain lightning\" lmao... and a gross net of five pieces of pizza.     im working on  areally long thing to go in here. be excited. its exciting. really.  '),(19752,2190377,'04,March,2004','2004-03-04 00:00:00','   Noticed I haven\'t written anything in this for awhile. hrmmm.  WOW I\'m sore.  Really sore.  My legs are starting to buckle when I walk even.  Forget walking up the stairs.  I\'m really pleased with myself though, because slave driver Coach P is pleased with me and maria (s. yes. another. i know too many marias). Yeshyeshyesh.  Yesterday we ate a milkbone.  :)   We found it sitting in the track lobby.  That pretty much summarizes my mental state right now lol.  Endorphins are good.  I have lots of story but don\'t feel particularly inclined to tell them :P  There is one at the back of my mind, and I can\'t think of it.  That angers me.  Some silly kid making a really amusing slip-up.  Speaking of slip-ups I think I\'m going deaf.  So today after track Megan DeBettencourt told me she liked my outfit or something. Somehow I heard, \"What\'d you guys do today?\"  so I said, \"hurdles.\" haaaaaaah  \"I like your outfit today.\" \"Hurdles.\" wow.  thats it.  wussy i know just go look at the xanga theres loads more than youd ever want to know about me in it.  since i know you stalk me  '),(19753,2190377,'26,April,2004','2004-04-26 00:00:00','   where to begin.. where to begin. i\'ll start by documenting relatively amusing convos.  xXrejectdtearsXx: i love ashleigh xXrejectdtearsXx: in a straight freaking way xXrejectdtearsXx: NO xXrejectdtearsXx: in a straight nonfreaking way. xXrejectdtearsXx: i cant win. spaceypiXiechick: :-)  spaceypiXiechick: i dunno its creepy that u and ur brother would flirt with the same girl MBC2010: that what happens when u r 3 yr\'s appart MBC2010: it\'s never happened b4 spaceypiXiechick: wait so hes a 7th grader and he flirted with a 9th grader?! MBC2010: u don\'t compete with ur brother?  if you speak german.. im sorry. lol and nora prolly wouldnt want me posting this but eh thats ok  virgogurl4: woof woof spaceypiXiechick: meeeow virgogurl4: jose ist heiss und er ist gut mit seinen Finger.   um thats enough. lala. ok so ive decided i want a tattoo on the inside of my lip.. one on the back of my neck (barcode a la dark angel :P) and one in that area by ur hip that dips. yaaay.  hah. except not because i actually dont like tattoos.. but mostly cause they stretch! anyway. im also in need of a bellybutton ring.  and eyebrow ring.. or anti-eyebrow. too bad you cant have those in the corporate world :( *nods* yeah... dont ask i have some weird obsessions.  Today I was in a bad mood.  Discussions with Jose, however, made me realize that the quality of my life is generally dictated by the quality of my relationships.  This is a very bad thing because I have many an unhealthy relationship.  Sitting in the car on the way home from track, I realized that there are so many things I absolutely adore that have NOTHING to do with shitty friends.    DISCLAIMER: This is a really long, boring list.  It makes me happy though, so you can just eat me.  I love... Running in the rain.   Waking up with a dog in your arms. Waking up with a dog next to you in a makeshift noose (curtain string thingee) Candy corn, conversation hearts, pixi stix, and Dunkaroos frosting. Jumping. Up and down... Off of things... To get a rebound...On people... Into sand pits... And hanging with the crazy retardedly cool jumper kids. Miniskirts. Wheat thins and four leaf clovers given to you by otherwise stiff coaches. Closets. Finally getting your ears pierced at age 15. Hot chocolate at 6 AM.   Waking up at 3 AM to get a glass of orange juice.  Believing that your TV is going to murder you and that the fridge is possessed(like in the Ring) once downstairs. Writing terrible poetry. Writing narcissitic blog entries. Writing. Being a girl with chickenscratch handwriting. Dyeing your hair fifty-million different colors... only to have it turn out red every time. No one understanding you--yourself included. Reading trashy teen magazines in purple tinted bubblebaths. Indie movies, chick flicks, bad horror movies, good horror movies, raunchy comedies, and ZOMBIE movies. Laying outside and just staring up at trees until you forget you even exist. Halloween. Listening to really terrible music and loving it.  Country.  Spice Girls.  Techno.  Cheesy 80s rap.  pOp-PuNk. EMO! Trying to bake people yummy cookies and failing miserably. Feeding people in general. Weight lifting. Having an awesome brother you can brag about. Having a ****ing psycho family in general. Thrifting. Subway workers knowing exactly what sub you want. The smell of track.  Not the lobby, though. Shoes.  Kitten heels, ugly shoes, mary janes, sneakers, track spikes, flip-flops, slippers.  Stealing your mom\'s shoes. Nauseatingly bright colors.  Preferably combined with an equally nauseating cartoon character. Pot-bellied pigs, teacup poodles, and huge slobbery Great Danes. VH1. Biology.  Discovery Health Channel.  Neurology.   Knowing that you\'re going somewhere.  In life that is.      I like how that list really turned into a train of happy thoughts.    also, HAPPY BIRTHDAY ELLA!!!! Well, belated.  Girl... you rock. :) :) :) Also, you\'re probably the only person i would ever say anything as retarded as \"girl, you rock.\" to.  which is good maybe? someday i will write more about ellas coolness.  '),(19754,2190377,'21,April,2004','2004-04-21 00:00:00','   AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAH. so scared. :-/  spaceypiXiechick: why am i such a pussy spaceypiXiechick: actually spaceypiXiechick: why am i the opposite of a pussy arvenmic: you\'re not butch spaceypiXiechick: i almost am spaceypiXiechick: if it werent for hte boobs arvenmic: nah, no way arvenmic: you reek, uh, feminine spaceypiXiechick: LOL spaceypiXiechick: thanks arvenmic: np arvenmic: and it\'s a compliment   apexfb54: what me being the lead screamer of a satanic band apexfb54: i could do that apexfb54: minus the satanic part apexfb54: be like apexfb54: yelling * happy shiny people happy shiny people or something like that apexfb54: that would be pure evil to evil people  uhm  yeah.  isdlkug \'yhik72 5\'2#TIHN  '),(19755,2190377,'19,April,2004','2004-04-19 00:00:00','   At the request of Jose I\'ve decided to describe my love life... my horoscope pretty much sums it up. sigh.  On the love front: You have the attention span of a gnat, so anyone who isn\'t on the ball is kicked to the curb before you even give \'em a fair shot. That\'s how you weed out the faint-of-heart, though - when you find someone who sticks in your mind, you know that sweetie\'s here to stay.   But honestly.. I\'ve kinda given up.  Besides track is much more interesting. to me...   For some reason no one else cares. heh. If I\'ve been babbling to you about it a lot lately I apologize.  but you gotta understand... im going to penn relays!! holy shit!! im gonna get beat up for it because I took the place of Appie i think... oooh man. shes really mad and really good but the coaches think i deserve to go because i work hard (translation: i have no life and always show up to practice.) Either way, yours truly is going to run in a 4X100 and a 4X400 this thursday.  does anyone know how hard running a 400 m dash is? holy shit. AAAAAAGH. SO FRIGGIN HARD. i wanna puke just thinking about it... but im so excited, just sad ella cant go :( what a cool kid. for serious. we had fun on the field trip today, yay for looking at carved bones and being watched by creepy 30 year old men of indeterminate sexuality!! and getting SUPER BURNED. im a redneck... and i mean it.  guys im so excited. now all i need is some action.. and another spring break and a pony.. and a new haircut. and a new HAIRCOLOR agggh.   also nora and jose are so cute. :)  matt isnt as much of a jerk as i thought.. hes just... ehhhhhh lol not to smooth when it comes to making people feel better i guess ?   and kelly and i are going to write a book about how to exploit boys. oh what an exciting life.   i could just pee my pants. for serious. okay maybe not.  also if i didnt mention ur name please dont kill me. im thinking i should just assign people arbitrary numbers or something.. \"so friend A was saying to me blah blah blah\"  '),(19756,2190377,'12,April,2004','2004-04-12 00:00:00','   survey-whore!!! :)  bold the one\'s that apply to you. ok so i was born in 88 but this is still fun  You know you were born in the 80\'s if...   1. You\'ve ever ended a sentence with the word \"PSYCHE.\" 2. You can sing the rap to the \"Fresh Prince of Bel Air.\"  3. You wore biker shorts under your skirts and felt stylish.    4. You yearned to be a member of the Baby-Sitters club and tried to start a club of your own. 5. You  owned those lil Strawberry Shortcake pals scented dolls.  6. You know that \"WOAH\" comes from Joey on Blossom.  7. Two words: M.C. Hammer . YESSSssss  8. If you ever watched \"Fraggle Rock.\"   9. You had plastic streamers on your handle bars.  10. You can sing the entire theme song to \"Duck Tales.\"  11. When it was actually worth getting up early on a Saturday to watch cartoons.  12. You wore a ponytail on the side of your head.  13. You saw the original \"Teenage Mutant Ninja Turtles\" on the big screen. (and i own them all on VHS)   14. You got super-excited when it was Oregon Trail day in computer class at school.  15. You made your Mom buy one of those clips that would hold your shirt in a knot on the side....   16. You played the game \"MASH\" (Mansion, Apartment, Shack, House)  17. You wore Jordache Jean jacket and you were proud of it. 18. L.A. Gear 19. You wanted to change your name to \"JEM\" in Kindergarten.   20. You remember reading \"Tales of a Fourth Grade Nothing\" and all the Ramona books.   21. You know the profound meaning of \"WAX ON, WAX OFF.\"  22. You wanted to be a Goonie. 23. You ever wore fluorescent clothing.  24. You can remember what Michael Jackson looked like before his nose fell off.   25. You have ever pondered why Smurfette was the only female smurf.  cause shes a skank!!  26. You took Lunch Pails to school.   27. You remember the CRAZE, then the BANNING of slap bracelets.   28. You still get the urge to say \"NOT\" after every sentence.   how about.. except not? cause i still do that :) 29. You remember Hypercolor T-shirts. 30. Barbie and the Rockers was your favorite band.   31. You thought She-Ra and He-Man should hook up.   32. You thought your childhood friends would never leave because you exchanged friendship bracelets.  33. You ever owned a pair of jelly shoes.  the blisters... 34. After you saw Pee-Wee\'s Big Adventure you kept saying \"I know what you are, but what am I?\"  35. You remember \"I\'ve fallen and I can\'t get up.\"  36. You remember going to the skating rink before there were inline skates. 37. You ever got seriously injured on a Slip and Slide.  38. You have ever played with a Skip-It.   39. You had or attended a birthday party at McDonald\'s.   40. You\'ve gone through this list occasionally saying \"Totally awesome\"  how about.. ahahaha YES  41. You remember Popples.  42. \"Don\'t worry, be happy\"  43. You wore like, EIGHT pairs of socks over tights with high top Reeboks. 44. You wore socks scrunched down.  45. \"Miss MARY MACK MACK, all dressed in BLACK BLACK BLACK\" Or, Ms. Susie had a steamboat, the steamboat had a bell (toot toot) Ms. Susie went to heaven, the steamboat went to... (\"with silver BUTTONS BUTTONS BUTTONS all down her BACK BACK BACK..\" the steamboat went to hell-o operator, give me number 9!)   46. You remember boom boxes vs. CD players.   47. You remember watching both \"Gremlins\" movies.   48. You know what it meant to say \"Care Bear Stare!\" (CAREBEARS!!!!!!)   49. You remember watching Rainbow Bright and My Little Pony Tales.  50. You thought Doogie Howser was hot.  51. You remember Alf, the lil furry brown alien from Melmac.  i hated him >=0 creepy turd. 52. You remember New Kids on the Block when they were cool.  53. You knew all the characters names and their life stories on \"Saved By the Bell,\" the ORIGINAL class.  54. You know all the words to Bon Jovi.  55. Wearing scrunchies around your wrist was the cool thing to do.  '),(19757,2190377,'11,April,2004','2004-04-11 00:00:00','   I find this slightly disconcerting.  \"ashleigh, your score is 19% consistent with a normal personality profile. You\'re least like others in the area of gender identity. \"  my results from \"The Multiphasic Personality Inventory\" test lol i dont even know what that means but.. eek  also disturbing is my new quasi-hick persona.  AAAAAAAAAAGh  Over break I: Sat through two hours of Deadwood Went to see the Alamo w/ the asian posse Watched Nashville Star Raided my brother\'s MP3s and listened to sappy country love songs for 3 hrs+ Watched CMT  Okay so maybe it was because I was really bored and watched a lot of TV... but.... I dunno  Be afraid.  What should I do to my hair?  '),(19758,2190377,'09,April,2004','2004-04-09 00:00:00','   somehow i forgot my password or username or something to my xanga. oooh well.  um. mindless blither time!! blither probably isn\'t a word.  Directions: 1. Put your birth month in an entry. 2. Strike out anything that doesn\'t apply to you. 3. Bold the six that best apply to you. 4. Put all twelve months under a journal cut.  Being cheap with journal space... im just gonna link this kid. im not really sure who it is though.. i was just reading random journals. heh  http://www.livejournal.com/users/aarfan/43845.html#cutid1 probably this link wont work. if ur dissappointed.. then um...you can just eat me. also since this thingee doesnt have a strike through mechanism. imma italicize the business that has nothing to do with me. shit this just isnt working out is it?  MAY:  Stubborn and hard-hearted.  Strong-willed and highly motivated. Sharp thoughts.  Easily angered.   Attracts others and loves attention.  Deep feelings. Beautiful physically and mentally. Firm Standpoint. Needs no motivation.  Easily consoled.  Systematic (left brain). Loves to dream. Strong clairvoyance. Understanding. Sickness usually in the ear and neck.  Good imagination . Good physical. Weak breathing.  Loves literature and the arts . Loves traveling.  Dislike being at home .  Restless.  Not having many children.  Hardworking.  High spirited.  Spendthrift.   I bet you\'re wondering: So. How\'s life Ashleigh?  The answer is: I don\'t even think about it anymore.  My existence as a sentient being is over.  Except maybe now.  This spring break was pretty... uneventful, but not particularly bad.  I think my body has started a bad trend of getting sick over vacations, though.  Ickyickyicky.  My tummy still hurts.  I went to track on wednesday and thursday and almost vomited.  Howeeeever, I\'m happy I went because now I can 4-step!! yaaay.  That means nothing to anyone and anyone it does mean anything to would know that it\'s not a kind of square dance OR anything to be cheering about really but I don\'t care because I PR\'d in hurdles in gym shoes, with shin splints, two days in bed, and alien spawn growing in my tummy! After seeing Dawn of the Dead today I\'m more likely to believe that it is not alien spawn, but, in fact a zombie baby.  Disgusting movie.  Oh how I loved it.  I think the defining moment of spring break was the conversation that took place in Jose\'s kitched at like 9 P.M. on Saturday.  I learned so much in an amazingly short period of time.  Did you know of the existence of the jeanis? geanus? geanis? not genus... I don\'t know how to spell it, but I will forever be paranoid of my jeans now.  Thank you so much Mr. Jesse Markowitz.  :) Man this entry is a downer.  Sorry.  Just picture my grandmother impersonating short black bootylicious women in her foyer.  bwahahhaa.  I\'m so proud to related to her... kinda  '),(19759,2190377,'02,April,2004','2004-04-02 00:00:00','   this week is beyond words.  most likely the moodiest ive been ... well for a long time.  not as in the worst week ever(although it was close) just up and down.. a friggin roller coaster of emotions. nyah. i dont feel like talking about it though  '),(19760,2190377,'31,May,2004','2004-05-31 00:00:00','   So guys, I have a confession to make. Really I\'m not very sane.  And I don\'t mean this in a cutesy way either.  I am one of those people Harry makes fun of because they have so little connection with humanity that they go write in their xanga etc. in hopes that someone somewhere will give a shit.  So maybe I haven\'t complained on this too much recently.  That doesn\'t change the screwed-up-ness factor of my life.  And don\'t flatter yourself by thinking what\'s bothering me is something about you although you did assist in putting me down this fucking psycho path (oh wonderous Jeremy etc.)  Holy crap why are people such jerks.  On another note, what happened to my so called intelligence? Anyone seen it lately?  So maybe I can\'t get pissed at people when they unknowingly label me as a vapid bimbo.  HowEVER, I think I wouldn\'t be this dumb if there wasn\'t so much drama.  Go away drama.  I think everyone could help me out by quitting with the stupid stupid lies they constantly feed me.  Someday all of the people like me that sacrifice their self-respect in an attempt to amuse others are going to lose it and blow everyone elses fucking brains out.  I\'m thinking maybe I shouldn\'t post this.  But honestly, does anyone read it? No.  But I can pretend.  btw, I\'m not actually going to shoot people so dont go reporting me to the school or anything  also im not that depressed but felt like expressing anger at a few people  rawr  actually this weekend was nice :) but that doesnt change the all-encompassing theme that is this life.  enough of this emo crap  goodddbyeee  '),(19761,2190377,'27,May,2004','2004-05-27 00:00:00','   stolen from bo which was stolen from sisy. yay.   If I Were... A month: march, because does anyone understand march? hell no A day of the week: thursday  A planet: um, earth? A direction: south A piece of furniture: a loveseat A sin: ummm.. whatever that circle of hell is that has gluttons and lusty peoples A historical figure: as if i know any... psssh. how about teddy roosevelt, cause he got to have teddy bears named after him. tooooo coool A liquid: tears. *sniffle* or koolaid. mmmm A tree: one of those droopy ones A bird: bluejay A tool: if i had a haaammer, id hammer in the morning... dadada. although i think a screwdriver would be more appropriate A kind of weather: not hot or cold and a tad windy, partly cloudy A mythical creature: dragondragondragondragon omg obsessed A musical instrument: harmonica An animal: a cat, because then everyone would spoil me and i could be as standoffish as id like A color: yellow A vegetable: CARROT! A sound: cicadas... jk. how about windchimes An element: helium >=) A song: video killed the radio star A movie: dazed and confused A book: thats tough, i only read scary books really eek.   A food: ice cream A place: a field somewhere quiet A material: polyester heh A taste: so sweet its almost unbearable A scent: pheromoooones. hah A religion: something thats more spiritual than organized religion-y A word: awesome :P An object: stuffed animal A body part: eh... imma be unoriginal and say eye like bo A facial expression: happily spacey  ill write actual stuff later. dooo dododo  '),(19762,2190377,'27,May,2004','2004-05-27 00:00:00','   Get to know the REAL you by  urlLink  crash_and_burn        Your Name           You Are A:    Rude Boy      Your Favorite Band/Song    Garth Brooks - Thunder Rolls      You Like To Read:    Cheap porn      You Firmly Believe In:    Free love      Everyone Thinks You Are:    The coolest person in history      You Were Conceived:    52nd Street, New York City      You Will Marry:    Britney Spears             Created with the ORIGINAL  urlLink  MemeGen  !       hopefully that works. i really like this one :) even tho these things are generally dumb  so  went to regionals, which is awesome sucked at regionals, which was bad  i miss track already. and my seniors :( what cool seniors man   so the other day at lunch we were discussing how far you would go in a relationship and harry says  \"Personally I would never give head in high school\"  i love that kid  '),(19763,2190377,'25,May,2004','2004-05-25 00:00:00','   guess what! im 16. yay. what a dork. and my family suuuucks about bday presents. sigh. dadadadadadadadadadada.  um yeah im really weird cause its 11 and i just woke up from a nap.  yes i know that makes no sense. geez this is sad, i have no life to report on. or rather i have no life id like you to learn of.  so, regionals thursday  so, ashleigh is a cripple  have you seen me walk?!?!?!  i might as well tell you what happened this weekend  friday: after track went driving around with matt killing time till we could meet rachel for dinndinn.  unfortunately the flowers he got me had sex with my boob.  pollen... aaaah on pretty purple shirt :(  sigh.  the flowers are nice, though. sooo we went to pacific cafe and it was yummy... virgin pina colada mmmm. then my house and weirdness. much weirdness.  really it was like lets make fun of ashleigh! then we went outsite and played with the frogs and stuff in the lightning. omg i love lightning.  saturday.. karin! bwahaha. ew i wont even tell you what we did. too much food and a four hour movie. karin is the bestest yay.... the pizza guy thought i had a tattoo. really i had just drawn on my hip with sharpee. and then he was like.. ohhh good i was gonna say tahts the most messed up tattoo ive ever seen. :(  moral of the story = if ur drunk or somethin, never let me tattoo you. just dont  sunday i went to friendlys with ed. more ice cream. eeee!!!!! his car is SO COOL. purple neon with biohazard seat covers and a really hot stereo system. im thinking of jacking it ala gone in 60 seconds style holy shit i just start spelling in german of course you cant know that because of the magical backspace button but  style= stil  and yeah.  SHUT UP ASHLEIGH!!!  '),(19764,2190377,'16,May,2004','2004-05-16 00:00:00','   i found this in the depths of my computer. sooo bitter. i make myself laugh :) it isnt finished but im too lazy to do anything about that    *entirely new topic*  You know those quizzes you take that are like, \"what\'s YOUR flirting style?\" Well, they\'re crap.  I\'ve decided to write profiles for each of the respective girl flirting stereotypes.   NOTE: I\'m not judging. Just merely poking fun at my gender.  Because as much as I hate to admit it, I do it to.  The \"Tee-hee Look at me I\'m so Incredibly Helpless\" Approach. These girls often begin with simple enough requests, \"I\'m stupid, can you help me do this problem?\"  while batting their eyelashes feverishly.  Later it ends up being the boy doing her homework,  the boy buying her lunch,  the boy waiting patiently as she fake-bawls on his shoulder. Run away while you still can.    The \"I am Perpetually a Five-Year Old and Shall Throw Rocks at you Until you Agree to Kiss me on the Swingset\" Approach Pretending to hate someone.  What a quick way to their heart.  Usually only high school aged boys pull this one, but occassionally you run into a girl that does it.  They\'re the worst.  For serious.  Signature moves include: punching, kicking, throwing, stealing, and then running away giggling nauseatingly.  The \"If I Don\'t Pay Attention to you, You\'ll want me\" Approach Sometimes, people will make a large effort NOT to talk to you, just to make you jealous.  This is surprisingly effective considering the incredibly flawed logic behind it.  '),(19765,2190377,'02,May,2004','2004-05-02 00:00:00','   Disorder      Rating        urlLink Paranoid :    High       urlLink Schizoid :    Low       urlLink Schizotypal :    High       urlLink Antisocial :    Low       urlLink Borderline :    Low       urlLink Histrionic :    High       urlLink Narcissistic :    Moderate       urlLink Avoidant :    Moderate       urlLink Dependent :    Low       urlLink Obsessive-Compulsive :    Low       --  urlLink Personality Disorder Test - Take It!  --        i wonder if that works. im so computer retarded  '),(19766,2190377,'02,May,2004','2004-05-02 00:00:00','   so its been awhile since ive posted anything. too busy really... i guess.  not really in the mood to share the details of my oh so interesting life with the public mayhaps.  it actually has been interesting lately though, and not always in a good way but i can deal. maybe. except for the losing 200 dollars worth of shoes part.  mmm mm m mmmmm.  so the cicadas are here. sounds like a really bad science fiction movie, cause theyre offf in the distance trying to get me agjaakgbjah!!!! cicadas + districts = crunchy hurdling. hopefully my body stops hating me by wednesday... and if not i dont care because my birthday is friday!!! yyayayayayayayayay.imma be SUCH A GEEZER. ewwwwwww.   also, im the proud mother of a fish!! actually, probably god-mother.  too bad he will probably die over the weekend. :( poor JAAAM.  my mommy got mad at me when i told her about him hehehehe.... oh gorsh.  funniest pick-up line of the week  fish-clerk: \"Oh, those ugly ones up there are the females.  It\'s only with humans that the females are more beautiful *shoots jess and me skeezy look*\" ahahaha.    so what happened was, actually no i mway too friggin lazy.  :)   HAH.    omg my love life is soooo confusing!!!!! HGhagkjh!!!!!!!!! *bangs head on desk*  '),(19767,2190377,'02,May,2004','2004-05-02 00:00:00','   :)  '),(19768,2190377,'29,June,2004','2004-06-29 00:00:00','   *Stolen from anita* also i love anita. yay. even though really i dont know her. and why did i steal it from anita? because andrew mentioned her and i was like ohhh man i LOVE ANITA. and he sent me a link to her livejournal.wee.   I\'m so HARDCORE  1. studded belt - 2 points: 2 2. dyed black hair - 2 points: 0 3. thick rimed glasses - 2 points: 0 4. tight shirts/pants - 2 points each: 2 5. gaudy belt buckle - 2 points each: 2 YEAH MAN. :) 6. clothing bought from a thrift store - 2 points each: holy crap that would give me fifty gazillion points. i will givemyself 8 7. having hair with bangs longer than the rest - 2 points: 0 8. trucker hats - 2 points: 0... i have hats but no REAL trucker hats. 9. messenger bag - 3 points: 3 10. livejournal/myspace/friendster account - 2 points each: 2 11. saucony/new balance/ converse shoes - 2 points per pair: 10. aieee 12. mountian climbing key thingy - 2 points: 0 13. X\'d up mosh gloves - 10 points:0 14. scarf - 2 points: 2 15. refering to bands as acronym - 2 points:2 16. vegan diet - 5 points: 0 17. vegitarian diet - for shame, no points: 0 18. food not bombs participation - 5 points: 0 huh? 19. straight edge - 5 points: 2.. lol. i deserve some pts. 20. aspiring photographer - 3 points: 3 21. using adj from the late 80\'s/ early 90\'s (i.e. rad, gnarly, rockin\', etc.) 1 point for each word in vernacular: haha \"AWESOME!!!!ehhh... rock. some other ones. totally.... etc. approximately 5 22. pins/buttons 1/2 point each: 4 23. plugs/body piercings - 2 points each: 4 if you count each of my earrings as one 24. vinyl collection - 10 points: 0 my dad has one i sometimes listen to but im not that cheap 25. moldy peaches fan - 10 points: 0 26. liking metal seriously - deduct 5 points: 0 27. liking metalcore - 5 points: 0 28. reading books over 300 pages long - 3 points: 3 39. riding a bike - 5 points: 5 30. participating in \"the mosh\" - 5 points: 5 31. tea aficianado - 5 points: 0.. i like coffe 32. writing poetry - 5 points: 5 33. attended 50 shows or more in a year - 2 points: 0 34. attending hell fest - 5 points: 0 35. being in a band - 5 points: 0 36. working at an indy record store/health food store - 5 points: 0 37. clapping durring midpaced part of a song - 5 points: 5   0-24 - poser! get out or be forced out. 25-44 - beginner. attend more shows, buy more records, you\'ll be fine. 45-74 - scenester. you\'ve been around awhile and you know what\'s up, good job. 75 - or more - GOD! i bow to thee with your knowledge of everything hardcore.   according to this i am hardcore. somehow i am very amused  '),(19769,2190377,'22,June,2004','2004-06-22 00:00:00','   urlLink lmao. omg. im sooo hot. -_- hehehehe. i really like mud!!!!!    other pictures from the party: (thank you helen for actually taking pictures...) heh.  urlLink ALSO CHECK OUT MY GRANDMOTHER  '),(19770,2190377,'19,June,2004','2004-06-19 00:00:00','   Today my dad made us clean for like four hours.  That sucked a whole lot.  Anyway, I found this in one of my journal thingees from seventh grade hehe.    \"hot pink and black toenails beneath my fucking preppy shoes it\'s hard to live this way confused about who i am and who i want to be unsure of what it is that others see try to be a punk, not a wannabe it doesn\'t quite work cause im only thirteen\"  i was so... ahahhaha. wow this is an embarassing post. I dont even need to point out the... aghajgah in that so you can just laugh at it.\"  MOVING ON!    SCHOOL IS OVER! I am excited but also apprehensive because it is the first summer I don\'t have someone to latch onto everyday.  Where is my siamese twin? aaieeee.   So yesterday was the party. It was rockin. :) :) :) Thank you friends I really love you a whole lot!!!!! Sorry for being loud.  Too much caffeine... or something.  I think the best was when I tried throwing a water balloon at matt, but completely overshot and fell backwards in my lawn chair, proceeding to roll around in the grass laughing hysterically.  Or maybe when Maria or someone dared me to \"hump the stump\" and i started to and hit my crotch REALLY hard on a root.  Or MAYBE mud-wrestling maria and having my mommy yell at me :(  I am never ever EVER running in a swimsuit again.   baywatch... aaaaah.  the water balloons with maria and joceline.  using bio terms to describe their um... special shape.  globular!! haha...   okay enough of that. :) dadadadadadada I got some pictures developed and will scan them soon.  Ask me if you wanna see some...  i actually dont have anything interesting to say. that is rather depressing. ive been writing too much in this lately anyway  '),(19771,2190377,'15,June,2004','2004-06-15 00:00:00','   spallderman: o know!!! spallderman: lol spallderman: whoops spallderman: i meant  . . . o nl!!! spallderman: damnit spallderman: i mena spallderman: mena spallderman: mean spallderman: o no!!! spallderman: shit that was badly done  ahahahahaha. oh harry. so silly. :)  today was SO FUN. sorry for being psycho guys  JwKbM: im Jung JwKbM: im not male or female JwKbM: im juss Jung... spaceypiXiechick: the third sex JwKbM: fo sho JwKbM: ;-) JwKbM: drum roll plz spaceypiXiechick: *beats on desk* JwKbM: lol im 20% a man JwKbM: duz dat make me 80% gurl?  spaceypiXiechick: ive done a belt once LedZeppelin1292: holy crap LedZeppelin1292: thats HOT  ^^ ur prolly best off not knowing heh  '),(19772,2190377,'15,June,2004','2004-06-15 00:00:00','   MBC2010: u easily fall into the category of MBC2010: \"most confusing girl i\'ve ever met\"  Mr Wizard 8891: sry, usually i try to play along, but a guy is just not a sexual thing Mr Wizard 8891: personally, i wish all my guy friends turned into furbys Mr Wizard 8891: that would be so much cooler than boys  *felt like clearing out profile*  http://www.collegehumor.com/?image_id=52648  http://bunnysnoog.cyborgcow.net/index.html  '),(19773,2190377,'11,June,2004','2004-06-11 00:00:00','   Reading my profile for this thing I noticed I had written 14,263 words.  That\'s a lot.  I should learn how to write that much about school related crap.  Today my grandmother said I talk too much too quickly ahahaha. ALSO frau vaden wants her to come in next year and tell us stories about WWII in germany. YEESSSS its gonna be so funny. huzzah!!  so basically, for me, school is over tomorrow.  i have one last bio project and frankly i just don\'t want to do it.  DONT WANNA DO ITTTTTTTTTTTTT.  Chem and math blew.  The highlight of my week so far, was when I was mowing the lawn... I almost ran over two turtles.  I know what you\'re thinking, \"O M G! Ashleigh!! You SICKO!!!!\" But see I didn\'t run them over which was very happy.  Then I looked closer and noticed that one turtle was sitting on top of the other one, or rather MOUNTED on top of the other one.  Did you know turtles had sex?  I can\'t imagine it would be very exciting.  Either way I didn\'t watch for long because I felt like a pervert, and maybe I am. Yay! brain.. hurts  goodbye.  please will sOMEONE RSVP??? :P  '),(19774,2190377,'11,June,2004','2004-06-11 00:00:00','   it\'s 6:40 in the morning and i\'m drawing pictures of mr. berry and mrs. eagen in mspaint lol.  so, math project.    i think school has driven everyone up the wall.   example:  yeahnd76: hyper spaceypiXiechick: how are you hyper? spaceypiXiechick: i dont like optomization.. too complicated yeahnd76: i just drank something yeahnd76: something yeahnd76: some thing yeahnd76: dunno what spaceypiXiechick: lol shut up :-P yeahnd76: :-P yeahnd76: :-P yeahnd76: :-P yeahnd76: :-P yeahnd76: :-P yeahnd76: :-P spaceypiXiechick: ANDREW!! WORK!! :-P yeahnd76: neeerobilagintanjaklog yeahnd76: thats like an indian fathers name yeahnd76: neeerobilagintanjaklog spaceypiXiechick: are u sure u didnt drink cough syrup? spaceypiXiechick: omg must work yeahnd76: neeerobilagintanjaklog srinivas yeahnd76: that would be his full name yeahnd76: and hed have some arranged marriage to some really weird indian sorceress with a dot on her head yeahnd76: and they ride elephants to work yeahnd76: Arabian Niiiiiiiiiiiiiiiiiiights! yeahnd76: im done  '),(19775,2190377,'05,June,2004','2004-06-05 00:00:00','   my priorities fluctuate by the hour.  Once summer actually comes... I\'ll probably go psycho without all the confusion.  I don\'t even know enough to know what to write about lol.  yesterday was my mommys birthday.  as well as screen on the green... err gym.  i had fun. now my quads hurt. next week = bone scan time. finals too? maybe. today sat 2. chem hw.. kinda.  no longer limping. or making any sense.  brother.. where is he? *worried*  friends... at this point, id rather say everyone than no one so thats how it\'ll be.  '),(19776,2190377,'02,June,2004','2004-06-02 00:00:00','   Today was very odd.  It could have easily been any day from last year... good I guess.  Funny things I haven\'t gotten around to posting lately:  spaceypiXiechick: what did that have to do with god anyway arvenmic: darwin spaceypiXiechick: ohhh arvenmic: and i was wondering how many ap bio people are religious arvenmic: heh, on a scale of 1-10, how much of a role does religion play in your life? arvenmic: jk spaceypiXiechick: :-P spaceypiXiechick: what have most people said so far arvenmic: wtf? about religion? spaceypiXiechick: yeah arvenmic: it was a joke spaceypiXiechick: oh spaceypiXiechick: lol arvenmic: ha arvenmic: ha arvenmic: ha spaceypiXiechick: leave me alone!!! arvenmic: ^laughing  surf in music: i spray painted ALL day for this hum project surf in music: and i finished surf in music: and then i did more hum surf in music: and then i did math spaceypiXiechick: ooh spaceypiXiechick: did u get high surf in music: heeeeeeeeeeeey surf in music: thats why everything is so blurry surf in music: man surf in music: i didnt even think about that surf in music: and i couldnt concentrate on other hw eithere surf in music: man surf in music: i forget what paint can do to you surf in music: but then i remember that it makes you forget surf in music: and thats why i never remember  yeahnd76: whats hotter, having bushes of hair cover your legs so they stay hidden, or showing the smoothness of them to the whole world? spaceypiXiechick: the latter yeahnd76: thats not a choice yeahnd76: what the hells a latter yeahnd76: ladder?  '),(19777,2190377,'17,July,2004','2004-07-17 00:00:00','   This entry will definitely make you think I\'m  off my rocker .&nbsp; And... well I am so I guess that\'s ok.     &nbsp;   I was reading this blog... and this guy had all of his compulsive behaviors listed.&nbsp; Then I thought,  shit!&nbsp;I do some pretty weird stuff too!!!!  The thing is, it\'s part of having attention problems, not OCD.&nbsp; Really I am not clean or orderly, just weird.&nbsp;      When I run warmup laps I constantly count my steps... It gets pretty high.&nbsp; With crunches I count too, except I can\'t seem to count to fifty... always by tens.&nbsp; Running on treadmills is even worse because then it gives me the opportunity to calculate my mph, constantly.    When I copy stuff at work I always have to copy in the exact same manner... same way to put the paper in, same way to pick it up and staple and paper clip etc.    When I can\'t sleep&nbsp;I don\'t count sheep, I do math.&nbsp; Or speak german in my head.&nbsp; I hate math so I guess it helps me fall asleep better.&nbsp; Still though, pre-calc at 3 am?     I always have to have a game of spider solitaire open on the computer.&nbsp; Also before I could ever start hw I would have to win a game of solitaire... which was a convenient way of procrastinating I guess.&nbsp;      heh. that\'s all I feel like sharing.    So, what have I been up to lately?&nbsp; Lots of working.&nbsp; Unfortunately I can\'t stay at work for more than five hours a day or I get really bad headaches from the insanely boring shit I do.&nbsp; aieeeeeee.... So I\'ve probably only made 200 dollars or soooo.&nbsp; The fact that I wake up at 11 every day probably doesn\'t help.&nbsp; I\'ve been sleeping over at Karin\'s so much in the past couple of weeks that my sleeping schedule is so out of whack.&nbsp; We always go to bed at about 430 and wake up at noon... I feel like I have a hangover.  ALSO MY DREAMS ARE SO FUGGIN WEIRD!!!!!!&nbsp;&nbsp;     I had one where a bunch of people I knew were going on a field trip somewhere and on the way back the bus stopped at Costco to eat (wtf?) then Luke insisted on buying potato chips but I was like, \"No way man potato chips suck arse.\"&nbsp; We bought them anyway.&nbsp; Then... something involving filling up canyons with edible sand.&nbsp; Yeah, that\'s when it stopped making sense to me too.    I had another dream where Elizabeth Kulajhgsheza whatever her last name is came to TJ as a junior replacement because Robinson was depressing and spilled her guts to me. lmao.      &nbsp;   Then by far the best and weirdest of the three was when I was hospitalized for something and they had me hooked up to the lil electrodes I have at physical therapy, but instead of shocking/massaging my muscles they were there to monitor my vitals.&nbsp; I rolled over in my sleep and the electrodes came off, so of course the machine thought I died.&nbsp; Out of nowhere the little machine that shocks you when you have a heart attack pops out from under the bed and shocks me, feeling uncannily like those invisible fences for dogs, then&nbsp;a nurse hears me screaming and comes in to turn the thing off.&nbsp; Then she\'s like shit why are you here you\'re just getting your teeth fixed.&nbsp; My friggin huge tooth falls out and she tells me I can go.&nbsp; So I get in the car with Karin (why was she there? who knows...)&nbsp; and we realize neither of us can legally drive.&nbsp; Because this is a dream and not reality we decide I should be the one to drive.&nbsp; It\'s dark and we\'re on Popes Head and I accidentally swerve on the wrong side of the road and nearly hit someone.&nbsp; Then I go, \"Damn! It\'s a good thing this Accord has such tight steering!\"&nbsp; WTF WTF WTF lol. We get to our destination and it turns out it\'s Rachel Miller\'s house.&nbsp; I\'m not sure what happened from there... something about how I needed to see her so badly I just drove over to her house with driving ability whatsoever.&nbsp; Then I think I might have professed my undying love of her.&nbsp; And then I woke up.      aieeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee.  &nbsp; SO WEIRD.&nbsp; But it\'s cool that I could remember my dreams in that much detail.     &nbsp;   Enough about dreams.&nbsp; Last Sunday we all went over to Maria\'s and had our very late 4th of july party.&nbsp;  AND THE COPS BUSTED US . ahahaha.&nbsp; I thought the car was Andrew\'s dad\'s and then I realized that there was a spotlight shining in my face and I was like ohhh shit.&nbsp; Somehow I couldn\'t wipe the shit-eating grin off my face, though.&nbsp; Fortunately he didn\'t kill us.&nbsp; Still I\'m impressed at how badass we are.&nbsp; OOH OOH and we made Raj wear makeup.&nbsp; Hoooooot.&nbsp; He looked like a cross-dressing Indian Cleopatra but boy was it  hot .     &nbsp;   Watching King Arthur made me want to go out and kill people with swords like a pseudo-amazon.  '),(19778,2190377,'14,July,2004','2004-07-14 00:00:00','   quite possibly the longest survey ever. dont read it :P     Name: ashleigh Rank: ... Serial Number: -_- Age: 16 Weight: no comment Height: 5\'5\" i think Eye Color: blue/grey Hair Color: brownish.. Marital Status: single!! State of Residence: va Occupation: student Religion: christian Siblings: brian... then chris and stephanie     Color: kelly green Animal: ehhh. after seeing 50 first dates it just might be walrus hehe Book: Shade\'s Children T.V. Show: I Love the 80s... and anything else on vh1 Movie: that is just way too hard to answer dude. Reality T.V.: the real world. and the spinoff on vh1 the surreal life :)  Food: everything. especially ice cream. mmmm and ramen Time of Day: 10 pm Room in Your House: basement hehe Time of Year: first two weeks of summer Website: durrr OS system: ...windows?  Computer Game: haha... spider solitaire!! :P Online Game: dont play any Video Game: hehe.. the whole spyro series, any final fantasy game... SNES games!! Brand of Clothing: uhhhh. dunno Type of Shoe: retro sneakers mmmm Store: target :) Toy: do bubbles count? Junk Food: ramen.!! Comic Book: online ones heh... like mall monkeys and diesel sweeties Cartoon: blues clues? Restaurant: does coldstone count Car: my beaaauuutiful honda accord Magazine: seventeen? anything stupid and girly Actor: scrumptious ones Actress: scarlett johanssen... or however you spell it Political Figure: the ones that arent stupid. oh wait there arent any. Author: garth nix, anne rice, m.t. anderson     Song when You\'re Happy: want you bad by offspring Song when You\'re Sad: anything sarah mclachlan Song when Your Angry: i dont listen to music when im angry.. i just kick stuff Song when You\'re Heart is Breaking: sappy country? Song when You\'re Feeling Mellow: the remedy- abandoned pools Song to Dance to: dunno Song to Sing Along to: everything Romantic Song: sappy country. Love Song: thats teh same thing you dipwad     Do you have friends: yes Are you sure: yeah Positive: yes Well, fine don\'t get cranky: rawr Best Female Friend: karin.. kelly Best Male Friend: mreehhh.. andrew? most of my friends are guys Best Online Female Friend: i dont talk to girls online lol. karin. Best Online Male Friend: my favorite person to talk to is ROBERT DU Smartest Friend: rachel.. boone Dumbiest Friend:me Friend you have the most fun with: YOU! Friend you don\'t really like: no one??? *shifty eyes* lol. WHY would i answer this on my blog where everyone can see it??? Friend that\'s always in trouble: umm kelly? Sadest Friend: ...??? Loneliest: ... dunno Happiest: ummm...???  Which will end up in jail first: haha ed. for lighting something on fire or something Be the first to get married: ... ashleigh??? Be the first to have kids: ummm.  Talented: .... dunno.... we\'re prolly all talented in some way... ;-D Artistic: ella Funniest: luke and his jokes :P Dependable: ummm andrew? Most like you: karin? Most unlike you: uhhh Trustworthy: trust no one. Gossip: me Quietest: i got a lot of quiet friends... brian luu, aaron Loudest: jessica lol    if i skip it, then i haven\'t seen it  Lord of the Rings: yay!! StarWars: yay? The Wizard of Oz: creepy Willy Wonka: omg i love that movie Gone in 60 Seconds: sex in the car... the sphinx, thumbs up The Last of the Mohicans: long. really long Dirty Dancing: :)  Top Gun: cant remember it Eddie and the Cruisers:  Harry Potter Films: long and .... well not very good, needa see the third Super Troopers: I LOVE THIS MOVIE ITS SOOO BAD IVE SEEN IT THREE TIMEs Girl Interrupted: pointless The Neverending Story: !!! i love the dragon thing.  i dream about him The Dark Crystal:?? Cruel Intentions: hehe. so weird. Spaceballs: meh/.. Mean Girls: gotta love lindsay American Pie: funny gross Pretty Woman: cant remember Twelve Angry Men:  The Land Before Time: cheesy First Knight: NO Legend: ? The Karate Kid: cant remember Wiillow:  Miracle on 34th Street: good around christmas An Officer and a Gentleman:  Dragon Heart: i cried.. lol How the Grinch Stole Christmas: gave me a really bad headache   >  The Ring: yay!!! Darkness Falls: i laughed my ass off Stephen King\'s It: also funny Psycho: creepy Halloween:  Friday the 13th:  Children of the Corn: creeepy The Silence of the Lambs: awesome but the books better Jaws: meh Poltergeist:  The Birds: hahahah. Signs: not scary The Sixth Sense: loveeee Thirteen Ghosts: scared the crap outta me     Lady and the Tramp: yay :) Alice in Wonderland: creepy Toy Story: ok Monsters, Inc.: i cried during this too Dumbo: BAAAD The Little Mermaid: i have it on dvd Finding Nemo: love! The Lion King: the best disney movie. Bambi: haaaate  Peter Pan: i dont remember the disney version but the one that came out recently was ok The Aristocats: cutecutecute Beauty and the Beast: cried Cinderella: cool mice Sleeping Beauty:so so     Smoke Cigarettes: no Smoke anything else: nooope Drink Alcohol: nuh uh... well... Abuse Alcohol: noppers Take Prescription Pills: yes Have eleven toes: eww Have twelve fingers: no Dance around the house naked: heh.. Sing in the shower: yupyup Cry yourself to sleep at night: no Wish you could run through a field of corn: no because the aliens would get me See dead people:nope Feel alone: naw Feel loved: by my friends! Love yourself: no Hate yourself: sometimes Hate this survey yet: its getting old Like yellow snow: only if its mt. dew Like snow at all: not really. Like rain: so i can go running!! Like the beach: hehe only when im with friends Like unicorns: :) Talk to yourself: errrr Answer yourself: errrr Believe in love at first sight: nonono Believe in true love: meh Believe in karma: yes Believe in fate: no Belive in reincarnation: actually kinda Belive in heaven: yea Belive in hell: yea Believe in god: yea Have a home phone: mhm Have a cell phone: no -_- Talk to your friends online: yeah Have AIM: mhmmm Yahoo Messenger: nope MSN Messenger: nope Use an online blog: haha noooo Spend more time online then in real life: noppers Have any pets: too many Like spiders: sometimes Pick your nose: ewwwwiw!! Curse: of course!     Scared of heights: no i love heights Afraid to die: hmmm.... Insane: yes! Still actually taking this: duhhhh Alone in this world: i dont think so Sad right now: .... ionno... kinda... ? Happy right now: ... kinda... Wishing you could beat me bloody for making this so long: hell yeah Hungry: ..... Tired: nah  Pretty: hahahahahaha no Popular: who cares? Wicked: lol... of coures... hahah jk ... im a little angel... hmm wow neither of those discribe me....  Fun: i thinks  OMG MAKE IT SHUT UP I GIVE UP THE RREST ARE KARINS ANSWERS    Apple: red Eye: danny... wow he had great eyes... i cant believe he left after freshman year! Bird: penguin Dog: puppy... cute Peanut: penut butter - yukky Black: hot pink Winter: snow Mother: insane Father: non existant Toothbrush: teeth Lawyer: mr. weiner Fear: scary movies Toenails: multi-colored Leg: hurts Bracelett: make-out Dinner: yummy Fat: sumo wrestlers Sneaky: ... mosquitos Mouse: .... cute? Grass: green Rock: and Roll Fan: dizzying Lightbulb: not good to look at directly Rabbit: fluffly... momo\'s bunny who is spending 3 weeks at my house while she is in china Pepper: salt Eggnogg: christmas     Red/Green: depends what type of green... if it is the bright green then green, but if it is a dark yukky pale green then red Winter/Summer: depends... winter is awesome cuz of christmas and snow but if its really cold and there is no snow then is sucks... but summer rocks cuz there is no school, but sometimes the hear is aweful... Rain/Fog: depends... rain is awesome to run in... but fog is awesome and mysterious to look at  Night/Day: depends... day is awesome cuz of hanging out with friends... but night is awesome cuz of sleepovers... Moon/Star: stars! wow this is the first answer where i actually decided on an answer  Daydream/Nightmare: daydream... duh Shirt/Shorts: shirt Long/Short: what?... depends on what the survey writer person is refering to... cuz some things are good long, like relationships, kisses, friendships, summer... but some things are good short, like schhol days, school year, guys\' hair... Cat/Dog: cat Hamster/Fish: hamster! Spoon/Knife: knife muahahahaaaa.... but.... spoons are also good for ice cream! Taco Bell/Arby\'s: both Would you like fries with that: yeah Banana/Peach: peach! hahaha Girl/Boy: boooys Silver/Gold: silver Tennis/Synchronized: tennis? Blue/Orange: depends on the shade of blue... if its bright blue, then blue... if it is dark navy blue then orange Glass/Metal: dunno... they are both pretty and shiny Beach/Wood: both are awesome Paper/Plastic: plastic Coffee/Tea: tea.... i hate coffee Carpet/Tile: carpet Long Hair/Short Hair: depends on who... Facial Hair/Clean Shaven: clean Rock Climbing/Sky Diving: omg deffinately both.... altho i still havent tryed sky diving... but i know i will soon... cuz i know my mom will make me try it soon Starve/Dehydrate: neither    held hands: as a couple? yeah kissed someone: yupyup hugged someone: of course made out: hehe. yeeeah given oral: nope recieved oral: no had sex: nuh uh     What time is it: 1.58 am What are you doing: this survey and talking to people  What are you wearing: my pink atitude shorts and my black \"satisfaction guaranteed\" tank top. lol What are you watching: computer screen and vh1 What can you see: my computer? What can you hear: mean to live - switchfoot... from ben\'s xanga because i forgot to close that window and thats where i got this survey from What\'s the last thing that you ate: ice cream... yummm What\'s the last thing that you drank: mountain dew! weee Who\'s the last person you spoke to: mom? Who\'s the last person you instant messaged: larissa Who\'s the last perons you emailed: megan davenport... lol Who\'s the last person you talked on the phone: ashleigh When\'s the last time you said thank you: dunno... When\'s the last time you said i love you: i say it all the time to my friends... but prolly to ashleigh the last time we were at university mall... so like 2 days ago... Who was it to: prolly ashleigh. hahahaha Did you mean it: yeah... i love her as a friend... lol... this is really weird...  Really: no. ashleigh i hate you and i never wanna see you again. lol. jk That\'s good thanks: ??? How far have you gone in a relationship: i\'m not going into that in a public blog... If I were a cockroach would you squish me: duh. cockroaches are gross Do you really just wish I\'d hurry up and end this: yeah If you could meet anyone alive or dead, who would it be: dunno Why: ... i didnt pick anyone... how can i say why? Who is your mentor: ...??? Who is your idol: don\'t have one Do you watch American Idol: nope What does Eeyore always lose: his tail... but he\'s soo cool Can you hear yourself thinking: ... do i think? What\'s your dream job: something to do with art, and or potography Do you want to be a millionaire?: yeah  '),(19779,2190377,'09,July,2004','2004-07-09 00:00:00','   aiieeee so much to talk about but really aren\'t surveys so fun?   1) Using names of bands, spell out your name ACDC Sixpence None the Richer Hootie and the Blowfish Leo Kottke Eve 6 Ikara Colt Goo Goo Dolls Hoobastank  Audio Adrenaline Live Lifehouse Enon NOFX   2) Have you ever had a song written about you? HAha. Actually I think so.. sorta 3) What song makes you cry? Once this song by Jagged Edge made me cry lol.. except i dont remember what it was called... oh and lots of country love songs do 4) What song makes you happy?  Disposable Parts- Enon  a b o u t . y o u [ full name ] Ashleigh Anne Allen yeah yeah triple A get over it [ birthdate ] 5/21/88 [ location ] Fairfax... [ sexual preference ] Often. haha aren\'t i so 6th grade :) [ height ] 5\'5\" [weight] hahah. 125. when im lucky [ hair color ] questionable. brown? [ hair type ] crappy... whatever it wants to be. wavy because i don\'t brush it very often [ skin color ] really pale pink  [ eye color ] blue [ piercings ] Ears... must have moreeee [ tattoos ] Ew, no.   m u s i c [ do you play any instruments? ] Cello [ if so, what one(s)? ] Uh...look above. [ favorite genres of music ] Indie!!!! weee. Indie pop... Rock... anything with keyboards... singer-songwriter fare, really hyper techno-y stuff  [ favorite bands ] I HATE THESE QUESTIONS!!! rawwwwrrr. I don\'t really have favorite bands... I like The Police, and Weezer, aaaaand Cake. yeah  Saves the Day...  [ do you go to concerts and/or shows? ] Sometimes.  [ what is the most you\'ve ever spent on a concert/show? ] I forget [ what is the least you\'ve ever spent? ] free [ do you think buying merch at a concert then wearing it there is corny? ]not if you\'re at a four day music festival :P [ do you listen to any bands that you\'d be ashamed to admit to listening to? ] I\'m not ashamed of anything.  BWAHAHHAa. CHRISTINA AGUILERA!!! weeee etc.  r i g h t . n o w [ pants youre wearing? ] shorts [ shirt youre wearing? ] blue.. from wet seal.. flowers and stuff on it [ underwear? ] yes. purple and striped [ music listening to? ] Nickel creek... the lighthouse song [ taste in your mouth? ] cereal  [ how are you? ] weee [ tired? ] ish [ happy? ] more than usual [ depressed? ] a little  l o v e [ boyfriend? ] nope [ girlfriend? ] nope [ sexuality? ] maybe  [ are you a virgin ] yuss [ children? ] are smelly [ current crush? ] harumph [ been in love? ] i guess [ do you believe theres a difference between \"love\" and \"in love\"? ] most definitely [ is it better to have loved and lost than to have never loved at all? ] depends [ are you romantic? ] noo [ had a hard time getting over someone? ]  ish [ been hurt? ] hasnt everyone [ cheated? ] No [ your greatest regret? ] ive got lots [ turn offs ] Facial hair!!!!!!!  [ favorite place to be touched ] anywhere but my armpits -_- well almost anywhere [ what song describes you now? ] Bubbletoes. lol  r a n d o m  [ do you have a job ] at my mommys office. although im bad and dont go very often [ your cd player has in it right now ] the postal service!! [ if you were a crayon what color would you be ] greeeen. oooh greeeen. [do you use internet shorthand (i.e. \'lol\', \'brb\', \'jk\')? ] screw you. yes. too often. I CANT STOP [ are you a people-pleaser ]Unfortunately yes    OKAY   So I have been busy doing things I don\'t normally doing lately.  Like.. having a life, writing, reading, having a life, volunteering to do not so fun stuff.  My social life is nonstop. I might dieeeeeeeeeeeeeeee.!!!!! Monday: Family took me captive Tuesday: Karin and I watched American Pie... a lot. -_- my mind was soooo perverted for awhile.  Wednesday: Went with Andrew to PG county to babysit my sister church\'s lil kids.  Oh my gorsh.  SO CUTE.  \"Ashleigh and Andrew sittin in a tree... kyssnaajgnana\" they couldn\'t really spell.  aaaand then we played tag in the parking lot and i didn\'t get hit by any stray bullets!! (stupid dad) Then we played simon says... and andre was all like \"simon says stick your finger up your nose, simon says stick your finger in your ear, simon says stick your finger in your mouth!!! the SAME FINGUH YOU STUCK IN UR NOSE!!!!\" aieee. omg too cute too cute too cute.  andrew got a picture of one clutching me and he was all like acting like i was his momma and it was so sad and so adorable and then he whips out the camera and the kid pulls out an east side symbol -_-.  haha ALSO \"simon says lay eyes on the pimp!!!\"  I made a new friend.  His name is Dashawne.  He is the coolest lil kid ever.  It\'s so sad cause they all have parents that don\'t give a crap and they don\'t have beds and shit and they were just so happy to have people pay attention to them.  I think imma go back because I had so much fun.    haha ok so that was a long wednesday.\"i can do a lot more than that with my socks big boy\" BUT ITS NOT EVER YET. then party at marias haaaah so funny. sooo funny. we played men are from mars women are from venus. bwahaha. um. and then karin slept over. and hten the next day i lost my dog! and then i found him that jackass. AND THEN i went to the mall with ed and his awesomepossum friends... coffeecoffeecoffeecoffee WE PLAYED WARIO WARE THAT IS THE MOST CRACKHEADED GAME EVER.  cool though.  Then I decided I might gauge my ears.. ooooh.    And then it was today.  And we went to the farm and chilled with supersuperusper cool old ladies and talked about stuff and i ate the best peaches of my entire life.  I am so serious. The peaches were better than the cake.  And it was chocolate cake  So today I realized that I am such a friggin dreamer when it comes to my goals.  i was like...   I want to write a book. I want to teach English in Japan. I want to become a doctor. I want to start my own business.    haha so the last one is really an fd up idea. I want to start a dog salon where you give them punk hairstyles and dye \'em and give em pedicures in bright colors.  siiigh  Um, so that was my week.  Well, about half of it.  PHYSICAL THERAPY!!! ELECTRODES!!! weeeeeeeeee.  Bessie gives good massages. :) i get to use the elliptical machine soon!! i am so out of shape its not even funny, but in a way its fun because i grew a BUTT!!!  This might just be the longest, most pointless post ever  '),(19780,2190377,'04,July,2004','2004-07-04 00:00:00','   HAPPY 4th!!! yay. it\'s my brother\'s birthday but he\'s sick.  this saddens me greatly. poor brian :( i kinda forgot it was independence day and all that jazz until my mom woke me up at 9 tonight hah... i had been reading rule of four and fwoosh somehow fell asleep.  i think its cause i was listening to postal service and that stuff is so surreal ur just like weeee dream world!! OH YEAH. last night i had a dream that i dyed my hair at this place... and they guy did my hair green. but the dye wasn\'t green enough so he added listerine.  GREEN listerine. -_-. really sometimes i wonder about me.  haha now i want to dye my hair though, not necessarily green but stiiill.  has anyone seen matt? i miss him.    this is a pointless entry i apologize for my retardedness  '),(19781,2190377,'03,July,2004','2004-07-03 00:00:00','   Haha. So this week has been interesting.    Yeah.  I\'m unable to sleep right now and am talking to Max who is also struck with insomnia. weeeeoooohwwweeee  Max is a cool kid even if he would rather not admit it.  Yesterday I spent most of the day with my brother which was cool. We went to Potbellies and his friend Stephanies house where i played with her whacko dogs.  she had this really senile sausage dog that would like snarl at you and then forget about it and start licking you... nonstop.  He was cute though.  ALSO SHE HAD A DOG NAMED WALLACE. TOO COOL.    then we went to the gym. i cant move the armmmsssss aaaaagh.  bought clothes for brian afterwards which was surprisingly fun.    i am sooo sore today from falling down a lot and the gym and shit.   man im not very witty at 6am.    dammit where is my other earring. and my other flip-flop.  and who left their flip-flops at my grandmothers house man.  '),(19782,2190377,'05,August,2004','2004-08-05 00:00:00','   Today I went to the optometrist to pick up my new  urlLink  emo   glasses.  Uber cute.  There we found out that the guy who always helps us pick out frames is moving on to bigger and better things later this month.  This may not seem like a big deal but  I love  this guy.  In high school he was the weird scrawny kid with long hair and a funny name that later turned into a babe magnet.  (Literally, as evidenced by a halloween costume he had one year... consisting of troves of barbies stuck to a magnet-y suit.)  People used to come up to him and go \"ARARARARARARAR\" because his name is Deano, pronounced like Dino... from the Flintstones.  Deano Ventura.  How much cooler of a name can you get? If you have a name like that you are obviously destined for stardom.  So he is leaving my very odd optometrist\'s office to pursue a career in music, inventing, and warehouse management.  Talk about being ADHD.  Apparently his band is going to provide music for CSI this upcoming season... too bad his band isn\'t called something totally rad like \"Deano and the ________\" (does it really matter who the other people are? nahhh.)  Tonight I will wake up at 3 AM knowing that somewhere in the wilderness of Manassas there is a 39 year old trapped in a 19 year old\'s body mountain biking until his legs fall off, hopefully he doesn\'t run into a deer.  That is the end of my tribute to the dude that picked out my glasses.   In other news, my grandmother had to put Fancy down.  For those of you who didn\'t have the  privelage  of meeting Fancy, she was my grandmother\'s decrepid, crotchy, old, Eewok dog.  18 years old.  I think it was actually already dead, but they were using experimental treatment that involves turning your dog into a zombie.... which means I should probably blow up University Drive Veterinary Clinic before tons of monsters created by the Umbrella corporation bust loose.  In all seriousness though, I love you Fancy.  You were a good dog.  R.I.P.  Mostly I feel sad for my gramma cause now she sits at home complaining of dizziness and cries sometimes.  :/  '),(19783,2190377,'01,August,2004','2004-08-01 00:00:00','   Brian just attacked me with electric nosehair clippers. What an exciting weekend lol.  Too much taco bell. Too much money being spent. Not enough sleep. However I am very happy because my family spent about $650 on electronics. We aren\'t like friggin\' rich either. Unfortunately I also blew my entire paycheck. Aieeee.     Canon Powershot A60 digital camera, its only a 2 megapixel camera but I don\'t think I need any better since I\'m just an \"amateur\"  Random camera accessories... camera bag etc so I don\'t break the damn thing  256 MB flash card (extra camera memory)  Color all-in-one photo printer dealie. seriously effing nice. LCD display and you don\'t need to hook it up to a computer since it\'s wireless.   Faster wireless router ... I don\'t really understand what these do  Spiderman and one of the Matrix games for Xbox, Prince of Persia for PS2  The nice cd player my dad bought my grandpa for christmas last yr, since he can\'t figure out how to work it (understandable considering the dudes 94)  My brother\'s old DVD burner since he\'s buying a new one... not yet in my possession but still   It feels like christmas!!! especially because this stuff is way better than what I got at christmas (i think... honestly I don\'t remember what I got)  Except it\'s not as cool because I financed about $200 worth of it.  We got the camera at Microcenter and I swear... it was like if you took TJ\'s computer science team, made them 1.5X dorkier, and fast-forwarded 10 years.  I was the only white female there, well, besides my mom.  So, the other day I had this really nice... sentimental entry written.  It took me at least an hour to write because it was the survey thing that was addressed to anonymous people.  Well, my session timed out and I lost everything I had typed.  So... sorry.  Basically it was all... blah blah I love lots of people.  But a couple suck. yep!!  blahblah.  end shallow happy entry.  boring yes.  sorry.  '),(19784,3448971,'28,May,2004','2004-05-28 00:00:00','       '),(19785,3448971,'27,May,2004','2004-05-27 00:00:00','       '),(19786,3448971,'27,May,2004','2004-05-27 00:00:00','    I always go to the internet. And hear music or urlLink  view pictures  about my country or even read about what\'s happening in my country.   '),(19787,3448971,'27,May,2004','2004-05-27 00:00:00','    My favorite website is a website about  urlLink Ecuador  and you can go see it now!   '),(19788,3448971,'26,May,2004','2004-05-26 00:00:00','    Here is a link to an  urlLink assembly  I went to last semester.   '),(19789,3448971,'26,May,2004','2004-05-26 00:00:00','    Hi! How are you doing this morning? I hope you\'re fine. And what did you have for breakfast?   I hope you had a good breakfast. Because it\'s good to eat a good breakfast, so you can start the day with a lot of energy. Okay Bye Now!!!   '),(19790,3448971,'25,May,2004','2004-05-25 00:00:00','    Hello World! Welcome to my new blog!   '),(19791,3448971,'14,June,2004','2004-06-14 00:00:00','    There\'re a lot of websites about Ecuador however I found a website that has everything about Ecuador,it has the history of Ecuador, all the languages that we have,everything about the people ,a map of Ecuador, Ecuador\'s history etc... so if you want to know about Ecuador and get informed then click on this link  urlLink ECUADOR .   '),(19792,3448971,'11,June,2004','2004-06-11 00:00:00','    Like I said today I am very happy because, today is Friday and also there\'re just two more days until school is over. However today  we have a math final and I am very nervous because I think the math final is going to be hard. Yesterday we had the English final and I think I did well. Then on Monday and Tuesday we\'re having our history and science finals. On Tuesday is our last day of school. Then I think we\'re coming back to school on September 11 or so.   '),(19793,3448971,'07,June,2004','2004-06-07 00:00:00','    This summer I\'m going to have my (Quinceanera) Sweet 15th party. That\'s why I\'m very happy. I\'m counting the days until the day of my party. From now until the day of my party there\'s 18 days left. My party is going to be on the 26th of June. And today is the 7th. I feel so happy because all of my friends and cousins are going to be there with me. Even though my birthday is July 3rd my party is goimg to be a week before.My dress is just like the dress that this doll has and it\'s being made it Queens and I have to go to pick it up on June 18th. I love my dress because it\'s exactly like I wanted it. It\'s tight at the top and very big at the bottom. On friday we were rehearsing and today we are going to rehearse too. We have to rehearse a lot so that day everything could work out fine.   '),(19794,3448971,'03,June,2004','2004-06-03 00:00:00','    Lung cancer is the number one cancer killing. At last the scientists and all those people working hard to defeat lung cancer have made a difference. The rates of lung cancer in women have decreased. The Scientists from the National Cancer Institute ,the Centers for Disease Control and Prevention and the North American Assossiation of Central Cancer Registries say that it was about time that the rates of lung cancer in women have declined.   '),(19795,3448971,'01,June,2004','2004-06-01 00:00:00','    Reggaeton is a type of spanish reggae. Reggaeton originated from Puerto Rico. There\'re a lot of good music groups of Reggaeton that I like. For example Zion y Lennox, Tego Calderon, Wisin y Yandel, Big Boy, Speedy, Don Chezina etc.  I love all of the Reggaeton singers and it\'s hard to choose from all those singers. However if I had to choose ,I would choose Don Omar because he is very good and he has a lot of good songs that I like a lot. If you want to see Don Omar then click  Don Omar   '),(19796,3208138,'30,November,2003','2003-11-30 00:00:00','   No, Rob, see the note to the right--this is your blog.  '),(19797,3208138,'29,November,2003','2003-11-29 00:00:00','   i just realized, i\'m in the wrong blog. i should be in the weaveryuleblog. my bad.  '),(19798,3208138,'29,November,2003','2003-11-29 00:00:00','   On audio cassette or in book form: The Ultimate Weight Solution: The 7 Keys to Weight Loss Freedom, by Dr. Phil ($15.60 at amazon.com)                                                                                                       Plain, pastel pink, women\'s crew-style socks (no design on them)            Movie gift certificates (yes, I copied Rob\'s idea) --- Regal, Edwards  World peace (or whirled  urlLink peas )  '),(19799,3208138,'29,November,2003','2003-11-29 00:00:00','   Did someone say something about a cat? That sounds delicious. Be sure to keep our water bowl filled. We might choke on the fur.  We also like Old Mother Hubbard\'s P-Nuttier Extra Tasty, Natural, Basted Dog Biscuits, available at Petco at the Quad.   And, those pigs ears that someone gets us at Christmas are tasty too.  '),(19800,3208138,'29,November,2003','2003-11-29 00:00:00','   Soundtrack CD for the movie \"Pirates of the Caribbean\"  Gift certificate to: * Pier I Imports * Borders *                  * Crate & Barrel  From  urlLink William-Sonoma  (with catalog # --- some items only available on internet or catalog, but I\'m not sure which ones): * Olive Wood Grater, #86-4759445, $29 * Slanted Dish Drainer, #86-4739553, $18 * Peppermint Hot Chocolate Set, #88-4821526, $29 * Peppermint Bark, #88-3491297, $20 * Toffee Trio, #88-1263680, $22  Stamp Stuff for a hobby... Contact Helen Homet:  E-mail: hmhomet@aol.com    Phone: (562) 945-4636 Stampin\' Up Consultant (free shipping if ordered by 12/10 --- saves additional 10%) Hobby Supplies: * Stamp Set: \"Snow Angels\" (#102692), $14.95 * Stamp Set: \"Best Borders\" (#102019), $21.95 * Paper Crimper (#101618), $19.95 * Stamp: \"Hand Stamped By\" (#103516), $19.95 --- add name: \"Cindi\" * Stamp Set: \"Good Times,\" (#100116), $28.95 * Gift Certificate in any amount  Thanks.  '),(19801,3208138,'29,November,2003','2003-11-29 00:00:00','   1.  Hat(s)   2. Gift certificate to  urlLink hat store  3. Gift certificate to a store that may not specialize in the sale of             hats, but has some hats in inventory 4.  urlLink Menoxidyl  5. Gift certificate for  urlLink greyhound bus lines  '),(19802,3208138,'29,November,2003','2003-11-29 00:00:00','   Ok, so my wish list might not be as funny as Rob\'s has been. But, it has live links:  '),(19803,3208138,'29,November,2003','2003-11-29 00:00:00','   Audrey--                                                                                blender, 9X9\" baking dish, 11x8\" baking dish, blue kitchen towels, leaf blower   Timothy--[a] vice, thread seal tape, Home Depot gift card, Battel\'s Hardware gift certificate    Daniel--  2001 Tour de Suisse video,                                   2000 Tour de France video, 1998 Tour de France video, 2000 Giro d\'Italia video, 1997 Tour de France video, 1999 Vuelta a Espana video, 2003 Vuelta a Espana DVD, 2003 Giro d\'Italia DVD, Crash DVD, 2003 Year in Review book, money, WCP (World Cycling Productions) gift card ( you can find all of these at  urlLink worldcycling.com )  '),(19804,3208138,'29,November,2003','2003-11-29 00:00:00','   Here are the gift-giving pairing for Christmas Eve gift-sharing fun 2003: Doris gives to Steve Glen gives to Dave Cindi gives to Audrey Beth gives to Cindi Steve gives to Doris David gives to Glen Rob gives to Beth Kyle gives to Richard Audrey gives to Jane Richard gives to Rob Jane gives to Kyle  Please post your wish list on this blog to help the person who is shopping for you. Thanks.  '),(19805,3208138,'29,November,2003','2003-11-29 00:00:00','   Here is the first sample posting. I will send everyone an invitation to join the blog. You may then sign in and post your wish list stuff for Christmas, post Christmas greetings for everyone to read. Please enjoy the first Peterson Yule Blog!  Blog is a contracted word--a combination of the word \"web\" and \"log\" --\"blog.\"  '),(19806,3208138,'27,December,2003','2003-12-27 00:00:00','   What a crazy place this is!   I have tried to post some of the pictures but the JobStarts server is not responding. I will have to get that working when we get home.   I had a great time with everyone. Hope we can do it again next year.  '),(19807,3208138,'21,December,2003','2003-12-21 00:00:00','   Dear Family,  I hope all are getting rested after the big day.  I want you all to know that I had a very nice time this Christmas.  I thought that it would be hard to see you all disperse on Christmas day but since we don\'t know the exact day that Christ was born we just can celebrate His birthday when the family can be together.  Missed Andrew and hope that he can come to warm So. Calif. sometime this year to visit.  God bless you everyone.  Mom, Doris, Gram  '),(19808,3208138,'21,December,2003','2003-12-21 00:00:00','   I love Sam but have never known him to be so articulate.  Also, while in PetSmart yesterday, my favorite checker said that whichever dogs I was buying for must be pretty spoiled.  Have Samson and Ashley ever had Pupcorn?  Samson may want to rethink his version of the \"Pool Story\" if he ever wants to try some...  Love, Auntie Beth  '),(19809,3208138,'20,December,2003','2003-12-20 00:00:00','   It seems that Glen is the only one who can post pictures.  Bummer.  I have a picture of the cat under the tree wishing us all a Merry Christmas (or that we would all disappear and leave her alone in HER house, can\'t tell the difference), but only the administrator person has those rights.  '),(19810,3208138,'20,December,2003','2003-12-20 00:00:00','   Now Louise sounds like a level-headed cat.  D.C., of course, is nowhere near level-headed.  At this moment she is lying atop a pile of unfinished work on my desk.  I have thrown her off numerous times, but she will not be deterred.  When I was scanning something using Steve\'s computer, she was all over his desk like a rash.  I believe she is anxious for us to leave so she can sleep on the bed without being thrown off.  Please do not get her anything for Christmas.  She wouldn\"t appreciate it.  She is a spoiled cat and reminds me way too much of myself at times...  Beth  '),(19811,3208138,'20,December,2003','2003-12-20 00:00:00','   This is a test. This will now show up on the blog.  '),(19812,3208138,'20,December,2003','2003-12-20 00:00:00','   I told Sam about the pool story on the blog from Beth. He asked that I give his version of the story--he can type, he just didn\'t want to hog my computer.  Sam seems to remember that Rob fell into the pool when he was chasing a squirel. Sam then remembers that he jumped in the pool which had a cover on it at the time and pushed Rob out, saving him from exhaustion and eventual drowning under the cover.  Thanks to Sam, Rob is with us today.  '),(19813,3208138,'20,December,2003','2003-12-20 00:00:00','   thought of some other useful things:  9v, and aa batteries  socks-any color, or type.  stereo headphones-for guitar amp  phone cards  '),(19814,3208138,'20,December,2003','2003-12-20 00:00:00','   meoooooowwwww!   I\'d rather not be dog stew for Christmas, something I just noticed that was published on the blog.  I do like all sorts of kitty treats you find at almost any store.  '),(19815,3208138,'18,December,2003','2003-12-18 00:00:00','   my fave christmas story was last year when my little brother came to visit, and i was able to hang out with him after being out of contact for three years. hope i can see both of my brothers this year.  '),(19816,3208138,'16,December,2003','2003-12-16 00:00:00','   One of my favorite Christmas stories happened when Glen, Cindi, Samson, Dick, and Jane came to P\'Ville for Christmas.  We were all around the table making merry when Rob appeared, much like Santa Claus, to annonce that Sam had gone for a dip in the swimming pool.  We all laughed and went back to our merry making, confident that Rob was just joking (Rob?  Joke?!  Who\'d \'a\' thunk?)  A few minutes later Rob and Sam both appeared soaking wet at the door...and we all felt a little guilty for not having believed Rob.  But we got over it.  Thanks to Rob and Sam for a great Christmas memory.  Love, Mom or Auntie Beth  '),(19817,3208138,'14,December,2003','2003-12-14 00:00:00','   Hello,  I believe I have already told the person who is an official gift giver to me, but just in case, here is the Official List:  Beth only wants things that can be use up like candles or coffee (check her closets and the rest of her house to know why...)  OR  things that are beautiful and useful like dishes, plates, etc. in shades of blue, gold, or yellow...  ALSO  Donations to worthy causes such as the Muscular Dystrophy Association, churches, or the Society for the Prevention of Reproduction in the Poor White Trash Community (SPRPWTC).  Yes, I made the last one up, but I couldn\'t help it after another day as a public school teacher in El Dorado County.  ACTUALLY  I am most looking forward to Christmas with all, or close to all, of my family.  Feliz Navidad, Beth  '),(19818,3208138,'14,December,2003','2003-12-14 00:00:00','   It would be a good idea for everyone to check and see what other people are getting other people to avoid duplicate gifts. For example: Rob has already recieved a clock radio CD player device. He in fact already has it so this won\'t spoil to surprise.  '),(19819,3208138,'11,December,2003','2003-12-11 00:00:00','   And, here are some images from tonight:     Dinner at the  urlLink Chicago famous Berghoffs .     The street near my hotel.     The lobby of the famous Palmer House Hilton--brought to you by priceline.com for $74 a night! WOW!  '),(19820,3208138,'10,December,2003','2003-12-10 00:00:00','   Hi,  I\'m hoping that Timothy\'s REAL wish is for a vise, not vice.  Makes a big difference...I\'m just bringing it up because I have not come up with my own list yet and I\'m stalling for time.  Congratulations to Audrey for starting us out!  Beth  '),(19821,3208138,'09,December,2003','2003-12-09 00:00:00','   It involves all of us meeting at a home on a warm ocean, with plenty of room and activities for all the different ages. It would help if a maid and cook were part of the deal. More importantly, it would be away from the daily routines and distractions of everyday life. Now if it was a home I owned and lived at that might be even better, but I don\'t buy a Lottery Ticket any more.  urlLink I am sure Glen could find a link for this one.   Santa note: Glen might like this link-- urlLink Mayan Palace .  '),(19822,3208138,'09,December,2003','2003-12-09 00:00:00','   These images are from Miami from the window of my hotel overlooking Biscayne Bay. [I\'ll check on that spelling later.] The forcast high today was 80.   Tomorrow I am back up to Chicagoland. When I was there on yesterday, it was 26.  '),(19823,3208138,'09,December,2003','2003-12-09 00:00:00','   Indiana Jones DVD Set James Bond DVD Set One or Three  '),(19824,3208138,'08,December,2003','2003-12-08 00:00:00','   adapter to play game boy games on the game cube Lord of the Rings Two Towers DVD Special Edition Metal of Honor Rising Sun Playstation 2 version Soul Calibur II for Playstation 2 Version Castlevania Lament of Innocence Playstaion 2 version  '),(19825,3208138,'08,December,2003','2003-12-08 00:00:00','   Beth Peterson 	* bethepete@hotmail.com 	 	  Cindi Peterson 	* mrspeterson@earthlink.net 	 	  Doris Peterson 	* dordorpete@earthlink.net 	 	  Audrey Wilson 	* engr@drecomm.com 	 	  kyle curtin 	* kylecurtin@comcast.net 	 	  rob weaver 	* toastgeek7@hotmail.com 	 	  Timothy Wilson 	* tgwilson@prolife.net 	 	  Glen Peterson 	* glenjpeterson@earthlink.net 	 	  Steve Curtin 	* s.curtin@comcast.net  Dick and Jane Remington * dicnjane263@earthlink.net  '),(19826,3208138,'08,December,2003','2003-12-08 00:00:00','   Richard and Jane Remington are also having trouble accessing the family yule blog. We will coach them once they arrive on the 17th. Then their lives will be complete. If you are getting a gift for them, contact me. Audrey should already have my mom\'s wish list on e-mail. Kyle needs my dad\'s, so I can help, if needed.  '),(19827,3208138,'07,December,2003','2003-12-07 00:00:00','   you guys are having too much fun. but that\'s not a bad thing... here\'s a better list than the one including minoxadyl:           a radiator-for a 1988 toyota corolla, if you have one lying around  gift cards/certificates for: target, tower books/records, mervyns, anywhere that sells food, guitar center, amazon.com, interpunk.com, cdnow.com, eastbay.com, amc theaters, blockbuster video, frys electronics, best buy, greyhound bus lines                   lakers/dodgers shirts  tickets for  urlLink lakers /dodgers games   urlLink manchester united soccer jersey-looks like this   since the cat idea was a bust, how about a pet rock?  pez dispensers-i already have c3po, batman, boba fett, and speedy gonzalez   urlLink a large, black or green one of these-    dvd-                                                                                                                          fletch, fletch lives, european vacation, anything with chevy chase, midnight cowboy  guayabera shirt(s)  cd-elliott smith:        , MIRV:                    , simon and garfunkel, santana: best of  pants-levis 501 preshrunk, dickies: tan, black or grey work pants without the cell-phone pocket, 34w 30L  cheese-mild cheddar, jack  '),(19828,3208138,'03,December,2003','2003-12-03 00:00:00','   I had to come all the way to Lancaster, PA for this image. I am sure that there are people who appreciate this kind of weather at this time of year. I liked how the sun was shining on this field, backlighting the trees and the fence. I am looking forward to being home the rest of December.  '),(19829,3208138,'03,December,2003','2003-12-03 00:00:00','   HELLO,  I am now posting as myself not as Steve on his computer while he is not at home.  More later.  Beth  '),(19830,3208138,'02,December,2003','2003-12-02 00:00:00','   Hello,  This will probably post as Steve, since I am using his computer (without permission, of course...) and I have attempted to sign in to Blogger but it will not let me.   I can access it as my user name and password as originally set up, but I cannot access the Peterson Yule Blog.  And, yes, Glen, these difficulties are NOT part of the Beth Plan as we all know and love.   If you could send be another invitation, I will attempt to gain access to the Holy of Holies, the inner sanctum, and share with the rest of the group.    Love to all, except Blogger... Beth, Mom, et al...  '),(19831,3208138,'02,December,2003','2003-12-02 00:00:00','   Santa Claus now has administrative previleges to correct and add html and links ;~)  And, now we can all add a title or headline. [You should be able to edit your own entries already if you need to change it after you have published. Just edit, preview, and publish again.]  '),(19832,3208138,'01,December,2003','2003-12-01 00:00:00','   subscription to   urlLink Wine Spectator                                    bookstore gift certificate:  Amazon Gift Certificates    music store gift certificate  urlLink Trader Joe\'s  gift certificate  '),(19833,3208138,'01,December,2003','2003-12-01 00:00:00','   Dear Santa: Trend Warning: personal items for women (sweaters, purses, etc.) are now featured with a monogram initial on them. (insert gagging noise)  Please be aware that I do not like this kind of thing. This trend can pass me by. And fast. Don\'t be tricked into thinking...oh, look, it\'s popular...I bet she\'ll like it. You\'d lose the bet.  '),(19834,3208138,'01,December,2003','2003-12-01 00:00:00','   This is my Christmas list for 2003  1.  urlLink Staples  Organizer desk lamp  $ 9.99   2.  A jigsaw puzzle made from one of the vacation pictures from Steve and Beth\'s      2003 Paris vacation.     3.  I can\'t remember the exact title but it\'s the biography of George W. Bush just      out.  [Claus note: this might be it:                      ]  4.  Time at the computer from those who exell and know how to cut and paste         etc.  '),(19835,3208138,'01,December,2003','2003-12-01 00:00:00','   i don\'t really want minoxadyl.  '),(19836,3208138,'04,January,2004','2004-01-04 00:00:00','   Wow!  Those Peterson girls look almost like sisters in the Formal Family Christmas Photo...(FFCP)  Beth (The Elder Peterson Girl, EPG)  '),(19837,3208138,'03,January,2004','2004-01-03 00:00:00','   Sorry for the size of this image.  I will work on that next, but it was a large and very beautiful tree picked out and brought home by David and Steve.  Beth  '),(19838,3208138,'02,January,2004','2004-01-02 00:00:00','   Here we are standing in front of wallpaper that looks like the Space Needle in Seattle. The real Space Needle (where this wall paper is located] has been painted 4 times since 1961. It now costs more to paint it [the real Space Needle] each time than it did to build it.  '),(19839,3208138,'02,January,2004','2004-01-02 00:00:00','     '),(19840,3208138,'28,February,2004','2004-02-28 00:00:00','   Beth and I saw Mystic River today. I am still feeling the tension created toward the end of the movie. That feeling that we can foresee something tragic about to happen and sit helplessly while it unfolds. That violence can\'t be rethought, retracted, or undone. That often in hindsight perspective is altered in ways that cause regret; intractable, irrepairable harm.  I am sorry to hear that Sam is struggling.  He is a \"good dog\" in ways that only a dog can be. My best wishes to Sam\'s recovery. Rob that struck me also about Spellbound. Those things said without question. Those \"truths\" that go unexamined.  \"The absolute value of love makes life worth while, and so makes Man\'s strange and difficult situation acceptable.\" - Arnold Toynbee  '),(19841,3208138,'26,February,2004','2004-02-26 00:00:00','   the contraversy, and fanaticism that has been brought out in americans by this movie is...interesting. i hear that it is indeed very graphic, and bloody. but there are at least three or four films at the blockbuster i work at which tell the story of jesus\' crucifixion. i wonder why this one warrants such strong (often fallacious) arguments on both sides of the issue, while these others, some made in the last five years, have slipped under the radar. i\'ll probably not see the new one, at least not on the big screen. i don\'t think there are any new discoveries revealed in the film, are there?  i was really sad to see sam struggling this afternoon. i fed the dogs, and he really didn\'t seem comfortable. he\'s the best dog ever, and he and i have many fond memories shared.  tomorrow is my twin cousins\' birthday bash. just kidding. but that is so cool how they have the same birthday, even though three (i think) years apart.  going back a few posts, i thought the diversity was the star of the movie spellbound. listening to the rancher who employed angela\'s father say that \"not all mexicans are reliable, but i guess there are a few mixed up in there,\" was really striking to me, and hearing the indian kid and his father who lived in san clemente rip through that list of words for hours was amazing. i cried when the boy from (i think it was missouri) was eliminated, and his younger brother proclaimed without irony \"he\'s still a champ in my eyes.\" my brothers are both champs.  Before God we are all equally wise - and equally foolish. -Albert Einstein  '),(19842,3208138,'26,February,2004','2004-02-26 00:00:00','   I saw the Passion on Wednesday in the afternoon.  I was really moved by the LOVE that Jesus had for each one of us that kept HIM on track with the plan for our redemption.  It was only His great PASSION for us, even before we were born, that gave HIM the power to tolerate that treatment without just calling it all off.  Thank you JESUS for your great love.  Doris  '),(19843,3208138,'25,February,2004','2004-02-25 00:00:00','   Our plan for gifts for Timothy and Daniel includes taking them to the  urlLink Emerald Nuts Quality of Life Expo  that is part of the  urlLink XIX Los Angeles Marathon  and the  urlLink Acura LA Bike Tour  and allowing them to purchase something within a pre-determined price range. The  urlLink Emerald Nuts Quality of Life Expo  has vendors from bicycling shops, athletic shoe manufacturers and lots of cool stuff. I think American Honda will have cars there--and if they want to apply our $20 to the purchase of a car, I will help them drive it home.  '),(19844,3208138,'24,February,2004','2004-02-24 00:00:00','   Maybe it\'s those photos of the Weber Grill Restaurant, but now I\'m hungry. Anyway, thought I\'d check in before I break for lunch here in the home office suites of Peterson & Peterson. Samson is improving. I\'m impressed and encouraged. Thanks for all your expressed empathy. The picture of Briana excavating her nose is CUTE. Beth - had to laugh at the \"psycho parents\" comment. A universal truth in education. I\'m going to see \"The Passion of the Christ\" tomorrow night. Peer pressure from the grass roots publicity efforts finds me swimming along with the church crowd. Curious to hear public response and to see if ticket sales keep up with opening day results. Nervous about the brutality. May have to come home and watch Sound of Music before falling asleep to calm down. Looking forward to a low-key Academy Awards party on Sunday night (not at our house, though, since Glen will be gone). Won\'t get to see Mystic River by then, I don\'t think. So my sentimental selections are: Best Movie - Lord of the Rings, Best Actor - Johnny Depp, Best Actress - Diane Keaton, and that\'s it so far. Don\'t know if they are worthy...just picked what I like. Happy Billy Crystal is hosting. That and playing fashion critic are my favorite parts. I laugh everytime someone says in pre-game interviews--- \"...and who are you wearing?\" I think that\'s the line. Maybe it\'s what are you wearing? Whatever --- it\'s meaningless and fun.  '),(19845,3208138,'23,February,2004','2004-02-23 00:00:00','   A lot has happened since I last visited the Blog.  I\'ve been spending the night aboard ship in San Francisco Bay with a bunch of 10 year olds and their psycho parents and in between doing about 3 full-time jobs.....  I\'m sorry to hear about Sam, and I want to reassure his parents that it is NOT the Pupcorn.  I can emphathize with the changes that age brings to health and bodily functions.   Yes, Rob is indeed an nice guy.  Beth  '),(19846,3208138,'22,February,2004','2004-02-22 00:00:00','   Does Timothy have a list?  '),(19847,3208138,'22,February,2004','2004-02-22 00:00:00','   Just watched Spellbound. Noticed the wide variety of families that these eight kids came from. How the different kids where invested for different reasons. Will watch it again. I would love to have the time to interview all of those kids now. Maybe there is a documentary in that.  '),(19848,3208138,'21,February,2004','2004-02-21 00:00:00','   Would he like J. S. Bach \"The French Suites\" Keith Jarrett on the Harpsicord or Deitri Shostakovich \"24 Preludes and Fugues op. 87\" Keith Jarrett on the Piano???  '),(19849,3208138,'21,February,2004','2004-02-21 00:00:00','   Money because I am saving for a pair of aerobars.  Cycle Sport issues: 	_ Sept â€˜03 (TdF) 	_ Sept â€˜01 (TdF) 	_ Mar â€˜04 (Current) urlLink    A couple of CDs of composers\' work, preferably from the Baroque or Classical time period like Corelli or Bach that no one in my family has already on a CD. It can just be something based around a musical instrument like a violin or harpsichord, or just a piece or pieces of music that I would like (interesting operas are fine, but I think drawn out opera is boring, and I guess I like music to be in there besides singing a large portion of the time)  \"2003 Year in Review\" book of pictures of the â€˜03 race season  urlLink  '),(19850,3208138,'21,February,2004','2004-02-21 00:00:00','   Posting a picture of a kid with her finger in her nose is taking the risk of offending others. If this picture is offensive please let me know and i will change it to one of her with her head on Beth\'s shoulder.  '),(19851,3208138,'21,February,2004','2004-02-21 00:00:00','   Glen your point was well taken. Here is a picture of Brianna picking her nose, not thumbing her nose at the rules, but digging away.  '),(19852,3208138,'21,February,2004','2004-02-21 00:00:00','   What could be better than a restaurant that cooks meat on giant Weber Grills indoors in the middle of Chicago? I can\'t think of anything better than  urlLink Weber Grill Restaurant  in Chicago.      These images are from their website and are hosted there. Therefore, when they change it, the images will disappear.  '),(19853,3208138,'20,February,2004','2004-02-20 00:00:00','   Editing of posts should be minimal. Restricted only to the cleaning up of typos and clarifying a few things. But to reduce the whole post from 3 paragraphs to one is so sad. It will be more fun if people take some risks here. So long as it isn\'t me! [Steve knows who I\'m talking about.]  I spent a couple of hours trying to figure out how to upload an image or two using the 10 mgs of \"free\" space that earthlink provides for my monthly subscription fee.   Ok, maybe it was only a few minutes. I will try again soon.  '),(19854,3208138,'20,February,2004','2004-02-20 00:00:00','   Glad to hear that there are treatments for Samson. Here\'s to his improved health and vitality. Looks like Beth and I are in for the Mexico weekend. Working out the pragmatics. Cindi makes Colonial Williamsburg sound enticing even in \"winter\",,, burrrrr. You know Beth and I, it doesn\'t take much to entice us to travel.  '),(19855,3208138,'20,February,2004','2004-02-20 00:00:00','   I made a trip to the vet yesterday. Samson had a scary wheezy breathing spell in the night, and I was concerned. It sounded like human asthma. He honked/coughed/gagged/breathed hard off and on all day. If you\'ve been around him in the last year, you\'ve probably noticed how he sounds when he is excited. Well, it seems worse, so we went to the Dr.  $202 later, we had a bunch of interesting information, several beautiful pictures of Sam on the inside (X-rays), and a bottle of magic juice, 4 oz. of which have to be squirted in his mouth three times of day. (Are you kidding?! Getting ear drops in Ashley\'s ear was bad enough! I have resorted to soaking a couple dog biscuits with the potion, having had no success with the doggy rodeo wrestling to wedge a squirt tube thing in his mouth!). Anyway, you\'ll be interested in knowing that in addition to airway disease (bronchitis) and allergies, he has a collapsing trachea. Who knew?! Yes, his air way is constricted halfway down...behind his chest bone. And it\'s very wide below that from the lungs trying to get the air out --- blowing. It\'s a genetic thing. He\'ll be 13 in August, so life is getting more complicated. I started the juice last night. He is doing better and is resting/sleeping easier when he needs to, so that is comforting. And Ashley scores a biscuit, too, everytime she\'s around for medicine time. I would have been a pushover as a parent.   Poor Sammy. But alas, he\'s his lovable self. Quite a trooper. The science of it all fascinated me. I enjoyed looking at the x-rays. I saw his bones (of course)...neck, head, and ribs. His very healthy heart. His narrowing trachea with pupcorn stuck in it (just kidding, Aunt Beth). And his lungs. No mass. No cancer. Thankfully. And if the medicine (it\'s new medicine --- just developed) doesn\'t work, he can get a stint put in (just like for heart problems). We\'ll be keeping our jobs to pay for his pharmaceutical habits. Because we love him. His head is on my lap as I write. Bless his paws. These dogs are great company with Glen traveling more. It\'s great to have Rob around, too. He\'s a \"nice guy.\" (Just had to throw that in to prove I read past blogs, too.)  '),(19856,3208138,'16,February,2004','2004-02-16 00:00:00','   I just caught up on my blog reading. It\'s enjoyable to read everyone\'s comments. Glen, I\'m in for the Mexico weekend. I know you know that, but I just wanted to confirm. Sounds fun.   Is anyone interested in Colonial Williamsburg some Christmas? I\'d really like to do that. I know it\'s not everyone\'s thing, but I love the time period and the area. I noticed that our timeshare exchange group has availablity there in December. This may or may not be the year, but with the timeshare, we could get housing for a group. Ahhhh. Christmas dinner at Kings Arms Tavern.   Valentine\'s Day was fun here. We celebrated at home, as is our tradition, on Friday night. We cooked dinner, as is our tradition. I took care of appetizer, bread, and made a salad...greens, feta cheese, dried cranberries, glazed walnuts, and rasberry vinagrette dressing --- one of my favorite salads (and homemade dressing...adapted the recipe from Houston Chronicle), and dessert (which had to wait). Glen made delicious salmon and filet mignon, plus a pile of broccoli. We didn\'t have dessert until Monday night! Too full on Friday.  I\'m working on a well-informed position with conscience and compassion about the whole gay-marriage issue. Want more than a shallow opinion, since I think it is complicated. Preparing an articulate response to the contraversy. I\'m digging in the Bible, too, to see what\'s true. I here some Bible comments made out of context to defend both sides, so I want to look for myself.  And hooray --- it\'s the weekend. Glen returns tonight from Chicago. That\'s a great thing.  '),(19857,3208138,'16,February,2004','2004-02-16 00:00:00','   Glen,  i saw your thoughts on the bike ride in Mexico.  I\'m with Steve about the ride but would enjoy a family get together in some nice place.  Mom  '),(19858,3208138,'15,February,2004','2004-02-15 00:00:00','   Beth and I had a great time in SF, the City by the Bay. Saw Noises Off and had dinner in Fishermans\' Wharf. Had time to roam the streets both by the Wharf and Union Square. Saw some wonderful art. I wouldn\'t recommend the Hilton as we could hear the conversations of our neighbors and the foot stomps of those above us. I didn\'t take many pictures. Was to busy having fun.     An electic streetcar in the Wharf area.  '),(19859,3208138,'13,February,2004','2004-02-13 00:00:00','   Steve, I guess that puts you in the: \"For those who like to see the sites and and just hang out--you can do just that\" category. That sounds great.   We can also do the birthday cake thing there!!!   For those who want to ride, more information is at the  urlLink Ensenda Rosarito web site .  '),(19860,3208138,'11,February,2004','2004-02-11 00:00:00','   I have clients who will not come to my office on a Friday the Thriteenth. I had one client who post-poned his surgery because of that date.  I only notice the date because Kyle was born on a Friday the Thirteenth.  Hope all enjoy their valentine\'s weekend. The September trip sounds fun Glen. There are a couple of people that have birthdays in that month. We will look at the pragmatics and get back to you. I don\'t think I could sit on a bike let alone ride one more than a few feet. I am already more than middle-aged and balding.  '),(19861,3208138,'11,February,2004','2004-02-11 00:00:00','   i also saw that this morning on the bbc website. i was under the impression that i\'d be a middle-aged bald guy by the time they released that. conveniently enough, it comes out a week before my birthday...  '),(19862,3208138,'10,February,2004','2004-02-10 00:00:00','   Did someone mention  urlLink this ?  '),(19863,3208138,'10,February,2004','2004-02-10 00:00:00','   In cleaning out my office at JSi if found so many interesting things. This quote was on a scrap of paper in my desk:  \"Success comes from experience; experience comes from failure. Eleanor Roosevelt said: \'We need to learn from the mistakes of others because we don\'t have time to make them all ourselves.\'\"   urlLink greatfailure.com  '),(19864,3208138,'10,February,2004','2004-02-10 00:00:00','   Here is an idea: September 25 is a 50 mile \"fun ride\" from Rosarito Beach to Ensenada in Baja Califonia Norte, Mexico. How about if we all meet there for the weekend. For those who like to ride bicycles--bring your bikes and ride. For those who like to see the sites and and just hang out--you can do just that. We could get a couple of hotel rooms or a suite somewhere. If anyone expresses an interest, I would be happy to make the arrangements. There is plenty of bottled water and other pure beverages and safe food there!  Note: Some of us are going on the April 17 ride and it is looks like some kind of guys thing. So the September date might be better for a family outing.  '),(19865,3208138,'10,February,2004','2004-02-10 00:00:00','   These are my next travel dates:  Chicago: February 17-20 Pennsylvania: February 27-March 3 Miami: the week of March 22 (exact dates not yet determined)  '),(19866,3208138,'09,February,2004','2004-02-09 00:00:00','   south america, europe, england, what a traveling bunch. that means pictures!!! everyone bring those cameras and remember you can post to the blog from internet cafes. beth and i will be in san francisco for valentines day, then in early april will travel to new orleans. six years ago i had no interest in traveling anywhere. i was more interested in having a home large enough to house anyone who wanted to visit. very much like the home in tuscany, from a movie i can\'t remember the title to, that that had multiple visitors. it may be that i want both now. i enjoy the long dinner, with interesting conversations, and the feelings that come with being immersed in conversations that make you forget time. i am learning from each trip what it is like to travel again. when to look through the lenses of a camera and when to just be with out working at being some place. i am glad that beth and i both like to wander, discovering as we go along what is there rather than looking for what should be there. i appreciate that when we can do that. to all of the travelers and travelers to be, please share your experiences here, it will be appreciated.  '),(19867,3208138,'09,February,2004','2004-02-09 00:00:00','   I also prefer calling you a nice guy.  In fact, that is what I do call you.  That is obviously what you are!  I was carried away by the fact that I actually stayed awake for a whole movie.  I rarely do that at home or in the theatre.  I think I was struck by the fact that I did not find any character in lit that I identified with, but the situation they were in, alone, in a strange placed that \"should\" have been exciting, all expectation with no payoff...I don\'t know.  I really like Secondhand Lions which we watched last night.  I cried when one of the old guys was threatening suicide, too close to home right now.  After all, as they say in the movie, \"What do you want me to do, die of old age?\" and my answer is yes.  We should all be able to do that...  '),(19868,3208138,'08,February,2004','2004-02-08 00:00:00','     '),(19869,3208138,'08,February,2004','2004-02-08 00:00:00','   besides, what\'s so good about having a complicated mind? our president doesn\'t even say \"nuclear\" right. anyways, i\'m going to carolina in my mind this summer. it\'s gonna be awesome. while i\'m there i\'ll probably watch the tour de france zoom by, and run with the bulls...in jacksonville. i\'m hoping i can afford this trip. actually, i saw a flyer at school for a three day trip to costa rica in july to study spanish. three days sounds brief, but that\'s just how i\'d prefer it. in rebuttle to beth\'s attack of my l i t opinion, i\'d like to reiterate that i thought it was \"nice,\" (see below) just that the climax at the end wouldn\'t be worth it to most. personally, i liked sofia coppola\'s film \"the virgin suicides\" better. there\'s some tweaked out relationships in that movie too, just like the uncommon one in l i t. the assertion that i am simple minded is quite falacious and irrelevant. er....irrelevent. irelevent. whatever. ah, forget all the fiction nonsense, spellbound was the best movie. ever. in the history of...my life. wait, that disney movie about v.d. i saw in my sociology class was pretty good. but that wasn\'t maid during my lifetime. i digress. keep it real everybody.  anyonebutbush \'04 ------------------------------------------------------------------------  '),(19870,3208138,'08,February,2004','2004-02-08 00:00:00','   Glen & I are planning a trip to Europe this July, pending frequent flyer miles. A friend of ours is returning to the Netherlands for her sister\'s wedding, so we can bunk there and tour a little with a native. That\'s what pushed us forward, beyond just expressing interest in a visit somewhere other than Mexico. Although, I am grieving the fact that I may not snorkel in tropical water this summer.   We are thinking about visiting London (and a shire around Oxford and a literary quest to where J.R.R. Tolkein and C.S. Lewis hung out), Paris (working at being in a town as the Tour de France zooms by...it begins on July 3rd in a small town, Liege, northeast of Paris...closer to the Netherlands, actually), and zipping by train for a day trip to Prague, as well. I want to get backpacks (for ease of travel and mobility...not for trekking Everest) and train from place to place, taking city busy tours. Then we\'ll come home to rest. I\'m excited, but holding my breath because costs seem prohibitive at this point in the planning. Just gotta sell a few more photo albums and stickers!  [Note to Glen...OK, now I\'ve blogged again. Can I return to the bedroom to sleep? Just kidding.]  '),(19871,3208138,'08,February,2004','2004-02-08 00:00:00','   We are also on a quest to see the \"Best Picture\" nominees. We still need to see \"Mystic River.\" I think we\'ve seen the rest. I think the Golden Globes have it right by splitting up genres. It\'s tough to critically evaluate comedy against drama, and same with the acting roles. I think Johnny Depp made that \"Pirates\" role so much more than it was, as opposed to Bill Murray\'s portrayal of LIT (since everyone refers to \"Lord of the Rings\"...LOTR). Evidently Sean Penn is the favored winner.  \"Master & Commander\" was a good story and I thought it was well crafted and a fascinating portrayal of the time period and military life at sea. Of course, Russell Crowe did well in the lead, but his character was not compelling to me. It did not engage me emotionally. I would see it again just to see the Galapogos Islands again.  '),(19872,3208138,'08,February,2004','2004-02-08 00:00:00','   In an effort to keep Ashley from having an \"accident\" where she sleeps in the laundry room, Cindi has been taking both Sam and Ashley out to the back yard just before bed where they respond to the command: \"potty.\"  Then they all crash back into the house for their reward biscuits. It is fun to watch. They just finished while I blogged about it from the couch.  '),(19873,3208138,'07,February,2004','2004-02-07 00:00:00','   Daniel Wilson told me this morning that he will ride in the  urlLink 10th Annual LA Bike Tour  with me on Sunday, March 7, 2004 at 6 AM.   Rob and I went on a nice 26 mile ride on Saturday and was asking me how severe the hills are in the  urlLink Rosarito to Ensenada Ride . So, I wonder if he is thinking of riding with us next time. That ride is on April 17 and is 50 miles. The hills, by-the-way, are steep and long. As steep as going up Turnbull Canyon Rd. and at least twice, maybe 3 times, as long. If you check out their website they give some details on the route and the steepness.   So far Rob is planning to go for the trip [hey Rob--ride too!] and David said he\'d like to go and ride. Also my friend  urlLink Ken Brown  is going--he is already registered. And, Bob Mendoza from church is thinking about it.  '),(19874,3208138,'07,February,2004','2004-02-07 00:00:00','   I would like to add Kevin Costner to the list of actors who, if not playing himself, every role is the same character.   Gods Must be Crazy I is also one of my all time favorites. I don\'t know that the actors had a lot of depth or could play any other parts, but that was not the point with this very orignal film. GMBC II did not take the story any further, just an ill fated attempt to build a franchise from the success of #1.  Lost in Translation lost me, not because of its complex plot, but because it failed to move me emotionally to empathize with the characters. I was  thinking that a better title for that film would be \"Jet Lag.\"   Note to printers of images: the resolution of the posted version of the picture must be rather low. I hope that there were other forms of the picture from all those cameras in play that evening. If someone wants a higher resolution version of that, or any, photo let me know so I can email it to you.   Note to Cindi: Try to remember your user name and password so you can blog some of those great thoughts.  '),(19875,3208138,'07,February,2004','2004-02-07 00:00:00','   I also wondered about the nomination for Bill Murray for his part in \"Lost in Translation.\"  He is one of those actors, like Clint Eastwood, who play himself in every role.  Wish I could get a nomination for being authentic in my own life...  Beth  '),(19876,3208138,'07,February,2004','2004-02-07 00:00:00','   Hi, Mom,  I will send you the picture.  I have two copies sitting on my desk right now.  Love, Beth  '),(19877,3208138,'07,February,2004','2004-02-07 00:00:00','   I still have not seen all the Academy Awards nominees.  That is my goal betwen now and February 29.  I will pass on \"Cold Mountain\" which was not nominated.  Anyone else out there who has seen that movie?  is it worth it?  I am not a Nicole Kidman fan, so if it has to do with liking to see her, forget it...  Beth  '),(19878,3208138,'07,February,2004','2004-02-07 00:00:00','   Just wait until you are middle-aged.  \"Lost in Translation\" will not be so \"fuzzy.\"    Love, Mom  P.S. Maybe what I am trying to say is lost in the translation...  '),(19879,3208138,'07,February,2004','2004-02-07 00:00:00','   I haven\'t seen a cute button either, but Brianna is cute. And, \"the gods must be crazy\" is one of my all time favorites. I am looking forward to seeing American Splendor, my catholic irish roots may be showing. look out when Weber is back.  '),(19880,3208138,'07,February,2004','2004-02-07 00:00:00','   How could you possibly get lost in \"Lost in Translation?\"  I used to think so highly of you and your intellect...It was, actually, a very slow starting movie.  I myself wondered if I could stay awake during the slow unfolding of the story.  (It is a well known fact that once the sun goes down, Beth is asleep.)  I was fascinated with the characters in the movie and was very satisfied with some of the choices they made.  Maybe being in the throes of my own midlife crisis made me sympathetic...  Signed, Love my brother anyway...  '),(19881,3208138,'07,February,2004','2004-02-07 00:00:00','   I am assuming that this is the way to leave a message.  I really like the family picture  that euther Glen or Cindi put or here.  I would like a copy but didn\'t know how to  print it from here.  Could someone let me know.  Thanks,  Mom, Gram, Doris  '),(19882,3208138,'07,February,2004','2004-02-07 00:00:00','   lost in translation was...nice, but ultimately, a lot to sit through for the fuzzy pay-off at the end. i still think american splendor was awesome, so nanner nanner glen. mystic river is my vote for most awesome movie last year. without lord of the rings money, and cgi, it tells a good story. it could be cited in an argument against capital punishment. it\'s my personal opinion that everyone should watch \"the gods must be crazy\" at least once a week. that way, if you are human, there\'s no way you\'d consider hurting another human. perhaps that\'s a little dreaming. oh well, \"we are the music makers and we are the dreamers of dreams.\" that brianna\'s cute as a button. can\'t say i\'ve seen any cute buttons in my life, so let\'s say she\'s cuter than one. look out when we get kobe back.  '),(19883,3208138,'07,February,2004','2004-02-07 00:00:00','   We saw Seabiscuit last night. It is a good story with solid character development. Special effects, gratuitious violence and chase scenes has to be cut for a lack of time.   I got lost in Lost in Translation. People wondering around looking confused. Can someone explain why Bill Murray is nominated for an award on this one? Maybe if I had just seen it in the original language, I wouldn\'t have been so lost.  '),(19884,3208138,'07,February,2004','2004-02-07 00:00:00','   Beth, Kyle and I watched Lost in Translation last night. Beth and I both liked it, but for different reasons. I thought it captured what happens when people drift spiritually and emotionally. The characters were using this vague relationship to give direction or substance to their experience. Beth wants to see Mystic River next. We rented Jet Lag a couple of weeks ago. A movie similar in someways to Lost in Translation. A french movie focused on chance meeting and relationship confusion.  '),(19885,3208138,'07,February,2004','2004-02-07 00:00:00','   hey everyone. i think that just includes steve and glen right now, but hi if you\'re not one of them and you are reading this. this is more or less just a test to see how this is going to work. so far, so good. so....um.....seen any good movies lately........anyone? i will now attempt to upload a picture.  '),(19886,3208138,'07,February,2004','2004-02-07 00:00:00','   Way to go Steve!  '),(19887,3208138,'07,February,2004','2004-02-07 00:00:00','   This is Brianna eating her dinner.  '),(19888,3208138,'07,February,2004','2004-02-07 00:00:00','   I am not using my web space at this time. I can set up that site so that anyone could ftp to it with the password. I will look up the address and find the password.  '),(19889,3208138,'07,February,2004','2004-02-07 00:00:00','   To remove the Christmas limitation to the blog. I have changed the url for our family blog.   Now I have to figure out how to let everyone know the new address.  '),(19890,3208138,'07,February,2004','2004-02-07 00:00:00','   Since I no longer work at the same place, we may loose access to the server where I have the photographic images stored. Does anyone else have access to a server for that purpose? I think, if we are really clever, we could also ftp all of the content to another place so we would not have Blogger advertising at the top of the blog. Any suggestions?  hmm, one other thought is that I think we could pay a fee to blogger.com to upgrade the service to remove advertising and have storage space from images.   I will look into these. Let me know if any of you computer geeks have any ideas.  '),(19891,3208138,'07,February,2004','2004-02-07 00:00:00','   I have taken the word \"Yule\" out of the title. This may help people know they can blog outside of the Christmas season. I will try to change the URL to be more season inclusive. I don\'t know it that kind of change is allowed. I will try to figure that out.  '),(19892,3208138,'07,February,2004','2004-02-07 00:00:00','   I think there should be more uses for this family blog. But that means people will have to blog!  A friend of mine who just went to work in Bangledesh has started  urlLink this blog . I\'d like to say that I got him started and coached him through his first posting of the images. --because I did do that!  I think we should keep up the postings. Especailly when something interesting happens. Like my work with MEDA was mentioned by some of our clients on KCET\'s \"Life and Times\" program last night. I don\'t know how long they leave it up there but today you can listen to the program on their web site:  urlLink here .  '),(19893,3208138,'07,February,2004','2004-02-07 00:00:00','   just checking to see if anyone else checks the blog. is there more purpose than just at christmas? take care steve  '),(19894,3208138,'31,March,2004','2004-03-31 00:00:00','   Brianna was here for two days. Beth and I took her for a walk both mornings. She is a lot of fun. The \"Sacramento\" Kings have several games on TV this week so I will watch unless they play badly. Kyle is applying for his first job this week. He wants to work at Boys and Girls Club. Help with computers, homework and video games. Pretty \"normal\".  '),(19895,3208138,'30,March,2004','2004-03-30 00:00:00','   . . . Is allowing me to access the Interent from the OCS conference at the Westin Bonaventure in Los Angeles. If anything good happens I will be right on top of it and blogging live.  '),(19896,3208138,'30,March,2004','2004-03-30 00:00:00','   I am working in Los Angeles this week. Today and tomorrow I am attending the DHHS funding conference at the Westin Bonaventure in downtown. Part of my work is to help younger organizations access larger funding sources. So, MEDA is providing technical assistance to some of our groups through a consultant. I had our consultant meet with 4 of our organizational clients in Los Angeles. Three out of the 4 showed up. So that is progress for them.  Cindi went with me as we took the MetroLink train and the Metro Red Line subway to the event. Cindi spent most of the day at the Los Angeles library--a very nice building worth a visit. If we are all really nice to Cindi, maybe she will blog about her day.  '),(19897,3208138,'28,March,2004','2004-03-28 00:00:00','   My co-worker\'s mom vanished on Tuesday...not like a trick in some Vegas magic show, but a tragic event to rip your world in two. My co-worker is 30. She had lunch with her mom that day, and no one has seen her since. Her car and her purse are still at home. She is 57 and lives in a very nice part of Pasadena. I am troubled because my friend/co-worker has to survive this ordeal. I\'m also troubled because her dad will be questioned thoroughly, since husbands are often involved in situations like this. I\'m troubled by the way the news stories stacked up on Channel 4 --- missing Pasadena woman, followed by sentencing of a man who was found guilty of murdering his wife, followed by the story of the search for a man exposing himself to teenage girls in the So-Cal area.  And I\'m sad that I have second thoughts about going places alone in this crazy world. Time will make me bold and brave soon, but for now I\'m quite distracted. I pray for a happy ending for my co-worker\'s family. It\'s all so curious to turn on the evening news in a major metropolitan area and find out you know the family in the lead story.  I saw the intellectually challenging and deeply moving epic movie \"50 First  Dates\" (Drew Barrymore and Adam Sandler) on Thursday. [tongue-in-cheek] To keep a marriage healthy, I let Glen opt out of these types of films. I did enjoy it, though I grew tired of the typical crass comments forced into the dialogue --- mostly from the other actor...I think his name is Rob Schneider. My expectations were low, but I was pleasantly surprised. No need to own it, but it was a pleasant distraction from a rough week.  We got our third set of photos of our new puppy and his brothers and sisters. This week they are walking and starting to eat solid food. They are beyond cute - chubby little balls of cute with stubby legs. Our puppy and Brianna in the same picture --- now that would be CUTE.  '),(19898,3208138,'24,March,2004','2004-03-24 00:00:00','   Here I sit in a Starbucks in South Miami, Florida. It is so boring here. There are students sitting around in large chairs and small tables studying, working on laptops, and studying for the MCAT. After my last meeting this afternoon I drove down to Key Largo. I think that would be a nice place to visit when I have more time and someone with me to do it. Actually, I\'d like to drive all the way to Key West and see where Ernst Hemingway used to hang out.   Starbucks has T-Mobile wireless access--wi-fi. So I am sure that I will be back here a few times before I head home.  [The MCAT studiers just left--one of them is the tutor of the other. The tutor was telling here about getting a physics tutor--not his specialty.]  My colleague from MEDA arrives tonight from Canada. He will write about some of the groups with whom we work here in Miami for the MEDA official publications. I don\'t know what time he arrives.  I am trying to decide where or whether or not I will eat dinner tonight. I am sure I will be hungry later on. It is after 7 east coast time as I write this.  '),(19899,3208138,'21,March,2004','2004-03-21 00:00:00','   I built a sandbox and Beth bought a swing. Here are pictures of Brianna enjoying both.    Brianna on the swing.   Brianna and mom in sandbox.  '),(19900,3208138,'16,March,2004','2004-03-16 00:00:00','   Brianna discovered dirt and plants on the deck tonight. Here she is putting dirt in a candle holder and eating fresh chives that she picked from the planter.    Tasting the chives.    We have the dirt in the little container.  '),(19901,3208138,'15,March,2004','2004-03-15 00:00:00','   Once again we are blessed with a visit from Brianna. Shawntea and her are spending the night. Kyle and grandpa are holding fort in the Curtin/Peterson/Weaver household. She is so busy I can\'t take a picture. For the record her hair has more curl and she is very busy. Beth is off seeing 21 grams. If you talk to her ask her about being stuck in the mud in a remote area and how she got rescued.  '),(19902,3208138,'08,March,2004','2004-03-08 00:00:00','   while the early morning loonies were out running/cycling, i was home sleeping. it was a blast. if i was at the event, i\'d have been one of the kids eating pancakes in front of the eager masses. (who knows how long they were to be there handing out food to the thousands of participants). perhaps it\'s still my instinct to defend those poor souls in the food service biz.   i went to santa cruz a few days ago for a concert. i\'ve been making that trip periodically throughout my stay here at the peterson residence, but this time, that drive seemed unbearably lengthy. i sensed that this last trip would be the last of it\'s kind for me, as i\'m an adult now, and perhaps i should be applying myself to other more ambitious goals.  dude, it\'s like 80 something degrees here! nanny na na, nor-cal!  still haven\'t seen \"the passion...,\" but i think i\'ll see it later, as the hype dies down a bit. i saw 21 grams last night, that was...unsettling i guess. good overall though.  this time of year is always my favorite as the weather changes, and baseball season nears. father\'s day weekend is sold out at dodger stadium already, since they will be playing the new york yankees. i\'ll have to watch the blue crew beat a-rod and co. on tv then.   okay, that\'s all.  wait......no, that\'s all.  '),(19903,3208138,'02,March,2004','2004-03-02 00:00:00','   Daniel Wilson and I rode our bicycles in the  urlLink LA Bike Tour  that is connected to the Los Angeles Marathon on Sunday, March 7. 15,000 plus riders assemble on Exposition Boulevard between USC and the Coliseum before 6 AM to ride along most of the Marathon route. At the end of the ride, my odometer read that we had gone 22.8 miles. That is not a long ride at all.     As the ride starts at 6 AM, as the sky starts a pre-dawn glow that announces that the sun will rise soon, riders are packed in tightly and it take 20 minutes or so before you really start to move. It takes longer than that until you can safely assume you can stay clipped into your pedals as cyclists with varying degrees of skill are weaving all over the road. The announcer on the loudspeaker system is cheering us on as county and city elected officials take turns making comments on the wonderfulness of their work in the community that made the event possible. Randy Newman&#8217;s &#8220;I Love LA&#8221; is blaring through the speakers too. It is a cute song that mentions a bunch of the Boulevards that we will be riding on-- Olympic , Pico, Venice, etc.     Some fun things: &nbsp;Cindi and my mom, Daniel&#8217;s Grandma, AKA Doris, found a spot along the route near a Starbucks to watch us ride by and cheer their little hearts out. It is fun that all the streets are closed down and without any car and truck traffic we can ride down some usually very busy streets as the sun rises above Los Angeles. There were a lot of funny looking people&#8212;including Santa Claus&#8212;riding their bikes. I didn&#8217;t see anyone crash this year. We finished our ride before it got hot! (the runners had a different experience with unseasonably warm weather exceeding 80 degrees F.)     Some less than fun things: Did see some ambulances coming and going from the route. It is hard to find people at the end of the ride with 15,000 plus people wondering around in the quad area at USC. Kids serving a pancake breakfast at USC to support a charitable cause, hung out eating pancakes in the cooking area while people who had paid for breakfast stood in a line watching them eat pancakes and sausages while the line just barely moved.     The whole Marathon phenomenon in LA for the weekend is energizing and is an experience to behold. I remember feeling this last year too&#8212;and haven&#8217;t acted on it yet&#8212;but I imagine the glory of training as a runner and completing the running Marathon. There are 20,000 plus runners, in addition to the elite athlete runners. A couple of people I know have run in this event&#8212;Lalo Diaz and Brad Peters for example. One woman from our church ran yesterday. I don&#8217;t know who she is but someone else told me about her. David Diaz and Lalo told me they would like to run. Maybe this year is the year to get those slow twitch endurance muscles ready to go.    Cindi and Doris got interviewed by an LA Times reporter whose article appeared in today&#8217;s paper. No mention of us though. The news coverage really only went as deep as the running event. Another reason to think about running I guess.  '),(19904,3208138,'02,March,2004','2004-03-02 00:00:00','   A fox on our deck.  '),(19905,3208138,'01,March,2004','2004-03-01 00:00:00','   Brianna was dropped off this afternoon with a runny nose. Steve did pretty well for a couple hours. Then Brianna got very cranky. Beth arrived in a nick of time. We will not post the a picture of the drainage. Just use your imagination.  '),(19906,3208138,'01,March,2004','2004-03-01 00:00:00','   Friday and Saturday of last week I was in an international training event hosted by USAID and my employer, MEDA. The training covered something called &#8220;Business Design Services&#8221; that is the preferred method of international community economic development used by the US and Canada. The trainees were mostly MEDA staff working in Peru: Lima, Maputo, and Tarapoto; Managua, Nicaragua; Dar Es Salaam, Tanzania; Mozambique;Khajand, Tazjikstan [former Soviet Union republic]; Lancaster, Pennsylvania; Waterloo, Canada; Haiti; and Los Angeles, California. Some staff from Haiti were unable to come to the training because of the threat of civil unrest in Haiti that escalated to Haiti&#8217;s President Arristide fleeing the country on Sunday, February 29. This two-day training event was held in a hotel in Manheim, Pennsylvania (near Lancaster).    Sunday, the group moved to a retreat center outside Lebanon, PA for an all MEDA staff retreat that occurs about every three years. The group is double the size of the training that I just came from and consists of approximately one third of the 150 or so people working on staff for MEDA world-wide. This event ends mid-day Wednesday. So far the agenda has focused on team building kinds of group exercises. The experience is somewhat less goofy that the AmeriCorps*VISTA supervisor training that I was coerced to attend in 2000. The group overall is an eclectic, multi-national, multilingual mix of youthful, energetic, intelligent, overachieving folks that make this an energizing experience that will be memorable. &nbsp;    MEDA&#8217;s international headquarters is in Canada. I will guess that over half of the staff here is Canadian. Hockey anyone? The big deal for everyone is that it is so unseasonably warm here at 15 degrees C. [That&#8217;s almost 60 F for us Americans.]  '),(19907,3208138,'30,April,2004','2004-04-30 00:00:00','   Well, I know everyone thought I was younger than my older brother, but, if he is watching a movie that has not even arrived at his home until after 8:00 P.M., I must be older.  Mom, when were you planning on telling us the truth.  It is now 8:48 P.M. and I have been admiring my own stamina at remaining awake as Kyle and I have been discussing the origin of the universe and the need for everyone to find a place in it when what i really want/need to do is get some sleep.  After all, tomorrow is another day.  My favorite movie line.....  '),(19908,3208138,'28,April,2004','2004-04-28 00:00:00','   Every review has been excellent. It was one of the movies on my list to see on DVD. Let me know what you think.  '),(19909,3208138,'28,April,2004','2004-04-28 00:00:00','   Does anyone know about this movie?                                    It comes out on DVD next week. I plan to watch a preview when Rob brings it home tonight. Don\'t tell anyone.  '),(19910,3208138,'28,April,2004','2004-04-28 00:00:00','   Doris, thanks for your questions. The map shows the different stages of the \"le tour de france\" bicycle race. It starts on July 3rd this year. A cable channel has a series of shows on Thursday that give background and detail leading up to this race. While watching it last week I saw bicycle racers on a section of cobblestones. Motorcycles that were filming the race fell in the way of some of the racers and disrupted the race. That is the picture that Glen posted.  The only trip Beth and I have planned is the one to Hawaii from July 26 to Aug 2. The American Psychological Association convention is from July 28 to Aug 1. One of my friends Kirk Schneider is getting an award for outstanding achievement as an independent psychologist. He also is talking about his new book on Awe and it\'s importance to our sense of well-being. I also want to talk to publishers about the two book projects that are important to me. One on the \"art\" of psychotherapy with children and the second on my understanding of Jim Bugental\'s teaching and writings. Jim is/was my mentor for two decades and I feel very strongly that his work has heavily influenced my work with all of my clients. I am hopeful that you will find someone who will say \"yes\" and join us in Hawaii. Thanks again for your questions.  '),(19911,3208138,'28,April,2004','2004-04-28 00:00:00','   Finally took time to check on the Bloog site.  Guess I shouldn\'t take so much time off again as so much has changed.  The Wilson \"boys\" and all the new pictures of all the young in the family, Sam and Brianna.  I do have a question about the map that Steve put on.  Are you going to be traveling on that map this summer?    I am so excited about the possible apportunity to go to Hawaii.  I just didn\'t think it would happen so soon as Mary and Bill are definitely not able to go this year.  Am still trusting that someone will say yes.  Did enjoy the updated pictures of Sam.  I didn\'t really understand the conversation about cobblestone vs ather types of roads but then I don\'t understand a lot of things and it doesn\'t really bother me.  '),(19912,3208138,'28,April,2004','2004-04-28 00:00:00','   Finally took time to check on the Bloog site.  Guess I shouldn\'t take so much time off again as so much has changed.  The Wilson \"boys\" and all the new pictures of all the young in the family, Sam and Brianna.  I do have a question about the map that Steve put on.  Are you going to be traveling on that map this summer?    I am so excited about the possible apportunity to go to Hawaii.  I just didn\'t think it would happen so soon as Mary and Bill are definitely not able to go this year.  Am still trusting that someone will say yes about being my traveling companion.  Did enjoy the updated pictures of Sam.  I didn\'t really understand the conversation about cobblestone vs ather types of roads but then I don\'t understand a lot of things and it doesn\'t really bother me. I didn\"t realize motorcycles were running while the bicycles were.  '),(19913,3208138,'28,April,2004','2004-04-28 00:00:00','   I was at the Kings - Dallas game last night. I am firmly convinced that the Kings want to give their fans heart attacks. What a roller coaster of emotion. It is so loud in Arco. I have no voice left and I can hardly hear. Will make me a better therapist today.  '),(19914,3208138,'28,April,2004','2004-04-28 00:00:00','   Well. I thought I was hiding my anxiety pretty well with not being able to reach David via cell phone between April 11 and April 22, but he did call me to tell me everyhting with him was great, it was his cell phone that was having problems.  I can now sleep through the night, well almost, and he wanted me to pass on that his mailing address remains:  David Weaver P.O. Box 928 Kings Canyon National Park, CA   93633  '),(19915,3208138,'27,April,2004','2004-04-27 00:00:00','   Well. having lost my head myself, I have empathy for the llama...Samson is too cute and should be outlawed in some way.  He is the desktop picture on m computer at school and has been very distracting during STAR testing this week.  I think I need to notify the state that I have administered the STAR in a non-standard way with students who need to look at and swoon over the Samson picture between questions.....  '),(19916,3208138,'27,April,2004','2004-04-27 00:00:00','   Samson and I are getting along OK without the one with the biggest shoes in the house. He makes so many sounds. (Sam, not Glen) I never knew puppies got the hiccups and burp. He even has a sound that mimmicks Donald Duck. Of course, the wimper-chirp noise is too sweet. But I practice tough love and don\'t run to him at those times --- unless I suspect it\'s bathroom time! He spends his outside time biting off flowers and other green and not so green plants. And every now and then I pull a piece of bark out of his tiny mouth. It\'s fun to watch him grow. Sometimes I have to carry him outside to do the business. That\'s just my excuse to cuddle him while I can. We call that strategy the \"swoop and poop.\" We swoop him up and get him outside just in the nick of time...or that\'s the plan, anyway. It\'s fun. And an adjustment. Having some little life to care for keeps us childless folks from becoming quite so self-centered. I like Samson a bunch. And yes, the llama had a seam split and suddenly poly-stuffing filled the room (almost) like a marine inversion (clouds). So llama is retired until he gets stitches.  '),(19917,3208138,'27,April,2004','2004-04-27 00:00:00','   I am glad that I am not the only posting \"cute\" pictures anymore.  '),(19918,3208138,'27,April,2004','2004-04-27 00:00:00','   It takes some persuasion for Sam to convince Ashley that she should play. This all seems to be related to which dog will become the \"alpha dog.\"     The tug-of-war game has become a favorite. I hear from Cindi today (I am in Chicago posting this) that the stuffed Llama has lost its head, so to speak.      At least Sam is the \"Alpha dog\" over the stuffed toy.  '),(19919,3208138,'27,April,2004','2004-04-27 00:00:00','   I am in Chicago again. I will be here until Friday visiting my CCF clients.  '),(19920,3208138,'26,April,2004','2004-04-26 00:00:00','   thanks for your impression of the song. it does have that quality.  when i was in vietnam we would sit on a ship off the coast, in the heat, and i often would find a place to sit and watch the ocean. it was one of those times im my life that i could physically experience something larger and more enduring than myself. float on reminds me of those times and the comfort of being part of something so much larger than ones self.  i hope you get to the beach. take pictures, post them on the blog for those of us who pine for the ocean.  '),(19921,3208138,'26,April,2004','2004-04-26 00:00:00','   so steve, you downloaded float on? it\'s a great song, isn\'t it? it has the sort of melancholy, and optimisim i feel like i\'d want people to leave my funeral feeling. i won\'t say any more, except that i love baseball. it gives me an escape that allows me to succeed at something vicariously on a daily basis. when things in baseball aren\'t going well, then my own endevours seem successful. it\'s great to pour your hopes into something other than my own tired routine sometimes. record high temperatures in the southland today. 99 in long beach. there was also 15 foot waves at newport beach today. wish i had a chance to see that. welcome wilsons.  '),(19922,3208138,'26,April,2004','2004-04-26 00:00:00','   Welcome Daniel.  '),(19923,3208138,'25,April,2004','2004-04-25 00:00:00','   Welcome to the Wilson guys.  '),(19924,3208138,'25,April,2004','2004-04-25 00:00:00','   Daniel\'s on thanks to Timothy (and Glen)  '),(19925,3208138,'25,April,2004','2004-04-25 00:00:00','   Welcome Timothy. Looking forward to your contributions.  '),(19926,3208138,'22,April,2004','2004-04-22 00:00:00','   Well, Timothy is now on with a working acct.! I beat Daniel  :+)  '),(19927,3208138,'22,April,2004','2004-04-22 00:00:00','   We are still trying to get Daniel signed up properly. We have tried a second or third invitation, let\'s see if that does the trick.  '),(19928,3208138,'21,April,2004','2004-04-21 00:00:00','   Hi, Daniel,  In order to post under your own name, you need to ask Uncle Glen to \"invite\" you to the Blog via your Mom\'s (or Dad\'s) e-mail address.  After you accept the invitation by creating a User Name and Password, you can log on to the Blog under that name from anyone\'s computer and be viewed as the \"poster.\"  I am posting from Steve\'s computer right now, and had to click on LogOut before I could log on as myself as his computer \"sees\" him as the default.. Let me know how it works out.  Aunt Beth  '),(19929,3208138,'21,April,2004','2004-04-21 00:00:00','     '),(19930,3208138,'21,April,2004','2004-04-21 00:00:00','   Brianna at Easter.  '),(19931,3208138,'21,April,2004','2004-04-21 00:00:00','   As long as I donâ€™t have to ride on the cobblestones, it is fine if other people do it. Even the rough sections on concrete or gravely areas of the bike path make my weary bones talk.  I am amazed at the way the TV reporters and camera folks are able to stay out of trouble on these races. I am glad they do such a careful job of it. I did see a picture somewhere on the internet of the motorcycle/scooters lying on the ground with the bicyclists riding around them. I was trying to find it to post here. I will look some more and post it if I find it.   I bet Mr. Liggett had a few comments on this if he was covering the race.   UPDATE:  '),(19932,3208138,'19,April,2004','2004-04-19 00:00:00','   It looks like the next tele coverage of bicycling on OLN is Thursday, tomorrow.   Sometimes I type my messages in a word processing program so I can get some spell check helpâ€”especially now that I know the word police are watching.  Speaking of word processing, my Microsoft Word stopped working sometime this week and I ended up having to call Microsoftâ€™s customer service. After the very efficient person on the line got all of my customer information, including my time zoneâ€”Pacific Daylight Time, I asked him what time zone he was in. The answer: India Standard Time. Interesting.  After I had waited on hold for 45 minutes, my problem with Word was resolved in about 10 minutes. Thank you very much!  '),(19933,3208138,'19,April,2004','2004-04-19 00:00:00','   Just checked the schedule and it was a replay of the Paris-Roubaix race from the weekend of the 11th.  '),(19934,3208138,'19,April,2004','2004-04-19 00:00:00','   Your right, banned was the correct word.  Very good answer.  I will have to look at the tv guide and see which race it was that I watched this am. Will let you know. It was painful to watch the leaders fall because a motorcycle rider fell in front of them. Having a cycle rider fall and disrupt the race seems different than have one of the motorcycles disrupt the race. Not fair some how, but Lance Armstrong might say the same thing about people with bags clogging the route. Thanks again Daniel and for letting me know who answered my post.  '),(19935,3208138,'19,April,2004','2004-04-19 00:00:00','   First of all, I am asuming that Steve meant \"banned,\" not \"band\" when he was talking about the motorcycles on cobblestone roads. For the bikes on cobblestones issue, I will tell about both points of view. The Belgians (or maybe just the ones from Flanders) really like them because it can allow a skilled rider to pull ahead by the path he takes through them.  On the other hand, my dad and brother (and probably most Americans, for that matter) disagree because they are really hard on the riders because they are so worn from hundreds of years\' use, are narrow, and therefore dangerous, etc.  Only two Classics, or World Cup races have them as far as I know (the Ronde van Vlaanderen aka Tour of Flanders, and Paris-Roubaix), which can be avoided even if you want to win the World Cup.  What a non-cycling addict might not know about is the UCI (pro cycling\'s governing body) will have what they call the \"ProTour\" in 2005, which states that if you want to participate in any event, you will probably have to ride them all (there\'s another long story behing that), which means that you will have to ride those races with cobblestones, which are major events on the calendar.  Now for the question about motocycles riding them at the same time, I also state two sides.  It would be good to clear the road more for the riders.  However, without the motorcycles, there would be no one to give us the live coverage, which the sponsers want because of all the ads on the riders and their bikes.  That means all the coverage we get is from the journalists that are right there with the riders to clog up the sides of the roads, which the spectators want.  Basically, I think the motorcycles should stay, and you should only have to ride the cobbled roads if you want to, however, I probably have not seen the races the others have because of a lack of OLN. (Posted by Daniel, who is still puzzled on how to get his own name on the blog under an article.)  '),(19936,3208138,'19,April,2004','2004-04-19 00:00:00','   Should bicycle races involve cobblestones and if they do, should motorcycles be band from those sections? After watching on OLN this am I wondered what Glen, Daniel and Timothy thought about that.  '),(19937,3208138,'19,April,2004','2004-04-19 00:00:00','   I believe now is the time to have a PWWCR Team Musette Bags.  If bicycle teams can have them, so can we. I even found a site to order them from. They would look like this except they could have \"PWWCR\" on them.  There is only oen drawback. The smallest order is one hundred bags. Not sure there are that many pople in this family. We might have to recruit more people.  '),(19938,3208138,'18,April,2004','2004-04-18 00:00:00','   All this talk about musette made me go to good old Webster and find out exactly what it was.  Low and behold it is either a bellows-blown bagpipe popular in France in the 17th or 18th century (Daniel could have used it for his report) or a small knapsack with one shoulder strap.  Glad I looked it up.  '),(19939,3208138,'18,April,2004','2004-04-18 00:00:00','   Beth, Kyle and I want to offically welcome Sam 2 to the family. He looks like he could steal all of our hearts.  '),(19940,3208138,'18,April,2004','2004-04-18 00:00:00','   Here is Sam 2.     He sleeps, romps around for a few minutes, eats, drinks. And, in between all of that he poops and pees. So far, mostly outdoors.  '),(19941,3208138,'16,April,2004','2004-04-16 00:00:00','   An interesting discussion at lunch today led to the idea of a school report on the  urlLink history of pizza . But I think the  urlLink history of bicycling  might be an interesting choice as well. But either of these two projects would do much to enhance Daniel Wilson\'s academic experience and that of all of his 7th grade colleagues.   By-the-way, Daniel may be joining the blogosphere soon. And hopefully we can get Timothy involved too.  '),(19942,3208138,'15,April,2004','2004-04-15 00:00:00','   I am so excited. Tomorrow is the Grand Day of the Puppy. Starting with a trip to the spa for Ashley, so she can be smelling fine and all clean to meet little itchy brother.Then will be off with Rob for the breakfast celebration of the Grand Day of the Puppy --- waffles, for sure. And off to Chino to meet our 3rd dog and bring him home. We may take turns driving home so that Rob and Glen can hold him, and not just me the WHOLE way home. I\'m sure Glen will explain to him about the Tour de France on the way home. Maybe we\'ll really name him Musette. Is that too girly?  I hope I\'ll be able to sleep tonight. Can we do it? Can we raise a social, well-behaved lab? Upon relfection, I\'d have to say we kinda spoiled our M.C. (Magnificent Creature, a.k.a. Samson #1). I\'ve been studying how to housebreak a pup --- quite a diverse topic come to find out.   I just can\'t wait to see which one we will bring home. And how little will he be at 6.5 weeks? We couldn\'t even pick out a collar yet because we aren\'t sure what size his neck is. We only bought him three toys. A rope (pup-size), and \"puppy starter kit\" of Nylabones, and a miniature tennis ball. We don\'t want to over indulge him. :-) The book said...one chew toy and one play toy. We must say we were slightly manipulated by the marketers...if the packaging says \"puppy\", maybe we need one. But we used great restraint.  We will be all consumed with cute, so please be patient with us. We are warmed up to that condition thanks to the arrival of Miss Brianna. She has taught us how to appreciate cute and find more love in our hearts.  I\'m lunching with Doris and Audrey today. Audrey\'s on Spring Break, so out we go for some girl talk and crepes.  Great to see the photos of New Orleans! Good for Steve & Beth for taking time to get-away and play as a couple. High priority. I have favorite parts...Preservation Hall Jazz Band, coffee and beignets at Cafe du Monde, and the food! And definitely people watching. Glad you had a great time.  And I\'ll be watching the OLN on Thursdays now. I\'ll check for a musette at Crate & Barrel (Steve, that made me smile). I think it is really a sweet young woman, but Glen must not know that. OK, probably not.  '),(19943,3208138,'15,April,2004','2004-04-15 00:00:00','   Hey I had no idea that  urlLink Phil and the postie boys are on Thursdays . Now that \"The Donald\'s\" show is over, maybe Cindi will let me check that out.  '),(19944,3208138,'15,April,2004','2004-04-15 00:00:00','   We are watching OLN\'s Road to the Tour every Thurs. night. Phil Ligget is in fine form and I wish I was in Georgia next week to see the \"Tour of Georgia\". Lance will be riding in that race. You noticed I said \"Lance\" like Beth says \"Bernie\". We are so familiar here in Placerville. We don\'t own a musette, can you buy one at Crate and Barrel, and what does it look like?  '),(19945,3208138,'15,April,2004','2004-04-15 00:00:00','   Only 79 days until the  urlLink Tour de France  begins from Liege, Belgium on July 3, 2004. Will you stay at home, order cable TV configured to receive OLN? Or will you head to France on an organized tour taking you daily to the various race venues with your lunch packed in your musette?  Cindi and I have decided to stay at home this year. There will be a new puppy who would be just too lonely if we were gone that long. So, warm up those golden vocal cords, Phil Ligget. We will be listening live in the morning and remixed for prime time. Who knows, maybe Uncle Glen and Aunt Cindi will invite people over to watch--especially if they bring food, gifts or even just go cycling with Uncle Glen once in a while.    urlLink Lance Armstrong\'s web site  has more information that you could possibly know what to do with.  '),(19946,3208138,'09,April,2004','2004-04-09 00:00:00','   So the  urlLink International Policy Governance Association  is putting on a conference--on what else but Policy Governance. The conference is titled  urlLink \"The Future of Boards: Realizing the Promise of Owner-Accountability\"  and will be held in Chicago, June 10-12, 2004. I am really looking forward to this learning experience from some of the experienced practitioners and consultants in the field today. I have been to several recent seminars on this topic and the approach is to talk about \"best practices\" in response to the obviously broken parts of the board governance system. They usually tell some stories about a board of director\'s dysfunctionality, make the audience laugh because everyone has faced some similar dysfunctionality and they the propose some fix for that particular problem.   In the Policy Governance model, the whole system is addressed and a new paradigm is built around some real or perceived value added that the board of directors brings to an organization. This is described as linkage between the ownership and the organization. The values brought by the board then are reduced to an explicit, written form of policies in four categories: ends, limitations or means, board-staff partnership, and board process.   Because I have had the opportunity to serve on several boards and work for one over the years I understand the need for addressing these issues with great clarity. I hope that my experience at this conference proves valuable for my work with MEDA providing technical assistance to community-based and faith-based organizations and for use with other organizations in the future. I will let you know how it progresses.  '),(19947,3208138,'09,April,2004','2004-04-09 00:00:00','   In the past, we would have some down time on the blog server and we just had to wait for it to start working again. I haven\'t noticed that for a long time now. I know that Blogger is a lot more capitalized now that Google bought them. So I would expect less trouble.  Let\'s see if it continues to work ok.  '),(19948,3208138,'09,April,2004','2004-04-09 00:00:00','   I am posting this because when I tried to go to the page with pwwcrfamily web name I get an error message.  '),(19949,3208138,'09,April,2004','2004-04-09 00:00:00','   Beth and I had a wonderful time in New Orleans. We listened to live music every night, walked down every street in the French Quarter at least twice and ate well!!!!! Even the rain and lightening was an event. What a place to people watch. What a delightful place to spend a couple of days.  '),(19950,3208138,'08,April,2004','2004-04-08 00:00:00','   Beth and I just returned from New Orleans. We found a house for sale, looks pretty inviting. Good Colleges nearby.     A nice little home in the garden section of New Orleans. $840,000.00      A great place to people watch until it rains.  '),(19951,3208138,'06,April,2004','2004-04-06 00:00:00','   this morning: there is talk that Mr. Parks may be running for Mayor of Los Angeles.  '),(19952,3208138,'05,April,2004','2004-04-05 00:00:00','   The kind folks at JobStarts put on an event Saturday at the home of one of the board members. It was nice to have mom, Beth and Rob show up. I think Beth had fun chatting it up with Bernard Parks, Eighth District Los Angeles City Council.  '),(19953,3208138,'04,April,2004','2004-04-04 00:00:00','   Today was their first day outside in the sunshine! Two weeks to go until we meet them and pick one to be Sam 2.  '),(19954,3208138,'02,April,2004','2004-04-02 00:00:00','   I think waffles would be a great idea!  '),(19955,3208138,'02,April,2004','2004-04-02 00:00:00','   Here as a picture I told last weekend. A beautiful spring day.  '),(19956,3208138,'02,April,2004','2004-04-02 00:00:00','   I have always like Calvin. If he could only join the Kings and stir them up.  Isn\'t there supposed to be \"chevy\" or \"ford\" under him. Oh, by the way Rob, we are expecting you to answer all the questions Kyle has about college. No pressure there but if he struggle\'s it will not be our fault. Besides there is not forgiveness for turning my son into a Dodger\'s hat wearing guy. E-mail any pictures you want on this site. I will post them for you. There has a be a way to post music so if you have a sound clip in the form of a wave file I would like to hear the song you mentioned. My funeral song is \"Time to Say Goodbye\" sung by Sarah Brightman. With all the trades I don\'t have a clue about who is on the Giants, except for that player who never used an illegal performance enhancing substance. The way the Kings are playing I will probably have more time to watch the RiverCats. And there is no giving up on anyone. ever. never. nada. though in those times I gave up on myself I was always greatful that others believed in me more than I believed in myself.  '),(19957,3208138,'02,April,2004','2004-04-02 00:00:00','   I saw this little guy on another blog and stole him. He is now dancing in our template for the time being. Any comments?  '),(19958,3208138,'30,May,2004','2004-05-30 00:00:00','   Bono received an honorary Doctor of Laws from UPenn. I like what he says about being a fan of America. I did not know that Ben Franklin went to Ireland for some months to study what a colonial relationship with Great Brintain looked like. There, he found poverty, mud and grass huts. He did not find anything good with the Great Britian-colony thing--so the rest is history. You can click on the title above to read the whole thing.  '),(19959,3208138,'29,May,2004','2004-05-29 00:00:00','   It was a good lunch at Glen and Cindi\'s for Glen, Cindi, David, Steve, and Kyle\'s b-day party. We called Steve on the phone to  wish him and Kyle a happy birthday.  I saw some of the Giro d\'Italia today.  Cunego, Gilberto Simoni\'s young teammate won it.      (Cunego is the one in pink.)  Glen says I am responsible to get him and Steve caught up with the cycling world, but I get all my news from velonews.com (below), and my  Cycle Sport  magazines.   urlLink For the most up-to-date news, click this line   For his Tour de France preparation, Lance is doing well, and Jan Ullrich (his main challenger) is behind schedule with too much winter fat still. Perhaps this could really help Lance win. Beloki, another challenger for Lance, is not recovering from his nasty fall last year which broke 3 bones.  '),(19960,3208138,'29,May,2004','2004-05-29 00:00:00','   Michael, my sister Cindy, Jamie visit.   Beth and Steve with Michael.  Michael is scheduled to leave for the Iraq in August. We had a nice BBQ with him and my parents. Beth got to talk to her brother, David and Rob. Made her day.  '),(19961,3208138,'29,May,2004','2004-05-29 00:00:00','   David Weaver arrived after 10 PM on Friday. He is now sitting here with Rob and I watching the puppy do his puppy thing. When asked for his best thoughts for inclusion in this post he says: \"huh.\"  '),(19962,3208138,'28,May,2004','2004-05-28 00:00:00','   Beth\'s fifth grade class spent the evening. Only one was pushed into the pool.  '),(19963,3208138,'26,May,2004','2004-05-26 00:00:00','   It is a loooooooooong direct flight from Philadelphia to Los Angeles. I was supposed to be on a connecting flight that went through Washington, DC, but that flight was delayed bcause of weather or crew delay so I was put on fight 99, direct to LA. That was better, but still long. (Did I mention that it was a long 5 hours plus flight?)  '),(19964,3208138,'26,May,2004','2004-05-26 00:00:00','   Thanks to Cindi for the faxed birthday greeting that said: \" Happy 25th Birthday! (May 25th--that is).\" All of my colleagues have enjoyed that as I have shared it with them.  Thanks to Beth and Steve who sent an electronic brithday card. I tried to use their system to reply, but I lost my concentration somewhere along the way and never finished it--so thank you here on the blog.   Jane Remington gave me a call on my cell phone---what a nice surprise!  Our group of MEDA staff and affiliated programs, about 20 of us, had a dinner together on Monday evening and they made me stand up in the middle of the restaurant and sang to me.   So, I think the frenzy of birthday celebration has slowed down a little bit.   Note to Daniel: I have tried to respond to your email, but for some reason it did not work yet. I would be happy to show you how to make hypertext linkages using HTML. Maybe we could look at it together on Sunday if you guys are coming over.  '),(19965,3208138,'25,May,2004','2004-05-25 00:00:00','   it won\'t be as humid here today as it was yesterday, but still warm. Yesterday was VERY humid and reached to 90! I am seeing may old friends here at the AEO conference and the MEDA related ASSETS program folks had dinner together last night. They made me standup and sang happy birthday. More later, or not. I need to make sure and get a lot of work done while I am here in addition to all the networking and activities of the conference.  '),(19966,3208138,'23,May,2004','2004-05-23 00:00:00','   Sat. we are having a BBQ for Michael (Steve\'s nephew). He is scheduled to leave for the Middle East in the next four weeks. Sun. we are attending a BBQ for Shannon Barger\'s son who is on leave from the Marine Corp. and will be here with his wife and stepson. Then Kyle returns on Sunday night. Beth is co-prinicpal of summer school and will be back to work on Tuesday. Flexibility is hard to come by these days. With planning we can be available for that phone call.  Beth, Kyle and I are looking forward to meeting Sam 2, visiting more colleges, and a Dodger\'s game. With schedules in hand we will plan a trip down to So. Cal.  Here is the best to the party goers. Happy Birthday to all.  '),(19967,3208138,'22,May,2004','2004-05-22 00:00:00','   Everyone in the family is invited this weekend to our Summer Kick-off & Birthday BBQ. David Weaver is rumored to be coming this weekend, too. We will be celebrating the summer birthdays...Glen, Steve, Kyle, & Cindi...our May & June birthdays. We will celebrate with or without you, but it\'s richer if you can all stop in. I just wish I could tell you a time and day, but you know how this goes sometimes. Depending on when David gets here and when he leaves...it may be Saturday night or Sunday night or even Monday afternoon (Memorial Day). As Gumby says, \"Stay flexible.\" Watch the blog for details. Glen and I will discuss a menu and cake priorities. Steve & Kyle - you can put in your vote on dessert preferences. Hope you can come! If not, maybe we can call you during the event, just to connect.  '),(19968,3208138,'21,May,2004','2004-05-21 00:00:00','   By the way, Sam-2 has been with us for a little over a month. We got him at 6.5 weeks and 7 pounds. He is now 11.5 weeks and 19 pounds. And much bolder and bounces higher. A delight in my life.  '),(19969,3208138,'21,May,2004','2004-05-21 00:00:00','   Wow! Sam #2 does look big in that photo. He is so fun and cute. He has a 10 minute lesson practicing \"come\" and \"sit\" every morning about 6:30. Not sure what I\'m doing, but he seems to be catching on. Next week, we\'ll work more on leash walking and progress into \"heel.\" As opposed to, \"heal.\" He, as of last week, wags his tail and runs to us. So I know he likes us now. I wasn\'t sure before. Maybe he even admires us, too. Isn\'t that why people get dogs? To project what positive feelings the dogs are having?  School ends 6/3. Teachers check-out 6/10. It\'s uphill right now. Lots of electronic reports to file and hours of proofreading for errors. I\'ll meet with all 31 students in the next eight days for the final Learning Record.   Glen grilled that Copper River Salmon yesterday. I\'m not a salmon fan, but this was better than most. With salt, pepper, and lemon, I almost enjoyed it. I\'d still prefer lasagna or a hamburger or beef wellington --- now we are talking! I so want to share Glen\'s passion for the pink fish, but alas, I cannot. If it\'s an acquired taste, I\'ll keep trying. Give me blackened mahi-mahi or just a tunafish salad sandwich like my Mom makes --- with chopped hard-boiled egg added (Glen taught me that). Must be lunchtime.  '),(19970,3208138,'21,May,2004','2004-05-21 00:00:00','   Sam 2 is looking great. Sounds from your posts that you\'ve taken him into your hearts.  '),(19971,3208138,'21,May,2004','2004-05-21 00:00:00','   Samson 2 with his favorite domestique.  '),(19972,3208138,'20,May,2004','2004-05-20 00:00:00','   . . . about bicycling competition on OLN.   Samson gladly wakes me up during the 5 oâ€™clock hour so that I can get my coffee and oatmeal out of the way in time to watch the coverage of the Giro dâ€™Italia with Paul Sherwen and  urlLink Phil Liggett  commentating. (BTW, Bob Roll and Paul Sherwen do a nice job with the Tour de Romandie in the evenings. What does Phil Liggett do on his free evenings?)  I am about to write a note to NPR to suggest that Phil Liggett do a regular sports essay on bicyclingâ€”at least through the summer months. I think he could make the whole thing â€œvery NPR.â€ And, it would rank up there with the old conversations that Bob Edwards used to have with Red Barber.      I really think that there would be more universal interest in the sport of bicycling than there is in horse racing, with all due respect to the horse named Smarty Jones. But, I donâ€™t know anyone who knows anything about the Kentucky Derby or the Preakness. And NPR covers those. Are horseraces an east coast thing? Or, just something for old gamblers? Maybe I just donâ€™t know the right people.  '),(19973,3208138,'20,May,2004','2004-05-20 00:00:00','   \"Doug Christie covered his bald head in his hands and fell to the floor in frustration. The image told it all\" according to Canadian Press (AP) writer, Janie McCauley. Our hearts go out to the sports fans of Placerville who were cheering their guts out for a chance to be defeated by the Los Angeles Lakers. But it was not to be.  I am sure that our Ploog friends of the land of 10,000 lakes will be tickled pink that their Timberwolves have now progressed to this glorious point in history.  '),(19974,3208138,'20,May,2004','2004-05-20 00:00:00','   I will travel to Philadelphia, PA for the  urlLink Association of Enterprise Opportunity (AEO)  annual conference and membership meeting. This year the event is in Philadelphia.    urlLink      In the recent past this conference has been in Wahsington, DC; Denver, CO; Oakland, CA; Ft. Lauderdale, FL; and Chicago, IL. I have gone to this event for the last several years and enjoy seeing friends from the microenterprise industrly. My friend Andres Martinez will be there--Cindi and I stayed at his flat in Mexico City a couple of years ago when our summer vacation took us through that city. He has moved since then.   Andres also stayed at our place in Whittier when he came through Los Angeles visiting the ASSETS related microenterprise programs that MEDA promotes. At JobStarts we started the BusinesStarts program on this model.  '),(19975,3208138,'19,May,2004','2004-05-19 00:00:00','   According to the sign at Ralph\'s, the Wild  urlLink Copper River Salmon  starts tomorrow. Last year I missed it all together our of just plain ignorance. My plan is to go down early to the store and buy some for lunch. I remember this salmon to have excellent flavor and texture. It is not farm raised so it has more of those omega fatty acids that are good for you.  '),(19976,3208138,'19,May,2004','2004-05-19 00:00:00','   I have decided that I no longer am interested in blog posts about blogging. I know that I invited such posts a while back. I know I have no way of enforcing the ban. Nevertheless, I hope that people will now be posting about the fascinating things in their fascinating lives.  '),(19977,3208138,'19,May,2004','2004-05-19 00:00:00','   I am having trouble with knowing when new comments are posted. I find myself reviewing every post that has a comment section and trying to remember if the number of comments has changed or not. Often I look at the comments because I can not remember how many comments were there before. I now realize that I have missed comments on some posts and reread comments from other posts. I find this feature very time consuming to use. I realize I have the option to not use it and just continue to post all my input as posts rather than comments.  '),(19978,3208138,'18,May,2004','2004-05-18 00:00:00','   I would like to know who taught a lttle yellow labrador retreiver puppy that the area by the stove in the kitchen is a cool place to deposit a yellow puddle. Hmmn.  '),(19979,3208138,'17,May,2004','2004-05-17 00:00:00','   I agree with Rob that the Bloog site seems less user friendly but maybe I\'ll get used to it soon.  I was finally able to print up a receipt for the trip.  (This not on the bloog ) They kept dumping me out saying I could try later.  I kept putting my pass word in and my secret code and also put a new one in, then when I would click on that they said I was already a member.  I finally admitted that I forgot my secret code and they sent it to me via e-mail and I realized how it got there.  Thanks Steve.  I\'ve been collecting names of places that we just must see from different groups of friends.  The Willows has come up from every group.  It\'s a restaurant and has been there since before the 50\'s.  Will sent things on as they become clear or any of you can send suggestions of places to visit.  The most important place is the Pearl Harbor Memorial then the Polynesian Cultural Center.  '),(19980,3208138,'17,May,2004','2004-05-17 00:00:00','   The company that is Cindi\'s  urlLink Creative Memories  business is running a  urlLink special promotion with Disney . It is a contest that will send some lucky customer and the consultant who sells the wining album to Disney World.   Cindi bought 15 of these special albums. 11 are already sold. So hurry in and buy your album to see if you win! The contest ends June 30. I think she will buy more if she needs to-- hurry!  '),(19981,3208138,'17,May,2004','2004-05-17 00:00:00','   It seems like whenever I turn on the  urlLink OLN , hoping for a tidbit of news about cycling anywhere in the world, all I find are bullriders  and fishing competition.   I have read that Lance has won the  urlLink Tour de Georgia . I believe he will be competing in some other races before  urlLink TdF . I am wondering what else is happening. Can anyone help us out a little bit here? What is the state of the  urlLink pro cycling  world? Post here to let us know.   (If anyone wants to, I can teach you how to make a hyperlink like the one above--that means you can click on it and it takes you to more information at another site.)  '),(19982,3208138,'17,May,2004','2004-05-17 00:00:00','   A Google search tool is now located in the right side column for your convenience. Many people use Google to search the web and now you can do it from this site.   This feature was requested by Beth of Placerville.   BTW--when you enter your own name into Google to search, it is called ego surfing. Have fun!  '),(19983,3208138,'16,May,2004','2004-05-16 00:00:00','   Rob, it didn\'t take another miricle shot for the Laker\'s to move on. They won with confidence.  '),(19984,3208138,'15,May,2004','2004-05-15 00:00:00','   this new format for the blog seems to be less user-friendly for the \"casual blogger.\" i don\'t like how you have to select a topic to read, instead of just reading it as a sort of running list of the blog entries. the lakers are going for the series tonight against the san antonio spurs. the shot derek fisher made the other night to win the game was like none that will ever be seen again. there was .4 seconds on the clock, where .3 is considered not enough time to catch the ball and shoot. sports fanatics get a little out of hand with this stuff. i saw a guy who painted \"(explitive) spurs\" on the back window of his truck. i wanna go watch the game at his house! sam #2 is increasing in size at an incredible rate. i just paid for my summer school CLASS, not CLASSES, and the cost was 96 dollars. that doesn\'t include the cost of the book, which thankfully, i\'d already purchased for the bargain price of $86.50. i\'m learning a lesson in math before i even take this stupid course. a collumnist who contributes to the los angeles times shared that it cost him 57 dollars for a full load to attend uc san jose for one semester. i\'m considering going back to polly\'s, or getting a different second job. does anyone know of any openings? i have experience in dealing with customers, food service, and i\'m a pretty good ticket-taker guy. don\'t worry about dangling participles glen... \"From now on, ending a sentence with a preposition is something up with which I will not put.\" -Sir Winston Churchill  '),(19985,3208138,'13,May,2004','2004-05-13 00:00:00','   I am impressed by the knowledge of some members of this family group regarding specific teams or sports. I am sure that the more knowledge you have about a sport the more you can appreciate participating or observing. I am more of a dabbler. Here to the more knowledgeble among us.  '),(19986,3208138,'11,May,2004','2004-05-11 00:00:00','   I had wanted to make a comment about subjective ness of sports in a post by Steve a while back.  However, time has slipped by and I am just now posting. As the readers of this blog are likely aware, I have not been much of a fan of watching sports. My thought on this is that if the subjective, emotional part was not a big deal, there would not be many fans. If the objective discipline and rules that set an immovable infrastructure against which to play, there would be nothing to be fanatic about [sorry this sentence ended in a preposition but it is too late to change this dangling participle].  Lakers are leading San Antonio by 10 in the fourth quarter as I type this. This may be the first time they have led in this game--an important one in a series that determines progress in playoffs. I am the only person left in the room watching this game. I think the true fans have been so affected by the pressure of the game and whether their preferred team will win that they have been driven from the room. My friend Lalo told me tonight that he has a hard time watching the Lakers--especially this season because of this whole passion issue.   Previously I had mentioned that I would comment further on the animated feature length French film Triplets of Bellville. So here it is:  1. It is extremely odd--exemplified by unusual lines and proportions on faces, buildings, streets, contrasting tall and short people, etc. and a topless dancer in some sort of cabaret and on and on it goes. 2. There is no real dialogue, the story is told entirely with visuals. 3. Interesting sonically--recurring sound devices like the dogs toenails against tile, the sound of a passing train, a whistle blowing, music, all seemed to serve to draw attention to a certain scene for a purpose. Much like the use of lighting a sound in a live theatrical production.  4. Visually interesting--combination of hand sketched 2 dimensional animations and computer generated 3 dimensional animations contrasted in the same scenes to make varying impressions on the viewer.   It may have been easier to use the comments function for part of this post, but it doesn\'t work on older posts.  '),(19987,3208138,'10,May,2004','2004-05-10 00:00:00','   Thankfully, Cindi is home documenting most every motion of two dogs with her camera. Anyone want to make a bet on how many CM albums will result from this new puppy?  '),(19988,3208138,'10,May,2004','2004-05-10 00:00:00','   Well, this is it!  We take Cynthia to the airport for a direct flight back to sunny California and the group.  Sad to see her go, but had a perfect time!  Welcome to anyone else who comes this way. Looking forward to this great Blogging idea of Glen\'s.  Lots of fun!!! Love ALL the pictures, and Beth you have me in stitches.  I look forward to your postings.  '),(19989,3208138,'10,May,2004','2004-05-10 00:00:00','   We started our Mother\'s day celebration on Saturday, Mother\'s Day Eve, with a brunch at the California Grill in Whitter.   Then, Sunday, we barbequed at the Peterson\'s Milton Mansion. Doris brought macaroni salad, Audrey brought a green salad with \"lime\" dressing that was a lot like guacamole.   Included in our celebrations: Doris, Audrey, Garth, Timothy, Daniel, Rob, and me. Hope the rest of you can join us next time.  '),(19990,3208138,'09,May,2004','2004-05-09 00:00:00','   Tonight Dick, Cynthia and I went to a Dog Training Class. There were between nine and eleven dogs with their owners learning what is expected in \"Competitions\".  Our eyes were opened to new fields of adventure and excellence.  There are dogs, and then there are dogs !!!  Our Sam 2 will certainly be a star!!  We are all excited again !  '),(19991,3208138,'09,May,2004','2004-05-09 00:00:00','   I was so shocked to sign into this and see a whole new program.  It doesn\'t take much to get me confused on the computer but after remaining calm and reading through the options I was able to fing out how to read the entries and also to make an entry.  I do like the temperature  '),(19992,3208138,'09,May,2004','2004-05-09 00:00:00','   Blogger has launched a lot of new features and a new look to the user interface section.  I have also, in a completely seperate act, inspired by the other changes, made some changes to the look of our site. I am not satisfied with several things yet; so I will keep trying.  Once I get a look that I like better, I will add those weather reporting links back in.   Your comments are weicome.   UPDATE: I have tried a couple of template changes that get the stuff I wanted to work--I kept getting various errors and formatting problems. Steve--I took you post of the steamboat off by making it a \"draft.\" The problem is that the picture is wide and was pushing everything on the page off of the side. If you can shrink the photo, upload it with the same name, and republish that post, everything should work. Let me know what you think.  '),(19993,3208138,'08,May,2004','2004-05-08 00:00:00','   Dick \'n Jane & Cynthia drove about 60 miles in the gorgeous weather (which we always have here in Ohio) to the lovely \"INN AT HONEY RUN\" near Millersburg, OH in the heart of Amish country. We had delightful brunch by the windows overlooking the woods. I had bananas foster with carmelized croutons and vanilla bean ice cream.  It was very delicious!  I had never had it before. I know it would have tasted even better in a \'ramekin\'.  Dick had a piece of most elegant chocolate cake with a strawberry that measured at least a \'cubit\'. Cynthia\'s choice was a fabulous sour-cream pound cake with berries and Chantilly-cream. On the way home we stopped at one of Cynthia\'s favorite places in Ohio - The Pine Tree Barn where she found some very unusual sun-catchers.  Hope the day was as pleasant for all of you. [written by Jane]  '),(19994,3208138,'08,May,2004','2004-05-08 00:00:00','   To all of the moms have a happy mother\'s day. According to the weather information you will all have a bright and warm day to enjoy it. Best wishs to all. Beth and I had lunch with my mother yesterday. I got caught up on all of the \"family\" news. My father enjoyed our company and was glad we took time \"from our busy schedules\". I think anyone younger than 86 would look busy to him. It was a plesant afternoon.  '),(19995,3208138,'08,May,2004','2004-05-08 00:00:00','   Audrey and Doris are meeting Steve and Beth in Hawaii. We will all have to learn the Hula. We will also have some interesting words to add to Glen\'s list.  '),(19996,3208138,'06,May,2004','2004-05-06 00:00:00','   . . . or should we just grunt and point? Here is a word that came up at lunch today:  ramÂ·eÂ·kin or ramÂ·eÂ·quin n 1.	a small ovenproof dish with vertical fluted sides designed to hold a single serving of a prepared food, especially one that is baked 2.	a portion of food cooked and served in a ramekin  Rob said he was curious about the etymology of the word ramekin, I think he was right about the Dutch: [French ramequin, perhaps from Dutch dialectal rammeken, toasted bread, or from Low German ramken, diminutive of ram, cream (from Middle Low German).]  Maybe you were wondering what a ramekin looks like?     I recently started a list of new words for myself. When I hear or read a word with which I am not familiar, I write it down, look it up and add the definition to the list.   After watching a movie that Rob loaned me recently, I added this to my list:   rumpus (n) an outcry or noisy disturbance   --and this--  obsequious (adj) Excessively eager to please or to obey all instructions  I might of added this one after going to the theater:   scrim n 1.	a drop curtain in a theater that appears opaque to the audience when lit from the front but transparent when lit from behind 2.	a durable open-weave cotton or linen fabric used for curtains, clothing, upholstery lining, and in industry  '),(19997,3208138,'06,May,2004','2004-05-06 00:00:00','   Cynthia is sitting beside me in Akron, OH - so we are breaking our silence and joining in the \"Family Blog\".  Sam 2 did not come.  We all miss him, but Glen and Rob are taking excellent care of him, so not to worry!  Picked Cynthia up at Cleve-Hopkins airport Friday  10:58 PM.  Lovely day here, have already been to the new STARBUCKS at the corner of our street - How sweet it is !!!!  '),(19998,3208138,'06,May,2004','2004-05-06 00:00:00','   Cindi went to Akron, Ohio this weekend for Mother\'s Day. That is because her mother is there.   Audrey and I will take our mother to lunch today, Saturday, to celebrate motherhood. We expect that Rob and Timothy will join us.   My friend Pedro, the owner of Uptown Mexican Cafe, told me this week that he expects to be busy from Friday dinner through Monday dinner with Mother\'s day revelers. Buddy at California Grill expects that they will be busy for Saturday breakfast and dinner and Sunday for lunch.   [Little Sam is snoring away on the floor behind the rocking chair right now--it made me laugh, and reminded me of Cindi.]  '),(19999,3208138,'06,May,2004','2004-05-06 00:00:00','   Something I started when I was working for JobStarts is a  urlLink one-day seminar with John Carver on Policy Governance  that will take place on June 5 Carver at the BP Amoco building in Los Angeles. It is now possible to register by credit card for the seminar through the Los Angeles Jr. Chamber of Commerce\'s website. Thas is really cool.   I am still acting as a consultant on this project. We have room for 90+ people. There are currently 19 paying customers signed up, all of whom I have personally signed up. We are a long way from breaking even at this point. I am nervous.  '),(20000,3208138,'06,May,2004','2004-05-06 00:00:00','   After a steady diet of sports news the last couple of months I want to change sports to a purely subjective endeavor. Teams would show up, demonstrate their skills for a set number of minutes, and then there would be no score or other objective criteria remembered. Spectators would only remember the event for is subjective \"emotional\" value. Sports would then be delegated to just another of those \"emotional\" moments that do not require any logical or critical thinking. Wait, sports writers and many fans only think emotionally about sports, they forget the scores or use the data from events for their own emotional reasons. Not much evidence of logic or critical thinking now. Forget I said anything.  '),(20001,3208138,'05,May,2004','2004-05-05 00:00:00','   Well Steve, I tried. I am sure that if I read more instructions, I\'d find out why the HTML is working in the right column field and not in the post message field.   Which one do you like?  The second \"sticker\" is identical to the one posted in the right.  '),(20002,3208138,'02,May,2004','2004-05-02 00:00:00','   urlLink   alt=\"Click for Honolulu, Hawaii Forecast\" height=108 width=144>  '),(20003,3208138,'02,May,2004','2004-05-02 00:00:00','   I am impressed Glen. Value added. All the weather sites show up.  '),(20004,3208138,'01,May,2004','2004-05-01 00:00:00','   It was really hot again today in Whittier. I think that the high was 110 degrees or more! Right now in Whittier this is the time and temperature:   urlLink      UPDATE: So as a value added, you can easily check out the weather for several of the places represented by those who read and post to the PWWCR Family Blog. Did I miss anyone?  UPDATE [an hour later]: Ok, after about 3 tries with varying degrees of sucess, failure, sucess, failure, (repeat ad naseum). I think this weather listing is going to work. I was putting the HTML in the wrong place and blogger was producing errors. I have now moved it all to where I think they wanted me to put it in the first place. It seems to have responded correctly now for a few repetitions. We\'ll see how it goes. If it works--there should be several boxes with weather for various places on the right.  '),(20005,3208138,'01,May,2004','2004-05-01 00:00:00','   Especially after an afternoon in the hands of the lovely people at United Airlines. I will have to watch the last part of the film to see how it ends. The film is unusual and interesting both visually and sonically. After I watch the rest of it I will try to post again with more.  '),(20006,3208138,'30,June,2004','2004-06-30 00:00:00','   A couple of weeks ago, when Dave was here, we had breakfast outdoors. In fact, I took the skillet out there and made the pancakes out there too.  Then, I captured this image.  '),(20007,3208138,'29,June,2004','2004-06-29 00:00:00','   When he finds something valuable, he will carry it around the yard, chew on it, and eventually put someplace safe, like the family room near the TV or his pillow.  '),(20008,3208138,'29,June,2004','2004-06-29 00:00:00','   If I recall correctly, and if it will be the same as last year, I think that the local Pacific Daylight Time start of the coverage will be at 6:30 AM with the Pre-Race Show and then actual race coverage starting at 6:30 AM--that\'s 9:30 AM Eastern US time. Then they repeat stuff during the morning, afternoon and evening. Thats got to be better than most professional sports! I will be up at the earliest show time. (At least on Saturday, we\'ll see how it goes for the rest of the race.)     Be sure to notice the time zone listed.  '),(20009,3208138,'27,June,2004','2004-06-27 00:00:00','   I am sure you are wondering what the weather will be like at the start of the race. I was looking for weather for Liege, Belgium. Weather Underground returned Bierset. I don\'t know how close that is to Liege.    urlLink       I haven\'t given up yet. But, I did find the weather station for the city of Charleroi, Belgium. That is the ending point of the July 4 day of racing.   urlLink  '),(20010,3208138,'27,June,2004','2004-06-27 00:00:00','     '),(20011,3208138,'27,June,2004','2004-06-27 00:00:00','   My work was mentioned in a MEDA Publication called  urlLink The Marketplace , a glossy magazine that goes to MEDA\'s membership. You can read the article by cliking on the links above.  '),(20012,3208138,'27,June,2004','2004-06-27 00:00:00','   Peets can be found at  urlLink www.peets.com  '),(20013,3208138,'26,June,2004','2004-06-26 00:00:00','   Who\'s next? Lance Armstrong\'s rivals are decreasing all the time.  First Beloki, and now Vinokourov.  T-Mobile anounced who will be in the Tour de France on their team, and Alexandre Vinokourov isn\'t on the list.  He crashed on a traffic island in a race and suffered a concussion, facial scrapes, and ligament damage to his shoulder. The shoulder injury required surgery, which will keep him off the bike for 3 weeks â€” the 3 weeks leading up to the Tour (unless he\'s like Lance and doesn\'t do anything his doctors tell him to).  He would have been helping Jan Ullrich in the battle to win his second Tour de France, so Ullrich wouldn\'t be called the \'Eternal Second\'.  I haven\'t heard about the wonderclimber Roberto Heras in a while.  Maybe he wants to just train instead of race so this won\'t happen.  See y\'all next Saturday!    (Vinokourov)  '),(20014,3208138,'22,June,2004','2004-06-22 00:00:00','     '),(20015,3208138,'20,June,2004','2004-06-20 00:00:00','   Enjoyed the unexpected visit from Beth and Steve this last weekend and enjoyed meeting Brianna.  What a sweetheart.  Doesn\'t really know a stranger.  Also thanks to Glen, Cindi and Rob for the impromptu event at their home on Saturday eve. Hope Sam 2 has recovered from the invasion of his kennel.  Steve, I thought we would see a picture of Brianna in the kennel.  Our lips are sealed as far as the child protective agency.  Doris  '),(20016,3208138,'20,June,2004','2004-06-20 00:00:00','   It was nice to see everyone. Here is a picture of Beth and Brianna at the beach. Brianna arrived safely and Jonah is out of the hospital. Thanks for the warm welcomes.  '),(20017,3208138,'20,June,2004','2004-06-20 00:00:00','   In case you don\'t think you will get enough of the Tour with OLN and a possibility of another book from Lance, you can pre-order a book that will tell all about what will happen at  urlLink velogear.com .  This might be what it will look like, because the riders on the cover are favorites--the German  urlLink Jan Ullrich , winner in \'97, and 5-time runner-up to the Tour, who won the Tour de Suisse (a one-week long race) by one second recently; the Texan  urlLink Lance Armstrong , 5-time winner and 3-time beater of Ullrich; and the \"Man From Marblehead, Massachussets\"  urlLink Tyler Hamilton , the one who got fouth place last year despite having a broken collerbone to deal with for almost the whole race, and who won the Tour de Romandie (in Switzerland) in May.   Perhaps this is what the podium will look like in Paris.  '),(20018,3208138,'18,June,2004','2004-06-18 00:00:00','   There are only 16 days left until the Tour de France, and the Spanish rider Joseba Beloki, Lance\'s main rival in 2002, and crashed out last year, announced he will not be racing in this year\'s Tour de France due to lack of fitness.  At the end of 2003, he signed to Brioches La Boulangere (a French team) for until the end of 2005, but then he decided to cut it short and end at the end of this year, then he decided to quit in June.  He is now training for the Vuelta a Espana (a major stage race in September).  Here is a picture of the coffee table book about Lance Armstrong\'s carreer that I mentioned in a previous post.   I never like to look at Lance struggling in a time trial like that and don\'t like to see him in that jersey design because I always think about that.  I hope you don\'t have that problem. Lance\'s team (United States Postal Service Presented by Berry Floor) will quit for 2005, but it will be replaced by the Discovery Chanel Team until the end of 2007 I guess.  At the end of this year, the team will have the Discovery Chanel logo on it prominately.  Read more at  urlLink Lance\'s web site .  Maybe this means we will have to watch lance and bike races on the Discovery Chanel instead of the current OLN.  Lance announced he will ride the 2005 Tour de France, but just to make the change go more smoothly. You can also read the entire first chapter of Lance\'s new book, \" Every Second   Counts \",  urlLink here .  There is a book called the  urlLink Tour de France Centennial Book  and is a coffee table book about the Tour de France with 360 pages of images and stories about the race in both black and white and color. It only costs $24.  A limited edition.  '),(20019,3208138,'17,June,2004','2004-06-17 00:00:00','     '),(20020,3208138,'16,June,2004','2004-06-16 00:00:00','   Here is a picture of Brianna eating a piece of Kyle\'s birthday cake. Don\'t tell me there isn\'t entertainment at our house.  '),(20021,3208138,'16,June,2004','2004-06-16 00:00:00','   We are back. Fell into bed last night around midnight. Flight went well, though a bit choppy at times. Weird thing...no window in our row. I was bummed. When I flew to Chi-town, as we Chicago residents say (pronounced shy-town), everyone pulled down the window shades so the movie screen could be better viewed in a darker plane. What is the point in flying when you can\'t see the view and the weather and the sky? Ugh. The movie eastbound was Starsky & Hutch. I opted to read and nap.  I\'m still doing piles of final reports. I check out of my day job tomorrow, and officially begin summer vacation. Audrey, Doris, and I will go to lunch at this cute spot in Los Alamitos called Shenandoah. Wish Beth and Mom-Jane could join us. I can\'t wait to be done with the school year, even though I do like my job. Next year will be year three!  Poor Lakers. But do I really care? I\'m planning the French toast family breakfast for the start of the Tour de France! Moving on to sporting events that really have some excitement for me.  FYI - the Remington computer is in the repair bay for a week. If you need to share news with the Ohio folks, do it the old fashioned way...by phone.  We had a nice time in Chicago. Hotel 71 is a contempo place --- kinda weird, but nice room and cool view. Maybe Glen will blog a picture. I got to meet with 10 good friends from the old days when I lived in Wheaton. So great to catch up. Sam looked HUGE to us upon our return. Wow! It\'s so nice to leave him in good hands with Rob when we travel.  Still so concerned for Jonah. Hope each day is a better one. And happy birthday to Kyle! How old now?  '),(20022,3208138,'14,June,2004','2004-06-14 00:00:00','   Jonah is surgery went well. No leaks into his abdomen. They resectioned his colon where it was ruptured. Beth, Brianna and I are off to see him now that he is off the ventilator. Beth is bringing Shawntea home while I stay with Jonah. That will give her a chance to spend time with Brianna and get a shower etc.  Thanks for your prayers and thoughts.  '),(20023,3208138,'13,June,2004','2004-06-13 00:00:00','   Jonah is off to surgery. The did a CT scan and he has a ruptured bowel. I will be off in a couple of minutes after singing happy birthday to Kyle.  '),(20024,3208138,'11,June,2004','2004-06-11 00:00:00','   Just a quick note. Jonah is having some problems with his abdomen. The medication he is on now requires that he is back in Intensive Care. I saw him late last night. Today he is hoping to go back to a regular room. Several of the firefighters are there to visit him when I called. The nurses are very good at letting people visit one at a time. Shawntea is going home for a while today and spend some time with Brianna. We will have again overnight. Here is a picture of the difficult life she leads at our house.  '),(20025,3208138,'11,June,2004','2004-06-11 00:00:00','   Beth, Kyle and I saw Jonah this afternoon. He is struggling with concentration and memory. Looking at the truck this am it is amazing to me that he survived. There is still part of the tree he hit embedded in the engine block. The tree is about three feet into the engine compartment. Quess where the engine is? No wonder his leg was shattered. He now has a metal rod from his knee to his ankle. The bone is strapped and screwed to together.  His head trainer arrived as did the chief of the largest fire department in the area. They were both supportive and made the point that firefighters stick help each other out. The chief of his department will be by tonight to talk to him as well. We will see him tomorrow. Shawntea stays at the hospital with him and we have Brianna again.  thanks for your prayers.  '),(20026,3208138,'10,June,2004','2004-06-10 00:00:00','   Jonah is in Sutter Roseville Hospital. He had surgery on his leg and will be in the hospital until atleast Sunday. Kyle and I will visit him this afternoon after getting his glasses from his truck. Not looking forward to seeing his truck, heard it is in much worse shape than he is. Thanks for air bags and seatbelts. Beth returned to work this am. She had Brianna overnight. We got to wake up to a little girl calling out for her grandma. Not like a puppy sniffing your nose, but nice.  '),(20027,3208138,'10,June,2004','2004-06-10 00:00:00','   Jonah was in an auto accident today. He was life-flighted to Sutter Hospital in Roseville. Kyle and I visited him tonight. He has a concussion and both bones in his right leg were fractured. He is supposed to have surgery to repair his leg sometime during the night. His girlfriend Shawntea is spending the night at the hospital. Kyle and I will try and find what is left of Jonah\'s truck tomorrow and also contact the fire academy. He will be in the hospital for about four days. Will post more as we know more.  '),(20028,3208138,'08,June,2004','2004-06-08 00:00:00','   Tonight in Chicagoland the lead news story is a tornado watch and flood warnings. One tornado has been sighted south of here. I am in the town of Boomingdale, Illinois. Just gentle rain out my window.  Regarding email. I have set up a new service on my email. In order to reduce the number of unsolicited emails (SPAM) that I receive, I am trying out a system that onlyl allows those who I proactively allow to send. I think I have allowed everyone who regularly read this blog. Thanks Steve for your patience, you are now allowed to send to my account. It is not uncommon that I have to hit the delete button 50 or more times in a day for emails that I do not want. Please let me know if I have inadvertantly not allowed your eagerly anticipated and appreciated emails.  '),(20029,3208138,'08,June,2004','2004-06-08 00:00:00','   Beth was to ill to go to work today. Sore throat, fever, and aches. Slept most of the day with Kyle keeping watch. Well when not focusing on his video game he did listen to guess whether she was asleep or not. She is hoping that she will feel better and be at work tomorrow.  '),(20030,3208138,'08,June,2004','2004-06-08 00:00:00','   I will be in Chicago through next Tuesday.   urlLink      The weather looks warm and humid tonight.   I will be at the  urlLink IPGA conference  and visiting my CCF clients.  '),(20031,3208138,'07,June,2004','2004-06-07 00:00:00','   Ask to see Beth\'s DFish tattoo next time you see her. I thought I got carried away by basketball. I am just glad the Lakers won tonight.  '),(20032,3208138,'07,June,2004','2004-06-07 00:00:00','   Today I visited Rudy Carrasco of the  urlLink Harambee Christian Family Center . The first time I heard of a blog was from Rudy and I have been reading  urlLink his blog  with some regularity ever since. Today, in his office, he captured an image of the two of us and posted it on his blog. He included a link to us, so If you are visiting from Rudy\'s blog, say hello with the comment feature here.  '),(20033,3208138,'07,June,2004','2004-06-07 00:00:00','   i too, have had to drop out of a lot of races this year due to lack of fitness. this year also holds a pretty important political race, one that ted kennedy has called the \"most important\" election of his lifetime. hopefully both candidates will get in better shape for that race. sad to hear about president reagan. jon stewart on \"the daily show\" wondered out loud if reagan\'s death was a conspiracy started by jennifer lopez to take focus off of her third (who cares?) marriage. peace to everyone.  '),(20034,3208138,'07,June,2004','2004-06-07 00:00:00','   Will Joseba Beloki, Lance\'s greatest rival in last year\'s Tour until he crashed out, not even start this year\'s Tour?  He has had problems of not being able to have as much fitness as he should for having the Tour start in just four weeks.  He has had to drop out of a lot of races this year because of his lack of fitness. Perhaps it will be Tyler Hamilton that will challenge Lance the most in this year\'s Tour.  He won the Tour de Romandie and is showing a lot of fitness.  You can get the  urlLink Official 2004 Tour de France Poster by clicking this line.  It\'s only 10 dollars.  '),(20035,3208138,'05,June,2004','2004-06-05 00:00:00','   In my last post I mentioned two books, but I forgot about another one that looks interesting.  It\'s called  urlLink The Noblest Invention , and has a forward by Lance Armstrong.  It\'s about how the bicycle chaned from its weird beginnings to what it is now.  There are over 250 pictures and essays about it.  Read more by clicking on the link above.  '),(20036,3208138,'05,June,2004','2004-06-05 00:00:00','   There is a new book out called   urlLink Lance Armstrong--Images of a Champion  . It\'s about Lance\'s life and has 239 photos of him by Graham Watson and was started several years ago. It will probably be outdated by Christmas because I think you can get it now, so this year\'s Tour de France won\'t be in it, but it looks like something I\'d put on my Christmas list.  There is also a new book out on the Tour de France by Bob Roll [You know, in the Tour de France days you a greeted by the commentators PhilLigget, the guy with one name and that TourDay France guy, Bob Roll (and Paul Sherwan)]. It\'s 192 pages at $11.  Here\'s what Bobchi\'s book looks like:    Happy shopping!  '),(20037,3208138,'04,June,2004','2004-06-04 00:00:00','   The word that a kid spelled to win the national spelling bee.  auÂ·tochÂ·thoÂ·nous adj 1.	used to describe a rock, mineral deposit, or geologic feature that was formed in the area where it is found 2.	descended from the original flora, fauna, or inhabitants of the region in which it is found 3.	produced or originating as a physical function or disorder in the place where it is found  EncartaÂ® World English Dictionary Â© 1999 Microsoft Corporation. All rights reserved. Developed for Microsoft by Bloomsbury Publishing Plc.  Click  urlLink here  for more news in the online edition of an Indianapolis news paper.  '),(20038,3208138,'03,June,2004','2004-06-03 00:00:00','   Any suggestions for the awards for the caption contest. I was more interested in the dinner that accompanies the presentations. Beth and I have stuff for a garage sale that might work. Easier to give away then sell. Still looking for more entries.  '),(20039,3208138,'03,June,2004','2004-06-03 00:00:00','   I am having fun with the captions. Liked the others as well.  '),(20040,3208138,'02,June,2004','2004-06-02 00:00:00','   I\'m wrapping up hours and hours of paper work for the next week or so. Students end their homeschool year tomorrow. I am swamped with work. Rumor has it Beth is back at it with summer school administration. I\'m looking forward to things winding down and reading a good book for fun again. Samson II went to the vet today for his next shot in the puppy shot series. He weighed in at 24.5 American pounds and was found to have dirty ears. I lost the lottery and so I will be putting drops in while he bites me for ten days...wrestle mania with a wagging tail.  '),(20041,3208138,'02,June,2004','2004-06-02 00:00:00','   I think this could be fun. Use the comment function to submit a caption for this photo.  '),(20042,3208138,'02,June,2004','2004-06-02 00:00:00','     '),(20043,3208138,'02,June,2004','2004-06-02 00:00:00','     '),(20044,3208138,'02,June,2004','2004-06-02 00:00:00','     '),(20045,3208138,'02,June,2004','2004-06-02 00:00:00','     '),(20046,3208138,'02,June,2004','2004-06-02 00:00:00','   Valueable prizes and lavish praise to the winners.  '),(20047,3208138,'02,June,2004','2004-06-02 00:00:00','   Beth and I had a nice dinner out. New Italian place in El Dorado Hills. Great food and lovely setting. Thanks for the pictures Rob. Wished we could have been there to enjoy the day with all of you. We are working on a schedule for this summer with Kyle. I am trying to set up a trip that includes college visits, a dodgers game and time with family. After looking at what already is scheduled, not sure when that can be.... Again thanks to Rob for the photos. I just picked a few, but let me know if you want others, I have saved them all to my website. Take care to all of you.  ');
INSERT INTO `msgs` VALUES (20048,3208138,'02,June,2004','2004-06-02 00:00:00','   i started summer school today. for six weeks, i will take four hours of math a day monday-thursday. it\'s going to be a challenging course. i feel kind of guilty that today i witnessed sam and ashley\'s gleeful play session in their new (and already punctured) kiddie pool, while glen and cindi were away, missing the priceless laugh riot. sadly, no photos were taken of the event, but i was able to capture several candid shots from the combined birthday celebration that took place on sunday. it should be viewable  urlLink here  . *note to blog administrator: i give permission for you to post any or all of these pictures on the actual blog site if desired. it may make for easier access.* the celebration was a lot of fun, although i had to cut out a bit early to go to work on time. sorry not all the birthday boys and (or) girls were not all able to attend, although they were acknowledged by their names on the cake.  '),(20049,3208138,'31,July,2004','2004-07-31 00:00:00','   I was lonely this week. Glen was in Miami. OLN no longer offered any company to me. Daniel only stopped by briefly once. The Democratic Convention held only mild interest. And Sam II did not learn his \"down\" command, even though we practiced every day over and over.  Thankfully, Rob was here, awaiting the start of school and helping me hunt down the mysterious, evasive bad smell in the refrigerator.  So, I worked a lot on getting photos in albums. I had a very good time. It is fun to get photos organized to retell the great stories, big and little of the year. Very fun for me. I\'ve been leaving off most of the design to speed things up. The photos and words/writing tell the story, so nothing else is really needed. And it\'s been a good month for my Creative Memories business.  I will be traveling next week to Minneapolis, Minnesota to attend the annual national convention for Creative Memories. It is always a top-notch experience. This year, I will be taking the \"plant tour\" --- traveling to St. Cloud, MN to tour the headquarters. Should be interesting. I\'ll be home on Sunday morning with many of the new products introduced in hand. That fact is very exciting to some of my eager clients.  Does anyone want to go to the next local Sheryl Crow concert with me? Not just to see if Lance rides through. I do like her as a musician, too.  I got bicycle shoes and gloves. Now I hope to get the training wheels off my bike by the end of the summer. I was really inspired by LA.  '),(20050,3208138,'30,July,2004','2004-07-30 00:00:00','   Glen posted something about me being in the race at Heritage Park, and that I beat the kids my age.  One of them is the State Champion.  We saw them riding on the trail yesterday, and apparently they turned around not long after to chase us down.  We were going pretty fast, like 24 mph for a while (with a tail wind), then Dad got a flat and they passed us.  We could have stayed with them if that didn\'t happen, and we could have stayed ahead of them if we knew we were racing (being chased).  He\'s going to do the US National Championships, which are in Utah, but I don\'t know when.  But the  urlLink national track championships  are happenning right now in maybe Carson.  For the pros, the Autumn Classics are happening, and on August 1st, the  urlLink HEW Cyclassics Cup  will happen.  Don\'t expect to see Lance in it because he has a press conference that day.  '),(20051,3208138,'30,July,2004','2004-07-30 00:00:00','     '),(20052,3208138,'30,July,2004','2004-07-30 00:00:00','   urlLink Asian News Channel International  reports: \"Belgian Christophe Brandt was the only cyclist to return a positive drug tests during the Tour de France, the International Cycling Union (IUC) said.\"  That means that the Postie boys, including Lance tested clean, as promised. Armstrong claims that he is tested more than any of the other competitors in the Tour de France.  '),(20053,3208138,'28,July,2004','2004-07-28 00:00:00','   yeah dude, where\'s the muumu\'s?! anywho, last night i was watching sportscenter on espn as i often do to find out about the outcome of the dodger game that day. one of the funny commercials that espn plays is where sportscenter analyst dan patrick is working in his office, when the power apparently fails. so there is a long shot of patrick walking down the hallway, down the stairs, and into the basement of the building, all the way to a door that would apparently be where generator might be, and there\'s lance armstrong on a stationary bicycle.  lance: \"oh sorry dan, i thought everyone had left for the night.\" dan: \"you okay? you want an energy bar or something.\" lance: \"no, i\'m good.\" (resumes pedaling) i thought the commercial funny.   three cheers glen! we know you were stressing the test, but we all secretly knew that it was in the bag. i bribed mr. carver with rentals from blockbuster. always looking out for you i am.   the daily show w/ jon stewart has been hilarious coming from the democratic national convention this week. one of their field reporters said he filmed all of his reports ahead of time and left early since the convention was so scripted and boring. he made sure to do reports on a few possible exciting scenarios just in case, including a polar bear attacking the delegates from kansas, and hazardous quicksand discovered on the floor of the fleet center.  '),(20054,3208138,'28,July,2004','2004-07-28 00:00:00','   Lance Armstrong says he is fully capable of winning the Tour de France for a 7th time.  \"Time is limited but I think I could win the Tour de France one more time,\" he said.  George HIncapie says he is as motivated as he ever was and needs to win.  \"It\'s [winning] something he needs in his life.  I\'ve known the guy for a long time and that desire to win is as strong as ever.  It\'s part of him, what motivaties him and keeps him going.  It could be a World Cup race, whatever, Lance is just like that.\"  Lance hasn\'t said when he\'ll give it another go, but he\'s not getting any younger.  '),(20055,3208138,'28,July,2004','2004-07-28 00:00:00','   If everything goes as planned, this will be my last day in South Florida. I will drive south to visit a client in Florida City--about half way to the Florida Keys from Miami. If things donâ€™t go as planned, I donâ€™t know what Iâ€™ll do. Find a Starbucks and order a venti!   Someday I will take the opportunity to keep on driving until I get to Key West. I will save that trip for a time when Cindi travels here with me.   Some of you know that I was taking a test last week that would qualify me for participation in the Policy Governance Academy with John and Miriam Carver. I received word yesterday that I have been accepted--I passed the test!   The weather here has been steamy. This is the low season for tourism because it is warm, humid and wet--lot\'s of rain. I am looking forward to a trip here in February 2005 with the board of directors. I will help facilitate a tour, along with the Vice President of North American Operations for the board members and their spouses. That should be fun. I have invited Cindi to come along with me on this trip.   I am looking forward to watching OLN on Sunday with Daniel--there is a Lance interview. Should be good.  P. S. I am looking for the promised matching muumuu images from the land of Hawaii.  '),(20056,3208138,'28,July,2004','2004-07-28 00:00:00','   Daniel just got a velonews magazine for free for having a racing license-- sent to him with hopes that he would subscribe.  Here\'s what the cover lookes about like (not this grainy):  '),(20057,3208138,'27,July,2004','2004-07-27 00:00:00','   in 2006. Matt Damon says he wants to play the part of Lance.  '),(20058,3208138,'27,July,2004','2004-07-27 00:00:00','   sometimes i pretend i\'m jan. and i\'m not talking brady, i\'m talking ullrich. maybe that\'s why i feel like an underachiever....  '),(20059,3208138,'26,July,2004','2004-07-26 00:00:00','   There have been at least 3 ads on OLN for powerbar.com with a guy pretending he\'s Lance and imagining how a commentator would commentate it.  But the pros also pretend they\'re Lance.  Here\'s a quote from one:  \"When I\'m taining I imagine I\'m Lance Armstrong and I go up the climbs at full speed pretending I\'m unbeatable\" --Fabian Jeker, who lost the Tour of Switzerland by one second to Jan Ullrich, so his strategy almost paid off, too.  I definitely do not want school to start anytime soon, Rob.  '),(20060,3208138,'26,July,2004','2004-07-26 00:00:00','   the administrator ought to know that i tried voting in his cute little poll for best oln staff member, and it told me that i had already voted, although i have no recollection of doing so. i tried voting for kirsten gum, cause i hear she is a babe, and she must be a lot better looking than al troutwhatever. so let\'s see...my mom, my step-dad, my grandmother, my aunt, and my uncle are all in the tropics right now. how exciting for them! judging by the enormous insects i often encounter on my bedroom walls, i would\'ve sworn that i too was in the tropics. so far, summer vacation is really lame. i find that i have way too much time on my hands. i just want school to begin again, so that i have a reason to get out of bed before noon. it finally feels like i\'m getting things done on my own terms, because I want to, and that is a nice feeling. i feel momentum. but not at the moment, since it\'s summer vacation. i can\'t believe that i even went to summer school already, that seems like a long time ago. last night i watched a movie called \"napoleon dynamite.\" it was super funny, and a really clean movie. (ie: no swearing, adult humor or violence.) i\'d heard of the movie a while back, and later read that it was made by some graduates of brigham young university. that had put me off initially, because i\'ve been conditioned to think anything mormon is either really weird, or really wrong. i\'m proud of myself for overcoming my own prejudices, and allowing myself to enjoy something that could be enjoyable to any subculture. just for clarity\'s sake, there was no mention of religion in the actual film other than a mexican character who had a lot of traditional catholic symbols in his house. i\'m starting to realize that i\'m not taking advantage of my surroundings enough this summer. it\'s almost august, and i haven\'t been to one dodger game or visited the beach this summer! something must be done about these things. not at the moment however, since it\'s probably really cold at the beach at 12:13 am. songs of the moment on my very own ipod: \"ocean breathes salty\" by modest mouse, \"over it\" by onelinedrawing, and \"goodmorning aztlan\" by los lobos. download away...  '),(20061,3208138,'25,July,2004','2004-07-25 00:00:00','   Some of you are wondering what iPods are.  I\'m sure Glen can give you a better answer, but it is a device that can store memory to back up your hard disc, or you can download music onto it for listening with ease.  In a magazine I got today, someone said, \"I\'ve put many of my 1000 CDs onto my iPod  so that I can have all my best music with me wherever I go.\"  '),(20062,3208138,'25,July,2004','2004-07-25 00:00:00','   In stage 3 in the Tour de France, the riders had to go over cobblestones that are usually saved for the spring classics Paris-Roubaix and Ronde van Vlaanderen (Tour of Flanders).  The riders were not too happy about this, and an overall contender, Iban Mayo, crashed and lost almost 4 minutes to Lance because of that.  Here are some quotes pertaining to them:  \"Cobbles are for Paris-Roubaix, not the Tour de France.  They\'re not going to put any mountains in Paris-Roubaix next year are they [Paris-Roubaix is always flat]?\" --Tyler Hamilton  \"We\'re bike riders, we should ride on the road and not be gladiators.  If someone loses it [crashes] in front of you, you run over him and you crash.\" --Jens Voight  Source:  VeloNews (velonews.com).  '),(20063,3208138,'24,July,2004','2004-07-24 00:00:00','   Lance Armstrong won an unprecedented sixth Tour victory this year while beating the crapes out of all his rivals (remember that add, OLN watchers?).  Today was a very easy day around the Champs-Elysees, with Lance in a special yellow uniform.  Tom Boonen won the stage, and Robbie McEwen got the green jersey for the best and most consistent sprinter.  Read more  urlLink here .  Lance in his different uniform.    Lance celebrates #6.    Tom Boonen celebrates his second stage win this year.    Winners of jerseys, left to right:  Robbie McEwen, sprinter; Lance, overall classification; Richard Virenque, King of the Mountains (for a record 7th time); Vladimir Karpets, Best Young Rider.  '),(20064,3208138,'24,July,2004','2004-07-24 00:00:00','   I really enjoyed the little bio about the Mennonite racer.  Hope he does well in the years to come  '),(20065,3208138,'23,July,2004','2004-07-23 00:00:00','   Lance Armstrong showed he is the king of the Tour today when he won the Time Trial to BesanÃ§on.  His teamates went out with marvelous times, then when \"Yawn\" Ullrich came around, he set the best times, with his teamate Andreas Kloden not far away from him.  Ivan Basso was next, but he didn\'t do too well (as expected).  Lance Armstrong came out of the starting gate with the fastest times at all of the check points, and beat Ullrich (2nd place) by 61 seconds!  This was his 5th victory in this year\'s Tour, and 21st victory total in the Tour de France.  Tomorrow is an easy 100-mile day ending on the Champs Elysees. The final standings are Lance Armstrong, first, Andreas Kloden, 2nd, Ivan Basso, 3rd.  (Ullrich, a five-time runner-up, was 4th).  Jose Azevedo was 5th, 14:30 behind his team leader.  American Levi Leipheimer (Rabobank) was 9th.  Thomas Voeckler, who was in the yellow jersey for 10 days, started in the white jersey today as the best young rider.  However, he lost that lead when he lost way more than the 45 seconds he was leadings the Russian Vladimir Karpets (Illes-Ballears).  Richard Virenque (Quick Step-Davitamon) won the King of the Mountains competition with 226 points, and Lance was 2nd, with 172 points.  T-Mobile won the team competition, with the Posties 2nd.  The Sprinting competition is led by Robbie McEwen (Lotto-Domo), but the Norwegian Thor Hushovd could beat him, because the competion continues on the Champs-Elysees, and he is not far off.  For the overall contenders, tomorrow should be a fun ride around Paris with smiles all over, and some Champagne for Lance, which apparently hard his easiest Tour ever, with 5 stage wins and the team time trial in his palmares.  \"I\'m happy, because it\'s finished,\" he added. \"I\'m tired mentally, have tired legs.\"  '),(20066,3208138,'23,July,2004','2004-07-23 00:00:00','   Beth is sitting here next to me chatting with Rob while I am watching the expanded coverage of T de F.  That\'s a nice surprise.  '),(20067,3208138,'22,July,2004','2004-07-22 00:00:00','   I friend and colleague of mine, Rudy Carrasco, has learned this week that his 4 year old son Samuel, has leukemia. Rudy is an avid blogger and has set up a blog for  urlLink Sam Sam here . I told Rudy about Lance Armstrong, who is a cancer survivor (leukemia is a kind of cancer). Rudy is now posting about the bicycle race, Armstrong\'s Foundation, and cancer survivorship on his regular blog called  urlLink Urban Onramps . So now I wear my yellow wristband for Rudy\'s Sam Sam also.  '),(20068,3208138,'22,July,2004','2004-07-22 00:00:00','   Don\'t forget to vote for who you think is Lance\'s best domestique.  The winner receives $10,000! Currently, George Hincapie has 55% of the votes, Floyd Landis has 24%, and Jose Azevedo has 12%.  To vote, click  urlLink here .  '),(20069,3208138,'22,July,2004','2004-07-22 00:00:00','   I have enjoyed Floyd Landis and his role on the USPS team to help Lance. A few months ago on a tour in Lancaster County, Pennsylvania, Floyd\'s family\'s farm was pointed out to us by out tour guide. His description was that Floyd\'s family is Amish. Had Floyd grown up on the south side of Hwy 30, he would not have been permitted to use a bicylce, the Amish are only allowed scooters on that side of the road.   So, I was curious when I read in Every Second Counts, Armstrongs autobiography that Floyd is Mennonite. Then, Robin William described the family as Mennonite. If you\'d like to know more, read  urlLink this . I guess this interests me because I work for MEDA (Mennonite Economic Development Associates). This link, thanks to Hugo Swyzer, a Mennonite in Pasadena whose blog I read ocassionally.   You can also read Floyd\'s  urlLink online journal here . I think Daniel may have already linked us to that one.  '),(20070,3208138,'22,July,2004','2004-07-22 00:00:00','     '),(20071,3208138,'22,July,2004','2004-07-22 00:00:00','   The riders in the Tour de France do not like the heat and even though they drink a lot, they can still get dehydrated.  Read about it in a journal entry by American Christian Vande Velde  urlLink here .  '),(20072,3208138,'22,July,2004','2004-07-22 00:00:00','   Juan Miguel Mercado (Quick Step-Davitamon) won Friday\'s stage of the Tour de France while Lance Armstrong retained his overall lead safetly in the bunch.  Jaun Miguel Mercado was in a breakway, and attacked with 12 km to go, and Garcia-Acosta followed.  Garcia-Acosta was expected to win because he was stronger and more experienced, but the Quick Step drafted him to the line and outsprinted him. This is the third stage victory for the Belgian Quick Step-Davitamon team.  Earlier in the race, Italian Filippo Simeoni (Domina Vacanze) bridged the gap to the breakway and Armstrong followed because of personal tension between the two.  T-Mobile or CSC (Lance\'s main challenger\'s teams) would definitely not let Lance go, and the break knew that, so they asked him to leave, but he said he would only leave if Simeoni left, so he did.     Not the easiest route, but nothing serious.    Tomorrow is the Final TT, which Lance has always won in the past except for last year.  Jan Ullrich could also make a splash.  '),(20073,3208138,'22,July,2004','2004-07-22 00:00:00','   The US Postal team aero bars (by Bontrager) are now available to the general public.  630 grams, $600.     The team\'s helmet, however, is not available yet to normal people.     Nor is the Trek Madone SL, the bike the Posties ride in races with cobblestones.  '),(20074,3208138,'21,July,2004','2004-07-21 00:00:00','   In  urlLink his last diary entry ,Tyler Hamilton paid tribute to his dog (and best friend, I suppose) Tugboat.  Tyler and Tugboat.    Yes, the riders have the same idea that you do on the crowds of the Tour de France.  In a recent entry of Christian Vandevelde\'s diary, he says, The crowds were insane from the start of the climb all the way to the top. At times it was scary, as I didn\'t really know if I would make it through the crazy screaming fans. By the time we started racing the fans had finished their lunches, finished their bottles of red, cans of beer and topped it all off with shots of grappa. When people are drunk, reaction times are slow and the noise is deafening-good and bad when you\'re racing up a hill with sweat in your eyes.\"  Here, you see the crazy fans.  '),(20075,3208138,'21,July,2004','2004-07-21 00:00:00','   Lance Armstrong showed he deserves a sixth Tour victory today when he won the 17th stage of the Tour de France (making it 4 stages this year).  US Postal set a fast pace, and the only ones that could keep up were Jan Ullrich, his teamate AndrÃ©as KlÃ¶den, Ivan Basso, currently in 2nd place behind Lance, and Lance himself led by Floyd Landis.  The stage ended with AndrÃ©as KlÃ¶den sprinting away at 1 km to go, and everyone followed, and Lance sprinted out and came out right ahead of him.  Jan Ullrich probably wasn\'t too happy about KlÃ¶den because he took the 2nd place time bonus (12 second), leaving Jan with just the 3rd place bonus (8 seconds), putting him at 8:08 back, in 4th place, with KlÃ¶den in 3rd, at 5:11.  Ullrich has never been worse in the Tour than 2nd place.  There were some early breakaways for the King of the Mountains competition and to be the first person to be at the highest point of the Tour, the 2000-meter Col de la Madeleine, which is rewarded with a 5000-Euro bonus (6000 dollars), which Gilbeto Simoni got, just edging out Richard Virenque, with the Armstrong group behind.  '),(20076,3208138,'21,July,2004','2004-07-21 00:00:00','   Lance Armstrong distroyed his rivals\' Tour hopes when he beat his normal challenger, a great time trialist, Jan Ullrich, by 61 seconds in just 15.5 km (about 9 miles), putting the German 7:44 back, and Ivan Basso, Lance\'s closest challenger this year, who is not a good time trialist, is now 3:38 behind Lance.  The time trial was up the legendary Alpe d\'Huez, which turned out to be a bad idea because of the 900,000 screeming and crazy fans lining the 15.5 km route.  \"I don\'t know that that is such a good thing for the Tour de France. I don\'t think it is safe. I think you\'d agree. I\'m sure the organizers agree,\" Lance said, and the organizers did agree.  Lance\'s time up the Alpe was the 2nd fastest in history, just one second off Marco Pantani\'s record.  Robbie McEwen, the sprinter\'s competition leader, did a wheelie when he crossed the finish line, which was a strange sight.  Lance goes hard.    Ullrich goes all out, but it wasn\'t enough.    Basso loses a lot of time.    It does look like Lance has the Tour in the bag, but he says the deciding day is tomorrow, a tough mountain stage.  '),(20077,3208138,'20,July,2004','2004-07-20 00:00:00','   This is from  urlLink Discovery Magazine . This is what they say about it:  --Today, Space Imaging\'s IKONOS satellite took a new image of the area covering Stage 16 of the Tour de France in the French Alps. The image of the 9.6-mile (15.5 km) race course from Bourg d\'Oisans to L\'Alpe d\'Huez was collected at 11:52 a.m. (local time) from 423 miles in space. American Lance Armstrong won this individual time trial stage in 39 minutes, 41 seconds - the only rider under 40 minutes.   The image shows the winding switchback road to the town of L\'Alpe d\'Huez lined with cars, RVs and hundreds of thousands of cycling fans. The town itself is also packed with tents, vehicles and people.  '),(20078,3208138,'20,July,2004','2004-07-20 00:00:00','   Did you know if you mix Daniel\'s name and Lance\'s name you get that! Cindi made up the Dance part. I added the Wilstrong part. (Daniel, has anyone ever said that before?)  Anyway, speaking of my nephew Dance, he was telling me about a  urlLink bicycle race  at  urlLink Heritage Park  last night. He rode faster than anyone else in his age group in a 20 mile ctriteirum. Hurray for Dance!  This post marks the 300th post to the PWWCR family blog! Huray for us!  '),(20079,3208138,'20,July,2004','2004-07-20 00:00:00','   First, I was shocked to find out this morning that Lance Armstrong\'s mom is my age. What?  Second, I am in awe of Lance\'s and the team\'s performances in each stage of this bike ride. They just seem to plan the work and work the plan. That is an inspiration for me with various personal goals. I\'d like to be there watching in France during the third week, but not with that kind of crowd.  What is up with the announcers starting to proclaim victory with Lance? Doesn\'t anyone else get a little nervous with that? What if he falls? Or what if a silly spectator interferes? Aren\'t there still some variables out there? And there doesn\'t seem to be any kind of big time gap between the leaders that make me relax. Am I crazy or should I start the Lance dance now?!  '),(20080,3208138,'20,July,2004','2004-07-20 00:00:00','   Squirt, I mean --- Sam did fine at puppy training. He performed to the \"Let\'s Go!\" command fairly confidently. He also embarrassed me with a new greeting. When the male instructor petted and talked to him at the start of class, he squirted a quick stream of liquid, let\'s just say, into the air. He did not hit any one. He was seated and leaning back a little, so the fluid just arched through the air with the greatest of ease. This happened one or two more times when various men in class petted him. No one was in the way.   Sam seemed oblivious to this new reaction. I\'m guessing his male dog \"marking\" ability kicked into gear without the control factor as a natural part of puppy development. Again, never heard of that before, but it must be typical.   The command for this week is \"Sit\" and \"Stay\" with hand motions to communicate what we want --- that\'s the tricky part. And practicing in the heat is no fun, especially on little paws touching the pavement.  '),(20081,3208138,'20,July,2004','2004-07-20 00:00:00','   Dick and Jane Remington will be traveling to Whittier and spending the week of August 17 - 24 to play with the granddog. A summer visit will permit some different activities. I am looking forward to introducing them to the wonder dog. I\'m sure a family BBQ will be planned at our house. Hopefully, Daniel will give them an orientation to the Tour de France via video highlights.  '),(20082,3208138,'20,July,2004','2004-07-20 00:00:00','   I think that I could stay awake easier in those early hours if Robin Williams was a regular with Phil and Bob. What a hoot this morning.  Can we put his name on the vote to the right?  '),(20083,3208138,'19,July,2004','2004-07-19 00:00:00','   Well, really? Lance Armstrong won today\'s stage of the Tour de France (making it 18 stage wins for the Texan) after beating Ivan Basso to the line as he did on Saturday\'s stage to Plateau de Beille.   This time, however, he took the yellow jersey from France\'s Thomas Voeckler because he lost more than nine minutes to Lance, and going from first place to 8th.  Jan Ullrich attacked on the 4th climb of the day (with six climbs total), a first catagory Col de l\'Echarassson.  The German gained more than a minute on Lance, but CSC and USPS (Basso\'s and Lance\'s team) reeled him back in and he lost three more seconds in the end, putting him 6:54 behind Lance.  \"Basso is riding so good that there is a chance I can lose time tomorrow,\" Lance said about the very difficult TT to Alpe d\'Huez.  Other news:  Iban Mayo did not even start today\'s stage.  This was the kind of terrain the riders did.    Armstrong celebrates his stage win.    Tyler\'s back trouble (the spots). Read his journal entry about it  urlLink here .  '),(20084,3208138,'19,July,2004','2004-07-19 00:00:00','   Basso says he will only attack Lance if Lance shows any signs of weakness, which the Texan has not shown so far. \"It\'s possible to win the Tour de France in one attack,\" Basso said. \"Last year at Luz Ardiden, Armstrong won the race in one attack. I don\'t know if Armstrong will have a bad day. If he does, we will be ready. It\'s not worth it to attack just to gain 10 to 20 seconds.\"  However, Basso\'s teammates got injuries recently from crashes etc, so that\'s a disadvantage to Lance with his strong team always there for him.  Domitilla, Ivan\'s daughter, sits in a happy father\'s arms.    A new name, Andreas Kloden of Ullrich\'s T-Mobile team, is expected to be pretty high up in the overall standings in Paris ahead of Jan Ullrch, the team leader, if he keeps up his climbing abilities and doesn\'t fall behind in the TTs (time trials).  '),(20085,3208138,'18,July,2004','2004-07-18 00:00:00','   Gravity Files, Outside Magazine--re-airing of stage 13 and a few of the Lance Chronicles. What will I do with myself today?  The Today Show seems sooooooooo unimportant [really when is it anything else].  Meanwhile in the real world--France--where the Postie Boys and their wannabe assumptive competitors are having a \"rest\" day.   urlLink      Looks pretty warm there. I am intrigued by the individual time trial coming on Wednesday--straight up a hill, 7.9% grade for 15 kilometers. Isn\'t Turnball Canyon about 6%?   Well, I think I still have a job. I\'ll go check. Rob wants to go for a bike ride sometime today.  '),(20086,3208138,'18,July,2004','2004-07-18 00:00:00','   Aitor Gonzalez (Fassa Bortolo) won Sunday\'s 14th stage of the Tour de France today after getting in a 10-man break and attacked 6 km from the finish (about 4 miles).&nbsp; Voeckler retained the overall lead, and Lance is still 22 seconds behind, with no major moves by any of the overall threats.&nbsp; Tomorrow (Monday) is a rest day.   As for Ivan Basso being on the podium in Paris, I was thinking of him being second (or third), but not first (because of Lance in the way).&nbsp; It would be nice to be in the leader\'s jersey, but we\'ll see where my life takes me.   Why does blogger change so much? It\'s hard to figure out how to do things another way so much. I think I figured everything out, though.  &nbsp;  '),(20087,3208138,'17,July,2004','2004-07-17 00:00:00','   I will be surprised if Basso wins. He has a history of miscalculations and mishaps. He also is not as good on the flats as Lance and is not expected to do well on the time trial. Yet many individuals in this race have not lived up to expectations and there is a lot of racing still ahead. We will see if Daniel\'s prediction is fulfilled. I do expect Daniel to be the Voeckler of the \"le tour de france 2014\".  '),(20088,3208138,'16,July,2004','2004-07-16 00:00:00','   I have really enjoyed the progress of Sam 2 in his school.  Keep it up Cindi ! !  Also enjoying the reports on the race, not that I can understand it all, Daniel,  Will be nice to see if your thoughts on the winners will be true.  Keep the information coming.  '),(20089,3208138,'16,July,2004','2004-07-16 00:00:00','   You can get a Lance poster by Giro  urlLink here .  '),(20090,3208138,'16,July,2004','2004-07-16 00:00:00','   Lance won very difficult stage 13 to Plateau de Beille with Basso, the last stage\'s winner, took second right behind.  It was a very dramatic day with Jan Ullrich losing another 2:42 to Lance Armstrong, putting him 6:39 behind Lance.  Tyler Hamilton quit because of severe back pain, with others quiting also.  Iban Mayo almost quit today after being dropped on a lesser hill, but he was persuaded to keep going by his teammates, other riders, his manager, and Basque fans, but he lost almost 38 minutes to Lance.  Stage 9 winner, Tom Boonen, crashed on a descent, when some part of his bike broke apparently, and overall threat, Roberto Heras, also crashed on another descent, and he also finished, but his overall hopes are history.  Be prepared to see Ivan Basso on the podium in Paris.  Lance takes the stage ahead of Basso.    Jan Ullrcih loses more time.    The yellow jersey is still on a very hard-working Voeckler.    Tyler Hamilton quits with back pain.    Mayo almost quits.  '),(20091,3208138,'16,July,2004','2004-07-16 00:00:00','   Samson #2 did really well today on his training task of \"Let\'s go.\" He stayed on my left side with slack in his leash...Not pulling nor dragging behind and a minimum of puppy play. I was very encouraged and he almost seemed proud, too. We got a \"training collar\" (formerly known as a choke chain), which seems to help the process. We practice with him each day, so it was a thrill to see such progress today.  &nbsp;  Sam marched around in his new pool yesterday. Just a molded plastic shell thing that cost hardly anything and can\'t be bitten and deflated. Glen went in, too, to demonstrate how to splash and enjoy the water. It\'s deep enough for Sam to be in to his chest, but shallow enough for him to step/climb in and out. The fun increased with the addition of pool toys: a little floaty ball and a piece of wood from somewhere in the garden. He pawed and chased those around and seemed to have fun. Soon he\'ll be ready for an actual swim somewhere. I had fun, too, watching him cool off.  &nbsp;  The Tour de France was good again today. I enjoyed the ending very much. The rain is something else! Can\'t imagine riding that fast, downhill, in the rain. I\'d be tortured. It\'s fun having Daniel here some days to watch, too.  '),(20092,3208138,'15,July,2004','2004-07-15 00:00:00','   Lance Armstrong showed just why he is a five-time Tour de France champion when he took second on today\'s stage, gaining two minutes and thirty seconds on his closest rival, Jan Ullrich, and 3:27 behind Tyler Hamilton, and 2:57 over super-climber and ex-teammate, Roberto Heras, now on the Spanish team Liberty Seguros.  In the overall standings, Thomas Voeckler (Brioches la Boulangere) made a huge effort and kept his yellow jersey for another day.  It was a day of much rain, making the descent dangerous, but luckily, no one crashed, and at the end the roads were dry.  The posties (in the rain) made the pack go in many groups  because of their hard pace.  Lance kept that pace and even sped up when his last teammate was through and made many suffer more.  Ullrich was just having his day in the first day of the mountains in his huge gear.  Lance could have won the stage, but he is trying to help Italian Ivan Basso\'s mother fight cancer. \"We\'ve been friends for a long time, and off the bike we\'re trying to work a little bit on his mom\'s situation, to try and see if she can win the fight against cancer,\" Armstrong said.  Basso wins the stage ahead of Lance.    Ullrich has trouble.    Hamilton also struggles.  '),(20093,3208138,'15,July,2004','2004-07-15 00:00:00','   What an exciting morning of racing. What an effort by team postal. Another example of how important working as a team is in cycling.  '),(20094,3208138,'15,July,2004','2004-07-15 00:00:00','   The Tour started out with 55-degree weather with rain causing many crashes, but now it is up to 100 degrees, \"But I don\'t think anyone is complaining after the [cold and rain-sodden] first week,\" said Lance Armstrong, defending Tour champ.  Because of this (and on-and-off back pain), Paris-Roubaix (a \"Classic\" with many cobblestones) winner Swede Magnus \"Maggy\" BÃ¤ckstedt, and also the heaviest rider in the peloton, had to call it quits.  As for the good things of the stage, there was a three-man breakaway that turned into a chase group when one the the three attacked (Frenchman David MoncoutiÃ©) and won the stage.  The overall standings remained the same with the Texan ahead of all his rivals.  As for Saturday\'s extremely hard stage, he said, \"It\'s very nasty. I think they should take it out. It\'s a violation of our human rights. I love it,\" showing his Texan behaivor all around.  But for tomorrow, Iban Mayo is expected to attack because it is in front of his home crowd (the Basque).  '),(20095,3208138,'14,July,2004','2004-07-14 00:00:00','   Tyler Hamilton, #11 in overall standings in the Tour after Stage 10 is pictured here with his 9 year-old golden retriever Tugboat.      Tugboat has been sick for some time,  diagnosed with cancer this week, passed away in France.  '),(20096,3208138,'13,July,2004','2004-07-13 00:00:00','   It looks like 3 guys withdrew from the Tour today. Watched SÃ©bastien HINAULT get taken away on a stretcher and ambulance. Yesterday a guy just ran out of juice, stopped, and handed-off his bike to a mechanic and got into the team car. Not to indulge in too much schadenfreude!  I am looking forward to seeing LA pull out of his #6 position. He was in good spirits yesterday, joking with the press.  '),(20097,3208138,'13,July,2004','2004-07-13 00:00:00','   Richard Virenque won the first mountain stage (and also the longest stage) of this year\'s Tour de France, while the overall standings have not changed much because Lance\'s main rivals finished not long after him. Thomas Voeckler retained the overall lead, but only because of his teammates\' help.  The Frenchamn Richard Virenque wanted to win on this day because it is  Bastille Day in France (July 14). He was in a two-man breakaway with Axel Merckx (Lotto Domo), but the Belgian could not hold Virenque when he upped the pace.  He says that he will really pay for the effort in the days to come like he did in last year\'s Tour when he took the yellow jersey and the King of the Mountains (KOM) jersey. He only took the KOM jersey this time.  '),(20098,3208138,'13,July,2004','2004-07-13 00:00:00','   Beth has Brianna for the whole day tomorrow. Watch for non-tour pictures coming to this blog. I would expect that they involve swimming or playing with water. You can see the temp on the side bar to the right. hot, hot, hot.  '),(20099,3208138,'11,July,2004','2004-07-11 00:00:00','   This is Robbie McEwen\'s second sprint victory this year.  There was a two-man breakway that lasted most of the ride, and unlike most of the breakways so far this year, they were certain of a victory, but the peloton reeled them in just in time for the sprinters to take over.  This is a little sad to watch, but at the same time, it\'s one of the best things in bike racing.     McEwen sprints to the finish line.     The General Classification (GC) remained the same.  Answer to Cindi\'s Question: A person that you\'ve never heard of can take the leader\'s jersey because of a smart breakaway like the guy on Brioches la Bolangere did.  The favorites are not scared because they know they will get the time back in the mountains, which start tomorrow.  Lance says there should be an individual time trial (ITT) so that this won\'t happen, and only real contenders will get the jerey, but Jean Marie Leblanc (the guy in charge of the Tour) likes to have it the way it is.  In order to have a exciting finish, he makes the real action  (the mountains, not the crashes) at the end of the Tour with a sprinter\'s first part of the race.  So anyway, it\'s because of a breakway with the times so close together already.  '),(20100,3208138,'11,July,2004','2004-07-11 00:00:00','   Well, we took Sam to the first day of dog training last night. And of course he napped in the mud to cool off prior to class and dirtied his new school fur. Off we went, helping him into the back seat (he doesn\'t quite get the jumping in part), and wondering what the evening would be like.   He is riding in the car well, now --- staying seated most of the time. It helps that he is big enough to see out the window, I\'m sure. The evening went well. He was very timid...tail between his legs, ears low, and body slouching down, when we arrived. He stayed very near to us and sat and looked around. To others, he seemed well-behaved, but we know the truth.   He was nervous about the first day of school. Would the others like him? Soon enough, he was sniff-greeted by a few others...the 9-week old rotweiler (how do you spell that?)and another BIG lab. The rot puppy was the only dog smaller than him, and her name is ---of all things---Delilah! Ha. Samson and...well, you get the picture. Sam stayed calm, sitting/or laying/lying/lieing (grammar patrol --- help!)and watching...often facing the wrong way for the instruction.   Then came the first walking on a leash practice with Glen. Well, he made great improvement by the end of the evening. Enough said. The end of class is \"free play.\" All the owners let go of the leashes and the dogs are allowed to interact. It\'s fun to see.   Sam seemed to get along OK, though he seemed uncertain at first. Then he romped gleefully with a...Great Dane! A sight to see. Quite the size contrast. The Dane got so tickled he started to prance/pounce on/at/around Sam. Sam just flattened out on his belly on the ground...like a pancake or like someone about to be arrested after a car chase. No one was hurt. I was nervous, but laughing.   And home we came. Sam got a big drink of water and fell sound asleep.  '),(20101,3208138,'11,July,2004','2004-07-11 00:00:00','   In order to keep harmony in our home, I am compelled to blog now and then. I do enjoy reading the family blog, though it seems to have a single focus these days. I\'ve been safely coasting along in the family blog peloton, but now I\'m breaking away to post again. My Mac cannot support the posting, and I\'m now using the PC. So I\'m back from the silent period of reading and not responding.  I am enjoying the Tour de France. Yesterday was a day off, and I was glad to catch up on other responsibilities. I found myself anxious last night, though, to watch the next stage. To me, it\'s been like reading a good book...a real page turner. The kind where I can\'t put the book down because I want to know what will happen in the next chapter, er -- stage. I enjoyed the stage today. The countryside and towns, the blue sky with puffy clouds, and the crash! What? Again. Yes, and this time the cameras seemed to capture the whole thing. A guy hit a thing in the middle of the road and some one hit him and tumbled hard. The second guy already had a BIG bandage on his forearm. I\'d be thinking...\"I could be on a swim team. This is crazy! Neosporin and an Advil please. And airlift me to the nearest...hot tub! But no, not the mighty cyclism warriors. Back on the mighty steel (or some speedy alloy?)steed with wheels and charge forth. Just for the chance to wear polka dots? Or some other color shirt. Just go to the closet and pick one! What\'s the big deal? [I\'m kidding. Of course, I get it and know it is not just to wear a new shirt!]. Can someone, however, explain to me how a guy who\'s name I never hear mentioned day-to-day during the race ends up keeping the yellow jersey? How does that work exactly? I am clearly missing something simple and obvious to the family experts, no doubt.  '),(20102,3208138,'11,July,2004','2004-07-11 00:00:00','   Monday is a day off for Lance and the TdeF riders as they will fly to the next starting point. The schedule says they will be in Limoges for their rest day.   urlLink      Now, according to Daniel, the interesting thing about Monday\'s coverage is that they will show footage of the race and the the significant events to this point. This will include a re-showing of the crashes up to this point. They will likely show these several times over. And, Daniel will come over early to see it all!  '),(20103,3208138,'10,July,2004','2004-07-10 00:00:00','   Thor Hushovd won Sunday\'s stage 8 of the Tour de France, after the field caught multiple breakaways (in the second picture you can see one of them), and Thomas Voeckler retained the overall lead.  There was a crash at the back of the peloton, due to a dog crossing the road at the wrong time, but no one expected to have any affect on the race was hurt.  The overall was not affected much at all.  Read more  urlLink here .          The riders traveled by plane to the next stop for tomorrow\'s rest day.  '),(20104,3208138,'10,July,2004','2004-07-10 00:00:00','   Get your 2005 cycling calendars now with  urlLink World Cycling Productions !  Only $15 each.  History of Tour de France or current cycling.  '),(20105,3208138,'09,July,2004','2004-07-09 00:00:00','   Stage seven of the Tour de France was won by Italian Filippo Pozzato, who was in a breakway that broke away at 2 km to go and opened up a 15-second gap.  He is a Tour first-timer and the youngest rider in the race.       There was another breakway that was caught well before the finish line with just two riders, Dutchman Erik Dekker and Belgian Thierry Marichal.       The overall standings remained the same with Voeckler retaining the overall lead 9:35 ahead of Lance, 10:11 over Tyler, and 10:30 ahead of Jan Ullrich.   There were more crashes, but the main contenders were not hurt.  '),(20106,3208138,'09,July,2004','2004-07-09 00:00:00','   Well, Tyler Hamilton crashed again, but at least he hasn\'t broken any bones yet.  He landed on his back and got \"significant bruising\", as a Phonak team spokesman put it.  Lance Armstrong did crash in an earlier pile up (just 14 km from the start), but is fine except for a few burns from a tire.       As Glen said, Robbie McEwen, the guy leading in the sprinter\'s competition, crashed, has bad abrasions on both hips and arms, and is out of the Tour.   Tom Boonen won the stage in a crash-marred uphill finish.     Voeckler is still the yellow jersey.     Read more  urlLink here .  '),(20107,3208138,'08,July,2004','2004-07-08 00:00:00','   I found this easy to understand  urlLink FAQ section  at  urlLink cycling news.com  wtih some of the basics for us who are new to the T de F.  '),(20108,3208138,'08,July,2004','2004-07-08 00:00:00','   After the massive pile-up at the 1 KM mark, I saw Robbie McEwen\'s leg and thought--\"hamburger\". The creative writing staff on the official T de F web site put it this way: \". . .Robbie McEwen who slid along the tarmac long enough for him to tear a big hole in his green knicks and make his powerful quadriceps look more like a pizza cooked up with the dangerous combination of speed and gravel.\"  It was amazing to see Juan Antonio Flecha so far in the lead for such a long time. What happened demonstrated how this is so much a team sport and not an individual thing. When the peloton swarmed down on him and passed him--I think it was at 8 KM before the finish--it looked like he had all of his energy drawn out of him. PL said it looked something like predatory birds swooping down on him as they approached. It was dramatic.  I hope we get the real commentary from Daniel today. Even though he didn\'t come over today to watch, he will have the inside track. Daniel--you might want to come over to watch the expanded coverage later today!  '),(20109,3208138,'08,July,2004','2004-07-08 00:00:00','   Wear Yellow--  The yellow wrist bands have arrived. I orderd 10, so come by to get yours soon.   The  urlLink Lance Armstrong Foundation  is raising money through the sale of yellow wrist bands and its  urlLink WearYellow, LiveStrong  compaign. I have bought 10 of them for anyone who want to wear them.  The cancer connection in our familiy is in our own grandma Pete who is a two time cancer survivor.  '),(20110,3208138,'07,July,2004','2004-07-07 00:00:00','   There was a five-man breakaway that gained 16:30 over the field, but the field got back four and a half minutes by the time the break crossed the finish line because there were strong winds keeping the peloton from going too hard. Stuart O\'Grady won the stage, but it was Thomas Voeckler (Brioches La Boulangere) that took the yellow jersey, who was fourth in the breakaway, and started the stage three minutes down on Lance, by more than nine and a half minutes.   (Armstrong is not concerned about him, and will not fight to get the jersey back too soon.)  It was another raining day, and there was a mass pile-up half-way through the stage, which took down three of Lance\'s teamates, two sprinters, Alessandro Pettachi and Robbie McEwen,  Bradley McGee, and Roberto Heras, a rival of Lance, but Lance himself was not hurt.    Top: you can see that the wind was quite fierce in the picture here because of how the riders are all single-file across the road; bottom right:  Portugal\'s Jose Acevedo, and the Spanish trio of Jose Luis Rubiera, Manuel Beltran and Benjamin Noval go down in a crash, bottom left: the new yellow jersey.  '),(20111,3208138,'07,July,2004','2004-07-07 00:00:00','   Someone was asking about women in professional cylcing on Sunday. There is a  urlLink Women\'s Giro \'d Italia  that you can read about at the link.   Yes, Steve: I left a really good comment on your post. It didn\'t show up, so I tried it again. Didn\'t work a second time either. Just know that they were really smart!  '),(20112,3208138,'06,July,2004','2004-07-06 00:00:00','   my last blog states there are two comments, but when I check there are none there????  '),(20113,3208138,'06,July,2004','2004-07-06 00:00:00','   Today was the Team Time Trial that USPS won last year and moved Jan Ullrich 43 seconds farther away from Lance.  Today?  Well, because of a new rule, Ullrich only got 40 seconds farther from the yellow jersey! What I mean is, US Postal won, and Ullrich\'s T-Mobile team took third. Good news, and another one of Lance\'s rivals, Tyler Hamilton, his team took second, more than one minute down from USPS.  I guess the new rule says that you can only lose 20 seconds from next team up, so Tyler just lost 20 seconds to Lance.  Lance now leads the Tour de France, although he doesn\'t want to defend it yet because it would strain the team too much for the mountains in front of them.  It was a rainy day with a tail wind, which made it very difficult to stay upright on the paint on the road.  Tyler Hamilton\'s team, Phonak, lost 3 or 4 guys due to this (and puntures), so they only finished with 5 guys, the bare minimum.  Here you can see how wet it was for US Postal, which had the worst weather.  Lance seems like a really happy guy!  '),(20114,3208138,'06,July,2004','2004-07-06 00:00:00','   neither rain nor snow nor fallen comrade will deter lance from winning today. not bad for a postal rider.  '),(20115,3208138,'06,July,2004','2004-07-06 00:00:00','   This likely is what those cobble stones look like up close:     They are available from  urlLink Belgain Block Co.  '),(20116,3208138,'05,July,2004','2004-07-05 00:00:00','   Well, someone did lose their Tour hopes on stage three\'s cobbles. Thankfully, it wasn\'t Lance, but a Basque threat, Iban Mayo, lost almost four minutes due to a crash on the cobblestones of Belgium and France (and Jan Ullrich\'s T-Mobile Team putting a lot of effort in at the front to make sure he wouldn\'t reach the pack).  Jean Patrick Nazon took the stage in a sprint, while Robbie McEwen took the leader\'s jersey.  He is no threat to Lance when the race hits the mountains.  Tomorrow\'s stage is a team time trial which US Postal won last year and is expected to win again this year.  '),(20117,3208138,'05,July,2004','2004-07-05 00:00:00','   \"I think everybody is worried about the cobbles - even the riders who are good on them are worried,\" Armstrong said.   \"You could be involved in a crash and that would mean the end of your Tour. Somebody\'s Tour will be over on Tuesday and I could be one of those people.\"   Are you nervous yet?  '),(20118,3208138,'04,July,2004','2004-07-04 00:00:00','   on July 6. 16 kilometers from Brussels, Belgium. The race is 210 kilometers to Wasquehal.     urlLink  '),(20119,3208138,'04,July,2004','2004-07-04 00:00:00','   To blog users:  I am finding that almost none of you have made yourselves known on the \"Contributors\" section on the right side of the blog.  In order for us to know you better, I think this would be a great way to do it.  To blog administrator:  When looking for the word \"comentary\" at dictionary.com, it says, \"No entry found for Comentary\" and says below, \"Did you mean Commentary?\" And yet this \"word\" appears on the side of the blog.  Another thing:  I notice on google.com that the \"Googe\" thing on top of the page is changing a lot.  Today for Independence Day, for instance, has a red, white, and blue Google icon with an eagle, but on the blog it is always the same.  Is there any way to change that?  Just curious.  '),(20120,3208138,'03,July,2004','2004-07-03 00:00:00','   Today AG2R\'s Jaan Kirsipuu won the first stage of the Tour de France.  Fassa Bortolo\'s Fabian Cancellara retained the overall lead, while Lance takes third place overall.  It was a cold and wet day, normal for Belgian racing, and had numerous crashes like I had predicted, but at least no overall contenders were seriously hurt (or was that a bad thing?).  Yes, Tyler Hamilton crashed again, but he got back on his bike immediately.  OLN advertises its race coverage as \"The Cyclism\", and after watching a day\'s worth of racing, I really want to go out and do something like what the pros just did.  Yesterday I sort of did a short time trial, and after I was finished riding, I was still thinking about it and completely forgot about what I was doing (and wanted to either keep riding or blog something.  Oh, the Cyclism!  '),(20121,3208138,'03,July,2004','2004-07-03 00:00:00','   I posted news about the Tour de France.  However, I forgot to write about an important aspect of it--I saw it at Glen and Cindi\'s house with wonderful hospitality.  It was a wonderful time breakfast at their house with Mom, Dad, Beth, Steve, Kyle, Rob and Grandma.  Too bad Dick and Jane couldn\'t be there.  Thanks again!  '),(20122,3208138,'02,July,2004','2004-07-02 00:00:00','   Lance took second place in the prologue while his biggest rival, Jan Ullrich, took 16th!  Good news so far.  Here\'s lance on his TT bike.  The most critical stage, a time trial, is on l\'Alpe d\'Huez, and lance rode it at least ten times in May to \"learn every centimeter of it.\"  There was one crash on the course when the rider started pedaling while he was on a corner (no one worth mentioning).  Actually somenone had a worse crash before the Tour even started when he was riding the course to know the corners as well as possible.      Eddy Merckx was the best rider that ever lived and won a third of the races he participated in.  He is also a good friend of Lance Armstrong and says he should try the to break a record for going as fast as possible in an hour around a track (Eddy did that).   All the pro cycling magazines have Lance on them because he is the favorite and he can break a record (and tradition).  Here are a few  of them:  '),(20123,3208138,'02,July,2004','2004-07-02 00:00:00','   Graham Watson is making a book of his photographs called \"The Landscapes of Cycling\"  and can be learned more about  urlLink  here .         For another book photographed by Graham, click  urlLink here .        Graham Watson also has made coasters with his pictures on them.  '),(20124,3208138,'01,July,2004','2004-07-01 00:00:00','   The 2004 Tour de France is meant to be an on-the-edge-of-your-seat route.  It does not favor our Lance Armstrong, but is meant more for an all-climber type. It does not have a long time trial (TT) prior to the mountains, which Lance has in the past dominated, and in the team time trial (TTT), the maxium a team can lose is 2:30, no matter how bad they are so that they are able to get the yellow jersey.  Lance has a good team for the TTT and is the favorite to win as they won last year, and ONCE (now Liberty Seguros), their greatest rival, in it isn\'t as good as they used to be (they finished third in 2003).  In 1999 or 2000, Lance gained 4 minutes with the TTT on his greatest rival.  So for the overall classification, we have Lance Armstrong, the five-time Tour winner; Jan Ullrich, the five-time Tour runner-up (and won in \'97); Tyler Hamilton, who placed fourth last year with a broken collarbone, and Iban Mayo, the Basque that won on Alpe d\'Huez last year and has made life difficult for Lance in pre-Tour races both this year and last year.  Some other possibilities for podium places are Haimar Zubeldia, which came in fifth in front of Mayo last year, and is on his team, and Roberto Heras, which was on Lance\'s team starting in 2001.  From right to left, that\'s Iban Mayo, Lance Armstrong, Jan Ullrich, Jose Luis Rubiera, and Haimar Zubeldia.     Key stages are: the stage to La Mongie, the Alpe d\'Huez TT, Plateau de Beille, and the final TT in Besancon.  For the King of the Mountains competition (KOM), the favorite is Richard Virenque that won 6 times already and is aiming for a record-breaking 7th title.  Allesandro Petacchi should be a good rider in the sprinter\'s competition, which he won in the Giro d\'Italia along with a record-breaking eight stage wins.  '),(20125,3208138,'04,August,2004','2004-08-04 00:00:00','     '),(20126,3208138,'03,August,2004','2004-08-03 00:00:00','   Do you want to be able to watch the 2004 Vuelta a EspaÃ±a like you watched the Tour de France?   urlLink Sign this petition  and hope to make an influence on those fishers at OLN.  '),(20127,3208138,'03,August,2004','2004-08-03 00:00:00','   Steve had a great time in Hawaii. There was time for play and time for psychological stuff.  My friend Kirk Schneider invited me to fill in for an ill speaker. That allowed me to talk about my work with children. I was able to attend two workshops, one on psychotherapy with males and one on psychotherapy with couples. Both were helpful and validating. It is nice to be reminded that you know something and that others are interested in your sharing. I value these experiences very much. I enjoyed the time in the ocean and at the beach. Audrey asked what was my favorite part of the trip? I first thought of all the things we did and tried to decide what I enjoyed the most. What I realized later was that it was having time for such a trip. Having time for myself, Beth and I, or with family feels like it is in such short supply. Much less than I like. So having the time for myself, Beth and family was the best part.  '),(20128,3208138,'03,August,2004','2004-08-03 00:00:00','   Hilton Village from the Beach.   Dancers at the Polynesian Cultural Center.   Wind swept ladies at Pali Point overlook.   Visiting the Battleship Missouri.   Visiting the Dole Pineapple Plantation.  '),(20129,3208138,'03,August,2004','2004-08-03 00:00:00','   Here are a couple of pictures from our trip.   The ladies on arrival.   A view of a rainbow from Beth\'s and Steve\'s room.   Hunauma Bay where all four of us got wet and Steve, Beth, and Audrey snorkled.  '),(20130,3208138,'02,August,2004','2004-08-02 00:00:00','   oh yeah, i tried to put my picture in my profile, but failed. am i allowed to do this blog masters? and if so, how can i make that happen?  '),(20131,3208138,'01,August,2004','2004-08-01 00:00:00','   today, after work, i spied a three day old piece of meat in the fridge that was begging to be cut up, and joined with a tortilla to become a burrito. i gladly, guiltlessly obliged, and it was pleasing to the lord. however, later in the evening, glen and cindi (my landlords) returned home from puppy bootcamp with the apparent goal to consume leftover foods from the refridgerating unit. in a sad turn of events, it turns out that i had eaten glen\'s intended supper! the guiltless glee that had filled my tummy quickly turned to shameful sorrow. glen had to settle for a nasty frozen pizza instead of a steak dinner. for this, i am eternally sorry, from the bottom of my cold, black heart. i\'d like to buy you another london broil glen, if i may.  i like telling sam \"down\" after he\'s already decided to lay down. it makes me feel accomplished, and fulfilled.   i got a postcard today from my pen pal in hawaii. it was beautiful, and made me jealous, and proud all in the same instant. can\'t wait to hear more stories.  looks like you can download speeches from the d.n.c. for free at itunes music store. i\'m thinking of downloading them, and playing them when i\'ve already tried counting sheep at night, in order to induce sleep. (yawn)  peace  '),(20132,3208138,'01,August,2004','2004-08-01 00:00:00','   Sam II goes to his fourth puppy class tonight. I have been doing \"down\" with him all week with little to no success. Sniff, sniff. I tell him down and wave my hand appropriately and all that. He looks at me and doesn\'t move. So, I go to level two...pull his paw out and roll him gently to that same side. Then I get big results. He resists, pulls back, and nibbles my hand. I wrestle him to the ground and say \"Good dog.\" What?!! Yes, I\'ve battled him to the ground everyday this week for the sake of \"down.\"  Is Sam testing me? Am I doing something confusing to him? Does it normally take more than a week to \"get down?\" And dear, sweet Ashley. She\'s come on the scene a few times. Sits promptly next to me when I tell Sam to sit. Goes down with grace and beauty in perfect Olympic form when I tell Sam to do the \"down.\" Ugh. So he figures if she\'s got it, he\'s off the hook.  Glen, oddly enough, observes from a distance and just laughs. With me of course. Well, I leave for Minnesota on Wednesday. He\'ll have the bulk of the training time this week. We\'ll see how it goes. This week\'s command will probably be something easy like \"leap on the sofa\" or \"chomp on an avocado\" or \"look sincere\" or \"beg for an ice cube\" (his new treat of choice).  Livestrong. Laughstrong. Downstrong, please.  '),(20133,3208138,'01,August,2004','2004-08-01 00:00:00','   Yaroslav Popovych signed for the Discovery Channel team (US Postal next year) for 2005-2008.  \"I\'m really excited to get Yaroslav on the team,\" said team director Johan Bruyneel. \"He\'s a young rider who has shown his talent on numerous occasions. ... I\'m sure he still has a great margin of progression (to make), and I see him as a future contender for the Tour de France in a few years. It\'s great to have a big talent like him joining us. He really believes in our program.\"   \"Riding for the Discovery team in 2005 has been my ultimate dream,\" Popovych said. \"It\'s a team in which I believe in and a team where I\'m sure to receive the proper education from directors such as Johan Bruyneel and Dirk Demol to become a three-week stage racer.\"   The team is also trying to get American Levi Leipheimer to sign, but no results yet.     No OLN? OLN won\'t be showing the Vuelta a EspaÃ±a as much as usual.  They said, \"Our current plan is to air a one-hour highlight show of the Vuelta on October 27 at noon Eastern time.\"  Bummer.  At least we shouldn\'t miss Lance because of it because he has never ridden the Tour and the Vuelta the same year.  Lance\'s fan club website,  urlLink thepaceline.com , says that the interview continues today on OLN. I was wondering if the OLN people (Glen and Cindi) know about this at all and when it is.  '),(20134,3208138,'01,August,2004','2004-08-01 00:00:00','   The PWWCR blog welcomes Jeff Barneson who has responded to the invitation I sent for him to join our blog. I hope that he will weigh in on the Lance mania and other bicycling chatter here. AND, I hope he is inspired to produce his own brilliant blog.   I told another friend about the blog, he got inspired and  urlLink blogged about his recent trip to Cancun .   BlogStrong; LiveStrong  '),(20135,3208138,'01,August,2004','2004-08-01 00:00:00','   Stuart O\'Grady won the   urlLink HEW Cyclassics Cup race in Germany  and just beat last year\'s winner Paolo Bettini.  '),(20136,3655173,'22,June,2004','2004-06-22 00:00:00','   Timothy M. Sledd 7652 Patton Road Bloomington, IN 47401 (812) 361-5851 tsledd@indiana.edu  Bar Admissions: 	United States District Court for the Southern District of Indiana  	State of Indiana  Education: 	Indiana University School of Law-Indianapolis (2001-2003) 	Doctor of Jurisprudence, May 2003 (Cum Laude) 	Order of the Barristers 	9th in Evans Constitutional Law National Moot Court Competition 	Member of Criminal Law Association  	Indiana University Bloomington (1996-2000) 	Bachelor of Arts, May 2000  Relevant Experience: 	Intern/Law Clerk: Indiana Federal Community Defenders, Inc. (January 2002 â€“ August 2003) 	Sat second chair with Jim McKinley on Jerome Steele trial 	Drafted motions, memoranda of law, appellate briefs 	Researched federal and state cases, statutes, and guidelines 	Maintained client relations through frequent visits to the Marion County Jail             Attended both trials and appellate arguments  	Student Legal Intern: Criminal Defense Clinic, IU Law-Indianapolis 	(January 2003- May 2003) 	Counseled indigent clients concerning Indiana State D-felony charges 	Represented indigent clients at court hearings 	Successfully litigated a bench trial (Judgment on the Evidence) 	Negotiated plea agreements with prosecutors 	Investigated allegations by visiting clients, witnesses and crime scenes  	Volunteer Intern: Indianapolis Legal Aid Society (June 2001- December 2001) 	Aided staff attorneys in handling the civil cases of indigent clients  Skills: 	Legal research and writing: 	Proficient in both Westlaw and Lexis Nexus search engines 	Experienced in using the United States Sentencing Guidelines and its commentary 	 Language skills: Completed college requirements for Spanish in 2000	   References:  James McKinley, Indiana Federal Community Defenders, Inc. 111 Monument Circle, Suite 752, Indianapolis, Indiana 46204.  Novella Nedeff, Indiana University School of Law Indianapolis (Criminal Defense Clinic) 530 West New York Street, Indianapolis, Indiana 46202. 317-274-1949  Fran Hardy, Indiana University School of Law Indianapolis (Criminal Defense Clinic) 530 West New York Street, Indianapolis, Indiana 46202. 317-274-1911.  Mary Jane Norman, Attorney at Law, 4911 E 56th  Street, Indianapolis, Indiana 46220. 317-254-1443  John Floreancig, Director, Indianapolis Legal Aid Society, 615 North Alabama Street, Indianapolis, Indiana 46204.  '),(20137,3655173,'21,June,2004','2004-06-21 00:00:00','   As of the date of this post, my profile pic is of three dudes wearing interesting outfits standing with their arms crossed. Let me explain.  As I have discussed previously, I am an instructor at Indiana University. I teach Brazilian Jiu Jitsu as well as an academic course, Criminal Procedure. Brazilian Jiu Jitsu is unlike most martial arts which award technical expertise with a Black Belt. Technical (book smarts) expertise in Jiu Jitsu is realized during the early stages of the blue belt level. At this point, a student should be able to speak with authority on the history, benefits, and detriments of Brazilian jiu jitsu (BJJ).   An upper blue belt (I would say 3 and 4 stripers) would be a black belt by most Karate, Tae Kwon Do, and Judo standards if those standards were applied to BJJ. However, this is definitely not the case for BJJ. Belt level promotions are controlled by several factors. First, one must spend at least a minimal amount of time with an instructor capable of promoting. There are relatively few such instructors in the U.S. Next, one must show not only technical prowess, but practical ability! BJJ is an art that allows you to train \"all out\" and test your abilities to their extreme without imposing significant injury to your opponent/partner. Thus, unlike Eagle Claw Kung Fu, we as BJJ practitioners cannot hide behind the statement \"well if we really showed you how much we know, you would be dead.\" LOL  So back to the explanation. The man on the left in the pristene white uniform is my friend Kelly Wherley (a.k.a. Que Si, KCDubya, Wherldiggity). The man in the center with the black uniform and a very worn black belt with frayed white edges is Professor Caique. He is a BJJ legend! He is my black belt instructor and friend.  The guy in the blue uniform and blue belt with three white stripes is Me. I had just received my third stripe that day and Caique and I discussed my continued pursuit of my jiu-jitsu goals which include remaining with his team and instruction.  So, I don\'t mean to hate on other styles of Martial arts, they each address different aspects of what Bruce Lee entitled the 4 ranges of fighting, but I will proudly boast that BJJ is the toughest style to achieve black belt, technical, and practical mastery of, and that is why I love it so much.   Human Chess.  Tim  '),(20138,3655173,'16,June,2004','2004-06-16 00:00:00','   Ella delia...  She is in there and \"ripe.\" I keep telling people that the 9th month proves that \"a watched pot never boils.\"   I want to meet her, see her, hold her, and love her. She is going to change my life! Tick, Tock...  If she knew the attention she is getting even before she has arrived I am sure she would smile. There is love and support walking her around. But she is cooking and waiting until her moment.  Delia, or Ella, or Ella Delia is still future tense. But for how long?   Tim  '),(20139,3655173,'16,June,2004','2004-06-16 00:00:00','   YOUR training family   Posted by Tim on Monday, 17 May 2004, at 11:07 p.m.  Today I was reading an article by Royler Gracie (one of the legends of BJJ). He reiterated the importance of forming a quality training family within your jiu-jitsu club.   Let me explain how I view MY \"training family\":   Grandpa Caique: Offers the soundest jiu-jitsu wisdom though he is not physically present all the time.   Greg Lucas: Is like a father figure when it comes to jiu-jitsu. This guy has laid the smack down on me many times and punks me anytime I get cocky!   Chris Hayre: Also a father figure with more patience for my questions than Greg (plus he lets me think I \"earn\" positions on him :))   Kedar: Definitely an older brother(though years younger in age and pounds lighter in weight). Here is a guy who can throw new riddles my way that make me think. We battle hard, have different approaches to the game, but I have MAD respect for the man! Kedar may be the first relevant member of your jiu-jitsu family... he may be your father figure... if so, damn you are lucky! This guy is stellar.   John Hill: Spiritual mentor who helps me keep reality in focus.   Matt Stratta: When it comes to a jiu-jitsu family, this guy is my close brother. He is silky smooth with his technique and has a steele trap mind for the details. In the perfect world, he would be watching every class I teach to remind me to show the one technique I slipped on.   Jeremiah Ashe: I have mad love and respect for this brother. He has helped my game significantly. Laid the beatings down repeatedly, and has the humbleness and control over his ego that we should all aspire to mimic.   Steve Corenflos: Another brother that I feel a strong sense of loyalty toward. At 135lbs, this guy is a tough MF\'er! His game is going to get tough the more he trains. Oh, yeah, watch out, this guy is going to Lucas for training! That extra effort will reap him significant rewards!   Paul Hogan: I was ducking you bro! Hogan tough as nails. I am happy to have him in my training family. He has legitimately tapped me freqently. He makes me play my game tightly, and even then it is not enough. At 19 or 20 whichever it may be, he is a force to be reckoned with. I promise that I will reckon that force every chance I get because I get better each time.   Jim VanAtta: New member to the family, but valued nontheless. Jim is strong, wirery, and has great mental prowess when it comes to jiu-jitsu. I enjoy teaching him the moves because his follow up questions force me to rethink the game.   I could stop here and be a very fortunate jiu jitsu student! But to stop here would be to leave out some other very valuable members that I call upon to help me train.   Little brothers include: Will Buck, Ryan Kloman, Andrea Robertson (yep a brother), Lobato, Joe Sunderhaus, Ben Markley, Rebecca Remilard (another yep a brother), Julie Kedzie (though many of you have seen her whip my ass with gloves on!).   I know I have forgotten others, but I trained with Lucas tonight which means I took a beating an my mind is not all there. I apologize.   What I want you to do is find a familial makeup that works for you. Some of you will need to have more brother\'s than fathers, some of you will be comfortable having many fathers and few brothers. (Gender\'s can be flipped for the women!)   We are very fortunate to have a jiu-jitsu club that is large and inexpensive. Make the most of it!   Tim   P.S. DTM  '),(20140,3655173,'16,June,2004','2004-06-16 00:00:00','   Preparing for Battle   Posted by Tim on Wednesday, 2 June 2004, at 10:21 a.m.  Currently our club is in preparation for an upcoming tournament. Of course not all of the members of the club are going to compete.   We don\'t require it! Nor do we expect it!   The strength of our club is the fact that even those who are not going to be competing show up to practice to push those who will be competing. It has been awesome to see the turn out to club practices and the efforts put out by all.   Some of the guys are going to compete to gain exposure and experience and have a good time. This is an important opportunity that competing in practice does not give. Approaching a tournament with this mind set allows one to compete without being concerned about being in PERFECT condition, or feeling superior. One can go with the flow and meet great grapplers, watch awesome matches, and become inspired to work a new aspect of their game.   Others in the club are looking to continue their tournament improvements. Their objective is to confront a rival, place a little higher, or simply survive the first or second round. I emphasize to these participants that conditioning and mat time are crucial.   Finally, there are some in our club that are going to compete to Win. This is the toughest challenge! In order not to be \"bs-ing\" one\'s self, the competitor must train like a champion. They must train not only their body but their mind. There must be no question that when you step on the mat you are SURE you are going to win!   If you are not interested in competing, don\'t feel lost or left out. Come help us train! I promise that you will learn some valuable practical concepts by being around. Our competition team will be eternally grateful as well.   So, the date is set, June 26th. The divisions are relative mysteries, so focus on making yourself as prepared to wage the battle you want to wage at the tournament, whether it is just to get some mat time, improve your finish, or take the trophy!   Train Hard and I will see you at club.   Tim  '),(20141,3655173,'16,June,2004','2004-06-16 00:00:00','   Welcome to the Sledd Zone. This is my first blog, so I am going to be experimenting with various formats and topics. Mostly what I want to get out of this is a place for me to dump information that I am feeling, thinking, realizing, or finding useful.   Hopefully this will a great way for me to express myself so that I can come to know myself better and those who wish to know something about me will be able to feel a pulse.  So... this begins the Sledd Zone, a portal to the strange, normal, sick, and lost!  '),(20142,3655173,'11,July,2004','2004-07-11 00:00:00','   Check out  urlLink http://www.timsledd.com    It\'s not much, but a source for pictures of my daughter, family, and Jiu-jitsu.  Tim  '),(20143,3655173,'11,July,2004','2004-07-11 00:00:00','   An explanation of Brazilian Jiu-Jitsu: >(as posted on my web page two days ago)       Brazilian Jiu Jitsu (hereinafter BJJ) is a martial art that focuses on minimizing the damage done to combatants. Striking of any sort is de-emphasized due to the potential for self-injury as well as the likelihood that the party absorbing the strike will be injured. You may be asking, how can BJJ ever expect to work against a \'real life\' angry person who wants to knock your head off? Answer: we carefully measure the distance between our attacker and ourselves. Then we move in to clinch or take-down the attacker to take away the space necessary to employ an effective striking attack.        Once the clich has occurred, or a take down has happened, the object of BJJ is to gain a dominant position on the attacker so as to either 1. convince them to calm down and cease their attack, 2. employ a BJJ attack such as a joint lock or choke, or 3. utilize short range strikes to overcome the opponent.         BJJ is also a sport. Competitors start on their feet and seek to defeat their opponent by gaining a submission (opponent taps the body of the competitor signaling desire to stop the match) or by points which are awarded for positional dominance. Matches vary in length based on skill level, ranging from 5-10 minutes with no breaks.            BJJ is unique in that its moves do not require great strength or speed. (Unlike striking styles, wrestling, or judo). Instead, the more advanced fighters win using a mastery of leverage, baiting, and techniques that emphasize the weakness inherent with the human body. Submission holds focus on the weak joints of the human body, chokes focus on the easy availabilitly of blood sources to the brain to be impeded, and sweeps take advantage of the awkward centers of balance in different positions. Through a mastery of these concepts employed in various techniques, a very small BJJ student can quickly render unconscious or incapable of fighting, a much larger, stronger, and tougher attacker.            Many find the allure of BJJ to be the complex mental components necessary to excel. It has been called \'human chess.\' When two fighters are at the same skill level, the one who will win the match will be the one who is thinking more moves ahead of the adversary. Unlike wrestling where a match is ended based on body positioning (i.e. the opponent\'s back is flattened to the mat), in BJJ you must cause you opponent to submit... they must signal defeat. This can be quite a feat!  Tim 7/11/04  '),(20144,3655173,'11,July,2004','2004-07-11 00:00:00','   Wow, since my last post (June 22) there have been some eventful weeks. On June 26, I competed in the Hoosier Open Jiu Jitsu tournament held in Indianapolis, and put on by Caique and Greg Eldred of Indianapolis Jiu Jitsu. It was a very well run tournament with great competitors. The purple belt divisions saw some great battles as Angelo Popofski took on Mike \"Joker\" Aref of McVicker\'s Competition Team. These guys battled three times in very technical and competitive fights. Bernie Ash of Louisville\'s Caique Academy won his Purple Belt division.  I was able to win my division! The first match was a hard battle where I was on the defensive nearly the whole match being down 3-0 very early. I was able to secure a sweep in the last minute of the match and pass the half guard into mount to make the score 6-4.   My second match was against a very strong opponent. I attempted several of my best takedowns to no avail. Finally, he pulled guard. I was able to pass his butterfly guard and secure knee on belly from two different sides making a score of 7-0. At one point he was able to get to his knees and ahold of one of my legs. Just as I had suggested in club, the sprawl drill came through for me. I was able to fend off his attack and front one of my own. I attacked his collars looking for a choke. I tried to get both hooks in, but stayed in control with one. Time ran out with a score of 7-0 me!  I called Angie and got in the car. (She was due any minute)!  That next week we got to meet our sweet daughter! Let me tell you the story:  The doctor told Angie that Friday July 2nd would be the day we could have Angie if we wanted. Everything was ready, we were simply waiting for the contractions. Upon hearing this, Angie got excited and her normal aches and pains decreased. She openly prayed for natural initiation before Friday, but the light was at the end of the tunnel we were finally going to get to see her.  Friday morning Angie woke up earlier than I did, and organized our things. When I got up, I took care of the dogs, and we were ready to go. Upon arrival to the hospital, we both were nervous, but excited. We ascended to the Maternity Ward and met the nursing staff who promptly put us in \"room 5.\" It was a nice room with a couch, two chairs, the fantastic birthing (transformer bed) and a stool. The room itself had woodfloors, a large window, and wooden cabinets hiding all of the \"medical equipment\" and monitors.  The room was comfortable.  Angie was put on the typical monitors (One measuring duration and intensity of contractions while the other listened and measured Ella\'s heart). Almost immediately after being placed on the monitors, the nursed asked Angie if she was aware of the contractions she was having at that very minute. Angie did not.  On time, the Doctor came in and broke Angie\'s water. From that point to nearly 10 hours later at 5:46p.m. Angie worked to deliver Ella, everything from walking lap after lap of the ward, to intense labor, to about 45 minutes of timed pushing. The process is an absolute miracle (cliche but absolutely true)!  There is nothing like seeing the birth of your child. The first, verified, facts I learned about Ella were that she had lots of hair. In fact, her hair was \'born\' quite a while before her head followed.   Another cliche about having children is that they will change your life. I was skeptical, but again, the cliche is true. I realized how heroic it is to be a mother, from the birth process to feedings, women are awesome!   My priorties have shifted as well. I will drop anything to tend to Ella. (This is not characteristic of my steadfast belief in seeing through of planned activities). She is much more important. I am excited for the point when I have adjusted to her presence and can involve her in my activities, but for now, she is running the show.  So, I may post infrequently, but there is a reason. A fantasticly beautiful little squawk box named Ella Delia.  '),(20145,3480958,'28,May,2004','2004-05-28 00:00:00','    This is my first time blogging, so god only knows how this is going to turn out. Many of you may think, \"Well what made her start to blog?\"  Well the answer to that questions is simple I have a lot of self-absorbed friends who never want to here about my day or problems, and gosh darn it I want something to be about me me me me me for once.  I am a 24 year old, well almost next Sunday is my b-day, who is a Administrative Clerk for a local police department.  Don\'t get me wrong I am far from straight edged though.  I feel like I am Dr. Jekyl and Mrs. Hyde half the time.  I don\'t act my age half the time, but like a teenager instead, even though my maturity level is at least the one of a normal 24 year old female.  LOL!!!  I just love to party and have fun and I love my men young.  At least 18 though.  LOL!!! Not trying to catch a charge you know.    My life is pretty simple I drag my self out of bed in the morning, put on my Dr. Jekyl hat and go to work where I am professional and very pleasant.  Then once 430 comes around off comes the Dr. Jekyl hat and here comes Mrs. Hyde.  No one inside of work knows how I am outside of work, and I would like to keep it that way.  Well I hope that I can keep this \"Diary\" going and hope that someone will find it interesting.  It will probably be filled with several sporatic trist with my \"friends\" and probably some complaining about work too, never-the-less I hope you will find it worth reading.  Til next time. '),(20146,3480958,'22,June,2004','2004-06-22 00:00:00','    I can\'t believe it I have found him.  I believe that everyone has a love that they have lost that they regret loosing. W.B. is mine.  He was the best guy to ever walk in my life. He was a perfect gent. but could still argue with me when I needed a good one.  He was also gorgeous, could dance, and even make a mime laugh out loud.  Of course I was young and stupid and screwed the relationship all to hell.  Nothing new for me.  Anyway we talked all the time afterwards, but then he changed jobs in the military and had to go across the US for school.  We managed to loose touch  for almost 4 years until Friday when my brother got his e-mail address from a website online for military people to find old military friends.  Well I e-mailed him and he e-mailed me back.  Nothin long.  It was short and sweet, but all the memories came rushing back, and I almost started to cry like a little bitch, but I didn\'t.  I wrote him a little to catch him up on my life, and asked him to do the same.  I guess I will have to wait and see what he has been up to.  But to know that I will now have the opportunity to show him how much of a better person I have become is worth the wait.  I think he would be happy to know that most of it is because of him. '),(20147,3480958,'21,June,2004','2004-06-21 00:00:00','    I swear if I could change the name of my blog now it would have to be Life with Kooki (Confessions of a Twenty Something Drama Queen).  Life seems to get more and more interesting yet mind racking everyday since I have been seeing R.T.  You know first of all he has way too much baby momma drama, and now she is trying to make it worse.  Telling everyone that I am pregnant and don\'t know who the father is and stuff.  I don\'t mind the talking b/c I know that it isn\'t true it is just aggrivating that someone would go through so much to try to hurt someone else.  She is even threatening to call my work and tell them that I sell drugs and that I am an addict. (By the way I work for a police department.)  I swear it just makes me want to grab her by her hair and beat her head into the concrete, but then I think of how much I do like my job and how much she is not worth it.  I understand that this is all b/c we can\'t stand each other and she doesn\'t want me around her daughter or her baby\'s daddy, but she is with my ex-boyfriend and has been with him for over a year and a half, so why don\'t she just be with him and be happy and leave R.T. alone to move on with his life with me or with anyone else for that matter.  UUUUGGGG!!!! Girls piss me off so bad. '),(20148,3480958,'16,June,2004','2004-06-16 00:00:00','    I am in such a mood today.  I don\'t know why.  Everything just irritates the shit out of me today.  I think I just want to scream at someone until I am blue in the face.  Maybe that would feel better.  Then again that might also get me fired so let\'s not do that and say we did.  Maybe I just need a little good loving.  LOL!!! I think I might try that tonight, even if I have to give it to myself.  LOL!!!!Hope the day gets better, if not at least it can\'t get any worst. (Knock on wood) '),(20149,3480958,'15,June,2004','2004-06-15 00:00:00','    Ok, I am not the fall in love too easily type, used to be but not any longer. Well you know this young guy R.T. I have been messing with right, well I am starting to really like this boy.  Not fall in love, but when I see him no matter how I am feeling about my day I automatically smile. Look just sitting here now thinking about seeing him is making me smile.  Last night was the kicker though.  We was laying in bed going to sleep just talking and we were talking about how he can never keep his hands to himself.  (He is always picking on me poking me with his finger tickling me and whatever.)  Well anyway I ask him if it is b/c he has a crush on me, you know when you were younger you always picked on your crush b/c you were too scared to tell them that you liked them.  He tells me he wouldn\'t call it a crush it is more like a Kooki fetish.  Well of course at that point I told him to come and get some Kookies to satisfy his fetish. Then it was the most incredible he has ever given to me. It wasn\'t the hard core fucking that I am used to.  At first it was sweet and passionate until we hit our climax then it was hard core fucking all over again.  I don\'t know about this though.  I hate having the feeling of helplessness and no control that a crush does to you.  I like being in control of my feelings and thoughts, and I like not having to worry about checking in, but I also love to have someone to rely on and always be there.  Well I am still going to take this one day at a time, and I am not going to get excited about this or get all attached.  I don\'t think that anything big will come out of this, but if it does I won\'t be disappointed. '),(20150,3480958,'10,June,2004','2004-06-10 00:00:00','    I hate funerals they always make you think about death and the what if\'s.  My grandfather\'s funeral was yesterday and you know with President Reagan\'s I have been filled with the what if\'s for days now and that is not good. I shouldn\'t think about these things, I am only 24 you know.  Then I start thinking well you never know.  Why is it that when you are subjected to something like a friend or relative passing you tend to go through these motions. '),(20151,3480958,'09,June,2004','2004-06-09 00:00:00','    Nothing interesting last night, nor so far today.  I have a funeral to go to and don\'t really feel like blogging.  Maybe tomorrow. '),(20152,3480958,'08,June,2004','2004-06-08 00:00:00','    Well Well weekend re-cap.  Friday night- Well me and a bunch of friend went out to a local bar/club.  We had a blast.  Between G showing all the girls a picture of his penis to everyone throwing shots down my throat I had a fab time.  Oh yeah and when I got home had a even more fab time.  That\'s right I found me some b-day bootie.  My god was it ever so good too.  It was a long time coming.  Me and A.T. always talk shit to each other but nothing ever cam out of it until Friday.  God was it ever worth the wait.  LOL!!!  Saturday- Hung over, very hung over.  I felt like shit all day.  I always found out about President Reagan and cried about that too. Didn\'t help the hang over at all.  Sunday- My b-day.  Sat around and was extreamly lazy, then went to go see my little sister for a bit.  Weather was crappy, so I really didn\'t do much but veg out.  Monday- Slept in, but it was bright and sunny so I got up and got out the house.  I went and saw my sis-in-law and played with her horsies.  Then went and took pics for the photo contest.  This month the catagory is farm building, and that is not hard to find around here.    Today is good so far.  Not compaints. Let\'s just hope it stays that way. '),(20153,3480958,'04,June,2004','2004-06-04 00:00:00','    Every guy that I have ever met always say that every girl that they have been involved with is psyco.  I always say that there is always a reason behind a woman\'s psyco habits.  You know a cheating man, a chemical inbalance, or in my case family.  LOL!!! My parents separated in February and I moved home to help take care of my father.  Not that he can not take care of himself, it is just that he is a truck driver and he would never be able to do his job and take care of the house, so since he was the one who took care of me for years and years I felt that it was my turn.  BIG MISTAKE!!! They now drive me crazy.  Sometimes I want to lock myself in a room and not speak to either of them.  Mom feels like I side with Dad all the time, and Dad feels like I side with Mom all the time, when in actuality I side with no one.  I try to stay out of thier problems as much as possible.  So, guys before you start jumping to conclusions and state that a girl is psyco or is a drama queen take a little time to find out.  They just might have an issue that you might possibly be able to help them with, or then again they just might really be psyco.  My godson graduates from pre-school tonight.  It makes me feel so old.  It seems like yesterday I was changing his shitty diapers and this fall he will be starting \"big kids\" school.  I am also going to go out and shake my ass tonight, and as mentioned before I am hopefully going to find myself a b-day present.  Have all the details for you on Tuesday.  I had to take Monday off.  You know to recover and all from the long weekend. '),(20154,3480958,'03,June,2004','2004-06-03 00:00:00','    Why do people feel that when there lives are not interesting enough they have to talk about other people\'s, or should I say invent stories about other people\'s lives.  I talked to E.Y. yesterday, you know the yummy yum yum from T.C.\'s wedding.  My god how could I ever forget that one.  My god my head was still spinning and my toes were still curling like three weeks afterwards, ok that is a bit of an exaggeration, but you get the point.  Anyway he proceeds to tell me that the reason he hasn\'t called is b/c B.C. told him that I was saying some not so nice things about his manhood and sexual abilities.  I spent forever trying to convince him that I never said any of the stuff he thought I did.  UUUUGGGGGG!!!! That stuff pisses me off so bad.  Other than that all in all if was a good night.  Should have went to bed earlier instead of staying up and playing video games with my best friends and her husband.  I am paying for it today.  I am so tired.  Good day at work so far.  Have about two hours to go and then I am out.  I have my god son tonight until 7:30 it should be fun. '),(20155,3480958,'02,June,2004','2004-06-02 00:00:00','    Well lets see yesterday was pretty good.  But, R.B. the Town Administrator is back today.  WOO!!! HOO!!! I am so excited. This place is so boring without anyone to talk to.  I have a dilemma today.  I have been thinking that it is time for me to cut this shit off with R.T.  Not that he is not a great guy, but for some odd reason I can see myself falling in love with this boy.  He is great in and out of bed.  LOL!!! I can\'t fall her someone who feels that the more girls you have the better.  I would rather wait until I can find a faithful one.  Until then I have my friends.  The closer it gets to this weekend the more excited I get.  I can\'t wait for Friday.  I has been awhile since I have been out to shake my tail feathers.  I need to.  A few drinks aren\'t going to hurt neither.  And who knows it might be like last year and I just might find myself something to take home for my b-day present to me.  :-)  That could possibly help me with this whole falling for R.T. thing too.  Kill two birds with one stone that has always been my moto.  LOL!!! I talked to T.D. yesterday I think maybe just maybe me and him can start picking up the pieces of our friendship and starting all over.  I really miss him.  See that is why you never have sex with friends, one almost always falls in love while the other doesn\'t.  That is what happened in our case, and I wasn\'t the one falling.  Imagine that!  LOL!!! I kinda like being a lone ranger.  You have no one to answer to and desides I still have a long way to go and there are plenty out there just waiting for me.  LOL!!!  Hey, someone has to think positive might as well be me. '),(20156,3480958,'01,June,2004','2004-06-01 00:00:00','    My long weekend in recap.  Friday- Nothing exciting.  I got stuck at work late, not enough hours in a day to get everything done.  Oh well.  Thank god for overtime pay. I had dinner with my Grammy and Dad and went to chill with my god son and best friend for a bit.   Saturday- R.T. came over first thing (Grammy had left to go back home)and my god was that fabulous.  I had waited three weeks, my head was about to explode.  You know you can only handle your own needs for so long before you need a little help.  Other than that, another unexciting day.  Matter of fact, I think that my whole weekend was unexciting.  Sunday- More time with R.T. but no lovin this time.  LOL!!!!  Then my brother and my new sister in law came over for a little BBQ with me and my Dad.  They just got back from there honeymoon and brought pics from there trip.  Can you believe that my brother took there lawn gnome with them on there honeymoon.  I know you have seen the Travelocity commercial with the Roming Gnome.  I so need to write them and say \"see what you have started you actually have people taking their lawn gnomes on vacation with them.\"  LOL!!! It was pretty funny though.    Monday- Memorial Day. Yuck! It rained all day.  I was so hoping to go to the beach.  Instead I stayed home and made it a lazy care for Kooki day.  I gave myself a manicure and peticure, and watched the Law & Order marathon on TV.  I love Law & Order.    So, in a nutshell.  Nice relaxing weekend to rest up for next weekend.  WOO!!! HOO!!! Next weekend is my b-day and it is going to be party, party, party. '),(20157,3519292,'23,June,2004','2004-06-23 00:00:00','   I hate people who bitch about thier problems all the time.I just hate that.No one cares what you think.Same thing for me. You thing anyone cares what I think?HELL NO!Anyway,I\'m at the store looking at video games and cd\'s and shit like that.Then some lady start\'s bitching at ME because I didn\'t stop HER SON from doing something. She\'s like\"WHY DIDNT YOU STOP MY SON FROM THROWING THIS??????HE IS A VERY TROUBLED YOUNG BOY!!!!!HE MISBEHAVES ALL THE TIME AND HIS DAD DOSEN\'T PAY CHILD SUPPORT!!!!\" When she stops bitching I say\"He\'s not my problem so i\'m not worrying about it.If you were a good mom you would take care of him and not bitch to other people about your problems.So stop complaining and shut the hell up.\" She goes away then 3 minutes later,wal mart people tell me to leave the store and I say no.So they go away.Nearly got banned all because of her and her damn kid.  The lesson:Shut the hell up and don\'t complain.  '),(20158,3519292,'04,June,2004','2004-06-04 00:00:00','   What the hell is up with this n-gage qd thing?First of all,what the fuck kinda name is that?Secondly,why make another craptastic hanheld thing.  For the lucky who have never heard of n-gage,it\'s suppost to be some hanheld system thing that doubles as a cell phone.That\'s pretty gay considering the damn thing costs $250.I have a much better idea.Go to walmart and buy a cellphone.Then go to a video games store and buy a gameboy advance sp. The n-gage had a great take off.Then that great take off went straight to hell and now it sucks ass.Graphics suck ass.Sound sucks ass.N-gage sucks ass.  What the fuck is up with the gaytastic advertizment for this thing?Ok I will guide you through if you have never seen it before(only the lucky few).A ninja is in a laundromat cleaning his cloths.His machine gets stuch.He punches it.STOP!If I was a ninja,I wouldn\'t punch it,I would whip out my badass dragon sword and slice the thing into cubes. Anyway,on with the commercial.Some chick comes over and unjams the machine he is using.He then says \"unstuck\".He is about to put his pussy ninja panties in there when he spys another ninja.A PINK ninja.A PINK DUDE ninja.He drops his panties and draws his sword.So does the PINK DUDE ninja.Then there\'s some shit about the n-gage QD.End commercial.Overall,a pretty gay ass commercial for a pretty gay ass piece of hardware.  '),(20159,3519292,'02,June,2004','2004-06-02 00:00:00','   Whats with all this church crap?People getting raped.Your \"father\" sending the collection plate around.That just pisses me the hell off. Jesus will still love you even if you don\'t give him money.So stop paying the goddamn collection every sunday.And riddle me this.Why is church on a sunday?Football games are on sunday. I would rather be really bored for 4 hours on a monday then football sunday.   A note to all priests:You suck ass.Stop touching little kids.Go out to a strip club and get some real poon like everyone else does.You make me sick.When you die, I hope you go straight to hell.AMEN!  '),(20160,3519292,'01,June,2004','2004-06-01 00:00:00','   I am getting pissed about this anti-smoking truth thing.Every goddamn time I turn on the tv, I see\"If you smoke,you could die.A lot of people do.Truth.com\"It\'s really pissing me off. I think we know the fucking dangers.You can stop advertizing it now. I see stickers of that cocksucking shit all over my school.Please stop the damn advertizing.EVERYBODY GODDAMN KNOWS ABOUT IT HIPPIES!!!  '),(20161,3519292,'01,June,2004','2004-06-01 00:00:00','   Why do people use so many excuses?Come out and speak the truth. Hiding it is not going to do anything.That\'s one of the main reason\'s  why I made this blog.The other one is because I was bored.I don\'t hide  the truth.I don\'t use excuses to hide true facts.Don\'t be a little bitch ass homo and say shit like\"I didn\'t mean to\" or \"It was an accident.\"That won\'t make a diffrence. If you kill someone, It won\'t make a goddamn diffrence if it was on  accident or pourpose.They will still arrest your ass no matter what.  '),(20162,3519292,'01,June,2004','2004-06-01 00:00:00','   I\'m posting a list of things that piss me off.I\'m going 10 to 1  because I want to.Disagree?FUCK OFF!!!!   10).Teachers  9).Jocks  8).Geeks  7).People complaining about thier problems  6).Punk Music  5).Goths  4).Hippies  3).Jehovis whitnesse\'s  2).Democrats  And the number one thing that pisses me off the most is...  1).All of the above  Yes I can post all of the above on my list.Its MY list.Not yours.  '),(20163,3519292,'01,June,2004','2004-06-01 00:00:00','   I hate hippies.All about peace and shit.The only peace I wanna show hippies is a 9MM right in thier face.They piss me off.Everytime I open my inbox I get tons of hippie shit.\"Don\'t eat meat\". \"Your killing Mother Earth\".\"Be kind to animals\".\"Join PETA\".It just  pisses me off.All hippies should be stabbed in the chest with an old  rusty dagger and be forced to eat meat.Then be pumped full of lead with an AK-47.  '),(20164,3519292,'01,June,2004','2004-06-01 00:00:00','   Country sucks ass.It\'s all the same.Just people singing about thier problems.Pisses me off.Why waste everyones time?Stop singing country. Be a use to scoceity.Flip burgers.Mow lawns.Be a stock boy. Don\'t be a pussy singing about your problems. No one wants to hear them so shut the hell up.  '),(20165,3519292,'01,June,2004','2004-06-01 00:00:00','   Don\'t be a retard and obsess about a movie.I hate people who do that. Wow star wars rocked.You don\'t need to make a damn religion out of a  movie.Same thing for those star trek assholes.I swear to god I\'m  walking down the street and these star trek retards dressed in these \"Star trek fan\" shirts come walking down the street.These kids have  pimples and fat everywhere.I mean come on i\'m 14 and I could kick  thier little fat star trek asses up and down the street.Don\'t be  gay and join some D and D cult either.That\'s even worse.Board game\'s  suck ass.  '),(20166,3519292,'29,June,2004','2004-06-29 00:00:00','   refer to this  urlLink thread at g4techtv.   THe japanese officially rule hardcore now. First they breed Shigeru Miyomato and Hideo Kojima and the ninja\'s and samurai now it\'s murderous icons.   In the words of Asmoron I     What the fuck is wrong with the japanese man. THe girl murdered this person and she gets her own anime and manga. So it goes to show that if you brutally massacre a person you get your own action toy in  mcdonald\'s happy meal. keep that in mind when the next time you decide to kill someone.   I\'m looking at you trig. '),(20167,3519292,'27,June,2004','2004-06-27 00:00:00','   Just another joke to take up space because we  never get anything else new on this page.     Ancient Chinese Torture    A young man was lost wandering in a forest, when he came upon a small house. He knocked on the door and was greeted by an ancient Chinese man with a long, grey beard. \"I\'m lost,\" said the man. \"Can you put me up for the night?\"  \"Certainly,\" the Chinese man said, \"but on one condition. If you so much as lay a finger on my daughter, I will inflict upon you the three worst Chinese tortures known to man.\"   \"Ok,\" said the man, thinking that the daughter must be pretty old as well, and entered the house.   Before dinner, the daughter came down the stairs. She was young,beautiful, and had a fantastic figure. She was obviously attracted to the young man since she couldn\'t keep her eyes off him during the meal. Remembering the old man\'s warning, he ignored her and went up to bed alone. But during he night, he could bear it no longer, and sneaked into her room for a night of passion. He was careful to keep everything quiet so the old man wouldn\'t hear. Near dawn he crept back to his room, exhausted, but happy.   He woke to feel a pressure on his chest. Opening his eyes he saw a large rock on his chest with a note on it that read, \"Chinese Torture 1: Large rock on chest.\" \"Well, that\'s pretty crappy,\" he thought. \"If that\'s the best the old man can do then I don\'t have much to worry about.\" He picked the boulder up, walked over to the window and threw the boulder out. As he did so he noticed another note on it that read: \"Chinese Torture 2: Rock tied to left testicle.\" In a panic he glanced down and saw the rope that was already getting close to the end. Figuring that a few broken bones was better than castration, he jumped out of the window after the boulder. As he plummeted downward he saw a large sign on the ground that read, \"Chinese Torture 3: Right testicle tied to bedpost.\" '),(20168,3519292,'25,June,2004','2004-06-25 00:00:00','   I think it\'s time I can start a chain of joke threads like this every week. One week that joke, this week this joke. hope ya get it   10 Things In Golf That Sound Dirty    1. Look at the size of his putter.  2. Oh, dang, my shaft\'s all bent.  3. You really wacked the hell out of that sucker.  4. After 18 holes I can barely walk.  5. My hands are so sweaty I can\'t get a good grip.  6. Lift your head and spread your legs.  7. You have a nice stroke, but your follow through leaves a lot to be desired.  8. Just turn your back and drop it.  9. Hold up. I\'ve got to wash my balls.  10. Damn, I missed the hole again. '),(20169,3519292,'24,June,2004','2004-06-24 00:00:00','   What the hell is up with this n-gage qd thing?First of all,what the fuck kinda name is that?Secondly,why make another craptastic hanheld thing.  For the lucky who have never heard of n-gage,it\'s suppost to be some handheld system thing that doubles as a cell phone.That\'s pretty gay considering the damn thing costs $250.I have a much better idea.Go to walmart and buy a cellphone.Then go to a video games store and buy a gameboy advance sp. The n-gage had a great take off.Then that great take off went straight to hell and now it sucks ass.Graphics suck ass.Sound sucks ass.N-gage sucks ass.  What the fuck is up with the gaytastic advertizment for this thing?Ok I will guide you through if you have never seen it before(only the lucky few).A ninja is in a laundromat cleaning his cloths.His machine gets stuch.He punches it.STOP!If I was a ninja,I wouldn\'t punch it,I would whip out my badass dragon sword and slice the thing into cubes. Anyway,on with the commercial.Some chick comes over and unjams the machine he is using.He then says \"unstuck\".He is about to put his pussy ninja panties in there when he spys another ninja.A PINK ninja.A PINK DUDE ninja.He drops his panties and draws his sword.So does the PINK DUDE ninja.Then there\'s some shit about the n-gage QD.End commercial.Overall,a pretty gay ass commercial for a pretty gay ass piece of hardware. '),(20170,3519292,'24,June,2004','2004-06-24 00:00:00','   Here\'s a rant about something in the mall today.   Ok I\'m just strolling along the stores to find something to waste my 175 dollars of b-day money, and then i see Hot topic. As most of you know (especially Tomcat here who\'s already had one rant about it) the place has gone to hell. Literally. The fucking entrance was a goddamned cave-looking entrance. 1600s anyone? No? you suck then....Well anywho, I just stand there in front of the entrance, gaping at the shear stupidity of what what it was, Evanescence-wannabes strolling by with chains that could make their boobs sag like an old lady\'s ass. Not that\'d I\'d know of course.......stop laughing.   Well one of the store managers sees me and tells me to move along because i was blocking the customers. there was like five feet left of space in the opening, just enough for some lady with a sign on her back saying \"I was Godzilla\'s ass in the movie\" and the nearest gothic chick was like 50 feet away. I\'m like \"What customers? Your store sucks! I mean, i wear black jackets and pants, but come on! Have you actually shoved that many punk cds up your ass that it reaches your brain?!\" Ok maybe so it was more of a remix version of that. She still got the point and sicked security on me. what security you ask? Gay dudes with ugly sticks that\'s what. they chased me about halfway across the wall until I lost them in JCpenny\'s. The only regret i had then was not taking some perfume and spraying their eyes. At least i could outrun two gay dudes who probably screw every fucking night. probably just getting old. Oh well. I\'m pretty fast anyway. Outran the entire soccer team once. But screw that. I need to think a little more about that. Chow '),(20171,3519292,'23,June,2004','2004-06-23 00:00:00','   I\'m currently working on a much more extensive rant, the details of which I won\'t discuss further until I post it, but since we are currently experiencing an insufficient supply of rants, I thought I\'d post one before I go to bed.  I can\'t stand it when people who are not educated on a given subject profess their extreme viewpoints as if they had actually taken the time to develop the ridiculous opinions they feel the need to express. They ramble on without regard for attempts by those with actual knowledge of the subject to point out the flaws in their erroneous beliefs, and refuse to acknowledge any reasonable argument against themselves. I\'m sorry but \"STFU\" and \"NUH UH\" are not adequate responses when confronted with major design flaws in one\'s theory.  If I am ever out of my depth in a certain topic or discussion, which does not happen often, I prefer to listen to those who are more well informed than myself and ask question to better my own understanding. I choose not to accost those I should be learning from with incoherent babble, but to conduct myself in a scientific demeanor, so that I might eloquently express myself to some tragically unversed fool in the future.  In closing, Please, you mighty offenders of civilized discourse, stop pretending like you know everything. You are annoying those of us that do. '),(20172,3519292,'23,June,2004','2004-06-23 00:00:00','   This rant is, well, about our ranting. Why the hell don\'t we have any yet? We have 5 FUCKING MEMBERS. Who do we need to get a rant?! Michael jackson in a tutu? Jerry Springer doing Britney Spears? Ben Affleck and matt damon?! Move your asses in gear people!!   thank you for your time '),(20173,3519292,'23,June,2004','2004-06-23 00:00:00','   The team is coming along great.   Actually it sucks major ass. I got two contributors and a whole bunch of people who email accounts won\'t let them join. This is totally fucked up. '),(20174,3519292,'23,June,2004','2004-06-23 00:00:00','   Yeah no rants on this yet. This is purely advertisement.   If you\'d like to join the rant team email me. '),(20175,3519292,'11,July,2004','2004-07-11 00:00:00','   Who the fuck ever thought this up?! Some desperate 19 year old uber-nerd who couldn\'t get laid if the world was on the line?! It was Bill gates wasn\'t it?! That sick fuck has always been conspiring crap! Look what he did to his house! he turned it into the first fully computerized home in the world! He doesn\'t even have to make his own food! It fucking cooks its own ass! Damn you Mr. gates! DAMN YOU I SAY!!  Back to the sick asshats who cyber-screw. YOU. ARE. SHIT. HEADS. Can\'t even get laid and they\'re over 18.......sick bastards. You should all burn in hell. I mean seriously! You all fuckin just write your screwing by roleplaying! I can  ALMOST  understand it when 13 year olds do it, since they can\'t get screwed even if they were the jessica Simpson of jr. high, but that just makes them REALLY fuckin perverted. Horny fucks......shame on you. Shame on you all! Losing your virginity to your fuckin comp...not right. basically, BURN IN HELL YOU SICK FUCKS. And if anyone who is reading this does cyber, shut up and take it up the tailpipe! Take your whiney ass remarks, and shove em up your ass. Then again, you\'ll probably use that in your next cyber-screw anyway.   Moral: Don\'t stick your dick in the fuckin :D drive or whatever, you sick bastard '),(20176,3519292,'10,July,2004','2004-07-10 00:00:00','   I usually try to avoid out and out ranting, but I\'m really mad, so you\'ll forgive me if I lack my usual eloquence.  Lowe\'s is the enemy. I enter the usually pretty crappily run hardware store expecting it to be the sanctuary of it has been in the past. This was usually a place I could avoid the bullshit the rest of the world feeds us on a daily basis, even if it was poorly managed, difficult to navigate, and usually misrepresenting its products one way or another.  Today however, something was different. Upon entering the store, I notice the usual displays set up in the main isle. A few how to books, a couple of specials, not thing much out of the ordinary. Just then, I see something so horrible, so disgusting and vile, I shutter to mention it. \"DECORATING WITH LIZZIE MCGUIRE!\" Infact, it was an entire series of home improvent and craft books from Lizzie right in the middle of my beloved saws and drills. WTF KINDA 9 YEAR OLD GIRL IS GOING TO REMODEL HER OWN ROOM! Licensing has gone too far, and Disney, once a respectable company from my youth, has become PURE EVIL!  Tell me now, WHO COULD POSSIBLY THINK THAT ANYTHING LIZZIE MCGUIRE BELONGS IN THE MIDDLE OF MY TOOL WORLD? In the future, you stupid store managing assholes do your jobs and keep that shit in the KID\'S DECORATAING SECTION if you must allow into my store at all.   LOWE\'S SUCKS ASS! '),(20177,3519292,'10,July,2004','2004-07-10 00:00:00','   At g4ttv there are these asshats who think that if you join at a certain time i.e.; 2002, it makes you cool some how. What the fuck is up with that? Just because you joined seven months before a person it some how makes you more superior than them? You know what fuck them and fuck people who think like them.   I remeber at techtv no n00b got made fun ragged on unless they totally deserved it. That\'s why I liked techtv everyone practially knew everyone and really respected eachother post count or no post count.   I admit I do that myself sometimes but most of the time I\'m joking. So the next time you rag on someone because they joined later than you do us all a favor and kill yourself.....yes I meant that literally. '),(20178,3519292,'09,July,2004','2004-07-09 00:00:00','   Ladies and gentleman of the blogging community, popular music is in dire need of you help. Even as record sales continue to rise, artistic expression and even involvement decrease with every Britney and N Sync clone forced into collective CD collection. Of course, music is about fun, and everyone has a right to listen to whatever they desire, but as these cookie cutter starlets achieve the bulk of our interest and music dollar an ever growing number of musicians making music from the heart are stymied and left in the shadows. Children force fed MTV\'s definition of music will never know the joy of listening to a real masterpiece that transcends mere trend to become a anthem for any generation. I plead with you, take an interest in music and pursue it. Listen to Jazz, Blues, Classical, and anything you like, but make sure you decide what YOU listen to. Don\'t let the record companies looking to make maximum profit from minimal effort decide for you. '),(20179,3519292,'02,July,2004','2004-07-02 00:00:00','   I just bought tickets to what I\'m sure will be a very crowded viewing of \"Spiderman 2\" tonight. I\'m sure some hick will think he\'s the only person in the building and offer his brilliant commentary on every painstaking moment of the film. In hopes that perhaps, that very redneck somehow has access to the internet and is viewing this blog right now, I would just like to remind everyone that talking, taking phone calls, and otherwise annoying your fellow movie enthusiasts is considered a faux pau.  The last time I was at the show, a family of three with no shoes walked in half way through. They sent their child of four years old to ask for candy from others, and then proceeded to call friends to brag about skipping out on the movie they payed to see and watching this one for free. This thing isn\'t common in my area, and I was rather surprised. If you have ever done this, please, refrain from entering public in the future for any reason and form an equestrian society in the woods behind you house. '),(20180,3692874,'30,June,2004','2004-06-30 00:00:00','    Hey hey... (yesterday 29.06.2004)went out with Vincent and Hong Guan to study... can u believe it? Hong Guan go study!!! hahaha... 1st time ah?well... hey... i recieved calls from my dad saying Bedok,Semei and Blah blah blah got blackout... well... i am very tired nowadays you all(readers) shld know the reason... Study la!! hahaha... Eating \"Ligo California golden seedless Raisins\"!!! it is my fav food when i was... hmmm... 3years old? hey! i am sitting infront of my computer eating Raisins and drinking Fruit Tree Fresh Orange...so sad right? suppose to go \"Region Hotel\" tonight de... my cousin wedding dinner!!but i not going... because is somehow taking over my sleeping time... cause i need to sleep and study.... cant be wakinging up for so long... somemore... dinner only ma!! hotel only.. anytime also can go... but \"O\" lvl different leh.... hahahah  urlLink    Ligo Raisins&nbsp; urlLink '),(20181,3692874,'28,June,2004','2004-06-28 00:00:00','    School started... woke up at 6am, continue to sleep till my father woke me up again at 6.20am... had a quick bath,brush teeth, stlye my hair(anyhow cause no time) and change to my blue suit(EVSS uniform)and the time is 6.50, my eyes wide open thinking, \"I am still in Bedok!\" hahaha... read newspaper as Silence Reading... after that we went to lessons and hey! our new SS teacher is not bad... willing to give us extra lesson on private groups!!! well.. after tat i went back to my seat... and i felt something not right!! my butt is like.. \"buring\" slowly... and my butt is feeling wet... i touched my chair and is wet... smell starts gettting heavier... (of Axe Oil) i jumped up immediatly... and hop around the class!! Oh... my poor butt... but then.. i must really thank tat person.. my ass is cooling!! (sprit ice feeling)  urlLink    Axe Oil&nbsp; urlLink '),(20182,3692874,'27,June,2004','2004-06-27 00:00:00','    Haiz... back from Elmo\'s house... 12.37pm... my mum was sitting on the Sofa... and looking emotionless... i was like... so confused... instead of scolding me... she kept quiet... i tried to break the ice twice.. but she simply just ignore me... ??? what happen? i told my dad to tell her that i will be home late... then what is keeping her? is she worried? most properbly... but hey... i nv went out fighting(cant fight),stealing(stop long ago) or mixing with bad companies(no bad companies what?)... well.. what ever she is thinking.. i hope she will be alright in a day or two... i hope it dun go on...  Ok... from now onwards... i dun think i can write anymore... maybe once a week or so... but not often... due to some circumstances... i must maake some sacrifices... hmm... Downloaded the new media player... player version 9(so called) and i saw player 10!! so fast... but i didnt download that.. cause is a beta version(testing) had a great day sleeping.. now is almost 6pm... i think i am going to study... media player playing \"Everything will flow\" by suede   urlLink    Suede&nbsp; urlLink '),(20183,3692874,'25,June,2004','2004-06-25 00:00:00','    Whee... meet up at 201 mac and den study wit lin and YT... and vincent suppose to meet us at 8pm... and he came at 9.45pm... how early!!! hahhahaha.... so... lin went off and left me and vincent... and soon after that i was like wanting to pee... hahha... so we took bus 38 and off we go to small mac there to meet sabby,Gundam(Song lim) and elmo(Jun Quan) to watch movie... end up having a mistake.. they thought we came for fun.. so they didnt buy our tickets.... but hey... is better not... cause it will be e 3rd time to watch \"around the world in 80days\" so we all went to watch \"The Day After Tml\" instead... the show is freezing cold and so are we... hahhahaha.... after the movie we meet Sab they all at the Small Mac and then i and vincent were like arranging the chairs infront of the small mac in a very neat rolls.. hahaha,... not bad huh? but it was like very fun.... and after tat me elmo and sl left vincent and yt alone... heehee.e.... one guy and one girl walking in the mid of the night... hahaha.... wad will happen? no one knows? haha.... so me went to elmo\'s house and sl took taxi home.... heeheehee...   urlLink    Day After Tomorrow (Twister In LA)&nbsp; urlLink '),(20184,3692874,'25,June,2004','2004-06-25 00:00:00','    This will be the 2nd time i watched \"around the world in 80days\" but no choice... after study no other place to go and no other movie to watch... so.. hahaha... well.. of cause not watch alone(Vincent, Yee Teng, Sher, Pris Lin) well.. the show is funny, of cause... hahaha... but that show is worth while to watch... therefore i am going another time with you, Clara (As I promised) Well... i am terribly sorry abt the NDP tickets... the balloting ends already...(13th June) well... is ok la... we have to study for our prelim anyway... so... dun be sad... next year ok? i promise u that... i will sure very \"Kia Su\" and go for e ticket... ok.. back to it... after the movie... we took 3 wanting to go to \"ponggul\"(Is it spelled this way?) den pris decided to get down for tuition... so i shant be a \"osram\" light bulb... and i get down wit lin and change bus and finally on my seats now.... hahaha  urlLink    Jacky...&nbsp; urlLink '),(20185,3692874,'24,June,2004','2004-06-24 00:00:00','    Finally is our 1st year le.. celebrated on the 23rd June.. unofficial... but we got no other choice cause the official date is 28th of June... but well.. is sch reopen le.. and 23rd June clashes with Song lim operation and i couldnt attend... so i take this chance to apologise to Ah Song.... well... ok..back to our 1st year event... starting e day by going to pasta mania for lunch(Chicken and Beef Balonise*)went to pasir ris park and enjoy the beautiful sea breeze and the beautiful girl(clara)... and after which we went to TM to eat Fish and CO!!! total cost of the meal is $39.65... and both of us added up is just nice($40) hahha... so that is the day... tiring but cool... P.S Found a free gym in Sun Plaza park  urlLink    Clara gave me this on our 1st non-official 1st year&nbsp; urlLink '),(20186,3692874,'22,June,2004','2004-06-22 00:00:00','    Whee.. it start of my day by waking up... of cause... reach sch at 7.10am.. and waited and waited for 10mins... still no sight of anyone... so i started msg YeeTeng... asking whether is she \"tua\" (kena sabo) me or not... den saw MeiHuey walking say is meet at 7.30... so i was like... haiz.. nvm.. forget it... hey!!! YT bought me a coke from japan... The picture below... heeheehe... cool eh? small small.. a guy like me who got \"big mouth\" can sollow all the liquid in it in just one mouth... heehee... but hey.. not drinking it.... (i am among the 10 lucky winners who got this can of coke from her... heehee...)  urlLink    The Coke Size&nbsp; urlLink '),(20187,3692874,'21,June,2004','2004-06-21 00:00:00','    Heehee... after english lesson today went to Plaza Singapura with my beloved clara... and of cause the movie is \" Home On The Range\" at 1st i was like so unwilling to watch cause i was planning to watch \"The Chronickles Of Reddick\" is it spelled this way? i hope so... well anyway many ppl say that my blog is rather short... but hey... I like short ones.. cause is easier to read.. and is like.. hmmm... Summary!! haaha... well... the story is nice!! is funny, is cute (the piggys and the Chicks)and the songs sang by the bad guy... hahaha... not bad la.. thought tat so called \"hero\" is a law enforcement guy... but actually he is a bad guy... he said... \"Follow me and u will get use to the other side of the law\" hahaha... and hey... during the late afternon, evening i mean... my mum taught clara how to cook eggs over the telephone.. hahaha... cool... not bad ah? heehee... and also... clara... i am sorry for watching the T.V. while talking to you.. i know is very rude... but i just couldnt help it.. i am sorry...  urlLink    home on the range&nbsp; urlLink '),(20188,3692874,'20,June,2004','2004-06-20 00:00:00','    Was partying last night all the while... and the girls came at around 8plus... after which they have left... we play MaJiong throught the night till 3am and finally elmo call for a stop... and we went to bed.. i went for a short bath and then join ELmo and vincent... i and vincent were talking on e bed... abt wad i also forgortten le... slept till 1pm!!!! from 3 plus to 1PM!! wow... in my 1st life time i slept for so long... and hey... we went for lunch... and after that... guess wad? my handphone fell... on the floor and to the \"Long Kang\" hahaha... cool? today is the last day le!! pack up everytihng... so tiring... after cleaning the whole house... we went to bath... and thanks VIcent\'s father for fetching me back... hahah... Toast to your father...  urlLink    My HandPhone... After falling into e \"long Kang\"&nbsp; urlLink      urlLink    Did the Barcardi evaporate? hahaha... u know wad i mean&nbsp; urlLink      urlLink    Packing up&nbsp; urlLink      urlLink    Lan gaming... 3 computer... sponsored by MuHua, Vincent and Hong Guan&nbsp; urlLink '),(20189,3692874,'20,June,2004','2004-06-20 00:00:00','    It started off the day by waking up early... took the 8am 31 bus number all the way from bedok to katong... not bad la... refreshing... got drunk in the night... cause the music makes me hot.. so i was so drunk and i was so blur and giddy... floating?? ya... floating... hahaha.. ate \"Taiwan Sausage\",\"Mexican Coffee\",\"Indian Prata\",\"American Hotdog\" and last but not least... \"latino Music!!!\" from my CD... hahahaha... heehee... we are enjoying ourselves here... lots of party for e last... den we have to switch back to our study mode... haiz.... vincent was talking on the phone wit YT... hope clara can tell her mum abt me... so i can have freedom of calling her at night... :(... darling... 1 yr le... is it time??/ heheeehee... i think tats all for e night lor... party.. bbq... all night..  P.S drank too much today...   urlLink    Taiwan Foods!!!&nbsp; urlLink      urlLink    Oh....... Sleepy?&nbsp; urlLink '),(20190,3692874,'20,June,2004','2004-06-20 00:00:00','    Holy Christ... my hair look like a nerd just because of the side of my hair... (Just a moment ago) well... wad to do... dun trust that shop anymore... hair secret better... hmmm... wad else? today DandT quiet boring... did almost nothing... just drill a hole in my acrylic... and thats all.. sit down there do nothing... shake leg.. hhaa..   urlLink    My Hair...&nbsp; urlLink '),(20191,3692874,'20,June,2004','2004-06-20 00:00:00','    Mid Night!!  It was like pitiching black now... wanted to bring camera to Elmo\'s house, but forgotten to bring it... haiz.. what to do? forgetfull guy ma... tml going back to school... once again for DandT... very de sian... but no choice la ah.. ok.. stop talking abt DandT... i am now in the masterbedroom of Elmo\'s house... so called the \"Sian School\" HQ.. haha... staying there for 5days.. till Sunday... but me not stasying untill tat... i very discipline de... i wun stay for long.. played MaJiong just now... aha... and i dragged a key board by its\'s wire from the 2nd floor to the 1st floor... hahaha... isnt that cool? and dragged it up again... hahaha.. den in the end.. me vin and MH throw the key board from the 2nd floor... taking turns to do it... hahaha... ate lots of food... and HG Vin Elmo went cooking Fried Rice... not bad!! is nice...just that the rice is sticky thats all... well... update more tml!! is realy to dark..  Morning!!!  Is kind of funny when i woke up in the morning... Alcohol still running in my veins.. haha... Miss my darling.... haha... kena scolded by my mum early in the morning... but nvm... is ok.. i understand... but me going back on sat.. and ton till sun... hahah...  P.S Barcadi Distilled Vodka \"40%Alcohol\" burns me up...  P.S.S There is always hope in MAJiong  P.S.S.S Is so much fun in the \"Sian School HQ\"   urlLink    The place where the keyboard being dragged up and down by me... and being throw down 5times by me,Vincent and MH&nbsp; urlLink '),(20192,3692874,'20,June,2004','2004-06-20 00:00:00','    Ha... yesterday was a bit kind of funny... really... Me and vincent were decussing on where to go out to study... then we saw sabby\'s nick online... and we were deciding to go her house... she agreed and me and vin meet up to buy some tibits to sabby\'s house... we were planning to do math in e beginning.... den suddenly got a change of plans... and we were like doing geog instead of math... and after awhile me and vin went running on sabby\'s \"running machine\"... hahha... after runnning we decide to play MJ!!! but chinese say... San Que Yi... meaning lacking of one player... so vincent called mh up.. and he arrive and we played!! out of 5 rounds... i won once!!! and sabby... vincent... mh.... witness my 1st time in winninng in a MJ game... hahaha... when i win... i shouted SO LOUD!!! like... FARK!!! I WON!!! den suddenly!! someone at the door... \"sabby!!\" we thought is sabb\'s mother... and all the three guys all ran to her brother\'s bedroom... and act act... den to our suprise.. is moon!! den we were like blaming her.. hahaha... i think thats all...   P.S: dun really like Fabian...   urlLink    Sabrina\'s House&nbsp; urlLink '),(20193,3692874,'30,July,2004','2004-07-30 00:00:00','    Nothing really happen this morning.. just me and my bag went to sch and it is friday!!! nothing much.. have some chat with the same people on the parade square and gave YongChye a CD containing Pictures of the brothers and sisters... he ask me to  keep it for him 1st and ended up he didnt take at all.. is still with me now... went to class and Mr.Sekhar review some of the questions from the DT and after which, BELL RANG!!! hahhah.. change in the class and prepare for P.E.. guys were kicking and throwing balls along the corridor.. just cant help stop playing... hahhaha.. and remarkably, KhengWei thunder a ball over to YeXiang and it hit a light bulb and the \"bowl\" covering it fell on to the floor and KhengWei need to go to the general office for some reports.. hahaha... hope he is fine... we did sit-ups and i ran 2.4km... and my timing is 12.42... ARH!! could have beat ZhongXuan(11.49) seems near.. but is very far!! after school i went straigh home with bus 67 and meet up my mum to POSB bank to withdraw 2000dollars.. for household use... I AM BROKE!!! but luckly still got some Ks in my bank.. well.. not going to use them as they are my future money... hahaha... went out with Clara after that... meet up at 4pm.. but god damn bus 67 just wun come so i am late.. so i went to a nearby flurist shop(Century Square) and bought 9roses for Clara... for being late.(wanted to give her for her birthday, but i didnt give.. so i take my chance) it was wrap with some white paper at 1st... but is so ugly.. and so i told the flurist to design it.. hahaha.. and alltogether cost $14... ($9  for 9roses and $5for decorations) went to meet her... people keep looking at me!! wad e hell man... wad is wrong with a guy holding roses?? so i was sweating.. while waiting at the small mc... saw her sister\'s friend(Sherlyn.. something like that.. comfirm will go and report to Cheryl!!) so waited for about 10mins and she appear and she was shock!! hahha... good for her.. she love it!! wow.. cool.... and we took bus 28 to Toa Payoh interchange and went to bedok for SUMO HOUSE dinner!! japanese food!! i treat her.. cause i owe her one.. hahaha... and after that we went home... tired.. and a bit feverish now... '),(20194,3692874,'28,July,2004','2004-07-28 00:00:00','    hmmm.... SOP!!! is the highlight of the day!!! hahah... SOP not bad la... i like.. not to skip lessons... but seeing people strolling out of the sch when they are surpose to be walking fast!!! ok.. Homec Room got Fire! Explosion! and Gas Leak! wad is it? bad things comes in threes? hahhaha.... ok... Watched SeanPual MTV.... \"Like Glue\",\"Give me the light\",\"Baby boy\" and \"Get busy\"....after school went to 201 MC and study with my baby... had Nuggets meal upsized for dinner... and i ask her question on math... and i learn alot.. but somehow.. we got some dispute and resolved!! haiz... well.. that is today... dun feel like saying alot... '),(20195,3692874,'26,July,2004','2004-07-26 00:00:00','    1st law:A robot may not injure a human being or through inaction allow a human being to come to harm.   2nd law: A robot must obey orders given by human beings except where such orders would conflict with the first law.   3rd law:A robot must protect its own existence as long as such protection does not conflict&nbsp; with the first&nbsp; or second law.   I.Robot is a&nbsp;remarkable show i must say... meaningful? ya of cause... especially the part when e Sonny and Detective Spooner shoke hand... is so.. hmm.. meaningful!!!well.. watch with KhengWei,HongGuan and Songlim yesterday...  Had a great day today!! maybe because Gana and Mok didnt come!!! hahaha.. so relax.. but not in Kelvin\'s lesson... wads with my smile?? he said i was on drug!! come on man.. wad the... wadever.. well... went to class 408 and did a couple of math questions and began slamming chairs and tables with Vincent... heeheehee... Sherylin and Sabby and YongChai were busy touching on their work piece... well... can see that i am disturbing them?? hahah... after which Me,YongChai,Vincent,Sherylin and Sabrina went to BurgerKing and have dinner... YongChai and Sherylin didnt buy any and we went home le lor... good day!!! laughing and talking in BK! '),(20196,3692874,'24,July,2004','2004-07-24 00:00:00','    \"X-Country\" is our sch annual event... For the pass 3 years(2001-2003)&nbsp; \"X-Country\" never rain before... but as for today, big wind started off around 6.30am and dark clouds flew around... many people were praying for the rain is just passing by(one of them is clara.. haha)...  Soon after, big rain came and the sch check up the weather forecast and it stated that the rain wun give way to us and therefore all of us ran and seek shelters... all of us were soaked and some lucky ones manage to grap hold of umbrellas and squeeze into some shelters... as for me and my friends... we ran to a nearby bus stop. No no no... we ran to another one... cause is all fully pack with kids...(seeing idiots who cant even squeeze into the shelter and&nbsp;still stand there and wait....) well... clara offer me her unbrella.. hahh... wow... so touch... but i think her friends need more than me... so i turn her down... hmm... well... went to Bedok Interchange and have breakfast with... Vincent,YongChai,KhengWei,HongGuan,Sabby,YeeTeng,ZhengXi,MeiHuey and Pris... (Treat the girls \"Zui Kuai\" which turns out awful... cause e Zui Kuai sux like... wadever)   In the afternoon went out to watch movie(mean girls)&nbsp;with Vincent,HongGuan,Sabby,YeeTeng and Pris... wasnt prepared at all and i also dun wan to watch \"mean girls\"... but out of the despression of getting out of my house... i got no other choice... and luckily the show turns out right and it is nice and funny... met SongLim after the movie and i saw my neighbour(Candy) inn TM and after which, we went to 201 mc and study!! HongGuan went off half way and came back to play basketball... i did my \"O\" Level math 2003 paper 1 and so i stop for a rest and went out to watch Vincent,HongGuan and his brother play basketball with Sabby... we had a nice talk abt poly stuff... but after tat, disaster came.. i drew on sabby\'s \"O\" level work piece... she cried and laugh... hahaha... and as for me... i laugh and say sorry... well.. she is kind of stress up but i guess now she is alright... okok... enough le... after that we went to eat dinner at the Round Market in Tampinese.. ate fried rice and Ho Fan... Somehow... we get greedy and YeeTeng gave a suggestion of going to DownTownEast and have satay.. so the guys went to HongGuan house and prepare(bath) and the girls went back to 201 mc and have some girls talk.... took 17 there but somehow miscommunication and took 3 instead... we reach there and have satay,chicken wings and talks on Singapore&nbsp;education and future jobs till 12 plus and sabby\'s father send us back!!! that is how i ended up sleeping at 1am...&nbsp;  To  sabby: I am really really sorry for spoiling your art piece... hope we still can be friends!(Hows the black teeth?) Vincent:Easy man... Pris: Holy Oil drip onto my hair... thanks ah...  urlLink     Bedok Reservoir&nbsp; urlLink '),(20197,3692874,'19,July,2004','2004-07-19 00:00:00','    Whee.. this is a post to follow up what i wrote yesterday... forgotten to thanks my brothers... especially KhengWei and VincentYap... cause without the two of you i would have fall dwon the stair and make it worst... i would also like to thanks all my brothers for faning me like a \"god\" on that day... hahaha... i dun wan tat treatment anymore... cause is nice but sickening somehow... head so pain sia.... bloody hell... nv had such painfull-ness before in my life... well... thats all.. manage to treat SongLim,KhengWei and HongGuan chicken wings... and the rest... if u want drinks ask from me lor... i mean... those who helped me la.. and also the sisters... same same la ah... hahahah... OKAY!!! today... early morning got 3 teachers greet me... haha.. cool sia... \"popular\" haiz... i nv faint... is just black out and i just feel floopy... where got faint sia? well... ate lots of food today... cause my mum say i am somehow under malnutrients*... hahaha... well... hope i will be fine... &nbsp; Shout Outs: KhengWei and VincentYap&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; : wow... that was close ah? Brotherhood&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; : wow... that was cool huh? Sisterspirit&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; : Never go hospital find me? hahaha... sorry i scare u all EVSS people who know i fainted:I NEVER FAINT!!! IS JUST BLACK OUT!!! '),(20198,3692874,'18,July,2004','2004-07-18 00:00:00','    Arts night... all preparing since 11am.... (YC sorry for e late comers) and then we left and went to sch for some decoration on hair... hahhaha... althought it doesnt match the colour of my shirt... but is still hip enough on that day.. all the Brotherhood members wore the Pink shirt on that day... only for e performers... so is like we all like very \"eye catchin\" when we were walking to sch and alot ppl were like looking at us... (is it good or bad?) but wadever it is... we feel great... went to bugis after the spray and styling of hair... (that is went my headaches began) went to bugis and zhong bought the pink colour shirt... so now all the brothers have the shirt le.. accept for Joel Foo and MuHua(MuHua???)... hmm... after zhong bought the shirt.. we go on seperate ways... Kw and Zhong went to SP to have lunch, Tan and YongChai went to arcade and me and vincent went to find YeeTeng... and den meet ZhengXi,BoonSiang,Pris and took bus 14 after tat to SP... half way throught we get down at Kallang and decided to tak e MRT because the bus ride took very long... and Pris got \"bus sick\" haha... so we reach there and arts night began and ya.. thats all.. dun wish to say the night... u guys all know...   urlLink    5from brotherhood... more to come urlLink '),(20199,3692874,'16,July,2004','2004-07-16 00:00:00','    hee... was playing with my hair after school and i \"waxed\" up my hair and sprayed with white and blue spray... the colour is nice... hahaha... i&nbsp;love white mixed with blue... but well... no matter how cool can it be... still cant win sabby\'s hair.. her hair stand up right!!! (erect) hahaha... heeheehee.... cool sia... like some goddess.... well.. tml is the arts night.... not involve... but somehow looking forward to it.. dunno why also... so.. lets hope tml will be a fun day!!!  &nbsp;  urlLink    Blue White hair&nbsp; '),(20200,3692874,'10,July,2004','2004-07-10 00:00:00','    went to visit Guan Ying Jie Jie again this saturday and something special happen today... rude singaporean dash into the cabin of a MRT, rude right? before we cld get out he push his way in.. and conflict happen... and SongLim left hand broken still use his hand and point the mid finger at that guy...(si bei Gui Lan hor?) but is funny!! cause tat guy is in the wrong not us... basic courtersy also dun have.. well... went to SimLim Square and SongLim bought a new MP3 player(all changing from manufacturing to Service industry le ah?) hahah... and then after which we went to bugis And bought a pink colour shirt each (Me KhengWei and YongChai)which is so called our brotherhood shirt...hahah... thats all for the day..   urlLink    Froggy \"NewWorldOrder\"&nbsp; urlLink '),(20201,3692874,'09,July,2004','2004-07-09 00:00:00','    Love  Mr. Right!!! women will say... hahha... every guy in this world is mr.Right... love is so profound, i dunno wad i wan... i currently trap in the web of love.. now i mean... being unable to understand wad i wan and wad i need... see couples around i get jealous(even the ones i dun know) i dunno why.. (anyone could ans to me pls tell me)  Life  Why so many bad guys are out in the public??? why cant we talk to each other nicely? why cant we speak in a gentle tone instead of saying each other??? why why why??? is it so hard? cant get the reason why...(Can anyone pls enlighten me?)  Finding  Currently i am finding for a button to restart this world... '),(20202,3692874,'07,July,2004','2004-07-07 00:00:00','    heeheehee... Me and HongGuan fought at play area2 after math triple S... it started off because i was sleeping on the bench(on sabby and denise bags) and he trigger me off by sitting on me and tickerling me.. so i decided to giv him some punches(useless to him) and we start to play lor.. punch here and there.. and then i we went to a corner and ta han 3 punches from HongGuan and one from vincent... wow... pain though... but i wanna train up my self defence... therefore i decided to do it.... get some hitting... to practice self defence...  Shout outs: Brotherhood:Jobs done(u know wad i mean) Clara      :smile when u teach me math... so i can remember clearly! HOngGuan   :Your wrist got cut by my nail in the mid of fight, i am sorry... Vincent    :Hey... i am serious abt taking 3 punches from ur left fist(master hand) before i start learning fighting '),(20203,3692874,'05,July,2004','2004-07-05 00:00:00','    Freedom... not as in being a freeman, not as in i am free to go anywhere and also not as in doing whatever i want... freedom to me is the 3simple letter word... Joy... so simple to spell so simple to say and yet is so tough to have it in Singapore... \"I want a life of my own... i am jason ang... no more\" those who can bring joy to me of cause goes to my girl and brothers and sisters... they are my life most important beings... my family? hahaha... not as much as them... bad boy? no no.. i am stating the fact.. they did bring me joy because if not them i wun be here and smiling everyday...  Messages: Clara:Thanks for the joys you bring to me.. Afraid i might die on the road(u know wad i mean) Brothers:Hahaha... u all shld know ah... hahahaha... other than party whole day wad else? Sisters:Hahhaha... same as on top... '),(20204,3692874,'04,July,2004','2004-07-04 00:00:00','    Meet up at 2pm at small mac and took 67 to Bugis to visit Guan Ying Niang Niang... On the way, we were talking and act like as if we are \"Singapore Bookie\" heeheehee is fun! and on the way, my Mp3 low batt!! how lame can it be... when we arrived, we went directly to the temple and grab 3 joystick and pray to the \"Ti Gong\" and then pray to \"Guan Ying Jie Jie\" manage to obtain a talisman... and i am thinking of giving it to clara... hope she believe in that... went to Parkway for some pool games and i watched them play(cause i got no interet in pool) bet 5dollars on Greece(dun disappoint me dudes) on e way home that time we saw bus 15... and we decided to catch it so we ran and ran and met a cross road junction... we pray tat e traffic light turned red... and it did!! it immediatly turn red!! and we manage to get on the bus... and that is it... sleeping soon to watch soccer match!!! 2.35am  urlLink    Guan Ying Miao&nbsp; urlLink '),(20205,3692874,'03,July,2004','2004-07-03 00:00:00','    \"No matter what I do, no matter how hard I try, is the one I loves who always be the one who pay. I want a life of my own, I am spiderman no more!\"  \"I believe there is a hero in all of us, give us strength, mades us noble. Even thought we sometimes have to give up the things we want the most\"  MJ   :This is funny, I am not sure where to begin... Do you love me? or not? Peter:I....don\'t MJ   :You don\'t? .... kiss me Peter:Kiss you? MJ   :I need to know something... Just one kiss  Spiderman2 is a great movie... watching the second time on monday... which means tml... the show is about live, love and action... therefore is a movie that is worth while to watch... my eyes got water when uncle ben spoke to peter in his mind and also the part when peter explain to MJ over a line-less phone... but most importantly, there is loads of meaning and good examples for us to learn from...  After the movie... i went to SunPlaza park to meet clara and jog with her for awhile and then send her home... went to  pasir ris to play sands!!  To: Vincent: Hey brother! hows ur hand going? Vin And Elmo: dun play sand in Pasir ris anymore! got glass HG: hey.. .come on.. i need to pei pei wo de baby de... dun always like tat la...   urlLink    Spiderman \"on air\"&nbsp; urlLink '),(20206,3692874,'03,July,2004','2004-07-03 00:00:00','    Haiz... Chinese listening compre... sickening... but is pretty cool! IS A LIVE TELECAST!! using a radio 94.2FM(is it?? too nervous to listen carefully)the last essay about reading is very tough! and is so complicated... pass is can pass... but the problem is whether i will get high or not... so scared... but forget it, what done is done... cant retake(unless next year!!) well... my new MP3 player arrived at my house yesterday at 10.03pm!!! cool sia... but den hor... my brother is holding it as 1st priority but is alright... i respect him...  urlLink    My New \"Creative NoMad\" Mp3 player&nbsp; urlLink '),(20207,3692874,'02,July,2004','2004-07-02 00:00:00','    WOW!!! is so.. Happening... went to TP to get my camera from my brother and went to sch with vincent and then to 201 mac wit Vin YT SL Sab MH Pris....saw the staff already preparing for party... And took some pictures(will be shown below)... 6 brotherhood members attend... 5 SisterSpirit members attend plus YT and Moon and also not to miss out... The 406 members... consist of Clara, Sab Sim, Tabby and woon han... and also 2 from the prefectorial board... saw Florence Lee and invited her.. but too bad she got something to attend(If yes, she will be dirty)... the party begins!!! wit the door lock by the staff(as planned because HG and KW had prepared a BIG CREAMY CAKE FOR PRIS!!!) cake cut and candle blown!! having a good time eating those meal we paid... and there it goes.... Hg got old on Pris and vin took the cake and SMASH!!! there it goes to pris\'s face... and then war started... everyone were like throwing cakes and drinks over each other.... and hahaha... we have to clean up the mess(dun wanna have a bad image on EVSS)... and then dismissed! took a \"family\" photo of the brotherhood and the sister spirit... heeheehee... it was so fun... played PokeMon toys wit HG and Vincent.... childish ah? but is fun... thats all for her birthday   Shout Outs Pris: Hapi Birthday From ur pal... JasonANg clara: sorry for making the balloons \"sound\" u know wad i mean sabrina: hitting ur head is a blessing!! relax... hahha Hg and Kw: cake is nice!!!   urlLink    birthday girl&nbsp; urlLink     urlLink    Escape??? NO!!!!&nbsp; urlLink     urlLink    Hahahaha... door locked... no escape!!!&nbsp; urlLink     urlLink    BrotherHood!!!! SisterSpirit!!!!!&nbsp; urlLink '),(20208,3692874,'01,July,2004','2004-07-01 00:00:00','    Today is a stressless day... but woke up in the morning is enough to kill me. because the day before i am tired to the extreme... (flashback:fell asleep in class. slept during Reccess time... so u know how tired i am) Went visit frienster after a month le!!! (So sherylin... pls understand for not writing ur testimonial.. ) have 5friends waiting for me to accept.. accepted 4 friends rejected 1. edited my photos in my friendster... few free to visit!! heehee... my family digital camera is still wit my brother\'s GF!!! when can i feeling very very very tired as usual dunno what happen!!! i must sleep more then 8hrs a day to be fully energized... if not i will fall sick... Go to \"Archives\" (left hand corner) to view my June blogs  P.S: Clara i am sorry for not companying you to study, cause i am indeed very tired now... P.S.S: YeeTeng... pls use HP with care... Mr.Aw is dangerous in a way or another... '),(20209,3692874,'16,August,2004','2004-08-16 00:00:00','    Hee... too long never write blog le.. nowadays busy mah... with what? study? maybe... but spending most of the time running and playing in the class... dun really care about study? hahaha.. cannot say that ah... but am trying hard to get my result... i hope i will get what I wan.. the result that is released at Aug really dishearten me and I dunno what to do.. feel so confused abt my academic... will i have enough time to study? Am i working hard enough? hahaha... u say leh? keep playing in class...    Ok... the day before yeaterday..i got $90dollars early birthday money !! $60 from Uncle Shin and $30 from Nian... thanks them so much... went to Charming Garden the day before yesterday... which mean it is 14th Aug 2004... celebrated along with my 2 other cousins... Kriss and Benjamin.... of cause the star of the day is Kriss... cause is her mum who organinse the event and her birthday lies 1st... hee... well.. anyway the food is nice.. love the fishes and the pigeon meat in bamboo tube? hahha.... not bad sia.. and the tea... hahha.. common ah, but makes me... \"Bright Open\" ahha... after the whole dinner... thank uncle Shin for giving the treat... and so after that we went to Ghim Moh Mc and have some coffee? but the adults left the 4 of us alone... My brother,kriss,benjamin and me myself...to have some cousins talk.. cause we seldom get together and talk... but we were like thinking so troublesoom and so all of us went to a nearby coffee shop and have coffee... my brother, kriss and I had ice milo instead of coffee... hahaha.. i guess next time in the future generation, milo will take over coffee as a common drink.... went home after the long day and slept at 2.30am.. thanks to the tea.. '),(20210,3692874,'15,August,2004','2004-08-15 00:00:00','    Went to charming restaurant yesterday and had a dinner... is actually a birthday dinner... for my cousins and me... i was like... huh? for me also? hahaha... well... dun really care abt my birthday also... hahaha... cause... i am a guy who dun really care abt my birthday... ok.. hmm.... well... went there with my dad and brother.. and met up with my aunties and uncles and cousins.. ok.. [my aunties: Esther ku ,Cat ku and Nian ku] [my uncles: Da ku husband,Nian ku husband and also uncle Shin]..  Some of my fav dishes are the soya fish.. wha.. shiok man.. so tasty.. ordered another rounds... hmm.. The pigeon meat in bamboo tube also not bad... 1st time eating it... hahaha.. hmm.. the whole session was so bored... cause it took quite long for each dish to serve.. at e mean time kriss msged me.. and so we start msgin each other.. cause she is sitting at the far side of the table... dun wanna chout across either... hahaha.. after the whole thing, we went to Ghim moh Mc and have coffee but we decided that coffee shop will do the work.. and like tat lor... wanna get a 700i also!! kriss got it.. hahaha.. is nice also... very good and no lack... camera also very de clear... wanna get it... one day ah.. not now..  ShoutOuts BH:Brotherships came from the heart... with or without the title BrotherHood is ok... SS: be more spotting! come out with us! hahaha Clara: Distintion is a want not a need... dun stress for it... it hurts me... Kriss: every morning sunshines SongLim:haiz '),(20211,3692874,'10,August,2004','2004-08-10 00:00:00','    ok ah... woke up in the morning and played a while of games... recieved sms from elmo regarding the time and meet up place(12.30pm at Songlim\'s house). I pack up everything and decided to go out at12pm.. but hey... was held up by my dad... so ended up at 12.17 then leave my house so no choice have to take MRT... hahaha... reached Song\'s house at 12.51pm and got some briefing from Song\'s mother telling us that we will be selling badges, one for $2...In my heart i was thinking ... looking at those pathetic badges u wan us to sell? come on... even if it is me i also wun buy... free i also wun get it...  We took MRT after the briefing and my bag was rather heavy because there are 10packets of badges in my bag.. sia la... like some drug trafficking like that.. went there and got a couple went up to us and told us whether we know \"Ma Li\" we were like \"huh? huh? huh?\" hahaha... is actualy Songlim\'s mother... mary in chinese... hahaha... got our attendance mark and went to \"cheers\" to get some drinks and foods.... and off we took a shuttle bus to kallang stadium... heeheeheee... from there i know why teenagers nowadays doesnt like to do chrity work... you wanna know why? because the people there sucks... make us sit at a place full of dirt... a place behide a carpark, a place beside people who walk pass us looking down at us, a place where the guards could just spit at us from the stair case... now i know why... bloody people... hahahaha... not going there again.. and do those kind of shit again.... sold 27++ badges and i call it a day le... dun care liao.. i buey ta han le.. not because of the tiredness... but is because... whatever... no pt ma... people just rush in like some ghosts like that... so they wun pay attention of us... well.. went home... just like tat... 31... hee... MediaPlayer playing \"RangWoHuanXiRangWoYou\" '),(20212,3692874,'09,August,2004','2004-08-09 00:00:00','    It started off by me meeting with clara at her house bus stop, we suppose to be meeting all the BH boys at 2.30pm but clara need to attend to something therefore we are meeting each other at 2.30pm and we went to TampineseMall to buy Famous Amos cookie for KhengWei but actually I decided to buy Shaver and Cream for KhengWei at 1st because me and Zhong were discussing what to buy for KhengWei...but because the Sisters wanted to get him that(Sabrina gave me a call) so i didnt get him the Shaver and the Cream...  So Me and Clara went to get somemore things.. hmmm food i mean to eat... we bought 2 bottles of mineral water and \"WangWang milk biscuit\"(baby\'s food) cause i love to eat them.. pretty nice.. and it cost abt 1.40 all together? cheap ah? so i called Kheng and ask him where to meet, he told me that the boys are watching \"ThreeOfAKind\". Called Sabrina asking her where is she and she said she is still at home... so we decided to go to town ourselves.. in the MRT we took picture of KhengWei\'s gift and slept for a while and finally we are in CityHall! we walk to SunTech City and eat.. on the way we saw Stanly... I am seems invisible by him... well whatever...  we bought a plate of Duck Noodle.. not so bad but cost us $4!!! sick.. sno choice ah... after lunch, I called vincent asking him where is him and YeeTeng and he told me to meet them at RafflesCity and so we did and ended up we went to Esplanade and watch the fireworks not bad sia... we reached there at 6pm and the bridge was already crowded with many people so we decided to go to the bottom of the bridge and watch... the spot we took is not so bad.. but the Merlion Kor Kor was blocking our view...his head is too big... hahhahah...  as the time gets nearer ... hmmm 8.20pm like that, people starts standing up and so we did too.. and people were squeezing and pushing to the front... dunno what the hell they want... Fireworks shoots up! not shoot straigh! IDIOTS!!! was looking up in the sky and I notice clara disappear in front of me... I pass the camera to Elmo and ran to find her... i saw her behind the crowd and pei pei her... she said she feel sick standing in the crowd so we stand behind lor... more space... hahahah... not bad... ended up watching the FireWorks just with her.. love it...  and the fireworks started on time and BOOM!!! there goes for a 5min fireworks.. waited and squeezed for 3hrs just for that 5min? NOT WORTH IT MAN... not going next year... hahhha... after that we got a problem le.. crowd again.... Luckly elmo is smart enough to rasie up HOngGuan\'s Lap top and so we follow it to the exit... the brigde is overwhelm by people and cars got stuck stuck just like that.... drivers complaining and traffic wardens were like so busy and running around... MotorBikers were like \"PEH PEH PEH\" all the way to scare the crowds off the bridge... hahaha.. cool man.. nv see such thing in Singapore b4... US all that got la... (BadBoys) hahaha.. after that we took a bus... heehee... uncle Peter ma... but hor... got baricade... we couldnt go to the bus... and so the boys started to think of  ideas and wanted to kick and push the baricades off... but.. better not ba.. later police catch us... hahha.. and finally some running and we decided to cross the road and ran up the bus.. hhahaha... phew... so tired... slept in the bus for a while and finally we reached le!!! steamboat and BBQ!!! wow!!! whee!!! and we were eating like siao... had fun \"frying\" the Meat... hahhaha.. clara told me i am like some chef... haha... hope so... so i can cook some nice dishes for ya... went home in my father\'s van and sending Clara,Zhong and Zhong\'s girl, Joel, Elmo and pris.... hahaha... tour around tampinese sia... hahahaha.... pris\'s mum was... whatever...  hahaha... too caring le... if me i sure rebel de... sure kup the fone and throw it in my bag... but too bad ah.. she is a girl... so... let the future decide.. ok.. finally back at home at1.30am... slept and dream alot of things...    ShoutOuts   Kheng: Happy 17th Birthday!  Clara: Next time you angry or pissed with me... tell me ok? dun hide hide... cause is not good for both of us..  BH: Sia la... not worth it man.. lame fireworks... but hor..since we went throught the \"tough time\" die die must say is worth it...  SS: er... nothing...  Joel: In love with me? hahhaha... I am handsome ah? hahhaha... u damn funny sia...  urlLink    Gift from me and Clara for Kw\'s birthday(Famous Amos)  urlLink '),(20213,3692874,'07,August,2004','2004-08-07 00:00:00','    Woke up in te morning at the same time for a 2hrs school day becuase of national day! well.. .as usual ah... every year National\'s Day is celebrated in a bored and usual manner... flag party march follow up by National\'s song and recollection up next is Mr.EE \"gnirob\"(Read Opposite)talk next will be the singing of some theme songs and then we will go back to our class and wait for further instructions and the InterHouse Final players will proceed to respective areas and so we waited in our class and finally we are able to walk around the sch to support the InterHouseGames.. but luckly this time there is a movie screening in my class... Zhong climb in to the class by the window and HongGuan brought the laptop into the class link all the wire so that it can be shown on the screen infront of the class! hahaha... 2Fast2Furious started just like tat... not bad.. nv watch that show b4 therefore i am quite interested...  When the clock struck 10am, we were all dismissed and so all the BrotherHood boys who are going to Sentosa(KhengWei,HongGuan,Elmo,Me myself JoelFoo and Song Lim) went home straight away and pack up... meet at 11.30am and went to parvillion to eat Lasi Lemak... headed there and reach there at around 1pm and we start taking off our top... play Volley ball for 10to15min and after that we ran into the water... THE WEATHER IS SO HOT!!! hahahha... the sea was some kind of polluted because sea weeds and wood planks are floating at the shore! oh my.. but we didnt care and we even help to clear the sea weeds... hahhaha... we even took a video on fighting!! hahaha.. is with SongLim.. hee... after awhile Joel took his leave because he need to work... we went to eat after that.. had LongTong and Lasi Lemak for $4 cheap ah? hahhaha... after eating and all we went to Bh HQ!!! hahhaha... i think last time le ba? cause heard from Elmo is selling? hahaha... nvm.. hahhaha... moving to Pasir ris? whahaha... thats all for today...  I am sick.. '),(20214,3692874,'04,August,2004','2004-08-04 00:00:00','    WOW... the coffee shop just outside my school sell Coke, Sprite and Grape for just 10cent per cup! wow... this is so cool man... well... their main pupose is to attract people to go la... hahaha... and also advertise their shop... well... ya... ppl will go around and tell people abt tat coffee shop! that is wad i am doing.. advertising free for e shop.. hahah.. heehee smart people ah.. so cheap... hahaha... cheap ppl do this kind of thing... Even i am advertising for tat shop... no choice ah.. haha.. ok... wad happen today? i am very sick today... had Bread and egg for my lunch and a cup of Chicken essence... went to sch as usual and i am kinda sick, dunno wad happen to me ah.. feel very tired and vomiting somehow... but i didnt.. tats good... was made to stand behind the hall during assembly because i said something to mrs mok... she scolded Rebakah for not wearing school socks and i said Mrs. mok also nv wear socks... and she made me stand behind... oh crap... so \"MaLu\" sia... stand behind... well... i stand very straigh and cross my fingers and put them to the front and stay at tat position for damn long... 10mins? hahaha.. very short la... and Mrs.Mok got soft hearten i think... i saw her keep looking at me at e left side of my eyes.. hahha.. was ask to go back and continue to watch the lame programme that was being ut up! wad crap... Survivor Singapore... oh.. nonsense... didnt really care abt those disgusting fellows up on the stage.. makeing fun of themselve? well... wadever..den after the assembly we went back to class and prepare for physic! and Mrs.Florence Lee was not in schook! Yippy!!! hahha... so i went back to my sits and i saw some unwanted newspaper and decided to tear them up to relive a bit of stress... hahhaha... Zhong start throwing at people and all of a sudden, people all start throwing newspaper\"ball\" at each other and the whole class immediatly split into two regiments.. hahaha.. the east and the west(i am on the right)... and we started throwing to the right side... haiz... the left side wasnt sporting and so we are like throwing rubbish at dead ppl... hahaha... i went over and fight for them and i got aimed instead... so sad... hahha.. who cares.. is fun anyway... changin of regiments... hahaha...ok ah... after school went to eat lor... eat \"tou hu\" e sweet sweet de... and also 8pieces of bread(ham and tuna) and den off i went to math triple S... Was like a mad man in class cause i keep playing with the pipe(black snake) and introduce people abt the rare species along wit Afif and people were listening ti us and then the Mdm Rubia chase me and HongGuan out of the class and we went to seek refuge next class.. hahaha... go there also do nothing.. after Math SSS i got a retest to sit and it went quite right.. can pass but B3? not sure... hahaha '),(20215,3692874,'02,August,2004','2004-08-02 00:00:00','    Is August already!! sick man.. it means, from today(2.AUG.2004) it will leave me to exactly 3 more months to GCE \"O\" level!!! oh holy... please if any god or goddess see this.. please.. bless me and bring me through this somewhat major examination!! i really need that... althought i have not been a goody buddhist but please help if u can... thanks dudes... hahaha... okok... lets talk abt yesterday...   Yesterday   was Sunday! Woke up in the morning and called my baby telling her tat i am desperate!! hahaha... dun anyhow think think ah.. desperate to ask her out to teach me math ah... sia la.. math got 10/60 man... that is crap.. well.. no choice ah.. been sleeping in class and nv do math for 2 years... last minute \"chiong\" of cause cannot get good result.. well.. i planning for a pass for MOCK exam and a higher grade for prelim!! will I get it? we shall see for the upcoming MOCK exam.. if i fail again? i really got nothing to say... haiz.. hahhaha... OK!! she told me she going to cut her hair.. i told her i will be following her to the salon and she insisted cause she said is weird when i am there to see her cut her hair... hahaha... well.. no choice lor... so in the end she dun wanna go cut hair le.. and so we meet up at 2pm and she told me bring graph papers.. but i rush out of my house at 1.20pm and forgotten everything... forgotten to bring graph papers and also... my money!!! my wallet is so empty.. haha... no choice ah.. forgetfull me... muddle-head? or stupid? most likely is stupid ba... Actually we surpose to go 201 MC and study ended up we took 28 and went back to my house and take graph papers and money.. so lame sia... and den we went to Bedok MC instead... hahahah... well.. we had 20piece nuggets and large Milo!! hahaha... total cost... $4.90+$2.50=$7.40.. not so bad ah... cause i ate most of the nuggets!! hahaha... heeheeheee... cheeky me.. cause she dun really eat much.. after eating we are down to seious work!! ok... i took out my math paper where i suppose to hand up... but i forgotten.. so she thought me math and her face got red half way through.. now i know how stupid i am untill she cant \"Ta Han\" me... well.. teacher is tough to be... so after studying is abt... hmmm 5.50pm plus? we walk around Bedok and i brought her to the Swimming Pool,Bedok Stadium and Silat Federation.... hahhah.. like touring around Bedok rite? no.. is all at the same place ma... hahhaha... well... went to the bus stop thinking of sending her home cause she will be eating dinner(steamboat). But she said to me that she doesnt want to eat steamboat and so she called her mum and told her she will be going out with her friends to eat... so we talked for about 30mins in the bus stop and i decided to go home and take money... cause i ran out of money le... saw Dominic and his brother at the bus stop too!! sia la... so coinsidence... hahahaha... well... i think there is a basket ball game or something happening... so well... dun really care.. so we walked home hand in hand... hahha... national day coming ma.. .hahahah.. went home and took 10dollars out and clara was waiting outside my house... filled up her bottle to the brim and off i went out with her to Tampinese! took 18.. and we went to Pasta Mania and have the same thing... she had beef balonisge(something like that)in linguine and i had chicken balonisge in puisilli... hahhaa.. all the spelling cock up de... cause i dunno how to spell!!! understand can already.. . so i send her home but i didnt get down the bus... cause she said that her parents will be home soon... hmm... so that is it... end of my day on Sunday happily...   Today  is Monday!! wait ah... early morning.. wad u wan me to write??? hahaha... later when i am back... ok? Ok.. i am back from NUH.. nothing much to me ah... just that i have some sort of major blood clot in my head... hahaha.. JKJK... heehee... nothing much.. the doctor analysis me and took down a lot of notes as i said and is kind of funny la... cause the doctor and I had a great time talkin... well... nothing hapeen la... is just that my veins react very slowly that cause blood to flow slowly and sometimes it contract too slowly and only a little bit of blood go to my brain causing me to faint or black out... phew... luckily not some kind of terminal illness.. well.. got some medication and i need to take down the number of time i got migrant! haha... follow up check up will be on 28th SEP 2004... hahaha.. after Prelim... cool... well.. that is it lor...   ShoutOuts  Baby: Heehee... getting to like you more and more... 1 year and 1 month and 4days le!! and my love for ya is still growing!! directly proportional!! DCbag: you ah... sia la... multi purpose sia.. can help me brighten up her day... BH: Onwards mmy brothers!! SP: BH gin na tan le lai... hahaha... jkjk... sabby: lol.. nice monday morning chat with you... '),(20216,3692874,'01,August,2004','2004-08-01 00:00:00','    Woke up in the morning feeling fine!!! therefore i pick up the phone and call everyone up(Same old brothers same old sisters).. hmm... ended up on Me,HongGuan,KhengWei and YongChye... wore white for that day, hahaha... bai ma wang zi ma...with open inverted commas... hahha... hmm... went to have lunch wit HongGuan and KhengWei first at and \"open-air\" coffee house.. somewhere at the 1st level of Parvillion and left side of BK(If u are facing BK).. wadever the place is, we had LasiLemak and Otak for just 2 dollars and also 2 cups of drinks.. hmm.. i took orange juice and HongGuan took IceLemonTea, kind of regret because the LemonTea looks much more bigger!! after which we went to meet YongChye(meeting us at 1.30pm) and off we went to Penisular somewhere in CityHall with the help of SMRT... IN the MRT me and HOngGuan were like keep on suckin KhengWei\'s water bag... cause inside got hundred plus ma... hahhahah and so the train arrive and HongGuan sat down on the floor and KhengWei follow suit... and YOngChye next.. so i was so reluctant to seat down because i am wearing white.. but no choice,have to sit down.. because is a long journey and we will be walking very long after that.. so i just sat lor.. no choice.. welll... reach there abt 2plus and we start walking around... 1st shop, not so bad... saw a big bag like very nice to me.. but they say dosent suit me.. so we went to the 2nd shop.. saw the same bag wanted to buy but is 75dollars!! I only got 40dollars in my wallet!!! so we went to the 3rd shop... and saw some bags.. the same i think, and HongGuan and YongChye helped me to bargin... but still cannot... so we went to the 4th shop.. wow! this is the place so cool.. HongGuan call the shop owner \"brother\" i see his face like happy like that.. and HongGuan Bargin untill $65..so i also follow ah.. \"hey brother... $60 can? please brother.. is nice...\" and it works!! it was sold to me at $60dollars!!! and so HongGuan helped me to pay $20 1st cool sia... i hope no other place sell cheaper than tat.. At first, the bag look kind of funny to me... haha... big ma... sia la.. but after awhile i feel comfortable with it.. and so we went off to Funun the IT mall and walk around... we were like hungry? and so we went around and look for IPODs and other MP3player... not planning to buy of cause.. after walking around and it is abt time we stop and sit down.. so we decided to go to the nearby ColdStorage.. and have some drinks... and hey!! we saw roastered chickens selling at great prices!! $3.95 for a whole chicken...and so me and HongGuan shared money for a whole chicken.. heehee... ate outside the store and haha.. so filling!! and nice.. blackpepper chicken!! shiok sia... went back to Tampinese to watch \"House OF THe Flying Daggers\" the story sux! but fightin scenes is cool!! so many effects and slow motions!! overall rating.. 2.5/5stars.. went to BK and have dinner after that we went home!! and HongGuan lost his EZlink card.. dunno whether is with him or not..  urlLink    My New DC bag  urlLink '),(20217,3252533,'24,April,2004','2004-04-24 00:00:00','    Today, I\'m gonna show the youth how blogging works! Hope it\'s gonna be successful. And tomorrow, I hope to intro this site to the church ;-) and get some support!  Heheh...   '),(20218,3252533,'23,April,2004','2004-04-23 00:00:00','    College today was fun.. 9:30 - 11am only. I went to internet lab 11-12:30pm.. The lecturers were having a meeting of some sort. So, half day of Telecommunications.. Was kinda fun, cuz Jason was funny. Later, as I was waiting at the bench for mum and dad, suddenly I got a call. They went out with Uncle & Aunt for lunch. I had to eat myself. I went to the restaurant operated my the Hotel and Tourism (H&T) guys at my colllege. No 1st class service, but decent service lah. I had this pie that costed RM6.. reasonable lah.  I met Sean Paul. He\'s from Sarawak, dad\'s a politician, mum\'s Filipino. He\'s 23 now, but just 1 year my senior. He was studying Civil Engineering at INTI Sarawak, but middle into year 2, he sort flunked a couple of subjects a couple of times. Quit INTI Sarawak and decided to come to KDU Penang instead. His girlfriend\'s in Penang, so.. that\'s the main reason he\'s here. Ahh... Neway, we had lunch together.  The apartment we were looking to buy wasn\'t good. It housed sick patients, lotsa \"tourists\" looking for a spot to live. The apartment is very nice, but unfortunatelly, there are drug addicts lurking nearby... and lotsa sick people! Dang.. they\'re like waiting for medical attention in the hospitals nearby. The lifts are darn slow.. you\'ll prolly have to wait at least 10 minutes everytime! So stupid.   As for now, we\'re staying at our present apartment. Its abit far from college, but I\'ll be driving soon, so that\'s good =)  Remember to visit  urlLink My site  and  urlLink My blog   '),(20219,3252533,'23,April,2004','2004-04-23 00:00:00','    This is the first post! Hi! If you\'re a youth, please register with  urlLink Blogger.com  and give me your e-mail address which you registered. Send it to me at pip@bmbc.info or tell me during youth meeting! Our forums is in the main  urlLink BMBC  page.  Hope this will be a success, and please remember to update often! (2 x a week minimum!) Thanks all!   '),(20220,3252533,'26,April,2004','2004-04-26 00:00:00','    The pictures are up. Its available  urlLink  HERE   or the links on the right. These are the sarawak pictures!   '),(20221,3252533,'08,May,2004','2004-05-08 00:00:00','    Phew, end of the week already. Lets recap.  Thursday afternoon was BM classes with Puan Suraya. Today we did some Fahaman, and believe me, its much tougher than O Levels. No doubt I got credit for O\'s, but if I were to take the test here, I would surely merely Pass. It is a great experience, I can testify. I met some Hotel and Tourism guys. They\'re all a fun bunch. Breaktime, we went down the cafeteria.. For once we all had lots to sembang about. Kenneth was a fun guy... with anti American setiments. His ambition is to be chef is large cruise ships. Also, he had experience working at Shangri-la. Ah well..  Friday telecom class. The supposed test didn\'t turn out to be true. But Jason promised this coming Friday we\'ll have a test. The new National Service girl came. Forgot her name... but she\'s the pleasant type. Nothing much happened except the usual tech / telecom discussions. We\'re into networking and networking fabrics now... (i.e. telephone wires, coaxial, and fiber) Its fun!  Saturday, today went for Micro-app. We did some MS Dos commands and played around abit. It was basic, thus deemed boring by someone as leet as myself. Everyone knew how DOS worked. As usual, the challenging part of this subject is to put our knowledge in words. Thankfully, I have Computer Studies experience. Woot!!  Later I\'ll be off to BM for the weekend. I received an SMS from Radiance that I might be needed as backup singer for the youth meeting tonight. Can\'t wait for that :P Then, Sunday we\'ll attend church, then go to granny\'s. We\'re gonna get the orchard cleared up abit. I might have to go college on Monday by bus. Take the 7:30 from BM to Komtar, then, hop onto another bus to Burma / Anson Road hopefully by 9:30am where KDU is located. It has been done before, and its quite an exciting thing... for one, I\'m a public transport virgin and for another... its like.. \"independence\" ya know.. Now that mum and dad are staying together with me, I\'m pretty much under scruntany.. not that it\'s a bad thing.. but I love freedom.  So as usual, I\'ll be afk (away from keyboard) till Monday afternoon after college. Happy weekend, everybody! And visit http://bmbc.info , ok? Sorry about piptech.org not working!   '),(20222,3252533,'07,May,2004','2004-05-07 00:00:00','    Ah.. blurry pics!      For more, visit  urlLink The camp photo file list !   '),(20223,3252533,'06,May,2004','2004-05-06 00:00:00','    http://www.hi5.com/i?l=QKTQ83YQ   '),(20224,3252533,'06,May,2004','2004-05-06 00:00:00','    Last Friday, Telecommunications class with Jason. We started the more technical parts of Telecom now.. with complex Physics terminologies and such. It was okay, just need some reading up. We\'re far behind the schedule, cuz usually when we discuss about something, other things pop up! Ev3r so often, we\'re distracted from the original thing. I think we were talking about Radio waves & Communications. Then, he said something about Astro. We all know that Astro is satellite, and that\'s another topic.  So, Jason actually worked for Astro for a period of time. He brought up the issue of hacked Astro cards and said that Astro actually knew those who are stealing their services. A mate of mine said that his whole neighbourhood are using hacked cards. I mean, pay a one time fee, and you\'re FREE from pesky subscription charges. It costs about Rm 130 a month for all packages here in Malaysia, so, in a year, its about Rm 1500 in fees. Then, he talked about spot beams... that Astro has a big \"footprint\", yes.. basically, if you\'re in China, India or Philippines, or even West Australia, you can get Astro reception. However, a concentrated \"beam\" is sent to an area (for example, Kampong Beribi in Brunei).. then, the computer calculates how many signals \"reflect\". Its sort of like a \"ping\" and then, they check with their subsciption records for people who paid. If they get more \"pings\", that means someone\'s not paying.  Actually, its kinda foolish for people not wanting to pay AT ALL. It\'ll be too obvious dammit! I wonder what guts it took you to go to the Astro office and say, \"I don\'t want your service anymore\". I mean, Astro is god. There\'s no other competition, and its programs are top dawg, other that the annoying local ads. If you don\'t pay a cent, you\'re telling them \"I have a dish at home... er.. I ain\'t gonna use it\". So, word of advice for hacked cards users \"pay the cheapest subscription (and use your hacked card)\". If you don\'t pay at all, they know something is going amiss.. If convincted, you\'ll be fined Rm 100,000 here in Malaysia and possible jail sentence; doesn\'t matter whether or not you produced the cards, as long as you\'re using it.  Btw, my Telecom lecturer happens to be an \"ex\" American grad phone phreaker, and hacker. He\'s Malaysian chinese but the accent, attitude and all is there =) Owns an audio shop, I believe, with RM60k turntables at his disposal and WiFi network at home :P Very fun guy. Today (thursday), there\'s gonna be a Engineering Society meeting with elections and all, basically an AGM. Hope I get selected.. thus improving my chances for a merit scholarship :P  Later that day, came back, rested. Arrived BM at 4pm, picked up a church friend, and set off for Sungai Petani. Its actually about 50 km away... but accessable via highway. In case you don\'t know.. the highways here have 110km/h speed limits, thus... its not \"luar biasa\" if you see cars zooming at 140-160km/h. Speed traps are set tho.. and you have the option to (a) bribe the cop (b) give lame excuse and escape altogether (c) pay the fine. Hahah..  Arrived at  urlLink Cinta Sayang Golf and Country Resort  by 6:30pm. Tonight tho, some church members hadn\'t arrived yet. That night, it was dinner by the golf driving range. FOUR!  Next morning, more friends came. Almost everybody was there. The Youth, little kids and adults were separated into their separate groups with Ps. Alex Chong ministering to the adults, Unc. Micheal and Aunt. Hannah with the Youth (fun guys) and their daughter Joanne playing with the kids. If you look through the  urlLink photos here , basically, it tells you the whole story lah.  Throughout Friday, Saturday and Sunday, there was non stop, high octane action and amusement. Certainly there was time of worship and bible teaching, then the games, and the swimming in the evenings. The thing about camps is that, for the first days, you really think its tiring and a bore, and when its time to head home, you miss it. Luckily, I was well prepared (mentally)!! And.. actually loved the whole experience.  urlLink Aiyah.. just browse through the pictures lah! .  I made new friends, and grew closer to older ones. It was fun. The activities organised were well planned and all. uncle Shannon, an entertainer by profession showed kids, youth and adults alike some of his \"geng\" magic tricks. Phew.. such fast hands to deceive the eyes, and the mind.  Monday, packed, and went home with a teary eye. Not teary because of missing the camp, rather the thought of gaining knowledge of the one and only God almighty and the lessons learnt. Arrived home by evening and was too tired.. just slept till 9pm before going for dinner. Tuesday, still in BM, we did absolutely nothing. Mum had the clothes washed, dad did the gardening, I watched TV (no computer there!!). Evening visited granny tho, and the Siamese temple my dad grew up in. Too bad I was out of \"ammo\" to take pictures of the temple. Recently our Chief Minister  urlLink Tan Sri Koh Tsu Koon  visited the temple and He and His guys were tasting some \"masakan kampung\". They liked the Thai dish \"kilek\" so much that they promised to return for more soon! Apparently, the temple\'s chef\'s secret was excessive \"santan\".... That\'s what granny said. Granny could cook better \"kilek\"! She\'s got a tree in the orchard.  Oh, and its durian season again. Finally, after a year of fruitlessness, those thony fruits have finally sprouted. Its not fully riped, but when it is, it\'ll automatically drop down. Durian tree is a special tree which requires minimal attention yet producing rich, tasty fruits. There were about 40 trees brearing fruits... kinda cool to see all those durians on the trees. Maybe in a month or two, we should have some fruits ;-) I\'ll report to you the taste and all. There was also campedak fruit and other small ones. We left granny\'s at 7pm, and tapau some kampung eggs. Small and cute, but yummier than \"normal\" eggs. It is said kampung products are always better than the mainstream. Its true! I wonder what about kampung \"chics\" ;-) ;-) ;-)  Wednesday (yesterday).. at 9am, we went to the IC department to collect mum\'s MyKad. After 6 months+ of processing, its finally done. I took control of the car, while parents lined up. And a minute later, I was lucky enough to land in a parking slot. Whopee. Oh, earlier the previous night, I drove from the dinner place (Tenang restaurant) back home. Its about 1km+, but my driving was near flawless. So smooth, so lancar. I\'m leet now. Afternoon, I rushed to this courier service to collect my mail. I won this voucher for Rm 100 in Thai food in KL. Gonna give it to my sis. Then, was slightly late for Maths class. There was a test today, so we were pretty spread out. It was all on Surds and Rational Numbers. I think I did okay, with 1 or 2 errors. But what the heck. Its 10% of the semester total, but yeah.. I did fine.  Thursday. My day started at 3am. Wanted to watch the Chelsea Champion\'s League match. Watched for a good 10 minutes then said to myself \"Argh.. i\'m no chelsea supporter! Everyone knows I love  urlLink Man United . What the heck? I went online instead and was surprised to find  urlLink San Fui  online. Talked to her abit. Apparently she was under exam stress. Sigh~ sometimes other people expect too much, and when numbers are programmed into your mind, your \"panic quotient\" just skyrockets. If you\'re kind hearted, ask her to calm down and think less about all the worries lah. Full time, Chelsea 2 - Monaco 2.. Monaco meets Porto in the Champion\'s League Final. Visit  urlLink Soccernet  for full report. Ah well, the team that narrowly beat Man United contraversially is now in the Finals. Nevermind. Its all right. Its 11am now, and I\'ll rest abit before going for BM classes this afternoon. After that, there\'s a 2 hour Eng. Society meeting, so look forward to that. I might report back tonight, if given the opportunity.  Thanks for visiting, and sorry for the server downtime.  urlLink Pinchpenny is cheap, but sucks . Oh, tell everyone you know about Piptech.org by adding this \"signature\" to your e-mail... or by word of mouth of MSN. Tell \'em \"Visit piptech.org today!\" =) I owe you one, dude and dudettes!   '),(20225,3252533,'05,May,2004','2004-05-05 00:00:00','    Sorry about the slighly blurred pictures. My hands were shakey... or perhaps I didn\'t know 100% how to use this digital camera. Seems like good ol\' fashion ones take better pictures. Sigh~ Advancements... one step forward, two step backwards!  I think I can claim that there are a good 100 \"clear\" pictures lah.. tell me which ones you\'re interested in ;) Good day, all! Mail me: pip@piptech.org if you are a BMBC youth who wants to try out this \"blogging\" thing.   '),(20226,3252533,'05,May,2004','2004-05-05 00:00:00','    Location: Cinta Sayang Golf & Country Resort, Sungai Petani Dates: 30th April - 3 May 2004  I had an exhillarating, extraordinary experience at this camp... the best I\'ve ever been! All good things have got to come to an end, but I\'m pretty happy overall.  BMBC Camp 2004 Pictures are available  urlLink HERE . I\'m sorry they\'re not sorted properly, but in due time lah ;) Now I\'m in college :P Just had a mini Maths test, I think I did okay, altho got wrong!  The full sized pics are available at RM 1 for a 4R print (buy 5 free 1) or RM 2 for 3 copies of 3R print. Sorry about the funny pricing stucture, but its still pending approval from church. Meanwhile, enjoy the online versions =)   '),(20227,3252533,'14,May,2004','2004-05-14 00:00:00','    Tomolo, Radiance is hanging me out to dry! I\'ve been given the honour to worship lead the youth meet tomorrow night. Not only have i gotta face the 10s of youth, I\'ve gotta face Simon Cowell of BMBC... aka Ps. Tamil. Gotta learn how to take constructive critisism.. Ah.. and I need your support oso lah.. :) So, sing along and lets have fun worship and praising God!  Sister Olivia is here on official business. She\'ll be going back before Church.. cuz got work Monday. Live is as usual.. and Computer \"Microapplications\" tomorrow morning b4 going BM.  What songs do you guys listen to (other than gospel / Christianity related)...? Just curious. And try out the Comment system lah ;)    A couple of my buddies doing Hotel and Tourism.. we\'re studying in Malay class cuz we didn\'t get Credit in SPM.. Actually, my O level credit isn\'t recognised. Gnite all!   '),(20228,3252533,'13,May,2004','2004-05-13 00:00:00','    This msg goes 2 evry1 esp. WILLIAM HUNG(He Bangs) !!   urlLink Click Me    Edited by Pip   '),(20229,3252533,'13,May,2004','2004-05-13 00:00:00','    Here\'s what I read today... Just want to share   Sometimes it doesn\'t take much to get us down, does it? An unkind remark from a friend, bad news from the auto mechanic, a financial set back, or misbehaving child can put a cloud of gloom over everything, even on the sunniest day. You know you should be joyful, but everything seems to be against you, making simple tasks a struggle.  David must have been feeling that way when he wrote Psalm 6. He felt weak and sickly, troubled, forsaken, weary, and grief-stricken. But he knew what to do when he was down. He looked up and trusted God to take care of him and to see him through.   When we look up and focus on God, something good happens. We get our eyes off ourselves and gain a new appreciation of Him.  Next time you\'re down, try looking up to God. He is sovereign; He loves you; He considers you special; He has a purpose for your trials.  Yes, life can seem unbearable at times. But don\'t let it keep you down. Meditate on God\'s goodness, talk to Him, and know that He hears you. That will give you strength to get up when you\'re down.  Come, ye disconsolate, where\'er ye languish - Come to the mercy seat, fervently kneel; Here bring your wounded hearts, here tell your anguish: Earth has no sorrow that heav\'n cannot heal.   '),(20230,3252533,'13,May,2004','2004-05-13 00:00:00','    Maybe we should all post our details: I\'m at Hi5 piptech@gmail.com and Friendster pip2110@hotmail.com... My e-mail is piptech@gmail.com, so all e-mail forwards, please go there. Soon, I\'ll have my pip@piptech.org e-mail ;)  Sigh~ The Palm Zire 31 still not in Penang, no stock. Its a wonderful device.  urlLink Check out this Zire 31 .. l33t stuff.   Btw, Keston, jgn lupa bawa CD, ok? And this Sat nite, you\'ll all get to see the camp photos in Fujicolor! :P Where is Daniel / Jason?   '),(20231,3252533,'13,May,2004','2004-05-13 00:00:00','    HI Everyone! Could you ppl please add me at the following...just to keep in contact.. Hi5.com, msn messenger :l_vivienne@hotmail.com Friendster : vivienne_vivacious@yahoo.com ICQ: 55190732  Thanks a bunch! Have a nice day ahead....   '),(20232,3252533,'12,May,2004','2004-05-12 00:00:00','    interesting !!!!   '),(20233,3252533,'12,May,2004','2004-05-12 00:00:00','    In Shakespeare\'s time, mattresses were secured on bed frames by ropes. When you pulled on the ropes the mattress tightened, making the bed firmer to sleep on. That\'s where the phrase, \"goodnight, sleep tight\" came from.  The phrase \"rule of thumb\" is derived from an old English law which stated that you couldn\'t beat your wife with anything wider than your thumb.  It was the accepted practice in Babylon 4,000 years ago that for a month after the wedding, the bride\'s father would supply his son-in-law with all the mead he could drink. Mead is a honey beer, and because their calendar was lunar based, this period was called the \"honey month\" or what we know today as the \"honeymoon.\"  In ancient England a person could not have sex unless you had consent of the King (unless you were in the Royal Family). When anyone wanted to have a baby, they got consent of the King & the King gave them a placard that they hung on their door while they were having sex. The placard had F.U.C.K. (Fornication Under Consent of the King) on it. Now you know where that came from..   '),(20234,3252533,'12,May,2004','2004-05-12 00:00:00','    I spent a good 9 years of my growing life in Brunei.. Here\'s what you need to know about this little country near Sarawak.  You might be a Bruneian if :  1. Your stomach growls when you don\'t eat rice for a day. 2. You believe any Lee Kum Kee sauce can turn bad cooking to gourmet food. 3. You think our country is a democracy. 4. You talk at the movie. 5. You use a water hose instead of toilet paper in the bathroom. 6. You eat fried rice/noodles in the morning. 7. You prefer Versace or DKNY jeans over Gap or Levi\'s. 8. You don\'t think Jim Carrey is funny. 9. You think Si Kudil is funny. 10. You think RTB (Radio Television Brunei) singers are kampungan/style lama(outdated). 11. You carry a 16 oz. jar of sambal to wherever you travel. 12. Driving a car cheaper than $20,000 embarrasses you.* 13. You think dangdut is stupid, but listen to it anyway, because everyone else does.* 14. You are willing to travel 36 miles to watch Liga Perdana.* 15. You think other Asian countries are poorer than yours. 16. You are very good at avoiding potholes and other road hazards.* 17. Your local McDonald\'s serves rice and sambal. 18. You think Maggi mee is a staple food. 19. You\'ve tried passing Malaysian coins in a vending machine or pay phones. 20. Your average height is between 4\'5\" to 5\'5\".* 21. You average weight is between 60kg to 200+++kg.* 22. You\'ve smuggled electronics, CDs and porn into Brunei.** 23. You do your shopping in Singapore/UK/Miri. 24. Your drivers license claims you are 5 years older then you really are. 25. You\'ve legally bought pirated software.** 26. You\'ve been forced to memorize the National Anthem.* 27. You think Kuala Belait is a foreign country. 28. You know exactly how many districts Brunei has. 29. You\'ve eaten something sold off a cart on wheels. 30. You realized that money is everything before you were six. 31. The first thing that comes to mind when hearing the word \"Brunei\" is \"rich\". 32. Someone you know has never ridden on a bus/taxi/public transportation in Brunei.** 33. Your daily commute includes thinking up new ways to get rich quickly, i.e.4D. 34. You don\'t mind people being late. You almost always are. 35. You think standing in line is a waste of time.* 36. You\'ve tried every Monday of your youth trying to avoid Assembly day. 37. You\'ve used a mosquito repellant that looks like a coil and is lit on one end. 38. You use the terms \"kan mati\", \"-lah\" and \"pukima\" on a daily basis.* 39. You know what MIB is, what it means and know it by heart.*??** 40. You complain that movies don\'t have sub-titles. 41. Your daily conversation may include enactments of TV commercials.** 42. You\'ve consulted a dukun or urang pandai. 43. Your whole class cheated on a test, and gotten away with it.* 44. You\'ve spent the night before an exam looking for someone who sells the questions.* 45. You like the smell of minyak attar. 46. You think the Liga Perdana is equal to the Super Bowl. 47. You talk loud thinking everybody is deaf but actually YOU are. 48. You\'ve a 16\' satellite dish in your front yard. 49. You\'ve never ridden in a motor vehicle with three wheels. 50. You miss your maid during laundry day. Other days you hate her guts.** 51. Your clothing has brand names printed on it that is visible from 50 feet away. 52. You attend weddings only a few minutes before eating. 53. You\'ve attended weddings with every piece of gold jewelry you own. 54. You go to \"kadai kaling\" to get your weekly supply of ketchup, salt, pepper and napkins. 55. You know more than one music group that stole the tune of Cranberries\' \"Zombie\" 56. You\'ve a can of Shelltox on your kitchen table. 57. You make major decisions based on who\'s doing what. 58. You don\'t take advantage of SALES cos\' you know it\'s not.* 59. Someone in your family has extra pockets in his outfit to hide souvenirs extracted from the palace.* 60. You\'ve paid more then $100 to get your own private number license plate. 61. When watching TV you regularly find that all the channels broadcast the same thing.** 62. You know more than 10 acronyms/abbreviations.* 63. You call every Chinese girl \"Moi\".**  You might be a Bruneian if (Part 2):  1. You think an old modified car with low suspension and large tyres are cool. 2. Sago (ambuyat) is your staple food and it is a necessity to have loads of it even when staying overseas. 3. You think dangdut is not stupid and converting them to chacha or raggae is cool to dance with. 4. You can\'t dance or talk properly while on TV. 5. You act like a rich tycoon in Miri even if your average income is below $1000.** 6. Wear as much makeup as possible. rosy red cheeks are in, natural looks are pale. 7. You will want to work in an air conditioned environment with ties hanging from your (pathetic) neck even if your academic qualifications are SH*T. i.e you want to be but not the indians. 8. You furnish your home luxuriously even if it means eating rice and kicap for the last 2 1/2 weeks before the end of the month. * 9. You love rosewood furniture.* 10. You have TV sets with separate decoders and VCR\'s in every room in your house. 11. You\'ve never heard of toilet hygiene. 12. Favourite cars - Mercedes and BMW, below that is cheap stuff. 13. You own more than 1 hand phone.* 14. Your favourite slippers - Jaguar. 15. You love to sing to the Karaoke even if you are out of tune regardless of what people think of you.* 16. You are always right.** 17. As per no. 6 - wear as much gold as possible. 18. Your idea of a balance diet includes, daily, curry chicken, beef rendang, udang masak asam pedas, sayur masak lemak and sambal belacan. Calories are your best friend. 19. Any one who is not obese is not beautiful. 20. You interfere when someone is talking. 21. As per no. 20 - you talk too fast. 22. \'Hutang\' (Owing) is your best friend. 23. You hate it very very very (x 100) much if a lady driver overtakes you (even in a powerful car) and you must, in turn overtake her no matter wht the risks are. ** 24. If you are a lady driver, you will overtake without looking into your rear view mirror first and you will ignore the sound of screeching tyres behind you. 25. For eg, if a star from Malaysia performs in Brunei, the next day you and every body around you will be playing his/her songs on the car radio full blast. 26. You make sure that everything you consume is halal but eat everything else when you are overseas.* 27. You leave Brunei International airport in baju kurung and wearing tudung (MIB concept) but come out of your hotel room in singapore wearing sleeveless T and tight jeans. (never forget your heavy makeup). * 28. You tie your infant up afraid that when he/she grows up his/her legs will bend. 29. You believe very much in pukau. 30. You must exaggerate if an incident occurs. 31. You dress up when going to Jerudong Park. It\'s the place to be seen. 32. Every child you have owns a fairly expensive car.If he/she is the only one you buy him/her a very expensive car. 33. You are always right.  (* : I have this on good authority that it\'s true :) (** : I REALLY REALLY am surethat it\'s REALLY true :)  Gnite everyone.   '),(20235,3252533,'11,May,2004','2004-05-11 00:00:00','    who needs a dance class?? we all r pros !! look for urself! luckily Joanne was there !   '),(20236,3252533,'11,May,2004','2004-05-11 00:00:00','    5 new faces = 5 new bmbc members if i\'m not wrong James is not in d pic !   WELCOME 2 BMBC GUYS !   '),(20237,3252533,'11,May,2004','2004-05-11 00:00:00','    Scam alert! I got this in my e-mail and I feel suspecious!  From Marc Hunter   reply to: hunter/xxxx... (lotsa stuff) Subject: Could you please verify your e-mail address!  Hi there, You are receiving this message because automated spam detection software has flagged your recent email to me as suspicious, and it did not recognize your email address. If you simply reply to this message, your original message will go through and you will not be asked to do this again. Sorry for the incovenience, and thank you for your understanding. Rgds, Marc   I warn against replying. I have a sneaky feeling that if you DO reply, you\'ll be subjected to bombardment of pr0n advertisements and spam. I know of no \"marc hunter\" and certainly didn\'t e-mail him. I bet that he e-mailed everyone on Hotmail, and if you reply, you\'re telling him \"Here I am... now, send me pr0n spam.\"  Good day, all.. I\'m off to college soon and meet up with  urlLink Ruth   '),(20238,3252533,'11,May,2004','2004-05-11 00:00:00','    A poor man in Ireland was plodding along toward home, carrying a huge bag of potatoes. A horse and wagon finally drew alongside him on the road, and the driver invited the man to climb aboard. After getting on the wagon, he sat down but continued to hold the heavy bag.        When the driver suggested that the man set the bag down in the wagon, he replied, \"I don\'t want to trouble you too much, sir. You are giving me the ride already, so I\'ll just carry the potatoes.\"       \"How foolish of him!\" we say. Yet sometimes we do the same thing when we attempt to bear the burdens of our lives in our own strength. No wonder we become weary and overwhelmed with anxiety and fear.       In Psalm 55, David spoke of anxiety he felt because his enemies were attacking him (vv. 1-15). But then he gave his concerns to the Lord and was filled with renewed hope and confidence (vv.16-23). That\'s why he could write, \"Cast your burden on the Lord, and He shall sustain you\" (v. 22)       When we recall the story of the man and his bag of potatoes, remember the simple lesson it illustrates: Rather than trying to bear your burdens by yourself, set them down in God\'s hands.       Give Him each perplexing problem, All your needs to Him make known;     Bring to Him your daily burdens - Never carry them alone!    Just want to share these words   '),(20239,3252533,'10,May,2004','2004-05-10 00:00:00','    Hey, if you want more Sarawak pictures other than Keston\'s Fotki contributions,  urlLink Click here . Its a 1 MB zipped file with about 30 pics taken in Sarawak December 2003.     Here\'s Dan playing the keyboard :D Update more, ok?   '),(20240,3252533,'10,May,2004','2004-05-10 00:00:00','    Well, am in college now. I\'m having a 10:15 - 11:15am break courtesy of the Semiconductor lecturer, Mr Babu. Apparently, some guys from Northumbria, UK are coming over to check the facilities and the programmes here. He\'s one of the main guys of this School of Engineering, so.. he has to be present. Btw, this semiconductor class is the most hardest subject this semester. Thankfully, the others are reasonably easy.  Driving today just improved. More lancar now. I\'m hopeful that by this coming Saturday, I will be allowed to drive to Church for the youth meeting :P woot woot.  Tests are coming up. We\'re almost approaching the middle of the term. Wow, time flies.  I\'m very happy because the blog at  urlLink BMBC Youth  is doing great. We have Jason, Daniel and Vivien on board now, so.. I\'m anticipating an active blog like  urlLink these girls . Maybe by the end of the month, we should have about 10 bloggers. Wheeee!!!  In other news, I\'ll be going mountain climbing (more like a hill) with Aunt Lily tomorrow morning. Class tomorrow only begins at 2pm, so that\'s good. We\'ll wake up at 6:30 and be there by 7:30am. Its gonna be a good exercise as the last time I remebered exercising was....  Anyhow, life has been great as usual. College friends are great, and I\'ve grown closer to a few. Basically, with that, I can get and offer help whenever necessary. Its good to build a network of study partners. Meanwhile, I need to work on my internet addiction. I think I spend close to 20 hours a week surfing the internet and on other net related activities (chat, surf, mail, blog, website management). Hope to cut that down before the exams.   '),(20241,3252533,'10,May,2004','2004-05-10 00:00:00','    Today, May 11th is Adele Poh\'s birthday!   In other news, there will be only 5 \"posts\" on the mainpage. This is to encourage fresh content. So, if you wanna see older posts, just click on the \"archives\". Glad to finally have Jason on. Hope that we all can post some encouraging e-mail and forwards on this blog.  As for the layout, I hope to get a nicer one soon.  urlLink Check this out ... hope to get help from these people :P Also, there is a function to \"Upload\" files, more specifically pictures. So, please resize pictures to MAXIMUM  320x240  pixels or else it would look ugly. Thanks!  So today I\'ll again be driving to college with dad by my side. This will go on for the whole week, then next week I\'ll be alone. Its fun, and abit nervy, but I can cope. Good day and enjoy the new layout!   '),(20242,3252533,'10,May,2004','2004-05-10 00:00:00','    I am in too.. Well, guess I would have one more thing to do every time when I go online starting from tonight, that is to keep updated with you bloggers. I\'m glad all of us pool our resources and time together to form a virtual cafe where we can keep in contact always. Sad :) I have one month time before I leave my hometown, guess it would be a long long time before I\'ll be back in BM after I left.  It would be great to see that this blogging thingy can last, hope it is publisized officially and permanently as our BMBc Youth\'s online cafe. \"Let your light shine before men, that they may see your good deeds and praise your Father in heaven.\" Matthew 5:16   '),(20243,3252533,'10,May,2004','2004-05-10 00:00:00','    Wah.. finally we have 3 members bloggin at BMBC Youth\'s blog! Its exciting! Hope you both update often!! Especially after you leave for Uni. I wonder when the rest will sign up..    Ah Vivien.. so funny =)    Me and Jonster. =)  Gday all!   '),(20244,3252533,'09,May,2004','2004-05-09 00:00:00','    Hey, I\'m in... Hello everyone!   '),(20245,3252533,'09,May,2004','2004-05-09 00:00:00','    Welcome to the blog, Vivien! Glad to have you on board! It would be great if we could invite Jason, Daniel.. and anyone not in BM in 3-4 months time to blog with us. That way, we can continue to keep ourselves updated on things and happenings in our lives.  Well, the Sarawak pictures are up  urlLink here  so do check that out. Otherwise, you still can see Keston\'s Contributions  urlLink here .  I\'m in college now, Learning computer fundamentals. Its really fundamental to me, but the class is going at snail pace.   In other news, I managed to \"hack\" one of our friend\'s e-mail. Its mainly due to weak passwords. SO a word of advice, create tough to guess passwords in the future.. like... ph1Lip90210 or some weird combination! :D Good day all!   '),(20246,3252533,'09,May,2004','2004-05-09 00:00:00','    Hi everyone!! Just testing the blog out.   '),(20247,3252533,'09,May,2004','2004-05-09 00:00:00','    Finally there is some life at piptech.org .. but guess what? They screwed up all my files! They\'re all gone! All my hard work in setting up piptech.org!! GONE!!   I\'m in BM now. I drove mum to The Summit, a shopping mall. Its Mother\'s Day.. and she\'s being pampered at the hair saloon. I\'m now at the Cybercafe lah. Wow.. I drove from church, to makan place and back home.. These are not ordinary big wide streets. These are the streets of ancient Bukit Mertajam.. Where.. there are NO rules. I survived and am glad.  Parking at the basement was thrilling. COntrolling the clutch and brakes. Thank God I didn\'t make any scratch on dad\'s car. My driving is leet. Tomorrow (Monday).. I\'ll drive to college, with dad by my side. Will see how it goes! I\'m overly excited! :P :P  Later, we\'ll be going home, and do some gardening. The orchard at Granny\'s is almost done... and maybe next weekend, we should be getting some durian now... This coming week, college will be exams on Friday and small quiz on Tuesday.  Oh, and church has some plans this weekend. Firstly, we\'re visiting the hospital and doing a dance to entertain the sick. Its all part of our church\'s community services. So.. its gonna be fun hangin\' out with my buddies after Sunday service. As for Sat. youth meeting, I\'ll be worship leading. Phew! Yesterday, I was backup singer.. now I lead? Big step.. but I\'m confident I can deliver! Heheh  Youth is a small group of 20ish. The pastor is Ps. Tamil.. he\'s the Charismatic type. A bit disturbing, but I\'m starting to accept him now. We were all uncomfortable with his style of ministrying to us youths, but I guess we can all compromise. So, next week, worship lead (whooppee).. and that\'s about it.  Gday all. And thanks for staying faithful (despite piptech.org being down).   '),(20248,3252533,'21,May,2004','2004-05-21 00:00:00','    What I read today...  Grant Murphy of Seattle was the active type, a man who ran at full throttle. Idling and coasting were not in his nature. \"One might even call him hyperactive,\" recalled a dear friend.  	The multiple sclerosis began to slow Grant down. First he needed crutches to get around. Then he was limited to sitting in a chair. Finally he was confined to a bed.  	Near the end, he was hardly strong enough to talk. His friend recalls, however, that \"he expressed only joy and thankfulness with a constant anticipation of being in the Lord\'s presence.\" Not long before he died, Grant whispered Romans 15:13 to a friend. He repeated the words \"in believing,\" then added, \"I can\'t do anything now.\"  	It\'s when we can\'t do anything that God does everything. And herein lies a profound paradox of the Christian\'s experience. Faith is simultaneously an exercise of our will and the impartation of divine strength. And from that marvelous mixture spring joy and peace and an abundance of hope.  	Are you in a total helpless situation? Strength gone? All options exhausted? If you have trusted Jesus as your Savior, God will Strengthen you to keep on believing. As you trust Him, He\'ll give you not only joy and peace, but also hope when all hope is gone.  When we are weak and in despair,  Our mighty God is near; He\'ll give us strength and joy and hope, And calm our inner fear.  No one is hopeless whose hope is in God...   '),(20249,3252533,'19,May,2004','2004-05-19 00:00:00','    Psst... Daniel, Vivienne, Jason, Keston where are you guys?   A sudden change of mind  My Dearest Susan,  Sweetie of my heart. I\'ve been so desolate ever since I broke off our engagement. Simply devastated. Won\'t you please consider coming back to me? You hold a place in my heart no other woman can fill. I can never marry another woman quite like you. I need you so much. Won\'t you forgive me and let us make a new beginning? I love you so.  Yours always and truly, John  P.S. Congratulations on you winning the state lottery.   The guide to wife translations   The wife says: You want The wife means: You want  The wife says: We need The wife means: I want  The wife says: It\'s your decision The wife means: The correct decision should be obvious  The wife says: Do what you want The wife means: You\'ll pay for this later  The wife says: We need to talk The wife means: I need to complain  The wife says: Sure... go ahead The wife means: I don\'t want you to  The wife says: I\'n not upset The wife means: Of course I\'m upset you moron  The wife says: You\'re ... so manly The wife means: You need a shave and sweat a lot  The wife says: Be romantic, turn out the lights The wife means: I have flabby thighs.  The wife says: This kitchen is so inconvenient The wife means: I want a new house.  The wife says: I want new curtains. The wife means: Also carpeting, furniture, and wallpaper!  The wife says: I need wedding shoes. The wife means: The other forty pairs are the wrong shade of white.  The wife says: Hang the picture there The wife means: No, I mean hang it there!  The wife says: I heard a noise The wife means: I noticed you were almost asleep.  The wife says: Do you love me? The wife means: I\'m going to ask for something expensive.  The wife says: How much do you love me? The wife means: I did something today you\'re not going to like.  The wife says: I\'ll be ready in a minute. The wife means: Kick off your shoes and take an hour nap.  The wife says: Am I fat? The wife means: Tell me I\'m beautiful.  The wife says: You have to learn to communicate. The wife means: Just agree with me.  The wife says: Are you listening to me? The wife means: [Too late, your doomed.]  The wife says: Yes The wife means: No  The wife says: No The wife means: No  The wife says: Maybe The wife means: No  The wife says: I\'m sorry The wife means: You\'ll be sorry  The wife says: Do you like this recipe? The wife means: You better get used to it  The wife says: All we\'re going to buy is a soap dish The wife means: I\'m coming back with enough to fill this place.  The wife says: Was that the baby? The wife means: Get out of bed and walk him  The wife says: I\'m not yelling! The wife means: Yes I am! I think this is important!  In answer to the question \"What\'s wrong?\"  The wife says: The same old thing. The wife means: Nothing.  The wife says: Nothing. The wife means: Everything.  The wife says: Nothing, really. The wife means: It\'s just that you\'re an idiot.  The wife says: I don\'t want to talk about it. The wife means: I\'m still building up steam.   '),(20250,3252533,'19,May,2004','2004-05-19 00:00:00','    I\'ve made a diagram of two friendships that I\'ve just \"made\"...  One, her name is Amy Lee from Sarawak and the other is collegemate, Ruth.      Its a small world, I tell ya! Oh, and thanks  urlLink Ruth for registering  at my  urlLink Pipboard   '),(20251,3252533,'18,May,2004','2004-05-18 00:00:00','    Youth forums is now available at  urlLink Piptech.org/pipboard . Sorry for sounding so vain.. :) but if you like, you can click the link on the top of the sidebar on the right.  You can post anonymously, but I recommend you to create a username so that other people cannot copy your username and post false stuff.. Enjoy the forums.  Keston, wishing you a speedy recovery. Nice pics u got there!  And welcome, Radiance!   '),(20252,3252533,'18,May,2004','2004-05-18 00:00:00','    minor op with 2 stitches thank God everything went fine i\'m all right and discovering well !   '),(20253,3252533,'17,May,2004','2004-05-17 00:00:00','    Hellooo everyone... Sorry for not blogging for a few days already. My com seem to have a problem. Currently using my mum\'s laptop. So, there is no guarantee whether I\'ll blog in a few days time or not. Sorry again! Yesterday was a good day, wasn\'t it? Visitation to the hospital.. to give some cheer to the ppl there..YOu guys sure did a good job for the Josh Groban\'s song. Now I know there sure is, many budding artists around. Anyone went for the Malaysian idol\'s audition in PISA ?  Yes, Philip..got your sms about the Brunei Prince news at 2 something in the morning..   '),(20254,3252533,'16,May,2004','2004-05-16 00:00:00','    Phew.. I made it to college without a scratch. The parking system is abit pathetic.. I\'ve gotta incert coins less than 50 sen in denomination to pay for the RM 2 parking. IT was a PAIN inserting so many coins.... and my hands are short!! I reached class at 9:40am & was 10 minutes late, but no worries.. no one\'s punctual!  I\'ve uploaded some MP3s using my college line for all to enjoy!  urlLink Not your Average Thug 3,924 KB   urlLink Linkin Park - Session 3,389 KB   urlLink Eminem - The Kids 4,768 KB   urlLink Dang Ni (Chinese) 3,178 KB   urlLink Hillsongs Australia - Touching Heaven Changing Earth 3,617 KB   Life rawks - cuz I\'m finally driving. Yeah, and I SMSed a few guys at 2:30am just asking them to visit piptech! I finally slept at 3:15am.. and I\'m still quite awake as of now!   '),(20255,3252533,'16,May,2004','2004-05-16 00:00:00','    Just want to share what I read...  In his book Christians in the Marketplace, Bill Hybels says that people outside the faith often say, â€œShow meâ€ before they say, â€œTell me.â€  	I knew a young man in Germany named Wolfgang who modeled Hybelsâ€™ principle at a building site where he worked. As an enthusiastic believer, Wolfgang always read his Bible during lunch. Though his fellow workers jeered, he didnâ€™t stop his daily reading. He simply prayed for a way to demonstrate Christâ€™s love for them.  	When the workers went home at night, they always left their muddy boots behind. Wolfgang began staying late after work to clean their boots. The men were puzzled at first but then realized that Wolfgang was the only one among them who would perform this humble service. Not only did they come to respect him, but sometimes they even asked him to read the Bible to them. Only eternity will reveal the full effect of Wolfgangâ€™s shining life. But this we know: When his co-workers saw his good works, they started listening to his God.   	Jesus said, â€œLet your light so shine before men, that they may... glorify your father in heavenâ€ (Matthew 5:16). If you long to lead the people around you to Jesus, radiate His Love by doing practical deeds for Godâ€™s glory alone.  My life was dark until the Light shone in, That Light was Christ, who saved me from my sins; His light that Iâ€™ve received I long to share In loving deeds for people everywhere.  A Christianâ€™s life is a window  through which others can see Jesus.   '),(20256,3252533,'16,May,2004','2004-05-16 00:00:00','    Well... a surprising news. A 30 year old soon-to-be KING marrying a 17 year old girl that  urlLink my friend, Fui  used to play \"Barbie Doll\" with. Wow.. =) I\'m shocked.  Congratulations to Prince Billah & Dk. Sarah binte Pengiran Haji Salleh Abdul Rahman. They\'re next in line to the throne.   urlLink from Brudirect.com   In other news, worship leading during youth meeting went well. We had a short 15 minute discussion about Conflicts. I\'ll type out the text when I have time. Very interesting. Anyhow, you can always search Google on conflicts.. :) Its good to know these sort of things.  The English premier league is over, and Ronaldo and Fletcher will be missing the final due to the red cards receieved during the last game of the season. That sucks. But I\'m sure we can beat Milwall this coming Saturday, 9pm on Astro Supersports ;)  Sunday came and went. Nothing unusual.. We went to granny\'s orchard. We ate some durians and now I\'m home. I think there\'s a test this week, now.. which subject it is..   '),(20257,3252533,'16,May,2004','2004-05-16 00:00:00','    Youth Meeting      Youth meeting, Saturday nights     Ben and Keston at youth meeting.     Our charismatic enthusiastic Pastor Tamil (youth)     BMBC Youth   Mothers\' Day      The BMBC kids during mothers\' day.     And more..     And more..   General pics      Shot of the church I attend in  urlLink Bukit Mertajam      Inside church     Bahasa congregation pastor, Pastor Jack with his \"axe\"!     Uncle Joel Lim     May 16th, Pr. Philip\'s daughters singing     Tenang cell group.. my home in BM     The coffee area after Sunday Service     Uncle Chew and Aunt Kee Kee     Fellowship after Sunday service   Pip related      My mate from Hotel and Tourism, Kenneth.. He\'s a GREAT chef.. gonna work @ Queen Mary *(cruise ship)     Closeup of Kenneth... He\'s Chindian..     College Cafeteria     My site\'s new Logo   Cool pics on the Net      How low can a plane go?     Another one..   '),(20258,3252533,'16,May,2004','2004-05-16 00:00:00','    Let your light so shine before men,  that they may see your good works and  glorify your Father in heaven.  Matthew 5:16  Live Life To The Fullest!!! Treasure Your Life. You Only Have One. Make A Difference!   '),(20259,3252533,'16,May,2004','2004-05-16 00:00:00','    Since I\'ll be back less than a month from now.. I guess it\'s best that I can keep in contact with you bros and sis in Christ. Might not like the idea of coming back one day and found that I am lost, so lost.. like what happened to me when I first \'landed\' in BMBC a month ago. Thank God, the Youths and church camp helped a lot to bring me back closer to you.   ICQ: 15558221 MSN: jasonlim83@hotmail.com  Friendster: jasonlim@mail.com Primary mailbox: jasonlim@mail.com   Handphone no.: 019-4447132  Birthday: 03-07-1983    That\'s all for now. Sayonara   '),(20260,3252533,'29,May,2004','2004-05-29 00:00:00','    God sure knows how to teach His children.  Just now, at class around 11am, the Microapp teacher was teaching us to use Excel.. and asked us to draw a plan (diagram) using Excel. It was as fun as eating sand! Well, after finishing, she inspected my work, and concluded that it wasn\'t complete.  With a sour face, and an inflated ego, I gave her the middle finger salute. Had a giggle with my friends, and yes... that WAS satisfying.  Driving home, near the \"underpass\"... I wanted to go to the left lane in order to reach home. After signalling for 10 secs, still they don\'t give way. Suddenly, I look in front and I was so close to the Merz\'s bumper.. I swirved left, and nearly hit this motor cyclist. Asshole gave me the salute!  Haha.. isn\'t it amazing how God works?  Oh, I won\'t be going to Youth Meet today. Dad and mum have Malay wedding reception to attend. Oh well, guess I\'ll continue my Shrek 2 in CANTONESE, Confessions of hot chic and of course, Troy.. VCD galore.. and of course, study for my Semiconductor exams!   '),(20261,3252533,'28,May,2004','2004-05-28 00:00:00','    Today after college, at 3pm ish.. I drove dad and mum to BM. This is the first time me driving across the famous Penang bridge. I nearly met an accident as there was a sudden \"brake\" near the Toll Plaza.. emergency braked.. luckily the air bags didn\'t pop up, or I get killed by the air bags!! lol..  it was frightening (for dad and mum).. I was reasonably cool and calm, altho abit shaken lah.. Overall, it thought me the \"peraturan dua saat\" I learnt once upon a time. Other than that, the whole journey to BM went smooth, I\'m now l33t.  Talking about l33t.... Liang, Neil (collegemates) and I were on MSN talking l33t. Leet talk (or elite talk) is a language commonly used by hackers so that their messages cannot be intercepted or searched using a search engine. If I were to tell you \"i\'ll be breaking into microsoft june 1st\".. then, the FBI or whatever can Google these terms and find my website (gosh, no.. I don\'t have the experties to hack MS).  So, h4ck3r$ t4lK 1!k3 tH1$... and u$3 t3Rm$ l!k3 ph34r, l33t, p4wned (search around the web, there should be a leet dictionary somewhere). Its a fun language to learn and use while MSNing.. Rather than lousy SMS lingo.  Hi to cousin Li Tinh, Ruth, Jen, Radiance and all da nice people who managed to drop by piptech.org (through the bulletin board). Sorry for inconvenicen.. Just wanting them to post something in the Forums b4 arriving here.  =0  Neway, congrats to Vivi for getting Biochem, Danny for getting Dentistry (hey, its a blessing too.. u may earn your million in 5-10 years.. if u work hard!) also to Andreen (Chinese lit, whoa.. respect dude.. everytime I *try* to speak Manglish, u sure / can laugh at me... cuz I speak like ang-mo!) and to Jon.. whatever course u might be taking.. all da best!  =)  Dad and I watched the American Idol replay, the Finals on 8TV.... first song for Diana deGemuk was great, 2nd one abit lousy and 3rd one was a disaster. She was intimidated by Fantasia\'s wonderful performance.... Fantasia didn\'t start well, but her 2nd and 3rd song was out of this world. I\'ll buy her album when it comes ;) Remember Ruben Studdord, Clay Aiken and that Guy Sebastian (yea, Rach, I know u like him).  Listening to Bone Thugs, and D12. Gnite all, college at 9am tomorrow, hence 8am have to leave home, hence 7am have to wake up. And as usual, I\'ll be mia, afk (missing in action, away from keyboard) for the whole of Saturday and Sunday in Bukit Mertajam. ;-) Thanks for visiting!   '),(20262,3252533,'27,May,2004','2004-05-27 00:00:00','    Heyya !!! Where\'s the good old bmbc.info/youth ??   '),(20263,3252533,'27,May,2004','2004-05-27 00:00:00','    Well, yesterday was eventful. After the usual 2-5:30 pm date with Ms Voon, the Engineering Maths teacher, I headed for Komtar.. it was abit of a crawl.. but eventually, I found the BSN (bank simpanan).. the place where the badminton was on.  It was only about 6pm, so I went to Komtar there and Cybercafe-d.. Wasted RM 3 for an hour\'s of access, but what the heck :P At around 6:45pm, went for makan. I had yin yong.. which was excellent - at New Lane area.... 7:10pm, met the gang. And I was surprised!  Apparently, almost all of them are from Engineering. KDU Penang has 6-7 schools, but School of Engineering was the dominant one.. A few ppl I know, they\'re my senior. It was fun stuff, really.. cuz all of them played very well.. Even the girls were skilled. I had loads of fun, and oh.. I beat them all :P [brags]  Next Wed / Thu, I\'ll probably have supper with them. They usually have a gathering after the 9pm game... so that\'s kinda fun. Today, I\'ll skip badminton, my arms hurt badly... Its been 1/2 a year since I last played, see? But I was still reasonably at the top of my form. Wheee =)  Later that night, wrote some testimonials on Friendster and had fun ;) Nothing much happened, and I slept a bit late lah... 2:45am, my alarm woke me up. I woke up, shut it off.. and went back to bed. FC Porto and Monaco don\'t excite me. Probably the worst finals in Champions\' League History! Ultimately, the class of Porto showed, and they won 3-0. Good for them... screw their coach! (who is apparently going to Chelsea next season).  Woke up as usual this morning... abit late, preparing my assignment (as below).. then 9:25am, went to the school computer lab to have it printed... Still got 50 pages in my Printing quota :-D must use puas puas! I\'m staying back today for BM classes later with Puan Suraya... The cool teacha.. =) Gonna meet my buddies from H&T.. and have loadsa fun learning about the history of the Malay language, and how to improve my language in general.  5:30pm, shud reach home. Plan to sleep well, and rest my right arm... Still kinda hurts, and probably will read up my Semiconductor and Telecommunications and do some Maths. My Maths sucked. Only 8 marks out of 15.. and its all on Surds and rational numbers nia.. me so disgraceful. I can justify that my buddies also got around that region, highest being 8.75 marks nia.. but hey, gotta do better!!!  Man U has signed Alan Smith for 7 mil pounds. Well, he\'s a good buy, except his temper. He\'s known to get angered very easy. Surely will get lotsa red card wan, and tunnel burst up with his own team mate or manager. He\'s a tough guy, so.. hope he can score and provide for Man United in 90 days time. Till then, its boredom 100%!!! No footie, only Euro 2004.... But that\'s equally good too.  Next weekend, Olivia & Anthony are coming to Penang for a wedding, and on Sunday, dad and mum are going to KL for a wedding... So, they carpool and dad and mum take bus back to Penang lah. I\'ll be driving them, so that\'s l33t.  I\'ve been starving myself the past few days.. eating less and all. Hopefully, my waist can reduce sikit to less then 36\' lah... (now its around 37\') and my weight, target: 160 pounds, 70kg... (now at 165 pounds oredi). I\'ve been badmintoning abit, so my leg muscles are leet now... So firm. =) Sigh, how I wish my whole body was as mascular!  Do post in the  urlLink Forums / Community Discussion Boards . Start a topic, reply to a post.. just say anything that\'s on your mind. ALL ARE INVITED and your post will make me smile ;-)   Thank you.   '),(20264,3252533,'26,May,2004','2004-05-26 00:00:00','    AS long as Radiance dun post (or is unable to post for some particular reason).. I shall be providing the visitors with \"entertainment!\"  Meanwhile, u can always visit the Forums at http://piptech.org/pipboard .. there\'s a BMBC youth section. I recommend everyone to register and post!! Lets tok!!!         enjoy!   '),(20265,3252533,'25,May,2004','2004-05-25 00:00:00','    Pip i think ur diagram should go as above it\'s really a small world   '),(20266,3252533,'25,May,2004','2004-05-25 00:00:00','    yupp, that\'s the site to check your application status. Do share the good news once u found out where u r heading to.. ;) all da best..  gonna miss all u guys wen u leave. sigh~   '),(20267,3252533,'25,May,2004','2004-05-25 00:00:00','    Today, as I was browsing the net, doing the usual online stuff, and of course the occasional thought about my Uni application feedback. This verse comes to mind :   29 He gives power to the weak, And to those who have no might He increases strength. 30 Even the youths shall faint and be weary, And the young men shall utterly fall, 31 But those who wait on the Lord Shall renew their strength; They shall mount up with wings like eagles, They shall run and not be weary, They shall walk and not faint.  ISAIAH 40:29-31  Wassup everyone?Hope you guys are doing great. As for me, things are just the usual..Hoping and Praying for the best... God bless ya everyone alwiz !!!   '),(20268,3252533,'25,May,2004','2004-05-25 00:00:00','    let a = 1 and b = 1 a = b a2 = ab a2 - b2 = ab - b2 (a+b)(a-b)=b(a-b) a+b=b b+b=b 2b = b 2 = 1 !  Huh? 2 equals 1?   '),(20269,3252533,'24,May,2004','2004-05-24 00:00:00','    Well, quite an uneventful weekend. Sunday, we went to granny\'s to grab some durians, 10 in all. Later there\'s a Computer Fundamentals exams.. hope I\'ll do decent! Church was great, this pastor spoke about prayer... very powerful message.. and hope that I can change, and spend more time talking with God.  I got my hair cut, its now abit short, I\'ll post pics this afternoon.   '),(20270,3252533,'23,May,2004','2004-05-23 00:00:00','    Woohoo.. Man U 3 - Milwall 0, FA Cup finals. We are the champions.  Congrats to May Lynn on passing her driving test. Piece of cake, wasn\'t it? Now for Form 6.. tough years ahead... wishing you all the best! Study hard ;-) ain\'t like form 5 no mo...   '),(20271,3252533,'03,June,2004','2004-06-03 00:00:00','    DeGemuk? So harsh on her, anyway she doesn\'t understand!!  Let me clarify It wasn\'t all because of \"symphaty votes\" the fact is LaToya London\'s hometown had a power failure so the phone wasn\'t working...... I read bout it from The Star few weeks ago  those hum-sup lous should cast their vote for the Miss Universe contest, not the American Idol! They ruin the whole show!!   '),(20272,3252533,'02,June,2004','2004-06-02 00:00:00','    Opinion deleted by request of WEBMASTER (Mr. Morgan)   '),(20273,3252533,'01,June,2004','2004-06-01 00:00:00','    deleted by request of WEBMASTER (Mr. Morgan)   '),(20274,3252533,'01,June,2004','2004-06-01 00:00:00','    Deep in a Wyoming canyon I came across the biggest porcupine I\'ve ever seen. As he lumbered toward me, I watched him closely and gave him plenty of room. I was not about to get near a guy whose quills looked like missiles. No wonder he was alone!  But he\'s not alone all the time. Every November and December, porcupines get close enough to produce offspring. During that time they choose to relax their quills, then they return to their prickly selves.  In nearly every church, there will be porcupine or two, with sharp quills of criticism or sarcasm or arrogance. We want to avoid them, but God places us in communities of believers for fellowship. He commands us to love one another - including the porcupine types. And in our honest moments, we have to admit that we have quills too.   John wrote, \"He who loves God must love his brother also\" (1 John 4:21). To do this, we need to ask God to help us \"relax our quills,\" even when other people are prickly. The Holy Spirit will help us stop being so defensive or critical or controlling, and enable us to love our Christian brothers and sisters. It\'s the way we show the world that we love God (John 13:35).  Some people can be difficult to love  And so we do not even try to care, But God says, \"Love them just as I\'ve loved you You\'ll bring Me glory as My love you share.\"  God loves you and me - let\'s love each other!   '),(20275,3252533,'01,June,2004','2004-06-01 00:00:00','    Ya Some say we \"halau\" Pastor !!!! Some say it\'s not entirely our wrong !!!!(Just some conflict Pastor can\'t handle !!!!)  For me ( i was there that day ) it\'s misunderstanding that spark the flame !!!  Let\'s forget it \"if we can\" and learn from it !!!  Look at 3 of our \"most wanted\" weaknesses 1) not punctual = late 2) come empty handed = without notes even some without bible 3) low attendance = very very low (maybe low tide) others are the worship .... etc....  Hoping to see everyone in the coming youth meeting Maybe without the usual Pastor Tamil but i got latest news that negotiation is going on .....  Bye C ya   '),(20276,3252533,'31,May,2004','2004-05-31 00:00:00','    Enjoy these 39 pictures. Weighing in at 600K, it\'ll take a couple of minutes for 56k-ers.     Looking up from my 15th floor     Breathtaking views from above     And more...     And more...     The guys at BM remidial class     The wedding I attended     More...     It was last Saturday     That\'s why I missed Youth Meet     I guess I\'ve missed \"THE ACTION\" back in youth meet     Things didn\'t went well, I heard     Last pic of the wedding =) Dad\'s friend\'s daughter     The cars. Sweet.     BMBC Tree of blockage.     The sign     The church     The dam     Mengkuang Dam     Keston     Mengkuang Dam     More shots of hometown, Bukit Mertajam     The gradient is making me look stunted!     Worse case of camera shyness :P     Jason and Andreen     Climbing the STEEP slope at Mengkuang!     Chillin\'     Andreen & Juliana     Dang.. I love Zoom! (This pic taken w/out permission)     The guys     The elderly having fun too!     Multi-talented girls =)     Adults havin their fun!     Having a break!     Uncle Joel     Games     Timeout!     Jacob the rapper... Vivi promoting peace =)     Picture of the Year... Kes acting funny!     Singing. =)   '),(20277,3252533,'31,May,2004','2004-05-31 00:00:00','    Okay, I admit it, I was lost whenever i type in the youth blog add. I was always directed to piptech something..the forum i think..And I can\'t seem to access the youth blog. My mistake was, i never log in to blogger.com. Haihs, getting so lousy at this,, cause i\'m use to looking straight for the page blog 1st....   '),(20278,3252533,'11,June,2004','2004-06-11 00:00:00','    Don\'t worry i\'m one big Malaysian Idol fan  here are some Malaysia Idol Wannabe (no William Hung though)   '),(20279,3252533,'10,June,2004','2004-06-10 00:00:00','    wat happen DAT DAY click below  urlLink KESTON\'s D-DAY   '),(20280,3252533,'09,June,2004','2004-06-09 00:00:00','    Hey guyssssss!!! Finally I can blog! Woooohoo! Thanks Philip! :)   '),(20281,3252533,'09,June,2004','2004-06-09 00:00:00','    Dear Youths Of BMBC,  It\'s been GREAT growing up and learning from all of you... I truly appreciate all the good times and bad times we have all gone through together as BMBC Youths. No matter what happens we know that God will always deliver us through each hurdle in our lives... The times that we have spent and shared, I will cherish it... Thank you all and may God bless all of you in everything that you do.  Until the time we meet again, I\'ll do my best to keep you all updated and in touch with me while I\'m in Kuala Lumpur pursuing my studies and my dream. I\'m going to miss all of you and the whole congregation of BMBC as a whole.   Take care and remember to \"Puji Tuhan\" always!!!  Bye for now.   '),(20282,3252533,'09,June,2004','2004-06-09 00:00:00','    WHAT IS THIS ABOUT?  Love. Forgiveness. Mercy. Grace. Goodness. Compassion. Kindness. Generosity.  Who in the world would be against these characteristics? Who could oppose such positive forces in a person\'s life? In other words, who could call Christians nasty names, consider them dangerous, and seek to put as much distance as possible between themselves and Christians?  Who? Many people. But why? It\'s because in the Christians they already know, they do not observe love, forgiveness, compassion, goodness, kindness, mercy, grace, and generosity.  They won\'t see Christ and His attributes if we\'re standing in the way. If we have a self-serving agenda or cause, they\'ll see our hate instead of our love, our grudges instead of our forgiveness, our indifference instead of our compassion, our harshness instead of our goodness. Our lives need to be consistent with the message we proclaim.  Christianity is about Christ and His loving, forgiving, and compassionate gift of eternal life. Anything we do that reflects something else gives others the wrong idea.  It\'s all about Jesus. Let Him shine through your life so that others can see Him. Show me the way, Lord, let my light shine As an example of good to mankind; Helps them to see the patterns of Thee, Shining in beauty, lived out in me.  Live So That Others Will Want To Know Our Savior.   '),(20283,3252533,'09,June,2004','2004-06-09 00:00:00','    Okay, Radiance can join the blog now. Thanks to pip, of course =) rad, check your mui_wan e-mail for instructions how to post!   '),(20284,3252533,'09,June,2004','2004-06-09 00:00:00','    I read the mail that Anonymous sent, and I\'ve something to say.    Well,I have a few comments for the BMBC youth !! I wonder how pastors have this church halau.I think it would set a new record in The Malaysia Book of Records. It\'s kinda terrible how you ppl treat pastors. If you ppl do not like how the way pastors preach, I guess you\'re to stuck up to admit your mistakes.And the worshippers or musicians who were ambarassed by Ps.Tamil , i think they should brush up on their musical skills if they wouldn\'t want to look like real idiots again.  Well, i guess there\'s nothing more to say accept halau-ing pastors are very wrong and very rude.Especially saying they suck.If you\'re that clever, why don\'t you try going up yourself and preach a message as bitingly good as theirs!!!!  So there!!!!    First of all, whoever you are, I hope you\'re reading this. I appreciate your concern for the youth group, and with you wanting to tell us what you think about the youth group is good, but please, if you want to do so, do it by also including your name, and your info. No, you cannot just send comments from your side of view/story and not allow us to defend ourselves and enlighten you with our side of the story. Don\'t you think you\'re being unfair? Like a kid who refuses to listen to anyone, stubborn & stone-hearted, not willing to listen to others.   Quote:  I wonder how pastors have this church halau.I think it would set a new record in The Malaysia Book of Records.    From that sentence, you don\'t seem to be one of the youths. Not someone who belongs to the church. If not, wouldn\'t you have known if there was a case of \"halau-ing pastor\" as you call it? But anyway, let me tell you, I\'ve been in BMBC all my life, and NO, we\'ve not chased any pastors away before. Don\'t be silly, our pastors are voted, discussed, put on probation before hiring, if we didn\'t like them, we can just NOT vote for them. Oh, you think it\'ll set a new record in The Malaysia Book of Records? Then I think you can also join their \"list-of-cowards\" for not having the courage to even tell us your name. Anonymous? HAH!   Quote:  It\'s kinda terrible how you ppl treat pastors. If you ppl do not like how the way pastors preach, I guess you\'re to stuck up to admit your mistakes.   Please, enlighten us. How do we treat pastors? You said pastors, plural. So it means we treat pastors terribly. If you have the guts, name the pastors and tell me how badly we treat them. Thanks. So, if WE don\'t like the way pastors preach, it means we\'re too stuck-up to admit our mistakes? Ah, what a nice definition. I wonder who taught you that? God? Are you saying that pastors don\'t make mistakes? Pastors are holy & righteous that they don\'t make mistakes? Let me tell you something. Nobody is perfect, and yes, pastors are not perfect either. Different pastors have different ways of preaching and YES, sometimes the way they preach just doesn\'t suit us, so we tend to dislike the pastor. It doesn\'t mean that we don\'t admit mistakes. I\'ve met so many pastors & I\'ve heard many of them speak, tell me why do I love some of them so much? It\'s just the difference of mankind. Don\'t be unreasonable.   Quote:  And the worshippers or musicians who were ambarassed by Ps.Tamil , i think they should brush up on their musical skills if they wouldn\'t want to look like real idiots again.   I agree with Philip on this. Why must we, avoid looking like real idiots, just because you say so, or Pastor Tamil says so? It doesn\'t matter how good we are, or how terrible we are. We are singing praises to our Lord, and that\'s all it matters. Our focus is God, not pleasing the audience. It doesn\'t matter if BMBC Youth are playing, or Altered Frequency, or Planet Shakers or Hillsongs. The one and only reason is to worship God. If God is unhappy with the worship, let Him deal with us. Who are you to say anything? If you don\'t like it, why don\'t you come and lead worship someday? I\'ll be glad to give you a chance. Just let me know.   Quote:  Well, i guess there\'s nothing more to say accept halau-ing pastors are very wrong and very rude.   I totally agree. But who are you talking about? As far as I\'m concerned, the youths of BMBC did not chase away Pastor Tamil. No, we didn\'t hold brooms & asked him to go away. No, we didn\'t say we didn\'t want him to come to our meets anymore. We did no such thing. Get your facts straight. Who said we HALAU-ED pastor?   Quote:  Especially saying they suck.    That was a mistake. He apologized, didn\'t he? The Bible teaches us to be forgiving, right? So what are you doing now? Keeping a grudge?   Quote:  If you\'re that clever, why don\'t you try going up yourself and preach a message as bitingly good as theirs!!!!   Once again, you\'re using plural. Who are THEY? Well, YOU think it\'s bitingly good. I don\'t. Sue me for having a different opinion. You\'re talking as if we can\'t have different views about other people. Are you just being totally unreasonable or are you trying to be God?   Yup. That\'s my opinion. Don\'t like it? Well, I don\'t care. Come talk to me face-to-face if you have the guts, and we\'ll talk it out. I\'m Radiance Leong, in case you don\'t know. I\'m the Youth Leader of the youth group, you have probs with us? Come talk to me. I\'m always open for opinions. -peace- God bless!   '),(20285,3252533,'08,June,2004','2004-06-08 00:00:00','    My server at Cybexhost experienced a breif downtime.. Brief enough for my friend  urlLink Han Jen  to notice and SMS me! Lol... Actually, they\'re moving servers and I hope its for the better! :P  All is well now.  I just went to the dentist just now, and did scaling.. My teeth are still abit yellow, but its an improvement. Dad did scaling. Now parents doing gardening, i\'m out, gonna have dinner soon before dad and mum goes to granny\'s to collect durian while I study for the Math test tomorrow.  Life\'s great, as usual.. Check out the  urlLink my website, piptech.org  and of course,  urlLink Radiance\'s Blog . Vivi\'s blog is a tad too tough, but it should be somthing xanga something vivicious :P ask her!   '),(20286,3252533,'07,June,2004','2004-06-07 00:00:00','    So call \"Anonymous\" !!! After reading Philip\'s blog , the \"Anonymous\" looks more like a coward to me !!! I wonder it\'s an adult or a youth or someone who is just taking advantage of this situation !!! Sending a comment and expressing your views bout the youth on this site will certainly help but not many youths suft the net !!! Can u please deal directly with Sir \"Chief\" Morgan , he\'s the \"Chief\" so u should tell him so that he can address to us about your comments !!! By hook or by crook , we are still One In Christ (OIC) !!! Once again \"Walk the Talk\"  Pls note, i\'m Keston a.k.a. \"z.OR.ro_VIR.a.GO\" at least i\'m not someone \"Anonymous\" my viewpoints are mine and only my own viewpoints !!! No offence intended !!! comments with \"contact number, name or e-mail address or at least a \"nickname\"\" are welcome my mail -> kestoncvj@hotmail.com   '),(20287,3252533,'07,June,2004','2004-06-07 00:00:00','    Here\'s what I got from an e-mail... was meant to be a comment to one of the postings of BMBC\'s Youth.    Well,I have a few comments for the BMBC youth !! I wonder how pastors have this church halau.I think it would set a new record in The Malaysia Book of Records. It\'s kinda terrible how you ppl treat pastors. If you ppl do not like how the way pastors preach, I guess you\'re to stuck up to admit your mistakes.And the worshippers or musicians who were ambarassed by Ps.Tamil , i think they should brush up on their musical skills if they wouldn\'t want to look like real idiots again. Well, i guess there\'s nothing more to say accept halau-ing pastors are very wrong and very rude.Especially saying they suck. If you\'re that clever, why don\'t you try going up yourself and preach a message as bitingly good as theirs!!!! So there!!!!    First off, I would like to thank \"Anonymous poster\" for his or her comments.  \"I wonder how pastors have this church halau\". Sorry, please rephrase.  \"I think it would set a new record in The Malaysia Book of Records.\" Well, if you think so, then, go tell them to make this new category.  \"It\'s kinda terrible how you ppl treat pastors.\" How do we treat them? I dunno myself. Tell me about it! Btw, do you come to the Youth Meets?  \"I guess you\'re to stuck up to admit your mistakes\" What mistakes? Tell us!  \"And the worshippers or musicians who were ambarassed by Ps.Tamil , i think they should brush up on their musical skills if they wouldn\'t want to look like real idiots again.\" Its EMbarrased. My personal opinion, and I might be wrong is that BMBC\'s musical talents are the finest in Bukit Mertajam.. I couldn\'t wish for any better. And who are you to comment on the Youth\'s musical skills? This is worshipping God! If God is unpleased with our worship or musicians, let Him deal with it. You shut up.  \"Well, i guess there\'s nothing more to say accept halau-ing pastors are very wrong and very rude\" No doubt you\'re right here!  \"Especially saying they suck\" Well, I take that back. Sorry. Apparently, Malaysia is not a free world after all. No place to express one\'s personal opinion. Btw, the church hasn\'t pay me my RM 250 for bmbc.info and the hosting space.. Hmm.. Mr. Morgan?  \"If you\'re that clever, why don\'t you try going up yourself and preach a message as bitingly good as theirs!!!!\" I would. I really would. Never did anyone say that he/she is more clever than the pastor. But as human with senses, we know an [insert favourite adjective here] when we see one.  \"So there!!!!\" What\'s your point?  Oh, and anonymous, next time, when posting, at least include a contact number, name or e-mail address or at least a \"nickname\" so I can address you properly!  Its no point listing down all Ps. Tamil\'s weaknesses and our \"beef\" with this dude. For He is God\'s servant. So.. yes, you\'re right in stating that we need to respect him. Like I said earlier, sometimes, us DEMANDING youths need someone we can relate to. And if one or two have a problem with a particular \"person\", then, its an isolated case. If almost everyone have differences with an individual, then, I don\'t see any sin in evaluating and acting on this individual.  Good night =)   '),(20288,3252533,'18,June,2004','2004-06-18 00:00:00','    Friday June 18 2004 Jit Sin teacher\'s day!  The highlight of the day is undoubtedly The Ever Ready Teachers versus The Kaki Bangkus Prefects in a entertaining high scoring match! The match was an one sided affair and the final score, an embarrassing 5-0 lose for the prefects! Even our own church leader Sir Ravi Samuel got into the act, scoring a brace and with his commanding defensive role,the students voted him the Man Of The Match. However there\'s no prize for the winner nor the Man of the Match.   '),(20289,3252533,'16,June,2004','2004-06-16 00:00:00','    Taken using the Hubble Space Telescope   '),(20290,3252533,'14,June,2004','2004-06-14 00:00:00','    The forums have been updated abit. Its easier to decide where to post to. Give it a try today!  urlLink Click here & register   '),(20291,3252533,'14,June,2004','2004-06-14 00:00:00','    Contraversies aside, Ps Tamil will make an appearance this weekend. See how it goes! Pls turn up, or else negative things will arise again. Try to be punctual, 7:45pm please :P  Good day, all!   '),(20292,3252533,'25,June,2004','2004-06-25 00:00:00','    Haha England lost!   '),(20293,3252533,'23,June,2004','2004-06-23 00:00:00','    http://img.piptech.org/  That\'s right, just sign up and get your pictures hosted. Come on now, share with us your favourite shots!   '),(20294,3252533,'23,June,2004','2004-06-23 00:00:00','    Pip , this site is getting deserted without the shoutbox !!!   '),(20295,3252533,'23,June,2004','2004-06-23 00:00:00','    Hey there!  How are you all doing? Fine I hope... Now doing my blogging in the new Dental Faculty Library. Really cool.   To update on myself... I\'ve just finished my orientation. Had a terrible week with only 3 hours of sleep daily. I\'m glad it\'s over. Puji Tuhan!!! I saw the Blog and I think it\'s cool and better in an unique way.   Oklah, gotta go and do more stuffs... Lots of assignment. Take care you all. God Bless!!! Hope to see you all soon. If I have a long break...  Bubye!   '),(20296,3252533,'23,June,2004','2004-06-23 00:00:00','    For those with broadband, check out the goals :)   urlLink http://piptech.org/euro/    urlLink Life is great .. nothing out of the ordinary happened of late.. And BMBC youth blog isn\'t really active, to say the least!  This weekend, hope to learn something from the Youth Meet... I mean, over the past few weeks, too many distruptions oredi.. And looking forward to the bible study @ 9am also..  Well, portugal play england on friday morning 3am... Mustn\'t miss it! Set your alarms, everyone! I\'m backing England (naturally)... and I really think we can beat \'em!  Other than footie, later I\'ll be playing badminton. I play Wed and Thu nites with my collegemates. They\'re good players, so, it kinda gets competitive. Usually reach home at 10:30 - 11pm lah.   '),(20297,3252533,'22,June,2004','2004-06-22 00:00:00','    Peace.   '),(20298,3252533,'21,June,2004','2004-06-21 00:00:00','    Here\'s the BMBC Youth\'s Parents\' Day Dance video. Pardon the shakeyness, but I assure you, its worth the 6 MB download.  (takes 20 minutes on dialup or 3 minutes on Streamyx)   urlLink Right click, then Save File As..   You probably need  urlLink QuickTime 6  to view this file.  Argh.. 15 minutes to go b4 lesson starts! Now @ lab uploading this video...  Have several nice days! Plugs to:  urlLink Radiance ,  urlLink Mei Yun ,  urlLink Daniel ,  urlLink Vivienne ,  urlLink Keston ,  urlLink The message boards  and of course  urlLink Piptech!!   '),(20299,3252533,'20,June,2004','2004-06-20 00:00:00','    Pictures time                               urlLink Group pic    urlLink click at ur own risk     urlLink Visit the Youth Forums and chat!   '),(20300,3252533,'30,June,2004','2004-06-30 00:00:00','    Where is everyone?  Other than, Viviene (who blog in HER OWN blog)... I can\'t seem to find where the \"online\" BMBC dudes.  Btw,   Holland 2  Van Nistelrooy 40, 88  Portugal 1 C. Ronaldo 90  2:45am!   '),(20301,3252533,'28,June,2004','2004-06-28 00:00:00','    i can open mine 4 days!!!  Ya Pip i\'m juz jokin add sum sos 2 it!!   '),(20302,3252533,'28,June,2004','2004-06-28 00:00:00','    I eat my words. We were robbed by the Ref!! Final score should have been:   England 2   Owen 3\' Campbell 88\'   Portugal 1  Postiga 83\'  BMBC Youth meet was great, wasn\'t it? :D Where is everybody else? Jason, Dan, Vivi (they got their own blogs?), Rad? Sigh~   '),(20303,3252533,'24,July,2004','2004-07-24 00:00:00','    Gosh, I had to miss last and this week\'s youth meet due to exams and family commitments! My cute nephew\'s here.. and going back Sunday noon... Gotta send them to the airport, so, it will be two weeks straight not going to Youth meet church.    Can someone reserve a copy of last week\'s teaching by Ps Tamil.. I really would love to know what he has to say about what you typed.. Nice.. :)    I\'m having my holidays now, for 3 weeks.. but there\'s 1 or 2 days I need to go back school for presentation and another BM test :P I\'ll be sleeping alot, and TV-ing and Internetting and doing the things I missed since I got into this exam pressure thing. Ack.. I think I didn\'t pass one of the papers, Semiconductors... Just hoping for the best!    Cya guys next week! Take care ;-)   '),(20304,3252533,'19,July,2004','2004-07-19 00:00:00','    Church History By Ps. Tamil July17,2004  All through the years the church was persecuted but they were growing in number, in strength, in holiness and spirituality. Until AD64, an emperor by the name of Nero came into power in the Roman Empire. This man masterminded the persecution against Christian. He sat the city of Rome on fire and then blame it on the innocent Christian. * Software giant Nero got itâ€™s logo and name from this incident! Then on, official persecution against the church broke out. Christian then became the enemy of the nation and they were to suffer all kinds of persecution.  Until AD324, Constantine came into power. â€œTradition or beliefâ€ supposed that he had an encounter with God. * But the bible didnâ€™t state anything alike. He was at the verge of losing a war when he saw a cross. *Because itâ€™s tradition, when â€œthingsâ€ passed down through the ages, common sense tells you to doubt it! So some dudes try to add some authenticity to it and say he saw two words which mean Christ. Miraculously Constantineâ€™s army won the war. So after the war, he made Christianity legal and Christian started to return to the church.  Around AD392, Theodisius was in full control and he gave privileges to those out there, if and only if he or she is a Christian. Thus this brought in non-Christian as they too wanted to be privileged and not to be saved. Worst of all, they â€œnaivelyâ€ bring in all types of idols into the churchâ€¦â€¦..  Ps Tamil also spoke about the papacy(pope), Lutheran, Ana Baptist, Baptist, Presbyterian, Anglican, Methodist, about the Pentecost and how the â€œtongueâ€ came about through the century of recovering the truth and the primacy of the scripture restored!  Sorry! I cannot or should not continue further! Join us every Saturday, 8-9.30 pm at BM Baptist Church from 17 July to 21 Aug! If you really canâ€™t attend those meetings, vcd(s) for every session is available at RM7 each. For enquiries, call: 04-3240307, 012-4142690.  Copyright reserved for all course and syllabus design.   '),(20305,3252533,'19,July,2004','2004-07-19 00:00:00','    Hie everyone, &nbsp; I\'m in sgp already. My contact number for the sgp hand phone is 98807165. Please add +65 to sms and if wanna call please add +02 in front. Don\'t worry, I\'m still keeping my 012-5344465 number. I\'ll use it when i get home to m\'sia. If u guys sms me, pls tell me who u are cos, i need to re-enter my phone contacts in the phone book. My address here is: &nbsp; 46-2-874, Hall of Residence 9, Nanyang Technological University, 24, Nanyang Avenue, Singapore 639811 &nbsp; Ok, take care everyone. will be able to blog more when i get my laptop...hopefully soon enough...   '),(20306,3252533,'29,June,2004','2004-06-29 00:00:00','   New York Minute anyone ?? I\'ll rate it an A- i\'ve read it\'s book and watch it\'s movie but the ending of the novel differ from the movie !! Maybe the director wants to shorten the movie time or the Olsen twins can\'t act for another day?! Cya Type whatever u like in the shoutbox on your right!!(Who cares) Btw, i\'m reading \"Brighter Than The Baghdad Sun\" now anyone reading this best seller ????  '),(20307,3252533,'28,June,2004','2004-06-28 00:00:00','   Back it up man! It\'s already days there\'s problem !! Sorry if the message box on your right is not working blame it on the shoutbox server !!  '),(20308,3252533,'27,June,2004','2004-06-27 00:00:00','   Well, I love Arsenal and France too( I hope France wins the Euro 2004 and Arsenal win the premier league) but I have to say that you are not being very fair to others(M.united supporters, Italy, and especially the big  amount of England supporters(not forgeting the Rooney mania going around now) and of course not forgetting our dear host Portugal eg.) But hey,it is your website and what I say does not matter. Anyway(to \'angkat\' you a bit)  I agree that Arsenal is( not the best) BUT \'ONE OF THE BEST TEAMS IN EUROPE\'(I did not say in the whole world because you must consider clubs like Boca Juniors, Corinthians, Sao Paulo etc.). However, France has 2 talented players. Zizou has recieved the player of the year award and he is the best player in the world. Henry had no choice but to come in second(He was just a few votes away) but he still recieved 3 awards which are(in case you have forgotten) PFA player of the year award, the Golden boot(highest scorer in England) and of course the most wanted Silver boot(highest scorer in Europe. Possibly in the whole world) and he should be proud of that. I\'m glad that you are also an Arsenal and a Henry fan like me(I thought his 2nd goal against Swistzerland was fabulously magnificently fantastic. Reminds me of the ever wonderful Arsenal striker). In case you do not know, Arsenal has bought a new striker Robin Van Persie from Holland. He is 20 years old and has scored 30 goals in just 60 matches. Next seasons premiership would be exciting as well as entertaining with all the new players and managers. Well, I have to end here. Send my regards to all and see you soon.  P.s. There will be a concert on 3rd July and will be held in PRai Baptist church at 8p.m. HOpe to see you there. One more thing, pls tell Adele thet her e-mail is expired again. Need to forward her a few things.                            In christ                                 John  '),(20309,3252533,'25,June,2004','2004-06-25 00:00:00','   Dude i\'ll eat my words! It\'s only their first loss in 2004! Most labelled it as a Greek fairytale or David and Goliath story I would say it\'s luck!!! I was \"that damn lucky\" too when i got an A for chinese in my PMR I hope luck will be on my side during SPM  Nevertheless, I took EST as a \"cover up\" for my chinese!  '),(20310,3252533,'25,June,2004','2004-06-25 00:00:00','   England thought they won it in 90 minutes Portugal thought they won it in extra time Then it all came down to the penalty shootout! Beckham scores the rugby style! Then another Aston Villa Old Gun - Mr. Vassell miss a penalty And Ricardo finish the game off by scoring the winner !!  Bring it on Portugal! France is coming your way!  '),(20311,3252533,'23,June,2004','2004-06-23 00:00:00','   Dude ! Caught up with the Euro fever or should i say ROONEY-MANIA ?? Haha The espn pundits commented that one thing Rooney does\'t have is good looks ( i guess Beckham will be quite happy with that ) For those guys(and gals) who support England, please pause for a moment and rethink !! Can a bunch of inexperience first-timers , one trendsetter and a Swiss steamroll England to success ?? Can they deliver under pressure from the media and the boo-ers Portugal fans ?? Joker! even Chris Colombus can\'t think of such a plot for a blockbuster megahit Hollywood movie! Man\"O\"Man believe me ! France will be the ultimate winner ! Think bout Henry, Pires, Zidane, Vieira ......... World Footballer Of The Year ! European Footballer Of The Year ! Golden Boot Winner ! ...... If there\'s one team on earth which is \"invincible\" (other than Arsenal) it\'s France ( which is mould out of Frenchmen and by Frenchmen) ! Without a doubt, I once again would like to emphasize that France and Arsenal are the best football teams around! God , let France and Arsenal dominate the football world! May your grace and loving kindness be with them and also with us always !!  '),(20312,3252533,'18,June,2004','2004-06-18 00:00:00','   Friday June 18 2004 Jit Sin teacher\'s day!  The highlight of the day is undoubtedly The Ever Ready Teachers versus The Kaki Bangkus Prefects in a entertaining high scoring match! The match was an one sided affair and the final score, an embarrassing 5-0 lose for the prefects! Even our own church leader Sir Ravi Samuel got into the act, scoring a brace and with his commanding defensive role,the students voted him the Man Of The Match. However there\'s no prize for the winner nor the Man of the Match.  '),(20313,3252533,'14,June,2004','2004-06-14 00:00:00','   Most importantly 2004 is my one and only make or break SPM year! Seniors say â€œif you doesnâ€™t do well, there goes your future!â€ But whoâ€™s stupid enough to try it out Even a daring dude has a pee brain to filter that SPM is no joke! Itâ€™s â€œkick butt, take-no-prisonerâ€ attitude in other words â€œDo or dieâ€ But Iâ€™m still relax, enjoying myself, typing this blog on the net rather than study! Whatâ€™s wrong with me?  '),(20314,3252533,'13,June,2004','2004-06-13 00:00:00','   2004 -> year of \"all years\" Euro 2004( I hope France win it with Henry bagging the golden boot) Malaysia Idol 1st season(may Malaysia produce less William(s) and more Aiken(s)) Thomas Cup(Malaysia flop again) Uber Cup(Needless to comment) Olympic Games(just wait and see) Elections (many many of them all over the world, but I rate Indiaâ€™s election as the best, as if Iâ€™m watching a movie with the exciting plots with trailers, dramas, what ever you ask for, they have it ) And many other â€œminorâ€ events like BMBC youth having a pastor(which they already donâ€™t have now and may have again) and still we are only mid-year and I bet thereâ€™s more to come!!!  '),(20315,3252533,'11,June,2004','2004-06-11 00:00:00','   Airs every Wednesday amd Thursday night 12 midnight on 8TV Starring Jason Lo as host and other celebrities  Patrick Teo , i guess Malaysian know about him !! Last night as a guest star , he had a whale of a time with Jason First up, he says he is curious why Jason always have the same group of people in Starbuck Sunway watching Latte@8! Do Jason pays them to do so ?! Certainly no, Jason says that they are the most loyal fan of Latte@8 So they talk about some nasty \"stuff\" in the middle of the show one part was cut and when it resume, Jason was talking about \"shuttlecock\"!! Common sense tells you thats very darn dirty but acceptable in today\'s world! The show goes on and it\'s really fun watching Malaysian hosting this shows and best of all hosting them very truly deeply madly well Jason Lo is one of the rare breed of capable Malaysian to do this  Can\'t wait Malaysia Idol starting tonight 9.45pm 8TV I hope there will be more \"Kelly Clarkson and Clay Aiken\" than \"Willian Hung\" -> America Best Reject ?? Then who\'s Malaysia Best Reject ??  '),(20316,3252533,'10,June,2004','2004-06-10 00:00:00','   Hi I\'m Keston Chew Vern Jiet from BMBC youth. I\'m glad God gave me the opportunity to log on the internet and share a few things with you. First and foremost, many will say it is untimely for me to comment on this topic, but i just want to clarify and make some \"things\" clear  I\'ll start off by sharing the whole incident that happen the night itself. At around 7.30pm, my mum fetch my bro and i to church. I was quite happy to be given the chance to play the guitar but no matter how hard i practice, i still can\'t get the tempo of play Jacob wanted me to play. So at around 7.50pm, Chun Keat and Chun Hean entered the church, i spotted Chun Keat first so i ask him to give me a hand while i proceed to play the drum! Pastor arrived exactly at 8pm. So Radiance \"kick off\" the worship by saying a short yet meaningful prayer. Man\'O\'man it didn\'t go as plan. Jacob\'s tempo beat on the piano just didn\'t match the guitar and the drum! And Radiance!!! She try to sing to the tune!!! that\'s all she could do!!! but to no avail and her tune didn\'t match Jacobs\' either. We then make our way to the nursery room. Before going in, Pastor have a word with me, not bout the worship but bout Daniel and other university-bound youths! I politely told him i don\'t know anything! Inside the nursery, Pastor\'s first sentence if i\'m not wrong is \"Y is dere so few youths attending dis youth meetin? I remember dere r 2 bro and 1 sis, d last time i saw dem was a month ago!\" I guess Pastor is talking bout the Thams and then Adele took out the attendance list and check for other absentees. Most have \"valid\" reason for not attending. E.g. Pip got to go for a wedding and so on. Then came the \"Fan flaming the fire\", Pastor wanted to continue his conflict topic but there is one very big problem - out of 10 youths only 2 brought their notes. Pastor then started to comment that our youths are not puncture for youth meeting and then he says \"According to ....... , not puncture for meeting and not bringing notes means those offenders don\'t respect others and the youth group as a whole\" He then added some other comments maybe because he wants to correct us. Then he said \"Y u all come for meetin like\'dat? Wastin time oni\" That phrase set my bro \"on fire\" and he told pastor \"If we continue like this and don\'t learn anything, it\'s really wasting our time. I\'m here for only the second time(since Pastor took over) and wants to hear the good news!!\" Pastor then answered \"U don\'t tell me wat to do , are u sayin my papers are unprofitable?\" (where did pastor got that from? No one said that) Kester replied that he is not saying the paper is unprofitable but Pastor kept insisting that me bro said it!! (Wat a conflict?! A big-time misunderstanding perharps) I just can\'t keep my big mouth shut and told Pastor what to do !!!! Ops!! I told Pastor to continue his conflict topic and today we\'ll discuss bout eschatology!! Pastor reluctantly discuss it but at least we learn something that day! Last few words from the Pastor is \"When the church take sides(against Israel), remember this day i told you not to go against Israel\" the then says this time will be his last time at BMBC!  witnesses ! if there is any wrong sequence or proceeding above , pls contact me 0125139020! Keston Chew Vern Jiet 871109-35-5467 BMBC youth (bmbc.info/youth)  one more thing, Mr/Mrs/Miss/Dr Anonymous, Pls reveal urself !!!!  '),(20317,3252533,'29,July,2004','2004-07-29 00:00:00','   This looks better!&nbsp; urlLink     urlLink  '),(20318,3252533,'24,July,2004','2004-07-24 00:00:00','   It\'s me!!!&nbsp; urlLink     urlLink  '),(20319,3252533,'24,July,2004','2004-07-24 00:00:00','   I brought some movies(most of them DVDs cause it\'s only RM6) last week n i\'ll rate them accordingly King Arthur - Great! but Lancelot \"should have\" survive Honey - Below Average! Jessica Alba can dance and choreographer??? I didn\'t know that You Got Served - Juz OK! Black guys with their trademark breakdance Prince & Me - Thumbs up! \"The fairy tale is about to get real\" Girls will love it and some may dream of it Mean Girls- Above Average! I like wild girls, it seems so Kill Bill 2- Good! Bloody Dirty \"Revenge is a dish best served cold\"! Enjoy!  '),(20320,3252533,'24,July,2004','2004-07-24 00:00:00','   \" It doesn\'t matter how often a married man changes his job, he  still  ends up with the same boss.  \" (Ghanchi-Boy)  '),(20321,3252533,'24,July,2004','2004-07-24 00:00:00','   Church History By Ps. Tamil  				        	                July17,2004   All through the years the church was persecuted but they were growing in number, in strength, in holiness and spirituality. Until AD64, an emperor by the name of Nero came into power in the Roman Empire. This man masterminded the persecution against Christian. He sat the city of Rome on fire and then blame it on the innocent Christian. * Software giant Nero got itâ€™s logo and name from this incident! Then on, official persecution against the church broke out. Christian then became the enemy of the nation and they were to suffer all kinds of persecution.   Until AD324, Constantine came into power. â€œTradition or beliefâ€ supposed that he had an encounter with God. * But the bible didnâ€™t state anything alike.  He was at the verge of losing a war when he saw a cross. *Because itâ€™s tradition, when â€œthingsâ€ passed down through the ages, common sense tells you to doubt it! So some dudes try to add some authenticity to it and say he saw two words which mean Christ. Miraculously Constantineâ€™s army won the war. So after the war, he made Christianity legal and Christian started to return to the church.   Around AD392, Theodisius was in full control and he gave privileges to those out there, if and only if he or she is a Christian. Thus this brought in non-Christian as they too wanted to be privileged and not to be saved. Worst of all, they â€œnaivelyâ€ bring in all types of idols into the churchâ€¦â€¦..   Ps Tamil also spoke about the papacy(pope), Lutheran, Ana Baptist, Baptist, Presbyterian, Anglican, Methodist, about the Pentecost and how the â€œtongueâ€ came about through the century of recovering the truth and the primacy of the scripture restored!   Sorry! I cannot or should not continue further! Join us every Saturday, 8-9.30 pm at BM Baptist Church from 17 July to 21 Aug! If you really canâ€™t attend those meetings, vcd(s) for every session is available at RM7 each. For enquiries, call: 04-3240307, 012-4142690.  Copyright reserved for all course and syllabus design.  '),(20322,3252533,'23,July,2004','2004-07-23 00:00:00','   Ben Yap! Opt for a \"girlish\" e-guitar  urlLink  '),(20323,3252533,'16,July,2004','2004-07-16 00:00:00','   Got The Fever ??&nbsp; urlLink     urlLink  '),(20324,3252533,'16,July,2004','2004-07-16 00:00:00','   1. Women are unpredictable. Before marriage, she   expects a man, after marriage she suspects him, and after death she   respects him.     2. There was this guy who told his woman that he loved  her so much that he would go thru hell for her. They got married - and   now he is going thru hell.      3. A man inserted an \'ad\' in the classifieds : \"Wife   wanted \". Next day, he received a hundred letters. They all said the   same thing : \"You can have mine.\"     4. When a man opens the door of his car for his wife,   you can be sure of one thing: either the car is new or the wife.     5. It\'s easy to tell if a man is married or not. Just   watch him drive a   car with a woman sitting beside him. If both his hands   are on the wheel, you can be sure he is married.     6. A man received a letter from some kidnappers. The   letter said, \"If you don\'t promise to send us $100,000 I swear that we   will kidnap your wife.\" The poor man wrote back, \" I am afraid I can\'t   keep my promise but I hope you will keep yours.\"     7. What\'s the matter, you look depressed.\" \"I\'m having   trouble with my wife.\" \"What happened?\" \"She said she wasn\'t going to   speak to me for 30 days.\" \"But that ought to make you happy.\" \"It did,   but today is the last day.\"     WOMAN   When she is 18 - She is a football, 22 men going after   her. When she is 28 - She is a hockey ball, 8 men going after her. When   she is 38 - She is a golf ball, 1 man hitting on her. When she is 48 - She   is a pingpong ball, 2 men pushing to each other.      MAN   At 20 - A man is like a coconut, so much to offer, so little to give.   At 30 - He is like a durian, dangerous but delicious.   At 40 - He is like a watermelon, big, round and juicy.   At 50 - He is like a mandarin orange, the season comes once in a year.   At 60 - He is just like a raisin, dried out, wrinkles and cheap.      Marriage Humour In the beginning,      -- God created earth and rested. Then God created man  and rested. Then God created woman. Since then, neither God nor man has   rested.  '),(20325,3252533,'16,July,2004','2004-07-16 00:00:00','   &nbsp; I\'m in school !!!!  '),(20326,3252533,'12,July,2004','2004-07-12 00:00:00','   BMBC youths ! Where r u ?? r u here ?? Our youth site is deserted !!! click the link below  urlLink  BMBC YOUTH   and get involve!!  '),(20327,3252533,'09,July,2004','2004-07-09 00:00:00','   Monthly test is back !! Next week ( Monday till Thursday ) Good Luck eveyone !!  '),(20328,3252533,'09,July,2004','2004-07-09 00:00:00','   When I was 10, my prized possession was my football â€“ I ate with it, slept with it and I polished it weekly â€¦ instead of my shoes. I knew about football â€“ but on some other things, like where babies came from, I was a little hazy. 	One afternoon, I was playing in the street and I lost my precious ball. I looked everywhere. I figured someone had stolen it! 	Eventually I spotted a woman who seemed to be hiding it under her jacket. So I marched up and asked her: â€œWhat do you think youâ€™re doing with my football stuck up your shirt?â€ 	It turned out that she didnâ€™t have my football â€¦ but that afternoon I learned where babies came from â€“ and what a woman looks like when she is 9 months pregnant. 	Later that day, I also found my ball. 	What fascinated me most of all, was why I had never noticed a pregnant woman before the age of 10 â€¦ and why, from then onward, I seemed to be surrounded by them!  Extracted from a book I read in the Jit Sin High School library  '),(20329,3252533,'07,July,2004','2004-07-07 00:00:00','   The 7 Ups!  1. Wake Up !!   Decide to have a good day. \"This is the day the Lord hath made; let us rejoice and be glad in it.\" Psalms 118:24   2. Dress Up !!  The best way to dress up is to put on a smile. A smile is an inexpensive way to improve your looks. \"The Lord does not look at the things man looks at. Man looks at outward appearance, but the Lord looks at the heart.\" I Samuel 16:7  3. Shut Up!!  Say nice things and learn to listen. God gave us two ears and one mouth, so He must have meant for us to do twice as much listening as talking.  \"He who guards his lips guards his soul.\" Proverbs 13:3  4. Stand Up!!   . . . for what you believe in. Stand for something or you will fall for anything. \"Let us not be weary in doing good; for at the proper time, we will reap a harvest if we do not give up. Therefore, as we have opportunity, let us do good...\" Galatians 6:9-10  5. Look Up !!  . . . to the Lord. \"I can do everything through Christ who strengthens me\". Philippians 4:13  6. Reach Up !!  . . . for something higher. \"Trust in the Lord with all your heart, and lean not unto your own understanding. In all your ways, acknowledge Him, and He will direct your path.\" Proverbs 3:5-6   7. Lift Up !!  . . . your Prayers. \"Do not worry about anything; instead PRAY ABOUT EVERYTHING.\" Philippians 4:6   A POSITIVE THOUGHT  If God had a refrigerator, your picture would be on it. If He had a wallet, your photo would be in it. He sends you flowers every spring, and a sunrise every morning.  Whenever you want to talk, He\'ll listen. He could live anywhere in the universe, and He chose your heart. What about the Christmas gift He sent you in Bethlehem; not to mention that Friday at Calvary. Face it, He\'s crazy about you.  '),(20330,3252533,'06,July,2004','2004-07-06 00:00:00','   Malaysia semua babi,  our government is so lousy,  make the college becomes no quality,  Chinese and Indian pun tak boleh masuk university.  It never had a choice for me, because no money cannot study at oversea, so I always feel i\'m so pity.   Every college are greedy, sini sana all need money, their pocket kosong our perut noisy, I am poor because there sucks my money.   Exam question full of difficulty, it has a reason only, which is to earn more money, because we retake money masuk lagi. The staffs always no aksi, they did wrong never say sorry, before asking please show your money, otherwise they wont be happy, we pay their salary still make us angry. that\'s why i wanna study hard to earn more and more money   Tiap-Tiap hari feel so lazy,  never try to wake up early, always hope college sekarang cuti,  everyday go to class feel so sleepy,  lecturer always scold me lazy.  do assignment just like to copy, before the final just feel worry, every nite study until crazy, because scared scold by mummy.   Why my class always no pretty, those elephant always show me how she\'s sexy, the monster wanna treat me nicely, like that i better mati.   my looking is **** ugly,  saya mari girl girl lari,  everyday date them they only say busy, semua orang tak ada hati, make me always so lonely,  only can watch porno movie.    CS always let people bully, maybe i am really so lousy, althought knife people is so happy, but i am always unlucky, kena headshot easily, after mati still mati.   This is just a small story,  typed it simply,  feel funny and happy,  actually i am not lonely, because I have many friends around me.  '),(20331,3252533,'06,July,2004','2004-07-06 00:00:00','   Curly Bin ! Hello Pixy ! .........  '),(20332,3252533,'04,August,2004','2004-08-04 00:00:00','   Pesanan Penting  Menurut laporan Polis seorang wanita telah dirogol oleh 5 lelaki disebuah kelab malam di Singapura pada malam Sabtu sebelum ditinggalkan. Wanita tersebut tidak berupaya untuk mengingati kejadian malam tersebut, namun ujian mengesahkan bahawa beliau telah dirogol berulang kali dengan kesan rohypnol di dalam darahnya dan progestrex, sejenis pil \"mencuci\" yang kecil. Dadah tersebut kini digunakan oleh perogol-perogol dan kemudian \"mencuci\" mereka. Progesterex dibekalkan kepada para Doktor haiwan untuk\"mencuci\"haiwan besar. Khabarnya Progesterex digunakan bersama Rohypnol, sebagai dadah \"Dating\" dan rogol. Apa yang mereka perlu lakukan ialah menitiskan Rohypnol ke dalam minuman si gadis, mangsa tersebut tidak akan ingat apa-apa pun esok paginya, apa yang berlaku kepadanya sepanjang malam tersebut. Progesterex amat senang larut dalam minuman, akan \"mencuci\" agar simangsa tidak akan mengandung akibat dari perkosaan tersebut dan siperogol tidak bimbang untuk perlu menjalani ujian air mani dan DNA kemudian. Kesan dadah ini adalah K*E*K*A*L terhadap si mangsa. Prodesterex asalnya adalah untuk memandulkan kuda, Mana-mana wanita yang mengambilnya TIDAK AKAN MENGANDUNG SAMPAI BILA-BILA. Si durjana boleh memperolehi dadah ini boleh mendapatkan dadah ini dari pusat haiwan atau mana-mana universiti, dan ia diramalkan akan digunakan dengan berleluasa di kampus-kampus.  Tolong edarkan maklumat ini kepada semua yang anda kenali, terutama teman wanita dan gadis remaja. Berwaspada apabila anda keluar, jangan biarkan minuman anda terbiar tanpa dijaga oleh orang yang anda percayai.     Azila Bakri, Social Worker,  All Women\'s Action Society (AWAM)  '),(20333,3992091,'21,July,2004','2004-07-21 00:00:00','    Ok this is my first post. Just testing and stuff. '),(20334,3992091,'17,August,2004','2004-08-17 00:00:00','    Look! Look!     Wow, don\'t they have lovely scarves. Wouldn\'t it be great if I could, say, knit one just like it?     Woohoo! It is so incredibly soft too, I love it!  I am a self-admitted dork but dorks are cool anyways. '),(20335,3992091,'14,August,2004','2004-08-14 00:00:00','    I finished a project! WOOOOOOHOOOO!  My Harry Potter scarf is now DONE. I just need to block it, which I will do when I get back home tomorrow.  I\'m very happy about it and for the first time ever, I\'m looking forward to the cold weather just so I can wear it. :p I will post pics in a couple of days.  ***  Now. I still have two WIPs, both of which I\'m a bit bored with. I\'m working on my illusion scarf right now (I have a bit more than a third of it done).  I will receive the Noro Kureyon beauties soon and they will be made into a bag.  But I also have one million projects in mind and I can\'t decide which one I want to make next. So here\'s me organizing my thoughts...  I have two knitted presents planned for the holidays.   Mom : I\'m hesitating. I can go simple and knit her a scarf. Her coat is white and red. So the colours would have to go with it. I know she LOVES my HP scarf so I may go with a similar style (in the round and all). The other option would be the Chinese charm bag from Stitch N\' Bitch. I already have the handles and she loved it when she saw the picture in the book. I just don\'t know which yarn I\'d choose as I don\'t know where to find Noro Gemstones. If not, I guess I\'ll choose another brand of variegated yarn. I just need to be very careful with the colours to make sure she will actually use it.   Sister : My first idea was to make her a scarf and if I have time, mittens or a hat to go with it. The problem is that I have no idea what kind of coat she\'s gonna wear this winter. So I wouldn\'t know which colour to choose. Hmmm. Option #2 would be felted slippers, like the Fuzzy feet from Knitty. It\'s marked as a \"tangy\" pattern though. But it really is just a huge pair of socks and I\'ve been meaning to try socks anyways. So I don\'t know.  I really need to figure out what I\'m gonna make and start soon since I\'m such a slow knitter.   Projects for me that I have in mind :  -A poncho AND/OR a shawl AND/OR a blanket. Basically, just something that I can bring at work to wrap myself in when I\'m cold and everyone else is hot so I can\'t turn on the heat. I just need to find the perfect pattern first.  -A sweater. I want to make my first sweater and I want to make it now. :p I\'m hesitating between To dye for, from the SNB book or the Surplice Cardigan from the Fall 04 issue of Vogue knitting. It\'s very pretty and marked as easy so maybe I could manage to make it.  -Socks. I really want to make socks. Any socks.   So many projects, so little time.  ***  Thing to add to the \"I want it but can\'t knit it yet\" list: Charlotte\'s Web. I wish. '),(20336,3992091,'12,August,2004','2004-08-12 00:00:00','    Lookie what I got on Ebay.... Three skeins of this:     This is so gonna be a bag. '),(20337,3992091,'11,August,2004','2004-08-11 00:00:00','    I just won 26 pairs of knitting needles on Ebay.  *happy dance*  Now I still need a set of Denises. BUT, I\'m using those as an incentive to stick to my diet (I\'ll buy them when I\'ve stuck to my diet for a certain number of days OR when I\'m to a certain weight that I\'m not disclosing here).  Oh and FYI: I\'m hoping for a FO sometime next weekend. Weee! I\'m such a slow knitter... '),(20338,3992091,'07,August,2004','2004-08-07 00:00:00','    I went to my LYS today. Now the place is small. Tiny. But so filled with yarn. I love it. And the place was crowded today (mostly older ladies). I spent so much time there, just looking at all the yarn, touching, trying to decide what to buy. I really just wanted to see if they had more of the yarn I\'m using for my HP scarf (they did!).  It was so funny too, there was this lady there who had the most charming accent (she said she\'s originaly from Egypt but she\'s been living here for 20+ years). She was talking to this other lady, telling her that she would once again get out of the store with too much yarn and a big dent in her bank account. \"There\'s just so many nice things in here!\", she said and then she looked at me and smiled and said \"don\'t you think?\". I just smiled and nodded. I looked at her shopping basket: it was FULL! I swear she must have had over 200$ worth of yarn in there (and the yarn is quite cheap there so that\'s a lot of yarn).  I wish I could have bought as much but my wallet said no. I bought one ball of yellow merino wool and two of the red one (I\'m hoping I\'ll have leftovers to make mittens or something). I also bought some type of novelty yarn (similar to the fun fur thing) in two different shades of purple. There was a scarf made with that yarn in the store that was just too cute.  And now is the time for a little confession. My name is Caroline and I\'m a scarf addict. I know I should knit other things (and I will, my next project will be a sweater, I\'m pretty sure) but I think I will ALWAYS have a scarf on the needles, whether it\'s for me or a gift. The thing is, I\'ve loved scarves even before I started knitting. I always buy too many of them in winter. I have three black coats (different styles, same colour) and I like to change the look with a different scarf. What can I say. So yeah. Scarves. My HP one is looking good even though I still have eight (?? I think) stripes to knit. Wanna see? I know you do!    It\'s all soft and warm too! Woo! '),(20339,3992091,'06,August,2004','2004-08-06 00:00:00','    Once upon a time (i.e. a few years ago), there was this girl -- we\'ll call her Caroline -- who wanted to learn how to knit. Why? This will remain a mystery, even to her. Caroline\'s mom knows everything, or so the knitter-to-be thought. \"Mama, can you teach me how to knit?\", she asked (except what she really said was \"Maman, peux-tu m\'apprendres Ã  tricoter?\"... Yes, Caroline speaks a weird and mysterious language sometimes known as Canadian French).  Much to her surprise, the mother replied that she didn\'t know how to knit. Crushed, Caroline put away all of her knitting dreams, convinced she would never learn the craft.  Now fast-forward a couple of years. We are now in December 2003. Caroline has a two weeks holiday break from work. Joy! Happiness! Boredom! Inspired, she suddenly decides that she would learn how to knit and that in fact, she would learn how to knit right at this moment. So off to Wallmart she goes. There, she finds a little \"I taught myself how to knit\" kit (as well as a \"I taught myself how to crochet\" one that has yet to be used). She also buys some yarn that fitted all the qualities she thought were needed in a yarn (i.e. the colours were pretty). So Caroline gets home, she sits comfortably on the couch, book in hand. She learns how to cast-on and then how to do the knit stitch. \"Woohoo! I can knit!\", she proudly thought.  Caroline puts the book down and decide to begin the obligatory first scarf project. Unfortunately, things don\'t go too well and what was supposed to be a long pretty scarf quickly becomes a sort of weirdly shaped piece of fabric. So our hero puts down the needles and doesn\'t touch them for a couple of months.  Then, in February or so, she picks them back up and start a brand new scarf with brand new yarn. This time, she decides to count the stitches on every row. After what seems like an eternity (it was really three months), she has her first -- still uneven but much better looking -- scarf (see Figure 1).     Figure 1. Caroline\'s first scarf   What I haven\'t mentionned yet is that while she was working on that scarf, Caroline had heard of a knitting book called Stitch N\' Bitch that seemed to be all the rage. Of course, curiosity got to her and she ended up buying the book. She fell in love with many of the patterns and the love affair still continues today.  But back to the knitting projects. After the scarf, Caroline decides she wants a bag. So why not try one that\'s in the beloved book? And with that, she starts her second knitting project: Zeeby\'s bag. It takes a month to finish and ends up not looking exactly like the original, but she likes it nonetheless and still uses it often (see Figure 2 and 3.)     Figure 2. Zeeby is resting on Caroline\'s lay-z boy.    Figure 3. Caroline is pretending to carry Zeeby for the photographer**.   Since then, Caroline hasn\'t stopped knitting but for some unknown reason, she can\'t seem to finish anything. She currently has three projects on the needles (two of them are almost done now) and she can\'t wait to show them off.  She also has an insanely long list of \"To-Knit\" items...  **There was no photographer, I made that up. '),(20340,3905634,'14,August,2004','2004-08-14 00:00:00','   -2 parts Vodka -3 parts cranberry juice -2 parts grapefruit juice -lime peel  Shake all ingredients with ice(remember to wash!) and strain into a tall glass.  '),(20341,3905634,'13,August,2004','2004-08-13 00:00:00','   King Edward VII introduced the custom of setting the 180 or so clocks on the Sandringham Estate half an hour early to allow him more time to shoot. As a consequence, all business when the King was at Sandringham took place in this unique royal time zone. George V maintained this tradition, but Edward VIII, when he acceded to the throne in 1936, reset Sandringham\'s time pieces in synchronicity with the rest of his kingdom.  '),(20342,3905634,'13,August,2004','2004-08-13 00:00:00','   I got down to thinking, its such a waste to spend six dollars on a grande size mocha frap at starbucks! Its an absolute rip off! So anyway, here is a recipe for one.  Take a blender, add in  -2 heaped tablespoons full of milo -1 heaped table spoon of cadbury hot chocolate powder -1/2 tablespoon Boncafe Mocca Coffee powder -1 near full cup of hl milk or any full cream or skim milk, your preference. -1/2 tablespoon condense milk (you can adjust as you try) -12-14 cubes of ice (Remember to wash the ice cubes first to remove any fishy, meaty smell that the ice takes on from storage in your fridge)  Blend until smooth, taste, and adjust taste as required.  Enjoy this simple drink! ^^ Adieu!  '),(20343,3905634,'13,August,2004','2004-08-13 00:00:00','   All right, i will be posting certain topis from time to time. They include: Aqua Vitea(drink recipes) La Gastronomique(for food), intelligentsia( stuff to enrich the mind) and finally haute couture ( Cultured stuff and gk for all you know it all bums ^^)  '),(20344,3905634,'13,August,2004','2004-08-13 00:00:00','   1) Unlike guys, they dont always talk about cars or soccer or trade ns stories! I mean... who on earth cares about cars if you cant afford them!!! Or soccer if you are not playing, unless of course you are betting heavily, then thats a different story.... And ns stories.....What in blue blazez!?!?!??!  2) Girls are definitely much nicer people to be with, ok...maybe thats a generalisation... on the same level of niceness (if there even is such a word) girls are much nicer. Why? Because they are able to pick up small signs of how you are feeling and will most probably know the right thing to say next. Of course it comes from them being hard wired as mothers but hey...just like testosterone for guys, its an advantage.  3) Girls can see effectively 180 degrees, well at least some. Definitely more than guys who only have tunnel vision. This is once again due to the fact that girls are nest defenders and thus have more visual cones in their eyes. But its still so cool! Thats why guys get caught all the time when they are ogling (did i spell that right?) at girls!  4) Girls arent always so (mind my crude term) horny as guys are! You know it gets Soooooooooooooooooooooooooooooooo irritating when you are talking to a guy and all he is doing is looking in the other direction because of a pretty girl or something. Even if hes looking at you, he will go like :Uh huh.... Then try to steal a glance at the girl. Grrrrr......  5) Girls are so much better at humanities than guys, they can appreciate good songs and poetry. Rare is the man who can finish this poem that begins with : Whats in a name ^^  '),(20345,3905634,'13,August,2004','2004-08-13 00:00:00','   1) Guys are usually more of the sporty type, you want a game of rugby, tennis? No problem, of course its due to the testosterone and all but hey.... Its still an advantage.  2) Guys dont always talk about shopping or the latest mascara and definitely dont gossip as much ^_^  3) Guys do not whine or cry as much, not that crying is a bad thing but going on and on about something and living in the past is one of the girl trademarks.  4) Guys use current evidence to argue and never bring up personal attacks unlike girls who bring up that you are stupid or gay or what you did last week that has absolutely nothing to do with the argument at hand and they always find a way to bring up something personal.  5) Guys, unlike girls, are able to appreciate lame jokes. For some reason or other, either girls are slow or they have no sense of a warped humour whatsoever.  '),(20346,3905634,'17,August,2004','2004-08-17 00:00:00','   Ambrose Bierce (1842-c.1914) was a remarkable man: a veteran of the American Civil War, a writer, poet, journalist, and most memorably, the creator of the The Demon\'s Dictionary. For an audience: enlightened souls who prefer dry wine to sweet, sense to sentiment, wit to humour, and clean English to slang. Here are some of the more pungent entries:  Achievement: The death of endeavour and the birth of disgust.  Actually: Perhaps; possibly.  Adore: To venerate expectantly.  Auctioneer: A man who proclaims with a hammer that he has picked a pocket with his tongue.  Bore: A person who talks when you wish him to listen.  Court fool: The plaintiff  Coward: One who in a perilous emergency thinks with his legs.  Dentist: A prestidigitator who putting metal into your mouth, pulls coins out of your pocket.  Envelope: The coffin of a document; the scabbard of a bill; the husk of remittance; the bedgown of a love-letter.  Erudition: Dust shaken out of a book into an empty skull.  Forefinger: The finger commonly used in pointing out two malefactors.  Ghost: The outward and visible sign of an inward fear.  Hope: Desire and expectation rolled into one.  Hospitality: The virtue which induces us to feed and lodge certain persons who are not in need of food and lodging.  Influence: In politics, a visionary  quo  given in exchange for a substantial  quid .  Insurrection: An unsuccessful revolution. Disaffection\'s failure to substitute misrule for bad government.  Language: The music with which we charm the serpents guarding another\'s treasure.  Mausoleum: The final and funniest folly of the rich.  Painting: The art of protecting flat surfaces from the weather and exposing them to the critic.  Pray: To ask that the laws of the universe be annulled in behalf of a single petitioner confessedly unworthy.  Price: Value, plus a reasonable sum for the wear and tear of conscience in demanding it.  Riot: A popular entertainment given to the military by innocent bystanders.  Twice: Once too often.  Zeal: A certain nervous disorder afflicting the young and inexperienced.  '),(20347,3905634,'16,August,2004','2004-08-16 00:00:00','   Hai... i just dont know what to do..... Today my class had a discussion on whether torture was ever justifiable... I had a very very long presentation....around 3 full pages of transparencies compard to the others half. Ahahaha..... I was trying to introduce to them the debating technique of using definitions to aid your stand, i used how justified can be defined, case studies and all......  The teacher was so interested...the students... well... some of my friends were paying close attention but the rest were just vegetating away.... And they still had the audacity to say \"So why arent you answering the question\" and \"Your point is?\" Omg..... I was giving them the answer on a silver platter yet none could see it.... Stoopid girl and guy...i shant name... Its Junior college! Its called tertiary education! When will they stop expecting to be spoon fed! Not only did i give them a full argument, i gave them the generic backing for other similar topics.... Oh well....  Must as well gripe about society since i am in the mood... you know...i am still waiting for a girl to come up to me, i mean as friends, curtsy, eat with some dignity...or even a guy who has proper table manners. Its absolutely horrible! I am placing bets that few know what it means to eat American style or British style.... your economic standing is no excuse for this form of ignorance, i mean...i am only in a middle class family.... Even if i do not ask so much...when will i see someone who eats without his arms so spread out, or without noise when drinking soup or cutting meat..... Singapore....My home.....We just need to learn a bit more about culture....Hypocracy? Do i hear sour grapes? Its called having some decency.  '),(20348,3905634,'16,August,2004','2004-08-16 00:00:00','   What do you know! The new week is here! Full of renewed energy and raring to go!!  On wednesday i have a chinese test, auditions for emcee for teachers day and a trip to tekong! On friday i have got a barbecue to attend to, guess whos cooking? ^_^ Hehe.....  The week started off pretty fine, just getting some backlog work done and doing up my pw (EE ^_^ aint that the bomb...). Kinda pooped after the long day....  Time to pray for the seventh month thingy, well not me...but my family! Ahahahaha.....  Poor J2s have got their prelims now...ahaha.... well my turn will be coming soon.  All rightz, its time for me to go to bed. Nine thirty already!  Adieu!  '),(20349,3905634,'15,August,2004','2004-08-15 00:00:00','   Ok....heres a simple marinade for barbecue. Make sure you leave the chicken to stand for 2 hours or so.  For a more mediter. flavour: -olive oil -garlic (around 1 whole clove) -chopped fresh herbs (oregano, thyme...etc, if you are unsure of the kinds of herbs, use the italian herb mix available at any store) -salt and pepper to taste.  For a more spicy flavour: -olive oil -garlic (around 1 whole clove) -chopped rosemary -ground chilli (as much as you want) -salt and pepper to taste.  Or you could use the chilli and take out the seeds for a tinge of chilli flavour instead of full spice.  Remember for more flavour, strew the coals (when they are all grey or glowing) with woody herbs such as rosemary sprigs or fennel stems.  Adieu!  '),(20350,3905634,'15,August,2004','2004-08-15 00:00:00','   I am just sitting in my room, listening to Kim Lian\'s \"Big girl\" Cd track...(Aint she just the right mix of sweetness and punk? *sigh*) For some reason or other i feel so hopeless, sure...I do blame it on my feminine side but since its a part of me...what can I do.... I just feel like i need someone right now... But i have noone to turn to because i have little friends, actually less than a handful whom I actually call friends (Yeah...this comes from being wierd)  I need someone to hug....but there is just my little cosy green room.... so i have no choice but to cuddle my little teddy (Whose name is Celi by the way ^^) The softness does offer some temporary respite....but its cold. Unfeeling. Cold.  But this is the way life is.... and how it will be for some time to come. I just wish my feminine side would stop surfacing.... it feels....wierd. But i do relish that emotion from time to time.  Looking forward to an afternoon of work, project work and mugging for physics. Allow me to make myself a baileys. Suddenly....it seems much brighter! It fades. I need noone. I have myself and my God ^^  Adieu!  '),(20351,3905634,'15,August,2004','2004-08-15 00:00:00','   A favourite Italian dessert which is found in many regions. Here it is flavoured with coffee and Amaretto, but you could also use Marsala and Maraschino.  Serves 4-6  -20-24 sponge fingers, about 150g -2 tbsp cold black coffee -2 tbsp coffee flavouring (extract) -2 tbsp Amaretto or brandy -4 egg yolks -90 g caster sugar -few drops vanilla flavouring (extract) -grated rind of 1/2 lemon -350 g Mascarpone cheese -2 tsp lemon juice -250 ml double (heavy) cream -1 tbsp milk -30 g flaked (slivered) almonds, lightly toasted -2 tbsp cpcpa powder -1 tbsp icing sugar   1)Arrange almost half of the sponge fingers in the base of a glass bowl or serving dish. Combine the black coffee, coffee flavouring and Amaretto or brandy and sprinkle over half the mixture over the fingers.  2)Put the egg yolks into a heatproof bowl with the sugar, vanilla flavouring and lemon rind. Stand over a saucepan of gently simmering water and whisk until very thick and creamy and the whisk leaves a very heavy trail when lifted from the bowl.  3)Put the Mascarpone into a bowl, add the lemon juice and beat until the mixture is smooth.  4)Combine the egg mixture and Mascarpone mixture until evenly blended, then pour half over the sponge fingers and spread out evenly.  5)Add another layer of sponge fingers, sprinkle with the remaining coffee mixture and then cover with the remaining cheese mixture. Chill for at least 2 hours and preferably overnight.  6)To serve, whip the cream and milk together until  fairly stiff and spread or pipe over the dessert. Sprinkle with the flaked almonds and then sift an even layer of cocoa powder over so the top is completely covered. Finally sift a very light layer of icing suger over the cocoa powder.  Salute! Buon appetito!  '),(20352,4218355,'21,August,2004','2004-08-21 00:00:00','   On Thursday, I made a  urlLink Summer Pudding .  It\'s a simple thing; just a pudding basin lined with bread, filled with a quick fruit/sugar compote.  (I have linked to the Delia Smith recipe, as mine only differed fractionally.  I made a 1 1/2 quart pudding, whereas she does a 1 1/2 pint one, plus I used 1lb blackberries, 12oz redcurrants, and 8oz strawberries, rather than her fruit choices.  A little more sugar, and bread too to cope with the larger quantities.)  Now this is a traditional refreshing summer treat, quite tart, but lovely and sweet too, made by many people all over the country.  However, I would not attempt this if the fruit did not come from the garden, for one reason.   Cost.   What is up with supermarket prices for berries??  Strawberries aren\'t too bad, a 1lb punnet cost me Â£1.80, mind you that was 1/2 price.  But redcurrants, and more over, blackberries and raspberries are ever so dear.   120g of Blackberries, ie. 1/4 lb, costs Â£2!  Redcurrants are about the same.  At that rate my pudding would\'ve had Â£8 worth of blackberries, Â£6 worth of redcurrants, and Â£1.80 of strawberries....  That\'s  Â£16 , before you\'ve added the sugar and bread!!  Alright, my pudding was rather large, and would probably serve 10, but even so....  Those fruits aren\'t  that  hard to cultivate, I think the problem must lie in keeping them fresh.  Their shelf life isn\'t fantastic.  I suppose you could buy frozen berries, you can get a mixerd summer fruit 1lb bag for ~Â£2, 2 of those with a couple of peaches thrown in would work, and cost about Â£5.  But the fruit needs to be fresh really...  Oh well, I hope Britains home gardeners keep up the good work, otherwise these sort of traditional treats are going to be priced out of reach...  '),(20353,4218355,'15,August,2004','2004-08-15 00:00:00','   Well my brother and I love mexican food, as does his girlfriend.  Now they are both devilish meateaters, :(, and regularly have burritos when they live away, but rather obviously make them non-vegan.  So I decided to try doing them as close to how my brother likes them, but vegan.  Surprisingly, it was exceptionally simple; no modifications hardly at all.  And guess what, my brother and his girlfriend like the vegemince and bean ones better than their normal ones.  They don\'t buy meat mince anymore. :D  pack of tortilla wraps (vegan -  urlLink Sainsburys own-brand  are) pot of salsa - make sure it\'s vegan, or make your own. 250g  urlLink RealEat Vegemince  350ml hot water fajita seasoning packet -  urlLink Discovery  brand is vegan by the look of it,  urlLink Schwartz  is as well. tin of red kidney beans 2 bell peppers - or mixed peppers to roughly that amount, small/medium strips 1 medium onion, finely chopped oil (optional) lettuce, shredded (optional) vegan  urlLink mayonnaise / urlLink sour cream   Get a large deep frying pan, chop the onions finely, and the peppers in to small/medium strips - fry them for a little while.  When the onions/peppers are semi-soft, add the mince and the fajita seasoning and toss the mixture together.  Add the water and mix well (some packet seasonings absorb water and need their own, add extra if this is the case).  Drain and wash the beans and add them to the mix.  Cook for about 10 minutes, or until there is no liquid left, but the mixture is still moist.  Warm up the tortillas in a microwave or under a grill for a short time, I find 3 together on a plate take about 30 seconds, don\'t do too many at once as they will all stick together.  Lay out tortillas, imagine the tortilla as being divided in to 5 equal width strips.  If you number these strips from left to right, 1 to 5, we leave 1 empty, put lettuce in strip 2, put the mince mixture in strip 3 (more mince than lettuce), put salsa in strip 4 (not too much, as too much liquid will make it fall apart), and then put a little mayo on strip 5 to help seal it.  Once you\'ve put all the fillings in, just roll it up, and tuck the bottom end inside. Voila!  NB. Do not put any thing too close to the edge, and don\'t overfill, or you won\'t be able to roll them up, and they won\'t stay like that if you do.  If you can\'t get Vegemince, which is the best in my opinion, then you could use soya mince, or even skip the mince, and replace the beans with a couple of tins of refried beans (check to make sure they\'re vegan).  Fajita seasoning often has some sort of milk fixer in it, like so many powdered things.  You can make your own, if you reckon you can judge the quantities you like.  They tend to be primarily: salt, chilli, paprika, onion powder, garlic powder, cayenne pepper, ground coriander, oregano, lemon/lime juice, cinnamon (hardly any) and a few other bits and pieces.  '),(20354,4218355,'13,August,2004','2004-08-13 00:00:00','   weed , n.    (a) A plant considered undesirable, unattractive, or troublesome, especially one growing where it is not wanted, as in a garden. (b) Rank growth of such plants.   Something useless, detrimental, or worthless.   Brambles are always referred to as weeds, as they are prolific over most of England at least, and have rampant growth, spreading out to all corners.  This description would hold with parts of the defintion of a weed (growing where it is not wanted, rank growth), and because brambles are covered in thorns, you could add troublesome.  However, I have always thought that the most important part of a definition of a weed, is the \"useless, [and] detrimental\" parts.  How can a plant that produces the most delicious fruit you can find be \'useless\', or \'detrimental\' for that matter.  They also look lovely at this time of year when they are in fruit.  The thorns may be an issue, but if I were a plant that produced such wonderful fruit, I\'d protect it too.  My nan has a thicket of brambles in her back garden, and she wants them all removed, as they are weeds. :(  I have been trying to show her that they are great, by picking a fair portion of the blackberries from them every week.  Then making some puddings from them.  To start off with I\'ve tried making a basic apple & blackberry filling, and using it to make a crumble and a pie.  To make the apple &amp; blackberry filling, I used the following.  8-10 oz blackberries (checked, and washed) 1lb 10oz cooking apples (peeled, then thinly sliced) 4 tbsp caster sugar 2 tbsp water 3 tsp arrowroot  Just dissolve the sugar in the water in a saucepan, add the apple slices, and cook down for a few minutes, then before the apples are very soft, add the blackberries, mash some of the blackberries down a bit.  Then add some arrowroot to thicken, stirring in.  Add more sugar if you need it, or more arrowroot if there is too much liquid.  (NB. Don\'t allow the apples to brown much, I put the apples in to the sugar solution as I cut them, this halts the browning a bit.  If not, put a tbsp of lemon juice in a bowl and toss the apple in it as you cut.  Then I just put the mixture in to a pie (using basic sweetemed shortcrust pastry) or in to a dish and cover with crumble topping (8oz plain flour, 5oz soft light brown sugar, 2oz butter (or substitute)).  Delicious they were!  If I get enough fruit at some point I might make some blackberry jam.  So brambles, they  are not  weeds, they are a great provider.  Nature loves them too, on my nan\'s patch there were ladybirds, wasps, hoverflies, leaf shieldbugs, spiders, flies, mites, even a frog had come! :)  '),(20355,4218355,'13,August,2004','2004-08-13 00:00:00','   Until this week, I have never been to the cinema alone. Sad thing to do I always reasoned, after all, the cinema and films are a social thing, something you do with friends. I could not understand how some of my mates managed to go and see films on their own without feeling awkward.  This has changed...slightly. :S  Recently, I purchased an \" urlLink Unlimited Card \", which is a subscription service for the cinema I use. Â£10/month for unlimited viewings. I\'ve had it for 2 weeks, and have seen 4 films so far, IIRC.  Now as I am a bum, and have lots of free time, thankfully free as I wouldn\'t be able to afford it if it cost me anything, I could spend my time at the cinema. However, lots of my friends don\'t have the card, and couldn\'t afford to go often. Besides, many of my mates have jobs, some people do apparently, and can\'t get the time to see some stuff. But I want to see things! Lots of things!  The last problem is my taste. I like many films, not a huge fan of serious realistic films, often described as \"gritty\", but I like the rest. This particularly includes \"kids\" films. Now last time I looked the age certificate for a film was a lower bound, not an upper.... Why does everyone grow out of good stuff? Mad people.  So I bit the bullet and went to see  urlLink Garfield The Movie  on Wednesday, and  urlLink Home On The Range  this afternoon (Thursday).  So I sat in my usual seat, on my own, looking like a pillock, but that\'s normal, amongst a cacophony of parents and little kids (the schools are all on holiday), and... enjoyed the show.  It was a bit awkward, but I wouldn\'t have got to see them if I didn\'t go by myself. So unless I get friends that actually share any of my interests anytime soon this might become a regular feature. Tons of films to see this month after all!  '),(20356,2536117,'21,May,2004','2004-05-21 00:00:00','   awww, the memories!  i shant ramble on about the last day again... for those of you that want to read reams of waffle, check out  urlLink my blog here .  here are all the prettiful pictures from that A-mazing day...enjoy them lublies, and treasure them forever!  OHOH i also STOLE (according to nazreney) pictures of the ADORABLE eliana, drapy\'s GORGEOUS ickle SWEETHEART of a baby, (apologise for the crapness of the quality, but taking a picture of a picture really isn\'t advisable unless in UNIQUE and SPECIAL cases...which this one undoubtedly was) but i thought it inappropriate to post those up on the internet, what with all the freakay PAEDOPHILES around... and no i don\'t mean farah. anyhoo, if you want a copy to KEEP for YOUR VERY OWN SELF, (yes, you can cut out and stick it on your wall if necessary - wow, we really ARE rather paedophilic), emailey moi or just ask and once i\'ve verified that you\'re a tiffinian (paedophile), shall sendy them asap with my \'fastest fingers of the west\', as dubbed by nazrene. (although technically, she deserves that title, her fingers are just like WHOOOSH WOAH WOAH BLURS!)  OH OH OH and also managed to clicky away and niftily get a photo of chris...YES THE 11 YEAR OLD JOHNNY DEPP IMPRESSION DUDE, PURE GENIUS - god it\'s THE BEST!! but likewise, didn\'t want thousands of adoring johnny depp fans following him around and demanding that he \"do it again\"! wait, that was us. and yes farah, you can have a copy of your beloved boy.  must also apologise for any offensive captionning, i know that frankie shall not be best pleased with the title \'le frank\' - in fact, i think i may be forced to change it, or suffer the very DIRE consequences of frankie\'s WRATH.  anyhoo, enjoy these little gems my fellow 11w-elites (and although we\'re very much THE BEST, i meant that in an \'e-lights\', we\'re just SO FAR OUT, WE\'RE OUT OF THIS WORLD kinda way...damn i make no sense.)  much love xxx   '),(20357,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    canbury gardens...oh my, aren\'t we a bunch of rebels!&nbsp; urlLink   '),(20358,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    the stairway to heaven...wait no, that\'s just malps\' bedroom.&nbsp; urlLink   '),(20359,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    who rules the school???!!! why, 11w of course!&nbsp; urlLink   '),(20360,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    wow prettiful ickle y7s...oh i am mistaken...tis the unruly 11w bunch!&nbsp; urlLink   '),(20361,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    the corner of cool!&nbsp; urlLink   '),(20362,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    shivani and charlie...the beautiful couple that they are&nbsp; urlLink   '),(20363,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    le frank et la flea...&nbsp; urlLink   '),(20364,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    sorry, this one\'s kinda dark...i\'m also squashing phil.&nbsp; urlLink   '),(20365,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    11w are a bunch of beauties!&nbsp; urlLink   '),(20366,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    11w ruling pizza hut!&nbsp; urlLink   '),(20367,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    woah freakay cheek suction! farah is the QUEEN of carazy photos (see below and you shall see what i mean...)&nbsp; urlLink   '),(20368,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    two fine young lollipop ladies naz baz and shivs&nbsp; urlLink   '),(20369,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    oooh, kinkay grinding!&nbsp; urlLink   '),(20370,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    annio dudette&nbsp; urlLink   '),(20371,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    the cheeky duo cath and rachel looking like angels...bootiful&nbsp; urlLink   '),(20372,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    that\'s more like it! show \'em how it\'s done annio and roo!&nbsp; urlLink   '),(20373,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    the charlie\'s angels three...no longer&nbsp; urlLink   '),(20374,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    terrified holly (!) and chaely&nbsp; urlLink   '),(20375,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    not so terrified holly and chaely&nbsp; urlLink   '),(20376,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    my drunken sailors! thankfully sober...&nbsp; urlLink   '),(20377,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    awww moi, shivani, saimy and alice&nbsp; urlLink   '),(20378,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    mes lublies - heh shivani :P&nbsp; urlLink   '),(20379,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    les cousinas preferee! aww look at nazreney\'s gorgeous smiley! OH AND ALSO ms davies shocking pink \'glamour\' folder on the desk...good choice nazreney!&nbsp; urlLink   '),(20380,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    the gorgeous smiles trio - chaely, becs and rajin :D&nbsp; urlLink   '),(20381,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    my lubly laurie&nbsp; urlLink   '),(20382,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    cutesy rachel&nbsp; urlLink   '),(20383,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    complete and utter cheese...but it gets me every time! my lublies shiv, malps, phil et moi&nbsp; urlLink   '),(20384,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    aww my favourite auntie et moi&nbsp; urlLink   '),(20385,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    the CARAZY ellie and carole in a rare moment of calm&nbsp; urlLink   '),(20386,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    awww just bootiful - look at me sneaking on up there with my camera!&nbsp; urlLink   '),(20387,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    awww my lublies sarah, lauren and freya&nbsp; urlLink   '),(20388,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    me and the funkay cuthbert... de-cuthbertised&nbsp; urlLink   '),(20389,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    alice et saimy - a picture of prettifulness&nbsp; urlLink   '),(20390,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    cuthbert, the hounslow rudeboy...apparently i\'m good friends with this stranger.&nbsp; urlLink   '),(20391,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    cuthbert avec cuthbertrice&nbsp; urlLink   '),(20392,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    heh...tis undeniable, we ROCK!&nbsp; urlLink   '),(20393,2536117,'19,May,2004','2004-05-19 00:00:00','   urlLink    GO WATSON GO! wooh, us and our defining watson signal!&nbsp; urlLink   '),(20394,2536117,'19,May,2004','2004-05-19 00:00:00','   heh... those were the days... no gcses to worry about... sigh.  anyhoo, these are the pictures i took when we were making that scrapbook for miss roberts. and now our dear room 16 is all burnt to ashes...  enjoy!  much love xxx   '),(20395,2536117,'16,May,2004','2004-05-16 00:00:00','   urlLink    shake it...shake it shake it shake it... shake it like the WATSON DRAPER!&nbsp; urlLink   '),(20396,2536117,'16,May,2004','2004-05-16 00:00:00','   urlLink    libby what are you doing?! trying to catch a couple of flies with your mouth open like that? (rachel, libby, holly, annie)&nbsp; urlLink   '),(20397,2536117,'16,May,2004','2004-05-16 00:00:00','   urlLink    heh...watch out for ms da cruz on the prowl holly, rachel, annie and libby!&nbsp; urlLink   '),(20398,2536117,'16,May,2004','2004-05-16 00:00:00','   urlLink    feel the funkay music! (holly, annie, rachel and libby)&nbsp; urlLink   '),(20399,2536117,'16,May,2004','2004-05-16 00:00:00','   urlLink    land ahoy! (capt\'ns rachel, cath and annie)&nbsp; urlLink   '),(20400,2536117,'16,May,2004','2004-05-16 00:00:00','   urlLink    ATTENTION!(nikki, ellie, philippa)&nbsp; urlLink   '),(20401,2536117,'16,May,2004','2004-05-16 00:00:00','   urlLink    mareen, malpreet and shivani, charlie\'s angels stylee...or perhaps not.&nbsp; urlLink   '),(20402,2536117,'16,May,2004','2004-05-16 00:00:00','   urlLink    da-da! the freya and nikki double act...coming soon to a school near you!&nbsp; urlLink   '),(20403,2536117,'16,May,2004','2004-05-16 00:00:00','   urlLink    heh... awww schlimey, you know that\'s not what a trophy is for!&nbsp; urlLink   '),(20404,2536117,'16,May,2004','2004-05-16 00:00:00','   urlLink    wow ellie...damn flexible!&nbsp; urlLink   '),(20405,2536117,'16,May,2004','2004-05-16 00:00:00','   urlLink    ahhhhhhhhh, scary farah! run away, run away!&nbsp; urlLink   '),(20406,2536117,'16,May,2004','2004-05-16 00:00:00','   urlLink    woah...freakay carole...beware, danger ahead!&nbsp; urlLink   '),(20407,2536117,'16,May,2004','2004-05-16 00:00:00','   et voila! mes lublies all looking bootiful, as usual. wait! oh dear god annie, what were you on?! the dead lizard pose is now renowned and shall be making it\'s way into a magazine near you soon!  enjoy these exotic pics! (oooh naz, exotic = kara, you\'ll fit right into the threesome now!)  much love xxx   '),(20408,2536117,'16,May,2004','2004-05-16 00:00:00','   urlLink    um, you can\'t really see us (saima, chaely, alice, naz et moi) properly here, but look! it\'s our coolio american surfer dude waiter in his hawaiin costume, yay!&nbsp; urlLink   '),(20409,2536117,'16,May,2004','2004-05-16 00:00:00','   urlLink    the gorgeous phil and saima, otherwise known as flea and schlimey&nbsp; urlLink   '),(20410,2536117,'16,May,2004','2004-05-16 00:00:00','   urlLink    aww my lublies&nbsp; urlLink   '),(20411,2536117,'16,May,2004','2004-05-16 00:00:00','   urlLink    beauties&nbsp; urlLink   '),(20412,2536117,'16,May,2004','2004-05-16 00:00:00','   urlLink    the birthday girlies&nbsp; urlLink   '),(20413,2536117,'16,May,2004','2004-05-16 00:00:00','   urlLink    the pouty pair&nbsp; urlLink   '),(20414,2536117,'16,May,2004','2004-05-16 00:00:00','   urlLink    my sweeties, lauren and cath&nbsp; urlLink   '),(20415,2536117,'16,May,2004','2004-05-16 00:00:00','   urlLink    my hula chickies and me... directly beneath the \'toilet\' sign. meh.&nbsp; urlLink   '),(20416,2536117,'16,May,2004','2004-05-16 00:00:00','   urlLink    oh dear god annie...&nbsp; urlLink   '),(20417,2536117,'16,May,2004','2004-05-16 00:00:00','   urlLink    saturday night fever!&nbsp; urlLink   '),(20418,2536117,'16,May,2004','2004-05-16 00:00:00','   urlLink    hula partay!&nbsp; urlLink   '),(20419,3631998,'04,August,2004','2004-08-04 00:00:00','   Well I got me a  urlLink gmail  today thanks to  urlLink JosÃ© Rojas . How did I get it? Go to  urlLink Gmail Swap  and find out. But I\'m here to talk about all the free stuff on the web.  What really is free in the world? I know of two things 1)  urlLink Sindome  and 2)  urlLink Starwars MUSH . They are the only two things that are really free on the internet that I know of.  Why? It\'s simple, every free service on the internet isn\'t free. There is a cost to maintain a server and when there are costs there is money. Let\'s take e-mails for instance, there are web servers, data servers and personal to pay for. But you aren\'t paying for them, as neither am I. Or are we? Yes we are. Every time we access our free online e-mail we generate renevue for the ones that give us free stuff.  Renevue is directly connected to the number of times a web page is shown, but in this case it is advertising. If you thought banners, you\'re right. Every time you go to  any  website that has a service that is free there is at least one banner on it. And I must say that I love banners, simply for the fact that they just take a small portion of the page and do nothing until you click them. But I hate popups, popups are the devil spawn and should be forbidden of existence. They are simply the most annoying thing on the whole internet (apart from 13 years old boys who think they know everything).  But back to the topic, there is no such thing as free stuff on the internet. Someone is paying for it.  Internet = cost . And whenever you see a banner you know that the page you\'re seeing right there was paid for the guy who ordered that banner. And by having that banner on the page, you increased that advertisement exposure by 1.  Multiply that one by the various pages an average user goes thru on a day and by the number of users that service has and you\'ll know that I\'m talking about a lot of money. That pays your free service and even generates income for the company, lots of income... Trust me if there isn\'t profit they wouldn\'t do it.  Now back to free e-mails, Google\'s Gmail will offer 1Gb (one gigabyte!!!) of space for your e-mails and that\'s a lot! Yahoo has raised it\'s quota to 100Mb in response. So what\'s the deal? Is Google using all it\'s money from selling shares on the market to buy storage space?  Not really. The space you have in any e-mail account is virtual. That means that they don\'t reserve 1Gb or 100Mb for each account. They raise the account space as it is needed. Something like if you have 2mb of space used you\'ll have like 4mb freed for your account once it gets close to that they\'ll get you another 2mb and so on... That\'s how they can offer thousands of people 100Mb of free space. Most of the users stay on the 50% usage of a 4mb account just so they can receive larger e-mails when needed. But with 1Gb people will start to go nuts and pass on larger and larger e-mails. How Google will cope with that I do not know, but I\'ll be testing a big fat email soon. Maybe I\'ll send myself a DVD rip of some movie I have or something... 600mb on one e-mail. [Edit] It appears that the maximum size of any email you can send or receive from Gmail is 10Mb Still good enough [/Edit]  The catch of Gmail is directed advertising, they want you to keep your e-mails so they have a good base on what interests you have and other stuff. The procedure is totally automated and scans for keywords on your mails. It\'s neat stuff, revolutionary and will be a success. Imagine having banners that are pertinent to what you like without having some type of worm or cookie installed on your computer? The rate of clicks by ads shown will raise sky high.  Way to go Google. '),(20420,3631998,'04,August,2004','2004-08-04 00:00:00','   I like black. I have black as the background of my computer screen, on the background of my NetBeans and text editors. I love it. But I am not fond of pastel greens over black, don\'t ask me why. Maybe it\'s just too much Cyberpunk influence, but for me green on black has to look like it will burn your eye. Full glowing green with the faint ones being at 70% when on the minimum.  And I like blue, and blue on black too so I took some time from my working day to make this thing blue on black. Gave it some fades and tried to make the colors on the sidebar not get more attention than the post itself. So there you have it, new colors and duotoneish style.  Oh yeah, if you can\'t read the darker stuff, your computer screen isn\'t set up correctly. It\'s not my fault. I can read them all. Fix your brightness and contrast, I won\'t change it back. '),(20421,3631998,'04,August,2004','2004-08-04 00:00:00','   Sometimes when I\'m playing RPGs (either online or not), I get the feeling that everyone is an elf. Really, most of people describe their characters as beautiful people, thin, tall and even  tanned ! And when I say most people, I\'m talking about a big fat portion of them, if the ones that don\'t have greek gods as their characters would be put on a pie graph, their slice would be slim.   I mean, come on! Why does everyone  need  to be gorgeous, sexy and fit? I\'m known to have some of the weirdest characters all around and I play mostly for the fun, than for giving my ego a boost up. People, please get real and break that barrier where your character is a representation of yourself like you would like them to be. Some of the players of both online text based RPGs that I play (Sindome and SWMUSH) are already there and it\'s a bliss, but even in those places much of the options on character creation are left behind and overlooked.   Why? People like to look good, they feel pretty... It\'s the subliminal connection between yourself and your character. You want to be accepted, you want to be among the \'cool people\'. Well guess what? You don\'t! It\'s a game, make your character so you get the most of it, and in the RP focus make him different. Always remember different doesn\'t mean better, never. On a world of elves the dwarf stands out. If everyone is gorgeous, tall and fit as a bull. Create something different, give them scars (and not those cool ritualistic ones), make a Sloth from goonies and see how much fun you\'ll have everyday instead of the same thing day after day.   But as far as looks goes, it isn\'t the biggest of problems. One of the biggest I found everywhere is what I call over-especialization. It\'s like you have one character and all that he can do is this and nothing more. He\'s the best on this but if he\'s ever alone and he needs to do that, he\'s doomed. It\'s like having a car jacker that only drives automatic cars, once he gets his hands on an expensive as hell ferrari he\'s done. He\'ll never take the car out... RPGs are all about diverse situations where you\'ll need to use lots of skills in order to get away in one piece.   On Neverwinter Nights there are loads of people that create a character that is the best at fighting and can kill a dragon alone... Guess what? One trap on the floor and he\'s history. Why? No variation, the same thing over and over again. See monster, knockdown, devastating critical, rinse and repeat. Even mages on nwn are repetitive as hell, the guy doesn\'t even change his spells to see what he can do. No it\'s bigby\'s, bigby\'s, bigby\'s, bigby\'s, missile storm, missile storm, rinse and repeat on the next target.   And people say that the game is getting dull. Why? Because of you ( and sometimes me too, I admit ). Try to broaden your RP by thinking outside the usual box and aiming for something that makes you unique. Once you got that, no day will be dull. '),(20422,3631998,'04,August,2004','2004-08-04 00:00:00','   I don\'t know if you noticed, but now I have a footer on every page that looks like this :         Unless otherwise expressly stated, all original material of whatever nature created by  urlLink Diego M. Diehl  and included in this weblog and any related pages, including the weblog\'s archives, is licensed under a  Creative Commons License .      Now what the hell is that? That is another bright idea that started on the internet and thru it became versuccessfulll. What that is? It is a copyright that allows the use of the work or part of it.  Let me explain this on a simple way. Everytime you create something, that automatically has a copyright on it. You made it, you spent your time in the creation, it\'s yours. No-one can copy, use part of it, distribute or any other thing without your consent. That is the \'Ã‚Â©\' next to logos and on every book you\'ll find on a bookstore.  But since the popularization of the internet one thing came up quickly : cooperative work. We suddenly have lots of people from around the world working on the same program/book/music. And there was the single person that made something that he wanted to be distributed thru the internet. Be it a free software or a book. With that raised the GNU license that has it\'s pro\'s and cons and normally only applies to software.   urlLink Creative Commons  created a few \"copyright licenses\" for normal people. That means no lawyers, nor confusions when trying to use someone\'s work. Each license has two versions, the easy-to-read-and-easy-to-understand one and the legal wording one. Both say the same thing one is for the average people (you and me) and the other is for the lawyers of the world that like to speak weird stuff and read 3000 lines affirming something that can be said in one line and without fuss.  So Creative Commons made the (CC), something that states: \"I made this, but you can use it if you do this and that.\" Simple, easy and fair to all. And better yet CC covers from text to music and other stuff.  Clearly everyone that has something online should check  urlLink Creative Commons  so they can assure their rights and still allow people to use their work. Like I do.  My type of CC License, says simply that you can use anything on this blog and by use I mean : copy, distributexhibitbit, execute, create something using this as a starting point, and even use commerciallylly. But you have to give me credit for what is my work and everything you do has to have the same License. It\'s that simple.  Use freely but add my name to the damn thing so I get the credit I deserve.  Cool?  I know I am. '),(20423,3631998,'04,August,2004','2004-08-04 00:00:00','   Today I updated my  urlLink Yahoo Messenger  and it came with a thing called  urlLink LAUNCHcast Radio  sure it is Yahoo pushing another of their services to their customers and shit. But since you can create your station for free and themusic quality is good, I gave it a try. And I must say this : COOL!!! &nbsp; Their casting is smooth and works well. I already have something like 2-3 Gb of MP3s at my work computer just to keep the variety going, but withbroadband internet acess and the radio I\'m pretty sure my horizon of musichas just exploded. Another thing on  urlLink LAUNCHcast  is that you can rate almost everything, from genre to artist, albums and song. With that in place they\'ll keep playing the stuff you like and avoid what you don\'t like. If they\'re going to sell that information, or use it to clean their buttocks I don\'t really care. I want a soundtrack to work with and I don\'t care about anadvertising here and there.So if you want to check my station  urlLink click here '),(20424,3499599,'31,May,2004','2004-05-31 00:00:00','   cuddle and a kiss on the forehead - you like to be close to your special someone and feel warm, comfortable, and needed    urlLink   What Sign of Affection Are You?     brought to you by  urlLink Quizilla      Nihilist Bear    urlLink   Which Dysfunctional Care Bear Are You?     brought to you by  urlLink Quizilla      You are a Siren. More adventurous than all with a voice like no other you sit on warm rocks and sing to the moon and sea. Yet sometimes shipwrecks find you and raving men want you. You are a bottle of talent and power. What the unknown is you seek to find, and a lover. You have the moon and stars as friends. There are a very few of you, what a rare find.    urlLink   What kind of mermaid are you? (Gorgeous Pics)     brought to you by  urlLink Quizilla  '),(20425,3499599,'30,May,2004','2004-05-30 00:00:00','   I\'m gonna go insane. Well, probably not. But I feel like I am.   I\'m wasting my hours away to stupid games and doing absolutely nothing. I don\'t know why I always end up playing games when I know I could be doing something much more productive.   My latest addiction is a free online game called Runescape. (www.runescape.com) It\'s the worst game I\'ve ever played. But I can\'t stop going back to it. I dunno why. I mean, it seems to collect the stupidest people I\'ve ever seen on one game. The n00bs are incredibly irritating...the elite assholes are a pain..and..I really have no idea why I play that game. Oh well.   Anyway. My name on there is The_Thiefess, if you ever want to find me. I normally play on World 20 because that\'s where the people I know are at.  So. Let\'s see. Something to write about...  Ooh.  Every summer, I always go through a I-want-to-start-a-garden phase. My Mom normally goes out and gets lots of flowers and stuff like that, and then I decide that I want my own little garden. And we go get some flowers and herbs that I like..such as basil, greek oregano, snapdragons, and..all sorts of things. It\'s fun. But, after about a week, I get bored of that lil\' project and move onto something else. It\'s aggravating. I can never stay with one \"project\" for any long period of time. I just seem to lose focus and start thinking up ideas for something else to do. And even if I test myself, by doing some activity just to see how long I can do it for...well, I always end up forgetting about it or being distracted. Sometimes I really feel like I have some weird form of ADD =p Well, most of the time I just feel like I\'m paranoid. But every now and then I feel like I have ADD. So. enh.  An example of my problem with projects...well, there\'s always my procrastination with projects that are forced on me. But I\'m talking about projects that I start and never finish. Oh! I used to be on a chat-program called Furcadia all the time. The site is www.furcadia.com if you\'re interested. You\'re basically an animal that can chat with other people-animal things and there\'s absolutely no point. But it\'s a lot of fun.  Anyway. You can make your own dream in furcadia. A dream is a map of your own little world, in simple terms. You can make patches to make pretty, extra graphics and use DragonSpeak to do HTML-ish stuff. It\'s difficult to explain. In any case, I used to have all these wonderful ideas for dreams. I would start planning how they would look..and what patches I would use..and what bots to make...and I would get started and work on my dream for hours on end. However...eventually, normally when I got half-way finished..I would start thinking up ideas for another dream and just completely forget about my current dream. And...that would go on for a long time. It was frustrating, to say the least.   I loved Furc, though. It just seems that Furcadia attracted a certain type of person that...if you get along with one person there, you get along with nearly all of them. Nearly everyone is intelligent in their own way...and everyone\'s funny and...oh, I suppose I was just spoiled by that chat program. Because..comparing the type of people on Furc..to Runescape..agh. It makes me want to go back to Furcadia..which is really bad. heh. I can\'t even begin to count how many hours of my life I wasted to that pointless game XD Well, it\'s not even a game. There is no point. At all. Whoo! Rock on.  Oooookay. Where was I? Oh. Gardening. That\'s kind of sad, how far off topic I got. heheh. Anyway. So my Mom had to go get some plants and I was slowly but steadily getting back into my I-wanna-start-a-garden mood and I kept asking her about all these different plants and..well..when I decided on what I wanted, she wouldn\'t give in. enh. I mean, come on..it\'d be a lot of fun to have a Weeping Willow. Or a rose bush. Those are just cool plants. Weeping Willows are just...well..really cool. And the rose bush I wanted was a \"Yellow Tea Hybrid\" or something and it was neat. But nooooo. We couldn\'t get either of those. bleh. And when I just said I wanted a little basil plant or a greek oregano..noooo...  It\'s probably for the best, though. ^^; I don\'t have a green thumb. Well, rather, I\'ve never stuck with gardening long enough to know if I have a green thumb. At least, it\'s very pink right now. Pink-ish. Kind of tan. Stupid summer. I always get a horrible farmer\'s tan due to tennis.  I think I\'ll just flood this little journal with lots of quizilla stuff later. I love quizilla ^_^  Oh, and every time I post I\'ll try to put at the bottom the song that\'s been stuck in my head all day. I always have a song stuck in my head for at least one day at a time..sooo.  Song stuck in my head today: Holding Out for a Hero (Shrek 2 Soundtrack) by - Frou Frou, Jennifer Saunders, and the original (that I dislike) is by Bonnie Tyler, I think. I get my music from Limewire so the names aren\'t always...ehm..accurate.  '),(20426,3499599,'30,May,2004','2004-05-30 00:00:00','   I am so very weak. Or at least, I feel so very weak. I have finally given in to this silly fad of \"blogging\" stuff that nobody cares about. Yay.   I\'ve decided to do this because.. 1. It\'s fun to read some of the few interesting blogs I\'ve seen. I just need to remember to fill this thing up with inside jokes and stupid stuff like that. Nobody cares \'bout that kind of info. 2. I need to keep a journal of some sort. If only to be able to read back over previous emotions and moodswings..in any case, if I stick with this, it\'ll hopefully help me get a better control over myself. 3. It\'s only my second day out of school and I\'m already incredibly bored. Sooo..why not start a blog?  Listing stuff always makes me feel better. ^_^ I\'m one of those people that organizing and cleaning helps produce endorphins just like exercising and that sort of stuff does.   I wonder who made up the word endorphin. That\'s just a really weird word. Wierd. Weird. Ok yeah.  Anyway.  To those of you who know me...some of these posts will make more sense than to random people just looking for time to waste by reading a stranger\'s blog. But...I don\'t know if I\'m going to even give out this link to my friends. And I have no idea if this will be something worthwhile to people that don\'t know me. enh.  I guess this is good enough for a first post. I\'ll probably write more later, considering there\'s nothing else to do and I\'m already feeling bored again, just by the fact that I\'m ending this post..oy..  '),(20427,3499599,'29,June,2004','2004-06-29 00:00:00','   Mean lil fellow, arn\'t you?    urlLink   What Monty Python Character are you?     brought to you by  urlLink Quizilla       urlLink                I\'m A 1970s Geek     You\'ve decided for the world that it\'s time for a change. JOIN THE GEEK REVOLUTION!      urlLink find your geek decade at spacefem.com             you are saddlebrown  #8B4513         Your dominant hues are red and yellow... most of what you do is motivated by your need to change things and have a good time, but you\'ve been known to settle down and think out a situation, too. You tend to surprise people just when they\'re starting to feel like they\'ve got you down.  Your saturation level is high - you get into life and have a strong personality. Everyone you meet will either love you or hate you - either way, your goal is to get them to change the world with you. You are very hard working and don\'t have much patience for people without your initiative.  Your outlook on life can be bright or dark, depending on the situation. You are flexible and see things objectively.         urlLink the spacefem.com html color quiz  '),(20428,3499599,'27,June,2004','2004-06-27 00:00:00','   So. I\'ve been hooked on Age of Mythology lately. It\'s fun to play online. Well, mostly fun. The stupid people..they\'re not so fun. And the laggers. And the people that bail, five minutes into the game. Ok ok ok the cons outweight the pros e.e And that\'s probably why I\'m my little play-AoM-24/7 spree is about to end..or I\'m going to make it end..  I got a little..furious..this afternoon. I was playing a rated game against a guy who used an incredibly cheap trick and rushed my village at the very beginning of the game. I barely stayed in there..I made new towncenters and lots of migdol structures (was playing as egyptian..Isis, actually)..but eventually my resources were gone and there was no point to keep on playing, since he kept killing off almost all of my people, in which I would have to rebuild in another remote location on the map..again. So I resigned. The after-battle conversation: Him: gg (good game) Me: That was not a good game. That was a cheap trick that you used at the beginning. Him: so u r a girl? Me: Don\'t try and hit on me because I\'m really pissed off right now... Him: how old r u? Me: AGGH Me: DAMMIT  And I left in a very bad mood ^_^ Was very angry to find out that my punching bag downstairs has been taken down and that I didn\'t have anything to take my anger out on.   heh. And now I\'m angry again. Too much testosterone in online games..bleh. There needs to be more girl gamers >  I think I\'m gonna go find some stupid quizzes to put in this blog now..  Song for today:  urlLink Killian\'s Red  by Nada Surf  '),(20429,3499599,'27,June,2004','2004-06-27 00:00:00','   yeah yeah ok..i got lost in some Tamora Pierce series for a few days there. enh.  and right now i\'m thoroughly depressed. just watched the new Peter Pan, which i love..but..it\'s still depressing. i don\'t wanna grow up ;.;  sadness.  oooh...but..i want  urlLink this   it\'s..pretty. waaaaant.  and  urlLink this   mmm.   Moving on.  I love  urlLink bash  ^_^  MasterG: .....................................................................           .................................. judas: where\'s pacman when you need him?  Nudger: The Internet is great, but the best thing it\'s ever done is when a bored schoolfriend of mine set the school\'s website\'s background to Goatse. Antifreez sighs Antifreez: So many memories attached to that song. Nudger: It was so great when the IT teacher opened the page on a huge projection screen, and without looking at it, said \"We, shall be attemping THIS, boys!\" Nudger: I swear, I was almost sick from laughing.  tRonz0r: I\'m $4 poorer, 1 bandanna richer and I still don\'t look like a pirate  tRonz0r: yarrrrrr, tis a sad day.  Elbonio: You know how women say \"I wouldnt sleep with you even if you were the last man on earth\" ? Well if you WERE the last two people on the earth, i hardly think sex is the first thing on your mind... it\'s clearly going to be \"let\'s raid Toys\'R\'us..\"  Rach: when i get my old puter back i\'ll send you some songs Olly: What happened to it? Rach: i broke it Olly: On purpose?   Rach: i wouldnt say PURPOSE Rach: just stupidly Olly: Ahh Olly: So you do know exatly what\'s wrong with it? Rach: yeah, it\'s full of water  Since Jacob\'s gone, I have nobody to paste bash quotes to...so..might as well flood my blog with \'em, since I\'m not going to do anything else with this stupid journal thing XD Yay.  aggh. had to change the quotes around a lil bit, so blogger would post them..all this for your viewing pleasure, folks -_-;  I feel better all ready. I love bash x.x  Oh, and go visit  urlLink Jeremy ...he\'s all feministic-ish suddenly...w007...heheh.   Song:  urlLink I am the Highway  by Audioslave (greaaat song ^_^)  '),(20430,3499599,'21,June,2004','2004-06-21 00:00:00','   urlLink Funny comic.  '),(20431,3499599,'20,June,2004','2004-06-20 00:00:00','   Enh. I won my first match, and lost my second. I was playing excellent-ish-ly this morning. I just got lazy after lunch o.-;  I don\'t really mind losing, though. I lost to a person that was nice. And she was polite and showed all the right courtesy for a tennis match. The other girl, though...agh.  1. She tried to hit me. Many times. When I would turn around to go back to serve and she had gone to get me a ball and was throwing it over, well, let\'s just say I had to dodge the ball a few times. \"Oh...sorry\" yeah right -_-  2. She questioned my calls, after she kept hitting them out and they were about a foot from the line and..oh, I don\'t know. But..well, after the match, my Mom said \"It\'s a really good thing she said that, because I knew you would be so pissed off that you would actually start playing.\" And she was right e.e That\'s my ultimate pet peeve. I mean, I try very hard to keep my line calls right. Occasionally I\'ll just let one slide, if I\'m not really sure about it. Aggh.  3. The people there to support her kept clapping when I would hit a ball into the net. My Mom did that a few times, at my first matches. I glared at her..a lot. I cannot stand it when people do this. It\'s okay to clap if someone hits a winner, or a really good shot. But..clapping just because a ball went out, or went into the net? That\'s my other ultimate pet peeve in tournament matches. It gave me enough anger to win, at least e.e;  4. Ok..the first set was 6-2, me. She was coming back in the second set and we were tied 5-5, and a couple of times she kept holding onto the tennis balls because she figured that the game was over..like..it would be 40-0, her. My serve. And instead of giving them to me, she would keep them because she just assumed that the game would be over the next point and..enh. I was just angry o.x;  Anyway. I won..6-2, 5-7, 10-6. I won the first set, she won the second set (I kept kind of screaming \"NOOOOOOoooooooooooo\" in my head during those last points), and then she just lost it when we did the 10-point tiebreaker. That was when the clapping was at its worst, and her \"Oh my GOD come ON Caroline! GET IT TOGETHER!\" peptalks were also at their worst. ^_^ It makes me laugh when people talk to themself like that, even though I know I do it too. heheh.   Oh yeah. My second game I lost 6-2, 6-0. x.x; Like I said, I got really lazy. And really tired. I\'m out of shape >  I think I\'m gonna go sleep now. Or just wander the net aimlessly. Neither of those require the use of my brain so..yeah.  Song for today: Pass That Dutch by Missy Elliott heheh. I found this by looking for the Mean Girls soundtrack. I like Missy Elliott\'s music ^^; My favorite part is near the end, when it sounds like the water-pipe part of Stomp.  '),(20432,3499599,'20,June,2004','2004-06-20 00:00:00','   too much family togetherness today. and too much chocolate. which leads to too many moodswings.  I\'m in a very depressed mood right now, when I was just in a very happy mood about..oh..ten minutes ago. There are a few reasons for my sudden depression that I can easily name, though...  1. I\'ve had to deal with my two sets of grandparents since 10 this morning till 8 p.m. It\'s very exhausting, answering the same questions every time when we\'re with the grandparents that practically have alzheimer\'s/dementia, and being at the second grandparents\' house, I\'m always criticized for every little thing I do. I hate these kinds of days..father\'s day..mother\'s day..well, basically any sort of day that I\'m forced to go visit relatives and get no gifts in return. I\'ve also had to deal with my parents in the car. Sometimes I wish they would just divorce already. They fight over every little thing, and it\'s not the kind of nice fighting where people apologize in the end and make-up. It\'s just..enh. It\'s enough of a warning for me to not want to ever marry, at least. I don\'t understand marriage. I mean, normally people \"fall in love\" because they talk to a person a lot and...well, if you\'re stuck with a person long enough, your mind can normally force you to find some positive aspect about them. But that being-together-ness for...say, twenty-five years? It\'s enough to drive anyone insane. The idea of marriage is just fucked up, in my opinion o.-; But then, I\'m having some raging hormone-induced thoughts right now that are also triggered by my overdosing on chocolate after I had lunch today..  2. I had a  lot  of chocolate. Chocolate leads to most of my chaotic moodswings. I\'ll probably be feeling either quite happy in a few minutes, or angry at the world, or just very sulky and sullen. I dunno.  3. I was reading  Trickster\'s Choice  by Tamora Pierce today, in order to avoid most of the pathetic attempts at conversation that my family tries on me. I love Pierce\'s books, but they just...leave me feeling so incomplete. Right now, I\'m pretty much yearning for my own Nawat Crow and wishing I could live in Tortall. >  4. I have a tournament tomorrow in Winston that..the girls in my category..well, almost all of them have beaten me once. It\'s..blehhh. wish me luck -_-  Ok. After typing all that, I feel better now. I think I\'m just going to publish this without reading over it, because if I read over it, I know I\'m going to edit some stuff.  Song for today: Hello Somebody by Handsome Devil I\'ve only been able to find this song, and Makin\' Money by a group called Handsome Devil..and I love both of the songs ^^; I wish I could find more, though. hn. I love the part..\"If I seem a little rude, please excuse me, I was raised on the dark side\"..XD The twinge of vibrato their lead singer has makes the songs great.  '),(20433,3499599,'19,June,2004','2004-06-19 00:00:00','   urlLink LMAO  '),(20434,3499599,'16,June,2004','2004-06-16 00:00:00','   So. I\'m still here at journalism camp. Even though I\'m here for literary magazine...it\'s just easier to refer to it as Journalism Camp. Whoo.  Right now, my fellow literary-magazine people are using PageMaker to create layouts. I\'m thoroughly confused and quite lost, since I\'ve never used the program before. I\'ve had fun tinkering with it, though. It\'s just getting a little boring. Sooo I felt like writing on here.  And..I have nothing to ramble on about. I could vent my anger about how  wrong  our literary magazine is, as I\'ve found from the lecture things I\'ve been sitting through..but..I can save that until tomorrow when I get home.  I\'m kind of homesick, but really I just wish I could be with some people I know. I don\'t mind being placed in situations where I\'m basically on my own and have to make friends..it\'s ok, if I\'m in a good mood. But I had to deal with new people last week..and this week I\'m outnumbered, and out of my area of..expertise, I suppose. Tennis camp is..a lot different than journalism camp, to say the least. At least I can still look forward to going back there in a few weeks and working on my slice with Juan Nunez. Yay!  I started reading  The Joy Luck Club  by Amy Tan this morning. I like it a lot. Chinese stuff interests me. And now that I think about the Chinese culture, I always remember the bracelet I nearly always wear that\'s on my arm right now. It\'s a jade bracelet, with jade that\'s truly from China. It used to be very green, but it\'s just kind of white now, with a faint, green tinge to it. It\'s got a really neat way of adjusting, but it would be kind of hard for me to describe it. The string is like..no other string in the US, really. You can only get it from China. My Mom\'s already tried looking for it..many times. The red string stands for Luck, and the green jade has a kind of interesting..background. Supposedly, it\'s a way to protect people. Well, rather, if you wear jade, and it breaks..then you \"know\" that the jade saved you from a situation that might have killed you. It\'s hard for me to explain. I can be superstitious at times, which absolutely no sense..but..whenever I\'m going into a completely new place and environment (ex: Journalism camp), it\'s just comforting to wear it. It has also helped me through many tests XD   Something to talk about...something..to..type and ramble on about, rather. Ummm. Ok. I give up. It\'s almost time to go, anyway.  I\'ll try to put in some big posts when I get home. ^^;  '),(20435,3499599,'15,June,2004','2004-06-15 00:00:00','   I\'m at UNC! Whoo hoo! But I\'m not posting anything really important on my blog because everyone can see this, like Justin who\'s watching over my shoulder. And I\'m sure Alexander will join him soon at looking at the screen so..forget that.  Journalism Camp is fun. Uh...coming here for literary magazine was the best idea I\'ve had in a while, I suppose. And Justin just said it wasn\'t my idea, and now, like I knew he would, Alexander is looking at the screen. Ok I need to end this.  '),(20436,3499599,'13,June,2004','2004-06-13 00:00:00','   Yay. My brain is working now. I\'ve gotten over the whole..longing-for-camp mood thing now. I\'ve now moved onto an apprehensive mood about going to journalism camp tomorrow, but mainly because I have no idea what to expect.  I\'ve been spoiled by Brenda\'s. I mean, last week there were only nine campers there. We had five counselors. And then one camper left, so..yeah. We got to do practically anything we wanted. It was  so  nice. And now that I\'m going to UNC for this camp...I just dunno. I don\'t know how many kids will be there..I have no idea what KIND of people will be there...I just..have absolutely no idea. So. It scares me a little. But I\'m still riding on my energy-high from tennis camp so I think it\'ll carry into tomorrow and keep me happy enough to be able to deal with some new people and a new environment.  Today has been really boring, though. I went and got some stuff for \"camp\"..like a notebook, a pack of floppy disks, a new clock..and some food, of course. Something about eating healthy at Brenda\'s camp and exercising so much..it just got me really guilty yesterday, after eating fastfood and getting soda and chocolate. So, I think I\'ll just going to try to keep myself from eating that stuff..as much as possible. I even had a salad today for lunch ^_^ And it was my own little masterpiece..it wasn\'t that bad, either. Some lettuce and spinach..(rabbit-food-ish stuff..enh)..mushrooms..chicken..feta cheese...Greek salad dressing..and, voila. It was all healthy. Whoo. I had some bread with my salad, too..and some muenster cheese ^_^ I love that stuff.   Ooh. I started making a quizilla quiz today. I got very aggravated when the editor kept freezing, though, so I haven\'t finished it. But..when I do finish it, I\'ll put up a link. I\'ve got all the questions and answers typed in, though. I just need to find some pics.   Hm. Hopefully I\'ll be able to update my blog at UNC, but..like I said..I have no idea what to expect from this camp thingie. I guess I\'ll find out everything tomorrow. Maybe I\'ll make some more friends. More likely, I\'ll make some more enemies. heheh. Oh well XD  Ok I\'m gonna go try and finish my quiz now. Oh yeah..I was looking up some more summer camps because I kind of want to find one on webdesigning and eventually I found  urlLink this  site. There are programs at UNC so..hm. I\'m probably going to order a camp catalog thing. It looks interesting. I would have one hell of a time deciding between the Web Design & Graphic Arts program and the Video Game Creation one >  Song for today: Without Me by Eminem (Someone sang this on Kareoke Night at camp..normally I hate Eminem\'s music. And this song just popped in my head yesterday when my Mom was driving me home from camp yesterday...heh. Guess who\'s back in Mount Airy. w007!)  '),(20437,3499599,'12,June,2004','2004-06-12 00:00:00','   Camp..was great. As always. And now I\'m incredibly homesick for camp..which is really aggravating. As soon as I finished checking somethings online here at home, I said to my Mom, \"Ok..I\'m ready to go back to camp now.\"  I\'m very happy because..camp was a lot of fun. And I made some new friends! Yay for me! But..I\'m just so sad because I love that place. It\'s really hard for me to explain. I always love going there and it just makes me so happy and..bleh..I\'m rambling on.   So. A few things that happened..  -Went and saw The Chronicles of Riddick (good movie, but made little sense since we missed the first fifteen minutes)  -Made friends with my counselor after first asking her if she got her shirt from Roots (It was a shirt with just the word \"Canada\" on it..kind of like the one I have)..and from there, we just had a lot of interesting conversations. Lets see..I found out that her parents are from Romania, which was great, because I got her to teach me a few new phrases..and told her some that I know that I\'ve learned from my own family. And then Brenda (Brenda Schultz McCarthy..it\'s her camp =p) told us how Karine (that\'s her name) dated Andy Roddick before he made it big which..that was really interesting.   -Made friends with a guy named C.J. from DC who..just reminded me of a lot of people. Part of him reminded me of my stand partner in symphony, in personality and the same age thing. Part of him reminded me of my best friend, Aaron (which was a good thing)..but then  another  part of him reminded me of my neighbor, Cole, which is not necessarily a good thing. He was fun to talk to, though. I needed someone like that. Otherwise I would\'ve gone insane from listening to the girls in my cabin talking about what they were going to wear each day and..oy. And he had the same PS2 games as me ^_^ Lotsa common interests, which was fun. Yay.   -I got a lot more..tan. yeah. Whoopee.  -My tan lines are now a whole lot worse. yeah. Whoopee.  -Had a lot of fun ^_^  Can\'t think of anything else to write about. I just..miss camp now. Maybe going to journalism camp on Monday..or whenever I go..that\'ll hopefully get my mind off of tennis camp. hm.   Song for today: I Won\'t Spend Another Night Alone by The Ataris (I love The Ataris..and the band No Use For a Name..they\'re my favorites XD)  '),(20438,3499599,'06,June,2004','2004-06-06 00:00:00','   Ok. Major headache. Will try my best so this post makes sense. hn.   Went and saw some movies last night. heheh. At the Drive-In! Whoo! One of the few left =p The first movie was \"13 Going On 30\" at 9..which my mom thought it was starting at 7..we still went, though. That movie was okay..it was cute..and all sweet \'n stuff..with just all..cute-humor and stuff..ok can\'t really explain it.   But the movie I really wanted to see was \"Mean Girls\"..ok still dunno if quotations are right for movies or italicization so just..enh. Mean Girls was really good ^_^ Really funny. The best part was at the beginning, when whats-her-name described homeschooled kids and there were those kids sitting on a tractor, explaining parts of religion and..yeah..ok that was just the best. There were a lot of really good jokes. I laughed...a lot =p  So.  Like I said..some other time..I go to tennis camp..well, today. I leave around 2:30..and..will be back home sometime this coming Saturday. I think. And then I go to journalism camp. And..yeah. Ok. Busy week. Er, busy week(s).   I\'m looking forward to tennis camp ^_^ I just got new Rook cards since I lost my old deck..and a new deck of normal cards..one of the best things about camp is that everyone knows a lot of card games..well, normally. And we just play cards..a lot. Yaaaay.  Well, we all need something like that to just..have fun and play around. Considering we play about four/five hours of tennis every day..enh x.x;   It looks like it\'s supposed to rain a lot this week in Woolwine...heheh..Woolwine, Virginia is where the camp is at. If it rains a lot, that means we\'ll probably play on the indoor court or go see a movie or go bowling or go to the skating rink or..well..we have a lot of options. The great city of Martinsville is close by the camp, so we have plenty of stuff to do =p  Ok..since  urlLink Jeremy  is a \"cute but psycho\" happy bunny and  urlLink Jacob  is a \"kiss my ass\" happy bunny..I had to find out what I was..so..I am..     you are the \"I hate you so bad\" happy bunny. You hate everyone and eveything and your not ashamed of it.     urlLink   which happy bunny are you?     brought to you by  urlLink Quizilla    Whoo hoo.  Ok. I really need to finish packing now >  Arrivederci...AdiÃ³s...Vaarwel...Au revoir...Auf Wiedersehen...etc. ^_^  '),(20439,3499599,'03,June,2004','2004-06-03 00:00:00','   My inner child is ten years old!    The adult world is pretty irrelevant to me. Whether I\'m off on my bicycle (or pony) exploring, lost in a good book, or giggling with my best friend, I live in a world apart, one full of adventure and wonder and other stuff adults don\'t understand.    urlLink   How Old is Your Inner Child?     brought to you by  urlLink Quizilla      Barefoot- free, rebellious, and wild, you hate boundries and rules.  You tend to be on the crazy side and often sweep people up along with you.  You are most likely the leader of your group of friends.    urlLink   What Kind of Shoe Are You?     brought to you by  urlLink Quizilla      Postatem obscuri lateris nescitis. \"You do not know the power of the Dark Side.\"  There are two possibilities: you are a Star Wars geek, or you are unreasoningly scary.     urlLink   Which Weird Latin Phrase Are You?      brought to you by  urlLink Quizilla    Ooh and..  Song for today: Spinning Wheel by Blood, Sweat, and Tears (old song!)  '),(20440,3499599,'02,June,2004','2004-06-02 00:00:00','   Normally I hate shopping. But..today was fun. Mostly because all I got was books. heheh. I loves me some books ^_^  Most of them were for summer reading. Here\'s what I got:  Much Ado About Nothing by Shakespeare All Quiet on the Western Front by Remarque The Joy Luck Club by Tan The Iliad by Homer Les Miserables (with the little accent over the e) by Hugo The Arabian Nights  I also got Alice\'s Adventures in Wonderland and Through the Looking-Glass by Lewis Carroll, since I\'ve been wanting that book for a while. I love Carroll\'s works. And he has my birthday, too =p Oh yeah..and I got The Inferno by Dante Alighieri because I\'ve been wanting that for a while, also. The only other book that I really want now is On the Origin of Species by Means of Natural Selection by Darwin..heh. I\'m going to wait \'till I can get a really good copy in hardback, though.  So. A summary of these books..  I\'ve already read Much Ado About Nothing..it was really good. I like Shakespeare\'s works a lot, fortunately. My favorite poem (for the time being) is \"The Seven Ages of Man\", which he wrote. He always seems to have one character that is very witty. I like how Beatrice is the one that\'s endowed with much wit in this story. It\'s always good to have a strong female in a story. Especially one that\'s very good at making fun of people and mocking people that are in love, and love itself. Whoo!  hm..I didn\'t want All Quiet on the Western Front, but it was required. And I didn\'t want The Joy Luck Club..but my mom insisted I get that. I already know the story in The Iliad by heart..but I still needed to get an actual copy of the poem.   I\'ve already seen the movie Les Miserables..it was interesting, I suppose. But..like The Iliad..it\'s something that I needed just so I can re-read it at times, and to have on hand.   Lastly..I just knew I had to get Arabian Nights because I love folk stories and fairy tales and..yeah. It\'s gonna be fun reading that. I\'m thinking it\'ll accompany me to tennis camp next week. That way, I won\'t be troubled with stopping points, since it\'s just made up of many short stories.   Ohh yeah. I\'m going to  urlLink tennis camp  next week.  It\'s my home-away-from-home. I went there the first year that it opened..the first week, actually. And I went there for two weeks last year. I\'m going for two weeks again, this year. I leave on Sunday (June 6th), and I\'ll get back on Saturday (June 12th). I then leave on Monday for Journalism Camp at UNC Chapel Hill..and..I have no idea when I get back from that. I think it\'s for three days, but I\'m not sure. I\'m going there for the literary magazine info, not for journalism, though. Soo. I have no idea what that\'s going to be like.  Ok. This post is getting long. I need to finish my comic-listing, however..  6.  urlLink Applegeeks  7.  urlLink Queen of Wands  8.  urlLink Megatokyo  9.  urlLink Little Gamers  10.  urlLink Mac Hall   Two honorable mentions..:   urlLink What I Learned Today  This comic just isn\'t old enough to know if it\'s really a keeper, I guess. There are some hilarious ones, though.   urlLink InkTank  I just found out that this comic was up and running again. There are many many funny comics in InkTank\'s archives. I just haven\'t gotten the chance to look over the latest ones.  Ok. Time to actually..do stuff. Like clean up. \"God forbid company comes over and sees our house in a mess\"...company meaning Aaron, who could care less if our house was messy...and..the quote being a..mocking of how my Mother goes ballistic the day before people come over, because the house isn\'t *clean* enough. Oy.  Oh yes and..I have no idea whether to italicize the names of books or to underline them or to put them in quotation marks so..I\'m just going to leave them plain like they are.  '),(20441,3499599,'02,June,2004','2004-06-02 00:00:00','   You\'re a Dialogue/Character Writer!    urlLink   What kind of writer are you?     brought to you by  urlLink Quizilla       You\'re Skittles!!!  You have a very interesting personality, you\'re so unique.  You\'re the kind of person who always thinks outside of the box. You\'re also a very accepting individual, and believe in inner beauty.    urlLink   Which kind of candy are you?     brought to you by  urlLink Quizilla       You\'re Gangsta Bitch Barbie.  You\'re tough and you like it rough, and of course you like to pop a cap in any wiggers ass.    urlLink   If You Were A Barbie, Which Messed Up Version Would You Be?     brought to you by  urlLink Quizilla  '),(20442,3499599,'01,June,2004','2004-06-01 00:00:00','   Ok. My Mom got me into making dogbones a few years ago to get extra money. I have a booth at our local emporium and...well, it seems like I\'ve been running low lately. So today I had to make some and I was going crazy. Instead of getting the kind of rye flour we normally have, my mother got some kind of natural...stone ground..whatever..rye flour. And every single freakin\' time I tried to use the cookie-cutter to get a dogbone out of the mess...it would crumble apart.   So. I spent an hour..an  hour ..trying to make dogbones. Eventually, my mom went downstairs so I just got the dough together and threw it out in the trashcan. ^^;  I was so ticked off, though. When I was yelling at the dogbones and my mom finally came to see what the matter was, she automatically asked if I measured everything right and..oy. That just made me so angry. Of course I measured everything right. I\'m not like my sister, who just guesses at all the measurements and does everything wrong when it comes to cooking. I try to be as precise as possible. I\'m a perfectionist, after all.  Enough of that.  For some good news..my other sister is coming into town! Yay! Aaron\'s coming by on Friday, hopefully, and we can just talk and play games and do stupid stuff like sing along to that yellow submarine song by the Beatles. heh. Or maybe I could get him to let me on his AIM name so I can confuse his friends. Ah..good times...good times....  For some bad news...I\'m going to try and pry myself away from RuneScape. Ok, that\'s not really bad news. But it\'s going to be hard for me. I finally got the picture after  somebody  sent me this  urlLink guilt-inducing slashdot article.   I think it\'s kind of funny, how this relationship is..I\'m the gamer..and Jacob just isn\'t really a gamer. But...I\'ve been thinking about it..it\'s a good thing that I\'m not dating a gamer. I mean, I already have a hard enough time getting myself away from games. If I had a boyfriend who was a gamer..well..he\'d just encourage the gaming and wouldn\'t help the addiction-to-games situation at all, I guess.   Oh well. I just know that I need to get away from RuneScape through my own strength of will. Jacob refuses to yell at me when I play RuneScape, unfortunately. I knew if I got him to \"yell\" at me, it would distract me from the game and I would be able to realize that I should talk to him instead of waste my life to a stupid, pointless game...  I\'m also getting quite tired of being asking if I am actually a female on RuneScape. \"omg asl?//\"...the phrase is getting a little aggravating by now. Of course, there was the time that one person kept asking me to be their girlfriend and when I said no, because I already have a boyfriend..\"o its ok, tho, u can have 2 bfs. i wont tell any1. it can b a secret\" That made me laugh a lot.  Ah, and if you didn\'t know, I have a boyfriend. His name is Jacob. He\'s an oxymoron. ^_^  Oooh. Lunch. Ok. Food > blog. Yeah. Math. I hate math. But..that equation is appropriate...  Song for today: Little Drop of Poison by Tom Waits  '),(20443,3499599,'01,June,2004','2004-06-01 00:00:00','   A GAME-BOY. You\'re like a tomboy without the love of sports. Reality sucks, but as long as you have your electronics you feel you can cope. Time goes unnoticed when youre locked in your room hooked up to your Nintendo, rocking to your favourite collection of guitar-driven albums.  Your virtues: Intelligence, sense-of-humour, individuality.  Your flaws: Inability to cope with real life, action-freak spirit, reclusive nature.  Your Personality type is the only type that would like this cool online gothic Game:   www.life-blood.vze.com    urlLink   What kind of girl are you?     brought to you by  urlLink Quizilla      w007.    -Perfect- You\'re the perfect girlfriend. Which means you\'re rare or that you cheated :P You\'re the kind of chick that can hang out with your boyfriend\'s friends and be silly. You don\'t care about presents or about going to fancy places. Hell, just hang out. You\'re just happy being around your boyfriend.    urlLink   What Kind of Girlfriend Are You?     brought to you by  urlLink Quizilla    So true ^_^    You\'re an Etch-a-Sketch!!  You\'re the creative, artsy type who doesn\'t need to actually utilize a single muscle group in order to have fun. Doesn\'t matter though, you\'re still cool.    urlLink   What childhood toy from the 80s are you?      brought to you by  urlLink Quizilla  '),(20444,3499599,'01,June,2004','2004-06-01 00:00:00','   I can\'t get the song \"Funky Town\" out of my head. It\'s very hard to have that song in your head, considering there..aren\'t many words..and...ennnnh. Yeah.   I\'m a little confused right now. Mainly about tennis and my geometry skills.    First..tennis.   I love playing tennis. It\'s a very complex, strategic sport. It\'s a sport I can actually connect with, really. You have to be a quick thinker, and also be fit and in shape. heh. I need to work on the being-in-shape part but...continuing.   I\'m just angry because I can never seem to be completely serious. It seems like I enjoy the game too much. I laugh at my mistakes...I laugh when people out-think my moves and use my own tricks against me. I comment on good shots...and...well, I just like to have fun when I\'m playing. I know that I should try and concentrate for when I play in tournaments but...most of the time I just don\'t think it\'s worth it, to be out for blood for a silly little game. Especially all these fanatic sports parents that push their kids to the edge and enter them in as many tournaments as humanly possible and make their children focus their lives on one sport...bah.   And..about geometry...hm. Apparently, I made a high score in a class I absolutely hate. It\'s not that I just hate the subject. Ooh no. I rather disliked the teacher, also.   So. It seems that I made a high score on the EOC...at one point, when I looked over my answer sheet..well, there are about six big columns with twenty questions in each or so. When I looked over it, I had one column almost completely blank because those were all questions I had skipped and were going to go back over. I panicked...and...gah. I hate math.  It\'s all..yes or no, wrong or right. You\'re just doing problems that have already been proven by so many other people over and over. What\'s the point of it? I just don\'t understand why they make us do math. And when am I going to use these skills in the \"real world\"..? I mean, by the time I ever come across a situation where I need to know how to calculate the length of an arc of a circle..you think I\'m going to remember the formula?! Give me a break.  Ok. So. I feel like listing stuff. Oooh. How about my favorite webcomics. Sure! That works.  I\'ll list top ten favorites, first being..my favorite, obviously.  1.  urlLink Something Positive     I\'ve just always loved this comic a lot. Maybe it\'s the sarcasm. Or the cynical-ish-ness. I dunno. It just always makes me laugh. ^_^  2.  urlLink Ctrl+Alt+Del     There\'s no need for me to write a reason why I like this comic. I mean..c\'mon. It\'s CAD. Everyone loves this comic. Well, every gamer. Well..ok..I dunno. Just..a lot of people like it.  3.  urlLink Sinfest     Again..it\'s Sinfest. It makes fun of religion, love, cats and dogs..random stuff like that. It\'s good stuff.  4.  urlLink Real Life     This is just simple humor. Which...everyone needs just a bit of simple humor every now and then.   5.  urlLink Penny Arcade     The legendary Penny Arcade. One of the best. When a comic is funny, it\'s lmfao-funny. When it\'s not funny, you just kind of sit there and scratch your head. You have to know a bit about games to understand most of the inside jokes, unfortunately. But..they always seem to end their last panel with something completely irrelevant that makes everyone laugh. Well, normally. enh. Tycho\'s the best XD  Ok I don\'t feeling like continuing this now. I\'ll do it later. Maybe. Procrastinating is always nice. ^_^  Song for today: Funky Town by Lipps Inc.  '),(20445,3499599,'25,July,2004','2004-07-25 00:00:00','   Gone to tennis camp. ^_^ Will be back on the 31st.  Here are some things for you to do till I get back..  Go read  urlLink a blog  that is never updated.  Go annoy  urlLink Jeremy  since I won\'t be here to do so.  Go see my  urlLink artwork  and add all my  urlLink pictures  as favorites.  Go watch  urlLink Foamy!  ^_^  Read  urlLink bash.   Go see the  urlLink Cruel Site of the Day.   Go look at  urlLink Something Awful.   Ok if that\'s not enough for you..well, pft. You\'re on yer own. =p  Song: Concrete Girl by Switchfoot, and The Beautiful Let Down by Switchfoot  Ohh..P.S. SEND ME AN E-MAIL! :D I forgot the mailto html thing so just..bleh. Send something to oOMeggieOo@hotmail.com  '),(20446,3499599,'21,July,2004','2004-07-21 00:00:00','   I have..nothing to do.  My Mother is going insane. Well, kind of. She\'s just going a little summer-crazy like I am. She made me a tinfoil hat. ^^;; She made one for Rose, too. Rose..did not like her tinfoil hat.  I had told her how I had tried to call Aaron yesterday at one point and intercepted someone\'s call. She then ranted for a while about all the stray microchips in the atmosphere that are out to get us and declared that, once we got home (we were in the car), that she was going to make everyone tinfoil hats. It\'s..very interesting to observe my mom when she gets into these moods.  hm. I played tennis this morning. I won my games..blah blah blah. Nothing to be very proud of. I wasn\'t playing against people close to my own level of skill \'n whatnot. It was fun, though, beating the annoying \'lil boys.   \"You can\'t hit hard! You\'re just a girl!\" I serve. \"Whoa. DID YOU SEE THAT?! That was so cool! She went like *whoosh* and like wow.\"  I couldn\'t stop laughing whenever I played against them. it was hilarious ^^; silly boys.  Ok I\'m gonna go..write..stuff. Yeah. enh. need something to do.  Quiz stuff..   My #1 result for the SelectSmart.com selector,  urlLink  Your ideal crime  , is  Kidnapping    My #1 result for the SelectSmart.com selector,  urlLink  World Domination Selector  , is  Hypnotise the populace    My #1 result for the SelectSmart.com selector,  urlLink  Who Were You In A Past Life  , is  You were Cleopatra!    My #1 result for the SelectSmart.com selector,  urlLink  Selecting the Right Programming Language for You  , is  CGI    Song: Punk Rock 101 by Bowling for Soup  '),(20447,3499599,'20,July,2004','2004-07-20 00:00:00','   Whoo.  &nbsp;  So I had a tennis tournament yesterday, with two matches in one day.  &nbsp;  First match went by easily 6-3, 6-4. It was a very slow match, neither of us really exerting ourselves and just playing quite..leisurely, I guess. It was getting rather hot, anyway, which makes the best of people just kind of slow down and forget scores.  &nbsp;  But then, I had to play the #1 seed.  &nbsp;  Which..didn\'t really bother me. I\'ve seen her in a lot of matches. And the last time we were at this tournament, she got out in the first round, then in the consolation round, too. I knew she\'d outpower me, but I figured I\'d at least try and give her a run.  &nbsp;  And I did. I did quite well, actually. My serve was going insane, but that\'s because my coach had told me the day before that I\'d been doing it all wrong and I hadn\'t remembered I had a tourney, so..I didn\'t really practice fixing it. Oops. I managed, though. The score of the first set, 6-2, doesn\'t say how we went to deuce every single time.  &nbsp;  Anyway. It was easily around 80-85 degrees outside and it was just getting..really hot, to play tennis. We had to play on a clay court that hadn\'t been watered in a while, so dust was rising everywhere, especially after long points when we both had to run down the ball. Ooh I did really good, though, because a lot of the time she\'d hit a very good angle but I would go diving after it, and she\'d be caught off balance because she figured I wouldn\'t get it and..yeah. Anyway. I was running a lot.  &nbsp;  So during the beginning of the second set, she was up 2-1 and I was..just..losing it. Mentally. Well, I hadn\'t realized it yet, but there was one point where I just dropped my racquet after hitting a great serve and stared at it thinking \"Oh hey, I dropped my racquet. That\'s pretty funny.\" and I hadn\'t really processed that yet when she hit it back...If I had been an onlooker, I would\'ve laughed at myself so much. I just stood there kind of stupefied.  &nbsp;  That\'s when my Mom yelled at me to get off the court and sit in the shade for a while but I yelled back No and that she wasn\'t even supposed to be saying anything. My opponent was pretty confused, not sure whether to laugh and keep a straight face. I don\'t remember much of the tennis after that. I just knew I was pretty dizzy, I was kind of talking to myself, and playing..bizarre-ish-ly.  &nbsp;  Half of the time when I was serving, I had to toss the ball straight up towards the sun and I couldn\'t see. So then I had those little traces on my eyes..y\'know..and I kept trying to blink so I could see. Which..didn\'t help much. And at that point, we were going to a 5th or 6th deuce on nearly every friggin\' game so I was just desperately trying to get it over with. Which..oh man, it was awful. I would just toss the ball up and hope I hit it, and I was hitting some great serves. So I\'d, say, get the first point, but then she\'d hit a winner when it was add-in for me. And then I\'d double fault, but then when it was add-out for me, she\'d hit the return straight in the net. \'Twas horrible.  &nbsp;  Aaanyway. Afterwards my Mom got mad at me and said I was having some sort of..heat stroke thing..which I wasn\'t listening, of course. I just had to lie down because I felt like I was gonna be sick. oy.  &nbsp;  Somehow in the second set, I got another game, even though I was going crazy. lmao. I started laughing during some points. And then at the ends of others, when I hit a bad shot or something, I\'d just start making up words and just..rant on stupid stuff under&nbsp;my breath. I don\'t think the person I was playing against heard me, sooo..um..that\'s good.  &nbsp;  Ok I\'m gonna go do stuff now. Maybe go crazy again. I dunno.  &nbsp;  Song: Lemon by Katy Rose  (I like her song Overdrive too) ^_^  &nbsp;  Ooh..and a quiz thing.    urlLink     urlLink  Which Super Smash Brothers character are you?  '),(20448,3499599,'18,July,2004','2004-07-18 00:00:00','   Been feelin\' artsy lately.. Been addicted to the  urlLink Deviant Art  site.. &nbsp; Go see my  urlLink stuff!   NOW.     &nbsp; ^_^ &nbsp; I had to go get new batteries for my digital camera today, since I felt like taking pictures then realized it was dead.  &nbsp; Wal-mart on a Sunday? &nbsp; Biiiig mistake. &nbsp; Everyone was dressed up in..their..church-clothes finery. it was scary. &nbsp; I had to constantly take the scenic routes through different departments because I kept seeing people I knew and  did not  want to talk to. o.-; enh. &nbsp; Ok I\'m gonna go take pictures of stupid stuff and call it art. &nbsp; Song: Meant to Live by Switchfoot have totally fallen in love with this song..neh.  '),(20449,3499599,'16,July,2004','2004-07-16 00:00:00','   I re-took the inferno test...  &nbsp;  &nbsp;   The Dante\'s Inferno Test has banished you to  the Ninth Level of Hell - Cocytus!    Here is how you matched up against all the levels:       Level    Score      Purgatory  (Repenting Believers)   Very Low      Level 1 - Limbo  (Virtuous Non-Believers)   Very Low      Level 2  (Lustful)   Moderate      Level 3  (Gluttonous)   Very High      Level 4  (Prodigal and Avaricious)   Very High      Level 5  (Wrathful and Gloomy)   Extreme      Level 6 - The City of Dis  (Heretics)   Very High      Level 7  (Violent)   High      Level 8 - the Malebolge  (Fraudulent, Malicious, Panderers)   Very High      Level 9 - Cocytus  (Treacherous)   Extreme       Take the  urlLink Dante\'s Inferno Hell Test    &nbsp;  ~*~  &nbsp;  I\'m so evil ^^;  '),(20450,3499599,'13,July,2004','2004-07-13 00:00:00','   Sooo. Yeah. Stuff. Been..doin\'..stuff. &nbsp; Ooh. The updater thing is all..different. This is funky. Yay. More options! Options are always a plus. &nbsp; I\'ve been having one of those \"leave-me-alone-I\'m-writing\" moods, lately. Yesterday I went and sat outside and just scribbled down some stuff..and today I went out for about an hour, and just got back in. I don\'t even know what I wrote down. I bet it was interesting, though. mhm. &nbsp; Lets see... &nbsp; I\'ve been playin\' some tennis lately, yup yup. Yesterday we had \"team\" practice and I was shocked when the *seniors* came out and graced the courts with their presence. Of course, they talked the entire time, then left without picking up any balls. But I had a nice time mingling with the freshman and turning them against the older players. Yay for me! &nbsp; I was really-super-extra-nice, too. I went and got a basket of tennis balls out for them, since I\'m only of the special people that knows the combination to the tennis hut..thing..whoo. At first, there was just a senior there and two freshman hitting against each other so I had to hit with the senior...which...I was just kind of thinking \"Ohhh..*unappropriate blog word here*\" But then all the seniors arrived and some other freshmen, so I saw the opportunity to leave the seniors by themselves and..it was just&nbsp;a good thing. Because then I actually got to practice! Yay! We played Canadian doubles, me versus the number 1 of the middle-school team last year and another girl..and they made me run. It was a very good work out for me, and I had to practice aiming the ball, alternating \'n whatnot so I didn\'t hit to one person the whole time.  &nbsp; I think the girl who was #1 on the middle school team last year will be in the top 6..I want to see her in the top 6, at least. She\'s good. She won\'t take over, but she\'s still good. I\'m hoping I\'ll be #1..but..I just need to practice more. I played horribly at \"open courts\" today with George (my coach) and..yeah, that was depressing. But it made me wake up and realize that if I don\'t get my ducks in a row, I\'m gonna get beat during challenges, and that would totally suck. So! More practice time for Megan, now. Lately I\'ve been going out and serving from 7-8 anyway. I\'ll just have to do that more,  along  with practices with the \"team\" (pft) and George.  &nbsp; Moving on.. &nbsp; I went and shadowed at the newspaper place yesterday. It was..kind of interesting, kind of boring. It would have been absolutely horrible if I hadn\'t shadowed the guy that I ended up shadowing. He knew what it was like to just watch a person do ads and that kind of stuff, so he tried to make it as interactive as possible, I think. I ended up doing one of those \"Happy Birthday\" announcements and..I won\'t be able to forget it. It was to someone named Gatlin..and..the person had listed the entire family in the announcement, which was bizarre..and come to find out, this Gatlin girl has a grandparent named Snook..and..I just lost it there. I couldn\'t stop laughing for a while..I mean, Snook. SNOOK! &nbsp; Anyway. I don\'t think I want to design ads for the rest of my life. It looks pretty boring. No way to really be creative, at least. But I\'ll keep an open mind. I mean, maybe my thing could be the magazine ads..or commercials. oh i love commercials. enh..It just seems like doing newspaper ads isn\'t that..thrilling.  &nbsp; In an hour or so I\'m going back to shadow the same guy doing page layouts. I really don\'t know if I\'m going to like that either, but..like I said..gotta keep an open mind. And this might be a nice opportunity to get a job next summer at the newspaper office. Who knows. &nbsp; I really think the literary magazine is gonna be my main project next year, and I\'m really excited about it. Honestly. I\'m not just saying it like OMG I CANT WAIT 2 DO TEH LITARARY ZINE! I mean...I think it\'s gonna be fun. I\'ll get to do page layouts in creative ways...pick what stories, poetry, and artwork goes in and what doesn\'t (bwahahaha)...be an  editor ...yay! &nbsp; To sum it all up..summer is getting boring. I\'m ready to be back in school. Doing stuff.  &nbsp; Song: Let\'s Get It Started by the Black Eyed Peas  '),(20451,3499599,'13,July,2004','2004-07-13 00:00:00','   urlLink JEREMY  IS A MEANIE >  Oh, it seems that I am now sneaky and conniving. Yay for me.   Song: Bug Bytes by Alien Ant Farm  '),(20452,3499599,'12,July,2004','2004-07-12 00:00:00','   urlLink       urlLink        A Tarantula is the pet for me!  Eight legs are better than two, right?  And what a wonderful pet for the alternative individual like myself.   My #1 result for the SelectSmart.com selector,  urlLink  What\'s your Hawaiian name?  , is  Kolohe - Mischievous, Naughty    My #1 result for the SelectSmart.com selector,  urlLink  Alice in Wonderland Selector  , is  Cheshire Cat    My #1 result for the SelectSmart.com selector,  urlLink  The Sorting Hat  , is  Ravenclaw          urlLink  So which letter of the alphabet matches YOUR personality, huh?  '),(20453,3499599,'07,July,2004','2004-07-07 00:00:00','   So..yeah. I\'m really really bored.  We went to Raleigh a few days ago...that was one of the worst little trips I\'ve had to suffer through. First, we had to wait two hours whilst my Mom and some tire people tried to figure out what to do with the flat tire and if they should put on the spare tire or put on another tire and where to put it and..oy. So, that took two hours. Then we actually got on the road! And of course, it\'s about three hours or something from here to Raleigh..which was delayed..a lot..due to much road construction and a lot of traffic near Durham. So, the trip took about four hours. We then had to wait for an hour to get food from P.F. Changs..mm..it was worth, it though. And I got pants from the mall. Yaay. They\'re nice.  Ooh..and I got some really good cake from the Cheesecake Factory..yeaah. that was good. Of course, then I had to listen to my grandparents talk about their trip in Saskatchewan, Canada..eh.  On the way back, though, we stopped by Winston and I traded in my games and got SnowBoard Kids..yaaaay! ^_^ It was really really weird, though...once I got in the car, I called Aaron to tell him I had found the game and wouldn\'t have to get it off of Ebay, but then while his phone was ringing, I checked the back of it and it says \"Aaron\" in permanent marker. Ok..coincidences like that freak me out. We had a nice time laughing about it, though.  Lets see..  We have kittens (again) and so far we\'ve only been calling them by their colors..like, \"Hey look at the grey kitten\" or \"MOMMM...DAISY\'S TRYING TO EAT THE WHITE-PAWED GREY KITTEN\"..etc. My sister and I gave \'em names yesterday...so...we decided to kind-of-sort-of going with artists. The black and white one is Escher..my favorite, the grey one with white paws, is Toulouse Lautrec...and we couldn\'t think of a good female artist name for the grey kitten that\'s a girl, so it\'s Phoebe. heheh.   Ah..today I went by the newspaper office to ask about possibly shadowing someone and seeing how our local newspaper works. whee -_-; I had a nice time talking to the editor, though, and I\'m going to go in on Thursday and Friday..one day to look at how they do page layouts, and one day to look at the ad designing. That\'s the stuff I like. I love advertising techniques. I can\'t help but watch commercials, captivated by trying to figure out how they try to get people to watch them. And I like analyzing ads in magazines \'n stuff to see how they\'re trying to sell their product. I think it\'s interesting. And I\'m going to need to know the eye-catching design work to be able to pull of the literary magazine this year...so..it\'ll be good practice, if only that.  And..nothing else has been going on. Well, I\'ve had a lot of fun making confetti out of all the republican mail we\'ve been getting, but besides that..enh.   so much boredom.  Song: Hey Little Girl by Unwritten Law  '),(20454,3499599,'07,July,2004','2004-07-07 00:00:00','   Someone  pleasey please please  buy me  urlLink this.    heheh. Or..LOL.. urlLink this.    Time to paste quiz results..these don\'t have pics, unfortunately..but there are a lot of these types of quizzes and they\'re all really interesting...    My #1 result for the SelectSmart.com selector,  urlLink  Your Spirit Animal  , is  Cat     My #1 result for the SelectSmart.com selector,  urlLink  Gods and Goddesses of Greek Mythology  , is  Athena    My #1 result for the SelectSmart.com selector,  urlLink  Anglo-Saxon or Germanic God Selector  , is  Loki    My #1 result for the SelectSmart.com selector,  urlLink  What myth beast are you?  , is  Dragon    My #1 result for the SelectSmart.com selector,  urlLink  Sandwich/Personality Selector  , is  Toasted Ham and Cheese  '),(20455,3499599,'02,July,2004','2004-07-02 00:00:00','   I have fallen in love with  urlLink Foamy   YOU HAVE TO GO SEE THAT SITE.  ^_^  Anyway.   That site is hilarious. If you ever saw Invader Zim, Foamy is like Gir..just as spastic..just as high-pitched...only with a very..dirty mouth, and twisted sense of humor ^_^ How could I not love it?  Sooo. Stuff that\'s been going on lately..  Aaron came over yesterday for the day..\'twas fun. It\'s always fun when my seperated-from-birth-sibling comes over. Yay! It\'s also lossa fun talking to people his buddy list that I don\'t know. Finding out afterwards that I actually know some of the people..well..so far, it\'s only happened once. e.e; I didn\'t technically know the person..but..when checking our symphony\'s end-of-the-year-picture, we were right beside each other and..after finding that out, we still had no clue who the other was..and..that was just a really freaky coincidence. Moving on.  Oh. I\'ve gotten off my AoM rush. After having a conversation like this at the end of a game, y\'can kind of understand why..: My opponent: gg Me: (I had just forefeited) That was not a good game. It sucked. You rushed me before we were even five minutes into the game, and I had to make do with the few resources I had for an hour. My opponent: lol Me: (After games, I always seem to get hit on..so..) Don\'t you dare hit on me, either. I\'m pissed off right now. My opponent: o, so u r a girl? how old r u? Me: AGGGGHHH  So..too much testosterone + too much stupidity = overwhelming of the Megan to the point where she can\'t enjoy her game of smiting people anymore.    Bleh. I don\'t like stupid people.  Let\'s see..  I\'ve been pretty bored. Nothing\'s going on this week. My Mom\'s gone to meetings a lot, but we\'re probably going to go to Raleigh on Friday (which means P.F. Chang\'s for me...YAAAAAAY..I   I have been downloading music like crazy today out of pure boredom. I go to Smartpunk.com and find some bands I\'ve never heard of that have interesting names..then download \'em. It\'s fun. heheh. I just have..very weird tastes in music.  enh. I feel like my summer is being wasted. I have a huge music theory book I should be going through, but I keep procrastinating. I should be practicing violin, but..again..procrastinating. I should be playing more tennis...I shouldn\'t be playing so many games...I should be working on ideas for the literary magazine...I should be doing  something!  But no..I just sit at the computer and watch flash videos or read bash quotes >  I\'ve always had a very bad habit of forgetting journals and long-term projects. bleh.   Enough complaining. I\'m gonna go play some games and scare people on Yahoo.  Song for today: Puttin\' On The Ritz by...I have no idea. It\'s a really good song, but I found a version by the Taco and by They Might Be Giants and it\'s the same..so..somewhere, along the lines, this song got mislabeled..enh.  '),(20456,3499599,'02,July,2004','2004-07-02 00:00:00','   urlLink  Ooh.   heheh: One academic came across: \"please click here for more information\" in the body of a student essay. Other students have forgotten to delete the: \"thank you for buying your essay from ...\" at the bottom of their work   urlLink  HAHAHHAHAAAA   lmao..  Serena follows law of the jungle  Fri Jul 2, 9:42 AM ET    By Bill Barclay   LONDON (Reuters) - Serena Williams graduated to the Eric Cantona school of sporting philosophy at Wimbledon this week.   \"The lion is the king of the jungle, but the tiger is the king of the forest\" -- that was the reply the defending champion gave when quizzed about her prospects for Saturday\'s final against Russian upstart Maria Sharapova.   The befuddling comment recalled the famous retort of French former Manchester United soccer player Cantona to reporters: \"When the seagulls follow the trawler, it\'s because they think sardines will be thrown into the sea.\"   Nonetheless it did take Serena a couple of attempts to get her philosophy right.   \"I will prepare in a way that I will not tell you,\" she told reporters with a mischievous smile.   Asked why, she replied: \"Because the lion is the king of the forest but the tiger is the king of the jungle.\"   Realising her error, the 22-year-old American hurriedly corrected herself.   \"I had it mixed up the first time. The lion is the king of the jungle but the tiger is the king of the forest,\" she said.   Asked why she was not playing the doubles this year with sister Venus, who lost in the second round of the singles, Serena hared off on another tangent.   \"I just had a deja vu, that\'s weird,\" she said.   \"I dreamt that Venus lost at Wimbledon and somebody asked me why we didn\'t play doubles. Then I was playing this young girl in the final. Isn\'t that weird?   \"I always have deja vus.\"  '),(20457,3499599,'02,July,2004','2004-07-02 00:00:00','   Well, stupid IE, rather. I have to press refresh at least five times to get any page to load and it\'s just..getting on my nerves, for some odd reason -_-;  eh. I got my schedule today. I have a problem with it, though, so it\'ll most likely be changing..soon.  First semester: Civics and Economics Algebra 2 Computer Applications 1 Journalism 2  Second semester: Honors English 2 Chemistry Show Choir 2 AP Music Theory  The teacher they\'ve signed me up to take algebra 2 with is the same one that I had for geometry that I absolutely hated and went to the guidance counselor about. Sooo..I\'ll hopefully be changing that soon. And while most of my teachers are listed beside my classes, I apparently have no teacher for Chemistry yet. heh. Kind of amusing to just see that blank spot where a teacher\'s name should be.   One thing I don\'t understand is that I have the same homeroom teacher I had last year, but I\'m not taking her for any of my classes..uh. I mean, I had her last her for Honors English, but this year I don\'t have her as a teacher..yet she\'s listed on here as my homeroom teacher..maybe that was a mistake. I dunno.  Sooo.  I\'m thinking about exchanging some games to get a new game..or maybe even  games ...just depends how much I can get for the ones that I don\'t want. I have a WTA (Women\'s Tennis Association) game that absolutely sucks and I\'ve only played once...it\'s for ps2..lets see..I also have the Harvest Moon game for PS2 which was really bad..aaand I have the first Sims game for PS2, which was AWFUL o.x I was so incredibly disappointed by that game e.e  I looked through my N64 games to see if there were any I didn\'t want, but I like most of the ones I have so..will be keeping them. heheh. My N64 games vary..ranging from Diddy Kong Racing to Zelda 64 (which isn\'t even mine..I\'m...*borrowing* it)...WaveRider 64...Super Smash Brothers (eee ^_^)..and..others. One of my favorites was Quest 64, but I never did get it. I did beat it, though. It had a very sucky ending.  I would  love  to be able to get the new final fantasy game for the comp with the money from my turned in games. I think it\'s down to $50 now, at some places. Buuuut..that game would suck my life away e.e; It\'d be enjoyable, losing my life to that game, but still...bad. e.e;; bad.  Or I could get the Age of Mythology expansion pack..the Titans thing..but I\'d also lose a large chunk of my life playing against people online with that. I mean, I\'ve already wasted countless hours playing just plain old AoM online against people that can\'t speak my own language...  hm. I\'m also trying to find the old N64 game SnowBoard Kids, with no luck. Aaron and I used to rent that game everytime he\'d come over to my house, so I really want it just for the nostalgic factor, I guess. It was a good game, though e.e I\'ve found it on Ebay a few times, but most of those are being shipped from the UK and I\'m not going doing that, with huge shipping prices \'n whatnot.  Ok. I\'ve rambled on and on..time to bring this post to a closing. Uh..bye ^^;  Song for today:  urlLink This is Not an Exit  by Saves the Day  '),(20458,2108549,'31,October,2003','2003-10-31 00:00:00','    It was a good Halloween. Much better than last year. Much different than last year too but very good.    The first week of work ended uneventfully, the way it should have.  I\'m glad the rain is here but it definitely means I needs more winter clothes!  Shopping! Well, not yet but soon.   '),(20459,2108549,'30,October,2003','2003-10-30 00:00:00','    Spam mail.  In a big way.  Especially when it is disguised as mail that I might actually want.  For example \"Per our chat last night\", \"Hi\", \"Job leads\", etc.  I got one this afternoon from a site that supposedly someone has information about available.  But I can\'t find out what the information is unless I pay for it.  Now I ask you, if the information were valid, which I doubt this is, wouldn\'t I know it already?  If this information were about me and truthful, wouldn\'t I, being me, know what it says?  Why would I need to pay for it? And to make it more obscure, the information is only listed by an email address I use for resumes.  Not by name, or a more often used email but by one used for job hunting.    Just another example of the many. many scams on the net.  So yeah, I am a tad irked by it.    In other areas, work is still good.  Life is still good.  45 minutes til the weekend starts!   '),(20460,2108549,'30,October,2003','2003-10-30 00:00:00','    I like coming back from lunch to find that they all locked themselves out of their accounts.  It gives me phone calls to take that are relatively simple and easy to fix in nature.   Work is good, slow Friday, but good.  Happy Halloween everyone!   '),(20461,2108549,'30,October,2003','2003-10-30 00:00:00','    Things here are so slow I almost want to fight over the phone calls!  I can\'t surf any more sites.  I check my email  all  the time and frankly I have thought about redesigning the company website.  So yeah, there ya have it.  I\'m bored.  Oh I love the work, love the group of people I work with, I just wish there was more of a workload.  :)  Its a nice feeling.   '),(20462,2108549,'29,October,2003','2003-10-29 00:00:00','    I just realized something this morning.  Since my life has gotten steered back in the right direction, readers have dropped.  Now while I have never really cared about the number of readers, I do find it interesting that when the drama subsides, and I stop whining that people stop reading.    It must be really hard to only want to see the bad things in people\'s lives instead of wanting to celebrate the good stuff with them too.   '),(20463,2108549,'29,October,2003','2003-10-29 00:00:00','    Day 4 and counting.  I am  so  looking forward to sleeping in on Saturday. The phone is gonna be turned off tomorrow night, the drapes drawn and if Freedom wants to stay he is gonna have to snooze with me.  Although I doubt that will be a problem.  I have thrown the poor boy into a tizzy this week with getting up so early.   I\'m trying to find that balance between all work and no play and all play and no work.  I\'m hoping that once my body settles into the routine, the mind will follow. I\'ve been eating a lot healthier this week and somehow I think that is making it worse.  *sigh*  I know.... its a process.   '),(20464,2108549,'29,October,2003','2003-10-29 00:00:00','    *snort*  I just broke my desk.  Well not the whole thing but the lil drawer in the front.  My chair arm caught it and ripped off the front panel.  So I fixed it.  With clear packing tape.  Always the resourceful tech girl am I.   '),(20465,2108549,'28,October,2003','2003-10-28 00:00:00','    Work is good.  I am really enjoying being back in the computer field and all of my fears were for nothing.  Once again, those around me were right and I was wrong.  :)  Gee, like that is an uncommon occurrence! Things are good though, I am just having a really hard time getting into this being awake at the butt crack of dawn thing!   '),(20466,2108549,'28,October,2003','2003-10-28 00:00:00','    And right now that beholder is me! The call volume is just too low to keep this girl active.  So talk to me!  I can be reached through any variety of chat programs (AIM, ICQ, Yahoo, MSN) Leave me your user name, who you are and why you want to talk and I\'ll add you!   '),(20467,2108549,'28,October,2003','2003-10-28 00:00:00','    Well its 11 somewhere in the world right now.  And there is simply no way I am going to be awake at 11 my time tonight. The day was good.  I am just so not used to getting up so dang early!    I do like the job though.  I like the guys I work with.  Eventually I will like the schedule.  The only thing I don\'t like is that my loafing time has been seriously cut in to!  :)    Basically though, I\'m a happy camper.   '),(20468,2108549,'27,October,2003','2003-10-27 00:00:00','    I am likin\' my job.  Albeit its a bit slow and a few times I think I may have close to the end of the internet.  Overall though I am happy here.  Its kinda nice.   '),(20469,2108549,'27,October,2003','2003-10-27 00:00:00','    I think it might be possible that I  will  indeed find the end of the internet.   '),(20470,2108549,'27,October,2003','2003-10-27 00:00:00','    Funny story of the day...  Ju and I went to walk at McKinley Park this evening after work.  Naturally, Freedom came with us.  Well, he was amped up and hyper as all get out when we first set out.  So being the oaf that he is, he can\'t figure out what side to walk on or which path or even how to walk in a straight line.  There is a raised concrete border around the dirt path, its probably 4- 6 inches tall, maybe 4 inches wide.  My boy, goofball that he is, kept jumping and walking on alternate sides of it.  2 or 3 steps on the right of it, 2 or 3 steps on the left of it.  When that got boring he even managed a few steps on this border like a little kid.  Oh how I wished for a video camera.    That dog is the biggest gift I will ever get in my life.   '),(20471,2108549,'26,October,2003','2003-10-26 00:00:00','    So the first day went well.  As usual, I was afraid for nothing.  None of the things I was scared of happened.  I wasn\'t and am not 100% comfortable there yet but com\'on its only been 1 day. My schedule is gonna be 7:30-4:30 M-F.  Its kinda nice its makes it so I miss a lil of the rush hour traffic.  It took almost an hour to get home tonight for what is usually a 20 minute drive.  With going in earlier it\'ll probably only take a half hour.  I\'m taking  urlLink this picture  with me tomorrow to put on my desk.  I am the only female there and the guys are really really nice, its kinda fun to be treated like one of the guys.  And I have the whole 2 stall bathroom to myself!  :D   '),(20472,2108549,'26,October,2003','2003-10-26 00:00:00','    First day of work and things are going good so far.  I like the guys I am working with!   '),(20473,2108549,'26,October,2003','2003-10-26 00:00:00','    I get to go to work tomorrow. Yes, get is the correct word. See, when you have never been laid off, never been too sick to work, never stayed home day after day, you think work is a right. Its like its something you have to do and you are entitled to do it. After being home for so long, after all the lay-offs... I  get  to go to work tomorrow. The day might suck. I might fail 100%. I may even hate what I\'m doing. But really... those are the worse things that can happen right? They can\'t take away my car, my dog, my sobriety, all they can do is fire me. And if those are the worse things that can happen, then I\'m ok if they do. I\'d really  really   really  rather they didn\'t but I can survive that stuff, done it before.   So I\'m looking forward to it. At least this hour.  I\'m sure it will change and I\'ll get scared again before 8am tomorrow but that\'s ok.  Freedom is a good smellin\' pup again, laundry is done, and I got good healthy stuff from Trader Joe\'s for lunch so I think I\'m set. Not much left to do but sleep! :)   '),(20474,2108549,'25,October,2003','2003-10-25 00:00:00','    This year has been such a whirlwind of good and bad.  It started out so horribly bad and looks like it is going to be ending so amazingly good. (These things have been published in this journal as they ahppened as well.)  In the last 2 months alone my life has flipped yet another 180 degrees. It all started after finding out Freedom need surgery.  I set up a website hoping to raise even just $100.  Five days later I had raised almost $800.    Three weeks later I got into a pretty bad car accident, and while I walked away pretty much unharmed, my car didn\'t.  For 2 weeks I went back and forth with the insurance company and body shop trying to get it fixed. The day I got a call saying it was to be totaled, I got a phone call from M.A. saying it was time to go car shopping and it was his gift to me.  I got a 1994 Explorer the next day.  A vehicle I have wanted for years and never thought I would own.  I was still stressed out about the difference between what I owed on my totaled car and what the insurance company would give me but at least I had wheels now and knew no one could take them from me.  A week later I got the call from the insurance company I had been dreading.  It was time to discuss the figures for my car.  To my amazement, they were paying me $1500 more than the most I thought I would get.  I called the finance company to set up a payment schedule for the balance.  Once again, I was being taken care of.  The difference between what I owed and the pay out from my insurance company was only $134. Not the $2000 I feared.    During that same week I got a call from a recruiter for a job I hadn\'t even applied for.  He simply saw my resume on  urlLink Monster.com  and called.  3 days and 2 interviews after his initial call, I now have a job back in the technology field.  My only requests were that this job let me have Sunday morning free for church and that it pay at least $12 an hour so I could keep my financial obligations and still move out.  This job is weekends off and well over that rate of pay.  Freedom had his surgery and is doing wonderfully.  He was back to his old self within a couple of days.   I won\'t go into all the horrible things that happened at the beginning of the year because they don\'t matter. Sure, they still hurt and I definitely still feel the sting from some of them, but the wonderful ways I have been taken care of this last couple months more than make up for that.  Sometimes you have to walk through the fire to come out alive. The things written about here are just some of the things that have happened, just some of the gifts I have been given.  There isn\'t enough time this morning to explain the inside changes that have begun to happen.    I haven\'t done anything to deserve the gifts I have gotten.  I haven\'t even begun to write about the all the gifts that have come simply by being sober.  The only thing I continued to do was keep on keepin\' on and to stay sober.  I cried, I threw temper tantrums, I wailed against fate, people and God, I surrendered and took it back but I kept on going.  That is the only thing I did right.  Everything else wasn\'t me doing it.  I guess the moral of this story is no matter what happens its going to be ok.... eventually and that God does take care of us but sometimes it takes a bigger smack upside the head for some of us.  A lot of human will was involved in all these things, but I believe it was God using them to help me.  I want to be able to be of that same service some day.   '),(20475,2108549,'25,October,2003','2003-10-25 00:00:00','    What an amazing, wonderful day. Went down to SFMOMA (San Francisco Museum of Modern Art) to see the Chagall exhibit. Wow.  Its the only thing I can say.  To try to describe the works I saw today simply lead to sensory overload all over again. So many many many colors and images and sights and sounds of the museum, the smells of the visitors, (not all good) it was just SO intense.  And something I wouldn\'t trade for the world.    If someone told me a year ago I would be in SF twice in one week, viewing  real  art each time, I would have said they were crazy. While art is still like many subjects in my life, I am sorta kinda starting to get it.  I definitely know what I like, definitely know what I don\'t like and the millions of works that lie between are eye candy right now.    It was just amazing. The day was so awesome, wonderful sights, wonderful company, wonderful conversation.  Just enjoying life with a couple of friends.  I\'m not sure it gets much better than that.    I am so very tired though and that extra hour of sleep tonight is going to be put to good use!   '),(20476,2108549,'24,October,2003','2003-10-24 00:00:00','    I went to a potluck tonight with a bunch of women I don\'t know.  It was fun though.  I dunno... my small talk skills suck in a big way and I never have anything to say.  I know it a matter of learning and stuff but, well, its gonna be a long process.  I am so sick of everything being a process.  What I wouldn\'t give to have a transformation.   '),(20477,2108549,'24,October,2003','2003-10-24 00:00:00','    blech.   that is all.   '),(20478,2108549,'23,October,2003','2003-10-23 00:00:00','    Its been a long day.  I have been on the run since 9:30 this morning and am finally home.  I\'m drained, emotionally, physically and mentally.  My brain is on hyperoverdrive.  I\'m overly sensitive now (yes, more so than usual) and I\'m hoping sleep tonight takes care of it.  It wasn\'t a bad day, actually it was pretty good.  Just busy.  Looking forward to hanging out with a friend tomorrow afternon.  Saturday is SF with Ju to see the Chagall exhibit and Sunday is getting ready for Monday.  :)   '),(20479,2108549,'22,October,2003','2003-10-22 00:00:00','    I am in a serious state of fear.  Things have gone too good too fast.  I feel like I am walking on that sheet of thin ice in that old board game \"Don\'t Break The Ice\". I\'m waiting for the other shoe to drop and I am terrified of what it will be.    So my choices are fear or anger.  Fear to me, is an emotion of inaction.  Its paralyzing in its grip and it causes me to freeze up.  Anger is an action emotion to me.  I run fast. I drive fast. I talk fast. I react fast. I do fast.  Fear is simply not an option so I\'m angry.  It makes no logical sense and this angers me even more.  That I can be so stupid, so illogical.  Knowing that my fears are illogical or at the very least, somewhat improbable.  Its sort of like the notion of winning the lottery.  Oh sure, its a possibility but not a very probable one.  I can intellectually tell myself this stupid all I want, it doesn\'t make the fear go away.   Yep, I\'m a big girl and big girls don\'t get scared and they don\'t cry. So what am I so damn afraid of.   I\'m scared that I can\'t do this job.  Scared that I can\'t adapt to the routine of working again.  Scared that I can\'t remember how to be a tech support agent. Scared that they won\'t like me.  Scared that they will find out \"who I really am\". Scared that everything negative I have ever thought about myself will be confirmed in one day.  Scared that this new way of trying to live is going to collapse and I will be left in a limbo of not being to able to live in the old ways and not able to grasp the new ones.  Just plain scared.  So my option is to freeze entirely or get pissed off about it.    I want to hold everyone at bay for fear if they see me like this they will run. I am tired of needing support and guidance, tired of needing period.  I am getting disgusted with myself and while my life is yes, finally pulling itself in the right direction, that scares me even more. It boils down to this: I don\'t particularly care for me therefore no one else can and once they realize how inadequate, bad, needy, stupid, incapable I am, they will leave anyway, so I might as well give them that push in the other direction.  Because you see if I push them away its my decision and it won\'t hurt as bad.   '),(20480,2108549,'21,October,2003','2003-10-21 00:00:00','    I got the job!  *happy dance*  The interview went ok but I had that gut instinct saying I didn\'t get it.  Yet I got a call from the recruiter and I start Monday.  Yippee!!   '),(20481,2108549,'21,October,2003','2003-10-21 00:00:00','    And the good news still keeps on coming....  I went to the job interview this morning.  It went very well, I liked the interviewer and I think he liked me.  We just clicked and that is always a good feeling.  The job sounds like a wonderful opportunity, more money that I would have asked for and \"normal people\" hours. I have an interview with the managers tomorrow afternoon and should know sometime tomorrow if I got the job or not.  The only scary thing is that it would start Monday and well frankly, I\'m scared. Its been a long time since I have worked and even longer still since I have worked in a call center technical environment. I know I have the skills.... they are just a bit rusty. There is no way to polish them up though without jumping right in.  Its an awesome opportunity and if I am offered it, I\'m definitely taking it.  I also spoke to the car insurance company and my finance company, the difference between what I owe on the Honda and what they are paying me for it is a total of about $140.  So that major financial burden is gone.  Once the contributors to Freedom\'s Fund are paid back, I\'ll owe very little and I can go back to school.  Hopefully in January.  The idea scares me silly though.    There has been so much good this week that I am looking forward with one eye closed waiting for something really bad to happen.  But ya know what.... I don\'t think its gonna.  God has taken care of me beyond my expectations and I have people in my life that help when needed.  Its amazing.  I\'m starting to feel lucky to be me instead of lucky to just have \"stuff\".   '),(20482,2108549,'20,October,2003','2003-10-20 00:00:00','    I got up and call my insurance company to find out how much I owe them.  Turns out through a snafu on their part, I haven\'t been insured the last week and a half.  3 hours later it got straightened out but sheesh  its been a while since I have been  that  frustrated. I also talked to M.A. who saw the post on my entry about the painting.  His mother in law is on the arts council and she got him a phone conference with Robert Watson.  I\'ll have a lithograph of my painting in 7 weeks!  I am way excited.  I can\'t wait!  That painting did something to me.  And now... I\'ll get to see it every day!!!!    I also got a call for a job.  I\'d be doing tech support for AAA\'s call center employees.  I go talk to the contract guy tomorrow.  I\'m not supposed to go back to work yet but.... in this job market, I can\'t turn it down if the money is right.  Its temp to perm with a 3 month commitment.  I\'ll get the work hours and pay scale tomorrow I guess.    All good stuff but definitely not any of what I thought would happen when I opened my eyes this morning!   '),(20483,2108549,'19,October,2003','2003-10-19 00:00:00','    While I doubt that it is my lot in life by design, I get the idea I will be part of the eternally frustrated.    I\'m frustrated with myself, my thought processes, my direction in life, my lack of... (not material stuff but inside stuff). I know that I have come a long way in sobriety.  I know I have come a long way this year.  Yet I want to be further along.  I want to know more and I can seem to learn it as fast as I want.  I want I want I want seems to be the motto and what is more frustrating is while at times that applies to material things, more often than not lately it is applying to inside stuff.  Stuff that cannot be bought or traded, things that cannot be bartered for.  I know, I have tried making deals and bargains and I don\'t seem to be getting it through osmosis.  I  know  its a process.  I  know  I have made progress.  I know too though there is still so much  more  progress to be made.   I want it all NOW.  More better more different more now.   '),(20484,2108549,'17,October,2003','2003-10-17 00:00:00','    I had the most amazing experiences yesterday.  I saw REAL art.  I saw Dali and Chagall and Rembrandt and Picasso. There was a gallery we walked by with the most amazing picture of a field of sunflowers and by this time I was starting to really like what I was seeing and even understanding the faintest amount so I went in and found a painting I fell in love with.  I can now say I understand when people say \"It spoke to me\".   urlLink This painting  (Turning Point II by Robert Watson) might as well have been a book from everything I read into it.  There is the person standing there, they have done some growth in life, coming out of a darkness and into the life. They have done footwork by walking up the steps and they are now at a turning point.  They look tired, weary from their walk but they have to decide.  Do they keep going, tired as they are farther into the light?  Or do they keep walking on the plateau they have reached into an abyss of nothingness?  Or there is the decision to turn back and return to the darkness. There is the white flag of surrender they will reach if they keep walking. Yet they have already experienced that light by simply starting to walk up the stairs.  There are also 2 shadows that seem to be in a face off if they walk into the abyss.  Its so amazing.    I might never own the original but someday, I\'m going to own a print. I couldn\'t stop thinking about it after we left the gallery and obviously can\'t stop thinking about it now.  I want to meet the artist, sit down with him and say \"this is what I saw, what did you mean? And what does this signify? And why did you choose these colors? And and and\" :)  I know my facial expressions always give me away.  Love, hate, sadness, hurt, etc have never been unreadable in my face through my eyes.  I can\'t conceal what I feel and I\'m sure the excitement and comprehension yesterday was no different.   '),(20485,2108549,'17,October,2003','2003-10-17 00:00:00','    More pictures!  The photos from the camping trip were also picked up today.  I haven\'t scanned them all in yet but  urlLink here\'s  the cutest of the bunch (all Freedom of course).   Early morning plans tomorrow so I am off to bed.   '),(20486,2108549,'16,October,2003','2003-10-16 00:00:00','    I got the photos back from Freedom\'s surgery.  The pictures can be seen  urlLink here .  Not a bad day all in all, went and had lunch with M.A. Afterward I was supposed to go meet with my sponsee but she cancelled so Freedom and I headed over to the park so he could get a little fresh air and exercise. Went through Sam\'s Club to get the pictures and now I\'m home.  I\'m so tired lately, I\'m hoping tonight will be a good sleeping night.   '),(20487,2108549,'16,October,2003','2003-10-16 00:00:00','    I think people learn how to treat us not by how we treat them, but by how we let them treat us.    Most of my life I have let people make me feel like I am less than them.  That I must be in some constant state of need or crisis for the friendship to continue.  What I am learning though is that the longer I am sober, the less crisis and need I have.  And what has happened is there have been times that I have needed to generate some crisis for fear of losing that friendship.  That doesn\'t need to happen anymore.  Through events in the last few weeks, the number of friendships I have is dwindling.  When I first realized this I was almost panicked.  \"What\'s wrong with me?\" and after more thinking and writing, I realized, maybe the unhealthy friendship have to go so that I can focus on the few healthy ones I have and begin other healthy friendships.    I don\'t think any relationship can ultimately last when it is based on need.  I think it has to be a relationship built on commonalities, and care and concern for one\'s well being without having to fill the need to control the other person.  More often than not, I have allowed myself to be controlled and while I didn\'t like it, fear of rejection or having to stand alone kept me from stopping that control.  Its something I don\'t want in my life and I guess identifying the problem is the first step.   '),(20488,2108549,'15,October,2003','2003-10-15 00:00:00','    The meeting in Coloma was so cool!  We saw 4 or 5 deer on the drive in off Hwy 50, there were coyotes howling during the meeting and the stars we so bright!  I\'m really glad I went. It was well worth the drive.   '),(20489,2108549,'15,October,2003','2003-10-15 00:00:00','    New layout.... once again. I just want a simplistic way to upload this journal so that I am not fussing and fretting with formatting and HTML scripting.  Hoping this is better.     I went this morning to get my pins out but.... the doctor said they need to stay in for another month. The bone isn\'t callusing or calcifying or cal-something the way its supposed to. And removing the pins at this point will lead to having to have a plate put in later. So the pins stay. One surgery and one set of metal things in my hand is enough for such a lil finger. I have a new cast though and I\'m not going to be able to work for another month.   I also got a call from the woman I was to start house sitting for next month and her deployment date has been moved to December at the earliest. I\'ll still get to house sit, just not yet. I spent the evening at Ju\'s last night and had a little of the emotional meltdown that has been building as of late. We relaxed and I ended up falling asleep while watching  urlLink Gilmore Girls . There is just something about the house that makes me feel all safe and sometimes safe can lead to sleepy when the lights are off and the TV is on.   I\'m supposed to go to a meeting tonight in Coloma but I am SO sleepy. I don\'t know if I\'ll make it. Gonna have to see how I feel in a few hours. Right now is definitely nap time.   '),(20490,2108549,'26,November,2003','2003-11-26 00:00:00','    Lately I can\'t seem to focus. On anything really. I\'ve been compared to a butterfly just flitting from here to there without really landing anywhere and I guess that\'s about accurate. Its not just the apartment or the job or anything really. I think its a mixture of too much at one time. I just can\'t concentrate. ADD child at her worst right now. Normally I could deal with that but with all the changes and because of all the stuff I\'m not sleeping well, hence the 4 am post. Its make me hypersensitive and I feel like I am on the verge of tears most the time. I hate feeling unfocused and ungrounded. I\'m hitting a meeting on my lunch hour today.   I have a bunch of guys coming over Saturday morning to help me move, or rather move me. Friday there is no work which is a mixed blessing. I could have used the money and no work means no pay. At the same time I have so much to do its kind of a good thing. Not sure exactly when this phone number gets turned off only that the new one will be active today I hope. All the utilities go into my name today and cable gets installed Saturday. I pick the keys up for the apartment on my way home tonight from work. Tomorrow is the Race for the Hungry and then I think I am just going to bring stuff to the apartment and veg there afterwards with Freedom. I was invited to dinner at a friend\'s place and I might run over to eat but I really kind of just want to relish in being *home*.   '),(20491,2108549,'25,November,2003','2003-11-25 00:00:00','    Things are good.   The day started out with the \"what can go wrong will go wrong\" type stuff but it has righted itself enough to turn into a good day.  I\'m so excited about moving, it seems to be the only thing I can really focus on these days.   Its going to be so nice to have my space back and to have my own place.   '),(20492,2108549,'24,November,2003','2003-11-24 00:00:00','    Its been fairly quiet here today.  This morning was busy but the last hour or so, I\'ve spent looking around online at curtains and towels and just stuff like that.  Made  urlLink wishlist  and looked at deals on ebay.  Its kind of fun imagining all the possibilities.   '),(20493,2108549,'23,November,2003','2003-11-23 00:00:00','    Ay vey.  I have spent most this weekend packing, getting boxes and obsessing on how to decorate, trying to get prices on blinds and curtains and the like.  I have a lot of ideas and little money.  And for once, that\'s ok.  This gets to be an ongoing project and that sort of makes it fun.  My only real have to have item right now is curtains and blinds.  The rest will come together with time.  Ju and I went and looked at it again last night and I still love it.  It just has a warm feeling when I look at it.  I can\'t wait to move in. Only 3 more days before I have keys and less than a week from now, I\'ll be moved in.   '),(20494,2108549,'21,November,2003','2003-11-21 00:00:00','    Things are good.  Good day at work.  Getting things set up for the apartment.  I have never had such a short notice on moving before and normally I\'d be so panicked but given that I am in such a liberal environment, I have gotten everything set up.  Its absolutely wonderful.  I\'m happy here and that is the best gift any job could give me.   '),(20495,2108549,'20,November,2003','2003-11-20 00:00:00','    Today I got the phone, PG&E, and SMUD set up, tomorrow I\'ll get the cable done and send in a change of address card.  I am SO excited!!   '),(20496,2108549,'19,November,2003','2003-11-19 00:00:00','    I have a new apartment.  Ok its really called a flat but I can\'t tell the difference.  Its 2 bedrooms, one bathroom, I have a washer and a dryer and a dishwasher and a fridge and I LOVE it!.  I can\'t wait to move in! I\'m getting the keys next Wednesday and think I might just spend Thanksgiving day in my apartment!    I also got the pins out today.  It hurt.  A lot!  But I\'m glad they\'re out and while I still have a cast, at least I know I am on the road to recovery!   '),(20497,2108549,'17,November,2003','2003-11-17 00:00:00','    Its good stuff but still a problem nonetheless.  Where I am working now, I am on a 3 month contract. Nothing is guaranteeing that I will become permanent. I like it there though, I have plenty of time to study should I choose to, I love the guys I work with. Financially even if I got that apartment, I could swing it.   It appears though that I might be offered a position with another company. As a database analyst. Something I would have to learn. Major stress there. But its about $5k a year more, permanent and with benefits. Its a longer commute as well. Its also a great opportunity and the money would be nice.   I don\'t know what to do. I am thinking about talking to my current boss, not as an ultimatum by any means, more though as a help! type thing. If they were to take me on permanent, even at the rate of pay I make now, I would stay there over going to this new job. I\'m safe there, comfortable there, happy there and the money just wouldn\'t be worth it. But I need benefits. I need sick days. I need vacation days. I need to know my job is definitely going to be longer than 3 months. Especially before committing myself to this apartment.  Help.   '),(20498,2108549,'13,November,2003','2003-11-13 00:00:00','    I\'m here.  I\'m sick.  But I\'m here.  Its been a busy but good weekend.  I think I found the perfect apartment.  Its a little more, ok a lot more, than I wanted to pay every month but its almost kitty corner to Ju\'s place, 2 bedrooms and I can have Freedom.  We\'ll see, I turned in an application because they weren\'t requiring an app fee sooooo maybe I\'ll get it.  I can swing it financially, it would just be tighter than I wanted it to be.  But it would be safe and its freshly remodeled and I love it.  So I\'ve done the footwork and if I an supposed to get it, God will provide the rest.   '),(20499,2108549,'13,November,2003','2003-11-13 00:00:00','    I\'m tired.  I\'m sick.  I have way too much going on in my life not mentioned in this journal that I have to deal with.  I\'m in effect shutting down for a few days.   '),(20500,2108549,'12,November,2003','2003-11-12 00:00:00','    House-sitting has been canceled. Military snafu.   '),(20501,2108549,'12,November,2003','2003-11-12 00:00:00','    I picked up the keys yesterday for the house I\'m be staying at.  Yay!  The day was incredibly long though and I woke up this morning with the beginnings of a cold.  Gee, Shannon, you thought you would go literally non stop for 2 weeks and NOT get sick? Duh.  I got a call though from a girl in a recovery house.  I spoke there a few months ago and she has a pass this Saturday and has to spend the day with another woman in recovery.  And she called me.  She\'s a sweetheart and has my name! So we are gonna hang out Saturday afternoon.  We can hit a meeting and then what?  Coffee I guess and then... I dunno.  It will be interesting.   '),(20502,2108549,'12,November,2003','2003-11-12 00:00:00','    Exhausted.  Overwhelmed. Bed time.   '),(20503,2108549,'11,November,2003','2003-11-11 00:00:00','    Sometimes I just want to be held.  That craving for physical human contact becomes overwhelming. Nothing more than simply being held, hugged, unromantic yet compassionately touched. Yet pride, ego and fear get in the way.  So many have ulterior motives, hidden agendas, unspoken feelings.  It becomes not worth it when the cost is considered.  Sometimes I want to disappear.  Not run away, simply disappear for awhile.  No commitments, no responsibilities, just nothingness.    Sometimes I cry at funny movies and laugh at sad ones.  So many scripts and books are written to elicit emotion, yet sometimes they are overdone and simply comical.  Sometimes I wonder if there is truly a way to rewrite the past. Not to change events or their outcomes but to change my perception.  To make it so that lessons were actually learned instead of having to be repeated.    Sometimes I long to be ugly.  While I am no show stopper, people tend to see me but not hear me.  How frustrating it is to know that intelligence has to be proven for someone to see past the outside.  Sometimes I drive fast just because I can.  Something akin to tempting fate.    Sometimes I smile when I only want to cry.  No one wants to be around someone who is always upset.  It pulls people down, its draining and overwhelming to be around all the time.  Sometimes I wish that everything I have done right I did wrong and everything I did wrong I did right.  Sort of a checks and balance system to see what side adds up more.  Sometimes I don\'t want to know me.  For what I see on he inside you could never fathom when looking from the outside.  Yet what you see on the outside I can never understand from looking within.  Sometimes I long to be ignorant.  To be happy with *less than* or under-achieving.  What I expect from myself, I would never dare to expect from others, yet the expectations rise daily.   '),(20504,2108549,'10,November,2003','2003-11-10 00:00:00','    I had dinner last night with a couple friends and all their babies.  It was nice.  They are both exceptional people with quantities of love and acceptance I can only hope to have one day.  I got on of the Jars of Clay CDs yesterday in the mail from a friend on  urlLink LJ  and while I haven\'t had a chance to listen to it yet, I can\'t wait!  Yesterday M.A. and I had lunch.  I realized as I was driving to meet him, just how much I miss having those weekly lunches!  Work definitely bites in regards to that.    Freedom has a staph infection, nothing major that antibiotics and a sterios shot won\'t clear up but he was definitely uncomfortable.  $100 my baby is non-itchy, and relatively happily back on his way to recovery.  I am just incredibly tired.  I have plans tonight and tomorrow night but Thursday and Friday are going to be reserved to getting some rest.   A friend and I were talking last night and apparently its time for me to grow up.  This wasn\'t said meanly or cruelly, nothing like that at all, simply matter of fact.  This is probably very true.  The problem is though that I don\'t know what the traits of a grown up are.  Does it mean that I do everything by myself and don\'t lean on people?  Does it mean that I don\'t talk to anyone about what\'s going on? That I don\'t ask anyone for help? I don\'t know what *grown up* looks like. Its not like I have had a lot of examples to follow.  And I think my fear of growing up stems from the fear of not doing it right, especially when I don\'t know what *right* is.  Its frustrating.  I just want to do everything right all the time.   And without a way of measuring right, I don\'t know how to do it.   '),(20505,2108549,'10,November,2003','2003-11-10 00:00:00','    It seems that when there is too much going on I shut down.  This has been the case as of late.  Too many plans, too many emotions, too much stuff - too overwhelming for me. Everything is going as it should I guess, its just been too much for me and when I have had the time to sit and write, I have instead just shut down.  Its been necessary.    I saw Jars of Clay last night with Caedmon\'s Call an they put on an awesome show.  Saturday night was Girls\' Movie night and it was nice.  We ended up sitting around and talking instead of actually watching a movie but the conversation was good.  I have the beginnings of a new life falling into place and its good.  I\'m scared though and that tempers the excitement. Studying, women friends that want nothing more than friendship, housesitting, new job, new car, new well.... me.  I guess its normal.    It\'s amazing to me that people actually trust me with things like housekeys and their house.   I start housesitting for 6 months on the 24th.  I\'m happy about it, a little scared about it, however I am mostly looking forward to the space it is going to give Freedom and I.  He\'ll have a huge backyard and I\'ll have a 3 bedroom house to roam around in.  Its a good thing.   '),(20506,2108549,'07,November,2003','2003-11-07 00:00:00','    Happy Friday!  Today is much better than yesterday was. At least emotionally to some extent.  It can\'t be perfect though!  If it was I\'d be bored!  Busy weekend ahead.  Good but busy. Other than that, everything is status quo!   '),(20507,2108549,'06,November,2003','2003-11-06 00:00:00','    I miss my dog.    How stupid is that.   '),(20508,2108549,'05,November,2003','2003-11-05 00:00:00','    I\'m cold.  I\'m cranky.  I didn\'t sleep well at all last night.  I\'m hungry.  This is not a good day.  Not a good day at all.   '),(20509,2108549,'05,November,2003','2003-11-05 00:00:00','    I am just too busy for my own good.  Its all good stuff but I am getting tired. Plans to be out every night are great however, this girl needs some downtime.  Next week I\'m scheduling it.  The flip side though is that I am grateful to have a life back.  After I started seeing M, I gave up a lot of it to sit and wait for him.  I lost touch with a lot of people and stopped making plans I should have made.  When he and I split up, I was left with nothing but a lot of time on my hands.  Moderation though is obviously a lesson I haven\'t learned yet!  :)   '),(20510,2108549,'04,November,2003','2003-11-04 00:00:00','    Things have been somewhat of a whirlwind since Friday.  This working thing has seriously cut into my social life! :) Friday I worked and then went and watched an awesome movie at Ju\'s.  Chinese food and a good movie and good convo equals a good night. Saturday was a little shopping.  I needed to get new clothes for work.  It was time and while I had enough to get through last week, there was no way I could go through the same 5 outfits again.  So I got a buncha stuff!  Almost everything was 50% off and I got awesome deals all the way around.  Sunday was dinner at an old friend\'s house.  A few of the pictures have been posted already.  Things are good just seriously busy.  I think I might have to let a couple commitments go.  I admit, I am a wimp.  I like my down time, in fact, my down is essential for me not getting sick.   Life is life though.  Its going really good.  So good that how it looked 6 months ago and how it looks now are two different beasts altogether!  And that my dear, is a good thing.   '),(20511,2108549,'03,November,2003','2003-11-03 00:00:00','    Busy busy busy day. It was a good day, just very busy.  Went and met B for coffee with Freedom after work.  We had a good time and once again, Freedom was so well behaved I was amazed he was the same dog!  Work was good, same ol, same ol.  And at this point I am exhausted and ready for bed.  Tomorrow is going to be even longer.   '),(20512,2108549,'02,November,2003','2003-11-02 00:00:00','    Pictures and a movie from Freedom after last night\'s dinner    urlLink  The movie       urlLink Say What?    urlLink Attack the Momma!    urlLink But Mom I don\'t wanna!   '),(20513,2108549,'02,November,2003','2003-11-02 00:00:00','    It was a good day.  I went to church this morning and then did a project I have been meaning to do for a while and then off to dinner with a woman I used to work at Apple with.  She even invited Freedom to dinner.  He was so well behaved. Far better than I thought he would.  It was a nice evening and now the rain is pouring and I am lookin forward to cozying up and falling asleep to the sound of it.   '),(20514,2108549,'02,November,2003','2003-11-02 00:00:00','    Our society is essentially made of *haves* and *have nots*.  I\'m not referring to a *have not* as someone who is driving a beater of a car and working at a lowing paying job.  I am referring to the *have nots* that literally have nothing but what they push in a shopping cart or carry in a bag.  They live on the streets of every city of every state in this country.    While a man in a BMW fiddling with his laptop, talking on a cell phone fights with his wife about what house they should purchase next, the man on the corner wonders when he will get to eat again.  A lot of people ignore them, for blindness is easier than caring.  It is easier to pretend they do not exist, than to admit \"that could be me\".  We\'re all guilty of it. Everyone has done it.  More often than not I haven\'t given to the man or woman holding the sign because I am greedy and cannot police the way they are going to spend my spare change.   What if though, every one of the people in the *have* group made one bag lunch every day and gave it to a *have not*.  Something simple, a sandwich, a juice box, some fruit and some chips.  The kind of lunch we resented getting when we were 13 because they weren\'t cool.  Its inexpensive to do, takes very little time to make and even less time to give away.  Wouldn\'t then there be a little less of the *have nots*?  Its hard to make a good impression for a job interview when you haven\'t eaten in 3 days.  Its hard to not consider shoplifting that sandwich if its the only prospect of food.  Sure, some are homeless by choice.  A lot aren\'t though. Through any series of decisions they made, they are living on a street with nothing.  I know I could have easily been there.    Last year it was a serious possibility.  It might not have been long term and I am sure any one of my friends and family would have let me stayed with them but not everyone has those options.   No single person can feed everyone.  But every single person can feed just one.   '),(20515,2108549,'31,December,2003','2003-12-31 00:00:00','    Happy New Year!   '),(20516,2108549,'30,December,2003','2003-12-30 00:00:00','    Ok so I got the job.  I am going to be an official Intel employee.    I start Jan 19.  Wooooooooohoooooooooo!!!!   '),(20517,2108549,'28,December,2003','2003-12-28 00:00:00','    My journal is going to be changing.  I have never made it a secret that all hits to this page as well as several others on my website are monitored. I don\'t care too much really about who reads what on here however when I see the same IP address hitting my journal five ten or fifteen times a day, it makes me uncomfortable. I have blocked access from thier IP using an htaccess file, they change their IP, I block the new one, they get a third one.  So I am in effect, going underground with my journal for an unknown time period.  If you wish to be added to the mailing list and have my entries emailed to you, please  urlLink email me  and let me know.  Maybe this seems a bit paranoid or defeating the purpose and to be honest, I don\'t care.  Its unnerving to have someone checking back so very often during the day to see what is going on in my life.   '),(20518,2108549,'25,December,2003','2003-12-25 00:00:00','    I worked yesterday until 1 and then came home, took a nap and B stopped by.  I gave her E\'s presents and got mine, and then I met Ju for church.  It was a beautiful service, candle lighting at the end and it really brought home the true meaning of Christmas.  We went to dinner at  urlLink Mikuni\'s , a wonderful Japanese restaurant that opened not too long ago. Oh man was the food good!  Came home and watched   urlLink Evelyn  , a total tearjerker, very, very good though.  I fell asleep during the second movie,   urlLink A Christmas Carol   .  Such a difference from last year\'s Christmas Eve.  I expected last year to be this big wonderful event and ended up being disappointed.  I expected this year to be uneventful and lonely and it turned out to be absolutely wonderful.  Today I slept in lately, went and saw a friend, her husband and their twins. Then my mom came over, we ate, I played with my  urlLink new video camera .  WOW!  This thing has SO many functions.  I can\'t wait to start recording stuff! Thank you again Dad.  You have spoiled me rotten!   This evening has been mellow and I am getting ready to head to bed.  Working the opening shift tomorrow and being there at 6:30 AM bites in a big way.  It\'s ok though, one early day and then a sleep in day.    I hope everyone had as wonderful a Christmas as I did.  Merry Christmas everyone.   '),(20519,2108549,'23,December,2003','2003-12-23 00:00:00','    Yesterday more than made up for all the down time we have ever had here.  The call volume was higher than probably a full week of calls all packed into one.  It was crazy.  I\'m still trying to catch up from this past weekend.  Christmas is sneaking up on me way too fast.  I got a new camcorder from Dad last night.  I can\'t wait to start using it!  I\'m so excited!  Watch for movies to be posted soon!   '),(20520,2108549,'21,December,2003','2003-12-21 00:00:00','    Exhausted.  Today was the same as yesterday.   Off to bed.   '),(20521,2108549,'19,December,2003','2003-12-19 00:00:00','    I got my index files back up and running, for full details of what happened, go  urlLink here .    I saw the ortho surgeon Wednesday and it seems like I will eventually need to have another surgery.  My pinkie is *scissoring* over my ring finger and frequently gets stuck.  While its not causing any extra damage, it is painful when it gets stuck and quite embarassing.   I spent last night dealing with the landlord and all his lies.  I don\'t like the situation and am looking into what options I have available to move to.  I love the apartment, I hate moving, I love the setup of the place as it was promised, I hate the landlord.  Its all a mess.   On a funky note,  urlLink this  looks fun.   '),(20522,2108549,'18,December,2003','2003-12-18 00:00:00','    The plans for this weekend were: Potluck tonight, Freedom\'s vet appointment tomorrow, another holiday potluck tomorrow night, church Sunday and cable Sunday afternoon with a meeting thrown in there somewhere.   Everything has now changed. And that\'s ok.   '),(20523,2108549,'18,December,2003','2003-12-18 00:00:00','    It appears that my site has been hacked.  I have back up files however it may take a couple days to get back online with them.  I apologize for any inconvenience and thank you for your patience.   '),(20524,2108549,'17,December,2003','2003-12-17 00:00:00','    I have a Christmas tree, lights, an entertainment center and a glider rocker, all compliments of Dad. Had a nice talk with the new neighbors.  They had no idea how loud it was.  They are really nice and as usual I prejudged without giving them a chance.  I think things will be ok now on that front.  Or at least workable.  Its a nice feeling, even if it means I was wrong.   '),(20525,2108549,'16,December,2003','2003-12-16 00:00:00','    Thanks to my Dad, I have a Christmas tree!  Woohoo!  And the best part is no tree died in the process of me getting into the Christmas spirit.  Its potted so after the holiday I can plant it or re-pot it or something so I can use it again and again and again!  I can\'t wait to decorate it.   '),(20526,2108549,'15,December,2003','2003-12-15 00:00:00','    The bratdog has an eye, ear and skin infection.  No idea what is causing it for sure other than possibly still the thyroid problem.  More ointments and antibiotics.  New neighbors upstairs.  Not liking them all too much as I don\'t enjoy listening to them hang pictures above my bedroom at 10 at night.  We\'ll have to see how it goes.  The apartment has been a disappointment in so many ways.  I love the feel it has, I love the inside however we are going on almost 3 weeks with no cable and no sign of it.  I am tired of the landlords excuse and honestly the whole thing is taking on the appearance of a joke.  *sigh*   '),(20527,2108549,'15,December,2003','2003-12-15 00:00:00','    The Freedommutt is sick again.  Another infection is appears.  I am taking him in tonight after work to have it looked at.  Oh its been a whole month so I guess there is not much more I could ask for.  Craving a cigarette pretty badly today but..... hopefully it will pass.  Bout way past burnt out on sunflower seeds.  Lunch with Dad tomorrow, Ortho appointment Weds, blood work for the mutt\'s thyroid Saturday.  Aye vey.   '),(20528,2108549,'14,December,2003','2003-12-14 00:00:00','    I got a concerned comment based on another comment (now deleted).  I have a family like few people I know.  There is only one blood relative in this family, all the rest have been adopted by me or have adopted me themselves.  And I say adopted not in the legal term but in the term of the heart.    It took me 27 years but I now have a Dad. Oh my biological one is somewhere out there.  Actually last I knew he lives about 30 minutes from me.  But neither he nor I have the desire to be a part of each others lives.  He chose not to be in my life as a dad for the first 16 or 17 years and I have chosen not to be in his life on his terms the last 11.  And that\'s ok.  I don\'t think he is a bad person, nor even neccessarily a bad father to my half brother, honestly, I don\'t know the man.  There are no hard feelings on this end and I wish him the best, however he and I both do better I think out of each other\'s lives.  I have a new dad though.  One that I met some time ago and he has adopted me into his heart as I have adopted him.  Any comments left by Dad are from him.    I appreciate the concern and care especially from someone I have never met, however, please let me assure you, nothing inappropriate has or will happen on that front.  Thank you for the concern.   '),(20529,2108549,'13,December,2003','2003-12-13 00:00:00','    All the curtains are up.  Woohoo!  Sponsee is coming over in a bit and then I have to run to the grocery store.  Ironing left to do and then I should be set for another week of work. All in all a decent weekend.  I am so not ready for Christmas.   '),(20530,2108549,'13,December,2003','2003-12-13 00:00:00','    Still sick.  It still sucks.  I still have boxes. I am still procrastinating on the whole theory of Christmas.  Freedom still takes up 75% of the bed every night.  I still love the apartment and hate the landlord.  There is still no cable.    Still trying to sit still in it all.   '),(20531,2108549,'12,December,2003','2003-12-12 00:00:00','    I have the flu.  It feels like an elephant sitting on my chest.  Darn good thing I am quitting smoking anyway, this would force me to if I wasn\'t.    I am so not a good sick girl.   '),(20532,2108549,'11,December,2003','2003-12-11 00:00:00','    I\'m here, not much has changed.  I got new sheets and new towels yesterday, thanks to M.A. I am all unpacked out for now.  I just can\'t seem to find the motivation to keep unpacking.  Housewarming party will be mid January and I\'m sure I can hide most of what is left by then.  :)   '),(20533,2108549,'09,December,2003','2003-12-09 00:00:00','    Yay!  Finally something went right!  I have DSL up and running.  I got the NIC card installed and everything all by my lil ol\'self.  Woohoo!   '),(20534,2108549,'07,December,2003','2003-12-07 00:00:00','    Ok so I have phone at home and I was supposed to be getting cable installed today.  However, the contractor never came back and installed the cabling inside the house that is needed. On Saturday I was told by the landlord and the contractors that it would be done Sunday.  After not seeing any workmen Sunday, I called the contractor and he told me to call the landlord because some guy named Steve needed to be contacted about it.  So I called the landlord and after acting like he\'s all put out about it, he agrees to call him.  K, hi, not the property manager, this is your shit not mine buddy. I explained to him that if its not going to get done, I need to know so I can cancel the cable provider appointment.   I didn\'t hear anything back from him, nothing yesterday and nothing this morning.  So Comcast shows up and they aren\'t happy.  This is their third time coming to this place only to find out nothing is in place.  Twice for me, once for the third floor. So I call the landlord and he tells me that maybe I should wait until the cable is physically installed before making the appointment.  Gee, I\'m sorry, I made the mistake of thinking if you said it was taken care of, it was actually taken care of. And then he has the nerve to tell me he is just as frustrated as I am about the situation.  No buddy, you couldn\'t possibly be.  You have a television that gets more than 2 staticy black and white channels.  The 3 bedroom windows, still painted shut, his remedy:  I  should scrape the panes to fix it.  Ummmm yeah, no.  Not unless we are talking about deducting more from the rent.  I took 10 days for him to get someone out to fix the telephone and the leaking sink and dishwasher, god only knows how long before the horrid patch job in the bedrooms on the ceiling gets fixed and yet I am supposed to fix windows too?  And pay top dollar for this place?  Oh I so don\'t think so!  So still no cable.  And no idea when.    Then B called and said they are talking about me and my position at my homegroup because I didn\'t show up to the business meeting this past Sunday.  Well, I didn\'t go because I didn\'t know about it.  Had I known I definitely would have been there but no one contacted me via phone, email or any other way to tell me.  Including B, who came by the new apartment this past Wednesday.    So I give up.  It is apparently one of those days.   '),(20535,2108549,'07,December,2003','2003-12-07 00:00:00','    Sort of.  I must say though, it feels dang good to have the laundry done, work clothes ironed, the floors mopped and vacuumed, Christmas lights hung and boxes thrown away and the dishwasher running.  Especially when its all by 8 PM.  All the essentials are pretty much unpacked.  I still need to hang the curtains in my bedroom but its not too big of a deal.  I feel quite accomplished although very tired.  Its time to go snuggle up with the mutt and watch a movie before bed.   '),(20536,2108549,'06,December,2003','2003-12-06 00:00:00','    I\'m back online!  I have phone service and will hopefully have cable Tuesday and DSL Thursday.  Woohoo!  Curtains are up, boxes are slowing but surely getting unpacked.  Its becoming home.  Even to the Freedo mutt.   '),(20537,2108549,'05,December,2003','2003-12-05 00:00:00','    Wow!  There is nothing better than sleeping in a bed with brand new sheets, a brand new dust ruffle and a brand new duvet all at the same time!  And even better to get out of the shower into big fluffy new towels!    This weekend I am hanging curtains and curtain rods, getting an alarm installed in my truck and unpacking.  Uneventful to some but fun to me. Hopefully I\'ll even get a working phone line.  Doubtful but I haven\'t given up hope yet!   '),(20538,2108549,'04,December,2003','2003-12-04 00:00:00','    The phone contractor came last night.  I am still phone less.  We did have a good talk though and it sounds like it will be a few weeks before I get cable.  Something has to be torn up and replaced and reseated and installed and a dohickey is needed in someplace and well, yeah.  Its got a long ways to go.  Hopefully I can get my phone up by Saturday night though.  Trying to unpack at least one box a night so far has been working but I am running out of easy boxes to unpack!  How tempting to just take them out and donate them!  :)  I got to have a long talk last night with my upstairs neighbor and she is a sweetie.  With him being a chef and her being so nice, we\'ll have fun.  Especially since I have the Freedo mutt to seal the deal.   '),(20539,2108549,'03,December,2003','2003-12-03 00:00:00','    So I just found out I am going to be a flower girl.  A bit old you might say? Well, sort of but not really.  I met this group of people in November of 1997.  It was a very eclectic group, all of having little to nothing in common other than sobriety.  Or in my case,  the desire to get some sobriety.  I didn\'t know then but 3 of the people in that group would later save my life.  I ended up staying with 2 of them, living with a third for a bit until I took care of her house when she moved.  At one point I could picture each and every one of these people as if they were people I had met in person. We were and I suppose still are, the Richard Cranium family.  We even made a family tree. I was, of course, the baby of the family and at 21 was by far the youngest.  Well 2 of them are getting married. Its a marriage we all have been waiting for and I am so happy for them.  And they have asked me to be the flower girl.  *dances in a circle*    It will be interesting but I can guarantee one thing for sure...  I ain\'t wearin\' pink.   '),(20540,2108549,'03,December,2003','2003-12-03 00:00:00','    Last night was nice, took Freedom for a nice long walk with Ju.  Got home and made dinner.  Yes, I actually cooked.  And it was yummy.  At least to me!  I got all kinds of stuff from M.A. too!  Curtains and sheets and towels and silverware and and and...  :)   Muchly loved stuff.    I have been doing a lot of thinking as I am a bit burnt on the radio and well, I don\'t know.  That\'s it I guess.  I don\'t know what I don\'t know and I don\'t understand what I don\'t know.    Very tired.  Tonight is gonna be an early one.   '),(20541,2108549,'02,December,2003','2003-12-02 00:00:00','    So with having no phone and no tv, I get a lot done in the evenings.  I went home last night, let Freedom out and then headed to the grocery store.  For being single and not buying anything for the animals, I still don\'t get how I spent as much as I did.  However I have a ton of food in the house now and many many many meals I can fix.  Spoke with the landlord today and the General Contractor is supposed to come over Thursday to fix the short in the walls for the phone and finish off making the entire building cable ready.  Things are settling in, last night I got my electric blanket on the bed and between the featherbed, the electric blanket, the down comforter and 3 cats and a dog, I was toasty last night.  I love having the house cold and the bed all warm snuggly.  It does make it so hard to get out of bed in the morning though!  Things are good, I\'m just blue I guess.  I don\'t know how to describe.  Everything is wonderful with my life and I have nothing to want or need for and really not a thing to complain about, yet I still feel vaguely down.  I can\'t explain it.  Maybe its just the time of year it is.    Freedom and I went for a walk last night sort of late and before I realized it I was in an area largely populated with homeless people and I felt a twinge of fear at first.  No one had done anything to me or even given me a reason to be scared.  But Freedom and I kept on and got to say hi to people we had never met.  Next thing I knew I was back home and we were safe and sound.  I think I like it there.  Maybe it\'ll keep me humble.   '),(20542,2108549,'01,December,2003','2003-12-01 00:00:00','    Sometimes its the little things in life that bring me the most happiness  Waking up to a passed out pooch that is snuggled up next to me  Watching that same pooch sleep in the way other women watch their children  Having the luxury of playing my music, the tv and mp3\'s on the computer all at the same time  Knowing that my house will look the same when I return as it did when I left it in the morning  Going to bed knowing that the laundry is clean and ironed, boxes have been unpacked and I am ready for work  Waking up and not having to rush to get ready  Sitting on the front porch, no traffic or people sounds, talking to God  Waking up with a kitten *fixing* my hair for me  Putting together a bookcase all by myself.... and it not crashing down  Driving an Explorer in the rain with brand new tires  The knowledge that I have everything I need and then some   '),(20543,2108549,'01,December,2003','2003-12-01 00:00:00','    I\'m here.  Sort of.  I only have internet access at work right now.  No phone and no cable at the new apartment.  The landlord screwed some things up and it might be a few weeks before those things can be installed.  I got moved in on Saturday and things went well.  I spent most of yesterday unpacking boxes and moving things around and getting things put away. Amazing how much more gets done when I can\'t plop in front of the tv or stay online all day!   Freedom and the girls are adjusting, I\'m adjusting and while Saturday was hard, I am back to loving living alone.  The first night there was weird.  I think I forgot how to be alone like that but last night was great.  I have a long list of things that need to be fixed but I think everything is going to turn out ok.  I met my new neighbors and they seem really nice so I think that\'s going to be fun.  He\'s a chef and invited me to come to dinner with him and his girlfriend one night.  They seem really nice and we\'re together in the whole no communication problems with the landlord.    So I am here, I\'m just not real available right now.  If you have my cell phone and need me, call that.  I check the new phone message center now and then but not with any regularity.   '),(20544,2108549,'26,January,2004','2004-01-26 00:00:00','    I\'m back online and fried.  The new job is good.  I like Intel and they seem to like me.  Things in the new apartment are slowly getting unpacked but I am just overwhelmed and tired and on the verge of meltdown most days lately.  So I am alive, I\'m back and I\'ll update more later.   '),(20545,2108549,'19,January,2004','2004-01-19 00:00:00','    New job started today.  Not a good first day, just feeling very overwhelmed.  I moved over the weekend and have no internet access for about a week.  Things are just basically biting right now.  I can check my email at Julie\'s but between Saturday night and tonight, over 800 emails came in.  Will be back on when I get back up at home, until then no emails please.    Thanks!   '),(20546,2108549,'13,January,2004','2004-01-13 00:00:00','    You go through life expecting things to look different, feel different be different and then all of a sudden you realize you\'re different.  I think I finally understand, truly understand what the term \"Its all about perception\" means.   '),(20547,2108549,'09,January,2004','2004-01-09 00:00:00','    This is my friggin website.  Just because there are crazy people out there that have no life and refuse to get one does not mean I should have to change my website.  I am tired of other people dictating what will and will not happen in my life.  I worked way too hard on this site to keep it and maintain only to let some asshole run me off from it. If they want to be neurotic about it, so be it.  That\'s on them. I have a life and if the only thing they can get out of their is to peer into mine well, they are someone to be pitied.  That being said....  This journal is back.    So here\'s the update:  I got the job at Intel.  I start in a week or so.  I had given the current employer the standard 2 weeks and last night I got a call saying they were ending my contract early.  My boss never said a word when I left yesterday.  Needless to say, my feelings are hurt, I am disappointed in him and I am confused.  I have been told it was a financial decision and I did nothing wrong but then my question becomes, what was the harm in letting me finish out that one week.  Let me retain my dignity and take my things home and say good-bye to the people I worked with. It makes no sense but the notion of going to work for a real company again is ten times more appealing than it was yesterday morning.   I am excited about the opportunity and scared about the financial loss for this coming week.    I\'m moving again because nothing with the landlord has been easy.  In the 7 weeks I have been here, nothing has been repaired yet.  I still can\'t even open the windows here.  I found a place that is a little cheaper every month, a little smaller but right down the block and the 2nd floor renters here are moving to the same place.  They rented the 2nd floor there, I rented the 1st floor.  We do well as neighbors together and I think a friendship is starting up there.  I just wish there was a magic spell I could cast to move all the stuff here to the new place.  *sigh*  How I hate moving.  So that\'s the latest.  For the most part things are well, the job thing though sure has thrown me for a loop.   '),(20548,2108549,'23,February,2004','2004-02-23 00:00:00','    I\'m here.  Freedom has yet another skin infection.  Lots of money later, we have more antibiotics and still no reasoning as to why he is getting them.  Not much really to report.  Just very tired.   '),(20549,2108549,'18,February,2004','2004-02-18 00:00:00','    I get to play mom in the morning to a 11 (almost 12) year old.  I kinda like it.  Oh its not much more than getting him up and dropping him off at school but I like the morning interaction and I love the kid.  Je\'s middle son and I get along great and I think I am going to enjoy the one on one time with him.    Had lunch with Dad on Monday and he is going to subject himself to my cooking Thursday for dinner.  Work is going well, things are just status quo as of now.  :)   '),(20550,2108549,'14,February,2004','2004-02-14 00:00:00','    Today was a good day.  I might be Valentine-less but I have good friends and a wonderful surrogate Dad.  Today is also my 3 year sobriety birthday so it made the day more special.  I\'ve never had this much time before!  Ju had a little mini birthday party/ movie night for me.  I got HUGE a print of Matisse\'s   urlLink Icarus  (my favorite of the  Jazz  series) and beautiful planted purple flowers and my chip came from Je, it was originally her 3 year chip.  Dad sent me flowers this morning and they came with a little teddy bear and a box of chocolate.   Needless to say, I am feeling quite the spoiled princess tonight and am off to sleep as a very happy girl.  A year ago, I never though I would have the people in my life that I do now, and I never thought the people that wouldn\'t be in my life would leave.  Through the seemingly good and the seemingly bad, change always happens and its up to me whether I fight it or accept it.   '),(20551,2108549,'13,February,2004','2004-02-13 00:00:00','    So, today being Feb 13.... I have 3 years  of sobriety tomorrow! Its funny, I still have the old mantras in my head that people said, \"You\'ll never make it to 1 year\".  I did it.  I proved them wrong and I proved myself wrong.  One foot in front of the other, I have managed to survve this last year and not only grow from it, but stay mostly sane as well.  This weekend will be a big weekend of reflection for me.... a year ago I thought I couldn\'t stand to live even one more.  Now, I made it and lived to tell about it.   '),(20552,2108549,'10,February,2004','2004-02-10 00:00:00','    Things are going well.  Work is good, I am learning a lot and getting on par with everything.  Of course my competitiveness has kicked in full bore and I want to be the best on the team. ;o)  While the chances are slim of that happening this week, maybe someday it will!  The new apartment is still wonderful, Freedom and I are adjusting pretty well I think to all the new noises.   Life is just well, life right now.  Some good days, some bad days and a whole lot of ok days.   '),(20553,2108549,'03,February,2004','2004-02-03 00:00:00','    The new job is wonderful.  I got evaluated yesterday and I am on the phones alone.  I like it and I think I am going to like it even better once I have a better grip on what I am doing.  Other than a new job and a new apartment, things are status quo.  I love the new apartment, it safer and has a much homier feel to it.  I am quite happy with how things are shaping up in the work and home arenas.   '),(20554,2108549,'10,March,2004','2004-03-10 00:00:00','    Ok so its been a while since I have done much of an update.  Things are going well.   Work is good, I survived my first corporate virus crisis, thank you  urlLink Bagle  virus.  It was crazy and some days were better than others, but I survived.  I am making new friends at working, getting closer to my new friends at home, just in general, doing life.  I am starting to find out that I am likeable, maybe even to the point that people seek me out to spend time with me rather than simply tolerating my presence.  Its an odd thing to learn this late in life. To think that people think I am optimistic and perky is just very strange for me.  It seems the person I was is fading and the person I am becoming is shining.   '),(20555,2108549,'25,April,2004','2004-04-25 00:00:00','    Ok so here goes for the real update....  Work: Going quite well I have managed to get my stats high enough to place in the top 5 most weeks on my team. I wrote and had published 2 articles for the self help web site for our employees. And for as stressed out as I was at the beginning of the job, I am already bored. I am hoping to get involved in some projects to learn some new skills and if nothing else, to have some to put on my review.  Home: Ju and I are still hanging out quite a bit and I have been hanging with my upstairs neighbors also.  We have been working on a vegetable garden in Ju\'s back yard. Today was tilling and a little planting. I am dog tired and finally clean but it feels good to have stuff that is gonna grow this summer.  Freedom: Is doing well, all the tests have been inconclusive or good results so.... no idea what is wrong with him. Perhaps his is just high maintenance. :)   I think that is everything. Things are just crazy busy and I have decided its time to cut back a little on the things I am involved in. I feel too pulled in far too many directions and like I don\'t have enough time for anything or anyone so its time to cut back some.   '),(20556,2108549,'06,April,2004','2004-04-06 00:00:00','    The vet called today.  Freedom\'s biopsy results are inconclusive.  She left a message and I didn\'t get a chance to talk to her.  I don\'t know if this rules anything out.  The only thing I know is that he\'ll need to have another biopsy done from a fresh sample site.  They\'ll take the biopsy from an area that has the lesions and healthy tissue side by side.  More info when I get it.  Just keep him in your prayers.   '),(20557,2108549,'04,April,2004','2004-04-04 00:00:00','    Things have been status quo for the most part.  Yesterday we noticed that Freedom\'s pads were literally coming off of his feet.  Took him into the vet, she is running a biopsy on the tissue.  We should get the results Tuesday or Wednesday.  She thinks its either his allergies in really heightened form or Systemic Lupus.  Say prayers for the silly mutt please.   '),(20558,2108549,'24,May,2004','2004-05-24 00:00:00','    Sometimes....  I get overwhelmed by the amount of choices I have.  I get frustrated by my lack of drive, focus, determination, stamina.   I want to give up, hide under the covers, run away from you all.  I want to stop feeling like I am always *on*, unable to break, not allowed to be down.  I think of living as a chore, a way to kill time, a form of entertainment for God.  I think that death is gift, a vacation from the montony, a freedom from feeling.  I can understand why you run, silently screaming, so far far away from me.   I want to fade away, quietly, invisibly   '),(20559,2108549,'16,May,2004','2004-05-16 00:00:00','    There just isn\'t much to update on.  Things are calm.  Things are good.  Work is good, I\'m bored already but its still good.  Freedom is good, the 2 mile walks every morning are helping take some of the weight off of him (too bad the same can\'t be said for me).  I\'m working on a few different freelance web design projects and as much as I try, the flow of creativity seems a little stifled.  My mom is going to Utah for a month next week to help my aunt after she has surgery.  The garden is growing very well.  This summer will be fun with all the homegrown veggies.  I honestly, just like knowing that something I bought in a little packet is growing and sprouting because of the work I put into it.    In other words....  My life is boringly wonderful.  :)   '),(20560,2108549,'11,May,2004','2004-05-11 00:00:00','    I got into my truck today to leave for work.  Started it up and heard POP and then smoke came out.   Its at the shop. It had issues a few months ago and I had to have the transmission modulator, hoses and the radiator replaced.  It is doing all the same stuff as it did then.  Needless to say, I am not thrilled with the repair shop but because they did the original work, thats where my truck is.    Then as I am waiting for mom to show up, I stood up from the brick ledge I was sitting on and hit my head on their sign.    Yeah, not going to work today.  Gonna deal with the truck and keep praying that everything that is wrong they don\'t charge me for.   On the flip side, I have an awesome boss and he was WAY cool about me not coming in.   '),(20561,2108549,'04,May,2004','2004-05-04 00:00:00','    Not much is new....  work is good.  I am doing well within my department and I still enjoy what I do.  How things are so different from my first call center experience. Freedom is still doing good and life is just in general well, boringly wonderful. I have settled back into the corporate world without issue.   My latest review reads:   Shannon,  What can I say? Awesome performance! Your AHT  (average handle time)  is really low, which would normally worry me because it might mean that you\'re not giving the customer enough service to be thorough. However, I see that your ACW  (after call work)  time is also REALLY low (it\'s included in AHT), which means that you\'re talking and typing and this contributes to your low AHT. Also, in call monitoring, I haven\'t noticed you giving customers poor, or rushed service. Just be sure not to over-focus on indicators, so that you don\'t burn yourself out.  I\'m so glad you\'re on the team!    Hope everyone is well!   '),(20562,2108549,'29,June,2004','2004-06-29 00:00:00','    Things are mostly good.  Work is good, home is good.  Missing MA a lot, he moved and while I know the intent was there, he hasn\'t kept in touch.    Mid-year reviews at work are going on, its hard to believe I have been here almost 6 months.  Seems like forever and only a short time all at the same time.  Praying a raise comes my way.  I love my mutt but he is too expensive.   '),(20563,2108549,'22,June,2004','2004-06-22 00:00:00','    So since the whole break up lat year, I have been hesitant to really want to get involved with anyone.  My heart had been broken and well, my and Freedom do just fine on our own.  And then I met a new guy.  I like him, quite a bit actually.  Lately I have been missing the relationship I had.  Not the man I had it with per se, but the relationship itself. I miss loving someone in that *awe inspiring, tummy flutters, light up my life* kinda way.  Not to mention I miss being loved that way.  aside from all the BS and deciet, that relationship was great in that area. If only for that reason alone I cannot regret that relaitonship.  It taught me a lot about being loved and loving someone else.  I don\'t know what\'s gonna happen or even anything will happen with this guy.  I\'m just glad that I finally feel ready to move on from the last one.   '),(20564,2108549,'13,June,2004','2004-06-13 00:00:00','    Back then... So when I was around 12 or 13 I started babysitting for the Ab family, M and K (the kids age 3 and 4), S and Ka, their parents were in Amway with my mom.  It was an easy $10 a week to baby sit for a few hours on Friday nights.  I did this until I was around 18 or so.  Every Friday night I would get there around 7, play games and do origami and crafts with the kids til they went to bed and I watched tv.  S and Ka were the parents I wanted.  They listened when I talked, they were interested and they wanted their kids to have every advantage in life.  For high school graduation, S built my first computer.  A 386, which at the time was the best of the best.  I broke that thing and had it fixed so many times I lost count.  That\'s what got me started in computers.  Bought a DOS book and thought things like fdisk and dblspace would be fun to try.  I got into the Grateful Dead, they got less active with Amway, the babysitting became less and less and I moved and we lost touch.   Presently... When I started at Intel I remembered S had worked there and looked him up.  He was still there so I shot him a \"Hi, Remember me...\" e-mail.  I never got a response and never really thought about it again.  Thursday I was walking into work and saw someone that looked somewhat familiar although I couldn\'t place the face exactly.  Then he started talking to someone else and I recognized the voice.  It was S. I said his name quietly enough that if it wasn\'t him, I wouldn\'t be humiliated.  He turned around and I did the \"Hi, \'member me\".  We talked and it turned out M was graduating that night from High School.  He invited me to the party yesterday as a *surprise guest*.  We had lunch Friday and got all caught up and made the plans for yesterday.  I showed up yesterday, completely nervous, second guessing if the kids would even remember me.  M opened the door, no recognition at all.  I asked for his dad and he went and got him.  As I was walking out in the backyard, Ka saw me and instantly recognized me, BIG hugs all around.  Then K saw me and did the same.  It was so good to be there yesterday.  Ka and the kids reminded me how much a part of my life they were and I theirs.  They have a third child, Pa, who is 7 that I got to meet. I was reminded of what an integral part of my upbringing they were.  It was Ka I told about my first kiss,  I confided in about the abusive boyfriend and the kids that kept me sane through it all.  We promised to keep in touch and I have no doubt that it is a promise that will be kept.  It was total God thing in each and every way.    As for the lake..... Well, J from work (not my other Ju) and her boyfriend R, invited me to the lake today to hang out and enjoy the sun.  Freedom and I got there and we were all enjoying his enthusiasm, and being that they are also both sober, the soda and water.  There were 2 single guys next to us with jet skis and then on the other side, a group of 18 or 19 year old.  About 4 guys and 3 or 4 girls and their bulldog.  Well yeah, there is a leash law in place but neither Freedom nor their dog was on a leash.  Freedom goes out to the water and back to me and that\'s about it.  He listens and comes when called and has never gone through anyone else\'s area.  Well, I wish I could have said the same for the bulldog.  He came tromping through our stuff, leaving water and sand on our towels and we ignored him.  I must say though when he lifted his leg on R\'s truck, we were all irritated and angry.  R told the guys to get their dog out of our area and back on a leash.  They did... for a short time.  Then the dog came back, this time to lift his dog on R\'s trailer.  So R picked up a handful of pebbles, tiny stones really and threw them near the dog to try to get him to leave.  Well the jerks next to us start yelling about Rick throwing things at their dog and this asshole picks up a rock the size of my hand and throws it at Freedom.  I yelled at him to knock it the f--- off and he picked up a larger one and did it again.  I lost it at that point, I jumped up and ran over without even stopping to think.  OK, I am a girly girl.  I don\'t know how to fight, I don\'t even know how to throw a punch other than what me and Billy Blanks practice with Tae Bo.  But it was Freedom and there was no way I could let anything happen to him.  The kid shoves me, knocking himself over in the process, R comes over to grab me and go after the kid and then the single guys came running over.  They were yelling at the kids, R was yelling at them, J had Freedom by the collar because he was coming to protect momma.  It ended about there other than smart remarks, dirty looks and the like.  I\'m not really proud of my reaction, yet... its Freedom and my reaction was far from thought out or planned.  It was a knee jerk reaction.  One that if it ever happens again, I hope I am better prepared for!   Work...  Is going awesome. I got a review last week and all are pleased with me, my performance and my skills.  I pretty much have a world of doors open and I just need to figure out which ones I want to go through.  I got a lot of compliments and it was nice to be able to talk about some of the issues we deal with just in the working environment, not necessarily about the work itself.  I think that if I stay on the track I am on right now, I have a very good future there.  And best of all, tuition reimbursement kicks in next month so I\'m goin back to school.  For what yet, I dunno but somethin!  :)   '),(20565,2108549,'25,July,2004','2004-07-25 00:00:00','    Good weekend if not a bit quiet.  Hung around the house, did laundry, played with the mutt, the usual.  Went to a birthday party yesterday for 2 year old twin boys.  Lots of laughs and kisses there!  Worked in the garden a bit and just in general kept to myself. My feelings have been hurt by a couple people and I know when it is more than one person, there is something wrong with me.  Regardless, they are still hurt and I am licking my wounds and trying to analyze myself.    Yeah... Shannon wrapped up in her own head is a muchly scary thought.   '),(20566,2108549,'23,July,2004','2004-07-23 00:00:00','    My new phrase of the week. Hey, its better than Oy Vey.  Well, let\'s see here.&nbsp; Work has been good.&nbsp; Boring, but good I suppose. I wrote another script for our department, I\'m ranked 10th out of the department and my peers like me.&nbsp; Imagine that.&nbsp; And what\'s more, we have a person on our team that is well, perhaps lacking in social skills.&nbsp; Or maybe its a matter of not wanting them.&nbsp; Regardless, when they moved him behind me, I decided I would make friends with him specifically because I did not like him.&nbsp; And we made friends and I found out he is a cool person.&nbsp; Well, his social skills or lack thereof, have gotten him into some manure with management.&nbsp; Oh how I remember the days.&nbsp; The conferences and meetings being told&nbsp;I was inappropriate, out of line, unacceptable, etc. etc. etc.&nbsp; It wasn\'t so long ago that I had an awesome manager that told me these things.&nbsp; I remember too the feeling that even if I did improve, all anyone would see was who I used to be.&nbsp; And that was the case and may be the case for him here,&nbsp; thankfully I don\'t know.&nbsp; My conferences have been nothing short of praise for my attitude, my willingness to help and my overall *perkiness*.&nbsp; Gag me.&nbsp; I am   NOT   perky. Well, maybe a little I guess.&nbsp; So the problem child on our team has now asked me for coaching on when he is out of line, sarcastic or inappropriate.&nbsp; Me? Really?&nbsp; Wow, a few years ago, no one asked me for anything like that.&nbsp;   Its amazing what a little recovery, a lot of God and fresh chance will do for someone like me.&nbsp; I am actually well liked here, not out of fear but because... they really like me.   So bizarre how things change.   '),(20567,2108549,'17,July,2004','2004-07-17 00:00:00','    The last few weeks have been so busy and yet so just, well, ordinary.&nbsp; It\'s been nice.&nbsp; Work is good, I am rated 10th in our department (out of 62) for my stats (call time, volume, resolution, etc) and I am rated first on my team (14 of us).&nbsp; Not bad for only being there 6 months!&nbsp; Mid-year reviews are coming up.&nbsp; I\'ve been told by my supervisor that I have nothing to worry yet of course I am still anxious to hear what he has to say. And while I am not first in our department, of course my head tells me I should be.&nbsp; One of the things about working in a call center is that so much of my performance is based on concrete numbers.&nbsp; Number that while I may not agree with the formula for calculation, because everyone is based on the same formula, there is no denying the good or bad in them.&nbsp; Numbers of course aren\'t the only measurement of success there but they do play a hefty part.&nbsp;  &nbsp;  My reputation at work is such a far cry from what it used to be at Apple.&nbsp; Same work, same environment (in a sense) so its easy to compare the two.&nbsp; Apple was my first corporate experience.&nbsp; I must say I was pretty hard headed about learning how to work and play nicely in that type of a setting.&nbsp; I had an awesome supervisor though and when I got out of line (which was often) she told me.&nbsp; As much as I hated to hear it, she was always right and because of those lessons, I am where I am today.&nbsp; The supervisor I have now is a gift from God.&nbsp; Literally.&nbsp; He used to be a pastor so its nothing for me to ask him to pray for me if I am having a bad day.&nbsp; He *gets* that a long with a ton of other stuff.&nbsp; He can read people and their reactions and know where they are coming from.&nbsp; He understand my need for perfection even while disagreeing that I should be perfect.&nbsp; He understand that all of us are human, subject to all the stuff humans go through.&nbsp; Its been a great experience there.&nbsp; And I seem to be on track for a good career there should I keep my current actions.  &nbsp;  I have requested information to go back to school.&nbsp; My company will pay for it, I need it and I want it.&nbsp; Of course, I am scared to death to think of it but I think it will be good.  &nbsp;  Freedom is still Freedom.&nbsp; Ever wondermous and my best friend.&nbsp; His allergies are in full swing but other than that, he\'s healthy again.&nbsp; He\'s losing the weight we need to get off of him and each pound we lose, I get another couple minutes added on to his life.  &nbsp;  Home is good.&nbsp; Our garden is flourishing.&nbsp; We have all kinds of tomatoes (at least 3 different types), squash, watermelon, cantaloupe, honeydew, corn, Chinese eggplant, green beans and bell peppers.&nbsp; Some we started from starter plants, some we planted from seeds.&nbsp; Its just amazing though.&nbsp; Its so cool to go out to the garden and harvest stuff and then eat it knowing that it was something we grew.  &nbsp;  So.... work is good, Freedom is good, home is good.&nbsp; Not a whole lot to complain about these days.  &nbsp;  And that feel great.   '),(20568,2108549,'15,July,2004','2004-07-15 00:00:00','    It funny how you can be going through life and then WHAM! someone from your past comes back to the present and everything gets all topsy turvy.  Anything emotional relating to them all comes floating to the surface.  Somewhat like an Exxon oil spill in the heart of the Alaskan waters.  Oily, dirty, smelly and foul.  Nothing good can survive in water like that.   I have lots of people in my life that I don\'t keep in regular contact with but are still in my heart.  Some going back to before I was a teenager.  The relationships survive, without a problem, us not talking for months and months.  When we get together, its like no time has past, quite comfy with the relationship.  I have others that for whatever reason (usually bad) are no longer in my life and contact from them is unexpected, scary and foreign.  I don\'t  like  being caught off guard.  I don\'t  like  being surprised.  I don\'t  like  the unknown.  And I  despise  being conflicted over how I feel about someone as far as like vs. dislike. I like knowing what is gonna happen next, I like consistency, I like stability.  Mostly though I like knowing where people stand in my heart at all times.   If someone rips something away from me, be it a dream, a fantasy, something material or nonmaterial... I fight to keep it.  I fight until my knuckles are bloody from hanging on to that thing for so long.  And then once I am worn out, tired of fighting, I accept the loss.  Usually not gracefully, usually not without a whole lotta tears.  But I do get to the acceptance part.  I can\'t unaccept that defeat just because someone wants to return what they took away.  At that point I all the fear of losing it again come flying back up.  Flying up like a blender with the top off on full speed blending tomatoes juice and milk.  It\'s just not pretty and its truly gross and way too messy.   I\'m at the point of fear and hurt and anger and sadness.  What could have been, never really was and what I lost, I never really had.  Wouldn\'t taking it back then mean that I was backing to holding on to something that wasn\'t mine and wasn\'t real?  Sort of like trying to hold on to air.  I\'ve looked over this ledge before, I think I need more than just air to stand on.   '),(20569,2108549,'13,July,2004','2004-07-13 00:00:00','    A reason to boycott the Olympics already   urlLink GREEKS \'TO GAS 15,000 DOGS BEFORE OLYMPICS\'   '),(20570,2108549,'04,July,2004','2004-07-04 00:00:00','    Got up went to church, came home, took a nap. Naps alone make it a good day.   I invited my mom to the church picnic/ 4th celebration last week. Dunno why I did but I think it was a God thing. We had a great time. I was in charge of the Bounce House *wicked evil grin* for the first couple hours, managed to go on *big kid* Slip-n-Slide twice. I quit after hitting my face on the ground. Ouch! We had a BBQ, great worship music, Ju and her friend La and her kids and well, basically a bunch of kids I know and love were there. Had a great time with everyone. TONS of fireworks at the end.   Basically a mmmmmmmmm yes, good day. :)  Happy 4th Everyone!   '),(20571,2108549,'04,August,2004','2004-08-04 00:00:00','    I have a case of the \"F--- it\" syndrome.  Lyrics that haven\'t appealed to me in quite some time are appealing.  My birthday is coming up and while I can usually rattle off a list of things I want, this year I had a hard time figuring out what to put on an Amazon wishlist.  I feel bad, my mom has been asking me for weeks what she can get me, but at the same time *f--- it*. Its just a birthday. Another year gone by to signal the fact that I have down little to nothing with my life. I dunno.  I care. Yet, I don\'t.  Screw it.  And screw them. I would say I am apathetic towards my life but I\'m not.  Its not a case of apathy, more like a case of ambivalence.  I\'m tired of it.  I want to run far far away where no one knows me, the problem though is not my surroundings but rather the problem is me. I burnt out on being me.  I just don\'t wanna be me anymore.  I would be perfectly content to not exist tomorrow.  I know this isn\'t a healthy feeling.  So I beat myself up for it.  Then I beat myself up for beating myself up. Which leads to a desperation of wanting to not be me.  So you see, my life is not the problem.  I am.   Its difficult to live in the skin of someone you don\'t like.  I try to be perfect and good.  The perfect friend, coworker, employee, daughter, pet owner... it just doesn\'t work. Yet, knowing the issue doesn\'t seem to fix the issue.  Talking about it grants me frustration from friends, rather than understanding.  It grants me weird looks, wide berth and pats on the head with the words *That\'s normal*.  F--- it.  I don\'t want to be normal.  I don\'t want to be me. I don\'t even like me.  Work is good.  I have a new project administering the log in ID\'s for our call consultants.  I am assisting in developing the training for a website our department will use and I am on the  Great Place to Work  team. They like me.  I do well there.  The structure, the distraction, the focus.  What do they see that I don\'t?  I just don\'t get it.  So I dunno.  Anyone have a brain they want to trade?   '),(20572,2108549,'02,August,2004','2004-08-02 00:00:00','    Saturday morning I went and picked up K.  (I\'ve known K since she was 8 months old, she\'s now 15.  She came back to my place, watched Uptown Girls while I picked up the apartment and then she, Ju and Ju\'s nephew E (age 10) and I walked down to Old Sac, ate caramel apples (a first for the kids) and wandered around.  Came back, BBQed and then the kids came back to my place to watch Dickie Roberts.  Ok, that movie was so funny.  I normally don\'t care for David Spade but it was way funny.  Walked E back to Ju\'s around midnight and then crashed.  Got up this morning around 8, made breakfast for K, and then we headed to church.  They were doing the water baptism today at Goethe park and it was way cool. While at church, we picked up V (age 19), the daughter of one of Ju\'s friends and she and I got baptized.  Her reaffirmation, the first time for me.  After that we ate lunch and headed back here.  K had to be home by 4:30 so V, K and I drove out to Antelope, dropped K off and then V came back with me and we watched In America.  Weird but good movie.  Very strange though. I dropped V off around 8:30 and I am headed to bed shortly.    I\'m beat, Freedom is worn out.  Off to bed we go.   '),(20573,4292654,'22,August,2004','2004-08-22 00:00:00','    There I was....in bed at 2 am, starting a new book....The Dork of Cork....a story if an Irish midget--reminded me of other situations in life where I find myself in odd situations and wonder how it is I always tend to manage it...is it normal?  I cant imagine im terribly unique that way...example-- if I am talking to someone really bright and well spoken....in my head Im wishing my mom had breastfed me or taken pre-natal vitamins so I could keep up with these thinking sort of people....my mind goes off on these tangents, thus im not paying close enough attention to even keep up...had i paid attention from the beginning maybe I would feel less *duh*. Since my bothers suicide in 2000 my mind seems to have left me... At 1st I couldnt read a book....or a pamphlet for that matter....or think things through w/o feeling overwhelmed....I can read again but im still scattered and flailing...I forget things....I think maybe I should try some weirdo chinese medicine to help my memory...haha yeah cuz THAT\'LL do the trick.... Its bizarre that a person can leave such a hole in a person...Im all for prolonged grieving etc whatever it takes... but I can honestly say w/o an ounce of drama that I will never feel the same again without him around.  Being touched by another human that way makes me feel very vulnerable of course.....what if my daughter dies....?  My sister....my last sibling.....the only one who can make me an aunty...??  Death-while perfectly natural-is a mindfuck...I havent talked about my brother in a while....Ive been keeping him in my head...perhaps why I have this free flowing wave right now to talk about him. His name was Glenn....he was funny....caring in his own hardass/sensitive way.....generous to a fault....a beautiful spirit...To me he was always this little boy I used to play with....the boy who thought Yoda was a really alien in our backyard, living in a woodpile....the boy who used to torment me when we had to share a room....the boy I would beat on...the boy I was taken from during yet another divorce.... We maintained a very close relationship despite distance and time...I truly loved him and my sister.....of course often you dont know just how much a person means to you until theyre gone....and wont ever be back again.   Sometimes knowing just how sad he was inside to take his own life rips and shreds my heart b/c I wish I had the foresight to do more....or turn back time and make him feel more loved...make him feel less alone in the universe. its a sick sick thing to know anyone you love suffers, physically or inside their own head of loneliness, insecurity, dread and sadness... Fuck...I wonder how many things i spelled wrong?  ha....no spell check dammit.... If I reread this I may vomit so I will take my chances....despite the lack of breastmilk and being over vaccinated I am literate....so plz ignore the typos....not that I expect any spectators......I dont really know how blogs work. Im winging it. til we meet again.....   '),(20574,4292654,'20,August,2004','2004-08-20 00:00:00','    I must go to work like the little underpaid pion I am....More to come...   '),(20575,4200075,'22,August,2004','2004-08-22 00:00:00','   today wake up wif sleepy eyes... wake up about 3 plus... so shoik as among the whole week.. last night was the only nite i can sleep my whole life back.. as ytd work full shift, today when i wake up still very tired sia.. haiz.. nvm.. when i wake up, i go out liao.. long time no update bankbook so today go update.. den go cmp QB house for a haircut.. so sianz.. den start work liao.. all the way crowd in until 9 sia... hate coupon promotion!!! still muz wait until 7 sep... haiz.. den after that me count cash liao when clearing cash me short $7... wah no money liao.. sianz.. nvm la not 1st time liao.. den after that zhong you sent me home.. on the way saw one lady.. everytime zhong you sent me home will see that lady one.. den we stop and see.. that lady will hire every taxi, open the door.. dunno talk wad to the taxi driver.. den close the door.. siao one.. lolz.. haiz so tired until nth to write liao... tml still have sch!! so sianz!!! haiz.. got to slp early liaoz.. or else tml cannot wake up liaoz.. nitez!'),(20576,4200075,'21,August,2004','2004-08-21 00:00:00','   today up, nv eat breakfast, go work liao.. today me replace celestine from 11-5 den me work 5-10.. poor thing rite? lolz.. 11am reach liao alot of pple liaoz.. den me work base... jam up base.. den jo help me.. den she ask me go help pack home d.. den rotate the cashier go drink water.. den me go pack home d again.. den go make whip potato.. den go base.. den rotate cashier go drink water again.. cannot even sit down and rest for 20 secs.. so b-u-s-y!! arrr...!! den me about 4 go cashier liao.. den about 5 go dress up as chicky and play wif the kids.. nabei wear the costume damn hot sia.. when i remove the costume hoh.. my clothings is all wet.. even my underwear =x.. den me very tired liaoz.. work cashier wif my half life.. den after work still hav to do closing.. me so tired.. den count cash.. den rot there wait for zhong you to sent me home.. lolz.. wait until about 1am de go home sia.. haiz so tired... come home write blog... den wanna sleep liaoz.. or else my bones is going to crack...!!! lolz... gonna sleep liaoz.. nite!'),(20577,4200075,'20,August,2004','2004-08-20 00:00:00','   today same loh.. go sch.. rot in sch.. den held up a meeting in class after sch.. den go back home.. today reach home quite early.. about 1.30pm reach home liaoz.. den me watch tv a while.. until 2.30pm.. den go cmp kfc... me today hav the star staff examination.. very scare.. coz if me never pass den create history in cmp kfc liaoz.. (1st person to fail 1 star staff examination) anli got tell me, although if i forget anything, juz smile smile and act blur.. lucky during the examination i got smile.. or else fail liaoz.. base almost fail sia.. lolz. lucky loh.. den after the examination, me work liaoz.. me work as cashier loh.. den got the edwin attach to me loh.. den he very blur... den me bek chek liaoz.. me dulan him until he knw how to pack... me den put him by himself to stand pos.. he can!! of coz la i teach one wad.. he only pack very slow only.. he put him alone.. den me go help top up all the things... lolz.. den after work.. zhong you sent me home again.. thankz! haha.. haiz me so tired.. feel like sleeping.. ok le write until here le ba.. nite!'),(20578,4200075,'19,August,2004','2004-08-19 00:00:00','   today same loh.. go sch.. come back.. after school me go back home loh.. den tell mummy i got work.. but den is go out to celebrate si hui\'s bday.. meet si hui, anli, guowei, daniel, celestine, yuling, derrick, jiehwee, willie, yishan, adrain and dennis at 4.. den go orchard.. buy movie tickets, den go makan at lucky plaza.. very funny loh.. we sit in one row.. den we pass the dishes all around.. den tok tok.. joke joke.. lolz.. after that go watch movie, the Twins Effect II.. before the movie, we buy drinks from kfc.. (becoz can free refill) lolz..den we go in watch movie liaoz.. the movie not so nice... but very funny.. dennis sit beside me, den kana me beaten until siao.. hahaha.. after the movie, we go kfc again to refill the drinks.. wahhaha.. den we go walk walk.. rot rot.. until about 10.. den i go seperated ways as me need to go back cmp kfc..  i reach cmp kfc about 10.30pm.. amazing all their closing done liao.. den me rot there.. until about 12 zhong you sent me go home.. so sianz.. wanna sleep liaoz..'),(20579,4200075,'18,August,2004','2004-08-18 00:00:00','   today is another sian day loh.. during morning assembly anbu say mdm tan not here... so we will have 3 periods of free period liaoz.. yeahh.. lolz.. after morning assembly is dnt.. sianz ar.. got 2 tests.. nvm loh.. after that is english and recess.. straight after recess is MT liaoz... woot! no nid to climb stairs we stay in the canteen do our self studies... den perlyn teo the all wanna discuss about the \'Be Yourself Day\' and after that i dulan liaoz i scold they all.. they seems to be asking very stupid questions.. so pple i now confirm that the theme of \'Be Yourself Day\' is  \'Funky\'..  den after school me go media.. den media haven end yet  han hui ask me to go home.. ok loh i go home.. me dun like to be in media one.. haiz.. nth to add on liaoz me wanna go nap liaoz.. byez!'),(20580,4200075,'17,August,2004','2004-08-17 00:00:00','   yeah...!! today is tuesday man!! finally i can take back my belongings liao.. today de lessons are nice man!! 1st 2 periods very sianz but after that 2 periods we song liao.. Mother Tongue, CME, PW, and PC.. woot song man..  today pw samuel\'s group present their project.. nice man!! lolz.. den as we cannot finish our lessons on time.. so miss neo took our PC periods to finish.. mrs koh den come to the comp lab and she conducted serveral surveys wif some of us.. den it\'s my turn!! as there will be a \'Be Yourself Day\' on Youth day.. miss goh wan us to dress up but according to a similar theme.. so.. i conducted a votes for the theme and the final theme is \'Funky\'..  Friends.. if u hav any suggestions pls voice out to me so that i can be able to make alternative arrangement.. the theme can be change.. pls voice out to me asap as the deadline is on friday.. we are also required to donate for \"The President\'s Challenge\" in order to facilite the event..  so.. the bell gone and we all need to do a survey again.. we go to the respective comp lab and we finish the survey.. me straight after the survey me go police academy to retreive my belongings.. song!! me take bus there and after which i took cab back home.. with my ez-link and my keys things are now easier to go along..  me today got training at 6.. den me and sihui go theory test again.. lucky the actual test date is postpone.. or else me sure fail one.. the theory test so difficult me forget alot of things sia.. haiz me muz go and study again liao or else sure fail one.. after work about 11 that time jun jie go and take bread from breadtalk back.. wahkao one black garbage bag full or breads..!!! siao! zhong you after that sent me home.. although i take only one Large TA bag of bread go home but when i squeze it into the refrigalator, me found out that it took so much space!! haha.. sianz liao.. wanna sleep liao.. gtg le.. take care friends!!'),(20581,4200075,'16,August,2004','2004-08-16 00:00:00','   today me go back to school sia... 2 weeks never go back... haiz... today go school wif black shoes.. lucky got ask bao sheng help me to bring one pair.. (hey! thank you bao sheng) me faster change den go morning assembly... today me very lost in sch sia.. all the lessons i miss out so much.. far far back behind the syllabus.. haiz..   me after sch rush to kfc liao.. den take wrong bus...!! actually wanna take 163/371/372 go one but me take bus 82!! arrrhhh.... me alighted at the bus stop near 299 den me walk all the way to compass point as me no more money to board another bus liao... ;( haiz.. when i reach compass point is already 3.40pm liao.. haiz.. kfc that time alot of pple no place to sit so me change into uniform and to in read the star 2000... actually wanna read up and memorise the content of star 2000 because the test is tml... but... when i finish reading it, anli come and she tell me that the test is postpone to friday!! arrrrr...!! me wasted my effort going down early sia... nvm loh... den me today same same work until 10 loh.. den jun jie and his family come liao.. me talk rot wif him... after that he go breadtalk take bread for us... so good of him...   me den pei daniel go wait for bus 27 den go home.. go home already 1++ liao.. bathed liao already 2++ le.. sianz ar so tired.. hiya me go sleep liaoz lolz...'),(20582,4200075,'15,August,2004','2004-08-15 00:00:00','   '),(20583,4200075,'14,August,2004','2004-08-14 00:00:00','   today very sian leh... wake up at 7am wanting to go npcc.. but... mother say stay at home.. but after that i sneak out and go to work... me actually work 6 one... but i reach there 3 plus... den jo ask me to help her work 4... ok loh me very steady one ma... lolz.. wah the customer queue today is as long as a dragon sia.. kaoeh the customer keep coming in sia non stop sia lucky got my papa (zhongyou) and jacq at base so products no jam... den hoh we is take order until no strength sia.. when wendy rotate us go drink water we drink water+ sit down in store room rest.. but me not so bad la me service champion wad... (lolz) jkjk.. me drink water liaoz straight away come out liao.. me today make a bet wif shawn loh we compare sakes today see whose sales higher... lolz... no need to say one wad me sure will win one coz i pos 2 he pos 1.. pos 1 is so called the customer service contiment station... can say almost every order got pple come take chillie one... that will slow down the service timing so i have the benefit of the doubt.. a bet is a bet.. me got confident i will win shawn one so as a service champion (me fang pee)... me of coz will fasten up my service timing one wad... that is why my sales hit $1000!! (applause)... lolz..  today de 10.30 cashier is derrick... he hoh deploy me do hot/cold well!! kaoeh me 10.00pm liaoz still at there taking order leh... but nvm friends ma me help him do closing because it is saturday ma... haha.. after work my papa(zhongyou) sent me home heez.. den daniel, guo wei and willie go supper loh.. too bad i cannot thon if i can i sure go one..  when i reach on i switch on comp sia so many of my collegues online sia (yishan, anli, dennis, sihui).. me chat wif them all the way until 3am den we go slp..  wad a hard day... LOL'),(20584,4200075,'13,August,2004','2004-08-13 00:00:00','   friday 13.. bu ji li... suay sia the hq pple call me and tell me that me can only retreieve my stuffs tuesdays onwards.. KNN me have to go buy sch shoes or else monday no sch liaoz.. nabei.. after that call me damn sianz.. so i early go down cmp kfc (4.00pm) help out..nabei sia enter kfc see crowd sianz liaoz.. go in store room saw all the stocks stack until no space to put sia..   wah today wait for sihui come for training can wait until sleep one.. she reach at 6.20pm sia.. kao eh.. nvm la... den later 1 star staff training that time 1st is theory test!! siao la me never even go study how to pass theroy test? me only got 50% for theory and theory muz at least get 90% den is pass one leh.. even sihui also got 87.5%... tyco la... lolz.. jkjk.. somemore POS stands for point of sales not product of service!! LOL!!  den after that is pratical.. i love pratical sia! it is so fun!! first go in do procedure fries.. den procedure zinger... den mayo.. kaoeh juz now top up mayo that time when me throw the empty mayo pack that time fall down!! kaoeh kings shoes leh should me safety boots ma... still fall down... haiz... 4get it la.. we top up mayo muz use straw ma... den me see si hui top up that time use the straw pok the air bubble pop until very bek chek... wahahahahaha... den after which me go walk walk wif sihui haha so sian.. den we two talk talk.. den walk slow slow back.. help out them.. den saw yuling come den after that kai ying also come.. den after work we talk talk until si bei song.. they wanna go punggol end ton but too bad me cannot go... sad ar...   friday 13 really bu ji li sia alot of pple kana bad luck.. me kana fall down den get back belongings only after tuesday onward.. alot of my friends also kana bad luck... my best wishes to u all!! smile always!   '),(20585,4200075,'12,August,2004','2004-08-12 00:00:00','   grrrr...!! siansation is the word!! having stayed at home for 24 hours.. and i spent 20 hours sleeping!! arrr..!! woke up 12.20pm today.. spend 5 mins brushing my teeth and 5 minutes switching on my tv.. =X lolz.. kidding la.. watch tv from 12.30pm-1.30pm.. den its\'s computer time.. haiz.. can rot at home liao.. my wallet is at ubin.. can\'t really go anywhere... without my keys, i can\'t even get out from my house!! siao liaoz.. now i\'m alone at home.. bored to death la.. haiz.. dunno wad to do sia.. play comp write blog... if nobody gonna call me or tok to me.. i\'m going to rot to death liaoz... ARRRRR......!!!!'),(20586,4200075,'11,August,2004','2004-08-11 00:00:00','   wah kao eh.... juz come back from the national camp as casualty. haiz.. the national camp is quite fun as i get to knw alot of pple from various uniform groups.. they are all very friendly.. but.. haiz got to come back as shoulder last time injured pain again..  wake up damn early loh.. den go the falg raising ceremony cannot see flag.. haiz.. P.trainging also cannot participate.. after that report sick the Dr siao come and see.. me shoulder already rotting sia..  Singapore!!! come back to singapore see doctor sia haha... sianz.. come back le wait at changi jetty... wait for the first aider go drive the ambulance.. come back wif 3 first aider, one gal 2 men. gal call shu hui.. quite cute la she got talk to me so that i dun rot there.. the other 2 dun care their name lolz... jkjk.. sit in ambulance inside is damn cold.. somemore the 2 male first aider talk cold jokes there.. such as later go hospital muz look for Dr Tua as juz now at ubin is Br Siao.. lolz cold until turn into ice =X..  reach changi hospital liaoz.. go the accident and emergency dept where i come be4 when i fall on 1st aug at compass point.. see doctor there quite fast also dunno why.. i think the doctor heck care sia.. go inside, ask me move hand, ask me got go sch anot, den give me mc.. den can go back liao.. den my mother ask me to go back myself.. but i cant!! my ez-link card is in my wallet and my wallet is in my bag and my bag is at ubin!! siao liao! haiz.. still muz wait until friday for my friends to bring back for me.. sianz ar..  the first aider later sent me home.. i told them my mum will be back home by 12 but my mum end up reaching home at 2.. i sit by the door and fall asleep.. minutes were passing very slo when i was by my hse door... haiz eat medicine liaoz come online pple ask me create blog.. den i create loh.. lolz my first blog written wor.. haiz nth really much to add on la.. hope i can pass my next two days faster loh so that i can get back my belongings..'),(20587,4200075,'11,August,2004','2004-08-11 00:00:00','   urlLink    cute rite?&nbsp; urlLink   '),(20588,4200075,'11,August,2004','2004-08-11 00:00:00','   urlLink    another pic of hebe&nbsp; urlLink   '),(20589,4200075,'11,August,2004','2004-08-11 00:00:00','   urlLink    i\'m also a dog lover!!&nbsp; urlLink   '),(20590,4200075,'11,August,2004','2004-08-11 00:00:00','   urlLink    u\'re my princess!&nbsp; urlLink   '),(20591,4200075,'11,August,2004','2004-08-11 00:00:00','   urlLink    darling~&nbsp; urlLink   '),(20592,4200075,'11,August,2004','2004-08-11 00:00:00','   urlLink    cute&nbsp; urlLink   '),(20593,4200075,'11,August,2004','2004-08-11 00:00:00','   urlLink    hello~~&nbsp; urlLink   '),(20594,4200075,'11,August,2004','2004-08-11 00:00:00','   urlLink    another pic of my album collection&nbsp; urlLink   '),(20595,4200075,'11,August,2004','2004-08-11 00:00:00','   urlLink    chio!&nbsp; urlLink   '),(20596,4200075,'11,August,2004','2004-08-11 00:00:00','   urlLink    wow dis pic look more cool!&nbsp; urlLink   '),(20597,4200075,'11,August,2004','2004-08-11 00:00:00','   urlLink    S.H.E cool hoh..&nbsp; urlLink   '),(20598,4200075,'11,August,2004','2004-08-11 00:00:00','   urlLink    S.H.E paradise~&nbsp; urlLink   '),(20599,4200075,'11,August,2004','2004-08-11 00:00:00','   urlLink    S.H.E rulez!&nbsp; urlLink   '),(20600,4200075,'11,August,2004','2004-08-11 00:00:00','   urlLink    photo i taken during the autograph session&nbsp; urlLink   '),(20601,4200075,'11,August,2004','2004-08-11 00:00:00','   urlLink    selina say hi!!&nbsp; urlLink   '),(20602,4200075,'11,August,2004','2004-08-11 00:00:00','   urlLink    she can also be very sexy!!&nbsp; urlLink   '),(20603,4200075,'11,August,2004','2004-08-11 00:00:00','   urlLink    selina also very cute..&nbsp; urlLink   '),(20604,4200075,'11,August,2004','2004-08-11 00:00:00','   urlLink    my foreva cute hebe&nbsp; urlLink   '),(20605,4200075,'11,August,2004','2004-08-11 00:00:00','   urlLink    4eva heBe~!&nbsp; urlLink   '),(20606,3097708,'29,April,2004','2004-04-29 00:00:00','    today was such a beautiful day!!!i think its maybe 80 degrees out right now..whoa i can\'t wait til summer!i love when everyone wears such cute outfits to school-it\'s so much fun!school wasn\'t bad today..n that\'s all im really gonna say about it. I\'m goin to the yellowcard and something corporate concert in an hour!!!i cannot wait-oh yeah and tomorrow\'s my road test, which i don\'t even think is gonna happen, considering the fact that my drivers ed teacher is like ill. But i\'m goin now-cyassss  \"consider the odds, consider the obvious..\" '),(20607,3097708,'26,April,2004','2004-04-26 00:00:00','    hello everyone-long time no post, eh?let me see-i don\'t think i\'ve posted since monday night, so let me think about what has happened since then- tuesday-we had a game against lakeland, yes it was away. Although they are a really good team, i thought we played really well against them, but we lost 10-6. Congratulations Kristen Brenner, who had her first goal, she was really excited. I had 2 right handed goals (im a lefty) and sar had a fricken awesome assist to one of them. Seriously-it was beautiful..she drew 2 defenders, while faking a crease roll, and one of the defenders she drew, which was mine, when to cover her, n i went to goal-it was awesome, and im glad she gave me the nod to go cuz i didn\'t even notice it-lol. today-it was a pretty good day, and it was soo nice out!I came in late..yes it was maybe my 32 time doing that, no exaggeration either. SO yeah like i was saying, today wasn\'t a bad day at all. It actually went by pretty quick. After school, i had to make up a math test, and then i was off to practice, just kinda late. My coach was bitching at me though cuz i\'m always late to school, n he told me that hes having my first period teacher write down whenever i\'m late n give it to him. He\'s such a homo when he does that shit, i\'m passing math easily so i dont understand why its such a big deal if i\'m late once or maybe twice a week. But whatever-after practice Jamie and I went to the mall to challll n get some stuff.We sat on the thru-way for maybe a half hour--badddd traffice..but neways  We literally went on a sandal shopping spree at old navy which was pretty fun, n went to hot topic, where we saw joe n agovino, i bought a shirt there, n then made her return it 10 minutes later..good times..good times. Then on the way home, she let me practice for my road test, which is friday, n i did some parallel parking, i actually wasn\'t that bad. But whatever-tomorrow is the YELLOWCARD and SOMETHING CORPORATE concert..i cannot wait!!!My friends and I are taking a 30 person stretch excursion limo there and back. It\'s gonna be sooooo much fun-basically all of north rockland is going lol. But i\'ll let you all know how that goes-i\'m going to bed..goodnight!!  \"..tell all my friends i\'m dead, it won\'t be long before you forget my name..\"  P.S-tomorrow\'s Amanda\'s birthday!!!!!! '),(20608,3097708,'26,April,2004','2004-04-26 00:00:00','    \"..and i  dare  you to forget those marks you left across my neck, from the nights when we were both found at our  best - i could make this obvious, and you, you could  deny  me all in one breath.. you could shrug me off  your shoulders... forget me, it\'s that simple \"  taking back sunday-your own disaster  if you\'re a tbs fan and you don\'t have it, i suggest you download it '),(20609,3097708,'25,April,2004','2004-04-25 00:00:00','    whatsss reallllllly good?haha-its raining, yes pouring, and i hate it. Although, it did get our lacrosse team out of practice at 345 instead of 515, which was awesome. All we did was do wall ball, quick sticks, and then run a mile and a half as a team, which wasn\'t bad at all.School wasn\'t bad, pretty boring though..ahh i almost forgot-  HAPPY BIRTHDAY JOE! i made him such a hott card, seriously, it was all pink, ahh too cute!-i was definitely proud, n he probably didn\'t even read it yet lol. Well back to my school day-gym was pretty fun, but i was kinda beastly when i played 3 pointer knock out with some seinor boys. I was a little embarassed cuz i was winning,so basically i looked like a man. Umm after that, nothing really important went on during the day. Welllllllll..right now i\'m gonna finish helping chris make macaroni and cheese..(yes he\'s 17 and still doesn\'t know how) and then eat dinner, n babysit..i will keep ya posted with everything that goes on tomorrow!  p.s- thursday  is the yc and soco concert!!can\'t wait!!     -and  friday  is my road test!!  \"..i\'m so deceptive..\" '),(20610,3097708,'25,April,2004','2004-04-25 00:00:00','    oh god-where do i even start..how about right here..this weekend seriously SUCKED. i always do this to myself-saying \"this\" i mean..getting my hopes up so high, and then everything falls through.ahh what can i do..    Friday-coach guccione let us out of practice at 3:45..which was awesome, n i seriously couldn\'t have asked for anything better at that time to be honest with you. All we did was run a timed mile, n then do quick sticks (which me n dams did horrible at..the freshmen beat us everytime lol)After practice, i went to gavs to see her before prom, and then i went to dom\'s with gavs family to see everyone. Everyone looked gorgeous, especially danielle. When i got home from Dom\'s..my brother, my mom, and I were watching homevideos..(yes this was a friday night) n then dams came over before we went out. We didn\'t go out until maybe 945 when Perry and Brittany picked us up, n we just went to wendys, n then to a&p. Ugh-that really sucked cuz it was pouring, so basically me dams n britt stayed in perrys car the whole time. Dams came back to my house n slept over n then we went to practice the next morning.    I left practice early saturday cuz it was my little brothers confirmation, and i was his sponsor..how cute, right? I could not find nething to wear, cuz none of my skirts \"should be worn in church\" according to joanne martin-so my mom made me wear this black n white express dress..it was cute but not exactly what i wanted to wear.The funny thing is, is that as i was getting up to go up to the bishop with my brother, sven lonergan tells me i have white shit on my ass, which is highly noticeable in a black dress. When danielle notices, she starts slapping my ass in front of everyone in the church..i was definitely almost crying with laughter. What good times we always make, even in church. But the confirmation wasn\'t that boring, i got to chill next to my brother, tj, n danielle..my bro n danielles bro were actually being really funny when they kept saying mrs.doherty was going up for seconds for the host-lol but neways..Joanne n Leo decided to give kevin a family party, so ofcourse the fam n some neighbors came over to celebrate, even though i was sleeping almost the whole time..(my mom never wakes me up).      At like 7:30, Caitlins dad came n picked me up cuz we were going to Jill\'s sweet 16 at the Comfort Inn. She looked beautiful!and it was a lot of fun. When my mom picked me cait dams n matt up, n we got home, caitlin n i soon came to find out that there was also a hotel party goin on there, eric n mike tell us that after we leave, which sucked cuz the rest of the night me n cait just ate all the leftover food from my bros party, n then went to sleep. It was a fun night.     Now i\'m getting ready for my driving lesson with Mr.Vann and sar, and then drivers ed with all my pals. I can\'t wait til Thursday, and Friday! Thursday is the yellowcard and soco concert and we\'re taking a 30 person excursion stretch limo, its gonna be sooooo phat, and friday is my road test!ahh i can\'t wait, i hope i pass but i dunno lol-i shouldnt have told as many people as i did when it actually was, cuz then if i fail, everyone will know-lol well i\'m gonna go get ready..cyaaaa  \"let your waves crash down on me, and  take me away ..\" '),(20611,3097708,'23,April,2004','2004-04-23 00:00:00','    ehhh today wasn\'t too bad. There\'s not much for me to say about it actually. After school though we had a lacrosse game against Fox Lane Highschool, which went pretty well. We won i think 13-4 n i had 3 goals..holla acha gurl!after our game i just babysat n i actually just got in..which is absolutely horrible news. I\'m gonna go to bed though cuz i\'m so extremely tired-goodnight '),(20612,3097708,'21,April,2004','2004-04-21 00:00:00','    heyyyy everyone-haven\'t posted in a couple of days, but dont worry, nothing too exciting happened. Lets see.. yesterday  was pretty good. School was alright, at practice though, we did a timed mile (6:33) yeah man not bad, but after it i felt so sick.But whatever, after practice my mom picked me n dams up and she took me to get my nails and toes done..which was awesome cuz i REALLY  needed  that- Today ..hmm-it was alright..I was kinda in an ambitious mood, but that didn\'t last til practice. I had a bball meeting after school before practice, which went pretty well. We were all just figuring out what camp we were gonna go to, ehh can\'t wait for that..right. Well, i just finished eating, im gonna get some homework done, and then go babysit. Tomorrow i have a lacrosse game at 6pm at the highschool..u fans better be there!   \"this awkward silence makes me crazy..\" '),(20613,3097708,'20,April,2004','2004-04-20 00:00:00','    sup?today was such a beautiful day!i\'m so excited its getting warmer so soon i\'ll be sooo tan!lol-well whatever, school was alright today. I woke up kinda late n had to take a shower this morning so i just jumped out n didnt even do my hair that good, but many people complimented it so i think i might do that more often. There\'s not really much to say about today besides we ended having a lacrosse scrimmage against Nanuet at the last minute. Although we weren\'t much prepared, we won 12-4 im pretty sure, and i had 3 goals! After my scrimmage my bro, ralph, and i went to wendys n taco bell which was really funny when my brother almost cursed off the man there. Then, i had to babysit at 730 n i ended up baking brownies for kaylas bake sale tomorrow, which wasn\'t bad at all. I felt so sick when i got home now, and now im about to go to bed cuz i just finished my homework-seeee yaaaaaa  \"..this is what living like  this  does..\" '),(20614,3097708,'18,April,2004','2004-04-18 00:00:00','    omg wow-this weekend was sooo much fun! sorry i havent posted in a couple of days but i\'ve been really busy-Hmm let me see..friday after practice and jill\'s pasta party Jamie, Amanda, and Gavin came to pick me up so we could all chall..we went to wendys til like 1130 n then went to cheety\'s house(is that how u spell his name?)well whatever..although cheety\'s party was a chip&dip party it was still soooo much fun!Everyone was dancing, so ofcourse, my friends n i went out there too. I even got seth and perry to dance, it was a little weird considering the people surrounding us, but it was definitely a fun experience, although they were both drunk. Saturday morning we had a lacrosse game against scarsdale-ehh it was a hike to get there but we won 8-3 and i had 2 goals..it was a really good game. Our team is getting better i think, but then sometimes when i trip over sarahs stick in the middle of the game, i dont know what to think lol- but dams and i had the HOTTEST double team..the other team probably thought we were dykes cuz after the whistle blew she kept slapping my ass saying..\"yeah kel you know how we do!\" and so on and so forth. After my game i came home n challed, got some lunch with my bro, and then my mom took me to get a new cell phone-its seriously such a hott phone, but the bitch at the store forgot to activate it, so now i have a new phone but i cant make outgoing or receive calls-it sucks but my dad will go fix that problem tomorrow. Last night was Jessica\'s sweet 16 at the hilton inn. It was sooo much fun, especially when the fire alarm went off cuz of the smoke machine the DJ had-everyone in the hotel had to evacuate, n the firemen came lol-good times, good times. After Jess\', some people went back to Brett\'s house to all chall. It was soooo much fricken fun! I took a couple of shots, 1 with seth, another by myself lol, n then everyone baked brett\'s room. I wasn\'t smoking, but everyone swears i was high-there were maybe like 5 blunts going around in his room lol. So ofcourse, like always, i fell asleep and woke up at 430 in the morning to danielle telling me we have to go to her sisters lol..so we all walked up to deli central (it was actually open) and then gav n i took a cab to her sisters n just passed out when we got there. I seriously had SOOO much fun lol-Now i\'m home, just got in from mcdonalds n the shoprite parking lot(like old times) with gav n her mom n all the girls r about to go chall at the park cuz its such a nice day out, then i\'m off to drivers ed which is always boring as fuck. But i\'m gonna get going, i will keep ya posted with everything!  \"..this all was only wishful thinking..\" '),(20615,3097708,'16,April,2004','2004-04-16 00:00:00','    today wasnt a bad day at all.-i actually stayed awake in all my classes and didnt even come close to dozing off once, which is probably a first.Umm..let me see-i slammed my nail in my locker today which hurt really bad so now i need to get them done sooner than i was expecting to..but what can you do. Human development was fun when kristina, erica, and i were telling mrs.newborn that oprah and the olsen twins were either anorexic or bulimic..she believes almost everything we say to her lol-she even thinks kristina saw a bald eagle on her deck one morning..which is hard to believe considering the fact they are extinct..but neways..after school mike agovino drove me home cuz i didn\'t wanna stay til 4 doing nothing so i went home n challed for a little then went back to the school for my game..(thanks to everyone who came!)..we won 10-0 but i just played so bad. 3 goals of mine didn\'t count cuz i was in the crease after and i was soooo pissed -but i finally scored a goal that counted which made me happier but i dunno i was just in such a bad mood and i felt like i was gonna cry if something else went wrong. WHen i got home, i went straight to the callanans to babysit til 10pm..yeah it was a late night there tonight for some reason, and now i\'m just online talking to joe alimo..its always fun chatting with him. I think im gonna go to bed though in a bit, mom said she\'d drive me in late on account of my insomnia..Ehh tomorrow\'s friday-i\'m happy but im upset cuz i know nothing is gonna go on cuz nothing ever does...hopefully i\'m wrong. Oh yeah-i heard about chittys..remember that in the beginning of the year?haha..well i\'ll keep you all posted.. p.s-can\'t wait for jess\' sweet 16 sat!!!  \"..i heard about your regrets..i heard that you were feeling sorry..i heard from someone that you wished you could set things right between us-i guess i should\'ve heard of them from you..\" '),(20616,3097708,'14,April,2004','2004-04-14 00:00:00','    i can honestly say that i never remembering ever being this tired. Every period today i felt like just falling asleep, but it\'s kinda hard for me to do that in school. Today seriously just sucked. We didn\'t play yorktown cause of the rain but we still had to do a time mile which i did pretty well..6:47-not as good as last time, but not bad. But seriously-i think my day got so much worse at practice when i got pegged in the boob n then in the chin-i wanted to murder someone..you have no idea how much that hurts-Now i\'m just like hanging out here waiting for my dad to bring me home pizza..then im off to babysit..hopefully when i get home i\'ll be able to go to sleep, which hasnt been happening lately-well wish me luck on that-i\'ll keep ya posted!  \"you\'ll be sorry..isn\'t that what they\'ll say? don\'t follow your heart \'cause it just seems to get in your way..\" '),(20617,3097708,'13,April,2004','2004-04-13 00:00:00','    will it ever stop raining? It\'s april, and I can\'t even wear cute spring outfits becuz of this weather. Ugh-it\'s so depressing. Well anyways, I think i\'m becoming a bad student. Dont get me wrong, I go to class every period, but now I just don\'t do anything. First period i have math, and all i do is eat my mini muffins n drink my yoohoo. I pretend im listening, but all i really do is just think of what i\'m gonna do the upcoming weekend. But today actually became exciting towards the end. Fifth period, Kristina, Erica, and I took a little trip to the bathroom to read the shit which is written on the wall about Kristina. Which is seriously soo gay. Let\'s see-nothing on the wall was actually true, and why would someone waste their time writing on a wall about someone, just say it to her face. That just goes to show how the girls in North Rockland High School are-but we all do it. Obviously Kristina got pissed and said something to the girl and they ended up getting into a fight. It sucked cuz i didn\'t get to see, but its kinda good cuz i\'m so convinced i would\'ve jumped in-but she didn\'t need me to neway. So the boxer Kristina actually got away without a suspension, which my parents do not believe me about ( i have no idea why ) but i am serously MAD proud of her..( yes i did say \"mad\" )cuz i can say that now since i make hott remixes. But overall, Danny saw it, n told me it was a good fight, and how he was so proud of Kristina. I could seriously picture him just standing there, cheering, n like saying.. \"ahh nice move\" or \"wow matt, did you see that?\" haha you all know he would say that. Our game against yorktown got cancelled, so tomorrow it should be happening-but its away. Thursday our game is home against Brewster High School, and you all better be there. Alright, i\'m gonna go eat dinner, then watch gilmore girls, then one tree hill-you all should be doing the same. '),(20618,3097708,'11,April,2004','2004-04-11 00:00:00','    I think i can speak for most people when I say that today was maybe the worst day. Oh yeah, let me clue you all in now, I will have bad luck for 7 years becuz i through my mirror at my closet door last nite-so that might be why i had such a bad day. When i woke up, it was just so gloomy. And yes, I hardly could wake up becuz i didn\'t fall asleep until 2am becuz of my insomnia that i\'m convinced i have. I did nothing in each class-just sat that, drew in my agenda n daydreamed. I got caught in global maybe 4 times too, which wasn\'t good. After lacrosse practice today, Sarah had a pasta party for our team becuz tomorrow we are playing yorktown (away..sorry!) It was a fun time i would say, we made our shirts, well actually Dams basically made mine, but they all came out good. But i\'m gonna go do my homework, and then talk to Eric, n then go to bed. Gotta get a decent amount of sleep for the big game tomorrow! I\'ll let you all know how that goes! '),(20619,3097708,'11,April,2004','2004-04-11 00:00:00','    Happy Easter!I just got back from church a little while ago, and now i\'m just like challen eating my chocolate. Today\'s gonna be a boring Easter cause my mom is cooking, although after dinner i\'ll be up at Kristina\'s house to hang out there. I am so upset to say that this spring break sucked so much. Even last night, all i did was just hang out with matt, danny, brian, kyle, and heather at heaths. But before that,  me, matt, danny, n brian went to kfc n we played the name game which i did AWESOME in, and as i said last nite, just being there eating, n playing that game with them was such good news. Tomorrow is school, which sucks so bad, but what can ya do. I was one of the people praying for snow days everyday, and this is my punishment. Well im gonna go, i hope all of you enjoy your easter day.   \"i called my mom last night-she said, \"Sweetie, you don\'t need someone, who\'s more fleeting than fall\"\" '),(20620,3097708,'10,April,2004','2004-04-10 00:00:00','    hey everyone-i havent posted in awhile..or maybe like a day n a half, it just seems like awhile. My parents went away thursday to Maryland so grandma peg has been here but she actually just left-Thursday night, we all went to Toms, and we dont really need to recap all of that. Yesterday, I was off from lacrosse which was seriously a treat cuz i\'m so sick of waking up early when there really is no need to. Last night, Danielle and Kim came over and hung out for a little before we went to the bonfire( or the valley..whichever u prefer). When we got there, we came to find that Damiani, Britt, Jill, Meagan, Danny, Joe, and Georgie were all maybe drunk off their asses. I normally don\'t like being around drunk people when i\'m not, but it was really funny watching them do stupid shit that they didn\'t even notice they were doing. We finally left there at around 11:45 and walked up to Stonys to get a cab but ended up talking to these weird old guys about accents, and how we\'re the future of america. The cab finally came, and i feel bad cuz Jill and Meagan were late, but what can ya do. Danielle, Kim, and I went to Jenn\'s to just chall and kim was mad pissed we wouldn\'t walk her home, but whatever kim u do it to me all the time. When kim left, me danielle n jenn just took a shot n talked til 4am with Jenns bro and i dont think ive laughed that hard in so long. So basically it was a fun night, besides the fact we squished 3 people in Jenns bed, which wasn\'t that fun. When we woke up, mrs.nicks brought us to bagel world so we could torture kim while she was working, and we just had breakfast there at 12:30..although it should\'ve been lunch. Well right now i think i\'m gonna lay down for a little and then take a shower and figure out what to do tonite..i\'ll keep ya posted! '),(20621,3097708,'08,April,2004','2004-04-08 00:00:00','    heyyyyyy guys-i just got in from lacrosse, and i\'m so happy to say that it might\'ve been the easiest practice i\'ve ever been to- But whatever-last nite was fun, i was with Jamie, Amanda and Britt. First, we went to like this bonfire thing and Britt loved it so Jamie and I left for a little and went to Ross\' to hang out with some people then. Then, after staying there for a little bit, we went back to the bonfire to pick up amanda and britt, but we dropped britt off at joes cuz she had to pee really bad. I hate being around drunk people when im not drunk- especially when brittany is screaming..lol i love her-But whatever, then Jamie, Amanda and I went to Pat\'s for maybe 15 mins but i think it was better than staying at that bonfire. Then i just went home talked on the phone for a little bit and then finally fell asleep at like 230am..i can swear i have insomnia. Fortunately, my parents went to maryland today(it\'s my dads birthday!) so they\'ll be gone for 2 nights so i should have fun-even though Peg is coming over. We will see what happens. Even though its the afternoon, i\'m gonna go wake up my brother and see if he wants to go to deli central..see yas! '),(20622,3097708,'07,April,2004','2004-04-07 00:00:00','    im sorry to say this, but spring break officially sucks-it\'s already tuesday n i haven\'t done nething great besides james\' party friday which was awesome. Today after lacrosse, my dad n i went to the mall and ate lunch at TGIF\'s..it was great cuz i love when me and him just hang out. Well whatever after lunch we went to abercrombie and i got a skirt and a shirt, n then to modell\'s to get lax gloves (which i really needed). I guess he was feeling brave cuz he let me drive home from the mall..on the palisades parkway, which always makes me nervous, so i can only imagine how nervous he was. Surprisingly, i did really well, so he told me to stop at Hoyers on the way home so we could get some icecream..it was a fun day. When i got home, I took my nap, and woke up finding out that there was absolutely nothing to do tonite. We all got our hopes up thinking Tom was gonna have people over but that didnt work out. But whatever, Dams, Sar, and I went to Saints house and ordered Planet Wings, which was a treat, and then we just talked. I just got in a little while ago and I think i\'m gonna go to bed cause its late..soo goodnight! '),(20623,3097708,'06,April,2004','2004-04-06 00:00:00','    this spring break better start getting better..  After lacrosse i slept until 6:00pm which was sooo gross..but it was definitely a great nap. I soon found out that i had to babysit til 9 but it wasnt bad cuz Colin was being extremely funny when he was acting like the HULK-hes 2 cute. After babysitting Kristina came over and we watched the Passion of Christ..or we at least tried to-somehow we kept getting interrupted..My dad brought us home some finnerty french fries and was being sooo funny when he kept leaving the room and coming back after he said he wasnt lol-you kinda had to be there..well anyway there was like mad tugboats here tonite, i have lax in the morning, and im really tired so i\'m gonna go to bed..hopefully tomorrow will be a better day..nite '),(20624,3097708,'05,April,2004','2004-04-05 00:00:00','    heyyyyy everyone- saturday night i think i totally jignxed myself by saying hopefully something will go on, becuz i think im safe to say that kristina, dams, and i definitely did nothing- although we did have so much fun watching damiani\'s and mine confirmation tape from the 8th grade. but whatever we slept over dams\' house and had fun-Yesterday:it was my parents 25th anniversary party at the fireside..so basically you know the deal with that-family, friends, everyone getting drunk (especially my parents) which is always fun to watch. Damiani came over before they got home n witnessed the whole thing- good times, good times.Last Night:we went to Ross\' house which whenever we r there i always manage to have so much fun. Even though it was snowing earlier, John and I still went in the hot tub, which was an insane thing to do, because when we got out it was freezing. It sucked taking a cab home with damiani, especially because i left my wet tanktop in it, yeah im pissed. This morning we had lax from 9-1130 n i think i literally almost killed myself. Not meaning to complain, but my toes were almost falling off, especially when a girl on my team stepped on them. Well whatever,i\'m gonna go lay down and hopefully something fun goes on tonite! '),(20625,3097708,'03,April,2004','2004-04-03 00:00:00','    hey guys-this is so weird, i wanted to create one of these things for the longest times n i finally got around to it-maybe because my comp is actually fixed! lol well neways..i\'m so happy to report that this vacation has started out awesome..last night we went to james\' house and i think im safe to say that we were all very much intoxicated-i cant remember the last time we all maybe peed laughing-we were in the cab, n to begin with, damiani slammed the door on my face so i started going on and on about how my face is bleeding ( this is what im told), and then we were making the hottest remixes..or maybe i was just studdering to rubberband man-orrrr walkin to the cab with saint n dams was MAD funny while we were singing..\"i dont wanna know..(what what nigga what)\" lol such a great night! then today..i had a lax game vs. albertus- we won 11-9 and i had 4 goals!i was really excited, but we didnt play that well towards the end. oh well, we\'ll get better.Well im happy to say that i just got my nails done w/ saint n my mom n im gonna get ready to go out..hopefully something fun is going on! '),(20626,3097708,'23,May,2004','2004-05-23 00:00:00','    ehh let me start off by saying that i am EVER so tired. It\'s almost 2 in the morning, and me being the huge loser that i am, is sitting posting in my journal. Hey, well i have to. Friday was so much fun. After school Cogar brought Amanda, Danielle n I tanning, and yes believe it or not, i actually did get tan. My little brother just asked me if that heart on my hip was a  tatoo  lol cute huh? Neways..after tanning, Danielle n I went to get our nails done, but first we stopped in at Sal\'s pizza to get some pizza and visit Joe. So whatever we got our nails done, n toes, and then i went home to shower n shit. Cait, Dams, n Danielle came back over n my mom brought us to that kid justin odells house, so we stayed there for a little bit, and thank god i had something to drink. Cogar picked me n danielle up after n we went to some college kids party, i dont know his name, but he was darins neighbor lol. Yeah a lot of people were there, including my husband James Sanjarourity.  He wanted it sooo bad lol . As we were leaving this party, we were walkin to cogars car, and austin was passed out in the front seat. It was maybe one of the funniest things i have ever seen in my life. But anyways, we went to DUNKIN DUNKIN (my favorite place EVER) lol with some people, and that was fun. We went back to cogars for a bit n then me n dan slept over sars house, it was definitely a fun night i would say. I was a little pissed though, but all in all, it wasn\'t bad.    This morning, Danielle and I went tanning again. As you all will probably to tell Monday, i actually am getting color lol. After tanning i went home to take a little nap before Amanda\'s Sweet 16. OMG-it was so much fun, and the cruise was soo beautiful. Amanda seriously looked gorgeous, i think she knows that cuz maybe everyone told her lol. So yeah that was really fun, especially when we were all screwing around and dancing, and making fun of the hott couples on the dance floor. It was definitely a fun time. We got back to Amandas at like 11:15 and some people just stayed there, and that was fun also. JohnnyO just actually dropped me off a little while ago, and i\'m very grateful he did, considering the fact i\'m really out of the way. Well i\'m gonna go to bed, i\'m extremely tired..laterrrrr   \"..do you even remember..?\" '),(20627,3097708,'20,May,2004','2004-05-20 00:00:00','    hi everyone. I\'m so glad that tomorrow is Friday, you have no idea. Today was a fun thursday though. School was alright, i guess i really can\'t complain. I didn\'t swim in gym today, which made it even better. Katie drove me home after school so i actually got home early, compared to getting home like 20 mins later when taking the bus. My bro drove me to amandas after i got home because i was really bored, so me gav n amanda just challed there for a bit.  I WAS SO HAPPY  when joanne gavin picked me n dan up n brought us to mcdonalds. Seriously, i was definitely  having the time of my life  being with them two, eating mcdonalds, and listening to sublime at the same time. Oh my god it was definitely amazing. So yeah, me n dan went back to her house n then saint came down. I can definitely say we were having funny times on the computer. A little bit later Eric picked me n danielle up n brought us to hoyers. He\'s such an awesome friend, even though he wouldnt treat lol. When i got home i went to babysit, and it was seriously like nothing. I just sat there n did my homework n watched american idol when the kids were sleeping. It was an easy night at the callanan house, surprisingly.      I really hope something fun goes on tomorrow night! Saturday is Amanda\'s Sweet 16 on the World Yacht boat. I serously cannot wait,  its gonna be so much fun . I\'ll keep ya posted!  --Meaghan Jean Rogers..i miss you so much, and i can\'t wait for next weekend boo!   \"..but i know you want me back now..\" '),(20628,3097708,'19,May,2004','2004-05-19 00:00:00','    its really late, im so bored, and yes i know i posted tonight already. I just really had a rough night with everything--my brother, my mom, and babysitting. Ugh..i can\'t wait til the weekend. But yeah so in the mean time i\'m listening to \"Lullaby\" by Taking Back Sunday, everyone should download it, its amazing. Oh yeah, the new Dashboard Confessional song \"Vindicated\" is on the spiderman 2 soundtrack, download it-its one of his best songs ever!oh yeah, thanks sean for letting me know about that.I\'m gonna go to bed, Gooodnight!   \"..and we can talk all night..\" '),(20629,3097708,'18,May,2004','2004-05-18 00:00:00','    heyyyy everyone-i just got in from the mall w/ jamie, and damn did we have an awesome time or what!!We needed to get dresses for amandas sweet 16, and yes we were successful. I, with my compulsive shopping problem, ended up buying more things then i actually needed. OMG-we saw the hottttest guy  EVER  at kids abercrombie, i think he was maybe one of the most gorgeous people i have ever seen in my life, just like that guy from the assembly today lol. SO yeahh school was ehhh today. I was in such a bad mood, becuz danny was giving me a huge attitude today, and then someone else just put me in a bad mood. Ugh, i really hate when people do that to me, and then im in a bad mood for the rest of the day, and then people begin to think im mad at them or something, just like katie kinda did at lunch. But besides all that bad stuff, the assembly my class went to today was awesome. The guy who was a drug addict was so fuckin  hott , and i know this may sound weird, but the fact that he was once a drug addict made him so much more attractive lol. So yeah, at the end of the assembly, my gay friends n i went up to him to get like information about the drug programs just so we could get a good look at him, lol we\'re so gay. After school i stayed after with my lax team to hand in our uniforms and stuff, it was ever so sad, and then me amanda n britt hung out with mr.krogslund til jamie came. Seriously, i really think he likes me, but amanda also thinks he likes her, but whatever. After we dropped manders n britt off, me n jamie went to the mall n did all that stuff which i already said, it was fun. I  love  whenever i\'m in her car cuz we always blast the taking back sunday cd and it NEVER gets old..ahh good times, good times.    I\'m sooo bored, but i\'m gonna go babysit soon, that should be fun. Laterrrr   \"..don\'t let me down..\" '),(20630,3097708,'18,May,2004','2004-05-18 00:00:00','    ahh tuesdays seriously suck, especially because i had gym today, and yes i am still swimming, which even sucks more. School went by so incredibly slow and i wanted to go home early the whole day but i couldnt because i had basketball after school..yes, lacrosse ended yesterday, and basketball (which is a winter sport) started today..crazzzy stuff huh? So yeah after school n bball, my mom picked me up n surprised me telling me we were goin to fiesta!i seriously love fiesta even though i hate mexican food n i get chicken fingers and fries there..lol. After dinner with my parents, my mom let me drive to amandas, and no i didnt crash. We were supposed to go to the wooden bat game but then we ended up not going, and i was just watching amanda drink-ha what an alcoholic. Jamie came to amandas after her practice n we were all just  challen  n screwing around with amandas clothes, like always. Jamie just dropped me off n im about to study for a little and then go to bed..sorry today wasnt very excitinggg...laterrrrr   \"..i wanna hate you so bad, but i cant..stop this any more than you can..\" '),(20631,3097708,'17,May,2004','2004-05-17 00:00:00','    lets be honest here, i haven\'t posted in awhile because my computer has kinda been a little messed up. I just got home from my lacrosse game against yorktown who is the best team in our   league.We held them really well and only lost 15-14..i had 6 goals!It was crazy but everyone played really well, and i was so proud of our team.This girl on the other team stick checked me in the lip, so yeah, now i have a fat lip. Let me go back to the weekend though-friday night mi madre n i went to mall and she bought me some things for my report card...she bought me this awesome purse and such hott thongs from aber..& some other things, but i was most grateful for those 2 gifts. After the mall i went to gavs n challed there with her for the night. Sat..i had practice early, then went to ciceros with meg, sar, n saint, n then they dyed my hair!!(which only some people are actually noticing)..Caitlin had a party that night at her grandmas, which is always  fun .-i was seriously drunk off of my ass and it was awesomeee. A lot of people started showing up, n you know pic got jumped, n then he went to fight at wendys so i went with a bunch of guys there..After that, I went back to austins with him, joe, n cogar..i think?? and then some people starting coming there. That was a good time also lol..aha..Austin fell asleep after awhile so it was just me danielle n joe chillin, and  danielle was definitely doing tripods for us ..it was soo funny. We finally got austin to wake up at 415am so he could bring us home..Gavin n I weren\'t allowed in my house at that time because my dad was up for work, so we went up to saints n chilled with these 2 disturbing cats in her basement..n then went to my house at 530am..umm yeah we kinda got a little caught in the morning by my mom also..ehh..but newaysssss..Sunday was my pasta party for my lacrosse team..but with my luck, my mom ending up bring my brother to the emergency room becuz his ankle like broke the night before..i was soo pissed and i was like flipping out. Dams and Sar came over early to help, n thank god for my neighbor Nancy(yes ill wills mom) for cooking the pasta and everything..she was seriously a huge help. So yeah the whole team came over, including my coach, and it was a good time, even though i felt a little hung over. So yeahh that was my weekend, it was so much fricken fun you have no idea!!!laterrrrr  \"listen trick, i\'ve had all i can handle..we won\'t stand for hazy eyes anymore\" '),(20632,3097708,'12,May,2004','2004-05-12 00:00:00','    what a disgustingly hott day..i was actually glad that i had to swim during gym cuz it cooled me off, and my hair didn\'t come out that bad..other than all of that, i really can\'t complain. Today, my mom came into the 4th and 5th(my class) period human development classes as a guest speaker to talk about her occupation. I was a little embarassed, but i guess i got over it-also, i felt a little bad when she asked me why i was being so mean to her, but i dont really think i was-i think she just heard george saying that i was a mean daughter..but newayssss...we were supposed to have a lax game today at nyack, so i came home from school n went back, n guccione told us an hour after the bus was supposed to come that the game was cancelled-i was a little pissed about that..tomorrow we\'re playing north though, that should be a good? game..hmm we\'ll see lol..i\'m actually about to go babysit soon, i hope the kids dont give me a hard time, ehh but they probably will- laterrrrrrr  \".. make it hurt ...\" '),(20633,3097708,'11,May,2004','2004-05-11 00:00:00','    let me first off say that i have to swim tomorrow and its goin to suck so bad..i\'m so not looking forward to it-but anyways..i am so tan right now, like you have no idea..thanks to practice in 90 degrees, it at least helped in some ways.today was an alright day..i was pretty pissed about my hair cuz i didnt really take a lot of time to do it so it was just kinda wavey all over, but what can ya do. My thumb is still kinda broke or whatever it is, and im kinda pissed cuz now i have this thing that i am told by carol i have to wear, but its not even on now so i bet that doesnt work for awhile. Practice was sooo boring today. But in the beginning, i must say, sarah pleasured me by stretching my leg awesome- omg it felt so good , not trying to be gross or anything lol. After practice me n sar went back to britts n jumped in the hot tub before the team came over for the pasta party. It was sooo much fun, especially when Baron kept putting his ears behind my head..that was soo hott!lol but yeah the pasta party was fun, and i just really got home. Gilmore Girls and One Tree Hill are on tonight, i cannotttt wait..god jess and lucas are soo fricken hott!im gonna go jump in the shower..chao!  \"..i begged you not to go, i begged you, i pleaded..claimed you as my only hope\" '),(20634,3097708,'09,May,2004','2004-05-09 00:00:00','    what can i say about today..let me see-it actually went by pretty quick, n it was so nice out. It sucked becuz i didn\'t realize that i had bad sunburn until i got to school and many people told me i did lol.Also..many people..including danny told me that i shouldve worn a pushup bra today which i was reallly pissed about!lol but u guys were right- but anywaysss you do not understand how happy i was when i got to bio n we were doing a lab the whole period, like seriously i think that made my day especially cuz jeff is my partner. Ehh..nothin really too special happened..but let me think-lunch was pretty fun i guess..Austin came to chat with me towards the end of the period, and hes gonna LOVE the fact that his name is in my journal..but anywayssss-i came home after school before my game n eric drove me back to the hs just in time..im so fortunate to have such a great friend like him! We played pearl river highschool and we fortunately won 11-5, i had 2 goals! and one of my goals my coach was pissed about..but whateverr what can ya do. Sadly, i think my thumb is either broke or dislocated, and mrs.coughlin agrees also. I got checked extremely hard, i almost started to cry but i held it in. I\'m gonna get going..i\'m looking forward to tomorrow, its gonna be really nice out, and britt is having a pasta party for our team..can\'t wait!!!   \"the jukebox is in the corner, my mouth is the speaker..it plays your favorite songs..\" '),(20635,3097708,'09,May,2004','2004-05-09 00:00:00','    yesterday morning was our class of 2006 carwash.I\'m sad to say that we weren\'t as successful as we were in the beginning of the year..maybe because hardly anyone was there. But anyways, i had a lax game vs new rochelle..they\'re horrible..we won i think maybe 14-0..i had 2 goals, but then in the second half we let our defense score goals, so that was fun. After the game we had a barbeque,. which was very nice of galati. I was looking forward to going to john gribbens after the game, but ofcourse his party was cancelled so i just went home n slept. That kid richie was supposed to have another biiiiigggg party too. don\'t you hate when you think a night is gonna be so much fun, and then it ends up not being all that you made it up to be?well that was last night, although i shouldnt complain cause i did have fun with jamie, katie, and michelle. we went to that kid richie\'s, but there were maybe 20 people there n it was pretty gay, so we left to go to the moontower.I seriously had to pee so bad, so i went in the woods, but by accident i stepped in quicksand n lost my flipflops..i was seriously sinking, n katie thought it was so funny so she went n took a picture while i was like crying. Thank god for alex, he got me out after like 10 minutes of me just standing there, n katie telling me to just walk out of it barefoot..she didnt wanna get her white shoes dirty lol. We went back to that party, it was kinda gay, but it was better than doing nothing. Jamie brought me home, n then i just went to bed..i was soo tired, i woke up with 4 missed calls n i didnt wake up at all when my phone rang.. this morning my mom n i went to get our nails done since it was mothers day lol. i lether pick my color out for me, i didnt think it was gonna be bad, but now i have a purple french manicure that kinda looks disgusting, but i shouldnt complain. We went to shoprite after, n my mom decided to be generous and let me drive home. Well so much for that, as i was pulling out of the parking spot, i crashed into the car next to me.. ehh it was really bad..the other car was all crashed n shit so my mom made me get out of the drivers seat so she could drive. i feel really bad lol but whatever..can you believe that i\'m supposed to have my license right now?im a little scared..alright well im gonna go spend mothers day with my fam..it should be fun---laterrr   \"..hey thanks, thanks for that summer..\" '),(20636,3097708,'08,May,2004','2004-05-08 00:00:00','    ahhhh i\'m soooo fucked up right now its awesome-tonight was sooooo much fun.....some kid rich had a keg party, which i got a lot of beer for my 5 dollars so im really happy for that. me sar dams n britt went to caits before we went, and then went there n met up with everyone. like seriously it was awesome-but there were some pretty gay girls there that shouldnt have been there, but other than that it was fucking awesome..especially when sarah pissed in the tub-oooooooooops!!!when the cops came i hid underneath a car n then one of my bfffffffs katie russo found me n drove me somewhere, and then me keri n dana went to wendys n then i went to gav\'s sisters house-thank god for that..i didnt really have a ride home, but brianna n katie said they\'d pick me up-like seriously i love having them as my friends, they are so helpful. tomorrow morning i have to go to the fuckin class of 2006 car wash, so evetyone better get their cars washed...and then i have a lax game at 1230..beeeeeeeee therrrrre!   \"but i know-you\'ll be calling, calling, calling me again..\" '),(20637,3097708,'05,May,2004','2004-05-05 00:00:00','    i haven\'t posted in FOREVER-seriously..well since sunday...and yeah ross i hope you\'re not reading this, that goes for him and anyone else who thinks my journal is boring-you dont have to read it. Monday we had a game against some team far away-i can\'t quite remember. We lost-i\'m really pissed cause me n jackie couldn\'t play since that was the game we were suspended for because of the concert. What can you do, my coach didn\'t HAVE to suspend us. Tuesday we went all the way down to Mahopac, and we were only able to play a 30 minute game, if u even wanna call that a game. Anyways, we lost 5-4..im so pissed cause we should\'ve won. I had 2 dingers, but i didn\'t really play my best, my allergies were killing me. On the way home, i unfortunately repeated a ride home just like soccer season. I had to go to the bathroom, like always, so the bus stopped at bear mountain-just like soccer season..(haha!2 everyone who knows im talkin about!) but i seriously felt bad, but i had to pee. Today was such an ehh day. Once again my allergies were acting up, it was horrible. My school day was followed by practice, which wasn\'t bad, and then actually coming home to my dad being pissed about my cell phone bill. Whatever, he can deal with it. I\'m about to go babysit..it\'s kiera\'s birthday so it should be fun over there. Sorry my entries haven\'t been really exciting lately..hopefully they soon will be   \"..i\'ll sing whatever song you want, for whatever mood you\'re in..isn\'t that what you expect?i can sing you to sleep..\" '),(20638,3097708,'02,May,2004','2004-05-02 00:00:00','    i\'d like to say that tonight was enjoyable. I slept til maybe 830, caitlin came over, n then my cousin brought us to get icecream, n then to the bonfire where sar, dams, saint, britt, megan, joe, and danny were. It was a fun time, although i was only there for maybe an hour. Dams, Saint, Sar, and I then went to tara coughs where her, heath, kaity, danny, matt, ross, mike, john, greg, and joe were. We had fun playing charades, although taras sister was the best. We all just hung out there for awhile, and then sars sis picked me, sar, saint, dams, danny, matt n mike up n went to wendys.It was a fun time-i think im gonna possibly go to sleep now, im reallllly tired-cya laterrrr     \"my heart is on the floor, why don\'t you step on it?\" '),(20639,3097708,'01,May,2004','2004-05-01 00:00:00','    omg-the yellowcard and something corporate concert was AMAZING-thats maybe not even the only word i can describe the experience with either.Like always, i was late getting to amandas, but fortunately the limo hadn\'t arrived yet. The limo, was awesome...except everyone was arguing about what song to put on, n who was gonna stick their head out the sunroof. When we arrived to the roseland ballroom, joe, kevin, jon, and i made our way to the pizza place and got a slice before they let us in. They planned on taking a bong or bowl (not sure which it was) in, so they have me put half of it in my bra, and the other half was down joe\'s pants, but the security guards found his n took it lol. When we got in, we made it so close to the stage, but then once the swaying started after the first band that opened, i felt like i was gonna faint. Steve threw me up, and i crowd surfed when there wasnt even a band on stage, lol-i felt like such a loser but i needed to get out of there. After i did it once, i couldnt stop doing it lol- I did it 7 more times, but fell twice, including one time my head got stepped on..it was really scary. Something Corporate, and Yellowcard both were incredible. The way home from the concert was kinda a little irritating. People wanted to sleep, and other people wanted to like sing n shit. so ofcourse, there were arguments. but all in all-it was a really awesome time. friday:my road test was cancelled..i was SOO pissed-ugh but whatever, it will be soon im hoping. School wasn\'t bad at all cuz i went in late, just like maybe everyone did who went to the concert..which was possibly 80 people from our school. Coach was pissed that we missed a game, so we thought practice was gonna be hard, but it really wasnt. We scrimmaged varsity the whole time-it was a good time too. After practice, i went home, showered, n then went to my friends house for a little bit-maybe til like 10. When i got home, eric picked me up, then we went to pick up gavin n then we went with him to get his brother(my future husband) from the train station..(he looked MAD good). Danielle and I got to my house at around 1215..and then Austin came to pick us up, and we hung out with him, stigel, joe, and brian..It was such a funny time, they thought we were high or drunk but we realllllly werent, thats why it was even funnier. We got home at like 230, and then saint came to sleepover too.  today-i had a lacrosse game..we played south and won 15-5? thats what i\'m guessing. I played pretty good, i had 4 goals, but i was sooooo tired, cuz it was soo incredibly hott outside. Sarah played really well too, i don\'t think she got as tired as me though lol. Now i\'m just sitting home, trying to find out something to do tonight-i\'ll let you know how that goes..  \"..oh should i tell you everything?i\'m feeling out of luck..\" '),(20640,3097708,'31,July,2004','2004-07-31 00:00:00','    last night was a girls night at gavs. Gav, saint, sar, dams and i were there for a bit, then drove sar home, n then watched confessions of a teenage drama queen. Going into the movie i was so excited cause i love lindsay lohan, but as the movie went on, it was maybe the strangest thing i\'ve ever seen. So yeah it wasn\'t that bad though. After the movie saint n i walked home, and then i just kinda went to bed. Tonight though should be amazing.  Amanda, Gavin and I are going to the city to see a show at the Knitting Factory . It should be sooooo much fun!I can\'t wait,  i\'m afraid of the city kinda, but i still like going there.  Maybe we can see my husband James while we\'re there..hmm..thats a good idea. Well im gonna go run to the bank in a bit, then start getting ready for tonight. I\'ll keep you all posted!     \"..to come around..\" '),(20641,3097708,'30,July,2004','2004-07-30 00:00:00','    ahh what a fun night!i just got in from 7/11, and prior to there, we were at stony point elementary school.  Yes it was so much fun , although i wasn\'t smoking. I was there with britt, and then like a bunch of guys from my grade, matt dillon and some of his friends, and then mike. I love hanging out with them cause they are seriously so much fun. Except i didn\'t have much fun when greg speared me on the grass, but he knows why he really did that. Anyways, today i just babysat for a bit, and i brought colin to danielles to swim cause she was watching her neice and nephew, and basically all 3 of them are friends (well we actually made them become friends).But yeah it wasn\'t bad. I\'m gonna go watch boy meets world. Laterrr...   \"..now who you gonna call tonight?\" '),(20642,3097708,'29,July,2004','2004-07-29 00:00:00','    so yeah its pretty late, or maybe i should say its pretty early in the morning. I\'m happy to say that we won our basketball game tonight against old tappan, which means we\'re competing for 2nd place i think. After my game, erin came over to pick me up and then me, erin, gav, n christine went to georges for a bit, literally maybe 10 mins if that. Erin dropped me n gav off at bretts which was fun cause  those guys are soo funny, like all the time.  That was really it for my night. I was actually home by midnight, it was kinda weird. But yeah, another day of babysitting tomorrow. I doubt it will be bad. Laterrr..  &nbsp;  \"..and wish the worst of what i was..\" '),(20643,3097708,'28,July,2004','2004-07-28 00:00:00','    i just got in from babysitting, it was kinda a funny day there. I put palmolive in the dishwashed instead of the stuff you\'re supposed to, and bubbles starting coming out of it all over the callanans floor. Pretty funny, but i cleaned it up, thanks to kristinas help. Yesterday i babysat once again, and then went to joes house to burn the new taking back sunday cd since my mom wouldnt let me drive to the mall. She let me drive to garnerville though so i dont understand the big difference. Then i went to pick up katie n we went to wendys where the employees thought she was an actress because of the get up she was wearing. We proceeded to danielles for a bit and then i went home to shower and stuff. Ehh last night wasn\'t all that fun. I have a basketball game soon though. I think i might take a nap before erin picks me up. Hopefully tonight will be fun, but i highly doubt it. Laterrrrr!  &nbsp;  \"..like it\'s where you wanna be..\" '),(20644,3097708,'27,July,2004','2004-07-27 00:00:00','    heyyyyy...i feel like i haven\'t posted in a bit. And i\'m actually in such a better mood then i was the last time i posted, which was maybe 2 days ago. Let me think..yesterday me saint gav n dams went to the mall for maybe 5 hours. As we all realized, we could\'ve went to maryland in that time lol. My bro and the zaps picked us up and we had a really fun car ride home squishing in chris\' backseat. After i got home n showered n everything, danielle, amanda and i went to coursens. Haha was that fun or what. Thank god taylor was there or i would\'ve left right when i got there. Ofcourse girls were trying to start shit, let\'s not get into that though, i might get some more rude comments in this thing. Today i babysat for a bit. Then, like the idiots that me saint sar n gav are, we decided to go to nanuet after we got our nails done. I bet they will all agree with me when i say that  we made some bad decisions . On our way to nanuet, there was a huge accident on the palisades so i just kept shutting my car off in the middle of traffic cuz i was wasting gas. It was maybe really funny. So yeah we found the tanning salon in nanuet, and then went to the nanuet hotel for some pizza. We finally got home to stony point at like 8oclock after 3 hours of being in the car. And surprisingly i wasn\'t in a bad mood when i was driving this time. So yeah we went to matt\'s all star game, and they did win, which is awesome! Me, sar, dams, n britt went to gynos...lol but me n britt didn\'t really stay there,  we walked to Annie\'s instead  to eat. Let me tell you, it was some awesome food. After some begging, katie turner gave in to picking us up,  which definitely saved our asses cause it was getting a little creepy there . Thanks to katie we got home safe. Katie Russo picked me up for a bit and i actually just got in a little while ago from that. We just really like drove around to catch up and stuff, although it seems like we catch up everyday lol. But i\'m really tired. I think i might go clean my room a little bit and then go to sleep. I\'m babysitting tomorrow, just like every other day. You\'ll probably see colin johnny and i at bowline. Until tomorrow!GOodnight everyone.  &nbsp;  \"..all i know it was the best conversation i\'ve ever had..\"   &nbsp; '),(20645,3097708,'25,July,2004','2004-07-25 00:00:00','    ahh i\'m in such a horrible mood right now you all seriously have no idea. Let me think about today.. uhh me saint gav n matt went to breakfast, then i went to gavs family party n then we all went to ross\'. Happy Birthday to him. I\'m gonna watch the Liberty game, it\'s already in the 2nd quarter. Laterrr..  &nbsp;  \"..can\'t you see my front is crumbling down?i\'m sick of being someone i\'m not..\" '),(20646,3097708,'24,July,2004','2004-07-24 00:00:00','    hehehe, just got in, kinda a late night, huh? I\'m hear with Danielle and we\'re just eating some chips and salsa. Today was such a shitty day. Although i did finish babysitting 2 hours then planned. I came home, took a little nap, then picked up dams n gav. We didn\'t do much, but then Danielle and I went to babysit for a little more, and then went to Matts,  where we played with Microsoft Sam  on the computer. And ofcourse we played \"asshole\" with him and his sisters friends. I would say it was definitely a fun time. Cogar and Jimmy picked us up for a bit n we went to this boys house, when he wasnt even home, just so they could smoke out of a bong in his basement. It was a little strange, but danielle and i stayed entertained with his camera phone. We then picked up my brother, and he was a little pissed that i was in the car wit his friends, but was still acting incredibly funny. He\'ll get over it soon i bet. We\'re gonna finish eating this shit and then go up to bed. Can\'t wait for sparkys in the morning!!!!!!!  &nbsp;   \"..i hear you..\"   &nbsp; p.s-hope you like the ashlee quote..haha a little different, but still cool  &nbsp;  &nbsp;   &nbsp; '),(20647,3097708,'22,July,2004','2004-07-22 00:00:00','    once again i don\'t wanna post, i just haven\'t done it in awhile.Tonight seriously sucked. Although i was with jamie n amanda, we were having fun when it was just us. Ugh i don\'t even wanna talk about the rest. But anyways.. i\'ve just been babysitting during the day, and thats never fun. Jamie and I brought Colin to the pool, and were trying to teach him to swim. It was funn. But whatever, i\'m gonna have my nightly devil dog, and then go to sleep. Hopefully tomorrow will be better?..i doubt it though    &nbsp;   \"..is pointed straight at your heart..\" '),(20648,3097708,'20,July,2004','2004-07-20 00:00:00','    heyyy everyone. It\'s really late, i just finished watching Trapped. I must say it wasn\'t a bad movie at all. I\'ve seen it a couple of times before, but i figured i\'d watch it again for some reason. Last night was kinda a boring night. I babysat for a bit, n caitlin came with me, n then after we went out with gillman, keri, and balones for a bit to this boys house. I would\'ve had a better time just babysitting all night i would say. Today i babysat again. It seems as if thats all i ever talk about in this journal, but hey it\'s my&nbsp; summer job. I actually had a basketball game this evening also, so erin picked me up for that n it was a pretty good time,  even though we lost by 25 points.  After our game, erin n i went on a little stalking trip, and then on a date to ciceros. I can\'t complain about that though.&nbsp;I went to damianis for a bit with saint n gav, n then we hung out with gillman n went to wendys. It wasn\'t bad i would say. But i\'m about to go to bed. Maybe tomorrow night will be a little more interesting to write about. So long. &nbsp;  \"..when you\'re down on the floor bleeding..\" '),(20649,3097708,'18,July,2004','2004-07-18 00:00:00','    sup!Last night was Meaghan\'s surprise sweet 16 party. I haven\'t talked about it in this thing obviously because of the fact that she would read it.  It was seriously soooo much fun, and she had no idea at all.  I think everyone would agree with me if i said james, john, and mo were being hilarious. Better yet, i think hilarious might be an understatement. With them continuously on the mic, i think meaghan\'s family was a little terrified, more so james n mo on the mic lol. So yeah that was so much fun. After meg\'s party, i went to britts for a bit, and then we all went to Ross\' for a going away party for James. He\'s moving to Atlanta tomorrow, i\'m gonna miss him soo much! I don\'t know why, i was just in a really bad mood when i was there. Everyone was kinda pissing me off, but i still had fun because it\'s impossible not to have fun when you\'re at Ross\' house. Gavin and I went to Danny\'s after we left Ross\'. We seriously maybe made ourselves die of laughter. Or maybe i was being funnier, so i was making ourselves die of laughter. This might sound a little weird,  but us 3 were actually playing Full House , you know the tv show. Although Danny was telling me how gay i was the whole time, deep down he knew that im just too cool to make up an awesome game like that. So yeah i went out for a bit after danny\'s, and saint was so kind to let me sleep over. I was very thankful for that. I just got in from her house now. I had a really bad sleep cause i shared the bed with dams n meg, need i say more? I\'m gonna probably go to sleep now for a bit after i get something to eat. Laterr everyone! &nbsp;  \"..you\'re not so happy, you\'re not secure..\" '),(20650,3097708,'17,July,2004','2004-07-17 00:00:00','    heyyy everyone. These past 2 days have been sooo much fun! Let me start off by saying that Brittany\'s sweet 16 was soooooo incredible. She looked beautiful, and it was just such an awesome time. It was maybe one of the best sweet 16s i\'ve ever been to. Although yesterday wasn\'t really that good of a day. First of all, i got my haircut, and yes, its ever so short. Then, when i was picking saint up from work, i crashed into the car in front of me at the red light. And, with the luck i have, brianna and riff were behind me  when i crashed  n they were maybe dying of laughter. Thank god i cried to the man so he said it was okay n didn\'t take down any info. I was extremely relieved. After britts brianna, katie, and riff picked up me n gav n we went to seths. It was pretty cool. A lot of ppl were there that i didnt really expect to be there. Danielle slept over afterwards, and ofcourse we didnt get home til maybe 4 in the morning. I can definitely say that it was an  extremely fun night . When we woke up, when it was maybe noon, danielle n i went to pick up megs n we went to sparkys for breakfast, well they had lunch, n then we went to gregs for a bit. I\'m about to get dressed n do my hair n stuff like that.. TONIGHT IS GONNA BE SOOOOO MUCH FUN..hmm don\'t you all wonder? &nbsp;  \"..they shine..\" '),(20651,3097708,'16,July,2004','2004-07-16 00:00:00','    sup dudes. It\'s really late, but i\'m not tired at all. I had a lot of fun tonight. But let\'s see-i babysat this evening, and it was kinda total chaos. Colin decided to go to bathroom on the callanan\'s new rug, n then after him shitting all over it, kiera decides to throw up because its so incredibly disgusting. I think i maybe almost cried because of this. Whatever, i got over it soon though. After babysitting me, dams, gav, britt, sar, danny, n joe went to the bonfire for a bit. After being bored for a bit, me, gav, n britt went to SPE to swing on the swings, even though we didnt do that. Matt Dillon was actually there, which was a real treat cuz i love seeing him, i just miss him so much. Gav, Britt, and I just hung out n did kuma n shit like that. Jamie n Amanda picked us up n then we went to fiesta cancun, which was maybe one of the funniest experiences ever.  Basically the mexicans were tryin to hook up with us , n they kept taking those pictures of us.&nbsp; Haha it was weird, but definitely fun. I\'m gonna go watch some tv with kevin now. I\'m EVER SO excited for britts sweet 16 tomorrow night, it should be tons of fun. I will let u all know how that goes...laterrrr &nbsp;  \"..what do you take me for?some kind of easy mark?..\" '),(20652,3097708,'14,July,2004','2004-07-14 00:00:00','    I can honestly say that this summer is so far, extremely fun. I think i maybe stayed home one night so far, thats it. Although, yesterday was just another boring day of babysitting. After that, i picked britt up n we went to the mall so i can get a dress for her sweet sixteen. I was successful. When we were eating, we saw ross, joe, kevin, james, and john, so we hung out with them for a little bit. They were being extremely funny, like always. Umm yeah so i dropped britt off at her cheerleading meeting n then hung out with dams for awhile.It was a fun time. Last night i hung out with George for a bit, then i went to brians, n then to my friends house.  All in all, it was a really fun night i would say.  Hopefully the people that had my company are able to say the same. It\'s kinda disgusting but i like just woke up. I have to babysit this evening for a bit, and then i have a basketball game which im debating on going to. We will see what happens. Laterr   \"..and i pray for you high health..\" '),(20653,3097708,'13,July,2004','2004-07-13 00:00:00','    i dont even know why im posting right now cuz i really don\'t feel like it. Maybe because i\'m bored and i dont wanna update the past couples of days tomorrow. Nothing really exciting has happened the past 2 nights. Although tonite i did see Anchorman with saint gav dams n sar. It was a pretty funny movie, just a little weird. But i\'m probably gonna go to bed soon, i have another thrilling day of babysitting tomorrow..Goodnight-    \"..or maybe..we will, i\'m dying to know..\" '),(20654,3097708,'11,July,2004','2004-07-11 00:00:00','    i can honestly not complain about yesterday. First off, it was brittany\'s birthday n i hope she had tons of fun. During the day though, it was a little boring. Although i did bring danielle kristina n lauren to coinstar, and then me n danny went to wendys for a bit. Let\'s see..for brittanys birthday me britt cait danielle kristina sarah dams and meg went out to ciceros, followed by hitting the carnival up. It was wristband night, so we made sure we went on the pirate ship as many times as we wished. George picked me britt sar n meg up n we went to Tim Keenans. We all got tanked, maybe sarah a little too tanked, but all in all it was a fun time. I think so at least.I didn\'t get home til 2:15, but thank god my dad was sleeping on the couch so he didn\'t even notice. But yeah that was last nite, i hope tonight is just as fun. I\'m gonna go try to get my moms keys so i can go to the mall. Laterr    \"..i\'m trying to get this right, yeah, cause i\'m ridiculous like that..\" '),(20655,3097708,'10,July,2004','2004-07-10 00:00:00','    hey everyonee..lax camp is finally over-and i\'m happy to say that i won one of the wormburner awards. It was a little fannie pack but still it was cute lol. Yeah so last night was pegs 80th surprise party. It was kinda weak, but gavin came with me so we made it fun. There was a weird witch there, and her daughter helen was like doing spells on me, danielle, and my cousins. Um yeah so after we went to that boy drews house but we werent able to stay there long. George dropped me off at home n then saint came down n my mom actually let me take out the car even though it was 1045, that was extremely nice of her. So Kristina and I just drove around for a bit screaming stuff at people out the windows. We actually went to dunkin and the alarms were going off in there n then soon there were a million cops and a fire truck there. It was crazy-so we just left n went to mickeyds, which was even funnier.  The girl at the window told us we should be on mtv  cuz we were just acting so funny, like singing ashlee simpson into the drive thru speaker, and hitting every curb on the drive thru. Good times, good times. Yeah so after that we just came to my house n watched a little boy meets world, and then kristina left. I actually had a fun night, believe it or not. But whatever, i\'m gonna go. I think possibly tonight i might go to the carnival since its wristband night again. We will see. Laterr   \"..but i\'m still trying..\" '),(20656,3097708,'08,July,2004','2004-07-08 00:00:00','    hey everyone-i just got in from lax n it actually wasn\'t that bad today. We worked really hard before lunch, and then after lunch we watched a dvd n then had a tournament. Pete and these two little boys were on my team and we won our game-ofcourse since i was the only girl i didn\'t touch the ball much, but i\'m happy we won. After lax i picked up matt n danielle n we went to bank n then to ciceros n mr.softees-i needed pizza and matt needed icecream, although i did end up eating both.  ...lets see-last night..i went to the carnival for a bit with some friends. It was kinda gay but i got one ride out of it-i definitely didn\'t bring any money so i just scapped tickets off of people. After the carnival i went to jamies cuz she had some ppl over for a little. I hope the carnival is fun tonight though. I cannot wait-it\'s wristband night lol. But yeah i\'m gonna go take a nap..i\'m so incredibly tired. Laterrr   \"..and when did your eyes begin to look fake?i hope you\'re as happy as you\'re pretending..\" '),(20657,3097708,'07,July,2004','2004-07-07 00:00:00','    although its getting late, i figured i would post before i went to bed. Tonight, Joe had ppl over to swim and stuff, which i found out when i got there. All the guys were swimming, so peluso n i decided to swim with them. With joe\'s shorts and shirts we constructed 2 kool bathingsuits i would say. So all was well while we were swimming, and when peluso n i felt the need to get out we put our jeans back on n borrowed some sweatshirts from joe. I definitely wouldnt have put nice warm clothes on if i wouldve known that john quevedo was gonna throw me in the pool. UGH i was soo pissed, although i did laugh at first. He\'s such a little fag but i can\'t stay mad at him. Because of him though,  i definitely think i have pneumonia .I was sitting outside soaking wet for maybe an hour-it wasnt a good feeling, trust me. But yeah it\'s really late n i have lax camp in the morning-i really wish someone could maybe call a bomb scare to it, then maybe it would be cancelled for a day. lol im so gay..laterrrrr   \"..think of all the times this jerk has fucked you up and let you down..\" '),(20658,3097708,'06,July,2004','2004-07-06 00:00:00','    i don\'t mean to complain, and it seems like its all i\'ve been doing these past couple of days, but i HATE lacrosse camp. I even think that the adjective hate might be an understatement.      lets see-last night, me britt gav saint n dams went to caits for a barbeque since today is her birthday.  HAPPY BIRTHDAY CAITLIN!  It was fun, i had a lot to eat, and the cake her mom made was awesome. After cait\'s, danielle n i went to austin\'s graduation party for a little bit, and then to timmys. I guess it was an alright night, it could\'ve been better though. I had some really bad sunburn, and it was a little noticeable.       well as for today, lax camp sucked-like i said 20 times already. But yeah, i have a basketball game at 630 at pearl river highschool, and everyone who said that they\'re were going to come, better come.  \"..and i\'m not so sure, if i\'m sure of anything, anymore..\" '),(20659,3097708,'05,July,2004','2004-07-05 00:00:00','    kristina stpierre is the coolest person ever...i love her so much. '),(20660,3097708,'04,July,2004','2004-07-04 00:00:00','    heyyyy everyone-omg i just got in from lacrosse camp a little while ago, and let me tell you all,  it was complete hell . I cannot honestly believe that i have 4 more days doing this. Our counselor/coach for the week is a huge bitch also. In the morning, we were all screwing around, and not catching like any passes, it was seriously hysterical and she was getting so pissed. AH but whatever-enough about lacrosse. Last night was a pretty good time. I hope everyone had a fun 4th of July, i know i did.  Joe\'s fireworks were awesome , although i was hiding for quite a bit when they were going all over the place. After a little bit, i walked down to Evan\'s even though hes in Ireland. His brother was having a party, so cait, amanda, and i went with my pretend boyfriend george. But yeah, all night i was back n forth from joes and evans. I would say all in all it was a fun night, now that i look back at it.     It\'s almost 4 oclock and i think im gonna go take a nap. I\'m so extremely tired from this camp, you have no idea. I\'m kinda excited for tonight, i think it will be fun. I might stop by austins graduation party, but i will have to see about that, it might be a little dangerous.Laterrr everyone   \"..let this mistake bring a vivid crystal to his* eyes..\" '),(20661,3097708,'04,July,2004','2004-07-04 00:00:00','    HaPpY 4th of JuLy!I hope everyone had fun last night, cause i did-lol. Well me gav saint kim danny n joe went to the bonfire for a bit so they could smoke, n me n kim just challed. Although i wasn\'t under the influence joe was being extremely funny with his car noises-we were alll dying. Cogar and them picked us up after a bit n we went to the O\'Connors- i had soooo much fun .My brother was a little pissed i was there i think, and i think he was more pissed that his friends brought me there lol-but whatever he obviously dealt with it by just pushing me whenever he saw me. But besides that, it was a good time. After the O\'Connors, we went to Johns to hang out with him, timmy, anthony, and kyle. I was really tired, thats y i didn\'t have much fun there-plus i think timmy was making fun of me the whole time. But anyways-i can\'t wait til the day gets started..i love the 4th of july, and i hope something good goes on. I\'m outta here though-laterrr    \"..i feel like shit, this isn\'t the way to treat old friends..\" '),(20662,3097708,'03,July,2004','2004-07-03 00:00:00','    heyyyy everyoneee last night was kinda ehh lets see.. well i went to coursens with britt and george but they decided to leave so i just stayed there, it was fun when everyone was there, but then i dunno i was just getting in a mad mood cuz of some homos that i hate. But besides that, i was chillin with my boyfriend george n then finally at maybe 2 in the morning reza drove me home, and i was  very thankful  for that. I didn\'t even go inside cuz danielle, jenn, timmy, n anthony were there to pick me up, and we all just hung out n i had fun with them. I got home soo late, i think maybe like 330, it was crazy. So yeah today me n saint went to the mall to go shopping n then on our way home we picked up gav n then went to amandas-but yeah im gonna go jump in the shower so im actually ready when my friends wanna go out tonight-i hope something fun goes on, i\'ll crosss my fingers-laterrrr   \"..enjoy this lonely sky with me, it will swallow us whole if we only let it..\" '),(20663,3097708,'02,July,2004','2004-07-02 00:00:00','    OMG-it has been forever since i posted, due to the fact my computer hasn\'t really been working-but great news..it\'s fixed, finally. SOrry i haven\'t been able to keep everyone updated with my extremely exciting life, and i\'m not gonna even try to go back n tell u what has happened since. I just got home from Wildwood like literally a half hour ago. Caitlin, Brittany, Gavin and I went for Caitlins birthday and  we had soooooooo much fun!!! When we first got there, we went to the beach and met these losers, and ofcourse-my asshole friends gave him my cell phone number so they were calling us all night to go to their party of \"7\" yes, 3 of them and 4 of us..some party huh?that night we went to the boardwalk though- I think we met maybe 100 guys, including some really weird ones. We actually met these workers ( no not carnies) at this store Fantasy, and became like bff with them. They\'re names were Canada and Simon, and simon and I really  hit it off . Canada wanted to show us his \"balls\" lol and Simon was just asking me some sexual questions. Everyone in the store thought danielle was an annoying fuck and wanted to kick her out cuz she was being extremely funny. But whatever, they gave us some free tatoos and wanted to pierce our bodies, which was a little weird. We also met this kid Antonio, omg he was the definition of gorgeous. He called caitlin later on that night and wanted us to go \"clubbin\" with him lol. Weird huh? Today, we went to the beach for a little n then went to the pool and were doing hott moves off the diving board and then went to the boardwalk to go on rides. We flirted our way onto the slingshot and ended up getting $15 off the price lol, and then the tilt-a-whirl was awesome cuz we stayed on for basically a half hour until some girl ralphed all over-funnnny shit.Then, we went out to dinner at this boathouse and we were singing along with the guy that was performing, it was amazing lol. But yeah i don\'t feel like getting into much more detail cuz im tired and i think i wanna go out for a little bit-but i had soooooooo much fun, and i hope caitlin did too!!! K well jenni and dan r gonna come over and challll ttyl!!!   \"..i am seeing in myself the things you swore you saw yourself..\" '),(20664,3097708,'06,August,2004','2004-08-06 00:00:00','    i feel as if everyday i repeat myself over and over. Possibly when summer is over this feeling might go away. Hopefully it does.  I babysat til 6pm today. I think  i\'ll just add also that it wasn\'t fun at all . It really sucks talking to a 2 yr old, and only a 2 yr old for 6 hours. I can\'t use my cell phone much anymore, and no1 ever wants to come over there and hang out with me. What can i do about that? Obviously nothing. I think even if i had a boyfriend, he wouldn\'t even want to come over there. It sucks, really sucks. So right now im being dead serious, if any of you that are reading this, and aren\'t busy tomorrow(Friday) please feel free to call my cell phone, you probably have the number, and invite urself over if i don\'t already do so.  So anyways, after babysitting danielle and i drove around for a bit. Well i drove, she just sat in the front seat and controlled the radio. It was fun, although we always have the same routes, i wouldn\'t be surprised if some people got worried that we\'re always near their houses. But yeah i saw klein n doug while klein was driving so i sped up quick and  we ended up having a race after talking for a bit.  No matter what they say,  the truth is that i won . The quest is a beast, just making sure everyone knows that. So yeah after going home, showering, visiting the stpierres before they leave, i finally went down to gavins around like 10. Danny and Greg were also there, and Joann let me take her car to get some movies. Sometimes i feel as if she trusts me more than my own mom. But whatever, the movie store was closed so we were all just fucking around and i let danielle drive, and i\'ll add that she did well. We went back to gavs and watched  Now and Then . The fact that danny n greg didn\'t see that movie ever before amazed me. Whenever those 2 are together they always find a way to make fun of me, and tonight they were saying i have red hair, which i obviously dont. But i\'m over it now. Greg left and it was just danny danielle n i. And ofcourse danny was pulling some funny business, but it was incredibly funny  just because i\'m so funny(not you ashlee, or danielle-whichever you prefer).  Danielle was getting so pissed at danny and i for not leaving but it was her fault for inviting us over. Shit, Okay before i forget, someone was  robbing her neighbors house , and we saw the whole thing. It was crazy, like honestly. Danielle could\'ve sworn it was the old lady, but it looked like a huge man, but i hope im wrong. So danny and i finally left at like 115 and i made him walk me home cause it was so late. I felt bad, considering the fact its totally out of the way, but he had to walk me home cause  there are bears on my block,  i know it. He told me he wasn\'t pissed, but i know he was, hes just tryin to be a good friend. Thanks danny. All and all it was a hilarious night. Just because my friends are awesome!!!lol im such a dork..   quote of the  night :    \"If you tell anyone about this, especially your brothers, i\'m gonna beat the shit out of you\"- Devin Sawa..  \"Wow, that\'s definitely Danielle..\" -Danny   that was maybe hilarious. Thanks ofcourse to now and then, and ashlee and mike. hahahahaha! so yeah im gonna find some food and then go to bed.  Please  take me up on that offer on visiting me while babysitting tomorrow. Goodnight everyone!         \"..i\'m already torn..\" '),(20665,3097708,'05,August,2004','2004-08-05 00:00:00','    once again, another boring day of babysitting..when i got home though i just went to sleep for a little bit and i woke up to my cell phone ringing, and it was my dad. I went downstairs and waited for him to come home cause he said he needed to talk to me, and he told me that i went $184 over my cell phone bill.  I felt extremely bad , cause last month i went $72 over. He warned me, next month if i go 5 bucks over, then my phone is terminated, so please don\'t call me too much everyone..Umm, after my talk with my dad, i got ready to go out, n right when i was about to leave, i got a call from carmita asking me if i can go to the callanans because the other babysitter didn\'t show up. I was extremely mad, especially because i spend most of the day there, n then i had to go back that night. Ugh whatever, i had to do her a favor though cause she always does me favors. So when i got home, my mom drove me to cjs, but i just ended up goin to wilsons with gav laur n jenn. I think i maybe broke kyles leg in the front seat cause we were all soooo squished, and ofcourse i ended up in the front-::sighs:: ever so dangerous ..but what can you do. On the way home it was a little more comfortable. Danielle, Timmy and I were all in the back holding hands. Like honestly,  i\'m so funny sometimes i can\'t take it lol . Yeah but i\'m gonna go get some milk and cookies. Tomorrow should be sooooo much fun though, i\'m babysitting until 6oclock, i think i might kill myself. By the way, feel free to stop by.     quote of the day:  \"Yo Kel, wanna take your 5 hr with me tomorrow?\"-Kim Jordan..      Shes maybe a little retarded.And no kim, my mom is never hiding in the back of the car when i pick you up, the license i have is real.        \"..well just get dressed, don\'t do this..you can\'t decide..\" '),(20666,3097708,'04,August,2004','2004-08-04 00:00:00','    last night was soo much fun. Jamie, Gavin and I ended up seeing The Village at like midnight. We were being hysterical in the movie theatre, and every guy near us was like dying laughing at us. After that i just came home and went to sleep, and surprisingly i woke up at 130 in the afternoon because i didn\'t have to babysit.  How disgusting is that?  I never sleep that late, i guess i was just catching up on the sleep that i haven\'t been getting lately, but im sure i\'ll be having more now.  Haha, i\'m too funny.  I bet none of you get that anyway. But whatever..when i finally woke up, i went down to gavs, and me, her, jenn, and kim watched the butterfly effect. It was pretty good, although i did miss half the movie. I went home to run some errands, and then drive around with jenn kim n gav.  Gavin and I ditched them  so we could go get a card for jenn\'s surprise birthday dinner tonight at fiesta cancun. It was soooo much fun, and she was ever so surprised!!We went to coursens for a tiny bit, and then cjs, and then right to James Hickeys. We stayed there for awhile, and i personally had a lot of fun. I\'m talking to kim right now online and i don\'t think shes drunk, but shes being incredibly funny.   Check this out:     kim 23054:    BORIQUA     kim 23054:    DOMINICANA     kim 23054:    thats u     kim 23054:    :-[ but ur not tan...ur burnt     who says it?apparently kim jordan--even when she is sober.       \"..why can\'t you just be happy?and i don\'t think i\'ll come back down..\"                            \"...CAN\'T YOU LIVE WITHOUT THE ATTENTION?..\" '),(20667,3097708,'02,August,2004','2004-08-02 00:00:00','    so yeah yesterday me, dams, danny, n greg went to the mall for a bit, they saw a movie, n dams n i just left after seeing if old navy had any more white flipflops. No they didn\'t so don\'t waste a trip to the mall like we did. Sar, meg, dams, n i went to ciceros for dinner, and then i picked them up after we all showered, went to hoyers, then to my house, then to james hickeys. It was alright. Not really my kind of people, but i wasn\'t complaining. Umm gillman picked me up from there, which was incredibly nice of him because i really wanted to leave. Danny and I were supposed to hang out cause i was a little upset, but he was a fag n went to steven lonergans instead, so basically he ditched me, even though he swears i was going to bed. I just got in from babysitting and i\'ll probably take a nap,  and then hang out with my friends . Jamies home from vacation, so it will be fun seeing her tonight. I\'m outta here..laterr       \"..i\'m lying just to keep you here..\"                                                                                                                     \"..Tonight Won\'t Make a Difference..\" '),(20668,3097708,'01,August,2004','2004-08-01 00:00:00','    okay yesterday was such a busy day. Gavin and I were driving around for a bit, went tanning, and then kidnapped Austin cause we wanted him to buy liquor for us for last night. He was a little mad when we wouldn\'t bring him home, but it was maybe hilarious listening to him yelling. But whatever, he knows he loved every minute of it. lol..So yeah amanda, gavin, and i went to the city to see  a show at the knitting factory. Let me just tell you, we made friends with everyone in each band. It was definitely so much fun. I think i definitely have a crush on Zach from Life Before This. He was such a hottttttie. When we first got there, they wouldnt let us bring our \"water bottles\" in so danielle  stuck the liquor in between her legs under her skirt  n we walked in..it was very interesting how she did that. But whatever, every like 10 minutes we went to the bathroom to drink, them a lot more than me, cause i hate the taste of that shit. But i think we can agree with eachother saying that we were basically trashed. After the show, we went out to dinner in the village at Arturos. This was maybe at 1 am, and u know when im drunk i fall asleep, i was sleeping on the way there. It was horrible. We ate soooo much, like honestly there was an incredible amount of food. Gav decided to go to the bathroom at this restaurant, and screamed to everyone how there was a bathtub in it, which was a little weird, and then started  tagging up the chalkboard in there with all our names . She was tanked. We finished eating around 2am and then walked around the village. Let me just tell you.. true fact: THE CITY NEVER SLEEPS.  I was petrified walking around, i seriously thought we were gonna get shot. I think i need to get over that stuff soon. Ahh i almost forgot.. i talked to my husband James , thank god for john amoroso giving me his number, but he called me and told me how he was in the hamptons so i didnt get to see him. Definitely next time though..!Me and danielle got back to my house around 345 to find we were locked out, but we finally got in through the garage. But anyways, saint, pel, jenn, kim, danielle, and i just went to fiesta. That was fun, especially pel just got back from mexico, it was awesome to finally see her after sooo long. But i\'m gonna go lay around, hopefully something fun goes on tonight. I feel as if i say that in every one of my posts?hmm.. Later!    \"...i could go all night..\" '),(20669,3724910,'23,June,2004','2004-06-23 00:00:00','    Rightio so after setting this stupid thing up im ready for entry number 2.  During the holidays I went on a 12 day seakayaking/rafting expedition with school.  This was seriously hardcore, we dropped in at a place called farm house creek up on the picton river.  Bloody cold but.  So yea anyway day was spent driving as far in as we could, then we had to get all our gear down to the river, set up camp, inflate the rafts ect.  Most of the 18 of us were under tarps the whole trip, however I took a tent for the first 5 days of the rafting, and ditched it when we got the sea kayaks.  Anyway, so yea we had a quick de-brief that night, where we talked about the leaders for that day, and our goals for the rest of the trip ect.  Day 2 we started rafting, and was one of the biggest days we had.  The water level was high so we had a quick ride.  However we took ages getting ready as it was our first day packin the rafts ect. then we had the safty talk, then we set off, with our eye on tahoon, however that was quite a long way away.... Turned out we were actually able to make it to Tahoon which was great because we were able to sleep in the hut there curtosy of the staff at the airwalk.  Day 3 was epic, we had to find a campsite SOMEWHERE along the river, and the front raft managed to find one by accident which was great. However it had this hugely steep bank to get to the site fromt the river so we had a handline which made for lots of fun....  So yea that was a bit of a weird site, quite cramped, and probably one of the shittest, but hey it didn\'t stop us.  That arvo we planned a walk down some foresty roads.  Anna, one of our big guides took us to a place which is being reserved which is great.  There were man ferns there like 5 metres high.  Walking back was great there was a huuuge full moon which we all howled at like wolfs, i wont ever forget what happened next.  We saw something I swear happenes once in a very blue moon.  We saw a night time rainbow, I kid u not.  The most amazing thing I have ever seen....  To day 4.  This day we didnt have that far to go, however the river was widning up so the water was moving slower than usual.  We ended up at a place we called Clancy\'s.  A paddock owned by a friend of Anna\'s.  On the way however we went up to southwood, a new wood processing plant to be made by the government.  not such a fun thing for anna, the conservationist, but we ended up even getting a tour?!?!  so that was a highlight. then as I said we pushed on to clancy\'s.  We set up ok, until Matt made us do what he called a \"group inititive activity\". this involved us setting up the tarps properly basically, so we could have another de-brief of the last few days.  I slept in my tent for the last night, which was warmer I\'m sure, because as we awoke on day 5, my god i havn\'t been that cold ever....  Day 5.  Ok i\'m talking serious cold....like u would not believe.  There was snow on low lying surrounding peaks.  Getting into a damp wet suit....my god it makes my spine shiver thinking about it.  neway back to the story, we had to make a place called rannala by 3 o\'clock to get our sea kayaks for the next part of our trip.  tho i think ill make that a seperate log.  So yea we got there in good time, after a quick run in with an electic fence.....dont ask.  anyway it was so great to get re-supply.  my 5 day old gurken dip was ready for the bin, and it was so good to get a fresh jumper....my god.  so yea it was nice, new food, rubbish drop, it was like starting over....  we then went into a nearbye paddock to set up one of the sickest tarps in the history of the world.  we lashed paddles together and stuff.  and then piched it against a fence.  that was a good night there, and my first under a tarp which was supprising great.  it wasnt too cold and it ment we had lots of room to move about and things.  so yea we had one more de-breif that night, after cooking up a beautiful meal, and had a good sleep.  ill leave the sea kayaking to the next post.  ....to be continued   '),(20670,3724910,'23,June,2004','2004-06-23 00:00:00','    G\'day all, I\'m Josh, more commonly known as Squish a nickname I aquired  a long time ago from my uncle.  Anyway welcome to my blog....This being  my first post I thought I might share my brief life story with you.  I  was born in Hobart, Tasmania way back in \'86.   Until I was 8 I lived  \"high on a hill\" up on Mt. Nelson, had a brillient childhood and went  to Mt. Nelson Primary school.  Bloody oath that school is tops.   Anyway, we had a good time there, lots of family holidays to Coles Bay  and visits up to the farm, then in 1994 or so we moved to Albion  Heights Drive.  I love my house so much, and we\'re still there.  I then  went to Friends\' High School and now im in Clemes in Grade 12.  I have  my license, and my car, a 1987 Honda Civic, 1.5 L, 5 speed manual.   Little pocket rocket it is too.  Life is good.  Plans at the moment for  the future are as follows:  Finish school, do well, get a score close  to 90.  Then I will be off to England working a GAP year there.  I will  hopefully do this with school and go to another Quaker school outside  of London called Saffron Walden.  If thats not possible, I do have a  interview with GAP this friday, so hopefully one of those will work  out.  After my year away I will be back in Tassie, get my job back or  get another, then start my degree at University studing Agricultural  Science.....after that what will I do.....your guess is good as mine.  Squish   '),(20671,3951300,'31,July,2004','2004-07-31 00:00:00','   I was with my mom riding to the store one night when I saw a guy, shabbily clothed and long dirty hair with a cap on to cover the top, and he was about to get on his bike, and I watched him as my mom was pulling in and he was putting his bags of food in his baskets located on the front and back of his bike and he was making sure everything was just right, and I felt a sadness dwelling in me, I felt his pain of no one knowing he even existed, everyone in their nice cozy vehicles passing him by as he tried desperately just to make sure he wouldnâ€™t lose anything upon his journey to his home or wherever he was going for that night. I think it is so sad that we Christians can be so Holy and so pure but we canâ€™t even stop to help someone who isnâ€™t so holy and so pure, and some times it makes me mad to know that we can go spend money to read our nice Christian books by great Christian authors but we canâ€™t even help someone who is tying bags to the rails of his baskets, so that on a bike ride home they will be safe. I think we American Christians have forgot the values of Christianity, I think we have forgot what Christianity is and why we are called Christians, and as we pop the trunks of cars to put in our nice good neatly packed groceries, we forget the guy on the bike is taking 15 minutes just to make sure he can keep his safe, because he canâ€™t lose anything, because when he loses something it may cost him more than just money it make cost him 3 or 4 maybe even 5 meals that were suppose to last him through the week. So go and live your American Christian lives the way you want, but I choose to live mine like Jesus, because Christianity is about being like him, not like Billy Bob or sally sue.  May God Bless You and Keep You, Bradley Floyd  '),(20672,3951300,'29,July,2004','2004-07-29 00:00:00','   Well I really love this story, it\'s one of my favorite\'s, so I hope you enjoy as well...  I remember, you know, you go to these parties on Saturday night and people would say, about 8:00 on Sunday morning \"Whoa, I have to go - I gotta get to church!\" People would say, \"Why do you want to go to church, all those hypocrites.\" And I say, \"Look, why do I want to stay here with all you hypocrites?\" (audience laughter) I never knew why going to church made you a hypocrite. They\'d say because you go to church and you\'re all \"Holy, Holy, Holy\" for two or three hours, and then you go home and sin. I\'d say \"exactly!\" (audience laughter) For two or three hours you\'re doing pretty good! (audience laughter) Maybe the problem isn\'t that you go to church, maybe the problem is that you go home!  I never understood why going to church made you a hypocrite either, because nobody goes to church because they\'re perfect. If you\'ve got it all together, you don\'t need to go. You can go jogging with all the other perfect people on Sunday morning (audience laughter). Every time you go to church, you\'re confessing again to yourself, to your family, to the people you pass on the way there, to the people who will greet you there, that you don\'t have it all together. And that you need their support. You need their direction. You need some accountability, you need some help.  Saying someone is a hypocrite is such a sore thing to say. It\'s kind of like saying they\'re human. We know there are people who think better of us than we know they have reason to. Funny thing about the whole thing. I love the story about Jesus calling Nathaniel. I think it\'s so funny. That he calls him and he says, \"O behold an Israelite in whom there is nothing false.\" And Nathaniel, I love this, because he\'s like \"Oh, how did you know me?\" And Jesus said \"Look, before you ever saw me, while you were still sitting under a fig tree, I saw you.\" And Nathaniel, who had so casually met the Lord, suddenly was crushed by the knowledge that God has of him.  I think it\'s funny that when Adam and Eve sinned, they covered themselves with a couple fig leaves. Then Nathaniel, who I think the Lord said sarcastically \"O behold an Israelite in whom there is nothing false,\" he was sitting under a whole tree of those leaves. I think we all do that too. And we have reason to. If you knew what I was really like, you wouldn\'t pay to be here. I\'d have to go out and get a real job (audience laughter).  The thing that\'s so funny about all that though, is that God does see us as we are. And you can\'t fool Him. And what\'s even funnier - you know the story about Jesus and the rich young ruler? I love that story. And you know what I think is really funny is that in all three of the synoptic Gospels, that story immediately follow the story of Jesus blessing the children, which I think is a hysterical story, because you think about the apostles and how they\'re good Hebrews and they\'re good monotheists and they\'re probably really struggling with the idea that God can become flesh. Lucky for Jesus they weren\'t Greeks - they never would have bought that.  They finally begin to get a hold of, they get a grasp that maybe God can become flesh and dwell among us, maybe God can be a man, and then they come back and not only is God a man, but He\'s acting like an idiot! He\'s hanging out with a bunch of kids. He\'s blessing them, you know, and you think \"How do you bless children?\" Well, the best way I know is that you pick them up and you just throw them as high as you can, and you catch them right before they splatter (audience laughter). You get down on all fours and you run around the room and you let them ride you and you buck them off. You beard them. You put your mouth against their bellies and you make funny noises. Here\'s Jesus probably doing all this business.  His disciples were humiliated! And they said \"You should not be making such a fool of yourself!\" And I love this, Jesus says, \"Here, look, look fellas. I\'ll call the shots here. I may be dumb, but I am God. And I\'ll tell you what else, if you wanna come into my kingdom, you\'ll come in like one of these or you won\'t come in at all.\" What is remarkable is that in all three synoptic Gospels, the rich young ruler comes up immediately afterwards and asks the stupidest question in the Bible, which American Christians in the twentieth century perpetually ask, and it\'s such a hack off. \"Good Master, what must I do that I might have eternal life?\" And the reason this is such a ridiculous question is because while this young ruler is coming up with this brilliant question for his big photo-op with the Master, the Master had already answered the question.  Jesus had just got done saying he must be like a child, or he cannot come into My kingdom. And this guy comes up and asks \"what must I do?\" It\'s not what you do. It\'s what you allow Christ to do. This young man was so arrogant that when God Himself spoke, he didn\'t bother to listen.  And here\'s the amazing thing in the Gospel of Mark, which you know is the briefest of the Gospels. And anytime you come across a detail in the book of Mark, take note of it, because he didn\'t waste time. He gives us three details that the other two synoptics leave out. He says this amazing thing - that Jesus looked at the man. Jesus, who was God, is so humble, that when this man, who was a mere mortal, who would like you and me someday be dead and no more, and would just stink up the ground for a while and then that\'s it. This man would not listen to God, but God is so humble that He took note of this man. He looked at him. And when God saw him, He loved him. And it was because God loved Him, I think, that He spoke.  Some of us are so afraid that God\'s not going to look at us. So we\'re out there doing all sorts of things to get God to take notice, but folks, God notices you. The fact is, He can\'t take His eyes off of you. However badly you think of yourself, God is crazy about you. God is in love with you. Some of us even fear that someday we\'ll do something so bad that He won\'t notice you anymore. Well let me tell you, God loves us completely. And He knew us at our worst before He ever began to love us at all. And in the love of God, there are no degrees, there is only love.  So tonight I\'m going to say to you what the angels said to every character in the Bible that they encountered, except for Mary, the mother of Jesus. They said \"Don\'t be afraid.\" We\'ve got a little while to go yet in this life, and it\'s a scary thing, but don\'t be afraid. Be of good cheer. He has overcome the world. And He has chosen to dwell within us. And we ain\'t all that big a deal, but our Savior is. He will walk with us through this life, and when it\'s over, He will raise us up again, and take us to be where He is. Not because of what we\'ve done, but because of who He is. Because of the love He has for us.  So go, and live in that awareness... love one another... read your Bible... wash your dishes... make your bed... and don\'t be afraid.   May God Bless You and Keep You, Bradley Floyd  '),(20673,3951300,'28,July,2004','2004-07-28 00:00:00','   Well I said I wasn\'t going to post I was just going to leave you with Rich Mullins, but tonight as I began to pray for my mother and her court case, I began to look up scriptures and all of a sudden I was becoming consumed with just trusting in God, and so I decided to post this hymn along with many, many scriptures that may be of help to you, they certainly are to me.  May God Bless you and Keep you, Bradley Floyd  I have decided to follow Jesus; (3x) No turning back, no turning back. The world behind me, the cross before me; (3x) No turning back, no turning back. Though none go with me, yet will I follow; (3x) No turning back, no turning back. Will you decide now to follow Jesus? (3x) No turning back, no turning back. From the song \"I have decided to follow Jesus\"(Composer unknown/Indian folk melody)   -----------------------Scriptures------------------------------------ The Anointing breaks every yoke ... and the yoke will be destroyed because of the anointing oil.Isaiah 10:27   But Jesus looked at them and said to them, \"With menthis is impossible, but with God all things are possible.\"Matthew 19:26   Ask, and it will be given to you; seek, and youwill find; knock, and it will be opened to you.Matthew 7:7  ... casting all your care upon Him, for He cares for you.1 Peter 5:7  But those who wait on the Lord shall renew their strength;they shall mount up with wings like eagles, they shallrun and not be weary, they shall walk and not faint.Isaiah 40:31  Yea, though I walk through the valley of the shadowof death, I will fear no evil; for you are with me;Your rod and Your staff, they comfort me.Psalm 23:4  Praise the Lord! Praise the Lord, O my soul!Psalm 146:1  Be strong and of good courage; do not be afraid, nor be dismayed,for the LORD your God is with you wherever you go.Joshua 1:9  Then she came and worshipped Him, saying, \"Lord, help me!\"Matthew 15:25  For where two or three are gathered togetherin My name, I am there in the midst of them.Matthew 18:20  ... I am with you always, even to the end of the age.Matthew 28:20   If God is for us, who can be against us?Romans 8:31  Until now you have asked nothing in My name.Ask, and you will receive, that your joy may be full.John 16:24  In God I have put my trust; I will not be afraid. What can man do to me?Psalm 56:11  This hope we have as an anchor of the soul, both secure andsteadfast, and which enters the Presence behind the veil.Hebrews 6:19  Oh taste and see that the Lord is good;Blessed is the man who trusts in Him!Psalm 34:8  Who is this King of glory? The LORD strong and mighty, the LORD mighty in battle.Lift up your heads, O you gates! Lift up, you everlasting doors! And the King of glory shall come in.Who is this King of glory? The LORD of hosts, He is the King of glory. Selah.Psalm 24:8-10  Don\'t worry about anything; instead, pray about everyting.Tell God what you need, and thank him for all he has done.Philippians 4:6 (New Living Translation)Be anxious for nothing, but in everything by prayer and supplication,with thanksgiving, let your requests be made known to God.Philippians 4:6 (New King James Version)  It shall come to pass in that day that his burden will be taken away from your shoulder,and his yoke from your neck, and the yoke will be destroyed because of the anointing oil.Isaiah 10:27   And Jesus came and spoke to them,saying, \"All authority has beengiven to Me in heaven and on earth.Go therefore and make disciples ofall the nations, baptizing them inthe name of the Father and of theSon and of the Holy Spirit, teachingthem to observe all things that Ihave commanded you; and lo, I amwith you always, even to the endof the age.\" Amen.Matthew 28:18-20  Stand fast therefore in the liberty by whichChrist has made us free, and do not be entangledagain with a yoke of bondage.  I have Decided to follow Jesus, there is no turning back, I have decided to trust in Jesus, he will prevail, he will prevail!!! Put your Trust in Jesus, for he will prevail!!!- Bradley Floyd.  '),(20674,3951300,'27,July,2004','2004-07-27 00:00:00','   My mother has day 2 of court tomorrow, And I really would appreciate your prayers, today was not a good day for her. But I believe with your prayers and god\'s help that she will be able to get through this court case, and hopefully she will win, nevertheless the LORDS WILL BE DONE.  God Bless, Bradley Floyd  P.S. Please keep me in your prayers as well. Thanks  '),(20675,3951300,'27,July,2004','2004-07-27 00:00:00','   Well I would just like to say that I have decided to take a break from my post and giveway to a great storyteller and singer Rich Mullins. Rich Mullins died in an automobile accident back in 1997. It was a loss to the Christian World. Rich Mullins is most familiar for his song \"Awesome God\" and&nbsp; \"Elijah.\" The following is from a concert in 1996, I hope you will find time to read this, it\'s pretty funny. Anyway\'s God Bless...  I went to this church one time where they wanted me to dance. Those churches always scare me (audience laughter). That\'s why I so glad there are denominations. I think it\'s wonderful, if you want to dance that you can go to a church that does that, I just don\'t want you to do it at my church. We just like to sit still. Anyway, I went with a friend of mine. We got there, and everyone\'s dancing around, and I was doing my Hoosier thing (audience laughter). This lady came up and said \"Dance brother, don\'t you have the Spirit?\" And I went, \"Wow, you know, I didn\'t realize that dancing was one of the fruits of the Spirit... or any of the gifts either. I didn\'t know that was the equivalent of being spiritually vital.\"  And she went, \"Well, David danced.\"  And I said, \"Well, right lady, in his underwear! How far do you want to take this?\" (audience laughter) So I think dancing is fine for those of you who like to do it. The most reiterated command in the whole Bible - you might not know this, but the most reiterated command in the whole Bible is the command to sing.  So, I figure when you find a command that is easy to follow and fun to follow you should follow it a lot, because not all of them are that easy. And certainly some of them are less fun than that. So, in the effort to up your chances here tonight, we\'re going to ask you to sing. This is kind of a song about growing up in Indiana, but it\'s also kind of a song about living in the presence of the God who is eternal, and who doesn\'t change. If you want to sing this, please do.  '),(20676,3951300,'27,July,2004','2004-07-27 00:00:00','   Well this is kind of a long entry, but it is worth the read... God Bless...   I remember one time, Beaker and I were hiking in the Appalachian trail and he met some friends of his and they all went in... (Rich immediately rewords himself)... they were all talking about school which I didn\'t want to talk about because I\'ve been in school for so many years it\'s not interesting any more... (Rich pauses while audience laughs)... and uh, of course that happened after third grade... (more laughter from audience as Rich continues)... so I walked into town. It was about a five mile walk from the campsite down in... down the trail... down into town... and when I got there, I went in and was having a steak, and this guy started talking to me and we had this great conversation; we were having a good time and he said, \"Hey look, it\'s dark and it\'s five miles up the road to your campground. Why don\'t I drive you up there?\"  And I said, \"Hey, great!\"  And uh, so we got in his car and just as we pulled out from under the uh, last light in that town the guy said, \"You know what, I should probably tell you that I\'m gay.\"  And I said, \"Oh! I should probably tell you that I am a Christian.\"  (audience applauds, but Rich continues...)  \"...and Christians and gays are really a lot more alike than most people think.\"  And he said, \"Well, what do you mean?\"  And I said, \"Well, I think that gay people are people who desperately need to be loved. And I think Christian people are pretty much the same.\"  (audience applauds loudly)  And he said, \"Well if you want out of the car...\"  I said, \"Why?\"  And he said, \"Well, I\'m gay and your Christian.\"  I said, \"It\'s still five miles and it\'s still dark.\" I\'m not crazy.  (laughter from audience)  And he said, \"I thought Christians hated gays.\"  I said, \"That\'s funny, I thought Christians were supposed to love. I thought that was our first command.\" (audience applauds loudly again)  He said, \"Well, I thought God hated gays.\"  I said, \"That\'s really funny, because I thought God was love.\"  And He said, \"Well uh, do you think uh, God sent AIDS down to punish gay people?\"  And I said, \"Well, in the same sense that he sends bankruptcy on farmers and presidents on voters.\"  (audience laughs at presidential remark and Rich pauses before saying...)  \"...that there are consequences for our choices and sometimes we have to bear the consequence of other people\'s choices.\"  And then he asked me the big one. He said, \"Do you think I will go to hell for being a gay?\"  Well, I\'m a good Hoosier... (audience laughs) ...and I puckered up to say, \"Yes, of course you\'ll go to hell for being gay.\" Now folks, I will tell you the truth. I got ready to say, \"Yes, of course you\'ll go to hell for being gay,\" and when I opened up my mouth it came out, \"No,... of course you won\'t go to hell for being gay.\" And I thought, \"Oh my god, I\'ve only been in New Hampshire for one week and I\'ve already turned into a liberal.\" ...(audience roars out loud in laughter)... What am I going to tell this guy now?  You know there are times when we think we are God\'s messengers, when we think we are the mouthpiece of God. And God really needs fewer mouths than he needs ears... (audience cheers approvingly)... And there are times when we think we\'ve got it all together. And we meet somebody. And they ask us the right questions. And we get to find out something about God that we learned when we were little because our moms and dads read it to us, and because they took us to Sunday school and we were told over and over the great truths of the scriptures. But we forget them because we get all caught up in how cool we are.  (Getting back to his story, Rich concludes...)  I said, \"No, you won\'t go to hell for being gay any more than I would go to hell for being a liar. That nobody goes to hell because of what they do. We go to hell because we reject the grace that God so longs to give to us... (audience applauds)... regardless of what we do.\"  '),(20677,3951300,'26,July,2004','2004-07-26 00:00:00','   Hebrews 11:1&nbsp;&nbsp;&nbsp;Now faith is the substance of things hoped for, the evidence of things not seen. Break Down - \"Faith is being sure of what we hope for and certain of what we do not see.\" - NIV  I usually don\'t care for the NIV&nbsp;Bible but I thought I would break it down with a NIV verse, because this certain verse is actually pretty close to the KJV 1611 edition.  I like this verse, it tells very quickly and very deeply what faith is. Faith is always misunderstood. You see faith is not about saying \"I know God will\", no! Faith is being absolutly sure of that which you have hoped for, and even though you may not see it yet, that it will come.  LoL I hope you get the point lol, just a short post for Monday  May God Bless You and Keep You, Bradley Floyd  '),(20678,3951300,'25,July,2004','2004-07-25 00:00:00','   Well I usually don\'t write much but I love reading and today I notice that Sunday\'s paper is always much broader and larger and so I thought hey I think I should do that on Sunday\'s Post, so that\'s just what I will do.&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; :)   So instead of 1 story I am going to give you 2.   My&nbsp;Sleeping Leg and Sleeping Spirit  Well when I started this post my leg started&nbsp;falling asleep lol. I like to assume the indian position while typing and it can sometimes cut off circulation to my left leg. I know your thinking what on earth does that have to do with God? LoL. Well Actually allot more than you may realize, see just like my foot, when our need for God becomes inactive our spirt becomes sleepy and when it falls asleep it&nbsp;becomes hard to wake up and can hurt in the healing process. So if you feel that you are falling asleep with God then I suggest assuming a new posture.   Getting the Picture...  Today My mother took a paint job, for a little while the picture looked great, but then she started painting the kid in the swing and from far off he looked disfigured, but from close up he looked great. I think we&nbsp;to often look&nbsp;up close at the picture in our lifes and don\'t want to change it but God is telling us to step back and see the&nbsp;disfigurement and correct it. All the Time i have friends who wonder what&nbsp;God&nbsp;has instore for them and they&nbsp;fail to see that God has a picture perfect life instore&nbsp;for them, but first they have to change.&nbsp;   May God Bless you and Keep You,  Bradley Floyd  '),(20679,3951300,'24,July,2004','2004-07-24 00:00:00','   Among ancient Greeks the runner who won the race was not the man who crossed&nbsp;the line in the shortest time, but the man who crossed it in the least time with his torch still burning. We are so often so busy with life\'s activities that we are in danger of allowing the torch of our spiritual life to become extinguished. A good woman once said that in the rush and hurry of her life she felt in danger of being \"jostled out of her spirituality.\" There is a real danger of being too busy to be good, of running too fast to keep our torch burning.   This week I have been real busy, busy with family and busy with band stuff and busy with the website you know and I know that they are spiritual thing\'s i am doing but yet my focus is not on keeping the fire lit but getting finished on time, it can really be a struggle.   So as we live our busy lives let us remember to slow down a little bit and put a little time in for god. I just finally got a second to write this blog lol. Now it\'s time to have some God Time.   May God Bless You and Keep You,  Bradley Floyd  '),(20680,3951300,'24,July,2004','2004-07-24 00:00:00','   Hey the site is still under construction but you still should check it out :).  Especially the Atom Archives area it\'s got a NewsTicker :) still playing around with the .xml file though :) well god bless  '),(20681,3951300,'23,July,2004','2004-07-23 00:00:00','   You are familiar, no doubt, with one of the most famous paintings ever done by any&nbsp;artist: The Last Supper by Leonardo da Vinci, that classic portrayal of Christ and the twelve apostles at the table.  Many students of art history believe that the painting, when first created, was somewhat different from the version which we now see. There was initially, it is believed,&nbsp;an exquisite lace border on the tablecloth. When, immediately upon completion, daVinci invited a group of art students to view his masterpiece, they were immensely&nbsp;impressed by the delicate design of that lacework. They studied it intensely and praised&nbsp;it highly.  Upon seeing the reaction of these young men, the artist took up a brush, dipped it,&nbsp; and made a few long strokes across the canvas, obliterating the lace. Then, with uncontrollable feeling, he shouted, \"Now, you fools, look at the face of Christ!\" Many times we look for Christ only for the good things he possess, just like the exquisite lace border on the tablecloth, we get traped in the good things around Christ that we forget to&nbsp;notice Christ.  '),(20682,3951300,'22,July,2004','2004-07-22 00:00:00','   546 times is the word love found in the bible. That\'s 546 more reasons to use it :).   Mark 12:30 \"Love the Lord your God with all your heart and with all your soul and with all your mind and with all your strength.\'[ urlLink 1 ] 31The second is this: \'Love your neighbor as yourself.\'[ urlLink 2 ] There is no commandment greater than these.\"   Pretty much that means to love God and to Love people. So go out and use that word!!!!  May God Bless You and Keep You. Bradley Floyd  '),(20683,3951300,'22,July,2004','2004-07-22 00:00:00','   Well I have been really Busy but I can officially blog again lol. The site is activated and will soon be up :). That\'s all have a good one  May God bless you and keep you, Bradley Floyd  '),(20684,3951300,'18,July,2004','2004-07-18 00:00:00','   Well as of late King David has really been in my thoughts, I just got done reading  PURPOSE DRIVEN LIFE  and a lot of times it talks about being more like Noah and Moses, and donâ€™t get me wrong theyâ€™re great guys and we should definitely be more like them, but also we should be like King David, he had an ambition that all the other ambitions covered up. We can call it his Secret Ambition ;). He had an ambition to have godâ€™s heart, a pure, loving, merciful heart, and although he failed many times to satisfy Godâ€™s quench for him to be clean and his own quench to be clean, he never gave up. The funny thing with David is, is he could do a very sinful thing one day, and then dance with zero dignity for God the next day. And thatâ€™s just what I want to do I want to be undignified for God. I want to be able to break the shackles off of me so I can dance. Godâ€™s purpose is not to watch you do bad and then whip you for it. NO!!! Godâ€™s purpose for you is for you to learn that he sent his son to die for&nbsp;you and to truly know what that means, also to not be embarrassed to talk about and to stand up for your God. Let us be like King David, let us break our shackles off and dance before the lord.   Irish Psalm -  Dance as if no one is watching  Sing as if no one is listening  Live like youâ€™ll die tomorrow, and die knowing youâ€™ll live forever!!!   I love that Psalm, but I can love it till I turn blue in the face, but until I live it, itâ€™s just a phrase.   May God bless you and keep you,  Bradley Floyd  P.S.  Be Godâ€™s Undignified servant!!!  '),(20685,3951300,'17,July,2004','2004-07-17 00:00:00','   urlLink http://thedigitalduck.blogspot.com/2004/07/forgiveness-from-sky-down.html  '),(20686,3951300,'17,July,2004','2004-07-17 00:00:00','   Brad Floyd - Will be helping Third Street Baptist out with VBS.  Location Kennett MO.   (Non-Preformance event.)   July 26-30.    Want to Schedule something for this week?&nbsp; July 25 and 30th are available. Email Jennifer at  urlLink sabshedule@netlane.com &nbsp;. thanks  '),(20687,3951300,'17,July,2004','2004-07-17 00:00:00','   Well I was really hoping this post would be an awesome Update, such as the bands domain being activated! But sorry to say it is not.&nbsp;&nbsp;&nbsp;&nbsp; :(&nbsp;&nbsp;&nbsp;&nbsp; Well in other news, in case you didn\'t read in an earlier update our scheduling service is up, please if you would like to set up a date and place e-mail&nbsp;Jennifer at  urlLink sabshedule@netlane.com &nbsp; please keep this strictly for Scheduling, all other mail can be sent to  urlLink sabinfo@netlane.com&nbsp; &nbsp;Thanks. &nbsp; Hey, on the next post I\'ll put up not this weeks but next weeks schedule for the band or band members.  &nbsp; Anywayz that\'s all I got. &nbsp; May God Bless You and Keep You, Bradley Floyd  '),(20688,3951300,'17,July,2004','2004-07-17 00:00:00','   Your love is perfect When I love you I have no Fear Your love is perfect When I love you I know you are near  Your love is perfect Perfect for me Y our love is perfect It\'s all love can be  Your love is perfect In all of your ways  Your love is perfect I will worship you all of my days  Because your love is perfect Because your love has no fear Because your love is perfect I know that you are near  ------------------------------------------------------------------------------------------------ This poem was written on account of this scripture -&nbsp;1 John 4:18 \"There is no fear in love; but perfect love casteth out fear: because fear hath torment. He that feareth is not made perfect in love.\" I have been in love with things of the world, they have left me lonely, but I am in love with God and he has keep giving me joy!  '),(20689,3951300,'17,July,2004','2004-07-17 00:00:00','   Well still waiting for activation of website, hopefully later this morning it will be activated. Added somemore links to the sidebar also, give a warm welcome to The Digital Duck, Joshua, he has been added to our members thingy lol. Anyway I\'ll keep posting as I learn more on the domain and etc. &nbsp; Want to set a date for our band to come to your church or worship event? Email&nbsp;Jennifer at   urlLink sabshedule@netlane.com &nbsp; please keep this strictly for Scheduling, all other mail can be sent to  urlLink sabinfo@netlane.com&nbsp; &nbsp;Thanks. &nbsp; Email List Coming soon!!! Join now by sending an email to  urlLink sabinfo@netlane.com&nbsp;   Subject E-List. The Email List is a great way to keep up with the band and they\'re tour dates, etc. &nbsp; And just added this in meet Jennifer i\'ll have her blog account up tomarrow, she is the online Scheduling Contact, she is awesome, and really friendly so guys don\'t be shy to&nbsp;do band&nbsp;scheduling with her.  &nbsp; May&nbsp;God Bless you and keep you, Bradley Floyd  '),(20690,3951300,'16,July,2004','2004-07-16 00:00:00','   \"I \'m Just A Nobody Telling Everybody About Somebody Who Can Save ANYBODY!!\"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; THE LORD\'S BASEBALL GAME&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Freddy and the Lord stood by to observe a baseball game.&nbsp; The Lord\'s team was playing Satan\'s team.&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; The Lord\'s team was at bat, the score was tied zero to zero, and it was the bottom of the 9th inning with two outs.&nbsp; They continued to watch as a batter stepped up to the plate named \'Love.\'&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Love swung at the first pitch and hit a single, because \"Love never fails.\"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; The next batter was named Faith, who also got a single because Faith works with Love.&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp; The next batter up was named Godly Wisdom. Satan wound up and threw the first pitch. &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Godly Wisdom looked it over and let it pass: Ball one.&nbsp; Three more pitches and Godly Wisdom walked because he never swings at what Satan throws.&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; The bases were now loaded.&nbsp; The Lord then turned to Freddy and told him He was nowgoing to bring in His starplayer.&nbsp; Up to the plate stepped Grace.&nbsp; Freddy said, \"He sure doesn\'t look like much!\"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Satan\'s whole team relaxed when they saw Grace.Thinking he had won the game, Satan wound up and fired his first pitch.&nbsp; To the shock of everyone, Grace hit the ball harder than anyone had ever seen!&nbsp; But Satan was not worried; his center fielder let very few get by. &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; He went up for the ball, but it went right through his glove, hit him on the head and sent him crashing on the ground;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; the roaring crowds went wild as the ball&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp; continued over the fence . . . for a home run! &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp; The Lord\'s team wn!&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; The Lord then asked Freddy if he knew why Love, Faith and Godly Wisdom could get on base but couldn\'t win the game.&nbsp; Freddy&nbsp; answered that he didn\'t know why.&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; The Lord explained, \"If your love, faith and wisdom had won the game, you would think&nbsp; you had done it by yourself.&nbsp; Love, Faith and Wisdom will get you on base but only My Grace can get you Home:\'For by Grace are you saved, it is a gift of God; not of works, lest any man should boast.\" Ephestians 2:8-9&nbsp;&nbsp;&nbsp; Psalm 84:11,\"For the Lord God is a sun and shield; the Lord will give grace and glory; no good thing will He withhold from those who walk uprightly.\"  '),(20691,3951300,'16,July,2004','2004-07-16 00:00:00','   Another Lonely Saturday night, ready for sunday to be in sight, anew week soon will start as a lonely saturday breaks my heart.  &nbsp;  That is how I felt, I guess sometimes we feel so lonely, we feel so bored, we just feel like we are ready to fall apart, but the whole entire time there\'s Jesus, we forget that we can just turn to him. He is so great, I think some of us&nbsp;feel that we cannot lean on the unreconizable, because he is so hard to get, but God is reconizable, he wants us so bad to turn to him when we are lonely, when we are barely holding on. I have forgotten many times I could do that and should do that. So tonight, although I am somewhat in a lonely state and just want to sleep the day away, I know that if I turn to the supernatual, to the reconizable,&nbsp;that I will no longer be in this mood. God is with us, God is in us and most of all God wants us.  &nbsp;  So tonight I\'ll turn to Jesus, tonight I\'ll turn a little Chris Rice on and fall on Jesus. And although my sky is dark I know that the light in the distance is coming for me. I just got to wait for it to carry me away.  &nbsp;   urlLink Genesis 2:18 And the LORD God said, It is not good that the man should be alone; I will make him an help meet for him.  &nbsp;  God doesn\'t want you to be alone.  &nbsp;  If your sad, Cry to Jesus  If your falling, Fall on Jesus  If your joyfull, dance for Jesus  And Live - Chris Rice  &nbsp;  May God Bless You and Keep You  Bradley Floyd  '),(20692,3951300,'16,July,2004','2004-07-16 00:00:00','   Well today is a sunny day here in Memphis, I am just down the road from the St. Francis of Assis Church, he\'s a mentor!!! I arrived at about 8:30 am&nbsp;from Sentatobia where I am staying, I got invited to play with a band Sunday, I\'m still debating it lol. &nbsp;I still haven\'t heard from the domain hosting services yet, but it is for sure purchased as the following  urlLink http://secretambitionband.com  although the link is inactive. Joshua&nbsp;our webdesigner has a couple of site\'s&nbsp;I would like to mention. For a great Christian Forum&nbsp;Community try  urlLink http://swordwalk.com  he and his friends have really made it great. Check out his blog at  urlLink http://thedigitalduck.com  also check out this awesome Scripture link up system, lol not sure if I said that right, but it\'s neat lol. I also want to thank Joshua he really is a great webdesigner and it goes without saying, but I\'ll say it&nbsp;anywayz lol, he is an awesome christian, and God really has used his knowlege in such a great way. So Thanks Joshua for taking the job and most of all being a great and loyal friend.&nbsp; Also all the supporters out there, thank you, without you guys none of this would be possible.  &nbsp;  May God Bless You and Keep You,  Bradley Floyd  &nbsp;  P.S.  Be God\'s.  '),(20693,3951300,'16,July,2004','2004-07-16 00:00:00','   For a moment, I forgot what I was suppose to do For a moment, I forgot how much I loved you  Just a moment can be a long time Just a monent is worth more than dime And I don\'t want to lose it, cuz it\'s the greatest thing I can find I always want my moment with you on my mind  For a moment, I was afraid of the sea For a moment, I forgot you were inside of me  Just a moment can be a long time Just a monent is worth more than dime And I don\'t want to lose it, cuz it\'s the greatest thing I can find I always want my moment with you on my mind  Just a moment, a moment with you Just a moment, not a minute to soon Just a moment, that will do Just to have a moment with you &nbsp; I wrote that one night when I felt so bad, I felt I had wronged God and I knew I needed to get on the right track, I\'m so glad I serve a God who gives 2nd chances.  '),(20694,3951300,'16,July,2004','2004-07-16 00:00:00','   urlLink http://secretambitionband.com  just bought it, should be up soon still looking for the host though, got some good leads but still not sure which to go with. &nbsp; May God Bless you and keep you,  Bradley Floyd &nbsp; P.S. My webdesigners official blog in case you were wondering is  urlLink http://thedigitalduck.com  '),(20695,3951300,'16,July,2004','2004-07-16 00:00:00','   Hey everyone this is Bradley Floyd, I just wanted to say hello and welcome to the bands official blog site, hopefully I\'ll have some images of the band up soon, thanks for your outstanding support. &nbsp; God Bless you and keep you, Bradley Floyd  '),(20696,3951300,'12,August,2004','2004-08-12 00:00:00','   Well I think this is really a struggle even with the non-christians lol buy probally more so with the Christians. Since we should know that these things of the world are wrong.  I think the great thing about being a Christian is we have an image to uphold, see we as Christians are the front-image of Christ since the world can\'t see Christ themselves we must show his love, his grace, his mercy and his life to all though us. And that scares me because all the time I\'m so scared of mis-representing God LoL, and at times I wish I could be of the world and just not have to face the reality of Christ with in me and Christ Before me. So with that I hope this next story blesses you and keeps pushing you to be more like Christ and less like the world.   Painted Sticks One Sunday a preacher told how, while sitting in his garden, he had watched a caterpillar climb a painted stick that was for decoration. After reaching the top, the caterpillar reared itself, feeling this way and that for a juicy twig to feed on, or some way to further progress. Finding nothing, it slowly returned to the ground, crawled along till it reached another painted stick, and did the same thing all over again. The preacher said: \"There are many painted sticks in the world-those of pleasure, wealth, and fame. All these call man and say, \'Climb me to find the desire of your heart, fulfill the purpose of your existence, taste the fruit of success, and find satisfaction, but they are only painted sticks.\'\" Solomon tried to find the purpose of his life in the world\'s \"painted sticks.\" He gave his heart to seek wisdom, but learned that it was \"vanity and vexation of spirit\" (Ecc_1:15). He then turned to the pleasures of the world for meaning in life. He built great houses, and gardens, and pools. He had servants and maidens; in fact, he had all that a man could desire. Solomon\'s comment on pleasure as a true source of happiness, however, was, \"All was vanity and vexation of spirit, and there was no profit under the sun\" (Ecc_2:11). After trying all that the world could offer, Solomon\'s final decision was, \"Let us hear the conclusion of the whole matter: Fear God, and keep His commandments: for this is the whole duty of man\" (Ecc_12:13).  May God Bless You and Keep You, Bradley Floyd  '),(20697,3951300,'11,August,2004','2004-08-11 00:00:00','   Greeting the King Noble men were gathered together in London waiting for the King of Great Britain. They all knew him personally, yet they all honored him as their king. When he entered, they stood solemnly to their feet. \"Take your seats, gentlemen,\" he said, \"I count you as my personal friends.\" And then joking he added, \"I am not the Lord, you know!\" Immediately one of the noblemen, a Christian, said, \"No, sir, if you were our Lord, we would not have stood to our feet; we would have fallen to our knees.\" - Brad  '),(20698,3951300,'10,August,2004','2004-08-10 00:00:00','   Whale Songs According to research conducted by the National Geographic Society, the 40-ton creation of God-the humpback whale-has a fascinating singing ability. Recordings have been made of the humpback whale singing in various pitches in solos, duets, trios, and choruses of dozens of interweaving voices lasting from six to thirty minutes. What an experience it is to hear, over one\'s own stereo system, songs from the ocean depths-sung by 40-ton whales! The passage of Psa_148:7 comes to mind: \"Praise Him down here on earth you creatures of the ocean depths.\"  Keep us in your prayers and May God Bless You and Keep You, Bradley Floyd  '),(20699,3951300,'08,August,2004','2004-08-08 00:00:00','   Well I decided tonight to put all the parts together.  Verse: Mat 8:20 And Jesus saith unto him, The foxes have holes, and the birds of the air have nests; but the Son of man hath not where to lay his head.  Jesus was a homeless man, he said it him self in Mat 8:20 when he said that he \"the son of man had no where to lay his head.\" WoW to think that our God, Jesus Christ, the one whom we call savior was homeless, now sure there were places he could sleep but never a place for him to call home. I guess this was because Jesus knew that he was only visiting for a little while, that he would soon be leaving, that he would soon be gone, and that this earth was not his home like wise this world is not our home, we are just traveling through this world as visitors. One day our father will call for us and we will go home. How AWESOME!!! But this verse \"Mat 8:20\" really humbles me and makes me reconsider what I have done and how I can be just like Jesus knowing that this world is not my home and therefore give all I have \'cuz it ain\'t goin\' to heaven with me anyway\'s. As the Shriners Moto goes \"No Man is Taller than The One Who Stoops so low to help a needy Child.\"  I hope you enjoy these stories. Please write back if this story has blessed you in a certain way. May God Bless You and Keep You,   The Higher the Bamboo Grows, the Lower It Bends A government official who came into President Lincoln\'s office was startled to find the chief executive shining his shoes. \"Sir,\" he gasped, \"surely you do not polish your own shoes!\" \"Of course,\" replied the humble President. \"Whose do you polish?\" The greatness of a man is evidenced by his humility. In the words of an old Filipino saying, \"The higher the bamboo grows, the lower it bends.\" All of this is illustrated in the life of Jesus. \"Jesus did not come to be served but to serve, and to give His life a ransom for many\" (Mat_20:28). The highest title God ever allows in His church is \"servant.\" Am I willing to serve people as Jesus did? I know, I cannot die for people to save them as Jesus did, but I can \"die to self\" to enable me to tell people that Jesus died for their sins, and thus serve them in the greatest way possible. When I humble myself, then God will give me the grace to grow into maturity where, I like the tall bamboo, can lower myself to serve those of \"low esteem.\" Notice, \"Do not be proud, but be willing to associate with people of low position. Do not be conceited\" (Rom_12:16). How tall are you spiritually? Tall and flexible enough to serve people in all kinds of situations?  \"Brokenness... God uses broken things. It takes broken soil to produce a crop, broken clouds to produce rain, broken grain to give bread and broken bread to give strength. It is the broken alabaster box that gives forth perfume. It is the broken Peter, weeping bitterly, who returns to greater power than ever.\"  '),(20700,3951300,'07,August,2004','2004-08-07 00:00:00','   Well I hate to end with just this story because there are so many great one\'s, but I promise to write only 3 parts :) so here is the conclusion, and this one really broke my heart.  May God Bless you and Keep You, Bradley Floyd  When Someone Is Needier Than You In a leper colony clothes arrived for distribution among these outcasts of life. One little leper boy was asked what he needed most. He replied, \"My hands are still good so I can mend my clothes when they need it, but there are some lepers who have no fingers, and some whose hands are so crippled or sore that they cannot use them. When their clothes get holes in them, they cannot mend them but must see the holes get bigger and bigger. I am thankful for my good hands, so please give the clothes to those who cannot mend their own. They are more needy than I.\" How we need such unselfishness. We should never complain when we know that there are others who are needier than we are.  '),(20701,3951300,'06,August,2004','2004-08-06 00:00:00','   Which Was Neighbor? A certain woman went down from Washington to Richmond and ran over a spike which punctured her tire and left her stranded by the side of the road. After raising the hood of her car and tying a scarf to her radio antenna, she locked the door handles and sat in the car, praying for the Lord to send help! By chance, there came a limousine that way with a bumper sticker that read, \"Smile, God Loves You!\" When the occupants saw the stranded woman, they passed by in the far lane-without smiling. And likewise, there came a sports car with a CB radio and a bumper sticker saying, \"Honk If You Love Jesus!\" The man who was driving passed by in the far lane without honking and without using his CB to tell the Highway Patrol about the woman\'s dilemma. But a certain working man, as he traveled to his job, came to the spot where the woman was and, when he saw her raised hood, white scarf, and flat tire, he had compassion on her. He stopped his old beat-up pickup-which had no bumper sticker-and crossed the four-lane highway, and offered to change the tire. The woman opened the door and gave him the key to the trunk. The man took out the spare tire, jacked up the car, removed the flat tire, and replaced it with the spare. When he had finished, the woman tried to pay him. He refused the money, saying, \"If my wife were stranded on the highway with a flat tire, I\'d want some Good Samaritan to stop and help her out.\" He returned to his bumper-stickerless truck, smiled, honked at her, and went his way. Which of these three was neighbor unto her who had a flat tire?  Christ doesn\'t want to be on a bumper sticker on a nice fancy car, Christ wants to be in your heart. Matt 5:8 Blessed Are the pure in hear for they shall see God.  May God Bless You and Keep You, Bradley Floyd  '),(20702,3951300,'05,August,2004','2004-08-05 00:00:00','   Once a young couple lost their only child, a beautiful three-year-old daughter, as a result of a freak accident. Unable to justify this tragedy with their concept of a loving God, they went to their minister for help. \"Why?\" they asked. \"It was God\'s will,\" the preacher told them. The couple just couldn\'t accept that the passing of their little daughter could somehow serve God\'s will, and they sought out another preacher friend. He quoted yards of Scripture, trying to give them comfort, but his verses fell on deaf ears. Finally, the young couple turned to a frail elderly woman in their church. Surely, she could give them a reason for the child\'s death, but the old woman didn\'t utter a single word. Instead, she tenderly wrapped her arms around the heartbroken couple\'s shoulders, and together they cried the hurt away. Sometimes when theological answers are not easy to the complex misfortunes of life, perhaps we just need to hush our babblings. Sometimes when answers aren\'t easy, our best statement is silent caring.  Tonight I want to take a moment to talk about Compassion International. This is a Christian Charity Organization, check them out at  urlLink http://compassion.org , that I really feel strong about and hope that while reading a seeing what they do that you will also develope this trust and strong feeling\'s toward these amazing people. Compassion International is a non-Profit Organization that not only takes the money that you sow and gives it to the kids but also spends times with these kids and teach them of christ. I hope that God will set it in your heart to do what you think is write.  May God Bless You and Keep You, Bradley Floyd.  '),(20703,3951300,'05,August,2004','2004-08-05 00:00:00','   Well this is last post, and I think I leave on a good note with this one. May God Bless You and Keep You, Bradley Floyd   It Makes All the Difference On a crowded street of one of our large cities, a young man was snatched from the path of a speeding truck, his life saved by a venerable-looking man. Still breathless from fright, the youth thanked the one who saved his life and then was lost in the crowd. Two weeks later in a crowded courtroom, an anxious young man stood in the prisoner\'s box to be sentenced for murder. \"Young man, have you anything to say before the sentence of death is passed upon you?\" \"Why! Yes! Yes, Judge,\" the youth responded, \"you know me.\" A silence moved like a shock wave over the courtroom. \"I\'m sorry. I cannot place you.\" \"Yes. Surely you remember. Two weeks ago. At Main and Seventh Streets, you saved my life. Surely, Judge, you can do something to save me now.\" A silence pervaded the courtroom. \"Young man, now I do remember you. But that day I was your savior. Today I am your judge.\" Today the Lord Jesus Christ wants to be your Savior. If you refuse Him and His grace, one day He will be your Judge.  '),(20704,3951300,'05,August,2004','2004-08-05 00:00:00','   Well I think god is wanting a part 2 to this story lol.   Christ the Way On a dark and stormy night, a child was lost in the streets of a large city. A policeman found him crying in distress, and gathering enough from his story to locate the home, gave him directions after this manner. \"Just go down this street half a mile, turn and cross the big iron bridge, then turn to your right and follow the river down a little way, and you\'ll see where you are.\" The poor child, only half comprehending, chilled and bewildered by the storm, turned about blindly, when another voice spoke in a kindly tone, \"Just come with me.\" The little hand was clasped in a stronger one, the corner of a warm coat was thrown over the shoulders of the shivering child, and the way home was made easy. The first man had told the way; the second man became the way. This is exactly what the Lord Jesus Christ has been for us. From eternity He has told us that He is the Way. He has to become our Light also, to lead us to the Way.  May God Bless You and Keep You, Bradley Floyd.  '),(20705,3951300,'05,August,2004','2004-08-05 00:00:00','   Well Yet again I have another post, tonight though I really feel like God is talking to me so I\'m going to keep on listening and keep on posting till he say\'s he done with me LoL   Christ the Savior A native of interior China wanted to become a Christian but couldn\'t understand how Christianity was superior to Confucianism and Buddhism. One morning he came to the missionary in a happy mood saying, \"I dreamed last night, and now I understand. I dreamed I had fallen into a deep pit where I lay helpless and despairing. Confucius came and said, \'Let me give you advice, my friend; if you get out of your trouble, never get in again.\' Buddha came and said, \'If you can climb up to where I can reach you, I will help you.\' Then Christ came. He climbed down into the pit and carried me out.\" It takes the Savior of man to do that. Only a Savior would stoop so low as to save a sinful soul like yours and mine.  May God Bless You and Keep You, Bradley Floyd  '),(20706,3951300,'05,August,2004','2004-08-05 00:00:00','   \"Well sometimes my life just donâ€™t make sense at all when the mountains look so big and my faith just seems so small, so hold me Jesus â€˜cause Iâ€™m shaking like a leaf, youâ€™ve been king of my glory, wonâ€™t you be my prince of peace.\" â€“ Rich Mullins  This just maybe one of my favorite Rich Songs it always get\'s me thinking of how God is using me and though I sometimes have the hardest times getting through the simplest things of life like learning to love and be loved, and getting stuff done in what time we have, I know that god is still walking with me and still carries me when i can no longer walk...  The Story of Footprints... As I prayed Jesus allowed me to look back in my life as I looked back I saw their were to sets of footprints in the sand, so I asked Jesus \"Who\'s footprints are these?\" Jesus Replied \"Those are mine, walking with you\" And then I noticed that sometimes there was only one set of footprints and so I asked \"Jesus, why is there only one set of footprints? Is this when you left me to go through life alone?\" Jesus calmly smiled and said \"No, those are my footprints, there is only one set because you were to weak to walk so I carried you\"  MAN!! That always get\'s me in tears to think that Christ loves us that much.  The Story of the heavy Cross... Jesus this Cross is to heavy, could you not give me a lighter one please? Yes my child go into that room and pick a smaller one. As I entered the room I saw all these huge crosses so I said \"Jesus, who must carry these? He replied These are the one\'s I carry, cannot you carry this tiny one upon your back?  Isa 53:3 He is despised and rejected of men; a man of sorrows, and acquainted with grief: and we hid as it were our faces from him; he was despised, and we esteemed him not. Isa 53:4 Surely he hath borne our griefs, and carried our sorrows: yet we did esteem him stricken, smitten of God, and afflicted. Isa 53:5 But he was wounded for our transgressions, he was bruised for our iniquities: the chastisement of our peace was upon him; and with his stripes we are healed. Isa 53:6 All we like sheep have gone astray; we have turned every one to his own way; and the LORD hath laid on him the iniquity of us all. Isa 53:7 He was oppressed, and he was afflicted, yet he opened not his mouth: he is brought as a lamb to the slaughter, and as a sheep before her shearers is dumb, so he openeth not his mouth. Isa 53:8 He was taken from prison and from judgment: and who shall declare his generation? for he was cut off out of the land of the living: for the transgression of my people was he stricken.  This is the God that we so often misunderstand, he has done so much already, if he choose\'s to do more it is through the kindness of his mercy.  May God Bless You And Keep You, Bradley Floyd  '),(20707,3951300,'02,August,2004','2004-08-02 00:00:00','   urlLink    :)  '),(20708,3951300,'02,August,2004','2004-08-02 00:00:00','   The Right to PursueSomeone cursed Benjamin Franklin, charging that the Constitution of the United States was a farce. \"Where is all the success that it guarantees us?\" he sneered. Franklin, smiling, answered, \"My friend, the Constitution guarantees only the right to the pursuit of happiness.\" Psa_34:14 tells us to not only seek peace but pursue it. Psa 34:14 Depart from evil, and do good; seek peace, and pursue it.  As we head towards another one of those elections that are so intense let us remember that as a nation we should seek peace and pursue it. It is not about who is the smartest nor who is the richest, it about who\'s after the pursuit of righteousness and the integrity of all nations to act on it. We live not in society that is given anything but in a society that is given the right pursuit so let us pursue, and though the battle for that pursuit may be long and vigorous let us remember our fathers words \"I am with you always.\" - Bradley Floyd  '),(20709,3951300,'02,August,2004','2004-08-02 00:00:00','   Well I\'m back maybe not for good lol but I do finally have a chance to post :). As some of you may already know we have an official record label deal with Javon Records :). We are very excited, and hope to have that album soon. That\'s all I have as far us updates. Please Check back for more post and another great story from me lol.  May God Bless You and Keep You, Bradley Floyd  '),(20710,3951300,'02,August,2004','2004-08-02 00:00:00','   urlLink    inside 3  '),(20711,3951300,'01,August,2004','2004-08-01 00:00:00','   urlLink    Inside 1  '),(20712,3951300,'01,August,2004','2004-08-01 00:00:00','   urlLink    Cover  '),(20713,3951300,'01,August,2004','2004-08-01 00:00:00','   urlLink    The Bands featured pictures, we are working on them for our flash video :)  '),(20714,3951300,'01,August,2004','2004-08-01 00:00:00','   Well as get ready to head for the sack, I thought I\'d let all you other lonely people out their know that you are never trully alone.  urlLink    Lonely tree in the midst of others.... Good Verses for this entry....  urlLink Psalm 77:7-13   urlLink Isaiah 53:4   urlLink Isaiah 55:8-9   urlLink Matthew 5:4   urlLink Matthew 6:11   urlLink Luke 22:41-45   urlLink Romans 7:18-19   urlLink Romans 7:24-25   You who live in heaven Hear the prayers of those of us who live on earth Who are afraid of being left by those we love And who get hardened in the hurt Do you remember when You lived down here where we all scrape To find the faith to ask for daily bread Did You forget about us after You had flown away Well I memorized every word You said Still I\'m so scared, I\'m holding my breath While You\'re up there just playing hard to get You who live in radiance Hear the prayers of those of us who live in skin We have a love that\'s not as patient as Yours was Still we do love now and then Did You ever know loneliness Did You ever know need Do You remember just how long a night can get? When You were barely holding on And Your friends fall asleep And don\'t see the blood that\'s running in Your sweat Will those who mourn be left uncomforted While You\'re up there just playing hard to get? And I know You bore our sorrows And I know You feel our pain And I know that it would not hurt any less Even if it could be explained And I know that I am only lashing out At the One who loves me most And after I have figured this, somehow What I really need to know Is if You who live in eternity Hear the prayers of those of us who live in time We can\'t see what\'s ahead And we can not get free from what we\'ve left behind I\'m reeling from these voices that keep screaming in my ears All these words of shame and doubt, blame and regret I can\'t see how You\'re leading me unless You\'ve led me here To where I\'m lost enough to let myself be led And so You\'ve been here all along I guess It\'s just Your ways and You are just plain hard to get Copyright 1998 - Liturgy Legacy Music / Word Music / ASCAP  '),(20715,3538351,'21,June,2004','2004-06-21 00:00:00','    It\'s amazing, isn\'t it, that someone\'s life can end so suddenly, while everyone else is forced to keep on living. How you can be with someone one night at a party, and then the next morning they die in a car accident. My friend Bradley C. Lybbert did just that. He was killed in a car accident on the way to his high school graduation on Sunday June 13, 2004. He had just been at a party, that I was also at, the night before. And then I find out the next morning that he is gone. It is events like these that make you realize that you don\'t know when your time is going to come to leave this life. It is times like these that make you realize that you aren\'t invincible and neither are the ones you love. Brad was a very good young man. He was fun to be around and talk to, he was a very good friend and a really kind person. He is going to missed very much. We will never forget you Bradley. Thank you for being a part of out lives and blessing us. We will miss you, but we will see you again in Heaven.   '),(20716,3538351,'04,June,2004','2004-06-04 00:00:00','    Okay yesterday I said that things come in 3\'s. Well I found out that it holds true for good things too. My third good thing is that I got into AP English for next year. I\'m really happy about that and I really wanted wo get into it. I however did not become one of the Vice Presidents for National Honor Society for next school year. I\'m kinda bummed about because I would ahve liked to get the chance to be a leader in my school. I\'ve run for class office twice without getting elected and now I won\'t have the chance to be a leader in NHS. I guess I\'m just not supposed to have such a big leadership role at my school. Maybe there is something else I am supposed to do ot maybe the Lord knows that i will be really busy next yeatr that He is protecting from extra stress. I don\'t know. Maybe someday i will know. But for now things, good and bad, come in 3\'s.   '),(20717,3538351,'03,June,2004','2004-06-03 00:00:00','    People often say that ad things come in threes. But do good things come that way too? If so, then I have 1 more good thing coming, but there are two things which I\'m trying for, so which will it be? Alomost 2 weeks ago I became a new member of the South Kitsap High School Acting Ensemble for the 2004-2005 school year. Good thing #1. Then that saturday I became South Kitsap\'s Junior Miss 2004. I get $1000 for any college I want and I get to go to state in January. I\'m so excited. Good thing #2. Now I would like to get into AP English for next school year and I would like to be a Vice President of National Honor Society for next year. So which whill it be? or will I prove the 3-theory wrong and have either 4 good things happen, or maybe only 2? I guess I\'ll find out soon! As for me I hope to prove the 3-theory wrong, with having 4 good things happen.   '),(20718,3331655,'24,May,2004','2004-05-24 00:00:00','   LOl... haiX.. today nothing to do..until 10.33.. cause.. i suddenly got a call from junjie. then he say wanna go cafe anot.. so i say..11.30 then can go out..then, unforturnaty,we go out anyway...we start to pass by hougang mall..then i sugguest that chicken rice is good.. after a few minutes of saying what to buy.. in the end, we still eat chicken rice :P..after that, we went to a nearby card shop. then only got 2 booster pack.. and as i was a good pokemon booster selecter.. i select a pack. And junjie open it out and inside got a ELETUBUZZ EX.. lol..that was good..after that.. i went home..then take up my deck to junjie house play.. he lose me a lot of time..then i started forming his deck to a new deck..in the new deck, he win me a lot of time.. and after that.. we go down play basketball a while..then go back to his house,bet card and win a lot of FOIL card.. lolz..in the end i went straight home..hehez  '),(20719,3331655,'21,May,2004','2004-05-21 00:00:00','   HaiX...YaWh!! wAU!@231@! 7.40AM ALREADY!! OH NO, I AM LATE.. As today got basketball match , i wake up late! lolz...lucky i dress up to my school uniform and then i rush to school.. as i reach school, all my friend thought i never come..and they worried alot..after the assembly, is our match i quickly dress my basketball pants and then dress into my school house T-Shirt.       The first match is vs the ultimate team,xaiver that team..we have no confident as never play as good as we have last time.. propably is that we even dun have a finger to lay on them.. Second, we vs junjie that team by a rematch..still we loss another 3 point by Joshua\'s tyco 3-pointer shot..hAiX we felt as we lose against junjie that team..i hate that.. I bet that my teamate also hate that..lol..nevermind lah...just a game.. that was a good game..we learn a lesson..we will try harder...      GREEN HAWK ROCK!! We will rematch again.. =>2B '),(20720,3331655,'19,May,2004','2004-05-19 00:00:00','   YaWn... WOw...6.30 already...getting excited as the basketball inter-house is going to start today..my team got Jacob,Desmond,Russell,Jiajie and me.. at first i go to school to discuss formation then started off playing..               Matches going to start the first team is Stingray(which is my class  group but not mine )vs Darkness(junjie group) They started off playing first so my team have to wait lolz :(.. but at last, the match have ended... the score is 2:2 that a draw..then secondly..my team(which is Snowflakes) vs darkness and we lose by Junjie\'s 3-pointer shot..while we only got 2 point for that..the 3rd match is my team(also!)vs stingray.   Unfortunaly,we win then 16:3... that was pretty good lolz.. haiX...then in the end becoz we trash stingray by 13point, we went into final. I feel sorry for my class team..which they tried very hard...  '),(20721,3331655,'13,May,2004','2004-05-13 00:00:00','   TodAy KinDa sIaNz MaN..Art ExaM iS my TodAy exAm...    The First 25min,i keep racking on my brain..but then..then i think of something...guess WhAt i HaVe DrAw?..(nO laH..nOt DrAw) is ShaDe..i decidE oN DoiNg ShApE pIc LiKe CirCLe, TrIanGle, And A lOt Of sQuArE. aNd AfTeR fINiSh DrAwiNg THe SqUaRe, I sTaRtEd to ShAde The sQuaRe..ANd THen The TriaNGle..anD aFtEr aLL,i FinIsh WitHin 1 Hour..tHe OtHeR HaLf An HoUr I slEep oN thE tAble        AftEr tHe eXam i StArt wIth GoiNg to LinYu hOuSe bUt He sAy CaNNot..So i STay bAck In SChoOl tO pLaY baSkEtBall..ShOcKiNg FoR me, i JusT sHoOt OnCe And I sCorE tHE PoiNt.WheN the MaTch End, i stArt goiNg tO FoOd COuRt To EAt.. FinIsh eAtiNg, i WeNt Home And HIbErNatIon UntIl 5.30..tHeN sTart to SIgN in To ChaT fOr A wHiLe thEN GO LoWeR coUrT pLaY bASKeTBaLL..I PlaY mAnY mAtCh In THe baSkeTbaLl cOuRt. TheIr waS a Xia Lan Kia KeeP scOldIng BAd wOrD.. HeNg aR..i PlaY baSkeTbaLL AlwaY EnJoy LaUgTHeR,TeAmwOrK aNd CooPerAtion..i Know A lOt Of fRieNd iN tHe baSkeTbaLL CouRt.. hOpe tHeir Stay as ~~~~PuNky And FUnKy Forever~~~  '),(20722,3711841,'22,January,2004','2004-01-22 00:00:00','    Autobiography    By Eric Heppner   The air was far too cold for it was a late march blizzard outside. The confining blanket of blue was not the same for it was not alive, it felt abrasive and rough compared warm softness from which I had just come. All twenty-two inches of my body shivered slightly from shock. My journey had been difficult to Fergus Falls, Minnesota, maybe because I had tried to take too much with me. Ten pounds and three ounces is a lot of baggage. My face was bruised something awful, my mother had always called me a prizefighter for that. I recognized a few of the voices surrounding me though I saw little besides soft white light and shades. I noticed one shade in particular. His voice was deep and frightening yet familiar and warm. I heard my name called out in song, I knew it because I had heard it all my life, Eric Ole Heppner. The shade spoke as if he didnâ€™t expect me to be there to receive it. I heard singing for the first time. It was praise, praise to God that I had been born. I did not know this. When the tremendous shade held me in his arms, as a reaction, my body, without my consent, released my stored waste upon him. This only dampened his mood for a second as his strong arms held me all the closer. His name was Steven.  I learned from a far a more familiar voice that called me away from his arms. Her name was Joan. I had heard it many times throughout my journey yet her voice was my voice, her breath mine, and her love was mine. She held me gently to her breast and I could hear her heart, the heart that I had heard for nine months. That moment on March 27, 1987 was the greatest moment of my life.  	My older brother Jon had spoken for me for the first year or so of my life. This made life very simple. I ate when he did, slept when he did, and got changed when he did. But I never had a first word. I had a first phrase. â€œI read a book.â€ What I wanted was for my parents to read to me. This was my first individual want, and I am glad that it was dedicated to a search for knowledge. My first actual word then would be â€œIâ€ and for the life of speaking afterward, I would be trying to explain that word to others. Even though I was now speaking for my self, Jon was still wanting for the both of us. We did everything together. Early in the morning Jon would climb into my crib and when I had more wants I would say to him â€œJonston leave me alone!â€ as I couldnâ€™t pronounce Jonathan. This would never abate him and I would soon be awake and on an adventure whether it was eating various berries outside or flying imaginary planes made out of large cardboard blocks in the living room.  There was no limit to my imagination and Jonâ€™s thirst to never sit still.  	One particularly large adventure came when we found the special advantage that men have over women. We found that we unlike women could pee any place and at any time we desired. This magnificent freedom sent us on a quest to water every plant outside our Staten Island, New York parsonage. We built a small following in this venture and our best friends from church, Marcos and Luis, would ease the burden of our quest and passion. Yet with this help there came a terrible cost and her name was Maria. She was old yet I can never remember her getting any older, and she was little. She had had six boys and a girl; seven children in all and I remember thinking that each one of them must have taken a piece of her to make because there wasnâ€™t much altitude left. Yet this had been attributed to her as her greatest weapon against our small band and our free ways for the closer she was to the ground the closer she was to her sandal. She wore leather sandals with everything, and it only lay at the most two feet from her hand that must have been made of wind. On one of our front lawn escapades, Maria caught us giving fluid to desperate shrubbery in the flowerbed. There she stood, all four feet of her magnificent power ready. As one can recall in the midst of great danger the motions of something so fast that it is uncaught by the natural eye.  I shall recall her motion. She flipped the leather weapon off her foot into the naked air in front of her and leaped forward. The sandal spun twice in the air and Maria caught the heel in her hand. She relinquished a scream that froze us dead to our very souls even if we had had time to go anywhere. I saw the face of Marcos get slapped and pushed out of the way as it headed like an arrow for its next target plowing Jonâ€™s face back wards as he fell into the flowerbed. Louis next received the punishment and I saw saliva rocketed from his mouth as contact was made. I watched with terror unfelt in my young mind as with a helpless stare I saw the sandal close in inevitably on my tender jaw. The contact didnâ€™t hurt at first yet surprised me as my slow motion vision had ended and those long mille-seconds ended as I spiraled through the air into the flowerbed in an instant. The pain rushed to my cheek as if I had no skin, muscle, or tissue of any kind left and the pain was simply flowing out onto the ground to create the next flood that would destroy the world. Luckily my cheek was still there. The shock of this left me dry for several days and nearly ended my infamous peeing crusade. Yet I have been notorious for not knowing when I have been vanquished.  	It had only been a month after the Maria incident and I peed alone. Groups where to conspicuous. On occasion Jon would help but I didnâ€™t feel safe with anyone else anymore. My legs were crossed when Marcos and Louis came. On this day I felt the irrepressible urge to water the honeysuckles that traipsed the fence of our yard. I stood on a solid fort built by my father to house Jonâ€™s and my adventures. On this second story hideout I did the dastardly deed again. My golden arch reaching the honeysuckles, the fence, and the sidewalk. This was my error. Luckily I did not make the apocalyptic error of hitting a passerby in NYC with anything more liquid and warm then a sideways glance, and even that was dangerous. My ramifications for my sin came from across the street. Gwenn, one of my dadâ€™s parishioners, happened to see my pernicious deed and I had barely gotten my fly back up when my mother, ear fresh from the phone, brought swift ramifications upon my backside and ears. Thus ended my illustrious urinating career and routed the last urination threat left in the community. I was to reform and forever after have been forced to use the head.   '),(20723,3711841,'21,January,2004','2004-01-21 00:00:00','    The Lilies of Connecticut    By Eric Heppner   Jordan flipped his car off the road during a late clear night in Connecticut. He got out and walked toward a house in the center of the woods. It was a small house, yet when he opened the door it showed a huge open plain that contrasted the night outside with the day within. He entered and the door stood open like a cut in the wind, bleeding wind and leaves through. Jordan walked through the tall grass to a large flat pool in the center. He bent over to the pool and lifted some water to a cut that had appeared on his head from the crash, the cut stopped bleeding and then disappeared. The crash left him covered in bruises and cuts and his arm was badly broken, in hope to placate his pain he slid into the placid pool. The walls of the lake were sheer and the clarity of the water betrayed the depth. He fell through the water as if it were thin cotton sheets. The water finally thickened and he gasped for air as he had been holding his lungs. The water slid into his through and the sides of his neck soothing him. He stopped moving and the water stopped flowing and he felt out of breath again. He kicked his legs but they barely moved and in desperation he arced his back moving a massive tail behind him propelling him forward at a great speed. He reached forward with his hands which were webbed to the first knuckle and very smooth. His ailments had passed away. He swam forward at an incredible pace, the faster he went the clearer his mind. Soon his tail burned as the muscles strained. He saw a light below him and fled toward it. As he approached he could make out a city,  a massive golden city surrounded by awe giving fields of purple and blue and green. The base to the golden city was coral that was nearly transparent as tiny creatures moved in and out strengthening it. He swam to the city and the men and women in the field approached him with awe on their faces. They kept they\'re distances and whispered among themselves. Several young children raced to the city yelling in a strange melodic voice, that Jordan strangely understood, they yelled victory, victory, he returns, he returns. The city poured with people of every size age and build all of them had wonder etched plainly on their faces. One old man came to him and held an old rusty metal rod. A worthless lump of rust that looked hewn from the car Jordan used to drive, the axle maybe that was so horrible it had cracked and flipped him. The elder presented Jordan the gift. He tried to refuse but the aged man was insistent. He reached out with his strong purple hand and lightly touched the metal and the most marvelous thing happened. As his fingers lightly caressed the jagged metal the flakes of rust melted into silver metal beneath. He grasped the growing bare silver spot with a familiarity that was returning to his hand from a long forgotten dream. The pole transformed into a flowing silver color that looked like a waterfall of platinum without the crash simply a steady beautiful stream. The scepter was fluted on all sides and at the summit was a grand pearl, the greatest in the ocean it held a slight pink hue within its swirling white wonder. From the staff his arm was coated in silver plates that decorated him. His chest was adorned with a single silver banner that flowed with the water around him, and on his head was formed a stunning gold crown that reached to the surface with seven fingers each crowned in a little pearl. He looked over his people of which he was king. They chanted his name yet in their language it took enough syllables to be a chorus of intense melodic beauty. Jordan looked toward the old man that had brought him and he saw the face of his greatest friend. He touched the face of the elder and the age passed from him. He found in the old man the general of his armies, the advisor to his thrown, and the keeper of the city in his absence, his name was Mathew. Mathew became not unlike himself in the beauty and grace of a people of a great age.  	Finally came to him one ragged old women. She had a tear in her eye that surpassed the beauty of the city, and the fields, and the pearls. He saw greatness in her. He reached out and touched her face with wonder and mystery and the hope that he would reveal her true beauty. The women shattered and fell away too the ground. The king was devastated at the loss of this beauty and chased after the shards. As they fell to the ground they broke again and again into a fine dust and as they touched the ground. They were eaten by the seaweed that then grew together and formed a glorious white bulb. The bulb broke and pedal upon peddle fell away before the kings eyes. Finally at the very core of the now heavenly flower there sat a maiden rained in white foam with a crown of purple fish swimming above her head. He swam to her and held her head in his strong hands. He lowered his head to hers and as their lips touched all the flowers in all the fields in all the oceans of the world bloomed a glorious white flower The king married his new queen that day on a platform of pink coral under a tent of translucent silver cloth. He then sat on his golden thrown and ruled wisely in ageless beauty with his glorious queen. Every summer after that glorious year the lilies of Connecticut yield white flowers in their honor.   '),(20724,3711841,'21,January,2004','2004-01-21 00:00:00','    Lifewright    By Eric Heppner   It was the forest Lifewright With boughs of green and gold That held the stars of heavens And shaded the men of old  Through it ran river faithful Which flowed through every root That quenched every fire And washed away the soot  The insects of Lifewright The mantis and the bees Kept the order of the forest And flowered all the trees  Still it was the druid Who sparked live in the wood Who loved the forest As only its father could  His eyes saw all the corners His mind foresaw the wind His hand started the beat of the heart And could stop deaths painful singe  He guided the trees of Lifewright Crafted each branch with his hand He brought up the golden kernels of grain And gave the fruit to the land  The forest grew tall and glorious Its roots grew down to core They filled all the soil with grasping hands And still they wanted more  They stretched the bounds of the river Their roots reformed the soil Its bottom sank into the waste And sunk its might to spoil  As the water fled the sun grew hot And turned forest plush to hard It dried the leafs to fire The forestâ€™s face forever marred  Now the forest was ablaze The faithful all were gone The trees fell quietly around him The druid was now alone  He gripped the last seed From when the trees were young and grew He groped his fingers through the blackened ground And hoped the forest would grow anew  He dug at the base of the river For the scent of a drop He dug till his strength left him But the water was for naught  He pleaded with his servants The mantis and the bees They flew far away  Yet never found the trees  His strength was gone And his tongue was dry His eyes were heavy He was about to die  He buried the seed Still clenched in his fist His last tears fell For the forest he missed  For an age he lay Unspoiled by the world His hand tightly closed As the seed unfurled  Watered by the tear That it had stored so long ago Its roots ran through the ground And the faithful began to flow  The tree grew slow and strong The faithful steadily poured And watered the trees first fruit The cycle now restored   '),(20725,3711841,'21,January,2004','2004-01-21 00:00:00','    \"Havelock\'s Underwood puts mistakes away, just plays the game\"    By Corey Friedman   HAVELOCK, N.C.--A short memory isn\'t necessarily a bad thing for Janiece Underwood.   urlLink Full Story   Published Thursday, Jan. 22 in the New Bern, N.C.  Sun Journal   '),(20726,3711841,'21,January,2004','2004-01-21 00:00:00','    Moving On    By Heide Hoffarth   Our time has come, a time to move on; A time to reflect on the past and look to the future. As we leave this place, may we be filled with a sense of His grace, As we embark on this new journey in our lives. When we think about the past, we laugh and cry, But now we have to leave, and I don\'t know why. We will leave our mark on the world, As doctors, painters, and businessmen. And as we think about the latter, we think about when We will see eachother again Face to face, after we have left our mark  And come to realize that the real world  Isn\'t all we thought it would be, But that we have the strength in and of our  God to help us see that we did the best  That we could- thank you Class of 2003.   The author is a student at Luther College in Decorah, Iowa.   '),(20727,3711841,'21,January,2004','2004-01-21 00:00:00','    The Last Bough    By Eric Heppner   Crash, falls the second to last One last limb, one last branch Just one arm Holding dear to the sky One more cut And the tree will die  One hundred years The tree has stood A few more seconds Dead rot of wood  End as beginning Death like birth Just one leaf Tying life to the earth  The last bough now Is bending down Soon to be scraping the leafy ground   '),(20728,3711841,'21,January,2004','2004-01-21 00:00:00','    Sweet    By Eric Heppner   The days outside Spent in the sun When the days were long And life fun  The cherry juice That stained the lips And sped life along On sugars whips  The secret hugs Behind sheds of wood When love was young And understood  The awe thatâ€™s felt At the spread of wings When it is realized how Big life really is  The speed that is found When pedal is moved Where freedom is born And confidence proved  The strength of atlas The softening bliss The simple sweetness Of just one kiss   '),(20729,3711841,'21,January,2004','2004-01-21 00:00:00','    Faith    By Eric Heppner    Iâ€™ve never seen the wind Iâ€™ve never smelled the breeze I cannot taste a zephyr But Iâ€™ve seen it move the trees  Iâ€™ve never seen my heart  Iâ€™ve never felt its heat Iâ€™ve never seen it break Yet I can hear its steady beat  I havenâ€™t seen a thought Thought pictures I can see I cannot hear my mind Though something talks to me  I havenâ€™t seen the weight of time  I cannot hear age I do not have its great wisdom But Iâ€™ve seen an elderâ€™s cage  Iâ€™ve never heard love I cannot see through its mist Iâ€™ve never smelled that sorrow Still stones beat when they are kissed  Iâ€™ve never seen God Iâ€™ve never seen his streams Which wash away the sins of men But Iâ€™ve seen him grow the trees   '),(20730,3711841,'21,January,2004','2004-01-21 00:00:00','    ABOUT THE RHUBARB CULT   The Rhubarb Cult is a for-profit organization of freelance writers based in Holiday, Fla.  The members\' affinity for rhubarb pie sparked the unusual name. Primarily, this blog will host poems, short stories, news articles and other works of fiction by Eric Heppner, Corey Friedman and other featured contributors.  The Rhubarb Cult accepts editorial submissions, but does not guarantee publication on the Rhubarb Cult Blog.  Responses to political and social commentary are encouraged, but we have one rule--nasty letters must be written in limerick form.  Excelsior, Corey   '),(20731,3711841,'28,January,2004','2004-01-28 00:00:00','    Venom of Lethe    By Eric Heppner   She is the spider that travels through the shadows of people standing still she seeks to steal their life away and put off dreams till another day  her victims wont scream or kick or writhe teh web is too soft and her kiss so nice  the will sit blankly with eyes ice cold until slowly their muscles turn from stones to mold  till the bones grow soft and the fat is plenty for that\'s when she\'ll come to drink them empty  and still they\'ll sit gently unaware that their life is being crushed away in apathy\'s vice  and as they stare upward lethe\'s venom still soothing their weak arms are tied down never again to be moving  finally the spider reveals her head showing the victims her full ugly name which is lethargy handmaiden to pain   '),(20732,3711841,'26,January,2004','2004-01-26 00:00:00','    Roller Coaster    By Eric Heppner   Tickety tock, go the rails before me As the car climbs squeakily its well tread course Heavy blow the bellows about my head As all anticipate the summit ahead The wind about goes silent The air thins as echoes fade The bellows hush their breath  And give a prayer of deliverance Tick tock stops as the rail ends As the car halts, the air goes dead I revel in the majesty of all Knowing to well I\'ll have to fall Fall from the crest, dive into life Scared of the twists and turns and strife Here I sit at the top of the earth Waiting for life to burst into earth   '),(20733,3711841,'26,January,2004','2004-01-26 00:00:00','    \"Making the cut\"  Eighth grader selected to represent U.S. in international volleyball competition   By Corey Friedman   NEW BERN, N.C.---Ashley Williams was not sure whether to be excited or nervous. Williams was apprehensive when she attended a Nov. 17 meeting after learning she had been nominated to represent the United States in an international volleyball competition in Australia this summer.   urlLink  Full Story    Originally published in the Jan. 26  Sun Journal.   '),(20734,3711841,'07,February,2004','2004-02-07 00:00:00','    THE CRITIC    Swimmer can be harsh on himself as he strives for perfection in pool    By Corey Friedman   NEW BERN, N.C.---New Bern swimmer Zac Smith expects to churn up some waves today as he sets out to silence his harshest critic--himself.  Disappointed with his time in the 50-meter freestyle, down just a tick from last year\'s, Smith is pinning his hopes on redemption today, when the Bears test their mettle in the 4-A regional tournament taking place this weekend in Raleigh.    urlLink Full Story   Published in Saturday\'s Sun Journal.   '),(20735,3711841,'03,February,2004','2004-02-03 00:00:00','    Search for Silence    By Eric Heppner   I long to find the music called silence i sit in the pianissimo hours of the morning longing for silence the click of a jumping anole outside my windo shatters the quest yet as i strain my ears in another pursuit the most beautiful sounds wnader about the room the leather chair clacks rhythmically  beneath my lethargic form the steady hum of electricity burns behind the lights turned off the sudden frump of dirty clothes that have waited all night to settle to place the wondrous footfalls of a cat who prowls the the early hours, never sleeping the snap of a flea\'s mechanical legs and the unsettling plop of it\'s landing space yes, i have searched for silence in the very dead of the morning yet as i open my ears hoping to find the room still i find the unexpected, unheard sounds hoping for an open ear   '),(20736,3711841,'02,February,2004','2004-02-02 00:00:00','    CONTEST CLARIFICATION   By \"miscellaneous prizes,\" we mean whatever cheap crap Corey finds in his trunk, or possibly one of Eric\'s old shirts.  Not that this should discourage you from entering.  We\'ll include some heavy-duty detergent.   '),(20737,3711841,'02,February,2004','2004-02-02 00:00:00','    RHUBARB CULT POETRY CONTEST   The Rhubarb Cult is proud to announce its first-ever poetry contest to inspire creativity and reward budding and accomplished poets with due recognition.    Entries may range from 75-500 words and may be written in any manner about any topic.  Please submit your entries via e-mail before Feb. 10.  Judging will take place on the 11th, and winners will be awarded miscellaneous prizes.    Winning entries will be posted on this blog for all to see.  Good luck, and get writing!   '),(20738,3711841,'11,February,2004','2004-02-11 00:00:00','    2004 Rhubarb Cult Poetry Contest Winner     False Reflection    By Heide Hoffarth    Why do I conceal, why do I hide? All these feelings I\'m feeling inside. Why do I feel so much pain? All the blood rushes through my veins. Why do I feel like I\'m hurting another? All I am is a mask with a cover. Why cover something I can\'t even see? All because I don\'t want you to see the real me. Why do I feel like I\'m wearing a mask? All of my feelings I no longer grasp. Why does all of this run through my mind? All the answers; too hard to find. Why try to look when I can\'t even see? All I see is a reflection of me.   The writer is a student at Luther College in Decorah, Iowa.   '),(20739,3711841,'20,February,2004','2004-02-20 00:00:00','    West Craven boys, girls fall in conference tournament    By Corey Friedman   DUDLEY, N.C. -- The West Craven Eagles need to forget Wednesday night almost as badly as Southern Wayne will want to remember it.  In a low-scoring contest between two dominant defenses, the hosting Saints squelched West Craven 48-42 in the opening round of the Eastern Carolina Conference tournament, sending the Eagles into next week\'s state playoff match-up against West Brunswick with a must-win mentality.    urlLink Full Story   Originally published in the Sun Journal Feb. 20, 2004.   '),(20740,3711841,'04,March,2004','2004-03-04 00:00:00','    2004 Tennis Preview    \"New Bern has great expectations for season\"    By Corey Friedman   It\'s time to replace the question marks with exclamation points.  Questions have been swirling around the New Bern tennis program for some time now; questions of when the team\'s potential will be realized, questions of when the Bears will contend for the chance to bring home a Big East conference crown.    urlLink Full Story    urlLink \"For Bradley, it\'s a family affair\"    urlLink \"Swinging into action\"    urlLink \"Havelock hoping to keep making strides in conference as a team\"   All stories originally published March 3 in the Sun Journal.   '),(20741,3711841,'29,February,2004','2004-02-29 00:00:00','    Where Iâ€™m From    By Eric Heppner   Bible stories, and nursery rhymes, Engine rumbles, and roadway signs, Quiet days, and happy lives, Gentle words, and pantomimes,  Shadows of an elder brother, Simple needs being covered, Shows of love to another, Satisfaction with the others,  Climbing trees, and grassy yards, Wooden slivers, and glassy shards, Enjoying the easy, and passing the hard, Endless nights of playing cards,  Nowhere specific, anywhere, home, Dead end friends that forced me to roam, Seeking redemption, seeking to be known. Nowhere specific, anywhere, home.   '),(20742,3711841,'29,February,2004','2004-02-29 00:00:00','    The Long March    By Eric Heppner   March, march, march Left, then left, then right. Death is beating across my back, But I will survive the night.  My boots are worn Through the soul, Yet my feet are tough, And my mind is whole.  The stench of death Runs down the road, But my eyes stay straight; My spirit stays bold.  I cannot die, For my will is too strong, My mind is too sharp, My life will be long.  I will see my homes again: The Southern sands, The northern mountains, For I protect those lands.  I stride with the confidence That I am far away From that aged field Where my deathbed lay.  I will not die in this war; My life is too great. I will change the world Before I meet my fate.  With this confidence,  My battles are won; My heroism shines; My medals are hung Back in my tent For another time, Another life, Another design.  For now there is duty That must be done; Victories to be had; Battles to be won.  March, march, march, Has led us to the field Where our valor is shown In the weapons we wield.  The enemy is here, Weak and scared, Our guns are fired. Their ranks are marred.  I laugh at them, Defying my death. As their bullets fly, One enters my breast.  A mistake has been made;  I fire to rectify The fool it was  That let it fly.  My breath is short, Cut small into gasps That spit up blood Along with my rasps.  It will not be; My day has not come, Yet I can feel the darkness Begin to overcome The will I had When the war had begun. I lie on my back; And canâ€™t see the sun..  â€œNo!â€ I scream. I will not die here. I will not give in To this cadaverâ€™s fear.  â€œMy life is not spent!â€ I yell. As I rise from the ground, I hear the victory shouts-- What a beautiful sound.  Later that night, In victoryâ€™s camp, I lie back waiting For my breath to revamp.  I look to the doctor, Yet no time can be spared For an able soldier With breathing impaired, Whose limbs are still strong, And trigger finger quick-- For many are dying, And many are sick.  There are legs to be severed, And bones to be set; Cuts to be bandaged, And makers to be met.  And in all the turmoil, I am transparent, With which my unwelcome Is made apparent.  I cannot feel my breath; My lungs feel halted; My chest seems still; Yet my senses must be faulted, For still my mind works, And my legs do their purpose, For my life is too important For ailments to surface.  I see a man crying, I ask what heâ€™s lost; He says, â€œwhy must war Have such great cost? Cause so much death, And pervert so many living? Why canâ€™t men find peace, Joy, and thanksgiving?â€ â€œBecause men are humanâ€ I say with a smile. â€œThere will always be that which is vile.. Which we must destroy-- To preserve our world And keep our homes Safely curled, Away from the destruction That would destroy the race And take the beauty of humans From the planetâ€™s face.â€ But the man keeps crying As if he had not heard, Or thought the reasons Iâ€™d given Were purely absurd.  So I move on past the man With my victorious mood-- Off to the mess To acquire some food, For inside I am empty, That I cam feel, And I need the comfort Of a good, hearty meal.  As I sit at the table, Behind potatoes, and meat, Corn, and green beans, And Cake as a treat, Behind baskets of rolls, And puddings, and pies, A whole plate of cookies, Yet all of them are lies. For as I bite into the meat, The flesh turns to soil; The potatoes to manure; The green beans are spoiled.  My emptiness roared As my bare stomach wrenched; My mind filled with fury, As my fists fly clenched, At the table of plenty And the food flies off the table Onto the ground, Around men who were able To enjoy the meal. Without convulsing in pain At the feel of food for the living On the tongue of lifeâ€™s bane.  I did not feel my heart, Yet it could not be true-- My life was too vital, So I flew from the room.  I am running far Without breath or weary; Without the feel of the wind, And the burn of muscle searing.  I run through the night, And with the first light, I find the harbor That would aid my plight, For it held the ship That could bear me home, Where I could assuage my fears And not be alone.  The ship is called sorrow; Which is aptly named, For it bears the souls That death has claimed.  I wandered the ship, Whose job was to ferry The unlucky ones home To sadly be buried.  At the end of my voyage, As I roamed through the hold, I came upon a box That added fear to my cold.  The box held markings That I was sure I knew, Yet I seemed to have forgotten For my memories are few.  As I stared at the letters Carved into the wood, The meaning came suddenly And I understood.  The words were clear: Robert M. James. I ran because It was my earthly name.  Over the side of Sorrow,  I do not care If what I run on is nothing but air. I have to find The one who would know What had happened, why my reflection wouldnâ€™t show.  When I reached my home, I ran to my girl-- She sat in a chair With letter unfurled. It read from my commander Who had to have lied. For it gave the message That in battle I had died. I try to lift The tears from her face, Or feel my love In her embrace. But the tears still flow; my embrace was cold my hands arenâ€™t there And my time is getting old.  I look at the face Of my darling wife. And can feel the tether break That ties me to life.  My spirit is failing I have nothing left. My mind is finished Deceiving death.   '),(20743,3711841,'09,March,2004','2004-03-09 00:00:00','    \"Havelock baseball rolls past SW Onslow\"    By Corey Friedman   HAVELOCK, N.C.--Rams head coach Chad Smothers said his team didn\'t play as well as expected Monday afternoon, but the scoreboard said otherwise.  After a sluggish start, Havelock ignited in the early innings to spring ahead by six runs, eventually easing past Southwest Onslow 8-2 to remain undefeated on the season.  \"It was good enough, but it wasn\'t a great, great effort,\" Smothers said. \"We played hard, but it wasn\'t as sharp at times as we need to be later on. That\'s a little disappointing, but I thought Chris Holden pitched pretty well--he did a good job today throwing strikes.\"    urlLink Full Story   Originally published and copyrighted in the March 9 Sun Journal.   '),(20744,3711841,'16,March,2004','2004-03-16 00:00:00','    The Last Bough     By Eric Heppner   Crash, falls the second to last One last limb, one last branch Just one arm Holding dear to the sky One more cut And the tree will die  One hundred years The tree has stood A few more seconds Dead rot of wood  End as beginning Death like birth Just one leaf Tying life to the earth  The last bough now Is bending down Soon to be scraping the leafy ground   '),(20745,3711841,'14,March,2004','2004-03-14 00:00:00','    Sweet     By Eric Heppner   The days outside Spent in the sun When the days were long And life fun  The cherry juice That stained the lips And sped life along On sugars whips  The secret hugs Behind sheds of wood When love was young And understood  The awe thatâ€™s felt At the spread of wings When it is realized how Big life really is  The speed that is found When pedal is moved Where freedom is born And confidence proved  The strength of atlas The softening bliss The simple sweetness Of just one kiss   '),(20746,3711841,'14,March,2004','2004-03-14 00:00:00','    Satisfied     By Eric Heppner   The breeze runs softly Its gentle race The tall grass softly Caresses my face The sun is low The sky is red The conforming soil Is becoming my bed The work is done The stress has died I close my eyes And am satisfied   '),(20747,3711841,'14,March,2004','2004-03-14 00:00:00','    Perspective     By Eric Heppner   A mole of atoms of gold Could make a nice ring A mole of pennies Would make anyone The richest person in the world For the sheer volume of copper  A mole of single grains of rice Would feed the entire world A million times over A mole of marshmallows  Would suffocate the sun Without melting half  A mole of earths Would draw the entire galaxy Upon itself And would look like a grain of rice In the lensâ€™ of the biggest telescopes Of the next galaxy over   '),(20748,3711841,'14,March,2004','2004-03-14 00:00:00','    \"Homegrown Products\"    By Corey Friedman   WINTERVILLE, N.C. -- DÃ©jÃ  vu took on a whole new meaning last Saturday.  Four homegrown athletes squared off on the baseball diamond last weekend as Pitt Community College hosted Louisburg College in a National Junior College Athletic Association doubleheader.  Wearing PCC pinstripes were former Havelock High School standout Phil Schandorff and Zach Williams, formerly of New Bern High. In the opposing dugout sat Kenny Flythe--dubbed K-Fly during his time as a Bear--and West Craven\'s Mike Ange.   urlLink Full Story    NOTE:  Corey apologizes for the appearance of \"hogging\" the blog, but if Eric would stop WRITING poetry long enough to POST it, we\'d have more of a balance between fictional literature and fake journalism.   '),(20749,3711841,'18,April,2004','2004-04-18 00:00:00','    Unalterable Ocean    By Eric Heppner   One day a man, With staff in hand, Marched arrogantly against the Ocean.  He cast aside the rod and said â€œYou will kneel before my head, For I will conquer you Ocean.â€  â€œHow can you be such a fool? Think me not a little pool, You cannot win for I am the Oceanâ€  Then bold, the man strode into the surf, Defiantly stomping to foamy turf, Saying â€œI will break you, Ocean.â€  â€œCome then oh tiny mortal, But beware for I give no safe Portal, Today you dance with the ocean.â€  Then the breakers assailed And the waters prevailed Casting the man to the briny depths  From crest to crest Poseidon gave him no rest As he screamed for unreachable breaths  Then Neptune regurgitated in disgust This weakling who could not take one gust From the lungs of the mighty Ocean  A broken man on the beach wept, â€œWhy oh why could not I have kept, My pride away from the Ocean.â€   '),(20750,3711841,'18,April,2004','2004-04-18 00:00:00','    Walking    By Eric Heppner   Out the window life goes by Images blur into another Itâ€™s a long lonely trip But my motor keeps whirring The motor that halts just long enough For me to walk around at night To walk the cramps out of my legs Yet when morning breaks My vehicle starts up again And I go off again Like it or not  Iâ€™ve seen the world Itâ€™s a big place I once that I could see it all But the first steps of my journey Claimed ultimately that impossibility Still I drive on against impossibility Hoping to see enough Before my motor quits Before my vehicle is too hold and broken Before I lose heart  Through my best efforts, age sets in And Iâ€™d still like to see that first museum again To catch the things Iâ€™d missed Though that is a very long list But the hydraulics donâ€™t work any more But my journey continues, slower than before  Motion is slower, Iâ€™m hardly moving at all Though I still get out at night and walk without walls I try to go as far as I can, before the morning calls me back To the vehicle that is old, and worn, and slack  One morning my vehicle, didnâ€™t start at all I came back from my walk and looked at it for a while Other vehicles came by, and buried my car I just kept walking, didnâ€™t matter how far   '),(20751,3711841,'18,April,2004','2004-04-18 00:00:00','    Away    By Eric Heppner   Float me away For far southern shores Dressed in white cotton On a boat without oars Bare me away To where it never grows cold To the air that is balmy And the wind that is bold Lets go away To where the high heat is soothing Where the dry sand conforms And the palm fronds are moving Flat away, far away With the sound of the waves Away from the noises To calm lulling lathes Slip away, slip away Without a thought in the mind To that place far away Where living is kind   '),(20752,3711841,'27,April,2004','2004-04-27 00:00:00','    Nede    By Eric Heppner   	There was light in existence for the first time. The translucent brilliance of the womb, tried my new eyes. There was motion as well. Motion and light enveloped me. Then finally, there was pain, pain beyond thought, pain that preceded the darkness. 	I awoke from the darkness into a light that did not hurt my eyes. Stepping forward by instinct, I felt my own form. It was bigger than I had though, or remembered. I could see, smell, and distinguish touch. What I saw I understood as naturally as a lifetime of familiarity. The solid trees, soft soil, and ever-moving water were familiar and beautiful to me. I wandered in the world, without pain. My mind gave me strange visions of a life I had never had. 	The people of the city were all like me. They all wandered in and out of the different houses that belonged to no one. There was all manner of food that appeared without work on long tables. Yet, there was no desire to eat it, for there was no hunger in the city. People slept in large soft beds, despite not being tired. No weary pulled on the peopleâ€™s eyes. No blissful ache invaded the muscle, no matter the toil or strain, of which there was none. The perfect cityâ€™s name was Nede. 	The people of Nede, like me, all knew the lives they would have had, if they had not come here. Some would smile, when they first saw the life of hardships that they had bee spared. All though, eventually turned sullen as they saw what could have been, the choices left unmade, and things left unchanged. 	I know nine musicians, whose lives would have created beauty on the Earth, where there was none before. There are athletes who have never known the heights of victory and the resilience of defeat. I see innumerable lovers who will never know a living relationship, the warmth of a loverâ€™s embrace, or the trust and fulfillment that comes from a loverâ€™s silence. I have the friendship of a woman who would have been a great healer. She would have cured the plague of cancer from the world. Now she sits under the trees. Unable to save the lives of the people she has never been able to meet.  	I am often asked, around the dinner table, what would have become of my life, had I lived. I would have been simply a teacher. I would have been great to a few. I would have had a wife and two children. Though now, I have nothing. I never made it past birth.   '),(20753,3711841,'28,May,2004','2004-05-28 00:00:00','    Satire    Kobe Bryant Crowned Lakers\' MVP    By Corey Friedman   LOS ANGELESâ€”His wide-toothed grin gone the way of his signature Afro hairstyle, a furious Kobe Bryant retreated from the Lakersâ€™ locker room to sulk yesterday after being unanimously voted the teamâ€™s Most Valuable Pervert and awarded a trophy depicting an electric toaster hovering above a bathtub.     \"I can honestly say that no one deserved this more than Kobe,\" said Los Angeles starting center Shaquille Oâ€™Neal. \"The little freak should do us all a favor and take the trophy\'s suggestion. But out of all the pervs in the Lakers\' line-up, Kobe is by far the most talented athletically.\"     The gutless worm is currently standing trial in Eagle, Colo. for the rape and sexual assault of a hotel employee. An insistent coward who refuses to own up to his repugnant actions, Bryant entered a plea of not guilty.     The MVP award was presented before a capacity crowd at the Staples Center yesterday afternoon. Child molester and musician R. Kelly was scheduled to perform at the ceremony, which was cut short when the brooding baller bolted.     Bryant priggishly refused to answer to reporters\' questions after the awards ceremony, acting like the repulsive bottom-feeder he is.     \"Even though looking at Kobe Bryant makes me want to vomit out my innards, thereâ€™s no denying his ability on the basketball court,\" said Lakers head coach Phil Jackson. \"This sicko has a gift for basketball. He truly is the most valuable pervert we regret having on the team.\"   '),(20754,3711841,'12,June,2004','2004-06-12 00:00:00','    All Night Murder  She picks up the pen, so she can write. Watching as the day turns to swift night. Blood vessels clench, hands turn white. Hate fills her soul, holds her so tight there\'s no breath, there\'s no air. Just the sensation of nausea and that feeling of despair as she lifts her head and looks at him bound by rope, gagged by sin. The rage holds her in grip so strong as she slowly brings her insturment down. No longer a pen, now only a knife built to stab and cut and slice. Strangled noises beg for release, until one swift jab causes him to cease.  He is silent now, just waiting for the end, as blood drips down and pools chalk thin. Agony builds, muscles strain, and only the gentle pouring rain covers the helpless screaming he\'s doing when she cuts open his skin in magnificent fury. Intenstines glisten, white bone gleams, and still she ignores the heartfelt screams. He swears he\'ll never do it again but she says nothing, and stabs him again. He promises he won\'t ever hurt her, As she turns the hours into all night murder.   '),(20755,3711841,'22,June,2004','2004-06-22 00:00:00','    One is one I came up with earlier tonight. well here it is.   Missing you is like missing your heart.  Because without your heart you have no Love.  Without Love you have no emotions that go along with it.  Fear, not there without Love you have nothing to fear about losing.  Anger,You are always Angry at someone you Love, its just you keep that loved locked away for a little bit.  Jealous, you are Jealous of those who get to have the thing you love or loved.  Right now I feel None of these, just blind rage. Dont you see? You drive me crazy, the way you Love and comfort me. But I dont have you right now.  I have no comfort.  I have no Love from you.  You arent alive right now to me. I would let a tear drop but then my memories of you would blur. One thing I am sure I feel is Pain.  The Pain of not having you by my side.  I am lost like a child in the dark, crying for his mother to comfort him, like you once did to me. I feel like a bug that got to close to the beautiful flowers.  Only to be poisoned by something bigger.  Now slowly I am dying.  My only Wish is to have you, with me as I die.  I Miss you Liz!Where ever you are   '),(20756,3711841,'22,June,2004','2004-06-22 00:00:00','    This one is called \"Our Love\" I came up with it this one morning before school.   Our Love is pure.   Our Love is not blind nor selfish.  Our Love is crisp as a spring morning.   Our Love is like the dew on a blade of grass in the morning.   Our Love is red like a rose, bright and livley.  Our Love is bud of a flower waiting to blossom to full potential.   Our Love is untainted.  I Love you Liz! I hope you had a good day   '),(20757,3332079,'17,May,2004','2004-05-17 00:00:00','    mY fWens are leavin mi...wads e prob wif mi...m i too sensitive...?? but i dun tink so.. its jux dat i wana b wif my fwens..yet dey find it irritatin or wadeva....i\'m sorri if i gav u dat kind of outcome...i didn\'t expect dat....all i wan is a fwen hu i can trust n trust mi...will b dere when i need dem...ish dat so difficult to find such 1.... i\'ve no idea...i tot i already found 1 when i met ya...but it turn out e other way... e first few weeks i\'m happi..i haf no isea if u r happi a not... but i tink u r not...otherwise u wun b not botherin mi tis few daes.... u hang arnd wif yr other grp...not carin fer mi le...ish tis e outcome fer treatin u too gd..?? i jux wan my fwens to b happi... ish dat a mistake..?? ish dat so hard to achieve...?? y cant i b lik other people... u tink i wana stick wif u if i\'m not happi..mi wif u happi so i follow u...mayb u find mi irritatin...or wadeva...now i dun seem close wif u le...we dun even tok fer more den 3 mins.....i dun wish our friendship to turn out lik dat... i duno wad tinks i did ish wrong..everi1 seems to b avoidin mi... not talkin to mi..i feel so lonely..pep sae i shld b independent.... i dun wish to b....its so lonely...   '),(20758,3332079,'15,May,2004','2004-05-15 00:00:00','    mi iSH a vEri sensitive person...i can sense mani tinks..most of e time ish accurate but ish not good news...mi sende dat...i realli got a problem...a huge problem..!! which cause mi to lose mani fwens...mi realli sad..realli duno wad to do... wad can represents my sadness..wo de wu nai...wo zhen de bu zhi dao yao zhou she mo..mi realli need to b independent...or else i will lose all my fwens in no time... but i dunoe how... can ani 1 help mi..?? i realli need help...!! jUx fEelin sAd aT dIs mOmenT..........(!o!)   '),(20759,3332079,'22,June,2004','2004-06-22 00:00:00','    What\'s love? When you are together with that special someone, you pretend to ignore that person. But when that special someone is not around, you might look around to find them.  At that moment, you are in love.           Although there is someone else who always makes you laugh,your eyes and attention might go only to that special someone. Then, you are in love.   Although that special someone was supposed to have called you long back, to let you know of their safe arrival, your phone is quiet. You are desperately waiting for the call! At that moment, you are in love.  If you are much more excited for one short e-mail from that special someone than other many long e-mails, you are in love.  When you find yourself as one who cannot erase all the messages in your answering machine because of one message from that special someone, you are in love.  When you get a couple of free movie tickets, you would not hesitate to think of that special someone. Then, you are in love.  You keep telling yourself, \"that special someone is just a friend\", but you realize that you can not avoid that person\'s special attraction. At that moment, you are in love.   '),(20760,3438987,'30,May,2004','2004-05-30 00:00:00','   I\'ve now lived in the space between an old way of life and a new for 7 weeks.  Rather than experience change as a single instance, I took a break between the end of my old experiences and the beginning of new ones.    Leaving the old wasn\'t hard but I\'m about to jump into the new (well, the short-term new) and it scares me.    I don\'t know what exactly is so scary about it-- I will be seeing familiar people in the next 2 months, staying in my country, and returning to a familiar city.  But I find myself crying for no reason, thinking at times of turning back, and almost dreading the fantastic summer I\'ve planned.  A known quantity is always less scary than the unknown.    This is the last deep breath before the plunge.  I\'ve done similar things before, and I know the pattern.  When I get in the car to leave DC, I will cry, maybe a lot.  Then, worn out and calm again, will resign myself to the next hours-- not excited, but doing it, doing what I had planned to do.  Slowly over those hours, the resignation will dissipate as I remember that not only did I choose to take this trip, it is what I dreamed of doing for 2 1/2 corporate years, what others keep telling me they\'d love to do, and what I may never have the chance to do again.  Before being on the road becomes routine, I will enjoy the juxtaposition of grief and excitement, of leaving the old and moving towards the new.  '),(20761,3438987,'26,May,2004','2004-05-26 00:00:00','   First hour on the road :  Depending on the time of day, some degree of pop is appropriate; happy-to-be-driving (or need-to-wake-up) music.  Try U2\'s Pop, Flaming Lips, or (gasp) top 40 radio.   In the groove :  Calming, Zen-- you\'re changing lanes with ease, thinking about putting it into cruise.  Try Beck\'s new album, classical if you\'re so inclined.   OK... :  You\'ve been driving awhile.  Well into the trail mix, you feel the need for a little energy-- angry energy.  Although you\'re not angry, you listen to keep from looking at every mile marker, totalling the number of hours your legs have been in a single position, or blowing up at the crazy truckers.  So many options here:  Def Lepperd, Bon Jovi, Radiohead, Eminem...truly endless.  You could drive angry for awhile  You think you\'re back to Zen :  and you might try something a little calmer.  But you\'re wrong.    Can Not Wait to get out of the car :  Nothing will do here but oldies/soul-- you\'ve got to belt it out with Diana Ross, Marvin Gaye, or Aretha.  As you pull off the highway, you\'re back to happy.  '),(20762,3438987,'30,June,2004','2004-06-30 00:00:00','   I imagine it is impossible to stand out in Portland, just as it is impossible to look like everyone else.  When I pierced my eyebrow and went to work at the Corporate Executive Board, I stood out.  Here nobody would blink an eye, and I might even fit in a bit better.  There\'s a weird sort of homogeny to a city full of nonconformists.    Yesterday I did everything Portland-- spent a lot of time in Powell\'s City of Books (where they have a ceiling-to-wall shelf full of books about FDR and Eleanor Roosevelt), bought fruit from a farmer\'s market, sat outside with a cup of coffee to watch people, went to yoga (where I almost passed out, then walked into the men\'s changing room...luckily I turned around in time to avoid too much awkwardness and luckily Portlanders are pretty chill anyway), got a $1.50 haircut (which the guy who\'s cut my hair twice in DC will groan over...he\'s been \"nursing my hair back to health\").  That evening I did everything Hostel-- sampling the current resident cook\'s masterpiece, helping out with a sidewalk chalk masterpiece, and talking for hours and hours on the huge porch.  I could stay in Portland for awhile.   I really wish I had brought a tape recorder to keep record of what I\'m thinking.  When I\'m not driving there\'s too much going on to journal intensively, and I can\'t possibly bore you all with the waves and waves of thought about life, travel, work, the US, etc. etc. etc.  If I could just talk as I drove (when I\'m thinking most anyway) I might be able to capture those thoughts (I\'m always brilliant when I don\'t have a pen around).  '),(20763,3438987,'29,June,2004','2004-06-29 00:00:00','   I can\'t imagine a better shower than after 4 days of driving, hiking, and sleeping in my clothes.  I was on the road north into Oregon today and couldn\'t handle it any longer.  For those who have never taken a shower at a truck stop (I was one of you until today), put away your skepticism.  Clean floor, clean towels, a secure lock on the door-- and only a bit unnerving to hear the voices of truckers right outside as I showered.    I hiked the old growth forest in California yesterday-- there\'s something comforting about those huge moss-covered trees that seem to melt into the soft ground. I also stayed at a campground where I could watch the seals play in the ocean, then watch the sun set and turn everything pink.  '),(20764,3438987,'27,June,2004','2004-06-27 00:00:00','   urlLink    I made it all the way from right to left on that US map!  An awkward self-taken at the San Fran harbor.&nbsp; urlLink  '),(20765,3438987,'25,June,2004','2004-06-25 00:00:00','   I made it!  All the way from DC (where I saw the Atlantic Ocean about a month ago) to San Francisco!  What a great city-- and even more interesting than usual on Pride weekend.  I\'ve spent a few hours wandering here, and will head north soon, making my way for the next several days along the coast of California then Oregon.    It\'s a pretty cool experience to be on highways across the country.  The hiking and friend-visiting along the way is actually the better part, but harder to describe (hikes in brief:  sky in shades of blue, mountains and hills, tiny alpine wildflowers and other colorful things, chipmunks, etc.  I spend much of my hiking time-- and driving, too-- gasping in wonder, or smiling with this goofy smile because it\'s so great).  Interstates might not show you everything about a state, but certainly something:    Maryland:  OK, I don\'t remember Maryland.  It was the first 3 hours of the trip.  I was playing Def Lepperd.  Pennsylvania:  what other state makes you pay $14 in tolls just to take the only major cross-state highway?   Ohio:  literally the best rest stops I\'ve ever experienced.  Michigan:  I\'ve done this drive so much I can hardly talk about it.  It is what it is.  Indiana:  Indiana is the signal to get excited for Chicago:  only 50 miles to go.  Illinois:  I maintain that Illinois is the absolute most boring state to drive through, even just east to west.  Even in Nevada you can pretend you\'re driving on another planet.   Nebraska:  this state has to borrow landmarks from other states (take this highway to get to the Badlands!) Colorado:  what to say about CO?  I love it!  Curvy moutain roads, high passes, some desert.  Everything wonderful.  Utah:  surprisingly varied-- there actually is a national forest in Utah, as well as salt fields and mountains.  Nevada:  like I said, at least you can pretend you\'re driving across the moon or something.  The only time I used cruise control and air conditioning.  California:  excessive signage telling everyone exactly how to drive.  One long and mostly beautiful downhill towards the ocean.  '),(20766,3438987,'25,June,2004','2004-06-25 00:00:00','   OK, I have to be honest.  For the first time yesterday, I didn\'t really want to be doing this anymore.    When I started planning the trip, I thought it might be hard somehow.  Actually, there\'s nothing hard about it-- I make decisions easily, don\'t need much to live on or be happy, and get to spend as much time as I want, doing the best things in the world.  Sometimes I think I could do this forever.    On the other hand, what am I doing, really?  In the middle of the desert I started to feel like Epicurus\'s collection of particles, bouncing around the country in my little particle car.  If I had a job, a destination, or just something to tie me down, I\'d feel secure with a sense of purpose.    I\'m experiencing a weird paradox in this freedom-- on one hand, I\'m the luckiest person around to be able to do this.  On the other hand, it\'s isolating and I struggle at times to find meaning.  It\'s funny how I\'ve been able to reduce life to the basics (OK, the basics plus cell phone and computer), but that hasn\'t really reduced the complexity of life.  I think I like it; after all, if life weren\'t complicated, where would the excitement be?  I wouldn\'t have anything to think about on the road.  '),(20767,3438987,'22,June,2004','2004-06-22 00:00:00','   urlLink    I made it to the top!  Met up with a group of 3 to hike/climb/stagger/crawl on all 4\'s up the mountain and met another friend to walk down with.  &nbsp; urlLink  '),(20768,3438987,'22,June,2004','2004-06-22 00:00:00','   urlLink    The clouds are to the left and I\'m making my way above them to the mountain on the right (climbing >14,000 foot Mt. Princeton). &nbsp; urlLink  '),(20769,3438987,'22,June,2004','2004-06-22 00:00:00','   urlLink    This is perfection:  organic bunny-shaped mac and cheese eaten at a campsite by a creek. &nbsp; urlLink  '),(20770,3438987,'22,June,2004','2004-06-22 00:00:00','   urlLink    This picture is cool for many reasons.  (1) because it\'s the first pass I went through in CO and my car survived, (2) because it\'s Hoosier Pass and I was born in Indiana, (3) because my Dad\'s a true Hoosier and I\'m wearing a sweater that was his before it shrank and it\'s a few days before his birthday and Father\'s Day and (4) because there are actually two mountain peaks in the background and I summited both when I was last in CO.  &nbsp; urlLink  '),(20771,3438987,'22,June,2004','2004-06-22 00:00:00','   I got to Crested Butte, CO yesterday when the sun was low in the sky, the photographer\'s favorite hour.  I made dinner at the isolated campsite and wondered who belonged in the only other tent at the site.  After dark, a car pulled in with a kayak on top and two men inside, who proceeded to get drunk and loud, and to do something involving hammering until late into the night.    I got up when the sun rose, as I always do, and didn\'t want to lay in bed any longer;  there was frost on the outside of the tent, and I\'d spent the entire night curling into a tighter and tighter ball in hopes of warming up in the sub-30-degree weather.  As I drove out of the campsite I considered laying on the horn for awhile to give the guys my farewell, but there were deer in the campsite I didn\'t want to scare away.  I went for a long hike-- almost 6 hours around the side of a moutain, giving me views of all the wildflowers and the Maroon Bells.    At 1PM I could think of nothing I wanted more than a nap.  Crested Butte has a nice park with baseball fields, tennis courts, and a picnic area.  I avoided all these and the playground, and found a nice spot in the middle of a field.  Perfect-- a nap in the grass on a sunny, 72-degree day.  I was in a warm, sleepy place when a buzzing noise started and the lawn sprinklers went on.  Somehow avoiding a direct hit, I staggered (lightly sprinkled) away from the sprinklers, and lay down again near a deserted basketball court.  Within minutes someone with an unfortunately inaccurate shot came by to shoot hoops.  I decided to ignore this, until I heard two of the playground voices come closer and a kid fairly close to me say \"Just a minute I\'m going to the bathroom.\"  At this point I decided napping in the park just wasn\'t a wise move.    There are all kinds of things to do in this town but for today I\'m not sure I have the wherewithall to do anything but wander around.  Maybe tomorrow I\'ll rent a bike.  '),(20772,3438987,'21,June,2004','2004-06-21 00:00:00','   For two hours today, I had a job.    Actually, I could probably have it back if I want it, and I may.  The regional campground director agreed that I could be a campground host of sorts.  I can\'t be a true host because I have a tent and not a hard-sided camper, and there have been bears around (I\'m not sure why campground hosts can\'t get eaten by bears when the paying campers can).  This means I don\'t really get paid, but do get to stay at my campsite for free.  Lest you think this may be a raw deal, know that they don\'t have cleaning tools for me there (so I don\'t have to do it) and the woman who runs the region will be doing the reservation-related paperwork.  So essentially I have no job responsibilities.  Also they don\'t really care when I\'m there as long as I\'m generally around on weekends.  From what I can understand, I\'m getting to stay at a campsite for free.  The only catch is that the site is at almost 11,000 foot elevation, so there\'s not a lot of warmth, but the director said I could try it for a few days and see if it\'ll work for me.    So after a brief stop at Wal-Mart to look for thermal underwear, I made my way to the campground.  Even though the sky was getting darker by the minute, I thought I could get set up in time.  I reached the top of Monarch pass and turned into the campground, just as it started to precipitate.  I\'m sure at a lower elevation it was raining, but at the campground, huge pieces of ice poured onto my windshield.  I watched the thermometer on my car drop below 40 and looked at the sky.  Usually in the mountains the blue sky is visible beyond the quick afternoon storm, but all I could see this time was darker sky.  The prospect of setting up camp in the hail and shivering in my Wal-Mart sweatpants sounded much less than appealing, even if it was for free.    So I drove to Crested Butte, 1 1/2 rainy hours away, to find the holy grail of hostels which unfortunately is so popular that they don\'t have a bed tonight, but the guy let me use the internet and pointed out some close campgrounds.  It\'s beautiful here, there must be some great hikes, and I saw an organic farm on the way; might see if they would like a volunteer.  And if I want a free place to stay, I know where to go.    Still trying to post those pics...now my computer\'s being difficult!  '),(20773,3438987,'17,June,2004','2004-06-17 00:00:00','   It\'s incredibly easy to meet people hiking, camping, hostelling.  I hiked up a mountain with 3 new friends, and down it with another.  The first few nights camping I had some wonderful solitude, but have also enjoyed the company.  Last night, the teenage boy in the campsite next to me invited me over to make s\'mores with his family, and then the guy camping with his sister, brother in law, and 4-month old baby brought me a plastic cup of wine and wanted to talk about hiking.  The campsite host felt very motherly towards me, according to her husband, and wanted to check in me during the night.  I had coffee with her this morning before I left, and talked about being a campsite host.    I might have a job for the next two weeks!  I\'m wicked excited about it and really hope things work out.  A campsite at Monarch in CO needs a host, and the woman in charge would love to have me there for a few weeks.  The only problem:  I have a tent, not a camper, and there are bears around.  She\'s trying to get it approved with the highers-up and I\'m hoping to hear in the next hour or two, so I can figure out what I\'m doing tonight, tomorrow, next week, etc.    Hopefully I\'ll find a wireless internet connection soon (instead of the library computers) and post some pics because I can\'t possibly describe it all.  I think I combed my hair yesterday but can\'t quite remember when I last looked in a mirror.  I love this!  '),(20774,3438987,'17,June,2004','2004-06-17 00:00:00','   There\'s a pub in Fort Collins where you can go, expecting that your friends will be there.  Start with a group of 4 and it becomes 7, 11...not because everyone\'s been frantically making plans via cell phone, but because that\'s where people go.  Walk your dog regularly enough and people will recognize you, greet you in the street, and stop to talk.  You might ask them to join you as you grab coffee.    One of the hardest adjustments for me after college was not having places to go where you could expect to see someone you knew.  In DC, an evening out requires some detailed choreography in advance if it involves more than two people.  I really like being in a place that lets you have an unscripted evening and still catch up with friends.    It was 54 and rainy here today-- I didn\'t build rain into the summer plans!  I\'m going to a baseball game tomorrow, then plan to head into the Rocky Mountain wilderness for several days.  If there\'s nothing new here for awhile, you can trust that I\'m happily munching gorp and wading in creeks.  '),(20775,3438987,'15,June,2004','2004-06-15 00:00:00','   I want to be a cowgirl when I grow up.  I did when I was 6 and when I was 15, and even now the open spaces, the long times alone and moving, the idea of getting somewhere on my own make me feel just right.  I can\'t tell you what I think on the road-- sometimes I\'m thinking about the big questions, just chipping away at what I think about life.  Sometimes I\'m completely zoned out.  Sometimes I\'m singing along with the music.  Sometimes I\'m replaying memories.  And somehow I can\'t pull apart one way of thinking from the other, they alternate with such frequency.  Really, I think it\'s my brain\'s happiest state.    I drove through Illinois and Iowa Monday, then Nebraska and Colorado on Tuesday.  Covering such distance in a short amount of time really makes the landscape changes clear.  I can see mountains now!  It\'s what I\'ve been craving.  '),(20776,3438987,'14,June,2004','2004-06-14 00:00:00','   When I\'m 84, I would like to enjoy life as much as Dave Brubeck seems to.  I don\'t normally enjoy piano jazz, but Dave Brubeck\'s love for the instrument shows up even in his recordings.  Live, he shuffles to the piano and bends over the keyboard until his neck is almost at right angles to his back, but his hands still move like I imagine they always have.    I went to Ravinia tonight to hear Brubeck and his quartet, who were not quite octogenarians like Brubeck himself, but all had white hair.  They all wore ties, and not one could stand still, even when he wasn\'t playing.     urlLink Ravinia   rivals the lakefront path as Chicago\'s best attraction.  You pay $10 for lawn seats on just about any night of the summer to hear anyone from Dave Brubeck to the Chicago Symphony to Rufus Wainright.  Beforehand, you\'d do well to bring a picnic and lawn chairs-- and during the show people sleep, read the paper, play games, continue eating (and some actually just sit and listen to the music).    If you haven\'t heard Dave Brubeck and his quartet play Take Five, do it now.  There\'s a snippet at  urlLink Amazon.com , but it\'s not a waste of money to buy an entire album.  '),(20777,3438987,'13,June,2004','2004-06-13 00:00:00','   Maybe once I leave Chicago I\'ll have more adventures.    Most (though certainly not all) of my time here has revolved around various family activities, and being with my family skews my idea of an adventure.  I saw an aunt yesterday who heard that I had left my job and said \"I was just starting to think that you were settled and now here you go doing all sorts of other things!\"  She and my uncle run the second generation of the Wunderink farm, and everyone already wonders who will take it over in the next generation.  My grandparents live on the farm with my aunt and uncle, and when we visited this weekend I climbed up to the attic to get a fresh look at where I had slept as a kid.  There were the same twin beds with the same comforters I had used, which were the same beds and comforters my dad and his brothers used as kids.    The farm has changed, even I can see it, but more slowly than the lives I see normally.  If I had stayed longer, I\'d want to ask my grandparents what they thought \"an adventure\" was.  They raised 8 kids on a farm-- really, I imagine every day was an adventure.  In comparison, mine feels a bit artificial.    But while I mostly describe these months as an adventure, there\'s a second component, which I\'ve been soaking up for the last 10 days:  making sure that I connect with my friends and family who live so far away, and whose lives are in such transition.  It was important for me to be with my family when my sister left for Kazakhstan, just as it\'s important to see the farm now that I\'m an adult (or more of one than years ago when I was last there).  It\'s also important to me to connect with friends who are getting married, starting school-- or even going through more subtle changes.  '),(20778,3438987,'10,June,2004','2004-06-10 00:00:00','   My sister never broke down or melted down, freaked out, or even teared up in the week before she left for the Peace Corps.  In fact, I haven\'t seen her so pulled together in a year-- she graduated in 2002 and spend a full year juggling part time jobs, hoping for 6AM phone calls asking her to substitute teach, and creating spreadsheets to figure out how to pay rent, when to take classes, etc.  Her life was up in the air in every way possible for a full year, but when she got her Peace Corps assignment, things all pulled together for her.  Not to say that all her plans solidified (the Peace Corps is notorius for its ambiguity), but the end goal was clear.    This morning it all came together-- all her underwear, boots, and business casual fit into two huge marigold LL Bean bags.  She showered, ate, did one final check through her things.  She hugged my mom and I, stood at the door and smiled a moment, then turned and walked down the hall without looking back.    My parents and I don\'t quite know what to do with ourselves today.  I\'ve thought a few times \"she\'s still on the plane,\"  \"OK, now she\'ll be in DC-- hope her bags aren\'t to heavy to handle\" and other things.  Come Sunday night I will no longer be able to guess what she\'s doing-- she\'ll have landed in Kazakhstan and begun her adventure on the other side of the world.  '),(20779,3438987,'08,June,2004','2004-06-08 00:00:00','   Even driving all across America I doubt I will find a radio station as kick-ass as  urlLink the Loop , Chicago\'s 97.9.  The Loop is the music of sweaty dark gym basements, a station whose idea of rocking is Black Sabbath, Ozzie, and Guns n Roses and who believes it is appropriate to rock all the time.  Its DJs are in it for the music, never checking around to make sure their version of rocking is up to date.  Its sets feature one-hit-wonder hair metal bands as well as the true classics.  The Loop is rock in its purest form.  Chicago deserves the Loop.  '),(20780,3438987,'07,June,2004','2004-06-07 00:00:00','   urlLink    My family during a hilarious game of speed scrabble (I had, without realizing it, spelled \"caddd\").  Note the lack of furniture in their apartment!&nbsp; urlLink  '),(20781,3438987,'06,June,2004','2004-06-06 00:00:00','   My parents moved last December from a suburb of Memphis to downtown Chicago.  This is the first time I\'ve stayed with them and noticed how they translated their life from one place to a completely different environment.  They sold all their Memphis furniture-- what fits well in a suburban Southern home does not in an urban apartment.  I like noticing what they kept-- the illustrated songbooks I used when first learning to play the piano and the big rocking chair we used to bash into the kitchen wall.  They were most worried that they\'d miss chances to be outside (and not feel stifled by the city), but their long weekend walks are on the path along Lake Michigan, and they drink coffee and read the Sunday paper in a city park.  Now, in addition to being less fashionable than my Mom, I also have good reason to feel less urbane!  '),(20782,3438987,'06,June,2004','2004-06-06 00:00:00','   Don\'t hold me to it, but here\'s the rough plan for my summer: driving west then back east, with stops in Ft. Collins, Denver, Salt Lake City, Eugene, Portland, Seattle, possibly Vancouver, North Cascades National Park, Simone\'s cabin, Kathryn\'s parents\' farm, Missoula, Glacier National Park, North Dakota, Minneapolis, possibly Madison and Chicago and Ann Arbor/Detroit again, Niagra, then somehow making it back to DC by approximately mid-August.    I\'m in Chicago right now, and will be until the end of next week, hanging out with my sister who will leave Thursday for 2 years in Kazakhstan with the Peace Corps.  '),(20783,3438987,'30,July,2004','2004-07-30 00:00:00','   Chicago just built a  urlLink huge park  downtown near the Art Institute.&nbsp; According to the Tribune, it\'s \"late and over budget, but nobody seems to care.\"&nbsp; The park features an outdoor amphitheater, gardens, huge walls of water that kids and some adults play in (ooh, I wanted to play in it), and-- my favorite--  urlLink the Bean , a kidney-bean shaped sculpture (or Tiffany neckace-shaped sculpture) that reflects the entire city.&nbsp;  '),(20784,3438987,'27,July,2004','2004-07-27 00:00:00','   urlLink    The Bean and the city!&nbsp; urlLink   (I didn\'t take this picture, I found it on the Web)  '),(20785,3438987,'27,July,2004','2004-07-27 00:00:00','   urlLink     Taking pictures in the Bean was a fun challenge. Here\'s my brother and I. &nbsp; urlLink  '),(20786,3438987,'27,July,2004','2004-07-27 00:00:00','   urlLink    Funniest campsite ever: there were tents at each picnic table (very close together) and groundhogs (or gophers? Can someone tell me the difference?) in their little hills between the tents. They were very friendly groundhogs. Also see the RV park behind the groundhog park. I was not camping in the wilderness at this point. &nbsp; urlLink  '),(20787,3438987,'22,July,2004','2004-07-22 00:00:00','   urlLink    Can you see how the rocks just cave into each other, and there\'s a waterfall at the fault?&nbsp; urlLink  '),(20788,3438987,'22,July,2004','2004-07-22 00:00:00','   urlLink    The Columbia icefields...just wow&nbsp; urlLink  '),(20789,3438987,'22,July,2004','2004-07-22 00:00:00','   urlLink    Me in my raincoat and plastic pants waiting for the storm to roll through the mountains&nbsp; urlLink  '),(20790,3438987,'22,July,2004','2004-07-22 00:00:00','   urlLink    Check out that rack!&nbsp; urlLink  '),(20791,3438987,'22,July,2004','2004-07-22 00:00:00','   urlLink    This is along the Going to the Sun road. If you haven\'t been to Glacier, there\'s a big hole in your life. &nbsp; urlLink  '),(20792,3438987,'22,July,2004','2004-07-22 00:00:00','   urlLink     Snow and mountain goats! I got some pictures closer up, but they didn\'t have the cool snow. &nbsp; urlLink  '),(20793,3438987,'22,July,2004','2004-07-22 00:00:00','   urlLink    Lake McDonald, a glacial lake. I took a quick dip in it, and it sucked the breath out of me it was so cold. &nbsp; urlLink  '),(20794,3438987,'22,July,2004','2004-07-22 00:00:00','   I had some great pictures to post...of a moose, my groundhog campsite, and the amazing Canadian Rockies.&nbsp; But those will have to wait until I get to Chicago (as far as I can tell there\'s no public wireless internet from where I am through North Dakota and Minnesota).&nbsp; But the itinerary is one more night in beautiful Montana, then to Theodore Roosevelt Nat\'l Park for just a bit, a few nights in Minnesota, a few days in Chicago, a night in Michigan, then back to DC!&nbsp;   This has been an unbelievable trip and this is by no means my last post but it feels like things are winding down and I\'m getting ready to go back.&nbsp; To varying degrees, I miss Rock Creek Park, my house and room, the guy in DC who\'s played millions of Ultimate frisbee games in the last 2 months, and everyone else I know in the city.&nbsp; My life there is by no means stable (read:&nbsp; I don\'t have a job!) but I do know where I\'ll be sleeping and where to get food and how not to get lost (mostly).&nbsp;   Anyway-- cool pictures will be up soon.&nbsp; If you\'re bored in the next few days-- call me!&nbsp; I\'ll be in the car!!&nbsp;  '),(20795,3438987,'20,July,2004','2004-07-20 00:00:00','   The mountains here are gorgeous, but Banff is a bit overdeveloped.&nbsp; I was hiking a trail when I noticed a trail marker for hot springs.&nbsp; The thought of a good soak was nice, so I decided to take the 1.7 km trail.&nbsp; It zig zagged up and up until it reachedâ€¦a huge parking lot and an even more obscenely large hotel/resort thing.&nbsp; I had made my way up one side of a mountain, the other side of which held a road and well-developed resort.&nbsp; Big sighâ€”no dangling my feet in woodsy hot springs for me.&nbsp; The resort wouldnâ€™t take the $5USD I had in my bag and Iâ€™d left my Canadian money (and my swimsuit) in the car.&nbsp;  &nbsp;  Later yesterday afternoon a thunderstorm rolled in.&nbsp; I hadnâ€™t set up my tent yet and I was too dirty to go sit in town, so I headed up to a scenic lookout, pulled on my plastic pants and waterproof jacket, and sat outside to watch the storms roll through.&nbsp; I guess not having a â€œhomeâ€ forces one into more extreme ways of passing time, but it was awesome to watch the storm move between the mountains (hopefully Iâ€™ll post a fun picture soon).&nbsp;&nbsp; &nbsp;  The Canadian Rockies are an entirely different mountain range than the Rockies in the US.&nbsp; Here you can see the striations of rocks going diagonally, curving, and generally painting a pretty good picture of which tectonic plate won.&nbsp; The glaciers just sitting on top of and in between mountains are amazing as well.&nbsp; I drove through the mountains for about 5 hours today and couldn\'t figure out how to react to such stunning and wild beauty with the mountains and powerful creeks and glaciers except by crying.  '),(20796,3438987,'19,July,2004','2004-07-19 00:00:00','   My tent has two poles.&nbsp; The base of the tent is rectangular and the first pole cuts the rectangle into two triangles, and the second crosses it to make four triangles like a PB&amp;J.&nbsp; When I walk around campsites I\'ve seen the marvelous contraptions that people sleep in-- tents with 5 poles, tents that you could do cartwheels in, trailers that look like they\'d make a truck tip over, trailers that pop up and out and people sleep all over.&nbsp;  &nbsp; I walk around my campsite pretty much every evening.&nbsp; It doesn\'t get dark up north until after 10, so there\'s plenty of time even if I get to a site late.&nbsp; Although the pattern varies depending on where I am, here are the several things I do pretty much every time I camp:  &nbsp;- Eat Cream of Wheat for breakfast, sometimes followed by a homemade mocha &nbsp;-&nbsp;Lay in my tent&nbsp;(either at breakfast or in the evening) and read in my underwear (this is a luxury of camping:&nbsp; lots of time to read in your longjohns) - Take a long, fast hike.&nbsp; Sometimes I head out intending to take a shorter or less fast hike and it never works.  - Eat an entire box of Mac and Cheese for dinner &nbsp; I just came up from Glacier Nat\'l Park back into Canada to visit Banff and Jasper.&nbsp; I love Glacier, and highlights of this trip include: - Waking up at 4 one morning and not being able to sleep, so taking the rainfly off my tent and watching the sun rise while I daydreamed - Swimming each night before dinner in the breathtakingly cold glacial lakes - Singing breathy renditions of REM to scare off bears during a trail run  '),(20797,3438987,'15,July,2004','2004-07-15 00:00:00','   urlLink    We just walked behind a huge waterfall-- it\'s to the left in this picture, but just shows up in pictures as a wall of white.  Brian and I are throwing rocks over to see where they land.&nbsp; urlLink  '),(20798,3438987,'07,July,2004','2004-07-07 00:00:00','   urlLink    Simone and Brian at the butte overlooking Mt. Hood&nbsp; urlLink  '),(20799,3438987,'07,July,2004','2004-07-07 00:00:00','   urlLink    We helped work on a deck at Simone\'s cabin (though I think we really just played)-- here\'s Simone in the rented Bobcat.&nbsp; urlLink  '),(20800,3438987,'07,July,2004','2004-07-07 00:00:00','   Although I loved pretty much everything about Vancouver (except the radio), two things especially stood out.  One was the newly-opened  urlLink Storyeum , which told the story of British Columbia on the stage.  I actually didn\'t watch the show, but wandered the lobby for hours listening to the performers and watching them walk through the lobby, looking at old pictures, and reading the captions and quotes from people who were there. The  urlLink UBC Anthropology Museum  complemented this well.  The curators have placed entire collections out for viewing, so instead of seeing one Peruvian basket or Inuit soapstone carving you see hundreds.  There\'s a notebook nearby each display with notes on each-- where it was from, use, condition when obtained, and other notes.  I found it much easier to get a real feel for the area or period by seeing many objects rather than one.  They also display objects that other museums have a hard time categorizing and often just leave in storage-- trinkets made in Turkish prisons, for example.    Today I head back to Oregon to spend a week at my friend Simone\'s family cabin.  This place is legendary among Northwestern folks and I\'m excited to spend a week jumping (and maybe sleeping?) on the trampoline, exploring Hood River, helping out where needed with the cabin, reading, running, hiking, and as my friend Brian says \"laughing so hard my belly hurts.\"    As with most wonderful trips, time picks up momentum as it goes and seems to barrel towards the end.  On one hand, I\'m not looking forward to it, but it is infinitely better to finish a trip I didn\'t want to end than to arrive back in DC sick of the car, sick of my tent, sick of my 3 t-shirts, and completely ready to be done.  At first I wondered how I could possibly fill 2 months and now I wonder how I could have thought I\'d get enough of mountains, cities, friends in that amount of time.  '),(20801,3438987,'02,July,2004','2004-07-02 00:00:00','   urlLink    This sums up my experience in Vancouver well, and I hope it speaks a bit to my life more generally.  It was in the lobby of the Storyeum.  &nbsp; urlLink  '),(20802,3438987,'02,July,2004','2004-07-02 00:00:00','   urlLink    These guys were awesome.  They walked around the Storyeum playing, and I walked around listening.  A woman on stilts dressed as an eagle also walked around and every once in awhile they\'d interact, with her dancing around them and them changing the song a bit to include her.&nbsp; urlLink  '),(20803,3438987,'02,July,2004','2004-07-02 00:00:00','   urlLink    Chinatown:  store after store with box after box of food.  I did not know what a single food item was.  Sigh...so Anglocentric and untravelled am I!&nbsp; urlLink  '),(20804,3438987,'02,July,2004','2004-07-02 00:00:00','   This is a blog for all of you.  I know you want to talk more about music (I guess because you assume since I\'m in the car all the time that I\'m always listening?  You are correct.), and I\'ve heard you want more pictures.  For #1, see pictures below.  For #2, check out  urlLink KEXP  in Seattle.  I\'ve never heard Belle and Sebastian on the radio before, let alone in the same set as James Brown and G-love.  In fact, it\'s been awhile since I heard so many new band names and so much interesting music on the radio.  I was about to start a list of bands to check out, but I realized I can listen on their website and I don\'t lose out on all this after I leave Seattle!  '),(20805,3438987,'02,July,2004','2004-07-02 00:00:00','   urlLink    Pike Place Market-- musicians watching musicians (the many ways to play stringed instruments). &nbsp; urlLink  '),(20806,3438987,'02,July,2004','2004-07-02 00:00:00','   urlLink    This is Portland. &nbsp; urlLink  '),(20807,3438987,'02,July,2004','2004-07-02 00:00:00','   urlLink    Not a bad place to chill and wait for the sun to set. &nbsp; urlLink  '),(20808,3438987,'02,July,2004','2004-07-02 00:00:00','   urlLink    What first looked like a heavily policed BBQ-- turns out it\'s Hell\'s Angels gathered to roast hot dogs somewhere along Route 1 in California. &nbsp; urlLink  '),(20809,3438987,'02,July,2004','2004-07-02 00:00:00','   urlLink    Awesome huge trees.  I felt like a rabbit in the forest. &nbsp; urlLink  '),(20810,3438987,'02,July,2004','2004-07-02 00:00:00','   urlLink    These are fields of salt-- huge chunks, bigger than kosher salt.  I didn\'t quite want to taste it to be sure it was salty. &nbsp; urlLink  '),(20811,3438987,'09,August,2004','2004-08-09 00:00:00','   OK, it\'s an adjustment to be back here.  I think the major change is the feeling that I don\'t have complete control over who I see, where I go, or what I do.  I have to wait for companies to call me back if I want to get hired!  I can only see friends when they\'re not working or otherwise occupied!  I even have to schedule time to see people!  I have to sit in traffic!  I can\'t walk around outside in my PJ\'s!  (scratch that-- I did it yesterday)  If I don\'t like a situation for some reason, I can\'t just leave town.  On the other hand, I drive 15 minutes to see people, rather than 8 hours.  I\'ve slept in the same bed for a week-- and it\'s a bed not a sleeping bag!  I actually run into people I know!  If I can\'t see DC friends this week, I can see them next week!  I can eat a variety of refridgerated and even frozen food!  I dunno.  It\'s an adjustment.  '),(20812,3438987,'02,August,2004','2004-08-02 00:00:00','   ...and I can hardly believe it\'s been 2 months, a nearly perfect 2 months. Somewhere along the way, concepts like Time and Home lost their meaning, and the question \"what do you do?\" signaled the start of a somewhat philisophical conversation. When I quit my job I had the sense that the summer wasn\'t an abberation but the beginning of a lifestyle, and I feel that even more strongly now.  Someday I\'d like to call a place-- not just a town or city, but a specific building, Home. Now, however, things like rent checks and furniture and accumulation of any sort seem distasteful in a way. I hope I have learned to hold tightly to the important things, just as I\'ve learned to let go of the inconsequential.  Introspection aside, here\'s a wrapup of the trip. Miles driven: almost 10,000 (average miles per gallon: somewhere around 48, best in the mountains and worst on the flat stretches) Total trip time: 2 months, 2 days New car name: Quark New CDs acquired: approximately 30 Number of states visited that I\'d be happy to live in: 7 Nights spent scared out of my mind (for ridiculous reasons): 2 New beers to recommend: Alaskan Ale, Fat Tire, Moose Drool My only regret: not having my bike Item I\'d like never to see again:  my green pants Item I didn\'t miss a bit:  blow dryer What I\'ll miss:  waking up in a tent, wearing a sweater in the summer, constant time outdoors, hostel lounges, fresh roadsite/farm/orchard fruit, friends\' couches and porches...on and on  By the way, I keep meaning to say this-- if I saw you on this trip and didn\'t write about it on the blog, don\'t be offended! You\'ll notice I didn\'t really write about anybody. Half the people reading this know you and half don\'t have a clue who you are. We had a fantastic time, but it\'s hard to describe it for an audience. Thanks for your hospitality and for spending time with me-- seeing friends really made the trip.  I\'m not sure what\'ll happen to my blogging habit now. I will probably continue to mull over the trip online, but don\'t feel the need to subject yourself.  You all are wonderful for keeping up with me!  Like I said, friends made the trip-- in person, over the phone, or email.  '),(20813,3682338,'23,June,2004','2004-06-23 00:00:00','   hey--i was bored so i thought i\'d come here and write a little somethin. not much happened with me today. i went shopping for a bit and then my aunt and uncle and cousin came over. we babysat my cousin for a little while but that was fun. i\'m still really upset that i didnt go to camp. its all i can think about. but i try to keep my mind elsewhere. i try to look at the bright side but its kinda hard to. oh well. i\'m still hoping for a job. hopefully someone will call tomorrow. but i doubt it. i seem like i really want a job but i dont. ive already wasted so much time and i couldve had a job months ago but i didnt try. i really need the money tho. but thats about all for now. i cant think of anything else. just the same old boring stuff.lol. oh well. its not like you had to read this. but anyways-im gonna go for now. bYe.   '),(20814,3682338,'21,June,2004','2004-06-21 00:00:00','   hey--today hasnt been too great. i\'m really sad and depressed that i didnt go to camp. i miss it soo bad and i gave up my only chance of ever being chief. i feel awful. but theres nothing i can do now. :( the only thing that i can do is try to take my mind off of it. and im praying that i get my call this week for a job. if i did it would make me feel ok about missing camp because then i could have an actual reason for missing camp. and it would make me feel 10 times better. so im praying for a job this week. lonestar BetTer call me. oh well tho. so other than moping around being depressed today, i went out with my bro morgan to the mall and some other places. it was better than sitting around at home. after that i came home and helped my dad out some outside. and now i\'m here. sitting at the computer. typing in this dumb little thing. even though i know no one cares about reading this stuff. lol. Oh WelL. thats all i can think of for now so i\'ll be back sometime. ByE   '),(20815,3682338,'19,June,2004','2004-06-19 00:00:00','   hey again--this is my second little post thingy and i know that my last one was also june 19, and it was, but it had just turned the 19th (because it was like 12 somethin in the night). i just wanted to let you all know that im not gonna be writin on here more than once a day. yeah. and its not like it really matters but i just thought i\'d let you know. lol. now im confusin myself but hey its all ok. so anyways--today was another boring day. i didnt do anything until around 3:00 and thats when i had to go see my family over at my grandparents house. it was my grandpas b-day and some family was in from tennessee. it was good to see everybody but it got boring...u know how that goes. today was all rainy and it sucked. and im still trying to decide if i should go to camp. i have ONE more day to decide. this is so hard!!! i know no one cares but--i dunno--its just crazy. i also found out some crazy news about my other side of the family but i dont think anyone cares to know. lol. hmmm--so other than that nothing happened today. just one hell of a boring saturday. i guess i should probably quit rambling on and on about my day. nothing exciting i know. so maybe i\'ll be back on here tomorrow and let u know if im going to camp or not. lol. but i\'m done for now. Later.   '),(20816,3682338,'19,June,2004','2004-06-19 00:00:00','   Hey! I don\'t really know why I got one of these things--but--I did. lol. I thought I would just write all the WonDerFullY wonderful things that happen in my boring days. haha. yeah, so today i had my second interview at LoneStar and i thought i was definitely gettin the job, BUT he said it went great and it was a shame i wasnt old enough to be a server but they didnt need a host right now but he was definitely gonna call me when one came open--so now i have to wait..and wait..and hopefully get the job. i dont want to work but i kinda have to. NOW, i have to decide if i want to go to 4H camp or not. its either--go to camp and be chief which i will never get to do again--or--stay home and wait on a call for a job which i need. i dont know what to do. i have two days to decide. it sucks ass. oh well. well i know this is really boring and its kinda makin me bored so, i think i\'ll stop about now. maybe i\'ll be back tomorrow--i dunno--oh well. Later.   '),(20817,3682338,'25,July,2004','2004-07-25 00:00:00','   hello. well i havent been on here for a few days and ill tell you why. because work sucks. yes. i had to go in on friday night instead of saturday because the manager had an emergency and had no one else to come in so she trained me for like an hour and left me all alone from 4:00 to closing and i was left basically clueless on most things. but i managed it. sort of. at least i will be gettin paid more than i would have! so i went back on saturday at 7 and she told me i didnt have to come till 8 and she was still gonna pay me for 7 through 9 so that worked out and when i did go back at 8 i didnt do a thing but talk with her and her husband. they were funny tho so it wasnt as boring as it sounds. so i got paid for 2 hours and didnt even do anything. but i dont go back until wednesday&nbsp;so thats&nbsp;ok. its not a bad job. its pretty boring but at least im makin money. well anyways right now im sittin here listening&nbsp;to music again. today was pretty boring because i didnt really do all that much. but anyways i think im gonna get off of here and find something&nbsp;else&nbsp;to do. ByE.  ~Gravedigger, &nbsp;when ya dig my grave,  could you make it shallow, &nbsp;so that I can feel the rain?&nbsp;---DMB&nbsp;&nbsp;&nbsp;   '),(20818,3682338,'22,July,2004','2004-07-22 00:00:00','   hey. today has been about the most boring day ive had in a looooooooooooooong time. first of all,&nbsp; it&nbsp;rained ALL day and i couldnt&nbsp;do anything outside and then i didnt go anywhere&nbsp;today. i just sat at home being bored. i just watched tv and wondered around the house trying to find somthing to do.&nbsp;oh well. maybe tomorrow will be better. i need to go shopping really bad. maybe i can do that tomorrow. but at 2:00 my moms babysittin my cousin again so i\'ll probably be here around then. hmmmm...i need&nbsp;to do somethin tomorrow&nbsp;because afer tomorrow im gonna be workin and who knows how much im gonna be&nbsp;gone doin that. at least it gets me out of the house tho. this blogs gonna be short. i cant think of anything else. i&nbsp;only came here because&nbsp;i was bored, but oh well. later.   ~&nbsp;So you lit her cigaretteYou\'re feelin\' pretty goodYou think you\'ve got a shotMost girls you probably wouldYeah this is that kind of placeBut she ain\'t that kind of girlYou\'re readin\' it all wrongLet me tell you about her/She only smokes when she drinksShe only drinks now and thenNow and then when she\'s tiredOf bein\' let down by menYou can give her a lightBut it\'s not what you thinkEverybody knows she only drinks aloneAnd she only smokes when she drinksDid you ask her to danceLet me guess, she told you noGot to take her some place quietAnd see how far that goesOh, don\'t take it all that hardWhen she smiles and turns you downFor a complicated girlShe ain\'t that hard to figure outShe only smokes when she drinksShe only drinks now and thenNow and then when she\'s tiredOf bein\' let down by menYou can give her a lightBut it\'s not what you thinkEverybody knows she only drinks aloneAnd she only smokes when she drinks--Joe Nichols'),(20819,3682338,'21,July,2004','2004-07-21 00:00:00','   well..hello again. im bored again and listening to music. lol. like always. this time im listening to another one of my great mixed cds. its not country tonight tho. this cd has a wide variety lol. its got--bush, rehab, ryan gosling ; ) , steve miller band, alice in chains, osbourne brothers (good ol rocky top) (i love my bluegrass) , blind mellon, megadeath, shinedown, rockapella ; ) , phish, nin, and lots of other great bands. yes. its a fun cd. lol. so anyways today i had to get up bright and early at 7:00 am so i could play with my lil baby cousin. then i was gonna go to the mall to talk to the girl about that job, BUT she called me and gave me the job. so now i will be working at the build a bear factory in the middle of the mall! lol. i start training on saturday night. so after that i still went to the mall with my bro to put some other applications in i had just in case this job doesnt work out.&nbsp;then i shopped for a while. after that i came home and my dad was here because hes takin off a&nbsp;few days of work : ) (he works out of town&nbsp;during the week)&nbsp;then we had dinner&nbsp; and then i was so tired that i had to take a little nap. lol. i woke up at 10:00 and made myself get up and then i came to the computer and here i am now. but i think im actually gonna go now because im sick of typing already. hehe. so goodnight!  *Cause I\'m a picker  I\'m a grinner  I\'m a lover  And I\'m a sinner  I play my music in the sun   I\'m a joker I\'m a smoker  I\'m a midnight toker  I get my lovin\' on the run --The Joker--StEvE MiLLer BanD   '),(20820,3682338,'20,July,2004','2004-07-20 00:00:00','   hey hey hey.&nbsp; guess what? im bored again! as you&nbsp;will probably tell by the time you are done reading this but its ok. so right now i\'m listening to music which im AlWaYs doing because its about my favorite thing to do. i would be lost without music. thats sounds all gay&nbsp;but its true. right now im listening to some country because its awesome. well today i got up around 12:00, got ready to go out and get some job applications and what do u know--i think ive got a job. wow. its about damn time. i should be getting my call tomorrow or else i will go back down there and have a little talk with the manager that told me i had the job damnit. i should be working at the bear factory in the middle of the mall by next wednesday if all goes as planned. and hopefully it will.&nbsp;so anyways--then i came home and had to protect some bunnies out in my bunny farm yard lol. i had to harass a cat that was harassing the bunnies. lol. it was awesome. i hate cats and when i can im gonna hit it with my protective broom. ok reading that back i sound like a big dork but i had fun and i dont care what you think lol. if you were here you wouldve done the same thing! hmmmmmmmm..enough about tne rabbits. george strait. hes awesome. and hot for an older guy. his song just came on next on my cd. soooooooooooo...after the rabbit incident i came in..listened to music, played the ol guitar, and then watched tv and talked to my mom. now im here. listening to more music and typing. its fun. you should try it sometime. well tomorrow my moms babysittin my little cousin which will be fun because shes soooooo cute! but i will also go back to the mall tomorrow. and straighten things out if i have to like i said i would. then who knows what i will do. i think im gonna find somethin else&nbsp; on here to do now because this is starting to get boring too. so bye for now!  *Sometimes I can hear this old Earth shouting Through the trees as the wind blows That\'s when I climb up here on This mountain To look through God\'s window Now I can\'t fly but I got two feet That get me high up here Above the noise and city streets My worries disappear--Mayberry--Rascal Flatts  *I\'ve dealt with my ghosts and faced all my demons Finally content with a past I regret Iâ€™ve found you find strength in your moments of weakness For once Iâ€™m at peace with myself Iâ€™ve been burdened with blame, trapped in the past for too long Iâ€™m movinâ€™ on--I\'m Movin On--Rascal Flatts   '),(20821,3682338,'19,July,2004','2004-07-19 00:00:00','   heello again! i\'m not up to much right now. that would be why i\'m writing on here! but anyways i was bored so thats why i\'m here right now. typing. i\'m also listening to some music. its awesome. i got out my old third eye blind cd because 3eb is great. and also the lead singer stephen jenkins is pretty hot.&nbsp; well today has been boring. i moped around the house about all day. i went to the store with my mom and then came home and moped around some more and was outside for a little bit. great fun. right now i\'m thinkin about plannin out my day for tomorrow. i\'m thinkin i need to be gettin me a job so i shall go get some applications tomorrow. it\'ll be fun. lol. hmmm..i dont have much to say on here tonight so i think im just gonna stop now.&nbsp; so goodnight! &nbsp; **The sky it was gold, it was rose I was taking sips of it through my nose And I wish I could get back there Some place back there Smiling in the pictures you would take Doing crystal meth Will lift you up until you break It won\'t stop I won\'t come down, I keep stock With a tick-tock rhythm and a bump for the drop--3eb--Semi-Charmed&nbsp;Life &nbsp;  &nbsp;   '),(20822,3682338,'15,July,2004','2004-07-15 00:00:00','   hey-i havent been on here for dayz. but now that i am i guess i could catch you up on things. ive been pretty busy. well, to start, im still mad at myself for not going to camp. i reeeeeally wish i wouldve gone but thats over now and no changing it soo..i also dont have a job yet. lonestar either hasnt called me yet or i missed the call but who knows and who cares. ill find a job elsewhere. i just need to try harder.&nbsp;anyways, &nbsp;some of the reasons i havent been on&nbsp; here for a while are:&nbsp;first, my bro was in a really bad car wreck (a hit and run)&nbsp;and luckily he came out completely fine. i dunno how he did because it seriously couldve killed him but i guess god was watchin over him. next, some family came in from ohio and we visited them a few nights at my grandmas and then some family&nbsp;came in from florida the next few days and we vsited them.&nbsp;also-i just didnt feel like gettin on here--mustve had better things to do! lol hmmm--what am i leavin out. i cant think right now. im sure theres something im leavin out but oh well. so for today--well, today i didnt do a whole lot. i woke up around 2:00 by accident because i didnt get to bed till like 5 or 6 in the morning. i couldnt sleep last night. so then i sat around the house for a while and then i played my guitar some and then had dinner and then talked on the phone for a little bit until my aunt and uncle and adorable little baby cousin came over to visit. after that i went to the store with my bro to look for some storage boxes because ive been cleanin out my room! AwEsOmE huh!?! jk. but i didnt find the kind i wanted so i shall search again for those tomorrow. so when i got home i went outside and harassed some rabbits. lol. no really ive been waitin to see when some little baby rabbits are gonna come out of their nest. my yard is like a bunny farm. you should see it. im not joking. oh well tho. so after that i sat around and watched tv and stuff like that and that leads me to where i am now. typin on here. and i bet you are just enjoying every word you are reading right now huh? lol. i know i wouldnt be. i know its boring but hey-i was bored and had nothing else to do. but now i think im getting real tired of this so im gonna go.lol. later gator!   '),(20823,3682338,'04,August,2004','2004-08-04 00:00:00','   hello there. i was bored again. u know those are about the only times i write in here so maybe i should just quit saying it? i dunno. oh well. so today was another boring day. i woke up, went to work, came home, had dinner, went outside to do some stuff, and now im here. work was boring as usual, but i had some visitors so that made the day a little better. : )and then theres a guy that i see every single day i work. hes at the mall EVerY day. he talks to me every time he walks by. hes a cool guy tho. im glad i met him. ooohhhhhhh wellll! enough about that. i feel all depressed right now and i have no idea why. yeah. oh well right?  so nothing much has happened the last few days that i havent been on here. lol. ive just been working and thats about it. not much time for anything else. well, im just gonna cut this one way short because i just started feelin REallY sick. no point in this one but hey at least i was here.  Later   '),(20824,3325039,'12,May,2004','2004-05-12 00:00:00','   OK, I got some squash now, so I can start typing. Now the big problem with this is the urge to spill absolutely everything in the first post. If I did that, however, I would lose momentum and impetus and this would go nowehere fast, thats the problem.  So before dropping some bombshells, lets start with something irreverant.  How do people get muscles?  OK, I\'ve never been a big guy. I\'m usually the rake thin, nerdy guy who can\'t lift, catch, or do anything that requires skill, strength and dexterity. However, for the last 3 years or so, I\'ve put on some weight, mostly through no longer being at  urlLink university  and therefore being able to eat properly.  So after splitting up with my last girlfriend, I decided to do something about it. The way I see it, if I\'m gonna atract someone, of either sex, I gotta look the part.  So I now have a lifting bench, with a cable and extensions and all osrt, plus dumbells and a barbell and a curl bar and alll sorts, I also joined a  urlLink proper Gym , but more about that later.  The problem is, that it\'s done bugger all. I\'m marginally musclier but not as much as I wanted given the time span.  Now, here\'s the thing. I\'ve not exactly Mr Testosterone either. I\'m certainly not hairy and I ain\'t gonna go bald any time soon. So, my warped brain gives me the idea of taking something to boost my chances.  ergo  urlLink 1AD  , urlLink Trenabol , and  urlLink 5AA   Now the 1AD I used right at the start of all this, and I kinda attribute the sudden results to it. This is all well and groovy so I\'m using it again.  I\'m also using the trenabol, as it\'s got some good reviews. It\'s very odd and is kinda like rubbing spunk all over yourself, and it\'s had zero effect so far ( apart from I slight scare I was growing boobs)  Now, the 5AA,I was taking about 700mg\'s of the stuff each day. And it\'s a whole world of bad news. From what I could feel it was like my prostate had grown to the size of a melon, the back of my neck was red raw from the glands flaring up, and my skin still hasn\'t recovered from the massive grease factor.   Anyway, since you can tell from this that I\'m willing to try anything, can anyone recommend things to me, like a certain set of exercises? The instructors at my gym are terribly woolly at times and stick a bunch of cardio and stabililty ball stuff in with my program, so I\'m getting a little disappointed with it all, which is making my slim motivation slip...:-(  '),(20825,3325039,'12,May,2004','2004-05-12 00:00:00','   So, I was in the bath, which I haven\'t had for ages ( I\'m a shower person), and I go through, in my mind, all the things that I\'ve thought of that could save the world, or entertain people, or at the very least generate some kind of response, which when you come down to it, is what we all want.  So, I think wouldn\'t it be at least some level of good if I actually put these things down? I\'ve tried a blog before, using Moveable Type, but there was more an exercise in me playing about, than actually putting my thoughts to screen.  So, this is my first post, I\'m gonna publish it, then go make a coffee or squash, and then come back and write something proper.  '),(20826,3403416,'24,May,2004','2004-05-24 00:00:00','   Look,  The next time I hear some talking head saying Nader is going to cost Kerry the election, I am going to puke on my fucking shoes.   The analogies to the 2000 election with Al Gore are tiresome and incorrect.  Ralph Nader did not cost Al Gore the election, Al Gore cost Al Gore the election.  If he only won his home state or Bill Clinton\'s state, Florida would not have mattered.  Al Gore looked like an idiot in the debates-rolling his eyes, sighing out loud, shruging his sholders, ect.  Al Gore ran away from Clinton\'s record.  Considering Clinton was popular despite everything that happened, the Al Gore ran from it.  Al Gore abandoned issues that Democrats care about: jobs, education, health care, and the enviornmet.  Al Gore was the one on stage with George Bush bragging how his \"plan\" would spend more money on the military.  Al Gore and his running mate were the ones who ran around the country with their religous talk, as if GWB didn\'t talk about it enough.  Face it, the Dems are abandoning their base of support.  They will loose again in 2004 to the most incompetent president in United States history.  Why?  Let\'s look at Kerry.  It is the same game. The \"nomination\" process only chose Kerry because they felt safe.  Kerry doesn\'t represent the Democratic party.  Even if he does, he will not fight for the presidency.  He talks about Democratic values, but does not fight for them.    The Democrats/Liberals/Progressives need to start fighting back.  Hannity, Limbaugh, Savage, and the rest of the blow hards cannot continue to spew their garbage unanswered.   For the intelligent Libs out there, it is time to answer these guys.  It is time to stand up for what is right.  '),(20827,3403416,'23,May,2004','2004-05-23 00:00:00','   Vice President Dick Cheney was in Houston last week for a fundraiser.  He took the opportunity to take a few swings at Kerry\'s \"war on terror\" credentials, again trying to scare the American people to vote for Bush.  Why anybody would trust this administration is beyond me.  Starting from the top, let\'s remember that neither Bush or Cheney served in a significant way(we can talk about the National Guard, and that is serving, but in comparision to others...).  I have no problems that they didn\'t serve.  The problem I have lies how this campaign notoriously smears those who did.  Remember John McCain in the Carolinas?   They portrayed him as a wacked-out veteran.  All of this aside, lets look at their Credibility.  Iraq is a mess.  Before the war began, Bush and Cheney made speeches implying mushroom clouds, millions of dead Americans, massive WMD buildups, and on and on......They stated they could do this war on the cheap, a small force will suffice, and Iraqi oil will pay for it all. They even led us to believe our troops will be greeted with flowers and such(like the liberation of Paris in WWII).  To date, everything this adminstration has said about Iraq is wrong.  From the justification of the war, the execution of the war, and the result of the war, they have been wrong the whole time.  And today they are trying to scare us in to voting for them.  I won\'t harp on the war forever, but it leaves me with a burning question.  What will happen when US intelligence identifies a REAL WMD threat in the future?  What if there is an immediate threat that required military intervention?  What credibilty does the US have?  I would trust Kerry over Bush on this war ANY day.  '),(20828,3403416,'20,May,2004','2004-05-20 00:00:00','   Welcome.  This is my first posting ever, so please be patient as I learn the culture and methodology of blogs.  I enjoy writing about anything, especially politics.  I want this blog to be honest and absent of jargon, cliches, and blind devotion to any party.  With this being my first posting, I want to keep it simple.  The issues I want to address are important.  As U.S. citizens, we need to be aware of two things.  First, the vast majority of us ignore our civic responsibilities, the most important being voting.  Very few people participate in civics, and for our democracy to survive, we must be active participants.  Our national and local leaders are elected by a smaller and smaller group of people, while the rest of eligible voters remain silent.  This must change.  It will only change at the individual level(i.e.  you!).  Second, we are CITIZENS, not consumers.  Corporations exist because we, the people, allow them to.  Corporate charters are granted by the people; corporate charters are not a right.  Elected officials are in office because we put them there.  We have far more power(and responsibility) than we exercise.  This, for the sake of our country, must change.  '),(20829,3665752,'18,June,2004','2004-06-18 00:00:00','   Well, I\'m still here.  I can\'t believe I\'ve almost started crying probably 3 times.  Why is that?  I loathe this company, but I guess it really is the people that make the difference.  '),(20830,3665752,'17,June,2004','2004-06-17 00:00:00','   Well, it appears that my transactions with this company has reached a conclusion.  My manager just told me that due to \"company-wide\" cutbacks, my services would no longer be required.  So, once again, me, and the 20+ other temps are out of luck, out of work and out of hearty meat sauce.  Why does temping suck?  For this very reason.  We\'re expendable and they all know it, and I know it too, which is the only reason that I can tolerate something like this happening.  Too much is going to payroll?  Well cut the guys who matter least!  I don\'t want to complain so much because I\'ve had a good run here and made enough money to keep my family afloat for another month, but inside, I\'m very sad.  I can\'t help but take any kind of rejection as a personal attack.    On the bright side of this, I\'ll have a really long weekend to work on my demoreel.  And goodness knows, I need the time.  I also need a job, so I\'ll call up the temp agency on Monday, fill them in on how this corporation pulled the rug right out from under me and hopefully the temp agency will be able to accomodate.  '),(20831,3665752,'17,June,2004','2004-06-17 00:00:00','   Who in the blue-eyed world buys Gogurt?  Who cannot help but be repulsed by their ridiculous commercials.  \"Lose the spoon\"??  What does that even mean?  Are spoons so uncool these days?  I guess they don\'t appeal to the hip up and coming generation of 9 year old wannabe street racers and ghat-packing G-Funks.  So, spoons, we\'ve had a good run together, but it looks like sucking on a tube of artificially colored yogurt is the thing to do.  I\'m afraid you\'ve been replaced.  In other news, so many commercials are so dumb, which makes the parent companies they represent look dumb.  For instance, I\'ll never switch to Progressive car insurance because of their stupid commercials with the lady who has voodoo internet powers.  I\'ll never eat Shedd Spread Country Crock because I don\'t like watching 2 talking hands converse about butter.  And then there\'s Subway.  Oh my lands, what in the world made them stick Jared Fogle in the forefront of their ad campaigns?  Was it his finesse on camera?  His on-screen chemistry with the other actors?  His rugged good lucks?  Probably all of the above.  And what about the blone guy who was trying to be his replacement?  He certainly showed some promise as the \"New Jared\".  What a bunch of boneheads.  Slubway is a sinking ship and it\'s all because of their commercials.  '),(20832,3665752,'17,June,2004','2004-06-17 00:00:00','   What a fine day for science.  I\'m already out of things to talk about, and it\'s not even 9am yet.  I had a nice bottle of rootbeer yesterday.  It\'s from the Thomas Kemper Soda Co. and it was fantastic.  Smooth and refreshing, tangy and salty, sweet and succulent, rich and meaty.  I haven\'t read the news yet.  But I\'ll be expecting to hear something about the commercial space/time travel race.  I\'m really looking forward to space travel moving into the private sector.  I believe that mankind is destined to control the shifting sands of time and that we are destined to move forwards and backwards through the millenia.  I saw a pretty cool movie the other day called \"The Day After Yesterday\".  It was about a group of hyper-intelligent time-traveling peacocks with a vengeance.  In the end, the peacocks won out, but WOW!  What a ride.  '),(20833,3665752,'17,June,2004','2004-06-17 00:00:00','   I thought I\'d get a quick entry in before I was off to the rat-races.  I saw a good movive the other day, I think it was called Hoary Powers and the Peon of Algebra.  It was good, although the algebra monster wasn\'t as cool as I thought it was going to be from the previews/trailers etcetera.  Right now, the movies I want to see are as follows:  Hellboy, Chronicles of Riker, Flimflam McGee and Silver Monkey and i\'ll throw in Kill Bill for good measure.  My kids are still out of town, and I\'m missing them pretty alot.  But they\'ll all be home today.  Now I just have to clean the house between now and Seven 30 am.  '),(20834,3665752,'17,June,2004','2004-06-17 00:00:00','   If you\'ve read any of what I\'ve posted earlier today, you\'d see that I had a craptacular day at work, but lukily for me, I\'ve got some good tunes to soothe my anguish.  It\'s not that I hate my job or the people there or what I do, but I don\'t appreciate when people assume I have no other ambition or capability to do anything other than deliver faxes.  I was playing some games this afternoon and it was a total waste of 17 minutes.  I like games and all, but I don\'t like wasting time, which is precisly why I have this blog.  '),(20835,3665752,'17,June,2004','2004-06-17 00:00:00','   I am not a fan of the corporate cube-farm office subculture.  I don\'t like the idea of spending 8 hours a day staring a screen and not playing a game on it.  Excel just doesn\'t have the same appeal to me.  Don\'t get me wrong, I\'m all for technology, computers and superstring theories etc., I just wish my job was more like an action movie.  I work above a well-known computer game developer and everytime I\'m in the elevator with one of those guys, I get so mad because I should be working with them.  It\'s a tough industry to crack, but I will.  '),(20836,3665752,'17,June,2004','2004-06-17 00:00:00','   I honestly can\'t believe I\'m taking time to actually update this thing, but that\'s because I have the blog FIRE!  I just have so many important things to write about.  So, my job, yes...well...I work at large North-western bank out of Washington in the wholesale division.  But, since I\'m a temp, I\'m essentially the office helper monkey.  And what a monkey I am.  It\'s true what they say about monkeys:  You can\'t teach a monkey to distribute faxes if you ain\'t got a monkey.  What I really want to do is animate.  And that is what I do in my spare time.  That and take care of babies.  '),(20837,3665752,'17,June,2004','2004-06-17 00:00:00','   Another first blog post on the internet.  Like we need another one of these...  '),(20838,3665752,'29,July,2004','2004-07-29 00:00:00','   I told these to my niece today:  What did the flea say to the dog? Nice to bite you  What do you call a tree on a farm? a farmtree  What\'s grey and blue and high in the sky? the sky  '),(20839,3665752,'28,July,2004','2004-07-28 00:00:00','   I found a cool site on MSN that talks about how stupid commercials can be.&nbsp; It\'s not entirely funny, but it is a cool place to go when you want to kill some commericals.&nbsp; Here\'s the  urlLink link .&nbsp; It\'s actually the search page for the Ad Report Card by Slate mazagine.&nbsp; Anyway, there\' s some good articles in there.&nbsp; That\'s some nice flutin\' boy.  '),(20840,3665752,'28,July,2004','2004-07-28 00:00:00','   Have you ever seen a commercial this is so stupid because it makes it seem like whatever the product they are selling is something that will bring your family closer together?&nbsp; On this one Slubway commercials of yesteryear, a father and a son are out on a boat fishing and eating slubway together.&nbsp; Slubway wants you to believe that eating Slubway products will make you a better person.&nbsp; Same with Pillsbury and Coke products.  Anyway, so what does this all boil down to?&nbsp; That Britney Spears has lost the battle of the spotlight.&nbsp; Hahahaha, I\'ve been saying it for years.&nbsp; She\'s on the way out.&nbsp; I predict the next wave of popular music is rock for girls.&nbsp; It\'s already sort of happening now, but it\'s going to get bigger.  And finally, here\'s today\'s Harry Potter predications:  -The Dursleys&nbsp;reveal themselves as powerful magic-users and defend Harry from dementors.  -The Order of the Phoenix is ambushed and mostly killed because of a traitor among them  -Hermione begins animagus training with Harry.&nbsp; She is able to turn into a raccoon.  -Harries father is revealed to him by Voldemort as a Death Eater who betrayed his family, but was killed anyway.  As you can see, my predictions are bullet-proof.&nbsp; I\'ve already read the next 2 books anyway.&nbsp; You can get there  urlLink here .  '),(20841,3665752,'28,July,2004','2004-07-28 00:00:00','   I was looking my blog just right now and it looks like I haven\'t blogged since this morning.&nbsp; Well, I need to fix that.&nbsp; I also entered a poem-making contest but when I tried to submit the poem, the web server crashed so I lost my poem.&nbsp; I was a bit mad.&nbsp; That poem took me the better part of the morning.&nbsp; I\'m so tired right now.&nbsp; I have not very much work to do here, but there\'s ton of stuff I want to do when I get home today.&nbsp; I want to play with my kids, do yoga, work on my portfolio and do some other stuff.&nbsp; This is a boring entry.&nbsp; But I guess the boring entry mirrors my boring life.  '),(20842,3665752,'27,July,2004','2004-07-27 00:00:00','   Someone once said that they have a problem with Hellboy. My question is Hellboy is a good movie with a good message about monsters and firegirls. In other news,  urlLink whiteninjacomics  is pretty funny.&nbsp; In other news, I watched the DNC yesterday and it\'s more and more obvious.   So in movie news today, the top movies of all time are The Neverending Story and The Goonies and The Conan and The Space Camp and The Howard Duck.&nbsp; I would have taken more anti-personality tests but some people have told me to stop taking them.&nbsp; Well, it\'s oblivious that they are one with the problems.&nbsp;&nbsp; Like paranoid-ia and other things.   And for all you Rocco-freaks out there:  urlLink http://www.cnn.com/2004/LAW/07/27/leisure.restaurant.reut/index.html .&nbsp; It seems like he\'s lost the battle for...The Restaurant.  And for all my super next top model freaks out there, Holler If Ya Hear Me!&nbsp; Last night was the riveting finale episode that pit 2 bonethin skankhoes against each other in a walk-off.&nbsp; One of the girls won, but nobody really cares anyway.  '),(20843,3665752,'27,July,2004','2004-07-27 00:00:00','   I have a strong impression that I know how it\'s going to end up, but I can\'t tell you except that it has to do with little bugs.  '),(20844,3665752,'26,July,2004','2004-07-26 00:00:00','   I\'m so tired because I\'ve taken too many personality tests.  I have so many problems.  I\'m a paranoid, narcissitic guy with multiple personalities and strong anti-social tendencies.  But I DO like sharks.  And anacondas.  You know that if I could a get a pet right this very minute and have it be at my house when I get home, I\'d have to pick a king cobra.  Do you know how awesome those things are?  They are AWESOME TO THE MAX-TREME!  I\'ve seen several in zoos and I just want to have one for my own because they just look so sweetbees.  '),(20845,3665752,'26,July,2004','2004-07-26 00:00:00','   Disorder      Rating        urlLink Paranoid :    Very High       urlLink Schizoid :    Very High       urlLink Schizotypal :    High       urlLink Antisocial :    High       urlLink Borderline :    Moderate       urlLink Histrionic :    Moderate       urlLink Narcissistic :    Very High       urlLink Avoidant :    High       urlLink Dependent :    High       urlLink Obsessive-Compulsive :    Moderate       --  urlLink Personality Disorder Test - Take It!  --       And while I wish that there were percentages here, I think the words speaking for themselves.  '),(20846,3665752,'26,July,2004','2004-07-26 00:00:00','   Do you remember the movie, The Glimmer Man?  That was a real rip-snorter with Zaz, Kapowza and Bing-Bang.  Anyway, the leading man has decided to pursue a career in music:  urlLink www.nonsoloblues.com/steven_uk.html .  And I just love french music played by an ex-action star.  Do you remember when Arnie started doing Hotwheels commercials in Russia or Halle Berry was doing Maggi noodles commercials in the Philippines?  Well, it\'s sort of like that except that Stevie is singing.  I mean, he sort of knows how to talk, but to sing?  I don\'t know about that.  And why hasn\'t he released his chart topper in the states?  Anyway, here\'s some more HP predictions for you:  -Ron doesn\'t die -Harry dies -Demetors are burned in the depths of the sun -Harry teaches the Dark Arts class -Hogwarts gets High Speed Cable Internet -Harry must return to see the Spider King for an extra special request -Hermione is kidnapped by the darkside and almost turns  More predictions later.  M to the C to the A.  '),(20847,3665752,'24,July,2004','2004-07-24 00:00:00','   Cattell\'s 16 Factor Test Results        Warmth   ||||||   14%       Intellect   ||||||||||||||||||||||||||||   82%       Emotional Stability   |||||||||   22%       Aggressiveness   |||   10%       Liveliness   |||||||||||||||   50%       Dutifulness   ||||||||||||||||||   54%       Social Assertiveness   ||||||   14%       Sensitivity   ||||||||||||||||||||||||||||   82%       Paranoia   ||||||||||||||||||||||||||||   90%       Abstractness   ||||||||||||||||||||||||||||   82%       Introversion   ||||||||||||||||||||||||||||   90%       Anxiety   ||||||||||||||||||   54%       Openmindedness   ||||||||||||||||||||||||||||   86%       Independence   ||||||||||||||||||||||||||||   90%       Perfectionism   ||||||   14%       Tension   ||||||||||||   38%'),(20848,3665752,'24,July,2004','2004-07-24 00:00:00','   Eysenck\'s Test Results'),(20849,3665752,'24,July,2004','2004-07-24 00:00:00','   urlLink What Classic Movie Are You?    urlLink personality tests by similarminds.com'),(20850,3665752,'24,July,2004','2004-07-24 00:00:00','   Personality Disorder Test Results       urlLink Paranoid    ||||||||||||||||||||   90%        urlLink Schizoid    ||||||||||||||||||||   90%        urlLink Schizotypal    ||||||||||||||||||   74%        urlLink Antisocial    ||||   14%        urlLink Borderline    ||||||||||||||||||   78%        urlLink Histrionic    ||||||   26%        urlLink Narcissistic    ||||||   26%        urlLink Avoidant    ||||||||||||||||||   78%        urlLink Dependent    ||||   14%        urlLink  Obsessive-Compulsive     ||||   14%'),(20851,3665752,'24,July,2004','2004-07-24 00:00:00','   Enneagram Test Results      Type 1    Perfectionism    ||   9%       Type 2    Helpfulness   ||||   13%        Type 3    Image Focus    ||    10%       Type 4   Hypersensitivity    ||||||||||||||||||    73%        Type 5    Detachment    ||||||||||||||||||||    88%       Type 6   Anxiety    ||||||||||||||    53%        Type 7    Adventurousness    ||||||||||    37%        Type 8   Aggressiveness    ||||    14%        Type 9   Calmness   ||||||||||||||||||    77%      Your Conscious-Surface type is   5w4    Your Unconscious-Overall type is   6w5'),(20852,3665752,'24,July,2004','2004-07-24 00:00:00','   Global Personality Test Results'),(20853,3665752,'24,July,2004','2004-07-24 00:00:00','   urlLink be sorted  @  urlLink nimbo.net   Well, it\'s official.  I\'m in Ravenclaw, my favorite house.  That means I\'m super smart, and every one RC thinks I\'m the most!  I saw the Bourne Supremancy this morning.&nbsp; It was the rootenest, tootenest Wild West thrill ride of the past year!  '),(20854,3665752,'24,July,2004','2004-07-24 00:00:00','   Making Harry Potter predications seems to be all the raze these daze. So add these to your harry potter stuff:   *Harry Potter is NOT the prisoner of Alzkatran  *Ron dates Hermione  *This whole series has been a dream and the last words of the last book will be \"and Harry woke up\"  *Ginny is flursmarshed by her two brothers  *Fudge aligns with Voldemort unbenkownsted to anyone by Dumbledore  *Harry does not defeat the dark wizard  *Dumbledore sacrifices himself for the good of Hogwarts  *Draco becomes Headboy   More to come on this later. I\'ve already read both books. They are available for download.  '),(20855,3665752,'23,July,2004','2004-07-23 00:00:00','   If you haven\'t noticed, I\'ve done some custom mods to the webpage here.&nbsp; There wasn\'t much I wanted to because it\'s already a great design.&nbsp; But I did mod the links and considered doing one of those \"What I\'m Doing\" things on the sidebar over there --->.&nbsp; So other than that, it\'s been an uneventful day here at work.&nbsp; The Frumix hasn\'t updated the blog yet so I just don\'t even know what to do about that.&nbsp; In movie news, much to my complete surprise and absolute shock, Catwoman is opening to not quite good, but throughly dismal reviews.&nbsp; I expected it to be the movie to launch Cats, not only on broadway, but in book form as well.&nbsp; Also, in other print mediums and comic books.&nbsp; Cats:&nbsp; The Movie is a terrific diversion from all these psuedo comic book movies gone comic books.&nbsp;  '),(20856,3665752,'22,July,2004','2004-07-22 00:00:00','   I was just at work, not working, when I thought that it might be the perfect time to create an entry in my blog.&nbsp; Yesterday I read a cool article on the 50 worst songs of all time.&nbsp; Check it out  urlLink here .&nbsp; I would agree with most of it.&nbsp;&nbsp; I thought some songs were missing from the list, particulary one from Eddie Murphy.&nbsp; I didn\'t agree with the number one song either.&nbsp; I think there are far worse songs than that.&nbsp; For instance, anything from Eddie Murphy.   Yesterday was a slow day for purchase orders.&nbsp;&nbsp; I only did one the whole day and I got it around 3:30.&nbsp; When I got home I started playing The Four Swords:&nbsp; Life in the Zelda Groove.&nbsp; It\'s a sweetbees game with sweetbees mutliplayer.&nbsp; Today I brought a sketchbook to work and I\'m going to try and show some discipline and actually draw something when I\'m at work.  '),(20857,3665752,'22,July,2004','2004-07-22 00:00:00','   We all know that Star Trek wouldn\'t quite be the show it was without the unpredicitiablitiy of Wil Wheaton.&nbsp; So, check&nbsp;out his  urlLink website .&nbsp; It\'s full of the Wil goodness that most of us have missed in the past 15 years.&nbsp; He really is a credit to humanity.  '),(20858,3665752,'21,July,2004','2004-07-21 00:00:00','   I\'ve recently contributed a large amount of time to studying the works of NK Stouffer, the woman who REALLY invented Muggles, Magic and Mayhem.&nbsp; If I had the money, I\'d donate to her cause.&nbsp; You can tell that JK Rolling is a fraud because her books are terribly written and the whole story behind the story of HP, the one where she made the whole thing up on a napkin in some coffeehouse or something, I mean, c\'mon.&nbsp; That\'s a fake napkin story if I\'ve ever heard one.  '),(20859,3665752,'21,July,2004','2004-07-21 00:00:00','   Here\'s a little something to think about.&nbsp; If you are in a car driving 30 mph on an interstate towards Sacramento and someone else in is a car driving 75mph on an interstate towards DC, what\'s the answer?&nbsp;&nbsp; The answer is you are going too slow so pull over so I can pass!!&nbsp; So far this morning, nothing to do at work.&nbsp; I won\'t bore you with the details, but trust me, it\'s boring.&nbsp; I\'ve decided I\'m going to take some more quizzes and see what I need to do with my life.&nbsp; Oh yeah, my IQ is 124.  '),(20860,3665752,'21,July,2004','2004-07-21 00:00:00','   Fight Club!     urlLink  What movie Do you Belong in?(many different outcomes!)    brought to you by  urlLink Quizilla  '),(20861,3665752,'20,July,2004','2004-07-20 00:00:00','   Well, for my next lecture, I\'d like to talk about The Neverending Story.&nbsp; What a delicious movie that has turned out to be.&nbsp; Yes, it has some funnyness in it, like when Atreyu first gets blown off the tree by Moola the Aged One, it\'s quite obvious that the boy that gets blown off the tree is not Atreyu, but a strange imposter of sorts who seems to enjoy taking the falls.&nbsp; Or the part where you can see the spinning support under the asteroid as it spins through the vast emptiness of what\'s left of fantastica.&nbsp; Or the part where you can see the dwarf that has hung himself inside the attic where Bastian is reading the book.  One of the things that we like to talk about here is The Brady Bunch.&nbsp; Another thing we like to talk about is the  urlLink mystery animal of Maryland .&nbsp; Now, I don\'t know about you, but to me, this is obviously Zuul from the documentary Ghostbusters.  '),(20862,3665752,'20,July,2004','2004-07-20 00:00:00','   What gives with that?&nbsp; Anyway, I\'ve noticed a trend on TV that is too stupid to be disturbing.&nbsp; There seems to have a been a 200 percent increase of corporate advertising bullroar.&nbsp; Lets look at Gilette for instance.&nbsp; What in the world is a \"micro shaving system\"?&nbsp; And what in the world is a \"micro pulse\"?&nbsp; Is this not just a razor?&nbsp; Does it really, I mean, *really* make the hairs on your face stand up for easier shaving?&nbsp; I don\'t believe that it does, or it does just enough, which could be just a tiny amount, just so they can hype it up on TV.&nbsp; Here\'s another one.&nbsp; Ganier, the hair people.&nbsp; What is a \"micro fruit wax\"?&nbsp; There is so much stupid corporate bullroar on the TV these days.&nbsp; One of my most hated commercials right now is the \"Tension Headache\" from Excedrin or Advil or Badpil or whatever.&nbsp; Very very annoying.&nbsp; I would expect this kind of thing in the Philippines, but c\'mon.&nbsp; Here?&nbsp; In the philippines there was this particular kind of orange drink, like Tang, but a different name, oh yes, it was call 8\'o\'Clock.&nbsp; Anyway, they advertised that it contained this special ingreidient called&nbsp;QX-99 or something.&nbsp; Obviously bullroar.&nbsp; Anyway, I think that consumers all over the world need to realize that the amount of commercial bullroar is at all time high levels.&nbsp; It seems like one of the only products out there that doesn\'t come with a lot of bullroar is Dove.&nbsp; 3/4 moisturizer.&nbsp; They don\'t even need to tell what the moisturizer is because that would make it corporate bullroar.&nbsp; So, Dove, thank you.  '),(20863,3665752,'20,July,2004','2004-07-20 00:00:00','   I love Miracle Lucky Time.&nbsp; Anyway, I was thinking about my brother\'s Atari Lynx.&nbsp; He had the newer model, not the big old one.&nbsp; His was sleek and sharp with clean lines and smooth graphics.&nbsp; Not to mention stereo sound!&nbsp; But I remember playing California Games and being the best in the world at surfing.&nbsp; I once pulled a quadruple 360!&nbsp; It was OFF THE HOOK!&nbsp; But then I would start playing hacky sack and I wouldn\'t even get 100 points because I wasn\'t any good.&nbsp; But in the half-pipe, that was a different story for sure.&nbsp;&nbsp; I used to be so good it was like I was going to fly away to the&nbsp; moon!&nbsp;  &nbsp;  If you want to watch a really funny movie, you should watch A very Brady sequel.&nbsp; It\'s super funny.&nbsp; Or for the decade impaired, it\'s totally groovy in a far-out happening kind of way.&nbsp; If you want to play a really cool videogame, you should try Mr. Driller or Rainbow Cotton.  &nbsp;  I still haven\'t actually received any envelope stuffing materials yet.&nbsp; But they did ask for 300 hundred more dollars.&nbsp; And since I gave them cash, maybe it takes a while to process the transaction.&nbsp; I\'m sure they\'re going to help me out so I can start stuffing envelopes for 3DOLLARS PER ENVELOPE!!!&nbsp; That\'s $3000 A DAY GUARANTEED!&nbsp; But for now I work at a small telecommunications reseller.&nbsp; And while I don\'t make quite that much, someday soon, I will be stuffing envelopes full of junk mailers for you, the reader.  '),(20864,3665752,'19,July,2004','2004-07-19 00:00:00','   I was just on Pollstar hoping that I hadn\'t missed the Rush concert for their 30th anniversary, but I did.&nbsp; And with that, I missed tasting history.&nbsp; Fleetwood Mac\'s dates have also passed.&nbsp; Face to Face is going to be here on Sept. 9th though and we&nbsp;all know how much fun their shows are.&nbsp; Right?&nbsp; Right?&nbsp; Anyway, I was thinking about NES games again.&nbsp; I think my number one game that should be redone is Kid Icarus.&nbsp; That game is legendary.&nbsp; You don\'t just play it.&nbsp; You experience it.&nbsp; I also think I should be the art director for the game.&nbsp; I\'m a good director and I love NES games.&nbsp; Why shouldn\'t I be?&nbsp; I\'ve actually done some rough sketches of what I think he should look like.&nbsp; It would be an awesome game. &nbsp; In other news, Michael Jackson is going to be a father of quadruplets.&nbsp; Those lucky kids.&nbsp; I can\'t believe I missed the Rush concert.&nbsp; I\'m so mad at myself.&nbsp; I just noticed as I glace around the office that because the 2 bosses are gone already, people are quietly starting to leave.&nbsp;&nbsp; And it\'s only 2pm!&nbsp; THAT\'S OFF THE HOOK!  '),(20865,3665752,'19,July,2004','2004-07-19 00:00:00','   I\'ve even mesmerized a phew kords.&nbsp; What do you think about that?&nbsp; Guitar is pretty cool, but my fingertips are hurting to the extreme!&nbsp; I can play \"Down in the valley\"&nbsp;and some other stuff.&nbsp; I\'d like to go electric eventually, but right now I just need to work on actually learning how to play instead of going straight to rockstardom like most of the kids nowaday. &nbsp; I have learned to play Discovery from 2112 from Rush from the album called 2112.  '),(20866,3665752,'19,July,2004','2004-07-19 00:00:00','   I was just thinking about something my brother told me the other day.&nbsp; When was the last time you saw movie that was really funny that wasn\'t a satire of other movies?&nbsp; I can\'t even think of any off the top of my head.&nbsp; I mean, Jem was really funny, but it\'s not on anymore.&nbsp; Speaking of Jem, it came out on Rhino DVD and I\'m seriously considering thinking about possibly looking into checking them out from the library. &nbsp; In other news,&nbsp; this&nbsp;blog is brought you to by Sugarman\'s Potted Meat Food Product. &nbsp; So I took the bus to work today.&nbsp; As it got closer to 8oclock, I was getting worried because I had no idea where I was.&nbsp; But as fate would have it, the bus dropped me off right in front of my building, sort of. &nbsp; &nbsp;  '),(20867,3665752,'19,July,2004','2004-07-19 00:00:00','   I\'m bored at work and I\'ve only been here for 7 minutes.&nbsp; So far this morning, I don\'t have a thing to do, but I noticed that there are 2 almost full boxes of pizza sitting in the lunchroom.&nbsp; Well, I don\'t think I have to tell you that I\'ll be tappin\' that before the day is out, if you know what I mean.&nbsp; At work yesterday there was&nbsp;a big to-do about a form called a Master License Agreement.&nbsp; I need the MLA to approve the approve the transaction, but the salespeople are suppose to get the MLA before they sell anything.&nbsp; So basically, they\'ve sold products/services that we can\'t legally give them.&nbsp; So, I\'ve been trying to get them to give me their MLA\'s so I can approve the transaction and move on with my life.&nbsp; Marvelous huh? &nbsp; In other news, The Goonies for Nintendo was such an awesome game.&nbsp;&nbsp; I never beat it, but I did have a lot of fun playing it.&nbsp; You might even call it the best game in the Americas.&nbsp; Here\'s some other games I really enjoy: &nbsp; Life Force, Legend of Scherezade, Stinger, Contra, Princess Tomato in the Salad Kingdom, The Guardian Legend, Castlevania and so many others.&nbsp; I get really nostalgic thinking about Nintendo.&nbsp; Oh yeah, Megaman 2 also.&nbsp; New games just don\'t have the same thrill factor.&nbsp; How many FPS\'s do we really need?&nbsp; Synergize!  '),(20868,3665752,'19,July,2004','2004-07-19 00:00:00','   I\'ve added a new picture of myself in the profile.&nbsp; It\'s actually from my fourth grade school picture.&nbsp; I was going through a transitional phase.&nbsp; So, anyway, I\'ve been reading about Hellboy for the past, oh, probably 6 hours, and I\'ve come to the conclusion that I\'m going to buy the first TPB and then I\'ll read it and then I\'ll think about what I\'ve read and then I\'ll talk about what I\'ve read to my blog who automatically records my thoughts. &nbsp; As I\'ve thought about things like Jem and Gummi Bears and Ducktales this past 2 days, I\'ve wished that I had a TV channel that had every cartoon show on that has ever been on TV.&nbsp; I\'d call it \"Cartoon Channel\". &nbsp; My IQ is 124 according to a scientific test on  urlLink Tickle.com .&nbsp; That makes me the richest person in Astoria.&nbsp; Speaking of Goonies, I  urlLink heard  they were making&nbsp;a Goonies movie again.&nbsp; It\'s about the Goonies finding a magic ring that grants wishes and makes you invisible.&nbsp; It\'s called Goonies:&nbsp; The Hunger.  '),(20869,3665752,'16,July,2004','2004-07-16 00:00:00','   Nostalgia is so dudical-ness.&nbsp; I was nostalgitizing this afternoon because I haven\'t done a lick of work since I got here and I just really like that feeling.&nbsp;&nbsp;In other news, I\'m going to the Fleetwood Mac concert in Seattle!&nbsp; In other news, I\'m not really, but I would love to go.&nbsp; I bet it\'d be way good.&nbsp; In other news, I WILL be attending the Rush concert on July 2nd.&nbsp; In other news, it would appear that I\'ve missed it.&nbsp; Do you know what movie I really want to see?&nbsp; It\'s called The Neverending Story.&nbsp; It\'s about Atreyu and his quest to save all Narnia from the power of the White Witch.  '),(20870,3665752,'16,July,2004','2004-07-16 00:00:00','   I\'ve already lost sooo many loans to Ditech.&nbsp; I\'ve been learning the guitar this weekend because now I have one to practice on where as before all I had was my air guitar, which rocked obviously, but the sound quality just wasn\'t there.&nbsp;&nbsp; In bigger news, my sister birthed a baby boy this weekend.&nbsp; I haven\'t seen him in person yet, but I did see the pictures over the internet.&nbsp; Fantasmic, that\'s what I say.  &nbsp;  So, about Gummi Bears.&nbsp; At work I was singing the Gummi Berry Gummi Berry Juice song and several of my co-workers joined in.&nbsp; I guess you never know where you\'ll meet other fans.&nbsp; It was a great experience.&nbsp; I think one of my favorite episodes&nbsp;was the one where Cubby gets the flute from Zummy, but it actually summons a huge water dragon that they fly around on and burn some of Igthorne\'s Castle.&nbsp; That was a cool episode, plus the one where they find Gusto.&nbsp; I never really liked Gusto so much, but the \"whole in the ocean\" was a cool sequence.  &nbsp;  &nbsp;  '),(20871,3665752,'16,July,2004','2004-07-16 00:00:00','   I just finished reading the episode guide for the Gummi Bear series.&nbsp; I remember most of those episodes too.&nbsp; One of my favorites was where Cubbi and a ghost knight had to stop a wizard from turning the doomsday clock on.&nbsp; It was freaky at the time.&nbsp; I remember the last episode too, but I can\'t remember how it actually ended.&nbsp; Did the Gummi\'s of G.Glen move to the barbarian gummi\'s abandoned city?&nbsp; Did they finally leave to New Gumbrea to be with the great Gummi\'s?&nbsp; I love that show.&nbsp; It was so entertaining.&nbsp; So much better than anything else from Disney on TV.&nbsp; Ducktales, of course, comes in second and all the other stuff sucked.&nbsp; Especially Darkwing Duck.&nbsp; Stupid stupid stupid.&nbsp; Talespin had some cool ideas, but that stupid kid was totally lame.&nbsp; Rescue Rangers was just ridiculous.&nbsp; I mean, talking rodents and time-traveling?  '),(20872,3665752,'15,July,2004','2004-07-15 00:00:00','   I was surfing the web right now and I found a cool site.&nbsp; Here\'s the link:&nbsp;  urlLink http://www.angelfire.com/la/Holograms/index.html .&nbsp; For anyone in the know, just by looking at the title, you can see the word Holograms spelled in plain language.&nbsp; And where\'s there\'s Holograms, there is most definatly, Jem.&nbsp; Jem was a hallmark moment in cartoon-dom because it fused rock stars with magic computers and schizophrenia all in one TRULY OUTRAGEOUS half-hour of tv!&nbsp; My brother watched Jem at 3:30 everyday and told me he was in love with Jem.  &nbsp;  Wow, all of this talk about Jem has made me want to talk about Jem some more.&nbsp; I remember one episode where Pizazz, the leader of the rival \"punk\" band, The Misfits, somehow comes across a lasergun.&nbsp; And also in the same episode she comes across a jetpack and starts flying around the city shooting her lasergun at signs of Jem.&nbsp; It was truly outrageous.  '),(20873,3665752,'15,July,2004','2004-07-15 00:00:00','   I was sitting here thinking about The Neverending Story.&nbsp; Remember the part at the end where Atreyu asks the Auryn to guide them to the Ivory Tower?&nbsp; Well, after the Auryn&nbsp;glows a bit, some asteroids finish moving out of the way to reveal the Ivory Tower.&nbsp; My question is this:&nbsp; Atreyu should have just been more patient.&nbsp;  &nbsp; Another thing about that great epic is towards the end, where Atreyu meets The G\'mork, The Gman says \"if you come any closer, I\'ll rip you to shreds\", but later, or earlier, tells Atreyu that he has the honor of being the Gman\'s last victim.&nbsp; So, my question is that the Gman shouldn\'t give conflicting advice.&nbsp; If he\'s going to eat Atreyu he should have said, \"Come closer, you who have the honor of being my last victim\".&nbsp; &nbsp; I think one of the cooler parts of that movie involve The Southern Oracle.&nbsp; I think they look pretty cool with all their un-earthly blueish glow.&nbsp; Too bad they give Atreyu the wrong information about the boundaries of Fantasia.&nbsp; If I were Atreyu, I sure would hate to have the wrong information about the boundaries of Fantasia, let me tell you! &nbsp; My next movie that I\'m going to see is probably going to be Spiderman 2 this weekend.&nbsp; I still have to get some ticket prices though, but I\'m going to try for the matinee tomorrow morning.&nbsp; Matinee prices are still slightly reasonable.  '),(20874,3665752,'15,July,2004','2004-07-15 00:00:00','   I have been thinking about Ronco the past few days.  I remember when he was peddling his products by himself, but I saw one of this shows the other day and his entire family was on including his older brother.  It was very rehearsed.  I\'m known for a lot of things, but one thing I\'m known for is being to tell is someone is fake laughing, and Ronco faked it every time.  His brother tells a joke about thin fillets of fish, something about \"you can fillet   the fillet!\"  Ronco just started busting up.  Anyway, it was really fake.  But I still love the Showtime oven and the Food Dehydrator just the same.  I don\'t personally own any Ronco products, but I have plans to buy a Showtime oven so I can cook perfect ribs, EVERYTIME!  Plus it comes with a lifetime supply of ribs too.  Or did they say ribwiches?  '),(20875,3665752,'14,July,2004','2004-07-14 00:00:00','   I cannot believe what I am  urlLink reading .  The Jared Fogle is set to reappear in Slubway ads.  This is for no joke.  The Jared is returning to your television set boxes.  You might recall, I had an interview with Jared Fogle, which you can view  urlLink here  and it\'s about halfway down the page under the date \"Wednesday in Mirkwood\".  But it\'s like I\'ve been saying for years, The Jared is a corporate spammer who possess no real marketable characteristics other than being a poster boy for a floundering organization.  If you want really good slubs, and you are in Florida, try Firehouse.  If you are in Washington, try Blimpies.  If you are in New York, try Machievelli\'s.    I\'m not going to lie to you.  I used to work at Slubway.  I worked there for all of about 1.5 months.  I did enjoy getting free slubways and giving discounted slubways to my friends, but the hours sucked, the pay sucked, Jared sucked and the building sucked.  '),(20876,3665752,'13,July,2004','2004-07-13 00:00:00','   Good morning everyone.  I slept ok.  I have a slight headcold from overuse of the fan at night.  On the calendar for today, Bejeweled, Hexxic and forum browsing.  For lunch today, pizza.  For dinner, Animal Crossing.  I\'ve almost got my final home loan paid off.  750,000 bells!  As usual, my mind is thinking about games even though I\'m suppose to be doing something else.  Oh, this is serious:  DO NOT BUY the Spiderman 2:  The movie videogame for PC.  If you have a nextgen console, buy it for that.  SM2TM for PC is a completely different game.  I played it on Xbox first and it was totally cool.  Just yesterday I got it for PC and what the crap, its a completely different game.  A game that sucks hardcore by the way.  I was totally shocked.  Anyway, Caveat Emptor.  '),(20877,3665752,'13,July,2004','2004-07-13 00:00:00','   My friend and I were driving around in mid december when we chanced upon a man shoveling road snow.  We\'re not sure who struck first, but we do know it was us that splashed roadsnow all over the guy with our car.  On that very same morning, we found another man videotaping the outside of a yet unopened store.  We didn\'t know what he was doing so we watched him from the car until he noticed us and left.  '),(20878,3665752,'12,July,2004','2004-07-12 00:00:00','   Listening to:  Dive--Sarah Brightman Last movie seen:  Underworld .125 stars Reading:  The Neverending Story PC Game:  No One Lives Forever 2 Console Game:  nothing  I would like to have this be a permanent part of the blog on the right side there so everyone can see what I\'m doing because everyone of course, wants to know.  '),(20879,3665752,'12,July,2004','2004-07-12 00:00:00','   Saved by the bell might be the most delicious show on the whole planet.  What other show captures the drama, the funny-ness, the angst, the drivers ed cars, the singing, the rapping, the beachhousing, the talent shows, the geeks and popular kids hanging out together and the Zack Attacks that really exist in real high-schools?  There was only one other show that could do that, Saved by the Bell.  In other news, William Shakespeare has made a lot of great movies based on his books.  Romeo and Juliet, Good Will Hunting, Alien 3 and House Party 3.  I saw a movie the other day called Big Fish.  It was the best tribute to the band RUSH I think I\'ve ever seen.  What other movie captures the drama of the French Aristocracy, the drama of the clash of a snow dog and the Prince of the underworld or the galactic takeover by a fiendish organization?  No sir, Big Fish knows drama.  '),(20880,3665752,'12,July,2004','2004-07-12 00:00:00','   Q:  Why don\'t you bathe with pokemon?  A:  Because they\'ll pikachu!  I do not take credit for the previous hilarious joke statement.  I heard it whilst watching TV one saturday afternoon.  I was thinking about other shows that have revoulutionized the airways and I thought of a few more:  The Live Action Super Mario Brothers for it\'s hilarious joke statements and high quality acting.  The Snorks for it\'s superior quality animation and true to life stories about smurfball.  I mean snorkball.  The Neverending Story movie for being the greatest implementation of the Auryn than any other movie.  Any sitcom with anyone from SNL because they are always the best shows and movies because they have the cast of the really really funny SNL show.  Punked, with your host, Ashley Kitchen  And the list goes on and on till about the previous entry.  As you can see, TV hasn\'t really given us much to laugh about in the past 20 years.  And who wants it to?  I\'m content letting Mark Burnett deciding what TV shows I\'m going to watch, because he\'s a genius with TV and realitivism.  '),(20881,3665752,'09,July,2004','2004-07-09 00:00:00','   Here\'s something I\'ve wondered about for a few minutes now.  Why do people idolize the hollywood stars?  Do people believe that they have a unique gift that no one else is able to possess?  Do you REALLY believe that Ben Affleck has a skill so rare that we have to pay him millions of dollars just so he\'ll use it?  Do you really really think that you could never ever be as attractive as the hollywood stars even if you had their money, their time, their personal trainers, their workout gear, their skin products, their stylists and etc. etc...This whole industry is built on a lie.  That lie is that you, yes you, could never hope to attain the kind of status they possess because simply, they are better than you.  The sooner we get rid of the hollywood star lifestyle, we\'ll see less suicides, less health problems due to dieting, less go-gurt commercials, more money in the pockets of working citizens, less anti-american sentiment abroad and better movies.  They don\'t care about you.  They don\'t think you are fit to lick the dirt from under their toenails.  They regard you as trash.  Why do we worship them?  Why are they our childrens role-models?  Why do we pay them?  Why do we let them get away with bloody murder and then applaud them?  There is something very wrong here.  '),(20882,3665752,'09,July,2004','2004-07-09 00:00:00','   I was thinking about SBTB while I was not working this morning at work and I was trying to think of my favorite episodes.  Here\'s what I came up with:  The one with the beach house and the now famous Leah Remini as Zack Attack\'s older love interest.  The one with the heart-wrenching, tearful rendition of I\'m So Excited by Jessie Spano when Zack Attack finds out she is addicted to Crank.  The one where they crash the drivers ed car into the lockers.  That really brings back memories of my high school, let me tell you.  The one where Jessie Spano\'s younger street tough cousin moves into her house and we here the now immortalized words, \"Names Zack Morris, when it comes to schemes, I\'m top dog.\"  The one where Bayside has an extra $20,000(US) sitting around so they decide that they\'ll have a contest between the most popular boys vs. the most popular girls to see who gets the money.  I remember that kind of contest at my high school too.  And probably the single greatest syndicated TV episode of all time, Zack Attack.  Zack Morris forms a band but loves the fame more than his own friends and fellow band members.  That episode is pure genious.  Friends Forever!  '),(20883,3665752,'09,July,2004','2004-07-09 00:00:00','   I\'m back at work.  It must have been vacation this weekend, but it\'s over now.  I was at a beach house in Lincoln City Oregon since Saturday morning.  It was pretty sweetbees.  My kids liked the ocean and the sand.  They tried to eat it.  We tried to barbeque some hotdogs, but it was so windy, we really just had sanddogs.  The house was nice too.  3 floors with an ocean vue.  We went crabbing too.  We caught about 10, but due to some crazy transaction with some locals, we didn\'t really end up with any.  But then for some reason, we did get some, but they weren\'t the ones that we caught.  It sounded like a scam if you ask me.  But the crabs were good eats.  '),(20884,3665752,'09,July,2004','2004-07-09 00:00:00','   Do you ever get a hankerin\' for some hearty meat sauce?  Man, I do.  I also wish I brought some movies to watch at work because since 8 am, I\'ve been playing  urlLink zone  games and reading Fark news and posting on my blogs.  And this whole time, I\'ve had no hearty meat sauce.  I don\'t even know what to do with myself.  '),(20885,3665752,'08,July,2004','2004-07-08 00:00:00','   If you have a great recipe for grilled cheese sandwiches,  you should enter at  urlLink www.teflon.com/grilledcheese  and you could win a ton of great prizes, plus grilled cheese sandwiches are a cook\'s best friend!  '),(20886,3665752,'08,July,2004','2004-07-08 00:00:00','   So I\'ll talk about The Neverending Story.  That movie has some funny stuff in it.  There\'s one part where Atreyu is woken by Artax, but you can see the sugar cube they bribed Artax with fall out of Atreyu\'s shirt.  It\'s very visible, expecially to the trained NES eye.  I\'m glad they aren\'t making anymore movies of that movie because from what I\'ve read and heard, the 3rd movie wasn\'t that great.  People say that the second one is pretty good though.  I haven\'t seen it yet because I don\'t want to wreck my impression of the first movie.  In other news, Jared Fogle has all but disappeared from the public eye.  Rumor control has it that he\'s taken another job with Miramax to be the star of the next big action movie called Submarine FIGHT!  It\'s high budget and also stars Mickey Rooney and Jessica Alba.  The musical score is going to be provided by Ace of Bace and the group of kids from Holes.  Those kids are so talented. Remember when they were rappin\' at the end of Holes?  Now that is pure talent.  Don\'t touch that dial!  Zack Attack will be right back!  '),(20887,3665752,'08,July,2004','2004-07-08 00:00:00','   Check out this new blog I\'ve found.  It\'s called  urlLink frumix.blogspot.com  and it\'s the coolest of the kool.  Oh yeah!  In other news, I read a news story about Harry Potter and the new book name.  Here\'s my prediction:  Harry Potter and the Tome of the Ancients.  I asked Batman what the name of his new movie is, and it\'s called:  Batman.  '),(20888,3665752,'08,July,2004','2004-07-08 00:00:00','   And now for my final question:  Does anyone know how the Harry Potter series is going to end?  I have a few thoughts about it.  First of all, Hermione is appointed Head Boy beating out Ron and Malfoy.  Dumbledore is slain by the nefarious Voldemort who is turn slain by Harry.  Mr. Malfoy is slain by Mrs. Weasley but not before slaying Mr. Weasley.  Moody slays a as-of-yet-unknown villian.  Snape beats McGonangall out for Headmaster of Hogwarts.  Hagrid marries the other Giant.  Harry\'s guardians reveal themselves as powerful wizards who have been taunting Harry to help him increase his magicbilities.  The Dementors leave earth.  Ron dates Hermione.  Harry becomes jealous and dates Ginny.  Ron becomes jealous and dates Harry.  Trewlawney and Snape begin courtship.  The Malfoy house is razed to the ground by the Order of the Phoenix.  The Weasley twins open up shop in Hogsmeade.  Harry becomes one of those wizards that hunt other wizards, turning down a desk job at the ministry.  Cho Chang is killed by an as-of-yet unknown villian.  Harry and Hermione aces the O.W.Ls while Ron fails with flying colors, sparking rumors of a secret H/H romance.  More to come on this later.  I\'ve seen the future of the Potter.  If I were to write a book, which I am currently doing, what should it be about?  I have a scifi background story that I\'ve been working over for about 3 years now, but writing a book with any real substance is a monumentous task.  If you know any writers who would like to write a book for free with no royalties or payment or warranty implied or otherwise or credits or any type of compensation, let me know.  I just might have a job for them,  Fogle Out.  '),(20889,3665752,'07,July,2004','2004-07-07 00:00:00','   Well, I have a new job.  It\'s only for 2 weeks, but at least it\'s something.  So as I was saying, I\'ve frequently been asked what the connection is between Jared Fogle and Catwoman.  Well, it\'s very simple.  In other news, I predict that Catwoman will be lame.  Lamer than Daredevil and Batman 4 combined.  '),(20890,3665752,'07,July,2004','2004-07-07 00:00:00','   Bastian!  Why don\'t you do what you dream!  In other news, I still don\'t have a job, but I have an interview tomorrow for a job at a window manufacturing place.  If all works out, I\'ll handle the ordering a shipping of parts.  MOONCHILD!  '),(20891,3665752,'06,July,2004','2004-07-06 00:00:00','   I noticed that for the first time ever, I have a comment on this blog.  So, thank-you, whomever you are.  Actually, I just found out who it was so nevermind.  I still don\'t have a job.  I think I\'m going to write a book about a mythical creature who is entrusted with a magic ring of pure evil to return it to the fires from whence it came.  '),(20892,3665752,'06,July,2004','2004-07-06 00:00:00','   I was at home today minding my own business when out of the clear blue yonder, I hear a knock at my door.  I should have known not to answer it based on what I saw through the porthole, but I haven\'t had any human interaction for 5 days, and even myself gets a a little lonely sometimes.  But I digress.  At the door was a younger man, 23 or 24.  He asked me to buy a newspaper subscription.  I said, \"No thank you.\"  He said, \"Can I ask why?\"  I said, \"I don\'t read much.\"  That\'s was pretty much it.  In other news, I just now tried some of that St. Ives Apricot Dry Facial Rub.  Let me tell you something;  I\'m exfoliated--TO THE EXTREME!!!  '),(20893,3665752,'06,July,2004','2004-07-06 00:00:00','   Have you ever sat down to watch a movie that you had previously not seen, only to find out it\'s the 11th or 12th time you seen it?  That happenend to me today as I was watching Underworld, a movie about cross-breeding.  Almost everything about it, from what I could tell, has been borrowed from other action movies in the post-Matrix era.  It\'s got the black leather and latex, the John Woo guns, the slo-mo, the greenish film tint in every scene and the unnecessary, yet laughable gore.  It\'s been said before, but is worth repeating:  Hollywood is out of ideas.  Sure, there were some cool scenes, but is 2 or 3 cool scenes able to save a movie from mediocrity?  It didn\'t save this one that\'s for sure.  If you\'ve seen The Matrx, Blade, Romeo and Juliet, Hard Boiled, Johnny Pneumonic or any action movie that\'s come out in the past 4 years, you aren\'t going to see anything new in this one.  Kate is a fine actress but I don\'t think she can carry a film, especially one that is so relentlessly lame.  In other news, I saw Underwo...oh wait, nevermind.  I had a good 4rth of July this year.  No roman-candle fights though which definatly took away from the excitement, but overall, it was good. I had a large clan of drunk russians singing next door, some kids outside lighting sparklers, BBQ hotdogs, Animal Crossing and Coke.  So, yeah, not bad this year, plus the weather was superb.  Well, back to work.  '),(20894,3665752,'05,July,2004','2004-07-05 00:00:00','   So here we are again gathererd around The Green Bottle to swap stories, sing songs and reminisce of days gone by.  It\'s a funny thing, that thing called Time.  How often it can be called a friend, but in a split second can turn on you.  It\'s a wonderful thing in abundance, but when it\'s in short supply, what a miserable task-master.  I\'ve been thinking alot of myself this past day.  I was beginning to feel old, actually.  Things that I once found great enjoyment in have quietly slipped into the background of a hectic and uncertain life.  I believe that with young children, my life has been irrevocably changed.  I think that one of the reasons I don\'t enjoy my old hobbies is that there isn\'t time to do any of them.  It would seem, at least for now, that there will inevitably arrive a time of self rediscovery.  I predict that it will be around the 2030\'s, when my kids will have all left home to pursue their own dreams.  What will I think of myself then, I wonder?  '),(20895,3665752,'04,July,2004','2004-07-04 00:00:00','   It\'s about 12:18 and next door is a fun-loving family (large family) of Russians.  Every holiday they gather for accordian music and the mythical Russian Vodka.  They are...feisty.  But, it\'s not like I\'m sleeping.  Oh, I wanted to update you on the status of my envelope stuffing job.  Well, the people called today and said that my hundred dollars didn\'t go though, so I said I would send them another check.  I was sure that I saw it go through on my online banking though, but if they get another one, I\'m sure they\'ll cancel the first one.  The man then asked if I could make it a rush money order instead, so I thought, what a nice man, trying to get me the materials faster so I can start working.  By this time next week, I\'ll be financially self-free!  '),(20896,3665752,'04,July,2004','2004-07-04 00:00:00','   And now for my final question, have you ever started liking a group or a type of music that perhaps in your youth, you shunned?  I would like to know why or what is making me curious about listening to musical music.  There\'s something about it that makes me *want* to like it, but at the same time, I\'m so embarrassed that I have to tell you all about it.  Anyway, I don\'t want to give the impression that I\'m moving away from my rock/punk/ska/rush roots, but I think in my old age, I\'m searching for something else.  After all, I\'ve been listening to essentially the same music for 14-15 years.  Bands have come and gone in my collection, but the core music is still there.  What\'s your damage?  '),(20897,3665752,'03,July,2004','2004-07-03 00:00:00','   Hans Christian Anderson!  I\'ve been working on my portfolio all day pretty much, but it\'s hard to see progress all the time.  I\'ve draw\'d stuff, and modeled stuff, but its still not complete.  Oh well.  I hope I get an internship on Tuesday.  I still haven\'t found a new band to listen to.  But here\'s what I\'ve been listening to lately:  Fleetwood, Guns and Roses, Sarah Brightman, Delerium and that\'s pretty much it.  Dangit, I had a question I was going to ask the internet but now I can\'t remember what it was.  I think it had something to do with...oh yes, now I remember.  In my perpetual quest for delicious hotdogs, if anyone knows of any hotdogs that are above average in deliciousness, please let me know.  Here\'s my list of favorites:  Nathan\'s, Hebrew National, Oscar Meyer XXL and Sabrett.  Those are, as far as I know, nationally available, but they are quite good.  But I\'d like to know of some other ones worth checking out too.  '),(20898,3665752,'03,July,2004','2004-07-03 00:00:00','   As I was sitting here listening to Fleetwood Mac and eating hotdogs, I thought it might be a good idea to clean the house.  It\'s a rare occasion that the babies are away, so I thought I should take the opportunity to clean the house.  Its important to clean the house so the twins feel more comfortable trashing it again.  The main question that I would like to propose to you this evening is:  If a tree\'s shadow is 67.3 feet long, and the sun is 23 degrees W, how tall is the tree?  For the record, even if there is a solution this problem, I wouldn\'t even know how to get the answer.  I might be able to reach the right conclusion if it had something to do with hotdogs or Jared Fogle or hearty meat sauce, but without those variables, I\'m pretty lost.  '),(20899,3665752,'03,July,2004','2004-07-03 00:00:00','   In all my years of blogging, I can\'t think of a better combination:  Guns and Roses and shareware games.  I\'m sitting here listening to Paradise City and playing Raptor: Call of the Shadows.  It\'s like they compliment each other, you know?  In other news, my wife and kids are not hear right now, they left for the weekend.  So basically I have nothing to do except listen to Guns and Roses and drink Coke.  And work on my portfolio and try to get a job.    In other news, in medical news, people that watch too many cartoons eventually become cartoons themselves.  It must be really late.  '),(20900,3665752,'03,July,2004','2004-07-03 00:00:00','   I haven\'t done any kind of sketching or drawing lately because I\'ve been too busy eating hotdogs.  But I thought tonight, I\'d pick up the ol\' hotdo...er pencil and start drawing again.  I\'ve got tons of re-skill-making to perform on myself though.  One more thing, I forgot to mention, if you see something that you think is really really cool, instead of saying \"cool\" or \"awesome\" or \"knarhly\", try saying \"Sweetbees\" instead.  It\'s a new word I picked up in the SoCal last month.  Down there, EVERYONE is saying it, so you know, jump on the bandwagon and say \"Sweetbees\"  I\'m in need of a new rock band to listen to.  It would seem that I\'m getting a bit tired of what I have now to listen to.  If anyone can recommend a good rock band, and let me clarify here, I\'m not interested in any modern sh*t with their 4 chord albums.  What I AM looking for is \"old skizool\" screaming guitars with guitarists that actually solo.  For examples, examine my profile.  I\'m craving something to eat, but I can\'t put my finger on it though.  This is because the table is slanted, and the hotdog keeps rolling towards the floor.  Here\'s a funny quote for you:  \"I didn\'t invent hotdogs\".  I\'ll never get tired of that quote.  Have you ever watched any TV shows, particularly, Informercials and become so upset with the host or hostess that you wanted to throw hotdogs at the face of the host or hostess?  Well I do.  But I can\'t find the remote because one of my kids was eating it earlier and I didn\'t do anything to stop him, so I have no idea where it is now.  '),(20901,3665752,'02,July,2004','2004-07-02 00:00:00','   This just in, I got a new job stuffing envelopes!  Yeah, a nice lady named Lorraine Harness sent me an email addressed to my email address and posed the question to me:  Would I like to be making more money, right now?  Well, I replied to Lorraine and said most emphatically, YES!!!  The best part is, it\'s 3 dollars PER ENVELOPE!  That means if I stuff 1000 envelopes per day, which they claim is the average, I\'ll be making 3000 (THREE THOUSAND) DOLLARS PER DAY!!  Well, I haven\'t started yet, I had to send them $100 dollars to prove to them that I\'m completely serious and fully motivated to stuff the envelopes for them.  So now I\'m just waiting for some materials that they said would be mailed to me so I can start stuffing envelopes from home!  I\'m so excited to finally end my financial woes!  I\'ll be able to pay off all my student loans in less than 2 weeks and start saving for a down payment on a new home!  Isn\'t this fantastic?!?!  '),(20902,3665752,'02,July,2004','2004-07-02 00:00:00','   Well, it would appear that I don\'t have a job yet, so I\'ll soothe my troubled heart by blogging.  \"It\'s the world\'s most popular hobby!\"  Who said that again...?  Oh, why that would be...me, actually.  Anyway, so that\'s right, no job.  I applied for an internship and I have high hopes, but many many many of my hopes have been dashed to nothingness in the past few months, I wonder if this time it will be any different.  It\'s hard to find anything to be happy about when you\'re worrying about how you are going to pay rent this month.  And for the record, as of this moment, I don\'t know how.  '),(20903,3665752,'16,August,2004','2004-08-16 00:00:00','   You are Ji Nu!   A Chinese stellar goddess. Charming, inquistive, and a dreamer.    urlLink   Which Chinese Mythological  Being  Are You?     I feel I should say that I did put that I was a male in the test, but it did not seem to affect the outcome.  '),(20904,3665752,'13,August,2004','2004-08-13 00:00:00','   I\'m all fired out about this!  urlLink    You are  Kusakabe Marron | Kaito Jeanne  You are generous and talented, though often widthdrawn. You tend to have just a few close friends, rather than many. You appear very cheerful and enthusiastic, but struggle to keep up with what others expect of you. In the end, your perseverance and sacrifices help you pull through.  urlLink Take the \"What Magic Girl are you?\" Quiz  '),(20905,3665752,'13,August,2004','2004-08-13 00:00:00','   Why haven\'t I blogged today?  Well, in conjuction with that question and the theory of Warpation, I have come to blog today.  And now for my final question, has anyone seen that they crappy US basketball team got trounced by Puerto Rico?  Now that\'s what I call a Real World Education.  In other news, a close friend and myself are making a game like the one from nintendo called Animal Crossing.  It\'s going to have things like collecting bugs, paintings, household objects and fish.  You\'ll be able to talk to animals and buy stuff at a store.  There will be a dump and a fireworks show too.  We\'re going to call it Animal Crossing 2-The Ripoff or AC2-The Hunger.  We\'re still doing the decision.  '),(20906,3665752,'13,August,2004','2004-08-13 00:00:00','   urlLink Link .  Good artwork makes the world a happier place to live in.  Here\'s another pic for your viewing delight.   urlLink Link .  In other news, quality anime-inspired art makes the world go round.  '),(20907,3665752,'12,August,2004','2004-08-12 00:00:00','   I had 2 hotdogs for lunch today.  It didn\'t really feel like I was walking on broken glass, but I\'m willing to try new things.  I have company coming today.  It should be a fun weekend.  In other news,  urlLink fanfiction  is so dumb.  If you are going to write a fanfiction, please make sure it doesn\'t suck.  In other news, it probably will suck, so don\'t write it.  '),(20908,3665752,'12,August,2004','2004-08-12 00:00:00','   But I was playing a little game I like to call,  urlLink Curve Ball .  It\'s highly addicting, so be careful.  Plus my days are spent trying to find the most awful art on the internet.  It\'s fun, and you can see tons of bad artwork.  Most of it tries to look like anime.  Which, by the way, just because you watch anime, does not mean the following:  a)That you can draw anime b)That you are an anime character and thus not confined to the laws of physics c)That you can speak Japanese.  Knowing what \"Baka\" means does not count d)That every Japanese woman will fall madly in love with you e)That you know can emit powerful energy pulses from your hands f)That you know karate g)That you can fly and finally h)That the country of Japan is made up of anime characters  Anime is cool and everything, but only for the discerning fan.  If you or someone you know falls into the previous category, you need to have a serious talk with them about what it means live in reality.  '),(20909,3665752,'11,August,2004','2004-08-11 00:00:00','   In my quest to produce bad journalism, I have come up with this simple scheme.  I\'ll take a news story from a reputable source, but replace all the names of people with a teen idol aka Olson Twins, Lindey Holan, Abanda Vines, Hillary Dumbhead or Lizzy Macguire and then I\'ll replace whatever is happening in the article with shoplifting.  I\'ll then sell these faux articles back to the reputable source as real news.  I\'ll make a fortune in practically a fortnight.  In other news, my envelope stuffing job has basically fallen through.  They didn\'t send me any products, so I\'m down almost $1000 and no envelopes.  I guess I\'ll have to look for another envelope stuffing company.  '),(20910,3665752,'10,August,2004','2004-08-10 00:00:00','   I just heard that JK Rollin\'s new book is going to be called Harry Potter and the Goblet of Fire is Lame.  No one blogs anymore these days.  So it looks like I\'m going to have to hold down the fourt.  In other news, it would appear that I\'m still Zack Attack Morris.  I just came back from my latest world-wide tour.  It rocked.  '),(20911,3665752,'09,August,2004','2004-08-09 00:00:00','   urlLink Which \"Saved By The Bell\" Character Are You?     But I\'ve known I was Zack Attack for years now.  Why would I need to take a test to find that out?  '),(20912,3665752,'06,August,2004','2004-08-06 00:00:00','   Ahh, you are \"Velveta Shells and Cheese\". You are expensive, no  denying this, but its the price you pay for quality. You  don\'t even require other food supplies. Water and a pan  is all you need to get by. That silver bag your cheese  comes in looks like something out of the 23rd century, but  it\'s cool! And those crevices in the shells hold the cheese  so well... You Rock!  Take the  urlLink What Kind of Macaroni and Cheese Are You?  Quiz  Created by LJ User RobProv222  '),(20913,3665752,'06,August,2004','2004-08-06 00:00:00','   urlLink         urlLink Which Golden Girl Are You?     I don\'t know how I feel about this.  I\'ve never really considered myself an old knife-wielding woman.  Yet I feel strangley empowered.  '),(20914,3665752,'06,August,2004','2004-08-06 00:00:00','   urlLink  '),(20915,3665752,'06,August,2004','2004-08-06 00:00:00','   You\'re plagued by feelings of abandonment and disgust for your backstabbing step-brother.  Oh, what might have been.\"  urlLink Which OS are You?  '),(20916,3665752,'06,August,2004','2004-08-06 00:00:00','   Here is a personality test.  I was starting to wonder if I really knew who I was as a person.  So, as you can see, I can still take personality tests with the best of them.                 Global Personality Test Results'),(20917,3665752,'06,August,2004','2004-08-06 00:00:00','   Hello Internet.  A comment was recently posted on the board about not finding anything funny on  urlLink DGreetings.com .  To that comment I reply:  do you mean to tell me that someone saying they are going to pop corn out of \"U\" isn\'t funny?  Well that makes me really fired out.  Obviously, you have a humorisity problem.  In other news, most stuff sucks.  But mainly Jared Flogul sucks.  Did you know they are bringing him back to slubway commercials because of his singular wit and superior on-screen chemisty with slubway employees?  I have yet to meet a slubway employee who has actually seen The Jared.  '),(20918,3665752,'06,August,2004','2004-08-06 00:00:00','   urlLink     You are  Haneoka Meimi  You are athletic and determined, though not necessarily the best of students. You have a strong sense of justice and are always willing to help those in need. You love to joke around, and have tons of friends. Though you love a good chase, you never like to get caught.    urlLink Take\'>http://strictlyshoujo.com/goodies/quiz.html\">Take  the \"What Magic Girl are you?\" Quiz  '),(20919,3665752,'05,August,2004','2004-08-05 00:00:00','   urlLink http://strictlyshoujo.com/goodies/quiz.html\"> %20src=\"   You%20are%20 Haneoka%20Meimi   You%20are%20athletic%20and%20determined,%20though%20not%20necessarily%20the%20best%20of%20students.%20You%20have%20a%20strong%20sense%20of%20justice%20and%20are%20always%20willing%20to%20help%20those%20in%20need.%20You%20love%20to%20joke%20around,%20and%20have%20tons%20of%20friends.%20Though%20you%20love%20a%20good%20chase,%20you%20never%20like%20to%20get%20caught.  '),(20920,3665752,'05,August,2004','2004-08-05 00:00:00','   Well, so far I haven\'t found any good links of the day and frankly I\'m tired of trying.  So you\'ll just have to wait.  In the meantime, here\'s a good recipe for hotdog pizza:  ingridients: -1 pizza -hotdog pieces instuctions: -put the hotdogs on the pizza  Here\'s a recipe for eggnog too:  ingriedients: 1 eggnog instructions: drink the eggnog  I\'m so tired and hungry.   Plus I\'m all  urlLink fired out  because my youngerest sister claims to not like Guns and Roses.  That makes me so fired out, the next time I see her,  urlLink this  is going to happen.  '),(20921,3665752,'05,August,2004','2004-08-05 00:00:00','   Everyone knows that when you put 2 delicious things together, it becomes twice as better right?  Like fried fish and grape juice, or red hots and Lucky Charms.  Well, hotdog pizza might be the most delicious thing on the whole planet.  And while I\'ve never had it before, it can imagine that it looks glorious.  Delicious pizza covered with generous sized pieces of quality hotdogs.  I do not however think is should like  urlLink this .  That looks revolting.  It looks like pigskin pizza.  And pigskin, while delicious when it\'s roasted on a spit over an open fire, is not good straight out of the microwave.  If you are going to make hotdog pizza, please consult me first.  '),(20922,3665752,'05,August,2004','2004-08-05 00:00:00','   A good selection of Hubble Space pics.  urlLink Link . Jared Fogle is super cool.  urlLink Link . Some high quality production art.   urlLink Link  '),(20923,3665752,'04,August,2004','2004-08-04 00:00:00','   The internet should not tolerate this kind of language, especially when referring to the Olympics and more specifically, drugged-up NBA gangstaz.  urlLink http://story.news.yahoo.com/news?tmpl=story&amp;ncid=1756&e=2&amp;u=/040804/483/xkoe11008042110   This story really shouldn\'t have made the news, but it was funny.  In other news, the US is going to get trounced in the Olympics in basketball.  '),(20924,3665752,'04,August,2004','2004-08-04 00:00:00','   Ah yes.  1993.  It was a banner year for digital gadgetry.  It was the year that the Compact Disc medium finally overtook cassettes in worldwide sales.  The MPAA was riding high on the waves of the digital future.  It was also a banner year for people who bought 8-tracks.  8-tracks while producing a superior sound than the phonograph, couldn\'t fast forward or rewind and you couldn\'t even really stop them from playing, other than to take them out of the host device.  So what I\'m saying is, that in 1993, no one should have bought an 8-track.  '),(20925,3665752,'03,August,2004','2004-08-03 00:00:00','   I\'ve decided to list all the links that I come across everyday that are cool and hip and good:  If you\'ve ever wondered about how Jem might look today go  urlLink here  Or, if you need a Jem Zip Disk cover, go  urlLink here   Well, that\'s all I found so far.  '),(20926,3665752,'03,August,2004','2004-08-03 00:00:00','   I liked the Neverending Story theme song (Turn around, look at what you see...that one) alot more before I knew that a man was singing it and before I knew what that man looked like.  I\'m not going to ruin it for everyone else, but if you don\'t like that song, you owe it to yourself to see who is singing it because it\'s not a girl.  It\'s a man.  '),(20927,3665752,'03,August,2004','2004-08-03 00:00:00','   I have nothing to talk about right now.   So I\'ll talk about the  urlLink Brady Bunch Movies .  I was pleased to discover that they have been pressed in Digital Versatile Disk format.  This is perhaps the best news since my parents bought a stereo with an 8-track player and my brother bought a Fleetwood Mac 8-track (Rumours, no doubt) and then I listened to it all the time.  Yes, that is good news.  '),(20928,3665752,'03,August,2004','2004-08-03 00:00:00','   I have come across 2 sites.   urlLink www.dgreetings.com  and  urlLink www.engrish.com .  Both have enormous potential to get you fired for laughing to hard.  At Dgreetings. com look at the e-cards and read some of the captions.  For example:  A woman is at a desk and she exclaims:  I\'m so angry I\'m going to pop corn out of \"U\".  Another example:  You make me fired out!!  Engrish is just that, engrish.  If you don\'t know what it is, check out the site, it won\'t take long to figure it out.  It\'s funny funny stuff.  I\'m full of appetite.  '),(20929,3665752,'02,August,2004','2004-08-02 00:00:00','   I  have not seen the Village.  I have seen the commerials though and feel as I know enough about the movie to write a review.  The main character in TV is the creatures.  And by creatures,  I mean the red-cape people.  They live in the forest and come into TV when the people paint the doors with red paint.  Why do they paint the doors with red paint?  I\'m not sure yet.  The commercials have eluded that the paint is used in some form of lumberjacking.  The paint marks the doors that need to be cut down.  The other main character in the TV is the yellow people, or the bee people as it were.  They live in Bee World and somehow have \"flown\" into our world by way of an interdimensional honey pot.  Because bees are attacted to honey.  It\'s what they eat.  That\'s why they are called honey bees.  Bees are a fascinating creatures.  They can dance and sing and play Gin Rummy.  I remember playing a non-alcoholic form of Gin Rummy called Rummykube whilst I was in my youth.  So, what is my verdict of the TV?  verdict:  false  '),(20930,3665752,'02,August,2004','2004-08-02 00:00:00','   I would like to apologize to the blogging community on behalf of myself for not blogging on Friday.  I understand that this is a difficult time for everyone but me and I\'m sorry for not coming through to those who need it most.  In other news, I want china buffet for lunch, but that isn\'t going to happen because there is no china buffet around here, plus it gets more expensive every year.  15 bucks per person?  That better be some mighty good duck sauce!  '),(20931,3393040,'23,May,2004','2004-05-23 00:00:00','   My girlfriend and I watched capturing the friedmans last night.  It was a disturbing experience.  For those of you who don\'t know the plot- it\'s a documentary about a father and son who were accused of extreme sexual molestation in the mid \'80s.  The documentary presents both sides, but definately makes a good case for the innocense of the father and son.  The father committed suicide in jail, but the son\'s now 34 and is challenging his conviction.  It\'s amazing how little things change- we\'re still capable of perpetrating salem witch trial-esque violence in this day and age.  Sad.  Anyway, Jesse Friedman (the son) has a webpage  urlLink here  '),(20932,3393040,'20,May,2004','2004-05-20 00:00:00','   Wired.com ran an  urlLink article  on anti-porn website, xxxchurch.com.   The guys that run xxxchurch.com are a bit inflammatory, but I really like what they\'re doing.    In other news, my other neighor (there are four of us on the hall) left floaters (note the plural) in the toilet again. If purgatory existed, he\'d be adding years upon years to his stay.  Classes are over here at Nashotah.  I spent the past six or so days reading all the stuff I was supposed to read for historical theology and systematic theology during the semester.  So all the material\'s still decompressing.  I really enjoyed reading Luther and Hooker (which surprised me).     Gas prices are high out here, although I understand they\'re not as high as some places.  Lee\'s got a really  urlLink funny pic  on his page about them.  '),(20933,3393040,'19,May,2004','2004-05-19 00:00:00','   Added several links today.  Nashotah\'s obvious.  Disinformation\'s one of my favorite websites, and toolband.com... Tool used to be one of my favorite bands.  I\'m not so keen on them now, but I\'ll still include a link.  '),(20934,3393040,'19,May,2004','2004-05-19 00:00:00','   Hey all,  Not sure of what I\'m doing at this point.  Will add more later.  Okay, it\'s later.  Check out  urlLink this  site.  It\'s where I get the blog name.  Only at Nashotah would I be called a liberal.  Still more information.  Check out my  urlLink neighbor\'s  blog.  Okay.  Managed to add a \"links\" section.  Can\'t figure out how to change the font for the \"links\" title, though.  Would appreciate some help.  I read the section in the blogspot help, but I couldn\'t figure it out from that.  The template I\'m using doesn\'t resemble the example they give.  Not to mention, I had to add the code for the links section, so how and where do I add the code to regulate the font for the \"links\" title?  Frustrating.  '),(20935,3393040,'02,June,2004','2004-06-02 00:00:00','   Trying to get everything coordinated before you go out of state can be difficult.    Nashotah\'s campus is pretty much empty.  Most people have left for the summer.  It feels weird to be here when no one else is.  No links for today.  No new news.  '),(20936,3845636,'05,July,2004','2004-07-05 00:00:00','   hee~ *ahem* WeLcOme to Fz\'s fIrst Blog...  Hmn... *Beaute De Rose*... I shall introduce this song...    **The Rose**  Some say love, it is a river  that drowns the tender reed.  Some say love, it is a razor  that leaves your soul to bleed.  Some say love, it is a hunger.  an endless aching need.  I say love, it is a flower,  and you its only a seed.   It\'s the heart afraid of breaking  that never learns to dance.  It\'s the dream afraid of waking  that never takes the chance.  It\'s the one who won\'t be taken,  who cannot seem to give,  and the soul afraid of dyin\'  that never learns to live.   When the night has been too lonely  and the road has been too long,  and you think that love is only  for the lucky and the strong,  just remember in the winter  far beneath the bitter snows  lies the seed that with the sun\'s love  in the spring becomes the rose.   ______..___8\"=,,88,_.__________________   _______8\"\"=\"\"8\'__\"88a88\'_______________   __.._.;88m_a8___,8\"\"_\"8________________   ___\"8\"\'__\"88\"__A\"_____8;_______________   _____\"8,__\"8___8_______\"8,_____________   ______\"8___8,__8,_______\"8_____________   _______8,__\"8,_\"8,_______8,____________   _______\"8,__\"8,_\"8mm\"\"\"\"\"\"8m.__________   ________\"8,am888i\"\'___I3,mm\"___________   ________,8\"___8\"__I3.m888\"_____________   _______,88P\"\"\"\"\"I3I888888______________   _______\"\'_________\"I888________________   ____________________\"I8________________   _____________________\"I8_______________   ______________________\"I8_____________-    Meaningful lyrics... a lovely song~   Basically... I think I\'m quite an approachable person but it takes time to understand me better. Hmn... quite emotional and sentimental too. I love the people around me. I\'m a perfectionist, so I guess I have lots of expectations. At times I may feel a little lost and confused, but thanks to those who have been guiding me on the way.  A hopeless romantic here too... hmn, enjoy star gazing on a starry starry night, loves the beach, loves roses, loves beautiful music, loves purple, pink, rainbow... , loves every single bit of the natural beauty on this Earth...  Thank God and the dear ones around me for what I am today.   \" Friends like you don\'t grow on tress. I know that this is true, but if friends were flowers, there is no doubt how qickly I\'d pick you\"   \"Friends are angels who lift us to our feet when our wings have trouble remembering how to fly. There is no beginning or end. Yesterday is history. Tomorrow is mystery. Today is a gift. Dream what you want to dream, Go where you want to go, Be what you want to be. Because you have only one life and one chance to do all the things you want to do...\"    A sad thing in life is when you meet someone who means a lot to you, only to find out in the end that it was never meant to be and you just have to let go. The best kind of friend is the kind you can sit on a porch swing with, never say a word, and then walk away feeling like it was the best conversation you\'ve ever had.  It\'s true that we don\'t know what we\'ve got until we lose it, but it\'s also true that we don\'t know what we\'ve been missing until it arrives.  It takes only a minute to get a crush on someone, an hour to like someone, and a day to love someone-but it takes a lifetime to forget someone.   Hmn... just a few reflections here. It may ring a bell because these sayings are rather common. However, I do agree with it totally. Read it again... and you will be able to understand how I feel...   That\'s all then... wow~ long entry! anyway... there will be more coming up! Special thanks to Denise aka Da Jie and meow meow for helping me to set up this blog... tata =)  '),(20937,3845636,'06,July,2004','2004-07-06 00:00:00','   Am I REALLY a happy person? This question just struck me not long ago... I have been pondering over it, but my final answer \"I\'m not sure...â€ Sometimes I think not. I have lots and lots of worries, problems... and it\'s accumulating daily. At times I feel really tired... tired about thinking too much~ Although some may advise me not to think too much... but I just can\'t help it at times. I guess I may seem optimistic on the surface, but deep down inside me, I\'m quite a pessimistic person. There\'s this sense of insecurity and fear that\'s always lingering in my heart.  Sometimes, I may appear to be a carefree person... someone who\'s humorous, someone who hasn\'t tasted any downs in life. I think that at times, people are unable to realise when one is unhappy and all these overwhelming worries thatâ€™s pulling one apart. Guess I can hide my feelings quite well?! Hee~  Today, I\'ve never stopped frowning ever since I stepped into school. I think I\'m not in a very good mood today. So hey guys, if I appear to be unhappy sometimes, it\'s not your fault, cos it\'s just me. Anyway, I had p.e today... while strolling on the tracks, I took it as a very good opportunity for me to sort out my thoughts. I recalled all those sweet memories... those times I spent with my CO \"family\" on the Pasir Ris beach, what we did, what each and everyone of us said. Each and every expression on everyone\'s face was captured and still running vividly through my mind. Guess these are the moments that will be able to brighten up my darkest days. I\'m a dreamy person indeed.  I love to picture my future... where will I be, and what will I be doing. It\'s not bad to think ahead... As mentioned in one of my fav. books \"Who moved my cheese\"... Pg 58 (Imagining myself enjoying new cheese even before I find it leads me to it)  Hmn... I may seem to be talkative most of the times. However, when I start to tone down, it\'s the time when I observe the actions and the behaviors of the people around me. Regardless if you are someone I know or a stranger from nowhere... I\'m just interested or I should say curious to guess what\'s going through the minds of these people. haha... maybe I should consider taking up psychology? hmn.... or maybe law appeals more to me still. ANyway... I think every other human being feels the same... we face lots of dilemmas, thus we must learn to be more decisive in order to overcome more hurdles ahead in life.  Hai~ maybe it\'s the weather today... that had instilled in me such a gloomy feeling. I think I shall learn to let things take its natural course.  Arrghâ€¦really miss my beloved hairpin too. This precious thing will probably be buried under the sand in the playground forever. Guess~ as what â€˜meow meowâ€™ saidâ€¦ treat it as something which holds all the unhappiness and it will be buried there. Recalling what â€˜Da Jieâ€™ saidâ€¦ maybe itâ€™s not all meant to be.  Anywayâ€¦ thanks for making me feel better. Iâ€™m truly touched, especially when you guys helped me to â€˜huntâ€™ for that pin.  Yeah! So stay tuned for the next blog thenâ€¦  '),(20938,3845636,'07,July,2004','2004-07-07 00:00:00','   I\'m seeing stars *_*... I wanna faint.Inexplicable emotions running through me... my energy seems to be slowly draining away. Guess I should try my very best to look at the brighter side of life... yes! I will continue to perservere... I hope to see the fruit of the labour at the end of the day...  Anyway words enough said... not much inspirations today. =(  '),(20939,3845636,'08,July,2004','2004-07-08 00:00:00','   Arrgh... not feeling very well lately... guess I should drop by the clinic for a check up. Lots of weird symptoms. I feel cold easily... even when the morning sun is shining brightly. My teeth starts to chatter badly the moment I enter audi. I have unusual aches everywhere. Oopps... hope it\'s nothing serious... I still wanna live a longer life. haha... Anyway, I love the weather today. Although it\'s a little freezy and you can feel the coldness penetrating into your skin like a thousand needles... this scenario reminds me of the time when I was in Hong Kong. hai~ gone are the days... when will the holidays be coming again? Hopefully this time I can travel to a Western country? Maybe for sight seeing and even for a shopping spree! That would be an extremely GrEaT idea yah...  Anyway... talking about co practice... I think I\'m looking forward to it. Because of the Matsuri piece, I\'m finally \'falling in luv\' with my zhongruan. haha... I miss those percussion beats that I used to play on my kulintang (malay xylophone) when I was in Angklung Band. I miss lots of people... all my juniors, my instructor and surprisingly even the PFA(Performing Arts)Department. Felt really touched when I attended \"Arts Alive\" the other time. I just wanna say... you guys really rock to the max. and I\'m truly proud to be an ex-senior of AB. Great music! Wonderful showmanship! it\'s really a tremendous improvement... and I really hope that they will continue to uphold the excellent reputation of AB. Hmn... but I think that there\'s 1 thing that\'s lacking in AB... and that\'s unity. Some of the members lack the positive attitude... hopefully under the guidance of the next batch of leaders... we will be able to see some improvements. Honestly, I do not have much faith in the next batch of leaders... so let time prove their capabilities. Perhaps, when one is appointed a leadership position... that person will learn to mature gradually and eventually is able to handle all the responsibilities in hand.  Hmn... I\'ve lots of expectations for the next batch of AB leaders. Perhaps it\'s due to my perfectionist nature again, that may give others some pressure at times. hahahahaha!~  Okay... that\'s all...hope that I\'ll feel much better tomorrow... or else... *_*  '),(20940,3845636,'09,July,2004','2004-07-09 00:00:00','   hmn... HoRoscOpes. How many of you believe in it? well, personally I think such stuffs can be quite subjective but after reading all those books about character analysis and all about the horoscope \'Cancer\' (Crab! crab!) I think it\'s quite accurate in my case... the predictions on a daily or weekly basis are quite true too~ This is scary man... I wonder how do these ppl read each other\'s minds or to even foresee the future.  Received an e-mail regarding such horoscopes... and here it is:   CANCER WOMAN    When she is in love, she will act both ways. First, Shy and polite trembling  to be near you. Second, Attach to you like glue and trying to be with you  all the times. She will try to go home with you after work, or have every  lunch with you. It is O.K. if you like her too, but if is is not the case,  you will feel very uncomfortable.   She hates to be talked about or gossiped by someone else. If she knows she  will be very hurt. In nature she is a shy type, except she has been  influenced by some other Zodiac. She is not a brave or daring type, so if  you like her then you better be the one who start first. She will not accept  her true feeling, so if you like her you better tell her first.   She is like a musical note always change in tunes, so one minute she can be  funny and cheerful, and one minute she can be sad and depress. Other people  may think of her as \"Over-acting\", or \"Over-reacting\".   When she is depress, she will go out and look for things to make it up. She  loves money, and thinks of having \"Money\" as \"happy\", not as \"God\". She will  not look down at you if you do not have a lot of money, but she will help  you make money, save money. She is not an extravagant person and sometimes  will tell you not to buy her expensive and not useful gift.   She is the type who enjoy a long and quiet walk. Cancer woman also  influenced by the \"moon\", so under the moon light she will be fascinating  woman.   She has a constant fear for many things. She fear of not being smart enough,  not pretty enough. Even if she is not fat, she will not be satisfy. Assuring  her of her look would help, because she can change mood 4 times a day.   She is not stingy, but you will not surprise if you see she collecting old  or broken junks. She sees that everything are useful to her. She will find a  way to re-use it again some day. She is not a jealous type,but possessive.   The best part of her is that she will sacrifice everything for her love one  with no limit. Don\'t leave her in times of troubles, she will never forget  it. She is not a weak type, even she looks like one, Example if you argue  with her, she might cry her heart out. Once you left, she will wipe her  tears and start clean up her apartment normally.   She is a very careful mother and will look after her kids every steps of the  way. If she is a mother of your children, you are at ease, but if she is  your mother in law, you are in pain. Not to worry, this type of mother in  law will not let her own daughter being an \"Old mate\".   She could be moody and argue with you in many little things like many women,  but she always wait and want to take care of you. If you argue with her and  disappear a few days, she will be waiting for you, but not for long O.K.  This kind of testing is risky, try not to do it.   The Cancer woman need 2 things to be happy which are \"Work\" and \"Love\". She  can be live in a dusty house, but she can not live in that same house with  no Love.    Haha~ after reading this results... maybe I would like to take back my words at the start of this blog. erm.. guess this character analysis is only 60-70% true. Anyway that\'s all then... tata  '),(20941,3845636,'10,July,2004','2004-07-10 00:00:00','   WaHahaha... had a wondeful... superb day today! Had a tricky listening comprehension exam this morning... followed by an interview by the ex co. of chess club.=X Hope that everything went well. hee~ Anyway, had a nice lunch today with my friends. It was a dramatic one though... with lots of climax in between. Firstly... clumsy Fz here spilled some barley onto her skirt but luckily the material is waterproof! So not so paisei... haha... then I accidentally bumped into an old Auntie behind me when I turned around to \'peep\' if my friends were still behind. Wow~ this \'accident\' had a great impact on me... I bounced up immediately... upon this collision. haha~  Besides this, I just returned from a gathering at Pasir Ris Park \"Fisherman Village\"! Had a sumptous dinner...although the food were quite expensive, but guess it\'s all worth while. In fact, this mini congregation was specially organised to welcome my friend who came back recently from Australia! Yu Yan... I miss ya man! haha~ not much changes... she\'s still as jovial and easy going as ever. This friend of mine... we know each other since pri.school days! wow~ so tonight is a \'childhood\' friends get together session!  Had lots of fun and laughter... At around 8 plus after dinner, each of us bought a drink and attempted to climb onto the \'breakwaters\'... haha~ this was the funny part... as usual we made lots of noise and ended up being the \"BRiGhTeSt\" lightbulb there. hehe~ became a nuisance to the couples who were enjoying the scenic view and doing some star gazing... when the sky is quite clear tonight. haha~ after a while, we decided to get down that platform. As usual.... clumsy ppl like us were all in slippers and we had to \'slide\' down the breakwater one by one. ahaha... we looked like \'country bumpkins\' man... This commotion we caused had resulted in 1 couple leaving the area. This shows that our \'wonderful\' presence is greatly felt! -_-\" hee~ Then... we decided to settle at another corner of the park... facing the beach. We chatted all night long.... had many lame jokes and shared different experiences in school. Goodness... I laughed whole night long... guess I have lost some voice. Oops... better keep quiet and rest my voice box a while before it deterioriates! hehe  At about 10 plus, we left the area. On the way out, I actually noticed a sign board reading \"Beware of falling coconuts\"!!! ahahaha... I nearly fainted upon reading the sign! Goodness me... such \'lame\' signs do actually exist. The evil thought that someone would be hit by a falling coconut... keep replaying in my mind! gosh... it will be really funny to see such a thing happen! haha... of course that victim will definitely not be ME!  Anyway... when we were approaching the bus-stop just a few km before.. we saw the bus! Then we decided to catch the bus... so all of us ran like a mad person being chased by a dog! We finally got onto the crowded bus... and throughout the journey, the bus was driving at top speed... guess the driver wants to get home in time for a rest! haaha... in fact such a scenario reminds us of the times in pri.school when we had school excursions to the Science Centre! hee~  WoooOo... today\'s outing is definitely an enjoyable and unforgettable one! AS i always say... CKpS6A\'99 rocks man!!! Luv you guys lots~   * Everytime I miss you... a star falls. So if you see the sky and it\'s all dark...It\'s your fault coz you made me miss you so much...\" ;P  '),(20942,3845636,'11,July,2004','2004-07-11 00:00:00','   What a lousy day... arrgh... lost 1/2 of my voice the whole morning. \"Crowed\" all day long. Thinking too much again... all the feelings bottled up in me. can\'t find a way to pour all of them out. Even if I had the opportunity... I\'m not sure who\'s the right person either. Anyway... I\'ll try to keep myself occupied with all the important stuff and I wish I can just chuck all these matters aside. I have got my focus right and I\'ll bang hard on my piriorities too...  Watched the taiwanese idol show \'qian jin bai fen bai\' today... Guess I agree with what one of the characters said.\" Meeting the right person at the wrong time \"... may not work either. Anyway... quite true~ Hmn.. personally I believe alot in fate. So what\'s meant to be may be there for you to discover. However, if one chooses not to treasure it, then this chance may just slip away... even without one noticing it.  yup yup yup... soooo lots of mix feelings today... haha... all these friendships and interpersonal relationship thingy... can be quite bothering to one\'s life at times. How do you differentiate between these two issues? hmn~ I think there\'s no definite answer to this. However, I\'ll not let it affect me... I must be strong yah. i\'ll try!  hope that I\'ll have my voice fully recovered tomorrow... and to sort out my thoughts properly soon... lala... that\'s all for today... *_*  '),(20943,3845636,'13,July,2004','2004-07-13 00:00:00','   hehe... finally managed to squeeze out some time to blog. Have been waiting for the mood to come by... Anyway yes.. Lately alot of things have been occurring simultaneously. So I\'m trying to balance things up a little and to try and keep up or to even increase the momentum. Did some serious self reflection too... to see how I can improve myself.  However, that\'s besides the point. Recently I made some new friends. That\'s interesting~ hee... these ppl didn\'t look friendly in the first place, but surprisingly they made the first move in approaching me. So, as usual... I\'m definitely ready to reciprocate to this friendship! haha... =P This shows us that we should never judge the book by it\'s cover yah?  hmn... feeling slightly better too... guess I\'ve learned that we should try to let things go at times. By accumulating them bit by bit... I\'ll never be able to handle them. It\'s just too much to worry about. Sometimes, by worrying too much about every little single thing... we tend to forget the most important stuff that we have yet to handle. So... we should try not to allow our overwhelming emotions affect us too much. All these should be set aside for the time being. I believe there will be plenty of time to solve all these problems later.  Feeling more lerthagic than usual lately... I\'m not sure why either. Maybe I should try to synchronise my body clock again...  I\'ve lots of questions in my head too... but I guess the answers will come by as time passes...  '),(20944,3845636,'15,July,2004','2004-07-15 00:00:00','   What a hectic day... tons of work awaiting me. Hai~ not easy being a student now adays. Who says that we just need to do our hw... get over and done with it? We still have our CCAs and pw... arrgh, this is getting on my nerves. I think lately I have been running into some bad luck or something. Lost 3 of my precious things... I lost $15 today. gosh... I spent the whole break frantically searching for it. However, what\'s gone is gone... I can\'t do anything now... =(  Yup... on the way home today. Something unexpected happen. Haha... I was nearly squashed in the packed bus that I boarded. GoodNEss... these passengers or rather lazy bums just refuse to be considerate enough to make way for me to alight the bus. Despite my many... and I mean MANY attempts to say \"Excuse Me!!!\"... the only person which was kind enough was an ITE student who actually helped me to press the \"stop\" button upon my request. So this comes to my point here... even if one is smart/intelligent... without the right character and positive attitude... that person = redundant. He really needs to do some self reflecton. ITE as some ppl may jokingly say that it stands for \"It\'s The End\". Not true... I definitely disagree to it. If you are willing to perservere, regardless of which ever institution you are in... with the great inner beauty... you will surely go far.  On a higher note... as I was walking home I passed by my neighbours house. Then I realised some giglings and some kind of mini \'concert\' going on. I saw 4 cute little girls acutally trying to set up a small stage outside their house. Creative enough... they actually used 2 hanger post as barriers. ANd one of these sweetie pies started to sing aloud...\" I\'m Falling in Love... di da da da...\" . haha~ but my sudden apearance caused that girl to blush immediately. I pretended not to take notice of them and continued my journey. Then I heard some giglings behind me... when her friends actually asked her why she stopped singing... she went like \"wait a while lah... there\'s someone....\" Upon hearing this... I could not help it but to burst out in laughter. wahaha... these girls really remind me of my childhood days.  I was no different from them. There I wore my big fluffy \"keng keng\" dress with huge agogo round dangling ear rings... slipped into my mum\'s high heels... borrowed her lipstick to touch up a little. And there I go... my little mini fashion show. hahaha... okay okay... I know it\'s pretty lame. But still it was really fun... amusing and \'self entertaining\' ... hahhaa =P  Hmn... there\'s CO tomorrow! haha~ this CCA makes my day most of the time... unless when I\'m not in the mood. Not forgetting my CO \'FamiLy\'... they are the ones that really brighen up my day. There\'s this magnetic thing between us that can never pull us apart. Hopefully... we will always remain as united. Hmn... lately did not really have some time for a mini congregation... but I guess some things can be felt. It\'s felt in the heart and actions can\'t say it all. Yup...I guess we can understand each other...  Anyway... a super long day tomorrow. Up to 7 plus... goodness me... I think I\'m going to fall asleep half way through the rehearsals. Hopefully everything will go on smoothly...  Through this cca... I can see my passion for music growing day by day. Really! I\'m proud to be from co. Yesh~ so let\'s pray that it rains heavily tomorrow too... nice weather to escape from P.E. wahhahaa... that\'s all then. WIll blog again when I\'m free. Really wanna make a chatbox and to load some music onto this blog. Guess I\'ll try to get it done asap... ciao~  '),(20945,3845636,'19,July,2004','2004-07-19 00:00:00','   &nbsp;  Ah!!! Super super warm weather today! I think I\'m going to suffer from heat stroke or dehydration sooner or later... argh... Anyway that\'s besides the point...    &nbsp; Although school was as stressful as usual, especially when there are lots and LOTS of rehearsals this week. I will have to endure this period of time and hopefully I will pull through. Sometimes I wish I\'m invincible and would just suddenly vanish from this world. However, guess I can\'t... because I\'ll definitely miss the dearest people around me. hee~    &nbsp; Had an enjoyable GP lesson today! Never knew that GP could be so SO fun man... Our new teacher rox to the max... haha~ although we had current affairs quiz today... we were allowed to discuss in groups. As usual... we came up with lots of lame and crappy answers just to fill up all 35 blanks in time! wahaha~ had LOTS of laughter. I couldn\'t stop laughing at all... there\'s this laughing bug in me... haha~ can you imagine... laughing until you feel breathless?! That was what happened... I reached a point of saturation whereby I couldn\'t laugh anymore already. My laughter seemed to be \'deadly\' contagious too~ seems like most of the people around me were affected by it. Ahahahaha... goodness me! -_-\'\'    &nbsp; Anyway, we watched an extremely cute show today during Chinese lesson! The little boy featured was extremely naughty... but on the other hand lovable. haha~ oxymoron yah? The story was about a separated couple who&nbsp;were busy looking for another ideal partner. However... their adorable kids were always there to cause some distruption during their dates and all... hehez~ quite a nice and touching story. Wonder what\'s the outcome? The story continues tomorrow~ =P   &nbsp; Watched the taiwanese idol show \'qian jin bai fen bai\' on SCV yesterday again! haha~ falling in love&nbsp;with&nbsp;that show sia~ so romantic... *_* Finally... both the lead actor and actress are together... hopefully they will be able to cross the hurdles ahead of them... hee~ I\'m contemplating if I should get this set of vcd! can\'t wait to watch the final episode.... hopefully it\'s a lovely one. Anyway... I only buy idol shows with a happy ending yah? ahaha....   &nbsp; A continuation for today... had some fun and laughter with the ppl in A03... they are really nice ppl... very fun-loving and sociable. Haha~ guess this class have quite a good repore with one another. Anyway thanks to meow meow~, if not I wouldn\'t have made such wonderful new friends. (\'\',) Hope to get to know them better in future...    &nbsp; So so so... tomorrow... rehearsals~ dreadful rehearsals. How I wish they can be more efficient and to stop being so draggy... it\'s so mentally draining... hai~ I think I will have a super \'black\' face tomorrow again. haha~ I\'m anticipating the worse... so that I won\'t feel that awful tomorrow. However... guess with the company of my beloved CO \'family\'... eveything should go fine. yeah~ so that\'s all then... will update asap. By the way... there\'s something wrong with my chatbox... arrgh...   &nbsp;  '),(20946,3845636,'21,July,2004','2004-07-21 00:00:00','   &nbsp; I\'m slEepy... I\'m dreamy... I\'m dizzy... I\'m floating... wahahaha *_*   &nbsp;----------------------------------------------------------------------------------------------------------------   &nbsp; a weird day ... filled with mixed feelings... beautiful windy weather...   Hehe~ had a mini and a little pathetic Racial Harmony carnival today. Was slightly late for the Racial Harmony&nbsp;talk in the audi... Oh my goodness! So paisei...&nbsp; my friends and I thought that we had to be seated by 10 instead of 9.45. Happily we marched in at 10 plus, by that time almost everyone was seated and the principal&nbsp;had just begun her presentation. Wow~ thousands of eyes watched us like hungery vulgers waiting to attack small little preys... However, luckily we managed to pull through this tiny \'ordeal\'. hehe...   &nbsp;&nbsp;Then my friends and I proceeded to the foyer after the talk... I&nbsp;participated in the Indian decor activity.&nbsp;Each of us took some coloured rice grains and laid them out according to the authentic flowery template that was placed on the floor. Wow~ wonderful experience... although it\'s&nbsp;a rather tedious job.    &nbsp; Following that... hey hey~ we&nbsp;took a stroll around the running&nbsp;tracks in school. So romantic right?! haha...&nbsp;we chit chatted along the way as the wind blowed in our direction. Goodness~ as though&nbsp;we were having a vacation on Cameron Highlands. It\'s&nbsp;definitely an experience that one should&nbsp;never miss out... I believe that such precious moments will&nbsp;definitely make&nbsp;up part of the sweet memories I have with my friends. In fact, such a scenario reminds me of the&nbsp;tittle of the book \'&nbsp; A walk to remember  \' ...&nbsp;I&nbsp;would encourage you guys out there to read it...&nbsp;   It\'s&nbsp;a sweet story about family,&nbsp;FRIENDS and the&nbsp;different characters you meet&nbsp;daily. How one gain acceptance by proving&nbsp;oneself... and to stay strong regardless of whatever&nbsp;setbacks you have faced. It\'s a simple yet&nbsp;complicated love story... Extremely touching~ &nbsp;  &nbsp;Yup yup... I\'m getting a little worried now... My dad will have to go for another mini op. tomorrow. I cross my fingers and pray hard that there won\'t be any complications... Hopefully the surgeon of Changi Hospital will avoid medical negligence&nbsp;unlike the last time. I can\'t accpect another blow... it\'s too painful to watch your love one suffer for nothing... . Hopefully&nbsp;after this mini op... another round of check up and bi-opsy will&nbsp;carry through smoothly too. I pray that there won\'t be&nbsp;any undesirable outcome... although chances are low... but there\'s still a slight possibility. Anyway, what I can do now is to give him the moral support and to provide him with some advices... which may be of some use. yesh...   &nbsp; tata... [=^_^=]  '),(20947,3845636,'24,July,2004','2004-07-24 00:00:00','   Our life revolves around love... true? I believe so...   Sometimes, I wonder it\'s better to be loved more than you love the person or it is better to be the other way round? Many people often say that it\'s better to experience the former... I used to think so too. However, sometimes I think that I may experience the latter in a relationship. I guess when one is in a relationship, alot of time and effort is needed to maintain it. But when do you ever know that you are ready for it? How do you know that he is the right one for you? Are you willing to give the other party a chance? Is it a risk in the first place? Or if you really like someone, are you brave enough to make the first move? I guess these all some common questions that lots of us who are less experienced ponder on.      Personally I feel that feelings is such an ambiguous thing. It\'s subjective thus there\'s no definite defination to it. It\'s something difficult to juggle at times... However I think that feelings will live with us throughout our life, so it\'s something we cannot escape from. It can be something that we live for too.      Sounds scary yah? But I think it\'s quite true. All these emotions and feelings will help to mould one\'s character... which makes up the most of us. \"Feelings\" is so powerful that we are controlled by it unknowingly at times. So should learn to control it well.     Anyway, I  have been experiencing lots of mood swings. Alot of things puzzles me at times. And words can\'t express it all...     Whatever it is... troubles and unanswered qns will always linger around... so perhaps it\'s a normal thing. Guess no one is trouble-free right? hee~   ----------------------------------------------------------------------------------------------------------------    Had CO performance on Friday night!!! hehe~ what a roaring success it was. Oh yah~ I received a rose from my senior! so sweet~ it\'s definitely a pleasant surprise! haha... thanks for all the support [=*_*=]    I felt quite nervous on stage initially. But what deterioriated it was the freezing cold surrounding. Gosh~ I could feel it penetrating into me like a thousand needles... giving me \'frost bites\' everywhere. haha~ okay okay that\'s a little exaggerated. But still, my teeth was already chattering badly. haha...     Yup... what a wonderful day. However the sad thing is that shi jie will be leaving co then. I\'ll definitly miss her presence! But I believe that will still be lots of time for us to congregate again. We have our whole life! so no worries... all the best in whatever you do! Must take care...      Yesh! Friends are very important to me... especially those close ones.(u know who you are... =P) yeah... although I don\'t say it often, but still... You guys are very VERY important to me k!!! *shouts* haha... will treasure you all!  '),(20948,3845636,'26,July,2004','2004-07-26 00:00:00','   When the daylights gone and your on your own    And you need someone, just to be around I will comfort you, I will take your hand,     I will pull you through, I will understand, And you know that                        I\'ll be at your side,    There\'s no need to worry, Together we\'ll survive, Through the haste and hurry,                         I\'ll be at your side,    if you feel like you\'re alone, And you have nowhere to turn,                        I\'ll be at your side,     If life\'s standing still, and your souls confused, And you cannot find what road to choose,    If you make mistakes you can\'t let me down, I will still believe, I won\'t turn around,                         I\'ll be at your side,    There\'s no need to worry, Together we\'ll survive, Through the haste and hurry    AiiShiTarU!  '),(20949,3845636,'27,July,2004','2004-07-27 00:00:00','   A Whole New World...(Aladdin\'s theme)   &nbsp; I can show you the world, shining, shimmering, splendid. Tell me princess, now when did you last let your heart decide? I can open your eyes, take you wonder by wonder, Over sideways and under, on a magic carpet ride  A whole new world, a new fantastic point of view No one to tell us no, or where to go, or say we\'re only dreaming. A whole new world,A dazzling place I never knew But when I\'m way up here, it\'s crystal clear That now I\'m in a whole new world with you.  Unbelievable sights, indescribable feeling Soaring, tumbling, freewheeling through an endless diamond sky  A whole new world, / Don\'t you dare close your eyes, A hundred thousand things to see. / Hold your breath it gets better. I\'m like a shooting star, I\'ve come so far, I can\'t go back to where I used to be /A whole new world, With new horizons to pursueI\'ll chase them anywhere, there\'s time to spare Let me share this whole new world with you  A whole new world, a new fantastic point of view No one to tell us no, or where to go, or say we\'re only dreaming. A whole new world,Every turn a surprise, With new horizons to pursue / every moment red-letter. I\'ll chase them anywhere, there\'s time to spareanywhere, there\'s time to spare Let me share this whole new world with you  A whole new world / A whole new world That\'s where we\'ll be / That\'s where we\'ll beA thrilling chase A wondrous placeFor you and me.  hehe~ really love this song to bits... lovely lyrics... melodious tune... it puts you into the dream world of your own. A song that has ironically been replaying in my mind lately. haha... maybe it\'s time for me to think back about the precious moment I hold... and to see how I can take a step forward&nbsp;into a whole new chapter of my life.  '),(20950,3845636,'28,July,2004','2004-07-28 00:00:00','   Hmn... read this if you wanna understand me better. I think this analysis is accurate to a certain extent... haha~ why on Earth will there be such a close resemblance? Anyway... yup... A tyipical cancerian here. Yeah~ canCer Rox!     &nbsp;   &nbsp;   &nbsp;  Ruby (Cancer)     &nbsp;   You like:  Spending&nbsp;&nbsp;time&nbsp;&nbsp;at&nbsp;&nbsp;home, family get togethers, making pizza with mates and listening&nbsp;&nbsp; to&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; soothing&nbsp;&nbsp;music. You dislike:  Violent movies, people who shout at u and being let down by those u love.  You\'re best at:  Making&nbsp;&nbsp;people&nbsp;&nbsp;feel good about themselves, listening to their problems and  being&nbsp;&nbsp;a trustworthy,&nbsp;loyal&nbsp; mate.  Deep down:  U&nbsp;&nbsp;want&nbsp;&nbsp;peace&nbsp;&nbsp;more&nbsp;&nbsp;than!&nbsp;&nbsp;anything.&nbsp;&nbsp;Home&nbsp;&nbsp;life is important to u and if  there\'s domestic squabbling, u become deeply depressed.  Your career:  Become a musician, journalist, interior decorator or chef.  Fashion:  Cancerians&nbsp;&nbsp;tend&nbsp;&nbsp;to&nbsp;&nbsp;have&nbsp;&nbsp;simple tastes: loose-fitting clothes in natural  fabrics and dark colours. You prefer unfussy outfits, but that doesn\'t mean  you&nbsp;&nbsp; don\'t&nbsp;&nbsp; like&nbsp;&nbsp; to&nbsp;&nbsp; glam&nbsp;&nbsp;it&nbsp;&nbsp;up&nbsp;&nbsp;when&nbsp;&nbsp;you\'re&nbsp;&nbsp;in&nbsp;&nbsp;the&nbsp;&nbsp;right&nbsp;&nbsp;mood.  As a mate:  U are a gentle person who may be quite possessive of ur friends. If someone  lets&nbsp;&nbsp;u&nbsp;&nbsp;down,&nbsp;&nbsp;u&nbsp;&nbsp;may&nbsp;&nbsp;brood in ur room for days. But when u feel safe \'n\'  secure, ur generous personality takes over and mates get spoilt rotten.  As a girlfriend:  U&nbsp;&nbsp;are&nbsp;&nbsp;sensitive&nbsp;&nbsp;and&nbsp;&nbsp;moody, so u like lots of cuddles and kisses to keep  balanced&nbsp;&nbsp;&&nbsp;&nbsp;chirpy.&nbsp;&nbsp;Quite&nbsp;&nbsp;secretive,&nbsp;&nbsp;u are shy to reveal ur feelings at  first but warm up once ur boyf gets to know u better.  If your boyf\'s a Cancer:  Tread&nbsp;&nbsp;carefully,&nbsp;&nbsp;girl. He\'s got an ultra-so! ft heart that breaks easily.  Not&nbsp;&nbsp;one&nbsp;&nbsp;for wild flings, he enjoys spe nding time alone with u & may even  be a bit clingy when u wanna go off on ur own for a bit.&nbsp;  '),(20951,3845636,'29,July,2004','2004-07-29 00:00:00','   Drowning sadness...sinking heart. Experiencing some domestic conflicts lately. Feeling quite upset but all I can do is to tolerate and to swallow down this unhappiness. I wonder why can\'t they understand me better?    Anyway... I\'m having lots of complicated feelings recently. I\'m a little confused... Sometimes... I think that there\'s no point waiting for the answer to my qns... it\'s really testing my patience and I\'m tired of it. Maybe I should try to accept the love and care another is willing to shower on me... or maybe I should persevere and wait for time to tell... these things are happening too quickly... too fast for me to comprehend.  My heart is a little lost too...  However, I think that  this could be part and parcel of our life... and inexplicable feelings just pop by... I believe that I\'ll be able to walk out of it soon.... yes.  '),(20952,3845636,'01,August,2004','2004-08-01 00:00:00','   ..:.. endless road ..:.. (By JJ)       the truth is tearing up my heart i can\'t recognize this place the endless road without a stop sign can\'t even find a stranger this time...   why am i still holding back my tears in this loneliness there\'s nothing to fear every chord still seems a wonder how we could be together every time i ask if this would be the last...   why am i still talking to myself hoping that you will have the keys to my cell every song might calm the weather but it draws me deeper how do i get out of this i think i never will...   a crystal forming in the eye maybe this would be the last the winding path down my face till i begin to taste the bitterness inside...   why am i still holding back my tears in this loneliness there\'s nothing to fear every chord still seems a wonder how we could be together every time i ask if this would be the last...   why am i still talking to myself hoping you will have the keys to my cell every song might calm the weather but it draws me deeper how do i get out of this   i think i never will...     Hmn... just bought this cd by JJ. Wonderful music...by chance I found this song. The last track of the cd. It was \'love at first sight\' kind of thing when I heard this song. haha... it can really describe my present feelings... my thoughts... my troubles and everything. The lyrics say it all...  Have been feeling really down lately... guess I\'m quite a weak person.. although I keep reminding myself to stay strong regardless of any setbacks I face or any problems that comes into my way. However still... I can\'t help it. A person who drops tears easily... that\'s all I can say. But I\'ll recover... I\'ll stand up again. I will show them that I can make it. I will exceed their expectations of me. I\'ll tolerate everything... I\'ll definitely pull through.  Another pondering qns... is it wrong to accept the love and care from another person knowing that you may not be able to return the same amount. I know that such things cannot be measured... but in this case isn\'t love unfair? Or we should always shower our love unconditionally? there\'s a weird feeling running through me lately... Although I\'m a perfectionist... but I realised that the world is not so perfect after all. I can try to make it one... but it takes 2 hands to clap and not everything will go according to your wishes. No one is perfect... yes... true. However... have I found the one that seems so perfect? Well, time MAY tell... I believe that alot of things are felt and words can never express it any better...  '),(20953,3845636,'03,August,2004','2004-08-03 00:00:00','   I\'m sick... hai~ guess my flu is getting from bad to worse... I can only depend on my antibiotic... my flu medicine may cause drowsiness, therefore I can only take it before going to bed. ahaha~ or else... I may just doze off anytime during lessons. arrgh... it\'s irritating to be sick. You sneeze all day long as though your nose is gonna fall off anytime. teRrible man~ spoil my mood sia. Anyway.... everybody out there must take good care wor... now adays it\'s a crazy trend to fall sick easily. Life is very unpredictable... so do treasure your own self and the people around you. By treasuring yourself will already be doing a good favour to your love ones! so ... that\'s all. My head\'s spinning... haha.. wait till I get better... there will be more bloggy updates to come.  '),(20954,3845636,'05,August,2004','2004-08-05 00:00:00','   DisaPpointed... forget it... I\'m sick of everything.  '),(20955,3845636,'08,August,2004','2004-08-08 00:00:00','   Had an enjoyable time yesterday at Marina Bay. The steamboat was quite nice... enjoyed the atmosphere and breeze over there. Beautiful scenery... as though all of us were on holiday. haha... we took several wonderful candid shots. haha... it turned out pretty okay... yeah. =P -----------------------------------------------------------------------------------------------------------------  hEhe~ just changed my bloggy\'s background. Yeah~ I\'m quite satisfied with the template.... for a change yah? At least this one looks brighter and more organised. hmn... enjoy reading the blog entries then... OoOpss... by the way how do I add back the links to the different blog pages? I think I have lost them all... =X  ----------------------------------------------------------------------------------------------------------------- A poem here to share...    Emily BrontÃ«    Love and Friendship   Love is like the wild rose-briar, Friendship like the holly-treeâ€” The holly is dark when the rose-briar blooms But which will bloom most constantly?   The wild-rose briar is sweet in the spring, Its summer blossoms scent the air; Yet wait till winter comes again And who will call the wild-briar fair?   Then scorn the silly rose-wreath now And deck thee with the holly\'s sheen, That, when December blights thy brow, He may still leave thy garland green.  urlLink    urlLink  '),(20956,3845636,'10,August,2004','2004-08-10 00:00:00','   ...FaTe...  Fate... this is what I believe in. It works all the time. We can make all the effort to keep our relationships going, but without this intangible \'fate\' , everything may not turn out well as expected. Of course there are many others who do not believe in fate. They believe that it\'s they who control their lives and not fate. true... I agree to what they say. However, there are times whereby certain things are not within man\'s control, isn\'t it? Thus I think we should always treasure every moment we spent with one another. It\'s something amazing that had brought all of us... the different ppl with different personalities and upbringings together. It\'s not something done deliberately... it\'s just so ironic and at times we can\'t explain why this had happen either.  I was flipping through my autograph book... the one I had before leaving pri.sch, when I was preparing to head for a whole new chapter of sec.life back then. Here... was what one of my teachers(CKPS) wrote \" Life is not a bed of roses but it is sprinkled with many learning experiences. Every step we take make us ever wiser! \" . Till now 5 years down the road... I can still vivdly remember the msg he left me. I believe that it had left an impact in my life... and it\'s an advice that will continue to guide me through thick and thin. I think sometimes we should never take things too lightly... I can\'t guarantee that all good things will last forever. However... we should always enjoy what we have... a step at a time. It\'s not a matter of being optimistic or not. It\'s a matter that one day... you would leave this place with little regrets. Knowing that you have once hold these memories and had shared it with the dear ones around you. By having faith in yourself and the friends you have... only then things will work out. Without basic trust in the first place... I don\'t see a point to start a relationship. Efforts will just go down the drain... knowing that you may have lost time meeting another wonderful person out there.  We lost a PE teacher (SACSS) lately... and I believe that he will definitely continue to live in the hearts of each and every student he had once taught. He\'s someone who believes alot in tough love and we were probably still less mature back then in sec.sch, to realise his real intention of disciplining us. He always believes that a school needs a \"Father\" to keep everything going and he has bravely taken up this role of a \"bad guy...\" Often... I wonder why must we wait till the person leave us one day... then we realise how important they are to us. I guess that\'s human\'s nature, to take things for granted. Although all of us avoid doing so... but sometimes it\'s hard, because we are influenced by the things around us...  I think reflections will help to guide us along the right track...and I believe that although time flies quickly... if we truly treasure it... then we won\'t miss out the ppl around us.  '),(20957,3845636,'11,August,2004','2004-08-11 00:00:00','   60 plus is what he is at the point of this time in his life. Today, then I realised how much he had aged over the years. Going through thick and thin with us and being the head of the family. I\'m glad to say I have such a wonderful him around. Although domestic conflicts do occur all the time... and I would say that it is like a daily affair... but through these misunderstandings and disagreements, we have got to know each other better. Do not think that within your own kins... you do not need to make the special effort to understand them better... actually the truth is that they are probably the ones whom we have always been taking for granted and we may tend to neglect them subconsciously. Sometimes he makes me cry... but I know the reason why. That\'s not a need for him to say much when he is annoyed or unhappy with me. A few words... a little change in his facial expression will send the message clearly right to the bottom of my heart. I realise how important my dad is to me...   He was there to guide me along and I can proudly say that I\'m glad to be from the Leong Family. I really do not know how long all this time and precious moments will last (not as though I\'m trying to be pessismistic)... the fact is, life is definitely unpredictable. We do not know what will happen next. Only someone of a higher power up there have things planned for us... what we can do, is to try to comprehend them and to deal it with the best of our ability. My dad just told me something today... which I felt truly touched... he said\" Let me fetch you when it\'s late... I do not know how much longer I can do so.\"   Of course upon hearing this... I said all those stuff like \'touch wood!\' and all. He tried to rectify his statement by saying that he may be going overseas soon... but anyway the point here is, whether I misinterpretated his remark or not... what he said had really struck something in my mind. Guess it\'s time I really learn to spend more quality time with them... and to manage time better so that I will have enough to spent with my studies and my love ones (family and friends...and all).   Although we don\'t say it often but as long as you sincerely use your heart to listen... you can definitely feel the love those dear ones around you are showering you with. There are times we feel very insecure... very uncomfortable... very lonely... but as these fears accumulate in us... it will gradually subside because we know that there\'s someone always there for u to give u the moral support! yup... had the sudden inspiration to write all these after hearing what my dad told me. yeah! He is my role model... and I try to remember the precious advices he gives me each time... hopefully things will continue to sail smoothly... =P Hee~  '),(20958,3845636,'12,August,2004','2004-08-12 00:00:00','   A quote:      The measure of love is when you love without measure.   In life there are very rare chances that you will meet the person you love that will love you in return...   So once you have it, don\'t let it go because it may never come back to you again...     --------------------------------------------------------------------------------     hee~ regarding this quote, I think it really makes sense to a large extent. Yes~ love cannot be measured right? How can you ever weigh it... in fact only your heart knows how important that particular \'someone\' is to you. Sometimes, it\'s a little mysterious... the more you know that \'someone\' has taken up a large space in your heart, the more we are so afraid of facing the truth. Therefore we end up trying to run away from it...      Hence, we may lose the one we greatly love and by that time it\'s probably too late to let that person know how much he means to you... the fact is, we can never turn back time. It\'s an irreversible hypothesis... thus we should hold on to what we have... and to shower it with love and care whole-heartedly...  '),(20959,3845636,'15,August,2004','2004-08-15 00:00:00','   My Heart Will Go On     Every night in my dreams I see you, I feel you, That is how I know you go on   Far across the distance And spaces between us You have come to show you go on   Near, far, wherever you are I believe that the heart does go on Once more you open the door And you\'re here in my heart And my heart will go on and on   Love can touch us one time And last for a lifetime And never let go till we\'re gone   Love was when I loved you One true time I hold to In my life we\'ll always go on   Near, far, wherever you are I believe that the heart does go on Once more you open the door And you\'re here in my heart And my heart will go on and on   You\'re here, there\'s nothing I fear, And I know that my heart will go on We\'ll stay forever this way You are safe in my heart And my heart will go on and on  '),(20960,3845636,'15,August,2004','2004-08-15 00:00:00','   Hee~ had CIP day yesterday. To say the truth, I didn\'t expect myself to enjoy it so much. We had a carnival made up of indoor and outdoor activities with different game stations and attractive rpizes to win/ This was all specially planned for our invited guests from the various children and old-aged homes. It was definitely an enjoyable one! We had our very own tpjc Radio Station with 2 lively DJs keeping the atmosphere to the MaX! We could make our dedications which consist of a mixture of Chinese, Indian and English pop songs ... which were played continuously throughout the whole event.   On my part... I definitely had os much fun netertaining them... Chasing after all those little kids, teaching them how to play the games and having to endure all the bursting of balloons that was surrounding.. (it was like a time bomb man~ ahaha) Anyway, it was really my pleasure to be given the opportunity to mix around with these kids. Although it wasn\'t enough to get to know them better, but I believe that my small effort had definitely managed to make someone\'s day better!   Maybe they are a little different from us, but their courage and determination to attempt the different games had really touched me a great deal... (no words can express it all k...)   I really look forward to such actvities in the near future, I believe we will be able to make it better each time. 3 ChEeRs for all Of us! =)  '),(20961,3558872,'09,June,2004','2004-06-09 00:00:00','    My neat son just called me from DC to hear the horses hooves as the processional carrying President Reagan\'s body went past.  It was strange because I was watching it on TV and they were going up a tree-lined path at the Capitol Bldg.  When I called Steve back, he told me he was under the trees, away from the crowds.  He\'s off to get in line for the viewing.  What tales he\'ll be able to tell is grandchildren!   '),(20962,3558872,'07,June,2004','2004-06-07 00:00:00','    I am sad over the loss of President Reagan.  I believe he was just one of us who went on to change the world.  I wish only the best for his family.  I pray they may get through this sad time and know that the nation truly loved and appreciated him.   '),(20963,3558872,'06,June,2004','2004-06-06 00:00:00','    I had to create this to be able to respond to my son\'s blogs.  I hope I remember my address so I can come back and add more!   '),(20964,3558872,'20,July,2004','2004-07-20 00:00:00','    Never paint your living room because it\'s never as simple as that.&nbsp; This past spring I decided to paint the grungy, fireplace smudged walls in my living room.&nbsp; So once summer came, and I had time to actually paint, I proceeded to my nearby paint store and brought home the pre-requisite 55 paint sample sticks.&nbsp; Needless to say, I could not decide which color to use, so my mother suggested I call a designer who could come up with a plan for my living space that I could work toward in order to have the Martha Stewart room.&nbsp; I met with Donielle and she said a plan would be ready within a week.&nbsp;&nbsp; Well,&nbsp;I needed to paint the walls, but Donielle was still planning my room, so I called her twice and finally pinned her down on a color , which turned out to be a two color, ragged-on paint application in my lovely large 25 X 14 room.&nbsp; I worried and fretted over this two step process, until my sister finally showed me how this was not brain surgery and I could always re-do anything I messed up, so one week later my walls look great, but unfortunately, this aggravated my carpal tunnel symptoms, so now I\'m looking at surgery sometime in August.&nbsp; Anyhow, back to the living room,&nbsp;&nbsp;I have now waited four weeks, and have no plan from Donielle, BUT my mother suggested I change the carpet to a lovely green.&nbsp; I did, and she was right!&nbsp; And not because she is always right, but this time she is actually right.&nbsp; Because the carpet is green and the walls are a darker shade than before, I need to change the curtains, and buy a new beige leather chair, as the blue stuffed chair I have is too dark, add a ficus tree in the corner, and order an upholstered bench for under the window.&nbsp; Oh, and because the living room is&nbsp; connected to the kitchen, I\'ll need to change the wallpaper in there, because it has no green.&nbsp; When I moved the furniture into the kitchen to have the carpet installed, my linoleum stretched, so now I need new.&nbsp; I\'m thinking a wood laminate might be nice.&nbsp; Also, my foyer now seems too light, so I\'ll probably have to hire a painter to paint it since I cannot reach up into the stairwell to paint.&nbsp; So, like I said, never paint your living room.   '),(20965,3380244,'28,May,2004','2004-05-28 00:00:00','    you know what i\'m the queen of?  making poor decisions!  yeah, that\'s right.  i feel as though, if i\'m making a decision, whatever the worse possibility there is as an answer, i\'ll choose it.  i suppose i\'ve learned something from all of my mistakes, and i wouldn\'t necessarily take any of them away... as i don\'t like to have regrets.  i make poor career decisions, i don\'t manage my time very well, i don\'t make good decisions regarding men, i push those away that i should keep close, and i keep close those that aren\'t good for me.  of course i\'m exaggerating, and of cousre i\'m all fired up right now so this whole thing is probably a bunch of BS, but that doesn\'t take away from how i\'m feeling right now.  right now i feel as though i must always fall on my face in order to learn.... maybe 3 times.   '),(20966,3380244,'25,May,2004','2004-05-25 00:00:00','    this post is probably going to come out sounding very bitchy, and i suppose i\'m intending it to be, because that\'s how i\'m feeling.  i am SO frustrated as of late with my life and it\'s craziness, that i feel as though i\'m about to crack.  i seriously don\'t know if i can take it anymore.  i\'ve been incredibly stressed out with my personal life, school, jobs, and future plans.  even though i\'m happy to have my friends and family care about me, i\'ve been getting so annoyed when people ask \"hey, how\'s the love life going?\"  or \"hey, how\'s work?\"  or even better... \"hey, do you know about grad school yet?\"  i know, i know... i should be thankful people care about me, but just let me vent... i have NO idea what\'s going on with any of it!  i don\'t know what\'s going on in my life because everything is up in the air right now, so i have no answers.  when i talk with friends, i wish we could talk about politics, the weather, what\'s going on in the world, or just chill out and have a beer together.  i don\'t want to talk about me.  me makes me crazy right now.  on another note... i\'ve really been missing sweden as of late.  i looked at my pics the other night and i miss that life so badly.  being so young and carefree, nothing to stress or worry about, no bills... just playing with kids all day, drinking beer and going out with my friends at night.  what more could you ask for?   '),(20967,3380244,'20,May,2004','2004-05-20 00:00:00','    yesterday was an odd day... i don\'t even know where to begin, trying to explain it.  though i did have one of the best days teaching yesterday, that didn\'t really make up for all the other weird things that happened.  though, i was happy to have a good teaching day.  i\'m a substitute teacher, and i always feel so inadequate.  i never truly feel like i\'ve grasped the whole teaching bit, but yesterday i did well.  yay.    so many odd conversations yesterday, and they all revolve around men that i\'ve dated, loved, or having feelings for.  interesting.  i emailed all day with *george, a guy that i crushed on for about 5 years.  he\'s now married and very happy with life.  i like that *george and i can go months without ever talking, or even thinking of eachother, then when we do talk it\'s like no time has passed and we can connect on a really close level again.  i like that.  i also thought things with *darryl might be going somewhere, but they\'re not.  i\'m not 100% sure how i feel about that yet, i suppose it\'s hard to tell because i haven\'t really \"lost\" him.  i\'m no where near ready for a relationship, as i\'m still getting over my love with *seth.  though, i had feelings for *darryl and clearly the same feelings weren\'t reciprocated.  i suppose that\'s good anyway.  i don\'t even know.  i have an amazing friendship with *darryl, something i don\'t want to lose, but when emotions are involved and then shot down, it\'s difficult.    i\'ve decided not to tell anyone i know about this blog.  therefore i doubt ANYone reads this.  i suppose its my own little ranting and raving place.  i don\'t want people i know reading this right now, the things i\'ve written are too personal.  and honestly, i don\'t want to be judged.  and as much as i truly love  my friends, i know they all have their opinions of my life and i\'d rather just not hear them right now.  i know i\'ll wind up getting hurt.. again... but such is life.  god i sound depressing.. i\'m gonna go grab breakfast.   '),(20968,3380244,'18,May,2004','2004-05-18 00:00:00','    righto... so i\'m not really sure what this is all about, but i\'m going to give it a whirl.  i feel like it\'s weird to start one of these things... i\'m not sure if i should introduce myself or just start talking about what happened today.  the thing is, i don\'t think any \"new\" people will read this, meaning... people that don\'t know me.  so i think i\'ve decided to just write about what\'s on my mind today, right now.    i subbed today at leary elementary school, which was pretty fun.  the class i taught yesterday was horrid, i mean, the absolute worst class i\'ve ever been in for!  so today i was greeted with 19 \"i\'m sorry\" notes from each student.  the teacher felt bad that her students were so terrible for me yesterday, so they spent part of their day today writing me letters.  in a way i still think they\'re little jerks, but it\'s also really cute.  i\'m feeling really stressed out as of late... about everything.  about money, jobs, school, relationships, friendships, this summer, my future, etc.  i\'m in such a weird state of life right now... definitely still figuring things out.  i always thought my life would be so planned out, as boring as that sounds.  when i was in high school i imagined that i would be married and in a career by now.  isn\'t that crazy?  i\'m SO young... i\'m about to turn 23 in a few weeks.  actually though, i\'m really glad i\'m not married and in a career right now, it wouldn\'t feel right.  i still can\'t decide whether to get the blue or red slurpee, how can i possibly be ready to  make decisions about marriage and career?  so anyway... as much as i bitch about my current state of life, i\'m also really happy with it.  i like being so unbelievably free.  if i want to skip town for a few days, i can.  if i want to go out to the bars and have a beer on a monday night, i can.  if i want a hug and to be with someone who cares about me, i have people.  i have so many wonderful people in my life, i don\'t thank them enough!    hmmmm... that all seems kind of random i suppose.  but it\'s on my mind.  just my life, what i\'m doing, and if i\'m happy.  the concluding decision is that i am happy; even though i complain and am really stressed out right now.  things are going okay.  okay, ciao.   '),(20969,3380244,'30,June,2004','2004-06-30 00:00:00','    :)  life is going well, folks!  my date with doug last night was   fabulous, he\'s such a cool guy and we have so much to talk about!  we went to this restaurant on park ave called sinbads, a place i\'ve been eyeing for a while now, so i was glad to finally go.  we ordered 3 meals between the 2 of us- haha, my kinda man!  after dinner we went nextdoor to a bar and talked for hours over some brews.  some of his friends were at the bar and kept buying us rounds of beers... we\'d finish one beer and another one magically appeared.  i love when that happens!  :)  the date was   wonderful, one of the best dates i\'ve had in a very, very long time!!!   '),(20970,3380244,'29,June,2004','2004-06-29 00:00:00','    well, i just came from my interview at DePaul... i was offered the position as a mental health counselor!  though, i\'m not sure i\'ll take it.  ya know, the whole commitment issue.  and the money... wow... if i take this job, it definitely WON\'T be for the money!  i\'ll keep ya\'ll posted.  so i\'m going on my date with doug in a little while, i must say... i like that he\'s planning it.  when he asked me out last night we didn\'t have that \"where do you want to go?\"  \"what do you want to do?\" conversation that sometimes annoys me... he just said \"hmm... i guess i wasn\'t prepared with anything cool to do, so give me tomorrow to plan, okay?  i\'ll call you at 7 and we\'ll go out!\"  :)  i liked that.  i like that he\'s putting effort into our date.  effort- something that i don\'t often find in men.    i spoke with ben a few nights ago... he called at midnight and asked if i wanted to come over.  you all may not think this is a very big deal, but i told him NO!  that was a huge step for me... i have a definite weakness for ben!!!  we\'re still friends, and that i like... but calling me over at midnight was \"friends\" not friends.  and lastly... i spoke with andre\'a today.  he\'s doing well in italy.  when he returns home, he asked if i would accompany him in a nice italian dinner with fresh mozerella cheese, and homemade pasta and wine- brought back from italy!  HOW CAN I TURN THAT DOWN?!   '),(20971,3380244,'28,June,2004','2004-06-28 00:00:00','    i have a date tomorrow!!!  his name is doug and i met him last wednesday night at woody\'s.  he seemed like a really cool guy, i think i blogged about him a bit last week, so i was excited when he called tonight.  we\'re going out tomorrow after my interview with DePaul, a mental health organization where i\'m applying to be a counselor.  wish me luck!!!  so listen to this... i\'ve been doing laundry much of my day today, when i go out to switch loads and realize a load of my laundry is MISSING!  now, i live in an apartment complex- there are 5 apartments in my secure section to this one washer and drier.  i\'ve moved people\'s clothes before if they were sitting in there for a while with no one coming to continue.  but i was in and out every 20 minutes or so!!!  the really weird thing is, they took the load that was in the washer!  not the clothes that were in the drier (they were towels, i suppose that\'s not something you\'d really want to take) and didn\'t touch the piles of dirty laundry i had in a basket, just the clean, WET clothes that were in the washer!  i can\'t figure this out!!!  so someone out there has a load of my wet laundry... what freaks!!!!!  i\'m still trying to figure out what clothes i\'m missing and i\'m NOT happy to be without them!  if i see anyone wearing my clothes- the gloves are off!!!   '),(20972,3380244,'27,June,2004','2004-06-27 00:00:00','    this is  my own little personal space to fill with my opinions, thoughts, and whatever... so i\'m going to take this opportunity to invite all of you to watch FAHRENHEIT 9/11.  i watched it yesterday evening, and it was one of the most powerful films i\'ve ever seen!  michael moore has many critics, though i am one of his supporters- i give him credit for being so bold, and for making people realize how silly they are acting in very matter-of-fact way.  the movie takes you on a roller coaster of emotions... one minute i was sobbing and the next i was pointing and laughing at bush on the screen.  i think it\'s a great film, and i think it can be used as just one tool for people to use when thinking about their vote this november.  thanks.   '),(20973,3380244,'27,June,2004','2004-06-27 00:00:00','    things that make me happy...  *dates that  actually  turn out well!  (yay for andre\'a) *catching up with alan from his china trip, addressing our drama, and dealing with it appropriately *conversations with laura LAURA-  \"i put new sheets on the futon for you.\" ME-     \"screw you!  i\'m sleeping in your bed with you tonight!\" LAURA-  \"hahahaha- i figured!\"  ME-     \"sorry if i spooned you last night, i do it randomly in my sleep sometimes!\" LAURA-  \"you DID!  didn\'t you hear me yelling at you!?, i thought i was going to fall off the bed!\"  *and having a relaxing, funfilled, sunday with my buddy laura!!!!!   '),(20974,3380244,'24,June,2004','2004-06-24 00:00:00','    so- i met a very nice guy tonight at mex.  he\'s italian... like, used-to-smash-grapes-with-his-feet-when-he-was-a-kid, italian.  it was sooooo wonderful to hear his stories from italy- his family not only makes their own pasta sauce, but their own pasta, olive oil, wine, etc.  hearing about his culture was so fascinating for me, his hometown sounds so beautiful.  when i visited italy in 2002 i went to the major cities, but would have loved to travel to the small villages and communities.  he\'s leaving on sunday to go home for a week- oh how i wish i was going to italy on sunday..... sigh.    i feel as though i\'m itching to travel again.  i crave people that enjoy my passion for traveling and seeing what the world has to offer.  there is so much beauty out there, yes even in rochester, and more people need to get out and enjoy it!  andre\'a and i were talking about jobs and work and he said \"i work to travel.\"  and i had to think about it for a second to really get it... but wow, was it powerful to me.  i mean, as my last entry mentions i\'m looking for a lot out of my intended career, but the idea of working just to have enough money to support travels is exciting to me.    on an unrelated note... patrick and i hung out today and got pedicures!  oh how i\'ve missed my patrick.  :)   '),(20975,3380244,'22,June,2004','2004-06-22 00:00:00','    what up folks!?  i just got back from \"ladies night\" at woody\'s... $.50 well drinks, DAMN that can get a girl in trouble!!!  i had a good time, met a really cool guy.  it\'s so refreshing to meet people out there that are down to earth, honest, nice, and  real  .  he didn\'t try to grope me, he didn\'t try to make out with my drunk ass, he was just super friendly and easy to talk with.  though i do still miss ben, sorta... i went out tonight and had myself a grand ol\' time!  :)  hooray for ladies night.  g\'night.   '),(20976,3380244,'22,June,2004','2004-06-22 00:00:00','    dear friends- whenever i am upset, depressed, sad, mad, angry, etc. please just tell me to go to the gym!  it makes the world of difference with my mood.  i can remember when i was dating alan, and we\'d have one of our (many) fights, he\'d say \"kimmy... have you gone to the gym today?\"  (sounds like, \"have you taken your medication today?\"  haha)  i used to get so annoyed when he\'d say that.  but it\'s true.  working out is my release, i gather up all these negative feelings and it affects my mood.  when i was in spinning class today, i felt like a superhero!  i worked my butt off, and it felt soooooo amazing!!!  i don\'t even know where all the energy came from, but there it was.  hooray for spinning!  i sent out resumes and cover letters for some jobs today... i\'m actually feeling pretty hopeful about it!  the jobs i applied for sound super cool- and definitely something which will inspire me.  i\'m super picky about jobs, i\'ve realized.  it\'s quite a commitment and i want to make sure i know what i\'m getting myself into.  i\'m looking for something that will be inspiring, a job that will be motivational, a job i won\'t \"get\" too quickly, a job where i actually want to wake up in the morning and go to, as well as a job where i can help people.  lets just hope i find this amazing position!  :)  ciao.   '),(20977,3380244,'21,June,2004','2004-06-21 00:00:00','    being home for father\'s day yesterday wasn\'t overly filled with drama- surprise!!!!!!  we went out to dinner to this really nice italian restaurant... food was GREAT!  yumm yumm.  it was nice to spend the day with my parents, usually these events are filled with fights, drama, and negative comments, but today\'s wasn\'t much of any of these things.  it was generally, nice!  phew...  though, when i got back to my apartment i had to deal with the ben drama.  we talked on IM, which i suppose made it a bit easier for us to say what we needed to say.  we\'re going back to being friends, and leaving the \"friends\" behind for a while (or forever?).  i suppose i saw this coming, dating doesn\'t last forever, and i really didn\'t see it turning into anything serious.  i am sad about it though, i\'ll admit that.  when ben and i were good, we were AMAZING!  we have this rare friendship that is something i don\'t want to lose... who knows how this will all play out.  he said yesterday that i\'m still really important to him, and for the most part, i believe him.  i guess we\'ll see what happens.  deep down, i understand this is probably a good decision.  i\'m still getting over alan... which will be a lengthy process.  as margaret said to me today... \"there are times in life where you need to go through it alone, i think this is your time.\"  margaret- i agree.  it won\'t be easy, but i\'m realizing the truth and importance in that.   '),(20978,3380244,'20,June,2004','2004-06-20 00:00:00','    well, ben and i are experiencing our first fight- that\'s for damn sure!  i probably shouldn\'t have gone over there anyway, because i was feeling sad about alan... but being that ben is one of my closest friends, i figured it would be nice not to be alone last night.  wrong!  last night was fine... actually it was really nice, we talked and laughed and my mood was lifted in no time.  but this morning ben was in rare form, just horribly moody!  i am   so   nice to him... you don\'t even understand.  i am a nice person, who enjoys treating my friends well.  but when you cross the line and treat me like dirt, watch out...   i\'m realizing that most of you don\'t even know who ben is... and to be honest i know who he is, but not what he is.  ben is my \"friend\" and we\'ve been dating for a few months.  nothing at all serious or exclusive.  weird- yeah, for me!  i know how to be a friend, and i know how to be a girlfriend, this in between shit confuses the hell out of me.  i go back and forth with my thoughts on dating... lately i\'ve felt really comfortable with what ben and i have.  it\'s not serious, there\'s no love... but we like each other enough to not want this to end.  i like the freedom and the space.  and before any of these added bonus\' came about, ben and i are   really   good friends, something neither of us want to lose.  but then WHY is he acting like such a moron lately!???  i was so upset this morning that i walked out of his apartment while he was saying some shit \"tell your father happy father\'s day...\"    *note to men* when you are being a dick and you\'ve had major attitude all morning with a woman, right when she\'s fed up and about to leave, don\'t try and save yourself and say something lame, while completely ignoring your behavior!    (where are all the nice men?)   '),(20979,3380244,'19,June,2004','2004-06-19 00:00:00','    feeling: there is nowhere i\'d rather be right now than in alan\'s arms...   reason:  i think i\'m just feeling sad right now because i waitressed a wedding reception tonight... enough said.          i\'m also feeling weird because i\'m supposed to be in china with alan and his parents right now, but i\'m not.          and lastly, i\'m getting anxious about the wedding alan and i are going to in july together.  i haven\'t seen him since we broke up and i\'m imagining it\'s going to be really difficult.  i miss him like crazy, and i can\'t wait to see him, but i also know that it\'ll be quite emotional.   '),(20980,3380244,'19,June,2004','2004-06-19 00:00:00','    hi.  i\'m super hungover right now... my poor liver.  i\'ve been so unbelievably horrible to it this past month, i hope it can forgive me, someday.  so last night was the east end festival here in roch-cha-cha... it was pretty fun.  i was drinking with my girls early on, so i was pretty drunk by the time we met up with ben at 10pm, which, i felt annoyed him.  oh well.  the festival was crazy... they blocked off east ave, and it was just PACKED with people, bands, alcohol, and religious freaks with signs telling me they were going to save me.  i feel i am spiritual, but not religious.  i\'m not a huge fan of organized religion, and definitely not a fan of so many rules and restrictions that religion often places upon people.  so in my drunken stooper, i decided i wanted to talk to these people.  i was respectful for the majority of the conversation, i\'m not a rude person and i can agree to disagree with someone when we\'re just not going to see eye to eye... but i couldn\'t handle it anymore when he started talking shit about gay marriage, interracial marriage, abortions- i FLIPPED!  seriously... it must take so much effort to be so closed minded, to hate so many people and to be against so much.  can\'t people just open their minds, love and respect the differences in others, and accept?  i realize i\'m probably sounding really naive, but seriously... i get so frustrated with people.  i hate that i let that guy upset me so much, but he did.  i got so emotionally involved in that conversation with him....   '),(20981,3380244,'17,June,2004','2004-06-17 00:00:00','    i went home today.  my grandpa has been really sick lately... and i\'m not dealing very well with it.  my other grandpa passed away when i was in sweden for a semester, and i wasn\'t able to return home.  i think that experience has scared me for life, so whenever my other grandpa goes into the hospital i go RUNNING home to visit him.  i want to be with him, with my family, ya know?  though, on a lighter note... a few weeks ago when he was in the hospital, they released him when i was there visiting him!  so i got to transport him back to grandma\'s house... she was so surprised!!!  and today when i drove home to visit- the same thing!  i was shocked though, he just had surgery on tuesday, but aparently the doc\'s thought he was ready to return home.  it\'s just so hard seeing him so skinny, so weak looking... it\'s not the grandpa i remember.    also... matt\'s tombstone has arrived.  matt was my cousin- we were the same age, and he lived nearby so we saw each other at family functions, school, and everywhere else.  when we were sophomores in college, he was killed in a car accident.  its been a few years, and i\'ve gone through a range of emotions and healing- though i don\'t think i\'ll ever fully understand or deal.  so when i was home today, my grandma mentioned that i should visit matt because his tombstone has arrived and she was telling me how beautiful it was.  and it is... it\'s gorgeous.  but it feels weird to use a word like \"gorgeous\" when describing one of the most horrifying objects in my life.  god... the writing on it, it  makes it so real.  his name printed out in stone, his d.o.b.- only a few short months before mine.  it gave me the chills... i stood outside in the rain, by myself, \"talking\" with him.  i can\'t even put into words how it made me feel...   (elaine, are you visiting?  i could  really  use a girl friend right now...)   '),(20982,3380244,'16,June,2004','2004-06-16 00:00:00','    where does ben get off thinking we\'re at that point where he can  dutch oven  me!?  i definitely don\'t think i\'m ready for that \"stanky ass\" stage yet... :)   '),(20983,3380244,'16,June,2004','2004-06-16 00:00:00','    1st grade student: hey, miss doll?  i feel bad for you, wanna know why?  me: uhhh, sure sweetie, why?  1st grade student: because you\'re not married!  and i know that because your first name is miss not mrs...  me: (laughing) well that\'s okay, i don\'t really think i\'m ready to change my name to mrs yet.  1st grade student:  oh, okay!   '),(20984,3380244,'14,June,2004','2004-06-14 00:00:00','    well now... i\'ve had a crazy, yet incredible past few days in atlantic city!  ya know those friends you have, but unfortunately do not keep in touch with as often as you\'d like to?  that\'s how it is with megan (the bride) and i, we don\'t call each other nearly as often as we should, yet when we\'re together it\'s like time has never passed.  her bachelorette party was simply amazing.  there were 5 of us crammed into villy\'s audi (with no ac) driving down to atlantic city, but we made the best of it.  gotta love girly road trips!!!  we partied like superstars at night in the clubs, and gave megan the rest and relaxation she needed during the day at the beach and spa.    aparently in atlantic city there are no male strip clubs, is that weird to any one else?  so we had to take her to see nekked ladies... i had asked one of the security guards in trump tower which nudy bar he suggested for us, and he mentioned that \"if 6 lovely ladies like yourselves walk into the naked strip club, you\'ll get harassed by guys, so go into the clubs where the girls have g-strings on and you\'ll be safe.\"  it was actually a lot of fun.  both nights we were out until the sun came up... i don\'t know how we made it, but i suppose lots of alochol and dancing can keep anyone in party mode.  during the day we layed out on the beach and walked around the boardwalk.  we even tried out for wheel of fortune... they had a contestant search that weekend in atlantic city, and though none of us made it, we still had fun!!!  we all met some very interesting male species... the guys that asked for my phone number (yes, believe it or not, there were actually quite a few!) i gave them pitt\'s (the groom) number.  he\'s gotten quite a few phone calls already asking for me.  muhahahahaha!  when i was sitting down with ben last night telling him my stories from the weekend, i realized that for the past several days i have been completely in the \"here and now.\"  a term often used in my grad classes, but also a term that i have a very difficult time understanding.  i find myself constantly thinking and preparing the future, and rarely enjoying the  moment.  my days are run by money and time... and i feel as though i need to be overly practical with both.  but this weekend, was different.  i couldn\'t really afford that trip, and when i should have driven right back to rochester on sunday night, i decided to stay the night in binghamton and catch up with pitt and lat.  monday morning i went out to breakfast with megan, pitt, and lat... then we went back to their house and looked at pictures and videos from parties at college.  good times.  it was great to hang out with them again and reconnect!!!  then on my drive back to rochester, when i was traveling through syracuse, i decided to try my luck and see if gigi was around.  so i stopped by gigi\'s work and hung out with her for about an hour!  i took my time getting home and spent time with friends, because that seemed to be more important.  even though i have a million things that need to be done... it didn\'t seem overly important yesterday.  pictures will (hopefully) be posted soon...   '),(20985,3380244,'10,June,2004','2004-06-10 00:00:00','    well... today sure has some interesting twists!  i met my good friend, megan for lunch today at pho 99... the best restaurant in buffalo!  mmmmmm!  after stuffing our faces, we caught up on all our girly talk.  i\'m a bridesmaid in megan\'s wedding this august, so we had plenty to discuss.  anyway, she mentioned that her bachelorette party is this weekend and there is still an opening if i wanted to go.  see... initially, i told her i could not go, because of work, money, traveling to china with alan, etc.  i felt terrible for not attending, but it just wasn\'t gong to work out.  well... as luck would have it... i can go, and it\'s really not a problem at all.  the god\'s were with me today, because i have the whole weekend off without even asking for it!  i\'m not going to china anymore (unfortunately) and i\'ve got a few extra bucks and off i go... to ATLANTIC CITY!!!  hooray.  i\'m super excited and i know we\'ll have a blast.  boardwalk and beach- here i come!!!  next order of business... STEVEN FISKE CALLED ME!!!  so yesterday i left him a phone message, and i would have bet my life that he wouldn\'t have the balls to return it, but he did!  come to find out... steven and lola did date for 4 years, but this past december steven found out that lola was cheating on him and they broke up.  because they\'ve been through so much together, they continued to sort of date on and off... as it was difficult for both of them to stop all communication.  so they still see each other every few weeks, but are not together.  according to steven, she\'s having a hard time getting over him and she can\'t handle him seeing anyone else and that they are NOT together!  funny thing is- i believe him.  now, i don\'t know steven overly well, i\'ll be the first to admit, but i honestly don\'t think he\'d lie to me.  i think he\'s telling me the truth about him and lola.  don\'t worry though... i\'m not going to go running up to toronto into his arms, because i didn\'t get out of a perfectly good long distance relationship to get myself into another one, but i think maybe next time i travel to toronto, i\'d call him up and see what he\'s up to?  i dunno....    life has a funny way of speaking to me...   '),(20986,3380244,'09,June,2004','2004-06-09 00:00:00','    dear blog readers... you\'re going to love this one!  remember that nice fellow, steven that i met this past weekend in niagara falls?  the one that said, \"you made me weekend, and just the thought of you makes me smile.\"  well i have an interesting story to update you about dear mr. steven.  i get an email (yes, i check my email while i\'m subbing) from a girl named lola and the subject is \"regarding steven fiske.\"  i open it up and it reads: Who are you, and why are you emailing my boyfriend?  well, OH MY GOD!  i\'m going to ignore that this chick is reading up on her boyfriends email, because this scum bag deserves her snooping!!!!!!  so miss lola and i have become quite close today, emailing back and forth telling each other every bit of detail we need to know.  i told her everything... i told her that her boyfriend kissed me this weekend and swept me off my drunk feet!  i told her that he\'s been emailing me everyday since asking me to come visit him in toronto and how amazing of a girl he thinks i am, blah blah blah!  she deserves to know the truth.  i\'ve been that girl who\'s been cheated on, and i would have appreciated the other girl to be as honest with me.  so honesty is what lola got, and appreciated.  it\'s too bad though... they\'ve been dating for 4 years!  so, men out there... DON\'T CHEAT!  my god, don\'t get your dumbass selves into relationships that you can\'t handle being faithful in.  if you cannot handle keeping your penis and lips off of another woman, then break up with her!  don\'t be a fucking loser and cheat!    STEVEN FISKE... you are filth!   '),(20987,3380244,'08,June,2004','2004-06-08 00:00:00','    \"I\'ve seen you stay with losers that don\'t deserve to clean your toilet... and seen guys that seem so wonderful go to shit... and through it all I never doubted how amazing you are.  And that I know... eventually... some guy will get his head out of his ass and not take you for granted, but only take you as wonderful.  Until that day... you\'ll stay so close to perfect... and I\'ll stay here and be there for you every time... through them all... and always there to hear anything that happens... \"  all i have to say is, wow.  thank you for that.  though, i think you are a bit harsh on my past boyfriends... it means a lot to know that i have friends that will always be there for me and believe in me.   '),(20988,3380244,'07,June,2004','2004-06-07 00:00:00','    \"you made my weekend, and just the thought of you makes me smile.\"  someone said that to me... do you believe that no one has ever said that to me before?  regardless if anyone has ever felt that, no one has ever put it into words and said that to me.  it felt so nice... so sweet and genuine.  :)  yay.  on an unrelated note... i\'m sick!  i\'ve gotten sick so much this year, i think it\'s from subbing and being around dirty kids all the time.  not that i completely mind, becuse i love the kids, but i could really do without being sick all the time.  booo...   '),(20989,3380244,'07,June,2004','2004-06-07 00:00:00','    well now... my birthday weekend is officially over, and it was an amazing one!  i\'m going to have to say that it was one of the best birthdays, ever... and that\'s tough to beat.  i wasn\'t overly excited about going to canada, because i was still bumming about the cancelation of the whole limo party, and the lack of everyone going.  it ended up being only me, tammy, amy, and amy\'s boyfriend- john.  but let me tell you, i had a  great  time!!!  we met these guys from toronto that were down for a bachelor party, great group of fella\'s!  i met one in particular that caught my eye, steven, he was very very nice.  i\'d like to meet up with him again if/when i visit toronto next.  we shall see...  the drive home from canada was tough... i was exhausted (as i got no more than maybe, 4 hours of sleep saturday night) and tammy was vomiting much of the drive back to roch-cha-cha.  poor girl.  i felt so sorry for her, being that hungover is not fun.  so i pulled into my apartment thinking i\'d hop into bed, snuggle with my teddy and take a nap all afternoon... but, instead i went to a greek festival with ben.  i suppose i\'m easy to convince, as my favorite phrase as of late is, \"sleep is for retirement!\"  so off i went... the festival was fun- we ate greek food and beer.  yum!  then we went to abbotts for some ice cream, no pistachio pineapple but i was still a happy camper.  we sat outside on park avenue, people watching for about an hour, was nice.    i\'m really contemplating moving to san francisco.  it\'s something i\'ve always wantend to do, and i don\'t want to be one of those people that talk the talk, but there is never any action to all that talking.  after living there last summer, i was convinced that SF is where i belong.  the people are so different... they\'re open minded, they\'re friendly, they\'re outgoing, they\'re into having fun, and diverse!!!  what more can i ask for?  it helps to have my brother and his girlfriend living there, so i know all of their friends, plus the friends i made when i was there last summer.  if things don\'t work out with brockport, i\'m gone.  and if things do work out with brockport, as soon as i have my master\'s degree... i still want to move.  this is what ben and i talked about while at abbott\'s yesterday... california dreamin\'     :)   '),(20990,3380244,'05,June,2004','2004-06-05 00:00:00','    what up folks?  so my birthday was yesterday... it was such a great day!  thanks for all of those who made it an extra special day for me!!!  i had a hot dinner date with tammy at lola\'s, such a great restaurant.  one of the best meals i\'ve ever eaten, they have the best chicken french!  mmmmm.... the last time i\'d eaten there was the last time alan visited, for our year and a half anniversary.  right.  anyway, after our yummy dinner we picked up ben and went to a carnival for a little bit.  i wanted to visit jason and melissa, they were working the \"big jumping cage thingie\", so that was boat loads of fun.  while meliss, tammy, and i jumped our hearts out, ben was inside playing poker... and won $50!  yay- go ben!  after the carnival tammy, ben, and i went to the geva theater for an improv comedy show, which was very very funny.  afterwards ben and i went to the alexander street pub for some dancin\'.  why o\' why do black men find me so attractive?  i don\'t consider myself a \"hot girl\" but aparently black men think i am.... maybe it\'s my ghetto booty?  who knows.   so, tammy\'s on her way over here now... we\'re off to canada for the night to continue the birthday celebrations!!!  yipeeee....   '),(20991,3380244,'04,June,2004','2004-06-04 00:00:00','    i\'m typing this between tears... i\'m such a sap.  birthday\'s mean so much to me and you\'ve all made it so incredibly wonderful.  things have been going so shitty as of late in my life, and i sort of lost hope for my birthday being a big deal... or even that people would remember it.  i was wrong.  i had a wonderful dinner with my parent\'s last night- i had my first shot with my parents!  sort of weird looking out of the corner of my eyes seeing drew and linda kick one back, but also pretty cool.  i\'ve gotten so many sweet cards from friends, emails from friends in germany, holland, sweden, and japan... geez- i feel SO lucky to have all these amazing friends!!!  and i was greeted at my apartment door with a beautiful bouquet of flowers from alan... and a card that made me sob.  you all mean so very much to me... thank you for today!!!  :)   '),(20992,3380244,'03,June,2004','2004-06-03 00:00:00','    okay, even though life sucks right now... i did find a reason to smile last night.  i went over to ben\'s last night around 9pm and we decided to go for a walk.  i hadn\'t eaten dinner yet, so we decided to stop at this place on park ave. that we\'d both heard so much about.  it\'s an asian restaurant that a girl whom also applied to my grad program works at, actually, her parent\'s own it.  so i heard about it from her during her interview, and ben had heard about it because one of the guys on his softball league goes there a lot, aparently it\'s really good.  well, wouldn\'t you know margarie (girl in my classes) and the softball guy (eek, forgot his name!) were both there!!!  it was so great.  so i talked with margarie and she kept saying how wonderful she thought i did atthe interview, and how after i went, she felt so intimidated because i did so amazingly well.  blah, blah, blah.  she was shocked when she found out i only got conditionally accecpted; she was accepted.  i\'m really proud of her, she did well... but it\'s hard hearing from SO many people that they thought i did really well, when i only got conditional.  UGH!    anyway, i ordered udon.  mmmm... one of my favorite dishes.  and they had BUBBLE TEA!  oh my gosh, i looked so hard for bubble tea in rochester (obviously not hard enough!) and i just didn\'t think there were quality asian places around.  i was definitely wrong!  i was looking for something homemade and small... not a big chain restaurant.  something that would taste similar to alan\'s mom\'s food.  something that felt comfortable and warm and inviting.  this place did!  it made me miss alan a lot... seeing margarie\'s mom... hearing her talk.  i got thinking about the china trip i was supposed to go on with him and his family this summer... i SO wanted to go.  i love alan\'s family so much... it\'s difficult when you date someone for so long, you don\'t just break up with them, you break up with their family too.  \"break up\" is such a horrible term... sounds so cold and lost.  which isn\'t alan and i at all... this has been the most healthy break up i\'ve ever been involved with.  anyway, i\'m getting too deep into that... i was happy to have bubble tea and udon!!!!!!   '),(20993,3380244,'02,June,2004','2004-06-02 00:00:00','    well now... as i wrote yesterday, things are not going well.  at all!  things are going amazingly horrible right now, and they only got worse last night and this morning.  i felt like a stranger with one of my good friends last night, that made me feel really sad.  i felt as though i didn\'t belong and i was no longer welcome.  which isn\'t the case, i do believe i am welcome... but i can\'t erase my odd feelings.  i didn\'t sleep, i coudln\'t.  i kept having weird dreams and weird happenings.... i wish i could give you more detail, but it wouldn\'t be right.  also... my one glimmer of hope, my lovely birthday party.  is now canceled.  the limo bus filled with 16 of my friends, partying the night away is now canceled.  right.  why would i assume i could at least have a happy birthday?  why would i assume such a thing?  life BLOWS!   '),(20994,3380244,'01,June,2004','2004-06-01 00:00:00','    wow- am i cranky right now!  i\'m not even exaggerating when i say that things are definitely not going my way right now.  in order for something to be taking place in my life right now, it must be both fucked up and complicated.  part of me wants to apologize for my entries always sounding so pissed off and negative... but then i think apologizing for that would be stupid.  this is one of the few spots i am able to rant, everyone see\'s me as this happy, friendly, cheerful girl... heck, i\'m getting my master\'s degree in counseling, so people assume, \"her life is so great, that\'s why she\'s the counselor, because she\'s got shit figured out!\"  WRONG!  i have problems too, i have bad days, days that make me want to cry myself to sleep, days that make me wonder why i even bother... i\'m human too, ya know?  so i\'m feeling really sensitive as of late... i can take a joke, i can.  i know i dish it out enough, i should be able to take \'em back.  but, the thing is... i get made fun of a LOT wheni i\'m home.  my friends, and especially my family- i am the butt to all the jokes.  i kid you not!  i can only take enough jokes, ya know?  why me?  why do they feel they can make fun of me, are they perfect?  does it make them feel better about themself to make fun of me?  i just wish the jokes could decrease, even just a little would be nice.    my 23rd birthday is on friday.  i don\'t have any plans (on my actual birthday).  i\'m having a really cool part/bar thing on saturday night, which i\'m REALLY looking forward to, but nothing for friday.  i realize it\'s my birthday, i should probably make a plan or something... but i sort of wished someone would suggest something, or offer to take me out.  i know that probably sounds really selfish... i don\'t mean for it to.  i think i\'m famous for wishing people could read my mind.  i don\'t know why i do it...   btw... donz thanks for the comment!  i didn\'t think anyone read this thing.  thanks!  :)  i\'m out.   '),(20995,3380244,'30,July,2004','2004-07-30 00:00:00','    within one hour yesterday i noticed 3 things that were  really  sweet and made me smile to myself... (1) i let spencer go. drove to the erie canal (his original home) and plopped him back in... he was soooo happy to be back. he kept bobbing his head up and down and swam like such a little champ. i swear he was across the canal in about a minute. i think spencer forgot he was a turtle and is supposed to be slow, \'cause he is quite the swimmer. spencer- i miss you, buddy. (2) i was in wegman\'s strolling past the flower section when i saw an older man picking out a beautiful bouquet of red roses. he was smiling from ear to ear, it was contagious. i found myself grinning just looking at his happiness. i walked right to the card section of wegman\'s (i love card shopping) and soon found mr. flowers. he, again, was smiling from ear to ear picking out a card. i\'m assuming the card and roses were for the same person and it made my heart melt. he seemed so happy and excited about the gifts... there  are  sweet men out there! (3) i was driving back to my apartment, stopped at a redlight. i glanced in my rear view mirror and saw 2 love birds in the car behind me. the girl was driving, and the guy was the passenger. he put his head on her shoulder and appeared to be singing... while she was smiling and leaning her head against his. they looked really happy.   '),(20996,3380244,'28,July,2004','2004-07-28 00:00:00','    good day, ya\'ll.&nbsp; so day 3 of work has come and gone and it\'s been going GREAT!&nbsp; i actually feel bad for having so many preconceived notions about DePaul, because the house i\'m working at now is completely different than my other one.&nbsp; the counselors i\'m working with at this house are all in their 20\'s and they\'re SO fun, easy going, sweet, and caring!&nbsp; its such a great work environment for me... i\'m really happy there.&nbsp; and the residents are beginning to warm up to me, they\'re all mentally ill and don\'t deal overly well with change.&nbsp; i\'ve spoken with a few of them 1 on 1 and i think i\'m really getting somewhere with them... the wall behind my desk is already half filled with pictures they\'ve made for me.&nbsp; :)&nbsp; yay!&nbsp; i just came from a depaul luau this evening.. all of the houses&nbsp;in rochester met at the administration building for a luau and it was&nbsp;a blast.&nbsp; i did the chicken dance and electric&nbsp;slide with a bunch of the clients- too cute!&nbsp; in september i\'m attending camping trip with my house and several others, i\'m really looking forward to it.&nbsp; it will be great for the residents to go camping, and also a great bonding experience for me and all the other counselors... so yeah, job is going VERY well!&nbsp;   things with ben have been going really well... he\'s definitely&nbsp;my best friend here in rochester.&nbsp; there are few people that i can completely be myself around, and ben is one of them.&nbsp; i can say anything to him and he\'ll listen.&nbsp; we can go from acting like brother and sister and teasing each other to having a really serious talk about love and life- a&nbsp;versatile friend, if you will.&nbsp; oh- i\'ve also been playing a lot of pool!&nbsp; and i LOVE it.&nbsp; ben&nbsp;is  very &nbsp;good, and he\'s been teaching me his skillz... i was at a bar last night playing doubles with doug and i made some amazing shots!!!&nbsp; i wish ben was there to see it, he would have been really proud.&nbsp; and speaking of pool- i\'m off in a few to go shoot some with ben and chad.&nbsp;   ciao ciao...   '),(20997,3380244,'26,July,2004','2004-07-26 00:00:00','    so- this morning ends my wonderful summer vacation, i begin work today.&nbsp; though i am very excited to get back into counseling and working with people, i\'m  really  going to miss all of my free time.&nbsp; my last week&nbsp;of time off was amazing... i really packed all that i could into it.&nbsp;   wednesday night ben and i went to a happy hour for the kickball team he played on (the game was this past saturday).&nbsp; everyone from the team showed up, which was really nice, so that everyone could meet and mingle before the big day on saturday.&nbsp; they made team t-shirts and were able to put&nbsp; names on the back... ben chose \"MR. RIGHT.&nbsp; (single and ready to mingle)\"&nbsp; which i thought was very funny.&nbsp; this other guy chad whom we quickly bonded with decided to change his name from whatever it was to, \"MR. WRONG.&nbsp;(scary, hairy, and ready to marry)\"&nbsp; it was hilarious!!!!&nbsp; so ben and i hung out with our new friends stef and chad all night and had a blast, though we were all disappointed that i wouldn\'t be at the festival on saturday to watch them play, because i couldn\'t afford to go.&nbsp; :(&nbsp; it was $30 just to get in the door (but you ate and drank for free all day!)... and seeing as though i had to work at 6pm, i didn\'t think i should spend the money i don\'t really have on that.&nbsp; so saturday morning i was ben\'s ride to the festival, since he\'d be drinking all day i said i\'d drive him there and pick him up so that he wouldn\'t have to worry about calling a taxi or whatever.&nbsp; but when we woke up he sort of pouted and said \"i really want you to be there today, it won\'t be the same if you\'re not.\"&nbsp; he offered to pay for half my ticket, which initially i felt really weird about, but he insisted!&nbsp; so- yay, i went to the 10 ugly men festival.&nbsp; weird name i know, but damn it was fun!!!&nbsp; the kickball tournament was a blast, their team did really well... even though the last several games, they were all drunk!&nbsp; there were people everywhere, bands playing, and lots of great free beer and food.&nbsp; i have this super hot professor that\'s in a band, and they happened to be playing at the festival saturday... yay!&nbsp; it was cool to see \"dr. hernandez\" all jammin\' out with his band... soooo hot!&nbsp; :)&nbsp;   great times, great friends, great party........... i\'m off to work.   '),(20998,3380244,'23,July,2004','2004-07-23 00:00:00','    hello folks... can i just say that i  love  rochester!&nbsp; when i first moved to rochester i didn\'t really get out much, i had a serious boyfriend, i had 3 jobs, and grad school and i had no great desire to go out and discover my town.&nbsp; lame i know, but it was the damn truth.&nbsp; i actually used to call it \"crap-chester.\"&nbsp;   i am a changed woman though, i have been out and about discovering all that rochester has to offer lately and i\'m quite impressed, delighted, and happy with my surrounding area.&nbsp; i was shopping at  urlLink http://http://www.parkleigh.com &nbsp;(parkleigh) this afternoon- enjoying their summer sale and all the fun things they sell.&nbsp; i could spend hours in that store... and i finished it off with a godiva raspberry truffle.&nbsp; Mmmmmm!&nbsp; on my walk back to the car, i passed a house with a sign out front, \"the dutch market.\"&nbsp; this was very appealing, because i have been to germany, and several scandinavian countries and i truly miss the food, smells, people, and chocolate!&nbsp; i stepped in and it felt like i was back in a small pastry shop in sweden... i bought some chocolate (and ate that, right after i finished my truffle) and picked up a menu.&nbsp; i want to go back someday for afternoon tea or a pannekoeken... anyone interested in dining with me?   when i returned home, i ate an extra large salad... felt i had to after all that chocolate!&nbsp; :)&nbsp; so tonight is the 2nd east end festival, i hope it goes well.&nbsp; i remember the one last month was less than pleasing for me, i went with ben and we ended up running into a group of girls that oogle over him.&nbsp; they gave me mean looks and were rude to me, and i remember in my drunken stooper trying to run away from ben and take a taxi home.&nbsp; basically, i was being a stupid drunk girl, and ben was being stupid horny boy.... we\'re not a good combination when we\'re like that.&nbsp; this evening i\'m attending the east end festival with ben and tony (his best friend from high school that\'s back in town- yay!).&nbsp; i think it will be fun, because the 3 of us are a lot of fun together... but i have those stupid memories from last month that won\'t go away.&nbsp;   in closing i will say this... rochester, i heart you!   '),(20999,3380244,'22,July,2004','2004-07-22 00:00:00','    i so rarely remember my dreams, but i must share with you my dream that i actually remember from last night.&nbsp;&nbsp;&nbsp;i suppose eating pita pitt right before going to bed wasn\'t the smartest idea, because it gave me some  crazy  dreams, YO!&nbsp;   i was sitting at a friend\'s house (not sure who\'s) on their couch... it was me and 2 friends.&nbsp; we were talking, being girls, and just hanging around when we heard this really loud noise!&nbsp; i looked out the big bay window that was opposite the couch i was sitting on and i saw fire and&nbsp;stuff flying&nbsp;everywhere.&nbsp; it was a BOMB!&nbsp;&nbsp;it eventually came through and broke the living room window.&nbsp; my friends and i were covering our heads, just trying to stay alive.&nbsp; the room caught on fire and we had to find our way to the door so that we could find safety.&nbsp; i don\'t really know what else happens after this.... but isn\'t that CRAZY!?&nbsp;   i looked up in an online dream dictionary what dreaming of bombs means and here is what it tells me:  Dreaming of a bomb means that you may be going through a potentially explosive and trying situation in your waking life. The bomb could represent repressed desires and unexpressed emotions that are likely to explode or burst if not dealt with soon. It could be something within yourself, such as the desire to explode with anger over an issue that\'s affecting you.  &nbsp;  ummm... whoa.&nbsp; i don\'t even know what to say about that.....................   '),(21000,3380244,'21,July,2004','2004-07-21 00:00:00','    yesterday was the best day i\'ve had in a very looong time!&nbsp; i crave being spontaneous and the past 24 hours have been just that!&nbsp; i left around 10pm on monday night and drove to ithaca to say goodbye to one of my bestest friends, diana.&nbsp; she and her boyfriend are driving across country and moving to california... diana already has an apartment in hollywood!!!&nbsp; wow.&nbsp; so i definitely needed to meet up with them in ithaca and wish my girl luck and have one last crazy night with&nbsp;Di before she migrates west.&nbsp; my time in ithaca was  aaaaaaaaaaaamazing!&nbsp;  it included numerous shots and beers with friends, sleeping at alpha delta phi one last time, lunch at viva taqueria with margaret, shopping in the commons,&nbsp;wearing my \"ithaca is gorges\" t-shirt while climbing one of the gorges (so cheesy, yet so fun), jumping into the gorges in my bra and underwear, dinner and pie a la mode at dories, a beautiful walk around wells campus, watching the sunset at the aurora inn, and jamming to  rusted root  and  cake  on the drive back to rochester.&nbsp;   doug accompanied me on my trip to ithaca.&nbsp; we had talked a few weeks ago about going to ithaca at some point this summer, so when i decided to go i called him up and invited him to go with me.&nbsp; yes, doug is the guy that left me in the ER last week, but... we\'ve talked about it and i think i understand his logic and thinking of that night.&nbsp; he feels terrible and wasn\'t really thinking clearly (obviously)... i haven\'t forgotten, but i\'ve forgiven.&nbsp; anyway, it was really nice to spend the day with him in ithaca... he\'s really adventurous, and always up for having a good time.   today i\'m going to lay out by the pool and enjoy the evening hanging out with ben... i\'m going to enjoy this last week of unemployment as best i can!!!   &nbsp;   '),(21001,3380244,'19,July,2004','2004-07-19 00:00:00','    ben is back in my life.&nbsp; this makes me really happy- i missed him when we weren\'t talking/seeing each other.&nbsp; it\'s interesting... you learn so much about a person in a state of urgency that you were completely unaware of when life is \"normal.\"&nbsp; ben has a heart of gold, but unfortunately for him, so few people are able to see that.&nbsp; anyway, i just wanted to mention that i\'m glad ben is back.  &nbsp; yesterday on&nbsp;my drive back from 6 flags melissa, ben and i stopped for some ice cream.&nbsp; i was really craving ice cream, so it was&nbsp;mostly my idea... i asked the man behind the counter if a certain ice cream had nuts in it (i am deathy allergic to most nuts).&nbsp; his response, \"no, i don\'t think so.\"&nbsp; i said... \"well, can you please check, because i\'m allergic?\"&nbsp; and he said... \"i\'m sure... there are no nuts in that one, it\'s just vanilla ice cream, raspberry swirls, and chunks of chocolate.\"&nbsp; so i ordered myself a cone of that and went along my merry way... 3 licks into the ice cream cone i realize that the chocolate chunks are crunchy, becuase they have NUTS IN THEM!&nbsp; my goodness people- this is sooooo not something to take lightly!!!&nbsp; people must think allergies mean- sneezing, coughing, or scratching?!&nbsp; when i say i\'m deathy allergic to nuts, they must think i\'m kidding?&nbsp; ben wanted to go back to the ice cream shop and smash it in that man\'s face, but we didn\'t.&nbsp; oddly enough, i get embarassed when this happens... it happens frequent enough, but i rarely make an issue of it. &nbsp; i\'m off to the chiropractor... yay!   '),(21002,3380244,'16,July,2004','2004-07-16 00:00:00','    i was lucky yesterday, folks!&nbsp; doug and i were driving back from my friend tammy\'s house when we got into a car accident...&nbsp;i was rear ended by a big dodge ram!&nbsp; \"oh my god!&nbsp; i\'m so sorry, my foot slipped off the breaks!&nbsp; oh my god!\"&nbsp; my little saturn was at a complete stop at a redlight when his huge truck slammed into the back of me, and i went forward and hit the car in front of me!&nbsp; we sure got bumped around.&nbsp; so, most importantly... everyone is alive and okay.&nbsp; and actually, the only car damage that was done was to his big honkin\' truck!&nbsp; which i find slightly amusing... so many people buy suv\'s or huge trucks because they feel safe in them, but in the accident yesterday, the huge vehicle is the one that had the most damage... his whole front bumper was smashed in and my saturn went off without a scratch.&nbsp; HA!&nbsp; because i\'ve been in a car accident prior to this one, and have had neck/back problems since my other accident, i went right to the ER yesterday to get x-rays because i was hurting right away.&nbsp; x-rays came out fine and my only diagnosis is \"cervical strain\" which is fine- could have been a lot worse- but i am SO sore and uncomfortable!!!&nbsp; it\'s more my upper back than anything else... it huuuuuuurts!&nbsp;  &nbsp;  i wanted to share a story though about the ER yesterday.&nbsp; so- i  really really really  hate being in hospitals, and especially emergency rooms!&nbsp; this probably isn\'t common knowledge about myself, but i don\'t know anyone out there that really&nbsp; loves  the place.&nbsp; so doug came with me to the emergency room, i thought because he was in the accident with me and was immediately complaining about back pain as well that he\'d also get checked out.&nbsp; we drove separately to the ER (yes, i drove myself there)... so imagine this... i\'m signing in and filling out paperwork when they put a neck brace on me and put me in a wheel chair.&nbsp; it looked a bit dramatic, but my blood pressure was sky high and that\'s probably standard for neck pain.&nbsp; so i\'m sitting there (unhappy and scared, because i  HATE  emergency rooms) with my neck brace in a wheel chair when doug leans in, kisses me and says \"alright... i\'m gonna get going, my softball game starts in half an hour.&nbsp; i\'ll call you later to see how you are, okay babe?\"&nbsp; i was in such shock that i couldn\'t even say anything!&nbsp; to be in an emergency room anyway scares the heck out of me, but to be LEFT in one all by myself?!!!!&nbsp; oh my god!&nbsp; i freaked out and demanded that i was wheeled outside so i could use my cell phone to \"phone a friend.\"&nbsp; i called my melissa (my cousin\'s fience, and my new roommate starting in a few weeks) and she drove over in a flash and saved the day.&nbsp; THANK YOU, MELISSA!!!&nbsp; so i find it humorous that the guy that everyone thought was soooooo sweet and cared about me sooooooo much actually left me there in the ER by myself.&nbsp; i realize i was mostly fine, and i wasn\'t dying or bleeding, but c\'mon.... have a heart!&nbsp;   '),(21003,3380244,'14,July,2004','2004-07-14 00:00:00','    wow... it has been a while since i\'ve blogged.  if anyone still reads this, i\'ve got lots to talk about!  my overall mood as of late has been utter confusion.  i\'m wondering when i gave up the power to make decisions about my life and my time and allowed others to make those decisions.  i just feel as though i\'m being pulled in so many different directions.  i\'ve been jobless for the past 3 weeks, so one would assume that i\'ve had lots of free time to do as i please... not the case at ALL!  everyone else has their opinion about what i do and who i should spend it with, and i\'ve had absolutely no time to myself.  complaining about this may sound stupid to some, but it\'s been difficult.  i\'ve started dating- which has been super nice!  i\'ve met guys that are truly great people, and so sweet.  doug buttered my banana bread the other day... and it made me realize, no one has ever done that for me before.  (i sound like that chick in america\'s sweethearts)  so dating has been very nice because i\'m finally being pursued, cared about, respected, and taken care of.  it feels SO nice.  though, it\'s tough because there is more than one and they both call SO often.  gosh, i feel like such a tool complaining about something like this, but it\'s the truth... i just feel pulled in too many directions.  from these guys, from friends and from family- i want to scream \"IT\'S MY LIFE!\"  anywho... i feel better now that i vented.  things with alan... i\'ll try my best to keep this short and sweet.  first of all, elaine visited on a tuesday (alan\'s x-girlfriend/my good friend) and it was SO WONDERFUL to see her and spend time with her.  at night we layed in my bed for hours talking about alan.  since we\'ve both been through the \"alan breakup\" we have a lot in common and a lot to talk about.  we laughed and cried and eventually fell asleep.  was really good for me to have her here and to prepare me for alan\'s visit, which was the following day.  alan arrived on a wednesday and it was really nice to see him.  when i first saw him i hugged him and it honestly felt like home.  so good to see him again.  the 1st part of the trip was blissful, he\'s changed... grown up a lot since our breakup.  and he was super sweet and caring with me- it felt so nice to have that come from him.  but then came the middle of the trip (he flew to NY because he was the best man in a friend\'s wedding) when we were in rome, ny for the wedding... and things got emotional.  in one of our many car rides i suddenly burst into tears that would not stop flying out of my eyes!  i felt so confused, sad, hurt, annoyed, and alone.  the following 24 hours were difficult... i was crying most of them, along with fighting with alan.  this was the relationship with alan that i was used to (unfortunately)... the fighting and crying were all too familiar.  i\'m not saying that\'s all my relationship with alan was, because i love alan and we had many, many, many absolutely wonderful times.  but towards the end of our relationship when we lived in different states, the crying and fighting were more frequent than i would have wished for.  :(  though, we did end on a good note... i realize that alan is an absolutely wonderful person and i still care about him very deeply.  but alan and i are not meant to be together as a couple- that is very clear to me.  and on an unrelated (to alan) note... the wedding was a BLAST!  i got to hang out with lots of his friends from high school again and i had a fabulous time.  they are some great folks and we partied like rockstars aaaaaaaaaaaaaaall weekend.    look for new pics,  urlLink   '),(21004,3380244,'05,July,2004','2004-07-05 00:00:00','    well hellooooo there!  thanks to those of you who responded to my job drama- i\'ve been thinking about it a lot, and i\'ve decided to take the job at depaul and go from there.  i\'ll figure things out as i go... but i do know one thing for sure... i need $$$$$$!  i have   very   exciting news to share with all of you, my brother (my most favorite person in the whole world) proposed to his wonderful girlfriend, brigid on july 3rd!!!  i\'ve been waiting soooooo long for this to happen, but it was well worth the wait.  i heard all the details from brigid and it sounds like my brother is a total sweetheart, and the ring is GORGEOUS!  good job, bri!!!!  i absolutely love brigid and i can\'t wait for her to be a part of my family (officially).    my weekend was pretty fun, after work on saturday i went out with a woman from work... she\'s 45 and spunky as a 21 year old.  ellen is a blast!  we met up with doug, so it was nice for her to meet him... she was quit impressed with him!  we\'ve spent much of this weekend together (doug & i), which has been really fun!  he\'s a great person and i love hanging out with him... but i am scared to death of anything serious right now.  the timing for a relationship just isn\'t right for me right now.  hopefully this doesn\'t end up hurting him in any way.  i\'ll keep you posted.    i\'ve got some visitors coming into town this week!  elaine is coming tomorrow, which i\'m super excited about.  she\'s actually alan\'s x-girlfriend... may seem odd to some, but honestly i\'ve gotta hand it to alan, he\'s got GREAT taste in women!  elaine is such a wonderful person- she\'s beautiful inside and out!  and i just can\'t wait to spend the day with her and show her my town.  :)  and alan\'s flying into town on wednesday night; we\'re going to his friends wedding in rome, ny this weekend.  i\'m looking forward to seeing alan, i won\'t deny that i\'m a bit nervous about how it will be, but for the most part i just can\'t wait to see him again!!!!!!!!  peace out.   '),(21005,3380244,'01,July,2004','2004-07-01 00:00:00','    okay blog readers... i need your advice.  i don\'t often ask for it... but i\'m really in limbo right now with the whole job situation.  so i was offered the job at DePaul, where i\'d work as a residential mental health counselor.   pro\'s of working at depaul:  i wouldn\'t have to be trained, because i worked there throughout college; it would be a stable income, money and work that i could rely on; it would be great experience for me; i would feel passionate about the work i do   con\'s of working at depaul:   i would make enough to pay my bills but not too much more, it would be stressful and i\'d easily get burnt out, and the hours kinda suck, i\'d constantly be thinking about how cool the other jobs i applied to are, and wish i was working elsewhere...  so because i need a job real soon... i\'m running out of money... part of me wants to take the depaul job for the time being while the other jobs that i applied for are making their decisions (it could be a month or longer until they do interviews and such).  these other jobs are AMAZING and the pay is about $12,000 more money that depaul would be!  but, then my conscious kicks in, and it would suck of me to work at depaul and then quit if something better comes along, right????  see- it\'s this whole commitment issue i have.  HELP ME PEOPLE- what are your thoughts!!!!!!!!!!!!!!   '),(21006,3380244,'09,August,2004','2004-08-09 00:00:00','    urlLink http://dreamofrain.diaryland.com/estrogen.html   :)  thanks lainy.  i love and miss you.   '),(21007,3380244,'02,August,2004','2004-08-02 00:00:00','    i  love  my apartment!!!  melissa moved in this past weekend and it has been soooo great for both of us!  it\'s really nice living with a girl again... i missed it.  especially someone so considerate, clean, understanding, and real- things are going to work out well.  we had a good weekend driving the huge uhaul all around corfu, oakfield, rochester, and webster- i felt super badass driving that thing.  :)  unfortunately we had no one to help us move our furniture in... it was sunday evening and all of my friends were at bbq\'s or completely drunk (it was park ave fest) so that was a bit tough.  we thought we could get our new furniture in the apartment by ourselves, but sadly... we were wrong.  luckily our neighbors are amazing and they helped us!  yay.   on another note... ben has a new girl toy.  her name is brigette.  she\'s really pretty (and knows it).  aparently her name means \"sex kitten.\"  and she makes weird faces when she\'s talking (about herself).  i will be the FIRST to admit... a portion of the reason i don\'t like her is because i don\'t see ben anymore, and that i miss.  but that is not even close to the majority of the reason i do not like her.  ben and i are best friends, i know him well- when we\'re all out together and he takes me aside to complain about how she\'s being selfish or mean, and then proceeds to sleep with her, kind of annoys me.  he\'s being super fake and thinking only with his penis about this whole situation.  there\'s so much  more, that i won\'t even bother going into... but i really miss my friends.  park ave fest was this past weekend, which is the best festival of the summer- and i didn\'t even hang out with any of them.. brigette always had to be there with ben (she can\'t leave his side), and i just get annoyed when she\'s around so why put myself through that!?  i have other friends, and i hung out with them... but i did miss those guys.  first and foremost ben and i had an amazing friendship... that i\'m not willing to give up, but i feel like ben doesn\'t care about it anymore.  i feel like having a hot girl to fuck every night is much more important than real friendship.  i don\'t like being shit on...   '),(21008,3380244,'02,August,2004','2004-08-02 00:00:00','    i slept home last night... alone, in my bed for the first time in 3 weeks.  it felt amazing.   '),(21009,3380244,'02,August,2004','2004-08-02 00:00:00','    i often find myself doing things that aren\'t mentally healthy for me.  this is recent.  i feel like i\'m falling apart.  i feel as if i started spinning months ago, and i haven\'t stopped.  things just keep happening, and i just keep going along with them, knowing full well that i shouldn\'t.  i\'ve lost my common sense, i\'ve lost my brains, i\'ve lost all sense of organization in my life.  i\'m scattered....   '),(21010,4105180,'07,August,2004','2004-08-07 00:00:00','   Life is confusing. Full stop. Really.      Sometimes there is a longing for someone to love, someone special, whom you love wholeheartedly, and who returns your love just as passionately. Though I have the love of family and friends, sometimes, even when I spend time with them, I feel that something, someone is missing. Sometimes it\'s so strong, it becomes a pain within, and I feel as though the whole world is slowly pressing me down, squashing me flat. It\'s an agony.      \"Some say Love, it is a river   That drowns the tender reed.   Some say Love, it is a razor   That leaves your soul to bleed.   Some say Love, it is a hunger   An endless aching need.   I say Love, it is a flower   And you its only seed.\"      Sorry, I don\'t agree with that, or at least, it doesn\'t seem like that for me, yet. To me, Love is still yet a hunger, the endless aching need. I\'ve watched him for three years, but he\'s so much older. Yet, his awkward, gentle nature allures me so, with his love for God, his humbleness, his humour, it is all so endearing. But he is a Uni student, soon to graduate, and who am I? Merely a fifteen going on sixteen with foolish dreams.      I hate rambling, but tonight, I just felt so empty, without any reason. I used to like another guy, more to my age, but I\'ve realised that though he eclipsed him for a span of time, there is nothing really nice about him, merely charm. What shall I do? Wait and see. The pain is back again, and I wish that I had not begun to feel like this so fast. Life seemed so much happier in the past. I sometimes wish it could return to that. But I know it can\'t, I\'ve changed, and it is irreversible.      God help me, I need Love so.  '),(21011,4105180,'06,August,2004','2004-08-06 00:00:00','   Wow, today was traumatic man... imagine over a thousand of your schoolmates creating a riot while you\'re trying to keep order, while at the same time, trying to get them to sing along with you... yep, I\'m never gonna do community singing on national day again... Oo might actually get her silver soon... must remember to give her STB... yesterday she convinced some daft maris stella guy that she was in interact club, and they do CIP by going online to chat... hilarious man... Went to her block today, spotty is sooooo naughty! He keeps dashing for the house to get Taro!! Poor Taro... Andre came over to practice Sunday\'s worship songs with dad in his army uniform... Had a new teacher in CAT. Am bored, which is why I took the trouble to blog this stuff... All crap really.  '),(21012,4105180,'02,August,2004','2004-08-02 00:00:00','   To all who dare attempt to view and comprehend these words of mineâ€¦'),(21013,3667102,'22,June,2004','2004-06-22 00:00:00','    Well, yesterday was a bust as far as the points go.  I didn\'t even keep track of them, and I am sure I was a little over.  However, I did not snack at all.  Went to the pool with Liz after work and actually mananged to do 6 laps.  Not one right after the other mind you, but, I did 6.  :)   I will stick with the 6 laps for now, and try to decrease the amount of time between them.  Once I have done that I will add two more laps. '),(21014,3667102,'20,June,2004','2004-06-20 00:00:00','    Well, here we are, it is Fathers Day.  A little depressing for me, but, the nice thing is that Liz gave me a little cash to take Gary out to dinner for celebrating both Father\'s and Mother\'s day.  Tomorrow I am beginning WWatchers again.  I think I have plenty of wise choices on hand, and I hope to get our pool passes tomorrow during lunch.  I have also started de-cluttering the house.  I decided that just because DH doesn\'t care about living in a pig stye, does not mean that I have to.  I will de-clutter and clean MY HALF of the rooms as I do them.  Lets see if that gets him moving. '),(21015,3667102,'19,June,2004','2004-06-19 00:00:00','    Hopefully this will be a restfull weekend.  So far, so good.  Found out today that Michael made All-stars, so, he and Gary will have practice everyday this week.  This will actually make it a little easier for me to regulate food etc.  On the food issue.  WWatchers will start on Monday, and I will be off to the store in a little bit to stock up on some things.  Although, I must say, that is the beauty of WWatchers, NOTHING is out of bounds as long as it fits into the points range.  I will however be limiting refined carbs, so, I will need to be sure to get plenty of whole grain bread options for lunches etc. I will be posting my previous days food log each morning.  I am hoping that this will help a little with the accountablility. '),(21016,3667102,'17,June,2004','2004-06-17 00:00:00','    Well, I was going to start this out as a log for my weight loss journey.  I guess that will still be my focus, but, I do plan to use this for other events in my life as well.  Daniel, my older son, just Graduated High School on Tuesday, what a tough time it is going to be, letting him go.  I NEVER thought watching one of my kids leave home to go to school would be so difficult. Oh well, life will go on, without as much humor as before, since Dan is our resident comedian.   Weight Watchers is the name of the game from this point on.  I am really not a fan of flex points, but, I REALLY need to get off this medication I am on.  The only way I will be able to accomplish this is to loose, loose, loose.  I plan on going to get our pool passes tomorrow, so no more excuses about exercise either. '),(21017,3590380,'23,June,2004','2004-06-23 00:00:00','   Horrible time in work today.  Our clientele is becoming increasingly aggressive, verbally and physically and we do not appear to be getting the support we are entitled to. Today, our \"closed to the public\" day, a drunk and violent man caused havoc at reception, a junior member of staff was told to show him to a cubicle and deal with him, even though this man\'s behaviour has caused our team leader to call the police already this week and even though this same man threatened to assault a member of staff with a stick a couple of weeks ago.  The member of staff who had to deal with him today is only (and forgive my ageism here) a youngster on a bare Â£160 a week and none of us is trained to deal with such a situation, even though we are frontline staff. The same grade of staff took an assault last Thursday when a client threw a keyboard at her under the protective (yeah, right) plexi screen in a cubicle... just before he kicked down the keycoded door to the office... Bloody stinks.  My blog, I can swear if I want to.  Senior management remained seated in its freakin\' ivory tower with a nice view of the river... on all occasions.  Bringing the state of mind right down to a relaxed level is another favourite blog of mine,  urlLink Mr. Dunstan Orchard , currently in San Fran and experiencing life to the max, as he always does. Sweetie.  '),(21018,3590380,'20,June,2004','2004-06-20 00:00:00','   Well, the day arrived that my daughter has soooooo been waiting for, the sleepover party to celebrate her 9th birthday!!! Nine little school mates duly arrived, a profusion of pink and glitter and proceeded to snuggle down in their sleeping bags to watch the DVD of  urlLink Freaky Friday  and stuff their little heads with as much garlic bread and pizza as I could cook.  Boy can those kids eat!! While all this was going on I was also tripping back and fore to the field behind our house to check up on no. 1 son Chris as he was camping out with some mates for the second night on the trot, again fuelling more than my own kids with various munchies!! Two of the little girls didn\'t seem to get on with each other too well but as I am very fond of them both (Lizzie and Jessie) it was relatively easy to keep the peace. I fell asleep with my youngest on my chest (no, not breastfeeding! She\'s six!!! eeewww!!) and woke at around 12.30 with an asthma attack, nothing major, but a klingon remnant of a rotten cold!! My living room now looks like a balloon cemetery and I just want to drink tea. Thankfully my Mum was a huuuge help pre-party, cleaning my kitchen and pegging washing out etc, feel a bit crap that I didn\'t get to spend much time with her but we are going to Bridgend for a friend\'s 60th party next week so will catch up properly then! Catch up later xxx  '),(21019,3590380,'09,June,2004','2004-06-09 00:00:00','   Having had a major 24 hours of shouting and tears from myself and my 2 eldest kids and my husband, I decided I needed to get away, but, in my insolvent circumstances, this is the cheapest, most appropriate getaway I could come up with at short notice!  I\'ve been reading a couple of blogs for a few months now and have become very fond of them... when I have learned how to link you to them, I\'ll put the addresses up on here, ok?... Oo... here goes... may have just learned this one...   urlLink Sian\'s blog  '),(21020,3590380,'19,July,2004','2004-07-19 00:00:00','   Dave was duly collected by two chaps from the Joint Forces Defence Animal Centre at Melton Mowbray, where they also look after \"all the Queen\'s horses\" etc.&nbsp; They deemed him a potential asset as a Search Dog, and said that people would be fighting over him as he looks so good.&nbsp; They have asked that we send a disposable camera so that they can keep us informed of his progress during training and have invited us to ring any time.&nbsp; A good outcome in the long run, but Chris, my eldest, is very down as a result at the moment, which is being exorcised by shouting and unpleasant behaviour.&nbsp; Can\'t say I blame him, but it is starting to pee me off now... I have a glazier coming this afternoon to quote for the window he smashed. &nbsp; Went to the doc this morning and he has given me a note for another fortnight.&nbsp; I expected to feel better by now, but am pleased that my reactions to situations are much calmer, less highly strung.&nbsp; All good stuff. &nbsp; Haven\'t heard anything about the job, so I presume I have not been shortlisted. That\'s the way it goes, so, their loss! &nbsp; Hah! Just found out how to change the time on these things!! There ya go! Am not midnight blogging any more!! &nbsp; xxx &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;  '),(21021,3590380,'12,July,2004','2004-07-12 00:00:00','   Me and Imodium, we\'re like  that , (crosses fingers and thrusts them in your face).    Well, Friday got better (tangible sarcasm here) when I drove to the other side of town to take Jasmine\'s suitcase over to Holly\'s house for her to stay the night there.  I stopped for half an hour with Holly\'s mum, Jackie, who is my nemesis! Always sweet, calm, mother-earthly, studying to be a children\'s nurse and she has five-yes-five kids! I went through a packet of paper tissues in that half hour, not through crying (I really AM a lot better now!) but by soaking them with , errmmm, nose stuff. Sorry.  I left Jackie\'s and was just turning into my street when the mobile rang, Mrs Gardner asking me to pick up Jasmine as she was unwell.  1...2...3...4...5...6...etc. Got to 10, took a deep breath and travelled almost all the way back to Jackie\'s to school to pick up poorly child.  Not poorly, doesn\'t like discos and there was the end of term party in full swing with veeerrry loud music.  Put Jas in the car \"I feel better now, can I go back in?\"  ...7...8...9...10!  I put my foot down with a firm hand and said \"No, we are going home and Daddy can take you to Holly\'s later on\".  We walked Dave (our soon-to-be-enlisted labrador, therein lies a whole new blog!) when we got home and made our way to the bus stop. Thank you Lord for this stress, this cold, this bad chest (I put my back out coughing!) and you obviously know I can cope with more \'cause you choose NOW, when my mother is an hour away from coming up to blow out her birthday candles on the Groovy Chick cake, NOW, to give me a  migraine  too!!! Woohoo!!! (Please be aware, I\'m not moaning, I just find this incredibly funny!!!)  Never mind!! I pack off to bed, having distributed the kids around various parts of the Cynon Valley and fend off numb legs, arms, mouth, blindness, nausea etc for an amazing two hours (these things used to last 12hours+ when I was a kid, so I am a professional migrainer now, shaving minutes off my time with each episode!).  The headache never appears, having popped a naughty 3 nurofen.   This said, I am, in all honesty, feeling positive and I have more clarity now about the direction I want to take, as a Mum, wife and colleague.  I just need this week to start getting some stuff organised on paper and I think progress will be made.  Two major decisions have been made, the main one is about our dog, Dave.  He is a delightful 2 and a half year old black labrador, fit and lean and bored.  He can leap the 6ft fence around our (large) garden and will not come back from the field behind us when called, preferring to obsess over an old stone, which CANNOT be taken from him, he is so possessive.  To cut a long story short, the Forces Dog Procurement Unit will hopefully be coming to assess his potential as a drugs and explos dog for the Army or the RAF, a good outcome for us and the dog.  The second decision may not sound much, but I have gorn and gort meself a credit card, too much month and not enough money means that new school uniform will have to bought this way this year! 0% interest for 9 months so all will be well, and no, I won\'t fritter!  Oh, and I have applied for a newly funded job in my department, I don\'t think I\'ll get it, I think they will appoint externally, but at least I tried! I have experience in all fields the job requires and I know I am capable, but sometimes these appointments are political, so who knows!   xxx  '),(21022,3590380,'09,July,2004','2004-07-09 00:00:00','   Thursday night is always a long haul in our house and last night was the longest!  It was parents\' evening at the primary school, so I duly waited to speak to Miss Morgan (Jasmine\'s) teacher and was delighted to read Jas\'s work whilst waiting.  How well this young lady has brought Jasmine on this year! She came to Beth\'s class frustrated at not understanding the building blocks of reading, guessing at words and getting upset, not able to grasp the base words of \"the\", \"and\", etc. well, she has in this time completed the reading course the school uses and is now on \"librarys\" as they say, I think she is now in the top 3 in the class.  Beth said \"I don\'t really need to see you, I just wish I had a class full of Jasmines\"  *beam*.  Sophie had a similiarly pleasant report, she has a particularly good teacher this year, but, like her two brothers before her, Sophie has felt a personality clash with this teacher.  Her work is good, but I wish this teacher would be a little more encoraging in her comments, rather than looking to the mistakes so often.  I hate to see red x\'s on a child\'s work with no explanation? Then after plying the two scholars with a Happy Meal each, a quick phone call en route home \"tell Nick to go wash ready for Confirmation Class\" in time for his 6 o\'clock class.  I\'m sure God won\'t really mind a grubby 12 year old but makes me feel better in a control freak way if I present a sparkling child before the Lord at least once a week! Next stop, Athletics club for Sophie, a 2 hour stint at the Sports Centre, having picked Nicky up from Church. Too tired to go home at this point, I call in on a friend with a bag of chips for Nick, pouring out my, by now considerable, woes. She takes me to the front of the house at 8.45, to wave me off, to pick up Sophie, and BUGGER, I drop my car keys down a drain. Trying not to panic, I realise that THE ONLY HOUSE KEY is now in the depths of the drain.  Gareth is in Torquay, an overnighter after interviewing somebody, and took both sets of house keys with him by accident. One bent meat skewer, a frantic call to another friend to pick Sophie up and some deep breathing later, I retrieve the keys, rinse off the muck and am delighted to find that the remote still works on them too ( By now, however, I am clinically insane)! I don\'t know how, but I get home at 9.45, fend off tantrum from Jasmine because there is no Ready Brek in the house and get the girls to bed, SugarPuffed, at around 10.00.  Surprise, Surprise, throughout all this I have developed ANOTHER frickin\' cold. I am a virus magnet.  This morning, I get shouted at at 5.30 by no. 1 son, Chris for waking him to \"shadow\" a paper round before doing it properly next week, then shouted at again at 7.00 for  not  waking him as he has overslept. Teenage logic. I have just had a phone call from him at school to bring down his PE kit as he left it behind.  Am very tempted to put in my frilliest bra and knickers just for spite.   On a high note, the Prozac is reducing my appetite, making me nauseous and giving me the shits. Should be a good weight loss this week then. Swapsies anyone?  xxx  '),(21023,3590380,'06,July,2004','2004-07-06 00:00:00','   ... and praying there is someone there to catch you.  Well, something had to give and thanks to my superb doctor, I now have 2 weeks off work.  \"just go home and think, do a bit of tidying, a bit of ironing\".  May sound like the wrong thing to say, but this chap knows me too well!! I didn\'t realise how crappy I was feeling in toto until I got home and more or less collapsed onto the floor after seeing the doc.  I think it was out of relief more than illness.  I feel terribly guilty that I have left colleagues with my work, part of me hopes it is still on my desk when I go back, but from what the doctor said, I think he wants me off work for longer than a fortnight. So be it.  I have had various forms of depression, teenage stuff, postnatal etc for over 20 years and I know when I am on the way down.  Unfortunately it took longer than usual for me to realise it this time and I have bridges to build with the kids, especially the oldest two.  I\'m the grownup, I should be able to deal with their emotions, they shouldn\'t have to deal with mine. Work has been a major player in this decline for the last couple of weeks though.  Not good to be crying in the car park in the morning, summoning up the courage to face those colleagues (I really am very fond of them, perhaps I care too much?) and try and get through another day of exasperation, underorganisation, lack of space, lack of guidance, lack of a trustworthy bloody computer system...  Anyway.  God Bless Prozac and all who fail in her.   xxx  '),(21024,3917555,'25,July,2004','2004-07-25 00:00:00','    (This title is gay..but hmm can u think of nething better?) Ya I kno u can, shut up...  I cant WAIT to try out for FH... and I hope i make it cause i just gotta new $$$1oo$$$ dollar stick! hehe And in case you didn\'t get the message: new sn :: FHXCgirl2009, but i\'m betting all of you got it...( I think i got the point across now didn\'t I?) Well, Maine was as usual not much to do weekend...ugh...and of course WE HAVE 2 go every SINGLE weekend!! Because, even if the weather is bad \"it\'s till \'relaxing\' \" Quote on quote mom. I spent my nights swimming in the pool and listening to the band, and listening to the neighbor\'s POLICE radio...its reallllyyy cool cause u can listen to whats going on around the town, on the ocean, and in the airplanes, like they have the Coastguard, the Police, the Ambulances, firetrucks, and the army plane things so its cool cause u could hear like who was hurt and accidents and such! Its really fun, and I want one really bad, but its like $$$200$$$...YiKes!  &nbsp; Sooo we were driving home and this traffic was sllow and it was just on a side road n we were like wat the f? There was a frikin CLYDSEDALE running along side of the road with a guy and a girl and i was like why the heck would that be happening? But me n my mom were like IDK! haha  &nbsp; ...When you smile, I melt inside...  &nbsp; -::-Sam-::- '),(21025,3917555,'22,July,2004','2004-07-22 00:00:00','    Well I started off my day with an Orthodontist\'s appt...great! :/ I have to wait another 3 months to get them off...(thats IF I wear my elastics right) hmm...tehe! Moving on to the next activity, Field Hockey camp, man tomorrow is the last day, and im really sad! This has been a blast this whole week!! We just worked on shuttles and weaves and stuff, then watched the bigger kids do a scrimmage and basically sat in the shade and talked about the game in general \'cause it was unbelievably  hot ... I thought I was gonna die!! Then the Rimol\'s dropped me off @ home..(felt good!) Then I ended up going over Jessi\'s house...we ate lunch and I watched her blow a kiss to her Orlando Bloom poster before we left for the pool..LoL! The pool was soo fun theres a big curly waterslide and a high diving board and then they dropped me off here! Jessi let me borrow the JoJo CD, so I loaded the whole things onto iTunes and now I just hhaavvee to buy it! hehe Happy Birthday mommy!!!  Mwaa! :-*  &nbsp; .:::*sam*:::.  &nbsp;  ...he told me he loved me, he called me his breezy...   ~JoJo '),(21026,3917555,'21,July,2004','2004-07-21 00:00:00','    I didnt blog yesterday...why? \'Cause i was being a lazy bum! (te he) Right now im just chattin to erica,&nbsp;lauren,&nbsp;marci&nbsp;n katie...look @ this crazy girl!     xohoney369ox:   dude...my dog ran down the stairs and just randomly licked my leg then ran back up again   HAHAH oh katie ily soo much! N erica...man i hope u feel betta don\'t worry about it! Marci-we love Otie!!   Neways...i just got done shopping online at mark..heres the link!  urlLink http://www.meetmark.com   ^some cooli makeup there babes!^ My mommy said that it might be easier to just get a catalog instead of ordering online, so im just gonna listen to her! teehee. Well, this morning i went to take the puppy dog to the groomers then i went to the doctors to get 1 shot out of the 3 i have to get! eekkk well theyre not really that bad so i dont kno why im so woorked up but hey haha! Then I went to Field Hockey camp and it felt like friggin 100 Â° out cause we were running and stuff but it was still a blast! My big sis and her oh-so-cool friend Tamra left this morning, yesterday was sooo fun because we went to like all the stores (abercrombie especially) trying on all the 100 dollar skirts that me n my sis wish we could afford to buy (tamra could, she just slapped that dough right out on the counter haha) but ya it was fun! I got lots of shirts from Aeropostale and some stuff from Pacific Sunwear (me and Katie\'s fave store!!!) haha ya right i like all of them but hehe w/e...That was basically my day so right now I\'m just ChiLLen @ da HoLidAe In...kidding! :)&nbsp;  Tomorrows my mom\'s birthday!!!!!!!!  OMG i gotta wrap her present! hehe    ...put your faith in what you most believe in...     &nbsp;   ~*Sam*~ '),(21027,3917555,'20,July,2004','2004-07-20 00:00:00','    Survey ...from  Erica...which came from Rachel\'s blog!  :D     What is your full name:: Samantha Jane Dargie    Spell your first name backwards:: ahtnamas (odd huh? I think its like greek mythology-ish! haha)    Date of birth:: March 21st, 1991    Male or female:: female    Astrological sign:: Aries    Nicknames:: Sam, Sami, Spam,( 2 scuba haha!), Mantha...w/e floats ur boat    Occupation:: uh, school, thank god its summer!    Height:: 5\'1\"Hair color:: Light brown w/ blonde highlights    Eye color:: Brown    Where were you born:: Hanover, New Hampster...haha    Where do you reside now:: Londonderry, NH    Age:: 13    Screen names::CCsami91...i have more but i dont use them    What does your screen name stand for:: Well CC stands for cross country, my sport, and sami is obviously duh my name, and 91 is the birth year.    Pets:: dog-Ivy    Number of candles you blew out on your last birthday cake:: 13    Piercings:: Ears but i want my belly button done!    Tattoo\'s:: None    Shoe size:: 8-8 1/2    Righty or lefty:: righty    Wearing:: HAHA im still wearing my shin guards from field hockey, my aeropostale shirt and baby blue terry cloth shorts    Hearing:: Crying Shoulder-GooGoo Dolls- love that song!    Feeling:: YAY! My sis is coming!!      Eating/drinking:: I just had a turkey, lettuce, and cheese sandwich in pita bread with carrots in ranch dip w/ ranch dip! Yum :)  Favorites    &nbsp;Color:: Light pink, baby blue, yellow...    Food:: err...i dunno...GRILLED CHICKEN!    Song:: Crying Shoulder-GooGoo Dolls    Show:: FULL HOUSE!! (gotta agree with Erica here!)    School subject:: English, i\'m so good at it;)    Animal:: horse, pig, dog, giraffe    Radio station:: 94.5 or KISS 108    Movie:: POTC, Mean Girls, Seabiscuit, Finding Nemo    Pair of shoes:: err my white strappy corky bottom- platform sandals    Cartoon:: Spongebob all the way man    Potato chip:: Err i dont really like them but...salt and vineager    Drink:: Water, milk, NETHING from Gloria Jeans or Starbucks, and grapefrut juice    Soda:: eWWWW i hate soda like out of this world hate :P plah    Holiday:: Hello, Christmas...i love the whole wintery, bonfire, skating, SKIING OH MAN YEAH LOON ERICA HAHA snowy happy feeling!    Pizza topping:: cheese, or pepperoni and mushrooms    Jello flavors::: gross    Lunch meat:: turkey    Card Game:: BS and Spit    Video game:: Mario Kart!!    Computer game:: SIMS    Cereal:: Lucky Charms    Dessert:: Either strawberry shortcake or cheesecake (there goes my thighs haha)    Clothing store:: Weathervane :(, Aeropostale, AE    Carnival game/ride:: Hmmm...    Candy bar:: umm...Snickers    Magazine:: GL, Seventeen, YM    Salad dressing:: caesar and french    Thing to do on the weekend::&nbsp; friends, mall movies, beach, yeah alot of stuff to do!    Hot drink:: Coffee and peppermint tea    Season:: WINTER ALL THE WAY DUDE    Sport to watch:: Baseball i guess...    loser/wannabe:: Ok then...    Doughnuts/bagels:: Bagels    Day/night:: Day    Heaven/hell:: Heaven    Make love/have sex:: uhhh, i dont really care    Coffee/tea:: Both..mygod how can u choose    Hamburgers/hotdogs:: I try not to eat that stuff...just chicken is the only meat    Rap/rock:: umm i like both    Britney/Christina:: Neither    Swiss cheese/american cheese:: american    Real World/Road Rules:: Real World    Backstreet Boys/*Nsync:: neither!!!    Silver/gold:: silver    Nike/Adidas:: Adidas    McDonalds/Taco Bell:: Can u say NASTY?    Sweet/sour:: sweet    Hot/cold:: WARM! muhaha    Winter/summer:: Winter!!    WinterSpring/fall:: Fall    Operas/plays:: plays    Read/watch tv:: Both    Cd\'s/tapes:: CD\'s    Dvd\'s/vhs:: DVD    Old/new:: Depends..err...idk    Shorts/skirts:: skirts    Pink/red:: PiNk    Colored pictures/black and white photos:: colored... Unless the black and white ones were taken by a good photographer    Meat/vegetables:: I would say vegetables    Mexican food/chinese food:: Mexican    Commercials/infomercials:: Both suck    Bikinis/one piece bathing suits:: Bikinis...ohhh yes    Sandals/tennis shoes:: both.    Dogs/cats:: Dogs, but kittens are cute    Water/land:: water...come on im on the swim team...    Sugar/spice:: sugar    Black/white:: not black, black is goth! WHITE...go white    Chicken/beef:: chicken    Colored christmas lights/regular white christmas lights:: white    Cars/trucks:: cars or trucks depends    Austin Powers/James Bond:: Austin...I DETEST JAMES BOND EWWWW    Popcorn/pretzels:: popcorn    Picture frames/photo albums:: both      Pens/pencils:: mechanical colory pencils!     Randomness    Name three things that are next to your computer:: A rose, my glitter eyeshadow, and my portable dvd player    Do you have any hidden talents:: hidden? what? I dunno    Do you wish MTV would play music videos:: Dont really care    Do you play any sports:: Field Hockey, XC, wanna ride horses...:( help me    Do you drive:: No:(    What is your dream car:: Silver lexus convertable...but i cant really dream bout it cause i already have my car for me when i turn 16...our RED MUSTANG CONVERTABLE...Yess all mine baby    Would you ever sky dive:: R U KIDDING ME NOOOO    &nbsp;How many rooms do you have in your house:: uhh...11 im pretty sure    Are you afraid of roller coasters:: hahahahaha!!! no.    Do you believe in God:: yes...but dont go to church    Do you believe there is a heaven:: yes..    Do you believe there is a hell:: umm im not really sure..nah    &nbsp;Do you own a pool table:: Nahh    Do you have a pool:: NO BOOHOO    Do you have a dishwasher in your kitchen:: Yeah, who doesnt these days?    Do you like chocolate:: YES! haha    Who/what is on your 2004 calendar:: Spongebob Squarepants    How many U.S. states have you been to:: uhh...11! there we go    Who is the last person to e-mail you:: MeLaNie..grrr    Who is the last person to IM you:: Kimberly Buttafuoco!    Are you a deep sleeper::Yep    Are you a good story teller:: Idk    Do you like to burn candles or incense:: both! hehe but my room is full of candles    Do you have your own credit card:: WHO NEEDS ONE?    Let\'s say you win the lotto. What do you do with all that money?:: Clothes, cars, and a MAC laptop...a huge beach house and i would redo my room all cooli!    Do you like your drivers license:: STOP INTAGINIZING ME! I DO NOT DRIVE!    Do you tan easily:: Sure    Worst feeling in the world?:: When you r really sad and ur heart sinks    Best feeling in the world:: Being loved    Is the glass half empty or half full:: Full    Last thing you downloaded:: Piolet    Do you catch yourself using online terms in your real life?:: yeah haha    What do you think people think of you:: don\'t know, maybe I dont wanna?    Do you need therapy:: Nooooo    Do you take medication for a chemical imbalance:: nope    Do you love your bf/gf:: I dont have one, but right now I dont care!!!! :):):)     &nbsp; '),(21028,3917555,'19,July,2004','2004-07-19 00:00:00','    .::Field hockey camp::.    Monday thru Friday...8:00 AM to 11:00 AM!!! yay! :)    I loooove field hockey camp...jeez Erica u should have gone...ur goin out for the team right? AnYwaYz...(whoops, anyway...(anyways isnt proper english says Melanie!) Yea so at Field Hockey camp its soo fun right and the coaches r hilarous omg they make u laugh all the time and it throws off my concentration! hahaha so yeah after camp i basically just hung around the house cleaning and stuff since my sister is coming down tomorrow from Vermont! Hip hip hoo~ray! LOL I\'m goin to the movies tonight with Kaily...I havnt seen this biatch for ages man! Aww baby! ;-) This was a short one...Now I have to go shower I have got my work all done and nowz tonight for a party! Yay! Catch y\'all lataz    Peace.     &nbsp;   ((~*Sam*~))     &nbsp;    ...And through it all he offers me protection, alot of love and affection, whether im right or wrong, im loving angels instead... '),(21029,3917555,'18,July,2004','2004-07-18 00:00:00','    Well, i havn\'t written in hmm i think its 3 days...well thats cause i was in boring Maine and blah blah blah blah blah...good news...KAILEENS HOME!  yay! &nbsp; Shes been gone like all summer and i miss her! Well here are some highlights of the Maine trip!        Shopping. (I bought a sticker thingey that was 4 FREAKIN DOLLARS (its erica\'s fish but its pink instead so ha 2 u erica! A necklace, and the book Seabiscuit. (its really good, if u get a chance you should read it.)    Old Orchard (always fun, cotton candy, rides, the pier.)    Riding the trolley!!!! :-)    Going out to dinner @ all the expensive seafood places      What else... THE BEACH!        And I think that\'s about it! I just got back from the neighbor\'s and I think I\'m gonna go on a bike ride with Lindsay a little later... Im starving and i need some food!      ...Everything went from wrong to right, and the stars came out and filled up the sky, the music you were playin\' really blew my mind, it was love @ first sight...        &nbsp;   >>** Sam ** '),(21030,3917555,'16,July,2004','2004-07-16 00:00:00','    Well this is gonna be long one!!    I didnt\' blog yesterday...(sorry!) I was over Kait\'s the whole afternoon  and  sleptover...but it was a nice lil catch-up cause i havent seen her in a loong time! So I got there and we practiced field hockey for about an hour, and boy oh boy were we  WHACKING  that freakin ball! Shes a great player, I\'m sure she\'ll make the team.&nbsp; She\'s going into sixth grade, so the team will have a good sixth grader for them!&nbsp; Then, we bounced on the trampoline and i took off my sox, and i just realized that i left them there haha! NewaYz... We played a trick on her mom, we put this lil rubber spider on her toilet seat! and in the process, the spider fell in the toilet and it like sank...i almost peed my pants!!!!!!!!!!!!!! :-) we used the toilet scrubber to get it out...god it was&nbsp; hilarous...so we were in her room and her mom opended the toilet seat and she almost sat on it and screamed oh u all should have been there it was hilarous!&nbsp; Then we went to her lil bros baseball game @ LAFA...n i saw Cristy there!&nbsp;&nbsp; iwass so happy to see u hun! Well yep that\'s the main highlights of yesterday!     &nbsp;   Today...well not much to say..I woke up @ the RIMOL residence at 9:45..and we played GameCube and made brownies...then they dropped me off @ my house and here I AM! (I fixed the time on the postings, ur welcome!) and I STILL Need help from Katie or Jenna, who ever...thanks erica for teaching me some more...but i still need a barn! cuz SBF if fancy pantsy...own a horse! IDONT THINIK SO HAHAHA     &nbsp;    ...look for the girl with a broken smile, ask her if she wants to stay awhile...      &nbsp; '),(21031,3917555,'14,July,2004','2004-07-14 00:00:00','    Ya know what i really don\'t get? People sign on at 1:30 in the afternoon and say ITS BREAKFAST TIME! I mean come on people! Do you really have to sleep in that late? Look @ Katie...she just woke up right now (1:30) and shes eating breakfast, when is she gonna eat lunch? Dinner time? haha that stuff just makes me laugh cause why can\'t you just wake up at a normal time? But anyway, I would like to thank Amber for helping me with the whole Britni thing-besides i started that from standing up for you!!!  I\'m really bored right now and yay! The  -::-sun-::-  is coming out!!!! Today is just another boring day of summer...and by the way WHEN IN THE HECK R WE GETTING OUR TEAM LETTERS? I really don\'t wanna wait nelonger, especially with the whole clothes/school supply bac-2-school shopping to do!! YAY! And I need some help from neone who rides *mm JENNA, KATIE?* (probably not all u SBF ppl *cough kaitlyn and erica* cuz u most likely don\'t know much about Londonderry barns! BAHAHHAAHHA muahahah but ilu guys too!) Hoped u liked all the quizzes from the  PrEviOus  entery...those r really fun 2 do when your bored...but im not so thrilled i came out as DORY! haha jk il dory but man what is that telling me?? That im a dumbo like her? haha shes really cool but i was hoping to be a lil smarter than that!!!!!!!!!!!!!!!! But w/e what do those quizzes know? Not much i tell ya!   -::- Sam -::-        If at first you don\'t succed, pick your self up and try again.:-)      Take a seat, take your life, plot it out in black and white       *Something\'s better on the other side* '),(21032,3917555,'13,July,2004','2004-07-13 00:00:00','    Yay! Lots of quizzzing today...the IQ test is soo fun! I have an IQ of 109! hmm i dont really know if thats good or bad! haha if you know, let me know!         urlLink    What Finding Nemo Character are You?    brought to you by  urlLink Quizilla           Raver Bear     urlLink  Which Dysfunctional Care Bear Are You?    brought to you by  urlLink Quizilla         My inner child is six years old!     Look what I can do! I can walk, I can run, I can read! I like to do stuff, and there\'s a whole big world out there to do it in. Just so long as I can take my blankie and my Mommy and my three best friends with me, of course.     urlLink  How Old is Your Inner Child?    brought to you by  urlLink Quizilla         You gotta believe, there\'s a reason we\'re here    -::-  Sam  -::- '),(21033,3917555,'13,July,2004','2004-07-13 00:00:00','    Well hey yea babysitting was great...haha jp!  No really they\'re great kids...most of the time!! haha no really they r! Well heres the skinny on them...Billy is going into first grade..and Maddie is going into Preschool...god i love them to pieces they r soooo adorable! Well so i was making them sandwiches for lunch and Maddie was like brrr im cold and i said why? and shes like the pool is making me cold and im thinking what the heck? the pool is outside far away from us in the kitchen how in the world can the pool be making you cold? I guess kids have theyre reasons! haha oh yeah and then maddie farted like really loud and i started laughing and she goes \" I make funny noises.\" aww baby it was funny i guess u had 2 be there hahaha well i really dont know what else to say except oh yeahhh im dumpin tyler cuz hes a freak and yep i like bein single and i dont even liike him so THERE! haha well catch ya all later on the flip side ( or sould i say *south side* haha if u dont get it read the stuff below haha*)    3      \'Tell yur daddy stop his hatin\'  Wearin braids and jerseys daily  Dont mean i\'m thuggin no\'    Love you all   .::::Sam::::.      faith can be broken, yeah hands can be bound  so open your hearts, and fill up the emptiness '),(21034,3917555,'13,July,2004','2004-07-13 00:00:00','    Hello people.  Once again, I had to make a  NEW  blogger!!! plahhh! I forgot the password, and my email wasn\'t working so I was like screw this S! neways, it\'s the morning and I\'m just about to go babysitting-Joy! ;-) This has been the suckiest summer yet I have to admit, the only things I have done was Canobie, Maine, Softball Camp, and Feild Hockey Camp (coming up soon.) I really really wanted to do FH but I can\'t cause its the same season as XC and lordie I wont miss out on XC!!! (Even tho i sucked last year haha.)  Nothing else to say right now...I will \'blog\' *lol that sounds funni* after I get home!  P.S.-All the times are wrong on the thing on the bottom where it says POSTED AT: this one says like 6 something but its realy 9 something! So if I dont fix it, just remember that its 3 hours  before  the actual time! OK! :-)         \"On the way down, i saw you  and you saved me from my self.  And i won\'t forget the way you loved me.  On the way down i almost fell right through  But i held on to You.\'   .::: Sam :::. '),(21035,3780716,'28,June,2004','2004-06-28 00:00:00','   Hola! Here I am today! Nothing to do and bored to death. I have school right now and that like takes up my day. I have to go to school because I didn\'t pass the Science TAKS...Boo!! Why do I have to be so dumb?! After my mom picked me up, we went to the driving place so I can sign up! Woo-Hoo...one step closer to driving! Well, everything has been so weir for me! I don\'t know what to do with myself! I have been so out of it...I don\'t even have the friends I used to! I have lost so much of my friends and I don\'t even know how I did that! Like I mean, I have lost touch with Krysten, Michelle, and Krista! I feel like I have lost my friends..even though I knew deep down inside Krista and Michelle didn\'t like me! I don\'t know what was up with that! It\'s like I went to a church where I no longer had friends. I felt so left out in everything I did! It was sad! Now, if go to church I end up going in Lytle or when I am with my dad! I miss everyone at church but like I mean, if they missed me too...they would call! I wonder what will happen with my mom and step-dad, Doug, and the whole divorce thing! I can\'t wait until something actually happens! Something need to happens now! I mean like, I get so tired of wondering what is going to happen! I can\'t stand any of my step-sisters! I mean like everything they have done has been so wrong. None of them are even the same anymore! They are all liars. Farrin is a two-faced idiot! Cassidy has a horrible attitude and Brandy is a WITCH! I wanna get away from them all! I wanna leave here. Get away forever! I hate where I am and I wish I could just go away-FAR AWAY!! But anyways- I have been pretty happy when I am able to go spend time with my brother and his gf\'s Family all in Lytle. I mean, it is so small over there, but I love it for some reason! I like being over there and it is just so fun! That\'s probaly where I will be the rest of the summer cause that\'s where I like to be! I wish I could move out of here! I wanna graduate and leave here! I wish I could drive! I wish I had a job! I wish I had money! My life has been screwed over because my stupid and wicked step-sisters! I need to get away! But I don\'t know- I have to stop whining and try to face what\'s happening now! I have to go for right now anyways! Bye and love you!  '),(21036,3780716,'18,July,2004','2004-07-18 00:00:00','   Hey...I am SOOO tired...I have to wake up so early in the morning and I am gonna have so much trouble doing that! I am gonna go to camp with my youth group at church. I will be gone until Saturday! I hope I have myself some fun! I really do because I didn\'t even wanna go but I told myslef the only way I would go was if Krysten went...and she is going so now I am going. I hope there is no DRAMA even though there seems to always be some! Today was a pretty boring day. Nothing too exciting. I was tired this morning because my dad had a gig last night. It was really good. I always enjoy going&nbsp;to hear him sing though. I love it! So yah-I wen tot church, then ate, then came home and them my Mesha calls (James\' gf) to see if my mom and I wanted to join them for dinner at Olive Garden. So we went and the food was GOOD! I like olive Garden! I got full fast though...I was so excited cause I got to eat shrimp Mmm Mmm Mmm :o) So yuppers-we came home and I had to pack my clothes for camp. I am GONNA BE SLEEPY!!! I am gonna miss everyone though! Uh huh...k, now I have nothing to say so I will go now...luv you mucho xoxo...  '),(21037,3780716,'05,July,2004','2004-07-05 00:00:00','   Holla-How are you guys doing? Well, I am ok and doing better than earlier. I was miserable last night...I cried myself to sleep :o( But, I woke up late and I chilled. Naomi and Paul came over and my mom BBQed for us. We just visited which was nice! We were just talking and talking. Later they left and I was once again bored. Today went by really quick though. I was actually planning on going out today but no one was around to. So, I just stayed here and did nothing. My mom was all crying and was saying that I don\'t love her or something like that. I didn\'t know what to do. So her and I ended up having this long conversation. She is better now. I was all depressed and emotional still, so I let some tears loose. You are probaly thinking I am a cry baby...but only sometimes-haha! Really, I am not always like that! But, the good thing is that I am doing a lot better. I finally got the picture posted. I was having trouble at first, but now I have it under control. Man-I need to go to the Valley soon. I haven\'t seen my family in a while! I miss them too! I also miss BOBBY! I haven\'t seen him in a long time! The most coolest thing is that we met at church camp. We met the summer of my 8th grade year! It\'s been a while and we are still good friends! That\'s what you call a friend. He is so sweet too! HI BOBBY :o) He is going off to college soon though. So, I need to get over there so I can see him! Man-I have no idea what I will be doing tomorrow...hopefully Lucas and Naomi can come over. I wanna do something with them-but more than likely they won\'t come over. So yah-I am gonna end up going to bed early cause Steve went to bed and I have no one to talk to...o well, I guess I need the rest anyways! I really hope I can get better...I don\'t like to be sad! I WILL :o) So yuppers...I guess I should head to bed in a lil bit...night!  Love you-M U A H xoxo  '),(21038,3780716,'04,July,2004','2004-07-04 00:00:00','   Hi-well, today sucked badly. I mean, I went to church and I got to see everybody and all. We went out to lunch with some family friends! We went to go eat at Souper Salad. It was ok. I was not feeling good so the whole time I was all out of it because I wasn\'t feeling good. Then my mom and I went to Seaworld to go meet Naomi! We got there at like 3. It was HOT! But yah-we went to go see some shows or whatever. I really liked the Fireworks...they were pretty! People were shoving a lot and I wasn\'t feeling good, so I was REALLY annoyed. I felt like the 3rd wheel with my mom and Naomi. They weren\'t even talking to me, so basically I was there all alone! I wish I would have had someone there with me. I hated it so much! I wasn\'t feeling good and I was being ignored! Then, I decide I wanna go to Lytle cause I don\'t wanna be HERE...and like there were some difficulties! There is this guy that is married to Lucas\' cousin. He is 20 and like supposedly he was looking at me a couple of times or whatever. I wasn\'t encouraging the situation, in fact-I was clueless until Mesha told James and he talked to me. From then on, I became uncomfortable. Well, long story short, he called me and James found out. He had just called to see how I was and it was wrong cause he is married but my brother doesn\'t like him and he got MAD. Now, when I get home from having a bad day, my brother calls to say to me that he thinks it is best that I just don\'t show my face in Lytle for awhile until they figure out what to do with him...and they expect me to be ok with it! YAH RIGHT!!! I am so emotional right now and I just started to cry :o( I was so sad! I feel like I have no where. I don\'t wanna be here and I can\'t be there. But I don\'t know what to do...it is gonna be so uncomfortable when I see him now. They said they were gonna talk to him and then that will put me on the spot. I just wish this wasn\'t happeneing! UGH-I hate it! So now here I am...sad! I am so tired but I don\'t feel good and when I don\'t feel good, I don\'t sleep! SUCKS!! Anyways-hope whoever is reading this had a great 4th! But yah-sorry this entry is so extremely boring. I had to say what was on my mind. And I think I may be in Lytle for a couple days...so if you don\'t hear from me, that\'s where I am! Well, I guess I am gonna go for right now! Bye guys and I love you! MUAH xoxo...  '),(21039,3780716,'03,July,2004','2004-07-03 00:00:00','   Hey...today I did not feel good at all! I ended up going to bed at like 5:30 because I was talking to Steve. We were like taking turns telling jokes or whatever. It was funny! I woke up at 11 because I was not feeling good. I had some bad pains. I didn\'t get to go to the funeral either because I could not walk and I felt so sick! I didn\'t have a lot of energy. I was so tired today that I ended up staying home all day! I watched TV and I played alot of games! I am tired of sitting at home. I am actually going to Crossroads tomorrow! It will be weird because I have not been there in SO LONG! I hope it isn\'t too awkward! Tomorrow, my mom and I are gonna go to Lytle and try to see if we can see my brother! I guess then after that, I am gonna go to Sea World and watch the Fireworks! Woo-Hoo! I hope I am feeling ok to go though! I hope I am atleast feeling a lot better! School is so close! I wish it wasn\'t! I REALLY don\'t wanna go back. I am not excited about going back to O\'Connor! I miss Taft! I wish I was keeping touch with all my friends! I think I have lost alot of them. YEAH-My hair has a grown a little. Not a lot, but hey-I am getting there. I am so tempted to cut it! I hate my hair. It\'s like I dont have a stlye! UGH...suck a nut! I have not been excercising like I had been wanting to. I get so lazy! I hope I can get fit soon. I don\'t wanna be all fat! So yuppers-I wanna hear \"Roses\". I love that song. I don\'t even know why! All I know is that I really like it! Everytime it comes on, I start singing and dancing! YEAH :o) Okay then, I probaly need to get to bed early tonight since I have to wake up Early. Nighty Night...Love you...M U A H xoxo  '),(21040,3780716,'02,July,2004','2004-07-02 00:00:00','   Holla! Well, today was a pretty long day! Today was the last day at school...thank goodness! Now all I have to do is take the test next week...I am so scared/nervous! But-it will be all good in the end! Anywho-I was so bored today! I went to go eat at Applebee\'s with my mom. It was good! Yummy...FOOD! I got to wear my new shorts today! I was so happy because me and Nessa wore them to school, it was cute! I also got a Frappuccino this morning! Mmm Mmm Good :o) Josuf came by today and visited me! He is so funny! We hung out for a little bit and got to talk because we usually don\'t get to...I played games today-it became a routine caus eI was so bored today! I had nothing to do at all! BORING! I am going to that funeral tomorrow! It is gonna be so sad. I feel so bad for the family. I wanna put my pic in this thing that is Hot or not thing...it is a webiste and supposedly people vote on whether you are pretty or not. I wanna do it because I am curious as to what people will think of me...So, that\'s what I will do for tomorrow..I just hope a lot of people do  not think I am ugly. That will be a discouragement! My brother was mad at me because there is this guy that he doesn\'t like that called  me and it got into a huge deal...I started to cry cause I thought he was mad at me. I hate when he is mad at me though! I also talked to Lucas, he is so funny! He makes me laugh so much and that is always a good thing when a guy can make you laugh a lot! Hehehe :o) So yah...I am talking to this guy Steve that I met through a weird way, but man, for those of you who know about Danny a.k.a the Faggot, I found of some stuff. He is so Stupid and so is his gf! I can\'t believe I ever believed anything he ever said to me. What a loser! But hey-we all learn from our STUPID mistakes! Or atleast I will! Never Again! It is so hard for me to trust poele though and that is something I am gonna have to work on. I just wish people would respect me enough to not lie to me! O well-it happens I guess. I wanna meet a guy that will truly care for me and like me for who I am! I mean, I want a guy that will make me laugh, smile, make me feel important and like, I dont know- I want A COOL GUY! My wish :) Yuppers...HI STEVE-SMILE :o) Thought I would say hi! Hehe...anyways-I am getting tired though. I still have no idea what I am doing on the 4th and I wish I knew! But yuppers, I have to go cause I am probaly being boring again :( Buh bye...Luv u\'s...Muah X O X O  '),(21041,3780716,'01,July,2004','2004-07-01 00:00:00','   What-up?! Here I am bored again in my stupid apartment! With nothing to do again! My days have been so boring! I can\'t stand being here anymore! I can\'t wait until school is over. 1 MORE DAY! Thank goodness! Well, yesterday I went with Nessa to Panda\'s Express and then to Old Navy. I saw the cutest shorts. I had to get them and I did! They are super cute. They aren\'t short...they come up to my knees I think, but when you see them, they are ugly, but I guess it\'s when you put them on they look adorable! I like them! I dont know what I am gonna do for 4th yet! I probaly will go with Naomi! I don\'t want her to go alone to Sea World. I really don\'t wanna be with my mom! I hope I am not stuck with her either! I have no idea of what I am gonna do when this week is over. I don\'t know whether to go back to Lytle or to go with my dad. I still need to decide! But all I know is that I am NOT gonna be here with my mom! Definitely not! I wanna go out and fun right now! I have some cute outfits that I feel like showing off right now! But I can\'t do it when I am under this roof!! Haha! Anyways-I wanna see all my old friends. I miss everyone! Man-I cannot believe that Lindsay\'s mom died last Saturday! I mean like everyone was kinda expecting it, but wow! I mean she is really not here now! It is so sad and I am going to the funeral on Saturday! I wonder how she is taking it! I hate going to funerals because they are just so sad! But, I feel that I need to go! I was close to them and I have known them for a long itme. I think it would only be right if I went! So, yah-I got to talk to Krysten the other day. I hadn\'t talked to her in forever and I was just so scared that she was mad at me! Thank goodness she wasn\'t! I really can\'t stand my step sis either. I mean she is so fake! I am glad I haven\'t seen them. I get mad when I think about them and everything! So yuppers- I feel like going to see \"White Chicks\". I wonder if that was a good movie or not. Man-o-Man...I wanna go shopping! Haha-I always want to though, so it really isn\'t a big deal. I need a job so that I can get money and go get some really cute clothes! It\'s on the to-do list! I wanna learn to play my guitar too! That would be so exciting! Maybe I should trying doing that soon! Okay so yah-I have to go...sorry this may be boring to whoever is reading this...I promise that my life will get interesting...haha! Luv ya MUCHO!   M  x  U  o  A  x  H  o  ');
INSERT INTO `msgs` VALUES (21042,4117248,'25,February,2004','2004-02-25 00:00:00','   So this past week has been a little weird, full of different emotions and feelings, and I guess it\'s pretty impossible to sum them all up into one entry, so I\'ll just stick with the big stuff. On Saturday, I invited Christina to the play and to spend the night at my house. She and Olivia and Nathan and myself, all went to the play, which kind of turned into a bit of a disaster. First, Olivia, as usual acted up like a tard, and made the whole situation stressful. Second, Nathan invited one of his other friends along, who Christina doesn\'t particularly like. At first I thought she was just being silly, but after I met Brittany, I began to think that I was feeling the same way as Christina. She seemed sort of snobby, and not some one I would really want to talk to that much. Then during the play, this whole tention feeling was going on, which made me angry, and in turn ruined the whole play for me. When we got home, I was complaining about it, because I hate it when Christina just sits there and gets angry, and sometimes takes it out on me by calling me stupid or something, and I was taking it out on Olivia, and then Christina came over to spend the night, and there was still some tention. So I told Christina right out, and it bugged me that she would just let Nathan anger her like that. We got to talking, and I actually told her how I felt. I even let her read last Wednesday\'s entry, we almost started to cry, but kept it down to a soft roar. It was a good night. Then on Sunday, Mom adressed Olivia and I about everything that\'s been happening between us in the last month, and how our arguing is really getting on her nerves. That was a good day, because even though we had some arguments, and were crying, we got out a lot of stuff that had been bothering us, and it was helpful. I think that Richard is moving out of my math class, and I\'m so sad about it. I don\'t want him too, even though it only reminds me of the things I can\'t have, I like looking at him, and thinking about what it would be like to have a boyfriend like that. He\'s really cute and really intellegent, and oh, I just want him to like me as much as I like him. I mean, just because I\'m fat, does not mean that I don\'t have a good heart and mind. I just want some one to like me, and if it were some one that I liked back, it would be a plus. I mean, I don\'t want a huge relationship, I just want to know that I\'m good enough for some one to like me the way I feel about some people, like I do Richard. I also like Chase a little. But, not as much as Richard, and I think I only like Chase, that way, because he likes me at least for a best friend. I might have some chance with him. I miss him a lot though. He moved away, and I don\'t ever get to see him anymore, and I really miss having him for a friend to lean on. I mean, I still talk to him online every once in a while, but it\'s different. I\'m going to visit him when I get my licsense. I\'m going to have a party on the 6th of March, and it will be so much fun. It\'s theme is Dr. Seuss, because we share a birthday. It\'s so weird, because it\'s my 16th birthday, and Dr. Seuss\' 100th birthday this year. Two very memerable years to turn. I\'m so excited. I\'ve already made the invitations, and now all I have to do is hand them out, and make sure that it all gets aranged. Well, it\'s time I high tail it, which is a car expression, which I\'ve been thinking about cars for a while, since I\'m turning 16, and I\'m going to be able to drive now. I\'m super excited. '),(21043,4117248,'18,February,2004','2004-02-18 00:00:00','   Today was so full of mixed emotions, and confused frustrations. It started off okay. Can you believe that I got up at 5:40am so that I would be ready to get out the door as soon as my ride came, and I was actually late. I woke up to my alarm clock right away, went into the house, took a shower, went to my room, got dressed, brushed my hair, brushed my teeth, and got my backpack together. I just should not have started resting. I wasn\'t laying down, and I wasn\'t sleepy. Once I\'ve taken my shower, I\'m not sleepy until halfway through Spanish. I was just sitting down, petting Angel, and relaxing, but when the time came to make my lunch, and breakfast, I was rushing, and I went out the door right, and I mean on the second, right as my ride pulled in. That was fine, and so was seminary, school was great. I was happy all throughout school, and all my classes, up until Math. That seems to be the class that I hate know. There are two reasons. One is that Richard is clear across the room, and we never talk, and I miss talking to him because I really like him. He\'s smart, not just smart, but intelligent, he\'s funny, and he\'s not horrible looking, not that that will change how I feel about who I like, it\'s just a plus. It\'s just, I know I can never have him. As much as people will argue, that I\'m so cute, and that there is undoubtedly some one in the school who would date me, I know it\'s not true. First of all, I\'m fat. Just a simple statement; I am fat. Plain and simple. I\'m not horrible looking, I\'m not so bad that no one wants to look in my direction, and I don\'t waddle from side to side when I walk, which is a plus, it\'s just... I\'m pleasantly plump, as Elizabeth Schulte would put it. Also, when I\'m around Richard, I act really stupid. I\'m not putting myself down when I say that, I mean, I literally act stupid. The kind of stupid you might call a blonde, or, I don\'t know, anyone who isn\'t smart. I act like I don\'t know anything, and that I\'m helpless. I just act plain senseless. It\'s my form of flirting, even though I know that it doesn\'t make me look any better than I already do, in fact, it makes me look worst, but, I want some one to like me so bad that I\'m making up different personalities for some one to like. I don\'t even know who \"me\" is. I really don\'t. I know that I enjoy certain things, and that there are certain aspects about me that will never truly change, but who I am day to day, is just what I don\'t know. So, I act stupid around Richard. The second reason why I don\'t like math, is because Elizabeth and Heather are both in that class with me, and instead of them each talking to me, they talk to each other. So, okay, I\'m jealous. I don\'t even think it\'s that, it\'s just, I now know that they would rather spend their time together, then with me. They don\'t even include me at all. That might help a little. It also doesn\'t help that Travis, Joe, Heather, and Elizabeth all sit in this little square, and Tim sits right in the way for me to be included. So know what I do in math, is something I haven\'t done since Eighth grade; I actually do math. Now, if only Richard could see that I\'m actually a little smart. That\'s a regular day in math, but today was worse. I was looking at my grade to see, well, what my grade was, and Heather and Elizabeth just left, when we all have the same class together after that. They just left without a thought or care about if I was coming. Writing this just make me cry because I don\'t even feel like my friends are really my friends. It makes me wonder if anybody would truly, and I mean truly cared if I just died. Heather has said multiple times that she wouldn\'t cry if I died, and Elizabeth is just too high in her league for her to care enough about me, and what I feel, and if I would die. I mean, it\'s ridiculous. I shouldn\'t be feeling this at all. So the next class was Theatre, and I just let Heather and Elizabeth do their own little thing on the opposite ends of the house seats, but of course, because I am known as the happy \'facts of life\' girl, I had to be happy. And I will be truthful, it is really hard for me to stay sad in the theatre. Also, I just am most of the time happy. It\'s become a habit, not that I really am having good happy thoughts, but its my facade, and it is super fixed and is going to be really hard to break if I ever find the need to. So, while we were reading Shakespeare, Heather and Elizabeth were on house right, and I was on house left, which are the two opposite sides of the theatre seats, and they didn\'t even ask me to come sit next to them. I didn\'t actually mind after five minutes of analyzing what we were reading, and learning things about drama. If there is one thing about me, that will always be the same, it is my fascination with dramatics. I have, am, and always will be in love with theatrics. It can make my mood so beautifully happy, that I can\'t help it. Which is why in Microbiology &amp; Genetics, I was so hyper. I was more than hyper, I was acting ADHD! I don\'t know how I went from super depressed to so hyper I had to shut myself up a million and one times, but it happened. I believe it was that the sub in theatre was trying really hard to get into it, and be a part of it, and make it fun and interesting, which is something sometimes even Cromwell doesn\'t do all the time. I mean, she does, but in a different way that the sub was. I really liked his style of teaching. I think the difference was that he was a sub, so he only taught sometimes, and she has had to deal with kids learning theatrics every school day for the past who knows how many years, and she\'s become accustomed to teaching us kids what a kid will grasp in theatrics, rather than teachings us like adults, and learning along with us, like I think the sub tried to do. I don\'t think he knew a thing about drama, but he was learning along with us. I really liked it. So all through Micro/Genetics I was hyper, making Heather hyper, because my frustration towards her just melted away with the anger of it all in drama, and Ashley Fanning became crazy, because she was our partner, and therefore was infected with our contagious hyper-ness, and I spent the entire period wanting to get outside. I was going to walk with Heather, and even Christina was going to walk, but then it was sprinkling, and Christina wanted to ride the bus, and I felt an overwhelming sense of commitment to Christina right then. For the past couple of weeks, month, the last year in general, I feel like we have been growing apart, still being close friends, but we were once much closer, and that is partly do to my lack of interest in life. I won\'t make excuses, but sometimes I just wanted to give up. I haven\'t told anybody that, not even Christina, because, I just don\'t know how to. I don\'t want to commit suicide, but, I just want to stop, give up, breeze through life as a nothing. And that meant leaving my friends behind too. But Christina has been trying for the past couple of weeks to get us back to the state we had been and could be, and I just finally noticed that she is trying to help me out. Make our friendship everything it can be. I\'ll admit, sometimes our views go in opposite direction, and sometimes I just want to beat her up, because we don\'t agree, but I love her to death. Not many people have been able to look past that rough exterior, and see what I have. She is able to be a bully, not as much anymore, but at one point in time, she was, but if you look into her eyes, and read her soul, you can see, I did see, that all she wanted was love. Love from friends, and boyfriends, and family, though she may not be able to admit it to even herself. And I want her to feel that I\'m here for her, that I love her, and that she can talk to me about anything. So I rode the bus, and I thought, this is it, this will be our turn around, and I will be happy again, truly happy, and I will want to keep going, and I will have something stable in my life, even if it is just one friend, because I know that she would do that for me. And then we got on the bus, and she started talking to Krys, a new friend she\'s made, and I wanted to poke my own eyeball out with a pair of scissors, because I don\'t want her to laugh, and have more in common, and have \"those one times\" with anybody but me. And because I haven\'t been there for a while, she\'s found others to have \"those one times\" with, and who can be there for her, because I wasn\'t. It was my epiphany, my wake-up call, the thing that shook me out of my deep sleep, and I knew that I had to tell her, that, I needed her to be my friend, and not have more in common with some one else. So, while we were in the library, I told her jokingly, in a round about way, how I felt. It came out more as I\'m jealous of these new people your hanging out with and hugging, and you can\'t be their friends anymore. I didn\'t know how to tell her, that I love her, and I don\'t want to loose her, and that I just want to her to realize that I have no one right now that I can turn to, that loves me as much as I need. That my father is an alcoholic who doesn\'t even come close to understanding what he puts me through. That my little sister hates being with me, doesn\'t even like me wanting to talk to her, thanks God for her friends, before me at a Thanksgiving dinner. That I have a mother who\'s TV programs always come before me. They when I need to talk to some one, no one is there. That I just need her. It was so hard to even look at her at that moment, because I was ready to burst into tears, because I was trying to tell her everything I was feeling, without making her feel bad because there was nothing for her to feel bad about, and that just coming close to the subject made me really emotional. I need to tell her these things, but I don\'t know how to without hurting someone, even myself. And I don\'t know when to do it either. I don\'t have the time to have her come over and for us to get in the comfortable point where we talk about everything. The last time we did that was after she broke up with Chase, and that was so long ago, I can\'t even remember. We walked home, laughed about things I was so disconnected from because I needed to be emotionally disconnected to be emotionally stable. I went to mutual, which was a lot of fun. We just played some theatre games, and hung out, had a doenut, and I was in even a good enough mood that I played the piano. And then I came home. All I wanted to do was talk to mom about how I was tied down with homework. I just wanted her to listen. But every time I opened my mouth, all that came out in response from her was that I was grounded for three weeks, which I know won\'t stand very firm, but still. I only wanted her to love me enough to listen and understand, that I can\'t breathe sometimes because I feel like nobody in the world loves me. I just need that reassurance, and I can\'t be guaranteed it. And so, I got into my email after crying to God, and hoping that tomorrow would be better, when a long expected email arrived, consisting of Christina\'s latest journal. It was packed full of everything that has happened for her that I have just ignored for the past year. I have been as emotionally detached in the past year as I was today, and I didn\'t even see everything she was going through. Maybe giving me her journal was a way to make me see, or maybe it was just an innocent gesture to show me some of the things that happened and the feelings she felt that I didn\'t really inquire about. But whatever it was, I just wanted to cry, because I had missed it all. I mean, I had been there physically, but emotionally, I was gone off in space. I had let my very best friend in the world down. Completely. I don\'t know why or how I could do it, but I did, and I hate myself for it right now. I just want to call her up and tell her that I love her, and I want to hug her, and tell her that I\'m sorry. But, it\'s 11:39pm, on a school night, and she is asleep in her bed, I\'m sure. Which is where I should be. I don\'t know how to end except to say, I should have done my homework, but I\'m glad that I chose to reevaluate my life instead. If I had to do this day over, I wouldn\'t change a thing. Except possibly my outfit. '),(21044,4117248,'02,April,2004','2004-04-02 00:00:00','   Does it make sense to like some one so much? No...but it\'s happening! I don\'t even see him, and I\'m freaking in love with him... I mean it... I don\'t want to sound all creepy stalker like.. but I find that I start to look for him in the halls in between classes, and I think about him all during math, where he should be, but of course, he was just too smart... The other day, I got in line for lunch,a nd he got right behind me, and all I was thinking was, \'I should say something\' \'No, he doesn\'t want to talk to you\'... but I just said hi anyways.. So I said hi, and asked how things were going, and he ACTUALLY talked back to me, it was amazing... Not that he didn\'t talk to me before, but before I didn\'t like him as much... I don\'t know why I like him, because I don\'t even know him, though I think I\'ve said that so many times before. It\'s true though. I mean I know his name is Richard W. and I know that he likes baseball and football, and that he is really smart, but doesn\'t like to be told that. I know that he listens when you talk to him, and makes an effort to talk back to you. I know that when he was in my math class, and I\'d ask him for some help, he wouldn\'t complain, or ask me what I would give him in return, he would just help out like a good person. I know that when I look into his eyes while he\'s talking, that it\'s sincere and genuine. I also know, that he will listen to what you have to say, even if he doesn\'t agree, and won\'t critisize or ridicule what you\'ve said. I also know, that he was raised in the church, but not baptized, and he knows the principles of the church, so he can\'t hate me because of my religion. He knew I was mormon in the process of us becoming aquantences, and he respected me, without mocking me, even though he didn\'t believe the things of the church. I know that if he ever asked me on a date, it would be sincere, and I wouldn\'t have to second guess it, though I\'m not saying that he ever would ask me, just an if... and that\'s what I know about him... I just wish I knew more. I just don\'t know how to confront him and start talking, when there are no times when I can... I just don\'t know... '),(21045,4117248,'29,May,2004','2004-05-29 00:00:00','   It\'s hard being an individual, like so many people are. I am so boring, just average. Another unseen face in the crowd, and that\'s how it\'s always been. I don\'t know why it matters to anyone out in the world, but I\'m afraid of heights and spiders. I can\'t sleep unless it\'s incredibly dark or we\'re in the car. I hate the smell of corn. I like cat\'s and dogs the same. My favorite flower is the water lily, and I love the smell of the ocean, the feel of the grainy sand in between my toes, and the feel of a wave crashing into my legs, the look of the foam coming in from the water. I hate the feel of seaweed, but love the look of it. I don\'t like any pig substance, but love Canadian bacon and pineapple pizza, and that\'s the only kind. Meat balls, meat loaf, and meat in gravy are disgusting. Chicken is best grilled and on a sesame bun. Cheese is wonderful, especially pepper jack cheese. Musicals are god\'s creation, and action films rank high on my list. Black and white photos are best, as well as black and white films. The silent scenes of a movie make it anticipating. I love the color green more than any other, but hate it on myself. I can\'t stand to try on clothes, and when I go shopping I end up not buying anything but crap junk food. I don\'t like happy endings in stories, and I appreciate the details that are far too often called boring parts of a book. I like to be a fan all the way, not half; I need to have merchandise and all their albums or books. I only watch TV for the commercials, and sometimes I will leave the TV on the TV Guide Channel to look at what\'s playing, instead of actually watching it. I love playing sports, but not really watching them. I love the feel of hard rain drops on my skin and soaking through my clothes, but hate it when it only partly rains, sprinkles or mists outside. I love the look of a lush green forest or colorful flower garden, but try to get me to plant anything, and I\'ll complain. I love the feel of moist soil on my feet. I love walking on the crunchy grass on the early mornings of winter, and I\'ll get up at six just to walk under the dark sky, and step on the crunchy grass. I think leaves are more beautiful than flowers. I like my hair long rather than short, but if my hair is long I won\'t wear it down, and if it be short, I\'ll let it spring into a life of its own. I type faster than I write, and my cursive is unreadable. I\'m afraid to excel but won\'t let myself fail, and if I do better than my friends I\'ll berate myself, same as if I do worse. Nothings good enough, but I\'m too lazy to try for anything better. I love acting and being on stage, but try to get me to sing as myself, play an instrument and give a speech and I\'ll shy away and shake uncontrollably. I appreciate humor. I\'m not very funny. '),(21046,4117248,'28,May,2004','2004-05-28 00:00:00','   My mother has been in California for most of this past month, and everybody has had to give up a few things. I\'ve had to give up my free time, my fun, my friends, my privacy, my sleep, my... the list goes on. I\'m not trying to say that I shouldn\'t have to do these things, because I know that I should, without a doubt. But when my mother came home, she was a total bitch to me, for no reason. The day she got back, she was yelling at me about anything she could fine to do. I went on the computer (after she yelled at me for being too close, and in her face) and she called me and yelled at me because she wanted me to be with her \"for a few minutes\" which turned into over an hour. So, after this hour passed, I came back to the computer, so I could, I don\'t know, maybe interact with a human being that didn\'t treat me like a stray dog. I didn\'t have my music up loud, or at all, I wasn\'t makeing any noise, I wasn\'t doing anything buy talking to Bethany on AIM. But of course, mommy dearest didn\'t want me doing that either, and she let me know. She let me know quite clearly. She yelled and yelled for all she was worth, than went to the bathroom, and yelled at me from in there, because the shampoo bottle had fallen from the edge of the bathtub, and no one had picked it up, and it was, of course, my fault, since I let the house go to waste while she was there, and just ignored messes, and stepped over them. So I get pissed, and go out to my room, and fall asleep angry, and nine thirty comes around, and my sister comes to my room saying that mother dearest want\'s me. I went to her, even though I was ready to kick my foot through the wall. She wanted me, get this, to go get her chinese food, because she was hungry. Boo hoo. Even though I\'d been asleep for an hour, and my eyelids almost refused to open all the way, I decided I would go for her, to squash any chance of her yelling at me further. I have Olivia come with me, because I don\'t want to go by myself, and I needed some one to come with me, or else I was afraid I might purposely drive myself off of a bridge. I get to the restaurant, and they are closed, and Kentucky Fried Chicken is closed, and every other placed is closed. So I come home, tell mommy dearest, and head out to my room, but she stops me before I can close the door behind me, and very sweetly, asks if I will MAKE her something! I was ready to just run at her and grab her throat, and give her something to be angry at me for, but I hold it in, and ask her what, and she wants a fried egg sandwich, with lots of mayo, but not fried too hard, and the bread not toasted. I swear I almost threw my hand through the window in anger. I never get these urges to want to inflict pain on anybody, but I couldn\'t respond, or she\'d yell some more. So, I climb into bet at ten, and, look at this, I can\'t sleep, because I am SO angry. I stay up the whole night hating being alive, wishing I were dead, and thinking up all the things I could do to help that last bit get into action quicker. Like I said before, I never feel those things, I never really want to kill myself. Everybody THINKS about it every once in a while; what would happen, who would cry and be sorry for the things they said to you. But no one ever really actually plans a way that it can happen. I did. I decided that I wanted to come in the house in the morning, and get all the pills that Tone is so stacked up with, and just take them. They wouldn\'t know, until at least four or five... but I don\'t, because I believe in something higher than myself. I go throughout my day, sometimes still seething from the night before, more often trying to be happy. I can\'t stand some people, like the people that I want to like, but just after they\'ve redeemed themselves for something they said last week, they do it again, only this time worse. Take Jason from Theatre as an example. The tampon incident just wasn\'t enough for him, now was it? No, now every time he sees me he has to sing the \"Fact\'s of Life\" theme song, and say things relating to me being a whore. I want to like him, because he\'s cute, and sometimes he can be bearable, but not today. I swear I was ready to kill him, when he said I was tacky. I\'ve never said so much as one rude word to him. Never. So I\'m in this horrible mood fifth period, and six period comes around, and I\'m still angry, but then I see Addison, and he really makes me happy. He makes me laugh, and he laughs with me so I don\'t feel like an idiot. I want to like him, more than just a friend, but he has a girl friend, and he is a junior, and.. I am me. So I decide I don\'t want to like him, even though just looking at him makes me happy, because he\'s so much fun to hang around with. So I leave school all happy, and pull into the driveway, and the bad mood hits me. And to make a very long, longer than I intended, story short, my mother does almost the exact same thing to me, and she yells, but this time, I react differently, I yell back, because I\'m not going to sit there and let her ruin my life. Too late. So, why do I hate myself so much, when she\'s the one that makes me so angry? ... Because I\'m the one that she reacts to in the first place. God, I\'m such an idiot, and I don\'t even know it. I don\'t know what to do anymore. Sometimes I feel like I can\'t breathe, and than mommy dearest comes a long, and I just want to kill myself. Get a knife, and slit a wrist, and get if fucking OVER WITH! I\'m not abusive, I\'m not crazy, I\'m not even depressed, I\'m just fed up. FED UP! '),(21047,4117248,'31,July,2004','2004-07-31 00:00:00','   The past couple of days, my family has had a booth at Summer Fest. The booth next to us sells little glass figurines, like elephants and dogs, ballerinas and piano\'s, ships and turtles. Not only do they sell them, but the man makes them, right there in his booth, using only a blow torch, a pair of Digidium glasses and the sticks of glass. I find the process to be such a fascinating thing; how it\'s just the movement of his hands and his imagination. I suppose I had little stars in my eyes, because he let me come to the back of the booth and watch him make a bullfrog (a frog with a bulls head; horns and all). I even got to wear the special glasses. I am still captivated by the skill and patience required to make such beautiful things. While I was thinking about it last night before bed, the image of two very large hands with bloodied scabs on them came to mind, and at once I knew they were the creators hands; the hands of the one who sacrificed his life for me. It\'s amazing what skill and patience the lord had with this world and its inhabitants. And then I realized how grateful I was that God had the patience he did, and that he was as forgiving as he is. I guess that\'s why he\'s God, eh? '),(21048,4117248,'29,July,2004','2004-07-29 00:00:00','   There is a reason that I haven\'t been writing in here lately; I\'ve been too depressed, too hurt to take the time to do so. I guess I should start from the beginning, eh? Okay, well, the beginning is Wednesday, the 21st. On Wednesday, I knew that Christina was going to go with another one of her internet boyfriends to the Drive-in, and I was getting so worried that something would happen to her, that I decided that I should go as well, in a different car, of course, but parking right next to her. So, I got home, made the plans, asking my mom if I could \"give Olivia a birthday present early\" by taking her to the Drive-in with me. Well, okay, and I wanted olivia to enjoy herself, so I let her choose one friend she could take. She choose a boy named Jeremy. That was my first mistake, because Christina\'s date canceled, but I have Olivia\'s hopes up, and I didn\'t want to break that, so I decided Christina should come too. So it was Olivia, her friend Jeremy, Christina and I, all going together in my car. That was my second mistake. Well, I had to put Tony to bed, and so Olivia, Christina and Jeremy were waiting in our front yard, and appearently Christina and Jeremy found some interests in eachother. I noticed, but younger boys will flirt, and Christina is a flirt, whatever. It didn\'t really bother me, until we were already at the Drive-in, and Christina insisted that she and Jeremy sit in the back of the car together. I gave in, and let them. My third mistake. The entire movie they were getting close to eachother, and ended the movie with making out. I ignored it, my fourth mistake. So, we were sitting in Spiderman 2, when they all decided that they wanted to go somewhere and park the car. By this time I had had enough, but I didn\'t want to be the party pooper, so I endured it. All night, until two thirty am I was listening to them making out, watching the road with total disdain, and wishing I were home in my bed, that I hadn\'t gone to the drive in at all, and that Christina cared about herself more than to get with a guy who was younger, smoked, was getting busted for drugs, and already had a girlfriend. Well, the next day, I was shoping with mum in Salem, when I got a call from Christina, asking if she could go into my room, because her whole house was super hot. I had an idea that she was going to meet Jeremy, but I wanted to be the cool friend again, so I said yes. My fifth mistake. So, I get home, and I purposely didn\'t knock on my own door, and when I swung it open, Jeremy has his shirt off, and Christina is whiping her lips with the back of her hand. I got in, apoligized, got my stuff, told them to leave soon and be quiet, and just left them to do whatever they wished. My sixth mistake. Then I left for the weekend with Emma and her family to Newport, along with Abbe Groh and Stephanie King. I had so much fun, and was being so carefree; enjoying the beach, and the company, and the never ending mint chocolate cake, when on Sunday, I got a text message from Christina saying, \"What would you say if I told you that Jeremy and I had sex on your bed?\" I replied with, \"Did you?\" I was in shock. I thought that after I had left on Thursday they had had sex, and it was my fault for letting them stay in there without me being there, or my parents knowing. \"No, but can we?\" She answered. \"No, I\'m sorry... actually, I\'m not sorry, because one day you\'ll meet some one that you\'ll think is special and who will make you fill special, and you\'ll regret giving away your virginity to some one who didn\'t care about you. I don\'t want you to hurt yourself like that.\" Is what my little speech was to her, but I guess it wasn\'t enough... Right then I should have called my mother and told her to make Olivia stay inside, and lock my doors, because a plan was being planned. I wouldn\'t have told her what, just to watch out; but I didn\'t. My seventh mistake. On Sunday night, I got a call from Olivia asking for Christina\'s number. I told her it, but then asked her softly, while crying, not to call her. She said she wouldn\'t, but I made her promise me that she woudln\'t call, and I begged her not to get involved. I just didn\'t want her to get dragged into anything that she didn\'t have to. I love her, and anything that can hurt her, I want to keep her from. Obviously, me crying for her to please not do something wasn\'t convincing enough for her not to do it, because when I got home on Monday, Tony told me that Christina and Olivia had been spending a lot of time together. I knew right then, that Olivia had broken my promise. I had asked her on the phone, \"Think about it, Olivia. When\'s the last time that Christina wanted to talk to you?\" I wasn\'t trying to be mean, but I wanted her to understand that Christina was using her for Jeremy, even though I wasn\'t sure how, I knew that it was the truth. I was right. After I got home, and Olivia showed up, I made her tell me if they had had sex in my room. They had, and not on Thursday; they had been using my room all weekend to do whatever the hell they wanted to do. I cannot describe how hurt I was. How used I felt, and still do felt. Now I know the whole thing, how and when, but I still don\'t understand why. What were they thinking to lie, and sneak, and.. use me and my parents. Both of them, no, all three of them, Olivia, Christina, and Jeremy. So, I was enjoying my innocent weekend with the friends who don\'t decieve me; meanwhile, in Dallas, Oregon, Carolyn Hudson\'s room, this is what was happening: Christina and Jeremy were using my room all weekend for whatever they wanted to do, and Olivia was helping them get in and out. On Sunday night, after I had told Christina that she could not use my room to have sex in, they used my room to have sex in. Jeremy gave Olivia a few ciggerettes to smoke, they smoked in my room, had sex in my room, un-virginized my bed and room, cheated, hurt, snuck, hurt, lied, and have I mentioned hurt people, all in the confines of my room. *Sigh* I still cannot process all the hurt and pain that I\'m feeling still. On Monday, I told Tony, called Christina, told her all the pain she had caused in my family, and felt like total shit (excuse the language) all night, hardly slept, and dreamed of only people decieving me. Then, on Tuesday, I made the decision, that I would tell Christina what she needed to hear, that I cared for her, and that she should care for herself, and how she had hurt me, and we were in there for at least an hour and a half, most of the time me talking. And then I decided, it was enough, she was sorry, she was reaping the consequences, and I had done everything I could. And so my mind was made up; I could forgive and forget, really, or I could stay mad, and hurt our friendship as much, if not more, than she had, because it would have been intentional on my part. I made the decision to take her out for icecream, and deal. Just because I forgave her, doesn\'t mean that it isn\'t effecting me. To say that, would be lying to only myself. I\'m so hurt, not just by her, but by Olivia, and the whole situation. I just, I cry to think that I could have stopped this, some how, even though it wasn\'t my fault in the least. Oh, lord! I still don\'t know what to do. I want to trust both Olivia and Christina as much as I did before this whole thing, but now I don\'t think I can. I don\'t think I can trust anybody right now, because I am still so hurt, I still feel so used. And I honestly don\'t know what to do. I just don\'t know, anymore... '),(21049,4117248,'20,July,2004','2004-07-20 00:00:00','   After I wrote that last bit, I promised myself that the change would happen right then. Well, every one who is smart understands that change takes time, I know that, but I wanted a change in my attitude; a change in my day. Of course I changed my attitude, for the better, but Satan knows me, like I learned at camp, and he knew exactly what to do to get me down again; only this time, I wouldn\'t let him! When we first got to La Pine, Grandpa promised he would take Christina and I out on a back road to drive the pick-up. Well, today he fulfilled that promise. Christina drove first and every thing went fine; a few cars passed and the steering was a little difficult for her, but other than that, all was well. We drove out to Twin Lakes, where we got a drink; then it was my turn to drive. Every was fine at first; I backed up great, slowed to the first stop sign great, started off great! And then, I looked at the speedometer. It was at zero- I knew I was going faster than zero, because I was actually going at all. Grandpa said it would kick in soon: he was wrong. I had asked Christina to tell me when to speed up and slow down because I was a bad judge of speed. She sure let me know, like I\'d asked, so I shouldn\'t have been upset, but it felt like she was criticizing me. She tends to be better than me at most every thing, so this just felt like another time I was doing some thing \"second best\". She\'s never said anything like that to me, so it\'s not her, it was me. So, here I was driving along, doing okay for not know what speed I was going, and being told to speed up every two minutes, when a junction sign came up. I asked Grandpa three times if that was the junction I needed to turn off onto, and the second time, Christina told me he would tell me when to turn. As we came almost right on it, Christina asked him. He didn\'t hear or reply until we were passing it. \"Okay,\" I told myself, \"I can do this.\" I knew Grandpa hadn\'t meant for me to miss the turn, or that it was my fault, but I still felt inadequate. Well, I missed the first road to turn off, so we could turn around, and then the second road came up, and I stopped really fast. And I mean FAST! It was because I was stressed and it\'s hard for me to calm down when something goes wrong. After getting off the highway, really fast and stopping bad, I\'d had enough; Grandpa turned the tuck around and I thought I was done. Oh no, he wanted me to finish. \"Okay,\" I told myself again, \"I can do this.\" I wanted to change, right? Well, giving up would not help that goal, so I got back in the drivers seat, and started off, only to find that my door wasn\'t shut all the way. Easily fixed, I knew I could do this, only now, I didn\'t want to. After, Grandpa had made the statement, \"Boy, when she stops, she really stops,\" to Christina. I don\'t think he realized how much that hurt me, after all those things had happened to me. I imagined that I would get home, go up stairs and cry my eyes out, as we made the last few miles home and then a thought occurred to me: \'Which of the things that had happened today were in my \"center of control\"?\'--something I\'d read from a book titled, \"The 7 Habits of Highly Effective Teens\" by Sean Covey. The center of control is what you control and what you don\'t, and it\'s good to know the difference, because it\'s a waste of your time and energy to get upset about things out of your center of control. So I decided that it was in my center of control that I\'d stopped fast, and yet that was under a stressful situation, and so it woudn\'t be effective to get upset over anything that happened today. I know myself and my limits, and stopping quickly is some thing I need to work on, but really couldn\'t be helped at that moment, so even that was in a sense out of my center of control, and therefore simply unnecessary to get angry about. It was out of my control that Grandpa had said what he had; It was out of my control that Grandpa hadn\'t been paying attention to the junction turn off; it was out of my control that the speedometer had broken. All of it, so getting upset, and crying over it all would get me no where, and I\'m trying to move forward, remember? I\'d been thinking all this, while on the way back, taking a chance to help myself, and hence was in a quiet meditation for the majority of the time. It had been decided in my mind, my center of control is over only me, my actions, and so getting upset wasn\'t the way. \"Okay,\" I thought one more time, \"I can do this.\" Only I didn\'t realize that as soon as I\'d made that decision, Satan would work on me just a little more, hurt me just a little longer, poke and prod at me just a little bit harder. When we\'d gotten home, we were telling Grandma about our eventful hour, when Grandpa said, \"The ride home was the only time she\'d shut up.\" Ouch! That hurt. I know he hadn\'t meant it to. I know that it was out of my center of control, but I knew the pain searing in me was louder than my reasoning; going upstairs and crying my eyes out didn\'t seem like too bad an idea after all. And then, something made sense; I was being tested by none other than Satan, and I was about to fail. If there\'s one thing I am not, it is a failure. I was not going to let Satan, that snake that can bruise a heal, get any satisfaction from me, so like a priesthood holder to a possessed soul, I bruised that serpents head, by saying to myself, \"... O thou child of hell, why temp ye me? Knowest thou that the righteous yieldeth to no such temptations?\" (Alma 11:23) I knew right then, without a doubt, that I would change. It will be just as hard, and take just as long, but I will not, and I repeat will not fail. The righteous yieldeth to no such temptations, and I am, no matter what I may do, say or think, a righteous person, because I love Heavenly Father and Jesus Christ, and have the Gift of the Holy Ghost, and I will repent forever, so I can stay righteous. This isn\'t a new concept, but a very helpful epiphany. God works in mysterious ways, and he helped me, in more ways than I can count, to come to this one conclusion. I have a center of control, and what some one says, no matter who much it hurts is not in it; and Satan is not in it; and I am not going to let myself be put into Satan\'s center of control. I will love God with every thing that I am. It\'s funny that I read about the \"center of control\" just a few days ago, and how it has changed the way I think and react SO immediately, and it\'s funny that Bekah Butterfield had told me her favorite scripture once (Alma 11:23) a long time ago, and I wrote it down, and remembered it. It\'s funny how such a bad day can turn out so good. \"Okay... I can do this.\" '),(21050,4117248,'19,July,2004','2004-07-19 00:00:00','   Today is just not my day... I break two of Grandma\'s glasses, mess up Christina\'s iced mocha latte thing, lead Andrew on and then drop him like a ten point bag of potatoes, and break the promises I\'ve meade to myself and God. And I\'m two hundred and fifteen pounds. Why, oh why, do I have to go through day\'s like this? I\'ve had just about enough of trying to be everything for everybody, and I\'m not getting anything from it. I\'m trying to be good and choose the right, but it\'s so hard when I can do something that no one will see... Only problem is that some one does see; Heavenly Father see\'s what I\'m doing to myself... I see what I\'m doing to myself, and I don\'t like it. I don\'t know what to do anymore. Pride is keeping me from seeking help. Well, not pride, because I feel no pride for myself or the things I\'ve done, only hatred. Fear is what\'s keeping me from getting help; Fear of being judged; Fear of being judged correctly. I don\'t want people to see who I really am; what I really think, because that would give me what I deserve, a whole lot of lonely days to think about myself, and openly hate myself more. Maybe I should just give up. Maybe... but I know that\'s not the way. It\'s just hard to live life by the principle that is the way. Like I told Andrew, on a some what different subject, but one just as important, I know what I should do, I just don\'t know how do to it... I guess no one always has \"happy, happy, happy day\'s\" and I just need to make it through one more day of a bad one, to come out on a hopefully good one... If anyone could read this journal and see the difficulty I\'m having at living life righteously and happy, that person would know me so well; not just my likes and dislikes, but who I really am, and I don\'t even know that all the way... Maybe someday I\'ll figure it out, just not today. I live almost differently than I write in here. A normal day will have me happy and carefree, hardly remembering Heavenly Father and Jesus Christ, hardly praying or reading his words, hardly being a friend a friend would like to have. That\'s something I want to change. Change again... Why does it always come back to that? Because I\'m finally realizing it\'s what I need to do for my life to be successful and what I will do. I never want to regret, so I have to make some changes... Just some changes. '),(21051,4117248,'19,July,2004','2004-07-19 00:00:00','   Last night Grandpa got me and asked me to come watch something with him. So I obliged, not knowing that he was going to have me watch something about \"The Mormons\" and their alleged \"polygamous acts\". Now, don\'t get me wrong, I love grandpa, but him trying to feed me anti-mormon things, just because a few cracked-pot jokers think incest and polygamy\'s the way, makes me upset. I know who I am, and where my life is going, at least in terms of my faith, so I don\'t need help re-deciding things. I\'m happy where I am with my faith and my beliefs and it\'s only causing me frustrations that he did that. I know he thinks he\'s doing me a favor, but... it only upset me. Another reason why it upset me so much was because I couldn\'t tell him that the \"Latter-Day Church of Christ\" is not the LDS church that I practice in, and that the church no longer practices polygamy. I couldn\'t bring myself to get up, tell him I wasn\'t interested and walk away. If I can\'t tell Grandpa, some one that\'s suppose to be close to me, that the anti-stuff on TV doesn\'t hold any interest for me, that I\'m happy with my beliefs, that I love my God and Savior, then will I be able to tell people on the street, people who may be one of God\'s \'ready to harvest\' fruits? I know what my answer is now, \"Yes, I would\", but I\'m hoping to change that into something as sure as my testimony of Christ. We made sour cream noodles today. It was a fun process for me, and something that requires a few people to do it, so I\'m going to try to do it with my family when I get back home. Also, making dinner with Christina the two times we did has helped me to realize that I can take a lot off of mommy dearest\'s shoulders by cooking once or twice a week for the family. I think I\"ll make that one of my Personal Progress goals, maybe as a ten hour project. And it will really help out mum. My scripture reading has sort of flopped while I have been here, but I\'m hoping to pick that right back up again. I\'m pretty sure I won\'t have the Book of Mormon entirely finished by the end of summer, but I\'m going to read enough to satisfy myself and Heavenly Father. I know my floundering in my thoughts and actions has created a fault in my relationship with Heavenly Father, but I also know that I\'m getting better, which helps with me wanting to get better. I prayed that I\'d see some results, and I\'m seeing them. Proof that I couldn\'t live without praying. I swear I am NOT obsessed with Andrew, it\'s just he\'s said some things and done some things that I\'ve taken a special notice to, and he inspires me to be a better person. I met him in February on an LDS chat room. At first we just chatted like two people getting to know each other, but then it continued to come to the point where he says *kisses* and *huggles* and he gave me his cell phone number, and we text message a lot. He makes me so happy- not in an \'in love\' way, but, just knowing him has made me a better person, and isn\'t that the type of people I should be around. I don\'t know... he\'s just awesome. I\'m both glad and sad about him leaving on his mission. I know I\'ll miss him. '),(21052,4117248,'18,July,2004','2004-07-18 00:00:00','   Andrew\'s going away and I\'m sad, because he\'s really helped me through the things I\'m going through... I\'ll miss him, but he said he\'ll write to me and tell me about the experiences that he\'ll have, and it\'s sure he will have some awesome experiences as a missionary. I\'m way excited for him, but still kind of sad he won\'t be on line every time I need to talk to him. Oh well. I can only pray for his success! Maybe someday in the future I\'ll meet some one as cool as him, closer to my age, and available to be mine! Maybe... '),(21053,4117248,'17,July,2004','2004-07-17 00:00:00','   A blot on a clear sheet of paper'),(21054,4117248,'15,July,2004','2004-07-15 00:00:00','   I know that being the Laurel class president means I have an increase in responsibility, and I readily accept those additional responsibilities, but there are some forces in my life in which I have no control over, and some of these forces have been working against me. I have a father who makes and brakes promises, and in doing so makes hope and breaks my heart, every time. He\'d rather be lost in a world of alcohol than hear my pleading voice with a sober ear. It\'s hard for me to be happy because of this and so I act like I\'m happy. Most of the time it works, but others, it doesn\'t. Sometimes I wish I didn\'t have to get out of bed, he hurts me so much. my mother, hard as she may try has a fast and angry temper. No one\'s perfect, I know this, and I also know that she loves me dearly, but that doesn\'t always stop her from yelling about the small things like a broken dish or a basket full of unfolded laundry. Because of this, my relationship with her is dwindling, and I find myself wishing that there were no relationship to salvage. I love my sister with all that I am, and yet, she still finds a way to upset me to the point of explosion. I don\'t care so much about the stained clothing or the broken thing-y that she borrowed as much as the attitude she has of not caring about how it makes me feel. I have no ro-model, except for Christ, and sometimes hoping to be like him seems like such a waste of time it hurts to breathe. Missing a meeting because my alarm clock is broken; being late because my mom had me go to the store for her; not calling because I\'m grounded for some missing \'Propel\" water... all these things are out of my \"circle of control\", and if I dwell on them any longer, I will explode. The only thing I control is myself and my relationship with Christ, and so I\'ll work on that as long as I live, but I must give up the things I have no control over. '),(21055,4117248,'13,July,2004','2004-07-13 00:00:00','   Yesterday was the journey to Christina\'s grandparents (I call them Grandma and Grandpa) house. It was long- but it was fun. I keep hoping the ten days we\'re staying will be just as fun, if not more. It\'s hard spending so much time with Christina and Cole, but it\'ll be okay. The only thing I\'m worried about is getting Christina angry with me before we leave, and that\'ll ruin my whole trip. I don\'t want to think about angering her but she is easily agitated and kind of moody. I haven\'t talked to her about her internet boyfriends yet, but again, I want her on my good side for the time being. And besides, it does kind of intreage me a little bit. I know it shouldn\'t but it does. I just don\'t know what to do or think anymore. Life isn\'t always fun, but I\'ll make it. I know I can and will. '),(21056,4117248,'09,July,2004','2004-07-09 00:00:00','   Once you\'ve gone off the path it\'s seven times harder to come back; it\'s like trying to be good in dog years. I\'ve learned this. But that doesn\'t mean I can\'t get back. I talked to Andrew again last night, and he really is rooting for me. He\'s one person who really is inspiring me to change my ways, because he was going through the same thing, and now he\'s going on a mission at the end of the month. I know I can do it, too! I just have to keep working at it, because if you\'re not moving forward you falling back wards and I want to quit falling and start moving forward again. '),(21057,4117248,'08,July,2004','2004-07-08 00:00:00','   So, I\'ve been thinking a lot about myself and the way I want my life to be, and how I want it is not how it\'s going. I want that to change. Change is a hard thing to do, that is certainly something I\'ve learned over the years, but not impossible, and besides, if it isn\'t hard, than it wouldn\'t be something worth fighting for, even if it is yourself that you\'re fighting against. So here\'s what I want to do: Change. Change who I am, how I think, what I do and say, and what I stand for. It\'s going to take work and it\'s going to be hard but it\'s going to be worth it. It\'s going to be eternally worth it. Now, I once read a very unique book called, \"For Time & All Absurdity\" by Robert Farrel Smith, and in it was this quote, \"... a goal not written down is just a wish.\" So, to follow through with my goal I have to write it first (or type). I want to stop hating myself; I want to love myself. I want to think eternally and banish all harmful thoughts from my mind. I want to pay a full tithe, not through obligation, but by wanting to serve God. I want to pray regularly to my Heavenly Father and read from his words daily. I want to be married in the temple to a man who loves me but loves God more. I will. '),(21058,4117248,'06,July,2004','2004-07-06 00:00:00','   Let me ask you: How much of a dork can one person be? Well... Now let me answer you: One person (me) can be the biggest dork in the entire world! Here I am, complaining about working for maybe forty hours a month and having to pay a couple of bills, when Nephi is crying out in repentance because he was praying to God in a nervous fear that his brothers were going to kill him, and that was supposedly wrong, because he\'s had so many blessings...? Man, I have got to straighten out my priorities. It says in 2 Nephi 5 (or 6) that the Nephites learned to work with their hands, which in biblical terms means to work your freaking butt off for almost nothing in return. New goal: Try to be positive about my work. Maybe then it won\'t be so hard... Maybe. (Most assuredly.) '),(21059,4117248,'05,July,2004','2004-07-05 00:00:00','   I suppose it\'s best I learn this now, but work sucks. I work for a minimal three hours a day, three day sa week, four weeks a mont, get three hundred dollars a month, pay ninety dollars a month for car insurance, ten percent of my paycheck a month for tithing, pay at least twenty bucks a month for gas and have to buy anything I want and have to save money, too! What does that leave me with? Nothing but a headache because I had to do math, and some tired fingers for typing it all too fast. I said it before and I\'ll say it again: Work sucks!... but it pays the bills!;) '),(21060,4117248,'05,July,2004','2004-07-05 00:00:00','   A few nights ago my family rented a couple of movies to watch together. They were: \"Along Came Polly\", \"The Secret Window\" and \"50 First Dates\". They were all okay, but I only really liked one; \"50 First Dates\" with Drew Berrymore and Adam Sandler. Now, normally any movie with Adam Sandler is pretty useless, funny as heck, but containing no real moral or value. This movie, however, was brilliant. Though it still had some crude humor, and unnecessary sexual innuendo\'s, the moral and ending were beautiful. The story line is a player (Adam) who gets with tourists as weekend flings, meets a beautiful, charming girl (Drew) and falls in love with her right away. The next day, he tries to sit with her at breakfast, she seems not to recognize him, and finds him to be frightening. After finding out that she has no short term memory, Adam begins to meet her every day, for the \"50 First Dates\" and gets her to fall in love with him every 50 times. He makes a video for her to watch every morning, so she knows that she loves him, and the story ends with them being married with a child, her still not remembering. The reason I loved this movie so much was because the moral is: true, real love, is waking up every morning, with a fresh start, and falling in love again and again, over and over, forever and ever. That is what I want out of my life, my marriage-when I have one. So, I have to choose the right guy to make me fall in love with him every day. Here are just some things I would look for in this man: *He loves me completely*'),(21061,4117248,'04,July,2004','2004-07-04 00:00:00','   \"My Dearest Love, I want you to know that I believe in you. I believe that you have what it takes to make your dreams come true, and I just want you to know that I\'ll be there beside you through both the good times and the bad times. I \'ll be your light and warmth when the road gets rough. I\'ll lend you a shoulder; I\'ll hold your hand when you hesitate. I know you can make it. Never stop trying! You never know what you can achieve or where your dreams will take you! I love you!\" That was an e-mail I got once from a friend, and I wanted to always remember those words. When I replied to that e-mail she told me that she hadn\'t written it, but it had captured everything she felt, so I sent it to my sisters, and it\'s a chain reaction, because I found a note Cheyanne had written to her friend, which had the same message on it. When I think about it, it\'s exactly what our heavenly father would say to us, if we would just listen long or hard enough. My goal is to share these words and listen to the words of God. '),(21062,3827259,'17,July,2004','2004-07-17 00:00:00','   After a gruelling two weeks of work n study, I finally got time to blog (brag?) about my trip to Washington D.C that I had refered (see CGI.pm for the spelling) to in my earlier post.     We (Vishwanath, Charan, Srikanth, Seshan, myself) started on a bright Saturday morning from Buffalo, NY in a Chevvy Cavalier. For a person who drives a Japanese car with stick shift, a Chevvy with auto gear was not very thrilling. After a hearty breakfast, we hit the highway at 8.30 and cruised along at a nice pace of 70 miles/hr. Then Vish took over from me. He had been wanting to drive ever since we started and after a lot of persuading and pestering I finally handed over the wheel to him after an hour and a half. He was nervous in the beginning but soon overcame that to drive smoothly. The route we took , US - 15 though scenic had a lot of traffic and we sometimes had to move at as low as 40 miles/hr. After a while, I took over again and madness took over me.    I like driving and I like driving fast but not dangerously fast. 80 miles/hr is a good speed and I normally drive at around that speed, sometimes touching 85. And thats precisely how, I should have been driving that day. I was doing 95, 100 and was in full control. The car was pretty new and was very different from mine. I was getting away with doing 100 and no cop in sight or thats what I thought. I was climing up this slope doing 95 and just as I was about to go downhill, I saw a cop waiting for his prey. I braked hard, but it wasnt enough. He had already made his kill for the morning. He caught me at 88 and locked me at 80.     I gave him some crap about the road being downhill and that the car needed a lot of power. He wasnt willing to buy any of that and gave me a ticket. The dreaded moment I dint want to happen in my life had happened. My record has been tarnished and I\'ll have to live with it for the next seven years. It was a good lesson for me and it has taught me to keep my excitement under check. I thought of contesting the ticket as I believed that I had a chance of getting away with this. But the drive was too long and I felt that I needed this to restrain myself from any harakiri in the future. But it did not hamper my spirits by a long way. We continued driving at good speed and reached Washington D.C in 8.5 hours which was not bad at all considering that we stopped atleast three times.             The start had been very disappointing and I was concerned about how it was going to turn out. It wasnt all that bad but it certainly could have been better. When we reached, I had forgotten about my ticket and was planning for the next two days of our trip which I\'ll recount in my next blog. In the meanwhile if you have any questions on how to get a speeding ticket, feel free to mail me. ;)                 Oh boy !! Am I proud of my narrative skills ? Laloo could write better !! '),(21063,3827259,'02,July,2004','2004-07-02 00:00:00','   Hello ppl,     This is my first blog. I am Sripathi Guruprasannaraj ( or long_name_raj), a Master\'s student at SUNY Buffalo. Its been a while since I have written anything, so I am left fumbling for words here. Anyway my passions are driving, programming, music and books and traveling. I like to visit new places and some places I have already visited. And thats precisely what I am doing tomorrow. I am going to Washington D.C with my friends for the long weekend. I have been there before and got attracted to the Capital instantly. Ever since I have been longing to go there and this is a perfect opportunity. Not only would it be good to go back once again, its also a great drive to there. So am looking  forward to the trip tomorrow. Hmm now I cant seem to stop. I want to write a lot more but I guess I\'ll stop now or else come back again later this evening.  Have a nice weekend folks !!  Raj  PS: If anyone from Washington read this, gimme tips on places to park and places to eat in the Capital. '),(21064,3466070,'26,May,2004','2004-05-26 00:00:00','     Welcome to my lil\' SickandTwisted home page. You may want to refrain from telling people you visit.    '),(21065,3466070,'31,July,2004','2004-07-31 00:00:00','     On the 29th I went to the Kittie concert. It was one hell of a show! There were five bands. One local Canadian band(there first time touring the states), 12 Tribe, 36 Crazy Fist, Canderia, and of course Kittie! It was by far the best show I have seen. Kitties mosh pit was crazy! Kittie had an awesome stage show. They had Black, Neon Green, and white strobe lights that went to the beat of the double bass drums. It was cool. The other bands were good to. The lead singer of 36 Crazy Fist stage dived into the crowd and let the crowd sing. It was awesome! I will have pics of the concert soon.    '),(21066,3466070,'30,July,2004','2004-07-30 00:00:00','     I have a lil\' bro by the name of Zach. He is real annoying, but today I have to reflect on all the good things we have done together. We have done some crazy shit man... Like light the hill on fire! LOL. Today is Friday the 30th of July and he going in to the docs so they can diagnose my bro for a heart disorder. He may need heart surgery, but I am hoping for the best. As much of a lil\' shit he is if anything happened to him I don\'t know what I would do! You besta take care MoFo!    '),(21067,3466070,'29,July,2004','2004-07-29 00:00:00','     &nbsp;       Yeah... What can I say? Girls in Korea like to make it look like there assets are peeking out of there pants... Weird. Although my mom thinks that the assets are printed onto the skirt; Sarah and I think that they are really mesh holes exposing there butts... Look at how the butts match perfectly with the legs... Is it a conspiracy or do there assets really hide beneath the fake ones printed upon there butts? Go to Korea and find out for yourself.  Kogal    um... Right... If the point was to look like a crazed fan of another bleached out Korean R&B singer then your on your way there, but if you were trying to look like another Penny Wise(Clown from IT) disciple... Well then you got a while till that happens. This is a sick world!  KOGAL Style    '),(21068,3466070,'28,July,2004','2004-07-28 00:00:00','     Well... I came to this conclusion recently because of some odd events. First I thought my mom was being cool about buying my condoms for me, but now she wants to invest into getting Sarah on birth control! Well it isn\'t that it is a bad idea, but the thought kinda freaks me out. Sarah thinx it is a good idea, but it is weird comin from my mom. We don\'t to have to rely on my mom for such a thing and hiding it is another problem and keeping up with taking them with school and work... It could be quite the hassle. She has been acting very weird though recently. Our van broke down the other day and she has been freaking out about how I need to stay home because we could be having a family emergency even though by lil\' sis and bro are hangin with there friends and I wouldn\'t be of any help. Just odd...    '),(21069,3466070,'27,July,2004','2004-07-27 00:00:00','     Once upon a time there was a girl and a guy who were really horny! They thought about what they could do when the guy asked if they could try to 69. The girl unwillingly said, \"Fine... But you have to be on top!\". As things started to heat up something bad happened, and no... A parent didn\'t walk in on them... It was far worse! The girl who must have been feeling so much pleasure that a little bubble in her tummy move up through her bowels. As she began to approach her climax she let out a little climax of her own from her butt! Yes she farted right in his face! Oh, but it doesn\'t stop there. There is such a thing as karma... yes they did 69 again, but this time the dude let out his own roar. yup, he farted right back in her face... lol. What irony.    '),(21070,3466070,'27,July,2004','2004-07-27 00:00:00','     I don\'t like Bush... I don\'t like Bush at all. He sux! I personally am a democrat. I hang with people who are liberal, hippies from the green house. I don\'t mind them, but Bush gave republicans a really bad name! If you wanna know what I mean then go see Fahrenheit 911. Once you see what that dumb shit has done you will see were I am coming from. In that movie it shows how we shouldn\'t have fought Iraq over weapons of Mass Destruction, because we are the ones who gave them the nukes! Bush... You dumb shit.    '),(21071,3466070,'26,July,2004','2004-07-26 00:00:00','     *Cradle of Filth- Hallowed Be Thy Name *Dimmu Borgir- Hybrid Stigamta *Hate in the Box- bloody Ballerina *Coal Chamber- Sway *Nine Inch Nails(NIN)- Head like a Hole *Nirvana- Come as you are *Kataklysm- Face to Face of War *Mudvayne- #1 *Metallica- One *Arch Enemy- We will Rise *Kittie- What I always Wanted *Slayer- Blood Line *Old Man\'s Child- Black Seeds on Virgin Soil *At the Gates- With Fear I Kiss the Burning Darkness *Eighteen Visions- Vanity *Static X- invincible *Atreyu- Lip Gloss and Black *Avenged Sevenfold- Unholy Confessions *Muse- Time is running out *Dark Tranquility- Monochromatic Stains *Kill Switch Engage- My Last Serenade *Children of Bodom- 18 Pounder *Lamb of God- 11th Hour *Strapping Young Lad- Devour *Morbid Angel- God of Emptiness *Behemoth- It is I *Sarah Jezebel Deva(with Cradle of Filth)- Throned,Queen of Winter    '),(21072,3466070,'24,July,2004','2004-07-24 00:00:00','     Well last night was the first time Sarah sleep over at my house. We had a blast. We watched Freddy vs. Jason and ate chili cheese its. Yummy...Although she had to leave at 9 in the morning we still got a good 5 hours of sleep. LOL. We gotta do that again.    '),(21073,3466070,'24,July,2004','2004-07-24 00:00:00','     urlLink Haloscan  commenting and trackback have been added to this blog.    '),(21074,3466070,'17,July,2004','2004-07-17 00:00:00','     Although many of you probably go off bragging about your new girl friend or boyfriend to your parents my girlfriend Sarah likes to keep it from her parents so they don\'t get involved. I didn\'t think we would ever get caught, but her pops caught us today. I had my arms on her and she was holding my sides when he busted in with a big \"Hey!\". I was startled and my partner was worried that he had caught us in the act of kissing. Thank god he didn\'t, but I feel this will the be the cause of my demise. I hope he will still like me(That is if he even did in the first place). I look back now I realize that it was quite funny.    '),(21075,3466070,'17,July,2004','2004-07-17 00:00:00','     urlLink Haloscan  commenting and trackback have been added to this blog.    '),(21076,3466070,'17,July,2004','2004-07-17 00:00:00','     Sarah and I  Me without Sarah   Guns and Roses: Don\'t Cry Tonight  Talk to me softly there\'s something in your eyes don\'t hang your head in sorrow and please don\'t cry i know how you feel inside i\'ve i\'ve been here before something is changin\' inside you and don\'t you know  Don\'t you cry tonight i still love you baby don\'t you cry tonight don\'t you cry tonight there\'s a heaven above you baby and don\'t you cry tonight  Give me a whisper and give me a sigh give me a kiss before you tell me goodbye don\'t you take it so hard now and please don\'t take it so bad i\'ll still be thinking of you and the times we had baby  Don\'t you cry tonight don\'t you cry tonight don\'t you cry tonight there\'s a heaven above you baby and don\'t you cry tonight  And please remember that i never lied and please remember how i felt inside now honey you gotta make it your own way but you\'ll be alright now sugar you\'ll feel better tomorrow come the morning light now baby  And don\'t you cry tonight and don\'t you cry tonight don\'t you cry tonight there\'s a heaven above you baby and don\'t you cry don\'t you ever cry don\'t you cry tonight baby maybe someday don\'t you cry don\'t you ever cry don\'t you cry tonight  I love you Sarah...    '),(21077,3466070,'17,July,2004','2004-07-17 00:00:00','     urlLink \"Cradle of Filth\"    '),(21078,3466070,'09,July,2004','2004-07-09 00:00:00','     I\'ve come to the conclusion that there is no innocence left in the world!  Josh is boinking his girlfriend-yeah, they\'re both too young...sixteen...I guess I\'m just getting old. I look at her and she\'s so sweet and cute... and I just want yell at Josh to \'leave her alone you penis-driven freak\'!  Thank god, that Pat gave me some good advice last year.   He said \'give Josh condoms now and tell him to keep them in his drawer, then when he needs more he can just tell you that he needs more and you will both know he is \'active\' without having to go into an embarassing conversation about it\".   I took this advice and bought the box of condoms for Josh to keep in his drawer. About three months passed...nothing.   Then they had a school dance.   A couple of days later Josh came to me. \"I lost my condoms, can you get me some more?\" \"You   lost    your condoms\"? \"Yeah, I had them in my wallet and Todd and David found them and started throwing them at each other.\" \"Well, these are not cheap you know. I bought them for one thing only...so don\'t go letting your friends play with them again.\" \"OK, I won\'t let my friends play with them again.\"  A week later...\"Can you get me some more condoms?\" \"Why, did your friends   lose    them again?\" \"No...so...can you get me some more condoms?\" \"Sure I\'ll pick up some on my way home from work tomorrow.\"  End of Conversation.    '),(21079,3466070,'09,August,2004','2004-08-09 00:00:00','     Today Aug, 9th was the first day of school... I got some pretty cool teachers like Mr.Housely. He is funny as shit!He cracks a lot of wise ass jokes, but oddly enough I am his favorite student in the class. He said, \"I don\'t know any of you little brats... Wait! I do know one of you...\". He points to me. \"You are the one guy who is always with my little Sarah! You are cool in my book...\" He was like I need to give someone in this class a nick name... No one raised there hands... So he looks at me and says, \"You are now J-Unit...\". So there. I am J-Unit now. I only saw Sarah 2 times today! We had one class together and then saw each other after school ended... I for a while thought she was avoiding me.  Turns out it was just a hectic day and she couldn\'t find me for awhile. It\'s ok though because now I have her schedule. We don\'t have lunch together this year... :-( That really sucks, but we both have an evil bitch of a teacher by the name of Ms.Aronin. She is evil. She asked me if I worshiped Satan then bitched because I pasted out in her class. Everyone else was cool though. I had some home work, but finished it in am matter of minutes. That is all for today... Thanx for your time.    '),(21080,3466070,'06,August,2004','2004-08-06 00:00:00','     If you would kindly go and look under my links to blogs you will see one that is called pillow talk. That is Sarah\'s so go to it. She feels lonely. Thank you for your time...    '),(21081,3466070,'04,August,2004','2004-08-04 00:00:00','     I will start using my blog as a journal once school starts. Seems better then stories.    '),(21082,3466070,'03,August,2004','2004-08-03 00:00:00','     Yup... Today is August 1st... I don\'t wanna go back to school yet. Summer wasn\'t long enough for my taste! Fucken summer school. Took my whole damn vacation! Today I gotta go in and do all the gay shit like pay for a locker that I will probably use once. Only thing I am looking forward to is hangin with Sarah at school and wrestling. Nothing else seems to really appeal to me. Well maybe orchestra. I\'ll think about that one.    '),(21083,3466070,'03,August,2004','2004-08-03 00:00:00','     Today Sarah and I saw The Village. It was not what it was cut out to be... At all! It was a good movie, but I just wasn\'t satisfied with it... You know. They made it sound really good when it was not the best M. Night movie. What else can I say?    '),(21084,3466070,'02,August,2004','2004-08-02 00:00:00','     Well lately I have been hearing a lot about suicide from a lot of people(especially from the punk and emo kids). They all seem to act like they have it so bad with all there issues with there parents and there painstaking lives in there $100,000 houses. I mean really! You don\'t even have it bad. Just because your peeps caught you beaten your monkey doesn\'t mean you have to kill yourself. Then the next problem is how they do it and what happens after. They always cut there wrists and may I say those are always scratches if I ever saw one. Then they turn around and brag about how depressed they are and show off there wounds like there were some kind of new fashion. Really! Wow that is soooo cool... Not! If you really want to die then you would realize you are trying to end your suffering... Then why would you go through the pain of bleeding to death from a slit in your arm? If you really are gonna kill yourself, then shoot yourself in the head. There you will die quick and easy... Just like you wanted. Really you should look at it like Sarah does... \"Be glad you are even alive!\". Every individual has some kinda special characteristic that makes them special. Take this advise and do what you will with it... Don\'t kill your self! Finally I said it. lol.    '),(21085,3466070,'01,August,2004','2004-08-01 00:00:00','     Yup... She is an awesome chick. She can smoke a bong, role a blunt and still have time to go to a hemp fest. That is a busy schedule. Oh yeah... My moms says she loves you Vader. Take care girl.    '),(21086,3466070,'01,August,2004','2004-08-01 00:00:00','     I am working for a man named Rob who owns a restoration clinic for cars like old mustangs. We restore all kinds of cars. I worked for him in the past, but he did some dirty business with my parents(charged them for a bad job twice) so I quite and now they are closing down. I get paid $15 bucks for 3 hours of work. Not bad, but I still get to finish up some of the old projects I never got to finish. It was pretty fun, but Taco Bell will pay better.    '),(21087,1916038,'29,July,2004','2004-07-29 00:00:00','   â€œMummy, Daddy, wake up! Weâ€™re going to the beach today, remember?!â€ Jonah shook his parents by their shoulders. Daddy stirred from his sleep and leaned over to the side table to reach for the clock.  â€œJonah, itâ€™s only 5.30 a.m, son. Itâ€™s too early,â€ Daddy replied groggily.  â€œBut Daddy, I want to see the sunrise,â€ Jonah whined. He continued to shake his Mummy.  â€œJonah? Whatâ€™s wrong darling?â€ Mummy asked him sleepily with one eye opened.  â€œI want to go to the beach now Mummy. I want to catch the sunrise.â€ Jonah urged Mummy to wake up.  â€œOh, alright then. Wait downstairs and let Mummy and Daddy change before we stroll through the beach,â€ Mummy replied as she pushed Daddy off the bed and signaled him to bathe. Jonah excitedly got off the bed and ran downstairs.   He had always loved the sunrise and today he had wanted to catch the sun rising across the waters. Though the drive from his house to the beach was short, he still wanted them to arrive slightly earlier. He waited as patiently as he could and got out the cereal box to eat his breakfast calmly. Just as he was about to finish, he heard his parents coming down the stairs. Like a bolt of lightning, he went to his parents and pulled them to the front door.  â€œLetâ€™s go Mummy, Daddy.â€ He urged them.  â€œJonah, we havenâ€™t even have our breakfast,â€ Mummy scolded him gently.  â€œNevermind Julia. Weâ€™ll have our breakfast later,â€ Daddy replied laughingly. So the trio went out of the door and went to their driveway. It was still dark but Jonah was used to the darkness â€“ so were his parents. Jonah climbed on the seat quickly and rushed his parents to move faster. Seconds later, they were driving through the lane which led to the beach.   When they reached, Jonah left his parentsâ€™ side and straightaway climbed on a large rock facing the sea. He huddled himself together and felt the morning breeze blowing gently across his face. He checked his watch and realized that after all the commotion, it was only 5.45am. He turned his head when Daddy called him.  â€œJonah, your mum and I will be getting our breakfast. You wait here alright? Weâ€™ll get you your breakfast too.â€ Nodding his head, he turned back to watch the sea. He was used to being left alone on the beach to catch sunrise while his parents went away. They had taught him to be independent though he was only eleven. He smelled the crisp air of Oklahoma and felt free. He loved the beach so much that his parents actually bought the house so that he could frequent the beach.   When a tinge of orange crept into the darkness, Jonah kept his eyes on it. Slowly, as if playing hide-and-seek, the glowing sun peeped from behind the sea. The warmth of the sun spread through the beach and pierced through the cold morning breeze, engulfing Jonah with its warmth. Jonah breathed in a fresh of fresh air and pampered himself with the feeling of being engulfed. He saw a few other people strolling together, mostly couples, hand in hand. Laying his head on his arm, he waited for his parents to come with his breakfast and they will be able to stroll the beach together.  But they never came back. Jonah waited and waited but there was no sign of them. It was his uncle who came to pick him up.  â€œJonah? Come here, son,â€ his uncleâ€™s voice rang out and he felt two hands taking his shoulder. He was then embraced by his uncle. He was puzzled as to why his uncle came instead of his parents and why there were tears in his uncleâ€™s eyes. He was also bewildered as to why his uncle was hugging him tightly. What happened?   Jonah still remembered how curious he felt to see his relatives crying. Never have he seen his uncles cry though he had seen his aunties cry. His uncle Johnny led him into a room and he saw two beds occupied but the person on each bed were covered with blanket all the way over their heads. Jonah became puzzled as to why Uncle Johnny showed them to him. He was not afraid of ghost or anything. Was this another prank?   But when Uncle Johnny brought him closer to the beds, he felt his hair stand and his heart felt heavy.  â€œWhat is going on Uncle John?â€ Jonah asked curiously as he looked up to look at Uncle John. Uncle John only stifled a cry before lifting the covers. He felt the ground swayed when he saw his Daddy and Mummy, pale-face, lying on each bed.  â€œMummy? Daddy? Wake up. Why are you sleeping?â€ Jonah asked furiously. He refused to believe or even think that his parents wereâ€¦ not coming back.  â€œWake up! You promised to buy me breakfast remember?\" Jonah all but shouted.  â€œWake up please Mummy, Daddy. Please donâ€™t leave me all alone,â€ Jonah finally cried as he hugged his parentsâ€™ lifeless body.  â€œPlease donâ€™t leave me all alone here. I donâ€™t want to be alone.â€'),(21088,1916038,'28,July,2004','2004-07-28 00:00:00','   15 years had passed since his parents died but at 26 now, he could still remember the tragedy like as if it was yesterday. His parentsâ€™ were killed at a car crashed. They crashed with a drunkard who was driving in a drunken stupor. Life had been hard after that. He still remembered asking for his parents sometimes when he had nightmares at night. But it was always his Uncle John who stood by him. It was also thanks to him that he received proper education and lifestyle. He really owned that old man a lot.   Now that he was a successful shop owner,selling books, he moved out of Uncle Johnâ€™s apartment and bought one of his own. Occasionally he would visit Uncle John, like today. He needed company sometimes. He\'s an old bachelor.  â€œUncle John? Where are you?â€ Jonah called out when he was at the porch.  â€œJonah? Here! Backyard young man!â€ Uncle John shouted. Knowingly, Jonah went to the backyard and saw the dear old man standing under the redbud tree where mistletoes grew from. His Uncle John loves to plant plants.  \"Atleast that will keep his mind off about getting me married,\" Jonah thought to himself. Jonah hurried to Uncle John and pecked his cheek.  \"Glad that oyu could come Jonah. I was getting bored being here alone. Though there is company, just next door. A new neighbour, sexy as hell, Jonah.\" Uncle john smirked. Shaking his head, Jonah sat down under the tree with Uncle John.  \"So how have you been ,Jonah?\" Uncle John asked, getting serious. Jonah could only shrug his shoulders. He couldn\'t tell Uncle John that he still had nightmares. He decided that it was best to keep quiet. The two kept quiet as they gaze across the lawn. Due to the weekly companion by Jonah, he was also able to help mow the lawn.   The silence was shattered by the slam of the door. It came from the new neighbour and seconds later, a figure hurried towards them.  \"John? You ok? I called your home but no one picked it up. What are ..\" her words trailed off. So this is the sexy neighbour Unce John was talking about. Jonah gazed at her up and down. She was wearing long pants and long sleeve shirt but still he could&nbsp;imagine that she has got a great body underneathe.&nbsp;He was instantly attracted to her. The thought of it got him worked up. No wonder old Uncle John gets worked up easily. Poor him.  \"Dana, meet Jonah, my nephew. Jonah, meet Dana.\" Uncle John introduced. Jonah only offered a quick smile before dropping his gaze. In that brief moment, he could see that her eyes were a light shade of green. Her bronze hair cascaded past her shoulders and her body was petite but it curved at the right places. In short, she was beautiful. He looked up at the mistletoes before looking down at his feet again.   The man called Jonah was what she thought as rude. Not even a handshake. But she did not take it into consideration much as she has a soft spot for the man\'s uncle. He\'s so sweet.  \"Well, since you\'re okay, I\'d better go then. See you around John,\" Dana said before retreating back to her home.  \"What da hell was that for, Jonah? Couldn\'t you be more friendly?\" Uncle John repriminded him. He didn\'t know what was wrong with him. Normally he would have be friendly with any lady that chatted up with him or he encountered but with that lady, Dana, beautiful name, his social entiquette just fail him. That lady sure has got some sort of magnetic pull in her.  \"Damn Jonah. What if she never comes back after that encounter? I\'ll be bored to death when you\'re not here,\" Uncle John complained exaggeratedly as he got up to go back to his house, it was getting dark.  Don\'t worry Uncle John. I\'ll make sure she comes back.\"  Jonah thought. Casting another look at Dana\'s house, he slowly trudged back into the house.'),(21089,1916038,'03,August,2004','2004-08-03 00:00:00','   i come home from work ... a little tired but very anxious to see my baby nura ... i open the front door and silently walk in ... i find the house very clean ... all except for a lone sock lying in the living room floor ... i (confused) walk over and pick it up and find that there is another at the foot of the hall ... i go pick it up and find a pair of your jeans and a dark red lacy shirt lying in front of our bedroom door ... i slowly and quietly open our door ... i find you wearing nothing but a black pair of panties and matching bra ... you look slowly and sultry up at me as you slide your hand out of your panties ... you bring your wet fingers up to your mouth and lick them clean before looking deeply into my eyes and saying \"I was just getting warmed up for you, baby.\" ... you smile at me and i smile back ... as i walk towards you your back arches slightly and you spread your legs for me to slide between ... i crawl slowly onto the bed and up between your legs ... and i feel your warmth on my body as you wrap your legs around me ... i lower myself to kiss you softly on the lips ... kissing again feeling your tongue passionately mingle with mine ... i feel my (ahem) getting harder through my pants and on your panties ... you squirm excitedely under me as you feel it too ... we kiss deeper and harder as you we slowly and softly grind together ... i fumble with your bra as we kiss ever more passionately ... finally your bra slides off of you ... i softly kiss your nipples and lick all around your breasts ... feeling them getting harder in my mouth ... your hands are at my waist unbuttoning my pants and slide them down slightly ... you take my pants down some and i slide them completely off ... i also take my shirt off so that i can feel your soft skin against mine ... we kiss passionately and i run my fingers down your sexy body ... i love the feel of my hands on you ... i love u ... i ask you if i can take your panties off ... and without a word you take my hands and slide them to the elastic of your panties and help me slide them off ... \"are you sure bout this ...\" but before i can get the words out you have already pulled me towards you ... i feel so .. so warm inside as i slide slowly into your hot wet pussy (hope word is okay) ... your so tight and as i get deeper and deeper inside of you you gasp and your eyes close slowly ... \"are u okay baby?\"  .... \"Yes Brian, don\'t stop!\" ... i pull out of you slowly and slide in gently loving the feeling as i slide deeper and deeper into you ... feeling your tight pussy sliding around my um ... um ... dick ... dripping wet and so very warm ... you begin to rock with me and i hear you moaning my name softly as i slide in and out of you ... i lean down and kiss and lick your nipples as i slide in and out of you ... feeling so good ... i begin to kiss you deeply, catching the moans before they leave your mouth ... the room is getting hotter and hotter and as it does a little sweat drips off of our bodies ... the sweat runs all over our whole bodies as we are one ... making it easier to slide deeper and deeper inside of you ... as i get a little harder you moan with slight pain but intense joy ... i ask if you want me to stop but you say loudly \"NO, Deeper BRIIIIAAAANNN!\" ... i slide deeper and deeper inside of you ... your moans get louder ... soft moans of \"Nurrrraaaaa....\" leave my lips almost involuntarily ... you arch your back as you moan even louder as your sweet juices run down onto our white sheets ... you pull yourself harder against me as i grow larger inside of you and slide even deeper into your hot pussy ... your moans turn into screams of joy  \"mmmmm Brian more Brian ... Harder ... Deeper ... mmmm Brrrriiiaaannnnnnn!!!\" at this time i feel you release me ... and i feel your hot juices explode inside of you ... as you cum i feel a strange surge of energy and sensitivity slide through my body as i cum deeply inside of you ... we both collapse ... our hot bodies lying side by side ... i hold you tight \"I love you Nura Radford/Harrison!\" ... with a gasp you say \"I love you too.\" ... and we pull each other close ... I lay your head on my chest as you fall asleep in my arms ... this is when i know that i\'m the luckiest guy in the world ... so happy to have you by my side !!! --'),(21090,3585158,'09,June,2004','2004-06-09 00:00:00','   I love men â€“ and I love women too.   I pop my eyes out ogling at the cosmo men, but that didnâ€™t stop me from staring at Angelina Jolie at capacity 24/7...  sometimes I wish I were a man but clearly most of the time I simply marvel at myself being a woman.  But then this wish of being a man just comes from this double standard society that I grew up on, or maybe because when we were all young we were taught to distinctly differenciate a woman from a man, thusâ€¦the world of difference.  I really donâ€™t understand.  Iâ€™m also not a firm believer of religion.  I had this mistake of telling my lack of sexual preference in a partner to a friend and I was looked down as the perforator of the destruction of Soddom and Gommorah. What Iâ€™m saying is that, câ€™mon, this society is still wary of my kind and they are definitely NOT ready to embrace my belief.    Imagine thereâ€™s no genderâ€¦ life would be much easier.  Of course I also think this is not a sexual perversion.  Have you ever really enjoyed talking with a girlfriend that you simply want to kiss her on the lips?  Not a smack but really kiss her?  Just because.  And not because I want to have sex with her.  Come to think of it--- Iâ€™d come out unscathed if I kiss my best male friendâ€¦ but my best girl friend????  As  much as I want a genderless society, I am still respectful of the differences between a man and a woman.  I just want NOT let these differences restrict our actions in any situation and restrict our preferences in any relationship, or be judged  wrongly and be misunderstood in any boy-boy, girl-boy, or girl-girl relationship we establish .  I hate the machismo of men not wanting to get caught in a tight emotional embrace with another man,  or the ultra-feministic view of  this colleague getting all worked up just because a guy suggested that women should be at the back of men in a blockade simulation.   I mean  no matter what this guyâ€™s intentions were, we have BREASTS to protect and their balls are much less likely targets, right?  I wouldnâ€™t waste my energy fighting something that is for my own good.  I am not that stupid.    Maybe some may think that Iâ€™m rebellious because things are not made up my wayâ€¦ not the way I would like things to be.  Yes, I respect everybody elseâ€™s opinion, but that does not and will not stop me from reacting and questioning  why things are as they are.  I still THINK and INSIST that women and men were made for both men and women, no matter what relationship or level of relationship they would want to establish. And no matter what certain feminists would say, If a man offers me a seat, Iâ€™d take that seat offered to me; not because I am a woman but because sitting in a crowded train is much comfortable.â€¦plus Iâ€™d really like to believe that if it was offered, it was out of good intentions.  '),(21091,3752323,'26,June,2004','2004-06-26 00:00:00','   Sundays are the bastards of the week. People like them on Friday but hate them on Saturday. On a Friday, you know that you don\'t have to go to work tomorrow and the day after tomorrow, but on Saturday, when you get up at 12 noon with a headache because of too many bloody maries yesterday, you are thinking \'Oh NO!\', only one more day of cessation of chores, after which, it\'s back to the grinding mill! HEEEEEEELP!!  '),(21092,3752323,'25,June,2004','2004-06-25 00:00:00','   Let\'s say that this guy named A kills another guy named B. Let us also say that A meticulously planned the whole thing and then carried out the plan to the T and was successful in murdering B. Mission Accomplished! The Police catch A and get hold of enough evidence to get a conviction. The judge sentences A to death. Let\'s call the judiciary apparatus C. Now, C is planning to kill A. C has made meticulous preparations to make sure that A dies. Then, C carries out this meticulous plan to the T and is successful in killing A. Pray tell, what\'s the difference between A and C?  	People sometimes say that the mere presence of capital punishment in the judiciary system of a state is a deterrent for people so that less number of murders and killings are committed. However, we can see that capital punishment has not deterred anyone from actually committing a murder because murders are committed in states with capital punishment as rampantly as those without. Therefore, this argument is nullified.  	Also, what I think is that the judiciary should not just vengefully mete out slaps on the wrist (or on the neck, for that matter) and make sure that justice really is served. If justice demands an eye for an eye then what\'s the difference between the civilized world and the Taliban?   	Let\'s say that a person is executed and somewhere in the future, new evidence comes to light that exonerates the person. What does the state do? How does the state make the person come back to life? The state has no right to take what it cannot give back in the event of a mistake committed on its part. If, however, this person were not executed but was kept in a pennitentiary for life (and I mean for life, with no parole), then the state could release them and ask their forgiveness for the time that the state took from them. This is, I think, better than saying sorry to the family of a deceased. 	 	So, it makes one wonder what the difference between capital punishment and murder really is.  '),(21093,3752323,'25,June,2004','2004-06-25 00:00:00','   This was a shocker right now! I saw news from India on NBC 10 news at 4! The news, however, was about the \"Rolling Seer.\" It showed a half-naked guy who has rolled (yes, literally rolled) the length and breadth of the country and hopes to get a visa to Pakistan so that he can roll into our neighbor (now that\'s a metaphor that I don\'t want to laugh at!). Anyway, this shock was somewhat subdued because Indians are traditionally portrayed in the US media as a country of mendicants and snake charmers and this news item only reinforced the fact. They did never show the flight of the GSLV, nor the LCA. They only show these sort of news that reinforce the stereotype.      I did, however, see a news item the other day in Philadelphia Metro that claimed something else. It said, and I quote verbatim, \"The consumer revolution has started in India, what with all the Call Center Workers making such a huge amount of money...\" So what do you say, people, have we graduated from being snake charmers to being call center workers?  '),(21094,3752323,'25,June,2004','2004-06-25 00:00:00','   Hi guys,     This is my first blog, so please bear with me (you are free to bare with me if you are the appropriate sex and age ;-) Anyway, my first blog is actually about why research is so damn important. Research is important because it gives the student money and the professor material to publish. I know, this sounds like the lamest explanation for research, but really, think about it... It\'s sort of a Catch22 situation. If there is no research, then there won\'t be enough students and in turn, there will be no research. Ergo, (this is my word for the day!) research is important to keep research alive!  '),(21095,3752323,'02,August,2004','2004-08-02 00:00:00','   I am totally tired of reading mails that project a holier-than-thou attitude and tell a small fable that is supposed to move us into cherishing our friends/parents/realtives/pets/asses/etc and therefore, I have created this real-life scenario. Read on...  Who\'s poor???  Following is the Holier-than-thou-attitude-wala story of a wealthy father and his wise-beyond-his-years son.  One day a father of a very wealthy family took his son on a trip to the country with the firm purpose of showing his son how poor people live. They spent a couple of days and nights on the farm of what would be considered a very poor family. On their return from their trip, the father asked his son, \"How was the trip?\" \"It was great, Dad.\" \"Did you see how poor people live?\" the father asked. \"Oh yeah,\" said the son. \"So, tell me, what did you learn from the trip?\" asked the father. The son answered: \"I saw that we have one dog and they had four. We have a pool that reaches to the middle of our garden and they have a creek that has no end. We have imported lanterns in our garden and they have the stars at night. Our patio reaches to the front yard and they have the whole horizon. We have a small piece of land to live on and they have fields that go beyond our sight. We have servants who serve us, but they serve others. We buy our food, but they grow theirs. We have walls around our property to protect us, they have friends to protect them.\" The boy\'s father was speechless. Then his son added, \"Thanks, Dad, for showing me how poor we are. \" Isn\'t perspective a wonderful thing? Makes you wonder what would happen if we all gave thanks for everything we have, instead of worrying about what we don\'t have. Appreciate every single thing you have, especially your friends! Please pass this page on to friends and acquaintances to help them refresh their perspective and appreciation. \"Life is too short and friends are too few.\"  Now, look at how the story would have really unfolded if an actually wealthy father had taken his city-bred brat to a village.  One day a father of a very wealthy family took his son on a trip to the country with the purpose of showing his son how the poor and downtrodden live a meaningless and wretched life. They spent a total of two hours in the shit-hole that the farmers called their home. The son started running away from the place as soon as he saw the food that was placed in front of them. The father, who ran behind his son, caught up with him and asked him, \"Now, son, do you see how the poor people live?\" The aghast son replied, \"Yes, dad!\" And he puked a little. The father asked his son, \"So tell me, what did you learn from this trip?\" The son answered, \"I saw that we have two healthy dogs while these farmers have 16 mongrels with attitude problems so bad that no one knows which one will bite you next! We have a swimming pool with clean, chlorinated water, while these wretches have a creek that\'s filled with industrial pollutants. We have imported lanterns in our garden whereas the hopeless excuse that they call a garden hardly HAS any plants! Our patio reaches to the front yard whereas these excuses for humans are still fighting about which land belongs to whom! We buy organic food whereas they grow shit-stuff full of pesticides and fertilizers that they sell on the market to our less-fortunate city brethren! We have walls around our property to protect us, whereas these assholes don\'t and that\'s because they don\'t HAVE anything worth stealing!\" The boy\'s father was speechless! Then his son added, \"Thanks dad, for being rich and not making our family live in such shitty surroundings!  NOW THAT\'S WHAT I CALL PERSPECTIVE!  '),(21096,2965205,'22,May,2004','2004-05-22 00:00:00','       '),(21097,2965205,'21,May,2004','2004-05-21 00:00:00','       '),(21098,2965205,'21,May,2004','2004-05-21 00:00:00','    San Francisco, CA -  An Al Qaeda plot to drive a van into downtown San Francisco and detonate heavy explosives was foiled on Friday by Shawna Jones, the morning dispatcher for U-Haul in the Potrero Hill neighborhood.  According to Jones, Reza Al Hagwani had reserved a U-Haul van for pick-up early Friday morning.  When Al Hagwani arrived at 9am Friday, however, Jones informed him that there was no U-Haul van available.  \"I told the man that we had no vans for him but that we might have one down in San Jose,\" said Jones.  \"I can\'t help it if we don\'t get the vans exactly on the day people needs it.\"  An incensed Al Hagwani stated that he intended to \"write a scathing letter to the management of U-Haul\" protesting his treatment.  \"I cannot believe that even with all the computers at the disposal of these workers, they are unable to deliver me with the van I ordered more than two weeks ago,\" said Al Hagwani.  \"There are many choices to me of rental truck companies, and the next time I will not be calling U-Haul, that is for certain.\"  Al Hagwani did make the one-hour trip to San Jose to inquire about renting a van from that location, but it is not yet known whether a van was available.   '),(21099,2965205,'20,May,2004','2004-05-20 00:00:00','    New York, NY   by Randall Rivers   urlLink      Mr. Rivers    Wassup?!?!   I am writing to you on my brand new Sony laptop, which is real dope. I could barely make enough money at my broke ass job to afford the Internet cafÃ©, but I won\'t even bore you with that. Check out what I did!   Ladies and gentlemen, I have figured out how to make some serious coin on the subway. Let me spell it for you. T-E-E-M-Work. Some glass-half-empty types call me a bum-pimp, but what I am doing is creating a union for the have-nots. Weâ€™ll have benefits, and paychecks, and hookers, but thatâ€™s later, when weâ€™re showing a profit and shit. I\'ll even rent that blowup rat and put it on the 6 train for a day if I find some dingy bum panhandling on the Lexington Avenue line without letting me know about it. No scabs in THIS subway, boy.  All I ask for is 2% of earnings and in return, you\'re a Union Man!  Know what I\'m sayin\'?   urlLink      The 6 Train, otherwise known as the \"Home Office\"    First off, I rounded up all the fellas offering hungry people a sandwich or a drink. You know the ones who say â€œI was homeless, and now Iâ€™m offering food to the hungryâ€ and you think to yourself, â€œwhat a nice manâ€ until you look up and theyâ€™re all scary looking. And Iâ€™m talking about the sandwiches, before you get the wrong idea. So I went around, and I said to them, listen up yo, noone\'s eatin that piece of shit sandwich, not least from a crusty old looking guy like you. At least wrap the shit in a Subway sandwich wrapper or something. Make \'em perceeve that it\'s a tasty fuckin\' sandwich, even if it\'s a raw burger patty, a green pepper slice, and a bunch of ketchup on a coupla old wonder bread slices. And noone wants to take a swig off your 2-liter of RC Cola when they\'re thirsty. Know what I\'m sayin\'? Offer the subway riders a frosty 7-Up or something. A toothpick and a mint wouldnâ€™t hurt either, but thatâ€™s for laterz, when we\'ve established ourselves, gotten a brand name and shit.   Then, I cracked the battery sellers. It\'s long been a standoff between the Asians who sell you the duracells, or the loud toys that break the instant they touch a howlie\'s hands, and God knows what else - and the brothers, who sell the fake duracells which make people say \"aw shit, fuckin\' con artist\" in their brain but their mouths go \"thank you, sir.\" Know what I\'m sayin\'? Anyway, I took one for the Sino-Brotha team by puttinâ€™ em to work with each other. I\'m a marketing man by nature, and my research reveeled that the asians had the access to the real batteries, but that customers preferred dealing with the brothers. Yo, check this out. I got them all together and I said, if you work in teams of two, one handling the cash and the other making the sales pitch, you\'re bound to double your money. It\'s working real well so far; sales are through the roof and we\'ve discontinued selling the robots that make a lot of noise. Fuckin\' time wasters, those things. We gave em to some little kids charity, and a bunch of letters from the charity have been coming by registered mail. Iâ€™m assuming theyâ€™re just thank-you letters, so Iâ€™ve been throwinâ€™ em out. Thank you\'s is not why I am in this business.   So now that I got the food folks, and the battery sellers all lined up, I axed myself, what next?  The singers! So I had an open tryout last week. I called it Subway Idle. Man I wish I got that shit on tape. If it wasn\'t horrible, it was hilarious. The winners would get free outfits, a free backup boombox, and a charming little 5 year old girl (my half-sister\'s kid) for a week to solicit bigger donations from the straphangers. It went over reaaaaal well. This beatbox kid, real fat, former M&M seller gone \"vocal\" as he likes to call it, beat out this Hasidic guy in a deputy\'s uniform and this Hipster for the title. It was b-e-a-YOO-tiful man!    urlLink      The runner-up, absorbing the fact that he\'d lost    Iâ€™m presentinâ€™ my bizness plan to the MTA next week. I call it â€œfood, commerce, entertainment â€“ the subway ride of the 21st centurieâ€. I figure if I throw the MTA a bone or two, that second avenue line will get done quicker and Iâ€™ll have a new â€œlineâ€ of business, know what Iâ€™m sayinâ€™? If approved, look for the peeps in the dark red t-shirts entertaininâ€™, sellinâ€™ and reprazentinâ€™ on a car near you!   Peace, yo.   '),(21100,2965205,'19,May,2004','2004-05-19 00:00:00','    Provo, UT  -  In what local promoters are labelling a \"communication breakdown\", Hollywood-area Tribute Band \"Dixie Chicks With Dicks\" played a 7-minute show at the Kiwanis Club in Provo, Utah last Saturday evening.  At 8:15 pm, the 3-piece band took the stage - a makeshift wooden platform placed on top of the club\'s stacked up milk crates.  As they launched full-throttle into the Dixie Chicks\' 2002 hit \"Travelin\' Soldier\", the audience roared its approval.  By 8:20, however, a general sense of confusion overtook the audience as they heard lead singer Tom \"Emily\" Schatz say \"thanks for comin\' Provo!\" in a distinctly baritone, un-\"Chick\"-like tone of voice.     urlLink       A recent picture of the band in happier times    Closer inspection revealed plenty: three days\' worth of chin stubble on all three band members, obviously dark brown roots revealing a crappy blonde dye job on the third \"chick\", and a mysterious bulge in the tall one\'s pants. According to eyewitnesses, by 8:22 the mood had soured: those who were not crying uncontrollably were yelling at the band to \"leave the darn stage.\"  To its credit, the band honored those requests, leaving immediately.  \"In retrospect,\" said Schatz, \"we probably should have avoided Kiwanis Clubs and the state of Utah in general.  It just seemed so convenient to stop there on the way to Vegas.\"    Promoters vowed to recoup the $96 they had wired the tribute band\'s manager three hours before the show.   '),(21101,2965205,'19,May,2004','2004-05-19 00:00:00','    New York, NY  -  In a long-anticipated study commissioned by New York City\'s Center for Homeless Research, the findings have been announced:  New York City\'s homeless are lazier than they have ever been.  \"I remember when the homeless of New York had, I don\'t know, a certain je ne sais quoi\" said Jennifer Skately, 24, of Washington Heights.  \"They\'d at least crack wise, do a little dance, or just take pride in earning the 21 cents you\'d half-heartedly fish out of your pocket.  Now it seems like they expect you to give \'em a buck for nothing else than being homeless.\"   urlLink      Sittin\' around, doin\' nothin\' - as usual    Quantitatively, it does appear that Ms. Skately\'s anecdote can be supported.  In a broad range of categories, it appears that the homeless are trying less hard than ever.  All-time low marks were earned in entertainment, creativity, and good humor.  Marks for aggravation, conniving, and public urination have bottomed out as well.  The only statistic in which they receive a better mark is \"sitting around and doing absolutely fuck-all.\"     urlLink      Get up and get a job? Not this guy!     Henry LaFontaine, spokesman for the Center, chimes in:  \"It seems that after 9/11, the average vagrant\'s desire to improve his or her lot in life just vanished.  Some may chalk it up to a crushing loss of hope, attributable to prolonged homelessness.  On the other hand, maybe since garbage pickup has been reduced to twice a week they\'re all better-fed these days?  I really couldn\'t tell you.  Nonetheless, I am afraid that the homeless, more than ever, just seem shiftless, bored, and unbelievably lazy.\"  I spoke with \"Joe\" (not his real name) to find out his side of the story.  \"Man, you go on the Subway and they look at you like you done [had intercourse with] their kids.  It\'s tough to find a [friendly] face.  And now, with the new mayor not making us go to a shelter at night, it just seems like the other bums got all lazy and [what have you].\"  Joe then leaned back against the wall and got comfortable.    Joe, comfortable    urlLink   '),(21102,2965205,'19,May,2004','2004-05-19 00:00:00','    VATICAN CITY  and  GREEN BAY, WI , May 18th  - After 2004 years of praying to our Lord Jesus Christ, humanity had reason to believe that, well, Jesus was a figment of someone\'s imagination.   No longer.   In a celebrity sponsorship scheme for the ages, the Vatican has announced a partnership program with Klondike Ice Cream Company. After securing exclusive rights to the real-life Jesus for 2005, the Vatican will work closely with both Heaven and Klondike to improve the image of the Church as well as the Ice Cream company\'s bottom line.   The sponsorship, still in its development phase, looks to incorporate Jesus into a variety of advertisements from TV commercials to print and banner ads. With a rigorous press junket planned shortly after Easter 2005, Jesus is expected to be on Earth for two months before prior obligations require His return to Heaven.    urlLink      Jesus, Our Lord and Savior,  Contemplating a Velvety- Smooth Klondike Bar    \"We\'re tremendously excited to welcome Jesus to the Klondike Team,\" Marvin Watson, VP of Marketing, said in a recent press conference. \"With the recent spate of Jesus-positive media, we felt that He was a perfect complement to our brand. We\'re excited to see, for the first time, not only what Jesus actually looks like, but what He looks like earning Klondike Bars: you know, doing the chicken dance, rapping in Polish, or my personal favorite, telling off-color jokes.\"   When asked for comment, Cardinal Abruzzi, Vatican spokesman, stated that \"We felt that Catholicism had taken a real hit in the last year with the various scandals. As an institution, the Church was rating only slighter higher than Janet Jackson\'s new  Damila Jo  album. Frankly, we\'d never really lived down the whole \"sun revolving around the earth\" issue from a few centuries ago. What better way to score some points by bringing back the Son of God? And what could be better than to combine His Holiness with a milk chocolate shell and creamy vanilla ice cream Klondike Bar? I can\'t reveal too much, but what I can tell you is that you better be prepared for the hilarity that is Jesus doing jumping jacks and screaming at the top of His lungs.\"    urlLink      If all goes according to plan,  Jesus will be enjoying  one of these shortly    Initial word has Jesus attending the premiere of \"Klondike: The Very Last Temptation of Christ\" next May at Mann\'s Chinese Theatre in Hollywood as the kick-off to the marketing campaign. There, it is expected that Jesus Himself will make a short \"comeback\" speech, sit and watch the movie, and then do jumping jacks and scream at the top of His lungs in order to earn His first Klondike Bar in two millenia. He is then expected to make His way over to the premiere party.   As of press time, Jesus was unavailable for comment.    The New Marketing Campaign from Klondike    urlLink   '),(21103,2965205,'19,May,2004','2004-05-19 00:00:00','    urlLink    &nbsp; urlLink    In a surprise move, Ralph Nader announced yesterday that he will not pursue a presidential candidacy in this year\'s election, citing \"four really heinous years of Bush\" as the major impetus for his decision.  \"Man, I really fucked things up last time, didn\'t I?  I am truly sorry about that,\" Nader told Meet The Press\' Tim Russert Sunday.  \"I don\'t even know what I was thinking.  You guys must think I\'m a real shmuck.\"  Nader concluded: \"Frankly, I just did it for a hoot.  I didn\'t think so many people would vote for me.  But I definitely learned my lesson this time.\"   '),(21104,2965205,'19,May,2004','2004-05-19 00:00:00','    urlLink    Laura Bush demonstrates proper fisting technique for Jay Leno. &nbsp; urlLink   '),(21105,2965205,'19,May,2004','2004-05-19 00:00:00','    urlLink    &nbsp; urlLink      Washington, DC  - On this week\'s \"Meet the Press,\" when asked how he would respond to demands from Iraq\'s foremost Shi\'i cleric Ali Al Sistani that US forces withdraw from the holy cities of Najaf and Karbala, Defense Secretary Donald Rumsfeld responded, \"I wish he would just quit his bitching.\"  Rumsfeld, who recently returned from a tour of the Abu Ghraib prison facility in Iraq, went on to say that \"The prisoners at Abu Ghraib should also just shut their traps.\"   '),(21106,2965205,'18,May,2004','2004-05-18 00:00:00','    Bloomington, IN -  Mass pandemonium was unleashed last Tuesday when an area toddler grabbed for and dropped a can of whoopass at the Kroger Supermarket on 10th Street.  Bystanders report that the can emitted a loud hissing sound before exploding into a black cloud of bitterness and rage on Aisle 3.   urlLink     Cans of whoopass. \"Justin\'s always grabbing stuff off the shelves and throwing them into the cart when I\'m not looking,\" said Marnie Wolff, the toddler\'s mother.  \"You can bet I gave him a big smack this time.\"  Assistant Manager Joe Shifflett remembers rushing to the scene and attempting to contain the spill, but says, \"I just kind of blacked out after that.\"  According to witnesses, Shifflett ripped his shirt off and began hurling cans of tomatoes at shoppers trapped in Aisle 3.  75-year-old Judith Martin, who was caught in the middle of the aisle, dropped to her knees and was heard chanting in a foreign language.  Bystander Matthew Pinole said, \"It sounded like Latin maybe.  Remember that kid in The Exorcist?  It sounded like that.\"   '),(21107,2965205,'18,May,2004','2004-05-18 00:00:00','    EDITORIAL by Freida Forlane  My favorite restaurant in town has this one very good looking, but very dum waiter. So dum, in fact, that when you ask him to bring you a glass of water, he asks if you want it flat or sparkling. What the fuck, dude? I want it COLD. I don\'t even know what he is asking me.  Anyway, so I think this waiter has a crush on me because it seems like I get him everytime I go into the restaurant. And it\'s fine, we usually recognize each other and say hello. If it\'s not busy, he\'ll even talk to me for a couple minutes which is ok but sometimes his breth reeks of whiskie or anchovies or something. Maybe it\'s something else even, I don\'t know.  Well, all that is well and good but a couple of weeks ago I went there with this boy I was on a date with. We both got the lazanya and it was truely amazing. As always. Well, mine was anyways. The boy I was with complaned that his was burned. And it was, real bad. So we asked our waiter to please return it, and he said it was not possible because it was too busy and there were too many other people who wanted to have the lazanya.  Well this put us in a bad mood. What kind of customer service is that? So I told him I wanted to see the manager. He said something like fan ghoul or something under his breath but I did not know what he said. Besides, I wanted to make sure we spoke with the manager.  I told the manager that we did not appreciate the waiter\'s attitude and that his breath was not very good, either. I said that I am a regular customer and that I expected him to be fired. The manager said he would look into it, and would my date like something else to eat? He said no, he wasn\'t hungry anyway since we got like 2 loaves of bread before dinner and he wasn\'t hungry anymore. I did not see the waiter anymore and I assumed he was let go.  So, anyway, I go back there last Thursday, and the jerk is still waiting tables there! What\'s more, he did not even look like anyone had yelled at him yet. And he shot me the dirtiest look.  What should I do?   '),(21108,2965205,'18,May,2004','2004-05-18 00:00:00','    Listen up, this shit is wack.  Selling M&Ms is no child\'s play.  I\'d have to sell like 200 packages a day - and it had best be 100 of peanut and 100 of plain.  Noone liked the plain though, so my backpack got kinda heavy.  Then I got kinda heavy.  Let me tell you, being a sweaty 12 year old kid on the subway, selling candy - even the little Chinese ladies selling fake duracells were more popular than me.  So we figured a way around it - We\'d blow off selling the candy, eat like 30 of them and get reaaaal full.  Then we figured out a nice little tradeoff - trade like 50 packs of candy for a couple of nice stinky buds.  Then we\'d smoke the buds, eat a few more packs, and then we\'d need to go home.  But momma wasn\'t going to be too happy about that, you see, so we\'d end up just pulling purses out of old ladeeez bags at the Union Sqaure Station.  That would usually get us a lot more than the money we needed for the candy, so we\'d either get more weed or we\'d go eat dollar burgers at this one place.  I got so fat I had to go to fat kids camp once.  Let me tell you, fat kids camp should have been called \"fat white kids camp.\"  Everyone asked me to do the beat box noises and shit.  That shit was embeeeerassing.   '),(21109,2965205,'18,May,2004','2004-05-18 00:00:00','    By Anonymous   So I am probably the only white guy in the United States to work at a Dunkin Donuts. Sure, it means a little good natured ribbing by my darker-skinned friends, but it\'s worth it. I work with two guys who are named Apu, no joke. Every time I so much think of the phrase \"grape squishy\" I laugh out loud. Eating day-old glazeds before they get tossed is pretty cool, too.   I think I\'ve proven my worth over the short haul ... I\'ve taught them how to not reuse the grounds three times per pot of coffee, so that the coffee we serve is actually pretty strong. I\'ve also educated them on how asking for milk and sugar doesn\'t mean a liter of milk and 9 heaping tablespoons of sugar. I have also given them free tutorials on American slang, nothing offensive of course. I\'ve even gotten to them to say \"holy cow\", which if you\'re not Indian, is pretty funny. At least they laugh a lot.   Anyway, so how did I keep my job?   This one night a few months ago, these two Pakistani dudes rolled in, all cocky and shit. This was not an evening to even so much mumble the word \"Kashmir\" for fear of an international incident. Anyway, one of them wanted a Coolatta, which was weird since it was so cold outside. I was so new, I barely knew where the employee shitter was, know what I\'m talking about? So you can imagine the scare I had in this totally-out-of-my-league request.   A few minutes later, when I clearly had proven I didn\'t know what I was doing, one of them was all \"come on now, we are not wanting to be long here\" in that stilted accent you all know and love.   I looked them right in the eye, and said \"hey listen, I\'m new here and I\'d appreciate it if you\'d cut me some slack.\"   This set them OFF.   In what I later came to find out was a Pakistani gang chant, one of them took out his wallet and threw a five dollar bill at me. He said, â€œyou will be dancing for me, my bitchâ€. Corporate creed or not, I was nobodyâ€™s bitch.   So I went in the back, and after knocking down a stack of styrofoam cups all over, finally got Apu #2 out of the toilet early, had him give me a quick tutorial on how to make Coolattas, and I went back. Despite the jeering from the customers, I worked on it and gave it the olâ€™ college try.   When finished, I asked the guy if he wanted any whipped cream. He said, yes of course.   So I went in the back, and proceeded to rub one out in the guyâ€™s drink.   The owner - Apu #1 -caught me! He\'s this really old guy, and I was totally fucked. So I threw myself at his mercy, and somewhere in my rambling it came out that I was just getting back at these Paki fuckers.   The owner laughed and laughed. \"Take it to those assholes\" he said. I didn\'t even know he knew the word asshole. He even added a goober to the drink. We high-fived, albeit very awkwardly.   We served it, asked the asshole who threw the fiver at me to give it a try, and he loved it! In fact, he liked it so much that he had his buddy try it. He loved it too! Gulp!!   I got promoted to assistant manager on the spot.   Next week I teach them how to add all sorts of â€œthingsâ€ to the donuts. It should be a grand time.   '),(21110,2965205,'18,May,2004','2004-05-18 00:00:00','    New York, NY -  The East Village was the scene Friday night of a pitched battle between rival urban tribes for rights to local watering hole \"O\'Malley\'s.\"  According to bar owner Jay O\'Malley, the bar was about half full when a large group of \"young professionals\" in their early 30\'s came in and occupied the back half of the bar.  \"We\'ve got some pool tables and a few couches back there,\" said O\'Malley.  \"It gets crowded fast - it only fits about 15 people or so.\"  O\'Malley reported that shortly after the first group entered, another group arrived and proceeded to the back of the bar.  A few minutes later, O\'Malley heard shouting and rushed to find out what was happening.  \"I couldn\'t believe it,\" said O\'Malley.  \"There were all these yuppies standing there yelling and shoving each other and popcorn all over the pool table.\"  Debbie Weisman, who was part of the original group occupying the back of the bar, says that they were celebrating their friend Josh\'s birthday and had been emailing each other all day about a \"pool marathon\" at O\'Malley\'s.  \"We are kind of like the group on \'Friends\' - we just do everything together,\" said Weisman.  \"We\'re practically family, you know?  We used to hang out at Sushi Samba, but when a few of us moved to the Village and found O\'Malley\'s, we decided to make this our new place.  Then these total jerks came in and started hassling us.\"  Christopher Boatwright, who arrived with the second party, disagrees with Weisman\'s assessment.  \"We\'ve been coming to O\'Malley\'s for five years, ever since we all moved here after business school.  We are here almost every Friday night shooting pool and catching up.  Then these assholes showed up and took our space.\"  O\'Malley says that he told both groups they could either cool down or leave, but neither side agreed to give up seats for the other group, so he told them he was going to call the police.  At that point both groups left the bar.  O\'Malley remembers the days when people came to his bar \"on dates\" or \"with a couple of people, max.\"  \"I think all this \'urban tribe\' stuff is bad for business.  You get enough of these people together and they think they\'re entitled to something.\"   '),(21111,2965205,'18,May,2004','2004-05-18 00:00:00','    San Francisco, CA -  In what some are calling the inevitable aftermath of the landmark decision to allow gay couples to marry in the city of San Francisco, spokespeople for Mayor Gavin Newsome are now reporting that divorce filings have logjammed the city\'s registrar\'s office for the second week in a row.  According to one source, demand for divorce paperwork is now greater than demand for marriage licenses among the gay community.  \"I just thought it would be fun if Mike and I got married with everyone else,\" said San Francisco real estate agent Jim Breyer, who filed for divorce on Monday.  \"We\'d only been dating for a couple of weeks, but it seemed like we really had a connection.  Then I saw him at Metro, dancing with his shirt off sandwiched between two other guys.  I knew right then it wasn\'t going to work.\"   '),(21112,2965205,'18,May,2004','2004-05-18 00:00:00','    Washington, DC -  At a surprise press conference this morning at the White House, President George Bush told a shocked press corps that he had renounced Christianity and had \"embraced the teachings of the Kabbalah.\"  Dressed for the occasion in a white track suit, Bush explained his unexpected conversion by saying that he felt that Christianity was \"limiting\" and that \"some Kabbalists believe that they are the inheritors of wisdom handed down from the time of the Biblical prophets.  I know now that I am in fact one of these prophets, and am thus uniquely able to unite this country and the world.\"  Democratic presidential candidate John Kerry was quick to dismiss the President\'s surprise conversion as \"an election-year ploy to detract attention from the state of the economy and the quagmire in Iraq.\"  When questioned whether his conversion was in fact a marketing gimmick to boost his sagging numbers and attract campaign contributions from Hollywood, Bush pointed out that there are many alternative spellings of Kabbalah, the two most common being Kabbalah and Qabalah.  \"Sometimes you\'ll also see Cabala, Qaballah, Qabala, and Kaballah,\" he added.   '),(21113,2965205,'18,May,2004','2004-05-18 00:00:00','    By Molly Wentworth  The other day I was at Lot 21 doing a quick key hit in the bathroom stall when I heard my old roommate Joyce talking to another girl about me.  She said \"I just saw a friend from college at the bar - I hope she didn\'t see me.  Major addict.  I\'ve heard she\'s turned into a total coke whore.\"  Joyce, first of all, that was so unfair.  I totally buy my own drugs.  That\'s an incredibly important distinction.  Second of all, I also don\'t think it\'s fair when you say I\'m an \"addict.\"  The difference is that an addict  needs  drugs, while I simply  like  drugs.  This is why I prefer the term \"aficionado\" to \"addict.\"  I don\'t even have an addictive personality.  Do you remember our junior year in college when I quit smoking for 2 months?  That wasn\'t even hard for me.  I\'ve also just switched from vodka to red wine.  Wine is full of antioxidants and hardly even gives me a buzz right away.  Just because someone likes to do something doesn\'t make them an addict.  I recall you getting up every morning to go running.  How would you like it if I called you an \"exercise whore?\"  I think that would hurt your feelings.  Anyway, I hope you\'re happy.  I am totally losing sleep over what you said.  I had to take two Ambien last night just to get 4 solid hours, which might not seem like a big deal to you, but now I\'m almost out and the Web site where I get them just got shut down.  Thanks a lot!   '),(21114,2965205,'18,May,2004','2004-05-18 00:00:00','    By Mike Itskovitz  OK guys, I know you\'re with me.  I just can\'t stand it when my girlfriend talks about being on the rag.  The last thing I want to hear when she gets in the car is \"these cramps are killing me!\"  Talk about shrinkage.  Not to mention the fact that I drive all the way across town to pick her up and then she has to tell me her \"flow\" is \"really heavy.\"  I definitely don\'t want to know that she is wearing a Tampon AND a pad!  I\'m all for women\'s lib.  I work with women and I got no problem with that.  I understand that they get their period and there\'s nothing I can do about that.  I just don\'t want to hear about it, especially on a date, OK?  In fact, if you\'re on your period, why don\'t we just take the night off.  Take a Midol and call me when you\'re back in the saddle.   '),(21115,2965205,'18,May,2004','2004-05-18 00:00:00','    By Paris Hilton  You know that saying - \"you can never be too rich or too thin?\"  Well I think that\'s not at all true.  For example, just last month I ate a ton of Twizzlers - I can\'t get enough of them! and I put on 3 pounds.  Just like that!  Can you fucking believe it?  I am purging this week so don\'t worry about it, but it was totally scary.  As for being rich, I don\'t think it\'s fair that people are jealous of me because I never have to work a day in my life.  Which by the way is also non-true.  I am a professional personality and it takes a lot of work to do that.  I am actually a pretty shy person and it\'s not easy being \"on\" all the time for the cameras and whatnot.  So that also proves my point.  Anyway, don\'t think the grass is always greener on the other side of the street.  When people say that I just think: \"walk a mile in my shoes please!\"  Between the TV show and parties my life can get really hectic.  So just please stop judging me.  I am who I am and being rich and thin isn\'t always the key to happiness.  Of course I wouldn\'t want to be poor and fat but then who would?  I guess the lesson is that poor fat people are sometimes unhappy, just like really thin, really rich people.   '),(21116,2965205,'18,May,2004','2004-05-18 00:00:00','       '),(21117,2965205,'18,May,2004','2004-05-18 00:00:00','    By Brian Sack  You bother me somehow. Perhaps itâ€™s that you litter. Or that you walk in front of me too slow. Maybe youâ€™ve come darting out of a doorway oblivious to the fact there are people on the sidewalk outside. Maybe you spat in my proximity. Or coughed or sneezed without covering your mouth. Maybe you walked diagonally or veered to the left instead of the right. Whatever you did, it bothered me.   It could be the way you stand in front of moving traffic with the attitude. Possibly it was the way you pushed through me to get to the bar. It could be your obsession with designer labels. Or the desperate cry for attention you call fashion sense. It could be that you didnâ€™t say thank you when one was warranted. Or that you failed to say please. Suffice to say, it bothered me.   It may have been the pantyhose on the head with the sideways baseball cap. Maybe itâ€™s because you do not belong in a shirt that exposes your ample midriff. Perhaps you called your son a loser in front of the whole subway. It might have been that you drew a swastika on a wall. Or that when you draw a swastika on a wall, you do it backwards. Rest assured, you bothered me.   Perhaps it was when you cut the line because you didnâ€™t bother to notice there was one. Or it was the wearing-sunglasses-indoors thing. Itâ€™s bothersome to hear your phone ringing in the nice restaurant. The kind of restaurant where phones shouldnâ€™t ring. If it wasnâ€™t that, it was the way you were shouting epithets to your friends on the subway platform. Or how you picked up some effete guyâ€™s dollar that he dropped and insisted it was yours.   It could be the hissyfit you threw at the deli because the slices werenâ€™t thick enough. Or that youâ€™re on Prozac yet taking comedy classes. Maybe it was because I could see your penis. Thereâ€™s a chance it was the lit cigarette you tossed on the ground. Or the open-mouth chomping of the gum, complete with snapping noises. Maybe itâ€™s the rainbow flag waving. Or that you have six kids and no spouse. Perhaps itâ€™s the fact that you couldnâ€™t spell, even if the spelling Gestapo were threatening to ship you off to Spellschwitz. It could be the thoughtless placement of your shopping cart. Or your misuse of the apostrophe. It bothers me.   Maybe I took offense to the way you called your significant other a whore. Or it could have been when you interrupted my conversation to ask for money. The way you decided that jumping in front of me was the proper leaflet-delivery method. Or the issue could be you choosing to ignore my having right-of-way. Perhaps itâ€™s your habit of spray-painting your creative impulse or politics on every surface. And putting your face in your own billboard advertisementâ€¦ that bothers me.   It could be that youâ€™re rich and make sure everyone knows it. Or that youâ€™re the laziest person Iâ€™ve ever seen. Perhaps itâ€™s because youâ€™re completely unemployable. Or it might be that youâ€™re an arrogant sociopath. Thereâ€™s a good chance it was because you donâ€™t get sarcasm. Or because you swear in front of children. It might have been when you didnâ€™t use gloves to make my sandwich. And you kept touching your forehead. Maybe itâ€™s because you held the subway doors open so you could continue chatting with your friend. Trust me, you bothered me.   It certainly may have been the puddle of poodle urine you allowed to happen in front of my apartment. Or the way you entered the subway car before we exited. Or that you sleep all day. Letting the door close in my face bothered me too. As did the riding of the skateboard down the sidewalk full of people. The 300 Gigawatt car stereo system bothered me at 4:30 in the morning. As did your business-killing smoking ban. The way you lie about yourself in your online personal ad bothers me. As does the fact you use online personals.   It could be the way you talk about therapy like itâ€™s normal. Maybe itâ€™s your insistence that everyone needs therapy. Or your belief that everyone is gay. Perhaps itâ€™s because youâ€™re gay, lesbian, trans-gendered or straight. Maybe itâ€™s that you drive a Hummer. You bother me.   It could be your liberalism or conservatism. Your blind patriotism and your contempt of the homeland bothers me. It could be that youâ€™re obese and ordering a venti mocha. Or it might just bother me that you make up new words when â€œsmall, medium and largeâ€ would suffice. Or it was how you blatantly stole my cab. Perhaps your clothes being filthy bothered me.   Itâ€™s possibly your lewd behavior. Your lousy taste in music. Your failed hygiene. Or how you blanket the entrance to the building with smoke. Maybe itâ€™s that youâ€™re selling drugs in my neighborhood. Or the way you give money to anyone feigning homelessness. Or how you feel the need to fondle your girlfriend so all can see. All of that bothers me.   Your racism bothers me. As does your multi-culturalism. And the fact that you didnâ€™t bother to rsvp to our wedding even though the stamp was provided. It could be that youâ€™re not competent at what you do. Or that you complain every day to the superintendent about fingerprints in the elevator and menus under the door. Telling people youâ€™re an actor when youâ€™ve never made a penny acting bothers me. As does the fact that you canâ€™t seem to run the Post Office.   Your junk email bothers me. The fact that you think Iâ€™m stupid enough to provide a Nigerian with my bank account information bothers me. And you can imagine how annoyed I get when you try to kill me because I think differently than you do. I also disapprove of you robbing banks and your adultery. Really, it bothers me.   I donâ€™t like your yipping little dog. Or that you became famous and aloof. It very well might be your purple hair. Or the tattoos. Perhaps itâ€™s that youâ€™re holding Hillaryâ€™s book. Or you talk too loud when itâ€™s not warranted. Maybe itâ€™s because youâ€™re an underpaid, undereducated bureaucrat and giving me grief. Or that it takes you 25 minutes to parallel park. That just bothers me.   Perhaps itâ€™s that youâ€™re 30 and still have posters taped to the wall. Or that youâ€™re obsessed with sex. Perhaps itâ€™s because you watch Showtime when HBO is so much better. Or that you honestly believe the earth is only 10,000 years old. It might be your atheism. It could be because youâ€™re a militant vegan. Or that you donâ€™t listen. Or held up the line. Your gum is on my shoe. You dinged my car. Your cable access show sucks. You canâ€™t speak English. Youâ€™re wishy-washy. You wear too much perfume. You try too hard. Youâ€™re paying with a check. You waste money at deli ATMs with $2.50 service fees. You hate your parents. You only date Jewish girls. You always want to share appetizers. You like jazz. You press the elevator buttons as if that will make it come any faster. You live in a 5th floor walk-up. You declared bankruptcy. You think youâ€™re so great. Youâ€™re pretentious. Youâ€™re weak. You show off at the gym. You pitch your sob story to a captive audience in the subway. You donâ€™t know the difference. You donâ€™t tip. You tip too much.   I canâ€™t put my finger on it, but something about you bothers me.   '),(21118,2965205,'18,May,2004','2004-05-18 00:00:00','    1) Twirling ass hair 2) Squeezing monster zit 3) Googling self 4) Licking orange cheetos from fingers 5) Cookin\' up 6) Not waiting for him to call 7) Recovering from liposuction 8) Reading superb new translation of \"Swann\'s Way\" 9) Humming cell phone tune 10) Plotting our revenge   '),(21119,2965205,'18,May,2004','2004-05-18 00:00:00','       '),(21120,2965205,'18,May,2004','2004-05-18 00:00:00','       '),(21121,2965205,'18,May,2004','2004-05-18 00:00:00','       '),(21122,2965205,'18,May,2004','2004-05-18 00:00:00','       '),(21123,2965205,'18,May,2004','2004-05-18 00:00:00','    By Leah Greenberg  Three weeks ago I  specifically   asked Mr. Gonzalez to be around on Sunday May 16th to help me with a delivery.  I needed him to help carry up 6 boxes of packing bubbles to my parent\'s storage apartment in our Upper West Side building.    My school is having a science project week and I needed these bubbles for my super-secret project.  Last year, the winning project got the winner into Princeton.  I mean, it might have had something to do with the fact that that kid was a total dork, but I need to get into an Ivy league school and my grades alone is not going to get me in there. Don\'t even ask me about my S.A.T.\'s.  So, I reminded the doorman again and again, to the point where he would say \"Audrey, when is that delivery again\" and start laughing in that fat little man sort of way that really annoys me.  Plus I don\'t like the fact that he\'s sometimes checking out my ass in the mirror.  My mother says not to worry, that in the Puerto Rican culture it is considered a compliment or something.  Anyway, so on Sunday, the buzzer rings and it\'s the delivery guys.  They\'re downstairs. I go downstairs, a little hung over from the night before, so I\'m already cranky, and Mr. Gonzalez is not there!  I ask the other guy, who is 82 years old (I think he played Blue in Old School but I am not sure) and he says that Hector had to take his daughter to Sunday School.  What the fuck?  Sunday school?   '),(21124,2965205,'23,May,2004','2004-05-23 00:00:00','    Jersey City, NJ  - For years, what have restaurants Houston\'s, Chili\'s, and Bennigan\'s shared in common?  Overpriced food, overly friendly service, and - time and time again - getting their lunch handed to them by 800-pound gorilla TGI Friday\'s.    Talk to management of any of the three completely interchangeable, instantly forgettable restaurants, and you hear a similar refrain: enough was enough.  Now they\'re doing something about it.  In what experts are calling the largest merger ever in the food and beverage industry, Houston\'s, Chili\'s, and Bennigan\'s will close their doors immediately and by early August, reopen under the new \"Hooligans\" brand name.     urlLink      The new, \"no-nonsense\" logo    The new restaurant, in an attempt to shed the humdrum, dime-a-dozen personality of its three predecessors, will be known for its value-priced menu, served to customers by an uncaring, belligerent staff.    \"We\'d just gotten fed up with being the red-headed stepchild [in the restaurant business]\", commented Frank diPizio, COO, of Bennigan\'s Inc.  \"As it turns out, I ran into some executives from Chili\'s and Houston\'s at a recent trade show.  Maybe we\'d had a little too much to drink, but after a few hours of griping at the hotel bar, we said we had to do something.  Friday\'s was just kicking our asses. And \'Hooligans\', I am proud to say, is the result of not wanting to be kicked around anymore.\"  The new corporate management team is interviewing regional managers now and hopes to have staff in place at each Hooligans location by mid-July.  A three-week training course will ensue.  After a team-building trip to get their heads shaved, prospective waiters will then be trained in such areas as staredowns, smashing people over the head cleanly with beer bottles, and affecting indecipherable cockney accents.  They\'ll also learn how to carry a large tray of food while blindingly drunk.  Not everyone is convinced.  \"If they think they can con the American public into eating greasy fried food in a sterile setting with a staff that couldn\'t care less about them, they\'ve got another thing coming\" said Brian Murphy, General Manager of the Appleby\'s on Main St. in Morristown, NJ.     urlLink      Meet the new waitstaff at Hooligans\' flagship location in Times Square, NYC!    One woman seemed to be in favor of the new idea.  \"It sounds great!\", offers Pam Snodgrass of Bergen County, NJ, mother of 4 boys.  \"Not only do I get to save a couple bucks, but if any of my own little \'hooligans\' acts up, I can ask for an assist from one of the waiters.  Personally, I\'m excited.\"  Company officials expect that just under 1500 restaurants will be opened under the Hooligans name in August, with another 100 opening by year\'s end.   '),(21125,2965205,'02,June,2004','2004-06-02 00:00:00','    Vatican City -  In a surprising blow to Christianity and religious faithful worldwide, the Vatican is now reporting that the second coming of Jesus Christ \"already came and went.\"  According to a press release issued by the Holy See on Sunday, Jesus Christ \"returned to this world to spread His message of peace and love, again.\"  The Vatican reports that the Lord appeared in a number of cities across the United States but spent most of His time at the Dunkin Donuts in Crawford, Texas.  Jeff Mack, assistant manager at the Dunkin Donuts in Crawford, said: \"I could tell the guy was different, even though he had a cell phone and wore jeans and a T-shirt like most people around here.  He asked me once if I knew how he could get onto the President\'s ranch, but I told him the security\'s pretty tight up there.  Guess he gave up.\"  Mack added: \"He liked crullers mostly.  Also eclairs.  Never gave us any problems.  Sorry to lose the business frankly.\"  The White House confirms that it received several phone calls from a man claiming to be Jesus Christ, but that it dismissed the calls at the time.  The President will hold a press conference later this week to provide further details.  However, a source close to the President speaking on condition of anonymity said: \"Of course we get hundreds of bogus calls every day from people claiming to be Jesus, Elvis, Gandhi.  There\'s no way we could have known that this one wasn\'t just another wacko.\"  Dr. Gordon Phillip, professor of religious studies at Georgetown University and a leading expert on Christian prophecy, says that he is disappointed that mankind missed a \"once in millenia\" opportunity to receive the lessons of Jesus Christ.  \"Without more information, it\'s hard to speculate on why Jesus came back at this moment, and even more difficult to understand why he didn\'t stick around longer.  However, I wouldn\'t be surprised if he just took a good look around and decided to wash his hands of the lot of us.\"   '),(21126,2965205,'17,June,2004','2004-06-17 00:00:00','       '),(21127,2965205,'06,July,2004','2004-07-06 00:00:00','       '),(21128,2965205,'21,July,2004','2004-07-21 00:00:00','       '),(21129,3745004,'05,July,2004','2004-07-05 00:00:00','    Hoshi wandered about the old mansion searching for her beloved Ukchan. Every footsteps she took picked up tons of dust swirling around her. Coughing she stumbled forward leaning on a wall till the dust cleared. After the dust was settled, Hoshi looked around and spied on a stairs she never noticed before. The girl sensed a strange delight creeping up on her as she made her way up the stars. Something was going to happen, she could tell. At the top of the stairs was a small trap door. Hoshi pushed it open and stepped into the crisp night air. A movement caught her eye as she emerged. Ukchan turned in the full moonlight to see who was there. His body had a crystalline glow around it in the moon as her turned and his short blue hair turned to diamond in Hoshi\'s eyes. As he spotted her. The deep brown eye of Ukchan drilled right into Hoshi\'s soul as he stood searching her. The females min and heart told her to run to him, but her body didn\'t listen. All was still. suddenly Hoshi came into herself and dashed forward, her legs becoming weaker and tears streaming from her eyes. She drew nearer and near when her legs gave out. Hoshi started to fall, fall right into Ukchan\'s strong arms. Her heart skipped not one, not two, but three beats as Ukchan stopped the fall. She didn\'t dare move for a moment feeling she might wake from a dream if she did. Ukchan gently wrapped his arms around her and pulled her tight to him. Hoshi help her breath as she buried her face in her soft shirt and linked arms around him. She relaxed when Ukchan;s torso softened and leaned on his chest. Hoshi was in a dream like daze as her worries fell away. Squirming a little in the arms that help her firm yet gentle she looked up and caught Ukchan looking down at her with a small curved smile. She lovingly smiled back at him with teary eyes. Taking then a chance Ukchan leaned down, closing his eyes and met lips with Hoshi in a kiss so heavenly and pure that the world instantly feel away from them. Their bodies grew hot as they shared their warmth and seemed to melt into one another, Hoshi closed her eyes and started to kiss Ukchan back putting slender arms around his neck. Both felt as if they could stay like that forever, but Ukchan pulled back after some seconds. Hoshi opened her eyes and looked at Ukchan with the slighest blush streaking her face. Nothing in the world could have beaten that first kiss. '),(21130,3745004,'05,July,2004','2004-07-05 00:00:00','   ((I wonder how far this will go... and uhh... sorry about the spelling.. they took Spell cheak off >  You start to fly hard, fly faster. Every fiber of your being is set on getting more of that precious life-giving substance. You could still taste the salty sweatness in your mouth and you savor it for as long as possible. The lust consums you in it\'s dark abyss. Soaring over the fiery forest of death you smell the blood of an innocent. The blood of a child mingled with fear, yet some bravery. The young one was drawing closer to you and you lose all control you may have had over your body. You swifty and silently float down to the forest floor and hid in a rose bush as red as the blood you lust for so infinitely. The child makes it\'s way into a clearing making it\'s self vulnerable to even the slightest of dangers. You studdy the child with your darking eyes and find little threat other then it carries a small dagger by it\'s side. The child stumbles a little to close to your hidding place and you get a good smell of it\'s blood and great fear. The lust takes a new killing hold on you and you leap out at it in a rush of rose pettles floating harmlessly around you. For a splite second the world is frozen in it\'s own evil as the darkness consums the child. It screams so earsplitting the next moment before you slash at it\'s windpipe killing it instantly. Blood poors out of it staining the ground red as you start to bite deeply into the wound, lapping up all the newly killed life withen. With bloodlust becomeing even more over powering you turn animal like in your feeding and gorge yourself so deeply that no other feelings can get though. '),(21131,3745004,'05,July,2004','2004-07-05 00:00:00','   ((I hope you don\'t mind, I continued the story.. add on as well! We could collaborate on it, that would be cool ^^)) Currents of warm air swell and gather underneath your ebony-furred wings, pushing you up, up, higher and higher in a spiraling motion until you think you shall faint from the height. Meaningless patterns of light dance on the inside of your eyes, a fireworks show only for you. You cough. Again and again you try to purge your throat of the sticky-sweet smoke, but it clings to your windpipe like molasses. Your mind starts to becoming weak, driven by the need for clean air it throws its last reserve of energy to your new wings, flapping madly to send you shooting out of the foul, odorous cloud. You break through. There it is- clean air! It seems like a miracle to you as you gleefully inhale the crisp, cold evening atmosphere. But something\'s not right.. your tongue runs along the top of your mouth before passing along your two canine teeth.. a salty substance trickles from the cut caused by the needle-like point of your tooth. And as the blood flows over your tongue your mind is sent into a frenzy. That was it.. you needed it! Blood! You craved it, you lusted for it, HAD to have it. It was your source of energy.. this newfound realization terrified the corner of your brain still conscious. But soon your mind is captivated by the lusting presence inside of you. Forgetting the forest, forgetting the flames, the corpses, the curse, you beat the air. Wind streams past your face as you cut through the air toward where your instincts told you blood was. Where humans were.. where there would be people to provide you with your means to live.. that is what you longed for. For now, you were not human but a creature of shadow, a vampire.. '),(21132,3745004,'05,July,2004','2004-07-05 00:00:00','   Far in the dark night. When all hope is lost. The screech of an owl, the howl of a wolf. You shake as you steep deeper into the forest. The air starts to get colder and yet somehow, more evilly delicious. You come to an ancient clearing showered with sliver from the full moon staring down at it, at you. Your skin prickles and you turn around. Bloody rotting corpses are hanging from each tree. Each body hung from organs strangled around their neck. They seem to be staring right at you. You run! Fast. But nothing could take the curse on you. Nothing will change you back. You start to go though a transformation. Your ears become long and pointed. Your K9 teeth become sharper and longer. As your nails grow great black wings burst the skin on your back.  Screaming in pain you keep running, right into a corpses hanging from a tree. It seems to put it\'s dead cold arms around you and not let go. Nothing could help you, nothing would save you. You shut your eyes trying to believe nothing is happening, the world is fine, but as soon as your eyes even slightly open you are still being held onto by the corpses. Turning back to the trees trying to break the grip your eyes grow even wider with fear. The whole forest is burning. The sky is alight with dancing flame. The corpses disappears as you try to run the other way only to find that you are blocked in. Trees fall ablaze to the ground starting the grass on fire. You can\'t get out. You will die and forever this evilly holy land will be your ashes grave. In your vain attempt to escape you leap into the heating air and your wings take flight. Higher and higher into the smoky air that\'s clogging your lungs. '),(21133,3745004,'18,July,2004','2004-07-18 00:00:00','   While I think of something to say, I shall mention that I love the name Mat. Thank you.  '),(21134,3745004,'05,July,2004','2004-07-05 00:00:00','   Ok so this is the dream I had last night. It is really mushy and stuff and the best non-vampire dream I\'ve ever had! ^^ Beware! Fluffy things included!  Oky so... It\'s like this: It started In school and stuff, right? Yea there was this ride thing in this building. Some of my friends and I went in it and like it was a conveyer belt with trash on it splitting into to was as it went. The trash goes one way and people go other way. But like we were on it and this really big bog of beer type thing went on it and got stuck with broke the ride so my friends and I just walked and sat watching the movie they were playing in there on this great pull down T.V thing. *Movie unknown*. Yea. Then we left and separated. *later* I went back in and sat down bored watching movie. *Beware! This is where fluffy comes in!* This insanely hot boy comes in and sits right by me! I mean WOW! Short black hair kinda anime style and stuff and eyes and tall and OMG! I can\'t describe his hotness (and in the dream he is also the person that plays Ukchan). I look at him knowing him and like, I have a crush on him. He looks at me and I look at him and stuff. He says to me in his warm voice type thing, \"I\'ve always liked you\" And I\'m all OMG! I love him! And stuff. And like yea. He takes my hand and like yea. He hold hands there and stuff. I lean on his shoulder. And then like yea soon we get up and like walk out holding hands and stuff and I\'m all ^^. I start walking in front of him holding both his hands with both mine leaning back on his chest and stuff. And we are just like walking that ways across the quad. There we meet his friends and he talks to them and stuff and I break away from him to go tell my other friends. And like I find friends and things (Bekka and Annie fer some reason) and I tell them and there just like \"No fair!-_-; My Ukchan\" and I\'m all (^-^)v yea. I go back  and another friend runs up to me and are all like \"How did you get Mat? (name of boy)\" and I\'m all like \"I was myself\" type thing. And then I go back to him and I spend the rest of the dream in his arms. The End!  '),(21135,3745004,'06,August,2004','2004-08-06 00:00:00','   Okay, so I finally had a dream!  It started out.. I\'m not quite sure. But it ended up in this farmer\'s market type of thing, where lots of inventors are selling a whole bunch of junk. So, I\'m there for a bit, then I go back to this.. camping.. place. Sort of. There\'s a barn-ish-shed-thing-of-doom where cars are parked and stuffs like that. And so we\'re all gathered by this freakin\' AWESOME anime guy, looks like Crim from .hack//SIGN only wearing leather.. black leather.. mm.. :9 Anyway. And he\'s like, in front of this hill, and saying that there\'s a fire coming our way. A few seconds later, fire comes ROARING over the hill. People are all screaming and running for their cars and all, so we (we being my mother and father and I) hop into an old-style white beetle and get to the road. Then I look over and I\'m all \"O_O OMIGOSH WHERE\'S BOY?\" And my mom just turns to me all calm-like and says, \"Oh, he must have died.\"  So I start freaking out, and my parents don\'t even seem to care. I hop outta the car and race back toward the barn-ish-shed-thing-of-doom, which is on fire. And people are just.. standing in it not caring. -_- So I start shouting, \"BOY! BOY! PETER! WHERE ARE YOU?\" He just walks out (Point of interest: he\'s got glasses on in my dream, but normally he\'s got contacts) calmly and gives me \'that look\' that shuts me up pretty fast. So we all get back in the white beetle and are driving down the highway when suddenly the \'rents turn off of the road onto a.. golf course.  We\'re just drivin\' down the golf course when suddenly it becomes a cliff. And at the bottom of the cliff we see.. more golf course. Well, we can\'t stop ourselves now, we go flying off of the edge, then land. Boy and I crash through the windshield and start.. uh..  flying.  Like we were shot out of a cannon, you know? And the land below us becomes so teeeeeny tiny, and we hit the ground and it doesn\'t hurt at all. Then we get up to find mum\'n dad.. then I woke up.  ...I have the oddest dreams.. >_>  '),(21136,3745004,'04,August,2004','2004-08-04 00:00:00','   Ok this odd dream has to do with wolvies!!! YAY! Many odd things in here! Funfun!!  So it starts off with me and my brother in my room just sitting when my mother and some random lady person comes in. They sit at a table that is at the foot of my bed in my dream and look at us. We look at them. Mother dear tells us that we re wolves!! And we are all like YAY! ^^ *bouncy bouncy bouncy *! And the lady person is all nice and gives us papers to fill out. Something about mating season. Well like. We don\'t have time to fill it out and we go to this place. We walk right next door to the old school, but as we are walking it changes to a forest. There we find wolves everywhere! hundreds of wolves! And like, most of them are really thin with their face and body so I change into a beautiful pure black wolf (most of the other wolves were brown) and like I walk down there. Here I look around and see the lovely brown wolf. Not thin like the others and not fat. Just, perfect. I run up to him \"Hello? Hi? Heeeelllllooooo?\" But before he can answer me some random werewolf in his human form blows a whistle and we all have to line up just like in the old school! Then like, I look for that wolf I was trying to talk to, but I can\'t find him. After that we went home and got the papers to fill out again. I started to fill it out like my brother, but he finished first (because most of the time he was trying to sit on my paper). When the lady comes back from where ever she was I showed her my paper and she gives me another one after telling my brother to not do that. I try to fill it out again, but for some reason I was not sitting at the table. I soon lay my paper down again just to take a break when my brother comes over and sits on it and tears it. I\'m all \"grrr\" The lady comes in and I show her my paper and shes all \"*sigh* I didn\'t think it would take you THIS long to fill it out.\" I get the paper and sit at the table and do the second question then look at the first. It says to write your parents and grand parents and great grand parents madden names and I\'m all \"What the heck? Do we have to do the first one?\" and the lady is all \"No, you can skip that one.\" So I do most of the paper before I wake up. OHH! Wasn\'t that fun? Being a wolf was the best thing I have ever felt. OH! I want a continuation of that dream!!!  '),(21137,3745004,'04,August,2004','2004-08-04 00:00:00','   Ok, so last night I had many odd dreams and I will tell you two here. So like yea. This one has pirates and kids in it! On with the dream!!!  It started I was on a huge boat in the middle of an ocean with my mom and like 11 daycare kids. All the daycare kids were like doing this course where they climbed things, crawled on things, and slide down things to nursery music. Then like my mom was wearing a scary red dress. Then after all the daycare kids were done we went down below the deck to uhh.... change into different outfits. Well my mom brought out this patriotic dress and then a really REALLY big american flag. Well I thought it was for twirling like in color guard, but then she took it off the poll and rapped it around my head like a bandana. So I\'m like ooook... Then like I went up with five kids and like these pirates started to board our ship. They grabbed the five kids for slaves then wanted more. I tried to fight them but they soon got five more. With these they took out these needles (sowing) and pocked each kid and each one died. One little boy was left. I took him to the front of the boat and told him to stay, but he followed me anyway. I got mad at him but then the pirates took out other needles (doctor) and stabbed him. Then I woke up. Isn\'t that nice? Well, yeah sure, I mean. The kids are gone now right???  '),(21138,3745004,'28,June,2004','2004-06-28 00:00:00','    Gomenasai!!  T-T  Chibichan returneth!!  ..I need to know if you remember Kenny \'Dahling\' the Hacker.. ohmigosh.. I really do need to know..    ~Chibichan Ryuchan Bekkachan   '),(21139,3745004,'26,June,2004','2004-06-26 00:00:00','    *Insert a long sigh here* Well, seeing as how our ikkle Chibichan Ryu has gone camping.. or so she told me for the weekend, it looks like I\'ll have to be keeping this place up to date. Humph.. I see how it is! >:P Just kidding of course. Anyway, I\'ll be here, posting, until she gets back, then she\'ll have to make some major postage! I\'ll keep my eye on her, make sure she\'s in line! >:3 Hmm.. moving right along, anyway.. I have nothing to say, currently, I really only posted so that people would know it\'s just me. Poor ikkle SDchan Hoshi all by herself. *sniffle* Oh well.. s\'all good. See y\'all around!   Love and Peace Out!  ~Hoshi-chan.. neh?   '),(21140,3745004,'25,June,2004','2004-06-25 00:00:00','    Well well well... here we are.. the Splat Duo has arrived! Everyone.. be afraid.. be VERY afraid.. When I tried making it, the thing didn\'t work.. so Chibichan Ryu did it and here we are! Heh heh heh.. I think it will be quite easy to throw the whole world into chaos and confusion.. eh? =^.^=  Love and Peace Out  ~ SDchan Hoshi.. neh?   '),(21141,3745004,'25,June,2004','2004-06-25 00:00:00','    *thunk* z_z   '),(21142,3745004,'27,July,2004','2004-07-27 00:00:00','    Ryu-chan has stayed up too late again..  It\'s 3:40, I usually get to sleep by 3... I got caught up in my music again -_-  Hoshiiiii? Tomorrow- er, later today, would\'ya purleeeeeaaase beat some sense into me? Tankoo. This goes for anyone else I know, please BEAT SOME SENSE INTO MY THICK SKULL.  *must go get SLEEP.. gotta wake up by 12\'n all -_-;*  \'night... I should get more sleep. -0- *thunk*   '),(21143,3745004,'24,July,2004','2004-07-24 00:00:00','    Picture yourself in a boat on a river, With tangerine trees and marmalade skies Somebody calls you, you answer quite slowly, A girl with kaleidoscope eyes. Cellophane flowers of yellow and green, Towering over your head. Look for the girl with the sun in her eyes, And sheâ€™s gone. Lucy in the sky with diamonds. Follow her down to a bridge by a fountain Where rocking horse people eat marshmellow pies, Everyone smiles as you drift past the flowers, That grow so incredibly high. Newspaper taxis appear on the shore, Waiting to take you away. Climb in the back with your head in the clouds, And youâ€™re gone. Lucy in the sky with diamonds, Picture yourself on a train in a station, With plasticine porters with looking glass ties, Suddenly someone is there at the turnstyle, The girl with the kaleidoscope eyes.   '),(21144,3745004,'18,July,2004','2004-07-18 00:00:00','    Well, they had a band and an orchestra. But everyone knows that band camp is so much cooler to say than band and orchestra camp.   '),(21145,3745004,'08,July,2004','2004-07-08 00:00:00','    Hahaha, mm watermelon... ^.^ Anyway, Kenny \'Dahling\'.. yesh, ah remember him! I don\'t think I\'ve ever spoken with him, but then again... I heard him mentioned on several occasions though. And like.. it was all.. amusing. And stuff. Yea, and stuff. Right. Anyway... Okay, so you\'re off to Band Camp. That means.. Hoshi-chan will be taking over again! Mwahahaha! Uhh.. err... uhh.. yea. x_x I wish I could go to band camp.. but Hoshi-chan plays the violin. Heehee!   Hoshi-chan... neh?   '),(21146,3745004,'06,July,2004','2004-07-06 00:00:00','    10 days Ryu-chan will be gone.  *thunk, disappears*   '),(21147,3745004,'01,July,2004','2004-07-01 00:00:00','    Ryu-chan has watermelon.  ^_^  *thunk*  ~Chibichan RYU ^0^   '),(21148,3745004,'07,August,2004','2004-08-07 00:00:00','    Well... if my memory serves me correctly... Chibichan Ryu will be gone for a week, so that leaves me to take over T3h Splat Duo Headquarters yet again. Lucky me. ^_^ I\'ll be here to post random nonsense that\'ll bore you all to death so badly, that you\'ll wish Chibichan Ryu was back! Heh... heh heh heh... >:3 Love and Peace Out! For now... >:F  ~SDchan Hoshi... neh?   '),(21149,3745004,'06,August,2004','2004-08-06 00:00:00','    Ashaman JoeÂ to meÂ Jul 19 http://24.15.220.212/sniper/irc-stereotypes.html  ..scizzors..Â to Ashaman Jul 19 I am SUCH a Cute Girl Stereotype ^0^  And so is Annie, and Jodi- sometimes -Sheep sometimes tho\' not as much, ummm... pretty much all of my friends except DF, and I have no clue what the fuzznuck HE is  Ashaman Joe to me Jul 20 He\'s the Erotic Furry. =P  -------  I don\'t even have a reply to the above message. All I can say is.. \"XD\"  ((Boy rocks, but when he has friends over it\'s a pain in the arse. *glares at loud people currently in room*))   '),(21150,3745004,'05,August,2004','2004-08-05 00:00:00','    Eh... heh heh heh... music is good. Yush, music is good.... >:3 Another awesome song. Go find it. ^_^ \"Cowboy Hat,\" by Nickelback.  Yes he wore a cowboy hat// Each bone broken in his back// All the money and all the gold// Couldn\'t buy my love at 8 years old// Now I\'m in love with a girl thats wrong for me// Ask her, she will disagree// Her father was a small time man// With small time dreams and small time hands// But// But I was there when there was nobody home// Please believe me when i say i dunno// Gone for good in fact but what does it mean// Within a year they always sold and bought//  Easy enough to just look at her// Whole world on a silver platter// Was that that bad for me// If I was there the scotch is free// Won round one but its not over// He don\'t care cause he don\'t know her// True love is not the same// If I could lend a hand in shame//  But I was there when there was nobody home// Please believe me when i say i dunno// Gone for good infact but what does it mean// Within a year they always sold and bought//  If I was big// So much easier to see// Don\'t hide from me// That which I\'ve already seen// You lie to me// More than occasionally// Don\'t lie to me// You lied to me//  I\'m as high as I\'ll ever be// I don\'t mind, Stay and see// They\'ve all gone, Why don\'t we// Holding hands, you and me// But I\'m as high as i\'ll ever be// But later on, lies to see// My past is gone, only from me// I\'m alive//  But I was there when there was nobody home// Please believe me when I say I dunno// Gone for good infact but what does it mean// Within a year they always sold and bought//  Remember when it was me// Driving come down just for me// And then we\'ll try as we\'ve been//  Heh heh heh... we lurv that song! It\'s so awesome! But then again... we have our other reasons for liking it, eh Chibichan Ryu? ^_^ Love and Peace Out!  ~SDchan Hoshi... neh?   '),(21151,3745004,'05,August,2004','2004-08-05 00:00:00','    Ah sowwee Chibichan Ryu! T-T!!!! Ah didn\'t mean to insult j00... *sniffle* SDchan Hoshi ish sowwee... mesa thought that psychos were koolness... meant it as a compliment. *sniff*  Oh wellses, if I offended j00 pwease tell da Hoshi-chan and j00 can beat t3h sense into her as much as you like! ^_^ Bweeheehee! Randomness astounds me. w00t! Ehh.. nothing else to say, really. *splat* Love and Peace Out!  ~SDchan Hoshi... neh?   '),(21152,3745004,'04,August,2004','2004-08-04 00:00:00','    A few words of wisdom to all of you out there who think we\'re moronic idiots with no life. \"Never pick a fight with an idiot. They\'ll drag you down to their level and beat you with experience.\" Trsut me. It\'s true. Heh. Plus... we\'ve got the evil genius who knows martial arts... and we\'ve got the insane psycho that\'s just... well... watch out. ^_^ Heh heh... we are T3H EFFIL! I lurv being evil. It rocks SO much. Bwee. Fankoo, Chibichan Ryu! *has sense beaten into her* Buwahahahahaha! Back to plotting to take over the world! >:3 Love and Peace Out!  ~SDchan Hoshi... neh?   '),(21153,3745004,'03,August,2004','2004-08-03 00:00:00','    Nekomori San (8:34:06 PM): Wheeeeeeeeeeeeeeeeeeeee! Whatcha up ta? Kyoki  Ryu Chan (8:34:36 PM): tetris Kyoki  Ryu Chan (8:34:41 PM): plotting to take over the world Kyoki  Ryu Chan (8:34:44 PM): y\'know, normal stuff  ---------  So anyway.. *beats sense into Hoshi-chan* ^_^   '),(21154,3745004,'02,August,2004','2004-08-02 00:00:00','    Yup. You guessed it. It\'s time for me to update all of my Blogs, and I.. DON\'T HAVE ANYTHING TO SAY. This is just a waste of your time. Don\'t bother reading it. It\'s just a little something I\'m putting here to make me feel better. It makes me feel like I actually updated all of the Blogs that I have. So... you can ignore this. Yup. If something interesting pops up, then I\'ll post it. Yep. So.. if you\'re still reading this, then you\'re stupid. Well.. not really stupid. You just have a lot of time on your hands. Am I right? Tell me I\'m right. Ah, jeez, I have WAY too much time on MY hands. I\'m such a loser. Chibichan Ryu... in your next post, beat some sense into me. Please? x_x;; Love and Peace Out, for now! ^_^;;  ~SDchan Hoshi... neh?   '),(21155,3745004,'02,August,2004','2004-08-02 00:00:00','    Okays, I\'ve been getting into the hang of updating all of my zoomyful Blogs at one time, almost every day, so now I\'ll be posting on here WAY more often. ^_^ Buwahaha.. okay, anyway. Here\'s a song that we like by... you guessed it! Sarah McLachlan! We loff/lurv her, yush we do. ^_^ She\'s really good at capturing the mood in her songs. Here you go! \"Full of Grace,\" go download it, OR ELSE!  The winter here\'s cold, and bitter// it\'s chilled us to the bone// we haven\'t seen the sun for weeks// too long too far from home// I feel just like I\'m sinking// and I claw for solid ground// I\'m pulled down by the undertow// I never thought I could feel so low// oh darkness I feel like letting go//  If all of the strength and all of the courage// come and lift me from this place// I know I could love you much better than this// full of grace// full of grace// my love//  So it\'s better this way, I said// having seen this place before // where everything we say and do// hurts us all the more// its just that we stayed, too long in the same old sickly skin// I\'m pulled down by the undertow// I never thought I could feel so low// oh darkness I feel like letting go//  If all of the strength and all of the courage// come and lift me from this place// I know I could love you much better than this// full of grace// full of grace// my love//  Well... there it is! Read the lyrics, download and listen to the song.. then reflect on this. It\'s good for your creative thinking. Maybe after deeply thinking and reflecting, you\'ll even get it. Nah. It\'s not that easy to understand. Your loss. Tata! ^_^  Love and Peace Out! ~SDchan Hoshi... neh?   '),(21156,3745004,'24,July,2004','2004-07-24 00:00:00','   Well, I got some art. Woo woo, yay!   urlLink Nekomori  Part of a trade with  urlLink Sekora , it\'s her character Nekomori. Just sitting in a tree, chillin\'. Haven\'t colored it yet because I don\'t know what colors she is (hint hint Sekora..) ^^ The only thing that makes me angry about it is the fact that when I inked it with my crappy cheap pens, the staff got messed up.. but I can fix that when I color it using Photoshop. So, overall.. not horrible.   urlLink Dalisedeni  My fan-made character from the musical CATS, looking rather cute and HAPPY. Deni\'s a rather HAPPY type of person... cat... Jellicle.. uh, whatever. The main point is that she\'s HAPPY. Only thing I don\'t like about this picture is the hand she has balled into a fist, it\'s just.. uh.. fingers = way screwed up. An\' I was trying to make it look like she was holding her arm out straight toward the hypothetical camera. But, uh.. doesn\'t look like that worked. Either way, though, Deni is HAPPY!!!  ^_____________________^ '),(21157,3745004,'21,July,2004','2004-07-21 00:00:00','   ((Just a free-verse poem I wrote a few days ago.. it doesn\'t rhyme or have any apparent rythm..))   Forever  You are like a beacon to me Your smile, your face, your eyes They call to me They tell of safety, they tell of love Tell me that all I need to do is to get to you To feel whole If I could I would run to you Never leave your side Tell me how to get to you, and I\'ll come Just tell me when...  Give me half a chance I\'ll fill you with warmth Just give me the courage I lack I\'ll stay by you Lead me through the hard times I\'ll give you all the good All I would ever want in the world Is to be here in your arms- Forever  I see you when I close my eyes It gets me through each day That taunting vision of someone so dear Could it be a sign? I need but to think of you To feel that things will be all right But that emotion isn\'t good enough I need to hear your voice telling me so before I know it\'s true I know if I was near you I\'d never have to fear Won\'t you please show me how?  All I need is a word I\'d give you all I am Just safeguard my soul for me It will give you the happiness you deserve Tell me it\'s all right, show me the way I\'ll make you feel whole day after day You\'re my breath, my sight, my hope And all I need is you- Forever '),(21158,3745004,'21,July,2004','2004-07-21 00:00:00','   Dark... it was... dark.  But the mottled light that cascaded down from the fissures in the wall, the flecks of light was what made it all okay..  It made it okay to the dragonlet that dozed, hidden from the casual eye by the overhanging rock ledge. She wasn\'t old per se.. well, not for a dragon at least. Somewhere between two centuries and three, if not three and four was she. But, compared to the rest of the dragon race, she was young. Considered thirteen, at best.. and only then if one was feeling generous.  But the little azure reptilian creature, she was not the proper size...  She was small.  Miniscule.   A runt, which is the name that many of the older dragons gave her. She had been undersized ever since birth. Due to one faulty gene, one gene in her entire body out of the millions housed there, the one faulty gene that had the ability to change her appearance completely. When one her age would be by now roughly the size of a small school bus, she was instead about the size of a house cat.  Size aside, though, she had the looks of a dragon.. blue scales made up armored skin, lethal spikes ripped from the very tip of her nose all the way down the base of her spine to her foot-long, whiplike tail. The two scaly bat-wings that shot out from her shoulder-blades, the span longer than her body itself but right now, folded tight against her back like a tortoise shell. The deft, four fingered hand-like paws, each nimble digit accented by sharpened keratin stakes of claws. The strong back haunches of the creature like loaded springs, feline.. almost. Her mouth, the elongated front canines that protruded from her mouth to give her the look of danger even when asleep, the row of needle-sharp chompers housed within.. Even the lizard eye, golden, ever-searching, now covered by a filmy eyelid as well as the tough outer lid, out of the same material that her skin was.  Yes, she was without a doubt a dragon. It would be nigh impossible to mistake her for anything but, if it were not for her size. One would immediately assume she was instead a fire lizard instead of dragon.. if they were not familiar with the species. One with just a touch of knowledge on the subject would know by the ridge of spikes, the shape of her head that she was indeed dragon.  A Blue Dragon.. the race of dragon most commonly dealing with liquids, with the life-giving substance on this planet.. Water..  ((TBC- to be continued..)) '),(21159,3745004,'21,July,2004','2004-07-21 00:00:00','   Ryu-chan\'s creative works needed a home for their own..  I\'m just picky that way.  ^-^  Whether or not I update this frequently, it will be a place to house my poems, stories, art... etc etc. '),(21160,3745004,'14,August,2004','2004-08-14 00:00:00','   jinopets.com has some cool looking pets, but not many! It\'s maps seem a little ill drawn and the games are few, but it also has good chat! Go see!  veropets.com is really odd. It has somewhat of a choice fer pets and you can have 3. The pets seem... odd. They seem mostly to be made from parts of other animals in life. Then the map. Most of the places dont work, but hey, it\'s a growing site!  furry-paws.com is a very odd sitee. It has nothiing but dogs, lots of dogs. Over 100 to pick from. These are realy life dog breeds. Then, after you find where to get them (It\'s kind hard to) you can change it\'s pic. The place has jobs you can chose from and games and stuff. Overall it\'s on odd oky-ish site.  muchopets.com is really odd. The pets are 3D-ish and moving! That\'s a new one. There are NO games and the site is confusing. It\'s a slightly OK site, but I wouldnt tell anyone to join it.  '),(21161,3745004,'14,August,2004','2004-08-14 00:00:00','   Usually at least one of us has a log-in name at these sites.. so, if you just wanna talk to us there, here\'s our usernames.   AfterMathZone  soramimi (Ryume)  Argopets  ryu-chan (Ryume)  Chibi Friends  soramimi (Ryume) Nekomori (Nekomori)  Chizo Pets  soramimi (Ryume)  Cool Camels  camelon (Ryume) Chichinegima (Nekomori)  Creature World  ryulon  eCritters  soramimi (Ryume)  Jinopets  ryu-chan (Ryume)  KiroPets  ryulon (Ryume)  Neopets  lupe933 (Ryume) dancingwilight (Nekomori)  Nutrinopets  nekomori (Nekomori)  Petnebula  ryuchan (Ryume)  Powerpets  nekomori (Nekomori)  Sensi pets  ryuchan (Ryume)  Setropets  ryu-chan (Ryume)  Subeta  nekomori (Nekomori) ryu-chan (Ryume)  Venetopia  ryulon (Ryume)  Veropets  ryu-chan (Ryume) nekomori (Nekomori)  XioPETS  ryu-chan (Ryume)  Zetapets  ryuchan (Ryume) dancingtwilight (Nekomori)  '),(21162,3745004,'14,August,2004','2004-08-14 00:00:00','   This blog, I belive, is for pet sites. Post all the pet site you know of on here.. or something ^^v HIA! I belive I am a staff person, thing here then! I\'m Neko, Nekomori. I am not as professional as the other in here, but I try > '),(21163,3745004,'14,August,2004','2004-08-14 00:00:00','   This will be edited as nessecary... Alphabetical, with short descriptions!   - A -   urlLink AfterMathZone  - Great site with very interesting pets, offers more than most virtual pet sites.  urlLink Argopets  - A nice little site.. cute pets.. but it\'s small, as of right now, and slightly buggy.  - B -   - C -   urlLink Chibi Friends  - A cutesy and very girly pet site.  urlLink Chizo Pets  - Nice site with nice pets.  urlLink Cool Camels  - Raise a pet camel!  urlLink Creature World  - Adopt a real-life species of pet and earn money..  - D -   - E -   urlLink eCritters  - Definitely more mythical-based pets, some interesting ones.  - F -   - G -   - H -   - I -   - J -   urlLink Jinopets  - Interesting pets, nice amount of stuff to do.  - K -   urlLink KiroPets  - Nice layout, nice pets, normal type of site.  - L -   - M -   urlLink Monopets  - Not the best site out there..  - N -   urlLink Neopets  - Pretty much the biggest and well known virtual pet site.  urlLink Nutrinopets  - DESC COMING SOON  - O -   - P -   urlLink Petnebula  - A bit lacking in the pet department, otherwise fine.  urlLink Powerpets  - Not liked by the VPO staff. Basic pet site.  - Q -   - R -   - S -   urlLink Sensi pets  - A nice site with interesting characters and a clean layout.  urlLink Setropets  - The art could use some work, but it\'s got its high points..  urlLink Subeta  - Nice looking pets, nice layout. However, it lacks games.. if games are your thing, this is not the place for you. Overall, very good.  - T -   - U -   - V -   urlLink Venetopia  - A pets site with a plot, for those who enjoy that.  urlLink Veropets  - Well drawn, well organized site in progress.  - W -   - X -   urlLink XioPETS  - Small but growing site.  - Y -   - Z -   urlLink Zetapets  - A nice, well-rounded pet site.  '),(21164,3745004,'14,August,2004','2004-08-14 00:00:00','   Simply, what this is is a listing, reviewing, and otherwise of virtual pet sites. Why..?  Because we can.  Feel free to submit any ones you might have!   ~Ryume  '),(21165,3745004,'24,June,2004','2004-06-24 00:00:00','   THIS IS MY MINDSPACE. I DON\'T NEED ANYONE TO FLAME ME.  Now that that\'s out of the way..  Yeah, this is just where I gab. I might post how my day went, how I\'m feeling, or any other random tidbit about my life.. if you don\'t like what I write, just close the damn browser window. I don\'t force you to stay here if anything I say offends you. It\'s my blog, and while I\'m honored that you\'re reading it I don\'t need to hear any complaints.  Thank you.  '),(21166,3745004,'31,July,2004','2004-07-31 00:00:00','   I went and saw The Bourne Supremecy today with Diane and Amy B. after the three of us stayed at Amy\'s house for the night.  Pillow fights, projector screens, loud movies, a raid-able pantry, an ELEVATOR(gonna live in it when I\'m living on my own), what more could you want?  Oh. The Bourne Supremecy?  Highly  recommended. I\'m still unsure if I like it better or worse than the first movie.. It\'s definitely got more fighting/car chase action scenes. Less of a \'Who Am I\' and more of a \'What Am I.\' It was fun.. but, uh, good times, good times!  '),(21167,3745004,'29,July,2004','2004-07-29 00:00:00','   I think this post pretty much sums up the  very  intriguing RP I had today...  Sasuke~ continued staring off into space until he blinked once, again. Someone was intruding upon his white haze..  his  reality. It blurred, before finally snapping into focus. He was back in the tavern.. he was back in the tavern, and he could hear Kaiuri\'s words. In disbelief he centered his gaze on the girl who had stuck her face in front of his. He just gaped at her, like he could not comprehend what she was saying. Uselessly his jaw worked up and down, no sound coming forth from his throat. Then he frowned, an angry look coming into his eyes. She was trying to take away  all  of his past... she was trying to wrench his  name  from him.  He would not let her!  \"I am Sasuke..\" he began, stopped dead by her words. \' I killed you. I ran you through with one of your own guard\'s weapons! \' \"No, I can\'t have, you don\'t know..\" \' You died that day. I watched your blood spill out like a waterfall cascading over the edge of a cliff. \' \"That\'s not, there\'s no way..\" But hadn\'t he awoken in a pool of blood, his own? No, it couldn\'t.. \' I watched you die. \' \"You didn\'t..\" \' I watched you take your last, shuddering breath of life. \' It can\'t have been.. \' I was there, you cannot trick me. I know you\'re not him. You\'re not Sasuke.. \' These words, and the rest, echoed in his skull haphazardly, each as if a separate voice shouting at him. \"No, no, no...!\" His forehead was touched by her hand. \' You are dead and I am alive! \' All her words between when she touched the spot where her kiss had landed years ago and when she declared that he was dead, had disappeared like the middle of his life. Just clouded by that white haze. Even before Kaiuri had finished condemning him to hell Sasuke was on his feet, the stool he had been sitting on knocked over in his anger. \"Shut up!\" he burst out almost childishly. Normally his shoulders were hunched, as if Sasuke just wanted to go unnoticed. But now, in all his fury, he was at his full six feet and eight inches, towering over Kai. \"Just shut up! You don\'t know, you can\'t, I\'m not dead! Can\'t you see? I\'m not dead! You never killed me, I don\'t even  know  you! I don\'t know anyone, I don\'t..\" his voice was dissolving into helplessness. \"I\'m not dead, I can\'t be, I can\'t.. I can\'t!\"  But hadn\'t he awoken in puddle of blood..? Didn\'t this girl\'s presence cause him to fade into the white haze again? But he wasn\'t dead... was he?  '),(21168,3745004,'27,July,2004','2004-07-27 00:00:00','   After Jodi\'s posts regarding this, I feel obliged to add my part in the Kenny escapade. Where Sheep\'s account got hacked, and everyone\'s mission was to mess with Kenny the Hacker\'s mind as much as possible. XD We did well, I would say.  7:33PM loves scizzors: ello Katie Marino: ello loves scizzors: ello 7:35PM Katie Marino: ello loves scizzors: ello loves scizzors: ello hacker! loves scizzors: ^_^ 7:40PM loves scizzors: so hacker-Ken, whats up with you hacking this account? Katie Marino: hey! loves scizzors: *snerk* Katie Marino: ? loves scizzors: hacker-Ken Katie Marino: damn, no boyfriend ish people are on loves scizzors: I daresay you might be wrong about that Katie Marino: oh really... loves scizzors: yes really loves scizzors: I COULD tell you who her boyfriend is loves scizzors: ...if I wanted to Katie Marino: please do loves scizzors: gimme one good reason and I will! ^-^ Katie Marino: becuz I won\'t bug u anymore 7:45PM loves scizzors: hmmm.... mayyybe.... lemme think about it Katie Marino: please? loves scizzors: ok loves scizzors: I will Katie Marino: really? loves scizzors: you know how you were wondering if I told the truth about me being a girl? hahaha I wasnt loves scizzors: Im her bf loves scizzors: hahahaha Katie Marino: really? loves scizzors: really Katie Marino: how dare u u idiot! loves scizzors: HAHHAHAH loves scizzors: that was so funny Katie Marino: How degrading to say u were a girl! loves scizzors: hackers are so stupid Katie Marino: U idiot! Katie Marino: its idiots like u who make girls think they can do everything we can loves scizzors: thats not nice to say to someone who just answered your question now, is it Katie Marino: so.. loves scizzors: well, you know, they CAN Katie Marino: ur point is? loves scizzors: my point is I might still be lying and youll never know! loves scizzors: I might actually be a girl! loves scizzors: but I might not Katie Marino: well, u know what fag? U fucking don\'t deserve credit if u didn\'t tell me this before loves scizzors: think about that, bozo loves scizzors: dude, you have a serious attitude problem loves scizzors: getting mad at people who didnt tell a hacker the truth? tsk tsk Katie Marino: o really Katie Marino: liers! 7:50PM loves scizzors: dude, honestly do you think that I even care what you, a hacker, thinks? loves scizzors: pfft. your so dumb. loves scizzors: ohhh, and cant even think of a childish retort. my my we ARE a bit below the learning curve today, are we? Katie Marino: nope loves scizzors: oh really Katie Marino: do you think i\'m gonna waste my knowledge on an asshole who doesn\'t desearve to live? loves scizzors: ahh, so maybe youre mediocre today. you can come up with a stupid and pointless retort, that has NO insulting power whatsoever. god youre pathetic 7:55PM Katie Marino: no insulting power only becuz u can\'t understand it loves scizzors: maybe a five year old could. too bad Im not a five year old, eh? Katie Marino: I know it is, are u still in kindergarden? loves scizzors: well gee I dont think Id be able to type better than you if I was. or maybe you\'re talking about your own level of intelligence? Katie Marino: oh, so ur typing better than me now... loves scizzors: at least I dont use the god-awful \"text type\" you do Katie Marino: hey, not my fault the fag set it as default loves scizzors: no, I meant the \"ur\" shortenings of word, dumbass 8:00PM Katie Marino: oh, and that\'s dumb Katie Marino: ur the dumbass u know Katie Marino: oh, ur not smart enoughto know that, sorry loves scizzors: this isn\'t even amusing how stupid you are. usually I can laugh my head off at idiots, but when ones so STUPID like you, it just looses all the humorous value Katie Marino: tsh Katie Marino: maybe... Oh, i don\'t know, BECAUSE I ACTUALLY AM SMARTER THAN U!!!!!!!!!! loves scizzors: back to the childish insults... this is so boring. can\'t you even say something SLIGHTLY intelligent for once? 8:05PM Katie Marino: why don\'t u give it a try? loves scizzors: because I\'ve been doing it all the time. I wont sink to your level    8:05PM Katie Marino: You are invited to: Katie Marino: fuck you all! loves scizzors: no thanks Katie Marino: fuck u dragongirlsirius: give me a break loves scizzors: I dont do that with assholes fangedwolfshadow has joined this chat. fangedwolfshadow: hey room loves scizzors: hey fangedwolfshadow: oh, he\'s here fangedwolfshadow: right loves scizzors: he is loves scizzors: Kenny the hacker Katie Marino: oh, and it matters to u dumbass Katie Marino: yo, not kenny, ken loves scizzors: heyyyy Kenny! Katie Marino: or are u not smart enough to remember? dragongirlsirius: Kenny! ooooh how cute! fangedwolfshadow: my my, someone\'s having a bad day loves scizzors: I am smart enough to realize this pisses you off and I should use it more frequently loves scizzors: ^_^ Katie Marino: u fucker! Katie Marino: fine Katie Marino: then loves scizzors: nope! havent uet loves scizzors: yet* Katie Marino: it don\'t matter Katie Marino: cus it\'s not my name Katie Marino: idiots loves scizzors: like I care what you think fangedwolfshadow: i really do say, u might be right ken loves scizzors: hahaha this is amusing again dragongirlsirius: oh poor kenny! so sad and morbid Katie Marino: ??? fangedwolfshadow: yes, we might all be idiots, but how are we to know? Katie Marino: oh screw u fangedwolfshadow: y must u hack in an account that obviously isn\'t urs? loves scizzors: back to childishness Kenny dahling? loves scizzors: oh, and if you were wondering, I lied to you again! heee heee! I really *am* a girl! dragongirlsirius: this will be interesting for school gosip loves scizzors: it will Katie Marino: fuck you fangedwolfshadow: veru fangedwolfshadow: very* dragongirlsirius: realy! what a bad timper you have Katie Marino: veru, what a dumbass loves scizzors: doesnt he? loves scizzors: veru 8:10PM loves scizzors: cant even type, eh Kenny? dragongirlsirius: cant you do anything but insalt/ loves scizzors: no Kenny dahling cant Katie Marino: shit head, i was repeating wolf fangedwolfshadow: i have yet to see it loves scizzors: ooooh, riiiight, we all believe you loves scizzors: last time I check my head was flesh, dumbo Katie Marino: fangedwolfshadow: veru dragongirlsirius: what a person you are! Katie Marino: so ha, loves scizzors: or did you skip science? Katie Marino: fuckyo loves scizzors: no thanks loves scizzors: not with assholes like you loves scizzors: ^-^ fangedwolfshadow: i don\'t do fuck fangedwolfshadow: sorry dragongirlsirius: he cant even controll his mouth! Katie Marino: screw u loves scizzors: sorry to disappoint you, Kenny dahling fangedwolfshadow: brb guys loves scizzors: cya dragongirlsirius: ooh kenny your my new friend! Katie Marino: fuck you all, i\'m leaving Katie Marino has left this chat. dragongirlsirius: hehehehehe fangedwolfshadow: well, that stinks loves scizzors: hahahhahahahahhahahahhahahhaha fangedwolfshadow: darn loves scizzors: HA HA HA HAH HAHAHAHA fangedwolfshadow: i\'ll brb guys loves scizzors: WHAT A STUPID IDIOT HE IS dragongirlsirius: oooooh that was ssooooooo much fun! loves scizzors: HA HA HA HA HA loves scizzors: I loved that! dragongirlsirius: see kt? loves scizzors: \"Kenny dahling loves scizzors: \" loves scizzors: ha HA HA HA HA dragongirlsirius: wait till school! fangedwolfshadow: back fangedwolfshadow: lol loves scizzors: omg I know loves scizzors: HEHEHEHEH fangedwolfshadow: that was fun fangedwolfshadow: lol dragongirlsirius: i hope he comes back fangedwolfshadow: ur a good insulter rebecca loves scizzors: thank you! loves scizzors: ^-^ fangedwolfshadow: lol dragongirlsirius: DragonGirlSirius: ooh kenny your my new friend! dragongirlsirius: nice name! i would have never thought of it loves scizzors: kenny dahling loves scizzors: hehe fangedwolfshadow: how fun loves scizzors: it was dragongirlsirius: come back to me please! kenny! i wish to mess up your head! loves scizzors: I wannt save the convo! dragongirlsirius: good idea! loves scizzors: teach him not to mess with us loves scizzors: heheheheeh 8:15PM dragongirlsirius: annie even talked to him loves scizzors: lol! loves scizzors: what she say? dragongirlsirius: he thought she saw the bf! dragongirlsirius: she was worried he might hack her account dragongirlsirius: she blooked him fangedwolfshadow: lol dragongirlsirius: bye fangedwolfshadow: she said he said hi honey to her dragongirlsirius: i g2g dragongirlsirius has left this chat. fangedwolfshadow: ok fangedwolfshadow has left this chat.  '),(21169,3745004,'26,July,2004','2004-07-26 00:00:00','   I went again today, so sue me. Partly because I wanted to go again, partly because I had a friend who wanted to go but under no circumstances would go by herself.. eenh. *mutterwhine*  Back on topic. So me, dad, and my friend Amy B. went. We also manage to drag along Boy, which made me happy. We get there, get tickets, go downstairs to wander. Wander wander. I\'m the only person there who\'s been there before, and similarly I\'m the only person there who actually will speak their own opinion. -_ NEVER  has an opinion anyway. *siiiigh* Off-track again >_>  First purchase: Azu Manga Daioh the manga #1 @ ADV Booth Cost: $8  It looked cute, I had seen the trailer and a music video that really amused me. I  had  to buy it. And I am SO not disappointed! ^^ Already read the whole thing, I love it to bits. I even got (gasp shock gasp) BOY to read it!! And his quote is: I never thought something so funny could be so disturbing.  Hee hee. I love Boy. He\'s an awesome beeg bruvver ^_^  Sooo, I\'m still trying to get input from Boy or Amy as to where to go next. I finally wander over to the Tokyopop store, other two in tow, because they don\'t seem to care.. >_  Second purchase: Marmalade Boy #1 @ Tokyopop Booth Cost: Around $9  To be honest, I like this series.. I\'m not quite sure why. Unsure as to whether or not I really  needed  to buy this, but.. I wanted to. Yes. Besides, I needed another thing to put in my uber-spiffy giganitc Tokyopop bag!! ^_____^ So, I bought it.  Yeah...  That\'s what I did. I finally decide Boy MUST see the booth of the people who made Munchkin, so I drag everyone over in that direction..  Boy\'s only purchase: Munchkin expansion set \'Clerical Errors\' Cost: $17  I know Boy enjoyed himself, because he bought something ^0^ Munchkin has gotta be the most HILARIOUS card game ever, I completely love it. And I still need to steal the new cards to read what they do- half of the allure of the game is all the cards, because they are so FUNNY. (+5 Chainsaw of Bloody Dismemberment.. +2 Boots of Buttkicking.. +3 Spike Codpiece.. XD) So.. yeah. Boy bought something!!!  Soo, it\'s lunchtime, so we meet dad over at the Penny Arcade booth. Go over to Hortons plaza, get some food (AMY NEEDS TO EAT MORE -_- That\'s my new mission: get Amy to eat REAL food, and NOT have just a glass of milk for breakfast! no frickin\' WONDER she\'s tired!), come back, buy Amy a new ticket after she loses hers and we failed at our attempt to sneak past the ticket-checkers. Eheh. -_-; Go over, watch anime. AZU MANGA DAIOH IS HILARIOUS. Ohmigosh. OHMIGOSH. I love it. \"Chiyo-chan, why are you flying?\" \"I\'m 10 years old!\" Dad\'s cracking up, even BOY is, as is the rest of the crowd. XD \"Why the heck? *whap* Why the heck? *whap* Why the heck? *whap* Why the heck? *whap* .............. Why the heck? *whap* Why the heck? *whap* Why the heck? *whap* Why the heck? *whap* Why the heck? *whap* Why the heck? *whap* ........................... Why the heck? *whap* Why the heck? *whap* ..... Why the heck? *whap* Why the heck? *whap* Why the heck? *whap* Why the heck? *whap* .......\" Heh. I, must, BUY!!!  Missions were failed, as much fun as I had. They didn\'t have anymore Princess Mononoke cards, I didn\'t find the cat-eared beanie I wanted, I didn\'t find the bunchy socks.. T-T Oh well. Next year!!  \'night all! *thunk*  '),(21170,3745004,'24,July,2004','2004-07-24 00:00:00','   I\'m not sure why I continue to post, when I doubt people actually bother to read it. e_e; I guess it\'s just the way I am. I like to log things, so that at later dates I can look back on these records and say \'Hey! I did that!\'.. yadda yadda, you know. Oh well.. it makes me happy, I suppose.  Well, yesterday was the Comicon. It was my first time going and, needless to say, I  loved  it. I would\'ve stayed longer than I did but.. I think the friend I went with was getting bored.. she\'s not as into anime as I am, although considers herself an anime artist.. enh, very confusing. Anyway, so we (we, being my dad, Katrina, and I) got there when the doors opened, but had to go get tickets. Noo problemo. So, we\'ve got our tickets, dad leaves us with a cell phone to go wander around. Finally we find our way downstairs, to the exhibit hall. Wow, what can I say about the exhibit hall.. it\'s HUGE! And  VERY  busy. So, Katrina and I start to wander the booths.  First purchase: A deck of Cowboy Bebop playing cards Cost: $5  Although the cards are starting to scare me with a few of their pictures. (If you\'re over at my house, ask me to show you what I mean..) So tomorrow since I\'m going again I\'m planning to get the Princess Mononoke set, just to be safe. XD So, more wandering. Swag, free stuff, = happy Bekka. I get a free painting in traditional- Chinese, I think it was -art style, with bold brush strokes. FREE! I love it.  Second purchase: The collected Tall Tails volumes 1-3, just because the furries are cute, and Egoraven because according to Daphne Lage there\'s only 25 copies out Cost: $15  This was pretty much a feel-good purchase. I liked the art, I had money, I bought it. Nothing really special to it. So, we wander on a bit more, and come across a LotR booth selling authentic-looking elven knives. Katrina just HAS to get one, \'cos she\'s such a LotR nut. I HAVE to spot her 11 bucks, \'cos she\'s low on cash. e_e;; But I don\'t really mind, so long as I get paid back. I mean, the knife is really awesome. XD Wander down to the other end of the convention hall, by now my bag is STUFFED with swag. Which made for a very happy Bekka ^_^ Apparently we missed the Megatokyo store the first time, so we figure out its number and go to visit it.  Third purchase: Megatokyo \'ph34r t3h cute ones\' poster, autographed by Fred, Sarah, and Dom. Cost: $14  This is probably the thing I loved most. I mean, AUTOGRAPHED! Besides, I\'ve been wanting the poster. What better time to get it? I was seriously considering getting Boy a shirt, because I\'m a nice l\'il sister, but- c\'mon. Nineteen BUCKS?! I\'m not THAT nice! So I didn\'t get him anything. ^_____^  Got lunch, ate lunch, went upstairs to watch anime until around 3, when Katrina wanted to leave. I had a complete and total AWESOME time. Seriously guys, it was so fun. There was even someone cosplaying Tsukasa (with paper mache staff and everything!), and two old ladies as Vash and Inuyasha! (I might add, Inuyasha had toilet paper stuck to his/her sandal.. >_>) It was so fun, I can\'t wait until tomorrow. I get to go again!  G\'night, internet!  '),(21171,3745004,'23,July,2004','2004-07-23 00:00:00','   Trip report from Arrowbear.  Arrowbear . Arrowbear Music Camp. Woohoo!  11 days up in a pile of sticks, near Arrowhead & Big Bear. Thus the name. I\'ll try to keep my report short, but then again I\'ve got a lot to say.. and I want to be able to read this at some later date and remember all the cool stuff. So, well... if it doesn\'t interest you, you don\'t have to read it. Also- it\'s like, near impossible for me to put all that I want to say in any coherent order, AND I have forgotten days for many of these things. So.. uh. Good luck!  -~-~-~-~-~-~-~-~-~-~-~-~-~-~-   Day 1, Wednesday - Arrival. The place looks so foreign! It\'s really crowded, near impossible to park. While mum dear registers Katrina and I, we go find a bunk. Main dorm- no luck. So, down a flight of stairs to a little porch with bunks placed on it. No luck, so down a bit further to the.. pile of sticks! A cabin on the porch, in the minimalist sense of the word. Rectangular cutout for a door, square cutouts for windows with loose screens on them.. except for one. Guess who gets the top bunk where they can be assaulted by mosquitos flying in through the open window? Mmmhm. Me. Katrina\'s on the bottom bunk. So, we dump our junk, audition, and get to know the people there. There\'s four bunk beds in the cabin (not room for anything else >_x) If coming in from the main porch, my bunk would be the far left. On far right was Taylor (bottom) and Lois (top). Near left was Lindsey (bottom) and Angie (top). Near right was April (bottom) and Grace (top). Outside of the door near to me was a little \'second\' porch, not the main one. As with the other, this porch was roof-less and wall-less. If coming from Cabin 1 (aka Pile \'O Sticks) near left on the single bunks was Liz the counselor, near right was a bunk on which Liz dumped her junk. Far left was Gwen, far right was the bunk on which Gwen dumped her junk. Then, at the end at a perpendicular to the other bunks, was Annie. And those were the porch-ers and the Cabin 1-ers. Although technically Cabin 1 was the people in Liz\'s group, which would be Liz, Annie, Gwen, Katrina, Lois, Taylor and I. Instead of having a cushy dorm like the rest (excluding the other -nicer- stickpile and the main porch) we had to face the elements and other disasters. e_e; We had a sort of pride thing going on by the end of the session. ((According to them this was a gianormous session with 150+ kids, I get the feeling they don\'t use the porch much))  First dinner was spaghetti. Our group sat in a circle on the ground, got to know each other after being forced to by the counselors, then learned about Liz the counselor. Liz is nearsighted, 21, and.. uh... stuff. I don\'t exactly remember anymore.  And thus ends an eventful day..   Day 2, Thursday - All that really stands out about this day is, when waiting in line for breakfast, I almost passed out. No, I\'m not kidding. I guess it was the elevation + lack of food + lack of drink + lack of sleep that did it. But, when standing on the porch outside of the mess hall, I felt very sick. No, VERY sick. Soon I felt dizzy, then black started creeping in from the edges of my eyes like has happened before- but that\'s more or less normal for me. Anyway, according to Katrina I took a step backward and fell into a tree. (( urlLink THE TREE )) But I was saved from falling by Katrina and one of the counselors.. Jeff, I think. After that I surprisingly felt better, although directly after regaining my footing the whole world was in black and white, I\'m not exaggerating.. I wasn\'t out for any amount of time, I didn\'t even lose consciousness. But, man.. it was exciting! After the fact, I can honestly say I sort of enjoyed it. Yes, yes, I am weird. But it was an experience. The rest was just a normal day, pretty much. We learned where we sat in band. I\'m a 1st chair. We had 1st chairs and 2nd chairs. Although the unofficial section leader of the alto saxes was Alex, the scarily-good owner of a silver saxophone. I sat on the end, next to the bassoons and this guy, uhh.. I think his name was Mike. Yeah. That was my day.  We also had an egg drop.. no eggies lived T-T LONG LIVE THE TREE! Half the counselors were \'aliens\' and half were \'dinosaurs.\' It was scary. Ergh-eth!  Oh, and a basic day at Arrowbear. Although I don\'t have all the times..  7:15 AM - Reveille courtesy of Toshi the trumpet counselor. Prepare for breakfast, which starts as soon as you haul yourself out of bed. 7:15 AM - Breakfast (KP on days you had KP) 8:00 AM - Clean-up (or bathroom duty, if you have it) ?:?? AM - Morning Assembly ?:?? AM - Rehearsal #1, lasting for an hour ?:?? AM - Break ?:?? AM - Rehearsal #2, hour-long 12:00 AMish - Lunch (KP on days you had KP) 1:30 PM - Lunch Assembly 2:00 PM - Rec Time (woohoo!) Usually lasts for two-to-three hours. During this time there was optional crafts, hikes to Happy Gap, and the snack bar was open. Also on certain days the pool was open. ?:?? PM - Rest (One hour, I believe). According to Liz you could either read, listen to music, sleep or pretend to sleep. e_e; Around 5 or 6 PM - Dinner (KP on days you had KP) ?:?? PM - Evening assembly ?:?? PM - Rehearsal #3, one-hour ?:?? PM - Evening program, variable 9:30-10:00 PM - Campfire 10:30 PM - Taps again by Toshi  Busy busy busy!   Day 3, Friday - (I think) KP. You get to cut the lunch line, but you also have to wash the stupid dishes. I had to put away pots and pans that had been washed, rinsed, and sanitized which really wasn\'t as easy as it might sound. I think I got the job \'cause I was tall and could reach where to put things. Anyway, we got a rather bad boy\'s group to work with (one girls cabin group and one boys worked on KP at a time). Of the four stations for pots and pans one was occupied by me, one by Katrina, one by drummer Clay, and one by viola Daniel. And do you want to know what those stupid little sevvies did to us? Got us SOPPING wet in sanitizer water!! I swear, at one point in time then took the sanitizer water nozzle, pointed it at Katrina and I, and turned on the stupid flow! As well as splashing us endlessly. It took ages for anything to get done. And I took several (if not more) sponges upside the head. At one point I calmly picked up the sponge that had been thrown at me, walked over, and wrung it out over Clay\'s head. It was very pro of me. The counselors applauded. Oh- that reminds me. The counselors were standing by, LAUGHING at us as we got soaked!! Grr... \\_/;;  And it was even worse. Viola Daniel had, at one point, gotten hold of a pair of tongs. And, instead of putting it in the sanitizer like anyone SANE would do, he started banging it against the metal sink rim singing at the top of his lungs, \"Na na na na, na na, na na can\'t touch this!\" AAARGH! I was dreeee-eeee-eeenched by the end of KP. Literally fscking dripping weeet. It pissed me off so badly T-T.   Day 4, Saturday - Bathroom duty. I had to do the toilets, which were surprisingly clean. So I was lucky. Now this gives me the perfect opportunity to describe the bathrooms.  They weren\'t BAD, per se.. but they could have been a hell of a lot better. There were 6 toilet stalls, only one having a lock that >worked   Day 5, Sunday - Inspirational. I had to draw a picture of someone who inspired me. So, I drew Boy. Okay, you\'re giving the monitor strange looks, but he\'s the reason I got into music in the first place! (He took up the trumpet in fifth grade while I was in third, I was jealous, by the summer after 3rd I had taken up the saxophone. Not exactly ideal, but..) So, that went well. Ish. Well-ish. Because we had to share in front of the rest of the camp.. But I found it amusing that almost all the boy cabins did skits while the girl cabins did artistic things. Heh-heh-heh.   Day 6, Monday - The first hike-by-number! With Kate and (I think) Ted we split into groups to do various activities. And so today was with Kate and.. Ted? I think. I think. So we played the Shadow Game. Two people had to work together to create shadows.. e_e; Boring.  Oh. And I think today was a musical.. but it might have been the second one o_O; People can sign up to play a piece in front of the rest of the camp. Yuss.   Day 7, Tuesday - Hike by number #2! I believe this was with our very own Lizzy and Jeff. We had to play this stupid counting game, where people try to count to twenty. People cannot indicate the order in which they go, and you cannot go at the same time as someone else. Veddy confusing. -_-;  During the lunch assembly the counselors taught us dance moves, since today was a dance. I\'ll list them, because they were really amusing.  \'The Kate\' a la Kate: Moving two fingers (disco style!) in random directions. \'The Mickey Mouse\' a la Toshi: Stepping from one side to another. If very daring, adding a spin. \'The Security Guard\' a la Brian: Arms crossed. Bored expression. Nod head to music. Look at watch occasionally. \'The Pogo Stick\' a la Brian: Put hands at sides, palms downward.. pause.. JUMP LIKE CRAZY IN RANDOM DIRECTIONS!! \'The Shopping Cart\' a la Toshi: Put hands out in front of self as if holding a bar. Walk along casually. Reach up and mimic pulling things down from shelves, putting them into a cart. Variant of The Shopping Cart.. \'The Shoplifter\' a la Toshi: Nervously walk, looking around self. Grab items, stuff into jacket, run for it!! Variant of The Shopping Cart.. \'The Lawnmower\' a la Toshi: Pull chain to start up lawnmower. Start walking along casually as if pushing a lawnmower. \'The Maniac\' a la Liz: Put arms out in front of self, together. Bend over slightly. Run in place. \'The T-rex\' a la Katy: Bring arms close to body, as if a tyrannosaurus rex. Two fingers on each hand, bent. Move one arm (and inadvertently whole shoulder) forward like attacking prey. Switch arms.  There were two dances during the session. Unfortunately, they didn\'t tell us, so I only had camping clothes to wear.. e_e; My first dance, eee! Yeah, I\'m pathetic. But I haven\'t been to any of the school dances. And besides, I\'m pretty much a flop when it comes to dancing. But I enjoyed myself, and yes.. loud music, it was nice enough.   Day 8, Wednesday - Hike by number #3! This was Brian (I think. Either Brian or Ted..) and Laurie. If that was her name.. >_>; Trust walk! It was amusing. You close your eyes and put your hands on the shoulders of the person in front of you. They do the same, you get the point. Then the person at the head of this (congo!) line, who has open eyes, leads the whole bunch on a hike. It was.. haaard. _._   And this was a musical, I think I think I think. If so it was the last one.. yes.   Day 9, Thursday - Hike by number #4!! With Tooooshi and Ana-banana! We had to sit in a circle and.. one by one.. go hug a tree. -_-;;;;;;; Then we had to say our name, where we come from, and our favorite band. THEN we had to say what the one thing we would change about ourselves is (Mine: procrastinating! You\'ll get a kick out of THAT, Sheep >_>). Fuuun, I guess.  ...And... Oi. Second dance.. I\'m sort of leery putting this in, but.. *swallows pride and continues* So we were dancing in our little group, right? A slow song comes up. Then suddenly Taylor runs off without any before words right up to David the Tuba, who >every The only words that passed between us= Him (trying to break the awkward silence as we dance): Having fun? Me: I guess. Him (a bit later): So you play alto saxophone? Me: Yeah. I mean, could I be any more of a loser?! Uhh, well, anyway, I didn\'t meet his eyes the whole time, I was that mortified.. I mean, really. And at one point Toshi, who like the rest of the counselors is frolicking around trying to get everyone spirited, runs up to us and puts one hand on either of our shoulders. \"Move closer!\" Us: *does nothing* Toshi: *runs off* .....e_e;  For the rest of the dance I was so quiet and depressed it was scaring even me. (Taylor ended up pulling the \'SET PEOPLE UP WITHOUT ASKING THEM FIRST\' thing on another of our friends.. then got angry cos people didn\'t want to be set up..) Uh, and anyway, when the second slow dance came up I was just milling around, and the OTHER David (the cool David who likes Cowboy Bebop and can Edwalk), the David Trumpet, was all \"Uhh, so, who wants to dance with me?\" Katrina looks at me, says \"She does,\" then wanders off to dance with someone. So, I danced with the other David, which was sort of okay \'cos he\'s cool if not hawt. But he liked Gwen, not me, so.. sad Bekka T-T.  And.. that was that..   Day 10, Friday - The last hike by number, #5! We went with Kimi and Cormak down to the campfire, where we played chief. Pretty much there\'s a circle, with one person off to one side with eyes closed. A chief is chosen, and they start doing an action which everyone copies. The person who had their eyes closed comes back to the circle, trying to figure out who everyone is following. They have three guesses. And to make it harder the chief changes actions often. It was amusing enough.  We had KP again today... but with a different group of guys. We did KP with David the Trumpet\'s group, which was sort of okay.. heh-heh-heh, with Muscles the Stalker too. Um.. LESS MESSY. Thank you.  COCONUT BEAR NIGHT!! All session long our Cabin had been working on a skit, and we did it for Coconut Bear!! Our group, dubbed Oblique Moose 5, was SO cool. We did a scene, with different themes. Okay, here\'s the normal version:  Katrina: *walking in the \'forest\'* Wow. Look at the trees, they\'re so green! The sky, it\'s so blue! I think I\'ll have a picnic. *sits* Taylor: *weaves her way over to Katrina* Sss! I am the snake! Sss! *uses hands to \'bite\' Katrina* Chomp! *moves offstage* Katrina: Ouch, I\'m bitten! I\'m getting dizzy, I\'m... *falls over* Gwen: *walks onstage* Hello? Katrina? Are you- ohmigosh! What happened? Are you okay? I\'ll call an ambulance! *pretends to call 911* Bekka: *waves arm above head as walks onstage* Weeooo weeooo weeooo! *pretends to step out of ambulance* What\'s the problem here? Gwen: She\'s been bitten by a snake! Bekka: Oh, this is very serious.. I\'ll take her pulse. *kneels and pretends to take pulse, then looks up* I think... I think she\'s dead... Katrina: *sits up* I\'m not dead!  Then we did.. VALLEY GIRL VERSION! Katrina: *walks preppily into forest* Wow!! Look at the trees, they\'re like so.. uhh.. PURPLE!! Annie the tree: Thanks, but I\'m green. Katrina: Oh, uh, green! And the sky, wow I feel really dizzy now.. Well like I need to eat because it\'s like lunchtime, hahaha! *sits* Taylor: Like Sss like ss I\'m like the snake! I\'m going to totally bite this person! Choo-ooomp! Katrina: Ow, that hurt! Wow, what\'s this red stuff? Hey, it looks good on me, hahahaa *falls over* Gwen: Hellllooooo? Like, heeee-llloooo? Like, what are you like doing like lying on the like ground? You\'re like going to like totally like ruin your hair like! Are you like okay? Like, are you like ookaaaayyy? *pause* I\'m going to like call an ambu... and ambula.. and ambulalala.. a pharmacy car, like. *dials* Like, hello? I\'ve got a serious problem like.. Yes.. Yes.. You guys don\'t know how much this like means to me.. Yes.. Nail appointment on Tuesday? You\'re a lifesaver! *hangs up and looks down at Katrina* Oh right, an ambu.. 911. *dials again* Hello like? My like friend is like hurt and like needs your help like. She\'s lying there like on the ground and like she\'s.. uh.. redding. She\'s redding. Thanks, byyyeee! *hangs up* Bekka: wooEEEEooEEEeEEo, wooEEeeoEEEeee! *hops over* What\'s the like, problem? Gwen: Like, it\'s YOOOU!! HIII! Bekka: Ahhh!! HiiiiIIIii!! Oh right, the like problem. Gwen: Oh, right, like she\'s just like there and like.. um. Bekka: Ooooh, that\'s baddd. She\'s lost a lot of blood. Gwen: Oooh, that\'s what it\'s called! Bekka; Yeeeaaah! *kneels and takes pulse* Oh oh.. I think she\'s like, a goner.. Gwen: Ohmigosh! But she\'ll like miss the like cheer competition! Katrina: *bolts up* OMYGAWD THE CHEER COMPETITION!!  Then we asked the crowd. They voted Pirate version. But I\'ll let you guys imagine that one o_o;! Everyone LOVED it! XD And we had done, when practicing, versions like Horror Movie, Monotone.. it was so fun.   Day 11, Saturday - Last day T-T Packed up, was DYING to see Boy & Parents again. They came, we concerted, then left. I MISS ALL OF THE OBLIQUE MOOSE 5!!!   First thing I had once returned to civilization: FRAPPACHINO!!!  -~-~-~-~-~-~-~-~-~-~-~-~-~-~-  IMPROV Since I didn\'t include it above, I will here. On most days during lunch assembly, the counselors had an improv skit. Well, not skit. Improv game is more like it. I\'ll try and recall some- they were HILARIOUS.  Bus stop- This was probably my favorite. The point of it was two people sat, as if at a bus stop. You tried to make the other person laugh by having such an ODD quirk. So, first it was Toshi and Ana-banana. Toshi was acting all bipolar, calm one moment and screaming the next. I\'m not sure what Ana was. Then finally they switched, and.. wull I don\'t quuuite remember. But up came Brian. Brian ROCKS. He was sitting there, hunched up a bit, pretending to very messily and strangely eat something. So someone comes up, sits. And Brian just looks over. And in serious the ODDEST voice ever: \"Do you want an anchovy?\" XD Like FIVE counselors attempted to go through that without laughing. Brian made \'em all laugh. Then STACY comes up and (this is hilarious). She walks up and starts talking. \"Hello, I\'m waiting for the bus so I can see the doctor. Because whenever I start to talk I people I just-\" She falls over as if fainted then gets up \"I just seem to-\" falls over again, but has managed to hold onto \'bench\' \"Fall over and I can\'t quite-\" has sat now, continues while staring at Brian \"Quite seem to stop it I just-\" She falls over, onto Brian XD It was SO hilarious. (Splat.. she was splatting.. o_o; )  Party- How to play: There\'s four people. Each person must have a mood like energetic, depressed, enraged, even constipated was one o_O. Just a mood. One of them is hosting a party. You choose what type of movie. So they\'re all acting they\'re mood when the first person knocks at the door. It\'s answered, and the new person comes in. Now BOTH people are acting like the mood of the new person. They talk for a bit, then another knock. The third person enters, and again there\'s a mood change. Same with the fourth. Then as one by one they leave, the mood switches to the third, to second, until it\'s only the first person acting as they had been. XD Very amusing.  Wull.. there\'s more but you don\'t need to hear about them.  -~-~-~-~-~-~-~-~-~-~-~-~-~-~-  I\'ll end this with piccies.. cos I really don\'t want to do any more heavy typing. Suffice it to say I had a BLAST!!   urlLink View From My Bunk 1  Facing the door to the main porch  urlLink Window-shot!  Out of the UNSCREENED window  urlLink Door-shot  Looking at the main porch.. you can see the NICER pile of sticks, barely  urlLink Another Door-shot  Looking out the door that leads to the small porch..  urlLink More Bunks  Far right and the tip of near right, as seen from far left.  urlLink People  Grace, Lois, and the back of Katrina \'Trinket\' \'s head!  urlLink Taylor  Thissis Taylor, the fellow manga-loffer!  urlLink More People (movie)  First shot: Lois (left) and Grace (with camera). Respective instruments: Flute and trombone. Pans to left to find Taylor, oboe!  urlLink Sunset  Ooooh... pretty.  urlLink 1  The tree.  urlLink 2  The TREE.  urlLink 3  THE TREE!!!  urlLink Flight of The Tree (movie)  Incidentally, The Tree died as well.  urlLink Before Rehearsal  Grace and Lois!  urlLink Working On Inspirational  Taylor and Katrina  urlLink More Inspirational  Annie!  urlLink Even MORE Inspirational!  Gwen and Lois  urlLink Blackmail  Katrina with her hair down. Need I say more?  urlLink Inspirational Again..  I love that expression.  urlLink Breakfast  Yay mess hall  urlLink KP(ugh)  Don\'t they look like they\'re having fun?  urlLink Pots \'n pans  Ew... vv  urlLink Washing thingy  I\'m really diggin\' this KP stuff.  urlLink Slackers  This is what counselors do while we slave away.. grr Jeff and Liz!  urlLink Food  Mmmmmm....  urlLink People at KP  David the Bebop\'n Trumpet (left) and Lois  urlLink Muscles (movie)  >_> the stalker..  urlLink Kitchen Tour #1 (movie)  From KP to the mess hall..  urlLink Kitchen Tour #2 (movie)  ...And from mess hall back to KP  urlLink Assembly  Pep Band Friday!  urlLink Assembly 1 (movie)  Kate has awesome hair. Es purple.  urlLink Assembly 2 (movie)  They made us do this EVERY STINKING DAY  urlLink Tour of the Pile of Sticks (movie)  Eeeee! Coming in from the main porch, ending in the small porch.  urlLink Rehearsal  This is where I sat.  urlLink TROMBONES  *covers eyes* BRASS T-T  urlLink Tuning (movie)  Es da stringies tuning!  urlLink Counslors  Liz, Cormak, KATE!!  urlLink Brian  Veddy stoned.  urlLink More Counslors  Brian, Jeff, and da Liz  urlLink Ana-banana  I love her makeup  urlLink Coolest Building EVAHR  Snack shop... >_>  urlLink Fuzzy  It\'s TED  urlLink Craig  Duuuuudeh, Craig rocks! He\'s the band director!!  urlLink Kimi  Kimi the hair-model counselor..  urlLink Angie  Fellow pile o\' sticks-er  urlLink Dennis  Owner of the camp  urlLink Wall-art  >_>;; I swear we didn\'t get in trouble, Liz wrote her name as well after I had taken this picture...  urlLink Happy Gap  Tres cool  urlLink Group Piccy- Lizzeh\'s group!  ACK!!!! I LOOK HORRIBLE! *cowers* Bekka... not photogenic AT ALL.. *shuddar*  urlLink Toshiro  ....-_-; We were sitting on the ground, eating. Toshi walks past. I needed a picture of him. This is what I got.  urlLink Counslors again!  Liz, Katy, Jeff  *whew*  -~-~-~-~-~-~-~-~-~-~-~-~-~-~-  Please won\'t you catch.. a shooting star for me.. And take it with you on your way. Though it seems that we\'ve just met You\'re the one I won\'t forget Till some kind wind blows you back my way  And I was thinking maybe someday later down the road After all our stories have been told I\'ll sit and think of you The dear friend I once knew Who shot through my life like a shooting star.  Sometimes I know, that a part of you will show Deep in my eyes or in my smile. There will always be a part of you deep inside my heart And I\'ll know just when to let it go  And I was thinking maybe someday later down the road After all our stories have been told  I\'ll sit and think of you The dear friend I once knew Who shot through my life like a shooting star.....   Arrowbear 2004  '),(21172,3745004,'21,July,2004','2004-07-21 00:00:00','   Anything artistic like stories, etc. will be housed @  urlLink B-i-r-d-l-a-n-D . Just so I can keep mildly interesting things like creative works seperate from my mindspace. ^_^  '),(21173,3745004,'18,July,2004','2004-07-18 00:00:00','   Ryu is back. Will type up a hella long camp report eventually. '),(21174,3745004,'06,July,2004','2004-07-06 00:00:00','   Off at Arrowbear for 10 days, leaving tomorrow. If I don\'t see any of you before then, I\'ll be thinking of you guys while I\'m away...  '),(21175,3745004,'04,July,2004','2004-07-04 00:00:00','   Despite the fact that I really can\'t get into the whole \'Mindless Patriotic\' spirit, I do appreciate how completely American fireworks are.  After all, what\'s more American than explosives, loud noises, sparkly things, and excessive pollution?  '),(21176,3745004,'04,July,2004','2004-07-04 00:00:00','   You know..  I really wish that I was the kind of person that had so much to say. Like, the type of things that really matter.. the types of things that would get somebody to really think hard, to start reflecting on their inner self and all of that good crap.  Well, you know what?  I\'m not.  And that\'s what makes me so.. so angry at myself. No, I\'m not a very deep person! Or, at least, I don\'t show it or know it if I am.. I\'m so unsure of myself. I\'ll think, \"Oh. I realize that I\'m not deep. That must make me deep. But wait.. Maybe everyone realizes this, and I just don\'t know it? That would mean I am in reality shallow. Or maybe all deep people think they\'re shallow and have the same thought process as I..\" It\'s just.. going in circles. Does that mean that there\'s something wrong with me? Or does everyone think like this? Am I just trying to fool myself into thinking that I\'m better than I really am?  Maybe it\'s just that I worry too much about these things. But lately I\'ve felt sort of numb inside. Maybe I\'m not getting enough sleep, and it\'s making me hallucinate that I have depth to myself.. or maybe I really am becoming apathic? Is it my tired brain trying to confuse me even further? Well.. I don\'t know. I just don\'t know. But, I\'m so.. so..  Dull. Empty. I\'m like a shell, although I bet I\'m not and just wishing I was. At least then I might be at least slightly interesting as a person.  Is it that I think that people actually care about things like this? No, I don\'t think they do. I\'m just whining about my lack of thought like the stereotypical Livejournaler..  *sighs..*  Am I really that shallow?  '),(21177,3745004,'17,August,2004','2004-08-17 00:00:00','   Soo. Yesterday I went to the mall with Anneh, bought a ton of crapola (and Tarot cards!), stayed up \'till about 3 am.. 4 am? Ennnh.. whatever. Spent a long time on the subject of \'food\' (XD still hungry, eh glutton? well, there\'s a seafood place, maybe you should- *falls asleep*) and also came up with a rather amusing thing concerning X-men.  Bekka: You know that part in X-men? Where, like, Mystique sticks metal into guy\'s butt? Anneh: He has... buns of steel!  Crude due to tired-ness... none the less hilarious XD  And, meh, deepness in general. ^_^ Thhheeeen I came home, got haircut (It\'s.. SHORT! *dies*) and basically lived life. Yannow?  On a completely unrelated note, I am about as angry as can be... for reasons... not at anyone who\'d actually bother to read my blog, don\'t worry XP But... Well,  some one knows what I\'m talking about.... *patience snaps* Grmmm... they don\'t... get it... Damn Fuzznuckers..  ~Ryume Of course, you know,  this  means  war !  '),(21178,3745004,'15,August,2004','2004-08-15 00:00:00','   -sheep- (5:36:11 PM): I want to throw my compy out the window. -bekka- (5:36:13 PM): XD -bekka- (5:36:14 PM): DO so -sheep- (5:37:06 PM): But it\'s my preeeeeeeeeecious. -sheep- (5:37:15 PM): And \'twas expensive. -sheep-(5:37:33 PM): Do you think if I were to throw my compy out of the window in frustration, I could sue AOL? -bekka- (5:37:37 PM): Yes! -sheep- (5:38:57 PM): YAY! -sheep- (5:39:00 PM): -throws- â€œ-sheep-â€ signed off at 5:39:10 PM.  -------------------------  Edited out screennames, just \'cos..  It amuses me.  ..... *walks off humming theme to Bill Nye the Science Guy*  '),(21179,3745004,'15,August,2004','2004-08-15 00:00:00','   Buuuuuut I had nothing to write.  So, I\'ll babble. I like to babble. Yodels and hopefully Sheep (we still need to ask her) and I have started a blog with links to virtual pets sites and stuffs like that.. um.. link in sidebar. *point* Needs work, but uh.. fun? ^^  And Azumanga Daioh is so amusing, I can\'t even begin.. watch it! NOOOOOWWW! *flail*  And, uh, stuff. I\'m bored. vv;; Nobody on CC, nobody on AIM, waaaah!  *thunk*  '),(21180,3745004,'14,August,2004','2004-08-14 00:00:00','   Back from the beach. Not much to say, really.. too tired. Uh.. leetles. Loud. Babysitting = money. Sunscreen is God. Might post more later, then again- maybe.  G\'night. *flop*  '),(21181,3745004,'06,August,2004','2004-08-06 00:00:00','   Explaining my \'mysterious\' almost two day disappearance:  Thursday I was banned from the computer. Yeah, nothing special there... \'tho I want to post about my day, because I  can .  Um, so, pretty much a friend comes over- Problems there, I was being such a bitch, but then again.. so was she.. and um vv;; Gonna have to apologize, anyway -then it\'s my lesson. After my lesson I go shopping for a one-piece swimsuit but end up getting this classy jacket for formal occasions and a white shirt to go under it, because.. um.. low neckline grrmm grmm guh. I also get a black tank top, because I  can .  Come home, time for Bekka to act domestic. -_-;; Cooking almond roca (carmel covered in chocolate and almonds- YUM), in an APRON. XD It... amused me, no I don\'t know why. But... APRON! :6 So, yeah.. too bad I don\'t actually get to eat the almond roca >_> It\'s a \"Thank-you\" gift for the Clows family.  Annnnnnd today I was over at Diane\'s house being domestic again, baking cheesecake with Diane for when we\'re at the beach.  That brings up something important. Tomorrow I\'m leaving to go to the beach for a week. I\'ll miss you all! T-T *huggles everyone*  Baaaaaiiiiieeeeeeeeee.  '),(21182,3745004,'04,August,2004','2004-08-04 00:00:00','   Well, I suppose I never really had been quite content with Sasuke\'s  urlLink original picture  (NOTE: Not my art. A link to where I got it from is coming, don\'t worry), so I went back to the site I got it from for two reasons. 1, I needed to find out if he was from a story or anime. (Point of interest: He\'s just a created character namd Alex, apparently) 2, I wanted to see what else the site had, because I was just scanning the place when I found the first picture. So, I was off to the  urlLink art gallery  of  urlLink BakaNeko  to see what I could find.  And I wasn\'t disappointed!  I stumbled across a  urlLink very interesting picture . (Again: Not my art) The character on the left caught my eye. It was promising, but there were two problems. The colors, and the other person in the picture. But he had the right shape to his face, and his clothes were right as well. Big problem? Well, I could always try and rely on the general populace of short-attention span roleplayers to read my profile once it came. But.. you know, people don\'t  do  that these days. They just look at the picture. So, it was steadily becoming more of a problem. But I can be  stubborn  when I want to be.  So what I did was dragged the picture into my trusty version of Photoshop. Short while later I came out with what I wanted.  urlLink Success!  It\'s still awaiting the final \'OK\' from the owner of the character Sasuke, but.. I dunno.. I think it\'s better than the one I had in the beginning. Much more \'Sasuke\' to it.  But pretty much all I\'m trying to say here is that Photoshop Is God.  '),(21183,3745004,'03,August,2004','2004-08-03 00:00:00','   I want:  urlLink This  and I want  urlLink This .  *counts money in wallet* ...T-T I can only buy the second one right now.. I\'d buy both but I have to save this month\'s allowance for next week, when I\'m off to the beach for a week with Diane.. *bugs father to order dvds*  '),(21184,3745004,'02,August,2004','2004-08-02 00:00:00','   1) I got a picture finally. T\'is Nyamo. I like Nyamo. Nyamo is very cool. Nyamo Nyamo Nyamo. Don\'t you like Nyamo? I do. ^-^  2) I need to take a poll. Am I the ONLY one who enjoys summer? o_O; All of my friends are like all \"I can\'t wait for school to start\" and \"Sssssoooo bored\"! T\'be poifectally plain I love summer to bits and wish it was longer >_  Err... or something along those lines..  *thunk*  '),(21185,3745004,'01,August,2004','2004-08-01 00:00:00','   Last night I did something I don\'t normally do. I turned off the computer early, even before Boy did, and went to go beat that  one level  on Tetris Attack. And because it amuses me, I shall relate to you the (mis)adventures of Bekka and the Super Nintendo.  So anyway, this level is the last one of the Stage Clear mode, where you beat levels one by one as they get progressively faster. Not hard, really. Well, that is, the levels leading >up  So, right, I\'m all hyped up for the last level. And then it comes... BOWSER!! Roooar. The flatulent King of the Koopas himself. -_- So, by now I\'m thinkin\' it\'ll be a breeze. (No pun intended) Then comes the chiller.. \"BWAHAHA! IF YOU DO NOT KNOW CHAINS AND COMBOS YOU ARE WASTING BOTH OF OUR TIMES!\" -_-;;; Isn\'t that dandy? He\'s got \"health\" and I have to use these \"chains and combos\" to \"attack\" him.  Well, it still doesn\'t seem  too  hard. I mean, chains and combos are easy, y\'know? So I\'m thinking \"Hah. My power of justice shall smite thee, yon toady-beast.\" You know, just cos it was easy to win the other levels. So Bowser laughs maniacally, the game starts...  Game over.   Game over ! I mean, the speed level was only 50, I had gone up to 100 before. So I hit the \'try again\' button.  Bowser laughs evilly. The colums of blocks start to rise again.   Game over!!!!  -_- So, I try again.  Same thing. Mwahahaha. Game starts. Game over. This goes on for a while, by now every time I lose I lapsed into Ed-mode. Screech unintelligible words, flail for a few seconds, angrily mash the button to try again. And every so often I\'d get a comment from Boy, \"Deep breaths,\" or \"Calm down.\" At which I would either pretend to hyperventilate or screech at him. I mean, I was getting killed by a  turtle on steroids !! I had a right to be torked.  \'cos, I mean, I\'d like get him way past half health then the stupid speed level would increase. And increase.. and..  I think I eventually started saying some rather unsavory things about Bowser\'s mother. Or at least I would have, but my mother was sleeping and wished to remain as such. So I just shook my fist at the tv. Repeatedly.  And it went on like this for, uh.. 45 minutes? I\'m sort of embarrassed to say that XD but it\'s true.  Mwahaha. Game starts. Game over. Screech. Flail. Try again.  Then finally I got a 4 chain and whipped Bowser\'s scaly tush. And lemme tell you, four chains are HARD. It\'s like, the set-up has to be PERFECT or you have to be pretty damn lucky. So I\'m thinking,  What do I win, what did I unlock?  Cos after that you damn well WANT to win something or unlock a new character or  something.   ...I got the ultimate privilege of watching the credits...  *screech, flail*  '),(21186,3792732,'29,June,2004','2004-06-29 00:00:00','   Hello all!  A friend recently sent me an invitation to join a blog, but for some reason I kept getting an error message. Anyway. So I went to the website, took the little tour, and was tempted to start my own blog! Resisted this temptation, however. But recently read about a man named Mil Millington (yes, Mil Millington, if I\'m remembering correctly) who turned his personal blog into a bestselling novel. You know, fictionalized it some. He\'s English. But this sounds like a freak accident. He was sitting at home one morning and Random House (or whoever) called and offered him a book deal.   Grr!!  So cannot blog til sundown every day and hope this happens. That would be nonsense.   Already the allure of a potential audience makes me giddy with questions: 1) Would people read my blog? 2) Would people like my blog? 3) Would people like me after they read my blog? 4) Would my blog be interesting to others or just clog cyberspace?   None of these questions are answerable, really, without first blogging. I myself have run across people logging their blog with gluttonous abandon, writing to all and sundry who will listen about late-night parties, skipping school, the odd romance, their day at the beach, what they had for lunch, and whatnot. And reading what an anonymous 15-year-old had for lunch is not all that interesting.   Or is it?  This is the frightening thing about the kinds of things I write, actually. When it comes right down to facts I\'m not at all sure the stuff I write for fun is sellable as a memoir/novel/short form item because more often than not I am talking about what I had for lunch (if metaphorically).   And for some reason, I know people who I wouldn\'t want hovering around my blog hoping to read what I had for lunch.  Is blogging for me? For all my apparent choice-driven inconsistencies, it\'s still tempting.  LA   '),(21187,3792732,'30,June,2004','2004-06-30 00:00:00','   Hello all!  My sister Nancy and I went to a bath house recently. It\'s a fully operational bath house in Hot Springs, Arkansas, a town famous for it\'s natural hot springs. In the early twentieth century the Hot Springs bath houses were flocked to for the \"curative\" powers of the naturally hot water. I visited Buckstaff, which is the last original bath house still operating today on bath house row. At least 4 women attendants saw me completely naked from all angles. I had to wear a thin white sheet between bathing stations that was soaking wet most of the time.   I sat in 100 degree water in a huge porcelain tub, scummy looking liquid soap foaming from my outboard-motor whirlpool and crusting around skin that broke the surface of the water. (This after being loofah-scrubbed by a plain girl named Tonya, four inches of water in my tub while she adjusted dials and fixed temperature, water level rising up the porcelain. We accidentally made eye contact as she instructed me to lay back, loofah scratching at my knees, my face undoubtedly flickering between discomfort (sandpaper loofah) and embarrassment (being completely naked). Tonya leaves and I\'m there 25 minutes, feeling my muscles relax and I try not to imagine the thousands of women, many seeking cures from diseases with the hot springs, who have bathed in this tub since 1917.  I sat in a mildly humiliating sitz bath cubicle (Me: \"What is this?\" Tonya: \"A sitz bath. You put your bottom in it.\" It\'s like sitting in a sink.) facing the bustling baths unshielded by a curtain, my front covered only by a stretched-looking hand towel.   I sat in one of those sweatboxes (Me, on approach: \"It\'s so hot!\" Tonya, ushering me inside: \"Yes it is, it\'s good for you.\")--metal with a hole cut out for your head, with steam billowing inside, sweat gushing (yes, gushing!) from every pore neck-down--that got so hot inside I questioned both my sanity and my ability to stay conscious. Sweat sluiced down my back, my front, my sides, the back of my knees, and my hands got so hot I had to press them between my legs. I had to slouch down so the metal box could settle comfortably around my shoulders without my skin touching hot metal. Tonya wrapped my sheet around the neck opening, and I lay my flaming face against the amazingly cool cotton and waited for release.  I lay on a plastic-covered slab of foam with hot towels wrapped around my knees and shoulders (Me: \"Those towels are so hot!\" Tonya: \"Yes, they are.\") and I could feel the burned skin sloughing from my legs, but I look later and they are intact. Tonya came at my face with a towel, which I tried to rebuff, but before I could it touched down and I realized it was blessedly icy.   I was ushered by a new lady, name unknown, to the needle shower. \"See if that\'s too hot,\" she says. I stick my hand in. \"It\'s too hot!\" I say. We slide down to the other end of the concourse, and I am presented to the other shower, which is cool. I see Nancy, wilty in a chair, looking slightly flushed in her white sheet.   \"I\'ve bypassed all the other things! I got too hot!\" She shouts to be heard over the din of running water.   I look at her in surprise. \"No sitz bath?! No sweatbox?!\" I am flabbergasted.   \"What\'s a sitz bath?\" she asks.   \"That sink thing over there!\" I point it out.   \"No,\" she says, \"no sitz bath. You actually got in a sweatbox? It\'s too hot!\"  I go back to have my massage. I lay in a small cool room with an overhead fan and am rubbed with hot lotion by yet another lady whose name I don\'t remember. This is the best part of the day. I decide to forego the bath next time and just get a massage.  After, N and I are faint from hunger and can only manage to drag ourselves across the street where we end up at some odd low-rent blues lounge. The outside looks respectable, like a place with character; we get inside and after a few moments decide it\'s not the kind of place you want to use the bathroom. It\'s almost deserted at 4.00 pm. Non-smoking is a joke, as the sunken velvet-covered (maybe velour) seats soak up everything. It\'s dim, the illumination coming almost solely from the deepset windows, and the walls are covered in horseracing posters. The menu has a racetrack theme. It\'s a squirrelly combo of racing and hard-drinking nightlife, the former infusing the atmosphere with cheap chain-store kitsch and the latter adding trailers, trash, and tattoos. But at least it has air conditioning. It\'s about 90 degrees outside and I worked up a sweat just crossing the street, ruining whatever cleanliness my bathing experience had generated.   Once seated, we look at each other and debate, but decide on staying. We don\'t have the energy to find somewhere else. I order cold roast beef and she tries the club with a side of onion rings.  She grudgingly gives one to me and they turn out to be the best onion rings I have ever had in my entire life. My fries are mediocre. The white bread on my roast sticks in my teeth and since I\'ve touched my straw with lotion-soaked hands I\'m wary of my water with lemon.  We manage to stroll the street after and have ice cream cones and coffee. Nancy gets two whistles from passersby, one from a truck full of redneck guys and the other from a semi-homeless-crazy-looking guy on a bike. On the way home N puts Garrison Keillor on the radio while she drives, and we float along smoothly, and I feel exhausted, and I doze.  LA   '),(21188,3792732,'30,June,2004','2004-06-30 00:00:00','   Hi.  I know a guy who does ventriloquism. You know, talking with a dummy and pretending the two of you are having a real conversation. The guy\'s name is Robby and the dummy\'s name is Raymond. I know Robby from work and Raymond from one of Robby\'s (what to call these?) performances. Robby is a children\'s speaker and uses Raymond as a teaching tool for the kids, which is good. Most kids like the whole ventriloquism thing, I think.  Except me. As a kid I didn\'t like watching the dummy half of a ventriloquism act talk or move. Ventriloquism makes me feel squirmy.   What\'s more, there is a promotional poster of Robby and Raymond in my workspace! It was there when I started working and for some odd reason I don\'t feel I have the authority to take it down. I have an open workspace, I guess is the thing. It\'s just a head shot of both, and from the way they are positioned Raymond appears to be sitting on Robby\'s lap. Raymond (the part of him I can see) is wearing a white buttondown, tie, and sweatervest. I can feel Raymond watching me a I type this (yes, I know, from work), round face and bulbous nose above his crisp collar. It\'s one of those snaps where the eyes follow you. And strangely there is  yet another identical poster across the room!     I know this sweatervest is green because Raymond was wearing exactly the same outfit when I saw him in person at Robby\'s talk. Raymond looks like Goober from the Andy Griffith Show. Goober dressed in white buttondown, tie, dark green sweatervest, grey flannel trousers, white socks, and black and white oxford saddle shoes. I wonder if Raymond has more than one outfit, but I don\'t dare ask Robby. It might seem like I\'m more than normally interested.   I had met Robby a couple of times through the course of my work, and before I saw him with Raymond he was a normal guy. Outgoing, friendly, professional. But now that I\'ve seen Raymond, every time Robby speaks I hear Raymond\'s voice in the background saying the exact same thing on a 0.5 second delay. It\'s like wearing headphones while each speaker intones the same thing in a different voice into either ear. Even the casual \"Hi, how are you?\" has Raymond\'s voice in eerie overlay.  The grandfather of a good friend died last week and his funeral was on Saturday. I didn\'t know this until I arrived, but Robby knew the guy. I spotted him from across the room as I walked into the church sanctuary, but there wasn\'t time to say hello. Then the service started.   And Robby gave the eulogy.   And so did Raymond.    LA   '),(21189,3792732,'01,July,2004','2004-07-01 00:00:00','   Hey.  8|   This is me. My eyeballs are glazed over and I\'m so bored I\'m expressionless.  I\'m glad to have somewhere to go every day, but I wish I were busier.   I\'m sooooooooo sleepy.  3|  This is me sleeping at my desk. I could go to sleep here at the computer right now. Even though it\'s a bit chilly in here.  See? I can close my eyes and type at the same time, and in a few minutes I\'ll just stop typing, slowly drift off, the clatter of the keys diminishing, each stroke becoming laborious, heavy, sluggish, sleeeeepy.  Haven\'t had any caffeine today. Maybe this is the thing. But I need to wean myself off caffeine anyway.   It\'s way too easy to close my eyes at my desk. When did this become acceptable????   So it alternates (in five-second incriments):   8| 3| 8| 3|   Then the phone rings, and   8|   LA   '),(21190,3792732,'02,July,2004','2004-07-02 00:00:00','   Hello all!  So Nancy and I are sitting in the dim theater waiting for Spider-Man to start. It\'s opening day, crowded. We manage to find seats halfway up the gallery, way over on the right, three seats from the wall. We juggle our jumbo Sprite (Nancy has this late-night caffeine adversion, late night being after 5 pm) and large popcorn (N: \"It\'s 50 cents cheaper if you buy a bigger combo!\") and plain M&Ms. I find going to the movies with Nancy to be somewhat economical, as she\'s got no qualms sharing a drink or a snack. It works well for both of us.   The couple in the row in front of us appear to be on a first date. I say this because that\'s how they looked to me. You don\'t often overhear people announcing things like that. They are youngish, but not teenagers. Early twenties? She\'s a mildly attractive blond in bright lipstick and red top, and he\'s a mildy attractive guy with dark office-worker hair and one of those three-button knit pullovers with a solid-color collar and striped body. Some muted earth tones.   His hair is thick and ends abruptly in the back as if the hairdresser failed to taper but instead mowed a line straight across his neck with clippers. Disconcerting, but okay then. They smile a lot into each other\'s faces with that I\'m-trying-to-impress-you-but-am-pretending-not-to-try-because-not-to-pretend-would-be-too-obvious-but-this-is-not-(although-this-maybe-kind-of-is) kind of way. Brief snatches of their conversation drift upwards and backwards towards me.   He\'s saying, \"You didn\'t know that about me? But yes! And of course I love my Chillow...\"  I poke Nancy in the elbow. \"Chillow!\" I hiss, trying to be covert. I gesture discreetly with the popcorn tub. \"That guy has a Chillow!\"  There\'s a look of incredulity on Nancy\'s face. She thinks  urlLink Chillows  are crazy, and is no doubt shocked to find that someone else owns one. \"Oh my gosh! There\'s two of you!\"  I feel an instant rush of goodwill and commonality towards him, and am suddenly able to forgive his haircut. I think the Chillow is a great invention. I purchased one for my mom for Mother\'s Day, but have since borrowed it and use it shamelessly.   \"And here\'s something else you don\'t know!\" he announces. Oh my. He\'s one of those guys who think it appropriate behavior to begin spouting lists of facts or ask you stupid job-interview-like open-ended questions while on a date instead of having a normal, flowing conversation.   \"And--the backseat--my car--.\" Drat! He\'s mumbling. \"I have this doll that--.\" I hear that loud and clear.   I poke Nancy again. She is now eating the top butterd layer of popcorn out of our tub. \"That guy has a doll!\" I sqeak at her, this time cocking my head in his direction.   Nancy pauses, hand poised above the popcorn, fingers glistening with butter, eyes hard and wary. \"Yeah, that\'s not working for me.\"  \"And the circus!\" we hear him exclaim with joy. Nancy and I look at each other, eyes wide with horror that anyone 1) actually likes the circus and 2) announces that type of thing on a date.   \"Look, she\'s in shock,\" I whisper, meaning his date. The blond\'s eyes have begun to glaze a little and her smile is on automatic.  All feelings of warmth and solidarity I feel over the discovery of another Chillow-user evaporate. Nancy fluffs the popcorn and angles it in my direction. I eat a few pieces and watch the crowd, wondering how many more bad first dates are going on. The lights begin to fade and the chatter of the crowd increases in anticipation. Applause breaks out and ripples across the gallery as the screen in front of us flares with light, and we watch.  LA   '),(21191,3792732,'14,July,2004','2004-07-14 00:00:00','   I\'m in my car, pulling up to the drive-thru at McDonald\'s (yeah, yeah), sliding into the narrow isle beside the building. I can see several cars in front of me disappear round the side of the building--the line begins to wrap around the corner. I\'m sliding into place, slowing, when a white SUV darts through the side of the parking lot--ignoring the narrow isle obviously marked Drive-Thru--and rolls to a stop \"in line\" moments before I do. Our front bumpers meet at a right angle, his gas-guzzler lining up perfectly behind the car in front of us, my car perfectly balanced on the cusp of the curve of the designated drive-thru isle curling around the building.   He is not making any eye contact. He is lighting up a cigarette. He is rolling down his window in anticipation of ordering. The nerve! He lets off the brake and rolls forward, bringing his front tires in front of my bumper. I now have a great view as he flicks ash and adjusts his glasses and looks hard-edged and defiant.  The line moves forward and he pulls up, but not far enough to let me get around the corner. His SUV now blends seamlessly in line, filling in the corner slot. We wait. I contemplate revenge, but can think of nothing satisfying yet inconsequential to me. Instead I try to transfigure those huffy feelings of indignation into smug self-satisfaction. I have done a kind deed. I have let someone take my place in line. I have....  In my self-deluding reverie I have not noticed the gap between our cars widen. Suddenly, out of my peripheral vision, I see a flash of green careening towards me. I snap to in time to see a green SUV squeal to a halt in front of me, squirming its front quarter into the small gap between me and the white SUV. There is a woman in the car. A woman on a cell phone.  Something inside me snaps, setting off an internal dialogue between me and my evil alter ego.   Evil: HONK!! HONK LOUDLY!! Me: (Gasp of uncertainty.) No! Evil: HONK!! Me: No! (I am angst-ridden.)  Evil: HONK!! Me: No! (But I think this feebly.)  Evil: HONK THE HORN!! Me: Maybe I can. Maybe I should. She certainly deserves it.  My hand hovers over the horn.   Evil: Yes! Just honk the horn! Me: Can I really? Evil: !!!!  Evil is speechless with frenzied desire to honk the horn.   Okay! I\'m going to honk the horn! I feel empowered and justified and completely right whilst at the same time seething with revenge. My hand moves to honk, and then something stops me.   Time. It\'s been too long. I\'ve missed that special window of opportunity. I instantly know that if I honk now it will mean nothing. It will not be the spontaneous and deserved response to an injustice against me; it will be a pathetic, aborted attempt at self-assertion that will look foolish. Or look like a joke. It will have the classic tell-joke-then-take-comic-beat-then-tell-punchline rhythm. Completely inappropriate for revenge purposes.  Suddenly it all seems like too much. I break past every sanity barrier and scream out into my windshield in a high, shrieking voice, \"YOU WILL GET YOUR COMEUPPANCE!!!\"  And instantly feel better.   My window isn\'t rolled down, and hers isn\'t either, so I doubt she hears me, but that doesn\'t seem important.   The line moves forward, and she jerks around the corner in starts and stops, like she is pressing on both the gas pedal and the brake at once. She gives her order. The line moves. I give my order.   We are now around the bend and I see the white SUV move up to take his place alongside the hand-out window. There is now a large gap between the white SUV and the green SUV. I see the brake lights of the green SUV disappear, and I hear a rumble in front of me. Suddenly, the green SUV lurches forward with startling intensity. And rear-ends the white SUV.   Immediately, the chain-smoker in glasses jumps out of the white SUV and postures with angry shock next to his rear bumper. He gestures with his cigarette to unseeable damages, smoke billowing from angry nostrils. I see the lady in the green SUV lean out of her window. She says something, but I cannot hear what. It obviously angers white SUV because he suddenly becomes dramatic, bouncing back and forth between her window and his bumper, gesticulating with his cigarette, head jerking back and forth on his neck, and saying in a raised voice, \"Watch where you\'re going you bitch, you....\"   I am immediately in awe at the inner workings of the universe. I feel one twinge of guilt, and then it disappears.   He spews more angry obscenities, jumps back into his vehicle, grabs his food from the hands of the McDonald\'s employee hanging out the drive-thru window, gets change, then drives off. The green SUV pulls forward to receive her food. It\'s my turn. When I get to the window the McDonald\'s folks are still shaking their heads and looking mildly shocked. I, on the other hand, am enjoying cosmic justice.     LA   '),(21192,3792732,'08,August,2004','2004-08-08 00:00:00','   Iâ€™m sitting here fully expecting prose to burst from my fingertips at any moment. Thatâ€™s the life of a writer. Long moments of quiet solitude filled with the dance of language and the feeling that everything you write will be nuanced and meaningful.  Instead I find myself waiting for inspiration, pondering the backs of my hands as they linger mournfully over the keyboard, and imagining myself writing, writing like a champ. Instead I find myself writing long, drawn out sentences of ungrammatical proportions, like Garrison Keillor, my new writer of envy. Only his sentences are grammatical and nuanced and meaningful and since they are so long on top of everything take on monumental importance and only show the beauty of the craft, if done right.  I think email has ruined me. Nothing comes now as easily as sitting in front of the Compose screen in my Hotmail account and pouring out my thoughts and feelings to loved ones. They appreciate all and criticize nothing. Unlike the rest of the world, they await my prose with an eager, open mind. Iâ€™ve come to realize a receptive audience is everything.  I havenâ€™t got one of thoseâ€”a receptive audience. Not even a hostile audience. Iâ€™m what they call unpublished. Well, okay, technically not. Technically I have three published pieces. I have sold one nonfiction essay to a website a friend works at. However, even though they have purchased my essay, it has not yet been posted on the website. I find it embarrassing now to have told friends and loved ones that I was being published on the web only to have my so-called debut delayed indefinitely. The second essay is also nonfiction, also published through a friend in a newsletter. And thereâ€™s my masterâ€™s thesis (fulfilling a degree in writing, no less) and my name is on the copyright page, but only two copies have gone into distribution, and both of those to the school I graduated from.  Iâ€™ve not gotten much further. But something will happen. I\'m working (thinking) on it.  Without this internal, self-investment-type praise, no writer would ever publish.   '),(21193,3792732,'09,August,2004','2004-08-09 00:00:00','   I met Garrison Keillor very briefly at a recent literary event. After a one-hour presentation, an informal book signing was staged in the lobby of a downtown auditorium. I went with Nancy (she\'s also a Garrison Keillor fan) and I was a little star struck, Iâ€™m afraid. We chatted for approximately three minutesâ€”an eternity in the book-signing lineâ€”as he inscribed my hardback copy of  Love Me , his latest book. Hereâ€™s what the inscription says: Leigh Ann, love and oranges, garrison Keillor, and then the date. Lowercase \"g\" and everything! I wonder now if he writes frequent emails (the nature of email being somewhat lax on capitalization).  Afterwards on the steps outside the building I spend several minutes trying to decipher that word that looked suspiciously like â€œoranges.â€ For some reason it seemed bad form to open the book and read what had written right in front of him. It must be some quality of the celebrity that made me think that Garrison Keillor might inscribe something of great importance on my personal copy of his novel, or impart some great wisdom during our 3 minute exchange. I left feeling horribly clichÃ©.  This is a paraphrased transcript recounted to the best of my recollection.  Me: Hello! (We shake hands.) It\'s nice to meet you.  GK: Nice to meet you. (He takes my book and opens the cover).  Me: My name is Leigh Ann.  GK: Let me guess. (He spells out loud): L-e, capital A-n-n.  Me: No, I\'m L-e-i-g-h, capital A-n-n.  GK: A high-rent Leigh Ann.  Me: Yes, I suppose so.  GK: Are you from around here?  Me: Yes, I grew up in Arkansas.  GK: So what do you do?  Me: Um, actually, I\'m looking for work. (At this point I desperately don\'t want Garrison Keillor to know I can\'t get a job as a writer.)  GK: As what?  Me: Well, as a writer, coincidentally.  GK: Oh really! You\'ll have to move away from here to do that!  Me: Yes--I\'ve heard NY. What would you recommend?  GK: I wouldn\'t recommend NY unless you have some contacts there already. I wouldn\'t just pick up and go. You need some contacts.  Me: Hmm. Okay.  GK: What kind of writing job are you looking for?  Me: I\'m open to all possibilities. I just graduated with my master\'s degree.  GK: In what?  Me: In writing. (At this point GK looks up from my book--he\'s still writing on it--and we make sudden eye contact. He lookes mildly surprised, mildly assessing, and later Nancy says he seems impressed.)  GK: Is it an MFA?  Me: No, it\'s an MA in professional and technical. I\'m trying to stay away from the technical and concentrate on professional.  GK: No, you don\'t want to do technical!  Me: I was teaching freshman composition--  GK: Well, maybe your parents could support you for a month or two? (He hands my book back.)  Me: Hmm, well, maybe. (I nod). That is a possibility.  GK: Or you could just have a nervous breakdown.  Me: Yeah, there you go. I could just have one of those. (I try to keep this as lighthearted as possible by my tone.)  GK: You have to write every day. Every day. And observe things. You have to watch, and to listen. And write every day.  Me: (Nodding.) Um hmmm.  GK: If only I could just hit you on the right spot, right on the side of your head (he gestures), knock something loose....  Me: (Smiling.) If only....  (A pause where we just look at each other.)  Me: Well, it was nice meeting you.  GK: Bye.  ---------End Transcript---------  I blacked out around the time he was talking about my parents supporting me, so I couldn\'t figure out if he was talking about supporting me in NY or elsewhere. I think it was the combination of GK telling me not to go to NY and the alternate track in my head that was running with all the things that I *should* have said that distracted me for that millisecond. I should have asked about what he meant by contacts.  It\'s all so horribly clichÃ©! The unknown/unpublished asking advice from the great author. At a book signing of all places! As Nancy and I walked up the steps to the auditorium before the program started (he did a presentation before the signing) I overheard one woman say to another, \"I\'ve been taking a creative writing course...\" and it made me sick. Everyone in that room probably took a creative writing course in their spare time! Everyone thinks they can be a writer! In the bigger picture I\'m no different from 100,000,000 other unknown writers in the universe, all of us asking Garrison Keillor how to become a writer.   '),(21194,3481803,'30,May,2004','2004-05-30 00:00:00','   Today didnt really do anything at all. I watched the Street Fighter anime last night and it wasn\'t too bad. Today I was gonna watch the ninja movie, but my tape player is missing. I think my mom took it with her last night (for some reason) and decided to leave it at her partying location.  I just played Rag all day, took a shower and washed my stank ass, and ate a fudgesicle. Life is peachy.  '),(21195,3481803,'29,May,2004','2004-05-29 00:00:00','   Today my dad woke me up at 7 (cant I get some DAMN SLEEP AROUND HERE!!?) and asked- no,  told  me to come help him install my aunt\'s new shower. It just started as a small leak in her tub, but now shes going nuts and demanding a whole new shower! Not to mention a new toilet and sink we\'ve yet to put in...  We head to Lowe\'s to get some caulk so we can... caulk up the shower I guess. Somehow, we get up on a conversation about what construction vehicles my dad\'s operated over the course of his life. He tells me that he\'s \"rode just about everything you see on the side of the road\". I\'m looking around: Doghouses? Trees? Garbage Cans? What the hell was he smoking?  Anyways, we\'re putting the doors on the shower and we realize that they\'re too big to fit the shower. So, my dad\'s stuck in the shower because the doors cant slide enough to make a big enough opening to get in/out of. I fall out laughing. Smokes pouring out of my dad\'s ears, like the demon he is, and i\'m laughing my ass off.  This evening, the folks are off to some lady\'s birthday party. So, i\'ll be here alone. Again. On the computer all night. Eating cold pizza.  '),(21196,3481803,'28,May,2004','2004-05-28 00:00:00','   Oh yeah! Instead of  renting  a movie, I bought some anime on VHS for only 10 bucks!! I got  Wrath of the Ninja  and  Street Fighter II: The Animated Movie . I certainly hope the Street Fighter anime is signifigantly better than that live action bullshit movie they made sometime in the late/mid 90\'s. If you dont know what i\'m talking about, BE GLAD! That movie damn near ruined my view of the Street Fighter franchise.  The Ninja movie, however, has to be good. You cant go wrong with ninjas.  '),(21197,3481803,'28,May,2004','2004-05-28 00:00:00','   Well, i\'m bout to go get some pizza (mmmmm pizza) and maybe rent a movie if I have the cash. But first i\'d like to record an asinine revelation I just had.  You know how your life is supposed to \'flash before your eyes\' right before you die? I got to wondering, is this life? Or just the flash?  '),(21198,3481803,'28,May,2004','2004-05-28 00:00:00','   Well, I guess this is my blog. Im going to be writing down random stuff whenever I feel like it. If you manage to get here, feel free to read it or not. I really dont give a fuck.  Today is my first day off on summer break (wee!). It isnt really that great because i\'ve miserably failed the grade, due to me missing somewhere around 50 days during the year. Personally, i\'m just pretty glad to be wasting more time on the computer. Mindlessly staring at the glow of the monitor during ungodly hours of the night. Ahh, my sweet addiction.  Oddly, I didnt sleep in today. I woke up at fucking 6 and couldnt go back to sleep. I\'m pretty tired now but theres better stuff to do than sleep at the moment. The first thing I did was start up the computer (naturally) and check out my favorite forums at  urlLink Animeleague.net .Nothing that new or exciting was going on there, so I hopped over to GameFAQs and voted on the poll. Since there wasn\'t shit to do anywhere and Kazaa was being a bitch like always, I decided to play some Ragnarok. After a few monotonous hours of killing creatures as a cute little swordsman sprite, I cut off the game and called up Rea (my close friend) as I promised her I would. We talked for 30 minutes then she went to go get some lunch. After screwing around back on the internet, I got the inspiration to start up my blog... and here I am.  '),(21199,3481803,'02,June,2004','2004-06-02 00:00:00','   Went and spent a day with my bud Cameron. He gave me a PS2 headset and i\'ve gotten back into SOCOM 2. So, lately i\'ve just been owning bitches and stuff online. You can find me on as ~Zombie~.  '),(21200,3599758,'29,June,2004','2004-06-29 00:00:00','    The other day, I was thinking about my friend who had a problem with their friends.  This story is almost true but I just added it some of them that arent true but you will get the whole picture... Let me make up the names for this friends... Jason, Mike, and Matt. Jason is the young guy and Matt is pretty much older than Jason.  Mike is about between Jason\'s and Matt\'s ages.    Jason and Matt met each other last year and became roommate but they did have a fling for each other for short time.  About few months later, Jason met Mike and Jason really like Mike very much.  They were getting know each other more and of course, they ended up have a \"fling\" together.  Later on, Mike grew out of it with Jason and he just want to become a good friend with Jason.  Of course, Jason didnt like it very much but he had to accept it.  Mike already met Matt because they met each other through Jason.  One night, they (Jason, Mike, and Matt) went out to nightclub with other friends. Of course, NIGHTCLUB!? DRINK DRINK DRINK!!! Mike and Matt ended up dancing all night and hitted off very well... Jason noticed that Mike and Matt started to hung out more than before. Jason felt left out by them and he found this is very annoying.  Jason noticed that Mike shared with his experience with Matt than Jason and he felt that he\'s unimportant but Matt and Mike really like each other a lot and they really want more further but they cant because of Jason.  Jason really like Mike a lot and Matt felt in the middle because he also like Mike...  Somehow, Matt and Mike felt that they are prevented to go out because of Jason.     So, my question to you, If you\'re Jason...  Will you let them go even though if you really like Mike very much?     Let me share with my opinion, If I am Jason, Of course, I will be jealous about that because I happened to like Mike and my roommate ended up with Mike but it will take a while for me to let it go even its hard to hide my feeling not to show them how I felt about them. But think about it, what if Mike and Matt are suitable to each other like \"soulmate\".  One question that I want to ask you all, pretend this, if you really like this person and you both ended up had a good time and all that \"fling\" together then later on, this person dont want go more further with you and decided to be your friend but you really like him a lot and thought he\'s the \"ONE\". But you have to let it go and move on but this person ended up with your friend and they are pretty much along very well... But you dont like it very much and try to find a way to keep them apart... Will you do that? hmmm? Will you fight to get this person? so I\'ll leave it up to you all...   '),(21201,3599758,'23,June,2004','2004-06-23 00:00:00','    I just noticed that lately people dont really forgive to each other so easily and I was like? what the fuck? GROW UP!  So, I just want to express my feeling how I see people really forgive...  So I just looked up the dictionary.com and I pasted it:  forÂ·give    ( P )  Pronunciation Key  (fr-gv, fÃ´r-) v. forÂ·gave, (-gv) forÂ·givÂ·en, (-gvn) forÂ·givÂ·ing, forÂ·gives  v. tr. To excuse for a fault or an offense; pardon.  To renounce anger or resentment against.  To absolve from payment of (a debt, for example).   What is forgive means to you? I often forgive people a lot because if I dont forgive people, I often felt so tense and I really hate this feeling so why not just let it go and move on...  Of course, you can just forgive them but you might will not forget what they did to you... that is all, is that so simple? Maybe I\'m wrong or they must have some reason why they refuse to forgive other people but I think you should forgive immendently...  Have you ever watch Queer as Folk last two sundays ago or something like that?  Deb (Michael\'s mom, the PFLAG mom), she said really harash to her brother, VIC, who has HIV poz.  She refused to apologize about what she said really harash on him.  Few days later, Vic died due to heart attack and Deb found out about that and she refused to forgive herself because she havent forgive him or apologize him so, its take a while for her to forgive herself.  My question to you? Can you tolerate that feeling if something happen to your friends or family members? SO, people, THINK! Also, why waste it? if my best friend and I been through a lot of shit happens, so we ended up not friends and refuse to forgive.  Why all the waste of years that we have been through? Can you just precious the friendship what you guys have been through? Hmmm?   I have several people who refused to forgive me what I\'ve done to them but I was willing to talk with them out and clear it up but they refused to do that so I had to let it go because they hold the grudge or they have to much pride? I honestly cant see what\'s people\'s mind...  Yet, I forgive them and move on.    What\'s the wonderful feeling if you forgive these people fast, its really peace of mind and you dont have to think about how we do have to deal with.  Just forgive them immendently unless if you want to have a time then communicate this person that you want to have a time and will deal with them later when you feel ready...  Its soo easy to do that, really... AGREE???  I like this quote and I\'m strongly support this, \"Communicate is a key to succeed\"   My question to you, people, CAN YOU REALLY FORGIVE PEOPLE EASILY IF THESE PEOPLE REALLY HURT YOU?  PEOPLE, PLEASE THINK CAREFULLY...   '),(21202,3599758,'11,June,2004','2004-06-11 00:00:00','    Well well, HAWAI\'I is soo awesome!   FABULIOUS!!!   Bradley Brown, David Fontes, and I flew to Hawai\'i to see Bradley Dale aka Bray for our vacation but not for David cuz he will have a LONG vacation in Hawai\'i! WHY? He\'s moving there to live with Bray as roommate (I miss David dearly and he\'s my best friend).  Well, Hawai\'i, How can I really explain about Hawai\'i...  hmmm, undescribable!! (if you really want to know about Hawai\'i? then you need to join with us for New Year Eve.  We are working on going to hawai\'i for nye)  When we landed on Oahu, HI and I told Bradley and David that not to expect to see luau girls to give us lei.  They were kinda disappointed about that and I told them,\"GAWD! I guess you watch t.v. too much!\" so we laughed about it.  So finally, we met Ollie Luer, Bray\'s roommate (former boyfriend) and guess what? He gave us the leis! heh...     First, I don\'t want to confuse y\'all.  If I said BRAY means Bradley Dale and Brad mean Bradley Brown.     Ollie took us to Bray\'s work at community college in hawai\'i... He\'s teaching ASL there and I finally saw him since two years ago and was very exciting to see him again. David and Bray never met each other so they\'ve been talking each other for like two years and FINALLY, they met!  (CHEER) We all went out to eat for dinner at cheesecake factory.  Brad, David and I were very TIRED! Hawai\'i\'s time is like 9pm and DC time is like 3am! So, Brad, David and I looked like a SHIT!    During that week, We went out to the beach, looking around the island and all that... It was beautiful there!  Especially BEACH! There\'s a gay beach in waikiki, its called \"QUEEN\'S SURF\" LOL! I KNOW! I KNOW! its soo gay! lol...    There\'s one thing I wanna tell you and I think this is cute one.  When we arrived and finally saw Bray.  Brad and I were puzzled to see him because his appearance and his clothing is not what we expected.  Remember this, Brad, Bray and I are ARIES!  Brad\'s birthday is on March 26th and Bray and I have same birthday, March 25th.  During that week, Brad let Bray borrowed his clothes and I had to fix his hair.  One night, I told Bray that he need to do something about his hair because its too PLAIN! Well, I have to admit that ARIES like to be FLAMY and we DONT like tooo plainly!  So Bray and I went to wal-mart and to get highlight kits and all that... Bray never had a highlight before so I have the cap on him so that way i can pull out to highlight on it...  He was kinda grumpy because its hurts!  Brad told Bray to SHUT UP AND ACCEPT! Bray was like, FUCK! and he has been patience for that!  I think its funny and then I added black tip on his hair so he has brown hair, blonde and little black on it...  HE LOOKED GREAT!  So the end of week, Brad and I looked at Bray and he looked GREAT!!!!  Brad and I looked each other and we said,\"WE DID GOOD JOB!\"     Few days before I went back to mainland.  Bray and I went to big island for one night as SISTER NIGHT.  Bray and I felt that we are soo SISTER (I know it\'s SO GAY! but SHUT UP!) We went into the plane decided to order wine and we were buzzed and we made a fun of some people (OOPS)...  When we landed Big Island from Oahu.  We rented convertible car and we felt so COOL at that moment...  We went to hotel and drop thing off and went out to eat resturant called,\"Bumma Gump\". Forrest Gump resturant! heh...  After that, we went to gay bar and that\'s only gay bar in Big Island... Not a lot of people there so we just went ahead and ordered drinks... WE DRANK A LOT and we had a blast time...  Next day, we are supposed to go to see volocanoes but somehow we have a $$$ plm due to miscommunicate! it was really HELL!!! we were so grumpy and it was so hot outside and all that so I had to have David sent us some money through western union and waited for like five hours!!!  PAIN IN ASS! We couldnt go to see volocanoes so we decided to eat Hard Rock Cafe and check the beaches out and it was very interesting!  Then went back to Oahu.  BUT THOUGH, we had a good time!   There\'s one beach that I love to go... Its a local beach called waimea bay beach...  They have a cliff that we can jump into the ocean and I believe its like about up to 25 to 30 feet to jump from the cliff...  We checked a lot of CUTE guys there!  When I swam far from the land and I could see the below the ocean and its soo CLEAR!  So finally we took a rest and we all fell asleep for like 45 to one hour.  I woke up and its a perfect timing, we were watching sunset! it was soo beautiful!     Well, I have a lot of thing to tell you but this is one is good enuff for me to finish it but one more thing, when the day I have to go back to mainland, Bray and I were sad because we already miss each other and also I miss David too... I might wont see him until NYE.  It was very emotionally for me to leave like that but I wish David and Bray a best of luck in Hawai\'i.     Well, that\'s all, FOLKS!  ALOHA!   '),(21203,3599758,'10,June,2004','2004-06-10 00:00:00','    BOY! Last night, my friends wanted me to go to green lantern/tool shed for a happy hour but I couldnt go out because I was too tired.  I decided to take a nap then go APEX instead of going to GL/TS.  SO I went there and had a good time!  I caught up the news with my friends because I was so busy all week from work.  Of course, I was really buzzed from 151 proof bacardi rum with pineapple juice.  so that\'s all about it last night, I dont really need to tell all the details about apex last night.   '),(21204,3599758,'10,June,2004','2004-06-10 00:00:00','    Hello everyone! I\'m so new at this so please bear with me.  Everyone has been telling me about BLOG thing and I tried to get into the website but I couldnt get in because some stupid internet service wont let me in so FINALLY I joined back as AOL member and tried to check the website and it came thru! (CHEER) so here i am! :) So, please bear with me and I will think of something to say... ALRIGHT? ENJOY...   '),(21205,3599758,'12,July,2004','2004-07-12 00:00:00','    Sorry guys! its been a while ago!  I\'m gonna tell you what did i do on lee\'s mini surprise birthday party and Jesse and Lee surprise birthday party...  On june 18th, Friday, Lee\'s 30th birthday...  The day before Lee\'s birthday party, I asked Scott and Emerson if we could throw a party for Lee and we agreed to do on celebrate his birthday...  So I emailed everyone to see if they could come over for his birthday. Some of them can make it to his birthday...  Next day, (on Lee\'s birthday) I left from work early and picked Emerson from work.  Chuck lived near Emerson\'s work so we picked him up and went home.  Chuck, Scott, Emerson and I are very HUNGRY...  we want to get something to eat for dinner so we decided to eat at wendy\'s...  Lee paged me and asked me if he could join with us because he was on his way to home to see us.  So, lee showed up and I could see him DRUNK! He was HUNGRY so I treated him for dinner...  I asked him where did he went and he said that he went out with jessy, jonelle and buck to swimming pool and had some drinks!  Lee told me that he only had three drinks and he\'s pretty much BUZZED! lol...  After that, Chuck joined with Lee to my place while Scott, Emerson and I went to Giant to get cake and beer...  I got people who paged me that they will come over.  Lee told me that he would go out tonight and I ordered him not to go out and he gave me his look that \"DONT TELL ME WHAT TO DO ON MY BIRTHDAY\". LOL! I covered up fast and I asked him what\'s in his mind for going out tonite and he has no clue about that...  So we went home and Cliff stopped by! YAY! Second guest came over!  We drank our beer in lee\'s bedroom and chat chat...  Of course there is something else we did but I\'m not very comofortable to say in this blog!  Then the more people come by that surprised Lee very well... He didnt expected to see these people...  Scott listed this people on his blog so if you want to see who are those people went to his birthday party...  We stayed up until like 3am or something like that but it was BLAST!  So I asked Lee if he had a good time and he has this wide SMILE and his eyes were widen OPEN and he said YES!! HAD A BLAST TIME!  *25* (touch heart) Glad that he has a good time on his 30th birthday.... Chuck and I were discussing about having another party for lee because some people couldnt make it to his birthday party so would like to make up on his birthday...  I told Chuck that I would like to have a pool party and I asked chuck if he could find one for me and he gave me some optional!  Thanks CHUCK!!!!  He\'s very helpful!  Somehow, Lee told me that Todd (Jesse Conrad\'s partner) and Lee had been discussed about having a surprise birthday party for Jesse...   So, Chuck had to talk with Todd and share our plans...  Todd like the idea of the party so Todd, Chuck and I\'ve been working on that for a week and made some plans for everyone...  Chuck found this public swimming pool in Laurel which is very nice!  They have four swimming pools, One is for toddlers\' pool, the other one is 3 to 4 feet tall pool (the largest pools of all), Two of them is deepest one, one is like 6 feet tall and other one is 12 feet tall for diving board and slide...  We only need to pay like 5.25 dollars from 1pm to 8pm, not bad!?!? On July 2nd, Gabe and Allan were on their road to baltimore from Philly...  They met me at my work to check out my office but unfortunetly, my office will be move soon! Ill explain the next blog!  FOCUS FOCUS ON PARTIES! (*GIGGLE*) After that, Allan decided to joined with me to pick Emerson up while Gabe followed me...  After I picked Emerson up, we went home and kicked back at my place... Of course, we did a bit party in our place in the afternoon...  I ended up really tired and went to bed early like 10pm while some of them want to go out SECRET...  AJ came over too...  CHADIVA moved to d.c. from louisana due his personal problem but welcome back!  (Chadiva is very happy to be here in d.c. check his blog..)   On July 3rd, We woke up like 11ish in the morning and we have to get ready to go swimming pool but I had to go to store and buy the cake...  AJ went to store with me because I dont want Lee to come with me but ended up Lee asked me if he could join with us and I was like SHIT!  But didnt want him to suspect anything so i had to let him to join with me...  Went to Giant to buy bigger CAKE!  Todd asked me if we could add spongebob on it so we did.  Lee suggested that we should add \"HAPPY BIRTHDAY JESSE\" and AJ and I looked each other and we smiled but I told him that I just want to be SIMPLE \"HAPPY BIRTHDAY\" so Lee was puzzled about it but he just not really noticed something fishy but he went along with us!  So, I asked people to meet me at my place at noonish...  I\'m gonna list who are those people went with us to the swimming pool. Lee Masa (ME!) Emerson Scott Jesse Todd Chuck Jeff AJ Allan Gabe Imran Chadiva Troy Jessy Jonelle Sara Victor Rene Renato Some of them will meet us at the pool...  After that, I had to meet Chuck since I dont know where is it so Lee and I had to meet Chuck then Chuck led us there to pool...  So, when we arrived, Jesse forgot to bring his swimming short but he did brought it but left in the car.  He joined with Lee to the swimming pool, he forgot to took it with him! LOL! Jesse begged Todd if he could give him a ride to store to buy new short! POOR JESSE! FINALLY, everyone showed up and I had to get everyone\'s attention. I said that we have some announcement..  I thanked for everyone to showed up and all that and the reason why we were here for two persons is LEE AND JESSE!  They both were surprised about it!  So after that, we swam and play around..  OF COURSE!  We are too expose to everyone at the public pool but it was a lot of FUN!  After that, we decided to go to TGIF and we ordered our dinner and something to drink! OH BOY! we were BUZZED! After the dinner, there is more people show up at my place, Cliff showed up and so do Derrick.  There is more people showed up but I couldnt remember who.  We had to get ready for Velvet Nation.  Before we went to VN, we served the cake and Lee and Jesse were surprised about it but Jesse were puzzled about spongebob thing on the cake and I told Jesse that Todd requested me to add spongebob on it because Todd mentioned me that Jesse didnt really like spongebob! lol...  No one wanted to eat cake because all they want to DRINK AND GET FUCK UP! So we all went to velvet nation, of course, you guys all know what Velvet Nation is all ABOUT!  We went there to FUCKED OURSELVES UP!  We bumped into Buck and his friend, Victor...  Billy showed up later, Terese showed up with her friend, Ruben...  We had a blast time together and after that, we decided to go home and kick back because we were still ALIVE.  Some people wanna join with us to kick back so they came over and we chat chat chat chat! Lee didnt show up until 830am.  Some people left and some of us were staying... We ended up play the playing cards called \"PHASE TEN\"... Its a lot of FUN!  but we stayed up until like 10 to 11am and I was SOO FUCKING TIRED!  so went to bed while emerson woke up! lol...  I slept until like 4pm. So that day is on july 4th, we went to CMRA cookout on 4th of july!  CMRA were surprised that I rvsp\'ed 20 people for 4th of july cookout...  So its a lot of fun!  We just relaxed and watching the fireworks after that went back home and we were lazy! lol! it was a good time!  I asked Jesse and Lee if they had a blast time and they said YES!!!  Had a wonderful time!  So that\'s what we did!  heh...   '),(21206,4144596,'11,August,2004','2004-08-11 00:00:00','    So, the project is off to a slow start.   urlLink Ebay  isn\'t being as forthcoming with cool project cars at a low cost as it has in the past; there\'s been a  urlLink Trabant  for sale on there, but it\'s a two-stroke engine and not the coolest of cars.  I know cool isn\'t the sole purpose of this project - in fact un-cool is kind of good - but come on, it\'s a Trabant.  Dad would disown me.  Local papers, ad trader thingy, Practical Classics magazine - all have come up blank.  There was a 60\'s BSA motorbike, in bits for the past 10 years, in the ad trader, but I can\'t ride a motorbike.  Would have been a decent project though, especially at Â£30.  No idea if all the parts were there or if it would ever go again - if it\'s in parts it\'s possible, likey even, that the engine components would have surface rust and would need machining.  You\'d end up having to get the engine re-bored and skimmed and all sorts.  The costs would spiral and a lot of the work would be out of my league.  Raises the interesting possibility of getting a motorbike though - bound to be cheaper (although the cooler the bike/scooter, the more expensive - even knackered old Vespas go for a bomb now).  The possible solution I\'ve been left with was raised by my housemate last night.  He rebuilds cars for racing and we were chatting about me getting a project car when he mentioned his business landlord has a  urlLink TR3  in need of restoration.  I\'ll see how difficult it\'s likely to be, and how much it\'ll cost, but it sounds tempting.  I mean, just look at the things:     Hubba hubba!  Bit frog-eye.  I\'ll let you know how I get on.   '),(21207,4144596,'06,August,2004','2004-08-06 00:00:00','    So, a new day, a new blog.  I suppose the reason behind this is twofold; firstly, and to excuse myself setting this up at work, it\'s to demonstrate blogging to the people I work with because I think blogging could be a handy tool for us.  I won\'t go on about where I work because I don\'t see that I need to and no-one\'s going to  be that interested.  Suffice it to say I just wanted a safe blog where I could demonstrate some of the capabilities without the shock of opening up a random blog in a meeting only to be presented with pictures of monkeys doing rude things.  The other reason, and the more compelling one from my point of view, is that I have a project in mind and I wanted to log it in some way.  Paper\'s so passe, dharlings...  so here I am blogging it.   The project   The idea is I want a car to do up.  I already have my little baby MG midget (about which more later) and I want something which breaks more so I can spend more happy days in my garage elbow-deep in car entrails.  I want a thermos and, to be honest, the current reliability of Monty just won\'t justify it.  To make it more interesting I\'m going to do what I can online.  eBay\'s the way to go, people.  More tomorrow!    Bought so far: Anglia Ad Trader - cost Â£1.50 Total costs Â£1.50   '),(21208,3496909,'30,June,2004','2004-06-30 00:00:00','   On the 19 june 2004[Sat]...i went to take part in the amazing g-race that Wr church organize... Heez... yesh we r the first team to finish the game althought we r the last team to set off...If was fun... still remember the bonus round in east coast park...Where they took out a packet of frogs[alive] n a packet red worms[alive too]...Wanted us to finish it[but actually is for the camera]...Yuckz...At first I thought it was real..but later they were laughing all the way... They keep the packet of frog n worms...And take out a bag which contain ONIONS & Bitter Gournd... Shit man...I don\'t take vegetable one[my friends should noe]...Kaoz...I hate onions & it taste horrible..But at least the bitter gournd still okay lor... Wr quite poor thing lah vomitting all the way...Haha...after that we bought ice cream...as a reward to our group lah..haha...The last station was at someone house[at loyang]...for dinner we had bbq food... but wr n mi we had bbq food but additional of onion taste...After the dinner we had prize giving...We won the 77th streets latest design bag... Quite cool... Haiz... after that we all go home...Really had a tired day but neber mind lah i really enjoy it lots... '),(21209,3496909,'30,June,2004','2004-06-30 00:00:00','   On Monday 31 may 2004 about 5.30pm go out with my mum...(real mum)... Take the bus 293 and reach the interchange about 5.45pm like that...Guess who i saw... my first love ...He was the first person i love deeply...And i suddenly got the feeling that i still love him...And my mum still say this...hey ger...this guy is tall...he match u(in chinese )...i\'m happy to heard that from my mum... but till now my mum still dunnoe that he was once my bf...but hai...too bad he got gf le...I can see that he love her alot......when i think back to the pass i really wan to kill my friend... she is the one match-make us... And also the person we separate us... shit her... feel like killing her that time... but it a good experience for not trusting her....Well if it\'s not her ... the breaking story will not start... and maybe now we are a loving couple...but well.. it\'s okay lah...It\'s all fate let god decide it...And Now I Found a person i really truly love...hee~ '),(21210,3496909,'30,June,2004','2004-06-30 00:00:00','   We are celebrating Mark\'s & Ee Ling\'s Birthday on the 16 june 2004...And we plan to have bbq...And on the 15 june... we went out to buy the food & some games stuff...that cost us about 100+ to 200 bucks... we went to giant, Ntuc, Prime & the paper plate shop... Can imagin that how many plastic bag we carry...here n there... n my house frigde was packed... cannot squesh in ani more..hahaha.... when we where shopping quite fun also lah... tok here tok there.... shop for things... woah so fun... can\'t imagin that next time we stay in the same house .... we can go shopping together haha... '),(21211,3496909,'30,June,2004','2004-06-30 00:00:00','   Haha...Everytime when i\'m sick... those nice nice dream will cum out...like the first one was my dream house...And that house quite big... Inside of that house were my school family members...There were 7 of us... we all stay together happliy...each of us have our own room... can imagin how big is that house right... a study room with 2 computers each of us have a laptop with a drinking bar with quite alot of alcohol... and band practice room with all our instruments ...hahaha... cool man ... but i think it really cost alot of money for it...but i think if we really going to stay together start with a smaller house...heehee... '),(21212,3496909,'30,June,2004','2004-06-30 00:00:00','   I\'m a person who don\'t usually say out my problem...Just can\'t find a person to sit down and talk...So even if I have any problem. I just sit down there and keep quiet..Or sometime I just smile even i\'m not happy...But if sometime that person ofference me... I confirm give that person a nice black face... And also those nice nice words will come out... Yupz... that\'s me... I last time had a quarrel with my friend..Don\'t wanna say the name out... I think got 1 month never talk to her... At times i felt guilty, sad and also angry... But i always think that it\'s my fault for starting that stupid quarrel... Why izzit my fault. Because if i talk to her. She will not get to know about his percussion guy in band... And also will not go into relationship with him... And as i can see, she love him with all her heart... But that idiot just don\'t know how to treasure her...In my eye he really treat her badly... And i hated him alot... If you were in my shoes and your friend is some how so call bully in that way will u like it...will you bare to see it... Up till now i still hate him... Maybe he had some good points...but i still don\'t like the character he has... Well it\'s the pass...but this guiltiness still make me feel bad... I still think that it\'s my fault... I cause this thing to happen... it\'s all my fault... '),(21213,3496909,'25,July,2004','2004-07-25 00:00:00','   On thursday 23 July 2004 Ite teacher brought us to Suntec City for heritage tours...The bus send us there... when we reach there i tell u the teacher ask us where is it... not we ask him... tell u that teacher is a loser man... Never shop suntec before...ask him go ask the information counter he also dun dare... he really hum... Then later we find for about half an hour...we finally found tat  place... Tell you it\'s damm bored... Exhibition on different culture...Hai... We walk for less then half an hour...all of us go toilet change into our home clothes... and off we go...haha... First we go shop around in suntec... go carrerfour play the electronic piano...play for quite sometime then some gers in our class got to go home...then left agnes, zi jun , mi , ah bing, jason n zhi cheng.... the 6 of us go arcade play... Drive the big big lorry...haha have alot of fun...We stay in that arcade for about 1 hour like tat then leave...Cause the guys wants to go sim lim square change something...We took the mrt down...then agnes go home cause she wans go n find her mother...When we reached sim lim...Kaoz... tell u we walk untill headache so many computers software sia...Woah lau cannot ta-han... We shop inside there for about 1 hour sia... cannot ta-han le even ah bing also say headache...then we decided to go makan...but zhi cheng got to go home le... then left the 4 of us... When bugis shop first...haha then go macdonald eat... haha... Sian everytime eat mac de... Haha... Sit inside got air con... but we sweat like siao... talk n laugh...laugh untilll face red red....laugh untill stomache...haha ta bu lay ta han...haha...then the ah bing n jason always very dirty minded wan everytimes say that kind of thing make the both of also abit dirty liao...tok n tok n tok...then time really flies.... very fast about 7.45 like tat late liao have to go home then we jus seprate from there lor...hahaha Really a fun day going out with my class friends '),(21214,3496909,'14,July,2004','2004-07-14 00:00:00','   In School many things happen... At first i get to know Alvin. Then slowly i get to noe Benjamin & Ping Loon... Everytime when teacher give us break the four of us will go and eat together... Although i\'m the onli ger in that group...Always feel good when i\'m with them...Then as days pass.. We get to noe this ger name Qiu Feng... Then Later we get to noe Jason that group...WOah tell u all.... they were really crappers...Always full of craps &amp; crappy stuff...that make people laugh untill stomache...Then i always dun noe y.. all of the sudden i go like my friend Benjamin(why all the things i like must always start with the Letter B..FOr example Band & Ben)...Well it think it\'s fate ba(sound so bu yao lian)... Then At the same this qiu feng also noe...At first qiu feng and i quite close den everything was doing fine... but later there is always something caught in the middle... when she sit beside ben... she always want to attrack attention from mi want to make mi jealous... And i really hate her... Then after awhile i stop toking to her... stop hanging out with all the gang i usually hang out with... Then I get to noe shi lei(dun say i never mention ur name), zi jun that group &amp; that loudspeaker... I told them except that loud speaker i like ben all tis stuff lah... Feel so happy to noe them... They are also tok crap sometimes...But they r not as bad as qiu feng... wanting to scatch ben away from mi... As days pass... i decided to tok to qiu feng... some how so call forgive her ba... jus dun trust her animore... Even if i still like ben now or i like someone else... i die die also won\'t tell her... Really scare that she will scatch away ben or someone again... '),(21215,3496909,'08,July,2004','2004-07-08 00:00:00','   When i\'m alone...i always think back...Think about the happy & sad things happen in the past...Many happy things happen in my life that i will never forget...For Example Celebrating Mark\'s [Lao Tai GOng]Birthday the way we torture him, Kana suan by my shsb family member or when toking crappy things...haha really got alot of fun...Then about sad things I will think about how my friend make my bf n i break up, SOme women trying to steal the guy i like away, or some idiotic people make life diffcult for either one of my shsb family member...I tell u i will make sure the idiotic pple life even worst then wat they giv to my friends...But after thinking all that... i feel that in life pple have to be happy & sad... But when u think about sad things better dun think of going commit suicide or watever thing that r stupid... After doing that i tell u regret also no use le... u r going to die... SO moral of the story is ask u dun do stupid things.. '),(21216,3496909,'08,July,2004','2004-07-08 00:00:00','   School is so damm boring... no more half life...:(Hai... everyday hear Aaron lecture untill wan to sleep... Lecture neber mind still talk so damm bloody slow...kaoz... '),(21217,3496909,'07,July,2004','2004-07-07 00:00:00','   My school life is so damm Fun.. everyday jus can\'t wait to go school... Want to know y... cause everyday go school play half-life...SO fun... This game onli can play onli the computer behind where we sit... Cause the front computer all the games kana delete by teacher... Everyday we jus sit there & play game even thought the teacher is having lecture infront...We will jus continue & play... Throw cockroach here throw cockroach there...haha...take basuca Shoot here shoot there...Muahahaha...So fun...Althought it\'s a violent game...Love school sia... Dun noe how we going to will survive if this game is delete by those lecturer... '),(21218,3496909,'07,July,2004','2004-07-07 00:00:00','   Stupid Chai Shen Yay (Mr.Harold)... Dun noe how to plan time table one...Plan All morning session neber mind...cause afternoon can go home sleep or go shopping..hehe...Kaoz...Den cum n plan tuesday 10am to 6pm... New teacher also siao... Give us 3 break a day... make us puke when we saw the canteen food today...Somemore...Thursday time table also...Freak... dun noe wat the shit he thinking... purposely plan 2pm to 4 pm...jus for p.e & Common Skills...Then later nobody Want to go for P.e & common Skill... teacher jus cancel it...Den ask us to go home... Woah lau waste our time...Toopid Chai Shen Yay (Mr.Harold) Neber use his head & think one... I think he only use his stomach to think... no wonder it\'s so damm bloody big... '),(21219,3496909,'07,July,2004','2004-07-07 00:00:00','   Woah lau...Did quite bad for my exam...Freak... I got 2.000 For GPA(Grade Point Average)... And earn 17 Credits...In order to go poly... i must get 4.000 & above for GPA... And In 2 years time when i finish ite i must get at least 73 Credits and above to go POLY...I Still have to earn 56 more Credits & above... Hai.. only top 5% students then can enter... If i can\'t get Enough GPA & Credits... have to go Ite Dover or tampines to get higher nitec certificate...Course Information technology...Waste another 2 years in Ite...Haiz... Aiming to go poly... '),(21220,3496909,'12,August,2004','2004-08-12 00:00:00','   Arh..... From Today 12 August 2004...Count Down 2 more days 14 August 2004 to my birthday...oh no i\'m geting old le... I\'m turning   17...  Very Old right...Jia Liat...hahaha...My Ite friends Celebrating For mi On Friday...Saturday the actual day my Shsb_family member celebrating for mi(  i think so no news from them yet...but neber mind lah guess they r busy...)...  Then also dun noe wat my ite friends doing for mi...Ask them they also dun wanna tell mi... Make my heart itch   ( Xin Yang Yang)  ... Woah lau onli on that day  (Friday)   then i noe wat\'s going on...  Hopefully they dun buy DURIAN cake for mi...Cause i hate DURIAN....DURIAN SUx...hopfully they won\'t sabo mi Also...But i still scare they will...Dun feel like going school tomolo le...haha but Den they say if i dun go school they come tampines find mi worx...woah lau so bo bian must go... Jia liat...very very cham... Then next Shsb_fmaily Memeber... if they really celebrate for mi i confirm tio sabo wan... Confirm won\'t let mi go so easy...Even Worst is  i think they really getting durian cake for mi...And even worst i think they will do something to mi like press my head down to the cake, throw whip cream on mi, force mi to eat onion again...   (Die liao Die Liao...Cham le cham le... Chi bye tai chi tua tiao liao... Jia Liat...)   ...   Shit i   think i should really get 2 mc on friday n sat... Friday for Ite_Friends &amp; Sat for Shsb_family....muhahahahahahah....... '),(21221,3496909,'03,August,2004','2004-08-03 00:00:00','   You might think y will i come out with this topic: \"Treasure ur life\"...It\'s because I have lost a friend 2 to 3 weeks ago... As she choose the death solution & committed sucide... Feel bad about it...Cause i jus lose a friend jus like tat...And i really think that it\'s not worth to do tat...n the world out there r waiting for u to explore...If u got ani problem jus find someone to talk dun always keep in ur heart n keep on thinking... if that\'s the case u will neber find a way out for it... Find someone that u trust n talk... i tell u... u will surely feel great after that... Next topic is: Treasure the love one &amp; Friends beside you... yesterday night (2 Aug 2004) went for my V.G friend\'s dad furneral... I tell u at that moment i really feel that i must really treasue the love one beside me... Whether if it\'s ur friends, ur relative or who ever... U must Treasure them when they r still around.. If not when they r gone u will really miss them...If u hate ur friend to the core...tried to forgive n forget... If not the next time when they r not around le... i gurantee u confirm feel guilt de... '),(21222,3753339,'30,June,2004','2004-06-30 00:00:00','   Work:  Margot\'s back.  Please.  Comments:  Margot\'s back!  We did our usual thing yesterday--lunch @ Makeda, going to the bank, running other errands, and basically catching up on things.  I graded like a madwoman last night, and a lot of the papers were very good.  I prepared for class, went out for a very late dinner with Margot, and then we hung around a bit more until she grew tired.  Not much was done, and I didn\'t write a thing (not to mention the fact that I didn\'t exercise).  I also spent a bit more money than I would have liked, but I don\'t think it will be the end of the world.  I did not hear from the insurance company today.  I watched part of a horrible Red Sox game, and mapped out my academic schedule until Thanksgiving--what a nightmare that will be!  If I can do it though...wow.  '),(21223,3753339,'28,June,2004','2004-06-28 00:00:00','   Another day, another weird sleeping schedule...I wish I could stay on track.  Falling asleep wasn\'t working for me, so I cleaned up the room, prepared my grocery list, and browsed some of my favorite Ethiopian and Indian recipes.  I\'ll do a cost estimate next time I go shopping--who knows, maybe I\'ll be able to make myself Atakilt Wat for dinner instead of paying $$$$ at Makeda.  When business hours rolled around (and I still hadn\'t slept), I called the insurance company to see if I could clarify things.  As with every person at every horse insurance company I\'ve ever dealt with, the woman was rude, cold and made me feel like I was wasting her time (and here I am just trying to be honest with them).  I faxed some stuff directly to an underwriter (and I actually operated the damn fax machine properly for once), so  hopefully  I will hear back from her tomorrow.  I slept for a few hours (four or five...which used to be all the sleep I needed in a night...I hate 23).  I feel like I\'m being picked to death by little things...I need to see this person, email that person, check this date, blah, blah, blah.  Erg.  I\'ve been sitting here in my \"office,\" waiting for students who \"desperately\" needed to see me, and never showed up--I could have spent this time napping!  However, I was able to make it through the book reviews in this month\'s journal, which makes me feel like I accomplished something academic, which is nice.  I stalled on Margot\'s gift; I hope to finish it by tomorrow.  At least class this week will not require a lot of thought from me--but I do need to finish all the grading (and by finish, I mean start) in time for class on Wednesday.  I feel constantly rushed because I am dimly aware that July, a month that usually drags, and that I hate, will go by very quickly this year.  At least I get random, unexpected Red Sox tomorrow...of course, with my luck, they\'ll lose.  '),(21224,3753339,'27,June,2004','2004-06-27 00:00:00','   Work:  Pages read: 0 Pages written (fic): 1 Pages written (non-fic): 0  Comments:  I really need to institute a \"page a day\" rule for my dissertation like I have for my fiction work--that would just make sense, wouldn\'t it?  Today didn\'t go at all as I imagined it would, and though tomorrow will suck because of that, I\'m not too ruffled about it right now.  Just as I was about to go to bed, my mother signed onto IM, and we had a nice chat--she sounded much more upbeat than she did in her email.  I keep on forgetting how much we\'re alike.  I was v. v. v. proud of myself, as I passed up a Coke and snack, and tried to go to sleep.  The problem was, I failed.  Since the weather was nice this morning, I took my walk at 5:30am.  It was  glorious ; I did stairs and all sorts of stuff, and didn\'t feel like I was going to fall over and die.  I finished exercising and decided I would sleep a few hours before going to Mass.  Let\'s just say that did not work out for me, as I woke up around 5pm.  There was a memo from Housing telling us to be out by July 30th (funny, I swear my contract was until the end of July; I should demand a refund just to be a bitch about it).  This means that my visit with Jessica and Kyle is going to be extremely weird.  Had a nice long chat with Sean that has been a few days in the making, so that was good.  I\'m ready to kill a few of my students--if you haven\'t done any of the work for the final, and you\'re in a group, why not ask your group members for help instead of whining to me (especially after I said, \"If X, then email me by FRIDAY\") and demonstrating that you\'ve put the project off until the last minute?  Oh well...they\'re not going to be happy with me tomorrow at all, as I don\'t have their papers graded.  Nichole\'s students are driving her insane too, so I guess it\'s just that time of the term.  I\'m trying to make headway on Margot\'s birthday gift, and what I have so far looks beautiful, but she needs to give me a few more bits of info before I can finish it.  It\'s a great idea though, and I\'m pretty sure she\'ll like it.  I will have to run errands and deal with the issue of insurance tomorrow, which will be wretched, I am sure.  I also need to clean the room before class, though hopefully I can manage some of that tonight.  '),(21225,3753339,'26,June,2004','2004-06-26 00:00:00','   My second-favorite grouch/contrarian (the first, of course, being my Dad) turns 24 today.    There must be something about these end days of June, because two of my absolute favorite people ended up being born on consecutive days.  Clay was the first person who was ever nice to me without having a discernible motive for being nice to me (i.e., he didn\'t need my help with schoolwork, we weren\'t a part of the same \"outcast\" group, I wasn\'t the friend of a friend--he had nothing to gain, in other words)--that\'s not to say that I don\'t have  great , enduring friendships with people who had a \"reason\" to be nice to me, but I remain touched by his extension of random kindness.  I enjoy his company, because as he once said, \"I bring out the best in you,\" and that is true in terms of humor, spirit and ambition.  One of these days I will send him, as a birthday gift, pink fuzzy dice for that car of his (this will probably happen when I get a job, \'cause these are the lean times for grad students in terms of cash)...either that or his very own Duke hat.       --Varenia  '),(21226,3753339,'26,June,2004','2004-06-26 00:00:00','   Work:  Pages read: 0 Pages written (fic): 2+ Pages written (non-fic): 0  Comments:  Today was marred by a sad email from my mother--it sucks when she\'s down and there\'s not much I can do.  I think I wrote her a good response though.  As is often the case when I am worried or concerned for someone, my productivity goes down.  I wrote fiction before I read Mom\'s email.  I\'m sitting here and I desperately want a Coke, but I know I should refrain, because I ought to go to bed as early as possible tonight so that I can get up and run, lift and shower before I go to Mass (I\'ll be writing more about Mass and things related tomorrow, I\'ll bet).  I\'m actually going to check out the Mass listings of other churches in town, so we\'ll see how that goes.  The things I  need  to do tomorrow are grade, clean up the room, and start poking around for Margot\'s present.  I have received many emails from my undergrads over the weekend--I need to have some of my fellow graduate students review my assignment guidelines to see if they could be clearer (or to the point, are they unclear?), because I have received a lot of requests for clarification on things I thought were clear (then again, I wrote the things, so one would hope I understand them).  I hope to hear from Aaron tomorrow, as I believe he will be back from his vacation with his grandparents.  I should give Margot a call as well to see when she is getting in from NC.  I also need to look around my room for my key card, as it has gone missing (I fear I lent it to someone, but I hope that is not the case).  '),(21227,3753339,'25,June,2004','2004-06-25 00:00:00','   I\'m such a sucker for this kind of stuff...           Conscious self'),(21228,3753339,'25,June,2004','2004-06-25 00:00:00','   Today is my \"right hand\'s\" birthday.  She is celebrating it with her family, as she should, but she will be back here early in the week so that we may cause trouble and do the things we generally do.    Words fail to express how much I appreciate what she does for me.  She knows all of my secrets, and instead of thinking less of me, she\'s always there when I need to talk.  She listens carefully to the fears and insecurities that everyone else dismisses, and will engage me in serious discussion.  She is gifted with patience, empathy and kindness--I am a much better person because she is my friend.  I remember when I first came to graduate school--I went about my usual \"I\'m in a new place\" strategy, which involved acting rather cool and distant so that I could evaluate everyone before acting.  In retrospect, it was a childish way to behave--making people come to me and draw me out--but she did just that, dragging me along to every social event she could (even when I\'d put on airs about not wanting to go).  Now I\'m more open and more confident, and I wonder what I would have been had she not been so insistent back then.  I\'m sure I would be less of a person.  I just hope I can be as good a friend to her as she is to me.  --Varenia  '),(21229,3753339,'25,June,2004','2004-06-25 00:00:00','   Walked:  3 days  Weights:  3 days  Eating:  Not bad, actually.  I was pretty bad on Thursday night--I pigged out while watching the NBA Draft--but other than that, I\'ve been able to keep it reasonable.  I\'m especially proud of myself for sticking to one Coke a day (down from 3-4/day in recent weeks).  However, I am starting to get bored with the types of food I got on my last grocery trip, so I\'ll have to make sure I get different things when I go this week.  Commentary:   I met my minimum goals for the week in terms of exercise.  I wish that I could have walked one more day this week, but I was unable to stare down the heat on Friday, and I overslept badly on Tuesday.  Summers are difficult.  Since the gym\'s hours don\'t mesh with my schedule, I have to walk outside.  However, I have terrible sensitivity to heat (which is even worse than it used to be because I\'m not in shape), and if it is sunny, more than 75 degrees, and there is any humidity, my energy vanishes.  It\'s awful.  On today\'s walk, the conditions varied wildly--when I started and finished it was grey and there was a bit of a breeze.  The sun came out in the middle, and when it was out, I felt exhausted--yet at the end of my walk, when the sun ducked under a cloud, I felt pretty good.  This means I should, ideally, walk before 9am (of the three walks, the most pleasant was at 6am (!), while the other two started around 10am), OR after 8pm.  Walking after 8pm is a sketchy thing because this campus is  deserted  after everything closes (between 4-5pm), and I\'d be all by myself.  I guess this means I have to haul myself out of bed earlier.   Today\'s Walk:    The weather was not ideal, but I found another \"course\" to walk, bringing my total to three (I think this is good, as variety should keep me motivated...if I don\'t feel like doing the Gibbons walk, I can do something else).  This one takes me back to the Douglass graduation field (that\'s not its name, but I don\'t know what it is called), and I do a few trips up and down the steps, which can\'t hurt.  Both the Douglass and Cook fields have sad brown spots where the graduation platforms were (they finally moved them, a month later)--it sorta sums up the entire University during the summer.  I know most of my fellow grad students love the quiet, but I hate it; it just makes me want to sleep and be lazy.  35/44  '),(21230,3753339,'25,June,2004','2004-06-25 00:00:00','   Work:  Pages read: 0 Pages written (non-fic.): 1 Pages written (fic): 1  Comments:  Today wasn\'t quite as useless as the \"stats\" make it look.  For one thing, I finally wrote a page of my dissertation--the first one I\'ve managed all month.  I spent at least a couple of hours dealing with students and class-related issues, and managed to make it to the library and Post Office before they both closed.  Although I only managed a page today, I\'m about halfway there in terms of writing chapter one of TGoP.  I do need to get my rear in gear in terms of non-teaching academic work though, because July is going to fly by, and I will be very busy in August (I am so glad I decided against going to the conference in mid-August--I\'d be ready to kill myself now).  My life is going to be so harried until January that I can\'t really imagine it.  The major hassle of the day had to do with horse insurance, which ends up being a pair in my rear every single year no matter what I do.  In smaller news, I installed Firefox today after reading about the latest  urlLink IE fark-up  on CNET.  Tabbed browsing is indeed very cool.  I also heard from a friend that had been \"ignoring\" me for some time--it was nice to get a long email from him...helped convince me that he hasn\'t suddenly decided to start hating my guts (good to see that I\'m so secure in myself).  His birthday is Sunday and my best friend in grad school\'s birthday is tomorrow, and she returns here on Sunday or Monday, so I need to find her a nice gift!  The boyfriend is still out of touch, so I probably won\'t hear from him \'til Sunday or Monday either.  I was also able to get in touch with Sean via phone, so that\'s another goal achieved.  '),(21231,3753339,'25,June,2004','2004-06-25 00:00:00','   As with everything I write here, this is all merely my opinion...  Over the past couple of days, there have been many  urlLink FARK.com  comment threads regarding marriage.  Three of my friends were engaged this semester, and another close friend of mine has already gone through a broken engagement, so the topic has been on my mind.  I have to admit, I was surprised at the hostility toward marriage--many people linked to and agreed with  urlLink NoMarriage.com  and similar sites.  I know that as FARK goes so does  not  go the majority, but I found the lack of hope very sad.  Then again, I\'m 23 and still involved in my first relationship, so what do I know?  I\'ll admit, there is still much I need to learn, but I feel that during the course of my relationship my thoughts on marriage have matured.  Here\'s where I stand now--  Complete independence (or dependence) cannot be completely satisfying:   I feel this is the most important thing I have learned.  Perhaps this revelation is bound up with the larger (and more general) issue of trusting people, but I\'ve finally learned to let down my guard a bit, and I feel more complete than I ever have before.  It\'s a bit Rousseauian--I give up my autonomy to become a better person.  Marriage is beautiful because it represents two people promising to forgo complete independence because they believe they make each other better.  On the other hand, marriage is not a necessity for a good life, and there are good dependencies and bad ones (unfortunately I\'m still sort in the \"I know it when I see it\" stage of definition).  One does not have to depend on other people--I was probably at my happiest when I was single and my dependence on God and faith was at its strongest.  This leads me to the second lesson...  Patience is a virtue:   I just can\'t understand the mindset behind a short dating/engagement period (I\'m not saying it never works out or that it is wrong, I\'m just saying I don\'t understand it).  I know that I don\'t have the faculties necessary to judge whether or not a person is right for me in so short a time.  I have a friend who\'s been engaged for a LONG time, and I admire her patience.  Most people are blessed with many years on this earth; there\'s no need to hurry into such a commitment.    Know Thyself:   Words to live by in all things, of course, but quite important in relation to marriage.  Before you can seek out people who make you better, you need to have some idea of what needs \"bettering,\" so to speak.     Don\'t settle--there is someone for you:   This has been the second-hardest thing for me to learn.  There\'s nothing wrong with wanting to marry someone that you are both compatible with and attracted to!  One should not feel guilty for wanting both.  Our culture has reached the point where it mocks those who (apparently) marry based on superficialities such as attractiveness, but I still detect a certain amount of guilt-tripping placed on those who want to end relationships with good people because they don\'t feel the physical attraction (I do believe there is a difference between true physical attraction/compatibility and merely thinking someone is \"hot\").  This may be the most \"immature\" of all my statements, but I wonder how many people are in \"guilt\" marriages and what sort of resentment they feel.  Children are OK:   True pain, but this is the lesson that assures me that I really have matured when it comes to thinking about this stuff.  I\'m Roman Catholic (at least I try to be), and the RCC demands that marriages be open to the possibility of children.  I\'m not a particular fan of tykes, so whenever the thought of marriage entered my mind, it was attended by such thoughts as, \"Gee, maybe I\'ll be lucky and be barren.\"  I\'ve re-evaluated that line of thinking and come to the conclusion that if I find someone that makes me better in the way that I believe a husband should, children probably wouldn\'t be the nightmare I imagine them to be.  These are not meant to be mind-blowing statements--just a summary of where I\'m at in thinking about the topic.  I may look back on this in a few years and call myself a fool, but I hope that\'s not the case.  --Varenia  '),(21232,3753339,'12,July,2004','2004-07-12 00:00:00','   7/11/04  Nichole moved out on Friday and she left me some things to give to Stacey.  Stacey had called me about these things on Friday, but her message had been on my phone along with the unpleasant one, so I pretty much forgot about it.  She called me on Sunday morning about these things, and when I picked up the phone, I misread my clock--I thought it was in the afternoon.  Hilarity (for her) ensued, as I tried to pretend that I hadn\'t just woken up.  I didn\'t do so well in that department.  The day was a comedy of me falling asleep when I should have been awake.  Aaron finally got me up for good around 4:30pm when he called about dinner (I had wanted to see  Prisoner of Azkaban  again, but farked it up by sleeping too late).  He and I had a nice dinner and then we went to Hickman to do battle with the copy machine.  Jay was in the computer lab, and he and I had a jolly old time laughing it up over various and sundry.  Aaron and I looked at HP theories (Snape as vampire?!), and generally screwed around until he decided that he had to go prepare for his class.  I felt so much better when I got home.  Mom and I had a nice conversation and I generally felt a lot better about things when I fell asleep (around 8am Monday morning).  7/12/04  I had decided to finish cookies on Sunday night/Monday morning, and I drank Coke and ate cookies as I poured over websites (during my particularly dark hours, I retreated into the escapism of fanfic, etc.).  However, I felt myself getting tired, and I knew I had to be up at noon to go visit Stip\'s parents, so I tried to sleep.  This was only partially successful, and when I had to wake up, I was miserable and felt sick to my stomach.  The drive was horrendous--I felt like I was going to toss my cookies (har, har, har) the entire time.  I got my orders from Stip\'s mom--so long as I can figure out the computers this shouldn\'t be too bad--and Stip and I had a great dissertation conversation (of course, I didn\'t work on the damn thing when I got home, but I spent most of yesterday being very tired) on the way home.  I didn\'t do much the rest of the day, and I ended up going to bed before 11pm because I was just zoinked.  '),(21233,3753339,'10,July,2004','2004-07-10 00:00:00','   Comments:  The less said about these two days, the better.  I got news so bad it left me in the \"lacking will to get out of bed\" stage.  Of course, I was unable to work on  anything , and the feeling of uselessness plus the feeling of guilt was a nasty combination.  My eating and sleeping became super irregular, so that was another nice feature.  I was almost completely reversed in terms of sleeping (i.e., I was sleeping all day and waking up around 5pm).   9P  '),(21234,3753339,'08,July,2004','2004-07-08 00:00:00','   Well, my exercising this week was terrible, because I had lots to do, it was quite warm for most of the week, and I had a couple of major down days.  However, my eating was also not as disastrous as it was last week--I went a little nutty on cookies on Saturday-Sunday-something (it was so late and I was so down I really didn\'t know or care what day it was).  I need to start exercising and continuing the decent and moderate eating habits.   35/44  '),(21235,3753339,'08,July,2004','2004-07-08 00:00:00','   Work:  Tuesday and Wednesday I did little work because I was preparing for class.  However, I did keep up my pace with TGoP.  On Thursday, I made a huge breakthrough in TGoP and I read quite a bit of Smith and Locke.  However, I did not write on my diss.  Comments:  Tuesday was an absolutely useless day.  I did nothing but dicker with my fiction writing and stare at my dissertation files, which I had up in a vain attempt to make myself write something.  Instead of preparing my closing remarks, I just grabbed last year\'s notes, which as you will see, was a mistake.  On Wednesday I was more industrious out of necessity.  It was the last day of class, so I managed to leave in enough time to check my mail and try to attend to business at Hickman.  However, when I arrived shortly after 4pm, the place was deserted.  I\'m sure it had something to do with the construction going on.  I sat around in my office, and no one came.  I was somewhat worried that class attendance would be dismal.  Before going down to class, I stopped by the third floor seminar room, and just stood in the doorway.  I remember when my entire world focused on that one small room with one table that had less than twenty chairs around it.  All I had to do was prepare myself to discuss things in that room, with my friends, three or four times a week, and I was successful.  I don\'t even know what the hell\'s happened to me since the start of Spring 2004--I feel so old all of a sudden.  I\'m looking seriously at jobs, and wondering how I will finish my diss., get a book contract, publish articles and teach enough classes to look good on the job market.  On Wednesday, I wished I could just sit in the seminar room and talk about things with my friends.  Once again, I fear that I am moving much too quickly, but I know that I have also managed to commit myself past the point of return.  I taught my last class.  They were dead (though 22/30 showed up), and my closing remarks were less than brilliant because I brought the wrong notes, though I still garnered a few smiles.  I don\'t know if it was just my mood, but I was very wistful in saying good-bye to them, and when the sentiment was seemingly unreturned, I was rather hurt.  I realized, after about five minutes or so, that I\'d forgotten something, so I returned downstairs, and they were all done with their teaching evaluations.  I was hurt that they\'d finished so quickly, so I retreated to the computer lab for a bit.  Nichole made an excellent observation on the situation--the students just don\'t conceive of us as humans with feelings.  Thursday was brightened by an email from Clay.  I told him about the MotS, and he seemed delighted.  He asked some questions that were funny, but I took them seriously, as is my wont.  I came up with some absolutely brilliant ideas in response to him.  He has ever brought out the best in me.  Aaron returned as well, and it was so wonderful to see him again!  We went out to dinner, and had a lovely, enjoyable discussion that was like an answer to my prayers in the seminar room yesterday.  It lasted for almost two hours, and it seemed like 45 minutes.  We focused our discussion around my NEPSA paper and a few of my later diss. chapters and how they relate to one of his potential dissertation ideas.  I am so lucky and blessed to be pursuing a job that I love so much.  I came home and messed around online for some time--some of my students were demanding their grades by email.  Ugh, I hate grades!  I IMed with Mom, and then I sat down to read.  I finished Mandragola, and I read significant parts of the First Treatise and the Lectures on Rhetoric by Smith.  I hope to continue the pace tomorrow, and to add in some writing.  '),(21236,3753339,'05,July,2004','2004-07-05 00:00:00','   No one\'s around, nothing\'s open, and the total lack of tactile contact with my species is starting to get on my nerves--ugh!  Nothing is updated--oh well, at least the heat isn\'t keeping me from anything.  On the other hand,  urlLink K is staying , hurrah, hurrah!  I love that man sooooooo much.    I also have to amend a comment I made yesterday.  While the day was physically boring, I did make some major strides in my dissertation and my fiction writing, and I came up with a felicitous way to describe life in RTP.  The day was by no means a total loss.  --Varenia  '),(21237,3753339,'04,July,2004','2004-07-04 00:00:00','   Work: Finished Ch. 1 of TGoP.  Wrote the first page of Ch. 2 Wrote approx. 4 pages of diss. Comments: Well, I worked well, but I didn\'t do much else.  Today was BORING.  I didn\'t get dressed, I didn\'t go outside, heck, I only saw one other human being (briefly).  I did get work done, though.  Once again, I did a lot of modding at TDD (no banning, sadly), and I read some of the Dukies\' old stuff.  Reading their stuff is great.  I remember everything about those games--the hype, the stupid stories, the idiotic quotes, the bad calls, and it all comes back so vividly when reading their articles.  They bring back the days when I was a Triangle insider--I appreciate the jokes on so many levels, because you really had to be there to \"get it.\"  I\'m so sad that I\'m not there right now.  I was thinking about \'01, and that happy day in March when Cassie and I got back into her car after a lovely dinner at Outback, with the North Carolina sun still shining, and the first thing we heard when we turned on the radio was Woody Durham lamenting the Heels\' loss to Penn St.  I had spent the earlier part of that weekend in Greensboro watching Duke win two NCAA games, and this was just the topper to the whole weekend.  It was one of those little moments that you cannot have unless you live in a place--really live in it.  It cannot be recreated; nothing that I love about RTP can be recreated.  That is why it is so disappointing to visit.  I try to put myself in the same situations and places that yielded those great moments, and they are always disappointing.  Life cannot be forced--and what I experienced in those two years was the beauty of living life, amplified by a place I loved and good people.  It was the beautiful irregularity of normal, day-to-day life in RTP that captured my heart, and I won\'t rediscover that until I live there again.  '),(21238,3753339,'03,July,2004','2004-07-03 00:00:00','   Ugh, what a disaster.  The wheels came off this week--Margot\'s return visit was a convenient excuse for me to do nothing.  I only exercised one day this week, and that was Sunday.  I ate poorly multiple times during the week, particularly the last two days where I sat and ate and drank Coke.  I need to turn to water instead of Coke, and I need to go get myself something to satisfy my sweet tooth cravings (all I have here are salty foods, and that\'s just not cutting it at all--I just eat and eat and eat, feeling unsatisfied all the while).  I\'m turning to food and its preparation out of boredom and a desire to avoid work, I believe, as the bad eating and continual stuffing of my face did not begin until Margot left and I really sat down to make myself work on my dissertation.  In order to prevent such a disastrous repeat, I need to a)drink water, or even juice, instead of Cokes, b)make myself go out--either morning or evening, it is my choice, but I must do it, c)next time I go shopping come up with some sort of snack or \"entertainment\" for myself when I\'m working (gum, mints, etc).  I need to just pick up again this coming week, and not let myself fall into old bad habits. 36/45  '),(21239,3753339,'02,July,2004','2004-07-02 00:00:00','   Work:  Fiction: 4 pages Non-fiction: 2 pages  Comments:  This was a quiet day.  No phone calls, no nothing.  I woke up around 2pm, and proceeded to eat way too much.  I worked on TGoP, and the first chapter is almost done--which is frightening.  My diss work didn\'t go as well; certainly I was not as productive as I was yesterday.  I spent much of my free time moderating at TDD.  I spent a lot of time reading \"how-to\" articles on Sci-Fi and Fantasy publishing--time that would have been better spent working, of course.  I wish I had more to write, but I really don\'t, as I didn\'t do all that much.  I am not going to bed yet, so hopefully I will be able to write more on my disseration.  My reading has suffered badly, and I should make sure I\'m up to date with all my students\' emails tomorrow.  '),(21240,3753339,'02,July,2004','2004-07-02 00:00:00','   Work:  Wrote four pages of my dissertation and two plus pages of fiction!  Go me!  Comments:  Said goodbye to Margot again.  V. sad.  Worked all day (I must call my sister tomorrow, as I was supposed to call her tonight and did not).  Have a lot of family things to do tomorrow...I hope I don\'t sleep so late that they don\'t get done.  I\'m quite proud of myself for banging out the pages on my dissertation.  I also modded heavily at TDD today, as the iiidjits are out in force, of course.  I should be quite tired, as I couldn\'t have received more than 6 hours of sleep, but I stayed up all day without taking a nap.  Unfortunately, I\'ve been flat awful, and I had three cokes today.  So very bad!  I craved sweets all night, and I didn\'t walk or lift--I\'ll have to do at least one tomorrow.  Obviously I need to work on balance--things fell apart in terms of diet/exercise once Margot arrived.  Erg.  I should clean tomorrow.  '),(21241,3753339,'02,July,2004','2004-07-02 00:00:00','   I despise summer with every fiber of my being.  I actually like my work (as an academic), and I also like my work environment to be active and flourishing.  Campus is dead as a doornail, even with summer classes in session, and it slows down the pace of my work, which is a Bad Thing when I have so much of it to do.  June was quite the struggle this year--it is usually July before I start to drag down.  The next six months will weigh on me heavily, and I want September to come around so that I can have people to lean on.  However, I can take solace in the fact that one-third of the Wretched Three Months has passed (and August will be much more active than it usually is, since I will be taking a little vacation).  July will be rough--it will be all work and very little play.  Still, if I can be productive during the month, it will go a long way toward making me a much happier person.  The month will end with a visit from Jessica, which is something to look forward to, and my move will also take place during this month, clearing August for the \"fun stuff\" of getting geared up for the next academic year.  So, at any rate, July, enjoy your reign.  --Varenia  '),(21242,3753339,'02,July,2004','2004-07-02 00:00:00','   Work: I wrote two+ fic pages on the 30th. Comments: I\'m getting old, I tells ya--I can barely remember what happened on the 30th.  That\'s probably because I didn\'t get out of bed until almost 3pm.  As I recall, I taught class, and then went grocery shopping with Margot.  It was a pretty uneventful day, since Gordon wasn\'t in Hickman for me to chat with.  On Thursday, I finally got the horse insurance thing resolved, and ran more errands with Margot.  Ugh, July is going to be a nightmare.  We went to NYC and saw the  Threepenny Opera .  It was quite excellent, and it played at a theater that does classic plays only (which is nice--hopefully they\'ll do  Macbeth  soon (unfortunately their Shakespeare offering for \'04-\'05 is  Romeo and Juliet ).  The train ride was a disaster both ways--more reminders that New Jersey is truly the Land of the Troll People.  Of course, the major news concerns Coach K...I can\'t fathom what I\'d do if he left Duke--I really can\'t.  He\'s coached there all my life.  I want him to retire there.  I know I have no right to be any part of this decision, and he needs to do what is best for him and his family, but...STAY K, STAY!  '),(21243,3483306,'31,May,2004','2004-05-31 00:00:00','   And I pay my respects to my late uncle Richard. Always the first one to punch me in the stomach. :) My grandma and grandpa of my dad\'s side. Miss ya all. On a side note, today might actually have some good in it. Got invited to a cookout by the Attracted One. Should be fun..  '),(21244,3483306,'30,May,2004','2004-05-30 00:00:00','   ...relationships happen. Nobody can. It\'s common sense. And here I have Peaches telling me to make things happen. Sure, I can make Devry happen. I can make Lawrence happen. I can make Network Administration happen. But I can never make a relationship happen. One, it\'s free will. Free will can\'t be controlled. Two, it take mutual emotions for it to happen. As far as I\'m concerned, there isn\'t one person I know now that I\'ll spend my life with. It just doesn\'t happen. I obviously haven\'t met her, and quite frankly I\'m getting very impatient. But what the hell, don\'t tell me to make something that I cannot control happen!  '),(21245,3483306,'30,May,2004','2004-05-30 00:00:00','   So I found out the start dates at Devry. Turns out there are 6 different start dates in a year. The next one is July 12th. Then it starts again in October. I\'d love to start in July, but there are some conflicts. First is my internship I have here. I get out of class in June, which would only give me a month for my internship. And internship requirement is to do 200+ hours. Plus I\'m not real sure if I can get all the paperwork done and turned in and all set up by that date. Plus I gotta have a job by then and a place to stay. What I think I would really love to do is go in October, although it\'s 5 long months away, but save money and get up there a few months early, kind of get aquainted a little with the new surroundings. That would mean that after I\'m done with my internship, I\'d have to get a second job. A day job. Taking all this into consideration, it looks like I have a hell of alot to do. Alot of planning and thinking and organizing. I just want to have this thought out so I won\'t be calling for help or something like that. It has to be carefully thought out and planned. But I do have one concern...how hard is it to get a job up there after summer? Let\'s hope the 3rd time is a charm...  '),(21246,3483306,'29,May,2004','2004-05-29 00:00:00','   Is wanting something now and not later pure excitement or impatience? Because I don\'t want to wait to get into devry. I guess part of it is because I want to get the hell out of this town. Don\'t get me wrong, it\'s a nice town. There just isn\'t anything here. I just found out that the next semester starts on July 12th. Even less time to get this done. Looks like I\'m still on the late and fast track. I think I\'ll eventually end up going to the one after this next semester, since they have 6 different start dates. I believe this is something that you have to do alot of planning and preparation for. I even started a new savings account, just for this one thing. How unusual of me. I must be pretty damn serious about this. Peaches is supposed to help me today, so hopefully she\'ll come soon...  '),(21247,3483306,'28,May,2004','2004-05-28 00:00:00','   They say patience is a virtue. It may be a virtue, but it\'s something I lack. I don\'t like spending to much time on anything. Why do you think I couldn\'t make it through Pitt State 5 years ago. Yet, here I am, thinking of going to Devry for 4 long ass years. It would be a miracle if I pulled it off. My life doesn\'t even have patience. People come and go quickly. But that\'s my life. The fast lane. Now if I could just slow it down, if I could just get people to stay around longer. If I could just...  '),(21248,3483306,'18,June,2004','2004-06-18 00:00:00','   ..I can feel it deep inside of  me. I can feel  the war that I am fighting with myself. And for the first time in my life, I am  truly winning. I am dominating my own mind. It\'s kind of weird how the path I took to get here. All I had to do was see my mistakes. All I had to do was lose someone important to me. And now my thoughts are forever changed, my way of thinking. I will no longer be negative.  I will always be positive. I will not worry that no one likes me. I will not worry that I will fail. I will not worry that I\'ll be alone. I will not worry wheather no one cares or not. I will not worry if I\'m accepted or not. I will not worry about anything. But most importantly, I will not let life get me down anymore. Happiness will flow through me one way or another, wheather I\'m alone or with someone. I  just want to say thanks Ash & Peaches. Thanks for making me see and teaching me. I wouldn\'t have seen this is it wasn\'t for you two. Peaches, I\'m glad I met you. And I\'ll miss you as you start life in Lawrence. Who knows, maybe I\'ll run into you up there. But for  now, let\'s just enjoy our convos.... And I can feel people looking at me as I act differently. With   more energy, with more excitement. Life can only get better!  '),(21249,3483306,'17,June,2004','2004-06-17 00:00:00','   ..Who is Blache23. I\'m the analizer, the observer of the room. The one that sits quietly and watches your every move, listens to every word you say, and  analizes  both sides of it. I\'m the one you see passing by. The one that lives his life with YIELD. I\'m the hometown boy. Going for the home teams. I\'m  the thinker. The person that takes everything in and tries to  think to much, before I do anything. I\'m  the italian food lover. I\'m the person that takes everything you say or do and never forgets, well rarely. I\'m also the one that makes mistakes and pays for them dearly, not the one that makes mistakes and gets away with them. I grew up  outside the world. So I\'m quite a ways behind in certain aspects of others in world. I see things differently, so much that it makes no sense to most. In the end, I\'m the person that is inexperienced in many things and have alot to learn. Hell, I\'m the one that wonders if anyone reads this...  '),(21250,3483306,'17,June,2004','2004-06-17 00:00:00','   ..In cases like mine. I don\'t know how I\'m going to do it since this goes all the way back to my childhood. The way I was brought up, the kind of father I was around. Just like at work, I was watching people flirting and playing around. And I thought to myself, is this the real world? I mean I was brought up to be serious and not to play around. That was the real world. I was brought up with hate, not love. Hate towards my father and my peers. And hate triggers anger. And too much anger and I blow steam. It\'s gotta change. I guess the first step is anger management or counseling. One way or another, it\'s gotta change..  '),(21251,3483306,'15,June,2004','2004-06-15 00:00:00','   Like getting arrested  for resisting arrest and speeding in from of you\'re classmates. Just makes me realize that the person   you\'re sitting next to isn\'t all  he/she claims to be. Everyone does stupid things, be we learn from them. Well, the ones of us  that actually pay attention to what goes on, that is...  Noticed I haven\'t been writing much lately. Yeah, my life has gotten a little busy. I  may be moving out in 3 weeks...away from my parents that is. In with a  really good friend. Good thing I have friends. Couldn\'t survive without them.   And   yeah, I\'m am an ex-smoker! What can I say, it\'s about damn time. It\'s like I always felt, theres a time to quit everything. Things will come...when it\'s time for them to come...  '),(21252,3483306,'08,June,2004','2004-06-08 00:00:00','   ...Towards the end of the semester. I dream of that day. But with one book left to go through in 3 weeks, everyone has decided that we should go through the first one again as well. Imagine that, going through two books at the same time. How much crazier is this going to get? I wonder...  ...Good news! I called and confirmed my tour date with Devry on Friday. I\'ve been so tired lately. Work is wearing me out. And it\'s not paying me enough. I think the problem is all the money I spend on gas. After I get out from class, I\'ll be able to save some money. I hope these 3 weeks go by fast...But then, that\'s the life I live. Look ahead, fasten my seatbelt, don\'t look back, and don\'t slow down!  '),(21253,3483306,'06,June,2004','2004-06-06 00:00:00','   ...That I got my pin number for my FAFSA. I think it totally freaked Khat out. I even freaked myself out a little. I mean I haven\'t been that excited or even showed it since I was in love. Now I gotta go get a copy of my parents tax returns. I\'m 23 and I still gotta have their tax info. How lame is that?!?! It\'s like you\'re not an adult untill you\'re 25. Car insurance doesn\'t go  down untill you\'re 25 either. Grr... But yeah, the good news is that I have a meeting with Devry on Friday, June 11th. Supposed to be an hour and a half or some long ass time like that. Gotta call and confirm it. Plus I\'ll be staying up there for the weekend. Going to a concert Khat invited me to. \'The Darkness\'...sounds hardcore to me...Should be fun!  '),(21254,3483306,'04,June,2004','2004-06-04 00:00:00','   ..Of doing Lynn? I\'m a better man than one that has to go to the past to get things. If I want sex, I\'ll get it where I am. Not some two hours away, back in the past. Just the fact that I thought about it makes me vulnerable! Grrr!!!!! Lynn, if you read this, yes, I still have feelings for you. I still care about you, love you, but I cannot do this! I firmly believe our time is over. Yes, I can and would love to be friends, but that\'s all that I can offer you anymore. I will not get back into the ways that we once knew each other. You come to me only because you\'re lonely. I am a better reason than that! So, if I was to see you, nothing, and I mean nothing will happen. We can be friends without benefits. We still connect...on some levels. That\'s what starts good friendships. Let\'s just see if at least this can last...  Another thought...My priorities are now set. No one will change them and I will live my life according to them. They are in this order: 1 - Graduate from Devry. 2 - Make a good income. 3 - Friends. 4 - Love. Now notice, I put love last. Those that know me know I am obsessed and impatient with love. I\'m sure that along the priorities of the first 3, love will come my way. I\'m just done being negative and worrying about it. It\'s all a matter of time...;)  '),(21255,3483306,'02,June,2004','2004-06-02 00:00:00','   ...Again. Eh, my ex fiance showed back up. Wants me to come see her. I just may. Not gonna happen. I think I\'ll stop by and catch up with her on life. For old times sake...  '),(21256,3483306,'01,June,2004','2004-06-01 00:00:00','   ...isn\'t easy. It\'s very hard. But sometimes you gotta do what you gotta do. One thing I will not be is vulnerable, no matter what anyone means to me. If I gotta be alone in everything I do, then by god, I\'m ready!  In more pleasant talk, I\'ve taken the first step in getting to where I want to go. Opened a new savings account today. Hopefully, I\'ll have enough to spare by October. Go me!  '),(21257,3483306,'26,July,2004','2004-07-26 00:00:00','   ...I go for the wrong ones. They say they care, they say all the right words. Why can\'t they mean it.&nbsp; Why do they have to lie? What did I do to deserve these people? To deserve this hurt? To deserve this anger? To deserve this hate? I just don\'t get it...Hopefully there is a woman out there that is loyal and truthful...but I wonder...is there such a thing?  '),(21258,3483306,'08,July,2004','2004-07-08 00:00:00','   ...End of this blog...going back to the orginal...I actually get feedback there...:)... http://premium.xanga.com/home.aspx?user=blache21 ...the invites open...farewell...  '),(21259,3483306,'07,July,2004','2004-07-07 00:00:00','   ...Again. Is it the same square one that I was at a few months ago? Yes, I believe so. How so? First, everyone I know has a distance from me. They are \'net\' friends. That means I\'ll be spending all my time sitting on my ass in front of this damn computer...just to talk...or not.. I know if someone doesn\'t come along soon, I just may be back to my old self...or not...I\'m still trying to figure out what to do...Oh, what to do?!? Lonliness is a bitch!  '),(21260,3483306,'06,July,2004','2004-07-06 00:00:00','   ...obsessed with getting out of here. As a matter of fact, I just applied for a new job. This job calls for 12 hour shifts. That\'s 72 hours a week. And the whole point in doing this is to pay off my car in a mere 4 months instead of 14. That\'d come in handy when I get bill I have to pay later. Today, I went to Computer Plus and see if I can get my internship there. The boss was on call when I arrived, so I should get a call sometime soon. Speaking of getting calls, don\'t you hate it when someone calls you and you don\'t know the number or who it is. I totally HATE that! Especially since I haven\'t give out my new number to many people. Grrr... In short, I have a busy life..thats about to get even busier...Hope I don\'t fall over dead before it\'s over!  '),(21261,3957509,'17,July,2004','2004-07-17 00:00:00','    Life is a precious gift. Hold on and celebrate life. I was just thinking about how quickly life can end.  How life changes with just one call. \"Mykal? Chris. Drew\'s dead. Fell asleep...I\'m at the hospital...can you come?\" \"I\'m on my way.\"  Just last week I had dinner with Chris and his partner, Drew. They have been together for 8 years. Yet, as of last night, Drew is no longer here - he fell asleep at the wheel, crashed into the highway divider, and was killed instantly. Only a mile from home! Hasn\'t it been said that many of these \"falling asleep\" accidents occur within three miles of one\'s home?  So, today is a day of memories, tears, anger, consolation, and sorrow.  Please keep Chris in your thoughts and prayers.  Peace!   Drew M 1968 - 2004 Requiescat in Pace   '),(21262,3957509,'17,July,2004','2004-07-17 00:00:00','    Just another James Baldwin quote: \"I imagine one of the reasons people cling to their hates so stubbornly is because they sense, once hate is gone, they will be forced to deal with pain.\"   '),(21263,3957509,'17,July,2004','2004-07-17 00:00:00','    Have you ever had one of those days when the world around you doesn\'t seem real? That things are not quite what they seem? And that you suddenly have more clarity than ever before? Well, today has been of those days for me.  When I woke up this morning, I had a strange feeling (no, it wasn\'t Colm doing something naughty, hehe). I couldn\'t quite put my finger on it; then it hit me! I felt safe and secure. What an awesome feeling! I thought about where I was personally in January of this year: things were good, but I was alone. Then, March happened - meeting Colm, starting a new relationship with him, and eventually making one of the more difficult decisions for a couple - moving in together. When I picked Colm up from the airport last Friday  my heart was singing, my pulse racing - this was it, ready, steady, go! The unreality of things, you may ask? Maybe not so much unreal, but surreal - life passing by, being connected, yet not; loving, and being loved. I sometimes think that I will wake from my dream state...but Colm assures me that this not a dream. He is a great man, and I love him with all my being.  Clarity attained? As I have as one of my favourite quotes by James Baldwin (an author that, as of May 2004, I did not know): \"Love him and let him love you. Do you think anything else under heaven really matters?\" This is my life. Thanks for your being a part of it!  Peace!   '),(21264,3957509,'17,July,2004','2004-07-17 00:00:00','    Summer has finally hit us! Heat and a rise in humidity are signs that our summer is here. Living at the beach does have its advantages, though.  Colm has begun his new job and is adjusting well. I am so happy that he is here and that we are together. I love waking up with my man next to me  .  Hope everyone has a great day!  Peace!   '),(21265,3957509,'17,July,2004','2004-07-17 00:00:00','    Day three of a new chapter in my life.  St Patrick\'s Day 2004 - a day that will live in infamy!  Amazing how life unfolds. I met him at a friend\'s party on that day. His name was Colm - straight (more like gay) from Dublin! From the first moment we shook hands, I felt a new type of warmth flow through me. No, it wasn\'t pure lust, it was something much deeper. We enjoyed the party, and then said good night. The next day, while working on the network, I received a phone call from, yes, you guessed it, COLM! He wanted to thank me for last night and that he wanted to have dinner. And the rest, they say, is history!  And now, we have begun our new life together. He has moved in with me and we are enjoying each other immensely. Life is awesomely amazing!  Peace!   '),(21266,3957509,'17,July,2004','2004-07-17 00:00:00','    It\'s good to be home!  LAX arrival @ 2:20a PST on Friday, 23 July 2004: man, is that late (or early, depending on one\'s perspective, I guess)!  Anyway, at home, to sleep in my own bed again.  Peace!   '),(21267,3957509,'17,July,2004','2004-07-17 00:00:00','    Well, here I sit, waiting....still!  Now the plane is not scheduled to depart until 12:30a CST. Talk about an interesting adventure. I haven\'t seen the movie  The Terminal , yet, but I think I may, by the end of this trip, be a little like the main character who lives in an airline terminal!  Anyway, here I sit, waiting...still! No worries though - it\'s all good.  Peace!   '),(21268,3957509,'17,July,2004','2004-07-17 00:00:00','    Well, here I sit,  - waiting! Airports are such adventures. My flight to Dallas-Fort Worth went off without a hitch. However, trying to get back to LA from DFW: a different experience altogether. No worries, though.  Flew into one terminal, told that my connection was in another terminal. Took the TrAAin (AA-DFWâ€™s transportation system) to the next terminal. Flight delay! Okay, no worries, right at that gate was one of those â€œAqua-Massagersâ€, so I thought, â€œHey, why not kill some time?â€ 15 minutes later, relaxed and refreshed, I look at the monitor â€“ time change for flight and also terminal/gate change, again. So, off I go to my next destination. And now, here I sit, waiting!  I am still on a high from my Toronto holiday and also visiting with friends in Cleveland. Now, am I ready to face the challenge of work next week?  Peace!   '),(21269,3957509,'17,July,2004','2004-07-17 00:00:00','    Well, I just got back from holiday in Toronto. What a great city - wonderful things to see, lots to do.  This was my first time in the city and I didn\'t know where to start.  Looked at the map book and decided that a walk up Yonge would be a great place to start. What energy in the city! I walked up to Bloor then over to Church and down through the most \"colorful\" neighborhood. I enjoyed a drink at Woody\'s and imagined the cast from  Queer as Folk  walking in.  On Sunday and Monday did the same walk and saw a few of my \"favourite things\" - boys from a certain condo there. Total shock in I was! Anyway, even cuter (and taller!) in person. Also, I met one of the objects of my desire, Bozz. (If anyone knows how to contact this guy, please let me know!)  All in all, a very pleasant visit.  Peace!   '),(21270,3957509,'17,July,2004','2004-07-17 00:00:00','    First rambling thought online.  Peace!   '),(21271,3957509,'17,July,2004','2004-07-17 00:00:00','    This past week has been a week of realizations. I entitled this entry \"Reality 101\" because I feel like I have gone back to the basics: live, love, work, play. With Drew\'s death, I have this overwhelming feeling that I have gone off course - I don\'t spend enough time with my friends or my family. Colm and I have our special time, but it\'s just not enough. His work schedule and my work schedule seem to conflict way too much. However, we did have the chance to spend some good, quality time together this past weekend. I had another job up in the Santa Barbara area, so that gave us the right reason for going back to that beautiful place. Here\'s a totally tangential thought - \"Karaoke Revolution\"! If you haven\'t tried it, do - you\'ll have a blast making a fool out of yourself and enjoying everyone else\'s foolishness! Peace!   '),(21272,3957509,'17,July,2004','2004-07-17 00:00:00','    Destructive behaviour: too much of this, not enough of that.  I have a tendency to eat all the wrong stuff when I am stressed: ice cream, creme brulee (my all-time favourite dessert), French fries, fast food. With Drew\'s death, I have been feeling very finite and mortal. Today when Colm got home from work, I was sitting on the couch with three pints of Cherry Garcia, scooping and stuffing! Colm walked over, gave me a kiss, took the ice cream away and asked, \"What are ye doing? Trying to kill yourself with that?\" He knows that I have been trying very hard to stay away from the \"bad things\" because of family medical history.  With those simple words, I cracked: tears flowed, heavy sobs came. Finally, release from all the pent-up emotion of the last two weeks. Colm just held me, and let me just melt into him. I tend to be an emotional \"stuffer\", i.e., throwing all my feelings and emotions into the bag I carry over my shoulder. Then, when the bag is full, POW!, mess everywhere. Mortal and finite, emotionally spent, tired of being the strong one for everyone else, I collapsed.  As I write this, I realize that I am very fortunate to have good people in my life who accept and love me as I am. Colm is a blessing, I just can\'t say this enough. I am grateful, not only for him, but for everyone in my life. I say this now, for all the world to see (or at least the few see who see this), I LOVE COLM WITH ALL MY HEART!!!  Peace!   '),(21273,3957509,'17,July,2004','2004-07-17 00:00:00','    There is such finality with a burial. We had the funeral for Drew yesterday, then went to the cemetery for the committal. Every time I see a casket lowered into a grave I feel so empty.  I was asked to share a poem at the funeral, and this is the one I read (someone gave this to me years ago, and I really love the last two lines):  \"When I am Gone\"  When I am gone release me, let me go; I have many things to see and do.  You mustn\'t tie yourself to me with tears; Be happy we had so many wonderful years  I gave you my love you can only guess How much you gave me in happiness. I thank you for the love you each have shown, But now it\'s time I travel alone. So grieve awhile for me, if you must;  It\'s only for awhile we must part, So bless the memories within your heart  I won\'t be far away, for life goes on.  So if you need me, call and I will come.  Though you can\'t see or touch me, I\'ll be near; And if you listen within your heart you\'ll hear  All my love around you soft and clear. And then when you come this way alone  I\'ll greet you with a smile and say, \"Welcome Home.\" --Author Unknown  I know that we\'ll see each other again, my dear Drew. So until that time, QaPla!  Hug your loved ones today!  Peace!   '),(21274,3957509,'17,July,2004','2004-07-17 00:00:00','    \"What you up to? Not much; how \'bout you?\"  This is how my conversation with Chris started this morning. The funeral for Drew is Saturday. Still can\'t believe he\'s gone. My head is still spinning, and Chris is so lost and hurting right now. I wish I could just take Chris\' pain, but I know that I can\'t.  Colm is such a wonderful man! When he and I decided to move our relationship forward, I don\'t think he knew what he was getting himself into. Life seemed simpler two months ago. No tragedies, no family drama. But since we said yes, my Dad went into hospital with a major heart attack, another good friend of mine was diagnosed with colon cancer, Drew dies; and yet, Colm has been by my side through all of this, being my strength, the shoulder to cry on, the arms to hold and comfort me. After the funeral, I think Colm and I need to get away for a few days.  Jumbled thoughts today from a frazzled mind.  Peace!   '),(21275,3957509,'17,July,2004','2004-07-17 00:00:00','    1 August 2004 - a new month has begun again!  It\'s hard to believe that we are now in the eighth month of 2004. Time seemingly begins to accelerate as we get older; this realization is a little frightening.  Colm and I did a little road trip yesterday after spending most of the day with Chris. We drove up to Santa Barbara. I have loved that place since I was seven. Every summer my family would spend a week up there. I just love the city. On the coast, about 85 miles north of Los Angeles. Temperatures are great throughout the year. Anyway, I digress...  We walked along the water, out onto the pier and just enjoyed each other\'s company. We then walked up State St and ate in one of the many bistros. I love this man with all my heart and am so grateful that he is in my life. He is funny, kind, gentle, compassionate, and so beautiful, on the inside, as well as the outside. I think about how devastated Chris is with the death of Drew. Ever feel guilty for having something that someone else whom you love has lost? That\'s describes a part of me right now. I confuse easily when shite happens.  On the ride back, I just enjoyed hearing Colm sleeping peacefully in the passenger seat. My heart swelled with love, pride, and admiration. Moments like these knock things into perspective. Hold your loved ones tightly and enjoy them!  Peace!   '),(21276,3806681,'30,June,2004','2004-06-30 00:00:00','   Thats what we all want to know right? Each day we wake up and visit the mirror, wondering what the day will bring and what we will be like. Yeah. Something like that.  Well lets see, im 24, a good ol taxpayer of the USofA. Born and raised in Oregon. On the surface im a cool guy. Not many enemies but lots of friends. I\'m laid back, easy goin, pay my bills, own my home, have a good job, never been married and have a phat ride. I enjoy sports, anything outdoors, snowboarding, music, goofing off, just anything that makes me forget about work and responsibility. Raised a Christian and I still adhere to it, its my faith. At least whats left of it.   Under the surface is a different story I suppose. I\'m still a cool guy. I still value things, friends and family, honor and respect. But things in my life have made it difficult to stand firm, to make myself solid. I have alot of hate, anger, and pain inside. Not at generalities, but at past experiences. The things ive seen happen to my family, my friends, and myself ... burns me inside. It\'s been 7 years since it all started. Since I began building that wall around my heart and bottling up all the hate. Watching from the sidelines as one thing after another hit me. Am I going to snap? Am I going to breakdown and end up on the front page? Doubt it. No matter whats inside I would still sacrifice my life for my friends or family without hesitation. That would be an honorable way to go instead of hitting the postal button and causing more chaos, right? I think.   Its a strange thing. Jesus taught us to have love for our neighbors. How can I be a Christian when I have such intense hatred for certain people? Have I failed that command? Who knows. Thats what I see in the mirror each morning. Conflict. Conflicts of my beliefs and raised perceptions against my past and what i\'ve become.   Ohwell. Life is, life. It makes us who we are. Our surroundings, our parents, the generation we grow up into... all of it shapes us, molds us. No matter what we take in mentally or spiritually, there will always be a part that makes us unique individuals. The tricky thing is that part isnt unique from birth. It gets developed by society or experiences. Sure we may have some persona quirks from the ma n pa, but even those dont get tested until our impressionable teen years. The years we begin to question with WHY\'s, and HOW\'s ...   Ohwell. Hehe, thats been my motto lately, \"ohwell\". Used to be \"What doesnt kill you, makes you stronger\" ... yeah, in my case its made me more spiteful, hateful, the pain getting worse.   Ohwell ... '),(21277,3806681,'30,June,2004','2004-06-30 00:00:00','   Sup! Just tryin out the new trend of bloggin. '),(21278,3619647,'12,June,2004','2004-06-12 00:00:00','    Well what a day...not really.  It consisted of my cousin and I playing video games and eating.  Oh yeah I took a nap too.  That is all for my first post, many more to come (hopefully).  PS. Yesterday I bought the Modest Mouse album \"Good News For People Who Love Bad News\" and it is quite amazing.  Love, Iggy/Alex (whichever you prefer)   '),(21279,3619647,'19,June,2004','2004-06-19 00:00:00','    I know it\'s been a few days or so since my last entry, so I figured I was due for another.  Yesterday I went to the Cubs game, and it was awesome.  The Cubs lost though, so that wasn\'t cool.  All in all it was a great event.  After the game I went to Peter Evans\' house and slept over.  I just got home and that\'s where we are now.  I think I\'m seeing  Dodgeball  tonight.  That is all for now, more soon to come.   I sure do love you...and pie   '),(21280,3619647,'16,June,2004','2004-06-16 00:00:00','    I had a fun day, although it didn\'t work out the way it was supposed to.  The plan was going as planned around 2 and even around 5, but then Jugal, Sean Zolfo, and James couldn\'t go see Saved! at 7:35.  Then I couldn\'t go when I asked again.  Plus, I called Craig to hang out, and he was not available for spending time with.  So here I was, just watching TV and trying to call someone to do something, and it didn\'t end up working at all.  Since that didn\'t work, I took a nap.  I just woke up five minutes ago because it is \"garbage night\" and I\'m on recycling duty.  Half asleep, I write to you, while every two seconds or so my eyes drift off to something more exciting to look at.  Perhaps they even close from time to time as I doze off into a partial sleep.  I leave you with that thought, and the thought of my grades, which happen to be terrible.  I\'m currently sitting at a 3.0 unweighted and a 3.083 weighted (cummulative grades).  That is all for now.  Until next time, I love you.   '),(21281,3619647,'15,June,2004','2004-06-15 00:00:00','    Well my day was pretty boring.  I played a ton of guitar, I talked on the internet a bunch too.  Quite a story unfolded today!  Sean McClellan text messaged me saying that he was stuck in a tree, so I replied saying, \"Where are you?\"  He responded with a witty, \"In a tree.\"  So that\'s when I had to call him.  I dialed his number, and he was really sorry about texting me.  I said it was alright, and that I had to go.  He again apologized, I again said it was alright.  Then it was over, what an interesting and exciting five minutes that was.  After that, I went back to my guitar and Modest Mouse.  Later on, I watched the Pistons beat the Lakers and win the NBA Championship.  For anyone who cares, it was the first Eastern Conference team to win the championship since the Bulls did it in 1998.  Go Pistons.  Then I talked more and watched part of  Bubba Ho-Tep .  It is one intriguing plot: a mummy invades a retirement home where Elvis and JFK wannabes are living.  Oh yeah, the JFK guy is an old black guy in a wheelchair.  Tis quite the story.  After  Bubba Ho-Tep , I talked online more.  That was my day.  Thank you and good night.  I love  you .   '),(21282,3619647,'15,June,2004','2004-06-15 00:00:00','    Oh how I love David Bowie.  Today I have just been sleeping, eating, talking via aim, and cleaning.  Oh yeah I listened to music all day too.  What a day.  Perhaps I will still hang out with someone, seeing as the day is still young enough.  Well, that\'s all for this entry. Goodbye.  I love you all   '),(21283,3619647,'13,June,2004','2004-06-13 00:00:00','    Today was what we in the business world call bittersweet.  I had a fun time golfing and playing poker, yet Bobby Powers is moving to Virginia.  What do I do?  I\'ll tell you.  I watch the Pistons win game 4 (84-76 Pistons right now).  Okay, I\'ll tell you my day.  James called me at 8:30 this morning after calling my house, and I said, \"Hmm, I\'m tired, wanna hang out at 11 or 12?\"  So then I went back to sleep, setting my alarm for 10:30.  James then called again at 10:15, waking me up yet again.  Finally, I go downstairs and watch TV.  Then at 11:20 my mom said, \"Oh James called this morning at 8:30.\" Me: \"Oh yeah he called my cell phone, wants to hang out around 11:20.\" Mom: \"It is 11:20.  You can\'t \'hang out\' because you have to clean then go to the going away party.\"  Me: \"Oh crap, I\'ll just call him and tell him I can\'t do anything.\"  Mom: \"It\'s kinda late now.\"  Me: \"No, I\'ll call him.\"  Mom: \"He\'s probably on his way already.\"  Me: \"No, I\'ll call him.\"  So then it happened.  I called James and he said, \"That\'s cool, lets just hang out during the week.\"  Me: \"Sounds like a plan.\"  James: \"Bye Alex.\"  Me: \"Bye James.\"  Then I continued to clean then I went to Sean Zolfo\'s house and then to golfing.  Afterwords, we played poker at Andrew\'s house.  I lost $5.  Duever was there and I didn\'t like him.  Brian was there and he was better than Duever.  We played wiffleball and I kept makin fun of Duever and he cried.  Then he blamed me for losing Mike\'s money, which I didn\'t do.  Then Ben Mebane, Connor, Mike, and I talked about movies.  Sean Zolfo was there too.  After that, we talked about music, then everyone went home.  I went to Zolfo\'s, where I currently am.  Woo Hoo.   '),(21284,3619647,'23,June,2004','2004-06-23 00:00:00','    Yeah...its been a long time since I wrote last.  What\'s been new?  For me, I\'ve just been hangin\' out at Swedish Days and eating and sleeping.  That\'s what my life has been reduced to now.  Anyway, this week is going to be all Swedish Days.  That\'s pretty much it for today.   Until next time, I love you.   '),(21285,3619647,'20,June,2004','2004-06-20 00:00:00','    I just returned from seeing  Dodgeball , and I must say, it was very entertaining.  The cameos were great, including Walker himself: Chuck Norris, William Shatner, and David Hasselhoff.  I\'m tired now, seeing as we saw the late show and I was up late last night and the night before.  So I leave you with this quote from  Dodgeball .    \"Fuckin\' Chuck Norris.\"    I still love you.   '),(21286,3619647,'02,July,2004','2004-07-02 00:00:00','    There\'s a reason behind all of this non-posting that has been going on.  I was in Milwaukee, Wisconsin.  The reason?  To visit with an old pal, and to attend Summerfest.  Perhaps I\'ll be visiting again on Sunday, July 4, 2004 to bask in the magnificent sounds of Rooney and Jet.  Well, I have to go mow the lawn right now, but I will hopefully write again soon.  God, I hate mowing lawns.  Why do they even grow......   '),(21287,3619647,'08,July,2004','2004-07-08 00:00:00','    Today was great, I guess.  I went to Coconuts and bought a Matthew Sweet cd entitled \"In Reverse\" for $5 for the fact that it was \"used\", but there wasn\'t a scratch in sight.  Cleverly enough, the case opens backwards and the booklet inlay opens from the left as well.  Then I came home and found out that my mom, of all people, has a Matthew Sweet cd also.  So the story of the day is:  two Sweets for 5 dollars...Sweet action, indeed.   '),(21288,3619647,'07,July,2004','2004-07-07 00:00:00','    I love Beulah.  RIP fellas.  Now that that\'s over with, we can get on with what we all want: more ramblings.  As you wish.  I\'m going to go see Anchorman on Friday, like it or not.  You\'re welcome to come, just call my house at 1-800-JAYMZ.  When does Napoleon Dynamite come out?  I\'d sure like to know.  I like music.  That\'s all she wrote.   Current listenings:  Polaris- \"Hey Sandy\", Beulah, The Beatles, The Postal Service. I love you.  \"Donde esta la biblioteca?\"-- Dodgeball.   '),(21289,3619647,'15,July,2004','2004-07-15 00:00:00','    My ps2 is okay after the fiasco, and the guy at EBGames let me exchange my game.  Marvelous.  Right now I\'m just talking online and listening to music, like I do much of my day, but perhaps later on I\'ll watch The Bourne Identity or The Big Lebowski.  That\'ll be a majestic array of good movies.  Until next time, visit my website at www.sonidostereo.tk.  Have fun, keep it clean, and remember to go to my website.  That is all for now, most likely I\'ll write again soon.  Good bye.   '),(21290,3619647,'13,July,2004','2004-07-13 00:00:00','    Today was good: Andrew came over and I got NCAA Football 2005 by trading in a few old football games, NBA Street, a Gameboy Advance game, and my old Gameboy Advance.  I didn\'t pay a dime.  But then it happened.  My ps2 \"fell\" (I accidentally knocked it over.) and now the game has a scratch on it and doesn\'t work.  Great.  I hope they exchange it.  Well I\'m tired now. Good night.   '),(21291,3619647,'13,July,2004','2004-07-13 00:00:00','    It\'s been a fun couple of days.  I saw Anchorman twice.  That was one hell of a movie.  Go see it.  I demand it.  Also, I watched the Home Run Derby tonight.  It was a grand event with Miguel Tejada winning, although I chose Lance Berkman to win it.  Well that\'s all for my sports ramblings for the night.  My current fixations in the music world are Matthew Sweet and The Postal Service.  Finally, I want a Texas-shaped belt buckle.  That\'s all I ask for.  Please God reward me with a Texas-shaped belt buckle.   There were horses, and I threw a trident.   '),(21292,3619647,'20,July,2004','2004-07-20 00:00:00','    Wow. I can change the color of my text.&nbsp;  Oh the wonders of science.&nbsp;   Today was fun, Taylor and Sean came over.&nbsp;   We \"rocked\", if you can call it that, anyways, we rocked your proverbial socks off.&nbsp;   It was grand.&nbsp;   Look out world, make way for this \"band\".  &nbsp;&nbsp;Yeah, we actually suck, considering we don\'t have a drummer and this was our first time with all three of us together  .&nbsp; We shall bide our time and make our appearance someday.....yes someday.&nbsp;  Anyways,&nbsp;it\'s going to be 92 degrees tomorrow and I\'m going to stay inside my air conditioned house. &nbsp;  Anyone is welcome to join me for movie watching and  lemonade  drinking. &nbsp;  Well,&nbsp;it\'s about time I stop writing. &nbsp;&nbsp;  Until next time.  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;   '),(21293,3619647,'11,August,2004','2004-08-11 00:00:00','    I\'ve had a revelation.  On the Mojo magazine website, I can sell this White Stripes Dvd for $89.75.  How cool is that?  I\'ll tell you.  Pretty cool.  That\'s right.  It\'s been awhile since I last wrote, for that I apologize, but I\'ve been busy.  No, not really, I haven\'t.  But I watched Kill Bill 2 tonight and it\'s incredibly good.  Go rent it or buy it.  Now.  I got Carvel cake today and I got a discount.  Woo-hoo.  Mr. Patel is nice.  I saw Jugal there, too.  Tomorrow/Today, I will invite people over to play poker and hang out.  It shall be grand.  Keep your eyes and ears open.  Big things are on the horizon.  Perhaps a new website design, perhaps not.  I hope so.  Well, I\'m gonna go talk to more people, then go to sleep.  Good night.   '),(21294,3889401,'23,July,2004','2004-07-23 00:00:00','   under construction  '),(21295,3889401,'23,July,2004','2004-07-23 00:00:00','   Billy Crystal\'s quote in When Harry Met Sally, \"A man and a woman can never be friends,\" is good enough to spark off a debate anywhere anytime. And he is the true example.  '),(21296,3889401,'23,July,2004','2004-07-23 00:00:00','   One of the most lovable guy in the wing. His killing spree by smile, sea-deep eyes and innocence of a baby. ;). Hence nicknamed \"BACHHA\" (The Baby).   &nbsp;  &nbsp;  '),(21297,3889401,'23,July,2004','2004-07-23 00:00:00','   Man with an intoxicating voice, currently riding the crest of success in corporate world.There is some good in the world and it\'s worth fighting for, and success is one of them.   &nbsp;  &nbsp;  '),(21298,3889401,'23,July,2004','2004-07-23 00:00:00','   Most debatable topic in India after Sonia Gandhi\'s Origin is if Prabhas is single or committed. A cherubic smile, a cool temper (even in presence of some people whom he woudl love to kill) is what makes him a true leader. Was Hall President and presided over everythign except hall. :P Best leader to work under. Miles to go for this young hunky-funky young guy and million gals to kill.  &nbsp;  '),(21299,3889401,'23,July,2004','2004-07-23 00:00:00','   Realistically speaking Unreal Tournament is in his blood, Age of Empire is his empire since ages, A computer boots up smelling his presence...Exceptionally Handsome..Good by heart.. But he is yet to be bagged, (gals listen...)!!!! Money is his true motivation he says, and he lies..One of the coolest guy, understanding, caring and bit erratic at times.  '),(21300,3889401,'23,July,2004','2004-07-23 00:00:00','   &nbsp;  &nbsp;  &nbsp;  '),(21301,3889401,'23,July,2004','2004-07-23 00:00:00','   This guy eats number, drinks number and sleeps (with !!) number. Insatible hunger for evrythign beautiful and meaningful; undying romanticism in those motivates him to the zenith of what he does. He defines his own rule, and breaks them when he wants. HE has his own world and yet that own world of himself always intersects with the common three dimensional space in which we live. He is his own fountain head and he is his own companion and competitor. This guy is intriguingly simple and yet, will compel you to have a closer look at him. You can hate him or love him, but definitely can\'t ignore him.  '),(21302,3889401,'23,July,2004','2004-07-23 00:00:00','   urlLink   If Vikram has something most special, that he is very lovable and dearly.I shldn\'t comment his knowledge on Physics as I\'m completely naive on that field and morever I don\'t have the standard to comment on his knowledg,it is sure that he\'s a genius with extreme kind of creativity. But one thing I can say he\'s the kind of friend one can aspire for. He\'ll always thr with u n care abt u whether u r in touch with him or not.   urlLink   http://jadhao.blogspot.com  '),(21303,3889401,'23,July,2004','2004-07-23 00:00:00','   urlLink    Life is no game, its all about sports. When it comes to LAN Comp Games, he is the one. And life is surprised at his sportsmanship. If you ever need to learn life, learn from this GURU.Don\'t say I love you to any gal, till you learn to say \"I\". This guy, hence, could never say that to any gal coz he says always \"WE\". One of the coolest guy around. Cricket was not the same before he started playing it. Silver Medalist, topper of his batch. I spend 3 years with him and I can say you need only three years to earn a friend of lifetime. Oh, yeah he is very handsome and very much single ( I guess !!).   urlLink   http://people.cornell.edu/pages/vr46  '),(21304,3889401,'23,July,2004','2004-07-23 00:00:00','   &nbsp; urlLink Panda  means Priest in Oriya, my mother tongue. If you want to know&nbsp; more about a Panda, the picture comes to mind is foul-mouthed, erutating bunch of highly respected people, who has the sole right to worship GOD for us in temples. But, for me, Panda is a depiction of lot of junk and not-so-very-junk stuffs arranged&nbsp;on a platform&nbsp;&nbsp;in a very confusing manner. He will be very raucous in words, but yet at times he would be very parsimonious in speech. Sometimes, he would make you drink till you feel like chopping off your head&nbsp;imagining how the next morning is going to be for you; and sometimes, he would initially coax you and then congratulate you for your teetotalism. Sometimes, he will say its sex and sometimes he would say it spiritualism. He would confuse you and then would take you out of the cloud to show you the clear world, which lies beyond everyone\'s imagination. He would encourage you to conquer the world, and sometimes would show you the pleasure in losing some battles in life.&nbsp; He is someone you would do anything for. He is someone about whom you would like to talk with  urlLink Vineet  and  urlLink Vikram , the guys who have seen him as closely as me.   And if you have an elder bro like him, you are exactly going to feel the way I do. He helped me a lot, always going out of the way. Thanks Panda, for everythign you did for me as a bro and as a friend.    PS: His blog is not activated till now, but will be very soon.  '),(21305,3889401,'31,July,2004','2004-07-31 00:00:00','   Its just not another day. I wonder how often this feeling is going to get me over, pick me up the ground and throw me to ectasy, teasing me softly and remaining as elusive as the mirage, telling me that the show must go on and then making me sleepy, making things appear beautiful and then stealing my every time to glance at them. Its just crazy. I can bear all the pain of burning midnight oils, reading someone else\'s PhD thesis and writing for paper for some journal, whose editor, you feel is an ****ole. Yeah, its happening. In my dreams and in my reality, in my every notion, I think of it. I amliving for that moment and that moment seems to be appearing in the distant horizon of december. I am graduating. I am, indeed, graduating. The underdog is surprising everyone. I will be graduting this december. I applied for graduation this friday and I have started believing that I can do it this december. And if providence is in love with me unlike past, then I may go for a PhD in Purdue. Shreekant and Amar all are all those fine buddies. I can\'t really wait to be with them.I don\'t know though, how to celebrate it. If it had been a year ago, I would have grabbed a budwiser, ok may be, more than just one and couple of calls to Vikram, Vineet and all those cell phone no.s lying in my phone book. But, my annual celebration for Teetolism is on, and I really don\'t want to again start drinking. I am a bad drinker I must admit. Or, rather I was, atleast. But, I made few calls to my friends I guess.  '),(21306,3889401,'29,July,2004','2004-07-29 00:00:00','   People, from back West Asia, especially, from India have this problem may be. But, it is really funny to watch some of the activities of them. This guy is there to get a ticket from Champaign to West Lafayette, Purdue and gave his credit card to the Afro-American Girl at the counter.  Girl: What is you father\'s name? Guy: Excuse me. Girl: I asked your father\'s name? Guy: (Paused Astonished. With an accented English this time) Why ya need tha? Girl: To issue the ticket. ( Girl is equally amused and astonished.) Guy: Why the hell you need my father\'s name? (Guy thinking...My dad is not single. He is happily married to my mom.) Girl: (Angry, annoyed and amazed) I need your first name , Sir. Not your father\'s name. Guy: Thats on my credit card. Girl : Oho...  Witness to this event were Mitrabhanu Sahu and Vikram Jadhao , both doing their PhD in Physics, UIUC. The guy is me. I guess I made a follow up to the blog written by Ankan.  '),(21307,3889401,'29,July,2004','2004-07-29 00:00:00','   Last night, I was listening to John Edwards, who is the vice precidental candidate for Democrats. He is a nice looking handsome chap and also talks good things about terrorism and economy although he is a lawyer. While he was delivering his speech, there was name of Dick Cheney. I was wondering all the time   \"Who is this Dick?\".   After that I came to know that he is the present vice president. I wish I have heard name earlier.  '),(21308,3889401,'26,July,2004','2004-07-26 00:00:00','   Jeet was with me when we were learning the car this weekend. And&nbsp;he was about to&nbsp; hit a girl. Jeet was the driver and the girl was furious. He is the biggest flirt I have seen and thank god,&nbsp; we escaped from the scene.   Girl: You ass**** !! What do you think you are doing? What do you think...  Jeet: I think you are amazingly beautiful.  Girl: Excuse Me !!!  Jeet: I think I was thinking about you and I hit you.  Girl: (Somewhat, soothing ..) You was about to hit me. You learning?  Jeet: (Smilingly) Yeah !!!  Girl: Drive safe, man !!!  Jeet: Yeah, I will.  Girl: OK. Bye and have a nice day.  Jeet: (Still smiling.) I have some other plan.  And we get into the car.  '),(21309,3889401,'21,July,2004','2004-07-21 00:00:00','   This is one of my friends, with whom I happened to go around the campus in some sunny day, few weeks ago may be. Donning the mantle of rebellion, he is a stud with a cocky confidence and hedonism with elan. But, there is this one thing I hate about him. With my limited intelligence quotient, I don\'t understand the need for him to look at those parts of a girl\'s body, though I have all my libido as a boy intact. But, every time??? &nbsp;May be the dress designer got the measurement absolutely wrong so that some of&nbsp;anatomy is not covered&nbsp;or may be it is some outfit malfunction !! (Anyway, that happens so often here in US that I wonder who teaches these people math here.)&nbsp;&nbsp;That is a too silly an excuse for me though. This thing continues when he is having a conversation too. Whatever may be the reason, I would kill the girl if she looks at my pants while talking to me. I just wonder if the girls don\'t feel the same way.&nbsp; I do believe girls deserve better respect than that.  '),(21310,3889401,'21,July,2004','2004-07-21 00:00:00','   The other day I was talking with&nbsp;someone and he asked me the reason for this draconian measure for preventing traffic from coming to my website. After complimenting for his own wonderful blog, this is not the least expected encouragement from a booboisie like him&nbsp;, but yet I asked why. He was very honest with me telling me that, my lil picture depicting some Afro-american guy is the reason.&nbsp; Yeah, may be. So true he was.   But this pic has a lot of stories, memories, tears ( !!!!)&nbsp; , which&nbsp;bind me with my Btech advisor Dr.Amalendu Mukherjee.&nbsp; The pic has been edited to show only me of course.Dr. Amalendu Mukherjee is one of those rarest of professors, whose company I enjoyed most. He would tell me stories about his son, when I asked him about the Nonlinear Dynamics class, he was taking. When I asked if entropy increase should be taken into account in the chemical reaction modeling problem for information storage, he galdly recited me some Urdu couplet, whose meaning only Khayyam can tell. When I asked him about writing my thesis, he talked about blackholes and distant galaxies. When I meet him in Park and looked into his misty red eyes, he took his eyes off me. When I asked for recommendation for my graduate studies in USA, he talked about the French delegation who were coming to Mechanical Engineering Department for learning Casting and Welding from Dr.Kajal Biswas. This is the happy part of the story.   The sad part began, when I got no scholarship to attend a grad&nbsp;school in US; when I got a B in project eval. When I went to say good bye to him and he asked me for working as PhD in Mech Engg Dept. After such poor performance, when he had given me B for project evaluation,&nbsp; it was very kind of him to ask me to continue my studies under him. I am really grateful, but I am not shameless enough to continue my study in IIT Kharagpur, where I have been such a loser and that to under esteem patronage of him.&nbsp;In the mean time, I was hearing from some of my batchmates, \"  Brother-in-Law, what happened  ?\". But, in between those happy and sad stories, there were tears. I cried for something, I cried for someone like a kid showing all tantrums and I got the biggest prize of my life.  '),(21311,3889401,'18,July,2004','2004-07-18 00:00:00','   In the middle of the night, I wake up to find there is no bed I am sleeping on, there is no one I am sleeping with , and there is no myself on the bed.&nbsp; I get up in the morning to get myself into the bathroom (or restroom !!)&nbsp; and stand before the mirror naked to find out, the mirror does not show me. Probably, lights even don\'t care to repflect and refract from my skin. I have become may be too porous for the light. My heart has melted into the air and soul, which I never found, was missing. With the open bright sunlight outside and the melancholy in the soothing breeze, I venture out my adventure to my lab. Its morning by now.  &nbsp;  There will be piles of unedited programs, randomly placed print outs and my lab mates would ask me \" Amit, where the fuck have you been?\" I have nothign but a synthetic smile to add, which would send them the reply that I have no interest in them. Then suddenly, someone would tell me my advisor has called me and left a messeage. I would visit him with slow but steady foot and he would say all good words for my effort. I will stick my head to the computer till the rest of day increasing my eye power from 2 to 3 diapter. And as I woudl feel sleep, I will think that its night now and I must go to the apt.  &nbsp;  Next day, when I come to lab and ask about the yesterday , everyone would say that I had not been to the lab yesterday and when ask them the thing they asked me yesterday, they would say they don\'t use the f-word in the lab. May be it was a ghost of a pseduo-man who visited my lab yesterday.&nbsp;  '),(21312,3889401,'16,July,2004','2004-07-16 00:00:00','   Finally the day has come when I can ego-google and can see my website and my name appearign on the web the same way I wanted. My blogger profile comes first, if you you search me. And then some other Amit Mohanty, who I have never met; and then lol !! comes my SIU webpage. Though it shows as my previous shity geocities page; actually you enter my SIU page. Well, I have placed a redirector link on my geocities page. Yeah my comment on  urlLink BD \'s page also comes in the search.      urlLink Wanna Check the TRUTH..Click here   &nbsp;  '),(21313,3889401,'16,July,2004','2004-07-16 00:00:00','   I really feel downtrodden with the situation I am currently having. I really hate when I can\'t blog. Thesis blue always changes its color and anticipates me to catch it up with me. Though I am done with 50% of my thesis work, many things yet remain to be done. Initially, I was afraid if I will be able make my thesis even a 100 pages or not. Now, I am having so much of work and as I am writing thesis, I have been poured more and more liquid work; I am afraid how I am going to keep my thesis with in 150 pages. I am writing a lot of shit in my thesis I guess. I would have loved to have a more organized thesis. Sensor Fusion, System ID..Man, come on ...Its not an encyclopedia of Mechanical Engineering.  After long time I met Kanchan today. He (yeah..He !!!) is also from IITKGP. With him I feel so much like myself. In KGP where the profanity starts with MAA and ends in the same polysyllabic word, it is a pleasure to talk with him with the same degree of profanity.  '),(21314,3889401,'16,July,2004','2004-07-16 00:00:00','   What Vijay is to Amitabh B, Rahul is to Sahrukh Khan. It is the most used screen name  for Khan films and those were immensly popular back in India and abroad. Being a matured Khan-fan, though I discuss with Vikram how ridiculous some of the Khan movies are; but I do admit that I like him. The way he carries himself off the screen and the way he miscarries on screen and the way he preaches love-funda in his movies.&nbsp; And, above all his love story with Gauri Khan, his wife. But, I hate this name Rahul.  &nbsp;  For someone like me, who pronounces   Cricket is my religion and Sachin is my GOD   , it is difficult to see Rahul scoring more than Sachin. I hate it.&nbsp; Again, it happened today. To everyone\'s surprise, UAE guys got rid of top three batsmen of India, but unlike soccer, a blazing instantaneous spark of excellence can\'t win you matches. You gotta be consistence. Rahul scored a century and was man-of-the-match.  &nbsp;  But, to my utter surprise, it did not hurt me that much as it would have hurt few days ago. Sachin failing and Dravid scoring. Fact is that, some people like Dravid more than Sachin and I love them more than I like Sachin.  &nbsp;  Anyway, just hope Saurabh dada stops thinking \"  Every match is a  test  for me \".And gets back Asia Cup.  '),(21315,3889401,'15,July,2004','2004-07-15 00:00:00','   Have you ever wondered that the thing which you complain about most, you can\'t really live without it ? I have , I think. You complain most when you have a lot of anticipation from something and when you have a lot of aniticipation, that indicates you need that thing most in life and can\'t possibly do away with it. I have a lot of anticipation from Rain. There will be a time, when you think Rain has two facets, may be three or may be it is a multi-faceted thing, but truth is that you simply can\'t live without it. Biologically, its important and its has a face of humanity and love too, apart from that. As  urlLink Narasana  goes, many priceless things are associated with it.  &nbsp;  With no offence to  urlLink BD , I must admit though the post by him about rain was a universal hit in blogger community; but the thing which was the masterpiece was the comment by Narasana. It reminds me about the fact in life that most of the priceless things in the world are free and just at a inch away from you. OK, a nano-inch away. But, its about opening up your eyes and have a look and then a grab. No point scavenging life for shitty things , as we do, or atleast as I used to do. My priceless thing was so close to me and just that I had to make a move and all happiness bestowed upon me after that.&nbsp;  &nbsp;   It was so nice to find many friends like Narsana and Mitra all on blogger community and writing pieces of their beautiful mind. &nbsp;This guy Saurabh is an awesome awesome guy and his English is better than that of an average American ( No offence Uncle SAM..Truth is truth is afterall. ) Its not about English only, guys are recommended to go to his blogger and find out more about this guy.  &nbsp;  Good work Narsana and Saurabh&nbsp;!!!!&nbsp;  '),(21316,3889401,'15,July,2004','2004-07-15 00:00:00','   The engineers and scientists out there in the industry and academia are  scratching there head after the dotcom blast and then its collapse. What is the next thing that is going to create the same influx of money into their pockets? Well, if the Wall Street is to be belived then its the NANOTECHNOLOGY. So guys who are researching in that field are advised to be elated, atleast for time being. According to Wall Street, the money into nanotech start ups are flowing in a same way as it was 10 years ago into the software start ups. But, again it also asked to beware of the hype that is created around the start ups.   So, probably after Microsoft, now its NanoHard which is going to make dollar greener. Size does not matter.  urlLink BD , are you listening?  Source: urlLink  http://www.hindustantimes.com/news/181_885325,00040006.htm  '),(21317,3889401,'13,July,2004','2004-07-13 00:00:00','   I have been messy since my childhood. Mama used to say life is a box of chocolates and I have been believing her words since then but never ceased to be messy. I scattered my books all around the tiny world of mine which Mama called our home. She did not say a word, but came and kissed my head. I ate sweets on the bed and thus making my dad\'s pillow haven for ants which would bite his ears. Dad was so fearful, but ironically he came smiled and twisted my ear. I swear Dad was so skillful that it never hurt me.   I never organised my books, I never had my notes ready. I wore someone else\'s T shirt sometime. I crammed stuffs from Rohit Singh\'s note book and during job interviews someone else\'s trouser, someone else\'s tie. Nothign never fitted me and I never fitted to any job, company folks thought.  I kept my recommendation letter for Stanford under the keyboard of my computer and last date elapsed then I realised I have not applied for Stanford though I wasted my dad\'s 80$ by completing the form online. By the way my dad earns around 200$ per month in back India. My dad came to know and laughed so much as if I said some joke. I really doubted his mental ability.   So if I am so messy, all fault goes to my parents and my friends who have always encouraged me to do so.   In a grad\'s life, where everything is procrastinated for eternity, guess how will you feel if you have to redo a old stuff being awake for two nights. I will go freak,  man !! And exactly that is what I am now. If you guys are kind enough to go to my research page which starts with an erotic line   \"Research is like sex...\"   you will find a link to the paper for IMECE conference. Suddenly after I lost count of where I placed the paper, my advisor asked for the paper and I was like I am dreaming some nightmare or what. Somehow I got hold of the paper in PDF format but yet I had to do a lot of work converting into Microsoft Word. ( If anybody needs any consulatation regarding this, he is most welcome.). Now loss of one night of graduate sleep in mental turmoil and another for working is what I paid for being messy.  But, surely one thing .. I love my parents a lot, and I swear I do, for spoiling me ; and I really like my friends for allowing me to remain unfitted for any good thing in the world. I wish I could say such words for other people.  Again, I have more serious incidents which I can tell you. Had those not been very personal, I could tell you what real price I paid for being messy.  '),(21318,3889401,'11,July,2004','2004-07-11 00:00:00','   As he graduated, one of my friends here in Carbondale moved to Detroit, Michigan to his cousins; so that he can again lead the same story of a new Desi graduate searching a job. Procrastination will not do anymore. So, we people went to Chicago to drop him off. Then he took a train to Detroit.  Before we could return to Carbondale we had some time to have a look at downtown Chicago. They say if you have not been to NY then you must go to Chicago. Downtown Chicago near the lakeshore is secondmost auspicious place for lovers I guess, Times Square being the Mecca. I heard someone saying \"  If you have my tears in your eyes and I have yours in mine; then that\'s more passionate that smooching  .\" But, if you are near Lake Michigan; you don\'t need no tears, you don\'t need no kissing. Every moment you spend  with your girl at that place will be a redefining moment for you. These are the few rare occasions when I get romantic and cruising down the shore in Chicago is one of them.  But, you can\'t love when you have an empty stomach. Searching a place , to be more specific a desi place, we drove to Devon Street. It is like India amidst USA. They have got things like Hamdard Medicare, Patel Brothers, Maa Laxmi Jewelleries. And mind you, the gold jewelleries which you find there are 22 Ct. unlike 10 or 12 like rest of US. There is this Gandhi Street and Jinnhan Street, named after father of our nation and father of a friendly neighbouring country Pakistan. There is this place GAREEB NAWAAZ I would recoomend every meat lover to go. The food is cheap and deliciously desi in taste.  Though all my friends were interesting in ogling at summer babes in Chicago, we had to return early , because we had procrastinated the weekend assignments for this Monday night.  '),(21319,3889401,'11,July,2004','2004-07-11 00:00:00','   urlLink Vikram  has been incorrigibly unpredictable yet a stable chaotic system since his birth. As his own word goes for all of us BAND OF NINE, we all are  miserably beautiful, erratically exciting  and  fuckingly cool  guys (including him!!!) . I have been ever jealous of him during my stay with him during my IIT days. My inelegant way of pronouncing the polysyllabic Hindi and English words was the source of constant of entertainment for my wingmates. While, unlike me, Vikram , though hails from the same part of India as do I, can weave webs of words around all of us in a very poetic manner and  his eloquence and etymological calibre  left me in  a unbelievable state of scariness . The phonophobia in me for English and Hindi reasons back me to him. But, time and again his surprising elements through his poems never dies. I got a snail mail from him few days ago (in this age of Internet and all !!!) and this poem accompanied that letter.   Totally New     Totally new to this I thought I would miss But then came the target It was a claen hit  Totally new to that I thought I was bad But then came the light It was a start act  Totally new to act I knew I would fart But flew past the first note And I sang till the last  Totally new to the vast sea Sure was I to first pee But then came the king wave And I stoked a killing spree  Totally new to the act of love I declared I was a sitting duck But then came by the prettiest girl And we redefined the meaning of fuc*  Totally New to the core emotions Somewhere inside was a starnger me But then came my first child A relection set a father free  Moments Unparalelled, My life unbounded Unfolded before me Generation of LOVE.  Everything I did I did for the FIRST time Every single thing Strange except one...  Totally new to this world line I thought I must have cried But them claimed my wonderful life Haven\'t a smiling man died !!!  '),(21320,3889401,'10,July,2004','2004-07-10 00:00:00','   urlLink BD  says procrastination is an addiction. Although I could not agree with him more, but I think the biggest addiction is the blogging. Every night, you feel like spending those few precious hours of graduate sleep over bloggin. When you are doing slow simulations in some P3 comp, then you open up blogger to read blogs of your friends unknowingly. And in the end,you realise you simulation has given some wired results, to interpret it you will probably call some PhDs or you will be fainted to unconsiousness thinking that you have to redo the simulation because of some goofy errors. Its all because of blogging. But, I can\'t do away with it I think. Just 24 hours but I have been addicted to it like ..like...say BD.  I was reading  urlLink Ankan\'s  blogs and that PhD qualifier stuff really got a hell outta me. I have yet to appear in PhD quals.   Since last 24 hours I have spent 60% of the time creating my webpage, rather helping BD creating my webpage. And then linking my bloging page to my webpage. Is this meaningful for a grad student? Answer for this question is another question. Do grad student do anythign meaningful? Anyway, I can say I learnt a lot about CSS Style Sheets and above all, it all fun and no work.  '),(21321,3889401,'10,July,2004','2004-07-10 00:00:00','   This is my first blog. I exactly don\'t understand my feelings now. May be the early morning laziness has made me numb.Now, should I say \"Hello World\" or just keep writing my first blog as if I don\'t know no-one? Last few days have been not so pleasant for me with my guide testing my endurance limit and me submitting to his wishes with a smile on face and pain in a**.  Anyway, but in between this dark patches of life, people like Biswanath Dutta comes in from nowhere. Just a phone call to say \"hi\" or to buzz me at yahoo Messenger. Our undergraduate days have seen us discussing books, SUTTAS, Engineering, SEX and LOVE. We discussed everything. I started doing so many thigns for the first time being inspired by BD, as we all fondly call him. Be it  LaTeX  or be it my  webpage . Yet again he has done it. I am writing my first blog being inspired by him.  So, no \"Hello World\" for me, its like \"Hello BD\" for me. I learn many thigns from everyone in the world and more so from you.  '),(21322,3889401,'12,August,2004','2004-08-12 00:00:00','   Twenty four hours, three states in US covered, driver lincense one week old, cops caught two times. Well, that sums upto make the voyage. Its no not any Arabian Night story.  6.30 AM -- Purdue, INDIAN.  I start off with a crusing speed of 80 miles as the traffic was low and cops were probably still having the hang over of last night just like me, not to catch me. Sometimes, I felt like a ball throw by Akhtar, speed was kissing hundred. I survived to reach my next destination in Champaign.  11.40 AM -- Champaign, ILLINOIS. I rode greyhound (the bus service) to reach Carbondale in 4.30 PM. I rented a car and started off for St Louis, Missouri.  5.00PM-- Mt Vernon. I took the wrong exit and was moving all inside Mt Vernon for fifty minutes. Finally, I got the right path after asking 72 people.  9.40 PM- Still on the route to St. Louis. Got a call from Tarini, BIotech, IITKGP, Batch 2004, that he has arrived at Lambert International Airport. Fcuk !!! OKAY, I said wait for sometime, may be for hour ..How about two hours? I would reach there.  10.10 PM- I missed the exit for Lambert and was inside downtown St. Louis. Four lanes. Speed 70MPH. I was wondering where the hell I was in. A truck passed beside me. I could see in slowly looking towards my right. OK..Truck ..It would pass in quickly. 10 seconds gone. 20 seconds passed by. What the hell it is carrying? KutumMinar or what? Now, another truck on my left and a big, sharp turn. Wow, it must be feelign great to be sandwiched between two slices of trucks. I would exactly feel the way a chicken breast feels when I have that in McChicken.  Just that  providence wished otherwise and I survived.  10.20 PM--  Some thign flashing on my rear view mirror. What the hell? Why these guys use these colored light flashing all above their head. Hmmmm..Coz they are cops , a lady cop to be specific.  My innocence put on my face saved me.  My life is not that expensive, I can probably take it for granted. But, while returning I was a better driver, coz Tarini was with me.  '),(21323,4101011,'31,July,2004','2004-07-31 00:00:00','    i\'m sure, now that i think of it , that your family was just looking for a reason to goad me.  seeing as how they\'ve never ever liked me and now look what i\'ve done and all.  and that\'s fine.  that\'s fine.  fuck that.  i don\'t know if you will read the rest of this email, because that paragraph i just typed is about as rational and levelheaded as i can be.  what\'s left is discombobulated angry mess.  it\'s all raw and terrible and you don\'t deserve it, i guess.  but it goes like this:  you can be such a pathetic, manipulative, vindictive, snide bitch, ____.  when i think of you i try to imagine how you were ever soft or sweet or good, because what\'s left is bitter and ridiculous and base.  you say you loved me, but you never fucking did.  you wanted some great girl goddess #9 to sweep into your little closeted life and find your cunt and make you alive.  so i did.  you wanted some sick mothering figure to tell you it was okay to act like a 5 year old and chase away your bad dreams and loveyouloveyouloveyou no matter what.  so i did.  you wanted to know what sex was.  so i showed you.  you wanted to kiss a girl.  so i kissed you.  you wanted it to be hot.  so it was.  you wanted to make love with some fantasy of a soft sensuous heifer.  you wanted my voice and my bluegrass songs.  you wanted my childhood.  i gave you these things.  you wanted to give me everything.  so i let you.  you wanted to validate yourself in a relationship and prove your sexuality.  you wanted to explore me and use my love as an inspiration for every other experience in your life.  so you did.  you wanted my chubby hands and my nose and my red hair and my cat and my sleep noises and my rainbows and my smile and my style and my dreams, ____.  i gave you everything i could.  you turned every fucking thing around until it was about you.  you made every situation a testament to your neediness, jealousy, or insecurity and i tolerated it.  i indulged you.  i did my best to put your fears to rest and keep your insecurities at bay because i wanted to care about you and i wanted you to have what you needed.  i let you pretend we were different people so that you could be okay, ____.  i made myself a different person.  i tried to help you be a different person.  fuck you.  fuck that, you are not a different person.  i am not a different person.  and this is the part where it really gets messy.  this is the truth that you already know that you nevernevernever wanted to see and i didn\'t want to show you.  why?  because we always thought it could be different.  i kept waiting for you to show me that moment, that glimmer that would prove to me that our relationship was worth the sacrifice and the compromise.  the moment in which we could be ourselves and actually understand each other.  the moment that i would stop regretting my choice, when we could be real.  i pushed us toward that moment but you always pulled back and it never happened, and it never could have happened.  the great big obvious truth is that we never should have been.  fundamentally.  i mean in that great philisophical way that all things happen for a reason and that we learn from all things, it was all fine and well that we wasted 3.5 years of each other\'s lives.  but actually.  really, truly, we never should have been together.  and in fleeting moments i miss you so strongly.  i don\'t know what i miss, any more than you know why you can\'t stop thinking about me when i obviously hurt you to your very core.  it\'s proof that we made a connection.  that\'s all.  it\'s moments of comfortability in which you think it\'s okay because there\'s a warm body or song to sing or a trip to take.  you think it\'s okay because you make plans or you spend money or you move around the fucking country chasing after ideas that you can only hope will finally come to be realities.  you think it\'s okay because there\'s nothing immediately better.  you think it\'s okay because you want it to be okay.  because you want it to be easy.  you think it\'s okay because you can cum.  or because you cook.  or because full house is on.  but nothing ever made it okay for very long, did it, hon?  even in those silly happy wonderful roadtrip sex-in-a-thunderstorm let\'sorderpizza idroveallnight ofcoursei\'llmovetosouthcarolina loverstobed littlesnack chiquitabanana moments there was a darkness.  here, you will think that darkness was ________.  you will think i could never be happy with you because of my feelings for him.  you think that because i thought of him, called him, missed him, visited him, wrote him, and eventually showered with him, laid naked with him, and \"SUCKED HIS DICK\" that i could never be happy with you.  but that\'s not it.  it wasn\'t because of him.  it was because of us.  so it was a great big fucking waste of fucking precious time.  yours and mine.  and all the fucking cocksucking i could ever do would never be as terrible a thing as you lying to yourself about who you are, as terrible as you pretending to be happy, as you suckerpunching our relationship with your detestable manipulations and fucking headtrips.  and what\'s ultimately terrible, the most terrible thing of all is that now you think it was me.  you think my untruthfullness brought this sad relationship to its tumultuous end.  you think that my lips on his penis erased so many false words falling from your own lips, that my secret private thoughts of him were more sordid than your secret private thoughts of other women, unrealized even to yourself because you\'re not strong enough to know who the hell you are.  i knew.  i loved you anyway.  it didn\'t matter who you wanted to fuck or what you might have thought about other people.  it\'s a shame that you never took advantage of my open-mindedness or acceptance.  it\'s a shame that you only milked me for your specific needs and demanded that i keep myself as sheltered and as blinded as you were.  it\'s ridiculous that you should have asked me to lie to myself just like you and deny who i am just so that i could be with you.  just because you were already there.  it\'s offensive that you think my bisexuality is a big fucking cruel joke and that i just like to throw it around to make you feel bad about being gay.  that i\'m a slut who would fuck just about anyone of any sex for any reason at any time.  beacuse i\'m like that.  i gave you moments of my life that meant something.  i could never trust you with the moments that meant everything.  that\'s not my fault.  you left me once and you took it back.  there were feelings of relief that i was afraid of.  what did it mean that i was almost glad you didn\'t want to be with me anymore?  what did it mean that you got drunk on the beach and suddenly realized that this was the most ridiculous summerville situation of our lives?  why, then, did you want so desperately to revoke those words?  why did i let you?  you left me once and you took it back.  you left me twice and i\'ll never see you again. it is with respect for all of the suffering we have caused each other, and all of the precious moments we can no longer trust that i say \"...finally\".   '),(21324,4173964,'20,August,2004','2004-08-20 00:00:00','   Fuck me am I beat to shit. Today was suppose to be my day off but 9 am rolls around and guess who calls? No not my dream girl...goddamn supervisor calling me in for a 8.5 hours shift. So i stumble up there and get to work. A few hours later I find out this little doofus Tristan isn\'t showing up. So me being infinitely wise I volunteer to work overtime. I figure it\'s good money so why not? By 8:30 I\'m beat to shit and decide to peace. But alas my parents aren\'t home so I have to drag my ass down home. Good god I can barely see I\'m so tired.  What else has happened?....Yeah not a thing. Just wanted to bitch. Gonna fall asleep to the sweet melodies that are Deathcab for Cutie.  Adios '),(21325,4173964,'19,August,2004','2004-08-19 00:00:00','   Dunno why I\'m here. I really have nothing much to say I\'m just uber bored due to the 3 day hiatus on gas pumping due to lack of scheduling. I think I\'m gonna trim my nose hairs then go for a jog. I must see Napoleon Dynamite. A friend sent me the dance scene and I cried I laughed so hard. I\'m doing it for the next dance show hands down.  I\'m in  a better mood today. Dunno what was wrong with me last night. Maybe it\'ll kick back in tonight and I\'ll figure it out.  I\'m watching Newlyweds season 2 right now. Jessica my oh my. How can something so hot be so stupid? Ok yeah easily I know I know.  I\'ve decided I don\'t need a girlfriend I just need to get laid. I\'ve got alot of pent uo sexual frustration. Need to let off some steam. Not saying I you know...would be rough or something like that...wow am I sounding like a creep here. CHANGE OF SUBJECT!!  I started listening to some of Johnny Cash\'s old stuff. Quality shit. Anyone who thinks they know everything about him because they\'ve heard \"hurt\", you don\'t know shit. I\'m getting tired of all these little 15 and 16 year old girls saying they love him and bands like  Zepplin and shit like that cause they\'re trying to be trendy. I went to Market Mall yesterday. UGH. So many wannabe trendy crap ass stores. All you see is what Cosmo says is cool these days. Even that \"Play\" is shit. trying to pretend they\'re indy but the second you step in the first album I saw was Jojo. I\'m not saying I\'m some indy freak. Just saying mainstream and indy are two completely different worlds and it\'s lame that they\'re trying to mesh them. Why can\'t people just be themselfs and stop having to be told what\'s cool or not. Gah lame.  Hey I bought Godsend yesterday. About to watch it. Looks scary as shit. Send in the clones oui? '),(21326,4173964,'18,August,2004','2004-08-18 00:00:00','   I\'m in a very strange mood tonight. I worked til 11 and had to walk home. All the way I listened to Dashboard Confessionals and it\'s made me very depressed. I just don\'t want things to change. I like how everything was during the school year. Even the bad stuff. At least I knew I had time to fix things. Now it feels like there\'s this huge weight pressing down and no way or time to change it. I just feel like I should have done more or resolved things more...resolutely. This summer has been an overall let down and for things to come to a close so abruptly and depressingly sucks. I just want to scream and let it all out but I\'m scared I\'ll never be able to stop. I know that sounds lame but it\'s the only way I can say it. Fuck it\'s too late for this. I need some time to think things through. Tomorrow is another day after all. '),(21327,4173964,'17,August,2004','2004-08-17 00:00:00','   Wow and now for something completely different. so I was originally suppose to head down to Kensington with my dearest friend Kirsten for a day of jugo juice and boy watching (for her not me I swear). She has to cancel but I end up heading down anyhoo. So I\'m wandering around looking for something productive to do and out of no where my friend Julia appears. We get to talking and she introduces me to her friend Sara I think her name was. We end up spending the better part of the afternoon together which was a blast. I head home and am not in the door more than 10 minutes when I get an invite from the one and only Vanya to come hang out with her and Christine. So I head \'er over. The night started off normally. The usual snide remarks from the duo about me being a closet gay and then them grabbing each others ass and randomly spooning each other. We head to Vanya\'s and hang out there. On the way I saw my dear friend Janine (I have way to many friends who are girls. Time for an infusion of testosterone and sex...With women) Anyhoo it was the stereotypical night of a regular group of kids. Which is the exact opposite of my usual night (ie sit at home and say hello to miss Palmela Handerson). On the way home we listened to Deathcab which is one of my all time favorite albums. There is something extremely surreal when you sing the lyrics \"She is beautiful but she doesn\'t mean a thing to me\" right next to the girl whom until the very recent past I was coocoo over. Whatever. Probably never gonna see either of them ever again. Which is quite depressing actually to think that that might be my last memory of either of them. I\'ve said it before but I\'ll say it again if you\'re reading this either of you. Have a goddam good life. You deserve nothing short of that.  *Heavy sigh* Goddam tired.  Found out my uncle whom I haven\'t seen in over 3 years (Dad\'s brother see Superfreak for an extremely abbreviated version of that story) is dying which is strangely depressing. The man was an asshole to me even when I was still seeing my father so why should I care? Fuck death is so depressing.  Whatever nothing much else to report. G\'night  Oh I started a nexus account. I\'m jumbojimbo for anyone who wants to add me. It\'s still shit I\'m gonna spend all day tomorrow editing it. '),(21328,4173964,'14,August,2004','2004-08-14 00:00:00','   So today was a day off work. I sat on my ass most of the day and watched \"Boys n the hood\". Gotta be on of the greatest movies I\'ve ever seen. Just mind blowing how powerful it is. I was planning on going to this play at Olympic plaza with that girl. At the last minute she calls and says she doesn\'t have the car so she doesn\'t want to do anything tonight. Whatever. I decide to go down and meet the people we were going to see it with alone. I can\'t find them anywhere so I end up sitting in between the two people who encompass every stereotype me and a close ami hate about theater folk (I finally mentioned you in here Kirsten now...DANCE WITH ME). On my right was an old woman who wouldn\'t stop sniffing and sneezing through the whole show. On my left was a guy who laughed at all the wrong places (because of them funny accents) and wouldn\'t stop talking to what I believe was his wife/girlfriend/goat...UGH!! Bad day.  Came home and went for a run with my sister. Mother is being a C-rag (y\'all know what that c stands for) as usual. Can\'t stand to be in the same room as her for more than a few minutes. She just randomly insults me for no apparent reason other than the fact that she is miserable and you know that saying... Yeah I actually can\'t quite recall it but whatever. Calls me weak and scrawny....Which I am in one regard but I could run circles around her fat ass... Bitch...Fuck it one more year left. Can\'t fucking wait. I\'m putting serious thoughts into saving up and moving into some cheap apartment in Bowness during the year. It would be better than living here. Wow is that a depressing revelation.  I watched Kill Bill vol 2 again. So radical. Sam Jackson...Oh how I love thee. Even if you only had 4 lines. I wanna watch the Shaft remake now. Or maybe the new Star Wars...Or how about (duh) Pulp Fiction or (my personal favorite) Jackie Brown. All have much more Sam Jackson goodness. hhhmmm what else  Yeah it surprising how little has gone down these last few days. I went to a party last night. Got drunk then bored and stumbled home. A bunch of girls there told me I was the hottest guy there which was a definite ego booster. I\'ve decided to myself this is gonna be the year of no regrets. Basically my whole life up until now has been a series of missed opportunities. I have enough regrets from the last 3 years alone to last a lifetime so I\'m gonna try and not add any more this year. I don\'t really know where to start seeing as how my life is pretty much in shambles as of this writing. Whatever time for bed. Got an early start tomorrow (have to be at work by 11 am MY GOD) and so I shall depart and fall into dream world aided by the sweet melodic voice of one Bob Dylan.  You stay classy '),(21329,4173964,'11,August,2004','2004-08-11 00:00:00','   WOW it\'s amazing what a few hours can do to completely change your mood. So I was all bummed out today thinking about my dad and such as I headed to work. I get there and find out I\'m already 10 minutes late for an orientation at the main branch (Breantwood). So me and my buddy Tyson (whom hadn\'t been told either god bless the managing staff) burn down there and get in just fine. The first half was boring shit (as is the first half of this blog HUZZAH) but when we got back from break it all went uphill. I was sitting beside this mondo hottie named Tannis (like tennis but with an a she says and I love her already due to the wicked ass name) and we start talking. Turns out she\'s pretty much the female version of me (but only slightly hotter ;) ) and we hit it off immediately. Many was there a time where we were chastised for talking by the speaker (who\'s sister happens to be one of my best friends girlfriends). Anyhoo the end of the orientation comes and I figured \"Oh well here\'s where it ends\" As usual right...BUT NO!! Out of the blue she slips me this piece of paper with her number on it and for me to call her tomorrow when I get off work. So now I\'m sitting on cloud nine and don\'t think it can get any better. BUT IT DOES!!! She lives IN VARSITY FUCKIN\' ACRES!! Which for those of you who don\'t know is the community adjacent to mine. Man oh man am I feelin good. Yeah that was my day...I\'m gonna go listen to some Deathcab or maybe some Trews or maybe something completely unrelated to either...OR MAYBE ALL OF THE ABOVE I\'MA CRAZY MAN!!!! Yeah I\'m really not....you get the ideer. Anyhoo this is James signing off. You stay classy Blogger.... '),(21330,4173964,'10,August,2004','2004-08-10 00:00:00','   Well the title of this blog has 2 meanings. First of all I just wanted to say RIP Rick James. Cocaine is a hell of a drug. I don\'t think enough people appreciate his contribution to music. Alright so maybe the Mary Jane girls weren\'t that great of a contribution but still....He will always be the man.  Secondly I guess it reffers to myself. Lately I\'ve just been very off and I think it\'s because I\'ve been trying to get stuff off my chest but won\'t because I\'m trying to not piss people off. I\'ve decided to stop pussy footing and to say what U want to. If it pisses someone off or offends them in any way they can go fuck themselves.  ...Funny thing is now I can\'t think of what to say or where to start....  Ok... I\'ll start with my father. I\'ve been working up the courage to talk about him for awhile now and I guess might as well do it now. Where to begin. I guess I\'ll do it chronologically so I don\'t have to go off on tangents.  At age 6 My parents get divorced. My father moves into the house next to my grandparents which happens to also be the house my grandfather grew up in as a child. My dad lives there for 4 years. During this time he meets and impregnates another woman. He eventually moves in with her in a house in the same community as me. He claims it is to be closer to me and my older siblings but in truth I begin to see less and less of him. My step-mother out right says in front of me if she had her way I wouldn\'t even be in their lives. My half brother is born when I\'m 10 and I\'m immediately labeled the baby sitter as neither of his parents are willing to put any kind of effort into his upbringing duringthe school year. During summers they pretend they\'re the perfect parents, taking him to birthday parties and movies and always dragging me along so they can talk with their friends and I watch the kids. At 13 I steal $90 dollars from my step-mother. I am caught and forced to go to couciling with my father. It\'s all a sham and is provided free by the school board. She\'s not even a real coucilor but my father is utterly convinced that I am a basket case and could blow up at any minute due to my \"cry for help\". This coming from the man whole thousands of dollars from my mother and grandfather during the divorce. During that summer my father sets up a bank account for me that he will deposit i think it was 20 dollars a month to prevent me from stealing again. The catch is I have no access to this money. Only he can get to it. I ask him for the money and he blows. An argument ensues (of coarse) and I don\'t talk to him for the next month and a half. Eventually things are smoothed out and for Christmas I buy him some Yanni cd or something like that. Some new wave shit because he thinks he\'s sooo deep. A week before we get in to another argument and while he is out I take back the cd. On boxing day he finds out about this from my grandmother (who\'s probably clinically insane) who also tell him she paid for it. He flips out and says I\'m stealing and yadda yadda. On top of this my step mother threatens to phone the police and say they were robbed due to my actions. I tell her to go fuck herself and my father kicks me out of the house. He says completely out of the blue that I was the cause of the divorce and that he\'s always happier when I\'m not around. Since that night we have passed a few letter back and forth. Mostly me asking for my things back and him saying no. I\'ve seen him once or twice and always he\'s been a cold hearted fucker. I realize I\'m not completely innocent and he\'s not completely at fault. Ya just gotta meet the man and you\'ll understand. He\'s never shown the slightest bit of affection to me. I dunno. It\'s alot bigger than this little paragraph. It\'s 17 years of bullshit. How do you boil down every little event into this one space? There\'s a shitload more to it I just don\'t feel like typing nor do I even feel like telling. Somethings should remain personal. Actually upon re-reading this all these problems seem rather trivial to me. I dunno. Don\'t all think I\'m some kind of pre-madonna who flips at the drop of a hat.  What more can I say about Vanya? I\'m just throwing this one out there maybe it wasn\'t even about me but I\'m assuming it was because i\'m an overly paranoid freakazoid. I wasn\'t trying to be creepy and know everything about you or something like that. I just honestly happened upon you blog mcblooger thingy (whatever ya call these things) and didn\'t mean to offend you in any way by doing that. I wasn\'t trying to be creepy and I\'m sorry if it came off that way. I realize I\'m probably still coming off creepy cause I went back. In all honesty I don\'t know what to say here. I\'m sorry if I upset you in any way, shape or form. That was the furthest thing from my intentions. In fact I didn\'t have any intentions when writting this stuff other than to get it off my chest. I didn\'t think you would ever read any of my entries nor did i honestly want you to. I realize telling Christine was probably the biggest mistake I could have made. No offence by the way Christine. Whatever I hope ya can forgive me for...whatever it is I did and just put all this weirdness behind us. If there even is any. Maybe i\'m just being like I said overly paranoid. but then maybe I\'m not. Bu then again, and i truly mean no offence to you in saying this, you are as everyone knows moving to Norway at the end of the month. Do you really think any of this bullshit is gonna matter over there? You probably won\'t even remember my name. That\'s not some kind of pessemistic comment towards myself. It\'s just what happens. Anyhoo I don\'t know if this really had any point I just thought if it needed throwing out there I\'d do it. And if not HEY more weirdness oui?    I dunno what else to write about so I\'m just gonna throw myself to the dogs and send this one out. '),(21331,4173964,'10,August,2004','2004-08-10 00:00:00','   Who the hell came up with the name for this thing? Blogger...The hell? Whatever. Wicked ass day at work. Barrels of premium oil for anyone who knows what that means ;)... What else.... I had McDonald\'s on my break and almost threw up after. Gotta break the habit and stop eating there. So gross.... Hey that reminds me I\'ve been meaning to rent Grosse Point Blank for awhile now. Great movie if ya haven\'t seen it go rent it now... That is unless you\'re going to the Rogers by my house in which case wait til some other time. Beat Tony Hawk 4 before work...yeah I\'m starting to realize why I be on the single side... I saw some friends of mine at McDicks and got invited to go partying sometime this week. Probably not gonna happen as usual but hey it\'s the thought that counts right? Heh I\'ve been listening to the College dropout album by Kanye West as I typed this and a funny little interlude came on.  \"Now beat that and your mother says go to college so you finish college and it\'s wonderful. You feel so good. And after al lteh partying and crazy and don\'t forget that drug habit you picked up at school hanging around your peers HEY now you\'ll get that 25 thou job a year and you\'ll spend all your money on crack cocaine. But it\'ll be your money. No more borrowing from mom or my hide. So now you get your degree tattooed on your back you\'re so excited about it. If you continue to work at the Gap after several interviews (OH MY GOD!!) You\'ll come in at an entry level position. And when you do that if you kiss enough ass you\'ll move up to the next level. Which is being the secretary\'s secretary!!! And boy is that great. You get to take messages for the secretary who never went to college she\'s actually the bosses niece. So now you\'re part of the family. You know what college does for you? It makes you really smart man. All you kids wanting to talk at the back of the class NOT ME I listened ok I was a hall monitor this was meant to be. Do you know how many extra classes I took? Extra classes extra classes. NO I\'ve never had sex but you know what my degrees keep me satisfied. When a lady walks up to me and says \"Hey you know what\'s sexy?\" I say \"No I don\'t know what it is... but I bet I can count up all the change in your purse very fast.\"  It\'s funnier when ya hear it I\'ll admit. But it\'s still amusing.  Wow was that a waste of time. Isn\'t that the story of my life eh?  So now that I\'ve wasted some more space and air I guess I shall depart and do what I do every night...Try to take over the world!!!! Or just you know....Go to sleep... '),(21332,4173964,'09,August,2004','2004-08-09 00:00:00','   I feel like being jessie today so I\'m gonna write a really lame poem. Here goes  My heart is like fine China A beautiful thing to behold yet just as easily broken And once it has shattered it can never be repaired ....I want a fat bottom girl.... *Tear* Why does it sting when I pee....  Yeah there ya go. My best Jessie impersonation. Anyhoo...getting ready for work again. I loves me that vehicular orifice pumping. Uuummmm stayed up all night and read the Beach by Alex Garland again. Best book I\'ve ever read. I\'m dead tired now but whatever. I\'ll just drink a shitload of coffee and coke and such at work. Kill Bill vol 2 comes out today. I\'m going after work to buy it. AND Alien vs Predator comes out this Friday. Could this week get any better? Well yeah absolutely but...you know... Yeah this has been pretty pointless I just felt like writing something. '),(21333,4173964,'09,August,2004','2004-08-09 00:00:00','   So today was my day off of work and I sat around on my ass watching movies. I went from the Last Samurai (Tom Cruise as the only samurai to survive is strangely depressing) to Dogma (Need i say more) all the way to Freddy vs Jason (You all know who won quitcher bitchin\' Freddy\'s a pansy). Anyhoo... yeah what else. Listened to some Meatloaf... rocked my world.  Goddamn computer keeps crashing. there it goes again. fuck. back... okie dokie  I\'ve had alot of time to think things through today and alot has become clear to me. I\'ve decided to quit smoking and probably drinking after the summer all together. All it does is pollute my body and cloud my judgement and I sure as shit don\'t need that right now. I see all the people at work who smoke and they\'re going nowhere in life and that\'s not where I wanna be at this point.  I\'m gonna take a step back and not be so hung up on being single. All things come in due time (teehee I made a double entendre). I don\'t need the bullshit of a girlfriend right now.  Everybody who\'s read my other posts and who know me really needs to take it down a notch on the whole Vanya front. Everybodies saying how I\'m not over her and I shouldn\'t bug her and shit like that. All I have to say about that is this... I wrote all that to get it off my chest. I wasn\'t writing it for you, or her, or anyone else other than me. I had to get somethings off my conscience and this was the only way I knew how. I know how I feel about her and I know how she feels about me and I was fine with that. I wasn\'t going to say or do anything about it. She\'s going to fucking Norway in less then a month for fucks sake. What could I expect. Not fucking worth it for such a short time frame. She\'s a great person and I\'d love to continue being friends with her long into the years and I don\'t want some goddamn teenage crush to get in the way of that.  I\'m coming up on my last year of high school and am absolutely terrified. My marks are nearly good enough and I have no idea what I\'m going to do. I\'ve always dreamed of being an actor. Not a big huge movie star I\'ve just loved performing for people and trying to make them happy. But now I\'m starting to doubt that. My drama teacher has told me I\'m capable but no one else seems to think so and, no offence Dean, but he\'s only a teacher and a great judge of ability *cough Elizabeth cough*. My step-dad is pressuring me inot being a cop. Why the fuck would I want to be one of those? He\'s utterly convinced he\'s going to brainwash me into becoming what he wanted to be if his life had gotten so fucked at such a young age.  Funny thing is I\'ve been going through all kinds of shit these past few years and I don\'t it\'s really even hit me yet how much it\'s affected me. A normal person would be a fucking basket case if confronted with half the stuff I\'ve experienced. I don\'t know what\'s wrong with me. When my dad kicked me out and did all his bit (someday I\'ll elaborate) not a tear was shed from my eyes. When my mom told me I was a mistake and she wished she had never had me... still nothing. I think I\'m dead inside. I don\'t know what\'s wrong with me. I always think about what it\'s going to take for me to break. I\'ve seen so much in such a short time that I can\'t even imagine what else life could throw at me. It just sucks that I have nothing to comfort me with. Most people turn to religion but I\'ve never been one to take comfort in words written thousands of years ago that may or may not just be all bullshit. I don\'t want to dedicate my life to something that may just be the figment of some loonies imagination. Why is God real and Superman isn\'t? The only proof for either are written word and pictures... what\'s the difference? I don\'t know what I saying here. I\'m really fucking confused right now. I just want to feel something different for once. I\'m tired. Tired of all this. I just need something different. '),(21334,4173964,'08,August,2004','2004-08-08 00:00:00','   You know who I\'m talking about here so I feel no need to use names. I posted earlier (I\'ve since just outright deleted it) how I \"adore\" you and the usual mushy puppy love crap (why oh why won\'t she like me *barf*) and have come to the realization that it\'s all bullshit. It\'s not worth it. I can\'t keep thinking in the mind frame that we could be together, that we could be something we can never be. It\'s too much for too little. Why am I literally laying awake at 2 in the morning thinking about all the things I want to say to you when in fact it\'s all just hormones. I don\'t love you, how could I barely fucking know you and am never gonna get the chance go go anywhere beyond that. It\'s not even the age thing, if that was ever a factor. You\'re just not what I thought you were. When I met you I saw everything so differently. I can remember the first time I saw you. I thought you were the most beautiful girl I had ever seen. When you started talking to me it was merely icing on the cake. I could never see anything ever again and be happy forever with merely that image of you in my head. And for a time we were happy. Fuck me I\'m making this bullshit into some kinda motherfuckin soap opera. AND NOTHING EVEN HAPPENED!! IT\'S ALL IN MY FUCKING HEAD!! I honestly thought what I felt for you was love. I\'ve never felt anything so strongly for another human being as I did for you. And somehow I managed to trick myself into thinking you might feel the same. I\'d pretend a sudden glance or a grazing of our hands was some kind of signal from you. Who the fuck was I kidding. GOD and now I\'m mad at you for not doing a thing. What the fuck is wrong with me. You didn\'t do a thing except not be attracted to me. Not that I can blame you. Even I\'ll admit there\'s only so many farts one person should be allowed to emit before they become bio-hazardous. I don\'t even know what the fuck I\'m trying to say here. I guess I just want things to go back to the way they were. Before I felt this, before all of it. I just want that moment in the library back. And I don\'t want to weird you out either. I only hope these last few weeks can be spent as friends not as two people too uncomfortable to even be around each other due to some phantom conflict that shouldn\'t have been there in the first place. For some reason I feel like that Robert Frost poem from the outsiders relates to this.  Nature\'s first green is gold, Her hardest hue to hold. Her early leaf\'s a flower; But only so an hour. Then leaf subsides to leaf. So Eden sank to grief, So dawn goes down to day. Nothing gold can stay.   I feel like our first moments were gold but then something happened whethere it was me or you it really doesn\'t matter. We can\'t ever have it back...nothing gold can stay.  Whatever fuck it.You\'ll probably never ever read this but whatever. I just needed you to know how I felt and that you don\'t need to worry about giving me the wrong signals and all that bullshit Christine fed me. No i\'m not mad at you either so don\'t go getting all...Christine-y. Hey so when\'s the cast party?....heh heh.... yeah bye '),(21335,4173964,'08,August,2004','2004-08-08 00:00:00','   So I got bored and had nothing better to do than post on here. Man do I need a life. I was listening to my Bob Dylan albums a minute ago and \"Forever Young\" came on. I don\'t think enough people realize just how deep this man truly is. He\'s a poet in the truest sense of the word. All you have to do is listen to that song and you\'ll understand. Songs like \"I shall be released\" and \"Shelter from the storm\" are other good examples. I know this is a pointless post but I had to get the word out. Listen to Bob Dylan NOW. Don\'t download it like a cheap little putz, get off your ass and go down to the closest music store and buy as much of his stuff as you can. As much as possible, he\'s like money ya can never have enough. '),(21336,4173964,'08,August,2004','2004-08-08 00:00:00','   Back from work... can\'t believe i\'m saying this but I think i\'m starting to understand what Blink 182 was talking about in that song. You know the one admit it. Had an epiphany whilst filling a car at pump 6... Being depressed about things as trivial as high school crushes or crap like that is just a waste of energy. Shit happens ya gotta just roll with the punches and...throw them \'bows and all that good time jazz. I guess what I\'m saying is I\'ve decided not to get hung up on some girl that I barely even know. Chalk one up for those damn glands right? Oh well. Listening to Bob Marley and really wanting to get drunk. Never gonna happen judging by the current state of my social life. I always get the same thing \"Oh yeah we totally should do something I call ya *insert day/time/date in general here)\" And guess what they never do. Whatever. Fuck it. I\'ll save my liver for pickling some other night. I still feel like a dork and/or a 13 year old girl when I post on here. \"Dear Diary tee hee\" puhleeze. Still it feels good to get some things off my chest. I\'m still working up the courage to talk about my dad. There\'s alot to sift through and I don\'t know if I\'ll like what I see when I get through. Listening to love songs when your single and desperate is really depressing I gotta tell ya. Well I\'ve got nothing else to say tonight so i think I\'ll go play some video games...and then I wonder why I don\'t have a girlfriend... '),(21337,4173964,'08,August,2004','2004-08-08 00:00:00','   Yeah so first post on this thingy WHOOO!! I\'ve been making fun of people who used this for the last bit of while and now i find myself...using it..??? Anyhoo.... yeah this summer has sucked my ass so far. Working at the co-op gas bar is the lamest job ever. Bad pay, bad shifts, lame managers the usual. In sticking with the emo mystique I guess I should be obliged to spill my feelings in some form of lame poem or something so how about this....  Raindrops keep falling on my head Yeah I already screwed this Haiku up (is that how ya spell it??)  I guess I can express how I feel using the immortal words of one 50 (fitty) cent. And I quote \"In this white man\'s worl I\'m similar to the squirrel. Looking for a slut with a nice butt to get a nut.\"  I told the girl of my dreams how I felt about her a few months ago but got shut down like...a guy getting shut down. I\'m just gonna say to her...You say no one wants to be with you but I did and you said no. Think about that next time you mention all that jazz... I just realized how egotistical sounds. Can\'t think of a better way to word it. sorry.  My parents....I\'ll save for another time. That subject deserves a post all to itself.  Anyhoo I\'m off to pump some vehicular orifice. Adios '),(21338,4051770,'29,July,2004','2004-07-29 00:00:00','    Kerry rocked the house tonight. Not only is it the best speech I have ever heard Kerry give, it was a speech that effectively debunked all of the perceived flaws of Kerry.&nbsp; Pessimistic? Pessimism is thinking Americans can\'t do better. Tax raiser? Yeah, on those who earn over $200,000 but cuts for everyone else. Waffler? He took stands on stem cell research, staying in Iraq, and resisting the privatization of health care. He masterfully weaved in his biography to display how \"tough\" he will be as president while reminding voters that HE knows the true cost of war first hand. And the sense of hope--the theme that the best is yet to come--bordered on inspiring.   Up until now I\'ve been hoping the Democrats would have a chance and fell behind Kerry as the choice offered to us. Now, I firmly believe the Dems just might do it--and I am, for the first time, EXCITED to be following Kerry into the battle for the White House.&nbsp; As the U2 song said that played when Kerry was done, \"It\'s a beautiful day!!!!!!!!\"   '),(21339,4051770,'28,July,2004','2004-07-28 00:00:00','    st1\\:*{behavior:url(#default#ieooui) }                        So he&#8217;s not as inspiring as   Clinton  , but I don&#8217;t care. My vote was locked when Edwards stepped on stage after Kerry&#8217;s speech and I could see him mouthing to Kerry, &#8220;You&#8217;re gonna be President.&#8221; 2004 is our year&#8230;this guy&#8217;s going to the White House and I can&#8217;t wait to see what he does once he gets there! I don&#8217;t care if his hair if 2 feet tall.           &nbsp;         Oh&#8230;and by the way&#8230;when Bush and Kerry debate I think Kerry&#8217;s going to wipe the floor with Bush.           &nbsp;         This is awesome&#8230;and I can&#8217;t wait to be in DC to watch it happen.           &nbsp;         Kerry/Edwards 2004!!!!'),(21340,4051770,'28,July,2004','2004-07-28 00:00:00','    Did you know that the Torres family, of whom Hugo is scion, has substantial connections to the House of Froeman?&nbsp; That\'s right Froeman, as in Abe Froeman, the Sausage King of Chicago.&nbsp; What do you think Hugo was thinking about as he read One Fish, Two Fish, Red Fish, Blue Fish on the morning of July 24th, 2004?&nbsp; And later he paid for shipments of Sausages from Sacramento to Chicago.&nbsp; Hmmmm.....   But, besides the House of Saud part, I thought M. Moore did a pretty good job.&nbsp; He tackled too much, to be sure, but I think his goal was to show that the lies, problems, and general inanity of the Bush White House runs far deeper than just one issue.&nbsp; The Iraq stuff was good, as was the Florida recount part.&nbsp; It was just too much of a romp to be cogent, or at least too varied to be a powerful singular message.&nbsp; It was maybe a powerful Cingular message--like how I call MY&nbsp;cell phone plan my \"more money for hookers--that\'s right, coke and hookers\" plan.&nbsp;   Anyway, I digress from the real issue:&nbsp; is Mary Kate Olsen going to start at NYU next year?&nbsp; Or will she still be rexin\' at her $9000 a week treatment facility?&nbsp; I\'ll leave it to Res Ipsa Loquitur\'s own NYU correspondent, Ms. Maggie D. Welles.&nbsp; Mags?&nbsp; What is it?&nbsp; Are the Olsen twins going to be at school next year?&nbsp; I must know.&nbsp; This is the burning issue that should be deciding the election, not some crap about Iraq or health care or terrorism.&nbsp; It\'s all about the Full House reruns.   '),(21341,4051770,'28,July,2004','2004-07-28 00:00:00','    Don\'t get me wrong. I\'m for sure voting democrat this year and every year. But, Michael Moore\'s latest film is trying way too hard, and trying to do too much. He could have easily blown open any one of a number of Bush White House snafu\'s, but instead he tries to blow them all open. I generally like Michael Moore, and I like his politics. But I think he\'s starting to head a little too deep into liberal \"left\" field to find huge conspiracies where there are just giant inconsistencies and stupid decisions. I worry that Moore\'s antics and tactics are getting tired and old. I worry too this his antics will ultimately detract from the overall message that he is trying to convey. He starts to look a little too Mel Gibson in \"Conspiracy Theory\" and less Al Fraken in \"Radio Free America.\" Again...his message is important, but not everything is Oswald in the book depository.  Just my two cents...'),(21342,4051770,'28,July,2004','2004-07-28 00:00:00','    It\'s great that the parties have opened up the conventions to bloggers. One good blog about the Dems convention is at Slate.com,  urlLink http://www.slate.com/id/2104296/   In particular, take a look at&nbsp;some interesting commentary&nbsp;on Steve\'s favorite Illinois candidate, Obama...   '),(21343,4051770,'27,July,2004','2004-07-27 00:00:00','    He\'s just great.&nbsp; He\'s gonna be president in 2016!  But Brian, I think you\'re missing Jason\'s point.&nbsp; I had a client last summer who can vouch for what the Daily Show was reporting.&nbsp; This guy (who was charged with perscription drug fraud) had sued Pacific Bell, and because of that lawsuit, the Clintons sent the Secret Service to steal his documents.&nbsp; This guy had to go into \"exile\" for a while, and when he returned, what happened?&nbsp; the DC cops trumped up this charge to keep him quiet.  I mean, c\'mon, Flock, if you can\'t trust a schitzophrenic who at a pre-trial status conference yells at the judge \"Look your honor, this trial is a sham.&nbsp; I know it.&nbsp; I\'ve FUCKED BARNEY FRANK!\", who can you trust?  Then again, this guy was a big Howard Dean supporter, so his judgment is at least somewhat questionable.   '),(21344,4051770,'27,July,2004','2004-07-27 00:00:00','    So&nbsp;I was at the San Diego Comic-Con this past weekend, the world\'s largest gathering of geeks and fans of all things entertainment related.&nbsp;While there,&nbsp;I attended a Star Wars presentation about the upcoming movie. The title of Episode III: \"Revenge of the Sith\" was met with cheers, the first of the prequel titles that fans seem to approve of. The coincidence of Comic-Con occurring around the same time as the Democratic National Convention has filled this correspondent\'s politcal geek heart with great joy. It has also prompted some thought as to how the Star Wars movies describe not only a galaxy far far away, but also our very own political climate. Think about it:    Episode I: The Phantom Menace    The Republicans harped on and on in the 90s about Bill and Hilary being the devil incarnate and how Monica Gate was ruining the Republic. But the real threat, unrecognized by all, was the bumbling good-natured presidential candidate known as Dubya, whom all though posed no threat...the good news at the end of the movie, a booming economy and an incumbent vice-president who won the popular vote, would soon give way to...    Episode II: Attack of Al-Qaida    A shocking attack on the Twin Towers sends Americans reeling, struggling to find a way to cope with tragedy and recover.&nbsp; Finding the time ripe to unleash his long standing plan, Dubya unleases his clone army...er...reservist army to invade Iraq. The Republic is plunged into a war it did not seek against the mindless droid armies...eh...terrorist armies, who did in fact seek such a war. Meanwhile, the principles of the Republic are undermined behind the&nbsp;scenes by&nbsp;Emperor&nbsp;Cheney and his&nbsp;apprentice, Dubya Vader,&nbsp;who slowly take away the freedoms long cherished by&nbsp;the people. But not all hope is lost...    Episode III: Revenge of the Democrats    The growing Rebellion scores stunning victories, uniting behind Clone War...er...Vietnam War veteran John \"Obi-Wan\"&nbsp;Kerry. The Rebellion is further bolstered by the addition of fresh faced and idealistic John \"Skywalker\" Edwards. Holding their war council meeting known as the \"Democratic National Convention\", the team brings in the legendary Bill \"Qui-Gon Jinn\" Clinton to lead the attack. &nbsp;Which leads to the appropriately named start of the second trilogy that shows George Lucas was more than a visionary film maker but also a politcal prophet...    Episode IV: A New Hope!   '),(21345,4051770,'26,July,2004','2004-07-26 00:00:00','    All right Jason...I know all the booze, drugs, and women have probably gone to your head, but I\'m gravely dissappointed with your un-American and anti- Democrat ic viewpoint. You\'re either with us or against us. It\'s time to root out the sleeper moderates in the democratic party. If you\'re not an ultraliberal I\'m going to have to bunker bust your attitude. We\'re at threat condition orange here people. If we don\'t move now we might have a republican in the White House for another four years. We don\'t have room in our party for division...so fall in line and be liberal. Support gay marriage, hefty social programs, health-care for all, and bigger government or move to another party. Oh...and by the way...don\'t even think about supporting Nader or I\'ll start breaking things. And they say us liberals are weak. Well we can strong arm just like&nbsp;the republicans. Have a hefty dose of fear and paranoia for motivation Jason! =).'),(21346,4051770,'26,July,2004','2004-07-26 00:00:00','    I just watched \"The Daily Show\" last night and was horrified to discover that, according to the founder of the anti-Clinton presidential library,&nbsp;Clinton was involved with literally hundreds of murders.&nbsp; Everyone knows he is secretly working against Kerry&nbsp;so as to position Hillary for a 2008 White House run.&nbsp; Then they can turn the government over to the undead Nazi\'s that Clinton has been working for in secret the entire time.   '),(21347,4051770,'26,July,2004','2004-07-26 00:00:00','    The excerpts from Clinton\'s speech last night that I\'m posting below are the reason I\'m proud to be a democrat! How is it we can\'t get this guy back into office???   \"We Democrats want to build a world and an America of shared responsibilities and shared benefits. We want a world with more global cooperation, where we act alone only when we absolutely have to. We think the role of government should give -- should be to give people the tools and to create the conditions to make the most of their own lives, and we think everybody should have that chance.\"   \"I think we\'re right for two reasons. First of all, America just works better when more people have a chance to live their dreams. And secondly we live in an interdependent world in which we cannot possibly kill, jail or occupy all our potential adversaries. So we have to both fight terror and build a world with more partners and fewer terrorists.\"   '),(21348,4051770,'26,July,2004','2004-07-26 00:00:00','    I\'d like to take a moment to plead for the repeal of the 22nd amendment, as I want Bill Clinton as my president again.  That\'s all for now.   '),(21349,4051770,'26,July,2004','2004-07-26 00:00:00','    Welcome everyone to our new blog. This little venture, dreamed up by Hugo and I is an effort to have a forum where friends can chat and rant. It&#8217;s a group blog&#8230;not a personal blog. So post and comment away. This should be fun!           &nbsp;'),(21350,4051770,'06,August,2004','2004-08-06 00:00:00','    Singer Rick James Dies at 56 By THE ASSOCIATED PRESSPublished: August 6, 2004 Filed at 3:25 p.m. ET LOS ANGELES (AP) -- Funk singer Rick James, best known for the 1981 hit ``Super Freak,\'\' died Friday, apparently of natural causes, police said. James, 56, died at 9:45 a.m. at a residence near Universal City, said Officer Esther Reyes, a Police Department spokeswoman. ``He died apparently of natural causes. We learned of his death after responding to a radio call,\'\' Reyes said.After his big hit, James\' fame began to fade as he became embroiled in legal problems and health troubles. James was convicted in 1993 of assaulting two women. The first attack occurred in 1991 when he restrained and burned a young woman with a hot pipe during a cocaine binge at his house in West Hollywood. He was free on bail when the second assault occurred in 1992 in James\' hotel room. James was sentenced to more than two years in state prison. In 1997, he released a new album, but a year later he suffered a stroke while performing at Denver\'s Mammoth Events Center, derailing a comeback tour. In 1998 he also underwent hip replacement surgery   '),(21351,4051770,'09,August,2004','2004-08-09 00:00:00','    st1\\:*{behavior:url(#default#ieooui) }                          Clinton       was on the Daily show this evening&#8230;and yet once he again he left me wondering what idiot passed the 22 nd  Amendment. He is dead on with his analysis that democrats need to start fighting back. Bush throws the first punch and the democrats just seem to take it. If we come out swinging we can win the election this year.            &nbsp;         By the way&#8230;my favorite line from   Clinton   on the Daily show was a quote from a Republican Senator who said &#8220;Well, if we fought fair you&#8217;d win all the time.&#8221;   Clinton   could mop the floor with Bush Jr.           &nbsp;         If you get a chance to see the interview on a re-run watch it. The best part may have been when Daily mentioned the news that Halliburton had recently been fined 7.5 million dollars by the SEC. His comment was &#8220;Travel-gate&#8230;come on.&#8221; Asking the President, &#8220;Doesn&#8217;t it just make you mad?&#8221;'),(21352,4051770,'09,August,2004','2004-08-09 00:00:00','    He was super freaky and super drugged out, but now he\'s super dead, and that\'s not super at all. Rick James was a legend and I hope he\'s found his way to that pimped out crib in the sky where Biggie, 2 Pac, Grand Master Flash, and Jam Master J are all kickin\' it with their ho\'s from different area codes. Keep it real Rick, keep it real. I\'ll pour a 40 oz. out for you and all the homies. Never slack on your pimpin\'.   '),(21353,4051770,'18,August,2004','2004-08-18 00:00:00','    first of all, Mr. Flock, 2 Pac is not dead.  it\'s all a conspiracy.  don\'t you know???  So i\'m back in the lovely city of Chicago today, and was greeted by high humidity.  makes me miss Seattle already.  And Alan Keyes is running for senate here.  The guy looks like a muppet.  wow.  and he\'s nuts.  good times, good times.   '),(21354,2751885,'30,June,2004','2004-06-30 00:00:00','   Well IM just listening to some old Belle and Sebastian, and IM thinking way too much. Its reminding me of times gone past. And IM getting too old. Funny story, my dad looks at me yesterday and asks me how old I am, and I say 24. He was amazed, he thought I was 21 going on to 22. Pretty funny. My birthday was a bust, I worked all day, went to the nutritionist at 8:30am, and I haven\'t gained anything, I really don\'t get it. Yeah so think I did nothing all night, My friend A gave me a call, he gave me two prints of his (really nice and contemporary) and some underwear. But all was well, Elliott wrote me an email whising me a good b-day, and offering to buy me a drink this weekend. IM trying not to be attracted to him, its hard, ironic don\'t you think. Well thats about it for now.   '),(21355,2751885,'26,June,2004','2004-06-26 00:00:00','   The weekend was indescribable. Friday night was the best. I met my friend Elliott at the bar, we were the only ones there. So he played his guitar for me, and then we played tribond with the bartender. After that we went to Donatella Versace\'s house were there were tons of familar faces, but I knew few names. So me and Elliott were being anti social and just chatting all night. I heart Elliott, I was giving him advice about meeting girls. But secertlly I wished he would ask me back to his room. As well we talked about a lot of things, art, third world countries, and being adopted. I must say it was a good night, I really like getting to know the straight boys. I went home around four thirty. And I did the responsible son thing and took a cab home. I just couldn\'t call my parents that late, even though I know they would have picked me up. Im spoiled like that. But now that i think about it, I should of went to elliotts room to use the phone, and maybe something unexpected would of happened, hmmmm I\'ll never know. Anyway that\'s about it.   '),(21356,2751885,'15,June,2004','2004-06-15 00:00:00','   So I just went through some of my old art work.  I usually do this one every year to remains of times gone by. And I must say I was impressed.  No head swelling here. I was impressed that I actually did all this work, I made these images.  And I think to myself \"where the fuck has this person gone. . . Who was he?\".  I don\'t know.  Lately I\'ve been wrapped up with work and trying to get my mind to land.  Its seems like I never get this low blood sugar thing.  Oh to make matters worse I lost a couple of pounds, grrrrrrrrrrrrrrrrrrrrrrrrr.  I would love to seriously just leave, that\'s it just stop everything and leave. Hmm. . .  Im the type to think way too much. Looking at the art that I produced got me inspired. I keep telling myself that I should rent out the studio. Well I should, I just don\'t know what\'s stopping me from doing it. That and applying to school.   '),(21357,2751885,'14,June,2004','2004-06-14 00:00:00','   I just happen to be fortunate enough to know the lead singer of the Doll Incident.  There this new feminist band out of Toronto.  Straight boys beware, the Doll Incident is coming to a neighborhood near you!! Good job girls.   '),(21358,2751885,'09,June,2004','2004-06-09 00:00:00','   well look who has ablog of their own!! it baddenbadenbaden....... he invited me to join and I had to write something even if it\'s not really anything good...and I\'m not a boy so it\'s more like rambelings of a grrrl...... that\'s it.....   '),(21359,2751885,'09,June,2004','2004-06-09 00:00:00','   hey ya\'ll hows it been? Me? Oh you know how it is. So this is it the big old blog.  Well then. I decided to hop on the blog wagon because it seemed like a nice way to stay in contact with my friend stacy.  Now stacey you can know all the things that are happing in my small life. oh and as well other people are welcome.  so this is it. I don\'t have much to say, well there is this one thing.  I think I find relegious guys to be really cute. What the hell am I talking about? you may ask. Well these two young guys came into the store that i work in and they were wearing white shirts, gray pants and ties, they were two cute, okay maybe to much acne but you know there was somthing very sexy aobut it. I think its because its so forbidden.  Hmm I think it would be so hot to seduce a little religous boy and make out with him in churh te he.  anyway thats my little bit of something for tody, maybe i should join a religous group hmmmm its an idea   '),(21360,2751885,'09,July,2004','2004-07-09 00:00:00','   hey just got back from the bar. and im listening to some sad music. fuxk what happened tonight. well first my hero fell hard, hard i say. hero, ha, boy was i wrong, i cant believe i looked up to omar, so he says he\'s ashamed to think that people might perceve him as gay because he\'s hanging out with me. Im really mad. i mean so what if people think your gay, get the fuck over it, i mean if someone is afraid if there gay, well theres def some issues theres.  Okay on a better note, i met two girls from trinnity bay, they were so cute, one was a peircer and the other was a student, they were the cutestest couple. i love them and thats about it, seeing them cuddling and making out without any cares made me wish i had a b-freind, someone to love, and again i thoght why am ihere. well thats about it, latter for now, ill be more clear once the booze wears off, cheers. the one and only b.   '),(21361,2751885,'30,June,2004','2004-06-30 00:00:00','    I have three blogs that I now have to be clever for...I don\'t think I have that much cleverness to go around. but we\'ll soon find out just how clever clever is!!!  so this band thing is really takeing off eh? Some of our friends are trying to get i on the band action, which is cool! but...you shouldn\'t be like \"hey, alisha said I could be in your band is that cool\"...when in fact alisha said nooooo such thing.like...dude start your own electro-raunch band and stop trying to hone in on our talents for writing great songs!!!!  heh...you know who you are! '),(21362,2751885,'28,June,2004','2004-06-28 00:00:00','    stacey! i thought it would be cool to have a communal blog where we could post together. in fact, we should get other people in on it too. or maybe it can just be for our band and for us. i don\'t know. but communal blogs rock. what do ya say? '),(21363,2751885,'24,July,2004','2004-07-24 00:00:00','    so...we on way back into the \"studio\" (meaning alisha\'s room) to rerecord some of our songs we didn\'t really like the first time around...such as it\'s so hard..and pornado. pornado will be a lot thesame with the same words and the same beat...hopefully, but will just be better. it\'s strange to me that people we know sorta but not really have been checking out our.  we had a guy we sorta know through school/friends (we both know him throught different people) came u[p to us and was like I really liked pornado!..that felt cool  it felt like that for one qucik second in time I was famous......  nah it was actually kinda weird and creepy that someone we didn\'t really know knew who we were. well I was creeped out anyway.  but it\'s back to the note books for me to start on some lyrics for some new songs. we need to get ths party started!!   w00t '),(21364,2751885,'11,July,2004','2004-07-11 00:00:00','    so, we\'ve been doing a lot of  recording ourselves being silly to regular beatas these past few weeks. It\'s been fun and some of it I think we\'re going to put as a hidden song at the end of our first album. we sing silly songs and make fun of each other and our friends. it\'s good times.  that\'s about it really. I think I might go over to alisha\'s house to rerecord some old songs that we don\'t really like that much. for example,...pornado can be sooo much better!!! '),(21365,2751885,'02,July,2004','2004-07-02 00:00:00','    we are gonna be superstars, and this new song is proof. '),(21366,2751885,'02,July,2004','2004-07-02 00:00:00','    NEW SONG!!! NEW SONG!!!  GO TO OUR PAGE AND CHECK THIS SHIT OUT!!!!! no, really go...it\'s our best song ever..you thought we were good before..now we\'re hot!!!!  urlLink CREAMIEST '),(21367,2751885,'29,January,2004','2004-01-29 00:00:00','   So here I am on a thursday night struggling to get the work I had all week to do in one night. I mean I\'m fucked really. I can never get all work done in time. but now here I am typing in my blog. My friend dave told me that I was a professional procrastinator. He\'s right...  And I hate to admit it. I hate being a procrastinator.And knowing it is way worse. B/c I know that I have work to do and I\'m just going to put it off til\' the last minute. I suck really.    I\'m in the mood to date someone casually. now to find someone who takes things as unseriously as I do but with all the drama.. B/c I do love a little drama.  '),(21368,2751885,'27,January,2004','2004-01-27 00:00:00','   so, It\'s fucking crazy out today and it felt like most ppl were \"smart\" enough to take the subway. Well I\'m coming home and the subway is packed like freeze dried apricots, and the train is moving so slow it was like we were going backwards, The Driver comes on the over speaker:  \"attn: all males on the train.......all males.........please get up .......and give you seat to a lady, it\'s going to be a long ride.\"  Everyone on the train was laughing. it was really cool. I mean nobody got up, but I have been on the subway at slower times and have never had that happen to me before. It made me feel so good. And when I was getting off he came on again:  Atten: to all passengers...This is the driver of this train. I just want to apoligize for the delay I know you all want ot go home. And it is taking a long time. I just want to say thank-you for being patient and I\'m really sorry it\'s taking so long.\"  This was the nicest guy ever!! Like holy fuck.....  I wish that ttc guy would read my random internet blog. I would thank him for being so fucking cool. It makes the trip a lot nicer and makes a person feel better!!!!  Man I hate the snow.......  '),(21369,2751885,'26,January,2004','2004-01-26 00:00:00','   I love this new look thus that I can thank alisha this new look. Alisha thank-you. If you to estalar on its anme then you go for the right its place. Which is fulled with all the types of the wonderful material. As it pictures and crap. Thus I sat down around my house all the night and ate the ice cream. I love the ice cream. I all seated and gave to attention the wind and snow and thought of warm thoughts and ate tub of caralho of all! this is right.......... approval, well that I ate very. Perhaps not it entire thing. This would make me sick. Cream swims likes it ice in one day cold, when its inside and the cold is for is. I never want to leave another time. He above turns the heat, all the lights on, he puts on mine sunglasses and suite to bathe itself, and imagines them that I live reallly in some warm place. As Mexico, in some place dosen\'t of the water only gives to me to the deirrhea or the some another one diease.  '),(21370,2751885,'26,January,2004','2004-01-26 00:00:00','   so I can thank  urlLink alisha  for this new look. Thank-you alisha urlLink  . If you click on her anme then you go right to her site. Which is filled with all kinds of wonderful stuff. Like her pictures and crap.   So I sat around my house all night and ate ice cream. I love Ice cream. I sat and watched all the wind and snow and thought warm thoughts and ate the whole fucking tub!!! that\'s right..........  ok, well I ate a lot. Maybe not the whole thing. That would have made me sick.  Nothing like Ice cream on a cold day, when your in and the cold is out.  I never want to go out again. Turn the heat up, the lights all on, put on my sunglasses and bathing suite, and imagine I reallly live someplace warm. Like mexico, only somewhere the water dosen\'t give me deirrhea or some other diease.  '),(21371,2751885,'25,January,2004','2004-01-25 00:00:00','   I just learned how to make titles. I\'m happy now.  '),(21372,2751885,'25,January,2004','2004-01-25 00:00:00','   Last week I came home to find my answering machine no longer spoke to me. It was mute...I was sad. I sat on the floor and had a little cry while cradling it in my arms. Then I called my friend alisha to help sooth my broken heart. After that was over, I went to the mall to replace my old friend.  I got the new machine home...Plugged it in....And...NOTHING. It was a dud. I was upset, but shed no tears, I mean after all we barely knew each other!!!  Today I went to get yet another answering machine. And you\'ll be happy to know it works. So you can call my house and 14 minutes of wonderful message fun.  I think I peed a little...........  So that was my day. The mall was hell as it always is. Anyway....  '),(21373,2751885,'25,January,2004','2004-01-25 00:00:00','   hi, my name is stacey janes. ppl often tell em it sounds fake.....like rubber cheese.  I often wonder if people are looking in my window at me when I sit here on my computer and type in the nude. And then I think if that\'s what they wanna do then maybe I should open my curtains and so they can get the full effect. Or I could look out the window and post a sign to invite them in.   or maybe not. I\'m a little self-conscience about that kinda shit.  '),(21374,2751885,'22,February,2004','2004-02-22 00:00:00','   so here I am at home in newfoundland.yeah it\'s all the fun I remember. I went to a \'rave\' last night.it was more like a HS dance party.sad really, but I must say I had a really good time.I did kiss this boy I had a super crush on when I was in HS. I know that was a long time agos, but it still felt like a small triumph. c\'mon you know what I mean.That whole you didn\'t want me then but I can have anytime I want now. I know it\'s ot really like that but it made me fel really good. he kept telling me I felt nice. I\'m not sure what that was suppose to mean...but it sounds good.It made my knees weak a little. good times.  yeah so the first two days I was here I was totally snowed in, it was a weird experience. I\'m not used to having to limb moutains of snow to rent a video. I really wanted to take a picture of all the snow....then in clicked in ....I grew up here, I should be used to this!!!..I felt really touristy. And that was followed by the need to still take a picture of all this fucking snow!!!  I\'m sitting here watching the christina agulira special. she\'s really dumb but wants ppl to think that she\'s not that dumb. She wrote a song about double standards.......but it\'s really stupid and makes no sense. Like brittany writing a song about what it feels like to be rich and how it\'s really hard!!!! wtf. I hate them both.   just b/c I hate boys dosen\'t mean I\'m a lesbian.......my friend baden read that somewhere.but it applys to me I think.   alisha and amy....I miss you!  '),(21375,2751885,'17,February,2004','2004-02-17 00:00:00','   I mean reallly.................if you were going to get hair transplants, why would get discount ones.If you were indeed going to go through all that troubel why wouldn\'t you just pay full price and know your getting a good job. So you pay half ...then all you hair falls out....what\'s the point?  '),(21376,2751885,'16,February,2004','2004-02-16 00:00:00','   I\'ve been trying to publish my top 100 things about me list and it dosent want to show you.....I hate this stupid blog.......but I love it!!!    the word pubic makes me laugh a little  '),(21377,2751885,'09,February,2004','2004-02-09 00:00:00','   so I went out on saturday with my good friend alisha.We went to the bovine sex club, and this guy started doing all these fucked up card tricks. it was craziness.I was holding the card....and then so was he!!!   ok ok...I was a little drunk it\'s true.But he really did all this crazy stuff and when I asked hi what he did for a living....he was in insurence!!! I was amazed and throughly impressed.  So I am so anti-love at this point. Boys never seem to know what they want and then when you tell them wht you want they always want something different but really want the same thing you want but don\'t want you to know because you may want something more then that!!!!  you got all that!  so after the bovine we went ot the velvet underground. It was last call so I went on up to the bar to get a last beer. I was wearing these cute little bunny ears...it was cute. anyway..there was this guy there and he kept saying things to me that I couldn\'r hear b/c he was all drunk and slurry. And then he gave me all this money..he may or maynot have thought I was hooker in my ears..and he kept trying to lick my neck. It was really gross. and then he asked me if I had a boyfriend and I said I did.....well I had a grrlfriend. so then I thought \"fuck the beer it\'s not worth it really\", and I went to find my friends.  so we are all dancing and what not. and ho appears...the gross creepy guy. he keeps grabbing my arm and pulling me. So I start making out with amy..my close and dear friend..but he starts licking me!!!I think I threw up a little. I had to sit in the corner and haev my friends around me before he would leave me alone. thank you friends.  boys are so fucking stupid. I wish I was a lesbian. I wish it everyday.  '),(21378,2751885,'09,February,2004','2004-02-09 00:00:00','   ok so for one my friend dave asked me if I ever write about him in my blog and I sadi no......he seemed a littl disapointed. So here it is Dave your 15 min. Your in my blog. I think your awsome and I would totally make-out with you even thought you think I\'m kidding when I ask you that repeatedly.  ok so here the real reason for this blog entry. I\'m addicted to 6 feet under. You know that HBO series. I don\'t have cable but DAVE owns the first season, and one day we were hanging out at his house we started to watch it....big mistake. I can\'t get enough and as soon as I start watching it it\'s really hard for me to stop. So everytime I go over there and he\'s like ok we can watch one episode....well it turns into like three.And the sad part is I never get bored. Sweet Jesus.........I could sit and watch it for hours. I could honsetly sit and watch the whole first season right through no breaks. I can\'t fathom having to wait a week to see the next episode!!! 3 hours of it just isn\'t long enough.  I realized yesterday I love jesus jokes. well not jokes about jesus.....but like when ppl say things like \"sweet jesus\" that is some good fucking pie!!!  you know what I mean.  '),(21379,2751885,'08,February,2004','2004-02-08 00:00:00','   yeah so I feel like I haven\'t been up to much so why update the ol\'blog if your life\'s boring as hell. but yet here I am. becca this is for you(b/c I know you read this)  yeah so friday I did the grl thing and sat aroud my house reading cosmo magazine. Man it\'s been awhile for the cosmo. I really forgot how fucking horrible it is. I always had the impression that it was the only decent grrl magazine. But no......  It just has all the same adds and \"dose he love you- here\'s how to tell\" quiz\'s. I was reading one such quiz.....how to tell if you in love. It asked the stupidest questions like if you would cheat and if you got giddy when you saw him and such. I mean c\'mon ladies it\'s not rocket science that if you would cheat on your guy, or if you dread seeing him, then your not in love with him. But way to pay 8 bucks for cosmo to tell you that.  I went to the bovine sex club last night. For the first time, I made add! Man that is the fuckin\' coolest bar on earth. It made me feel cool just to be there.Threr were all kinds of grrls there who had crazy coloured hair...It was Awsome. Of coures there were the old burnout rocker guys that were super gross and just came to pick up young grrls. but on the whole it was a pretty cool fuckin place. And to the bartender with the pink hair....I Love You!!!!!!  oh man!!! there was this guy there who we think may have been famous...(but we\'re not sure.) He was like, really ugly but in that way that you can\'t stop staring at him. I couldn\'t keep my eyes off him. he was intrigiingly hidious. Yeah he probably thought I was into him. he looked like a cross between chunk (from the goonies........hheeyy yyoouuu gguuyyss!!!!!!) and Axcel Ross. It was truly starable.  ahh...good times, good times. I also love if you think I\'m sexy by the revolting coks!!!Man I\'m telling you I\'m addicted to it. I can\'t get enough.  '),(21380,2751885,'04,February,2004','2004-02-04 00:00:00','   yeah so I had my ex call the police on me....why?  well I don\'t really know why. I call him on friday to ask a question (it made sense at the time), and he called the police on monday, becasue aparently I was harassing him!!!!. It was really strange because I hadn\'t talked to him in three months. We didn\'t fight or have anything angry to say to one another.  But I figure he\'s crazy. and that\'s the end of that. weird he called the police on me for nothing..........What a waste of time. I sent him a strongly worded e-mail...and made sure not to put in anything he could use against me in a court of law. But that was the end of it ...........I hope.  '),(21381,2751885,'01,February,2004','2004-02-01 00:00:00','   why are boys so fucking weird all the time? Answer me this.  I don\'t hate them but......I just don\'t get it.  '),(21382,2751885,'31,March,2004','2004-03-31 00:00:00','   so, almost everyday this strange asian womans calls my house. When I pick up the phone she automatically starts talking in some language I don\'t understand. I feel like I\'m interupting her just to tell her she has the wrong number. But she dosen\'t say \"can I speak to..\" or at least I don\'t think she dose? Anyway there is no break for a long time so I alaways have to cut her off mid-sentence to tell her she has not called who she think she called. She also leaves long strange messasges on my answering machine.  If a day goes by and she dosen\'t call....I miss her a little bit.  '),(21383,2751885,'27,March,2004','2004-03-27 00:00:00','   you kow when your a teenager and you dye your hair so much that people often say that your hair is going to fall out? ok....if your someone like me you know what I\'m talking about... well it\'s true. I bleached it and it started to fall out ...then I had to get it al cut off. Now I look kinda like a little boy with pink hair. I like the pink part ...but not the looking like a boy part.  '),(21384,2751885,'27,March,2004','2004-03-27 00:00:00','   ok, so I orderd the BLACK EYES cd from the internet, b/c you can\'t buy it in the store!!! and anyway not only is it the best cd/band ever, but they included all the words in the cd sleeve. I just want to say I love it when bands/singers do that. We all love to sing along, (yes you too!) I just felt like I was getting abonus in the cd.I think everybody should always include the words.ALWAYS. It really is the curtious thing to do  if your fans are going out of there way to buy/order you cd off the internet, then the least you caould do is allow me the pleasure of singiing along. Ok the truth is I would sing along anyway and would listen to the cd enough to learn the words, BUT......but you never know if it\'s the right words and by providing them it saves you the trouble of looking them up. Or in the case of black eyes....looking,not finding and crying in frustation.  I also want to add that black eyes is my new favorite band. They are soooo fucking A-m-a-z-i-n-g, but I want to know who dose all the background screaming, I\'ve e-mailed dischord (label) and asked...but as of yet ther has been no response? and there is no grrl listed with the band...could it be  boy? it sounds like grrl screaming, but I just can\'t be sure. If you know anything about this ...please let my know...I would love you forever and send you a nice card in the mail.  thank-you  '),(21385,2751885,'25,March,2004','2004-03-25 00:00:00','   So I was searching the net..like a loser. And I came across  urlLink this  It\'s really quit funny. Can you imagine sending your friend poo in the mail. Well I would never do that to my friends...but there a re a few people I have in mind...  I mena if you could send someone poo without them know ..would you? I certainly would. Are you fucking kidding me. There wuold be no better way to get back at my ex then to send him and his girlfriend a box of poo!!! that would be fucking great...  And then I got to thinking...who mails this stuff out. Who could yuo possibly pay to wrap, put in a box, and deliver someone else\'s poop?   And then threr was  urlLink this  Who would want to know all about poo? beats me. But I got to say I was interested enought to take a look. It is really fucking gross I gotta say, \'I really love the word \'poop\' but actually have to read about it or getting it in the mail is kinda gross.  If any of you send me poo........ I will hate you and never talk to you again!!!!  '),(21386,2751885,'23,March,2004','2004-03-23 00:00:00','   So I\'ve been finding this past little while I refer everythign to being a train....like \"your riding the ice tea train\" if you are indeed drinking a lot of Ice tea. So it\'s like I\'m on the train train. yeah. I\'m also on the digital camera train. Becca baby has been letting me play with hers, so I just take pictures of all the randome things in our apartment and put them on my computer and then delete them. Or I have been practicing taking pics of myself..putting them on my computer and then..deleting them.I never keep any of these random pictures, but it\'s fun to play. Digital cameras were made by jesus..I swear. Nothing this great could come out a mere human.It had to be jesus.  speaking of jesus...I downloaded a jesus cursor. it\'s a little jesus head. It\'s fucking  jesusilicious!!! I could surf the net and look up whatever I want(porn) and know that he is with me!!!   I think I\'m also on the jesus train.....I talk about him a lot, not in a \" I\'m giving my soul to jesus\" way, but more in a not capitalizing his name jesusilicious way. you get what I\'m trying to say?  not to knock jesus...I\'m sure he was a great guy and all, and he did bring back the whole sandel thing.......birkenstock got nothin\' on jesus....  I just realized....(well not just ) that I am going to hell. Hopefully in a handbasket,a pink one.  '),(21387,2751885,'22,March,2004','2004-03-22 00:00:00','   This is me and becca baby at the fetish party. This is becca baby,me, and kelly..not the best one of kelly but it had to go up there..she\'s so cute mid-blink   this is some grrl who spent the night getting tied up.   This is amy\'s friend rachel..she was the fire thrower...how cool is that!!!   This is me kissing beca...I love her so much..she\'s one of my best friends..   This is kelly and amy getting all ..I\'m not sure what they are getiing..before the bar?   This was the very cute burlesque dancer..   These are some very cute boys we met in the room af unusual things.   bec, kelly and me..before we left the house..   This is some guy becca met, we don\'t know who he is or whet his fetish thing is suppose to be...guy in a old suit with stuff on his eyes???   Again becca baby with cute boy..   Half alligator...half boy....it was really gross looking in reall life..eewwee   this was some old guy he was all dressed up, and he was soo cute!!! he looks all serious in this picture. but he was really nice!   this guy was so weird..he came over and was like \"I just NEED to know what you guys just pinky swore about!!!\" and then he hit on us for a little while..he was totally creepy!!! ew  ok that\'s the fetish gallery.  '),(21388,2751885,'20,March,2004','2004-03-20 00:00:00','   yeah so I hd to delete all those fetish pics. they were really too big for this little blog and since I am kinda slow when it comes to fixing stuf,,and I don\'t have photshop, then it was eaasier to delete them. If you want me to send them yout to you I can..just give me your e-mail adress. I act like ppl I know read this blog..I\'m pretty sure it\'s just alisha and spagetti.lol.  so anyway..that\'s it .  '),(21389,2751885,'20,March,2004','2004-03-20 00:00:00','   So we were all hanging out at amy\'s last ight and the topic of children came up...I mean at our age normal people think about haveing kids..anyway we all agreed that we have been feeling the biological urge to have kids. Why do womyn feel this way? Do men feel this way? Do men get in there mid-twenties and think to themselves. you know i would really like my girlfriend to get pregnant? somehow I doubt it, I\'m not say ing this doesn\'t happen b/c I\'m sure some men feel that way. But I now for me it\'s like this deep seeded feeling in the back of my head and the pit of my ovaries, that I constantly have to retell myself all the reasons I don\'t REALLY want kids. I\'m not responsible enough for a child for one thing..And for fuck sakes that\'s what I do all fucking day long!! And if I ever had kids they would be so fucked up. And I just know I would tell them all these lies to see if they believe me and then laugh about it to myself when they do.......heh...heh  anyway then we were talking about if we ever have kids we would never be able to go anywhere and do anything.Except amy...she has alll these sisters that would take her kid is she wanted to have a night out.So me and alisha deceded that amy should fill all of our biological urges and have some kids that we can play with and hang out with....and then go to our childless home!!! I mean, it would really be more for alisha, she has little to no contact with children on a daily basis. I guess I have the best of both worlds, I get to hang out with kids all day and then get to do my own thing the rest of the time!! that really is a sweet deal. If I could only make my ovaries see how great we got it, then I couild just stop having my period all together b/c it\'s not going to get my anywhere I want to go , if you know what I mean.  Wouldn\'t that be great if you just have a one-on-one with your ovaries \"hey you can stop producing eggs for awhile I really don\'t want to reproduce\" and they would be likr \"no problem\". And just like that it\'s all over...painless.  yeah..when I was a child I really thought there were little men inside my body running the whole operation. I would eat some food the first guy would send down my throat and then to the next stomach guy and he would do his thing and so on..they would all talk about me and you know, do there thing.I would actually picture this little factory in my body and when I would get sick I would thing it\'s b/c I did something to upset one of them.....I was a weird kid.  '),(21390,2751885,'19,March,2004','2004-03-19 00:00:00','   So yeah alisha urlLink   fixed all the comments on my blog,apparently they were under her name or something stupid like that. But now they look super cool. BUT.....All the old comment were erased....Boo that. But she has them and read them to me.Which was nice.   Idecided I want o have a menial job when I grow up. Me and alisha were just talking about all the shit we do, and how at our age pple just don\'t do shit like that anymore...Going out drinking several times a week..Meeting strange magicians at bars..You know the usual. Anyway, I figure if I just have a job where I have no real responsibility then I can still do whatever the fuck I want...Right. I mean I would never have any  money and would really really hate my job. But I would have a lot of fun and I would never have to really grow up and be a serious adult. Being all grown-up seems really boring to me in soo many ways.Staying at home and the hightlight of your week is to rent a video....boo that I say..boo that!!!  The only other job I can think of where that would be ok is Rock Star!!! But I eman really....I don\'t think I have the talent to be a rock star. Unless I could become a famous bongo drum player...That I can do. Ok change of plans...I\'m going to be a rock star!!! woohoo...livin-on-the-bus-drinking-beer-everyday-puking-out-the-window..here I come!!! Are you excited..cause I\'m excited!!  oh yes spaghetti I got your messages........your so cool!!!  '),(21391,2751885,'18,March,2004','2004-03-18 00:00:00','   so ahh I\'m trying to write in this thing like everyday..but I must confess I\'m really not clever or witty enough to write something good everyday.And it brings down my self esteem to reread them and see all the poor spelling.It makes me feel like I should go back to grade three , when I had to stay after class and write out vocabularary, yep. Every word I got wrong I had to write it out 100 times. you\'d think I would be  a better speller as a result..but no. The spelling part of my brain seems to have blocked out that aparently horrid memory. Which is why I\'m such a poor speller. Man who needs a shrink!!!!!!  All the answers to the univeres anre hidden in my brain and I have subconsciensly blocked them out...now we\'ll never know the secret to time travel.  If I could travel back in time I think I would hav taken this semester off and looked for myself somewhere. Who cares where....As long as it was constructive  man sometimes I think I\'m getting stupider....  '),(21392,2751885,'16,March,2004','2004-03-16 00:00:00','   so...I have neglected to do any work all semester and now I\'m fucked. What am I going to do!!!! It\'s not even liek I could just sit home for days and make shit up b/c I need to have everything signed by my school supervisor. Now I have nothing signed and nothing done and it\'s coming to an end and I\'m fucked. JUST FUCKED...and yet here I am writing in my blog when I should be doing work..  so, you know what really gets on my nerves...people who stand in the subway doors during rush hour. you know who you are! You stand in the doorway and refuse to move when people get on or off. I just want you to know that when running for the train and you think yuo can make it in the doors right before they shut...YOU are stopping me from doing so.That\'s right I\'m talking to you, All you door lingerers. Get out of the fucking doorway man! I need to get on that train or I wouldn\'t be running!!!! just move in just a little bit for fuck sakes!!! I know you want to lounge there leaning on the door way all cool like you don\'t care about everyone...WELL I CARE ABOUT YOU!!!! just move in man, that\'a all I\'m asking...just move in for the love of god.  I obviously need to calm down..the train runs like what, every three minutes.   I like the word procrastinate....it\'s a theme with me lately....  '),(21393,2751885,'16,March,2004','2004-03-16 00:00:00','   so I went ot the mall to buy pants, which I already hate doing, and ther was like a million pl there. And I remembered it was spring break!!!! NNNOOOOO  Really have ppl nothing better to do with there 8million kids then bring them to the mall!! they\'re all running around being annoying.........knocking stufff over..smelling bad. I was sitting in the food court eating my lunc and ther was this kid in front of me and he kept turing around and looking at me while I ate. hey man, I don\'t like that!!! and then this kid in le chatau was all \" hey mom...I see a girl with pink hair..\" and then she described EXACTLY what I was wearing..outloud..to the entire store.  for fuck sakes....GET CONTROL OF YOUR CHILDREN!!!!  it was a tramatic experience. and I work with kids everyday..but I really think the mall makes them go all crazy, it\'s like there\'s something in the air.....  '),(21394,2751885,'13,March,2004','2004-03-13 00:00:00','   So I have this crazy assignment to do and I just have not started it yet. Why you ask?? I don\'t even fucking know. I\'ve been sitting around in my towel for two days eating vegan food.   SO my mom bought me this vegan cookbook for x-mas. which is nice. I was trying to look through it yesterday to find something to make. but like what the fuck!!! They want you to have all this crazy stuff.......like \"and then use your electric yogurt maker...\" ok..who has an electric yogurt maker?? And all the recipies call for protein powder!!! kinda gross if you ask me....  so yeah I have to go to school today. And yet again I will be showing up empty handed with no assisignment. I\'ll have to stay up all night t get it done.Then I\'ll be bitchy for days and days. Cause I know everybody loves that!  I love my mom.  '),(21395,2751885,'13,March,2004','2004-03-13 00:00:00','   So I was looking up words in the dictionary for my room mate at Dictionary.com, and I decided to look up the word fuck. So yeah I know it\'s childish but what else have I got to do?  Anyway the results surprised and entertained me.   fuck     (fk) Vulgar Slang  v. fucked, fuckÂ·ing, fucks  v. tr. To have sexual intercourse with.  To take advantage of, betray, or cheat; victimize.  Used in the imperative as a signal of angry dismissal.   v. intr. To engage in sexual intercourse.  To act wastefully or foolishly.  To interfere; meddle. Often used with with.   n.  An act of sexual intercourse.  A partner in sexual intercourse.  A despised person.  Used as an intensive: What the fuck did you do that for?   interj.  Used to express extreme displeasure.  Phrasal Verbs: fuck off Used in the imperative as a signal of angry dismissal.  To spend time idly.  To masturbate.  fuck over To treat unfairly; take advantage of. fuck up To make a mistake; bungle something.  To act carelessly, foolishly, or incorrectly.  To cause to be intoxicated.    -------------------------------------------------------------------------------- [Middle English, attested in pseudo-Latin fuccant, (they) fuck, deciphered from gxddbov.]  Word History: The obscenity fuck is a very old word and has been considered shocking from the first, though it is seen in print much more often now than in the past. Its first known occurrence, in code because of its unacceptability, is in a poem composed in a mixture of Latin and English sometime before 1500. The poem, which satirizes the Carmelite friars of Cambridge,  The whole thus reads in translation: â€œThey are not in heaven because they fuck wives of Ely [a town near Cambridge].â€  '),(21396,2751885,'11,March,2004','2004-03-11 00:00:00','   pink haitr was my dream...now it\'s a reality. I officially have pink hair. And I joined the gym. Soon I\'ll be hot mama.....not that I\'m not already hot(cause I am) but I\'ll be even hotter....my milkshake brings all the boys to the yard.....  by mlkshake I mean my BODY!!!  oh yeah.  I want to date a boy that\'s like me. none of this metro-toocoolforschool-insensitive-trying to be sensitive crap. I want someone real...who will talk to me and tell me what the fuck is really going on. I\'m not good at guessing shit like that. you know? or aybe not.  So if elvis was asian..and 5feet tall...and had grey hair....then I seen him on the subway. He was sitting across from me and I couldn\'t stop staring at him. I know it was mean...but dude he WAS elvis. it was more then freaky.   So  urlLink alisha  has kindly fixed up my blog for me, seeing I\'m too stupid to do it for myself. So cudos to her!!! you know I love you pussy girl!!!  '),(21397,2751885,'10,March,2004','2004-03-10 00:00:00','   I was on the subway and I had ths great idea for a blog entry but now that I sit here I totally forget what it was.well then...I guess this is it.  I went to McDonalds for lunch today, and ther was this guy there who was buying his dog a hamburger as a treat. His dog was waiting outside.That was all he ordered, I thought that was really cute. to take your dog for a walk and then buy him some lunch. it\'s kinda like your kid...or a bad date.  I really want pink hair......red is nice but pink..that\'s hot!!!  And pink would match my new tattoo.  '),(21398,2751885,'08,March,2004','2004-03-08 00:00:00','   so I ride the subway like a million times a day. And It\'s like a second home to me now. I have seen some weird shit on the subway...and I\'ve only lived here for two years. so anyway, today...I saw hitler. It reminded me of \'The Andy richter Show\' do you remember that show? well if not then you missed out b/c it really was the funniest fucking show on t.v.....so anyway there was this one episode where andy\'s friend was dating twins but she didn\'t know it, and then andy was telling all their friends and they were like \"andy is this like the time you thought hitler was the muffincart guy?\" and then it flashed back to him running down the hall at the (hitler looking) muffin cart guy yelling \" HHIIITTTLLLEEERRR NNNOOO!!!\" and then andy was like \" what you think when hitler comes back he\'s just going to march in wearing his uniform?\" and then his friend was like \"WHEN hitler comes back? and if hitler were alive today he would be like 124(or some ridulous age)\" good times.....  yeah so I seen this guy on the subway and I thought of \'The Andy Richter Show\' and what would happen If I ran over and tackled him. what the subway ppl would do....?  I really don\'t think they would do much more then stare. but that\'s me.  I got a tattoo in my ear. it\'s a pink star. and it\'s cute as helll. I\'m in love with it.  '),(21399,2751885,'06,March,2004','2004-03-06 00:00:00','   Yeah so it was my birthday yesterday. I realized that I\'m no longer going to have birthdays. I just want to forget that the day I was born is significant in any way. I mean really. Nothing good ever comes out of my birthday, they are usually really sucks, I get older and rarely get cool stuff!! (except for alisha\'s cool \'book of ideas\'). I mean my birthday is pointless. And for now on I\'m just going to stay in bed on my birthday. I\'m just sooo glad it\'s over and I can get on with the rest of my life, which by having a birthday, I realized is slowly coming to an end.  So besides that.......I really like ice cream. I don\'t just mean I LIKE it....I really really love it. I could eat ice cream in the middle of winter sitting in a snow storm. I just like it that much. I can never get enough ice cream.   I feel the same way about rice.  yeah....So, I just want to go to sleep since I didn\'t sleep last night, I stayed up to watch infomercial about weight loss, like that ever works. I sit in front of the T.V. thinking I deserve to be that skinny. But yet I have no motivation to actually lose weight.  '),(21400,2751885,'03,March,2004','2004-03-03 00:00:00','   I hate to fly. not the flying part but al the other ppl who fly along with me. They always seem to want to talk to me. And I hate that. I always have to pretend that I\'m asleep or really into my book.I know it seems bitchy but there i only so much smal talk you can make before you run out of things to say and it gets all awkward and weird and then you feel like you need to keep talking but neither person wants to. At the same time you don\'t want them to feel that your ignoring them but that\'s all you really want.   I mean who really finds true love on the airplane. And if you do is it really love??   you should not be able to recline your seats on the plane. B/c I already have like an inch of personal space.....and when the bitch in front of me reclines her seat her head is now in my personal space. Sitting for two hours on a plan looking at the top of someone else\'s head is not fun. I just want to say that right now. And to all the ppl who recline their seats with no thought of the person behind them........I hate you.  let\'s face it, I hate everybody who dosen\'t give my candy, or other edible goods.  '),(21401,2751885,'25,April,2004','2004-04-25 00:00:00','   Well then..Hello there!!!! So spring has sprung...heh. I first want ed to address the complaints about my poor spelling. Did you not read the post about the really mean grade school teacher!!!!!! Well go back and read it !!!. I\'m trying to be better..I am now able to use spell check ever since I got cable internet. It no longer keeps me from changing my template or using spellcheck!! woohoo to cable.  I don\'t really have anything funny or witty to say as usual (haha) I just felt that I have not posted in some time and I needed to do so. so here\'s my post..POST!!!! Post..Post...Post!!!  done..Thanks.  '),(21402,2751885,'22,April,2004','2004-04-22 00:00:00','   I keep writing titles even thought I know they are not going to show up on my blog anyway!!! Silly grrl. So anyway I was searching around soulseek for new bands as I always do late at night with nothing else to do.. And I came across a very old band that I love.  the beatles!!! Man, I remember driving in the car with my dad, listening to the beatles singing along. It was one of those things where I just always knew the words. I don\'t ever remember trying to learn the words to \"I wanna hold you hand\", or \"black bird\", it\'s like this instinct that I just always knew the words to all the beatles songs. Music constantly keeps me in awe... how soo many people can write so many different songs...You\'d think they would have all been done by now. And then a band like black eyes comes along and opens everything all up again. Is that really cheesy?? I think it may be a little cheesy.  music is just ....Music! It\'s just the way it is. People try to analysis it everyday and it never works out b/c it\'s just....Music. It speaks to some other part of your body...It just dose. It just never need explanation.......Like toes. You know they are there for a reason, without them you would fall over all the time. They somehow keep your balance. But you never think about why they are the way they are...Ugly. There just there...Your toes. Maybe that was a bad comparison......  I don\'t know..It\'s late and I\'m tired. And I don\'t want to go to bed in case Amy comes by....Maybe I\'ll just sleep next to the door in case she buzzes up....*yawn* I can feel myself becoming nonsensical.....  '),(21403,2751885,'19,April,2004','2004-04-19 00:00:00','   ok, the title has nothing to do with this post. I just really like to write the word fuck. I like the way it looks and the way it sounds when you read it out loud.  So as you all see I fixed my blog up really good. I got this template fro maystar designs(the link is at the bottom) I\'m going to ask alisha how to change the colour, she told me it was just a numbers thing. Speaking of numbers...They area  really great band!!!\"insomnia!insomnia!....\"  So I thought I had a virus in my computer But then it went away..What dose that mean? My computer was all turning off and then turning on again... And It would do this all night unless I unplugged it. But then ..I awoke one day and it just stopped. I turn it on and it will stay on till I turn it off. And vice versa. I love it. I mean even though it\'s suppose to do that.  So anyway I havn\'t reallly done much in the past little while. I have mostly hung out with alisha and her new bo....who, I may add, I totally love! they are so fucking cute together it makes me sick! well not really I\'m really happy for her but it makes me want to be with someone too!!! and as we all know you never get what you want when you want it, so I\'m not looking forward to this game of looking for someone to date me. B/c for one:..boys never want to date me..and two: when a boy dose want to date me , I get all mean.. and do things to push him away until he leaves me alone. I don\'t know why, it\'s like this test ot see if he really likes me. yeha I know.. that\'s really fucked up and makes me seem crazy.and now we all know why I am alone. I say that I choose to be this way..and for awhile I really did chose to be this way. but it seems lately everyone has someone to be with and I gettnig left home more and more. Or becoming a third wheel.  Fuck you and your springtime coupleness(is that a word?)  '),(21404,2751885,'18,April,2004','2004-04-18 00:00:00','   ok... somehow..my super cool template got erased?? I don\'t know how b/c I never even go in there and even look at those things let alone delete anything!!! I\'m really mad. I had everything just the way I wanted it and now it\'s all normale and plain. Who wants to look at that?? hmm...hmmm?? nobody..that\'s who!  '),(21405,2751885,'17,April,2004','2004-04-17 00:00:00','   ok, so if you read alisha\'s blog you are aware hat we are thinking of  starting a band. All we need are instrumaents and a little bit of talent!!! no seriously. so here are some band names we have thought about: -filthy panties -honey slut -fresh panties -soyalicious -cuntalicious -pink panties -glamour cunt - CUNT/BITCH/SLUT/PANTIES/DOLL/PINK(ok...this one was kinda a joke) -president juice(kinda boy name though) - Chai fetus (ok we went for chai lattes and I was adding cream to mine and when I put the stick in to stir it there was this big giat thing in it...it was this enormous tea bag!!! and it looked kinda like a meatball or a chai fetus) -alisha\'s knees.  ok so those are just a few of the names we came up with. we were thinking filty panties b/c then we could be really poppy sounding but have really feministy and rauncchy lyrics....\"let\'s wipe our cum on my parents bed\"- Bikini kill.  anyway...what\'s your opinion???  '),(21406,2751885,'15,April,2004','2004-04-15 00:00:00','   Ok, so... We got tickets to see the pixies!!! whoa. What the fuck ...the pixies, the band that is responsible for bands such as nirvana, weezer, etc.. Really the list is endless. And e are going to see them live!! It\'s not for like another seven months or something, but knowing I have tickets is good enough for me!!! So I bought an extra ticket. It\'ll go to the highest bidder!!! While we were in line the grrls I met there told me someone bought a ticket to the winnapegshow for 399US dollars!!! So I holding on to it to see what\'s going to happen. Maybe I\'ll sell it on eBay.....Maybe I\'ll have a raffle........ nah. I\'m going to sell it on E-bay. Unless one of you wants to give me $399Us...Then the ticket is all yours!! haha just kidding. I doubt I\'ll get that much for it. But I only paid 45, so even if I get a hundred it\'ll be worth it.   So ok, ticketmaster thought it would be a good Idea to sell pixies tickets as well as Christina agularia tickets at the same time. So there was a line of about five pple...I kid you not.. And only one guy was the for CA tickets. They were giving out wrist bands to draw for the person who was going to get tickets first. So me and the other three ppl there to get pixies made a deal that if any of us got picked we would buy all the tickets. Makes sense right?(they did sel out in Winnipeg in less then two minutes) So I say to the guy... If you get picked do you mind if we go first b/c if we wait we might not get any tickets at all. And he was all like no, I want good seats to CA.....  ok...WHAT THE FUCK IS THAT!!! Dude...You know your going to get tickets...No matter what! But if we wait we might not get tickets!! That is like pure selfish ness right there. Stupid christine agulaira fans!! but alas..Good wins over evil and he did not get picked!! So the fine ladies I met in line bought tickets on my behalf so me and two of my companions could enjoy the pixies. Thank heavens for nice grrls who love the pixies!!! THANK_YOU!!!!!  GIANTIC...GIGANTIC.....GIGANTIC...A BIG BIG LOVE!!!!  '),(21407,2751885,'13,April,2004','2004-04-13 00:00:00','   ok, I got my very first cell phone yesterday.Yes, now I\'m available at all times, people can reach me where ever I go. Unless I go underground to hde away from the world.....or take the subway. But the super cool thing is...I have \"here comes your man\" by the pixies , as my ring. And those of you who know me.. know that I ahev a special place in my heart for the pixies. I also downloaded \"should I stay or should I go\" and \"tainted love\" three very good god choices in my opinion.  So I will now be the grrl dancing whenever my cell phone rings..... do do doooo do do do do here comes you man do do do do dooooo dooo do.... woohoo  '),(21408,2751885,'12,April,2004','2004-04-12 00:00:00','   If you just can\'t fiond it in the websters dictionary...you may just find it  urlLink here  I mean I foudn the definition to . urlLink gutter slut  '),(21409,2751885,'10,April,2004','2004-04-10 00:00:00','   Peopel who say sex dosen\'t change things is wrong. not that I hear many people say that...but you know. For a long time I had this dilusion that I could fool around with certain friends and not have it interfer with the friendship. It was really the ultimate dream, to have a friend that you completely loved as your friend and you did friend stuff, you joked around with each other, had you own seperate lives where you did whatever you wanted without answering to the other friend, and still have sex nad what not. but this dose not exsist. I have finally accepted that reality ...you just can\'t have it all. there is no way.Someone always gets weird. You can say \"it\'s just sex\" all you want but without good communication...all fails. Someone gets weird about the whole thing and the other person is left thinking...what the fuck just happened here? you lose a good friend and you lose sex. it\'s just not worth it!!  so the moral of this story is.....if your going to have casual sex..do it with someone you never want to see again, or someone you care very little about. Friends + sex = no good.  '),(21410,2751885,'10,April,2004','2004-04-10 00:00:00','   Now I can play dress up jesus!!!w00t     with his wonderful wig and beard set, or for a more casual look,   just the wig.   I gotta tell you...these creepy heads are more scary then the fact that someone out there spealizes in jesus wig sets. But if your looking you can find them here at the  urlLink wig outlet.  Again serving all your jesus wig needs.   Ok....that was my LAST jesuslicious post...I swear to jesus. I am sooo going to hell.  '),(21411,2751885,'09,April,2004','2004-04-09 00:00:00','   I got this picture from my friend erik...hi erik!!!  '),(21412,2751885,'09,April,2004','2004-04-09 00:00:00','   ok...I just made a t-shirt that says that!! w00t! and I found............... urlLink dress-up jesus  YEs that\'s right...the website we\'ve all been waiting for. Where you get to dress up jesus in an array of different outfits according to the season. So go, have fun with your jesus today!  '),(21413,2751885,'09,April,2004','2004-04-09 00:00:00','   man I got the most kickass shoes today. not only are they kickass....but they were only $25!!!!!!! Me and beccca baby went to this really cool shoe shop down at younge st. BUT...ths biggest sized they carried was 81/2.......what the fuck is that? I mean really...most people have larger feet then that. I just felt really lucky I haev average 8 1/2 sized feet.  '),(21414,2751885,'07,April,2004','2004-04-07 00:00:00','   urlLink this  is just really really weird...and really ridiculous.  '),(21415,2751885,'05,April,2004','2004-04-05 00:00:00','   And then I found  urlLink this  when looking up info about said item...   Dear Mrs. Mills,   I have worn a monocle on and off for the past few years. I now feel totally comfortable   and indeed confident with it. However, I am anxiouis to know the correct etiquette, if   any, that goes with the wearing of an eyepiece. I do entertain on a regular basis --  luncheons, dinners and so on -- and I\'m desperate to ascertain the correct procedrues   when receiving dignitaries and, of course, female company.   DPC, Lincs   Dear DPC,   While the monocle should be worn whenever needed for seeing clearly (ie, reading,   shooting, neurosurgery), it is also your responsibility to maintain its silly-ass yet,   paradoxically, rakish image. So, for instance, it should be worn when eating soup so   that you can exclaim \"Gosh\" (or \"Crikey\" in extreme duress), put on a surprised   expression and allow the monocle to fall into the bowl -- a minor coup de thÃ©Ã¢tre that   will give you the moral high ground. On the other hand, when being introduced to a   lady wearing a low-cut dress, never fail to screw the monocle tightly into your eye   socket and draw! \"Hellloooo\" while examining her cleavage closely. Extra kudos is   obtained by sounding like Terry-Thomas.   Mrs. Mills  '),(21416,2751885,'05,April,2004','2004-04-05 00:00:00','   So I went to the eye DR. Last week to get the perscription in my glasses updated. I have a severe stigmatism in one my eyes and when the Dr looked at it he was all \" wow that\'s really bad\" and then he started to tell me what was wrong with my eye. I have had the same problem all of my life..........but I think it made him feel good to tell me...so I let him. anyway.... later that day I went to go get contact lenses (my first pair in nine years!!!) and the guy told my my right eye was fine  and that I only need one contact lens. If I had my time back I would have got a monocle. Alisha was talking about monocles on her blog and me and alisha already had this conversation. But I could actually get a monocle!!! I only have bad sight in one eye. I am only going to need to wear ONE contact lens! Man, how cool would that be....When you want people tom know your really really serious..you just take of your monocle and tilt your head a little to the left.....Powerful!  and then I can go out and start solving mysteries!!! or become a strange butler of some kind!!!haveing a monocle calls for a whole new career!!! I could just sit around looking prestigious!  So when I go back to the DR. I\'m going to inquire about getting a monocle......I\'m curious to see if they still make them and how much they cost.....I\'ll let you know the skinny as soon as I find out..  '),(21417,2751885,'05,April,2004','2004-04-05 00:00:00','   yeah so I haven\'t been updating as much as I would like to. But whatever nobody really reads this anyway......at least I don\'t think so....  Fudge urlLink   ....That\'s what I\'m doing today, making some fudge to amke myself feel better. Loooks good dosen\'t it? you want some don\'t you? well your welcome to come over and have some....mostly alisha b/c she\'s the only one who lives in this area who reads thid as far as I know. but if anyone else out there would like some fudge you can e-mail me....... urlLink  '),(21418,2751885,'02,April,2004','2004-04-02 00:00:00','   urlLink  I adopted a cute lil\' baby jesus fetus from Fetusmart! Hooray fetus!     so I adopted this fetus from fetus mart. It\'s baby jesus, and you know how I love jesus!!! Anyway I\'m not sure what it dose or if it will grow or what. but it sure is funny !!! I got this link from alisha (link to your right). She always knows where to find the cool weird shit. YOUR THE BEST PUSSY GRRL!!!!  '),(21419,2751885,'01,April,2004','2004-04-01 00:00:00','   So I was on the subway (surprise) and the woman next to me fell asleep on me. Yes, you always think \"that\'ll never happen to me\" and then...it happens to you. What is the priopoer subway ediquitte in that situation? do you wake her up and be like \"hey lady your drooling on me\" ,give her a pillow?, subtly lean into her so her head falls on the person on the other side of her?... I really didn\'t know what to do..so I sat there until it was my stop and then just got off. I\'m assuming she woke up and was like  where the hell am I? you know how it is.  '),(21420,2751885,'31,May,2004','2004-05-31 00:00:00','   so I just got a new kitten!!! I was going to name him Mr. ridiculous...but I figured he\'d get teased enough for having a pink coller with spikes(*wink), so I gave him a very many name instead...darwin. ok so it\'s not all that manly but sure is a nice name!!!! I love him soo much i could just ssssqqqqqquuueeezzzeeee him. But that\'s not really moral and I could get charged with abuse. imagine that, getting put in jail for loveing someone the wrong way. Well I guess if that way kilss or harms someone that your fucked upa dn shoudl be put in jail!!!  ok, enough of that.  bu seriously, darwin is the cutest cat ever made! there is only one thing about him I don\'t like...he\'s sooo needy!! he need my constant attention, if he\'s asleep and I leave the room he gets uo and follows me then goes to sleep whereeve it is that I am. I knwo it\'s cute. he walks right under my feet and I\'m always kicking him by accident...it\'s really annoying, I always feel soo bad to see his little tiny body go flying across the floor and get right up only to walk between my feet the next time I take a step...and the cycle continuse.  anyway It\'s late ad I need some sleep ...don\'t you worry I haev soo many cat stories....  '),(21421,2751885,'28,May,2004','2004-05-28 00:00:00','   somehow my template keeps getting erased....does anyone know why this keeps happening to me!!! it really fucking frustrating!!! and I want it to stop. it takes a long fucking time to cstomize this shit.....so if it\'s you...STOP!!!!  '),(21422,2751885,'27,May,2004','2004-05-27 00:00:00','   So the saga of the crazy cat lady continues....  so, ok, we have this door that leads out onto the roof, and other people have windows onto the rook and the way it is our \"neighbor\" in the next building has been feeding stray cats out there window. So, all these cats and a door leading into our apartment only spells bad news. Our apartment smells like pee...Cat pee! So we complained to our landlord about this cat issue, and I guess he complained to the pple next door to us. He also took away all the food and cleaned up the place not to mention told these pple not to feed the cats out there.   but...  the next day this crazy guy comes out the window and threatened the life of my roommate!!! I kid you not...Over cats he doesn\'t even own!! I know, craziness, then when the food did not return he dumped cat litter by our door!!It\'s really fucking gross I\'ll tell you that much.  so I called the police, and now we wait........dum dum ddduuuummmm  '),(21423,2751885,'23,May,2004','2004-05-23 00:00:00','   soooo. how\'s it goin???...I don\'t really have much to talk about but I haevn\'t writen in here in about a week and I think that\'s a little unfair.So I went to go see my friend luis\'s band play the other day. They were really good!! it was in his highschool...yes that\'s right I went back to high schoool...It was way different then I remenber highschool being. I\'m not sure if it was b/c it was like, six years ago or b/c I went to high school in newfoundland....  but I got harrassed by  teachers and it was a pain in the ass. we were told that if we left we couldn\'t come back in and all this other shit. I was kinda irratated by the whole thing...I\'m a 24 year old woman and if I want to go out for a smoke..then I should be allowed to do so. I understand the whole, pple with guns might try to break in here, but I\'m a adult and I should be able to go out of a highschool and come back in....  and also most of the time when kids go into schools with guns...they already go to school there. they are already mad at the way they\'re being treated in school.....  but really they were treated these kids like they were all fucking morons..they didn\'t know any of the kids names and they were hearding them around like sheep...it was really awful, very disorganized and out of control in everyway...  man, I don\'t miss highschool in any way what sooo ever....  '),(21424,2751885,'17,May,2004','2004-05-17 00:00:00','   So, Someone new has moved into our house. I don\'t know how he(orshe) got in but I would like him to pack his stuff and move out. Not that I\'mn afraid of mice...but I don\'t want them crawling on me while I\'m asleep. I gotta say he IS pretty cute.   So I called my friend chris last night to help with the whole mousey mouse situation and he was all like \"why are you freaking out? What\'s the big deal?\" and then he gave me some ideas on how to get rid of my mousey mouse friend,:1) get a bucket, make a ramp and put food on the ramp and trap it....this is a bad Idea b/c I saw the mousey mouse climb a wall..2) leave a trail of food to the door and leave the door open...this is a bad idea b/c I have some really fucking creepy neighbors and I do not want them in my apartment, or I don\'t want to let the mousey mouse in someone else\'s apartment.   I don\'t want to kill the mousey mouse. I just would like him to find a new place of residence. I want to catch him and set him free in the park or in the subway, mice seem to like the subway. OR, catch him, put him in a cage, feed him, love him, call him fabio and he\'ll be mine forever.........yeah, not so much the last one.   any ideas about catching my mousey mouse friend in a humaine way?     NO KILLING!!!!those traps scare the crap outta me...  '),(21425,2751885,'15,May,2004','2004-05-15 00:00:00','   cassandra smells like poo....nah she dosen\'t really. but she\'s sitting here so I have to say that. that\'s it.that\'s all I had to say.  '),(21426,2751885,'14,May,2004','2004-05-14 00:00:00','   so. I\'m having one of those months where I would just do anything to go to sleep before 5am, and then I have been forcing myself to stay in bed past 8:30. It\'s hell. I feel like  my body is all hopped up on something and I seem to have forgotten what it was!!! It\'s like if I get more then fours hours of sleep the night before then my body goes all nuts and is thinking...Well that\'s enough sleep for one week, don\'t you think?  NO I don\'t think!!! I just want to be able to go sleep like a normal person again. If there is such a thing as a normal person with normal sleeping patterns.  sleep,    If your out there and your reading this...I MISS YOU!!Please come back. I love you I really really do. I know I have taken you for granted in the past..but I promise I\'ll change!! I  '),(21427,2751885,'11,May,2004','2004-05-11 00:00:00','   I\'m so hot..I feel like my skin is melting off.   please god kill me now.  '),(21428,2751885,'11,May,2004','2004-05-11 00:00:00','   Hey there ho there........ so I went to he bar on Friday and when I went o grab my bag to pay..somefucker stole my fucking purse. So I assumed at first that it fell off the back of my chair so I was asking all the people around me if they seen it, if it was under their table or whatever..but then this chick stood up and described my purse and told me that she seen someone walk out with it. Why you wouldn\'t you say...\"hey that guy may have stolen your purse\", is beyond me. so later that night the bar called my house and was like we were cleaning up and we found your purse..WooHoo. but they did take my wallet. so I figure the grrl who stood up and was all, I seen that guy take your bag..Well she took my bag. What a bitch!!!  Lucky for me my wallet was all she fucking took, my cell and shit were all still in there!! I was freaking out..I filed a police report and all that other pain in the ass stuff. Only to call back and cancel it. I felt like a moron.   yeah so that was my weekend...  and a cute boy kissed me..It doesn\'t get any better then that.I found my bag and I was kissed!!!!! THE MAGIC EIGHT BALL SAID IT WAS TO BE SO, and as we all know ..the eight ball is never ever wrong.  '),(21429,2751885,'08,May,2004','2004-05-08 00:00:00','   so...tonight. The unicorns and lerhosen lucile are playing at the . urlLink drake hotel !!! I\'m soo excited seeing as I love both those bands. And it\'s only 10 dollars, and ...it\'s like this thing called \"connect the dots\". It\'a a bunch of canadian artists , photographers, musicians..all doing one show. It\'s going to be really fun..you should come!  '),(21430,2751885,'07,May,2004','2004-05-07 00:00:00','   Ok...So a few month ago my room mate..becca baby went to Australia for a few weeks and while she was there she made a whole new discovery...They put fun facts on the wrappers of pads. Bathroom boredom has finally come to an end!!! They call them \"Odd Spots\" ..Get it!! here are some examples of such \"odd spots\"..  Odd Spot #122:  Arachibutyrophobia is a fear of peanut butter sticking to the roof of you mouth.  Odd spot #70: There are three golf balls sitting on the moon  Odd Spot#80: The youngest pope was 11 years old  Odd Spot#129: From the age 30 humans gradually shrink  Odd Spot #42: An iguana can stay under water for about 28 minutes  See you learn something new everyday..Well in Australia they want to make sure girls learn something at least once a month!  '),(21431,2751885,'06,May,2004','2004-05-06 00:00:00','   hi, sorry about the lack of posts. somehow my template keeps getting erased, and I find this very frustratinig. it\'s kinda turned me off blogger for a bit. And I don\'t have the time right now to fuck around with it. I\'ve just D/L the second and third season of six feet under and I\'ve been busy watching them in between work. so ....when I get things together....there will be more posts. I promise.  '),(21432,2751885,'30,June,2004','2004-06-30 00:00:00','   why yes nerina..I would fucking love for you to show me how to add links to my blogola. I\'m soo blog-illiterate I make myself want to throw up a little.   So I started writing a new song today. I must say...it\'s gona be a keeper..!! I\'m giving myself a pat on the back right now!!!  OMG, I locked my cat in the fride for like 20 min on the weekend!! it was an accident....he always crawls in there when I take syuff out and this one time I was doing a bunch of stuff and didn\'t realize what I had done till I sat done and heard him crying from the fridge...hahaha. when he came out he was all cold and sad!!! but he\'s ok now.     I\'m a bad mommy.  '),(21433,2751885,'29,June,2004','2004-06-29 00:00:00','   so toronto had it\'s gay pride weekend last weekend. It was all out gaytastic!! woohoo. we did a running comentary of the events that we particapated in....here is a small sample(keep inmind we did this almost evey hour all weekend)  THURSDAY: 6:55pm- Theresa rears ends a car on the danforth on the way to becca and stacey\'s house  7:05pm - Becca and theresa smoke a joint  7:40 pm - took a cab to buddies and bad times to watch \"tits up\" fearless feminist porn  8:20 arrive at buddies  8:22 some randome girl falls on her ass in the rain and we try to not laugh  FRIDAY:  6pm - stacey quits her job  9:30pm becca gives notice to her job  9:55pm seen creepy boys on then subway, they all had really ugly feet and as theresa was getting off one of them was like \"after you hottie\" ..........ewweee  10:00pm seen hot girl on the bus and  followed her to the vatican   10:05pm arrive a \"cheap queers\" it was this gay comedy night but most of the performers we have seen the night before....it was not so funny  ..........and so it goes on and on..  yeah pride was great we had a really good time and I got a tattoo, Evan did it for me and he designed it for me right on the spot. I know your thinking why would you get somethign you\'ve only seen for two minutes tattoo on your body for the rest of your life...well..  it\'s super fucking cool. and it\'s absolutly beautiful!! I will hopefully have a picture soon so you can all expericen the beauty I get to experience everyday for the rest of my life!!!.....whatever.  '),(21434,2751885,'25,June,2004','2004-06-25 00:00:00','   so it\'s pride weekend in toronto!! woohoo. lots of hot gay and lesbian action goin on!!! not that I\'ll be in on any of that..but I fully supprt those that will be particapating in the festivities!! we went to a feministy-porn poetry thing last night and tonight we\'re goning to some other lesbian thing....  theresa is here so it\'s dyke action all over the place!!!!  '),(21435,2751885,'21,June,2004','2004-06-21 00:00:00','   Amajohn said : I\'m bored, and I remember a drawn out conversation me and The Shaun had a few years ago about a diner with celebrity-themed foods. I can remember some of the ones we came up with, and have some new ones - contribute, dammit!   Robert Donairo  Al Cappuchino  Eugene UnLevyed Bread  Clair Danish  James Van Der Beer  Martin Scorsausage  Jean Claude Van Ham  David Lunch Special  Halle Berries  Pretzel Washington   And some that don\'t need to be changed:  Iggy Pop  John Candy  Kevin Bacon    Trevor said : cant forget these either:  Chocolate Mussolini  Fidel Fadle  Lenin Meringue Pie  oh and not a dessert but how about,  Pol Pot Pie.  I know there is more but i cant think of them right now.    Trvor said Message: i got some more:   -millie vanilla milkshake  -samuel l. flapjackson  -david cronenburger  -johnny wonka (sorry john i had to)   well thats it for now. im getting fuckin hungry.    Anyway, if you guys read this I hope you don\'t mind that I posted this on my blog...but it\'s fucking hi-larious. if you wanna add your fav famous foods go  urlLink here.  '),(21436,2751885,'19,June,2004','2004-06-19 00:00:00','   he there. I started my nanny job today. It\'s great, I really love it! One of the kids is kinda heavy and I put my back out. I feel like an old fucking woman,, al hobbleing around!! But the good thing is..there is a starbucks right near where I work!!! I know starbucks is truly the devil...it\'s liek one of those things where it\'s soo bad...yet, soo good. You gotta say. nobody make soy chai latte like starbucks makes a soy chai latte.  I\'m a corperate whore!! damn them!  I need to start planning for halloween, I know i\'s early but ..IT\'s MY FAVORITE DAY OF THE YEAR!!!! I\'m a loser it\'s true.   The doll incident recorded a new song..and I gotta say it was just a silly song for our friends.We wrote everyone into it. I\'m in the process of writing some new songs, and I know alisha\'s doing the music thing on her end. but it seems like every time we get together, we just end up hanging out!! nothing gets done, a lot gets drank.....it never ends!!!  ah whateever...we\'re going to be famous..the magic eight ball said so..and we all knwo the magic eight ball never lies!!!  '),(21437,2751885,'19,June,2004','2004-06-19 00:00:00','   so we were supposed to go to this party thing happin\' this fine fine evening! but ALAS!!! the pary was canceled!! I got my face on and even washed my \"good\" underwear in the sink!...you never know who\'s gonna see your undies. c\'mon you know you\'ve done it. So maybe THE DOLL INCIDENT will do some recording tonight.   speaking of incident...anyone want a cat. he\'s real cute. Darwin has my whole apartment trashed. he has gotten into a fight with my guinea piga, chewed up two pairs of my shoes, chewed up the edges of all my book, scratched up my legs, AND ate some of my kohl (that cannot be good for him!) I\'m at the end of my tither...I really don\'t know wherwe o go from here!!!  Please god...kill me know!!  I love him...yet, I really really hate him!!  '),(21438,2751885,'18,June,2004','2004-06-18 00:00:00','   so, I found this super cool bog about these friends who moved to california to make in hte film industry... It interested me b/c I have found that lately a lot of my friends are also inspiring to move the the wonderful world of heat...  Anyway.   urlLink CALIBLOG  '),(21439,2751885,'16,June,2004','2004-06-16 00:00:00','   you know you when you look at porn (oh c\'mon now) and all these crazy windows keep popping up and you can\'t get rid of any of them....well we wrote a song about it...PORNADO!!    you can go on our site and download it for your very own. I think we woke amy up today...there was a little screaminess going on..and I felt bad about that..  oh yeah...I hate the bank. what\'s so wrong about keeping all my monety in a mattress anyway?? I lost my client card so I went to get a new one and the woman was a fucking moron about it. When I went to use it to buy something it didn\'t work. so I figure no big deal I\'ll go back to the bank. Well not so easy, as the bank hardly ever is, she had canceled my card all together!! why on earth would she do that!!! now I have noooo....money..  see!! the bank is stupid, and I\'m stupid for using it!!  that\'s it...I\'m getting a steel mattress with a lock. It won\'t be comfortable but at least I know that when I need five dollars all I have to do is go home and get it!!!  I\'m gonna post the lyrics for all our songs up here soon so you can sing along at home!!  '),(21440,2751885,'14,June,2004','2004-06-14 00:00:00','   So, here it is at last..MONDAY. I don\'t actually hate Mondays, but I usually don\'t work them either, so I guess Monday doesn\'t really hit me crazy hard or anything.  we wrote another song yesterday, \"Pornado\", we have yet to record it, but don\'t worry it\'s only a matter of time before we get this thing on the go!! woo hoo.  My cat is driving me INSANE.......He dose this thing now where he wakes me up in the morning by licking my face until it\'s all raw and hurty.......It hurts me, but he just so cute!! I love him ..Yet...Have a extreme dislike for him as well.   OMG, we went to go see the  urlLink The white cowbells they were fucking crazy, the music was great but they put on this fucked up show!! They had a stripper ( featured on the blog) and the guy played guitar with his penis!!! Yes, there is a picture of that as well...It was soo fucked up, he had this glass rod over it and he played the guitar with his penis!! fucked up I know! They even sawed up giant stuffed animals with a chain saw....Craziness all over the place. I will defiantly go see them again...It was really great!!  '),(21441,2751885,'13,June,2004','2004-06-13 00:00:00','   so we recorded a new song today...and I must say it\'s WAY better then the last one!!! me and alisha actualy sat down and wrote it....like thought about lyrics and shit.It\'s not that bad, and as we get to knwo what we\'re doping we\'ll be getting a lot better. we do have another song in the making...it\'s a bout pornado\'s..  are you excited yet?? go check out our new song!  urlLink The doll incident  '),(21442,2751885,'13,June,2004','2004-06-13 00:00:00','   Here we are in all our glory.......... urlLink The Doll Incident . And you can Download our song...our one and only song ...for FREE!!!! do it,  do it now !  '),(21443,2751885,'12,June,2004','2004-06-12 00:00:00','   urlLink'),(21444,2751885,'12,June,2004','2004-06-12 00:00:00','   ok, so we wrote and recorded our very first song yesterday, sure it\'s not that great the vochals could use a little help... BUT GIVE ME A BREAK IT\"S OUR FIRST SONG EVERRRRRRRR . so If yuo don\'t like it then I don\'t want to know...ok? cause it\'s mean to crtisize a person\'s first attempt on art.......ok, I haev to make an audio post and then record it form there...it\'ll be on the next post..  so we were sitting in alisha\'s room and cassandra was like\"I\'m cold I need a sweater\" and then we made a beat and alisha was like sing...so I did. and this is what came off the top of my head.....  '),(21445,2751885,'12,June,2004','2004-06-12 00:00:00','   so, me and alisha got the ball rolling on some sing-a-long fun last night. We actually sang a song and recorded it!!ok, that may not seem like a big deal to you, but believe me...it\'s the farthest we\'ve ever come to getting this band thing going. What suprises me is that I think we are really going to do it!!!! So we\'ve thought to name our band  The Doll Incident .Alisha came up with it after I told her a traumatic story of my childhood.....but I don\'t want to talk about that.   So I woke up this morning to my cat licking my entire face,it was cute at first then it really started to hurt a little bit and then I think he was trying to lick my skin off!aawwweee trying to lick my skin offff...how sweet. He\'s like a really cute little demon cat. he constantly scratches up my legs when I\'m sleeping, so I can no longer wear shorts, or open snadels. he also has my neck and chest all torn up. so, basically, I\'m all blody and gross.......ok, that may be exaggerated a little bit.  I do love him though...I can\'t imagine what it was like before.....  '),(21446,2751885,'11,June,2004','2004-06-11 00:00:00','   this is great,I laughed so hard I fell off my chair....what you say could make you fall off your chair..... urlLink this. It\'s an anti-masterbation article.......  \"Left unchecked, this progression can result in debt, sexual crimes, perversions and other harmful effects\" \"Masturbation conditions our bodies to respond to self-stimulation, which is self-centered.\" \"Masturbation poses the threat of lifelong bondage and a handicapped spiritual walk.  Jesus was resurrected so that you also could live in a new life, free from the slavery of sin.\" \"Masturbation causes you to be \"mastered\" by sin\"  I, on the other hand think masterbating is normal and I encourage you to do so. but that is my unprofessional opinion. but then again that whole article is some other persons unprofessional opinion.  wouldn\'t it be cool if masterbating was a profession......well porn I guess...but I mean they hired you to sit in an ofice and masterbate all day then write a report about you felt about that.....  good times, good times  '),(21447,2751885,'10,June,2004','2004-06-10 00:00:00','   Wasn\'t jesus a feminist? oh....maybe I was thinking of someone else.Check  urlLink this out . Strange but true.........  '),(21448,2751885,'10,June,2004','2004-06-10 00:00:00','   ok, so I was reading the archieves of  urlLink Alisha\'s  blog and I cam across  urlLink googlism . So I tried it on myself yet again (for those of you who don\'t know that is where \"all jane no dick\" came from in the first place)but this time it seemed to be way funnier. so her it goes::    stacey is reading  stacey is smart  stacey is mad at her parents  stacey is now a bay city bomber in the ardl the 2000 quakes  stacey is the queen   stacey is a clean teen hottie    stacey is nude in that car    stacey is also a member of an exotic dance group and loves to  move slowly and erotically in front of men   stacey is the last bond woman to feature in a view to a kill   stacey is having a bad time    stacey is the 1999 roller derby queen   stacey is there  stacey is the 211th most popular female first name in the USA  stacey is more raw and a bit earthier  stacey is developmentally delayed  stacey is waitressing   stacey is a woman to be reckoned with and not to be  slighted   stacey is going on a special trip   stacey is done hiding behind oversized sweaters and  thigh    Stacey is popular on the internet   stacey is determined to solve ian\'s problem  stacey is a teen girls fancy dancer and is 15 years old   stacey is a born again christian and believes in saving life  rather than terminating it    stacey is one of a kind   stacey is now married to that man who convinced her to stay in england  stacey is the doll that falls between the smallest sized kelly and  friends line  stacey is currently playing right wing on the rockets  stacey is now almost 8   stacey is hot   stacey is a great pal and i love her  stacey is a joy to hear and to watch  stacey is a really good book   stacey is the aries of all jane no dick    stacey is a native of illinois   stacey is a beautiful young woman who has it all   stacey is sure to go far in a world where only men traveled just a  few years ago  stacey is going on a special trip  stacey is a martial arts expert and \"by the book\" air force   stacey is a teen girlâ€™s fancy dancer     stacey is popular on the internet   stacey is a hard working  stacey is always dealing with people issues   stacey is all of these things and much more  '),(21449,2751885,'10,June,2004','2004-06-10 00:00:00','   urlLink  this is my cat'),(21450,2751885,'09,June,2004','2004-06-09 00:00:00','   so I tried to leave antother audio post but I kept deleting it cause I thought I sounded like an idiot. maybe I\'ll do one of my cat or somethign stupid like that...  I went to go see  the Day After Tomorrow , let me tell ya it was really crappy, like independace day with lots more ice and less alien action.\"look at us we\'re americans, everytrihng bad always happens to us but...no worries we\'re all strong white citizens who can overcome anything\"......oh yeah... and according to the way the movie ended, even global warming has a happy ever after!!  hollywod fairytales don\'t get better then this.....unless they involve keanu reeves!!  '),(21451,2751885,'09,June,2004','2004-06-09 00:00:00','   urlLink'),(21452,2751885,'07,June,2004','2004-06-07 00:00:00','   I went to go seee the new harry potter movie tonight. It was ok, not as great as the first two, mostly because the effects were all the same. And really with a movie like that it\'s why you go!! or maybe you just like the harry series...who am I to judge you on what you like....hm?  so I went with my friend luis. He, seriously, is one of the coolest guys I know. we\'re sitting in the theatre and luis looks over at me and was like \"hey stacey, check that out\" and he kinda nods to his right ....... low and behold I see these two bare feet jutting out from behind the seat. can you imagine feeling so comfortable in the theatre to take off your shoes and socks and just put your feet up!!!. I just felt bad for the guy sitting next to those two feet.....eewwwee!!! your all watching the movie and you look beside you to see two smelly feet right next to your head.....yeah I would have to move cause there is no way I\'m going to enjoy the movie after that!!!  someppl have no clue about what\'s disgusting!!  '),(21453,2751885,'07,June,2004','2004-06-07 00:00:00','   At the Aching-Heart Diner    She\'ll flavor her coffee with both cream and sugar and tap on the window as she mentions the weather, tossing off sparks when she pulls off her sweater. The waiter will come. She\'ll give her order bluntly: \"A hot chicken sandwich and a tall glass of water.\" She\'ll spatter her french fries in grand doses of vinegar and the point she is making, emphasize with a gesture, her steak knife held high like a gravy-stained scimitar. And the salt that is scattered when she topples the shaker she\'ll toss with a flourish across her left shoulder. I\'d like, I will say, to get to know you better. I\'ll look down at my clubhouse, so we don\'t look at each other as I pull out the toothpick that holds it together.    this is a poem by david o\'heara. I read it on the subway today as I haev many days. I don\'t know I just like the way it flows......  '),(21454,2751885,'02,June,2004','2004-06-02 00:00:00','   ever have one of those days where you feel like your in the way of eveyone you know?  I\'ve kinda felt like that for the past few weeks. I don\'t know why, I love my friends and I know they love me........  hmmm....Alisha come back from her travels afar!! she brought me back this tin box and it says \'box to make you pretty\' in french, and it this picture of three little girls and the one in the middle is pulling the hair of the other two...it\'s really cute. I put all my misellanious make-up in it today....now it\'s all full of  \"my face\" (that\'s what I call putting on my make-up,putting on my face).  my cat was alone most of the weekend, I cam back to check on him everyday but he was alone most of the time. Anyway now that I\'m home I don\'t think he has left my side for two seconds. he was waiting by the door for me when I come home....awwwweee. what a little cutie!!  when becca baby come back with her digital camera I\'ll take some serious pictures of the little cutie pie and post them all up for the whole world to see......I\'m going to be one of those crazy fucking cat people...with all the cat figurines and t-shirts with pictures of them holding their cats on it...  you know who you are....  \"this is my cat scampy...isn\'t he the cutest thing you\'ve ever seen. I made us matching outfits to wear to the movies\" and I would have to talk in that really high baby voice that hurts your ears a little bit...  '),(21455,2751885,'02,June,2004','2004-06-02 00:00:00','   so I went to go see  urlLink super size me   saturday night. wow!! They guy ate  39 lbs. of sugar!!  that is the grossest fucking thing I haev ever seen. I gotta tell you, when he threw up out the window of his car I ewas done with fast food for ever. no, really. But ....I think the portions in the states are way bigger then they are in Canada. But you know, it\'s not like the US to do things in excess!!!C\'mon, who the fuck needs half a gallon of soda and a LB. of fries.  Sweet Jesus!!!  gluttony at it\'s best.  So I\'ve had this copy of  urlLink Fast Food Nation  for what feels like forever and keep telling myself that I\'m going to readt it but yet never manage to start. I gotta tell ya\', I\'m going to read it now for sure!!!  OMG, and then I was telling this guy at work about SSM and he told me this story about his mom\'s B/F. This guy is a plumber, so, his company gets a call from McD\'s to come and fix their coke machine. So he takes the top off to look inside, you\'ll never guess what came out.....are you ready for this...no, really.......ok....thousands of  MAGGOTS .  No, I\'m not kidding you. I\'m not even sure how that would\'ve happened, maybe a fly got stuck in there and layed eggs?? ok, so now think about all the coke that was poured and drank before the plumber came to fix the pipes.God only knows how long it was like that.   now, what do you think about that?  '),(21456,2751885,'25,July,2004','2004-07-25 00:00:00','   when I was in elementry school CC was my forever-best-friend. She lived doen the street fomr me and I spent&nbsp;almost every weekend of my youth at her house playing dress up with her sister cloths. We would put on these fashion shows for her parents. Her dad would give us&nbsp;funny names and sing silly songs about us. He was always saying or doing silly&nbsp;things to make us laugh. He gave me a valuable lesson in humour, one I carried with me to adulthood.  It\'s ok to be ridiculous. It not only makes you feel good but it makes other people feel good too.  My mom called me early sunday morning to tell me that CC\'s dad had died of a massive heart attack. He was&nbsp;young, it was&nbsp;so sudden. I never realized how much he affected my life until that moment. That seems to be just the way it is I guess.   I\'ll miss him very&nbsp;much. &nbsp;&nbsp;  &nbsp;  '),(21457,2751885,'24,July,2004','2004-07-24 00:00:00','   What I want&nbsp; is to stop hearing that stupid song on the radio...I also want to get this new apartment that I want. And I would like some to tell my cat to stop peeing on my bed. yes that will be all for right now.   so what is it some cat thing where they are like \"spray me with the water bottle will yah...well I\'ll show you with my incredibly rank cat pee who\'s the boss up in here!!\"   and then he pee\'s where I like to sleep....In my bed!!!!! Can you say I don\'t wanna sleep I now wanna vomit!!   cat peee is like the worst smell imaginable. I Know, I can now smell it all the time. well, I got this recipe of like peroxide and baking soda and hand soap that I sprayed on it to make the smell go away...And it worked for the most part. It\'s still there really faintly but you have to actually stick my mattress up your nose to smell it. Which is not sooo bad, cause it doesn\'t really fit up my petite nose anyhow.   I had this dream last night that my evil ex faked his death to get back at me. It was really weird. He was shot by a gang.and then I found his body wrapped in newspaper. Then when I went to school the next day he was there and I was mad b/c him and his g/f had planned it all out to make me feel bad.   it was fucked up. Maybe my dislike for him is all coming back to me in sleep. That\'s too bad.  '),(21458,2751885,'23,July,2004','2004-07-23 00:00:00','   so, I\'m on my way to do laundry which I have put off for soooo long that I am now wearing the ugliest cloths I have and I almost had ot go comando, but lucky for me I found this pair of teeny tiny panties in the back of my panty box. now I\'m going to do laundry in the most uncomfortable way!!! and I have soo much shit to do before baden comes tomorrow!!!  but the good news is that I got us on the guest list for the fly this weekend!!! weehww. It\'s like a real gay bar and I\'m excited to take baden to meet some cute boys.&nbsp; well hopefully it will be all good and he\'ll have a good time!!  ok, laundry time it is!!!  my cat peed on my bed (where I was keeping my laundry..) and now my laundry bag smells like cat piss!!!!  gross.  '),(21459,2751885,'22,July,2004','2004-07-22 00:00:00','   this is my 100th post!!!! woohooo!!  ok, back on topic. I seen THE funnest movie tonight.  urlLink http://www.apple.com/trailers/fox_searchlight/napoleon_dynamite.html &nbsp; Napolean Dynamite is the funnest movie I have seen in such a long fucking time. I laughed from the time the movie opened right up until the closing credits. It was soooo good...there was clapping at the end..I kid you not. People in the audience clapped. I know you always feel like doing it but nobody ever dose....well people did. and I gotta tell it...this movie really deserved it!!   it\'s about the guy we all knew in hight school who is so obviously weird but the only person who dosen\'t know it is him, because he just dosen\'t care!!!   \"do these chickens have large talons?\"  '),(21460,2751885,'21,July,2004','2004-07-21 00:00:00','   I\'ve seen super size me, and now I have read fast food nation. Both were valuable pieces of information ....  so why then do I still feel the need to eat at mcdonalds???? is it the cheap food.....no.  is it the great taste................NO!!  I don\'t know why I eat there, I know that they feed dead chickens to the meat, along with animal litter. I know that a great deal of the meat we consume contains fecal matter which could cause e.coli, and that could lead to death. I know they market high fat, high calorie, low nutrition food to our children.They pay there workers considerably low wages,&nbsp;cheat their workers an are STRONGLY against unions.  yet when I walk into a McDonalds, I don\'t see any of those things. It\'s like the place I know all these awful things about is somebody else\'s McDonalds. It couldn\'t be mine...I eat here all the time!! all those years of fast food couldn\'t be that bad for me...could they?? could all the things that happened to morgan spurlock really happen to me, Stacey Janes??  I know your thinking......what is your point. well I\'m not going to eat fast food every meal for a month that\'s for fucking sure. but I think starting today, I\'m gonna boycott fast food for real this time. it\'s really really gross. and I think I have to come to realization that what\'s happening to fast food in the states is happening here too. and that the McDonalds down the street from my house is part of that large corperation that supports and encourages near slave labour in the meatpacking industry. I need to start thinking that Ray kroc lives down the street from me at my McDonalds .  today.......day 1.  &nbsp;  '),(21461,2751885,'18,July,2004','2004-07-18 00:00:00','   So yesterday we watched the entire marathon of \"WB\'s SUPERSTAR\" you Know the one where they tell the people that they are rally good when they actually suck. It\'s sooo crazy mean ....Yet entertaining....hmm.  &nbsp;  so we decided that the girl who won, at the end when they told her it was a hoax but then gave her a recording contract anyway....Well....They didn\'t actually tell her it was a hoax, they kinda said it all vague like and that was MEAN!!! I mean C\'mon do you not think she\'s gonna figure it out when she see\'s the show and gets humiliated all over again?? It was soo awful, they convinced this poor girl she was gonna&nbsp;be a star...but yet she\'s never going anywhere b/c they LIED ...and she\'s actually the worst singer in america! &nbsp;  but man, these people were really awful. And it was THE worst show I have ever seen on TV. The worse part was once we started watching we just couldn\'t stop!!! It was like a really gross wound that you can\'t stop yourself from touching, or the weirdly ugly guy on the bus that you can\'t stop staring at...&nbsp; &nbsp;  you know what I\'m talking about!!!!!! Don\'t try to act like your above all that...Cause I know your not!!!!!!  '),(21462,2751885,'16,July,2004','2004-07-16 00:00:00','   I\'m talking about the shot...but I mean you know..... &nbsp; So I was talking to a friend and he told me he dated this girl who could have orgasms in her sleep!!! I gotta say , how much would that fucking rock!!! I eman I love sleep as much as one person can...but to have my favorite awake activity become my favorite sleeping activity!!! whoah....I just don\'t think it can be any better then that for real!mixing sleeping and cumming .....SWEET JESUS!!!! &nbsp; I have to stop thinking it.....b/c I am not ont of the fortuate few who has ever had an orgasm in her sleep and I just envy those who have! &nbsp; we went to the fun hause last night (we thought it was modanna night) but it was this hard core metal thing...and there were girls dancing around in 3\" heels...to metal........weird. yes honey, we can all see that your trapped in a box and would like someone to let you out, and god knows these guys all want to touch your box, if you get me.&nbsp;  &nbsp; ok, time for bed!  '),(21463,2751885,'16,July,2004','2004-07-16 00:00:00','   becca\'s fish however are not doing so good. one of them died and I didn\'t notice and the other ones ate most of it! so when I went to fish(no pun intended) it out....it was sadly only a head! yeah it was kinda gross and sad. &nbsp; yeah that\'s all I got for right now...I dyed my hair back pink.&nbsp; it\'s all my dreams come true, I lost weight and I have pink hair....look out boys here I come!!!!!  '),(21464,2751885,'11,July,2004','2004-07-11 00:00:00','   So we went out friday night and got our drink on. i was good times. we talked to some people and what not....you know the usual good times to be had goning drinking on a friday night downtown.   but I gotta say the highlight of the night was the ride home on the street car!!  we go to sit in the back and there is this guy back there all by himself...he\'s like \"you can\'t sit over there there\'s a boi-hazard on the seat\"   it was a BLOODY SOCK!!!! it was this tube sock cover in blood strewn across the seat. he was like \"I went to tell the street car driver and she\'s gonna wait till she gets in the yard and call the cops\". so we all sat back there with him to help protect people from the bloody sock on the seat. People were getting mad b/c chris was like \"dude you can sit back there it\'s a bio-hazard...you don\'t know how many people\'s blood was on that sock!!!\"  the driver kept anouncing \"please be aware there is a bio-hazard on the back of the car\"  the saddest part of that whole thing (besides the person who bled all over the sock and what happened to them) was that we know if we hadn\'t of refrained people from going back there....someone would have just moved the sock and sat down.   that\'s gross.  '),(21465,2751885,'09,July,2004','2004-07-09 00:00:00','   so I\'m a nanny right. and I take my ladies to a near by park and until today I thought this woman pushing a carrage was just a homeless women. but no! she actually thinks the bag full of stuff in her cart is a baby. I heard her talking to it today, asking it if it seen my babies and weren\'t the cute! she always sleeps in the park with a blanket over the stroller, I thought it as so noone would steal her stuff....bt no, it\'s too not wake the he plastic bag...  a old man stopped me today and asked me if the ladies were mine and when I said no, he was liek \"I thought so, your wayyyy too young to have two babaies like that\" what if they were mine.....he would have openly judged my \"youngness\". that\'s just wrong. I\'m 24 I could totally have two kids under the age of two and be married and all that other adult shit. I mean realy, my mom did it!  '),(21466,2751885,'07,July,2004','2004-07-07 00:00:00','   my mom told me today (via e-mail) that sje wants me to go to therapy....like a real shrink type deal!!  dose that mean I\'m fucked up, or is that the whole point of going to therapy, for them to tell you fucked up and then try to fix it?  I don\'t think I\'m gonna go though...I have enough going on in my little brian without some so-call-doctor filling it up with stuff. I know full well why I\'m fucked up without somebody telling me why they think I\'m all fucked up!  I\'m a big girl who can over analyze things all on her own!  '),(21467,2751885,'02,July,2004','2004-07-02 00:00:00','   soooo....WE HAVE A NEW SONG!!!!! I know you wann hear it right..? so go go now, urlLink the doll incident   just go to the music page and go to creamiest...that\'s the name of our new song. it\'s super excellent!!!!!  '),(21468,2751885,'01,July,2004','2004-07-01 00:00:00','   you HAVE to check this out. I found this drunk guy walking game...how far can you walk when your three sheets to the wind!!!!  urlLink TRY IT!!!!  '),(21469,2751885,'01,July,2004','2004-07-01 00:00:00','   so,this is my idea for a t-shirt and underwear making company. well the underwear thing was amy,alisha, and amy\'s idea, but the t-shirts were all me. so I looked on the internet and wanted tosee if anyone had yet comeup with the name \"venereal design\" and nobody has!!! it\'s all mine baby, and now I have written it in my blog so nobody can ever steal it b/c I have a company named that now!!!  if you would like to place an order for a offensive t-shirt...e-mail me and let me know. so far we have the following phrases... \"lick my clit\" \"I can\'t answer with your dick in my mouth\" \"go! vaginas go!\" \"drunk on jesus\"  so yep..... oh yeah...HAPPY CANADA DAY!!!!!!!  '),(21470,2751885,'07,August,2004','2004-08-07 00:00:00','   urlLink'),(21471,2751885,'06,August,2004','2004-08-06 00:00:00','   urlLink'),(21472,2751885,'04,August,2004','2004-08-04 00:00:00','   urlLink'),(21473,3870139,'14,July,2004','2004-07-14 00:00:00','    ... then i just found out that little tears are forming on the corner of my eye. Can\'t believe it, been crying for a movie. Well, hard for me to resist but honestly, i\'m just carried away. Bet you take it as corney and a cliche-type, but for me, it is worth piercing to the hardest entity of our human emotion. I myself who is proven to be numb(numbing me?!) got carried in every scenes especially the last excerpts of the movie. I find it inspiring to live the life to the fullest as what am I doing now. It gave stresses my point of \"Life no matter how short or long is still life, love is the element that prolongs and shortens our finite life\".   '),(21474,3870139,'09,July,2004','2004-07-09 00:00:00','    Hey I\'m with my classmates here(Bootc@mp). Purpose? Wasting time!!! Hehehe, honestly, wasting time. You know, am waiting for my girlfriend. I\'m waiting till 7:00 and it\'s still ten mins. before five. Well, gotta wait for love\'s sake. It\'s just a simple sacrifice since I have nothing to do in the house. And it\'s my pleasure to post here so nothing to bother right? But most of all, I want to thank my two companions, Kate and Nollie. Thanks my friends!!!   '),(21475,3870139,'08,July,2004','2004-07-08 00:00:00','    God dammit ?*#@@@! Who the heck he think he is. Yeah, another bad day from one son of bitch. We\'ve been in a bullshit computer shop with one son of a bitch for an owner. What does he think he is, he can put his filthy hands over my face. He even doesn\'t have the right to speaj on me like that. Well, don\'t want to talk about it anymore. It\'s just another bad day that will be left behind tomorrow. It\'s getting the shit out of my ass.   '),(21476,3870139,'08,July,2004','2004-07-08 00:00:00','    I\'m tired of being  what you want you want  me to be Feeling so faithless Lost under the surface I don\'t know what you\'re expecting of me Caught under the surface of walking in your shoes Every step that I take is another mistake to you  I\'ve become so numb I can\'t feel you there Become so tired so much more aware I\'m becoming this all I want to do Is be more like me And be less like you  Can\'t you see that you\'re smothering me holding too tightly Afraid to lose control Cause everything that you thought I would be has fallen apart right in front of you Every step that I take is another mistake to you And every second I waste is more than I can take  And i know I may end up failing too But I know you were just like me with someone diasappointed in you   '),(21477,3870139,'07,July,2004','2004-07-07 00:00:00','    Been seeing me quite so happy. Nah, you saw it wrongly. Been seeing me that sad, well you might be right this time. I\'ve been doing this for a decade. Showing up masks to hide my weaknesses. Hard to show up what is the real me. Frighted to let others know who am i for real. Don\'t take it as badly as you feel now. I\'m not fooling you that way. I\'m just being precautious of showing up my personality to everyone. Reasons? Yeah... I really mean things for the reason. Hard for me to say but need to give a clincher as a must. Well, it is my experience who gave me the cause for this practice. That way you might understand me. Hope so, well don\'t take this seriously, it\'s part of my collection of mask. Got you again.   '),(21478,3870139,'07,July,2004','2004-07-07 00:00:00','    I was here, somewhere, nowhere you can find. Indeed, I don\'t know where am I too. Just been here for nothing. Nothing? No, been here for one thing. Well, it\'s my favorite chore to do. Yes, you read it right, my chore. It\'s my way of releasing myself, sort of an outlet where I can reach out for my personality. I\'m sorta wander from street to street or alley which I find to be my refuge of being alone. Alone and away from society, outcast maybe. Well, I\'ve been enjoying this for an hour so cause after this, I\'m back with my social life. Another part of me, another part of numbing joseph. The dark side or lighter. Don\'t know...   '),(21479,3435900,'27,May,2004','2004-05-27 00:00:00','    On my way to the city yesterday I saw a really bad car accident. The kind that you aren\'t really sure that everyone will live through. There were like 3 or 4 ambulances and fire trucks and so on. I didn\'t look for details because I would have fainted and caused another one. Well this nasty accident reminds me of how good we really have it here. Sounds stupid but... In 1998 I was in Nigeria. It was still under military rule and just to drive a few miles down the freeway you were stopped, questioned, scared shitless, and extorted for as much as they could get from you easily. They all carried machine guns you know the \"I will fucking kill you \" kind. I never saw anything like it. OK OK to the point. I was traveling at night from a small town called Ile Ife to Lagos. There were 6 of us in an older Mercedes, so it was a full car. As we were going we hit some traffic on the freeway, it was dead stopped actually. I could see a fire on the other side of the freeway and heard people screaming. As we got closer to the fire I could see a minivan on its side engulfed in flames. The driver of the van was dead, on fire, his body still in the position on sitting even though it was laying on the ground. I was sick with panic and then it just got worse. People that were in the van that had escaped with their lives were walking along the side of the road crying bleeding, their clothes ripped off along with some of their skin. I started crying and begging my husband to let some of them into the car. Like I said it was full really full I was on his lap in the front seat. I said lets get out and let Daddy take them to the hospital.  This was also in the middle of nowhere. The sides of the freeway were virtually jungle and there was nothing within miles and miles. Like I said before its not the safest of countries to be in to begin with and not the best place for a  lone white woman to be standing in the middle of the night along the side of the dark spooky animal infested jungle. Still I wanted to get out. I wanted those people to have my seat. I couldn\'t convince anyone that my idea was even close to sane. Humph. For once my whiteness was an issue. I said \"where in the fuck are the ambulances goddamnit!\" Well my husband being from Nigeria as well as the driver, his uncle, they were not fazed as I guess they are accustomed to watching people die HOLY FUCK SHIT MOTHERFUCK. The passengers in the back seat were 2 of our friends from El Salvador and my father in law. I was LOOSING MY SHIT, while everyone else was mildly irritated. I said give them money at least and my husband said something along the lines of \"If you give a monkey a diamond he has no use for it\" his ass is notorious for having some dumb ass shit to say at the most inappropriate time.  We eventually passed the accident, we passed to injured crying,DYING people, and drove off towards the next stick up station that for the first time in weeks was way way too far from where I wished they were.  That was over 6 years ago and I still feel sick at the thought of it. There are many many things that I learned on that trip but this one really makes me appreciate something I probably took for granted everyday of my life before that trip. I mean we all kinda think we have it good here but you might not really \"know\" how good we really do have it.   '),(21480,3435900,'25,May,2004','2004-05-25 00:00:00','    How is it that someone so adorably cute and wonderful can actually be the casing of what I can only equate to a tasmanian devil with a severe crack addiction? Well my daughter (2) has a runny nose and woke up last night at 1am and wanted to come into my room to sleep. I am slightly retarded and thought poor baby come sleep with mama. Well lets just say that after a full night of twisty turny kicking snorkles I managed to get at best 2 hours of sleep. Damn cute genes can trick me into anything. Humph.  OK I have to say it. I am a faithful watcher of the Sopranos, I have been since the first season. OMG is all I can say after Sunday\'s show and watching it again last night. I mean we all knew that it would eventually happen but damn, I mean really damn. I am fully comfortable with all the shit that is done on this show but I don\'t think they have ever done a show that the \"victim\" knew that they were on there way to well the end. It was just dirty and I never thought that Christopher would be so in on it. Wow. The power of a fictional whacking. I\'m still bothered by it. I love this show.  Another kind of strange thought that has been going through my mind is a recent visit to my great Aunt in San Diego who just turned 94. I took my son (18) and left my daughter with her dad, and my mom also came along. Well lets just say that at 94 she is not exactly a full blown racist but she slips up and says some funky ass shit from time to time. I know that she tries to bring her remarks down when I am around because both of my children are bi-racial and I have, well lets just say I don\'t have the ability to not talk shit to anyone when they get all stupid and act like ignorant asshole racist fuckwads. She informed me that the little \"colored\" clown on her table was given to her by a neighbor. A not really \"colored\" neighbor but a mixed up one. She then went on to make the watermelon comments to which the younger members of the family never heard, as in black people love watermelon, and they (the younger kids)who by the way seem to think that they ARE black but are almost clear they are so white, went on to say that\'s not true they like 40\'s and rap music. I cant believe that I actually am related to such idiots.   '),(21481,3435900,'24,May,2004','2004-05-24 00:00:00','    Well here it is. My first post. I am going to start here by giving you the 100 things about me.  1. I have 2 kids  2. I have never been arrested 3. My favorite color is green 4. I hate black licorice 5. I have been to Africa 6. I stopped smoking cloves 7 years ago 7. I don\'t like nightclubs 8. I am a registered Democrat 9. I was a teenage mother 10. I live in Northern California 11. I have 2 tattoos and want more 12. I don\'t drink beer 13. I don\'t wear makeup everyday 14. I believe the epidural is the worlds greatest creation 15. I hate the smell of tar 16. I went to summer camp 17. I lived in Capitola for almost 20 years 18. I was born in Hawaii 19. I dye my hair 20. I swear like a sailor/truck driver 21. I see at least one Shakespeare play a year 22. I have a garden 23. I was a tomboy 24. I had my kids 16 years apart 25. I have a crystal ball 26. I met all the members of NWA 27. I like my kids more than anyone else 28. I hate people that hate people 28. I love the smell of Jasmine 29. I wear Birkenstocks 30. I don\'t understand Republicans 31. I married my husband after knowing him for 2 months 32. I am soft 33. I have fired a gun 34. I wear glasses 35. I am an atheist 36. I believe in love 37. I drink way too much Diet Coke 38. I have never been on a cruise (yet) 39. I have seen almost every Disney movie 40. My favorite number is 17 41. I have been online since 1992 42. I play cribbage 43. My eyes are grey and change color 44. I have seen George Carlin live 45. I am a compulsive gambler 46. I love Champange 47. I learned to drive when I was 12 48. I got my first car at 14 49. I like rap music 50. I have visited people in jail 51. I am 5\'3 52. I have slept on the beach 53. I have been to Vegas more than 5 times 54. I like my kids more than any other people 55. I did not graduate from high school 56. I cant stand cigarette smoke 57. I want to go to Paris 58. I have had my eyeball licked 59. I think men should not have opinions about abortion 60. I don\'t believe in the death penalty 61. I have broken the law 62. I was kicked out of kindergarten for fighting a boy 63. I don\'t use drugs 64. I like to watch people laugh 65. I can swim 66. I have a TV in almost every room in the house 67. I like to be in the woods 68. I don\'t watch soap operas 69. I love Frank Sinatra 70. I was a bad kid 71. I still am 72. I don\'t have a cell phone 73. I know how to crochet 74. I like roller coasters 75. I wear a size 7- 7 1/2 shoe 76. I have 8 piercings 77. I love EBAY 78. I see George Benson in Saratoga EVERYTIME he is there 79. I played the saxophone in elementary school 80. I never laugh when people fall down 81. I am a slob 82. I don\'t know my multiplication tables 83. I am an Amnesty International member 84. I don\'t like caviar 85. I love Prince 86. I wear sunscreen everyday 87. I didn\'t like oatmeal until I was 32 88. I think Al Franken is a stud 89. I don\'t know if I will have anymore kids 90. I believe in monogamy 91. I have never gotten a car with a big bow on it as a gift 92. I was at the last Giants game at Candlestick park 93. I have never been stung by a bee 94. I had my appendix removed when I was 18 95. I am a great cook 96. I didn\'t read a complete book until I was 19 97. I want more diamonds  98. I had fake ID at 17 99. I love men in Tuxedo\'s 100. I have taken Prozac   '),(21482,3435900,'24,June,2004','2004-06-24 00:00:00','    Well I have been missing for a few weeks now, but am now back in action. I took a little vacation to well the looney bin. I have suffered with major depression all my life and I finally decided to go an try some different medication. Well its GREAT its better than that even. Its called Lexapro and if you haven\'t had some get on some, well that\'s if you are depressed I guess. I mean I love it, I am so fucking happy I am singing happy songs. I feel like a superhero. I am seriously thinking of changing my name to LEXAPRO! I cant ever remember being so completely content. I have taken several things over the years but woo hoo this is it. For now anyway. So I am back, I will be back reading all my favorite blogs, and starting a cult to worship the \"drug\". Life is good.   '),(21483,3602354,'26,June,2004','2004-06-26 00:00:00','   I can\'t bear it!!! I\'m sad!!!! Really sad!! And i can\'t help it, i\'m crying, i\'m sad. Why did he do that???? Was it so bad??!! I\'m sure there could have been another way out... i feel shit!!!  \"Ich verlasse heut\' dein Herz Verlasse deine Liebe Ich verlasse dein Herz, dein Leben deine KÃ¼sse, deine WÃ¤rme deine NÃ¤he, deine ZÃ¤rtlichkeit\" I can\'t stop crying!!!!! This shouldn\'t have happened, not this soon... why do i have such a bad luck??!!!! Why does this life hate me!!!! What have i done to deserve this shit??!!! I had all this amazing dreams and hopes about it!! And suddenly all of them are taken away from me!!! It\'s not fair!!!! I\'m not angry i\'m sad!! I have to stop this suffering!!! but how!! i can\'t!!! I want to be with him!! i love him!!! I just one my happines to last... at least just once!!!  \"Suchte ich die Herrlichtkeit der Zweisamkeit und fand nur die lÃ¤cherliche Einsamkeit\"  '),(21484,3602354,'25,June,2004','2004-06-25 00:00:00','   All my fears related to the little boy are gone, but not precisely in the way i wanted them to go. Its really simples to explain the fact but not the feelings which came along: He dumped me... just like that. He said he didn\'t want the commitment of having a girlfriend, he didn\'t want to do all which that implies, it means: call me everyday, be with me the whole day... and all that stuff. But it is nonsense, why? Easy: BECAUSE I\'M NOT LIKE THAT!!! I don\'t ask all of that, because i hate that kind of dependence. But he just imagine he had to do all of that, and got sick of it, although, like i said, i wasn\'t going to ask him all of that. Of course i would like him to call me, but if he doesn\'t and i really wanted to talk to him i would call him. Also with the \"being together all the time\" thing, i didn\'t ask him, that and neither was i going to. Anyway, he made his decision, so there\'s nothing else i can do, but mourn and weep a bit and keep on with everything else. Definitely i\'ll take a big break with love. At the end i always get hurted, if not because of me, because of somebody else, but always hurted. I\'m not going to stand it anymore, i don\'t have to. I know i always say so and never keep the promise, but this time i will, for my own sake. I want to this break to last at least until i go away from here, then it will start another story of my life. I hate love, i really do... but i have to bear it all the time, because of my stupidity. Too bad for me, but i\'ll fight against it... or at least i\'ll try.  '),(21485,3602354,'25,June,2004','2004-06-25 00:00:00','   Maybe now you don\'t know i\'m talking about you, but as long as you find out you\'re the only who knows this page, you may realize that this has precisely written for you. (Too many \"you\'s\" in that paragraph jaja)  Anyway, let\'s get to the point. You may think i\'m also blind (like you) because until today i didn\'t realize you had posted two comments, well it was only one but it said there were two, one deleted and one there jaja. Well the matter is that i just saw it today and it really hit me, not in a badly way but in a weird one. When i started reading what you wrote (after the first paragraph, the one with the explanation) i got this feeling, a nice but weird feeling. I really liked you telling me your thoughts about what i had written, it was great to get a friend\'s opinion, and even greater coming from you, one of my best friends (if not the best of all jeje). So all this is to thank you for writing to me, and telling me what you think and feel about what i think and feel.. And thank you for the comment about buying all my books jaja the first one may come out in a few months, so get your wallet ready jaja. Well, that was all i wanted to say to you: THANK YOU SO MUCH!! See you later... soon i hope.  PS: by the way, you still have something to tell me, remember!  '),(21486,3602354,'16,June,2004','2004-06-16 00:00:00','   Finally, i\'m with the one i wanted, but there\'s a bit of a problem... i haven\'t told anyone, and i\'m afraid to. But i don\'t even know why (like with all my other fears), maybe cause i think it\'s too much early after the other guy, and also maybe cause i\'m may be criticized by hanging out with a 13(almost 14)-year-old boy... i know that in many ways he\'s still a child, but i really like him, i have a very good time with him... anyway i\'m still afraid. The other thing that frightens the hell out of me, is that we are already boyfriend-girlfriend, so i just don\'t want to hesitate no more, because that end up being the reason of breaking, and i don\'t want it to be this soon. I really want to last at least a couple of months with him... IT\'S ALL I\'M ASKING!! I have to trust myself and see what happens. Damn it just seem so difficult, and that\'s why i\'m scared... i don\'t want to end soon with him, i really like him... (i have already said it twice, it must be something.. no, it may not) I wish i just had someone i could talk about it without repressing my emotions or anything... i even had to lie to one of my best friends!!! Just because i was afraid of telling her the truth about him, that i like him and that we are already together. Anyway, now i\'m WITH him, so i have to enjoy it without worrying about it.. looks hard but i\'ll make it!.... hope so.  '),(21487,3602354,'16,June,2004','2004-06-16 00:00:00','   Sometimes i see my life like through a glass to give me the chance to appreciate all the things this fucking life has given to me.. and i can see that i have too much to live for, i guess. Any way, sometimes i feel extremely worthless and i don\'t care about those things, i just want happiness fallen from the sky... which isn\'t gonna happen. I\'m a mess, no doubt about it... what the fuck am i doing?? I don\'t even know what the hell is going on with me!! I just have this feeling of being lost and going nowhere. I just want to do something that fills my life, but i feel it so goddamn empty! I want out of here, somehow... don\'t know! Why am i so scared of everything?? Sometimes i think i just want to be like those \"risk-takers\", those who doesn\'t care what\'s gonna happen next, they just do what they want... but for some reason i\'m not like that... i\'m always afraid of what may come next!!! That just sucks! I\'m sick of being afraid!!! Fear has become a part of my life, and i don\'t want it to be here any longer... what is for me to do!! I wanna cry forever and don\'t give a shit about nothing!! I want to love, to dream, to write, to sing, to do what i like... without boundaries!!! But that\'s just not gonna happen.. it really sucks!! I SUCK!! Well, i shall better be going, i don\'t want to get angry nor hate me.... i just hope it isn\'t too late...  '),(21488,3602354,'15,June,2004','2004-06-15 00:00:00','   Demonios!!! Al paso que voy nunca voy a llegar a ser escritora, soy un fracaso!!! Nunca hago nada en todo el dia y nomas no escribo, asi no voy a poder ni siquiera completar la publicacion del libro, se supone que debo tener aproximadamente 15 textos para elegir entre ellos, pero apenas llevo 10, y todavia me falta pasarlos corregidos, apenas llevo dos corregidos... y los otros son los que necesitan arreglos mas cabrones... buaaaaaa!!!!! En fin, solo me queda esperar a que encuentre el momento ideal para escribir, creo que los ultimos tendran que ser poemas cortos porque si no, no voy a terminar. Lo de la cancion de wany y de inzunza ps creo que tendra que esperar jeje porque ya nada mas me quedan 2 o 3 veces de ir al taller para corregir textos y entregarlos para la publicacion... y todavia falte este lunes... no no no En fin, ya vere que hago, por el momento tengo que disfrutar mis vacaciones jeje y ps ver que sucede con este tipo... \"es cuestion de tiempo\" ScheiÃŸe!!  '),(21489,3602354,'11,June,2004','2004-06-11 00:00:00','   El domingo fui a ver a Haggard.... no hay palabras suficientes para expresarme... realmente necesitaba liberar estres y toda la tension que traia. Estuvo excelente, realmente genial :D!!!! Si tuve pedos para entrar jejej pero a pesar de eso lo logre!! Tocaron como 2 horas, estuvo increible. Logre tener muchos autografos, fueron como 10 u 11 de los 17, no esta mal jeje y una foto con el guitarro!!!! Queria comprarme una camisa pero ya no pude... fue muy triste!! Pero ya despues me comprare una jeje a la mera cuando vaya a Lacrimosa!!!! Ya falta poco mas de un mes, ya no aguanto!!! Nomas que esta bien esperarme porque tengo jodido el cuello jejeje Insisto en que el concierto estuvo de huevos jajaja bueno pues lo dejare por la paz, por mi salud mental... ya no se ni que digo pero bueno.. YA QUIERO QUE SEA LACRIMOSA!!!! (pero antes que se arregle mi cuello jajaj)  '),(21490,3602354,'11,June,2004','2004-06-11 00:00:00','   Tengo que escribir la rola que me pidio wany :S:S y no se ni como hacerla...  Tengo que escribir algo tambien para la rola de inzunza y tampoco se como!!! Tengo que corregir los textos para el lunes!!!! Y lo peor.. TENGO QUE ESCRIBIR MINIMO OTROS 5!!!!!  AAAAAAAAAAAhhhhhhhhhhhh voy a llorar... no tengo inspiracion, nada pero nada... y tengo que seguir pensando... aaaaaaaaaaaaaa que hueva!!! En fin, tarde o temprano tendre que hacerlo... AAAA que coraje mi vida es un pinche desmadre, y tengo miles de cosas que hacer antes de arreglarla.. es el colmo. Bueno ya que... en dos dias voy a disfrutar un concierto de huevos!! asi que ps ya despues me preocupare jejeje  '),(21491,3602354,'10,June,2004','2004-06-10 00:00:00','   What on earth am i doing?? One minute i was begging for love, in the next one i was leaving it behind. Anyway, it isn\'t the worst part of it all: everyone is getting weird ideas into my head, but why am i listening to them? I really have to work hard on that, i can\'t let them tell me what to think or feel. We are not so like the perfect couple or anything... that something i have to get clear. We are too much alike but that\'s exactly what makes us so different... what a crap! Sometimes i just wish i could be kinda normal... but when i think deeply into it, i realize i really don\'t wanna be so, it\'s nice to be weird, to be a freak... although it causes me so much troubles sometimes. Again the subject has suddenly changed... scheiÃŸe! I\'ll shut up my mouth for a while, i think that\'s better  '),(21492,3602354,'29,July,2004','2004-07-29 00:00:00','   That\'s the real me..&nbsp;but because&nbsp;i like more Tarja\'s picture than mine that one is in my profile and not this one.   Getting back to important stuff... I\'M HARDLY WAITING FOR TOMORROW AT 8PM!!!! (or is it 9??) i really think i\'m gonna cry... well that isn\'t too hard to guess since i\'ve cried in almost every concert i\'ve been to... jeje i really get excited. Let\'s see how everything goes... meanwhile i shall make a stop...    (doing the laundry)   i\'m back now... let\'s see, where was I?? ahh well the concert theme is over i guess. My mantra!!!!!! Cradle of Filth&nbsp;- Satanic Mantra...  Ahh i really don\'t know what to talk about, i have too many things in my head and i\'m not thinking very clearly... so for everyone-who-read-this-thing\'s sake i shall shut my fingers up...  Don\'t die... or die.. as you please, right now i don\'t care.... well i do, so don\'t die  '),(21493,3602354,'28,July,2004','2004-07-28 00:00:00','   My&nbsp;heart is killing me... or is it my mind?  It doesn\'t matter, at the end it is just the same: pain I know i won\'t die for loving you, i\'m still alive I know i won\'t cry for your love, it\'s already mine The only truth i know is that there\'s darkness around us bluring our sight and blinding our souls. I also know my aim is to be at your side feeling your warmth and breathing your life. But&nbsp;i had been cursed with an eternal tormented love. I\'m afraid of bringing your life down into ruins i\'m scared of ripping your heart with my unmercyful hands i&nbsp;don\'t want my despicable soul to be in touch with yours i don\'t want my acid&nbsp;tears to melt down your body I\'m under the spell of fear and despair but underneath both of them, i love you. I will just love you in the darkness of&nbsp;death... were i\'ll be able to.  '),(21494,3602354,'24,July,2004','2004-07-24 00:00:00','   Lacrimosa is coming!!!! Just thinking about the whole scene of being there in the middle of their music makes my heart beat out of its place...  Although i\'m tired... exhausted really. Taking care of my mother was harder than i thought.. it\'s not that i wouldn\'t have done it, i would have anyway, but its horrible to sleep at 12am and wake up (or being waken up) at 7am or even earlier... well, i don\'t want to talk about my awful days, it has been enough by living them.. i\'m just pleased that my mother is getting better, its a great relief.  I just wanted to talk about something... write about something. I\'m sick of this, i think i want to die after the concert.. it would be the most amazing death i could get.... aaggrrhh no!!! i won\'t talk about my death though those images are coming up to my head again... hate it!!  Let\'s talk about.... my writing.. i hate it too, but there\'s nothing else to talk about. I\'ll write down here somethings i wrote the last week:   \"...sentÃ­&nbsp;mi alma cayendo, congelada, a los pies de la esperanza (...) dentro de mÃ­ habÃ­a un espÃ­ritu lleno de fragilidad, al borde de la desesperacion...\"   \"El corazÃ³n dentro de mÃ­ derramÃ³ una lÃ¡grima de ira pero se contuvo de gritar su dolor. Me sentÃ­ tan sola, tan abandonada, sin una mano de ayuda ni un brazo de apoyo. SÃ³lo mi madre y yo; ella sufriendo, yo agonizando; yo cuidando de ella, nadie cuidando de mi.\"   To end this writing i\'ll say my mantra to give me strength:  \"Archangel, Darkangel,  &nbsp;lend me thy light  through death\'s veil  until we have heaven in sight\"  '),(21495,3602354,'24,July,2004','2004-07-24 00:00:00','   I\'ts been a while since i haven\'t written anything in here, but i\'ve been out of this business a litte. I just want to post something to remember i have this thing.  A couple of minutes ago, searching for my nickname i found spmething kind of nice somewhere inside my mind: \"Alone and afraid... breathing despair and walking beside death... With no one to love... nevermore.\" Sad but true.  Since some days, i\'ve wanted to cry with someone, because crying alone does not calm me down anymore, at least this time it didn\'t. Bad luck, there is no one to cry with, no one who can listen to my sorrows, no one to hug me... just no one. The only thing i can do i wait in here, wait for that which i know will not come.&nbsp;I\'ll&nbsp; try to be strong and bear my pain. I\'ve been longing for some hope, for perfect love... such a foolish illusion! Anyway&nbsp;i know i\'ll be here waiting for some of it to come, a little hope, a bit of love. Meanwhile, i\'ll be trying to carry on with this process of time called life, or&nbsp;trying not to die...&nbsp;or at least enjoy the effort.  '),(21496,3602354,'17,July,2004','2004-07-17 00:00:00','   Some quotes not from me:  &nbsp;  \"She imagined a specter who told her that, if she wanted, she could still walk to the square, she could still find a man who could ask: \"Do you want to be my girlfriend?\" But the silence and the night inside her house said: No one.\"  &nbsp;  \"The world\'s last explosion. My blood splattered on the wall. My body, my shattered head, fell on the floor.\"  &nbsp;  \"Almost in silence i say: \"love\". The word hovers before me for long into the night, amid sounds, amid gestures.\"  &nbsp;  \"The word stops and i remember you like a penetrating knife. The infinite blade of a knife infinitely thrust into me. An eternity has passed since you left me alone among the shadows of the night.\"  &nbsp;  \"Now, in this moment. I don\'t know where you are. I imagine you doing so many things. I imagine you not remembering me.\"  &nbsp;  \"I saw you leave. Your footsteps drawing away from me. You entered your house as if you were leaving me forever. You were leaving me forever.\"  &nbsp;  \"I lost myself in the dreams we had. I gave myself up. Never again will we be able to dream. Never again.\"  &nbsp;  \"I see you. I don\'t know if this face is you or the imagine of you in my memory.\"  &nbsp;  \"So much to live for. So much to die for. If only my heart had a home.\"  &nbsp;  \"If you have no one, no one can hurt you.\"  &nbsp;  \"Suchte ich der Herrlichkeit der Zweisamkeit und fand nur der lÃ¤cherliche Einsamkeit.\"  (\"I searched the splendor of the companionship and found only the smiling loneliness.\")  Some of my own:  &nbsp;  \"Amor mÃ­o, eres tan frÃ­o como la luna, tan etÃ©reo como un dios y tan falso como mis sueÃ±os.\"  (\"My love, you are as cold as the moon, as ethereal as a god and as fake as my dreams.\")  &nbsp;  \"No estÃ¡s en mi memoria, ni lo estarÃ¡s. No estÃ¡s en mi presente, ni en mi futuro. No eres nada, tan sÃ³lo tristeza y soledad, la mÃ­a.\"  (\"You are not in my memory, neither you will be. You are not in my present, nor in my future. You are nothing, but sadness and loneliness, my own.\"  &nbsp;  \"..otra noche mÃ¡s de tantas, sin mayor valor que la de ayer, sin menor esperanza que la de maÃ±ana...\"  (\"... one more night, with no more worth than the last one, with no less hope than the next one...\")  &nbsp;  \"Â¿CÃ³mo puedes sostenerme siendo la brisa mÃ¡s sÃ³lida que tÃº? Â¿CÃ³mo puedes mantenerme viva siendo la muerte mÃ¡s real?\"  (\"How can you hold me if the breeze is stronger than you? How can you keep me alive if the death is more real?)  &nbsp;  \"Pensando en tÃ­ olvido todos mis pesares, pues tÃº eres el mÃ¡s grande de ellos.\"  (\"Thinking of you i forget every one of my sorrows, because you are the biggest of them.\")  &nbsp;  \"Es difÃ­cil sostener una promesa, pero mÃ¡s difÃ­cil es vivir con la culpa de haberla roto.\"  (\"It\'s hard to keep a promess, but is harder to live knowing it&nbsp;has been broken.\")  &nbsp;  \"DarÃ­a mi todo en soledad para ser la nada junto contigo.\"  (\"I would give my everything in solitude to be the nothing with you.\")  &nbsp;  \"Yo puedo pensar en una idea pero una idea no puede pensar en mÃ­. Yo puedo amarte, odiarte, idolotrarte y olvidarte, pero tÃº no puedes ni imaginarme.\"  (\"I can think&nbsp;about an idea but it can\'t think about me. I can love you, hate you, worship you and forget you, but you can\'t even imagine me.\")  &nbsp;  \"Te amo, te odio... y para tÃ­, ni siquiera existo.\"  (\"I&nbsp;love you, i hate you... and for you i don\'t even exist.\"  &nbsp;  I share these words with you (whoever \"you\" are) ... i hope you enjoy them as much as i do. By the way, i thought&nbsp;they would be&nbsp;a few, but&nbsp;there were too many of them that i like. &nbsp;  '),(21497,3602354,'08,July,2004','2004-07-08 00:00:00','   This was a nice day, i can\'t deny that... but i was really excited about that \"party\" (if i can call it so). Despite the fact i wanted to get a bit drunk hehe... i wanted to go there.. i don\'t really know why but i was excited about it. I went to the movies, again i got there before almost everyone... so i sent a message to that girl to see if she was going to stay at my house tonight... but oohh surprise!! There wasn\'t gonna be any party!!!! Damn it!! I wanted it!! Anyway, i put my disappointment aside, and waited for the others to come. After a while people started arriving... we saw the movie... not bad but not so good. I was a funny day... its really amusing to see people watching us like freaks hehehe really funny... just like scaring little boys haha poor of my children... that\'s why i\'m not planning to have any... hahaha. I\'m laughing a lot, and i\'m alone.. i may look like a madwoman.. but i don\'t care, i\'m worse than that hehe. Well i don\'t know anymore what i\'m talking about... i\'m just writing because of writing... I\'ll finish this here, before my friend falls asleep waiting for this to be published. \"Good night, good fight\" (I remembered this quote from Celebrity Deathmatch\", how great was that program hehehe the best of MTV-shit)                            THE END  '),(21498,3602354,'07,July,2004','2004-07-07 00:00:00','   It\'s been a while since i hadn\'t posted anything. Not because i don\'t have anything to say, it\'s just that my living schedule has been a little weird. I sleep early in the morning and wake up at midday, so everything changes for me... i write too much a couple of hours before sleeping and in the whole day i have no inspiration. Anyway, i said these where going to be some quick words, so be it.  Tonight another letter shall be written, for the Uncoming One. Tomorrow a movies-afternoon and later... some drinks.. hope so. Now, i gotta get started with the 3rd letter to... that guy... or whatever it is.. or may be... hmmm i\'d better shut up. See you later.  '),(21499,3602354,'13,August,2004','2004-08-13 00:00:00','   Hey you! You\'ve really let me down... somehow i thought you wouldn\'t do that to me, you looked so sure about not doing it, that i actually believed you... bad for me, i shouldn\'t have, i should have known this would happen... in fact i knew, but i just blinded myself from the truth. I wanted you to stick around forever, and be there when i needed you to be there... but i was so selfish, and now that i think deeply into it i see that i really don\'t want you to be there just for me, that\'s stupid... it\'s better (and healthier) for you to be there for somebody else. You\'ve been there for me too much time, more than anyone could bear, so i thank you for it from the heart of my rotten soul. I think it may be appropriate to ask you not to get stuck on me again, it\'s been too much for you and i don\'t want that to happen to us again (i hope you know what i am talking about) I think our friendship gave what it could, and it went as far as it could go. Thank you for everything, but just forget every feeling you may still have. \"Ich danke dir fÃ¼r all\' die Liebe ich danke dir in Ewigkeit!\"  '),(21500,3602354,'11,August,2004','2004-08-11 00:00:00','   Everything seems to go with the flow. Time passes by and looks back and sees me -the one behind everyone else-. He points at me and laugh... he goes faster and leaves me alone. So here I am, in the middle of the darkness of my mind, no warm no light... no time. Blessed by the demons of the night, cursed by the eyes of the Great One above me. I can see my fate beyond me, so hazy... so sad. I don\'t want to keep moving, i want to be here in this nowhere... in here i\'m safe, so far away from everyone. \"Oh Death! Thy power so great! Release me from my pain and lead me into the burning flames of endless peace! With thy unmercyful hands rip my heart to love no more! Tear my skin to shreds as i don\'t want to feel any more! Smash my body and let my soul be thine... forever\" In the darkness shall i remain until time comes back for me, to raise my body from the dead and give my soul new life... therefore i will reign over the shadows of my mind.  '),(21501,3602354,'06,August,2004','2004-08-06 00:00:00','   I don\'t need to know any kind of witchkraft to forsee this would happen... it always happen, even when nobody wants. When one of my friends, even my dearest and closest friends, get \"connected to someone\" i see myself moved aside. It is something inevitable, it is something that comes in the \"next-step-relation pack\", so what can i do but accept it and moving on. The only thing that bothers me the most is that it was such a nice... thing. Well i shall say no more, there\'s nothing bad about it. It was meant to be. And for the sake of every one of us it is better like that. It\'s not healthy to impede something that has been written in the pages of the fate... I\'m glad that it happened, i just wish the feeling to go away soon... before anything happens with me, in me. Things are better now, the sacrifice has been done, there\'s nothing left to do, except watch time go by...  '),(21502,3602354,'05,August,2004','2004-08-05 00:00:00','   Robots... since i remember people has always thought about robots and all those cybernetic issues. I don\'t like them at all, when i was a child i could, because they were something like fascinating to me, but now i really appreciate natural life. Could you imagine getting into a building and hearing a voice who welcomes you and tells you what to do next?? I would became absolutely insane!! Even more... is that posible? I would kill myself before getting used to it, some weird humanoid telling you what\'s best for you, or helping you with everything... aahh!! How useless would the human race become!! I understand now why the machines would raise up against humans... they would do everything, so they would controll everything. That would give them the right to rule over us... what a fucking mess!! I insist, i love living in this time... i guess it is better than how it may be.. or at least like they show it in the futurist movies... Anyway, i \'m writing stupidities, i don\'t even know well what i am writing.. i\'m almost asleep, so i shall better go to bed, before one creepy humanoid robot comes to tell me to go to bed, because it is not healthy to be awake at 1:39 a.m. See you, and beware of the robotic artificial life...  '),(21503,3602354,'02,August,2004','2004-08-02 00:00:00','   It\'s broken!!!!!! At last... i had to bear the truth: there was no way out, it was going to break anyway, no matter how much i take care of it... :\'( I was so sad.. more than 1 cm!!!!! I got a new record!! But i wanted it to grow more... or at least not to break so much!! Anyway, it has already happened... i\'ll just have to wait for it to grow again. So i hope i\'ll see it that long again in 3 months or so. Too sad but bearable... i guess. Well there\'s something good about fingernails, they\'re always growing...  '),(21504,3602354,'02,August,2004','2004-08-02 00:00:00','   Why do i feel? Why do i think? Which one is the best? Sometimes i believe i think too deep; some others i believe i feel too much. When i think i see only for myself, what\'s best for me, what makes me feel better and more comfortable... i think about  me . But when i feel i just worry about the other people involved, what\'s best for them, what may make them feel better and more comfortable... just worry about  them . So what the fuck is going on with me??!!! I know everything comes from the same place: my mind. But then, why is it so much trouble with two \"things\" which come from the same place?? Why can\'t i think and feel just what\'s necessary; think when i have to and feel when i want to. But life can\'t be so beautiful, nor my mind. I have to be complicated then i wouldn\'t be a woman, right?? Or rather wouldn\'t be a human. Anyway, this is me... this is my life, so i\'d better get used to it, used to the continuous fight between Feelings and Thoughts...  '),(21505,3602354,'02,August,2004','2004-08-02 00:00:00','     '),(21506,3602354,'02,August,2004','2004-08-02 00:00:00','   No words for it.... absolutely wonderful, magnificent.... i guess the best concert i\'ve been to! Just amazing... everything was great, unbelievable. Having this genious a few steps away from me was so exciting! The music!! The voice!! Everything was perfect!! But when they started with \"Ich verlasse heut\' dein Herz\".. my goddess!! it was incredible! I was shaking like if i was having an epilepsyc attack jeje i screamed like if someone was gonna rape me... well not really, no one has ever tried to rape me so i don\'t know how that scream would be... Anyway, that song... and the other... and everything was awesome... i loved that concert! Everytime i hear one of the songs they played i inmediately remember the whole scene and everything so beautiful... if i die one of these days i woulnd\'t do it so sad and miserable jeje i\'m getting better. Although, i don\'t want to die yet.. too many other concerts to go to  '),(21507,4224651,'18,August,2004','2004-08-18 00:00:00','   After a few months had gone by, and the dead of winter was now upon us, it was time for my father to fire up the coal stove and generate a little heat for us old-school New Englanders. The reason for the coal stove was that Dad never believed in electric heat because it was too expensive so he hired a service to come deliver loads of coal about 2-3 times throughout the season. It was effective enough, I suppose. Dirty, because one would have to shovel coal into the stove and afterwards, remove the ashes left by the coal that had already burned, but still effective. My father\'s only rule was this: \"Nobody shovels coal into the stove but me. If you\'re cold at night, take out another blanket from the closet and put it on.\" Needless to say, we spent many a cold night at Chez Loubier. But none so cold as the nights spent during that winter.  My mother had placed the candlesticks she got from that tag sale on the mantel above the coal stove. Sometimes she\'d take them down to burn a candle or two, but mainly they were put there for the \"ambiance,\" as Mom liked to say. I really don\'t think the ambiance she got was the kind of ambiance she intended to achieve. I think it started one night in December of \'84. And it scared the hell out of us. '),(21508,4224651,'13,August,2004','2004-08-13 00:00:00','   So we get to this tag sale, and there\'s about 40 people walking through and around makeshift rows of shit left by the person who lives in the filthiest-looking house I think I\'ve ever seen. The yard looked like it hadn\'t been mowed this century, the brown paint was flaking and peeling off of the front of the house. There was a maroon-colored car parked in the driveway (had to be a Datsun) and one next to it camping out on the grass (probably another Datsun). All around me I saw what to me looked like junk. A couch, an old bookcase, a rusty table with four equally rusty chairs, picture frames, a lava lamp (if you can believe that). There was some old pottery pieces that looked as if they had to be from the early 1900s. But in the end, it was the silver candlesticks that caught Mom\'s eyes. Not too much longer after we had gotten there, my mother spotted two identical candlesticks. Very antique. She was in to that kind of stuff. She liked the whole old-fashioned, rustic look about certain things. She was in the middle of a very intense examination of said candlesticks when the old woman walked over.  \"You like them?\" she asked. My mother was a bit easy to startle so the question made her jump a little. The old lady was short and walked with a cane. Her shoes looked like the kind you wear when you\'re in a hospital and her dress resembled those smocks you used to wear in 5th grade art class. Those smocks sucked. I hated those friggen things.  \"They\'re beautiful,\" Mom replied. \"Are they old?\" \"Let\'s just say I got them when I was about your age.\" She spoke very slowly, almost with a natural stutter, but one could tell it was because of her age. She didn\'t wear any glasses which I thought was weird because I just assumed that all old people wore glasses to help them see. All I can remember thinking is maybe this old bag had some kind of Superman eyesight. My father and I poked around at some other junk. Dad was basically trying to appear as if he was really looking around for something while I was busy trying to find out if there were any G.I. Joe\'s in this pile of garbage.  \"There\'s no price on them, are they free?\" Mom laughed as she asked this. Always the packrat, trying to snag a freebie. The old lady, however, who seemed pleasant from the beginning, did not find this funny. \"Oh, I must have missed those,\" the old lady said. She reached out slowly and took one of the candlesticks from my mother\'s hands and held it in her own. She stared at it for a few seconds as if waiting for the price to slap her in the face. But to my disappointment (and my father\'s), there was no magical force leaping at her face and slapping the old lady in the mug. Not on this day at least.  \"Give me three dollars and you can take \'em,\" the lady said after a while. Mom reached into her purse and pulled out a five dollar bill and handed it to the old woman. Again, my mother attempted to make nice with the lady. \"Take the whole five,\" my mother said with a smile, and the old lady, indifferent in response to this gesture, turned and slowly walked away. Perhaps toward wherever she was keeping the loot she made off this tag sale. Meanwhile, my mother turned with a what-did-I-do look on her face and motioned to my father to let him know \"we\'re leaving now.\"  That was about all the time we spent at the old lady\'s tag sale. How odd, I thought, that she did not have an old man accompanying her at her home? In fact, I didn\'t see anyone there who looked like they were helping her sell her old junk. Maybe her husband passed away some time ago. The thought left my mind about as quickly as we made it to the interstate.  It was a long ride home on I-91 as I felt we had accomplished nothing (no G.I. Joe\'s, no video games), but my mother was happy with her candlesticks so I suppose one could say \"mission accomplished.\" She had no idea how soon she\'d be getting rid of them. '),(21509,4224651,'13,August,2004','2004-08-13 00:00:00','   It was really just an ordinary October day. The leaves were changing (actually they had already changed, now they were falling), the grass stopped growing for the year and browning in some spots, and from the highway, as I watched from the backseat of my Dad\'s old Mercury Cougar, all the houses in the town below had wisps of smoke coming from the chimneys. It reminded me of a valley full of candles that had just been blown out.  It was 1984, and I was tagging along with Ma and Pa Dukes for the day (had no choice, I was 5 years old). Apparently, someone out in Greenfield, Massachusetts was foolish enough to have a tag sale (or a yard sale, or a rummage sale, depending on what part of this lovely country you\'re from) on this brisk Saturday morning in October. So naturally, it\'s my mother\'s job to drag her husband and youngest son off through the wilderness to pick through someone else\'s crap just in case there\'s something of value to her.  Mom was a bit of a talker as well. She\'d chew my Dad\'s ear off the entire day and he would just turn his head and smile then and again just to show her he was listening. Now and then he would throw out the occasional, \"hmm,\" or \"oh yeah?\" but for the most part, he\'d pretty much stare straight ahead and watch the road in front move toward him at a steady 55 mph clip. Someone once said my mother could talk a dog off of a meat wagon. I didn\'t understand what that meant (again, I was five).  From the backseat, in between battles with Shredder and the Ninja Turtles, all I could offer was the frequent, \"how much longer?\" \"Ten more minutes,\" my mother would say. Mom was infamous for replying with \"ten more minutes.\" Ever notice how every time you asked your parents how much longer, it was always \"ten more minutes?\" It didn\'t really matter where you were going. You could have been 2 hours from the destination, but it didn\'t matter. I think every time I was told \"ten more minutes\" it had already been twenty minutes from the last time I\'d asked.  About 45 minutes later, we arrived at the tag sale. '),(21510,4224651,'13,August,2004','2004-08-13 00:00:00','   I can\'t remember how many times I\'ve been told this story. I used to think it was to keep my brothers and I from staying up late at night after my parents went to bed so we could watch old Family Ties reruns or soft-core porn on HBO. But you know how you start to believe things after you\'ve heard them spoken of after so many times? Well, that\'s why I believe this story. That and the fact that I\'m a grown man now and my parents still swear that the following events did indeed happen. After all, they never gave me a reason not to believe them. And I never gave them a reason to tell stories. None like this... '),(21511,4206637,'17,August,2004','2004-08-17 00:00:00','   urlLink Turkmen Leader Orders People to Stop Chewing Chicken Crap - NEWS - MOSNEWS.COM   Wow! There are many people who chew chicken excrements. Unbelievably! It\'s appeared to be a very common habbit in Central Asia. '),(21512,4206637,'13,August,2004','2004-08-13 00:00:00','   urlLink German Chancellor Adopts Russian Girl - NEWS - MOSNEWS.COM   How lucky the girl is.  I\'d like to extend my respect and congratulations to Shroeder\'s family for managing to go through all the bureaucracy barriers that Russian goverment sets in front of the people who want to adopt children. Taking into consideration how many orphans there are in Russia and what frightining conditions they live in it is hard to understand why our goverment does everything to complicate the adoptation process. Right now only seriously sick children can be adopted by a foreigner with relative ease. Everyone else has to spend their childhood in run-down orphan homes, often exiting them with grave psychological problems. '),(21513,4206637,'13,August,2004','2004-08-13 00:00:00','   Every year  an  urlLink International Air Show   takes place on Zhukovsky aerodrom, 40km away from Moscow.  Zhukovsky was founded in 1947 as a closed military town with secret aerodrome where most of aircraft experiments and tests took place. The conspiracy was so high that the railroad station of the town was named Otdih (Rest), to pretend that it is another dacha\'s area in suburbs of Moscow. The funny thing is that the station is still called Otdih and even more, in description of how to get to the show the organizers don\'t mention station Otdih the nearest to the aerodrome. Conspiracy is still on a level.  If the weather won\'t spoil i highly recommend visiting the show. Although there are always crowds of people everything is very well organized and flights are breathtaking. '),(21514,4206637,'12,August,2004','2004-08-12 00:00:00','   urlLink Police Operation Interrupts Pirated CD Network   It is like a drop in the ocean.  Targeting individual pirates only helps their competitors, without actually causing any decrease in operation volumes.  Most of CDs and DVDs that are sold in Russia are pirated. You can find them in kiosks near every metro station in Moscow. Even in large music stores one finds shelves with pirated discs.  St.Peterburg is quite famous for very cheap pirated discs. People from Moscow go for one day trip just to buy CD\'s or you can order them in Internet and they will be delivered by mail.  Needless to say, that it takes most serious involvement of law enforcement bosses\', to protect this industry from any substantial damage, despite its high visibility to whom it may concern. '),(21515,4206637,'11,August,2004','2004-08-11 00:00:00','   The following 8 news articles made for most popular reading among MosNews.Com visitors in August 2004:     urlLink Winged Cat From Hell Put to Death in Central Russia  â€” July 27th    urlLink Russian Researchers Say Debris of Alien Spaceship found in Siberia  â€” August 10th    urlLink Russian Museum to Exhibit Rasputinâ€™s Penis  â€” April 28th    urlLink Boy Raised by Dog Found in South Siberia  â€” August 3rd    urlLink Shark Attack Nearly Costs Far East Man a Leg  â€” July 30th    urlLink Tennis Star Myskina Sues GQ Over Topless Photos  â€” August 9th    urlLink Russian Orthodox, Muslims Battle Mormon Building in South Russia  â€” August 9th    urlLink Spacecraft Maker Says Russia No Longer to Carry Supplies to ISS for Free  â€” August 4th  The list will be updated, as soon as new articles attain comparable popularity with those listed above. '),(21516,4206637,'11,August,2004','2004-08-11 00:00:00','   urlLink Russian-Israeli Mafia Boss Murdered Near Moscow  â€” MosNews.Com reports. This poor kid had once made headlines, becoming the first and the last Russian mobster, who was extradited from Israel on Russian authorities\' request. Needless to say, it was extremely easy for Mr. Zhuravlyov to buy his way out of Russian jail: there is no such thing as juror protection in Moscow, so if any gentleman on that jury could not be bought outright, he could be easily and cheaply intimidated by thugs working for the defendant.  It proved a more difficult task, however, to obtain adequate protection from Mr. Zhuravlyov\'s own colleagues. Unsurprisingly for a man of his trade, he used to have too many enemies, and they proved more efficient in dealing with him, than both local police and the courts.  May we remain in peace, after the passing of this belligerent compatriot of ours. '),(21517,4206637,'11,August,2004','2004-08-11 00:00:00','   Roman Abramovich, of Chelsea fame,  urlLink has reportedly set his sights on an Italian resort , where Persey Bysshe Shelley used to drown back there in his young days.  Eco del Mar might be quite a distance from Chukotka, of which Abramovich is the guv, but so is the borough of Chelsea, so who cares.  What\'s interesting, though, is that we keep hearing about all sorts of Abramovich acquisitions lately, and some 80% of these rumours lately prove totally unfounded. It seems like there\'s someone behind the stories â€” obviously a ressourceful someone, who badly needs to infuriate both Roman\'s poorer compatriots and his Kremlin partners with all those headlines.   If this plot ever succeeds, Abramovich will have to chose between the paths of his two ex-partners: Mr. Berezovsky, who fled to London, and Mr. Khodorkovsky, who chose jail in his hometown. Regardless of Roman\'s ultimate choice (I\'d bet on him fleeing, 100:1), his carefully amassed wealth, worth billions, will come up for grabs, and our Ressourceful Someone expects himself to be among the first invited to the looting festival... '),(21518,4312005,'24,August,2004','2004-08-24 00:00:00','   urlLink    Albanians or what?                                          From left to right : Mirela Manjani (alias Mirella Tzelili), Pirro Dhima (alias Pyrros Dimas) and Luan Shabani (alias Leonidas Sampanis) '),(21519,4312005,'23,August,2004','2004-08-23 00:00:00','   Olympics are being held in Greece, that\'s one of our neighbors, one of our bordering countries. Albanians have had a long big time, mainly informal conflict with Greeks. Itâ€™s all about territorial subjects; Greeks claim that actually the southern part of Albania is Greek, while Albanians (Tchams) were kicked out of northern Greece with their houses burnt 100 or so years ago.  Relationships between Albania and Greece looked like were going to get better after the elections in Greece, because the new-coming Greek Prime Minister claimed that Albanian immigrants (theyâ€™re a lot) would have a better treatment after he came in power. He came in power but there was not any big change.  So what does it have to do with Olympics?  One Albanian athlete and two Albanian weightlifters played as Greeks at the current Olympics, as they have been doing for 10 years by now. You have Mirela Manjani, that holds (till 1 year ago, I donâ€™t know about now) the Pike-throwing world Record. Pirro Dhima (naturalized as Spyros Dimas in Greece), who won some weightlifting gold medals in the last two Olympics and won bronze this year. And last but not least, Luan Shabani (Leonidas Sabanis I think :D) who is also a great medal holder, but who  unfortunately  was disqualified for doping (everyone was happy in Tirana about that).  Yes, Greece offers them much more money that Albania does. Yes, Greece gives them much better ways to train that Albania does. Yes, Greece gives them the Greek nationality which is much more useful that the Albanian one. But that does not mean you have to revolutionize your nationality. Mr. Dimas stated that he is proud to be Greek; reality check Spyros, you were born in Albania, like it or not you had and will have Albanian blood in your body for all your life, you are Albanian.  In the other side you can see Slave athletes playing for Greece. Well they never say they are Greek. Actually they always say they are from Serbia or Macedonia or every other country they are from. They accept the fact that they are playing for Greece, and they like it, but they donâ€™t fuck with their nationality.  Last paragraph is about the Greek team that takes part in the Olympics that are being held in Athens (Greece). Out of 8 weightlifters that take part in the Greek team, only 2 of them are Greeks. And what about the Greek baseball team? There is no one there who was actually born in Greeceâ€¦That makes you think :)!  Thatâ€™s it for now!  j. '),(21520,4312005,'23,August,2004','2004-08-23 00:00:00','   urlLink    \"The block\" during daytime... '),(21521,4312005,'23,August,2004','2004-08-23 00:00:00','   Itâ€™s an original title, isnâ€™t it?  Well the thing is this, people in Albania, and mainly in the Balkans consider drinking coffee a sport :). Every teenager has more-or-less the same agenda during summer afternoons. At about 18:00 you start getting yourself prepared for going out. You will need 15 minutes if you are a male and an hour or so if u are a female.  Than you go out of your home, take either a bus or a taxi, and go to \"The Block\". Yeah it looks like a fancy name, actually it was the place where communist leaders had their homes during the communist regime here, today it is totally transformed..! After you arrive at \"The Block\" and meet your friends, you go to your favorite bar/coffee-shop/cafÃ© and stay there for at least 1 hour and a half. And what do you do? Just chat!  So how will I survive without my 1~2 daily coffee-drinking hours? Where will I find a place to spend at least half an hour drinking a damn cup of coffee so I will feel like home...Will I really have time for that?  j. '),(21522,4312005,'23,August,2004','2004-08-23 00:00:00','   Hi Everyone (anyone)!      It\'s 3:53 in the morning at this moment, and I just decided i have to start a blog! I live in Tirana, Albania. I have been living here for 6325 days, 151800 hours...I was actually born here :)! Well here it comes, i\'m moving to Toronto, Canada the 8th of September...  This is just the first post but i would like to say so many things. I\'m sure a lot of people will not consider my moving a big deal. I think that\'s normal. At this precise moment, i don\'t consider it a big deal neither so your question will be: -\"Than, why are you creating this blog?\". Well the answer for me is simple. I\'m sure that when i will actually go to Toronto and settle there my feelings, my emotions, my thoughts about this moving will change.  I love Tirana. I love Albania. I love Albanian. I love Albanians. But i\'m bored of each of them.  So the imminent result of this blog will be to describe my last 14 or so days in Tirana, and than my new life in Toronto.  That\'s it for the first post!  j. '),(21523,4086028,'31,July,2004','2004-07-31 00:00:00','   in inquisitionibus adventis tibi fortuna secunda sit vivite felices... et tibi parta quies  Two roads diverged in a yellow wood, And sorry I could not travel both And be one traveller, long I stood And looked down one as far I could To where it bent in the undergrowth;  Then took the other, as just as fair, And having perhaps the better claim, Because it was grassy and wanted wear; Though as for that the passing there Had worn them really about the same,  And both that morning equally lay In leaves no step had trodden black. Oh, I kept the first for another day! Yet knowing how way leads to way, I doubted if I should ever come back.  I shall be telling this with a sigh Somewhere ages and ages hence: Two roads diverged in a wood, and I --- I took the one less travelled by, And that has made all the difference.  '),(21524,4086028,'30,July,2004','2004-07-30 00:00:00','   I\'m Daniel. Having read Vanna\' s post I felt it my moral obligation to reassure my ego that my mind hasn\'t become redundant. Unfortunately there\'s no alias, pseudonym, and low and behold no AKA in the name Daniel. I\'m just not that cool.  Four weeks from now we\'ll have been through Beijing and have settled into the \'cosy\' flat that will embrace us for a year. That is to say we\'ve been updated to the role of teacher, transitioning across the great expanse that is the void between the teacher and the taught. Now weigh yourself down with the slight discomfort of only being able to say hello, ni hao.  For the next three weeks I\'ll be living where&nbsp;I have been for the majority of my life. Muswell Hill, London. In terms of bags, we\'re talking the mutts.  Waigouren in Zibo simply refers to the status I will adopt on stepping foot in China. Quite simply, you are different. You are a foreign Devil.  This may be my first and last post, but I\'m not generally one for computers.   I hate to see you go, but I\'d love to watch you leave&nbsp; - Nick Cage  '),(21525,4086028,'30,July,2004','2004-07-30 00:00:00','   Zibo Institute of Foreign Affairs.  My main work is teaching English to college students. All I have to accredit my status is a yellow card which reads  \"This card confirms that the carrier is a bona fide PROJECT TRUST VOLUNTEER\".  That and my ununique enthusiasm for something different, that selfish desire to feel like you have deserved. I\'m not that different from my friend Vanna, but we will come away with two different tales. This is your opening to shimmy to the left :  urlLink http://www.thoughtsfromchina.blogspot.com/  I\'m not under the illusion that my going away will benefit anyone more so than it does me, but in the last twenty years more than two hundred and fifty different regions and cities have fallen suddenly open. All the people I\'ve spoken to have opinions that are not far-removed from my own. Would it be insulting of me to believe everything that I hear? Initially I was seized by the foolish idea that each one of them would withhold some secret from me - some simple, perfect illumination. Because that is the foreigner\'s obsession in China. A billion uncomprehended people. I don\'t want to fall suspect to the nagging feeling that I can. I don\'t want to teach for a living, I want to learn. China, teach me.  '),(21526,4086028,'30,July,2004','2004-07-30 00:00:00','   urlLink    muswell hill  urlLink'),(21527,4086028,'09,August,2004','2004-08-09 00:00:00','   Morning Glory&nbsp; urlLink     urlLink  '),(21528,4086028,'09,August,2004','2004-08-09 00:00:00','   Stealing a thought from one of my closest friends  \"It\'s one thing to go away for a couple of months but a whole year is some serious time, especially considering life-changing decisions and events can happen in a fraction of a second.\" It\'s wierd how a \'small\' statement can bring together things that have shied away from the frontal lobe.  Revalation seems a bit heavy - more an encroaching sense of recognition. Even more daunting are the times  a journey comes to an end. Often your first fledgling steps can appear cumbersome on hindsight, but given the opportunity would you change them? Having reached the end of the path to China my mind is unwilling to take in that there is no longer the adventure of reaching the adventure. This  is  the adventure. It\'s odd when you saunter back and observe just how institutionalised you can become into the life that you know. When you see that every little habit that you have has evolved from some small niche unbeknownst to you, how can you even tell what will happen when the safety net dissolves.  '),(21529,3368285,'28,May,2004','2004-05-28 00:00:00','   i guess i\'d say one of the stupidest defense mechanisms that we ever made was intolerance... me , i\'d rather hurt than be hurt... that leads to being hurt more... a very vicious cycle. i wish mine would end... i\'d have better relationships then. i really wish i could go back in time... specifically a month ago... everything was calmer. i hate my freakin pride. sometimes i just hope i could be peaceful, even for once... haa, that is a load of bull...   '),(21530,3368285,'27,May,2004','2004-05-27 00:00:00','   as i was on my way to work, this old lady noticed that i was drenched in sweat... she offered me tissue. i didn\'t know her, she was just a face in the throng of strangers . imagine if everyone would just be like that. we start off being inherently kind, but somewhere along the way , we get jaded. that is so so sad, we should never compromise principles. they\'re the only thing we can hold on too in this life. try to search for that which makes you happy... and content.   '),(21531,3368285,'27,May,2004','2004-05-27 00:00:00','   i guess mine is really eating me up. had three hours of sleep, almost no food... and probably have been staring into the ceiling most of the day. it really does hurt. and i thought i was strong , haa. i guess that wasn\'t true. i really would do anything, anything... ha haa, i guess. but of course , rationality and logic should always be a part of everyone\'s decision making process. somehow i wish i wasn\'t so rational at all...   '),(21532,3368285,'25,May,2004','2004-05-25 00:00:00','   \"sorry . i don\'t know why this is happening. all i know is, i miss the times that we were happy. we were. i guess the situation really wasn\'t on our side. i hope i could just go back to the days when you kinda missed me too. you know, it\'s so funny that i know i need you now more than ever but i have to go because you want me too. i want you back , but i don\'t know how. i need you that much i know. again i am sorry, for i can\'t help myself missing you...\"   i recalled this letter. i probably just don\'t know ... or maybe i know better.   '),(21533,3368285,'24,May,2004','2004-05-24 00:00:00','   hah, finally, i am cynical. another company called me up... gave me a phone interview, and guess what? the cynicism in the air was as thick as the smog in san fran. i thought nothing would ever make me feel that way.sigh, i wish everything was as before... everything was calmer, more serene... ever thought of reading catcher in the rye 12 times straight? that should give you an idea...   '),(21534,3368285,'24,May,2004','2004-05-24 00:00:00','   stupidity is the theme. in my work, the people i know... i guess for everyone else. i am not self righteous or anything but i just can\'t help notice that. i hope this stupid phase would just end. it\'s so like a contagious disease.... and life goes on....   '),(21535,3368285,'21,May,2004','2004-05-21 00:00:00','   my last post was probably portentious. the time has come to make a decision. and i have made mine. i can\'t honestly say this would not pain me... i am choosing the least painful path. at least i\'ll have more time for me, and my son... i can\'t help but just feel envious of people who could be so very selfish. i wish i could be .... my day started when it ended. i raise a glass to love , for once again it has won the day...   '),(21536,3368285,'20,May,2004','2004-05-20 00:00:00','   familiar? of course... choices are hard to make. some lead to eternity, and some lead to ruin. some people choose using their reason, and some by emotions. i used reason a lot in the past. but now, i don\'t know... perhaps this is just a phase. did someone really say i\'d rather be a king in hell than be a servant in heaven? i guess so. the choices ones make are a measure of their soul... i wish when the time comes, mine would shine, as bright as the sun.   '),(21537,3368285,'19,May,2004','2004-05-19 00:00:00','   called and talked to her ... i really notice how my heart would go thump-thump when i\'m with her. also this mona lisa smile that creeps on my face , ha haa. maybe it really is true that love is temporary insanity. it went well... she is happy , i am happy , i guess. that\'s all that matters now , does it? i hope things would always stay this way. but , knowing me... i\'d probably start doing something to ruin something perfect, ha haa. but i do want to stay in this temporary bliss forever.   '),(21538,3368285,'19,May,2004','2004-05-19 00:00:00','   sticking with my no angst rule... is great here. i guess i\'m wasting my talent here, sitting down. working from 9-8. i leave home 6 even... probably the only reason i\'m here is her , ha haa... lately my routine was work, home, play with my son for an hour or so, wake up ; wash, rinse, repeat. i\'ve gotta do something about my sleeping habits. have been sleeping for only 6 hours tops lately.   '),(21539,3368285,'18,May,2004','2004-05-18 00:00:00','   it\'s raining hard. cold too, perhaps too cold. i can\'t help myself from thinking of her, dang... there is something amiss. some psychoanalists say that when you fall in love you suffer from temporary insanity, haa. probably the control-freak side of me just can\'t accept that. ha haa, i kinda smile when i can\'t stop thinking of her with me , through rain or sunshine. i pity people who love. oh yeah, here\'s something to think of \"the greatest thing you\'ll ever learn; is to fall in love , and be loved in return\"   '),(21540,3368285,'18,May,2004','2004-05-18 00:00:00','   called her up she was asleep. i told myself i would never express any angst. but why do i feel this way? i really miss hearing her voice... i really don\'t know what\'s happening, feel like i\'m falling too fast, too hard. i thought i knew myself. a certified control freak, i never should have let it come this far... heck i can\'t even decide if i want to go through with this or not... it\'s as if she\'s slowly taking over me, ha haa... i kinda missed this...   '),(21541,3368285,'18,May,2004','2004-05-18 00:00:00','   she hasn\'t e mailed me. this is the first time this happened . her reason being she was busy. somethings afoot. called and talked to her on the phone though, she asked me to go with her to a job interview monday. let\'s see what happens then...   '),(21542,3368285,'17,May,2004','2004-05-17 00:00:00','   okaay, i met this girl. she works for the e mail arm of our company. initially it was just small talk through im. message here and there. which ended in me walking her home twice, seeing here at her office once, dinner (if you would call it) once, a stuffed toy and a boquet of stargazers. i start falling for her, maybe coz she bared out her soul, and started talking about personal stuff. and she is so responsible, something i\'ll never be. and yes opposites do attract. i guess she is my exact opposite in every way which makes her even more enticing. i hope you get the picture because i would never delve into any particulars. and it is SO SO hard to talk without specifics. let me see, fri we had a tiff... haven\'t talked to her since yesterday. where we sorted out our differences. and , apparently were making progress. she is something else, i am falling for her... i hope i could weather this though....   '),(21543,3368285,'17,May,2004','2004-05-17 00:00:00','   how apt to start my blog with hi. and probably introductions are in order as well. i am francis nor capule, a heterosexual single dad of a three year old son that works for a call center somewhere in makati. twenty seven years old, single, poetry dabbler , conversationalist , and bull-shitter-par-excellance. i even tried selling the sun to one my former trainors, which i did , technically, he hee. i\'m gonna have to think of ways so my blog doesn\'t become an angst filled, psychoanalyst\' treasure trove of discontent. welcome to my world...   '),(21544,3368285,'10,June,2004','2004-06-10 00:00:00','   i came across this on a forum i frequented  I hate the way you talk to me and the way you cut your hair.  I hate the way you drive my car.  I hate it when you stare.  I hate your big, dumb combat boots.  And the way you read my mind.  I hate you so much it makes me sick â€“  It even makes me rhyme.  I hate the way youâ€™re always right.  I hate it when you lie.  I hate it when you make me laugh â€“ Even worse when you make me cry.  I hate it that youâ€™re not around.  And the fact that you didnâ€™t call.  But mostly I hate the way I donâ€™t hate you  Not even close, not even a little bit, not any at all...   yes, it\'s from \"10 things i hate about you\" kinda made me smile, and reminisce... but things has been set in motion that couldn\'t be taken back... everything is so surreal now, it\'s like i\'m living a day-to-day existence... this is the best recourse, the best option... when would i find a unicorn to call my own?   '),(21545,3368285,'07,June,2004','2004-06-07 00:00:00','   attending job offers from companies lately. might be i\'m looking for growth? one of the companies is promising though... 4th in it\'s industry... not bad.... i have to take control... focus , focus... in the end everything will pay off.   '),(21546,3994022,'31,July,2004','2004-07-31 00:00:00','   Dragon Boat Festival  Saturday I attended the  urlLink Dragon Boat Festival  at Sloans Lake in Denver. Not far from my home the event is Colorado\'s premiere showcase of pan-Asian culture, combined with a unique, thrilling athletic competition -- Dragon Boat Racing, an ancient sport that\'s catching on worldwide.  This is one of the boats used in the race.    Some children cooling off near the boats.    The Dragon!      Fixture of the Sloans Lake Neighborhood.    A perfect end to a great evening.  '),(21547,3994022,'30,July,2004','2004-07-30 00:00:00','   Fly Fishing on the Big Thompson  Friday I took a day of rest and headed up to Estes Park for some fly fishing. About an hour north of Denver on the outskirts of Rocky Mountain National Park is the Big Thompson River. It was sunny and cool at 7am and there was little to no activity on the surface. In the first hour I landed a Rainbow on a RS2 Nymph.    Next came this lovley Lake Trout, I think, on a RS2.    After 10am I switched to a Blue Winged Olive dry fly as there was some surface activity. I landed two Browns near the Dam in the next 2 hours. All in all it was a relaxing day.    A herd of Elk decided to cross the river several yards in front of me. All protected by this big buck.  '),(21548,3994022,'22,July,2004','2004-07-22 00:00:00','   Google, the greatest search engine ever engineered, shows us what is was like back in the day. Photo courtesy of  urlLink fury.com .  urlLink Google Circa 1960 . Please allow 4-6 weeks for results!  '),(21549,3994022,'21,July,2004','2004-07-21 00:00:00','   Andrew Anderson is now a U.S. Army Captain.  For those of you who have not heard Andy Anderson was promoted to the rank of Captain after returning from his tour in Iraq. Congratulations.  '),(21550,3994022,'21,July,2004','2004-07-21 00:00:00','   Dad to comatose son: I\'ll buy you a Corvette if you wake up  Jason Stackiewicz\'s $75,000 new car will arrive at his home in September.  urlLink Boston Globe  '),(21551,3994022,'21,July,2004','2004-07-21 00:00:00','   The 4th of July in Colorado Springs.  Linda, Shawn and Scot headed down to Colorado Springs to visit Karen Andy, Bree and Katy for the 4th of July. It was good to see them all and relieved to know that Andy is  safe and back from Iraq.  '),(21552,3994022,'21,July,2004','2004-07-21 00:00:00','   Hi-Just signed up.&nbsp; At work in union office. &nbsp; Shannon at home and helping mom with garage sale they made 86.00 yesterday.&nbsp; Sale on today as well but only until noon.  '),(21553,3994022,'21,July,2004','2004-07-21 00:00:00','   I hiked the Herman Gulch trail on Sunday and did a little Fly Fishing. Herman Gulch is located in the Arapaho and Roosevelt National Forests near Idaho Springs. The trail is 5 miles roundtrip and starts at an elevation of 10,400ft and ends at 12,705ft at Herman Lake.    A nice sized pool, formed by a beaver damn was my first query and there were 10-15 small trout feeding on the surface. Here I landed a litte Greenback Cutthroat in a matter of seconds.   Several minutes later I landed this tiny Brooke Trout. Small but damn tough when death is on the line. Literaly.  '),(21554,3994022,'21,July,2004','2004-07-21 00:00:00','   My company,   urlLink threefour studios  just launched a website for the real estate development firm, Dakota Ridge Development.  urlLink Check it out.  '),(21555,3994022,'21,July,2004','2004-07-21 00:00:00','   If you want to put a picture up on one of your messages - you can upload one onto the internet for free here:  urlLink http://www.imageshack.us/ &nbsp; Then you can paste the link onto your message&nbsp;like this&nbsp;:  &nbsp;  You don\'t have to sign up or anything! I did it!&nbsp;Then I&nbsp;lost&nbsp;40 inches on my waistline!&nbsp;   See!:::  '),(21556,3994022,'21,July,2004','2004-07-21 00:00:00','   urlLink Who\'s taking swear words out of \"Murder, She Wrote\" books?  An unknown culprit in Mormon country has crossed out the nasty words and replaced them with milder words, like \"darn,\" \"gosh\" or \"heck,\" written in black, purple, green and even pink ink. One library patron says: \"It bothers me \'cause I\'m trying to read a book. It\'s distracting.\" (Deseret News)  '),(21557,3994022,'21,July,2004','2004-07-21 00:00:00','   Name this blog. What should we call this \"web log\"? You decide, submissions due immediately!  urlLink Email me.  '),(21558,3994022,'11,August,2004','2004-08-11 00:00:00','   A birthday gift from the Mrs.  Craig Sween on Lake Michigan with a fat Chinook Salmon. Mmmmm, Mercury!  '),(21559,3994022,'11,August,2004','2004-08-11 00:00:00','   He likes mousse.  Proclaims this young food experimentor. Silk Tofu + melted chocolate chips + blender + 30 minutes in the fridge = mmmmmmmmm good!  '),(21560,3994022,'11,August,2004','2004-08-11 00:00:00','   A true Texas Cowboy and American . Not to mention faux President.  urlLink Will Ferrel\'s take on George W. Bush.  '),(21561,3994022,'31,July,2004','2004-07-31 00:00:00','   Dragon Boat Festival  Saturday I attended the  urlLink Dragon Boat Festival  at Sloans Lake in Denver. Not far from my home the event is Colorado\'s premiere showcase of pan-Asian culture, combined with a unique, thrilling athletic competition -- Dragon Boat Racing, an ancient sport that\'s catching on worldwide.  This is one of the boats used in the race.    Some children cooling off near the boats.    The Dragon!      Fixture of the Sloans Lake Neighborhood.    A perfect end to a great evening.  '),(21562,3994022,'30,July,2004','2004-07-30 00:00:00','   Fly Fishing on the Big Thompson  Friday I took a day of rest and headed up to Estes Park for some fly fishing. About an hour north of Denver on the outskirts of Rocky Mountain National Park is the Big Thompson River. It was sunny and cool at 7am and there was little to no activity on the surface. In the first hour I landed a Rainbow on a RS2 Nymph.    Next came this lovley Lake Trout, I think, on a RS2.    After 10am I switched to a Blue Winged Olive dry fly as there was some surface activity. I landed two Browns near the Dam in the next 2 hours. All in all it was a relaxing day.    A herd of Elk decided to cross the river several yards in front of me. All protected by this big buck.  '),(21563,3994022,'22,July,2004','2004-07-22 00:00:00','   Google, the greatest search engine ever engineered, shows us what is was like back in the day. Photo courtesy of  urlLink fury.com .  urlLink Google Circa 1960 . Please allow 4-6 weeks for results!  '),(21564,3994022,'21,July,2004','2004-07-21 00:00:00','   Andrew Anderson is now a U.S. Army Captain.  For those of you who have not heard Andy Anderson was promoted to the rank of Captain after returning from his tour in Iraq. Congratulations.  '),(21565,3994022,'21,July,2004','2004-07-21 00:00:00','   Dad to comatose son: I\'ll buy you a Corvette if you wake up  Jason Stackiewicz\'s $75,000 new car will arrive at his home in September.  urlLink Boston Globe  '),(21566,3994022,'21,July,2004','2004-07-21 00:00:00','   The 4th of July in Colorado Springs.  Linda, Shawn and Scot headed down to Colorado Springs to visit Karen Andy, Bree and Katy for the 4th of July. It was good to see them all and relieved to know that Andy is  safe and back from Iraq.  '),(21567,3994022,'21,July,2004','2004-07-21 00:00:00','   Hi-Just signed up.&nbsp; At work in union office. &nbsp; Shannon at home and helping mom with garage sale they made 86.00 yesterday.&nbsp; Sale on today as well but only until noon.  '),(21568,3994022,'21,July,2004','2004-07-21 00:00:00','   I hiked the Herman Gulch trail on Sunday and did a little Fly Fishing. Herman Gulch is located in the Arapaho and Roosevelt National Forests near Idaho Springs. The trail is 5 miles roundtrip and starts at an elevation of 10,400ft and ends at 12,705ft at Herman Lake.    A nice sized pool, formed by a beaver damn was my first query and there were 10-15 small trout feeding on the surface. Here I landed a litte Greenback Cutthroat in a matter of seconds.   Several minutes later I landed this tiny Brooke Trout. Small but damn tough when death is on the line. Literaly.  '),(21569,3994022,'21,July,2004','2004-07-21 00:00:00','   My company,   urlLink threefour studios  just launched a website for the real estate development firm, Dakota Ridge Development.  urlLink Check it out.  '),(21570,3994022,'21,July,2004','2004-07-21 00:00:00','   If you want to put a picture up on one of your messages - you can upload one onto the internet for free here:  urlLink http://www.imageshack.us/ &nbsp; Then you can paste the link onto your message&nbsp;like this&nbsp;:  &nbsp;  You don\'t have to sign up or anything! I did it!&nbsp;Then I&nbsp;lost&nbsp;40 inches on my waistline!&nbsp;   See!:::  '),(21571,3994022,'21,July,2004','2004-07-21 00:00:00','   urlLink Who\'s taking swear words out of \"Murder, She Wrote\" books?  An unknown culprit in Mormon country has crossed out the nasty words and replaced them with milder words, like \"darn,\" \"gosh\" or \"heck,\" written in black, purple, green and even pink ink. One library patron says: \"It bothers me \'cause I\'m trying to read a book. It\'s distracting.\" (Deseret News)  '),(21572,3994022,'21,July,2004','2004-07-21 00:00:00','   Name this blog. What should we call this \"web log\"? You decide, submissions due immediately!  urlLink Email me.  '),(21573,3994022,'11,August,2004','2004-08-11 00:00:00','   A birthday gift from the Mrs.  Craig Sween on Lake Michigan with a fat Chinook Salmon. Mmmmm, Mercury!  '),(21574,3994022,'11,August,2004','2004-08-11 00:00:00','   He likes mousse.  Proclaims this young food experimentor. Silk Tofu + melted chocolate chips + blender + 30 minutes in the fridge = mmmmmmmmm good!  '),(21575,3994022,'11,August,2004','2004-08-11 00:00:00','   A true Texas Cowboy and American . Not to mention faux President.  urlLink Will Ferrel\'s take on George W. Bush.  '),(21576,3994022,'31,July,2004','2004-07-31 00:00:00','   Dragon Boat Festival  Saturday I attended the  urlLink Dragon Boat Festival  at Sloans Lake in Denver. Not far from my home the event is Colorado\'s premiere showcase of pan-Asian culture, combined with a unique, thrilling athletic competition -- Dragon Boat Racing, an ancient sport that\'s catching on worldwide.  This is one of the boats used in the race.    Some children cooling off near the boats.    The Dragon!      Fixture of the Sloans Lake Neighborhood.    A perfect end to a great evening.  '),(21577,3994022,'30,July,2004','2004-07-30 00:00:00','   Fly Fishing on the Big Thompson  Friday I took a day of rest and headed up to Estes Park for some fly fishing. About an hour north of Denver on the outskirts of Rocky Mountain National Park is the Big Thompson River. It was sunny and cool at 7am and there was little to no activity on the surface. In the first hour I landed a Rainbow on a RS2 Nymph.    Next came this lovley Lake Trout, I think, on a RS2.    After 10am I switched to a Blue Winged Olive dry fly as there was some surface activity. I landed two Browns near the Dam in the next 2 hours. All in all it was a relaxing day.    A herd of Elk decided to cross the river several yards in front of me. All protected by this big buck.  '),(21578,3994022,'22,July,2004','2004-07-22 00:00:00','   Google, the greatest search engine ever engineered, shows us what is was like back in the day. Photo courtesy of  urlLink fury.com .  urlLink Google Circa 1960 . Please allow 4-6 weeks for results!  '),(21579,3994022,'21,July,2004','2004-07-21 00:00:00','   Andrew Anderson is now a U.S. Army Captain.  For those of you who have not heard Andy Anderson was promoted to the rank of Captain after returning from his tour in Iraq. Congratulations.  '),(21580,3994022,'21,July,2004','2004-07-21 00:00:00','   Dad to comatose son: I\'ll buy you a Corvette if you wake up  Jason Stackiewicz\'s $75,000 new car will arrive at his home in September.  urlLink Boston Globe  '),(21581,3994022,'21,July,2004','2004-07-21 00:00:00','   The 4th of July in Colorado Springs.  Linda, Shawn and Scot headed down to Colorado Springs to visit Karen Andy, Bree and Katy for the 4th of July. It was good to see them all and relieved to know that Andy is  safe and back from Iraq.  '),(21582,3994022,'21,July,2004','2004-07-21 00:00:00','   Hi-Just signed up.&nbsp; At work in union office. &nbsp; Shannon at home and helping mom with garage sale they made 86.00 yesterday.&nbsp; Sale on today as well but only until noon.  '),(21583,3994022,'21,July,2004','2004-07-21 00:00:00','   I hiked the Herman Gulch trail on Sunday and did a little Fly Fishing. Herman Gulch is located in the Arapaho and Roosevelt National Forests near Idaho Springs. The trail is 5 miles roundtrip and starts at an elevation of 10,400ft and ends at 12,705ft at Herman Lake.    A nice sized pool, formed by a beaver damn was my first query and there were 10-15 small trout feeding on the surface. Here I landed a litte Greenback Cutthroat in a matter of seconds.   Several minutes later I landed this tiny Brooke Trout. Small but damn tough when death is on the line. Literaly.  '),(21584,3994022,'21,July,2004','2004-07-21 00:00:00','   My company,   urlLink threefour studios  just launched a website for the real estate development firm, Dakota Ridge Development.  urlLink Check it out.  '),(21585,3994022,'21,July,2004','2004-07-21 00:00:00','   If you want to put a picture up on one of your messages - you can upload one onto the internet for free here:  urlLink http://www.imageshack.us/ &nbsp; Then you can paste the link onto your message&nbsp;like this&nbsp;:  &nbsp;  You don\'t have to sign up or anything! I did it!&nbsp;Then I&nbsp;lost&nbsp;40 inches on my waistline!&nbsp;   See!:::  '),(21586,3994022,'21,July,2004','2004-07-21 00:00:00','   urlLink Who\'s taking swear words out of \"Murder, She Wrote\" books?  An unknown culprit in Mormon country has crossed out the nasty words and replaced them with milder words, like \"darn,\" \"gosh\" or \"heck,\" written in black, purple, green and even pink ink. One library patron says: \"It bothers me \'cause I\'m trying to read a book. It\'s distracting.\" (Deseret News)  '),(21587,3994022,'21,July,2004','2004-07-21 00:00:00','   Name this blog. What should we call this \"web log\"? You decide, submissions due immediately!  urlLink Email me.  '),(21588,3994022,'11,August,2004','2004-08-11 00:00:00','   A birthday gift from the Mrs.  Craig Sween on Lake Michigan with a fat Chinook Salmon. Mmmmm, Mercury!  '),(21589,3994022,'11,August,2004','2004-08-11 00:00:00','   He likes mousse.  Proclaims this young food experimentor. Silk Tofu + melted chocolate chips + blender + 30 minutes in the fridge = mmmmmmmmm good!  '),(21590,3994022,'11,August,2004','2004-08-11 00:00:00','   A true Texas Cowboy and American . Not to mention faux President.  urlLink Will Ferrel\'s take on George W. Bush.  '),(21591,3363764,'30,May,2004','2004-05-30 00:00:00','   I don\'t quite get the fascination, or obsession, that some people have about someone they really know very little about.  Upon encountering a celebrity, as I have a few times in my life, I always think, \"Oh, I recognize that person from such-and such,\" and that\'s about it.  I don\'t feel compelled to introduce myself to them, comment about their work, or touch them.  They are people whose image just happens to be copied several times and distributed through different media.  They\'re just people.  They are not newly- discovered forms of life of extraterrestrial origin.  I\'m not going to line up for several hours to get pictures taken with them, if I have any choice in the matter.  If only people showed as much respect and admiration for their own friends and family; maybe even their enemies, in some form or another, perhaps this world would be a better place.  Are you more excited to see a person whose image you\'ve seen over and over again, whom you have never had a conversation with, than someone that you really know and love?  '),(21592,3363764,'27,May,2004','2004-05-27 00:00:00','   Looking at some of my entries, you\'d probably think I\'m generally a miserable, negative person, full of malaise and ennui.  Sorry, felt like using some French.  Actually, I\'m not.  It may be because I only write things on here when I\'m bored or depressed.  Which makes sense, I guess.  Actually, when you average everything out over the last few weeks, I have probably been the happiest I\'ve been since I was about 10 years old, believe it or not.  Well, maybe happiest is not the right word, exactly.  I should say most content with who I am, where I am.  Or, I should say, least unhappy, if that makes any sense.  Least anxious.  Least lonely.  Sometimes I feel lost, but it\'s nowhere near the feelings that I used to get.  I recover quickly; I realize that it\'s a mood that will pass.  I realize that bad moods are a part of life.  I think I\'ve done okay for myself, overall.  A lot of what I have is much more than what most people on the planet will ever see.  And I do something that I love for a living.  And I\'m self-sufficient, I\'m healthy, I\'m intact...  Am I whole?  Well, I don\'t know yet.  Is my life going to continue in a straight path and follow this pattern until the end?  Not sure.  And I guess I never really want to be completely sure.  Good moods are a part of life, too.  Really good, euphoric, what-the-fuck-is-happenning good moods.  And those make it all worth while.  '),(21593,3363764,'24,May,2004','2004-05-24 00:00:00','   I need a freakin\' vacation so badly now... Jesus.  And the fact that the weather around here has been totally freezing at night isn\'t really boosting my morale.  Do you ever get that completely bogged- down feeling when you think \"Yeah, I can go on vacation, but the vacation can only last so long, and when I come back there will just be an even more tremendous pile of crap that I\'ll need to deal with.\"  God, I don\'t know what\'s going on, I don\'t know what I\'m doing, I don\'t know where my life is headed.  And I still have this annoying freakin\' headache.  I just needed to complain.  Sometimes I don\'t know what the point of any of this is.  '),(21594,3363764,'20,May,2004','2004-05-20 00:00:00','   Shit, why does my boss\'s ex always call up the office if he\'s never here, and she knows his cell phone number?  For fuck sake, leave me alone and don\'t bother me with that shit right now!  '),(21595,3363764,'19,May,2004','2004-05-19 00:00:00','   I feel like I need to cash in some of my sick hours and take the rest of the day off of work.  What will I do with my extra time?  I was thinking about having a good cry.  I haven\'t cried in years.  I think I need a good one, the when-it-rains-it-pours, thunderstorm-in-the-desert, shuddering sob, snot-bubble, doubled-over-in-a-fetal-position kind.  If I\'m not able to manage that perhaps I will take my dog to the beach.  There\'s nothing like the power of the Pacific Ocean to remind you of how inconsequential and insignificant your problems really are.  Ocean saltwater has exactly the same consistency as human tears.  I wonder if marine mammals cry, or if that problem is pretty much taken care of with the water.  I\'m hoping no one reads this, but I\'m writing it here anyway, I don\'t know why, I just had to dump these thoughts someplace else other than inside my own head.  It may be deleted just as fast as it was written.  It\'s purely an internal problem, no one can help me with it, I don\'t know why I would want anyone to read about it, it\'s no one\'s fault but my own, I guess I\'m just an emotional exhibitionist or a fucking sedomasochist.  '),(21596,3363764,'19,May,2004','2004-05-19 00:00:00','   karma, coming into balance- aha!  there\'s the pain i remember i knew you would eventually show your ugly fucking face around here again, sooner or later such is the universe  '),(21597,3363764,'19,May,2004','2004-05-19 00:00:00','   how the hell am I supposed to keep my focus on this shit right now?  '),(21598,3363764,'18,May,2004','2004-05-18 00:00:00','   I\'ve been having too many good things happen to me lately.  This is definitely not fitting into my normal pattern of bad luck.  Today I got another random raise at work.  And I\'ve sort of been procrastinating lately.  Imagine if I were operating at 100%?  I feel like I\'m due for a disaster pretty soon.  This much good luck with no effort on my part has to be bad for my karma.  Maybe my splice has occurred after all...  '),(21599,3363764,'17,May,2004','2004-05-17 00:00:00','   When I was little I was very religious.  Probably up through freshman or sophomore year of high school, in fact.  I used to say a set number of Hail Marys and Our Fathers per day, plus a bunch of other ones my parents and grandmother taught me.  Somewhere along the line I got really pissed at Jesus.  There were times that I thought he had ditched me or was purposely being a jerk to me.  instead of prayers, there were actually times when I would literally look up at the sky and say, \"Dude, why are you being such a dick?  Remember a few years ago, when I was your best friend?  Let\'s see if I ever talk to you again.\"  Well, whether or not he could actually hear me, whoever he worked for or whoever\'s son he supposedly was, I think that\'s incidental.  What I was saying goodbye to was an ideal- the paradigm of a human being.  I was saying goodbye to my own naivete.  Having learned more about the world, I stopped believing that anyone actually gave a fuck about living up to such an example.  I started to think, what\'s the point in trying to follow the example of a historical figure from 2000 years ago, if it only means that people in the late 2nd/ early 3rd millenia would just walk all over you for it?  What kind of reward do you get for \"following Jesus?\"  Let\'s be practical for a second and think about it strictly from a scientific standpoint.  Well, lately I\'ve been realizing that perhaps I\'ve already been given my reward, and am continuing to reap rewards, without even noticing.  Maybe there\'s reward enough in contributing to the survival of your species.  Maybe there is such thing as good Karma, and maybe Jesus isn\'t such a dick after all.  The jury\'s still out on his dad, though.  Can you really call the underlying infinitely complex mathematical order of the universe a dick?  And would it care if you did?  '),(21600,3363764,'17,May,2004','2004-05-17 00:00:00','   What if you could choose the exact traits that the perfect member of the opposite sex (or the same sex, if that\'s your thing) would have.  What would they be?  Let\'s say that you can choose 50 traits that he or she would have, in terms of personality, intelligence, and interests.  Let\'s say you can choose, oh, 10 physical traits.  What would it be worth to you to be able to meet a person with all those traits- down to the very last one- tomorrow?  Can a price even be put on that?  Is it worth different things to different people?  Maybe to some it\'s not worth much at all, and of course to others the question isn\'t even applicable.  What are the chances that a person like that even exists?  Now, the big question is, what are you supposed to say to him or her when you finally meet?  Let\'s say he or she doesn\'t know about all the traits that you have chosen.  How are you supposed to explain them to him or her without it sounding like a load of bullshit?  Now, what if he or she decides to convince you that he or she does not possess all of those traits?  What argument can you make?  A pretty tough puzzle... perhaps requiring lateral thinking...  '),(21601,3363764,'17,May,2004','2004-05-17 00:00:00','   Have you ever thought that you would put everything you thought made you happy on the line for a chance at what would make you really happy?  Do sticks and rocks matter? ;)  '),(21602,3363764,'17,May,2004','2004-05-17 00:00:00','   If my splice hasn\'t occurred yet, I\'m imagining that it will soon.  '),(21603,3363764,'17,May,2004','2004-05-17 00:00:00','   Mondays.  For obvious reasons.  Although my current job does make it easier for me to deal with them.  On this particular Monday, as with many Mondays recently, I am finding it hard to concentrate and hence I find ways to procrastinate.  No, I am not charging clients for this, of course.  Oh, the other thing I\'m not comfortable with.  Crushes.  They\'re like hiccups.  I thought I was all done with them as of a decade or so ago.  I\'ve never had hiccups in ten years, and I was hoping that I would never have another one, nor would I ever have another crush.  The similarities between the two?  Lack of control.  Will power has nothing to do with it; they are involuntary bodily functions.  And you always feel stupid during and after they are happening, especially if they\'re loud and obvious.  I shouldn\'t be having any crushes, doing any crushing, at this age.  Is that what this is?  It\'s different than before.  I thought crushes, by definition, were contingent upon physical attraction, but I guess I was wrong, because apparently I have a crush on a woman without even knowing what she really looks like.  Uh oh, you\'re thinking, this guy has lost it.  And that\'s my whole point.  Remember that movie \"You\'ve Got Mail?\"  Cheesy and sappy come to mind.  How do I get myself into such situations...  '),(21604,3363764,'17,May,2004','2004-05-17 00:00:00','   First time blogging.  Initial entries will probably be cursory at best, until I figure out how to be witty and somewhat entertaining to those who don\'t know me while simultaneously being benign and, um, careful, to those who do know me and have somehow found this blog.  I must have told you about it while drunk.  Either that or I must really like you.  '),(21605,3363764,'16,June,2004','2004-06-16 00:00:00','   My concept of time is warped right now.  It feels like weeks since I\'ve written on here but I just wrote something four days ago.  Maybe it\'s because so much has happened between then and now.  Driving hundreds of miles for my sister\'s graduation, partying, eating, and drinking to excess down there, and a busy day at work yesterday and even today, though it is supposed to be my first day of vacation.  I just felt bad that I had one piddly little thing to do before receipt of permit for this one project, and the clients have been waiting so patiently.  Little did I know that driving to do that one piddly little thing and waiting a few hours in total for red tape to untangle itself would take basically the whole day.  So I guess I have an extra day of vacation somewhere down the road.  Although I have so many freakin\' days piled up I can\'t keep track of them anymore.  But that also means that I\'m almost a day behind moving all my crap from my old apartment to my recently acquired house.  Oh well.  I guess I\'ll just enlist a few more people to help me.  After all, I did get all those offers.  I\'m going to shoot for being out of here by noon on Friday.  Then I\'ll have a bunch of days left of vacation to enjoy the house and hang shelves and pictures on the walls.  It could be that my concept of time is warped due to sudden exposure to the sun for prolonged periods of time after living like a vampire for several months.  I think it made me loopy.  Oh well, at least I have a nice, dark tan now.  Wow, this is pretty much the most normal and sane thing I\'ve ever written on here.  That must mean I\'m happy.  Could it be the sun?  Could it be the length of days?  Could it be my house?  Could it be my ridiculously- friendly-to-the-point-of-being-Midwestern new neighbors?  Could it be my vacation?  Could it be that work is going pretty well?  Or is it something else?  Hmmm...  '),(21606,3363764,'11,June,2004','2004-06-11 00:00:00','   How did people communicate across oceans before there was copper wire and fiber optic cable?  How did people find long- lost friends without the internet?  It must have been agonizingly slow.  Waiting for a letter from overseas, complete with wax signet seal, must have been excruciating.  Do you remember life without the internet?  Granted, it was invented in the sixties, and I wasn\'t alive back then, but it didn\'t come into widespread use until, what, the late eighties?  I know I never used e-mail until I got to college.  Who would have thought that almost nine years later I would have become so dependent on it?  Remember having to find a particular item, for a history or science project in junior high, and actually needing to use the yellow pages to call businesses one at a time, or, worse yet, go door to door?  You know, many may say that the internet is a source of evil and corruption.  But it sure as hell makes it easier to find the things you\'re looking for, regardless of how unlikely the the chances of finding them seem.  Millenia from now people will be wondering how 21st century dwellers dealt with the tediousness of communicating between planets with simple beams of electromagnetic radiation.  Imagine, waiting over a twenty- minute delay when talking to your friend on Mars.  How long would a phone conversation be?  I wonder, if two telepathic people were placed light years apart from each other, would their communication be instantaneous or limited by the speed of light?  If the speed of light is 310,000 km per second, what is the speed of thought?  '),(21607,3363764,'04,June,2004','2004-06-04 00:00:00','   I don\'t know why I can\'t leave well enough alone.  '),(21608,3363764,'02,June,2004','2004-06-02 00:00:00','   I\'m not entirely sure anyone is reading this blog at all, so maybe I\'m just talking to myself.  That\'s okay, I do that a lot, and my fingers can use the exercise as opposed to my vocal chords.  I wonder how long blogs last, or how long they theoretically can last.  I mean, technically there\'s no reason why a blog shouldn\'t last until the end of time.  Someone on a distant planet several thousand years from now might find this in the archives of the internet somewhere.  They may delete it.  Or they may read it and dismiss it.  What\'s the point of writing it, then?  Well, if   one is going to pose that question, one must also ask what the point of living life is.  I really don\'t know what I\'m talking about.  I am rambling as usual.  That\'s the whole point of this.  It\'s nice and hot today.  Finally.  Solstice approaches.  Go hang a salami, I\'m a lasagna hog.  '),(21609,3363764,'02,June,2004','2004-06-02 00:00:00','   Someone to go to my stupid movies with so I don\'t feel awkward sitting by myself in the theater (I have only done that once or twice in my life- as misanthropic as I may be, I just cannot sit in a movie theater by myself).  Someone to do things with that no one else really wants to do, like walk to the top of Half Dome.  Someone to go to my favorite steakhouses that don\'t offer takeout with, so, once again, I do not feel stupid sitting by myself.  Someone to fly to a distant and new place with me so I don\'t have to start up conversations with strangers on an agonizingly long flight.  Someone to sit on the train with me in Europe so that we can sleep in shifts and gypsies won\'t steal our stuff.  Someone to accept chocolates and roses from me on Valentine\'s Day so that they won\'t go to waste after I buy them in an effort to feel like part of the crowd.  Someone to put sunscreen on my back.  Someone to massage my back when I\'m so tense that it spasms really hard and makes me feel like I\'m growing a second goddamn spine.  Someone to give my friends a reason for why I always put the toilet seat down.  Someone to use as an excuse when I just don\'t feel like hanging out with said friends.  Someone to go to a beach on a Hawaiian Island with so I don\'t feel like the only lone freakin\' creature on the planet and I can just enjoy the sunset.  (Even the damn fish I see while snorkeling are in pairs.)  Someone to bug me to clean up and be more organized.  Someone to take up the extra space in my house.  Someone to help me use up all those stupid two-for-one coupons.  Someone to go to my high school reunion with so I don\'t have to pretend to get along with all the shallow people.  Someone whose picture I can put in my wallet so it looks like I have a life.  Someone to be with on New Year\'s Eve.  Someone to listen to me bitch and moan.  Someone to bitch and moan to me.  Someone to call me up at work to give me reason to procrastinate.  Someone to paint a portrait of.  Someone to split the bloomin\' onion at Outback with because it\'s way too big to finish by myself but I love that sauce that comes with it.  Someone to suck the venom out when I go to the actual Outback and get bitten by a snake.  Someone to read ramblings such as this.  Someone to give my mom assurance that I\'m not gay.  Someone to make everyone stop asking why I\'m single.  Someone to cook breakfast for.  Someone to wake me up when I don\'t hear any of the three alarm clocks.  Someone to sit next to me on a roller coaster.  Someone to be around on Friday and Saturday nights when no one else seems to be.  Someone to ride in the car with me between 6 AM to 9 AM and 3 PM to 6 PM on weekdays so I can use the carpool lane.  And cross the Bay Bridge without paying toll.  Someone to get sloppy drunk with.  Yeah, someone to drink an Irish Car Bomb with.  Someone to split those 3,000 calorie restaurant desserts with so that I don\'t get fat nearly as fast.  Someone to whisper to me until I fall asleep.  Someone to pretend to like me.  '),(21610,3363764,'20,July,2004','2004-07-20 00:00:00','   Aha, seems that we are now able to alter the font, size and color of our entries, and   the html!?  Wow, now I can put links on my entries like a real grownup.  Of course if I weren\'t so apathetic I\'m sure I would have found a way to do that in any case.          This shed I am building is getting to be a pain.  It was just supposed to be a simple little shed.  But no.  I had to go and make it complicated.  Seems like I complicate a lot of things in life that should be straighforward and simple.  I like making things hard on myself.  I didn\'t know it was going to take several trips in a little car and two trips in a giant SUV to get the materials necessary to build this so-called little shed.  or make that three trips in the SUV, because I have to go back and get more stuff tomorrow.  Maybe it wasn\'t such a smart idea to move almost a ton of material by myself.  Is that why I\'m sore?  I\'m a pain in my ass.    And now here\'s what I just wrote in Webdings:      Aha, seems that we are now able to alter the font, size and color of our entries, and   the html!?  Wow, now I can put links on my entries like a real grownup.  Of course if I weren\'t so apathetic I\'m sure I would have found a way to do that in any case.        This shed I am building is getting to be a pain.  It was just supposed to be a simple little shed.  But no.  I had to go and make it complicated.  Seems like I complicate a lot of things in life that should be straighforward and simple.  I like making things hard on myself.  I didn\'t know it was going to take several trips in a little car and two trips in a giant SUV to get the materials necessary to build this so-called little shed.  or make that three trips in the SUV, because I have to go back and get more stuff tomorrow.  Maybe it wasn\'t such a smart idea to move almost a ton of material by myself.  Is that why I\'m sore?  I\'m a pain in my ass.     If I type in Webdings, can I count that as another foreign language on my resume?  '),(21611,3363764,'14,July,2004','2004-07-14 00:00:00','   humming and humming I stare blankly at the damn refrigerator  '),(21612,3363764,'10,July,2004','2004-07-10 00:00:00','   I feel like my mind is on too many things at once, even though comparitively I\'ve been a complete slacker lately.  Actually, it might be that a good portion of my brain is preoccupied, and the little part that\'s remaining is overwhelmed by simple tasks such as home improvement, feeding animals, and writing in my blog.  procrastination awkward silence dominates anticipation  midnight reflection across the largest ocean my dreams shall set sail  through my mind\'s valley an intangible fog creeps soothing and calming  a beautiful voice is heard across the darkness the dawn approaches  dazzling daylight dark corners of my heart thaw the long winter ends  my fears put to rest relief and joy fill my lungs this must be a dream  whispers in my ear will tell me to awaken this can\'t be real  '),(21613,3363764,'06,July,2004','2004-07-06 00:00:00','   censorship sucks http://www.petitiononline.com/hole64/petition.html  '),(21614,3363764,'05,August,2004','2004-08-05 00:00:00','   distracted always dirty thoughts run through my mind perverse yet loving  '),(21615,3065484,'28,March,2004','2004-03-28 00:00:00','   Hi my name is ??????????? you want to know my name so I give you some hints (Listen:I am a Korean person and I\'m 14 years old and Grade 7 in NJIS.) Do you know my name now?? My name is KWang Wook,. And I can play golf so my golf scores are 95,and handicap 40. But next year I want to be golfer. and i will be better than Tiger Wood. He was famous at LPG golf competition. And next I like to play computer games too much. Because computer makes people happy and exciting. I can  play Clarinet because I like  urlLink music . Clarinet makes to good sounds and feels comfortable. my favorite subject is English and science. my e-mail was \" urlLink skw92000@yahoo.co.kr \"   urlLink  '),(21616,3065484,'28,March,2004','2004-03-28 00:00:00','   Hahahahahawhahahahahaha>>>>>>>>>>>>>  Who Am I???  My name is Kwang Wook Shin, and I am korean person.  so I come from Korea. I got 1 older sister(she\'s now university).  Me.....ummm I\'m 14 years old and Grade 7 in NJIS.  My favorite sports and hobby was playing golf and basketball and my hobby was enjoy to playing game.  AND I like to play Clarinet, because it was make good sounds and feels comfortable.  My best friends are Kennith, and Dong san.  Before i have one puppy dog in my house.  '),(21617,3065484,'25,April,2004','2004-04-25 00:00:00','   1. Do you feel this would be something you would like to use? => Yes, because I could buy aynthing that i want it easily.  2. Do you think society is ready for technology like this? => No, I think society was not ready for Technology because there isn\'t development techonology yet, so i don\'t like technology now.   3. What would be the benefits of this technology? =>  It was development more technologythis year because all people want to use Technology..  4. What could become problems with this technology? => There are so many problem with Technology because like sometimes the technology are makes not good for them. 5. What other thoughts or opinions do you have regarding this technology. =>???????????????????????????????????????????????????????????????  '),(21618,3065484,'15,April,2004','2004-04-15 00:00:00','   1. Do you feel this would be something you would like to use? => Yes, because I could buy aynthing that i want it easily.  2. Do you think society is ready for technology like this? => No, I think society was not ready for Technology because there isn\'t development techonology yet, so i don\'t like technology now.   3. What would be the benefits of this technology? =>  It was development more technologythis year because all people want to use Technology..  4. What could become problems with this technology? => There are so many problem with Technology because like sometimes the technology are makes not good for them.  5. What other thoughts or opinions do you have regarding this technology. =>???????????????????????????????????????????????????????????????  '),(21619,3065484,'15,April,2004','2004-04-15 00:00:00','   Hello Njis Students tomorrow we are going to YOGJAKARTA. So i think there was fun. I already have been YOGJAKARTA 1 times. There are so hot and so funny places.                                                   BY: Kwang Wook SHin.....    em> and every body enjoyed the fieldtrips............................  '),(21620,3065484,'15,April,2004','2004-04-15 00:00:00','   Today I have swimming and I learned butterfly strog. at first we swimming 7 laps and then later playing \"Tube WaterPolo.\" and I ride on the Tube and starting play. I like swimming at PE times.  '),(21621,3065484,'04,May,2004','2004-05-04 00:00:00','   1. How does the increasing use of cameras affect privacy? =It could affect us, people will watch us when we are sleep, eating, Running, talking or other else... 2. Where do you think it is fair to have cameras installed? Not fair? = I think, it is fair to have camaras installed in house or places because when we stay at home we can take a picture of family and it can show my house pictures. And when in other place we can use for important things.... 3. What is your opinion on this subject? =In this opinion i want to tell the technology was development with them... *^---^;*  4. If you feel that this is an invasion of privacy, how do you feel governments should help increase protection from the inclining terrorist threats? =????????????????????????????????????????????????  '),(21622,3860954,'26,July,2004','2004-07-26 00:00:00','   It\'s amazing how much shit you can accumulate in such a short period of time. With my move back so close, I\'ve been going through everything I own in order to scale my vast collection of irrelevant crap down to a reasonably-sized, easy-to-move-around collection of irrelevant crap. I\'ve kept letters, pieces of paper, a marble, a photo of some unidentified person\'s shoe, someone else\'s ID, collars of past dearly loved pets and train tickets, all for the sake of memories that mostly&nbsp;I cannot even remember.&nbsp;Going through all of this stuff made me realise how sentimental I am. But the time came, I had to move on, take what I needed, and bin the rest.  On a lighter note, Poe learnt how to say \"fuck you hunny\", in the gayest accent that a bird can muster, and that\'s kept me amused for the last few hours. Leaving him here will be difficult, but unlike my puppies, he can come up and live with me in another 9 weeks.  Peace, love and bad-mouthed homosexual birds,  Bethie xoxox '),(21623,3860954,'15,August,2004','2004-08-15 00:00:00','   The young child next door has an imaginary friend. I often hear him outside speaking with it. The other day I realised that he was referring to it as \"Mum\", and I thought... \'stupid Beth, he just has a soft-spoken mother, and you cannot hear her respond\'. However, recently I learned that his mother died a few years ago. It makes me sad to listen to him now, chatting away to thin air as if someone is lovingly looking on as he plays with pegs and toy cars. Peace, love and absent parents, Beth xoxox '),(21624,3860954,'10,August,2004','2004-08-10 00:00:00','   Your nothingness is my everything, your heart is my never-ceasing passion. Everything is broken. Feeding this corrupted, fitted puzzle. what you search for in confusion only makes more sense. Your black world, my white face, surrounded by this dissimilarity, they all become the same. Sharing opinions you all wish to find likeness. Where is your precious uniqueness now? All looking for pity, your insecurity, your suicide, your deadly attention. You say you refuse to conform, to disregard the drones. By doing so you join the outcasts, this labelled group who are so-called unique - unique is none the same, yet you feel as though you\'re joining something - you feel a part of something. This destroys your feeble attempts. Feeding off this darkness, you\'ll find that nothing satisfies. '),(21625,3860954,'01,August,2004','2004-08-01 00:00:00','   I think that my cigarettes prefer to be smoked at night, when no one else is around. Well, at least that\'s how I enjoy them. With the faint shaft of light that seeps outside the backdoor from inside the house makes the cancerous smoke dance, and I stupidly imagine that it is my friend, flirting with my imagination like it\'s about to write me an intimate letter, expressing it\'s gratitude for my addictive loyalty in the air before me... but it never does. My deadly companion, I sacrifice myself for you daily, and you do not respond. You only make me smell bad. I crave you, and you float away with the wind, infecting others, with not so much as a thank you for my attention. '),(21626,3860954,'01,August,2004','2004-08-01 00:00:00','   -It is no surprise that a generation of children of divorce have grown into a world of extended adolescence, in which so many of them have slept with one another and remained friends, have put the conflicts of sundered relationships aside for the sake of maintaining a coherent life. Divorce has taught us how to sleep with friends, sleep with enemies and then act like it\'s all perfectly normal in the morning. - I have so much respect for myself but have very little respect for most of the people I sleep with, particularly my subs. Is it that I think I am closer to perfection and less susceptible to temptation than the \'typical chauvinistic males\'? I have stronger grudges against women than men, but the biggest grudge I have is against myself. I am an actor, a dramatised projection of a few minor traits in my personality, and I use this facade to appear genuine. It works, but I never call them again. Should the guy be saying that? - Bethany is a dirty hussy. '),(21627,3860954,'01,August,2004','2004-08-01 00:00:00','   I\'ve always wondered about certain jobs. Especially those guys who drive the cars in front of trucks with a large yellow sign stating \"Oversize vehicle ahead\", with the flashing lights. It\'s like \"Hey, what do you do for a living?\" ... \"Oh, I drive one of those cars with flashing lights that say \"oversize vehicle ahead\" \"... I wonder how you\'d apply for such a job... License: check, IQ under 90: check.  Everytime I see one, I cannot stop laughing. It\'s the same with people who lock the cemetary gates at night. \"What do you do for a living?\" ... \"Oh, I lock the cemetary gates at night\" ... \"Oh, so you unlock them in the morning?\" ... \"No, I am not trained in that respect, but sometimes I get extra shifts and lock the public toilets at night... you know, clear out the random gays declaring the site as their new beat, or occasionally a corpse resulting from a drug overdose\". It brought me to a thought... before I actually embarked on my shitty job as a Deli girl, I had never considered those people who served me at the supermarkets, just took them for granted. I suppose quite a few people feel the same way, as I am not treated as a person in my job, only by the occasional thoughtful senior citizen. Today was my last day (or should I say... technically, yesterday was my last day), and as it was a Sunday, my nine hour shift was particularly slow, and I found myself purposefully (and quite obviously) fucking up orders so that my time would be filled with something that to me appeared constructive. Through this job I have become so morbid. Moreso than usual. As I have never liked meat, it was expected that I would be bitter about such employment, but my mind has been playing nastier tricks on me. Occasionally if someone asks for \"silverside\", I swear they are saying \"suicide\". Not long ago I found a bacon rasher that looked like Frankenstein\'s monster. When opening a new packet of cocktail frankfurts, they are shaped like a noose. Every so often, the weight will come up as 66.6g, or the price as $6.66 (only with particularly evil customers), and it keeps me amused, otherwise I am calmly bitter/cynical/sarcastic toward customers, and find no joy at all in working in an edible mortuary. This all sounds stupid, trust me, I realise. But when stuck in the redneck shithole, little else amuses me. I move back to the city tomorrow, and aim toward finding a new job asap, involving serving alcohol to cockheads and sleazes. Again, I will instantly find issues with that, as well. I should get some intense training as a barista, then I would only have to hear customer complaints second-hand, and not have to bite my tongue to stop myself from articulately telling them to fuck off. Like chicken drumsticks, or cuts of gammon, I find myself comparable to the pristine frozen supermarket meat. But unlike the plastic coated innards, hearts which can be preserved forever, I am just a mass of stinking flesh As life rots me from the inside. Don\'t worry, I love myself (in a non narcissistic way, I just scribbled down some thoughts on a scrap of Deli paper one afternoon...) Peace, love and baby octopus, Beth xoxox. '),(21628,4165509,'07,August,2004','2004-08-07 00:00:00','   Lets start with my parents wedding.My mom was fooling herself, she married my dad just to marry him.She thought she had to marry,she did have 7 sisters that were marrying or already married.She was the youngest..Than again you cant blame my mom,she never really remembered a true marraige.Her mother died when she was real young, And her farther re-married an evil bitch..Who was worse than the stepmother in cinderella.Unfortanly for my mom she didnt get a ever-after happy ending like cinderella.        She and my dad married,last for seven years.I was about 5 almost 6,When they divorced..I remember,sad really.I have issues remembering thangs,but I remember this. I was the one who sugested them getting divorced.What was I thinking? You try being 6 year old,the kid at school everyone picks on,yet The trouble-maker..And have to go home to parents fighting,throwing stuff,and hardly ever seeing your mom,and your farther forgetting your alive.      Well they got divorced by the time I started 1st grade..By the time I was about to begin 3rd grade.I lost my grandma*to my mom,the only one Alive that I had..my stepgrandma dont count as the family, mom got real issues with her.* And my grandma and my aunt were kinda like my reall close family.My aunt is still alive,and she\'s cool, But she lives out of state.      Well as the years went by..My dad kinda got in trouble, drank, and had anger issues everyonce in awhile..Lost his house..We moved into an apartment for cheaper than normal.Wasnt bad either, he got a specail deal to..Met some friends, infact two of my best friends now.Well one, we\'re kinda distant,but we still talk..She\'s real nice..The other she\'s fun and cool,but I bit to sarcastic at times..Even when she\'s joking you cant help but take some thangs personal.Not I,but others.Heh I\'ve learned to ignore or sarcastic commements..      By then My dad and I we\'re cool,I was playing soccer..Being the son he never had,and I was taking care of the son He did have..My little brother isnt an atheltic or smart kinda boy..IM the one who has to do so well in sports,and grades.I dont mind it,Gives me something to do,and at least that way My dad knows who I am.      My dad was done with some of the bad shit he was into and etc..But than he lost alot of money..Lost the apartment..Lived with his brother,my uncle who\'s really sick..Infact he got better,but is sick again..Poor uncle mo.      Well My dad and bob,both getting reall down on money wise..looked,got a house togethere, got new jobs that pay alright enuff..And are living togethere..Now bob is disguesting! He smokes, drinks,parties, and is a pervert..But he\'s not that bad of a guy, and he\'s my dads best friend...THe bad thang bout this is, my dad is drinking again, gambling..We still keep in touch some.Oh by the way,Im living with my mom now.And she\'s stressing alot over who knows what..      As thangs started to seem good,they went downhill..Funny eh? well once your down thiers only one way to go, back up..So Im not gonna get depressed bout this, But sometimes It kinda gets overwhelming..      anywho thats the background..and now we\'re to the future. Me living with my mom,who\'s become an over-stress, raging mom..who\'s drinking occasionally to.. To me starting at a new school,with my old friends, but funny thang is, I only have a couple classes with my old friends.Time for new friends,and a new beginging..I just gotta tie up loose ends..      Oh I know its not the smartest thang to do, nor the safest..But I\'ve met some friends online, from a distances..Jason, he\'s a great friend..and My new boyfriend Joey.      But the problem is I have a best friend thats a guy,down here..Who can be such a temptation at times,and he wont shut up..always liking someone else, and yet likes me but will never go back out with me..and Im not sure,but I just cant let go of him..Which is an issue..and I dont know if being such close friends with him will work out..so I think Im gonna try and push him away..      I started to,but he resisted..Argh..How annoying..But Nonetheless..if I want my new bf joey to work and to be honost and true with him..I need to stop holding onto Dan.         Okay Now that you know whats going on,and up-to date,and now a basic review of my past..We can talk some real shit,But Im not in the mood right now..one of my best friends*a chick one* signed on msn..SO IM OFF to chat. Toodles! heh..That is just a funny word to me..I like it dunno why..TOodles!  '),(21629,3894623,'11,July,2004','2004-07-11 00:00:00','    Testing, testing... 1 2 3 '),(21630,3894623,'31,July,2004','2004-07-31 00:00:00','   Well I grew up in the 80\'s, and like all 80\'s fight fans I was a big supporter of Iron Mike Tyson. Then all the craziness that has become his life started to set in. Last night he began his umpteenth comeback, this one brought on by the IRS. He needs to pay back 38 millon in back taxes and other assorted fees. He might have to get a regular job, or go back to jail or something, because the way some British slob in boxing shorts man-handled and ultimately knocked \"iron\" (or is it lead these days?) Mike Tyson out in 4 rounds; it looks like his days of raking in the money via a squared circle may be behind him. It really is sad to see how his life turned out for him. Even sadder is to think of what a washed up, ex-champion, ex-con; with no savings, and no friends,  has to look forward to in life... Only time will tell.   '),(21631,3894623,'26,July,2004','2004-07-26 00:00:00','   urlLink    Still looking for my GTO. I think I have ruled out a convertible... Sure would be nice though.&nbsp; urlLink   '),(21632,3894623,'21,July,2004','2004-07-21 00:00:00','   urlLink    Just another peak....&nbsp; urlLink   '),(21633,3894623,'21,July,2004','2004-07-21 00:00:00','   urlLink    Take That A-Rod. If the Red Sox played like that every game we would be running away with the Division now!&nbsp; urlLink   '),(21634,3894623,'21,July,2004','2004-07-21 00:00:00','   They really don\'t care about customers. I bought a Sony Vio laptop less than a year ago. They told me that it\'s out of warranty because they stopped making the model in July and even though I bought it in August, they start counting down the year when they beginning churning out the newer models. They certainly don\'t compare to Dell or Gateway when it comes to working with their customers. My advice? Buy something OTHER than Sony...   '),(21635,3894623,'18,July,2004','2004-07-18 00:00:00','   Well with a&nbsp;little less than 1/2 the season left the Red Sox are still stinking. They have to get things together and learn to win on the road soon. Or else it will be another long, sad offseason for Red Sox Nation.  &nbsp;   '),(21636,3894623,'17,July,2004','2004-07-17 00:00:00','   Well saw Spider-Man 2 again tonight. It was even better the second time around. I highly recommend this movie to all that enjoy a great show. Go now! \"I\'m back, I\'m back, I\'m back.... MY BACK!\"   '),(21637,3894623,'16,July,2004','2004-07-16 00:00:00','   I am extremely bummed out. I was bidding on a GTO on eBay, and I got sniped by 100 bucks... I wanted that car so badly too. Oh well... I\'m on the hunt now, I even have my liscence plate picked out.... Stay tuned.   '),(21638,3894623,'11,July,2004','2004-07-11 00:00:00','   Here are some links to some of my families\' Blogs... Check them out and leave a comment or two.    urlLink My Mom\'s Blog     urlLink Blobby\'s Blog    Well I am being summoned away.   '),(21639,3894623,'11,July,2004','2004-07-11 00:00:00','   urlLink    Back in my carefree days!&nbsp; urlLink   '),(21640,3894623,'11,July,2004','2004-07-11 00:00:00','   Hey I figured I should jump on the bandwagon too! If you\'re not in the lead... you\'re off the back.   '),(21641,3894623,'04,August,2004','2004-08-04 00:00:00','   urlLink    Man there will be more on this later... but I gotta get to work now...&nbsp; urlLink   '),(21642,3894623,'01,August,2004','2004-08-01 00:00:00','   urlLink    Well it\'s all over for #5 of Red Sox Nation. In a trade that will mark this season, hopefully in a good way, Nomar was traded to the Chicago Cubs. So he went from one historical loser to another. I hope he can be happy there because he was miserable here in Boston this last year plus. Wouldn\'t that be irony if we did seem him again in Game 7 of the World series and he slapped the winning homerun for his new team...&nbsp; urlLink   '),(21643,3325171,'12,May,2004','2004-05-12 00:00:00','    urlLink TAKE THAT!!!!!!!!!!!!!   '),(21644,3325171,'12,May,2004','2004-05-12 00:00:00','    Since you all think it\'s funny that I\'m a little asian girl that lives on a farm, here we go with my online farm...   '),(21645,3325171,'13,May,2004','2004-05-13 00:00:00','   urlLink Games people play - May. 13, 2004 : \"CNN/Money\"  Never heard of the game....GOD im getting old?  '),(21646,3325171,'12,May,2004','2004-05-12 00:00:00','   urlLink Boston.com / Business / Scrambled channels irk cable viewers : \"Comcast Corp. is cracking down on a free lunch many of its Boston cable television customers have been enjoying for years.\"  (L) users?  '),(21647,3325171,'12,May,2004','2004-05-12 00:00:00','   urlLink Mario 3 beaten in 11 min  '),(21648,3325171,'12,May,2004','2004-05-12 00:00:00','   urlLink Click here Movie Quiz  '),(21649,3325171,'11,May,2004','2004-05-11 00:00:00','   urlLink \'Spam King\' gets restraining order against SpamCop | The Register : \"A bulk mailing company headed by notorious spammer Scott Richter has won a restraining order against anti-spam reporting service SpamCop.\"  There goes the neighborhood :)  '),(21650,3325171,'11,May,2004','2004-05-11 00:00:00','   urlLink New Jersey 101.5 FM Radio : \"THURSDAY  80s Similar to Wednesday some sunshine with the chance of a late thunderstorm. The highest temperatures in the low or mid 80\'s but as usual cooler near the water   Meteorologist Alan Kasper   Every 10 minutes, 7 days a week! \"   Waht a storm tonight :)  '),(21651,3325171,'11,May,2004','2004-05-11 00:00:00','   POINTS TO PONDER   1. Who was the first person to look at a cow and say, \"I think I\'ll  squeeze those dangly things  and drink whatever comes out?\"    2. Who was the first person to say, \"See that chicken there... I\'m gonna eat the next thing  that comes outta it\'s butt.\"   3. Why do toasters always have a setting that burns the toast to a horrible crisp, which no  decent human being would eat?   4. Why is there a light in the fridge and not  in the freezer?   5. If Jimmy cracks corn and no one cares,  why is there a song about him?   6. Can a hearse carrying a corpse drive  in the carpool lane?   7. If the professor on Gilligan\'s Island can  make a radio out of coconut,  why can\'t he fix a hole in a boat?   8. Why do people point to their wrist when  asking for the time, but don\'t point to their  crotch when asking where the bathroom is?   9. Why does your OB-GYN leave the room  when you get undressed if they are  going to look up there anyway?   10. Why does Goofy stand erect while  Pluto remains on all fours?  They\'re both dogs!   11. If Wile E. Coyote had enough money  to buy all that Acme crap, why  doesn\'t he just buy dinner?   12. If quizzes are quizzical, what are tests?   13. If corn oil is made from corn,  and vegetable oil is made from vegetables,  then what is baby oil made from?   14. If electricity comes from electrons, does morality come from morons?   15. Is Disney World the only people trap  operated by a mouse?   16. Why do the Alphabet song and  Twinkle, Twinkle Little Star have the  same tune?   17. Stop singing and read on..........   18. Do illiterate people get the full  effect of Alphabet Soup?   19. Why do they call it an asteroid when  it\'s outside the hemisphere, but  call it a hemorrhoid when it\'s in your ass?   20. Did you notice that when you blow  in a dog\'s face, he gets mad at you,  but on a car ride, he sticks his head  out the window?   21. Does pushing the elevator button  more than once make it arrive faster?  '),(21652,3325171,'11,May,2004','2004-05-11 00:00:00','   urlLink I don\'t even know..this is just weird  '),(21653,3325171,'11,May,2004','2004-05-11 00:00:00','   urlLink DEA Agent Shoots Self During Gun Safety Class : \"ORLANDO, Fla. -- A federal drug agent shot himself in the leg during a gun safety presentation to children and his bosses are investigating. \"  what a jackass....  '),(21654,3325171,'11,May,2004','2004-05-11 00:00:00','   urlLink C FLizzo aka C Flo - One of Hip Hop\'s finest   I think this guy goes to DeVry.....  '),(21655,3325171,'11,May,2004','2004-05-11 00:00:00','   urlLink   Naughty gnomes made to cover up  Wed May 12,11:17 AM ET 	  BARNSLEY (Reuters) - A man has covered up his lewd garden gnomes with painted-on swimwear after police warned him he faced arrest for causing public offence.  	   While most garden gnomes fish or enact scenes of bucolic tranquillity, ex-army sergeant Tony Watson\'s models in Barnsley, South Yorkhire, bared their breasts and buttocks, prompting complaints from the public.  \"It is an offence to display something that is insulting or likely to cause distress,\" a police spokeswoman said on Wednesday.  \"Although some people view the gnomes as a bit of harmless fun, we have to take complaints from members of the public seriously.\"  One of the gnomes now sports a polka-dot bikini, said local resident John Threlkeld, who passes the gnomes every day on his way to work.  \"Tony used filler and paint to cover them up,\" he said.  '),(21656,3325171,'11,May,2004','2004-05-11 00:00:00','   urlLink More on Beheading  '),(21657,3325171,'11,May,2004','2004-05-11 00:00:00','   urlLink InfoWorld: Sasser a warning of things to come: May 07, 2004: By Paul Roberts : SECURITY : \"Sasser a warning of things to come Don\'t wait to remediate, experts advise\"  you crackah\'s already seen this duuuuuuuuuuddddeeeeesssss  '),(21658,3325171,'11,May,2004','2004-05-11 00:00:00','   urlLink Google   Tough  '),(21659,3325171,'11,May,2004','2004-05-11 00:00:00','   urlLink Google   What is with the stuff today?  '),(21660,3325171,'11,May,2004','2004-05-11 00:00:00','     '),(21661,3325171,'21,May,2004','2004-05-21 00:00:00','   urlLink Extreme Ironing Bureau :: Ironing under the sky target=\"_blank\"  : \"Welcome to the Extreme Ironing Bureau\"  '),(21662,3325171,'31,May,2004','2004-05-31 00:00:00','   urlLink Australian Open Source Group Dismisses SCO Linux Claims - Computer Business Review  '),(21663,3327285,'28,May,2004','2004-05-28 00:00:00','   OK, its the final hour before the end of the work week and a 3-day weekend.  You\'d think I would be excited.  Well you\'re wrong.  This day is going by so slow that I can actually feel myself getting older as each minute passes.  The day started off fine.  The first 2 hours flew by.  But then the hour before lunch basically dragged, knowing full well that I was going to cherish lunchtime.  And from there, this day has just dragged on, kicking and screaming.  And for what?  I don\'t have any plans.  I\'m not going anywhere.  Its not like there is anything for me to be excited about that would make me want to leave.  No, its just the fact that this is the final day before a 3-day weekend.  That\'s all it really is.  And its pissing me off.  Come on...all you need to do is say 4:00 and I am gone.  Is that so difficult?  For the love of all things small and furry....GET ON WITH IT!!!!  '),(21664,3327285,'25,May,2004','2004-05-25 00:00:00','   Is anyone actually reading this?  I mean besides those who I asked to read it?  I guess its hard to expect someone to just up and read your blog, especially if no one knows it exists.  Perhaps I should put a counter at the bottom of the page?  Yeah, then I will know just how pathetic I really am because the damn counter will never change.  Sigh.    Anyway, I guess what really tipped off this latest rage is the current state of music.  I just don\'t understand.  Are there really that many idiots in this world?  I was reading about how experts predict that this season of American Idiot has become the most popular and they expect the finale to be the biggest yet.  Are you serious?  There are really that many retards in the world?  I mean come on, who gets the satisfaction of knowing that glorified karaoke singers are competing for a million-dollar contract and a guaranteed tour?  I certainly don\'t.  There are many unsigned and struggling bands out there that would kill for the opportunity to even open for an established act, and yet, these darlings of the television are being handed a contract and record with a minimum of going platinum at least twice (well, except for that Justin guy...but seriously, did you expect him to be any good with hair like that and a stupid smirk on his face?) and all they had to do was sing cover tunes for a television audience.  Yeah, real deserving.  And if I have to get into another argument with someone over how \"the pressure they suffer is so immense that they deserve to have a record and tour\" I am going to go Manson and stab someone in the belly with a fork.  Seriously, who can justify this argument?  Look at the facts:  1.) Kid stands in line for several hours for a chance to \"audition\" in front of an \"expert\" panel of \"judges\". 2.) If the kid can impress a fat, uninspiring album producer, a washed-up, has-been choreographer (who has no business telling someone they can or can\'t sing when she couldn\'t hold a tune herself), and a bitter, pretentious label head (who only got his notariety by doing that which makes the most sense...telling kids with no talent that they, in fact, have no talent), they are off to Hollywood...well, their parents, partners, whatever will take them to the airport where they will fly to Hollywood. 3.) Once in the land of stars, they will, once again, \"audition\" for the \"judges\", only this time they do routines and duets.  Ooh, that\'s difficult.  Singing with another person.  YAWN!!! 4.) They get selected and spend the next 8-10 weeks (if they survive) singing covers 2 nights out of the week, and spending the rest of the time in a tricked out mansion in Malibu.  Oh the horror.  This group right here are already guaranteed a single on a compilation album, and a tour later in the year...wow, I don\'t now how they can take all this pressure. 5.) The final 2 will perform live with a song that they have been \"working\" on for the whole time.  America (read: 30-million screaming girls and gay boys speed-dialing and text messaging Fox) votes on who they think is better, and boom, instant #1 single (because radio forces it on the airwaves 24/7 for a month), album, tour, etc.  Now, where in there is the pressure of having to deal with an audience that hates your music?  When do they get to see what its like for a band to have beer bottles chucked at them?  How are they deserving?  How are they stressed?  As far as I am concerned, all they do is sing some tunes, live in a \'Real World\'-type house, and have absolutely no responsibility whatsoever.  Fuck, they don\'t even write their own songs.  And here\'s further proof of how bullshit this show is.  That ugly red-headed kid, whatever his name is, did not have any business being in there.  He had no talent (pretending to croon like ol\' Blue Eyes is not talent...I can do that while taking a shit) to speak of and was about the whitest thing on Earth next to Carlton from \'Fresh Prince of Bel Air\".  I\'m sorry, but this needs to stop.  Something has got to give.  When will America see that all they are doing is flooding the airwaves with more disposable tunes and less-talented singers.  I don\'t mind pop music.  There are some really talented acts out there (Christina and Jessica, Outkast, etc.), but then there are these damn fake pop princesses that think they are the top of the world (I blame Britney..she has a cute face, nice ass, perky tits, but she can\'t hold a note to save her life and her songs are obviously written by a 54-year old man), and its these \"stars\" that have tainted the music world and have allowed the insurgence of crap that is...American Idiot.  '),(21665,3327285,'18,May,2004','2004-05-18 00:00:00','   So what started out as a pretty shitty start to the week turned pretty positive.  Monday I came into work pissed off from the obvious oversight that my boss made the week before in handing my job to someone other than me and realized that I had a 3-hour website developer\'s class to attend.  So that was a bit of a plus, seeing as this meant I didn\'t have to be in my building for the first half of the day.  Anything to keep my distance from the ball of negativity that has built up over here.  Anyway, so I attend this class, fully expecting it to just be cut and dry nonsense that I already knew, when I came to find out that the company is going to be creating a web developers position, and that I was the lead candidate for it.  Naturally I was high as a kite with this info.  I just came off of a bad weekend, and now, suddenly, I forgot the travesty of not getting the promotion.  So this made for a very good morning.  Then, once class was over and I got back to work, my boss pulled me into the office and asked me to pick out my co-worker.  Basically he felt that since I KNOW what is best for my department, that I should be the one to pick the team.  So I got to pick my co-worker that I have been trying hard to get over here.  Bonus happiness for me.  Now, I hear that not only do both of my girlfriend\'s children have ear infections, but now the youngest has a rash caused by the medicine she was given to combat it.  Not only that, but my grandmother informed me that my grandpa had a pacemaker installed to help regulate his heartbeat.  Great, so now I can only hope that equally momumental highs are in store for me, otherwise the title for this blog is quite false, wouldn\'t you say?  '),(21666,3327285,'16,May,2004','2004-05-16 00:00:00','   Well, this post finds me in a foul, yet vulnerable mood.  Recently two constituents of mine had been removed from their positions in my department for illegal behavior that cost the company some dough.  One of these people happened to be my immediate boss.  Now this wasn\'t some run-of-the-mill asshole by any means.  I have worked with him for several years (6 to be exact).  The same goes for my co-worker who was also let go.  She was a firecracker in her own right, but she was one of the guys nonetheless.  Getting to the point, with my boss\' removal, this opened up his position.  Now, not to take anything away from my ex-boss\' performance, but he didn\'t exactly make HIMSELF look good, it was a combination of efforts between myself and my ex co-worker.  I had worked under him in another department 2 years prior, and it has been brought to my attention NOW that it was my efforts then that made him look good enough to get the job that he has recently been removed from.  I practically did his job, and all at the same time I made him look good.  Of course, little voices coming at me have led me to believe NOW that after all I did for him, he spent a good part of his tenure trying to debunk my efforts and, in essence, have me fired for incompetency.  What the hell is this?  After all I did for him, all the whining I endured, all the time we spent hanging out and grabbing drinks had all been for naught.  Well, this left me feeling pretty shitty, but also it instilled in me the power to regroup and try for his vacated position, a feat that should have been rather simple...If it weren\'t for one person, the ass-kisser.  I had the job pegged.  It was mine.  My supervisor, the one doing the hiring, even told me the job was written for me.  But, there was foul work afoot.  See, I work in an off-site building from my company\'s main campus.  We are a small, autonomous operation that employees 8 people.  We have our own system of leadership that runs very well, and we are a pretty tight group of people.  But, the particular operations I am involved with here is relatively new to this group.  And one of the people that started in this building, an ass-kisser by nature and reputation, has been hard-up for a leadership role since the my group showed up.  He was always promised, but never saw results.  He was ready to leave, and tried unsuccessfully several times prior to secure a leadership position outside of our building.  When the firings went down, you would have thought he was a starving dog salivating over a freshly-cut side of beef.  He was about to sink his teeth into it, until I stepped up and noted my desire to lead my group.  Then he stepped back, feeling the hurt of knowing that someone was more qualified than him to get the position.  A short victory for me that would never be celebrated.  When he announced to the super of his reluctance to apply for the job because of me, the super should have accepted it as the best thing to do and left it at that....but no.  He told him to apply anyway, as he wanted a variety of good candidates to choose from.  I saw this, originally, as a proving ground to stake my claim in once the position was handed to me.  What I should have seen was an obvious ploy to hand over MY job to the resident ass-kisser that the super has had a boner for since the day they started the operations.    No, he\'s not gay...he just sees himself in this kid.  I say kid because I am 3 years his elder.  I also have 3 years seniority over him.  So I should have rightfully received this job by default...but this is not the case.  In the end, ass-kiss gets the job and I get the bullshit cookie-cutter explanation of \"He\'s had the leadership role in his eye a lot longer than you, and while you have the technical know-how and this position was practically made for you, I have to follow my instinct and go with what I feel is the better candidate.\"  Are you fucking kidding me?  Did he just say that?  This job was made for me...I am the most-qualified for the job...but because I didn\'t have the desire to be a leader (unbeknownst to said super, I have been trying to leave the building for some time in a leadership capacity, but nothing ever materialized from my leads), I lost out?  Are you on drugs?  How the hell can you begin to tell me something like this?  I should have seen it coming, but I was filled with so much pride from the number of people who wished me luck and assured me I would get the job that I was blindsided by this obvious case of favoritism.  I\'d take it to HR, but in all honesty, it would cause me more grief from my super thinking I hate him, and my new boss thinking I am trying to screw him over and cost him his newly-acquired job.  So, what do I do?  Simple...start looking for something new.  Either inside the company, or out, I am poised with my schooling and my skills to find something better...and to get out of this obviously lop-sided rule that is going to run the building into a brick wall.  I used to have pride in my job...now I\'m embarrassed I even work there.  Pity, because I make a good living and have a lot of time in it.  I just can\'t STAND to see undeserving people get a high-paying gig with responsibility they will obviously squander, or hand down to me for them to take credit for.  Ah, office life.  '),(21667,3327285,'13,May,2004','2004-05-13 00:00:00','   You know what really pisses me off?  Stupid people.  And I am not talking about those who occasionally have a brain lapse and do something stupid...no no no.  I am talking about people who thrive in their stupidity.  These people have as much stupidity as I have blood.  If stupidity were a currency, they\'d be billionaires.  I think you get my point.  There are times where I feel like God put these people on Earth just to piss me off.  They get in my way, talk their talk, and then argue to no end about their stance, only to annoy me and make me bow out in default.  Want an example?  My bank made an error in their favor that cost me roughly $60.  This made no sense to me, so I called up the customer support department, being that this is the only way to get an answer without listening to 20 different recordings all telling me how much they value my service and which button I should press to achieve nirvana.  So this half-wit answers (How do I know she\'s a half-wit?  Easy, she has a nice, thick Kentucky accent, that\'s how.)  I calmly explain my quandary and ask that it be rectified immediately.  I give out all my pertinent information she asked for, and I sat patiently as she fat-fingered my account number into the computer 3 times (I know she fat-fingered because no one hits 7 keys all at once without screwing something up...it was that loud..I heard them over the phone).  Anyway, so she types in all this information and then says I overdrafted my account, not once, but twice, and they charged me $60 in fees.  Here\'s the kicker...there isn\'t anything on my register or transaction log that shows where these could have originated from.  I put just enough money in this account to pay bills, and no more.  There should have been about $17 left in this account.  All my transactions added up correctly, and the total balance was $16.74.  I mention all this to her, once again in a nice, friendly tone.  She then goes on about how I had two debits, one for $21.20 and one for $1.00 in that overdrafted my account.  I asked who they were from, a little annoyed now because who the hell debited my account without my authorization?  She stated that these were from Verizon Wireless.  Ah, the cell phone company.  But what the hell, they got their money for my bill 2 weeks ago, what\'s this?  At this point I am logging onto my online banking page to see, in detail, what they debited.  I log onto the page, and there aren\'t any debits listed.  I ask her to verify her information, and then to make sure she was looking at my account.  She states everything correctly, but these two debits do not exist on my account.  I am getting a little steamed at this point because she\'s just typing away, telling me I owed them this money and that there is nothing that they can do to help me out.  I ask her why these aren\'t showing up on my account, and she says its because Verizon cancelled these debits before they posted to the account.    Great, so I shouldn\'t have overdraft charges then, right?  Wrong.  She claims that my account still overdrafted.  I ask her to prove it.  I told her to add up all that was on there and then subtract it from my beginning balance and tell me what she gets.  What was her answer?  $16.74.  What did she say then?  \"Well, then when you factor in the $60 in fees, its -$43.26.  WHAT?!?!  DID YOU JUST SAY THAT?!?!?!  YOU JUST TOLD ME MY ACCOUNT WAS IN THE POSITIVE, YET YOU STILL ADD IN THE OVERDRAFT CHARGES?!?!?!  DOES THIS NOT MAKE ANY SENSE TO YOU?!?!?  DID YOU TAKE MATH IN SCHOOL?!?!?!  She gets upset (and rightfully so) and tells me I need to speak with her manager.  \"Good, I was beginning to actually feel my IQ slip below 140 during this conversation,\" I so eloquently stated.  I heard her sigh as she put me on hold.  Ah the power of being an asshole.    So the manager gets on, tells me that I have upset his employee, to which sent me into another tirade that basically left him breathless and without any cocky/witty comeback.  He asked me to tell him what was wrong, I explained, and wouldn\'t you know it, he credited my account for the 2 erroneous fees, and issued a stern warning of \"Next time you call, try to be a little courteous to our staff as they are only trying to help.  If not, we could be forced to close your account.\"  I scoffed at his warning and told him, \"Do something stupid like this again and I will not only close my account for you, but I\'ll see your asses in court for incompetence, harassment, failure to provide service, and theft,\" and quickly hung up the phone.  Ah, the fun I have in my life with stupid people....what bliss.  And all this on a Friday.  Oh what fun will the weekend bring?  '),(21668,3327285,'12,May,2004','2004-05-12 00:00:00','   Well, it only took me 5 years to build up enough curiosity to look into this blogging phenomenon.  Hmm, and I thought I was on the cutting edge of technology-driven pop culture.  There goes my ego, deflated and bruised.  Ah, anyway, I guess a brief introduction is in order for all those who choose to view this atrocity of a blog (is this even a word?).  My name is Josh.  I was born, raised, and probably will die in Indiana.  I live a pretty normal life.  Well, I guess that\'s an exaggeration.  My life is quite unique.  I am blessed with the ability to sponge up any and all useless knowledge and store it away for use at a later date.  This is especially helpful when I pop into a pub with the fellas and someone rants on and on about a particular line uttered in a movie or what kind of music a particular media icon might enjoy.  Some brandish me a know-it-all, but I like to think of it as an eccentric breadth of knowledge that they are too jealous to appreciate.  Perhaps I\'ll be on Jeopardy one day.  Anyway, I work for a major, international diagnostics company based here in Indiana.  I make a good living, earn good money, live in a great apartment, and have a very beautiful girlfriend.  Oh, I\'m 25...just to clear that up.  Yeah, I know...I should be living in a house.  Well, that\'s another story for another time.  I think I have summed up the basics that will get you through these posts....oh wait, I almost forgot.  I am a rambler by heredity, so these posts will probably be pretty long-winded and bounce all over the place from topic to topic.  Dammit, I\'m eccentric, get over it.  Also, I am a pretty funny guy.  Most people say things like that and really aren\'t, but I seriously am.  I crack myself up sometimes with the words that fall out of my mouth.  Perhaps some of that will spill over into this........then again, this is just a glorified diary...the only comedy in that is the pathetic attempt at trying to make yourself seem witty when all you really are doing is talking to yourself, pretending someone will actually read this and think, \"Hmm, this guy is pretty cool, I wish I was like him.\"  Sigh, why do I even bother.  Perhaps I shouldn\'t have done this after all....curse you, insane curiosity.  Why do you torment me so?!?!?!  '),(21669,3327285,'29,June,2004','2004-06-29 00:00:00','   I feel like I am on an episode of Star Trek and this is my little captain\'s log.  I don\'t really have too much to write about at this point, but I thought I would drop a line on here for those precious few who are actually reading this (I got your emails...I am updating this finally...but for the sake of my sanity...please just leave a comment).    The first of my summer adventures has come and passed, and I miss it.  This was the family reunion trip.  Things didn\'t happen they way I wanted it to, and we only ended up spending about an hour and some change at the actual reunion.  But all in all, we had a good time.  The cabin we stayed in was absolutely atrocious.  It was obviously a fisherman\'s cabin for the lake we were on, and the cleanliness depicted this fact nicely.  The beds were big and comfy, but they were brown from obvious years of use.  As for the funtime merriment...well, there wasn\'t much.  It rained for much of Friday, leaving us stuck inside with nothing to do.  There wasn\'t any cable to speak of, so watching TV was not happening.  We ended up going to my family\'s house and hung out there for a while, basically having a mini-reunion if you will.  Because of the state of the cabin, we had to get out of there, so Friday night we spent driving around, trying to pickup a cell phone signal to make a reservation to a hotel.  I know what you\'re saying...\"Why don\'t you just go home?\"  \"Its not like you were far away or anything?\"  Well, we didn\'t want to go home.  We wanted an indoor pool for the girls to go swimming in, and I wanted to stay somewhere nice and different for the remainder of the weekend, so we reserved a room at a new Fairfield Suites in Clarksville, IN (just across the river from Louisville) and stayed there.  The girls loved it, and we slept like babies that night.  So in summary...we didn\'t go boating on the lake...the cabin was disgusting...we only spent an hour and a half at the reunion...and we ended up staying the last night in a hotel.  Sounds like nothing went according to plan and I should be upset, huh?  Well, you\'re wrong.  If there\'s one thing people have come to learn about me, its that I love the unexpected.  Sure, I love to have everything mapped out, but I allow myself to deviate from that plan if I have to.  And as far as I am concerned, we had the best time possible, and it made my love for my future wife that much more.  It was great...so what\'s next?  Our trip to New York has been cancelled.  Cyndi\'s (my woman) mother has decided she isn\'t going to watch Cyndi\'s children while we go to New York, so this has left us in a pickle.  So instead of breaking the court order to not take the children out of state without the ex\'s permission (which he will obviously deny just to be a cockhead), we have decided to just not go...which is fine considering that it would have cost us damn near $1000 to do it.  Money well saved.  So that frees up that weekend.  What we have decided to do is take the girls to Lake Michigan to go swimming that weekend.  I think they will love it.  It sure as hell beats spending a weekend with their grandmother.  Other than that, all the other plans are still on and I can\'t wait....bring it on...summer rules.  '),(21670,3327285,'13,June,2004','2004-06-13 00:00:00','   Well, this was by far the most whirlwind of all weekends, believe me.  Three shindigs of varying importance, and several events set to transpire in the coming months, all happening within a 72-hour time frame.  Below is a listing of what happened and what\'s to come.  You\'ll get a pretty CLEAR depiction of just how busy things have gotten.  Parties  Friday - My bride-to-be and myself attended my grandparents\' 50th wedding anniversary.  It was held in a quaint little room in a good-sized banquet facility.  The price of the food during normal dining is listed in the local paper as moderate, so you can imagine the type of fare we were expecting to have.  But never in my wildest dreams did I expect to be exposed to food of this caliber.  Lets put it this way, for those who aren\'t too keen on massive details...it sucked.  It was basically a buffet dinner, which I am not opposed to in the slightest.  The salad was the traditional roughage, with broccoli and cauliflower thrown in for good measure.  The choice of dressings was either fat-free Ranch (barf), a paltry excuse for French, or a dismal and bland oily Italian (the little lady hated it).  The dinner was worse.  Rock hard rolls with no butter, grease-infused chicken, very wet cole slaw, bland scalloped potatoes, and the worst cafeteria-style mashed potatoes I have ever tasted.  It was horrible.  But, it was quickly forgotten as we celebrated this huge milestone in the sacred bond of love.  Aside from the food, it was a fabulous occasion.  Saturday - My lady\'s older daughter celebrated her 4th b-day today at a bash we had thrown for us at The Children\'s Museum of Indy.  An expensive effort, but rewarding nonetheless when we saw her face light up in joy.  It was marred by an accounting error and a lack of cohesive organization on the part of the organizers.  Basically they wanted $50 more than they were owed and because they took too long to play dress up with the children, we had our party go over and the present-opening was cut from the itinery.  Pissed us off, but it was still fun.  Sunday - Went to the woman\'s friend\'s shipping-off party.  Her friend is heading to the Navy on Tuesday, so this is their send off to her.  Had some good food, mingled, watched some insightful TV, and learned that if you pet a male Pug dog on the back of the neck just right, his little pink rocket will pop out for the whole world to see.  A learning experience for everyone to indulge in.  Plans  Listed here are things we are supposed to do for the next 2 months.  This is in no particular order, but is on the agenda: Weekend camping trip Weekend trip to Kings Island with another couple Family Reunion in Southern Indiana Surprise 50th b-day party for my woman\'s aunt in New York My niece\'s 5th b-day party Fourth of July cook-out with my family Company picnic at the zoo Company outing to Kentucky Kingdom  This is all things I have learned of planned in the past week, and summer hasn\'t even officially begun yet.  Oh what fun lies in the near future.  Also, stick in a potential trip to Cedar Point with the lady\'s boss, and yet more hilarity ensues.  Ah, summer...just a couple days ago I was yearning for the old days...now I feel like my wish is granted...more things to do than I have time/money for.  '),(21671,3327285,'08,June,2004','2004-06-08 00:00:00','   Hello all...me again.  So another Memorial Day weekend has come and gone and with it, the childish splendor that is summer vacation.  I remember a time when I loved Memorial weekend.  The Indy 500, barbecues, no more school, road trips with the family.  Times were a lot simpler and fun.  But now, I look back and I feel empty.  A void that has replaced that warmth and happiness.  Why?  What am I missing?  Ah yes....youth.  Looking back, I feel old.  I know it was only 12 years ago, but when you think about it, that\'s a really long time.  That\'s infancy, toddler, elementary school-age, puberty, and pre-teen, to put it into perspective.  And thinking about THAT alone is enough to make me sick to the stomach.  Gone are the carefree days where we can wake up at noon and go to bed at 3am.  Gone are the days of calling up your friend and popping over to watch their father\'s pornos while talking about the intricacies of a boner and how big a certain schoolmate\'s breasts are.  Gone is the excitement of your parents telling you that the whole family is heading to **insert theme park name** for the weekend or to **insert family gathering** in **insert state** for a week.  And gone is the passion and desire of a young couple\'s summer love, tinged with sexual innuendo and heavy petting in a conditioned and controlled \"exploration period\".  Yeah, while some people look at Memorial Day weekend and think about warm weather, swimming pools, and camping trips, I sit back and sadly recall a time when I didn\'t care about responsibility, and I didn\'t worry about how much money I have or if I will be able to pay all my bills.  I solemnly reminisce on a period of my life where I genuinely felt happy, with no need for outside companionship or monetary stability.  Most people love this time of year....I dread it.  It brings up painful memories of a doomed first relationship.  It leaves me pondering the kind of summers I\'ll be able to provide my girlfriend\'s children (and perhaps our own if we ever make it that far), but most of all...it leaves me longing to be young again, in the thick of adolescence, where my only concerns and worries are who\'s house are we setting a tent up in and how much trouble am I going to be in because I am 20 minutes late in coming home for the night.  Ah, growing up....how I loathe your rank embrace of adulthood.  '),(21672,3327285,'06,August,2004','2004-08-06 00:00:00','   Well, it was bound to happen sooner or later...I post another entry.  As you can tell, I am not really jumping into this like its an actual diary or anything. Instead, I am using this as a tool to vent my frustrations. My lack of posting has to do with the fact that I have been, for the most part, not very frustrated. Sure, there\'s the usual crap that puts a frown on my face, but for the most part, things have been pretty close to tip-top.  So then why am I writing now? Why not shut up and go away until I have something juicy to scream about? Well, because I am bored at work and have no better way to spend my break than to jot some things down. A few \"projects\" in various forms of completion have crept into my social life, leaving me pretty winded. What are they?   Internet Radio Station  That\'s right, I am working on my very own internet radio station. Now, its not going to be anything huge like  urlLink www.knac.com  or anything of that nature, but it is a springboard that will hopefully launch a side career for me. As you may (if you know me) or may not know, I used to be involved in radio heavily. I worked at a couple stations in Cincinnati, and I even received certification from the Ohio Center for Broadcasting. So as you can see, I am qualified to make this happen. Basically what this entails is running a server from which I will broadcast various types of music through. Am I going to DJ? Well, that depends. As for right now, I am just running music through automation, and that provides for uninterrupted music for as long as I leave the station up. Sometimes its an hour, other times its for a day. It all depends on what I have going. Right now, I mainly use it to listen to something at work, and there is enough slots for 1 other person to listen to what\'s playing. So, once things get finalized, you\'ll find out here what\'s up.   E-zine  OK, this isn\'t that huge, but basically I am in the process are starting up an e-zine with a couple friends of mine. What is an e-zine exactly? Well, its really something similar to this. Basically its a web-based newspaper that is going to cover various points of interest revolving around the arts. Aside from being the webmaster, I am also going to be a bit of a contributing editor. I\'ll have a couple columns of differing topics, and perhaps my own review section for albums, movies, books, whatever I can bother myself with. My personal life takes a front seat to all this, and therefore I need to make sure I have my priorities going before I dedicate time and money to that venture.   Novel(s)  Yeah, you read that right. I am working on several novels of different subject matter. I am not going to drop titles or anything as of yet, seeing as I put quite a lot of thought into these and I would hate for someone to steal these titles and make them their own. Screw that, this is my work, dammit, go create your own. I have a semi-autobiographical in there, as I have lived quite an interesting life, mixed with some fiction to spice up boring parts. I am also working on a comedic space epic in the vein of Douglas Adams\' fantastic \"The Hitchhiker\'s Guide to the Galaxy\" series and Monty Python. I am also working with a couple friends on a reference guide to all things related to rock. Its not an encyclopedia of rock music that you usually think of...no no no. Its actually a listing for up and coming bands who want to try to get into the biz. Basically I was having conversations with a couple friends of mine who are in bands and how they are sick of not being able to cut a break because of lack of connections. Well, considering I still have a few friends in radio, I tend to have some pretty good connections in this regard, so I decided to put together, with their help, a guide for young, aspiring bands to help get them in the door. I have venues, promoters, studios, rehearsal spaces, and small, independent record labels all listed with write-ups on them. This is going to be a must-have for bands that are deciding to take it serious but not knowing what to do next. This has the potential of being released first, as there are a lot of interested parties who\'d love the info we\'ve gathered. Well, there\'s no such thing as a free lunch, so you\'ll have to buy it like everyone else.  Anyway, that\'s what\'s been up. I just realized that I didn\'t say a word on this blog for the whole month of July, so I guess this update to what\'s been up was a bit rewarding. Until I get pissed again, this is your old pal Josh.  '),(21673,1225881,'09,April,2004','2004-04-09 00:00:00','   So I have decided to spruce up the old blog (Laura and Savannah have finally bugged me into doing it...)  I\'m going to try to update, but I\'m not making any promises.  What\'s new with the Brantino you ask?  Well, nothing much; here\'s the quick and dirty:  Still no solid prom date and anxiously awaiting to hear from GSP so I will know if I will be losing five weeks of summer.  I have to get reading glasses in order to prevent my eyes from getting to the point where I really need glasses.  I got flipped off while driving...for cutting off another car. OOPS!  My sister claims to have a running freshman poll (I\'m thinking about 2 people) that if I grew my hair out long, I would become a \"hottie.\"  Hmmmmm...  If I ever do change my hair style, it won\'t be because of a freshman poll.  5 weeks of school left...Summer plans anyone...???   '),(21674,1225881,'12,April,2004','2004-04-12 00:00:00','   I just spent almost 2 hours reading the 9/11 timeline mentioned today in Leander Kahney\'s blog.   It\'s really scary  to think that so many people, including high ranking government officials, could have had knowledge of the attacks before they happened.  I leave tomorrow for Louisville and will not return until Thursday.  I started to work on my homework, but, of course, got sidetracked.  That leaves me with two chapters to read in my AP book and a  TON  of review problems for math.  Fun.   '),(21675,1225881,'11,April,2004','2004-04-11 00:00:00','   I had to get up REAL early today.  I went to Easter sunrise service at Mount Pisgah and had to get up at 4:30.   Yuck.   Lot of people from our church came as well as Lindsey and crew.  When I got home I sat down on the couch at about 10 and ended up falling asleep and sleeping until 2.  I really didn\'t mean to do that!  This afternoon Greg, Torrie, Joe Paul and Katie came over to hunt Easter eggs and Greg, Torrie, Mom and I ended up doing to Francis BBQ where my arranged prom date works.  I was afraid the situation would be a little awkward, but it really wasn\'t that bad.  Mary (The arranged prom date) was  REALLY  funny and I think we will have a good time.  Should I call her and ask her if she wants to do something between now and then?  I think I might try, but she works a lot.  Ah, decisions, decisions...   '),(21676,1225881,'30,April,2004','2004-04-30 00:00:00','   I have come to the conclusion that I am uncool.  No after prom plans, no life, no after prom plans, my tux is all messed up and errrr....I don\'t know what to do after prom is over.  Plus, I haven\'t been posting.  How uncool.  I applied for FBLA Reporter because I probably won\'t get to go to camp...and somehow I got VP of Competitive Events.  I\'m happy, I don\'t really care what office I get, I just hope no body gets mad.  Prom is gonna look pretty cool.  I hope it rocks.  I also I hope I don\'t get slapped in the face with a steak at Showgun.  I never eaten there before, but I have heard that they throw food.  Maybe not the best place to eat before prom?  I should sleep....kinda got a big day tomorrow.  Pray that I make it though the day....and that my date won\'t want to kill me.  I might just croak from all this prom crappage anyway…  tino   '),(21677,1225881,'12,May,2004','2004-05-12 00:00:00','   I\'m mad at my car.  I\'ve been nothing but good to it, I even washed it yesterday, and what does it do?  Freakin short circuits.  I can still drive it, my power lock on the driver side door just went out.  This may seem small, but it keeps blowing the fuse, which also has something to do with what locks the gearshift.  So now every time I break my gearshift clicks telling me it\'s ready to change gears, at 50 MPH.  Lets not think of the cost of replacing the little motor inside the door.  Gillie Hyde doesn\'t even sell wiper blade for under $100.  (If I paid that much for wiper blades they had better be gold plated or something.)  The other day I found out I didn\'t get Senior Engineer, which isn\'t a big deal at all.  It\'s just that a few of the people that go it don\'t seem like they would be good team leaders or are overall insane.  I\'m working on a diabolical plan to sabotage it all next year...can we say wireless access point on the same frequency as their wireless network hidden in my backpack? (They\'ll never find it! MAH HA HA!)  In all seriousness, I wish everybody next year great success and hope everything goes well.   '),(21678,1225881,'11,May,2004','2004-05-11 00:00:00','   One week left... then graduation... then GSP... Wow everything is flying by.  I thought it was after your senior year that everything was supposed to change, but it seems like everything is changing now.  My dad quit his job, his last day is Friday.  I\'m happy for him that he will be able to take some time off an do what he wants to, farm, ect.  It just feels weird because my dad has worked at the same bank since before I can remember and I can\'t fathom the idea that he won\'t be working there anymore.  I guess I\'ll get used to it...  Graduation will be hard this year because I know so many of the seniors.  At the same time it will be a reminder that we only have one year left.  It doesn\'t seem like it was that long ago that we were Freshmen.  I guess I\'ll get used to it all...   '),(21679,1225881,'17,May,2004','2004-05-17 00:00:00','   It\'s late and I can\'t talk much.  Tommorrow is the last day of school as well as Graduation.  (Good and bad)  I leave you with this random thought:   '),(21680,1225881,'16,May,2004','2004-05-16 00:00:00','   2 more days guys...2 more days.  ...and half days at that!  The review sheet in chemistry just reaffirmed how much I want out of that class.  I\'m probably not even taking the final in any of my other classes.  I\'m bad, I know.  I can\'t wait for summer is all I have to say.   '),(21681,477665,'20,September,2002','2002-09-20 00:00:00','    salsa dancing....  cafe se villa is the place to be on a thursday night if you wanna learn salsa. it never really interest me but i\'m open to new things. so when my girlfriends suggested we go, i was in. it was pretty cool because they give you lessons before the dancing really begins. you know me...i\'m a little shy but the atmosphere was so welcoming and friendly that dancing with strangers was natural.   here\'s the difference between a salsa dancer and a clubbin\' guy: the salsa dancer is a lot more respecting...not trying to hit up on you, just wanting to salsa and teach you how. a clubbin\' guy (not all)...will try to hit on you, grab your ass, caress your back, kiss your hand, pretend to trip onto you, eye you up and down in the most perverted way and if you turn him down, he moves right along to the next girl and does his routine all over again. that\'s just my personal observation and i\'m sure many of you would agree.     anyway back to my story...salsa dancing is a lot of fun especially if you have a good partner. i was spinning, turning, twisting, dipping, smiling, and laughing...just having a great time. oh yeah and a lot of hip action is involved too. it was cool because you\'d switch partners all night long. all the guys i danced with were hella good at salsa...they made me look good..haha me and my two left feet.   anyways, i think my clubbin\' days are catching up to me. i have a bit of cold right now...sucks since i haven\'t been sick in almost 2 years. i must have jinxed myself some time ago. i\'m sorry kim, i promise not to make fun of your everlasting cough ever again...please take it back...i don\'t want it!    it\'s friday again and i should be going to LA for antionette\'s birthday, the big two-three,  but i feel like shit and i have to move this weekend...i\'ll be homeless for a few days...anyone wanna take in a poor girl??? i\'m a good roomie,really!!  well, i\'ve been waiting all week to party in hollywood...damn, planning never works for me. i\'ve been lucky so far but recently nothing seems to be going my way. oh well \"whatever happens, happens for a reason\"...ummmm...don\'t know what the reason is but hopefully i\'ll find out soon.   '),(21682,477665,'16,September,2002','2002-09-16 00:00:00','    actions vs. words...  i\'m a very strong believer that actions speak louder than words. but at one time in my life, i remember thinking why couldn\'t he express how he felt, why didn\'t he write me poetry or even letters, why was it so hard for him to tell me sweet nothings. in a way, i think i felt i was missing something because he didn\'t do any of  those things and with this mentality, our relationship went a rye.  and that was the beginning of the end.  it\'s been 9 1/2 months since we broke up. i know he was hurting alot and to this day, i still feel that he\'s bitter...not sure if that\'s a good word choice, but i still feel the anger and resentment that he still has for me everytime i see him.  when we broke up, i really felt that we could be great friends, that things wouldn\'t change because i considered him my best friend for so long but that didn\'t happen. when we broke up, i lost him, i lost my best friend...the one person who knew my in\'s and out\'s and at times believed knew me better than my own family. even though he had a hard time expressing himself, i knew he loved me with every cell in his body. i knew he would do anything for me,  i knew he would never break my heart. he\'s one of those rare guys out there that is actually good and what i mean by good is that he truly has a loving and trusting heart. not many out there. let\'s just say, whoever he may end up with is one lucky girl.   this whole experience has been a very painful one. i\'ve learned a lot about people and have come to realize what i should have known 9 1/2 months ago. i\'ve reached a point in my life where words mean absolutely nothing until they prove that they can deliver. people can tell you one thing and really mean something else, people can lie to your face, people can twist your words, people can just be simply heartless when they want something so bad that they would say anything to get it. shut up already...heard it before, don\'t need to hear it again. i use to want someone who could talk and express words easily but after knowing what i know, i rather have a mute then a broken record. my only conclusion to this is...stay away from chatter boxes..you know the ones that tell you  everything  you want to hear.   '),(21683,477665,'15,September,2002','2002-09-15 00:00:00','    hmm....another way to look at it  \"Simple Truths\" by Kent Nerburn \"It is a mystery why we fall in love. It is a mystery how it happens. It is a mystery when it comes. It is a mystery why some loves fail. You can analyze this mystery and look for reasons and causes, but you will never do any more than take the life out of the experience.   Love is more than the sum of the interests and attractions and commonalities that two people share. And just as life itself is a gift that comes and goes in its time, the coming of love must be taken as an unfathomable gift that cannot be questioned in its ways.  Too often, when love comes to people, they try to grasp the love and hold it to them, refusing to see that it is a gift freely given and a gift that just as freely moves away. When they fall out of love, or the person they love feels the spirit of love leaving, they try desperately to reclaim the love that is lost rather than accepting the gift for what it was.  They want answers where there are no answers. They want to know what is wrong with them, or they try to get their lover to change, thinking that if some small thing were different love would bloom again. They blame their circumstances. They blame each other. They try anything to give meaning to what has happened. But there is no meaning beyond the love itself, and until they accept its own mysterious ways they live in a sea of misery.  You need to treat what love brings you with kindness. If you find yourself in love with someone who does not love you, be gentle with yourself. There is nothing wrong with you. Love just didn\'t choose to rest in the other person\'s heart.  If you find someone else in love with you toward whom you feel no love, feel honored that love came and called at your door, but gently refuse the gift you cannot return. Do not take advantage, do not cause pain. How you deal with love is how love will deal with you, and all our hearts feel the same pains and joys, even if our lives and ways are very different.  If you fall in love with another who falls in love with you, and then love chooses to leave, do not try to reclaim it or to assess blame. Let it go. There is a reason and there is a meaning. You will know it in time, but time itself will choose the moment.  Remember this and keep it in your heart. You don\'t choose love. Love chooses you. All you can really do is accept it for all its mystery when it comes into your life. Feel the way it fills you to overflowing, then reach out and give it away. Give it back to the person who brought it to you. Give it to others who seem poor in spirit. Give it to the world around you in any way you can.  Love has its own time, its own season, and its own reasons for coming and going. You cannot bribe it or coerce it or reason it into staying. If it chooses to leave, from your heart or from the heart of your lover, there is nothing you can do and nothing you should do. Be glad that it came to live for a moment in your life. If you keep your heart open, it will surely come again. \"   '),(21684,477665,'12,September,2002','2002-09-12 00:00:00','    just one of those days...  today sucked, let me tell you what a crappy day i had today first of all my business trip to vegas in october didn\'t get approved...my boss said i could go yesterday and then called me today while he was at a seminar to tell that my trip got denied by the ceo for some lame excuse like they needed me here during that time...b.s. he didn\'t even have the balls to tell me face to face...when i  found out i was gonna go to vegas, anxious tiff booked her flight on southwest.   secondly, i scheduled a virtual seminar this morning to demo one of the products that we\'ll be implementing for hr/payroll. i failed to read the email that the seminar would start at 10:30 eastern time....EASTERN time!!! they should have bolded it or something.  i brought my laptop and projector down into our conference room and got set up for our clients. logged on at 10:30 OUR time and got an error message saying \"the meeting is over\".  it actually started at 7:30 am. damn i felt stupid. so with my head down i packed everything up and brought it back to my cubicle...all of the sudden it hit me that i needed the laptop and projector in that same room for a training session that afternoon...haha.   oh and then my car starts making weird rattling sounds...i found out it\'s the brakes...hasn\'t been changed since 1998. called my hero, my dad and he came to the rescue on that one. i guess shit happens. well all i know is that tomorrow is definitely going to be off the wall...it\'s friday...you all know what that means...shake that bootie!!!   '),(21685,477665,'11,September,2002','2002-09-11 00:00:00','    My favorite...  movie- Seredipity...reminds me to believe in fate and destiny restaurant- those who know me....onami all the way...anyone wanna challenge me to an eating contest?? hehe number- 037...leo spelled numerically backwards...it\'s my sign place- i would have to say everywhere and anywhere with the right people jewel- ruby baby...ruby or maybe..diamonds are a girls best friend sport- snowboarding and yes it is a sport food-sushi: shrimp tempora roll... NO eggs.... yummy drink- powerblaster, surfer on acid, lemon drop, malibu and pineapple juice...hey who said i had to pick only one. ice cream- rainbow sherbert, what else?! fast food- in and out tv show- definitely friends! not even a cute boy can keep me away from this..haha flower-stargazers tulips/lily and long stem red roses w/baby breaths music- hip hop, trance, reggea...pretty much anything that i can dance to color- red pastime- vegas with the girls...underage...who said we couldn\'t gamble... song- as of right now..\"i\'ll fly with you\" by gigi d\'agostino, it\'s been on repeat forever... saying- \"whatever happens, happens for a reason\"...wait i have another favorite saying too... \"to the world you may be one person but to one person you may be the world\" club- hotel ibiza...my first  ee xperience and definitely not my last  k, i can\'t think of any other favorites but had fun making this list. until next time...   '),(21686,477665,'16,October,2002','2002-10-16 00:00:00','    identity theft….  have you ever had a gut wrenching, heart pounding, head rushing and hands shaking experience all at the same time? well yesterday i felt all this and more.  i felt so sick to my stomach that my entire afternoon, i felt like vomiting.  here’s the story….  on october 15, 2002 i received a phone call from a man name nick perkins, a bank manager at the washington mutual branch in houston, texas.  he tells me that they have a lady there who claims that i wrote her a check for $2500.00 and ask if i did in fact do this.  my heart stopped for a split second as i tried to grasp at what he was saying. i couldn’t believe this was happening to me. with a shaky tone, i tell him “NO.”  all of the sudden, i felt my mouth rambling on and on with endless questions of who, how, when, what, and why.  the answers I received became more of a nightmare and more dissatisfying then expected.  here’s the information i received on this lady:  name:  Jamie Clay  gender:  Female  ethnicity:  African American  age:  Mid twenties  drivers license:  Arizona  apparently, jamie clay tried to cash a fake check at this washington mutual bank in texas. she knew my social security number, my mother’s maiden name, my driver’s license number, and my date of birth. god knows what other information she knew. when the bank teller questioned her about the check, she claimed that we had a fender bender one night at a club and I paid her off.  luckily the bank teller was suspicious of her and notified me before cashing the check.  he asked if I wanted to press criminal charges and of course i answered, “YES!”  they called the police and once the police arrived at the texas bank, he gave me a call.  here’s the officers’ information:  name:  Deputy Schields  gender:  Male  department:  Harris County Sheriffs Department  phone #:   deputy schields asked me a series of questions and after I confirmed everything, he said he had the lady under arrest and that he was going to take her to jail. damn hooker, she deserves everything she’s getting. anyway, he gave me the case # and said he would be in touch.  down at the rancho san diego washington mutual bank, i spoke to naomi who was in charge of fraudulent activity division. she closed my accounts and gave me instructions on how to proceed with getting outstanding checks paid so they would not bounce.  now i have to go through the hassle of canceling all my outstanding checks and everyone knows how horrible i am at balancing my checkbook so this will definitely be a pain in the ass. i’m seriously thinking about switching banks; i have nothing against WAMU but better safe then sorry.  i also had to cancel my credit cards just in case that information was also stolen. i’ve called all 3 of the major credit bureaus- equifax, experian and transunion so they could put an alert on my accounts.    i’m so sick of this bullshit….why can’t people just work hard for their money like the rest of us???? i still have a lot of phone calls to make but hopefully it will be all over soon. i have mixed emotions right now…i’m pissed that my identity has been stolen, my private information compromised, i feel so violated, i feel sad that there are evil people in this world that would go against what’s right to make money, i’m puzzled to how and who else has my pertinent information, i\'m overwhelmed with stress, and now more then over i’m suspicious of people. i hate feeling like this. i know I trust a little to easily, which is one fault of mine.  i’ve been told many times that i’m too soft hearted. well no more! i’m fed up with being taken advantage of.  so don’t mess with me or you’ll end up like Jamie…in jail. that bitch definitely messed with the wrong tran!   '),(21687,477665,'10,October,2002','2002-10-10 00:00:00','    to the skies....  you know sometimes i really amaze myself. i never thought i would actually carry out a new year\'s resolution and for the first time in 23 years, i did it...yes ME!!  in 2002, i decided that i would exercise a lot and also do something i\'ve never done as a child...travel by plane.  i swear it was only 2 years ago that i actually flew on a plane...i guess you could say i was an airplane virgin.  but this year changed dramatically. i\'ve been on more airplane rides then i count on all fours. everytime i\'ve gone, i would meet the most random people. twice i ran into Noel, this guy i use to live with in the dorms and we even flew to the same destination...twice!  i also remember this one time (and no not at band camp) i sat by this women in her 40s, never married who tried to give me advice on men. she said \"they\'ll no good and if you wanna be happy, never rely on a man\". the duration of the trip consisted of male bashing. she\'s definitely been burned way to many times in her life. and then i met this other guy and a bunch of his friends on the plane. they were all from san diego. we exchanged numbers and have been friends ever since. i still remember the night at bar dynamite...giggles galore! but now my southwest buddy has moved to the city (sf) which i guess is cool since you all know how much i love to visit san francisco.   i\'ve never had a dull adventure at the airport, on the plane or at my desired destination. it\'s a memory i will soon not forget. i love flying, i love traveling, i love meeting new people. just being in a different environment, surrounded by unfamiliar faces, and a little power blaster...i lose all inhibitions, which isn\'t bad..not bad at all. i guess in a way, i could compare running to traveling. those who know me, know i love running outdoors. there\'s something serene about running in the open. all is see is beauty, a beauty so pure that i don\'t want to stop. as the wind blows gently across my face and as i start spacing out into the world before me, i feel free. no worries, just free. and so i guess running and  traveling gives me the same high. it\'s nice to get away...away from the real world.   so my next destination, vegas baby, cha-ching!   '),(21688,477665,'29,April,2003','2003-04-29 00:00:00','    rub-a-dub-dub...  i\'ve been told by a certain individual (i won\'t mention any names but u know who you are!) that i suck at giving massages. dude, how hard could it be? i mean, i\'m taking time out...time where i could be sleeping or eating...to what?!?..give you a back rub and you complain?? once, i was told on how to give one...\"you shouldn\'t press on the spine..that doesn\'t feel good\"...what the heck do i look like...a professional masseuse..if you want a professional massage, this isn\'t the place to get one!! anyways, i\'m not sure why i\'m blogging about this because it\'ll probably come back and bite me in the ass. oh wells..so continuing with my story..this person had the nerve to tell me that and ever since then...i stopped giving massages. my excuse: \"well someone said i suck so i don\'t give \'em anymore\".  well it didn\'t work...my excuse.  the response i would get back would be \"i don\'t believe you, here, give me one and i\'ll tell you the truth, i\'m sure you\'re really good.\" damn..i thought i got away with that one.    it\'s weird...for the longest time, i wouldn\'t want to give anyone a back rub. they would ask and i would say me first and fall asleep or when my turn was up, i\'d say \"i\'m hungry...let\'s go eat\" and then the person would forget all about the massage. ok..i know that\'s awful but i\'m lazy...hey at least i\'m honest.  anywhooo, something has come over me...something strange. i find myself doing things i said i would never do. here\'s a classic example...those who know me, know i don\'t watch movies in theaters especially if the guy thinks it\'s a date. i haven\'t seen one since july but about two months ago i did. here\'s another example...i said i would never take anyone to my solo spot..yet i\'ve been contemplating about making it into a hang out spot. and now this...here it goes...(*takes a deep breath*) i volunteered to give a massage!! it\'s shockin\' i tell you. but here\'s the twist, i still didn\'t give one yet i received one. awww...ur just simply amazing.  well i guess i\'m just not meant to give massages but you know what..i\'m a sport at receiving them...haha!!!  have a great day everyone!!   '),(21689,477665,'25,April,2003','2003-04-25 00:00:00','    a prayer...  days would past by, where u didn\'t hear from me...weeks gone by, and you still havent\' seen me.  a quick phone call here and there to let you know i wouldn\'t be coming home was all that was given to you.  when i finished my last course, i know you thought i\'d be home more cuz i didn\'t have school as an excuse to not come home. but the complete opposite took place and the visits and phone calls became less. it had been awhile since i\'ve been home, i\'ve was too wrapped up in all the fun..never wanting to plant my foot on the floor..always on the go.   last week my older sister came home from stockton to visit. i finally had a reason to go home not that i never wanted to, i just figured everyone was busy but doing fine and that i wouldn\'t be too missed if i skipped a few weekends.  what i went home to was a complete shock..a shock i could have lived without. instead of coming home to a house filled with love, laugher, smiles, and food...i came home to a house that was empty, no laughter, no smiles, and no mom\'s cooking. i decided to call my sister..thinkin\' they were all actually at my grandma\'s house.  i stepped in my parents room to use the phone and to my surprise, my mom was plopped on the bed all wrapped in her blanket. it was only 3:30 in the afternoon. as she lifted her head to see who was there, i jumped on the bed, hugged her tight and said \"mommy i\'m home!!\" but the response i got back was a weak sigh. i took another look at my mom and realized she wasn\'t feeling well. if you all have seen my mom..you know she\'s a heathly women..a very strong women. but today was not the case..actually it\'s been a few days that she\'s been like this. she looked drained, she looked weak, she looked sad, she didn\'t look like mom. i asked her what happened, i asked why, i asked her when but she was too tired to talk. i laid in bed with her as i held her in my arms. we feel asleep together and was awoken 2 hours later by my sisters.  they said they got the prescription for vicodin in order for her pain to subside.. the pain..wait...what pain? what was she feelin? what happened? why was no one was talkin? everyone kept quiet when i spoke, endless questions flew from my mouth..yet no response.   i still didn\'t know what was wrong with her. i don\'t like being in the dark. i pulled my sister aside and asked her again to tell me what happened to mom. she said \"while you were in vegas, mom went to the emergency room.\" i closed my eyes as she began to tell me the story. everything and nothing started running through my head. as i opened my eyes again, all i could say was \"why didn\'t you call me..i don\'t care if i was out of town..i need to know these things. i want to know if anything happens to anybody in this family. i have the right to know.\"   my mom is a fighter..it was only 6 years ago that the doctor told us that we might lose her but we didn\'t. she fought that battle and won. this is just another battle that lays on her shoulders now and the outcome of this one will be the same.   it sucks that something bad has to happen in order for you to see what really matters in life. all the mini drama episodes..not so important anymore. i\'m gonna go home a lot more now, i\'m gonna be there for my family, i\'m gonna do whatever i can to help my mom get better and when she does, i\'ll still be going home.  i just hope this is enough. Not a day goes by that i don\'t think about her or pray for her. mom,  i\'ll always be here for you..u can count on that. i luv you.   '),(21690,477665,'24,April,2003','2003-04-24 00:00:00','    a days dream...  daydreaming is what i do best. if i had to pick a talent (besides eating), i would have to say that i\'m the best daydreamer alive. from the moment i click snooze, and i\'m half awake, i\'m daydreamin\'..i don\'t really care to share these dreams but i find myself having the same dreams all day long...like right now.  daydreaming is purely heaven to me. i can\'t snap out of it...my mind wonders off and i don\'t care to come back. many times i find myself daydreaming at my solo spot. i found it one day as i was daydreaming in my car...something led me to this place and ever since then, i\'ve been going back just to daydream. ahh..i think it\'s time for my nap :-)   '),(21691,477665,'23,April,2003','2003-04-23 00:00:00','    peas n\' carrots  it was a year or so ago when we decided this was it. little did we know that fate had something else in-store for us.  i can still remember the long talks of dreams that we held on to for the life of us.  always thinkin\' things would change..it would get better, that it had to. time after time..we were proven wrong. our circumstances were different but our problems the same. the tears rollin\' down our cheeks as we drove to work, the gut-wrenching pain we shared when the hurt was unbearable, the sadness in our eyes as we tried to escape the world...but you were there to share this with me and i with you...a burst of emotions, a few long-lasting silence, a shoulder to lean on, an ear to listen, and comforting hugs, but ultimately what did it was a kick in the ass back to reality.   our story doesn\'t just end there, our paths became one in the same. a life changing experience effected both our lives. there was no more of \"them\" as we both closed our eyes and buried our heads in our lap and cried like the rain that falls upon us on a winter\'s day. we were now alone. our hearts crushed, our dreams shattered and our hopes dead.  little did we know that this was the best thing to happen to us...  it was just the beginning..a new beginning for the both of us. we became eachother\'s side kicks..inseparable at the hip. cruised up and down southern california...san diego, irvine, santa ana, la, and hollywood. we eventually made our way further north...destination...well vegas, san francisco and tahoe baby.  i\'m telling you..if our parents knew of what we did..they\'d shoot us. 2 girls driving all the way to la just to club..and then drive back the same night...i swear there were so many times that i would pinch myself just to stay awake at the wheel. the clubs were fun, we met a lot of strangers and we\'d like to keep em\' that way...who cares if they thought we were stuck up or bitches...as long as they didn\'t touch..we were cool with em\'. *news flash* *news flash**..we go to dance not to get picked up on!!!  i swear i think we heard every line out there...you boys must all read from the same book. but besides that..we always had a good time..if we weren\'t eating, we\'d be clubbin\', if we weren\'t shoppin\', we\'d be clubbin\', and if we were drinkin\', we\'d be clubbin\'!! we became clubrats.. i can\'t believe i just called us that but you know what..we were. the music, the vibe...it made us happy and that\'s exactly what we needed.    it\'s crazy the way things work out. we\'ve learned alot and i think this whole experience has only made us stronger. if we were to ever fall again, it won\'t ever hurt like it did before. we\'ve already reached the lowest and we both know that once you\'re down..the only way to go is up. and that\'s what we\'ve done..picked ourselves up and looked at the world right in it\'s face. so now that we\'ve overcome our fears and opened our hearts again, we\'re lettin\' in a whole new level of uncertainty and vulnerability. but you know what?!?  i don\'t care cuz i know myself now. i know who i am. and i know what i\'m worth but i will tell you this, i\'m glad we\'re going through this together! i wouldn\'t have it any other way!!   \"A true friend is someone who reaches for your hand and touches your heart.\"  --you are the very meaning of this quote.   '),(21692,477665,'22,April,2003','2003-04-22 00:00:00','    A poem for us girls....   (i never really read forwards but i thought this was hillarious so i decided to post it...haha! enjoy! )    \"I shave my legs, I sit down to pee.  And I can justify any shopping spree.    Don\'t go to a barber, but a beauty salon.  I can get a massage without a hard-on.    I can balance the checkbook, I can pump my own gas.  Can talk to my friends, about the size of my ass.   My beauty\'s a masterpiece, and yes, it takes long.  At least I can admit, to others when I\'m wrong.    I don\'t drive in circles, at any cost.  And I don\'t have a problem, admitting I\'m lost.    I never forget, an important date.  You just gotta deal with it, I\'m usually late.   I don\'t watch movies, with lots of gore.  Don\'t need instant replay, to remember the score.   I won\'t lose my hair, I don\'t get jock itch.  And just cause I\'m assertive, Don\'t call me a bitch.   Don\'t say to your friends, Oh yeah, I can get her.  In your dreams, my dear, I can do better!   Flowers are okay, But jewelry\'s best.  Look at me you idiot... Not at my chest??  I don\'t have a problem, With Expressing my feelings.  I know when you\'re lying, You look at the ceiling.   DON\'T call me a GIRL , a BABE or a CHICK .  I am a WOMAN. Get it?, you DICK!?! \"   '),(21693,477665,'22,April,2003','2003-04-22 00:00:00','    happy birthday!!!!!!  \"Go, go, go ricky It\'s your birthday We gon\' party like it\'s yo birthday We gon\' sip Bacardi like it\'s your birthday And you know we don\'t give a fuck It\'s not your birthday!\"   -haha... there\'s my birthday shout out to ricky. hmm... the big... 2 .. 8 ... nah that\'s not old at all... it just means you\'ve already passed your prime time, you\'ve celebrated your golden birthday, you\'re reaching the age where your friends around you are getting married, you find yourself requiring cat naps here and there cuz you can\'t function properly without one, you\'re noticing foreign lines around the eyes, your only 12 years away from being over the hill,  27 years from retirement, ....umm..ok..i take that back..you are old..haha!!! just playin\'! well i do hope you have a wonderful birthday..what\'s better than spending it with your favorite friends (sorry \"big boy\"..i think you\'ve been replaced..hehe...jk) ok birthday boy. it\'s your day...haPPPPYYYYYYYY butttdayyyyyyyyyyyyyyyyyy!   '),(21694,477665,'21,April,2003','2003-04-21 00:00:00','    lucky charms....  if i had to sum up in 1 word to describe my very eventful weekend, i would have to say \"lucky\". those who know me, know that i\'ve always considered myself a jinx but i guess that only applies in san diego.  but that\'s not the case when i go to vegas... i\'m a lucky girl!!! there we were, my turn to shoot, all eyes on the dice...i take a deep breath and before i know it, the dice flies from my hand and fate lands it on table. the crowd smiles and cheers as i just made them a little richer. and so the minutes go by and i\'m still shooting, the crowd cheering and everyone winning. at one point, i crapped out but that didn\'t stop me from shooting...the table requested that i take another turn and shoot once again. wow..i became everyone\'s bestfriend.  i think we played craps twice, once at MGM and another time at Luxor..both times we came out as winners. and of course i couldn\'t leave vegas without playing the wheel of fortune slot machine. i love seeing the wheel spin as i scream \"wheel of fortuneeee!\" and so i won a little money there too.   i had a great time everyone. you guys made this trip very memorable. the rio buffet was awesome as usual. pebs..i think we might have scared some of our newly found friends with our talent..haha. hard rock hotel--babies...definitely chill..and it did rock. a lot of dancing, a lot of chillin\', a lot of long beaches, and group jager blasters definitely made the night. can some one please say \"strong\".  i still can\'t believe the bouncers led us to the VIP room...ahhh.. fresh air..exactly what we needed!  hahaha. and so the night ends with us dragging ourselves back to the hotel room to only wake up with a headache.   sat afternoon, we went to eat at emeril\'s..you know  \"bam!\" and to tell you the truth, i don\'t see what all the hype is about but i really did enjoy the company of all of you...and that\'s all that matters! sat night, we went to the palms..club rain.  the line was ridiculously long..luckily a.j. and his girlie got their at 9 pm for us to cut..so it was all good. club rain was off the wall. the lights, the stage, the atmosphere, the patio, the music, the vibe, the club itself..dope!  but the drinks...EXPENSIVE. oh well.. so i think this was a very unforgettable night but to some of us...*ahem* tazzy..i\'m sure we would like it erased from our memory..hehe. but next weekend that\'s a different story and i know you\'ll make up for it.   anywhoo..it\'s monday morning..i\'m still a little tired, a little groggy and my eyes sting a bit but i\'m happy... i named the title of this blog \"lucky charms\" not because i won at gambling but because i definitely think i found myself one... you are my lucky charm... and this weekend it finally hit me how lucky i really am. i am very fortunate to have you in my life, to have been able to share this weekend with you..all of you. so...my lucky charm... if you don\'t mind, i would like to keep you..for it is you that has opened up my heart to believe again. 2003..the year of the goat..yep, it\'s definitely my year. :-) luv you guys!!!!   '),(21695,477665,'17,April,2003','2003-04-17 00:00:00','    and she\'s off...  what is it about vegas? i mean..really?  the food, the clubs, the hot guys, the hot girls, the lights, the slots, craps, blackjack, free drinks, strip bars, shows, shoppin\'..i dunno..must be the whole sha-bang. i find myself there a few times a year.. and  guess what kids????..here we go again...vegas baby, vegas!!  i\'m not into gamblining cuz i\'ve realized after trial and error that i suck at it, but that\'s if i play with my own money... i think my friends/family felt sorry for me at times cuz i was such a fat loser and say..\"here\'s $20..$40 bucks...try your luck again.\" low and behold, i\'d win..double, triple, and even quadruple my money...no joke. i\'d be on a winning streak..winning for myself and winning for the whole table.  i guess you could say i have a track record of winning with other people\'s money. so i\'ve decided to put my own money where it belongs and that\'s buying food to feed my tummy...(mmm\' rio seafood buffet) anywhoo, i decided to let the whole world know my secret on how to win at gambling and that\'s to let me play with your money..if you don\'t believe me..give me $20 or however much you want and i\'ll play but here\'s the deal..if i lose..opps..sorry...but if i win..i\'ll keep the winnings, give u a big hug and say thanks...hehe.  i\'m so excited..i can\'t even sit still.  i love hangin\' out with all of you..but once again we are missing one c-toon link. sux but trust me..we\'ll definitely be thinkin\' about you sweetie.  ok..the count down is in..20 more hours to go and i\'m outta here...byeeeeeeee!!!!!   '),(21696,477665,'16,April,2003','2003-04-16 00:00:00','    spring fever...  i think i\'m gonna try to get back into this bloggin\' trend. i started one about a year ago and never really kept up with it and then i started this one \"if you only knew...\" and i think i did pretty well but things got a little too busy...you know doing important stuff...hehe...well lets just say i was free minded and spirited like there was no tomorrow. it was all about having fun and doing what i\'ve always wanted to do.  for the first time in my life, i can honestly say i finally know where i\'m going. my head is on straight and my direction is clear. my goals and dreams can never be shattered..not now..not ever.   let\'s see..the last time i blogged was in october...it\'s been 5 months..damn i must have been livin\' it up..haha.  so much has happened. it\'s weird to think that in a year\'s time frame, my thoughts have transformed like a catepillar in a cacoon to a free flying butterfly.  words don\'t phase me, what bothered me before, i could care less about now, the emotional roller coaster that once plagued my life has been lifted from the smiles and kind words of others. i dunno..i just think my life has taken the turn for the better.  c-toons...i wanna thank you for always being there for me..through thick and thin...good times and bad...tho i can\'t really recall any bad times we had with each other. i guess i\'m refering to more of the ugly drama that unravels when the opposite sex gets involved..haha no offense boys.  anywhoo...pebs, tazzy, mimz, and livvy...from the bottom of my heart..i want you all to know that the times spent with you guys are unforgetable.. i can\'t remember ever having a dull moment....even when we\'re just chillin\' like villians. just imagine..it\'s only april and we\'ve already made a thousand memories..who knows what\'s in store for us..you never know..we all might end up back at miyagi\'s dancin\' on the bar...haha..next time...pebs and mimz..it\'ll definitely be your turn to shine.  sorry readers for all the sappy stuff but aunt flo must be visitin\' soon.. well girls and boys..that\'s it for now...to be continued..(i have 5 months to catch up on..haha)   '),(21697,477665,'29,May,2003','2003-05-29 00:00:00','    gemini twins..happy 21st!  i can\'t believe you girls are 21. all i have to say is you better enjoy it while it lasts cuz it only goes down hill from here..hahaha!!!!! ok..please don\'t follow your sister\'s foot steps....drinking is bad for your liver, dancing can increase your calf size, and staying up late from a night of partying can cause wrinkles. too bad you girls didn\'t turn 21 earlier because i\'m pooped...no more clubbin for me..ok maybe after this weekend. well here\'s a little note: drink, drink but don\'t forget to think and here\'s another one: my friends are brutal when it comes to making your birthdays an unforgettable one. drinks all around! happy 21st jujubean and panda bear. love you guys!   '),(21698,477665,'23,May,2003','2003-05-23 00:00:00','    happy friday!!!!  yeah...it\'s finally here. i\'m so happy. i love fridays especially when i have a long weekend ahead of me. just wanted to wish everyone a happy memorial day weekend. eat lots, drink lots, get a tan and EAT LOTS. haha!!   ok bestfriend...it\'s also your birthday this weekend!!! you gotta use ricky\'s infamous line...\"but it\'s MY birthhhhday\"..use it to your full advantage. it\'s funny how people can say those magik words and the person they say it to becomes putty in their hands. it\'s the only time of year you can whine and get whatever you want without being called a whiner. geez..only 2 more months for me..watch out world...\"BUT IT\'S MY BIRTHDAY!\" haha..just practicing. k pebbles...i know you\'ll have a great birthday...what could be better than spending it with your friends, family and of course your sweetheart...can\'t wait for the bbq...mmm\' delicious!!!!  k  birthday girl..you\'re only 15 years from being over the hill..so much to look forward to...imagine all you\'ll accomplish in the meantime..hehe. happy birthday babe! muah!!!!   '),(21699,477665,'21,May,2003','2003-05-21 00:00:00','    history repeats itself...  i experienced a little piece of history last night..it was a horrible remembrance. i\'m still in awe. it was happening again...all over again. i felt trapped..no where to run, no where to hide, and no way out. it felt like a slap in the face, a punch in the gut, a knife in the heart. it hit me hard...knocked me back to reality. what i realized after this experience was that i didn\'t learn from my past mistakes...i forgot about it..i forgot how it made me feel and i forgot the hurt and sadness my eyes once felt. i guess it\'s true that those who do not remember their mistakes are bound to repeat them. if i only remembered, then maybe...just maybe this reoccuring event would subside...   '),(21700,477665,'20,May,2003','2003-05-20 00:00:00','    the one...  after a couple of attempts to watch the matrix..we successfully accomplished our goal last night. we started it on sunday but i fell asleep 20 minutes into the movie...hehe (that\'s why i don\'t like watching movies at night :-)... so we picked up where we left off last night.  i watched it when it came out in the theaters a few years ago but i felt like i was watching it for the first time. i didn\'t remember a lot of it but that\'s no surprise. i just remember liking the movie. anyways, the question came up after the movie about how trinity could say neo was the \"one\" when they never even laughed together. it never really crossed my mind..i just thought she was one lucky duck to be making out with keenu.  just kiddin but to answer the question,  i said \"the oracle told her that she would love only one man and it would be the one so that was why.\"   well that was my initial response but after putting more thought into it,  i guess it would be hard to love someone you barely even know but then again, you can\'t choose when love comes. it\'s unpredictable. yeah they didn\'t laugh together (which is a key ingredient to love), but they did fight the bad guys together and that brought them closer.  i dunno tho..i guess she felt \"that\" feeling that everyone feels when they are in love.   i\'ve been contemplating about seeing a pyschic. i\'ve always wanted to but just never had the chance. i think it would be really interesting to see what the fortune teller would say about my life and my future. hmm..maybe she\'ll tell me that the \"one\" will make me a princess..haha. ok well i\'m all pumped and ready for matrix reloaded. i\'ve heard good and bad reviews but if the special effects are anything like the first one, i think i\'ll love it.   '),(21701,477665,'16,May,2003','2003-05-16 00:00:00','    memories...  though i don\'t always write about the good times, doesn\'t mean i don\'t remember them! the memories of the warmth, the light, the smiles, the laughter, and the joy are all ingrained in my mind. and even if i chose to block it out, scared that the bad memories would re-surface, it\'d be impossible and undesirable. those events that have taken place during the course of my life has shaped me into who i am today.  i wouldn\'t change it nor would i ever want to.   words that were left unsaid, were left unsaid for a purpose. and i agree, there is no point to hashing up something that has been buried for sometime.  in a way, i guess some of the memories i write about truly embody the phrase forgive but never forget...there\'s no deeper meaning..it\'s as simple as that.   '),(21702,477665,'15,May,2003','2003-05-15 00:00:00','    must be an asian thang...  during the early college years, the place to eat late at night was fridays (TGIF) for happy hour which started at 10 pm. we\'d gather up our friends and spend the late night chowing down on half price appetizers. that\'s where the asians were at. after a year or so of that, someone introduced us to callahans....25 cent wings. what a deal!! you could order all the wings you wanted as long as you ordered in groups of 5 for only a measely 25 cents or maybe it was 30 cents..i can\'t remember..it\'s been so long. anyways, the wings came in 3 different flavors... you could get mild for the weak, hot for the bold and fire for the brave.  so there we were, every wednesday night chomping on wings. hey it was cheap and we were starving college students.  i guess the word caught on because soon all the asians were eating there as well. you know...it was kinda like e-street..you started see the same people every time...  so anywhoooo, the reason why i\'m blogging about this is because last night we had another revelation. i think another trend has started.  no more half price appetizers and no more cheapo wings. we\'ve moved up in the food chain ladder. it\'s all about the lobster now.  $9.95 for 1lb lobster at the boathouse (located in shelter island) every wednesday night. my friends and I have gone so many times that last night someone (i won\'t say his name..*ahem* ricky) got a little to excited and ate way too many lobster cabezas and got sick. :-(  awww...poor baby.  but that\'s not the point. the point is we saw so many asians there last night. u should have seen it...asians gallore..i think we\'re taking over the world. all you have to do is mention the word \"free\", \"half off\", \"buy one get one free\", \"sale\", or \"$9.95 lobsters\" and they\'re there. it was too funny. the word is definitely catching on. so we\'re now eating at upscale restaurants. we\'re moving up. *sigh*.. maybe onami or todai will follow the trend and throw in some specials on thursday nights. something like \"if you\'re name is tiffani..the meal is on us :-)\"..ahhhh...what a wonderful idea! but i doubt it..only in my dreams right?!?   '),(21703,477665,'14,May,2003','2003-05-14 00:00:00','    distance makes the heart grow fonder...  i was chatting with my girlfriend the other day and we were comparing our different views on missing people. she said that she loves to miss her toy..i mean boy. haha. anyways, she said that the feeling of missing someone portrays how much you really like the person or how much that person really means to you. she loves that feeling of missing cuz when she finally sees him, she\'s on cloud nine. i on the other hand did not like that feeling at one time... i didn\'t care much for the whole missing saga. yeah...i thought it was great when you got to see him again after sometime but the lag time in between the next visit..awww..definitely a killer.  let\'s see... my friend see\'s her man everyday...and i really mean everyday cuz they work in the same building. so the only time she doesn\'t see him is when she\'s home, so the time in between missing him is fairly short cuz she\'ll see him again the next day.  she enjoys missing him and i\'ve come to believe it\'s because the time in between is not long enough where it starts to hurt.   i\'ve been in one long distance relationship and at first it wasn\'t too bad. i saw him very weekend but then slowly the visits were less and less. it became every other week, every three weeks, and eventually it seemed as if it was once in a full moon.  k..i\'m exagerrating but it was hard to be apart and i didn\'t like missing him. but i also think it had to do a lot of with trust issues. i didn\'t trust him but then again..he gave me reasons not too. and the gut wrenching feeling i would have everytime we talked..gave me bad vibes but that\'s a closed topic. anyways, getting back on the topic...i think it\'s true that distance makes the heart grow fonder but only if you have complete trust in other person. if you don\'t have it, it will definitely hurt because you\'ll always be wondering if the person is doing you wrong behind your back or what not.  trust is what makes a relationship hold true. it\'s the peanut in the m&m, the sourcream in the carne asada burrito, the lobster dipped in butter, the boba in the avocado shake and of course, the pumpkin in the pumpkin smash.    this time around, my feeling towards missing someone is a bit different. i don\'t feel the pain, i don\'t feel the hurt, and i don\'t feel the suspicions. what i do feel is a sense of sadness when he\'s away. i don\'t see him everyday like most people who are in relationships but the time that we are apart, he is still in my heart and that makes me happy, so i smile.  i guess missing someone isn\'t  so  bad. :-)  the distance has definitely made my heart grow fonder.   '),(21704,477665,'13,May,2003','2003-05-13 00:00:00','    life after college...  i remember the all-nighters i use to pull, cramming for exams and working on projects. i use to get real sick all the time right before finals. i think the stress really did a few numbers on me. i also remember how much i wanted school to be over so i could finally enjoy my life. i felt my life was at a stand-still until i was done. i always had school on my mind. if i was hanging out with friends, i never gave my undivided attention cuz in the back of my mind, i\'d worry about if i finished that homework assignment or how much work or studying i\'d have to do.  it sucked. at times i wanted to give up, especially during my master\'s program. anyways, it\'s all behind me now and you know what?!? my life is almost as peachy as i thought it would be. i\'m traveling here and there, seeing things i\'ve never seen, doing what i\'ve always wanted to do, studyfree weeknights and weekdays, working on my tan, spending time with loved ones, experiencing a whole new side to life, daydreaming into tomorrow...the only problem now is WORK.   i guess i never really thought about how much of your time is put forth working. you spend most of your time there, 10 hour days, 5 days a week. ewww.. but at least i get every other friday off. that\'s probably one of the main reasons why i\'m still here.  i can\'t even fathom waking up at 6AM everyday to go to work until I retire.  i\'m getting tired just thinking about it. it wasn\'t too long ago where i loved coming to work, maybe because i had the coolest boss. i would come in at 9 AM work a couple hours, go to lunch at 11 for about 2 hours, and then come back to the office only to work another 2 hours before he let me go home. now that\'s what i call a job.  but those days are over and i\'m now working in what they call corporate america. i think i wouldn\'t mind it so much if i was treated like an adult but they still treat me like a kid. the old ladies still hate me, the old men still gawk at me, and i\'m still under-paid.   i\'ve made a huge decision...i\'ve been contemplating this for awhile and i keep coming back to the same outcome. here it goes... i  want to become a princess. yeah me! it would be so much fun. i wouldn\'t have to work and i can play all day long.  i think i\'d be a good one too. yes..my career goal has definitely changed..old career goal: be the CEO of my own company. new goal: become a princess. some my friends already call me that and i would very much like to fill those shoes..they have such high expectations ofe..hehe. i can\'t let them down... i must become one.  princess tiffani.  has a nice ring to it huh?!?. ok i know you all think i\'m crazy or dorky but i\'m only doing what i was taught and that is to dream big.  life as a princess...now that\'s what i call life after college.   '),(21705,477665,'08,May,2003','2003-05-08 00:00:00','    commitment phobic..who me??  at 23, i feel like i\'m still a kid. yeah, i\'m done with school and i have a good job and the next logical step is to find a mate and get married..well that\'s what my grandparents keep telling me. yet i know my parents are so not ready to be in-laws and i\'m not ready to be anyones wife. everytime i see my grandpa, he\'s always telling me that i need to find a husband..geez, i\'m only 23 and finding the \"one\" that i\'m meant to spend the rest of my life with, well isn\'t that easy. lately, the topic on marriage has been floating around. some of my friends are married or are getting married real soon. it looks like everyone is coupling off.  it\'s weird. it happens in bunches..first everyone is single and partying like there\'s no tomorrow and now it seems like everyone has someone.     for awhile, marriage and starting a family was so far fetched. not my favorite topic of discussion. i just couldn\'t understand why anyone would want to settle down so soon. maybe cause i\'m not ready. i was talking to my friend stacy about that. she said that she knew she would marry her man a few years ago and she was just waiting for him to ask her. she\'s only 25 and she knew he was the one 3 years ago.  i\'m amazed at that. i\'ve been in 1 long term relationship and maybe a couple shorter ones worth remembering and i don\'t think i was ever ready. i would question the relationship, i would question the love, i would question him and i would question myself.  my friends would tell me that the reason i had so many doubts wasn\'t because i didn\'t love him, but because i wasn\'t ready and when i am, i will know. well i guess i wasn\'t ever ready cuz i never felt that feeling they were trying to describe to me.   don\'t get me wrong...i don\'t think i\'m marriage phobic, but i do think i still have some growing up to do. it\'s not like i never want to get married even tho after my break up, i was anti boys for the longest time..hehe. but i\'m over that now and i can finally see a glimpse of my future and i\'m looking forward to the journey that is before me.  i\'ve always known what my wedding day would look like all the way down to the itty gritty details.  i know it will be purely magical and will be everything i\'ve always imagined. it will definitely be my dream come true and hopefully his as well. but there\'s no rush because like i\'ve always said, timing is everything and when the time is right, everything will slowly fall into its place.   '),(21706,477665,'07,May,2003','2003-05-07 00:00:00','    timing...  i\'ve come to accept that everything has to do with timing. everytime something would happen or would not happen, i\'d say it had to do with the time and it just wasn\'t our time, and with this mentality, it keeps you sane. i\'ve been through my share of heartaches and all the tears i\'ve shed has only made me stronger. each time my heart gets broken, it never really completely heals, and when it breaks again, the cracks are always different.  people always say that once you get hurt, it doesn\'t hurt as bad the second time around. i agree with that belief but only to a certain extent.  the second time, the third time, whatever the case may be, the pain still stings, the tears still burn, the heart still aches. it doesn\'t hurt any less but the way you cope is different. you know better, you\'ve learned from past experiences and you know life goes on. so dealing with the pain is easier but the hurt still stings to the bone.   i think this past weekend proved to me that everything really does happen for a reason. but sometimes the reason may be different than what you expected and it just takes time to see that.  a few months ago, the timing was right for that purpose, a month ago the timing was not right but yet we knew our time would come so we held on.  after awhile confusion, misunderstandings, innocence began to plague our minds. and with a blink of an eye..it was dark again. dark and alone. i couldn\'t understand what happened or why it happened but i knew that it happened for a reason. i came up with reasons why which helped me deal but the fact was that after some time, the reason for it happening became crystal clear. it was in the cards, our cards. and even though we both felt pain, it had to happen for us to open up our eyes and see what really matters. and in all honesty, i\'d suffer that pain again if that meant that it was the only way we could get to where we are right now.  i guess you could say the timing couldn\'t have been more perfect.   '),(21707,477665,'01,May,2003','2003-05-01 00:00:00','    40 days and 40 nights...  so for lent my roomie and i gave up carbs...not all, but the major ones like bread, rice, pasta, potatoes, and sweets (fudgcicles, cookies, peanut m&m\'s, chocolate, cake, etc.). we got this brilliant idea after watching \"40 days and 40 nights\" (josh hartnett..mmm\' definitely a hottie). anyways, after watching that flick, we decided to give up something that we both love and you all know what i love most..EATING. so we crazy girls decided to suffer 40 days without yummy, delicious, high carb foods to prove that we have will-power.  yeah i know i\'m not catholic..shoot, neither is she...but this gave us a reason to cut out the bad carbs and at the same time, lose a little weight.    week 1 - first week without any high carbs..yup no rice...heaven knows how much we love rice and a day without it..it was indeed a sad day. but we survived. we started eating lots of green veggies and chicken.   week 2 - we started feeling a little weak. we were slowly runnin\' out of energy but that didn\'t stop us. we kept to our word and also started exercising regularly. running about 3 times a week around miramar lake. that kept us going...  week 3-  hmm..not so bad. don\'t miss rice, don\'t miss pasta, don\'t miss potatoes, but damn what i would do for a loft of bread and my thin mint girl scoot cookies that are hidin\' in the back of my freezer.   week 4 - we decided to eat bread at least once a week because someone told us if we eliminate it from our diet completely and start eating soon after our diet, we\'d blow up like balloons...heck..that scared us so bread it was..but only once a week.  week 5 - we\'re still runnin\' almost every other night. our motivation still prevalent. we felt good about ourselves because we were eating healthy and exercising. only 2 more weeks to go...  week 6 - uh oh..i think i started breakin\' down by week 6..it had to be disneyland..i couldn\'t eat anything else but high carb foods... that\'s all they sold. it\'s notmy fault..so it was ok. funnel cake, churros, ice cream, cheeseburger, fries, chilli dogs, potatoe chips..man i was in heaven. so there we were..eating everything insight. yeah i felt a little guilty but it tasted so good that the guilt went away real fast. :-)  WEEK 7 - last week and where am i?!? in vegas..damn..there goes my will power...mine and tazz. we only had 2 more days to reach our mark and we couldn\'t even do it. i needed pizza..a fat mushroom, pepperoni and olive pizza. mmm\'. it was simply heaven. and so our lent ended a couple days before our target date of april 20th (for those who think that i lost at week 6, remember i\'m not counting disneyland..it wasn\'t my fault..u don\'t expect me to starve do u? :-)   since then, we\'ve been eating like crazy and we stopped exercising. i really do miss running around the lake but these days it seems like i have an excuse for everything and anything not to go. our motivation is completely gone. also during those 40 days, we said we\'d stop eating rice even after lent..but we didn\'t. it\'s back in our tummys. i think i need to find something else to do..i\'m bored. i wanna do something spontaneous..something crazy. the only problem is..i don\'t know what. maybe for now,  i\'ll just chill and watch the sunset and the sunrise...*smile*   '),(21708,477665,'19,June,2003','2003-06-19 00:00:00','    and we\'re off!!!!  our long awaited vacation is finally here. in about 16 hours, we will be on a plane to the happiest place on earth..walt disney world. :-) yeah i know i\'m still a big kid and i think i\'ll always have the kid in me. what can i say, i love dreaming and this place makes dreams come true. i think we\'ll prolly hit up all the major attractions...magik kingdom, epcot, mgm, and pleasure island which is located in downtown disney. you know what that means...clubbing woo-hoo!!!.  who knows, we might even end up taking a cruise to the bahamas. i\'m seriously thinking about it. i have 10 days off..might as well make the most of my trip. yeah i know i plan trips way in advance but i never think about the details of the trip until last minute.  grrrr!!!!  and i get in trouble at times for not thinking but i guess that\'s just the spontaneity in me...at least that\'s what i\'d like to think.    well i hope everyone has a great weekend and hopefully when i come back from florida, i\'ll brag..i mean blog about how tan i am. :-) byee!!!   '),(21709,477665,'17,June,2003','2003-06-17 00:00:00','    my ideal date...  i hate dating. it\'s definitely not my forte. i\'ve always had a boyfriend or i was seeing someone...well since the high school days anyways. but last year i experienced the greatest thing ever, it was the best thing that could have ever happened to me and the timing couldn\'t have been more perfect.  that was being on my own. i was single as can be... a free spirited girl ready to take on the world. i had so many crazy moments...so many unforgettable memories.  i learned so much about myself and i realized how wonderful it was to be alone and to do things on my own and not feel one bit lonely.  i actually enjoyed spending time by myself.  i spent a lot of my days alone at the cliff on the beach. i dunno...there\'s just something so tranquil about watching the waves crashing on the shores and the sun setting across the horizon. it\'s amazing. it\'s beautiful. it\'s astounding. it made me happy.  anyways,  as much as i liked having my time alone,  i still hung out with the opposite sex. it\'s a funny thing. the 10 months i was single, i \"hung\" out with a few guys but never once did i consider them dates. yeah, we might have done the dinner thing but i made it clear to them..it wasn\'t a date. i guess i was a little harsh but i really didn\'t want to date anyone. there were times where i had to straight out tell them that we were just hanging out cuz they called it a date but it sO wasn\'t in my book....no no dates for me. i guess i wanted to make it clear to them cuz i didn\'t want to lead any of them on.  i really just wanted to chill and be single.    well as much as i hated dating, i found myself wanting to when i met ricky.  everyone knows how much i hate watching movies in the theaters. my girlfriends tried to get me to go forever but didn\'t succeed. but once ricky came into my life, i didn\'t even think twice about it. come to think of it, i think i asked him to the movies..haha. anywhoo..i\'m here to prove a point that wanting something sucks cuz you just might get it. i was anti-dating for the longest time and finally when i felt i was ready, i think my prospect was scared to go on a date with me cuz we never hung out alone. we always hung out in big groups. yeah that was fun and everything but i really wanted the whole wine and dine me thing.  i guess that\'s what i get for telling guys i don\'t date. my bad. :P    i guess the main reason why i\'m blogging about this is to share with everyone what my ideal date is. yeah i enjoy the whole wine and dine sha-bang so don\'t rule that out! haha. but what i imagine is him and me on top of a huge green hill...a waterfall behind us. the grass is a perfect shade of green, the sky is a sharp color blue with fluffy white clouds surrounding the gleaming warm sun. so there i am with him, butterflies in my tummy but all the same time i\'m comfortable. he makes me laugh. we smile and we laugh again...we laugh until our stomachs hurt. it hurts so much that we both decide to roll down the soft green hill for fun. there we go tumbling down and finally when we get to the bottom there\'s a pumpkin smash jamba drink awaiting for us. we sip out of the same straw and splash our feet in the lake below us.  he then decides to take me to my favorite spot. a place i didn\'t even know he knew about. we sit on top of the cliff with our feets hanging over and watch the beautiful sunset. he looks into my eyes and smiles. i melt. and then to end the date, he gives me a warm bear hug and kisses me gently on the forehead.  *sigh*...now that\'s what a call a date. :-)   '),(21710,477665,'12,June,2003','2003-06-12 00:00:00','    just something to think about....  \" Life is not measured by the number of breaths we take,  but by the moments that take our breath away.\"   '),(21711,477665,'09,June,2003','2003-06-09 00:00:00','    taking risks part II  well if you guys are wondering what my dilemma is all about, it\'s the decision to move to seattle for a great job. those who know me have heard me say how much i love traveling. i love meeting new people, i love doing different things, and i love new challenges. this position would require me to relocate to seatlle, washington and would require me to be on 3-4 week assignments at a time in all different states. now this is what i\'ve always wanted to do..travel for work. i\'d be a consultant/trainer for companies/utilities that want to implement an enterprise wide software. it\'s a great job opportunity and what i\'ve learned thru the years is to never let an opportunity to pass by, but in this case..i just might.  my career is very important to me. i\'ve felt lost for awhile. i wasn\'t sure what my next steps were in life except to get married and start a family but i\'m not financially ready for that. and emotional???...as much as i think i am..i\'m not. but i know it\'s my career that i need to be working on. it\'s not something i can put a time frame on like school. with school, you can decide how long you want to take and you know when you will reach your goal . you have control over that. with your career, you don\'t know when you\'ll reach your peak. you have less control. you can work your butt off but you don\'t decide when you get a raise or promoted. it\'s not in your hands.   i just think the timing is all off. i just bought a house. i\'m painting. i just put in shelves in my garage. i\'m getting wood flooring this summer.  i\'m having so much fun planning what i need to do next to the house. it\'s my baby. but if i leave, i\'ll have to rent it out because i don\'t want to sell it.  but my house is not the only thing holding me back.  i\'ve recently completed my graduate studies and i\'m ready to just relax and spend quality time with my friends and family. i want to go on more vacations. i want to sit back and read a good book. i want take naps on a hammock.   maybe i\'m too much of a daddy\'s girl. i had a pretty deep conversation with my dad this past weekend about the job offer and he didn\'t seem too thrilled. even tho i\'d get paid sO much more, he doesn\'t care...he wants me here and the more i think about it..i want to be here. my life is in san diego and my future is in san diego. i\'ll build my career right here in sd!! but i know that if i decide to go, they will always be here for me... both family and friends but i don\'t think i can stand being so far away. i\'d miss everyone. *sniff*  but we\'ll see what happens.  this was my dream job but sometimes you have to let go of one dream to fulfill another. and right now, i\'m definitely living one of my dreams.   '),(21712,477665,'06,June,2003','2003-06-06 00:00:00','    taking risks...  what do you do if opportunity knocks? do you take it or leave it? if you take it, your whole life will change. if you don\'t, you\'ll never know what could have been.   something big is going to happen and if i choose to go down that route, my life will never be the same.  it could be so risky that i might lose everything. am i willing to drop it all and take off? am i willing to be completely on my own? it\'s a big step for me but i think i\'ve finally found what i\'ve always been dreaming of. but then again dreams change....there are past dreams, present dreams and future dreams right? whoever said you couldn\'t change them.   i think i just need some time to sort out my thoughts, lay out all the factors on the table, list the pros and cons. it\'s a big step and who knows if this step will take me ahead or bring me three steps back. i dunno. i guess time will only tell...   '),(21713,477665,'04,June,2003','2003-06-04 00:00:00','    middle child syndrome  here\'s a little background on my childhood...5 girls and 1 guy. no brothers, 3 sisters, one older, the twins younger and a mom and dad. yep that\'s my family. we are one big happy family... the majority of the time but like all familes, we had problems...hmm..and what could that be???..yep..me..i had the middle child syndrome.   looking back, i really think i made more mistakes than the average teenage. you wouldn\'t believe half the things i did and got away with and the other half that i got busted for but i\'m taking those secrets to my grave. hey..kids will be kids. anywhoo, i always got yelled at, always got in trouble, and i always rebelled. i was such a brat. i can admit that now. but you know what the funny thing is, everyone at school thought i was the sweetest person, the most innocent of them all..i even got selected as sweetest personality for my highschool yearbook. haha..what a joke. k..don\'t get me wrong..i\'m not a meany but i don\'t really think anyone really knew the real me except for my family. they saw me at my worst, they saw the spark in me die, they saw the blank stares in my eyes, they saw the turmoil in the pit of my stomach, and they saw me grasping for air. and no matter how much i hurt them with my actions, they were always there for me and i know they always will be and vice versa.    this is hard for me to admit but i\'m gonna suck it up and let it out..i think the reason i made so many mistakes not just because i was ignorant but because i was jealous...jealous of who you ask..well my sisters. here\'s the scoop..my older sister..smart as hell..valadictorian of her highschool, never got in trouble, always obeyed the rules, went to UCSD, never a disappointment. my younger sisters..well..they are twins..that should explain it all but i\'ll elaborate anyways. being twins..you have instant attention, plus they were the babies in the family and they\'re super cute. and that leaves me..the middle child..not the smartest, not the cutest, and definitely not the one to follow rules. i guess the only way i thought i could shine was to make my parents life miserable..at least i got their attention right? heehee..jk..yeah i was an attention seeking brat. but i don\'t blame my parents...they loved us all the same...even though now i call myself their favorite..haha. so that\'s my story of my middle child syndrome. maybe now you all will understand me a little better and give me all the attention i need. you guys can even spoil me..i don\'t mind..not at all. haha :-) have a wonderful day everyone! *wink*   '),(21714,477665,'02,June,2003','2003-06-02 00:00:00','    \"If You\'re Not The One\"   by Daniel Bedingfield   \"If you’re not the one then why does my soul feel glad today? If you’re not the one then why does my hand fit yours this way? If you are not mine then why does your heart return my call If you are not mine would I have the strength to stand at all  I never know what the future brings  But I know you are here with me now We’ll make it through  And I hope you are the one I share my life with  I don’t want to run away but I can’t take it, I don’t understand If I’m not made for you then why does my heart tell me that I am? Is there any way that I can stay in your arms?  If I don’t need you then why am I crying on my bed? If I don’t need you then why does your name resound in my head? If you’re not for me then why does this distance maim my life? If you’re not for me then why do I dream of you as my wife?  I don’t know why you’re so far away  But I know that this much is true We’ll make it through  And I hope you are the one I share my life with And I wish that you could be the one I die with And I pray in you’re the one I build my home with I hope I love you all my life  I don’t want to run away but I can’t take it, I don’t understand If I’m not made for you then why does my heart tell me that I am Is there any way that I can stay in your arms?  ‘Cause I miss you, body and soul so strong that it takes my breath away And I breathe you into my heart and pray for the strength to stand today ‘Cause I love you, whether it’s wrong or right And though I can’t be with you tonight And know my heart is by your side  I don’t want to run away but I can’t take it, I don’t understand If I’m not made for you then why does my heart tell me that I am Is there any way that I could stay in your arms\"  this is exactly the way i feel about you. never in a million years did i think i\'d find myself here. but here i am and here you are. i guess it\'s true that if you wish for something long and hard enough, it will come true. all those nights praying, all those pennies at the bottom of the well, all my stars that have been wished upon, and all my daydreams  have finally paid off. i\'ve fallen hard, harder than i could have ever imagine and i\'m not scared. for the first time, i feel free...free to let my heart live, free to let it run wild, and free to let my heart love. all the endless possibilities with no restrictions. for the first time... i\'m not scared..i\'m not scared.   '),(21715,477665,'29,July,2003','2003-07-29 00:00:00','    dreams...  i rarely dream..yeah i know experts say that everyone dreams but i seriously never remember any of mine. but the times that i do have dreams and i can recall them, they\'re normally nightmares. i wake up sweating, i wake up scared, i wake up in shock.  i can clearly remember dreams of being attacked by bears or having a homeless man naw at my leg. i once dreamt that rats where mulitplying in dozens and i was trapped in my old room at my parents house.  i also have dreams of people being mean to me and in real life they\'ve been nothing but sweet.  i\'ve dreamt that i was being chased by guys with guns. i\'ve dreamt of family members being really ill. i\'ve dreamt of death in the family. i\'ve dreamt of dead people. i can go on but i\'m kinda freakin\' myself out right now so i think i\'ll stop documenting my nightmares. anyways, i borrowed a dream book from a friend and i\'ve glanced through it a few times but the way the author interprets my dreams, i don\'t fully agree with. i think a lot of it is mumble jumble..in my case anyways.  maybe i just need to get another dream book...nah..i just want to stop having nightmares.  i rather just stick to daydreaming...that\'s what i do best.   '),(21716,477665,'28,July,2003','2003-07-28 00:00:00','    another year older...  i\'ve just celebrated my birthday...yup, i finally turned 21, three years in a row now..haha :-)  anyways,  i just wanted to thank everyone for the birthday cards, letters, emails, texts, phone calls, gifts, and of course the party.  special thanks to ricky for organizing the party. if it wasn\'t for you, i prolly wouldn\'t have celebrated my birthday this year. i\'m just sorry that we got kicked out of the pool area (screw those neighborhood watch spies)but i think everyone still had a good time at the house.  i think i\'m gonna write a letter to the home owner\'s association about those nosy neighbors.  i swear they have nothing better to do then to crash parties.   anywhooo...i had such a great weekend. again..i would say one of the best weekends. we made our way up the coast, wined and dined at my favorite restaurant..crusteaceans. mmm\'!!!  strolled around beverly hills and of course made memories in the city where stars shine.  we also went to a wedding on saturday...a beautiful one in deed. stacy was just glowing. we danced the night away to music that would just take your breath away...(inside joke..*wink*)and then ended our night at the yardhouse downtown to catch up with old friends.  sunday was a blast, a gathering with some of the best people in town...party till the sun disappeared into the moonlight. i don\'t think it can get any better than this but i\'m more than willing to accept it so keep the fun and happy times coming this way. thank you guys for making this year\'s birthday an unforgettable one. i couldn\'t ask for better friends.  so here i am, another year older and enjoying every moment of it....   '),(21717,477665,'24,July,2003','2003-07-24 00:00:00','    i got that loving feeling...  u know that feeling that you only hear about and pray that one day you\'ll feel it to??    i always hear girls or guys say \"oh u just know..u get this feeling.\" man, in every relationship i\'ve been in, i\'ve always waited to see if i\'d get \"that\" feeling and i never did. i think in away i was envious of people who did feel it. i wanted it. i wanted it to happen to me. i wanted to feel that magical feeling where you know he\'s the one. but this time around..i didn\'t expect anything. i took it as it came and only hoped for the best. i find that this approach to be a better way of living. if it happens, it happens but don\'t go looking for it.   i don\'t know what\'s come over me but i feel the sparks, i feel completeness, i feel true love.  it\'s changing my world around in ways i\'ve only dreamed of. could it be? is it happening to me? when you least expect it? when you\'re not looking for it? i dunno but i\'m not gonna question the feelings i feel. i\'m finally happy. happier then i\'ve ever been and that\'s all i ever wanted...   '),(21718,477665,'18,July,2003','2003-07-18 00:00:00','    t.g.i.f.  this week has been the longest week ever. i was in training every fridgin\' day and when i got home from work..i slept till the next day. and worse of all.. i didn\'t even get to see my huney. i\'m just so sO happy it\'s finally FRIDAY!!!! yeah..happy friday everyone. i\'ve waited all week to hangout with you girls and even tho it was suppose to be a girlz night out...i\'m sure it\'ll still be a blast. it\'s been a long time since all of us c-toons have partied together...it\'s long over due!  tazzy, pebs, livvy and mimz...you guys are the craziest girlz i\'ve ever met and also the hottest...hehe. yep, let\'s work it..haha. jk.   but seriously... i know it\'s been hard to get us girls all back together again...everyone doing there own thing lately but we gotta make time for the c-toons.  yeah! c-toons forever..shoot..i sound like a little kid...yup that\'s me..i big kid. :-)  alrighty..let\'s go meet up with mr. lemon drop, mr. surfer on acid, mr. power blasters, and our newly favorite mr. long beach! bottoms up everyone!!!!!!   '),(21719,477665,'16,July,2003','2003-07-16 00:00:00','    everytime by janet jackson  \"I\'m afraid I\'m starting to feel What I said I would not do The last time really hurt me  I\'m scared to fall in love Afraid to love so fast Cuz everytime I fall in love It seems to never last  But every time your love is near And every time I\'m filled with fear Cuz every time I see your face My heart does begin to race everytime  One half wants me to go One half wants me to stay I just get so all confused  I\'m scared to fall in love Afraid to love so fast Cuz every time I fall in love It seems to never last  It\'s everytime It\'s everytime Cuz. . .  Everytime your love is near And every time I\'m filled with fear Cuz every time I see your face Could it be that this will be the one that lasts? The fear does start to erase every time Oh could it be that this will be the one that lasts For all my times Ooh yeah For all my times\"  this song reminds me of how i felt when we first started dating and i can still remember how scared i was when i realized how hard i was falling. i was practically in tears. but my fears subsided as our love grew stronger.  i\'m no where near being a pro when it comes to relationships. i still say the wrong things, i still act on my emotions, i still let the little stuff get to me. but i will tell u this. i\'m trying my best. i\'m trying the best i know how and that\'s all i can do. i don\'t expect much anymore but i do have a lot of hope. and i do hope that \"this will be the one that lasts.\"   listen to your heart, feel my heart...look into my eyes and feel the vibes.  i want you to know how i feel and want you to feel what i feel and then you\'ll know that what we have is real...so very real.   '),(21720,477665,'08,July,2003','2003-07-08 00:00:00','    a sleepless nite...  this is ridiculous. it\'s past 1 am and i can\'t sleep. so here i am..blogging again. i think this is a record for me, 3 blogs back to back. well i don\'t have much to say except for that i\'m very tired. for those out there who are still awake..you know who you are...i\'m sorry. for those who are tucked fast asleep in their beds, dreaming of absolutely nothing...that should be me too. :-(   i guess i\'ll start counting sheeps or something.  nite nite everyone!   '),(21721,477665,'07,July,2003','2003-07-07 00:00:00','    fun times  i had the longest weekend this past 4th of july and might i add, it was indeed one of the best weekends i\'ve had in a long time.  it all started thursday afternoon. i wasn\'t feeling my best so i called in sick that day. ricky called me early in the morning to see if i was doing better cuz the night before, i felt awful. i was a bit snappy due to the exhaustion from the pain i was feeling but he understood. anyways, going back to thursday afternoon, ricky informed me that he was taking me out to a nice dinner. it\'s been awhile since we\'ve gone out and i was just so happy we were gonna spend some quality time alone...didn\'t matter where we were going, just as long as we went together.  he got to my house around 6:30. he rang the doorbell. i remember, my heart was beating a bit faster than usual and i felt the little butterflies flutter inside my tummy. as i opened the door, i was pleasantly surprised by a dozen long stem red roses with baby breaths. awww...one of my many favorites.  my eyes grew larger and all i could do was give him the biggest hug. i held on to him and didn\'t want to let go.  he then took me to this restaurant i\'ve been wanting to take him to (him unaware of this)...ruth\'s chris steak house down by the harbor. i\'ve never had such great steak and lobster before. we drank wine...ok..he drank most of the wine and we had the best view of the ocean. the waitor was excellent. everything was perfect. we even got our picture taken and a complimentary dessert from the waitor.  we went home that night, and all i could think about was how lucky i am i have him in my life. my roomie asked me why he did all that and if it was for a special occasion.  i smiled and nodded no.  it was just an ordinary day and that\'s the best thing about the whole thing.  he never ceases to amaze me. he always keeps me on my feet. he always makes me feel special. i know he\'s probably reading this right now thinking i\'m way to sappy but i\'ve never been so happy...for realz.   friday got even better. ricky threw a picnic for all of our families and friends  at poway lake. i finally got to see some special people that he had been telling me about. and he described them exactly as they were. we ate a grip, had cantelope juice for the first time, we got some sun, tazzy fell asleep, the boys played football, i was the cheerleader, we also played volleyball and yes our team won, 3 out of 5 games.  and of course i got to see fireworks with my sunshine. i loved every moment of it. the smiles on everyone\'s face, the jokes, the horsing around, the keep away from ricky game...you have to admit..that was pretty funny, and of course being introduced to everyone as \"ricky\'s girlfriend\".  geez, i think i heard that a million times..\" that\'s ricky\'s girlfriend everyone!\"  just wait until you go to one of my family parties..the last one didn\'t count!  the tables will definitely be turned.  anyways,  i had a blast. this is definitely one 4th of july i\'ll treasure forever.   it didn\'t just end there, spent saturday afternoon with a group of friends and watched charlies\' angels. it was very entertaining. kept me laughing for awhile. we ended the night at the cheesecake factory and i had the most delicious cheesecake..banna cream...mmm\' delish!    alright, sunday finally rolls around and i go ocean kayaking with another group of friends. it  was my first time and i have to say, it was super duper fun. ricky did most of the rowing, so it was pretty easy.  we went into the caves and saw orange fishes. arnold decided to jump out of his kayake and go snorkeling...let\'s just say the water was filled with kelps and it looked kinda dirty but hey..whatever floats your boat. :-)  we had kayake races, and some of us had a little too much fun getting other people soaking wet. just wait till we go again. i\'ll be prepared!  after that, we went home and took a 6 hour nap, ricky made me dinner and we watched \"the little princess\". a perfect way to end a perfect weekend.   so there you go...that\'s my very eventful weekend with some of the people i love most. i wonder what next year\'s 4th of july will be. i think it will have a hard time beating this years. thanks ricky for the fun times. you definitely know how to bring a smile to this girls\' face. :-)   '),(21722,477665,'07,July,2003','2003-07-07 00:00:00','    -the cycle of friends-  it’s weird to think how friends come and go. there was a time in my life when it was filled with their laughters, filled with their smiles, filled with their spirits and filled with their love. some of my fondest memories…just to list a few…  the drives up the coast, the spontaneous trip to magic mountain at 3 pm  in the afternoon, the karaoke night when we sang duets, the nights at bar dynamite, the conversations that lasted for hours, the 10 mile runs around calle cristo ball, the nights partying until the toilets became our best friends, the underage gambling in vegas, the 2am house painting brawls, and of course the most important…the friendship itself.     this list is not just memories shared with  one person but a collaboration of memories shared with friends that have come and go in my life. i’ve come to accept that friends are hard to find and good friends are even harder.  i think i’ve been very fortunate enough to have shared my life with so many people. but as another year goes by, another friendship will lose momentum.  for awhile,  i was questioning why this was happening, was it something I said or  did, or was it on their end. where did the fault lay, would things ever be the same,  would they ever  be a part of my life again. if they didn’t move away, would things stay the same. if their work wasn’t taking them in another direction, would there still be time for me,  if they weren’t tied down, would we still be friends. the questions are asked yet the answers are already known. so why do I grasp at straws trying to find these answers?  sitting here, my mind wonders, my thoughts typed out, my heart being heard.   i’ve come to believe that the reason why friends come and go is because they’ve already accomplished their purpose. whatever the reason may be for their presence, i’m grateful they came and touched my heart and left a mark because without each imprint, i wouldn’t be where and who I am today.  find yourself very lucky to stumble upon some friendships that last a lifetime but honestly, friendships like those are rare to find. As for the other friendships that come and go with the change in the season, i don’t find them any less important. for the short time they were in your life, the effect may have been greater than someone you’ve been friends with your entire life.  i can’t keep questioning. i just have to accept that and I have. i know they came in my life for a purpose and i’m not bitter that they’re gone. maybe a little sad but the memories I hold will always be ingrained in the back of my mind.  so with this said, my mind is more at ease. my thoughts are clear, my heart is free. and just to let you know, if you ever need a friend again, you’ll definitely find one in me.   '),(21723,477665,'26,August,2003','2003-08-26 00:00:00','    my sassy girl...   last night just took my breath away. god answered my prayers. there\'s so many firsts that i want to do with you, so many places i want to see with you and through your eyes. the possibilities are endless. day dreams fill my mind with so much hope. hopeless romantic that i am.  i keep telling myself that i need to take one day at a time but i can\'t help but to look into the future and what i see is you.   the weather was a perfect breeze, the location lit by love and innocence, in the midst of prospect. as your words left your lips, my heart skipped a few beats. my mind wonders off for a split second. a shock overcomes me. it\'s me that he feels for...and only me. i can\'t help but smile. i can\'t help but feel so very lucky.   everyday, i feel like i\'m falling...deeper and deeper. there\'s no way out, no where to run, no place to hide.  and strangely enough, i wouldn\'t have it any other way.   last night left a very special memory in my heart. a smile, a glow, a forever image of true love in the making. it\'s not just me, my familiy sees it  and has said that they\'ve never seen me happier. i only hope that i make you as happy as you have made me.  \"my sassy girl\" is a great movie. it sparks ideas and if taken with imagination, it may be your destiny.  i\'ve been pointed in the right direction and fate has got me here but now it\'s up to me to control my destiny...   '),(21724,477665,'21,August,2003','2003-08-21 00:00:00','    a word from the wiser....  a good friend once told me \"There are three important dates in your life. Today, tomorrow, and yesterday.  Your future belongs with \"Tomorrow\"  Learn from the past, remember what you did \"yesterday\" to improve how you live your life \"today.\"   thanks for your encouraging words. i think your optimistic views are beginning to shine through me. let\'s keep the hope....   '),(21725,477665,'15,August,2003','2003-08-15 00:00:00','    it wasn\'t the flu...  geez..i\'ve been home all week...didn\'t go to work at all and you\'re prolly thinking i took another vacation or something...nope..i wish tho..i\'ve been sick..sick as a dog. i thought i had the flu but i just got back from the doctor\'s today and she said that i have strep throat. that kinda thru me off. i had a fever all week and my whole body was aching but my throat wasn\'t bothering me at all so when the doctor said it was strep, i was shocked. well...she\'s the expert so i guess i have strep. hopefully these antibiotics will make me feel better. i\'m such a baby when i\'m sick..wahhhhhhhhhhhh!!!!   i just wanted to thank my sweetie for taking care of me. i don\'t know what i would have done without you. staying up with me, getting me medicine, feeding me, loving me when i look a mess, trying to make be freeze to death when i had the chills, the cold showers, the pampering...i\'m so lucky to have you. thanks baby!   '),(21726,477665,'10,August,2003','2003-08-10 00:00:00','    if only you knew...  a blow to the heart, beating ever so slow the sound of her name kills but you don\'t know.  a painful remembrance of nightmares that came, dreading the sleepless nights but who\'s to blame.  unbearable at times, but i have to stay strong, my mind\'s not at ease, it\'s been going on for so long.  any negative remarks and a defensive mode appears, my stomach turns which is followed by hidden tears.  a love so rare that even the sun shines at night,  yet comparisions expressed, doesn\'t feel right.  there\'s only one diamond that lives in this heart so true, as for the gem in your heart, somehow makes me feel blue.  with a drop of a coin, my thoughts meander and fold, i can only hope that this feeling will soon be left in the cold  the touch of your lip, a sweet kiss of honeydew, i\'m so in love with you, if you only knew....   '),(21727,477665,'26,September,2003','2003-09-26 00:00:00','    change of colors...  haha...i changed my mind...i actually went with \"desert floor\"!!!!    like you guys care about the color i chose for my living room but  I DO  and i\'m sO happy with it. man..who would have ever thought someone could get so excited about a paint color...i think i\'m gettin\' old...oh well.   '),(21728,477665,'22,September,2003','2003-09-22 00:00:00','    playing host...  here we go again...another weekend filled with fun, fun, fun! another bay area friend is coming to town. oh how we love playing host huh ricky??  anyways, it\'ll be great to see her again for she is the main reason we\'lll be hitting up the clubs and of course getting all our friends together to parrtttay! bottoms up everyone!   '),(21729,477665,'22,September,2003','2003-09-22 00:00:00','    colors beyond the rainbow...  i love home depot...it\'s one of my favorite stores (never in a million years would i have ever fathomed saying that).  but yeah, i really think it\'s the best  thing to have ever happened to homeowners. ever since i bought my townhome, i\'ve spent more days at home depot than my usual spot..the mall. you won\'t ever see me there anymore but guaranteed, you\'ll find me at home depot late at night.  i\'ve been neglecting my store for a few months now only because i was preoccupied with other stuff...oh how i love detours..haha. but now, i\'m back in action...ready to paint again, ready to re-model, ready to brighten up my place and give it some TLC.  you all have no idea how motivated i am this very moment and there is a reason for my madness. everyone around me is purchasing property...they\'ll are gonna be homeowners if not one yet. on sat, i went to see a friend\'s sisters townhome in carmel valley.   it looked just like a model home. they used earthtone colors for each room. and a couple weeks ago, i went to drew and michelle\'s condo located downtown...their home was also painted in amazing colors and fully decorated. it just got me thinking that i\'ve lived in my place for almost a year now and all i\'ve done was paint 3 bathrooms and these people have painted all their rooms and have finished decorating!! man, i\'m so behind.  so last night ricky and i were up till 2 am, masking my living room but prior to that, we stopped by home depot for tape and books on paint colors, and then we stopped by barnes and nobles for decorating tips and paint color ideas and OMG...i can\'t believe all the colors we can choose from. there\'s like a thousand shades of white, yellow, red, blue...all the colors of the rainbow. we spent hours looking through the magazines, books and searching the net for colors. i thought it was hopeless to find the perfect color for my living room, but low and behold, the brocheres i brought home from home depot proved to be promising. i found the color...my color...the perfect color for my home and i found this in home depot\'s brocheres. yeah me!  tonight we\'re gonna paint...i hope it\'ll be a beauty...how could you go wrong with \"nature\'s whisper?!?\" yeah!   '),(21730,477665,'17,September,2003','2003-09-17 00:00:00','    wednesday...  happy wednesday all! just one more day of work for me and i\'m outta here. whoopie!!!!!  i can\'t wait till the weekend. there\'s so much i want to do and i\'m gonna do all of it. i\'m not sure why i\'m so cheery this morning...maybe it\'s because i slept like a baby. i love sleep but then again if you sleep all the time, you miss out on alot of action. but i got my 7 hours of sleep in...could\'ve been 8 but i stayed up so i could say good nite to my huney. *sigh*   '),(21731,477665,'16,September,2003','2003-09-16 00:00:00','    ...live with it  it\'s been almost 2 years since i broke up with my ex.  i can remember it like it was yesterday and it\'s even stranger to think that so much has happened in-between.  it wasn\'t a good breakup. i tried to stay friends but he wanted nothing to do with me. i think i tried for almost a year before i gave up. i was caught up in a one way strive for friendship. looking back, i feel foolish for thinking he was mature enough to stay friends with me. but i don\'t regret trying because at least i can say that i tried on my part.   to this day, he still talks to my sisters, they still hang out ever so often, and he stills calls them.  at first, i was very bothered by this. if he didn’t want me in his life, why was he so adamant in staying friends with my sisters. that leaves a connection to me and as long as he is friends with my sisters, we’ll always be connected. i didn’t like the fact that he was still in my life when he made it clear he didn’t want to be. it would be a different story if he stayed friends with my sister and didn’t ask all sorts of questions about me. but he did. my sisters told me that every time he would call them, i was the topic of discussion. i know at times it made my sisters feel very uncomfortable and ultimately they became annoyed with his countless questions about my life. i don’t get it! if he just agreed to work on being friends after the breakup, i’d tell him about my life myself and he wouldn’t have to hear it from others. if he wants to hang out with them, i don’t care anymore but just be their friend and don’t try to get info out of them for your own benefit.   in away i feel sorry for him and i know shouldn’t. it just makes me a little uneasy that he knows my business.  he lost all rights to it when he decided he was better off not being my friend. i don’t really care anymore and i don’t think that’s harsh to say. i’ve tried all I could for almost a year and just got really sick of trying. i decided 1 year ago that it was his lost and not mine. he’s my past and that’s where he belongs. if one day, our paths should ever cross again, i’m not one to turn away from a friend but i’m also not one to take friendship lightly. it means a lot to me and i can tell you that it’ll take a long time to re-build a friendship that’s been destroyed. but one thing for sure... i’m definitely not holding my breath…   '),(21732,477665,'15,September,2003','2003-09-15 00:00:00','    religion...  everyone has there own beliefs...and regardless of what a person believes in, it\'s their own prerogative. we live in a world full of ignorant people who think there religion is the best and condemn those who don\'t believe in what they believe in. we live in a world with people who could care less about religion and think it\'s a crock of shit. and then there\'s people who are in between, the ones that believe in God but that\'s the extent of it.  i think i fall in the category of inbetween. i\'ve never been very religous and honestly i don\'t really care to be. i grew up in a loving family and even though religion was not our strongest hold, we still managed to live our lives in a positive and loving environment. i think we all came out better than ok.  i guess what i\'m trying to say is that regardless of what religion you are and if you\'re a strong believer or not, never knock it because everyone is entitled to believe in whatever they want. at one time, i thought some religions were cult like but who am i to say that it\'s wrong. i know absolutely nothing about certain religions but i what i do know is what\'s right from wrong and good from bad and if i don\'t agree with it, that\'s my problem.   i\'ve come to realize that certain things i once believed in, no longer holds true. but the one thing that stayed consistent and strong was someone who\'s been there for me thru thick n\' thin. you were my eyes when i was blind. you were my strength, when i felt so weak. you were my mind, when i was wrapped up in confusion. you were my heart, when i needed to forgive. you were my family, when noone was there. you were my friend who believed in me when others didn\'t. and for that, i will always be grateful. you\'re definitely a star in my eyes and every thing that you\'re going thru right now...school, work, personal issues, your studies, family, and friends, it amazes me that you still have time for me. you\'re a beautiful person and i just hope life treats you like the ruby that you are. with that said, i hope you find what you\'re looking for thru religion that i\'ve found thru you, my family, and my love ricky.   '),(21733,477665,'08,September,2003','2003-09-08 00:00:00','    what a relief!!   there\'s so much going on in my life right now, i don\'t even know where to begin. i guess i\'ll start off by saying i just had the best sunday ever. who would have ever thought that sharing your happiness with the ones you love most could only bring a smile to your heart and theirs.  i can still feel my heart racing, my thoughts trying to make sentences, my lips trying to express my thoughts but the only thing uttered were \"ums\" and \"uh\".  i needed some help and suddenly, there you were... a vision of love from up above...i felt you with me. i felt you holding my hand. i could feel your presence and because of this,  i was able to share the most wonderful thing that has ever happened to me to someone who gave me life and showed me nothing but love.    i have to say that i feel a lift off my shoulders. it\'s free. i\'m free. the unknown is finally known. i feel total and complete happiness. and i couldn\'t wait to share the news with you.   i love to make my friends and family smile. i love to see my one and only smile. it gives me joy to see that because what was said, it made your heart lit up like the gazillion stars in the sky. with my handful of true friends, my amazingly wonderful family and now you...my life could not be more complete.   a chapter of my life is now closed and a new chapter begins. it only gets better from here and i do have to say, i\'m ready for whatever comes our way.   '),(21734,477665,'04,September,2003','2003-09-04 00:00:00','    good times..   you northern cal boys really know how to wear this so cal girl out...hehe. but i must say you guys are definitely a fun bunch and i\'m very happy that you were finally able to make it down to good ol\' san diego.  here\'s the scoop...  last weekend, i played host along with my love ricky. curtis, bren and sanford came down for the labor day weekend to visit. it was curtis and bren\'s first time being in san diego so the pressure was on. i had to show them what i\'ve been bragging about for nearly a year now..how great sd is!  but anyways, what i thought would be mission impossible, actually was very obtainable.  i was pretty nervous to see the guys again...it\'s been so long but i knew that once we saw each other, the comfort zone would kick in and that it did.    day 1:  the first night they came into town, we partied with all of the san diegians in pb.  my girlfriends and guyfriends all came out to welcome the newbies to sd. we ate, drank, chilled, and ended our night dancing till our legs turned to jello. we had so much fun. the girls went crazy, the boys stared, and we got our groove on.    day 2:  the bay area people are very deprived...i\'m sorry to say!  so therefore i had to take them to grub on \"REAL\" mexican food. we all ate california burritos from santanas...mmm\' delish!  from the look of things, i think the boys enjoyed it cuz once we got back in the car, everyone looked really tired and that can only mean that your tummy was well fed.  then we made our way to la jolla...chilled at the living room, walked around prospect street, went to the cove and of course watched the little seals hit on each other. it was pretty funny.   as the sun went down, the nightlife began. we strolled downtown and let the boys experience one of the very few \"good\" clubs in sd (if there is even one). we had dinner at on broadway as the band played in the background. i thought that was pretty cool. we had front row seats too. hehe. anyways, i hope you all enjoyed the club.   day 3:  sunday...the day we got bren on rollerblades. it was definitely a kodak moment and i jumped to the opportunity to take those pictures...hmmm blackmail...jk!  but seriously, all of us went blading on the boardwalk down at mission beach. it was hilarious to see you all skate and ohhhhh...the fall...hehe... don\'t worry, i won\'t mention any names. *wink* anyways, the weather was perfect and curtis finally got to see the beach on a sunny day.  well i hope you guys all had fun cuz all my friends here really enjoyed hanging out with you guys. i promise next time we\'ll fulfill bren\'s needs and wants...we\'ll definitely take you to the zoo bren, so no more whining! :P good times...yup definitely!  all i have to say now is mission completed...... :-)   '),(21735,477665,'02,September,2003','2003-09-02 00:00:00','    running on empty..........................................................................................................................   '),(21736,477665,'31,October,2003','2003-10-31 00:00:00','    happy halloweeeeeeeeeeeeen!!  for those out there who still have the halloween spirit, i wanna wish ya all a happy and safe halloween!!!  i love this time of year, the change of the season...which can only mean... cold crisp nights, warm fires, hot chocolates, getting lost in your bed with the covers and pillows, snuggling up to someone special, leaves changing colors, snow, snowboarding, and of course the best thing of all pumpkin smash. :-)  it\'s back guys....my all time favorite \"happy\" drink. i think i really need one right about now. it does the trick every time. it always brings a smile to my face. i remember last winter, i had pumpkin smash twice a week. i tried so hard to o.d. on it because i knew it was only a seasonal thing and i didn\'t want to miss it but my mission failed. my addiction only grew stronger.  people think i\'m nuts that i can get this excited over a drink but if they only knew how yummy it is and how much happiness it can bring. i blame my obsessiveness on the one and only \"munkie boy\". ever since my first smash two years ago, i\'ve been addicted to it. he\'s the blame to my insanity. now munkie boy...if you are reading this and fully agree with me about the drink...then i know i\'m crazy. hehe. k..gonna get some smash...(*grin*)   '),(21737,477665,'30,October,2003','2003-10-30 00:00:00','    random...  nothin\' ever stays the same and though you wish that it would, things just don\'t ever work out the way you think it would. this year has definitely been a challenge for me...i take that back...every year is a challenge and i guess that\'s all part of growing up.   it\'s 5 am and i\'m awake...this time change really sucks...i can tell already it\'s gonna be a long day. i\'m tired of blogging. i can\'t think with this fire. i can\'t breath with all the smoke i\'m inhaling. that reminds me... i had a dream the other night that my parent\'s backyard was on fire....thank god it was only a dream....sometimes i think my whole life is a dream...yup..that\'s me..tiffani living in her own little dream world...damn realilty bites!   '),(21738,477665,'15,October,2003','2003-10-15 00:00:00','    it\'s all part of life....  i\'m a very strong believer that everything happens for a reason but sometimes i wish that things never happened. last nite i had dinner with my ex. the first time i\'ve seen him in almost a year. he lost a lot of weight... not that he was ever fat but his reason for losing weight was because he wasn\'t with me anymore. geez..i guess i do eat a lot which makes whoever i\'m with eat a lot too. :-( but that\'s besides the point...my mind has been spining in circles and my heart can still feel whatever little pain he still has left. i know he\'s trying to be my friend and that\'s all i ever asked for. it took him 2 years to finally let me back in his life and after talking to him last night, i understand why it took so long. i know i didn\'t make it any easier for him all those times i would call him but i can\'t do anything about it now. i just wish the best for him and i hope he knows that. things are finally moving on.  we both got our closure and we both know where we stand...as friends.   '),(21739,477665,'25,November,2003','2003-11-25 00:00:00','    my skylla...  sunday marked my first day of snowboarding this season and i have to say boy was i rusty!  i got to the top of the mountain and as i looked down the slope, i froze. holy shit...what the heck do i do??? i totally forgot how to snowboard! how sad is that?!?  but i did make it down the mountain with the occasionally tumbles which was cool because i wanted to get a feel for the snow you know?…heehee.  so after that, we went on the lift again, back to the top but this time i did a lot better. i was able to pick up my snowboarding abilities that i left behind in tahoe last winter but i’m sure some of you would beg to differ. :P  anyways, we got in about 6 runs, which is pretty good considering it was a busy sunday.  oh yeah, i rode my new baby too, it’s a k2 skylla board and it’s purty, a real purty red. k..that\'s it...see ya all on the slopes!   '),(21740,477665,'13,November,2003','2003-11-13 00:00:00','    let it snow, let it snow, let it snow!!!!   this season, we\'re gonna take over the mountain and this time i don\'t mean *tumble* *fall* *sky*, which has been my destiny before. lol!  anywhoo, i\'ve been itching to go since the summer. i even got season passes a couple months ago and then we went to this huge snowboard sell and got free tickets, new googles and gloves and tomorrow we\'re going to snow jam to get me a new board and butt pads (ummm...butt pads...not for me but just incase someone else may need it..*grin*) and of course to get more free tickets. i love this feeling...you know...everyone in the snowboarding spirit. it\'s like christmas but better cause it can last for months just as long as the snow gods are on our side. let it snow, let it snow, let it snow. Amen!   '),(21741,477665,'12,November,2003','2003-11-12 00:00:00','    good for the soul...  last night i went to see \"elf\" with some friends i haven\'t seen in awhile. i felt a little uncomfortable and a bit nervous at first. i wasn\'t sure what judgements they have made if any at all. the awkward feeling stayed with me and i felt it radiating from them as well. finally the movie started and everything was forgotten. for 2 hours, i got lost into the movie. it was hilarious. i can\'t remember the last time i laughed so much from a movie. i love comedies and ooohhh how i love romantic comedies even more. anyways, after the movie, we all re-grouped but this time the awkwardness dissipated. everyone\'s spirit was lifted. i think the movie did us all some good. yup laughter is the key to a happy life and laughter with friends, even better. so if you wanna enjoy some good laughs, go see \"elf\". it did me wonders. :-)   '),(21742,477665,'10,November,2003','2003-11-10 00:00:00','    in limbo…  what if you knew you were doing wrong but it felt so right? could doing what’s wrong be right? Or could doing what’s right be wrong? my heart and my mind do not coincide.  opposing teams both striving to triumph over the other for at the end a decision will be made. it’s usually anyone’s game but this time it appears that both teams may lose or win. i’m hoping for the latter….   '),(21743,477665,'06,November,2003','2003-11-06 00:00:00','    from within...  and  through all the tears and the  sadness and the pain comes the  one thought  that can  make me  internally  smile again, i  have  loved.  November 6th, 2003...enough said.   '),(21744,477665,'03,November,2003','2003-11-03 00:00:00','    rain. it rained. I fell. it rained I loved. it rained. I lost. it rained. It hurts. I rained. rain.   '),(21745,477665,'29,December,2003','2003-12-29 00:00:00','    where\'s my head??  i think i\'ve developed a case of stupidity. there must be a huge whole in my head where my memory use to be stored. i can\'t seem to remember anything.  i\'ve done a lot of airheaded things lately...you know like looking for my keys when it\'s in my hand or even worse, when it\'s already in the lock to the car.  get this, i did it twice, once at my house and another time at my parents...yup...on christmas eve, i left my keys in the lock to the trunk of my car.  one of my neighbors knocked on the door around 9pm to point out that some idiot left the keys on the lock to the trunk. let\'s just say my parents weren\'t thrilled by the visit.   then i lose/misplace my free southwest ticket for utah. yep..snowboarding in park city! i tore up my room looking for it. i was seriously stressed out for a week. i called southwest last week to see if they could help but they said i would have to contact rapid rewards but they wouldn\'t be open until monday...that\'s today. so i called early this morning. anyways, to make a long story short, my free ticket turned out to be ticketless so a ticket was never really sent to me...it was just a notification letter. dammit..i knew i should have read that letter. yes...i\'ve been worrying and causing problems for nothing. i\'m sorry. =(  wait, there\'s more...but do i really want to share it...nah..i think that\'s enough for today. i really hope this is just a phase i\'m going through and that my memory will come back so no one will ever compare me to jessica simpson again. how sad is that?!?   '),(21746,477665,'24,December,2003','2003-12-24 00:00:00','    looking back on 2003...  the year started off with a bang! with not a care in the world, single as can be, i danced the night away with kimball collins. met a few strangers, got silly drunk and still managed to make it home solo.  the winter months consisted of partying every weekend. that was our love..we had the energy, we had life, we all had something in common, a joy in dancing, feeling the vibe, and letting go. at times, i miss those days.  spring blossomed a love so beautiful, which can be compared to the rarest diamond but even with the most exquisite diamond, there\'s imperfections which this beauty definitely had. but it was all worth it and if i had to go through it again, i\'d do it in a heartbeat.  fall was a cold month and i\'m not just talking about the weather. the truth revealed ugliess i didn\'t know existed. i could have lived without this but it happened for a reason and that i have to remember.  and now here i am again...approaching the end of the year. it\'s been a heck of a ride. a year of change. a year of love. a year of loss. a year of happiness. a year of rekindle friendships. a year of accomplishments. a year of life!!! i\'ve learned a lot this year and i\'ll take everything i\'ve gained with me in the new year and the baggage will definitely stay behind. and with that said, i want to wish   everyone   a very Merry Christmas and a Happy New Year!!!!! i wish you all the best in 2004. XOXOXOX!!!!!!   '),(21747,477665,'23,December,2003','2003-12-23 00:00:00','    the day n life of a beginner...  day 1- if it\'s steep, double diamond steep, i\'d plow, if it wasn\'t, i\'d be riding on my heelside. great weather. good conditions. ate it many times.   day 2- if its\' steep, double diamond steep, i\'d plow, if it wasn\'t, i\'d be riding on my heelside and at times toeside. tried the rainbow rails many times. fell everytime.  did some jumps. landed a few.  day 3- if it\'s steep, double diamond steep, i\'d plow, if it wasn\'t,  was carving the mountain ever so slow. attempted more rails. longer rails. ate it again.  did more jumps. also ate it again.  day 4- if it\'s steep, double diamond steep, i\'d plow, if it wasn\'t, i was carving the mountain at a faster pace. ate it a few more times. landed my rainbow rail. landed more jumps.   day 5-if it\'s steep, double diamond steep, i\'d barely ride it down the right way. if it wasn\'t, i was carving the mountain at a much faster pace. attempted a box rail. landed more jumps.  day 6-if it\'s steep, double diamond stepp, i\'d attemp to ride it down the right way. if it wasn\'t, i was catching more speed. no rails this time..all jumps. tried a trick. great weather. a bit icy.   day 7-  i will ride down the steep mountains the right way. i will do more rails and land them and i will get more air on my jumps.  to be cont...   '),(21748,477665,'10,December,2003','2003-12-10 00:00:00','    where\'s the spunk in me?  ok, last night i went to bed at 8:30 and woke up at 6:30. the night before, i did the same. that\'s 10 hours of sleep!!  on average, i sleep about 8 hours a night. things are definitely changing. sleep is becoming more important to me...or more of a necessity.  i remember not too long ago, i slept 3-4 hours a night....insomnia plagued my mind or i was just out playing. i miss those days where i didn\'t need to sleep. i remember doing so much more after work but nowadays, i come home...whip up something to eat, shower, and jump into bed and snooze till the next day. the only thing that keeps me going is the thought of the weekend. i love my weekends...i cherish every weekend. these are the times where i have my fun and feel alive again.  i\'m just hoping my week days won\'t be like this for the rest of my life cuz that\'s no life.  oh, did i mention that my weekends will soon consist of only saturdays and sundays ...no more fridays off. =( so basically i\'ll be sleeping a whole lot more and having a little less fun since fridays are now shot.  somehow i need to find away to make my weekends longer. hmmm....i guess i\'ll be calling in sick a lot starting january. =) haha.   '),(21749,477665,'04,December,2003','2003-12-04 00:00:00','    drifting along...  i can\'t think of any earth-shattering events to blog about. i have nothing to share. no exciting news. all i do is work, eat, sleep and snowboard. and to tell you the truth, that\'s all i care to do. don\'t really wanna club, don\'t really care to shop, don\'t wanna go to the gym...i\'ve become a lazy bum. maybe it\'s the weather. maybe it\'s the people. maybe it\'s the life i\'ve made for myself.  i\'ve already accomplished most of the goals that i\'ve set out but the one goal i\'ve been striving to achieve for over 24 years have still not panned out. i\'m tired of trying, i\'m tired of workin my ass off and not getting what i deserve, i\'m sick of broken promises, and i\'m disgusted by politics.  sometimes i just feel like throwing up my hands and walking away but i won\'t. i won\'t let them win.   so i guess my life will keep truckin on untill that day comes when life is free...when i\'ll be my own boss...   '),(21750,477665,'23,January,2004','2004-01-23 00:00:00','    blah, blah, blah, blah...  \"blahging\" is a powerful tool. it can help get a message across through a fairly easy channel. but what if the person you\'re trying to get the message to doesn\'t even read your blog? or maybe that person does but you\'ll never know because that person chooses to stay quiet, which kinda sucks. i feel so BLAH!!!!!!   '),(21751,477665,'22,January,2004','2004-01-22 00:00:00','    the year of the monkey  yes today is the day of new beginnings for me. no more year of the goat...what a relief!!!  they say in the chinese belief that if you\'re born the year of the monkey and this is the year of the monkey...it\'s a bad year for you.  i\'m glad i\'m not a monkey! anywhoo, happy chinese new years.   '),(21752,477665,'16,January,2004','2004-01-16 00:00:00','    i\'m no stick?!?  so last night i went out drinking with the guys and i have to say they\'re definitely a crazy bunch. who knew that older men could have fun and not be perverted. a group of men at my work go out all the time for happy hour and always invite me but i always turn them down. but yesterday was different. i decided to join them and i\'m happy i did.    i was a bit uneasy at first about the idea of partying with my coworkers especially the top level managers and with two of the chief officers. but as i sat down, i felt very comfortable. everyone was happy, everyone was just being themselves. no politics, no talking shop, and no stress.  it was great. so i ordered myself a fruity drink. everyone was shocked. no one thought i drank and when they saw me, everyone just cheered me on, so we toasted and toasted and toasted...to what...to toast so i\'d drink up.  anyways, i started having a side conversation with one of the managers and he was telling how one of the ladies at work would give me dirty looks behind my back. he said that when i\'d walk by, she would look at me and grunt...\"uhhh\".  he told me this particular lady in his department was jealous of the younger ladies at work. sorry that i\'m not an old fat fart but geez...there\'s no reason to be nasty. then he looked at my plate and said \"you don\'t eat much huh?\" lol....little does he know i\'m a P-I-G. i tried telling him that i do eat a lot but he wouldn\'t listen...he just kept calling me skinny and then he said \"You\'re a STICK! Ever since i\'ve known you...you\'ve been a STICK!\"   what the hell?? that\'s like the worse thing you can say to a women.  yeah..i\'ve been sick a lot this year and lost weight but i don\'t think i\'m sick-stick looking. :-(  i\'m sure there\'s guys out there that like the olive oil look like popeye but i personally like meat on a women and i\'m sure most guys do too.   anywhooooo, it was fun..quite interesting i have to say. i think the highlight of the night was when one of the chiefs asked if i order a \"new england ice tea\"....NEW ENGLAND???? i think he meant LONG ISLAND...it was hilarious...you just had to be there.  trust me...i won\'t ever let him forget that.   till our next happy hour escapede...cheers!   '),(21753,477665,'15,January,2004','2004-01-15 00:00:00','    a perfect fit...  i\'ve experienced something so amazing the last couple weeks that i\'m just simply overwhelmed with such a desire to somehow explain my revelation. i\'m not sure if i completely understand it myself but i think by seeing it before me may help. so here\'s my first shot...  since highschool, i\'ve always had someone..either a guy i was crushing over, a guy i was dating, a boyfriend, or just a bunch of random guy best friends. i\'ve always been surrounded by friends and family. i was never really alone and if i had a broken heart,  i knew my friends would be there to help pick up the pieces and fill the empty void.  i know that i never wanted to be alone. i was always on the go...always had something to do and somewhere to be. but lately my motivation has fallen and i just don\'t care about a lot of things that were once important to me.  alone and lonely are two distinct words. i\'ve never been alone but i have been lonely and these past weeks have been very difficult for me to breath. i was finally alone. just me, myself and i.  honestly...it was very hard at first. i couldn\'t sleep and i couldn\'t eat...now those who know me, know those are my hobbies.  but after a few days of being alone, things began to change. my outlook transformed before me and being alone became more of a blessing.  last saturday, i spent the day on my porch. it was a beautiful sunny san diego day. i plopped myself on my papason chair and let the sun consume my body. i felt so at peace. my mind wandered and finally things started to make sense. the puzzle i\'ve been playing with all these years finally came together except for one piece. but that piece is not in my possession and only the person who has it, can complete the puzzle.   but there is no mystery who this person is. he\'s been with me all along even when I was alone. only now has it become crystal clear.  i\'m no longer in a confused state of mind. i\'m not alone. i\'m more sure than ever about my feelings, about myself, about what i want and what i need. all those days of trials and tribulations have finally been defeated. i\'m not gonna fight it because i\'ve already won.    being alone has proven to be the best thing that could have ever happened to me. yes, i was miserable at first but like they say...without sacrifice there\'s no such thing as success.  i\'ve realized that you are the piece to my puzzle and i hope one day, you\'ll take that piece and put it together because in my heart..it\'s a perfect fit.   '),(21754,477665,'10,January,2004','2004-01-10 00:00:00','    day 3...  still coughing, still congested, still feeling quite shitty...i really hate being sick without having someone take care of me.   '),(21755,477665,'09,January,2004','2004-01-09 00:00:00','    day 2...  *sniff* *pout*...tissue anyone??   '),(21756,477665,'08,January,2004','2004-01-08 00:00:00','    day 1...  it\'s only been a day...it feels like months....wahhhhhh come back home!   '),(21757,477665,'26,February,2004','2004-02-26 00:00:00','    snow, snow and more snow!!!  finally the time has come. in about 4 hours, we will be on our way to mammoth. hanging out with great friends, the love of my life, and on top of that...we get to go snowboarding.  what more could a girl could ever ask for?!? i\'m ready for the big jumps and high rails! oh the possibilities.... boy oh boy...i can hardly sit still...mammoth, here we come!!!   '),(21758,477665,'23,February,2004','2004-02-23 00:00:00','    \"With You\"  The real me is a southern girl with her Levis on and an open heart Wish I could save the world, like I was super girl The real me used to laugh all night lying in the grass just talking about love But lately I\'ve been jaded life got so complicated  I start thinking about it, I almost forgot what it was like To know what feels right \'Cause with you I can let my hair down  I can say anything crazy I know you\'ll catch me right before I hit the ground With nothing but a T-shirt on I never felt so beautiful Baby as I do now Now that I\'m with you  With you, with you, with you Now that I\'m with you  You speak and it\'s like a song And just like that all my walls come down It\'s like a private joke just meant for us to know I relate to you naturally Everybody else just fades away Sometimes it\'s hard to breathe Just knowing you found me \'Cuz I start thinking about it, I almost forgot what it was like To know what feels right \'Cause with you I can let my hair down I can say anything crazy I know you\'ll catch me right before I hit the ground With nothing but a T-shirt on I never felt so beautiful Baby as I do now Now that I\'m with you With you  Come and take me Love you save me Like nobody else Now I can be myself With you   '),(21759,477665,'03,February,2004','2004-02-03 00:00:00','    a prayer...  i see it in her eyes. i can feel it from her touch.  i know it in her heart.  the light shines but all the same time, the light dims. it\'s dark.  she\'s lost. she can\'t find her way back. it\'s dark. she believes she\'s dying. her body becomes numb. she\'s grasping for air. the pounding of her heart multiplies. the pain becomes unbareable. someone help her! God help her!   i sit and feel dazed and a little angry at times. i sit and wonder and ask myself over and over \"why her? hasn\'t she been through enough?\"  all i\'m asking is for this to stop. she doesn\'t deserve it. we all know she\'s a fighter. always have been but how many times does she have to prove herself. can\'t you see...it\'s killing her. it\'s killing us. and the worst part is not knowing what\'s wrong and even tougher is not knowing how to help. i wish i could take her pain away. i wish i could make it all better. i would trade places with her in a heartbeat. i see her before me and i\'m dying inside but i put up a front for her sake. i can\'t crumble.  she needs all of us to be strong. we will try our hardest but that means you have to pull through this one.  i believe in you like you believe in us.   at one point in my life i stopped believing in God. but as i\'m getting older, i know that all the bad things in my life was not God\'s doing. it was my own. i don\'t blame him...not anymore. i\'m a believer and although i do not go to church or temples...i still pray every night so that must count for something. so please out of the thousand prayers i say each night,  i really hope you listen to this one.   '),(21760,477665,'05,March,2004','2004-03-05 00:00:00','    i\'m back..back on my traveling kick. i won\'t be going on any long trips but definitely i\'ll be on mini vacations as often as possible. who knows where i\'ll go. the skies the limit. i\'m aiming for utah...again...and hopefully this time..mission will be completed. i love traveling..it\'s my passion and to have someone who loves traveling as much as i do, is just heaven. with you, it doesn\'t matter where we go..it doesn\'t have to be extravagent. i don\'t need europe, i don\'t need asia, i don\'t need the tropics. that can happen later and if it never does, i don\'t care. what makes a great trip is the experience of something new with someone worth it. and all these mini weekend getaways that we take, are awesome...simply unforgetable. you\'re awesome!! let the adventures continue!!!!   '),(21761,477665,'04,March,2004','2004-03-04 00:00:00','    i almost forgot that there are genuine people out there. i almost forgot that someone can touch your heart without really knowing they did. i almost forgot that friendship can be forever lasting. i almost forgot that i really do have true friends. i\'ve met a lot of people and have made a lot of friends along the way. i\'ve seen the ugliest side of some and the beauty in others. it hurts to know that some friends stabbed you in the back and never looked back. it\'s like you served a purpose for them and now that you\'re not needed, it doesn\'t matter. too many fakes out there i tell you. and then you start thinking that maybe all your friends are capable of deception. but then one day, you open your eyes and see that not everyone out there is like that. a couple nights ago, i had a heart to heart conversation with my girl. and she started talking about her life, boys, friends, and family. we laughed and we cried. i felt a connection and the light went on...she\'s been there all along and i didn\'t see it. she was the pillar to my strength. my crazy, sexy, cool partner in crime. she loved me for all of me and had my back the whole time, especially when others looked the other way. i think i\'ve made a mistake by believing that friends like that do not exist. and now that i see what i should have seen months ago, i know i\'ll never let go. it goes deeper. the rarest of friendships. i don\'t think she knows it but i know she feels it. it\'s nice to know that even when words are not said, you know the friendship is forever there. i love ya girlie girl!!   '),(21762,477665,'03,March,2004','2004-03-03 00:00:00','    ok, this had to be the best snowboarding trip i have ever been on. it was perfect. it snowed all week long and stopped when we got there so we got to board in perfect sunny weather and on fresh powder. i got to know ricky\'s friends a lot better...more than i ever wanted. at least i\'ll be prepared for next year\'s trip. Note to self: No chilli dogs for the guys. I swear to you...our cabin stunk with fart bombs the entire weekend. anywhooo, what shall i share with you folks??? ahhh...i got wind burned and looked like a lobster. rielyn got drunk and also turned into a lobster. she became my twin. i found out that the guys take way longer to get ready then the girls especially keith. he is definitely a snowbunny. eric sucks at jenga..i think he had the most shots...poor poor boy. ricky is the chef of the group. jon smells and is two tone. joanne has the cutest googles...velvet..i think i want one too now. mary likes to board alone. no jumps for sherman. lynnda loves to drive. hey ricky!! she needs to go on all the long trips with us. hehe. damn..i had so much fun. the only thing disappointing were the parks...not enough beginner jumps and rails. maybe next year, mammoth will have better parks because i love to hit those jumps. i guess big bear will have to do until then..yes...i\'m ready to hit the slopes again! one more thing to sum up this trip...\" urlLink EAT QUIZNOS SUBS!!!!!!!!!!!!!!!!!!!!!!!!\"  here\'s the link to the clip in remembrance of our trip. play the clip...lol :P   '),(21763,477665,'28,April,2004','2004-04-28 00:00:00','    i love diet v8 splash, tropical blend. it\'s the best juice to have ever touched these lips. remember the popsicle that they use to sell back in the days, big sticks. that\'s what it tastes like...melted icy cold big sticks. mmmm\' yummy!   '),(21764,477665,'27,April,2004','2004-04-27 00:00:00','    ok..i know this post is a little late but i\'ve been busy. just wanted to thank you again for coming to seattle to be with me on your birthday. that was awesome. well enjoy your last year of being in your twenties..you old fart! heehee. you have about 3 more days to use that line \"but it\'s myyyyy birthday\" to get what you want then it\'ll be my turn. yes! happy birthday sweetie!   '),(21765,477665,'14,April,2004','2004-04-14 00:00:00','    they say this city only has 60 days of sunshine a year. they call it the city that forever rains. they call it the suicidal city. i beg to differ. last week i went to seattle and just fell in love the first day i was there. the weather was nice and grass was green. i went on a rampage around downtown seattle. i hit the many malls, starbucks, pikes, space needle, toi\'s, dalihia\'s lounge, salty\'s, alkine beach, cinerama, and much more. seattle is just beautiful. from downtown, i could see mt. rainer and yes it looked like a huge scoop of vanilla icecream. and ohhhh the tulips..how i love tulips...they were everywhere and in every color you can think of. every night i was up till the wee hours in the morning and if you know me, i sleep hella early. i had a tough time falling a sleep...yup...i was sleepless in seattle. by the time my week was over, i was exhausted. i think one night i was up till 4am and had to get up at 6:30. it was bad but well worth it. i love traveling but man..as much as i love it, i also love being home. i missed my fishies, my house, and mom\'s home cookin\', but most of all, i missed my BED! there\'s simply no place like home. but i shouldn\'t put my suitcase away just yet, i think i hear southbeach callin\' my name. :-)   '),(21766,477665,'14,April,2004','2004-04-14 00:00:00','    &nbsp; there are way too many reality tv shows out there but i can\'t help to watch them. ever since i got cable, i\'ve been watching more tv then I have in the pass 3 years. i\'m addicted to american idol. jasmine rocks!!! linda thinks her georgie will win but come on on..his best move is the squat dance. anyways, i also watch newlyweds...it just cracks me up. Real World/Road Rules challenge is a must-see. and if that wasn\'t enough, i\'m now stuck on this new show, \"the swan\". the swan takes average looking people and transforms them into beauty queens...it\'s kinda like extreme makeover. well i watched the episode on monday and i was just flabbergasted. this women. needed a tummy tuck after having only one baby. you should have seen her stomach..it was nasty. there was a lot of loose skin and i mean a lot! it just hung over her and sagged so far, i felt sick. ok..if this is what my stomach will look like after having a baby..i don\'t want one. she was only 27 and had just one kid and her belly looked like that. given, my mom had 4 kids but i\'ve never seen her stomach. she always refused to show me. i asked her a million times but she always said no...now i know why. for all the women out there that are planning to have babies, here\'s a bit of advice..wear tight tops, short tops, halter tops, see through tops while you can before it\'s too late...cuz once you have that baby..say good bye to your nice stomach and cutsie tops unless you don\'t mind wrinkeled, sagging, excessive skin hanging out of your clothes. or you can always get a tummy tuck. :-)   '),(21767,477665,'25,June,2004','2004-06-25 00:00:00','    a smile says it all. :-)   '),(21768,477665,'21,June,2004','2004-06-21 00:00:00','    i\'ve had a lot of embarassing moments in my life. some events i would never mention to anyone even if my life depended on it but for most of these moments, i think they\'re worth sharing especially if i\'m still laughing about it.   on saturday, my friend invited me to a surprise birthday party for one of his friends. and although i\'ve never met any of his friends before, i thought it would be fun to check it out. when i got to his house, my friend wasn\'t there yet and neither was the birthday boy. i walk into a room full of strangers and this guy introduces himself to me and then introduces everyone else in the room. he showed me around the house and shortly after, i plopped myself on the couch waiting for my friend to show up for the surprise. i looked down and i noticed that i was wearing my black sandals but i swore i was wearing white sandals. so i looked down at both feet and i saw my white sandal on my other foot. holy shit! i had a black sandal on one foot and a white one on the other. i thought to myself...no f*cking way am i wearing two completely different shoes. they weren\'t even the same height or the same style! one was dresser then the other. how did i not feel the difference? by now my heart was pounding, my mind was racing, and i could feel my face turning a rosy pink. all of sudden, i blurted out \"hey look..i\'m wearing two different shoes. it\'s the new trend. hahaha. ok.. (by now, my face has turned a bright shade of red) oh my god...um..i\'m so embarassed. um..i live in my car and i slipped on the wrong shoes. ok.i\'m gonna go to my car and get the other matching pair... hahaha. i\'m such a dork.\" i then take off to my car to grab the other shoe. i know for a fact that after i left, the laughter probably increased by 110%. yes i\'m an idiot. so i come back to the party and everyone is just laughing and i can\'t help but to laugh right along with them. it was definitely an ice breaker. i was now the center of attention and everyone was asking me questions. the rest of the night consisted of great food, good music, and the constant teasing of my new trend. it was great...i\'m happy everyone had a good laugh even at my expense. for sure they will never forget me.   anyways, word has already gotten around about the incident so i wanted you all to be the first to hear it from me before you hear it on the street. damn..i room full of strangers..what a great way to make a first impression. :)   '),(21769,477665,'16,June,2004','2004-06-16 00:00:00','    ok..so i got this fish tank for valentine\'s day and i really wanted to have a pet shark but instead a got a bunch of feeder fishes. for some reason, they all got sick but one. the runt out of the group survived. she outlived all her brothers and sisters. then one day, i got a few guppies added to my tank but that didn\'t last long because they miracuously disappeared while i was on vacation. so now all i have is one guppy and a gold fish name wanda. i ran out of food a week ago...i\'ve been neglecting them. i never really got to attached to the guppy but wanda, i did care about and i didn\'t want her to die. well, i was at my parent\'s house the other day and i saw my sister\'s turtle and thought about wanda. i asked my sis if i could have some of her turtle food for my fishy. yes i am lazy and didn\'t feel like going to the pet store. well when i got home, i crushed the turtle stick food and put one in the tank. wanda was so excited cuz she hasn\'t seen food in awhile. she went in for the food and she choked. the stick was too big for her little mouth. she was gagging. i didn\'t know what to do. i was so scared for her. i could see the food in her mouth. it was too hard for her to swallow it. a tear came to my eye...i felt my fish slipping away from me and if she died, it would have been all my fault. i called my friend and told him that my fish was choking. he was no help. he just laughed at me. so i hung up the phone and prayed that the food stick would just get soft so she could either swallow it or cough it up. after about 5 minutes of gagging, she coughed up the food. i was so relieved but then i saw her going for the food again but luckily it was soft enough for her to swallow. go figure.   well that\'s my fish story. i hope i never have to live through that again. it was pretty traumatizing. but i did learn something...that maybe i shouldn\'t get a shark ..and not to feed wanda turtle food. :-)   '),(21770,477665,'13,June,2004','2004-06-13 00:00:00','    I\'m so tired of being here  Suppressed by all my childish fears  And if you have to leave  I wish that you would just leave  \'Cause your presence still lingers here  And it won\'t leave me alone   These wounds won\'t seem to heal  This pain is just too real  There\'s just too much that time cannot erase   When you cried I\'d wipe away all of your tears  When you\'d scream I\'d fight away all of your fears  I held your hand through all of these years  But you still have   All of me   You used to captivate me  By your resonating light  Now I\'m bound by the life you left behind  Your face it haunts  My once pleasant dreams  Your voice it chased away  All the sanity in me   These wounds won\'t seem to heal  This pain is just too real  There\'s just too much that time cannot erase   When you cried I\'d wipe away all of your tears  When you\'d scream I\'d fight away all of your fears  I held your hand through all of these years  But you still have   All of me   I\'ve tried so hard to tell myself that you\'re gone  And though your still with me  I\'ve been alone all along   When you cried I\'d wipe away all of your tears  When you\'d scream I\'d fight away all of your fears  I held your hand through all of these years  But you still have   All of me.....   '),(21771,477665,'25,July,2004','2004-07-25 00:00:00','    i have butterflies in my tummy. in about 10 hours, i\'ll be flying in the air. i still can\'t believe i\'m going to do this. i\'m excited. i\'m nervous. i\'m scared. i\'ve talked 4 other friends into doing this too. what have i done? i know they\'re scared shitless right now and i think it\'s rubbing off on me. i can\'t sleep. yep...i\'m gonna do it!!! i\'m gonna jump out of that plane. i\'m gonna shudit, relax&nbsp;and musn\'t forget to breath now... *long sigh*&nbsp;   '),(21772,477665,'21,July,2004','2004-07-21 00:00:00','    so yesterday morning i was awaken by the light shining through my window. it was&nbsp;the crack of dawn.&nbsp;since i was up, i decided to go to work early (disgusting huh?).&nbsp; so i get ready and head out singing along to a mix cd...and might i add it is the best mix cd i\'ve ever received (thanks j).&nbsp; so i\'m on the freeway trying to get on the 94 when all of a sudden i hear this loud rumble. i wasn\'t sure what it was but i looked around to see if they were doing some sort of construction on the freeway. i kept driving and the rumble became more intense...the noise&nbsp;was following me. \"oh crap...i think it\'s my car!\" so i pull over to the side of the freeway and i smell&nbsp;something burning.&nbsp; i get out and&nbsp;notice that my front tire is flat and the rims are&nbsp;bent. WTF!!&nbsp; so i sit back in my car and&nbsp;cry like a baby...ok....jk, i didn\'t do that but trust me, i wanted to.&nbsp;&nbsp;what really happened was that i changed out my own&nbsp;flat!!!!!&nbsp; lol...ok..that didn\'t happen either.&nbsp;instead i&nbsp;got out my cell and started scrolling down my numbers and the first person that i thought of, i couldn\'t call. the second person, i knew would do it but probably think i like him or something, and then your name came up....ahhh yes..my buddy jay to the rescue! so i call jay and he comes and saves the day, changes out my flat and of course i\'m watching him do it just so i can do it myself next time.&nbsp; :-)&nbsp; i\'m very thankful that he was up that early and took the time to help me. thanks jay!&nbsp; well my first flat tire and i hope it\'s my last cuz who knows... next time i might not be so lucky...in other words...i might not get a hold of anyone. lol....   '),(21773,477665,'19,July,2004','2004-07-19 00:00:00','    every now and then i found myself back to where no one can hurt me.&nbsp; and although my mind is cluttered with what if\'s, should\'ves, could\'ves and would\'ves, this place clears my thoughts but only for as long as i\'m on the run. i\'ve always enjoyed the outdoors especially the lakes, the beaches, the cliffs, and the mountains. some of my favorite&nbsp;solo spots, i\'ve&nbsp; never had the chance or taken the time to share it with anyone. in away that\'s good because lately i\'ve been hanging out there an awful lot but to have memories there with someone you love as well, would only defeat the purpose of my solo spot. a place to&nbsp;clear my mind and feel at peace with myself.&nbsp; &nbsp; it\'s been awhile since i\'ve&nbsp;put on my running shoes&nbsp;to&nbsp;get away from it all but for the&nbsp;past couple months..that\'s all i\'ve been doing...running.&nbsp;&nbsp;i guess it\'s the forrest gump in me. i don\'t ever want to stop.&nbsp; i want&nbsp;to keep on running until my mind is free and&nbsp;my heart is&nbsp;whole again.&nbsp; today i ran 6 miles and i think i sprained my ankle.&nbsp; my knees are getting weak. they\'re not as strong as they use to be. but&nbsp;that doesn\'t stop me...my run turns into a jog, then a walk, and ultimately&nbsp;a&nbsp;&nbsp;limp...i don\'t care...whatever it takes to&nbsp; keep going. i start to feel the pain even more. it hurts. &nbsp;at one point, i\'m stopped by a roller blader. i\'m not sure if he was hitting on me or actually wondering if i was ok but it doesn\'t phase me...i kept going.&nbsp; &nbsp; &nbsp; i\'m not sure how much longer i can do this. my body&nbsp;says to stop, my mind says to keep going, and&nbsp;my heart calls out your name.&nbsp;i dunno...i&nbsp; guess i conflict with myself but what i do know is that&nbsp;i\'m definitely running on empty....   '),(21774,477665,'18,July,2004','2004-07-18 00:00:00','    one touch is worth a thousand words....   '),(21775,477665,'16,July,2004','2004-07-16 00:00:00','    1 Corinthians 13:4-8 \"Love is patient, love is kind. It is not jealous, it is not pompus, it is not inflated, it is not rude, it does not seek its own interests, it is not quick tempered, it does not brood over injury, it does not rejoice over wrongdoing but rejoices with the truth. It bears all things, believes all things, hopes all things, endures all things. Love never fails.\" &nbsp; &nbsp;   '),(21776,477665,'02,August,2004','2004-08-02 00:00:00','    there\'s so much i want to say and so much i can\'t say. so i\'m leaving it up to fate with a little help from me. i do believe that events in our lives are predetermined but it doesn\'t mean it can\'t be changed. As long as you are in control of your life, you have the power to change your fate. With that said, i\'m hoping that by making this decision to post this blog, my fate will be in favor of my destiny.  so here it goes...  our 100th day together brought us to a place where nothing matters. the focus was not on the world but on us. feelings were carefully thought out and written on paper. i can still remember how nervous i was and how excited i was to hear your letter. as your words flowed freely from your lips, i held on to every word that you said and kept \'em near to my heart. i was simply amazed. and then it was my turn. my voiced trembled as i read my letter to you. i couldn\'t focus on my letter because your words were still lingering in my mind. i couldn\'t breath so i paused and with your gentle caring touch, i was able to continue. at the end, i think we both knew we had something special, something worth every strength to hold on to and that\'s why it\'s so hard to let go.  if fate is on my side, you will find yourself back to my blog and read this entry. on august 13, 2004, 7:30 pm, i will be sitting at the place were we exchanged our letters. i am also hoping that fate will bring you to that same place. i will wait there for you and if you show up within the hour, i will believe that fate brought you back to me for a reason. and if the only reason is to talk about what happened or to get closure for the both of us then so be it.   there are so many unknowns right now and i know the only person that can get hurt by this is me. i have no idea if you still read my blog and there\'s a pretty good chance that you don\'t. i have no idea how you feel about me or if you even think about me anymore. i have no idea if you are working that night or have other plans. but what i do know is that my pride has gotten the best of me and i\'m done with being stubborn. i want to do this...it\'s something i have to do and if you don\'t show up, i will know that my destiny is not you.  i\'ve always believed communication is vital. it has definitely been lacking between us for the past few months. so now i\'m leaving it up to fate. it\'s a chance i\'m willing to take. i\'m not hoping for a good outcome or expecting a bad one. all i\'m wishing is that i will come to accept that this is the way it was suppose to be or not be and with that realization, i\'ll be able to move on and that\'s all i can do.   '),(21777,477665,'01,August,2004','2004-08-01 00:00:00','    yep..it\'s over. it lasted for a good week and now it\'s back to reality. i\'m spent i tell you. my eyes still burn from lack of sleep. my body aches. i\'m too tired to actually explain all that happened but i think if i don\'t blog about it now, it will only be memories in my mind and you all know how forgetful i am these days. so here\'s a brief summary of my vacation...  monday- july 26: i never experienced anything so amazing. i had my first  air gasm. skydiving has been on my list to do for quite some time now and i finally did it. i can still remember looking down at the word as i jumped out of the plane. it was such an adrenaline rush and i want to go again!!!! i think everyone should try this. it has definitely put another spin in way i think and i\'ve gained another perspective on life. oh yeah, i got a pedicure and manicure too. we then ended our night downtown at RA.  tuesday- july 27: off to New York City! whooohoooooooo!!!!! so excited, i was coerced into taking sleeping pills on the plane in order to catch up with the time difference. bad mistake!  wednesday- july 28: the marriot, queens, shopping, brooklyn bridge, rosa\'s pizza, met a bunch of new friends, rockefeller center, trump towers, mtv trl, time square, central park, lounges-bob, music, drinks, korean restaurant.  thursday- july 29:roy\'s restaurant, world trade center, century 21, statute of liberty, elis island, staten island, hot dogs, indian food, kats pastrami sandwich (where harry met sally was filmed) free show at the park, comedy club, hot dogs again.  friday-july 30: saw usher live at bryant park, soho, chinatown, koreatown, subway, virgil, broadway muscial \"Aida\", went clubbing at the \"park\" - $200 bucks for a table! ouch! late night eats of course.  saturday-july 31: back to sd, da kine in pb, caught up on sleep.  sunday august 1: birthday party in irvine!  that\'s all i can remember right now which i think is pretty good considering how tired i am. new york is crazy. i think i slept 2 hours each night..if that. i ate like a pig...i think more than i usually do. hard to believe huh? every restaurant we went to was sO freakin\' good. it was heaven. the lounges/bars were awesome, the clubs were kickin\', and the vibe was great. i\'ve met so many wonderful people there and they were all great hosts! thank you! if any of you ever find yourself in san diego, i\'ll definitely take good care of you!  what a vacation! we crammed so much in the short time we were there. it was extremely busy at times, rushed, a little overwhelming, very crowded, too much traffic, hot and humid, and sleep deprived but damn...i loved every second of my trip. thank you!!!!!!!!!!!!!!!!!!!!! XOXOXOXO. how will i ever top next year\'s birthday?!? so the question is will i be back??? of course but not anytime soon. i still need to recover from my vacation. it\'s one i\'ll never forget and as much as i loved new york, i\'m so happy to be home in good ol\' san diego. ok...need sleep...yes...sleep.   '),(21778,4103386,'15,August,2004','2004-08-15 00:00:00','   Wheeha...im so high today!!! Got water baptized =) Won the gold medal for orienteering =) Got the hillsong more to life CD =) Had a nice dinner with my great friends =)  All day roxX  Verse 1: I don\'t care what they say about me It\'s alright, alright I don\'t care they think about me It\'s alright, they\'ll get it one day  Pre-Chorus I love you, I\'ll follow you You are my, my life I will read my bible and pray I will follow you all day  Verse 2: I don\'t care what it costs anymore Cos\' you gave it all and I\'m following you I don\'t care what it takes anymore No matter what happens I\'m going your way  Chorus: All Day All Day now All Day  Bridge: Anyone around can see just how good you\'ve been to me For all my friends that don\'t know you I pray that you would save them too  Ahahaha rockyy!  The water baptism was cool...literally!!! xD Lots of my friends were there...bernard, yeow, em, eliza, bronson and abel. Some cell pple came down to watch too xP  I luv God I luv church I luv cell '),(21779,4103386,'07,August,2004','2004-08-07 00:00:00','   Yay no more Malaysia trip!!! Can celebrate national day in church! Long live Singapore! Today was s0 fun, i went for kayaking in the morning, church in the afternoon and L-cell in the night. In kayaking, we paddled under Merdeka Bridge to a deserted beach. Then, we beached our kayaks and took a swim out to sea(with our life jackets of course). The filthy water was so salty that everyone was choking and gagging on it... 0_o As usual, i found that i had fibreglass splinters from carrying the kayaks. The rough sand on the beach had worn off shards of fibreglass and they got embedded in my arms and hands as i carried the kayak back to base. They were stinging me badly...my arms were covered with red spots... At church, bryan gave me some scotch tape which i used to stick the fibreglass out from my arms(wow that hurt)...the wounds closed up amazingly fast ^^ After church we had L-cell. We walked to the botanic gardens to have dinner at food haven by the park. Guess who we met there? My mum!!! Argh ~.~ And we were sitting close together too...Mel tortured me...  After dinner and cell discussion, we proceeded to the park for a game. Since it was Charlotte\'s and Rachel\'s birthday, we decided to play a trick on them. We pretended to play a game of sardines, where one person had to go hide and others and to go find him. But in this case, we all had a pre-arranged meeting point. Soon, the two girls were wandering around alone in the dark park, looking very lost. Meeting up, Bro. Adrian decided to send Yeow and I to lure the girls to the \'ambush\' point. Things went so liquid smoothly! The girls were feeling skittish after wandering around the deserted park and wanted to follow me and Yeow around looking for Bro. Adrian...we led them to the pavillion and the others jumped out and yelled,  \"HAPPY BIRTHDAY!!!\" and strted singing the birthday song....it was so cool and funny :) L-cell is so fun :D '),(21780,4103386,'02,August,2004','2004-08-02 00:00:00','   Going to have to miss church this saturday because of my NCC Melecca trip...what a lame way to spend my national day holidays : '),(21781,4103386,'01,August,2004','2004-08-01 00:00:00','   Sigh...end of weekends already! Havent done much work...NCC tml ahh!! Individual marksman training...hope i can get my marksman badge its so shiny and cool -drool- NCC is taking up so much of my time...wonder when i will get round to doing my cheena compo? Ive been procrastinating for almost a week now... Bah, bring back the saturdays : '),(21782,4103386,'01,August,2004','2004-08-01 00:00:00','   Haha, set this up cause everyone\'s doing it :P '),(21783,4213897,'20,August,2004','2004-08-20 00:00:00','   Hello there I,     it\'s friday again!!!   i get to take my daughter to work today (she\'s 12 & she can take care of herself while i\'m working...so, no problema...), also my boss is not in, the other big cheese is on vacation &amp; it\'s casual day too!  we get to hang out in the city after work ( which is an early day for me...i\'m off by 4pm) so what better way to spend this day but with my daughter?!  Moms gotta get a break too once in a while, of course Dads too should be allowed to take their kids (not toddlers of course or babies unless the company offers back-up daycare, which is great)...We just don\'t get enough time to spend w/ our kids when we work full-time in an ofc., the weekends are just too short!, & before we know it they\'re off to high school &amp; dating & have no more time to spend w/ their cornball parents...boo hoo....anyway, if i can find a home-based business or some part-time career...that would be ideal!     anway, my daughter is rushing me...not nice), just wanted to jot down some of my to-do today:          1) try to work till about 3:30 or 4pm...keep my girl busy, she\'s gotta get on this computer & play some games.     2) after work, go to Chi-town - my daughter wants to go get a hair-styling/trim of her long, long hair.     3)  whle in Chi-town - get some dimsum snacks & bubble tea maybe.  buy some goodies for the folks at home...     4) go home, & hang-out w/ my little boy & my hubby.          (& the start of the weekend)  hooray!          ok, that\'s it, gotta sign off...guess what song is playing right this very second....     my fave - JT\'s \"Fire & Rain\"  -      (? could the day be moreperfect than this???)     ....\"thought i\'d see U one more time again?  JT asks, & i agree...i know i will see U one more time again...at least before my life expires (wow, here\'s the drama queen again :)          till then,  au revoir mon amour, '),(21784,4213897,'18,August,2004','2004-08-18 00:00:00','   Good Morning I,     not much to say today. just been thinking that i should change my blog title...i mean if i were to write a book i wouldn\'t call it - \"& the journey continues...\", it\'s too cliche, over-used &amp; beaten down the road littered with wanna-ba-writers hung-over w/ their own self-worth & obsessive thoughts...oops, not referring to me of course. so, i was debating what is a perfect title for this journal>>>          1) my secret love affair (sounds great &amp; mysterious but corny at the same time...)     2) \"Truth is stranger than fiction\" - but this has been used a million times!     3) Who Am I? - is good, but too existentialist for those ordinary day--to-day musings...     4) \"raison-d\'etre\" - cool, but i\'m not even French!     5) voices in my head - this has a certain ring to it...i kinda like it, but now for the next step...          So, it\'s just one of those boring days, we all have lots of those...     not much to rant & rave about, &amp; too tired to really try...i\'ll just take it easy today.     too early in the morning to be tired already!     well, here\'s one good thing, this little song playing in my head this morning was:     \"Honey Pie\" by the Beatles ( & a cover done by Tuck &amp; Patti which is great too!)     ...i love this song, for those cloudy,gloomy days...it\'s sure to pick u up.     ...anyway, here goes: \"honey pie, u are driving me crazy, i\'m in love but i\'m lazy...     la, la,la, la ...la, honey pie, my position is tragic...i\'m in love but it\'s ....\"     ( i don\'t know all the lyrics...sorry).     but the melody & the rhythym sure is good...     so till then,       happy singing ,      now it\'s 5:10 pm (NY time)     i just changed my blog title to :  voices in my head....  ( not the most original i know, but   somehow this resonates more for me right now...i\'m sure i could change it again later on, during the many more blogs, months &amp; months later...) ohh, i forgot another good one, \"talking to myself\" or \"writing for herself\" (but that\'s not in 1st person, which I AM.) this is so great, i don\'t feel lonely & alone knowing so many people out there are also blogging their funny tragic life stories...     life is amazingly simple &amp; complex at the same time...     well till the next blog, SEE Y\'All. '),(21785,4213897,'17,August,2004','2004-08-17 00:00:00','   good morning I, sometimes i think only my kids \"get\" me. nobody else in the world can get me all the time. my 12 yr old girl knows exactly when i\'m getting grumpy before i even recognize it myself...(talk abt an early warning device!) she\'s great that way...she goes: \"Mom, why are u grumpy?\" & i yell back \"No, i\'m not!, i\'m just...Wow, ur right, i am a little pissed off...&amp; then i stop & think...\" so she\'s wonderful to have around. &amp; then my little one, has a way of melting my heart. never fails. whatever mood i\'m in, he heightens it into a positive note...every morning, before i leave for work i have to kiss him & greet him a beautiful morning even tho. he\'s still sleeping...just looking at his face makes my day. but on the weekends, forgedaboudit! he becomes some kind of a little dictator around the house...yelling orders like: \"I want cartoons!\", \" I want ma-mam &amp; i want ma-mam now!\", or \"Mommy, let\'s play!! U be this soldier, & i be that\"....\"or U be Rocky &amp; i be Billy! (from his Rescue Heroes)....& wakes up the whole house to serve him...&amp; now that he\'s potty trained, he goes to the toilet by himself like a big guy, my swashbuckling cowboy/soldier...I got him allright. i really should count my blessings...at least there\'s 2 people in the world who can totally take me for who i am...not even my husband can get me.   ...u know those moments when ur only trying to communicate something & all these other stuff come up, &amp; u just wanna give up & shut down...my husband doesn\'t get that all i\'m saying is:    \"can u pls. understand me thru all this confusion?\" instead of judging me...asking why, & bringing back what was said &amp; done before...saying stuff like, \"don\'t be starting something again...\" when i\'m not trying to start anything really, just reaching out...can U hear me?      that\'s why, God blessed me with these two very intuitive & sensitive kids who know me before i even say anything...they GOT what it is i\'m trying to say or make happen, &amp; when something is not happening, they also know there is something missing...& it melts my heart every time one of them asks me: \"why are u mad Mommy?\", &amp; both of us get stopped on our one sided-track ( a train wreck waiting to happen) & we go, \"no baby, Mom &amp; Daddy are just talking...\" or some lame excuse... \"trying to have a conversation...\"  & we become quiet, lost in our own train of thoughts.  that kinda summarizes our 5-year marriage...still trying to have a conversation...a real conversation.   ...hmm, what brought this up? i don\'t know, just missing my kids during some free time at work.  i\'m so pathetic!      but hey, i don\'t want to dampen anybody\'s day that\'s turning out to be sunny &amp; gorgeous after all...despite the weather forecast,   so, \"keeping on the sunny side\"...   till then, '),(21786,4213897,'16,August,2004','2004-08-16 00:00:00','   it\'s too early in the week to be daydreaming i know, i usually reserve these for night time solitude...      but, like i always say, what else can i do? (it beats sitting here & staring at a blank screen w/ the blinking cursor... ) what do i daydream the most about?....Hmmm, that\'s easy... ...i always imagine that \"ideal love\", that \"ideal man\" that i\'ve never met, or maybe... bumped into once in my life...but just don\'t have the courage to rearrange my whole life for...to see, if perhaps, he truly is the One...if there ever is a One?... well, not to be cynical, i do believe there is a One, a soulmate, but that there are also \"soul-mates\" in one\'s lifetime...your best friends, your EX-es, your kids, your family, your soul-group...&amp; depending on the Karma that you need to relive and reinvent...you either are blessed to be with that soulmate or doomed to walk this earth like a heart without a proper home...     ... i know, i\'m really vague when it comes to matters of the heart...that\'s my weakness, my \"achilles\' heel\", that\'s where i go giberrish... did u see that movie \"Troy\" ( i mean, it\'s hollywoodized of course) but, the most touching scene for me was, the last moments of Achilles when he realized his true love, & ready to defend his loved one to the death, ready to change into a less bitter man...but as fate would have it, the prince Paris shot an arrow to his heel not knowing he was on their side... that became his downfall...the most celebrated warrior of that time... It\'s a tragedy of course, like most Greek myths...but it makes me think... what if?     what if, by my own undoing, because i am so jaded about relationships...&amp; because i still don\'t know what that \"ideal relationship\" is...ok, so there\'s no ideal relationship, u gotta keep working on it...well,     what if there is one...where it is less work because you are the perfect match, like peas in a pod...like the Yin to your Yang & vice versa... so, here i go believing in my heart that there is a possibility of finding that true love that is complete...without too much trying...it\'s just there... it just IS, and how do U know for sure?      for no reason but, that u feel it most when you\'re together...&amp; even when apart...it\'s just there....       (sigh) but like the jaded Achilles, i might be walking around, one of the many wounded (by war of the hearts)...like a lost soul searching for its harbor...seeing a vague vision of a lighthouse from a distance, far-away from her reality...but visible through a fog...what if? that light is what\'s real, & this reality that can crush the spirit from life sometimes, is what is not...because it is something that i have fabricated from all the mis-steps, like a mistake built on top of a mistake, to cover-up the fact that i might have made a bigger mistake?! oh God, how does one get out of that one?! (read to the tune from Rolling Stones \"Paint It Black\")...          trying to make right what\'s wrong... or trying to right a wrong from a negative place??...it only works in Mathematics right?... 2 negatives equals a positive?...but 2 positives is even better!... i hate dichotomies!      it doesn\'t have to be so black &amp; white...      well, that\'s the sort of daydreaming that i find myself getting lost in once in a while...not too often of course...or i will be walking towards the deep end again my friend...&amp; no, i\'m not gonna go there...          till then... still dreaming, '),(21787,4213897,'13,August,2004','2004-08-13 00:00:00','   good morning I, it\'s rainy again in NYC, i was falling asleep in my commute...nice relaxing ferry ride is always better than the packed-like-sardines-subway. it just takes a little more time...time enough to read a book, read ur newspaper, or as for me this morning, just thinking of nothing (for a change) & just dozing off to the sight of the waves. ...someone once told me that he always pictures me close to the water, maybe i look like a mermaid? (or perhaps sing like one, i wish?!) ...well i am from the islands, my earliest memories of weekends with my family were always spent picnicking on sandy beaches...i\'m sure if i had enough money, i wud live by the water...a little beach cottage &amp; i wud be really happy :)  well, how was my weekend? it was good.  i did a lot of my things-to-do, shopped for back-to-school clothes & supplies for my daughter, went to a children\'s party on sunday (lots of family &amp; food), did my laundry, cleaned around the house a little, watched some Olympic games, what else...cooked a little, that\'s about it... & now it\'s back to monday, rainy from the remnants of Hurrican Charley &amp; Bonnie from the South...  i think i\'ll take it slow today...what else can i do? it\'s really great when i can have lots of time to think, & not stress out. that\'s why i \"opted out\" from a corporate to a non-profit job i guess. my real reason was just to spend more time &amp; energy with what\'s important to me more than my job - my Self, my family... & around that time i got pregnant with my second child &amp; got married again- all when i was just starting out here on this job...also working here makes me feel good - the knowing that i can contribute in my own small way to making a difference in people\'s lives...those in need.  but my next goal is to make a more direct connection to making a difference - a career where i can really see & experience firsthand how i can help people, i wanna focus more on children...cause when i\'m around kids, my own &amp; other little munchkins, i get so much JOY & working with them will not feel like work at all.  So, my goal is to change careers, maybe become an early education teacher, or special ed. (which i hear is very challenging) or maybe become a school counselor...i have lots of experience in dealing with troubled kids - starting w/ my self, then my hubby & all my other Ex-es! so, i\'ll take it step by step, &amp; hopefully one day, i will be writing here &amp; logging in about my days in school...  for now, here in the school of life... one baby step at a time, '),(21788,4213897,'13,August,2004','2004-08-13 00:00:00','   good morning I, good morning All...after i checked my work emails this is the first thing i opened...i think this could definitely grow on me like a bad habit. (not that my boss will mind, at least it makes it seem like i\'m so busy & serious on some project)...anyway, today i felt like treating myself so i got my favorite starbucks coffee grande w/ soy &amp; the marble loaf! Hmm, & as i\'m enjoying it, i meditate on what\'s new today...   well, aside from noting that it is friday - whoopee!, the 13th! no big deal, i don\'t believe in superstitions...the conversations i had w/ some regular folks on the bus &amp; ferry seemed to revolve around the next President (hopefully!) John Kerry...also where we\'ll be during the coming dreaded Republican convention here in our city...why of all places did they have to pick NY?! i mean haven\'t we gone thru enough already?! after 9-11, the deadly SI Ferry crash, the sudden unemployments, the Bloombergization, FOX 5 biased news (thank God i hardly watch that news, i\'m an NBC fan)... It\'s no wonder why I along w/ a lot of other working parents i get a chance to speak w/ are considering packing our bags & taking the U-Haul to search for a new life...a better life from here? who knows. only time will tell. my brother lives down south (of Georgia) &amp; so is my hubby\'s sister\'s family so...at least there\'s equal representation from both families there. If push comes to shove, there\'s always some place to run to.   but as i was saying, i\'m glad that there is a renewed interest in this voting process...but we still talk abt. the bitterness after all that confusion & whatever-happened-to-the-Florida-votes?...maybe people have grown cynical? ...Maybe. but after 4 years of nothing &amp; this stupid purported Iraq war!, maybe people are at the point of vomitting...enough is enough.    i know they\'re talking abt the politics of \"hope\" but if that\'s the only thing left to keep working class folks such as me going, feeling like maybe a change is gonna come...well then i am all for HOPE...   i always remember where hope came from, especially for me as a kid growing up w/ a lot of \"family drama\", bitter images of 2 loving parents going thru their own crises, staging their tragic-comedies not realizing the little ones ( the formative not so little teeny sons & daughter) had to sit at home &amp; witness this passion play...while at the same time trying to stay sane & make the grade in school...well, u know what i mean....   ...I always turned to that greek mythology about Pandora\'s box, where all sorts of evils &amp; monsters have been opened & released to the world....yikes, kinda reminds me of where WE are right now in our times...look at all the funny evil stuff shown in the movies...(i just can\'t be bothered watching horror flicks anymore...why raise my adrenaline for that??) ...anyway, what remained in that box, waiting to come out &amp; redeem what\'s lost, what\'s forgotten...the last thing that remained in that box, the only thing that could give back Love & Peace to people is HOPE...   so, whatever they wanna call it, Politics of Hope, \"Help is on the Way\"... well i\'d rather believe in something good...that there is still some inherent goodness in all of us, &amp; that we will go out there to cast our vote for Hope - John & John, isn\'t that something? it\'s almost a sign...they have got to win this time. I don\'t care anymore whatever vile (which when scrambled up - comes out w/ the word evil)...stuff i hear in this coming republican convention...i just hope it\'s not gonna be violent around here...hey this is my city too, u know, i live &amp; work here....what about us little folks?! pls. don\'t cause more trouble than what we\'ve been through...we have families to support too u know.   anyway, there was even 1 older guy on the bus, who listened to the Democratic convention &amp; liked what he heard....who used to be a republican, w/ a son in the Navy or something like that...who thinks that this time he will swing more towards becoming a democrat...Well, that is hopeful.   ...maybe a change is brewing, a good brew just like my coffee...hmmm,   till then, happy trails,   :) '),(21789,4213897,'12,August,2004','2004-08-12 00:00:00','   i guess i just have too much time on my hands right now...my job is kinda seasonally busy...& it seems we\'re getting more &amp; more into the not-so-busy seasons...yeah, i know it\'s gonna be my problem soon, but i already started to cover my bases on that one...i\'ve started sending out my resume, posting online at all the job sites, Monster, careerbuilder, yahoo hotjobs, ajc.com, co. sites u name it... i do get tired trying to make 1 cover ltr seem better than the next one...& my work experience? my 1-page resume that\'s supposed to be able to tell a future employer just how great a worker i am, &amp; just how lucky they will be to hire someone like me...(who mtm, loves to float around w/ her head in the air, writing journals saved on my c:drive, now this...the freedom of blogging has found me!...oh boy, wait until u hear all i could write about...my life, people\'s lives, my fantasies, & the realities, the little cubes we revolve in just gets too cramped for me...i can\'t even express myself anymore...i meant to say, it\'s too damned cramped for me! i probably just need this space &amp; VOILA, here it is. Thanks to the founders of this blogging revolution... )      well, where was i? the questions why...  why am i blogging now...aside from having too much time, i also got too much energy in my brains that don\'t get much practice. i suppose, reading books in my morning & afternoon commute, my daily relating with my family, the household chores, the chats with friends, the boxes of \"things to do\", what else? the music that\'s been put on the shelf, the website that is somewhere sitting in my c:drive with tons of cool stuff...waiting to be ...what? waiting to be worked on?....no, more like ...waiting to be recovered, discovered, whatever...since i\'ve recovered from my \"dark\" days...i\'ve been taking my time...taking long deep breaths instead of hurried and stressed from too much energy that needs to come out...cause came out it did... out in the deep end where nobody dare go...Only the brave can come back from where my mind have been traipsing...travelling i call it... my mini vacations...but that\'s in the past. anyway, so why again? because, i can...right now, right here. &amp; if for some paranoic reasons (of mine)...i won\'t be able to anymore (that\'s in the future)     ... Alas (with heavy drama)...i\'ll probably turn to reading Shakespeare and sporting black reading glasses, become a teacher\'s assistant or a librarian or something very obscurely safe like that...I love safety, i love anonymity - that\'s why i\'m here (how ironic!)... But yes, i do love anonymity...even with all my talents that were waiting to explode, & sometimes they did...like little sparks, or shooting stars...i still love to be able to just BE... just be with mySelf, with all my contradictions...i don\'t anymore try to find reasons for them cause that only exhausted me &amp; everyone trying to understand or the better word is...trying to put me in a box... at least here, i\'m in a limitless box, within a box....so, i\'m almost like, i\'m imagining being under beautiful crystal clear blue waters, my eyes open & seeing everything there is to appreciate &amp; see...but really, i\'m in my cube...like a mini-aquarium...I should\'ve named myself after some favorite fish...but i didn\'t think of that before....hmm, cause maybe i don\'t know the names of my favorite fishes (they\'re tropical fishes, i also love goldfishes, someday i will have my own pond with a Japanese garden)...that\'s my dream...      for now, i better sign off...i still need my paycheck to paycheck, until the next job/career comes along...i know, it won\'t just come along...of course i\'m searching but i just need a little creative mindbreak right now... my thanks to the bloggers.com\' founding fathers & mothers again... what else wud i have been doing if not this right now?.... (probably getting bored, checking more job-searching sites, how much more can i search, before i have to start paying some Agent to find me another job that i\'ll probably get tired of too)      well, maybe if i just sit here &amp; talk to myself...that \"dream job\" will come to me...yeah right.  maybe it will come in a dream! ...maybe, if \"Muhammad can\'t go to the mountain, maybe the mountain will come to Muhammad\"...i\'ve always believed in that...oops, that should be the other way around isnt it?     \"...if the mountain doesn\'t come to Muhammad, Muhammad will go to the mountain!\"...right on. happy trails, till then, still searching, '),(21790,4213897,'12,August,2004','2004-08-12 00:00:00','   i feel so lucky to have found this \"free\" and \"free-ing\" space...by some stroke of luck my friend sent me a link to a website \"HotorNot\" so i can check out her new guy & i checked it out &amp; found it really stupid, but funny, serves its purpose i guess...but at least it led me to this blogger site & now i\'m happy. i got enough time on my hands at the moment just to think and blog...maybe i should just make this totally private & nobody can read my mind, but i figured what the heck i am ms. anonymous anyway...i\'m afraid my husband will find out about my thoughts &amp; discover i\'m becoming more of a stranger to him, yet we live together under the same roof deciding it\'s best for our kid, doing our best to stay together for our boy - he\'s so wonderful and kids are so fragile that it\'s difficult for me to make that move again, but, trying to stay in love? or is there still enough love to see us through???...i\'ve been through it once before in this same lifetime...going thru another divorce just seems so daunting, like another mountain to climb...altho. i was a mountaineer back in school, that was a longg time ago...now i\'m full of insecurities. I\'m afraid that my actions of wanting my freedom again, will create a bad domino effect on my kids...what if, & all the other what ifs  come up, &  i hear them constantly... the noise around my home...when i\'m just trying to build the perfect home for me, the perfect space & room to grow with my kids... i\'ve made my mistakes...but who hasn\'t? so what if he finds out about my new \"escape\", my blogs, then what the heck, maybe it\'s about time he meets the \"real\" me, not the one he wants me to be, his wife...the stranger. I wish Life was much simpler than what i\'ve created...&amp; i wonder, why do we make the choices we made despite the warning voices in our heads? the nagging doubts...when will there ever be no doubts? That\'s what i\'m searching for... till then, '),(21791,3224168,'28,April,2004','2004-04-28 00:00:00','   Ever since the day  Laura  came on my course...I knew there was something that I didn\'t like about her. She just walked in that room thinking she was so cool, like she was above everyone else. She thinks getting drunk is the only worth while thing to do on earth. I take it she\'s never experienced much in life...and she hates me for some reason.  The thing about Laura is she\'s violated just about every rule in the \"rules\" book. Yet she still is on this course. She turns up at course stoned...does the teacher not notice? It\'s very obvious to me, I can smell marijuana from a mile away. One thing I really hate about Laura is how she picks on this one particular girl in my class. This girl has done nothing to her! Also...Laura\'s always slutting up to the guys. She does the stupidest(is that even a word?) things, like pretends to hump Gary (a dude in my class) from behind. She already has a boyfriend but of course she doesn\'t care. Oh and...Laura is the worst distraction ever! I have ADD(Attention Deficit Disorder) and believe me, Laura does not help it at all, she\'s worse than my 10 year old brother.  You know what? I hope Laura gets kicked out of course (if you\'re under the influence of drugs, that\'s instant dismissal), and I hope my  friend  (supposedly) Tara gets kicked out too. Tara\'s just not a nice person anymore, now that she\'s been hanging out with Laura. She\'s been leaving me out majorly! I really want to get back into contact with my friends, I haven\'t talked to them in ages! I used to talk to Rose and Raph all the time...but now I\'ve just lost contact. It really sucks. They\'re WAY better people than Laura and Tara.   '),(21792,3224168,'27,May,2004','2004-05-27 00:00:00','   I found a good description for Laura \"obnoxiously loud\".   We were trying to watch a video last week that we rented for course. Through the whole movie she TALKED really loud, started stupid conversations with people and kept walking around the room. She can\'t sit still for one hour, even someone with ADHD is more behaved during a movie. I can barely concentrate in class with her around. Sometimes she randomly gets up and dances to stupid songs on the radio, she flirts with boys during class time (although she already has a boyfriend), passes notes with her \"friend\" (well she was my friend...but now she ditched me for Laura) and says really nasty things about my new friends.  She is so horrible. She makes everyone feel like crap about themselves. I really hope she gets  kicked out of course. I heard she has 2 warnings already (once you get three you get kicked out for good). Actually, I hope alot of people in my class get kicked out. Yesterday only four people in class weren\'t stoned (including me), oh and the teacher wasn\'t stoned. The other five, were all stoned (not a very big class).  I wish my teacher would notice already! These stupid stoners are nothing but a distraction and they aren\'t following the rules! Marijuana is ILLEGAL and it\'s against the course laws. If you come to course under the influence of drugs/alcohol, you get instant dissmissal.  So why haven\'t these juvenile obnoxious posers been extradited!!??   '),(21793,3614153,'29,June,2004','2004-06-29 00:00:00','   Getting used to working in hair profile salon again.....yeah not much time to think so much already....girls....got or dun have....dun really matters to me now....just wanna concentrate on my career once again.... '),(21794,3614153,'26,June,2004','2004-06-26 00:00:00','   after losing 5 kg in 5 days.....i decided to work....be it for her or for myself....got career....no need fear....after many days and night of suffering...tears have dried...friends consoled...especially desmond and francis....time to pick myself up..the recovering....i m really grateful to all those that helped me up.....finally 1 thing come to my mind......girls r not everything....put in too much = hurt yourself....treasure your friends.... '),(21795,3614153,'13,June,2004','2004-06-13 00:00:00','   this song POWER! by zhang jing xuan- duan dian   é™é™åœ°é™ªä½ èµ°äº†å¥½è¿œå¥½è¿œ  è¿žçœ¼ç›çº¢äº†éƒ½æ²¡æœ‰å‘çŽ°  å¬ç€ä½ è¯´ä½ çŽ°åœ¨çš„æ”¹å˜  çœ‹ç€æˆ‘ä¾ç„¶æœ€çˆ±ä½ çš„ç¬‘è„¸  è¿žè¿™æ¡æ—§è·¯ä¾ç„¶æ²¡æœ‰æ”¹å˜  ä»¥å¾€çš„æ¯æ¬¡è·¯è¿‡éƒ½æ˜¯æ™´å¤©  æƒ³èµ·æˆ‘ä»¬æœ‰è¿‡çš„ä»Žå‰  æ³ªæ°´å°±ä¸€ç‚¹ä¸€ç‚¹å¼€å§‹è”“å»¶    æˆ‘è½¬è¿‡æˆ‘çš„è„¸ã€€ä¸è®©ä½ çœ‹è§  æ·±è—çš„æš—æ¶Œå·²ç»è¶Šæ¥è¶Šæ˜Žæ˜¾  è¿‡å®Œäº†ä»Šå¤©ã€€å°±ä¸è¦å†è§é¢  æˆ‘å®³æ€•æ¯å¤©é†’æ¥æƒ³ä½ å¥½å‡ é    æˆ‘å»è¿‡ä½ çš„è„¸ã€€ä½ åŒæ‰‹æ›¾åœ¨æˆ‘çš„åŒè‚©  æ„Ÿè§‰æœ‰é‚£ä¹ˆç”œã€€æˆ‘é‚£ä¹ˆä¾æ‹  æ¯å½“æˆ‘é—­ä¸Šçœ¼ã€€æˆ‘æ€»æ˜¯å¯ä»¥çœ‹è§  å¤±ä¿¡çš„è¯ºè¨€å…¨éƒ¨éƒ½ä¼šå®žçŽ°  æˆ‘å»è¿‡ä½ çš„è„¸ã€€ä½ å·²ç»ä¸åœ¨æˆ‘çš„èº«è¾¹  è™½ç„¶ä½ ä¸åœ¨æˆ‘çš„èº«è¾¹  æˆ‘è¿˜æ˜¯ç¥ç¦ä½ è¿‡å¾—å¥½ä¸€ç‚¹  çŸ­å¼€çš„æ„Ÿæƒ…çº¿ã€€æˆ‘ä¸è¦åšæ–­ç‚¹  åªæƒ³åœ¨ç¡å‰ã€€å†å¬è§ä½ çš„ã€€èœœè¯­ç”œè¨€ '),(21796,3614153,'12,June,2004','2004-06-12 00:00:00','   Itâ€™s going to  be a long story...  Once upon a time, a hairstylist got to know a lady who is sweet and pretty. Sadly she was married to a man that has the occupation of a ship broker, a profession that requires long hours and endless night of entertaining of clients,  that includes bringing clients for sexual activities.  Despite the nocturnal lifestyle, the lady remains true, has very strong faith and trust in her husband.  The hairstylist told her, Wow ... if his girlfriend is half as nice as her, he would have melted. As she told the stylist what she did for her husband, the stylist knows by each word she said this kind of lady is rare, to the brink of extinctionâ€¦  A couple of months passed by and the lady (referred to as Ling from now onwards) went for her next hair appointment. She was slimmer then before. When the stylist asked her how is her husband and if things are going smoothly for them during the past months. Ling replied in a mild tone that her husband and her are no longer an item anymore... The hairstylist was shocked into silence. Oh my gosh, how could that have happened when they were such a happy couple... Like most stylists, he tried to console the poor damsel. He can see through her facade of putting up a strong front as he can feel the pain resonating from within her.  The stylist tried to dress her hair to the best that of his abilities, hoping that it would make Ling feel better. Ling find that stylist is quite caring and asked for the stylistâ€™s number. The stylist obliged as he thought maybe she just need someone to pour her woes to.  They started to go out more often for coffee, dinner just like what courting couples did. Despite the age difference of seven years old (Ling was 30 and the stylist, 23), they are in love.  As Ling is an air stewardess (kong jie), her stay in Singapore is not regular. The stylist is always with her whenever she needs him. They are like a happily married couple. Ling showered him with love like no other. The stylist feels that he is the happiest man in the world. He feels like a man ...a husband. He never felt this way before in his life and he actually moved in to stay with her (her place with her ex husband) while secretly promising to himself that he would never let Ling be heartbroken again.  Ling is a very caring, faithful person, loving her man to the max. She told Stylist that she will never give others a chance if she got a boyfriend. She will never say no to her man. Stylist is so touched that he again secretly tell himself never ever be unfaithful to Ling...  Stylist was a notorious flirt. He had been called names like jerk, bastard etc. Imagine the kind of love that Ling showered on him to make him to decide on something like those. Stylist is cool and is the type of person that donâ€™t sweet talk a lot and say â€œI love u â€¦ blah blahâ€ because the stylist knew that those words would work magic when it comes out from him. He told himself that he will never leave her as long as she is the same, while taking into consideration her age and how she will look ten to twenty years down the road. The stylist knew this time he is committed to Ling so much that it scares him.  They are so in love, calling each other everyday without fail. The stylist would be always reachable as he has promised Ling that he wonâ€™t give her the insecure feeling that her husband used to give her. They were in each othersâ€™ minds and hearts all the time. The couple will spend all their time together whenever Ling is in Singapore. Cooking, eating, sleeping, bathing...in other words, they were always together.  Lingâ€™s family consist of her parents and two younger brother. Both brothers are older then the stylist. They were shocked and asked Ling if she is sure she want to go out with him as they feel that he is too young for her. These are the comments by the people that is around Ling.  Stylist know that the odds are stacked against him and he tried to make friends with everyone. He is so much in love with Ling and he tried his best to make friends with her friends and especially her family even though he knew that they viewed him with prejudiced eyes.   Around a year later, Ling moved back to her motherâ€™s place as she wanted to save some money without taking much of stylistâ€™s feelings into consideration. Anyway, stylist loved her and understand that if she donâ€™t rent out her condo, she will probably end up with nothing at the end of 5 years when she sells her condo.  A year and few months later,  her flight stewardess contract was up. Stylist thought that they are finally going to really spend some quality time together as stylist was alone on christmas and new yearâ€™s eve. Stylist was looking eagerly forward to it. Meanwhile, Ling continued her career doing car sales.  Things went smoothly until after a year and a half. A heated argument erupted between the two of them until it is so bad that Lingâ€™s mother stepped in and asked them whatâ€™s wrong. Stylist just told her in a respectful tone that they can solve their problem by themselves but Lingâ€™s mother persisted. Stylist lost his cool and replied in a firm tone again saying that they can solve their own problem.  Her mother began to make nasty remarks... really nasty things that would make a normal personâ€™s blood boil. Stuff like with your looks and qualifications, surely you can  find a better man then the Stylist...things like that. That was the last straw as Stylist had a really bad day as he was asked to leave from his part time car sales job. Stylist raised his voice, â€œAUNTIE!! ASK YOUR DAUGHTER IF I BULLIED HER OR NOT!â€. Her mother replied. â€œWHY ARE YOU SHOUTING!!â€ without realising  that she was the one that was talking loudly all the time. Stylist donâ€™t want to argue anymore because he knew he made a terrible mistake. He shouldnâ€™t have shouted back as he saw that Ling was on the brink of bursting into tears.She was torn between the man she loved and her mother whom she respected. Stylist was too lost to say anything more and left.  Things begin to take a turn for the worst. Suddenly, the whole world seems to be going against Stylist. Lingâ€™s brothers advised her to make use of this chance to end the relationship as they think that he is too young for her. They also said that he is rude and immature. How is he capable of taking care of her... stuff like that. Stylist thought that the brothers already knew that he serious about their sister and after a year and a half of trying to be friends with them. Stylist tried to be as helpful as he could to Lingâ€™s brothers as he knew that since he want to be with Ling, he got to make her family accept him. Stylist felt utterly betrayed as he already feel that he is part of their family.  Ling continued to receive lots of negative comments. Since the day of the heated argument, her family and friends think that Stylist is a rotten guy. I think you guys can imagine the scenario. Everyone is pressurising her to leave Stylist and this made their relationship turn sour. They are arguing almost everyday...  One fine day, Ling told Stylist why donâ€™t they go back to the days when they are just be friends. Stylist was shocked as he didnâ€™t see that coming. He asked her why she is doing this to him and what did he do wrong? Ling began to tell Stylist of how much she actually suffered for the past two years. It is really quite a lot so please, bear with me.  She said that Stylist complained about buying food for her! Because Stylist bought her food back from work and when he reached home he said to Ling that he has to walk quite a distance to buy her food and he is persipiring all over. Is it worth it?(Are you guys getting the message that Stylist is trying to tell Ling?)  Despite her age, Ling is still attractive. She donâ€™t look a day past twenty five years old, so naturally she got a lot of admirers going after her. A guy actually bought her beehoon soup and proceeded to separate the soup and beehoon. Knowing that Ling is a heavy sleeper that need lots of rest, he went to get a slow cooker and actually bought black chicken tonic from those chinese medical hall. Ling was touched by the things the guy did.  She told Stylist that she donâ€™t feel much love from him and she is always the one giving. She didnâ€™t remember that despite being tired after work, Stylist went out of his way to buy her noodles and cook dinner for her (it is all time, sweat and effort) if she want to compare. Guys... just spend some time to thinkâ€¦ Which is more time consuming? Cooking tonic or cooking dinner?  Ling is actually complaining now. Stylist was just so devastated and he knew that guy is going to be a threat.  She told Stylist that she felt that he is a good man... Someone that can actually love and take care of her better. Besides, he is more financially stable and mature. He told Ling of his plans of starting a family with her blah blah... GUYS,  WHEN SOMEONE YOU KNOW LESS THEN A MONTH TELL YOU I AM GOING MARRY YOU!!! HOW CAN SUCH A GUY BE TRUSTED?  Well,  guess that you guys know what is the motive of that guy already. The sad thing is, how can Ling forget about the things Stylist had done and sacrifice for her. How can a two year relationship compare with a less then one month thingy.  Stylist felt so lousy. He told Ling that he knows that she is not young anymore and from the start, she knows that he is young. He is twenty three when they started. Now he is twenty five. If age is a major problem, then they shouldnâ€™t have started. But Ling has changed. At the beginning of the relationship, Ling told Stylist that she will never give other guys a chance as she is faithful. What she needs from a man is just sense of security (meaning faithful to her). But Ling had really changed, what she needs now is different. She needs someone that is ready to start a family and is mentally and financially ready.  Stylist have plans on marrying her and have a baby without telling her his plans. He thought that he will tell her when the time comes. It will be a wonderful surprise for he. Stylist is that kind of person that donâ€™t sweet talk much. He is a bit too frank, firm or cool maybe.  Stylist begged Ling for a chance. He told Ling that he donâ€™t mind her changing, but she must give him time to adjust to the changes. He will save up and marry her once she has settled everything with ex-husband. If she like sweet talking, he can do it better then other guys. Ling replied. â€œIsnâ€™t it too late? How long do you want me to wait for you? I am no longer young... Why do you want to torture yourself...? Stylist said, â€œBECAUSE I LOVE YOU!!! FOR YOU, I DONâ€™T MIND CHANGING... I JUST WANT TO BE LOVED BY YOU!!!  Stylist loved her so much. Imagine a notorious flirt transforming to a faithful man. What can make a flirt turn over a new leaf and settle down and have a family. After seeing how Stylist begged her, she canâ€™t bear to hurt him anymore.  Stylist is a totally changed person. He knew this time he needed to really wake up... for he knew that the person he loved has changed. He knew that if he is not going do something about it, Ling is going leave him for good. He promise Ling to save like a madman, talk in a more sweet manner... but deep down in Ling, the feeling is no longer that strong ...  Stylist began showering love, lots of love by doing things that he usually wonâ€™t bother to do. He actually takes the MRT, (he always go around in a cab) and ate the cheapest food he can find. But things are just different already. Stylist can sense it but he told himself to give her some time to recover. It was their second year anniversary and he made Ling a card saying nice sweet things but things have changed. Stylist didnâ€™t see what he wanted to see. The expression on her face is totally different from what he had seen during the first year anniversary.  The following morning, both got up and prepared to go to work. While on the way, Ling told Stylist, â€œLing, (she called Stylist Ling too as it is the short form of darling) the love that I had for you is no longer there... Letâ€™s just end it...  Stylist was devastated... He thought Ling was going to give him a chance, but how to give chance when the feeling was no longer there? Stylist knew he canâ€™t work for that day already. He need someone to talk to. In the end. Stylist went back to Lingâ€™s home and began pouring out his feelings to Lingâ€™s mother.  Eventually, Stylist went home.  Stylist began to cry...thinking what did he did wrong. All the while he is faithful and actually changed for the better... Stylist had already let the whole world know about him and Ling like his friends, ladyfriends and even ex girlfriends... Stylist found the pain in his heart unbearableâ€¦  Stylist called Ling and told her, â€œAlthough I am is feeling terrible, I donâ€™t blame you as it always take two hands to clap. But if you are going to leave me for that man, do think twice. If you can find happiness with another man, I will give her his blessings because I love you and want you to lead a happy life. I donâ€™t want you to get hurt as I know what a man wants from a lady. Judging by what the man has said and done, I know what is he after... I just hope that you wonâ€™t get cheated. I am still holding on to the hope that you would come back to me. I will always be there when you need a listening ear or advise. I will be there for you when you need someone to talk to. Darlingâ€¦ I love you so much...  Sorry guys ... I am not able to carry on anymore... I m worn out by writing this blog as my tears streamed down my face so many times... if you canâ€™t understand what I m saying....it is because you are just reading and not feeling and reading it seriously... '),(21797,3614153,'24,July,2004','2004-07-24 00:00:00','   Ok comfirm going to Shanghai on sept 20 liaoz...2~3 weeks ...hopefully i come back with really good and new good styles...arrhhhh really looking forward to it ...   nothing much happened recently.....all the same old stuffs happens whole day .....cut cut .....color color....by the time knock off..... go home....a bit boring heh?   ok new goal for this month ......save $$ and !!!!! QUIT SMOKING!!!!!! i hope i can.....or at least ....smoke lesser....   ya shit.......my right side wisdom tooth hurts again!!! damm..!!!! need to go through the same old operating shit again ....omg!!! arrrhhh another $5oo + off my medisave again......ill go extract them on early august.... '),(21798,3614153,'14,July,2004','2004-07-14 00:00:00','   just recived a call from shanghai .....saying the hair cutting course on august 30th is fully taken up ...sianz....actually i was thinking going with pinky and another friend....TMD!!! she said the next avalible course will be on september .....sianz ar!!!  saw a friend in mrt yesterday morning when going to work ....he has been doing hair for the pass 5 years.....and still an assitance....i asked y r u still wasting your time?? its time u really think ...u want to be a cutter or ...u wanna stay like this ...what u really want....if u want to be a cutter...go somewhere that has training  rather then somewhere that only make u work like an trainee rather then giving u chance to promote to stylist ....  after knocking off ....tommy(a hairstylist) zen and me went late dinner together....outside far east plaza(the viet resturant)..talked about lopts of things....old times...i asked ....wil u be ready to start a salon with me when the time is ripe? blah blah ....well its still my dream .... '),(21799,3614153,'11,July,2004','2004-07-11 00:00:00','   just came back watching spiderman 2....i think spider man 3 is coming....damm...they must be making tons of $$$$  work is busy as usual...saw patsy walking pass my salon few times liaoz....guess she likes the food in far east...dunno y i got this feeling she became taller....haha...  cutting hair and doing color became part of my everyday routine....suddenly felt like learning something new again....heard from kenn(boss) that pinky is going shanghai Vidal sasoon to learn classic and comtempary cutting...hmm...should i go with her?? 2 weeks course is gonna cost me $5+ K ......$$$$$$$$ $$$$$$  maybe i should go ....maybe a holiday and learning something new will be good.... '),(21800,3614153,'06,July,2004','2004-07-06 00:00:00','   ok 1 a bit tomboyist customer came in....with hmm...about shoulder length hair...ok typical singaporean student ...no makeup no styling..brought a friend along....dunno what she wants except something short ....i suggested how about something like mine? she said ok pls do it...cut cut ....after near finishing ...which was about 1 and a half hour...altering and craps...i said to myslef omg i thought we agreed on something like mine....she said ....i dunno what i want mah....just cut then i think think and alter lor....heng i not that busy or ill surly go crazy ..  ok she brought a friend along....her friend is rude ....especially 1 part ....she was trying to pluck out a grey hair on my client....then i just casually said erm ...theres another 1 here ....and 3 more here if u wanna pluck out more....*i was styling my customer hair that time and i dun like to be disturbed.. she said in a very firm tone....*hey u dun tell that to a girl ok!... i was thinking omg u r plucking grey hair in front of me and as a hairstylist blah blah ....u think i cant see the grey hair?some girls r just fucking sick ....well she prolly didnt like the idea i told her theres more grey hair...and plucking wont help...i felt like telling them *eh bitches.....its only $28 and u r asking for a $250 service? even i think my ex boss Kim Robinson wont entertain them much ...time is $$ .....firstly u burning lots of my time...which in the begining i dun mind....but u should be grateful...nvm ....u cant pls everybody ....some ppl r just sick... '),(21801,3614153,'05,July,2004','2004-07-05 00:00:00','   getting used to working in hair profile again....lots of my old customers came back to me ....so busy until sometimes cant even find ppl to shampoo for me even going to toilet or a smoke...  the argument between desmond and leekun just subsided...hope they argue again....we r good friends...nothing cant be solved....its whether u r willing to solve or not...  Leekun just passed me his geforce4 :D....although the graphic looks the same ...but it doesnt lag that much hur hur....  leekun going to ord soon...he is still considering to start building his career or going to university...desmond is still considering should he go HK for 3 years and he just went for a match making LOL aranged by her sister!! omg what era r we in huh? anyway i think that girl impressed him.....as for francis...glad for him to find a gf.....or maybe not hur hur ....he told me their relation is messy ....anyway glad he is a happy person again....hmmm zhiguang....still some time away from his ORD....hopefully he can endure for another few mths bah.........good luck in u guys career and love life  how fast time flies....took a good look at my dad and mum....they r aging ....mummy just had a operation....and shes recovering....sometimes i wonder when can i provide for her and let her enjoy some life....as for my dad ...he still drinks everyday ....but after all ...he is still my dad...seeing them growing older bit by bit ...makes my heart breaks...i told mummy....ill be a good boy...i hope to be able to provide for u soon.... be nice to your parents ....dun be rude...they r the ppl to turn to ...even the whole world abandons u .....after all ...they wont be with u forever....so be nice to them...respect them and let them enjoy some life ....coz they slog day and night to bring u up .... '),(21802,3614153,'15,August,2004','2004-08-15 00:00:00','   had been taking mrt since i started working full time again...... always wondering ....y dun those ppl wont offer seats to those who need them more....those old ppl standing looks so ke lian....maybe thats the reason i got tired legs.... '),(21803,3614153,'02,August,2004','2004-08-02 00:00:00','   æœ‰å¤šä¹…æ²¡è§ä½ ,ä»¥ä¸ºä½ åœ¨é‚£é‡Œ.åŽŸæ¥å°±ä½åœ¨æˆ‘å¿ƒé‡Œ,é™ªä¼´ç€æˆ‘çš„å‘¼å¸.æœ‰å¤šè¿œçš„è·ç¦»,ä»¥ä¸ºé—»ä¸åˆ°ä½ çš„æ°”æ¯,è°çŸ¥é“ä½ èƒŒå½±è¿™ä¹ˆé•¿,å›žå¤´å°±çœ‹åˆ°ä½ .è¿‡åŽ»è®©å®ƒè¿‡,æ¥ä¸åŠä»Žå¤´å–œæ¬¢ä½ ,ç™½äº‘ç¼ ç»•ç€è“å¤©,å¦‚æžœä¸èƒ½å¤Ÿæ°¸è¿œèµ°åœ¨ä¸€èµ·,ä¹Ÿè‡³å°‘ç»™æˆ‘ä»¬æ€€å¿µçš„å‹‡æ°”,æ‹¥æŠ±çš„æƒåŠ›,å¥½è®©ä½ æ˜Žç™½,æˆ‘å¿ƒåŠ¨çš„ç—•è¿¹.æ€»æ˜¯æƒ³å†è§ä½ ,è¿˜è¯•ç€æ‰“æŽ¢ä½ æ¶ˆæ¯,åŽŸæ¥ä½ å°±ä½åœ¨æˆ‘èº«ä½“å®ˆæŠ¤æˆ‘ä¸€ç”Ÿçš„å›žå¿† '),(21804,3660667,'29,June,2004','2004-06-29 00:00:00','   This is my eleventh blog post.  I had a very good day.  The day included eating, reading, watching TV, etc.  Also J.K Rowling lady announced the title of the sixth Harry Potter book today!  It\'s going to be called Harry Potter and the Half Blood Prince.  This is awesome news.  I\'m a pretty big fan of the series.    I liked the fifth book.  It was great!  Harry was starting to get bitchy at everyone!  His friends, godfathers...anyone really.  He was just so uptight and cranky.  It was like he was taking female hormones or something.  Best Harry book so far.  Hopefully the sixth book will continue with the Harry being a little bitch trend.  I predict that by book seven, Harry will be a miserable hermit living in the forbidden forest writing crazy rants and mailing them to people.  Like Dumbledore or Hermonie.  Maybe Ron.  But the movies suck big time.  They make me fall asleep.    A movie that won\'t make me fall asleep is Fahrenheit 9/11.  I haven\'t seen it yet but does it ever look great!  I\'ll probably see it in a couple days.  If it\'s playing in my area.  I\'m not even sure if it is.  I\'ve been reading Michael Moore\'s book \"Dude, Where\'s My Country?\" And It\'s ok.  Not great.  It\'s a little boring.  But other than that, it\'s just great! '),(21805,3660667,'28,June,2004','2004-06-28 00:00:00','   This is my tenth blog post.  I had a very good day!  I cleaned my room and found a lot fo things that I thought I had lost. One of the things was some glue.  Luckly, the cap was still on tight, and the glue still useable.  I put it away for later use.  After cleaning my room, I didn\'t do anything.  Then I watched a movie.  The first Godfather movie.  It sucked.  Just kidding!  It was awesome. '),(21806,3660667,'25,June,2004','2004-06-25 00:00:00','   This is my ninth blog post.  I got up very early today. I made myself cereal.  And went back to sleep.  After getting up later in the day, I ate some other things.  And read a book.  Then I did various things.  I had a good day. '),(21807,3660667,'24,June,2004','2004-06-24 00:00:00','   This is my eighth blog post.  I didn\'t do a whole lot today.  But it was still good. '),(21808,3660667,'23,June,2004','2004-06-23 00:00:00','   This is my seventh blog post.  I had a fantastic day!  I did a few things.  There were a little bit important.  And I also did other things.  http://www.local6.com/entertainment/3447172/detail.html  Also, Mary-Kate Olsen has an eating disorder and is in the crazy hospital for it.  I think it\'s kind of funny only because she sucks. '),(21809,3660667,'22,June,2004','2004-06-22 00:00:00','   This is my sixth Blog post.  I had a good day!  I did several very important things.  And also things that weren\'t very important.  But they were all very fun.  I won\'t list any of them. '),(21810,3660667,'21,June,2004','2004-06-21 00:00:00','   This is my fifth Blog post.  I had a very nice birthday.  I recieved many things that I wanted.  All of them very nice.  I won\'t list them all as I am a very busy person.  I\'ll just list my most favorite gift.  My best gift was the Game Boy Advance SP.  It\'s a blue one.  It looks nice, and is nice.  So far, being 16 isn\'t a whole lot different than it is to be 15.  I guess I have to wait a couple of months.  Or something.  Since I\'m not as busy as I thought I would be earlier, I\'ll type my second favorite gift.  It\'s the Gillette M^3 Power.  It\'s a razor thst vibrates.  It\'s awesome!  I\'m also still not busy so my third favorite gift is some DVDs. I have Pirates of the Caribbean, Forrest Gump, and Bob Dylan: Don\'t look Back.  I like them all equally.  My fourth favorite gift is this machine thta makes shredded ice and it comes with this juice that you put on the shredded ice to give it flavor. I think they call them \"snow-cones.\"  My fifth favorite gift is a Playstation 2 game which is Lord Of the Rings: The Return of The King.  My Sixth favorite gift is Cat\'s Cradle by Kurt Vonnegut.  It\'s a book.  That\'s pretty much what I got.  And it\'s good. '),(21811,3660667,'19,June,2004','2004-06-19 00:00:00','   This is my fourth Blog post.  Today was good.  I didn\'t do a whole lot.  That\'s good.  I watched some pretty bad TV.  There wasn\'t any good TV today.  What I also did was play video games.    Tomorrow is my birthday.  I am pleased.  I will get gifts, etc.  It will be a fun day. '),(21812,3660667,'18,June,2004','2004-06-18 00:00:00','   This is my third blog post.  I woke up early into the day.  This was odd because I usualy sleep late into the day.  I had a breakfast of French Toast.  French Toast is when you put the egg stuff on the bread, and then cook it on the stove.  That is what makes it good.    After eating the French Toast and all that, I went for a swim in my pool.  Only I don\'t have a pool so I used my hot tub.  This was just as good, and plus I had jets.  Lots of them.  So it\'s even better than a pool except that you can\'t swim in it. Kind of.  I also played video games on my Playstation 2.  Not Playstation 1, but Playstation 2.  I can hardly tell the difference.  If Sony spray painted all of the playstation 1s black, it would be just like a playstation 2.  Maybe they did.  I just don\'t get why they\'re so expensive.  Hell, I could probably build one myself. It wouldn\'t be that hard.  I could just wire a bunch of radios and CD players together and BAM--A playstation 2!  Those Sony suckers are seriously fucked up in the head if they think I\'m going to buy their little Playstation 3.  Fuck that.  I\'m going to build one.  I\'ll just use my old playstation 2 and spraypaint it whatever the next color will be.  And maybe add a few radios.  My birthday is on June 20th.  I\'ll be 16 years old.  I think 16 is a fantastic age!  Much better than 15. '),(21813,3660667,'17,June,2004','2004-06-17 00:00:00','   This is my second blog post.  I had a fantastic day! It was fun and productive.  I did many things, I slept very late into the day.  After waking up, I took a shower. After doing that I ate some cereal and watched TV.  There wasn\'t anything good on though.  Sometimes there\'s never anything good on the TV.  I put the blame on those damn reality shows.    I also had a good amount of fun on the internet.    Summer is the best time of the year.  With all of this free time, I can do various things in a day.  This leads to a more productive life.  Which is good for various reasons.  But sometimes I forget to brush my teeth everyday. '),(21814,3660667,'16,June,2004','2004-06-16 00:00:00','   This is my first blog post.  I like blogs.  They\'re a whole lot of fun!  So this blog will be no exception! Hahahaha!     I\'m currently on summer vacation.  This a great for various reasons.  The reasons are that I can sleep in later, watch more TV, Spend more time on the internet, and sleep more.  So far I\'ve accomplished all of the previously mentioned shit.    Today was pretty awesome.  A lot of great things happened! I\'ll put them in a list:  1.  http://www.cnn.com/2004/ALLPOLITICS/06/16/911.commission/         The 9/11 commission found that there is no \"No credible evidence\" of the link between Iraq and al Qaeda.  This was one of the reasons we want to war with Iraq.  The other reason was the WMDs. And they haven\'t been found and probably never will be found, as they don\'t really exist.  So we basically went to war for no good reason.    On Monday Cheney said this:   \"He [Saddam Hussein] had long established ties with al Qaida.\" -Dick Cheney = liar.  Cheney didn\'t even back this up with any evidence.  What sucks is that Bush is still going to get elected.  :(  2. http://www.degrassi.tv/index.jsp    I watched some great TV today.  There\'s this cool show called \"Degrassi\".  It\'s about these Canadian teens in Canada.  The show\'s one of those drama type shows where everyone is always messing up.  Sometimes they cry about it and/or get killed.  It\'s really great and I recommend it to anyone! It\'s on The N.  I think you have to have satellite or digital cable to get that channel.    3. http://www.amazon.com/exec/obidos/tg/detail/-/006052569X/qid=1087440087/sr=1-1/ref=sr_1_1/103-7492914-0123826?v=glance&s=books    I\'ve also read a good deal of this book I got called Bob Dylan: Behind the Shades.  It deals with the life of Bob Dylan.  For an asshole, he\'s had a really interesting life.  Someone should shoot him.  Or, if not shoot him, then maybe just make sure he never sings again.  I saw him a couple of nights ago in concert.  Man, what a loser!  Talk about a raspy voice...  4.  I slept a lot today.  I like that.    Summer vacation is awesome.  If I were always on summer vacation, I would be able to get a shit load of stuff done! I bet I could read at least twenty books, watch 1,000,000,000 hours of TV, and Sleep for 1,000,000,000 hours!    I would also like to stat how much I hate this fellow blogger: http://misterfahrenheit.blogspot.com    While he did spell the name correctly, he failed to mention where he got the name.  The name was created by Freddie Mercury of Queen, and is featured in their song \"Don\'t Stop Me Now.\" Also, he hasn\'t posted a blog since May 31st.  That sucks because I wanted that name and now I have to use Misterfahrenheit2.  Which is very lame.  He is a liar. '),(21815,3660667,'20,July,2004','2004-07-20 00:00:00','   This is my fifteenth blog post.     I had a very productive day.     I was reading the newspaper, and I saw an ad to apply in person at the Kyoto Japanese stake house. Needless to say, I headed over to apply. In person. Just like they asked. I applied for the position of \"bus boy\". I think it went over very well, and they told me they would call.     They won\'t.  I\'m sure it\'ll just be a repeat of the Best Buy incident.  But I could get hired.  I sure hope they hire me.       After doing that load of shit, I went to Blockbuster to rent the following movies:     Something\'s Gotta Give -- It was ok. I don\'t have much feelings for it. The story was kind of lame. I give it 3 out of 7 stars.     Along came Polly -- It was alright, I guess.  If you like that sort of thing.  4 out of 7 stars.     Masked and anonymous -- This movie rocked. Best movie I\'ve seen in awhile. It was partly written by Bob Dylan and Larry Charles. The movie is just so fucking great. I would recommend it to anyone who wants to see a fantastic movie. 7 out of 7 stars. '),(21816,3660667,'18,July,2004','2004-07-18 00:00:00','   This is my fourteenth blog post.      On the 15th my family and I left Virginia to travel to Wildwood, New Jersey. It was a long drive in the car, and I slept for most of the trip. When I was little, I spent many vacation with my family at the Jersey Shore. It\'s a very nice small town with fantastic sea food restaurants  and such. When we arrived, I found that little has changed. Some newer hotels have been put up, and a couple more condos, etc. But still pretty much the same.     Many of the hotels have a very 60s feel to it. With the colors and such, and just the way everything looks. It totally goes with that era. This is because it was that time era when the hotels were being built.     The reason my family and I were going there this summer, was to celebrate my Grandpa\'s 90th birthday. So I had lots of Aunts, Uncles, cousin\'s, 2nd cousins, 3rd cousins, etc. Lots of people I\'ve never even met before. It was similar to a family reunion.     It must be very interesting to be 90 years old. He\'s lived through a lot. Two world wars, and the great depression. He was around my age during the great depression. He\'s had a very full life. He\'s extremely religious for some reason. He almost entered the priesthood, like many of his brothers did but instead chose to be a teacher. He taught at a University of some kind. Very full life.     I spent much of my vacation there at the beach, in the pool, and talking with my aunts and uncles. I haven\'t seen them in nearly two years. They\'re pretty cool. One of them lives in New York City. I\'ve been to New York City many times because my Father used to commute to work there everyday, and sometimes I would go to work with him. There\'s many tall buildings, people, and homeless people. I\'ve never lived there, but I think it would be nice to live there.     I forgot to mention the boardwalk in Wildwood. It\'s all of these boards nailed together and you wall all over them. They also stick all of these game booths and stores and shit on the side of them. Lots of rides too! Like rollercoaster and about million other rides of some sort. The boardwalk is always a great place to be. '),(21817,3660667,'07,July,2004','2004-07-07 00:00:00','   This is my thirteenth blog post.  Best Buy sucks.  I went to my Best Buy job interview only to discover that it was a huge waste of my time.  I blame the manager who may have been a homosexual.  Not that there\'s anything wrong with that!  I\'m just saying...  Anyway, After he asked me the dumb questions and shit, he told me he wants experience for the job.  Which surprised me because it clearly says on my job application that I have no experience at all.  Why would he call me to the job interview if it says right on the application that I\'ve never had a job before?  Is it to waste my time?  Because I can\'t see any other reason.  Maybe it\'s because I wasn\'t cute enough for him.  I\'ve decided that because Best Buy wasted my time, I won\'t be shopping there anymore.  Those people suck.  I also sent an angry email to Best Buy explaining what had happened.    It\'s ok.  I didn\'t want to work there anyway.  I also put in an application for Walgreens.  The same thing will happen over and over again.  Work experience, etc.  I blame Bush.  If Kerry were president, I bet I would have a job of some sort right now.  Damn liberals! '),(21818,3660667,'06,July,2004','2004-07-06 00:00:00','   This is my twelfth blog post.  I had a very good Fourth of July.  The main reason is because My family and I set off those small firework things.  It was great!  Today was also very good, and included many events.  I have recently been looking for gainful employment.  I sent an application to Best Buy over the internet a few days ago, and today they called to set up an interview for tomorrow at 12.  This is good.  I like Best Buy and I hope they hire me.  I decided that there was no way I would stand a chance of being hired with hippy hair so I got it all cut off.  I guess I look more like a Best Buy employee rather than a fucking moron.  I think I\'ll get hired.  Why wouldn\'t they hire me?  Because I have no work experience?  Probably.  Work experience is the biggest crock of shit since Crystal Pepsi.    I\'m pretty nervous about the job interview.  I\'ve never been to one.  I think I figured out what they\'ll probably ask me and what I\'ll say.  I hope I don\'t screw it up.  I wouldn\'t mind an employee discount for Best Buy.  If there is one.  I think there is.  In other news, John Kerry finally chose his Vice President pick!  It\'s John Edwards.  I have no feelings for that at the moment. '),(21819,3660667,'15,August,2004','2004-08-15 00:00:00','   This is my seventeenth Blog post.  I have kick ass news! My parents bought a really awesome 51 inch Widescreen TV. It\'s really awesome. I can\'t wait to watch TV and movies on it. The old TV was good. But this one just shits all over it! It comes on Tuesday.  Also, I decided to try again and get a job. My mother is a medical transcriptionist at a hospital. Apparently they have some jobs available like cleaning work and kitchen stuff. Hopefully I\'ll get some fucking work. Getting a damn job is stressful. It really sucks. Also, I think, this shitty doughnut place is hiring. I don\'t like doughnuts that much.  Having a job is important. I really want one. But it\'s difficult to get one. I blame our capitalist society! With its work experience and economy that\'s kind of sucks but it\'s a little good.  Sometime this week I\'m going to take the god damn test to get my driver\'s permit thing. For the second time. I FAILED the first time...by default. I accidentally submitted the wrong answer that I wanted to submit. It sucked. I knew the answer, but I went too quickly and fucked up. It won\'t happen again. It will.  My school year starts soon.  I\'m not sure how I feel about this. '),(21820,3660667,'13,August,2004','2004-08-13 00:00:00','   This is my sixteenth Blog post.  Getting a job is tough work. Either they don\'t call, make a mistake in advertising for the job (Like not saying you need to be 18 even though you do, so you go but they tell you to leave), or they don\'t look at your application where they can very easily see that you have no experience but waste your time by calling you in for an interview anyway. This is complete bullshit.  Why the hell do places not call? They say they\'ll call. Can\'t they just call and say I suck at getting a job, and that they don\'t want me anywhere near them? Of course they could! But they don\'t! Because they\'re all assholes.  Fuck getting a job.  I quit.   I went to Blockbuster today.  I got two movies.  13 going on 30 -- This was an ok movie but it couldn\'t keep my interest very well. It\'s about this girl who wishes she was \"thirty, flirty, and thriving.\" She says it over and over again. Anyway, because of this god damn dust that her friend, the not-so-cool Matt, got for her.  It works!  Now she\'s a big time magazine editor. And thirty! Also flirty as she tries to date a 13-year-old kid! Whoa, pedophile! I thought the movie would go in a much more interesting direction had her friend not stopped her, and allowed the relationship to develop. Nope, instead she hooks up with her old time buddy, Matt. He\'s still not cool. :-(  Anyway, Matt is a very unlikeable character. He seems uncomfortable in almost any situation he\'s in! Fuck, man. This guy\'s weirded out by anything. He also likes Razzles. Razzles are candies that apparently make your tongue red. And bring old friends together.  For some reason she goes back to being 13. And is about to have sex with Matt, but then rips upt his girl\'s paper for no reason at all! What the fuck!?!?!? You god damn whore! Then she goes back to being thirty. But this time she\'s thriving, as she is married to Matt. And then they eat Razzles.  I really only got this movie because it had Jennifer Garner in it. She\'s pretty. But the movie lacks a logically sound plot, and therefore sucks. I give it 2 out of 7 stars.   The next movie I got was Kill Bill volume 2.  It\'s a very good movie.  A little better than the first.  7 out of 7 stars. '),(21821,3849830,'08,July,2004','2004-07-08 00:00:00','   fineally i got 2 play with christian. after like a month and 5 days or somtin like that. we play at his house first and they at mine.  then at 6:00 we went 2 the mall it was really fun but im way 2 tired 2 write about it c\'ya   '),(21822,3849830,'06,July,2004','2004-07-06 00:00:00','   these web sites r sooooooooooo cool if u want 2 make 1 like it just like it go 2 www.blogger.com and make 1 it is sooooooooooooooooo cool.   '),(21823,3849830,'05,July,2004','2004-07-05 00:00:00','   1 of the funnist films of the summa it is funny as hell(puss in boots)the cuitist cat eva   '),(21824,3849830,'05,July,2004','2004-07-05 00:00:00','   katie and jake r boy friend and girl friend... u could say that they best of friends and even more than that... they both are my best friends when they  r away from each other but when the r 2ether they only hang out with each other they dont hang out with any other of there close friends... for example when we all (chris, james ,me, jake katie, kimmy , amy and anna)went camping 2 lion contry safari,and before james, chris, and jake got there me, kimmy, katie, amy & anna where going 2 go 2 the park on our scooters, katie saw the ardells camper coming. she turned around with out telling us she rode her scooter as fast as she could.  she jump off the scooter falling in the road hurting her self.  we all ran down there 2 help her but half way there katie looked up and saw that the ardells car was coming closer.  she got ^ and ran 2 the car just 2 see jake.  she did not even tell us that she was ok.  katie practicly broke here hip, she scraped both elbow. she did all that just 2 see jake.  they still r 2 of my best friends but they have gone way 2 far.  also on that trip they both ditched me, chris, and james.  they only did things 2gether nothing with us, and we all where very made.   '),(21825,3849830,'05,July,2004','2004-07-05 00:00:00','   luke is jake the gay mofos little brother. him and james made a lanuage. A phil is a guy with 3 nipples and a phillie is a guy with 1 nipple and so on.  jake his older brother is going out with katie ross purrrrrrrrrrrrrrrrrrrrrrrrrrrrr.....   '),(21826,3849830,'05,July,2004','2004-07-05 00:00:00','   my friends r kinda weired.  First u got the neighborhood friends.  Jake, a good friend...(when he wants 2 be..J/K)lol... then theres jakes girl friend...Katie she has been 1 of my best friends 4 like 6 years ...then there is kimmy 1 of my best friends 4 6 years also... then there is james he says he is pimp and he dress\'s like it...lol Blake, he plays the drums and is like 2wice my size lol... then there is luke him and james have there own GAY language... 4 ex. a phill is a guy with 3 nipples or sothine like that... theres corey or should i say corney j/k  lol he is like 20 times my size and i all joke around and call him a big oaf... Then there r the school friends 2 start there is christian he is like my best friend he has red hair like me and was born 1 day an 6 min after me... then there kate and brie they r like giants compared 2 me lol they r like sisters but there nt and they r like my best friends also. well thats nt all of my friends but my hands r geting tryed... peace out   '),(21827,3849830,'05,July,2004','2004-07-05 00:00:00','   hey sup just tellin u guys 2 check out my big bros sit www.spitcherry.blogspot.com   '),(21828,3849830,'05,July,2004','2004-07-05 00:00:00','   What about ether the good or the sux.... but they r on a losing streak. they gave up there 1st place spot 2 the stupid phillies....hey i thought that we r suposed 2 be the world champs...well nt > now at least we will most likly make it back in2 1st place a gain.   '),(21829,3849830,'05,July,2004','2004-07-05 00:00:00','   about the worst skool eva. it suxs t looks like shit and most of all the teachers sux!  And miss costa !!!!!  plz she walk tike a duck and has a nose like a bird.   '),(21830,3849830,'05,July,2004','2004-07-05 00:00:00','   Hey this is my new site tell all ur peeps 2 come on~!!!!!!!!!                     Peace Out   '),(21831,3235391,'01,May,2004','2004-05-01 00:00:00','   Grr.....my landlords just wanna piss me off don\'t they? Well, my dad was okay today. I\'m not as mad with him as I was yesterday. But my mum...grr. All she wants to do is wallow in self pity and scream at me. She thinks that she\'s the only one with a troubled life. But because of her troubled life (and of course my fathers), I have a troubled life. See, my father was raised in a middle class family where he was tought that love just gets in the way of everything. So the only we he shows his emtion is by screaming them at me and my fellow tennent when it all just gets to hard to keep inside. This is a...lets just call it...yearly volcano. My mum came from a REALLY high class and rich family who spoiled her rotten. And now, they are both fighting and on the verge of a divorce (which I hope so badly for) and they\'re pulling me into it. Telling me to take sides. Pissing at each other behind they\'re back to me. I hate it. I hate them. I hate life.  The landlords want rent....taliho!  '),(21832,3235391,'30,April,2004','2004-04-30 00:00:00','   The landlords want rent....taliho!  '),(21833,3235391,'30,April,2004','2004-04-30 00:00:00','   Hehe. Thanks for helping me make this Sarah. This is awsome! But today wasn\'t. Grr... My landlords (parents) are trying to mold me into something I\'m not. They want me to be smart and perfect in math and science. They just don\'t get me. I don\'t think they want to. I think I scare them because I\'m different. I really want to get away from here. Far away. To a place where people who actually love me are. God I can\'t wait till I\'m older. I\'m gonna move outta here the second i turn 18. Oh and my public speaking. Grr.... I\'m just never good enough for my mum OR my teacher. it drives me insane. Why can\'t they just settle for what I can give them. I don\'t even want to do this. And dad. Man!!!!! He\'s not gonna eat for two days just because I couldn\'t control the nature of traffic and couldn\'t get home to finish my schoolwork. See if I care dad. You can just go melt in molten lava! Just to warn you. Since I seen really perky and possitive in real life, theres gonna be lots of bottled up emotions that I\'m gonna spill here. Lots of...mad stuff. blah. I dun care. The only person whos gonna have my link are Sarah and...me. hehe. Ta for now  ~Prakruti  '),(21834,3235391,'07,May,2004','2004-05-07 00:00:00','   I feel so left out. My mother is so serious when my bro was at school. But the second he stepped into this house it was \"Oh my little baby!...\". For once I want to feel loved. The only things my parents say to me is that I\'m a  complete disapointment. It makes me...sad, that I can\'t be the one, for once, to put a smile on my parents faces.  I had a nice buble bath and seriously considered drowning myself in it, just like Sarah. But once I read her post describing what she had did. I don\'t know what happened. I thought about it and figured, if I killed myself, then what will I miss? My first kiss, my sweet sixteen. I\'m guessing Sarah figured all this out (\'cept the kissing part)  after  she found she didn\'t have and bubbles. She has a tough life, I have to say, but if only she knew what I go through. My family is torn apart and it was all because I was born. We came to Canada and I apperantly changed.   I have my public speaking competetion tomorrow. I really love doing this. The rush you feel when everyone is looking at you, it is unbeleivable. And your thinking to yourself, \"they\'re all here to listen to me\". Last time I won, there was a smile on my parents face one day. The next day it was \"Prakruti, did you finish science?\". I hope I can find someone who really does appreciate me. Not because I\'m they\'re friend or they\'re my friend. Which is usually the case. I don\'t know, I think I\'m being to picky. Oh well. I have like this mind that is just way, way, way to passionate. Love is like the ONLY thing I can ever think of.  I want to get a good rest before tomorrow. I\'ll write and tell you what happens!  '),(21835,3235391,'03,May,2004','2004-05-03 00:00:00','   Hiho. Let me just start of wit this. MY MUM IS JUST...grr. She just wants me to be perfect. She even yells at me if I don\'t do things according to \"the plan\" she and father made for me when I was born...study, study, get good grades, don\'t have a life, get good grades, be a loner, get good grades, get into med school, get good grades, get married and.......get my kids to get good grades. They don\'t realise me as an individual just as a younger version of low life...nerd. Grr.  I have toastmasters tomorrow. I have to talk in front of people on my speech. Ah. the wonders of my mother. Why does she have to make me talk in public. I have good social skills! I have changed schools since the past 8 years, I know how to talk to people I don\'t know. Ah dash it all. I\'m gonna go practice.  '),(21836,3235391,'15,May,2004','2004-05-15 00:00:00','   I went over to Sarah\'s today. It was SO good to actually see her. Her sisters boyfriend picked me up in his car and my mum was like \"Oh my. I am NOT sending you with him!\" I was scared because well...it was scary.  We saw Peter Pan. Oh god I LUV Jason Isaacs. Witty, charming, who wouldn\'t like him? :D Sarah let me borrow like 5 million books (okay well five). But I let her borrow my bible so we\'re even.  She\'s leaving for Whitehorse on Monday. I feel so sorry for her cause its like, 2 weeks with mum for company. *Shudders at very though*. I wanted to sleepover but when I asked my dad he said I had enough fun and it was time for me to come home and study...sad excuse dad, sad excuse. You know what I found? When Sarah and I are like \"virtually\" talking on msn, we can talk about anything. But when we are actually around each other \"physically\" we can\'t talk about anything THAT depressing. Its kinda weird but i guess we\'re not really morbid people, or we both want to seem as the non-morbid type. Ah well. I\'m gonna go since nothing else of importance happened today. I might change my layout since I\'m just not feeling it. Well, I\'m kinda jelous of those two BABIES having their first kiss when I can\'t even get a guy to talk to me let alone kiss me. But I want my first kiss to be special with someone special. Specifically Ronil. I will write more on him later tonight since my wrist is in pain \'cause I\'m using the main comp and not the laptop since the laptop has a nicer keyboard. :D The landlords want rent....taliho!  '),(21837,3235391,'14,May,2004','2004-05-14 00:00:00','   As I did my \"thang\" on my audio blog yesterday. My landlords are fighting...again. I think it really got to my mum. This morning she wouldn\'t talk to me. And then later when I tried to make convo with her ( I asked where I should stick my TEMPORARY tatoo tat Sarah A. gave me for my b-day) and this was her exact words...\"PRAKRUTIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  I awakend the spirits in her.   I honostly don\'t want to really awaken the spirits in her. That\'s because she used to be abusive. Not like, i got black eyes and stuff...no, no. She gave us slaps when she was mad at me and my bro. Which was especially usual with me since I was like a minature rebel. Now I\'m just a...rebel. Anyways, I got really scared. In fact, I still am. And...Sarah please tell me your not going to the aquarium cause I NEED like seriously NEED to get outta this house before I get hurt. Or just really pissed. And plus I need to see your trampoline...:D. I\'ll write more later if there\'s something worth writing. Taliho!  '),(21838,3235391,'13,May,2004','2004-05-13 00:00:00','   urlLink'),(21839,3235391,'12,May,2004','2004-05-12 00:00:00','   urlLink'),(21840,3235391,'12,May,2004','2004-05-12 00:00:00','     '),(21841,3235391,'09,May,2004','2004-05-09 00:00:00','   urlLink'),(21842,3235391,'09,May,2004','2004-05-09 00:00:00','   Wow. God really does see me out of the corner of his eye, (sometimes). Here I\'ll explain it to you. You see, in my public speaking, there was a first, second, and third plase winner who got recocnized. The first and second place winners go to the final district competetion and if and of those two get into a freak car accident and dies, then third place winner goes in their place. Now I know your wondering which one died cause I\'m going to district baby! *Ahem* well back to what I was saying. Neither Zaarah nor Ray died, but I hoped dearly that they could\'ve plundered the depths of hell. No, no. You see what happened was, the VICEpresident of the optimist club saw me speak and found it , and I quote, \"utterly revloting that I got toped by a girl who couldn\'t even remember to say her name.\" Hehe. Zaarah is going DOWN.   So the district level competetion is in Seattle. I am going there on a plane with, Ray and Zaarah. They can snog their heads off while I actually put my time to some use and read my beautiful book which has now become my bible \"A great and terrible beauty.\" (More on this later). I am going to have a roommate and stay in a hotel. Of course my parents will be there but not all in my face going \"Prakruti do this!\" and \"Prakruti do that!\". Score! I\'ll show everyone that I DO have talent. There\'s going to be a banquet as well, like one of those semi formal banquets your brother goes to at New Years with his \"med school buddies\" and comes home drunk.   I have more good news...SARAH IS BACK!!!!!!!! I am just beyond happy. I read her blog. She seemed really happy to be back. I suppose being stuck in a car with your parents for 4 days straight gets to you. She gets me! She read my post and she understood what I meant. I thought that even she wouldn\'t get it! I feel so...I don\'t know. I feel happy. She understands that being needed is a good, nay, great feeling. It just makes you go \"Wow you like me...you really like me!\". Shes right we do have our own way of getting the message across without useing somewhat complete sentences.   Okay, now I shall tell you about my new bible. \"A great and terrbile beauty\". I know what everyone who reads my blog is thinking. What about Harry Potter? Well I like Harry Potter, yes, but I don\'t feel the pain of an enitre community depending on me to survive. In this book, the four main charecters make an oath. They say what their heart desires most. And all of them say what my heart desires.    ~*~  \"To find true love.\" \"To be beautiful.\" \"To be too powerful to ignore.\" To understand myself.\"  ~*~  So as I said, Ray and Zaarah can snog each other senseless on the plane while I connect with my fictional soul sisters. As long as they don\'t suggest a threesome. *Shudders at the very thought*.   I should really get back to my work. Not that I\'m working, just reading fanfic. hehe. Well, bye!  1:22 (later in the day)  Do you ever have that feeling where your being controlled by everyone around you? Like you don\'t have ANY say in anything you do. I really hate this feeling but apperantly I\'m going to have to live with it until I\'m like....married. I just wish Prince Charming would come and carry me away QUICKLY.   1:24   Do you also ever have that feeling that some people are your friends just cause your rich? I really hate bragging about myself but my family really is rich. And I don\'t know, I think Ashley just likes me because of that.   1:25  Do you feel that sometimes your being watch. I swear my house is like completely haunted! *looks around nervously).  1:26  Do you sometimes feel like you need to get away from your family even if just for a week. Just hav some time for yourself? *gives time for readers to think*. I do to....Sarah take me to Chritina Lake!!! (Is it Chritine or Chirtina....I confused.)  1:27  Do you think that sometimes you need to buy premeire ticktes to Harry Potter POA ahead of time. I shall go buy it...I\'m going with Sarah to see it right?  Alright I\'ll right more later!  '),(21843,3235391,'09,May,2004','2004-05-09 00:00:00','   I got third place. I suppose thats good. But I know I was the best there. I should have gotten first place. Instead, idiotic Zaarah got it. The judges picked her cause she was loud. And second place...Ray. I could just squeeze him sensless. Not that he has any sense in him that is. I wasted my time there when I could have spent those precious minutes with Ronil. He\'s falling for Jayna. I know it. I know it because life is just not going my way. Now, I\'m really gonna go hurt myself. Life has no meaning. Just a puppet for my parents and a fake happy smile for my friends. I\'m breaking, I\'m broken. No one needs me anyways. What good am I to the world? Just a couple of good answers to some disturbingly annoying questions. I want to run. This is why I like soccer so much. When I run, I feel freedom. No one controls me when I run. No one can control me after I die. I\'m scared of these thoughts. I don\'t really know what to think of them. No one understands me, so no one to talk to. Sarah has gone...somewhere. I\'m eagerly waiting for her to come. Check how well my spelling on here is and make me laugh away my misery. Thats what shes best at.  I have to go help mother. Bye.  '),(21844,3235391,'16,May,2004','2004-05-16 00:00:00','   MY DAD IS SETTING UP MY LIFE FOR ME AND I DON\'T EVEN KNOW WHAT HIS PLANS ARE!!! Its scary. He said I don\'t have enough time to think. Not that he\'s giving me time. And even if he did, he\'d merge and mold until he would be satisfied. Doctor scary thought but MED SCHOOL...even scarier thought.  I really want to be a special needs teacher you know, help other kids. But my dad said I can persue that \"rifraf\" as a hobby. Sometimes he just makes me so boiling mad.  I know his heart is in the right place but sometimes I wish he\'d be like a normal dad. The one who tought me how to play b-ball when I was going through my TomBoy stage. Instead he bought me pink frilly dresses with laces. And now all he talks about is \"the future\" and \"my role in it\". Its scary. I\'m gonna stop ditching on my father and turn to another big issue at hand...SARAH IS LEAVING ME!!! She\'s going to....a place...far away...VERY far away...TOMORROW. :\'(. It\'s evil really, the time I need her most she\'s going. The final round of Public Speaking is coming up and well....there\'s lots of stress in this kinda stuff. Its sad really. Truly it is. But ah well...I\'ll tell her when she comes back.  Thats all I can think of wirth writing right now. But if and when I think of something else I\'ll come and share it. TutTut and Fare-well!  '),(21845,3235391,'29,May,2004','2004-05-29 00:00:00','   Sorry I havn\'t updated in a whie. I\'ve been busy being lazy. I just came back from a veiwing of Shrek 2! It was simply first class. I went with Sarah....FUN! hehe...we\'re still quating the movie.  Well....my public speaking competetion moi no win...but I did meet a real good friend who lives in Oregan. Maya. She and I got on real well. My room mate Nikki was er....well erm....okay I\'ll be frank, she was a real bee-yatch. Ah yes and Ray, he\'s not so bad once you get up close to him and notice his cutness! hehe. Next week is Harry Potter! I\'m going with Sarah (big surprise there) and I gonna sleepover after. Our first sleepover since my b-day. hehe.  I\'m not sure what else to right as I am still busy being lazy so...Sarah live with this..\".\" Taliho!  P.S. I\'ll right more tomorrow....  '),(21846,3497723,'30,May,2004','2004-05-30 00:00:00','   This is my first attempt at keeping any sort of journal. I have never been very good at keeping any sort of record of my life. I have always had the fear that once an experience is put on paper, the brain no longer has a need to remember it and the memory is thus discarded, or rerouted for use in another way. Therefore, I am rather skeptical to put much of note down here too soon. Perhaps my mind ramblings will be just that, ramblings, nothing more, nothing less. Or, perhaps, I will learn that the more I jot down, the more I will be able to remember. The brain never ceases to amaze me. It remembers things long since forgotten, in a way...Perhaps that is where DeJaVue comes from. I am constantly in a state of rememberance, that is the way that I know I am on the right track in life. Whenever I don\'t get a memory of something that I am doing that I know I have never done before, I know that what I am doing is that which I am supossed to be doing. My life, otherwise, might get very complicated and then where would I be?  A friend from my childhood has had a very trying life and started blogging to sort things out. I am very proud of her for that. She is who made me decide to try it. I don\'t really have an outlet for my life, my husband has a hard time understanding my need for higher and higher education. That is not to say that he is not supportive or encouraging by any means, it is just that he felt a Bachelors degree was all that was really needed. I, on the other hand, feel that I need to get more. I am in the process of completing my BA in Molecular Biology and a minor in Chemistry, only to persue a DVM/PhD combo in the next 6yrs.  Perhaps I will obtain that goal, but for now I am just looking forward to finally obtaining my first degree...and it has only me since 1996...3 years sabatical for work and my son...etc, basically life in general... Well, that is it for my first blog, perhaps next time it will be more insightful.  '),(21847,3639495,'29,June,2004','2004-06-29 00:00:00','   Let\'s all pause for a second to give it up for the unsung heroes of the human race:   TRAFFIC COPS!  No, I\'m not talking about the beat cops who apparate in your rearview mirror out of thin air, Dumbledore style.  I\'m sure there\'s a time and a place for forcing people to obey basic speed law, but they never DO seem to be at the right time or place, do they?  Sure, when you\'re trying to slink off to school at 7:50, there\'s no shortage of black and whites, but when your car nearly gets tidily-winked off the freeway by a bass-thumping, low-riding, erector-set-detailed rice rocket.   Wow.  How old was that sentence?  Nevertheless, it is therapeutic to wave your fist in the air and yell \"young whippersnappers!\" in digi-rant form.   No, I mean the \"standing in the middle of the intersection with the white cotton gloves and scary polyester pants\" traffic cops.  An integral cog in the machinery of modern society that I\'ve always taken for granted...  Until last Saturday.  Dateline: Devore, California.  A little patch of Appalachia nestled amid the junction of the 15 and 215 freeways.  Nothing but dirt and monster trucks for miles around.  And then, rising like an oasis of pure rock & roll in the wilderness, it stands: the 40,000+ seat Hyundai Pavilion at Glen Helen.  My sister and I headed out to Glen Helen this weekend to see the No Doubt/Blink-182 concert--a double bill about which Jennifer wryly observed \"we would have killed to see that show... seven  years ago!\"  It DID have a major 1997 vibe up in here.  Nonetheless, it was a rockin\' show, and we had a great time.  It turns out music was pretty strong in the last millennium.    So: encore, No Doubt is coming out for their final bows, and Jen and I decide to hightail it out of there to get back to the car as quickly as possible.  Unfortunately, about a thousand other people had the same idea, but that\'s okay: that\'s still a head start over the other 39, right?  We\'re hauling as fast as possible back to a parking lot that is somehow exponentially farther from the venue than it was when we left it earlier that evening, trying to save precious seconds by cutting in front of slow-moving couples, bobbing and weaving through people-traffic like the aforementioned rice rockets that irritate me so, but the possibility of making it onto the freeway two seconds earlier than the rest o\' the dirt people somehow excuses our inconsiderate crowd wading behavior.  Finally, success!  Back at the car, squeezing into the tiny gap formed when we open the doors without scratching our neighbors\' rides, throwing it into reverse without backing into the drunken hordes streaming out of the concert, and then: hurrah!  We\'re facing out, we\'re heading up the aisle, we\'re on our way, one of the first out, nothing can stop us now!    And then: a car about five spots up pulls out into our lane, and the two of us are trying to get out.  No problem.  Then, a convertible beamer with a truly heinous lady inside pulls in front of us.  But that\'s cool, we\'re still rolling...at a snail\'s pace.  Then another car pulls out.  And another.  And so on, and so forth.  And they tell two friends, and they tell two friends.  And all of a sudden, EVERY car has been pulled out of its spot, but there\'s nowhere for ANYONE to drive, because we\'re just a huge line of cars, pulled out of their spots, with nowhere to go.  Our brilliant plan of beating everyone else by sliding out juuuust at the end of the encore turned out to be  everyone else\'s plan , too!   But that\'s no problem.  Jennifer and I are good times, we can entertain ourselves, we own some kickin\' tunes, we\'re in no particular hurry...this line will start moving soon enough, right?  Half an hour later, after having moved maybe THREE car lengths total, we turned off the engine to preserve precious battery juices, which means: the small comfort of occupying ourselves with kickin\' tunes must go down the drain.  Lots of other cars started getting the same idea, so now, we\'re a whole line of parked cars that only occasionally grind on the engine to roll forward a precious few inches forward, so as not to let any stragglers back their cars into the tiny pocket created by unaware drivers.  Still, we\'re still good company, right?  Jen and I have the best of times...when we\'re in a happy mood, which is 99.44% of the time, a la the purity level of Ivory soap.  (David Letterman once said the other .56% of Ivory soap is filled with pure jam preserves...crack open a bar and check it out for yourself!)   You know what gets us into the other .56% of the time faster than anything else?  People.  Lots of people.  Large herds of agitated, obnoxious, drunken, mildly frightening, brain dead 909ers who haven\'t the foggiest notion how to MERGE properly so as to get out of the parking lot safely.    Merging.  Huge, tough concept that Los Angeles natives take for granted.  Every other car and you\'ll get there soon enough, we PROMISE.    But guess what?  If NO ONE agrees to let anyone go in front of them, there ends up being an eight-car can-can line at a standoff in front of the two-lane driveway, with no one able to squeeze their cars out of the spot because no one\'s willing to give an inch and let anyone squirm ahead of them.    And each of those eight cars in the can-can line have an ENORMOUS line of cars behind them, equally unmoving, except to close tiny gaps between cars into even tinier, nonexistent, wafer-thin slices of air between bumpers, so as to ensure NO ONE take their spots, either!  And the longer Jen and I dwell in that .56% of our personalities, the more we start becoming agitated and unsafe 909ers ourselves, such that we soon developed a rule for ourselves that, as long as we could see the printing on the top of the next car\'s license plate frame, we weren\'t too close.  Bad.  Bad, bad, bad.  As we are parked in this car...not even feigning hope that one day, movement may happen, Jennifer unbuckles to slither up out the car window and survey the landscape.  Nothing but parked cars for miles around.  No one is moving.  Lots of angry red brakelights, and even angrier fists slamming against steering wheels that really never did anything to deserve such violence.  She also sees: a few people who have given up entirely, and have pulled back into a parking spot to take a nap and wait out the traffic; the occasional pod of tailgaters numbing the pain of our hellish standstill with countless empty 40s littering the back of their truckbeds...yeah, I really want to share the road with those guys in a few...years, or whenever we\'ll get out of this lot; and oh-so-worst-of-all, a few hair-tossing, giggly, slithery-looking girls who have decided to pass the time by making their dates grateful that they don\'t need to pay attention to the road because SO many more interesting things can happen in a  parked  car!  Ick.    Yuck.    Ew.    Noooo!    Remember that convertible in front of us with the truly heinous lady inside?  Yeah, well, so will everyone else who walked up our aisle on the way to their cars that night.  Which, we must confess, did make a decent diversion for a few minutes as we watched other people comment outright, exchange horrified glances at their friends, flash their highbeams or honk angrily at the silly trashy woman.  Oh, and to her man\'s credit, he didn\'t seem too into her either, because seriously: silly trashy woman!  Convertible!  No!  Long story short, (YES, this IS short, compared to the actual agony of being there!) can you imagine our abject horror when,  two hours  later, we\'ve made it to the can-can bottleneck of cars at the driveway, and we realize that the problem was that Glen Helen Pavilion hires NO parking lot attendants, no traffic cops, no one with the foresight to think, hmmm...if we told which cars to go and which to stop, we\'d ALL get to go home a little sooner...  Anarchy in the parking lot = total automotive stagnation.  Drivers left to their own devices = unparalleled stupidity.  Dirt people suddenly evolving into polite sensible individuals who give others the right of way = as likely as when pigs fly at the speed of sound and finally master the art of outrunning Farmer John before he can saute them into full-fat bacon.  Mmmmm, bacon.  Never mind that, once we\'ve made it out onto the street, it\'s still a sub-1-MPH crawl the final stretch to the freeway, where we somehow managed to anger the truck a few lanes over into a double-fisted flick off, followed by proctologically-close tailgating, high-beam flashing, and attempted faux-ramming worthy of that silly 70s B-movie,  Death Race 2000 .           Do you see how driving REALLY ought to happen faster than a mere 500 yards in 2 hours?  Maybe it\'s just me, but don\'t venues have an obligation to ensure their patrons can LEAVE the concert they\'ve just paid an arm and a leg to attend?  Wouldn\'t it just be good business sense to have your customers leave satisfied?  And how much trouble would it be, really, to have someone don a pair of white cotton gloves, and do the pinwheel arm motion (go go go!) or the palm out motion (stop stop stop!) at tens of thousands of drivers?  What a waste of hosting a truly rockin\' concert, Glen Helen, because  everyone  stuck in your parking lot that night would rather get their eyeball caught in a beartrap than endure your particular brand of psychological abuse EVER again.   Moral of the story: if Jennifer and I had to live out the mind-numbing agony of the Glen Helen parking lot, we\'d hate to see all our pain and suffering go down without helping our friends and family avoid getting their limbs caught in the meatgrinder as well.  I\'m sharing our pain as a public service, a cautionary tale, and a call to action!    Boycott Hyundai.  Shout angry things out your window as you\'re cruising up the 15--or, at the very least, save those angry fist-shakings for the young whippersnappers who designed the traffic patterns there.  Write strongly worded letters to the proprietors of Inescapable Labrynthine Parking Lots, Inc.  But most importantly: do yourself a favor, and never, EVER buy tickets to any concert at Glen Helen Pavilion.    Think I\'m just whining? Blowing things out of proportion?  Flexing my muscles in the Hyper-Bowl of hyperbole?  Try this on for size:     urlLink http://www.nodoubt.com/band/journal.asp?newsID=13454   When was the last time rock stars flitted down from their gilded cage to apologize to  you  for forcing you to take a foretaste of the third circle of Hell?  Of course,  none  of this could\'ve ever happened if we had a decent traffic cop on our cotton-gloved hands...  I\'m just saying!  '),(21848,3639495,'21,June,2004','2004-06-21 00:00:00','   (Or at least, Musings Sprung from a KCOP Broadcast, but what kind of sexy title is that?)    I\'m watching the Dodgers/Giants game right now.  That is, it\'s on directly behind me and I keep having to swivel my chair around to catch any exciting plays as I am not an owl and cannot swivel my head around 180 degrees.    Pity.  \'Twould be such a fine fan skill.    You\'d think it would be a brilliant teaching asset, but the truth is, my students have NEVER seen the back of my head!  I\'m of the ever-watchful school of thought.  As far as I\'m concerned, the best teaching advice ever is \"You know, kindergarten is like the ocean...you don\'t want to turn your back on it.\"  Trust me, students in kindergarten and junior high are very similar: ready to climb the walls the second you let your guard down.    P.S.  Guess where that brilliant piece of educational theory came from?   Kindergarten Cop  (1990).  You know, that pre-gubernatorial Schwarzenegger flick far more famous for lines like \"It\'s not a too-mah!\" when the little kid thinks his teacher\'s headache may be cancerous?  I swear, that\'s one of the best teaching films EVER--my Mommy and I both love the part where he collapses onto his bed from sheer exhaustion after his first day of teaching.  That\'s, like, a true-life account of our lives EVERY SINGLE DAY after school.  Brilliant movie, you should go rent it or just sit still long enough for them to run it with umpteen commercials on TBS.  It may remind you of the charm that enabled Arnold to bamboozle California voters.  Oooh, snap!      But the good news is: this is baseball, so I have little need for head-swiveling and plenty of time to write between exciting plays.  Don\'t judge me, Vin Scully!  You know baseball is slow.  There\'s no need for constant TV staring like you have to do hockey-style lest you miss a second of excitement.  Strike that.  Like  I  have to do hockey-style. I think normal people are capable of movement during games, trusting in the unbridled goodness of instant replay.    Amateurs.  [By the way: I choose, at this early point in the summer,  not  to whine about the possibility that there may not be any hockey next fall because of a \"strike/lockout nobody wanting to work\" situation.  Look at me, showing restraint!  But I reserve the right to whine in the future, because SERIOUSLY, no hockey for me is like you getting your arm lopped off in a tractor accident.  Or in whirring helicopter blades, like that snarky bald man on ER.  Good times.  End tangent.]  Tied at 2, bottom of the ninth, Giants have a runner on first, in case you\'re curious.  Hey, guess what?    You are a  GRAMMAR GOD !   If your mission in life is not already to preserve the English tongue, it should be. Congratulations and thank you!    urlLink   How grammatically sound are you?     brought to you by  urlLink Quizilla    Yea, me!  Frankly, it would be bad news for my students if it didn\'t say that, right?  That\'s from a grammar quiz that my friend Laura mentioned on  her  blog:   urlLink http://www.livejournal.com/users/tree_and_flower                                        I may be a divine grammarian, but WE (the Dodgers, remember that\'s how this started?  Oh, yeah!) just walked two guys, \'cuz we do that, so now bases are loaded.  Dumb.  I wish I had angry sports voodoo like my Great-Aunt Mary.  She rocked the crazy mojo or juju or whatever you prefer to call it.  She\'d sit at the kitchen table, pushing string beans through this tiny razor-like contraption that split them into skinny little slivers...julienne beans or whatever, while muttering curses in Spanish at a fuzzy TV set or fuzzier transistor radio.  I should go buy a housecoat and some beans.  Maybe I could work up the serious anger with some Mary-like props.        Some guy named Ransom just hit a lame single that my sister could easily stop despite being three seasons rusty with all her softball skills, meaning we lost.    See why I need voodoo?  More than that: see why I need writing to distract me?  But the good thing about baseball, that even Saint Hockey can\'t offer, is that there\'s always another chance tomorrow.  Hurrah for hellish working conditions for my boys in blue, off days are for wimps when you get a six month winter break!   Still, boo for making me waste an evening of head and chair swiveling on a San Fran win.  Ewwwwww....  '),(21849,3639495,'15,June,2004','2004-06-15 00:00:00','   \"The time has come,\" the Walrus said, \"to talk of many things.  Of shoes, and ships, and sealing-wax; of cabbages and kings.  And why the sea is boiling hot.  And whether pigs have wings.\"  - - from Lewis Carroll\'s  Through the Looking Glass .   Out of context, that section of the poem has become a charming little cliche for \"I have a lot to talk about\" or \"Gee, I have a myriad of interests!\"  In context, actually, it\'s a bit less cute.  The English teacher in me wants to pique your interest enough that you go look up the whole thing and rediscover some poetry.  Sufficed to say, the walrus and the carpenter are using cutesy small talk to lure unsuspecting baby oysters to their death.    Really!  Check it out for yourself:  urlLink http://www.jabberwocky.com/carroll/walrus.html      And if you\'re too lazy to read, (well, you wouldn\'t be having this little exchange with me, now would you?) go rent the Disney version of  Alice in Wonderland .  In it, all the oyster-slaying barbarism is acted out for you by little cartoon walruses (walri?) and carpenters. (That one I know, a la the 70s brother-sister act of \"We\'ve Only Just Begun\" fame. By the way: I make it a point never to sing love songs to my brother--but then again, I  always  make it a point to eat on a regular basis, so I guess Karen Carpenter is not my role model in many things.) Ooh, and the \"barbarism\" has a friendly \"G\" rating, no reading involved.  The time is long since past for me to talk of many things, it\'s just I finally have the time to do something about it.  And seeing as I\'m going to be staying in the teaching biz for at least one more year, I need to have an outlet for all this pent-up writing energy.  While all my colleagues run out into the real world brandishing new Journalism School degrees and an armful of clips to gain entry at elite newsrooms in Boise, Bakersfield and Barstow, I\'m left behind trying to get my writing out into the world.  The cult of the journalist, just like any job worth doing, is wrapped up in a major Catch-22: you\'ve got to be published to get published.  In other words,  my  J-School degree means very little on its own.     The Polish side of my family has a saying: \"Find a way.\"  It sounds much fancier in Polish, but then you wouldn\'t know what I was saying, would you?  As long as the writing bug is in my soul...and it\'s in there good, I tell you...I\'ve got to find a way to get my ideas out there.  So if no one will publish me just yet, I\'ll publish me.    Really, it\'s not so silly as it sounds!  Just ask Izzy Stone.  Izzy--officially I.F. Stone--was an amazing investigative journalist who worked just about everywhere in New York throughout the 30s and 40s muckraking to kick up the big stories.  But it wasn\'t until the 1950s when he started his own paper, the  Weekly , that Izzy was able to make the biggest impact.  His small paper became a cult source of information for politicians, city bigwigs, and just about every journalist at those big papers he left behind because his work was so thorough, passionate, and unfiltered.    That\'s every writer\'s fantasy, even if they\'re too faux-modest to admit it.  So Izzy?  Here\'s  my  Weekly in weblog form..give or take a steady timetable.  So Dad?  I\'ve found a way.  So everyone I\'ve lured over to this site for a first look at this blog?  Don\'t let the title scare you off.  Sure, being a demi-Pole, there\'s bound to be some talk of cabbages.  And okay, being a hockey fan from Los Angeles, there will undoubtedly be talk of Kings.  But beyond that?  Movies, politics, friends, brilliant ideas, complaints, celebrations, half-formed thoughts, real-life, and everything in between.    Get set for talk of many things.  Some of it important, some of it Seinfeldianly insignificant, but all of it pure, unfiltered me.  Cheer, cower in fear, scratch your head, read, skim or ignore...but know this: the time has definitely come!    I can only hope this opening post strikes you as a charming little cliche to let you know I have lots to say, and not a sunny diversion just before I lead you into a barbaric slaughter by unchecked rambling that even those poor, devoured baby oysters wouldn\'t envy...  '),(21850,3639495,'02,July,2004','2004-07-02 00:00:00','   Hey...I hear those people who are able to regularly update their blogs do so because they actually PUBLISH things, rather than building one epic rant piece over several evenings.   I guess it\'s time for  VH1\'s Behind the Blog: Maimed Snail .  Reliving the parking lot experience kept making my mind so weary, I\'d fall asleep.  It happened, like, four days in a row.  Uncool.  Must get better at this.  Okay, that\'s all.  Happy Fourth of July, everyone.  Happy 228th birthday, Declaration of Independence.  Only a John Han-cocky person like me would bother taking up so much room on an uberimportant historical document, or take up so much room in your busy day reading a fulmination on the poor driving habits of dirt people.  Ugh.  The thought still gives me chills.  If you don\'t think Hancock was an obnoxious fellow for writing so big, then you must have the proper tolerance for me.    If so, read on...  '),(21851,3639495,'10,August,2004','2004-08-10 00:00:00','   So, with an eye on dusting off old rants and raves (see: August 9th for the full-on blustering explanation of why these old ideas are just finishing now), here\'s one bit of schlock I started way back on July 16th...     Check this out:      \"Five charming sisters on the gayest, merriest manhunt that ever snared a bewildered bachelor! Girls! Take a lesson from these husband hunters!\"    -- An honest-to-Pete quote from the honest-to-Pete trailer for the 1940 version of  Pride and Prejudice , starring Greer Garson and Sir Laurence Olivier.   Wouldn\'t I love to take credit for finding that gem on my own? If only. Actually, I just finished reading  The Jane Austen Book Club , by Karen Joy Fowler, and she has this whole Reader\'s Guide worked out in the last 40 pages of her novel with 200 years of responses to about Austen\'s work. That quote was the definite winner.  [By the way, a close second in the race for pure excellence is \"Every time I read \'Pride and Prejudice\' I want to dig her up and hit her over the skull with her own shin-bone.\" - Mark Twain, 1898. Note that, as torturous as he finds Austen, he  does  subtly \'fess up to repeated readings!  ...One can only dream that my writing will be so repugnant to the authors of the future that they will long to exhume my rotting corpse to beat me down with my own femur (which is a sturdier bone and could do MUCH more permanent damage than a delicate shin--I know enough of Twain\'s brilliance to presume he would choose THAT bone instead, given the chance to choose again!)]  So tell me, fellow citizens of 2004: how do you like  them  apples?  Manhunt? Husband hunters? Snaring a bachelor? Taking a lesson from their example?  Oh, the gall of MGM! I\'m so shocked!  Right?  Well, I was. Actually caused me to drop the book in surprise and run it to Jennifer for show and tell. \"Can you believe this?\" and the like.  It\'s condescending and chauvinistic. Tacky and trite.  Not to mention just plain silly.  And then, with an inward gasp of outright horror, I realized something:   I\'m  a husband hunter.   Of course  I am. Not only is that idea unsurprising, it\'s expected. Understood. Implied.  It is a truth universally acknowledged, that a single woman in possession of good breeding hips, must be in want of a husband.  That\'s not the shocking part.  The shock comes in openly admitting such a hideously antifeminist thing. In 2004. With an Immaculate Heart education, no less!  ...that\'s just NOT done!  Which is a very antiquated notion in and of itself. Subjects being taboo, admitting such a thing is unladylike, everyone\'s thinking it, but no one\'s saying it, yadda yadda yadda.  It\'s \"Elephant in the Room\" syndrome.  You know, like that little dance we all have to do when we try to describe people without reference to race, body type, physical disfigurements, Siamese twins welded to their foreheads, etc.  Instead, we learn to refer to people as, \"you know, that guy who was standing to the left of that girl who was wearing orange that one time?\" Rubbish. When it would be far easier just to bust out with \"I know a man with a wooden leg named Smith...\" without fear of seeming like an a-hole for pointing out his general pegleggedness.  Poor Smith the Wooden Leg. The PC police won\'t let us speak of him anymore. Curious how being subtle about something is far preferred, albeit entirely less efficient.  \"I\'m a husband hunter\" is the PC-equivalent of \"Well, he\'s black\" on the tact-o-meter. No one with an ounce of class would ever say that.  Which is precisely why  I  can.  Although (to be fair) I\'m of a generation that has traded in cotillions for  Girls Gone Wild , so I\'m not exactly a pith-helmeted, tranquilizer-totin\' big-game hunter compared to some ladies...  And thank goodness for that! But:  Even a person who seems Zen as can be--sitting in her rowboat, casting a line out onto a seemingly fishless glassy pond--is ultimately hoping to pierce through a trout-brain or two herself! Don\'t be fooled by the lack of outward-bound showiness...hunting is hunting either way; \"getting your hooks into someone\" isn\'t an expression that appeared out of nothingness!  As I was saying: I  was  shocked when I saw that MGM quote--especially the bit about \"Girls! Take a Lesson!\"  Jot down notes, learn their moves step-by-step, do as they do!  \"Ew!\" said I, for that is how I was raised to react.  Except we really do, you know. We take down notes.  Not out loud. Not in any notebook. Not in the secret depths of our girlhood diaries and journals. Not even in a way that we ourselves notice or recognize. But deep down (where we all know those fish are hiding!) we\'re taking  mental  notes. Imprinting ideas, thoughts, reactions, attitudes, responses deep in the subconscious levels of our brains.  We\'re using pop culture to program ourselves for future mating rituals. It\'s all very anthropologically sound. I\'m sure, had I the formal sociological training, skills, knowledge and information, I could perform a formal study that would arrive at that very conclusion. We upload tidbits onto our own cerebral diskettes at a teeny-tiny age.  Ergo: everything I know about snaring bewildered bachelors, I learned from watching TV. And movies. Reading books. Or any time I  stopped  singing along with pop music long enough to comprehend the meaning behind the words.  Maybe I never enrolled in Olivier\'s school of man-snaring, but there\'s a whole stockpile of course notes coiled deep in my cerebrum from sources as far-flung as  Anne of Green Gables , The Beatles,  When Harry Met Sally ,  The X-Files, The Baby-Sitters Club  and the Colin Firthization of Jane Austen herself.   Pride and Prejudice  went on to be a monster hit in 1940 and is still in constant rotation over at Turner Classic Movies to this day. Most people would credit the widespread popularity of the source material. All apologies to Miss Jane, but that\'s fantastically naive. Advertising! Pure and simple. Savvy, match.com \"Your ideal mate is waiting!\" SPAM filtered into the inboxes of moviegoers 60 years before email.  In the age of political correctness, the League of Women Voters or Planned Parenthood or Gloria Steinem or Oprah\'s Book Club would tackle MGM in a second flat for throwing a movie tagline like that out into the open.  That\'s a bit of a shame. Still, I don\'t think I\'d be comfortable in a culture where everyone felt free to wave their every inappropriate thought about completely unfettered. I prefer to be the only one lacking an internal monologue.  Although I can\'t have completely abandoned tact...after all, I made it all the way through this post without  once  mentioning MGM\'s description of their movie as the gayest manhunt ever.  Aw, phooey.  ... almost  made it!  '),(21852,3639495,'09,August,2004','2004-08-09 00:00:00','   Have you ever had someone disappear on you for such an unbearably long time that the morbid half of your brain pauses now and again to wonder if said someone is, in fact, dead?  Yeah. Me too.  Sorry to have been the cause of your latest death-musings.  -------------------------------------------------  Seriously, that would make a  sweet  greeting card. \"Sorry I faked my own death\" and the like. Very popular among Mafia goons, eccentric billionaires and soap opera characters.  Not only am I alive, I am chock-full of interesting things to say and share and contemplate. So why, you may ask, haven\'t I written for the last month?  Two things, really:  a)  I\'M BUSY!  No, really. It looks like I\'m on vacation, but I haven\'t had a free day in a loooong time. And YES, I know I can\'t whine, MOST jobs never stop at all, blah blah...  Society as a whole has a big ol\' chip on its collective shoulder about teachers getting a summer vacation when the rest of the civilized world has to work nonstop. Trust me--not only have we earned our time off, it\'s the reason why more students aren\'t killed by their teachers each year in an act of frustrated rage.   Know how I know? \'Cuz golly by golly, that almost happened to me...but that\'s another post.  Anyway...even without official work happening, the days keep filling themselves! Sadly, the bank account does not. So in that sense, it will be nice to get back to work, but it is kind of a fun existence being faux-unemployed. Like I could fulfill some excellent professional hobo riding the rails fantasy because I have time on my hands, without falling into actual, crushing unemployment that involves having to search want ads and employ head hunters.  P.S. This is also why, if you live within the sound of my voice (no, really, I\'m uberloud, that IS all of Southern California, stick your head out of the window right now and see if it isn\'t true!) you not only haven\'t read anything I\'ve written lately, but you haven\'t seen much/any of me, either. My existence at this point is something of an urban myth. Blurry photographs exist, and a few people claim to have sighted me at the gas station or In-N-Out or something, but really, at this point, I\'m like Elvis: only my rabid fans hold out enough hope to believe those stories.  Which brings us to:  b)  I DID WRITE!     No you didn\'t, I\'m looking at your blog right now and I can SEE nothing\'s been written since July...do you think us totally daft or what?  Bear with me, irritated readers!  I really did write--in fact, right on time, within a week of the last post, I totally brought an entire off the wall essay together, and then I  lost  it! No lie. I went to check a preview of my post to see what it would look like (because I\'m so vain and previews are the mirrors of the writing world. In case you\'re curious, or feeling smart-alecky, YES, I do think that song is about me. I\'ve gone on record several times about that fact!) and when I came back, the page was BLANK. Oh, it was tragic. That was a gorgeous post, it was.  So, I set out to rewrite the thing, and in doing so, a lamentable thing happened: I got too tired to type. Combo of two things: I write very late at night as it is, and I was frustrated at not being able to replicate the gorgeousness of the OG post. Which is when I discovered a VERY dangerous thing:    Save as Draft   As in...start this now, come back to it when you feel like it, I\'ll hold your place \'til then!  \"Save as Draft\" is a sexy little icon at the bottom of the typing page I use that reassures procrastinators like me and gives us free reign to put FINISHING our work off for infinite mananas.  Very dangerous. Because when I came back, I didn\'t feel like THAT post anymore, so I started another. Got tired. Saved as draft. Came back. Had lots of good ideas, but really wanted to get this new idea out into the world. Started a third post. Got tired. Saved as draft. Went on vacation. Came back. Wanted to write about the vacation...you know, \"while it\'s still fresh!\" Got tired. Saved as draft.  Yes sir. Yes ma\'am. I have four works-in-progress on my behind-the-scenes page. And remember than unfinished for me is like an encyclopedia for a normal kid, so we\'re talking about a decent chunk of hidden work.    See , I\'m  not  a nasty old liar. I  did  write. I just never got around to actually PUBLISHING, is all.   Oh. That\'s an important step, I guess.  So what did I do when I signed on this evening? Why, started a fifth post, of course! You see, I needed to explain what all this silence from the cabbage-front was about. But fear not, loyal readers: I\'m not allowed to Save as Draft here...I need to be strong and face my fear of endings.  As school is starting again soon, I have to pull together all my summer vacation loose ends--and THIS is one of them. Hidden work must be given a chance to play in the sunshine!   Stay tuned over the next week for posts on: 1) Why sexism is really quite refreshing, 2) The worst form of psychological torture, 3) The future will be a scary place, 4) The band so incredible I asked security to rebound my poorly thrown panties back toward the stage.  I kid you not.  And as I do better with deadlines, even artificial ones, I figure if I finish up one unfinished symphony per night, I\'ll be all caught up by the weekend. Which should give me plenty of time to write one more new thing: \"Germans: Not just for world domination anymore!\"  And then, seriously: don\'t ever expect that kind of flurry of activity from me ever again. Once a week is plenty, thanks!  Please consider this my contract with you, dear readers. Then, if I don\'t deliver...you can come after me with a pitchfork and lit torches.  Until then!  '),(21853,3805349,'30,June,2004','2004-06-30 00:00:00','   I am in Des Moines right now. I have been here at my sister\'s apartment for about a week now. I have been having a lot of fun. One major downfall though is that I have had sunburn on my shoulders. I couldn\'t move very much without feeling pain. The burning went away and now it\'s all pealing. My skin is dry so when my shirt gets moved around it hurts for a bit. Sleeping is a pain. I guess I am somewhat out of order.  I miss some stuff about home. I love to play games and I am a big fan of the  urlLink Warp Pipe Project.  I left my Gamecube and Broadband Adapter at my house. If I come back this summer I will be bringing that along with Mario Kart: Double Dash!! because my sister has cable and I will be able to get a good game going. We don\'t have cable available in our area yet. MediaCom has it available in the city but we have our house in the country so we have to resort to low-grade DSL. It does the job, I guess.  I am going home later today. In a way it sucks but really, it isn\'t that bad. I was on a trip in Washington DC for a week. Then I came back home for three days and then came down here for a week. I haven\'t seen much of home.  Well, I don\'t have much more to say! I just want to emphasize how cool  urlLink Warp Pipe  is and how everyone should look at  urlLink Warp Pipe  instead of nasty XLink. The  urlLink Warp Pipe Revolution  is near and you should check out  urlLink the Warp Pipe Forums  to find a download of Test Builds of the Megaton, also known as  urlLink Warp Pipe v0.4.  '),(21854,3805349,'30,June,2004','2004-06-30 00:00:00','   This is my first entry in the Undies Drawer. I should start out by telling you a little bit about myself.   I am a 14 year old male who enjoys computers, football, games, and lots of music as well as movies. I like hanging out with my friends and family. I play guitar and am working on learning C++. I am interested in persueing a career in the field of software engineering.  '),(21855,3805349,'06,July,2004','2004-07-06 00:00:00','   I\'m back in Des Moines at my sister\'s apartment. They have cable internet so I was able to play a nice game of Mario Kart with Chad, the project manager of  urlLink Warp Pipe Technologies , the best frickin\' thing on this planet, at around 1/2 speed. It was the fastest game I\'ve played.  We went to an animal adoption thing and adopted a dog. It was stray so we got to name it. I suggested Meat or Bone but everyone else said that Maddy was better.  My brother in law designed a tattoo for me. I can\'t get it put on but it looks cool. It\'s a Japanese character, but we don\'t exactly know what it means. I thought that a tattoo with \"GONZO III\" on it and then some Japanese writing under or after it would be cool.  '),(21856,3805349,'01,July,2004','2004-07-01 00:00:00','   Well, this morning I woke up at around 7:30 to head home. It had been a great night. My sisters and I went out around town and drove up to fast-food places. We\'d park up by the person who give you your food and we\'d blast Micheal Jackson\'s \"Beat It\". I don\'t like MJ, personally. He\'s quite creapy. That was pretty fun.  We got back to the apartment at around 12:30 and watched \"50 First Dates\" and fell asleep. I was on the floor of course. No complaints though.  We got back at around 11:15, just in time for me to shower and get ready for my dentist apointment. I had no cavities, which proves that it was a waste to come all of the way back home for a measly dentist appointment.  I am running out of things to keep me entertained. I should go upstairs and check whats going on.  Peace out.  '),(21857,3805349,'03,August,2004','2004-08-03 00:00:00','   It\'s been a while since I\'ve updated. I guess I\'ll do it now.  Well, last night I started working on the graphics for my site. I have the table made but the problem is that I have a large banner at the top so the table is offset. I have to cut the image up so I can make everything fit nicely. I\'m going to start doing that really soon.  I forgot how much fun web design can be.  I haven\'t had any websites since those that I made when I was younger. There was one for 4x4 Evolution, an offroading game that I used to play, and one for video games in general that my friend and I tried to set up. They all have failed. This one probably will too.  One thing that I\'m doing differently, though, is I\'m setting everything up, getting some content, and then I may buy a domain and a hosting plan from Dreamhost.  I have chosen Dreamhost because Chad Paulson, the Project Manager of the  urlLink Warp Pipe Project  had possitive feedback. There is also a affiliate program where you can earn $65 for each referral you get. If this is actually true, I figured it was an opportunity to get a domain and a good hosting plan, while at the same time, support the Warp Pipe Project, and all their great efforts.  Well, I\'ll let you know if I ever buy a domain.  '),(21858,3860294,'08,July,2004','2004-07-08 00:00:00','    Day3(the Final Day)  Once again Todd woke us up at 6:30, grrrrrrrrrrr, we packed up, went to eat in the cafeteria. The guys got together in the gym for like an hour. We had a \"guy talk\" about sexuality and how it should be respected. haha the guy was stupid who gave the talk. but had some good stories like, one guy loved life so much, but one day he got his arm stuck between a boulder and a clif side while hiking ALONE. So he waited 3 days. He eventually cut his arm off w/a sharp rock. it took him HOURS! but he lived because he loves life so much, his arm means nothing. After that we headed over to the Conference. We had the worst seats! but it was ok. We sang praise and worship, by now i was beat. i was tired as i could be, and my clafs hurt from jumping. Then Mass started it was ok until the homily, geeze the guy could talk for forever! lol but it was soon time for communion. but during the blessing of the host, a girl was rendered unconcious by the holy spirit...in regular day MASS! how power is God...wow...haha...but she was unconcious for like 20. Then she woke up and we were all Dancing and SCreaming HAVING THE BEST TIME! the the bus ride home was fun lol, i watched people ply Maffia  '),(21859,3860294,'08,July,2004','2004-07-08 00:00:00','    Day2  Ok Todd woke us up at 6:30. I went to the restroom cleaned up, went to breakfast, went to the conference and saved some seats. Then Praise and worship began. You should have seen everyone. they were all into it so much. It was really amazing. And my group looked really kool in our matching shirts, lol. Praise and worship lasted a while. Then it was eventually time for mass. Yeah Mass was long especially communion. But i LIVED! Lunch was right after that. mmmmmmmmmmmmmmm food. we had sandwhiches then. Lol richard made a pyrmid out of the boxes...i mean he is like 6\'+ and he could barely be seen over the top. Then the group moved its way back to the Conference. More Praise and worship. Then we split up. I stayed to hear a Father talk about the Media and hop it was affecting us all. Yeah it totally changed my veiws. It was nasty too. half naked women in ads. i cant believe its not illegal, but i realized what i needed to do. I have decided i will go into the media and forever change it. Thats my life goal now. Next the \"young apostles\" had a skit thingy where people ate things ina  certain amount of bites. the a girl had to eat someSPAM! lol she backed out right away and called up Fr. Jose. he ate it! i was amazed lol. after that, Adoration. Adoration is when the Bishop brings in the body of christ(most people will say its a piece of unlevened bread) but catholics believe its Jesus, but its when you express your intense regard and love and devotion to God and Jesus through deep mediation and prayer and ask for help if needed. But i couldnt believe my ears and eyes. Steubenville adoration is like no other. People opened themselves to demonic forces, others were rendered unconcious by the holy spirit. People were crying hysterically once the Bishop entered for no reason. People started laughing, for no reason. But really the holy spirit had entered and caused all this. I didnt realize what i had revieved from the holy spirit at the time. What did i recieve? I recieved belief. I had been dobuting God, but now, i cant think why i did. He is real! HE REIGNS! DONT FORGET!, after adoration we danced praised like no other lol. then My sis and I went to get emily a present and so we did. Then our group talked about our adoration expreience. it was a fun night. But once we got back to the dourms it was almost time for bed. but Todd bout all they guys a pizza lol. nothing better than Steubenville pizza! but then i heard something amazing. som guys were making their own music and dancing just cuz they wanted to! it was SO awsome. they were making a beat w/ empty bottles and some\'s guitar. It was fun. b4 i knew it, the time was 12:30, lights out. End of Day2.   '),(21860,3860294,'08,July,2004','2004-07-08 00:00:00','    Day1 I packed up made sure i had everything. Kelly and I go to the church, find out we would leave like 30 min later, but i had a perfect time to look arround and see who is comming. I didnt know anyone that was lol. But then we packed up and went on the Charter busses and set off. About an hour into the trip someone asked Todd(my youth group leader) to put in Finding Nemo. You would think YAY! FINDING NEMO! but now everyone was like NOOOOOOOOOOOOOOOO! lol AGAIN they said no because Todd lectured us on how Disney has a deeper meaning. A bad meaning. I mean it makes sense but come on i have heard that story like 5 times. lol. So eventually it went on. We made it into Lousiana arround 3 and stopped in Shreveport. Fun time. I went to Wendies to eat it was good. Then I went back on the bus where I called sarah. Neways the next thing i know i hear yellling. lol. Austin spilled some one\'s frosty on the seat. haha. it looked like poo. lol. We moved on to Alexandria from there. Nothing happened until we arrived at our Dourms. All the guys got off and went to their rooms. i found a nice spot set my sleeping bag and my boat up(lol my inflatable pool boat kept my comfy from the hard floors). Next we all went to the place where they held the Conference. We did some praise and worship for about 5 hours. then we went back to the dourms. we had about 30 min till we went to bed so i learned a card game. then it was 12:30 lights out. Day 1 came to an end.   '),(21861,3860294,'08,July,2004','2004-07-08 00:00:00','    Hey im bored, no one is talking to i think i will tell the whole story of my Steubenville weekend. O.K.   '),(21862,3860294,'08,July,2004','2004-07-08 00:00:00','    Hey this is new so ill try to keep it simple. This is now the summer of 04. Summer is almost over, and Steubenville has passed, not much to look foreward to anymore, but hopefully the effects of steubenville will last. If u ever need help on anything im here ok? Just IM me at TheMatrixIsDown. The song on my site i love it so much. It makes me happy every time i hear it haha. well ill be stopping now.   Talk to you next entry.    Sven   '),(21863,3860294,'15,July,2004','2004-07-15 00:00:00','    WHEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEE... *ahem* ok well... &nbsp; Today 6flags, emily came. WOAH! HAHAHAHAHA! i got u to go to Mass! :P lol she hasnt gone to mass in idk how long but today i got her to go hehehehehe. Then while i was at 6flags, bobby and i went to duel on DDR obviousley he beat me cuz i suk at the metal pad. but yeah then we finished for then bleh i was about to throw up. then bobby made me \"scout\" haha he wanted girls to come up and talk to us tooo bad! then i hung w/emily geeze it wasnt fun until 3. Emily and I had a waterbottle fight, she won, i&nbsp;got soaked. then&nbsp;lol we all went on a log ride. bobby splashed us like no other. we were soaked. you could see a lil through&nbsp;the back of&nbsp;emily\'s tshirt, obviousley i ignored it. Then we went to dry off on the Bobsled ride. big mistake. we all went at once. i had to get emily to sit too close for comfort for me BLEH! (sorry if ur reading this emily it wasnt that bad&nbsp;haha) but i wont go into detail. then at the end of the ride WE GOT STUCK! bobby and i could just stand up in the cart and walk out but they wouldnt let us! GRRRRRRRRRRRR. 30 min having&nbsp;my Ex&nbsp;in my lap not the best day haha. but yeah. then we left and emily was all sad and wont tell me why&nbsp;still! but i think i know why. its prob not true.  &nbsp; &nbsp;   '),(21864,3860294,'14,July,2004','2004-07-14 00:00:00','    WHEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE  Just been so happy haha  WHEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE  lol and 6flags tomorrow!  WHEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE  ttyl   '),(21865,3860294,'13,July,2004','2004-07-13 00:00:00','    Hey nothing really happened today. i just woke up, played DDR (what a suprise!) then i went to the church to help bleech moldy chairs (bleh) then came home to DDR AGAIN! but this time i was rocking! lol nothing like Heavy mode...haha...bobby is sooooooo gonna beat me. Well im excited 2 days till 6flags haha! yay! but idk if sarah will go now...but yeah neways. I also finished the steubenville video to make it full length. yeah but nothing happened today yeah.   bye  '),(21866,3860294,'12,July,2004','2004-07-12 00:00:00','    It is only in texas where you meet the Misquitos on Steroids, and also lousiana. It is only in Texas where you are driven to the limits by the Misquitos on Steroids to the point where u spray bug spray in your room. Only in Texas. HAHA. Only in texas youll find you most common friend, the cockroach, and they fly too!. HAHA. Only in texas...(haha good times Chris)  Well its been fun. Guess what!? you all missed 7/11 day w/FREE, yes i said FREE SLURPEES! hahaha so fun! but yeah im keeping a log for alexandra cuz she is in france...lucky...but yeah today has been, idk, weird. Well i found out some things i may have not wanted to know. (im so worried zach! PRAY!) but things will move on, just dont forget the big guy above. then when i was at Driver\'s ED. it was ok. not the best day (\"whats that sign!?\" \"what sign?\" \"that sign that says wrong way\" \"ooooooooooooooo\" haha good times kelly) but it was fine. Then when i was gone brandi IMed me. I really need to talk to her! (dont do it again brandi). Then DDR. HEAVY MODE BABY WAHOO!(right on ZACH! and Bobby, no way i can play that song (my generation) its not even on my version, silly bobby) wow alot of insiders. well i dont think there is much more to say cept,  RISE UP AND WALK IN THE LIGHT WAHOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO! YEAH! ya hear me!? WOOT!    .......haha html style im such a geek   '),(21867,3860294,'11,July,2004','2004-07-11 00:00:00','    Hey guys, stressed out i am. So here i gooooooooooooooooooooooooooooooooooooooooooo.   Well i just got a letter from my youth group (LifeTeen) it reads \" You probably just arrived home from your weekend at stubenville. Hopefully this weekend will springboard for a very meaningful spiritual journey (haha yeah it will). God has given u a gift this weekend...yatta yatta yatta(a ha ! here we go the point i was going to make) Add some more prayer time,...blah blah blah, What ever you decide, it is your time with your best friend and your creator.\" there. I feel like im being pulled away from him. Lately i have been stressed. Thinking too much. Worrying about my social life. I need to stop focusing on it. I have been getting worse about it. The online thing also. I have been mean to people lately (sorry sarah) but i cant explain. I need a break from all this. I know more people also have this problem. Just want to get away from it all. And school will be starting soon. wow. but yeah. if yall need help or w/e with that stuff im here. im feeling ya. But right now. Social life, it shall BURN! hahaha. to bad it cant. for too long i have kept Christ back  i need to rise up go against the flow, walk in the light NOT LET THE DEVIL IN! YA HEAR ME!? AGREE YEAH! WAHOOOOOOOOO!    bye   '),(21868,3860294,'23,July,2004','2004-07-23 00:00:00','    Well this week. VBS all week! wow it got tiring. But i made a video for each day ( 4 OF THEM!) and eachday someone asked me\"how much are you getting paid\" (NONE!), Did you really do that?!(YES!), ori hear your going to be come a producer.(well maybe.) but yeah people were impresse by my stupid videos. haha only if they paid attention to my errors in them, bleh! well neways its been a long week im tired. Idk i am tired of this 0 comments crap. COME ON! i made it so everyone could comment...w/e.   Lately at VBS, i love what im doing making videos that proclaim god is there! its something im being called to do, well i think...but yeah i have gotten compliments left and right geeze makes me feel good. im going bye   '),(21869,3860294,'20,July,2004','2004-07-20 00:00:00','    hahaha im in a&nbsp;strong bad mood, hahaha and remember the time when...\"Hey im trevor and i am a vampire\" hahaha...were gonna miss u trevor :\'( ...and then...\"i got u a fashon card \'oooooooo how did u know?!\' hey cheerleader i got you a gizel carcue *silence* YAY GIZEL CARCUS!...hahahaha...The system is down the system is down, dowa down, dowa down, hahaha and then \"It is with my heavy heart to inform you that your son has exploded in a firey ball...thats was visiable from space, and did noway beat the everliving snot out of me\" hahaha \"I HOPE YOUR HAPPY KEVIN!\" hahahaha...then...\"Radioactivity is here, Radioactivity is here! Growing arms out of heads is here, today!\" hahahaha, and then...sorry last 1 i promise... \"way to go texas\" hahahah whooooo good times... &nbsp;   '),(21870,3860294,'19,July,2004','2004-07-19 00:00:00','    well i found someone found this poem and showed it to me, wow i think it really speaks for its self how guys and girls communicate... &nbsp; -BOY: i SAW HER TODAY GiRL: i SAW HiM TODAY -BOY: iT SEEMS LiKE iT`S BEEN FOREVER  GiRL: i WONDER iF HE STiLL CARES -BOY: SHE LOOKS BETTER THEN BEFORE GiRL: i COULDN`T STOP STARiNG AT HiM -BOY: i ASKED HER HOW THiNGS WERE GOiNG GiRL: i ASKED ABOUT HiS NEW GiRLFRiEND -BOY: i`D CHOOSE HER OVER ANY GiRL i`M WiTh  GiRL: HE`S PROBABLY REALLY HAPPY RiGHT NOW -BOY: i COULDN`T LOOK AT HER WiTHOUT STARTiNG TO CRY  GiRL: HE COULDN`T EVEN LOOK AT ME -BOY: i TOLD HER i MiSS HER GiRL: HE DOSEN`T MEAN iT  BOY: i MEANT iT  GiRL: HE DiDN`T MEAN iT -BOY: i LOVE HER GiRL: HE LOVES HiS NEW GiRLFRiEND -BOY: i HELD HER FOR THE LAST TiME GiRL: HE GAVE ME A FRiENDLY HUG -BOY: THEN i WENT HOME AND CRiED GiRL: THEN i WENT HOME AND CRiED -BOY: i LOST HER GiRL: i STiLL LOVE HIM &nbsp;   '),(21871,3860294,'18,July,2004','2004-07-18 00:00:00','    -Yeah that\'s another symptom that all girls experience. i call it \"exaggeration of everything to make the ex jealous\" -ITS WORKING! &nbsp; haha thats all i have to say... &nbsp; o and also.. new DDR PAD WooT! and also... NEW-DDR MAX! WOOT!!! &nbsp; so fun, sweating all down my back for the love of the game...want a challange. i think not! hahahaha! &nbsp; &nbsp; &nbsp;   gtg to bed VBS tomorrow making videos...haha   '),(21872,3860294,'27,July,2004','2004-07-27 00:00:00','    yeah i have been...  ...  ...  ...  Bored lately, so i havent really updated much so yeah w/e. Well, u know, i did change my site background music, so i did accomplish&nbsp;something new.&nbsp;School is comming up soon, itll be fun, i just cant wait to c my old friends. then also idk, i just, idk lol. but itll be fun at clark. I found out i only have 1 class w/rebecca and 1 class w/sarah, o well atleast were in a class together, but w/e. recently this one person i have talked to online, i wont mention her name, she is kinda weird to me. she gave up her faith, but in a situtation u would think u would want to be stronger in your faith. doesnt make sense to me, but neways... i tried convincing her to go into her faith again, but then again, most people dont veiw thing as the way i do. it just bothers me that someone cant believe in anything. i guess ill just have to accept the fact people can go through life clueless. O! i heard this on the radio, this might help out some people who read this: Is it a sin to swear? well, it isnt written in the commandments, but the only thing that relates to it is \"Thou shall not use the Lords name in vain\" by yelling out jesus christ and o my god as a swear word. but it does say in some parts of the bible that a person should not say words that would insult or make others feel bad....well yeah i gtg to Drivers Ed. now...       P.S. If yall want to email me, like about something to write about or w/e or make a crappy story, just email me or leave a comment...(my email=  urlLink shotpaintballer@comcast.net )   '),(21873,3860294,'03,August,2004','2004-08-03 00:00:00','    Last night i was lying in my bed...ARGH! i yelled as i strained and struggled endlessly. i couldnt sleep. i streatched maybe to help me relax, ARGH! i hurt even worse. I closed my eyes. Thinking of something nice in my mind nothing. everything was bothering me. It was too hott. There were too many arrows. it was too...too, i never found out. Why was this happening to me u say? well too much DDR was simply to blame. I just came back at 10 pm from Main Event. The DDR machine there was nice. I left my house to go to main event arround 3. stayed till 10. sweating like crazy after the 1st 15 min. yeah it was all good. now i dont know what to say, nothing really to explain, cept for the pain i am in. well yep, nothing left to say...        hopefully i can sleep better now...   '),(21874,3955951,'24,July,2004','2004-07-24 00:00:00','   Just talked to my girlfriend \"Katie\"&nbsp;about a lot of things.&nbsp; One of which is my having sex with other women.&nbsp; She asked me when was the last time I was with someone else and I told her about \"Debbie\" and our hooking up yesterday.&nbsp;&nbsp; She said that she was glad that I was getting it from someone regularly close to home (hopefully regular anyway).&nbsp; The distance between Katie and me is a bitch.&nbsp; She\'s in a shitty marriage as well.&nbsp; We can\'t get together often enough to keep either one of us satisfied. She asked about me having a double standard regarding sex, that is is it ok for me to fool around and not her?&nbsp; I told her if I could do it she could do it with my blessing.&nbsp; Then she proceeded to tell me about a guy she met near her that she\'s been chatting with for sometime now.&nbsp; They\'ve talked about sleeping together.&nbsp; I\'ve got to tell you I find this hot.&nbsp; The thought of her getting porked by another guy just arouses the hell out of me.&nbsp; Even if I\'d never be able to watch her go at it just hearing her&nbsp; tell me this is so erotic.&nbsp; They might be getting together soon. God I\'d love to have a spy-cam. '),(21875,3955951,'23,July,2004','2004-07-23 00:00:00','   Holy fucking shit!&nbsp; That was just awesome.&nbsp; I met \"Debbie\" at her work.&nbsp; Well actually I intentionally went in to show her who I was and who she\'d be getting ravaged by in a short while.&nbsp; I could see the sparks flying between us when I looked in her eyes.&nbsp; We got to the motel and in no time we started pawing at each other.&nbsp; OMG!!! what a receptive pussy she has.&nbsp; I had her cumming time after time fellating her wet hole.&nbsp; Finished her off by fucking her doggy style, drilling it home good and proper.&nbsp; When I came I let out a groan and slammed it in once more, grinding my hips into her ass.&nbsp; Even with a rubber on she felt the throbbing and pulsating of my pole when I reached my breaking point.&nbsp; God what a lay!&nbsp; What a way to spend a Friday afternoon.&nbsp; Fuck buddies indeed.&nbsp; This is going to be a steady fuck for me. '),(21876,3955951,'19,July,2004','2004-07-19 00:00:00','   from the timeline to fantasize about tomorrow\'s tryst. I just rejoined AFF Monday before last and I\'m already getting laid. That\'s quite a response for me. She answered my email right away that night and we talked on the phone. We chatted for awhile breaking the ice. God she\'s got a nice voice. After awhile she started talking dirty making me REAL hard, asking me how big I was getting. She brought me to orgasm just listening to her. I\'m gonna drill her good.  God how I need regular sex with a woman that excites me.  Depending on how we hit it off, this may be a regular fuck-fest.  Good, a fuck buddy! About time by god. ');
INSERT INTO `msgs` VALUES (21877,3955951,'19,July,2004','2004-07-19 00:00:00','   The greatest fear of many a man is that he passed up the woman who may well have been \"the one.\"&nbsp; Looking back at my single years that very well may have been true.&nbsp;&nbsp;I didn\'t date a great deal but on reflecting upon my past I realized that there were many fine women who were interested in me.&nbsp; It\'s just that I was so thick-headed that I totally missed the signals they were putting out.&nbsp;  &nbsp;  While in college, I remember going home night after night past lovers sitting on park benches in heavy embrace, some in various stages of lovemaking under the stars.&nbsp; I saw guys leading their girlfriends to their rooms for a night of heavy sex.&nbsp; As for me, I was alone&nbsp;fucking my fist.&nbsp; I convinced myself that I was all alone and that no one wanted me.&nbsp; For someone who should\'ve been sewing his wild oats at that time this was quickly driving me nuts.&nbsp; Self-doubt and questioning my attraction to women was a daily occurence.&nbsp;&nbsp; True or not, it seemed to me that I was in the minority in sleeping alone at night.  &nbsp;  I remember working at a variety show put on in the school of music.&nbsp; I was a stagehand that year.&nbsp; A woman in one of my classes was in a skit.&nbsp;&nbsp; She wanted to talk, and I couldn\'t spit out a word.&nbsp; She was quite attractive, damn near beautiful.&nbsp;&nbsp; Another potential lover I met as a freshman in bowling class.&nbsp; We hit it off extremely well in class.&nbsp; I made her laugh, which to me is a good sign. Best aphrodisiac in the world, right?&nbsp;&nbsp;Even though she was practically eating out of my hand,&nbsp;I couldn\'t get myself to ask her out.&nbsp; Another I met at a Christmas singing event.&nbsp; The sparks of attraction were flying, but once again I couldn\'t get myself to close the deal.  Even with&nbsp;a woman that I had met all the way back at freshman orientation who would\'ve been an easy lay, I found myself backing away at the crucial moment.  &nbsp;  I honestly cannot recall all of the ladies I passed by through the years before getting married.&nbsp; A number of them stand out in my mind.&nbsp; One in particular&nbsp;absolutely stole my heart with just her presence.&nbsp; \"Lori\" was the lady who made my knees weak.&nbsp; I felt a chemistry with her that was so strong it made me ache.&nbsp; Sure enough, she asked me out for drinks after work and I said that I had school in the morning.&nbsp; I backed away again.&nbsp; I spent the next two years trying to convince her that I really was interested while she dated others.&nbsp; I was eating myself alive inside thinking of what she was doing with them and knowing that I had blown my chance.&nbsp;&nbsp; Me being a passive-agressive didn\'t help at all.&nbsp; She did give me opportunities to start out but each time I was too dense to see it.&nbsp; I eventually moved away with her making one last ditch attempt to start someting on my last day in town.&nbsp; I went to my old workplace, saying goodbyes to all that I still knew there.&nbsp; There she was, as gorgeous as ever; her fair Scandanavian features and lush blonde hair shining for all to see.&nbsp; When I looked in her eyes she was just beaming.&nbsp; But I left her behind, never seeing or hearing from her again.&nbsp;&nbsp;Leaving her behind is&nbsp;probably the major regret of my life.&nbsp;&nbsp; My one wish in life is to have a chance to go back and do it over with her.&nbsp; She is without a doubt \"the one.\"&nbsp; '),(21878,3955951,'17,July,2004','2004-07-17 00:00:00','   At the wedding I had a sick feeling come over me.  This was not  the person I wanted to spend my life with.  The sex was ok for a  few weeks.  But beyond that I can honestly say that I never felt  \"in love\" with the woman I\'ve called my wife for almost 20  years.   God help me when I say this but I can definitely  relate to those who\'ve left their lover at the altar cooling their  heels.  Even though such an incident would be a fiasco when it  happened, it would be far better than spending your life with someone you didn\'t love. '),(21879,3955951,'14,August,2004','2004-08-14 00:00:00','   Well today was another slight.  At the checkout at Hy Vee I was cashing in bottle deposit refund slips.  The girl barely said anything to me, hardly looked at me at all.  The guy behind me was a rail-thing guy with a beard.  You could see her eyes light up and hear her chatting away with him about this and that. Jesus H. Fucking Christ, can it get anymore obvious?   I had the same thing happen several years ago at the bank.  The teller was all squinty-eyed with me and curt and polite. Then a slim blonde guy behind me got to her window and she just lit up.    Some would say that I\'m being too sensitive.  No I\'m just good at picking up things like that, little non-verbal clues to whether they are happy to see me or are disgusted to be around me.  I don\'t know how many fatties are at the board.  It\'s just that those that\'ve posted pics there are mostly in good height-weight proportion and some quite attractive.  And no one is sitting there hitting on me because I\'m down or feeling vulnerable at times.   No one wants to be near a fat guy. I don\'t even bother to post a hookup thread.  Don\'t make me laugh.   If I were a woman or a gq guy.....................................  It\'s just time that I do something about this weight for more ways than just being attractive to women.   This anger I feel is what\'s going to drive me away from overeating and more exercise and weight loss.   Screw the fat-hating world.  But when the weight is off................................I can be as shallow as the next thin person.   I\'ll pick and choose as I please.   I\'ll also remember those who belittled me, those who rejected me.   Even a dog remembers who kicked him.   I\'ll carry that grudge to my grave thank you very much.  I\'ve used food for too long as a tranquilizer, to numb the pain of an oppresive upbringing.  But do most thin people look at that?  No all they see is a fat guy.  Cancer, Aids you name the affliction,  the outpouring of sympathy is there.  But being overweight?  All they see is a person who can\'t control himself, a person they\'d never love intimately, a person they\'d never know would worship the ground they walk on.   A person who\'d risk life and limb for them. But that\'s fine.  That\'s ok.  This anger, this fiery-hot rage in the bottom of my soul will see me through.  When the lard is off, fuck the world.   I want a T-shirt with a fat pic of me on the front saying, \"To those who rejected me when I looked like this Go to Hell.\"  I really do.  That is my deepest wish.     God I hope there\'s sex in heaven.  I sure as hell ain\'t getting my share down here. '),(21880,3955951,'12,August,2004','2004-08-12 00:00:00','   Sometimes I don\'t know where to begin. I\'ve NEVER had a relationship where I loved someone so deeply that I couldn\'t live without them AND have them love me back the same way. One-way romances for me. I wish I knew where to look for someone. People are alone all over the country. Why can\'t I meet someone who just blows me away? Isn\'t it my turn pretty soon? I\'m just afraid that if I don\'t find an attractive woman who\'ll accept me the way I am now, after all the struggle I\'m going to go through losing all this weight, I\'m going to be a callous as those that reject me now. After doing all that work I\'m going to feel a sense of entitlement woman-wize. Don\'t come to me with your kid problems, I\'m out the door. Anything less than I visualize a woman to be figure wise, forget it. I\'m afraid that I\'ll exclude just like I have been excluded. I\'m going to go through hell to change myself for the better. Nothing, make that ABSOLUTELY nothing less than the ideal woman for me will do. No tradeoffs, no compromises. '),(21881,3955951,'11,August,2004','2004-08-11 00:00:00','   I don\'t know if anyone from the board actually came over here or not.  Looking back on it my decision to link it to my profile there was wrong.  This is a place where I can bitch and moan about my love life and life in general. It can get downright ugly at times, but I feel I must let it out.   It would truly poison the atmosphere if it were to be posted at another place.  I have NO animosity towards anyone there, and if that was the impression someone got if they happened to wander over here I am deeply sorry.  You\'re not going to see porn pics here.  No pics of people screwing; not even of me.  No erotic stories, no love poems.   Just little \'ol me ranting about my feelings of being alone in this world and having no one that I truly love. '),(21882,3955951,'11,August,2004','2004-08-11 00:00:00','   I know I shouldn\'t expect to be the life of the party.  Hell afterall I\'ve been conditioned to \"be realistic\" in life.  A soft-spoken fat guy doesn\'t have much notoriety unless it\'s odd looks or pig calls behind your back, being rejected by women.   As God as my witness I\'m going to lose 100 # and get the hell out of my marriage.   But the thing is I can be as cruel as those that rejected me or wouldn\'t even consider me as a lover.  Fuck them.  It just plays into my cynicism about the world.  A pretty face gets all the attention whether they\'re as dumb as a fucking pile of hog shit or not.  You know, I\'m kinda glad this blog gets no attention, no comments.  Gives me a chance to say my real feelings.    If I were a reasonably attractive woman or a hot guy people would throw themselves at me.  Cry about my problems and someone would want to dry my tears. Not me.  No wonder I\'m a loner.   Whatever I have in life no one has lifted a finger to help me.  No one showed me the path I should take.  I\'ve done it ALL BY MYSELF.   No wonder I\'m so good at masturbation. '),(21883,3955951,'10,August,2004','2004-08-10 00:00:00','   I seem to be in one of those moods again.  The realization that I\'m in a marriage that is not in the least just hits me like a ton of bricks.  I\'ve put in nearly 19 years with this woman and I don\'t love her.  She\'s not a bad person.  Quite the contrary she\'s a very good person.  But I cannot find the least bit of sexual desire for her and haven\'t for a long long time.    I just cannot forgive myself for settling for this kind of life.  I\'ve wasted all these years with someone who\'s not more than a roommate to me.   I\'ve been addicted to a new message board and post there quite a bit as a way to release some of this tension I feel.  It\'s theraputic in a way to reveal things to people you barely know from the \'net.    The frustrating thing for me is the fact that some there have met and actually fallen in love.  Wish lightening would strike me like that.    But what bugs me is the fact that since I\'m not a woman in a shitty marriage, or a real cute guy (I suppose a 325# man is cute to some) I don\'t feel anyone there reaching out to me.  I know people message each other and form private offline friendships and affairs.  One woman in particular has written about her marriage being on the rocks for some time.    She\'s very attractive and a heart of gold.  So I know whatever guy she fell for on the board was undoubtedly one who saw that she was hurting.  He saw a vulnerable woman and formed a bond with her.   Nothing wrong with that.  But what makes me such a leper?  Why won\'t anyone reach out to me when I\'m hurting?  Vulnerable fat guys just don\'t make any lady diddle their clits.  Now if I was a pinup or GQ looker, I could be silent and brooding all I wanted and women would be PM\'ing me nonstop, if only just to talk.  No fucking wonder I\'m a loner.  I\'ve got a universe of love to give and no one to give it to. '),(21884,3955951,'10,August,2004','2004-08-10 00:00:00','   Well, Sunday out of the blue I got laid again.   \"Rita\" was another woman I talked to from the web.  We were just emailing back and forth getting to know each other a little when she suggested that she wanted me now, as in RIGHT now.  I said sure, I\'ve got a bed here at work and when I\'m done we can go at it.   She showed up  a little after 5 and we kissed passionately.  I pressed her against the wall dry-humping her hard, grinding my hips into hers, showing her the force of my thrusts. But that kiss was the last time she even looked at me.  When we were in bed together I had her cumming quite a number of times, orally, finger-fucking, and tit massage as well.  Even when I mounted her she had her head turned to the side, refusing to look at me.  Oh well it was just like fucking a whore, cause that\'s what she was.  No emotion in that cunt at all.  Hey it was an experience, a spur of the moment type of thing and in that respect it was arousing as hell.  Just meet and fuck.  Her fucking loss. '),(21885,3358017,'20,May,2004','2004-05-20 00:00:00','   I have been blessed by having good friends who have been there for me.  I would hate to think what I would be doing if it weren\'t for the good people in my life. Happiness fills my heart when I\'m with my friends. The company reminds me to continue down on the goals that I want to achieve and encourage me to keep going whenever I have fallen. The group that would help me when I would come to a dead end.   Did you ever notice how much a person\'s life changes when you pay them a compliment? It\'s the small things that make their day better.  It\'s also a good way to forget past faults and errs. Did you ever hurt someone in the past and wish you could have done anything to take it back? I\'m sure we\'ve all put our foots in our mouths sometimes. Mending fences is sometimes hard to do. To keep in contact with people is even more difficult when you lose common ground.  Friendship is a two way street and it can be lost if you take the wrong road or don\'t pay attention to the signs of when your losing it. However it can be salvaged. My mom called some friendships a patchwork quilt. Like a blanket, friendship could be a nurturer to some, and comforting.  The blanket (meaning the friendship) may be well worn or it could be brand new, or so badly abused that it has a few holes in it, however if your take out the bad parts and mend it together, you would have a quilt where all of the memories are positive and the bad ones are forgotten. Today I am going to make my own patchwork quilt of friendship and remember all the good memories that I have with my friends and take time to thank them personally for being a good friend.  '),(21886,3358017,'16,May,2004','2004-05-16 00:00:00','   Where the road would take you? How far you would have to go to reach that destination? Did you ever go somewhere without thinking about where the path would take you? And what it would lead you to?   This is what my life has felt to me for a while. I\'ve come to a place where I am in a direction that I don\'t want to be in, but I\'ve put in so much effort into what I have accomplished that I feel that it is necessary for me to keep going so that I can achieve more in my life. Work has been difficult, I have always chosen the harder path when it came to it. Spending extra time at places when I didn\'t have to, making sure that everything was just right, so that I could reap the rewards. But the rewards aren\'t coming.   So I am in a transitional stage with work right now, that\'s okay. I want to get out of my job. It doesn\'t pay me enough to make it beneficial. I never knew that I could put myself down into such low degrading work, and something that made me feel I have so much more potential. There are so many other things that I want to achieve.  Did you ever go back in your life and regret certain things that happened in the past? Like if only I did this, then maybe....   Someone would be around today  I would have an ideal job with a better education  I would support myself with a good income  I would do something I love.  I would look better and take care of my health.  But then we are told that certain things were meant to happen so that we would learn from the experience. However, we can change what is happening in our lives at any time that we want to. So what\'s holding me back from doing what I want to do? Life itself.   Wouldn\'t it be great if we could change out of our crummy jobs?  '),(21887,3495545,'31,May,2004','2004-05-31 00:00:00','   well, i\'ve been skipping work for several days now. but today was different. I actually went to office plus got some work done too. But lets see whether i can maintain this. Now that\'ll be something to brag about. smaller victories mean nothing, its the big ones that count. _______________________________________________________________________________________ one more achievement.. i\'ve been trying to start watching Smallville for some time now. And luckily today i have started watching it.. seen the first two episodes of the first season. Was something really worth seeing.   All in all.. had a real good day today.. . but i\'ve been better.. and those days will never return i think.  '),(21888,3495545,'30,May,2004','2004-05-30 00:00:00','   urlLink    a little something i did in 2001. maybe i was mad at someone.. or maybe i was gonna kill the PC..&nbsp; urlLink  '),(21889,3495545,'30,May,2004','2004-05-30 00:00:00','   urlLink    fishing garbage: (c)2004.lonu - 31/5/2004&nbsp; urlLink  '),(21890,3495545,'30,May,2004','2004-05-30 00:00:00','   urlLink    march,13,2004. location:ha.hoarafushi&nbsp; urlLink  '),(21891,3495545,'30,May,2004','2004-05-30 00:00:00','   well, the day was done for.. went to sleep.. (even forgot to blog before going to sleep). As usual, kept twisting and turning in my bed. I just had to do something. So, just turned my room upside down. Place looks better now.... like a junkyard.  '),(21892,3495545,'30,May,2004','2004-05-30 00:00:00','   well, nuthin much happened.. burned a few CDs, watched some Shrek2 and mailed my blog url to a friend.. the day is going pretty much boring i guess.  '),(21893,3495545,'30,May,2004','2004-05-30 00:00:00','   urlLink    bat&nbsp; urlLink  '),(21894,3495545,'30,May,2004','2004-05-30 00:00:00','   its late afternoon and i just had a coffee (well.. was cold chocolate with sandwiches) with a close pal. came back and sat infront of the PC. read some blogs and my hands did the rest.. ended up creating an account of my own and start bloggin... cheers.. i\'ve blogged ma first blog!  '),(21895,3495545,'05,June,2004','2004-06-05 00:00:00','   In the distant future...  With technology spreading across the internet, information sharing had a sense of value and importance which enabled easy communication. However, there were a number of cases in which there were communication problems. Crimes committed by network hackers began to spread onto a worldwide scale due to the disperse of secret information. Will we be able to keep track of where this is going?  One such case happened on a certain day when the stock prices repeatedly jumped, which lead to depreciation in the New York Stock Exchange. The hacker who was responsible was detected by the server through the transaction, but he has yet to be arrested. The establishment was under the United Nations WNC (The Worldwide Network Council). The danger in the network still continues to increase.  And it happened suddenly... In the year 2005 the main computer that was connected to the network for the entire world, stopped working on December 24th. The communication controls and the other networks connected experienced a network collision in a worldwide scale. After this incident, the network recovered its performance 77 minutes later. The automatic retaliatory system of the United States began to operate erroneously and was dangerously moving to a point of crisis, which was connected to a terrestrial existence. The \"Pluto Kiss\" virus which attacked was created by a mere 10-year-old elementary school student...  Among the few, the \"ALTIMIT\" OS was not affected by the \"Pluto Kiss\" at all. Eventually, the WNC announced that the majority of the network servers of the entire world will switch to the ALTIMIT OS, in 2007, 2 years after the incident.  In addition, the restriction of the network utilization of entertainment system was removed in the same year, December 24th. Simultaneously on the same day, the MMORPG \"The World\" was sold. The United Nations declared the safety of the network and \"The Mother Mary\'s Kiss\" was named and was known as an international holiday.  And on December 24th of 2007... 0 o\'clock and 0 minutes, the worldwide simultaneous download of \"The World\" began. \"The World\" will bring disasterous incidents years later, without anyone realizing it now...  '),(21896,3495545,'02,June,2004','2004-06-02 00:00:00','   yaaaawn!!! so sleepy  '),(21897,3473164,'29,May,2004','2004-05-29 00:00:00','   So, I got a little serious yesterday, and part of me wants to apologize-- probably the same part of me that is scared sometimes-- but I am not going to apologize.  Apologies, fear, and lack of faith in the individual are to be avoided.  Well, not all apologies, but... apologies for being serious and committed and a visionary.  You shouldn\'t apologize for that.  So there.  Just  urlLink breathe .    Today is my last day selling  urlLink MINI Coopers  for a baseball legend.  I never thought I would sell cars, and I never thought I would work for Hank Aaron, but now I have done both.  Life has a funny way.    Selling cars has taught me a good bit, including how to drive a manual, which makes me feel more competent as a human.  I have been driving a MINI, and now I have to give it back, which makes me sad.  My  urlLink real car  doesn\'t have airbags or a CD player and only the left-side speakers work, so I will be unsafe and without good tunes.  Here are some heroine hints to buying a new car: 1.  Say no to dealer prep fees, bank fees, and generally any other fee that is not included on the Munroney Label, the official label from the manufacturer that by law must be displayed in new cars.  These fees can be negotiated.  Now, most dealers do have a documentation fee, and here at MINI we never waive it, but there might be some negotiation at other dealerships. 2.  There is generally flexibility in the finance rate that an auto dealer offers you, so even negotiate your interest rate if possible.  Car dealerships want your finance business and will generally do what they can to gain your finance business.   3.  Be sure to thoroughly inspect your car when it is delivered.  Look for dings, scratches, etc.  This is not to say that they will be hidden from you, but sometimes even a dealership will overlook something, and it is better to take care of it before you leave.   4.  Do your internet research.   urlLink Edmunds  is a good place to start.   5.  Look at several dealerships.  Comparison shopping works well in the car business.   6.  Be nice.  Remember that car sales people are people too with families to feed, or in my case a cat, and for the most part they are just trying to earn their keep.  Most work on 100% commission.  7.  Buy a  urlLink MINI  or  urlLink Scion XB .  Maybe that will be of some help to someone, even though I don\'t think anyone is really reading my blog.  Which leads me to ponder how does one force people to read their blog?  Okay, well maybe forcing isn\'t really the right verb, encourage, how does one encourage.    I am feeling a bit nervous about being unemployed.  I will officially be part of that faceless 5.6%, but I will go voluntarily.  I was supposed to talk about Israel today, but I am busy trying to figure out how I am going to celebrate my last day at work.  Figures, doesn\'t it?    Maybe I\'ll  urlLink krump .  Watch a  urlLink movie .  '),(21898,3473164,'28,May,2004','2004-05-28 00:00:00','   Peace concerns me because I do not think that people should blow-up, shoot, maim or otherwise hurt each other if they are in search for solutions.  Death is not a solution; it is a sentence.  And, no one has ever been able to explain to me how violence solves problems.    The obvious answer:  \"Violence solves problems by killing everyone who is a problem and scaring everyone else who is thinking about becoming a problem.\"  Interesting proposal, but it has never and will never work because there is always someone who is above fear, someone who is willing to die, someone who doesn\'t give a poot, and someone who fills deeply wounded by previous violence and therefore someone who seeks retribution.  Hasn\'t anyone watched \"The Godfather\"?  Violence begets violence.  What are we supposed to do?  Sit around and wait to be attacked?  Well, here is an idea I have been kicking around.  Attack and death are inevitable as there will always be someone who does not value life. There are two answers to this sad reality.    One:  Go out and try to kill everyone who might try to kill us first.  Problem:  In the process of trying to kill all of the evil-doers, we kill some innocents, and these innocents have friends and families.  They are people that do not believe in violence and probably would have never taken up arms IF their loved one had not suffered needlessly.  After the death of their loved one, they become filled with anger and seek retribution creating a new threat that we (of course) must confront.  We (being America) send over our guys and gals in the army, and they heroically fight for our freedom.  Another group of innocents lose their lives (albeit voluntarily) and back home loved ones blame the faceless enemy.  Where does this cycle end?    The cycle breeds more violence than the status quo.  Regardless of what happens, there will be people who hate and people who kill, but meeting their violence with more violence does nothing to prevent these people from existing and actually increases violence because it transforms non-violent individuals into fighters.  Result:  We have the violent status-quo plus the violence we have bred with violence.    Net loss  Two:  Do what we can to prevent violence through non-violent means and educate.  Problem:  The argument goes that there are people out there who hate us and will stop at nothing to ruin our way of life.  They do not care if you are Democrat or Republican, peaceful or hateful; they want us all dead.  I grant that.  I concede that these people exist, but they will always exist regardless of what we do, and if there is any chance of changing them or making sure that their beliefs are not passed-on that chance does not lie in blowing up their countries and killing their citizens.  Such violence only supports their belief that we are dangerous, hateful people who need to be stopped.  Attacks provide fuel for their fire of hate.  Attacks win new recruits, recruits that previously were on the fence or did not want to be involved in the violence at all.    People say that war is the price of freedom; the death of a soldier is the price of freedom.  That I live in an ivory tower and take my freedom for granted while the sons and daughters of America sacrifice their lives for me.  And I say, wait one minute.    If I could bring back every soldier that is overseas, I would.  I do not want them there.  I do not believe they should die for my freedom.  I would rather take the risk of dying here because they were not there than have them assume the risk of dying there.  Simply:  I would rather die here at the hands of a terrorist because my country adopted a policy of non-violence than to selfishly demand the sacrifice of others.  I will die for my freedom, but I will not seek my death and I will not ask that others die for me.  Violence is not the American way.  We do not condone violence in lieu of the democratic process.  We do not condone gang warfare.  We do not condone preemptive strikes.  If we did, we would live in a \"Minority Report\" inspired landscape where people who were prone to commit crimes would be locked up.  Don\'t we believe in innocent until proven guilty?  So, how can we say that we are fighting for the American way?    Ahhhh.  I am tired.  There is more.  There is always more.  I didn\'t even get to the part about Bush and Israel or about education.  I was watching  urlLink \"The Dennis Miller\"  show last night (What kind of website is that anyways?  It isn\'t even a website.), and a member of the panel said one reason he is voting for Bush is because he has been the most Israel-friendly president in recent history. Israel, another interesting topic when it comes to peace and the American way.        I wish I had some links for my thoughts.  It looks more exciting when there are  urlLink colors .                     Maybe, tomorrow.  Hopefully, forever.  '),(21899,3473164,'27,May,2004','2004-05-27 00:00:00','   There is this little part of me that wants to post a blank blog and that little part is fear.  Posting my thoughts, my ponderings, and rants seems a little presumptuous.  The act assumes that there will be a reader, and I am not really sure who would want to read.  Then, there is the other part that screams to be heard and demands she will be read.  I recently read an article about \" urlLink Blog Lit \" and I was left thinking, \"Well, why not?,\" which seems to be my latest approach to life.  At the age of 25, I am planted firmly in a \"quarter-life\" crisis.  The title again presumptuous as who is to assume that they will live to 100, but I did not create the label, but the category has created a new market segment.  And the segment has  urlLink a book, a website, and now a conference  .  So, the quarter-life crisis.  I haven\'t read the official diagnosis, but for me it means floundering about searching for a passion, scared that my life will end without accomplishment, working half-heartedly at jobs I feel are meaningless, and beating myself up for not being the 25 year-olds who organize  urlLink Bonnaroo  or the 15-year old who started her own internet consulting firm.  (Her name is  urlLink Laima Tazmin .)  Obviously, if greatness hasn\'t found me, then I am not meant for greatness.  That is what my quarter-life crisis dictates.  I have quit my job as a Motoring Advisor, aka MINI Cooper Salesperson, after quitting my jobs as an investment banking analyst for Lehman Brothers, a waitress for the Atlanta Fish Market, and a nanny for an unnamed family in Buckhead.  Perhaps my greatness is my ability to quit.  Somehow I don\'t believe that will garner great accolades in Horatio Alger\'s homeland.     But, I am here, and I too need to be heard, or perhaps I just need to be written.  And so it begins, the ponderings of a self-proclaimed heroine who worries about world peace, the rise of women, the fall of children, and how one becomes a Top 40 rock star.  '),(21900,3868312,'12,July,2004','2004-07-12 00:00:00','   Of course the first week went by... not juz wnet by but with events that made me realised that teaching is no longer teaching.  Firstly, 34 periods as a begining teacher. Really heavy. More than double when compared to my practicum dayz. But i took it in a good stride. I took it more of a challenge and I have to live up to it.  What I am even more surprised is that, I have to work more than 15 hrs a day. Is this crasy or what. The first week, the only day that I managed to leave sch early was on fri, which was to go for fridy prayers. This is because on the 30 and 1 of July, there were induction programmes for the new teahcers. Thus I was not able to prepare lessons for my coming day and I have to continue work at work at home till 12in night. This cycle stated on tues 29 of june where the teahcers had a contact time. Thus my work load on tues was carried on to wed and wed to thurs and thurs to fri. I feel unfair to my students because I personally think I could have prepared a more interesting lessons for them.  But then again when i think abt it, the induction programme could have been carried out much ealier or during the holidays when we came back for the workshop where the entire induction programme could be squeeze into one day.  While, I enjoyed my classes. I love to go to class and present my lessons and see the kids literally grwo in terms of their knowlege and moral development. Thus I realised why there is a high percentage of teachers leaveing the work force because teaching is no longer teaching. There are lots of admin stuff which begining teacher like me have to managed.   Fortunately I have a very helpful department which keep in me check with the stuff that is going ard in the school, like booking of the AVA room or the computer lab. Despite all the heavy first week, I look many new things which have kept my teahcing passion alive. One of them is the love for the students. Seeing the ability to see them grow is something I feel no amount of money can buy.  I guess i need more time to get mysekf acustome to the sch policy and work culture. We see how it goes in the second week then. Till then.. signing out wery and still hopeful.....  Begining Teacher in AMKSS  '),(21901,3868312,'08,July,2004','2004-07-08 00:00:00','   The first day went by...  '),(21902,3937692,'15,July,2004','2004-07-15 00:00:00','    hey all hows everyone doing   heres my testimony   Hello everyone my name is alice i was born in london ontario. when i was 2 my dad left my mom I never heard from him again. My mom had schrezphrenia was unable to look after me so i was put in a foster home. Then i was adopted when i was 4 by my grandparents and moved to timmins. Growing up my grandparents were christians although they never told me. They never really talked about their faith. I grew up as a catholic. I went to sacred heart church and was baptized when i was 12 there so i could attend sacred heart school as my grandparents felt it was a good school to get into. I also went to church at sacred heart church and i would listen but never really felt like i belonged there. Me and my grandpa would go from church to church trying to find the right one we never seem to settle down in a church. We went from church to church for a while and then finally we gave up altogether going to church because we couldnt find that right one. I was never a happy child growing up as i was teased alot.I was born with epilepsy and grew up having seizures all my life. When i entered highschool it stopped and then came back again when i graduated. After highschool i started hanging out with the wrong crowd getting involved in fights getting in trouble bars drinking etc. It was a life i enjoyed at the time but was making me really depressed. Then after that i started to work at lords kitchen a place for people who have no food to come and eat weekly. It was there i met a guy and we started dating. It was pretty good for about a year but then he never made time for me anymore so i ended it . He was my whole life , nothing else mattered to me thats why when i feel the relationship ended.  God showed me that i needed to change and that wasnt the kind of relationship he wanted me to have and not a good way for me to live to depend on someone else for happiness. We broke up after a year together. At that point i felt my life was falling apart, i wasnt getting along with my grandparents as we were always arguing i was in special olympics at the time and i wasnt getting along with any of my friends, so i felt like i needed somebody i dont know where i came up with the idea to do this but after the breakup and the loss of friendsips and the not getting along well with my family i went to my apt said a short and asked jesus into my life and prayed that he would help me sort out my life. Not long after my grandma called the baptists church looking for christians groups to hang out with thats when i found out about c&c i started going there at bfa. Then i started looking into joining a church would go regularly between bfa, faith community trying to decide between the two which one i would attend. Then after a while i finally decided on faith community but wasnt really sure that i wanted to a member right away and wasnt really ready to be baptized yet as i was still sorting out my life and taking one day at a time not really making any commintments. So i started attended faith community and going to college and careers it was going good i was meeting new friends praying and doing bible studies i was learning a lot about god and the christian life. My mental health wasnt doing so well. I wound up twice in the hospital because i was feeling very suicidal and it was god who kept me from doing it getting me into the hospital before i did something to myself. The second time i wound up in the hospital some friends came to visit me when i was there. They gave me some tough love, which i never had before and aksed me a lot of questions to make me look back and reflect on my life and what i was doing with it. It gave me a lot to think about. As at that point in my life i had nowhere to go the drs and councillors didnt want me living in a homeless shelter as i left home looking for an apt and they were looking into places in which i could live , which was even more things i needed to decide i had to make a lot of decissions and quick so this is the time when i needed god the most, and when he made himself very clear in my life and helping me out in the most difficult time of my life. After my friends left the hospital i spent my week in the hospital praying and reading the word trying to find the answers and solutions i needed to fix my life around as i was so unhappy with the way things were going and i needed to do something drastic to change my life. I did a lot of reading in the psalms that week as the ppl in the bible were all in down times and in depression which is what i was going through so i could relate to them. I knew that i was tired and didnt wanna end up in the hospital it was not how i wanted to live. I wanted to be happy and joyful and live for the lord. So from then on i decided to change my life around to try better to get along with my grandparents not to hang around with the ppl who got me in trouble and to live life for god. I also decided to leave c&c for awhile even though it was a good group i felt i needed time alone to heal and get counselling and come back when i felt i was ready. I took about six months off i really got the chance to heal and also had a weekly hangout with some ppl just to keep in touch with some friends i had made from the group since i couldnt see them at the group for a while. I\'m glad that i had that time off really felt like i needed it . I\'m back in the group its going very well, on the right medication doing well with family and friends and living my life for god. It was a hard thing to go through but i learned a lot and all the time god was right there with me and never gave up on me even though i gave up on myself. '),(21903,3700178,'22,June,2004','2004-06-22 00:00:00','   \"I like to keep my mind limber by a strict regiment of drugs\" he said.  \"I just don\'t understand why you just dont quit\".  One thing I\'ve learned in my experiences with drug people and drug culture we cant be convinced or tricked even its a realization we come to alone or not at all.  I\'ll go to rehab but when i get out it\'ll still be there.  Unless i wanted to go there its not going to do anything for me.  \"i just cant handle the fact of thinking that someone else is somehow controlling any part of my life in any way\" \"you cant even do it for me\" \"no darling im sorry i cant let someone else reign over my destiny\" I have always had a problem with controling people and as a consequence i have had quite a problem with authority figures in my life.  Held up and oppressed by their fascist rules.  My step-father the worst offender.  It turned my life into a vicious cycle of being turned out on doorsteps.  I would think oh itll be better at this place and find myself not even being able to handle the most miniscule of rules.  Then it would start all over again with me testing the waters at my newest place.  I finally realized i needed a solution to this problem when i lived with my uncle for around six months funny finding clarity where the murky water originated.  His parting words of wisdom dont turn your mind into shit.  I spent my first week in an orgy of marijuana mixed with any other drug i could possibly find.  Then i decided it was time for my solution get my own place and live life by my terms.  I went to the medicine cabinet and popped four aderall its an attention defecit disorder medicine when taken in bigger doses it gives a rush like speed.  No where near as pure and long lasting but still enough to clarify my mind.  Now that my mind was clear and not bogged down from its usual concotion of alcohol and marijuana i decided now was the best time to put in job applications.  I went down to a local pizza place sort of like chucky cheese.  I walked in smoothly and unhindered my limbs so full of life they felt like they were floating.  I walked to the register along the left side wall and asked nicely for an application, yes total control i had total control over everything in this situation.  I knew exactly everything i was doing, filling it out was a breeze and later i went over to the taco bell better chances with multiple applications.  Taco bell was more of a last resort i had worked there before and bitterly hated it at that.  Survival was more important than my pride.           work in progress... '),(21904,3700178,'21,June,2004','2004-06-21 00:00:00','   no site is complete without qoutes  i believe in a prolonged duration of the sense - jim morrison  to tell the truth i dont ever remember being born it must have happened during one of my blackouts - jim morrison  if the doors of perception were cleansed everthing would appear to mas as it is infinite - william blake  who is the happier man? he who has braved the storm of life and lived or he who has stayed safely on the shore and merely existed - hunter s thompson  the possibility of physical and mental collapse is now very real no sympathy for the devil buy the ticket, take the ride - hunter s thompson  we could do that, but it would be wrong - richard m nixon  the best laid schemes of mice and men go aft astray and leave us naught but grief and pain for promised joy - robert burns  the more experience and insight i obtain into human nature the more convinced do i become  that the greater portion of man is purely animal - henry morton stanley  man given free will questions trys to better himself animal accepts which of us is truly free  man in the constant battle for perfection or the animal who simply reacts man is always trying to be the best, and failing the key isnt winning or losing its making the attempt the man is what i ought to be sure its scary but whats the alternative stagnation a safer more terrible form of death not of the body but of the spirit an animal knows what it is and accepts a man may know what he is but he questions he dreams he strives changes grows - chris claremont '),(21905,3700178,'21,June,2004','2004-06-21 00:00:00','   some things i wrote when i was tripping  the clock on the wall is frozen boredom knaws at my mind an uneasy truce quivers the silence to be broken  explosions and laughter all around yet i cant put the past away take the alcohol lure your mind to forget leave it behind better off without it  is home the place where you spend your time live sleep and eat or the place where your soul is at ease comfortable in your chair  or in your head your turned away from home your soul is left to wander comfort is no where to be found  and the past never dies suicide crossed my mind aint worth the time scarred for life  with gone by mistakes  think therefore i am but what if i forget? if everything is taken from a man a man he remains?  and i drew alot of pictures '),(21906,3700178,'21,June,2004','2004-06-21 00:00:00','   i have a picture of you it rests on my dresser my comfort in dark times when i have nothing i remember ive got your face why or where did you go i remember the good times late night discusions instead of two people we were as one but you vanished now im left with  this picture of you and memories just to hear your voice  once more please '),(21907,3700178,'21,June,2004','2004-06-21 00:00:00','   remember the good times forget the bad dont forget the past never goes away  in this time of sorrow dont you be sad your future is coming be on your way the ride is over your finally free now you call the shots you reign destiny time to construct your life the way youll be this school always to be a part of me '),(21908,3700178,'21,June,2004','2004-06-21 00:00:00','   night has come, a time i look forward to my dreams once haunted by foreboding doom your visage now holds sway the devil slew you cast demons out from my mind of gloom revel in your love once again resume if but my dreams were true id be cured i wish you would join me in my white room freed from past sins, there no longer endured there we come together fatefully lured i willl hold you close never to let go and there stands out love eternally moored  id die for you no better cause i know so i ask of you this healing favor  help me regain myself, be my savior '),(21909,3700178,'21,June,2004','2004-06-21 00:00:00','   hearing laughter as it courses my stream dosed not for death, to mess my mind inside hapiness to float off this induced dream full of this dexmethorphan hybormide nothing really matters, after it hits is expansion of the mind truly crime leave them to swirl in there unlighted pits i wont abide them not me not this time live the lie horatio alger pure my own espitamological choas it suits me just fin, as you know im sure '),(21910,3700178,'21,June,2004','2004-06-21 00:00:00','   i love you so much i dont care you know thoughts of you plague me often, been too long the thing i miss most, that would be your smile the time of my return is coming fast our tears will be that of a crocodile together we will reconcile the past although im sure it will need some sutures and i have absolute faith in your hand in the coming days i see bright futures in your favor once again i will stand in this new found love i hope to smother for id smile to die for you my mother '),(21911,3700178,'21,June,2004','2004-06-21 00:00:00','   along the road side an angel we saw hitch a ride to the edge of dusk we asked tears streaming revealing major flaw on our way we went, her fading to past for this wounded beauty a strange yearning the money meant nothing i wish id lost just to hear your name my desires burning and unfourtanetly i won he lost as we ventured deeper into the dark thoughts return to wary angel, the lark forget it let the memory flatten if you ever find a love breif unknown youll know what i mean secret your emotions shown '),(21912,3700178,'21,June,2004','2004-06-21 00:00:00','   CRUSH dear sweet secret love your eyes always take focus nothing else ill view your smile draws one on my face wished courage to speak to you  Desert soul come view desert souls  strange and decadent they are exiled to this place coming soon ill return home desert still holds a part of me  Army fuck the gay armyy alot of retarded guys biggots from the south there job is to defend me i would never join that shit  Marc marc has funny hair he says fuck you, you faggot and he likes his mom  incest his favorite pass time hes a good freind to me though  Danielle shes a card player even though shes a cheater my job escort her her favorite game is hang man shes the best friend ive got '),(21913,3700178,'21,June,2004','2004-06-21 00:00:00','   how it may feel you may one day wonder to have your heart ripped from your very chest your soul left to wander no place to rest and whos responsible for this blunder  your hearts been stolen a pirates plunder the last youd expect, the close tricksters best no concentration makes it hard to pass but you can handle this strange decadence youll want to explode, but try not to cry next youll be found thinking in tranquill grass youll find yourself thinking this doesnt make sense maybe youll fail but you still have to try '),(21914,3700178,'21,June,2004','2004-06-21 00:00:00','   ny forbidden long forgotten secret love vivid dreams of you enter my mind this night fantasy appearance of you the dove your love upon me as if some strange blight destiny set from the fateful crossroads her looks a last imploring entreaty my heart yearns for you in loads only desire to see you and you me beauty reminds me of a blooming rose i long for your voice in the midnight hours and on the next morning when i arose i returned to you bringing flowers our many wrongdoings formed to be piled the forbidden love to be reconciled '),(21915,3700178,'21,June,2004','2004-06-21 00:00:00','   oh,to return to the place where i belong showing an act of painful contrition necessary suture of civil wrong internal wounds with pin prick precision on this apparition ill seek ammends the past left no shortage of transgressions a chosen retribution to transcend hearts desire acceptance of confessions day is coming very soon you will see when the suns risen i rehearse no more time to offer my dilatory plea nearing the end of this internal war my effort in this matter unsurpassed at the end ill crucify my cursed past '),(21916,3700178,'21,June,2004','2004-06-21 00:00:00','   although it pains me to do so i have to cut down on some weight before my return trip and i figured the best place to start was with all the fucking stacks of papers in my suitcase so i decided id publish those things on here so here they are for your viewing pleasure i hope you enjoy em kids '),(21917,3700178,'21,June,2004','2004-06-21 00:00:00','   so um i got one of these online journal things hunh so i can write whatever i want in this bitch and the only thing that comes to mind right now is bud light and velocihyenas so thats pretty cool right yeah i know it is bitch well if youll excuse me i have a certain marlboro light calling my name un hunh and dont you forget it either '),(21918,4321513,'24,August,2004','2004-08-24 00:00:00','   I hate expectations. Today, in first period, we played Jeopardy! (a cheap version that tests nothing, nor that expects the students to even hear the entire question) Anyway, last time around-we play before each test-I answered almost EVERY question for my team, thus earning all ten of us a +1 on our test. (I chose not to take the point, but oh well.)  So today, the same people who received points for the (continuous) raising of my hand last week, EXPECTED my to repeat. What they didn\'t know, the poor souls, was that I entered the classroom committed NOT to answer any questions. In fact, I was hoping to be the speaker so it would be obvious that I couldn\'t answer the questions.  But I failed. I wasn\'t speaker so I just sat and listened. My team, which since the teacher can\'t count had a 10-8 advantage in bodies, NEVER answered any questions. So what do they do when the score is 320 to ZERO? They snap at me. ME! Me who knew the answer to all of the questions that went right over their heads.  So I answered questions. A few at first, and more as the game continued. I botched one, but I must admit I don\'t give a damn. Ultimately, my team AND I brought the score up to 470-530. Then the teacher announced that he didn\'t think we were going to have time, so he didn\'t bother to offer a sixth category to choose from \"Jacksonisism.\" (Yes, that IS supposed to be \"Jacksonianism\")  So after that we go into FINAL JEOPARDY! 650-670. The question involves treaties of the 1800\'s in America. Instead of consulting the team though, the TEAM sticks me with the pen and paper and has me right down the answer. I know it (The Treaty of Guadalupe-Hidalgo, and the Gadsen Purchase) but I toy with the idea of botching it, knowing that the other team will bid to win. I didn\'t, but wouldn\'t it have been fun?  So now, I failed myself. The day started out depressing and by first period was a crap shot.     I HATE EXPECTATIONS! THEY ARE THE DEVIL!     Additionally, in Earth Science, a 7th grade level science course mandated in NC high schools, I learned that in order to do well, I should forget everything I learned in my advanced classes-biology and chemistry-since \"we don\'t care about orbitals everyday.\"  Not that it matters, I am listening to Oasis\' \"Wonderfall\" and it\'s not bad.  Today was, without a doubt, a    gray    balloon... '),(21919,4321513,'23,August,2004','2004-08-23 00:00:00','   I have failed myself...again.  I succumbed to the pressures of my contemporaries and have created this-this monstrosity of which I had for so long been so adamantly opposed.  I have nothing intelligent to say, no deep thoughts, and no greater meaning.  I have failed!  I therefore doubt this will ever be updated, however I, at this moment in which I declare my hate for blogs, find myself rejecting food, a hobby, to type.  Shit, I\'ve lost my mind!  This will, then, be for me. All poor souls who stumble upon this may be better off looking away- far, far, FAR away- deep into the shoes of the most isolated of clowns. It will be a study; a more independent view of myself. I may try my damnedest to be honest but since that is what I scope to learn I may struggle.  Again I say, If you have found this random, assorted, vile collage of thoughts, RUN!!! If you care to read on, do so, it obviously doesn\'t bother me.  Today, in case you want to know, my poor soul, would have been a    yellow    balloon... '),(21920,3997969,'24,July,2004','2004-07-24 00:00:00','   sasha has just sworn never to make a blog!!  o my!! now sasha has vowed that he \"might make a blog\"!!!!!  sofia is has more of this thrilling story at the scene....  thank you bill... yes here in st. paul there is compete outrage and chaos.  i\'m here with bob shmob, a concerned citozen mr shmob, these two clashing statements by the ever (non)important sasha, how do they make you feel?  What are the general thoughts and feelings that you have noticed in the comunity?  well prdy much everyone is apauled and kinda heart broken...that our leader would do something like that.... its just awful...i just dont know where to turn to.  '),(21921,3997969,'22,July,2004','2004-07-22 00:00:00','   im waiting.....grr michelle is taking for ever!! ok so heres the deal michelle and i are both writing on our blogs for twenty min to see who writes more in that time... stupid i know... but oh well.  i\'ve always wanted to live eather on a farm in the middle of no where or to live in the middle of the city. i have always wanted to have lots of brothers and sisters, but in truth, i only have one little brother.  His name is Sasha... well we just call him sasha, his real name is Alexander.  but he has never really gone by that.  Michelle has a brother the same age as sasha, his name is zach.  She also has a little sister, ariana.  Michelle is staying over tonight, we have already had dinner.  My dad grilled some pork chops and we also had asparugus.  michelle just told me a story about how zach once tried to pick a lock on a car with a stem type thing and it got stuck in the car latch thingy... i dont know how to describe it... oh well.  A couple of years ago (maybe it was last year), i gave my brother (sasha) the nickname sushi, and michelles brother (zachary) the nickname zucchini.  that actually stuck for a while but i dont think anyone calls him that anymore (sushi), the just call him sasha.  his screen name is something about sushi though.  My dog is being strange... whatever.  Like i said before, michelle is sleeping over.  I hope we do something fun... we dont really have any good movies to watch so i dont know what were gonna do about that.  Maybe we can go to blockbuster or some place like that to rent a couple movies... ill ask my dad later.  my dad\'s name by the way is scott, (that means of scottish origin, i dont think he is though, and if so, just a little)  he has a brother, a half brother and two step sisters.  he was the oldest in the family growing up (like me), and then there was kurt, but he died of krib death when he was about five months old.  Then the was steven, he was three years younger than my dad.  And then when my dad was about fourteen, his parents got divorced and he spent his summers with his dad and the rest of the time with his mom.  A while later his mom got remaried (by the way her name is myrna) and she got maried to ray.  He was also devorced and also had two kids, two girls, mary and margret.  so after they got maried my grandma had another baby, annother boy, robert (he is actually here now too).  Uncle rob came for dinner as well.  Sasha is being all obsessive about cleaning the \"amazingly cool\" bottle.... it is glass and it had \"pom\" in it, it was a mix of pomogranate and blueberry juice.  He and i split it a couple of days ago, and anyway he keeps putting this blue dish soap in it and it and makes the water in it blue, its really stupid, but whatever.  ------ that was twenty minutes worth of writing.  but michelle is writing something about politics so she wants to add another five minutes and i said fine.... speaking of politics... i am a hard core democrat.  i also was  into skateboarding for about two years or so.  But i quit last year, i was getting kind of bored with it.  I love to draw but im not very good unless i spend about two hours on a tiny little sketch and then you can almost tell what it is ( jk im not THAT bad- but prdy close).  Im reading a book now, that my dad is making me read.  its called spies and its by michael frayn or something like that.  its not the kind of book i usually read.  mostly i read fiction but sometimes i go for some contemporary stuff.  i really like the book sabriel by garth nix, annother book i am reading now.  I tend to read literally five or six books at a time, which reminds me that i have to get my books back to the library some how before i leave for camp with michelle.  '),(21922,3997969,'22,July,2004','2004-07-22 00:00:00','   \"Traffic In The Sky\"  There\'s traffic in the sky and it doesn\'t seem to be getting much better there\'s kids playing games on the pavement drawring waves on the pavement shadows of the planes on the pavement its enough to make me cry but that don\'t seem like it would make it feel better maybe its a dream and if I scream it will burst at the seams this whole place will fall to pieces and then they\'d say...  well how could we have known? I\'ll tell them it\'s not so hard to tell if you keep on adding stones soon the water will be lost in the well  puzzle peices in the ground but no one ever seems to be digging instead they\'re looking up towards the heavens with their eyes on the heavens there are shadows on the way to the heavens it\'s enough to make me cry but that don\'t seem like it would make it feel better The answers could be found we could learn from digging down but no one ever seems to be digging instead they\'ll say...  well how could we have known? I\'ll tell them it\'s not so hard to tell if you keep on adding stones soon the water will be lost in the well  words of wisdom all around but no one ever seems to listen they\'re talking about thier planes on paper building up from the pavement there are shadows from the scrapers on the pavement it\'s enough to make me sigh but that don\'t seem like it would make it feel better the words are still around but the words are only sounds and dno one ever seems to listen instead they\'ll say  well how could we have known? I\'ll tell them it\'s not so hard to tell if you keep on adding stones soon the water will be lost in the well   [Thanks to lordygee14@ahoo.com for these lyrics]   \"Fall Line\"  And by the way You know that hope will make you strange Make you blink, make you blink Make you sink It will make you afraid of change Enough to blame the box with the view of the world And the walls that fill the frame I turn it up But then I turn it off Because I can\'t stand When they start to talk about The hurting and killing Whose shoes are we feeling The damage and ruin And the things that we\'re doing Gotta We gotta stop We gotta turn it all off We gotta rewind Start it up again Because we fell across the fall line Ain\'t nothing sacred anymore Nah nah nah nah nah nah nah nah Somebody saw him jump Yeah, but nobody saw him slip I guess he lost a lot a hope And then he lost a grip And now he\'s lying in the freeway In the middle of this mess Guess we lost another one, just like the other one Optimistic, hypocrite that didn\'t have the nerve to quit The things that kept him wanted more \'Til he finally reached the core He feel across the fall line Ain\'t nothing sacred anymore Nah nah nah nah nah nah nah nah   [ www.azlyrics.com ]    --just a couple of songs of jack johnson that i really like... but i couldnt find lyrics  THAT WOULD LOAD  for my fav song by him: cookie jar  '),(21923,3997969,'22,July,2004','2004-07-22 00:00:00','   dame im bored... and ppl are pissing me off.  mostly my family.... grrr being stupid.  I wish ppl could just chill out insted of being all anal and uptight about everything.  all i have to soothe my anger are to poor my emotions into a computer and to listen to my brother squeak on his clarinet. joy.  '),(21924,3997969,'22,July,2004','2004-07-22 00:00:00','   the other day i went to the movie fahrenheit 911 (without michelle....SRY!!).  it was prdy good, very funny, at parts and at others it was really sad and prdy graphic.  i highly recomend it to anyone who is remotely democratic... and even if you\'re not, its still a good idea to see what the true story is.... even if it is told in a bias fasion.  '),(21925,3997969,'22,July,2004','2004-07-22 00:00:00','   I thought it would be a good idea to explain my name to those who don\'t know what the heck it means.  First off, Ozzie is my dog.  He\'s a Westie (a West Highland White Terrier) and i love him.  He is absolutely adorable. Secondly, Zlatibor: its a charming little town in Yugoslavia that I\'ve visited at least a couple of times with my grandparents.  I am half Serbian (for those who didn\'t know),and my mother grew up in Belgrade so its a big part of me.  I do speak Serbian, not very well (any more), I actually learned it before i learned english... but I kinda lost some of it along the way.  '),(21926,3997969,'22,July,2004','2004-07-22 00:00:00','   yup like all depresing storys starts with a girl, ____.  no one really knows who she likes, its imposible to know.  and if you want to find out, its as hard and dangerous as trying to dissarm a nuke. thats a leasson for all those guys out there.  peace G,s.  '),(21927,3997969,'22,July,2004','2004-07-22 00:00:00','   I am sooo excited!!  I might be going to France to be a forgien exchange student....eep!  If that doesn\'t help my problems in french... nothing will.  Well anyways, I know the family that i would be staying with, they\'ve got three little kids, a girl and two boys.  We actually went on a family vacation to visit them a few years ago.  They\'re really nice people.  I might be going as soon as second semester!!  So Pumped!!  '),(21928,3997969,'22,July,2004','2004-07-22 00:00:00','   wow.... well i thought it might be fun to make a blog... dont really know anything about them tho.  ehem.... so...this should be... well *ehem* interesting.  '),(21929,3997969,'31,July,2004','2004-07-31 00:00:00','   so a couple days ago, while cleaning my room, i found a video that we (myself and my classmates) made in 3rd grade french class.  we talked about the weather and stuff like that... o yea and we did a fasion show... haha i remember that too.  i decided to tie my lime green and black shirt into a belly shirt and i put my hair up in a pony tail on the very top of my head. and marian, i finaly got to see your hair down.  i had forgotten that you used to wear it down when we were little.  lauren, omg you were so funny, you looked so nervous in front of the camera, ill have to show it to you some time, lol.  every one looked so funny... so little, it was cute.  o yea, and i also found a audio tape that i had made when i was 11, i didnt even recognize my voice because it was so high pitched, my brothers was even worse (he was 8 at the time).  ha. i was saposed to be doing my french homework but instead i decided to sneak off with my little tape recorder and make a tape.  speaking of cleaning my room, im not really cleaning it, its more like remodleing.  it needs it too, i havent changed anything (with the aception of putting up posters and stuff like that) since i was about 9... and im 15.  you would have to have seen my closet and desk to truely apreciate the news that i have succeded in conquering my desk (i emptied it out, thats right, there is noting in it, NOTHING!) and the same is true with my now walk in closet, where as before, it was literally knee high in crap.  well thats about all the news for today.... o yea, i forgot to say, im going up to menogyn (its a ymca camp), im going to do a canoeing trip for 8 days, with michelle.  i havent started packing though and im leaving on sunday....hehe.  im really bummed though, because i have heard that starting this weekend they might outlaw any campfires up north, because of the fact its so dry...grr, i hope they change their minds... or have some kind of monsoon before we get there...  '),(21930,3997969,'30,July,2004','2004-07-30 00:00:00','   6.25 %          My weblog owns 6.25 % of me.  urlLink Does your weblog own you?     ok.... i guess thats a good thing... i mean its not 100%  '),(21931,3997969,'27,July,2004','2004-07-27 00:00:00','   yo yo yo... who ever it is who is reading my blog...  well, i really dont have anything to talk about..... let me think... -michelle is lazy -im bored -marian is prepared -im bored -theres nothing to do -im bored -no one is on line -im bored -do you get the picture?? -im BORED  '),(21932,3997969,'27,July,2004','2004-07-27 00:00:00','   ha. well my lil bro (sasha) got a battling bush punching bag....hes been quite amused with it.  on it it has bush with boxing shorts and boxing gloves (which squeak) and hes standing on top of the world. at the bottom it says BATTLING BUSH in big capital letters then to the left side by his knee, its got bush sr. (titled DADDY) saying GO GET EM BOY!  i was rather sadened when i heard that there was also a battling kerry.  '),(21933,3997969,'01,August,2004','2004-08-01 00:00:00','   well, like i sayed before, im going to go to a ymca camp in the boundry waters... actually i have to get up in about 4 hours to go.... ehem... so i best get to bed. i wont be back until sunday (the one after this one).  wish me luck!!  toodles!  '),(21934,3499629,'30,May,2004','2004-05-30 00:00:00','    yupyupyup.today was kinda busy. we went to santy (china town) to get some flowers for anne\'s deceased great grandma i think. them FOREST LAWN cemeteries are beautiful. you would think of it as a rich man\'s yard,but really,though,it\'s a cemetery.pretty crazy shit ryt thurr.   im trying to get used to this blogging shit. back to the subject,when we went to santy, we drove past a whole shitload of shop before we actually reached the block for flowers.damn, los angeles is one hell of a place. i luv it up there.even though it may seem like the ghetto kasi maganda parin yung place, kasi ang daming quality clothes na ang sarap isuot kasi uber pretty na nga,cheap pa.hehehe.i think pwede na akong i crown na queen of bargain clothing. oh hell, im proud of it. atleast nakaka save me ng money!hahaha!im listening to kanye west\'s i cant live my life this way, i love that guy!i love his rhythm and his drive.  back to the subject,then i saw some cute flowers,and as usual...day dreamed about getting them from my dreamguy.even though joe and ray flashed through my head, a couple of times, i couldn\'t quite decide who i wanted to get flowers from... i mean, who i wanted to get flowers from first... anyways,enuff na ella, your gonna go into a reverie agen. so after some thinking and day dreaming and more arguing(anne was convincing mwah to buy them already since i kept on going back to them and kept bugging her about how pretty them thangs are. ) so anne bought some flowers for herself and i did too. then on the last min she decided to give it to her stepmom instead. to make the long story short, it seemed like i was the desperate bitch who bought flowers for herself...  hey,maybe i am...i mean,does it really seem like it?hahahah.  o well...hu cares,them flowers would look so pretty in mah room! can\'t wait to put them in the vase and make my rom pretty. incase youre wondering,the flowers are pink carnations... pretty.....real pwettie...  they aint mah fave, but they were cheap and pretty at the same time so i juss went ahead and bought em.hahahah.  NOTE TO Y\'ALL WHO WANNA IMPRESS ME ON THE FIRST DATE:  I LOVE,ADORE AND ALMOST COMPLETELY VENERATE PINK LILIES AND TULIPS!THEY\'RE THE BEST! BUT IF YOU WANNA GO THE CLICHE-ISH ROMANTIC WAY,(EVEN THOUGH LILIES AND TULIPS WOULD SAY MUCH MORE THAN THAT) THEN GET ME SOME PINK ROSES, otay???  see,i\'m not such a high-maint. gal,now,am i???hahahaha.  anyways,we came home tired and watched TFC... acshalli,ive been watching that shit all weekend.  asteeq nga eh, kasi atleast updated me sa mga happenings sa pinas,ryt?  lam nio ba na si ganito ganyan ganyan ganito?!?!?magchismis ba daw???  ang haba na nito ahhh...are y\'all\'s eyes white yet???hope so,este,sorry, didn\'t know my material was so interesting(CONCIETED GURL ALERT)  BYE NA.  lotzaluvinfrommetoyou.peaceout.postcommentsifyallwant.ayt?teartheplaceup.ella.   '),(21935,3499629,'31,May,2004','2004-05-31 00:00:00','    omg!i have so much to do.i hate this kinda stuff.mah god! i dunno what to do first.anyways,gez what? i went to mah friends\' party, kc and marie celebrated thier bdays kasi today...  gez whu i saw? BS! no, not bullshit, but BS, yeah, the guy i i\\used to swoon over....hahaha....some crazy shit ryt thurr. anyways, i gotta go get some more stuff done so buhbye.  noelle\'s here kasi she needs to use mah computer siince hers is acting up on her.i interrupted nga her eh. ok,anyways,i said i was just gonna be quick,so eto na...  wish me luck...sana lang naman talaga maging maganda grades ko for graduation,noh?at sana grumaduate me...  bye peepz.prayforme.ihopeimeetpromotionrequirements.ella.   '),(21936,3499629,'05,June,2004','2004-06-05 00:00:00','    my mom came in while i was washing the cup i used for ice cream...told her about how i wanted to go kick it at jeramie\'s agen...and how alberto came to ask...she apparently didnt notice it was the same story i told her earlier,and started naggin the shit outta mah bootie.  it\'s ayt though.im frickin used to it.it\'s a frickin after work thing she does. everyday after work,she gotz to nagg about some stoopid shit or her day aint complete.  anyways,enuff of that....she brought home some filipino desserts as a \'pasalubong.\'turon,bico,ube halea,and huh??? macaroni salad???confusing but whatevers....  now she\'s worrying about what to have for dinner. honestly,im not frickin hungry. paano ba naman,i\'ve been eating all day noh...nyaha.yupyupyup.  breakfast:  a goblet full of vanilla ice cream and caramel               sauce...  lunch:  rally\'s grilled honey chicken sandwich and some          seasoned batter fries,and to top that off,          extremely refreshing raspberry iced tea.yumyum.  merienda:   goblet full of mint chip ice cream... i scream,              you scream,we all scream for ice              cream...hahaha....   dinner:    adobo daw!finally,momma decided na.            naku,she changed her mind, potstickers nalang.             i love them thangs.specially with hot soy sauce...  the spicier the better!expect to see some serious piggin out at dinner!lol.  weellll,she said na yung food daw take home from the party na this pinoy fam had... my mom works as activity director sa one of the convalescent homes in Glendale kasi,i guess, somehow,my mom got to converse with the peepz there and asked one of the guys where he went to school and crap, and it just so happens that the guy went to the same school i used to go to in seventh grade and most of 8th grade...he said din na we used to be in a class together....i remember him,and i think i had him for p.e. that year...but here\'s the funny thing,his sis knows mwah too,but i don\'t know her,ain\'t that about as bitchy as it gets??? funny too.   gezthisblogendshurr.going..going..gone.ella.   '),(21937,3499629,'05,June,2004','2004-06-05 00:00:00','    im halfway through mah cup of ice cream... the winner? *drumroll*  MINT CHIP!  after several debates with myself and after checking the freezer...i found that mint chip was the one that was more \"exposed\" kasi that was the first one i saw, and so i juss picked the box up and scooped \"some\" out. after that,i ofcourse,licked the scooper~yeppeeeeeee! isn\'t it weird how the weirdest ideas  turn out to be the best ones too?  i mean,this chick wanted to make chocolate cookies,but had no cocoa powder,however,what she did have are some choco chips,so she put them in hoping that they would melt and mix in with the other ingredients,and thus,invented the choco chip cookie...  sorry,juss had to give some trivia... anyways, im feeling the love..it\'s a quarter past 5...and i still havent heard from alberto...gez he aint kickin it with j\'mie either,or probaly fell asleep and forgot.  i wonder what happened to that exercise routine daisy and i planned for ourselves??? we were supposed to go running together..around the block lang...i gez were moving it to....uuummmmm..... whenever we remember??hahahahaha.o well.  earlier,i went to kuya marco\'s blogger,and i noticed na he has africkin chatterbox!kuya,if you\'re reading mah posts, how do you get one ba???help naman o!all this blogging shit\'s  getting to me na talaga huh....gez im not much of a techie...  i wish i was though...nyehe.  recently,blogger\'s been acting up on me...can\'t post crap, whatevers.i think i\'ve been bit by the issue bug.hahaha, kasi lately i\'ve been having some serious isshies,hahaha. or is it juss becauuse im menstrual??? oh goodness gracious... can\'t wait to go into menopause...hold up, isn\'t that  shit worst?!?? ahhhhhhh!!!!!!!!!!!!!!!!!!!!  peaceout.hopeyallarehavingabettertimeanalyzinglifethaniam.   '),(21938,3499629,'05,June,2004','2004-06-05 00:00:00','    sencha na peepz,ahh?all these shitty ass complications occured kasi eh.inis.well...mah mom juss called.was juss checkin up on mwah.mah nigg alberto came and knocked on mah window,kanina and asked if we were still go\'n go kick it with mah homie jeramie...i said na maybe...i still dunno... it was kinda awkward,i couldn\'t show mahself fully,kasi i was just wrapped in mah towel...hahaha...anyone who must\'ve saw him woulda thought na he was talking to himeself...hahaha. you couldn\'t see me kasi i was barely peeking through mah blinds.o well....i have really bad cramps.aahhhh!!!im going whacko on mahself.omg.should i still go?? my momma said it would be ayt,but said it would be better if i go later nalang,or something like that,she was talking so fast talaga.hahaha. i mean,if i go, then i\'ll be limp and juss altogether moody,but if i dont,then i\'ll be sittin at home wanting to get mah ass off the couch and would probably start pitying mahself and start hating womanhood. yupyupyup.  im sure j\'mie and alberto would understand...bec. they\'re real cool with the girlie shtufff we ahve every month,because they both have sisters,and i gez because they both suffer from my moodswings everymonth anyway... and because they\'re just great!period.  i want soem ice cream...lookie hurr... strawberry cheesecake,mint chip or vanilla???which one?  ahhhh!shit!look...costco\'s all american chocolate cake! 8 layers of chocolatey goodness!damn! and there\'s also that kfc choco bundt cake....which one???  i can\'t pick...ok...ito nalang, one scoop of ice cream, one slice of cake...but which to choose???darn it!HEELLPPP!!!!!!  stillconfused.andwrappedinatowel.hahaha.ella.   '),(21939,3499629,'05,June,2004','2004-06-05 00:00:00','    hey you guys,i know i havent been blogging in such a while na....dont worry,ill make it up to yall...ayt? i was gonna give you the sitch sa happenings in mah life,the other day,but the computer was acting up on me. i know.it sucks.anyways,ill think about how im gonna get to tell all and not write a  novel at the same time,ok??? but before that,i will step in the w.r. for a quick shower.    catch y\'all when i do.ok???peace + love = harmony.ella.   '),(21940,3499629,'06,June,2004','2004-06-06 00:00:00','    hhaaayyyy!as i awoke this mownin,i found a note next to mah bed, bob left it for me and it said were going out to dinner daw kasi birthday ng mom ko... and honestly,shocked ako... because even though i knew her bday was oon june 6th, i gez i had \'so much\' things going on.... i feel like such a bad daughter....what am i gonna give her?whahhh??????  i am so feeling so unworthy right now.... and i acshalli call myself hher daughter???  i feel horrible.  man....   imma have some breakfast.hopeifeelbetter.ella.   '),(21941,3499629,'08,June,2004','2004-06-08 00:00:00','    hhhaaaayyyyyy!i am absoluutely positively entirely and extremely tired. why? i hung out with the guys after school.... food,pillow fights and video games leaded to so much fun.... it was sooo.... cool! i feel like such a kid...  hahahahaha.....wish you all were there...  after that though,i got into a shitload of trouble for coming home late.o well.that\'s why i can\'t use the phone tonight.hahaha.  bad gurl,bad gurl,whatcha gonn\' do?whatcha gonn\' do wen dey cum fo\' u???  harhar...just had dinner...some pizza with A LOT of toppings, and a nectarine.nyehe,so much for starting that diet.  i\'ll talk to y\'all again when  do,ayt???  peace out.ella.   '),(21942,3499629,'09,June,2004','2004-06-09 00:00:00','    as of 5 mins ago, anne and j\'mie are at war... weeelllll, it was all because of my bestest friend anne....they\'re not really at war or whatevers, but anne\'s just in a crazy,,hyped up,uber  loka mode ryt now... i was talking to jeramie,ryt, and she juss rings on j\'mie\'s fone, and started telling him off and shit...telling him to hang up on me and stuff....hahahaha....iwas cracking up....gez she wants to talk to me that bad..... i feel so special  im talking to anne ryt now and she is compairing  the animal world,and the human race, and asking which sounds better, bllleeeeggghhhh....or blech.....hahahahaha  now she\' s singing mr. suave and spaghetti...pamela blah blah shit....i think she\'\'s going jologs on me....hope not....  now she\'s talking about what she wants to be wen she grows up and stuff sabi niya, magiging \'conductor\' daw siya...hahahahaha/.....  now she\'s doing tongue twisters, pasko paksiw...pasko paksiw.... trying so hard to sing it\'s freakin hilarious....  now she\'s looking for a song to make bulabog mwah  hint,hint:a very bad rendition of \'will you say that you love me\'  better talk to this byitch before she gets sent to the mental hospital...baka we would both end up there later...hahahah....  kasi she\'s crazy naman,tapos she\'s gonna drive me crazy pa ayun,eh lam mo na,crazy na nga,lalo pa naging crazy,tuloy,siguro we will share a room nalang sa mental...hahhaha....  laterz...loveyallalotza.peacepeoplez.mwah!ella.   '),(21943,3499629,'09,June,2004','2004-06-09 00:00:00','    my friend was acting all weird today.not really today-today,but i mean,juss  after school.he still hasn\'t talked to me about it,though...but its ayt...it\'s his secret to tell, and i shouldn\'t, can\'t,and WILL NOT puush him to spill the beans, kasi it\'s up to him naman eh...it IS his life ya know...  i juss wish i knew though...called him earlier,and i gez he was at the galleria playing ddr,(dance revo) and stuff,so i juss sed to myself,juss leave him alone...let the dude breathe.... and i juss tld him,ya know, i was juss wondering if you were in a better mood yet and shit like dat...blahblahblah...  o well.... hope he feels better though....   '),(21944,3499629,'10,June,2004','2004-06-10 00:00:00','    hahahaha.funny day today...i have like, 2 shirts full of autograohs from peeps....it\'s cuz it\'s the last day of school for people who aren\'t marching....thank god im graduating,and i got into the mode. so i just started asking people to write me up stuff on mah shirt....hahahahaha! im not gonna write up a whole buncha shit today cuz my wrists hurt!both of them... it\'s cuz i signed people\'s yearbooks and shirts too, so yah...  datsit.goinggoinggone.ella.   '),(21945,3499629,'10,June,2004','2004-06-10 00:00:00','    hhhaaayyyyy! I LOVE FOOD! im trying to make a great salad. it has spring greens,(mixed) chicken potstickers,and a home made thousand island dressing i made myself....so good!y\'all want some???  imma enjoy myself muna,ok???   laterrz.byeah.ella.   '),(21946,3499629,'16,June,2004','2004-06-16 00:00:00','    anyways,i totaLLY chickened out on the rides... but i had to go anyways,and i did!i went on extreme scream.....it\'s like the worst ride in the park!you\'re like hanging off your chair,with only some belts and this thingy supporting you.and also,youll feel like being freaking tossed into freaking space...like flying....except you\'re going about 150 miles per hour!shitless!that\'s what i was! fudge!  and it\'s hard to scream when your being switched from lyting down tobeing switched over,head down,feet up,and juss plainhysterical noh!  the first time on \'x\' was with samantha, and i noticed this hot ass...but we weren\'t in the same batch nga lang, and i was too busy convincing sam i wasn\'t brave enuff for this ride...blahblahblah! anyways, we got off the ride... blahblahblah...so and so....and decided to go agen... to make the long story short,i ended up going on thier line,and meeting aaron,the hotass, his friend justin(sam thinks he\'s hot,i agree with her though...babyface!) annd met this guy named zack...(cutie too!) anyways,i asked justin if he wants to ride with sam,(i didnt even ask sam,i just went ahead and said it,cuz i know she didnt mind anyways.) and thus,i was \'stuck\' riding with aaron!!!!!!! yeheye!!!!!!!!!!!!! i had fun screaming my ass off with him!hahahaa.  then after that,we kinda just waited for our friends to get off,and he introduced me to his friends!hahahaha.  and then.....as i was gonna ask for aaron\'s digits, my friends come after me,and  start walking me off the line... whatevers... and gez whuttt???  ZACK COMES AFTER ME AND ASKS FOR MY DIGITS! i gave em,since he\'s pretty darn cool,and since he\'s kinda cute too!!!hahaha!   '),(21947,3499629,'16,June,2004','2004-06-16 00:00:00','    juss got off the phone with mah mom.gosh!all this staying at home crap is juss plain shitty.and people are starting to become shitty too.before mah mom called i was talking to jamie and joe.he asked if we were still going to the mall. he also said his gurl\'s gonna be there,blahblahblah.  don\'t get me wrong,she\'s ayt,but it\'s just that jamie\'s not the same guy when she\'s around, i mean, atleast that\'s what me and my friends think.she\'s all up on him,and he\'s all up on her,and it\'s kinda annoying. sometimes we all feel like saying \'hey, get a freakin room!\'  jamie juss called... i read this to him over the phone...hahaha. he knows about it...but doesn\'t  care anyways. o well.i told him about how i dunno how to get to mah gym, for my first session/orientation crap whatevers... he said the place is right around his crib, and that he\'ll pick me up and take mwah to the gym....oohhhh...im nervous na talaga...   hope they treat me gently. hope i could still walk after everything.hahaha.hope my trainer calls na.  wish me luck.imma dance mah fears  off.get ready to see booty shaking.ella.   '),(21948,3499629,'16,June,2004','2004-06-16 00:00:00','    on friday we left school at 6:00, the fuckin call time was 5; and everyone started getting all pissed off bec. we were all gonna be behind sched, thus, get less time to spend at the park.  the darn bus was all crowded too. j\'amie and joe thanked mwah for saving them seats,right in the back of brittany and i\'s seatsfuck, when we got there, everyone had better buses...holy shit...tv\'s crap like that...leather interior,not to mention.  anyways,while getting thurr joe and i had some moments...haha. back massages,etc... so much for being friends! but me and jamie did too...we shared  his cd player and he cracked up when i got the words all mixed up on \"the reason\" and some people were nice enough to wave at us... this sweet old lady waved to me thrugh her car... and this absoluutely adorable little kid waved at jamie!!! he was so cute!!!mah gossshhhh!  anyways,jamie and i had a little  fight before enjoying ourselves.hahaha.  so we split up and he hung out with his \'girlfriend\'annd her crew,and ofcourse,the ever faithful joe.  anyways,i hung out with samantha,alberto,and souha... ofcourse,throughout the whole night sam and alberto started being themselves,which didnt work out for them...it kinda did for me, though cuz im kinda used to them already,but i gez not that much when they are bitchin and blabberin crappy shit to each other.souha though,was just chilled out.i mean,yeah.  we had fun.i\'ll have to continue this blog next time,ayt??? part 2 will be coming up soon,i juss gotta tell y\'all something else....   '),(21949,3499629,'18,June,2004','2004-06-18 00:00:00','    well, sherry, my trainer and i had a talk today over the phone,and she said were meeting at 5 pm monday...   and yesterday,when i got the envelope my diploma was in, it was suupposed to have mah summerschool sched...but unfortunately,it didn\'t! and what\'s worst is that summer school starts on monday!!! FUCK IT! that damn bitch from GHS didn\'t sign mah ass up!now i have to register by mahself... how???i have no shittin stinkin cow bownin idea!  fuck this shitty ass life!im pissed as fuck!   thankgodformusic.ifnotforitidgocrazy.holdupiamcrazyaminot?  laterz...gonnadancemahassoffandtrytoshakeoffallthisstress.ella.   '),(21950,3499629,'18,June,2004','2004-06-18 00:00:00','    well,gez what peoples???i made it through middle school!!! i was the first one to  walk out into the field,was extra nervous,and almost got my diploma from the wrong person,thank god i didnt though!  in the end,mah mom ended up going back to work,and she handed down the responsibility to jamie. she wasn\'t very cool about it though.its cuz we were supposed to go to the mall,and crap,and mah mom really didnt trust him that much,(consider this:peirced tongue and ears...oist,my mom\'s pinay noh!)and she had a \'talk\' with him before she left the school...  she\'s like, where are you guys going?(even though she already knowzzz!) how\'re ya gonna get there?? what time are you guys gonna go home, ARE YOU TAKING HER HOME? how?  and jamie was juss all calm and looked into her eyes the whole time!  .::pRoPs::. .::pRoPs::.  .::pRoPs::. he\'s like, to the galleria, blahblahblah... here\'s mah fave part...this answers the last two questions...yeah...i\'ll walk her home~  hahaha.so after that,i think my mom just gave me the dough and the cellphone...  she definitely trusts jamie more now... maybe she\'ll akshalli let mwah go places with him more.   '),(21951,3499629,'20,June,2004','2004-06-20 00:00:00','    yesterday teresa and i went to megan\'s block party. i got picked up at rally\'s around 12 pm, when the call time was 11~  go figure!  anyways,we had a lot of food,and danced our asses off. then,we sneaked teresa into her house. finally,i went home at 10 pm and then had the tolandway people come over to deliver the virgin mary and had to pray. i was sssoooo tired! my eyes were watery and i just felt soo sore from all the ass shaking. akshalli,i still am.  oh gosh!  hayden and jose were outside.we talked for a while. imma go now.gotta get ready to go to my buddy chris\' house. jamie and baby joe might go too.but daisy and alberto surely are.  mywholebodyhurts.ihopeifeelbetter.huhuhuh.laterz.keepitlocked.ella.   '),(21952,3499629,'26,June,2004','2004-06-26 00:00:00','    hhhaaayyyyyy...yesterdday...hhhmmmm?where do i start??? well,in the morning i went to mah foods class,and i was head chef,we made some gorgeous looking BILTMORE APPLE PANCAKES,and after snack, we made FRITTATAS *sausage and vegetables, (frittata=italian style omelet)annd didint really get to eat it! i kn0ow! it\'s fucked up, but it was already 10 mins after the bell rang,and i didnt wanna be late for the bus,  and gez what?the peppol i take the bus wuth already left,and i figuured the bus already left,ok,honesty,i juss didnt wanna go to the bus stop alone...there! and then i saw joe and jamie!thank god for the them!then i gez i asked joe if i cuud hab a ryd home, he said yeah...but we had to go to the office first, cuz thats where his mom was, getting his cell phone back and junk...anyways, his mom was juss great! she\'s so young and basically, she\'s juss a ray of sunshine!WE HAD SUCH A LAFF! she\'s so like joe and  jamie,it\'s hard to think she\'s a mom!  then we went to mcdo and i gots a fruit and yogurt parfait and \'babala\' was like, you don\'t want anything else? she\'s like, what about a sundae cone?ya want a sundae cone???i was like,oh,no thanks...it\'s ok...she is juss really nice!  THEN THEY DROPPED ME OFF AT HOME...BLAHBLAH.. then i went to go see noelle,robin and karla at the galle, and then we went to this store blahblahblah....then we went to go get my piÃ±a colada slushee and they got some nibbles,i then saw bianca,and we left for a while to go see if jamie nd joe were there yet,then when we came back,some other peeps had our table,and i saw kc , robin and noelle walking to us...asked em where mah stuff was(i got a whole buncha accessories,and told methey left it at the table,\"cuz we thought you had it,and because you juss left it there...\" \"you didnt tell us tyo bring it\" blahblahblah, right now im like, pissed still, cuz i had a lotta accessories in that bag,but oh well. it aint gonna come back to me.so what the fuck,freakin fuck it!  then jamie and joe came...and then i \'introduced\' them to kc,robin and noelle, they were juss like, hi,and then they turned away and left...i honestly dunno what they were actin like that for, but fuck it,im not the one who looked stupid,right? i went to tower recs and told noelle they were actin stoopid,whattever,but she totaLLY washed her hands off! fuck it i say! Hu CuRrZ??? anyways,then we all split up, and noelle went to macy\'s and kc and robin went to the 181 bus stop,and i went to the #2 beeline bus stop, and then bianca and maria and joe and jamie waited for mwah to get on the bus before they got thier tickets,and before i went on,i gave em all big hugs,and ran,then jamie said something about how i might get lost agen (im notorious for going on the wrong busses) and joe annd them waved goodbye...  so much for a \"fun\" day! i gotta go clean the house,before the people who will pick up the virgin mary come and pick it up, plus,i still gotta take a shower....   peace.love.happiness.prosperity.serenity.itsinyou.lookinside.ella.   '),(21953,3499629,'01,July,2004','2004-07-01 00:00:00','    yep.changes. so much has changed these passed few days.so much.  yesterday,i got a letter from one of my friends, revealing how \'he\' really felt about me.hhhaayyy! whattodo?dunno.dont wanna worry about it yet...  IM TOO YOUNG FOR ALL THIS SHITTY ASS DRAMA!!!  yesterday,i met this chick and had a real serious conversation about her about how she likes mah other friend, and i gez she hates me now because she thinks that i put her down and made her lose her confidence to ask him out. hey...honestly,i didnt mean to do it. i just told her to take shit easy,and she lost her confidence na? mah gosh.  it hurts me how my friend always sacrifices and tries to hide his feelings,and then,just blows up \'all of a sudden\'  also how even though he sacrificed the most,he also got hurt the most.  it hurts me how the most undeserving people in this world get all the blessings, and yet they still dont realize na it\'s ALL there!  it sucks how they dont realize that they don\'t really have to look anywhere else, because it\'s right freakin there!  it makes me so upset that i feel what i feel, but im too stinkin scared to admit that im feeling it.  it gets me pissed when i realize that when people gimme the same exact peice of advice i gave em when they were in the situation, i realize that now that im in it, its not so easy.  NO MATTER HOW OFTEN I TELL MYSELF TO DO THE RIGHT THINGS, I END UP DOING WHAT I DIDNT WANT MYSELF TO DO IN THE FIRST PLACE.  it sucks how i learn so much from my exes, but how i doubt if they even learned anything from me.  it sucks how i try to give a peice of me out everytime, and how people act like its \'nothing\'  BUT WHAT SUCKS THE MOST IS HOW I ALWAYS TRY TO ANALYZE LIFE, ONE DAY AT A TIME,ONLY TO FIND MYSELF SAYING,LIVE LIFE, HAKUNA MATATA.  GOSH,I GOTTA END THIS SHIT NA BAGO PA AKO MAGING TOO DEEP THAT EVEN I WOULDNT UNDERSTAND KUNG ANU NA SINASABI KO.      pRaYfOrMwAh.PaRaNgAwAnIonAnAmAn0.ella.   '),(21954,3499629,'02,July,2004','2004-07-02 00:00:00','    when i first met you when i didnt know you when we were barely strangers when i wasnt your lover?  ya told me ya needed me ya said you COULD go out with me but why did you suddenly change? how come ya went away?  now you\'ve left me confused it hurts cuz i cant make you choose because now i know exactly what you\'ll say after that, you\'ll surely go away!  i cant say i dont wanna stay i cant tell you to go away i just wanna go back to yesterday when you werent so far away...   '),(21955,3499629,'02,July,2004','2004-07-02 00:00:00','    You used to talk to me You used to tell me I used to think It was worth the suffering  But now ya have HER Yo eyes are just glued to HER everything\'s changed now I dunno why it had to  Now i ask myself Now i really regret some shit we did back then way,way,way back when???  i guess im alone now No you to stand beside me No you to tickle my fancy? No you to ask me...   '),(21956,3499629,'02,July,2004','2004-07-02 00:00:00','   well,this blog\'s for y\'all!!! this blog is where y\'all will get to have a complete update on mah life, read about the stuff i speak about, and, maybe get to say stuff too!!!!!   peaceout.speakout.ella. '),(21957,4058677,'28,July,2004','2004-07-28 00:00:00','   Step up to the crease!  Here is the first of many overwhelmingly interesting posts, serving to inform the world of my futuristic vision.  Special mention needs to go out to my fellow visual prophets from the South African Institute of Visual and Aesthetic Significance (SAIVAS). Word to Nic Grobler, Sven Uhlig and David Bouwer!  Check out the following links:   http://nicgrobler.typepad.com/saivianation/  http://nkomazi.blogspot.com/  May the practice of online documentation serve as a tool that promotes living for a higher goal.  '),(21958,4058677,'30,July,2004','2004-07-30 00:00:00','   Mr White, you made my day with your little \'stuff-up\'!  Hetzner must be reckoning that we really like to play our cards open. ;-) Maybe they\'re thinking that we hope they\'ll offer us a better rate if they know we\'re gonna get it FREE somewhere else.  I also think that exploring the FREE hosting is a good option to explore. If it doesn\'t fulfill our requirements we can easily change our hosting. Our URLs will remain the same if it ever comes to that. I just want to make it clear that I don\'t want to be Mr Pink. I want to be Mr BurntUmber or Mr OliveGreen. Okay, lets just make it Mr Green.  Mr Green. '),(21959,4058677,'27,July,2004','2004-07-27 00:00:00','   \"Don\'t ask for a pay rise. Ask for a 4 day week.\"  Work. Work. Work. Work. Play. Play. Play.  Check out www.howies.co.uk '),(21960,4058677,'27,July,2004','2004-07-27 00:00:00','   dudes... I\'ve found the mother of all free web hosting http://www.ihostfree.com/ This is like an equavalent of an R279 monthly account at hertzner. With no apparent catches. 500MB of space for each of us!...we\'ll have to test this out.... my credit card is not going through to register a domain name, some address inconsistency... maybe I\'ll check it out over the weekend sometime.  Vibe me your guts bru\'s. '),(21961,4058677,'27,July,2004','2004-07-27 00:00:00','   This one is not just free... it\'s quite fast. The question is..., can saivians live with ads on their sites? '),(21962,4058677,'19,August,2004','2004-08-19 00:00:00','   Hey Sir Snev, since I have registered my own hostname, I am battling to set up the hosting by myself at ihostfree, do you have any tips where I can start? Is there someone I can e-mail to ask them about it? '),(21963,4058677,'04,August,2004','2004-08-04 00:00:00','   Fellow SAIVIANS,  Just a quick note to say that it was really a pleasure to have spoken to you this morning. Its decision time for me again and talking to you guys will most definitely help me reach the right decision quicker. So thanks for that. You rock!  Just to let you know, I have registered the domain name www.svensven.com with ihostfree. To date there is nothing on the site, but at least I now have an address I am happy with. I will start posting bits and bobs to the site within the next two weeks. Itâ€™s a start and hopefully momentum will be gathered as we all start moving along. Iâ€™m looking forward to sharing my presence with you on the web in the near future. As Mr Grobler rightly said this morning: â€œOur future is fertileâ€   Carpe diem SAIVAS! '),(21964,3580741,'29,June,2004','2004-06-29 00:00:00','   Okay - didn\'t exercise today because coworkers from London and Sydney were in town, we went to  urlLink Triumph  and I had a semi-late night.  Plus I was so worn out from all the exercise that happened yesterday.  Updates on the girls... Rachael - nothing from her, I\'m thinking of closing that one out.  Laura - what the fuck is up with her.  She\'s about to move in with \"her boyfriend\" what the hell would she be emailing me for?  Christ - the girls I know really suck.  I\'m sitting at home on the couch, watching pbs - blogging, and working at the same time.  I think that I\'m going to go to bed early tonight, I\'m really starting to get tired.... I\'ll just read until I fall asleep - right now I\'m reading Timeline by MC.  I\'ve read it before and enjoyed it, so I thought I\'d read it again.  Ok well - I\'ve got to finish a couple things for work... signing off.  '),(21965,3580741,'28,June,2004','2004-06-28 00:00:00','   Okay - did another 30 minutes of running this AM... however, it is a little dangerous to be running on quakerbridge roady - I have to run on the median in the middle of the road, and while it is at least 15 feet wide, its still a little scary.  Tonight I meet with the landlord from my old place to do the final walkthrough.  Devin, who I used to live with, had put a hole in the wall upstairs that was never really fixed right, and wasn\'t painted.  That\'s the only major damage that we might have to pay for.  I guess I\'ll find out tonight!  After that I\'ll be heading to the  urlLink rock gym  for some more climbing.  '),(21966,3580741,'27,June,2004','2004-06-27 00:00:00','   Ok - I did 30 minutes of running, and plenty of rock-climbing.  I just need to keep some sort of moderate exercise up every day... alternating to avoid too much stress in one place at a time... and gradually increase the intensity a little each day.  Made pasta with scallops in alfredo sauce for dinner tonight... mmmm and there are leftovers for tomorrow night.... scallops will still be okay tomorrow I hope.  Oh well, time to clean up.  Still nothing from Rachael.  Also still no hint from the girl from  urlLink this post  ... referred to as Laura from here on out.  We\'re just emailing a little.  Oh well.  Looking forward to my vacation coming up in July...  8 days of diving... sweet.  '),(21967,3580741,'26,June,2004','2004-06-26 00:00:00','   Someting my Mom introduced me to....  Breakfast Smoothie: 5 large cleaned strawberries 1 1/2 cups of non-fat vanilla yogurt 1/4 cup of orange juice 1 bananna  Throw the above into a blender and blend until it\'s smooth... then drink it!  Quick and healthy... mmmmm and it tastes great.  I\'m going to stop now before I start sounding more like a commercial.... and make my self one - maybe I\'ll throw in some peach too.  Then it\'s off to the old house to clean it up for the last time so it will be ready for our old landlord to walkthrough. I\'m also going to pickup on a physical fitness routine again so I can get back into top shape.... oh, and still nothing from Rachael.  '),(21968,3580741,'26,June,2004','2004-06-26 00:00:00','   When I was little I always had those thoughts the the whole world was a stage and everyone around me was putting on a play just for me.  Maybe they were robots, or maybe I was in some MATRIX-like world... but that movie came out after I had grownup so I didn\'t have the benefit of seeing that movie.  But later on, I began to read books and see movies that at least confirmed that other people had thought the same thoughts and feelings that I had as a kid.... if other people have thought of things as strange as that before, is it even possible that I\'ve had an original thougt ever in my life?  But then when you add my unique experiences then it must be possible.  Like has anyone ever thought about what it would be like to fly a plane while getting a back massage and eating jello?  Doubtful.  :-) That\'s all I have to say.   urlLink Mr. Roboto  '),(21969,3580741,'26,June,2004','2004-06-26 00:00:00','   I\'m writing this at 8:30 am, the cable/internet guy is supposed to come sometime between 9am and 1pm.  Hopefully I\'ll finally have internet access sometime after that!  So I met Rachael last night.  I thought she was incredibly cute.  She\'s 21 and I\'m 25, but I think once you pass that age of 21, the difference really doesn\'t matter that much.  She admitted to being kinda nervous before we met on the phone, and I was nervous too.  So I think that me admitting that to her helped us meet without too much pressure.  The \"date\" was a little weird, we didn\'t hit it off as well as we did on the phone, and I think that\'s because of the circumstances.  Anyway, after the movie she had to run off to a friends party - I\'m not sure if the party was really happening, or she was just using it as some excuse to get away.  I gave her my business card, she said she\'d email me, but we\'ll find out.  Oh well - only time will tell! :-)  I think we could get along pretty well - she is just as nuts as I am when I\'m comfortable with people.  '),(21970,3580741,'25,June,2004','2004-06-25 00:00:00','   So Rachael, the girl I talked to the other night, called me last night just as I was putting my phone in the charger.  We only chatted for a couple of minutes, but it was nice... in fact we are actually meeting tonight.. :-)  We are supposed to see  urlLink Fahrenheit 9/11  tonight - I think the movie topic might be kinda harsh for a first date, but she seems excited about it, so I\'m sure it will be fine... if we can get a seat! :-)  I can\'t decide if I should shave my head again or just keep it really short.  I like it really short, but its tough to get it that way without using the bic and letting it grow a day... maybe I\'ll let it grow a little bit and see how it goes.  Anyway - I\'ve got to clean out my car just in case we decide to meet someplace else and drive to the theater together or we decide to go someplace afterwards - maybe dinner or someplace where we can talk.  I\'ll clean it out at lunchtime - please don\'t rain!  '),(21971,3580741,'24,June,2004','2004-06-24 00:00:00','   it\'s been a day, and I still can\'t get that bestie boys song out of my head.  maybe its because their new album just came out, and i will probably buy it or rip it off the internet someplace.... no, wait, thats only metallica that I would take off the internet.... sue their fans... christ, what the hell is wrong with them.  so I talked on the phone last night with someone from an internet dating site (i know, those things are weird, but it helps me meet people).  Her name is Rachael and wow - we talked for a long time.  I think that we seem to be on the same page when it comes to being crazy - she laughed at some of the dumb jokes I made, and she had me practically die of laughter for something she said... don\'t remember what exactly it was but it was funny :-)  I\'m going to call her back tonight - I know that is supposed to be against the rules or something, but I don\'t play by the rules anymore!  And the cool thing was - just as I was starting to get tired and wanting to fall asleep, she said she was getting tired.  It was perfect timing... though she seemed to have so much energy, I think she could have talked for hours.  '),(21972,3580741,'23,June,2004','2004-06-23 00:00:00','   so what do you do when someone that completely broke your heart a year prior suddenly comes out of the blue and says whats up....  I should just ignore her, I know, but there\'s always that thought in the back of my mind saying why couldn\'t it have been different.... why is she writing to me now unless she feels that she made a mistake before or something.  So I just wrote an email back - not really giving any real information, just sort of testing the waters - if she has any intention to do anything else, then she\'ll just have to say it - I don\'t want to be played with again....   Here\'s a little picture for the viewing pleasure of my readers - just some random girls :-)     Girls - all I really want is girls And in the morning it\'s girls Cause in the evening it\'s girls  I like the way that they walk And it\'s chill to hear them talk And I can always make them smile From White Castle to the Nile  Back in the day There was this girl around the way She liked by home-piece M.C.A. He said he would not give her play I asked him, \"Please?\" - he said, \"You may.\" Her pants were tight and that\'s ok If she would dance - I would D.J. We took a walk down to the bay  I hope she\'ll say, \"Hey me and you should hit the hay!\" I asked her out - she said, \"No way!\" I should have probably guessed her gay So I broke North with no delay I heard she moved real far away That was two years ago this May I seen her just the other day Jockin\' Mike D. to my dismay  Girls - to do the dishes Girls - to clean up my room Girls - to do the laundry Girls - and in the bathroom Girls - that\'s all I really want is girls Two at a time - I want girls With new wave hairdos - I want girls I ought to whip out my - girls, girls, girls, girls, girls!  '),(21973,3580741,'22,June,2004','2004-06-22 00:00:00','   I\'ve spent alot of money recently fixing up my new apartment.  It\'s actually starting to look pretty good now.  My room is almost all done - I still need to get the internet setup.  When I get home tonight - probably pretty late - I\'m going to make a huge dinner, or a huge  urlLink breakfast casserole  and stick it in the fridge.  I also need to pickup a couch, chair, and rug from my Aunt\'s house... since its going to be so late, that will have to wait until tomorrow.  My parents are in Alaska on a 2 week trip - so cool, I think I\'ll go in a year or two.  '),(21974,3580741,'21,June,2004','2004-06-21 00:00:00','   ok, so i survived the movers.  they were kinda crazy - but at least they were good at moving heavy stuff quickly.  i keep on buying stuff at the stores, and there is constantly more things that i need.  however, priority 1 is getting the internet working.  '),(21975,3580741,'18,June,2004','2004-06-18 00:00:00','   not pictured is the laundry room or the garage....  but I\'ve got them.  '),(21976,3580741,'15,June,2004','2004-06-15 00:00:00','   Halfway moved into my new place.  All the small stuff is moved, I just have the big stuff to go.  I\'ve decided to ask some movers to help me - but it is a little sketchy... they called me and we setup an appointment, however the confirmation email never arrived.  I called them back last night and they said they sent the email already.  I still didn\'t get it as of this morning, so I have been calling the 800 number they gave me all morning.... still no answer!  I\'m wondering if it is some scam that they setup, the truck shows up at your house and steals all your stuff or something.  Well, there will be about 10 other people at my house tomorrow morning for the move, so they will have a hard time killing me or stealing my stuff.  But just in case - The company is supposedly called \"Excellent Movers\" and I reached them at 866-638-6683.  If there is no more blogs put up on this site, it\'s because I\'ve beened kidnapped or killed, so please let the police know to start looking for me.  Thanks! :-)  '),(21977,3580741,'15,June,2004','2004-06-15 00:00:00','   whatever i did - it really hurts.  its so incredible i can barely breathe!  must have been when i picked the tv up this morning.... i did it in a weird way so i wouldnt get 2 years worth of dust all over me - i guess that was a mistake.  but i did pickup the keys and garage opener of my new apartment this morning.  i did a quick walkthrough and dropped off a few things - it looks really nice... trying to decide if I should setup a marine aquarium or not... ok - so i\'ve already make up my mind, but I\'m trying to decide what type and how big.... :-)  '),(21978,3580741,'14,June,2004','2004-06-14 00:00:00','   Van Wilder is probably one of the funniest movies that has been made in the last 10 years.   Also Zoolander....  I shaved my head on Sunday this week instead of Saturday - everyone said that it looked really short - I guess the one day does make a difference.  Gotta finish this last miller lite and then 4 or 5 more, and then off to bed...  I pickup the key to my new apartment tomorrow, so I\'ll be going into work a little late, but they\'ll survive without me :-) Tonight its a little steamy - I think I\'ll blast the fan on me.  Anyway - have a polly wolly crappy day.... and if you don\'t have seatbelts, just try to go limp.  urlLink Season 4  '),(21979,3580741,'13,June,2004','2004-06-13 00:00:00','   I hope that we have many more days like yesterday this summer.  Played golf and stayed cool... 75 degrees.  I\'m almost all packed up and ready to move - just have to take care of a couple more things.  '),(21980,3580741,'11,June,2004','2004-06-11 00:00:00','   cloudy friday.  tonight i have to pack the basement up for my move next week.  ive got so much junk that im thinking about putting a table out on the street with a big sign that says FREE.  im too lazy to do a yard sale, and i really dont want to waste time.  as long as nobody steals the table, and i have nothing left to clean up, ill be happy.  heading to the rock gym now.  since this is the first mention of it, i rockclimb at rockville climbing center.  really cool place with a cool staff.  urlLink Rockville Climbing Center  '),(21981,3580741,'10,June,2004','2004-06-10 00:00:00','   Okay - so here\'s living situation story.  (yes, I\'m sitting in class right now, but again - it\'s accounting, so who cares)  I\'ve lived in Hamilton for 2 years with the same friends I lived with before.  I was looking for a house to buy for a few months because of the low interest rates.  But in Mercer County, I\'d be buying a shack for the same price of (virtually) a mansion in other places.  I really didn\'t want to live in a craphole, so I decided to rent.  Now I\'ll be moving into a very nice apartment behind the Quaker Bridge Mall in Lawrenceville.  3 bedrooms and 2 baths.  I\'m having my brother and a friend of his live with me, so that will help cut down on my costs.  But at least I still have my own space.  June 15 is the first offical day, but I won\'t really be moving in until the saturday afterwards - June 19.  '),(21982,3580741,'09,June,2004','2004-06-09 00:00:00','   Just had lunch with my friend Nancy from work.  mmmm McDonalds.  I was sooo hungry.  Do you think it might have something to do with the strain I put my body through last night at the race?  I say yes.  Anyway - it is slightly cooler today, but alas I do not get to enjoy the weather, I\'m stuck going to class again.  I\'d really like to play golf, or even just sit at home and do nothing, but instead I\'ll be driving down to Camden for class.  This is a review class - the first exam is on Monday, so I do really need to go - the prof will probably give out some secrets or tips for the exam.  Well - I\'m going to sit back and let my stomach do some digesting - #3 with rootbeer.  grumble grumble grumble.  '),(21983,3580741,'09,June,2004','2004-06-09 00:00:00','   How do you say exhausted....  It was so hot last night at the race, and my time just wasn\'t up to par.... Of course the alcohol from the night before didn\'t help either.  So much for the JPMorgan Chase Corporate Challenge 2004.  I\'m still exhausted - of course that we went from the finish line to the pub certainly was not too nutrious, but we really needed the beer!  Today its back to the office, and then back to school for my summer class.  My Dad\'s birthday is on Saturday, but I can\'t see him until next week - I\'m setting up a tee time so he can see my new apartment and then play golf.... probably Mercer Oaks.  Yes - that\'s right, I\'m moving soon - but more on that later.  '),(21984,3580741,'09,June,2004','2004-06-09 00:00:00','   urlLink JPMorgan Chase Corporate Challenge   Ok - so its going to be 95 degrees today or something outrageous like that.  I\'m running in the New York JPMorgan Chase Corporate Challenge this evening, so I better bring plenty of water...  Last year they took my picture and put it up on the website.  I\'m in the very left edge of the photograph looking very tired even though there is some little old guy right behind me!!  Ok - so I started the race at the very back end of the crowd, but I managed to finish somewhere near the front:     I\'m feeling a little sluggish today, could it be the bottle of wine I chugged last night?  Or maybe it was the huge breakfast I had at the diner on the way to work this AM....  either way, I think I\'ll just jog in the race, I don\'t want to get too hot.  Besides, it won\'t make the trip to the bar come any faster since I\'ll have to wait for all my co-workers to finish the race too.  I leave work at 2:30, hop on the train and head on up...  '),(21985,3580741,'08,June,2004','2004-06-08 00:00:00','   as this is my first serious blog - i\'ll just have to watch what i post until i get the hang of it.   urlLink Microsoft fires worker over weblog  '),(21986,3580741,'08,June,2004','2004-06-08 00:00:00','   today i\'m taking the first accounting class of my life.  the class actually started on June 1, but since this is my first log of this blog... well you know.  coming from a science background, accounting is so strange.  I just want to write mathematical expressions to settle companies accounting issues - but no - we need to follow these conventions set by these \"important\" people around the world... right... well - oh well, I\'ll have to press on and deal with it.  the class ends on June 24, so just grin and bear it Jason!!  '),(21987,3580741,'28,July,2004','2004-07-28 00:00:00','   It has begun.  3 fantasy teams - 2 pools.  Here\'s one.  urlLink 2004 NFL Season One-On-One Pool  '),(21988,3580741,'27,July,2004','2004-07-27 00:00:00','   Went over to Sue and Devin\'s house last night - Christine and Niffers (Jenn) were there too.&nbsp; We hung out and just drank beer, Devin had just replaced the keg, so there was gallons and gallons to drink...&nbsp; needless to say I had to spend the night - so basically, I haven\'t seen my bed in almost 2 weeks.... It\'s strange - I don\'t really like drinking that much or getting drunk too often, but last night the beer flowed like water... nay - like ice cold water on a hot summer day in death valley.... waiting for me when I get home also are my bags which are still packed.&nbsp; I made a stop at petsmart looking for aquarium stuff - didn\'t find what I was looking for but what did I find instead?&nbsp; Cute kittys just begging to be adopted!&nbsp; It took some serious willpower to walk out of there without getting one.  '),(21989,3580741,'26,July,2004','2004-07-26 00:00:00','   OK - gotta get up and head to the airport soon.&nbsp; As you can see it\'s early - definitely too early to be playing with the camera.&nbsp; Just thinking that in 24 hours I\'ll be doing my morning run near the mall, or maybe I\'ll run into lawrence square village, or down village road to mercer oaks... I\'ll just let the wind direct me... I certainly won\'t be running along a beach in paradise like I\'m about to do now.&nbsp; Is Jason naked in this picture - well - I\'ll leave that to the imagination of the millions of lovely ladies who read this post every day... :-)&nbsp; Hah!  '),(21990,3580741,'25,July,2004','2004-07-25 00:00:00','   7/25/2004 9:30amSo we have a nice quiet Sunday morning on the last full day of the trip.&nbsp; Tomorrow afternoon I\'ll be back in rainy Princeton.&nbsp; Yay!&nbsp; :-)&nbsp; I\'ve realized that my brother is a big slob.&nbsp; OK - so I knew this already, but still - how can he let all those dishes sit for days!!&nbsp; That would (and is) drive me nuts.&nbsp; Doing some deep dives this afternoon - a little bit later than I\'d want considering we won\'t have a full 24 hours until we fly - but I\'m pretty sure we\'ll survive.&nbsp; 21 hours... 24 hours... not much difference.&nbsp; I also went to the ATM and put together the tips....&nbsp; I\'m pretty sure my brother is too poor/cheap to give tips to his 2 instructors or the captain... so I put a little extra in the envelopes.&nbsp; These people work and a good portion of their income depends on tips, so I\'m always careful to make sure I give them what they deserve.&nbsp; Oh - and in case you don\'t think that my brother is sloppy and I\'m just making it up, here\'s proof:  '),(21991,3580741,'24,July,2004','2004-07-24 00:00:00','   Only did one dive trip today (2 dives in the morning) Jeff\'s ear is bothering him.&nbsp; We did some deeper reefs - on the slope from 60 feet to 100 feet.&nbsp; We hovered in about 70 feet partway down the slope and let the current take us.&nbsp; It was nice to do some drift diving, except for the vis which wasn\'t the best.&nbsp; But we saw some nice size jewfish and sharks.&nbsp; The guy I was diving with even found a nice Penn reel on the reef - it was all covered in growth, but he should be able to clean it off pretty well.&nbsp; We covered some good distance, so it\'s a good thing we did the drift, otherwise it would have been tough to fight the current constantly.&nbsp; Said goodbye to Jan and Bob - they are driving back home.&nbsp; I liked those guys - I gave them my address and email address, so hopefully I will hear from them.&nbsp; They also took some video and pictures and promised to send it to me.&nbsp; Here are some pretty spiral worms on coral... these things shoot into holes faster than you can blink if you get close to touching them - so it\'s fun to watch them hide.  '),(21992,3580741,'23,July,2004','2004-07-23 00:00:00','   Met this cute girl who was diving with her brother and Mom.&nbsp; The three of them hadn\'t been in the water for 3 years.&nbsp; I was sorta turned off from her at first because she didn\'t seem to want to talk... but who was I to judge without even trying talking to her.&nbsp; Pat - the diveshop owner had me watch the three of them as they got in the water, she didn\'t want to trust them with their own safety, but once I saw that they had remembered everything important about diving.... don\'t hold your breathe, ascend slowly, and that stuff, I let them go on their own.&nbsp; Later I was talking to girl, and found out she was a grad student in Tenessee.&nbsp; She seemed to be pretty smart - I told her my story and my job, and she was interested.&nbsp; She was studying theology - and I asked her what she wanted to do - and she said anything that has nothing to do with religion.&nbsp; I\'d have married her there on the spot after she said that!&nbsp; :-)&nbsp; But in the end, I gave her my card and said to send me an email - but of course I\'m not expecting to hear from her... plus she\'s in Tenessee, and I have no business in that area except for a quick rockclimbing trip or something.... ugh - can\'t stand country music.&nbsp; But if she were to contact me, then that would be something different... :-)&nbsp; Attached is a picture of me, the man-whore.  '),(21993,3580741,'22,July,2004','2004-07-22 00:00:00','   7/22/2004 11:30PMJeff caught so much fish - we ate between 2 and 4 pounds of fish.... each for two nights in row.&nbsp; I fried some, and, baked some, and I basically steamed some in wine.&nbsp; The fried fish came out the best, but maybe that had something to do with the frying in vast quantities of oil and butter.&nbsp; While eating at the hotels is much cheaper than going out to eat, I do miss the good service and really good food....&nbsp; but my brother is poor, so why not let him save some money.&nbsp;  '),(21994,3580741,'20,July,2004','2004-07-20 00:00:00','   Wow - the wreck dive today was great.&nbsp; Nice white ship hull that could be seen from the surface - the water was just so clear.&nbsp; There was a current that was pretty ripping though.&nbsp; Almost no current on the surface, but from about 10ft to the bottom it was pulling along at about 2.5 knts.... but in the opposite direction of the prevailing current, so that was a little weird.&nbsp; Jeff went out fishing since he couldn\'t do the dive this AM, but he\'s said he\'s caught some yellowtail, so hopefully I\'ll have some good fish on the way!!&nbsp; I have to go pick him up in a little bit.  '),(21995,3580741,'19,July,2004','2004-07-19 00:00:00','   Dove alot today with Jan and Bob&nbsp; Also, Jeff finally finished his OW certification.&nbsp; If only he had remembered to take the car keys out of his pocket before he jumped in the water.&nbsp; I guess he\'ll have to do his advanced certification on another trip.&nbsp; We did some easy shallow dives.&nbsp; Bought stuff at the supermarket for dinner and lunches.&nbsp; That should stop the cash drain somewhat.&nbsp; Oh well, what do I care, it\'s vacation!&nbsp; Diving the Speigel tomorrow with Jan and Bob.&nbsp; Jeff can\'t because it\'s a little tricky dive, so he\'s going to go fishing.&nbsp; Not sure how I\'m going to get home after the dive, but I\'ll figure it out.&nbsp; Here is a picture of Jan and Bob... very strange guys, but they like to dive as much as I do.  '),(21996,3580741,'18,July,2004','2004-07-18 00:00:00','   Jeff is in SCUBA class.&nbsp; Went diving today with a really cool guy named Mark from LA.&nbsp; He was diving with this other guy who really sucks down the air... and thus really sucks... turns an hour dive into 30 minutes....&nbsp;&nbsp;&nbsp; Mark from LA was a little nervous this morning on the Eagle.&nbsp; I could tell it was his first deep dive in a while, or maybe it was just because he was not familiar with the wreck.&nbsp; Either way, I showed him the cool stuff all over the wreck.&nbsp; We did a simple swimthrough - through the wheelhouse of the wreck.&nbsp; I traded cards with him, so maybe we\'ll keep in touch, and I can send him some pictures.&nbsp; Below is a picture of me... proof that man can live underwater without the need of air or a mask.... ok, so I only took those off for the picture - you caught me!  '),(21997,3580741,'17,July,2004','2004-07-17 00:00:00','   Ahhhh some nice snorkeling yesterday just to get my feet wet a little to prepare for all the diving to come.&nbsp; This morning I ran along a dirt path the locals use to walk to a natural jetty where they can fish from.&nbsp; It\'s sorta out in the middle of nowhere.&nbsp; The guys that ran the snorkel trip today offered to take me on a night dive, but since my brother can\'t go, I\'m going to hold off.&nbsp; Besides, there are so many potential night dives coming up, I\'d rather go with people I know.&nbsp; Dinner tonight at Lazy Dayz - chef Lupe\'s stuffed chicken... mmmmmmmm.  '),(21998,3580741,'16,July,2004','2004-07-16 00:00:00','   Nothing like waking up 100 feet from the beach...&nbsp; the air at 6am is still and the sun is just about to rise.&nbsp; 4 mile run along a white sand beach in nothing by the boxer shorts I fell asleep in the night before...&nbsp; then after that a swim in the ocean that is warmer than the air... ahhhhh yep - it\'s true, I\'m on vacation.&nbsp; Now I\'m trying to get my brother out of bed so we can eat breakfast and get going.&nbsp; Today we have some shopping - I need a new dive mask, and I would like to replace the battery on my dive computer just to play it safe... there are still 2 out of 4 bars on the battery meter, but that\'s the type of thing you don\'t want to lose during a series of dives.&nbsp; Sun is shining brightly into the hotel room now, and I\'m really hungry.&nbsp; There was some cafe one block north, I\'m going to head over there and get a coffee.  '),(21999,3580741,'14,July,2004','2004-07-14 00:00:00','   Obey my dog!  If there\'s anything that this horrible tragedy can teach us, it\'s that a male model\'s life is a precious, precious commodity. Just because we have chiselled abs and stunning features, it doesn\'t mean that we, too, can\'t not die in a freak gasoline fight accident.  A eugugalizer:  one who speaks at funerals, or did you think I\'d be too stupid to know what a eugugoly is?  I feel like I\'m taking CRAZY PILLS!!!  How can the children learn to read, when they can\'t even fit inside the building! *smack*  There has got to be more to life than being really, really, really, ridiculously good-looking.  96% of the earth is water. The other half is wheat.  '),(22000,3580741,'13,July,2004','2004-07-13 00:00:00','   Ok - so I\'m sitting in bed right now, trying to get up, but my head is still so foggy... no real reason for it, I went to bed not too late last night.  I was even reading until I passed out, so I don\'t even remember what time it was.  I\'m going to run about 5 miles today... not sure how it is going to be since I just ran less than 12 hours ago inside on the treadmill.  It was raining so hard, that\'s why I took the treadmill - normally the rain doesn\'t bother me, but everything around here was completely flooded, including some of the streets.  Oh well - so this blogging is waking me up - and I think I can roll out of bed now.  Just remembering that in exactly 60 hours, I\'ll be on a plane to paradise.... and then I\'ll be swimming alongside this fella...!  '),(22001,3580741,'12,July,2004','2004-07-12 00:00:00','   So my old hometown makes the national news once again... 5000 people in it and we have more freaks per capita than anywhere else.  I knew the woman charged with murder - I worked with her on the Ambulance Squad - we saved lives together... kinda weird she\'s now a killer. urlLink CNN.com - Woman charged in fatal antifreeze poisoning - Jul 12, 2004   Oh well - I don\'t live there anymore!  Whew!!  '),(22002,3580741,'11,July,2004','2004-07-11 00:00:00','   Scorpio! He\'ll sting you with his dreams of power and wealth. Beware of Scorpio! His twisted twin obsessions are his plot to rule the world And his employees\' health. He\'ll welcome you into his lair, Like the nobleman welcomes his guest. With free dental care and a stock plan that helps you invest! But beware of his generous pensions, Plus three weeks paid vacation each year, And on Fridays the lunchroom serves hot dogs and burgers and beer! He loves German beer!   Check out what I found on East State Street.  Now I know where his secret lair is!!!  '),(22003,3580741,'10,July,2004','2004-07-10 00:00:00','   Went running - did 6.5 miles this morning.  Pottery Barn and Borders for some stuff.  Picked up the soundtrack to Gladiator... wow, pretty decent sound.  Went swimming, now I\'m able to swim 2.5 lengths of the pool underwater on one breath.... not too bad considering I could barely do one length just two weeks ago.  Then I went bouldering at the gym...  Just got out of the shower which puts me at the present moment.  I\'m going to finish up a load of laundry, read a little, and then head on over to KC Prime for the biggest piece of meat that I can find.  Mmmmmm.  Yesterday was vegetarian, today I guess is the opposite!  '),(22004,3580741,'10,July,2004','2004-07-10 00:00:00','   Last night was a blast.  It was so cool to be climbing in the dark with my friends.  I didn\'t buy any lamps, so I\'ll just have to pick something up in florida city on Friday.  fla city is the last town before the keys begin on route 1.  Today I just got back from running, and I\'m heading on over to the pool.  Then its off to best buy to buy a vacuum....  I think they have like 20 different kinds, I\'ll probably be at a loss.... as long as the vacuum SUCKS then I\'ll be in good shape.  Tonight I\'m headed over to my friend (and she happens to report to me on my team at work) Diane\'s house.  She got a new dog named Louie, so I\'m going to meet him.  Then we\'ll go have dinner or something.  '),(22005,3580741,'09,July,2004','2004-07-09 00:00:00','   Ok - night rock climbing event tonight 10-12 at Rockville.  I\'m going swimming directly after work, and then I\'ll probably take a nap so I can stay awake tonight.  The vendor at the rock gym is Princeton Tec, and they make some good diving lights too, so maybe I\'ll pick up something for the wreck diving I\'ll be doing... starting NEXT FRIDAY!!! YAY!!!  So tonight all the lights will be out, and we\'ll have lights attached to our heads.  We\'re supposed to top-rope, but I think I\'ll just be bouldering.  '),(22006,3580741,'08,July,2004','2004-07-08 00:00:00','   Just got back from Borders... I bought... 1. Brave New World  2. Hot Plastic  3. Babbitt  4. The Etruscan Chimeria and some others.  I think that should be enough reading for my trip coming up... not that I\'ll be able to stay awake enough - I\'m usually so tired after diving that I can\'t stay awake, but we\'ll see.  I\'m freezing a bunch of bunches of bannanas tonight to help them keep for a long time.  Exciting evening, I know. :-)  '),(22007,3580741,'06,July,2004','2004-07-06 00:00:00','   Running.... Work... Rockclimbing... swimming... busy day!  '),(22008,3580741,'05,July,2004','2004-07-05 00:00:00','   Played golf this AM at urlLink Green Tree Golf Course   it was so nasty sweaty humid hot that I didn\'t think I\'d be able to make it through the course... but I did.  My score wasn\'t that great, but at least I wasn\'t working.  Oh well - I think I\'m going to get an early start on the road and hit a couple of stores on the way home to get some good deals.  Either that or I\'ll just bomb home really fast and just chill out in bed reading the last of a book.  Or, if I\'m not exhausted by that time, I\'ll hit the rock gym and do some bouldering.  '),(22009,3580741,'04,July,2004','2004-07-04 00:00:00','   Sunday evening, 4th of July.  Did alot of swimming in the ocean today off of the boat.  Did a little bit of fishing just about a mile off of the beach right in front of  urlLink Lucy  but all that we caught were skates, or sea rays.  It wasn\'t a really good spot for fishing, but we didn\'t care, just wanted to relax away from the marina for a bit.Then I had crab legs, shrimp, a cheeseburger, and some other stuff for dinner.  Now I\'m sitting in my house back in Margate, listening to this horrible... nay... disgusting band on the beach.... most of the time it is REALLY nice having a house on the beach, but right now I wish it was as far away from this band as possible.  Guess I\'ll go out in about 20 or 30 mintutes to see the fireworks, they should only be about 5 blocks away, and in addition to Margate\'s, I should also be able to see Atlantic City and Ocean City fireworks too.  Still no sunburn.  Finally - the band has taken a break and is now playing my favorite radio station.  Seriously - I\'ve traveled all over this country, and the best radio station is  urlLink  Digital 102.7.    They play stuff you can\'t hear anywhere else.  Ok - time to find a nice spot in the sand.  '),(22010,3580741,'03,July,2004','2004-07-03 00:00:00','   So I\'m here in Margate now, ready to begin the relaxing weekend.  I opened all the windows last night because it was getting pretty cool, though still really freakin humid last night.   Think we\'re going deep sea fishing today with some friends - the ocean is flat, no waves so we could probably go out pretty far without taking too long, I\'m not sure how close the gulf stream has come in yet this year.  So here we go, off for an exciting day!  oh - no, wait, I do need to do about 20 minutes of work first... doh! :-(  Oh well, I shouldn\'t have put it off last night when I wanted to be out of the office by 5pm.  '),(22011,3580741,'02,July,2004','2004-07-02 00:00:00','   How weird is this - I\'m at this liquor store in Lambertville buying beer for Kim\'s party... I come out of the store, and driving by, staring out at me from the window is Laura\'s Mom.  Laura, the girl that is trying to screw with my head, moving in with her boyfriend from a couple posts ago....  On top of her truck is a mattress, I guess today was move-in day or something.  I know that her boyfriend that she \"cheated\" on to see me a couple of times lived in Lambertville, I guess he must be the guy.  Oh well - screw it, I\'m not going to dwell on someone who couldn\'t see the storm through the rain, or whatever.  '),(22012,3580741,'02,July,2004','2004-07-02 00:00:00','   okay - so tired.  Wednesday night .... moving a couch turned out to be just Devin and I getting a big couch into a very small apartment.  It was a tight squeeze - wouldn\'t really fit in the stairs, so we had to keep it above us.... but evenutally we forced it in.  Then we played drinking trivial pursuit...  so needless to say I had to sleep for a little bit there.... on the couch that we just moved in.  The couch was pretty comfortable, and I had to fight with the cat for a spot practically the whole time.  Then I got myself up and dragged into work.  Thursday night.... party at Kim\'s house.  Pretty fun time - I never thought that a 100% vegetarian night could be so tasty.  I never did break out the beef franks or steaks, there was no need. :-)  Oh well - somehow I\'ve got to stay awake today, I need to last until 4pm when I can get outta here...  '),(22013,3580741,'12,August,2004','2004-08-12 00:00:00','   Ok - so he resigns because he had a gay affair... I still don\'t see the point.  Wouldn\'t that be between he and his wife?  Why didn\'t he resign from all the corruption that he had taken part in or turned a blind eye to over the past couple of years?  Oh well - whatever.  '),(22014,3580741,'11,August,2004','2004-08-11 00:00:00','   Wow - I wonder what all this is about!  This came completely out of the blue for me, I guess at 4PM I\'ll have to find out more... urlLink wnbc.com - Politics - NewsChannel 4: N.J. Governor McGreevey To Resign  Rumors he is gay is what I\'ve heard - but so what!!!  Why should I care if my governor is gay or not?  I mean seriously - what the hell.  And what\'s this business about California voiding all the gay marriages that happened earlier this year?  Why does it bother so many people!  I certainly don\'t care.  I think that once some of the more \"densely prejudiced older people\" start dying off this country will change for the better...   Maybe this weekend I\'ll tell everybody where I sit for all the important issues, and someone can tell me if I\'m a democrat or a republican!!  '),(22015,3580741,'11,August,2004','2004-08-11 00:00:00','   A Toyota Emil saw a slime God saw I was dog Naomi, did I moan Red rum is Pepsi murder Dammit, I\'m mad A slut nixes sex in Tulsa Solo gigolos Are we not drawn onward, we few? Drawn onward to new era Straw Warts  urlLink more  '),(22016,3580741,'11,August,2004','2004-08-11 00:00:00','   urlLink Big Big man    I have nothing to say but... \"holy shit.\"  '),(22017,3580741,'10,August,2004','2004-08-10 00:00:00','   Ok, I\'m addicted.  A friend of mine at work brought in these little packets of white pills.  They could have been crack/meth/coke/etc. and I would never have known it until it was too late.  Anyway, they taste like mint, and I can\'t stop eating them.  If I stop showing up for work, it\'s probably because I\'ve found another source of these things, and there would be no need for me to go to work anymore.  Ok, so it turns out these things are legal and won\'t kill me... but I still think it\'s kinda funny... they look like some type of street drug.   urlLink Complimints  '),(22018,3580741,'09,August,2004','2004-08-09 00:00:00','   A spider\'s web was in the way as I walked through the dark wood. To the right another web exists, though this one is a tight tangle of sticker bushes. To the left, thick deep brown mud. The sunlight glints on the silk, one bright strand in a sea of Verde. I see the spider - staring at me.  Is she going to be mad at me? Does she know I plan on walking through her web?   I mean, really what are my options.   I have new clothes on - I really have no business being in the woods with these clothes. In the end, it is the pitiful look of the spider, whose only method of catching food is her web. I take a sip of my chocolate milk, and a bite from my turkey sandwich. And I walk through the mud.  '),(22019,3580741,'08,August,2004','2004-08-08 00:00:00','   still lying in bed... on my left side... tv on cnn... I didn\'t wake up with a headache today.  I\'m going to go running in about 15 minutes.  Then I\'ll wash my sheets and clean the kitchen... but by then the mall should be open.  It\'ll be time to go blow a couple hundred dollars on clothes...  I spent about 150 yesterday, and I like what I bought, so I think its time to continue the trend.  This time I\'ll buy stuff that I can wear to work.  I also want to find a headboard for my bed... I\'ve never had one before.  Hopefully its going to be sunny and warm today so I can go swimming this afternoon.  I\'ll bet that I can swim the length of the pool underwater back and forth 3 times.  Changing the subject completely... here is the first fire department I ever joined:  urlLink Indian Mills Fire   urlLink  '),(22020,3580741,'07,August,2004','2004-08-07 00:00:00','   Ok, so I met Misty for a late lunch.  She was actually really cute!  Lunch only lasted a little over an hour, but then we went walking in the Franklin Mills Mall - if you\'ve ever been there, then you know it\'s a pretty large mall.  I think she and I got along pretty well, she is slightly taller than I am, and while that doesn\'t bother me, I know some girls like guys to be taller.  Oh well - we\'ll see how things work out, I\'ll call her Monday night and see if she wants to go out again.  After our date, I drove through New Hope, then went to visit my Geocache to check on how it\'s doing:   urlLink GC24B Washington\'s Crossing 1 by Jason Malatesta   While flipping through the log book, I found an entry that Laura and I had left in Feb 2002.  Oooof.  I had completely forgotten about her until that moment.  Oh well, I\'m sure I\'ll forget her again soon enough.  Isn\'t it funny how they keep turning up one way or the other?  Anyway, I\'m going to make myself a nice pasta dinner, and then I\'ll head out to the rock gym and do some climbing.  Peace out.  '),(22021,3580741,'06,August,2004','2004-08-06 00:00:00','   Ok - you have to love that you can try to tell the president to do anything you want him to.  If he understands the command, meaning that it is not too many words, and is something a 2nd grader could understand, he still won\'t do it!  urlLink Subservient President   Anyway, try typing in some instructions and see if GWB will comply!  '),(22022,3580741,'05,August,2004','2004-08-05 00:00:00','   Woke up early today, but I just couldn\'t drag myself out of bed in order to go running.  So instead I just picked up my book and read a little.  This latest book is quite boring, and I\'m hoping it is going to turn around and be exciting or interesting soon....  I\'m going climbing tonight, my first bit of doing something physical since I got back from vacation.  Being sick sure just took the fun out of my life!!  I\'m glad that I\'m better and I can do some fun things again!  '),(22023,3580741,'05,August,2004','2004-08-05 00:00:00','   Got an email from a girl named Misty...  We\'ve been emailing back and forth for some time now, but I\'ve finally got her phone number... and what\'s more... she has agreed to meet me this Saturday!  :-)  I\'ve seen her picture, and she\'s pretty cute!  Out of respect for her, I won\'t post her picture or any more details about her here unless she gives me permission.  I\'m looking forward to meeting her!  On another note, I think I\'m finally past the illness that I\'ve been suffering from.  Today I feel good.  '),(22024,3580741,'03,August,2004','2004-08-03 00:00:00','   Today I was sick.  Stayed home and worked on the couch while watching movies.  I didn\'t get to pay attention to any of the movies because work kept me busy, but oh well - that\'s how I get paid!  I also kept Max on me, she seems to like to sleep an awful lot, but I guess that\'s normal, since she is a cat.  '),(22025,3580741,'02,August,2004','2004-08-02 00:00:00','   I\'m sick.  I\'ve been sick since Friday night.  Saturday Sunday Monday.  But it\'s not some really horrible flu - I have a really stuffy nose, a headache, and I get dizzy.  If there is a fever, it is really slight, or else my body just can\'t tell it\'s own temperature any more.  Anyway - my ears are all clogged up too.  I haven\'t really been sick in, well - I can\'t remember the last time - must be years ago.  I should probably be glad about that, but I\'m not really in the mood to be glad about anything - I just want to feel normal so I can go back to work.  I\'ve also decided to try and pickup the first pretty girl I see someplace no matter what the circumstance (except for someone at work of course!)  But it could just be me thinking in my current state of mind that is all fogged up from being stuck in my apartment all day.  '),(22026,3580741,'01,August,2004','2004-08-01 00:00:00','   She\'s such a little baby!  :-)  Named after Max Planck... she\'s a black cat... and her name is a tribute to Planck\'s work with thermodynamics and blackbody radiation.... get it?  No?  Ok - well just call her Max and don\'t worry about the reasons why. :-)  '),(22027,4061665,'30,July,2004','2004-07-30 00:00:00','    Introverted ( I ) 62.5% Extroverted (E) 37.5%  						Imaginative ( N ) 64.71% Realistic (S) 35.29%  						Emotional ( F ) 63.64% Intellectual (T) 36.36%  						Easygoing ( P ) 73.33% Organized (J) 26.67%  					  				    	 Your type is:    INFP    	 		      						You are an  Idealist , possible professions include - information-graphics designer, college professor, researcher, legal mediator, social worker, holistic health practitioner, occupational therapist, diversity manager, human resource development specialist, employment development specialist, minister/priest/rabbi, missionary, psychologist, writer                urlLink get yours .   '),(22028,4061665,'29,July,2004','2004-07-29 00:00:00','    I can\'t wish it was how it used to be. It\'s been this way for a while.   '),(22029,4061665,'28,July,2004','2004-07-28 00:00:00','    Time doesn\'t take me seriously.   '),(22030,4061665,'28,July,2004','2004-07-28 00:00:00','    I\'ve fallen into myself again.   '),(22031,4061665,'28,July,2004','2004-07-28 00:00:00','    When all the houses have fallen down, and there\'s no one else around. I will stand on my own two feet, and fight to stay incomplete.  Then I thought I heard a sound, water drops fell to the ground. I turned around to pick up pace, tears mixed with rain ran down my face.  I gathered what was left of my strength, and held it out at arms length. I set it down and ran away, I will come back to fight some other day.   '),(22032,4061665,'28,July,2004','2004-07-28 00:00:00','    I write. Not to remember. I hate. The month of December.   '),(22033,4061665,'28,July,2004','2004-07-28 00:00:00','    Why must I paste this look on my face and act in ways I don\'t feel.  When did the world decide it\'s better to hide so that no one will know what is real?  It\'s hard to see how it would all be if we took off our masks and began to heal.   '),(22034,4061665,'28,July,2004','2004-07-28 00:00:00','    I try to relax and let it all go. I die for a while, it\'s the only thing I know. Sometimes I just stop, I can\'t take anymore. Human beings can be such a bore.   '),(22035,4061665,'28,July,2004','2004-07-28 00:00:00','    Learning from the bravest souls to fall Iâ€™ve found in life, that time isnâ€™t time at all So it goes, lying here in my nest for naught Ordered, assembled and paid for, without a thought  If the night will, Place itâ€™s gentle arms around me, Wrap me up and keep me.  If the night will, Fight this sad war without me, And carry on despite me.  My comfort alone, the only comfort I seek Tellingly, outside of mind, it\'s the labels that speak Forced by choice with eroded conscience rare Much better, for tomorrow, with closed eyes would I fare  12:23am â€“ 01/12/04  c   '),(22036,4061665,'28,July,2004','2004-07-28 00:00:00','    For Melissa...  1. WHAT COLOR ARE YOUR KITCHEN PLATES?  Usually white (paper). If not paper, then they are white. But I don\'t through those away when they are dirty.  2. WHAT BOOK ARE YOU READING NOW?  A Heartbreaking Work of Staggering Genius by Dave Eggers.  3. FAVORITE BOARD GAME?  Trivial Pursuit  4. WHAT\'S ON YOUR MOUSE PAD?  Don\'t have a mousepad...I use a tablet.   5. FAVORITE MAGAZINES?  Q Magazine is the best music publication ever. I also read the occasional NME. I like Print Magazine and Relevant as well.  6. LEAST FAVORITE SMELL?  Rotting corpse...  7. WHAT IS THE FIRST THING YOU THINK OF WHEN YOU WAKE UP IN THE MORNING?Â Â   It sure would be nice to be sleeping...  8. FAVORITE COLORS?  Surfer blue. (CMYK 50,30,15,0)  9. LEAST FAVORITE COLOR?  All colors do have their application. Â    10. HOW MANY RINGS DOES IT TAKE BEFORE YOU ANSWER THE PHONE?Â   Depends on what I am in the middle of and who is calling.  11. FUTURE CHILD\'S NAME?  Girl: Kate or Katie maybe. Not sure. Boy: Aidan or Hayden.  12. WHAT IS MOST IMPORTANT IN LIFE?  Real relationships. Love wins you know?  13. CHOCOLATE OR VANILLA? Chocolate on chocolate, covered with chocolate, with a tiny bit of chocolate on the side.  14. DO YOU LIKE TO DRIVE FAST?  Are you a cop? Cause I was just passing that old lady because she was driving \"unsafe\". I just wanted to get past her and slow back down to my normal 5 below. Really. It\'s true.  15. DO YOU SLEEP WITH A STUFFED ANIMALSÂ   I don\'t sleep, but I love stuffed animals...  16. STORMS - COOL OR SCARY?  Bigger is better, right?  17. WHAT TYPE WAS YOUR FIRST CAR?  1990 White Astro Van with fusia and teal heart beat decals on the side.   18. IF YOU COULD MEET ONE PERSON DEAD OR ALIVE?  Frances Farmer.  19. FAVORITE DRINK?  Water for the day, Alaskan Amber for the night.  20. WHEN IS YOUR BIRTHDAY?  25 July.   21. DO YOU EAT THE STEMS OF BROCCOLI?  Wait, brocolli is edible?  22. IF YOU COULD HAVE ANY JOB WHAT WOULD IT BE?  Studio engineer.   23. YOUR HAIR COLOR?  Blonde.   24. EVER BEEN IN LOVE?  Yes.Â Â Â Â    25. ONE FAVORITE MOVIE?  Impossible, but I will just go with The Royal Tennenbaums.  26. DO YOU TYPE WITH YOUR FINGERS ON THE RIGHT KEYS?Â   Yes, because if I didn\'t, Ms. Wright would smack my hands with a ruler until they were swollen and bruised.  27. WHAT\'S UNDER YOUR BED?  My bed is on the floor. So, carpet, cement, dirt and then China... I think.  29. FAVORITE SPORT TO WATCH?  NCAA Basketball.   30. WHAT IS YOUR SINGLE BIGGEST FEAR?  Never finding my soulmate.  31. SAY ONE NICE THING ABOUT THE PERSON WHO SENT THIS TO YOU.  Melissa takes great History notes!  33. FAVORITE CD?Â   Once again, impossible, but I am listening alot to The Delays. Sounds like James and Suede a bit. I think I like it because it sounds like the Beach Boys a bit. Happy and shiny.  34. FAVORITE TV SHOWS?  Most things on the History Channel and Discovery Channel.  35. KETCHUP OR MUSTARD?  Ketchup on everything.  36. HAMBURGERS,HOT-DOGS OR VEGETARIAN?  Hamburgers from Bonnie\'s Place and Hot Dogs at Yesterdog and then, yeah I guess, if I was still hungry, I would eat a vegetarian...  39. BEST PLACE YOU HAVE EVER BEEN?Â   Gull Lake in the winter time when everyone else is gone.Â    40. IF YOU HAVE A SCREEN SAVER WHAT IS it?  It is called Flurry. It is just one of the Mac OS X screen savers. Spinning color wisps.  41. BURGER KING OR MCDONALDS  Burger King. Though I still plan to launch my franchise called McBurger Queen any day now.  Â 42. FAVORITE HOLIDAY?  4th of July.   43. FAVORITE DAY OF THE WEEK?Â   Sunday. It\'s the only real day left, isn\'t it?  Â 44. FAVORITE SEASON?  Fall.Â    45. DO YOU KNOW THE BOOKS OF THE BIBLE?Â   Most of them. Though I could not recite them in order by any means.  46. IF YOU COULD DO HIGH SCHOOL OVER WOULD YOU?  No, I think I was lucky enough with how things played out the first time around.   47. LOCAL BAKERY, DUNKIN DONUTS OR KRISPY KREME?  Local bakery. I like dutch, old world style donuts.  48. WHAT IS YOUR LATEST HOBBY?  Building a tiny getaway in the woods to camp out at.   49. WHERE WOULD YOU GO ON YOUR NEXT VACATION?Â   I am going to northern california this saturday, but I would love to go to europe somtime in the next couple years.  50. WHAT DO YOU LIKE TO DO WITH YOUR CHILDREN?Â   Don\'t have any just yet.  Hmmm. I guess that\'s me then...   '),(22037,3418474,'31,May,2004','2004-05-31 00:00:00','    Man I haven\'t written one of these forever.  This past weekend was pretty cool.  Youth was awesome I loved the worship and the small groups were cool.  Saturday I went and helped the food bank that was rad.  Umm sunday I went to church and then set up a trampoline and then I played softball later that night.  My Life with God is pretty good:)...It\'s kinda tough not having many people to talk to.  I wanna be the light Jesus has called me to be but it is tough.  God\'s been keeeping me real busy lately sorta to get my mind off the friend situation.  Right now...SOnja is kinda my only good friend and it sorta sucks but at least she\'s a rad girl:).  I\'m trying to live this thing for God and just be the light in the darkness and just try and live for God with my all.  I find it way easier to hang out with people who actually give a crap bout what\'s goin on in my life.  I love it when people just take time to talk with me and stuff.  I find myself being really fake when I\'m with people that already have a bunch of friends and I mean nothing to them.  I love it when I act like the Real Jordan It feels great:).    I\'m not quite sure of God\'s purpose or his plan...But everyday I wanna live Christ:).  It\'s totally worth it to surrender your all to God cuz he\'ll only do great works in u.  I haveta stop being of the World...and just be fully like Christ.  My life has a pretty bright future in that It\'s gonna involve Jesus and God\'s Plan soo it has to be bright.  I haven\'t changed soo dramtically yet that people wouldn\'t know it\'s me...but that\'s what I\'m planning on doing.  Changing into a true christ follower instead of one of those kids who have things good in their Lives but they never effect the rest of the world.  ANd just let Jesus take hold of everything.  It takes alot of Faith and I am afraid...I shouldn\'t be but I am...I wanna become fully real and act the same around everyone.  I\'m hoping that over the summer and stuff I can really get some friendship booming and just also become really close with God and become real.  People usually fall away from God in the summer I wanna do the opposite and just stick to God like glue.  ANyways guys...If u need prayer just write it in a comment or email me:) That\'d be great...It also be Rad if u guys all pray for me and that My walk with CHrist would turn into a sprint:P ANyways cya all later bubye and GOD BLESS AMERICA:) haha not GOD BLESS U:P   '),(22038,3418474,'24,May,2004','2004-05-24 00:00:00','    These past few days I\'ve been pretty busy...Me n Josh built a soccer net yesterday after church...It\'s pretty sweet:D.  Today I went boating with my family and that was really fun.  My dad basically let me drive the whole time it was awesome.  My life is seeming more and more purposeful.  I just love being in God\'s presence and worshipping him it\'s amazing.  I\'m hoping that everyday I live my life a lil more true to who I\'m supposed to be.  I\'m sick of being fake...and sick of putting up fronts:S.  I\'m aiming to be totally real and just being who God wants me to...Instead of being soo stupid and being who I made up myself to be.  Right now My life is pretty distraction free soo it\'s pretty awesome.  I don\'t have any friends tieing me down...or any girls I\'m likeing or any girlfriends:P.  It makes God way more accessible and makes it way easier to find time to get close to him.  I\'ve spent soo long just at the same place:S...Like where I wanted God with all of me but there was soo much holding me back.  I was afraid of stuff and now I\'ve given it to God.  Life\'s looking brighter everyday.  Idn what my future is but I know God\'s plans r gonna be played out. anyways:P...If u care to comment do it..otherwise w/e haha God bless ya guys cya later   '),(22039,3418474,'22,May,2004','2004-05-22 00:00:00','    Another day another dollar...I\'ve done nothing today:P cept some chores and I also watched Peter Pan the movie:D haha it\'s actually pretty good it kinda exactly copied the disney cartoon version tho.  My life right now is good for the most part.  I\'m trying with all of me to focus on God with everything I do.  I\'ve made soo many mistakes in these past few weeks and it\'s been a real bummer cuz I\'ve lost alot of people and also...grades:P.  I\'m slowing finding the real me.  The real me is who God\'s created me to be not who I was trying to be.  Haha these past few weeks I\'ve been also really fake just cuz I didn\'t know how to handle the pressure:s...and there again I lost alot because of it and gained a whole bunch of things I don\'t need.  I want to be exactly what God created me to be.  I know this is a bold statement, but I wanna just do w/e God asks and give my life to him...I was afraid to do this for soo long thinking maybe he wouldn\'t get me the girl I want, maybe he wouldn\'t get me the friends I need, maybe I\'d just end up as some pastor somewhere and just not having fun and just not having very much money:P these have all been doubts of mine, but now...Me and Jesus hand and hand are gonna walks this....and if I fall he\'ll pick me up...If i\'m tired he\'ll carry me...if I\'m down he\'ll lift me up...If life feels hopeless he\'ll give me hope:).  God Rules!!!  regardless your situation just embrace God...Life can\'t get any better than when u do.  Another thing...Don\'t fall into Spiritual Pride. Idn who this is for...but God put this on my heart to just write this:P...haha Idn who reads these prolly not many but just don\'t think ur strong enough to do things that u have a weakness towards like say go to a party where there\'s alcohol...or do stuff with friends that aren\'t on fire for God...Or maybe even just u got a weakness with the opposite sex:s but just don\'t think ur strong enough cuz ur not.  God can be ur strength and get u through anything...but if u open up a door to your weakness then Satan can take control of your life....and once he does this you\'ll fall flat on your face. Alright well...Idn if u want things to be prayed for just tell me anyone who cares and reads this:)...I wanna pray for anyone who wants me to. Anyways God Bless u guys:) take care   '),(22040,3418474,'21,May,2004','2004-05-21 00:00:00','    haha I decided to go with the fad:P(y) and make myself a blog.  Haha life for me this past while has been really rough.  Idn...I guess I just lost my footing a bit and God became less of a thing and my friends, school, and parents relationships became more priorities because everyone of those things was screwed:P.  I\'ve made a committment and I\'m giving my all to God.  I want Jesus to be my bestfriend...and man I\'m sick of lukewarm living.  I\'ve been really talking to God lately...and man I\'ve just given like all these problems I have to him...and like nothing has really been solved...but I\'ve got a peace about it.  My relationship with my parents has been kinda junky this past while...I\'ve been constantly having to humble my self and apologize, but I\'m finding God gives me an obedience that I can\'t help but do what they ask.  The Friends Sitch is pretty trashed:P...I kinda blew it a bunch with a bunch of different people.  Man  Idn who God has for me as friends...but I\'m hoping God brings someone quick.  And I know u think like o u got tons of friends:P like...I\'m looking for close friends...not shallow hang out friends.  Once u\'ve had a best friend u can\'t go without one if u know what I mean...It\'s just like someone is there and they care soo much for u and they just wanna talk to u:)...and just know how ur doin.  Jesus and me r gonna get super tight...but if u know what i mean...it\'s kinda like u need someone who u can see and someone who\'s physically there:P Idn.  Welp Anyways...I\'m really excited bout what life has in store,  I\'m doin this thing for God...and I\'m gonna do w/e that takes:)...GOD RULES EH???  Blessed be his glorious name!!!:D haha reply if u wish:) haha comments are nice   '),(22041,3418474,'22,June,2004','2004-06-22 00:00:00','    Man summer is here Yea baby!!!!!:)  I decided to leave all these thing up to God again.  I\'m feeling great about the friends situation...God will provide all I have to do is trust in him.  TOday I worked down at my dad\'s job site and that was kinda wearing on me cuz i went to bed at one and I think I woke up at 6 this morning cuz my dad woke me up to get me to work:S and I didn\'t even know I was gonna work:S. Then I came home and put up lawn furniture and skateboarded:P...haha weird eh I think I might start skateboarding again.  :P Haha that was my day and most people already probably heard an exact replica:P.   It\'s soo hot out it\'s amazing:) I wanna get myself a wicked tan.  I\'m taking off to Cali tomorrow to ride the waves:P...haha I wish naahh I\'m goin to Saskatchewan to ride the tractors hahaha:P bad comparison.  But yea hopefully I have fun down there or over there.  Grrr :@ k well I\'m absolutely done with living even the tinyest bit away from my Jesus.  I\'m gonna go for it all and get as close as I can.  I\'ve given alot of thought to the things I\'ve said on this blog and I\'ve decided that I should fulfill em.  I know the right way and since i do I\'ll be judged accordingly by God on the day when I die.  I want to have a wicked crown when I get to heaven haha maybe not wicked cuz wicked is from hell:P just awesome and yea man I wanna get myself treasures in heaven.  I wanna live this thing for God:).  Man I\'d kill to here the words \" well done my good and faithful servant\":) that\'d be soo soo amazing.  This has been long soo far but keep reading:P...  God has my life in his hands now...I\'m given myself to him. I wanna live by the spirit and bye what he wants.  God kicks Satan butt. The chorus of one thing by Finger 11 is killer to describe what I feel:  If I traded it all If I gave it all away for one thing Just for one thing If I sorted it out If I knew all about this one thing Wouldnâ€™t that be something  Like Man i wanna trade it all away for one thing...It\'s tough tho  I\'m sick of my living tho soo it\'s my only option and I know it\'s the right one cuz I feel the holy spirit telling me to give myself away to God and trust him.  Alright guys that\'s a lil peek into my life I know it\'s long but comments are great:) I hate joke comments:P...Unless u can really make me smirk then I\'m fine:)  GOD BLESS YOU ALL and to all a goodnight:P  Oh yea...guys if ur up for the bible study tell me...I wanna arrange a couple meetings this summer for it(y) Adios   '),(22042,3418474,'19,June,2004','2004-06-19 00:00:00','    Today was pretty rad:D it was the last day of school and I finally finished exams and yea I hung out at Lorne\'s for awhile with some people then I went to uth. Wow I\'m tired:S not just physically, but mentally and spirituallly. School has been a drain on my mind...cuz I had to finish alot of hw.  I\'m sick of all the on going battles inside me:S.  Man Jesus called us to be either hot or cold and right now I\'m just sitting in one place.  My pastor said there\'s no such thing as staying in one spot because it\'s an uphill battle, soo if ur doin nothing ur only goin downhill.  I\'m soo spiritually hungry right now...I need to have my tank filled.  Man I want Jesus soo badly right now.  I\'m soo lost with the friends situation:S I have no clue what goin on.  I want Jesus like crazy, but it makes it hard when the friends who were all going for God kinda ditch u and then u haveta become friends with all these people who rn\'t too into the things of God.  I\'ve worried about friends alot this past semester of school.  It\'s taken alot of my mind off of other things.  Now I\'m just putting it into God\'s hands and waiting.  There\'s nothing I can do.  W/e I try to make happen I botch it up.  I need the Lord to help me in all things:).  I\'m sooo glad it\'s summer now...finally I can take a moment to breathe:P and just think about what\'s goin on.  I wanna get my life totally in sink with what God wants.  It\'s gonna be tough to do.  I just need God to provide me with some awesome friends again like I had b4....:PIn these stupids blogs I\'ve talked bout how I want to live for God and stuff and all and I meant it, but I haven\'t changed much since and i want a dramatic change to happen.  Something soo dramatic that people will recognize I\'m a christian by only looking in my eyes.  I want Jesus written all over me. :S I\'m tired of lukewarm living. I\'m tired of ups and downs and I\'m tired of worrying about my future when God has control.  I need to live christ.  Man:s the battle is insane.  Be praying for me :) I need as much of it as I can get.  I want to learn my life purpose and what God has planned for me and what direction I\'m supposed to head.  I\'ll pray for anyone who sends me a prayer request:D just email it to me.  Whoever did the Bible study or worship thing with me...Do u guys wanna do that again sometime??? that was cool.  Like...Mark, Joel, Steph, SOnja, Nicole, umm Lorne haha whoever reads this and who\'s up for it we should do something.  Anyways God Bless You All and COmments are great:) haha especially when u don\'t put it anonymous(y) Anonymous ones suck:P  Haha gosh this is long cya later   '),(22043,3418474,'08,June,2004','2004-06-08 00:00:00','    Wow! summers almost here:D!!! It\'s wicked.  Gosh man Life\'s changing soo rapidly...Man I kinda got a slew full of different friends.  Haha I don\'t know what happend with all my friends:S...It\'s kinda a bummer letting go of alot of them. But they don\'t seem to miss me or nothing.  I\'ve been trying to be more real with myself:)...and it turns out that if I\'m real with myself I\'m not all that great of a person haha  actually crappy is a better term...It\'s almost like I saw myself like alot without God and I realized like man the real me without God is terrible and I gotta have him fully in my life.  :S I\'m talked soo much bout living this totally for God but man it\'s been tough:S.  I\'m doing fine...Like I\'m not off the target...I know what\'s the truth and now I\'m heading towards it with more intensity then I have been.  I get caught up alot of the time with Self- rightousness and Spiritual Pride and that\'s cuz I am striving for God:S...Idn Mr.Kennedy talked about how striving for God isn\'t what we should be doin it\'s more like we gotta live by the spirit and like that\'s how in turn we\'ll get close to God.  It\'s not like u keep up living a moral life and u get close to God.  That\'s what we all buy into.  we all end up trying to \"live for Christ\" yet we just basically limit oursleves to some dumb rules and crap and we don\'t go anywhere:S.  Man we gotta live this for Christ but the way we do it is living by Faith and living by the Spirit.  God rocks...but don\'t limit yourself:)...U\'ll always fall short with Stupid rules.  Just live by the spirit and let the holy spirit decide choices for u.  Thinking bout it that way I think it\'s rad cuz it\'s almost like God is making decisions for u:D haha that\'s awesome.  Haha he could make u something way better than u make yourself.  Give ur all to him.  Don\'t be afraid:)...that\'s what Satan wants.  Don\'t give in to the thought that God doesn\'t have ur best intentions in mind.  God bless u:)and if could comment that\'d be great...And if u comment I\'d like it alot if u put your name to it:)   '),(22044,3418474,'03,June,2004','2004-06-03 00:00:00','    Oh man it was awesome outside today:D!!! It was soo sunny and beautiful:).  I got out early today and soo I just basically did homework till lunch time then I went out to mcdonalds and then came back and did more hw:P.  I\'m really trying to boost my marks up in this final stretch and do all the assignments I\'ve missed.  I\'ve been pretty busy with that and I\'ve been kinda tired from staying up too late:P.  Life with God is pretty good...I\'ve told about what I want and who I want to be in Christ and now I haveta walk it out.  Reflecting Christ\'s image is tough cuz man he\'s perfect...but that\'s kinda my aim to just have Jesus written all over me:).  I\'ve said this lots like who I wanna be in christ...but I haveta fully take action and just give my all.  Sometimes It\'s hard to give ur all and I\'m really working on it.  Satan just grips us with fear:S and he\'s makes us pretend to be things we\'re not and makes us live in insecurity and fear of our future.  Giving your all to God frees us from the fear and being afraid what\'s goin to happen.  Man...ANother thing that\'s kinda bumming me out is how fake everyone is:( It\'s just soo lame.  I just rather people be who they really r and not put on fronts.  I rather people not say they\'re christians if they don\'t even want God until like it\'s uth group or something and they put on a front \"o I\'m tight with God\" but really they\'re not close to God at all.  People all deserve a chance don\'t get me wrong...but if u rn\'t even trying there\'s no point in faking it.  Live life like there\'s no tomorrow.  Do this thing for Christ:).  Be the real u whoever that might be.  Don\'t feel pressured into christianity.  If ur not real about it either get real or just be the person who u really are acting like out of church and out of the christian groups.  Be the person u\'d be like when u hang out with Non- christians...cuz then that\'s more where ur heart is and it\'s less fake.  I want u guys to all go for God.  But be real:)  God bless u all and guys if u ever need prayer I\'ll pray for anyone just send me prayer requests or something:D take care   '),(22045,2824507,'29,February,2004','2004-02-29 00:00:00','    It\'s February 29th...that\'s something we only get to say once every 4 years.  That\'s right, it\'s Leap Day and I feel bad for all the babies around the world who will be born today.  They can either choose to celebrate their birthdays on Feb 28 or March 1, or they can only have birthdays every four years.  If they choose the latter, then they would be alive for 84 years before technically reaching their 21st Birthday.  Not a bad way to stay young I guess.  Update from Haiti.  Aristide has fled the country and will soon be replaced by another regime just waiting to be overthrown.  Today I\'m going to New Jersey to visit some family members and (gasp) might not be home in time to catch the beginning of the Academy Awards.  Oh well, that\'s what VCRs are for.  After all, can\'t miss Billy Crystal\'s opening monologue.   '),(22046,2824507,'28,February,2004','2004-02-28 00:00:00','    Rumors are FLYING that the United States has already captured Osama Bin Laden.  Those rumors have come out of Iran, always a highly reputable source.  The U.S. and Pakistan have strongly denied those reports since Iran is one of those bad countries who lies all the time.  But we did admit to having intensified the search for Bin Laden which leads one to wonder if Bush has an October Surprise in mind.  Bill Maher thinks that the Republicans are going to drag him onto the stage of Madison Square Garden during their convention.  I say that if you\'re going to bring Bin Laden to MSG, do it at WrestleMania in two weeks so he can get his ass whupped by the WWE superstars!   Big mess going on in Haiti.  Very messy.  What a mess.  We should help them out right?  Nah, they don\'t have anything we need.  Forget it.  The big winner at the Razzie Awards was...  \"Gigli\"!  Worst Picture!  Worst Director!  Worst Screenplay!  Worst Actor for Ben!  Worst Actress for J-Lo!  Tomorrow\'s Oscar contenders can only hope to do so well.   '),(22047,2824507,'27,February,2004','2004-02-27 00:00:00','    Goodbye, Richard Hatch.  The original Survivor winner was unanimously voted out last night and was his cocky self until the end.  As predicted, they first merged into two tribes of seven.  After Ethan successfully caught a fish, his new team may have gotten the impression that heâ€™s a good provider, making Richard expendable.  (That and Richâ€™s constant nudity may have gotten on their nerves)  One person who sure didnâ€™t appreciate Richâ€™s nudity was Susan Hawk.  He not so inadvertently rubbed up against her during the immunity challenge, causing CBS to flash a disclaimer warning.  (They sure are jittery after the Janet incident)  Previews for next week show Susan ranting to Jeff Probst about it so maybe the network is worried about potential for a sexual harassment claim.  Richard just canâ€™t avoid controversy!     The Academy Awards are Sunday and below are my predictions as to who will win.  Picture : â€œThe Lord of the Ringsâ€ Director : Peter Jackson   â€œThe Lord of the Ringsâ€ Actor : Bill Murray   â€œLost in Translationâ€ Actress : Charlize Theron   â€œMonsterâ€ Supporting Actor : Tim Robbins   â€œMystic Riverâ€ Supporting Actress : Renee Zellweger   â€œCold Mountainâ€ Adapted Screenplay : â€œMystic Riverâ€ Original Screenplay : â€œLost in Translationâ€ Foreign Film : â€œThe Barbarian Invasionsâ€ Animated Film : â€œFinding Nemoâ€ Original Song : â€œThe Lord of the Ringsâ€ Original Score : â€œThe Lord of the Ringsâ€ Cinematography : â€œMaster and Commanderâ€ Art Direction : â€œThe Lord of the Ringsâ€ Film Editing : â€œThe Lord of the Ringsâ€ Costumes : â€œThe Lord of the Ringsâ€ Sound : â€œThe Lord of the Ringsâ€ Sound Effects : â€œPirates of the Caribbeanâ€ Visual Effects : â€œThe Lord of the Ringsâ€ Makeup : â€œPirates of the Caribbeanâ€ Documentary : â€œFog of Warâ€   '),(22048,2824507,'26,February,2004','2004-02-26 00:00:00','    I just don\'t have anything I feel like saying today.  But on second thought, it\'s Thursday which means that when nothing else comes to mind, I can always give a \"Survivor\" prediction.  With 14 castaways left, they\'re teasing a surprise.  Methinks that they are going to merge them into two tribes of seven.  Not knowing what the new tribe makeup will look like, I cannot predict who will be voted out!  But it should make for a good episode.   '),(22049,2824507,'25,February,2004','2004-02-25 00:00:00','    Itâ€™s Ash Wednesday.  Party down.  Lots of controversy about â€œThe Passion of the Christâ€, which opens today.  I dunno, when it comes to movies about the crucifixion, I think Iâ€™ll stick with Monty Pythonâ€™s â€œLife of Brianâ€.  Remember to â€œAlways look on the bright side of death...Just before you take your terminal breath.â€  Disturbing news yesterday that Bush is backing a Constitutional amendment banning gay marriages.  The good news though is that its not going to happen.  Republican Congressional leaders have pretty much conceded that they donâ€™t have the votes.  So why did he do it?  Most likely to rally his electoral base.  That and a recent speech where he blasted John Kerry and John Edwards shows that he is getting worried!  Three more primaries and caucuses were held last night, all of which were easily won by Kerry.  The most amusing part was that in Hawaii, the Democratic Party leader conceded that they would probably get a low voter turnout because everyone would be staying home to watch and vote for a Hawaiian contestant appearing on â€œAmerican Idol.â€  Hey, at least they know which elections matter!  On â€œ24â€ last night, the terrorists released their biological virus into the air vents of a major L.A. hotel.  Too bad that Kim wasnâ€™t in the hotel.   '),(22050,2824507,'24,February,2004','2004-02-24 00:00:00','    Today is Fat Tuesday.  Party down.  Gay marriage.  Itâ€™s the great civil rights issue of our time.  Whatâ€™s happening in San Francisco is downright inspiring as gays from across the country are making a pilgrimage to obtain the same legal right that straight people have always enjoyed.  Mayor Newsome deserves credit for having the courage to go through with this.  Not surprisingly Bush is making noise about a constitutional amendment banning it.  (Way to endorse discrimination.)  I donâ€™t know how long the marriages will be permitted to continue but I believe that one day, as Americans open their minds and their hearts, such marriages will be commonplace.  We will then look back upon these days as we currently look back upon the days when other minority groups were discriminated against.   '),(22051,2824507,'23,February,2004','2004-02-23 00:00:00','    Women all over the country, including many of my friends, are mourning the end of â€œSex and the City.â€  I never watched the show during its first five seasons, but decided to see what all the hype was about this year.  I found that beyond all the shock value, it was a genuinely well acted and well written program.  Each of the four main characters demonstrated considerable growth in their characters and last night we saw just how mature they have become.  Charlotteâ€™s determination to have a child despite her inability to conceive was rewarded with the news that she would be adopting a baby from China.  She was so assured that the adoption process would work out that she was not upset to learn that a previous donor had backed out.  Mirandaâ€™s initial reluctance to settle down and embrace love diminished as she not only embraced living with Steve in Brooklyn but even volunteered to take in his ailing mother.  Samanthaâ€™s battle with cancer led her to recognize the good thing she had in front of her and, for the first time, embrace a monogamous relationship.  And Carrie realized that her Parisian fantasy with Petrovsky was not quite the fantasy she had envisioned and returned to her true love â€“ New York City.  Plus, with Mr. Big (or should I say John) moving to NYC to live with her, it leaves the storylines open-ended enough to sustain an upcoming movie.  So donâ€™t worry, my â€œSexâ€ obsessed friends, the movie should indeed happen...one of these days.   '),(22052,2824507,'22,February,2004','2004-02-22 00:00:00','    \"After careful thought and my desire to retire our supremely selected president, I\'ve decided to run as an independent candidate for president,\" Nader said during an interview on NBC\'s \"Meet the Press.\"   Unbelievable.  That quote demonstrates just how deluded the man has become.  On to happier thoughts.  As pitchers and catchers report to spring training, thoughts drift to nice weather and the upcoming baseball season.  Below, I offer my predictions as to what the standings will look like.  Don\'t hold me to it!   NL East  1. Phillies 2. Marlins 3. Braves 4. Mets 5. Expos   NL Central  1. Cubs 2. Astros  (Wild Card) 3. Cardinals 4. Brewers 5. Reds 6. Pirates   NL West  1. Dodgers 2. Giants 3. D-Backs 4. Padres 5. Rockies   AL East  1. Yankees 2. Red Sox  (Wild Card) 3. Blue Jays 4. Devil Rays 5. Orioles   AL Central  1. Royals 2. Twins 3. White Sox 4. Indians 5. Tigers   AL West  1. Mariners 2. Angels 3. A\'s  4. Rangers   Playoffs  Phillies d. Astros Cubs d. Dodgers Yankees d. Royals Red Sox d. Mariners  Cubs d. Phillies Red Sox d. Yankees  Cubs d. Red Sox    OK, so I\'m partial to the cursed Cubs and Red Sox franchises.  But they came so close last year and have each improved themselves in the offseason.  If they fail to deliver this year, then they truly are cursed.   '),(22053,2824507,'21,February,2004','2004-02-21 00:00:00','    Tomorrow, Ralph Nader is going on \"Meet the Press\" and will announce that he is running for president again.  To which I say to him, \"You motherf---ing son of a b---h!\"  There, I feel better now.  Actually no I don\'t.  Nader\'s narcissism astounds me.  Is he in the side pocket of the G.O.P. or something?  So many Democrats and liberals are begging him not to run, hell even the Green Party won\'t have him on their ticket!  Without Nader in the 2000 race, we win Florida, we win New Hampshire, we win several other states that would have put Gore in the White House.  All we can hope is that those who voted for Nader four years ago refuse to do so again and that he becomes a fringe candidate with zero appeal.  Speaking of fringe candidates how about that Lyndon LaRouche?  If you\'re asking \"Who?\", I don\'t blame you.  He\'s basically a nutjob who\'s been running in the Democratic Primaries.  Thankfully he was not invited to any of the debates and has received no mainstream attention in the press.  The only attention he did get is when one of his supporters began heckling Howard Dean at a rally several weeks ago.  He had to be subdued by many of the attendees, including Al Franken.  Which reminds me, I highly recommend Franken\'s book \"Why Not Me?\" - a comedic look at what would happen if Franken ran for president and won.    Did you hear the latest from Bush\'s budget office?  They want to consider fast food jobs at places like McDonalds, manufacturing jobs in order to inflate the U.S.\'s job gains statistics.  That\'s called cooking the books - supersized!  Cathy, the comic strip character, got engaged this week to her long-time boyfriend Irving.  Immediately after that, Marcie and Peppermint Patty visited the San Francisco marriage licence office.  Ba-dum-dum.   '),(22054,2824507,'20,February,2004','2004-02-20 00:00:00','    This site is now the 4th selection on a Google search for \"Phil Abramson\".  Moving on up.  Take that, \"Close Encounters\" Phil Abramson!  That reminds me of an article I read yesterday about online search engines.  It said that Google and Yahoo are going their own ways.  So when we search for something from now on, we\'ll get a different array of options.  Before, Yahoo had just used the Google search results.  Unfortunately, the new Yahoo does not list this page.  The nerve!  Thus, my conclusion is that with their increased search range, Google is the way to go.    Last night on Survivor, I was right in predicting that Chapera would lose the immunity challenge but I was wrong in who would get voted out.  The unlucky one was Rob Cesternino.  Can\'t say that I\'ll miss him.  For someone who was supposed to be such a great strategist, he miscalculated in trusting Boston Rob.  Plus, he should have insisted on participating in the puzzle challenge rather than leaving it up to Boston Rob, Big Tom, Alicia and Sue - neither of whom are any great brains.  With 14 contestants left, they\'re promising a twist.  I think they\'ll realign the tribes and make two teams with seven members each.    Have a good weekend!   '),(22055,2824507,'19,February,2004','2004-02-19 00:00:00','    WrestleMania XX has crept up upon us.  It is only 24 days away and I will be there with the olâ€™ nVo crew.  Briar, if youâ€™re reading this, yes youâ€™re travel plans for that weekend are fine with me.  In the spirit of merging wrestling with politics, I am providing a link to an article I wrote in 2000, likening the Florida recount mess with a pro wrestling tournament.   urlLink Check it out!   Time for Survivor predictions.  I think Susan is toast.  Last week, Big Tom called her â€œa hag from hell.â€  She then went on to distance herself from the rest of the Chapera tribe.  If they lose the challenge, she wonâ€™t have any allies to protect her.  If Saboga loses, Ethanâ€™s status as a past winner might do him in.  The Mogo Mogo tribe is harder to call.  They lost a member last week without voting so itâ€™s unclear who is alligned with who.  But then again, I donâ€™t think theyâ€™ll lose the challenge.  Stay tuned to this blog in the upcoming days when I make my predictions on the 2004 baseball season!   '),(22056,2824507,'18,February,2004','2004-02-18 00:00:00','    In light of his disappointing third-place finish in Wisconsin last night, Howard Dean has returned home to Vermont.  At 1:00 this afternoon he is going to announce the suspension of his campaign but, fear not Deaniacs, he will remain on the ballot!  Well duh.  Even Carol Mosley Braunâ€™s name has remained on the ballot.  For all intensive purposes though, this is a drop-out.  In the span of 30 days, he fell from front-runner to also-ran who couldn\'t win a single primary.  Quite a story for the political almanacs.  Itâ€™s not over for John Edwards though who came in a strong second last night.  Just as he was beginning his speech, Kerry took the stage, bumping Edwards off of the major news networks.  Now thatâ€™s playing hardball.  So it looks like this campaign will continue at least until March 2 and I say that this is good news for Kerry.  He needs to hone his campaigning skills and Edwards will provide him with healthy competition.  Plus, Edwards is not likely to go negative on him.  Plus, heâ€™ll remain in the news and if he wins big on March 2, the media will make a bigger deal of it.   At this point theyâ€™ll take all the free press that they can get.  The only thing I heard about Kucinich recently is that he was trying to get a date on the Tonight Show.  I knew that was the only reason he was running for president... to get chicks.  Alexandra Polier, the girl who was falsely alleged of having an affair with Kerry, is on Friendster!  The Daily News ran a story copying her profile where she described herself as â€œJust another hot piece of ass with a philosophy degree...â€  It appeared to be real though as the profile was created in May, 2003.  I logged on this morning and found out that I am connected to her with only three degrees of separation!  Of course, by this morning she took down all of the details in her profile and all that remains are the basic â€œname, age, occupationâ€ stats.  Last night on â€œ24â€, Nina Myers got whacked!  Gangland style.  Although I wish it were Kim who got whacked instead, Ninaâ€™s execution was highly satisfying.   Jack : â€œYou donâ€™t have any more information, do you Nina?â€ Nina : (whimpering) â€œYes, I doâ€ Jack : â€œNo you donâ€™t.â€ BLAM!  BLAM!  BLAM!  BLAM!   '),(22057,2824507,'17,February,2004','2004-02-17 00:00:00','    Fame is growing!  A Google search of the term \"Phil\'s Spin Zone\" reveals that this site is the first choice.  And a Google search for the name \"Phil Abramson\" reveals that this site is the 7th selection.  We\'d be higher if it weren\'t for the other Phil Abramson, the guy who did the art direction for \"Jaws\" and \"Close Encounters of the Third Kind.\"   '),(22058,2824507,'16,February,2004','2004-02-16 00:00:00','    In previews for tonightâ€™s episode of â€œ24â€, theyâ€™re hyping a confrontation between Nina and Kim with hints that Kim might be murdered.  Please donâ€™t tease us like that!  Most â€œ24â€ fans (yours truly included) have been rooting for Kimâ€™s death for the past three seasons.  If they fail to deliver, weâ€™ll be mighty disappointed.  Yankee fans sure are excited about A-Rod coming.  If he doesnâ€™t hit in April though, will they call for his head as they did Giambi?  Today is the Wisconsin Primary.  Will Dean finally drop out or will his delusion continue?  My vote goes for delusion.    The Drudge Report is usually a fun place to visit while surfing the internet but they caused some anxious moments last week for the Kerry campaign.  Drudge spread rumors that Kerry had an affair with a young woman.  No proof whatsoever mind you, just rumors.  Well the young lady released a statement yesterday denying that any affair took place.  Now the story can go away and the G.O.P. can search for something else to smear Kerry with.  The comic strip of the day is â€œThe Boondocksâ€... Hueyâ€™s Friend : â€œWell they finally found a couple of people who saw George Bush serving in the Air National Guard during those months in 1972 when he allegedly never reported for duty.â€ Huey : â€œReally?  Who?â€ Hueyâ€™s Friend : â€œRay Charles and Stevie Wonder.â€ Huey : â€œYou stupid.â€   '),(22059,2824507,'16,February,2004','2004-02-16 00:00:00','    Another freezing cold day in NY.  Good thing I\'m off today.  The winter hibernation continues.  Congrats to Eddie and Chavo Guerrero on respectively winning the SmackDown Heavyweight and Cruiserweight championships last night at No Way Out.  They lie, they cheat, they steal!   '),(22060,2824507,'15,February,2004','2004-02-15 00:00:00','    Anyone up for a friendly game of Russian Roulette?  I watched \"The Deer Hunter\" this weekend.  Wow - what a powerful movie!  It was 25 years ago that it won Best Picture, drawing a vivid depiction of the horrors of Vietnam.  Apparently there was some controversy at the time because there wasn\'t any evidence that the Viet Cong actually used Russian Roulette as a torture device.  It would be reassuring if they didn\'t because that is some scary shit!  Especially the part when Christopher Walken, traumatized from his P.O.W. experiences, goes AWOL from the Army and starts voluntarily playing for fun.  Shiver...  If you want to see a happier movie, see \"Miracle\", a fun retelling of the triumph of the 1980 U.S. Olympic hockey team.  The New York baseball world got pretty shook up over the weekend as the Yankees have acquired A-Rod from Texas.  Now I\'m a Mets fan so all I can do is sigh as Steinbrenner opens his wallet for another superstar.  No doubt A-Rod is awesome and the Yankee lineup is going to be imposing this year.  But if the movie \"Miracle\" taught us anything it\'s that championships are won by teams, not collections of All Stars.  The \'03 Marlins and \'02 Angels didn\'t have any \"stars\", instead all 25 guys played well together collectively.  That\'s the challenge for the Yankees, ensuring that all of these guys egos don\'t get in the way of the team\'s goals.  No work tomorrow.  God bless the Presidents who made Presidents Day possible!   '),(22061,2824507,'14,February,2004','2004-02-14 00:00:00','    Yay!  My blog now has a section where people can leave comments.  Not as if anyone is actually READING this!  :)   '),(22062,2824507,'13,February,2004','2004-02-13 00:00:00','    Whew!  A weight has been lifted from my shoulders.  I can\'t go into the details but there\'s been a matter that has concerned me for the past two months.  It was resolved earlier today to the best possible conclusion.  Relief!  Sorry to be cryptic but hey...   '),(22063,2824507,'13,February,2004','2004-02-13 00:00:00','    What is your life rated?  G?  PG-13?  NC-17?   urlLink Take this online poll to find out.   (Mine was PG-13)   '),(22064,2824507,'12,February,2004','2004-02-12 00:00:00','    Itâ€™s Friday the 13th.  Boo!  The front page of the â€œDaily Newsâ€ proclaims, â€œBarbara Bush as youâ€™ve never seen her before!â€  Yes, it vividly shows Wâ€™s daughter in an NYC club... DANCING!... with a BOY!!!  Oh the scandal of it all.  OK, so her dancing involves much dry humping.  Big deal.  Paris Hilton would be proud.  There has been a big controversy at St. Johnâ€™s University where several members of the basketball team are facing expulsion over engaging the services of a prostitute.  Now one of the players, Elijah Ingram, protests being targeted claiming that he was never told of any standards expected of the schoolâ€™s athletes.  â€œYou mean I canâ€™t screw a ho?  Says who?!?!â€  Last night, on â€œSurvivorâ€, Jenna Morasca left the game to be with her ailing mother.  Turns out that the rumor I heard was true.  Unforunately, her mom passed eight days later.  My condolences to the Morasca family.  In better news, Richard fought a shark...and won!  The sucker latched onto Richâ€™s arm for a good half hour until he beat him to death against a rock.  Now thatâ€™s good TV.   '),(22065,2824507,'12,February,2004','2004-02-12 00:00:00','    I almost forgot.  Happy Birthday, Abe Lincoln!!!  (The greatest president of all time according to the Murray-Blessing survey!)   '),(22066,2824507,'11,February,2004','2004-02-11 00:00:00','    This morning I was at the chapel of Greenwood Cemetery.  Itâ€™s a truly impressive piece of architecture.  Although non-denominational, it clearly has predominant Christian overtones.  That\'s a contrast from Vassar Collegeâ€™s non-denominational chapel which seemed to be more balanced.  (I recall being surprised to see Jewish stars engraved on the windows)  Anyway, the purpose of the visit was to prepare for the Parks Commissionerâ€™s annual review of the borough next month and Greenwood was selected for aesthetic purposes.  I look forward to soon touring the cemetary and seeing the graves of notorious NYC legends such as Boss Tweed and Bill the Butcher!  (Both characters were prominently featured in â€œGangs of New Yorkâ€)  Today is Thursday which means that itâ€™s Survivor night.  Since there are three tribes, Iâ€™ll make three predictions as to who will be voted out assuming that each loses the immunity challenge.  On Chapera, Boston Rob and Alicia have butted heads at camp but it appears that Susan Hawk is the most alienated tribe member.  Last week she barely said a word which leads one to believe she is keeping to herself and has not made any alliances.  On Mogo Mogo, the odds are against Jenna Morasca.  Sheâ€™s a past winner and rumors are also circulating that she might have left the show due to her motherâ€™s illness at home.  On Sabaga, they clearly cannot afford to lose a third consecutive immunity challenge.  If they do, Jenna Lewis will target past winner Ethan.  The only question is if Jerri and Rupert will go along with her.  I donâ€™t think they will.  I think theyâ€™ll keep Ethanâ€™s strength and vote off Jenna.  Thus, my three picks for elimination (depending on who loses the challenge) are Susan, Jenna M. and Jenna L.   '),(22067,2824507,'11,February,2004','2004-02-11 00:00:00','    John Kerry\'s Jewish Roots...  Democratic presidential candidate John Kerry, thought by many to be a \"Boston Brahmin\" of Irish-Catholic ancestry, was the grandson of Czech immigrants who had concealed their Jewish heritage.  The story begins in the hamlet of Horni Benesov on the tenth of May 1873--the day Benedikt and Mathilde Kohn had a son they named Fritz. Like his father, Fritz became a simple brewer. Yet it was difficult for him to succeed in an area dominated by German-speaking Catholics. Many Jews hid their religious identity, posing as Gentiles. \"It was easier to do business as a Christian,\" says Prague-based genealogist Julius Miller, who specializes in tracing Jewish lineage. \"Many Jews just stopped practicing Judaism during this period and had no belief at all.\"  On March 17, 1902, shortly before his 30th birthday, Fritz took his wife Ida and infant son Erich to a government office in Vienna and changed their family name. Fritz Kohn would henceforth be known as Frederick Kerry.  John Kerry is his grandson.   '),(22068,2824507,'10,February,2004','2004-02-10 00:00:00','    Iâ€™m not a big fan of cell phones.  Yes, they are extremely useful in case of emergencies and I believe everyone should have one for that purpose.  But more often than not, they are used frivolously and are a huge distraction.  I constantly see people yakking on them while walking or driving through the City, while acting oblivious to their surroundings.  Itâ€™s gotten to the point where some people value these stupid phones more than life itself.  Case in point, Lina Villegas, an 18-year old who yesterday, dropped her phone onto a subway track, jumped onto the track to retrieve it, and was then killed by an oncoming train.  Itâ€™s splashed across the front page of todayâ€™s â€œDaily Newsâ€ as a huge tragedy and, while it is indeed tragic, I have no sympathy for such stupidity.  First of all, odds are good that the train would have glided over the phone with no damage done and she could have called a member of the Transit Authority for assistance in retrieving it.  But even if it couldnâ€™t have been retrieved, it sickens me that one would risk their life for such a meaningless materialistic item.    Later today, Wesley Clark will announce that he is withdrawing from the Presidential race.  General Clark, I salute you!  I thought he was a fine candidate who would have fared well in a general election.  However, as a novice to politics, his candidacy stumbled in the primaries and he failed to gain any traction.  This became apparent when John Kerry surged ahead of the pack, using his Vietnam veteran credentials to his credit.  Suddenly Clark wasnâ€™t the only candidate who could argue that he was a decorated military vet.  Hopefully, the general will remain active on the campaign trail and who knows?  If we get a Kerry Administration, he might qualify as a good candidate for Secretary of Defense.   '),(22069,2824507,'10,February,2004','2004-02-10 00:00:00','    The winner of the dog show is Josh the Newfoundland!  Which means that the poodle lost.  Yay!  So Josh, mazel woof on your victory.  Drink all the toilet water you want tonight.  You earned it.   '),(22070,2824507,'09,February,2004','2004-02-09 00:00:00','    This just in!  The missing Russian presidential candidate has been found.  And he\'s alive!  I guess he just took a wrong turn somewhere.   '),(22071,2824507,'09,February,2004','2004-02-09 00:00:00','    New York City has gone to the dogs.  The Westminster Kennel Club Dog Show has taken over Madison Square Garden.  This is the 128th annual show, which means that the first dog show took place in 1876 â€“ roughly the year in which most of the judges were born.  I donâ€™t mean to be cruel but some of these judges look like theyâ€™ve been thawed and dusted off just for this occasion.  And why, out of all the adorable dogs to choose from, do they always pick the standard poodle?   They look so unnatural and unnerving.  Maybe the poodle has a special financial arrangement with the judges.  How else do you explain such a subjective decision?  If the poodle wins Best in Show tonight, I wonâ€™t be happy.    My issue of \"Newsweek\" arrived in the mail yesterday.  On the cover it says \"Who Really Killed Jesus?\"  For a second, my eyes didn\'t quite register and I thought it said \"You Really Killed Jesus.\"  Great, there we go blaming the Jews again!  Now Dean is backing away from his statement the other day that he would drop out if he loses Wisconsin.  Perhaps he saw the poll that has him trailing Kerry by 50 points.  Thatâ€™s OK.  He can stay deluded until March 2 when New York and California will seal the deal for John Kerry â€“ The Real Deal!  My friend Ben directed me to  urlLink Lawrence Lessigâ€™s blog .  He is a law professor at Stanford and has some good commentaries on Ralph Nader and how egotistical he would be if he were to run for President again.  Nader always attacked big businesses for not accepting the accountability of their actions.  Now, Nader refuses to accept his own accountability in contributing to Bushâ€™s election in 2000.  Itâ€™s a shame to see what this once admirable man has become.  Tell me what you think of this blog!  Send me feedback at noyes1@aol.com.   '),(22072,2824507,'09,February,2004','2004-02-09 00:00:00','    And we Americans thought that our electoral system was wacky!  In Russia, presidential candidate Ivan Rybkin is missing.  That\'s right, he\'s MISSING!  Now if this were the old Soviet Union, it wouldn\'t be unreasonable to suspect foul play since any dissenters were routinely rounded up and sent to Siberia.  But in this case, am I too naive to think that Russia has moved beyond such days?  Rybkin was not a serious threat to Putin\'s reelection prospects so it might just be a coincidence that he\'s missing.  Or maybe he\'s hiding on purpose in an effort to get some attention and maybe some sympathy votes.  I just hope that Karl Rove doesn\'t get any ideas.  Speaking of Russian leaders, I\'m reminded of a funny pattern that existed throughout the 20th Century.  The pattern is that every other leader had hair and every other leader was bald.  It goes like this :  Lenin (bald), Stalin (hair), Khruschev (bald), Brezhnev (hair), Gorbachev (bald), Yeltsin (hair).  Then Putin (hair) came in and the pattern got messed up.   '),(22073,2824507,'08,February,2004','2004-02-08 00:00:00','    Was I the only one who felt that not enough breasts were bared at the Grammys last night?  It seems that CBS set a standard of expectations last week.  The least they could do is live up to it.    I hate Mondays.   '),(22074,2824507,'08,February,2004','2004-02-08 00:00:00','    W on \"Meet the Press\" this morning : \"Saddam was a danger... Dangerous man... Could have made dangerous nukes...  Had to elimate this danger.\"  I think Jon Stewart said it best the other day, mocking how the administration keeps repeating the word \"danger\", by exasperating... \"Could you say that a few more times?  I\'m not retarded yet!\"   But the best part was Bush\'s insistence that there was no way that an Islamic regime would come to power in Iraq. How can be so sure?  He was assured so by the Iraqi Governing Council\'s Adnan Pachaci.  He of the Iraqi TV series \"Pajoanie Loves Pachaci.\"  OK, that was another joke I borrowed from Jon Stewart.    Last night John Kerry won two more caucuses in Michigan and Washington.  It\'s looking as if Edwards and Clark might flame out this Tuesday and Dean might call it quits the next week.  Then when I go to the polls in New York on March 2 I\'ll have a choice of Kerry or Al Sharpton.  Oh wait, there\'s always Dennis Kucinich.  Can\'t forget about Kucinich.  Saw two very good movies yesterday : \"American Splendor\" and \"Capturing the Friedmans\".  Thanks to Netflix for providing me with home delivery service.  If only there wasn\'t such a long wait for \"Lost in Translation\" and \"Spellbound.\"  It\'s too damn cold out.  My winter hibernation continues...   '),(22075,2824507,'07,February,2004','2004-02-07 00:00:00','    Anyone reading this needs to go out and get an XM Satellite Radio.  It\'s an amazing invention.  You get 100 digital quality channels - 68 of which are commerical-free music stations.  And they represent every genre imaginable.  It puts anything on AM or FM radio to shame.  XM is available by purchasing a receiver and then subscribing for the low price of $9.99 a month.  (Their competitor, Sirius, offers fewer channels for a higher price!)  Plus, my brother works for them!  Thanks for reading this plug.   '),(22076,2824507,'06,February,2004','2004-02-06 00:00:00','    Rain Rain Rain.  Itâ€™s been raining all day with no signs of letting up.  But hey, itâ€™s better than the ice storm we had earlier this morning.  Despite the bad weather, I spent the morning at Coney Island for our first meeting of the year to discuss the Beach Opening!  Nothing like a little bad weather to put you in the mood for some good weather.  The Krusty the Clown â€œDonâ€™t Blame Me. I didnâ€™t do itâ€ game continues in Washington as CIA Director George Tenet denies ever saying that Iraq posed an imminent threat.  OK, but isnâ€™t that what the White House told us?  No!  Word from the White House is that they never used the word â€œimminentâ€ either.  Now I donâ€™t know if they ever used that specific word but, if not, they sure used a hell of a lot of synonyms.  Let the fallout continue!  In sports news, Lennox Lewis, boxingâ€™s heavyweight champion of the world, announced his retirement today.  This creates a vacancy in the heavyweight division that will surely be filled with exciting fighters like...  um... uh...  nope, none come to mind.  Tyson and Holyfield are washed up.  Roy Jones has gone back down to light heavy.  James Toney is out for the year with an injury.  Chris Byrd and John Ruiz are the most boring â€œchampionsâ€ in recent memory.  Corrie Sanders has only one big upset win to his credit.  The only promising fighters on the horizon are the Klitschko brothers, Vitali and Wladimir.  The Ukranians just might be the second coming of Ivan Drago!  â€œIf he dies, he dies...â€  Nah, theyâ€™re not really that good.  But you can bet that HBO will market them that way.  And on â€œSurvivorâ€ last night, Iâ€™m sad to say that 75-year old, NAVY SEAL Rudy Boesch was voted out.  Canâ€™t blame them for doing it since he was a huge liability during the challenges but it was still sad to see him go.  Have a great weekend!   '),(22077,2824507,'05,February,2004','2004-02-05 00:00:00','    This just in!  Howard Dean has told his supporters that he MUST win the Wisconsin Primary and that he will drop out if he loses.  To which his supporters replied, \"We\'re coming in droves to help you out!  Now where the hell is Wisconsin?\"   '),(22078,2824507,'04,February,2004','2004-02-04 00:00:00','    Yesterday the Massachusetts Supreme Court affirmed that homosexuals have a right to marry as they wrote â€œThe history of our nation has demonstrated that separate is seldom, if ever, equal.â€  Props to the Court for recognizing that the issue boils down to equal protection under the law.  Of course, W is in a tizzy over the ruling, decrying the actions of â€œactivist judges.â€  Um no, Mr. President.  Activist judges would be those who put you in office by stopping the Florida recount.      Now in my line of work we receive many inquiries from the public on how they may organize special events in public parks.  These events have ranged from concerts to picnics to ritual sacrifices.  Yesterday I got a phone call from a woman who wanted to hold a Breast Feeding event.  To which my co-worker replied, â€œJeez, do they have to shove it in our faces?â€  The babyâ€™s sentiments exactly.  Gov. Jesse â€œThe Bodyâ€ Ventura is going to teach at Harvardâ€™s Kennedy School of Goverment!  Is it too late for me to enroll?   Iâ€™m still waiting for him to move to California and challenge Gov. Schwarzenegger in the ultimate battle to settle the score left over from that classic film... â€œPredatorâ€.  Joaquim de Almeida is sure getting around!  Less than 24 hours after his Ramon Salazar got bumped off on â€œ24â€, he showed up on â€œThe West Wingâ€ as a South American diplomat wooing C.J.  Now thatâ€™s an impressive resurrection!  My pick to get voted off of â€œSurvivor : All Starsâ€ tonight : Jenna Morasca, the Playboy & swimsuit model who won in the Amazon.   '),(22079,2824507,'04,February,2004','2004-02-04 00:00:00','    Wednesday, February 4  Let it not be said that I was never one to jump on a bandwagon.  Much as my presidential allegience shifted from Joe Lieberman (from Dec. 2000 â€“ Summer 2003) to Wesley Clark (Summer 2003 â€“ two weeks ago) to John Kerry (two weeks ago until hopefully the next eight years!), I am joining the ever-so-popular trend of publishing a web blog.  Why?  Why not.  So thanks for coming!  And I hope to entertain you for many years, months, weeks or until whenever my attention span wears out.   In this space you can find my humorous observations on news, politics, the media, or whatever I feel like satirizing at a particular moment.  I begin today by sending my best wishes to Joe Lieberman â€“ the first Jew to run for president â€“ and, as of today, the first Jew to fail in his presidential bid.  Just as Jews around the world continue to wait for Messiah, we continue to wait for a Jewish president.  Wait, what if they turn out to be the same person?  Now that would be cool.  And by winning 5 out of 7 primaries last night, John Kerry looks more and more dominant in his quest for the nomination.  Now think about this.  Kerry is tall, gaunt, statuesque.  If you put a beard and a top hat on him, who would he remind you of? ...  â€œFour score and seven years agoâ€ ...  Yes, itâ€™s Lincoln!  Kerry should run as if he is the second coming of Lincoln.  Heâ€™d win in a landslide.  Well, maybe not among the NASCAR dad demographic...  In TV news, on â€œ24â€  they killed off the Salazar brothers last night but not before a footchase ensued over possession of the virus.  It seems that everytime I see a footchase on TV these days, I think of the chase music from â€œBenny Hillâ€and I expect to see horny old men chasing after bikini-clad women.  But thatâ€™s just me.  And Kim Bauer continues to act like the biggest moron on television by pulling the Director of CTU away from a huge crisis to discuss her co-worker Chloeâ€™s childcare issues.  For those who do not know, Kim is the dumbest character to be portrayed on any series past or present and she will receive no free passes in this forum!     Anyway, whereas â€œ24â€ has become a bit lame, â€œSurvivor : All Starsâ€ has arrived!  (That was the REAL highlight of Sunday night, not the football game or Janetâ€™s peep show).  Australian winner Tina Wesson was eliminated on Sunday as past winners are being targeted.  Look out Richard Hatch!  But the two most popular survivors ever, Rudy and Rupert, have formed an alliance that hopefully will take them far.  In the meantime, after three days on the island, noone has any food, water or fire.  Letâ€™s see how long they can last before Jeff Probst has to save their sorry asses.       Until tomorrow!   '),(22080,2824507,'31,March,2004','2004-03-31 00:00:00','    If youâ€™re in New York, be sure to tune in to radio station 1190 AM today at noon for the premiere of â€œThe Oâ€™Franken Factor!â€  This is the debut of Air America, the liberal talk-radio network.  I have my walkman with me so I am all ready to tune in!  Word has it that Al Frankenâ€™s first guest will be Michael Moore so it will certainly be entertaining.  And if you donâ€™t live in a market that carries the station, it can also be heard on XM Satellite Radio.  Today is the five-year anniversary of my very first interview with the Parks Dept.  It was March 31, 1999 and the weather was fantastic.  I was a senior at Vassar and woke up early that morning to take a Metro North train into the city.  I first had a 10:00 A.M. interview with Primedia, the publishing company that produces several magazine lines.  In retrospect it was a good thing I didnâ€™t wind up working for them, because shortly after graduating, Primedia wound up laying off many of their employees.  After finishing up with Primedia, I grabbed a quick lunch somewhere (canâ€™t remember where) and then took the F-train into Park Slope for a Parks Dept. interview.  The job I interviewed for at that time though was very different from the one I accepted two months later.  The division I initially interested with did not interest me as much and I left the building that day figuring that I would not return.  Funny how things work!  Afterwards I traveled back up to Poughkeepsie where I had a senior class council meeting that night.  It was also one of the Passover seder nights but considering everything else I had on my plate, I skipped out on the holiday that year.  I did not wake up at 5 A.M. today to watch the baseball game.  Instead I got up at 6:30 and listened to it on the radio as I lied in bed.  The Yankees got even from yesterdayâ€™s loss, beating the Devil Rays 12-1.  The good news for me is that my guy, Carl Crawford, got a hit, a run and a stolen base.  That should move me up the fantasy standings.  After work tonight Iâ€™m seeing â€œEternal Sunshine of the Spotless Mind.â€  It got great reviews and Iâ€™m looking forward to it.   '),(22081,2824507,'30,March,2004','2004-03-30 00:00:00','    Long day.  This is the first opportunity I\'ve had to sit at a computer and write anything.    It started at 5 A.M. when I got up to herald the 2004 baseball season.  The Yankees were playing the Devil Rays in the Tokyo Dome and it counts in the standings as the first game of the season.  Being a frustrated Mets fan, I was amused at watching the Yankees lose by a score of 8-3.  When it comes to my fantasy league, I had one player who I was rooting for to put me on the board and hopefully gain an early lead.  He was Carl Crawford.  Naturally, he went 0 for 5.    Then I went to work where the annual review with Commissioner Benepe at Green-Wood Cemetery finally took place.  All went well as our two months of preparations paid off.  My boss, Marty Maher, sure worked tremendously to put it together.  It was especially nice to hear Commissioner Spiegel thank me by name in his speech.  I could talk about the meeting and the presentations but all of that is boring shop talk and I don\'t think anyone reading this blog would be interested.  But I do want to give Chief of Operations Charlie Gili credit for the anti-littering public service announcement.  It was a great piece merging the \"Crying Chief\" PSA from the \'70s with footage of Charlie and his staff, teaching a Brooklyn-style lesson to a litterer.    After the meeting, the Director of Green-Wood Cemetery took about 20 of us on a walking tour of the grounds.  It was amazing to see the care put into many of the stones that were created in the 19th century.  We passed Leonard Bernstein\'s gravesite and I was surprised at how simple the footstone was.  All it had was his name and the years he was born and died.  There wasn\'t even a headstone.  We also visited the grave of Edwin Litchfield.  Litchfield is the man who in the 1850s constructed and lived in the building where I currently work, Litchfield Villa.  We also stood on the highest altitude level in Brooklyn and from there we were able to look down upon the Statue of Liberty and the Manhattan skyline.  It\'s an amazing place, full of NYC history. I intend on returning to Green-Wood on May 29 to go on their \"Gangs of New York\" walking tour.  That will be based around the gravesites of Boss Tweed, William Cutting (Daniel Day-Lewis\' character) and victims of the Civil War Draft Riots.   '),(22082,2824507,'29,March,2004','2004-03-29 00:00:00','    The stomach virus is gone.  Yesterday was the first day in three days that I didnâ€™t throw up.  Too much information for you?  I donâ€™t care!  Iâ€™m just psyched that Iâ€™m feeling better!  Noticed another Vassar reference in pop culture last night.  During â€œThe Sopranosâ€, Tony and Carmela went to A.J.â€™s high school to visit with his college advisor.  Inside of his office there was a very noticeable Vassar poster hanging on the wall.  Whereas Meadow may have been admitted, A.J.â€™s 900 SAT scores kind of meant that he was out of the running.  Speaking of Meadow, her TV movie where she plays Heidi Fleiss got TERRIBLE reviews in todayâ€™s newspapers.  Normally Iâ€™d be psyched to see Jamie-Lynn DiScala prancing around in her lingerie but David Bianculliâ€™s zero-star review gives me pause.  Ah, the hell with it, of course Iâ€™ll tune in!   '),(22083,2824507,'28,March,2004','2004-03-28 00:00:00','    \"Lip my stockings.  Lip my stockings!!!\"  This quote is from the scene in \"Lost in Translation\" where a...shall we say...escort visits Bill Murray\'s hotel room and volunteers her services.  After a hilarious back and forth, he realizes that she wants him to rip her stockings.  This scene and many many others make \"Lost in Translation\" a joy to watch.  I initially saw it in November and thought it was very good but I was even more impressed with my recent second viewing.  Sofia Coppola\'s writing and directing almost redeems herself from her \"Godfather III\" disaster.  Almost.  But seriously, she does a great job portraying the isolation one feels being alone in a strange land, and the relationships that can develop between similar lost souls.  Scarlett Johannson\'s predicament was similar to Carrie Bradshaw\'s in the \"Sex and the City\" finale.  Don\'t be surprised though if you feel jet-lagged at the film\'s conclusion.  Still in a Japanese mood, I caught some of the Yankees exhibition game today.  They played the Yomiuri Giants at the sold-out Tokyo Dome.  It\'s amazing how the Japanese people are even more enamored with baseball than Americans.  One day I will have to make sure to visit the country, only after my trips to London and Rome though.   '),(22084,2824507,'27,March,2004','2004-03-27 00:00:00','    Not too long after I wrote yesterday\'s blog entry, I started feeling ill again.  Since I already finished the work I needed to do, I left work early and took a nice 3-hour nap at home.  Still felt nauseous throughout the night but have been feeling great since waking up this morning!  In the meantime I\'m being careful about what I\'m eating.  The latest Netflix films to arrive are \"The Rundown\", \"Intolerable Cruelty\" and \"Open Range\".  Reviews to follow!  Baseball season starts at 5 A.M. on Tuesday morning as the Yankees play the Tampa Bay Devil Rays... in Japan!  I\'m all ready for the season as my fantasy team is ready to go.  Special thanks to Tim for serving as commissioner of \"Fantasy\'s Finest\" on yahoo.com.  My team is called \"Thugz Inc.\" and it features an all star lineup of catcher Jason Kendall, first baseman Todd Helton, second baseman Marcus Giles, shortstop Jimmy Rollins, third baseman Mike Lowell, outfielders Luis Gonzalez, Jim Edmonds, Shawn Green, Carl Crawford, Carlos Lee, and utilityman Desi Relaford.  The starting pitching is my weak link but the ten guys that I have are Livan Hernandez, Brian Anderson, Steve Trachsel, Jae Weong Seo, Horacio Ramirez, Kirk Rueter, Tomo Ohka, Ramon Ortiz, Elmer Dessens and Ryan Franklin.  The bullpen consists of Robb Nen, Arthur Rhodes, Rocky Biddle, Shigetoshi Hasegawa and Rafael Soriano.  C\'mon guys, win me some money this year!   '),(22085,2824507,'26,March,2004','2004-03-26 00:00:00','    Oy.  I feel sick.  The body aches continued throughout the night and into the morning.  Iâ€™ve had no appetite to eat anything and even threw up this morning.  (Hindsight says I probably should not have had a beer last night when I went out with my coworkers)  The temptation was great to stay home today but I had to write our biweekly report.  So I trudged into work and am actually feeling a little better since Iâ€™ve arrived.  Still have no appetite though.  But I already got the biweekly over with so I no longer have to worry about it if I happen to take a turn for the worse.  The annual White House correspondents dinner was held on Wednesday night.  Itâ€™s traditionally an opportunity for the president and other politicians to make fun of themselves.  Bush took part in that tradition by showing a video where he is wandering around the Oval Office, looking underneath the furniture, saying â€œThose weapons of mass destruction have to be here somewhere... Nope, no weapons over there.  Maybe under here.â€  Itâ€™s funny because itâ€™s true!  The families of Iraq war victims donâ€™t seem too think itâ€™s too funny though.   '),(22086,2824507,'25,March,2004','2004-03-25 00:00:00','    The Parks Commissionerâ€™s annual review of Brooklyn kept getting rescheduled and rescheduled.  Its new day is Tuesday and today we held a rehearsal session at the Salt Marsh Nature Center in Marine Park.  (The real meeting will take place in Green-Wood Cemetery)  The rehearsal went well but my back was killing me during the whole thing.  Itâ€™s a reaction to the rain.  Usually, when itâ€™s rainy or damp, I get muscle pains in either my lower back or my legs.  A sign of getting old I suppose.  How great are these testimonies before the 9/11 commission?  OK, theyâ€™re not â€œgreatâ€, but they do make for interesting TV.  Albright, Cohen, Powell, Rumsfeld, Tenet have all gone and now the heat is on Condi Rice to testify.  Richard Clarke socked it to the Bush administration yesterday by apologizing to the victims families for the governmentâ€™s failures.  The Bushies are furious, especially after one woman who lost her son praised Clarke for being â€œbrave enough to talk openly about the truth.â€  John Kerry needs to send Clarke a thank-you card for putting Team Bush on the defensive and distracting their efforts to negatively portray Kerry.  A 16-year old Palestinian teen was apprehended by Israeli forces before he could execute a suicide bombing.  The kid later admitted that he was paid off with $22 for his family and the promise of 72 virgins in the afterlife.  So sad.  Besides, the kid wouldnâ€™t know what to do with 72 virgins.   '),(22087,2824507,'24,March,2004','2004-03-24 00:00:00','    Life has grown stagnant and it has become frustrating.  Changes are on the horizon.  Time to shake shit up.  Last night I decided to lay off the deep thoughts that have plagued me for too long.  So I watched two mindless movies, both of which were a great distraction!  The first movie was â€œOld Schoolâ€.  It reminded me of those college frat movies they used to make in the 80s.  This was even more enjoyable because I could identify with the 30-year old characters who resisted settling down, in favor of some good old-fashioned partying.  Will Ferrellâ€™s comedic skills were a high point of the movie.  I also enjoyed Vince Vaughnâ€™s shtick of signaling for his kid to cover his ears by telling him â€œearmuffsâ€ and then feeling free to curse like a drunken sailor.  And although many of the supporting roles were tired movie stereotypes...the evil dean, the fat black kid, the horny old man...they were still amusing.  The second movie that I watched was â€œParty Monsterâ€.  Wow, Macauley Culkin sure has come a long way from â€œHome Aloneâ€.  His performance and the movie-at-large was similar to a train wreck...a complete mess yet I couldnâ€™t turn away.  It seemed as if Culkin was trying too hard to be over the top.  But that could have been an intentional effect of the story since those Club Kids were the definition of â€œover the top.â€  I particularly enjoyed the efforts of Seth Greenâ€™s character in writing a novel.  Two years to write two pages.  One week to write a sentence.  Almost reminds me of someone!   It looks like tonightâ€™s â€œSurvivorâ€ is going to be a recap show.  Hence, no prediction as to who will go.   '),(22088,2824507,'23,March,2004','2004-03-23 00:00:00','    You know what I hate?  When you call someone and they tell you that theyâ€™re currently on another call but will call you right back.  And then they donâ€™t call you back.  Grr...  One of my officeâ€™s phone numbers was misprinted in the Yellow Book as being the Brooklyn Marriage Licence Office.  Every day we get our fair share of calls from hopeful newlyweds and, in a monotone voice, we dryly inform them of the correct number.  The temptation is too great though to answer their questions and possibly give them misleading information.  Thereâ€™s also an impulse to answer, â€œYes, we DO issue same-sex marriage licences.â€  Wouldnâ€™t that create a nice barrage of publicity?    WWE held a draft lottery last night, shuffling up the Raw & SmackDown rosters.  The coolest part was the Pier 6 brawl at the end between the red shirts and the blue shirts.  Havenâ€™t seen that big of a brawl since the old WCW/nWo days.  A new blog has recently made its debut!  Check out  urlLink Sarahâ€™s Entertainment Spot  for reviews of all things entertaining and not-so-entertaining.   '),(22089,2824507,'22,March,2004','2004-03-22 00:00:00','    Sheik Ahmed Yassin, founder of Hamas, was assassinated last night by Israeli forces.  After engineering the murders of scores of innocent civilians, I say good riddance.  Many readers of this blog may decry Israelâ€™s action claiming that it wonâ€™t help matters and that assassinations are wrong.  That may be true.  It certainly wonâ€™t help in the short term as the followers of Hamas are sure to seek revenge attacks.  But it may very well help in the long term.  Peace in the middle east will never happen as long as terrorist organizations such as Hamas exist and are bent on the destruction of the Israeli state.  Hopefully, by cutting off their head, their influence will eventually diminish.  The U.S. should be so lucky by murdering Osama Bin Laden.    I saw â€œMonsterâ€ this weekend.  It was a good film with an excellent performance by Charlize Theron.  Her physical transformation was astonishing and we got an excellent glimpse of her emotional torment.  Christina Ricci was solid in her supporting role.  I also saw â€œDuplexâ€ on Netflix.  Bad!!!  I wanted to like it since they filmed it a few blocks away from my office in Park Slope and I often walked over to the set to watch.  But it just wasnâ€™t funny!  Ben Stiller and Drew Barrymore tried their best but the screenplay just wasnâ€™t delivering.  â€œThe Passion of the Christâ€ was defeated at the box office this weekend by â€œDawn of the Dead.â€  If they make a sequel of â€œPassionâ€, that could be its subtitle.   '),(22090,2824507,'21,March,2004','2004-03-21 00:00:00','    Click  urlLink here  for a great article about recent statements from former White House terrorism advisor, Richard Clarke.  Clarke maintains that the Bush administration ignored the signficiance of Al Quada for over eight months and even after 9-11, tried to pin the blame on Iraq.   '),(22091,2824507,'19,March,2004','2004-03-19 00:00:00','    Prospect Park looks amazing right now.  The snow is falling and has settled on all the trees.  It was quite a sight to see as I walked toward the park this morning.  Sometimes Prospect Park reminds me of the idyllic Vassar campus but that might be because there arenâ€™t too many other open green spaces in NYC.  I have my radio alarm set in the morning for the Don Imus show on WFAN.  This morning I awoke to Colin Powellâ€™s speech from Baghdad, where he extolled the virtues of the U.S. invasion and occupation.  (Usual talking points from the administration)  After going on for a while, Imus cut him off and began ranting about the lies and the failure to find WMD.  It was pretty funny.  Unfortunately I fell asleep again and didnâ€™t awake until 10 minutes before I had to leave for work.    My dad is in the â€œDaily Newsâ€ today!  Well, sort of.  He sent an e-mail to the newspaper with his opinion on Martha Stewartâ€™s sentencing.  Somehow they got his name wrong.  Instead of â€œStan Abramson from Brooklynâ€, they printed â€œSteve Abramson from Brooklyn.â€  The quote that they printed was â€œShe is not a serial killer and she should not be treated as one.â€  Way to go, Steve!   '),(22092,2824507,'18,March,2004','2004-03-18 00:00:00','    Batten down the hatches.  More snow is coming tomorrow.  We still have Tuesdayâ€™s snow lying on the ground.  I guess winter is getting its last licks before spring starts on Saturday.  In NYC educational news, Mayor Bloomberg fired three members of the new Education Panel because they disagreed with him on his new policy of holding back any third graders who get the worst score on a standardized test.  The proposal passed after he replaced them with three new members who agreed with him.  Now, thatâ€™s playing hardball!  No doubt, Bloomberg is accustomed to such maneuvers from his experience in the private sector.  But this is the public sector where change is resisted at every turn and heâ€™s getting lots of heat over it.   Fellow Vassar alum and fellow Jew, Ethan Zohn, got voted out of â€œSurvivorâ€ last night.  Iâ€™m bummed about it.   His departure was not unexpected though as he was targeted from the beginning because of his status as a past winner.  Now all of the past winners are gone.  With 10 members left, Chapera has a 6-4 lead and are in a position to vote out the remaining Mogo Mogos one by one.    The neighbors next door are renovating their house.  All I hear when Iâ€™m home is â€œBANG BANG BANG.â€  How long will it go on?   '),(22093,2824507,'17,March,2004','2004-03-17 00:00:00','    You might notice that I have made some changes to this blogâ€™s visual interface.  I first decided to go with a white screen as it is easier on the eyes.  I feel a little bad taking the green screen down on St. Patrickâ€™s Day and all but whatever.  The second change is that I decided to do away with the comments feature.  It became a compulsion of mine to constantly check the blog and see if anyone left any comments.  Now I have one less annoying compulsion to deal with.  Besides, my e-mail address is still at the top of the page so yâ€™all can still hit me up if you got something to say.  Whatâ€™s up with the freaking snow storm?  At least I got a good workout last night shoveling.  Best part of â€œThe Sopranosâ€?  When Carmela put in the videotape of â€œCitizen Kaneâ€ to show to her film group (consisting of mob wives), and watching their uncomfortable reactions as the FBI Warning appeared on the screen.  The NCAA basketball tournament starts tomorrow.  Which means that â€œSurvivorâ€ airs tonight!  Iâ€™m still thinking that Alicia will get voted off from Chapera for the same reasons I talked about last week.  If Mogo Mogo goes to tribal council again, theyâ€™ll be in deep trouble as theyâ€™d go into a merge down 6-4.  If that happens, either Ethan or Jerri will be targeted.  My guess is Ethan.  I know, I know.  I watch too much TV.   '),(22094,2824507,'16,March,2004','2004-03-16 00:00:00','    There was an old â€œPeanutsâ€ cartoon where Lucy bemoans, â€œAnother holiday season has come and gone and as usual I didnâ€™t get what I really wanted.\"  Charlie Brown asks her, â€œWhatâ€™s that?â€  Lucyâ€™s reply : â€œReal estate!â€  That about sums up how I feel.  Iâ€™ve become addicted to this program on HGTV called â€œHouse Huntersâ€.  It follows people as they search for new homes or condos.  More and more as I watch the program I think, â€œHey, I want one!â€  But hereâ€™s the dilemma.  I love New York City and want to stay here.  On the other hand, real estate values are in the range of ridiculous.  Right now I have a basement studio apt. in Sheepshead Bay, Bklyn and it currently suits my needs.  But for how long?  Do I need to wait until Iâ€™m in a serious relationship with someone before Iâ€™m in a position to buy or at least pay one of those higher rents?  A single guy who works in the public sector just doesnâ€™t have too many viable options.  Not that Iâ€™m complaining about where I am now but I guess itâ€™s just human nature to always want more.   '),(22095,2824507,'15,March,2004','2004-03-15 00:00:00','    Beware the Ides of March.  Now that that Julius Caesar reference is out of the way, it\'s time to record my observations on last night\'s WrestleMania event!  A great time was had by all.  Briar, Bartow, Jon, Emily and I first met for dinner at Jack Demsey\'s on 33rd & 5th.  It\'s a great place to eat but we\'re worried if they\'ll ever let us back in.  Briar fell off his chair after drinking too many waters, but not before Jon and Emily asked him to be best man at their wedding!  Isn\'t that cool?  Bartow and I will also be on hand as part of the wedding party.  Don\'t worry Emily.  Despite our teasing, it won\'t turn into a wrestling wedding.  No cakes are going to be thrown if I have any say about it!  It\'s all happening on Memorial Day Weekend 2005 and I\'m looking forward to it.  From there we walked to MSG screaming \"Whooo!!!\"  all the way.  The line stretched around the block but it moved pretty quickly as there was no security checking people.  We got to our seats at around 6:45 and it appeared to be the section reserved for people who either works for WWE or knows someone who works for WWE.  We fell into the latter category.  Thanks, Tony!  The Harlem Boys Choir sang \"America the Beautiful\", Vince\'s national anthem.  We then saw a video opening montage that culminated in a shot of Vince, Shane and Shane\'s new baby.  I can\'t wait for the power struggle in 30 years between Shane\'s baby and Stephanie and HHH\'s eventual kid.  First match was John Cena winning the U.S. Title against the Big Show.  Good idea to start with Cena as he got the crowd going.  Then Booker T & RVD retained the World Tag Titles.  Nothing too special there.  Christian then defeated Chris Jericho as Trish turned heel and aligned with Christian.  Poor Jericho has now lost at three straight WMs.  Trish should be fun to watch as a heel though.  I couldn\'t hear JR\'s commentary but I imagine that he screamed \"What a Jezebel!\"  The next match was Evolution defeating Mick Foley and The Rock when Orton pinned Foley.  The crowd was solidly behind Ric Flair during the match.  Right above us a banner was draped across the rafters that said, \"Whoooo!  Thank You Ric Flair.\"  Unfortunately, before the PPV started, MSG ushers took the banner down.  They had an old lady usher do it because if it were anyone else, they might have gotten beat up.  Sometime around here we saw the Hall of Fame inductees.  Great to see Bobby Heenan, Jesse Ventura and Pete Rose.  Good to see Tito Santana, Don Muraco and the kin of Big John Studd and the Junkyard Dog.  Indifferent to see Sgt. Slaughter since he\'s always on TV.  The fifth match had Sable and Torrie beating Stacy and Jackie in an evening gown match, which degenerated into a bra and panties match.  Standard T&A fest.  Then, the cruiserweight open took place.  Chavo Guerrero won it but the highlight was Tajiri inadvertently spraying Akio with his mist.  Akio was then unable to enter the match which led to a fan behind us to yell \"You Screwed Akio.\"  Bartow\'s response, \"No, Akio Screwed Akio.\"  During the match we chanted the names of cruiserweights we would have liked to see included such as Super Calo, Silver King and La Parka.  The next match was perhaps the most bizarre match in WM history.  Goldberg defeated Brock Lesnar.  The problem was that both guys are leaving WWE.  They say that only 10% of the fans read the internet, but every one of those 10% must have been in MSG last night.  Fans chanted \"You sold out\", \"Nah Nah Nah Nah, Hey Hey Hey, Goodbye\", \"This match sucks\", \"Stun them both\", \"Austin Austin\", \"We Want Bret\", \"Hogan\", and lots more.  Neither Goldberg nor Lesnar looked too pleased with the crowd reaction.  The next match was my requisite bathroom break as Rikishi and Scotty retained their Tag Titles.    Upon returning from the bathroom I saw a promo promising \"Edge is coming.\"  Funny, I thought Val Venis was coming.  The next match had Victoria defeating Molly.  Molly then had her head shaved.  She now looks a little bit like the bald chick from the first Star Trek movie.  Then, Eddie Guerrero defended the WWE title against Kurt Angle.  It was a very sound match from two of their best.  The next match featured the return of the Dead Man.  That\'s right, the Undertaker came back with Paul Bearer and Druids.  This was to Bartow\'s chagrin who wailed \"No good can come from druids!\"  Indeed, the druids kept dropping their fire and a lackey had to keep running out with his fire extinguisher.  Oh yeah, Undertaker won.  The final match had Chris Benoit winning the World Title from HHH and HBK in a triple threat match.  This was by far the best match of the night.  The crowd was solidly behind Benoit and we went ballistic when HHH tapped to the crossface.  I almost lost my voice in all the excitement!  To close the show, Eddie Guerrero came down to congratulate Benoit as confetti fell from the ceiling.  Benoit\'s wife (Woman) and children also came down.  Five hours after the show started we made our leave.  Except for Jon and Emily who left after the Undertaker\'s entrance.  Special thanks to the entire nVo crew for treating me to my ticket.  Its a much appreciated birthday present.  If I can ever give you a nice B-D present five months after your actual B-D, I\'ll be glad to.  For another amusing take on last night\'s festivities,  urlLink read this link from the ESPN page.   '),(22096,2824507,'14,March,2004','2004-03-14 00:00:00','    Yesterday was the first day since I started this blog that I didn\'t update it.  Today\'s update will also be brief as I am soon going into Manhattan for WrestleMania XX!!!  A full summary can be found here tomorrow.  Word life.   '),(22097,2824507,'12,March,2004','2004-03-12 00:00:00','    In New York, the Deputy School Chancellor was forced to resign over a nepotism scandal.  In order to ensure that no such scandal happens again, they picked a replacement who is not related to anyone.  WrestleMania XX is this weekend and the band is getting back together.  Looking forward to seeing the entire nVo crew as weâ€™ll be there in MSG.  For a full history of the nVo and our adventures dating back to 1996,  urlLink read this!   My â€œSurvivorâ€ prediction was accurate again as Colby was voted out.  His teammates cited the same reason I did yesterday, that they needed to vote him out before a merge, otherwise heâ€™d win all the individual immunity challenges.   '),(22098,2824507,'11,March,2004','2004-03-11 00:00:00','    No longer will talk radio be dominated by the ultra-conservatives voices of Rush, Hannity and their ilk.  Beginning on March 31, â€œAir Americaâ€ goes on the air!  This is the long awaited liberal talk radio station.  In New York, it will air on WLIB (1190 AM).  Highlights will include Al Frankenâ€™s noon â€“ 3 pm timeslot.  His show will be called â€œThe Oâ€™Franken Factor.â€  Iâ€™m sure Oâ€™Reilly will not be pleased.  Also, comedian Janeane Garafolo gets the 8 pm â€“ 11 pm timeslot with her show â€œThe Majority Report.â€  And Robert Kennedy Jr. will get a weekend show.  Iâ€™ll certainly tune in from time to time.  Now all we need is for them to get on the air in swing state markets!   John McCain made some news yesterday.  When a reporter asked him about the prospect of being Kerryâ€™s running mate, McCain refused to rule it out.  They have gotten along very well in the Senate due to their Vietnam experiences and there is no love lost between Bush and McCain.  But the idea that McCain would even consider switching parties had to be jarring news to the G.O.P.  The good senator from Arizona is highly respected by most Americans, including those who are independents.  A Kerry-McCain ticket would be devastating.  But it wonâ€™t happen.  A Kerry-Tom Brokow ticket... that could be good too!  Itâ€™s time for my weekly â€œSurvivorâ€ predictions.  In Chapera, Boston Rob and Amber have made an alliance with Rupert and Jenna.  They also have an alliance with Big Tom.  Only Alicia is left out.  Rather than betray one of their alliances this early, they should opt to vote Alicia out.  Then Rob and Amber can keep their options open as they head into a merge.  In Mogo Mogo, Lex has emerged as the power player.  He seems to have alliances with everyone and everyone trusts him so far.  Colby is on thin ice though.  Heâ€™s already rubbed Jerri, Shi Ann and Kathy the wrong way.  Ethan is also weary of him since Colby presents tough competition.  Since Colby was a master at winning immunities in Australia, I think theyâ€™ll want to get rid of him before the merge happens.  Therefore, depending on who loses the challenge tonight, I predict that either Alicia or Colby will be voted out.   '),(22099,2824507,'10,March,2004','2004-03-10 00:00:00','    I am annoyed that the BlogSpot banner ad that can be seen at the top of this screen currently features Republican propaganda.  That is inconsistent with the political beliefs of this blog!!!   '),(22100,2824507,'09,March,2004','2004-03-09 00:00:00','    The Game Show Network has renamed itself â€œGSNâ€.  This is in the tradition of The Learning Channel renaming itself â€œTLCâ€ and The National Network renaming itself â€œTNNâ€ (before changing that one to â€œSpike TVâ€)  Donâ€™t you dare call them by their full former names, otherwise an irate entertainment CEO will come after you.  I love acronyms that donâ€™t stand for anything.  Last year I dumped Cablevision and got a satellite dish instead.  Itâ€™s been great.  More programming choices, clearer reception, lower rates.  Now Iâ€™m really glad that I chose DirecTV because their competitor, the DISH Network, has dumped all of the Viacom stations during price negotiations.  That includes CBS.  How insane is that?  This standoff can only hurt DISH Network.  Much more than the standoff between Cablevision and YES.  DirecTV better not pull any stunts like this in the future!    In an effort to make their menu healthier and to defend themselves against lawsuits, McDonaldâ€™s is dropping the option to â€œSuperSizeâ€ your order and they are increasing marketing for their salads.  Because whereas a cheeseburger has 11.5 grams of fat, a Chicken Caesar salad has 18.4 grams of fat.  Way to go, guys!  In celeb death news, the guy who played the painter on â€œMurphy Brownâ€ passed away yesterday.  Didnâ€™t it take him something like 6 years to paint Murphyâ€™s house?  He was ALWAYS there.  Maybe there was more than some paining going on, nudge wink.   '),(22101,2824507,'09,March,2004','2004-03-09 00:00:00','    The latest celebrity who died is Paul Winfield.  Winfield was a celebrated actor with a long career but the role that I most distinctly remember was in â€œStar Trek II : The Wrath of Khan.â€  Winfieldâ€™s character joined Chekhov in their visit to City Alpha V.  After being captured by Khan and his cronies, they had those disgusting looking beetle larvae inserted into their ears.  The larvae were first extracted from this huge insect that Khan kept in an aquarium tank.  They were then put inside Winfield and Chekhov\'s space helmets so they could crawl along their faces and into their ears.  That scene made such an impression on me when I was a kid.  My friends and I would act it out, mimicking Winfield and Chekhovâ€™s horrified expressions as the larvae attacked them.  Ah, great moments in film history.   '),(22102,2824507,'08,March,2004','2004-03-08 00:00:00','    Lots of news this morning.  First of all, the Iraqi Governing Council signed a constitution that will supposedly go into effect when Iraq regains their sovereignty on July 1.  Iâ€™m feeling all patriotic about Iraq now.  I wonder if they have a national anthem.  If they do, Iâ€™d like to sing it.  (Iâ€™m reminded of Nikolai Volkoff, the evil Soviet wrestler who became a good guy when the USSR abandoned communism.  Suddenly his renditions of the Soviet national anthem went from â€œHow dare he sing that!â€ to â€œAw, isnâ€™t that cute?â€ WWE can do the same thing today with a â€œreformedâ€ Sgt. Slaughter, Col. Mustafa and Gen. Adnan.)  In other wrestling news, Hercules Hernandez (real name, Ray Fernandez) died over the weekend.  Another ex-wrestler added to the dead pool.   John Henry Williams also died from leukemia over the weekend at the young age of 35.   This is Ted Williamsâ€™ son.  The same guy who had his father frozen in a cryogenics laboratory.  Maybe now that heâ€™s out of the picture, Ted Williams can be given a proper burial.  In better news, someone who we thought was dead, isnâ€™t!  That girl who disappeared on her way from Poughkeepsie to Brooklyn resurfaced over the weekend where she was chilling in the Hamptons.  Hmm.  Well, it is spring break, right?  Newly acquired Mets, Karim Garcia and Shane Spencer are in trouble.  They got drunk at a pizzeria a few days ago and Garcia took a piss on the sidewalk outside of the restaurant.  When a delivery boy chased after them, they confronted him, causing him to fall over some bushes.  Now the delivery boy is threatening to press charges.  Letâ€™s Go Mets.  â€œThe Sopranosâ€ came back last night with a bang!  A highlight was a waiter getting whacked when he demanded a bigger tip.   '),(22103,2824507,'07,March,2004','2004-03-07 00:00:00','    Just got back from a successful shopping expedition.  Among my purchases were a new watch (have been without one since my old watch broke on New Years Eve), a new keychain (my Vassar bev-key held up for five years but no more), the \"Monday Night Wars\" DVD, and Caren Lissner\'s new book \"Starting from Square Two.\"  Caren is a great new writing talent who I had the pleasure of meeting a few months ago.  I\'m glad to give her a plug, so  urlLink buy her book!   \"The Sopranos\" starts its new season tonight.  Can\'t wait!   '),(22104,2824507,'06,March,2004','2004-03-06 00:00:00','    Martha Stewart - guilty!  Or as she would say, it\'s a bad thing.  Well at least she\'ll have the best decorated cell on the block.    Mayor Bloomberg reportedly told a roomful of gays and lesbians that he personally supports gay marriage.  Good for him.  But then his press secretary started backtracking.  It\'s a shame that the mayor won\'t go on the record with his opinion.  Instead of genuinely helping the movement progress, he comes off as pandering.  Saw two good movies recently on DVD.  \"Pieces of April\" and \"School of Rock\".  Highly recommended.  \"Pieces of April\" actually got to me a little bit.   '),(22105,2824507,'05,March,2004','2004-03-05 00:00:00','    I was wondering how long it would take after Bush started airing 9-11 images in his campaign ads before there would be a backlash.  The answer turned out to be...not long.  As soon as the ads hit the airwaves, 9-11 families charged him with politicizing the tragedy and a big protest is planned for this afternoon at Ground Zero.  All of a sudden, what was supposed to be a positive ad to kick off his campaign has caused the Bush team to go on the defensive.  The thing is, without 9-11, he doesnâ€™t have too many other cards to play.  Bush will need to invoke it all he can in order to avoid talking about the economy or Iraq.  Speaking of the campaign, Iâ€™m very depressed this morning with the national poll results.  Bush and Kerry are in a dead heat with Nader pulling 6%.  Thatâ€™s 6% that could have given Kerry a decent sized lead.  Whatâ€™s wrong with these people who would still consider voting for Nader, especially knowing what happened last time?  Itâ€™s too depressing to think about.  Turns out I was right (again) with my â€œSurvivorâ€ prediction.  Susan Hawk is gone, having had a nervous breakdown and quitting the game.  Sheâ€™s got some emotional issues to work out.   '),(22106,2824507,'04,March,2004','2004-03-04 00:00:00','    Big announcement from NASA the other day.  The Mars Rover found a rock.  No big surprise there, if thereâ€™s one thing Mars has a lot of, itâ€™s rocks.  But this is no ordinary rock, ladies and gentlemen.  This rock has evidence on its surface that water once existed atop it.  There is water on Mars!  Or should I say, there was water on Mars.  As Jon Stewart points out, this leads to the inevitable conclusion that Mars once had waterparks.  Next thing we know, the Mars Rover will send back photos of a Slip â€˜N Slide.  As we get excited about amusement parks on Mars, thereâ€™s lots of tumult going on in the Disney empire.  Michael Eisner was stripped of his Chairman position by the stockholders and he is being replaced by George Mitchell.  Yes, George Mitchell, the former Senator and broker of the Irish Peace Accords.  Is there anything the man doesnâ€™t do?    Funny but true story from the Daily News...  THREE SENIORS HURT IN SALAD-BAR SLUGFEST  Thereâ€™s nothing like a good old fashioned food fight.  Just ask any kid in junior high school and the old folks at Spring Haven Retirement Community. A brawl over the salad bar send three seniors to the hospital during the weekend. Words were exchanged when a 62-year old man spent a little too much time picking through the greens.  He â€œonly likes a certain kind of lettuce,â€ said his 80-something mom, who also lives at the home. When the other folks at the home berated the man, he punched an 86-year old in the nose and bit a 78-year old man on the right arm, the News Chief newspaper reported yesterday. The lettuce picker has now been evicted.  â€œThis is the first time in 25 years Iâ€™ve ever heard of something like this happening,â€ said Jill Andrew, spokeswoman for the facility.  Susan Hawk is going to be the next one eliminated from â€œSurvivorâ€.  Take that to the bank.   '),(22107,2824507,'03,March,2004','2004-03-03 00:00:00','    Congratulations to John Kerry â€“ the Democratic Presidential Nominee.  With wins in 9 of the 10 Super Tuesday states, Kerry is the last standing major candidate as John Edwards is expected to drop out today.  The only primary Kerry didnâ€™t win was Vermont as, despite dropping out, Howard Dean picked up his home state.  At least the good doctor will always have Vermont.  President Bush called Kerry last night to offer his congratulations, to which Kerry responded, â€œGo fuck yourself.â€  Or maybe he really said, â€œI look forward to debating the issues of our nationâ€™s future.â€  Either way itâ€™s going to be a long and nasty campaign.  Chris Matthews remarked, â€œ8 months?!? You can have a baby in nearly 8 months.â€    Aristide is bitching from his exile that he never left Haiti voluntarily and that he was the victim of a U.S. coup.  Maybe so.  Itâ€™s hard to condone removing a nationâ€™s elected leader but the bottom line is that we saved the guyâ€™s life and prevented a hell of a lot of violence that would have broken out if the rebels stormed the capital.    Barry Bonds, Gary Sheffield, Jason Giambi...  steroid junkies!  This is bad for baseball, very bad.  Itâ€™s been obvious from looking at the size of these ballplayers that somethingâ€™s been up but now the shit is hitting the fan.  Giambi is now a fraction of the size that he used to be and looked downright sheepish when reporters questioned him yesterday.  And Sheffieldâ€™s rep released a statement that he never â€œknowinglyâ€ took steroids.  Didnâ€™t take them long to break out the legalese.    This use of illegal steroids makes Pete Roseâ€™s gambling seem insignificant by comparison.  Yet he is the one who is barred from the MLB Hall of Fame.  Enter Vince McMahon!  Rose is being inducted into the WWE Hall of Fame at WrestleMania.  From 1998-2000, Rose made cameo appearances by getting into scuffles with Kane.  So on March 14, Kane will present Rose with his plaque.  Jesse Ventura will be on hand too.  Canâ€™t wait to see the wacky shenanigans that will ensue!   '),(22108,2824507,'02,March,2004','2004-03-02 00:00:00','    Yesterday at work, an Orthodox Jewish employee entered the office of an Irish employee.  After a brief meeting, the Orthodox Jewish employee left and accidentally left behind a plastic bag.  A half hour later, the Irish employee first noticed the bag and said, â€œOoh.  Whatâ€™s this?  Did someone leave me a present?  What could it be?â€  He then opened the bag and examined the object.  â€œHmm.  Not quite sure what this is.  It looks like an ammunition belt.â€  He then walked out of this office to show his staff his new ammunition belt.  After a long laughing fit, we informed him that he was holding tiffilin, a ritual prayer accessory used by Orthodox Jews.  The tiffilin was soon returned to its rightful owner.  Itâ€™s Super Tuesday, which means that after months of reading and watching news coverage about other states primaries, I get to vote in one myself!  After work, Iâ€™m heading to P.S. 206 â€“ the same building where I attended Pre-Kindergarten through the 5th grade, and am going to cast my vote.  If Kerry sweeps, as is expected, Edwards will likely drop out tonight.  So today is probably the final day of the primary campaign.  Sniff sniff, what will we talk about for the next few months?   '),(22109,2824507,'01,March,2004','2004-03-01 00:00:00','    I got 17 out of my 21 Oscar Predictions correct.  Not too bad.  The only categories that I missed were Actor, Adapted Screenplay, Sound Effects and Makeup.  Although I thought Sean Penn was fantastic in â€œMystic Riverâ€, I wasnâ€™t sure if the Academy would honor him because of his controversial reputation.  Fortunately I was wrong as Penn is a very deserving winner.  I was also surprised that â€œLord of the Ringsâ€ won every single award is was nominated for, tying the record set by â€œBen-Hurâ€ and â€œTitanicâ€.  Otherwise there werenâ€™t too many surprises on Oscar night.  This morning I have the song from â€œA Mighty Windâ€ stuck in my head and it wonâ€™t leave.  Damn you and your comic genius, Eugene Levy!  I cannot get over how warm the weather is outside.  No more heavy winter coat!  The deep freeze of winter has finally ended...although there are still patches of snow that can be found here and there.   '),(22110,2824507,'29,April,2004','2004-04-29 00:00:00','    Bush and Cheney met with the 9-11 panel.  Alas, it was in private with no TV coverage and no written transcripts.  I would imagine that Bush sat on Cheneyâ€™s knee as the VP did his best impression of a ventriloquist.  Maybe he even made a talking Bush voice as he drank a glass of water!  We had another Parks event today.  It was a groundbreaking at Harry Maze Playground in East Flatbush.  The fun part was when our sound system went on fire a half-hour before the event started.  It wasnâ€™t as bad as it sounds, but lots of smoke did emanate from the speakers.  So my boss ran to Radio Shack and bought a Karaoke machine which served as our sound system for about a half hour until it went dead in the middle of some guyâ€™s speech.  (Kinda appropriate though since the guy was pontificating for way too long)  So my boss returned the Karaoke machine as soon as the event was over and is presumably shopping for something else at this moment as we have another event tomorrow for Arbor Day!  Yipee.   Last week I predicted Shii Annâ€™s departure on â€œSurvivorâ€ and she certainly would have been voted out except she won immunity.  Barring another miracle, she is certain to go tonight.  Then we will be left with Boston Rob, Amber, Rupert, Jenna and Big Tom!   '),(22111,2824507,'28,April,2004','2004-04-28 00:00:00','    Saw â€œWickedâ€ last night.  It was great!  (Warning â€“ Spoilers Below!)  The musical â€œWickedâ€ is quite different from the book â€œWickedâ€.  The first act has some slight differences that work well for the stage medium in the sense that it moves the story along more fluidly.  For example, it skips over Elphabaâ€™s upbringing and jumps right into her entry to Shiz University.  It also has her sister Nessarose beginning school at the same time as Elphaba and Glinda.  This way we meet all the main characters at once.  The show then focuses on the relationship between Elphaba and Glinda with many duets between the two.  Whereas the book had heavy political and metaphorical content, the musical is more driven by the relationships.  The many atrocities of the Wizard in the book, are boiled down in the musical to his suppression of animals.  In the second act though, the musical veers wildly off the novel and becomes a story unto itself.  But itâ€™s a damn good story!  We see the origins of the Cowardly Lion (alluded to in the book), the Tin Man (itâ€™s Boq!), and the Scarecrow (itâ€™s Fiyero!).  In this regard it does a great job of connecting Gregory Maguireâ€™s world with L. Frank Baumâ€™s world with some surprising twists along the way.  Finally, Elphabaâ€™s fate is dramatically satisfying.  Whereas her affliction to water is all too real in the book, she plays it up to her advantage in the musical, leading to quite an ending.  I just donâ€™t know if I can shake the image of the Wicked Witch and the Scarecrow together.    The performers in \"Wicked\" were on top of their game.  Idina Menzel and Kristin Chenoweth are remarkably talented actresses with some amazing pipes!  The main reason that I wanted to see \"Wicked\" was to see both of them perform in the lead roles.  They both deserve Best Actress Tony Nominations.  Chenoweth might win because she is the bigger name, but Menzel was the heart of the show.  And how about Joel Gray?  72-years old and he\'s still doing his thing.  Glad to see his in such great shape.  Finally, the Playbill said that the actress who played Madame Morrible used to appear in \"The Cosby Show\" (one of my favorite late-night cable rerun options).  After wondering who she played, I remembered - Rudy\'s teacher!  The one who used to give Kenny a hard time.  Between her and Phyilicia Rashad appearing in \"A Raisin in the Sun\", it\'s a Cosby takeover of Broadway!  Or something like that...   '),(22112,2824507,'27,April,2004','2004-04-27 00:00:00','    So much work, so little time.  This week we have a groundbreaking ceremony on Thursday and then two Arbor Day events on Friday.  Plus thereâ€™s the Coney Island events poster Iâ€™m designing and mooring permits that need to go out.  It can be overwhelming so Iâ€™m taking it slow and focusing on one thing at a time.  Meanwhile one of my less-experienced co-workers is on the verge of a breakdown.  Ah, rookies.  Tonight I am seeing â€œWickedâ€!  Canâ€™t wait.  It is a Broadway musical that is a prequel to â€œThe Wizard of Oz.â€  I read the Gregory Maguire book that it was based on in November and look forward to comparing this adaptation to the original work.  A review of the show will be forthcoming.   '),(22113,2824507,'26,April,2004','2004-04-26 00:00:00','    Canâ€™t blog.  Working.   '),(22114,2824507,'25,April,2004','2004-04-25 00:00:00','    Taking full advantage of my Netflix subscription means ordering every single feature film release under the sun.  One case in point was \"Cheaper by the Dozen.\"  It made over $100 million at the box office so how could I not check it out?  I\'d give you a plot synopsis except that there\'s not much of one.  Steve Martin and Bonnie Hunt want a big family so they have 12 kids.  Yes, 12.  They also want careers which involve moving the family to a different town.  The kids (brats that they are) aren\'t too happy with this and spend the better part of 90 minutes acting out.  I suppose it was supposed to be funny.  It\'s as if the producers said, let\'s put Steve Martin with a bunch of kids and see the wackiness that will ensue.  Wackiness indeed.  My next Netflix film to watch is \"The Matrix Revolutions.\"  I actually started it last night.  I watched a half-hour before my eyelids became very heavy and I turned it off to go to sleep.  I\'ll probably finish the movie at some point today.  Congratulations to Vitali Klitschko, boxing\'s NEW heavyweight champion of the world.  Last night he won the title vacated by Lennox Lewis.  In his corner was his younger brother, Wladimir, who has been having his own problems in the ring.  If this was wrestling, Wladimir would turn heel on Vitali and demand a title shot, ala the Bret Hart vs Owen Hart wars of 1994.   '),(22115,2824507,'24,April,2004','2004-04-24 00:00:00','    I\'ve noticed that when Saturdays come around, I never have too much to write.  Maybe this is my blog\'s way of observing the Sabbath.   '),(22116,2824507,'23,April,2004','2004-04-23 00:00:00','    We always get a nice turnout from the local press at our Parks events.  The Daily News ran a story about the Wyckoff House event that I described yesterday.   urlLink You can read the article here.   Just got back from Coney Island where we met to discuss the annual beach opening ceremony.  While there, I took a walk along the boardwalk and admired the four new performance gazebos that were installed.  These pavilions are named after the amusement parks of yesterday (Dreamland & Steeplechase) and the amusement parks of today (Astroland and Denoâ€™s Wonder Wheel Park).  When we have the ceremony in late May, the big photo-op will be the unveiling of one of the gazebos, as a band plays inside it.  In the meantime one of the big projects that Iâ€™m working on is designing a Coney Island Events Poster that will be displayed on our 9 boardwalk kiosks.  Jon Stewart was hilarious on The Daily Show the other night.  He was making fun of our indecision over who we would be handing Iraqi sovereignty over to on June 30.  At his press conference Bush had told a reporter â€œWeâ€™ll get back to you on that.â€  Stewartâ€™s reply : â€œWait!  Donâ€™t tell me!  Iâ€™ll wait up bright and early on June 30th, tip toe down the stairs in my footsie pajamas, and see who will be the new leader of Iraq sitting under the tree.  Wait, what if itâ€™s me?  Just think, I could be the leader of Iraq!  Oh my god, that would be the best June 30th ever!â€  On â€œSurvivorâ€, Shii Ann was marked for elimination last night and the only way she could be saved was if she had won immunity.  So, she did!  Without the option of voting out Shii Ann, the original Chapera members turned on their own and ejected Alicia who they had earlier indicated was the â€œleast deserving of being an All-Star.â€   '),(22117,2824507,'22,April,2004','2004-04-22 00:00:00','    This morning we had an event at the Wyckoff House.  Not only is it the oldest house in New York City, itâ€™s the oldest house in New York State.  It was built by Pieter Claesen Wyckoff, a Dutch settler, in 1652 on land that was recently taken from the local Canarsee Indian tribe.  Wyckoffâ€™s descendants continued to farm the land until 1901.  In the 1960s it was declared a National Historic Landmark.  The house exists today as a museum that is owned by the Parks Department.  Todayâ€™s event was to herald some new improvements that are coming to the grounds.  The weather was great, the speeches were short, and the school kids in attendance were very knowledgeable about the siteâ€™s history.    Lauren Sullivan, one of my favorite co-workers of all time, told me today that she is leaving Parks.  Her husband is a musician in the band Guster and she is forming a non-profit organization that works with musicians to promote environmental issues.  Iâ€™m happy for her but very sad to see her go.  Fortunately she will still be living in Brooklyn and has promised to visit and do lunch.       My â€œSurvivorâ€ prediction is Shii Ann, last of the Mogo Mogos.   '),(22118,2824507,'21,April,2004','2004-04-21 00:00:00','    Iâ€™m getting a raise!  Yay!  OK, itâ€™s not a big raise, but itâ€™s something.  Last night, the Mayor made a deal with the Union for 5% increases.  This comes as a slight surprise as there had been no indications that a deal was forthcoming.  The Mayor had previously taken a hard stance that there would be no raises without productivity improvements, and no retroactive raises.  He backed away on both points.  My guess is that he wants peace with the labor unions during his upcoming reelection campaign.  His poll numbers have been improving and this  might  help.  (Yes, there are negative parts of the deal.  Itâ€™s less than the 8% increase given in 2001 and it reduces starting salaries for new hires)  But all in all, something is better than nothing and up until this point I had been expecting nothing.   '),(22119,2824507,'20,April,2004','2004-04-20 00:00:00','    Today is the anniversary of some sad events.  Itâ€™s 13 years ago today that my Grandpa Murray passed away and itâ€™s 5 years ago today that the Columbine shootings happened.  To top that off, itâ€™s Hitlerâ€™s birthday!  Letâ€™s hope that this 4-20 is more uneventful.  Great weather this week.  Spring has finally arrived and the parks are bustling with people.  Phones are ringing off the hook, special events are being planned.  It should be crazy like this until... oh Iâ€™d say October.  Todayâ€™s Daily News features The Ten Worst Songs of All Time, as decided by Blender Magazine.  They are...   1)  â€œWe Built this Cityâ€   (Starship)   2)  â€œAchy Breaky Heartâ€   (Billy Ray Cyrus)   3)  â€œEverybody Have Fun Tonightâ€   (Wang Chung)   4)  â€œRollinâ€   (Limp Bizkit)   5)  â€œIce Ice Babyâ€   (Vanilla Ice)   6)  â€œThe Heart of Rock & Rollâ€   (Huey Lewis and the News)   7)  â€œDonâ€™t Worry Be Happyâ€   (Bobby McFerrin)   8)  â€œParty All the Timeâ€   (Eddie Murphy)   9)  â€œAmerican Lifeâ€   (Madonna) 10)  â€œEbony and Ivoryâ€   (Paul McCartney & Stevie Wonder)   '),(22120,2824507,'19,April,2004','2004-04-19 00:00:00','    Good luck tomorrow to all of those third-graders taking the exam that will single-handedly determine if they are going to become fourth-graders.  I think it\'s utterly insane that the promotion of a child (a CHILD!) will boil down to their performance on ONE standardized test.  Yes, standards need to be established but c\'mon!  So I\'ll be rooting for you little kiddies.   '),(22121,2824507,'18,April,2004','2004-04-18 00:00:00','    My new favorite word is â€œriff raffâ€.  As in describing those less than desirable elements.  Such as... â€œThe bed and breakfast in Park Slope does not have any identifiable signs in order to keep away the riff raff.â€    There was this guy on the subway this morning who looked a little bit like my old political science professor, Mr. Born.  Especially the face and glasses.  Except where Mr. Born had short, thinning hair, this guy had long straggly hair that went down to his shoulders.  So he sort of looked like Mr. Born had he gone to Woodstock.  Big â€œSimpsonsâ€ night last night.  Skinner and Krabbapel almost got married but she instead wound up having an affair with the Comic Book Guy.  Then they reran the classic episode spoofing â€œEvita.â€  Then on â€œ24â€, Ryan Chappelle got whacked.  Meanwhile I taped â€œThe Sopranosâ€ and watched it afterwards.  Turns out that the actor who played Chappelle is also on â€œSopranosâ€, as the conflicted priest, Father Phil.  Now that his â€œ24â€ character is dead, heâ€™ll have plenty of extra time to be on â€œSopranos.â€    So much work to do this week.  The busy season is upon us!   '),(22122,2824507,'18,April,2004','2004-04-18 00:00:00','    I just saw \"House of Sand and Fog\".  What a great movie!  Battles for real estate have never been more dramatic.  Jennifer Connelly plays a woman who was wrongly evicted from her house, as the county claimed she failed to pay a tax that she didn\'t even need to pay.  But Connelly could have prevented it from happening had she bothered to open her mail!  Meanwhile, Ben Kingsley, a former Iranian colonel buys the house at auction and refuses to leave despite learning of Connelly\'s predicament.  I loved the depiction of Kingsley, his wife and his son.  Despite his financial stubbornness, it\'s clear that Kingsley only wants whats best for his family.  That makes the devastating events in the last half-hour of the film especially heartbreaking.  Go see this movie now!   '),(22123,2824507,'16,April,2004','2004-04-16 00:00:00','    Just recently got back from a night out in Manhattan.  Don\'t feel like sleeping so I\'ll blog instead!  I\'ll add this to my previous post about chemistry.  There\'s nothing you can do about it.  It\'s either there or it\'s not there.  The best approach is not to take it too personally if it isn\'t happening.  For example, I\'m confident that I\'m a decent looking guy who is a good conversationalist with a witty sense of humor.  I can only be myself and, if that\'s not enough, then it\'s just not meant to be.  Can\'t think of anything else to say on the matter seeing as how it is after 2 in the morning!  Will blog more tomorrow!   '),(22124,2824507,'16,April,2004','2004-04-16 00:00:00','    Chemistry.  Iâ€™ve always hated it in school.  Definitely my worst subject.  So once I fulfilled my requirements, I did not take any additional courses in it.  Now, years later, chemistry has again come back to haunt me in the dating world.  Thereâ€™s nothing more frustrating than going out with someone, thinking it went well, and then being told that they donâ€™t want to pursue anything because they â€œdidnâ€™t feel any chemistry.â€  Grr.  Thatâ€™s one of the perils of online dating sites like JDate.  If two people meet in the real world, then thereâ€™s a better chance of establishing mutual sparks before setting up a date.  But even if two people establish a good rapport online and on the phone, thereâ€™s no guarantees until they first meet in person.  Although it can be frustrating, Iâ€™m doing my best to stay optimistic and to keep trying!   I guess it pays to watch a TV program before offering a prediction on it.  â€œSurvivorâ€ addict that I am, I correctly forecasted Kathyâ€™s departure.  But having never watched â€œThe Apprenticeâ€, I chose the wrong guy.  Turns out that Bill won, not Kwame.  My bad.   '),(22125,2824507,'15,April,2004','2004-04-15 00:00:00','    Update :  Timber!!!  A huge 80-year old tree just fell in front of my office building in Prospect Park and caused quite a noise.  A combination of its age, health, wind conditions and the fact that its roots failed to sustain it.  It\'s a good thing noone was standing underneath it when it fell!  I also wanted to point out that there is a striking resemblance between Ariel Sharon and Tommy Lasorda.   '),(22126,2824507,'14,April,2004','2004-04-14 00:00:00','    Thereâ€™s one more Bush quote from Tuesday nightâ€™s press conference that I forgot to mention but is too irresistible not to :  â€œI wouldnâ€™t be happy if I were occupied either.â€  Yes, Mr. President.  We sure wouldnâ€™t want you to be occupied.  Thatâ€™s why Cheney does all your work for you.  The rains continue.  April showers ain\'t no cliche.  \"The O.C.\" is such an addictive program.  It\'s so soapy but I can\'t turn away.    My \"Survivor\", who gets the boot prediction is Kathy.  With Lex gone she should be the next target for the united Chaperas.  That\'s assuming that the Chaperas remain united.  I think they will for at least one more vote.  I\'ve never watched \"The Apprentice\" but, after reading a newspaper article about it, I\'m predicting that Trump will hire the black guy.   '),(22127,2824507,'14,April,2004','2004-04-14 00:00:00','    More fun excerpts from Bushâ€™s press conference last night :  * A reporter asked, â€œIn the last campaign, you were asked a question about the biggest mistake you\'d made in your life and you used to like to joke that it was trading Sammy Sosa. You\'ve looked back before 9/11 for what mistakes might have been made. After 9/11, what would your biggest mistake be?â€  To which, Bush then looked like a deer caught in headlights as he stuttered â€œHmmm. I wish you\'d have given me this written question ahead of time so I could plan for it... You know, I just â€” I\'m sure something will pop into my head here in the midst of this press conference with all the pressure of trying to come up with an answer, but it hadn\'t yet... You know, I hope I don\'t want to sound like I\'ve made no mistakes. I\'m confident I have. I just haven\'t â€” you just put me under the spot here and maybe I\'m not quick, as quick on my feet as I should be in coming up with one.â€  They then cut to a shot of Karl Rove, Condi Rice and Andrew Card who looked mortified.  Had they not prepared him for such an expected question?  Or had they prepared him and Bush had forgotten what he was supposed to say?  * When asked who we will be transferring sovereignty to in Iraq on June 30, Bush responded â€œWeâ€™ll get back to you on that.â€    * I loved when a reporter asked him â€œWill (Iraq) have been worth it, even if you lose your job because of it?â€ and then when another reporter asked â€œIf you feel in any way that youâ€™ve failed as a communicator?â€    Now we know why Bush does not have too many press conferences.    In other news, congratulations to Barry Bonds for hitting his 661st home run last night and surpassing Willie Mays for 3rd on the all-time home run list.  If anyone is interested in the 19th century plagues that wrecked havoc in Europe, read  urlLink Sarah\'s blog .  Her recent bout with bronchitis has inspired a litany of commentary on the subjects of consumption, tuberculosis, ague, and many more!  And if anyone is interested in commentaries on todayâ€™s dating scene, check out the webpage of  urlLink The Anonymous Blogger .  Be sure to visit the always happening comments section.   '),(22128,2824507,'13,April,2004','2004-04-13 00:00:00','    Who is this boob currently appearing on all the TV channels?  Oh yeah, Bush. He\'s currently in the middle of the press conference and at least the press is asking him some hardball questions this time around.  Their softball questions at his previous press conference (before the Iraq war) was inexcuseable.  When the reporter asked him how his faith served to inspire him I wanted to throw my shoe at the TV.  But that was last time.  This time, they\'re challenging him more over his mistakes and Bush is stammering through his talking points.  One noteable exchange was when he was asked why he will only appear before the 9-11 commission sitting alongside Cheney.  Bush answered that he looks forward to answering the commission\'s questions.  But when pressed why he wouldn\'t appear by himself, as the commission requested, he only repeated that he looked forward to meeting with them and moved on to the next question.  What a dodge!  Even the dumbest viewers had to pick up on that one.  More commentary on Bush\'s performance tomorrow when it\'s all said and done.   '),(22129,2824507,'12,April,2004','2004-04-12 00:00:00','    Skies are overcast.  A mist is falling.  I walk through the streets wearing my black trenchcoat, looking like Morpheus and vowing to overthrow the rule of the machines.  How much do I despise George W. Bush?  Never mind that he stole the 2000 election.  Never mind that he dragged us into the quagmire of Iraq with no exit strategy.  Never mind that he blew the Clinton surplus by giving tax breaks to the rich.  I despise him because his press conference tonight is going to preempt â€œ24â€!!!  The guyâ€™s got some nerve.   Improvement in fantasy baseball.  I shot up four places in the standings and am now in third place!  Iâ€™ll refrain from giving a daily commentary on this subject but will certainly mention if I reclaim my first place position.  Premiere Magazine published a list of The Top 100 Movie Characters.  The Top 10 are...    1)  Vito Corleone   â€œThe Godfatherâ€   2)  Fred C. Dobbs   â€œTreasure of the Sierra Madreâ€   3)  Scarlett Oâ€™Hara   â€œGone With the Windâ€   4)  Norman Bates   â€œPsychoâ€   5)  James Bond   â€œDr. Noâ€   6)  Annie Hall   â€œAnnie Hallâ€   7)  Indiana Jones   â€œRaiders of the Lost Arkâ€   8)  Ellen Ripley   â€œAlienâ€   9)  Jeff Spicoli   â€œFast Times at Ridgemont Highâ€ 10)  Gollum   â€œThe Lord of the Ringsâ€  Other noteables...  12)  Charles Foster Kane   â€œCitizen Kaneâ€ 13)  Atticus Finch   â€œTo Kill a Mockingbirdâ€ 14)  Randle McMurphy   â€œOne Flew Over the Cuckooâ€™s Nestâ€ 22)  Travis Bickle   â€œTaxi Driverâ€ 27)  Marge Gunderson   â€œFargoâ€ 31)  Norma Desmond   â€œSunset Boulevardâ€ 38)  Willy Wonka   â€œWilly Wonka and the Chocolate Factoryâ€ 47)  Mrs. Robinson   â€œThe Graduateâ€ 59)  Lt. Kilgore   â€œApocalypse Nowâ€ 63)  Carrie White   â€œCarrieâ€ 75)  Dr. Strangelove   â€œDr. Strangeloveâ€ 84)  Darth Vader   â€œStar Warsâ€  (should be higher!!!)   '),(22130,2824507,'12,April,2004','2004-04-12 00:00:00','    What is going on?  The baseball season is one week old and my fantasy team has fallen from 1st to 3rd to 5th and now to 7th place!!!  At this pace I\'ll be in last place by Thursday.  Grr.  Even the real-life Mets are doing better than me.  Congrats to Phil Mickelson.  He\'s the golfer who won the Masters yesterday.  I like him because he has the same name as me.  And also because he seems like a genuinely nice guy who until yesterday had been the best player to never have won a major tournament.  Yay, Phil!  Some guy in Las Vegas bet his life savings on the spin of a wheel!  He bet over $100,000 on red.  And the winner was...red.  So now he has over $200,000.  Not a recommended method of investment though.  I think I\'ll stick with my 457 plan and Roth IRAs.   '),(22131,2824507,'11,April,2004','2004-04-11 00:00:00','    Happy Easter.  Not a whole lot going on today.  Took my parents out to dinner last night to celebrate my dad\'s birthday.  Then I helped him set up his new DVD player which was a gift from my brother.  It was very frustrating to hear incredible distortions in the sound.  After wracking my brain over what was wrong, I realized that I had plugged the red cord into the wrong red outlet on the back of the player.  You see there were two reds.  If they would only come up with a variety of different colors we wouldn\'t have such problems.  All is now well on the technology front.  I also watched two Netflix movies this weekend.  The first was \"Something\'s Gotta Give.\"  This was a fun movie, although a half hour too long.  You can\'t go wrong with Jack Nicholson and Diane Keaton.  The second movie I watched was \"Mona Lisa Smile\".  I can understand why it got bad reviews since much of it was cliches, but I liked it anyway.  It took place at Wellesley College in 1953 and as a Vassar alum, I\'d imagine that Vassar was very similar during that time.    I read an article in the newspaper about Anne Hathaway, star of \"The Princess Diaries\" and \"Ella Enchanted.\"  It turns out that she is currently enrolled at Vassar but her movie career has seriously curtailed her advancement towards her degree.  She was originally enrolled in the Class of 2004 and, whereas her friends and classmates are currently getting ready to graduate, she has only completed three semesters.  I really wonder whether or not she will make the effort to continue.  Fantasy Baseball woes.  After two happy days in first place, my team has slipped to 5th.  It\'s all because of my pitching.  10 of my starters have made 14 starts thus far and I only have ONE win from them.  I\'m counting on them all to rebound with better performances next time.   '),(22132,2824507,'10,April,2004','2004-04-10 00:00:00','    One year after toppling the statue of Saddam Hussein, Iraq looks messier than ever.  Now a hostage situation has developed.  Three Japanese civilians are among the hostages and their captors are threatening to burn them alive unless Japan withdraws its troops.  Given all of this chaos, I don\'t see how we can implement a transfer of power by June 30.  The situation seems to be getting far worse before it gets better.  Sad news at the New York Aquarium.   urlLink Their 35-year old beluga whale, Kathy, passed away yesterday.   If you click on the link to its article, you\'ll find a mention of Aquarium spokesperson, Fran Hackett.  I work closely with Fran every year when planning events at Coney Island so, although its a sad occassion, its great to see her name in the national news media.   '),(22133,2824507,'09,April,2004','2004-04-09 00:00:00','    Happy Birthday to my dad!!!  The streets of Brooklyn are utterly deserted today.  Must be a combination of no school and the religious holiday.  Lex got served last night on â€œSurvivorâ€.  Iâ€™ve gotta say that Lex deserved it, having played the game very poorly.  He engineered the demise of those most loyal to him (Colby, Ethan, Jerri) and gambled everything on the assumption that Boston Rob would include him in a Final 4 alliance.  Wrong!  Rob cut him loose, preferring to stick to an alliance with his original tribe.  Lex never should have trusted Rob and now he paid for it.  Did you hear that 19,000 Hasidic Jews rented out Madison Square Garden yesterday for a private showing of the Ringling Bros. Circus?  All female performers were excluded, the men and women sat separately in the audience, and kosher cotton candy machines and other kosher food concessions were brought into MSG.  Must have been quite a sight.    I wish a Good Friday to all the goyim.   '),(22134,2824507,'08,April,2004','2004-04-08 00:00:00','    Update : More thoughts on Condi Rice\'s testimony.  After denying any prior knowledge of Al Qaeda attacking within the U.S., the commissioners revealed that President Bush was presented with a memo on August 6, 2001 entitled \"Bin Laden determined to attack inside United States.\"  Wow.  What was the response by our national security advisor?  That it was a historical document, detailing PAST efforts.  Um, then shouldn\'t it have been titled \"Bin Laden USED TO BE determined to attack...\"?  Now all 10 commissioners are urging that this document be declassified.  Dr. Rice also was questioned on why they didn\'t follow through on plans to attack Al Qaeda that were drafted by the Clinton administration.  Her response?  That there were no such plans, merely several actionable items listed together.  Um, isn\'t that a \"plan\"?   '),(22135,2824507,'07,April,2004','2004-04-07 00:00:00','    Update : Condi Rice has finished testifying before the 9-11 commission.  As expected she refused to acknowledge that the administration made any mistakes or did anything wrong.  I particularly enjoyed the exchange between her and commission member Bob Kerrey.  Rice kept insisting that, prior to 9-11, Bush said he was tired of â€œswatting fliesâ€ with Al Qaeda and wanted a greater response.  Kerrey asked her â€œWhat flies did the administration swat?  Give me one example of when you swatted a fly!â€  That was not terribly surprising since Kerrey is a Democrat.  But then James Thompson, the most partisan Republican on the panel, chided Rice for the administrationâ€™s failure to attack Al Qaeda after the bombing of the U.S.S. Cole.  She said that a â€œtit for tat responseâ€ was not appropriate.  He then asked her how many of our destroyers Al Qaeda would have had to attack before we would have deemed it appropriate.    All in all, I have to give the 9-11 commission credit.  It seems they are genuinely interested in determining what went wrong with our intelligence and the failure of our leaders to act upon known threats.  (That includes Clinton.  Even though Iâ€™m an admitted partisan Democrat, his administration must also be held accountable)  It will be very interesting to read the final content of their report.   '),(22136,2824507,'07,April,2004','2004-04-07 00:00:00','    So Iâ€™m currently listening to Condi Riceâ€™s testimony on my walkman.  Sheâ€™s reading her carefully constructed opening statement.  It will be interesting to see how she responds to intense questioning by the panel and whether she relies on ambiguous statements like â€œIn my opinionâ€ or â€œTo the best of my recollection.â€  It will also be interesting to see how long it takes before she tries to decredit Richard Clarke.  â€œSurvivorâ€ is down to nine people, all of which will at least be on the jury.  CBS said that the merge will finally happen tonight and Boston Rob is in a position where he has made alliances with all remaining members.  The previews indicate that he is going to have to betray one of his alliances.  If the others had any sense, they would get wise to Robâ€™s antics and vote him off.  But they havenâ€™t shown much sense thus far.  Despite Steve Traschelâ€™s awful performance last night, Iâ€™m still in first place in Fantasy Baseball!  My hitters have a combined .380 batting average so far.  That makes up for my pitchers 5.something ERA.   '),(22137,2824507,'07,April,2004','2004-04-07 00:00:00','    Passover Seder, Round Two was held last night in New Jersey.  Special thanks to my cousin Pamela for leading the seder.  It was really great to see her, her husband Mike and their two children, 2 and a half year old Grace, and 1 and a half year old Matthew.  Since they live in San Francisco we donâ€™t get to see them as often as weâ€™d like.  Those kids are so cute and intelligent!  It was also fun to hang out with my cousins Russell and Lisa.    And Uncle Norman is recovering well from the operation he had last month.  Now thereâ€™s nothing left to do but eat the leftovers.    (Pamelaâ€™s Passover tidbit : An orange slice can be put on a seder plate because when women were attempting to be ordained as rabbis, a prominent male rabbi angrily proclaimed â€œThe day women are allowed on the pulpit is the day oranges will be put on seder plates!â€.  So since women have indeed achieved positions as rabbis, itâ€™s appropriate to add oranges as a nod to their achievements.)    (Russellâ€™s Passover tidbit â€“ if they do a spoof of the movie â€œOld Schoolâ€, they should call it â€œOld Shulâ€)  How bout dem Mets?  I did not see the game last night but I heard that they kicked major ass.  A 7-2 win over the Braves with hot performances by Tom Glavine and Kaz Matsui.  And even better â€“ my fantasy baseball team, Thugz Inc., is in first place!  I donâ€™t know how long that will last so I will gloat as long as I can.   '),(22138,2824507,'06,April,2004','2004-04-06 00:00:00','    One Passover seder down.  One to go.  Kudos to mom for making a great dinner last night.  It started with the traditional reading of the \"New Hagaddah\".  It\'s called \"New\" even though the books we use were published in 1948.  The seder began with this funny exchange...  Dad : \"We\'ll begin on Page 3.\" Aunt : \"Where\'s Page 3?\" Brother : \"It\'s after Page 2.\"  50 pages later and after a hearty singing of Dye-anu, we finally ate.  The menu featured gefilte fish, matzah ball soup, turkey, baked potatoes, corn, carrot tzimas, haroset, pickles, veggies, matzah, and all the Manischevitz you could drink.  Dessert had ring jells, marshmellow twists, fruit slices, and various sugar-free stuff for the diabetics.  Round two is tonight in New Jersey.  Front page of the Daily News proclaims that J-Lo\'s mom won $2.4 million in Las Vegas.  Like she needs it!  Some good news is that Ralph Nader failed to make the ballot in Oregon.  He had only 700 supporters turn out last night when there was a 1000 person minimum in order to qualify for the ballot.  At the same function four years ago, he had 7000 supporters turn out.  Maybe people are finally getting it.   '),(22139,2824507,'05,April,2004','2004-04-05 00:00:00','    No work today because it\'s Passover!  T-minus a few hours until family descends upon Brooklyn for the first night\'s seder.  Stay tuned to this blog for a complete food review.  Today I am officially 26 and a half years old.  Does anyone still celebrate half-birthdays?  There was a great episode of \"The Sopranos\" last night.  It focused on Tony\'s infatuation with Adrianna and how they got into a car accident together at 2 in the morning from swerving off the road to avoid hitting a racoon.  Adrianna\'s fiance Christopher became enraged, thinking that the cause of the accident was oral sex.  There was a great montage of mobsters gossiping on the phone, embellishing the story of how it happened.  Chris then took the unprecedented step of attempting to assassinate Tony and wound up almost getting whacked himself.  Enter Steve Buscemi\'s character who urged everyone to drop their weapons and think rationally.  Buscemi took them to the hospital where Adrianna was treated and got the doctor to tell Chris that she had abrasions on her body from the seatbelt.  Hence, she could not have been engaged in any such acts and Chris had to apologize to Tony.  The episode ended with Chris, Adrianna, Tony, Tony\'s wife, Buscemi and Buscemi\'s mother all going out to dinner in public, encouraging the perception that there is peace.  The previews for next weeks episode indicate that its going to be all about Buscemi.   '),(22140,2824507,'04,April,2004','2004-04-04 00:00:00','    How come nobody told me that today was Daylight Savings Time?!?  Boy was I surprised when I woke up this morning, thinking that it was 10:00 but that it was really 11:00!  There\'s a big rain downpour going on right now.  I was caught outside in it earlier and am now drying off.  Saw two Netflix movies yesterday.  The first was \"The Rundown\" with The Rock.  It was mindless yet enjoyable.  I met The Rock back in 1997 when he was known as Rocky Maivia.   urlLink Click here to see a picture of me and Rocky.   (It\'s the second one)   It\'s good to see him succeeding now as a big movie star.  He\'s certainly entertaining enough.  The second movie I saw was \"Intolerable Cruelty\".  Hated it!  It was hyped as a screwball comedy but there was nothing screwy or comedic about it.  I can\'t believe it took four people to write the screenplay to such a piece of dreck.  Avoid at all costs!   '),(22141,2824507,'03,April,2004','2004-04-03 00:00:00','    Not much to report today.  Passover is approaching and family seders are being held on Monday night and Tuesday night.  Looking forward to matzoh ball soup and those fruit slices coated in sugar.   '),(22142,2824507,'02,April,2004','2004-04-02 00:00:00','    Three of my friends have birthdays today so happy birthday to Jon Gans, Gabe Anderson and Andrew Chazen!  Yesterday I forgot to make a â€œSurvivorâ€ prediction.  The unlucky castaway to get the boot was Jerri the Bitch.  OK, maybe she wasnâ€™t as bitchy as she was in Australia three years ago.  But she was still pretty damn annoying. With 9 players left, Rob and Amber remain in control as they have seemingly made alliances with everyone.  In other TV news, the shit finally hit the fan on â€œ24â€.  Or should I say, a biological virus hit the fan.  After chasing after the virus all season, it was finally unleashed within a hotel in L.A.  People inside the hotel didnâ€™t waste anytime showing symptoms so it looks like thereâ€™s going to be quite the casualty count.  Unfortunately, Kim was not inside the hotel.  And be sure to check out reruns of â€œSouth Parkâ€ this weekend.  The episode was titled â€œThe Passion of the Jewâ€ and is a great satirical look at Americaâ€™s interest in â€œThe Passion of the Christ.â€  The episodes main conclusion?  That Mel Gibson is insane.  So far so good with â€œAir Americaâ€ radio.  Iâ€™ve been listening to â€œThe Oâ€™Franken Factorâ€ for the past two days and itâ€™s pretty good.  Franken has a professional radio host co-hosting the show with him because heâ€™s a novice to the genre.  But the content has been mostly satisfying.  So far theyâ€™ve interviewed Bob Kerrey, Hillary Clinton, Michael Moore, and Richard Reich.  Al Gore and Ben Stein called in and Bebe Newirth provided the voice of Ann Coulter who, as the skit went, got locked inside of a closet in the Air America studios.  Hopefully the station can soon get on the air in the swing states.   I hate to say it but it seems that the Mets are really going to stink this year.  Theyâ€™re losing all of their spring training games, Jose Reyes is out with some mysterious leg ailment, Kaz Matsui is having a lot of trouble adjusting to U.S. pitchers and their starting rotation is still undecided.  Iâ€™d like to amend my original 4th place prediction.  I now say they will come in last.  But perhaps its best to enter the season with low expectations so any successes they do have will be a nice surprise.   '),(22143,2824507,'01,April,2004','2004-04-01 00:00:00','    Today is April Fools Day.  That means that I donâ€™t trust the veracity of anything that anyone may say to me today.  That includes you!  Last night I had a good time hanging out with Sarah, as well as her cute dog Zoey!  We saw â€œEternal Sunshine of the Spotless Mindâ€ at the AMC theater on 42nd Street.  I thought the movie was very well done.  Jim Carrey proves that heâ€™s a very effective actor when acting in serious, dramatic roles.  This has to rank up there with â€œThe Truman Showâ€ as one of his best performances.  I also enjoyed how Charlie Kaufman constructed the screenplay and the visual effects were unique as we witnessed the erasure of Carreyâ€™s memory.  Thereâ€™s a subplot with Kristin Dunstâ€™s character that I could have done without but the central drama between Carrey and Kate Winslet is what pulls the film along.  I highly recommend this movie!  The Washington Post is reporting that on 9/11/01, Condoleezza Rice was originally scheduled to give a speech on the state of our national security.  It was to have focused on long-range missile defenses and would not have mentioned Al Qaeda, Bin Laden or Islamic-extremist terrorism at all.  I think this effectively proves that, prior to 9/11, terrorism was not the slightest priority for this administration.    Overheard on public transportation this morning : Girl : â€œWhat are you reading?â€ Boy : â€œThe Autobiography of Malcolm X.â€ Girl : â€œYouâ€™re reading that on your own?â€ Boy : â€œYeah.  What you think Iâ€™m retarded?â€ Girl : â€œNo.  Iâ€™ve just think its great that a guy is reading without being forced to.  Iâ€™ve never seen that beforeâ€ Boy : â€œNah, this is great stuff.  Everyone should read this.â€  That exchange put a smile on my face.  Even though Iâ€™m not the biggest fan of Malcolm X, Iâ€™m glad to see that reading isnâ€™t entirely forgotten among urban youths.   '),(22144,2824507,'31,May,2004','2004-05-31 00:00:00','    Sorry for the lack of recent blog postings this weekend.  I\'m too busy relaxing.  Fans of this blog (if there are any!) will notice that I have posted a picture on the right-hand column, alongside my Blogger profile.  Yes, it\'s the same picture that\'s on Friendster.  I\'m hoping to purchase a digital camera soon and once that happens, look for more pics on this site!   '),(22145,2824507,'29,May,2004','2004-05-29 00:00:00','    I stayed up late last night finishing reading \"Angels and Demons.\"  Just could not put it down.  It\'s a great book by Dan Brown, author of \"The Da Vinci Code.\"  However, having read \"Da Vinci\", I am familiar with Brown\'s writing style and was therefore able to decipher the twist - who the villain was - halfway through the book.  Even though I had a good idea who the bad guy was, I still enjoyed how the author got us there and there was one final twist at the end that had me saying \"Holy shit\" out loud.  I know that Ron Howard is currently prepping a \"Da Vinci\" film and having now read \"Angels and Demons\", I really hope that a film version of it follows.  All I ask is that Ben Affleck NOT be cast as Robert Langdon.  Today they are dedicating the World War II Memorial in Washington D.C.  I\'m watching it right now on The History Channel.  This year on Memorial Day, as we enjoy our beaches, BBQs and clothing sales, let\'s take some time to honor the sacrifices of all of our veterans, in wars both just and unjust.   '),(22146,2824507,'28,May,2004','2004-05-28 00:00:00','    The beach opening went off without a hitch.  The weather cooperated, unlike last year when it rained on us.  I even got to meet the Mayor and he gave me a kazoo!  One of the many giveaways was souvenir kazoos donated by Astroland Amusement Park.  As Mayor Bloomberg walked in my direction, I stuck out my hand and introduced myself.  He shook it and said â€œHello.  I am going to give you this.â€  He then handed me a kazoo that someone else had previously given to him.  Quite the encounter!  Otherwise it was a long day but we got a lot of positive media coverage so Iâ€™m happy.  Since I subscribe to DirecTV, I donâ€™t get the station NY1 anymore but I was told that I could be seen in one of the shots, with my back to the camera!  I also met a guy who works for the Disney Corporate Office in Florida.  He was vacationing in NYC and seemed very impressed with Coney Island.  Maybe heâ€™ll go back to work and suggest that they invest, ala Times Square.  It was also great to have worked with the volleyball teams of Grady and Wingate High School, arranging for them to take part in the event by playing beach volleyball.  Those kids and their coaches were thrilled to be there and I was happy to have them.  But the best parts were what happened after the formal event ended... an unlimited lunch buffet courtesy of Greg & Paulâ€™s and a ride on the Wonder Wheel.  Yay Coney Island!    As is the tradition, the beach opening ceremony is always followed by Memorial Day weekend, which is the perfect time to crash and burn.  I intend on doing just that.   '),(22147,2824507,'26,May,2004','2004-05-26 00:00:00','    Today I bought a Lotto ticket.  I just watched the drawing.  I didn\'t win.  Shit.   '),(22148,2824507,'25,May,2004','2004-05-25 00:00:00','    â€œ24â€ went out with a bang last night.  Or should I say, a chop.  In one of the more gruesome scenes in the seriesâ€™ history, Jack was forced to chop off Chaseâ€™s hand with an axe.  You see, he had no choice since Chaseâ€™s hand was permanently clamped to the deadly biological weapon that they had been hunting all year.  In other developments, President Palmer decided not to seek a second term, Tony is off to serve at least 20 years in jail, Saunders was killed by Gaelâ€™s widow, and Chase and Kim decided to live a calmer existance.  The final scene featured Jack privately sitting in his car, overcome with tears... the first time he ever cried on the show.  A quiet yet powerful conclusion.  â€œ24â€ will not return until January, 2005 when FOX will then present it without any preemptions.  They might do so with Kiefer Sutherland but without many of the supporting players.  Busy week at work.  Busy busy busy.  Tomorrowâ€™s blog entry will be entered a bit later than usual in order to provide a full recap of the Beach Opening Ceremony.   '),(22149,2824507,'25,May,2004','2004-05-25 00:00:00','    Bush gave another speech last night.  Donâ€™t know why I bothered to watch.  Maybe it was because nothing else was on.  Maybe it was because he was supposed to have laid out some actual specifics as to what our Iraq endgame would be.  Silly me.  It turned out to be little we havenâ€™t heard before.  â€œWe will transfer sovereignty by June 30... We will rebuild their security... We will rebuild their infrastructure.â€  Yes, but how???  In terms of transferring sovereignty, Bush said that an envoy from the UN will decide who will be the new President, Prime Minister, VPs and cabinet members.  Sounds like that envoy should have been the one giving the speech instead.  Of course no mention was made of the very real possibility that a new Iraqi government would ask U.S. forces to leave the country.  Then what?  More speeches will be forthcoming in the next month and I really hope they are more illuminating.  Last winter I read â€œThe Da Vinci Codeâ€ and really enjoyed it.  Now Iâ€™m reading â€œAngels and Demons,â€ the first Robert Langdon novel.  Itâ€™s very much in the â€œDa Vinciâ€ formula and it reads like an season of â€œ24â€, fast paced, taking place in one day, with a race to stop a catastrophe from happening.  So far, so good!   '),(22150,2824507,'24,May,2004','2004-05-24 00:00:00','    Was at Coney Island earlier this morning getting ready for the big Beach Opening event, which is on Thursday.  Today I was putting up the 2004 Event Calendar posters that I made on all of the informational kiosks.  So if youâ€™re at Coney this summer and you see the good-looking promotional posters, know that it was all me!  As Cher said in â€œMoonstruckâ€... â€œI donâ€™t believe in curses.â€  To which the old Sicilian woman replied, â€œNeither do I.â€  Itâ€™s obvious that neither of them are baseball fans.  Because true baseball fans know the power of a curse.  The plights of the Chicago Cubs and the Boston Red Sox are one example.  Another is that no Mets pitcher will ever throw a no-hitter.  Case in point â€“ yesterdayâ€™s game.  Tom Glavine came within 4 outs of a no-no, only to be denied.  Through the years, the Mets have had many great pitchers... Tom Seaver, Dwight Gooden, David Cone.  All of whom never threw a no-hitter for the Mets but did so later in their careers after they left the team.  When I was a little kid I used to say that I would erect a statue in my frontyard of the first Met to throw a no-hitter.  As the years go by, it seems that is one statue that will never be built.  â€œThe Sopranosâ€ was AMAZING last night.  Full of action and drama!  Perhaps one of the best episodes ever.  Yet also disturbing.  It culminated in the whacking of Adrianna, a regular since the seriesâ€™ inception.  Adrianna had been giving information to the FBI for the past two seasons and once the family found out... well letâ€™s say she knew exactly what was going to happen as Sylvio drove her into the woods.  It was also a noteable episode as Tony Soprano refused to hand Steve Buscemiâ€™s character over to the rival mob boss for a retribution killing.  Now it looks like full-fledged war between the two families.  Or to put it another way : Johnny Sack said to Tony, â€œYou Got Served!â€  To which Tony replied, â€œNo, You Got Served!â€  To which they resolved, â€œItâ€™s On!â€  Now if this were â€œSouth Parkâ€ it would culminate in a dance-off.  Donâ€™t think thatâ€™s gonna happen though...   '),(22151,2824507,'23,May,2004','2004-05-23 00:00:00','    Yesterday was busy.  Went to not one but two birthday parties.  Go social life!  One of the parties was in the afternoon and the other was late at night so I had some time to kill in between.  So I went to the McDonalds on 42nd Street and ordered a Number 3 Meal - SuperSized!  After gorging myself (and only being able to finish half the fries and soda), I then went to the AMC 25 theater and saw the film \"Super Size Me.\"  It was awesome!  As I sat in the theater suffering from the effects of my own Mickey Ds binge, I watched Morgan Spurlock as he only ate McDonalds three times a day for thirty straight days.  As a result he gained 25 pounds, his cholesterol shot up, he suffered liver damage, had problems performing sexually, and suffered through bouts of depression.  But it sure tasted good!  Actually, only part of the movie dealt with his experiment.  The rest dealt with the American obesity epidemic in general.  It\'s a great film and a good companion piece to the book \"Fast Food Nation.\"  Oh and on my way walking from the theater to the 2nd party, I passed three McDonald\'s.  They really have taken over Manhattan.  And speaking of documentaries, Michael Moore\'s \"Fahrenheit 9/11\" won the Cannes Palmes D\'Orr Prize last night as the best film.  Will definetely see that one as soon as it comes out!   '),(22152,2824507,'21,May,2004','2004-05-21 00:00:00','    Did everyone hear about the interview with Condi Rice where she answered â€œAs I said to my husb...I mean the Presidentâ€  Yes, Ms. Rice was making a reference to Bush and the first words that almost blurted out of her subconscious when thinking of him were â€œmy husband.â€  Does Little Miss Condi have a crush???    The comic strip â€œThe Boondocksâ€ has jumped all over the flub, joking around that theyâ€™re having a steamy love affair.  In yesterdayâ€™s strip, Huey is watching a TV news interview with Rice and she tells a reporter, â€œAs I was saying, weâ€™re all horrified about the treatment of the detainees.  Why, I was just saying to my snoogumcakes...â€ (stunned silence)  â€œI â€“ um â€“ meant the President.â€  To which Huey yells out â€œSHE DID IT AGAIN!!â€   '),(22153,2824507,'20,May,2004','2004-05-20 00:00:00','    The news is so gloomy.  Iraq is a mess.  The 9-11 hearings are contentious.  The Mets lost yesterday.  Nothing good in todayâ€™s paper to talk about.  Except for a humorous story about a insurance executive who was billed for $28,000 at Scores Strip Club.  Heâ€™s suing them for wrongful charges.  The spokesperson for Scores said that the lawsuit boils down to his â€œpartying like a champ but waking up the next day feeling like a chump.â€    The expense list is itemized as follows... $16,000 for Champagne   (5 magnums of Clos Du Mesnil at $3,200 each) $7,000 for Lap Dances   (350 for $20 each, including dances by up to 12 women at a time) $5,000 for more alcohol, food and tips  Well, I hope he had fun!   '),(22154,2824507,'19,May,2004','2004-05-19 00:00:00','    Happy Birthday,  urlLink Sarah !!!  Now you are as old as me!  (Give or take seven months or so...)  NYC has made the Final Five cut for the 2012 Summer Olympics.  However in the voting, we only came in 4th, behind Paris, Madrid and London.  (Moscow is the other finalist)  I donâ€™t think the odds are good that NYC will be chosen.  With the 2010 Winter Games in North America, theyâ€™ll probably choose one of the three European cities for 2012.  Iâ€™m disappointed though that Havana, Cuba failed to make the cut.  How great would that be?  Castro would never let anyone leave!  Congratulations to Randy Johnson on throwing a perfect game last night.  Not bad for 40 years old!  Itâ€™s even sweeter that he did it against the Atlanta Braves.  Theyâ€™re not so tough this year, are they?  Add that to another 9th inning come-from-behind win by the Mets and it was a good night for baseball.  Televisionâ€™s greatest villain, Sherry Palmer from â€œ24â€, was killed off last night.  She was the quintessential manipulater, always scheming against her husband and plotting her way to power.  All of her schemes finally caught up with her though as she was gunned down by one of the women who she betrayed.  In three years on the show, Penny Johnson Jerald did an amazing job in her portrayal of the character and sheâ€™ll be missed!   '),(22155,2824507,'18,May,2004','2004-05-18 00:00:00','    I cannot wait to see â€œFahrenheit 9/11.â€  Itâ€™s the latest documentary from Michael Moore and it got a 20-minute standing ovation at the Cannes Film Festival.  Not surprising since the screening was in France and it really tears into the policies of the Bush Administration.  It includes a rarely seen 7-minute clip of Bush being informed that the second tower was struck, and then sitting in a kindergarten chair with a vacant, baffled look on his face.  It also shows war atrocities in Iraq with many of our own soldiers questioning what weâ€™re doing.  Plans are for the film to be released in theaters this summer (provided that they find a distribution company) and then on video and DVD in the fall.  Hopefully it will be seen by many voters as a wake-up call.  Sad to say that Tony Randall of â€œOdd Coupleâ€ fame passed away this morning.  I met Randall in November, 1999 at the Marriot Marquis Hotel in NYC.  I was working with then-Parks Commissioner Henry Stern at a banquet in honor of City College.  Stern and Randall were both guest speakers and I had the pleasure of an introduction.  I wonder if Nick at Nite or TV Land will show an â€œOdd Coupleâ€ marathon in his honor.  Vassar Collegeâ€™s 5-Year reunion for the Class of â€™99 is coming up!  June 4, 5, 6.  If you havenâ€™t registered,  urlLink do so now at the AAVC home page!   '),(22156,2824507,'17,May,2004','2004-05-17 00:00:00','    A word about the new Blogger system for comments.  In order for your name to appear as having left a comment, then you have to sign in as a Blogger member.  If you are NOT a Blogger member, you can still leave a comment by clicking the link to post anonymously.  But to let me know who you are, please say so in the text of your comment.  Thanks!   '),(22157,2824507,'16,May,2004','2004-05-16 00:00:00','    The front page of todayâ€™s Daily News asks â€œWould you pay $1,000 for an omelet?â€  My answer is... no, I would not even pay $1 for an omelet.  I do not like omelets.  I do not like green eggs and ham either.  But apparently, Le Parker Meridian Hotel on W. 57th Street is charging a grand for an omelet that consists of eggs, lobster and lots and lots of caviar.  It was also reported that noone has ordered it yet.  Hilarious stuff on â€œMeet the Pressâ€.  Colin Powell was being interviewed by Tim Russert via satellite from Jordan.  When Russert asked a hard-hitting question about using false information to justify war, Powellâ€™s staffers moved the camera off of him and pointed it toward some palm trees.  His deputy press secretary told Powell that the interview was over but Powell yelled at them to put the camera back on him so he could answer the question.  Powell later called Russert to apologize.    Great Mets game yesterday.  After being shut out by Roger Clemens for 7 innings, the Mets got even with the Houston bullpen in the 9th as Mike Piazza hit the game-tying homer.  For Piazza its an indirect form of revenge against Clemens.  Who will ever forget the beaning he took in the head during the 2000 season?  Then in extra innings, Jason Phillips hit the game-winning home run.  â€œThe Sopranosâ€ was certainly surreal last night.  Much of the episode was an enactment of Tonyâ€™s dream, breaking all rules of time and space and bringing back many of the showâ€™s former cast members who had been whacked in previous episodes.  The dream reflected much of Tonyâ€™s inner torments and anxieties and foreshadowed the troubles he is soon going to face.  Steve Buscemiâ€™s character (Tonyâ€™s cousin) murdered a member of a rival mob family and now Tony fears that his family is at risk for retaliation.  The last two episodes of the season should be great!   '),(22158,2824507,'16,May,2004','2004-05-16 00:00:00','    This morning I went to Lundy\'s (a Sheepshead Bay institution) for a breakfast in honor of Congressman Anthony Weiner (who used to be my boss many many years ago!)  It was great to see him and catch up with many of the people who I used to work with including Jeremy Halperin, Brian Gottlieb and Suzanne Bressler.  Now the only question is if Weiner will run for Mayor next year.  Hmm...  Speaking of Sheepshead Bay institutions, Roll \'n Roaster has been a hallmark on Emmons Avenue for the past 33 years.  They just recently opened a restaurant in Manhattan, on 3rd Ave & E 11th St, and I ate there last night.  Tasted the same to me!  Saw \"Mean Girls\" last night.  Sharp, funny movie that was highlighted by Lindsay Lohan\'s... ummm... assets.  When does she turn 18?   '),(22159,2824507,'15,May,2004','2004-05-15 00:00:00','    Special thanks to Marivel for e-mailing this to me...  Have you heard of the NO-CARB Diet for 2004?  NO C-heney NO A-shcroft NO R-umsfeld NO B-ush  and \"Absolutely NO RICE!\"  This is a diet that will take you successfully through the next election.  -----------------------------------------------------------  Yesterday\'s Parks event was a ribbon cutting at Police Officer Reinaldo Salgado Playground in Bedford-Stuyvesant.  It was for a new synthetic turf ballfield, replacing an asphalt one.  Definetely a good thing for all the kids who will be using it.  Got up early this morning and watched \"The Last Samurai.\"  Lots of exciting battle scenes and an engrossing historical epic of Japan\'s transition in the late 19th Century.  And yeah, Tom Cruise was good too.   '),(22160,2824507,'14,May,2004','2004-05-14 00:00:00','    There is nothing more soothing than listening to the rain fall as you settle down to sleep.  Last night was great in that regard.  When I heard that Bruce Ratner was going to build a new arena complex for the Nets in Brooklyn, I originally felt bad for the homeowners that would be displaced.  Now I hear that all property owners have been offered over $1 million to move.  Suddenly I don\'t feel too bad for them anymore.   Another Parks event going down today.  It\'s at 1:00 though which will severely cut into my lunch.  Hmm, will have to grab something on the way.  Am determined to see a movie (or two!) this weekend.  Don\'t forget to visit  urlLink VassarChat !   '),(22161,2824507,'13,May,2004','2004-05-13 00:00:00','    RUPERT!!!!!!     Rupert won the million dollar vote, no doubt thanks to all the votes that me, my dad, my aunt and my co-workers threw to him.    Yeah, Rupert!!!  OK, no more Survivor talk for a while.   '),(22162,2824507,'12,May,2004','2004-05-12 00:00:00','    Saw an excellent movie the other night.  \"Osama\", the first movie made in Afghanistan since the fall of the Taliban.  It\'s about a family of women who are starving to death because all the men in the family are dead and the Taliban bans women from working to earn a living.  So they cut the hair off their youngest daughter and have her pose as a boy in order to work.  Meanwhile, all the young boys in the village are rounded up to serve in Bin Laden\'s army - including her.  It\'s an amazing film and very reminiscient of \"Europa Europa\", where a Jewish boy pretends to be German in order to escape the Holocaust and even serves in Hitler Youth.  Terrible, terrible pictures out of Iraq.  First there\'s the prisoner abuse photos.  Then there\'s the beheading of the American.  It\'s as if the terrorists are saying \"Oh you think you\'re sick?  Watch what we can do.\"  Voting for the next Survivor million dollar winner has closed with the winner to be announced tonight.  I voted for Rupert a few times.  I tried to vote for myself but did not see my name on the ballot.   '),(22163,2824507,'12,May,2004','2004-05-12 00:00:00','    Fiddling around with Blogger.  Changed the visual interface again.  Testing, testing, one two...  And oh yeah, the comments are back.  Contain your excitement.   '),(22164,2824507,'11,May,2004','2004-05-11 00:00:00','    Announcing a brand new website...   urlLink VassarChat.com   An opportunity for alumni, current students and  prospectives to come together and talk about  Vassar, dish gossip, discuss future plans,   promote causes, talk about the news and much much  more!  Log on now, register with the site and leave  comments on our easy to use message board.  Then  tell your friends about it and keep coming back!   urlLink VassarChat.com  is run by fellow alumni and is  not affiliated with Vassar College.  See you there!   '),(22165,2824507,'10,May,2004','2004-05-10 00:00:00','    Said goodbye to Tom Stofka today.  His funeral was the first Catholic one that I have ever attended.  Literally hundreds of people were there, with Parks employees filling the church pews all the way to the back.  Across the street from the church, a forestry climber & pruner truck was parked with lettering that read \"We Love You Tom Stofka.\"  An honor guard of Parks supervisors and enforcement officers stood at attention at the front doors.  The priest and Commissioner Benepe both spoke exceptionally well in their eulogies.  Tom will certainly be missed.  He was a highly thoughtful and insightful individual.  He and his wife would leave notes to each other every so often with positive affirmations on life.  The day that he died, his wife found a note from him that he apparently wrote that morning before he left for work.  A coincidence or did he have an inkling that something might happen?  This is what he wrote...  - Want more for others than yourself - Be an appreciator - Stay in rapport with source energy - Understand resistance - Contemplate yourself surrounded by the conditions you want to create - Meditate connected to source - Stay matched with the field of intention - Think from the end - visualize results - You can resolve a problem by condemning it - Be in a constant state of gratitude - Practice radical humility  Thank you Tom for leaving us with those final, valuable words.  And thank you for living the life that you led.   '),(22166,2824507,'09,May,2004','2004-05-09 00:00:00','    Woah.  Blogger sure changed its visual interface last night.    And the winner of \"Survivor : All Stars\" is... Amber!  I picked it!  Rupert was first to go, then Jenna, then Amber beat Boston Rob by a 4-3 vote.  Voting for Amber was Shii Ann, Lex, Alicia and Tom (the latter three were out of spite for Rob).  Voting for Rob was Rupert, Jenna and Kathy.  Then on the live reunion show, Rob proposed marriage to Amber and she accepted.  What a Survivor love story.  I hope she signs a pre-nup. :)  The \"big surprise\" was something of a letdown.  It\'s that CBS will award a second million dollar prize in a \"viewer\'s choice poll.\"  You can vote by logging on to  urlLink CBS.com.   '),(22167,2824507,'09,May,2004','2004-05-09 00:00:00','    HAPPY MOTHERS DAY!  Can\'t wait to go out to dinner tonight and eat lobsters with mom.  Tonight is the big finale of \"Survivor : All Stars\".  All season I\'ve been offering my weekly picks and it all culminates tonight.  The first elimination will be interesting.  Rob and Amber are an alliance.  So are Rupert and Jenna.  Will it be a 2-2 tie?  If so, what\'s the format for a tiebreaker?  Will one of the players abandon their alliance to secure a 3-1 edge?  Then when it comes down to 3 players, it\'s all about who wins the final crucial immunity challenge.  The unpredictable nature of the final four makes predictions difficult but here goes.  I\'ll say that Rupert is the first voted out.  Then Boston Rob.  Then the jury will choose between Amber and Jenna and they\'ll give the win to Amber.  CBS is hyping huge surprises for tonight, which could be anything!   '),(22168,2824507,'07,May,2004','2004-05-07 00:00:00','    Thanks to John and Spencer for throwing a supremo BBQ tonight!  Though, a BBQ before Memorial Day?  Isn\'t that sacriligious?  Mmm, sacrilicious...  So after meeting David McCullough on Friday, I went to the bookstore and purchased his book, \"John Adams\".   I enjoy his references to the journals that Adams kept when he was in his late teens & early twenties.  His journal entries sound no different than the blog entries that me and many of my friends write nowadays.  Just think, if John Adams was alive today, he\'d have a blog!   '),(22169,2824507,'07,May,2004','2004-05-07 00:00:00','    UPDATE : I just met David McCullough.  The historian and author of \"John Adams\", \"Truman\", and many other novels.  He came into our building today as he tours historical points of Brooklyn.  And I introduced myself and shook his hand.  I feel giddy.   '),(22170,2824507,'06,May,2004','2004-05-06 00:00:00','    Still devastated about Tom Stofkaâ€™s death.  But life goes on.  A bunch of us went out last night for food and drinks, and we talked about Tom as well as other things.  I think the camaraderie helped us cope.  In his honor, American flags in Brooklynâ€™s parks are at half-mast.  So how about the shit hitting the fan about the abuse treatment of Iraqi prisoners?  Rumsfeld is in trooouuubbbllleee...  Apparently, Rumsfeld knew about it for months and hid the reports from Bush.  Yesterday, Bush had to actually use the word â€œsorryâ€.  This administration never apologizes for anything!  We very may well have a new Secretary of Defense before itâ€™s all said and done.  And now a few TV commentaries.   1) Big Tom Buchanan was voted off of â€œSurvivorâ€ last night.  He could have played the role of a crucial swing vote but neither alliance felt they could trust him for the duration.  The final four is now Boston Rob, Amber, Rupert and Jenna with the finale to be aired on Sunday night.   2) I caught the last half-hour of â€œFriendsâ€ last night and was not impressed.  I never really watched the show to begin with and couldnâ€™t care less that itâ€™s off the air. 3) The season finale of â€œThe O.C.â€ was the other night.  For the first time, a party on that show did not end with a fistfight or someone getting thrown into a pool.  Very sad ending though with Ryan going back to Chino out of obligation, and against his wishes.  I think itâ€™s pretty clear though that heâ€™ll move back to the Cohenâ€™s house very early next season.  And what was with Seth sailing off into the sunset?  Gabe raises a good point, heâ€™ll probably wind up drowning! 4) â€œ24â€ is down to its last few episodes and Tony Almeidaâ€™s character has absolutely lost it.  Blackmailed by the bad guy, Tony has betrayed his country and directly put the lives of millions at risk.  Canâ€™t wait for Jack to dole out his brand of justice.  A litany of bullets in Tonyâ€™s head perhaps?  Worked for Nina.  How bout those Mets?  They still stink but at least they swept the San Francisco Giants.  Guess the Giants arenâ€™t much of team when Barry Bonds doesnâ€™t play.  And my Fantasy Baseball team is even doing a little better.  If you can call 5th/6th place a little better.  I really want to see \"Super Size Me\", the documentary about the guy who only eats McDonalds for 30 days.    Have a good weekend!   '),(22171,2824507,'06,May,2004','2004-05-06 00:00:00','    Well this has been some day thus far.  This morning, there was a groundbreaking ceremony at Sternberg Park.  The event itself went well.  However, I then returned to the office and learned that one of my co-workers, Tom Stofka, died this morning of a massive heart attack.  He was Brooklyn\'s Forestry Director and a great guy who I\'ve worked with for five years.  I\'m still in disbelief as I type this.  I last saw him yesterday and come to think of it, he did look a little flush.  Then, five minutes before I learned of his death, I \"cc\'d\" him on an e-mail.  So right now we\'re all just sitting around the office, stunned.  Tonight we\'re supposed to go out and have drinks with Lauren, who left the agency last week.  I\'m still going to go out but I don\'t know how much of a celebration its going to be.  \"Survivor\" is down to five.  Anyone could go at this point.  Big Tom is a crucial swing vote as Rob and Amber, and Rupert and Jenna seem like they\'re going to stick together.  My prediction is that Tom sides with Rob and Amber and chooses to vote off Rupert.  But I hope that doesn\'t happen.   '),(22172,2824507,'05,May,2004','2004-05-05 00:00:00','    The following was posted today by  urlLink The Anonymous Blogette .  It\'s a great piece and I have decided to reprint it in its entirety.   SINGLE PERSON\'S BILL OF RIGHTS   1. I have the right to do something on my own without being subject to A) incredulity and B) pity. If I say I did something alone, do not look at me as if to say, \"Awww, you didn\'t have anyone to do it with?\" Maybe it was my choice. Maybe it was easier than dragging someone else along who was not as into the activity. If I stop at my favorite restaurant to eat alone right after work, I may have not wanted to base my schedule on someone else\'s on that particular eve.  2. I have the right to give, as an answer for why I can\'t meet up with you, that I\'m busy and not have to justify it. Just because I\'m single doesn\'t mean each evening is free. If I were married, you wouldn\'t be holding me responsible for having to turn down some of your invites. You\'d assume I have a life. Well, I am single, and I want to do things with you, but I *still* have a life. Just because I don\'t have to change diapers today doesn\'t mean there aren\'t things I need to get done. Don\'t quiz me on what I am doing if I turn you down. It\'s nothing personal. I may have planned some alone time. Please keep asking, as I would do the same with you.  3. I have the right to, on the one hand, value my alone time, and on the other hand, still want to seek out someone to share my life with, without being called a hypocrite or judged on it. I can be happy alone, but I might be happier with someone, IF it was the right person. These are not warring impulses. How hard is that to understand?  4. I have the right to expect that you will be looking out for someone for me to date. Yes, I do...Listen up: Are you a true friend? Well, the best way for people to meet is through friends. If you are a true friend, rather than doing #1 and #2 above, you will, upon meeting someone who is single and whom I might like, find out more about them and pass the info to me or see if they might like to meet me. It might be the nicest thing you can do for me.   I am trying to do this on my own as much as I can, but could use extra pairs of eyes. Dating services are awkward. Meeting appropriate people and getting to know them is HARD. Just look at  urlLink the blogs of my brothers and sisters in arms.   5. I have the right to complain about dating. If you haven\'t been out there and experienced it in a while, don\'t criticize me. Instead, re-read #4. And if I find someone, I won\'t be complaining about dating anymore.   '),(22173,2824507,'04,May,2004','2004-05-04 00:00:00','    The website I alluded to the other day is coming along nicely.  Stay tuned to this blog for the grand announcement of its domain name.  It will herald a new era in the Internet!  My apologies to Jay Bartow.  Upon arriving home last night, I switched off my cell phone.  Itâ€™s a habit since I usually use the cell during the day and then my home phone at night.  Besides, my cell doesnâ€™t get good reception in my apartment.  Anyway, I later turned on the cell to find a voicemail message from Jay yelling â€œIâ€™m driving through Brooklyn and am completely lost!  Help me find a way out of here!â€  Oops.  Sorry I didnâ€™t get your message when you needed it, Jay.  Hope you found your way out sooner or later.  Hilarious sketch on SNL last weekend!  It featured Rachel Draitch as â€œDebbie Downerâ€, a woman who always knows how to ruin a good mood.  While everyone else is having a good time she adds comments about diseases, natural disasters, terrorism and how she canâ€™t have children.  It was supposed to have been played completely deadpan but the entire SNL cast including Lindsey Lohan kept cracking up and breaking character throughout the entire sketch and that made it even funnier.  Sure to become an SNL classic.  When 9-11 happened, I figured that it would end Hollywoodâ€™s habit of making disaster movies.  Surely, after witnessing the actual collapse of the Twin Towers, who could stomach such similar fictional scenarios?  Well itâ€™s been two and a half years since 9-11 and disaster movies are back.  Case in point, â€œ10.5â€, a miniseries aired for the past two nights on NBC.  Yes, the acting was terrible and ridiculously melodramatic but noone watches disaster movies for the acting.  We just want to see shit get destroyed.  Well â€œ10.5â€ gave us the destruction of the Seattle Space Needle, the Golden Gate Bridge, and the entire coastline of Southern California.  Soon Hollywood will give us â€œThe Day After Tomorrowâ€, where the entire world goes to hell!  So soon after 9-11, is this in the best taste?  I guess the ratings and box office will tell.   '),(22174,2824507,'03,May,2004','2004-05-03 00:00:00','    Woke up, put on clothes, went to work.   '),(22175,2824507,'02,May,2004','2004-05-02 00:00:00','    Maybe I should have placed some money on Imperialism after all.  Although he did not win the Kentucky Derby, out of 18 horses, he came in a very respectable 3rd.  Supposedly you can bet that a horse will come in \"place\" which means that you\'d win $$$ if they finish in 3rd or higher.  Guess I should have done that.  Ah well, that\'ll be my plan for the Preakness Stakes!  Saw two good movies on Netflix yesterday.  One was \"Shattered Glass\".  Just as Hayden Christensen plays Anakin Skywalker in Episode II as a whiny bitch, he plays Steven Glass, the New Republic reporter who faked his articles, as... a whiny bitch.  Guess it works for him.  The second movie I saw was \"Win a Date with Tad Hamilton.\"  Standard formulaic teen romantic comedy...but it had heart!  Or something like that.  Now I need to drag my ass to an actual movie theater and see \"13 Going on 30\" and \"Mean Girls\".    Stay tuned to this blog for an announcement of a new, innovative website that will be of interest to most of my readers.  I will write more when more details become available!   '),(22176,2824507,'01,May,2004','2004-05-01 00:00:00','    Yesterday was so busy that I didn\'t have time to blog!  It was Arbor Day (a big deal for the Parks Dept) and we had two events to celebrate it.  The first was at Carroll Park in Carroll Gardens and the second was at Kaiser Park in Coney Island.  I was only supposed to have been at one of them but wound up at both when the programs were left behind at the office and I was the only one available to bring them over.  The weather was great though so it was ideal to have been outdoors all day.  Then my mom had knee surgery yesterday on a torn miniscus.  All went well and she is currently home and recovering nicely.  I\'ve never bet on horse racing before but I was walking past an OTB this morning and I became highly tempted.  Today is the Kentucky Derby and I considered placing a bet on a horse.  I ultimately decided not to, seeing as how I know nothing about the horses and that it would likely be a complete waste of cash.  However the horse\'s name that caught my eye was \"Imperialism.\"  So if Imperialism does indeed win today, then I might have to trust my gut in the future.  If Imperalism loses, I\'ll just forget the whole thing.  I just heard that U.S. Supreme Court Justice David Souter got beat up last night by a bunch of young punks while he was jogging.  That sucks.  Couldn\'t they have targeted Scalia instead?  (Kidding!)   '),(22177,2824507,'30,June,2004','2004-06-30 00:00:00','    Today is the last day of June.  It has been a very long, busy and enjoyable month.  Seeing as how it is the end of the fiscal year, I came into work this morning and found a plethora of new office supplies waiting for me at my desk.  Stationary, envelopes, staples, paper clips, white out...  it was like Christmas morning!  Nicole Kidman, goddess of the cinema, is having problems meeting men.  Her words from a recent interview : \"I\'m hoping to meet someone and be happy with them. But that\'s not as easy as it sounds. I\'m a 37-year-old woman with two children. Men aren\'t beating a path to my door.  I don\'t want to sound like a woman from a lonely hearts club and I don\'t want to advertise. The children are my priority. I take them around with me... and that\'s not so appealing for any new man on the scene, is it?\"  Câ€™mon guys, any volunteers???  The final four women left at Wimbledon are Sharapova, Davenport, Amelie Mauresmo and Serena Williams.  Earlier this morning, Serena spanked Jennifer Capriati 6-1, 6-1.  And the pride of Great Britain, Tim Henman, lost again.  A perennial semifinalist, this year he lost in the quarters.  Poor Brits.   '),(22178,2824507,'29,June,2004','2004-06-29 00:00:00','    The debate over â€œFahrenheit 9/11â€ continues!  Commentaries can be found on Wendyâ€™s blog and Valâ€™s blog.  (see the links on the sidebar)  I am also attaching a link from  urlLink Denis Hamillâ€™s article from todayâ€™s Daily News .    This just in!  The Supreme Court ruled that underage kids can look at porn on the internet.  Their reasoning had to do with free speech.  I don\'t fully comprehend it but I cracked up laughing when I heard that the five who voted in the majority were the four liberal justices and CLARENCE THOMAS!  Perv...  Iâ€™ve been in a mood for documentaries these past few months.  Last night I watched â€œThe Fog of War.â€  It is an interview with former Secretary of Defense Robert McNamara and his take on the Cuban Missile Crisis and Vietnam.  Some of his insights are remarkable.  With reference to the Cuban Missile Crisis, itâ€™s amazing how close the world came to nuclear war.  Itâ€™s a good thing that Kennedy and Khruschev were both rational individuals, as Castro admitted to McNamara years later that he had urged Khruschev to use the nukes!  And with reference to Vietnam, itâ€™s a classic case of two sides not being able to understand the otherâ€™s point of view.  America viewed it through the prism of the Cold War.  The North Vietnamese viewed it through the prism of a civil war.  Now that McNamara is in his 80s, he is able to speak openly about his regrets and it is refreshing to hear a policymaker admit that they even have any regrets!   There is a new 17-year old tennis talent from Russia named Maria Sharapova.  Today she won her quarterfinal match at Wimbledon and is now in the semifinals where she will meet Lindsay Davenport.  Sharapova is being called the second-coming of Anna Kournikova as, not only is she a young talent from Russia, but she is also quite the looker.  Letâ€™s hope that, unlike Kournikova, the fame wonâ€™t go to her head and she can concentrate on her tennis.   '),(22179,2824507,'28,June,2004','2004-06-28 00:00:00','    It was a very fun weekend what with the â€œFahrenheit 9/11â€ outing, the day at Coney Island on Saturday, and a return to Coney on Sunday to take my parents to a Cyclones game.  This time, the Crappers beat them by a score of 6-5.  To which I said â€œOh Crap!â€  But thereâ€™s lots of news to talk about on this Monday morning.  After months of anticipating the June 30 transfer of sovereignty in Iraq, they instead did it two days early this morning with little fanfare.  How anticlimactic.  Itâ€™s like changing Christmas to December 23.  Now I have to call the caterer and cancel the party I was going to throw on Wednesday.  Kidding of course.  No doubt, the date was moved up and not publicized in order to prevent terrorists from attacking during the ceremony.  How effective will the new Iraqi leaders be against the insurgents?  Can they convince other Arab nations to embrace their legitimacy?  Or will they just serve as American puppets?  Lots of questions remain to be answered.  In other news from Iraq, an American Marine of Muslim descent was captured and is threatened for a beheading.  The terrorists are claiming that they infiltrated one of our camps and captured the marine.  The Pentagon is claiming that the marine deserted days ago and must have been captured elsewhere.  Who to believe?  Michael Moore is having a very good morning.  â€œFahrenheit 9/11â€ was the #1 movie at the box office this weekend.  This is even more impressive considering that it was only playing on over 800 screens whereas the #2 flick, â€œWhite Chicksâ€, was on around 2500 screens.  Whatâ€™s even better is that the filmâ€™s popularity is not confined to the blue states.  Theaters were sold out all over the red states that voted for Bush in 2000.  Its success is now prompting distributors to release it in even more theaters.  This film looks like its going to have some considerable staying power!  Hereâ€™s an interesting story.  Tiffany Schley, the valedictorian of the High School of Legal Studies in Brooklyn, gave a graduation speech where she criticized the school for its failure to provide resources and address studentâ€™s concerns.  This is obviously a very smart and astute girl as she will be attending Smith College on a full scholarship.  After giving the speech, she returned to the school to pick up her diploma and was barred from the building and denied the diploma out of sheer retaliation for her speech!  How repressive is that?!?  Only after the Daily News exposed the story did the school relent and grudgingly agree to give her the diploma.  Mayor Bloomberg had it right when he decried the schoolâ€™s administrators as â€œbozosâ€ who tried to deny her freedom of speech.   '),(22180,2824507,'26,June,2004','2004-06-26 00:00:00','    Saturday was a great day as I gave  urlLink Sarah  a complete tour of Brooklyn.  We started in Prospect Park, visited my office on the way, lunched at Two Boots, took the train to Coney Island, watched the Mermaid Parade, gawked at barely and strangely dressed people, walked on the beach, went on the Wonder Wheel and Thunderbolt, and then saw a Brooklyn Cyclones game.  They beat the Mahong Valley Scrappers, 4-3.  Or as Sarah likes to call them, the Crappers!  Pictures are posted below.   '),(22181,2824507,'25,June,2004','2004-06-25 00:00:00','    urlLink    Sarah & I at KeySpan Park&nbsp; urlLink   '),(22182,2824507,'25,June,2004','2004-06-25 00:00:00','    urlLink    Sarah at the Mermaid Parade&nbsp; urlLink   '),(22183,2824507,'25,June,2004','2004-06-25 00:00:00','    urlLink    KeySpan Park & The Parachute Jump&nbsp; urlLink   '),(22184,2824507,'25,June,2004','2004-06-25 00:00:00','    urlLink    Dick Zigun, Founder of the Mermaid Parade, and Marty Markowitz, Brooklyn Borough President&nbsp; urlLink   '),(22185,2824507,'25,June,2004','2004-06-25 00:00:00','    urlLink    Ah, street dancers...&nbsp; urlLink   '),(22186,2824507,'25,June,2004','2004-06-25 00:00:00','    urlLink    The Mermaid Parade  (The nudity should get this webpage a bunch of new hits...)&nbsp; urlLink   '),(22187,2824507,'25,June,2004','2004-06-25 00:00:00','    urlLink    Fahrenheit 9/11&nbsp; urlLink   '),(22188,2824507,'24,June,2004','2004-06-24 00:00:00','    \"Fahrenheit 9/11\" is just as amazing as I thought it would be.    It goes like this.  The Bush Administration seized power by a fraudulent vote in Florida and exploited the September 11 terrorist attacks to pump up fear, tamp down dissent, enrich its cronies and, ultimately, to launch an ill-advised war against Iraq - on the dubious grounds that Saddam Hussein was somehow in league with Al Qaeda.  Michael Moore\'s telling of this narrative is enthralling and always entertaining.  He begins by showing Al Gore\'s victory celebration on Election Night 2000 and wondering \"Were the past four years just a bad dream?\"  We then see details of how Florida was rigged for Bush and his inauguration on a rainy day with huge protests in the streets and eggs being thrown at the motorcade.  Prior to 9/11, Bush is shown as laid back and ineffective.  His vacation for the entire month of August, 2001 is emphasized by the song on the soundtrack : \"Vacation, all I ever wanted.  Vacation, time to get away.\"  (The song remained stuck in my head for the entire film)  Then 9/11 is handled well by showing a black screen with the sounds of the airplanes crashing.  Video is shown of Bush learning of the second plane hitting the tower and then sitting in a Florida classroom for a full seven minutes, with a vacant stare on his face and reading \"My Pet Goat.\"  We then learn how the administration made sure that all members of the Bin Laden family were safely flown out of the country without being detained for questioning.  We also learn about the Bush family\'s ties with Saudi Arabia and the Bin Laden family in their oil business.  Moore raises the point : \"If you earned $400,000 in salary from taxpayers as President, and millions from Saudi Arabian oil interests, who would you be the most loyal to?\"    Moore then explores the Patriot Act (most members of Congress didn\'t have time to read it before voting for it), the color coded terrorist threat warnings (used to induce fear), the insufficient job of defending the homefront (a national parks staffer complains that their budget is nonexistant), and the military\'s emphasis on recruiting among the poorest classes where the military is their best shot of going to college.  Finally Moore gets to Iraq where precious resources needed to hunt Bin Laden in Afghanistan were diverted.  It is demonstrated that the administration pushed for war with Iraq from the beginning, even though the case could be made that Saudi Arabia was more responsible for 9/11.  We see first-hand the devastation suffered by innocent Iraqi civilians.  We also see the death and injuries of many of our own troops.  The kicker is his interviews with the mother of a U.S. soldier.  In the beginning, she supports the war and the administration.  But after her son is killed and she receives his last letter where he writes that Bush sent him there for no reason, she changes her mind and asks what the point of this war was, going as far as traveling to Washington D.C. to stand outside of the White House gates and demand answers.  It\'s heartbreaking and riveting.  Even if Bush feels that he doesn\'t owe anything to Michael Moore, he does owe a damn good explanation to each of the families who lost their loved ones in a war that was fought under false pretenses.  To sum up, everyone must see this movie.  I went with my friend Kerry and some of her friends and, to us, it felt like Moore was preaching to the converted.  What we need to do is to bring Republicans and undecided voters to see this movie, especially those who live in the swing states.  Kerry raised a good point in that some people might be turned off by the partisan manner in which Moore mocks Bush.  But even if they don\'t buy some of Moore\'s points, there are others that will give them serious pause about Bush and his cronies.  On a side note about the theater I saw the movie at.  It was the Loews on 34th Street and 8th Avenue.  It was the first time I\'d been to a theater where they had reserved seating up front, with food menus for the concession stands and ushers who take orders and bring food to your seat!  At the movies!  Unreal.   '),(22189,2824507,'24,June,2004','2004-06-24 00:00:00','    With one week to go until the Iraqi transfer of sovereignty, all hell has broken loose.  This morning alone, there were a wave of coordinated attacks that have resulted in at least 83 people dead and 233 wounded.  Threats have been made on the life of the new Iraqi Prime Minister.  All in all Iâ€™d say that things are going badly.  Not exactly the rosy scenario promised by Cheney and Rumsfeld before the war.  As a matter of fact, Democratic Senator Joseph Biden had a recent meeting with Bush and Cheney.  When asked what he would do about resignations following the Abu Ghraib prison scandal, Biden replied, â€œLook, Mr. President, would I keep Rumsfeld?  Absolutely not.â€  Then he turned to Cheney and said â€œMr. Vice President.  I wouldnâ€™t keep you either.â€  He went on to tell Bush that Cheney and Rumsfeld were â€œdead wrong on every major piece of advice theyâ€™ve given you.â€  What was the reaction from Bush and Cheney?  According to Biden, â€œThey said nothing.  Just sat like big old bullfrogs on a log and looked at me.â€  Maybe theyâ€™re in a bad mood about something.  Iâ€™m a huge tennis fan.  Every year I go to the U.S. Open and at this time of year, I always enjoy watching coverage of Wimbledon.  Except this year, the rains have hit London pretty hard.  This is the fourth day of the tournament and they are still playing 1st round matches.  The big story of this Wimbledon has to be the comeback of Martina Navratilova.  Sheâ€™s 47 years old and the other day she beat a player half her age.  Letâ€™s see how she fares in her second round match.  And speaking of tennis, a movie called â€œWimbledonâ€ is coming out this fall.  It will star Kirsten Dunst and Paul Bettany.  Should be good!  (I hope!)  Last night I was talking with my new friend Caitlyn and we were discussing recent vacations that weâ€™ve been on.  It occurred to me that I havenâ€™t had a vacation where I went somewhere that lasted more than 4 days in YEARS.  Part of the problem is that I usually do not have a traveling companion so Caitlyn suggested that I use this blog as a forum to gauge if anyone else has the traveling itch.  So here it is, as I really need to get out of this city!   '),(22190,2824507,'23,June,2004','2004-06-23 00:00:00','    Al Qaeda beheaded the South Korean.  Bastards.  So I started reading my long-awaited Bill Clinton book last night and I wasnâ€™t five pages in before my eyelids started feeling heavy.  What separates this from say â€œHarry Potterâ€ novels is that in â€œHarry Potterâ€, itâ€™s an adventure where you canâ€™t wait to turn the pages and find out what happens.  In this case, you read about Clintonâ€™s life and say â€œOh yeah, I remember that.â€    Last night, the American Film Institute presented their list of the Top 100 Movie Songs of all time.   urlLink The full list can be found here.   No surprise, is that â€œOver the Rainbowâ€ was #1 and â€œAs Time Goes Byâ€ was #2.  It seems that each time AFI does a Top 100 list, â€œWizard of Ozâ€ and â€œCasablancaâ€ rank at the top.  Some other selections include... 35-  â€œAmericaâ€   (West Side Story) 40-  â€œFight the Powerâ€   (Do the Right Thing) 58-  â€œGonna Fly Nowâ€   (Rocky) 74-  â€œRainbow Connectionâ€   (The Muppet Movie) 80-  â€œSpringtime for Hitlerâ€   (The Producers) 85-  â€œCome What Mayâ€   (Moulin Rouge) 93-  â€œLose Yourselfâ€   (8 Mile)  The Mets are one game over .500!  They are now a winning team!  Enjoy it while you can folks...   '),(22191,2824507,'22,June,2004','2004-06-22 00:00:00','    Guess who was the first person to buy Bill Clintonâ€™s autobiography at the Barnes & Noble in Park Slope?  Me!  I walked by just as they were opening at 9:00 A.M. and made a beeline to the Clinton display (along with about 15 other people).  But I was the first to pay for it!  And it was cheap too.  In addition to it being marked for 30% off, I got an additional 10% for being a member, and then an additional 15% on top of that from a B&N coupon that I got in the mail.  I canâ€™t wait to read about his battle with Voldemort... oh wait, wrong series.  Well in lieu of a Harry Potter book this year, this will probably be the publishing industryâ€™s big hit.  And if it sucks, I can always use it for weight training purposes.  The cartoon on my Dilbert desk calendar :  The scene is a restaurant.  A waitress is taking Dilbertâ€™s order.  Dilbert :  â€œGive me $35 worth of food, including your 15% tip.â€ Waitress :  â€œIf I bring you a penne pasta, will you promise to not build a log cabin on your plate?â€ Dilbert :  â€œI canâ€™t promise that.â€ Waitress :  â€œWell then.  We have a situation here.â€  And this was yesterdayâ€™s cartoon.  Dilbert is sitting at the restaurant table by himself :  Dilbert :  â€œIâ€™m not a loser who canâ€™t get a date.  Iâ€™m a business traveler.â€ Waitress :  â€œWhatâ€™s your girlfriendâ€™s name?  Iâ€™ll call her and check your story.â€ Dilbert :  â€œMaybe I should order.â€ Waitress :  â€œMaybe you should.â€   '),(22192,2824507,'21,June,2004','2004-06-21 00:00:00','    Happy Belated Fathers Day!  A good time was had yesterday as I went out to dinner with my parents, brother, aunt and uncle.  My brother drove up from Baltimore on Friday and went home last night after dinner.  In between meals and movies, much of his time was spent working on fixing my dadâ€™s computer.  It seems that computer problems always arise when he is in town.  Two years ago, my computer crashed and my brother had to install Windows XP on it.  Itâ€™s good to have a tech wizard in the family!    Iâ€™m sickened by the beheadings in the Middle East.  First Nick Berg.  Then Paul Johnson.  Now a South Korean translator has been abducted and will also be beheaded at sundown if South Korea does not withdraw their troops.  Although Berg and Johnson seemed resigned to their fates, the Korean was videotaped bawling in hysterics.  To hear his cries on the radio this morning was devastating.  And now, doing a total 180, how about them Mets?  Winners of four in a row, a three game sweep of the Tigers, two game-winning hits by Mike Cameron, the debut of Richard Hidalgo, the return of Jose Reyes, an even won-loss record and they are only two and a half games out of first place.  Letâ€™s hope they can improve from here!   '),(22193,2824507,'20,June,2004','2004-06-20 00:00:00','    Romantic comedies aren\'t exactly my favorite film genre, but I just saw an amazingly sweet one.  It\'s called \"Love Actually\" and it\'s a great piece of ensemble storytelling.  Each plotline is unique yet compelling in its own way.  There\'s Hugh Grant as the Tony Blair-ish Prime Minister, who falls for one of his aides.  There\'s Emma Thompson married to Alan Rickman, who is tempted by his secretary.  There\'s Liam Neeson who, after his wife passes away, bonds with his 10-year old stepson and advises the boy through his first puppy love.  There\'s Laura Linney who longs for her co-worker yet denies her own needs for those of her mentally disabled brother.  There\'s Keira Knightly who marries one man yet is longed for by the man\'s best friend.  There\'s the aging pop singer who realizes that he\'d rather spend time with his loyal manager than to any of the women he can have.  There\'s the awkward English guy who goes to America to pursue his goal of finding hot American chicks who will sleep with him just because he\'s British.  And there\'s many many other story lines that all weave together very well.  Highly recommended!  Today I took a LONG exam (nearly 4 hours) for the title of \"Staff Analyst.\"  For a couple of weeks prior to the exam, I was studying past questions dealing with flow charts and their crazy symbols (i.e. a diamond is a decision, a rectangle is a process, a triangle is storage).  Fortunately, none of that was even on the test!  I guess they decided to update the test from its 1930s incarnation.  While waiting on line to get into the test, I think I was hit on.  I\'m not sure, I\'m bad about figuring those things out.  But she gave me her business card and wrote her cell phone number on it.  But then she said something about her boyfriend.  I\'ll never understand women!   '),(22194,2824507,'17,June,2004','2004-06-17 00:00:00','    urlLink    Litchfield Villa.  This is the building where I work.&nbsp; urlLink   '),(22195,2824507,'16,June,2004','2004-06-16 00:00:00','    urlLink    The infamous Root Boot!  (Two Boot\'s root beer & vanilla ice cream float)&nbsp; urlLink   '),(22196,2824507,'16,June,2004','2004-06-16 00:00:00','    urlLink    Ah, the safe streets of Brooklyn.  Next to Lauren are are a couple of toddler bicycles LOCKED to a front gate with a black CLUB-like device.  &nbsp; urlLink   '),(22197,2824507,'16,June,2004','2004-06-16 00:00:00','    This time of year is just insane for me.  Never mind all of the ribbon cuttings, park openings and check presentations.  Never mind the increased number of permit requests.  The craziest part of this time of year is that June 30 is creeping up on us.  No, Iâ€™m not talking about the Iraqi transfer of sovereignty.  Iâ€™m talking about the end of the NYC Fiscal Year.  Itâ€™s a mad rush to get all of the paperwork in to document expenses and ensure that they are credited to this yearâ€™s account.  It may not be exciting but itâ€™s sure been keeping me busy!  The 9-11 Commission announced yesterday that there is no evidence that Iraq had anything to do with Al Qaeda or with the terrorist attacks.  Did you hear that, Mr. Cheney?  Iraq had NOTHING TO DO WITH IT.  Yet, the VP still insists it in his speeches and, according to an informal CNN online poll, 31% of Americans still believe it.  Câ€™mon people, pay attention!  Bill Clintonâ€™s autobiography is coming out soon and I cannot wait to purchase it and immerse myself in the happy Clinton years when there were no wars and the economy was strong.  Ah, how I miss the days when the biggest scandals focused on irrelevant land deals and Oval Office BJs.  So much for the Metsâ€™ hitting streak.  Yesterday they went back to their old ways, only scoring one run.  Guess they better fire the new hitting coach!  Office lunch outing to Two Boots today.  Iâ€™m looking forwarding to ordering the Root Boot â€“ itâ€™s a root beer float with vanilla ice cream, served in a mug, shaped like a boot!  Mmm...   '),(22198,2824507,'16,June,2004','2004-06-16 00:00:00','    Itâ€™s too damn hot out.  Yesterday we had the ribbon cutting for the new playground at Brower Park.  The event was a success but was it ever hot.  And today is no different.  At 7:30 this morning, it was already broiling.  Consider me cooked.  The Mets fired their hitting coach yesterday because none of their  players could remember how to hit.  (But when have the Mets ever hit?)  So in the very first game after the firing, the Mets win by scoring 7 runs on 14 hits.  I guess it really WAS the hitting coachâ€™s fault after all!      Itâ€™s time for Summer Reality TV!  Yesterday was the premiere of â€œThe Joe Schmoe Show 2â€ and tonight is the premiere of â€œThe Simple Life 2.â€  Is it mindless?  Yes.  Is it perfect for the summer when mindless rules the day?  Absolutely.  Iâ€™ve made plans to see â€œFahrenheit 911â€ next Friday night, the first day it opens.  That\'s Michael Moore\'s new doc and Iâ€™m excited to see it.  They had the big celebrity screening a few days ago.  Although most of the crowd was liberal, Bill Oâ€™Reilly also showed up.  He left halfway through, which means that it HAS to be good!   '),(22199,2824507,'14,June,2004','2004-06-14 00:00:00','    Haven\'t been able to think of a single thing to say today.  Happy Flag Day?  Nah, that\'s lame.  Although the cartoon guy in my \"weather pixie\" has been replaced by an American flag which for some reason I find to be funny.  Oh that reminds me of something.  Speaking of flags, the Supreme Court ruled today that \"Under God\" should remain in the pledge of allegiance.  Not because of any logic pertaining to the issues at hand.  But because they ruled that the plaintiff was not the proper legal guardian of the girl he was suing on behalf of.  Three cheers for technicalities!   '),(22200,2824507,'13,June,2004','2004-06-13 00:00:00','    Happy Birthday to my Mom!!!  Last night I attended Tim Wan\'s birthday party at the Carriage House on East 59th Street.  It\'s an ideal place for a party - great drinks, great food, fun interactive jukebox, pool table & dart board, not too loud or crowded, and a cute bartender / waitress too!    Thanks to Jess for e-mailing me more Reunion pics!  They have been posted below.   '),(22201,2824507,'12,June,2004','2004-06-12 00:00:00','    urlLink    Holding the Class Banner at the Alumni Parade&nbsp; urlLink   '),(22202,2824507,'12,June,2004','2004-06-12 00:00:00','    urlLink    Me & Ben at the Saturday Dinner&nbsp; urlLink   '),(22203,2824507,'12,June,2004','2004-06-12 00:00:00','    urlLink    Jess & me at the Saturday Dinner&nbsp; urlLink   '),(22204,2824507,'12,June,2004','2004-06-12 00:00:00','    So I watched the tape of Reagan\'s funeral.  Pretty powerful.  My heart goes out to Nancy.  She\'s been a rock of strength this past week.  Although the eulogies by the politicians at the Washington Cathedral were well done (particularly the love shown by George Bush Sr.), I was most impressed by the speeches of Reagan\'s three children at the burial site.  I was most struck by Ron\'s comment, \"My father was a religious man but, unlike many of our politicians, he never wore his religion on his sleeve.  He viewed his convictions as a responsibility but not as a mandate.  And believe me, there is a difference.\"  Wow, that statement seemed to be a subtle jab at Bush.  Between Ron\'s statement and Nancy\'s push for stem cell research, would it be too hopeful to wish for the Reagan family to join the Kerry campaign?  I\'m sure it won\'t happen but the signs are there that they are not in lock-step with the current Republican administration.  And right now, the heart of the nation is with Nancy Reagan.  It has been a busy week at the Parks.  On Wednesday, we held an event at Taaffe Playground in Bedford-Stuyvesant.  Children from the public school across the street, spent the year collecting $1000 in pennies in their Penny Harvest Drive.  The kids then decided that they wanted to use the money to fix up our playground.  So at the event, the Principal presented the Borough Commissioner with a ceremonial check for $1000.  It was a very sweet gesture.  Then on Thursday, another check presentation ceremony was held at Grand Ferry Park in Williamsburg.  This time, we received a check in the amount of $300,000 from the New York Power Authority to fix the park up.  Next on the schedule is a ribbon cutting this Tuesday at Brower Park in Crown Heights.  On the sidebar of this webpage, I have added a feature with links to other friend\'s webpages.  Basically, there are two ways that you can get your page listed there.  You can either link my page to yours or, failing that, you can leave a comment about one of my posts.  It\'s all about reciprocity!   '),(22205,2824507,'11,June,2004','2004-06-11 00:00:00','    In honor of Ronald Reaganâ€™s funeral, today is a national day of mourning.  Federal offices are...closed.  State offices are...closed.  City offices are...open.  Sigh.  So while Mayor Bloomberg goes to Washington to attend the funeral, the rest of us remain behind to do the work of the city.  No worries, I have my VCR set to tape the show.  (Was I just insensitive to have called the funeral a â€œshowâ€?  Yeah, probably.  But thereâ€™s so much pomp and ceremony, itâ€™s hard not to be wowed by it all.)  On top of that, Ray Charles died yesterday.  My favorite Ray Charles song has got to be â€œNight Time is the Right Time.â€  It was performed on â€œThe Cosby Showâ€ by the Huxtable family as they lip-synched it to Cliffâ€™s parents on their anniversary.  The best part is when Rudy belted â€œBAAABBBYYY!!!!!â€  I can watch that one in reruns over and over again and never fail to crack up.   '),(22206,2824507,'10,June,2004','2004-06-10 00:00:00','    Happy Brooklyn Day!  When I was a kid, I used to get this day off from school.  Although Iâ€™m still in Brooklyn, I donâ€™t get it off anymore.  Rats.  I saw â€œHarry Potter and the Prisoner of Azkaban.â€  It was my favorite of all the books and, thus far, it is my favorite movie.  I enjoyed Alfonso Cuaronâ€™s direction.  Rather than show us every single scene from the book (ala the first two films), he just focused on the main themes and showed what was needed to move the story along.  Whatâ€™s impressive is that even though the book was longer than the first two, its film version was shorter than each of the first two.  Besides, my alltime favorite scene in all of the Harry Potter books is the confrontation inside the tree between Sirius Black, Lupin, Pettigrew, Snape, and the kids.  It was as much fun to watch as it was to read.  Now, with â€œGoblet of Fireâ€ schedule to open in November 2005, the films are catching up to the books.  I hope J.K. Rowling can write the last two quickly!   '),(22207,2824507,'09,June,2004','2004-06-09 00:00:00','    I hope everyone has enjoyed the recaps and pictures of my college reunion.  It put me on a serious nostalgia kick that has continued all week, what with keeping in touch with old friends, swapping pics, reading the books I bought about VC, etc.  But life goes on and so must this blog.  So starting tomorrow I will resume posting film reviews, political rants, news from the parks, and lots more!  Now if you\'ll excuse me, I must go back to my book about Vassar\'s transformation to coeducation...   '),(22208,2824507,'08,June,2004','2004-06-08 00:00:00','    urlLink    Me, Casey, Nicole, Graham&nbsp; urlLink   '),(22209,2824507,'07,June,2004','2004-06-07 00:00:00','    urlLink    \"Three Men and a Baby\" (Gabe, me, Jonty & Jonty\'s baby daughter)&nbsp; urlLink   '),(22210,2824507,'07,June,2004','2004-06-07 00:00:00','    urlLink    Me at the Alumni Parade&nbsp; urlLink   '),(22211,2824507,'07,June,2004','2004-06-07 00:00:00','    Vassar Reunion Recap for Saturday, June 5 and Sunday, June 6  My first night sleeping in a dorm room bed in 5 years netted four and a half hours sleep.  Not having an alarm clock, I left the shades open to ensure that the sunlight would wake me up.  I awoke at 8:00 A.M. and then took my first shower in a dorm room bathroom in 5 years.  Jay Briar complained that my blog posting about Friday didnâ€™t have any details on the bathrooms so Iâ€™ll make up for it here by saying it had very good water pressure.  By 9:00, I was ready to go.  Ben was also up so we went to the college store.  On the way, we met some elderly alums and talked about the Japanese elm trees in the quad.  Even though the store had only been open for 15 minutes, it was already very crowded.  I made three purchases.  A Vassar Alumni sweatshirt.  A book about the history and architecture of Vassarâ€™s buildings.  And a book about how Vassar spurned Yaleâ€™s offer to merge and became a co-ed school in the late 1960s and early 1970s.  Those two books will certainly keep me busy for a while.  As I was waiting on line, I stood next to a woman who said that she flew in all the way from Saudi Arabia to attend reunion.  Thatâ€™s amazing.  I left the store and ran into Jon Gans and Emily Jacobs in the retreat.  I wasnâ€™t sure if they would be coming so I was very happy to see them.  (I am going to be in their wedding party next year)  Pictures were taken with them as well as with ADK and Stu in the retreat.  I saw Marguerite Mugge too.  (Not Marguerite Moreau though.  She was supposed to attend but couldnâ€™t at the last minute)  Then it was back to Davison to drop off my new purchases and to get ready for the Alumni Parade!  All the classes lined up for the parade on the path in between the Chapel and Main.  Each class was adorned with banners and balloons in our class colors.  Lots of pictures were taken and I had the honor of holding the Class of 1999 banner, along with Becky.  The parade started with Fran, her dog, a band and then a golf cart driving the lone member of the Class of 1929.  Think about that... 1929.  If she was 20 when she graduated then that would make her 95 today.  Amazing.  More golf carts followed with members of the Class of 1934 and 1939.  Then other classes followed, marching behind.  Standing on the sidelines, all of us cheered for them and they in turn cheered for us.  Many of them looked at us and said â€œYouâ€™re so young!â€ or â€œLook, thereâ€™s men!â€  Donâ€™t forget, Vassar was originally an all-womens college so it was great to see all of the alumni from that era.  The Classes of 1948 and 1949 marched together because the classes were very small, as World War II prevented many women from leaving home.  One of the alums that I spoke with earlier in the quad went over to me and we gave each other a high five.  Then we saw the Class of 1974 with a banner proclaiming that they are Vassarâ€™s first fully co-ed class.  As we cheered for all of the classes ahead of us, it became our turn to march.  But since we were the tail end of the parade, there was no one left to cheer for us!  Thankfully, Graham Campbell and Pedro Amador broke away from the group and ran along the sidelines yelling â€œYay! 99!â€  The parade made its way to the new athletic building next to Walker.  Inside there is a new gymnasium that was set up for a rally.  Our class was assigned the side bleachers.  Two huge video screens were set up next to a stage.  What followed was a presentation announcing how much each class gave to the college.  The Class of 99 gave the most money for a 5-Year Class in 28 years, as our gift was around $17,000.  Thanks to Annie Thottam and Carolyn Brooks for their hard fundraising efforts.  The 50th Reunion Year class gave over $2 million!  And the total for all classes was over $9 million.  Amazing.  After those announcements we were shown a very sentimental video combining footage from Vassar at the turn of the century, with that of the 1950s, and with that of the present day.  Everyone got a little emotional during it; a testimony to how well the institution is loved by its alumni.  Many people, myself included, said during the weekend that we would love to drop everything and do another 4 years all over again.  Finally, the rally concluded with a singing of â€œGaudeamusâ€, the traditional scholarâ€™s song.  After that event they were giving out box lunches, but I wasnâ€™t interested so I went back to the Retreat and consumed a roast beef sandwich, just like they used to make it.  Then it started to drizzle and, since there was some downtime, I decided to check out some of the new renovations.  I went into Noyes, my home for three years, to see the renovated Jetson Lounge.  I also checked out all of my old rooms to complete the nostalgia trip.  Then I went into Jewett, which was completely gutted and beautifully redone.  It doesnâ€™t even resemble a dorm anymore; itâ€™s like a hotel.  Melissa Walker said that her old room is now a stairwell.  Andrij Dobriansky said that his old room is now a study hall.  The halls, walls, windows, furniture are all impeccable and if I lived there Iâ€™d be afraid to touch anything!  I wonder if this is their idea in renovating all of the dorms.  After my self-guided tour I went back to Davison to catch up on some rest.  I tried to nap for a half-hour but who can nap during reunion?  So I chilled on the front steps of Davison, talking to Matt Newton, and waving to all of the late arrivals including Jill Cadre.  Ben soon came by and to end our boredom, along came a golf cart being driven by a reunion staffer.  She asked if anyone needed a ride so Ben and I said â€œSure!â€  She told us that she had 20 minutes left in her shift and would drive us around campus for that long.  Turns out that our host was Emma Angevineâ€™s younger sister from the Class of 04!  So we sped through campus and checked out the new South Commons facilities.  Along the way, we saw older alums slowly making their way around campus.  When we asked them if we could take them anywhere, most declined.  I think they wanted to prove to themselves that they could still walk around campus, unassisted.  Finally some alums did want a ride to the new Center for Drama and Film so Ben and I were kicked out of the cart.  While that was happening, my favorite political science professor, Richard Born, was giving a lecture.  I did not attend but Jon Gans e-mailed me to fill me in on what happened.  Jon wrote, â€œSo Born\'s talk was pretty good and wicked funny.  The same old antics, clothes, and notes.  He was more enthusiastic than I remember... He was like :  â€˜I wanted to call this talk Delay does Dallas (pause) and Houston and Fort Worth..., but I didn\'t think they would print it in the program.â€™  He told Mike Riggs and I to forget all the nonsense he said in class about ticket-splitting, the decline of political parties, etc.  He said he was wrong.â€  So basically, my political mentor just openly admitted that everything he taught us was wrong!  Canâ€™t beat that one.  After that class, Austin Brown invited Mr. Born to attend the Class of 99 cocktail reception in Raymond.  Bornâ€™s response was â€œWell, I donâ€™t know if Iâ€™m really supposed to go.â€  Austin convinced him though!  More on that later.  Anyway, after Ben and I were kicked off the golf cart, I made my way back to my room to change for the Presidentâ€™s reception for donor societies.  I was invited since my gift contribution just barely qualified me as a Cornaro Fellow.  So I put on some nicer clothes and made my way over to Franâ€™s house.  She had a big tent set up on her lawn and the first person I recognized was Ken Wong, from the Class of 2000.  Ken said that even though it wasnâ€™t his reunion he decided to come up to campus for Franâ€™s donor reception.  ADK and Stu then made a beeline in my direction as there werenâ€™t too many 99ers there so we have to stick together.  A gin and tonic and a red wine were consumed.  We waited on a receiving line to pay our respects to Fran and I had a nice little conversation with her.  She said how much she appreciated the large turnout of our class.  (Almost half attended)  I mentioned how well organized the parade and rally was and how nice it is for the different generations to have the opportunity to interact.  Fran then encouraged us to go inside of her house and look around.  So we did, admiring her artwork and books.  In one room we saw her TV and personal video collection.  Amusingly, Adam Sandlerâ€™s â€œThe Wedding Singerâ€ was included.  I then went back into the reception and met up with Jill and Annie Thottam.  We hung out for a little while and then made our way over to Main.  Each parlor on the second floor of Main was being occupied by various small receptions.  I peeked into the Faculty Parlor and saw some people I recognized so I went inside and started saying hello to people.  I then asked one person what this reception was for and he said that it was the Gay and Lesbian Alumni Association.  To which I said, â€œCool!  Iâ€™ll see you later then!â€  Heh heh, canâ€™t have any rumors going around now!  Jill, Annie and I then visited the reception for the Latino/Latina Alumni Association.  We caught the very end of some speeches and then partook in some festive fruit.  At that point I made my way up to the 3rd Floor of Main to see the room that I lived in during my senior year.  I also checked my e-mail at the computer cluster and that was when I found out that Ronald Reagan had passed away.  It would become a big dinnertime conversation topic that night.  Back to Davison to watch the Belmont Stakes.  The room was packed with members of our class and the Class of â€™89.  I still canâ€™t believe that Smarty Jones lost.  He had it!  During the race I chatted with former Davison residents, Megan Marks and Adam Perrelli.  Megan was quite devastated when Smarty lost.  Many were.  Then it was over to Raymond for the Class of â€™99 reception.  All I can say about that one is â€“ nothing says Vassar better than red wine in a paper cup.  The room was packed and it was a bit difficult to move around.  I did see Mr. Born though as he accepted Austinâ€™s invitation and attended.  We had a nice conversation as he remembered where I was working and whom I used to work for.  Austin then invited him to go golfing with us at the 10-Year Reunion.  After that it was on to the Saturday night dinner.  Each class had a tent set up outside of the dorms.  Ours was on the lawn in between Chicago Hall and Joss.  This was a sit-down dinner and they checked for tickets this time.  (Good thing too since we spent $35 on em)  The dinner consisted of a choice of chicken, tuna or pasta, with potatoes & veggies, a salad, cheesecake, and yes â€“ red or white wine.  So that marked my third red wine in three consecutive receptions.  Reunion Chair Michelle Laredo got on the microphone and got cheers when she said â€œDonâ€™t worry.  I wonâ€™t be asking you for any money.â€  Becky Corona then got on the mic and acknowledged the efforts of her fellow class officers - Marivel (applause), Jill (applause), and me (the room ERUPTED in applause).  I donâ€™t mean to sound cocky â€“ on the contrary I was very humbled by it.  I couldnâ€™t get over the overwhelming reception I got from my classmates and Iâ€™m very thankful for it.  Becky then said that we are accepting names for new officers and anyone who is interested should contact her.  I suppose weâ€™ll have an election via the mail or something.  As for me though, I am officially retired!  After dinner, there was yet another fireworks display.  This one emanated from off-campus.  Coincidentally there was a street fair going on in Poughkeepsie.  After lingering in the tent for a while, a bunch of us made our way back to the Davison MPR to relax.  Gabe, Jonty and I ordered chicken parms from Naps, thus fulfilling my eating goals for the weekend.  We also watched the end of the NHL game and the news coverage about Reagan.  The night died down after that.  Some people went back to the Acrop (I couldnâ€™t even look at food anymore) so I visited the Mug for literally five minutes before getting sick of it.  So it was back to the dorm to hang out until 2:30 A.M. at which time I called it a night.  Slept for seven and half hours uninterrupted.  Woke up at 10 and at 11 made my way over to ACDC for the brunch.  It was so surreal to be walking to ACDC again with a group of people.  Serious deja vu.  The selection of brunch foods wasnâ€™t as impressive.  Stephanie went into the main dining area to get some cereal and an employee bitched her out because it wasnâ€™t â€œpart of the brunch.â€  Stephanie said, â€œI paid $15 and you wonâ€™t let me eat CEREAL?â€  The worker conceded but that was just ridiculous.  Anyway the brunch was a somber affair as everyone was saying goodbye to each other.  Before I left though, I checked out the new second floor of ACDC which is now the new Students Center.  Very nicely designed.  The weekend then ended as Ben drove Jess and me back to the train station.  Jess and I finished our trip the way we started â€“ on the Metro North.    All in all, a great weekend.  Thanks for reading this very long recap!   '),(22212,2824507,'06,June,2004','2004-06-06 00:00:00','    urlLink    Hanging out in the Davison MPR.  Thanks to Gabe for providing these photos.   urlLink For more, see his blog! &nbsp; urlLink   '),(22213,2824507,'06,June,2004','2004-06-06 00:00:00','    urlLink    Watching the Friday Night Fireworks&nbsp; urlLink   '),(22214,2824507,'06,June,2004','2004-06-06 00:00:00','    urlLink    Walking to the Friday Night Dinner&nbsp; urlLink   '),(22215,2824507,'06,June,2004','2004-06-06 00:00:00','    urlLink    Friday Night Misc Reunion (Left to Right : Shrey, Ben, Stu, ADK, me, Melissa, Stephanie, Jen, Stephen, Sally-Anne&nbsp; urlLink   '),(22216,2824507,'06,June,2004','2004-06-06 00:00:00','    And now, a recap of my 5-Year Vassar Reunion!  Due to its length, this first post will focus on Friday, June 4 with other posts coming soon.  It started on Friday when I met up with Jess Roth in Grand Central Station.  While waiting for our train we consumed soup in the food court (matzo ball for me) and we saw our first Vassar people, Kate Lane and Jerry Decker â€“ both of whom were also eating!  We took the 3:12 train to Poughkeepsie and arrived a little bit after 5:00.  Upon leaving the Po-town train station, we were greeted by the standard posse of Jamaican cabbies who yell â€œVassar?  Vassar?â€  One very aggressive cabbie got nine of us, including me, Jess, Jerry, Kate, Dave Giles and Tom Pyun, to cram into his van.  I barely even fit as instead of sitting I was kneeling in between the seat and the door for the entire ride.  But thatâ€™s all part of the Poughkeepsie Cab experience.    We were dropped off near North Lot and our group proceeded to wheel our respective luggage toward the registration in the Aula.  On the way there I saw a new fountain and garden and said out loud, â€œYo, what the hell is this new thing?â€  Then when I looked closer and saw the plaque inscription â€œDedicated to the victims of September 11, 2001â€, I felt very foolish and said, â€œOh.â€  Once arriving at the Aula, I registered and got my welcoming packet including keys to Davison 327.  Several items were for sale including one clever t-shirt that said â€œVassar Football.  Undefeated since 1861.â€  I didnâ€™t buy it though.  I called Ben Horst on my cell phone and he informed me that he was hanging out in the College Center Circle so I said â€œSee ya laterâ€ to Jess and the others and met up with Ben.  We then went over to Davison so I could drop my bag off in the room.  On the way we saw â€œEasy Eâ€ Erich Walter, Pedro Amador and some others kicking around a soccer ball in the quad.  Everyone was already feeling very relaxed and comfortable!  On to Davison it was and I was quite impressed with the size of 327.  I learned later on that itâ€™s one of the prime singles in the dorm and is usually taken by a House Officer.  I donâ€™t know who put the housing assignments together but they obviously know that I rate!  Even better was that Ben was right across the hall from me.    We then went down to the Davison multi-purpose room where a welcoming reception was being held for our class.  We were given t-shirts, steins, and beer.  There would be plenty of more beer to be had over the weekend.  And wine.  And mixed drinks.  Itâ€™s pretty obvious that they were trying to get us all drunk â€“ possibly so weâ€™d give them more money!  Among the people I talked with at the reception were Michelle Laredo-Torres, her husband Efrain, Becky Corona, Marivel Oropeza, Nikki Ruth, Allison Harris, Colleen Fox, Annabelle the desk messenger and our friendly student assistant.  Thatâ€™s also when Gabe and Jen Anderson made their arrival!  There were many many other people I talked with there too but I canâ€™t be expected to remember them all.  By now it was approaching 7:00 and it was time to head off to the Friday Dinner at the Walker Field House Tents.  I walked to Walker with Ben, Gabe, Jen, Shrey, Jonty Yamisha and Jontyâ€™s wife Dina and 5-month old daughter Maydene.  The baby is so cute and so well behaved!  They took her to many of the events and she was always smiling and having a good time.  Anyway once at Walker, we sat at various round tables.  Ben and I joined Brandee and Rouslan at a table.  I also spent some time sitting with Jess, Erin Condy and Carrie Schultz at another table. And with Tara Knowland and Noel Baxter (too bad that Noel\'s husband, Joe Bonk, couldn\'t make it - he was my student fellow!) Table hopping is ultimately called for in such occasions.  I also had an animated conversation with Cody Kushner about baseball!  It was a great buffet dinner including unlimited drinks and unlimited Vassar Devils.  Those are the great desserts consisting of chocolate cake underneath vanilla ice cream, hot chocolate syrup and many toppings.  The only problem was that we spent 20 bucks on our tickets and noone even bothered to check them!  Easy E didnâ€™t come to the dinner because he didnâ€™t have a ticket.  When I told him that they didnâ€™t even check, his response was, â€œWhat a gyp!â€  (FYI â€“ they made a point to definitely check tickets at the Saturday dinner)  After the dinner there was a cool fireworks display and we then made our way to the College Center.  In the College Center, Sally-Anne Moringello reserved room 237 for a smaller reunion for members of the Miscelleny News Ed Board.  I was never on Ed Board so I crashed!  Our group was then joined by Stephen Gribbin, ADK, Stu, Stephanie Litos, Melissa Walker and Tye Wolfe.  What great fun we had looking at current Misc issues and picking out all the typos and horridly written articles.  They even spelled the word â€œStudentsâ€ wrong in one of their big-fonted headlines.  Sigh...  Apparently the Misc people used to take lots of votes in their meetings and Litos was saying that we needed to vote on something.  So we all voted that Melissa would be one of the new class notes correspondents, as the terms of me and Litos are now over and done with.  We also drank beer and consumed cookies with vanilla frosting that were generously provided by S-A.  Stu made a point to remind me to eat lots of cookies.    While in the College Center, I checked my e-mail on one of the many iMacs they have on the second floor and was pleasantly surprised to see that someone made VassarChat.com the default web setting. Between that and all of the VassarChat flyers that were circulated, the whole campus is bound to know about us soon!  From there it was on to the Mug!  Let me tell you something.  It hasnâ€™t changed.  Itâ€™s dark, loud, crowded and the floors are as sticky as ever.  The sight of a bunch of alums sitting on the Mug Line was hilarious.  Although the DJ played some newer songs like â€œHey Ya!â€ and my personal favorite, â€œIâ€™m a Hustlerâ€, we also heard many of the same stuff that was played from 1995-1999.  With all the noise itâ€™s also the perfect place to have a conversation : â€œHowâ€™s it going?â€  â€œWhat?â€  â€œHow are you?â€ â€œWHAT?â€  Despite not being able to hear very much I think I had some conversations with Austin Brown, Hayden Courtland, Graham Campbell, Amar Reganti, Nicole Zahka and others.  I had thought it would be all 99ers in the Mug but there were many from 94, 89, 84, there was even one guy from the Class of 74 with a big white bushy beard and he made a real beeline to the bar!  Great to see all of the diversity.  Itâ€™s now 1:30 A.M. and hunger is setting in.  I had two eating goals for Reunion weekend.  One was to go to the Acropolis Diner.  The second was to order a chicken parm from Naps.  The first was done on this night as Ben, Gabe, Jen and I set off for the Acrop.  We walked there and I didnâ€™t remember it being such a big walk!  Old age setting in.  Sitting right across from us were Erin, Matt Wilson and Mattâ€™s fiancÃ©e Stephanie.  So yeah, I table hopped again.  Since it was technically morning, I decided to have breakfast â€“ a scrambled egg with bacon!  The best part was when our waitress mocked the owner (her boss) right in front of him.  I wonder if she still had a job the next day.    After eating off the alcohol that was consumed, it was back to Davison.  Itâ€™s now 2:30 A.M.  By that time most of the people I knew had gone to sleep.  I walked around for a while looking for some action and sure enough I found it in the shape of a loud party on the 4th Floor of Davison.  But I didnâ€™t stay too long and retired to my room and to bed at 3:30 A.M.  After all, there would be a very big day tomorrow!  Stay tuned for the recap of what happened on Saturday, June 5.   '),(22217,2824507,'06,June,2004','2004-06-06 00:00:00','    urlLink \"Avenue Q\"  just won the Best Musical Tony Award!  Hell, yeah!!!  The show is awesome.  Get thee to New York ASAP and see it!   '),(22218,2824507,'03,June,2004','2004-06-03 00:00:00','    Guess where I\'m typing this right now?  At an iMac in the Vassar College Center!  Reunion is so amazing.  The past two days have been incredible and I am going to write all about it on this blog very soon.  But first some thoughts on non-reunion related matters.  Ronald Reagan passed away this morning.  After ten years of being ravaged by Altzheimers, I\'m glad that he is no longer suffering.  Although I am no fan of his presidency or his policies, his passing is nonetheless a sad moment for America and I extend my sympathies to Nancy.  Speaking of Nancy, I have to give her all the credit in the world for voicing her support of stem-cell research.  The Bush Administration\'s opposition to it makes no sense - yet another case where policy has been hijacked by religious ideology.  But Nancy realizes that the research opens a world of opportunities - who knows, maybe it can cure Altzheimers.  Hopefully in the coming media frenzy over Reagan\'s death and funeral, Nancy will stress the importance of this matter.  Yet again, a horse wins the Kentucky Derby, wins the Preakness, but chokes in the Belmont Stakes.  This time the choker was Smarty Jones.  The worst part was Smarty had the race won!  A four-leg lead that all of a sudden vanishes from a horse that came from nowhere to surpass it.  And the hopes of all horse racing fans are dashed...again.  Don\'t worry, a full detailed reunion recap is coming to this blog very soon!  Stay tuned.   '),(22219,2824507,'03,June,2004','2004-06-03 00:00:00','    I canâ€™t get over how perfect the weather is (as everyone can see from my Weather Pixie!)  I could be outdoors all day long with weather like this.  As it was, I had a convenient excuse to be in Central Park this morning and was sure to take some time to appreciate its beauty.  And I can I tell you how mobbed the zoo was with classes on field trips?  It was pretty mobbed.  George Tenet resigned today as Director of the C.I.A.  The reasoning behind it?  According to Tenet, â€œIt was a personal decision, and had only one basis in fact: the well being of my wonderful family, nothing more and nothing less.\"  Riiiggghhhttt...  Anytime an elected official or a highly appointed official resigns for â€œfamily reasonsâ€, itâ€™s all a load of crap.  Itâ€™s a code for being shown the door.  For Tenet, the intelligence failures related to 9-11, WMD and many others, have finally caught up with him.  Now can Rumsfeld resign next?  Who says that spelling bees are not a dangerous sport?  Earlier today, a 13-year old boy appeared to faint and collapsed on stage, drawing gasps from the audience.  After a few seconds, he gathered himself, stood up and immediately started spelling â€œalopecoid.â€  He got it right and then received immediate medical attention.  This might be the last blog update for a few days.  Iâ€™m off to Poughkeepsie tomorrow and most likely will not have access to a computer.  Even if I did, I donâ€™t think blogging would be on my mind!  But I will return next week with a full recap on what happened at the Vassar Reunion!   '),(22220,2824507,'02,June,2004','2004-06-02 00:00:00','    Saw â€œFiddler on the Roofâ€ last night at the Minskoff Theater.  Very good show in a comfortable theater.  Rarely do I see a show on Broadway that I am SO familiar with, inside and out.  I was literally mouthing the words to myself as they sang them.  Alfred Molina played Tevye and, although he was good, I donâ€™t think he measured up to Topolâ€™s performance in the movie.  It seemed that Molina was trying his best not to mimic Topol and to make the character his own, but I kept thinking after each number that Topol did a better job.  Oh well, he spoiled me.  The actor who played the tailor Motel Kamzoil was fantastic though with impeccable comic timing.  He was appropriately nominated for a Best Supporting Actor Tony.  The dream sequence was very well staged with lots of cool special effects including actors descending from the rafters.  The actresses who played Tevyeâ€™s daughters all did a good job.  I never fail to tear up during â€œFar From the Home I Love.â€  And there was a new musical number in the second act that was not included in the original production or the movie.  It is called â€œTopsy Turvyâ€ and has Yente and a couple of the other women lamenting the changes.  The final change came at the end of the show.  Instead of the fiddler being the last one to walk off the stage, he passed the fiddle off to a young boy and they walked off together, symbolizing that as they move to a new world, the traditions are being passed on to the next generation.  Got an e-mail from AAVC last night saying that more members of the Class of 99 have registered for reunion than any other 5-Year class in their history!  Should be an amazing turnout.  However, due to those numbers, our class will be housed in Davison and not Main as originally planned.  I am very familiar with Davison though as I spent a lot of time there during my freshman year.  Only two days to go!   '),(22221,2824507,'01,June,2004','2004-06-01 00:00:00','    Is it June already?  Dang, 2004 is flying by.  Itâ€™s a three-day work week, thanks to the holiday yesterday and my going away on Friday.  I love short weeks.  Iâ€™m thinking of various different ways to improve this website.  In addition to the picture on the right, you can now see a Weather Pixie, detailing the latest weather conditions at JFK International Airport.  If anyone knows of any other fun do-hickeys that I can add to this site, please let me know!  Sometime today I have to write my final Class Notes column for the Vassar Quarterly.  I inherited the job a few years ago from a former correspondent who reads this blog (long story!)  Itâ€™s been a fun job with minimal work â€“ only about an hour, twice a year â€“ but even still, Iâ€™m happy to be done with it.  The novelty of writing the column wears off after a couple of years.  I hereby pass the torch to whoever else wants to do it.   Tonight I am seeing â€œFiddler on the Roof.â€  Although I have seen the show many, many, many times, this is the first version that Iâ€™m seeing on Broadway.  Thanks to Jessica for providing free tickets!  A review will be forthcoming tomorrow.   '),(22222,2824507,'31,July,2004','2004-07-31 00:00:00','    I saw They Might Be Giants perform a free concert last night at Prospect Park.  It was so much fun.  It was also a big Blogger party as I was joined by  urlLink Sarah ,  urlLink Caren ,  urlLink The Anonymous Blogger \'s substitute,  urlLink Claire ,  urlLink Daryl  and  urlLink Sydney .  Those who joined us that didn\'t have blogs were Anonymous\'s friend \"Daisy\", Daryl\'s friend Annie, and my friend & co-worker Eryn, with her friend Cary.  A great time was had as TMBG is always entertaining.  Most of their best songs and biggest crowd reactions came from the \"Flood\" CD, which reminded me of the Simpsons episode where Homer attends a BTO concert and screams \"Play \'Taking Care of Business\'!  Nothing from your new album.  No new crap!  Just \'Taking Care of Business\'.  And skip to the chorus!\"  Some pics from last night are posted below...   '),(22223,2824507,'30,July,2004','2004-07-30 00:00:00','    urlLink    Eryn & I @ TMBG Concert&nbsp; urlLink   '),(22224,2824507,'30,July,2004','2004-07-30 00:00:00','    urlLink    Caren, Sarah & Claire @ TMBG Concert&nbsp; urlLink   '),(22225,2824507,'30,July,2004','2004-07-30 00:00:00','    What a convention.&nbsp; Four nights of rousing and inspired speeches.&nbsp; Over the weekend I will chime in with my thoughts on the week but for now...  A recap of Thursday nightâ€™s speeches at the Democratic National Convention    John Kerry   â€œMy name is John Kerry and I am reporting for duty... â€œWe have it in our power to change the world again. But only if we\'re true to our ideals â€“ and that starts by telling the truth to the American people. That is my first pledge to you tonight. As President, I will restore trust and credibility to the White House... â€œNow I know there are those who criticize me for seeing complexities â€“ and I do â€“ because some issues just aren\'t all that simple. Saying there are weapons of mass destruction in Iraq doesn\'t make it so. Saying we can fight a war on the cheap doesnâ€™t make it so. And proclaiming mission accomplished certainly doesn\'t make it so. As President, I will ask hard questions and demand hard evidence. I will immediately reform the intelligence system â€“ so policy is guided by facts, and facts are never distorted by politics. And as President, I will bring back this nation\'s time-honored tradition: the United States of America never goes to war because we want to, we only go to war because we have to...&nbsp;  â€œAnd on my first day in office, I will send a message to every man and woman in our armed forces: You will never be asked to fight a war without a plan to win the peace. I know what we have to do in Iraq. We need a President who has the credibility to bring our allies to our side and share the burden, reduce the cost to American taxpayers, and reduce the risk to American soldiers. That\'s the right way to get the job done and bring our troops home... â€œI defended this country as a young man and I will defend it as President. Let there be no mistake: I will never hesitate to use force when it is required. Any attack will be met with a swift and certain response. I will never give any nation or international institution a veto over our national security. And I will build a stronger American military...  â€œAnd tonight, we have an important message for those who question the patriotism of Americans who offer a better direction for our country. Before wrapping themselves in the flag and shutting their eyes and ears to the truth, they should remember what America is really all about. They should remember the great idea of freedom for which so many have given their lives. Our purpose now is to reclaim democracy itself. We are here to affirm that when Americans stand up and speak their minds and say America can do better, that is not a challenge to patriotism; it is the heart and soul of patriotism... That flag doesn\'t belong to any president. It doesn\'t belong to any ideology and it doesn\'t belong to any political party. It belongs to all the American people... â€œFor four years, we\'ve heard a lot of talk about values. But values spoken without actions taken are just slogans. Values are not just words. They\'re what we live by. They\'re about the causes we champion and the people we fight for. And it is time for those who talk about family values to start valuing families. You don\'t value families by kicking kids out of after school programs and taking cops off our streets, so that Enron can get another tax break. We believe in the family value of caring for our children and protecting the neighborhoods where they walk and play. And that is the choice in this election. You don\'t value families by denying real prescription drug coverage to seniors, so big drug companies can get another windfall. We believe in the family value expressed in one of the oldest Commandments: \"Honor thy father and thy mother.\" As President, I will not privatize Social Security. I will not cut benefits. And together, we will make sure that senior citizens never have to cut their pills in half because they can\'t afford life-saving medicine. And that is the choice in this election. You don\'t value families if you force them to take up a collection to buy body armor for a son or daughter in the service, if you deny veterans health care, or if you tell middle class families to wait for a tax cut, so that the wealthiest among us can get even more. We believe in the value of doing what\'s right for everyone in the American family. And that is the choice in this election... â€œAnd let me tell you what we won\'t do: we won\'t raise taxes on the middle class. You\'ve heard a lot of false charges about this in recent months. So let me say straight out what I will do as President: I will cut middle class taxes. I will reduce the tax burden on small business. And I will roll back the tax cuts for the wealthiest individuals who make over $200,000 a year, so we can invest in job creation, health care and education... â€œThe story of people struggling for health care is the story of so many Americans. But you know what, it\'s not the story of senators and members of Congress. Because we give ourselves great health care and you get the bill. Well, I\'m here to say, your family\'s health care is just as important as any politician\'s in Washington, D.C. And when I\'m President, America will stop being the only advanced nation in the world which fails to understand that health care is not a privilege for the wealthy, the connected, and the elected â€“ it is a right for all Americans... â€œI\'ve told you about our plans for the economy, for education, for health care, for energy independence. I want you to know more about them. So now I\'m going to say something that Franklin Roosevelt could never have said in his acceptance speech: go to johnkerry.com... â€œI want to address these next words directly to President George W. Bush: In the weeks ahead, let\'s be optimists, not just opponents. Let\'s build unity in the American family, not angry division. Let\'s honor this nation\'s diversity; let\'s respect one another; and let\'s never misuse for political purposes the most precious document in American history, the Constitution of the United States... â€œAnd let me say it plainly: in that cause, and in this campaign, we welcome people of faith. America is not us and them. I think of what Ron Reagan said of his father a few weeks ago, and I want to say this to you tonight: I don\'t wear my own faith on my sleeve. But faith has given me values and hope to live by, from Vietnam to this day, from Sunday to Sunday. I don\'t want to claim that God is on our side. As Abraham Lincoln told us, I want to pray humbly that we are on God\'s side... â€œAnd now it\'s our time to ask: What if? What if we find a breakthrough to cure Parkinson\'s, diabetes, Alzheimer\'s and AIDS? What if we have a president who believes in science, so we can unleash the wonders of discovery like stem cell research to treat illness and save millions of lives?... â€œI learned a lot about these values on that gunboat patrolling the Mekong Delta with young Americans who came from places as different as Iowa and Oregon, Arkansas, Florida and California. No one cared where we went to school. No one cared about our race or our backgrounds. We were literally all in the same boat. We looked out, one for the other â€“ and we still do. That is the kind of America I will lead as President â€“ an America where we are all in the same boat.â€     Max Cleland   â€œI ultimately realized that although I had lost a lot, I still had a lot left, and I resolved to make something of my life. I decided to run for the State Senate in Georgia. I won, but when I got there, in 1971, I was a lone voice. Then I heard this young veteran on TV speaking about the war. It was John Kerry. He put everything I was feeling into words. Tonight, Iâ€™d like to let you know, that even before I met John Kerry, he was my brother. Even before I knew John Kerry, he was my friend. Even before I spoke with John Kerry, he gave me hope...  â€œAs I look back over the last 36 years, I now realize John Kerryâ€™s service to his country didnâ€™t end in Vietnam. It began there. Since Vietnam, John Kerryâ€™s life has become an object lesson in what was once described as the true definition of patriotism â€œthe long and steady dedication of a lifetime.â€... â€œMy fellow Americans John Kerry has never let me down. Heâ€™ll never let you down. He is an authentic American hero. He is the next captain of our ship of state. And he will be the next president of the United States.â€     Wesley Clark   â€œWar. I\'ve been there. Heard the thump of enemy mortars. Seen the tracers fly. Bled on the battlefield. Recovered in hospitals. Received and obeyed orders. Sent men and women into battle. Awarded medals, comforted families, attended funerals. And this soldier has news for you: Anyone who tells you that one political party has a monopoly on the best defense of our nation is committing a fraud on the American people. Franklin Roosevelt said it best: \"Repetition does not transform a lie into a truth.\"...  â€œThis hall and this party are filled with veterans who have served under this flag - our flag. We rose and stood reveille to this flag. We saluted this flag. We fought for this flag. And we\'ve seen brave men and women buried under this flag. This flag is ours! And nobody will take it away from us... â€œUnder John Kerry we will attack and destroy the terrorist threat to America. He\'ll join the pantheon of great wartime Democrats. Great Democrats like Woodrow Wilson, who led us to victory in World War I. Great Democrats like Franklin Roosevelt and Harry Truman, who turned back the tide of fascism to win World War II. Great Democrats like John Kennedy, who stood firm and steered us safely through the Cuban Missile Crisis. And great Democrats like Bill Clinton, who confronted ethnic cleansing in Yugoslavia, and with diplomacy - backed by force - brought peace to a shattered land. My fellow Americans, Democrats are leaders. Democrats are fighters. John Kerry is a leader and fighter and he will be a great Commander-in-Chief.â€&nbsp;   '),(22226,2824507,'29,July,2004','2004-07-29 00:00:00','    A recap of Wednesday night\'s speeches at the Democratic National Convention    John Edwards   â€œDecisive. Strong. Arenâ€™t these the traits you want in a Commander in Chief? We hear a lot of talk about values. Where I come from, you donâ€™t judge someoneâ€™s values based on how they use that word in a political ad. You judge their values based upon what theyâ€™ve spent their life doing. So when a man volunteers to serve his country, and puts his life on the line for others, &shy;thatâ€™s a man who represents real American values. This is a man who is prepared to keep the American people safe and to make America stronger at home and respected in the world... â€œThe truth is, we still live in two different Americas: one for people who have lived the American Dream and donâ€™t have to worry, and another for most Americans who work hard and still struggle to make ends meet. It doesnâ€™t have to be that way. We can build one America...  â€œI have heard some discussions and debates about where, and in front of what audiences we should talk about race, equality, and civil rights. Well, I have an answer to that question. Everywhere. Everywhere.&nbsp; Everywhere. This is not an African-American issue, not a Latino issue, not an Asian-American issue, this is an American issue. Itâ€™s about who we are, what our values are, what kind of country we want to live in...  â€œAnd we will have one clear unmistakable message for al Qaida and the rest of these terrorists. You cannot run. You cannot hide. And we will destroy you...  â€œSo when you return home, you might pass a mother on her way to work the late-shift &shy;you tell herâ€¦â€¦hope is on the way. When your brother calls and says that heâ€™s working all the time at the office and still canâ€™t get ahead&shy; - you tell himâ€¦â€¦hope is on the way. When your parents call and tell you their medical bills are through the roof - you tell themâ€¦...hope is on the way. When your neighbor calls you and says that her daughter has worked hard and wants to go to college - &shy;you tell herâ€¦â€¦hope is on the way. When you talk to your son or daughter who is serving this country and protecting our freedoms in Iraq&shy; - you tell themâ€¦â€¦hope is on the way. And when you wake up and sit with your kids at the kitchen table, talking to them about the great possibilities in America, you make sure that they know that John and I believe at our core that tomorrow can be better than todayâ€    Elizabeth Edwards   â€œThis has been quite a year for John and for me. We started last January crisscrossing America, talking about how, with determination and vision and optimism, we can end the injustice of two Americas. And this January, it\'s going to end when we move two great friends, two great Americans, John Kerry and Teresa Heinz Kerry, into 1600 Pennsylvania Avenue...  â€œYou know, I married the smartest, toughest, sweetest man I know. And in two days, we will celebrate 27 years of marriage the way we always do. We\'ll do it the way we always do at Wendy\'s. Whether it\'s Wendy\'s or Washington, I\'ve found that it\'s true: It\'s not where you go, it\'s who you go with... â€œBut none of the things I\'ve mentioned are the reason that I married John Edwards. I married him because he was the single-most optimistic person that I have ever known. He knew there was a brighter day ahead, even as he swept the floors in the cotton mill as a high school student. He knew if he worked hard enough, he could be the first in his family to go to college. He knew that he could outwork and out-tough any battalion of lawyers to find justice. And he continued that fight in Washington, courageously, eloquently, with one simple goal: to make the opportunities of America available to all Americansâ€    Al Sharpton   â€œI\'m convinced that at a time when a vicious spirit in the body politic of this country that attempts to undermine America\'s freedoms -- our civil rights, and civil liberties -- we must leave this city and go forth and organize this nation for victory for our party and John Kerry and John Edwards in November... â€œMr. President, I heard you say Friday that you had questions for voters, particularly African- American voters. And you asked the question: Did the Democratic Party take us for granted? ...Let me answer your question. You said the Republican Party was the party of Lincoln and Frederick Douglass. It is true that Mr. Lincoln signed the Emancipation Proclamation, after which there was a commitment to give 40 acres and a mule. That\'s where the argument, to this day, of reparations starts. We never got the 40 acres. We went all the way to Herbert Hoover, and we never got the 40 acres. We didn\'t get the mule. So we decided we\'d ride this donkey as far as it would take us... â€œMr. President, the reason we are fighting so hard, the reason we took Florida so seriously, is our right to vote wasn\'t gained because of our age. Our vote was soaked in the blood of martyrs, soaked in the blood of good men, soaked in the blood of four little girls in Birmingham. This vote is sacred to us. This vote can\'t be bargained away. This vote can\'t be given away. Mr. President, in all due respect, Mr. President, read my lips: Our vote is not for sale... â€œAnd it occurred to me as I heard Ray Charles singing, that Ray wasn\'t singing about what he knew, because Ray had been blind since he was a child. He hadn\'t seen many purple mountains. He hadn\'t seen many fruited plains. He was singing about what he believed to be. Mr. President, we love America, not because all of us have seen the beauty all the time. But we believed if we kept on working, if we kept on marching, if we kept on voting, if we kept on believing, we would make America beautiful for everybody. Starting in November, let\'s make America beautiful again.â€&nbsp;  &nbsp;   '),(22227,2824507,'28,July,2004','2004-07-28 00:00:00','    A recap of Tuesday night\'s speeches at the Democratic National Convention     Teresa Heinz Kerry    â€œMy name is Teresa Heinz Kerry. And by now I hope it will come as no surprise to anyone that I have something to say...  â€œI know how precious freedom is. It is a sacred gift, sanctified by those who have lived it and those who have died defending it. My right to speak my mind, to have a voice, to be what some have called â€œopinionated,â€ is a right I deeply and profoundly cherish. My only hope is that, one day soon, women &shy;who have all earned the right to their opinions&shy;, instead of being labeled opinionated, will be called smart or well-informed, just as men are...  â€œJohn is a fighter. He earned his medals the old-fashioned way, by putting his life on the line for his country. No one will defend this nation more vigorously than he will &shy;and he will always be first in the line of fire...  â€œThe Americans John and I have met in the course of this campaign all want America to provide hopeful leadership again. They want America to return to its moral bearings. It is not a moralistic America they seek, but a moral nation that understands and willingly shoulders its obligations; a moral nation that rejects thoughtless and greedy choices in favor of thoughtful and generous actions; a moral nation that leads through the power of its ideas and the power of its example. We can and we should join together to make the most of this great gift we have been given, this gift of freedom, this gift of America.â€     Ron Reagan    â€œA few of you may be surprised to see someone with my last name showing up to speak at a Democratic convention. Let me assure you, I am not here to make a political speech, and the topic at hand should not, must not, have anything to do with partisanship...  â€œI am here tonight to talk about the issue of research into what may be the greatest medical breakthrough in our or in any lifetime: the use of embryonic stem cells â€” cells created using the material of our own bodies â€” to cure a wide range of fatal and debilitating illnesses...  â€œNow, there are those who would stand in the way of this remarkable future, who would deny the federal funding so crucial to basic research... A few of these folks, needless to say, are just grinding a political axe and they should he ashamed of themselves. But many are well-meaning and sincere. Their belief is just that, an article of faith, and they are entitled to it. But it does not follow that the theology of a few should be allowed to forestall the health and well-being of the many. And how can we affirm life if we abandon those whose own lives are so desperately at risk?...  â€œWhat excuse will we offer this young woman (13-year old diabetes sufferer) should we fail her now? What might we tell her children? Or the millions of others who suffer? That when given an opportunity to help, we turned away? That facing political opposition, we lost our nerve? That even though we knew better, we did nothing?...  â€œWhatever else you do come November 2nd, I urge you, please, cast a vote for embryonic stem cell research.â€     Barack Obama    â€œMy parents shared not only an improbable love, they shared an abiding faith in the possibilities of this nation. They would give me an African name, Barack, or â€œblessed,â€ believing that in a tolerant America your name is no barrier to success. They imagined me going to the best schools in the land, even though they weren\'t rich, because in a generous America you don\'t have to be rich to achieve your potential... I stand here today, grateful for the diversity of my heritage...knowing that my story is part of the larger American story, that I owe a debt to all of those who came before me, and that, in no other country on earth, is my story even possible...  â€œGo into any inner city neighborhood, and folks will tell you that government alone can\'t teach kids to learn â€“ they know that parents have to parent, that children can\'t achieve unless we raise their expectations and turn off the television sets and eradicate the slander that says a black youth with a book is acting white...  â€œFor alongside our famous individualism, there\'s another ingredient in the American saga. A belief that we are connected as one people. If there\'s a child on the south side of Chicago who can\'t read, that matters to me, even if it\'s not my child. If there\'s a senior citizen somewhere who can\'t pay for her prescription, and has to choose between medicine and the rent, that makes my life poorer, even if it\'s not my grandmother. If there\'s an Arab American family being rounded up without benefit of an attorney or due process, that threatens my civil liberties. It\'s that fundamental belief â€“ I am my brother\'s keeper, I am my sister\'s keeper â€“ that makes this country work. It\'s what allows us to pursue our individual dreams, yet still come together as a single American family. E pluribus unum...  â€œYet even as we speak, there are those who are preparing to divide us, the spin masters and negative ad peddlers who embrace the politics of anything goes. Well, I say to them tonight, there\'s not a liberal America and a conservative America â€“ there\'s the United States of America. There\'s not a Black America and White America and Latino America and Asian Americaâ€“ there\'s the United States of America. The pundits like to slice-and-dice our country into Red States and Blue States; Red States for Republicans, Blue States for Democrats. But I\'ve got news for them, too. We worship an awesome God in the Blue States, and we don\'t like federal agents poking around our libraries in the Red States. We coach Little League in the Blue States and have gay friends in the Red States. There are patriots who opposed the war in Iraq and patriots who supported it. We are one people, all of us pledging allegiance to the stars and stripes, all of us defending the United States of America.â€     Howard Dean    â€œI was hoping for a reception like this. I was just hoping that it would be on Thursday night, instead of on Tuesday night...  â€œI may not be the nominee, but I can tell you this: For the next hundred days, Iâ€™ll be doing everything I can to make sure that John Kerry and John Edwards take our country back for the people who built it. Because tonight, weâ€™re all here to represent the Democratic wing of the Democratic party...  â€œWeâ€™re not going to be afraid to stand up for what we believe. Weâ€™re not going to let those who disagree with us shout us down under a banner of false patriotism. And weâ€™re not going to give up a single voter, or a single state...  â€œTo everyone who supported me youâ€™ve given me so much, and I canâ€™t thank you enough. But this was never about me. It was about us. It was about giving new life to our party, new energy to our democracy, and providing hope again for the greatest nation on earth.â€   '),(22228,2824507,'27,July,2004','2004-07-27 00:00:00','    A recap of Monday nightâ€™s speeches at the Democratic National Convention     Bill Clinton    â€œNow this state, who gave us in other times of challenge John Adams and John Kennedy, has given us John Kerry, a good man, a great senator, a visionary leader... â€œWe Democrats want to build a world and an America of shared responsibilities and shared benefits. We want a world with more global cooperation where we act alone only when we absolutely have to. We think the role of government should be to give people the tools to create the conditions to make the most of their own lives. And we think everybody should have that chance... â€œOn the other hand, the Republicans in Washington believe that America should be run by the right people â€” their people â€” in a world in which America acts unilaterally when we can and cooperates when we have to. They believe the role of government is to concentrate wealth and power in the hands of those who embrace their economic, political and social views, leaving ordinary citizens to fend for themselves on important matters like health care and retirement security. Now, since most Americans aren\'t that far to the right, our friends have to portray us Democrats as simply unacceptable, lacking in strength and values. In other words, they need a divided America. But we don\'t. Americans long to be united... â€œFor the first time when America was in a war footing in our whole history, they gave two huge tax cuts, nearly half of which went to the top 1 percent of us. Now, I\'m in that group for the first time in my life. And you might remember that when I was in office, on occasion, the Republicans were kind of mean to me. But as soon as I got out and made money, I became part of the most important group in the world to them. It was amazing. I never thought I\'d be so well cared for by the president and the Republicans in Congress. I almost sent them a thank you note for my tax cuts until I realized that the rest of you were paying the bill for it. And then I thought better of it... â€œNow, we tried it their way for 12 years. We tried it our way for eight years. Then we tried it their way for four more. But the only test that matters is whether people were better off when we finished than when we started. Our way works better... â€œDuring the Vietnam War, many young men, including the current president, the vice president and me, could have gone to Vietnam and didn\'t. John Kerry came from a privileged background. He could have avoided going too, but instead, he said: Send me... â€œNow, everybody talks about John Edwards\' energy and intellect and charisma. You know, I kind of resent him... â€œTheir opponents will tell you we should be afraid of John Kerry and John Edwards, because they won\'t stand up to the terrorists. Don\'t you believe it. Strength and wisdom are not opposing values...  â€œNow, again, it is time to choose. Since we\'re all in the same boat, we should choose a captain of our ship who is a brave good man, who knows how to steer a vessel through troubled waters, to the calm seas and the clear sides of our more perfect union. That is our mission. So let us go in tonight and say to America in a loud, clear voice: Send John Kerry.â€     Hillary Rodham Clinton    â€œWe meet at a moment of great peril, but also of great hope... â€œ(John Kerry) will lead the world, not alienate it. Lower the deficit, not raise it. Create good jobs, not lose them. Solve a health care crisis, not ignore it. I know a thing or two about health care. And the problems have only gotten worse in the past four years... â€œBeing a Senator from New York, I saw first-hand the devastation of 9/11..I hope no American ever has to witness a sight like that again. That tragedy changed all of us. I know it changed me. And every day now, as a mother, as a Senator, and as an American I worry about whether we are acting as wisely as we can to protect our country and our people. Last week, the bipartisan 9/11 commission issued its report. It was a sober call to action that we ignore at our peril. John Kerry understands whatâ€™s at stake... â€œJohn Kerry is a serious man, for a serious job. So letâ€™s work our hearts out and send him to the White House in 2004. And Iâ€™m optimistic we will because I know a great leader when I see one.â€    Jimmy Carter   â€œTwenty-eight years ago, I was running for president, and I said then I want a government as good and as honest and as decent and as competent and as compassionate as are the American people. I say this again tonight, and that\'s exactly what we will have next January with John Kerry as president of the United States of America... â€œToday our Democratic Party is led by another former naval officer, one who volunteered for military service. He showed up when assigned to duty and he served with honor and distinction. He also knows the horrors of war and the responsibilities of leadership. And I am confident that next January he would restore the judgment and maturity to our government that nowadays is sorely lacking... â€œAfter 9/11, America stood proud, wounded but determined and united. A cowardly attack on innocent civilians brought us an unprecedented level of cooperation and understanding around the world. But in just 34 months we have watched with deep concern as all this good will has been squandered by a virtually unbroken series of mistakes and miscalculations.... â€œYou can\'t be a war president one day and claim to be a peace president the next depending on the latest political polls. When our national security requires military action, John Kerry has already proven, in Vietnam, that he will not hesitate to act. And as a proven defender of our national security, John Kerry will strengthen the global alliance against terrorism while avoiding unnecessary wars.â€     Al Gore   â€œIâ€™m going to be candid with you. I had hoped to be back here this week under different circumstances, running for re-election. But you know the old saying: you win some, you lose some. And then thereâ€™s that little-known third category... â€œTake it from me, every vote counts. In our democracy, every vote has power. And never forget that power is yours. Donâ€™t let anyone take it away from you or talk you into throwing it away... â€œI sincerely ask those watching at home tonight who supported President Bush four years ago: did you really get what you expected from the candidate you voted for? Is our country more united today? Or more divided? Has the promise of compassionate conservatism been fulfilled? Or do those words now ring hollow? For that matter, are the economic policies really conservative at all? For example, did you expect the largest deficits in history, year after year? One right after another? And the loss of more than a million jobs? By the way, I know about the bad economy. I was the first one laid off...  â€œI also ask tonight for the consideration and the help of those who supported a third party candidate in 2000. I urge you to ask yourselves this question: Do you still believe that there was no difference between the candidates?... â€œWe have to be crystal clear about the threat we face from terrorism. It is deadly. It is real. It is imminent. But in order to protect our people, shouldnâ€™t we focus on the real source of this threat: the group that attacked us and is trying to attack us again: Al Qaeda, headed by Osama bin Laden? Wouldnâ€™t we be safer with a president who didnâ€™t insist on confusing Al Qaeda with Iraq?â€  &nbsp;   '),(22229,2824507,'26,July,2004','2004-07-26 00:00:00','    If you look at  urlLink my Friendster profile , you will see that one of my favorite hobbies is \"Brainstorming Ideas for Novels I Never Have Time to Write.\" This is also true for screenplays. Sometimes Iâ€™m not sure if a particular idea would work better as a novel or as a screenplay but when I visualize ideas they often play in my mind like a film. Anyway I wrote a couple of screenplays in 1998 but currently hate everything about them. I know I can do better and over the weekend I started to jot down some notes on a new idea. Itâ€™s original in the sense that this particular story has never been done before, yet if I were to give you the pitch, youâ€™d say, \"Oh, itâ€™s just like ...\" Thatâ€™s because in my brainstorming process I have a tendency to focus on a well-respected film - take its basic themes and structure â€“ but then completely change the situations, settings and characters. So instead of being about \"X\", itâ€™s about \"Y\". Anyway weâ€™ll see how this one develops.  Since our office air conditioning was busted, we were given this giant industrial fan to use until the A.C. can be fixed. The fan is positioned right next to my desk and the best part is that itâ€™s whirring drowns out most other sounds. Phone rings? Sorry, canâ€™t hear it. Someoneâ€™s trying to get my attention? Sorry, canâ€™t hear you. Someone complimented me on my haircut this morning and thought I was dissing them by not responding. But I wasnâ€™t being rude, I didnâ€™t know that she said anything! Itâ€™s a very Larry David / \"Curb Your Enthusiasm\" situation.  I was asked this morning who will profit from the proceeds of the 9/11 Commission report, currently available in a bookstore near you.&nbsp; My theory is that Tom Keane, Lee Hamilton and the rest of the commission is going to use the proceeds to take a nice relaxing vacation in Tahiti.  On the speaking schedule for the Democratic Convention tonight : Al Gore, Jimmy Carter, Hillary Clinton, Bill Clinton. In other words, itâ€™s nostalgia night. (Or in Hillaryâ€™s case, the future?)   '),(22230,2824507,'25,July,2004','2004-07-25 00:00:00','    Mini film reviews...  I saw \"Bad Santa.\"&nbsp; My opinion of that movie can be found in its title.&nbsp; And my opinion ain\'t \"Santa.\"  Saw the first 20 minutes of \"Malibu\'s Most Wanted\" on HBO.&nbsp; Would have watched more but my attention span was too short.&nbsp; Did like the bit where the rabbi asks the kid \"What theme do you want for your Bar Mitzvah?&nbsp; Star Wars?&nbsp; Baseball?&nbsp; Superheroes?\"&nbsp; And the kid replies \"An O.P.P. theme!\"&nbsp; Cut to the bar mitzvah where the 13-year old is grinding with some hootchies to the tune \"I got an O.P.P., Yeah you know me!\", as the horrified guests look on.  And speaking of strange bar mitzvahs, the Daily News ran a story yesterday on people who give bar mitzvahs for their dogs.&nbsp; They dress their dogs up in mini-yarmulkahs and mini-tallises.&nbsp; I know it\'s a sign of affection for their pets but I still think these people have too much time on their hands.  Attended&nbsp;a doubleheader yesterday for the Brooklyn Cyclones.&nbsp; They lost the first game.&nbsp; Then they were losing the second game and we decided to leave early.&nbsp; Shortly after leaving, the Cyclone came back and won the second game.&nbsp; So typical!  And yes, as I\'m sure everyone is wondering, my barber got back from his vacation and I finally got that haircut yesterday.&nbsp; Now I\'m shorn for the summer.  Had some strange dreams over the weekend.&nbsp; On Friday I dreamt I was a member of the press and was flown to Baghdad to cover the next court appearance of Saddam Hussein.&nbsp; Then last night I had a dream that Manhattan became engrossed in riots similar to those in \"Gangs of New York\" and everyone fled to Brooklyn.  Democratic Convention starts tomorrow.&nbsp; Yay!   '),(22231,2824507,'23,July,2004','2004-07-23 00:00:00','    \"Philâ€™s Spin Zone\" returns to more serious political commentaries...  After months of research, interviews, and stiff resistance from the Bush administration, the 9/11 Commission released their final report yesterday. They concluded, \"This was a failure of policy, management, capability and above all a failure of imagination.\" Rather than going after individuals, many government agencies were blamed including both the Clinton and Bush administrations. The commission issued many recommendations that they believe will improve our intelligence capabilities and I believe those recommendations should be taken seriously and implemented as soon as possible. Unfortunately, House Speaker Dennis Hastert said that there wouldnâ€™t be time to consider the changes until next year. How is it possible that in the face of all these color coded threat warnings, that Hastert and others can be so indifferent to our shortcomings? That he dismisses the commissionâ€™s recommendations as \"another layer of bureaucracy\" ignores the fact that this was a bipartisan commission that explored every angle of the 9-11 tragedy to the fullest extent. They are legitimate and they deserve the trust and respect of the U.S. Congress.  Today is the 1-Year anniversary of James Davisâ€™ assassination. Davis was a Brooklyn City Council Member who was murdered last year inside of City Hall by a deranged member of his entourage. I admired Davis and was fortunate to have met him on three occassions. The first time was shortly after his election to the City Council. I was eating lunch in a Fort Greene diner when he came in to shake hands and introduce himself. The second time was at a Parks ribbon cutting event where I took pictures of him going down the slides. The last time I saw him was at Mayor Bloombergâ€™s State of the City address at the Brooklyn Botanic Garden. His assassination last year was shocking and has led to enhanced security measures inside of City Hall.   '),(22232,2824507,'22,July,2004','2004-07-22 00:00:00','    This is my 200th blog posting.&nbsp; Hoo-ray!!!  David Wrightâ€™s Met debut... 0 for 4. But he was defensively sound and the Mets won the game. Of course, they only won because the Expo first baseman dropped a routine throw, but weâ€™ll take whatever wins we can get.   Convention mania is heating up. The Democrats hit Boston next week and \"Philâ€™s Spin Zone\" will provide a complete recap of what happens inside the Fleet Center. Then the Republicans come to NYC in late August, at which point I will avoid Manhattan like the plague.   Last night I got my butt kicked in a hardcore game of Scrabble. It had been many years since I played and my opponent showed no mercy. All the high-point letters were used against me... Zs, Qs, but sadly no Os. Revenge will be mine though. Oh yes, revenge will be mine.   '),(22233,2824507,'21,July,2004','2004-07-21 00:00:00','    Thereâ€™s this great cartoon circulating the internet spoofing Bush and Kerry. Itâ€™s quite bipartisan as each candidate -- thanks to animated photos and voice impersonations -- trades character assassination accusations to the melody of \"This Land is Your Land.\"  urlLink View it here! &nbsp;  &nbsp;  Just as soon as the Mets start winning, the Mets start losing. Itâ€™s going to be one of those years. They lost their second game in a row to Florida last night, putting them back under .500. But theyâ€™re only three games out of first place in the crowded and mediocre NL East. Tonight they call David Wright up from the minor leagues? Will he meet the expectations or will he be yet another of their over-hyped prospects? Stay tuned.   For those wondering how Iâ€™m doing in Fantasy Baseball (and Iâ€™m sure youâ€™re ALL wondering), my team \"Thugz Inc.\" is in 7th place in a 12-team league. But Iâ€™m within striking distance of 6th!  &nbsp;  urlLink Check out this link from the NY Daily News about the park event that I talked about yesterday.   '),(22234,2824507,'20,July,2004','2004-07-20 00:00:00','    We renamed a park today in the Cypress Hills / East New York section of Brooklyn.&nbsp; Formerly Verona Playground, it will now be known as George Walker Jr. Park.&nbsp; Walker was a community activist who fought against the proliferation of drugs on his block.&nbsp; He helped the City seal off a condemned building that became a drug haven.&nbsp; Once the building was knocked down, it was converted into the playground that now bears his name.&nbsp; It was a nice event, with Walker\'s family in attendance.&nbsp; My favorite part came after it was over though.&nbsp; It was another hot sunny day so we turned on the spray shower and I ran through it, clothing and all!&nbsp; I spent the rest of the day soaking wet but it was worth it.   In honor of George Walker Jr., I believe the following quote from Ali G. is appropriate... \"Drugs.&nbsp; It\'s not something&nbsp;you should do.&nbsp; It\'s something you should don\'t.\"   '),(22235,2824507,'19,July,2004','2004-07-19 00:00:00','    Today is the last day at work for one of my co-workers, Paul Thomsen. He is off to Kansas where he will be pursuing his Ph.D. in military history. Paul is also a historical writer and he has a novel coming out in September. Itâ€™s called \"Rebel Chief : The Motley Life of William Holland Thomas\" and itâ€™s about a white man in the 19th century who became the chief of an Indian tribe. I am plugging his book especially because Paul was good enough to list me in his acknowledgments. So,  urlLink preorder it here!   '),(22236,2824507,'18,July,2004','2004-07-18 00:00:00','    Look &nbsp;   at &nbsp;  the &nbsp;   pretty &nbsp;     colors&nbsp;  and &nbsp;   sizes...     &nbsp;  Enough playing with the new Blogger toys now.&nbsp; Back to business.  &nbsp;  Well, it\'s confirmed.&nbsp; The reason VH-1 left out so much in their \"I Love the 90s\" series is that they are planning for a sequel, \"I Love the 90s, Part II\" or something like that.&nbsp; Remember when channels like MTV and VH-1 actually played music videos?  &nbsp;  Very slow weekend.&nbsp; Attended the Siren Music Festival and&nbsp;a Cyclones game at Coney Island but the music sucked and the Cyclones lost.&nbsp;&nbsp; And my&nbsp;barber is still on vacation so&nbsp;no haircut for me. &nbsp;  Tonight it\'s time to enjoy one of my guilty pleasures.&nbsp; No, not that...&nbsp; I\'m talking about \"Da Ali G Show\" on HBO.&nbsp;&nbsp;&nbsp; Check out this exchange from last season\'s first episode.&nbsp; Ali is interviewing former Attorney General Richard Thornburgh...  &nbsp;  Ali: What is legal?  Richard: Well, I think most conduct most of us engage in on a day-to-day basis is legal.  Ali: So, what is illegal?  Richard: What is illegal is what the elected representatives of the people define as crimes.  Ali: What is barely legal?  Richard: Well, that\'s where you get into technicalities and you have trials.  Ali: \'Cause me saw dis film called Barely Legal Three and it was about these two naughty college girls and them have done their own work and then as punishment they had to have a three-header with their supervisor, this teacher. Is that to do with the law?  Richard: Uh, it\'s hard to say. That\'s probably governed by the rules of the institution, the college.   '),(22237,2824507,'16,July,2004','2004-07-16 00:00:00','    The first order of business is to clarify something that I wrote in \"1995\". I mentioned that I helped to write the musical, Sing. This is true but \"help\" is a vague word. I was part of the committee that came up with the concepts, themes and settings of the show. The bulk of the writing was done by Danielle Weinstein and Rachel Levine and they deserve credit for it. Happy, Danielle? :)   And now the conclusion of my own \"I Love the 90s\"...    1998    1998 was the year of the Syndicate. What\'s that you say? The Syndicate was formed by me, Dave and Ben who had just returned from Beijing. This triumverate evolved into an oranized crime group, thugging and pimping our way across campus. But while all that fun was going on, there was some seriousness too as my father had a serious quadruple heart bypass operation near the beginning of the year. Fortunately all went well with it. I also made my film debut, performing in Dave\'s student film \"Time.\" I had the lead role, of a college student diagnosed with leukemia. Dave got an A and my performance was acknowledged as a big part of it. I then made my electoral comback getting elected Secretary of the Class of \'99. So what if I was running unopposed? Everyone else feared going up against me! In May I said goodbye to many of my Class of 98 friends who were graduating such as Bartow and Briar. Jon and I would have to carry the nVo mantle ourselves. The summer of 98 was relaxing. I spent some of it working on Weiner\'s congressional campaign. I also worked on two screenplays and visited Vassar in July where both Ben and Dave were staying there for the summer. In the fall, I headed back for my senior year. Ben, Dave and I were unable to get a TH or TA in the room draw so we settled for a suite in Main that was promptly renamed TH D-Funkt. We spent the year feuding with our neighbors, The Turkish Harem. One time those girls snuck into our living room and created a maze of yarn in it, ala \"The Parent Trap\". I also got to know my fellow class officers, Becky, Marivel and Jill. Together, we decided that our old class account e-mail password would be \'3L1J\', representing our ethnicities... three Latinas and one Jew. We interviewed for our class council and appointed a wide variety of personalities including Homeskool!&nbsp; We planned the campus-wide Halloween party and it was awesome.&nbsp; I dressed as Sting, the former WCW wrestler.&nbsp; Also that year I took a swing dancing class with Kristin. And more intramural softball. 1998 was fun.&nbsp;    1999      \"And tonight we\'re gonna party like it\'s 1999.\" And I did. With most of my serious academic work out of the way, my final semester at Vassar did resemble a big party. The year began with a cocktail / housewarming party in my suite. Which was followed by many parties in the THs. And weeknights at the Mug (where I found out they served free sodas!) Then I served on the committee to find a commencement speaker and we wound up getting James Earl Jones. Awesome choice. I also had a political career to finish and I did that by serving as Chair of the Board of Elections. Jon was my assistant chair and we had a great time. I only had to disqualify one candidate and he brought me before the school\'s judicial board because of it! (P.S. I won the case.) Another of my jobs was picking who would do the annual commencement musical. After one Sarah Ackerman told me that she didn\'t have time to do it, I wound up with two proposals - \"Pippin\" and \"The Fantasticks\". I picked \"Pippin\" as I thought the themes were more fitting for a commencement. Then it was senior week with the booze cruise, the winery brunch, the Star Wars Episode 1 premiere, the lazer tag and other events I can\'t remember. Before I knew it graduation weekend arrived and so did the family. Saturday was a fantastic day culminating in the bonfire where I burned the books I hated most. But on Sunday it rained, putting a slight damper on the actual graduation ceremony. Then it was back to Brooklyn and the very next day I interviewed with the Parks Department. Got hired two days later. Started two weeks later. And a new era began. Blah blah blah. Work work work. The decade ended amid Y2K anxieties and a New Years Party at Jill\'s place.   Hope everyone enjoyed my look back at the 90s!   '),(22238,2824507,'15,July,2004','2004-07-15 00:00:00','    VH-1\'s \"I Love the 90s\" has been pretty good but there have been some glaring omissions, including stuff that they advertised would be included.  So far they\'ve covered 1990-1995 and there\'s been no Seinfeld, no Friends, no ER, no Beavis and Butthead, no My So Called Life and no Simpsons.  Maybe they couldn\'t get the rights to show the clips?  Maybe they\'re already planning for \"I Love the 90s Strikes Back\"?  It\'s odd.  To the anonymous person leaving comments - identify yourself!  E-mail me if you\'d prefer.  And now the series continues...   1996   1996 began with an awesome New Years Eve with high school friends while seeing a They Might be Giants concert in NYC.  Then it was back to campus during a snowstorm.  And for months it snowed and snowed.  Very depressing.  I didn\'t see grass until April.  I went home for spring break and spent it shoveling snow.  But in late March I was elected president of my dorm!  In a three way race, I got 53% of the vote.  My two opponents lived on the same floor so they split the votes of that floor while I carried the other three floors.  Now that\'s strategy!  Best part of being elected president was picking the best room in the dorm for myself for sophomore year.  I picked a big double, right by the center stairwell and with windows overlooking Noyes Circle and opening onto the mushroom. I think I declared political science as my major sometime around then.   Upon returning to campus for soph year, I was busy giving orientations to the new freshmen, enforcing (and choosing when not to enforce) house rules, and planning social events.  Like jello wrestling!  Yes, that was our big party.  Outdoor jello wrestling.  Besides the cold, I remember participating in a house president battle royale.  Also in 1996, the nVo was formed.  That\'s the New Vassar Order consisting of me, Jon, Bartow and Briar.  We sat together in Mr. Born\'s class on the American Presidency, had lunch afterwards and soon started attending pro wrestling events!  The history of the nVo, featuring all of our exploits, can be found on the  urlLink nVo history page.    1997   1997 began with another New Years party with high school friends but as our lives were diverging on different paths, it would be the last time I would see several of them.  Back to Vassar to finish my term as Noyes President.  My next step was running for Vice President of the VSA.  But I lost to Tara, a friend and a house officer who had worked for me.  So that sucked.  I rebounded by getting elected Treasurer of the Jewish Union but that was just a consolation prize.  Most fun of 1997?  The Mens Seder.  In 1996, the Mens Seder consisted of me, Bartow, Moshe and Getnick, taking over the Jew House, downing bottles of Manischvitz, eating a full tray of chicken each, watching porn and making crank calls to our friends.  Getnick had so much fun that he flunked his finals and had to drop out of Vassar.  In 1997, the Mens Seder was expanded to 15 people.  No porn or phone calls but plenty of Manischevitz.  Our party then proceeded to the Mug wearing red Turkish fezzes.  In the summer I worked for Councilman Weiner and then returned to Vassar for junior year.  Many of my good friends were JYA - Jessica in London, Ben in Beijing - so I became closer with Dave, the nVo guys, and a bunch of underclassman who lived on my floor.  Although I originally had a small single, I got to move into another huge double for myself when someone moved out.  It was also during this semester that I made my debut playing and managing an intramural softball team.  Our team name?  Nobi\'s Horde of Doom.    Oh and to those of you who wanted juicy gossip, sorry.  Too many people I know read this thing.  Stay tuned tomorrow for the conclusion of this series - 1998 and 1999!   '),(22239,2824507,'14,July,2004','2004-07-14 00:00:00','    More from my own \"I Love the 90s\"...   1994   1994.  The year of college applications.  The process of looking at schools, visiting them, applying to them and interviewing for them seemed like it completely consumed me for the entire year.  And that includes taking the SAT and doing all the prep work for it including Saturday sessions of The Princeton Review.  The first college that I visited actually took place in late 1993 - SUNY Binghamton.  The rest of my visits took place in the summer of \'94 and they included Princeton, Brown, Amherst, Vassar, Haverford, UPenn, Cornell and Wesleyan.  I wound up applying to all of the above except for Haverford and UPenn.  Princeton was my first choice and I applied Early Action but was deferred into the regular pool of applicants.  Of the other choices, I was most impressed with Vassar.  Elsewhere in 1994, I rejoined the legitimate student government by winning an election for Vice President.  This was also the year I felt the most comfortable in my social skin, spending time with a great group of friends including Joslyn, Lisa, Dan K, Dan H, Adam, Lauren, Jessica, Melissa and others establishing ourselves as the \"second floor\" crowd.   And I should also mention visiting the Holocaust Museum in Washington D.C. right when it first opened.     1995   What a year this was.  The end of high school.  The summer transition.  The start of college.  The year started with some sad news as my Grandma Rose passed away, my last remaining grandparent.  Then with college applications in the mail, I was able to relax and participate in Sing! (an original musical production).  I helped write the script and had several roles in the show including Elvis!  Then I was casted in another musical, \"The Roar of the Greasepaint, The Smell of the Crowd.\"  In the spring I was disappointed to get a rejection letter from Princeton but happy to get an acceptance letter from Vassar.  It was clear that I much preferred Vassar to the rest of the schools that accepted me so (once the financial aid details were worked out) I made the decision to move to Poughkeepsie!  The rest of senior year included a weekend trip to Quebec City, a prom on the Spirit of NY cruise ship, and a graduation at Madison Square Garden where I again got to speak!  The summer was spent working for Anthony Weiner again and preparing for the big move.  Many weekends were spent with the parents at stores like Bed Bath & Beyond and Linens and Things.  Then in late August it was off to Vassar!  I was glad to have Jess with me from H.S. to help smooth the transition.  Once there I quickly made friends with my own hallmates in Noyes House and with Jess\'s hallmates in Davison House.  In October, Hillary Clinton came to campus to speak at freshman parents weekend.  I also had one roommate, Otis, drop out after only six weeks and another roommate, Jon, replace him.  Oh, and Jon brought his loud bird with him.  It was an exciting semester!  Stay tuned tomorrow for 1996 and 1997 - juicy Vassar gossip!   '),(22240,2824507,'13,July,2004','2004-07-13 00:00:00','    My personal version of â€œI Love the 90sâ€ continues...   1992   The year started with my seeing the movie â€œJFKâ€ and becoming utterly convinced that Kennedy was killed in a coup de tat orchestrated by Lyndon Johnson.  I even read a poem that I wrote about it aloud to my English class.  And for this high school freshman, thus began my political career.  It was followed by my election to the position of â€œSophomore Senatorâ€ for the next year.  I beat an Asian kid who could barely speak English by a very significant margin!  I also entered the working world, doing a summer internship for Chuck Schumer.  But the best political news for me was seeing Bill Clinton get elected President.  To see a Democrat win the White House for the first time in my life was euphoric.  And speaking of euphoria, I fell absolutely head over heels for this girl who sat next to me in English class, named Joslyn.  I would carry a serious thing for her throughout my four years in high school.  Quite an interesting history there.  Elsewhere in 1992, I went on my last family vacation to California.  This time we visited San Diego as well as L.A.   1993   After interning for Chuck Schumer in 1992, and being one out of a gazillion interns in the office, I decided to go somewhere where I could be of more use.  So in the summer of â€™93 I began working for then-Councilman (now Congressman) Anthony Weiner.  Having just been elected to the City Council, I was Weinerâ€™s very first intern and would continue to go back for many exciting summers to come.  Back at school, I decided to prematurely run for President of the Student Alliance.  It was an ambitious and, in hindsight, stupid move to run against an upperclassman and I got my ass handed to me.  Bitter till the end, I decided to become a little revolutionary.  Since student government has no power anyway, I banded all of the other election losers together and we formed our own little counter-government, challenging the real oneâ€™s legitimacy at every turn.  Such fun!  Since I had some free time I decided to enter the world of drama.  I auditioned for my schoolâ€™s production of â€œBrighton Beach Memoirsâ€ and came in second for the lead role of Eugene Jerome.  Although disappointed, I went on to audition for the next musical â€œHello, Dolly!â€ and this time was casted.  I played a dancing waiter, the court clerk and several other ensemble roles.  Clearly I was casted for acting skills though, and not because of singing or dancing!  But most importantly in 1993, the Abramson family finally got cable TV.  Stay tuned tomorrow for 1994 and, the most significant year of the decade for me, 1995.   '),(22241,2824507,'12,July,2004','2004-07-12 00:00:00','    Every night this week from 9-11, VH-1 is showing â€œI Love the 90sâ€, a celebration of pop cultural events throughout the decade.  I completely enjoyed â€œI Love the 80sâ€ as that decade comprised my childhood.  My brother even taped the show onto DVDs and Iâ€™ve watched it over and over for the past year and a half.  VH-1â€™s â€œI Love the 70sâ€ program didnâ€™t thrill me as much since I was barely alive in that decade and donâ€™t remember much from the 2 years and three months that I did see of the 70s.  But the 90s was quite the decade for me as it covered junior high, high school, college and the working world.  I canâ€™t wait to see how VH-1 depicts it.  Tonight they will be covering 1990 and 1991, tomorrow will be 1992 and 1993, and so on.  So what I thought Iâ€™d do with this blog is to take two years per day and summarize what happened in MY life during those years.  My own personalized version of â€œI Love the 90s.â€       1990   1990 can be summarized by two words.  Bar Mitzvah.  Having never gone to Hebrew School, I was tutored at home every Monday night by Rabbi Eli Orgel.  He taught me how to read Hebrew and about Jewish culture.  It culminated in a perfect reading of my haftorah that completely impressed the synagogueâ€™s resident rabbi, the congregation, and my family and friends.  To this day it might be one of the proudest moments of my life.  And the reception on the next day was so much fun, held in the same reception hall where my parents were married.  The weekend lived on throughout the decade through the wonders of Paris Video.  I also spent the year attending the bar mitzvahs of my friends â€“ Joel, Andrew, Ryan, Howie, David, Eric, Eli, and Iâ€™m sure there were others.  1990 was also notable for the junior high school experience.  What an awkward, peer pressure environment that was.  Basically, if you didnâ€™t wear Champion sweatshirts and Nike Air sneakers, then you just werenâ€™t considered â€œcool.â€  I spent much of the year struggling to conform so that I too could be part of the cool crowd.  My crushes that year included Jaime, Stacy and Debbie.  The end of the year class trip was to Boston.  And oh yeah, I wore my hair completely spiked.    1991   In 1991, I ditched the spiked hair and decided to comb it normally.  And in my last six months of junior high, I succeeded in becoming part of the cool crowd.  Naturally it would only last for a few months until graduation, officially ending the ridiculous junior high social structure.  That year I became very globally aware as the Persian Gulf War broke out.  Although Gulf War I ended in a month, it was the first war for my generation and we were worried that it would expand into something much worse.  We also went on a great class trip to Montreal and Quebec City.  But unfortunately, my Grandpa Murray died that year.   Our junior high graduation was held at Brooklyn College where I gave a speech, thus beginning my affinity for public speaking.  That summer was my last completely free summer of my childhood where I didnâ€™t have to work.  The family vacationed to Los Angeles, visiting my aunt and uncle who lived there at the time.  In the fall, I began Murrow High School!  I remember being physically exhausted after the first day and asking â€œDo I have to go back tomorrow?â€  But I loved Murrow and did indeed go back.   '),(22242,2824507,'11,July,2004','2004-07-11 00:00:00','    This weekend I committed myself to cleaning my apartment.  So much useless junk has accumulated over the past few years and it must be purged!  Particularly since I am considering moving to another place one of these days and do not want to have to deal with all the junk at the last minute.  I have a fun idea for this blog.  Stay tuned tomorrow and throughout the week to see what it is.   '),(22243,2824507,'10,July,2004','2004-07-10 00:00:00','    urlLink    Van Voorhees Ballfield Ribbon Cutting&nbsp; urlLink   '),(22244,2824507,'08,July,2004','2004-07-08 00:00:00','    So it\'s been a sobering week.  The unexpected loss of a friend in the prime of her life caused reflection on appreciating each day and enjoying the small, seemingly insignificant moments that comprise \"life.\"  All too often I find myself worrying about the future at the expense of appreciating the present.  Maybe this will help put things in better perspective.  Yesterday, we had a ribbon cutting for a new synthetic turf ballfield in Brooklyn.  (a picture is included above this post)  I was concerned about the prospect of a very hot day.  Thursday was brutal and the heat would have been magnified on the turf surface.  Fortunately, Friday cooled off a bit and it was a pleasurable day.  (Wendy, I guess you were right after all about the cool front - although it\'s hot again today!)  Saw some good movies this past week.  \"In America\" was a nice surprise, about an Irish family that moves to Hell\'s Kitchen NYC in the early 1980s, after the accidental death of their son.  The girls who played the 10 and 6 year old daughters were adorable.  Best line was when the 10 year old said to her father, \"Don\'t call me \'little girl.\'  I\'ve been carrying this family on my back for the past year.\"  I also enjoyed \"Cold Mountain\" and \"The Terminal\".  Both were a little slow and dragged in their middles but the stories and fine acting made up for it.  I was going to get a haircut today but my barber of 15 years is on vacation and I won\'t let anyone else do my hair.  So it will have to wait until next Saturday.  Currently playing on my XM Satellite Radio - Milli Vanilli, \"Blame it on the Rain\"   '),(22245,2824507,'08,July,2004','2004-07-08 00:00:00','    I want to extend my warm welcome to all of the new visitors to this blog.  Many of you are likely finding this page by doing google searches for Erin.  Yesterday alone, there were over 125 visitors to this site (far exceeding the typical 20-30 hits I get per day.)  The entry about her can be found a little further down this screen at Monday, July 5.  Feel free to add to the wonderful comments that have already been left there.   '),(22246,2824507,'07,July,2004','2004-07-07 00:00:00','    Hot.  So hot.  And the air conditioner at work is barely working.  Donâ€™t... know... how... much... longer... I... can...  (collapses at desk from heatstroke)  A slight breeze awakens me as I grasp for the water bottle at the other end of the desk.  So close, yet so far.  Can\'t... quite... reach... ah, got it! Gulp... gulp... gulp.  Ah...   Still hot though.   '),(22247,2824507,'06,July,2004','2004-07-06 00:00:00','    John Edwards has been selected by John Kerry as his Vice Presidential running mate.  Although the choice is not terribly surprising, it may have been the best choice that Kerry could have made.  Edwards gained prominence during the primary campaign and most Americans seem to have a positive opinion of him.  The worst that can be said about him is that heâ€™s inexperienced.  Then again, so was George W. Bush.  Looking at Kerry and Edwards together, one is reminded of Abe Lincoln (Kerry) and John F. Kennedy (Edwards).  Or at least Iâ€™m reminded of them.    The New York Post has lots of egg on its face this morning.  Their headline read â€œKerryâ€™s Choice : Dem picks Gephardt as VP candidateâ€, with a big front page photo of Kerry and Gephardt together.  Dumbasses.  They credited an unnamed source who was either misinformed or had the Post completely played.  It wouldnâ€™t surprise me if the Post was deliberately tipped off in the wrong direction just to make them look bad.  After all, they never resist an opportunity to bury the Democratic Party in their front page headlines.  Godspeed to Kerry and Edwards!   '),(22248,2824507,'05,July,2004','2004-07-05 00:00:00','    urlLink    Erin Schlather,  1977-2004&nbsp; urlLink   '),(22249,2824507,'04,July,2004','2004-07-04 00:00:00','    It pains me to write that one of my classmates from Vassar, Erin Schlather, passed away over the weekend.  She was involved in a single car accident on Saturday morning in her hometown of Ithaca, NY.  I have some wonderful memories of her.  She was exceptionally kind, funloving, intelligent, and always had a smile on her face.  Whether it was hanging out in Noyes House, or taking Bob Brigham\'s seminar on the Vietnam War, I was lucky to have known her.   urlLink Click here for the obituary.   '),(22250,2824507,'04,July,2004','2004-07-04 00:00:00','    HAPPY FOURTH OF JULY!!!  Actually, the entire notion that America was born on the 4th of July is a myth.  The Founding Fathers voted on adopting the Declaration of Independence on July 2.  In his letters, John Adams wrote that July 2 would be celebrated by future generations.  The Declaration wasn\'t signed on July 4 either.  That was a gradual process that occured over several months.  But don\'t let those facts get in the way of the celebrations!  The best part of the 4th of July has got to be the hot dog eating contest at Coney Island.  It\'s become so famous that this year, ESPN made a one-hour show out of it.  And for the fourth year in a row, Kobayashi from Japan, a man thinner than ME, won the thing.  He broke his own record by eating 53 and a half dogs in 12 minutes.  His prize?  A years supply of hot dogs.  The Mets swept the Yankees this weekend!  Friday nights game was an 11-2 blowout.  Yesterday they won 10-9 and today they won 6-5.  The best part of today\'s win was that it happened on George Steinbrenner\'s birthday.  Take that!  And finally, Roger Federer won the mens title at Wimbledon, defeating Andy Roddick in 4 sets.  Roddick was ahead when the rains came, forcing a delay.  When play resumed, Federer took over and established himself as the undisputed #1 player in the world.  I remain partial to Maria Sharapova though.  Doesn\'t the trophy that she\'s holding in the picture below look like a huge Passover Seder plate?   '),(22251,2824507,'03,July,2004','2004-07-03 00:00:00','    urlLink    Maria Sharapova, 2004 Wimbledon Champion&nbsp; urlLink   '),(22252,2824507,'02,July,2004','2004-07-02 00:00:00','    Congratulations, Maria Sharapova!  The 17-year old triumphed at Wimbledon this morning, confounding my expectations by upsetting Serena Williams.  At a score of 6-1, 6-4, it wasn\'t even close.  Afterwards, she proved to have a highly engaging personality, delivering a heartfelt and humorous interview to the BBC reporter.  She even tried calling her mom on her cell phone in the middle of Centre Court but complained that she couldn\'t get a good signal.   '),(22253,2824507,'02,July,2004','2004-07-02 00:00:00','    So last month I got this thing called â€œSiteMeterâ€ put on my blog.  Itâ€™s fascinating.  It tells you how many hits this page gets a day, what time zones the hits come from, what the domain addresses are that access the page, and other fun stats.  For example, the most popular domains are AOL, IP addresses, Road Runner, â€œNForâ€, Verizon and Optimum Online but there are some other unique ones that I noticed such as â€œharvard.eduâ€, â€œsanantonio.govâ€, â€œsunoil.comâ€ and â€œbuckeye-express.comâ€  One never knows who is lurking in the heart of the internet!  There were two great womens semifinals at Wimbledon yesterday.  Maria Sharapova beat Lindsay Davenport in 3 sets.  For Sharapova, it is her first grand slam final and it already establishes her as a better player than fellow Russian hottie, Anna Kournikova.  Impressive as she is, she is likely to be overwhelmed in the final by Serena Williams who defeated Amelie Mauresmo in her semi.  When healthy, Serena is the best player on the womens tour and she should destroy Sharapova.  Doesnâ€™t mean I wonâ€™t be rooting for an upset though!  On the mens side, the semifinals were delayed by rain and will be completed tomorrow.  However, its on track to be a Roger Federer / Andy Roddick final.  They are the top two ranked players in the world and it should be an amazing match.  Iâ€™m all set for Breakfast at Wimbledon!   '),(22254,2824507,'01,July,2004','2004-07-01 00:00:00','    For the first time since last December, new photos were released of Saddam Hussein.  Finally, the networks have footage of Saddam other than those of his being inspected for lice!  Freshly cured of his lice infections, Saddam appeared this morning before an Iraqi court.  The big bushy beard is gone, replaced by a thin white beard.  Those present reported that he remained defiant, claiming that he was still President of Iraq and that the trial was nothing more than theater.  If it is indeed theater then I believe we can call this play â€œThe Execution of the Dictator.â€  If anyone can come up with some snappier titles, let me know!  Matt Drudge, from The Drudge Report, often takes small rumors and blows them up into major headlines on his website.  Therefore, take this news with a grain of salt, but he reported yesterday that a â€œhigh level sourceâ€ believes that John Kerry is very seriously considering naming Hillary Clinton as his running mate.   urlLink The article can be read here.   I doubt this will happen, but if it does I say it would be very good news for Kerry.  Yes, Hillary is a polarizing figure but the Democratic base loves her.  In 2000, New Yorkers stood on lines that stretched outside of the polling site, around the block, just to vote for her.  With the country split 50/50, the election will be determined on turnout.  If Hillary gets Democrats into the polls in droves, it can only help Kerry.  Those who hate Hillary are probably likely to vote for Bush anyway.  Plus, naming Hillary would give Kerryâ€™s campaign a lot more media attention that it otherwise would.  That being said, heâ€™ll probably pick someone safer and someone who wonâ€™t eclipse him in popularity.  The names John Edwards, Richard Gephardt and Tom Vilsack are named as likely candidates.  Although each can help deliver crucial states to Kerry, none particularly thrill me.  An official announcement is expected on Tuesday.  One of my co-workers is visiting the new Museum of Sex in Manhattan this weekend.  She is excited about some Asian exhibit.  I told her to let me know if they had any â€œhands onâ€ exhibits.   '),(22255,2824507,'11,August,2004','2004-08-11 00:00:00','    The Blogger Party. Coming soon, to a condo or apartment near you. Check out the webpage and particularly the comments section of  urlLink The Anonymous Blogger  for details as they develop.  My team, \"Fay Wray Dead â€“ King Kong Arrested,\" came in 4th place last night at Trivia. Disappointing because we were tied for 1st at the intermission. But we were sunk at the final visual category where we had to identify obscure character actors. Perhaps if my teammates werenâ€™t so busy passing junior high school notes! Val did a good job with hosting and sheâ€™ll be back again to co-host next week.  Poor Tom Glavine. Not only is he stuck on a losing baseball team that always fails to give him run support, but yesterday he was involved in a car accident that caused him to lose his two front teeth. He was riding in a cab and an SUV smashed into them. Fortunately noone was seriously hurt. Amusingly, the cabbie complained that while all the police and medics attended to the celebrity Glavine, noone went to check on him until afterwards.   '),(22256,2824507,'10,August,2004','2004-08-10 00:00:00','    What is wrong with Jason Giambi? The Yankee first baseman came to spring training carrying far less weight and mass than usual. It was whispered that he was a user of performance enhancing drugs and his reduced size resulted from his no longer using them. Then he goes on the disabled list with an intestinal parasite. It was assumed that he ate something bad in Japan as teammate Kevin Brown also had a parasite. But then, after the antibiotics failed to help, he announced that he has a benign tumor but instead of operating on it, it would be treated with medications. His return to the team is questionable. Although Yankee brass are optimistic that Giambi will return by next month, his friend Eric Chavez said in an interview that after speaking with Giambi, he doubted heâ€™d be back this year at all. So what exactly is wrong with him? When the Yankee G.M. was asked if he knew where the tumor was, he said that he did but that he would not disclose it to the public, in keeping with Giambiâ€™s privacy rights. I completely respect their desire to keep Giambiâ€™s ailment private but the mystery just raises some anxious questions. The Yankees didnâ€™t hesitate to announce David Coneâ€™s aneurysm, Joe Torreâ€™s prostate cancer or Darryl Strawberryâ€™s colon cancer. So, what is wrong with Jason Giambi?   '),(22257,2824507,'09,August,2004','2004-08-09 00:00:00','    Blog reader  urlLink Tex  referred me to this website,  urlLink Election Projection . It analyzes polls from each individual state to determine who would win the Electoral College if the voting were held today. Basically, it all boils down to the swing states of Florida, Ohio, Missouri, Iowa, Pennsylvania and New Hampshire. Whoever wins the majority of those states wins the election.  Over the weekend, the City Council and its speaker released a report slamming the cleanliness of NYCâ€™s beaches, particularly Coney Island. Itâ€™s ridiculous. Iâ€™m at Coney often and can attest as to how clean itâ€™s kept. Parks crews thoroughly clean the beach each night between 9:30 P.M. and 6:00 A.M. During the day, there are thousands of trash cans on the beach. It reeks of political motives for the speaker of the City Council to release this unscientific and unfounded report. Oh yeah, heâ€™s running against the mayor next year.   '),(22258,2824507,'08,August,2004','2004-08-08 00:00:00','    I took a little break from blogging yesterday as it\'s hard to think of things to put on here everyday. Especially on Saturdays when I wake up from a long sleep with my mind operating in slow motion. But Sitemeter tells me that my blog viewership decreases drastically during the weekends so I suppose it\'s no big loss if I skip a day or two.  The weather has been great in NYC this weekend. Unseasonably cool for early August, which is just fine with me. So I went into Manhattan both yesterday and today just to relax and hang out. On the recommendation of Gabe, AKA Chris, and The Anonymous Blogger, I went to see \"Harold & Kumar Go to White Castle.\" Pretty funny stuff! Although it wasn\'t wall-to-wall laughs as my referrers claimed, it was still plenty of laughs to qualify as an enjoyable flick. (My personal favorite scene : The worker at the burger place that replaced White Castle screaming that he was going to burn the place down) I left the theater STARVING, and craving White Castle. But I didn\'t know of any that were around so I ate at Wendy\'s instead. I actually do know of a couple of White Castles in Brooklyn but they\'re in ghetto neighborhoods. So I\'ll stick to the microwaveable White Castle burgers that I can buy in the supermarket.  An excerpt of dialogue from tonight\'s episode of \"Entourage\"... \"Hey Turtle, where\'d you guys meet?\" \"Uh... JDate.\" \"Isn\'t that for Jews?\" \"So?\" \"She\'s wearing a cross.\" \"She\'s reform.\"   '),(22259,2824507,'06,August,2004','2004-08-06 00:00:00','    \"Our enemies are innovative and resourceful, and so are we. They never stop thinking about new ways to harm our country and our people, and neither do we.\" --- George W. Bush, August 5, 2004  I couldnâ€™t make this stuff up if I tried.  My efforts to do more new things continue. Last night I went to a local Democratic club, to visit a friend from my internship days and to introduce myself to the local councilman. It was a cool place but not somewhere where I can see myself spending too many social hours. Basically, the politicians and judges hobnob with each other and the senior citizens stuff envelopes. I might have to look into some social events for  younger  Democrats.  I have a relaxing weekend coming up with nothing planned in particular. Perhaps Iâ€™ll see some movies (i.e. \"White Castle\") or visit some museums that I havenâ€™t been to in a while. Visiting the Met last week has put me back on a museum kick.   '),(22260,2824507,'05,August,2004','2004-08-05 00:00:00','    Two of my favorite topics in recent months have been politics and blogging.  Combining the two is this  hilarious  article that I just saw in \"The Onion\".  I\'ve pasted it below for all to enjoy...   CIA ASKS BUSH TO DISCONTINUE BLOG     WASHINGTON, DC-In the interest of national security, President Bush has been asked to stop posting entries on his three-month-old personal web log, acting CIA director John E. McLaughlin said Monday.  According to McLaughlin, several recent entries on PrezGeorgeW. typepad.com have compromised military operations, while other posts may have seriously undercut the PR efforts of White House press secretary Scott McClellan.  A July 24 posting read, \"Just got back from a lunch with Colin and Adil Moussa (one of Prince Saud al-Faisal\'s guys). Colin wants the Saudis to send some troops to Najaf-so some of the soldiers are Arab, I guess. This Moussa guy sure wears a lot of jewelry. A golden chain, a golden ring with his initials or something, and some other sparkling stuff-kinda effeminate. Anyway, best of luck in Iraq, Iyad.\"  McLaughlin, normally hesitant to express public disapproval of the president, said the blog was \"ill-advised.\"  \"I would hate for the president to inadvertently put American soldiers at risk,\" McLaughlin said. \"We work hard to maintain the integrity of state secrets. When we see the president posting details of troop movements, international counter-terrorism negotiations, and even the nuclear launch codes, as he did on Monday, we have to step up and say something.\"  Bush said he could not understand McLaughlin\'s anger, characterizing his blog as a \"personal thing written for friends and family or whoever\" and therefore \"none of the CIA\'s business.\" Nevertheless, U.S. Secret Service director W. Ralph Basham objected to the blog, as well.  \"He is compromising his safety and the safety of those in my department,\" Basham said, citing a post from last Thursday in which Bush revealed that he \"had to go to some secret meeting with Norquist at some Marriot [sic] over in Virginia.\" \"Someone could uncover some serious state secrets, if they took the time to wade through all of those photos he posted after he got that digital camera in June.\"  On Saturday, Basham asked to pre-screen all blog activity before Bush posts it online. Bush rejected Basham\'s request and later that day wrote in his blog that \"Some people who shall remain nameless apparently do not know there is such a thing as free speech in this country.\"  Members of Bush\'s re-election team have urged the president to exercise caution with his blog, perhaps because of posts like the one dated July 8, 2004: \"Another long day of speeches and fundraisers. Met with all these phony media company execs. Had to promise them some bill next term and shake a lot of stupid hands, but they did bring in two or three million or so. Whatever. Karl keeps a list. I got big laughs during my speech, so I\'m happy.\"  Republican National Committee chairman Ed Gillespie said he spoke to Bush about the blog last month.  \"After he mentioned our Monday message-of-the-day in a Saturday post, we\'ve really been pushing him to not talk about campaign strategy,\" Gillespie said. \"He\'s not that involved in the planning anyway, so it shouldn\'t be too much to ask.\"  \"We\'re not trying to stifle the president\'s creativity,\" Gillespie added. \"We think it\'s great he\'s taking an interest in writing.\"  Bush maintained that he\'s doing nothing wrong.  \"I know so many people, but I\'m way too busy to keep in touch with all of them,\" Bush said. \"Whether I\'m talking about our strategies in Gitmo or my dogs down in Crawford, the blog is an easy way to let everyone know what\'s been up with me. If I\'ve just had a really good lunch at a new restaurant, or something funny happens in a briefing from the NSA, I want to let my friends and family know about it.\"  McLaughlin said it\'s likely that Bush will eventually agree to submit his blog for review by the Secret Service.  \"Right now, the president insists it\'s his right to have it, as long as he doesn\'t work on it during White House work hours,\" McLaughlin said. \"But I believe we\'ll be able to convince him, if we let him calm down. And even if we don\'t, frankly, I can\'t see the blog holding his interest for too long.\"   '),(22261,2824507,'04,August,2004','2004-08-04 00:00:00','    My trivia record is 1-0 after a big win in my debut last night!  Thanks go out to my smart teammates Val, Anonymous, Claire, Eric, chemistry professor and Republican dude in a suit.  When it came to the team name I said that we should make a pun off of Paris Hilton\'s black eye and bruises.  Val then came up with \"Paris is Bruising\" and we won the best team name prize.  Then we won the trivia contest, despite my shih-tzu / lhasa apso mixup, which my team took great delight in berating me for.  I remind them that Claire and I gave us all of our points in the Liberal Arts College question!  A fun time was had and props go out to Sarah and Caren for co-hosting (even if Sarah didn\'t know that the 9/11 Commission Report was paperback)  I noticed a few cute girls at several of the tables so I just might have to go back!  Bob Murphy sadly passed away yesterday.  He was the radio voice of the Mets from 1962-2003.  He was a pleasure to listen to and will be missed.  Bob, here\'s to many \"happy recaps!\"   '),(22262,2824507,'03,August,2004','2004-08-03 00:00:00','    Tonight Sarah and Caren will be hosting Tuesday Night Trivia at the Baggot Inn. Iâ€™ve been hearing about this event for months but have yet to go... until now! With these two co-hosting, itâ€™s got to be fun. I just hope Iâ€™ll be able to harness all of the useless knowledge Iâ€™ve accumulated and answer some of their questions.  The rest of this blog posting will be me bitching about the utter downfall of the Mets. Fred Wilpon said that his goal for the team was \"to play meaningful games in September.\" Well, the best they could do was play meaningful games in July. A winning record at the All-Star break, only two games out of first, has dissolved into a record of six games under .500 and nine whole games behind the first place Atlanta Braves. Sigh, why should this year be any different? The players may change but the results stay the same. In better news, my Fantasy team is making a comeback. Iâ€™ve moved up from 7th to 4th place. Even if I donâ€™t win, Iâ€™m definitely in contention for the \"Most Improved\" prize.  I\'m currently eating the best tasting onion roll I\'ve ever had...   '),(22263,2824507,'02,August,2004','2004-08-02 00:00:00','    I forgot to mention yesterday that upon leaving the theater where I saw \"The Village,\" a Loews staffer was handing out free samples of Advil. To which I said to him, \"Hey, the movie wasnâ€™t THAT painful!\" All around laughed at my ingenious one-liner.  With the Democratic Convention all said and done with, I thought Iâ€™d offer my list of the Top 10 Speeches. They are...   1. Barack Obama   2. John Kerry   3. Bill Clinton   4. John Edwards   5. Wesley Clark   6. Jimmy Carter   7. Al Sharpton   8. Ron Reagan   9. Max Cleland 10. Teresa Heinz Kerry  Another hot and humid day in NYC. The dog days of August have begun. Woof.   '),(22264,2824507,'01,August,2004','2004-08-01 00:00:00','    The plan for today was to see \"Much Ado About Nothing\" performed in Central Park.  Unfortunately, by the time Jess and I got there, all of the tickets for this evening\'s performance were gone.  Sadness.  But Jess will go back later this week at an earlier hour and hopefully we\'ll see it before it closes!  Instead we went to the Metropolitan Museum of Art and we saw \"The Village.\"  The film was certainly different.  And interesting.  I\'m not so sure how good a film it was but it sure was... interesting.  For that reason alone, I\'d recommend it.   '),(22265,664485,'31,July,2002','2002-07-31 00:00:00','   urlLink    urlLink    urlLink    urlLink    urlLink     I had a really good cry tonight.   I’m not really sure what exactly it was about, just a culmination of things I suppose. I miss grandma, I don’t know, there are so many things to it, I want her to be okay but I also want to be okay if she needs to leave as well. I just wish that sometimes, I could say more things, or do more things…I know she’s doing better for now but doing better is just being okay ‘under the circumstances’. It hurts because I don’t know, it just does. So many times I just want to reach out and fix everything.  Then theres him and us and tonight. I really hurt him the other night, I know what our conversation did to us, created that space, because we haven’t really talked since then. I miss our uncomplicated friendship. we still go out but now when we’re alone there’s this dreadful silence that suffocates the air, everytime we have a conversation there are so many unspoken words and so much tension and it is so hard because I really want us to be good but it’s just difficult right now. I need to be complete on my own before I get into this. I don’t want to need him or lay it all on him you know? there’s that whole thing about me not being completely his but beyond that, I want my heart and soul to be in us and only us before anything goes on. I don’t want to risk having us break down and break apart because there were things from the past or things we needed to do. I think too much for my own good but everytime he makes promises there’s a nagging feeling that asks till when? Or how long? Or what happens?  I reread my france-singapore journal. I read the things I went through and the hopes I had. The things I wanted and needed and the conclusions/solutions I had penned out. It hurt more because I’m right where I started again, discouraged. Exhausted by everything and letting things get to me.   I might have upset jared too. I seem to be good at that. I don’t even know what I want anymore. I feel half here and half there, half able to trust him and half not. I am torn because I remembered how he told me he’d be here when it ended, it hasn’t ended but he’s left already. Sometimes, confusion really gets the best of me for I am stuck in between not understanding what he wants and knowing that it is an emotional rollercoaster. I have known now that my constant violent swings from different extremes breaks us but I don’t want to learn to depend on him. I need to undo this comfort he brings. I’ve moved on but my foot is only few steps out this half opened door that at times seem so easy to just turn around and walk into. I wish things were as easy as just saying lets try again or let’s break apart. There are no more agains just the aparts but the aparts come and go. It is hard to still need someone, it is even harder to try and suppress that and be afraid of needing him. Should we attempt to challenge the hands of fate and form a bond of friendship? Or do we just leave this at the door and close this chapter and all the wounds that come with it unhealed. They both seem so easy on paper, so easily said than done. some days just good, some days just bad. How come there isn’t the in between…or rather how come there is the in between to realize how much you miss or how much you’re missing.   I am convinced that I’m complicating things more than they really seem to be. I think it comes with the uprooting. I have too many pros and cons to consider that the time I’m given allows me too overanalyse and beat myself up about it. I’m booked in to see grandma next week too, I think that that might have a lot to do with it. I don’t know what to say. I don’t want to say bye. I don’t want to leave knowing that it might be the last time I’ll see her again.   I feel untouchable and invincible. Like no one can touch and as if I’ve reached a new realm of understanding. Yet why do I still get broken late at night sometimes? Why is there still so much of debris left?  I am staring at my new chapter in the eye. I am watching this beginning unfold and am elated with the results yet there is a but in it all. The but of knowing that things might crash and fall at any time and the knowing that I have the ability to do so. It is the insecurity that bites my ass. it is the insecurity I need to conquer.   I should really get to bed now   urlLink  “If I had it all, you know I\'d fuck it up”  '),(22266,664485,'30,July,2002','2002-07-30 00:00:00','   jeezums, it\'s only half past eight but i think that if i just lie down i\'ll be knocked out for the next ten hours. or maybe twelve. yes, it has been that kinda day. everything and anything coming at you but all good honey! :)  the kids had fun though so that\'s all i really care about...i\'m tired as shit and need to get all this paint off me (i reckon i got painted more on me than on the shirts) and just soak my aching body for a long while but janice leaves tommorow so we\'re going out tonight. pah.   and i think if i have to eat one more pack of nasi lemak i\'m gonna puke.   blech.  we\'ve been on this eat malaysian be malaysian drive trying to somehow absorb in everything so that when we go back we won\'t miss it (fat chance) but yeah, so in the last few weeks i\'ve eaten all bloody types of nasi lemak. scary.   and i don\'t even like rice!   madness.   in other news though, my cdr works again! yay!!  :)  will do this later when i get back...hugs.  '),(22267,664485,'30,July,2002','2002-07-30 00:00:00','   But because gramps is doing really good and i\'m talkative and added that i like the feel of my keyboard clicking at this hour I shall tell you about today’s other half minus the workshop incident and parking lot.   Be warned though, it is as irritating and whiny. (yes, /you/ may fuck off now)  See, drew works at inane hours and so sometimes we all meet up at the mall(the gym’s in the mall) after the guys sweat it out and do all their shit in the morning, and so today was one of them. And see, because I terminated my membership to go back and am much too lazy to be bouncing off walls at eight in the morning, I normally just wait for them and read a book or journal at the coffee place.  So anyway, hooi ying, our normal coffee girl went someplace somewhere and got replaced by hot ass dude with the most flawless skin. Damn.    I have to add here too that jeff is retarded and desperate and has a tendency to go for exotic cocky condescending bastards. (can you tell I didn’t like “caleb”?) so anyway, the moment he steps in and sees hot ass our WHOLE conversation is dominated by the texture and size and suppleness. I kid you not.   Then there’s drew, who is a love em and leave em type and has chicks spurting out of his ass but hooi ying is Hot. Fair, thin, fragile looking, you know that type and so drew’s all depressed and upset and jeff is so fucking horny and chris is just being a total asshole. Me? I’m just plotting on ways to kill brainless drivers.   But anyway, it finally reached a point where there were three and a half different conversations going on at once and it was so stupid because we were really just talking to ourselves out loud in a group. And we’re all in our own little worlds when someone finally realizes that there’s a fucking cat underneath the bloody table.  I don’t like cats. I don’t like furry ugly things. And there’s a huge big difference between kittens and cats. Cats are ugly and mean, kittens are not.   But anyway, this is a mall. We’re in the inside section and so how in the world it got there, I don’t know. And it’s a stray I think because it looked nothing like the ones you see at pet wonderland that’s in the mall as well….but guess who found it.  Fucking jeff shrieks. Jumps up and drops his ciggie on my fucking toes. (note to self: do not wear open toed sandals in vicinity of smokers freaking out). So I scream and we look like bloody clowns because drew knocks over his coffee and so we have coffee all over, shitty dirty mean looking cat ready to draw some blood and burnt toes in an obsolete corner with see through glass. Then as if we don’t draw enough of attention, there’s construction work going on right next door.   Needless to say, hot ass had a good laugh.  He did have to bend over to mop up the mess though.   But anyway, that was my day today…now who wants to go romp around the parking lot and leave some sarcastic ranty notes, get arrested and have coffee tomorrow?  No takers?   Damn, you’re no fun.   Hmph.  :P   urlLink  “Who needs $150 an hour therapy when you have me? Thank me later. Tell your dog to thank me, too.”  '),(22268,664485,'29,July,2002','2002-07-29 00:00:00','   See, I’m not a guy so I obviously do not understand the obsession with upgrades and what nots and horse power bullshit so please spare me and do not attempt to explain it to me because it comes out as gibberish. Get it? Got it? Good.  Because my mechanic does not.   My reverse sensors have a mind of it’s own because they seem to beep at the weirdest times, there can be an object ten miles away and the bloody thing will start beeping. This can get awful frustrating and ass painfully irritating at traffic lights and knowing kl roads, there is no place you can go without six millions cars around you. fanfuckingtastic.   So I go in to get it fixed and this dude starts telling me about something or another and blah blah blah blah blah blah blah blah blah and he won’t fucking shut up. I’m late and it only takes about fifteen minutes to readjust the wires. I spent half a bloody hour listening to his nonsense. Gahhhhhhh.  Screw the independence. The next time I’m just going to let whoever take it in.   And still onto cars…:0  All you bastards that park in the disabled spots should be castrated, shot, wringed and put through the steamrollers. Six thousand and seventy two times.   I was very very tempted to write mean sarcastic notes to the assholes today or perhaps accidentally bang their cheap ass sardine can cars but then I was running late so perhaps I will do that tommorow. In any case though, I hope some day you’re crippled and you have to park 10, 000 spots away from the elevators just so you get a taste of your own medicine.   Meanies.   Man, I have something stuck up my ass…and that just came out totally wrong.  I think all this abstinence has made me horny, I humbly apologize….but it keeps the hits coming doesn’t it?  Boobies! Boobies! Boobies! (thank you evan)  Okay, I shall be returned to my den now.    urlLink “How the hell could someone not like boobies? Stop the planet and let me off. Christ.”  '),(22269,664485,'29,July,2002','2002-07-29 00:00:00','   our excuse for humans.    i /really/ wasn\'t going to give a flying fuck to come back here and explain things but seeing as i wrote that i would go to sleep three hours ago and am still running up and down, left right and center and with a shit ass tired and easily irritated mind, i am going to explain myself.    number one.  the last i checked, it was *my* site and *my* oppinions and *my* fucking love life. fuck off if you don\'t want to hear it.   number two.  i have never claimed to be a godly christian. i believe in god, we have a good arrangement and lets keep my dealings with him private okay? fornifuckcation is none of your bloody business and if i go to hell, glad to know i\'ll see you there.    number three.  my take on the whole exes at my wedding thing was merely me trying to make a point backing my reason that it is good to stay single, to narrow down options, to be choosy and picky and anal and not just bang any tom dick and harry. i did not in any way mean to offend whoever i mentioned, you don\'t even bloody know who they are and if they are upset because i just listed our affairs and our dealings out into the bad wide world of the internet than i repeat, OUR business.      number four.  i am not ready for a relationship. what\'s it to you?   number five.  yes, i am a drama queen, i have never denied that but no, everything i said i meant.    number six.  i have never understood people like you. i thank god i don\'t because i\'m far too nice to make voodoo cutouts of you and stab you right where it hurts the most. it was an excellent idea though.   and lastly, because i\'m \"whining little fuck\", this is my place. my heart and soul. i\'m sorry you don\'t like it and i\'m sorry you find it so damn difficult to muster it down and if so, then why bother? save me the trouble and the bastards at hotmail from deleting my other far more important mail okay? i am not even going to begin reasoning with you, i think we\'re beyond that now, neither am i going to apologize for what i write. i write what i write and at least i have a life. you need one too.   and now that that\'s off my chest, i\'m going to bed. yes, naked.  '),(22270,664485,'28,July,2002','2002-07-28 00:00:00','   urlLink     forgive me if I blubber and stumble.   In between of all the bullshit that’s been going on in the last week, someone made me think of it last night. Someone special. Someone I gave a part of me to, someone I know will have that part of me for as long as I live. Someone irreplaceable.  I imagine my wedding day, I imagine The One and I pledging eternity and till death do us part but with us are the others, the exes of our past, the people we once promised forever to. And then I realize and it hits me that when I am to be married to him that my heart and perhaps, neither is his completely ours. That hurts. For everytime you get into a serious emotional relationship, you give a part of your heart away.  I have no regrets for the pieces that are etched in me of them for it still brings smiles and comfort some days but it is hard nonetheless to know that I am not pure, not complete, not truly a hundred percent only his and his alone.   Abstinence. I have never really practiced it, it makes me tainted and dirty. Not a reflection of god, not a good example, perhaps even, not wanted. But then in society and the generation that I have grown up in it has become so acceptable, so easy, so okay. I have always believed in sleeping with only people you love, trusting someone enough to reach that new level. There are benefits but what can mar it all is when sex is used as a bait rather than a way of affection.   I’ve had lovers who were in the relationship for the sex and used sex as an object. Violated that trust. Actually, I’ve only had one lover like that. He broke my heart but he’s still in my life. It is like that itch you can’t get rid of. Like eating fast food and that instant gratification and it feeling so good yet consciously, bad for you. he’s left me high and dry, he’s left me more times than I care to remember but yet each and every single time he’s begged I’ve let him in again. I’ve been hurt and hurt continuosly but yet some faulty mechanism in me caves in every time he needs me. I am a game player, I am emotionally mental and highly indecisive, I am your version of the worst girlfriend you can ever get yet he is exactly like me. perhaps that’s what attracted us to each other, your novice game player, the challenge. the sucking up what we needed from each other selfishly.  I’ve known him for much of my whole actual dating life and something tells me he will never leave.   I’ve been the bitch and I’ve been the mistress. I’ve been the one being cheated on and I’ve been the one that’s dumped. I’ve been the one that dumped and I’ve been the one that left. I’ve been all of those things…have they made me wiser? Maybe, maybe not.   I met him when I was sixteen. Seemed like a young age but back then, everything accelerated. When you were twelve you were going on fifteen, when you were sixteen you were going on eighteen and when you are eighteen, you’re going on twenty one. Perhaps that explains why I’ve never dated someone younger than I am, that or I could give you the crap about how women mature faster than men.   He was everything I’d ever imagined in a man. The One you know. Yet we had gotten out of volatile relationships and were still being constantly drawn back into them. He was cute and he was charming and sweet and funny and kind. We flirted, we talked, we shared but it wasn’t till much much later that things started becoming concrete. Sure there were lets have dinner together nights and sure there were moonlight walks, stolen kisses and conversations that never ended but something always distracted us or pulled us apart. Then when we finally did get together it was magical. Surreal. It was blissful and perfect and so amazing. I swear the flowers were livelier and the skies bluer and everything was just so divine and heavenly and everything was just nice. We reached our comfort zone but bit by bit we started growing apart. The distance didn’t help yet we tried again and again to make it work. We flew in fortnightly, we wrote long mushy emails talked daily, we left little gifts around, we made plans and tried to work on it but it reached a point where there was just nothing left. We were overexhausted and tired and so it ended. It didn’t become crazy, we just had nothing left in us. It hurt a lot but it ended beautifully…if there was the simplest most honest okay break up, it was that. I slipped out and we didn’t talk much to each other for a long while. I missed him because he was a good friend above everything else.   I’ve found him again and have begun to rediscover him, we know that we cannot work it out being together again…not at that level for we are moving on in our own separate lives. We talk about things with less red tape, with blunt, sometimes suffocating honesty because I have given that to him, I hesitate to think that this honesty might ruin the future. Too much honesty is bad sometimes.   History repeats itself.  I am in a funny place right now. I am right in between swearing off love yet wanting to try it again. I don’t know if I want to go through another relationship and give away again another part of me to someone else for safekeeping. I don’t know if I want him to join that line of befores and should things not work out, the line at my wedding day. It scares me whenever he’s said I love you, it’ll still scare me for a while more. I get funny when he says forever. Because I think about all the forevers that have fallen apart.   I know I cannot get into a relationship right now for the timing is awful yet I do not want to stand here and push away this amazing friendship that we’ve found. How do I keep it at this level of closeness and not let anything get beyond this. How do we keep this relationship strictly platonic? I am not done with my issues from the past, I still wake up to tears some nights and I need to get away from that before I can give my all in this. Will it be too late?   Hope is a strange thing yet hope also confuses a lot of things. I have hope for the future to work things out but so much evidence is laid to know that we might never ever work out but how can I ignore even that little slither of hope? I suffocate it with thoughts and with best laid out plans of all the anger and hurt but I have already let go of most of that in hopes to move on with my life. Those are my overbearing radar signals that emit out the fuck off status. Am I ready to face the world alone without protection?  I have digressed from my original point, I can’t quite remember how I reached this nor really what my specific point in this post was but rather that I don’t want to date just for the sake of dating you know? I don’t want to go out, meet some guy, go out for a few months then end it again, I don’t want to do the game playing, the ping ponging of different extremes and emotions, I don’t want to try try again when it is pointless. I don’t want to shop and not buy.   I want to love and be loved. And I want it forever.   I want there to be no compromise or broken promise for that.   But I know you cannot provide that, and that’s why it’s not alright.  That’s why I need to be alone.     \"Pain is inevitable but misery is optional.”  '),(22271,664485,'28,July,2002','2002-07-28 00:00:00','   urlLink    urlLink    urlLink    urlLink    urlLink    urlLink     you know what i hate? i hate it when i come back stinking of stale ciggarettes and booze. so frustrating because it gets under your skin and the whole way back in my semi intoxicated mind it\'s just chanting, i need a bathe, i need a bathe. i swear i need to soak in peroxide or something.  there were so many fucking people tonight. i don\'t like people. well not right now anyway, and i don\'t like talking to random strangers and i don\'t like being nice to people i don\'t know and holding conversations that are so stupid and repeated and dumb. i also don\'t like dancing in public and swinging my fat ass and being sociable. yet despite all that, i had fun tonight. it was a good distraction, thank you you. and you. for making tonight the first night in a while that i\'ve really laughed and smiled.   i got coerced(conned) into watching sophie ellis baxter perform. actually, i just got forced, period. because this bastard decided to ring and then because i\'m being cranky and frustrated and lazy i don\'t pick up my phone and he doesn\'t leave a message and so the next thing i know i\'m talking with a couple of friends online and he\'s standing at the door telling me to get dressed. the best thing being, he roped in my brother as well so i couldn\'t exactly back out of it. it was fun and loud and noisy and she had one very yummilicious dancer which everyone is convinced is gay because really, aren\'t they all? and to that i say, good for them then.   i also got picked up by a chick tonight. that tickles me, because i had fun with sexy white tube top girl. i questioned her advancement though, it must have been pretty obvious that i was straight yet she very admiringly convinced me to dance with her. yes, me dancing. stage. lights. dancing. i blame the drink. she might have tried to grab my ass a couple of times too. but it\'s cool even if  urlLink you  might be jealous.  :P  they served some moet which was good but i hate moet so i just smiled and nodded as the much too cute waiter dude person handed us drink after drink. it\'s very funny when you\'re pretty sober and you\'re watching your friends get pissed drunk. it\'s even funnier when they start talking nonsense and telling you stories.  i must have been really quiet tonight because pompous ass bitch noticed. well, that was when she wasn\'t snogging every other moving thing that is.   it\'s been a good day but i am so tired i\'m actually falling asleep typing this.   i wanted to blog this to remind myself how i feel right now. peaceful.  haven\'t had that in a while.  thank you all who made today so much better.    \"There is in friendship something of all relations, and something above them all. It is the golden thread that ties the hearts of all the world.\" - John Evelyn  '),(22272,664485,'26,July,2002','2002-07-26 00:00:00','   blogathon.   sorry i got too caught up in my shit to support you guys. i promise you that i will run through all the archives once everything settles donw..soon.  you\'re  beautiful, talented  and  loved.   thank you  urlLink michele ,  urlLink joy  and  urlLink jasmine  for doing what you do.   hugs  '),(22273,664485,'26,July,2002','2002-07-26 00:00:00','   I had to song lead for Sunday worship today, so last night I had to go to church for band practice and run through the whole thing for the third service today. Songs I had chosen to be sung last week seemed so inapplicable, I wish I wrote right now that somehow god reached out his hand and touched me and everything was alright again but it wasn’t because last night when I came home I was more miserable and depressed then ever.   I wanted to sing off key and tell them to bang on the drums harder and break the guitar and run around screaming but I ran through the motions, ignoring all that was within and concentrated on putting on a good show. People stopped to talk, to ask how things were, to hug, to pray, to give in words of comfort but they were just merely what they were, just words.  Then because I have to (start) keeping my promises, I’d promised the kids last week that we’d paint t-shirts for the kids at Cares. I have to do t-shirt painting with the cancer kids next week at the playroom and so the other kids that are on the bed don’t get anything and so the kids at the shelter want to paint them shirts as well. If that’s too complicating, I basically had to face up to the kids yesterday, something I really did not feel like doing at all.   …I felt like crawling beneath the sheets and blocking out everything and anything and going to sleep or crying my fat ass to death. But I didn’t…couldn’t. had a life to live, had promises to keep, had more things to do and people on my back.   I spent too long at the shelter than felt even sorrier for the world and then because everyone’s so fucking worried no one let me go home till I had about all I could really take from anyone. I think there was about 15 seconds in between when i would have really lost it and just burst into tears.   I finally came home at three in the morning and chris wanted to stay over and keep me sane for the night but I couldn’t take anything from anyone anymore. I love him to death but it was just so suffocating because everyone hounds you and questions you and I’m okay and everything was just difficult. I hate it when people see me at a weak moment, and I hate being so exposed and so…I don’t know. just too much attention. I don’t want to need someone you know? I just don’t want to do that whole thing again. It hurts too much.   Ironically, I did need him last night. I needed him to take it all away and let him just catch me but I know he’s not mine anymore so it’s okay. I’m still okay. I’m still alright. I’m just angry at the world right now. angry at myself. angry because this should be a new day.  '),(22274,664485,'26,July,2002','2002-07-26 00:00:00','   My grandma went into respiratory arrest this morning. I don’t know how I feel exactly at this moment, I think I am numb, I understand what has happened but I haven’t accepted it. I feel disconnected, strange, like is she really my gramps? Is that grandma?   She’s in the icu right now. my parents flew in straight away, I’m not allowed to go. I want to scream but I can’t, I’m not sure if I want to see her the way she is now. does that make me selfish? I’m sorry, I don’t know, I don’t know how to make sense of this.   I’m trying not to cry, I’m trying to be strong. I have to be strong but I’m so afraid. I don’t want her to die. I just…I don’t know. I love her. It’s so hard. I have pictured this a million times but now as it happens every part of me is so lost.    I hope she’s okay…please be okay.  '),(22275,664485,'24,July,2002','2002-07-24 00:00:00','   faceless.    urlLink    urlLink    urlLink    urlLink    urlLink     appearance and reality...i find it all too easy and available to simply mask and be... the person i am is multi layered, exposing the different textures in me that can exist. does that make me shizo? does that make me assholic? does that make me indecisive? perhaps so, but it also makes me comfortable.   yet there are times like today, when i look and see and laugh cynically yet ironically at the person i am on the outside and inside. i laugh because it is funny that things can be misinterpreted so and judgement crosses the path i walk on already molded and formed. i cracked up rereading  urlLink my testimonial  today, i wonder if perhaps she had seen me inside if everything i have ever achieved mattered. things look too perfect on the outside, too put together. she writes only a couple each year, specifically for the over achievers. how in the hell am i one of them?  then i laugh at me. i laugh because i am a walking talking breathing joke. such a fucking farce...yet i am not angry anymore. i was bitter, angry, and assholic before about having too many facades and having too many faces, i can only accept them now as how i am. this is my version of honesty.  a friend told me once that as long as i was comfortable in my skin, i could be whoever and whatever i wanted to be. selfishness sometimes too is an important part of survival. i take the brickbats, just like i take the backstabs, i take the compliments just like i take the criticism, i take days and hours and weeks away from reality, just like i take those times manifesting in it. i take what i can get for it is me.  i am not sorry that i was upset yesterday, for it made me a better person today. so often we focus on what\'s easily available, easily conquered, choosing simple over better.   i challenge you to choose difficult. i challenge to change something in you that you\'ve needed to change a while. i challenge you to climb that mountain, cross that sea, swim that ocean. i challenge you to be yourself. just you.   fuck whoever and whatever else. if nobody loves you for just you than there is no point at all.  no, this is not a i\'m mad at him because he wants me to be someone i\'m not retaliation but rather a i\'m me and i can only be me and i will be me no matter how many faces i put up and how many different walks i\'ll go though for in the end it is just me and so you can only love me for me and not love me but want me to be someone else type of thing.   if that made sense.   :P  nice long understanding working out talks are good.   in other news though, i only had one extra jab today so it was nice..and the kids are doing really really well and we\'re making shirts next week! :)  you doing good?   urlLink  \"But she felt she truly saw some of his depth that night. She understood so much more. Things she once thought were frivolous suddenly made sense to her. And that was the moment she knew what he could do for her. He could love her, comfort her, keep her safe & warm.... but more than anything he could show her how to feel alive again. \"  '),(22276,664485,'23,July,2002','2002-07-23 00:00:00','   empty.    urlLink    urlLink    urlLink    urlLink    urlLink      \"no one said it was going to be easy, no one said you won\'t wake up on days where you will feel like screaming your head off because it\'s so frustrating and no one said there won\'t be days where everything will hurt so bad and so you\'ll just cry and cry and cry somemore. no one said it was going to be easy. but there is hope. and hope is beautiful.\"   i took a walk alone tonight. just me. it was dark outside, darker than usual, gloomy, i was gloomy. i was down, for some reason. just not in my best spirits. the street lights were out, the moon looked terrible. i walked fast fast faster sometimes, and then others slow slow slowing stopping stride. only it wasn\'t really a stride. it was a just one more step, just a couple of streets more, just a while longer..you\'re okay ones.  i took pictures. nothing came out right..it frustrated me, so i walked somemore.   i looked and peeped into other people\'s lives. i watched the kids play. i smelled the staleness in the air. i passed by the candles lit up for the funeral on the next street. i looked up to the heavens, the clouds, my most favorite thing in the world but there was nothing. only plain cloudless skies. only a half heartedly shining moon. only an empty street.   i looked inside. i searched for the person i was inside but i was empty, i was tired. i was drained from all of today\'s going ons. i was hurting and holding back the tears that threatened to flow. i was holding on and keeping strong, humming my desolate tune. breaking me it wailed...breaking me.   upload the pictures. write another post. go on away.  i have emails to write. phone calls to make. a bathe to take. things to do.  but for this very moment, i am just curled up in my little corner letting the tears free.  i cannot be strong tonight.  i am only me.   just stupid retarded me..and i am hurting.   just having a bad day.    urlLink \"If I am not free to fail, I\'m not free to take risks, and everything in life that\'s worth doing involves a willingness to take a risk and involves the risk of failure...I have to try, but I do not have to succeed.\"  '),(22277,664485,'22,July,2002','2002-07-22 00:00:00','   someone said drama?    urlLink     Because I know you like crazy ass maid stories and laughing at me is always good for you :P   So crazy maid two decided to poison my mom today. With her spit.  Absolutely disgustingly abhorringly revolting. * blech *  See, crazy maid two is a bit senile and retarded and so today, when my mom told her that she did not need to go grocery shopping with nice maid one, she flipped…vehemently insisting that it was of national crisis that she went. My mom of course told her that it was not and that she should just continue doing whatever maidly things she was doing and obey as she was told because that’s what a good maid does right? She listens to her commands. (I feel like god now btw :P)  Then because she’s in charge of making breakfast she  brings  slams my mom her coffee and whatever else and my mom, thankfully, notices a gob of congealed phlegm floating ever so serenely on her nice black coffee. My mom looks at it kinda in disbelief and then tells her to pack her bags.  Lol, it would have been funny had it not been spit and had it not been my maid but considering it is my maid, and it was spit, I can only find this funny ten hours later.   Crazy maid starts yelling and screaming and spitting all over the damn place and all this drama begins of course when I naively, sleepily and regretfully come down to the dining room for what I thought would be just another boring stupid day. Picture me, in my nice bright red pants and all my glory, stepping into the room, nonchalantly going to my seat and then this whole explosion of fury and words and body fluids. Oh yes, it was a great morning indeed.   Nice maid then comes over and kind of protects me because at this point crazy maid is really starting to flip and she’s is screaming. And I mean really screaming. Like seven hundred and sixty five decibels screaming about anything and everything in a flurry of English, Malay and Indonesian and just really flipping. And I’m still trying to figure out, what the fuck?  The guards of course then come running down because of all this inhuman screaming that is coming from the house and then uncle ahmad kinda like restrains her with the rest of them. I swear I felt like I was in a fucking movie. Then suddenly, crazy maid changes tactics and starts crying. Bawling. Sobbing. Begging. She goes mmmaaaaammmmmmmmm, then she goes kaakkk rraacchhhheeaaaallll, then she goes ssssiiiiiiirrrrr and I’m all like, Get. Away. From. Me.   Scariness.  Then of course we find out later that she had to see some guy at the market and that she’d been seeing him for a while now and that they’ve been going out together. I look at my bed and I pray to god nothing happened there. Then we also find out that she’s been giving away our things to the other maids and the other dude as well to sell or use and that she’s also been having parties for the rest of the neighborhood maids on Sundays when we’re at church and sneaking whoever and whatever lunch and dinner…I wonder if theirs were laced with spit as well.   But the worst thing is not knowing if all this while I have been eating my maid’s spit. Nice maid does all the cooking and evil crazy maid is in charge of helping her to prepare though. I’m almost glad I throw up what I eat because this suddenly makes me even more sick. Then if I really wanted to mess things up further I could conclude that it was all a conspiracy by the nice maid because we have previously screwed her in her past life and so crazy maid has been framed. Of course then I’d also believe that I saw purple spotted cows today….but then again.   So that’s another crazy stupid evil maid story for you to be stored into the archives of the out to get racheal list.    Madness.    urlLink  \"I\'d slap you but shit splatters.\"  '),(22278,664485,'22,July,2002','2002-07-22 00:00:00','   ready for the world (again).    urlLink    i really should be banned from blogging at this hour.  i think too much for my own good too.  my new found tom friend is cool. he  urlLink writes  pretty damn good.  i made a huge mistake the other night. guilt is definitely not my good friend.   five player scrabble games are cool.  my test results come out on wednesday. i hope i don\'t need another jab or course of something. it is such a pain to be sick.  i am such a whiny little brat.  it is hard to end a game when you are stuck with all vowels.  i have to bribe the literati gods.  pop tarts at two in the morning make me laugh.  i\'m sleepy.  i have a long rant in my head but it won\'t come out.   i really talk too much for my own good.  my dad\'s throwing me a party.  i don\'t really want a party.  sheesh.  and on a more serious note..  i\'ve noticed lately that a lot of bloggers seem to be coming clean. i did that today. i gave an old friend the blog address because she asked if i still had that thing online. i really wasn\'t sure how she would take it, i am anticipating that she might take it really badly, not because of what i\'ve said here but rather the connection of racheal in the real world and racheal online.   i\'ve struggled lately, to really sorta come to bring them together. that comfortable place where i do not need to pretend nor am i insecure with how much i am revealing.    and it. is. Hard.   i wonder if i want to give out this address to people in my real world. my mom used to read my old blog, i think it was too painful for her but today, she asked me again if it was okay if she came and read this one. i haven\'t answered her, i don\'t really think i can prevent her from coming in. it is more of like a trust thing, she knows i will know if she came in here and all i have hanging on it is her word to not come in here and read what i have to say unless i\'m okay with it.  i guess i\'m really sorta overprotective then. because this really is my baby. i designed it, i made it, i pored over it but most of all, everything inside here is a mirror image of me. i know every nook and corner and what it represents. i have found writing to be my best tool of honesty for faces can decieve and lie.   i am slightly shaken and unsure about this but yet the trepiditation(sp?) of being free to be who i am and what i am excites me. i vowed in my old blog to only write for me but then as the numbers increased and as i began to realize that my written word affected people i started censoring myself. i really hope that does not happen again.  i do not want to use my blog to hurt people. i have done it countless other times before. because writing is mostly my safe place and writing it out is how i make sense of the situation, me using the words that formed sentences onto the screen through my mind at that very moment has made a mark and causes a reaction or a significant effect.   i want this to be a medium of self discovery yet i also wish for it to be a platform for the things i believe in. i like meeting new people and the brilliant new and old friendships that have formed and bonded through this medium has never ceased to amaze me. i am thankful and grateful yet there are times where i wish i could turn back the hands of time and begin again.  i can\'t but can only turn over this chapter and rewrite my story.  will you rewrite it with me please?  jared and i talked. it hurt but things are starting to look up again, it\'s ended yet this is a good ending. i think i\'m beginning to like life again...now isn\'t that scary?   urlLink  \"but there\'s nothing wrong with that. its your site.  you can write long stuff if you want.  you can write short stuff if you want.  you can write \'cheese\' in six inch high letters if you want.\"  '),(22279,664485,'21,July,2002','2002-07-21 00:00:00','   self-discovery.    urlLink    urlLink    urlLink    urlLink    urlLink    urlLink      edited.    \"sometimes love is at its best when it digs in its heels and holds tight.\" - William Coleman  '),(22280,664485,'21,July,2002','2002-07-21 00:00:00','   check-mate.    urlLink    urlLink    urlLink      i wonder if everyone at church feels like the sermon was specifically meant for them or if it\'s just application or if the pastor was just good at it?  i.  feel.  like.  typing.  like.  this.  why?  because.  it\'s  irritating.  isn\'t.  it?  okay, i\'ll stop now. :P  i\'m going to laugh.  ha ha ha ha ha.  see, funny.  :P  i think the heat is affecting my mind.  it\'s become hot again. no wait, scrap that, it\'s become very very unbelievably crazily insanely my flesh is burning hot again.  i have a pretty black silk skirt on.   i like it cause it\'s flowy wiggly.  and nice and shiny but not hookery type retarded.  i can\'t remember when i bought it.  good buy though.  yay.  i saw a lovely ring yesterday.  it was only 16, 000 ringgit.  which is only about six months of working.  damn.  dah-yum.     and i don\'t even work anymore.  this is not good.  i think i lost my cell phone too.  because i can\'t find it.  i\'ve rang it but it\'s on silent so i can\'t hear it.  maybe if i put my head to the floor i\'ll feel the vibration.   maybe i left it in another bag.  maybe someone stole it.  maybe it walked away.  maybe, i don\'t know maybe.  it was a nice phone though.  a good phone.  rest in peace little phone.   i\'m bored, can you tell?  the sermon was good though.  tell you about check-mate later.  i\'m too busy kicking jeremy\'s literati ass.  wa ha ha ha ha   :P    urlLink \"purple monkey dishwasher.\"  '),(22281,664485,'20,July,2002','2002-07-20 00:00:00','   brainless post because it beats talking with my heart.     urlLink    urlLink    urlLink    urlLink    urlLink    urlLink    urlLink       \"After you’ve been to bed together for the first time, without the advantage or disadvantage of any prior acquaintance, the other party very often says to you, Tell me about yourself, I want to know all about you...You tell them your story, or as much of your story as time or a fair degree of prudence allows, and they say, Oh, oh, oh, oh, oh, each time a little more faintly until the oh is just an audible breath...Well, one of you falls asleep and the other one does likewise with a lighted cigarette in his mouth, and that’s how people burn to death in hotel rooms.\" - tenessee williams  '),(22282,664485,'19,July,2002','2002-07-19 00:00:00','   you\'re   \"special\"   special.   lately, i\'ve realized that whenever i have a proper meal i feel nauseated after that. i\'m trying to figure out if it\'s just a side effect or if there is really something wrong with me. and so i searched it under  urlLink google  and well apparently, it\'s common. i find that absurd yet it makes sense too though it does make me very cranky and frustrated. because it just means that i can\'t eat whatever and whenever you know? like i have to really ration things. which irritates me and throws my whole thing off balance. but i\'d still rather opt for that any day than looping, which unfortunately, i did today. but then i\'m also trying to defend it by saying that i was already nauseated and so it didn\'t really count because i was sick already. except then how is that different from bullimia when the point was to keep it in?   i read  urlLink this  tonight. it felt really weird.   i don\'t even know if i should be posting this. it\'s scary to think that whoever reads this might know that i have a problem. and it\'s scary to think that judgement will be passed. and it\'s scary to think that someone will analyse me and attempt to figure out or underline my faults and flaws. and it\'s scary to think that i am willingly exposing myself to this.   i can\'t remember how this started. it began a long time ago i think. around the time junoir high school started, when everything great was going even better and then suddenly, like all things, insecurity crept in. i worried too much and thougth about too many things. i\'ve always tried and defended it by saying that i\'m just made like that, to be oversensitive you know? but it\'s bloody destructive, that much i\'ll tell you.  i was aneroxic in the beginning though i didn\'t really know i was at that time. it just sorta fell into place you know? like i had too many activities or things i had to do in a day that i couldn\'t find the time to eat. i\'m not sure if i did it subconsciously or if i really did not know at all. i am thinking that perhaps it was a little of both. it felt great when i started losing weight, because then people commented, and i got more jobs, i was the right look, the right type, in other words, i was another ignorant stupid dumbfuck sucked into an illness, but it was working. then i noticed. and it scared the living shit out of me.  i think it went downhill from there.  when i was young and people talked about being aneroxic or bullimic, i always thought that i\'ll have the upperhand, i\'d always thought it was the stupidest most dumbest thing anyone could ever do. i never understood why or how and what the hell ran through these silly people\'s minds but then i became one of them. and i saw. and understood.   and i saw how at times, no one really cares about who you are inside but rather what you are on the outside. i saw that if i was small and thin that directors hired me because i appealed to what was society\'s take on beautiful. i saw that it was normal to go into an office full of walking talking skeletons, people that encouraged you to throw up because apparently it was good for you. it was a fucked up world...but it was my world.   they sent me to theraphy when it ended up with having a can of diet coke a day. that was my grand meal of the day. but then i could only hide so much..and so i found out that when i came home for dinner, i had to pretend to eat. then throw up after that. so that\'s my introduction to bullimia. because i didn\'t and wouldn\'t and couldn\'t face up to it and especially not with my parents. i didn\'t know if it scared me more to know that they\'d be angry or if it scared me more to know how much they\'d hurt if they knew.   i became so good at lying to the world. it was my craft after all..masking faces. the people that knew were convinced i was healed. inside, it was just an even bigger spiral to hell. friends would ask and cajole and my best way of pretending was doing the exact opposite of what i\'d been doing before. i\'d take the time out and sell my new image to them and completely squashed their doubts and made lunch dates with them and had a balanced meal with dessert and shit and then throw up after that. i threw up and threw up untill i was physically exhausted. i threw up whenever i had a meal. big or small. i didn\'t binge much, just threw up.   a while ago, i realized how i really didn\'t want to do this anymore. and so i\'ve been \'good\' the last couple of months. maybe even a whole year. been clean except when i go for long dinners, or parties, or gatherings. somehow, so much of food still scares me.   i\'ve never admitted this to anyone but i\'m afraid. i\'m afraid if someday i kill myself doing this and i know enough that it is a possibilty. and i\'m afraid of hurting my parents but at times it is really out of my control. like it eats you from inside. the oh my god you fat little bitch.   but then i have good and bad days. so it\'s just ups and downs i tell myself. just looping.   i have to accept who i am but some days that is difficult.   this isn\'t a i\'m depressed entry, nor is it another one of those shitty things racheal likes to talk about. i\'m still doing good you know? i\'m still on this whole road to holy atonement. still in the \'right frame of mind\'.i don\'t know why i wrote this. i think i was feeling guilty. i think i just needed an outlet.   i think i just needed to tell myself that i just had a bad one time only. that it was okay.   i hope i\'m okay.  because it\'d be really stupid to die from bullimia and get through cancer.   excuse me if this disgusts you, i will return hyper racheal tommorow.  hugs   urlLink  \"Today I am still not cured from bullimia. I do not expect that I ever will be. It is an ongoing daily process of increasing self worth and decreasing fear of food, my body and the relationship they share.\"  '),(22283,664485,'18,July,2002','2002-07-18 00:00:00','   love you.    urlLink    urlLink    urlLink    urlLink       dear che che,  just wanted to tell you that i\'ll miss you so much. thank you for being the che che i never had, for all the good, bad and fantastic times. love you.  kisses, mei mei.   she taught me how write my name in chinese. how to speak in a horrible irish accent. helped me design my prom dress. braided my hair and tucked me in at night. she\'s loved me all these years and been there for me in the typical asian subtility(is there such a word?)   she flies back to belfast again tonight. she\'s officially a mrs now. gonna have cute little semi irish babies. gonna eat chips and vinegar again.  went shopping today. says she\'ll fly in from london for my birthday. it\'s my eighteen. thinks i\'m special enough and worth it. makes me want to cry.   miss her already. words are so hard between us. we have to stop crying, can only hug and cry some more..it\'s absurd. she tells me to be good. tells me to take care. i tell her all the things i\'m supposed to tell her. and with that she is gone. we wave and stand there a while, just till after she\'s passed the dreaded customs. she\'s on a flight tonight. alone. waiting to meet her husband again. she is happy yet sad. i am..i don\'t know what i am. i am missing her.  have a safe flight home.  '),(22284,664485,'18,July,2002','2002-07-18 00:00:00','   bodoh.    urlLink    urlLink    urlLink    urlLink    urlLink       urlLink blogger  hates me.   or so i think.  that or it has a very sick sense of humor.  ...or maybe i played too much with my template and it got tired of editing and publishing and editing and publishing and editing and publishing and oh well, you know..  or maybe it\'s just angry with me and i\'ve somehow overloaded it with so much of rubbish html codes and i\'ve somehow stringed together in what seemed like codes but weren\'t really real codes and were just a result of racheal thinking it was a code when it really wasn\'t a code.   but then of course i only came to this conclusion after i changed computers. three times. because heaven forbid should the ever efficient and user friendly blogger have a nervous breakdown.   so i took a diskette, a great big effort in itself because who the hell in this house uses a diskette when you can just burn it into a cdr. well not me said i. and so after looking in between cracks of bookshelves and under my bed and in my drawers and all around the bloody room, i finally found one. a virused one that is. because largely emblazoned on it in tip ex was \'retarded\'. my room is 22x75, and you\'d think that with everything packed up and with all of that bloody space that i\'ll find at least one unvirused diskette right? well apparently, i was wrong. because i ended up going next door and asking for a diskette, which of course brought on the whole cdr issue. again. for the six hundred and seventy second time.    but i\'ll get back to my point again.  so i take my diskette, and copy my files and label them and that\'s all fine and dandy till i bring it to the other computer. everything loads and opens up nicely till stupid  urlLink blogger  gives me an error message. again. of course, any normal person with a brain would have figured out that hey, maybe blogger\'s just acting up. but nooooo, i have to be convinced that is my freaking computer and connection. so what do i do?  i use the laptop.   only that the laptop needs a password.  and i can\'t for the bloody life of me remember the darn thing.   i tried  shit . then i tried  hairy   . then i tried  stupid  and  slut  and  dumb  and  moronic  and  love  and lastly, even  god . i kid you not, at some point in time, those were actual passwords to accounts because every bloody thing needs a god damn password now doesn\'t it? and i can\'t bloody centralize it because see,  al ,  evan ,  jared ,  jasmine ,  kara  and  chris  have the blogger passwords so they can figure out whatever shit i\'ve put in there that isn\'t working. and  bryan ,  tim ,    sakthi ,  chris  and  drew  have my email passwords because they send out emails for me or download them for me when i don\'t have time to or like the other day, change my passwords for me because i really should be banned from sending long annoying stupid emails which i will regret later. don\'t even get me started on the chain emails. and  steve  with  drew  has my work account password because god forbid, should i die or anything, he would know how to handle all the crap in there. then there\'s my phone password which half the world has and knows because..i don\'t actually know how they know but they just seem to. then there\'s the password for the house alarm, and the door alarm, and  urlLink various   urlLink other   urlLink chat   urlLink mediums  and so really if i were to centralize, would not nothing be just mine anymore? but then i have another rant stored for that and so i will continue on with this one...which is that with all these bloody passwords, how the hell am i supposed to remember what i put in for the password on the lap top considering i only use it when i\'m taking a bath and feel like watching a movie, or late at night when my fat ass can\'t walk across the room to switch on the pc, or when i have to use a different line, or when i feel like it. in which i don\'t. because i don\'t know, it just doesn\'t work for me and i hate hate hate hate the mouse and can\'t stand the screen nor the keyboard but yeah, what\'s my fucking password?  so then admist all this i forgot to ring steve back to say that we\'d go for the  urlLink mib  thing tonight. which was dumb and ignorant and insensitive of me because oh, we only planned on going to this thing since last month but well, i forgot. which is no excuse. so why didn\'t he ring instead? because when he did earlier i was asleep thanks to some pills i had to take for the thing last week and so he thought that i needed rest and that i didn\'t feel like going and let me sleep instead. doh.  and so, i still haven\'t watched  urlLink mib , i still can\'t access my template and i am still tired. minus the fact that i\'ve slept for a good portion of today but yeah, whatever!  and so just now, when i sat down to write this, i remembered the bloody thing.   \"dumbass.\"   i think i will shoot me now.   in other news though, i got some very cute pants fr  urlLink tommy  today. and my scanner works again. and i talked to  urlLink koren  at six in the morning because some people don\'t know how to plus eight hours. oh, and i had a popsicle for dinner. which was very very good. like a slushie cum popsicle cum orange tasting syrup thing. :) and i made a  urlLink few   urlLink new   urlLink friends  too. hm...that and  urlLink he\'s  bloody adorable. :P  now, if only i can get this damn template done.    urlLink  \"No, Lisa, you cannot pull that bitch out of her car, run her over, then drive away leaving her on the street. Just beep and give her the finger”  '),(22285,664485,'17,July,2002','2002-07-17 00:00:00','   meshes.    urlLink    urlLink    urlLink    urlLink    urlLink      \"dear you.. it\'s taken you a while to get here hasn\'t it? so much has changed..so many crazy turns have ensued from no where and so i hope this comforts...guides, provides hope if and when you need it.  you need to learn to be more aware and responsible for your actions. you\'ve made mistakes along the way getting here and so you know what i mean when i say every action has a reaction. no doubt, there are things that are sometimes, not up to you at all, that are beyond your capabilities or control but there are others too, that are. and living your life well means taking control of it and having enough of strength and dignity to know that you can get through to the next day with a smile and a pieced together solid heart. it\'s knowing that at the end of the day, come what may you can hold your head up high and be proud of who you are or even just be grateful for that small little thing and not give up.   you have a tendency to believe the things in your mind. you\'ve always been set in your ways..determined little tyrant, you\'re used to getting things done your way and i know that that very brute and obscene strength that puts you where you want to be is also, a lot of times, the reason why you\'re stuck in that hole. the reason why you fall. that balance is precariously positioned that it can only be achieved by one who deserves, by one who does not use her ass to think, by one who is calm and rational. by one who thinks.  your very disabilities can work out to be your attributes if you twist the right key, if you turn the right knob.   you are sensitive. i guess we\'ve always known that. i know things hurt you easily, and i know how easy it is to just fall back into routine. to do the things you\'re used to. to give up halfway because it gets so damn difficult sometimes. i know it\'s hard and at times you are so tired of everything and feel so lost, know that at those times, the father up above keeps you safe and is watching, that he loves you even if you\'re weary, haggard and displeasing to the eye. that he\'ll carry you through footprints in the sand and pick you up when you are down. that itself is the true meaning of love and trust and faith.  every little hurdle, every trial and tribulation, every damn obstacle was put there for a reason...and it\'s an open path. you still have to make choices and you still have to make decisions to decide the very road you will walk on, the very path you will take.   when no one else believes in you, i will. when no one else listens, i will. and when no one else is there, you\'ll still be here. alive.   live it like it\'s meant to be lived. live it like there is no tommorow. live it with no regrets but most importantly, live it for you.   relationships fall apart sometimes. whether it was your fault or his or the both of you, it is time to move on. let go because if that is the only advice i\'ll give, then it is the best i will ever give. let go because nobody starts a painting on a dirty canvas. nobody builds a house on a crappy foundation. and nobody applies make up to a greasy face. this is your platform. this is your life. this is your decision. this is your turn that only you can make. it\'s your time, your moment to decide if you will break and fall, or make it and dance.  i know some parts of you that don\'t understand and perhaps even, don\'t want to anymore..then don\'t hang on to it. you\'ve done a good job the last few days and it is something i can be proud of. because unlike other times where the dirt was just swept under the rug, i bear no grudges nor condemnation but rather bear not anything at all. endings and closure merge together to give in to new skin, new life, to rejuvenate, to rediscover, to relive. to start afresh. new. not forgetting where you\'ve been but merely using it for the better.  live one day at a time, let love rule over hate or anger. let trust rule over being afraid. and know that whatever happens, happens.   take care of yourself and know that in your life, there are many many people behind you, willing to help if you ask. know that while you cloud yourself with the less fulfilling events, that you\'ve been blessed abundantly. and so, it isn\'t really a solution for there is no one specific queastion but merely just a hodge podge of a series of events linked together spurred by actions and strenghtened by humanity.   hold on to humanity spread love if not only because love is how things will fall into places, for you to make love and not war.  if love were only a will to possess, then it is not love; for love is patient, love is kind. It does not envy, it does not boast, it is not proud. It is not rude, it is not self seeking, it is not easily angered, it keeps no record of wrongs. Love does not delight in evil but rejoices with the truth. It always protects, always trusts, always hopes, always perseveres. Love never fails. \"   i wrote this last night before i went out because while waiting, i came across an email i wrote to adrian when things were falling apart for him. and then i figured, if i could tell that to him and want him to believe me when i say that things would be right, how can i turn around and not believe it myself? and so, the above letter was written. the things i knew in my mind but needed to write out. just things filtering through my mind, keeping me sane you know?   and it was strange too, because last night, when we all got together, we were all different and changed and completely just sorta dazed a while because we were almost like strangers. and so, it took a good twenty minutes before the smiles became real and genuine, not put on, before the sync was found and we talked and talked and talked. i had not lost them at all but merely just needed time to know them again. and the thing that struck me the most was when someone said that it was hard to just forget. that we had grown up together, lost together, learnt together and cried together. and though we\'re all in different stages, different parts of the world, that no matter where we are or where we\'ll ever be that these nails in my heart are embedded forever and love is how we\'ll stay together. we\'re all strange and kooky and crazy and silly but it is those things that you hold dear, it is that brainless foot in the mouth responses that you\'ll always remember, it is the irritating continuos slap on the back that comforts you, it is that annoying high pitched cackle that makes you laugh out loud too. and i guess it was really just to tell myself that while a chapter closes that things will still be right if You hold on. if you believe.  and so i do.    urlLink  \"Now she does her best to move forward, still awkwardly clutching to the hurt. She knows she needs to let it go. In time she will. The words she thought she didn\'t have come spilling forth onto the computer screen. She tells the truth, no matter what they think. \"  '),(22286,664485,'17,July,2002','2002-07-17 00:00:00','   take what you want and leave.    urlLink    urlLink    urlLink    urlLink    urlLink    urlLink      via  urlLink carlos .  Blue Skies - Ella Fitzgerald.  blue skies, smiling at me nothing but blue skies do I see.  bluebirds, singing a song nothing but bluebirds all day long.  i was blue, just as blue as I could be ev\'ry day was a cloudy day for me then good luck came a-knocking at my door skies were gray but they\'re not gray anymore. blue skies, smiling at me nothing but blue skies do I see.  bluebirds, singing a song nothing but bluebirds all day long.  never saw the sun shining so bright never saw things going so right noticing the days hurrying by when you\'re in love, my how they fly.  blue days, all of them gone nothing but blue skies from now on..  urlLink ---words and music by Irving Berlin    i have red striped pink toenails. for some reason, they are staring back at me and making me laugh. madness. :D  will this be short lived i hear you ask? how long will good and nice racheal last? i\'m planning to keep her around, she makes things fun and sorta interesting to be with.   i\'m going to go see the kids, then tonight, have dinner with some old friends and i plan on having a good night.   i could twist it all around, and be upset because i\'m leaving. and feel fucked up because at least two kids out of the ward won\'t make it out alive. i could feel sorry for myself because i had a conversation i didn\'t want to have today. i could be pissed with gibson because i have very short hair and it feels funny, and i could be angry at god because everything in the last few months have just been crazy...  or i could be grateful that i\'ve had a chance to be abroad, to see things, to experience things...i could go in there and smile and make them laugh and play with them, and if they go, they go to the greatest place around, and it\'s for so much better, and if for every boo boo they\'ve had, they get a million rewards up above, that while they\'re still around, i\'ve had time with them. i could take my conversation and stick it in the back of my mind, and understand that it just didn\'t work out. and know that i\'ll be alright, and know that it is beyond my control now, that everything that made sense from before still does, i just need to balance it out and see the postive rather than the hurt. if i can place all that energy and time spent brooding over the things i cannot control anymore then it will be right because a closed door just opens up to another world. and learn from this. sure, it\'s another broken relationship, another tear in the heart but oh well, i\'ll live you know?...and about the hair, it\'s been pretty interesting. hair grows anyway and i sorta like it. the whole messy spiky look, gives me another reason to change my wardrobe eh?   and to god, you better bloody have things in control because hell yeah, i\'m relying on you to have it all figured out.   so it\'s just life. :)    urlLink \"I sat stunned for a moment, then continued on my trip, driving much, much slower.\"  '),(22287,664485,'16,July,2002','2002-07-16 00:00:00','   letter to self.   dear stupid side of me, would you please kindly go away and leave me alone tonight?  thank you. yours sincerely, your desperate better half.     urlLink [ bryan+adam = eclectic result of sticking japanese, chinese, malay, english, a guitar and two very silly sods together ]  '),(22288,664485,'16,July,2002','2002-07-16 00:00:00','   urlLink    urlLink    urlLink    urlLink    urlLink    urlLink     I’ve had a nice long shower and couple of days to think about what I wanted to write in here. It isn’t exactly that I have nothing to say, but rather that I have so much I need and want to say cluttered up in my mind and I can only pray that random coherence and words will do me justice. Will do this explanation that I have pushed away for so long some sort of meaning…understanding.  I didn’t exactly want to write it this afternoon because I had dinner with him today and I wasn’t exactly sure how I would feel after that. If the same positive outlook would be replaced with a fuck this whole thing I’m hiding again persona. and I needed to write some emails that I had avoided facing up to…just things that needed to be written out and said and put into words…just things that I needed to do but didn’t know how to. I still don’t, but I tried, and lately, I’ve realized, that’s all that matters. If I try and try again.   We talked...and one of the first real things he said was ‘what did you do to yourself?’ it was that tone, that same tone he used in the hospital when I had taken more pills then I cared to, it was an I care tone and it sorta scared me yet pleased me. I have lived my whole life trying to please someone or another, trying to be things I wasn’t, masking everything I had inside because I was afraid or always told to protect myself…made to protect myself only that this time, I am truly tired and done with it. With that.   This is who I am.  I am racheal sarah. I am eighteen, and I am proud to be eighteen. I am a child, and I am an adult capable of making decisions on my own. I am afraid at times but it doesn’t mean that I will always run, I have an eating disorder, something I am not proud of and something I know, will be fixed in time. I like me, at times, yet there are others, like the last few months where I am in a rut. An indecisive turn. Where I don’t know the roads I must take and fall deeper and deeper, losing myself and losing the things I know I must fight for. Losing my dreams and hopes and aspirations…and that is a very very scary place. Because suddenly, being alone means being responsible for yourself and the things you do to yourself. Suffice to say, I handled it badly, I broke promises and broke the trust of others around me by just being mean. Or cold. Or distant. Provoking a reaction that could hurt. Provoking others around me to leave, to go away, telling them I wanted to be alone…needed to be alone. Wanting to see the person I had become, wanting to see the demons in me I unleashed.   And so I did.   I went away and the first couple of days felt like eternity. The part where your flight takes off and paranoia sets in and suddenly you remember all the things you needed to do and say and worrying if it’s too late now. And you dwell on everything, on every event, and it plays and replays in your mind, that every significant turn and crossroad that made you you. And somewhere in the midst of that, you realize you’re just you. That no one is perfect. That perfection was fairy tale to a child’s eye yet perfection also came with acceptance. And so as days turned into nights you found little things about yourself you never knew and rediscovered others. You found that somewhere deep in you, you still could love and be loved. That somewhere deep in you, you knew you were loved and all you had to do was open your eyes and see.  And so you saw. You saw the hurt you caused when you were hurting, you saw their arms out stretched to you, you saw the people that cared for you just you no matter who you were or what you were, and you saw that it was okay to trust and it was okay to believe in people. That it is okay to believe in yourself because that’s the most important part of it all. To cross that mountain because you know you can. But most of all, it was knowing that you had made a mistake, and that people make mistakes and that you were willing to fix it.  This is my attempt at fixing it. At saying I’m sorry for the person I became. I was going to call this a new beginning but it is really an ending to eight months of ups and downs and in betweens. Of eight months of trying to find myself. Of eight months of giving and taking. The end of this month truly signifies a new beginning. One which I have longed for and thought of for a long long time.  When I didn’t get to nz, I thought it was merely just me being stupid and emotional and using my ass to think but I see now that perhaps I made the right decision after all. That everything I have always said about how time mends all things is true after all. And so here I am, eight months later, having been at places and stages I never would have been had I gone, learning and loving, believing and trusting, letting go and moving on.  I wrote  urlLink this  about a month ago and it made a lot of people question and upset but it was what I felt and what I had written as a source of theraphy. As a way of understanding or at least trying to make sense. It all does now. I know where I’m at and I know where I am heading. I know the things I want and I know the goals I will strive to achieve.  Yet there is a big fat But in it.   It isn’t that I will never have a bad moment again, and it isn’t that I am super confident that everything will work out but rather I am ready for anything that might come and more acceptable to anything that might not. That I need to take this step forward and try my very best to not look back or at least not run back to this hole and cry just because it is comforting and perhaps when I am in here, everything bad is expected and so ironically, it is easier. But what I\'ve realised is that the better person, the person I want and hope to be takes this challenge and faces up to it. Fighting to the last breath and not giving up halfway. Believing in myself and perhaps, trusting in the big guy up there to have things in control and that I can only dictate so much and plan so much for truly it is only up to the hands of fate to decide. The gods above. Destiny.  And so I will run to my destiny with a smile and an open heart...understanding what is not mine to take and accepting the things and lessons and people that envelop me even if it hurts a while.     Here’s to life again, for everything it’s worth.  200 and 2 percent.  hugs.    urlLink \"she broke my heart and ruined my life. hope that answers that.\"  '),(22289,664485,'16,July,2002','2002-07-16 00:00:00','   i guess many people know me better than i do...because much as i REALLY :) tried to stay away..lol, i\'m back. so my so called yet well worth \'intermission\' was short lived by the sudden need to come online and write a long ass post and talk about myself. :P  where the hell have i been?   i don\'t quite know really. so many places and back. so so many things has happened since july the first and it\'s hard to believe that it has only been sixteen days. there was a little scare of a heat stroke in between but after a proper rest and lots of tlc, we were both back on our feet, on the sand and enjoying the breeze. this is all the reasons why i want to live near the beach again. :) but a bay will do i suppose...i am excited, oh wait, make that assholically crazy excited to go back...it\'s been so long and getting in touch with all the people i used to know and talking to them again has sorta quenched a lot of the fears i had. thank you mom and dad.   i did a crazy thing yesterday. i can\'t exactly say it was a spur of the moment thing because i went in there with an appointment and everything and said work your magic! hehe, five painfully long hours later (yes, vanity, thy name is woman blah blah) i came out with a wonderfully textured extremely short cropped streaked two toned hair. as said earlier, photo evidence will come later! be warned though, it\'s  s h o r t !  we picked up the new ford linux?..lynx? whatever the hell it is...c a r yesterday for the mib promotion and it was ass funny because it really just looks like any other normal car untill of course you put one bum at the drivers seat and let him take you to the track and see what his baby can do. i know now all the reasons why a safety belt is so damn important :P   oh, then we also tried to lure  urlLink jasmine  out, dumbfully not realizing that she doesn\'t work in the building anymore. of course it only turned out to be funny for me. but in any case, it was really really good food. i normally don\'t really rant about what i eat y\'know? i\'m  particular  but i don\'t gush over shit ass great tasting food like i\'m about to. and it was cheap! well, relatively anyway...and there is nothing more malaysian and kiasu than having mouth watering lobsters that are cheap and good! heheh...though the picking thing you use to take out the \'succulent\' flesh thing really reminded me of the same thing gibson used to pick out the strands of hair to be streaked from the cap (instead of the alluminum foil way-lah) and so much for that, i had dessert soon after.  :)  other \"big thing\" than being back in this place with a dodgy layout is that i purchased a domain the other night. wa ha ha ha. everything is still sorta under wraps but a hot sexy dude with immense unbelievable graphic capabilities has been manipulated into designing it for me. lol, all i can say is that it looks really good and i\'m piss excited! the space at  urlLink thewarmcompany  from  urlLink chris  will still work as the main ftp-ing place something though...i\'m not too sure about the \'technical\' side of it seeing as i\'m pretty crap with the whole html side of it but basically it\'s a blog cum personal space cum kids update cum friends join thing with some pretty nifty shit and it\'s all good honey! ehhe..but i should just keep my toes crossed and not build castles in the air eh?   and if none of that just made sense, go  urlLink give    urlLink money!    i\'m sorta caffiene and pretty good day induced right now minus the fact that i really really need to pee but can\'t step away because i KNOW that this stupid thing won\'t be here when i\'m back so yeah...but rest assured though, you\'ll get serious more mellow racheal tonight. in other words, i have a nice long essay and shit of the last few days/weeks/months/years all in my mind but have no patience or time or sense to write it now lah!   so i leave you with photos...bad photos. burnt cow, bitches and wet feet photos. big huge asses too.    urlLink    urlLink    urlLink     urlLink    urlLink      urlLink    urlLink    urlLink    urlLink    urlLink      urlLink    urlLink    urlLink    urlLink      urlLink    urlLink    urlLink    urlLink    urlLink    urlLink      urlLink    urlLink          urlLink \"Writing is like religion. Every man who feels the call must work out his own salvation.\"  '),(22290,664485,'21,August,2002','2002-08-21 00:00:00','   don\'t even wnat to talk about the hurt...maybe because we\'ve talked about it so many times it\'s pointless now, maybe because it just makes it more painful, like rubbing salt on the wound.   you said it wasn\'t love. perhaps that stings most of all, as opposed to we lost us, as opposed to we just fell apart, as opposed to just don\'t know each other anymore because then i guess that we never had anything to start with then. you tell me you don\'t need me anymore, you tell me my words don\'t hurt you, and you tell me i\'m all the same, that i havent changed, you say things and you say them and it hurts. it hurts right there, right where tears or anger or fear doesn\'t even come close.  remember when we talked? remember when i found it hard to say words because i was crying so much, remember when you said it was okay for me to trust you again. remember when you said those words i didn\'t want and didn\'t need to hear, remember that night, remember how i told you this would happen but you wouldn\'t listen, well it happened only that what i percieved then is nothing like what it is now. it is a million times worst with your words ringing in my ear and my heart twisting and choking on my sobs and trying so hard to hide it all in.  i really don\'t know you anymore. i almost thought i had us figured out, see, i told you not to tell me things, i told you not to make me need you, why did you make me trust you? why did you make me need you? love? you\'re telling me it doesn\'t exist, you\'re telling me that everything that has happened is a lie, is not real, perhaps you win the big fat ribbon then becuase i don\'t even know what it is anymore. why am i hurting so much and how come i can\'t numb it?  you, you were my best friend. you were my hero, my knight, the person that puts his hand over my heart and heals it, the person that tells me things are going to be okay, the person that makes me laugh when i\'m tired of the world, you were my friend.  now you\'re, i don\'t know who you are anymore. now everytime i think about you it hurts. now, you tell me mean things and you tell me you hate me and you tell me things about myself i wish i never told you, now you make me hate me, now you make me sick of us. now, you make me cry.  now, it all just hurts.  it\'s your birthday today. i didn\'t send anything. didn\'t say anything. your gift sits wrapped and ready to be sent out in the corner. i found your cd and i think that maybe perhaps love went into making that but then maybe not because you hate me now, but then maybe not beause you tell me it was just a phase and that you were just confused and that everything was not as it seemed. i want to break the cds, i\'ve already torn your letters, already deleted every conversation with you, already trashed every email, i hold it and it\'s so hard.   i don\'t know how to get rid of you. how to get you out of my system. i need to forget you and forget we ever existed and forget all of us but then i keep on tripping on myself and falling.   you got upset because i mailed you an email telling you i wanted you in my life. you got upset because i wrote you and you asked me why i was rehashing the past when we could just bury it and start from wherever. because the past was where i felt most secure, because the past is where i ever felt close to you, or knew you, because the past is where i ever felt like you actually liked me, because the past is where i ever felt like i meant something to you or that you actually really wanted me around, before it all changed. before everything changed.  i tell you it\'s different you say it\'s not. i tell you about changing and changes, you tell me it\'s all the same. i try to tell you things happening within me but you will never forgive me, never accept me anymore. so instead you react by telling me i\'ll always be the same, the same disgusting whore, you tell me i\'m just the same and i never change, reminding me your thoughts and takes on me, i don\'t need to read it, i don\'t need to see it, i know it by hard, i know what you think of me, you tell me those are just stupid moments when you\'re angry with me but it\'s what you think..what you still think, what you will always think of me.  it\'s like spears that cut right through you and restab again and again. like icicles that frostbite you. ironically, you\'ve changed me so much.   i don\'t know what i\'m doing here. on your side of the world. i don\'t know why things remind me of you, or why i see you in buildings and faces, i don\'t know why i miss you or that i wish you were here, i don\'t know why the other extreme takes over and i never want to see you, i don\'t know why i don\'t even want to ever step into where you\'re at let alone hold a conversation with you again. i don\'t know why i write things like this, maybe just so i know how i feel and write it out to understand then hopefully forget all about it the next day. maybe i do this to exhaust me so i finally get some sleep.  i miss you jared. i miss us. i said your name.   i love you even if you don\'t love me and never loved me or whatever it was. i care but it hurts so much to care. if i could, i would hold you, just to feel you and know that we were real or that some part of us actually existed. if i could, i would ask you to save me tonight, ask you to not be mean tonight and not say the things you say and just not hurt me tonight, if i could, i would tell you but i can\'t.   i can\'t because you won\'t hear it. i can\'t because everything i say comes out tainted to you. i can\'t because everytime i think of you or look at you or try to tell you something it hurts. because i\'m so so afraid of you coming close and ripping it all apart again, because i\'m afraid if you tell me all those things and take it back again. i can\'t because i know i\'m not worth it, i can\'t becaus ei know you deserve better, i can\'t becaue i know you want me to fuck off.  i lost you.  it hurts a lot.  '),(22291,664485,'18,August,2002','2002-08-18 00:00:00','   end the damn thing already.   i woke up last night and had a good cry. it was tiring and frustrating but it was something i needed to do. it hurt so much but then i suppose that you hurt so much to a point where it\'s just numb and i hope i reach that point soon. this is exactly where we\'ve been before, i don\'t even know why i tried again. i don\'t believe in you or us but most of all, i don\'t believe in me, and that hurts a lot. how oculd you tell me those things and make me feel safe and secure and give me those things to hurt me so bad? why is it you? who are you and what did we have and who am i and why did i ever trust you? damn you, damn you for putting me here again, i was doing fine ..alright..i was doingokay why did you shatter everything?  i am so sick of everything it literally makes me sick. i keel and curl up at the thought of anyone coming this close again. i like this barrier it\'s safer than being scarred alive. i\'m bleeding, i\'m broken, im hurting, fuck you for telling me you\'d never do this. for saying the things you said and doing the things you did. i\'m me, and i\'m human, and i have feelings, and i can break.   i dont want to ever come close to anyone again. i don\'t want to ever tell anyone how i feel or what im feeling or be honest or transparent. i like my layers, i like hiding, i like people not knowing how i feel and people not coming too close to know where it hurts the most. i hate myself, but i don\'t hate me as much as knowing that you wlaked away because i was who i was with you. that when you stripped me away and found me you turned away and told me how disgusting i was. you will never know how much that hurts. you will never know how you shattered everything i ever believed in and stole trust or faith or love from me or whatever the hell it was. i don\'t even know what it was. do you know what it\'s like to not know what it was?  all the other times i turn around and tell you and everyone else that i\'m going to be fine. that i will take whatever shit life hands me and i will and i will tell you this time that i don\'t need you and i don\'t need anyone that i am just fine alone but that i also want to do this just alone. i don\'t want anybody and i don\'t want to need anybody. if that makes me selfish and hateble and bitchy and an asshole to be around with then so what\'s new? call me names, laugh at me, tlak about me, make fun of me, it\'s all the same now, it still hurts but i\'m getting used to this hurt and i just have to ignore it and put it away and i will be fine. it\'s a painful reminder but it keeps me safe and okay for the future. so i don\'t need anyone for the furture and know why.   i don\'t know what possesed me to think that things will be okay. i don\'t know what possesed me ever to think that i could trust you or that i was safe with you because in the end you\'re just like the rest of them. i\'m sorry i let you see this part of me, i\'m sorry i let you see a lot of things and told you a lot of things, i was mean but you were too. i\'m winding up, putting this away, starting again. funny how i am here but i have no desire to be anywhere near you at all, i don\'t want to ever listen to your promises or the things you say because we had our conversation, and we talked but i guess it was just talk to you then. i showed you whatever and you walked away again. your very last sentence hurt so much more than anything in the whole wide world, you\'ll never know why but it cut right there, right where it would hurt the most.  you did know where after all. i wish you never.   and so this is it. the ending for a new beginning. i don\'t need lies, i don\'t need anyone and i certainly don\'t ever need to tell people how i feel anymore. i\'m sorry i\'m taking this down yet not sorry at all. perhaps i\'ll keep a part of me just for me after all, take care of yourselves.  i meant to come back here and start again but i realize i need change and this isn\'t helping the slightest bit. life is a damn bitch but guess i just have to deal with it then.  you\'ll do just fine, that much i know.   whatever really, people don\'t change. they just get fucked up inside then fuck up outside too. and we all know what a huge fuck up i am. god knows, you\'ve told me enough of times. so this is ending it, i moved, i ran away, i packed my bags and left and i\'m never ever stepping in here again, i am never letting myself get vulnerable or ever trusting anyone again, why? because it\'s not worth it. nothing is worth it, that much we all know.   and so it ends.  '),(22292,664485,'05,August,2002','2002-08-05 00:00:00','   he told me to open my eyes and reach for beyond what my mind could ever imagine. i laughed, snickered, smiled our secret smile as if to say, i understood yet i was wary(weary) then took a deep breath, spread my wings and  flew  soared.   it has been so good to be back here (malaysia!). the last few years have inched forward and sped by. multiplying and teaching me so many lessons and things i could never even begin to describe nor tell you. they were like treasures interwoven into one another, each unleashing it\'s own magical powers through the hands of time.   this whole week alone has signified many new beginnings and tied up a vast amount of endings....there are people i will never forget, that have touched my life in their own special ways and have forever marked their place in my heart. i cherish, and am grateful for all the moments and times lost in eternity spent making those. xie xie.  it\'s hard to imagine that a week from now i will be back across oceans to a place i once knew. i am terribly excited, yet just a little intimidated by the roads before me, \"make the right decisions racheal, make the god damn right decisions!\"...but yet before that i will see grandma, very possibly for the last time. it is hard sometimes to say goodbye but it is inevitable. i will miss her just as i will miss a lot of life here too.   take care of yourselves you hear me? i should be back here by the eighteenth or so with a new design(!), stories to tell, retardedness to share and bullshit to phantom. so in the meantime do the chicken dance and go bug these fantastic people...  hugs!      urlLink  +  a small victory  urlLink  +  angela  urlLink  +  austin  urlLink  +  barbtries  urlLink  +  bottled bliss  urlLink  +  brickchick  urlLink  +  brooklynkid  urlLink  +  chris  urlLink  +  closetboy  urlLink  +  cracks in the pavement  urlLink  +  daria  urlLink  +  darkmoth  urlLink  +  devon  urlLink  +  digital kara  urlLink  +  digitalnap  urlLink  +  empress  urlLink  +  evan  urlLink  +  fredshead  urlLink  +  ghostcassette  urlLink  +  gray, looking like an ashtray.  urlLink  +  jared  urlLink  +  jasmine  urlLink  +  jay  urlLink  +  josh  urlLink  +  loveisabitch  urlLink  +  lovesadrug  urlLink  +  manicmoods  urlLink  +  mark spark*  urlLink  +  mighty mitch  urlLink  +  missing pieces  urlLink  +  moby  urlLink  +  mystery mona  urlLink  +  nadablog  urlLink  +  planting impatients  urlLink  +  reality remixed.  urlLink  +  rion  urlLink  +  sandee  urlLink  +  screw me  urlLink  + shaynanagins.  urlLink  +  sze  urlLink  +  s!ren  urlLink  +  subinev  urlLink  +  tim  urlLink  +  tinman  urlLink  +  tom  urlLink  +  torn wide open  urlLink  +  trabaca  urlLink  +  wil wheaton  urlLink  +  wishful thinking      \"the smile on your face is the light in the window that tells people you\'re home..\"  '),(22293,664485,'04,August,2002','2002-08-04 00:00:00','   :)  oh come on, like you really didn\'t know this was going to happen!   yes, i\'m redesigning.  no, i still have a life outside of this.  yes, i will be back...tentatively on the 16th...mmhmmm.  no, i am not bullshitting you.  and lastly yes, my ass is still huge.  '),(22294,664485,'04,August,2002','2002-08-04 00:00:00','   urlLink    urlLink    urlLink    urlLink    urlLink     yes, you\'re shit out of luck because i\'m still here.   i\'m quite afraid that should i play my scratched up album of screaming teen angst and an eclectic mix of the sappiest most pathetic ballads my neighbours might snap. bad. because i\'ve only played it all morning and yesterday night. oops.  and if you look at the time and the fact that it\'s monday then you should have had it all figured out that we didn\'t get to the beach. which i really did sorta expect except that well, i was kinda looking forward to it. i think. (sorry)  in other news though, i think i got a lot of shit settled somewhere in between friday night and three hours ago. it feels like i have felt every extreme since then and i\'m not bitter anymore, just merely unfeeling.   i wrote a lot of things earlier about i felt but then realized that i am at no liberty to talk about it. in fact, i might have already spoken too much.   things are going to okay. and today, that\'s all i really care about.   urlLink  \"the heart believes in something else that the logical, calculating mind knows to be true. that i can\'t carry this forever. that, while you have no spiritual believes at all, and that it may bring you no peace, it must be said. to bring peace to itself. it knows that even though it makes no real difference, that it brings you no closer, and perhaps even pushes you further, that it must come to pass.  you are forgiven.\"  '),(22295,664485,'04,August,2002','2002-08-04 00:00:00','   i am wheezing. it\'s not very serious but it\'s frustrating. very very frustrating.  it\'s like a ten pound fat chicken on my chest restricting my airflow.   i also have a headache. i took an aspirin but all it\'s really done is just dull my senses and made me sleepy.  i tried to go to bed...but lying down just makes it harder to breathe. i am giving it half an hour more before i tell my parents.  this is not going to be fun.   i still want to go out tommorow!  and no way in hell are they going to allow us to drive out to the beach if i\'m sick. damn you acupuncture. you were supposed to make me feel good. argh.   my inhaler is being an ass. i don\'t want to overpuff that i might shake. but then me wheezing here doesn\'t help either. i really don\'t feel like putting together the nebulizer(sp?) either.   fuck you weather.   it\'s all your fault.   so in the mean time, i\'m going to think happy thoughts and ley bryan entertain me while hopefully my berotec will start kicking in.  so fucking frustrating.   gahhh  '),(22296,664485,'03,August,2002','2002-08-03 00:00:00','   urlLink    urlLink    urlLink     so apparently last nights dinner didn\'t go down quite as well as it was supposed to. in a bid not to consume anything too foreign(she suggested pate! *blech*) and strange we stuck with potato lamb and wine.   ...but guess who woke up with sarajevo in her tummy?  ah..but that\'s not the best part..  the best part being that i just realized i\'ve been taking  urlLink banned  substances.  uhuh.   because i went out to get some and the lady gave me a funny look then carefully explained to me that it was bad fr you. oh yeah, like now you tell me.   pah.   crazy druggies.   and now it feels like i am coming down with something thanks to the shitty poluted air.   p/s this morning i woke up in another bed, i fell asleep last night in the guest room somewhere in between the hundredth trip to the bathroom and the telly is bigger in this room so then it was disorientating as i tried to remember all of last nights events, i remembered something vague of an email i received, then snippets of conversations that took place last night, i remembered the city lights and the drunk scantily clad girls, then i remembered posting in the middle of the night last night. amidst some stupor of \'getting the truth\' out. so then i did the next best thing i knew what to do, rush in here to delete it. see, it\'s easier to just sweep things underneath the carpet, get it out for that moment then collect it back the next day..i question if i am using my blog to say things again. everytime people write me they always tell me how insanely honest it is, how fucking private it can get..and i keep on hearing these things, people keep on remarking but i\'m just thinking how i\'ve lost so much of innocence. how my very first blog compared to this has so drastically changed. how very possible too that i might have changed.   some days i think this change is so good. yet some days like today i wonder where that part of me went.   is she still somewhere in me or did i lose her in between all the questioning?  i look at my stats and see all these people. it\'s a nice feeling yet it\'s also really like \'oh fuck, i\'m overexposing\'. sometimes i think i blog to one specific person, sometimes i blog to no one at all, sometimes i blog just because.   it\'s time for a redesign. i am so tired of this layout. i hate it when i have images in my head but they don\'t transcribe on paper(screen.)  i am thinking of a hiatus too. you think, so fucking soon?   i say just till get my bearings right. when i settle down so i don\'t end up posting mad insane stupid posts for the next whole week because i\'ve just realized last night too that it is my last week. and i\'m looking at my schedule and thinking how am i going to fit in all these people, how am i going to tell them all how much they mean to me, how am i going to say goodbye.   i have banned specific people from the airport, we have agreed to say our separate goodbyes, i have might even banned goodbyes.   but the strange thing is, i like the thought of going away and starting again. my ass is itching from all the beautiful sunsets i\'ve missed and my rooooooom, and all my bloody things that have already been sent there, i think about all the people i can\'t wait to meet and going back to school! oh my god have i missed being educated, have i missed learning something substantial that did not have to do with how i feel or emotions, have i missed just doing assignments..doing something with a point and not letting my brains rot away.  damn, i am such a nerd.  but you know what? i am. i am a stupid nerd. a pseudo uber geek because i\'ve been taking lessons. :P  i\'m glad i\'m going home. i\'m glad for this change. i\'m glad i\'m rewriting. i\'m glad there are second chances. i\'m glad i make mistakes so i learn from them.   i\'m glad i\'m going to be okay alone.   actually, i\'m just glad i\'m going to be okay, whether or not alone.   thank you you for making me realize that this morning. thank you for mad insane fucks that force you to rethink and rehash your whole entire life and don\'t let you go back to sleep. thank you for people just behind eery nook and corner.   and now i\'m sappy so i\'ll shut up.    \"When the dream is big enough, the odds don\'t matter.\" - Chuck Yeager  '),(22297,664485,'02,August,2002','2002-08-02 00:00:00','   i want to piss you off so you go away.  i might be very close already, possibly even pushing it.  i don\'t know why, it is just like before, me testing you again.  please break please break.  i am afraid to trust you.  ...please break. please break.  are you broken yet?  will you scream at me now?  will you hurt me so i can bleed?  can i cry for a reason now?  what\'s the fucking point again?  i\'m afraid.  '),(22298,664485,'02,August,2002','2002-08-02 00:00:00','   urlLink     “I do not suppose you’d be interested in my superficial ignorance. In my yet another rant. I do not suppose you would be interested in knowing about my day, or how things went or what I did. Because I’m not either.  I am interested in what I want to do for the rest of my life. I am interested in world peace and conservation. I am interested in the children of the world, of today and tommorow. I am interested in making things right yet the focus is recklessly abandoned, easily wavered. I am interested in fixing things, between us, between me and him, between me.   She’s at that fucking thing again. Suck up your thoughts and reposition somewhere else missus. Get your grubby paws off it and prod another.  Take it all in but don’t let it out.”   We went to the short film thing today. I was almost afraid I’d bump into you, but then I knew you knew that I’d be there tonight, and I knew you had the sense to not go tonight either.   Tell me again why this space upsets me. It was what I wanted wasn’t it?   Tell me again why I’m upset, tell me again why I can’t decide….tell me again why I don’t want you but I miss having you around.  Went for drinks at sangria, just below. Did what we were suppose to do last night, did everything we were supposed to do last night only that it wasn’t with you. Because last night we talked again, and last night you decided that it would end here and now but yet today you came back in again.    You’re yelling. Only that it’s a silent yell. It is the kind that gets beneath your skin and penetrates your mind.  Your click on my voice box hurts. It would have been easier had you just slammed the phone down, your terse voice that tells me to have a good day is bitter. You are angry. I know why but I am angry too. You know how I feel about it, you know the problem, you know how it all works…then tell me why this is still upsetting?  I know you understand. I know you know why and I know you know how. Believe me when I say I’m trying to work it out too…believe me when I say I’d like to know what the hell I want either.    In betweens.   Just the cracks.  Smoothen it out please.    urlLink  \"almost everything in life is easier to get into, than out of.\"  '),(22299,664485,'02,August,2002','2002-08-02 00:00:00','   urlLink    \" Difficulty Explaining. You know you can\'t ask me a close ended question anymore? I mean like you can\'t ask me questions that require me to give you a definite answer  Because I won\'t be able to I think that sometime from now I\'d think that there was something wrong with us because of that maybe I don\'t see you enough to be completely honest with you maybe that\'s it. This sounds like one of those online rants that I hate reading I\'m using our relationship to write something that is appealing Part of the failure must come from insincerity then, I mean when an amateur writer is unable to transcend the barriers of creation. I am not palatable. I am thunder and flash and a few soundbites. I am not cinematic. I am not theatrical. I am words but not images. You like what I have wrote. We crawled from all of us to where You are and where I am. You still want to Leave when I really want to go Back.  \"  I’m tired of listening to the things we talk about. I\'m tired of reading the bullshit I write. I should apologize for the whining bitch I’ve been the last few days, the bout of depression I’ve wringed myself through, the ugly side of me I’ve exposed. Yet again. The road not taken but the past revealed again. I have no explanation.   They say change would take place, they say I will go through phases and mature. Somehow I still feel like a dorky insecure kid, my very words here underlilne the doubts in me. What if I have no talent at all, no forte. What if I’m a mesh of bullshit, what if I have but not enough?  Confused. Disorientated. Lost.  Here I go again.  -  I write this the next morning, I fell asleep on this. Too tired beyond words, exhausting myself out. The creatures that come out at night to haunt me scares me, I seem to always break at night. When the world isn’t watching, my cloak of confidence evaporates.   We go out and party, we socialize, we make do as if everything is alright, we laugh as if joy comes freely from our soul, we smile as if the world is beautifully serene, concealing all the turmoil that is within. Concealing because it is easier to not explain or concealing because you don’t trust yourself with them or then again, concealing just because.  Yet the world is still perfect, is still beautiful.   Hope still surges on despite all that is happening. Today is a new day, all my tommorow worries for yesterday are here today but yet I am still here, still standing.   Somehow, things might just be alright.    Somehow today is just today, somehow I am just a small part of the world, somehow, my problems are only as big as I make it, somehow, small steps make the world.  Just watch out for the signs.    urlLink  \"You\'re not getting off so easy, Ms. Lack of Responsibility.\"  '),(22300,664485,'01,August,2002','2002-08-01 00:00:00','   my day started out brilliant with my sun roof deciding to fuck up on me then subsequently, rain. yes, it\'s crazy ass hot for a good whole week and the one day my sun roof decides to die on me it bloody rains. damn, i must  really  be pissing off someone above.   ahh..but that will not deter me for i went in (afuckinggain!) to get stabbed. :) of course i just squeezed my eyes super tight and chanted and chanted and chanted till they stuck everything in and we played video games till it was over. all good honey! i\'m not really sure what todays session was for though, because unlike yesterday where i was anally sensitive, today, the voices in my head blocked the doctors whatever out completely. i also happen to have that god damn bishi bashi theme song in my head i swear i\'m going insane.   we\'re starting on me lungs on monday. oh hell yeah this is going to be fun.   :P  in other news though, it\'s a friday night!  grace wants to go get dinner at some new place she found.  my brother wants to go clubbing.  chris wants to go for some short film festival blah thing downtown.  auntie jesse wants me to go for cg. (mmhmm)  steve wants me to make chicken soup. lol :P  and i...want to go to bed.   i\'m sleeeeeepppyyyyyy.  why the hell do i have a feeling that i will do none of the above?   urlLink  \"No.. I am not just busy, I am in Hell!! If I am ABLE to pull all this off without a hitch, it is impossible that I am human!\"  '),(22301,664485,'01,August,2002','2002-08-01 00:00:00','   i might have majorly worried a few people last night. i don\'t know what possesed me but i lost it, it\'s almost funny now, still kinda stings but good to know that i\'m still breakable. i really don\'t know why i had such a bad night because essentially, it was a pretty good day. there was a lot more of tears after that post but they were tears i needed to get out..it felt like there was so much pressure bottled up or soemthing. in any case, things are...i don\'t know, strange right now, trying to find my balance, will get there eventually i suppose.  in another realm completely though and almost uncharacteristic of me, i went in for alternative theraphy today. the method of torture? acupuncture. it was...interesting. i generally don\'t like needles but then someone suggested that it might help and so we went in and got it done. i freaked out for a good half an hour before anything really happened but i was still piss terrified none the less. nothing like watching needles stuck down you and connecting volts to it.   i still believe it\'s all in the mind, but oh well, if it makes me parents happy.  i didn\'t really feel like doing anything today. to be honest, i wanted to just get away from everything. i think i got burnt out from so much activity and thinking in the last few weeks that i might have tired myself out physically too. and you know that when you\'re tired just about everything else is wrong too.   but then sakthi and archana came over. i forgot i made plans last week so we just went and got dinner. it was nice to be with them. it\'s nice to be with old friends you know? people that have known you for what seems to be like your whole entire life. when i\'m with them i almost feel unjudged, it\'s just a nice comfortable feeling. it\'s almost strange how possibly that might have been what i really needed tonight. just us girls against the world. just us, just me, and i can be sad if i want to or i can be bitchy or just cranky and i know it will be alright. they almost numb the pain if even for a just a bit.  we\'re going to watch hooverstan(?) next week at hard rock. i don\'t even know the bands name but apparently they\'re good. and even if they\'re not we have to get arch a single hot chick with mind blowingly amazing positions. possibly even an orgy too. i\'m just going to shut up and smile, me opening my mouth has caused me quite a bit of trouble the last few days.   right now, david is trying to convince me to go to hk for the weekend. in all honesty, it doesn\'t seem like such a bad idea but then i don\'t really want to do that either. running away merely just multiplies the confusion and mess. that much i\'ve learnt. so how to get to the root and fix it?  beats the shit out of me.    urlLink  \"She wanted him. Badly. She turned and walked away. Her mind knew it was not the right time. He was not the right person. Some day her body would win over her mind, but not today.\"     \"you cry on the inside and hurt so very terribly that your heart feels like it\'s about to explode\"- al    HAVE YOU EVER------------------    * Ever been so drunk you blacked out:  probably but not that i can remember..i generally don\'t exceed the i\'m going puke for the next four days stage...i think.   * Missed school B/coz it was raining:   missed school just for the heck of it, period. anything and every excuse comes later...though i am extremely excited to go back to school though, strange.    * Put a body part on fire for amusement:  hmm..well i did burn my fingers the other day because some people don\'t know how to light up their lighters properly, then there was my toes with jeff and i bet you that you prolly did it when you were little and burnt up strands of hair just cause it was silly. shit, you mean that was only me?   * Been hurt emotionally:  i think at times that could really be just my fault for being too sensitive or letting things affect me but yeah.   * Kept a secret from everyone:   don\'t we all have things kept in the innermost valves of our hearts?  * Had an imaginary friend:  she had pink and purple curly hair and talked in swahili. we had fun together, she and i. good times i tell ya, good times.   * Cried during a Movie:  more than i care to remember...really.    * Fave Ice Cream:  it\'s all his fault that i\'m addicted to green tea fr haagen daz because really, who the hell eats green tea ice cream?   *Fave Subject:    varies between art, english or science.   *Fave drink:  i don\'t think there is a specific drink i\'m completely devoted to..prolly coffee or coke or juice?  * Fave Persons to talk to online:  i have no idea..the list is pretty long. or at times, really short. :P  -------------------RIGHT NOW------------------        * Wearing:  top and shorts i dug out from the unwanted pile of shit i have lying around.  * Hair is:  wet, short and colored  * I\'m feeling:  melancholic, grateful, sleepy, anxious, lonely, stupid, bored.  * Eating  : nothing..unless you count the ice chips i\'m chewing on..?   * Drinking : orange+carrot juice   * Thinkin bout : so many things and trying to not think about them at the same time  * Listening to : dmb  * Talkin 2:  no one in particular except a three way with david and terence. don\'t feel like talking much tonight so am hiding.    ---------------IN THE LAST 24 HRS------------------  *   Cried:  yeah. a while this morning, a short one at the doctor\'s and a retarded one in the shower. it\'s been a long day and i\'m emotional and hormonally imbalanced, forgive me.   * Worn a skirt : nope, today was a jeans day   * Met someone New  : special new? no. new new, yeah. the asshole that backed into my dad\'s car this morning. dumb cow.    * Cleaned your room:  clean? well, i sorted through a couple of things tonight but no, generally i don\'t really help to clean up much. sorry.   * Done laundry : that too.   * Drove a car:  didn\'t drive today either, took the safe route and stole uncle ahmad. my dad\'s car got fucked anyway so it didn\'t make much of a difference. all these fucking excuses..pah.  ---------------DO YOU BELIEVE IN------------------  *   Yourself:  i have my days  * Your friends:  yeah   * Santa Claus : only cos he lets me have his other cookie  * Tooth Fairy : nope, she\'s mean.   * Destiny/Fate : i believe that we all have a specific path we need to go through but i also believe that there are also choices and decisions that you make along the way that determine your destiny  * Ghosts:  yes, especially the big fat ones that haunt me late at night  * UFO\'s : i swear i live with them sometimes.  -----------------FRIENDS AND LIFE------------------    * Do you have a boyfriend/girlfriend?:  no  * Like anyone?:  what do you think?  * Who\'s the loudest:  trying to think and the best name i can come out with is lai jin. i swear her voice reverberates(sp?)  * Who\'s the shyest : chris? only cause it takes him so long before he\'ll have a complete conversation with your friends and be comfortable  * Who\'s the weirdest : :)  * Who do you go to for advice : i don\'t know, you don\'t really need to go anywhere, it just comes..   * Who do you cry to ?: i don\'t know..myself? god? him?    * When you cried the most:    grandma   * What’s the best feeling in the world:  believing in yourself. loving life..?   * Worst feeling:  when everything feels like it\'s crashing down on you and there is absolutely nothing you can do about it. the feeling of nothingness.  * Who will respond to this email the fastest?:  considering it\'s not an email...i wouldn\'t really know.  * Who did you send this to who won’t reply : no one. i\'m just trying to distract myself really, this is apparently supposed to help..   * Who sent this 2u:  albert lim   * Do you want all your friends to do this and send it back to you:  sure, if they read it i guess?  '),(22302,664485,'24,September,2002','2002-09-24 00:00:00','   closure.  '),(22303,664485,'09,September,2002','2002-09-09 00:00:00','   night falls.  i wonder what it would be like to kiss the sky and dance among the stars. floating hazily in between consciousness and swiftly  wading in between dreams. seeing people, faces; things i use to know but they don\'t connect. i recognize, but it is only familiar,  deja vu, but nothing significant. nothing i really  know .   good days and bad days. ups and downs. inside and out. open and in. honest and hiding. things will be okay, they have to be,  they must be.   i cry when i hear your voice, it tells me not to cry. your will to protect me pushes me away further, i am scared of life today...i  cannot do this. courage: what the fuck is that? no, but i am strong, i am strong, i am fine, i will get through this, i am always good,  i am fine, things will be right, it is okay.  i toss and turn wanting and forcing my mind to shut down and sleep but i feel the tears well up behind my eyes and i hold tighter  and tighter, and breathe deeper and deeper, and then i let go. sob. again. why baby, why?  fanfuckingtastic show. perfect. utter brilliance. but i don\'t see your face. i don\'t get your congratulatory hug. i don\'t get your laugh  that rings in my ear, i don\'t hear your voice, i don\'t see your face.   i am alone.  auditorioum. people. friends. dates.  still alone.   but i wanted this didn\'t i?  un-attached. single. open relationship, hell yeah.  but still alone.   and that\'s good right?  '),(22304,664485,'09,September,2002','2002-09-09 00:00:00','   carys.   You think the images might fade one day…some day, it will all be over you tell yourself, things are really okay, just look for the good, just look for the good. You watch their faces, so damn innocent, you watch and you guide and you find and search for all the things in you to make things right but you have no bandaids, no miracle cures, nothing to stop the pain or the tears so you cry with them and hold them close and tell them it’s all going to be okay, that things are alright and you be strong even if you have no idea, you pretend, you piece their jigsaws back together and mend the broken hearts because you know fixing them fixes you too.   Things change around you, you adapt, swim or sink, life goes on whether you like it or not…days pass and months go by and things can go so wrong and so right together and all at once in a go so you look out for the good days and bury the bad ones. You learn to let pieces in you go even if it hurts so much, you know you need to do the things you need to do even if everything is harder this way, you try and bleed and try again. You cannot give up, no, life is too short for that.   And then you find the ones you love and tell them how much they mean to you. And then you find the ones you’ve hurt and tell them how sorry you are. And then you find the ones you don’t like but become real friends anyway because you remember what you were taught as a child, to treat others exactly the way you would like to be treated and you ask yourself what would jesus do and what is right and how to fix things even if honesty hurts so very bad, you look deep in you and learn to fix the things you can and accept the things you cannot.   It is a violent game of extremities, at times you think you will go insane from all the pressure, at times you think it would be so much easier to just give up but you realize you cannot, and you realize how wrong you are, how hypocritical it would be…you realize how you’ve been here before but how you never followed through. You promise yourself to make it this time, you promise yourself change and rainbows, you promise yourself hope.   Hope is a beautiful amazing feeling; I forget sometimes. At times I think I should learn to deal with criticism a whole lot better, and at times I think about how I am too proud. I want to say that to love someone is to love it for everything it is but then again, how exactly do you love something that is unlovable? I am tired of thinking and questioning and at times I wish my mind would just give me a break and at times, it does. Too bad I forget those moments and focus on the wrong things.   The bay was extremely peaceful today. I took out all the tealights I had and lit them up one by one and let it float. It was so serene and calming, I could not hide my tears and so I let myself cry, I cried like a little baby because it was so beautiful, I cried because right at that moment, I knew things were going to be okay no matter where it ends or where life takes me now. I don’t know what the future holds for me tomorrow nor do I know how today would end, I am scared and a little intimidated of what is before me and all that I have left behind but just like someone told me a long time ago, you gotta grab the bull’s horns by it’s goddamn balls and screw them.   I am not perfect. I never will be. And so when I strive for perfection I need to remember that no one is either. I have been asked to sing ‘God Bless the USA’ for the memorial service this Wednesday. Sometimes I wonder if all this is really happening, I sit through rehearsals and think how surreal it all feels, they play the clips and I am numb. I wonder if that makes me cold or if I just don’t know how to process it. In a room filled with different ethnicities, race, color, gender, and age, I wonder if I am alone or if I am one of them. I tell myself I have too many fears, that I worry too much….i worry if I break down on Wednesday, I worry about what to say to the kids, I worry about what to say to the adults, I worry if they ask me anymore about muslims, I worry if they’ll hate me, I worry if I’ll fuck up, I worry everything to death and back but it still doesn’t stop me from worrying. I think too much.   I guess I just came back to say that I love you. And you. And you.   That each and everyone of you is special and beautiful and loved whether I know you or not. If you came in, you either came in because you were looking for some machoistic entry that is always promised here or because you cared. Either way, I love you in anyway. Sept 11 taught us a thing or two, remember the promises you made that day, remember the dreams you need to fulfill and know that it will come true someday. I am just one small stupid person in a huge universe filled with every different color, touch and smell but you make up a part of my world and I believe in you.   I guess then that I am really shutting this place down for good. It is bittersweet but not a bad thing…I am learning to trust myself, god forbid, love myself. It is an interesting battle. I am growing up…now isn’t that scary?   Take care of yourselves and please know that if you ever need anything or just a friend, I hope you know and would consider me one too. Be good because the angels are still watching…do something nice for humanity today okay?   Corny, but we still need world peace.   Everything is going to be okay…it has to be and it will be, I know.    sing jie pao chu ni.    *hugs*  '),(22305,664485,'05,September,2002','2002-09-05 00:00:00','   it is 4:17; i have a million things to do but i walked past the comp lab and suddenly had the urge to savor this moment.   i am free. okay. surviving.  i am understandable, lovable, improvable.  i am here, and now, and living everything.  i am happy, i am content, i am free.  '),(22306,664485,'30,October,2002','2002-10-30 00:00:00','   note to self: do not, under any circumstances at all, ever attempt to put together a halloween costume 27 hours before. almost if not eeriely like christmas shopping the day before christmas.   collectively, we were going to go as a group.  first we came out with  scooby doo  characters...which got tossed out right around the time we realized no one really wanted to be doo at all.  then there was  captain planet.   and  archie.   and  smurfs.   and  snoopy.   and even them  disney characters.   then we thought, hey, since we\'re all dumb artsy people we\'ll go as musicals.   rent : too fucking overdone.   cabaret:  slips and bras?   rocky horror show:  white bra, glitterry stuff, lots of ass.   carousel:  inside joke and only the cast members will get...besides, no way in hell is our propmaster going to risk us destroying the costumes one week before opening.   little miss sunshine:  what the fuck?  so then we went back to the concept of just normal halloween type things...but the girls wanted to be pretty and the guys wanted to be, well, guys.  so this is what we came out with..   spice girls:  we had three chicks and two guys in our little circle of people that stuck their asses up their heads in other things then realized halloween was tommorow. the guys debated being transvestites but guess which side didn\'t win?   moulin rouge:  too much ass, too much booty and hell that\'s some damn fucking  body  fats we\'re showing. fantastic idea though.   pimps and whores:  doesn\'t that answer the suggestion itself?   cheerleading squad:  not funny at all.   ravers:  neon very very bright.   gothic shit:  only about everything and everyone\'s theme this year   hackers:  inlines?   buffy:  not hot enough   christina aguilera through phases:  the scarf, the genie, the whore.   madonna through phases:  i don\'t care anymore, it\'s six and we have an hour.  so the final end product after five painfully tedious long hours, six hundred and seventy shops and four thousand and six times back into it; we\'re hookers cum sluts cum ravers cum rocky horror show meets jerry springer thrashy things.  (now hey, wasn\'t the point of halloween to dress as far away possible as you are again? :P )   lol, and so i have neon blue punk highlighted hair, a kick ass butterfly glittery top in which i realize i have absolutely no and retarded boobs, held together by what hopefully is two very strong pieces of string and black pants with slits in them to reveal my  neon  cyan blue fishnet stockings and heels.  hell yeah.   amanda  has neon purple pink and silver hair, kinky kinky \'keep-it-down-boy\' victoria secrets lacy type corset top and very sexy leather shorts/skirt(she hasn\'t decided), black fishnet stockings and some kick ass boots and hoochie mama boobs.   cory  has...the most hideous ever pink bathrobe type material dress with some sort of strange white collar, neon yellow cum lime green hair, same neon colored fish net stockings and pink fluffy slippers with hearts on them because it is close to impossible to look for a size twleve women shoes that look cute below the cost of twenty dollars.   jessica and will  left early for rehersals but apparently they\'re coming as the dj and cabaret dancer.   we were going to go trick or treating but rehersals starts at six and only ends at ten and i have classes before so there goes that idea..we will be bringing candy for the kids to trick or treat around the cast though. :)  no specific plans but hell a lot of places to fit, oh, only about sixty people. oops.  camera comes out tommorow!  :)  i haven\'t celebrated halloween in so long it\'s almost ridiculous. hehe.  but now, back to the thesis.  gah.  p/s because i thought it should be announced, i\'m fucking tired.  '),(22307,664485,'29,October,2002','2002-10-29 00:00:00','   urlLink    urlLink    urlLink    urlLink      urlLink    urlLink    urlLink    urlLink     my brother is a turd.   \"anyways  you have one month to stop smoking  otherwise ill tell your parents  put on yr website your progress in quitting smoking  its for your own good  and if u wanna know, my children\'s own good (\"Aunt rachel is smoking so it must be cool and healthy\")  everything else ill keep my mouth shut.   love  daniel\"   in other news; i had hell of a good cry today.   isn\'t it always fun how you have to evoke certain things to play your character?  i deserve my ciggarettes. i cna\'t justify it but yeah, oh well, i\'m a bad ugly person.   bite me.  '),(22308,664485,'29,October,2002','2002-10-29 00:00:00','   it rained last night; it poured.  [10/28/2002 5:32:33 PM | + racheal]   monday came, and left and somewhere between it all i gained an extra hour..though i can\'t explain it something happened. weird feelings meshed together with thoughts and actions but it\'s all still very surreal, the whole outer body experience cum deja vu like feeling.   monday was shit. must have been the universal day to feel like crap all day and hate the world.  ...i let three people cry on my shoulder and wiped away more tears than i cared too. monday, i was unemotional and numb. some spark in me fused or burnt out but i did not feel, and that scares me much more than being overemotional. i wanted to cry but the tears would not come, wanted to feel the pain but just emptiness remained.  monday, i know now my little problems are almost nothing compared to the heavy load others have to bear. while i am worried about my stupid dumb feelings, someone else is worried about their next meal. while i am worried about my body and getting my nails done, someone else is worried about staying alive. does that put that into perspective? a hell lot.   signs.  because last night i realized what \'family\' truly meant. my dad on the phone with me at some god forsaken hour his time comforting me, my mom listening to everything i have to say because she knew that all i really needed was someone to listen. and so she did. my brother who i love with all my heart that has taken me a long while to see that he cares for me so much.   people i hurt without meaning to.   i am lost in this funny place of wanting to do what\'s right. growing up i guess, all these damn responsibilities and pressure. i am stressed, i am up my ass in trying to control every single fine damn detail of my life.   but i can\'t.  i have to trust someone and trust the people around me and trust god.  ...but what if i don\'t want to?  i don\'t know what i want exactly. i just have the rough draft, a hazy idea and it\'s like that last ten minutes of reaching somewhere, like waiting in line for the tickets you\'ve waited to get your hands on the whole day, like knwoing something is wrong but not knowing what exactly.  i want my \'exactly\'. i want instructions that come with a 101 book. i want to fucking understand, that\'s what i want.    \"i shall never believe that god plays dice with the world\"  '),(22309,664485,'28,October,2002','2002-10-28 00:00:00','   i think i should just return to long rambly posts about cows and atmore as opposed to my life is tart.  i thought so too.  :P  stay away from the fumes racheal.  '),(22310,664485,'27,October,2002','2002-10-27 00:00:00','   end of the bargain.  maybe because i\'d like to think that somewhere out there the world stopped turning and people aren\'t watching and words didn\'t hurt.  get out of my world.  '),(22311,664485,'27,October,2002','2002-10-27 00:00:00','   who\'s your bitch?  i am trying to decide if i should laugh my ass off or whine the hell away. (i have a pretty good feeling that i\'ll be able to do both..but anyway)  he says all the right things and does all these unbelievably sweet things like the  urlLink teddy bear  and  urlLink roses  and coffee and waffle bloody house and just crazy ass nice guy stuff that sometimes makes me want to scream and jump on him yet i want to go like you\'re too fucking good for me, i\'m going to tear you rip you change you and i don\'t want that responsibility, don\'t want to hurt you or lose you thing. yet at the same time i am happy with him but there\'s something missing..or something is wrong?   we talked last night and in retrospect, it went really well, but if i have to think about it...it went terribly disastrous all at once. you know how the more honest you get the further away you feel disconnected? i don\'t know exactly how to pinpoint the problem...but i have a rough idea where and what it is about which doesn\'t help at all considering it\'s stupid and unexplainable.  i just need to be over it and secure with myself before i\'m going to put my everything in this because i don\'t want to be thinking about someone else when we\'re together or have him deal with shit from my past you know?    oh well.  the set is coming together pretty damn well and i\'ll be lying if i told you i wasn\'t excited. it\'s almost an outer body experience to try and understand that our opening night is in two weeks but all at the same time i don\'t necessarily feel prepared enough or rehearsed enough minus the fact that i\'ve been in rehersals every week this night but there\'s just ass a lot to do..plus the monologue i have to work on aptly titled \'suicide\'. madness.  i just had an epiphany.   or something to that effect anyway.  ....  i. i\'m writing for me, this is my place and i am safe here.  ii. i will not let you affect me.  iii. i am fucking worth it.  iv. fate just happens and so whatever will be will be..  v. patience is a damn nice virtue. (time fixes everything)  vi. i am okay, blessed, loved and going to bloody make it.  vii. unsinkable ships sank and unbreakable walls crashed, look at it as nothing is impossible rather than nothing is certain.  viii. carpe fucking diem  ix. life is beautiful.  x. my koko is pretty damn something. :)  ----  so yeah, my thoughts.   grr.  :P  '),(22312,664485,'26,October,2002','2002-10-26 00:00:00','   the notion.  you know how there are certain songs that remind you of specific moments and evoke these complex unexplainable feelings?  yeah.  i don\'t know if i am in love, or i am in love with the notion of being in love, or i am infatuated, or if it\'s just pure sinful lust.  today was...  confusing  tiring  frustrating  productive  eventful  okay.  or maybe beyond okay  ...yet just about three steps below that.....it\'s hard to explain.  somehow, between three weeks ago and now i had convinced myself that i needed to be in a relationship to keep me from going insane, not beating around the bush and stringing along bullshit nor wanting someone but not having him because you\'re being a bitch yet all at once just not very sure at all.  it\'s been going really well.   and that scares the fuck out of me.  i got all crazy panicky last night when he held me. i don\'t know what it was, maybe the way he was looking at me, or maybe that strange aura that enveloped us but i freaked out. i don\'t know, i just felt so weird, and strange. it doesn\'t make sense at all but i realize that i\'m not really in love in love with him. or if i am i\'m pissing afraid of it.  i know, i know, it\'s the same damn bloody story again but precisely that. fuck, if i have to go through the last god damn year again i\'d rather just kill myself. it\'s just that in the end there is so much hurt left and so much pain that i don\'t know exactly if i am comfortable with having to trust someone so much or having to need someone that\'s going to tell me to fuck off and make me hate myself type of thing.  i don\'t know, it\'s just last night when he touched me all these alarms went off in my head and there were like flaring red light bulbs that exploded and i didn\'t say a single bloody solitary thing.   i worked on the set the whole day today then had rehersals so we havent really talked talked. we were supposed to go out tonight but my body feels like it\'s been through the wringer..hell, my mind is going crazy.  i\'m just. i don\'t really bloody know at all.  i don\'t know if i came into this relationship to \"move on\", or if i\'m prepared.  it\'s not a relationship, relationship. somehow, the notion of having a boyfriend right now just doesnt seem right at all yet i know he\'s more than just a friend. he\'s very special to me  ....but it\'s still different.  damn it.  the show opens in two weeks, the midterms went fucking well but it also means the finals are coming soon, i still have no bloody time for anything, i need a fucking ciggarette and he\'s coming over now.  maybe we\'ll just have wild explicit sex and i\'ll forget about everything.    hell, there\'s always a maybe right?  \"sigh\"  '),(22313,664485,'24,October,2002','2002-10-24 00:00:00','   i have dreams of you. sometimes i lie awake in the middle of the night fighting of tears and nausea, but all is good i still say, all is good. i wonder if the sometimes nightmares, sometimes dreams, are trying to tell me anything at all, or if, like maybe you and me, they were just a friction of my imagination. maybe, perhaps, i don\'t know really. back to moving on..  people tell me things. sometimes i think it\'s because i seem to have this cloak that makes them figure that there is nothing that they can tell me that i would have never heard before. i don\'t flinch, don\'t blink..and when all is done, smile and produce more reassuring hugs and offer another round of coffee. but i lie.  i am afraid.  i hurt me, and beat myself up, and starve abuse myself but that\'s just me and that\'s okay.   i don\'t know. it doesn\'t make sense really but i want to protect everyone and all around me. but i can\'t...and that\'s scary.  i hate hate hate vulnerability.  i just want everyone to be okay and live happy ever after and see them the way i see them and realize how damn amazing they are.  you do know how precious you are right?    urlLink \"I want to drink and dance and sing, with my friends, Rick\'s friends; I want to burn wood and smell the smoke and piss in the sea, shout at the moon, cry at the beauty of this world and swear eternal opposition to those who would paint it with blood, again, I want to hold you my friends in my arms and circle around our wounded brother and sing and weep and shout and break the shell of the words. I\'m drunk again, by god, and I\'m typing with tears running down my rugged goddamn manly cheeks, and I feel so useless. I want to bellow in pain like some big dumb beast, and I want to hold my friend and tell him that he has nothing to fear.\"     rest in peace rick.  '),(22314,664485,'22,October,2002','2002-10-22 00:00:00','   [10/24/2002 1:55:41 PM | + racheal]  i need..  to get my nails done  to finish up the set  to get my damn assignments done  to work on my grammar  to rehearse my lines  to schedule the classes for my next term  to balance my checkbook  to finish up my speech  to get on a diet to get the photographs printed  all the my, my, my\'s and get get get\'s  terribly conceited now aren\'t we?    [10/24/2002 1:57:48 PM | + racheal]  every time i have a ciggarette i wonder how much of my lungs it blackens now  then i think about how if i\'m being anal about it how i should just quit  then i do for the next six hours or so and start being a bitch  then i cave in and finish a pack  (well not really, but more so than i should have anyway)  then i start again  and think again  and so this vicious cycle fucks me up again.   now fucking indecisive are we?    [10/24/2002 1:59:10 PM | + racheal]  i smell of paint, reek of ciggarettes but am filled with hope  what a fucking oxymoron  or not really  whatever the word is anyway  i like honesty on paper  or words in this case  but it\'s funny to take it all back  when i think i\'m boring you to death  or conscious of what i\'m saying  pah  this is retarded anyway  harrumph.  back to the painting..   edited to add: you think that one year of this shit and i would have learnt how to hit \'publish\' now won\'t you?  '),(22315,664485,'22,October,2002','2002-10-22 00:00:00','   urlLink    urlLink    urlLink    urlLink      urlLink    urlLink    urlLink    urlLink      \"i drift away and fall into this space where i am trying to comprehend everything that is happening within and around me. wondering how in the world i got here yet feeling like there is so much more to this i have to discover. ever get the feeling of itching to get somewhere but not knowing where you\'re going? or knowing something is happening but not knowing what it is? yeah, exactly.\"  '),(22316,664485,'21,October,2002','2002-10-21 00:00:00','   my  urlLink host  is an amazing person...i don\'t tell him very often but once in a while, when i\'m feeling lovey and nauseated, i write posts like this to tell him he\'s special. :)  besides, who else ass am i going to kick in literati and share gory bits of \'paramedics\' with?  or yell about my day and expect him to fix the site when i\'ve shoved it up my ass and churned out broken links and garbled html?  and who else can i go to to share the feeling of sweet ecstacy as you sit back and enjoy that fucking well earned drink and  malboro  salem light?  i think he\'s a cool guy...and i\'m very very happy he\'s back.  :)   \"[5/25/2002 5:14:29 AM | chris raettig] once there lived a village of creatures along the bottom of a great crystal river. each creature in its own manner clung tightly to the twigs and rocks of the river bottom, for clinging was their way of life, and resisting the current what each had learned from birth. but one creature said at last, \"i trust that the current knows where it is going. i shall let go, and let it take me where it will.\"   p/s: i think  urlLink she  is a hot sexy amazing chick too....but don\'t tell anyone i said that. she might find out and kick my fat ass. :P   p/p/s  urlLink girls suck!   (....but men are still evil anyway! )  '),(22317,664485,'21,October,2002','2002-10-21 00:00:00','   i feel like god right about now...updated and fixed  *all*  the sections thank you very much. :P  ironically (i love that word!), rehersals were cancelled today so all the extra time to myself has made me deliriously happy..hehe. and this would be the part i tell you again...life is goood.  so is chocolate fudge cake.  and starbucks skinny cafe mocha.  and stupid crazy retarded rain that begins right as you step out of the car.  but most importantly, stupid retarded friends that will dance in the rain with you.   :)  and because i love filling in these stupid survey things..   i am:  happy, content, single and not looking :P   i think:  thinking is bad for you  i know:  everything changes.     i want:  to move on, to be the best person i can be, to try try again.   i have:  love, friendship and guidance.   i wish:  i understood me a little better.  i hate:  being angry with myself.   i miss:  playing with the kids  i fear:  getting hurt again, myself, the asshole i can be.  i hear:  voices in my head? :P  i wonder:  how god manages everything..i mean, that\'s a lot of lives to be working on..   i regret: .  fucking up  i love:  hope. (and orange juice)  i ache:  late at night when the world isn\'t watching...or so i think.   i care:  too much and not at all  i always:  smile even if it hurts  i am not:  perfect  i dance:   for a living  i sing:  in the shower  i cry:  when i think no one is watching  i do not always:  tell people how i feel.   i fight:  because i want to. i need to i guess, for me. i\'m tired of being pushed and broken and being used, i need to do this for me now. be the person i know i can be and hopefully, be someone that has integrity, self-worth and is finally enough.  i write:   to escape  i win:  sometimes  i lose:  when i am tired of trying and give up  i confuse:  the living shit out of myself   i listen:   very rarely because i\'m a stubborn idiotic selfish bitch. :P  i can usually be found:  talking to the little green men.  i need:   balance  i am happy about:  getting in the musical, moving on, making new dreams and trying again, fixing my life even if it\'s the six millionth time i\'m doing it again, my friends and the amazing people they are, hope and the things it means to me.  i should:  get my ass off this thing and start working on my homework..hmm  have a good night..  hugs   *edited to add : oops, i lied...forgot about poetry! :P  '),(22318,664485,'20,October,2002','2002-10-20 00:00:00','   out on the rooftops.  ever just sit sometimes and wonder about life and all it\'s twists and turns?  amazing.  crazy really, but interesting.  so  urlLink he\'s  left. it\'s strange, but i\'m glad he finally did it. and i hope whatever he\'s doing now, he\'s happy.  life isn\'t too bad you know, chisel off the crusty dirty layer...you might find something worth holding on to beneath it.   emotionally fucked is a good term to describe a lot of situations i get myself tangled into, but that\'s just me you see. sometimes the very things you try and run away from are the things you are made of.   my new start isn\'t really a big bang or anything, in fact, if i was really anal, it\'s pretty much the same. just trying again.   so yeah, i make mistakes. i fuck up. i fall down. big fucking deal.  :)  things are good, now hell, That\'s scary.  today feels strange. or maybe being content is just a whole new experience, closure somewhat does make you see clearer i guess. so does sleep.   oh yes, sleep very very good.  thinking is normally an evil thing, but i\'m glad it happened today. :)  i\'m so tired of being a cranky bitter bitch and to choke on my own words, life is precious and too short.  hugs   **ironically, the moment i hit publish, the site fails on me. and as forewarned, will fail quite a bit too in the coming days...pah. but the selfish bitch thanks you anyway..:P  '),(22319,664485,'19,October,2002','2002-10-19 00:00:00','   hey jersey  there are six things she needs right now.  she notices his eyes first. they are beautiful, like a swirl of green, gray and brown put together. she\'s shy, very quiet, trying to make up for it by exchanging banter with the \'common friends\', she\'s pissing wow-ed, he asks for her name..and so the game begins. it doesn\'t take very long. no, not too long at all, the attraction is apparent. chemistry? damn, that\'s a strong word.  back and forth. each time drawing her deeper with every word he says, his intelligence draws her nearer...closer. falling to edge and it\'s hard to get off this ride. lights up, smiles etched too deeply on this face...a certain aura that radiates when he\'s near her. a certain feeling.   not wanting to screw up so bad. nice, slow and easy. they have so many conversations without words, like driving around no where and listening to music, like watching the first light of a brand new day break in or looking out into the world beyond in moonlight to the sounds of waves crashing. things he\'s dying to show her, new things he helps her discover every day.  he watches her. talks about her. her friends say he\'s a keeper. they\'ve seen them together, they\'ve seen him talk about her.  he sees when no one is watching. protects her if she\'ll let him. nights she steps out to be alone a while; she always knows there\'s big comfy hug waiting for her to come back into, sometimes, he finds her too. they talk about a lot of things, time flies by so fast and beyond. he drives careful, makes sure he takes her home safely each night. takes her to her favorite places, doesn\'t pressure her, no not at all, take your own time sweetie, it\'s all yours.  doesn\'t promise her anything. as a matter of fact, knows to avoid all conversations that veer towards subjects that make her run, he\'s seen her run before. he knows that feeling, she\'s afraid of him running, she\'s afraid of her running, but when they are alone, it\'s almost too comfortable. you have no idea how time flies.  she wishes he would talk forever, every fucking thing he says pulls her in deeper. when he talks about history or the world.things that go beyond a cute butt and a sexy body....yet it\'s not the talk at all. it is the person and the way his eyes look into yours, that old soul behind it that takes you back a moment, that moment where you pinch yourself and wonder, is this real? yet she still never talks about it. he\'s tried once, right after, the night he told her she was beautiful and she cried. the people from her past should remain there she says, and so they never talk about the past. nothing there that you need to know, men are evil, very very bad..jokes but actually believes it.   odd isn\'t it how things that cloud you the most inevitably become the subjects you avoid the most. he\'s seen her without even saying anything at all. is she transparent or is this just the magic that happens?  thier eyes lock, i don\'t know if a second passes by or if twenty minutes has passed already. he reaches in to kiss me and i don\'t know if to draw back like so many times before because this actually feels kinda nice. right almost. i like his arms around me, or the way he whispers in my ears and his breath tickles me. she\'s a huggy person, they\'ve always hugged, but something has changed. the way he says i love you is different tonight. or how he picks her up and swings her around and around and around as she giggles and laughs.  how he brings her coffee or watches her at rehersals. or rings her and leaves her messages to make her laugh. how he kisses her forehead and hugs her so tight it\'s hard to breathe and says, \'good job baby\', how she lights up when he\'s around. that buzz, a high.   he takes her shopping. and she shops a whole lot. how six hours later he\'s still smiling, offering to play another round of \'the dancing game\', how he doesn\'t mind being goofy with her. how he doesn\'t mind her being goofy. or stupid. or retarded.  how they fight to pay every time and they sneak paying things. making up for things she says as his mock angry face only tickles her further. she says she\'s being the man so she has to be in control. secretly, she knows, she just doesn\'t want to fall. i don\'t think this is helping either.  kiss me harder, again and again. tantalize me with your mind and touch. brings his arms around my waist and plays with my hair. tickles me, breath on my shoulder and the breeze is so damn cold tonight as i get chilly and shiver. feels my tummy groan against my free will, hunger pangs invade as he pulls me in tighter. runs off and drags my fat ass to get food. the waiter laughs, \'oh it\'s the both of you again\'.  she flinches. his finger toying with her belly button ring as she falls deeper into (t)his spell. they kiss, and each kiss is a magical explosion of it\'s own.   someone said inseperable tonight.  she almost laughed. it scared the piss out of her.  so many pressures in her life right now. stress of venturing into an adult world and picking up responsibilities yet he eases that tension. someone she can actually study with, someone that makes her laugh a lot, someone she feels safe with.   the first person she\'s dated in a long long time.  the first person she\'s actually wanted around in a long long time.  too long she says.  where have you been all my life?  '),(22320,664485,'17,October,2002','2002-10-17 00:00:00','   urlLink    urlLink    urlLink    urlLink      urlLink    urlLink    urlLink    urlLink     i figure this cannot possibly be good for me; i am currently running on two hours of sleep, caffienated diet pills, homework, homework and homework, rehersals, school and attempting somewhat a social life. fuck the hell yeah.  i also attempted to quit smoking somewhere in between the hours of seven last night and three this morning. very bad idea. cranky grumpy sleep deprived racheal without nicotine? evil almost. :P  i\'m tired. no wait, tired doesn\'t even come close to it, i\'m bloody ass exhausted. pooped. dead. whatever you fancy really.   though i really like life right now, i just wish it\'d give me some proper time to enjoy it. don\'t get me wrong though, it\'s still dandy and i\'m still going out, still dating (LAUGH), still doing fun stuff, still socializing but oh my god.  i also need to find a halloween costume.   i think i\'ll go as a prostituite.   which reminds me, i find out last night that one of my costumes is a freaking corset. damn.   racheal? corset? disgusting.   so i\'m on a diet for the next six weeks till the show is over. this is going to be fun....especially since my main meal of the day happens at two in the morning at waffle house.   but hey, living for me and doing what i want right?  bring it on, you assholes of fate and beauties of life, bring it on.   grrr.  :P    \"Writing free verse is like playing tennis with the net down.\"    p/s the other post dissappeared!! don\'t ask me how...it\'s a conspiracy i swear!  '),(22321,664485,'16,October,2002','2002-10-16 00:00:00','   urlLink    urlLink    urlLink    urlLink    urlLink     well, apparently, blogger is deciding to be a bitch to me but hah! i\'m still going to rant anyway..  today was fucking ass amazing! :)  i\'m so crazy excited right now i don\'t even know where to bloody begin..it\'s just crazy how everything falls apart and gets mended again and you\'re watching all this beauty in your life so you sorta take a step back and go whhhooooaaa...  i was driving home from school today and i cross the bridge that separates bfn alabama and florida and it\'s raining but not pouring, just kinda drizzling flirting type of thing and there\'s a rainbow. i can\'t even remember the last time i saw a rainbow and it\'s fucking gorgeous. i want to stop the car and just sit there and be mesmerized by it and all the subconscious things it\'s telling me but damn. wow. i think it\'s scary because we had this conversation the other day about rainbows and promises and hope and all that stupid crazy shit you think about in your mind and so it was just breathtaking and thrilling all at the same time. :)  i got my biology paper back from class today too and i got a fucking 96. racheal does not do biology shit and racheal is fucking ecstatic. just like, fuck, that\'s my damn paper! it\'s like a i-can-actually-do-this type of thing and it\'s this indescribable happiness that just runs through you, just wow.  then as if god isn\'t being nice enough, we got let out of rehersals early today and the last run of the review we did for the blocking last night went perfect. like just fucking brilliant. it\'s that feeling where you\'ve done something right and it\'s that other feeling of sharing that moment with someone and smiling and laughing because you\'re truly just buzzed and mad excited.   see, for the last few weeks it\'s just really been hectic and crazy with so much of pressure and work and school and trying to get my life sorted out and all that bullcrap you put your mind through you know? finding your place in the world and trying to act responsibly and \'right\' and so i was extremely grateful for that extra two hours of just time. time to do nothing and anything i wanted. :)  so we went and got coffee till it closed and then went to go watch \'the transporter\' and it\'s funny because we just bought tickets for whatever was showing next and had absolutely no bloody clue on what it was about and it\'s like chinese! azn pride and all that but i was fucking proud.   it\'s one of those nights where you are so buzzed you just laugh and laugh till it hurts and you want to dance and jump and sing all at the same time.   i\'ve missed being spontaneous and living for the moment and doing what i feel like doing at that fucking moment and not really giving a rats ass at the same time.  like being free.  and finding out new things.  and discovering.  and learning.  and smiling. lots of smiling.   and having someone smile with you. laugh with you. dance, sing and shout with you.  that bond. connection. closeness.  just wow.  fooh.     urlLink \"I sometimes wonder if I\'m crazy. It\'s possible, isn\'t it? I mean, you don\'t really know how other people think. You can estimate, but you can\'t know for sure. Huh. \"    p/s finished up \' urlLink about \' during break time but stupid me forgot to bring my network card so couldn\'t upload untill tonight but yeah, just four more to go!  '),(22322,664485,'15,October,2002','2002-10-15 00:00:00','   urlLink    urlLink    urlLink    urlLink      urlLink    urlLink    urlLink    urlLink    urlLink     me thinks racheal needs to haul her fat ass into bed now to get to early class tommorow.  me thinks right.  :P  and yes, i do reallize that all the pop-ups are broken...maybe because i havent even made them yet. just a thought you know? :)  sleep good.  early class bad.  early biology class evil.  grr.  night night.   update- did archives! one down, six more to go....  '),(22323,664485,'14,October,2002','2002-10-14 00:00:00','   urlLink    urlLink    urlLink    urlLink      urlLink    urlLink    urlLink    urlLink     last night we took a drive to the beach and stayed a while to watch the sand shimmer after the waves have washed upon them. it was so beautiful; one of those things that stay in your mind forever like the first sunset you watched or something like that. fucking. amazing.   \"there\'s nothing you can do about the past, because it\'s gone. and there\'s nothing you can do about the future, because it\'s not here yet. so live today and live in the Moment.\"   :)  pinecones are pretty damn spectacular too. i don\'t remember them from when i was little, not before malaysia. come to think of it, i don\'t even think malaysia has pine cone trees...do they?  the mid terms are this week, there is the pressure to study and do brilliantly..grr. days like these that are nice and cool make it spectacular for cookies, coffee and talk. there\'s a little bunch of us that get together after class that have sort of bonded, it is interesting how i\'ve been here for three months exactly today but it feels so odd, almost deja vu like and eeriely familiar yet all at the same time so different. strange how time slips past by you as you impatiently wait on it.  ?  i have rehersals tonight..it is so tiring to live away from everything. sometimes i think all i really need in life is just to be content, or satisfied or something...be happy with who i am and what i am becoming and what i have achieved sort of thing which in a way i am, strangely enough.  our character excersizes today was a bunch of questions and it\'s weird.   part i - character structure   when i wake up in the morning...  i drag my lazy ass out of bed and take a shower   ever since i was a child...  i learnt that trust has to be earned   when i look in the mirror..  i don\'t know who i see   i can remember...  things that used to be important to me   the child in me...  laughs easily and does silly stupid things   i need...  balance   i hate..  being alone and thinking, sometimes.    strength to me means...  never giving up   weakness to me means...  breaking, crying, falling..that sorta shit   right now i am aware...  that i have responsibilities    sometimes i want to cry out to people...  because sometimes i think no one sees me at all   pain to me is...  destructive   i love..  hope. tranquility. peace of mind   if i could be free to do what i want to do...  i would be a pead emergency doctor   my body....   disgusts  is a part of me   if you could hear the music in me..  it would be unpredictable  ------   part ii - character emotion from inside out   the most beautiful part of my body is...  ?   happiness to me is...  being content    the thing i most want to do before i die is..  donate an organ   the most embarassed i ever was...  .......?   the ugliest part of my body is..  me?   the thing i like best about myself is..  the little things   pain to me is..  inevitable and teaching   the most secret thing about me is...  the person i am, the person i can be and the person i want to be   i can hear my father\'s voice speaking through my own when i tell myself...  trust has to be earned   love to me is...  not there   the thing i am most proud of is..  the kids   i want my epitaph to be..  gone home: exit upstage.  it\'s odd. just too much thinking on one page you know? i have to write a thesis on character submission and traits and it makes me think too much. scary.  hm..i am starving, and for some strange reason hot and spicy ramen noodles sound really good right now..  yum.  :P  '),(22324,664485,'13,October,2002','2002-10-13 00:00:00','   urlLink    urlLink    urlLink    urlLink      urlLink    urlLink    urlLink    urlLink     0 so i guess you\'re asking why and what the fuck and how and would you please make up your fucking mind you indecisive cow :) the truth is, i don\'t have the answers. i don\'t know why i\'m back and i don\'t exactly know why i want to do this again...  maybe it has to do with the long talk we had last night about forgotten dreams and things that made you passionate, maybe i missed the a place to come, maybe i just missed writing and talking to no one and everyone at the same time, maybe i just missed this and everything it meant to me...maybe my way of trying again is ending the things i didn\'t succeed to the last time. grab the bulls balls eh?  yet so much has happened all at once. i am eighteen, legal. an adult. so many changes has come about within and around me yet essentially, painfully, i realize that i am still very much the same person.   my heart still gets broken once in a while, a lot of the times because i have blindly led myself to that door but it\'s okay because i want to learn from this shit, i want to hurt so badly now if it means that i\'ll be fanfuckingtastic later...if that makes any sense at all anyway.  i am not perfect.   it hurts to say this but i am not who i wanted to be. i failed at dreams i tried to build, and while i\'m falling into this abyss and drowning myself in a void of emptiness and fury i realize i am tired and that i can\'t go on anymore. i tried and tried again, made the same mistakes, drew back to the same conclusions. sometimes some things don\'t work out and you can beat it to death but some questions just don\'t have answers right?  content.  i have an amazing life. i have amazing friends and people that care and love me. i am making the grades and doing the things i want to do, fulfilling the little dreams and things do work out.  so i still don\'t know why i\'m back, because i\'m habitual i guess. i am a stubborn selfish bitch and this is my place, my home. my corner on this side of the world and i\'m just me.   this is my voice.  my side.  leave if you don\'t like it but i have to accept me.  i\'m trying.  '),(22325,664485,'13,October,2002','2002-10-13 00:00:00','   my fat yellow ass needs to be kicked hard. :P  want details?   patience; damn the virtue.  life is good.  ...apparently :P  '),(22326,664485,'30,November,2002','2002-11-30 00:00:00','   four years.   i don\'t feel liberated, if anything, i feel worst. like i\'ve handed my parents a collection of shit on a silver platter.  but i\'m better now. no more nights, no more crazy extremes, not much throwing up.   ...but still sick.  still small, stupid, still need help.  and a part of me is thankful that i stepped out and asked for help.  but a part of me questions the consequences.  changes. strip everything and rock the comfortable and work at it even if means tearing it apart to start again?  scary.  is this the part i say, \" hie, my name is racheal and i\'m bullimic \"?  i guess so then.  '),(22327,664485,'28,November,2002','2002-11-28 00:00:00','   happy turkey day.  it\'s thanksgiving. it\'s 9:58 in the morning, it\'s wet, cold and noisy outside. but in here, in here it\'s a different world. in here it has lasted forever and i\'m down to my sweats, picking at my fingers, wanting to throw up and needing a ciggarette. in here i am wasted and crying. maybe for god to save me, maybe for me to save me. maybe for just any damn thign to save me.  he says i\'ve fucked up.  i know i\'ve fucked up damnit. i know. i know. i know.  i know it when i wake up in the morning. i know it when i\'m fixing dinner and the stupid osbournes soothes me in the background. i know it when i curl up and sleep at night. i know it every moment. i know i\'ve fucked up.  ten thousand dollars in three months. he says i can\'t afford to feed your addiction. i can\'t do this to myself and it\'s wrong and i need help. that i need to change. and he tells me it\'s about disciplie and willpower and some fucking common sense.  he wants to know if i have anything to say to him.   i shake my head but forty eight hundred miles away where it is dark over there and a single solitary phone line connects us he doesn\'t see me.  he wants to know how and why and what.  i don\'t know, daddy, i just fucked up.  and that\'s not an excuse and i wish i could tell you what the fuck was wrong with me.  he tells me i\'m intelligent and that i\'m talented but that i can\'t do this to myself.  he tells me he knows i\'ve changed and he knows i am changing.  and i want to say did i have a choice?  and i want to say was it really up to me?  but i stop myself. but i stop because i don\'t know how to tell him. i don\'t know how to explain this in between where i am confident and secure with myself yet these very things somehow push me further too. and i hide, i cover, i inexplicably deal with it and they put up with my nonsense.  they put up with my nonsense because they love me. because more than anything in the whole wide world they want to see their little spoilt princess happy.  but there is a fine line.  and i don\'t want to be her anymore.  he says he\'ll bring me home if it continues. like it\'s a threat. like it\'s bad or something.  i like it here, i do. it\'s nice, and comfortable and like home somewhat...but it\'s not home at all. it\'s so different. so fucking different and strange and i feel so stupid and colored and dumb. like there\'s nothing i really know for a fact here. and i like the starting fresh and the building bridges again, and i like the quiet of suburban lilfe and new experiences but it just gets difficult i guess. i\'m dumb.  it\'s been three months and i should be used to it now. i have a great boyfriend, excellent grades, a thriving career and amazing friends. why am i complaining? i shouldn\'t because i\'ve came so far from nothing to something. because i\'m thinking about the homeless man outside and i should count my blessings.  because last night i looked beside me and sitting next to me were two people that loved me very much and at home were my \"parents\" and a million miles away, i had people that prayed for me each night and loved me with their whole hearts good or bad.   but i let those people down and it hurt.  and then the pastor asks us what we\'re grateful for this year. and amanda breaks down beside me because her dad\'s left home and she\'s turned her plane around from virginia because her mom rings her broken down, crying because she doesn\'t want to spend thanksgiving alone. and she says, i\'m thankful for family and friends.   and i look beside me and here is a man that\'s willing to love me for all my flaws. and here at home in a different country is another set of parents letting me in again and here is god saying, what the fuck are you complaining for?  it\'s thanksgiving.  and frankly, it\'s time for some changes.  a new beginning.  it\'s time to leave the past behind and start again.  '),(22328,664485,'27,November,2002','2002-11-27 00:00:00','   reach out, tonight; my home is yours.    a little early but oh well...  burning firewood in amber flames of orange glows as the turkey roasts and all of today comes to rest tonight. i thank god for gifts so sweet and bitter, for learning experieces as little memories jingle in my mind and lonely tired weary eyes cry for the very last time. safe in someone\'s arms again, secure, life today seems beautiful and challenging.  for family; all nights and days of torture and torment having to be bloody related to me. for people to listen to me and laugh when the day is over, for when days melt into dark empty cold nights and everything is uncertain, for at the end, standing right behind me are the people that have known me my whole life. are the people that will always know me the best no matter how far i run or how much i hide. thank you for loving me, for caring, for showing love in your own special ways and meeting all my needs.   for friends; both near and far. for terrible nights i have hated this place, thank you for friends at home that have pushed me forward again and again, for friends here that have made this whole experience a better one, for helping me fit in and slowly taking the time to explain moronic things that everyone around here should know. for sharing a part of their culture and making it so much easier. for being with me to discover new everyday things. there are friends and there are  friends . thank you for being my \"friends\".  this year, there are no long overwhelming emotional schizo speeches. i am grateful to be alive, i am grateful for healing and little miracles. this year has been a combination of good and bad events. whatever doesn\'t break you makes you stronger.....for not breaking.  an online community that has become home for a while now. thank you for space, for words of wisdom and for inspiration and encouragement to be whoever i want to be.   outside, on the cold empty streets lies an old homeless man that begs for shelter. hungry pangs invade his shrinking bony frame...what is his story? how can i help? what can i do?  tonight, my home is your home. my coat yours. my meal is never too little to be shared, never too big to be small.  reach out.  bring in.  i\'m only one in a world of billions.  thank you, everyone, anyone.  '),(22329,664485,'26,November,2002','2002-11-26 00:00:00','   let down.  it\'s thanksgiving tommorow. a part of me has a multitude of things to be grateful for, things to shout about, things that have completed my life yet this week has been awful. wretched. this week i have felt invisible and nothing, yelled at and beaten to bits and pieces. i just got off the phone with my dad, another conversation, another telephone call.  and he questions my priorities. and he questions the person i am.   \"think about what\'s really important to you\"  \"you fucked up.\" plays about six hundred and seventy two times in my mind.  i let my father down. i let the people i don\'t mean to let down everytime. dissappointed again. racheal fucked up…now just how many times have i heard that? have i said that? it’s not even a valid excuse anymore.  a part of me is trying to move on. a part of me that\'s confident is weary at the same time too. i am not depressed or having a string of bad days, i am dissapointed. maybe because i know i could be so much more. maybe because i know i let myself down on purpose.  i want to go home. as much as there are so many good things here, i want to go home.  i thought maybe i\'d get to go home for christmas break.  i thought maybe someone would say, you know what, screw everything else, want to go home?  because i miss my mom and dad. i miss grandpa and grandma and talking to them. i miss playing with the kids and watching their faces light up. i miss my cousins and my relatives, their screams and shouts. i miss my friends.  i am home sick and it’s terrible.  but i like it here. i swear i really do. i came for the wrong reasons but i\'m slowly beginning to change, it gets hard sometimes. it gets confusing and lonely and frustrating. when things don\'t go your way and when no one understands you. literally.  he said if you go home i don\'t suppose we\'d work out huh?  i don\'t know.  i know i like you. and i know you make me want to laugh and cry at the same time because i feel so amazing and special at the same time. i know i like it when i find your arms around me at night, when i see your smiling face at the end of the line. i know all these things and a part of me has felt like this before yet it is different. hard to explain. do i love you?  is this what love feels like?  i don\'t know.  do i want to stay?  i don\'t know.  i know a part of me has to. i know a part of me doesn\'t have a choice.   i am grateful. but why do i feel so lost at the same time?  i am tired of my past. i am tired of what has happened and i am tired of dealing with things. i just want everybody to forget it and let me try again. i know i fucked up and i\'m sorry and i know i\'ve fucked up a lot of other things to but when the whole wide world constricts you and reminds you constantly that you\'re a failure it gets to a point where it\'s really hard to keep on going. i know i am headed at the right path but you have to stop bringing me down damnit.  i don\'t know what to tell my father  i tried \"i just fucked up.\"  but he just questioned more. and yelled more. and stamped it in more.  i don\'t know how to tell him, i don\'t know. maybe i\'m dumb. maybe i\'m stupid. maybe i\'m not good enough to be your daughter. maybe you should just give me away.  again.  '),(22330,664485,'26,November,2002','2002-11-26 00:00:00','   emo.  i\'d be empathetic, but that\'ll only mean i have to open up further, or care, or risk falling into places i\'ve known before. a full circle? fuck that. this is exactly the same as it was before only now the tables have turned. you talk about empowerment, does it make me any stronger? any more elevated? maybe...or not.   does it make me feel good to see someone i love hurt, no, it never does. as much as i bitch about it and as much as you\'ve hurt me, i\'ll never wish it on anybody, no, not even you. especially not you.  i still care, but some part of me can only care so much before i get hurt again.  am i happy? yes, yes i am. and i deserve to be. i haven\'t been in so long and i know i\'ve come so far. am i proud of myself? yes. for once in my whole entire goddamn life and i can solidly stand here and say, i am enough.  relish all the nights i came back to you and grovelled and cajoled and pleaded. relish all the nights i lay on my empty bed, with my cut up soul, and the dead tone at the other end of the line that tells me you\'re not here anymore. take away all that pain and all the nights and days i\'ve woke up crying, it\'s made me stronger but no, i would not like to ever remember that again. no, not your cold steely voice at the other end, no, not that click, no, not another email, no, not all the times you\'ve put me down and grinded me into pulp, no.  never again the times i\'ve waited for you to come but you never did.   if i was stronger i would tell you fuck that and lets try again. but so many parts of me on so many levels wants never to do this ever again, you don\'t go back running to the stove once you\'re burnt and trust me on that one, i\'m fucking burnt. charred. ashes to dust and all that. a part of me is dead, a part of me is shut off, closed up, ripped apart and it feels good. it feels good to not go through so much pain, it feels good to disconnect, it feels good to move on and let go and not be so vulnerable and fragile.   it feels good to finally have control of my life again.  and yes, it feels good to have someone next to me again. you fall and fall and fall and you try to look for something to hang on to and your strong person and your big guy, the one that promises to be there forever and the one that promises he\'ll love you forever, then you watch as he pushes you away, then you watch as he screams at you, then you watch as he tells you it wasn\'t love, then you watch as he tells you he didn\'t know what the fuck it was.   then you feel that emptiness. then you feel that ugliness. then you feel that dirtiness.   that i am nothing, and disgusting and all my fears.  then you listen to him confirm that.  once.  twice.  three times.  ....i lost count.   then you don\'t invite it because hell, you\'ve never invited it. and it just comes, right when you don\'t expect it, it comes. and so does healing. and so does understanding. it comes and it scares the piss out of you but this time you\'ve learnt your lesson and you say none of that self doubting shit and none of that \'fucked up little child\' business. you work at it and you work at it again and you work at it even if it means you have to reopen certain scars so they bleed just right and heal properly because there\'s only so long you can blame it on someone else and only so long you can feel sorry for yourself and only so long you can lock yourself up in a tiny dark empty cell that lies in the recesses of your brain.  there\'s only so long your prejudices can shut out love.  and you finally realize and know, it\'s okay.  it\'s finally goddamn okay.    urlLink  \"maybe the mistake was assuming that love solves everything. that it takes everything away, makes everything else not matter anymore. \"  '),(22331,664485,'25,November,2002','2002-11-25 00:00:00','   urlLink    urlLink    urlLink     i am thinking about changing my major. i\'ve been having second thoughts about journalism for a while now and after having my papers torn apart consistently for about twelve weeks, i am thinking that i might not be cut out for this writing thing at all.  right now, i am taking biology, biology lab, english 1101, creative writing, acting i, public speaking(shit class), rehersal and performance and am making pretty good grades in all of them. i enjoy school, for the most part, and the big nerdy part of me actually finds all this hogwash interesting BUT i\'m not so sure if i like this whole writing deal anymore.  originally, i was enrolled to do broadcast journalism but i need to take all these basic journalism courses first and damnit, racheal hates history, geography and any other shit of that kind. half the requirements and prerequisites take up about all those damn subjects. pah.  i\'ve been offered a scholarship to be in the jazz choir next term but i am really looking forward to being in the drama department again. i\'ve always been interested and involved in the \'entertainment\' and \'artsy-fartsy\' side of things that to think about a real actual job and covering shit like the elections is exciting but it doesn\'t really make my heart pump you know?  bob wants to know if i\'d like to work on the set again for the next show and that makes me excited. :) i think i am somewhat creative more than i am intellectual. sure, i\'ll sit around and talk to you about seven hundred and forty two real things but give me a paintbrush or a script and i\'m there in a heartbeat you know?  or maybe i just don\'t have much going on up there.  whatever.  well anyway, as i was saying, i have to start registering for the spring term. mr cashwell has listed me in for english 1102, algebra, mass communication, earth science, classic mythology, acting ii and/or jazz ensemble. all my credits should be transferable within the state but i sorta need to be certain about my major and i really don\'t want to do a general thing.  so how?  i can\'t go into theater because to be quite honest, there\'s nothing to fall back into. and the whole point of me getting a degree would be so i had a degree and when i went to go do my shit and get burnt out again and all that fun stuff then i\'d still be able to do journalism or some shit right? wrong.  maybe i should do marketing?  public relations?  god knows i work them prostitutes well.  what i\'d really like to do is join the air force and be a flying paramedic or an emergency peadaetrician...but seeing as i can\'t even spell it, maybve not.  speaking in which, kirk got called for active duty. he\'s leaving for the turkey border in march, i don\'t know about you but i\'m sick of all this war and iraq nonsense.   pah.  p/s sudafed bad. :P  '),(22332,664485,'24,November,2002','2002-11-24 00:00:00','   urlLink    urlLink    urlLink    urlLink      urlLink    urlLink    urlLink    urlLink      urlLink    urlLink    urlLink    urlLink     so the weird day continues..  friday was one of those days where you truly go \'what the fuck\'?  friday i found money, had a long ass conversation about the religion and god, got stalked, scared the piss out of myself, talked to the ex, discovered the new starbucks at bauer, went to watch people  wrestle  beat the crap out of each other in fucking thirty degree weather, got knackered and played pool then couldn\'t tell the solids from stripes, drank too much, got extremely horny, was very emotional, almost got deported, got stuck in a mosh pit, watched him play to a way too packed audience, watched his ex try and rape him, needed a cigarette, fell asleep about forty times in the truck, had a weird ass dream about funerals and weddings in which i woke up freaked out and crazy then he psychoanalyzed me, made hot chocolate and told me to go back to bed, fucked up the imaging on the computer, had my crazy cousin ring me and tell me to go to arizona next month to drive out to la with him to meet the parents and all that good stuff, understood life a little better, learned a hell fuck more, then tried again.  very interesting.  '),(22333,664485,'22,November,2002','2002-11-22 00:00:00','   lol, i just had a weird moment?  hm..franklin and i have one class together. creative writing. and in that class, it\'s like a bitch fest. you basically take everyone\'s work and tear it apart and feed it to the dogs and bitchslap each other left, right and center.  martha is also in this class.  martha is my administrative coordinator. see, people that work on campus are allowed to take classes there as well. so it\'s very strange to have someone like your teacher come down to your level in class and have to critique their shit.   well anyway, martha calls.  it is very scary when martha calls because as nice and wonderful as a person she is, she has also the power to deport me back to malaysia. very very scary.  hell no am i leaving bloody fuck nowhere pensacola now.  but anyway, she wants to know if i\'ll swing by her office on monday.  i don\'t know bout you but a million red bulbs are ringing off in my head.  holy fuck what have i done now?  but she says she\'d like to work with me on my latest assignment because apparently it is good.  i want to laugh.  because in that class i have been reduced to tears definitely more than once without anyone knowing. except the obvious of course but that\'s an entire different story completely.  my stories are shit. i am not a writer. i don\'t write well. i suck at english. and my grammar can go kiss my big fat hairy yellow behind.  so that was a nice interesting phonecall.  and now i\'m all weirded out.  remind me again to tell you about this class, boy are there some fucking interesting stories to tell about that damn stupid moronic dumb class.  people are weird.  very weird.  oh wait, shit, i\'m in that class.  '),(22334,664485,'21,November,2002','2002-11-21 00:00:00','   racheal needs some nicotine in her system. amen.  '),(22335,664485,'20,November,2002','2002-11-20 00:00:00','   in between being sexually harrassed by marc and studying for the dreaded biology exam tommorow, i\'ve just discovered the wonders of taco bell\'s fajita grilled stuft burrito...or whatever the hell it\'s called anyway.  yum.  :)  i\'m bored and sick. he\'s at band practice and hopefully, on the way here soon.   yay!  haha, i\'m retarded and ooh, i\'m vibrating too.   :P  guess who?  '),(22336,664485,'19,November,2002','2002-11-19 00:00:00','   i just spent the last forty-five minutes laughing my ass off. oh yes, today is shaping up to be a nice one...:P  i accidently stumbled in, just running through some links and \"whooaaa\", besides, it was way way too tempting to not go in but yeah, it was interesting. last week though, last week, i would have been all emotional and shit but right now i\'m \'pahhh\'...hehe...which is funny. because he is actually quite an amusing and interesting guy but hey, fucked that up! so yeah..  back to this world:  things are surreal-ly comfortable it\'s weird. not even like an oh shit, i\'m pissed ass scared comfortable but just a \'comfortable\', peroid.   it\'s cute.   :P  i found roses all over my bed the other night with sushi and a midnight walk down the bay with a jazz cd. now, if that doesn\'t make you want to hug the universe and thank all my lucky stars then wonderful intellectual sweet conversation will. and i\'m the lucky girl!   whoop. whoop.  didn\'t do much today, worked on the duets for acting and did some stupid grammar test for english and am now waiting to be picked up so i can crawl into bed and sleep forever. i\'m pretty much tylenol happy right now so it\'s all good.    life is Good......! hell bloody fuck yeah!  '),(22337,664485,'19,November,2002','2002-11-19 00:00:00','   urlLink    urlLink    urlLink    urlLink      urlLink    urlLink    urlLink    urlLink     i guess i needed some time to think about things, just space i suppose and it\'s been pretty productive and comforting...as always, life has been interesting. :P  thank you for all the kind words and thoughts though, the wonder of technology never ceases to amaze me...yay!  we unbelievably sold out both weekends and so it was a hectic rush to perfect everything and all the damn pressure! hence, striking the set was a bittersweet joyous occassion. the people i had worked with were absolutely beautiful and immensely talented and i am fucking proud and honored to have been a part of the cast.  much ad-libbing though was added to my monologue presentation last week but strangely enough, stan enjoyed it so it worked out pretty nice, a!  i\'m unfortunately down with some bug and of all the weeks to choose to quit smoking, it is this week. but i\'m somewhat sticking to my balls and doing this together..so yeah, :S  don\'t know what else to say...have quite a bit of homework to do and lots of sleep to catch up on. hope the week goes good for you!  hugs  '),(22338,664485,'12,November,2002','2002-11-12 00:00:00','   wording out differences, things that make me upset. wording out feelings onto paper, onto screen. this is my exit; the pedestal that i can stand on and proclaim i am free, i am free. shouting raving lunatic, shut up you stupid indecent moronic bitch, this is my place, my world, my corner away from the world.  i wonder if anyone sees me as i unmask and unlayer, make up scrubbed away, digging into raw flesh, i bleed to show you my inner self, to reveal the broken pieces in me and the torn up soul i beat from. love me for who i am i say, leave if it digusts you. dishonesty is a fucked up moronic feeling, dishonesty is a horrid horrid hurtful lie. do not pretend. do not lie. do not hurt.  you are what you write.  i write and rant and go on seven hundred and fifty two times about him and him and him. i cry because it hurts, i laugh because i\'m happy.  a year yesterday i met him, today this year i loved him. the two he\'s, play them together in a quintet and a quartet and string the harmonies together to see the racket that will come out. he tells me to move on and fuck off, i ask him questions that bleed on solid cut off walls that intrude, he leaves again.  but i don\'t care anymore. i cannot care. i don\'t want to care.  i attempt to fix and let go, welcoming changes and letting wonderful exciting things to happen to me.  this harmony, happiness, euphoria seems vaguely eeriely familiar. almost dejavu. do all evil men drive trucks, come from california and have a good sense of humor?  i don\'t know, i\'m about to find out. invest everything in this? hell no. run? absolutely not. try again? say what? then what the fuck are you going to do then?  i don\'t know.  but i\'m trying. and investing. and running.   oh wait, that\'s just me.  dumb moronic fat bitch.   i am happy, but one part of me wants to question \'how long\'?  i am happy, and one part of me says, fuck everything else; i\'m happy!!!!!! fanfuckingasstastic!@!   but it\'s a good day. a quiet day. it\'s a strange day.  one year.  time creeps up on you doesn\'t it?  boo.   urlLink  \"Some days you feel like you are talking to a wall. Some days the wall talks back to you. That\'s when you put down the crack pipe and go to bed.\"  '),(22339,664485,'10,November,2002','2002-11-10 00:00:00','   listen. smile. trust. love.  strong word; funny things that evoke weird things that offset my heart. sparks and chemistry that seem so right and magical scares me. wanting to curl up to you and watch tv forever and play footsie, wanting to have your hand in mine, and wanting you to hold me tighter and tighter and never let go.  giggling and laughing and smiling and feeling so happy it\'s stupid.  i laugh because you say all the things i want to say and mean it. i laugh because there are tokens of what you mean to me all around my room, roses on every inch for separate occasions but everything remains superficial to the other gifts you present me.   you help me believe in things i have forgotten to dream about and make me want to do better, come out stronger and fight. you make me patient and trusting, learning to understand that not everything is always my way and that sometimes, deep inside, i am a someone and that someone is worth it.  you push me forward but pressure me in right ways. taking care of me as you say, i want to say i\'m fragile and watch out because i break and i want to say so many things, things i think about late at night when the world is sleeping and i am worried but then i look into your eyes and feel your body next to mine and the thoughts go away and the hurt slowly fades away.  healing me.  i\'d like to think i\'m terribly courageous in this step but it actually feels kinda nice and good. right.  i don\'t know what grows on what. if love is love or how friendship blossomed but i know that tonight, i\'m just happy i have you.  thank you.  '),(22340,664485,'09,November,2002','2002-11-09 00:00:00','   hmm..word play comes to my mind but a million beats to every sentence kinda makes it hard. it has been Amazing.  :)  i hesistate to take down the previous post because every time i come in and look at it, it makes me giggle. :p  we had a sold-out opening night and pretty much the whole weekend! hehe, it\'s just crazy and funny at the same time i guess, frustrating also because guess who couldn\'t get tickets?!?  haha, nevermind the fact that he took time off work but hah! lol, he\'s been making up for it though so it\'s all good.   hid band had a show on friday night too and i think it\'s kinda cute how we both had good crowds so we were both on this unexplainable typa buzz like the world is so fucking great thing.   damn, i like this guy.  SHIT!  hehe, i take back all i say about men being generally evil because they\'re not. and i\'m just stupid. and retarded.  but it\'s all good because it\'s just kinda nice being just me and not being the insecure fuck head i tend to be so it\'s a really nice experience and ahhhhhhhhhhhh. :P  weeee!  there are a lot of things that seem to hold me back from this but you know how some things just feel \'right\'?  it\'s as if i am sitting here and for once in my whole bloody life i can tell you honestly that i am not worrying it to death, not because there isn\'t a point but because there\'s something about being together and that feeling; i don\'t know if it\'s love, and i don\'t know if it\'s forever, but i know safe. and this feels safe. warm. secure.   very nice.  about time damnit.   :P  '),(22341,664485,'09,November,2002','2002-11-09 00:00:00','   i\'m happy.  very very happy.  =)  yay racheal!  '),(22342,664485,'07,November,2002','2002-11-07 00:00:00','   tired.  i\'m about ready to scream...perhaps even throw a tantrum. which to this nicotine deprived tired mind sounds like a fucking great idea.  i\'ve just taken off only about seven layers of clothes and forty-two layers of icky stage make up and am extremely exhausted yet i still have a paper due tommorow and a shit load of studying to be done. gah.  painintheass.painintheass.paininthearse.  grr.  :(  ....save me;  someone make me dinner? someone make four more extra hours in a day? someone hand me their prozac?  harrumph.  '),(22343,664485,'06,November,2002','2002-11-06 00:00:00','   sleep?  we played pool tonight..it was funny. six games to one, oops. oh well, at least i played eh?  i\'m getting kinda droopy and falling asleep but  rehersals went...okay i guess? stan didn\'t give us notes because of last night\'s event so we only get them tommorow.   so many things to say but not really sure where to begin much less if it should even be talked about at all.  hm..i\'m tired.  goodnight.  '),(22344,664485,'05,November,2002','2002-11-05 00:00:00','   things happen if you try..     hmm..i just woke up.   i strongly, do not at all, recommend running on an hour and a half of sleep a day.  rehersals came out fucktastically late last night because of the run through with the orchestra, we\'re doing costumes tonight then adding make up tommorow and hopefully, the show should kick ass by friday.  i have to say though, it\'s been really nice to have him around during breaks; sanity prevailing and comforting so it\'s good. i guess? :P  we\'re trying to keep it open, fun and getting to know each other right now, sometimes, when we\'re walking there\'s this strange feeling that makes me feel like i\'ve known him forever but it\'s interesting.  bill came to one of our rehersal\'s one night and brought me a sweater i apparently left in his place...nevermind the fact that he\'s from uwf and won\'t stop ringing..stalker alert? uhuh.   it\'s always nice to watch the guys get it on and turn on a notch, sorta silly though but the semi confrontation was fun. at least on my part anyway.  private number good.   :P  over the last week i think i\'ve pretty much managed to wrap my past in a little ribbon and tie and packaged it neatly into the corner. it did sorta hurt and make me cry a little, (okay, maybe a lot) but it\'s all good because i think i\'ve walked away from this a better person, stronger.  friendship runs it\'s test and sometimes you win, sometimes you lose eh?  kick me hard if i start feeling sorry for myself again, i can get kinda wrapped around my fat arse sometimes.  well anyway, i need to go take a shower and put on some warmer clothes...note to winter: kiss my yellow behind.  :P   p/s the server is expected to continue having pms throughout the week or so...so, uhm, have fun? sorry!   urlLink \"in short; i think i\'ve largely got my head around the evolving idiosyncracies of the medium. sometimes i still get burnt. i definitely get the balance wrong sometimes. but i tread the line as best i can.\"  '),(22345,664485,'04,November,2002','2002-11-04 00:00:00','   i wonder if friendship grows on you...:)  it\'s been a weird day..but good though, just strange.  damn, look at me, all giggly and shit...not gooood!!!! oh wait, i take that back, all good baby, all good!  and no, i\'m not drunk.  and yes, i am a slut. deal the fuck with it.  hah!  *pffffft*  '),(22346,664485,'01,November,2002','2002-11-01 00:00:00','   beware of the fat self-obsessed needy bitch.     so it\'s been an interesting night..   lol, i started this post two nights ago and consequently fell asleep midway through a conversation, thinking out this post and d/l the pictures from the camera. someone say exhausted? i think so too.  :) it\'s just been one of those weeks where your days and nights all string together and you\'re kinda like \'what?\' but it\'s been fun. i mean, sure, i\'d like some alone time and time to do nothing but i\'ve realized how disgustingly unproductive and depressing that gets to racheal so i pick this i don\'t have a life outside of all this shit type of thing anyday.  besides, learn to be grateful for the moments eh?  life is really strange right now..but a good type of strange you know? i\'m fucking excited about the musical and i\'m at this stage where things are comfortable and i have my routines that make me feel safe and normal and all that so it\'s kinda nice.   school is coming along pretty well and it\'s hard to believe that there\'s only five more weeks to the end of this term. i feel like i\'ve come so far since before so there\'s this surreal magical type of aura that surrounds it. one year ago today i was preparing to leave for nz and i can still remember all the feelings that came with it...i\'m thinking back to how that feels so damn long ago it\'s crazy.  you know, there are a lot of things in my past that i wish didn\'t happen and a lot of regrets that i wish i could somehow mend or change, but the other night when i was trying to figure things out there was a tinge of realization that everything happens for a reason.   perhaps this happy with life racheal doesn\'t last very long, perhaps the good spells come and go and perhaps i might come back tonight again and be crying and upset with myself and hate the world. but perhaps so many other things might happen too. perhaps i might get hit by a car, or perhaps between now and tommorow six hundred things might have changed, don\'t live for the perhaps and live for the right nows right?  the easiest route is to just give in and give up.  the strongest is to heal.  so i don\'t really understand half of what\'s going around me sometimes..but it\'s okay, because nobody does anyway.   :P     \"I asked God to take away my pain. God said, No. It is not for Me to take away, but for you to give it up.  I asked God to make my handicapped child whole. God said, No. His spirit is whole, his body is only temporary.  I asked God to grant me patience. God said, No. Patience is a by-product of tribulations; it isn\'t granted, it is earned.  I asked God to give me happiness. God said, No. I give you blessings, happiness is up to you.  I asked God to spare me pain. God said, No. Suffering draws you apart from worldly cares and brings you closer to Me.  I asked God to make my spirit grow. God said, No. You must grow on your own, but I will prune you to make you fruitful.  I asked for all things that I might enjoy life. God said, No. I will give you life so that you may enjoy all things.  I asked God to help me love others as much as He loves me. God said... Ahhh, finally!  you have the idea.\"      urlLink \"Basically, the bottom line is this: I am me. This blog is me. There is no variation between the two. I will never, ever back down from my convictions or from being who I am because it makes someone nervous or uncomfortable. Love me or leave me, as the saying goes.\"  '),(22347,664485,'21,December,2002','2002-12-21 00:00:00','   Ever feel like beginning over?  Trying, maybe even accepting everything and just starting again?  I wish I could sometimes.  Erase.  Delete.  Start Over.  Maybe  '),(22348,664485,'19,December,2002','2002-12-19 00:00:00','   i\'d like to believe the world was smaller....i really do.   then maybe we\'d be connected stronger, understood deeper, thoughts decipher. i\'d get you, and you\'d get me. i\'ll fill you, and you\'ll fill me. i\'d fix you, and you\'d fix me. and there would me one soul mate for every one soul, and one lover for every heartache, and an extra smile for every tear. maybe, if, perhaps.   it was dark, grossly warm, stumbling to take a shower, half open, half asleep eyes that slip into the outfit put together the night before. hastily tying and untying, smudging lipstick and dragging a black knapsack with odds and ends sticking out. distinct familiarity, the hushed forced sounds of dawn, slipping away into darkness, a drowning feeling of running away. driving off into nowhere.  birds not even up yet, eyes are tired. tired because it\'s sleep just slightly too deep than a refreshing 15 minute doze and sleep not deep enough to be determined sleep.  cried the night before. made decisions today. after christmas. new year. i\'ll fix my life. i\'ll make new resolutions. i\'ll be alright.   exit ramp onto the freeway. eighty miles an hour. wind in her face, hands on the wheel, music. the damn fucking music; drown out the voices.  exit onto life.  exit into freedom.  exit into escape.  '),(22349,664485,'18,December,2002','2002-12-18 00:00:00','   laughing because she doesn\'t want to scream, smiling because she doesn\'t want to feel, numbing because she doesn\'t want to cry. and no one sees her.   and no one hears her.   she\'s quiet. neatly hidden.   tipsy on the wine. white, transparent liquid that splashes on carpet. falls but the glass doesn\'t shatter. no smithereens to collect tonight, just mere sour stench of pulped grapes and damp odor.   for a moment she is lost. transported. remembers.  remembers picking up remnants of their glases broken on the marbled slippery floor, and how the red spread so fast and flowed so fluidly. it was art in creation and drama in explosion.  tension.  she was going out tonight. been asked. begged.   her last night in town. her new found status of being single, exploiting that freedom. she was happy, she was renewing, she was trying again.  drastically improving. piecing together little pieces to form the bigger picture..she\'d be okay, she\'d be alright.  she was secure.  only that getting ready she stared too hard at the being in the mirror. she found the flaws again, she found discrepancies and ickiness.   holds that small torn notebook lined paper. numbers hastily scratched on to it.   tired of feeling like this. tired of trying trying again. tired of excuses and lies and bullshit.  just call the number.  just ask for help.   fix yourself and do one solitary right thing.  going insane. so many questions. too many answers. nervous breakdown reaching it\'s climax.   i just want to be okay, i just want to be fine.  479-9411.  i want to. i know i need to. i have to but something stops me. something scares me. something about coming out in the open and saying i have a problem messes me up. something about fixing this then fucking up and having nothing else to blame but me. something about being afraid to realize that i might not be strong enough for the consequences, the truth. that i might not stand alone secure.  and i\'m lost.  confused.  and i\'m small.   big.   tiny.   fat.  and i\'m tired.  alone.  hiding.  where\'s my season of hope?  '),(22350,664485,'18,December,2002','2002-12-18 00:00:00','   just three decibels too low, deep...soothing. an angel\'s decree from heaven, let there be light, let there be peace. sleep tight now my child, sleep tight.  it\'ll all be over soon. it\'s just a bad dream, just a bad night, just strange situations and not enough inside to fight. just intoxicated and looping.  fallen streams of salty teardrops caress fleshy cheeks, lose twenty pounds she tells herself. criticizing, analysing, destroying.   she has so many fears. too many fears. why does she feel so schizopherenic? so manic depressive? the extremes, the pretending, the violent extremities....laughing because she doesn\'t want to scream, smiling because she doesn\'t want to feel, numbing because she doesn\'t want to cry. and no one sees her.  and no one hears her.  she\'s quiet. neatly hidden.  drip. drip. drip.  crimson blood flows, hurting herself, tearing apart flesh and wound, differentiating, or at least attempting to. which is physical and which is mental? torture. abuse. punish yourself you sorry excuse for a human.  packs her bag. not going home. going away but not going home. alone. alone. alone.  christmas presents. more christmas shopping. tidy bows and colored tissue. tis the season to be merry. forgiveness. forgive yourself.  dream. let be. fantasize. carpe diem.  she knows who she is yet large overwhelming questions cloud her mind. searching for herself. searching for bits and pieces to string together. looking for the bigger picture, putting the puzzle together. clearer now, clearer. haunt me, don\'t distract me. pull me, don\'t annoy me. try me, don\'t abuse me.   tipsy on the wine. white, transparent liquid that splashes on carpet. falls but the glass doesn\'t shatter. no smithereens to collect tonight, just mere sour stench of pulped grapes and damp odor.   for a moment she is lost. transported. remembers.  remembers picking up remnants of their glases broken on the marbled slippery floor, and how the red spread so fast and flowed so fluidly. it was art in creation and drama in explosion.  tension.  but she puts on the happy face for them to see. she always has a face for anyone. the you can\'t touch me looks, brushes away the sweltering harassing remarks from far too overgrown men, the kind she attracts all the time anyway. fat. old. rich. balding. men.   and the ones that are worth it she pushes away. shies away at even the slightest chance, flirtation, banter. she\'s single again, they chase, she runs. further and further away it\'s not even a game anymore.   it\'s some sick deprevation in her mind.  it\'s some solitudal need to be independent crap she\'s drilled into her mind  yet she wants to love. she wants to be loved. but it comes out foreign and weak. she\'s not used to being so open, she\'s not used to expression, she\'s not used to love.  what is love?  the bolder stronger renewed side of her wants to find out. accepts the offer.   she is dressed. preparing for a good night out, the last in town. preparing for lots of dancing and swearing. for waffle house at three and too many ciggarettes, too many drinks. yet getting ready though, getting ready slowly disturbs her. and she doesn\'t feel like it anymore. and she looks, and stares. she sees ugliness, she sees pain, she sees the things she cannot hide.  she needs to lose weight. she needs to escape. she needs the answers.  she needs help.  479-9411  '),(22351,664485,'17,December,2002','2002-12-17 00:00:00','   ahh..there\'s absolutely nothing like driving, shopping, and queue lines the weekend before christmas.   kinda makes you jealous of me doesn\'t it?  cause i get to sit around in traffic in seventy eight degree weather (yea, you go figure, thirty for a week then it\'s in the seventies), listening to the harmonious sounds of honking traffic and broad daylight finger excersizes while carefully analyzing the shitty sorry excuse for christmas music the little town of three actual listenable stations play while ever so gently realizing the itch and sudden urge to pee.  yes, yes, there\'s just absolutely nothing like christmas.  oh and btw, would someone please tell santa i\'d like world peace for christmas?  thank you.  :)  '),(22352,664485,'17,December,2002','2002-12-17 00:00:00','   Maybe it was because I read what he had to say about us. And I felt that shitty icky crap that always seem to come hand in hand with it and realized, rather pathetically, how different I’d become yet at the same time, how emotionally sort of strange I still was in a way.   There’s always going to be something weird and strange about jared and I, perhaps even some sort of cosmic fatal attraction type of drama but something about Saturday night made me realize that as soon as I said the words, “maybe” I fucking ran. I ran and screamed and yelled and realized I couldn’t and didn’t love him in that way anymore. Perhaps a cling to desperation and wanting to soothe healing wounds. You know how when a wound heals it itches and the battle is not to scratch it? Exactly.  A short, somewhat short anyway, recap of the last few weeks is complicating and weird.  I am single. And I can honestly actually tell you, I’m okay with that.   Somehow, it wasn’t that franklin invaded my privacy but rather the attachment of having someone so close while not really having it all together really bothered me. It wasn’t that I didn’t have fun, or that I did not feel comfortable with him or trust him but rather a this-is-cute-but-really-what-and-who-are-we-kidding type of thing.   In a bid to escape everything that seemed to close in and constrict me closer and closer, I drove out to Mobile on Saturday for a concert. I am pseudo and at times fanatically religious. I may not practice a lot of the teachings and my faith may get broken once in a while but I believe there is an immense eternal being out there and all that good shit. It was uplifting and at the same time terribly eye opening, I wasn’t sure if it was just because the person we had gone to watch was the same artist that I saw in Malaysia or if it was just the being able to step away from all this shit and think thoroughly thing. Well anyway, four hours of driving and being in some place else and thinking clearly for once in a long while was good apparently.  I came home, physically whacked but mentally fresher than I’d been in a while. he drove over needing to say the things he had to say and so we talked. And I explained. And somehow laughter, smiles and a big great hug emerged out of it. Minus a pillow fight, lots of tickling, and midway distraction of course.  But anyway, it was weird because everything was out in the open rather that understated subtle shit that had to be read in between the lines and it felt really good. I watched his truck pull away and it wasn’t like the other night where I had told him I had needed some time alone but it was rather quite drastically different.  Then I went in to ring jared. Yeah, I know. smart, racheal, really smart. Doh.  I rang perhaps not realizing all the things I had told myself before. Lost in some moment I suppose, that desperation I clearly remembered and talked my ass of for a good few minutes into oblivion and it ended. Somehow, having said everything and opening the door had brought in closure.  I cried and sobbed and did the whole she bang. Feeling sorry for myself because I couldn’t go home for Christmas and just letting go. And it felt good. Hell, it felt really really good.   Marc shot me yesterday and I somewhat enjoyed it. It was weird and different, yet it was pleasing in some odd sense. I fucking aced the biology paper which seems so ironic when it was to be one of my most difficult subjects and it’s nice how this term has wrapped up and ended.  I think I made some pretty good grades and there are a lot of little side projects I have going for me and being planned that it’s nice to look forward to.  Tomorrow I leave for a long awaited vacation to Orlando and Melbourne, which is I guess rather coincidentally planned to fit quite nicely into my frame of mind.  So nothing has turned out like it was supposed to and nothing ever will you know?  Just go with the flow I guess  Strangeness, but I actually think I’m happy.  Madness.  '),(22353,664485,'17,December,2002','2002-12-17 00:00:00','   [12/17/2002 9:40:16 PM | + racheal] shagging. I feel sick.   I want to go throw up and throw up and throw up.   I don’t know what it is…remnants of dinner flushed down with red wine and biscotti feels trapped in my throat. But that’s the way it always feel, that’s the way the story goes. Right before I do myself the glorious perverse pleasure of pumping the volume up to stick my fingers down my throat and drown out the gagging. To drown out the sounds of making myself sick, simultaneous relief and disgust at the same time. I’m making myself sick.   I’m late. I’ve never kept track, never worried about it. I’m not a sleeper, hard to believe but I don’t exactly fuck around. No, it doesn’t mean that I’ve never had sex, it just means that the meaning of a casual fuck has never quite rang anything to me, make love though…make love I have. Only that it’s different, I’ve never had to worry about the actual conceivement of an unwanted child.   I’m late. And it scares the piss out of me. I’m stressed and I’ve half blown smoke up my ass convincing myself that stressing it out, being underweight, and having a shit system for a body contributes to it. Abortion. Fucking A.   I’m not pregnant. I’m not pregnant. I’m not fucking damn pregnant.   I just cannot have that much of amazing luck. To an extent, I think we have been very safe, yet there are still questions. Loopholes. Condom? Check. Spermicidal? Check. Birth control pills? Not checked. Heck, the last thing on earth I expected when I came here was to have a relationship much less sleep with someone so why would I have brought my prescription right? Wrong. I’m a dumbass.   I had one of those piss on a little shit piece of foam and let it tell you if it detects hormones or not tests but it came out negative. Those three minutes were my three  minutes in hell. And I’m still late.   Why beat yourself over such a simple matter right you drama queen bitch.   For one, it’s not simple. In fact, it’s about as complicating complex as it can get.   Add in an ex-girlfriend, an ex-boyfriend, lots of shagging, and the whole thing about wanting to just be friends into it and you get a turmoil of bullshit and drama.   Only that I’m the major one freaking out.   Why?   Because as always, I’m the only fucking person that knows.   Hell, I could always just ring Her up and tell her right.   Or Him.   Or the Ex.   Whatever really.   Oh don’t mind me. I’m just the crazy insane bitch freaking out as usual.  '),(22354,664485,'17,December,2002','2002-12-17 00:00:00','   sanity prevails every once so often....i have a shoot today, somehow, that brings back distant memories which i am unsure if it would be better left forgotten.  we\'re going to fort pickens by the beach. i have three costume changes, two looks, one set. marc is photographing. salty air, breeze and finally, some warm sunshine. my only requests have been roses and fruit. maybe even alcohol. but then i\'ve been there before, boozed up on yet another bottle of chardonnay, puffing on another pack of ciggarettes, bitching at another placid flaw of the human system.  i have pretty nails today. a nice new black leather coat, dyed jet black hair (failed attempt of red streaks) and a cute ass skirt. i aced my biology exam today, the finals, the last one i\'ll ever have to take and breaths of relief choke me. yes, i am that pleased. i walk down the hallways and guys whooop but my new shirt says, \'i want attention...but not from boys\'.   apparently, i am gay now.  rainbow, fruity, happy vixen.  or maybe i\'m bisexual.  whatever really. label me whatever you want, however you please.   is this position good for you?  '),(22355,664485,'14,December,2002','2002-12-14 00:00:00','   i don\'t know why it stopped. or how it ended. or why we became what we became.   i wish i did though, i really wish we tried again. i wish we got ourselvews together in time to save us. but last night you left, last night you hurt me again like countless other times before, and i don\'t wonder if my disconnection this time will really last at all because maybe i\'m just tired of thinking.  maybe i\'m just tired.  i\'m really tired and sometimes i really wish god will just stop and let me off this world a while. sometimes it\'s hard and there\'s so much damn pressure.   i don\'t exactly know how i feel about me right now. i feel ugly.  i am disgusted.  disgusting.  revolfuckingting.  small. i feel really small. and insignificant. and stupid. and everything you said frokm before.  i feel pointless.  am i pointless?  '),(22356,664485,'12,December,2002','2002-12-12 00:00:00','   urlLink    urlLink    urlLink    urlLink       urlLink    urlLink    urlLink    urlLink       merry christmas and have a safe new year.  '),(22357,664485,'10,December,2002','2002-12-10 00:00:00','   fucking finals.   excuse me while i shove my head up my layers of fat and scream.   oh, and uhm, have a good one too.  '),(22358,664485,'07,December,2002','2002-12-07 00:00:00','   changes.    urlLink      pick up your sorry ass from the floor, racheal ; it\'s starting to mold.  '),(22359,664485,'07,December,2002','2002-12-07 00:00:00','   It always seems easier to refer to yourself in third person. Somehow, it seems more distant, less painful, not so real. Maybe?  So they talk today. They talk and somehow she finds herself crying and sobbing. Breaking down. She hasn’t cried like that in a while, not since him. Not since she told him how he hurt her, not since she remembered how he hurt her.   She tells the boyfriend, just like she told him, that she’s not strong enough for heartbreak again. Promised herself and owes it to him to not have extremes, but knows enough to understand and protect herself yet she’s a different person now, she’s stronger.  But why does she question though? Why are there always some nights where she lets him get to her? And why are there always some nights she feels so terribly insecure?  Dumb really.  The boyfriend wants to help her. Begs her to stop. Watches her, attempts to guide her, sometimes even intrudes into her bubble uncomfortably, desperate to heal her I guess. “Don’t do it baby,” “It’s going to be okay, I fucking swear to you,”   Bulimic fucking reject.  She feels ugly and old. Fat. Oversized and disgusting. Terribly disappointed in herself and the person she is, the person she became.   Why does she let him get to her? Why do the wounds still bleed when he attacks? Why isn’t there such a thing as forgive and forget?  She watches.  She sees.  She understands.  Time to move on, these walls have crumbled and its foundation torn down. He hurts her again.   Live and learn.  Let.Go.  '),(22360,664485,'07,December,2002','2002-12-07 00:00:00','   green goblins; knock yourself out with tylenol.   hmm..the boyfriend has a show tonight.  but racheal gets to sit at home, play video games, study for the finals and harvest all the good shitty viruses in her.  how pleasant.  ahh...life is stupid.  :P  '),(22361,664485,'06,December,2002','2002-12-06 00:00:00','   just a thought...   my father used to make me chocolate-chip pancakes with little smiley faces on saturday mornings.  i miss that.   hmm...do you know if  chocolate-chip pancakes  daddies are fed ex-able?  '),(22362,664485,'06,December,2002','2002-12-06 00:00:00','   wise words for the day:-   life is like a dick; when it\'s hard, fuck it.  '),(22363,664485,'03,December,2002','2002-12-03 00:00:00','   hiatus.   finals + trying to sort my life together.   do the chicken dance with me?  hugs  p/s be good!  '),(22364,664485,'03,December,2002','2002-12-03 00:00:00','   construction.   he rolls over. half asleep. half dreamy and silly and whispers those three words.   i laugh, snort, giggle and scream at the same time. wanting to yell at him and say, \"what the fuck?\" but you know what they say about people being asleep and being compared to blatant honesty when you\'re drunk and it\'s slipped out.   holy shit. holy shit. holy shit.  ----  breaks down today, tears in his eyes and hurt aching through. i want to hold him in my arms and take away all the pain and draw him close and soothe him till all is gone. his one and only grandparent left and now he\'s alone. pallbearer he tells me. i say it\'ll be okay and he takes that long drive there and back, sleeping in between heartache and headache. it sucks to be sad and sick at the same time and i wish i could find the words. i\'m sorry.  you don\'t have to be strong all the time, it\'s okay. i promise.  ----  finds me when i\'m tired, cranky and sick of the whole fucking world.   when someone else has broken me down, he picks up scraps of me to dust me off. and we talk.  we talk like we\'ve never talked before.  we talk about life and love and letting go and moving on and mistakes. finality and all things that must come to an end. we talk about  my addiction and his. we talk about our demons and the things we fear. but yet we laugh a lot. we laugh till our sides hurt and laugh somemore. we laugh because it\'s silly. we laugh maybe because it\'s love.  ----  he bangs and sings and strums and jams. he plays to his little sold out crowd and signs for his little fans.  runs around and sets up and catches me in between cigarettes and photographs to sneak in a kiss and a hug or two. and runs along again doing his own thing so i have time to do mine.  and he tells me he believes in me and my dreams and that he knows i can do it. and we talk about two years from now. and we talk about our careers and lives we want to lead. and i tell him how immensely fucking talented he is, and he thinks we happened for a reason, that i\'ve completed his life. but i know, no, he\'s completed mine.  he\'s made me finally worth it.   i thought friendship meant more things. maybe i thought it would be stronger, that the part about being friends come what may would ring true but last night, desperation on some new level won out and sold out. i\'m tired of backlash and nit-picking, saying the wrong things and doing worst. it\'s over.   perhaps we should leave it at nothing.  you break and break and let someone break you so many times it gets old. redundant. same bloody damn story again. only that i\'ll try this time, and fight back and tell you no, not this time. not anymore times.   funny how you put the question in my mind. the what if. the maybe we just needed to go do our own things on our own for a while and come back and it\'ll be perfect only i realize now it\'ll never be perfect. there\'ll always be something wrong, and something fucked up and something not right.  you\'ve never seen me smile or heard my laugh. you\'ve never held my hand or wiped away a tear. you don\'t know the things i do nor have you tasted my cooking, heard me sing, watched a show. i\'ve never seen you, never touched you, never really known you at all. maybe we\'ve been caught in a fucking lie, perhaps you were right after all, maybe we weren\'t real. maybe it was just bullshit.  well keep your bullshit then. i\'m done with you.   and i\'ll work on me now. the fucked up little child gets help. the fucked up little child is finally going to be something.  so i get depressed, so i get extremes, so i throw up and starve myself. so so so...so what?  i\'m tired of this life and i\'m tired of the person i am with you. i\'ll find me a new life now. i\'ll make me a new life.  i\'ll change and fucking bleed stuffing myself and smiling myself full. i\'ll learn about confidence and about being stronger because i know i\'m strong enough for this. i know i\'ll fucking make it and i know i\'ll bloody do this.  and you can stand on your edge and laugh at me. and you can stand on your side and say more hurtful things and pull me down a while but i know now that being nothing and everything at the same time is not what i want.  you question why i ring you. stupid me thought i was comfortable with you, stupid me thought we\'d actually make it this time. well fuck that, fuck the past right? simply just forget and let go eh?  and i remember you telling me that you hope im sorry someday.  and i am.   i\'m sorry because you\'ll never see me for the things i am and the things i can be. i\'m sorry that revenge gets in the way of trust. i\'m sorry you\'ve lost my trust and hurt me again. i\'m sorry this is again.   call this immaturity. call this being the bigger person. call this finally doing something we\'ve talked about forever. call this letting go.  i know what i want and what i need now. i think i\'ll go away a while and work on me, and dreams i\'ve finally let myself dream again.  endings don\'t always need to be painful.   they just need to be simple.  goodbye.  '),(22365,664485,'31,January,2003','2003-01-31 00:00:00','   About five minutes ago, I was horny. Now I\'m just trying to stay the fuck rational. Hell knows I don\'t want to go to sleep crying tonight, after all, the Muse is happy tonight right? Was anyway.  Mind against Heart.  Don\'t let it get to you, let it go.  It\'s not worth it damnit, it\'s not worth it.   I try to remember all that was said in yesterday\'s session, how blatantly honest I had become in those few minutes. I found all the answers I had been looking for, subconsciously knowing they were within, perhaps just not knowing where to look.  Perfect tore away at Passion, I like all the parts I had talked about changing. Essentially a good person really, just gets too carried away. Guarded now, Control has to lay frigid hands to rule, keep in check. Never try free falling again, why pick away at old scabs?  A test. A measurement to see how far you\'ve come...Sorry some day. Tommorow, Yesterday or next Week? Forgiveness is a key to Kindness, she tells me Mind needs to be gentle with Heart. That the insecure messed up person in me is really okay; that session scared me. Opened me.  The initial high came crashing down at  urlLink Marc\'s , thoughts jumbled up and a meal desperately attempting to be digested. Palm on skin, tumor beneath flesh, sanity prevailing sinking on the line.  My insecurities can bite my ass, tonight I am Enough.  Tomorrow I am Beautiful.  Forever I am Special.  I believe that, I think I might really.   I am good enough, strong enough, Enough.  and I won\'t let you convince me again I\'m not  ...even if you still Hurt me  Tear me apart, feed me to the wolves;  So I am evil, so I am bad, so I am all the things you constantly tell me I am, so what?  at least I am still Fighting, and I will still hang on to my Dreams.  You have no idea how precious these dreams are to me...Life.     urlLink \"let it go-the smashed word broken open vow or the oath cracked length wise-let it go it was sworn to              go   let them go-the truthful liars and the false fair friends and the boths and neithers-you must let them go they were born              to go   let all go-the big small middling tall bigger really the biggest and all things-let all go dear             so comes love\"  '),(22366,664485,'30,January,2003','2003-01-30 00:00:00','   I’m busy, busy, busy it’s not even funny. Hell you try maintaining eighteen credit hours, getting fit, keeping the social ‘in’, staying sane and doing the whole movie thing. It’s interesting, crazy ass hectic and tiring as stitch but all good honey!  Auditions on Monday was just absolutely beautiful…the director seemed really nice and it was the best feeling walking out of there knowing I kicked ass. You know how racheal’s confidence level gets but dahyum, that was a good audition. :P  It’s funny how people try to hide how excited they are and mask it with all the talk and crap and you’re sitting there just thinking, hell yeah, you want me don’t you? * giggle *  I do have to say though that being Asian helped seeing as I didn’t see too many other Asians audition for Thai’s part and it is supposed to be apparently written for an Asian girl but I won’t get too excited. :P  It’s nice to feel so happy though. Proud, like I did something right and it feels really good!  Whoop. Whoop.  Now, to memorize those two damn monologues, 55 pages of homework and dinner tonight.   Hehe, god I’m delirious like a little school girl!  :P  I can’t begin to tell you how good it feels to be smiling again after that god of all fuckers week I had…we did this whole new approach with Gayle last night and that whole social stigma thing about seeing a therapist? I can only say that up yours because if I get to figure out as much as I did last night and feel as good as I do now, I’ll deal with it!  :)   urlLink  \"Our bodies are made for pleasure. For the pleasure of breathing freely and walking and climbing and dancing and touching others. My body is made for pleasure.\"  '),(22367,664485,'29,January,2003','2003-01-29 00:00:00','   Sometimes you don’t need reasons, you just try.  Everything fades and reappears again, reality searches back for you and peels away layer after layer. Orange skin that dehydrates by the sunlight by the mirror, reflection stares back at you and you break down.  Invade me, come into my corner and join me, understand me.  sarah, where the hell have you gone child? Where the hell have you been?  I used to know you when you were different, when you’d twirl around in your little pink dress and sing to yourself. Hum a tune and prance around with so much joy, so much heart beneath those eyes. What’s wrong honey, what happened?  I watch you now and you are so broken, so confused. I watch you tear at your hair and push away the tears, the anger, the hurt. Mind versus Heart. You question, I listen to you abuse, rack yourself but the eyes never lie honey, sound travels. You pretend you’ve hidden it so well, but you’re still so naïve child, still so much to learn.  '),(22368,664485,'29,January,2003','2003-01-29 00:00:00','   Sleepy as fuck and going to bed.  Thank you, you.   Maybe I needed that good cry, or you behind the door knowing I was hurting and begging to be let in. Maybe I needed that extra pack of ciggarettes and that violation of trust I fucked while yelling at you. Screaming at you not to touch me and just go the fuck away then driving back to your place many, many hours later to tell you I was sorry. The glimpse of acceptance in your eyes that welcomed me home, as you shared my phone calls and celebrated in my joy.  You are a gift to many, your friendship is as fragile as porcelain as solid as rock and I am thankful.  You are beautiful. I have your fingers tugging at my heart, your hands at my soul and you are among the reasons why my living hell seems so much more calmer when you are around. I watch while you are singing, proud of you...jumping up and down saying, did you see that? did you see that?   I hurt while you are hurting and I\'m sorry he still breaks your heart time and time again. The painful tears you cry when you are tired of the world makes me hurt too. Fathers are a weird specimen on their own aren\'t they? Don\'t let him take you down silly, don\'t shy away.  You are enough and complete as a person. Your dreams are still there just like they were before. You make me a better person, having you by my side teaches me I am worth it. I tell you these things, begging you to constantly believe me yet I violate your trust too when you mention his name. When you bring up pain I turn around with spite and bitch.  I\'m healing though, and I know you see it too. One day, I knew I\'d see just like you saw me too.  Thankyou.    \"she made me a sandwich she walked with me she let me touch her fingertips...\"  '),(22369,664485,'29,January,2003','2003-01-29 00:00:00','   yellow monkey haven.  '),(22370,664485,'28,January,2003','2003-01-28 00:00:00','   urlLink  “You are a child of the Universe... No less than the trees or the stars themselves, You have a right to be here.”    Nightly visits from the Shallot, my heart dreams of breaking free from this tirade, this charade. Open masks that melt like chocolate…so sweet, so wanted. Eternal ecstacy that breaks from happiness, the need to get stripped and naked torments. The need to peel off the layers and come home hounds; who am I exactly?   urlLink Colorgenics  tests I make everyone take, he vehemently denies my theory…only to find an email telling me the truth. Right, right, right, wrong. Honesty is beautiful isn’t it?  It’s as if we are all hidden in our own layers, all locked up into our own fears. You aren’t the only one, you are not alone, it’s okay. Being there and feeling safe, the moments he sheds the image and shows heart. You find the not so strong, not so big side he’s willing to show you…and it is beautiful.  Midnight beach football, soft Spanish music that plays in an old forgotten diner. He sings you beauty in a foreign language, shows you love in friendship.   Broken down again, in the middle of a crossroad trying to find your way. Your place in the world, your mark, your strength. So many externalties cascade you from different angles, and so many fears compound your insecurities.  Casting call today, headshots prepared with only the flaws apparently only your eyes see. Tick, tick, tick, the seconds loom. You are afraid, excited, monotone. Worked out the stress and pressures, ex-boyfriend on campus, gym equipment at my disposal, self-abuse in regalia. Getting in shape, kicking some ass, healing with love.  Safe people surround me, carry me when the days are bad.  I am a child of the Universe yesterday, today and tommorow.  I am a product of Love.  Loved.     urlLink \"It is here for no reason but to make you ask for a reason.\"  '),(22371,664485,'27,January,2003','2003-01-27 00:00:00','   “her ex boyfriend (the one who treated her like shit no less, raped her, and dumped her when she had cancer)” Familiar?    I am immortalized by his being, his words ignite fire in me. Translucent and clinging to the air I breathe, clasping on fluidly to gentle words, lips forming sounds, coherent thoughts, indecision, and conflict. Drama, bigger than life itself, I smile.  Sexual tension is beautiful. Selfish, but giving, Frustrated, but understanding, so everything not only harbors on Lust…don’t want to strip again, don’t want anyone to come close.    I am angry. Incredibly frustrated with myself, my being. Not able to forgive and still affected ; he breaks me down again. He pierces through the layers the others have helped heal and stabs again. The tears fall, against my will, swipe away…swipe away…fall. You’re not worth it. And that’s all that’s left chanting in my mind, nevermind how far I’ve come, his disapproving, debilitating stare has gunned me down. Pawn.    He holds me, hugs me. Rubs my back when my heart pounds, pounds, pounds against my will. As if it will explode any second and my chest contracts beyond all pain imaginable and I cry. I scream inside but physically too scrunched up to do anything. Only these attacks tell me I am still alive, only these attacks tell me I still have a heart. I cling on to comfort, but shield away from security. I’ll build my own walls, thank you very much, I’ll fight my own battles. Alone.     I am left in the center. The beginning of an ending that feels shortchanged. We are prisoners locked in our own worlds, but yet he chooses to take my key…or maybe I gave it away too soon. Or maybe I’ve lost it forever. He’ll always hold it won’t he? He grins. Superior power in his hands ; and he knows it. That ability, the one he questioned me about, Revenge. Ego. Submission.     We drive. We drive for hours and hours, on long winding roads, and half assed bumpy paths only to find ourselves on this never ending journey. I have found a friend, I have found someone that cares, loves…but I cannot seem to love him.  I let him in, but with every brick that I take down, four goes back up. It’s as if with everything emotional, everything that constitutes thoughts and a heart; I take one step forward and twelve steps back.      I throw sticks and stones. Wanting to turn my back and not care when he’s hurting, not be bothered when he questions.  I remember Christmas, and New Year. I remember he is selfish and unforgiving, spiteful and indifferent. That he hates me and I deserve better, deserve someone who loves, believes, trusts in me.      Go Away. Run. Leave. Let Go. Search. Move on…whatever.    It’s as if God looks down and mocks me. Irony spears through and little boys and little girls of every type and nature feed off this insecure, broken, extremely confident, happy, sad, crazy aura that surrounds me. I am beautiful in disharmony, strong in breaking down.  Attractive? Scary.  I’m indecisive, I don’t want to make decisions…commitments. I run,  and even they know it.    Just thinking out loud.       urlLink  \"The sounds of water, it sings to me; the smell of water, the fragrance refreshes my soul; the silky soft current weave in and out and in and out gracefully from between stones, I’ll never forget that beauty. I laugh, I smile, I drop the pretty wild flowers in my hand, I dip my fingers into the cool water, when I look up at you again, and you’re gone already.\"  '),(22372,664485,'26,January,2003','2003-01-26 00:00:00','   Yeah, yeah, yeah.  :P  Friend of a friend of a friend is coming over.   Hm.  '),(22373,664485,'25,January,2003','2003-01-25 00:00:00','   LAST MOVIE YOU SAW IN A THEATER?  gangs of new york ; violent, graphic, beautiful, pointless.   WHAT BOOK ARE YOU READING NOW?  absolutely nothing.  though i just finished  urlLink  augusta, gone  by martha todd dudman  ; interesting poignant emotional shit   am waiting for  urlLink amazon  to send me a couple of books i ordered the other night..  urlLink  his bright light  by danielle steel  (gave my copy away to my therapist in malaysia as a parting/thank you gift – the ultimate i’ve let you in to my world thing ; I write by the borders thoughts that run through my mind, pictures and shit like that or however the writing is affecting me type of thing and I read that during a very weird time, trying to understand depression, coming out of aneroxia and dabbling into bulimia, being fourteen and absolutely convinced the world was against me so it’s got a lot of honest hurting me in it and la di da. i miss having that copy but i’m happy it’s with her too…it’ll be fun to compare the scribbling from this to the other one ; have i changed at all? )   urlLink  prozac nation  by elizabeth wurtzel  (looked interesting, been hearing quite a bit about it, got  urlLink rementioned , peeked at it, liked it.)   urlLink  wasted  by marya hornbacher  (i normally get a little intimidated by what i read if only because i’m afraid it fucks me up further, but the internet only tells you so much and I’d like to beat this thank you very much so bring it on I say)  it’s rather morbid and dark reads i know, but i’ve always like reading about people…or maybe it’s just the comfort of knowing i’m not the only one eh?  i don’t know : stupid girl.   FAVORITE BOARD GAME?  scrabble…taboo..la di da…might actually need the time on hand to play    FAVORITE MAGAZINE?  in style, cosmo, q, glamour   FAVORITE SMELLS?  the rain, hugo boss, coffee, fag?    COMFORT FOOD?  * smiles sheepishly * i don’t know really, waffles? fruit? pasta. pop-tarts, cookies and pretzels.   WHAT ARE YOUR FAVORITE SOUNDS?  waves crashing onto shore, quiet tranquil relaxing music (think jazz/alternative), the hum of my car on drives, the steady background noise (television/music) to soothe/distract/evoke me.    WORST FEELING IN THE WORLD?  not knowing what’s wrong with you, having no diagnosis and getting more paranoid by the second…midnight crying.   WHAT IS THE FIRST THING YOU THINK OF WHEN YOU WAKE UP IN THE MORNING?  here we go again..   WHAT IS YOUR FAVORITE FAST FOOD PLACE?  good question. hmm…oh wait, i knew this, taco bell!      FUTURE CHILD\'S NAME?  little boo and big boo   WHERE WAS THE LAST TRIP YOU TOOK?  wisp skiing resort, maryland   FINISH THIS STATEMENT, \"IF I HAD A LOT OF MONEY, I WOULD....  start a foundation, bone marrow research, make-a-wish, buy the new merce, start a production company    DO YOU DRIVE FAST?  is the pope catholic?   DO YOU SLEEP WITH A STUFFED ANIMAL?  i sleep with boys and i sleep with girls but I don’t do the animals alright?   WHAT DO YOU EAT FOR BREAKFAST?  coffee, juice, pills, snack bar   FAVORITE DRINK?   non-alky:  it wavers between the tazo ginger peach tea and my good ol trusty skinny café mocha  alky:  corona? i’m going through a beer phase..nooooo   FINISH THIS STATEMENT, \"IF I HAD THE TIME,\"  i would go on a holiday   DO YOU EAT SUSHI?  yup   IF YOU COULD DYE YOUR HAIR ANY COLOR, WHAT WOULD YOU CHOOSE?  hmm..red with orange streaks.   HOW MANY DIFFERENT CITIES/TOWNS HAVE YOU LIVED IN?  how long do you have?   DO YOU BELIEVE IN GHOSTS?  somewhat…i have my beliefs…just like the green men.   WHERE IS YOUR FAVORITE PLACE TO RELAX?  it depends, sometimes it’s home, sometimes it’s the bay, sometimes it’s the beach and sometimes it’s the coffee place.   FAVORITE SPORT TO WATCH? On TV?  x-games, and yes, it is a sport.  live?  english football   ONE NICE THING ABOUT THE PERSON WHO SENT THIS:  loving and always there   WHAT ARE YOU MOST THANKFUL FOR?  eighteen years, family, friends, the kids, Hope.    OH MY GOD - scariness.  Results-   urlLink  \"At this time you are really feeling quite exhausted by all the conflict and quarrelling that is going on about you and you are looking for some sort of protection from this state of affairs. Ideally you are seeking a peaceful condition and a tranquil environment in which you can be afforded the chance to relax and recover.  You are looking for excitement and stimulation and you are ready to try anything - but be careful not to take too many risks.  You are a perfectionist in everything that you put your hand to. You are demanding and very exacting in the standards you apply to your choice of colleagues and friends -perhaps you demand too much from people. That perfection you seek in a particular person is illusive - perhaps it does not even exist.   From every direction there appear to be unwarranted restrictions on your freedom of action and this is producing considerable stress. You\'re really looking for independence and freedom from any restriction and therefore avoiding any obligations or anything which might prove hampering. You are being subjected to considerable pressures and you would like nothing better than to escape from them but you tend to lack the necessary strength of purpose to succeed in this. Whichever way you turn you are being frustrated. You need to be free to do your \'thing\' in your own way.  You are fed up with other people trying to influence you and you also feel that it is necessary to protect yourself from the threat that your independence and freedom may be restricted. You would just like to be left alone.  Be it through unfulfilled emotional requirements, whatever the circumstances you are experiencing considerable stress, be it mental or physical. To your credit you are attempting to escape from this by endeavoring to create a semblance of peace and serenity by refusing to allow yourself to be involved. You have the strength to \'pull through\' and all indications are that you will... perhaps sooner than you even believed possible.\"     er, yeah, exactly.  '),(22374,664485,'25,January,2003','2003-01-25 00:00:00','   I would laugh, or on normal shitty days, let the tears fall…but maybe you build up mechanisms in you to shield you from all the soul-searching, heart-wrenching bullshit.  Tonight was…weird. Weird into a new context.  I was invisible and right there, in the middle and to the sidelines. All eyes on you but hiding, that god damn bitch in you that says, oh fuck off. Deal with it. Whine to somebody else. La di da…woe is me, skip di lee doo da.  Hating the idiots that blow smoke up your arse and blame little insignificant problems on everything and anything, getting angry because the smart ass didn’t learn her lines and we spent too much time on so short a scene, wasting a perfectly good Saturday afternoon. One that should have been spent with friends, basking in just being alive, whooping in the audacities collegiate life offers.  Feeling the effects of a whole week of stretching muscles and relieving fat, of Friday’s swing dancing and too many late nights. Nevermind the twenty-six hour sleepless binge being rung up at odd hours this week with three hundred and seventy two million things/pound on my shoulders that I so gallantly, and proudly, carry.  Ice skating tonight, the glorious feel of sliding…as if floating. So swift, so tender, wind through your hair, smooth white ice beneath you and a rink full of friends. Partners with old souls, coffee and forbidden cigarettes with the books and scripts over heart-to-hearts and ex-boyfriends. Mouthing off to the bastard that proudly, bitchily walks over and hurts him. staring him down with the, ‘if I ever as to so far see your damn skanky ugly white thrash face again..’ and ending it off with him knowing that the friend is far better off without him.  And I thought all gay men were nice.  ---  Sober now, sorry cartwheeling ass dragged out. Boogie-ing, door thumping in the lavatory as you patiently wait and grin, your knowing eyes that flashes to the next person behind you in line, there are beds upstairs you know. Keg-stands, ass-kicking and big band swinging, frat boys and sorority girls, induction and seduction. A nice hodge-podge mix really, an interesting combination.  Affected but you lie. Smile through your damn adversary bitch.    I’d really say more things, but taking the whole blog into consideration, the last whole year and a half has been an emotional rollercoaster. A drama-filled whiny decaying insecure daily speech. Woe is me, pity me, I mock them…only to find three fingers pointing back at me.  Becareful of what you preach.  Be wary of what you write.  Go to sleep delusional, the sun has already risen.   urlLink  \"Sometimes I wish I wasn\'t such an introspective navel gazing thinker. \"  '),(22375,664485,'25,January,2003','2003-01-25 00:00:00','   The girl says I intrigue her.  She holds my hands in hers, clasping it tightly knowing that I might jerk away at any second. At any time, my walls may come haltingly up again.  Her eyes stare deep into mine. Her big round American blue eyes, she pierces me with her feelings, her deep hidden thought and emotions. She is a person of action, not words yet everything she has to say is brain candy. Intelligent, brilliant, beautiful.  She is breathtaking...  She holds me close again, and we lie down, covered up, bunched together in the cold and look at the stars. Clear, cold, cold, night….every asteroid is burning so brightly, moonlight encases us into a fairy tale, an enchanted evening.   ---  We have that relationship conversation again. He wants a special place in my life, a piece of my heart, a window into my soul.   I laugh and tell him he’s silly, stupid. I am not a girl he wants to love, I am not a girl that’s capable of loving or being loved, but it doesn’t mean that I don’t enjoy it either.  He’s caring, he makes me dinner and walks with me. Nature looks so much more tantalizing through his eyes, life seems so much more interesting with him around. Larger than life, more vibrant, more colorful, just bigger….better.  His mind attracts me the most, nevernmind that he is quite aesthetically pleasing too. Yummy.  My lips have never touched his though. Never really lead him on, I hope. If anything, just made it clearer and clearer boundaries. Invisible lines; but you can still feel the tension.  Things are out between us, but he slowly chisels off his way into the inside. I push him away ferventently, continuously.  Ask Jared, he’ll tell you why you don’t want to know me. See me. Love me. He’ll tell you why it’s bad.   So we talk, laugh a while. Silently waiting, silently wanting.  Emptiness filling and caving in more, Afraid.  ----  Competitors. Behind, at the back of my mind, I wonder if it’s all just a game. Predator, Hunter, Pawn, Bait.   They come and go in succession. Each needing to be filled, each wanting to be loved. Outgoing, shy, arrogant, giving, caring, loud, flirtatious, quiet, silly, funny, intelligent, talented. Brewing, melting, cascading pot of interests. A plethora of personalities. One to another to the next and a day, some are in classes, some are random, some are friends of a friends of a friends. Some go dancing with me, some play video games, watch movies, work out and some, just dinner and parties.  Sheep. Blind deceived sheep. Oblivious to the being within, they fall into the image. The life. The show she puts on, the layers she brushes on, the glittery wrapping paper and the expensive bow for the Dollar General gift.  ---   The old man. The best friends. The safe people that see her at night, when she is heaving and sobbing, the ones that put their hands on her heart and stop the blood from gushing. The ones that see her when she’s weak and lends her their strength a while. The ones that understand and help her through another night, carries her through another day.   The ones she goes to when she needs to feel special, when she needs to know that at least one person on earth knows who she is within inside. When she can’t pretend anymore, when she is tired of being the happy, laughy, giggly person. When she is breaking.  Hurting.  ---  The boy that still remains so hugely a part of her life. Suppressing thoughts and feelings even if she still cares; life moves on, let go damnit.  Untouchable.   Other side.   Disconnected.  Broken up.  ---  The words that come flowing out of her mouth, the thoughts that leap onto paper and castrate themselves. The touchy-feely emotions she pens out daily in regression, in anger, in depression, in ecstasy.  They think she is manic-depressive. All the extremes, the highs and lows. As if a solitary diagnosis explains everything, why she feels like she’s failing and falling, the eating disorder, the insecurities, the relationships. Growing up too soon in a large, tormenting, cruel world.  The flip side of being sensitive, she is an actress. Vivid, poignant, dramatic. She escalates and drives a character through, she gets deep within and feels it from the core. She fits right into the theater group, all weird, all  different  yet she disguises and weaves her way too into the musicians, and the painters. The cheerleaders, and the nerds, the geeks and the jokers. The bums and the lost. She feels, she is sensitive, she chooses pain over numbness, she chooses creativity over nothingness, she chooses extremes over the in betweens. She chooses to be driven instead of just being content, instead of losing dreams and aspirations.  They say people choose their own poison and that my life is a testament.  Eighteen years of existence, so much and nothing to testify for.   So much left to live for. Dream for.  So much more that I can be and I can try.   So many decisions left to make and paths yet to take.  Will she make it? Will she crumble? Will she smile with the little tiara on her head?  Time, the essence, as always will tell;  I haven’t turned out yet.     urlLink  \"Maybe falling back on memories isn\'t such a bad thing anyway. Maybe clinging onto a past that has long forgotten itself isn\'t as selfish as I had once thought.\"  '),(22376,664485,'24,January,2003','2003-01-24 00:00:00','   I must have at least fifty layers on…nevermind that right as I run the bath I use up  all  a significant amount of our hot water but yeah, eighteen degrees outside? Damn it, if it’s going to be this cold, I should be able to make snow angels! Gah.  :P  on another thread though..   do ut des   naked imperialism says that apparently my torturous limbo constitutes the all important question of theology against theatrics. I mean, fuck my faith, but damn it, how far does your basic necessity to believe the being up there take you? Semitic thrash really, thoughts that invade my pagan mind when running across campus with the crazy, screaming, nut chasing after me, hitting me with his bible full of verses and challenges.  Sure, I humor him. Smile as I hurry in to avoid the cold, yeah, laugh if you may at how many pounds I’ve put on just by the coats but we’ll see who’s sorry ass freezes eh?   But he questions.   “where’s your faith, little child, where’s your faith?”   I don’t know mister.  (half open, half closed]  Sacred versus Vulgar.  A good shit novel of deals with god; promises, agreements, discussions.  I smile again, at your latest entry. Figuring out your codes, encrypted. Running through in my mind and pointing out who Rhett is, or the Strad.  I wonder how you’re doing old man. Are you tired of the crazy, crazy world too? Or did Love find you along the way and introduced you to Hope. Hope is a friend, an aquaintance. She’s nice to me sometimes, she finds me. Or I find her: same difference.  Morning classes are fun, nothing like starting off the day discussing black literature with the Substitute. Have I told you about the Substitute?  The substitute rocks my socks. Rox my sox.   The weekend is here. FINALLY here. You wait and wait and wait, swing dancing, parties, clubbing, friends, aimless wandering and Freedom. On the way to the gym now, going to work out with the ex and his ex and me and ‘friend of a friend of a friend’s competitor. This should be  interesting.   I talked to the Boy last night. The main Ex.   Waffle House sessions won’t even come close to the theories I have about this one, it’s all a conspiracy I swear, bring down the bitch!  I love this sporadic, half not there writing.  No more cancer sticks on me, sob-sob, I was going to tell you about the little girl.  She feels fragile now. There but not there you know? Healing but afraid, happy but impatient, content but searching. All them oxymorons, all them extremes. Oh well.  Barren platonic redundant talk, they find obsecinity in raping my soul. That’s what I told her yesterday, that reading this place was like tapping into the birthplace of my insecurities. They call it honesty, I call it suspended disbelief.   The moments I escape, let go, and run free.  The moments I’m just me, and it’s okay right?  You all think you know me, as if you’ve figured me out with my idiosyncracies and problems and stupidity. You think you know, but you don’t even come close to anything at all, frankly, just because I don’t either. But then again, you know how it’s always clearer from the outside looking in? Maybe.  But she\'s different in Reality isn\'t she? She\'s more secure, more hidden, more Confident.   I\'ll find that bitch and make her tell me her secrets.  Then I\'ll be more secure too, more hidden, more confident.  sounds good.  Now what were you up at  urlLink this  hour, silly?  I\'m flattered.   urlLink  \"She stands still, afraid to move a muscle; afraid the movement will ripple as a pebble thrown into a calm pond, causing the dream to fade.\"  '),(22377,664485,'23,January,2003','2003-01-23 00:00:00','   Still celibate, because I know you thrive and jerk off my (non-existent) sex life, but very content.  That whole need to be romanced thingamagic thing we talked about? Well, tonight we put aside all fears, all wants, all needs, shoved it next to our summer bikinis and roughly discarded open toed slippers at the bottom of the wardrobe. Tonight we were little lovers, giggling, losing ourselves into our own little worlds.  I could pretend he was him and he could pretend I was her. In our own world we held hands, and he hugged me from behind when we went to get soda and fought like little kids laughing all the way and throwing popcorn. Acting like silly retarded juveniles stuck in grown up bodies. We made people jealous, we did just like you explained and made everyone want to slap us hard.  I was happy tonight. I even put on a baby pink shirt on top of the white ensemble to humor you. I felt like a princess tonight, perfect and beautiful. Enchanted as if untouchable, so cried out, so broken torn that all we could do from here was mend. Was fix. Was be happy.  And we even played the dancing game and kicked some good sorry ass.   Two points!!  It was weird driving there though, the broken cd player as if a testament to last week’s events. The empty ackward silence a while, not knowing what to say or where to start. Apologizing profusely on both parts, him because he cornered me into some place I wasn’t ready for, and me because of the way I reacted.  So we hugged and made up. And he asked about Gayle and I asked about Mary. As if this whole seeing a therapist thing was normal, well, that’s California for you. We did highs and lows, and he apologized again for compromising friendship and I laughed and talked about just how damn irresistible I was.   And he promised to make me my Chinese dinner because it’s that animal instinct thing again. The jealousy, the living up to what the other does. The I’ll give you bigger, better, stronger. I laugh and tell him he’s stupid and silly because I sure as hell would not like to be compared to the other girls. Hard to explain uniqueness and difference to people that won’t listen, hard to explain individuality and how one person differs from another.  I can’t make him my Jared, and I sure as hell don’t want him to be my Franklin. I just want to enjoy his company, who he is, his special qualities and bask in that a while. It’s fun. I don’t want to get too close, but I enjoy getting to know someone again.  There’s so much shit from everything else that sometimes I wish I could just shove it someone else’s ass but I’ll deal with it. I don’t have a choice but I’ll be right.  It’s the coldest it’s been in quite a while, everything freezes over and pauses a while, as if caught in a trance. Everything is put on pause or moving so extremely slow it is dictated. I may be hurting, and I may still feel the need to cry sometimes, but every day is another day of getting better, of feeling better. With spring comes dying and growing again remember? Hope.  I’m okay, don’t worry silly-billy.    urlLink  \"kiss yourself if you\'re in a warm place.\"  '),(22378,664485,'23,January,2003','2003-01-23 00:00:00','   I do this once every few months…clean up my inbox. And for the most part, it is fun, fun to see how I’ve changed and how my thoughts have differed from how they are now. Sure, there’s always going to be pain but gotta take out your trash you know?  It must be the coldest day on earth today, well at least in Florida anyway. I hate the cold weather, I hate how the chill gets into your bones and fucks you up that way. Or how you shiver, shiver, shiver.   It would be nice to have someone to share hot chocolate and marshmallows with, someone just to curl up and read by the fire. Fall asleep to. Him.  I miss abominable snowman days, when his great big coat will suffocate my little frame and I’d get lost somewhere in between. Or how we’d laugh and play in the cold, making out, creating music, exchanging ideas and thoughts.   You know what? Maybe, just like how you were saying last night, I am something. The whole intimidation thing? Well, I don’t know how baseless your theory is, or even if there’s any spine of truth to it…but what I do know is that it is nice to be the other person. It’s fun to come into school and have a selection to pick from, guy a for Monday, guy c for Tuesday and girl b for Wednesday.   All harmless flirtation really, everyone’s just looking for their soul mate. Company. Friendship. Everyone just looking to be understood and loved, looking to love.  And I do understand that, which is perhaps for the most part, why I play along with it. It’s a nice feeling to be desired, I shy away when you say such nice things but everytime you do, you know it puts a smile on my face too.  Thank you I guess.  You know when I’ve been in to see Gayle. I am so much calmer from this morning, less edgy, less manic, just sorta almost peaceful. I don’t sleep much I think because my demons tear me up, they tell me things like I need him, and sometimes I end up believing it. Bad. Very bad.   Because I don’t right?  Because I am enough on my own and what was it about me having to find someone who could love as much and appreciate me? Someone who can understand the cycles and the chemical imbalance, saratonin bullshit was it?  You think he/she exists? You think I’ll find that one mad crazy person that’s going to love the other mad crazy person and we won’t ever tear each other apart?  Well, I guess I’ll keep my fingers crossed then.  Want to know what the crazy thing is? It’s that I’m sitting here, completely dressed in white with my hair tied up, my nails done and just enjoying this moment.  I’ve had a nap since the twenty six hour binge and sleep does wonders don’t you think?  I am intelligent, and I am attractive. I have dignity, and I have worth. I may not always be right, but I am not always wrong either, I am allowed to have opinions and thoughts. They shape me to be who I am. I am emotionally highly strung out sometimes, but that just adds to loving me as a person right? I cannot apologize for being who I am, for my eating disorders, my chemical imbalance and being sick, but I will apologize for being the not so nice person I become when I’m like that. I may not always like who I am and I may always push me further and further to achieve more and unrealistic greater things, but that differentiates me from a spineless drop out with no dreams that I fear myself becoming.   I laugh, because no one has the answers. We can search, and cry, and scream about it, but when all is done, still have to find it. Live through it. Deal with it.  I refuse to sit here tonight and write another sad entry, so instead I will go out and have fun. Superficial or not, I like my laugh. And I like myself a lot when I’m laughing, even more so when I can laugh at myself.  I am not perfect. I just forget sometimes.  That little girl that searches in the mirror? She’s still there, she still exists. She just gets tired sometimes, that’s all. But she’s fighting back, and being the smart ass she is, you know she’ll get what she wants.   I am secure enough to know I’ll make it, but still shaken enough to know I need to press on and stay strong.  I love you back, thank you for the entries.   urlLink  \"We can laugh as it dances in the breeze. Hold my hand, little girl. Let\'s walk.\"  '),(22379,664485,'22,January,2003','2003-01-22 00:00:00','   I’m not going.  Another reason, another dagger, another good cry.  Sometimes I wonder if anybody even understands me at all? So many times I find myself here again, searching…alone. Trying to fend off these beasts that violently torment me, these vulcans that devour my living flesh, picking at me till I am empty and torn, picking at me that I have no place to run, no place to hide, no solitude to seek..  I started crying twelve hours ago. I let the tears touch my pillow and soaked it, I let my heart cry out and soothed it. Wanting to cry it all over and still wanting to be strong at the same time.   I don’t understand why.   I try to look at myself in the mirror this morning and remember what bryan told me a long time ago. Tell yourself you’re beautiful each morning sweetheart; I see tired worn out eyes, I see a soul that’s hurting, fighting against time. Against herself, against the rest of the world. I see a little girl that’s afraid to be seen, that really wants to please, that’s so scared of trying. I see someone who’s trying to make it and  pretends so beautifully that she does, I see someone that used to believe in herself and had hope. I see all these things, but I can’t find Beauty. Would you please tell her I’m here?  Would you please tell her to find me?  This morning she rings, just like every other morning. Begins with wanting to know if I’ve had my pills, if I’ve had anything solid in the last few hours. Same old, same old. Asks me about my day…I lie, pretend as usual. It all went good, Mom, it all went good. Don’t worry, Mom, I’m okay, don’t worry, Mom, I’m doing fine. I pretend to be irked that she’s rung me so early, and she explains again, like every other morning, that it’s midnight their time and she likes talking to her babies before she goes to bed.  Then she tells me they don’t think it’s a good idea at all.   I am lost, as if confused. I thought maybe someone would be proud of me, I thought maybe someone would say, you did good, Racheal, you did good. If I didn’t have his congratulatory hug, I thought maybe my parents would give it to me. But they didn’t, I guess it reaches a point where nothing shocks them anymore huh? No big deal, what’s the Honor’s program and being selected to go got to do with it?  All this bullshit about getting well and being alone in the world.  I  am  alone in the world. I’ve been alone for so long, so hard now.   I don’t tell her, but I want to come home. I want to forget everything and give up. I want to run away and hide from the rest of the world a while…I want the nightmares to be over and I just want to be okay. I want to shield myself from me, and draw strength from someone else a while.  I want to really, honestly, just be okay.   ”Smile for me little princess, sing for me like a child. “   I am feeling icky, icky, bad today. Not strong, not happy, not even anything enough to pretend. I am hurting, and I can’t even remember all the reasons why, except like the crying, I’m afraid that if I start I cannot stop. And I haven’t.  Drug me.  Saratonin levels bad? Racheal bad? Racheal bad boo boo mistake not worth it?  I don’t know. I don’t care.  It hurts.  I’m hurting.  I said it now, I’m honest; why won’t it go away?  Why am I not worth pride, and love, and understanding?  Why am I, once you reach my very core, not worth anything?  I try again, but this time I am only putting a bigger band aid, on a bigger wound.    urlLink  \"Please use \"I TALK COCK IN THIS SITE\" instead. \"  '),(22380,664485,'22,January,2003','2003-01-22 00:00:00','   I have given you everything I am, taken everything you are.   And the questions? The questions will always confuse and the answers always linger…but sooner or later, it will fade. Sooner or later, when bare, naked necessity dictates it, it’ll all come around.   Gullible imp, if anything I am. Deceiving witch, if anything I could be.   We knock and scrape at heaven’s door, begging the gods to release us of this curse. This curse of uncertainty and disarray. Make me a mark of misogyny you say, make me hate thine passions thirst with a wrath so great it dispels need. Or trust. Or love.  Remember this when I am weak and tired again, remember this when I fall and break, scratched up and bleeding.   I am alone.   I\'ll let you break your promises, strip me of my pride and make sense of us by tearing me apart. I\'ll let you mock me, and take cheap shots at who I am as a person, and who I\'ll be in coming days. I\'ll sit here and stare, and when my putrid salty tears touch the ground, I\'ll let it rot the soil, and kill the earth; all the while masking this face of strength and dignity, all the while masking this emotionless face of disregard.   You come back, as they always say, and beg me for forgiveness. You\'ve taken time to finally understand the reasons why I pushed you away; afraid of love, insecurity, disorder. I want not to blame it on my disease, to blame it on the state of my being, but what other explanation is there? I remember the fucked up child business, I think it hurt most because I actually think you were right.   It takes a while for me to trust you again. Someone else is in my life, putting on more layers, like a band-aid over a broken wound...only temporary. Takes me a while to realize too.  But you leave again. You\'ve pushed me to unleash these feelings and emotions that have taken months and months to bury. Phone calls that drive me to explain everything, questions that question and question and question. You want to know why I lied, you want to know why I fucked up, you want to know if I still have a soul, if I am still human.  You break this sheen of hardness that protects me and I sob. I break down and you tell me that for once in your life, you finally understand why I couldn\'t do it. Why it hurt so bad. Why I couldn\'t go through us again, the loneliness, the emptiness, the stabbing. I didn\'t want to lose you again, I didn\'t want to have to go through losing you again.   Then for a while we are good. Or I am good rather. I\'ve come back renewed, the wounds healing and something in me has changed. Something in all of us has changed. I am good for a while, but not long after that, the nights begin again...only that this time, these are your nights. I cannot complain, I cannot whine, I cannot fight back. I have no voice, I try patience but you\'ve always known the right words the pierce through, I understand where you\'re coming from but comprehension doesn\'t disregard the right of emotions.  I feel.   I\'m human too you know? I forget sometimes too.  One night, as if by epiphany, I realize I’m still in love with you. And I still love you. I realize when he tells me to move in with him, and I know I can’t. I can’t move off some place else, as much as I’d like too…I’ve tried that before, that’s why I’m here remember? Because I was trying to run, escape. So I packed my bags, and moved continents, I thought maybe, perhaps, I don’t know. being in the same place didn’t help either, I’d severed us long before, too much damage done I suppose.  I am the idiot now. I’ve always been haven’t I? When I had it all, I fucked it up. I changed you, you say. I made you into the cold, hard, cynical beast that you claim you are, that you want to be. I still see someone I used to know sometimes though, I still see him sometimes…when he lets me see him I suppose. I remember all the nights you asked me why you weren’t the one anymore. And I remember all the nights you asked if I still loved you and if I would ever love you again and I told you, no. I told you I couldn’t love you anymore and that I’d fucked up forever. That I couldn’t handle just being enough and just being enough didn’t cut it. I wanted to be special in your life, I wanted to be everything I was before I destroyed it.   I lied. You’re still that guy. I realize now, but then I know being the hard headed ass I am, I’ll reach that stage some day and let you go. I don’t want to hurt you anymore, frustrate you, piss the living bejesus out of you.   I hide, go away and take you off my lists. I attempt to cocoon myself into security, see no evil, hear no evil right? I miss you though, I hate it that I do. I hate it that I have to go through this bullshit of feeling…wouldn’t the world be just nice if we didn’t have to feel?  I still read you from time to time. Sometimes, it feels as if I’m violating your privacy, but then again I cannot help but selfishly wonder, are certain things directed at me? Well, they work pretty well then, because when you put on that cloak and throw sarcasm at me, it still hurts. It’ll always hurt, some part of me will always want approval from you, will always want you to be proud of me. Some part of me is always going to want, at the end of the day, to tell you about it.   The stupid part of me still cares. When you talk about doing stupid things to yourself, I worry. And when you talk about hurting and being alone, I wish we could forget our stupid fights and disagreements and I could put my arms around you and give you love. Or security. Or whatever you needed, filled that void for you, mended that crap for you. Lay it on me. Lay it on me.   Everybody wants me to start dating again. Some days, even I do. Sure, it’s always nice to have company you know? everybody says move on, get off my sorry butt and have fun. I’m young and there are plenty fishes in the sea right?  Maybe so. But I want this fish, this fish I killed. This fish I mutilated.   I fix me and it’s as if according to Murphy’s Law and I get sick again. I don’t even care anymore this time, sure, I’d like to be all healthy and shit but to go through all that again? I don’t know, maybe not huh? But I’m seeing someone and getting help, and the whole damn world and their mothers say that they are proud of me. My real world finally knows I have a disease now, it’s funny how supportive they are.   Damn mind games in my head that still play. It’s still hard for me, I still want someone to hold my hand and on nights which I’m not feeling too good about myself, say it’s going to be right. But I don’t want to worry anyone, so sometimes I just cry it off and take your remedy of sleeping and sometimes, I just give in.   You asked me the other day about my body. I don’t think I’ll ever be enough for you. I used to be shy around you, now I’m just ashamed.   I’m doing really good in school. Sometimes, I remember about what we used to talk about. I booked my return ticket five months ago, I thought maybe I’ll fly in through LA. I have to cancel it tomorrow…I get to go to Italy, and France, and Switzerland. Get to see all the things I’ve learned and put them into existence.  I wish things were like the way they were before, and I’d write you, hell, even call you to tell you about it. History Honors program. Pretty damn ironic huh?  I guess in some ways everything still happens for a reason. Fate decides what’s not meant to be I suppose. Lately I’ve questioned my faith, I used to be strong and I hate to say this but gullible. I liked that crappy innocence then, when nothing used to worry me and I’d climb mountains without worrying about falling. It’s sad I guess, but it’s okay. I’ll survive…you know I always do, and I know you will too.  I see you take more steps away with each passing day. Some part of me wants to cling on so bad, wants to just curl in your arms and feel you but then I realized that sometimes, loving means letting go. I’ve kept you for so long now, kept you in my craziness, my bitchiness, my insecurities and you don’t need that. You are right, you do deserve better. You do deserve someone who’s going to make you feel good and laugh all the time. I guess it hurts to say goodbye, and we’ve done that quite a few times.  I don’t know why I wrote this tonight. Maybe just to know that I’d be okay and maybe just to understand why I had to let you go again.  It\'s okay, you\'ve left anyway. I understand, I\'m not angry, not feeling, just sorry.  sorry.  '),(22381,664485,'22,January,2003','2003-01-22 00:00:00','   you\'d almost believe that i was happy wouldn\'t you?  '),(22382,664485,'22,January,2003','2003-01-22 00:00:00','   exactly.  '),(22383,664485,'22,January,2003','2003-01-22 00:00:00','   We’ve been playing phone tag for as long as I can remember…and so she rings tonight, I am busy, away. Soaking away in bubble glory trying to maybe drown out all my troubles. I don’t hear the phone ringing, forget to bother to check. It’s been hard to explain my distance these past few days, I don’t feel like talking.  I ring in and run through my messages, hastily deciding that I prefer my hermit life better. Haven’t had this in a while, haven’t had alone time…but do I really like it? I don’t know, that remains questionable.  I can’t really make out what she’s saying, just sobs and wretched empty cries. I ring back and dial again and again, convinced that even her parrot might pick up next.  I wonder if she’s doing fine. Of course not silly, she needs you.   I care, and I want to help. I want to listen and fix things for her…but I can’t. I don’t know how to, I am not strong enough to love her. Not weak enough to say no.   It is…complicating to say the least. You feel that two hour cross county drive coming soon? Well honey, let me tell you something, every night this week and last and the week before would have been perfect candidates.   I came home last night, and I don’t know if it was just hormones acting up or the actual realization that I was so alone. No one’s arms around my waist to hold me tight, no lips that would sneak kisses, no security…absolutely nothing; because I was afraid? I had it all and fucked it up. He leaned in to kiss my cheek and I brushed him away, then turned around and locked myself back into my four walls again.  How do I fill this emotional void and not get hurt again? How do I not let myself hurt me again? I just want it all to be over with, I just want it to be done.    But would that justify me letting someone into my life now? Hell no. I always tell you men are so evil and treacherous and awful but they are all the same. Men, women, all bad. Maybe I have commitment phobia, or maybe God just really messed up when he made me, couldn’t tell you.  Make-up works wonders you know? Make-up and a cold silver spoon on swollen, puffy eyes. Crying silently, softly to sleep again. She breaks, her bouts of depression pound in stronger and stronger; it seems as if the more structured and perfect her life gets outside, the more torn down she is inside. Maybe I can’t handle it, maybe I won’t accept that I can really be someone because that only brings in expectations and dreams, or maybe I’m just hormonal.  I see Gayle tomorrow, but I want nothing to explain this. I want to keep mum and cross my legs and tell her about how ‘good’ I’ve been this week. Tell her I haven’t looped, that I’ve made the conscious effort to not practice regurgitating this week. That for the last seven days, I have digested.  And I won’t tell if you won’t tell.  I’ve worked out every morning, every afternoon and every night this week. I’ve ran, danced, lifted, bladed, cycled, kick-boxed and walked off all my very sore muscles. I’ve slipped in the diet pills as well as lived on my one meal a day plan this week. Dinner, because dinner is with the family and I can avoid breakfast, and avoid lunch but I can’t avoid dinner. I made a promise to not throw up so it’s the masochist in me that finds the need to do this. That has the perverse pleasure/pressure in getting down to double digits, just seven pounds more to go. i want that whatever it\'s called bone to show, i like that. not the ribs, got that down two years ago, not the hips, got that down a year ago, not the collar bones...but the in between your breast bone and that. whatever it\'s called anyway.  Want to know the worst part of it?  Is that I feel guilty for it. Is that I have my two masked face and my two sided heart. That my brain is able to decipher that this is wrong and I need to fix it. That I am losing it and abusing the shit out of my body. Taking all my shitty selfish problems and making it all exterior. Things I can work on. The part that pisses me off?  That there isn’t just a book I can study off, or a script I can memorize. There is nothing concrete but just all about emotions and feelings and all that bullshit. All about what my self-worth is to me at that point, or how secure I feel about my own being. It doesn\'t help when you\'re in this industry, it doesn\'t help when you are attacked.  I lie when I say I’m stronger now. Because I think that this strength, or at least this battle of staying on top, has made me weaker.  Do I really want to do this charade? of dating and getting better and being happy?  No. I don’t know.  I only do it so no one bugs me about it.  I only do it so everyone will shut up and leave me alone and quit trying to fix me.   I am so different out of this html work. I am so different out of the Internet.  In ‘reality’, I am confident and happy. In reality, I am giggly and laughy, independent and secure. I am so many things that people want me to be, expect me to be, play me up to be.  Will you do an interview with the newspaper? Do you mind if we do a story on you? Would you consider joining our program? Are you interested in saving the world?  Sure. If it’ll make you happy, okay, why not?  Ever get the feeling where your whole heart feels so numb and painful at the same time? I think it feels a little like this right now. Crying but not really crying, crying but there’s no more tears left kinda thing. I understand, I get it, but things are always easier in theory you know?  Maybe I’m crying for punishment, maybe I’m crying for sickness, for innocence. Maybe I’m just crying out because I can, because it’s not fair. Because I’m at the prime of my life and I damn bloody well should be enjoying this. Because I damn bloody well really want to enjoy this.  I don’t even know what my point was in the first place. Except I keep on going on all these things and pretend as if I’m really interested, and maybe I am. Maybe I find them intelligent and charming and amazing. Maybe I do want to get to know them but do they really want to get to know me?  I don’t know. I don’t know the answers…which started the disagreement the other day. If I never let anyone in then how will I find someone to fill up this void? How will I let love in if I continuously push it out? If I continuously make up reasons not to be loved. If I continuously make myself unloved.  Because I don’t want to hurt someone again. Because I don’t want to reach out into thin air and grasp onto nothing. Because as hard as I try to forget and as hard as I do, I’ll still remember that steely click. I’ll still remember the loneliness. I’ll still remember losing someone, and having someone not there anymore. Not exist.  Why am I hurting tonight? couldn’t tell you really.  It’s been so long since I’ve been happy. Really truly completely happy. I think I just lie to myself some times…when I have streaks. Because really, it all comes crashing down anyway.  I don’t know why I’m here again.  I hate not knowing where I’m going. I hate not knowing how this works. I hate feeling so damn sorry for myself.  I hate being able to fix every damn thing else but myself.  ...but the good part is I\'ll probably be okay tomorrow. I\'ll be fine and bouncy again, just don\'t catch me at night though.  Not when I\'m alone.   urlLink  \"this is the time in my life where everything is falling apart, and at the same time it\'s all coming together.\"  '),(22384,664485,'22,January,2003','2003-01-22 00:00:00','   “Forgiveness is a gift you give yourself,”  …but what if it is the unforgiven. The sinners, the dirty, tainted shit. What now?   Ku menanti, seakan tangisan bidadariku menguasai langit terserlah yang menjadi bumiku, Ku berbisik, bertanya…adakah harapan di dalam dindingan ini? Adakah jawapan di semananjung gerakan ini?  Menanti…seakan mencari detik wajah yang melinangi air mukamu Menlintasi batasan yang terbahagi-bahagi di antara maruah and keperluan Diketahui, tidak dinyatakan, tidak disoali Ku menhitung, mengira Melapiskan wajah yang tidak tentu, tidak benar.  Anda terlepas  Menyoal..  Tidak tahu.  …aku kian faham  Anda bukan gerisan yang kucari lagi  Anda dipertunjukan.  Pertukaran.  Aku Tentu.    Oligopoly exists. A synergy that moves, conforms by the superior hands of media moguls. The vanishing powers of an individual silenced. Protests, but hypermass = dumber² . A mirage of sheep following, being fooled into a scary mass of marketing strategies and promotional gimmicks. The pop culture: don’t take away our MTV, our Happy Meals, and our mainstream bullshit and hand us independence. Don’t silent us with banning, and censorship, and communism. Thrive in this conforming, cookie cut-out Britney Spears / Hugh Heffner wannabes. So says the different that ends up being the punk that falls into a category anyway. A stereotype.  Ahh..see this is what happens when you stick racheal into a mass-media class that debates four thousand and six conspiracies where every damn secret government agent is against us and everything we are being fed is programmed. Encrypted into a code somewhere, dictated by an unknown source. Watching the Beautiful Mind and finding more ways on how the KGB and the CIA is against me, more chips, more androids, more sheep.  Damn sheep.  Dr Schuler offered Italy/Switzerland to me today. History Honors program, ladies and gentlemen, that would be your biggest farce ever. The coolest stunt I would have ever pulled. English Honors, I understand. Math Honors, I still understand. Thespian Academy, doh. But history? Holy shit fuck I am whinging it!   So this summer I get to go to Italy. The selected few, it’s the Cambridge program or the Italy shit. Ahh, what sweet nectar I am sucking from. Parading in.   This is huge and big and incredible at the same time. My passport into freedom and excellence and learning. Teaching, growing, developing. What my heart sets out to do, my mind will get. Every other aspect is perfect, everything on the surface is polished and gleaming.    All that glitters is not gold    I feel stifled and oppressed. Pushing my limits, and striving to be everything I am not.   I don’t know how to justify this, or that. I speak with my head up my ass, I am amazed at this opportunity. Stunned even. But it is expected, make the cut, get the grade, live the life.  What if I’m not enough?  What if I’m not perfect, and not smart enough and I can’t explain Plato’s symposium and to the damn fuck don’t know much about Medea. What if who I am that you think I am is not at all?  What if I just question too much?  What if I should just go home and celebrate my  sorry  fat ass and continue memorizing my two bit monologues for my two bit auditions that I two bit avoid anyway? My silly bimbo mentality. What if some night when I’m asleep some fairy will fly it’s wings over me and wave it’s magic wand of confidence and security?  The dreaded what ifs again.  I am who I am. Live with it right?  Be proud of who you are, what you have achieved, who you can be.  I made the cut, Mom. I’m on the honor roll, I have a crazy crowd of friends, and I am almost but undeniably satisfied.  But I want more. So I push further and further, and test my limits, and set new goals.  So I try again and if I fall, damn bloody work harder at it.  Then why, I ask, am I not enough?  My poison, he says. It’s all your own doing. What makes you breaks you.  Understandable.    urlLink  \"The Muse \"needs her excitement, and drama, and craziness.\" - Yeah, right. All you need is a couple of \'C\' sized batteries for your vibrator. (Don\'t be suspicious - it was a guess.)\"  '),(22385,664485,'21,January,2003','2003-01-21 00:00:00','   I sent a part of my soul to the Technology Doctor today. I feel as if a part of me is missing, as if I have given away the intricate core of my being that pieces my life together. I figured, what’s a few days without writing, blogging, photographing, editing, imaging, emailing, surfing, and talking. Ah, you see, it is a huge part of me, it’s sad, I do admit, I’m a huge nerd.   So you see, I am curled up on the sofa with a borrowed laptop, rummaging through volumes and volumes of some sort of off-beat trance type British bullshit sorry excuse for music which in my humble opinion, is an unfair judgement of the type of sounds that come from the other part of the continent but what can I say, I have weird friends.  I am sitting here, sharing a cigarette with legs propped up with the intoxicatingly orgasmic scent of coffee parading around the room, seeping into his futon, getting underneath the covers with us. Getting my foot massaged and dinner prepared…thinking hey, maybe being a woman once a month isn’t too bad after all.  Nevermind the cramps, and the moodiness. Nevermind the strong desire to punch someone in the face and the sudden urge to wail my sorry ass off. I smell dinner, and damn it is good. But then again, that strategy of starving so anything tastes good always works.  I laugh because it feels like we are an old married couple though really, nothing about us is conventional. Perhaps lets start with the beginning, how the hell I managed to come home with a guy instead of a girl from the gay bar, don’t ask me.   I remember a part of this a long time ago, okay, maybe just two or three months. When He was still around, that comfortable “Honey, I’m at Wal-mart, do you need milk?”, but then as always, comfortable is never good enough for me. Racheal needs her excitement, and drama, and craziness. Racheal needs her volatile screaming and bitching and yelling. This is the part I see my mother shaking her head, whispering under her breath, racheal, racheal, racheal. All together now…  Siiiigggggghhhhhhh.  Oh wait, I forgot my point.  Er..yeah, Him was it? Well, really, there’s nothing much left to analyze there. There’s the ex, there’s pushing him back to his ex, and there’s excuses. The master of conspiracies and planning is at her best again, I still get my daily calls, my emails and updates, and I am happy for him. His life is balanced, and perhaps that’s what I need too….balance.  The friend of a friend of a friend’s place is typical. Clean even. Lots of black, white and silver here, abstract art there, Lucy Liu (yes, I’d fuck her too) poster tacked up  and the modern versus eastern vibe going around. Yes racheal, smart move, guy number seven with asian fetish. Hmph.   So maybe we’ll have sex tonight. Shit, I thought my gay-dar was pretty accurate but apparently, I am wrong. I take that back though…absolutely not, stupid thought dummy, I am staying celibate remember?  Stop laughing Marc.  I am attracted to a nice mix of humans right now. Just the right breed too, Men. * Gasp *   note sarcasm.   And I have absolutely forgotten my point of this entry, except maybe to say, I wimped out on Gayle today and rescheduled for Thursday again. Somehow, all the talking and explaining and feeling guilty didn’t sound like such a great idea today. Besides, how else was I going to knock this one down and fuck the human race again?  Oh well, there’s always tomorrow. New day, new idiot to diss, new morning to bitch.  Oh Wee.     urlLink  \"I forget where I\'m going with this.\"  '),(22386,664485,'20,January,2003','2003-01-20 00:00:00','   I used to be a writer…an artist. I used to sit and type for hours and hours, reliving tales and baring my soul. I was less cynical then, more opinionated then worried about being exposed. Art was my means of expression and so daily, I wrote in thoughts and opinions, daily, my feelings were transcribed into words, my moods captured in slides and the very  core of my being dictated my strides. My strokes. The beauty of heavy, wet paint brushed onto a white, clean canvas.   Free to be explored and altered, free to be manipulated and created. Free to do anything for everything was in my hands.   I was nimble and open, but years passed by and made me older. More cynical. Less forgiving. The shitless porcelain vase that stands broken thinly held together by only a film of adhesive...I am that vase that borders on insecurity. Will I fall and break and like Humpty-Dumpty not be able to put the pieces back together? Or will I waver between both sides and always be saved right before? Right before I crash and burn and fall like China cups on cold, wet, tiled floor…guaranteed them smithereens.  Stroke me.   Very gently, rub my back. Touch me. Get inside of me.  Make sweet, hard, soft, passionate love to me.   Don’t touch me.     urlLink  \"There i sat, bleeding thumb held back away from my hand in an attempt to forget that it was conveniently opposable, pressing together the clay i had just so painstakingly cut apart.\"  '),(22387,664485,'20,January,2003','2003-01-20 00:00:00','   I feel like the dirty slut, with her long overdrawn secret rendezvous, down and tired with  everything. Noticing all the loopholes, changes, insecurity, seeing with that damn naked eye again.  Some part knows I have unleashed a whole new world to you…one that will undoubtedly bring you much pleasure and pain. This secret hideaway from the rest of the world that feels so good and so bad at the same time…you’ll understand a little bit more now; this perverse pleasure of hiding, of having the world in your hands, of that control, that desire, that ability to say anything you want and whatever you please and not give a rats ass about it.   How people can judge you here but really, do you care?  Some day I’ll grow old, tired, and weary. Some day my skin will wrinkle, and my eyes will lose it’s sparkle…that some day looms near. That some day is here and so far away…I have my days. All the wonders of being a woman they say, all the wonders of pms.  I play with that fucking heater control knob in my little room with nothing of my own and everything of my own at the same time. Always about the two tones, the two faces, the double lives.  I walk around naked, in my own skin around this room, I am secure. Cold, but secure. Avoid the mirrors I tell my goddamn self, don’t abuse yourself tonight, go out have yourself a shitless amazing crazy night.   I put on my little black dress and slowly, apply my layers. Concealer, foundation, powder, shimmer, blusher, eyeshadow, mascara, bronzer, lipstick. Holy shit it takes a lot to make me presentable.  I want to stun tonight. I want to step out and have him take a step back and say wow. I want him to think, holy damn shit hell, what did I do to deserve this? I want him to spend the whole night looking at me, wanting me, captivating me. And I know he will.  It is that power thing again. That control I crave. And I can act coy, and silly, and laugh. And I can play along into his fantasy of everything I am not, as if this outside racheal that people so damn like is truly so damn amazing.   Hah.  Oh well, and so I’ll have myself another night of pure lust and seduction. Of getting to know someone then come home and analyze all the reasons why this wouldn’t work. I don’t eat Starbursts anymore, I don’t take wine with my sushi, or listen to the ba-da-boop beats of my heart. They lie, I’ve learnt. Don’t fall this time, don’t fall again.  So then I live in my little world of the in between. Of craving for love but not wanting it at the same time. Of feeling the empty loneliness that comes with just choosing your own poison, of choosing this fate and running to a self-proclaimed destiny. I have fun when I need it but hard-earned lessons have taught me, don’t put out, don’t give out, don’t let go.  I am my own being. I am not enough on my own but I am all I have.    urlLink \"Well, shit. These kids think I\'m some kind of polymath. Like I know it all. Look - I\'M A MAN - we don\'t know from crap, but we just talk like we do.\"  '),(22388,664485,'19,January,2003','2003-01-19 00:00:00','   Paper entries penned in daily- anger driven thoughts layered with putrid heartless emotions, tears that smudge the black ink and causes streaks to run along the smooth, crisp, surface of an intricately bound sixty-two dollars Hobby Lobby/Barnes & Noble put together journal.  Coffee sits on the old oak table, getting colder by the minute…second. Getting stale, obsessive nights of Starbucks and Salem lights, heavy crystal ashtray emptied daily. Daily, the trash is taken out.   Scattered albums of childhood pictures lay around the floor, picture cds of ranging emotions that lie deep within her, mindframes of different occasions captured on camera. She thinks she’s almost invisible, her the subject though escape the reality.  Overexposed shots in a row. The naked eye does not see all that lies within it’s context. The series of bright, stark, harsh light washing her out remains to be her favorite. Transgressed, perhaps that is how she feels.   Soul, torn up and bleeding. She is breathing again, beating…but nightly, she draws back to nestle into this cocoon of security she used to know, how do you break the cycle? How do you stop all these emotions and thoughts?  Her brand new angel wraps herself around her broken wings, begs her to wave her magic wand and make it all disappear. Her hundred dollar an hour therapist is where she goes to search for more answers, ask more questions.  But nightly, she still gets sick. Nightly, when the world stops watching, she cries.   It hurts. It hurts to be so wanted  but feel so empty. It hurts to lock yourself away into a tiny little cell in your mind, it just gets so damn cold. So damn alone.  Yes, she does want to feel love, wants to love, wants to need. Yes, the opportunities are presented, are argued. But she doesn’t understand why, but she is afraid how…afraid of her own being, her own doings. Afraid of the person she is, afraid of the person that begs her back in the mirror.  She ruined herself way before he ruined her and all these ruining has left her with scars. Scars that control her actions and dictate her feelings.  There are too many extremes in living her life, apparently, her indecision, her fears, her sensitivity is what makes her her. Makes her desirable, makes her beautiful, makes her loved.  She doesn’t understand. Not one solitary bit at all.   Why does letting go mean coming back?    urlLink  \"i want to cut my heart open to see what\'s inside. who am i, or what am i? is she me or am i she? i don\'t know, i don\'t know.\"  '),(22389,664485,'19,January,2003','2003-01-19 00:00:00','   I know the pope/bishop/the wrong reverend and the all important big guy.  I just thought you\'d like to know.  :)  '),(22390,664485,'17,January,2003','2003-01-17 00:00:00','   I am in one of those moods where I am so talked out and tired, I almost have no more bitches and frustrations in me to rant about. Notice the all important Almost.  The trips, to say the very least, was quite eventful. I really should probably just stop at that but seeing as I feel a strong desire to press on with this need to analyze things, I will talk about it.  So I figured out a lot of things, hell, you think that with the last whole week my damn mind would have at least been in overdrive already.   But Nooo, coming home I decide that on a four hour drive with a dog, sex talk and no music because I brilliantly destroyed the CD player on the way there, we would talk about feelings, and emotions, and bullshit about relationships. Now what started that spiel, I don’t know completely.  I am swerving in between lanes, attempting to get used to the car and having long overdrawn protests on why I stand for the things I do or at the very least, attempting to reclaim what little dignity I have left when it dawns on me to just god damn shut up and keep my opinions to myself.  It’s like one of those things where people talk and talk and talk all the while; judge and judge and judge.   To a certain point, it almost reaches unbearable.  Never, ever, have I been so glad to cross the Pensacola Bay. Or that retarded street along Pine Forest that says N.A.S. oh, god forbid do I ever curse that sign again.  But anyway, it is that tension so thick it is silent that permeates in the air, all laughter from the night before is now destroyed with the difference of opinions and judgements. I hate that I am affected by it, I hate that the whole damn car is suddenly full of drama and tension it is like a bad run of The Bold & Beautiful with fake smiles and pats to go all around.  I feel as if any minute now, some cheesy theme song is going to burst out. And it does, in different forms, just make it through the next twenty minutes and you’ll be right.  I ring and say, I’ll be home soon, take me for a drive. Take me away. Calm me down.   I finally reach and it’s that damn dam again and all the pent up, bottled shit just spurts out forcefully. It is painful and mortifying all at once, it is embarrassing.  We drive for a while, and I need someone to say the things I know I need to hear tonight. they are said, and two hours of driving around one side of the town to the other later, I am sane again.  We reach and go in for dinner. Diner smoke-free and harsh bright lights, fading into a corner…I eat. I am smiling again…laughing. Comfortable.   At peace.  I might think too much, so sue me.  '),(22391,664485,'17,January,2003','2003-01-17 00:00:00','   Cubicles intrigue me.  Think about it, you know, how psychology is all about analyzing the human mind and how people project certain things…or rather how people take your basic functions and put them into context about who you are as a person?  Like which row you choose, the left aisle? The fourth computer in the middle? Do you like your chair higher, lower….how everything is nothing as it seems.  How, apparently, if you cross your legs a certain way it decides who you like better? Acting is so interesting, how you study your characters and decide how it’s suppose to live or function or walk. Talk.   I don’t know, it suddenly just came to me. I’m easily amused…so go out, pick your seats, play up your idiosyncracies and somehow, human evolution spines through and we all end up having the same thoughts. Doing the same things.   Damn sheep.  So what separates you from one being to the other? Are they your values? Your opinions? The way you view the world or the god you believe in?  I honestly to the damn fuck don’t know.  I also curse too much.  But yeah, just food for thought…or in this case, thought for the next four hours.  Baby’s going to Tallahassee!  : )   urlLink  \" i felt kinda bad, but jesus, what a view it is\"  '),(22392,664485,'16,January,2003','2003-01-16 00:00:00','   “We will not talk about anything that will make you feel uncomfortable,”   but then I ask her, taunt her to hurt me…ask me…push me into a corner and weep. I know what I do is not right, I know my decisions at times are bad and really terrible. I want that feeling, that muck with being who I am to be realized. I want to go in and feel guilty, I want to do whatever it takes….whatever the god damn hell it takes to be sane.  I talk and it feels like this big great dam in me has been unleashed, all the pressure of trying to stay on top…of having to be on top. I talk and talk and talk, I talk till my voice is hoarse and throat is dry, cry out. Then I turn around and realize I haven’t done that at all, I’ve hidden yet again, I’ve continued this act of being strong and have ironically, come out strong.  Stupid, lying bitch.  Psychologist. There’s such an eerie ring to that, the social stigma; mental health benefits.   “I am seeing someone,”  Instantly, you assume it is of a romantic nature, free sex for all and lots of romance. After all, that’s what I do best innit? Play the games, work the circuit, get my entertainment. I make it seem as if we are all pawns in a match, as if it is just a challenge. Hook, line, and sinker.  But I don’t mean that. I mean seeing someone to get help. Get better. Work on things and get through stuff. Icky, icky stuff.   I hate this seeking for help part. I am uncomfortable with admitting it, I am a damn conformist. I too believe that psychologists are for really really bad people, crazy people. People that kill their parents and drown their kids…I’m not one of them am I?  But then again. I am bad, I do bad things, do evil, nasty things to myself, to people, to anyone that comes close. I am bad.   Punish me.   Hurt me.  And when all is over, forgive me.  Is there a god? Am I on my way to meet him? is there some immense, eternal being at the pearly gates…..is there such a thing as the pearly gates?  i….don’t know.   I hate knowledge.  Fuck you Eve, again.   urlLink  Maybe it’s time to re-prioritize things in life again. What is it to feel love and not feel loved? Hurt. Emptiness. Void. But life continues, right or wrong. Things that weren’t given priority will surface in my life again. Design. Books. Friends. Life. So many things have happened, but none without a good reason. Ah, one of life’s greatest mysteries. To soar high and fall back to the ground and to renew the faith. Afterall, what’s the hurt if not joy?  '),(22393,664485,'15,January,2003','2003-01-15 00:00:00','   No sunlight pierces through, only the sharp stinging ice drops falling. It is bleak and somber, slow…as if almost melodramatic surrealness. I’d like to believe the world stopped turning a while, took a break too…teardrops from heaven they say; why is god crying?   “God because you look at everything and still smile,”   Surprisingly very sober now, flashes of last night makes breaks her into a smile. Their little rendezvous into forbidden land, rolls of raw fish consumed washed away with laughter and good company. Twirling around, she is dancing.  She feels good and bad at the same time, the cold doesn’t help. Her bones ache as she shivers, curls up closer to the fireplace but it’s okay, she reminds herself. As always, things will be right.  All these people coming and going in her life; bad pick up lines that make her giggle and charm that makes her blush. It’s interesting to see how the animals devour their meat, fend off allies that become equal predators….basic human instinct. Fight.  But she sits at the sidelines and observes. Poor silly things, she’s not looking; does that mean she’s leading them on? No, she’s been quite clear about it, she remembers, yet some part of her can’t help but wonder why his hands finds hers as she jokingly yanks it away and wags her finger.   Cock teaser.     It’s Thursday.  …and I know that apprehension. That anxious feeling that makes you want to run.              Especially when everyone who “knows” says you have to do what you don’t do. And you know you don’t – so why don’t they know you know?  ‘Cause sometimes what we know – we don’t.  It’s just talking.              Blogging out loud.  To a stranger’s face that, let’s face it – you might never see again anyway.  Don’t be too smart. (You are sometimes – and you know that.) Too smart “for your own good” – whatever the fuck that is.  It’s just talking.  So talk. You dazzle when you talk. You impress without pressing too hard.  By the time you leave…they’ll know why you’re so confused sometimes.  (Laugh as you drive away…now it’s their problem!)  Love you      you remembered;  .....thank you.      urlLink \"The most seemingly difficult decisions to make are actually quite easy. Once you make the decision it seems natural and you wonder why it was ever so difficult to decide in the first place.\"  '),(22394,664485,'15,January,2003','2003-01-15 00:00:00','   “A desire for control and to avoid responsibility for growing up.”   This is an irrational world. This earth, these expectations and these beings are harsh…tormenting. Voices in my head, that sounds so damn psychologically sick but damn voices in my head.  I think I have it all figured out, this game of life we play. Checkmate. You lose, I win, the devil says.  I’ve been feeling shitty the last few nights, the prognosis has been getting better and better but I feel weaker and weaker. I don’t know if I am physically tired because of the illness or if I am depressed, in any case though, I skipped out on auditions last night. I figure, I’ll just pay for the consequences later…  But later is today, when I think about all that’s going right in my life and how damn insecure I feel about it. Scrubbing this harsh muck of dirty grime off and not quite very sure of what lies beneath, wanting to take control but unsure of where I’m suppose to head.   I am coming out and feeling good and being happy but all these emotions and feelings scare the bejesus out of me. It draws back to that question of whether I am enough, and when there is nothing else externally to blame, if I can accept who I am as a person and if that person makes me good or not. Enough or not. Secure or not.   I am so damn fucking irritated with myself. Angry with myself for questioning and being unhappy with myself. It’s a chicken or an egg thing. The more icky I feel, the worst physically I am, and the worst physically I am, the more icky I feel.  Some days I want to go against all that I’ve been fighting for and say, just drug me. Just make me better, just take this shit away. But then it takes away everything and fucks up all the effort and time I have spent going against it. Maybe, sometimes, we all just reach a stage where we learn to question?  Damn Eve for eating the forbidden fruit. Damn Pandora for opening the box. Damn the gods for feelings, and growing up, and salvation.  I realize every time I reach a stage where things are going exceptionally good I fall. Or breakdown. When everyone pats you on the back and says wow, and people around you make you feel so important; you feel the opposite.  Stupid damn thoughts.  Stupid damn questioning.  Stupid damn mind.    urlLink \"blah. icky mondays.\"  '),(22395,664485,'14,January,2003','2003-01-14 00:00:00','   I accomplished a lot of things today.  Yet in some sense, I did absolutely nothing for me as a person at all.  I read, and continued reading, even as the pain ripped through my heart, the remnants of it anyway. I don’t know how love went so wrong, or how something that seemed so right and perfect could have been fucked up so bad. And the only answer that remains is that it was all my doing. My bad. My fault. As always, my desire to make things right messed things up further.  I opted for no treatment. I didn’t realize the symptoms would come so fast, so hard. Maybe I’m rethinking my decision, wondering if this is truly right.  I honestly do not know what I want, it feels as if it gets more and more confusing by the day. I am physically so worn out my mind is tired too.  I read, felt. You know how you watch certain movies and feel the protagonist/antagonist pain and understand all that shit that runs through. It felt worst to have understood rather than the guilty pleasure of being secure. Or right.  I watched at how he tore apart me as a person…and question who I was. Who am I anyway? Is everything he says true? I watched how he mocked my relationship, our relationship, how it seemed so easy for him to talk about being raped, getting sick, being left.   As if each nail he picked up only drove straight to the right point. It hurt. I hate that it does, I am strong enough for this damnit. I am okay.  You don’t need him. you don’t want him. you cannot love him anymore. He’s not the one, no one is. You’re not enough.   Maybe some day he’ll find out…and maybe some day he’ll know. understand. Will it be too late then?  Probably.  Why am I here again? Why is he so unreachable…untouchable? I just feel so not there as a person right now, my chest hurts like a damn fuck and my ten pound chicken sits securely on it. It is frustrating because I know it would be easy to fix this immediate problem but I’ve done this alone before…I don’t want to do it again. It’s ironic how he has become my steve. Raped, left, broken, begging again.  I’m tired of begging, of explaining. He didn’t care, he won’t care anymore. So maybe he will always love me, but maybe he just loved the racheal that was never racheal with him. that maybe the only racheal anyone will ever love is the thin, secure, happy, healthy racheal. Maybe she’s the only one that deserves love. I seriously thought I was leaving last week. I went skiing instead, went to go continue dreaming, relax. And I came back so fired up, so ready to go, so complete…but I am a complete absolute mess tonight. Crying and hurting, throwing up more times than I care to remember, feeling so nothing and bad about myself…it’s so wrong.   I know I am a someone. I know I have hope and I’m beautiful and I’m strong. I just wish that he saw that too, that he’d say I see you for who you are and you don’t have to pretend anymore, I accept you.  Funny isn’t it how the one person you want is the one person you can’t have? Is the one person that you should probably stay away from.   It feels as if there is so much riding on me as an image. The person on the outside, nothing bad ever happened when I was pretending. Having problems doesn’t justify cheating on your boyfriend, I perfectly finely understand that…but I’m just fucked up and I don’t know how to fix me anymore. I don’t feel strong enough or good enough for this.  Part of me just wants it all to be over. Take me and be done with it. Throw me into the fiery pierces of hell or whatever, I don’t care. I’m done.  Why the fuck do I have to question that?  Fuck you cancer.    ”and nobody will ever know. i\'m good at keeping up appearances, don\'t you know? just like a clockwork orange.”  '),(22396,664485,'13,January,2003','2003-01-13 00:00:00','   urlLink  \"She gave her flesh. She gave her heart. She gave her soul.  And in return, She was left with nothing but the memories. The bruises. The nightmares. The scars.  Discarded. Left to hang helplessly from the rafters within the dark recesses of her mind. Once more discarded into the black, forcing the familiar journey back to the light.\"    Two points for determining my factor of existence tonight. It’s been a long week; one of those where your days and nights are strung together and your basic necessity of staying alive is coffee and endurance. Classes filled to the brim with people, anonymity comforting, nothing to touch bases at the end of the day. Reflection in the mirror ugly to the eyes, analytical mind criticizes, tear apart my being. Highs and lows from the environment around.  It’s been a good week though; parties, friends, people, dates, school, stepping out, being safe. I’ve had the security of people around me the last few days, backing me up, the strong arches that rectify my sometimes not so very solid foundation. It is nice to feel like I’m wanted sometimes. It is nice to feel wanted.  My appointment is this Thursday. I am scared and excited; it feels as if my life is finally going somewhere but at the same time, I am kinda displaced and thrown off about this whole new world I am heading into. It is hard to stay calm and focused when daggers penetrate deep into the skin from every angle and turn. It is even harder sometimes, to just stay alive.  Beautiful things are being weaved into my life though, little miracles that almost make everything okay. I keep on reminding myself to not fall in love, as if it is up to me. I will not put myself into a position of having to need someone again. It’s terribly selfish I know, but I need to be strong for me now….need to do this for me.   We’re all moving on and stepping into the new year. Finally grabbing hold of these reins to take charge, I will not make this another wasted, unaccomplished, pointless year. The initial surreal wonder of having slipped into another year has faded, all that’s left with resolutions is fixing. Living.  I like all these new people around me. Some perverse part of me enjoys being told I’m beautiful, enjoys being taken out and feeling like the most important diamond in the world. Maybe it’s the little crazy part in me that has fun dressing up and giggling like a little school girl, blushing profusely when complimented and feeling good. Is it pride? Is it reassurance? Or is it just human nature to like being chased after? Liked.  It is confusing to the naked mind why I would choose getting mentally better over physically better. Why not do both? Why not fix both? Is it a mind over matter thing or is this just another one of my not so great decisions? I’d be lying if I told you that I knew the exact reasons why, it just feels right I guess.  Beggars can’t be choosers.   urlLink  \"if you\'re reading this, don\'t reply. because, not only will you have ruined my long-standing perception of friendship, you will have also ruined my rant.\"  '),(22397,664485,'12,January,2003','2003-01-12 00:00:00','   Booyah.   urlLink  When a cute guy asks you out, don\'t respond with, \"But, I didn\'t think you liked girls!\"  '),(22398,664485,'11,January,2003','2003-01-11 00:00:00','   Hmm..so we’re back to the sending your ‘friend’ off at six in the morning again.   :P  Question mark, raised eyebrow, harmless query  …all of the above.  Go figure.  Life is…very interesting right now.  Ahhhh.  (Whoop. Whoop.)   urlLink  \"A normally lonely little girl lay naked on pink silk bed sheets with another girls fingers playing adult games inside her burning hole. Heaven.\"  '),(22399,664485,'11,January,2003','2003-01-11 00:00:00','   You tell me I’m oblivious to beauty…that my four walls of security has confined me. And I blush, stumble over explanations and words, hell bloody shit why am I shy now?   Things are changing in my life, things I cannot predict nor really visualize. At times, it feels as if I am falling and leading myself into bigger trouble, slapping myself in the face with more warts and wounds to cloud my judgment. I am my own being, whatever my decisions, determine the outcome; all my actions are of my own right and doing. I cannot put the blame on others anymore, I must take responsibility….but damn, that is a big load.  I am equipped with different gifts. Talents they say. But many times I have felt as if it was all just a show, a façade, because many times I have failed. My grades come from hard work and the shows honed from countless hours of practice…precision even. It does not come easy, the different languages a barrier…translation goes on in my mind…I feel retarded. Disabled almost.  But I am not. I am whole in being and breathing. Heart beating and life living. I am here and my moments are being executed and made, every second of my life is being recorded in some book above…will I live my life without regrets?  I am a person. I am a being with feelings and emotions…perhaps even at times, way too much sensitivity points. I am person that can be beautiful, can be something, can be someone. Sometimes I want to say I just need you or someone to believe in me, and I catch myself constantly evaluating the people around me, not realizing that all the insecurity lies within. I make my own person, it is me that needs to believe in me and me that needs to know that I can do it. That I am someone to somebody and have so much more ahead of me. Is life gorgeously enthralled in ecstacy? Hell no. but it is quite damn close to at times…the times that I  believe.   urlLink  \"I\'ve never been this happy about someone proving me wrong. \"  '),(22400,664485,'10,January,2003','2003-01-10 00:00:00','   So I’d really like to tell you that I had wild, explicit, back-to-back cosmic orgasm sex last night…but no.   Hehe, instead we went dancing, listened to some bands play, had brilliant conversation, laughed and had a good time. Tadahhh!  And so it’s one down, two more to go…I can’t stop giggling..it’s retarded really. I feel like a million bucks right now, I guess what they say is true; confidence does attract the masses.  Ah well, too bad I’m not one hundred percent interested, or rather i\'m in that limbo of just enjoying the company not looking for a relationship stand thing…though it does allow leeway for some fun then eh?  SLUT!!!  :P    urlLink friday five  blatantly stolen from  urlLink jasmine :    1. Do you wear any jewelry? What kind?  yup…mostly silver or artsy shitty weird things I pick up..i don’t know. it really just depends on the mood and whatever I’m wearing….though I’m also the idiot with purple straps and a hot pink shirt on right now so maybe I should just shut it eh?    2. How often do you wear it?   quite often actually though lately I have been obsessively wearing these gorgeous silver hoop thingamabobs besides the usual hoohaa of rings and bracelets…I like to make noise :P   3. Do you have any piercings? If so, where?  eight piercings..you go figure.   4. Do you have any tattoos? If so, where?  no :( and damn fucking pathetic too! I’d really like a butterfly at the base of my spine but it’s been known to hurt like a bitch plus the current social stigma of every other tom, dick, harry and their mothers having a tattoo. Oh well.   5. What are your plans for the weekend?  I have a date tommorow night…(Whoop! Whoop!) with some guy that I’ve sorta been talking to for the last few months…in fact, come to think about it, he was one of the first people I met but then because I’m such and dork and he’s shy and we never got our shit together…it’s taken us this long. Yes, retarded, I know. but then like I said earlier, it’s funny how it works either that you’re instantly attracted but shy as fuck or it grows on you. So anyway, there’s that, swing dancing tonight, and maybe, if the cast can coordinate their schedules together, a movie on Sunday! Oh, and if you get me drunk enough maybe even some studying.   Right.  '),(22401,664485,'09,January,2003','2003-01-09 00:00:00','   Hmm…is hotmail not working or is it just the on-running conspiracy against racheal?   Pah, well anyway, guess who has a hot date tonight?  Yum.   :P  Ahh..being single is good.  *nngeah!*   urlLink  “I will not sulk about having no boyfriend, but develop inner poise and authority and sense of self as (wo)man of substance, complete without boyfriend, as best way to obtain boyfriend.”     Yeah, whatever.  Woman.  :P  '),(22402,664485,'08,January,2003','2003-01-08 00:00:00','   Ah…isn’t it nice to just hang out and fervently discuss your sex life with the ex?  :)  hmm..so I  canceled  rescheduled the appointment with Gayle. Why? I don’t know really… just felt like it  :P   nah…for real though, I seem to have this little problem of taking on too much at one time and this would be a very good example.  In this week, I’ve had to unpack, get settled in again, explain myself, get my ass out of the hot seat, start school again, schedule my classes, get the right books and find the right classes, figure out how things work out and all that good stuff. To sugar coat things in a nice way, I wanted the experience of getting better to be an experience of it’s own rather than a hodge podge mix of everything else and between.    I wrote this quite very poignant and emotional eulogy last night for Brian. In my drunk stupor and rocked out good self, I had captured all these emotions I had inside and somehow I lost that post. I’ve had a few days to sorta think about it now and I know he’s in a better place. Heaven’s gained another angel and maybe he’ll rest in peace now…  The whole saga with getting sick again is still around somewhere. Maybe just not quite as dramatic as it was before this. Once again, something else that I can’t explain…giving it up to fate or god or whatever really. Reality check today was watching the insurance rocket up just a little higher and taking the smug smile off the bitch from hell needle stabbers face.  So school started today…to say it was a good one would be an understatement. It was nice to walk the halls, pass friends by and be stimulated mentally again. I have sixteen credits again this term and maybe with the shadow box production whether it’s set design or actual acting, there’ll be an extra one. I am in no hurry to graduate but it feels good to be productive and moving somewhere. Aiming for something at least.  We arranged it so I only needed to be in school on Mondays, Wednesdays and Fridays as opposed to being there every damn day for too many damn hours in between like last term. It’s pretty much basically back to back from eight till half past one in the afternoons. And note to my stupid self, it is not a smart idea to go out the night before and have to wake my sorry head pounding ass at six in the morning.  I do have some pretty exceptional lecturers though. Much thanks to the beautiful people that get my back…ahh…it is pretty darn nice to feel special. I have the head of the history department teaching me classical mythology. Professor Hagedorn for comp 2, head of the communications department for survey of mass comm. plus even the same director for college publication. Then there’s mr dean, the director of the musical, for acting 2 and Washington for algebra. My classes are pretty much packed to the brim with people but it’s interesting….:) I am excited and love this feeling of having something to look forward too. Remind me again though when I’m up my ass in homework and reading, five out of my six classes have comprehensive reading and writing…ah…what great fun.  Hmm..it’s strange but I feel loved today. And shit yeah it’s a good feeling.  Bite me!  I love you  '),(22403,664485,'08,January,2003','2003-01-08 00:00:00','   About two weeks ago I met an amazing person.   His name was Brian. He was funny and kind and one of the first things I ever remembered him telling me was to not give up.  See Brian knew my battle, and my fears. Somehow, through emails, I had found someone just like me…quite lost and confused but supported. We talked about how it wasn’t easy, that good wishes only brought you so damn far and that when you were puking your guts out and staring your sorry sick disabled body in the mirror…it only remained so.  That very week, darrell, his twin brother, emailed me…telling me that even if I felt that I was losing to remember all that I said to him when I first posted. I had written this long overdrawn entry on how hypocritical I felt and how hopeless and pointless it all seemed to be back here again. My prayers seemed fervent but yet, reading that and thinking about how someone I did not know at all and someone I hadn’t even met was telling me to press on and not give up, touched me.  A week ago I wrote them saying that I had had enough of this shit. That I was ready to die and give up. I wrote and begged, asking for some merciless soul out there to shoot me, perhaps even kill me. They were posts on a forgotten teenage journal with angst and fury to boot. I was crying out and breaking down.   The last whole week has just been breathtaking. I built again, dreams I had thrown away and found courage, strength and peace. This is seems so awfully cliqued but it’s been one of those weeks where you’re going and going, running and running, and you really have absolutely no shit idea where you’re going or what you are doing and you finally realize all the things you truly want and how badly you will go to get it.  I wrote last night, summary of all that has happened in small, stoic posts. I had chosen to be away from everything, disconnected. Coming in last night I had thirty-seven emails, fifteen messages and a dozen people to talk to. News spread around, the bitch was back in town ready to kick some ass and start again, I had boxes and boxes to unpack and a thousand and six things to do on my list. As always I was too busy, too preoccupied with pointless shit and too caught up with my drama, and my life, and my shit.  I found out today heaven gained another angel on new years. My last words to him was new year, new resolution, new hope. He didn’t make it past midnight of December the thirty-first. I am…I don’t know what I am right now or how I make of it.  Shocked is a good place to begin.  I don’t know how he crashed so fast and so hard.  A week ago we were talking about remission and biopsies. Of going home and the holidays. Merry fucking Christmas right?   Life really is that fucking short.   This is so surreal. I want to scream and shout…maybe even cry. But my feelings aren’t coordinated enough to make sense of everything right now. Fight harder or give up? Is this how you would feel too if I didn’t make it? It’s just unfair right now. How? Why? What? Who? When?  Zplaid1, you will remain in my heart forever.  For the person that fought/pressed on to the end, I wish you peace.  Thank you.   ....sorry you were robbed.  '),(22404,664485,'07,January,2003','2003-01-07 00:00:00','   Ten states.  Eight and a half days.  Two years.  No connection.  One conclusion.  Apparently, Wednesday I go in to see Gayle. I’ve been thinking a lot about it, what I want, who I can be, what I want to be…where I’m going, you know, all that horse bull crap you get with the end of a somewhat indefinable year and the beginning of something else not quite understood yet.  I wondered if this would be the very last time I said this, that I was a better, bigger, stronger, incredible human being. That I was, among everything else, quite possibly beautiful. And some days, I believe it, some days, my head is stuck so far up my ass that I walk the walk and talk the talk. I am gorgeous, and thin, intelligent and charming. I am distractingly amazing and I am absolutely everything you want.   But nobody has, has me. Perhaps even as far as to say, nobody can get me.   It started a little like this, I begin to say, explain everything that has happened in the last seven days. Hell, even trying to put the last twenty four hours into perspective is hard…and you want me to do the whole damn week?  Maybe my little heart told my little brain that it was dying. And maybe my little mind told my little body to just give up. And so I did. I gave up, packed, lost it for a good couple of hours and then came home. Or ‘home’?  Ridiculous.  Stupid.  Insane.  Fuck waste of money.  In the last few days, I have seen mountains and ridges, snow and sunshine, land and air, people and faces. But the best discovery was finding a place in me to lock away and cry. Yes, cry.  Cry because I am nothing like the person I want to be. cry because I keep on letting the ones close to me go, cry because I want them to come back but I know now that I can’t have them do that. cry because I know that despite how dirty, ugly and tainted I get, that I still deserve love. And love is what makes me worth it. Love makes hope. And I have Love.  Skiing and sledding. Climbing and tracking. Driving and horse back riding. And I learnt about trailers honking and blowing soapy liquid bubbles. Rainbows after the rain, snowmen (sometimes women) and angels from heaven.  He asked me where angels came from. I said deep within. And so he ran to mommy and said nana was made in heaven. I rubbed his boo-boos when his three year old uncoordinated knees scraped the ground, and he fell asleep to lullabies that I sang him, wanting to protect him from growing up, from wanting to find out who he was and where his place was in the world.  Scares. Getting knocked up. Breaking up. Running away. Getting sick again. Doing so damn well and so damn horrible at the same time. Pushing, testing, accelerating.  I’ve been running on empty, and empty has been an understatement.  Growing bald in my slowly emaciating body, wheezing, it’s hard to breathe….abuse me further. I dare myself to.   ‘It’ is under control. Monitored. Whatever the fuck that means really…but if it means more time here than hey, that’s good…right?   I promise to take care of myself. To quit smoking. To eat properly. To fight the best I can. To go in when I have to and speak up when I need to….but do you think that’s easy?  And nobody said it was easy. And nobody said I’m not going to come back some night tired, sick and alone. But nobody also said that it was impossible.  Maybe as long as there is still all a hundred and five pounds of Chinese shit mass of me then I might still make it.   Nobody said I couldn’t try.  And nobody should ever make me feel the shitty way I do again.  But the question is, it isn’t them, or him, or they, or her.  It is the being in me that gives up, breaks down and chops me into pieces. It is my doing that I am who I am.   Pray for whatever I want but the work needs to be done right?  Admit that I’m sick. Admit that I’m bulimic. Admit that I have problems. Admit that I can be pretty damn amazing. Admit everything and back, but a year ago I made the same promises, the same resolutions. A year from now…will it all still be the same?  Change doctors, change treatments, change prescriptions. Can I change bodies and minds too?  I don’t know. but I’m alive. And still living. And fight your god damn sorry self he says.   And fight my god damn sorry self I will.  I am beautiful goddamnit;  …fight you little fucker, fight.  '),(22405,664485,'28,February,2003','2003-02-28 00:00:00','   at the old man\'s house...hiding away from the rest of the world.   (cast party downstairs..i said my hie\'s,  showed off, then walked my butt upstairs and so, here i am)  down time i say...thinking about the performance. it\'s significance. when it gets bad, do you think they\'ll send me to a hospice?   i wonder sometimes.  weirdness.  cigarette break.   i promise i\'ll come back mom.  '),(22406,664485,'27,February,2003','2003-02-27 00:00:00','   now if i fall asleep in class tomorrow, blame the smartass in maryland.  :P  p/s old man is safe/sleeping/alive/sane. yay.  '),(22407,664485,'27,February,2003','2003-02-27 00:00:00','   old man...where are you?  i\'m worried.  call me.  :(  '),(22408,664485,'27,February,2003','2003-02-27 00:00:00','   something about certain thursdays that breathe like sundays. the laziness...carefree nothingness that hangs lightly in the air. floating through as if just another breeze...simplicity at it\'s best.  a fresh pot of coffee made, utensils thrown around, overalls splattered with paint and hard-worked fingers dirty with paints and crayon specks. calligraphy tools and bottles and bottles of ink stacked across the shelves, old tattered t-shirts strewn around as rags used to soak up spills and wipe off clumsiness..broken glasses, pencil shavings, cloth pegs and superglued velvet used to explore different mediums.  camera recharged and solid prints lie scattered on her desk. strips and strips of photo paper cut-out fight for space against the other seven hundred and forty two projects she has lined up. disks with a hundreds of pictures scanned in for her latest endeavor, a personal challenge she\'s taken up on herself. she must be bored really. or just dumb.  just don\'t rub it in when she starts tearing her hair out frustrated and whining...  :P  the television hums in the background, more depressing news reports yet she keeps it locked on cnn, as if delibrately reminding her of the \'real world\'. real things out of the space she resides in, out of the four corners of her own world. her selfishness and feelings about wants/needs/dreams/hopes falls shallowly among worldly matters.  someone is hungry.  someone is dying.  someone is homeless.  someone is fatherless.  motherless.  soldiers deployed.  terror reigns.  ...another debate on war. another point battled and fought over again. she slips on covers and sheds off layers. intellect battles, points that have to be made...some sort of agreement to be on the same page and yet his intelligence has only charted up more points.   she wonders what this means.  her heart rather skelteredly opens. tests waters.  she knows she enjoyed her. him. she enjoyed the craving sufficed. being held. loved. she enjoyed being close to someone again, physically satisfied, mentally complete, sexually recharged. she\'d begun to think her libido had went into the non-existent, nothing attracted her and she had shut off contact, emotions and intimacy completely.  one by one she shot them down. never opening up. never giving in to the possibility...yet she realizes what a vast mistake that was. could she have slipped that person by? the one that would complete her and make her happy forever?  now as she realizes how amazing it feels to be needed. be loved. be close to someone. feels. she wonders..  nothing will shut up the voices but she\'s not too sure if she wants to shut them up anymore. the extra pounds in the mirror she sees? she (learns to) love them now.  a part of her.  so she\'ll be fat and loud and bitchy and opinionated and talk up and through her ass.  ...big fucking deal.  so what?  :)  hah.  ....renewed.    urlLink  \"muse has a date....of the bubble gum kind.\"    heh.   :)  '),(22409,664485,'26,February,2003','2003-02-26 00:00:00','   i picked up a box of oil-based crayons today.  you know, the type with vivid colors that merge quite nicely together but leave your fingers greasy and speckly?  :)  wal*mart trips at two in morning after a waffle house session with everyone and their mothers is great. stupid, silly, retarded and definitely unproductive but there\'s a distinct feeling of pure utter joy in watching your friends pull up rather coincidentally and having made friends with the table next door that just so happens to be friends of a friend. it\'s a small world indeed.  as the old man pulls up a chair next to the giggling bobsy twins and escapades are rehashed and retold. colored brightly and painted into someone else\'s world, the jukebox slammed with quarter after quarter of late eighties gold her father taught her to listen to, country bumpkin crap she\'s learned to appreciate and a host of pop songs she\'s too embarrassed to buy cds of.  i just killed my streak of eight wins. i fell asleep.  yes, my literati ass crayoning at this obscene hour just dozed off lying on the floor with my fluffy soft pillow.  well, that sucks. but oh well.  i\'ll get over it.  :)  scattered around are \'expressions\', they call it \'art\', i call it boredom at five am and a fascination with colors.   inspired.  so i had this phone conversation today....  :P  '),(22410,664485,'26,February,2003','2003-02-26 00:00:00','   so i wouldn\'t exactly say that i was \'arbitally forced\' into writing this post..  though \'coerced\' might not be too bad a word at all...:P   urlLink  \"Girl, you wear that outfit again, and we\'re gonna have to sell tickets just to walk down the hall behind you. There was more \"backfield in motion\" going on there than the Tampa Bay Buccaneers.  So, who is it?  DETAILS, girlfriend, DETAILS...  And I don\'t want any of that eye-rolling, tongue-clicking, breathy-sighing bullshit either.  Shut up, sit down, and let\'s talk.  What\'s his/her name?  (I love the idea that I can\'t even guess the gender with you, these days...Oh, yeah, you can fake this exclusive lesbian thing with the less-attentive, but I notice, sweetheart. I notice. Your front door is swinging both ways these days...I can tell.)  So out with it.  Those of us whose sexuality isn\'t threatened by you exploring yours want to know.\"    see...now how can anyone resist such shameless ass kissing and not revel in it?  :P  i did recieve a couple of interesting emails with regards to that silly post i put up this morning.  and just for the record, no, i\'ve yet to sign the deal with showtime just in case you were wondering...but at the rate the spamming hits my mailbox, and should i decide to \'enlarge my breasts the safe and easy way\' and lose all senses, you\'ll be the first to know.  :P  hm...i also reconsidered putting up the rant and good stuff people had to say because i figured why rub salt on the wound eh?  lol, too much typing anyway. and hell, i want to talk about  *me* !  it\'s about a quarter to three on thursday and i\'m already into the next day. it\'s been one of those silly crazy retarded days. grins plastered and silly giggles to no one and everyone..it\'s nice. :)  went to go bug ryan at work then caught the three-dollar movie (complete with nachos, chips/fries/potato-things with about fifty gallons of ketchup and jalapenos that i swear are fighting against the butterflies in my tummy!) after church. yes, she went to church, again too at that! hehe, i even agreed to agonize our choir director with my majorly transposed version of \'requiem aeternam\'. good shit i tell ya..  and ryan.  damn.  no, i mean dah-yum.  :)  in another realm completely though..the head doctor and i had a pleasant chat today. i actually feel good and strong. laughing and making jokes and taking cheap pisses at her lousy green sofas and lack of comfortable pastel colors in the room. brought in my paper journal and read out some recent entries...it\'s quite eye-opening what self-realization does to you. :P well, at least she\'s taken me off the treatment center list...baker acts are damned piss scary.  to be quite honest, i\'m actually feeling rather mentally sane. it\'s strange, do you think it takes acceptance to reach this tranquility? or a busy schedule, bizarre budgetting techniques, scrabble games, and silly kooky loving friends to do that? i don\'t know...but whatever it was, it apparently worked!  yay for mental health care.  :P  hm..i have to go in and settle out the transition/moving documents with the university tomorrow and start looking for a place + all that good bullshit. i can\'t wait for my mom to come!  i don\'t even care about the hospitals or sticks and needles anymore. i just miss her...i miss having a mom around. or a dad too. and i figure it would be so nice to just sit and talk over a warm soothing cup of tea even if it rains outside but to feel the sunshine from within. i\'ve really learned/matured a lot in these last six months about life and my crazy silly self but moving so far away will do that i suppose.  some part of me is afraid of losing a part of me that\'s slowly being left behind but i know i\'ll be alright...and things are genuinely okay. love is always such a beautiful feeling and love heals.   it\'s funny how it takes loving you through the rain to realize how capable or worth of love you are. i am.  and i am.  stronger, slightly older, slightly wiser but nonetheless, better.  :)  it\'s been a good day...  two points!  '),(22411,664485,'26,February,2003','2003-02-26 00:00:00','   smart ass.  that\'s all i\'m going to say because it\'s a good day today.  *sniggers*  /delete/  whoop. whoop.  there\'s something very wrong about finding this liberating. or right.  what do you say wrong reverend?  :P     \"and racheal...well...i don\'t know if i can put a positive spin on that one. so instead i\'ll make a joke. in my life, i\'ve never met someone who\'s sucked so bad at suicide. i shouldn\'t make jokes about that, but honestly. if you\'ve tried to kill yourself as many times as she has, and you still haven\'t done it right, i think you need to find another hobby. i say starring in pornos would do her good. it\'s obvious she wants attention (as anyone who repeatedly \"attempts\" suicide does) without the permanence of death...and she apparently likes fucking random guys. that way she\'d get attention and dick and money to boot. it\'s win win win for her.\"   show time video...here i come!  '),(22412,664485,'25,February,2003','2003-02-25 00:00:00','   i met someone tonight....by pure coincidence though i must say, it was quite nice.  i think i\'m a little flushed. flabbergasted even.  it\'s that nervous shiver you get after a long kiss on the second date...that anticipation that sorta rumbles in your tummy as you try and keep yourself from squealing or giggling out loud. the excitement as you catch yourself smiling to no one from no where and just because.  it\'s a good feeling really.  if only it didn\'t involve conscious thoughts and primitive actions. if only it didn\'t involve a heart, a set of limbs and a mind multiplied by two.  i wish. i wish. i wish.  i should go to bed, perhaps not think this to death.  yeah fucking right.  and he respects me too.  good fucking god, what is this world coming to?  so i\'ve waited for the beautiful ribbon stringed bow package to arrive..and apparently it has. i feel as if i should kick and throw it around to just make sure it\'s not going to explode on me then implore all it\'s horrid contents on to my now cleaned a step before faded jimmy choos.  we talk about bubblegum lesbians and he strips me off my sexuality. i explain how i enjoyed the brief fling exploring her body two days ago but it was extremely different. nice soft different but still nonetheless, different. and i crave for this closeness, this new found freedom. intimacy.   i tear my insecurities to pieces and laugh along with him as he rehashes anecdotes to politically correctly address his ex. i don\'t know why but there is a distinct feeling that i would have been more comfortable with this conversation had she had blonde hair, blue eyes and big huge boobs. no, instead she was japanese...and already my sirens are blaring. albeit young and perky, cute sexy thing la di da...i know this drill very damn hard.  i joke around and mask myself while unlayering so fucking well it\'s almost as if an undefined talent. he talks about writing and the people inside...and i shrug and just say, well shit happens. some part testing waters to see how much shitforbrains still means to me and i cower a little, attempting to rectify the mistake of letting him bring it up and want to just smile it off instead. forget the hurt that idiot has ever caused me because hell knows, i\'ve did my fair share too. so i focus again on our conversation and smile a little, giggle even, as he rather noticingly stares in awe.  i skip, dance, twirl around. choke the butterflies in my tummy to death then grin again.  oh fuck me now.  '),(22413,664485,'25,February,2003','2003-02-25 00:00:00','     '),(22414,664485,'25,February,2003','2003-02-25 00:00:00','   i curl into different positions...hoping at least to accidentally slip into one that might hopefully be comfortable.  watch, as silent sheep pass me by and the city sleeps, falls down into tranquility and wraps around.   \"Somebody oughta take you in Try to make you love again  Try to make you like the way they feel  When they\'re under your skin Never once do you think that they would lie When they\'re holding you Then you wonder why they haven\'t called When they said they\'d call you  You\'ll start to wonder  If you\'re ever gonna make it by You\'ll start to think  You were born blind\"   i\'m frustrated.  there\'s just this icky heavy feeling that angers me from deep within. it\'s that itch that won\'t go away. a certain feeling that creeps beneath my skin and irritates the hell out of me...and the ironic thing would be that it\'s all me.    me.   go the fuck away voices.     urlLink  i think that we are setting out to either be nuns or become lesbians. yes, i know muse, you will be the first i call.  '),(22415,664485,'25,February,2003','2003-02-25 00:00:00','   he\'s got me figured out..or so he thinks. all seven hundred and fifty two fronts he says. fronts.  i think maybe he does too. a little.  and perhaps that\'s why it hurts. or stings so badly.  because he knows.  sees.  this self-protection shit is fun. absolutely mind juggling(boggling)...keeps the brain cells active i suppose.   i\'m fucked.  sigh.  but hey, where else can you get an asian-american-non-chicken-eater-nut-hater-language-muse-stupid-retarded-ass for a rollercoaster ride eh?  '),(22416,664485,'24,February,2003','2003-02-24 00:00:00','   i\'ve come to realization that when the people at claiborne tell you to dry clean only, they mean  dry clean only . i sit here with black fluff shedding off my sweater and clinging to my skin from the static. it is not a pretty sight let me tell you. nevermind that my lazy ass was trying to save a few dollars..(okay yeah, i was just lazy but play along with this story will ya?:P ) but nooo...sheer curiousity got the better of me.  gah.  '),(22417,664485,'24,February,2003','2003-02-24 00:00:00','   i know you exist somewhere between these parallels. your lips moistened by falling rain that pierces through another atmosphere, i hesitate to say drenched because that seems so weak, and you are definitely not. the water soaks your shirt and sticks to skin, your guises are naked, your breasts swelled and peaked. it\'s cold outside but your smile is warm. inviting. you sway and motion a little as your fingers beckon towards me...taunting me. your sparkling eyes flirt and your sexy hips dance.  your lips mouth silent o\'s and a\'s but they emit nothing. in this world, the only rhythm i hear is the rain slapping against the pavement and even that has a melody. a pattern. a chant that evokes a trance-like, perhaps peaceful embodiment.  the sun sets at the end of the boulevard but right before, a single lonely ray pierces through the droplets and forms a rainbow. a promise. seven colors of change. seven colors of hope.  seven colors of love.  '),(22418,664485,'24,February,2003','2003-02-24 00:00:00','   i feel.  i hurt.  i touch.  i see.  i love.  i need.  i give.  i try.  i heal.  i stand up.  i brush off.  i walk again.  that lonely mile.  that cold solitudal wind.  that forceful voice.  why does it haunt again?  why does this cliff seem so easy to just jump(fly) off rather than to traverse down?  why am i weak?  selfish?  endings and beginnings.  leaving and coming home.  going and returning.  one minute happy, the next minute sad.  one minute confident, the next minute scared.  one minutes that never get past twos.  hiding, trying, falling, searching again.  i\'m tired.  i\'m angry.  i\'m broken.  i, me, her, she, saya.  empty.  nothing.  fallen.  small.  '),(22419,664485,'23,February,2003','2003-02-23 00:00:00','   there\'s something very distinctively soothing and disturbing about matchbox on rotation.  i don\'t know what it is...these bitter tears aching to flow are intoxicating. distracting and therapeutic. something about the half scattered magazines on the floor and the wafts of perfume samples too fucking familiar against the accoustic guitars, drum beats and his raspy sexy voice. lyrics that simply just hurts. hurts because it\'s so damn relevant or hurts because it\'s everything i want/need/have to say...about going home...about finding myself...about all these things..\"things\"  sacred, untalked about, hushed bullshit.  stigmas.  diseases.  wanting to scream and scratch styrofoam together to hear that squeaking, piercing, irratable whine of friction mounted together. sado-masochism perhaps..i don\'t know what it is..  i don\'t know what it is..  maybe just unwell.  tired.  but still breathing..  it\'s hard to be in the middle.  '),(22420,664485,'23,February,2003','2003-02-23 00:00:00','   urlLink  \"I think I need to stop hanging out with heterosexuals.\"    male lesbians that whip me, hit me, and call me edna especially..  :P  kids are amazing aren\'t they?  i want to hug, kiss and swing her around..she\'s torn the whole bitch-hates-the-world attitude i\'ve managed to carefully brush on today...just so i have an excuse you know?  but as she runs forward to the candy machines and her tiny hands clasp my big silly ones, i am equally dumbfounded and mesmerized by her innocence  she can\'t decide between the choices and chooses soda instead;  root beer to be exact...then we gang up on picking on the old guy.  we spin quarters and she tells me about powerpuff girls, and the evil mojo jojo, and how she\'s the fairy godmother today.  her four year old eyes sparkle under the harsh flourescent light..  she granted me my only wish today..  i was happy.  free.  content.     ...amazed.  '),(22421,664485,'23,February,2003','2003-02-23 00:00:00','   i could, but i wouldn\'t, but i\'d have to won\'t i?  it\'s way too fuck early in the morning to be  going to  in school.  School.  blech.  and thinking..  thinking bad.   sigh  '),(22422,664485,'23,February,2003','2003-02-23 00:00:00','   still finishing up long, long, loooooonnnnggg, paper.   oh, and kicking chris\'s literati ass of course.  :P  sleepy as hell.  grr.  '),(22423,664485,'23,February,2003','2003-02-23 00:00:00','   somehow, writing a paper and stalking the bubble-wrapped boy on the webcam seems morally wrong.  '),(22424,664485,'22,February,2003','2003-02-22 00:00:00','   just uploaded the template back up again..  am i afraid?  am i nervous?  why did i continue writing?  why did i still have so much to say?  why didn\'t i just hide and suck it in?  bring it on you condescending fuckers.  pore through and analyze my fucked up mind.  see if i care.  and if i do, see if you\'ll break me.  and if you do, see if i\'ll rise.  and if i do, see if i\'ll fall again.  and if i do, see if i\'ll finally die.  because fuck, i\'m still alive.  fuck, my pills are regulated.  fuck, i have a therapist, a psychologist, a psychiatrist, a gastrologist, an oncologist, a physician, a nutritionist, a nephro-can\'tfuckingspellit-logist and a radiologist.  so fuck you.  and fuck me.  and we\'re a happy family.  \"exactly.\"  '),(22425,664485,'22,February,2003','2003-02-22 00:00:00','   it\'s amazing how little it takes to runaway.                                          breathe in.  and out.  and in again.  '),(22426,664485,'22,February,2003','2003-02-22 00:00:00','   Hold it in, I whisper to myself later.  I watched the car pull away, grow tinier and tinier into a distance then disappear. I retreat back alone into this now cold and empty house and walk into disrupted calmness.  I feel the teardrops slide off my face but I do not cry and smile instead. I think about the luxurious weekend I’ve had and the lushness of her warm body beneath the sheets. Her long, thin fingers wrapped with petite little rings and the way her curls frame her face.  We went to a traditional Catholic mass this morning. Her faith against mine I suppose, I’m Lutheran, or at least my parents are…and though the services are very much similar…there’s still quite a bit of differences in them. I loved the incense and candles, the ritualistic/holistic like atmosphere, the father-forgive-me-for-I-have-sinned kneeling and bowing, holy water…and the constant fervor of blessing yourself and making that cross thing. It was beautiful; to feel so lost in a crowd and found.  To have the world disappear before you…retreating into a lost island in the middle of February and have tornadoes crashing behind your porch yet feeling so surreally calm is breathtaking.   Midnight runs on the beach layered up in four layers and an anorak, laughing and feeling your Nikes sink into the sand. Stopping to pick up sand dollars and feeling the ferocious wind push you forward. Her increasingly passionate moans with the crashing of waves outside your bedroom window as her fingers find yours and you discover love together.  The ‘out of town’ rendezvous packs and your magical weekend ends in twenty-four hours.  You step in and you have to pick your troubles back up at the door but you breathe in deeper, and take stronger whiffs…just so to see if the smell of her is still on your clothes.   It is, and you smile again.  Tired, weary of the world….but walked out a while.   Had a break.  '),(22427,664485,'22,February,2003','2003-02-22 00:00:00','   waiting for another outlook..dressed again..somehow i\'ve never put together this \'look\' before and it gives me a very \'la-la\' feeling...i\'ve mixed some knits and cotton with a pair of pin-striped slacks..my hair\'s tied up in that messy attempting to be sexy look and i have these silver hoop earrings on. i can\'t decide between heels or mules and the make up says casual, gothic and bitchy at the same time.  perhaps this is what it\'s all about...reinventing yourself...being able to put on a \'facade\' with whatever you have in your closet.  new major said the other day of the outfit, the one that screamed \'look up, please\'...it was a just over the knee length dress, with this tying strings..choker like crap..and i\'d put on a white one shoulder shirt on. it was \'artistic\' but was it to call attention...be different? or was it just cause i wanted to feel \'pretty\' that day?  i walked into \'gap\'  yesterday and realized what a fucking shopaholic i was. everything i picked up, i\'d seen before and the boys knew me by first name basis. i stepped out depressed and to a certain point, almost ready to cry. frustrated with myself.   i have such little control over my finances and it\'s pathetic how unthrifty i am. my concept of money should be flushed down the drain...i want to blame my upbringing but i realize this discipline shit and being responsible can only be up to me. my parents can only live so long and no matter how many millions, there are crude people out there and within me. irresponsible and just shitty with it, i\'m tired of being the spoiled fucked up brat you know?  i don\'t know...it\'s a saturday and shit like this should not be psycho-analyzed.  the girl just arrived...we\'re staying at the beach house this weekend. just to \'get away from it all\' or something. on a tornado watch...that\'s smart..  too many thoughts but i\'ll write it down later..gotta pack.  '),(22428,664485,'21,February,2003','2003-02-21 00:00:00','   dressed, sleepy, sorta lazily lounging around, cuddling and not doing much...  she\'s getting ready and plays around with layers and faces. grins at me through the mirror as i watch her in awe...mesmerized. god, she is beautiful.  i flip channels...and for a while, i almost fall into the category of \'the husband\'.  flipping, waiting for her...watching the news...singing to myself.  more talk about war, more talk about rhode island. i turn away to hide the tears but then the glint from hers reflects against the mirror too.   it\'s silly, two girls dressed and made up crying over a news report..  i almost want to say lets not go tonight...lets just stay in and we\'ll make pasta and run over to blockbuster and maybe rent a movie or something?  but i know i need to get out...or at least she makes me anyway. she does her little jig, and i cannot help but fall into fit of giggles.   she\'s almost ready now...i get to drive tonight. (i\'m in control!) she\'s tipsy, little alco...but i log in to say i need to do  something first;  it\'s pouring outside. i\'ve never liked thunderstorms and this whole fire/club thing just hits too close to home.   be safe tonight okay? wherever you are...take care of yourselves.  it\'s hard to comprehend...the site is on \'hiatus\' right now...part of me likes it this way yet the other itching to get it out again. get out of my corner and preach. drone you with long, long irritating whining posts  i love that a part of me is in this world...it always makes me feel so warm inside to \'come home\'.   is this for attention? i don\'t know. but i got it anyway...perhaps i do like it. perhaps i do like going away every once in a while to be alone and find myself. discover the world around.   take care of yourselves people, it\'s a cold, vicious world out there.  be safe.  i love you.  '),(22429,664485,'20,February,2003','2003-02-20 00:00:00','   you know, there\'s nothing like going to the medicine cabinet and finding it empty.  time to go play hide and seek...  or lock and keep?  whatever really.  i\'ve decided i\'m going out tonight....i have five places to go to...two in which i don\'t really care for but have to.  i want to go take a long bath and start getting ready...but there is this imminent fear in having to look myself in the mirror and being afraid i might not be so okay later..  taking a bath normally also means a lot of things..sure, i figure i could just fuck myself and give myself a good time in steaming hot bubbly candle atmosphere...but why the fuck do people have to get naked to take a bath?  i wish i just didn\'t need to see myself, and just ignore. not feel so overweight and ugly and disgust myself.  why. why. why.  the odd thing is that i\'m sleeping over at the girlfriend\'s house tonight. and she has all these things planned for us..i just feel so numb and sorta blah right now.  i\'m not depressed though, just unfeeling.  my choice of music isn\'t exactly helping either..come on, melloncollie and the infinite sadness?...dumbass.  i had a really  beautiful  fucktastic conversation with the old guy tonight though...kissing and making up is fun. painful but really nice..  and for some really odd reason, the new major and i ended up having lunch then went shopping together. very interesting. nice though..i enjoyed it anyway..  hm..time to make voodoo cutouts of the ex...  bitter? me?  hell no.  '),(22430,664485,'20,February,2003','2003-02-20 00:00:00','   one of the drugs they were thinking about putting me on is called effexor...it\'s an anti-anxiety and anti-depressant..i just took a look at some of the side effects and already all my doubts and fears are starting to resurface again..  too much. too much.  what if it doesn\'t make me feel better? what if it makes me feel worst?  scary...i don\'t know. just strange.  \'may cause aneroxia\'.  excuse me, but isn\'t that like giving a knife to the suicidal?  though really, that might be it\'s strongest point right now...  must.  lose.  weight.  i had this meal today...and i feel so icky. indigestive. like my chest hurts and it\'s hard to breathe. i half don\'t want to get sick, half need to get sick.   \"oh what a fucking dillema.\"   holy shit you\'re self-absorbed sometimes racheal.  i think i might go worry about the war now.  (insert too many complex morbid depressing thoughts)  or   not.  '),(22431,664485,'20,February,2003','2003-02-20 00:00:00','   raindrops splash on suede..200 dollar coat ruined.  layered, black against white, against black with splashes of silver. some rings, hoops, chains.  tied in at six in the morning, two hours of sleep and caffiene running.  thoughts clutter and consume  brain freeze - i wish.  '),(22432,664485,'20,February,2003','2003-02-20 00:00:00','   in class now...it\'s odd how \'school\' has become my escape again instead of wallowing away in self-pity.  they talk about different medication, i fight, but after a while...i just say fuck it.  whip me, slap me, call me edna.  i\'ll do whatever it tkes now..  i almost don\'t even care..  it\'s as if there is this \'don\'t care/don\'t feel/don\'t want\' thing that envelops you and you fall like sheep  just follow  tired of conflict  of thoughts  of opinions  ...so you just say  \'yes\'.  \'okay\'  \'whatever\'  '),(22433,664485,'20,February,2003','2003-02-20 00:00:00','   i bite back tears, bitter angry hurting tears..the big round drops that squeeze out and the choking horridly painful sobs i hate...i hate crying for you..about you..to you.  but here i am again, as always, with my heart ripped out into shreds and my soul bleeding...left to your transgressions, your verbal abuse, your meaningless/meaningful attacks.  i want to say remember when...but i\'m afraid, the past stays in the past. to realize what we\'ve become will hurt too much...but i don\'t forget, and that\'s why it gets bad.  i won\'t lie, this attention is great. sure. it\'s uplifting to find hope at the end of a torrid storm, it\'s fantastic to find the rainbow.  but you forget about this disease, you forget about how it consumes you and eats you alive  you forget about what little control you have over it and how extremely horrible it is  you forget what it\'s like to be the little person trying  or the weak one giving up  you forget the insanous crave for escape  she is sick. sicker than you\'ll ever know.  the hope is gone.  her brain fried.  her livers damaged.  her body failing on her.  her mind weeping.  but do you know?  do you care?  hell no.  you\'re too damn bent on taking her down.  too damn bent on proving your point;  that she wasn\'t worth love anyway.  well guess what mister  she waits.  and waits.  cries.  begs.  she looks silently at the phone again  and draws it back in  cuts herself  you\'ll never know will you?  you\'ll never see.  too blinded again.  she searches for answers.  and they put her \'on watch\' again  she screams and says...what the fuck difference would it make?  i hold my fate in my hands goddamnit.  i decide when i die  her demons devour her alive  she begs for you to come save her  see her  but you\'ll never know  never know  because she puts the strong face for you  the bitchy slutty one  truth be told  it\'s just a lie.  she pretends for you  so you don\'t hurt  so you don\'t see her detoriate.  so you don\'t see her die  watch her take herself  feel her heart slow down against your palm.  you laugh, and you laugh again.  i finally pull the trigger.  you won\'t even come to the funeral would you?  '),(22434,664485,'20,February,2003','2003-02-20 00:00:00','   they all talk.. she tries to put on her \'i don\'t care really\' mask but inside, a tiny tiny voice screams; i care. i give a fuck.  she wishes morbid occurences, pain that he\'ll feel her loss that when she\'s gone he\'ll realize why, and how he always did know her best she said, he said..they all said; they understood each other so perfectly it was torturous. and so they both wait waiting to see who will end up pulling the trigger first waiting to see who will weep at who\'s tombstone  death pounds them imminently now how ironic death started this  and death will end it.  if only you knew.  '),(22435,664485,'20,February,2003','2003-02-20 00:00:00','   just walked in from the cast meet, started another diet and i don\'t think my abused lining is quite liking it...ouch.  parents freaking out..the new maid ran away this morning..eloped apparently..  very very long, semi interesting, tiring, exciting day.  time to do homework.  '),(22436,664485,'20,February,2003','2003-02-20 00:00:00','   - writing when no one is looking is very liberating.  - trying to look for an outfit for tonight is not.  - having the distinct feeling that your other mom and birth mom are about to claw each other out is not a good feeling.  - having a distinct feeling about clawing your other dad on the other hand though, is very satisfying right now.  - being stifled and being treated like shit is a horrible feeling.  - walking around without underwear on, being a bitch, and wearing only black while using the movie as an excuse is a good excuse. (...because i get to be somebody else or because it\'s an acceptable excuse(?) i don\'t know.)  - being excited to meet the rest of the cast tonight is pleasurizing.  - actually, just being excited about anything is pleasurizing.  - using the word pleasurizing seems very wrong.  - feeling like i have the reins back in my hand is calming. very special..  - talking a lot and overblogging obsessively to escape is even better.  - reading good email makes me fuzzy all over.  - eating real food makes me kinda feel like god. i\'m nourished.  - not throwing up is even better.  - i have purple-green bumps on my thigh.   - time for some concealer.  - and a nice, cold, stainless steel spoon.  - perhaps even some lemon.  - instant messaging is evil.  - jennifer love hewitt is fucking gorgeous.  - i need to get ready...put on some layers.  '),(22437,664485,'19,February,2003','2003-02-19 00:00:00','   Perhaps there are a lot of things I’d wish to say…but words often betray me and I am left in a standstill of trying to decide between what\'s right and wrong, good and evil. Paths that I don’t necessarily want to take, but know I have to.  This getting better shit frustrates me, to say it is an uphill battle is an understatement. I understand the beauty and immense joy I will feel at the peak, there are just too many times thought where when slipping, it’d seem so much easier to give up. And given up I have, plenty and plenty of times.  This journey that goes through climaxes takes me through never ending landscapes, vast continents and faceless demons that I encounter. Angels that hold my hand and lead me along the way, I stumble and scrape myself, find pleasure in poking at my bruises; half to see if they’re real, the other half just to remember what physical pain feels like…and how it differs.  I tell myself to be grateful, and for many times, I am.  I have learnt that Tylenol is not strong enough to kill me, and Pensacola doesn’t really have many high-rise buildings, that the drivers here have too much damn etiquette and forget to just run me over when I step onto the street and honk instead. That Gillette razor blades are really quite hard to hold especially when ‘the pain in your heart isn’t as strong as the pain in my mind.’ I’ve come to realize that my asshole insurance company does not cover suicide attempts and so if I’m going to do it, to do it right.  But see I’ve also learnt that no matter how sad I am, no matter how broken, hopeless and ‘out’ I feel, that if I just press on….tomorrow will be a better day. Or the day after. Or the next day into it. I realize that I can make my super hard, super heart psychologist cry when I’m being honest and it’s okay. I’ve learnt that when I actually go for therapy and agree, things get so much easier. I get so much easier.  I am not perfect. I never will be. I’ve never come close to and I never will. And I want to say that if I just take down the bars a little I will be so damn much happier…but part of me knows I’ll never forgive myself for that; for lowering the standards, for reaching anything less than perfect…and so that internal conflict battles through and sometimes I am discouraged. Sometimes I am tired and lost. Very confused and hurting. Perhaps that it my biggest problem then…trying to be way more than I can be, trying to carry way more than I can handle. Trying to be Superwoman, Sheera and Storm at the same time.  Fuck me.  You know that feeling of tranquility that sort of envelopes you after a good cry? I sit here sniffling, sorta sleepy but extremely peaceful…and it just feels so good. So right. So something you could snuggle up to and sleep. and sleep for me is hard to define…but I could just let go, lie down, and sleep now.  It’s a good feeling.  Fixing…being able to know I’m being fixed is nice. Being able to know I’m fixable is nice.  I reached out and pushed back the other night.  The most important lesson I’ve learned is that I do this not only for myself but for the one’s around me.  …I started to list down the things people have said to me in the last few days…but it got long, too sappy and retarded, unless the other person and I read it. They were nice things though, good, solid, heartwarming things…and I am grateful.  I don’t want to return in a coffin, and my liver/stomach can only take so many sessions before it won’t even be my conscious decision anymore. I’m not sure if I regret it, but it was definitely stupid.   I want to say I’m sorry, thank you, and I’ll be okay at the same time.  The best thing would be that I’ll mean all of it.  Pretty cool huh?  I’m okay.  '),(22438,664485,'19,February,2003','2003-02-19 00:00:00','   is it good or bad...to tell you i love you?  to tell you the situations and ask for your opinions?  to get out of my own world and venture into others, maybe even yours, and see what the world looks like from the other side?  is it safe to hold my own very heart and place it next to yours, in your kind and gentle hands?  is is safe to trust?  i wonder if, when i sleep at night, the kids i used to play with in hospitals look down from above and smile...  or do they still make scrunchy faces when having boo-boos because i forget to keep my promises of going on or staying strong..  i wonder if heaven looks as neat, and as cool as i described it for them..  i wonder if i really believed that or as always, if i was just bullshitting my ass. like smiling so the world thinks you\'re okay and get off your back..  i wonder if talking is right and mentioning is okay.  if saying out loud and fixing problems instead of hiding is right  because i\'m still afraid.  it still hurts.  this cut from where i\'ve fucked myself over   and i wonder if it\'s safe to come out and play now  i wonder if it\'s okay  i wonder if you mind  i wonder if you\'re still angry, still upset, still hurt  because i didn\'t mean to  and if i could i\'d take it all back  and if i could i\'d let you hold my razor and slice me all the ways you were hurt  i\'d let you make big blown up posters of mine as targets and even volunteer to be made fun of in your weekly column  because truth be told (heaven forbid)  i do care about you  and i deny it, but you mean so much to me  you are important in my life..and you touch so many others  i want to say please don\'t leave me  i want to say tell dopa and sara to shut up or racheal will come and kick their asses...because fear does that you know, fear makes you do crazy things  and so i want you to promise me you won\'t go  i want you to be there when i make it  when i grow up  i want you for all my selfish reasons because you mean /that/ much  don\'t do anything stupid.  and i won\'t.  and we both snort at that statement...  because it isn\'t about going away really, it\'s about escaping.  i\'m trying to say all the things that i need to say  but none of it, i know, will really actually convince me when i\'ve made up my mind  so here\'s my heart  my soul  i love you  sorry it\'s all imbalanced screwy now  you are an amazing, beautiful, intelligent, wonderful, caring human being.  and you make this world such a nicer place to live in.  you\'re granpa, and dad, and surrogate daddy, and uncle, and all the things a title can bring you.  but most of all you\'re /you/  and we all need the old guy around.  to laugh  smile  giggle  play  drive  eat  fight  condescend  be sarcastic to  pose  poke  irritate  agonize  talk  listen  watch  learn  understand.  just wondering..  have a good day okay?  thank you.  '),(22439,664485,'19,February,2003','2003-02-19 00:00:00','   She smiles…finally.  Two weeks over, her cloud has passed. Restored, she opens her gates.  She welcomes,  beckons…let sweltering rays implode her but stands tall and says  I will take this.  I will accept.  I will move forward.  '),(22440,664485,'19,February,2003','2003-02-19 00:00:00','   i\'m in ecstacy...  i just got off the phone with my mom...she\'s flying in and we\'re going to work on looking for a new place for me to stay in, fix my medication and get me sane again.  i am in complete, utter, disbelief.  it is amazing.  wow.  :)  '),(22441,664485,'19,February,2003','2003-02-19 00:00:00','   mental slow down..things wrap up and to be continued. rest now, sleep tight.. tommorow...fight again? new day?  just breathe. just live. just make it through another night.  '),(22442,664485,'19,February,2003','2003-02-19 00:00:00','   everytime you leave, it hurts all over again.  i tell you so many times i don\'t care anymore...but behind this sheild, so many times i want to reach out to just hold you. to just touch your fingers or hear your heart beat against my ear.  when i cry, or hurt myself, some part in me always wishes you were near...but you never are anymore  i cannot find you.  and so days when your user name signs off, my heart hurts and i don\'t know why.  i want so bad to reach out beyond this hiding and let you see me, but i\'m more afraid of the consequences.  so i just watch you from afar, and keep you close to me.  i\'ll smile when you\'re happy, and weep when you\'re sad. but i know i cannot come out of these shadows, protecting you, selfishly, maybe even me, and maybe look down from up above...or down below.  i wish you could save me, but i know now to not wish on lofty dreams.  i have no dreams.  ...only love.  i genuinely, really loved you.  i genuinely will always love you.  i guess i went first huh?  perhaps you\'ll be at peace now.  or amused at least  '),(22443,664485,'19,February,2003','2003-02-19 00:00:00','   does it hurt because it\'s the truth or does it hurt just because?  this very pain that rips through me tastes almost distinctively familiar, nevermind that you were not a lover; in essence though, you came too near and broke too many walls. you tore right through and touched raw, bleeding flesh.  sometimes, it\'s when you pretend not to care, not to listen, not to give a fuck when you learn the most.  two points for me but you wanted to be a drama lecturer. like stan you said...and that whole thing about you\'ll be any age you\'ll be anyway so why limit yourself?  you see, i do listen. i do care...i\'d just rather pretend not to...if only because i figure caring is evil and attempt to shut that shit off but fail desperately anyway.  i worry when you\'ve had a sometimes bad day, when you\'re preparing for a role and it calls for being depressed.  you say i share your disease; in which case, then i think or even cannot begin to imagine what the ups and downs are like.  we talk about him. it\'s ironic how you tell me i just need to look for someone younger who can love me like you do?  love me like you do?  i\'ll smile because i\'m afraid it will hurt.  i had three choices yesterday. get help, don\'t get help, die.  i chose the last option but couldn\'t go through with it. i panicked, and in the heat of that moment, rang up amanda.   \'i\'ve been stupid...very very stupid,\'  and i broke down again.  i don\'t know how when falling, it seems to crash so much worst.  i close off but crave so much for someone to just understand.  i think about taylor and desiree, and perhaps even better times. i read again and i realize how painful it is to bear your heart and soul. it must be me then...the masochist in me flips between both sites and cuts with every line.  you both have a way with words.  love hurts.  '),(22444,664485,'19,February,2003','2003-02-19 00:00:00','   she can\'t do it.  it only angers her futher.  too weak to leave, too weak to stay.  she\'s confused.  she\'s alone.  she\'s fucked.  '),(22445,664485,'19,February,2003','2003-02-19 00:00:00','   i sit here listening to keyboards being tapped.   unfeeling, almost numb to a certain extent. afraid to cry in case i\'ll never stop.  fought with the therapist, she screamed emergency room and i put back all my walls and stood up. i laughed, even if it hurt, and smiled even if it was fake...but no one could tell you see, no one knew.  i talked about being locked up and at forty degrees with only the bare minimal three layers of clothes on, i said no. fighting with the other voice on the phone, i paced around, jumped up and down and danced as they watched, trying to look for signs. she took the pills goddamnit, where are they? how many? what the fuck did you do to yourself racheal?  flushed my system and i was good to go again. \'good to go\'.  but what they don\'t know is that i stopped by in between classes and walked straight in to purchase another bottle.  extra strength tylenol, 500g, 18 years old, no breakfast, 103 pounds.   lets call the poison control center again.  hie, my name is sarah. it wasn\'t intentional. i am fine. (i am scared)  do i want to die? do i want to die? do i want to die?  plans being formulated in my head, now that there are no more dreams...i plan my death.  tonight. (after church?)  later on. (after gayle?)  tommorow morning. (after the gym?)  but i have a cast meet tomorrow night. i need to be there, i have to be there. i still have so much i tell myself, i still have a lot more to live for. live for racheal i ask myself almost condescendingly?  hell bloody fuck no.  i\'m fucked. i\'m fucked.  someone fix me.  i write to no one.  the site\'s shut down.  no one reads this but me. some perverse pleasure of seeing pain typed out. keeping memories. tabs. this was when i tried to kill myself. this was when i had a nervous breakdown. this was when i fucked up.  but i\'m afraid.  do i want to die?  do i really want to die?  or do i jsut want to escape?  but i cannot afford to. i have school and i have the movie and i have all these damn bloody fuck things going on.  then why? why? why?  i\'m going to be quiet now.  i just need to go to sleep.  i just need to rest.  i just need to shut down a while and things will be okay later.  later.  later.  will later come now please?  '),(22446,664485,'18,February,2003','2003-02-18 00:00:00','   take a step forward from hiding beneath the shadows because i figure fight pain with pain. how much more can honesty hurt now that it\'s been unleashed.   you tell me not to use any references to me, but it\'s the only thing i know right now, selfish on my part i know, but just being me, living me, consumes all of me. because precisely like you said, if i let someone in, i fuck things up. you hurt me too marc, i won\'t attempt to measure that.   you\'re the only guy, real person, i\'ve really let down my guard to. the only person i\'ve managed to ring when i was afraid, the only person i\'ve told when i was hurting, the only person i\'ve let in to see me cry. if that is not trust, i don\'t know what is.   i don\'t want to beg, or plead. you have your rights, and i should walk away from this knowing that i have crossed a line. it wasn\'t somethng i did on purporse, and far nothing was meant to  that intention.    i actually care about you. hard to believe but i do, i\'m sorry you never saw any of that in me. i\'m sorry i have become so ugly, so tainted, so immature to you....and perhaps i am.   so many thoughts run through my head, i want to run up to you and shake you and ask you what in the god damn hell happened yet i am afraid to be in the same room as you are. i\'m afraid i\'ve overexposed and let you in and all this shit hurting feelings just isn\'t good right now.    i try to look for all thewords to say but my mind betrays me.    and so i lock up, freeze and fuck up again. i wonder if you feel as much pain, not because i want to curse it to you but just if i meant as much as you did to me.   do i regret anything? would i take back anything?   no. because as much as it hurt to turn the other way and let go of your hand, i treasured this little path i had with you.   i\'m sorry you hurt.   i\'m sorry it seemed like i never cared.   i\'m sorry i messed up   but i guess being sorry doesn\'t fix anything sometimes does it?    pain matures you. i guess i\'ve learned a lot.   i didn\'t think you were any of the things i said. it was just a careless thought the people outside thought, perhaps i\'d figure you knew where i stood.    apparently not.   sorry   I have a rose and half a heart-shaped box of Ferrero Rocher chocolates sitting here by my computer.  How could something so beautiful go so wrong?  Maybe it can\'t. I dunno.  Time heals.  I love you, Racheal Khong (said with an accent). You will never know how much.  I swear I do.  Marc  (PS. Get your lazy butt up and go to school.)  '),(22447,664485,'18,February,2003','2003-02-18 00:00:00','   so i finally did it.  i emptied all the bottle\'s content and swallowed. lying on the bathroom stall, i wreaked and threw up...sick and tired of it all; i reached into my bag and jangled. walked ever so calmly to the water fountain, betrayed all my idiosyncrasies and drank from it. two gulps, that\'s all it took to keep me numb. that\'s all it took to feel the pills slide down my throat....straightened myself out then went to class.  i didn\'t panic, if anything, i was more peaceful than before. was this how i would stand before death\'s door? calm and accepting, self-depricating.  i felt my heart slow down, my feet get heavier, my eyes harder and harder to keep open. the tears came in a flash, in a moment, i begged for help...but when help came i ran.  threw up again.  i tried to kill myself again today, this is not good.   i just want to go away.  '),(22448,664485,'18,February,2003','2003-02-18 00:00:00','   i forget that if you turn towards the red knob and sit in the water you\'ll scald your skin...and so i have angry red splotches on my body. i forget that if i don\'t eat my fingers start to shake and this light-headed feeling is not just an extension. i forget that i have to get dressed or get out of bed and i cannot tell if it\'s night or day because i haven\'t stepped out of these four walls. i alternate between hours of crying and passing out, break and smash my pills, pull my hair and cry again.  i burn myself forgetting that my cigarette is lit and walk out into the cold with only my knickers on to bring the mail in. i curl back into the fetus position and rock myself to sleep and sleep again. i sleep so much i am numb and when i finally am i pick up phone calls, open gifts and reply emails. i answer the phone and turn in homework before the hour comes around and i crash again.  flesh against the cold reflection, i poke at another bruise and curse myself. i hate the girl in the mirror and this time i don\'t even bother to look for that bitch hope. i only stare back at this other bitch that looks at me and tear her whole being apart. i crave for touch and feelings but flinch when they find me. i reach out but my hand betrays me and pulls back.  i am alone. alone. alone. pain consumes me and i don\'t even know why. hurting, letting my disease eat me alive. feed the starving child.  afraid of food, afraid of praying to the porcelain gods. afraid of praying.  will you help me please?  don\'t let me kill myself.  don\'t let me hurt me.  don\'t let me be stupid.  please.  '),(22449,664485,'17,February,2003','2003-02-17 00:00:00','     '),(22450,664485,'17,February,2003','2003-02-17 00:00:00','   I probably  should  hold my peace but here are my few words instead.  Went to the head doctor today, it was interesting to say the very least.  An ‘emergency’ session they called it, help! the girl was losing it. Fucking up the mental and physical, oh screw that.  She wasn’t suicidal she kept on repeating, she was okay, damnit, just believe her she begged,  she was okay . Really.  But she still heard the jangling of pills in her head, still felt her heart pound out from her chest as adrenaline and anxiety flowed through her…still heard the thoughts in her head that said jump, stand in front of that car, fly, cut, swallow, sleep.  She wasn’t crazy. She wasn’t mental.  She was just Normal and Real. Just like the other girls, just another ordinary girl…  But she shook so hard last night she bled, so utterly lost and confused. Screaming, yelling and catatonic, she reached out into the abyss unknown, drew back shades and cried. Shrieked for help and so they brought her in. Tranquilized her as she became manic, panicked and nervous again.  As she broke down and was wretched in pain again they shot her.   Blood levels too low, no sugar, no substance they fed her through saline and she felt immense pleasure in the pain and absolute horror. They were feeding her, they were keeping her alive, they were making her stay.   Putting nutrients in her because she turned crazy with the sleep and food deprivation. It fucked up the chemicals in her brain and made her so depressed. So unable to function. Live.  She wouldn’t feed her demons so in turn, they tormented her.  Her heart went no, no, no but her mind begged please, please, please.  Please just get well Racheal, please just let them fix you, she cried with herself. But she couldn’t and it was so hard, she felt so wasted and worthless. So valueless and hurt. There was just too many people to deal with, too many problems to stomach in a night that she was already cracking.  But then the hundred dollar an hour therapist sits down with her and cajoles. For two hours they talk and she wavers in between crying, being silent, whispering, talking and screaming. She alternates between the raging emotions in her and her disappointment. In herself, in others, in the world around her. God and the being above, the people she lived with, the girl she slept with, the image in the mirror. They worked through so much and nothing at all, and finally, very slowly, they made progress again.  And so she lies down here watching the big hand cross the little hand and signal a quarter past the hour. She is sleepless again, but stable. Sane.  It feels so good to be sane. It feels so good to be numb.  Subdued and chick-flick, pill-popping, cigarette-smoking, happy.  She finally smiles again.  She runs.  Hides.  Tired of the world.    \"I prayed last night, I crouched in and held myself as the demons played around in my head and begged for you.\"  '),(22451,664485,'17,February,2003','2003-02-17 00:00:00','   and she smiled @ the lifeless image in the mirror. we had learned what we had seen. nothing remains.    \"my redeemer employed all of her arts, but what could a woman, master of limited number of souls and bodies, do, faced with my friend who was always changing- and always identical to herself in her incessant metamorphoses.\"  octavio paz my life with the wave  '),(22452,664485,'16,February,2003','2003-02-16 00:00:00','   I don’t tell anyone because I think it only scares them, or hurts them. It scares me too, it scares me to think about how absolutely little control I have on these reins when I am falling.  I tried to kill myself twice this week.  I don’t know how it came or why. Only that in those moments, I was looking to escape being Racheal a while. I’m not sure if I wanted to die, or if it was my way of saying I needed a break. I needed to Stop. Escape. Run. Hide.   The coward’s way is suicide. I am a coward. I am bad with this feelings shit and taking care of myself. Bad with handling things and loving, trusting, accepting myself.  I wish I could tell you that I was just being stupid and that I was okay now, the truth is, I don’t know.  I don’t know what happens to me when I get sad or abusive, I don’t know what happens to me when I get terrified of the world around me and when I am so afraid, angry and disappointed with myself. I cannot tell you what happens or why it happens. If it’s just the chemical imbalance in my brains or hormones or the lack of nutrients or if it’s just fucked up me.  I cannot tell you why I tear myself apart or have such dark morbid thoughts.   The truth is, I don’t know why I get so happy and sad in a split second. I don’t know why I crave being alone and hate loneliness in the same heartbeat. I don’t know why I have become such a child, a bitch, an asshole and a despicable whore.  I don’t know why I am not capable of loving or being loved. I wonder sometimes what is so wrong with me and why I cannot just make the face in the mirror disappear. Go away. Fuck off.  I am in so many pieces. Shards. Smithereens.  Tonight, I held a razor blade against my skin and finally felt the pain before it cut me. I didn’t want to kill myself really, I just wanted to see what it would be like. Some part of me was just crying out, reaching out. Some part of me just hurting.  The thing is though, I’ve been doing so fucking well. I’ve been more sane and happy in the last few weeks than I’ve ever been since I’ve had this problem. I’ve been more okay and confident with myself then anything else. I’ve stood up for myself this week and walked out the door.  Yet it feels like I’ve also slipped back to where I started from. It feels as if I’m fourteen again and running on empty and being two people at once. The public Racheal and the private Racheal. The similarities are definitely there, I’m sick again. Killing myself slowly, painfully, but is this pain truly worth it?  I have so many material things going for me, and everything is truly fine and dandy on the surface. The pressure mounts and I crack when nobody is looking, I fall and stumble and scrape myself. When the curtains are drawn, I peel off the masks and out emerges this terribly fucked up child. And out emerges Me.  I don’t know what is wrong with me. I don’t know how I ended up so lost and confused. I don’t know why I’m so lost and confused.  Sometimes it’s as if I’m looking from another dimension and I see all the things I already am, can be, and will be. I see hope in big bold print but when I reach out to touch it it’s as if it disintegrates before me and falls to the ground.  I hate burning from both ends and feeling as if I’m tugged from both sides.  I rang my parents up tonight and we talked. I sobbed so hard I couldn’t breathe and rocked, rocked, rocked, as the pills in the bottle made a jangling sound and soothed me. It felt as if I held my fate in my hands and it was all going to be okay. But would it be okay if I swallowed it or was it the other way around?  I reached out.  I am still reaching out.  I’m trying to be strong and say this is who I am sometimes and if someone would just say it was okay to be weak. If someone would just say it was okay.  But I hate it that I am writing this.  That I am showing this side of me that doesn’t have something smart to say or isn’t bitchy and laughing about something. This side of me that isn’t strong.  But I can\'t handle it and I am so disintegrated.  I am begging/pleading/praying for help.  But I don\'t know what for or what it\'s worth.  I\'m pathetic.  I really just want to be okay.     urlLink  To do the useful thing, to say the courageous thing, to contemplate the beautiful thing: that is enough for one man\'s life. - T.S. Eliot  '),(22453,664485,'16,February,2003','2003-02-16 00:00:00','   cheese.   something she said upset the photographer tonight.  she hurts.  perhaps this trust thing just feels so threatening now. she swears it isn\'t personal, just a mere folly of liking her brick walls put up. or letting the questions bombard her mind so she\'s \'safer\'.  she trusted him once, and now as she is dying alone she wonders where he is.  it hurts.  her fault.  her fault.  her fault.  ugly, dirty, undeserving child.  she wishes she could do something for the other person that walked her through hell..as his world falls apart she wishes she had superglue too.  wants to say \'it\'s going to be okay\' with the same assertiveness he used with her but knows she wouldn\'t believe it too.   his words strike chords again.  she\'s confused tonight.  time to retreat to her little hole.  time to run.  she\'s been here too long.  she wants to cut herself. wants to see blood spurt out of her veins to know \'pain\'.  crashing down tonight, she falls.  she walks.  she meets her maker.     ...he turns the other way.     urlLink \"if i\'m going to suffer i need to suffer alone. \"  '),(22454,664485,'15,February,2003','2003-02-15 00:00:00','   Maybe she remembered what it felt like to have the cold square steel in her hands...  or how liberating it felt..  after a long break of between three months, she picked up the camera again. She\'d forgotten what it felt like to see the world through these lenses and show the others what her world looked like.   She ran a bath for herself..shuffled in various dark melancholic albums and switched off the lights.   She craved being in the dark and hiding. Something about the room glowed, something about having it\'s only lightsource to be a collaboration of flickering flames soothed her.  She stripped, and as her clothes touched the floor, so did her teardrops fall.   Her guiltiness having thrown up her dinner ten minutes before gnawed at her as she furiously belittled herself and attempted to shut the voices. When all was done with the square case, she laid it on the side next to her glasses.  Blinded, naked and empty, she climbed in and read.  Cleaned.  Cleansed.    urlLink  \"It is more of the bittersweet feeling that I get. That intense feeling that doesn\'t fully show itself. It\'s like an equal mix of ecstacy and deep melancholy that leaves you balancing precariously somewhere in the middle, swaying.\"  '),(22455,664485,'15,February,2003','2003-02-15 00:00:00','   there are times where she looks so old, aged and to a certain extent haggard. tired and weary of the world as you spot the eye bags and the diminishing sparkle, the weary limbs and the breaking body.  then you look closer and closer or maybe just search harder and you realize she\'s just a little girl forced into womanhood. her eyes are still so naive, her body still growing. you watch her mind develop yet there is still that immature selfishness that kids hold on to and then you realize, she\'s really trying quite hard.  so you scale down the difficulties or help her figure them through and suddenly she realizes there\'s such a thing as hope and there\'s such a thing as living. trying.  she smiles again and runs. dances in her little red outfit and tumbles through the grass asking if she can have another scoop of ice-cream as she licks the melting chocolate whipped cream off her fingers and grins.  still a kid yet so much a woman.  if only she could figure it out..  '),(22456,664485,'15,February,2003','2003-02-15 00:00:00','   When you wake up with an ass throbbing head pounding ache, it’s about the same time you commit to that resolution of never getting wasted again.  And so last night, when they were at the  urlLink sexy boy’s  place (who still owes her Arbys and nude pictures), she very intelligently only had little to drink but much to get silly over.  Because she’s already plugged in one half of the duo, here’s  urlLink the other one  that had to keep Mr. Happy away from Mr. Peepee while being the asshole. Who plus also gave her the obligatory bonus hug at the end…Whoop. Whoop.  They stumbled and laughed over tits and obscene amounts of lard ass to be voluntarily chosen to be put into a film and sniffed ‘ginger’ and ‘peach-something-something’ shampoo as the dirty eye was given when remarked at how ‘airport-ty’ the place smelled. Nevermind that extra painful lengths with scented candles and a single solitary pink rose (yes, I noticed) were taken to unmask the  ‘We\'re just frat boys in a bachelor pad’  mask.  The chauvinist that didn’t stay long enough to watch the girl kick  urlLink fancypant’s  behind in Halo as the tetestrone level was being pumped out to balance out the whining bitching blockbuster customers in his head. He might have very well found his match though… :P  …as he knocks at five am to show the new code he’s deciphered.   Ah, and sweet bliss in giggling myself silly at the clink of every bottle on the table proceeds to naming that ‘Amy’s my favorite’ and various other more interesting rules that result with everybody getting more and more ‘happy’.  Minus the long-winged cockatiel and the missed silence of Yen-Yen the chinese-american Eskimo rubbing up against me, it was a good one.  But damnit, still no smiley-faced  waffles  pancakes..  Hmph.  :P     urlLink  \"Valentine\'s Day. VD. Venereal Disease.  no coincidence.\"  '),(22457,664485,'15,February,2003','2003-02-15 00:00:00','   “I guess you’re one of those girls that are used to people falling in love with you huh?”   Classic.  So fucking overused it isn’t even funny.  I play stupid so damn well I deserve an Oscar. Or a nomination at the very least..  My comfort space intruded in so rudely, backed into a corner with no damn choice or reasoning. Irritated and pissed, damnit, I have other better things to worry about.  Grr.  Upset for not a good reason at all, lets handle this like two perfectly mature adults as he tightens the strings on my leash and condescends me.  (Well, you’re (definitely) not the type of guy I fall in love with are you?)  What isn’t said beneath the surface is horrifically fiery. Hell hath no wrath like a woman, I remember that now. But I keep all my sarcastic remarks and painful anger to myself and wallow instead at this god forsaken hour and ask why me?  This isn’t some puppy love cute attraction shit, this is a whining, obsessing, compulsive, egoistical, head-up-my-ass, won’t leave me alone attraction.  Perhaps my last saving grace and worst thing is that it was a good friendship.  Why, goddamnit, why?  Being coerced and having to pry wandering forceful hands off me is not my idea of fun thank you very much.  I’m Pissed.    urlLink  \"And I\'m going to bed and I hope I\'ll die and rot and stink enough so the maggots will come crawling all over me and swarm through my bodily cavities and eat me up so that everything thats left of me is a stupid little skeleton white and pure just the way I\'m supposed to be.   Haha.\"  '),(22458,664485,'15,February,2003','2003-02-15 00:00:00','   Candle flames flicker as it’s glow resonates throughout the room, falls over encapsulating it into an almost sacred territory. Spiritual?  Beer bottle in hand and late nineties rock music he once introduced to her fill up the picture. She lies with chest to the ground and rests, her dress sits crumpled at the end of the room, her shoes hastily taken off the moment she returned back to her sanctuary. Barefoot and tired, she poeticizes to the demons of her night. Charming, as if dangling a carrot to the terror nights before and two-hour sleep days. She runs on nothing, not food, not rest and definitely not sense…but she loves it. She rises, stumbles awhile but feels victory at the end. Tastes it on her lips.  She is not intoxicated as per say, she still can walk a(somewhat) straight line and countdown from a hundred, but she is definitely Numb. Unfeeling and blind almost, blind to hurt and feelings but she is still able to understand. To define. To write.  Roses sit slowly dying on her desktop; she says they are to dry but secretly probably has a perverse pleasure in seeing it wilt away. Of letting go of materialism and all these bullshit notions of courting rituals.  She comes home at around two, considerably pretty damn early for a Friday night. Exceptionally especially for her. Fridays are nights filled with excitement and anticipation. Of swing dancing, clubbing, karaoke-ing, driving around aimlessly and sharing…laughing. Friday nights most importantly though, represent Freedom. When there are no early classes the next morning, no homework that needs to be handed up in three hours but most of all, Friday nights represent this inexplicable need for Habits. Rituals. Practices. Friday nights almost by law must be good ones.  But last night she chose to go home early. They had broken her ritual and gone to a pretty little restaurant, ate and slow danced instead, driving to beach and downtown and to various other places in a synergy. It was a night of celebrating Love and everywhere she went there was red. She liberated in it and the catcalls and wolf whistles of the party that once made her giggly now only made her anxious.  Consistent begging only irritated her more, she had had a good day but was ready to curl in and rest her thoughts. They flew through faster than she could handle, the masochist in her spoiled the party and said, home. Home. Home.  Not even waffle house at four am, or taking up the offer of the party that was supposed to end the night of all nights. Too tired she claimed, but everybody knew why, she’d been in the real world for too long and she was sick of it.  Well, it is Saturday now and my aching bones feel smashed, pulped and wringed. I must be growing old…or maybe it is the lack of nutrients and rest against the other piling problems. Yet yesterday was an extremely long one too…oddly enough, there are still men out there.  They just make me nervous that’s all.  They clamor and intrude into my bubble and I shy away or hold my ground and drive them to the door. It is hard to have straight thoughts on a day that represents bullshit beginnings like this…  I wake up to the shrill sounds of the phone ringing…it’s a quarter past eight and I’ve just now fallen asleep. I want to yell and scream but think instead about the effort he’s making. My head throbs and my body is sore, my mind frozen. I go through motions as he tries to pry out details from last night ; that jealous bitch. :P  In retrospect, I had a pretty interesting day…or had my ego boosted at least.  I sat at the dock, feet dangling and mind slightly not there and watched the seagulls play and the mullets leap as the first break of dawn emerged. I read and wrote, then thought to myself how amazing God’s creation was as the first rays of light cascaded me.  I  walked  stumbled back in, a sunny beautiful gorgeous morning, and slept.    urlLink  “it was a \"st. valentine can fuck off\" celebration.”  '),(22459,664485,'14,February,2003','2003-02-14 00:00:00','   BORED. BORED. BORED. BORED. BORED. BORED. BORED. BORED. BORED. BORED. BORED. BORED.   waiting to be picked up...fa la la la la la la. i\'ve had enough of this institution for the day. school = blech.  i have to get ready damnit. hurry up!  eeyaahhh, eeeyyaahhh, baby, andale andale...  still not working?  sigh.  i need sleep.  rest.  being tired makes me silly.  and i AM silly.  which makes me double silly.  one little, two little, three little...  home.  now.  going.  out.  four   hours.  helppp!!!      urlLink \"It just means that I’ve got to work harder, got to eat, sleep, breathe beautiful, and that I\'ve got to hope to dear God that Mariah Carey wasn’t bullshitting when she said that we could, indeed, make it through the rain. \"  '),(22460,664485,'13,February,2003','2003-02-13 00:00:00','   I just don\'t post it that\'s all! :P  February 11, 2003 - 5:31am   Yes, she is capable of love she says. Hidden far beneath this dark, deep, complicated exterior lies innocence still untapped. She is jaded, skeptical of the ones that are around her. Is it just sex that you want? Is it just Lust that makes you stay and not Love? She wonders sometimes…  Laying around and listening to Fountains of Wayne play obsessively at this hour intrigues her. Reminds her about the beauty he once forced on her, the way his fingers would trail her flesh and rub against her bones. She’d cower and scream, running to switch off the lights before he’d catch her again and tickle her endlessly…oh, how their laughing and giggling would ring through out the house and resonate through the walls.  He’d call her sexy, and beautiful and tell her she was intelligent. They’d make dinner together and go through mundane everyday motions but those were the best times she remembered. She didn’t care much about the expensive diners and holidays but she loved the warm fireplace and falling asleep next to him.   She found out about love with him. Long-distance love.  He rings her to see how the day is going. The site bookmarked and thumbed through obsessively, she feels like a thief in the night, catching him as he plays the same game. He knows she knows, and she knows he knows. But they keep silent anyway…still playing. Still restrained.  It’s been three years since they’ve ended but he still keeps his promises. His integrity in being there for her never ceases to amaze her. He envelopes her with love so deep that she finally understands the true meaning of love forever no matter.   How naïve they once were…so innocent. So free.   She keeps with her the lessons he has taught her. He’s earned a chapter or two in her Book of Life and carefully, once in a while, she reads through it. She escapes back into memories with the yellowed pages and lost snapshots. The candid moments and his face after a long flight home.   They’ve gone their own ways and are with different people, but she knows she’ll always have his love.  And so his, hers.    (happy now? :P)     urlLink  \"Roses are red Violets are blue If you duct-tape me Then I\'ll duct-tape you\"  '),(22461,664485,'13,February,2003','2003-02-13 00:00:00','   1. Explain why you started to journal/blog.  inner me wouldn\'t accept outer me and outer me hated inner me. i seem so god damn different on the outside, so damn happy, hyper, ecstatic, \"perfect\" and it always feels as if it is just a layer. a show. i couldn\'t find me, and couldn\'t love me and some how, amazingly, this has helped a lot. i guess it\'s always nice to come to some place and say, you know what? i\'ve had a bad day and fuck that and fuck me and i\'ve hurt myself, and i\'m hurting, and i\'m in so many pieces with so many thoughts and no \'safe\' place.   i shudder to think about how much i\'ve changed yet i relish in that. me the bullimic, sick, fucked up child can finally challenge the other person. the one on the outside that\'s sucessful and strong and bitchy and \"confident\" and all these other programmed crap i use as a defence mechanism.  the crazy thing is that, it feels really good.    i guess i also wanted to bitch, bitch, bitch a lot.  a LOT.  :P   2. Do people you interact with day to day or family members know about your journal/blog? Why or why not?  mmhmm.   becaussseeeeee!  pah, well, for a long time i wrote under a pseudonym. it was nice and beautiful and special and all that crap but i was doing exactly all the things i did not want to do.  i did not want to hide.  i did not want to run.  i did not want to have all these different parts of me.  i did not want me to be separated into different categories and hell, i felt schizopherenic. (i also can\'t spell if you can\'t tell already...)  so..  i took little steps.  and more steps.  and soon i learned that if someone could not accept me when i was deep shit ugly and hating myself then was it really worth it at all?   i put so much damn pressure on myself not realizing that it\'s okay to be silly, and incredibly insecure sometimes.  besides, sometimes, it\'s easier written out than said.  this is my legacy, whenever...i just wanted someone, anyone, to have a part of me or at least know me on some level before i die you know? (holy shit i said the d word.)  even having said that though, i still get piss ass jittery and berate myself for the rubbish i put up in here when i know the people that are reading, are reading.   oh well.    3. Do you have a theme for your journal/blog?  home.     4. What direction would you like to have your journal/blog go in over the next year?  i honestly, have no idea.   if you realize, i go into moods where i just need to be alone or if i feel as if there are too many people watching, or if i\'ve opened up too much and so i go on hiatus from time to time. i think about going home this summer and how that\'s going to turn out...or with my health and shit like that. i think about how that\'s going to affect me and all the words i might say in here. i just never want this place to not be safe anymore i guess.  there are no direct specific paths, perhaps that is the beauty of keeping this thing...?    5. Pimp five of your favorite journals/blogs.  why? why? why? do you do this to me?!!?!?  go  urlLink HERE . they are all shittingly good, and ferociously loved.  so hah.     urlLink  \"just some opinions, for what they\'re worth. just pretend i\'m a loudmouth in a bar who takes his opinions too seriously.\"  '),(22462,664485,'13,February,2003','2003-02-13 00:00:00','   She smiles. Wakes up into a lifetime of memories relived in 60 seconds, slept in an hour, dozed off in 90 minutes. She looks around and slips into another brand new outfit, the crisp sound of plastic being torn off as loud eighties music bombards the early morning air. It is beautiful outside; seventy fucking degrees! Silly pink and red paper strewn all over the room from an impulsive visit to the late night dosage of Mass, Wally\'s world. Colored tissue thrown around with streamers to mark this happy unpleasant day..she giggles to herself as she catches a glimpse of the outfit for tonight.  Black stilleto strappy heels, flowing wishy layered black dress fitted just right...something in her size(!) and black beaded choker with diamonds that could knock your ass out. Special, she feels good.   She throws up.  But this time it\'s just from the medication and she is well. No, I mean  Well.   She skips off to school, too excited for breakfast and finds herself at class boinking away at the keyboard recalling tales of the night. Homework and charts all nicely and exquisitely mapped out..math equations solved and papers written.  Teacher bitching(lecturing) now. Sneaks off and visits the rest of the world, sick of her voice and her whiny pathetic crap but laughs anyway.  (Excited!!)  :P  Happy fucking Valentine\'s Day!!!  '),(22463,664485,'13,February,2003','2003-02-13 00:00:00','   Home.  but more importantly, Safe and in one piece.  voice hoarse, pot-smelling hair, tired aching feet, smiles plastered, booty shaken, hockey game high, hands crossed and marked underaged, ears ringing, fifty two men and mr. happy humping my ass, karoake sung out, ego restored.  naked. bare skin on carpet, arms propped as fingers type.  no.no.no.no.no.no.no.no.no.  pretty and ugly. silent and screaming. smiling and crying. dancing and limping.  she walks, runs, wanders around.  heads for the next destination and drives right through...eighty seven hundred miles an hour she flashes by and strips. she peels off layers and struggles to apply on more, she learns the ultimate lesson of confidence and breaks it out to the boys. dances with the girl. talks to the strangers. answers to no one.  only the voices in her head and the pounds of her heart. they clash and to one\'s own, battle to death.  she inhales, breathes in a breath of fresh air and walks out again.  ready, unsure, dellusional.   insane.  but finally right.  no.no.no.no.no.no.no.  why.why.why.why.why?  lets go dancing again.  lets cross our feet and chant again.  lets sing, and cheer, and laugh.  winks. pouts. snorts. huffs. sticks out tongue and wags.  she grabs the key and heads home. says her goodbyes as eyebrows are raised and hesitant shy thoughts are exchanged.  smug and intimidating.  she thrives on being a bitch.  oh wait, look up bitch and you\'ll see a picture of her touching herself.  smirking.  fucking slut but at least she is \"stronger\" that way.  defensive.  never falling again.  thoughts so fucking scattered at half past two.  time to do some homework. time to do some real work.  time to be superwoman again.     urlLink  \"Life [one year / month / week / day / hour / minute at a time] seeps through her fingertips as she wastes yet another day with another Frappuchino, yet another complaint, yet another thought of CHANGe. It\'s hard to  reflect  speculate upon the  past future \"  '),(22464,664485,'13,February,2003','2003-02-13 00:00:00','   oh,  hello world.  '),(22465,664485,'13,February,2003','2003-02-13 00:00:00','   If I could pull out just one moment out of the time frame we’ve shared; it would be that one night at Waffle House with our sorry intoxicated drag happy asses looking for quarters to play the ‘Yo Te Voy Amar’ song. It would be that time when we kicked the jukebox as it replayed obsessively ‘Sweet Home Alabama’ and being our ditzy selves, we danced instead. Nevermind that we were in full stage make-up glory at three thirty in the morning with about a quarter of our cast singing and grills going on. Nevermind that it was see-through windows and harsh fluorescent lights that frankly, did nothing to our complexions, with half eaten smiley faced whipped out waffles and too many tears cried.  When the eight of us group hugged and wiped each other’s tears off and made each other smile.  It was that time when we ran through the mall laughing so hard we had to stop to breathe and try and stand still from the convulsion of giggles we got just by looking at each other.  …or playing lets see who can fly at the steps in front of the theater and messing up each other’s dressing rooms. Playing cards in between scenes with obscene pictures on them and finding every moment we could to take a sneak at the audience’s reaction.  Fun times I tell ya, fun times.  This show, as the spring production comes around I sit here remembering all the times we’ve shared and instantly, a smile lights up my face as your name flashes on my caller id and the queen of drama commands me to get off my sorry butt and go galavanting.  A hastily scribbled note found among stacks and stacks of notepaper filled with doodles and sayings.   “Live only in the moment: don’t look back to the regrets of the past, only look to the possibilities of the future,”   I remember now. I forgot.  I get ready and take a nice long piercingly warm shower and tickle myself with forgotten and newly made memories. I hum to Ben Harper and say fuck yeah at the moments. I scrub off the grime and slather on lotion onto moist skin renewed. I pick up my outfit and twirl around in front of the mirror and I am happy with the person that dances in front of me.  I leave my face bare except some gloss and perfume and click off the lights as excitement and anticipation for the rest of the night fills me.  Another adventure in the life of the Bobsy twins as I whip out my camera and you give me reason enough to take pictures again. If only so we can look back at this in print and laugh at the boys or how damn good we look tonight. :)  I never thanked you for the rides or the good cries. For allowing me to be your friend when you were hurting and yours mine. I laugh at how your face says everything as you naïvely, unsure, step out of the dressing room and we gush at each other’s outfits. Or the way you trust me enough to say fuck that and the way I trust you enough to tell you I was bad today.  I remember when you walked up to him and quite closely almost slapped him for touching me, or gave him the dirty look as you told him what your opinions were. I remember how you snatch away my phone when I am feeling vulnerable and you appear at my house when I tell you I’m breaking.  When I joke about the songs you have to sing at my funeral and kid around about what I’d like to be inserted in. I remember the first day we became friends and how far we’ve come now…  I remember shit like that and think now hell yeah I’ll be just fine alone.  If only because you believe in me enough to tell me I can do it and I trust you. That Rome wasn’t built in a day and sordid love is not the only type of relationships I’ll have.  Your perky little damn ass runs in and says Lets go! Lets go! as you stub out my cigarette and drive over to terrorize the ‘photographer’  You know, perhaps double dates just aren’t too bad at all.  :P     urlLink  \"I will however, wear underwear on my head, jump around all spastic and yell \"gleeb eep norp bop doo waaaah earp snob\" at the cokcateil. =) \"  '),(22466,664485,'12,February,2003','2003-02-12 00:00:00','   5:30 am -  bitched out and exhausted beyond reasoning...zzz.   6:05 am -  wake up to the shrilling sounds of my alarm clock having orgasms.   6:10 am-  stumble into the shower only to realize that i don\'t have the maids to restock my shampoo here and i was supposed to do it about a quarter of the bottle ago. blah. salvage around for the mini -travel pack shampoo that comes with super tangling strength.   6:45 am-  fuck around with something to wear. might i remind you that should you need to know how the weather is going to be like for the day...just dress the opposite of what i have on. really.   7:00 am-  oh hell yeah this day is shaping out nice...take my breakfast, shots and colorful array of pills only to realize that i have to run to the bank and i\'m already late! shit!   7:05 am-  of course the atm is down. hell forbid you give racheal money! pah.   7:06 am-  head for school   7:25 am-  still heading for school   7:45 am-  have my first cigarette in what feels like a hundred days. proofs my theory on how things always seemed better in my head.   8:00 am-  english comp ii class with professor hagedorn, a pot smoker, six perverts, and forty-two pms-ing women debating faulkner. funn.   8:15am-   puke session number one. bye-bye breakfast. hello anxiety/perfectionist attack.   8:50 am-  crammed in four weeks of mythology notes in ten minutes.    9:00 am-  dr. schuler walks in with a disney tie, green pants and a mean ass test in hand.   9:50 am-  nine brief tales of lovers conquered. how fucking damn appropriate.   10:00 am-  salvage a couple of dimes and some change for coffee. ahhh...caffiene. :)   10:05 am-  much much happier and sane. whoop. whoop.   10:06 am-  late for my mass comm class but i\'m never the only one so hah.    10:15 am-  get stuck in a heated ass debate about iraq and saddam hussein. argh. argh. argh.   10:50 am-  walk out with blue haired guy and english boy stuck in another debate of another conspiracy. once again, ooh funnn.   11:00 am-  lug around heavy ass algebra textbook and plop myself next to the very, extremely, perky girls. decipher solutions and mess with each others calculators as the back of my hair stands while a discussion of the different shades of pink ensues. like oh my god! like really? like get out! like can you believe it?! nuh-uh. uh-huh. nuh-uh.   to which this point i start laughing uncontrollably out of fear that i might die.   12:00 pm-  finish class. run across to the other side of campus. return seven hundred and forty two messages to people that don\'t understand the meaning of class and hunt marc\'s ass down.   12:08 pm-  yell, scream, freak out, hug, kiss, exasperate marc the assistant director and work on monologue that oh, only needed to be done the day before. :P   12:15 pm-  freak out somemore and panic. get dragged around by amanda to proudly display our Big Stick. hehe   12:25 pm-  get my hands tied and my hair messed up in tickling contest.   12:30 pm-  present monologue and kick ass. hell yeah i play a good deranged lady macbeth. :P   12:55 pm-  get out of acting and finish up college pub homework with zac\'s tantalizing burger in my face and amanda screaming in the background.   1:00 pm-  finally get to go to the bank and feel more liberated now that i have more than a bunch of plastic cards and seven dollars.   1:20 pm-  decide we deserve fat free low fat soft serve parfait. terrorize the workers at tcby with the bobsy twins. already high and they still feed us sugar?!   2:00 pm-  bore myself shitless at college publication class.   3:05pm-   puke the fuck out of my guts somemore. bye-bye lunch. hello gayle.   3:15pm-  sit around and tell her about the whole week. and falling, breaking. things i need to do, things i want to do, things i have to do. wants, needs, must haves, craves. punishing. attacking. laughing. pills. breaking. going crazy. pressure. tired. insomnia. sad. happy. sad. confident. sad. insecure. happy. happy. happy.   4:15pm-  she keeps me in for another hour and i say i only have a hundred dollars with me? and she says, it\'s not about the money. it\'s about getting better and i think to myself, she really wants to help me. and i think dah-yum.  and so for the next hour i let her in some more and she tries to rationalize with me. she comes not as gayle the psychiatrist but just gayle. she asks me why i didn\'t make the list and i say i did. i\'m just not sure if that\'s what i really think or if it\'s really real. i tell her more about what i really did this past week and how i hurt myself. i tell her i am afraid of death and darkness and going to sleep. i tell her about escape and standing strong and fighting. i am afraid of failing and letting myself go. run free. fall.   of losing control. losing sense of who i am. allowing myself to be all these things and not be so hard on myself.  she says i\'m mean and vicious.  that i need to take a break.  and she ends with you\'re not superwoman racheal, but you damn well bloody can be.   5:35 pm-  she picks me up and i scream and tell stories with my hand. hyper from the session, smiling and happy :)   5:45 pm-  we shop for outfits for tomorrow\'s dancing and friday. ah, friday. hm.    6:45 pm-  we head back for church. yes, you read right, Church.   7:28 pm-  church is well...like church. too much talk of war. too many people crying. i\'d be depressed but i am too calm. i can only bow down and say a prayer. sing another praise. read another verse. Faith.   8:30 pm-  we get out and she wants to get coffee but she gets into an argument with the boy and we\'re stuck on our side of the island in a parking lot. it\'s oddly amusing how emotional things get towards the pms stage. we end up laughing at how silly we\'re being and head home. her exhausted from a day full of random shit, me just tired.   9:00 pm-  i come in and there is a package on my bed. i shriek and laugh and giggle. my first thoughts are always him but fuck that. it never is anyway.   9:05 pm-  a fucking good card, a letter, chocolates, candy, roses, a teddy bear and silly lingerie(you pervert) later...i feel tears not the sad icky type but just good tears and i am overwhelmed.  \"just thought you needed to feel special.\"  i ring and scream and want to say so many things but it\'s only about five in the morning his time and i say thank you thank you thank you. remind me not to mess up this friendship.  i start out to write this but the aching body protests and so i soak till everything is wrinkly old. warm, bubbly and soothing. i fall asleep and this time it\'s good.  i sleep.  :)   urlLink  \"There was no exhiliration, no impulse to celebrate. Just an overwhelming sensation of closure, as if all my bones dissolved and I could only sag in relief.\"  '),(22467,664485,'11,February,2003','2003-02-11 00:00:00','   Hit me, slap me and call me Edna.  That’s how weak I am around you, shitfaced and fucked.  You attack, as if your words would pierce through this thick skull of mind and hurt it, it doesn’t. It only stabs through my heart and fucks with my emotions.  I don’t remember why I let you do this, only that when I am vulnerable, you are never there.  …and you wonder why I don’t come close anymore.  You give me no reason to. No reason to stay, no reason to try, no reason to like myself around you. That’s always kinda hard when your favorite game is to tear me into pieces and watch my flesh slowly be made into jerky. Wringed, dried, processed and slowly wilted away. Preserved in some sordid state, scrunched up consumed with guilt.   Valentines day makes it about a million times worst. It makes me think of past lovers and before the words are even out of my mouth I see your condescending bastard attitude that labels me again. It’s funny how I almost swear you are so uncannily right sometimes, that when you choose to be an asshole, there is almost a perverse pleasure in playing your game.  Lets see who breaks down this time. Lets see who pushes the other over the cliff. Lets see what great things we can hurt each other and accuse the fuck out of each other.  It’s become a sink or swim situation.  …quite literally too.  You sink as you fight to get back up to shore or drown which in case means you’re fucked anyway or you bloody stand up and hold your ground only to swim in your tears. Can’t win anyway.  It’s not a game damnit.  Yet it doesn’t hurt as much when you call me names anymore. Whore, slut, bitch, fucked up…I run through the list and they’re starting to just become words to me.   Instead of weapons you used to hold against me, instead of missiles you use to aim straight at me.  Being in a warzone, a crossfire.  Why am I so used to that?  It’s almost a ‘here we go again’ dejavu like existence. I wait for the day when you will truly know what it has been like. What it was. When your blindness doesn’t cloud your judgment anymore and you learn that attacking is not always the best answer.  And I would know considering I’ve done all that before, attacked, yelled, screamed, bitched, realized…and here I am again. And again. And again.  I am tired now. I’d like to go to bed. Tell your demons in my head to shut up. Tell the gnomes playing trickery in my heart to die. Tell your voice that pounds my mind to fuck off. Tell your empty words and lies to take flight.  Tell my sorry self to disappear again.  I have no personality around you. No name, no face, no heart. I am merely just the girl that torments you and fucks you up.  I\'ll break it down for you; you’ve left every other time and wonder why I won’t trust you. Wonder why I am horrid around you yet I can be so lovable, so wonderful around other relationships. Why do I not treat you like the rest of them? Why do I actually give a fuck time and again?  I’ve shown you my soul, and you’ve raped it again, and again, and again.  Yet here I am, bearing soul when the only thing you want is flesh.  You don’t understand me. You don’t know me. You use hatred as an excuse. Too cowardly to admit that I am something and I am far beyond anything you imagined. I am unclean, dirty and sick but at least I have passion, have heart and even still my soul.  You think I’ve sold it to the devil.  Well think again, because no matter how short my life turns out to be, you’ll always have to live with turning your back on love. You had it once, twice, forever but you’ll never have my soul.  I don’t trust you.  You will probably end up scarring yourself further by attempting to bring me down again.  You claim you don\'t give a fuck but really, at least I\'ll admit it.  I hate feeling the way I feel about myself around you. Poison, you are like deadly nectar to my system. I swear I can almost taste it now.  Take flight before we hurt ourselves some more.  Walk.  Please don\'t return.  Leave.  I almost don\'t care if it hurts.  '),(22468,664485,'11,February,2003','2003-02-11 00:00:00','   having things to say at this hour probably means i should just suck it up and go to bed.  '),(22469,664485,'11,February,2003','2003-02-11 00:00:00','   I scour through a chockfull of boxed up memories and pull out this distinct moment. The aura that surrounds it, this smell, your touch, the emotions in it are still vividly there…your heart still pounding beneath my rested ear, your voice still clearly audible beneath the whispers.  It was our first road trip alone, sixty miles for a gig out of town. It was a good one wasn’t it? Holding our breaths going through the tunnel and that damn traffic, convinced I was going to be cyanotic any moment now. Watching you play to yet another full house, smiling as you looked out across the room and smiled into mine, I’m proud of you silly. Getting your ass plastered, knackered beyond reasoning and having the rest of the band’s drunk asses to deal with, the little groupies that traveled out to watch you too and dreams slowly being realized.   Learning to drive your big ass stick home as you handled the gears and I focused on clutch, gas, clutch, gas, oh fuck me now, gas.  We had a good laugh didn’t we? Yes, we’d laugh so hard and silly it was stupid. They’d be so sick of how incredibly pussy whipped you were behind that rock star drum banging punk exterior.  You made me so bad and angelic at once, so naughty and good.  …That was when I still took pictures, when I used to show people the world through my lenses. I captured those moments only to be put in your inlay later, only to be published and shown to the whole wide world with pride. You grinned your silly grin and brought the pictures around saying Look at this! Look at this! Loving me as if I was the best girlfriend in the whole entire world.  Next to the work of professionals and promotional bullcrap you took pride in mine and that made me feel so special. Important. You always seemed to make me feel so big and tall, so valuable.  But it was that very same destination, driving home from that very same city that made me realize that I was more in love with being in love than it was actually Love.  When you’d have to soothe me late at night when the evil attacks came. Or when you’d listen through the door with the sound of the water running and me puking my guts out. When you’d wait after every doctor’s appointment to see how badly things had progress or gone better and begged me to take treatment and I’d always plead with you to just shut up. To be quiet, to respect my decisions, to let me live it my way.  You’d always try to tell me I was beautiful but I was so broken then that I’d never believe you. Never let you touch me, see me completely. I’d always say No, or switch off the lights, or turn the other way. And it was always Him.  Funny how I always blamed it on Her and thrived on how badly we’d both rebounded. We’d crack jokes but we always knew. It was over before it started yet I know what we had was Real, just not Ordinary.  Breaking up with you through a pillow fight and sending you off again. Avoiding phone calls and distancing myself away, telling you we needed to go back to our own ways and this wasn’t Right. It’s odd how it felt like friends with benefits. A brief period of extended friendship and being cared for. I didn’t really feel the pain or the hurt, I don’t think we ever reached the hurting stage? I do miss you though, I miss you for all the little crazy things you’d do, we’d do.  I have so many stories to tell you…things about how I’m doing now and how great things are going. Things I know would make you really proud of me, about standing up and standing tall. Still no treatment although I’ve started to heal mentally though, fuck the physical eh? I’ll deal with it later..:P  I took that drive again for a sorority meet and thought of you. Right down government on daulphin too. Those were some crazy ass times weren’t they?  Damn Snake. :P  Take care you.     urlLink  \"So I feel like a big bucket of stupid now. But at least I\'m a happy big bucket of stupid.\"  '),(22470,664485,'11,February,2003','2003-02-11 00:00:00','   I just talked to Pam.  His Ex.  Who wants to know why I am avoiding him. Hurting him.  I swear you can never win. Fucked if you talk to him, fucked if you don\'t.  Oh shit fucking yeah...and you wonder why I need therapy.  '),(22471,664485,'10,February,2003','2003-02-10 00:00:00','   How is it that Lucky Charms have  less calories  than Golden Grahams?  I’m being psychotically obsessive again; analyzing the back of packages while stocking up at my weekly three-in-the-morning wal*mart binges. I have little packets of cereal, lots of soup, yogurt and more fat-free light shit of any kind plus a good load of poptarts to suffice my cravings.  And that, is my basic diet really.   Oh and orange juice. The pulp-y kind + coffee.  :)  Somehow my body doesn’t like Real food, or Real products. The eggs, the meat, the proteins, the fucking carbohydrates only go in to come back out the wrong way later.   You’re  talking to  reading manic Racheal now.  The girl that goes to the gym with the ex and his ex who is the current squeeze and with ‘boyfriend of the week’ in tow at the crack of dawn with ‘my ass if falling out! my ass is falling out!’ pants.  The one that sings loud cheesy pop music while driving down long, long depressive curvy roads to get back to the next state/home/porn shop.  She’s also the one that acts like a bitch and expects everyone to regally bow down at her. Hell yeah, you ass, grovel.  She’s the colder, meaner side of what it means to live in this head, to put up with that depressive whiny annoying bitch. She is the side that says so fucking what?  The one that will damn bloody well kiss a girl if she wants to and not beat herself up for being the way she is. The one that is proud and confident, that smiles, laughs and yells at the same time, the girl that makes three leaf clovers with her tongue and makes silly faces.  She’s a bitch and I really should hate her but you know what, I think I’ll just keep her around.  Crazy turd.  All this early morning fat fighting blood flowing-boosting workout makes me hyper.  Hmm.     urlLink \"Bwhahhahahahahahaha. That was fun ;) \"  '),(22472,664485,'10,February,2003','2003-02-10 00:00:00','   You look at me as if I could be so many things to you and I tell you to stop staring and laugh uncomfortably as your eyes look deep again into mine. You reach out to push the stranded pieces of hair around my face, another excuse just to touch me you say, and you tell me again how incredibly turned on you are by my glasses. You call me sexy from across the floor and I answer to beautiful.  You twist and twirl me around, dancing in beat to the rhythm that pounds through this place. The glasses have come off and out emerges the girl that everyone’s used to. At least here in the Real world anyway, that confident bitchy chick. I have on this low cut slinky black dress, yes, more black and another dress, another outfit. The girl is layered with white and black, then white then black again…strappy silver stilettos and a necklace that fits just right and almost becomes a part of me. My make up after twenty hours in the bathroom is finally right, smudged and applied on as you periodically came in and putzed around, as you put your arms around my waist and I giggled as your lips kissed my neck and slowed me down some more.  We sit by the steps at the American Legion and we share lights as we take a breather from all of the craziness going on inside. You pick up the bill when I am not looking and steal kisses in the rain, you laugh when it hurts and your face scrunches up when you’re angry.  I come home to you and we get past the door to a whole new world inside. I let you hold my hand and lead me to some place else. I trust you though my trust is questionable in itself.  I let your fingers trace my bones and your palm rub my back. Your toes meet mine at the foot of the bed as you slip deep into me and I moan. I moan with pleasure, with pain, with uncertaintity. I moan with escalating desire and scream with lust. Like slapping my ass and fingernails scratching your back; how can this feel so right and so wrong at the same time?   Hours slip by and soon the fifth disc in the changer is playing as sleep slowly consumes me and I smile to myself as your arms wrap around me tighter in that very second. We get underneath the covers and soon, exhausted and orgasm worn out, fall fast asleep.  It’s funny how eating twisted Fritos and listening to Jimmy Eat World can bring back such odd memories.    urlLink  \"Your sweet sweet skin lingers on my mind. Your laughing voice tickles as it tortures. The glint of streetlight off the corner of your eyes as I fuck you lights the way. Your tits warp my vision, my view bending to suit the curves of your innocent flesh. Your taut face, stretched in a grimace of pleasure, mouth agape, throat muscles corded and trembling with passionate anticipation of the next moment. Your hips slide and thrust in thirst for each and every inch of cock. Your belly, warm and smooth, rubs and rubs and tickles mine. I get lost between your tongue in my mouth and my dick in your cunt. My hand on your back. My hand on your throat. The hissing draw of breath inbetween thrusts. The desperation of our embrace mirroring the vulnerability of our passion. Our legs intwined, our fuck fast or slow, our scratching or clasping, our kisses voracious or tender, in and out - in and out - in and in.  I thought I was lost until I found myself in you. I never spent a moment so happy as the one I lived losing myself in you.\"  '),(22473,664485,'10,February,2003','2003-02-10 00:00:00','   \"...it was a beautiful shoot, marc....... ....................it was honest, it was deliberate, it was Real.\"  It was that, love. It was that.  And \"the others\"? ...welll  Just as important, sweetheart.  You see  It seems to me that no one picture tells the story. As soon as the shutter clicks, another story gets told. And I have to save that one, too.  Some stories are sad...some thoughtful, contemplative. Some are cute, sexy moments of getting caught in your bra, biting a nail, and looking up with that all-too-adorable grin that says your titilated, but self-consciously embarrassed at being \"naughty\" - a word that even sounds naughty. Some are confident, strong, independent. Some are just plain bitchy. Some are silly. Some are playful.  They are ALL -- you.  Some say \"I\'m sad.\" Some say \"I\'m happy.\" Some say \"I\'m too fat.\" Some say \"Do I have a nice ass or what?\" Some say \"My tits are too small.\" Some say \"Fuck you - they\'re MY tits, and I\'m proud of them.\" Some say \"This is one of my best friends in the whole world.\" Some say \"I\'m bored.\" Some say \"My photographer is SO DAMN clever.\" Some say \"The Old Guy with the camera is an idiot.\" Some say \"I wish...\" Some say \"I hope...\" Some say \"I AM.\"  The thing is...  They all say:  \"I\'m Racheal.\"  Two hundred, twenty-five pieces of a puzzle I still haven\'t completely fit together. I don\'t think I ever will. For you see, I\'ve noticed something.  The puzzle keeps changing.  Every time you think you\'ve just about got it all together... Something\'s not cut to fit. And you have to start over. Or back up, at least.  Take more pictures.  Tell more stories.  Fit a few more pieces in.  But that\'s just the way Life is, isn\'t it?  A step forward. A step back. Shuffle to the side.  Pick up a piece, and snap it into place.  It\'s frustrating, I know.  But the thing is  It IS worth it.  Because, for all of the boredom and bullshit...  Every once in a while...  A piece just slides into place.  And you step back, and have this epiphany.  \"That\'s what that is!!!\"  (click)   I’ve always wondered…sometimes with purpose, sometimes just because and sometimes even against my will; what did I do that was so wrong, so right, never loved, so damn adorable and so strong sometimes?  Why have I become a puzzle filled with a million pieces, a tale with a million sides. Songs.  Why does the music only play in my mind and when I trace, only changed much later. Or right now. Or a while more?  How do I become hot then cold, furious then angry, obsolete then saddened, disappointed then heartbroken? How do I manage this existence of hiding then letting go, giving up then moving forward, giving in then breaking down?  Why am I, over millions and millions of people, the one that gets sick, the one that’s fucked up, the one that’s so damn sensitive to things around her? Why am I tired when I’m not done yet, happy when I was sad just a moment ago and dysfunctionally perfect. Why do my flaws make me special, and my idiosyncracies make me wanted?  Why don’t I have size 48 DD breasts with long blonde hair and blue eyes? Why is the other person always better and more beautiful? Why do I berate the person in the mirror and let the voices control me?  Why am I so weak and stupid?  Why do I question so damn fucking much?  Because I can, that’s why.  Because Eve let us into the garden of Knowledge.  Because Pandora opened her damn box.  Because I have a brain and it’s working harder than it probably should.  Because I’m racheal, and this is what being \"racheal\" is.  This is what loving racheal means.  Because it\'s four in the morning and I\'m good at analyzing shit like that.  Or so I\'d like to think.  '),(22474,664485,'10,February,2003','2003-02-10 00:00:00','   i sit here sipping chickenless chicken soup and i log in periodically into my server. internal error-\'cannot find server\', it\'s been acting up the last couple of days. in some sense, it really does feel like a child to me, the constant need to be cajoled and taken care of...or how it sits pathetically when i go through binges of being so sick of the internet world and avoid it.  i don\'t know how i feel about my new layout; perhaps that is the beauty of it. that i am undecided if it is too much or just right.   i remember that shoot though, the selection of pictures i have chosen to put up and not put up have to a certain extent been deliberately done so. i am not completely stripped as per say but a lot of times, i feel as if my soul is raped in here. as if i have unlayered and am completely bare; my thoughts, my wants, my needs.  i remember crying and breaking down. being stuck in a corner in darkness and music played just right, just enough to capture the moments. evoked just painfully enough to snap the series of breaking down.  it was a beautiful shoot, marc.  it was honest, it was deliberate, it was Real.  and that\'s what i loved most of all about it.  ...now the others are well, questionable.  :)  perhaps i\'ll put together the portfolio and actually upload it later.  we\'ll see.    urlLink  \"I\'ll eventually get a wild hair up my ass and feel the need to write more. Thanks for being there.\"  '),(22475,664485,'09,February,2003','2003-02-09 00:00:00','   There are times you just have to shut up and do the chicken dance.  Oh, Oh, Oh!  :P  '),(22476,664485,'09,February,2003','2003-02-09 00:00:00','   The three year old that runs around in her half falling diapers chased around by the maid that cusses in the foreign language and laughs at the same time. She is cheeky and extremely playful, energized and never sleeps. Thumb sucker and independent, no pussy ass pacifiers for her she says, she colors the world in bright beautiful hues of every damn kalaedoscopic color and strives to be just like her older brother. The cool one yet she fights him to death and pulls his hair and torments him rightfully just like a little sister should. But late at night when she says her prayers, she thanks god for mommy and daddy and yes, even that idiot.  Smart ass.  ---  Fourteen: Belligerent and still opinionated. Still a little girl at heart thrown into a world of hormones and indecision. Too little hours in a day she splits between her different worlds. The entertainment industry that sucks her in and matures her indefinitely, scars her for life yet teaches her all these valuable lessons. The pretend world she lives for the ones around her that thinks she’s always in control, always happy, absolutely beautiful. And the one she keeps to herself, the spiraling depressive bitch that tears herself apart and abuses her already tormented body. The suicidal girl that deals with all the diseases that inflict her already broken body, she scratches her nails and feels the pain of it being ripped off as she injects brutal force into it.  Angry at the world, she deals with the world around through mediums. Discovers herself and the people around her, she learns the ultimate lesson on Life.  ---  Seventeen and burned out. She gets back in schedule of the Normal life. The preppy girl, captain of the cheerleaders, the one that fucks with all the rules in school and still manages to suck face with the teachers. The prefect, the one that always gets away with marking her independence and screwing the uniforms. The ones around her with a cult like obsession, the ones in her clique fuck around with each others minds, the smart lost ones. She is calm, calm, calm but curious. Still questioning, dabbles her foot back into the industry and can’t decide if she’ll stay or not.  Has people wait on her hand foot, craves on the control…the power. Her dark abusive rituals have become more apparent now, four years of eating disorders and depression becomes obvious after a while. always excused after a meal or doesn’t eat at all. She writes away her pain and finds herself. Constantly searching for herself, the being she is, can be, will be.  Graduates and tries to be like the rest of them but she’s never been Normal. And this time she appreciates it. Her difference gets her into clubs, gets her on lists and takes her places. Her difference gives her the nice car she drives and the fucking fantastic clothes she gets, her difference makes it a perfectly reasonable excuse for the person she is. She accepts.  ---  Eighteen. All the things they have drilled into her is finally beginning to make sense. She can now look in the mirror and not want to smash something against it, she can look at pictures and not tear it apart. She still has to silence the harsh whispers though, she still has to tell herself to shut up and take a break.  Set backs propel her forward, tumors invade but she stands back tall and strong and says fuck that.  She’s changed and changing…everyday still goes through fast like the rollercoaster. Up and down, up and down, down and slow in anticipation of the highs, and high and afraid in preparation for the lows.  People still come and go in her life. Love fills around her and invites her into a different realm. She breathes with life and dances to her own, drums to her own beat. The difficult persistent ones stay, but those are the friends for life. Thick and thin, weak and strong, she grabs on and laughs as she slides down this ride…but it is only what it is, just another ride.  She smiles at the girl that looks back at her and knows things will be right. Her ending is still unchartered, her life still in threads…she doesn’t know when the day will come, or how it will; but she prepares the best she can.  She lives.  She smiles.  She’s growing up.    urlLink  \"It is morning and once again the world has changed.\"  '),(22477,664485,'07,February,2003','2003-02-07 00:00:00','   urlLink http://agedandconfused.com  I watch you reach out to thin air and cry. I watch the tears bleed out and the broken smithereens of your heart in your hands; and for a moment I wish I had all the right things to say, I wish I could hold you in my arms and make you feel safe a while, I’d listen to your screams and tell you it’d be okay…and I’ll do all the things the others do for me and tell you,  urlLink  “hold my hand little girl, we’ll fly away.”     Or at least try to.    urlLink http://karac.com  You hide your pain with a cynical mask somewhere amidst the broken lonely girl. You call us the Internet whores, the fake people. Designed, instigated and filled with putrid thoughts, but what you don’t see or perhaps understand a whole deeper more is that we are real too. Still people. Still human, still with feelings. Feelings to love, to care, to understand, to hurt. And we are still Real.  If only you’d understand.    urlLink http://karac.com/~jared  I am your whore, your slut, your bitch. I am the girl that can’t keep her legs together and fucks you over from time to time. But what you forget, is that I am still all the things you once saw in me, I am still her yet maybe you never knew me at all. I can take this abuse, and I can hurt a while but with every break grows a stronger child, with every beating comes a bigger resolution. So tear me down, so cuss me out, so yell, scream and hit me again with the words you use so well and when all is over, rest.    urlLink http://brickchick.net  You question the hands of fate. You analyze the importance and value Love is to you…is it worth it? Is he the one? Can this be it? You ask again, unsure…scraping on this surface of obscure existence. What are the answers? Is this love? Do I love him? Will we make it? I don’t know honey, but Try.    urlLink http://jasmine.blogspot.com  You avoid the rest of the world and hide. I’ve seen you take the back step, find your way in the world again. You mask yourself with being cold and callous, but I still see her. I still hear you break sometimes in the middle of the night, you want to know how long more he’ll wait. If love transgresses all things…you move on then come back again but know you are better each time.  twenty one years.    urlLink http://racheal.thewarmcompany.com  You pass judgement on others…but I see you too in the middle of the night. When your insomniac mind questions and questions and tears your being apart. You put little steps in to endeavour into that world called Love again, but with every good potential try you push away. You smile and arrogantly dismiss them all, twirl around and dance to your own little tune. You work all these problems and fix them with a little bit more than superglue and some tape…and then you remember that if you can’t duct it, fuck it. Your sorry little roaming ass heals, yet you still fall down into this foray of darkness whenever you get tired of the world. Outside, behind/beyond these four walls lies Life. You aren’t the only one screwed.  Choose your poison and Live it.  Hope.  '),(22478,664485,'07,February,2003','2003-02-07 00:00:00','   I sit here trying to figure out why this end seems so familiar. An almost suffocating dejavu like existence; I threw up tonight. all that I had been going for, all that I had worked on came crashing down after the game.  The  game.  Victory is always sweet,  we  got the stick, big guy. Artistic impression, foam thrown together, ideas clashing, blue+green+yellow hues for the last couple of hours have proved themselves, big winners in the end. Strong competitors. Empty existence of a non-energized fuel. Hit the delete. Hit the delete. Delete.  Rolls and rolls of pictures. But  urlLink this series  hurts the most. I remember all that was felt in that moment; crouching, reaching out to hold him yet defensive…crying and letting go at the same time. Knowing that within those four walls I might have been safe yet still sucking it in, only the betrayal of random teardrops that escaped my eyes.  Spotlight that shone through the blinds, created light in the complete utter darkness. Deafening silence only intruded by your shutter speed clicking against the torturous melancholic crap I’ve played. Now who the hell needs inspiration when sadness is my forte?  Fell asleep out of pure utter exhaustion, all the sobbing in a day plus the onslaught of aneroxic attacks I feel coming on…or have they came already? I can’t tell anymore. The fidgeting from absolutely no nicotine in the system, caffeine and gum with that damn perky bitch boinking up, down and around running through aisles and aisles of restocked crap in wal*mart and playing in the ice rain. Not snow, not salt, but ice rain. In fucking Florida.   Home now. Safe in the confines of this room I have come to know well, sleep is almost foreign now. Insomnia is my best friend as I tear apart and recreate this place again, reaching out to tell you I’ll be the fuck okay because I know you’ll catch me if I’ll fall. Because I know it’s okay to fall.  The therapist wants me to make that list: things I like about myself.   I’ll get there…eventually. I think of a couple than laugh myself silly, who am I kidding? To feel so confident and insecure at the same time is a bitch, but at least I’ve stopped crying now and sucked it in.   At least I’m stronger now…making it.  Watch me.  Watch me.  I’ll make you so fucking proud.     urlLink  \"This whole thing started because the Muse has a problem with her \"Delete\" button - she refuses to use the damn thing. So she\'s all in a blue funk, calling me at odd hours, as if a smiley-faced waffle is the Great Balm of Life.\"  '),(22479,664485,'07,February,2003','2003-02-07 00:00:00','   I wake up to lips on my forehead and someone lying beside me...holding me.  He smiles as I turn over and he rubs my back. We make love, bare skin on bare skin, yet it is without words. Without sounds.  He looks deep into my soul and sees Beauty, he believes in me. I don\'t need to question, I feel.  Draws me in close and I weep like a little child.   I sob, and hurt, and shed more and more faces. Facades. I tell him I can be so many things, that I have so many dreams and I scrape and scrape on hard rock cliffs barely hanging on, climbing back up. Trying.  I cry out to all that is within and around me and tell him I want to give up so bad.   ....and then I rub dreams away from my eyes.  Outside, raindrops are falling, invading this barren winter earth and cascading it into life.   It is a good forty degrees or less;   i\'m layered up and ready to go again.  Fuck you world, I\'ll sail my ship alone.  '),(22480,664485,'06,February,2003','2003-02-06 00:00:00','   Fear grips me as blood rushes to my head and my jaw drops open: yell, scream, anything at all but nothing audible is heard, just my vocal chords going against the panic, near hysteria in my mind. My stomach in big giant knots, I feel like I’m about to throw up and pass out at the same time…only gripping tighter. Gripping so tight my knuckles aren’t even white anymore, I taste the salty streams running from my eyes but I don’t remember crying.  And there it goes again, another gunshot.  A loud deep rumbling    BOOM   I want to step on the gas and run every red light that stops me, I am too afraid to curse the long dark winding roads, and attempt to just focus on going home safely. Out of nowhere, it comes, and I am screaming and screaming and I cannot stop.  The harsh glow across my dashboard tells me it is a quarter to two. I pray right there and then for a cop, for some highway patrol person, for Lillian Highway to cross Bauer  Now.   Two pairs of headlights behind me; Was it them? Are they tailing me?    What the fuck is going on?   I ring him and I am yelling, crying and not saying a single solitary word that makes sense at all. He yells with me, 911 on the other hand…what the fuck is going on?  I am afraid to turn, afraid they are following me. The pain so much more worst tonight, my medication sits rotting away by my bedside drawer…I honestly do not know what to do.  I tell the other dad to meet me outside. It’s two in the morning, what the fuck is happening?  I run in and scream and pass out at the same time. He makes me take my pills and get a shot of steroids, my bloodstream is fucked over but I am finally calm. Breathing.    “It’s all going to be okay Racheal, you’re safe now..it’s alright honey.”   No I’m not.  No I’m not.  I’m scared.  …..I  want  to  go  home  Now.  '),(22481,664485,'06,February,2003','2003-02-06 00:00:00','   urlLink Tears In Heaven – Racheal  ‘Out of Tune and Bored’  Khong   This is kinda silly…but significant I suppose?  My dad used to play ‘Tears In Heaven’ for me when I was little, and we’d sing it together...I miss my dad? :(   For some odd reason (boredom most probably) Amanda and I decided we needed a break from our mundane studying shit and so, loaded up on caffeine and math formulas, we went and made our own entertainment. Cracking ourselves up with the karaoke booths at the mall then laughing ourselves silly with ice-cream and whaddya know, more coffee!.  :)  Unfortunately, she won’t let me upload her fuckingly good voice but let me tell you, DAMN that girl can sing!  Lol, other than the constant  \'now how the hell does this go again?\'  plus crazy ass very audible jitters, I had fun. I have to say though that next to that woman I’m like the squeaking rat. Hehe. So if it’s god awful and cracks your windows…I told you so! (sorry)   urlLink ‘Stay’  had a lot of weird chokes and laughs in it too because well, it  is  a damn mall and we had to record it in one go….we being the two screaming high semi claustrophobic girls together…justified my picture yet?  Hmm…bet you didn’t know my voice was so manly eh? The alto against the descant, smart move innit?  :P  In other news though, guess what I found when I came home?  One fucking dozen roses…Wa wa wa wa!  Thank you Terence! Whoop. Whoop.  Hehe, yes, being Thai is good indeed.  It has been an amazing week…well somewhat anyway. Like for all the bad moments I’ve had, all the little ones like this have made up for it you know? or something to that effect anyway..  Cast read-through tomorrow. I’m so excited I think I might piss in my pants!  Baby got the role homeboy…  Hell fucking Yeah!  :P   P/s I had a long ass talk with God today, got some shit settled so it was nice. Now what do you have to say to that,  urlLink the Wrong Reverend ?    urlLink  \" \'It\'s a bagel,\' she said. \"    I want a scone. No, I mean a Real Scone.  gah.  '),(22482,664485,'05,February,2003','2003-02-05 00:00:00','   Hee. Hee. Hee.  I have  something special  a surprise for You!  :P  '),(22483,664485,'04,February,2003','2003-02-04 00:00:00','   Happy Birthday  urlLink Jasmine !!  '),(22484,664485,'04,February,2003','2003-02-04 00:00:00','   I have a couple of threads in my bag…  One being about the shuttle disaster and the other about this whole loving myself business and getting better crap…which for some reason tonight, feels just like what it is, indecisive crap racheal spurts out through her ass once in a while.  School has been insanely and tremendously taxing the last whole week. I honestly swear I spent at least eight hours in school yesterday in classes and studying, figuring out the language that I am apparently supposed to be good at but have been struggling with. I hate to fail, I hate getting a B when I damn bloody well know I can get that A. Damn word problems I tell you, damn metric system.   We had to go for this speech by Leonard Pitts at the auditorium last night in conjunction with the Black History month for extra credit. (Ten extra points? hell yeah baby!)   I scribbled down a lot of worthy quotes amidst all the other notes I had to take to get my paper done…I guess a part of me has always realized how important color was to this world but not really quite as bad as it hit me last night.  I am Asian, and I am proud to be. I dish it out almost quite as much as I get it, I’ve learn to laugh about it mostly, if only just because, I have to accept the fact that people are ignorant sometimes. I cannot control the person that judges me just because of who I am before looking into  who  I am. I cannot help it if someone thinks I am only good at making clothes or making ‘flied lice’ if only because that’s all that they have been exposed to.  But what I can do is smile when I feel like slapping the shit out of someone that takes me down just because I am yellow. Turn around and face the other way when someone decides that I am not smart enough, good enough or intellectual enough just because Malaysia is a third world country. I believe in karma, and I believe that actions always speak louder than words.  The only thing I really didn’t quite agree with last night was when my generation was referred to as the  valueless generation.  Somehow, even with all the bullshit and rubbish “my generation” comes out with…it still seems acceptable. It’s sad nonetheless but if only those condescending critical eyes would see the beautiful people I meet everyday. The people from my era that have accepted me and supported me for just as I am.  Loved me.  Then maybe they’d understand why I take offense to that…because contrary to that, I think we still have our values.  Anne Frank said a long time ago,   “In spite of everything, I still believe people have a good heart,”  and it’s still true isn’t it?     urlLink  \"eh, it\'s something to hope for.\"  '),(22485,664485,'04,February,2003','2003-02-04 00:00:00','   \"I\'m changing my name to \"Frosted Cheerios\" in the morning...\"   Dumbass.  :P  so I turned down this extremely yummilicious ass today…and I’m sitting here trying to figure out why the hell I did that and I’m drawing a very sorry blank.  I swear confidence or at least this sorta happy/whoop-de-doop the world is so perfect attitude attracts people…which doesn’t really make a lot of sense if you think about it?  Why? Why when I’ve sworn off my rights and declared celibacy…WHY?!  Hehe.  Oh well.   I think there might really be seriously something wrong with me though…I’m normally very fascinated by the whole romance, sweet-talking, gentlemanly crap but now it all just seems irritating. It’s as if something in me just really can’t stand it and I just want to yank my hair out and scream, stay on that side you idiots! Leave me alone….i can appreciate it now!  Gah.  Tired as fuck…spent a good portion of today studying and in classes then had to go to some event for extra credit. It was fun but I’m much much too incoherent to analyze it now. (we have to do a write up for it) Back to back coffee load up in racheal’s system is bad…you’d think I would’ve had known it by now eh?  In other news though, I’m going to take a nice long, long, hot bubble bath.  Excuse the Bitch.  My bones/muscles are aching…and it’s Tuesday tomorrow too..  Damn.  Dah-yum.  Must.   Get.   Fit.  Arrggghhhhh!!!!!  think happy thoughts, think happy thoughts.  batthhhhhhhh.  '),(22486,664485,'03,February,2003','2003-02-03 00:00:00','   algebra can go kick my fat hairy yellow behind.  and frosted cheerios can go lick it.  '),(22487,664485,'01,February,2003','2003-02-01 00:00:00','   The world waits, wings arched…perched just above a very thin ledge. Obscure + waiting…peacefully serene; calmness and confidence needed to take flight, to soar above.  Mission launched, then aborted, then fought back and launched again. Guiding over tall peaked twin escalades, a bid into the future. A signature of feminine strength, vitality…it ages along the years. Wiser yet dethroned: dancing spirals of steam creates atmosphere, stirring…an oblivion into the world unknown. I wait.  Desires give birth to dreams. Slightly stronger with every backstep, trip?pun intended: it is a fall and a journey together.  Gorgeous divine implications, a bruised knee scraped creates blood flow and heals.  ReBirth.   I feel like an inferno champion.    in·fer·no ( n-f? n )n. pl. in·fer·nos 1.	A place or condition suggestive of hell, especially with respect to human suffering or death: the inferno of battle. 2.	A place of fiery heat or destruction.      We went to the Pilots game tonight…very, very interesting. I swear there is absolutely nothing like caffeine/booze fueled loud cheering fans and violent anger management needed players on the ice. Fun, fun, fun!! I might have lost my voice and a good chunk of my hair from the shoot out and nerves, but shit yeah it was worth it.   Whoop. Whoop.  Besides, there was skating afterwards and landing on my ass is always amusing…you know they don’t call me Grace for nothing eh?  And for god’s sake, I’m all about this whole Beautiful, love-yourself shit but damn, just how many times can you rotate that song on your airwaves in an hour? I swear I have reached a point where I will seriously scream and stick my fingers up my ass if I have to listen to it again. And again. And what the hell, just one more time cause we can huh?  Someone kill me if you find me still humming that song. Good song though, and like every other pop song out there, true dat.  I should probably go to hell for eating pop tarts at this hour but oh well, I’m hungry. Hah. And contrary to popular belief, no, the hotdogs at games is not quite edible at all. It was cute though, he offered to go over to Copelands and get me real food. Lol. High maintenance bitch.  I’m sorry, I just have class.  …says the idiot in a big old tattered t-shirt munching on her cold pop tart.  Yeah.  Uhuh.  Whatever, woman.  Bulimic manic depressive retard.  Classy indeed innit?  And just for the record and future references, I’m stronger than a stick of no good tobacco thank you very much.  …as she twitches and pops in another Tylenol.  Gah.  Withdrawal symptoms: ooh fun!  Just you wait for the bitch to come out?evil snicker *  Okay, time to get back into the strait jacket.   p/s yes, I did realize how ungodly long and obscene the last whine/rant was but yeah,  my  space! :P    urlLink  \"how do you say high priest sex god from heaven? \"  '),(22488,664485,'01,February,2003','2003-02-01 00:00:00','   You see, the beautiful thing about the Internet is that somehow, the world seems so much smaller in here.  At last count, twenty four emails from strangers after that last ‘happy’ post…   I sit here wondering how so many days, weeks have melted into such a period of time; a spiraling depression that apparently no one but me could take out of this rut. The long whining nights, the early dawn phone calls, the running of tears and continuing sobs that rang through the night, hid through the day. Why have I been so sad? How is it that I have been so dysfunctional and not really realized it?  I didn’t phantom how much me as a complete and whole person affected others? I didn’t realize how critical and vicious the voices in my head were…how easily rectifiable this problem was. Listen, Understand then Learn.  I owe an apology to anyone that reads this site, I haven’t been much of me lately. That stinking depressive blinded bitch…I shot you down, one after the other each and every time you tried. Protecting, as if guarding from Hurt again only to understand that I had confined myself in wraths of wretched Pain all on my own.  I am soothed by the chanting, flowing sounds of sigur ros…it feels like an emotional journey down some lane. The lane of the Unknown.   Life?  Put in a trance of solitude, nothing can go wrong today…this week…anymore. It’s not if I am saying I’m never going to feel shitty again, but rather it would be a manageable amount of shittiness. I just never want to fall so far down no one can reach me again, or I can’t reach them.  Loneliness is a terrible torture. I don’t need to be alone, I just choose to be most of the time.  “Loneliness and being Alone are two different things sweetheart,”  Are they? Somehow, I must have mixed them up. See I figured as long as I didn’t get into a relationship then I couldn’t hurt anyone and no one could hurt me…never ever have I been so wrong before. I crave for that touch, for that emotional bond unsevered….I crave for someone to wake up to, and someone to fall asleep by my side. I realize how perfectly easy that would be to fix that but I hate that one night stand shit, and these people that I’ve been dating? I don’t know, it’s hard to grow close to someone and not want to let them go, not want to take the next step.  (Of course all he thinks is that I just need to get laid.)  but it isn’t that.  I wish I could be comfortable enough in my skin and it wouldn’t have to be a face. I know I am all these wonderful unique things, it isn’t as if I don’t see it but sometimes it just gets hard to believe that shit. When the voice you’ve been listening to all this while shouts above all others and controls you. That voice is embedded in you, without it you are lost and too emotional. Without that voice, you have no common sense or ration, and so you keep it around. Silencing it when you need to yet always submitting to it.  The fight back up is a difficult one. Shut up Voice, person, whoever you are. I am beautiful, and I am enough. I am all the things that is enough to be a precious stone, just unhoned yet. Just need to be worked on but some day I will be all the things you told me I never could be.  I’m on the Honors program, and your little voice that told me I was stupid, dumb and undisciplined can go to hell because I made it. All on my own I studied and worked on, my very own self…I did it on my own.  Casted for the leading part, when your loud voice boomed in my head what an untalented ugly whore I was. Pinpointed all the little things during the audition and making me feel so much smaller, so insecure about the rest of the world. You laughed and cackled when I told you of my dreams, when I said that I could make it. Without Drew around you said I was pointless, no one to organize things for me but I did it on my own two feet and I will damn bloody well stand tall and proud.  You said no one would ever love me. That I was undeserving of love, tainted, ugly and disgusting. You told me I would never be worth it or enough and I wonder if you ever knew me at all. Voice, I believed for so long that you knew me best and was blinded to the fact that I was a somebody. When they came in succession, I pushed them away because I didn’t want them to see the inside me…not knowing that they had already seen things in me I never acknowledged. You taught me to let my prejudices push away love, that Love was bad and evil.   Five months and sixteen days. You said I wouldn’t survive, that my spoiled little fat behind wouldn’t make it with having to go without my car, without the maids, without my friends, and without the doting parents. You said my sorry ass would come home the next month and I did want to but this equally fat and equally stubborn behind persevered and here I am, almost reaching six months and counting. I learned to do dishes and laundry, I learned that cooking was very good therapy and have been closer to my parents than I have ever been in the last eighteen years.   You called me a mutt. A breeding of sorts and made me feel less important. Cultureless as if I had no homage. But then it took me a while to realize it only gave me character, somebody to be proud of. I hid away and didn’t like speaking, being made fun of because I felt so different, not realizing that this difference was what made me so special. Just me. It wasn’t that they didn’t accept me, that the harsh cruel eyes the world I thought I lived in that tore me down but it was me that wouldn’t, wouldn’t accept or love me just because.  Yet I thank you Voice. I thank you because you’ve pushed me further than I would have ever gone. You challenged the boundaries and scraped beneath skin to get me to the core. You brought me real people, friends that had to go through shit to be a Friend. You showed me things I could fix and made me god damn bloody well fix them and wouldn’t get off my back. You let me enjoy many beautiful places whenever you suffocated me so much I had to run, get away from everything. You taught me about late night telly and how calm the bay was at night with the bright of the stars reflecting off the water.   I know you haven’t left yet, I know your loud, obscene, criticizing voice is still somewhere around but it’s really okay. I don’t need to change everything about me just cause you say so, I just need to accept me. Then I’ll understand and grasp fully the love me or leave me concept.  I don’t need to be afraid of Honesty. It’s really not such a bad thing is it?  And for once you know I am Right.     “In three words, I can sum up everything I’ve learned about life.  It goes on. ”  '),(22489,664485,'01,February,2003','2003-02-01 00:00:00','   The bitch got Thai!  Wahahahahahah…!!!  Leading female lady ; uhuh, uhuh, *breaks off to song and dance *   The callback apparently wasn’t really a callback but rather a test on reliability. Nothing like the sweet ass grin I carried for the rest of the evening when director dude said, ‘welcome on board!’  Heeheehee. I’m feeling pretty darn good about myself now it’s scary…I spent the rest of the day first trying to muster down as much lunch as we could before the thing, then studying some at the coffee place before we went back to lesbian lover’s place and got ready.  She dressed me up in this little white dress with sparkly butterflies on them complete with stilettos and shit. It was interesting…but damn that dress was short. Hell, you try swing dancing in some heels and a short ass skirt….helppp! : )  Marc took  me  manic panicked racheal out this morning too..there is absolutely nothing like bitching the first thing in the morning and ‘getting it all out’. Working aimlessly at a monologue that in the end had no significance at all seeing as the cast had already been decided. Oh Well. No complaints here!  I sit across, looking at her. She is beautiful. Straight....pity innit?  The boys ask for my hand and dance with me, throwing me around and forcefully. Coordinated, as if almost a farce on it’s own…we move in synchronity, bop to the beat and smile, laugh, smile.  Freaked out at the other coffee place, my little punk rocker hang out. Watched the live set for a few minutes or so and contemplated between moshing or shuffling a deck of cards…Both equally torturous.  Ladies night at the club next door…somehow dragged in and it is military plus galore. Coerced into dancing and being bought drinks off the table, as if fresh blood has entered this den. Remain courteous and smiley awhile before excusing myself ever so politely to leave.   Woman in control. The girlfriend holds the mind but I hold the keys.  We drive for much passion must have been felt. We fall back into supper with good friends and partners in crime. Reminding me again, what a beautiful world I live in.  “I’m quite a rascal ain’t I?”  Home at last, tired as shit but happy like a saint.  Time for bed!  whoop. whoop.  ZZZzZzzzz  '),(22490,664485,'31,March,2003','2003-03-31 00:00:00','   she wrote the note in a haze of confusion, lost, blinded...searching i suppose. hoping to reach out.   she\'s thinking about taking this down again, it\'s becoming redundant and she thinks she needs change...or at least a break a while. she\'s gotten too used to falling into depressive states and writing things off in a non committal way...she\'s thinking she needs to live in the real world a while. experience the consequences and strengthen her actions.   actions speak louder than words   an afternoon with herself and the old man. a special gift. thank you you.  '),(22491,664485,'30,March,2003','2003-03-30 00:00:00','   he writes to tell her he has ears to listen to her a while, a heart to sympathize and a mind to rationalize with her. and she is touched.   thank you.  how are things? i don\'t know. how\'s the boy? i don\'t know. how\'s your life? i don\'t know. how\'s your mind? i don\'t know.  i think i started crying on wednesday was it? i don\'t remember now, the days just mesh together and become one long overdrawn hour, that invisible hour of desperation, the hopeless hour of darkness, the emptiness of a moment.   and she weeps.   alone.  pillow to chest clinged tightly, curled into a ball...and the strangers are outside, the voices inside. shut up a while head, be still a while heart.   she\'s lost daddy, and no one can find her. she\'s stuck in a maze she\'s crafted on her own, of dead ends and never ending roads, strides into never never land. she\'s broken her wings.  the girl in the mirror laughs at her, she sees the tears in her eyes, feels the wet streams down her cheek..cries again, but all is trite now. old.   redundant, her diseases bind her; locks her in a destructive cycle. she cannot get up, cannot get out. she cannot find her way...would you help her please?  she\'s dying.  frustrated, agonized, disappointed with herself. so amny things surround her and pressures. she has no answers. solutions.  she\'s sick again. her demons grow rampant, her dragon burns.   tormented.  alone.  by choice...don\'t bring anyone down with you.  she\'s leaving.  she walks away.  ...lingering smell of fresh roses in her car, unanswered messages, invisible man.  she hides.  '),(22492,664485,'30,March,2003','2003-03-30 00:00:00','   the term comes around and things finally begin to fall into place. of ups and downs, highs and lows, a certain tranquility finally washes over...though not quite in the middle nor in between, it is a limbo that flows fluidly, gently this time. of finally learning to tug harder, or softer, scream louder or whisper, push harder or take a breather and though things still worry and frustrate me at the same time, there\'s a sorta peace that underlies it all and that is kinda nice.  my budget is out of whack again with the home phone being added to my bill. somehow, i was very slyly conned into that one and am now paying a bulk of the calls. because we are separated by about three miserable miles, all my calls are mostly state-to-state and hence, long distance charges. painful, painful, charges. i\'ve been trying to keep my international calls in check but the pensacola calls just take a beating then...it feels like a see-saw sometimes, one or the other...though i\'m getting smarter about it so yeah, takes time i suppose. every bloody thing takes time. gah.  met up with marlene and a bunch of the other girls to watch the dance explosion downtown and it was good. albeit entertaining but definitely memorable. oddly enough, i got a phonecall today about another indie movie then got offered by dominique to do the vocals on his next project. electronic clash something or another he was explaining, but considering this afternoon\'s drinking binge and the wine we had during dinner...quite a bit flew past my head.   :P  tentatively, as it stands right now, i might or might not move in with marlene. apparently, i have to put together a \'business proposal\' for my dad. what living outside involves, promises, health shit, budgeting, drama, school and the person i\'m going to live with. there is a lot of pressure on myself to straighten out my life and to get out of the desperate pathetic route i\'ve taken the last couple of months and to basically, \'get my shit together\'.   i get so pissed sometimes, at myself, because i haven\'t really mastered the \'real world\' techniques of being independent and it frustrates me to the bone. i can get there mentally but physical crap from being a real spoilt brat my whole life just bugs me. all the trivial miniscule details you have to look into, separate the whites and colored, how the fuck do you slice beef?, checkbook has to balance, utillity/cellphone/phone/school/lunch/going out budgets, i hate being dictated by all these stupid things. i suppose the immature part of me still needs to learn about these things and it gets sorta overwhelming at times.   i wonder if i\'ll get back to that routine this summer. and have to readjust again.   \"i want my fucking car. the damn maids. my parents. and not have to worry about money.\" - now, that would be the brat in me. though the other part says, i want to be independent. i want to work this summer and make my own money, iron my own clothes (hah!), open my own account and budget the bitch out of me so i\'ll grow up. so i\'ll stop messing up. so i\'ll be ready for the world eh?  he rang earlier from work...it was really nice. shit, it made me giggly and plastered a smile the size of texas on my face. and no, it wasn\'t just the alcohol. there\'s something very special about knowing you\'re the one he rings, or that he cared enough. bothered to. there\'s something very special about being attracted to someone that is attracted to you and sharing that bond. i feel like i\'m rediscovering this, understanding from what\'s beneath skin deep and learning about this again. being wary yet at the same time allowing myself to feel this, to be honest with myself. to know there are issues and there are problems but also to understand that there are solutions and there is such a thing as working it out together.   so if things don\'t work out...then they don\'t. big fucking deal. if my heart gets smashed, pulped, twisted again...so be it. i\'m stronger, different, changed and all that now...but it isn\'t too bad. i like who i am, i like the person i\'m becoming...and whether he helps me shape up to be that person or forces me into it..i don\'t know. but he\'s here now, and he\'s in my life some way or the other and that isn\'t up to me.   i\'ve come to realize, my head and my heart are entirely separate divisions and so i can think this to death and yet my heart will still have other opinions. so yeah..  life.   i\'ll be alright i suppose.  :)  '),(22493,664485,'29,March,2003','2003-03-29 00:00:00','   she came in to take things down, but right at she signed on he sent her a message saying he\'d just read through it all. he wanted to talk things through, to work things out. perhaps, maybe, i don\'t know was the question of the day as it\'d been a while.   lets talk. lets hang out he says. lets try and work this through, lets give us a chance, lets figure this through together.   being away from her he says has not been good. he says, he figures, he thinks, thinking is evil. thinking hurts.   then she launches into a cascade of a million thoughts unpolished, unrehearsed and so everything comes out in a confusing mess. things are left hanging in the air and words pieced together weirdly, choice of sentences kinda screwy but she needs to lay her cards. she needs to be honest with how she feels so he\'ll be honest with her. so it\'s fair game, so they strip off notions and facades and bullshit and work at this.     and she does want to work at this.   she doesn\'t want to let another one pass by, doesn\'t want let another one go just because she was afraid, because she was selfish with how she felt.   this one\'s different, special. a lot like her and not at all.  it\'s almost amusing how eeriely similiar their fears are. three magaritas, odd eighties rock music and countless of scrabble games to pass time later, she knows this. wants this.  defies gravity, she stumbles and steps on stains, snakes, demons. two burnt defrosted dried roses, one off-white the other red; now pale okra yellow and the other a deep maroon, both entirely different but they look so well together. she twines the stalks together and presses poetry, prose on it.   creates art.  lets make music together; dance with me.  '),(22494,664485,'28,March,2003','2003-03-28 00:00:00','   it\'s funny how it takes an eating disorder, a conversation in the bathtub, scented candles and rational thinking to get her in this position.  she\'s going to be just fine..  really.  :)  to barker she says, i don\'t have the answers, and i don\'t know how, there are limitless choices to limitless problems..actions speak louder than words. i am hurt by your indecision but understanding of where you come from, i don\'t hold it against you but rather hope you\'ll come to a conclusion soon. wavering in between gets so hard after a while, the need to crawl back into your arms and push you away frustrates me...but i think i\'m going to be okay. actually, i know i\'ll be, it just makes me tougher i suppose...stronger.  i\'ll work on the movie with you, such close proximity might seem to bring problems but lets not throw this out of the window because we screwed up okay? work is work and play is well, play.  it\'s hard to handle hot and cold. it\'s hard to be secure when i don\'t know where you stand, becuase sometimes you seem so there and sometimes so far away it\'s hard to reach you.   the hardest part is knowing i\'ve been here before...and seeing how it ended.   ---  to the boy, ironic isn\'t it that right as i mentioned it, it all came crashing down. did you make the voodoo cutout you promised to make a long time ago? and if you did, damn, you worked it well.   want to know the odd thing? i\'m at the flipside of how our relationship used to be, and i see now the crap you had to go through. all this being afraid bullshit...pahhh. we are such stupid idiots sometimes, letting fear dictate us. sigh. i\'m stupid.   i\'m surprised we hung on for as long as we did. such stubborn assholes..  hm. so now not only do you drink and party but even do the socializing thing? yay! and one nighters...whoop. whoop. which i suppose draws back to the see-saw where i\'m out of it and you\'re in it, i\'m happy and you\'re sad, la di da, that type of shit.  men are evil but women? women are scary.   and i would know.   :P  and yes, life is a bitch.  ---  we came by your place this afternoon, to check out the apartments you know? i\'m moving out...and that\'s such a \'grown up\' thing to do. sure, i\'ve been \'independent\' the last few months...but to move out? move out? hm.  i didn\'t shave tiki though, and i left the \'hugs\' at home, so i guess i\'ll come around again another time.  i\'ll keep my yellow ribbon up for your brother okay?  thank you for caring, sorry i\'m such a lousy silly friend.   lets go to the beach and play music again sometime.   p/s and the answer is coke.   ---  your mother, father and i have been talking a lot these couple of days.  and i mean a lot.  and as your younger less achieving sister i say, would you please quit making the bar so damn high?  i miss you.  ---  old man, how did the show go tonight?  sorry i missed your call, fucking pilots lost again. too much commotion from the game and too many people around to talk. yes, \'talk\'. but i\'m okay now, i promise.   we need to go to waffle house again soon, i have a math review i need to get done by monday.   my waffle house thank you very much.  and maybe wal*mart too...what do you think?  :P  ---  the demons have rested, the girls are tired, the dragon\'s passed out.  time for bed.  time to recharge for a, hopefully, new day tomorrow.  '),(22495,664485,'28,March,2003','2003-03-28 00:00:00','   she paints in rituals to feel comfortable, to know what\'s expected and how things go yet there is a certain boldness to it, a extraordinary difference..space to intercorrelate with her demons to pry. to mess up things a while, to change how the circumstances are and to deny all plausibility.  she ends scrabble games and exercises her mind, keeps it fresh and humming yet she drops a class today. an unexpected decision, a shock to most...her? but why? the stress? but you\'re making fucking excellent grades and i don\'t understand? her teacher refuses to sign the sheet, pleads with her...some sort of understanding at least the professor says, but she only shakes her head and stares back down at the linoleum, piercing it with her eyes wishing the ground would create a hole she could sink in.  they say you can only run on emptiness for so long before you have a breakdown...she\'s getting close to that. she empties her  stomach, damages her throat and let the acid gnaw at her insides. the scars on her wrists from the slits a while ago have now faded, except if you look closely, except if she lets you come close enough. perhaps she\'ll even tell you about them...perhaps she\'ll even show you. let you in.  she pops in pills and searches for instant relief. what a fucking time to quit her addictions, to go cold turkey. not exactly a smart move but hell she\'s executing this to the end, she screams. stands her ground. the very same determination and bitch in her that disintegrates her being is the same one that keeps her alive.  journal entries from a year ago, six months, three days are all almost painfully exactly the same. hurt, betrayed, left, running, scared. all that perfect bullshit yet everyone from her past that sees her says she\'s changed and she doesn\'t know how to take it. changed good? or changed bad? she doesn\'t know...just changed.   hell, even the pictures look different.  behind her eyes of a more recent photo illuminates hope and certain jaded dullness. there is something brighter and more naive about her younger years...before she moved across the world, before she got her heart broken, before she tried to piece things together again. and she still hasn\'t, well not exactly, but well that her life isn\'t complete yet. that there are still so many holes and booby traps, so many unknown hazards in this journey and sometimes, she truly struggles.  she tries sometimes, yet sometimes, it\'s so hard to just do that.   to face another day nor to face herself.   people come and go, what makes this person so important, so special?   she doesn\'t have the answers....perhaps it is a gut feeling.  her biggest fear is latching on, substitution, clinging on to something that isn\'t what it is but making it to be.  is she capable of love?  should she try again?  '),(22496,664485,'28,March,2003','2003-03-28 00:00:00','   ever feel like you have so much to say yet nothing at all?  like when you come down to it, all you have is a whole bunch of words, a couple of random tears, and a shitload of emotions while keeping a little bit of common sense with you yet losing all sense of sanity.   it gets like that sometimes, and you can\'t worry anyone and try to protect the whole wide world so you cry at night, when no one watches. so you hurt at night, when no one sees. it isn\'t that you\'re being ignorant, it\'s two parts really, the first being you suck at asking for help, suck at not being alone. the second because you\'re afraid of hurting someone with your pains, afraid of making them worry, afraid of making them cry with you...maybe because you don\'t think you\'re enough, maybe because you aren\'t enough. i don\'t know.  i didn\'t want to push you in a corner tonight, i didn\'t mean to force you into anything. i know it seems like that, i replay our conversations, and i hate myself and the callous bitchy way i was, like the world revolves around me and i\'m the only one feeling icky. at least i hope i\'m not anyway.   i don\'t know what it was tonight. i went through all of yesterday convincing myself that i needed space, that i couldn\'t handle being withing three feet of you or having you touch me. hug me. because i knew it would hurt, and i would miss us, miss waking up to you or finding your arms around me at night. it would hurt to remember that we were over and if only i just pretended that things were okay, i could be numb and blinded, i could just smile and laugh and smile again. pretend. they don\'t call me a drama student for nothing you know? i wanted to get as far away from you as possible, maybe because i was hurting, i\'m sorry.   some part of me wants to work this out. wants to because it\'s worth and maybe, perhaps, we are too. but see, i can\'t tell you that for a fact, i can\'t tell you everything\'s going to be okay and we\'re going to stay strong through this. i can\'t tell you that there are not going to be moments where i want to run or you want to run, i can\'t tell you that i\'ll never intrude into your bubble nor can i say you\'ll never. but i know that this much is true, i\'ll make space, if you\'ll make space for me too. i\'ll learn to not piss you off so much, not irritate you...i\'ll learn to push the right buttons and attempt to make this work. i just need to know that you\'ll try and you won\'t give up without a fight. i need to know for me at least that when it gets hard, you aren\'t just going to run because i know how fucking liberating that feels....to run away from it all. i\'ve ran my whole life and frankly, i\'m tired of running. tired of not giving enough or giving too much.  yes, you did hurt me. perhaps mostly with the thought that i came into your bubble. that hurts the most, i don\'t know why. i think because i was fighting urges to run and it\'s painful to know someone didn\'t exactly want you around. that i violated some law i considered sacred. here\'s this dumb bitch, being a burden. i\'m sorry.  are there issues? yes. there always will be. there also always will be more excuses though. somehow, i want to believe that if we both want this bad enough, we\'ll make it through. whether i\'m american or not, whether i lose an arm in some freak accident, whether you grow blind, whether i move or stay or you move or stay...i don\'t know. at least i want to believe that anyway, i just need to know, you want this. but then again, that\'s prolly the million dollar question.   here\'s the deal big guy, we need to decide if we want this, and if we really do, then we can go from there. so maybe there\'ll be time where we take one step forward and three steps back but hey, you know what? shit happens. besides, whatever doesn\'t break you makes you stronger eh?  i don\'t know, barker.  you tell me.  i\'m tired of playing games and sticking my head up my ass. you know where to find me, my number and all that shit, you decide. loving me is a challenge and i\'m not going to sit here and try and sell myself or tell you lies, it\'s going to be hard. it\'s going to be shitty. and it\'s going to be icky.   but maybe that\'s not all there is to it?  i\'m here if you want this.  and if not, it\'s okay to walk away.  it\'s okay to just be friends.  it\'s a whole lot better than being unhappy.   and just for the record, it\'s okay to question. it\'s okay to have questions and i am not mad, or angry with you. it\'s only human, and self-protection mode always kicks in for preservation. and yes, i\'m glad we tlaked too.  take care you.  '),(22497,664485,'27,March,2003','2003-03-27 00:00:00','   i want to say lets try again.  i\'ve felt the bitter aftertaste of what it feels like to be on the other side, of trying to deal with someone else that is so much like me.   they say revenge is sweet and this is almost ambrosial. high lows, up downs, not mine anymore. spiraling into an abyss that i absolutely have no control over, his decisions, his wants, his needs.  lets go to the hockey game together, let\'s watch a sunrise at the beach again, come play at the bay with the dolphins with me. come talk to the angels and dance with the fairies, come lay down with me.  hurt only takes you so far, the fear of pain should not eradicate something that could be so magical.   there are options, solutions, problems and instances. there are real things, actual hard shit to get through, long-distance for summer, shooting in the fall, graduating in a year, moving to orlando..will you come? will you stay? parents, trying, us, fears.  afraid of letting go yet needing to trust, to let in. let me in.   i\'ll show you mine if you show me yours.  i need you to trust me just like i need to trust you.  i need you to learn to let me in just like i need to learn to let you in.  it\'s scary, it\'s enticing. it\'s good thoughts one moment and scary, suffocating moments the next. it\'s wanting to run but enjoying this at the same time. yes, i\'ve tasted that bitter medicine before, and just like before, it killed the relationship. letting fears consume, drive, and overwhelm.  maybe that\'s why i understand, maybe that\'s why it hurts, because i know you\'re trying, and you\'re confused, frustrated...and i wish i wasn\'t the source of that. and i wish i could be your friend, that i could put my feelings aside and be strong enough for you. support you and help you figure things out, make things nice for you, be there for you.  but sometimes, i\'m afraid too. and i get a little scared and insecure, i\'m pissing afraid you\'ll leave me. pissing afraid you\'ll come to realize i\'m nothing like you think i am, i\'m afraid history repeats itself and the patterns continue, that some day we won\'t be friends anymore and i\'m going to lose a friend, a companion, a lover.   and i really don\'t want that.  and so i don\'t know where this goes, or how things are going.  except that i like having you around.   and you make me laugh, smile, dance in the dark, and sing to the nymphs.  so nothing is perfect, so nothing is anything, but sometimes, i suppose, exceptions are made.  and nothing is ever as it seems.  sometimes, you just have to toughen up and try.  sometimes, we just have to try.  you think?  '),(22498,664485,'27,March,2003','2003-03-27 00:00:00','   \"sorry, i wasn\'t around when you rang? / sorry, i was in the shower? / sorry, the hairy green polka dotted purple monsters under my bed ate my phone? / sorry, i just didn\'t feel like talking?\"  i rehearsed what i was going to leave on your answering machine, convinced i was going to be lucky enough to get it...i guess i didn\'t expect you to ring again. or try. but then, in true fashion, you rang just as i clicked on \'send\'...and i wasn\'t expecting that.  and so i didn\'t know what to say.  and so i couldn\'t find the words.  i wanted to say, yes, let\'s go watch a movie tonight. yes, lets go climb a mountain. yes, let\'s go do something together...and talk..and have coffee...and all that good stuff...but that\'s scary.   and now puck\'s on the phone yelling about 50\'000 dollars and a pilot game, and here i go whining again.   making up excuses.   urlLink  \"If you’re not the one then why does my soul feel glad today? If you’re not the one then why does my hand fit yours this way? If you are not mine then why does your heart return my call If you are not mine would I have the strength to stand at all  I never know what the future brings But I know you are here with me now We’ll make it through And I hope you are the one I share my life with  I don’t want to run away but I can’t take it, I don’t understand If I’m not made for you then why does my heart tell me that I am? Is there any way that I can stay in your arms?  If I don’t need you then why am I crying on my bed? If I don’t need you then why does your name resound in my head? If you’re not for me then why does this distance maim my life? If you’re not for me then why do I dream of you as my wife?  I don’t know why you’re so far away But I know that this much is true We’ll make it through And I hope you are the one I share my life with And I wish that you could be the one I die with And I pray in you’re the one I build my home with I hope I love you all my life  I don’t want to run away but I can’t take it, I don’t understand If I’m not made for you then why does my heart tell me that I am Is there any way that I can stay in your arms?  ‘Cause I miss you, body and soul so strong that it takes my breath away And I breathe you into my heart and pray for the strength to stand today ‘Cause I love you, whether it’s wrong or right And though I can’t be with you tonight And know my heart is by your side  I don’t want to run away but I can’t take it, I don’t understand If I’m not made for you then why does my heart tell me that I am Is there any way that I\"    i don\'t know, ryan  i don\'t understand either.  close, but not quite.  damnit, someone find a solution.  someone find an answer.  someone find a way.  arsh.  '),(22499,664485,'27,March,2003','2003-03-27 00:00:00','   \"Well I blew it. I drove away from her in more ways then one tonight. I\'m no good at knowing what to say or knowing what to do. I broke a promise tonight also. I hurt her. I may have passed up the opportunity of a lifetime. For years I\'ve pissed and moaned about wanting to be in a good relationship. For years I\'ve whined about no woman wanting me. And here she came with a red ribbon wrapped around her and I returned to sender. What the fuck! I worried whether or not I could handle the relationship. I questioned my feelings for her. I drove away and couldn\'t decide whether I should be relieved or upset. I came home and looked at the mirror and I saw a newly emerged Jerk. So what do I do Is it Miser time or do I get her back?  Was she my last chance? The one perhaps? I don\'t know. But it was the same indecision that got me dropped. She is hurt and I am an ass. What do I do? I guess I sleep and figure things out later.   -SIDEBAR- I write subconsciously knowing she\'ll read. ;(...\"   ---  \"you wanted to know why i still go. i\'m not sure. do i care? do i want ammunition? is it just a habit? i couldn\'t tell you. i\'d like to say i\'m at a place in my life where i no longer have the desire to see you suffer, but i\'m not entirely sure that\'s true. i\'d like to say that being vindictive doesn\'t suit me, but in reality it suits me to a t. i don\'t know what i\'m doing, with you or my life or anything. lately i\'ve been trying to branch out and be more sociable. the me that exists now is more different than any other me you\'ve known. i\'m more of a \"guy\" now than i ever have been in the past. i\'m sure you can tell from the last two entries just how things are going. but then there are nights like tonight, where it all comes crashing down. i realize that the things i\'ve been filling my life with are empty. empty, yet at the same time somehow fulfilling. i have a lot of fun. it seems like for the first time i\'m actually living my life. and while alcohol and no-strings evenings are fun...in the end my heart still only wants one. i\'m still lonely and in the end i still want a companion. and of course, i\'ve picked someone far out of my league and currently taken. i never was a smart man, and alcohol seems to be killing what brain cells i have left. the best part is that my best friend knows i like her. he also knows that she finds him very attractive. so he realizes that he can have what i want at nearly any time. while he wouldn\'t, he likes to fuck with me and push my buttons on a daily basis.  i don\'t know about any of that healing bullshit. getting calloused isn\'t really healing. something bothers you so you develop a thick skin. i still have to skip everything about him. it still gives a little white hot flash of pain. a brief flashback. quickly forgiven, but never forgotten. i\'d like to say that i wish you both the best, but i\'m not sure how sincere i\'d be.  i don\'t know what else to say. i\'m not the man you fell in love with. and yet, i\'m also now more that man than i ever have been in my life. healing? moving on? it seems to me that we\'re just running around in circles.  life\'s a bitch\"  ---   i should have known, i understand. i know fear, i know about running away. retribution. ironic how she steps forward to talk about it and it bites her in the ass.   her world comes crashing down and \"her girls\" go into a frenzy.  school. barker. her family. going home. medication. being sick. war. parents. moving. trying. shitforbrains.   same old brand new shit.  it all overwhelms her, too many things going on at once. too many obstacles, too hard a course, and she wishes to just escape reality a while, and she wishes to just go away. to not be her. to not face the world and shooting grounds.   the hardest part is having to realize, she wasn\'t even expecting this. this fall, this crash, was not set up...and so it hurts like a bitch. and so it\'s six times harder. just like the eight hundred and seventy two million things in her life that need to be solved right now and then and need to be figured out.  well fuck figuring out, and fuck her.  yet she knows she set herself up for this one.   she really, really should have known.   she\'s using up space.  she really shouldn\'t be here.   how fucking repetitive. a pattern, she has a problem. the problem is loving her.   she needs to go away.   she craves for an escape.  afraid that when she starts, she won\'t stop.   she cries.  weeps.   sobs.   whatever.  '),(22500,664485,'25,March,2003','2003-03-25 00:00:00','   he rings.   i told him to come back when he was ready, to get his shit settled then come home and he tells me he wants to. i walk by aisles and aisles of clothes and shoes but i walk empty handed, something that used to make me so happy isn\'t even working anymore. i only see faces, and signs, and emptiness etched in everything i touch, feeling when i really shouldn\'t.  i feel the phone vibrate against my skin and fight back pangs as his name flashes across the screen. i walk faster, a little hyperventilated and fidgetty, click on silent and walk again. run even. and it beeps to tell me to check my voice message, and it beeps to tell me he\'s on the other side. the other line.   i hide but i need to know, i want to know. does he hurt like i do too? is he as much in pieces, afraid to cry, scared?   i listen. i click. i press \'send\' then question what the fuck i\'m doing but the dial tone soothes me; this connection, this knowing he\'s not so far away, this knowing he\'s on the other side and i feign happiness and sarcasm. i smile and laugh and make conversation. he poses questions which i only avoid, he iniates conversation but i have nothing to say, afraid to say. afraid to put out again.  you\'re never had a three o\'clock have you? you\'ve never cried yourself to sleep and woke up to tears, you\'ve never been so afraid of sleeping, of being alone, of having to face that few minutes with yourself before you fall asleep. never learned to love someone, let someone in, let him change you, let him affect you, let him into your life and make space for him, need him, never come to a painful, scary, truthful realization that it\'s not safe anymore. that he hurts you. that he\'s changed. that you\'ve changed. that things aren\'t okay and you hate yourself, and you hate your being and you hate everything there is to hate about you.  tired of being an adult. tired of being me. wanting escape, wanting to runaway a while, wanting to go somewhere, go someplace else, wanting to be oblivious. wanting to pretend everything is okay and all will be right. wanting to be strong but not strong enough. wanting love but not capable of being loved. wanting space but being afraid of being alone. wanting to try but scared that you\'ll give up.  i\'ve let you break down walls, make love to me, come into my world. the hardest part is knowing it isn\'t quite the same for you, that you didn\'t specifically feel the same way, perhaps that hurts the most because here we go again. letting someone in, i really, really should learn my lesson by now. i\'m not worth it. i\'m not enough. i\'m not anything as far as anything goes. when will my thick ass numb skull learn this?   you miss me you say.  i ask how it\'s any different.  you say i\'ve gone away.  you\'ve left too.  you drove away.  i stayed.  i\'m not strong enough for this shit.  not good enough.  '),(22501,664485,'25,March,2003','2003-03-25 00:00:00','   twist, turn, hurt again...ring old man and say, \'daddy, it hurts,\' then draw back and question. hiding, covering..\"don\'t want to talk about it\", \"i should have known shouldn\'t i?,\" \"how come i don\'t bloody learn my lessons?\"  and he says, \"i don\'t know, honey\"  and i say, \"me too.\"  '),(22502,664485,'25,March,2003','2003-03-25 00:00:00','   so things are a little chaotic at the set of deliverance. too many dragons to slay, too many demons to silence as everything that once touched was gold collapses into shreds of steel that screech with every flake. an old brass, a tarnished silver, a worthless unpolished, uncrafted gold...the sirens are wailing, the walls are whispering, the voices are screaming; get me out, turn away, search for an exit, runaway.  the bay soothes me as i fall into piercingly cold water and swim. i swim to no one and into the heart of the ocean, i kick and splash and splutter in my flabby fat legs and my washboard chest, i drown and play with the dolphins and the mullets and the crabs. perhaps even the invisible sharks that never come to play, the ones that he convinces me are there...sometimes i wonder what it would be like to die ripped apart?  you would be proud of me today, i did not fear death but rather, welcomed it. hell, the both of you would have been.   you, who i awake to and fall asleep to at night. you, who makes love to my body and soul and calms me down...rationalizes the fuck out of me, smiles and laughs the crap out of me. i talk exhaustingly of the art you create in my life, of the works you thread into my quilt and the colors you introduce to my life. i blog, i rant, i scream, i dance to your music, to your voices, to your hopes and dreams of us but yet i stumble, and find myself searching for the nearest available exits, looking for signs, searching for independence and not letting go but holding on to familiarity.   i find myself turning the other way at night, putting my back towards you when the voices haunt late at night.  \'you\'re not good enough, he\'s blinded, you\'re ugly, run before you fall in too deep\'  but then your damn arms always hold me tighter and your damn lips always find mine, your eyes always shines with a smile and your heart always loves like a child. innocently, unwaringly, naively towards all the damn hurt i could cause you, towards all the damn shit i could put you through, towards how unbelievably impossible it would be to love me.   you smile, even as my mom comes off her spiel of us sleeping together. you say, it\'s okay, that it\'s alright, that everything\'s going to be fine. you take blame even if a good half of it is my doing and damn, you look so bloody good in a shirt and tie. stepped off the pages of gq and into my life, she says we look cute together and i say, me? like hell no.   you look like you could be with some beautiful exotic chick with long thin legs, ravishing shiny curly hair and a smile that could dazzle a million hearts. a woman that could launch a thousand ships. and then i wonder again, what the hell are you doing with me again?  but sometimes though, sometimes i think we fit like spoons. the way your hands curl around my waist just right like gloves or the way i\'m not tall enough but short enough so my head lies on your chest and feels your heart beat. the way i sometimes know what you\'re saying and even if i\'ve used seven hundred unnecessary words to explain something, you\'ve understood it with the first word.   you make fun of my accent and the funny way i dissect my food, you poke at my funny views on life and take charge of my odd ways. you don\'t always say the right things, and you won\'t always do them...but the greatest part for me is knowing that you will always try.  thank you.  ---  perhaps then there\'s you. old man.  person that started me on this journey, person that has helped me along the way. the one that wipes away my tears and soothes the sobbing, the one that tells me i\'m beautiful when i\'m nothing, the one that listens to me bitch, whine, assholically rant about my problems and my needs and my wants and my everything then sits down and figures them out for me. a herculean task, an impossible god damn fuck me now thing to try and keep her happy, try and keep her sane, try and fix her life but you courageously do it, but you willingly do it. why?  perhaps i\'ll never know.   perhaps i\'ll always wonder.  perhaps i\'ll never understand the way your fragile mind works nor the voices in your head i ignore half the time because i\'d like to talk about mine. you show me wrong, then right, then the in-between. you show black, then white, then the grays. hell, i can say, for sure, that you\'re the man in my life that will call the cops on me, cry when you see me, make fun of me, laugh at me, and tell me i\'m a stupid retarded asshole all in a span of an hour.    you make jokes and laugh but i know that trick too. the yankees won because they were distracted by the stripes, we\'ll win because they\'ll be distracted by the smiles, and the laughs, and the pretend-to-be\'s.  don\'t hurt yourself old man, i\'ll listen. i care.   i\'ll be your three o\'clock if you\'ll be mine. the mother is here and i am excited and intimidated of the prospect of the both of you meeting up. like barker, i wonder what it\'ll be like...what you\'ll think? what she\'ll think? what you would say? what she would say?  i tell her about how you wait and take me to the head doctor when puck is busy, how your tired ass sits in the car and waits, then waits somemore, or how you sit with me at waffle house and i yell at you because you don\'t know how to do my bloody homework too, or how i drag your sorry ass for knowing me at three, sometimes, four, even six, in the morning for a cup of coffee and a smiley faced waffle at MY waffle house and not the plaintiff\'s nor yours. i talk about how i make you sit through countless and countless hours of varying monologues making you search for them to find the right one for me and find the right tone, the right voice, the perfect expression because racheal khong has to have an a and nothing else is acceptable.  i don\'t tell her about the cops though, but i suppose that will be a great story to tell the grandkids someday. it might give her a heart attack or she might die laughing...of course, i can only laugh about it now eh?  i think about how insane it is that i\'ve only known you for a couple of months and i wonder what it would\'ve been like if i wasn\'t casted, if i didn\'t go to the auditions, if i wasn\'t in the musical.   scary thought but what ifs always are.  i love you old man, you take care now.  ---  everyone would scream at me for talking to you but i want to. i don\'t know why, i\'ve been thinking about you...i always seem to don\'t i? i suppose you did hurt me quite a bit, inflicted a little bit of damage they say but as always, just as my blog mirrors yours, i stirred up some shit in your life didn\'t i? changed you, you say...made you into the being you are now, created hate in a place that once had love and i know you feel so damn sorry for the boy...you snicker and laugh and point fingers. wondering what the fuck i\'ll do to this new victim next then not bother at all.   the truth is, i don\'t know either. sure i\'d like to tell you we\'ll make it, sure i\'d like to tell you i\'m more secure than i\'ve ever been, that my meds are fixed and i\'m almost okay, but then that condescending cocky route always leads me into trouble. always gets my ass burnt.  sometimes i wonder what the hell you\'re doing here, why you still come in...and sometimes it bothers me. bothers me because you care or bothers me because you\'re laughing at everything i say and mocking every word i come out with. i don\'t know. the phone rang the other day, and she asked about you...i laughed it off and said you were busy with the perfect woman that\'s slipped into your life now, then that section ended with me hoping that you were okay. knowing that you were doing better, getting better, moving on.  some days i wonder if we\'ll ever speak again, somedays, i don\'t want to.   but nights like tonight, i wrote to just say, i hope you\'re doing fine. i hope you\'re okay. i hope you\'re happy now.  nights like tonight, i wrote because i couldn\'t call, couldn\'t mail, too personal, too real, too there.  you walk three lonely miles and turn back to notice, you\'ve changed a lot since then. you\'ve grown up, but you\'re still the same person, and you\'ll always be her.  some part of me will always want to be somebody else, will always strive to be somebody else, yet at the same time, i know, some part of me will learn to love me. and accept me. and forgive me.  so anyway, this was half your post. almost directly, almost to you and you would know it too. you\'d know it was you without having me say your name..so it\'d be out there. so i wouldn\'t have to hide. wouldn\'t have to pretend.  we\'re moving on. we\'re growing up. we\'re learning.  we\'re walking, slowly and painfully away.  we\'re healing.   separating.  going our own ways.  ---  there\'s so much talk about next term, about whether i\'ll come back or not, or move out, or move back, or move in. there\'s talk about the war and medication and growing up. talk about school and travelling and trying harder, immaturity and maturity issues, of hiding and pretending and not letting them in enough. they sometimes just ambush through and my defenses aren\'t hard enough but i think perhaps sometimes, i want them to come in, i just don\'t know how.  and so another conversation ends with my father, and he says we love you again. collectively as a unit even if my mother is in the other room, they function as one. separated by far too many oceans, their love still prides me. twenty five years and they\'re still going strong, perhaps that\'s what it takes to keep this family going, perhaps that\'s what it takes to keep her going. and her dragons. her demons. her fears. her wants. her needs.  so goddamn selfish racheal.  pride and joy my ass as she does another stupid thing that pains them. as she runs the water and throws up again. as the pain washes over and she breaks her pills again. as she does another stupid thing and her mom says, why can\'t you be more like your brother? why can\'t you grow up? why can\'t you just color within the lines?  i wonder sometimes.   trying to be different i suppose.  trying to be me.  being stupid.  being worthless.  being a bitch and selfish.  being retarded.  sorry mom.  sorry dad.  sorry koko.  ---  the walls raspy whispers awakens the sleeping child; she searches...moans...screams. awakens, in deep slumber, she startlingly stumbles around..searching. slaying her dragons, silencing her demons, she ripples the threading earth.   she falls.  '),(22503,664485,'24,March,2003','2003-03-24 00:00:00','   it\'s that giggle you get when you\'re being silly, just a tad out of line you know? a little extraordinary feeling that makes you do the \'wee-wee\' dance as puck and blue-shirt have the longest conversation ever known to mankind and he gets just a little more inebirated with the girls. they drive off to go stalk barker\'s ass and sneak tulips to his car, watching out for the capped boy in box and making sure the cops don\'t get rung on their asses.  she\'s a little tired, a little cranky, a little defensive.   a little out of line and in need for relief. numbness. distraction.    cookies don\'t turn out, dinner\'s left a mess to clean up and school fucking starts tomorrow. an empty house she walks without clothes on and dutifully does her chores in pure commando style only that there is no cover up. she thinks perhaps there is subtle and conscious decisions to do the things she does; to get comfortable in her skin...to shed off layers. open up, strip, be loved, love.  finally gets to his post, his email, his things as words wash over the screen and her heart dances.  and yes baby, i like.  (woof.)  :P   \"I am expecting a different set of experience in Pepsicola but not less interesting and enriching so who cares about what war.  I am going and going with full of joy to see my only daughter.  She will give me the usual set of headaches but I love her all the same just like I love my lor koong passionately and my only son dearly.  You are all that matters so I’ll hop around and find my own piece of fun.\"    her mom arrives in a couple of hours, she\'s unsure about how she feels. excited, intimidated, nervous, preoccupied, retarded...ready for changes yet happy to just stay with familiarity..the coming together of what seems right but feels wrong. strange. of medication, and stepping up, and try, try, trying again.  he\'s excited to meet her, decked up and ready to go. so is old man and a bunch of her friends, excited to show her how much she\'s grown, moved on..lived independently.  last night, she questioned. she broke down and had a good cry...the first in a while. it\'s been a while hasn\'t it? since she\'s looked in the mirror, since she\'s had to be with herself...her mind...her soul.  \'let me in, let me in\'  she thinks about it constantly....where she would have been had she took that flight. had she stayed. had she not come here.  she wouldn\'t have met barker. or puck...and her hair would still be brown and not the flaming orange it is now which hopefully, she\'ll get fixed by tomorrow...just a surprise for the mother you know?   she moved here seven months ago, and in seven months she\'s done quite a bit hasn\'t she?   she remembers the way his fingers trail her skin and she shivers...because this feels so real? because he seems so honest? because they genuinely look like they could make it? i don\'t know.  but there\'s something about the way she grins when his name flashes across her phone, or how secure she feels when she falls asleep in his arms. something about the way he holds her hand and holds her tight, softly, gently and firmly all at once. something about the way he makes love to her and doesn\'t just fuck her, doesn\'t just make her feel like an object or a toy.   there\'s something nice about knowing she has the icing on the cake, that she\'s the lucky one. that at the end of the day, they can sit and exchange stories and not get overbearingly protective or jealous. she laughs sometimes, when she\'s alone remembering the way his face lights up or unexpected surprises along the way.  he\'s hot. so fucking sexy and adorable at the same time. but more than anything else, she respects him, feels up to parr and sometimes even lower and not have that pressing need to hide things, or pretend. that it\'s okay she\'s done the things she has, that she is the person she is, background, job, person, whatever. it\'s just a bunch of hogwash with them because what matters in the end is what they are made of, what their heart, soul, and mind represents.  learning a little bit more of herself, him, and the both of them together as days go by. the things they like to do together, apart, constructively. as summer rolls around and he starts working on the film obsessively and she goes home, what would things be like? or shooting, the spaces between, schedules conflicting and tired cranky bums at the end of the day. moving out, moving in, will things be easier? harder?   something about his willingness to make sacrifices scares her and makes her heart soar at the same time. for me?   you\'re retarded. stupid. silly.  all mine.  :P  '),(22504,664485,'23,March,2003','2003-03-23 00:00:00','   urlLink      i am still the same person, stuck in a different body, a different face, a marred figure.   i am still beautiful in the inside, just as i was before i got hurt. i still have the same name, the same eyes, the same body...perhaps just a little different. but you see, i\'m still me.  i talk funny, look funny, act funny, it hurts to touch, to feel, to see sometimes. it hurts to accept.  so tonight, as you drive home from a good weekend, as you celebrate the coming of spring and have a couple of beers, please drive safe.    puck, madonna, curly sue and maria cruise around picking up hot young things. unsuspecting victims to night filled with crazy endeavors of waffle houses and sworn insomniacs. of course it is the coldest night out but the top still  stays down, just so there\'s still access to random shouting and catcalls. wolfwhistles and offers, the \'strip\' is busting, flowing, bursting. knickers are in a twist, girls are out to party, the waves roar and crash in a distance, the jeep hums, the people dance. stalk. pick up. drink.  they curl up behind, madonna\'s head on her left and puck\'s on her right. relaxed, screaming, singing. it is an interesting night but she notices again, how careless people are with drinking, how horridly pathetic the need to get high overcomes the need to be safe.  if only someone would explain.  if only the kids understood.  if only we\'d mature.  if only.  '),(22505,664485,'22,March,2003','2003-03-22 00:00:00','   broken hearts, weeping tears, a united family tearing...crying. a lost son, a missing father, a dying soldier.  we are a proud batallion. we are a strong nation. we are a rallying world.  too painful to watch, to hard to not see. go out? runaway? ignore?  only so far, only so long before you come home and watch obsessively again. judge. formulate. try to make sense.  hurt by one of your own, a cameraman, a marine officer, a newbie fighter...too stressed? just cracked?   fucking betrayed.  i finally agree, or see rather, the need for war. accept...but yet my heart bleeds. everytime a plane goes down, the phone rings. was that kirk mom, was that kirk?  special forces.  kuwait.  you\'re fucked being one of the best in your field.  but you stand proudly defending your country, for the stripes and stars, for freedom that stems from the brave.  this is my dream, my passion. to stand at battlefield and report, to understand, to enlighten. to bring coverage, to stand in dangerous ground, to go places no one wants to, to challenge the unknown. there\'s something sinfully guilty about, something disastrously beautiful.  i am afraid of death, but it is that fear that drives me. that need to achieve all my hopes and dreams, that need to do the things i need to do. before i die. before i walk through the pearly gates. before i meet cerberus. scares, i woke up this morning and willingly took my medication...on time for that matter.  i sat on the countertop while he made coffee and we talked. perhaps there was something very enticing about the aroma but it was a good feeling as he told me more stories and opened up a little bit more. the strange amazing intoxicating part of this is that we\'re still learning about each other and it gets better each day.   walking out the door repeatedly for about the next five times in two minutes, each time returning his silly little ass forgetting about something while i stole another kiss and the jangle of his keys led to the door. tapped on keys, a short stoic post, a scrabble game, chatting to random strangers and replying email, i watched the short hand touch six then seven then crawled back into bed at a quarter to eight.  hid back beneath sheets that smelled of him and fell asleep.  drifted off to fairyland and dungeon monsters.  waking back up to him opening the door...smiling.  \"hey you.\"  it\'s a big world out there, the girls are going out to party at the beach. spring is finally here, the shops have opened up and the boardwalk comes alive. so go out, live life, dream....stay safe.  come home.  live.  '),(22506,664485,'21,March,2003','2003-03-21 00:00:00','   we\'re at the boy\'s house again, a night of pure intense horror, a scary gory pissingly bloody but good movie; something to keep the heart pumping i suppose plus a night out with her friends while barker joins them for a while before going home. he\'s had a horrible night and she irks for him, wanting to make things better as his fingers find hers in the now familiar darkness of this quaint cinema. a single forgotten rose slips in between as they watch the ending together; her in his lap.  she slips in brownies specially made for blue-shirt, the silly girls swirling caramel and figuring out between olive and vegetable oil. sleeping in tranquility, singing to puck tinkering on the grand piano, twanging between different rhythms and sounds. making music together before puck horribly plays and sings to a long, long, thankfully forgotten song.  getting dressed together, running around in knickers and bras, sorting out different cuts and styles, jeans and pants, capris and hipsters, chokers and chains, make-up and lotions. television blaring, stereo pounding, cellphones ringing, chaos is welcomed in this world.  a great night out and it\'s a quarter to half past three, she comes \'home\' to him. stroking onyx on his lap and she purrs and her striking eyes demands territory. candle lit, music, monitors purring..he\'s working on a couple, she on his. searching for words to describe her feelings right now; they share an ashtray. a quiet that\'s nice, random words shared in between: it\'s kinda nice how they can do their own things together. yet it\'s beautiful, sweet, perhaps almost \"perfect\" in some sense. a dejavu like feeling washes over her, the security she constantly bitches about, the one she craves, it\'s here now. she\'s found it.  '),(22507,664485,'21,March,2003','2003-03-21 00:00:00','   it\'s when you can make love slowly, seductively and feel every solitary sense on every inch on your body come alive that perhaps things might be okay.  when he grabs you forcefully, like a crescendo in a masterpiece as he climaxes and his eyes rolls back, his mouth opens and lips wet with desire. when he strokes you to sleep and cuddles up next to you, protecting arms shielding away from unnecessary thoughts. whispering lyrics, as sleep washes over you and you are transported into a surreal dimension, a comfortable place...a place you\'re beginning to know. to learn about.  when tears stream down your cheeks and pain deep within spells out fear. fear of intimacy, fear of commitment, fear of leaving...fear of seeing. it\'s when you\'ve heard all these words before and trusted all this sincerity before and you get confused, and afraid, and sick. it\'s when you\'re trying hard to not be sick, be nauseated, it\'s when you\'re making a very painfully conscious effort to digest food and it\'s not sitting well in your tummy. it\'s when you know how much it would mean to him...how much it would mean to you. but it gets harder, and you are angry with yourself because you have become so weak, you\'ve let your addictions eat you alive. tear you down.   you peel away into a corner and sob quietly, softly so he doesn\'t hear you. so he doesn\'t get worried or so you don\'t lock him into a corner, a place where he has to say something nice about you, a place where he has to tell you you mean something, a place where he is forced to say things and make promises he might not keep.   and so you don\'t show, and so you weep alone.  except of course for a phone call to the old man. a call of desperation and you whisper so the man on the bed doesn\'t wake up. he\'s had about two hours of sleep in the last few days and has had to work his ass off so you want him to get as much rest as he can possibly get. it\'ll be another long day at work tomorrow, it\'d be another twenty-four hours to worry about product placements, and cast meets, and drop-outs, and responsibilities and that damn stupid girlfriend of his.  the phone dies right as you say hie, and hello, and how\'s it going. but by that time, the old man has heard your voice, heard that scream and cry for help beneath the surface. he\'s read between the lines but your phone\'s died out. the battery is dead and you are too naked to walk your yellow ass down to grab the phone, too shy, eyes too puffy. so you wipe away the tears and curse at the phone, reach for your last cigarette and climb into bed with barker.  his sounds. his moans. his whines.   his random hugging and repositioning, his arms that find their way around your waist and his head that grows heavier and heavier on your side as he slowly drifts off to sleep again. or the way his fingers find your hiding place and want to play. the way his lips find yours and the soft kisses he pecks on your forehead. stroking your hair. falling asleep and waking up together.  then comes interesting.  then comes spectacular.  then comes irony, drama, coincidences, and irresponsibility.  the fucking police are at the door.  911 is on the damn phone.  eight new voice messages on the cell.  identical twin brothers with mistaken identities and stats copied. taken down.  she\'s somewhere. about 5\'3. 108 going to 98 pounds. chinese. malaysian. single dimple of the left cheek. talks rather retardedly and looks quite ugly.   where is she? i think she\'s in trouble. please find her, find her please.  is she alone? is she scared? is she lost? is she hurt? did she take more pills again? is there a bathtub full of blood? goddamnit racheal, what the hell have you done?  other parents on the phone. grown men talk. amanda. lindsay. bryan. ryan. marc.  oblivious to sounds, crashed for the night, thrown up, tired, sick...she sleeps.  24 hours for a missing persons report, no probable cause to have a search warrant, last whereabouts....like pussy and her cell they come together; she\'s not picking up. where are you little girl, where are you?  finally the cops come pounding but the house is empty, a hit, but the twins come home and thump loudly on the door. awakened, startled like the night before and that horrid dream. a name mentioned she remembered, throws on clothes and runs downstairs leaving beauty to sleep on the bed.  screaming, shouting, crying. she\'s safe, come over, we found her, she\'s alright, she\'s alive, she\'s in one piece, it\'s okay.  call made to dispatch, no baker act. no baker act. she\'s fine. just a misunderstanding, just losing grip on reality, just pure simple immaturity.   she goes back up to get barker. baby, i think we\'re in trouble. baby, i think i just fucked up. baby, come hold me.  he is....i don\'t know. i can\'t read him right now, he is confused. upset. a little angry. a little agitated. a little annoyed. he is barefoot, standing on rocky pavement in t-shirt and jeans holding me, looking at things with patient eyes. simple he says, it\'s going to be okay. but i\'m a drama queen, and all five people on that sidewalk are drama people. theater, film, everything.   nothing is simple.  nothing is black and white.  racheal needs red, and oranges, and bright bright pink.  his eyes look so tired as i look back at the old man and see that his too is weary. worried. madonna and puck? i don\'t know.  i come home and finally plug the phone in.   i talk for a long, long time with the other parents. i talk about moving out and barker fitting into my life. i talk about the old man and how he\'s saved me too many times. i talk about trying to take my life and the war and being irresponsible. i talk about the night before and how he walked in on me hurting myself. i talked and talked and talked and talked. talked till there was nothing else left to say, you see honey, it was quite simple...just tell the truth.  it was so much easier.  rather than pretending. rather than lying. rather than concealing.  i\'ve just listen to the voice messages. my birth dad left me one just to see how i was doing. it said, \'hi racheal, papa here, can you please call me back? we love you,\" and it was simple. it was kind. it was daddy.  papa.  she hasn\'t called him that in a while...   even puck said love you. and puck never ever says love you voluntarily..and so that meant a lot. and she\'s never felt so loved in her life. so special. she made people worry...she didn\'t mean too..people actually worry about her?  she still has fears. doubts. things to be afraid of..the future, barker, school, coming back, the war, getting cured, all sorts of good shit to blog about. all sorts of things to scare her with...but you know what?  she\'s going to be okay.  she\'s going to be just fine.  with or without, can or cannot, does or does not, will or will not.  it\'ll be right.  i\'m alright.  :)  thank you.  '),(22508,664485,'20,March,2003','2003-03-20 00:00:00','   we argue about \'making love\', \'fucking\', \'sex\'. would you rather have, \"baby, please make love to me,\" or \"i want you to fuck the shit out of me,\"...then we laugh it off, cuddle again...your sexy very naked body leaning against mine...complimenting my frame. making our reflection look nice. wonderful. almost kinda amazing.  you make me feel beautiful and special at once, a gem, as if i could be so precious and valuable. so worth holding on to...i feel your fingers caress mine and the way it spells out genuine all over it, the way it spells out trust and i am happy. and i am loved.  you entice me. you make me wet in waves of ecstacy and delirious with desire. you stimulate my brain muscles and force me to question, to think. you replace cannot with can and must not with must have. you look me in the eye and tell me black from white, real from fake and draw me away from the grays. draw me away from \'me\'s.   you keep things simple, drama-free. you keep things in order, but not in an anal controling way...you are like the place i come to for balance, for stability. the place i come to to feel safe even if sometimes, letting you in scare me. you don\'t push away my fears but instead help me figure them out with you.   i love how your arms find their way around my waist and how your breath slows down as you fall asleep. how you slowly get softer and softer and the lyrics not quite audible anymore as my eyelids get heavier and heavier and my heart dances to your song. your rhyme. your beat.  he sings me to sleep.  '),(22509,664485,'19,March,2003','2003-03-19 00:00:00','   it\'s spring break this week, a holiday, a celebration...a break from shit in school. academic hoo-ha, preparation for the finals and time away to do things for your own. students walk the streets, the beaches are filled with party goers, light shines in their eyes, smiles abound, laughter rings through.  puck\'s just come back from georgia, a road trip to a theme park somewhere, and she screams with her stories as i go to pick her up. we talk then stop by barker\'s workplace to leave a rose on his car, so he\'ll find it when he gets off work, so maybe it\'ll put a smile on his face after a long day at work. we stop by the mall for dinner, more news on war but it\'s not \'war\' yet. yet.  we\'re driving, singing, laughing, talking and suddenly, it\'s real again.  cellphones ring off the hook.  come home now.  come home now.  come home.  there is a lump in my throat and tears continue to stream down her face, it hurts, it\'s scary, i\'m afraid. i find words but i don\'t know how to comprehend, i can only pray, i can only just continue driving. take saufley, take michigan, left at bauer, right at kaoz, get home. hide. i don\'t know.  mom don\'t come in, daddy needs you. i\'m okay, i\'ll be fine, just get off that god damn plane. tense, so fucking tense in london my brother says, jittery, scary. waiting.   waiting to be bombed, waiting for a nuclear attack, waiting for september eleven again.  stay safe, buy duct tape, act paranoid, wear a mask; i don\'t give a fuck just please please stay safe.   cnn, bbc, am radio, fm radio, the news comes pouring in. the calls come ringing through..but my thoughts don\'t formulate. my mind so fucking scattered, my heart so fucking jitterry, i sit here with half a mug of cold coffee from this afternoon with the tv blaring in the background, staring at my screen trying to comprehend.  what happens now?  why is this so real?  why does this hit so close to home?  and i curse myself, for every time i was rude to a navy boy. for everytime i snubbed someone from the military. for everytime i took freedom for granted, for every time i put me before others.   this makes you stare yourself in the mirror and say, what have i achieved? what can i do? where do we go now?  this makes you want to curl into a ball and cry, and weep, and sob for all the lives that will be lost.  for all the terror that will reign through.  for all the crap christina\'s five year old eyes have to see just because greed and evil has clouded our judgement. that in two thousand and fucking three, we still need missiles and war to dictate our teritory. that we have resorted to war.  i\'m not against or for, i just wish it didn\'t have to come down to this.  i just wish this wasn\'t happening.  '),(22510,664485,'18,March,2003','2003-03-18 00:00:00','   it\'s when you check your voice messages and someone leaves you a message about home, about grandma, about terrorists and the first thing you think about is are they safe?  when war apparently is supposed to break out in a couple of hours, you shield away from extremities and hide away from talk, like september eleven you keep your own private pains and maybe voice out once in a while, each time remembering the horror, each time crying for the unknown, each time feeling fear run through your very veins and numb you.  it\'s when you know your mom is flying and you are afraid. afraid because you know she doesn\'t want to take that flight either, but does it for you, because your needs, your wants, you were more important. what if? you try and not think about it...but what if? you stupid selfish bitch.  singapore, the neighboring country. your home away from home every holiday, half your family, grandma. mysterious illnesses and you cry. you cry because you\'re afraid, you cry because you don\'t know what\'s going on and the world is too damn big. you cry because you\'re angry, because goddamnit, that\'s my family.  you cry because you feel threatened, because intercepts at the friend of the family that works at the fbi feeds you horrible things but doesn\'t say anything, private and fucking confidential with only a note to say, carry a mask if you fly through. fly through? you stupid idiots, my grandmother lives there, my cousins, my aunts, my uncles. are you stupid or crazy or just so god damn dumb? they\'re there and it\'s real...it\'s so damn real.  we\'re malaysian. i\'m malaysian. i come from a muslim country but i\'m chinese, it\'s fucking peaceful and tranquil there but no one knows. everyone is ignorant, everyone holed up in their own little worlds and no one cares to understand, no one cares to listen. i am not a terrorist. i am not a \'threat\' just because i understand this culture better. how do i explain? how do i make sense?  there is fear in the adult\'s eyes. there is talk, there is anger, there is fear.  fear for children on the battlefield, fear for still naive eyes that might mature, that might be forced to grow up too fast. fear for the bodies that will return, fear for the ones that won\'t come home or those that will with missing arms and broken hearts.  my dad\'s home alone.   my mom\'s flying between two kids.  my brother\'s in london. the capital. another threat.  and i am here.  with news reports and constant growing fears.  i don\'t care about stupid artificial superficial shit anymore, just please let everyone be safe.  just please don\'t take anymore lives.  just grant us peace.  please.  '),(22511,664485,'18,March,2003','2003-03-18 00:00:00','   does this scare you? do i scare you? did what i say scare you? will this scare you? he asks in between reasonings, questions...vouching. venturing into space unknown and afraid i suppose, worried.  me too.  old man points out how she\'s always had that problem with intimacy and running; that bugs her. that intimacy thing; he feels so good next to her, in her, with her..but yes, it does strike her mind sometimes to run. to step away. to walk away just in case...hurt them before they hurt you, you know? leave before they leave.  she loved the feeling of waking up next to him, bare skin on skin and his fingers stroking her hair, kissing the tops of her nose and forehead...holding her in their little cocoon of security. falling asleep to his rhythm, his breath and heartbeat and the warmth of the sheets and his body.  she craved his touch, the way his fingers would slip into her and his tongue explored her. liberating, she felt free. she didn\'t let go, but she wasn\'t held back either..she stripped and took pride in it, if only because he made her feel beautiful. special. if only because he made her feel safe.  almost.  his words, the passion that drove him to reach his dreams seduced her. his maturity, his kindness, his thoughtfullness, his generosity..the fact that he had hopes and wants and needs and was not another \'random ass\' intrigued her. that he wasn\'t from the navy, or a forty-year old bastard hanging out at a college club nor her normal attraction to the bad boy asshole with an attitude that makes her think that\'s the only guy she\'ll ever deserve entices her.   and as day light rays shone in through blinds, their once only candlelit room became a sanctuary, a retreat to hide awhile. bright harsh morning light, light that would underline her flaws, light that would let him see her...it was strange awhile, ackward...but it was okay. like her \'neck thing\', it took time. trust.  maybe she was beginning to...trust him. was that possible? was that  right ?  she wonders sometimes. but then she remembers the car pulling away and the look of his face in the driveway, or how he notices the little things in the big scheme of things, the way his face looks so peaceful as he cuddles up next to you and you think, holy shit, he\'s so calm, so everything, so happy, so excited, so silly, so goofy, so cute...because of me?  and then she figures  ...just maybe.  we\'ll see.  :)  '),(22512,664485,'17,March,2003','2003-03-17 00:00:00','   hm..so i\'ve come to the conclusion that i seriously attract men above the age of thirty or so. this is of course proved by the cherry cokes (people really drink that shit or is it just the illusion of funky desirable crap you can come out with the cherry?) and strange, funny looking drinks that flooded out way. or really, maybe it\'s just my huge non-existent tits.   either really..  very interesting.  :P  barker\'s upstairs doing something or another..putting roses in water, finding aspirin, sheets on the bed and creamer for the coffee.  a walk by the beach, karaoke singing and dancing...relaxing.  ringing up random waffle houses and terrorizing patrons, hunting down spicers...but it worked didn\'t it?  brushes off sand from her toes, her pockets and feet...a green monitor, a pounding headache, a different turn.  pla-fucking-cebo.  :P  '),(22513,664485,'17,March,2003','2003-03-17 00:00:00','   walking...hazy illusions cloud my mind: foggy, as if there could be a mist somewhere, a change in temparature. i walk, bdly, but my feet knows where i\'m going; my heart leads me.  i\'m blinded...but running.  trailing...but dancing.  crying...but smiling.  hiding...but laughing.  i am me..  trying.  Hard.  '),(22514,664485,'17,March,2003','2003-03-17 00:00:00','   hmm..hockey game and emerald city tonight.  old man\'s evening out with the girls maybe.  if he agrees.  barker\'s getting off work and joining...maybe.  that\'s a whole shit load of maybes innit?  '),(22515,664485,'15,March,2003','2003-03-15 00:00:00','   i thought i\'d seperate melancholic with questionable.  so i\'ve met someone..well, perhaps not really just met someone but that i\'ve known him a while. we calculated a couple of months last night, it\'s strange. i\'m a little taken aback, a little in awe...amazed.  his mind fascinates me, though it isn\'t just that. there\'s something intangible about him, something quite distractingly drawing, motioning, tugging...and i wonder sometimes. beneath those looks he gives me, beneath that goofy grin, beneath the way his fingers trail my body...eh?  it\'s new, it\'s exciting, it\'s pleasuring...is it just right now? just before?   long forgotten hidden fears haunt me again, it is that lump in my throat, the sharp intake of breath that suffocates..worries things to death. and i\'m confused, scared really...of letting go. of wanting to hold on to facades and layers, of wanting to stay being liked, of wanting to still be loved, be wanted, be needed.  will things change as we learn to get to know each other?  will he realize what others have noticed before?  will he learn to point out the flaws, draw out the ickiness, stab the silent. will he master his weapons to fight against mine, will he get to know the weak spots and attack? will it be safe? will he be safe?  he\'s a good guy. a hardworking smart bastard. out of my league? i\'ve learned to shut off those doors damnit, i\'ve learned to close away on emotions, to shut out on feeling. enjoy it a while...just don\'t fall. but there\'s something different about this one, something distinctly frustrating, a voice i cannot shut off.  i remembered too, our first somewhat \'real\' conversation. i remembered putting down the phone and that \'wee-wee\' shit. i remembered that night at your house after the first cast meet and how i specifically kept some space between us, how i\'d talk to anyone but you except maybe just to punch in a smart ass bitchy remark, except maybe just to ask where the soda was or if it was okay to smoke in here.  yes, i quite remember that. and the conversation in the car with amanda after.   and we\'d agreed to not pursue it, to shut up, to just smile and laugh it off. some things just can\'t happen for a reason.  it\'s fighting against being attracted, fighting against these feelings that feel so awfully guilty, i made long long mental lists in my mind, i couldn\'t, i must not, i wouldn\'t, i just fucking cannot.  and it worked a while. see no evil, hear no evil eh? i kept so fucking quiet i deserve an oscar,  and now as puck, madonna, blue-shirt, old man, curly sue, fruity, and the rest find out about it they go oh my fucking lord. and laugh. and giggle. and go, holy shit i should have figured and awww and really?!?!  and i don\'t even say anything so they continue their little valley conversations on their own and i just want to run. put on my nikes and speed for the next available ramp, because this feels too real. you\'re breaking down too many walls and i feel exposed. naked. ugly. you\'re coming too near and it\'s dangerous, it\'s bad to let people in, it\'s not good to trust, it\'s evil to let someone in. it\'s just plain awful bad.  but i can\'t help it.  and i\'m stonking ass pissed with myself.  because i\'m such an idiot.  dork.  i want to say lets do something today but then, i say don\'t fall. don\'t fall. don\'t fall. and keep my distance.  so yeah.  hm.  we\'ll see.  question marks.  '),(22516,664485,'15,March,2003','2003-03-15 00:00:00','   so it\'s been a while, yes, i\'ve noticed too.  i suppose it\'s just been absolutely absurd trying to piece together puzzles, figure out distances...work out shit you know? a lot of blind spaces and empty blanks that needed to be filled..i wrote this strange homey post last night...it was pouring in mid-march, with flowers springing up from every imaginable spot...an unexpected surprise. things changing  and evolving around me and trying to put together scattered thoughts and ranging emotions....almost suceeeding even.  then of course, the computer crashes.  falls apart.  and c:/ can\'t be located.  there\'s some scan error.  some horrible monster that lives in my computer was disturbed.  so there it went...my little post, the first in a while, into never never land.  and here i am, four o\'clock in the afternoon on a monday the day after wondering. figuring. thinking.  ...posting.  they must call this blogger\'s block, for i can\'t seem to find the right words and nothing seems to really flow. my brain searches for the right things to say, my heart it\'s current pains but nothing seems to be able to transpose onto screen, it\'s as if point a can\'t and point c can\'t find b and so it is lost and confusing between. empty.  i\'m pissed afraid.  i\'m pissed excited.  i\'m giggly and laughy and i want to cry out a storm.  run. fly. fall. crawl. search. hide. find.  hm.  '),(22517,664485,'15,March,2003','2003-03-15 00:00:00','   so he asks tired, weary body masked by over-perkiness in his voice, masked by the plastered smile, the raised eyes...lets do something together sometime. can we talk? i\'m hurting. life\'s too fucking interesting, girls are too damn weird, boys need to go screw themselves, heart needs to go away...  and i say honey, it\'ll be okay.   and i say sweetie, it\'ll be over soon...things will be right.  ---  ..demon-free, resting. peaceful.  beautiful.  i miss this.  ---  talking with hands, with smile, with eyes and words with double entendre  dashboard confessional playing..(thank you) she remembers the laughs. the heaving of breath in between puck, silent, and barker surrounding her around...nestled in a cocoon of security. attempting to achieve their overlying marathon of watching day break at waffle house, of between the shifts changing and the morning after.  it\'s funny how much you can learn about each other in a night.  a bad day going extremely good with every breathing moment, with another chuckle, giggle, random burst of lucidness. mad crazy retarded people all with a very sick sense of humor. they connect, on different planes...throw in sexual connotations, throw in movie lines...throw in seven hundred and sixty two under the table, in the shizzo, down with the posse remarks.  she breathes.  walks over to the bathroom.  she\'s eaten tonight.  smiles at the girl in the mirror.  walks back...lights up, she shush-ed the dragon. they helped her. comfort. freedom. space. friendship.  but she watched his eyes, the way they lit up with joy, with cheekiness, and she watched those very same eyes hurting, strumming the guitar..playing melodies. tunes created...she watched his eyes grow silent, retreat into another world.  how does she tell him he\'s really an amazing guy?  how does she tell him he\'s a great damn mother fucker?  falls asleep on the drive home..the wistful sounds on the player, to conversations and directions.   she\'s resting.  she\'s sleeping.  she sleeps.   thank you.  \"bff\"  ---  friday night started out with the ice hockey game, of number twenty one and ten. she laughs in silliness, she walks silent in her thoughts. beach condo, canadians, six girls piled up in a car, two stalkers, one cold, one naive, the other questionable, the girl frustrated.  bags sent to the vault, ice-skating too wet, too many people, strange kooky people trying to hook up. they leave early, unplanned yet up to every precise detail...too rehearsed. too right down to the point it comes out fake; everyone\'s sick of each other, the car and it\'s contents have gotten off on the wrong foot. too many hormones in this car, though i should be rejoicing eh?  football on the beach again. scrapped, she might really have to hit them. hurt them.  old man to the rescue. old man always to the rescue...saves them. but not before she makes a phone call. reschedules. medicine adjustment, psychiatric evaluation..keep me out of the hospital. no more blood, no more nights...she gets better now. she tries harder. she wants it more...craves it.  and the girl supports her. \'go get em\'.  the girl lights up and comes back to life with blue-shirt, silent and barker. her disposition of walking anger sheds back into giggly valley shit. transitions back into the retarded silliness of a bunch of silly people together.  bird on boob. creamer in mouth. fingers tapping on table.   these are all the reasons you love her.  these are all the reasons she\'s special just for who she is  ....god only knows how many other people keep big sticks in their car and shares concealer with the bitch.  :P  ---  the gaping questions.  she wonders sometimes too.  silence them, don\'t overthink.  thinking is bad.  ---  in less than a month they\'ve removed more layers than she\'s normally used to in this frame of time...as much as she bitches about stripping, she\'s not very good at it. she\'s better with the fronts, with the smiley faces and the silly remarks. she\'s better at being bitchy than hurting...at least in the \'real world\' anyway.  barker\'s made cotton candy and guessed blue-shirt\'s right age. she\'s tickled with glee tonight...he actually looked comfortable. okay. like he was having fun or something. did you?  voices, impersonations and music blaring in the overcrowded ghetto parking lot. nail scratching and tickling..reflection on the windows and half expecting a lime-green grasshopper to scare them shitless again, chairs with holes and ashes on pants.   he\'s taking her out on sunday..won\'t let her know what they\'re doing. her suspicion\'s aroused, curiosity killed the cat. ah, she\'ll be quiet then..and wait. and wonder.  hm.   \"I care and I\'m not just the flavor of the week trying to get in your pants.\"     oh?  :P  ---  the old man sleeps. awakens. smiles. cries.  the old man weeps.  the old man goes through things just like us..but he\'s wiser. more patient. slightly more mature about it, slightly more tolerant. enduring.  and so they smile again. they giggle. in their tiny little girl voices they say thank you.  thank you \'daddy\'.  ---  the mother arrived safely in london today. me dad\'s lonely, called at some godforsaken time to talk...chit-chat, you know, do the father-daughter thing. and i know how that feels, to feel lonely, so i woke my ass up and talked. and smiled. laughed.  and told him i love him.   it\'s been a while. in between all the bullshit that\'s been going on, we haven\'t really talked. and i mean talked, like superficial bullshit that didn\'t involve fucking up or feelings or other good silly shit of that nature. we forgot about soccer games (liverpool won. *BLECH*) and new hi-fi systems, or brand-new some new obsession furniture. my dad the artist, the business man, the church dude.  my dad, the super cool mofo.  ---  the other parents have now smelled the sneaky suspicion that she\'s moving out. she\'s yet to say anything, leaving hell\'s fire to when the birth mom arrives in two weeks. to when she returns back to school again, to when they regulate her medication, to when they go to the doctors, to when they settle her transcripts.  mail arrives, letter after letter...of confirmation and receipts.   she\'s moving up. growing. independence.  or so she thinks..  running?  not really, just striving.  ---  she lies down, head on pillow and bolster by the side. watching tv together: strange reports, complete with sarcastic comments. they go through lists at hot or not; bitch, nit-pick, laugh, ogle.  a lazy afternoon, half slept away into tranquility though she\'s only had two hours.  sleep deprived,  she left the girl sleeping in her bed at half past ten this morning..went to church to practice her solo. this should be interesting..  singing. god. witnessing. emotions.  hm.  came back in an hour later, fixed lunch and plans for tonight. perhaps get her nails done later...dinner tonight.  then the girls want to do something.  and maybe kidnap blue-shirt and barker again.  hm.  :P  '),(22518,664485,'14,March,2003','2003-03-14 00:00:00','   my hands caress her soft, smooth skin...her tiny frame and the rich glowing tan, her supple breasts so round and full, erected as she arches with desire. aroused and titillated, my tongue licks her every inch. from her equally frantic tongue, to the plane of her stomach, her protruding sexy hips as her clenched fingers scratch my back in desperation...her eyes roll back with pleasure. my name escapes her lips in a soft raspy breath, she chants, she sighs, she begs, she moans. deep cat-like moans that purr, she purrs to be filled. touched. made love to.  hard, raw, passionate, moans.  my lips trace the nape of her neck, planting kisses. sucking on skin. she is so damn sweet; her taste lingers in my mouth. her eyes flirt with mine as my tongue now finds her womanhood. her box. her opening. my hands massage her, my fingers explore her...she spreads apart further as she grabs at the sheets in the building climax, orgasmic. her voice is terse...her legs tense with desire...trying to be quiet against the hard cold tiles of the basement. lost in a foray of freshly laundered white sheets. it is cold, but her skin is moist...she is wet.  coming.  i scream  ...pleasurely.  wet.  horny.  climaxing.  '),(22519,664485,'13,March,2003','2003-03-13 00:00:00','   forty questions.   1. first grade teacher:  pn sinna - this nice indian lady that wore glasses and smelled of jasmine.   2. last word you said:  uhm. (is that considered a word?)   3. last song you sang:  manbai - kau ilhamku   4. last person you hugged:   hm...grandma, before dinner.   5. last thing you laughed at :  amanda and her latest toy-boys   6. last time you said \'I love you\':    this morning?   7. last time you cried:    at the head doctor\'s..though really, it helped.   8. what\'s in your CD player:    hm..let me check..sigur ros, smashing pumpkins, counting crows, george winston and matchbox   9. what colour of socks are you wearing:   not wearing any...though somebody i know has my little pink piggy socks. no one in particular...just somebody you know..:P   10. what\'s under your bed:    monsters!   11. what time did you wake up today:   i might need to go to sleep first...   12. current taste:    starbuck\'s mud pie ice cream   13. current hair:    layered, down, in-between.   14. current clothes:   knickers and my bca top   15. current annoyance:  idiots that send silly notes and talk in that pathetic \'4u b hooking\' type of crap    16. current longing:  wouldn\'t you like to know?   17. current desktop picture:    urlLink walking down the beach a year ago..    18. current worry:   life   19. current hate:  time   20. story behind your domain:    urlLink chris  is my god. i needed change, space, comfort, and an outlet so he provided...and so, eight months later..here i am. ...still. :P   21. current favourite article of clothing:   hm..black layered swishy dress.   22. favourite physical feature of the opposite sex:   eyes/lips   22. favourite physical feature of the same sex:   hip bones   23. last CD that you bought:  good question. oh, tony williams - machine dreams   24. favourite place to be:   home / beach   25. least favourite place:   miss jackson\'s class   26. time you wake up in the morning:   the ungodly hours between 5:45 - 6 am   27. if you could play an instrument, what would it be:   violin / saxaphone   28. favourite color:   hm. let\'s try least favourite colour- GREEN!   29. do you believe in an afterlife:   it\'s questionable..   30. how tall are you:   fucking short. my license says 5\'2   31. current favourite word/saying:   \'bloody hell\'   32. favourite book:   green eggs and ham   33. favourite season:   autumn / summer ?   34. one person from your past you wish you could go back and talk to:  grandpa + the kids   35. where do you want to go:   heaven?   36. what is your career going to be like:   sucessfull hopefully...?   37. what kind of car do you have:    i don\'t own insurance in the states but in malaysia, a merce. :) lucky bitch.   38. type a line you remember from any book:   \"the end.\"   39. a random lyric:    I am covered in skin No one gets to come in Pull me out from inside I am folded and unfolded and unfolding  I am colorblind Coffee black and egg white Pull me out from inside I am ready I am fine   40. identify some things surrounding your computer:  it\'s on my bed..so the pillows..bolster..sheets..comforter..my ass  '),(22520,664485,'12,March,2003','2003-03-12 00:00:00','   Pink: You see the world in bright pink. The world is a happy, happy place! You love all people and things!! Life is great! You\'re just like a happy child. Spread the cheer.    urlLink   What color do you see the world in?     brought to you by  urlLink Quizilla    forgive me if i laugh my ass off.  :P  '),(22521,664485,'12,March,2003','2003-03-12 00:00:00','   it\'s strange. a very weird feeling, i wonder if i\'ve been here before..some very specific aroma in here, almost like saffron; distinctive smelling yet not pungent. not too suffocating. but it\'s nice...calming. :)  it\'s odd to not walk in a cloud of sadness, it\'s weird to not have a pressing need to comfort pain, i\'ve moved on and today when you asked how it was i could honestly tell you i didn\'t know. didn\'t know because i haven\'t been there in a while, didn\'t know because i\'ve cleared my list, didn\'t know because there are new people in my life and i don\'t feel guilty about it.  i still questioned though, i wondered...but then again, i\'ve learned to understand that i always will.  curious bitch.  :P  so apparently the girls got uhm, \"happy\" last night..  lol  more episodes about the hockey boys and baseball dudes..she gets the sporty ones, i\'ll terrorize the artists. problem? where?  the interview went good though, much talk about tantalizing sounds and being anal perfectionists. his beats that bounced off mine, talk about photography and music. web sites and designs to impromptu sketching in the little six-by-six cubicle which then veered off to funky silly rhythms and machine dreams.  cups and cups of coffee swirling, steam filling up the room...amidst \"intellectual\" talk about the world. shoulder and neck; fingers explore, hands meet...midnight madness, falling asleep. dreaming.  bullfuckingshit.  horny.  puck slept in someone\'s bed last night.  lobster crawled in with red hair.  ginger stocked up with lips.  beauty decided canadians suited her better.  and barker? barker wrote inspirational emails last night. (thank you)  yet amidst it all, she dances alone.   seven hundred and sixty two matches, boys...girls...she smiles. tickled.  ego, head blown.  silliness.  hmm.  hungry.  hah.  '),(22522,664485,'12,March,2003','2003-03-12 00:00:00','   crying, tears drawn is nothing. you cry, but the tears can only get you so far. remembering; harping on disappointment and being angry at myself will only lead me into the golden gates of depression, will only spiral, lock me down further. actions, actions speak so much damn louder than words and that intricacy i have not mastered. that special transition into maturity, i have not learned. have not honed.  pick, i pick at mistakes but i don\'t put into proper going on. i don\'t put into sequences and revert back into fucking up, into breaking and tearing again.  it is not an excuse.  it is not an answer.   it is a decisions i do not put enough action into, a baseless dream i think about too much but put nothing into. irresponsibility is my middle name and i have lived up to it. screwing up again, my name is racheal. i beg with myself, my being, to do it. to do right. i don\'t talk to the ugly, consecrated, shitty reflection, the girl in the mirror. i talk to racheal khong, to all of me. to the living, breathing, walking, everything me.  harp on fixing, on finding for a solution. what i deem as an unexpected emergency, a surprise was a path, a walkway i subconsciously walked through. damnit, racheal, damnit.  strip pride and walk on fire. strip masks and walk with the truth. i am not who i think i am, i am not who i want to be, i didn\'t try hard enough, i didn\'t apply the same anal assholic torture i put my mental self through..into..i wrote it off. i fucked it up. on my own, i wrote this story.  it isn\'t an epic, it isn\'t just drama. it is my life. i struggled, it was my fault. i didn\'t try hard enough, i didn\'t prove my point. i didn\'t stand my ground.  i couldn\'t do it or i didn\'t do it.  my second chance: my last change.  i cry bring on the wrath but i\'m afraid. terrified. i say lets deal with it but i want to hide. want to pretend but i know i must not. this limbo, this destruction, i played it on my own.  i have to take responsibility, i have to admit i\'m wrong. i have to face the consequences.   this is who i am.  i have disappointed.  i am a disappointment.  '),(22523,664485,'11,March,2003','2003-03-11 00:00:00','   confessions are weird things..  there\'s an empty spot next to where i lie, some part of me searches. searches to be filled, be renewed...searches for something. something special that\'s going to complete the picture, my icing on the cake, the gold trimming..an understanding.  it\'s been a rather heavy day, a day filled with too many questions. what ifs. why nots. how comes.  i looked the other day, deep within into questions i\'d been hiding a while. feelings i had suppressed.  i do want this.  i do want to try, fly, fall, crawl, climb.  it isn\'t easy. it isn\'t easy.  so dang hard.  so dang many things.  i strive to be responsible. to be an \'adult\'. i get so sick of being a grown up sometimes, i get so sick of growing up.  don\'t you?  i get so sick of having to put the pieces together, having to mature through learning, having to brush off and say, it\'s okay.  i get so tired of saying it\'s going to be okay when i don\'t feel like being okay a while.  sometimes all you want is just to yell, to scream, to be as immature and as bratty as you can possibly be. to climb into daddy\'s lap and hide a while, be protected. feel safe.  the real world\'s too big.  too dangerous.  evil.  the real world seems way too strange. way too off-balanced.  i\'m off-balanced.  ---  \"princess\'s happy post\"  so midnight football was interesting..  puck and talon had a blast.   what began as a couple of people(lol, like 20?) surged into a good number, but then again drawing random people from the dock was about half the fun...:)  being lifted and turned upside down though...now that was not.  and for that, /you/ still owe me cotton candy.  i had fun though, american football is a terribly brilliant concept.  run, dash, scream, hit, grab.  i like.  great on sand with waves crashing and club music in the background too.  i was going to say something about jc but then i realized that would have been too much time spent on it already and i might get my arse kicked.  but anyway  thesilygrashoper: Once upon a time In a dark world that reached beyond the boundries of earth, there was a spec of dust known as a pixie.  The pixies heart tried to bring her to places of grand splendor, but the pixies body trapped the heart and forbit it from going.  Till one day the pixie began to fight in an epic battle.  Trillions were lost in combat but the pixie and her heart prevailed along with a few others. These formidable allies to the pixie soon became part of her and thus without realizing it the pixies heart had tricked her into going to the magnificent place of wich she had been forbiden.  Her heart in essence had \"been in the right place at the right time\"  :)  ');
INSERT INTO `msgs` VALUES (22524,664485,'10,March,2003','2003-03-10 00:00:00','   come crashing and falling  \"everythning happens for a reason, it does. it really does...\"  '),(22525,664485,'10,March,2003','2003-03-10 00:00:00','   a lame attempt.  unleashing, as if orbiting into a swirl of known and unknown paths.  of knowing that the coffee as always, has two creamers and two packets of substituites but never quite sure how it will taste.  not quite distinctively sure of it\'s meanings.  it\'s representation.  contents no doubt pretty much the same...but no doubt, still questionable.  still pokable.  lingering...the mind wanders. the heart slices. the pounding ponders.  hm.  '),(22526,664485,'10,March,2003','2003-03-10 00:00:00','   notes on napkins and messages left on the phone.  the girl squeals.  he\'s learned to use her number.  he\'s learned to rub her the right way.  she giggles.  then smiles.  they meet.  his blue-gray eyes sparkle, glint...  another conversation.  a decision.  next!  :)  excited.  conceited.  selfish.  bitchy.  she plays.  she dates.  she emerges.  hell fucking shit yeah, she dances.  flirts.  puts herself out there...news travels fast doesn\'t it?  she\'s returned to the other side.  swings really..  she plays it her way now...  don\'t you dare try to stop her, take her down.  she\'s on a rampage.  she\'s on a high.  '),(22527,664485,'09,March,2003','2003-03-09 00:00:00','   lol, so we\'re playing midnight football at the beach tonight.   mmhmm.  get that straight, biatch!  :P  puck\'s posse against mine? HM.  the lesbos against the straights.  the opposing against the rights.  the ogle, ogle, google, run into her boobs, doogle.  it\'s a beautiful day outside too...the sun\'s all bright and cheery. barker and beret have gone fishing, lamed attempts at dragging my ass down to the pier has obviously not worked.   :P  persuasions of skinny-dipping and magnet-poling added with deep sea watching have failed...my ass is staying in my \'mermaid skirt\' and smiling. watching. laughing.  it\'s a cute day outside...  so much so that we had mythology class out by the lawn....in one with the nature. our teacher with his flamboyant styles and a class full of, pardon the saying but, weirdoes..artsies, \"interesting\" people, people with brains spewing out of their ears, people with ideology myths and unrealistic notions..all bathed in the glorious wonder of nature discussing cheif seattle, and wordsworth, and my favorite monk of all times, francis.  opinions discussed. the honor students bring down the oldies and the little prep blonde \"heeheehee\" girls.   the old man listens to an argument about the appointment with the hairstylist. colors..cuts...dye jobs, foiling jobs, boob jobs, manicure...whatever you want really. all with a price. all with a date. all with a sacrifice.  my thoughts go ha ha ha, malboro lights does nothing for me...last box of the damn month. or week. i don\'t know.  trying to quit.   oh wait, screw that.  quitting.  i inhale, my lungs protest quite violently, my stomach churns, my throat bleeds.  i am an eight.  i am on a high.  i am falling.  flying.  crashing.  dancing.  feeling alive.  '),(22528,664485,'09,March,2003','2003-03-09 00:00:00','   people go strange as they get older..  hm.  so ryan\'s all propped up in front with the camera and everything. set to go with bryan..:P  we\'re on first name basis now....and i\'m proud.  the world awaits the unveiling,  i giggle awhile. flirt with the dangers...and smile a little.  grin really.  smug smirk on the face, i\'m in the know.  uh-huh.  '),(22529,664485,'09,March,2003','2003-03-09 00:00:00','   ever feel like you\'re meeting all the right people at the wrong time?  or all the wrong people at the right times?  ever feel like you\'re not complete and empty?  ever stare blankly and cry wondering who you are...what you are..where you\'ve been and how much more to go?  there are so many parts of me that wonder tonight, so many turbulent questions pounding..  i wrote this out earlier, but a mis-click of the mouse led to the whole entry being erased, vanished. lost into the deep dark world of the internet...and i wonder what i said. i know i cried through that, but i can\'t remember anymore. only that it was painful to write, to realize i\'m changing.  i don\'t know if i like that.  i dont\' know if i\'m comfortable.  there\'s this whole new confidence in me that scares the hell out of me. i\'m still insecure but it\'s a nicely balanced insecure, a down to earth modest insecurity as opposed to the destructive shit i\'m used to. and that\'s weird...becuase i actually feel it, and live it.  it\'s as if i actually know i have a reason for being here, as if i have a right to this.  like i\'m moving on or something and some part wants to cling on to familiarity so bad but the other just itching to go. the other mesmerized, amazed, excited.  so many people..  could it be?  it\'s as if a veil is lifted, a sheen is cleaned and i finally see.  i realize.  and then i know  i\'m not that bad  i\'m not that evil  i\'m not that horrible  i\'m actually kinda okay  kinda nice  i\'m actually kinda worth it.  heart on the line again  hopes escalate weirdly  thoughts crossed  people come and go  interests and dislikes  turn ons and turn offs  acceptance.  it\'s strange.  confusing.  two in essence.  can it be?  i want to be completed.  i want to be loved.  i want to be me for me.  i want love.     ...i have it.  what now?  decisions.  '),(22530,664485,'09,March,2003','2003-03-09 00:00:00','   anna nicole needs to go strangle herself.  '),(22531,664485,'07,March,2003','2003-03-07 00:00:00','   it\'s a quarter to six in what must be the strangest weekend i\'ve had in quite a while..  of course, strange and racheal pretty much goes hand in hand together but yeah..  staying pretty much bed-ridden and trying to keep entertained in between all those hours can be pretty interesting. hell, i\'ve not done anything remotely productive nor stepped out of the house to do \'fun\' stuff since friday afternoon...minus of course the midnight run to wal*mart because hey, now that\'s just wal*mart and it doesn\'t count alright?  so after much persistence and sticks stuck up my arse, i posted my picture on hot or not on friday and the amount of shit that has emerged out of that has amused me endlessly.  the sudden cum well-thought out decision to start dating again just seems like a big hogwash laughing matter right now. holy mother of god, i am very tickled. it\'s silly, i run through these people that send me notes and emails and they have all these double matching blah which is just retarded.  it\'s cute though, and i just want to go awwwwwwwwwww.  see, there\'s this whole spectrum.  first we have the interesting ones. the ones with similiar interests, cute adorable sexy on the same level ones. the ones you have nice cute conversations with and get all flustered about..  then there are the drop dead gorgeous HELLO ones which are pretty much out of my league..the idiots with 9.8 ratings and i\'m like...me? nahhhh...  then comes the best ones which are all about , what\'s up my pretty sexy gorgeous laydee, you be hooking up with me soon and we be chillin\' sometime? okay, so i make a bad impersonation but you get the picture. the \'sex,sex,sex\' ones and sleazy icky greasy dudes.  of course, there are also the 30-50 year old paedophiles that cite young girls in their profile send stupid messages which really just want to make you scream and shudder. gross.  apparently also, i attract the bondage, ropes, chains, group.  hm.  then of course, there are also those that just stalk you down.  *tsk* *tsk*  BOTH of you!  helllllpppp!!  hehe...weirdness.  '),(22532,664485,'07,March,2003','2003-03-07 00:00:00','   when i was younger, my mom always told me that if i had nothing good to say, to just not say anything at all.  and for the most part, i probably should have a lot to say except that there are a couple of things worrying me  the first being:  i threw up blood last night.   it\'s been going on for a while now except that last night really scared me more than the others. something doesn\'t feel right or at least soemthing seems to be sending off signals some where and so i\'m a little worried...but i\'m okay, i just don\'t want to go to the doctor\'s that\'s all.  i\'ve been pretty lethargic the last couple of days with the bleeding and pain so it hasn\'t been too sunny on this side of the world. it\'s also been raining like a bitch so that hasn\'t helped any but oh well, no big deal. i\'m a strong girl and damn, it\'s hard to kill me.  :P  and yes, i would know.  other than that, i had to cancel last night\'s dinner plans because getting up felt like an act of god much less cooking so that was kinda disappointing and liberating at the same time.  things are very strange right now, \'playing by ear\' like somebody says and it gets kinda silly and retarded but it\'s interesting nonetheless.  i think i\'m ready for a relationship...but i don\'t know if it\'s because something just feels so right when we\'re together or if i\'m just plain lonely and tired of playing the single game. in the end, i do still want to have someone to come home to, someone to laugh with but most of all, someone to love.  i miss loving someone. caring for someone and going out of your way to make them happy you know? i miss the closeness of two beings, or the understanding and friendship of lovers. i miss having a lover.  i guess it\'s been a strange couple of days. i passed on swing dancing friday night and stayed at home again saturday and so that seems to have worried a lot of people. the crazy bitchy staying home on a friday night? holy shit fuck there must be something up! :P  though really, i just wasn\'t up for it. i\'ve just been so tired and in pain, i\'m either passed out numb, or cranky bitchy pain.  but i\'m alright though.   it\'s hard to explain, but therapy the last couple of months have seemed to help quite a significant bit. is it just medication or is it because i\'m able to think more rationally and in the right direction now? i don\'t know. but something has changed, and that change is weird.  different.  foreign i suppose.  sometimes it gets a little frustrating because this is so damn new to me. there are times where i really would like to just curl into a ball and lock myself away then sob endlessly but this whole being changed thing challenges to find another alternative because consciously, the tears don\'t come anymore. nor does the anger nor pain.  does that mean i don\'t get sad anymore?  (i wish)  it\'s just a few levels less depressive. a few levels not so bad.  but it still comes. no magic pill or anything, no fucking great epiphany, no amazing damn revelation but it\'s getting there...i\'m getting there.  and that\'s very strange.  but then again, i\'m strange.  :P  all of you have a safe weekend and drive safe okay?  hugs  '),(22533,664485,'07,March,2003','2003-03-07 00:00:00','   uhm.  yeah.  exactly.  '),(22534,664485,'07,March,2003','2003-03-07 00:00:00','   hm.  i reinstalled/replaced pms ridden trillian with the new version today.  apparently, it\'s not compatible with aim.  brilliant...just brilliant really.  hmph.  '),(22535,664485,'06,March,2003','2003-03-06 00:00:00','   stolen from  urlLink jasmine:    1. What was the last song you heard?  liebestraume - juliet the orange  fucking great album..   2. What were the last two movies you saw?  catch me if you can / drumline  it\'s a wednesday night thing at silverscreen, cheap three dollar tickets + snacks.  cute little quaint theater with old re-runs and funky ass paintings..rituals honey, it\'s all about the rituals.  oh, and not to mention the sexy boy that works there too. :P    3. What were the last three things you purchased?  ooh! ooh! we went to this craft shop last night and it had all these amazing textures, colors and materials in it. bloody a-wow! (kinda like an independent hobby-lobby?)  hm.  this tube top..pink + white + orange!  and uhm. does coffee this morning count?  (p/s can you tell i\'ve had coffee? :P)   4. What four things do you need to do this weekend?  study for midterms. finish up all my half-assed art projects i\'ve been doing but have been procrastinating over. WORK OUT! make dinner tomorrow night.    5. Who are the last five people you talked to?  amanda ryan lindsay lindy daniel  hmm...look what i just found in \'drafts\'..   \"[12/7/2002 11:36:09 PM | + racheal] i hesistate to use love. love is...shit, let\'s not even go there. but it\'s kinda interesting you know?   when he respects you for who you are even if he doesn\'t necessarily agree with the things you do or the opinions you stand for.   when he tells you you\'re beautiful even after you\'ve had the world\'s most insane assholic day and your hair isn\'t even apart of your own anymore, your black shoes against your brown bag put together in the bare five minutes you had this morning, your nails chipped and your eyes tired, your mood horrific, your menstrual cycle being all hormonal and shit, he smiles, hugs you and you suddenly forget why you were so bitter and angry with the world and smile.   when you\'re laughing so hard your guts hurt and you can pillow fight and play 123 with your thumbs in the middle of nowhere. when you\'re lost in your little world and he\'s in it. or when you\'re in his little world he\'s sharing with you.   when it\'s three in the morning and you\'re having a nervous breakdown, he drives right over even if you\'ve repeatedly said no but secretly wanted him to. talks to you and calms you the fuck down then makes ramen noodles together with you and tucks you into bed.   when you roll over to find his arms locked around you and tranquility washed over his face. safe protective arms that hush you from a bad dream, cuddles you in and kisses the boo boos away.   when you\'re watching him play and all these people shout and scream for him and that immense pride in you that goes, that\'s my baby! and the secret part of you that takes a deep breath and goes, whoa.   driving home, literally, figuring out the stick shift because he\'s let you touch his other love of his life and is now laughing at your silly face in bfn mobile trying to get home with the engine choking and dying on you about a million times saying clutch baby, clutch!   having real conversations about god, and dreams and \'real\' things then making fun of \'superficial\' and rather retardedly, talk about shopping and manicures and stupid crazy shit like that.   but most importantly, coming home and finding that he\'s found your favorite game, the one you used to play when you were little, and there\'s that look in his eyes that spells out how much you mean to him. and it\'s crazy.   when he says i love you for the first time and your heart melts.   when actions truly speak louder than words.   that\'s love. \"   *sniggers.*  muahahahahahhahahahahahaha.  :P  '),(22536,664485,'06,March,2003','2003-03-06 00:00:00','   twelve hours later.  my head finally touches the pillow, my fingers tap the keys and my eyes finally start to open. to see. to watch again.  i finish up my half-written post from the afternoon before and smile to myself.  a night out, an escapade.   flirting with the dangers of brinking on insanity. drama. music. dancing. singing. stripping butt naked in the middle of the street.   commando style.  it\'s been an interesting night.  the boy scores a home run.   the girl runs an extra lap.  the kids celebrate.  '),(22537,664485,'06,March,2003','2003-03-06 00:00:00','   i deal with being me everyday...sometimes it is not so much a task, and i actually do have fun. sometimes i smile, and the girl in the mirror smiles back at me too...then we hold hands and walk forward. yet sometimes, sometimes it gets really hard. sometimes she points out things i force myself to be oblivious to and i battle her demons..i reach for her heart that bleeds so damn desperately and horridly, but it is never just an episode or a continuation, it is a frame found in seconds and minutes from time. from life. from trying, then trying again.   i wallow in depression, in disorders and in \'state of minds\'. diseases that invade my thinking space, uninvited ghosts that haunt me..destroy me.   i cry and hurt and fall and bleed.   i brush and smile and laugh and dance.  i live in different extremes and so many times i crouch into a ball, my fetal position and cry why me? why me? why me?  then i tear my being apart and forget.  and grow cold.  numb.  my obsessive compulsive, manic depressive, bulimic retarded self brushes off the ickiness and finds hope.  the problem is...i always find hope.  i always get back up.  i always swing and traverse and try again.  for the most part, that is solitarily my biggest frustration.   to fly and fall and have no space in between.   ---  george winston plays softly in some realm between reality and my world. my universe out of these four walls. out of this space that comforts...protects me.  rubbing out x-es and crosses.  stamps.  ink on the back of my hand.  marked.  the clock reads 3:58 in the morning, i lug in reels and reels of excess baggage..of change and make up.  my body screams physical torture, my mind yells for attention of it\'s own and my brain demands for an all-nighter for the mid-terms tomorrow.  a scene prepared, a story read, a math exam formulated.  time to live in mythology land a while, time to escape reality.  time to discover mnemosyne.  '),(22538,664485,'05,March,2003','2003-03-05 00:00:00','   so it was ryan, puck and i last night.  terrorizing the old man with seven hundred and fifty two voice messages that span from orgies from the back seat, to cheers, to foreign jabber that only happens at three in the morning.  by the time we\'d left to all drive home, catatonia was playing about tinkerbell and the sun had already risen. the new waitresses from the next shift had already come around and dusk had somehow metamorphosed into dawn.   we laughed so hard, peed so hard and buzzed.  highs from a million things, the rehashing and relayering of a thousand stories, a million tragedies, a trillion connotations.   an it was like swimming in a mind of absolute extremes. silliness, kookiness, intelligence, retardedness and horniness.  and kinkis.  lol.  good night out..whoop.  hm.  i\'m in the honors room and everyone around me is diligently studying for the mid-terms and i\'d rather just jump on the table and just go shoop ba dah doo wop.  whoop. whoop.  hehe...seville tonight!  :)  '),(22539,664485,'05,March,2003','2003-03-05 00:00:00','   so we had this conversation about god today..  me, the old man and puck.  puck the christiany christian god believing-saving wonderful saintly one.  old man, the atheist something that sounds like antagonist spiritual non-believing dragon slaying one  and me, well, me.  me the, i\'m not really sure where i stand person.  the i know there\'s a god but i\'m not exactly sure what my thoughts on it are...  it\'s that questioning of faith that people go through...the change from where a devout believer learns to ask? learns to contemplate. learns to wonder..  am i good enough? is there really someone out there? does he see me? does he listen?  how?   what?  where?  when.  and then the questions role around again and i wonder.  i question.  i float around and twist around in circles and get dizzy.  lost.  but i know...and i believe.  i understand.  i just don\'t know where i  stand .  questioning....like bisexuality really.  '),(22540,664485,'04,March,2003','2003-03-04 00:00:00','   i\'m uncomfortable.   investigative reporting seems very wrong to me. or let me rephrase that, the article i\'m doing and it\'s consequences is not exactly something i\'m comfortable with. perhaps it is because i don\'t believe in it enough, that i cannot really get the complete feel for it. the angst behind it. the full-out passion towards it to question.  i\'ve grown up in a country where you don\'t ask questions. really, you just don\'t.  they shoot you for that shit.  they put you in jail with all these internal security acts.  i am not used to this freedom of speech...not used to questioning out loud.  sure, i am not the easiest person to get along with. sure, i am loud, bitchy and have an opinion with just about anything but to write it down? to put it out? to give it to an audience?  do i want to.  oh wait, can i not?  no.  and that\'s a hell no for you.  i talked to my teacher about it. but he\'s seems very set on having me do the story. the big drama hoo-ha piece. and i want to say thanks but no thanks.  but it\'s a class.  and i need the credit.  and it\'s part of the finals.  it just seems so \'wrong\'.  like swinging a carrot in front of the donkey to lead it to the slaughter house.  gah.  ...journalism bing bang kiss my ass.  harrumpphh.  '),(22541,664485,'04,March,2003','2003-03-04 00:00:00','   i swear there is nothing like going to watch a hockey game and retreating to the sports bar downtown later with our very own private party.  very interesting.  rest assured, there were more than enough numbers to come out through me arse.  :P  and amanda\'s rendition of the star spangled banner?  bloody amazing.  and the look on their faces?  fucking a priceless.  old man has the pictures though...perhaps when he returns my lipstick i\'ll confiscate some of the pictures and post it.  yummy canadian hockey boys.  very subdued almost cold, mean and bitchy racheal.  very rough waves @ beach.  very pleasant memories rehashed.  wet.  memories.  :P  but anyway;  three cars.  four girls.  a baseball team.  eye contact.  an ice hockey game.  lots of cheering.  screaming.  yelling.  stomping.  a couple of random phonecalls.  strip poker.  dollar bills.  stalking.  cheap throw-away kodaks.  expensive retarded working digital cams.  good time.  wah bah dah boo bop.  and on another note though, what the hell is it with the almighty shitty weather?  '),(22542,664485,'04,March,2003','2003-03-04 00:00:00','   i\'m changing.  i\'m not quite sure how much i like that but i\'m changing. i\'ve changed.  ...i\'m different. i can\'t define it really, just that i feel it in my soul, that deep within me it rumbles and  i know .  sometimes i pretend not to notice, not to care. sometimes i just shrug it off, as if it is just another piece of this whole great puzzle of life. this inter-changeable correlation of depth and perception, another stage, another turning point...just another significant change of course.  another walkway...another journey...another road to be taken. traversed. lived.  i am proud of me, and the things i have achieved. i am proud of the person i am now...the person i\'ve honed, mended, fixed and am working on. the person i\'ve been working on for eighteen years and the funny thing is that i can lie here and tell you, you know what? it\'s not that bad. i\'m not that bad.  ...and  that\'s  a good feeling.  but sometimes, sometimes though i miss my old self.  i miss the dork. the stupid silly sensitive person in me that tries to save the world. the part of me that\'s so god damn naive it\'s incredible. i miss crying alone in the middle of the night, that after feeling of a good cry. that after feeling of knowing it can only get better from here. getting back up, putting back both my feet on solid ground and standing. smiling.  there\'s something distinctively missing in me. that feeling. passion.   selfishness.  greediness.  that voice that tears you apart is the very same voice that drives you and i have lost that edge. or at least my heart seems to have kept it away a while.  i am calm. almost numb.  i am dying.  and i\'m not afraid anymore.  and that scares me.  i\'ve changed.  '),(22543,664485,'04,March,2003','2003-03-04 00:00:00','   twelve hours later, she sits with a bowl of lightly salted-fat free-buttered popcorn and wonders what it would be like to recreate snow and toss the whole bowl into the air.  ...that and she hasn\'t had enough of sleep nor does she make sense.  she alternates between outfits and has her nails done. soaks in a bathtub and relaxes in bubble salts and bath oils. big night out tonight, puck\'s big day and half the world is coming out to see her. mardi gras game and party afterwards.  proud of the girl and wants to smile nice and big just for her.   \'things happen for a reason\'   more things to do, more people to see, more shit to antagonize.  but she\'s good.  and she smiles.  and she dances.  weeeeeeeeeeeeeeeeeeeeeeeeeeeee.  :P  '),(22544,664485,'03,March,2003','2003-03-03 00:00:00','   she\'s happy, then sad...in between, then silly-giggly laughy again.  betrays her inner sanctions and reaches for tiny capsules. it\'s like being diabetic really they tell her, they need their insulin and she needs her saratonin. just a disease, another cancer..just life.  starving at three in the morning. an empty stomach with acid gurgling away; her demons torment her yet she pleasantly resists. fucks with them and tells them to go to bed. she talks to no one and to herself, her four walls. locks herself in a villa with no windows encapsuled in her mind.   conversations vary and people interchange.   bitter sad tears bite behind her eyes but then she smiles again.  she\'s retarded.  she\'s \"special\".  she\'s just her.    hm. time for some scrabble/literati kicking..  '),(22545,664485,'03,March,2003','2003-03-03 00:00:00','   boop sha ba doo wah.  frat boys +  urlLink maxi pad  alert!  :P  '),(22546,664485,'03,March,2003','2003-03-03 00:00:00','   \"If she digs you because you KNOW you dig her. You dig her chili sooooooooo goddamn much. And you don’t want her to know it. You don’t want her to know you think of wrapping her in your arms, caressing her hair and her back, parting your lips to welcome hers. Because women,….. well they can smell an agenda. They can smell fear. And they can smell desire. J’elle desire. J’elle desire beaucoup. Mais je….Je ne sais pas…..Je suis stupide\"   it\'s that giggly stupid feeling that makes you walk down the hallways with a grin the size of your ass. a smile that floats you through the day and a laugh that completes you.  it\'s that smug smile you give him when he asks \'what\'s up sunshine?\' and you just giggle and laugh somemore. smiling.  yet it\'s that feeling that rumbles deep inside that wonders..  why me?  how?  where?  what?  which cosmo, which deity, which goddess shines down on me today?  the enticing aroma that swirls through with you, his aura that glides off yours and god, that fucking eyes.   that face.  his smile.  touch.  shivers...your body floats in ecstacy and all senses are in overdrive.  you play the uninterested bitch but oh my god, if only he knew the way your heart pounds against your chest and the squeals that go off in your head when he comes to find you. and you are noticed. and you are complimented. and you are appreciated.  the thrill of the chase. the hunt. the starry look in his eyes that makes you question in amazement, me?  little things that you slowly learn about each other, it\'s beautiful how things correlate.  he\'s in the class next door now. thirty minutes. he\'s \'walking me to my next class\'. and we\'re doing \'lunch\'...(or coffee).  i can\'t remember what he\'s wearing. it\'s ironic how people get so enthralled they remember every fucking thing in detail...he can tell me what i wore the first time we met (?) but i only remember how nice he was. his smile and those damn eyes. piercing through...and meeting again.  i can only remember the hush of his voice this morning and the way his face lit up walking towards him.  and i smiled.  i graced.  i played it \'cool\'.  really.  and here i am...  freaking out.  weeeeeeeeeeeeeeeeeeeeeeeeeee!  \"details girlfriend, details.\"  :P  '),(22547,664485,'02,March,2003','2003-03-02 00:00:00','   i just deleted half of my paper.  why?  it\'s that evil paper clip thing that fucks with you i swear. so i \'x-ed\' it off.  and with it.  my very hard-worked paper.  thank fucking god for notes.  but still...  arrrrgggghhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhh.   :(  '),(22548,664485,'02,March,2003','2003-03-02 00:00:00','     '),(22549,664485,'02,March,2003','2003-03-02 00:00:00','   it\'s been a tranquil quiet day today. peacefully calm almost. kinda like a la-di-da day for myself; not a giggly moronic hyper one nor a depressed the world is a bitch fuck day...just a nice. swooshie day.  :)  well, i also did only get up at noon so only having half of the day to deal with might have something to do with it too..  hm. went to the gym and got out all the bad chi by running for about half an hour than taking it out on the machines. if my muscles hurt like crap tomorrow, i\'m just going to laugh instead. it was good though, needed to sweat things out?  i lost two more pounds. yay racheal. :) my weight\'s been fluctuating so weirdly in the last couple of weeks i think it might have been detrimental to my health...or was it the other way around? whatever.  went over a couple hundred of math questions at the math lab with robert for the test tomorrow... now hopefully, i will kick ass. and if i don\'t, oh well, i tried. :P  f circles fucking g and it\'s domain? gah.  i did finish up my paper at barnes and nobles though. which is quite surprising really considering that about everyone and their asses were out there \'studying\' too. hehe, i didn\'t mean to be rude but damn, if you have a 1800 word paper to be completed by tonight and you\'re starting from scratch, hell yeah you\'ll be taking your fat yellow behind to the corner and hiding with your mugs of coffee too.  lol, you know you\'re habitual when the girl at the counter recognizes your order and the boy that makes it asks you why you weren\'t here the other day.  but i like my rituals thank you very much.  important lesson of the day?  aurora was a fucking nymphomaniac and nymphomaniacs are very  very  different from kleptomaniacs.  '),(22550,664485,'02,March,2003','2003-03-02 00:00:00','   oneactor101: bye  oneactor101: i love you  oneactor101: (two points)   yay.   :)  '),(22551,664485,'01,March,2003','2003-03-01 00:00:00','   coming home, i had so much to say to you. so much i would yell and scream at you for.  but then i figured it wasn\'t worth it.  you made your choices.  you chose your actions.  you voiced your opinions.  i am pissed.  i am livid.  i am angry.  i am furious.  but   shit yeah  it hurts.  you\'re a fucking fool.  i hope you just realized what an ass you were tonight, i take account for the inebriated bitch i was, but jealousy is really an ugly green monster.  not the first time, not the second, won\'t be the last either.  i have an ex boyfriend to play games with me with.  so no thank you on the offer.  no thank you on being yelled at.  no thank you on being treated like dirt.  no thank you for feeling like dirt.  sod the hell off for all i care.  the ironic thing would be, it\'s funny how i had so much fun tonight.  hugs, dancing, bouncer-ing, hanging out with your friends...chilling. at ease. smiling.  you peeled out, perhaps only to reconfirm my thoughts on what jerks men can be. perhaps it never crossed your mind that my feet hurt, my brains were tired, and i only live 23 miles away. perhaps it never crossed your mind that other than a few frat parties here and there plus the random get together, i knew no one. perhaps it never crossed your mind that as much as i asserted the cold bitch can take care of herself doesn\'t need anyone role that i was still a girl. and i still have no fucking damn idea how to get home from fox run and god only knows with the directions.  i admit to being slightly more intoxicated than i probably should have been. i admit to submitting to horrible disgusting cancer sticks, hell, i even admit to warming up, getting to know people, talking to people instead of giving people the \'leave me alone, don\'t touch me, what are you looking at\' stare.  my point would be tonight i put myself in your shoes. and instead of just humoring myself and you, i let go. i realized what a fucking amazing person john is, and how fucking sweet (and sexy) brian is, i ogled at becky\'s turkish belly dancing, and made a conscious damn effort to be nice to everyone. to not be a wallflower and stick with my posse. you\'ll blame the evil bottles but go figure. tonight, it wasn\'t so much of an effort anymore, but that i enjoyed it.  i understand where you\'re coming from but do you see where i come from too?  i\'d give you that much credit for coming back.   but hell, to sit forty-five minutes in the car with someone that was going to go at eighty miles an hour and be pissed at me? been there, done that.  i\'d give you that much credit on checking to see if amanda came to get me.  but what if she wasn\'t around? what if the boys were fucking god know whats and i didn\'t get home safely. would you have a conscience?  i wonder sometimes.  and so as you sit there, sulking, reading this.  ( lie one: you do read the site..i know computers remember? ask brian/john/chris sometime about referrals, site tracking and cache files. )  go figure why i looked the other way when you said you were sorry.  go understand why i\'m pissed.  go try and rationalize my immaturity against your outburst.  go shit yourself because i\'m not just a thing with legs and arms.  i\'m human and you hell fucking know i\'m sensitive.  '),(22552,664485,'01,March,2003','2003-03-01 00:00:00','   on my way out.  need to pee.  fifty million things to do.  '),(22553,664485,'01,March,2003','2003-03-01 00:00:00','   i\'ve been having dreams lately..extremely strange dreams.  take for example the other night\'s dream. somehow, with chris leaving and everything, he\'d magically appeared to have been working in the same office with me. now, if you want the major flaws in that dream it\'s:  i. we\'re not even in the same line, let alone fucking country! ( london  yorkshire and florida? hm.)  ii. come on, me at a 9 - 5 job?!?  well anyway, apparently i wasn\'t too flustered by these facts and it was quite normal. hell, we even had coffee breaks in it. gen, his p.a. was in it too and something around a book my dad had published, \'if i really wanted to be happy i would...\" something-something and my new stiletto shoes. la di da.  strangeness, chris was in a suit too.  i must say though, it was a cute little office. lots of people, plants, funky colors and hip furniture.  :)  which brings me to the next dream..  apparently the girl and i are now \'attached\'  and we have this beach house..it looks like something out of a miami home, a massive slab of concrete with about a hundred different gazebos and additions. it\'s fucking gorgeous and it\'s this yellow-okra-orange color with a pool that has a waterfall in it(?) too much of mtv cribs for racheal i say..  we alternate between walks at the beach during sunrise/sunset and midnight runs while talking to random strangers and we seem perfectly happy making dinner together. we bake cookies and chop up vegetables while the curry (yes, curry.) simmers on the stove. i am happy and she\'s smiling and it\'s strange.  and at night we curl up together and fall asleep to the waves crashing...and when i wake up in the middle of the night, i\'m not there anymore.  instead, i\'m in my room.  alone.  cellphone flashing voice messages.  the clock reads 3:49 and i don\'t know if it\'s night or day because all my blinds are shut and my body feels heavy, as if i\'ve slept forever.  i\'m still tired and so i spend the next few minutes in melancholy before sleep sets in again and i\'\'ve escaped.   and now this escape has revolved into a party somewhere and lots of smoke, completely drunk asses and loud music. we\'re dancing and walking our sorry intoxicated loud selves around, these are all familiar friends and everyone has a joke(or jokes) to tell. we\'re laughing ourselves silly and the added alcohol in the system plus random tiredness at what seems to be dawn only adds to the \'good time\'.  we have a waitress, and perhaps we\'re in a mixture of nouvo and fridays. our crackers are stale (i had stale japanese crackers in real life yesterday) and the bitching starts. we ring up the manager and complain about such poor service...our dinner\'s taken too damn long to arrive and the mics don\'t work. (karaoke?) we make noise but in drunken stupor, enjoy it anyway.  the time frame changes and so has the location apparently. it looked like brian/wayne/john/chris\'s apartments at first...somehow, it\'s become this lavish posh house. don\'t ask me, my brain cells are overworked...that\'s why i cna\'t tell my head from my ass...but anyway, the party is over. it\'s messy and there are beer bottles all around, i\'m passed out on the couch. tired and inebriated. soft jazz music plays from this amazing surround sound system by the corner. we\'ve come home and continued apparently, and it\'s quite a pretty though sterile house. very modern. steel and pewter all around.   well anyway, friend of a friend of a friend\'s competitor is here. he\'s picked me up to tuck me in and my half-asleep brain tells him to climb into bed with me. even in drunkness, she\'s lonely...afraid of being alone. hmm.  he comes into bed and we are cuddled together. i start telling him that this doesn\'t mean a thing, that it\'s just for mental purposes and not heart...and if there was a climax in the dream it would be this. i open my eyes to see him again and i see pietro.  it is that same exact feeling. telling him it meant nothing and that i wasn\'t ready for a relationship. all that good stuff and platonic crap.    then i get up, sober now? and walk out.  alone.   now, there are about a million metaphors and subconscious meanings in the dreams but holy damn fuck mother of god.  no more sleeping for racheal.  too overrated.  grr.  :)  '),(22554,664485,'01,March,2003','2003-03-01 00:00:00','   my body feels physically whacked. like if i had to sit down and list down everything that was misfunctioning and fucking up, i wouldn\'t know where to begin.   and so i lie, half on my back, half to my side, trying to ignore the pain and strong urge to run myself against a wall. hell knows, we don\'t need any more repeats of the tylenol incident so i\'m \'roughing it out\'. really.  i just got off the phone with \'her\'.   i remember her smile. my little prep \'friend\'. i swear my face lit up like a light bulb when i saw her today...in the midst of all the chaos with curly sue and puck, navarre bum and the rest of the acting group....she saunters in in all her glory and goes, \'baby...\'  she holds me from behind and hugs my waist as she watches me get ready. i hate running late but she says no big deal, plays with my makeup...dances around to the lame-ass (i\'m pathetic) sappy country station i\'ve set it to before that song comes on. our eyes meet and i don\'t know whether to cry or scream so i kiss her instead.  we forget about the little tiff, a fight about nothing because she\'s leaving...and i\'m staying. because this casual fling has taken a new adjustment and it\'s hard letting go...about being angry with each other and no one.  monday. monday. monday.  you know how i emit that bitchy, cold exterior? she plays that well...inside, she\'s this sweet, amazing, loving, kind person but somehow, she enjoys the company of the ice princess. my walking gap commercial strutting her sexy self...i show her off at the theater tonight as if she is my prized possesion. i want to say, look at us! look at us! i want to say, mine! mine! mine! and spell out r-a-c-h-e-a-l-\'-s on her forehead and tell the idiot guys to stop making passes and it\'s rude.  we walk to the parking lot and i have to send her home. i almost want to cry, it\'s so hard to say goodbye. but i tell myself i\'ll see her tomorrow...and sunday.  but what about monday?  and tuesday?  and wednesday, thursday, friday?  she says we can make it work if we want to. and i want to. but i don\'t. but i do. and i don\'t.  so i walk back alone into the theater and slip on the smiley face and drive over to the cast party with the old man. ah..the old man. the old man did an eeriely fucktastic job today, i was scared. scared of death and of seeing my friends acting out death because death has been so imminent in my thoughts.  now i sit here giggling because when as she was pulling away after her goodbye kiss, she said the magic words.  perhaps i realize now, you don\'t choose love but that love chooses you.  which reminds me..  how in the hell did i ever get this lucky?  '),(22555,664485,'30,April,2003','2003-04-30 00:00:00','   and blogger publishes!  :)  '),(22556,664485,'28,April,2003','2003-04-28 00:00:00','   building an ectasy.  '),(22557,664485,'28,April,2003','2003-04-28 00:00:00','   sometimes, i wish to see what others see in this world, wish to hear what sounds they listen to and wish to feel what things that inspire them. change is inevitable and a subtle flow of picking metaphorical instances that create rhythm in the beauty of life and the ever-present feeling of discontinued murmurs, an ever-present crochet of patched quilted dreams bottled up through see-through glass, waiting to shatter and spill over.   boundaries are where you set them and the things i see, the things i hear, and the things that inspire me i know, are the things that truly make me unique and special, \"special\" for just i am. and sometimes, just sometimes, you find the faces that know the crevices of your heart, the visions of your soul in a crowd. and it\'s at these times that you see it is the bond that links and represents you in this world. this universe.  '),(22558,664485,'28,April,2003','2003-04-28 00:00:00','   paper journal blog entry   so we\'re back to this existence; and i write again using the same typeprint i\'ve forgotten a while..it still comes easily though, almost fluidly...perhaps old habits truly die hard. i almost sent a text message last night that i would have regretted, in between puck\'s panic attack, sexy pants tryst with jumpoffit and the booming and tinkering of bizarre sounds that vibrated beneath us of a plethora of senses being kicked into overdrive and random drag queens strutting their stuff with dollar notes flung into the air : almost home they say, a certain paradise..  i might have been lying to myself a while, really desperately, wanting to believe but the subconscious mind is a brilliant thing innit?  i\'m falling asleep writing this, my eyes are droopy and the pain killers are setting in. aroma fills the cafe as notes are sprawled all over the table and joints are weary exhausted; gap box lying on a chair next to my purse with it\'s new treasures inside. this boy,  a little muscular, a little yummy looking sits across the room from me but i catch him stealing glances and i smile. in barker\'s orange cargo pants with it\'s straps and ties and velcro that makes this swish-strraachh--wishy-washy sound and a tied black shirt with sleeves pushed up to show my new tan and the glistening diamond star of my belly button ring. black framed glasses and the customary silver jewelry: hoops, bracelets, rings, with and a manicure of striped lavender shades lined down with crystally sparkly things. lotus position, back to the math and mythology notes, back to the comparison of mamet and beckett, back to debating contemporary ethics and the law of the media industry.  a sip of coffee and soft jazz music that plays subtlely in here yet clearly, soothingly. rumbles and mumbles of discontinued chatter that fills the air, a picture of barker fixing a bath for us and nynkempoops curls, the indistinct scent of fresh flowers and the yellow-orange golden hues of a sunset at 6:49.  the day coming to an end and a night just beginning of journeys, of hopes and of dreams. cackles and silly giggles ensue, yummy boy heads to the table and i\'m seeing the head doctor tomorrow. i have a feeling she might be impressed.   =)  '),(22559,664485,'28,April,2003','2003-04-28 00:00:00','   so it\'s a quarter after ten, and i stink of ralph lauren.  :)  on the way out to the cast meeting and waiting for sexy pants to come pick me and puck up. it\'s been a long ass crazy day, got my classes for the fall registered and had a nice long pleasant talk with mr. cashwell. changed my major and rethought the direction of my life and where these currents were taking me. confirmed my plane ticket and picked up a couple of books, redrew money and wrote a nice long entry in my journal. attempted to paint a little before paramedics got way too exciting and distracted the heck out of me, the medicine field is so damn interesting...perhaps later in life when i\'ve achieved all my immediate dreams i will pursue that. you never know. i have one long ass running post right now, typing as fast as i can and making four hundred and fifty two typos in between.  i get to see the chick! i get to see the chick! later anyway, work first then comes the pay off.   damn, gotta go get my dollar bills.   heh.  i look hot tonight, and you know what? i know it.  :) whoop. whoop.  '),(22560,664485,'27,April,2003','2003-04-27 00:00:00','   yes, i am pretty darn good at putting off things eh?  i suppose i could tell you about my day and rant about my scene and acting and the theater scholarship la di da. then i could also tell you about how old man\'s gone to tampa and the weather is being a bitch and sexy pant\'s at the beach with seizing people or about the talk with ninkempoop this morning and her fucking unbelievably sexy new hair and the way it smells and you just have to take a whiff of it as she leans in close to give you one of those hugs in which you nonchalantly just smile and laugh with her.   or i could tell you about chick from emerald city and how we\'re going to see her tonight. dancing. talking. laughing. hobbling around and sometimes how things seem so right and so wrong at the same time. i suppose there\'s always the infamous barker and how we\'ve fallen into this sorta odd existence. a booty call here and an intrusion of space there, it\'s wonderful and magical and silly and stupid and taxing in it\'s own little ways. some days, it feels as if we\'re growing apart and some like the other night, feels as if we\'ve just gotten to know each other and i can\'t wait to get to know him more. discover him further.  then there are others, perfect moments in which everything is right and so damn good it\'s not just about making love to each other anymore or having another body to curl into to sleep at night. there are times where everything is held on a pedestal and as if a symphony as been unleashed into a catastrophic melody that just plays so damn right but yet it\'s wrong. there is friction, and sometimes it\'s sexual tension...like trying to make love with your clothes on and the build up to it where it all explodes into a huge giant collaboration of rhythms and movements, in, out, around, licking, teasing, wanting, needing, fucking, grabbing, holding back, stopping, coming together.   it\'s odd in a good and bad way, but shit like that, i should know by now should just be left unanalyzed.   :)  in any case though, i have a cast meeting tonight. and maybe we\'ll go home, or maybe we\'ll meet at waffle house and share a drink together, or then again maybe we\'ll just talk on the phone and share a ciggarette later.   he leaves in a couple of weeks, and so do i. we\'ve shared our worries and hopes and dreams, praying they\'ll all correlate and work out somehow. that there is somehting bigger out there and that we will reach the infinite beyond.   times when he tells me he\'s lonely, my heart breaks and i wish he\'d had things i\'ve been blessed with that we both know i take for granted, and so in those very moments that i look back upon right now, i realize. i am selfish, and i am very far from perfect...but that i haven\'t turned out yet and i\'m going to.  i\'ve done well for two weeks without the head doctor and the strange beautiful thing is that it shows. i look back at the pictures from the other night and i see so many things and sometimes, i think maybe, it\'s just me but it\'s magical when someone else looks at it and says you\'ve changed. and says you\'re going to do well. and says you\'re worth it, you\'re beautiful, you\'re special.  and for those moments, it carries you through because you do feel special and you are.   i\'m looking forward to being with the kids again, i suppose they keep me grounded. i run through a battery of tests when i go home but the kids keep me strong because for every one procedure i go through, they try and make through ten.   i\'m going home to work, to be with my friends again, to prove to my family that i can make it...that all the times they\'ve believed in me, that all the times they\'ve stood by me is paying off. that i\'m going to be just fine and that i\'m doing just fine. that while they may still worry, it isn\'t so much a concern anymore, i\'m determined as hell to cross this road, i\'m determined as hell to make it.  i\'ve learned to speak up, to stand up; to say this is how i feel, this is what i want to do and this is what i\'m going to do. i\'ve learned that there are times i should not take no as and answer and times i have to get off my ass and realize i\'m not the only person in the room, that i have to accept things and situations and people for just who they are.   life hands you lemons, you make lemonades. life hands you racheal, you smile, kick back, yell, scream, cry and dance.   look mama, i\'m making it.   i have straight a\'s for a course i thought i\'d never survive, i\'m on the honor\'s roll for something i thought i\'d never get my shit together enough to achieve, i\'m offered jobs people only dream of having, i have friends that go to hell and back with me, for me, and because of me. i have love, i have trust, and i have wants, and with all that is in front of me, all that is on my plate, i realize, i can\'t do this half-assed, i can\'t do this half-way because life\'s thrown me opportunities, talent, and chances i should never take for granted.   because i\'ve finally reached a stage where i\'m growing up and hopefully, blossoming into an adult with worth, integrity and maturity. or so that\'s the plan anyway.   how\'s that for monday afternoon thinking?  ...time to watch some cartoons.  '),(22561,664485,'27,April,2003','2003-04-27 00:00:00','   \"there she stands all alone he reads her like an open book he knows what to say to her to make her smile she studies his every move wondering what hes thinking but its not possible never will she know what goes on in there for she is trying to figure out that mind of his see the problem is,  hes not using his mind... she needs to be figuring out whats in his pants for that is what boys think with\"    good one amanda. ten points for you.  '),(22562,664485,'27,April,2003','2003-04-27 00:00:00','   so the much awaited, planned surprise party for madonna was last night. aptly themed the infamous pimp and ho party, it was quite interesting with three drag queens, spoonfulls of altered sexual preference and very very plastered people.  i\'ve just gotten off the phone with steve. holy mother of god, this day is strange.   i suck at this blogging thing, i\'m beginning to lose interest in it and have instead preferred writing, lately, in my journal. hand-written shit just seems more honest, open, and sincere. i\'m doing pretty good and i still have a couple of stupid moments but other than that, it\'s been going pretty steadily.  things seem to be tying in together and i don\'t have much time for anything but that\'s not such a bad thing because it isn\'t a chaotic type of shit but rather a nice flowing series of events in an odd kinda way.   i sit here listening to the radio, my staple of pop and part of me can\'t stop smiling, grinning, and the other cries because it\'s overjoyed, loved, frustrated, confuses, worried and happy. it\'s a beautiful night, and the stars are shining brightly...everyone, for the most part, are doing pretty well.   summer rolls in and this computer hums to a tune of it\'s own, fresh roses fill the room...this is peace. security. tranquility.   this is the life.   my life.  and i am thankful.  thank you for being a part of my life for a while...i think i might go off for a bit. finals are this week and i might just go around the country a bit, see some parts of the world, live in a different culture fr a while...perhaps discover myself. perhaps share my life with someone else. let someone in and all that crap.   redesign in about a week or so and perhaps an update or two from wherever the wind takes me.   i\'m coming home and i\'m kicking ass.  '),(22563,664485,'25,April,2003','2003-04-25 00:00:00','   barker makes me post.  old man too.  so does nynkempoop and sexy pants.   chick from emerald city creams, a twisted ankle, legs spread wide open...climaxing. eyes rolled back with desire...or pain. glazed.   barker fucks and inferno. scrubbing dirt, and grime off in watermelon desire, sweet pungent addictive tasting...talks.   mind flies at a million miles an hour  here\'s my dumb silly post asses, i make no sense.  '),(22564,664485,'24,April,2003','2003-04-24 00:00:00','   \"i try to avoid life as much as possible. i\'m sat in my little fantasy world where i\'m quite happy. surrounded by furry zebras and bathtubs filled with brightly coloured power-tools. and the door to my fantasy  world stays lock and bolted. \"  '),(22565,664485,'23,April,2003','2003-04-23 00:00:00','   chick from emerald city rang last night.   yum.  heh.  :)  whoop.  doop.  dee.  wee.  '),(22566,664485,'23,April,2003','2003-04-23 00:00:00','   so i had nice big plans to go out tonight.. untill i passed out.   yeah, it\'s never good to wake up and realized you just knocked yourself out.  very freaky.  hm.  '),(22567,664485,'21,April,2003','2003-04-21 00:00:00','   i forgot what we call keith here, hm. how about sexypants? lol, well anyway, sexy pants and i went out shopping yesterday. very very fun, and definitely lots of interesting buys not to mention silly laughs but all good though.   we picked up a couple of things for barker, curiousgene, old man, madonna and \'chick from emerald city\'. heh. chick from emerald city story later. but yeah, my thoughts are all sporadic and spastic today, been waking up and going to bed at absolutely strange hours, and in between going home to barker and going home to bum fuck alabama..it\'s been a ride. :P  ooh, so here comes the part where i rant and gush about barker.   guess what that silly dumb sweet ass did? heh. (no, i\'m not telling them about  that  silly. :P)  we went shopping the other day, just out of random boredom on the way home from his place, and i\'d picked out this pretty little sundress and a pair of white capris to go with it. the semi indian dothi looking thing for summer you know? but yeah, stupid me didn\'t have much cash on me and i\'ve always had my thing about getting people to pay for me so to make the long story short, i only got the dress and only told that to barker later.   so fast forward a couple of days and in between kayaking, canoe-ing, tubing, and random accidents of fate, barker springs on the surprise yesterday and yanks out my pair of apparently very hard to find white capris! woooo...two points for the boy! and it the right size too!  :)  now is that just not absolutely adorable or what?  *insert a fit of silly giggles*  hmm...barker\'s on his way to orlando for a day and i\'m feeling lonely. when the \'boyfriend\' is away can the pussy come out and play? *tsk* *tsk*  okay, enough of silliness...i meant to put in a serious proper update but i\'m being distracted by seven hundred and fifty two people with four hundred and seventy six things that need to be done now.   will continue later....oh, and don\'t let me forget chick from emerald city!  :)  strange shit i tell ya, oh and bitch from the doctor\'s office too.   gah! so much to do, so little time!  whoop. whoop.  :P  '),(22568,664485,'21,April,2003','2003-04-21 00:00:00','   details with endless twists but not much time to elaborate. needless to say the weekend was absolutely strange. good and bad and weird and interesting...all good though.  finals in a week.  holy mother of god.  '),(22569,664485,'21,April,2003','2003-04-21 00:00:00','   waiting for the weekend.  '),(22570,664485,'20,April,2003','2003-04-20 00:00:00','   she hasn\'t been to bed since friday night, and at the wee hours of the morning at that. she\'s a little woozy, extremely exhausted, tired as fuck, a little melancholic and obtusely content, pleasantly appeased and somehwat satisfied. she likes the waves she\'s riding on right now, the way in which her currents flow.   she went out with the girls last night. the three of them versus the whole of pensacola and their mothers at the boardwalk by the beach. they salsa-ed, merengue-d, hip-hopped, booty danced, ass-humped, flirted, seduced and ran away. convinced though that virtually every man above the age of thirty tried to pick her up that night...must have been the asian porn star look she was going for. :P  they stopped by at blue shirt\'s first to kick things off though, chinese boy and a couple of his frat buddies and little sisters were there and in an odd sort of way, it was interesting. but as usual, three pms-ing women put together? drama. big fucking oh my god drama. nothing new though, nothing new.   typical even, to a certain extent, laughable. laughable because it was so silly and getting old? or laughable because she figured hey, let\'s not ruin the rest of the night eh...i don\'t know. questionable.   next stop on the way was barker\'s. stealing her ciggarettes back and sneaking in a rose while he was at work...which of course then went on to stopping at his work place to say hey baby. to say hie sexy. to say, just felt like seeing you. confirm plans. talk a little. watch you smile a while. :)  fuck the rebel flags dude, look what i scored.   then when all hell was done breaking loose among the girls, they headed to the beach.   ah, but i digress.  something was up with old man last night too though. so many things on his mind, so many thoughts, so many indecisions, so many feelings and so much numbness. hard to find the words to say, afraid of threading on dangerous ground, of saying the wrong things, of aggitating him more. upsetting him. i love you though old man, no matter what. and i meant all that i said last night, you are the staple that holds together a lot of our lives and so incredibly special and loved to everyone of us.   and now that that mushy shit\'s over, fastforward to after the boardwalk. and the two guys that went to waffle house with them plus the dragging curiousgene and barker\'s ass out at four in the morning for a texas cheesesteak. it was good though, all good. very cute.   driving the girls home though, it was of utter importance to stop by at bedlums downtown and freeze our asses off waiting outside the club to meet their friends. yes, friends that also left me weird strange sexual messages on my cellphone at five in the morning. i should be flattered i suppose...still weird though. perhaps almost kinda freaky.   madonna ends up staying the night and right about when she\'s done with her bath and ready to curl into bed, it\'s time to leave. easter bloody sunrise. yay. literally home, bath, change, out.   and they ask her to read the scripture for the day, a special priveledge, an honor...and so she does. then rings daddy after that to say hie, happy easter and i love you.   drags madonna out of bed and drives them off to barker\'s place, heh. :) and she found her way too!  drama, drama, drama for the rest of the day. some fun, some scary, some different and some just plain silly. she had fun though, a lot of things ran through her mind, things she is at no liberty nor strength to talk about right now. she\'ll figure it out though, later, but for right now, she\'d just like to curl up and die. pass out for exhaustion.   that would be nice wouldn\'t it?  :P  p/s don\'t you dare fucking die on me.  '),(22571,664485,'19,April,2003','2003-04-19 00:00:00','   hm..so it crossed my mind today that this will be the first easter in five years that i won\'t visit my grandfather\'s tomb. is that weird? is that strange? is that releiving?  i was painting easter eggs, and with all that time and dye on your hands you sorta let your thoughts wander around even with the tv in the background and screaming, crazy, hyperactive, adorable kids running around.  there are a lot of things that surrounded my grandfather passing away, his idiosyncracies, my grandmother, my father, the family, our doctor, my brother and just little quirks and painful remembrances that sorta hit you all at once. being at that place today reminded me of working with the kids at home and having to stay strong, stay happy, stay  altogether.   we\'ll see.   :)  in any case, have a great easter everyone.  '),(22572,664485,'18,April,2003','2003-04-18 00:00:00','   i just about peed in my pants. i thought i lost the last post.   hah.  :)  '),(22573,664485,'18,April,2003','2003-04-18 00:00:00','   \" Winter.  I.          Twist              turn                  flip                     and                      fall                    nothingness                  matters              yes -          it does...            it does               love                  is a beautiful lie,                              yes it is...                                  it is...                                      . . . II.                                      I suck the teabags                              and let its empty taste                          fill my mouth                  (with bitterness)                      My tears don\'t flow -                              instead they                                  flood stagnant                                          in my chest                      weeping for the parts of me                that die day by day          who is there left to      accept me tonight,          but the cold starless sky                              and an empty                                          coffee cup                                                       . . . III.                                                      she                                                  left                                              me                                          cold                                        so cold                                          lost                                             in this                                        meaningless                                      blizzard                              of thoughts                              and this                              indifferent                                  month of                                      December   ---   Rapunzel.  (the lady in the tower)   Rapunzel...  She looks down from her high tower  The sun catches her face and kisses it all over  She shines and shines like a magical song  But her face is tear streaked  Like a porcelain doll  Weeping saltwater from its smooth perfection   And I just stand here  Looking up at her...  Looking up at her looking down on me  She talks to me  She talks to me like she\'s talking to a frog or An ant or a rock on the ground.  I might be a lost prince  but I am a prince no less  I should have worn my crown before I left my castle  Nevermind,  I like the way she sees through me as she talks  She\'s comforted with my voice  And the knowledge that I am listening to her   She is so alone and cold up in her lofty prison  Her tower is a wicked one  It holds her so tightly  That thoughts of escape do not exist in her mind  How can I make her see  That an escape is possible with me  I could carry her away from her tower  And even though I am lost now,  We will find our way together to another kingdom   Rapunzel, Rapunzel let down your hair...   The truth is, I am not a prince any more...  I ran from my land and burnt my crown  I have nothing left to lose  And I should be happy with what I have done unto myself  But nothingness makes a man hungry  So, my fair maiden, will you trust me ?  Do you, like me, ache to run from it all ?  You seem to think that I am a scoundrel unworthy  Who will cut or tear your sweet hair  But my lady I promise to be gentle, I swear  I will be a better man,  A more gallant prince  And I will be worthy  I will be worthy...    Rapunzel, Rapunzel let down your hair...   Let down your golden tresses,  I will climb up to your loft  And keep you warm  I will break down the doors which hold you there  Together we’ll escape the cruel tower  You will be free and I will be found  No longer lost or bound will we be   So, Rapunzel, Rapunzel, let down your hair   ---   Strings.   I lay her on my chest to feel my heart beat softly with hers,  I feel her soft weight upon me, her curves molding delicately with mine,  Our bodies/souls entwine in a beautiful three chord harmony,  She sings me her melancholic melody (speaking from the very depths of me)  As my fingers make her cry out for more love.  I lie in the darkness of our room as the love we make  creates sweet sinful sounds, echoing through the still night.  At times, - she is my meaning, she is the reason I seek,  each time we\'re apart - I ache to hold her, hear her voice.  and each time I feel the world start to eat at my flesh,  I find my sanity in those six strings she wears,  thank you love - for everything you\'ve been to me  thank you for your patience with my inexperience and inadequacies  and thank you for the songs we shared, played tenderly (with love) on your strings   ---   Not one lip.   Crucified on your altar,  I forced myself a smile to brighten your day fingers ache this ledge might break and leave me  crashing- grasping for a rope to hold  myself to solid ground  you stand there an ache personified desire a wish for something to touch and feel something tangible to hold  through whispering nights  i kept my pen in my pocket  keeping the poems that would  spurt forth as vague hopes as faraway as two lifetimes a promise not even there  not even one lip and when my fingers would hold no more we hung up the lines  and sought solace in the forgiving arms of sleep needing dreams that would  pacify this need distract me from this  bittersweet sweet reality (... you bitch...) i need you...  ---   shallow.   i have sunk to the depths of shallowness the thoughts that ache in my mind a confused mixture of childish ideals and pretentions the many visions of a shallow utopia and nirvana the whispers of a perfect complete state of mind like a childish song playing over and over in my head in my small minded quest to open my mind, i cheat myself of the very maturity i seek  just because my feet do not touch the ground,  does not mean that the waters i swim in are deep indeed i am only floating in shallow waters deluding myself in the thought that i have swum far into deep oceans   ---   Never... unless you\'re sure .  Her handwriting\'s changed her point of view as well she\'s lost that shy innocence, she\'s more coy than ever her hairstyle\'s changed too she does it up the way he likes it when she smiles at me, its not the same smile i never knew a year could be so long... her letters- now yellow,  mine... she\'s burnt them i think ... he\'s nothing like me... NOTHING like me... certainty has turned to doubt to self loathe, regret and now stupid scribbled poetry i never knew a year could be so long\"   a couple of things..  monologue posponed till monday.  photo shoot posponed till....(good question)  growing apart sucks ass big time.  it\'s amazing how people write from their souls sometimes. so fucking beautiful..and soul searching. gut wrenchingly perfect, and so incredibly mesmerizing, euphoric, deppressing, sarcastic..seven hundred emotions that flow through at once. seven hundred different senses evoked at the same time.   kinda like someone peeled back layers so delicately carefully kept, a deep dark secret hidden it is somber, painfully quiet and hushed in here. so fucking fragile. so fucking careful...like treading on broken wine glasses, bleeding.    \"our friendship  no matter how loosely tied, how strained, how intangible, how fucking strong and time-tested, how down right oddly related, i treasure the fact that we are friends who inspire each other everyday.\"   get that straight bitch.   sorry, i love you, i\'ve moved on, and i\'m letting go, i\'m trying, and i\'m happy, and i\'m sad and i\'m alone and i\'m alive and i\'m so so many things but i can\'t find the words and can\'t find the right exact perfect pieces to thread them together to form sentences that will make sense. that will strike deep.   would you like this new look?  jet black hair, heavy on the eyeliner, striped nails alternating between red and black with white flowers(maybe daisies) and sparkles in the middle, silver jewelry...oversized hoops, leather cord, odd pendant that alternates between beaded chokers and randomly tied strings. long (fake) eyelashes brushed with mascara, body jewelry adorned. make up just right..almost. rosy cheeks, dark burgundy lips and gold shimmer on the eyelids. tanned, minus the lines from a tennis game in a sports bra, bitchy but most importantly, confident. almost scaringly cocky.   playing dress up by the mirror. twirling in odd ends and pieces from the bottom of the closet to the drawers shoved in between. a silk dress pulled up to be made into sexy lingerie, add a belt and it\'s glam. a twist of material here and add of a look there and something else is created. so many ideas, bursting out with creativity, liking what\'s reflecting back...having a good time. twenty \'european roses\' sit in an overcrowded vase in a far little square in a distance yet noticable. the reds, the yellows, the off-whites. \'overcrowded.\' she smiles.   gigglies are so damn embarassing, and odd.   funky shoes left in her locker at school with that little black number and another yellow tulip. she tries and pieces her schedule for saturday together and her song comes on air. she hums, sings, then oddly feels a tear slide down her cheek. good tears? bad tears? she doesn\'t care anymore, why make a big deal out of something that is so beautiful? why make a big deal out of feeling? why be afraid of feeling?  a bowl of blood, a syringe of medication, a chapter in a book, a series of events. a realization. a confirmation.   too many random thoughts to organize. too many questions to find answers...things she will never know. things she must learn in time.   she finds out for herself.   she finds herself.  '),(22574,664485,'18,April,2003','2003-04-18 00:00:00','   \"She looks at me with those...  huge, vulnerable eyes,  and all I\'m thinking is :  God - ... I can\'t reach her,  Outside looking in and ...  I must say -  I think I\'m in love,  The colourful Sundays and  the sad Tuesdays,  I want to have all that with her,  Sucking popsicles on a swing,  watching grass grow,  ...God - I can\'t reach her,  I am her lapdog,  one look - one glance and I sink  to the highest high.  or float to the deepest depths  She doesn\'t see me,  the window only lets you look in,  but the sign on the door says \"closed\",  and - God -I can\'t reach her.  I can\'t reach her.\"     it\'s funny how people hold on to things...i pulled a \'taylor\' thing and switched monologues because smart-ass keith has my notes in the trunk of his car from beach, tennis and cast meet the other night but anyway, it\'s this rather ironic, flipped out, bite-in-the-ass piece by cheryl west and i was working on it last night and this morning and it\'s significance just hit me. it\'s weird how monologues are itsy-bitsy threads of who you are inside and it takes certain things in you to pull it off.  in any case, here\'s to hoping i do good anyway.   speaking in which, the theater auditions are up again next week for the fall and spring term of 2003. i hell yeah better get well soon because i need to pull off a serious monologue and a song. not quite sure what i\'m doing yet, just throwing a couple of things in random air.   we\'ll see.   :P  hm. i talked to hung over curly sue earlier today too while switching locker things and getting my dress and shoes back. it\'s scary how alcohol is almost a staple of collegiate life. to a certain extent, pathetic too...but seeing as i\'m equally in the same boat and escaping reality or extending altered perceptions are nice once in a while...it\'s all good i suppose.  hey, in all defense, you only live once eh?  i\'ve been having glimpses of amnesia. i start to say something but i forget halfway, or i mean to do something then it slips my mind. i must be going old, or i\'m on so many things my limited brain cells can only muster enough of strength to get my ass up.   hm...the hormones or raging. the crazy bitch hides beneath the surface and the happy girl comes out to play. the childish one, not the brat, just the playful silly girl. and it\'s cute a while...you bask in her glory, you dance in the melody her voice carries...and she lets you in a while. invites you to play with her.   things are drastically changing, drastically improving in her life, so many new things on her plate to sort through, so many exciting things to look forward to. love is universal where ever you are, where ever you may be. and to others, she\'s seen heaven in their innocence before, the cherubic faces she cannot wait to get back to...yet death waits patiently close by, constantly reminding things aren\'t forever but hope shields her. so long as there\'s hope, she\'s safe...no matter which world she is in. no matter what state. no matter who she becomes.   she\'s emerging, and sheddding. and she has many to thank, the old man says begin with him, barker says what about me, puck says me too, me too, but then again, she chooses them all, but stops to look within herself. selfishly, she twirls around alone for a while, then thanks herself.   she did good this time. she gave in to continuity.   she gave in to her voices.  she\'ll make it.  her brother laughs, mocks her and says, \"oh here we go again,\"  she never liked playing by the rules anyway.   :P  '),(22575,664485,'17,April,2003','2003-04-17 00:00:00','   you suck!   :P  wa-hey...and yes, i cheated -   hé sexy  trois choses nous avons besoin de faire bientôt avant que nous partons:  Prendre des images ensemble, l\'amour de marque ensemble sur la plage, et aller danser.   Alors peut-être aller le travail et par notre eh de lignes?   Vous avez eu un bon jour au travail?   m\'appeler quand vous recevez ce bête  lol, this cracks me up.   talk to you later...go figure.   :P  '),(22576,664485,'16,April,2003','2003-04-16 00:00:00','   so, a little of today;  the procedure went, predictably, quite smoothly and it\'s nice to be able to breathe again. as terrifying as it seems, you do so much of this shit so many times it becomes uneventful after a while...minus of course the new faces, and strange buildings?   unfortunately though, i have this horrid ulcer (canker sore, depending on where you\'re from) that\'s underlining my accent and making it worst because i\'m taking extra care in trying to not irritate it. it\'s pathetic i tell ya...:P   but yeah, i also have a monologue to present tomorrow and another shot on monday, depending on, in which both the old man is going to help me with.   i. because he\'s good at that shit.   ii. well, because he\'s nice like that. and hey, i know i lot of people would willingly watch somebody jab me with a couple of needles but really, who else would voluntarily listen to me whine/bitch/complain through it eh?  oh well.   in other news, guess who went grocery shopping with me today.   and even better, guess who drove? :P  all the way through town, around town, and back!   woohoo. ten thousand gazillion points for the girl.   heh, so the chinese guy formerly known as friend-of-a-friend-of-a-friend\'s-competitor is now reintroduced back into the picture. why? i don\'t know really, just cause i guess. forgiveness is a good thing i suppose. besides, barker might have helped with that and so did the dj thing on friday night. in any case, with all the shit that went down on friday night, it\'s best to be friend\'s with him a while..just in case you know?  so we went through the list of food we needed to get on the diet and trying to somewhat eliminate the dairy products and the weird looking, evil tasting stuff, running from one end to the other because i brilliantly didn\'t bring a pen on me nor really separate the sheets. oops.   but yeah, i had fun. other than looking like a bum with shorts and a top...but then again, it\'s wal*mart so big fucking deal. :P  we bought bones for yen-yen, strange yellow tulips for barker and blue-shirt, surprise for old man (you\'ll find out tomorrow if i manage to stalk you down) and of surprise, surprise, healthy food for the bullimic.   excuse me while i keel over and die laughing. or puking.   :P   puck and madonna are off to wendy\'s and are rubbing it in my face. evil bitches.   nevermind that they got tans at the beach today with fancypants and singingboy too.   pah.   so, it\'s been a pretty good day i suppose. kinda nice, kinda subdued, kinda relieving really.  oh.   besides, i got to talk to a certain boy in yorkshire today too.   (i told you you\'d miss me bitch. :P )  but yeah, i thought i\'d insert a nice update instead of the feel sorry for myself whirly one from last night. hm. yeah. last night.  oh nevermind...   it ended up good though. good cries are always nice, especially when someone holds you through it. all that needing to be alone bullshit? just drama, random overthinkingness, and crankiness while being sick and feeling sorry for one\'s self.   :)  oh, and in other news..   \"crazy shit man. at least i have to work tonight. that will keep me out of at least some trouble. i live an amazing life. who knows if i\'ll call this girl back. i mean i should, she was fucking hot. but i dunno. i don\'t really care. it\'s kinda like fishing. the fun is in reeling them in. since i\'m not motivated by a desire to fuck every girl i see, i\'d be totally fine not seeing her ever again. so do i keep her, or throw her back? eh, not something i\'m overly concerned with. in keeping with my aquatic analogy, there are plenty of fish in the sea. .\"   lol, otherwise known as the thrill of the hunt. stability is all nice and dandy but there\'s nothing like looking hot, knowing it, and basking in it i suppose. yum-yum to that.   wait till you get to the part of bitch fighting. the shit gets better i tell you. :P  ...and here she goes on another tangent again. i forgot my point, but then again, i always do. oops.   but yeah, just to say it went okay today and i\'m alive and all that good shit. ooh...and this too...     :)  ((hugs))  '),(22577,664485,'16,April,2003','2003-04-16 00:00:00','   he wanted me to come on tonight, to talk he said. he missed me.   and so i did.   i dragged my fat yellow overweight sick ass out of bed at exactly a quarter to twelve and started putzing around.   blatantly making it obvious to be \'okay\', to protect, to make sure he didn\'t know that i wasn\'t feeling too good. i\'m sure he heard though, he always does. even before i posted, he\'s got his angels (spies), he\'s got his information (sneak attacks) and sometimes i wonder why we even bother at all.   i read an email he wrote me about a month ago. the last trauma, the last hooha. the last email i was convinced would end everything...and then it reminded me again of the other site, and how the ending of one hate mail would lead to two others. and then i wondered, why even bother at all?  i had the chills last night. chills because of my temperature or chills because of the nightmare i had, i\'m not quite sure. but someone tried to kill me, and i was duped. it was a nice man and that nice man was secretly evil and it was so goddamn cliqued i mocked it in my head but i was so afraid too. my heart pounded so fast, so fast, and my sweat beaded profusely. fear is a great fucked up feeling i tell you.   i\'m guessing, three to one, that it was because i fell asleep to \'one hour something\', that movie by robin williams and my overdrugged cranky self.   oh well. i thought i\'d write it down anyway.  i\'m feeling a little sorry for myself right now. i hate hospitals, but more than that, i hate it that i\'m doing this alone again. by choice.   or maybe not.  i don\'t know really.  i should probably go to bed and get some rest.  i should probably sleep and escape into fairy tales of love and tranquility.  i should probably do so many things i will probably never achieve.   i should probably believe in myself.  i can\'t though, not tonight, tonight, i\'d just like to return back to comfort a while.  tonight i\'d like to fall back to something familiar.  tonight, i\'d like to dissappear into my own world a while.   or so i say.  it\'s funny how people come out with theories to justify things.  to erase the gnawing guilt that drives at you.   i miss you too you dumb stupid assholic whore bastard loser fucker bitch retard inconsiderate fuckass.  and just because you need me tonight, doesn\'t mean that i need you too.  because i never need anyone remember?  because i sail that fucking ship alone.  because i\'m good alone.  because. because. because.  i   am  alone.  '),(22578,664485,'16,April,2003','2003-04-16 00:00:00','   ever notice that? that you have some of the most amazing conversations just because one or the other, or if you\'re lucky, both, are on their tippy toes and trying to impress?  hmm.  well anyway, old man took me out for egg drop soup. not chicken soup mind you, because a. racheal isn\'t exactly conventional, and b. fuck the chicken. so while i lay in bed trying to burn zank and join onyx in kitty hell, sniffling, head pounding, coughing, moaning and whining, puck and the old man work on a \'very special surprise\' (not at liberty to divulge information) while ruminating little boy bush\'s speech in missouri.   it\'s funny how many thoughts and comments can swirl through your head but not much follows through.   so if you\'re wondering, yes, i\'ve been sick as a dog (i almost misspelled it god there) the last couple days and no, it hasn\'t been fun at all. and just for the record, because i am asian and have a flu, it doesn\'t mean i have sars thank you very much. now try telling that to my doctor.  grr.  i have to go in so they can put a tube down my chest tomorrow and do \'oxygen\' treatment.  i\'m trying to whing my way out of it with orange juice and pills but it isn\'t exactly working.   isn\'t it nice to see that i still have a sense of humor though?  ooh, i had a scene due today. or rather a monologue by ken friedman but i\'ve been so shitty the last few days that when i stopped by stan\'s office to make up excuses to not present it today....he gave me a big great hug. :) isn\'t that nice? hehe, that put a grin on my face other than it somewhat freaked me out that people were talking about me in class and had told him i wasn\'t feeling too good?   but yeah, back to what i started off earlier.  alltel and i had a nice pleasant long assed conversation this afternoon...he stopped by after work to  \'see how i was doing\'  because  \'he was in the neighborhood\'  and  \'that\'s what friends are for\' . and just at the right time too because i was in so much pain and discomfort and starting to feel all bad and nostalgic and cranky that he made things better because the boy is pretty good i tell ya.   i skipped the head doctor today too, which was probably a bad idea because dealing with everything that has been happening, is happening, and about to happen, i probably shouldn\'t skip any sessions but oh well, big fucking deal.   i talked to a certain someone the other night about barker. perhaps it didn\'t really sink in till today about how i feel about shit nor what i am really doing till today. something about all the things that were said in that conversation came back to haunt me and question everything i stood for. perhaps it was the way things were yesterday, and those goddamn noisy kids and pool work at eight bloody o\'clock in the morning. or maybe...i don\'t know.   how does working with someone correlate into what we are doing? how do we keep proffessional and play apart? how do you separate director and friend? you just fucking do that\'s all.   there are levels and things that need to be done for the sake of my sanity and his, there are secrets, there are pretend to be\'s and there are backroom shit. i am not going to sit here and take crap from anyone for anything because i feel the need to do things that make me happy or things that i like so fuck you turds, but thank you for your opinions. you have yours, and thankfully, i have mine too.   do i know how things are going to turn out? hell fucking no. do i understand everything that\'s involved and how things are going to work out? of course not. do i confidently know that i have all my shit together? you stupid?   do i still have open-ended unanswered questions? hell bloody yeah.   does that mean i just give up and say oh well screw this and continue my existence pathetically knowing that i\'m a wimp and a baseless gut? i\'d rather prefer not to.   god, i don\'t even know how this started, or how i went into my hour long debate with alltel because it suddenly dawned on me all the questions that had plagued barker, and i figure if anything, i would get constructive bullshit from alltel , so i sat there sniffling at the roses he had brought over (but i couldn\'t smell), coughing, lying around in my sweats and watching some ass weird 80\'s movie looking very darn attractive i\'m sure, and we went over baseless thoughts and backed up voices. or rather it was me ranting, screaming, crying, shouting, laughing, yelling, talking, as he pieced together words quietly, calmly, and confidently. these men i tell ya, so damn irritating.   :P  hm. i know i dragged my lazy butt off the bed to write soemthing good here, how i felt and what it\'s been like the last couple of days. but then as usual, i went off to a tangent..and now i don\'t really know what i was trying to say or what i started off to write about except maybe, you know, i don\'t know what\'s going on in my life, i don\'t have it all pinned down, and i don\'t have it all neatly arranged in categories. but hey, i\'m okay, and things are pretty good for the most part.   i am still (barely) alive even after that little emergency scare, school\'s still going well despite missing some classes (friends are great aren\'t they?) and though certain issues aren\'t resolved...i honestly think they never will be. i just have to remember it isn\'t the end of the world when my heart hurts and the tears weep. yet, even with everything, i hope some day he\'ll know it was real and that love, no matter how tainted it ended up being, was real.   and now that i\'m sappy and indecisive (fucking libra) i\'m going back to bed.   goodnight all.   (thanks for the mail)  '),(22579,664485,'14,April,2003','2003-04-14 00:00:00','   cast meeting + auditions tonight.  hm.    \"Why do all superheroes wear spandex? If a missing person sees their picture on a milk carton that offers a reward, would they get the money?  Is sign language the same in languages other than English?  Why do you have to \"put your two cents in\" but it\'s only a \"penny for your thoughts\"? Where\'s that extra penny going to?   How is it that we put man on the moon before we figured out it would be a good idea to put wheels on luggage?  What would happen if: Everyone was to flush their toilet at the same time?  If the weather man says \"it\'s a 50% chance of rain\" does that mean he has no idea if its going to rain or not?  Is there anything easier done than said?  What do you say when someone says you\'re in denial, but you\'re not?  If you own a piece of land and there is an volcano on it and it ruins a nearby town, do you have to pay for the property damage?  If Pringles are \"so good that once you pop, you can\'t stop\" why do they come with a resealable lid?\"   ...stolen from  urlLink joy .  '),(22580,664485,'12,April,2003','2003-04-12 00:00:00','   \"a random thought about the search for the right.  today, sitting in the desert i had time to think about why we are here. after seeing the war at first hand : both sides of a struggle for freedom and pwoer, i know now. the desert has a way to isolate you in a crowd. it\'s desolating heat, it\'s barren sands can make you feel like no one would ever want anything but leave it\'s wrath yet people are here, and they want to be here. it is home to them. that is when it hit me...a memory hit me like bullets to the heart. it was so powerful i began to see it all very vividly through my mind\'s eye.  i was remembering a convoy that i was on. it was a resupply run in a safe zone, there was no fighting there, just a lot of scared people from both sides. whole families with nothing left but what they could carry on them. soldiers and marines from the allies huddleing wherever they could to get out of the blistering sun. it looked like something from the bible, bodies were sprawled out in the shade to keep cool and empty water bottles were all over the dusty desert. exhaustation had set in and all of a sudden, everyone of the soldiers and marines got in their vehicles and took off like lightning had struck.   it was time to go to another fight.   all i can say is wow! it was amazing to see the dedication to the cause. those people later lost some people to the war. they were tortured and murdered...that is all i will say about that.   what i remember though is this, these men and women shared their precious shhade, their water and food with the native people who had congugated around them, this was not for show. it was the right things to do, even more impressing on me was the fact those allied military fighters didn\'t have any extra food and water, and they didn\'t think twice about sharing what they had. that was the right thing to do and that is all that matters.  we are here to do the right thing, no matter what anyone nation or world organization thinks we are here for. politics mean nothing to those with no freeson. the oppression these people endure are of the most disgusting in the history of the world, bar none. the search for the right ends with a rebirth of nation: it will be a sweet nectar of freedom and liberty. the search for peace has come at a high cost, and that cost comes in blood. but it has always come in such ways. tirany will fall here, and for good.  crying eyes are dried by the dusty desert reminding us there is no time for tears for those we have lost, there is still a job to do. and we will complete our task to give this country back to it\'s people. during this war, the blood and sweat has poured out of our bodies to flood this land with freedom, it washes our souls of doubt to see it first hand.   war has one enemy; our memories are all too short. we have to remember why we are here and what we do here. it is not a pat on the back, it is a lesson in history. then and only then can we take the time to grieve, when the lesson i over. it is time for out country to grieve for us not, we will remember our fallen comrades now and forever which is a duty we bear and cherish readily when the time comes for us to do so. but for now, we live to fight for all the people of the world to know the freedom they deserve as human beings.   when the tears stop falling, when we are back home and peace has come, we must all the world realize our responsibilities to one another. no matter what the causes, the effect of this war, the casualties of it do not care. they are dead, do not forget them or the lessons learned from this conflict. make the world a better place with the blood shed on this battle ground. consecrate in your mind to always try and do the right thing. it is the duty we have as human beings to not only appear as good but be good. every minute of the day we must ask ourselves, what can i do to help me fellow man.  sometimes it will be hard and we will make mistakes, just find to find the right and never forget why. we, the people who protect our freedom will always know the price of freedom. that is why our mottot is so important to us in my line of work, it is the rescueman\'s slogan, \"these things that we do that others may live,\"  to add to it makes no sense so i won\'t.  -SSgt Kirk L. Bobleter USAF\"   i\'m sorry i didn\'t understand earlier. but today, i stood beneath the yellow ribbons we put up for you and said a prayer. the whole neighborhood is flooded with ribbons, from the front mount of our addition to the back end and the dock. uncle bruce prays that mark comes home safely too, and we watch cnn daily, perhaps hoping to catch a glimpse of you.   i watch, sometimes hear her cry at night, she misses you. you\'re her youngest, her precious one. the naughty cheeky ass with a sharp tongue and it was silly, but the other i thought it would be supercool to just have you come home. but then i realized you would never ever do that, you\'re great at what you do. and i think it is safe to say, i am fucking proud to know you. you\'ve watched me grow, turn into the little bitch i am, and i am amazed.  i respect you because you\'ve changed from playful idiot to matured soul. wow.   take care you and don\'t kill me when you read this.   p/s i figured you might like the extra attention anyway.  go save some lives and come back home safe okay?  '),(22581,664485,'11,April,2003','2003-04-11 00:00:00','   so last night...  last night..  hm.  how do we put this?  lets just say  i. frat boys and pledge bitches. (i love you)  ii. making out with puck.  iii. having unbelievable great drunken sex with barker. (ow, ow!)  iv. chicken mc nugget trading while driving with number 201 random cute guy on the bloody road!  v. cameras, video taping, crazy ass shitty things you need shots for.  vi. puck\'s boobs.  vii. blue shirt\'s goggles/shades.   viii. titilated braless madonna and open containers in a parking lot.   ix. old foes make friends.  x. white boy dance.  oh my fucking god, i\'ve turned silly, stupid, blonde, and retarded.  but i had fun so hey, all good honey!  :P  yes, last night was an interesting night too. hell, this week has been fucking interesting.  and now that i\'ve repented for all my sins after the two and half hour long rehersal at church (palm sunday performance tomorrow!) that started at bloody nine o\'clock and had my ass up at eight, i am amused.   ah, but i had to leave sleeping beauty in bed as i left this morning though, grr. ooh, and we had great plans to go down to the beach but rehersals ran late so no go and we need to rest tonight so hah. i think i might go lie on the grass outside and write though, i don\'t have much to say but i figure it would be nice to capture these feelings on paper. for rememberance and all that good shit eh?  i\'m rambling.   heh.  i haven\'t thrown up in two days, the last time i cried was thursday night and i have a big ass grin on my face. my mind is pretty sane and i don\'t have a hangover, i\'m not on drugs, nor am i high right now. this is all me baby, this is all me.  wa ha ha ha ha ha ha ha ha ha ha ha ha.   :)  good streaks fucking rock.  '),(22582,664485,'11,April,2003','2003-04-11 00:00:00','   barker says:   \"When a guy likes a girl and a girl likes a guy but knows that the guy likes her, the girl no longer wants the guy until the guy no longer wants the girl, THENNN the girl once again wants the guy.\"  \"Skin color based racism..that\'s so absurd it\'s almost quaint!\"    \"The funny thing about the word \"LOVE\" is that there is such a large range of emotion involved in it that there is only one word to describe it.\"   R-S  '),(22583,664485,'11,April,2003','2003-04-11 00:00:00','   barker is so damn sexy and we need to go to bed.  '),(22584,664485,'11,April,2003','2003-04-11 00:00:00','   girls are so damn   odd  strange   vicious   indecisive   needy   beautiful  insecure  loving  selfless  bitchy  selfish  honest  amusing  gentle  amazing.  perhaps that is why i am attracted to them...because they are sensitive and everything at once. because they aren\'t afraid to show you when they are hurting...because they come to you when they\'re hurting. no matter how hard, how thick, how goddamn strong the wall, girls break. get broken. cry. seek comfort.   girls get honest and strip layers and make love to you in the simple more passionate trusting amazing way.   giving.   the way the hip bones are just right, bony but not malnourished, but then again, i am the bullimic and thiness is beauty and strength. the way their lips swell up and breasts beg to be touched, be taunted, be teased.   for some odd reason, i am attracted and aroused by this. by some very odd things i\'ve discovered in the last hour. perhaps it\'s the softness in the way she writes, and i can imagine her lips, and the words that form on her mouth. her voice echoing, her body naked next to mine. i think this must be highly illegal somehow, because the tension is great. because the tension has been there for more than a year now, and to have her say it out loud is almost dangerous. like playing with fire.   out in the open.   because i\'ve just realized that i desire, lust, want to fuck her probably as bad as she does. herself and me included, self-pleasure. give in to the glories of getting wet, yet i will miss the entrance. but then again, that\'s what dildoes are for i suppose.   if you can\'t tell, i\'m horny as a jackrabbit at the peak of the season. it must be the hormones i say, it damn well fucking must be. because right now, i feel like a nymphomaniac. which is almost odd and ironic considering the week i\'ve had, of absolute amazing crazy highs and down right shitty depressing lows.   i have a fuck buddy.  i try and ruminate this and put it into my categories and i think i actually might have.   and you know, that\'s sort of fucking cool.  there was a 6963 guy this morning, and i put out my ciggarette by the entrance way and talked to this small slightly older man (60\'s?) that just wanted to chat and probably just needed to a friend, i reached in and did one of those crazy moment things and gave him a hug. he looked like he was going to cry, like he hadn\'t had a hug in a while..and my heart just nearly about broke. little things i take so much for granted, i have love, i have laughter, i have tears, i have life.   and i told him to take care, stay safe. and i felt good, inside, i felt amazing. like my sorry ass had did something good, and all i really did was listened, shared a ciggarette and gave him a hug. but it felt good, and i felt good, and i know it sounds retarded, but it was sorta magical.  but hey if anything, i also learned about alka-seltzer and curing hangovers?   i imagined someone doing that for my grampa, and that slid a tear down my cheeks. i tried to brush it away, before it became this full blown incontrollable nostalgic racheal thing and so i rang him. and it made him so happy. i rang to say i love you, i rang to say hie grampa, i miss you.   and that was kinda cool.  then i went for classes. and sat a while as the lecturers brought me on a journey of myths and dreams, and hopes and needs, and wants and desires. and i thought to myself, life is pretty damn something right now, life is pretty damn beautiful.   i went for my last class, and handled drama in between. yelling girls on the phone, each screaming to tell me about boy a and boy b and boy abcdefg from last night. each getting their own spotlight, and fifteen minutes...but the boys wanted a share too. and so another fifteen minutes gone by, of strange inconcievable rambling and bullshit, somewhat comedic value but then i realized i enjoyed this. that i was someoen to come to, that i was important enough to someone for them to want to tell me about their day...and i thought to myself:  pretty damn cool.   and it was a cool day too. it was fucking gorgeous, it was cold but not too cold, sunny but not blazingly hot. i looked good and the strange thing was, i knew it too. i had earrings just right, a platinum flat necklace with the twisted \'artistic\' pendant, heels, a very fucking great bra (thank you victoria and silicone), a halter top just right and pinstripe pants. and not to mention the shades old man and i bought the other day...i had numbers flung at me, hello\'s and how are you doing. and now i\'m going to go against what all the lying bitches tell you and say, hell yeah i liked the attention.   only that here comes the ironic part.  i gimped out on a date and somehow ended up back at barker\'s place. refusing to place my ass in anymore shit that i\'ve already flung at myself. it was nice to just lay there a while, a breathing, beating, pulsing body next to me, on top of me, behind me, under me. it was nice to feel wanted, desired, loved, needed, craved, lusted after.  not just an object, not just a thing, not just a girl with tits and a hole, but an actual person.   and it felt good...like things fell into place a while. as if all the thoughts that swirled constantly in my mind got trangressed and translated into little tiny cubicles and lighted pathways.   :)  and so i smiled.  did some homework.  rode back home with the old man and powerpuff girl who tells me she\'s waiting for her angel, her wings to fly. it\'s amazing how much innocence a five year old can have, it\'s amazing how old her soul is.   puck is yelling on the phone to get off my butt and go party.  there goes my great friday night plans of vegging out in front of the tv in underwear and magaritas.   :P  i\'m sure i\'ll update later...there always are stories when puck and madonna are involved.   have a great friday night!  '),(22585,664485,'10,April,2003','2003-04-10 00:00:00','   there are mornings where you wonder why in the hell you woke up, ...this would be one of them.  '),(22586,664485,'10,April,2003','2003-04-10 00:00:00','   \"yeah, you capture my sentiments exactly. We all seem to be pretty much screwed and for the same reasons, work, work, work, work. Doesn\'t matter where you are, it\'s the same thing. God, this place is crazy, it\'s living hell. There\'s so many things going on at the same time that sometimes I just feel like crying, in fact I have cried, I\'ve never been this stressed before in my entire life. I have the most wonderful relationship of my life and I have hardly enough time with him. This college makes you look so old. What was it again in the brochure..\"a highly challenging course designed for motivated students\"....HAH...HAHA...HAHAHAHHAHAAHAAHAA everything suddenly seems so funny. You\'d think in a college with only 200 students, chosen from all over the world for their excellent outstanding character and abilities, you would have no thefts, no people punching holes in the wall, no stupid little petty things like that. And then everyone will sit in the dayroom and complain about it, but it still goes on, and all these little things, you can\'t get away because you\'re always stuck on campus, your life becomes the campus, you get sick of the same people, seeing the same faces, hearing the same whispers about who and who is sleeping with who and who, because everyone knows everything about everyone on this campus, including the faculty members including the principal, who fucked who last night and where.  Hypocrisy, hypocisry, the amount of competition, 200 power hungry individuals who never knew defeat for the greater parts of their lives, were always the top in everything back home, crammed together in a tiny sardine tin and marketed as an \"international experience,\", a \"united world.\" And the war starts and in this place the only thing you hear are anti war statements, nobel laureates come to speak against war, bad bad bad, but never a war supporter. Only pacifists are allowed on campus, brain washing, brainwashing. And there are good things about this place but at the moment I can\'t remember them...but of course they are, and even if there aren\'t, how can it be a waste of two years of your life ( i haven\'t even gone to university yet!) when you met a guy that you know you\'re gonna marry?   but then again that\'s what you always think.....  it is the best and the worst time of my life too.   Then of course there\'s the personal realisations of the inadequacy of yourself and the discrepancy between who you lived with for 19 years and who you are now, and what you are, of dreams burst and of naivete lost, of childish confidence gone replaced by a disillioned reality that is true but breaks you to face it, and you need to start learning how to walk all over again, which is the good part, and the only support you get is from him.   well now since i\'ve done my share of depressive outpourings..time to go do depressing lab reports on the depressing experiment on the quality of hydrogen peroxide in Wal Mart...  Florida ain\'t that difffernt from New Mexico huh. Love, Rina \"   you are beautiful, amazing, intelligent and brilliant rina  ...inspiring.  loved.  (very much)  '),(22587,664485,'09,April,2003','2003-04-09 00:00:00','   \"Does he have any idea, I wonder, just how blessed he is to be  allowed so close to one so precious as you? \"   i wonder sometimes, but then i heard it in his voice tonight.   it made me cry, but not in a bad way. okay, maybe it did hurt a little. and maybe we are going apart some, realizing and searching our own destinies.   i missed you though. falling asleep in your arms. maybe for a moment, i could pretend i was everything this void in me needed to be filled. and maybe i could be everything your emptiness needed to be completed.  i honestly don\'t really know. the flight out to la is booked, it\'s been arranged down for quite a while...except maybe i just didn\'t get round to cancelling it. except maybe i didn\'t want to do it, i don\'t know. you see, it is that same exact feeling i get when you talk about ena and rhonda. it\'s hard on my part too, i don\'t know why. i shouldn\'t be affected really, we\'re just friends.  you and i. a little more than just fuck buddies, friends with benefits. special.   i don\'t know.   it\'s really just a name. a title. a definition. a little box with a catergory.   things have changed. drastically. i remember how naively simple and silly things use to be, then how it goes intense and crazy a while, then maybe a nice type of comfortable and now a weird sort of imbalance. but we take it as the day comes, and another sunrise rolls around, another wave crashes on the beach and another seagull swoops down to knock over an ice-cream cone.   life does go on when you aren\'t looking. when you aren\'t paying attention. when you don\'t expect it to.   i hate this feeling.   happy, sad, in-between and content.   but you know what, everything happens for a reason.  and you are my reason.  i want to believe...but then again, i wonder if we made a mistake too?  '),(22588,664485,'09,April,2003','2003-04-09 00:00:00','   someone say insomnia?  :P  but then again, after a night/day like tonight/today...it isn\'t that surprising. it might take a while more to calm down from all of today\'s events. suffice to say,  a lot .  some good, some bad, some awful, some ridiculous, some amazing, some silly, some retarded, some stupid, some brain-numbingly dumb, and hell, some just plain beautifully inspiring.  kinda like a wow-huh-fuck day.   i\'d go into details but i might have to kill you.  heh.  :P  '),(22589,664485,'09,April,2003','2003-04-09 00:00:00','   saat-saat begini, ku terlintas kepada malam-malam yang terbiar di dalam ingatan dan bisikan hati yang menyerapi fikiranku.   ku sememang-memang cinta padamu tetapi kian terhempas harapan kehidupan mu di dalam dunia ku. ku teguh, sabar, melintasi perjalanan ini yang mengusut-ngusut, tergores, mencabar.   menghitung kikisan-kikisan yang ku terhempas, jatuh..melukai diriku.   besar kemungkinan ku tidak akan memasuki pintu duniamu lagi, besar lagi harapan yang ku tidak akan membiarkan diriku mencintai anda selain daripada tempat yang jauh. tidak pintas. tidak terus.   menakuti, ku kian tidak tahu dirimu apatah lagi, anda kepada ku lagi. sememangnya, wajah kami and air muka tidak banyak berubah tetapi pemikiran, hati, jawapan kini kian kompleks, dan tidak nyata lagi.  ku rindu malam-malam and harian dibiar bersamamu. ku rindu menembusi persona dirimu, mencabar pemikiranmu. dan walaupun air mata kian melintasi mukaku, ku tetap berdiri tinggi, ku tetap menari tarian bidadariku dengan hanya bisikan yang terlepas..  ku tetap and menyintaimu selama-lamanya.   ku rindu padamu.  ku inginmu tahu.  ku perlu menyimpan.    -yes, the malay is rusty, i know.   bite me.  '),(22590,664485,'09,April,2003','2003-04-09 00:00:00','   \"I got home, so tired last night. So tired I wasn\'t even hungrry.  In the garage, I parked the car, closed the door, and started to get out. Then I realized. The car was full of you. A delicate fragrance that pulled me back inside. I sat in the car with my eyes closed, smelling Racheal. Then I remembered: you left your clothes in the car. Your perfume, barely noticeable, had made itself home in the air, my air.  I took the bag inside, got undressed, and crawled into bed in the darkness. No TV. No computer. No light. The blinds drawn on a cloudy night. Darkness. I took your clothes out and laid them neatly on the pillow beside mine. Closed my eyes. And the sweet, subtle scent that I know to be Racheal Khong filled my head, my bed, my room, my world.  I suppose I\'m a bit envious of him. He whose fingers have traced the soft contour of you breasts, the firmness of your nipples yearning for his touch in the night. He whose hands have held you so close in such darkness and found there way to every secret place on your body. Does he have any idea, I wonder, just how blessed he is to be allowed so close to one so precious as you?  I drifted off to sleep a sleep that I rarely know anymore. I didn\'t dream, that I remember. Not even my Racheal Dream, of you and I lying in bed together, the only sound is your voice, talking. Whispering in the night. I slept for six hours, straight. I never sleep that long anymore. It was \"the sleep of angels.\"  And now, I have awakened. So relaxed. So refreshed for the experience of it. A slightly sad, silly smile on my face at my own childish romanticism.  Thanks, Racheal. Just for being you, and letting me into your world, to the extent that I\'m allowed. Despite my constant, clumsy attempts at pseudo-seduction...it is enough. I am truly grateful at being The One in your life, every once-in-a-while.  Thanks.  With Love, e=mc2\"   no, thank  you .  it\'s strange to hold love at a beckon, to have it laid spledidly at my feet. it makes me blush and giggle, scream and laugh.   but still though, i thought i\'d say, it\'s an honor to be held up so high when i know you\'ve seen my god awful thrashy bitchy ugly moments.  thank you for being just you.  ...here\'s your ten points.  '),(22591,664485,'08,April,2003','2003-04-08 00:00:00','   strange headaches otherwise known as \'the morning after\' are uncomfortably assholic, like a wedgie of your favorite pair of sheer knickers stuck riding up your crack...the soft feel of it rubbing against your box and the god-awful feeling of your ass being split into two. the lesser of two evils. the generalization of an extreme.   i have layer-y soft flowing black dress on and it fits just right, though i could stand to lose a couple of pounds more...shed off some excess crap from late night \"half\" waffles with whipped cream and the bloody huge lunch i\'ve had this afternoon. of quesdillas and mexican rice from los racheros and the whole bunch of us just like old times, perhaps missing just a few. nonetheless, it was good. and i feel good. and hell bloody yeah that feels awesome.  :)  i wish i could take a picture and show you the world from my eyes today. but the laptop cord has breathed it\'s final intake and so the programs are all in it, it is an amazing day. cold nonetheless, but beautiful. inside, my heart smiles and dances to wonderful songs played around me. i join in and partake in it as it\'s conjoining rhythms draw in more people. a happy crowd of angels surround me, a wonderful circle of friends.   alltel rings during lunch and wants to know what i\'m up to tonight...oh, nothing much i say. and smile at the nervous polished voice he resonates. i could play this game but i won\'t. i\'ll take the backseat on being a slut this week. i take a backseat on jumping head first into obscure strange trite relationships. relationships? hah.  it\'s funny how i seem a whole shit load more attractive when i\'m not dating.  it\'s ironic how the numbers come pouring in when you\'re tired of that game.  it\'s stupid how the timing is never right...but hey, perhaps the epitome of life is getting things to corrolate just exactly in time.   hitting your sharps and flats at the right beat, in the right frame, at exactly the right bar without a metronome. knowing exactly what comes next and how it goes.   here we go again, up this rollercoaster ride.  it\'s not that bad, really, it\'s not that bad.   i could stand to like myself a while.   i could get used to being happy.  :)  '),(22592,664485,'08,April,2003','2003-04-08 00:00:00','   nynkempoop professed undying love to me today.  over a lesbian dominoe club gathering at starbucks.   then we had dinner, and i got romanced to a grand she-bang(in more ways than one) and while pushing around the leafy greens of my salad, ruminated over what was to be called the situation of barker, alltel, shitforbrains, sera with an \'e\', and . . .   i came to a nice long empty blank.   which i suppose to many, might be the beauty of it all. of hanging your hearts and sticking it with clothes pins while memories are being crocheted onto it. run, dazzle, frazzle and dance with the flow i suppose, though my opinions vary in awfully trite conflicting terms when it boils down to a decision. decision? what fucking decision?  i\'ve had a caffeine overdose in the last few hours, starbucks will do that to you. starbucks and sobering up always does apparently.  my lazy ass needs to get deported.   i worked on my monologue with nynkempoop and it was beautiful. creative energy spent together is always good what more a very intellectually artistic mind. i\'m making straight a\'s in a couple of classes which is funny considering the shit sense of studying i can attribute to it. i spend so much damn time in those four walls it isn\'t even funny anymore. an amazement nonetheless, but definitely way too stuck up the ass to be a laughing matter anymore.  the girl gets serious. now isn\'t that scary?  so while we shared a cigarette and slammed poetry off one another it became fucking apparent that my tits were getting cold in the wind. not very funny at all thank you very much. and seeing as the old man had an encounter with racheal and evil purple pants today, i figured, oh how fucking great. inebriated is a very smart way of dealing with things, what more, drastically simple. and so, we laughed it off. our rumbles, snorts and giggles carried away into the wind that smelled of freshly cut grass and rain.  isn\'t the smell of rain just absolutely orgasmicfyingly fantastic?  hm.  it\'s getting cold again, the war coverage seems to be dying down, i\'ve lost an hour and the azaleas are at the end of their bloom. my mom flew to london today, my dad\'s awful lonely, and my brother has a broken heart. barker\'s at work again, and puck\'s been on the phone for an hour with me. the old man went shopping with me today and madonna\'s kidnapped my cigarettes. noisy asian bitch(i lost by two votes) is dating someone new and boyfriend-of-the-week-who\'s-forever is somewhat making new major happy, infatuated.   the battery on my cell is dying and so it emits these god-awful annoying beeps and shrieks every couple of seconds which isn\'t helping the headache any. i\'m hyper and tired as hell so i\'m rambling at the flight of an express train and the swirling chopping rotators of a lifeflight helicopter. my eyes are peeled on a special documentary about starving kids in africa which is pretty darn ironic if you ask me.   the stuffed snake that\'s been fucked on countless of times sits lifelessly in the corner of my bedroom it\'s bright orange and black hues contrasting and it\'s slithering bright red tongue mocking me...and to think we were going to give it away to the children.  unfinished sketches and binges of crayon/oil-paint bullshit lies scattered throughout the bedroom, some hiding at the bottom of my closet and the other pair of trainers i\'ve been searching for and another covering a pair of lacy knickers from miss victoria, a gift from yamato. what fucking redundant bullshit i write as this hour, as i survey and delete the last couple of whiny, depressive, silly posts....how immensely great it is to feel this power in your hands. the fantastic control of a click, gone forever. history erased. new pleasures discovered.   lavender lotion, that makes me think of his lotion and it\'s purposes. a long warm bubble bath complete with candles, a book, and funky weird jazz music from the ex\'s cd that makes me think of fresno. and him showering. and the talk of cold water and it\'s effects one night. weird, strange moments of different relationships, the good things, and the happy laughy memories that continue to shuffle in my mind like tim missing all his connecting flights, and the taxi cab window reflection on my case and fucking pachebel at three in the morning over naan and fucking tandoori chicken.  terence throwing me onto his shoulders and spinning me around and around and around till we all fall down and tumble down wet grass holding on as bloody tight as we can, sera\'s infinite gift of beauty and how warmth, sunshine and  niceness flows from her and being made love to with waves crashing just outside the window. nigel and watching the stars at that funny park just outside of the towers with the harsh city lights behind us and the stars dancing around the moon..the fountain splashing, lighted next to us. jared and starbursts and waiting up at the wee odd hours of the morning stuck in a hotel room trying to keep drew away and wanting so damn bad to jump on him.    jae and her god damn sexy body dancing in the middle of a closed restaurant with only the two of us as she sings to me and tickles me to the ground. look up eyes that light up and you\'ll see hers there. steve and his silly goddamn ass in the studio looking out through panels of glass, screaming on the microphone and purposely messing up voiceovers, failing miserably at keeping sexual tension at bay. chris, mr tough guy, saying seven hundred and fifty two things at once with tears running down his cheeks in the middle of fucking starbucks and that damn stupid cat that appeared out of nowehere and made all things right. screaming i love you in the middle of the mall and the christian dior show together. pietro and positive tones and all that it meant working there and sacrifices, and emotional drama work bullshit to deal with and taking a breather and understanding the thin, very fucking thin line, of friendship and \"friendship\" and the lengths you would go for your job. franklin and his yamato cooking skills making dinner and pookie jumping around us on the loveseat watching that movie about notes on skin and forgetting things and falling asleep to his heartbeats and him stroking the nape of my back.     ryan and that goddamn cute silly mood he gets when he\'s happy and that bloody piercing of his. the way he\'s concern with what you want and how you want it and the moans and whimpers, growls and touching bare skin on bare skin caressing. the way he\'s strangely naive and so fucking old at the same time, the ways his eyes are of a little bity boy and an old wise man talking, dreaming, hallucinating, and fighting. whispers through the night and singing as you go to sleep, excited. so very different from the other, coming from two very different worlds, it\'s funny how very similar things sometimes are.   an odd stage. questioning. of knowing yet having a somewhat grip on things, an odd balance.   i think i\'ll keep you around for a while, you make me happy. laughy. silly.  do stupid things and have \"intelectual\" conversation without being too ackward.   we\'ve gone around the block and back, i don\'t know what the hell love is anymore and frankly, i\'m not sure if i care to find out. i have scars and a whole slew of shit to prove, old man gives me every damn disorder there is while you blatantly disagree with the idea of it, we swing from one end to the other, another extreme but from captains blog and mine, i watch how things stabilize and i know where i am, where i\'m going and how i\'m getting there.   i am who i am. you are who you are. he is who he is. they are who they are. and you know what, not one of our single solitary asses can change that.   so at three twenty in the morning with only two and a half hours left to sleep, an exotically blended orange dying out cigarette on my ashtray and diet bloody ginger ale on my desk with the fucking beeps from the cell i say; welcome to this brand new wednesday bloody morning, may it grab me by my tits and swing me by the balls and may it be a fanfuckingtastic ride.   bring it on you fuckers.   i\'m ready.  i\'m bored.  i\'m confident as a bitch.   i\'m crazy as a loony.  i\'m letting hell freeze over.   i\'m getting my share of a piece of ass.   i\'m \'getting my shit together\'  i\'m doing good.   \"now back to you, tom.\"   ....IT\'S QUANTITY NOT QUALITY DAMNIT!  '),(22593,664485,'08,April,2003','2003-04-08 00:00:00','   nynkempoop and i had dinner together.   hm.   yes, hm.   :P  '),(22594,664485,'08,April,2003','2003-04-08 00:00:00','   i might have seventy things to say, yet fifty will be censored, twelve will be cornered and the remaining eight will remain unspoken. perhaps just a half given away here and there, perhaps just a thought shared in between.   he says i come alive when we are alone. that my bariers fall and my masks break...i wonder sometimes too. subconsciously and consciously...has it become routine?  '),(22595,664485,'07,April,2003','2003-04-07 00:00:00','   she left this morning.  yuck.  :(  '),(22596,664485,'07,April,2003','2003-04-07 00:00:00','   she\'s been questioning, yet she hasn\'t quite told anyone. been noticing differences...perhaps the rose colored glasses are broken now and visions are fixed, sharpened.   they talk about it into the wee hours of the night, exchange opinions, provoke thoughts.   understanding a little better..allowing herself to be a stepping stone.   you never know, the voices whisper...even if she is quite certain...but it\'s okay, they\'ll wade it through.   instances; now she can blog about it.  '),(22597,664485,'06,April,2003','2003-04-06 00:00:00','   days i am needy, forgiving, wanting, fulfilling.  days i am sick, tired, deppressive, manic and defensive.   days i am just enough  ...and a little bit more.  '),(22598,664485,'06,April,2003','2003-04-06 00:00:00','   the head doctor always ends our session with, \"how do you think we did today racheal?\" and sometimes i say, bloody great..sometimes it\'s just an okay...and sometimes an i don\'t know, i\'m not sure. days range from ups to downs, manics and depressives, goods and bads, okays and normal...but what is normal really?  i can feel the chocolate-chip cookie i\'ve had after bingeing act up. i know i\'ve thrown up tonight, perhaps that is a worst feeling when you\'re watching events that are so close to heart replay on television. behind me, my mom packs to leave tomorrow. she\'s watched the documentary in her room and me mine. i couldn\'t watch it together; it was too hard. and she wants to know if i have anything to say to her, if i want to talk or anything...and the only thing i can do is say no, is to deny it tonight. i\'m alright mom, i\'m okay. don\'t worry.   nobody worry. nobody freak goddamnit i am fine. i am alright. i am okay. i am normal.   everyone has fucking issues with it. barker, old man, puck, madonna, the whole bloody world and i want to scream and scream and scream. cry. leave me alone, leave me alone, leave me alone.  she suggests the treatment center, psychiatrists, slaying the demons head on and i am worried, afraid. what\'s the point i sometimes say...what\'s the point when it always returns anyway. when i know it will come back...when it is a fact.   so many issues tonight, so many questions. and i\'m going in tomorrow, one last session with the mom. and i\'m afraid of the things we have to talk about, and i\'m worried about the shit we have to go through, the wringer i know that i will be made to step in. and i am nervous, anxious. sweaty and fidgetty...what do i do now? what do i do?  i need to cry.  i need to go throw up.  i need to go be normal.   i need to hide.   ...but i\'m okay, really.  '),(22599,664485,'06,April,2003','2003-04-06 00:00:00','   i\'m feeling bitchy, horny, fucktastic and amazing.  funny how a phonecall, an email and a good day out changes everything.   hm.   :)  '),(22600,664485,'05,April,2003','2003-04-05 00:00:00','   so in comes the sexalicious alltel guy again.  once a month, i get my fix of gq goodness, american charm and bankruptcy applications while bringing my fat yellow ass down to the store to get the half a thousand down. only that today, we come to this big fucking discovery that allbloodytel has been mischarging my account for the last three months.   ironic? coincidental? serves my ass right?...i don\'t know. i\'ve never been one to scrutinize payment to payment and so i just assumed i was paying for my overcharged minutes and long distance roaming, not really paying attention at all to the vast differences in accounts nor the huge surge. stupid me, i know.   but hey, for consolation, i got a 300-400 dollar rebate, a new fucking amazing plan, person eating out of my palm plus a confidence/happy booster.   puck, you would have been proud of my damsel in distress foreign bitch i had going on. excuse me? but please talk english, no minutes, mobile, long-distance, wireless, international, nights and weekends, morning, all time, bullshit jargo talk thank you very much.   heh.  sexual tension is a great feeling.  '),(22601,664485,'05,April,2003','2003-04-05 00:00:00','   having  losing a whole extra hour to ruminate and let thoughts and opinions swirl in my gray mass is evil. pure unadulerated evil.  so we\'re..i don\'t know what we are really. and perhaps, that is the beauty of it. that it is undefined, uncategorized. not quite a relationship yet a little bit more than friends; dating, yet not exactly going out together...but somewhat exclusive too.   it\'s nice in an odd way...but it leaves the loose ended strings untied. easily returnable to the past, creates space...second thoughts. thirds and fourths.  nostalgia is a bitch.  '),(22602,664485,'05,April,2003','2003-04-05 00:00:00','   he rolls over, sun-washed and bathed in morning rays, sunlight piercing through the blinds, birds chirping and radio going off in a tangent...blowing. awakened by my half-asleep fidgetting and ciggarette smoke from comfortable stupor. fingers trail bodies and lips meet each other; saving, reaching, consoling, finding.   unconscious mouthing of words, lyrics as the fall asleep spooned in together, talking. questioning, theories bashed around and sides taken, stances. he talks about her diseases, she talks about his idiocyncracies and they figure out together. of finding a position that\'s comfortable, of sacrifices and indulgence.   he\'s made the bed. he\'s brought orange soda and inserted the stereo. she lights the candles and strips him down.   \"fuck me,\" she says.   and the old man\'s advice rings in her ears. what if? could it be? what do you think?  and the voice of her mother that nags about accidents and her father that wants to know if he\'s tested and the ex that just shrugs it off. they had a pleasant conversation the afternoon before, gallantly referring himself to the \'best\' ex he claimed as he told her he missed her much and wished to work things out. she could only laugh, smile smugly and brush it off.   time to move on, honey.  and so they fall asleep to odd strange dreams and wake up to tranquility. hyperness. discord. harmony.   she\'s learning as they go...she\'s standing as she moves...she\'s reaching higher.   somewhere out there a soldier fights for his life, and the people of his nation...but here she lays in peace with hushed worries and new found dreams for tomorrow.   she believes.  '),(22603,664485,'05,April,2003','2003-04-05 00:00:00','   odd how everyone either wants you or diswants you at the same time.  '),(22604,664485,'04,April,2003','2003-04-04 00:00:00','   okay..so i\'m trying to not gloat..actually, i take that back: HELL SHIT YEAH i\'m going to gloat!  :)  baby made a hundred and ten for the math\'s test!  boo wop seven hundred and fifty two hours at barnes and nobles poring over algebratic formulas that make bloody no sense at all actually pays off once in a while.    \"you\'re really killing it, girl,\"   and so the bitches in my class that stare me down with their textbooks and the condescending fuckers with their apparatus can go kiss my yellow (slightly tanned) butt because who\'s the asian chick, the only one to make a hundred without extra credit on the exam?!  huh?  huh?  *insert evil laugh*  :P  BOOYAH!   ---  \"well done,\" he writes on my carefully drafted somewhat intellectual garbage i spew and write for about fifteen pages long on the comparison between greek style performance demands versus shakespeare...i write till my hands are tired, sweat pouring from my forehead because this is important and i\'m nervous as hell at trying to remember all the specific points, honing on non-existing photographic memory to try and remember all them spider crawls and colorful charts i\'ve drawn up the week before and analyze many, many nights and afternoon before.   i look around and there are only three people left from the twenty \"elites\" and i think, oh shit. oh shit. oh shit.   twenty minutes left on the clock; adrenaline pumps in and i go into crazy mode where it is my mind versus my heart that says, you\'re sucking ass in this shit. you know nothing of what you write about. you are horrible, you are stupid, you are useless.   his jolly exterior and eyes that glint pass out the papers. comes to me and says, \"so you think the dimple can save you from anything huh?\"  i smile. unsure.   (is that good or bad?)  the papers are folded, the grade is on the back sheet. my fingers shake and i think about how i\'m going to go home a hopeless reject and how i\'m god\'s worst creation on earth and step out of class to finally read my dictation.   97 and a well done.   say what?   97 and a well done.    ---  she aces both papers.  she gleams.  she did good.  \"the best in the class\"  she laughs.  ---  thank god for days like this.  a failing dying child breathes life again. a turning point in life, a decision, things look up again.   she\'s found her dreams again...she believes.  '),(22605,664485,'03,April,2003','2003-04-03 00:00:00','   urlLink  Cigarette Lighter Love Song  I know you\'re not asleep I can feel you moving over there You\'ve been playing with the seam In your worn out underwear My lips are raw as hell From biting on them just to stay awake It\'s not like I\'m gonna need them, you won\'t be around To see them bleed and break  Chorus: All that I do, comes back to you So I\'ll just think about you \'til there\'s nothing in my head All I can do, is try not to screw this up again And just be friends, I\'d rather be dead  I drove out of east Atlanta With a headache the size of my car I called to say I was okay \'Cause I know how you are I\'m like a movie without an ending You know I\'ve got nowhere to go And it makes me wanna throw up To see you wanna give up More than you\'ll ever know  (Chorus)  Bridge: Everything\'s supposed to have a happy ending But the record keeps skipping and the needle keeps bending Like the road I\'m driving to the bridge that has no end I wanna take back everything that I\'ve broken But the bridges behind me are burning and smokin\' I guess this is the end    they lie beneath blankets and sand in the crevices of their arms. bikini strings pulled and a back rub ensues, kisses trail her slightly warm, slightly shivering body as the waves crash just a few feet away from them.  \"look at the ocean!\"   \"look at the seagulls\"   \"look at the sun, and the gentle way it hides behind the clouds..flirting to come out and play...shying away\"  \"look at the way the different hues paint the skies, a little bit of blue here, some orange there, off-white in a distance and kaledoscopic in the middle.\"  of cream cheese with chives on toasted onion bagels..orange soda and root beer from bottles and glasses. sitting on the kitchen counter top ticking off days and marking off events, riding in the car ruminating thoughts and problems and solutions. listening, hearing, talking through touches without sounds...comforting.  barker\'s beautiful mind goes into overdrive, i\'ve never told you this but you amaze me.  the careful way you schedule things without being too assholic. it\'s nice to have stability, it\'s nice to have someone know what\'s going on. you don\'t overdo it in a charts and nonsensical garbage way...but gently. thoughtfully. intellectually.  the everpresent easily attainable smile i know is always there. where when everything crashes and fall you manage to detach shit and make things simple...like they should be.  \"being happy and being a good person\" your motto.  you question about people\'s goals and dreams in life. you wonder and you ask...trying to save the world. you\'ve saved someone silly, you\'ve saved me.  and to put it in \"yorkshire\'s terms\"...let go of the twig and run with the currents.   we\'ll see how this goes...maybe it will, maybe it won\'t.  but hey, we tried.   and hey, sometimes it\'s nicer to just go with the flow..  \"keep things simple\"  :)  '),(22606,664485,'02,April,2003','2003-04-02 00:00:00','   urlLink m o t o r c y c l e   d r i v e b y    summer time and the wind is blowing outside in lower chelsea and i don\'t know what i\'m doing in this city the sun is always in my eyes it crashes through the windows and i\'m sleeping on the couch when i came to visit you that\'s when i knew that i could never have you i knew that before you did still i\'m the one that\'s stupid and there\'s this burning like there\'s always been i\'ve never been so alone and i\'ve never been so alive  visions of you on a motorcycle driveby cigarette ash flies in your eyes and you don\'t mind and you smile and say the world it doesn\'t fit with you i don\'t believe you, you\'re so serene careening through the universe, your axis on a tilt, guiltless and free i hope you take a piece of me with you and there\'s things i\'d like to do that you don\'t believe in  i would like to build something, but you never see it happening and there\'s this burning, like there\'s always been i\'ve never been so alone and i\'ve . . . i\'ve never been so alive  and there\'s this burning, there is this burning  where\'s the soul i want to know new york city\'s evil, the surface is everything but i could never do that someone would see through that and this is the last time we\'ll be friends again i\'ll get over you you\'ll wonder who i am and there\'s this burning, just like there\'s always been i\'ve never been so alone and i have never been so alive  i go home to the coast it starts to rain i paddle out onto the water alone taste the salt and taste the pain i\'m not thinking of you again summer dies and swells rise the sun goes down in my eyes see this rolling wave darkly coming to take me home  and i\'ve never been so alone and i\'ve never been so alive   ---  \"  Act V. THE TRUTH. (Love is hard work. And, sometimes, hard work can really hurt.) Love is a game. If they didn\'t tell you before, we will tell you now. Love is a game and if you play you either win, lose, or get ejected before the game is over. There are no ties. Maybe you\'ll lose and learn some great, meaningful answer from it all (like if it looks to good to be true, it is). It\'s easy to love something when you don\'t have to work at it. It\'s harder when it asks something of you. You just might be afraid to give. Give it anyway. The heart is the most resilient muscle. It is also the stupidest. So if this love you\'ve found is good to you, hold it, keep it, shout about it. If it isn\'t, then maybe you should just become very good friends.  \"    i think about how some day the memory of him won\'t hurt quite as much as it does now...that some day, we\'ll get together and have conversations like we used to. only that we are mean now...we test friendship once a while, give it a go, before everything comes crashing down again and we remember all the reasons why we grew apart. why things ended the way they did.   reading what you said brought it all back, i suck in the tears that stream on my cheeks against my will, but against the oddly coincidental song i\'ve cried to a lot of nights that has mysteriously come on air...i am a hopeless case. i violently swipe away, as if i could just push it all away, as if i could just pretend that all was well. i always do anyway. i\'ve learnt now it doesn\'t pay to show people who you are inside, it\'s scary. it\'s easier to just pretend, be numb, it always is.  i am lost for a while. disorientated and confused, i forget about the scrabble game that i am kicking ass online, i forget about the 1700 word research paper i have to turn in tomorrow nor about the seven hundred and forty two worries i\'ve been trying to figure out. no, right now, all my thoughts are consumed in what you are saying. and maybe perhaps how my heart is screaming and bleeding...the masochist in me enjoying this torture.   i remember how he still comes in and reads my soul, my thoughts, the things i have to say for the day...and he comes too close. or how he uses it against me later, and it hurts. i hurt.   i am seeing someone new but he\'s scarred me for life. i take equal blame, i was a bitch. a slut. a whore. a i-don\'t-know-what-else, whatever name of the day he\'s decided to bestow on me i suppose. i used to think i knew what love was too...but if love is what i think it is, then i don\'t want it. and so i get afraid. head for the nearest exit while another person i\'ve pushed away comes running after me, corners me and i yell. i scream. i fight it to death and it\'s too late. i\'ve drawn them to fear, to understand, i\'ve drove them away. i always drive them away...maybe i\'ve learnt the best way how to.   this new guy fits almost everything i\'ve wanted in a man. artistic, sensitive, sensible, kind, gentle, understanding, loving, patient. director, producer, actor, talent, friend, companion, lover...he is so many things i could wish for yet nothing at all. i am afraid, and this fear drives me insane. i am questioning if i am enough, because the other person has taught me how to. so nights like tonight, when i weep for all that i\'ve lost, all that i\'ve gained, and all that i\'m giving away...i remember again that hearts do break. and i\'m not alone.  so thank you.  '),(22607,664485,'02,April,2003','2003-04-02 00:00:00','   urlLink  It is with deep sadness that I announce the demises of The Muse, Puck, one-who-I-think-is-already-named-Curly, and some punk friend of theirs who, like the three of them has this uncontrolled, all-encompassing death wish.  April Fools Day?  APRIL FOOLS DAY?  A P R I L .. F R E A K I N \' .. F O O L S .. D A Y .. ????????    lol.  oh come on old man...  we love you too!  :P  p/s gotcha!  heh.  ---  missed you at the head doctor\'s today.  :)  ...she does good.     and it isn\'t just the medicine.    two points?  though she talks again about having more pills to play with. the girl stumbles around admist theories and \'working on things\' ; for once a week, she sits on this offbeat randomly colored room to get her head fixed. what her mind cannot do, her heart sets out to mend. they feed her a teaspoon of confidence and happiness each week, not too much, not too little either. they guide her through mazes and lets her pick out crayons to color her world, steps to help her dance and she is grateful. one more session while her mom is around, they bring in reinforces...they teach her how to open. to share. to let go.   they teach her it\'s okay....she\'s okay.  they teach her the things she\'s afraid to know.  she stays.  ---  barker\'s taking her out tomorrow.   reverting back to what he originally set to do on their first date.  she smiles, he\'s so damn adorable sometimes.    silly boy.  '),(22608,664485,'01,April,2003','2003-04-01 00:00:00','   stereo blares: this very place again..machines whirring, television humming, me on his lap. we alternate, ignore the gnawing voices...tonight we make peace.   cast meeting goes better than expected, whispering, running through lines with his lips just a hair away from my ears. giggling, making faces, sharing a cigarette, changing voices, accepting directions.   welcoming again...in the kitchen. peeling hands away from each other, wanting to touch, to kiss, to feel, but afraid. but hesitant. drawing closer, coming closer...too dangerous to be left alone. why the hell are we alone?  words exchanged, perhaps a bit like slam poetry only rickety, nervous. unsure of boundaries or rules, afraid of wrong choices of words...dead roses in the corner of the eye. sentimental reasons? she smiles, but cries. they look so dead / he kept it.   wow.  he shows her the long forgotten song..and the other about going down to the river. ani difranco settling in the room, hash in the corner...a blind eye turned.   he stands his ground.   she walks again.  ...he holds her hand.  '),(22609,664485,'27,May,2003','2003-05-27 00:00:00','   it\'s interesting how it\'s three in the morning and i\'ve stuck chinese food in the microwave with full intention of throwing it back out later.   it\'s even more amusing to sit with your friends at the wiped off counter of waffle house being served by a woman that\'s become your insomniac little group\'s surrogate grandmother then looking down to a plate of breakfast foods and making out markers in your mind of the reverse order in which this will come out. it\'s like two people in your head debating how much slack the other would give.   so what\'s it guys?  full blown out, puke till i bleed tonight? or it\'s okay, we\'ll let you off easy, just reach the eggs racheal night?  the sad thing would be, i\'m willingly writing this post without even being slightly ever intoxicated and i can lick the dried tear stains off my cheek and feel my nose bleed beginning to clot up somewhere within my nostrils. not a pretty picture innit? ugly. say it bitch. you\'re ugly. fat. retarded. sick.  bring it on you fuckers.   it\'s not like you have to be me anyway.  '),(22610,664485,'26,May,2003','2003-05-26 00:00:00','   it\'s a quarter after seven and lost rays from the sun going down pierces through my windows and i feel so lost in my own little world. i have tears running down my cheeks, smeared mascara lines and lots and lots of pictures smiling back at me from me desktop. i\'m used to freezing my butt off in here but tonight there is nothing, i am neither cold nor hot, neither comfortable nor terribly uncomfortable. the hours tick off slowly but surely...so little time left, so many things to do. some hours, the excitement mounts, some minutes, they just go all the way down completely.   i am feeling melancholic and very unbalanced right now, as if i\'m the peak of my cycle and i charting out the course, watching as i fall down fast and hard too, waiting for the crash. tylenol has been my god lately, i think it\'s a stage people go through in life where you look to chemicals to attempt to numb the pain. it\'s as if my mind has it set that i need pain-killers, for six million pains and aches my body comes out with but yet an unfilled prescription of prozac sits admidst my box of whacked out thrash, memorabilia of moments.   it\'s funny how people think a pill is going to make them normal.   it\'s funny how some people want to believe it is.   yet it\'s even funnier how some people are just stupid and don\'t know what they want nor what they believe in.  there are days where i am so high, so high....flying and soaring, tearing up and ripping through everything i see, everything i touch, everything i feel. nights like last night where it was so amazing, and i was so happy. intoxicated maybe, but i could feel and i smiled. laughed. where we walked through streets and felt the sand beneath our feet, where we played and dared and went from balcony to balcony, tit to tit, addiction to addiction. where we rolled down the top and belted on top of our lungs and honked at strangers and flirted with disaster. where we made friends along the strip and watched the little old guy kissed the forehead of the little girl and the couple with blooming roses and the street partiers with the jacked up, hyped up, boomers and bass.   where we got to know so and so and this and that happened, where we switched waffle houses and had a massive array of variety and to where we rode home together and toned down with classical and opera and discussed life long dreams and hopes and needs.   in two days, i leave and don\'t return for another good three months.   it\'s an odd feeling.   my mom bitches on the phone, i\'ve pissed her off somehow because my manic mind has taken over and i cannot seem to concentrate and she\'s getting so frustrated. i\'m frustrated too, but i don\'t know how to tell anyone, i don\'t know how to deal with it. it\'s as if i\'m spiralling down and i honest to god, do not know what in the hell to do. to hold on and prolong the madness or to crash so things start looking up again?  good question really.  and here i sit in the dark cos darkness is my friend and that pisses the shit fuck out of me. i\'ve had a good streak for way too long and i\'ve gotten used to it that this almost feels new. i don\'t know what to do with myself. i\'ve been throwing up again, and hurting myself not too good, i\'m startign to feel the way i\'ve felt before...when things were on their way to bad you know? don\'t get me wrong, i\'ve still been doing the shit i do it\'s just that this time, when i do it, it\'s not out because of a daily routine but because i cannot stand to look at the girl in the mirror again. i look like a massive orgy of ugly fat people that decided to hook up with psychosomatic retarded bitches. i hate me right now, which isn\'t right, because a while ago, i was really really fine. i was happy. i loved me. but most of all, i was proud of me.  i look at that piece of paper that justifies everything and the first and only thing that comes to mind is bullfuckingshit racheal, this isn\'t you.   and then i think to myself honestly, i don\'t know who i am. i don\'t know what i\'m worth nor who i\'m really all about. quite frankly, i don\'t really give two hoots about anything. yet i do all at the same time.   we have philosophical talks and talks about religion and education and third world countries. we make jokes and have a good laugh. we play tricks and decide to walk down the boulevard butt naked then take the boats out and skinnydip. and i still don\'t fit in. or at least my mind doesn\'t because i don\'t feel like i\'m completely all there anymore. like i\'m losing it and i\'m never completely one hndred percent anything.   i don\'t know. i\'m talking way too much for someone that has to be ready in an hour, and hell knows, i need more than an hour.  '),(22611,664485,'24,May,2003','2003-05-24 00:00:00','   sweat glistens off sparkling skin, lavender slathered pored out as running shoes squish on a moist pavement and the alarm beeps as the grandfather clock that greets her sounds twelve rings. midnight; the ending of an exhausting, productive, stressed, retarded, interesting day...the sounding of a brand new morning. a two hour jog across the bridge, around the bay, into wally\'s world and in pitch black darkness...there\'s something awfully enticing about the dark innit?   cramps, shortness of breath, searing pains, drugged out body but the brain jets through at six million miles and hour and her thought process works on overtime, overload. little categories formed in little bits of her mind, trying to get it all together...trying to put things and feelings together to understand. understand...that is a nice word isn\'t it?  help me, my name is stressed. i am the demon that creeps into solitary emptiness and feeds on it, i am the fear that terrifies and keeps you up and about at night, i am the reason you bounce of walls and run off into tangents, i am the reason you are imbalanced, insecure, frustrated.   help me, my name is insomnia...and i am the one that generates energy even after half a bottle of pills, i am the image that plays in your mind of your fucked up self, i am the saving grace of your other being and the wonder of your amazement. i am the reason you are alive.   help me, my name is bullimia. i am the ugly image of your otherwise perfect bullshit life. i am the one that drives you to take a million baths, the one that scoops you up caffienated ice cream after a good throw up to soothe your bleeding stomach, your bleeding esophagus. i am the reason why the back of your knuckles are calloussed from the backs of your teeth biting in as you force your fingers down. i am the bitch that feeds you jacked up pictures of magnified poportions of your every flaw. i am that face you stare back in the mirror that says, fat bitch. ugly bitch. retarded bitch. i am the being you come to for comfort, i am the reason you seek comfort.   help me, my name is suicide. i am the cheap stupid journey you attempt to take every couple of months, sometimes weeks...depending on how fast i work with the other halfs. i am the half version of what your selfless ass attempts to do but your cowardly mind stops you. i am the person that works hand in hand with depression to bring you down. down. down. i\'ll bring you down one day, i\'ll shatter your defenses, i\'ll rule your world one day.   help me, my name is depression and i love what an achoring place i hold in your mind. i love how i can creep into your innermost thoughts and take the simple, most silliest darnest thing in the world and use it to my advantage. i love how i can torment you night or day, dusk or dawn, january through december. i am the protecting arms that hold you in the still of midnight and encase you into all that i can drive you into. i am the tears you cry, the scars you bleed, the questions you seek. i have every capability of turning all rights into wrong, and then somemore. i amthe cause of instability in your life, the reason you run, the reason you fly, the reason you love, the reason your past comes haunting to bite you in the ass but most of all, i am the reason you are the way you are.   help us, we are the voices in her head. we dig our claws in deep and penetrate her every being, her every inch that covers her body and her every pound she counts meticulously. we are vicious, and we are stern. we yell, we scream, we throw horrible temper tantrums and rule her life. we take her through these amazing highs...in control all the time, only to reveal ourselves and the bigger plan. we drag her by her hairs and enjoy the scraping of skin, the bleeding, every tear drop that falls. we lick her every whimper, cry, and thrive on it. she feeds us with her addictions, she feeds us with her commitment. the more she shakes us off, the harder we come back the next time. her constant repetition of taking ten steps forward and falling back a million provides us more than we\'ll ever need but yet, like everything else in this world, greed consumes us and we are selfish. always wanting more. always testing to see how much further we can go. always enjoying another good meal.   help me, my name is racheal and i cannot stop myself from hurting tonight. i am numb from everything but still hurting inside. my throat is parched, my lungs are stretched, my heart is tired and my whole fucking body feels like a dead log. my thoughts are swirling, my bags are packed, my fears are escalating by the every minute.   i am leaving. i am afraid.   excited.   i hate this bi existence. this living of two worlds.   where the, what the, who the fuck am i?  do you know me?  do you need me?  am i part of your world?  am i a point to this existence?  am i of any use?  are you as sick and tired of me as i am?  whining, crying, yelling, screaming, sarcastic, dumb, stupid, long assed, shit piece of crap.   i am not strong enough to slice my wrists yet even weaker to live in this world.   hormones and chemical imbalances don\'t go well at all but i\'m alright. i\'m always alright. just give it a couple of days, a couple of hours more of being awake, another six messages and i\'ll be fine. i\'m always fine. i am fine.  '),(22612,664485,'23,May,2003','2003-05-23 00:00:00','   chrisr  (2:08 AM) :   i don\'t think anything of great worth was ever done by a well balanced person.  balance is overrated.  hence the mad scientist stereotype.  einstein wasn\'t the epitomy of well balancedness.  nor was bach.  to be very good at something, you have to neglect other things, and hence create unbalance.  i consider it a definite plus.  also; i\'ve reached a stage where i quite genuinely don\'t care what other people think of me.  unless they know me well enough to respect how i am.  so, no, it doesn\'t bother me.  i\'m very happy to be unbalanced, though one has to be aware of the downsides to that, and everyone else can cheerily go to hell :)  in a nice way, of course.  '),(22613,664485,'22,May,2003','2003-05-22 00:00:00','   i. the point of having british spelling and american spelling is really just a conspiracy to screw people up.   ii. the cat and dog is snuggling up on the bed behind me.....what\'s wrong with that picture?!  iii. i make fucking good chinese food.  iv. having a pseudo teenage little brother that worships and thinks you\'re the coolest thing on earth around is kinda cool.   v. i like being a chocolate hoarder, literati ass kicker, sleeping couch potato and an absolute waste of space.  vi. racheal - caffiene = bad idea.   vii. my tits get hard when it\'s cold, and it\'s fucking ass freezing in here.   viii. i am not pregnant.   ix. i leave in 144 hours....i\'ll do the minutes tomorrow. :)  x. disneyfuckingland, ready or not, here i come!....again!    lol, so i have an almost confirmed flight to orlando when i come home along with some ass hard shopping money i intend to make over the summer...hell yeah! i also have a confirmed job for when i go home with an advertising firm, oddly enough, this lady down in washington fixed for me. to add to the weirdness, i had a long, overdrawn, complex dream about malaysia, the actor\'s studio, positive tones and my maids going amok in the middle of winter with my non-existent little sister?   everybody go HMMM.   :P  off to go stare at the ceiling and chase more butterflies...  *kiss* *kiss*  '),(22614,664485,'21,May,2003','2003-05-21 00:00:00','   hey babe,   yes, you have every right to kill me, castrate me, and feed me to the starving kids in somalia.   lol, so i absolutely bawled my eyes out listening to your messages last night then realizing quite pathetically that i\'d missed you leaving and that you\'d tried so many times but her royal stupidness had switched off the phone. for some odd reason, i was bloody afraid, wondering...be safe babe, i\'d like to see you again.   but yeah, so here i am, publically admitting that i am a very stupid cared for girl that needs to put her brains together fix things.   thank you for everything you, and for the emails.   take care...i\'ll see you around.   lets shoot a movie!  :)  '),(22615,664485,'20,May,2003','2003-05-20 00:00:00','   don\'t worry.  '),(22616,664485,'20,May,2003','2003-05-20 00:00:00','   i\'m lonely, it\'s hard to explain something like this. it\'s like being lost in a crowd, being surrounded by so many people, being with so many people yet still being the only one. the faceless one. fading away, sinking in, losing grip on something real, something there, something to know.   i keep on needing to remind myself this is the beginning of an ending and things are going great! going fantabulous! going absolutely amazing! i keep telling myself that if anything, i have to be alright, be strong for him, for the others. stay alright for a million people yet it\'s hard when things just don\'t feel right and so many emotions run through you.   tears run, anger lashed, laughter rings.  afraid of changing, craving for difference, for breath to feel so fresh, so alive. for every draw of air that enters through your lungs to give life, to feel like it\'s giving life and that you are not just dying slowly...bit by bit. pieces. breaking away.   it feels like such a horrible existence, to be flying and crashing at the same time. to try and hide scabs and scars as you primp and perform. to dance among the pixies and feel like a fairy...your feet so light and wings like a butterfly fluttering yet to find the flipside of it where your nails are broken, your insides bleeding, your sanity questioned and you, as a being, so unloved. unwelcomed. unwanted. by your own self. you can\'t look into the mirror, it hurts more to know you do sometimes, it hurts more to be up and down and left and right. to hold a knife in your hands, to shake a bottle of pills, to find the easy way out because you\'re so goddamn sick of trying. fighting. to be so goddamned sick.   the therapist says you need to go to tc for a while...what do you do? you walk out and never return. mental health smanzy, bullshit melodramatic excuses anyway. so you\'ve been off for a month and a half now, it\'s like the meds you keep stored away...what the fuck ever you say. you throw it, you crush it, you deny it. it pisses you off that you are not enough, that you aren\'t strong enough. it hurts you that you aren\'t complete, that you have to depend on such things to stay alive...and so you rebel, as always, you rebel against it. fight it. strive for independence. i\'m okay. i\'m okay. i\'m okay.  don\'t need anything.  don\'t need anyone.  don\'t need a damn thing.  fucking needy bitch.   it\'s late and the world starts to tune down as the little town begins to sleep, nothing is left but bare empty silence of nothingness. of emptiness. there is a gaping hole from where the tears begin, to where i am calm. tugging, being torn apart and pulled from every angle. the conscious mind versus the unconsious, the rational versus the crazy, the manic versus the depressive. what the fuck is wrong with me?  little things. small things. miniscule retarded pathetic things that throw everything off balance. hormones. chemicals. emotions. people. thoughts. feelings. understanding.   packing, returning, knowing. i\'m going home, i cannot wait...i cannot wait to come home. fall into familiar arms, show myself, find myself, try myself. i cannot wait to climb higher mountains, soar different skies but yet i\'m so afraid to. some days i am so bloody hessistant, some days i don\'t know what things are going to be like, some days i am absolutely embarassed because i am so different now and it scares me to be alone again. it scares me to step into this world i\'ve lived in for long yet gone away from too....but i want to. i need to. i must.  it is something i know, at times even, it is the only thing i know.   my life is so dang perfect right now it\'s frustrating. i list down all that i have and the sheets of paper i have are not enough, yet why do i get so damn upset on nights like this? why do i get so hard on myself? why do i still feel so incomplete?  is it love i\'m searching for? is it education? is it a career, a degree, a perfect bloody thing to hang on? what exactly do i need? what exactly am i looking for?   i don\'t know, i don\'t know and that frustrates me.  i lay with my head on double pillows and tonight there is no one to comfort me. i wait for his arms to find my waist and squeeze me, i wait for his kisses to trail my body as i am sleeping, i wait for his voice to soothe me to sleep as he tells me of his dreams, and sings, and talks about nothing and everything. i twist and turn finding for his chest to let his heartbeat calm me to sleep..for his smile to greet me as i awake and for his body to keep me warm as night falls, for protection as the night demons haunt.   but i\'ve let him go. i\'ve sent him off. we said goodbye today...our final one as he leaves on a long journey to jersey and me back to malaysia for the summer. funny how we tried to categorize our relationship last night and drew nothing, we laughed...we smiled...we played. not exactly fuck buddies yet not forever either. almost like an in between, a teacher, a guider...him for me, and me to him.   he calls me gorgeous, and sexy, and beautiful. he sits and tells me about the things i mean to him, the things i mean to the world and the things i have to do...my part of the deal in life. he holds my hand and he holds it tight...we do lovey-dovey corny make me sick things together, for each other. we are emotionally attached yet still very much single. i call him the person that fills my void, the friend that steps in. the fucking lover.   i tell him about the people i am meeting and he laughs about the silly things i talk about, the names i give them. he gets a little jealous but then he knows he has no right to be. as i have none. we step away from our existence and examine comfort. he tells me he\'ll miss me, i say me too.   it\'s real, it\'s genuine, only that we know we have things we need to do.   i think about the people at home. i think about the kids, my family, the exes, me friends. i think about acceptance and continuance and redundance. i think about change and difference and ignorance. what now? how? i know it\'s going to be different but how different is different going to be?  i think about how i leave in less then a week. i am excited and sad, i am afraid and happy, i am unnerved and fucking flying. califuckingfornia in a week. shit.  it\'s taken me two hours to write this, in between instant messages and literati games...in between trying to calm the fuck down..running from bored to pissed to sad to crying to calm. and that\'s where i stand now, calm.  i threw up tonight. bled.   i guess this shit never really ends does it?  i\'m okay though, i\'ll be alright  ...i\'m just bloody fantastically fine.  '),(22617,664485,'20,May,2003','2003-05-20 00:00:00','   it\'s funny how she chooses to ignite fire at all the wrong places...yet sometimes, sometimes the amber burns just right and she dances in the orange glow of a passionately beautiful fire.   take care of yourself in jersey silly, i\'ll see you in a couple of months.   changed.  different.  grown.  stronger.  :)  '),(22618,664485,'19,May,2003','2003-05-19 00:00:00','   my moon is your moon.   that just astounds me...  because i thought i saw the most incredibly beautiful full moon the other night and here you are, millions and trillions of miles away telling me the very same thing. it makes me smile that you noticed, that you\'re not too caught up in the drama of life, in the evil corporate influences and that you have time to stop and smell the flowers, watch a sunset, dance in the moonlight. that comforts me.   we\'ve been talking a lot the last couple of days, running up the phone bill, running up the internet bill, running up charges that begins at the top of our heads to the ends of our asses. we\'re both of stature, both from about the same background, hell i\'ve grown up with you and you\'ve played in my neighbourhood as i have yours plenty of times. we\'ve dated on and off and on again. you were my first everything, well almost anyway. the first person i ran from, the first person i fell in love with, the first person i left, the first person that loved me, the first person i hurt, the first person i let go. you were my first of everything, the first time i discovered i was a woman, the first time i had a \'real\' relationship, not the silly kindergarten la di da we play in school, my first valentine i hoped for....and received.   there\'s this fine line of where you stand, there are always tests with me. i offered last night, to bring you back anything you wanted. i\'d asked this before, half my suitcase is filled with things for other people, things they need, things i have to take into account...and you, like any other, i expect to run off into a whole list of things. it\'s funny how people take a lot of things for granted...yet i know i am guilty of that crime too. but anyway, i get so used in having to pay for something, buy something, bring back things, do a big huge favor, purchase expensive gifts and i waited for you to name of six million things. nevermind that i knew you had expensive taste too, in a lot of ways, we are so alike and i can never picture you in a thrift shop getting excited over a shirt that\'s going for a quarter...maybe it\'s ignorance? maybe it\'s just realizing that value is what you give it. i guess i\'ve changed quite a bit and in a lot of ways i\'m afraid of going home, and in other ways, fucking proud to step off the plane.   but anyway, all you said was for me to come home safe. then you told me you missed me.   it was so incredibly corny and genuine at the same time i burst out laughing and crying.   nothing can ever take away that moment because i realized then all the reasons why i love you and will always love you. friendship is a beautiful thing and i\'m glad we have had nothing come in the way of that. it\'s funny how we can not speak for a couple of months and pick things back up where they ended. it\'s even funnier how close we are without trying, how there\'s always that unspoken tension, how the taboos of society are what ultimately separates us, separated us.   life is one funny fruitcake and never in a million years would i have ever expected myself to be sitting here, so many miles away, reminiscing over silly times..okay, i lied. i do that constantly, but anyway, it was all good.   and it\'s a good feeling going home to someone that you know will accept you for whoever, whatever you are.   thank you because i know you read this from time to time.   you are amazing!  :)  '),(22619,664485,'17,May,2003','2003-05-17 00:00:00','   i\'m tired of making up excuses...ever get there you know? where you have to justify everything you say, everything you do. every smile, every tear, every laugh. it feels as if when i\'m happy, there is a voice in the back of my head that waits for me to fall and when i\'m sad, it\'s just the same old thing. there are so many nights where i am alone, alone in this torturous existence, alone in this chemical imbalance, alone in this shit of a disease they call manic depressive. i listen to music, i play music, i sing to bloody music but music only spurs me further...it elevates me, it moves me, it spears me right through the heart. it\'s what i breathe, it\'s my best tool of communication, it is the thing that represents me, speaks for me yet it is the very thing that tears me down.  i tell him i\'m going crazy. that in my little mind and my little thoughts i have come out with six million disorders. barker says every damn bastard on the street has something now, that they\'re feeding his brother a.d.d. medication and that they could prolly label him with six thousand different things because of his childhood, because he is who he is. as he casually tells me in bed one day as lie down that he fits in all the symptoms for depression and he somehow finds that amusing.   i\'m tired of depression. i\'m tired of the damn word, i\'m tired of the disease, i\'m tired of every dang thing it represents. ever feel like something is eating you alive? that\'s what it feels like, whether i\'m super duper happy, on a high, or very low...it\'s always there. it\'s a pain in my big fat butt and i am so god damn bloody sick of it. i wish it\'d go away. i wish it\'d suck my yellow ass and fucking mutilate itself.   i\'m frustrated as a piss fuck right now.   this is apparently, according to the big wigs in white suits, \'part of the disease\'. part of the disease my ass and i\'m supposed to just accept it? just deal with it? well, i can\'t. and it\'s fruystrating. because i\'m pissed for no apparent reason, crying for no apparent reason. you know what the world feels like in here? it feels like it\'s crashing, screaming, scratching nails on a chalk board unbearingly fucked up and i am so goddamn sick of it. i want to throw something, hit something, cut myself, throw up, jump up and down and run into walls.  i write this and there are two sides of me. rational and irrational.  rational says what the fuck are you doing racheal? you sound like a crazy psychopathic bitch.  irrational says i am a crazy psychopathic bitch.  goddamnit, i am so angry and i don\'t even know why. why i am the way i am. why i am never ever happy, or satisfied? why does everything need to be so damn perfect in my life? why i am never ever happy with what i have?   this is me being an immature selfish brat.  this is the ugly side i try to hide but can\'t contain.  this is the awful part of me i hate showing anyone.  this is falling, and crashing, and licking the ground because it\'s sweeter than tasting myself.   i am angry, pissed, livid....the hardest part is not knowing why. the worst part is not knowing where to start.   fuck me.  '),(22620,664485,'17,May,2003','2003-05-17 00:00:00','   i\'m bored as fuck and shit smells.  '),(22621,664485,'17,May,2003','2003-05-17 00:00:00','   it\'s odd how you think you know someone, then you scrape a little harder or flip to another side and you discover this whole new realm you never knew; there\'s a little nervous shiver filled with excitement as you dig a little deeper, a little harder...hoping, searching, wanting to discover more and more and more till everything overflows though you don\'t think you\'ll ever reach that stage, you don\'t think you\'ll ever reach a point where it\'s too much, or it gets too trite; no, life is too damn complexing beautiful for that. life has too many dang intricate details to be overlooked, put away, given up upon.   i stand on my high stool preaching, rambling six million thoughts that cross my mind at this very second purring, slinking away. lights turned down low, so low it\'s dark yet enough to put off a glow that illuminates the room. little candles, big candles, broken candles, burnt out candles...an aura so enticing, so comforting radiated. this is home, this is me, this is life.   it\'s the single breath of life that feeds you everything you need, the voice on the other end, the smile on a person\'s face, the twinkle of an eye. it is the first time you watched a butterfly perch onto a blade of grass, the first time you flew a kite, the first time your bare feet touched soft white sand...the second time you woke up to someone loving you, the second time you saw an absolutely amazing sunset, the fifth time you felt tiny arms around your waist begging to play, the millionth time you thanked some entity above. it is that crack in the middle of the night, the run down empty fluorescent lighted restocking wal*mart newly polished floors at the break of dawn, the ecstacy of driving a machine at any rate you want, challenging the other insomniacs on the road..making friends. it is that last cigarette you share with a friend as you laugh, cry, rant, bitch together. it\'s the shots you down together, the salivating salt that\'s followed by the stinging of spirits and the tearing of lime. or lemon. or whatever you prefer. it\'s the virgin asking if you\'d be the one, it\'s the lusty, excited, scared, tempted look of evil. it is the starting and ending of a promise, the beginning of a journey and an ending of a lifetime. it is being forgiven, forgiving, letting go and being let go of.  it is having the best conversation ever on the phone, one which you will smile about and laugh at for the next couple of days, it is that feeling of knowing you are loved and that you love, it is that reciprocation of care, and trust. it\'s open arms waiting for your return and smiling faces at the end of an exhausting trip. it\'s when your days all string together and everyday feels like a dream, like a new day, like a holiday.   when the dog sits at your feet with the cat by it\'s side and they both want you to play catch with them, it is the first time you discovered love, the second time you fell in love, the third time you gave love. it is the fourth, the fifth, the sixth, the eigth because they are never the same. it is his look, his expression, her eyes, her lips, his carefully chosen words, her legs, your fingers, his arms.   it is sitting down on a saturday night going through films, thinking about what the head doctor said, holding a plane ticket in your hand, marvelling at the retribution of hard work paid off and enjoying the silence.   loving the silence.   dancing in the silence.  :)  hugs.  '),(22622,664485,'16,May,2003','2003-05-16 00:00:00','   chrisr  (5:02 PM) :   i can do that kind of thing, because i don\'t really exist.  i\'m purely a figment of your caffeine tanked imagination.  you\'re actually talking to yourself.    racheal  (5:03 PM) :   nooo  you\'re one of them too  well actually since you\'re on the other side already, could you please tell my coke can to stop dancing around..it\'s distracting my scrabble game....oh, and while you\'re at it, tell the salems to shut the fuck up.    chrisr  (5:04 PM) :   I\'M INSIDE YOUR HEAD.  *tap, tap, tap*  LET ME OUT!  ITS DARK IN HERE!  ---  ....and you thought /i/ was crazy.   :P  '),(22623,664485,'16,May,2003','2003-05-16 00:00:00','   it\'s about a quarter past five into saturday morning and the hormones are raging. it\'s a little uptight in here, scarily predictable even...puck\'s spending the night here and littleboy\'s been up and about the whole dang day! war is going on on the orlando camp and things aren\'t looking too cheery for the mickey-puck saga but we\'ll see...i don\'t know. all these teenage love dating crap. hah.  :P  call me cynical but good luck honey!  barker\'s somewhere off in missouri feasting over \'the best ever\' cashew chicken which just sounds horrid to me. and just like i\'ve said a million times, my ass is missing me some barker. pah. and you dongo, i see you laughing over there, \"funny\". now get up off your ass and ring me now! heh. that\'s right bitch. :P  we\'re putting together a thing or two for the boys....have i ever told you how dang brilliant it is to go shopping at wal*mart at some obscene hour? vanderful.   ooh, i got two very nice phonecalls today. in fact, i think that\'s why i\'m so happy and cheery this morning on nothing amount of sleep and very little nicotine in me (quitting i swear)...both my best friends from home were telling me about shit and all these feel good things along with silly little nit-picky random stupid things it was fun. very fun. perhaps even too much fun.    well anyway, i\'m tired as crap and am falling asleep mid-way. sleep is overrated i tell ya...:) but puck\'s in bed and i guess i should get my ass in there soon before she steals all my pillows and i have to pay for it.  two and six togehter, you figure.  ...just missing madonna but that\'s okay, after orlando, i don\'t think we\'ll be seeing madonna anytime soon. anytime soon.   lol.  talk to you later!  '),(22624,664485,'16,May,2003','2003-05-16 00:00:00','   top five places of the moment...  x. kuala fucking lumpur and absolutely everything and anything it represents. x. that one spot on the beach where we go to  watch a sunset/sunrise/make love  do a lot of things x. disneyland! x. slow dancing in the moonlit rain at the gazebo x. with the kids.   top five bands of the moment...  x. marvelous three x. dashboard confessional x. sigur ros x. lizard\'s convention x. jonatha brooke   last cigarette:  about forty five seconds ago. chain smoking is a good habit.   last car ride:  i just stepped in actually..  destination:  the wonderful world of wally.  ...and no, i don\'t need a reason, though doughnuts, caffiene and nicotine works too. :P   last kiss:   barker pulling away from the theater and running back for one last snog for another couple of days...*blech*   last good cry:   uhm. shit, i suppose i am making progress considering i have to think about it....oh wait, i know. outside the boys apartment on amanda\'s birthday after we got wonderfully knackered, intoxicated in a million ways, senses driven and thoughts messed up big time..talking to barker about going home and missing home and life lessons learned. hell of a good cry i tell you, especially considering someone had to carry my passed out on the curbside ass in.   last library book checked out:  holy shit, i haven\'t been to a library in a long, long, looonng time. the girls and i go into barnes and nobles pretty often and i think the last book i bought was prozac nation.    last movie seen:   in a theater: x2, at home with andrew: my big fat greek wedding. babysitting, watching movies and making popcorn is fun innit?   last book read:   silverfish 1 and 2, a collection of short stories by malaysian writers me mom brought over.   last cuss word uttered:  oh fuck me. (i accidentally let the cat out...which considering i am not very fond of it, is perfectly fine with me)    last beverage drank:  lol, my version of orange soda ; orange juice, ice, and diet sprite.   last food consumed:   ribs, bread, salad, potatoes. chewed, swallowed and digested too! ten points for the girl...:)   last crush:  hm...hehe, a certain bed buddy in orlando. if only because intelligence is a huge turn-on and being nice person with a cute body helps too. :P   last phone call:  amanda freaking out, giving me a hypothetical situation with charles making up disorders. ah...good times i tell ya, good times.   last tv show watched:  hehe..surprise, surprise! paramedics. i\'m pathetically predictable, i know. fart on you.   last time showered:   sometime tonight, i went the whole nine-yards too...candles, flowers, bubbles..a girl needs to be pampered you know. and hell, after you\'ve lived with boys for the whole dang week and a half, some tlc is much needed.    last shoes worn:   uhm. my black sandals?   last cd played:  proposed soundtrack for \'standing at the clock\'.   last item bought:  groceries, ciggarettes and a couple of magazines.   last downloaded:  lol, i caved in a redownloaded trillian into this piece of crap, i\'ve been laptopless for two months now and it\'s a beautiful thing to not be technology reliant, or be addicted to something. i don\'t know, i\'m weird.    last annoyance:  that my room is too dang big and i don\'t have my pictures developed yet so i can\'t finish up the walls.   last disappointment:  this is going to sound odd...but that\'s a hard one. i\'ve been pretty good to myself lately...maybe just that i pissed off the old man.    last soda drank:   god forbid caffiene-free diet coke, or technically, if you count my jacked up orange soda; diet sprite.   last thing written:   eh?   last key used:  car keys?   last word spoken:   shit you. (still talking to the cat)   last sleep:  sleep? now what the hell is that? :P   last im:  oddly enough, jared. though chris is also making sporadic comments and gary can\'t decide if he\'d like to go to bed or not as the annoying closed and open doors keep on sounding.    last sexual fantasy:  heh. i did take a bath tonight, and i do have fingers that like to come out and play, and i did see barker last night, and i do have a very interesting postcard from ron jon\'s...and there is that beauty from the disney parade. hm. you put two and six together.   last weird encounter:  talking to both exes in a night. double whammy! :P   last ice cream eaten:  hehe...and i know this because i live by it. starbucks mud pie ice-cream on an oreo cone. hell yeah!   last time amused:  talking to the exes, one drunk, one tired and both very heart-to-heart shits..and the amusing part would be i have a smile on my face. it\'s kinda nice. :)   last time wanting to die:   the morning after amanda\'s birthday. shit.    last time in love:  still am :)   last time hugged:  lol, anthony came in earlier to say goodnwight, wacheal.    last time scolded:  heh...it\'s been a looong time. thankfully.   last shirt worn:  this maroon shirt i got at the thrift store that says \'up yours\' in yellow print. upped my dyke status any?   last time dancing:  with jeff in orlando at some club    last poster looked at:  that i remember? pirates of the carribean at the rave with barker the other night. because you know he\'s sooo obsessed with those things.   last show attended:  shit. i\'m not really sure and i might be wrong about this one but ramani\'s dance showcase?   last webpage visited:  davinia\'s blog.  urlLink (http://dkj.blogspot.com)   and there you go...don\'t you just feel so much more dang brilliant now that you know more trite silly things about me?  :P  '),(22625,664485,'15,May,2003','2003-05-15 00:00:00','   so i\'ve been stuck in a haze of pissiness and being cranky and in between dusting pixiedust off my chests and chests of drawers, thumb tacking random pictures up on my wall and \"artistically\" arranging dried roses onto different portions of my bedroom wall, i\'ve been bored shitless, restless, and frustrated.   i do however seek retreat once in a while (try every ten minutes) and change my away messages about a billion times a day promising to give head, lick asses, fuck someone\'s brains out, castrate assholes and fly across the moon.   packing is a bitch and i\'ve accumulated quite a bit over the last nine months..apparently, i am a sensitive, pathetic, sentimental fool. it\'s sad.   :P  on a side note however, my room is looking spankingly clean and i am quite fascinated with the redesign. i took art pieces from stuff i worked on in the last few months, dried roses, random scraps of fabric as well as odd ends and pieces of metal and tacked it all up on the wall. it is now a clash of cultures with stars and stripes versus asian print...kind of a familiar theme innit?  :P  kim\'s leaving for melbourne tomorrow and andrew\'s bitching for me to go along, and just for the clueless, melbourne is an hour out of orlando which is a long eight hour drive and just for the uninformed is where i\'ve been for the last week. somehow reliving that nightmare/dream seems kinda terrifying and exciting at the same time. nonetheless though, my ass is staying in pensacola.   i\'ve been sworn to a couple of favors and promises that needs to get done in the next couple of days.   nice thought though.  in another thread, I MISS BARKER!  :(  this sucks.  which reminds me, i have photo evidence of his sexy body being developed. life is sweet...  thank god for heavenly visions.   :P  talk to you asses later...send me some love.   or late night 1800-hot calling cards.   either one really.   ...maybe even batteries.   arf.  '),(22626,664485,'14,May,2003','2003-05-14 00:00:00','   barker leaves for missouri tonight.   it\'s not a good feeling.  i have so much and nothing to say, this has been a rather odd week. we spent much of yesterday and today together, in between sleeping and talking and singing, i didn\'t get much done. soaking up after being away from each other a while, and soaking up for the upcoming long months ahead.   he strokes my hair and i catch him looking at me when i sleep, i find his fingers to twine with mine...his muscles give me the security and strength i need. he smiles, giggles sometimes, and laughs at me, only that he tries to convince me that it is with me. and i roll my eyes and mock him and we play fight in bed.   he gets shy sometimes and you watch as his cheeks and ears turn red, red, redder. and he sorta does this giggle laugh thing and hides beneath the blanket even if it\'s obvious that many would die to be in my place...yet he still calls me beautiful. and sexy. yet he still thinks i\'m special.   i watch him pull that white t-shirt over this morning as his muscles flex and his abs say hel-lo. it is hard not to jump on him sometimes, he calls me his nymphomaniac. it\'s equal game though, equal addiction so hah i say.   it\'s going to be weird to not have someoen to come home to or make dinner for me, even weirder when there are no roses anymore. i\'m going to miss him as he assures me he will too...and i know he will. it is the gentle way he holds my hand to let me know he cares, it is that look in his eyes that look into mine that make me feel so important and special all at once, it is the little things he does in the big scheme of things...it is his mind, his intelligence, his wit that turns me on the most. that after all that\'s said and done, the person, the soul, the being deep within is what i adore the most.   11:23.  he got off work an hour and a half ago. left early...i\'m not even exaggerating when i say that boy works like a damn dog. he has his dreams, his hopes, his visions and so many can concoct them but not many live them. he\'s getting there and every day he works on it a little more, i have nothing but absolute faith in him, he\'s one of those people that do what their heart sets out to do...if anything, those are the reasons why i step back. because my reluctance becomes a damper in his journey.   we are different people...yet sometimes so alike. we are strange and odd and very muchly so queer. we have our own thoughts about the world and the private pains we hide at night. he talked about crying last night, then told me this morning that he did. i\'ve never seen him cry before, and i felt like shit knowing that i didn\'t wake up while his teardrops fell next to me. i was knocked out completely from the road trip and from using every single ounce of energy left, he tells me stories at night....and sings to me. or he downloads some new song he thinks i\'ll like or something new to share. his back hurt like a bitch, and without medication it drove him to tears. i wish i could have soothed him the way he\'s soothed my nights...and i\'m sorry i wasn\'t there. or that i was, but not really, as he apologized this morning in case he woke me up.  he lets me peek through windows into his world, it\'s a different game completely there. he shows me what it\'s like to be him, and live his life, he lets me walk through his footprints and pushes me forward when i am reluctant. he teaches me about simplicity and intelligence, priorities and understanding, he teaches me about barker.   in three months, we will change. i will go home and he\'ll be back in jersey and we\'ll learn more about ourselves again. we touch on how it\'ll be to come together again in the fall to shoot the movie...how strange, how odd, how weird things would be. how much would we have learned? what would we have achieved? how different would we be as individuals? how odd will things be for us? will we fall in love over the summer? will we find our special someones?  he mocks him and that hurts me. i turn away and cry to silence as he whispers in my ears and begs again. he doesn\'t understand, and neither will i ever know about her. we try and convince ourselves of things he tells me, but somehow, i know and believe, someday that would be me. someday i will be strong enough, big enough, worth enough...and you know what, that someday looms near because i am enough. and strong. and hell bloody worth it.   \"walk away.\"  i\'ll make you proud silly, just as i know you already somewhat are.   that made you smile didn\'t it?  take care babe, i\'ll talk to you later. drive safe and i miss you.  '),(22627,664485,'13,May,2003','2003-05-13 00:00:00','   i want to..   dance in the rain.  climb a mountain.  fly every countries\' flag on everest.   sooth a crying child.  save a dolphin.  train a killer whale.  slow dance in the middle of the night with that special someone.  get a big great hug from god.   find the cure for cancer.  take pain away from the world.  find a home for the homeless.  love the world.  make enough money to donate to charities.   shop till i drop.  make love in a great big tub of chocolate ice cream.   graduate on the honor\'s roll.  get my shit together.   talk in a million foreign languages.  take a semester at sea.   be an artistic director for disney.   photograph a story.  raise awareness for aids.   teach my friends about my country.  learn about my country.  be proud of my country, my culture, and all that good stuff.  learn to forgive, forget and move on.  get casted as a voice for a cartoon.   play with pluto.  become a good mother.   learn to love selflessly.  take a long walk down the beach.  live in a beach house.  watch the sun rise and the sun set every evening.   throw away my watch.  make malaysia and america a whole lot nearer.  make my parents proud.  talk to my brother.   get fit.  lose 15 pounds.   change the industry.  help someone like me.   become a great friend.   love myself.   give hope to someone else.     what do you want to do?  '),(22628,664485,'13,May,2003','2003-05-13 00:00:00','   mustang, puck and i watched an incredible display of fireworks the other night at epcot, so spectacular, so amazing, so dang fantastic as towers lit up to cues of music and fountain springs danced, globes turned and projected across a million people as lazers beamed through the sky illuminating lit up mesmerized faces, enchanted smiles and the little kid lost in the magic of disney in all our eyes...it was beautiful.   a moment so genuine, so innocently naive shared but lost at the same time...of adjustments made with the chemicals in my brain and work lights flooding thoughts as they came together into a massive attack of symphony, and sound, and correlation as the night ended and we wrapped up our trip.   one day late, but one day gained. an extra day of basking in the sun, giggling with friends, swimming and dancing, splashing in cool waters soothing the parched sunburned skin as thoughts met eye to eye and conversations took off, jokes played, and an ending to beginning formed.   new goals, renewed dreams, as magic carpet rides on this journey of leaps and loves, of drops and soaring, of ups then downs...so much to do, so much to see, so many emotions...such simple amazement.   things happening, things changing, things shifting from extreme to comfortable to maturity plugged into the life long collection of the lessons in life. love lost and gained, thoughts realized and concocted, the basic virginity of being scared, excited and amazed at penetration for the first time; a beautiful, absolutely gorgeous feeling of coming to terms.   mother\'s day, comfortable in my skin; stripped naked and bare but beautiful...loved. kisses trail, skin on skin..learning about myself, oozing confindence it\'s almost disgusting. driving to cocoa and back with nothing on but smiles and laughs, a new discovery of self...a vacation of mind, a lesson learned and a lesson taught. taking off with wings soaring through the sky as a new chapter begins and i am old and young again.   moralities, integrity, wants hopes and dreams questioned by evil, putrid desires as confidence stands back and says this time i win.   ---  old man and barker greets us, familiar faces at the end of a long, long eight hour journey. at first there were four, and then there were two. two tired, hyper, melancholic girls step off the greyhound bus, buzzing with stories and tales and dreams with picture frame rememberance in the minds of the boys waving goodbye and mustang\'s hands pressed against the glass. taurus and puck beginning on an uncertain journey of discovering friendship, then lust...and maybe love.   bongo and his \'bongoes\' recovering from a week of crazy screaming lunatics. waiting this morning for bongo to bust out some weird one liner and tell us more stories as i awoke and for caress to fix us the breakfast of champions as he fixes himself another samolian dinner.   thanks for the eight days guys, contrary to the other two and the backlashing, bitching, stabbing, screaming, yelling...we had an awesome experience.   thank you mustang for disney, for 12-7 talks, cigarette breaks and being a great bed buddy.    my body aches with muscles stretched and tanned skin makes me kinda giggly. it\'s a quarter after six in the evening and barker and i have slept the whole day away after a night of intense grouping. \"rejuvenating\"...whatever the hell you wnat to call it really. it was damn good fun though.   barker leaves in four days as he sits behind me ringing up the cast and putting things together before missouri and new jersey take place. summer\'s here and i leave again in a couple of days, it\'s a brilliant feeling...a little scared, very excited, and a little sad.   ---  old man, i love you.   ---   urlLink  \"I am not a child now. I can take care of myself. I mustn\'t let them down now- Mustn\'t let them see me cry. I\'m fine. I\'m fine. I\'m to tired to listen. I\'m too old to believe: All these childish stories. There is no such thing as faith, And trust, And pixie dust. I try, But its to hard to believe. I try, But I can\'t see what you see. I try. I try. I try. My whole world is changing, I don\'t know where to turn. I can\'t leave you baby, But I cant stay and watch the sitting burn; Watch it burn. \'Cause I try, But its so hard to believe! I try, But I can\'t see where you see. I try. I try. I try and try, To understand, The distance in between: The love I feel, The things i fear, I dearly say good dream.   I can finally see it. Now I have to believe: All those precious stories. All the world is made of faith, And trust, And pixie dust. So I\'ll try, \'Cause I finally believe! I\'ll try, \'Cause I see where you see! I\'ll try. I\'ll try! I\'ll try! I\'ll try- To fly.\"  '),(22629,664485,'08,May,2003','2003-05-08 00:00:00','   hey strangers,  how\'s it going?  it\'s been a crazy couple of days and i\'m finally plopped down on jeff\'s computer by force to be made to blog...it\'s an odd thing to have not much to say and everything to blog.  :)  i burned my paper journal last weekend, it was strangely very soothing in some perverted sense, like giving away and letting go a cetain part of you...all the nights, all the midnight thoughts and midday drama penned down, all the hushed secrets and forbidden ink on paper, scribblings, writings. poetry, prose, of soaring and falling and climbing back up again, it was building a dream and escaping midflight in ecstacy to a place unknown, it was falling down, bruised and broken, hurting from a million places and tear stain pages greet you, smudges from where teardrops touched paper and candle wax dripped on. it was liberating and depressing, as if something i had to do to break chains but hurt me at the same time?  in any case though, i started one in between the hours of five and six am monday morning and haven\'t written in it since. there\'s somehting soothing about being able to keep things in and not feel the pressing need to explode, it\'s empowering and it feeds you a certain strength of believing in yourself and confidence. a mini confidence pill i suppose?  orlando is unbelievable and nothing like i expected. the boys we\'re living with are the cleanest, nicest, sillliest, craziest alcoholics i\'ve ever met. and this coming from someone who spent eight hours on the road with equally insane, hyper, dramatic, needles up their buts and strings all gone loopy people in the car. suffice to say, lots of mooning, flashing, nudity, honking, craziness went on on the way here and something tells me, it\'s going to be three times more when we drive back on sunday...  we did a little bit of modelling wednesday night and that was interesting for the most part and nobody ever says no to nice clothes and money eh? i love, miss, adore, revel in the big city yet we\'ve gotten lost a good number of times but i suppose it\'s all good. adventurous pompous assholes. :P shopping here is about as close to getting whipped cream licked off that i can get, high-rise buildings, highways, speed demons, traffic jams, cars, noise, pollution...heaven! whoop! whoop!  i did get pissingly knackered last night though and the bits and pieces along with the headache and rumble in my tummy tells me it was probably not a pretty sight. i remember talking to barker, body shots, ass grabbing and people dissapearing into rooms plus puck\'s and madonna\'s massive orgy in the bathroom but i look at invisible tearstains and remember little of what i must have been bitching about. little bits of conversation float into my mind progressively throughout the day and i think, damn i am a sucker.   lol, she\'s a college student!   what a bloody trite excuse, in any case though, apparently we\'re going skinny dipping and sunbathing now. will blog again maybe later, but yeah. life is a goddess on crack.  '),(22630,664485,'04,May,2003','2003-05-04 00:00:00','   so i leave for orlando tomorrow and should be gone for about a ten days or so..i\'m not really sure what to expect out of this nor how \'together\' my thoughts are right now. it\'s hard to believe that after tomorrow, i am done with school and would have completed a whole year....it\'s even harder to believe that i survived that whole damn year!  my fears and concern soemtimes mount up in neat little piles and because so much constantly goes on in my head it becomes living hell every couple of weeks and we talk about the dangers of spiralling and crashing and looping, depressive one day and manic the other, independent for a couple of weeks and needy the other. it\'s a delicate balance that at times, i truly wish i did not have to deal with but hey, it\'s okay you know?  barker and i are growing apart. i haven\'t really said it out loud and have been putting it away but we are. we came to a decision a couple of weeks ago about how we both had to grow on our own and that as much as we enjoyed each other\'s company, we weren\'t in love. perhaps that\'s the recipe to it all, figure out where you stand and you will soar. for the most part, we do really good. he knows how i function and i know what ticks him off, keeps him happy, makes him safe. he knows which buttons not to push and when to let go of a subject yet there are times where i miss the magic of a beginning. the discovery, the amazement of sharing something new, the excitement of revealing/finding out something new...perhaps that\'s what explains itself that as much, and as simple as it would be, we aren\'t really meant for each other, neither are we really the one.   in some sense, i have a general rough sketch, and it\'s hard to live with your past. it\'s hard to open three hundred doors before you and come to terms with the person you are but so many times, i look into the mirror and as much as the very same image that stares back at me utterly revolts me and pisses me off, i cannot help but love her, i cannot help but want to help her. help me. accept me.   we all have dreams in life, and when you fulfill one dream and watch it come to life, you build more and more and see them come to life too. my life has been a series of fulfilled dreams and sometimes dashed hopes. i get up and i get down, i watch myself bleed and let myself cry and the other night, i just felt so utterly sorry for myself and so i cried. no big deal at all, something that i would have normally felt so dirty and so wrong for i accepted.   i cried and i sobbed and it hurt but the pain soothes away and tears dry off. i climbed into bed with barker already asleep, looking so peaceful he waited for me to come but i stayed back behind downstairs to finish up a couple of things and so many times i think, it\'s absolutely beautiful how mundane things fit in so nice. sleeping on the couch together, taking a drive, going grocery shopping, walking the dog. picking up chinese food...stupid shit like that i crave for. non-drama stuff, non-special things that become so special in their own right. it\'s silly but i look forward to nights where we just lay in bed and talk, or to days that we don\'t have to do anything extraordinary at all...thrift store shopping, red-light snogging, falling asleep to a warm body. skin on skin.  i have had issues i\'ve needed to deal with and trust me, there\'s been a couple. it hasn\'t been high flying all the way and it certainly hasn\'t been easy. we\'re such different people yet we\'re alike in some odd way which just irritates the piss out of me at times because i can\'t just yell and scream and throw a tantrum, because i know how it\'d feel. now, doesn\'t that suck? racheal having to be all reasonable and shit...fuck that! :P  we both get our nights where the world just seems so dark and wrong, or drives where it\'s hard to just concentrate on going on, and there are so many times i try and push things away and not let them get to me and those times, undoubtedly, are really, really hard but hey, i\'m here ain\'t i? melancholy is truly a fucked up feeling, missing someone, messing up, trying again and all that crap is such damn bloody hard work to sift through but then he holds my hand, and i grab on to his, but then he looks into my eyes, but then he holds me close and he says we always can talk. and i know now the things i wish i knew when i was younger, and i know now that above everything else that we\'re friends and we always will be.   how do i justify us sleeping together? how do i justify this relationship? how do i justify all the issues that surround all the shit that life always has to come with? i don\'t know. i just make do. i just figure it out as we go along...  no, we aren\'t together. yes, i do care for him. no, i don\'t know what\'s going to happen over the summer and yes, we need our time apart.   see, the crazy thing about it is that we both have dreams, and we both have things we need to go off and do for ourselves, both great big unresolved shit we need to find closure to and i know it\'s so hard to be on the outside looking in but clarity is something that just makes life boring and trite. some day, when i grow old, i\'m going to look back at these crazy times and realize i had one hell of an amazing life..because as pathetic as this may seem, i look back at my tumultuous years and think, wow, i\'ve had an amazing life and i am grateful. i am special.   i need to be, as i\'ve needed for a while now, to be my own person. to learn about me, to live in my head a while and not feel the pressing need to kill myself...and so far, having barker around has helped. we get what we need from each other, some call it using, some call it dirty, and some call it wrong. i call it being there for each other, i call it friendship. i call it caring. he cares, i care, he needs, i need, he gives, i give, he takes, i take. and so that\'s how that works.   we shoot the movie when i come back in the fall...what if we fall in love over the summer? now how ackward would that be? yes, it would be strange, but then again, everything about us is strange and like everything else, we\'ll deal with it. don\'t mix work with play, don\'t mix directions with friendship, don\'t mix lust and love....and so we won\'t. and so we\'ll make it. i think. we\'ll survive. :P  he leaves for jersey about a week before i do, resumes the bits and pieces and puts it all together again. carries his responsibilities and the things he knows he needs to do for the movie and himself, discover himself, live his dreams, learn more about the world...and like wise i\'ll go home a little unsure, a little scared, very excited and certainly different. nine months away is a bloody long time and there\'s been so many twists and turns to this whole living thing it just gets crazy sometimes but you know what, i kinda enjoy it. as much as i bitch, as much as i moan and swear i hate this living thing, it gets kinda nice sometimes.   the site has picked up a couple more visitors in the last few months, friends i\'d like to say..and so sometimes when i check my stats and get to know these numbers, ip addresses, internet servers, it\'s almost a game we play. silent lurkers, odd random emails, crypted notes sent at the right times, blinded. they come in succession, sometimes all at once, and sometimes none at all.   i know i\'m different than the sarahkate that posted a year and a half ago, and that i\'ve changed from the girl that wrote lost love letters online and the girl that played out drama on the internet. i\'ve evolved from the bulimic, sick, psycho bitch but i am still her. i am still that person and she is still so much a part of me and it\'s brilliant to go in and realize someone else is saying, hey, i kinda like you just for who you are. hey, thanks for trying. hey thanks for helping. and then you stop and say holy shit, and then you stop and say, i\'m just like you, i\'m exactly like you and maybe that\'s why it hurts when i hurt because whatever i can do, you can do to. wherever i reach in life, whatever this point of nirvana is...you\'ll reach it too. i believe in me, i believe in you, i believe in that selfish bastard that lives next door to me and that little old lady that lives down the street, the asshole that steals my food at every break without fail, the idiot that bums cigarettes off me and the anarchistic dumb bitch at the corner of the road.   and so when you tell me how you\'re proud of me i can only tell you how grateful i am and i wish you knew how much that meant to me. when someone writes because they\'ve felt like i could understand the pain, when someone writes to say you\'re beautiful, when someone writes to say i love you, you step back and wonder how the hell you pulled off this shit. how the hell you garnered this and  say hell yeah!   :)  hm...so i had a point to this but then i got side-tracked? i\'m leaving, and it was really just to say, i\'m doing pretty okay and don\'t worry about me. what with living with two very sexy (ow!ow!) boys as house mates, a city filled with a million sinful pleasures and an itinerary that looks like it should be on mtv, along with the crazy, definitely trouble-making, people i\'m going with...this should be interesting.   take care all, i love you, and if i die half-way...remember to kiss my ass when they bury me.   heh.   \"smart ass\".  '),(22631,664485,'03,May,2003','2003-05-03 00:00:00','   heh. so it\'s been an absolutely brilliant day..i\'m not quite sure where to begin though? we watched x2 last night, which reminds me, that chick in daredevil? very, very, yummy.   spent some \'therapy\' time with the old man which is always good. off to go do some papers and study a little..i\'ve had a couple of things to say the last few days, maybe even weeks but haven\'t had the time, maybe even courage to get my shit together to write it down. maybe tomorrow...always a tomorrow eh?  hugs.   have a good one Y\'ALL.  :P  '),(22632,664485,'02,May,2003','2003-05-02 00:00:00','   he played a song the other night that seemed so ironically perfect and melancholically on the dot to how i felt. it was strange..then he read me my posts, he said like how he reads them when he\'s alone, tinkering on the computer..scanning my thoughts. i let teardrops fall and swallowed pride, it seemed so right then, it seemed so altogether..the tone in which he soothed the scars, the force in which he steered my pride, the strength in which his arms held me upright. it was so gentle, so kind, so sweet...it was so real.   i was afraid, then happy, then angry, then just not very sure.   he sings me to sleep sometimes, and my thumb pressed into his, my ears to his heart and i measure our heartbeats together and tune in to the rhythm.  dup-dup.  bop.  dup-dup.  bop.  dup-dup  dup-bop. and i climb into bed and curl into his arms, let my head rest on his shoulders and let our bodies contort to a comfortable position. we make love in the middle of nothing sometimes; his fingers find me and i help him fill that need. the need to be touched, the need to be stroked, the need to be fucked, and then again. the need to escape the demons of the night...they haven\'t visited awhile; will they come knocking soon?  i\'ve never been one to have nightmares or dreams, but in the last few months they\'ve pounded incestantly. i\'m either getting killed, raped, chased or something or the other. he tries to wake me up though, when i\'m squirming around only i don\'t realize it\'s a dream and it\'s so hard to fall back asleep after that...constantly waiting to be attacked, constantly waiting for something to creep beneath the shadows and fuck me over. fear is a very odd feeling. but anyway, i started this off to say i had one entirely different last night, very strangely odd in fact. it was about being caught up at a tcby store (i\'m allergic to tcby ice-cream, and yes, you can laugh) that was shaped after a kfc fast food restraunt yet they were handing out mc donald\'s food. and the manager, this malay guy (malay guy?! ) that sorta looked like someone you\'d pass by in a crowd yet there was something in his eyes that sorta made you stop, and wonder, was telling me about milkshakes and how i shouldn\'t order the chocolate milkshake and definitely not the special because the potatoes were bad.   what the hell?  it\'s vivid, stuck in my mind in a very strange way. so god awfully weird but then again, what\'s new eh?  madonna, puck, sexy pants and i are tkaing a trip down to orlando next week and spending a couple of days there. it\'s of utter importance that we analyze the details now. barker\'s talking about going to missouri, and then i leave for la on the 28. the parents want me to go down to virginia for a visit and hang out with the family, we\'ll see. there\'s too much on my plate to figure out silly things like this now, i\'d much rather not give a care frankly. take it as it comes eh? live for the moment.   the one trip though that i\'m extremely looking forward to is a certain one that flies me into kuala lumpur.  hm.  oh, and just because you counted, three months, four days was it? and still counting.   interesting.   let\'s play with fire.  '),(22633,664485,'01,May,2003','2003-05-01 00:00:00','   urlLink  \"A lot of times I look at this page and wonder musingly if this is the last post I\'m going to make. If things will just sort of fade away and people will forget that I\'ve kept journals. I wonder to myself if the visitors will stop visiting and one day I\'ll come here and see old words and my thoughts will echo in the emtiness that is the web. I\'ll look around and see myself alone, and maybe I\'ll have the courage to really say a couple of things that have been on my mind.  But most of the times I come here hoping to vent. Hoping to let tired thoughts escape my weary mind and instead dribble flows from the tips of my fingers and I spray out so many syllables with not one little ounce of feeling behind it.  A lot of times I think this is simply because I spend so much time these days feeling things. Things I never thought I\'d feel. Things I was both afraid of feeling and didn\'t think there were any nerves left to soak it all up.  I guess honestly my past has never been very guarded from anyone who\'s ever really paid any attention. The fact that it left any scars is the last thing anyone would be surprised to hear. The daily fairy tale is a shock, and yet is a ray of light and hope in a life I\'d of been happy to give up on so many times during so many years.  I can\'t convey how weird it is, to wake up groggy next to a warm loving body next to me. For a heavy arm to fling itself across my chest and flesh press into my flesh and be joyous. The heaven of being able to feel the wonder of skin against skin, for that waft of morning breath to drift across my nostrils and for my body to sigh in releif and drink in the little stink happy that it\'s even there. Okay so even the best parts of life have stench, but to be joyful about it is something I am happy with.  The saddness is still there though. I am surrounded by love, and still the saddness clings to me like dirt that cannot be washed away at times. I remember the hard times, although now they are faded as they never have been. Remembering how hard it was growing up in a home where love was a rumor and happiness was dubious at best and assuredly fleeting.  I find myself wallowing this week. I go through the motions but it is a definite effort and I can\'t quite explain why. It makes me thankful of the arms that hold me though the nights and the dreams I can\'t quite escape. Strong loving arms, that I only hope feel the love I have for them.  In the mornings I face a stranger in the mirror. I\'m not sure who this person is. She goes with me everywhere though, and it\'s odd to know that this is the only person a lot of people here know. My reflection gazes back at me unflinchingly, the face somewhat familiar, hard in places and soft in others. And the eyes stare back at me and occasionally I look at them as if looking at a strangers eyes and wonder what those eyes have seen. But then they peer knowingly at me and I remember. I remember all the things those strange sad eyes have seen.  Mostly though, my eyes stare back wearily. I wonder when I\'ll give up all these ghosts. I wonder when I\'ll get past the scars. I wonder when I\'ll be happy inside my skin. I wonder when I\'ll be content. I wonder if my particular slice of heaven has a time limit. And then I wonder if my heaven might someday blow away like so much dust in the wind.  Then my eyes meet my reflection and I stand taller, straighter, and I look like stone and my thoughts fall out of my head as if rehearsed. If all I have is a moment, I\'m going to try to live in it. I\'m going to try to stay in that moment. I\'m going to suck every ounce of juice out of the moment as if I were dying of thirst in the middle of an arid desert. And if my moment does end...maybe the moment will sustain me though whatever comes next. This is what I tell myself. This is what I want to beleive to be true.  This is what I think before I drift off to sleep.\"    hm..so i suppose these are only a month late but hey!   lol, in any case, drew better start getting his ass to work...if only because as retarded as this might seem, i actually missed that bullshit. heh, so yeah, here they are.:P  they\'re a little silly and i\'ve put on weight (oh my god!) but oh well, shit happens eh?                                                                                                             chicken dissing bitch!  :P  '),(22634,664485,'29,June,2003','2003-06-29 00:00:00','   urlLink       urlLink       urlLink       urlLink       urlLink       urlLink       urlLink       urlLink       urlLink       urlLink       urlLink      i remember the street performers...and the funny way the air smelled that day. it was too bloody cold for the beginning of summer, the heralding of a new season yet many thronged the city. we walked through aisles of concrete manmade bullshit, sales and whatnots...brand names splattered haphazardly on every available surface atop the display windows. people, lots and lots of people and that funky cool bubble machine spurting out liquid once in a while by the pier. oh how i long to join the little kids that ran gleefully catching irridiscently lit bubbles in the air. how i longed to be young, care-free and obnoxious again.    these little amazing craft stores run by illegal looking immigrants snuck into the great big land of america. a new journey, a brand new beginning, a lifeless mermaid, scuplted clay, a charcoal drawn potrait...random lucid thoughts spun around as bikini clad hot chicas strutted their stuff. did i happen to mention that it was freeze ass tit hardening cold?  but the best of all were the people, the people that stopped to watch. the cute waiter that served us our coffee and biscotti. that winked and smirked because he knew he was hot. the little kid in pink that ran through the streets with her newly purchased equally pink animal twisted helium filled balloon...and her parents with little brother in tow chasing after her. the million and two people lining up to use the restrooms that seemed to only appear after every five miles or so, the smart entrepeneurs that charged for entry, the tango-ing bride and groom on the streets. the shocked faces of awe and selfless regard, the cheering, and the stupid selfish bastards that walk away without dropping even a nickel or two into the buker\'s guitar case, the girl\'s violin treasure, the b-boys macy plastic bags.   it was beautiful, to be lost in a crowd snapping pictures, taking in the sights, walking hand in hand.   it was amazing, just being invisible a while. to blend into the crowd....to not stick out.  '),(22635,664485,'25,June,2003','2003-06-25 00:00:00','   [ Sun Jun 29, 06:32:58 AM | + racheal | edit ] chrisraettig: hey babe.  hell yeah.    :)    urlLink    urlLink    urlLink    urlLink    urlLink    urlLink      urlLink    urlLink    urlLink    urlLink    urlLink    urlLink     i. lol, at the airport in florida with amanda about two hours before the flight was schelduled to leave...i love you puck. thank you for the memories, the nights, the fights but most importantly, for little crazy things you do. for being a bitch, and a bloody charming, fantastic, beautiful woman. :P   ii. sexy pants, the stupid muse and puck :P lol, and just in case you were wondering..we bought/traded/fucked all our tops together..god, these inside joke things are hard to explain!   (read: amanda\'s shirt says \'how to steal a boyfriend\' and oh god it\'s pink!!!)   iii. in la somewhere doing something. love the top. hehe, surprise package from me mom!...and the butterfly pants embroided with beads in GREEN!!!! :) yes, i\'m bloody fat, i know. shut up.    iv. captains unite! yee ming, 2002-2003 dynamitez captain + beverley, co-ordinator of all tommy hilfiger and man u gigs a.k.a. big kahuna woman person.    v. my girls!!! being retarded after the prize giving cermony. :) me a very, very, very proud mother hen. first runner\'s up in the nationals...someone say fuck yeah!!  oh, and emilie\'s mom attempting to shove the hampers in?   vi. making fun of the press. making a even bigger fool of myself. lol, me famed box split! wahahha...which btw also equals a LOT of pain without stretching. ( thank you girls.  )   vii. boom dynamitez! hehe, more lets mengada-ngada with the press.    viii. at universal studios being stupid. (so what\'s new?)   ix. notice the similarities?   x. some chinese theater thingamagic.    xi. santa monica pier.   xii. dinosaurs and racheal. you make the connection..:)  '),(22636,664485,'25,June,2003','2003-06-25 00:00:00','   something feels wrong...i don\'t know what it is. i feel so unbalanced at the moment, like i\'m crashing and falling and i\'m afraid to look down. i\'m scared i think, or something. nervous maybe, i dont\' know. i don\'t know what started it, i don\'t know why i suddenly just felt so icky, so hopeless, so bloody disgusted with myself. is this right? i didn\'t think so either, and i am so dman frustrated of dealing with shti like this. i\'m so tired of being tired, so sick of all the bullshit and drama and up and downs, highs and lows. ever feel like this? ever feel like you\'ve fallen off a cliff and the best thing being you don\'t even know why? or who pushed you? or if you slipped or if it\'s just a horrible dream. is this paranoia? is this mania? is this just part of being me?   i\'m crazy, i swear i am. i look at myself in the mirror sometimes and wonder, what the hell did my parents do to deserve this? why the hell am i here? and i think about all the things i wish i had done, i think about all the things i could have done, should have done, must have done. i think about everything and i will myself to cry because i am so numb. i am numb to a point that i am afraid, afraid i\'ve lost all sense of sensibility, afraid i\'ve lost all sense of hope. what is hope? is hope the thing you hang yourself on, the tiny piece of thread you walk on? i am home. what the fuck is home. i don\'t know this place, sure, i know these people and the backroads and a little bit of everything but this isn\'t me anymore. i feel like a stranger to my ownself, i feel so lost and fucked and ugly and bastardy. there\'s just so much pressure, so much damn pressure to be perfect, to be right, to be okay, to be nice, to be of a certain conduct, to be of a certain person i am not. i am not her...i\'ve never been her. it isn\'t me. i am just this sorry excuse i\'ve layered. i am a product of bullshit. leave me alone i sometimes want to yell, but i\'m afriad of being alone. just like i\'m afraid of the dark. but i am so claustrophobic at the same time....too near. too close. too fucking bloody close for comfort.   and so i hide. suicidal, bullimic, deppresive, manic, insane, retarded, fucked. i throw up till i bleed, it\'s funny how that scares and soothes me at the same time. i want to leave but i cannot go, i have too many commitments, too many lives hanging on me, too many people depending on me to be around, too many hearts i cannot let down but what about me? can i please be selfish too? can i please just hide away from this a while, can i not be here for this second, this minute..just a while? it hurts. i hurt. and not at all. i am just flying. and soaring. and falling. and crashing. ....only that things are so perfect; just exactly like they should be. i am everything and i am on the way to bigger better greater amazing things because i am this amazing person. or something that goes like that anyway. it\'s called the extremes. i hate the extremes. i hate it with a passion so great i\'d do anything to get rid of it, sometimes i think even dying. sometimes i think to rest in peace would be nice. sometimes i slap myself silly because i am silly. i have a great laugh, a nice smile too. have you ever seen it? i don\'t know where it goes when i get all ugly like this, i whisper to the silence and beg it not to take me, i will the emptiness into insomnia. cripple myself with exhaustion but the body can only take so much. so many doctors to see sometimes, so many things to accomplish, so many things to do. so much expected...and i must want these things but sometimes i feel like i\'m on life-support. a useless limb kept alive by the people around me. ever feel like that? like i\'m just wavering in between.  i am nineteen this year. almost an adult, and in some countries...already one. i feel like a complete utter bloody failure. sometimes i think, sure yeah, i\'ll make it. i\'ll be all the things i\'m supposed to be, all the stuff i\'m expected to be but some part feels so sure that i\'ll always be caged up, locked into an existence like this. it\'s always going to be drama and something is always going to hit one after the other after the other. it\'s perverse but sometimes i think it\'s good to get cancer. that way i can slip out and no one would be too angry. it\'s stupid. i should be slammed hard on my head for saying that...i don\'t know. maybe i\'m just tired, it\'s 2:42 in the morning after all, i am out of ciggarettes, i am anything but driving on the wrong side of the road tonight and it is fucking freezing in this musuem. i am cold, i am stripped but layered and i am going crazy.  '),(22637,664485,'20,June,2003','2003-06-20 00:00:00','   something\'s wrong..i don\'t know what it is. perhaps it\'s intuition or plain racheal paranoia but something feels off. unbalanced.   it\'s eight minutes after two and my nerves are twitching, my mind going at too many million miles an hour and insomnia hits hard. mania too. i\'ve chain-smoked through a packet, taken a cold shower, freezed my ass off, worked out, thrown up, written, organized, cooked, eaten, cried, talked but nothing feels right. nothing feels okay or anything right now..and i am scared.   i\'m scared because everything\'s fantastic at the moment....then why the panic attack? why this feeling like i\'m hyperventilating and drowning at the same time? why the feeling like i\'m falling and i\'m useles, ugly and disgusting. i feel hopeless at the moment...terribly dissappointed; but i don\'t know what caused it. i don\'t know why i\'m going insane in my mind, i don\'t know why the sirens keep on going off in my head. am i going crazy? or am i already mad?  just a little going on inside;  i am a sorry excuse for a human being. i feel overweight, obese and disgusting at the moment. i don\'t know what it is, but it gets harder and harder to wake up in the mornings...much worst to go out. it\'s hard because it\'s that feelingof failure again..the why the hell am i back here? i cannot seem to find anythign that suits me, soothes me. there\'s something about the pressure, there\'s something abotu having to dress up, to be perfect, to have the clothes, bag, make-up, belt, earrings, shoes and god knows what else fit perfectly into the picture. i\'m in this rut again, this cycle...and it\'s destructive. and i\'m destructing.   i\'m tearing apart everything i\'ve worked on for the last couple of months. i think about what if i was stronger. what if i was thinner. what if i was more complete. what if i was better, smarter, prettier, nicer..whatever. sometimes it feels as if i am never enough, sometimes as if i am never going to be enough. know what i mean?  i am a sorry excuse.   i am a fuck hole living.   i am a pointless piece of retarded shit.   i am a waste of space.   i don\'t know what else i am but at this moment, i am icky. i am sad.   i don\'t know what\'s wrong.  '),(22638,664485,'20,June,2003','2003-06-20 00:00:00','   she wills me to speak, to say something. to utter meaning behind this destructive, once-protective barrier. she pierces blue-green eyes to mine that trickles out painful drops of betrayal....why, how, when, what? she wants to know, she begs to know.  i can only look the other way.  i can only toughen up, shut up, walk away.   write;  six million entries of prose and poetry...seven million anguished thoughts. did i mean it? was i dreaming? is this another one of those \'lessons\' you file in that great big storybook of life and flip back to reminince once in a while, to recreate all the drama to know you still feel, to pierce a little, stab, scratch the surface and sting?  it doesn\'t seem to work that way anymore...it\'s sad, perhaps a little ironic yet comforting at the same time.   a void filled. a place taken. a story ended.   what is this desire? what is this trust? what is this obsession?  with love? with trust? with commitment?  taking in hard fought battles, old, trite, boring...so fucking repetitive. so damn redundant, i am amused. laughs...cackles a while. a bloody big joke, a huge facade of shadows, a constant change of securities.   they say admit her. she says no....let me think about it.   she distracts.   shopping, alienation, socializing. coffee, charity do\'s, airline tickets and holidays. make me beautiful; her splendid array of animal coated products at her dispense, money. money. money.   what the hell does this shit prove anyway?  a hundred here. a thousand there. a couple of quids in between, pounds? aussie? ringgit?  a ball kicked through different holes, a specific amount. a random card chosen, a tarot flipped. choose your destiny, decide your future...this power. this surge of control. building up an itinerary...irony hits like a steel brick shoved up the ass.   talking to three travel agents, mastering online purchasing, sucking at connections, and liabilities. fly this for that for something or another.   three weeks later and the crave for independence begins again. i\'ll walk, i\'ll drive, i\'ll pay, i\'ll plan, i\'ll take care of myself, i\'ll be fine, i want to be left alone.  '),(22639,664485,'20,June,2003','2003-06-20 00:00:00','   she\'s screaming cause it\'s a quarter after midnight on a friday night and we\'re spent lazing on the couches watching shit satellite, playing a rather tacky game of mah-jong and sitting in our knickers over a pint of fucking delicious green tea ice-cream. the kind that melts just right in your mouth and enters your fucking system so damn smoothly it\'s like making love in chocolate syrup.   and the great thing is; we could very well perfectly be down at liquid or nouvo someplace in skimpy litte outfits with makeup and hair done just right partying to the latest dang gweiloh dj flown down, getting absolutely knackered and driving off in our little cars, talking on our little cellphones then picking up a one night stand, meeting the person of our dreams, or coming home to each other in the secluded security fucked, cops crawling, balisque shit of a hideaway we have and make love. but instead, we\'ve chosen this. is it right? is this real? is this what i want?  yet ashes lie scattered over the newly purchased home bloody-rip-off depot coffee table from an equally useless but so goddamn gorgeous ashtray. her eyelids flutter and her doeful innocent yet naughty seductive bullshit eyes pierce into my soul and commands my every thought, my every pore, my every decision. goddamnit, three weeks home and we\'ve started up this old horse again. hm.   but seven days ago she lay nursing me, three days ago she rubbed my back as i prayed to the porcelain gods, last night she wiped my tears spent from agonizing over medical drama and every morning, then noon, and night, she counts out the pills, lays them out, and chokes me down with them. is it that possible to hate and love someone at the same time?   she says run away with me. she says stay. she says lets have fun.   i just listen.   and sometimes i answer....sometimes i don\'t.  sometimes i raise my voice too loud and end up screaming.   sometimes i think things too deep and get emotional, curling into a ball of tears.   yet sometimes, i am in absolute peace. i am gracious and smiling, patient and understanding.   sometimes, i am forgiving.   it is rare, but sometimes, i am a gem. a not quite completely honed, brilliantly cut yet in need of a polish exquisite rock. diamond sometimes, ruby, sapphire, opal....anything that crosses my mind at that moment. any smile that washes over my face in that instant.   there are times i am beautiful, and these are the times i am grateful because love makes me beautiful.   ......i accidently stumbled across the old website we used to keep together today. a little after i read his mail, i didn\'t quite expect new posts, heck, i didn\'t even expect mail. but there it was, a two page letter, some photographs, and kisses in an envelope. so full of expectations, so full of hope, so full of things i am not sure of.   i read through the last few things you wrote, they were dedicated to me. i should be honored....i am honored. this is love, but love is a challenge...is love enough?   i think about the last couple of months...approaching six now, and i think what bliss, what a nightmare, what a fucking adventure. i think about waking up to you and falling asleep in your arms. i think about having you move in me, and the gentle way you kiss me from head to toe, the funny way you whisper even when we have the whole house to ourselves. the way you come in me and your lips kinda part....or the way you feel so hungry when you reach to grab me at the end of the day.   i think about sending you off and you waiting for me. i remember the way your face looked so sad and so happy for me, the roses just right and the music that always played even when we were fighting. i dance again alone remembering the way you keep me safe in your arms and move to the music with me.   i remember the passion in your eyes, and the standoffish almost arrogant attitude you sometimes give, the ways you know or the ways you think you know. and i\'ve learn about giving and taking but i know that there are many more lessons ahead.   i know your soothing voice as i go through three minutes in hell. the one that says you\'d like to be the father of my kids, the one that says it\'s okay, the one that says we\'ll take it one day at a time.   i also know the voice that agrees even if my argument is patetically weak. and i wonder what this break for summer is all truly about. did i need time away or did you already sense me pushing. is it like she says and makes a mockery of me for; am i that shitty at commitment? am i so pissingly afraid?  i say i need time. you say how much. i say i cannot. you say you don\'t know. i say things are brilliantly better off this way. i\'ll see you in three months and we\'ll see how things go from there. from where you say. will i fall in love? will i stay? will i come home?   i am in love.   i\'m staying for the rest of the trip.   and yes, i am coming home.   summer is our excuse for sin....and hand-written letters make me smile. and i smiled nice, huge and big today.   what the fuck is life all truly about? good question bozo, i wish i had all the answers for you sometimes. maybe i\'ll make the perfect wife, the perfect mother...maybe. you never know. shit, we never knew this would last as long as it has, as deep as this. i take out paper to write you a reply but i am so goddamn distracted, so many things to say....everything and nothing to show you.   flights, tickets, we always seem to do this halfway thing. no tahiti this year but i suppose the elephants would do eh?  this feels so fucking strange. almost surreal.   just in what in the hell am i doing?  just exactly which game is this?  i am bombfucked.   i am loved.   i am me.  '),(22640,664485,'19,June,2003','2003-06-19 00:00:00','   oh shit me. i just wrote the most amazing entry on post lesbian sex and blogger bloody fucks the entry.   fucking typical innit?  '),(22641,664485,'16,June,2003','2003-06-16 00:00:00','   i think you might like this. or not.   it\'s a digression nonetheless, not necessarily what i expected to happen as i analyzed every god damn thing about me last night while getting ready yet honestly, what the hell do you expect when you meet up with the one that shall be known as the goddess.   kuala fucking lumpur is an amazing city i swear.   :P  '),(22642,664485,'16,June,2003','2003-06-16 00:00:00','   it\'s a quarter after midnight and i\'ve tossed out enough of shit to drag my lazy ass to sit down and blog. blog something good i tell myself..write something real, something that has meaning, something that\'s honest, and kind, and true.....about yourself.   let\'s see; i am...  this gets a little hard doesn\'t it?  it\'s been hard to look at myself in the mirror the last couple of days. i can\'t quite explain it, no, it isn\'t the dozens of photos that stare back at me with steely eyes that irk me the most...i think it is the subtle pounding into my conscience that screams out you fucking double crossed bastard liar. you bloody good actress you. it is the acceptance that lead great double lives, sometimes even three.   the me inside, the one that cries, that hurts, that\'s for most part, an irreravant piece of clumsy shit.  the me outside, the smiling, amazing, popular, wonderful lifed bitch.  then the me in here, the chain smoking, foul mouthed, unbalanced, manic, depressed, bullimic, suicidal, retarded, silly little whore that pleases herself in churning out sentences that brings her closer to herself. closer to being real she says, closer to bringing the inner and outer together.   he says you\'re beautiful.  she says believe that and you\'re the dumbest girl on earth.  he says you\'re amazing.  she cackles and her laughs ring in your ears.   he says so many things but she knows all the counter attacks.   what are you really racheal? what are you made of? where do you belong? what is your ancestory? how does fame and glamor or whatever the hell you do the whole damn day mask the inner you? how the fuck do you hide the millions of skeletons in your closet? what makes you think you\'re truly worth anything, and of all things, what makes you think you\'re worth love?  because they think so damnit. because i know so.   right?  you are one fucking fruitcake racheal.   welcome to my world bitch.  '),(22643,664485,'15,June,2003','2003-06-15 00:00:00','   it\'s a little bit obscene at the moment. actually, fuck that, it\'s insanely obscene...  a little constrictive, a little daring, a little avant-garde meets my grandmother.   i am very amaused and very disenchanted by it all, a little jaded, a little naiive, a little insecure. i\'m not sure exactly...hm.   what are six words i would choose to describe it all?   uhm. beats me.   i\'m getting a little tired of this really, all the writing gets a little annoying, and so i think along with the holiday i might fill the empty lots with further more productive things than sit and bitch about my life eh? or not.   you never know, i might return after summer with a tan, a definite bet of a million stories to tell and an engagement ring or two. at the rate shit\'s going right now...this summer is heck of shit going to be interesting. :)  the pixie\'s are calling...excuse me now.  '),(22644,664485,'12,June,2003','2003-06-12 00:00:00','   \"i love you.\"  say fucking what?!  my mind explodes to about seven million thoughts and i am not sure if i should scream, gigle, laugh, cry or hide. and so i just remain silent...breathe in, breathe out, breathe in, breathe out i remind myself. i am composed goddamnit, i can handle this, feelings? emotions? drama? bring it the fuck on damnit, bring it the fuck on.   and so i scream, giggle, laugh, cry and hide all at once.   oh shit.   what the fuck is it with the ex-boyfriends in the last whole damn month? i am confused.   it\'s a fucking conspiracy i swear. i absobloodylutely swear.   decisions. decisions. decisions.   yay.  '),(22645,664485,'06,June,2003','2003-06-06 00:00:00','   lol, it\'s only about a quarter after six in the morning and i am WIDE AWAKE. being horribly jet lagged is an awful, awful thing, especially when you\'re knocked out by six in the evening and by noon you\'re ready to kill someone or just drop dead. in any case..being back home has been amazing!!  i\'ve been on a reading rampage from all the books i brought back from la and i still haven\'t gotten through it all! the manuscript is coming up along fine and we\'ve been working on it some. it\'s funny to relive so many stories and yet be so emotionally involved in a draft...very weird.  but a good kind of weird.  :P  hmm...the last three days have just been a constant flow of people, alcohol and shopping! lol, i\'ve been meeting up with friends from my high school, friends from nowhere, friends from work and friends from every bloody place. it\'s retarded because i feel like a broken tape recorder saying the same things, showing the same pictures, telling the same stories again and again and again. it\'s fun though!!!   that  ex also proposed to me two nights ago.   i think i will leave it at that.   ...or maybe explore it later but yeah, suffice to say, i was shocked, honored, livid, pissed, irritated, and bored. it was funny, fucking hillarious from another world. the guys got a good kick out of it though...:P   and yes, i did say no thank you though a slap across the face might have done a better trick. in some odd sense i was floating in another world as he told me bullshit about waiting for me, me being the one, you\'re the only one that has ever mattered blah blah blah and the only thing that crossed my mind was oh really? and too fucking late. and you don\'t know me and neither do i know you and i\'m not quite sure if we should do that either. in any case, he\'s convinced that actions speak louder than words and that i just need a little persuasion. i was too hungover, jet-lagged, tired, and frustrated to tell him about the other half of the story and where i stood. i suppose i will though, eventually. we\'ll see.   you see, it isn\'t that i don\'t give a fuck about our friendship and all that good stuff but that i don\'t get sparks or romantic notions and all that jazz when i see him anymore. my heart doesn\'t beat faster, my palms don\'t get sweaty but rather my mind sorta wanders around and that i finally see him for everything he is. flaws and all. and it isn\'t very nice of me to say but shit, i got fooled a long while didn\'t i? no doubt knowing someone for five years is a long, long time...and i know i\'ve grown with him, loved him, and all that but that moment has just passed and i don\'t want to return to it. if only because, or most of all, i\'m not the same person anymore.   back onto nicer things though....FOOD! fucking food. holy bloody mother of god, it\'s been a succession of food, food and more food since i\'ve arrived. that whole thing about malaysian\'s and their food? hell bloody true.   and OH MY GOD! the roads!!!! the drivers! the solution to it all, keep your attention to the screen, strap in, and pray really hard. lol, i\'ve given myself two weeks to a month before i\'m attempting to drive on the other side of the road again. good fucking luck. (thank you uncle ahmad, red and white cabs, and the swooshing trains)  hehe, want to know a funny thing? my room at home is about four times the size of my room in pensacola and so the first night, after balking at all the extra additions to the house and the brand new entertainment system in my room (thank you daddy) i absolutely did not know what to do with myself. lol, and so at four in the morning when i woke up (regardless of the fact that i only had two hours of bloody sleep, but did my body care? hell no. it woke me up anyway!) i had all this bloody space and i\'m like whoa. lol.  oh, and for the first time in a lot of months, my clothes are fucking ironed, my room is clean clean clean, my bed is made, and i have breakfast, lunch, dinner, tea, and supper prepared! hell yeah, this is the life. lol.   oh and btw, because i promised a mention, drew is the most amazing, wonderful, gorgeous, talented, whore master in the whole entire world.         (or so he claims.)   :P  ooh, and closet space!!! hehe, i have this tiny little white closet in pensacola that when opened, bursts out with six million things that aim straight for the most painful places to hit or smother you and back here i have a walk in and it is brilliant! clean, neat, categorized and all nice and pretty.   hehe. nick said the other day, oh my god you\'re one happy bitch. and i am!!! it\'s crazy. i got off the plane and i\'m like asian people! asian people! my cousins! my grandparents! my mom! my family! yay!!!!  other than the crazy, crazy heat (we\'ve had the airconditioning at full blast) malaysia is absobloodylutely amazing! :)  lol, and today we\'re going down to chinatown, which is always good, and i\'m pissed excited. hehe.   of course next week also starts a whole bunch of tests and appointments to tell me how horrible i am at taking care of myself but for right now, the only thing i can say is, HAH!  i\'m happy. this is good.   whoop! whoop!  talk to you later!!!  (i\'m getting to the emails soon i rpomise. just requires me sitting down for a period of time that\'s all.)   big hugss!!!  '),(22646,664485,'01,June,2003','2003-06-01 00:00:00','   so i\'ve started on what can only be deemed as the ultimate life-long dream of writing my book. what the hell is it about? a memoir? the first hand look from the fuck up girls mind? a sad  modest  desire in me to see my name in print...or just the longer, printed out, mass marketly available version of my blog? i\'m not sure exactly..but a couple of days at the book convention thronged with six million publishers, editors, authors and people trying to convince you that writing is the ultimate thing might do that to you.   or not.   we\'ll see. :P  how\'s la been, everyone seems to ask?  or at least all 368k of you that took up my inbox space that hotmail threatened to kick my ass and stick needles up me.   it\'s fine. it\'s beautiful. it\'s sinfully amazing.   i\'ve spent more hours in the sun, basking in the glow of very deliciously tanned boys from god knows where and from every corner of the earth..getting in line with everyone and their mother\'s being sucked into fucking great marketing from the people at universal, rediscovering my love for mickey, the anaheim ducks and bloody basketball and every damn news coverage of the downtown la area and back plus getting to know the bellboys, the housekeepers and the man that stocks up the mini bar daily very well. life is beautiful and things do seem to be getting better for the most part.  i take these long walks late at night after the partying is over and everyone\'s returned back into their rooms...and i walk these streets and streets that seem so ancient and so emcompassing at the same time. i take a million pictures of the beggars lining up the streets and the drunks returning from a good night out with feathery boas and weird dress sense and make friends with the new faces from different parts of the world all here on holiday or for the book convention. it\'s quite strange really.  i\'ve been to quite a few of these things growing up. my dad\'s business things but suddenly i am eighteen, sorta kinda grown up now and put into a different light. i meet these publishers that have seen me grow up and i run into their arms for the great big hug and i feel secure. papa rold especially and he shows me his cellphone and how my names imbedded in there and it\'s wonderful. like coming home really.  and me dad. god, where do i even begin? i cried and bawled like a little child when i met him off the plane at lax. he\'d flown in from narita and here i was, 6 hours later from pensacola, tired of the plane and fucking anxious to see him and i just burst out crying. i swore i wouldn\'t do it and even as i said good bye to the other parents, amanda and keith...walking in to board my plane and all that good shit i was still doing fine. laughing, making stupid jokes, tinkering around and being silly....but then i saw my dad and suddenly i felt so many things and it was as if some switch in me had turned on or turned off rather and i was my dad\'s little spoiled princess child again and i didn[\'t need to be in self-protection bitch mode because i was home. and safe. because it was that feeling of being loved for whoever i was, whatever i\'d become and whoever i was going to be. it was a good feeling really and i\'ve missed my dad and it\'s funny how much you realize what love truly is and means.   there\'s been so many little things in the past few days that have just sorta taken me aback and made me so thankful for my family. from the little surprise package my mom sent with my father to the gentle caring glow that reflects in their eyes. the way he sorta boasts and tlaks about me to his friends/business people and the way i know he\'s proud of me and i can be a cocain crack homeless absolutely useless addict shithead and he\'ll still find something to love. that just amazes the hell out of me and i am blessed. truly.   i tlaked to jared the other night too and it was odd. there\'s a lot to be said about the ups and downs i\'ve been through in the last few months but i feel sorta grown up yet naiive in some way now. some part of me is so ready to conquer the world, live out my life to the fullest extent i can live it too and yet at the same time i\'m sorta chilling and taking things one thing at a time. it\'s weird.   my brother\'s going to balk at this but here i go again tlaking about change as i move on into a high. bloody manic depressive freak. :P  some little part in me is beginning to believe in myself...it\'s as if these last few days i\'ve seen all these people believe in me and seen how i am someone in someonelse\'s life and it\'s weird. i don\'t know.  i\'m also a little worried about ryan at the moment but i suppose i\'ll ring up jessica tomorrow. i\'ll be home in less than a week and that\'s a crazy crazy feeling.   wow.   i\'ve also clocked in at least 200 laps in the last few days at the pool so that\'s kinda cool too. my bullimia is somewhat under control but it\'s still somewhere there but you know, you take it one step at a time eh? i don\'t know, but i\'ll be right...:)  hm. i need to get back to writing the damn thing and sending out some postcards so i suppose i\'ll talk to you guys later.   i love you and take care okay?  i miss you!  '),(22647,664485,'31,July,2003','2003-07-31 00:00:00','   i crept into his mind last night, and begged to stay a while. sobbing so hard i was suffocating, and said i miss you.   and said i\'m tired.   and said   nothing   at   all.   ---  because my resident  urlLink gracie poo  and  urlLink sexy meesh  are posting happy-happy things i figure i\'ll jump on the bandwagon and attempt to stop feeling sorry for myself.   hah.   :P   eighty(+2) things.    * being with the kids.  * laughing so hard you can\'t breathe.  * hot lazy afternoons spent with friends on the veranda.  * the rush of skiing down steep ass hills.   * slow dancing in your own world.  * watching the seventeen million colors of a beautiful sunset and making picture memories.  * reflection and the acceptance of change.   * watching someone you love light up.  * meeting someone new.  * the mak nenek across the street, the gentle way she cares for you.   * fresh laundered sheets.  * candle-lighted, incense filled, peaceful tranquility.  * long, long baths in the middle of the night.  * trigger-happy insane water gun battles with the car next door.  * pinching and appreciating a cute ass. :)  * finding out about yourself.  * discovering something new about someone else every day.  * unravelling a box-full of intellectual treasures.  * having all your senses poked and aroused.   * finding the perfect dress for a date.   * it\'s on sale.  * and you have matching shoes.  * making someone happy.  * understanding.  * loving like you\'ve never been hurt before.  * discovering your first same-sex experience.  * accepting your sexuality.  * learning to master a piece you\'ve been dying to play for a long, long time.   * actually doing it.   * bumping into random friends at the mall and the warm feeling it glows.    * making new friends.   * stupid inside jokes that leave you keeled over with laughter.  * going home.   * feeling safe.   * simplicity.  * reading a book and being transported into it\'s words and sentences.  * having little tiny arms around your waist.  * wiping tears off, and soothing a broken child.  * waking up to loving arms.   * making an elaborate meal together and enjoying it.  * showing your private own world through your lenses.   * taking the perfect shot.  * running down fluorescent lighted allies in wally\'s world at some obscene hour going crazy.  * long drives with the old man.  * the loud, raucous, insanity of an ice pilot game.  * working on a show, the excitement of being on stage.   * doing something you love.   * watching the chemistry between barker and blue-shirt, and the horniness on sexypant\'s face.   * making puck nauseous.  * midnight beach football with strangers and new made-up rules along the way.  * getting your cheography down just right.  * stolen kisses.   * hot chocolate and marshmallows on a cold, freezing, frigid winter day.  * exploring greek myths in the colorful explosion of a bookshop cafe.  * coming home to someone.  * working on your monologues together.   * first draft scripts and anticipation.  * crying, and being allowed to cry.  * butterflies, fairies and pixie dust.   * knowing the difference of being alone and loneliness.  * filling a void.  * learning to love.  * being needed.  * letting go.  * coming back.  * trying again.  * feeling strong.  * faith.  * trust, learning to trust and having someone trust you.  * long conversations about everything and anything.  * making love on the beach.  * tickle monster and pillow fights.  * sharing a home.  * cigarette lighter love song.  * finding a \'kuala lumpur\' shirt nestled in between racks and racks of clothing in an obscure thrift shop in pensacola.  * finding  urlLink friends online .  * reading a good post and feeling about every damn emotion run through you.  * the release of passion and believing in yourself.  * the  urlLink blogathon  and all it means.  *  urlLink 26 things  and the thrill of finding meaning.  * writing up a whole list and still having more to be happy, thankful, cheery about.   :)  thank you all, have a great weekend.      urlLink      urlLink  \"Call me Moody Swing Betty, but I feel spectacular today.\"  '),(22648,664485,'29,July,2003','2003-07-29 00:00:00','   it\'s about a quarter to two on a tuesday evening, rather early to be blogging but nonetheless, a typical melodramatic racheal post.   i\'ve just read random bits and pieces of \'prozac nation\', a gift from my father that i found laying outside my door sunday morning. pretty damn cool, i must say. the bastards at amazon ran out of stock so this gift from my dad is very special. in a million ways he will probably not know.   i have cold honey lemon tea sitting by my bed side, frankly, i am more interested in the pack of cigarettes that lay untouched stashed away in my drawers. i am trying to get rid of this nasty cold and all the gunk in my throat, i need to go in for shots sometime this week that they won\'t give to me until my immune system is up.   i also leave in two weeks, ten days short of when i had been expecting to go.   and once again, i still don\'t know how i feel exactly.   saw the head doctor today, dragged my ass downtown and she readjusted my medication again. it must have been the suicidal bitch in me that raised alarms....but who gives a hoot anyway? i am such a piss ass whiner  sometimes.    i have this strange habit of underlining phrases and writing on the sides of the books that i am reading. i date them too..and so i sit with a couple of yellow and black lined 2b pencils and a sharpener with a box of tissues and an afgan to crawl into. the thermostat is at some obscene temparature but the thought of moving across the room just seems like too much effort.   i read through quotes and poetry that my distorted self-destructive mind seemed to have penned in the last 48 hours. i didn\'t sleep for 36 hours yesterday, some manic insomniac phase taking over and the need to just be doing something seemed so much important. so many things to do with my life, so fucking little time. i must be going crazy, i am crazy. fuck me. fuck me. fuck me.   as i think about barker\'s hands slipping beneath the covers at night or spooning close with bare skin on bare skin and the surrounding beauty in the simplicity of being naked. i, the person that can only make love in the dark, that hates looking in the mirror, that throws up and starves herself, is okay. is safe.  or about him moving deep in me..making love to this broken body and tattered soul, the gentle soothing way listening to his heart thump against my ear drifts me off to sleep as he whispers in the middle of the night or hums along to whatever we have playing on cd.  i think about how amazing it feels to have someone to come home to, and how lucky i feel.   i am such a wanker to be feeling so sorry for myself, get me out of this rut. get me out of this whole. get me out of this bitch of a state.   and memories locked in some sacred haven comes tumbling down as i remember.   i remember going to the show with my brother last year. watching sophie ellis-baxter slink her way through and let her powerful beats move through my body. or that horrible flourescent lighted, rat infested, dark back alley mamak glory lane we went to after that. taking the tube with koko and catching up on life. on london and of all things, trying to soak in the comfort of being the little sister again. a different relationship now, it\'s changed. i kinda like it.   or the three in the morning sprees back in florida at wal*mart. and the alltel babe that we always manage to somehow con with amanda\'s boobs. i do have a great calling plan though, boobies do serve their purporse sometimes. :P  i remember mark from boston, as we snacked on ramen noodles and binged on snicker bars from the pantry. of taking off and landing, of twenty sum hours in between and nestled between our parents and movies. of being stuck in a plane for way too long, travel scrabble and air hostesses with oddly colored suits and funky smelling toiletries or the odd way strangers connect on a long flight back home.   or the new friends i have met, and how insecure i feel about me sometimes.   i remember the death sentence. and the caving of your insides as you try and breath and calm the fuck down.   or right after the surgery when my face was so bloated up and how fucking damn ugly i felt. a shoeboxful of get well notes and memories, all locked up, all looked at from time to time....reminding myself i have to be strong. i need to be strong.  i think about the insane craves i have for scones or a cup of teh tarik. for marmite or baked beans and toast. magee mee goreng or someone to just bloody know where kuala lumpur is.  i remember, and sometimes it is hard remembering. sometimes, there are so many little details you start to forget, or leave out.   i forget the way my grandfather smells, or if akari liked that bitter kerepek thing. i forget because sometimes, all the memories hit you at once and it gets kinda displaced while sorting it out, and it\'s confusing.   and i am sad but happy, excited but afraid.   forgive me for the manic-depressive sorry whining bitch i have become in the last few days.   i love you all.   and i am just rambling at some hour of the morning, just to say something, just to write.   just  to   remember.      urlLink  \"A whole shitload. Everything around you falls apart, like those buildings stuffed with TNT, demolition.   But, is it worth losing everything?\"  '),(22649,664485,'26,July,2003','2003-07-26 00:00:00','   the blogathon has started...i\'ve spent the last couple of days trying to come out with a decent entry but it\'s been hard to say the least. i bite back the tears that make me feel so weak and insecure even as i write this, i try and come to terms with life and all that surrounds it...i suppose there is a point to it all. i suppose there will be reason and understanding beneath the layer of tranquility that i will find eventually.   i counted out pennies and quarters, spare change...wanting so desperately to make a difference. a little part of me believed that things could go on, my few dollars would make a big difference, there would be world peace. and i could save the kids. and no one else would hurt.  i sit here now, flipping through hundreds and hundreds of pages. i have an insane urge to just run from it all, the stories, the words, the pictures leave a sour bitter after taste like rotting food or bad sex. it is hard to muster how i feel or swallow the tears, but still, i keep on reading. and keep on crying. and keep on trying.   cancer scares the shit out of me.   i\'ve said it out loud. but seven mere words don\'t even come close.   it is a dragon i fight with my demons. it is sanity i plead with the gods up above, the entities, whatever supreme being that rules this earth.   i am afraid. afraid of so many things, things i don\'t have control over at all. i am afraid of my disease, i am afraid of myself and growing up, of not knowing my place in the world and not knowing what i have to offer.   i leave in a couple of weeks. i don\'t exactly know how i should feel, i feel...confused. and tired. and melancolic. i feel...excited. and weary. and laughy.   i am picking up things from my life that i have left to rest a while. my theater degree, my hopes, my dreams. him. and coming home to him, and working on all that we are...wiser over the summer, realization hits like ten pound bricks; painful. but nonetheless helpful.   he says   \"Hang in there baby. I need you. \"   i whisper,   i need you too.   and crumble, fade alone into the darkness.  i think i\'ll go away a while. be by myself for a couple of days. i\'ve been avoiding people and faces, it just seems so wrong. i feel so guilty.   i didn\'t take my medication on thursday or friday either. things seemed pointless and hopeless at some stage. i felt like i was cheating myself, making myself believe that i would be okay, that i would be cured, that i would be healthy just by a couple of pills.   i don\'t think it helped.   i spiralled so far down i don\'t think anyone could quite reach me.   fell.  stumble.  cried.  hurt.  scraped.  bled.   i begged god, asked if he was listening. wanting to know if he was there.   i remind myself constantly that suicide is not an option. it is never, and it will never be.   i\'m just tired sometimes. i just want a break.   my throat is so sore from throwing up, but i cannot help myself. my stomach so empty, my thoughts so jumbled, my room so horribly musked up with cigarette smoke.   i am not feeling too well today. i don\'t know what i want. i just wish....i don\'t know what i wish.   i\'m okay.   i\'ll be alright.   i\'m a big girl now, i\'m okay.   i just want to hide in the corner a while.   disappear.  '),(22650,664485,'23,July,2003','2003-07-23 00:00:00','   akari stopped treatment three months ago to go back home to indonesia. they buried her today.  unexpected? not really. i don\'t know. i just didn\'t think it\'d be so soon, i just didn\'t think it\'d really happen. i just didn\'t think her....her.   i don\'t know exactly how i feel right now.   i think i need to cry, but the tears don\'t come. i just remember the dreadful silence, and the caving of your insides. the be brave, smile, be strong front you put on for the other kids, and go through the routines, and make them smile, make them laugh, calor another butterfly.   she\'s dead.  fuck you god.   fuck you fairness.  fuck you life.   these are the times you come back home and ask just what the fuck you\'re doing? these are the times you come home and the only thing you really want to do is just make it all better, take the pain away, the screaming voices, the wailing children, the catatonic so fucking drugged faces out of your mind. these are the times you question, question, question why you do this, and you want to know how to be strong, and you want to know all the damn answers, and you really just want to know why?  i sit here, a blank screen before me searching for the words to somehow describe how i feel. i cannot seem to find any adjectives, and verbs, anything at all that comes even remotely close. i am so many things at once, i am, i don\'t know.   you get close to the kids, whether you like it or not, there are always the special ones that stick out. the ones you keep close to your heart forever, the ones that make your whole world seem so surreal with stupid bad news like this.   i sit here remembering all the different things we used to do together, how she\'d cling to my waist when i left, how we\'d play with wigs and hats, or color and color and color. she made this origami thing of 180 folded pieces with her mom for me before i left last year and gave it to me for my going away party. it was for luck she said, it was for good wishes, and dreams.   she is still seven. and she will never be eight.  and she will never be all the things she fought so fucking hard for.   what the fuck is wrong with this world?  what the fuck is wrong with this whole bloody shit of a system life that we have?  i just want to know why.  i just want to know how.  i just want to know.  she still has big grown up prom dresses to pick, first crushes to kiss, ambitions to fill, barbie dolls to play.   i want to live in denial for a while and pretend this isn\'t happening, but i cannot. but it\'s real. but she\'s gone.   you try and get the sick nauseating smell of the hospital out of your mind, fresh in your memory. or the sores, or the patches of hair falling off, or the horrible horrible nightmares.   akari died today.   why?  what do you say to someone who\'s just lost their only daughter?      ---   i\'m okay, don\'t worry about me. if you believe in god, pray for her okay? and her parents...she has two other older brothers and her mom is expecting. it\'s too early to tell if it will be a girl or boy, maybe that was akari\'s last gift to her parents.   thank you so much for all the kind words. for some odd reason, i couldn\'t cry yesterday. i think my medication does that, but tonight, tonight i had a good cry. and i wanted to thank all of you for being there, for all the comforting, reassuring, calm words. i don\'t think you\'ll ever know just how much it means. thank you. terima kasih. xie-xie. nandri. danke sehr. arigato.   i went on a rampant search for pictures of us together last night too. and it was good to remember her smiling, i still don\'t know why crap like this happens but maybe there is a reason, i don\'t know. thank you all again.   thank you to the  urlLink bloggers blogging for cancer  too. for providing hope and filling up the void.   hugs.  you guys be safe and take care.   i\'ll be (al)right.  don\'t worry.    p/s remember to tell someone you love them today okay?  '),(22651,664485,'22,July,2003','2003-07-22 00:00:00','   lunch with the ex-girlfriend.   urlLink   \"And it hurts me To look into the mirror at myself And it hurts even more To have to be with somebody else And it\'s so hard to do And so easy to say But sometimes Sometimes you just have to walk away\"    the morning sun blazes it\'s way through the afternoon, and the single solitary tear from god cascades into a kaleidoscopic ray of colors; a gorgeous rainbow that spreads throughout the valley...we light up watching the magnificence of mother earth and humanity.    \"is it okay if i hold your hand?\"    i lace mine with yours, assuring you.   of course it\'s okay.    urlLink  they  wail into an ironic apt afternoon.   \"Running in circles Chasing tails And coming back as we are  Nobody said it was easy Oh it\'s such a shame for us to part Nobody said it was easy No one ever said it would be so hard\"    an almost endless stream of silence and conversation. the little old man at the table beside us enjoying his scoop of ice-cream, glee in his eyes and his quiet, patient, loving wife, with her metallic hair pushing his wheelchair as if holding on to his hand.    \"i want to grow old with you.\"   ouch.   the restaurant rings with us in the air. the ambiance so serene, peacefully calm and tranquil in so many ways....maybe it is the waterfall behind us, and the gentle gushing sounds. or perhaps it is the lighted candles: of different sizes, of different colors, of different scents to intrigue our minds. fresh orchids in clear crystal vases,  lips on bare naked skin, rays of light piercing through the windows, clouds of smoke surrounding, misting us.   it\'s ending.   i\'ll miss you, i whisper into the air and everything that surrounds you. afraid to look you in the eye, afraid i will cry and it will never stop. afraid my heart and yours will break with letting go, smiling. you\'re so gorgeous, so amazing in so many indescribable ways and i cling on to you as if i could cling on to you forever. holding you close, stroking your hair, listening to your heart beat against mine and kissing your tears away....almost too desperately begging you not to break down. it\'s tough, and i giggle into your ears, and you listen to my heart.    \"you smell so, so, so familiar.\"    i love you.     urlLink  \"Reached out and caught you.   Yes, I want all that.  With you.\"      urlLink  \"Here I am. Yes, I\'m still here.\"  '),(22652,664485,'20,July,2003','2003-07-20 00:00:00','   urlLink    urlLink    urlLink    urlLink    urlLink    urlLink    urlLink    urlLink     she slips away into the dark at the break of dawn, catches the sunrise just as the day breaks and she\'s flying out of the city into the mountains. to look for peace she says, to relish in solitude a while, to disconnect from the outer world, the people around her, to be alone with her demons a while and maybe contemplate on life a little. or the quality of it rather.   she doesn\'t pack too much; her time-worn tattered precious journal of too many years and too many stories, a flask of hot coffee, a couple cans of diet soda, some fruit and an armfull of cds and mds. a little bit of krall and sondhiem, a little bit of dashboard and sheik, a shitload of sp,marvelous three and sigur ros. heck, she even threw in disney just for good measure and a whole lot of self-made, friend-made, compilations.   it\'s the crack of dawn and she fills up the tank to go up the windy roads. she hasn\'t driven in a while, but i suppose she needs the driving therapy this weekend. she goes in to see the head doctor again tomorrow so it is only a day trip, slightly less than eighteen hours. and as she stops at the gas station, she catches the most amazing sunrise she\'s seen in the last couple of days, weeks even. away from the city and nestled in the valley, she sits and cries her silent tears. meditates, prays, smokes her cigarettes, steps on the gas, passes by cars and cars of happy children, excited students, vacationing families.   it\'s cold as fuck up there. misty, foggy and chilly too, but she loves it.   layered and bundled up she checks into the hotel and has tea with scones. so bloody very english it almost makes her laugh. and so she does, she giggles to herself, hums, the smiles return after a while. the rainbow comes after the rain, the clouds clear up, the thoughts decipher themselves and make sense somehow.   she had a strangely disconnected saturday. from waking up to falling asleep. pretending, smiling, laughing, hiding. don\'t show your tears, disguise your pain, disconnect.   she listened to them sing, screamed i will survive, and had seven thousand and sixty two messages come in from her. the questioning still continued, and as night fell, she watched the ex-girlfriend break down.   sometimes i don\'t know if being tough, and strong is all at all worth it? it drew blood last night, to see her cry, to see her confusion, to listen to her....to consciously feel her falling and losing grip on her fingers. but i couldn\'t give her the answers she needed to hear, i couldn\'t be the person she needed me to be, i am not the person she needs me to be.   i kissed her, as if it would make things better. i held her hand and held her tight, a part of me never wanting to let go, another knowing i had to walk away. it was surreal almost, something i would definitely not have guessed at all...the roles have fucked themselves over and changed again.   i hurt her.   she shed her ms tough girl act.   i walked away.   i tossed and turned the whole night trying to justify my actions. wanting so badly to reply her messages, wanting so badly to tell her it\'d all be okay but consciously, or subconsciously, or whatever the hell it was, i knew it wouldn\'t happen. it couldn\'t happen.   and so in the aloneless today, i found myself again. taking away all the emotions and drama, taking away all the comforts and dealing with the brutal truth. i love her too much to do this, i love her too much to have us hurt that crazily much again.   and barker laughs at the irony of it all. ina came to jersey this weekend, we had a long conversation when i got back. about things that had happened, about things that were happening, about the future, about our pasts, about the currents and the current.   this summer has been so strange i\'m not quite sure how to even start organizing it in my mind frame.   but for now, there are post-it notes.  --- thank you callista, grace, irene, hani and ash for listening to me croak and singing so beautifully. you made certain parts of the weekend bearable and meesh, you are beautiful and loved, don\'t ever let anybody take that away from you.   the other kl bloggers and the gender genes group, for making kuala lumpur so special to me and opening up a whole new world.   big two, for talking some sense into me, for conversations, for your amazing voice, i hope it all works out for you. for looking out for me, thank you always.  chris, linux and cake, lime ice-cream, drama, drama, drama, bitching. fix your bloody site!  sloppy,  i love you no matter what happens, no matter where we end up, no matter where we go from here. you will always, always be a special someone to me, and thank you for the lessons. the road forks and we carry our angels eh? take care of yourself bimbo, you mean the world.   puck, recieved the hand-written long notes in the mail yesterday. thank you for the emails, the blasted phone bills, the long drive to the postoffice and sexypant\'s update. see you in three weeks!  old man, i hope you\'re doing fine. i know it\'s been a tough couple of months for you but hang in there okay? we all love you and i miss you much. hope the eye is healing up and the smiles are returning.  babe, i miss you too and i think we both knew going into this that it\'d be hard. too many skeletons in our closets, too much going on, too many emotions and feelings and all that good stuff. thank you for believing in me, trusting me, loving me, and whatever happens, happens i suppose? expectations smanzy, we\'ll do our polka dotted tango together when i get back and finally shoot the fucking movie eh?  koko, i can\'t wait to see you and no, you can\'t have my prozac. i love you silly dinky and thank you for letting me share this side of me with you and not bursting the bubble. have fun with the prada and the gucci.    ---   and so, even if i, the fat lady, sang yesterday...it still ain\'t over yet.   :)    urlLink  \"This site.  This journal.  This niche.  This is the mirror of my stupidity; the reckless life of a youth and the irony of my maturity and naiveness.\"      urlLink  \"I find myself looking forward to tomorrow.\"      urlLink  \"So here are my feelings. In black and white. For the entire internet world to see. A fountain of emotions, thoughts, wishes, and feelings. Wrapped up in one blog with mistakes and holes everywhere. I could do this for hours and hours and hours.   Something\'s right.\"      urlLink  \"..... then, everything falls into place perfectly.  And she lived happily ever after.\"  '),(22653,664485,'18,July,2003','2003-07-18 00:00:00','   urlLink    urlLink    urlLink    urlLink     i wrote one long ass-humping entry last night sobering up but in which i brilliantly screwed myself over and somehow managed to delete. i swear my stupidness amazes me sometimes.   --  i accidentally ran into her last night. actually, i take that back, i ran into Her last night. Her that denotes a big fat capital fucking h, her that started the whole ball rolling of gender issues, and crossing the lines, and questioning sexuality. her that brought me into a different realm, the other side, so to say, equally her that means so much in such a little way. or not. i\'ll tell you when i figure that one out.   i was not prepared to meet her, to say the least. we were at yet another one of those five million places in kuala lumpur where they had seventeen million clubs in one place. i can\'t remember what exactly happened that i ended up going, despite repeated attempts of seven hundred and six excuses that went unheard.  we were tequila happy with shot after shot consumed. if i have to think about it, i suppose i stopped counting after six or seven, too busy chasing the purple monkey on the dance floor that i swear winked at me. now, as i look back at it, i know there were signs. signs all around, signs up my ass, signs spurting from the backalley and the sidewalks.   attempting to walk straight, getting picked up by strange random men, screaming, laughing, waiting for the asshole jockey who parked our cars seven thousand miles away to come back. the night seemed to be coming to an end, adjourning to birthday girl\'s house, moving on to the birthday suit party.  but then the phone rang. flashed.   her.   on the next street.   right   there.   and i\'m not quite sure what i was thinking. i\'m not sure if i screamed, or if i had a grin the size of my ass plastered on my face. or maybe it was both. i think oh fuck, shit, hell yeah, uhm, damnit, uh-oh, and yay ran through my mind a couple of million times in those nanoseconds.   i might have ran into her arms. we might have exchanged six million words, and danced around a little. stilletto shoes and all.   we sat outside of absolut, now progressing to double shots, and talked. and exchanged words, i\'m sure i slurred. i\'m sure i said a lot of stupid things. i\'m sure i just smiled and pretended to comprehend. i\'m sure i pretended not to notice that she\'d switched from malboro to salems. or that she still wore that chain. or that i could not keep my eyes off her. or that above everything, i would always be in love with her.   i know she questioned that. hell, i know last whole night was a bloody load of questioning. of trying to understand where things stood. of seeing someone you\'ve thought of plenty and plenty of times before. of runnning again into someone you\'ve run into a million times in your head before, and what you would say, the things you would do, the questions you would ask, the answers you would give.   i pretended not to notice when her hand touched my back as we crossed the street. or when her arms brushed past mine as my left hand and her right hand drew in breaths of our salems. our cancer sticks. or the way her lips would move, her brilliant smile. damnit, to be mesmerized by her again. to question.   and there are so many thoughts that go through my mind right now but the headache isn\'t exactly helping. i have three weeks left, and that\'s three painfully too short weeks.   humpty mouthing \'ryan?\'.  birthday girl screaming, yelling, taking pictures.   ass-happy getting more ass happy and flashing thumbs-up signs and getting more sloshed.   and the rest of the other people at the house, inside. around.   and us.  and her.   and this.    i don\'t think i\'m quite making sense at the moment, nor do i know what is going on. or how i feel. or what i want. or how she is. or if i\'m fine. so i think i\'ll climb back into bed and nurse the hangover, raw eggs and all.   damnit.   urlLink  \"Whatever causes these things, whether it be an amazingly humbling image, the vision of a heavenly creature, a moving song, a familiar scent, a stolen kiss or an unexpected smile, they\'re the kind of jolts that we need once in a while to remind us that we\'re alive.\"     urlLink  \" But I also hope it won\'t be too hard to leave it all one more time to come back in the fall.\"  '),(22654,664485,'17,July,2003','2003-07-17 00:00:00','   she scribbles away in private ecstacy, of purging one\'s self of thoughts and voices, opinions and accusations. the immediate relief as solid black ink touches crisp white lined paper, a journal, a diary, a fuck-it-all.   wrinkled, tear-stained, yellow entries mock her. reminding her, bellowing out at her, screaming of what was yesterday, and what is today, can never quite be forgotten. she broke down in many, many pieces last night...they questioned her. they asked, prodded, begged, violated, screamed, yelled, cried, soothed, calmed, tried to understand her.   a million questions left unanswered, a cave of rotting secrets and hard-kept facades and masks exposed, violently thrown in their faces, as if daring them to accept her, as if daring them to love her. love her? the harsh, falsetto lined, cackling, mocking voice laughs at her. you\'re so stupid sometimes silly girl, you\'re so purely, insanely dumb.   she fights back, but she is tired. sick already, and so her defenses aren\'t quite as strong as they should be, she scratches but it is only a surface, she  pierces but only the outer layer.   she looks at swollen eyes in the mirror, bruised body from broken veins at the gym, throat sore and bleeding and a  half cracked smile wades it way through eeriely, she\'s okay though. she\'s alright. notes, poems, prose scribbled all over the books she\'s brought back that lies on her night stand. of thoughts and actions, of important words underlined, and someone accidentaly read it last night, and someone accidentally intruded into her private pain. i don\'t think they meant to, they just did.   you are so fucking morbid she tells herself. you are so fucking weird. as dreams invade her already sleep-deprived insomniac mind, the little three hours that she spends tossing and turning into a nightmarish jumble of people and faces, instances and things. she runs and runs, but there is no exit. she burns alive. she burns breathing. she burns inside.   alone now. afraid to trust, afraid to step past this ledge to see what\'s below. love thrown at her feet but her hands cannot seem to grasp it. bend over. pick it up. how?  she doesn\'t feel good today. maybe it\'s just another one of those days. maybe she\'s just tired. maybe, damnit, there are so many maybes. maybe writing as a third person throws her into an oblivion a while, of pretending that it\'s all okay. maybe that is her defense mechanism, the ability to step away from it all. the ability to just ignore. the selfish immaturity of just turning the other cheek.   she\'s still a kid lost in a big big world. and like a kid, she still has hope. hope of finding, hope of dreaming, hope of so many things and beyond. pixie dust, fairies, god, angels and all that. someday she tells herself, as tears put out the burning candles, i\'m alright. no big deal. just another day.  let me hide in the darkness just a while.    urlLink  \"Her heart raced at a lunatic rate, bells rang, the voices laughed, somewhere in the not so distant past a baby cried, and above all the silent noise she whispered\"     urlLink  \"because no one listens and so she cries (like the sea and waves) because everyone hears but no one listens\"  '),(22655,664485,'16,July,2003','2003-07-16 00:00:00','   because i didn\'t have much to say and some weird stranger wanted to know. :P   urlLink      eighteen random things you really didn\'t need to know from racheal about life.     i.   sometimes, when i cannot sleep late at night, i sit crosslegged on the floor in the middle of my room and light up my ciggarettes in complete utter darkness, just because.    ii.  i also seem to learn a lot of about life that way. thinking.    iii.   some lawyer in london related to me is going to start blogging here. :)    iv.  i turn nineteen in a couple of months. i think i\'m scared and quite excited at the same time.    v.  as wrong as this sounds, i refer to knackered as being piss drunk. (which happens against my will more often than not.)   vi.  i think starbuck\'s mud pie ice-cream on oreo cones along with a side of great company is absobloodylutely terrific. though whipped-cream smiley faced waffles come in a close second...    vii.  i think the idea of discriminating against sexuality is absurd.    viii.  i have a thing about eights. don\'t ask, i don\'t know why either. lol, just strange like that i suppose? :P   ix.  i am pretty bad with the whole controlling emotional behavior la di da though i also think i\'ve been doing not too bad lately.   x.  but then again, that might also be the prozac talking.    xi.  i have very very strange dreams.    xii.  is it possible to be needy and distant at the same time?   xiii.  i come out with so much shit sometimes it\'s amazing.    xiv.  it\'s stupid but sometimes i apologize for being who i am.    xv.  and other times, i\'m such an obnoxious little confident bitch which gets interesting i suppose. :)   xvi.  i\'ve been blogging for two years now...and it\'s damn hard to shut me up! :P  sarahkate: discovered. even angels fall.  karac.com/~just me hiding. sobrevivir : learning to survive. heartstrings and shiznit. mistated; life is like a dick: when it\'s hard, fuck it. she = hopes: there are moments when whatever the attitude of the body, the soul is on its knees.  sarah @ warmcompany. latin for love. mnemosyne: stripped.    xvii.  my redesign dateline is two months. now whether that will happen, is a different matter completely.    xviii.  exit upstage: hope is a damn blasted nice feeling. :)      urlLink  \"I got up so early and everything felt so different.\"  '),(22656,664485,'15,July,2003','2003-07-15 00:00:00','   uhm. yeah.   even having said that though, yes, please.   :P   p/s sextitlating/incriminating photo  stolen  borrowed off  urlLink msn .  '),(22657,664485,'15,July,2003','2003-07-15 00:00:00','   i stumbled into their bedroom somewhere between the hours of dawn and dust. it couldn\'t have been more than five in the morning, perhaps around three or four. i knocked on their locked bedroom door and set off many, many alarms at that. some infra-red shit sensor had caught me and that definitely woke them up. i felt like a thief.   i said i wasn\'t feeling well. eyes welling up with tears clutching to my body pillow, slanting against the wall. i said everything hurt and i felt like shit.   and she put her arms around me and i cried.   i cried and sobbed for the longest time ever, not quite knowing if it was just the physical pain that was making me so delirious or if it was something more. something more like i only had four weeks left, something more like i couldn\'t decide, something more like i missed being cared for, something more like i didn\'t feel like being responsible and grown up in that moment.   and then i think we both realized that i was still a child. that as much as i tried to be tough all the time, the smart ass that i am...i still had so much to learn. that all the facades and all the masks faded; i think she saw a different me last night, and i think as much as my demons are fuck heads in my mind, i think they helped me show her a little of my world last night.   and so someone did tuck me into bed.   and made me fresh orange juice at four in the morning.   and laid out my multicolored pills to take.   and now, i do feel better. a little stronger at least, a little more renewed.   but still wondering. still waiting. still trying to figure out life.   still growing up.   thanks for the get-well wishes you bums, this hard-headed annoying whining bitch gets all touchy-feely sometimes.    hugs  :P     urlLink  \"As the final dust motes settle in the moon\'s silver glory, she looks at the chances and the stories left untold.\"  '),(22658,664485,'15,July,2003','2003-07-15 00:00:00','   the server is having problems of it\'s own. big time.   and about the only bloody thing worst that having your head pound, your insides feel like mush, your nose give the tap a run for it\'s money and every single solitary part of your body hurt is THE FUCKING HORNY NOISY SCREAMING BLOODY CATS next door.   it\'s 1:49 in the morning and i\'m running a temperature.   i am not in a mood to be fucked with at all.   i look like shit and i feel like shit and nothing is helping.   i am lonely.  i am tired.  i am cranky.  and i am sick of being sick.   right about now, i\'d like someone to cuddle me, wipe my tears and tuck me into bed.   :(  goddamn this distance.    p/s is it terribly against the law to pour acid(my cooking) on cats?  '),(22659,664485,'14,July,2003','2003-07-14 00:00:00','   i dreamt about her last night.  i remember bits and pieces now but they are rather haphazardly put together; like an old, old black and white film restored, jumpy, missing bits, sound and picture quality not quite right.   we were holding each other close in this stark white empty room and there were no words. only us breathing together. only us waiting.   the floor was layered with fine white silk except for a splattering in the middle of red taffeta cloth; i\'m sure it must have meant something, heck, i\'m sure the whole dream must have meant somehting. she was so close though, so close that i could smell her. smell the faint scent of her perfume, sexy body perfume i used to call it because of the bottle shape that it came in and i miss that smell. i took it all in, the smell, her smell, being in her arms again....love.  i could feel my back against her breast, so smooth, so naked, so aroused. by me? but yet there was silence, and yet we did nothing, we just sat there...waiting...though waiting for what i wasn\'t sure.   her white almost translucent skin against mine, a long long line that ran from her navel down to the tips of her toes. her curly sun-streaked hair, her smile, the sadness in her eyes, her blood red pouting lips.   and i missed her.   i stirred awake, wanting so desperately to hang on to that dream, wanting to ask her so many questions but she was already gone. i took a deep deep breath in, wanting to comfort myself with her smell again but there were only traces of blood on my pillow and a mind numbing headache to wake up to. she wasn\'t here, it was only a dream.   it hadn\'t happened.   then why?  then what?  stepped back into the polluted streets, the raging motorcyclists, the beeping cars, the sweltering heat. stepped back into reality, into the stark oppresion of sexual prefferation, into the blinding aftermaths of progression. regression.   i swear the drugs are making me hallucinate.    urlLink  \"People like to ask. For once in my life, I say I don\'t know, and I like it that way. \"  '),(22660,664485,'14,July,2003','2003-07-14 00:00:00','   so i was telling stinkypoo last night how amazing i felt, how healthy my new vegetarian diet was ; how this bodybalance thing i took at the gym that had yoga, tai chi and some other ancient asian practice plus the working out, getting the toxins out of my body, changing back into this dermalogica aromatheraphy soothing smanzy shiznit and all that bloody bullshit was making me feel so good and all that.   i wake up this morning and not only do i have cold, a head pounding shitting headache but that i also have a nosebleed.  fuck.  that.   my throat feels about as good as sandpapering your insides.   :(  we also have tickets to go see the bolehwood awards tonight.    life is swell innit?  hope your day\'s going a whole bloody lot better!   love, the bitch from hell.  '),(22661,664485,'14,July,2003','2003-07-14 00:00:00','   so i\'ve decided to sponsor three amazing people for the  urlLink blogathon  this year.    urlLink michele  because i love her and it\'s become sort of a tradition.    urlLink peter  because i know he can do it and world education is always good.   and newly found  urlLink aphoristic  because he\'s helping get rid of the \'dragon\' and there are knicker promises.   on a side (sappy) note, this is also for  urlLink brian  , wherever you are buddy, whatever you\'re doing, i hope you\'re looking down from heaven with the kids and smiling, at peace, and rocking it out. i\'m fighting my demons, dragons, whatever and whatnots down here and thanks for everything.   ....i\'ve been working with the kids for about three years now and it is a challenge sometimes. an extremely rewarding beautiful experience nonetheless, but heartbreaking too. in some ways, it has shaped me into who i have become and made me stronger.   i was looking through the archives of the other place and attached a random cut-paste collage of things i wrote about the kids. it\'s a lot of drama and what-not but oh well.   hugs.  thank you michele, peter, and keith, for doing what you do.   if you have a little bit of change, or words of encouragement, or anything at all...i hope you guys would consider it?    ----  urlLink  [ Mon Dec 10, 2001 12:20:21 AM  ] facing up.   \"i realize how i have so little time left with them, how time is precious is so many ways, especially for them. i realize that for those few hours yesterday i was genuinely happy, genuinely content. that for those few hours yesterday my brain wasn\'t over reacting and my heart didn\'t feel so broken, in fact i think it was dancing.   i was dancing because i managed to make someone forget about their pain for awhile. me, this totally insecure crazy manic wimp made someone smile, someone happy, made them laugh. laughter is bliss in there, above the suction noises and constant bleeps that harmonize after a while. those irritating heart machines that beep every few seconds provide security to the moms and dads that sit by them, that wipe up the puke, that rub their backs, that cries silent tears...those irritating heart machines that beep provide hope....and i like hope.   hope is good. these are my kids, and everything is worth it. everything. \"     urlLink    urlLink    urlLink    urlLink    urlLink  '),(22662,664485,'13,July,2003','2003-07-13 00:00:00','   i.  dont, under any circumstances at all, stick a pen down the socket when trying to fit a two-plug down a three-plug thing. you get electrocuted.    ii.  always switch phone to silent mode when in church. people might start giving you evil looks. especially the old ladies with afro looking hair and bad b.o. very traumatizing.    iii.  when going to eat at the pasar malam, don\'t ask what\'s in the food. you /really/ don\'t want to know.    iv.  pinch ash\'s yummilicious butt.    v.  hunt meesh, hani, and grace down so you can go down to the mamak stall, order teh tarik(lol, or todi?), eat a lot,  bitch, chain-smoke,  admire their boobs , take stupid pictures, and mengada-ngada.   vi.  don\'t tie a string to your finger to remind you to do something because a) you will forget, b) your finger turns blue, c) you look stupid. :P   vii.  thank profusely when recieving gifts from the way too friendly pak-cik no matter how strange sounding, weird looking, funky smelling the object is. or in my case, the shirt-cum-jacket-cum-some-poor-puppy\'s-furr. it makes them feel special and you can always donate it to salvation army later. people like \'someone you know\' will definitely find a use for it. :P   viii.  you don\'t want corn, red beans and peanuts in your ice-kachang even if the weird people around you tell you you must have it that way because you will remember all the reasons why you don\'t like having it in your ice-kachang in the first place and you will not be happy. ruined ice-kachangs always makes you sad.    ix.  if the stupid lady with the big white hat tells you it\'s going to be just like an ant bite, RUN.    x.  don\'t send chris songs about linux. send microsoft happy songs instead.   last but not least, don\'t post bullshit like this is your blog. it makes you look very bodoh and cham.   :P   \"sunday-lah....relekleh brother!\"    urlLink  \"You dance with the Devil, the Devil don\'t change, He changes you.\"  '),(22663,664485,'12,July,2003','2003-07-12 00:00:00','   would you trace the line that runs down my body and let your fingers love me in? would you kiss my forehead and sing me to sleep, read to me, tell me about your day? would you let me grip you tight and gyrate deeper in you as i come? would you let me see your heart, your smile, each night, every morning, whenever i open my eyes...it\'s hard to explain, this empty bed, even colder heart as you try and make our worlds seem so much closer than it actually bloody is.   i\'m worried, i suppose because i\'m insecure and silly sometimes. i don\'t believe that you believe in commitment, i have little issues i\'ve carried from before but nothing quite as extreme. six months with someone and you learn a lot about yourself, about compromise, about recieving and perhaps most of all, about giving.   summer with definitely more than a couple of obstacles in the way, i think it\'s been harder than you expected it to be. i hear it in your voice sometimes, when you sorta get all quiet and soft with me and you almost whisper, i miss you. sometimes, i say i told you so, and realize that we needed this. that whole saying about treasuring the things you need around when they\'re gone is amazingly profound innit?   i love listening to you laugh. it\'s this giggle, hearty, happy little chuckle you have and it fills me up with so much joy. :) i love it when you say love. i love it when you say my name and i can hear you smile across the line. but i suppose, the thing i love most about you, is who you are.   you see, i don\'t need the dozens and dozens of roses, or the little things you pick up along the way, i don\'t need the simple thoughtful unexpected surprises, or all the crazily planned out carved out dates that are always so perfect. no, i don\'t think that;s quite it that makes me burst out laughing and singing at the same time, i think it is the essence of you. of the person that\'s beneath all the jokes, all the sensitivity, all the seriousness, and all the loving, i think it\'s about you. and the funny strange way you fit into my life.   it\'s not those things that you think are important to me, but it is so much more.   maybe it\'s love.   only you would clean the whole entire bathtub with a toothbrush and liquid soap for us, and only you would dress up with me to have a cute night out even if it\'s a weekday and we\'re both tired as hell. only you would make (and deliver!) cotton candy, drive out to wal*mart at three in morning because i can\'t sleep and piggy-back me down the aisles. :P or tell me i\'m gorgeous even at four in the morning or ten at night. waiting up for me till some obscene hour for me to come home when i go out clubbing with the girls and nurse my sorry drunk ass. or encourage me to try out new things, believe in my dreams, and do jello shots with me tubing down some hide-away river with snakes and still find it fun(ny) as i yell at you because my face contorts in so many directions.   i sit here and so many memories flood through my mind and i want to list them all, put them in print and yell on top of my lungs bursting with pride. :) but there are so many things like the coffee machine, or jessica\'s little red box, or running through thrift stores looking for a size fourteen gown.   but even as i do that, i remember our latest conversation and how it sorta reverts back to i wish you were here as you tell me about the mall you went to today, and the super duper cool ceiling thingamagic they had that i had to come see or the evil looks you get if you open up your mouth to say one more word of me. heh. :P   i miss you ryan spicer. and i hate it.   mostly because it sucks balls and i could be with you but i\'m just being a dork.   i don\'t know how being on the medication again is going to be like, i don;t know what all will change. i\';m sure you\'ve noticed the change already. i\'m distant and we don\'t talk about it.   you worry because you don\'t know if the way i feel has changed, and you don;t know if i will return too different. and to be honest, i don\'t know either.   admittedly, considering the circumstances, it\'s crossed my mind many, many times to let things go a while...even with us back in pensacola. and i don\'t know how that\'s going to be, i suppose i wasn\'t sure most of all if you\'d find someone back in jersey or if you\'d fall in love again. i have your security but mine isn\'t strong enough to understand it.   sometimes i think i know exactly what i want, like tonight. yet there are others, like the last couple of weeks where i know exactly what i don\'t want. i don\'t want to jump into this again too soon, i don\'t want to do anything stupid that might jeopardize the movie. i don\'t want to act too fast, speak too soon, or have to much faith. and i know you understand that, and i know you know what i mean, and i know it doesn\'t make it any easier.   i guess tonight, i just wanted you to know that i miss you. and that you are a part of my life, whether i like it or not. and six months is a long time to forget.   ...you ask me to try again.   move in.   love you.   try us.   again.   i smile.   i giggle.  i weep.   i fear.   distance works miracles sometimes.   i\'m glad you\'re still around.  '),(22664,664485,'12,July,2003','2003-07-12 00:00:00','   heh, you know you\'re in trouble when your mom stomps up into your room three floors above and yells out your full name. and i mean your full. name.   so it\'s around the bloody crack of dawn (i lied, it was 9) and me mom, apparently sending beeps and rings for the last half an hour, charges upstairs into my little area and wakes up the dead. she does it almost subtlely though, just sorta saying seventeen million things at once in which to my sleep deprived mind is rather comical.   this pagan needs to get her arse off the bed and go to church in 15 minutes.   right.   :P  and so once again, i find myself throwing things together, desperately trying to look somewhat decent and angelic cursing the hellfuck out of syrensex because of her countless seducing techniques in getting me somewhat happy last night. of course despite my lame attempts that well, pretty honestly, went unheard.   church was, well, church.   now, it isn\'t that i don\'t believe in god or anything, i\'m just...i don\'t know. losing it? i think it\'s a phase, as i\'m sure mickey will whip up an email about god\'s wrath and all that good stuff but yeah, onto other things.   my dad is very quirky. it\'s hard to explain...he\'s like this artistic, nice, gentle, shy person that rottenly spoils the people around him. and i speak from experience. :P  so he says we have to do lunch together and that there\'s some japanese place that someone or another said was good and so he had to take his wife and daughter there plus there\'s a flea market nearby, and flea markets are always good because i\'m strange like that.   so here i am, nineteen (okay, eighteen) years old and all in the most godawful uncomfortable pair of shoes in my own little world a few steps behind my mom and dad strolling along hand in hand and jabbering on. and my dad stops by the little shops and shows my mom things and tells me about the different things of his week and my momplays along and humors him and i feel all warm and cozy in our little family except something\'s missing and my brother\'s not here.   of course i could make a sarcastic(jealous) comment about how he\'s backpacking through europe but i\'ll say this instead, dang i miss you koko. pah.   hm. i did however manage to find soem pretty cute ethiopian looking trinklets and what nots to bring back. there was a massive conglomeration(sp?) of coca-cola collecting freaks that paid 16 ringgit for a can of coke from america that you can prolly get for a quarter at wal*mart. madness.   haggling is fun though.   and so is hanging out with the old folks once in a while.   heh, after all, i can\'t blow my slut-bitch-snob-asshole cover now can i?  :)   urlLink  \" \"oh well, with age comes senility.\"  '),(22665,664485,'12,July,2003','2003-07-12 00:00:00','   the world waits in single silent eclipse sometimes...and there are so many things you want to accomplish in life. some rather comical desires better left to your inner self to explore, the others, perhaps more realistic, more real...more \'normal\'. like finding a good job, like graduating with good great fantastic grades and climbing a mountain, or scaling the sky, jumping off some buildings or traversing the great beyond. you see, somewhere between all the pressures of life, all the talk, all the drama and all the glitz, the lines get blurred and you cannot help but ask, is faith just an ignorant selfish part of you that presses on, believes, to get the things you want? then if so, than the things that you have set out to do has been accomplished hasn\'t it?  you sit, in some cubicle of surrounding cubicles in a multi-storey building, filled with drones, office workers, people just like you, little pawns in the game of a metropolitan city keeping together so the building stands. and sometimes, you sit and wonder, at your cluttered desk tacked with the many pictures of loved ones, of happy joyous occasions, of little things from different people; quotes, inspirations, a plant, and you wonder.   you\'re eighty-seven today.    urlLink  \"I can’t seem to indentify it with myself anymore, even though it was once such a defining part of me.\"  '),(22666,664485,'11,July,2003','2003-07-11 00:00:00','   heh.  for the blind: you have bigger fonts now.   (so bloody big-lah. i don\'t like, like the many hundreds and thousands of words are screaming at you? )   for the needing to complain: you have comments.   for the bored: there\'s much of bullshit to read.   for the horny: welcome to the club.   :)  '),(22667,664485,'10,July,2003','2003-07-10 00:00:00','   urlLink    urlLink    urlLink    urlLink    urlLink    urlLink      she rushes home, somewhat unsure if she should go..it\'s been an up and down tiring kinda day with the medicine taking affect, the drowsiness, and the irritating dizziness that pounds into her head every time she moves faster than a snail\'s pace.   ...but yet she figures, since the plans are already made and the flight\'s already postponed; what much harm can she do to herself really right?  besides other than the alluring sexy, beautiful, people she\'s meeting up with tonight there are also the long found friends that she knows she\'ll have a good time with...  and so she takes a long soothing piercingly hot shower, scrubs herself off grime and other disgusting things she must have picked up in this awful humidity and terribly polluted city, but then again, the city has her heart at it\'s beck and call and there is something so enticing about it that she cannot seem to run away from.   she walks around, she\'s having an almost shitty kinda day. the kind where she doesn\'t feel like looking in the mirror much and the kind where she doesn\'t want to pick out something to wear because nothing looks right. she feels the ten extra ppunds she\'s put on, it\'s in her guts, in her flabby arms, in her cellulite thighs and her vastly expanding tummy. no, she isn\'t really having a low day, just a day where she just doesn\'t feel too nice about herself that\'s all.  she wishes she were taller, thinner, prettier eyes, nicer smile, whiter teeth, bigger boobs, smaller ass and so much more that she could list on forever. but she picks up her self-depricating ass and grabs a dress, puts on some make-up and goes anyway.   she goes to pick up some cigarettes at the mall but they are out of her favorite kind. she stops to look at the happy faces around her and stops for a while, as if to just say, chant, she\'s going to have a good night.   takes a cab, her first cab ride since she\'s been back and for some strange reason she is pissingly nervous. scared shitless almost.  the media hasn\'t exactly been too modest(?) with their coverage of crime. rapes, murders, robbery, charred remains, you name it, they\'ve prolly reported it in the last four weeks she\'s been here. evil lurks out there and some flash of wanting to believe in humanity and being betrayed goes through her mind as clips from last night\'s late nights with astro patch adam\'s screening flickers.   ba da boop. ba da boop.   but then she walks down to smiling faces with open arms to greet her and damn, how she\'s missed them so. and they all sit around with their coffee and theatrical talk and catch up. and comment about changes. and marvel at beauty and innocence, but most of all friendship.   bigger two who is big shot director, producer, singer, super multi-talented amazing friend in which whom she loves to death. instructor, teacher, flamboyant and talented, honest, and upstraight young dude. beautiful kittie mom, who\'s trying to sober up, smiling, interesting, bloody good memory. kopitiam dress wardrobe designer person that she doesn\'t quite recognize in an instant but remembers nonetheless. the other three are missing, called away, the empress had a collision. hip mama and singing shy boy joins them later though at karaoke where they belt, scream, yell, croak(her) their way through the night.   she\'s dumbstruck by the band that\'s playing at the wine place tonight. they\'re fucking bloody awesome as she kicks herself for her cynicism. perhaps the local music scene hasn\'t quite died yet as she kicks back, smiles, draws in a breath and exchanges crazy travelling stories with humpty.   10:30 comes, and 10:30 goes. should she ring? is it okay? can she just gatecrash?  she walks dowm to meet them anyway, she is very much amazed at their kindness and open attitudes. like sure, although i have no idea who you are, what you are and all that, lets meet up anyway! lets go clubbing! lets play paintball!  and there is the girl, with whom she had a rather self-obsessed pathetic conversation last night with, looking kinda rather zonked out. sad almost. is something up? is she okay? then there is the other guy, the one she exchanges e-mails with, bright and smiling as ever over his plate of nasi lemak and cheeky grins. the boy that\'s ironed his special shirt, that smiles this warm inviting smile, and you smile back at these people and think, these are some pretty great people here.   they promise to meet up later and half an hour of frantic non-stop calling in between seventeen different cellphones results in a rather interesting clash of people.   the bloggers. the new friends. the rather interesting variety in which they sorta kinda understand the addiction to releasing yourself online. the blogging, the bullshitting, the random silly notes posted at any hour, at any place, of any topic. the people she\'s sorta come to know and am glad she knows.   other american girl and her definitely surprising buddy. the ex-girlfriend of the ex-boyfriend. she swears someone could have taken this out of a script. lol, first she works out with the other ex-boyfriend\'s ex and now clubbing with another ex-girlfriend of another ex-boyfriend? hm. interesting nonetheleast, doesn\'t quite puts a twist in her knickers but rather strange nonetheless.   new york in glasses and this red top that screams hot mama embraces her. it\'s funny how you sometimes meet people you know instantly you can be friends with.   along trots dancer, after rehersals and you run down telawi street in your stillettos screaming because you\'ve missed him so much. the one that brought over ice-cream after school, the one that shared plenty of nights by the pool smoking and crying on each other\'s shoulders, the one that got you onto that terribly addictive dancing game and the one which you\'ve had cup after cup of teh o ais limau with. yeah, the same dude that decided he\'ll get hitched in australia last year only to call it off, the one that sends you strange pictures of the french pianist and tells you explicit details of his sex life. yup, that would be him.   or the time you were co-captains of the cheerleading team and disagreed about anything and everything only to agree on everything else after. the only person you would drive out six million miles out of the way to shoot some tv program in malay with, the person you\'ve held hands with in and out of the closet and through his many, many, many countless obsessions with boyfriend of the week. :P  one year is a long ass time to not see someone isn\'t it?  he worries about his rather \'rehersal-ly\' attire. his gym bag and tap shoes for the competition this weekend, as humpty and dancer run off to put things in the car parked in some garage seven thousand miles away because six thousand and seventy two people decided to go out tonight.   and the musical theater people want to go salsa.   the bloggers want to go clubbing.  the friends don\'t quite know what they want to do.   and so you look like a retard running in between these three cliques but it is the highlight of your week! so happy, so happy, so damn bloody fantastically happy and you\'re not even drunk.   the salsa bar is closed and after bouncing around making an absolute fool of yourself at the club with the bloggers (who are much better dancers btw) you move on to join the mt people down at the karaoke place. you sorta smirk because where there is an mt meet, there is singing. :) and bloody good singing at that.   the night ends with having the streets again to the three of you, and it\'s the very same three of you in a car again minus the other two who are missing because of a certain indian musical instrument and the evil exams. you hug and kiss and sing and smile about and you wonder just what it\'s going to be like when you get back to pensacola and there are no more wednesday\'s or karaoke or random blogger meets.   oh well, i\'m sure there are plenty and plenty of things to keep you occupied like waffle house and wal*mart, barker and puck, blue-shirt and sexy pants, old man and madonna, stupid things and odd hours though there is one thing for sure and it\'s that you\'re definitely going to miss nights like this.   hm.     urlLink  \"I am awake, but soft and blurry around the edges, of my vision and of my voice.\"     urlLink    urlLink    urlLink    urlLink    urlLink    urlLink  '),(22668,664485,'10,July,2003','2003-07-10 00:00:00','   urlLink      how is it possible to not to be absolutely yonkingly in love with people that can make goofing around in an elevator fun?  :)  '),(22669,664485,'10,July,2003','2003-07-10 00:00:00','   would you trace the line that runs down my body and let your fingers love me in? would you kiss my forehead and sing me to sleep, read to me, tell me about your day? would you let me grip you tight and gyrate deeper in you as i come? would you let me see your heart, your smile, each night, every morning, whenever i open my eyes...it\'s hard to explain, this empty bed, even colder heart as you try and make our worlds seem so much closer than it actually bloody is.   i don\'t know honey, now, didn\'t i tell you thinking was evil?    urlLink  \"Another short night in a lonely bed, leading to another long day in a lonely head.\"  '),(22670,664485,'09,July,2003','2003-07-09 00:00:00','   urlLink     so i got my prescription filled today; 20 mg daily so i didn\'t do too bad i suppose. he hates the idea, i\'m just really undecided.   whatever.   babe, this one\'s for you and i miss you much.   :(  hugs   urlLink \"better now.\"  '),(22671,664485,'09,July,2003','2003-07-09 00:00:00','   i see the new head doctor at noon tomorrow. some woman named dr.aw who\'s going to regulate my medication.   this should be interesting.   expect much rubbish, mood swings, or insane hyperness tomorrow.   hm.    urlLink  \"But there was nothing I could change, because I am young, I am naive, I don\'t have the means, I don\'t know how. As I got back, the night changed into day, and looking at the world as it was, with all its flaws, and thinking, I\'d vow to make a change somehow, this world is where I\'ll be living, and I know it\'ll never be perfect, but I\'d be damned if I didnt do anything to make it a better place.\"  '),(22672,664485,'08,July,2003','2003-07-08 00:00:00','   urlLink    urlLink    urlLink    urlLink    urlLink    urlLink      urlLink    urlLink    urlLink    urlLink    urlLink    urlLink     so technically we started out taking photos of me hair....but seeing as the lighting in my room is pissingly shitty: we just took random pictures anyway.   a 230 ringgit hairjob + sucker = mushroom head.  lol, when i go back in next week to have the streaks done (something about needing the dye to set in first because of color mixture thingamagic la di da...in other words, come spend more money!!!)...i think i might have them crop it shorter into a pixie cut. hm?   i don\'t know.  vanity, thy name is woman!   but in all bitchiness and glory, it was done for a good cause.  (a percentage of money is pledged to the malaysia hiv/aids foundation with  of every bottle of hair dye they use...or soemthing like that anyway)    he double toned it once by mixing these two browns together then bitched about how i had previously had it dyed black and something about the dye being hard to set, then of course he said, hey, what about this and that and that and this?  i\'d totally forgot about the wall to wall being in bright yellow, the psychedellic blinding selection of nail enamels, the insane gossipy six thousand and seven shampoo girls though the soothing calmness of flowing streams from the waterfall and the sexy moans of diana krall did make things better though.  hehe, i sat for three and a half hours straight, me, sitting still, mmhmm.    of course, there was a certain white-suited slick chick that made everything almost a breeze simply by being there. :)  we came home, showered, putzed around for a little then went off to this extremely adorable place in tmn tun that had a fusion of east-west menu, a nice little place yet very cosy too.   we managed to sneak in a picture before the camera(battery) died..mostly my fault from all the 772 pictures we took in the room.   but yeah, pass or not?    urlLink  \"Sweet!!! I\'m being stalked. =P\"  '),(22673,664485,'08,July,2003','2003-07-08 00:00:00','   so it\'s official now, i am a hog-boarder.   i am also a ditz, a klutz, and a very stupid ass.   in other news though,  i  am  bored.  (if you can\'t tell already)  i also have an appointment with the hairstylist in about an hour and i have yet to get ready simply because i am lazy and am enjoying random messages from some dumbfuck in ireland who thinks i\'m hot. or not.   see meesh, you don\'t have to be knackered to  be stupid  talk cock; i\'m living proof.   :P   urlLink  \"sometimes I wish I had a cock just so I could grab it and tell people to suck it.\"  '),(22674,664485,'08,July,2003','2003-07-08 00:00:00','   i wrote that almost a year ago and think about how much has changed yet realize how vaguely indistinct everything still remains. it tugs at some heartstring nonetheless, and i feel again, tears line down my cheeks only that there is an uncertainty if these are bad or good tears. perhaps both really.   i still see the same child, the heart--broken, confused, lonely, over-worked, tired girl. the same girl that looks back at me when i stare into a mirror too long sometimes, the reflection that\'s insecure and longing so much to find her place in this world. that reflection, that little girl, that stranger...will always be me.   yet i have found my niche too. i have found the person i can be, the same person that i have always been. i have found that though confidence isn\'t always around, that pride was an overrated thing anyway, for i learned so much more scraping from the bottom.   i take nineteen credits next term, that\'s back-to-back classes, the fall production, set stuff, the movie shoot and trying to deal with all my crap at the same time. a year ago i would have said, sure. uhuh. yeah. a year ago, i would have taken all of this on because it was expected of me, not because i wanted to. this year though, this year i take all of this on because i want it.   i am on the honor\'s list because i can do it. because i am capable and i am alright. i am in the fall production because i know i work hard at it, because i will sit and build the set, paint the shit, memorize my lines and live in the theater if i have to. i\'m going to do the movie not just because i want to but because i believe in it. experimental or not, sundance or not, i have auditioned for it and i have gotten the part. it is a commitment and i will not fall shy of it.   i take care of the \'other things\' because at nineteen i know i can no longer hide from my demons nor my dragons. fess up, eat up, shit up. time to get healed, time to be sane again, time to be a good girl.   and so tonight when my dad sits me down to talk about things i don\'tfidget quite as much as i used to. i don\'t quite have the tough barrier i put up all the time, in fact, it was kinda nice letting him in. telling him i wasn\'t too sure but that i would try. asking him if it was okay if i didn\'t do so well. questioning if he\'d hate me if i wasn\'t perfect.   ironic things like the transition from aneroxia to bullimia. the lengthening of things to come and prophecies i could have told you a long time ago.   the angels? the angels are doing fine.   i see them on wednesday. it\'s been a full-fledged year now, save the random emails, cards and letters...i\'ve missed them. the sars scare was a deadly threat, immune systems down and everything, they were risks i was willing to take but knew i couldn\'t. responsibility and all that, and like they tell you constantly, you can\'t help if you need help.   i think about how i missed mun joon\'s funeral, i think about how \'mc donald\'s\' boy is doing now. i think about chemo and all the things i have to say to them, all the things i know deep down in my heart is right now. i think and think and think, then i ask myself what the fuck i am doing, only to run back to the store and get more balloons to practice twisting them again.   i\'ve fought my dragon and i\'ll fight it still.   dimples aren\'t just proof of identity you know.   :P   i think about him sometimes. not too often now, i think about how last summer he was all i ever thought about. all i ever dreamt of, all i ever wished for, all i ever breathed about.   ....and then i think about the other people that have come into my life, no matter how briefly.   i think about barker, and the crazy insane way he laughs. the smile i genuinely look forward to seeing each night, the arms i want to hold me at night. i think about this summer and how long-distance is a bitch. i think about having him around, then not, then the in between. i think about how oddly ironic that we\'re closer now then when i left. or that he\'s so much more real, that i am so much more trusting.   i think about love, and how i am still not quite sure what it means. or what it represents. but that i am sure, beneath everything, beneath how my life has shaped up and above all things, that love was somewhere entwined into it all.  i thank god for friends and family. for trust and patience. for hope and adversary.   life isn\'t always so easy, but life isn\'t too bad either.   and so tonight, even if i am still rather not too good, and even if i am a little cranky and tired, i am happy i\'ve made it but most of all, i am happy i have tomorrow to look forward to. and the day after. and the day after, after, after.  ----    [Thu Jun 27,2002 06:33:28 AM | + racheal | edit ] Paradox.  I wish to say I didn’t care, but today when I came home alone, I wept again to myself, swallowed pride and sobbed into my soft old pillow…comforting. I remembered again promises for tomorrow and things we had said, I remember again conversations that played through my mind again and again. And as music vibrated and put me to sleep, I dreamt of autumn and the changing awesome colors and falling leaves and how I was looking forward to seasons again.   I looked at my blistered feet and stretched my body, looping and twisting it around, contorting it so that it remained flexible…feeble. Put aside my jazz shoes with bare feet on the ground I trained to loud annoying beats and rhythms and twirled and twirled, lost in a trance of another world I let myself be free. But soon it became work again as I sharpened my moves and became anal, a perfectionist. drenched in sweat and a cloud of exhaustion, just me, a sponsored adidas sports bra and a pair of shorts in a too cold room and a ruined pedicure. I really liked nike anyway.   I wondered if my no show of emotion came because I am ready to leave and move on and change. Because there’s only so much I can take and something tells me my bubble has burst already. I am sick and tired of this place and the people and all these emotions and excess baggage. The next few weeks promises to shake it all off and clean that slate again…and I am thankful. But yet tonight when I listened to him speak and heard his every emotion I walked away, I refused to let him see me cry nor comfort me, I drove in silence and bit back the tears, keeping it all in me for I knew I would have to sit for dinner with my parents and I have mastered the art of protecting them. And so I choked down dinner and played with my food and commented on the lovely wine and the great food and talked about my entirely fictional ‘great’ day and I listened to them tell me theirs and even stayed to watch the beginning of lord of the rings for the six hundred and seventy second time with my dad.   Then I quietly and subtly excused myself and collapsed into my room and with it came the torrents of racking sobs quietly hidden by sade’s loud moans and drones. I like sade. I let and felt ever tear stream down my face and cried for everything that had fallen apart. For new zealand. For friendship lost. For good times and memories slowly slipping away. For broken promises. For hurting words. For losing the people you love and for being robbed of the ones that were meant to be forever. For angels and heaven. I cried for my imperfection and missed the person I used to be yet a part of me never wants to be her again.   I have lost certain dreams and am afraid of building new ones and that upsets me…because I used to be confident and brash and care free. I used to not have any hang ups and used to know what I was made of and what I wanted. I used to know myself but lately, I am a stranger to me.   I go through preset routines like smoothies and coffee in the mornings and cleansing, scrubbing, masking, toning, curing, moisturizing and protecting at night. like using chopsticks with my left and a eating with my right, like picking out the peanuts in a snicker bar and anally portioning my food and it comforts me but I need a change of environment and to get rid of routines.   I need to stop going in to read what he has written and I need to stop caring. I need to let go and let be. I need to move on.   …and this is a hard way to it but hey, cold turkey might bring good.  I am hoping that this is for the best not worst and I am hoping I come back changed for the better.   Perhaps I might, perhaps I won’t but no one can predict the future but neither should you stop believing in it. I am not perfect, I have many many flaws and many many wounds to heal but someday, someday, it will all be okay and things will be fine…and someday, I will understand.   Someday I will understand why I have to do what I have to do and someday I will know that it was right.    first thoughts after a nap.   lord have mercy.     urlLink  \"I don\'t know what will happen, and the Unknown is terrifying. I would rather sit on my little boat for as long as possible, rather than risk exploring the land before me, filled with possibilities that could lose me everything I have gained.  But if I don\'t move on and leave my boat, my boat would eventually rot beneath me and sink me into the sea. I would be left with nothing anyway.\"  '),(22675,664485,'07,July,2003','2003-07-07 00:00:00','   :)    urlLink grace  is absolutely wonderfully fanfuckingtastically brilliant!   i have   urlLink  links    now! i have   urlLink  links  !   weeeeeeeeeeeeeeeeeeeeeeeeeeeeee.  lol, see, you don\'t exactly need your codes stuck in pensacola when you\'re in malaysia, you just need gracie! weeee!  :)  i\'m happy. can you tell?   this also means i do not need to painstakingly go through six million links to read up on the bloggers.  everybody say hell bloody yeah!  :)  go visit! go visit!  yay!!!   \"graceshu: *licks u* graceshu: :D NYAM graceshu: spread ur legs and let my holiness penetrate your being\"  '),(22676,664485,'07,July,2003','2003-07-07 00:00:00','   so i absobloodylutely forgot that i had to meet up with the director person last night. of course i remember this at eight in the morning the next day. right. holy shit, i am in trouble.   now why didn\'t he just ring to remind me?  because the same brilliant ass that\'s sitting here having a slice of cold pizza for breakfast switched off her phone yesterday. lol, i swear, everytime i decide to have some peace and quiet something always crops up. conspiracy!!  hm. and so now i\'m hammering off an apology on e-mail which is prolly the whimsical way out but oh well.   time for a cigarrette break.   btw. how is this possible that i am sitting in malaysian heat trying to not kill myself by being on stupid roads and me brother is on the way for a holiday in italy?   life\'s a bitch.   .....but i love it!!!!   :)     \"imasouthernqt2: well the night that you get back i promise waffle house and walmart  imasouthernqt2: OUR waffle house!!!!  imasouthernqt2: with the spanish song and michigan\"   wa-hey!  '),(22677,664485,'06,July,2003','2003-07-06 00:00:00','   i think he forgets sometimes that i am no wonderwoman. he wants me to be so many things at once yet i cannot question where he stands for soemwhere beneath this lies tranquility and faith. in knowing that the person that helped make you knows you.   perfect.  '),(22678,664485,'06,July,2003','2003-07-06 00:00:00','   urlLink    urlLink    urlLink    urlLink    urlLink    urlLink      assam(one s or two?) laksa, a good conversation, shopping, a massage and strange people never fail to amuse me.   :)  p/s and fuck yeah it was good noodles!   urlLink  \"We chatted, we socialized, we mixed, we shared, we laughed and at the end of it all, we became the friends that we thought we could be.\"  '),(22679,664485,'06,July,2003','2003-07-06 00:00:00','   it\'s funny how a couple of comments sparked on by jealousy laced with sarcasm can awfully change things around. two hours ago i was ending up my post preparing to throw together some things and share a drink with humpty but i got side-tracked. i saw it flash across my cell screen and knew instantly that this couldn\'t be good, but being the dumb narsistic word-for-people-that-like-to-hurt-themselves person i am; picked up anyway.   and with it came a surge of adrenaline, a blast of foul shitty hurtful accusations, a string of alphabets and letters, sentences and words all aimed quite directly at my being, at my heart, at who i was as a person...as who i was, who i am, and who i will be.   i took it all quite well really, rather surreally in some sense though i am quite proud. i listened, as always, but was too tired to fight back. to try and distill the bullshit, to try and deal with it and so i let him scream till all hell broke loose and he was done. scream, yell, curse, cry till he had nothing left to say.   i cradled my phone in my hands, not quite knowing what to do. not quite understanding all the feelings he had put in me, the hurt, the rage, the pain, the understanding, the calm, the everything. i didn\'t know what the proper etiquette was to be, should i be hurt that my ex-boyfriend, someone i hadn\'t talked to much less seen in the last three years suddenly felt the compelling need to tell me that i was bitch. that i had treated him wrong. that i had left without saying a word, that i never understood, that i had broken promises and had fallen. really, how /do/ you expect me to react? realize i love you when you\'re yelling? right.   and the odd thing is i sit here with marvelous three blasting through the speakers crying for you. crying because i know you\'re hurting, crying because there\'s not a single bloody thing i can do, crying because i know that tomorrow when you wake up and realize what you\'ve done you will ring again, and apologize profusely...then go through the whole depressed hate myself thing. i wish i could hold you and tell you that it\'s just life. that in your drunken stupor, i understood that you said things you hadn\'t meant to say. i understand that ignorance is bliss but that i cannot tell you about how you\'re not the one. that what we had was just what we had. jesus, you\'re ten years older than me and i\'ve been through hell to get this far, i\'ve been through hell injust moving on.   i\'d tell you about my nights but i figure why add salt to the wound? i\'d tell you about how much i hurt a long time ago, about sobbing and feeling like a miserable piece of shit. i\'d tell you all my stories if i knew that it would help you but i cannot.   sometimes, i lie awake in the middle of the night wondering what the hell happened. how things ended up so complicating when things had gone on so well. we were pretty good friends even when things were over, untill of course, when i came back. and that night. and the proposal.   ironically enough, the affair with you split up what i thought a long while ago was it. but it wasn\'t, and he painfully drilled it in each and every time as he reminded me of the many things you did to me. i used to be so bitter, now i am just tired.   tired of the drama, the lies, the bullshit.   it all seems so juvenile, so trite and drawn-out....  you ask me rather strongly, am i happy with him? is he the one? can i sleep at night?   and to be honest, i don\'t know. if anything, i am still a stupid teenager unsure of the world, looking for my space in the world, my little corner.   i am not ready for marriage and kids, not ready for long, long term and retirement plans. we take it a day at a time but for now, for now it seems right. for now it fits.   for now, being with someone that loves me, respects me and needs me, completes me.   why am i sorry that i\'m okay?      urlLink  \"There is a moral in there for us all somewhere, I’m sure.\"  '),(22680,664485,'05,July,2003','2003-07-05 00:00:00','   there are so many things i want to blog about but like all things, i seem to be a little retarded in trying to find the words or if anything, just being in the right frame of mind at the moment.   perhaps i\'ll just sit here, cross-legged buddha like style and pray that a sense of peace will wash over me as i go into a chanting mantra that spraypainted green purple dishwashers are safe. or not.   you get the gist anyway, i\'m rather incoherent at the moment. blame the caffiene, blame the insomnia, blame whatever the hell you want to blame but the little crazy old woman down my street that does these tai-chi kicking flying things every morning at a quarter past six in her front yard.  blame the little gwei-loh kids down my street that seem to be constantly craving durian as the toxic waves makes its way down over and a morning smoke turns into a nightmare. or just blame my pagan lazy ass that couldn\'t get up this morning for the 10:30 service.   i haven\'t been out of the house at all today, minus the brief six or seven minutes i got stuck talking to a neighbor that wanted to know if she could steal some herbs off our plants or something to that extent anyway.   i missed the fourth of july celebrations but nonetheless, we had a party of our own at home last night: a day late but a day right technically. i can\'t exactly say that  we  i got absolutely knackered but a little happy might be the right term. no fancy fireworks though and much to mickey\'s credit, definitely nothing like the fantastic fantasmic. pah. :(  ....somehow, i don\'t think a massive array of candles on a cookie cake decorated in red, white , and blue splashed with \'happy fourth!\' quite counts either...  hmm...it\'s half past eleven and i am absobloodylutely bored. nevermind that i\'ve slept most of me day off and cannot seem to figure out the template codes for the current page(apparently, the permalink is down as it has painfully been for the last few months..) and can\'t access the original codes because they\'re stuck in the hardrive somewhere 100000000 miles away in  pensacola  alabama. hm.   bagaimana? (lol, how\'s that for a little malay eh?)  speaking in which, kl bloggers...i\'ve told everyone and anyone about the meet yesterday and strangely enough, i think i\'m more excited now. (!!!) lol, i flipped through the links and what-nots the main page, read through the group discussions and think, HELL BLOODY YEAH! *grins*   and btw, about all the incriminating pictures uploaded: harruummpphh. it wasn\'t me...really.   ooh, humpty-dumpty just rang. mamak calls.   fooooooooddd!!!  :P     urlLink  \"I cuss. Fairly often.\"  '),(22681,664485,'05,July,2003','2003-07-05 00:00:00','   urlLink    urlLink    urlLink      i\'m rather a little intimidated. i wonder what they\'ll think, who they are, if they\'re half as excited, if how bloody nervous i am will show....but then the caffiene takes effect and i am off speaking at seventy million miles an hour; coherence and sense not quite making it and it\'s rather a little odd. i laugh a lot to hide, giggle really...and soon i am just rambling. ah well.   :P  she is quite as amazing as i\'ve expected her to be. to the other buddy that talks to me about newts and ipecac, it was absolutely brilliant meeting you today! i am every bit as honored and dumbfounded...it\'s strange how this whole meet people on the internet thing works...ironic really if you think about it. but yeah.   i didn\'t quite keep track about the number of people that went...enough to fill up three tables and a side though. a cordoned off area, four bold square artpieces splashed with black swipes on the wall directly mocking me. distracting but nonetheless appealing....i find such fascination in the bloody weirdest things sometimes?  today i felt everything and nothing.   i felt what being different was like and i\'m not quite sure if i enjoyed it or was upset by it. i couldn\'t help but want to hide, layer, shit around and not bring things up. it\'s hard being here and having to play the seperate facades. sure, i\'m bold and bitchy, stupid and loud here but despite all the revealing, rather naughty pictures as the current layout now, i felt stifled. there was something uneasy about the room, perhaps it was just me, my disease, my demons mocking, screaming, protesting....perhaps it was just being up at an ungodly hour, perhaps it was just insecurity.   things are a lot more conservative here, i was embarassed. shy really of this place and i hated that feeling. that i wasn\'t too okay with being honest, with being who i was. i wanted to hide my american self and that life, i wanted to say so many other things i could not verbalize.   i happen to think that i\'m also rather a little melancholic after a long drive home from the airport. after a week, seven whole bloody days, i am a little displaced and unbalanced to see them go. it\'s been a good week, and with all that has happened in the last couple of days....i\'m doing pretty well. or so i\'d like to think anyway.   i don\'t know what it is. here we go again where i draw back to the part of trying to explain things. i don\'t think i quite fit in but i was definitely welcomed. i suppose we choose our own destinies and feeling the way i feel right now just oppresses all proper thoughts of sanity.   in any case though, i would do one of these things again. i enjoyed it so if you\'re reading this and you happened to be there today....thank you for sharing my saturday morning with me. it was a pleasure and definitely eye-opening.   terima kasih.    urlLink  \"I whispered to him that if we don\'t have time for our friends, then our friends will not have time for us. \"     p/s my group photo didn\'t quite turn out?  :( so much for mr.dome and point-shoot cameras! anyone uploading theirs for free?  '),(22682,664485,'03,July,2003','2003-07-03 00:00:00','   he laughs and giggles on the line and sometimes, there\'s ackward silences. moments in which we\'re both sorta like uhm, then comes the funny laugh. the look at us we\'re being retarded dorks ones.   we talk about summer and how things are going in our respective worlds. i tell him how it\'s strange being back here....and how insanely different it is. he tells me about jersey and the boardwalk, his friends, the people, his life and i cannot help but smile and hold him close. it\'s funny how little things can make big moments.   he says i\'ve been thinking.   and instantly i know, here we bloody go again.   i want to stop him before he utters those words, i want to stop him before he adds more swirls to the already overlapping confusing spin.   and so i say, uhm.   and i laugh again, we tlak about other things.   escaped another day.   our roles seem to have switched, it\'s a little bit disturbing and comforting at the same time.   ---  i am a needy person. i have a million weaknesses, a couple of hundred thousand flaws. yet i find myself striving to be independent the last couple of days. not wanting to need anybody and it\'s strange.  but anyway, jokingly (seriously) i said, lame attempts at trying to be non-chalant about it...as if it wasn\'t an issue to me at all; what would you do if i came back 20 pounds heavier and ugly?   \"kiss you.\"    lol, and he makes it a point to mention that he doesn\'t need to think about it.   i am relieved, that he\'s still the same person and that i did not concoct him into life, but at the same time, i am rather cynical. i suppose i\'ve done that my whole life; been a damn well bloody difficult lover.   but wasn\'t that so cute, adorable and absolutely corny at the same time?   :)  what i meant to say though was that i\'m in a rather strange twist at the moment....i feel rather surreal. odd.   i sat down the other day in the waiting room pissingly afraid thinking, i really really want to do this. i really want to live.   every six months we go through this, and every six months i reaffirm my need to press on in this world, to live on, to renew dreams. and every six months i reach some stage where i worry if i might not make it at all.  challenges are posted all around you, the right person, the right being, \'the one\' would make it through. i drift away and keep my distance...i\'ve been doing that a lot the last couple of weeks. it\'s as if all the crap i learned in pensacola has disappeared into another realm and i\'m once again stuck in this position where it is not okay to just be.   i always seem to find it much easier to just be whatever in florida than it is back here in kuala lumpur. maybe it\'s the stigma, or the environment. maybe it\'s just my state of mind. in any case, it\'s different.   denouncing the thought of a relationship or anything of that commitment when i am here. i suppose i am more trusting over there, more open, a whole lot nicer. i am very cold here, hard bloody bitch though in all honesty, that\'s prolly the stark opposite.   as much as i am more insecure in florida, i think i am a whole lot stronger.   over here, i am a lot harder on myself....a lot more abusive and it scares me. i broke down last night because i was bleeding again and i was wide awake having a panic attack. afraid of dying but even more terrified of living.   there\'s a huge stigma attached to mental health care here. it\'s a big conspiracy, a bloody taboo and so things are kinda hard and there is an enormous pressure to just keep things right beneath the surface. to hide.  i have an appointment next week with the doctor and i know i need it yet it still doesn\'t make anything easier. i know i have to take my meds but that doesn\'t make it any easier too.   i have been breaking down privately a lot in the last couple of days. it\'s part of the depression and i understand that but it just gets so hard when the demons in your head are screaming and the dragons are pounding. when you\'re crying so hard you\'re sobbing, and sobbing so hard it\'s almost impossible to breath but it\'s all done so quietly, so masterfully it\'s scary. when you know you\'re okay but you cannot seem to let yourself accept yourself for who you are even if you know that on some level you already do.   nights like this, when i\'m still wide awake even if it\'s four in morning and thinking....i wonder if things will get better. in some sense, in some hope, i know they do. i know things have improved and i know that things will continue to get better but sometimes, like all rides, i worry about the dips that it needs to take to get on top and that scares the petunias out of me.   someday i\'ll look back at this and say remember when?   and someday i\'ll say, good job.   in any case though, i\'ve been thinking too.  i don\'t know if that\'s good or bad.     pp/s HAPPY FOURTH OF JULY Y\'ALL!!!! :)  '),(22683,664485,'02,July,2003','2003-07-02 00:00:00','   to the girl in russia that deciphers my soul and periodically sends me random emails, thoughts sputtered out, crafted into perfect flowing sentences, verbalizing tones in rhythms of poetry that brushes out the kinks. kind words, supportive...opinionated nonetheless but honest....perhaps even brutally at times.   she stares, heavily mascara-ed eyelashes flick over deep set blue-green eyes, a reflection of the ocean, a sense of peace of the place i\'ve ventured off a while. return home, walk on water, traverse the seas; no one is here.   glamor case lies shining in the corner...collecting dust like the chinese like to say. i am a proverbial bullshit master; quoting, random data stored, strange inkling towards unknown territory.   dating, a little bit thrown into another world completely. pictures, real actual photographs taken on print...it\'s been a while hasn\'t it?  i miss your voice: does that sound stupid? i miss falling asleep with my ears pressed against your chest soothed by the rhythm of your heartbeat. i miss waking up to you and falling asleep spooned closely, protected, with caring, gentle arms around me. i miss listening about your day and making coffee in the morning. the odd way you try and differentiate between subject and matter, the strange way you categorize your demons.  it\'s different here, i know some obsessive intrusive part of me will desperately try and explain it but it\'s difficult. how do you put into words feelings that even you can\'t decipher?   his eyes light up....i can already see it happening:  could it be me he wonders?   you miss me?   yes you stupid, i miss you.  it\'s different when you\'re not around and seven hundred million miles away.  you owe me cotton candy.  --  three messages left on the machine. \"prospectives\" but i am not quite interested, i want to be held, yes, but meaningless sex and stupid relationships just doesn\'t seem to be my thing anymore...god, i\'m growing old i tell myself. soon i will be a crotchety old lady bitching and nit picking at every bloody single thing....she laughs beside me and says you already do that silly.   she\'s here.   picked her up and we\'re having a brilliant time. it\'s amazing, fun, loving and it\'s a bit of everything that i need right now in some odd sense. we can enjoy girl things together, things heterosexual relationships can never reach. of facials, and spas, and shopping. of baking extra chewy chocolate chip cookies and eating cookie dough while counting the calories.   but she is also the big fat ex. and the word ex in itself says everything....it\'s gone good so far though.    to the special girl from kl, because you are beautiful in so many ways i don\'t even think you know. i am excited, nervous quite a bit, but this might be good i tell myself, i am a little insecure. intimidated. not quite sure. but we\'re doing this nonetheless, i will go through with this damnit.   search for words to say. what? i don\'t know. do i say it\'s alright or do i say you stop that right now? do i become the world\'s largest hypocrite or do i just support it because that\'s what friends do?  one of the biggest things of coming back here is realizing how many of my friends have eating disorders. was i blind earlier or just i just turn away and pretend not to notice. i see all the signs, she\'s aneroxic and she\'s bloody disappearing in front of me; the most disgusting thing is knowing that some part of me wishes i had the willpower of her strength to shrink so small too. it\'s odd what perverse pleasure we find in control. in being thin.  i understand the circumstances in which the kids, the girls, my friends, the others go through to battle with this raging disease. it is never ending and society is so damn cruel but still, there are times i look beyond that and think, what wasted years and trauma. what influential scars this emotional battle will hurt.    i made a new friend today, she stood beside me in the 6:30 powerstep class. she said she watched me in bodybalance the other night...said i was very flexible and happy. at peace, i looked so serene at yoga. i laughed and giggled, shook hands and said hey.   she\'s australian but a lot like me in little ways, of course i still don\'t know much and three hours of working out with someone doesn\'t exactly constitute judgement...she\'s nice though and i like her. she gives me good vibes but in a good friend type of way.   --  we went to watch full throttle tonight. salivate, gush, wow-ed.   then we ended up at that nook by bangsar and had nasi lemak, satay, mee goreng, fried fish thing and teh tarik. it could have been heaven but heaven would be digesting.   i\'m tired, perhaps it\'s time to go to bed before i ramble somemore. it\'s been pretty okay though, the demons still crop up once in a while but things will be right evventually i suppose. he says i\'m close to blowing a hole in my esophogus, i tell him to kiss it. it doesn\'t just end, just goes away, just be done with it...i bloody wish too.  drag me out of bed, choke me down with chemicals, dance for me.  up. right. forward. shuffle.   bellkay rings.  '),(22684,664485,'01,July,2003','2003-07-01 00:00:00','   sometimes, i feel so lost in this city. alienated...as if i don\'t belong. they say you\'re so ang-moh, so damn gwei-loh, and i want to stick my tanned skin, brown eyes, black hair into their faces, i shove durians into my mouth and proove my cultural identity, i want to say; i can too-lah  ...but sometimes, they don\'t listen.   what am i? where am i? who am i?  so attracted by the sounds, the streets..it\'s people. so enamoured by the food, so amused with the culture. a stranger in a place i apparently belong. too different, too american...this place is so conservatively protective sometimes. this place is living hell.   i\'ve been here for about three and half weeks now, but still they look at me like i\'m a tourist. foreign. it doesn\'t help, nothing helps. i am alone, i am invisible, i am so obnoxiously intrusive.   ...but it\'s okay. i\'ll be alright i tell myself...it takes a while. i\'ll learn, i\'ll eventually understand, i\'ll get used to it. what-the-fuck-ever.   i walked through chinatown today and i must have had \'sucker\' plastered on my forehead. we went and bounced from shop to shop, bought shirts, trinklets, fruits, gold, bags, and six million kadoodle other things. a little pack of strangers in blistering malaysian heat afraid and i felt so out of place.   they\'re on the same boat and we talked about how we seemed so out of place sometimes. and i don\'t know what it is but i sit here writing now and i am so frustrated and contented at the same time. i long to go back and stay here forever.   i want to run to waffle house and bug the shit out of michigan. i want to curl in barker\'s arms and have him hold me to sleep tonight. i want to make fun of puck and take long drives with the old man. lie on the beach, visit mickey, go down town and bask in the emptiness of nothing. watch a hockey game and scream my lungs out....or soemthing.   but i enjoy this too. this constant flow of energy, of seven hundred things going on at once. of people and more people. of people that look like me, speak the same language but not at all. i love the smell of this place, the comfort this place reeks of. peter hoe\'s, bukit bintang, midvalley, chinatown and the markets. the orangutan shop in malacca and the nasi kandar stall in the middle of nowhere.   i fixed dinner the other night and the maid had a good laugh. i found and understood absolutely nothing in the kitchen. my diet has changed strangely enough, i don\'t know what i want to eat because there is so much to eat...and it is all very diferent.   i\'ve missed the kids, the girls, and my friends. but soemthing feels so off about the other picture, something feels so bloody off-coursed about this frame. like i\'m about to embark on a journey to hell, like i\'m betraying, hurting, and lying to myself, to him, to the people around me. like i know i should not do this, if only, in the end, to protect myself.   i\'m tired and they\'re jet-lagged.      urlLink  \"Everyone has a million stories in their head, all of them connect and form a safety net of experiences that are tagged and filed away under quick reference words, or smells or colors or whatever. You\'re sad when you smell freshly cut grass, but only you know why. When you see clovers you remember your first year of kindergarten, and a wave of nostalgia passes through you and for a moment you feel nothing. And for a moment you like feeling nothing.\"  '),(22685,664485,'01,July,2003','2003-07-01 00:00:00','   you bozo.  '),(22686,664485,'28,August,2003','2003-08-28 00:00:00','   monday drag show night...  four fairies, a dyke, three breeders and two confused. a good crowd i must say, especially when you\'re trying to negotiate between the feuding fairies who have now decided that they will exchange partners.   yet again.     \"but honey, he was soooo last week!\"   uhm.   yeah.  which of course brings in to my next big question in which a certain breeder could not keep her hands off me. strange? very. confusing? even more so.   saw the ex-girlfriend in naked glory, which of course pulled me aside to show me her new tattoo at a place i would rather really not ever remember touching, all this while i smiled and said oh?  and she looked good too.  damn bitch.  she wore a mesh shirt with nothing beneath but black electrical tape. i\'d forgotten about her nipple rings, her and her statements but damn she was cute.   and the best part was probably that she knew it too.   imagine this, the couch, dorothy oz(formerly known as chick from emerald city) and \'other breeder\' locking me in.   and it\'s almost amusing to see the competition between them two.  one rubs her legs again mine.   the other puts her arms around me.  one wants to show me her new tattoo.  the other sticks a lollipop down my throat (?)   one plays with my hair.   the other strokes my arm.   and the only thing i can think about is WHAT THE FUCK IS GOING ON?!?!  screams.   lots of them.  it\'s like lesbians galore in full strength and everyone run for hug and i repeat my summer story six million times but it\'s okay i suppose. because the boys are just cute like that and are always the best listeners.   (how do you shut a gay guy up?  you tie his hands together :P)  then come the pictures.   and more drama.   and x\'s and dollar bills.   of ricketing, swattng indecision and clap, clap, claps.   i keep my distance and am very uncomfortable,  one because she\'s straight for crying out loud and the other because well, let\'s not go there again.   and she questions me, as we are leaving the club and i am desperately thinking of ways to avoid the inavoidable goodnight kiss that will turn a competition yet again, don\'t i miss being with a girl?  say what?  of course i do.  i miss the intimacy.  i miss the closeness and the vulnerability of a woman. i miss her hair, her eyes, her touch. i miss a lot of things, but i\'m beginning to think that i\'ll be okay, that i\'ll be fine down this path and yes, it scares me. it scares me a lot.   i draw back to the conversation with the head doctor in which i tell that i feel like i have lost a significant part of me. that in between growing up and all that good shit, i have become less confident, much more vulnerable and a little bit more jaded and naive at the same time.   and for once, i can tell her that i don\'t feel as much of pressure like i used to. i don\'t wake up in the middle of the night suffocating and panicking because i am afraid i will never be enough.   right now, i think i\'m beginning to let myself be okay, tell myself that it\'s okay if i\'m not prefect. that it\'s okay if i am not enough, or not thin or not pretty. that it\'s okay....that i\'m okay.   that the world isn\'t going to crash, fall, sink beneath me because i let myself be.   ...and it\'s been seven months since i started dating barker.  it\'s been a rough summer.   i\'ve moved out, in more ways than one, and that signifies a million different things.   it took us six and half to admit we were falling in love, that we we needed love and whatever stereotypical things couples are suppose to do.   and he broke down when i finally said those words.   and it was strange.  and it was weird.   and it was very freaky.  but it was also very nice.   and very comforting.  seven months seems like a long ass time.   but it\'s not enough.   but it\'s scary.  but we\'re going to make it.   the swingers are together, how much more crazy can this world get?  the strangest part is knowing that he understands. understands that craving, the girl days, the swinging....or at least makes a good show of it anyway.   and i miss him moving in me. i miss the way his lips trail every inch of my body and makes love to me. i miss the explosions of passion and lust, desire and trust. i miss curling into a warm body and feeling safe, i miss skin on skin and an erection so hard it makes me giggle.   it makes me giggle because i did that.   it makes me giggle because he wants me.   it makes me giggle because i am loved, i am needed, i am wanted.   and suddenly, it is okay.  and suddenly, i don\'t feel so ugly around him.   and i let him strip me.  and i let him undress the layers.   and come close.  and lock the door behind him.   and open my heart.   he comes home in two weeks, and last night was the first night we didn\'t talk. my pills have been acting up on me so i took a wrong dosage and knocked myself out with it.   waking up to find it was twelve hours later and sunlight had risen.   i am home, in a foreign place, but home strange enough.  school started on monday.   i\'m making dinner tonight.   labor day weekend starts tomorrow.   life  goes  on.   :)   how are you doing?  ((hugs))     urlLink  \"I do think I need this trip though the underlying reason for it is no longer as pressing or obvious.\"  '),(22687,664485,'26,August,2003','2003-08-26 00:00:00','   \"it could be well that i\'m going mad; or, on the other hand, that i am simply confused and unhappy; or, on the third hand, that i know exactly what i want but cannot bring myself to do it because of all the pain it would cause, and the tension between those two states of being makes me want to explode. but when david touches me in that way, with tenderness, with love and concern, it all dribbles away to nothing, and i just want to be with him and my kids for the rest of my life. i don\'t want to touch stephen, i want to row about what david may or may not have said to other people, or what he certainly said to molly and tom. i just want to do my job do my job during the day, watch dinosaurs in the evening, sleep with david at night. nothing else matters. all i need to do is hold on to this feeling, and i\'ll be fine.   we go and sit in the car for a while, and david lets me cry.   ....it doesn\'t occur to me for a moment that for david my confession marks the beginning of something, not the end of something, and just because he has known me for twenty-five years, it doesn\'t mean that he knows me or understands me now. he\'s quiet for a moment, and then he says, \"can you come home tonight?\"   -nick hornby  '),(22688,664485,'17,August,2003','2003-08-17 00:00:00','   i had biscuits and gravy at this hideaway diner place for breakfast today. awake at the crack of dawn, taking a fifteen minute walk down familiar streets, way too friendly neighbors and trying to soak in all the warmth of a bright summer morning.   if i had to complain, it was too hot. and i was up too early after restless hours fighting with the night demons, but yet, i knew i had to do this. had to walk down, had to paint on a brave face, had to go give and recieve hugs. and so i did it. and so i was happy, smiling, laughing, amazing, perfect racheal a while.   and i felt her tiny arms squeeze around me, as they all rushed to say hie to racheal. as their tiny squeaky voices said i miss you, and they sat around me as i braided hair, colored strange objects, and folded origami for them. i listened to summer stories and watched as each child fought for attention....a swift clumsy action and there was orange juice everywhere. orange juice on the table, orange juice on the chairs, orange juice just everywhere on everything.   and she starts crying. wailing. screaming. i pick her up and walk her down and attempt to calm her down. soothe her. it\'s all okay honey, it\'s alright. lets go get more okay? that\'s my girl, it\'s okay sweetie. here, it\'s alright. smile. smile. smile. smile. smile.   and i don\'t know if it is the shock of having orange juice splashed all over in a split second or just realizing that in her little three year old mind, orange juice is that important to her. and i wish i was three again, and i wish all it took to calm me down was hershey kissies and a walk outside a while. and some place in me fought to keep the tears down, some hole in me begged to be filled.   and soon it was ten o\'clock, and the kids asked if they could come over and swim, or hang out a little more, or bake chocolate chip cookies, paint t-shirts...everything and anything. and that made me so happy and sad at once, to be so needed, to be so loved. it was strange, i don\'t know why, and so i let them come over and hang out a while.   and played more of perfect, happy, responsible racheal.   and when they left, i cried for the longest time. cried for the emptiness inside, cried for uncertainty, cried for fear, panic and disorder. cried because last thursday i was scared with so much news coverage again, cried that another anniversary is rolling around.   i am only me.   and sometimes, i think i will never be enough.   that i will never fill my parent\'s shoes.  that i will never be a somebody.   that i will always be just plain stupid retarded ugly fucked in the head racheal.  and my heart hurts.   and i hurt.   i hurt because i used to be so much, so happy, so free, so confident. and somewhere along the line, the kid in me took a vacation and i became so cynical, so ugly, so hard. somewhere along the line, i became this horrid existence and i find myself constantly fighting to stay alive. to keep peace. to just hold on for one more day, one more hour, one more second.   afraid i will break.  afraid i will go so fucking insane.  afraid i will let my friends down.  afraid i will never make my parents proud.   afraid i will embarass my brother.   afraid to be me.   and so i swallow the plethora of bullshit they feed me, praying for chemical balance, and the next appointment with the head doctor is in nine days.   but already i am running.   hiding.   fighting.   questioning.      urlLink  \"So here\'s to friendships (old and new, lost and found). Here\'s to pictures and yelling and naughty antics. Here\'s to the White Queen, flitting about, dropping fairy dust, dimpling her way into our hearts. Here\'s to more meetups, and more shared laughs and tears. And finally, here\'s to life. Wonderful, glorious life.\"      urlLink  \"The conversations... the screams... the curses... the profanities The laughing... the drinking... the smoking... The sweet talk... the naughty talk... the cuddling... bordering foreplay\"      urlLink  \"How can anybody describe what it was like last night?\"      urlLink  \"Jamie, I swing every way.\"      urlLink  \"But let me learn, let me make my own mistakes, brush the grit and dirt off and try again. I might cry, I may be difficult, and I may whine about never wanting to live/love/try. But let me wallow till I can wallow no more, and I will be fine. I will get up and try twice as hard. \"      urlLink  \"I was at the Light Side of the Force table before shifting to the Dark Side.\"      urlLink  \"How long did it took you to figure out the tap in the washroom?\"      urlLink  \"the weekend was warped, and the only reminder i have that anything happened at all is the black nail polish on my pinky, a slightly sore throat and exhaustion. and my hair still smells slightly of cigar...\"      finally catching up, thank you everybody much.  '),(22689,664485,'14,August,2003','2003-08-14 00:00:00','   i love this picture of us. there\'s something about it that says i am happy with you. something about the arrogance and bitchiness on my face and the joy on yours. something about having you hold me close and something about the way we are grinning and how i am safe, happy that night.  i\'ve been stuck in my pj\'s for the last two days...wavering in between sleep, paranoia, depression, hopelessness and crying. having phone conversations with barker that start when he gets off work till six in the morning my time, and seven his when our free minutes are used up and we resort to i miss yous and trying to keep strong for each other.   i talked to you last night too, it was strange. the ending of summer love or whatever the hell that was. you say so many things at once, and i am suddenly stuck in the position of the \'man\'. the irresponsible heart-breaker jerk that pulls the strings and won\'t commit. i keep thinking this is wrong and that we\'re seriously screwed up in some way, you, the essence of you, the warmth of your wet knickers on my thigh, the brilliance of your smile when you see me, the adoration, the protection, the regality in which you hand to me. i am your princess and you are my goddess.   but i cannot be with you.   i cannot hurt you, love you, need you the way you need me to.   we cannot hurt her.   we cannot hurt them.  and i say i\'ll see you at christmas.   hiding my private pains and smiling, laughing for you.  i keep thinking my parents don\'t deserve this, don\'t deserve to have this royally fucked in the brain kid as a daughter. and i begin to think that no amount of psychotherapy nor drugs will ever cure me, i begin to think i am incurable. that i am just fucked as fucked would be.   so fucking selfish in my own world, wanting to escape, wanting to run, wanting to only decide what\'s good for me.   still in my pj\'s.  still on malaysian time.   still hiding in a corner.   still crying alone.  i really hate me sometimes.     ( blogger pro changed up and i can\'t figure out how you attach the fucking thing. here are the  urlLink other pictures  anyway. thank alex. )    (sakthi\'s hand?)  '),(22690,664485,'11,August,2003','2003-08-11 00:00:00','   i just called but you were on your way out. why didn\'t you like the electric soft parade? bits of paranoia and wisdom.  i can\'t find the counting crows md that you made out of my cd which you took. you won\'t remember but you actually bought that cd for me to replace the one i lost. It was  the third incarnation of that cd (i lost it twice before and bought em again) There\'s just alot i never told you and you never asked. i wanna tell.  today i went to bar med with irene.met her would-be man who was reassuringly self-confident from travel and age. Phillip came along as well. you\'ve probably never met phillip? Ahh..someone you should marry ..an ethical lawyer. i saw W in groove syndicate. She used to stay with sl some of me wanted to ask the wheres and whats but most of me did the wiser and stayed away.  your godmum found the ideal mid sized firm for me. she said i should call tomorrow. i\'m thinking i should also call my godmum which i\'ll do tomorrow. Both senior partners are women and Joseph and his wife (your parent\'s friends from church) said that would be a problem. Now Joseph and his wife are cool no doubt but surely they did not expect me to rely on stereoytypes, misogyny and old fashioned patriachalism to make the very important decision of deciding where i am going to work.  So he tried a different approach with the next nugget of unintentional bigotry when he said \"you should also avoid small indian firms\" because of the reputation of underhandedness that comes with it.   What\'s a indian lawyer with ambitions and capital for a small firm gonna do? Mamakrize i suppose.  Having said that we are gonna have dinner with them tomorrow because there are obviously volumes of bigoted conservative values we must go through. (admittedly i begged for this)  i call it bigotry now but will i say the same when i\'ve been in there long enough?  you have to keep me sane because no one else will.  '),(22691,664485,'11,August,2003','2003-08-11 00:00:00','   i sit here at the changi international airport in singapore, soothed by the sweet aroma of freshly brewed coffee lingering in the air and the distinct unmistakable smell of his perfume. i wonder who it is that\'s turning me on and churning my insides at the same time...could it be that gweiloh asian dude sitting across me, or that idiot in the terminal next door that keeps on staring at my breasts(thank you miss victoria)...perhaps it is that sweet old man with silver hair that smiled as me as i tried so hard not to cry.   if i could, i would take you all in my arms and hold you tight to thank you.   thank you for a great night out, thank you for friendship, thank you for coming out to the boondocks to see me off...thank you for the four in the morning chats and the long long unmistakable drives in the middle of the night. thank you for helping me make fun of myself and appreciating the darkness and laughter with me.   i search for so many words to explain how i feel right now but i\'m not very sure at all.   i am sad, and happy. i am excited, and tired. i am so many things and nothing at all.   there is the absolutely seductive loneliness of depression that comes knocking on my door, the unmistakable call to wallow in self-pity and hide in the corner a while. i\'m trying so hard not to cry now, it just turned three in the morning here and i am sitting alone at a desk writing out an entry into the big wide world.   tomorrow, later, for 36 hours, i will be on a plane back to pensacola.   i will run into barker\'s arms and hug the old man for the longest time ever.   i will go and poke stupid jokes at puck and scream at our newly purchased ringside ice hockey season tickets.   i will go back to school, take on a full load and try to make the best out of the next few months...the next few years of my life.   i am going back to hell and returning to heaven at once.   confused, heart-broken, tired and extremely melancholic.   this little indian girl that couldn\'t have been more than a few years my senoir leaned her head towards the window and kissed malaysia goodbye as we were taking off. i listened silently from behind to her weeping. the air stewards and stewardresses were nice though, so damn fucking sweet i wanted to stand up and hug them all for doing such a good job. i know how she felt, for somewhere in me, i was that little girl too.   and i still feel so small. like there is so, so many things i need to do and have to. i don\'t know, and i am scared. i am so fucking terrified of the world and my life ahead, and i feel so stupid. i should never poke fun at second chances, never take anything for granted...for if i lost it all today, what would i have gained? what lessons could i take with me? what concrete inspirations beyond materialistic superficial bullshit could i stand for?   i saw my hungarian goddess for the last time and for the last time, she called me princess, i felt my heart break.   i hugged me mom and dad tight, as my dad whispered into my ear, \"we love you,\".   and secretly i told my brother, take care of them for me okay? you be good, and entertain them and keep them happy and all that good stuff.   but i don\'t think i thanked him for flying out to see me.   i don\'t think i told him how much he means to me nor how much i respect him.   thank you koko, for everything, for being my friend and for being my brother.   i\'m going to miss malaysia.   i\'m going to miss catching up with the bloggers and painting bangsar red.   i\'m going to miss all of you and i thank you again, for everything that you guys are to me. for coming out, for sending messages, for making phone calls, for throwing a great party, for letting me be drunk and stupid.   i thank you and love you guys for being just you.   for the lurkers, for the strangers, for the random people that come in every once a while....thank you for being a part of my life and for making dark empty days feel so much better with comforting emails and silly comments.   i have a silly grin on my face while the tears run down my cheeks, my plane leaves in three hours.   and i know i told some of you i was going into hiding a while.   and i know i said i\'ll shut down, settle down, redesign and open up again.   i just wanted you to know that before that, i love all of you.   as i run through my cell phone list to say goodbye.   as i wish again, that i could hold all of you close and hug you guys tight.   thank you for everything.   i\'ll catch you on the other side of the world..   love, peace, and big warm hugs,   stupid crazy silly alco horny retarded fat ugly hole-in-the-head racheal.  '),(22692,664485,'09,August,2003','2003-08-09 00:00:00','   eighteen words for a sunday and the rest of my life:  i am a piece of shit and i should really learn a thing or two about life.   goodnight.  '),(22693,664485,'08,August,2003','2003-08-08 00:00:00','   thank you all semua orang everybody person.   :P  FUCK YEAH~!!!  :)  delte all p0ictures!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  nooooooooooooooooooo.  :P  love you bandyak bnayak!  hehe  :)   hUGGIES!!!  pensacola time: 12:56  malaysian fucking time: 3:55 AM  gah.  jae got sick on the tripel tequila finishing up bottle shot.   you should all feel guilyt!   :P  toaal damage tonight;  corona  jd  tequila  vodka  thank you all for frriends. hehehe  :)  LOVE KOVE LOVE LOVE YOU!!!!!!!!!!!!!!!!!!!!!!111   :))))))  theome to aput the goddess to beddy bye.   my turn1!!!  towoohhowowo!  i havd fun.  thank the qyou you you/  please say you haad fun too?  uhm,. please?  i kiss kiss?  '),(22694,664485,'07,August,2003','2003-08-07 00:00:00','   (eight points for a great day.)     i.  just came home from seeing  jae  my hungarian goddess, finalizing out tomorrow\'s details...btw have i ever told you how fucktastically amazing she is?  exes smanzy, it all works honey!  want to know the best part?  you guys get to meet her tomorrow! yay!  :)  btw. we bought a round of shots or alkie (hello mr. jd) for everyone but any extras + food is all you yeah!  (disclaimer! disclaimer!)   ii.  she ran around the restraunt calling me princess! princess!   lol.    iii.  the well-needed and fucking well timed massuese came over today and while trying to sort out my messed up limbs said, \"you stressed ah?\"   lol.  bangang.     iv.  four hours till i pick my brother up! :)    v.  i got picked for  urlLink the traveller project !   heheheh.   \"Because we thought your request was so charming, you\'ve been chosen to take part in speedbump\'s Traveller Project.\"   *ahem*  i don\'t even bloody remember what i wrote!  oops.    vi.  it\'s friday! it\'s friday!   i\'m seeing all the bloggers/friends/cohorts/strangers in 20 hours!  thank you all, for making this possible.   i swear it makes me all emotional and stuff.  thank you!  hug?   vii.  i\'m going home! (on a positive note)   viii.  i am so very blessed.   *pulls everyone in for a nice long hug*  see you tomorrow!  (or not. and if so, then you will be missed. :( *hmph* )   p/s can i say thank you again?    urlLink  \"When there\'s so much beauty in the world, sometimes I feel like I\'m seeing it all at once, and it\'s too much, my heart fills up like a balloon that\'s about to burst. And then I remember to relax, and stop trying to hold on to it, and then it flows through me like rain and I can\'t feel anything but gratitude for every single moment of my stupid little life. \"  '),(22695,664485,'06,August,2003','2003-08-06 00:00:00','   so guess who\'s ass i\'ve managed to drag along tommorow?  *giggles with glee!*  arriving tommorow morning from london is..  *cheng cheng cheng*  my favorite(and only) brother!!  wa ha ha ha ha.  shit yeah i\'m excited, happy, unbelievably amazed and damn, the gods are being nice to me!!  wooohhooooo!!!  oh crap,  does that mean i have to be good now?    heh, confirmed in  no particular order  alphabetical order are :-  aiz from  urlLink v.o.i.  and his sayang, baizura :) alex from  urlLink yat.ch  ash from  urlLink the insane ox.  calista from  urlLink empress...just life.  daniel khong from racheal\'s house. :P gracie poo from  urlLink you is stupid.  hani from  urlLink hani\'s honey.  jae the hungarian goddess. josh from  urlLink josh lim & associates.  kaz from  urlLink digital fugue  and girlfriend(?) khalilur from  urlLink malaysian idiot . maeve from  urlLink lizard tales.  meesh from  urlLink the meesh experience.  tv smith from  urlLink dua sen.    non-bloggers.  anne(?) bryan drew ee lynn elaine faizal fariz jan jeffry lai jin nicholas pietro rina sakthi steve (yeah, yeah, yeah)   unconfirmed (still trying to convince/drag along):  boon teong from  urlLink warung ikan bakar.   dinesh from  urlLink alphaque . evan from  urlLink that freak . fooji from  urlLink warung ikan bakar.  irene from  urlLink ireneq . jikon from  urlLink jikonlai.com  najah from the  urlLink malaysiana digests.  shals from  urlLink girl:disemboweld.  sharizal from  urlLink perpetual permutating perceptions.      disclaimer: i\'m not really sure if these are all the people. gracie poo and meesh should have better details since my ass is irresponsible and unorganized. :P  woohooo!!  :)  /me happy.   (thank you all for doing this)   p/s those in other states, countries, want to fly in too? :P  '),(22696,664485,'04,August,2003','2003-08-04 00:00:00','   we went filming last night; putting together a series of digital hoo-ha linebacked with shots of the city and everything that surrounded it. the crowded beggar infested streets of kuala lumpur, the hookers and food galore of jalan alor, the mist evaporating power hungry people at klcc, the brightly lighted up towers with the dancing fountains and calm serenity of the park next to it.   and admist it all, i could not help but feel lonely. slightly out of place.   i wanted so much to show him a part of my world, the places i had run barefooted through, or that rest area secret place we trained for many weeks leading into the finals...there were so many things, so many stories to tell and so much to capture but unsure of where to start from.   we filmed bangsar, with a slight introduction to spg(s) and their ang-moh clan. all done from a distance, candidly, rather unobtrusively from behind somewhere. or next to someone. or in the middle of something.   a few jittery frames here and there, a couple of messed up shots...commentary going stale and unscripted, just trying. just showing.   the world through my eyes.   the things i noticed.  the people i saw.  like that fucking clown walking down the telawi three.   (did you see him?)  or the akdemi fantasia people at starbucks.   the sexy salem girls moving in a green trailed synergy.   i wish i captured it all.   the sounds.   the smells.  the atmosphere.   but i tried.   the end product, a hundred and twenty minutes on tape unedited.   of me, me, me.   and my world, my life, my country.   but there is so much more out there....so fucking damn much.   not enough time, not enough drive, not enough passion.   not enough commitment, not enough people, not enough faith.  not enough love, not enough trust, not enough humanity.   emblazoned on the front page, chilling reminders of 9/11.  that fear.  that knot in your stomach.  those tears.    p/s a gift from barker.   :P    \"She could feel the heat of her own breath reflecting off the vibrating, beautifuly shaped, pussy of the stranger girl as she dug her tongue up the girls soaked valleys.  The vibrator on low but deep enough inside to give very little grip from its\' end.  She pressed with the tips of her fingers against the low humming toy and grazed it with the tip of her tongue as she made her second pass, this time drinking as she tasted.  \"Was there anything more delicious?\" She thought.  The girl grabbed her head and pushed it firmly between the girls muscular legs.  Lapping and kissing, she devoured every surface of the strangers sweet pussy.  The taste was only made incomparable by the act.  The girl yelped and moaned squeezing both thier breast preparing for cloud 9.  She drank faster now with vampiric lust bringing the honey to her mouth.  A memory flashed across her mind of a boy.  A memory of not long ago.  And as she shook her head and spread her tongue making a precious memory for this little girl, she remembered back to when she was the girl.  They had met at an audition.  She with false self-consciousness and he with false confidence.  In a blink as if no time had passed in between she was touching his hand, giving him her gift while he hesitantly gave his.  The girl\'s smooth under-clothes were nothing in comparison to the girl under the smooth clothes.  His fingers ice skating over her silky knickers pressing her milk to his tougue.  Then moving inside her, playing in the waves of delight as they lay on the beach with the waves of the night.  He lay her down to bed.  Both were naked and confidently free.  They played for the night a sexual game.  She was a goddess in the night\'s blue light.  Her eyes were honest, even as they closed.  He slid his tongue over her perfect breast.  Then sucked on her delicious tits, running his tongue from her chest over her tummy to her thigh.  He loved her little pussy.  She could tell by the look from his eyes.  He had used his tongue as a toy and moved deep, drinking, just as she was now.  He stood hard for her then.  His clumsiness masked by her feminine touch.  She grabed him and forced him within.    \"This was it.\" she had thought as she moaned.  She wasn\'t even bothered by his cheap cologne.  The feeling inside her was like a massage of the most important muscle.  That night and many others she made him come for her and he made her happy.  As she thoroughly licked the young pussy of the stranger she remembered all this.  And looked foreward to someday soon when she would once again be.....The Girl\"   heh, right on honey!  '),(22697,664485,'04,August,2003','2003-08-04 00:00:00','   so i officially only have about six days (malaysian time) left...    OH SHIT!  might be a good way to describe it?  in any event, come  send me the fuck off,  ogle at sexy people, do tongue exercises, get absofuckinglutely knackered, have long conversations about philosophy, pray to the porcelain gods, dance, talk absolute bollocks, smoke cigars, salems (or your damn dunhills ;p ) or just meet random friendly strangers this friday @ bangsar, kuala lumpur, malaysia.  and yes, it\'s at \'paprika\' because hungarian bartenders are just sexy like that.   yum.  :)   for more details...go  urlLink here  and bug   urlLink her.  or  urlLink her .  (my foxy bitches!)  :P   p/s paprika is halal and serves non-alco drinks as well just in case you were wondering.  pp/s there\'s also valet parking down the road at tamu lounge.  ppp/s melancholic emotional sad grateful miss you like hell post to come later.    so how?  '),(22698,664485,'03,August,2003','2003-08-03 00:00:00','   my  urlLink 26 things project :  a little late, a little tardy, a little here and there.    animal.   urlLink     pensacola, florida.  at joe paddy\'s, a staple seafood market downtown.     authority.   urlLink     back home in kuala lumpur, malaysia.  an antique chinese chair used for traditional wedding ceremonies.     colour.   urlLink     in bed with racheal - malaysia  a splash of summer! crossed halter neck dress.     communication.   urlLink     paprika club, bangsar, kuala lumpur.  malaysian bloggers getting horny.  :P     construction.   urlLink     shah alam indoor stadium, malaysia  national cheerleading championships, dynamitez team.     empty.   urlLink     red chamber, bangsar, kuala lumpur.  flickering in the darkness.     food.   urlLink     rasa penang(?), one utama, malaysia.  assam laksa: a typical malay dish; noodles in a spicy fish broth?      footwear.   urlLink      kicking it up with stinky feet and beaded shoes!     home.   urlLink      home for many reasons with the obvious being a picture of barker at home back in florida and the photoframe that consists of the kuala lumpur skyline.      light.   urlLink     home, pensacola, florida.   long warm baths, candles and roses.  join me?   :)     little things.   urlLink      knickers for charity!     love.   urlLink     damansara jaya, malaysia.  about love,   holding on,  and letting go.      money.   urlLink     santa monica, l.a.  street performers doing their thang. :)     monument.   urlLink     l.a. airport, usa.  tribute to 9/11.    new.   urlLink     santa monica, la.  a newly wed couple dancing in the streets..celebrating.     numbers.   urlLink     home, pensacola, florida.  we both smoke too much.     scape.   urlLink     one utama, malaysia.  space. area. definition.      self.   urlLink     bandar utama, malaysia.  a collage of faces..frames. pictures.      signage.   urlLink     santa monica, l.a.  some obscure pottery shop downtown?      sound.   urlLink     home, petaling jaya, malaysia.   cd/md (sanity) player.      sunset.   urlLink     barnes & noble, pensacola, florida.  an evening of books, coffee + friendship.     symmetry.   urlLink     some elevator, bangsar, malaysia.   when people get easily amused by their reflection..:P     time.   urlLink     at the fair in pensacola, florida.  one of the rides that seem to make all sense of time disappear. (or seem too painfully long.)     transport.   urlLink      boarding passes going home!     water.   urlLink     harem of delights, bangsar, malaysia.   water in both sense...the drink as well as water used to smoke the infused tobacco.      weather.   urlLink     gulfshores, alabama.  hot lazy afternoon.     (pass-ah?)  i feel silly for being nervous and excited.   did you like it?  *winces*  :P  '),(22699,664485,'01,August,2003','2003-08-01 00:00:00','   urlLink    urlLink    urlLink    urlLink    urlLink    urlLink    urlLink      urlLink    urlLink    urlLink    urlLink    urlLink    urlLink    urlLink     photography shi-sha high with strangers and friends.   all talked out, and pasting  urlLink excerpts . it\'s not all that bad.  how was your weekend?     urlLink  \"I love when you\'re out drinking and drinking... and you\'re doing fine... and then you try to stand up.  Whoa. Hello, Mr. I-Almost-Fell-on-my-Ass.\"  '),(22700,664485,'01,August,2003','2003-08-01 00:00:00','   urlLink    urlLink    urlLink    urlLink    urlLink    urlLink     hm.   it\'s never easy to start off a post where you willingly admit that you got shitfaced drunk as a motherfucker last night.   it\'s even more retarded when your ex-girlfriend, the bartender, tells you this morning on the way home that she spiked the other kid\'s drinks. with fucking tea.   and so while i had way more double fucking shots than i remember, a certain  urlLink sexy girl  and  urlLink intimidating smartass cutie  had shots of TEA.   i also got very acquainted with paprika\'s bathroom floor, the ceiling of jae\'s kitchen, and bits and pieces of last night\'s bullshitness.   i haven\'t been this drunk since...  since gulf shores.   two maybe three and a hlf months ago.  and caeleb.   and my first (and last) one night stand.   of feeling like i was raped again. going against everything that i stood for, screaming to say no but agreeing on the surface.   of feeling so dirty on the ride home.   and wanting to hide in the corner for the longest time ever.   but that was then, and this is now.   steve came last night.   you can say it was a nice big happy fucking family.   numb.  maybe i was looking to numb.  or escape.   or something.   i don\'t know.   as he whispered i love you.  i miss you.  i need you.   i said leave.  i said go.  i said touch me and you fucking die.   and i called him all sorts of wonderful names.   then walked back in and smiled.  and pretended.   and ordered another round of double shots.   as the ex-girlfriend came over to see how things were going.  as she whispered in my ears too.   as she held my hand.   as i let the evils of alcohol pollute me.   a tip off, a fucking raid. and jae swings into well, jae-action.   by this time, i am professing my love for everyone.  freaking the bloody hell out of my other  urlLink favorite gay depressed blogger . (who was very yummy btw)  we slip out the back door but i am only spinning and spinning. and keeling over wishing to die.   she rubs my back, and holds my hair, soothing me. calming me down, yet i am still thinking. yet i have not escaped.   the driver stops for coffee which i do not touch. my salmon salad for dinner comes out almost as fast and as easily as i would remember.   but yet it doesn\'t stop.   i am still spinning, and my mind is still going.   she carries me. my strong ass woman, lifts me. as i tell her i cannot negotiate the stairs, and i want to cry, scream and yell at the same time. escape me, escape racheal, escape tonight.   she tucks me into her bed. gets me changed, and strokes my back as i pass out.   and she is up for a good part of the night.   from when i wake up to throw up again, or sip water, or just need to be held.  to be told the world is an okay place.  to know that i am safe.   somewhere between dawn and morning the bastard calls again.   there are six million messages on the phone from friends.   and we talk.   i talk to him in jae\'s lap.  as he apologizes again.  and bullshits.  yadah, yadah, yadah.  i am angry.   i am hurt.   i am tired.  sober enough to know what\'s going on, yet still intoxicated to not be sane or adult about it.  i scream i hate you.   i scream you fucking lying cheating bastard.   and jae still rubs my back.  and jae still holds me tight.   and jae still says it\'s okay.  i hang up, wanting to throw my phone across the room, wanting to collapse into a fit of tears, wanting to escape into oblivion.  and i feel so stupid and dumb.   affected.  why?  bastard.   as i try to remember the good times from last night.   of bringing  urlLink meesh  to the other side.   and horny  urlLink dinesh .  and the fucking adorable  urlLink jikon .  of learning about monte cristos and cigars rolled by virgin thighs.   or having meesh pinch my sexy ass by gracie poo\'s demand.   or meesh and the kissing galore.   and raping jikon.  or terrorizing dinesh.   and i did have fun.  lots.  i\'m just leaving in ten days.   and there\'s just too much fucking shit to deal with.   drama, drama, drama.   somebody hand me a bottle of aspirin.      urlLink  \"aarrgghhh... Meesh and Racheal are so drunk... I\'ve got two lessies on my hands. This is so not for real...\"  '),(22701,664485,'28,September,2003','2003-09-28 00:00:00','   he lies, eyes closed for the first time in many days...so many things rocking and shaking, coming down in his world.   mom in the hospital again, and he is angry, upset, frustrated...so many things at once.   he holds my hand, i bleed, he calms me down.   afraid of the phonecall.   he has to be strong for everybody, the man of the house, the brother, the cousin, the nephew, the son, the boyfriend, the friend, the director, the producer, the manager, the everything and anything in between. yes, stressed would be an understatement.  and so many times i wish i could take it all away, but instead i just lie with him when he needs me to.  and stroke his hair.  kiss his lips.   love him.   it\'s five in the morning, and he\'s just finished unloading another box, unpacking another shelf, fixing another meal, securing another person.   i slip aside into the world and thank god for another day. precious, amazing life, i learned all about it today.   everyday.   he smiles at me from across the room, and says, what\'s up baby?  i say, oh nothing much.   he blows me kisses.   it\'s time to go to bed honey.   he sleeps.   we sleep together.  '),(22702,664485,'26,September,2003','2003-09-26 00:00:00','   urlLink  \"This is just my silly, trivial little gift to the world. It\'s all I have to offer.\"  '),(22703,664485,'26,September,2003','2003-09-26 00:00:00','   lay awake; fall silent beneath angel wings that carry you over crystal clear jaded tears that flow between eternity and history flashbacks of bittersweet memories that surreally come pounding in on reality pixies and broken butterfly wings that lie drying(rotting) on the pavement this must be a soothsayer\'s dejavu experience this must be losing it in tranquility.     \"we are the one percent.\"   the good news? : it isn\'t cancer.   the bad news?   : it\'s  pre cancer.  the smear came back bearing big fat chunks of alphabets printed in black ink ; abnormal.  abfuckingnormal you.  but i was calm.  just somewhat shocked really.  and he kept on saying, it\'s good news baby, it\'s good news.   i go in for a scope and biopsy on the tenth, two days after my birthday.  the good side, i\'m going to turn nineteen. one of my biggest fears and one of the things that kept on going off in my mind...that i wanted to graduate, that i wanted to turn nineteen, that god damnit, i was still a fucking teenager.   the bad/strange/weird side : happy fucking birthday indeed.   precancer.   precancer.   precancer.  what the hell does that mean?  i am going to be okay right?  i\'m going to be just fine.   then what?      urlLink  \"the day went so fast and in such a blur that I lost my dimensional orientation \"  '),(22704,664485,'23,September,2003','2003-09-23 00:00:00','   urlLink    urlLink    urlLink      we somehow ended up hosting four lovely japanese exhange students on friday night and to say it was interesting, kooky, fun, testing, and exciting all at once would be an understatement....put together a hodge podge of foreigners that speak only about a sliver of english, a little white boy and his dumb girlfriend and it results in lots of laughter, swing dancing, the dollar store, vietnamese food and terminator 3  ryan took care of the movie tickets in which the girls thanked us profusely for, this after unleashing them into the dollar store where they went squealing away in japanese and loading up on everything and anything for a dollar each!   perhaps one of the most amusing parts of the evening was attempting to teach them swing dancing. the instructors were going too fast for them (classes are an hour before the live music plays) and so they just watched and made their own moves! twisting around and jumping and turning....!?!?  there\'s something indelibly special about watching everything come together, the dancefloor spinning around with splashes of color, ethnicity and age...and the smiles. the hardwork, the confident and the shy, the layers that all merge into a picture that makes you step back and wonder at the glory of the universe.    his highlight was when a typical american pop song came on air and they knew all the verses and lyrics to jewel\'s \'standing still\'. we struggled to explain \'brat\' earlier and there they were mouthing, singing, busting out english songs. it tickled us quite a bit in which we then of course went into japanese cartoon theme songs...fuck yeah!    \"ahn, ahn, ahn, do rae mo na chi su chi, do rae mon!!\"   :)   we ended the night with sci-fi gore and arnie. just what every damn exchange student needs eh?  oh, that and rolling a house. my first induction to america and it\'s neverending use for toilet paper...  i had fun.   a full circle with someone else filling my shoes at where i was a year ago.   such a delight, such a joy, such a feeling that grows deep inside that calms you down and makes you realize you\'re already home in some odd special way.   ---- seven hundred and fifty two events all happens at once, things that need to be done now, now, now and people that have needs to be fulfilled right this very damn second. and in between it all, you stumble upon treasures and glimpses into a life you know you want, where you are happy and calm and relaxed, but most importantly, at peace.   my auditions were last night and they went pretty good i think. choosing at the very last minute to do a familiar favorite instead of the piece i\'d been working on for the last few weeks, finding confidence within and even maybe my voice. i was nervous as heck, gritted my teeth and smiled, and somehow made it through. pretty damn well too apparently.   i have a callback tonight, and it\'ll be a good thing because oddly enough, the boyfriend doesn\'t need to be at work (surprise! surprise!) and is going to go with me.   the last couple of weeks have really been one of those things where you break or make a relationship, not seeing each other for months, his mom getting sick, moving, the stress, school, my auditions, his movie, his work, my work, the dog, the cat, the family, our schedules, trying to stay sane, living in four different houses, me getting sick, me going crazy, me trying very hard, him trying even harder, us trying hard together and it\'s been insane as hell to say the least.   there\'s a different way he touches me now, and it\'s not something you really physically understand, just a sense of feeling that elevates the place you\'re both in. it\'s this strange thing called love.   and i love it when he whispers i love you as i\'m falling asleep.  or the way he is constantly holding me, touching me, kissing me, smiling...it\'s so fucking disgustingly adorable.   or thrift store shopping and never agreeing on any furniture then finding something we both like.   i am comfortable, and oddly enough, i\'m beginning to think that things might be okay.  that everything might be okay.  i found a whole disc full of pictures from the summer the other night downloading stuff into the computer, and it was strange, funny, weird, fucking weird, but good. we sat and laughed through pictures of jae and i together, reminiscing and fuck yeah she and i look cute together.   i talked to jae the other night, and i know she comes in here. like they all do, and they all take peeks into my world inside, and sometimes it\'s hard, sometimes you don\'t know what to say...how do i explain the summer? do i miss her? yes, very much so, but she has someone else and so do i. i don\'t know how to explain it, perhaps it\'s decisions, perhaps it\'s just knowing that you\'re with someone you want to be with for a long long time. and it\'s hard for me to admit that, he has to practically hold a knife up to my neck for me to say the dreaded words \'forever\'. but i do think about it.   every so often.  it\'s strange to have someone so supportive, someone who you know really believes in you.   this thursday i go in to the doctor\'s and to say i\'m pissingly afraid is an understatement. the other dad wants to take me there and i really don\'t know how i\'m going to handle anything if anything comes. ryan has the day off (again!) which means a lot to me because whatever the results, i want him to be there with me. i am scared and calm at the same time, it might all be nothing, and if it is anything, i\'m a big girl and i\'ll be right.   there are so many things to think about. i talked to mom the other day, trying to constantly decide between staying and going, and she says i should stay here for treatment if anything, and i was thoroughly surprised at her decision. there\'s been a lot of stress lately, from trying to decide when to graduate, to what school i\'m going to, to deciding my goals in life, to just wanting to make it through the next few months, next few years even.  but it\'s been good. i go in to see the head doctor tomorrow, and it feels as if there are so many issues but none at all.   what\'s going to happen to everyone and anyone and everybody?  i don\'t know, and right now, it seems silly but i don\'t really care.   i just want to enjoy life.   and be in love.   and sing to the trees.  and be eighteen.  :)    urlLink  \"and no-one would give a pair of donkey\'s kidneys, as it very much should be.\"  '),(22705,664485,'21,September,2003','2003-09-21 00:00:00','   urlLink      urlLink      urlLink      urlLink      urlLink      urlLink      urlLink      urlLink      urlLink       i woke up to a morning so beauitfully calm and frantic at once. in jammies and a rising sun, i stood out on our dock and took pictures of a country that shook and grew stronger two years ago.   two weeks later, when the computer\'s finally hooked up again and the files are transfered, a single usb connection downloads a series of photos from a peaceful morning not so long ago.   a morning i thanked god for life.   and friends.   and loved ones.  and special moments.   then wept for the departed.   hugs    urlLink  \" i\'m in a position to begin taking stock and slowly righting what needs making right. it will take a little time, but i\'m gradually moving forward. there look to be interesting times ahead. i\'ll be here to document them. more soon. \"  '),(22706,664485,'18,September,2003','2003-09-18 00:00:00','   day light breaks on a tangle of bodies in the back porch. the \'florida room\' they call them, where we have laid out a bed, some sheets, the mini fridge and candles in our screened in haven. away from everything, in the middle of nothing.  at one with the stars and the beauty of the universe, making love to the faded sounds of the night and the freshly awakened sounds of the morning.   s  l o w.  thenfastfastfaster.  and  s l o w  a g a i n.   he watches me dress in the morning, under the covers, away from the sunlight; hiding.   he whispers and nuzzles my ear, holding me tight...and i shiver.  ....and it\'s not just from the cold.   ----  he\'s off to see his mom, and i sit at home waiting.  cleaning up.  doing proper girlfriend shit that proper girlfriends do.  and he comes home with a gamut of emotions.   it must be hard, a dozen roses sit in a vase.   he envelops me with feelings that i want so badly to transcribe into words.   like being needed and needing.   comforting.    ---- moving, packing, putting bits and pieces and months of people, and a person, and life into cardboard boxes and sealing them tight.   renting u-hauls and driving up and down, back and forth. lugging a hot tub hooked onto the back of the trailer...into the big pretty backyard with the wide expanse deck.   a week left.   and every now and then, he looks at me from across the room and mouths, \"i love you,\"  and his sister laughs.   his cousin smiles.   his aunt and mom gossip.   we collapse together at the end of the night to bits and pieces of everything and emptiness but it is enough.   ---  moving, and everthing seems filled and disarrayed at once.  '),(22707,664485,'15,September,2003','2003-09-15 00:00:00','   some nights you pray for sleep to come and never leave. like bouts of paranoia and depression, it hits you like concrete bricks knocking you over, flung with the capacity of killing from a seventy story avant garde building shaped after a bamboo in a cosmopolitan city.  and you begin to wonder what death must feel like, back here almost certainly, failing yourself and the ones around you; you\'ve given up already, yet the actual action of life sucked out of you has begun but once again, like everything you are, a complete utter bloody disgust.  a comparison of the she-devil in your mind, and every word pierces through the soul like old, rusty, bluntly sharp as ever knives. but you\'re all bled out, and all semblances of breath and humanity seems to shirk away. your middle name should be medea, you are midas in reverse. everything you touch crumbles and falls beneath you.  you useless fucking piece of shit.  and i cry.  i cry out to someone and everyone and nobody at all.  i want to run, run, run and never return.  i want to hide just a little while, beneath these safe arms that will hold me but i can\'t trust myself, i can\'t seem so let myself go.  and i suck ass.  up the medication to 60 mg, all the drugs from the treatment, and the shots, and the transfusions, and the hormones, and everything that\'s possibly fucked up about this body is angering the demons in my head. they return much more vicious than ever; this time shrinking me smaller and smaller into bits and pieces that i am convinced will never be repairable.  he wants to hold me tonight, and i crumble into so many pieces but i feel so numb. it\'s as if the pain just envelops and bubbles it all in. i want to share, i want to tell him so many things....like how much i\'m hurting, like how everything feels so wrong, like the mania and the depression is a runaway train at full speed with no brakes(breaks) and i can feel it cominng, coming, coming and i am scared, and i am panicked and i am absobloodylutely terrified.  and he holds me close.  and he holds me tight.  i am sick.  i am sick on the couch, in the trashcan, dry heaving in the bathroom and crying on the bed.  so  fucking  out  of   control.  there\'s something about bullimia and being comfortable with the puking and something so fucking terrifying of losing control, and having absolutely no control of your body at all, or how it works, or what the hell is wrong with it.  but i just sit here pathetically and cry, cry like there is no tommorow. cry like i don\'t know who i am, cry like i\'m scared of myself and so fucking disappointed. cry like a little child sobbing and sobbing and sobbing.  and this whole other part of me, this part, the strength, the whole part of racheal that used to be so happy, so determined, so fucking everything has gone away. and i\'ve felt bits slip away with the years gone by and now i am nothing. i am a recluse of emptiness and emotionally fucked in the head, in the heart, in the body, in the soul. i am one fucked up little child.  and i want to scream and scream and scream forever.  inaudible incoherent sentences consume me, and i feel so weak. so weak from the throwing up, so weak from the medication, so weak from thinking and feeling and crying, and so i just fall.  and he catches me.  he cries.  and i don\'t mean to make him cry, i don\'t mean to let anyone see me this way but he\'s here, he\'s always here, and he keeps on seeing me, and keeps on unlayering, and keeps on holding me and i still want to run. i still want to be independent and cynical but i need him.  i need his security.  i need to listen to his voice soothe me, calm me the fuck down from hyperventilating and panicking, and just plain old losing it.  i need him to change my bandages, and wipe up the blood, and do the dishes and hold me to sleep, and protect me.  i need him to go scream at the crazy administration people to get me an appointment right. now.   or put my pills together in the morning, hold me up for 30 minutes, 8 oz of clear fluid, 8 oz of milk, and fiure out everything between because i am so useless and i so fucking bastardy.  how can you still want to make love to someone\'s that so disgustingly ugly and painful to look at?  how can you still get in bed at night and hold me, and stroke me, and tell me you love me?  how can you still want to spend the rest of your life with someone who fucks up constantly, who\'s so fucking useless she\'s not really worth keeping around on earth, how can you still be here?  and it hurts him to see me hurt.  and i try so hard to hide it from him.  i try so hard to only cry when i know he\'s sleeping.  or smile when it hurts.  but he knows.  he knows every damn bloody thing.  and it\'s frustrating.  and i\'m frustrated.  tired.  giving up.  harnessing.  just trying to keep it together.   i love you so, so much.     urlLink \"Sometimes a song, the breeze, the sound of the wind chimes and the smell of incense could still make me feel light and weightless inside, the girl who was a teenage witchbaby. But more often than not I just wish I could fall asleep and not wake up. And all I really want is to be a witchbaby again. All I want is to feel infinite one more time. \"  '),(22708,664485,'13,September,2003','2003-09-13 00:00:00','   i sat at the porch tonight, watching the brilliantly scary display of crashing lightning and thunder. synchronizing the pounds of gallons and gallons of rain splashed onto to pavement, the heavy continous rhythms of torrenting rain drops on the roof....and i hate thunder. i hate lightning.   i\'m a fucking terrified pussy when confronted with such uncertainty, when letting go to allow the universe to take control.   but something drew me so strongly to it tonight.  i can\'t find my camera charger, i don\'t have my lap top cable, the transfer programs are on ryan\'s, blogger pro is shutting down, chris is on leave, jikon\'s taken a sabbatical, i am questioning and so fucking nauseous....the world seems to be spinning too fast and not shaking enough.  i use the back of my hand to put out my ciggarette. and the first layer of my skin burns off with the glowing ash i\'ve pressed into it. my white flesh beneath it makes an appearance....and i wonder why i don\'t bleed.   it doesn\'t hurt, not much anyway. i\'m on enough of painkillers to jump off a sixty seven foot building and still not feel the pain. something inside is stirring awake and i am in a labyrinth spinning around in circles and coming out at every available dead end.   i have homework to do, so much pressure to succeed. i miss my mom and dad, i miss my brother, i miss grandma and my family. they\'re over in singapore for the weekend, and i can\'t help but feel the maid has taken my spot. she went with them, and i am a million miles away wishing that i all i could do tonight was curl in with my parents on their bed and fall asleep. and feel safe. and be at peace.  it\'s the mooncake festival back at home. and i remember when we were younger, everything it signified and how we\'d light up the garden leaving trails of candle wax on every available surface. burning at least one paper lantern if we were lucky but feeling so much pride at the end of the night, when everything fell into place and was so perfect. when my cousins and i ran around and squealed delightedly, when the adults gathered at the porch and overlooked us playing. and the whole place would be warm in a yellow glow of chinese paper lanterns, little candle sticks, and smiling faces.   peeling away only the outside and never liking the different insides that were mushy like peas. or the different animals on the lunar calendar that came as cookies in little plastic woven baskets tied together.  i miss home.   i\'m beginning to think that i\'m beginning to forget what baby isabel smells like, or if my granpa still likes the same brand of tea because people change you know. i\'m afraid that while i am growing up and changing and maturing here that i miss out on the important things that mean so much to them. like forgetting the little things. like wanting a hug from the kids. like blogger meets that are tantalizingly chaotic.  selfishly, some part of me is so relieved to be back. it\'s quiet here, and when i want i can take a swim at the bay...play with the dolphins and ducks and mullets. or take a boat ride. or drive into town and watch the local bands play at the cafe downtown.   and there\'s always a church open anyday, anytime. always a service for possibly anything and everything. of coming back to some place i used to know and i swear i want to get to know again....but it just takes time. and accceptance. of holding hands in the back of the church with some stranger from another part of the world, or another tourist and saying the lord\'s prayer and marvelling at how sometimes the world seems so beautifully connectingly small.  but most of all, i like how i can slip into the shadows. be absolutely invisible, yet i hate how fucking alone it makes me feel.  to want this so much but crave for ups and downs at the same time. for the thrills, the amazing highs and bouts of mania, the crashing, falling, weeping pain of losing grip on reality and coming to terms with yourslef.  i\'m really okay, and if i have to think about, i\'m doing much better than i expected to be...but then again, i\'ve always been miss superwoman so well. yeah.  but it\'s alright.  i\'m going to be okay.  i\'m okay.    hugs  '),(22709,664485,'12,September,2003','2003-09-12 00:00:00','   some nights you swallow your fears and choke back the tears...yet some nights, you just let it all rip apart and crash on you. falling freely with the ground caving in beneath you into an oblivion somewhere. a destination, a journey, an endeavor of hope, trust, heartache, confusion and slaying your dragons, silencing your demons.    she wakes up in morning grogginess, the bright red light of the digital clock beside her reads the breaking of early dawn. she tosses and turns, willing herself to return to the other world, willing herself to escape reality for another hour, another couple of minutes. willing to escape.    you think about the parentless children around, the little five year old that needs a heart...that wavers in between. a beautiful symphony of life and smiles stripped away.....just letting precious minutes waste by, waiting. waiting for another five year old kid somewhere else to journey to heaven, so he can finally have a healthy heart. so he can finally do five year old things. like play in the sand, and tag the dog, and chase after soapy liquid bubbles on the boardwalk. so he finally run, and dance, and spin around.     fumbling around for her lenses, knocking over herself a couple of times as jarring pain envelopes her. she falls back and curls into a fetal position, desperately crying. so fucking alone.     remnants of nine-eleven come and gone again. second time around and yet life has rushed by and remained in a standstill. a roomful of strangers of every ethnicity, of every gender preference, of every religion and of every background crowd in a little park nearby. lighting candles, joining hands, shedding tears, saying prayers.   yes, the pain is still there.     she peels back layers and works so hard. she tires herself endlessly with school, with friends, with the partner, with the head doctor, with her meds, with her parents....she tires herself in trying to remain sane. or at least some semblance of it. she wears herself out being normal.....and she is tired.   again.    the terror of the day a nation fell apart and joined together. a day that signifies the world begging for humanity.   i swallow it down, but suddenly everything is so overwhelming and it all just seems too much to deal with at once.   ....and i wonder if this is the ride back down again?  but i don\'t want to watch the rollercoaster on the sideline, i want to fall and scream and yell and live.   i want justice, and fairness.   i want equality and understanding.   i want to save someone.  i want to be someone.  i want to hold something.  i want to paint a picture.  i want to sing a song.  i want to smell the trees and play in the rain.  i want to reach out and hold your hand.   i want this world to be smaller, closer....nicer.   maybe.  someday.  maybe...      urlLink  \"It\'s the children\'s voices that make this so much harder.  At the World Trade Center, the month, the day, and the minute come, and their voices cut through the sounds of the city. They sing our anthem: clear and strong and beautiful\"  '),(22710,664485,'12,September,2003','2003-09-12 00:00:00','   being awake for twenty sum hours on painkillers, caffiene and nicotine is definitely detrimental to the soul.      urlLink  \"Driving to work today, I glanced over at her lithe presence in the passenger seat. She\'d kicked off her sandals and folded herself into an accordion shape, knees mashing into boobs, adorable bare feet up on the dashboard. Her head was tilted at an appraising angle and a slender arm reached around her legs, painting her toenails a glittery opalescent silver. A beautiful girl in a beguiling moment, phasing from the familiar to unfamiliar and back again. I feel like I\'ve been with her forever. I feel like we\'re just beginning.\"  '),(22711,664485,'12,September,2003','2003-09-12 00:00:00','   long fucking ass post \"summed up\" in ten points.        i. blogger is a fuckturd.   they shut me up for a whole fucking damn ten days!   ii. blogger is amazing.   thank you tech-support.   iii. i\'m being tested for cervical cancer. whoop dee doo dah!     \"the world is a funny paper read backwards, and that way ain\'t funny,\"   so this is this week\'s drama.  and i seem so nonchalant i know,  but maybe i\'m in shock.  maybe i\'m just used to it.  maybe, i don\'t know, just maybe.   maybe it is nothing.  maybe it\'s just an infection.  maybe the slides she examined weren\'t mine.  lets just harbor on maybe, it\'s the second anniversary of 9/11, and i still have hope.   i\'m going to be okay.  i\'m going to be alright.  because i\'m a fucking fighter, because i will make it, because i\'m just good like that.     iv. be grateful for what you have and what you don\'t have.   ever notice how extremely fucking far buildings are on campus when you\'re on heels and you have ten minutes to get into class?  and this is going to sound so wrong, but pap smears are so fucking disgustingly painful, stupid, and retarded.   how is it that the answer to everything is sticking something up you, poking around, and aggravating an already equally aggravated thing?  blech.   blech.   pooi.     v. my question was if you missed me and not if i was pregnant.   but just for the record, HELL NO! and i was puking like tap water this morning so that was the first thing that came to mind. (thank you depo-provera)   vi. why in the goddamn hell am i so cheery?!?!    vii. we had a quiet night of getting stoned on homemade magaritas and \'fruitful\' conversation the other night. maybe i\'m all talked out, or as we learn in psychology. in fucking denial.   really though,  how many damn times have i watched this movie?  change already!    viii. to the one that completes me, and lets me complete him ;     so often i know it\'s crazy. so many parts of us, so many different experiences and meanings to us are derived from other people, other pictures from other places in our lives. our past sometimes knowingly, sometimes unacceptingly twined into us and now.   \"we do good together, you and me...or i...well, you know,\"  and there is so much left in between that stands preciously and precociously to be unravelled.   waits to be explored.  pussy whipped and all things great.  little things, silly things, that suddenly, crazily fall into place.  with time.  with learning.  with ups and downs and lefts and rights and befores and nows.  but eight months later, going on nine...we are still together.   stronger than ever, more vulnerable than anything.   but i love you.   and love can be so many things and nothing at once.   the tears you cry from a whisper, a touch, a breath of silence on thin pulses..  three simple complex words.   i whisper, i scream, i bellow.   wipe away, it\'s okay silly, we\'re okay.  i have you.  and you have me.   welcome home, sexykins.  welcome home, luv.  we made it through summer.  ......even if i break.  even if i fall again.  even if i scrape my knees, break my neck and never walk again.   or so i say.   now.  or so i do,  now.  or so i want to be,  now.   and now is everything.   and everything is now.   building, rebuilding, tearing down, mending.   and a whole lot of loving.   :P  i feel your breath on my skin, and and you move deep inside me, beneath where skin on skin touches, you touch me right there.   and i am at a crossroad again.    reaching for your hands that beckon my way.   looking in your eyes that gently guide your soul.  we celebrate an anniversary...and you hesssitantly spurt out \'i love you\' in my language.   my laughter fills the air, rumbles in my lungs.   i love you too, silly.  i love you.   thank you for being there, here, in my life.    ix. if the social life and medicine doesn\'t kill you, the homework will.   5 papers and a presentation to finish THIS weekend.   x. and lastly, to sum up everything and anything, life is like a dick, when it\'s hard, you fuck it.    :P    have a great weekend y\'all!  hugs, love and peace.   (i want  hugs  all three damnit!)  oh come on, humor the dying bitch will ya?   :P    urlLink  \"But, i think, if there were more hours in the day, then there would be no sense of urgency.\"  '),(22712,664485,'02,September,2003','2003-09-02 00:00:00','   okay, big news, small news, mini news, mundane news and la di da news.   !!  but for now,  thank you blogger for taking the stick up my arse and deciding to work again.   miss me?  :P  huggies!  '),(22713,664485,'02,September,2003','2003-09-02 00:00:00','   i\'ve found out.. that the reason i believe is because there are people to inspire me everyday.  and thoughts, and pictures, and seeing the world a little bit different from my sometimes rose colored glasses aren\'t too bad.  that while my perception of things are admittedly rather dramatic, they are also sensitive.   while i panic in the middle of the night and toss around through restless dreams of insecurity and indecision, that in the end, i am who i want to be and i\'ll be whatever that\'s meant to be.   i sat through my technical lab class today, an hour and half long of lectures and constant brickbats of having to come back with something smart to the other five burly guys in my class. defending the little asian girl that\'s playing with the power tools, building the set, and figuring out the light management.  i have as much heart, as much commitment, and as much perseverance as you fuckers damnit.   so i prance around in my stillettos, and take my bundle of work clothes and tennis shoes, albeit, they are red and white with plastic stripes and rather funky to say the least. and strip my jewelry, and slap a baseball cap on, and tie my hair up, and prove my worth.   the wolf whistles, the constant fending off of people staring at your breasts (what breasts?) and making snide comments.   but i am like the rest of the guys, a little bit different in shape, a little bit different in size, but special in a whole lot of other ways.  and that is truly kinda nice.   and that is truly kinda special.  :)  signs you\'re in redneckville geriatric town.  going to the annual mullet festival and tossing fish the furthest, target shooting, and being served hush puppies.   how\'s my day going?  great really. (and that\'s SCARY!)  a lot of pressure to succeed, but i\'m going home for christmas.  and suddenly the mere thought of getting on a plane and returning home is a lovely feeling.   that simultaneous calls from the hungarian goddess and barker put a grin on my face minus the odd instance when the colliding weirdness of barker and her discussing how i taste was just weird beyond strange.   but then again, things are always a little catastrophically different and strange around here anyway.   somebody put up a bid to buy my pictures the other day. and so i spent most of the weekend putting together an attempt of somewhat organizing my portfolio. and it was good money too.  but it wasn\'t the money, nor anything else that i really cared about.  it was that for those brief moments, i believed in myself and saw what i could do and what i could be and what i was made of.   that while i passed by stan today he hugged me as a friend and not as a director and made my skanky ass damn bloody try out for the auditions in three weeks.   because someone believed in me.  because someone thought i could do it.   and last night, it all dawned on to me that even with the 18 credits, the movie shoot, the set work and the musical preparations....that i might be able to do it.   that i might be worth something.  that i might be special after all.  or just for the next few minutes at least.  or just for a little while more.      urlLink  \"And you are sad because it has been so long and so loud and so lonely. In this brief moment of clarity you realize what you have missed. All this time you thought you just couldn\'t grasp it, it was too far away, but it was right in front of your face and you couldn\'t hear it. \"  '),(22714,664485,'01,September,2003','2003-09-01 00:00:00','   tori amos and ani difranco interchanges between wafts of stale cigarette haze and dying souls. no more politics, or talk of things that pains my heart so. another anniversary looms, but the hurt, fear, pandemonium still remains.   a photo exhibition set up, a shoot prepared...cameras aimed and ready to capture, to elaborate, to draw into the innermost chambers of an unknown spirit and find faith.  labor day weekend, of tornading hurricanes and humid summer nights.   of wanting to be free, alone in the dark, dancing naked in the warm bay waters crystally emerald clear and the white sand beaches. beneath the bright twinkling stars, beneath everything that seems so surreal and perfect and wrong all at once.   granpa says i love you.  an emerging familiarness in a crowd of faces. of strangers and things you used to know but can\'t remember now. Of memories you\'d like to tap into but cannot find, cannot sink(drown) into.  the seduction of emptiness cascades the immediate need for something, for anything.   just anything at all...but nothing at the same time.  indecision and the constant pressure that pounds and locks you in.  what are the demons that drive you into this oblivion?  you watch the mullets leap around you and the waves crash and float surrounding you. the salty immersion of tears and sea water, the bouyancy of too much weight(fat) on your shoulders(body).  you feel invinsible, and no one really listens at all.  you are the cute girl in the class, the friend of everyone, the believer of everything.  you are nobody.  you are nothing.  i miss you malaysia, i miss you certainty, i miss you everything i used to know and don\'t know at all.  i miss you understanding.   i miss you happiness.   and normal-ness.  i miss you, me.      urlLink  \"A howl crescendos, deafening, but it drowns out the groans of these ancient pillars. A sign of failing strength and one can only wonder when they will finally give way. That’s when reality comes crumbling down around you. \"  '),(22715,664485,'23,October,2003','2003-10-23 00:00:00','   someone asked me the other day what it is to be normal....and i honestly didn\'t know the answer.  normal now isn\'t like it used to be.  normal is now comprised of panic attacks, or episodes of mania. normal is now crying myself to sleep, or wanting to hurt myself. normal is now letting everyone down and starving myself. normal is now going at a hundred miles an hour on a ride that takes you nowhere.   it\'s that feeling you get that you\'re letting everyone around you down, or that you are just a big fat pain in the ass. it\'s the dissappointment and anger towards yourself, it\'s that hopelesness in the middle of the night, middle of the day.  it\'s looking at the ones around you and knowing deep inside that they deserve so much better.  it\'s coming in here and writing a stupid entry to strangers, it\'s coming in here and hoping to write it all out and achieving nothing at all.  it\'s coming in here and having a good cry, only that there is nothing good about it. except that you are losing it, and some part of you cackles and mocks you. laughs at you again.  what bloody made you think you deserve to be here anyway?  do you think that if anyone ever saw you inside they\'d love you?  hell no bitch.  and you know that by hard.   so sometimes you tell yourself to just die already.  and sometimes you just give up.   sometimes, you just want it all to end.     you are nothing.  '),(22716,664485,'21,October,2003','2003-10-21 00:00:00','   took this last night out at the beach watching the sun go down. there\'s something about this picture that makes me tear up and calm down at the same time, i don\'t know if it was because of the many things i felt while taking it or if it is the simpleness in it.   it\'s strange to be alive.       urlLink  \"The last paragraph. Suddenly there\'s so much I want to say. Things I want to remember, all those blog posts I began but never finished. People I need to thank for their kindnesses, for sharing in return. Goodbyes and more goodbyes. I wish I didn\'t have to go. But this is a happy ending. Not the happy ending I would\'ve written for myself when I started this blog two years ago, but life is full of surprises. Even a plain life like mine.  Take very good care of you.\"  '),(22717,664485,'20,October,2003','2003-10-20 00:00:00','   air mata melinang gelora angin yang membisik kepada hatiku. kusemacam tidak percaya diriku lagi. terlalu banyak yang mengguris fikiran, terlalu banyak yang membebani jiwa. adakah tempatku di dunia ini?  '),(22718,664485,'20,October,2003','2003-10-20 00:00:00','   they discharged me, it\'s sorta odd to get back into routines and things.  ---  i\'m currently pissed at shredder @ the escape artist @ let\'s piss and shit everywhere @ the not so silent dog.  mr smarty-pants just about gave me a heart attack...i had him in the back part of the living room and had our doors wide open to let the wonderful cool night breeze in, but mr. dog, the brilliant fucker that he is, found out a way to escape and comes charging, and i mean charging, out, jumping all over me. bear in mind we live in a corner house, and bear in mind that our street interconnects with 9th, and bear in mind it is a busy ass road.   so in a state of panic, hysteria and madness...i scream.  fuck me.   pah...still love the damn fucker though.   ---  i\'m doing better for what it\'s worth. we went out to the beach to watch the sunset tonight...the bad thing about not being in alabama anymore is that the bay isn\'t in my backyard, the good thing about being in pensacola though is that the beach is about 10(?) minutes away or so. sweet hell yeah.   well, that and having someone to come home to every night.   :)  ---  we were at the c.s.u today, and it was the first time i was in the front reception area. that was some strange feeling...all these metal detectors, people searching up your crack and shit..and the guy up front says, see how much he loves you?  he came every night, and sometimes even at strange hours to bring me stuff i needed. even if visiting hours were over and i didn\'t get so much of a glimpse of him. we were driving home, and he stops to say, i love you so so much racheal.   he pulls me close to make sure i\'m warm...and kisses my wounds. he worries, then worries somemore. making sure they check this and that, that and this.   am i still kinda in the funk?  yeah, a little. it gets depressing sometimes, it\'s so weird to be in touch with reality again.   you\'re holed up for a couple of days and it\'s strange to pull into your driveway. or think about what to make for dinner. or even just to breathe in and out the smells of something familiar. something you know. it\'s strange to go back to your life, to go back to being racheal khong, to go back to writing and school and learning and trying. it\'s strange to return to \'normalacy\'.  --  i sit here typing this, blogging...something so simple. something so easy for me to do, bitch about life, hate myself, say what\'s wrong. and part of me still wants to sob like a little girl, and part of me is just strangely serene. not an absolute calmness but just more of a blank. like where the hell did my whole entire week go? or what is going on inside of me? or even just something as easy as where do i go from here?  i want to thank all of y\'all that mailed me. that sent flowers. that stopped by just to say hie. for a hug. for all of y\'all that made me smile, that said it was okay. i wanted to say thank you, and thank you again.   i was...very scared, i think i still am. perhaps some part of me will always be. afraid of the world, afraid of me, afraid of life.   my schedule is jam packed from trying to pick up things from where i left them. i have rehearsals every night as we lead towards our opening in a couple of weeks, i have homework and assignments from school, i have laundry to do, emails to return, sorries to apologize. for drifting away, for trying to take my life, for not being strong enough, for just being me. six thousand appointments to meet, two on wednesday and friday when we find out the results from the biopsies they took two weeks ago. it\'s hard to imagine that that was two weeks ago isn\'t it?  it\'s hard to imagine how life has moved on, and changed, and stood still at the same time.  it\'s hard to be stuck in the same body, but to go through a spectrum of emotions. to have a myriad of thoughts.   how do you convince yourself that suicide isn\'t the answer? and if you don\'t, how do you deal with being a failure. yes, a failure even at taking your life.   i think you\'re going to laugh at that one. it\'s the here we go again ride. and maybe i\'ll never be okay with myself, never be really proud of me. and i know you\'re going through some shit of your own right now, fuck me for even making any reference to you eh? but i read through what you said, and i recognized that insecurity. that i hate the world and the world hates me, yet i also knew that little part inside that still said, i\'m okay damnit. i\'m fine. i am alright. i have every damn bloody right to be here. you\'re a good person,and yes, you are okay, you\'re alright. and maybe some day we\'ll both learn to love ourselves. or something to that effect anyway. i\'m here if you ever need anything, and that goes out to anyone that reads this, i know i may not be much but if you need someone to talk to, bitch at, compare shitty lives with(i win!) or just play scrabble with. i\'m here.   hugs.  take care all.   thank you again.      \"I thought of you today and prayed that you were feeling better. I felt fresh tears on my cheeks, as I thought of you there---alone, in need, unable or unwilling to reach out---as your life started to fade away. Thank God they came in time.   I want to tell you that I’m here for you and that I am not mad at you. It’s not your fault. Sometimes the pain seems too much to possibly bear… I remember how that felt. I want to reach out for you and silently hug you tight. I’d like to say I know what you are going though but that wouldn’t be correct, no one knows that but you, however I can say that I understand. Possibly more than you will ever realize.   As I sit here I find myself looking at a mental scrapbook of the years you have been in this world. Years that I have known you. Your bright, quick smile. Your brilliant mind. Your natural beauty. I can only smile. Our lives have been enriched simply by your being in them.   I know the tunnel is dark right now and there doesn’t seem to be a way out. I know that everyone in your life is telling you about all that you have, all that you are, and what the future can hold for you so I’m not going to. It’s not my place. It’s yours. Eventually you will find your inner voice and this voice will begin to teach you. I only ask that you at least try to hear it out.  Just remember that this is about moods, about emotions, about feelings. It’s not who, or what, you are. It’s a thing, something that can be addressed, treated if not cured. Give yourself that chance. Please don\'t walk away without a fight.  Reba\"      urlLink  \"I need mental help. I know I do. \"  '),(22719,664485,'18,October,2003','2003-10-18 00:00:00','   i wrote something absolutely profoundly stupid the other night.   i thought i\'d share that with you.  it went along the lines of a cherry and toe nails.    \"cherry, cherry, line my toenails,\"  (???)  i suppose pretty good signs that the sanity was going.   it\'s an interesting thing to be in the \'crisis stabilization unit\'.  even funkier to be woken up at six every morning for vitals and bloodwork.  .....the last couple of days have been rough, good though; i\'m supposed (trying) to view this positively. getting off my meds wasn\'t exactly the smartest thing to do, even more underlined when you\'re locked in, baker acted, and have an increasing medical bill to boot. i have bruises from where they\'ve continually drawn blood from, tired wankingly weak legs, a broken body, a faltering though system, and the extremes one mind takes you to.   i am not good, or great...but i\'m okay.  the psych techs on this unit is excellent...i am comforted and somewhat secure. a little bit more sane and rested than the couple of weeks leading into this.   it\'s still a sad place though. heartbreaking really, i spend almost every waking moment caged in with the other patients (or as pc america likes to call them, \"clients\") that have lost their mind, on the way to losing their mind, or are just about to.   and i wonder what the hell i am doing here.   or which category i fit into.   why so many people sick? why so much hurting?  ...emergency rooms are scary places to be.   the early crack of dawn, being asked seventeen thousand things and not being there at the same time. smiling faces, crying hearts, delusional dying depressed fucking up you.   the boyfriend holds the puke bucket where black vomit continuosly comes out. he gets water, another blanket, the nurses. he holds your hand and assures you it\'s okay, he spends twenty hours holding you, wiping you, stroking your hair, talking to the doctors, assuring the parents. he takes the role of someone that you know for a fact now genuinely loves you. cares.   he sings this to you as you fall in and out,   \"Let\'s Stay Together Al Green  (Al Green - Willie Mitchell - Al Jackson)   I, I´m so in love with you  Whatever you want to do  Is all right with me  ´Cause you make me feel so brand new  And I want to spend my life with you   Since, since we´ve been together  Loving you forever  Is what I need  Let me be the one you come running to  I´ll never be untrue   Let´s, let´s stay together  Lovin´ you whether, whether  Times are good or bad, happy or sad   Whether times are good or bad, happy or sad   Why, why some people break up  Then turn around and make up  I just can´t see  You´d never do that to me (would you, baby)  Staying around you is all I see  (Here´s what I want us to do)   (repeat to fade):  Let´s, we oughta stay together  Loving you whether, whether  Times are good or bad, happy or sad\"   and you wonder how you ever got so lucky, and you wonder why he stays, and you wonder why he still tells you you\'re beautiful in a horrendous green gown, an ashen face, and tubes stuck everywhere.   a diastolic heartbeat, falling blood pressure, sticky tapemarks on your chest, your tummy, your arms from the monitors.   yet you know nothing and everything that\'s going on around you.  you want to block out the last few days.  and every night, as highs and lows tie in together and closure to another day comes; you crash. so pill happy c.s.u. feeds you another tranquilizer and anti-anxiety medicine, and they threaten to give you a shot if you don\'t at least attempt to digest this little green pill.   you cry, you cry so hard you hyperventilate, you cry so hard you are in your own world, you cry so hard you never stop.   and every hour, they let you out of the locked rooms, in the still caged facility, to the tiny garden to get some sunshine. or breathe in some fresh air. or freeze your butts off. they take the strings off your pants and hoods, they take away anything that might possibly harm you, even your pen, even your sanity.   they say they\'re just trying to keep you safe.   just trying to keep you alive.   they have groups around the hour, but you just sit in the corner and weep. or cry. or just stay silent, watch the others...everyone in here feels the need to talk about how they feel, struggle for attention, yet you are still alone in that you\'re the only one that feels safe yet wants to get the fuck out of here at the same time.  up by six, bed by ten thirty. restricted, regimented, everything is planned. noted.   at 4:45, racheal sat on the bench.  at 4:50, she watched tv.  at 5:00, she had mash potatoes for dinner.   yes, even what you eat. what you hear. what you say.   and someday, maybe when you\'re bettern, you\'d like to look at those charts, yet some day, you hope you never ever see them. read them. know how absolutely fucked up you are.   i am afraid, and scared.   i am pressured, and piss fucking frustrated.   i am getting better, and doing fine.   i am in a fucking damn crisis unit.  i don\'t want to live.   but i want so badly to see my rainbow.       in.  between.     death is a s l o w,   s   l   o  w,  agonizing process.     \"just try, try, then try again,\"  '),(22720,664485,'12,October,2003','2003-10-12 00:00:00','   too many things on my mind and i don\'t know where to start.  '),(22721,664485,'10,October,2003','2003-10-10 00:00:00','   so to whoever the smart person that discovered that when you covered an area with a vinegar solution, the abnormal cells would turn white....fuck you.  because after that they take a biopsy of it, in which i had three samples taken, one which gave me shitty cramps, and two other that required cutting of a portion of it.   think vinegar on a freshly cut open wound.   exactly.  five people in a much too tiny room is a little too claustrophobic for me.   and i don\'t think the blood circulation has returned to ryan\'s fingers yet.  the procedure was this morning and i just woke up six hours later, very very sucky.   i am in pain and i\'m cranky.  '),(22722,664485,'08,October,2003','2003-10-08 00:00:00','   (because you deserve corny sappy shit every once in a while..)  for pixiedust, faith, and stupid reality shows.   for laughing at my silly, absolutely stupid jokes and wiping off my tears.   for comforting me in times of solitude, in times of fear, but most of all, for being there when i feel so absolutely alone.  for allowing me the strength and understanding to figure out love, and patience, and trust everyday, every hour, every night.  for inviting me into your home, your heart, for letting me love you and rediscovering what love is.  for eating every single last morsel of my concoction of chinese food, or western, or driving me to wal*mart at some obscene hour in the morning because i really need a whisk. and i really mean right. now.  :)  for surprising me everyday, beginning with being in my life, and helping me realize what happiness is.   for kinko, and shredder, and anjing. for listening to my shitty ass opinnions on what we should name the dog, even silent god. (we still don\'t have a name!!)   for the roses, the flowers, the i can\'t even pronounce the fucking thingamabob for our dinner party, for always looking out for me.   for driving me and picking me up for school, and rehearsals, and the once in a while visits to bum fuck alabama and everywhere i need to go in between.   for the master bedroom, the walk in, the full fledged his and hers bathroom, for building the bar, helping me make this place mine.   for excursions and surprises. for hiding me and having six thousand and seven people \"protect\" me from finding out the \"big plan\".  for the fireworks display(!) and stopping by god knows where to get them.   for talking to my mom, and trying to understand my culture, for learning chinese and malay and crazy stupid english words i mispronounce all the time.   for saying saya sayang padamu at the most absurd times.   for having pillow fights, and tickling matches.   for elmo cookies and downloading all my chinese, malay, spanish, god knows what songs.    i could go on and on and on, you bring so much joy, so much happiness, so much peace into my life. i don\'t even know where to begin and i think that even if i tried, i wouldn\'t do it justice.   thank you most of all, for LOCKING ME UP IN THE COMPUTER ROOM (!!!) because you have to \'set things up\'?   :P  i can\'t wait to spend the rest of the forbidden f with you.   i love you sexykins.   i swear this is the craziest birthday ever....and it\'s only been two hours!!!!!!  :)  :)  :)   p/s well, okay, an hour and fifty four minutes but hey, who\'s counting?  p/p/s you are one crazy fuck. :P  p/p/p/s thank you, thank you, thank you.   so very very much.   :)  terima kasih baby!  :)  '),(22723,664485,'05,October,2003','2003-10-05 00:00:00','   crashing. an hour ago i was high as a kite, excited about the new layout, rambling, smiling, then it all fell down. i fell down, and my tummy hurts, and i\'m bleeding again.  i started crying, and now i can\'t stop. and i\'m sobbing and sobbing and sobbing and everything hurts all at once and it feels like i am going crazy and a fucking terrified out of my mind. i\'m afraid, i\'m pissingly scaringly fuckingly afraid. i am just another shit terrified loser who\'s really just taking up space, it feels so wrong in here sometimes. i feel so wrong.   i turn nineteen in three days and for some odd reason, i am so unhappy. i am sad.   i call my mom and dad, and cry on the phone and ask them what\'s wrong. and beg to know why i can\'t stop, and beg to know what\'s going on in my body, why it\'s all fucked up and why is god so angry with me. and my mom tells me stories of when i was little, hoping to cheer me up, hoping to give me hope. she tells me of how i was so full of life when i was younger, she tells me about how i was always on top, and everyone loved me, and i was always so happy, and such a good kid. and every word she says pierces even harder, because i am a dirty ugly disgusting hopeless stupid child now. i hate me.   i am confused, and i shouldn\'t be. i am sad, but i have so much to be grateful for. i am fucked up and i am so angry. i am angry with myself, i am angry at the world, i am angry that i saw an old man walking home alone today and that hurt me, i am angry because i watched 25th hour and they had the tribute to ground zero, and i couldn\'t help but think that someone\'s mommy and daddy and koko and poppi and baby died there. died there because some motherfucker decided to be god. and i am angry at god. angry because i have no one else to be angry at, angry because i am angry with me.   i have this fear, this big great black wad of fear in me. there\'s so much i want to do in life, but sometimes i am so tired. and sometimes, i am so afraid i stop thinking and look for immediate escape, i suck at that but somehow, i am still afraid that one day i will finally succeed and it will be accidental. that i don\'t really want to die, i just want to escape being me a while. i just want to be okay, and breathe, and smile.   someone help me.  someone tell me everything\'s going to be okay.   i\'m in alabama tonight. so far away, so disconnected from everything.  i am lonely.  i am alone.  i am sick an i am fucked and i\'ve stopped taking all my medication because i don\'t think they\'ll ever work. and maybe that\'s why i\'m feeling shitty tonight, because i\'ve been off my medication for a week now. because i turn nineteen on wednesday and i am so unhappy. i go in for a biopsy on friday and scope. where they slice you apart and dye you and take samples and cut off and do whatever the fuck they want to you. i am scared, i am sad, i am alone.   no one understands, and it\'s not they don\'t try, it\'s that i\'m in too deep, too far down that no one can get me.   someone save me.  please?  still crying  still can\'t stop.  still hiding.   ryan gets off work in two hours, and i have to keep me from going insane is to tell myself again and again and again that killing myself wouldn\'t make it any better. that there are craszy stupid people that need me aorund here, that i can\'t just go. that goddamnit i have things to do and i am going to fight this dragon and i\'m going to be okay.   i am so  stupid  one stupid fuck.   fuck me. fuck me. fuck me.  does anyone really read this shit?    \"\"I can do this. Nothing in my life matters except this moment. Nothing in the world matters except this moment. I was born for this moment, and if I fail in this moment. I will die in this moment\"  -Raistlin Majere-\"  '),(22724,664485,'04,October,2003','2003-10-04 00:00:00','   a while ago..   we spent the night in with homemade magaritas (extra tequila and salt), my grand concoction of leftover-steaks-turned-into-fajitas-turned-into-i\'ve-had-a-long-day-at-school-so-it\'s-this-or-mc donald\'s and the whole she-bang of \'catch me if you can\' for the six millionth time on telly.   i let him clean up because i\'m good at whining like that but most nights he volunteers though. sometimes, we do the dishes together but the floor and our clothes probably get more \'cleaned\' that the actual dishes themselves.   we were talking about our childhoods the other day, something along the lines of not really being a kid or having a childhood and taking all that for granted. living it up now, and i see a lot of that in us. or where we are in life now.   maybe we have more time to slow down and smell the flowers, not as much a pressing need to safe the world and grow up.   some nights we are children together discovering the treasures of life, blowing bubbles in our backyard and piggy back riding on the bay. some nights, i see him stripped of everything and i see the little guy within that still believes in faith and pixiedust. and i smile.  i smile because he makes me smile.   because he holds my hand in public and makes it seem as if i am so important.   i smile because he makes my dark world seem not so scary sometimes.   i smile because he comes out with the corniest shit and the most silly romantic surprises.   i smile because he completes me.  so the drinks were great, the sex fantastic and the curling to sleep with a warm body holding you the best.  he whispers in my ears, i love you so much, racheal.  and i nuzzle closer into him.   i close my eyes and relax.  breathe in, breathe out.   a little bit of armani mania, and sensi, our odd retarded indulgence of his and hers, plus a little bit more.   worn out, arms lazily over chests and legs tucked in together; i fall asleep wanting to remember this moment forever.   wanting to make those invisible picture memories to store in that old, tattered book for a rainy day.      it\'s a quarter to four in the morning right now and i am absobloodylutely awake.   and instead of being the rude bitch i normally am and making him entertain me, i thought i\'d tap a post into the world to see how it was doing.  or how i was doing.  same difference.   :P  '),(22725,664485,'04,October,2003','2003-10-04 00:00:00','   hm...  :P  '),(22726,664485,'04,October,2003','2003-10-04 00:00:00','   so i slept in and had this great idea for my redesign together, it seemed to all piece together solidly, profoundly, in my little idealistic wannabe mind. there were splashes of sunny yellow, and different hues of colliding blues and blacks, shades of pinks and purples.   i emptied out the trash cans, and cleaned out the ash trays, wiped settling dust off table tops and danced alone in the cold of a newborn fall.  patiently, like waiting out the waves of nausea...or studying so hard your brains break apart for an exam, weaning off prozac, acting out a scene you know so fucking well because you\'ve rehearsed it seven million times before.   and i thought, if only, what if.  but i know i can be, i am.   a survivor.   i  am  at  peace.  happy even.  :)  '),(22727,664485,'03,October,2003','2003-10-03 00:00:00','   urlLink    urlLink    urlLink    urlLink      urlLink    urlLink    urlLink    urlLink      so i haven\'t written in a while, and she writes to ask me if i\'m enjoying the silence.   congratulates and wonders.   \"why no more comments?\"  \"why no more fire, and passion, and craziness?\"  \"why no more tales of drunken debauchery, and midnight beach football or waffle house endeavors?\"  \"why no more stories of the fairies and dykes?\"  \"why so quiet?\"  no time, i tell her. no time at all..  !!!  in between rehersals, and living in between, and settling down, and homework, school.   in between trying to get better, and constantly not having enough of blood to function  in between having summer turn into fall, and cooling temperatures.  in between furnishing, and painting, and photography, and writing, and singing, and dancing.....  in between making love, and sleeping, and doing laundry, and cooking.   in between take-outs and buffets, dollar menus and sit downs,  in between fiddler and the movie, in between life and everything that happens in between. in between living, and surviving, and smiling, and hugging.   we watch a movie in bed together, him with my croutons and crackers and his bloody mc chicken shit and me with my salad takeout and bluecheese. yes, that would be bluecheese and salad thank you very much.   me back on my half diet, determined to lose some god damn weight this winter, and me begging him to do the ironing. especially that white skirt with fifty thousand pleats in it.   at the doctors, the pregnant woman around me coo and giggle, congratulating us, saying what a fine young gentleman he is with his rose and my smile. insert hysterical laughter,  a. he isn\'t my husband.  b. we aren\'t expecting.   just in for treatment honey, just in for treatment.   but of course the appointments run over and the scary pregnant bloated ladies have joined us.   whoop dee doo dah.  and so i have conversations with expecting seventeen year olds and their fathers, forty year old women who look like they\'re in a lot of pain, and (equally) scared as fuck women.  and the boyfriend says, fuck yeah!  and i say, erm.   i don\'t (bloody) think so.  too young.   too inexperieced.   still a kid.   still want to achieve a lot more things in life.   but it\'s a nice start though.  a pleasant one.   building our lives together.   :)  i\'m back in alabama tonight, home for the weekend.   things sure are different around here.       urlLink \"wo hen lei >O\"  '),(22728,664485,'01,October,2003','2003-10-01 00:00:00','   first night in the new place!  :)  '),(22729,664485,'28,November,2003','2003-11-28 00:00:00','   happy thanksgiving y\'all   it was special, in a very simple sort of way....the ways things should be i suppose. uncomplicatingly amazing. right down to our bubble bath, our abundance of food to last six lifetimes and as always, the excursions. watching a movie together, curled up in the arms of of someone i love and of someone that loves me. i still find for the words to describe it, the little big small huge things. it\'s in his eyes, in his touch. it\'s in the way he moves inside me, the way he says i love you, i need you, i want you.   peeling potatoes with the moms, baking with the cousins...no chaos, no screaming, no fights. nothing like last year, no bickering of the other parents. it was a nice thanksgiving indeed. no pressure to be perfect or anything at all, in my jammies just being me. just being a part of the family. this is what it\'s all about i want to scream, i want to cry, i want to thank the great big guy up above again.   he talks about our basement, what it\'s going to look like. and we plan out the little things, picking out furniture and deciding on schools and what-nots. what i\'ll do when we get there, what he\'ll do. planning holidays and summers, white christmases and joyful springs. the doctors, the hospitals, the psychiatrists, he\'s got it all worked out in his head, all the little pieces coming together...being pieced together.   and they both look so happy. going home. and i know that feeling only oh too well.   it\'s funny how much can happen when you put four brains together. a big thing, i suppose almost like getting married. our 2.5 kids and the dog. got the dog covered plus jessica and rob. this is so strange, i don\'t know if that this feels so right that it\'s strange or if it\'s because the whole concept of it all just amazes me. me?  there\'s this warmth when he\'s near. just this insane amount of love he radiates. this is what love truly is, and everyday, even as i wake up to arms flung on my chest and his breath tickling my ear i thank someone again. thank god for this gift. him. us.   i am only nineteen and him twenty and in the back of my mind i am constantly worrying. what if we\'re just kidding ourselves? what if it happens all over again? what if? what if? what if?   this just seems too perfect...and it\'s as if i\'m waiting for it all to cave beneath me.   it\'s ironic how boringly in tune we are. honey, pick up some eggs on the way home. you never wear any bloody thing i buy so why bother? take the fucking dog out will you. you do the dishes and i do the laundry? ramen noodles, mac and cheese or mc donalds tonight?   i have all these hours in between school and when he comes home from work. and i clean, or organize, or humor myself...watch tv, paint, play scrabble as i wait for the magic hour. making dinner in between, trying to prevent it from getting cold, little surprises.   and he comes home with flowers and a smile.   and says hey baby, i missed you.   and i say i missed you too.   i missed you too.  '),(22730,664485,'27,November,2003','2003-11-27 00:00:00','   seasons changing...shaping up into things your heart and soul truly desires, the night before thanksgiving. going out to buy the perfect christmas tree, picking out tinsel and lights, lights from everywhere and anywhere spurting out your butt, purples and yellows, reds and blues, greens and whites. it\'s all so pretty, it\'s all so gay and you scream and jump up and down like a little girl. he says hop on my back, and he carries you up to place the star at the very top. all seven feet of it.   he smiles and whispers into your ear, leading you to dance to christmas grinch music, wrapping presents, making pies. life could be so perfect, life is so right, right now. here. together. us.   you want to cry and laugh and cry again. cry because it\'s so perfect, cry because it\'s so beautiful, cry because you\'re so damn blessed. he talks about your future, our future. we plan the move, and the house, and the everything in between.   it\'s like being lost for words but having so many similes to describe it. it\'s the i love you at intersections, it\'s the hugs that make you feel safe and secure, it\'s the talks in the middle of the night, it\'s the coming of two into one but still being your own person at the same time.   could i be this lucky?   coming home to a house of love, a home of us.   the fireplace, snow, the dog, the basement. dreams shaping up from absolutely nothing into everything. missouri, florida, kuala lumpur. here, us, now, this, everything.   and tomorrow the fmaily will sit at the table and the turkey will roast. the house bustling with activitiy, god there is so much to be thankful for. so much to just stop and say wow.   it\'s the realization that you\'re okay. that beneath all that\'s fucked up, beneath all that lies uncertain, unbalanced that there is hope. that this is a season of love and thanksgiving. that this is a season to accept.   it\'s strange to build your lives together. it\'s feels so damn out of place but right at the same time. the times your insecurity rears it\'s ass ugly face and says you\'re stupid. the times you don\'t want to think, or feel, or anything at all. but then there are the ones where he feeds you, and bathes you. where you\'re crying and sick and tired and he takes it all in with a smile. anything to make you feel better, anything to make you alright. it\'s the times you realize you are so in love with this man that you will do anything for him. it\'s the times you both laugh and cry together, it\'s the times i pull and he pushes, it\'s the times i encourage and he questions, it\'s the times we love together. discover.   thank you, someone, everyone for this. for everything that i am, can be, will be.   it\'s thanksgiving, and it\'s good to know that i have a list, that i have a prayer, that i have so much to thank for tonight, tommorow, the day after.   thank you mom and dad for invaluable advice. for the three in the morning hysteria, for still holding on, still pushinig forward. thank you for carrying out my hopes and dreams when i lost sight, thank you for all the hardwork, the physical labor, the emotional toil. i know it isn\'t easy having me as a daughter, but i sure as am ass thankful for having you guys as parents. i honestly don\'t think i\'d ever make it this far without you guys. thank you for letting go and holding close. there are times i think i don\'t know how, or where. and it\'s ironic to think now that i truly trust your decisions, your judgement. thank you for giving me wings to fly, to soar, to fall, to touch. everything i have learned, i have learned from the both of you. of compassion, of forgiveness, of selflessness, of joy, of pain, of respect, of everything i know. i love you both so much and it\'s crazy to think of the things i\'ve learned. thank you for being my best friends and my mom and dad at the same time.    koko, for tormenting me and giving me crap when i need it. for knocking reality into my mind and saying i can damn well fucking do it. for being someone i admire and look up to, for setting a good example, for playing with me. i keep your essay framed up to remind me of home, to remind me of all the love i have across the oceans when i\'m alone. thank you for being my partner in crime, for talking to the parents when i\'m being crazy or they\'re being crazy. for just being you, thank you for deep dark secrets and stupid retarded ones too. :) i\'m glad you\'ve found love, i\'m glad you\'re happy and you will always be my makr of approval because your opinion makes a big difference in my life. who the heck else am i suppose to pinch and irritate right? love you.   miss you guys very, very much too.   we were putting up the christmas tree and it was this melancholic strange feeling...like i was happy and sad at the same time. i was happy because i had someone special to put our christmas tree together but sad at the same time because i missed you guys. and i thought about christmas shopping, and parties. i thought about christmas eve service and christmas morning....wow.   it\'s about to hit midnight soon and i felt the urgency to write this down. to let myself go cry awhile, to let myself just sit in darkness but illuminate light at the same time. we\'re going to bed soon, and tomorrow will be thanksgiving. i am loved, i am confused, i am happy, i am okay, i am me.  just me you know?  '),(22731,664485,'24,November,2003','2003-11-24 00:00:00','   do you feel the breeze dancing in it\'s tantric display? do you feel the empty silence, from this bitter cold night? do you feel? do you touch? do you see? i don\'t know.  i don\'t.  '),(22732,664485,'21,November,2003','2003-11-21 00:00:00','   it\'s funny how your defenses are stirpped and you sit so vulnerable puffing on another stick of cancer. searching, trying, not knowing what to say and feeling a myriad of thoughts impound you.   going crazy again...the midnight walks, the mania state giving in to the depression giving in to the just pure madness. you wonder what goes on in deep within the mind, you wonder how your thoughts are put together. you scream at the voices but there is nothing concerte...the voices you hear are your own. the battle between good and evil, the battle between right and worng...the battle of miind and heart.   there are times, deep in the still of the night, you curl into a little ball and beg for some entity to release you. that or beg for the solutions, the answers. sometimes though, sometimes i don\'t think i really want to know.   seventeen million misdiagnoses, just trying, trying then trying again. but somehow those words don\'t quite bring you comfort much like they used to.   you have panic attacks, and they seem so real...so delibitating. so there. you can\'t breathe, it hurts, you can\'t think straight to save your life...just want to run, run, run. escape.   there are times you know you need to go back in again, but you can\'t. you can\'t because you\'re trying to save the world again...protecting anyone and everyone. you try to focus, to concentrate, to find meaning in something else but what else is there. you feel like nothing and empty and the dirty crud you try to brush off so hard from your soles. (souls)  physically you\'re fucked, and mentally too. so in your mind it makes so much sense to just let go. just give up. but everyone around you screams and screams. screams so much you don\'t know what the answer is anymore. heck, do you even know the question?  '),(22733,664485,'20,November,2003','2003-11-20 00:00:00','   scared, crazy, angry, insane.  '),(22734,664485,'14,November,2003','2003-11-14 00:00:00','   i performed for 2800 people over the weekend...yet out of three standing ovation curtain calls, not any one of them were of my family in the audience clapping, screaming, whooping. you know you search in a crowd for something familiar, something you know? you how you search for your dad so proud of his princess? or for the (good) tears streaming down your mom\'s face...you know how you look out for you brother, wanting to know what he thought, what he suggests, things you should do, things you shouldn\'t have done?  this is my second production here without them around..it\'s strange. it\'s sad. it\'s makes me want to curl up into a ball and stay there a while.   you know how there are people in your life you would do absolutely anything for? you know the people in your life that you want so much to make them proud? for me, most of the time, is me trying to do right. me trying to not screw up. for my parents....because a lot of times, at the end of the day, i think to myself that my parents don\'t deserve such shittiness.   and so there are days you take long long walks when you\'re in a manic state. impulsively wanting to fly and die at the same time. you sit outside eckerds and pizza hut too tired to move, too tired to cry anymore, too tired of life and you sit on the side of the road and count the cars that whizz by you. 100....200.....250..and finally he comes to rescue you, save you. screams at you, yells at you in the parking lot...WHAT IN THE FUCKING HELL ARE YOU THINKING?   i\'m not.   you can\'t explain it. can\'t explain the demons that corrode your mind, can\'t describe the empty harsh mean voices that churn inside you. but yet you tell yourself you can\'t be admitted again, you can\'t go back in there. you either die, or you fucking bloody make it. no in between bullshit racheal, you yell at yourself.   you cannot do this to the ones around you again.   and the doctors struggle around your head medicine and your body medicine. they struggle with drugs, then they struggle with disorders. they know you are definitely depressed, and that you definitely have manic tendencies too....they know you starve starve starve yourself, or that you throw up till you bleed. they even count the scars on your body, or ask again and again how this new bruise appeared, or cut, or ciggaratte burn. and you just shrug and say, i don\'t know. i was hurting. i\'m strupid. i\'m fine. and they want to evaluate you for borderline personality disorder, and they want to do this and that and this and that and at some point you scream. you break down. you give up.   the other docors drag you in because you have blood in your urine, because you are in fucking pain all the time, because your body, like everything else, is giving up on you.   yet there are good days. and fuck yeah, i mean they are good.   there are days you come home to a meal prepared, a night planned, a back rub and a place to come home to. there are nights with the dog and playing catch. there are afternoons with the kids, or evening at the shows. there are times with him, that i feel safe. alright even.  there are times.  and these times make up life.   me.  and it\'s never easy, never easy at all.  so you just try  and try  and try  and try.  '),(22735,664485,'09,November,2003','2003-11-09 00:00:00','   it\'s that feeling that nobody listens but everyone hears....something in the way the world looks funky around you, almost out of place. as if you\'re suspended in some god-awful reality, a nightmare pounding in over, and over, and over again.   so i haven\'t written in a while, it\'s not much different is it?   the emails come at first, truckloads by the day, and you have good well intentions to get to it, but you can\'t seem to face this, to face the real world, to face being sick, to face who you are, what you are. or maybe you\'re just lazy.   but slowly you fade off, behind thesewalls. slowly you disappear. you go day by day, and play your roles, push the buttons, go up and down, smile, laugh, cry, love, understand, feel, want, need, nothing, everything.  '),(22736,664485,'24,December,2003','2003-12-24 00:00:00','   (failing little shit.)  it\'s coming on so strong now, i can feel it in my very pores. the submission of antagonizing fear and there is nothing, nothing i can do. the tears they flow like torrents quenching a desperate drought. crying as if i could do this for the next five hundred years, crying so desperately, painfully. i don\'t know what i am anymore.   as i stand, one hand pulls me back to fall back and sink slowly into the ground. i am not me anymore. if anything, i don\'t even know who i am?  why do i do this every other day? every other month? why do i sit here and question, question, question? why is it sometimes just so goddamn hard to love myself. tell me, why? why can\'t i love me?  why do i hate her?  why do i hate the girl in the mirror so much?  it\'s as if you go through these emotions...sometimes fear, sometimes calamity, sometimes pride....and sometimes nothing at all. i am so ashamed of her, it\'s as if the distorted image in my mind is magnified by a hundred billion times more. and i want to do is tear her apart, and scream. scream. scream.   i hate you. i hate you racheal sarah khong chui mei. i hate every single shitty existence of you.  want to know the truth?  i\'m really not all the strong. really not all that i\'m cut out to be.   i am Nothing.   and sometimes, it seems as if nothing doesn\'t even describe it.  i am angry with myself. angry because i\'m not good enough, not smart enough, not pretty enough, not a good friend, not a good girlfriend, not a good daughter, not a good sister, not a good anything at all.   it\'s all just a big pretend.   i am angry because i am a sordid waste of space. i am angry because i keep on trying and trying and i swear i really do. i try so hard my heart breaks, my skin is calloused and i am so confused.   i am tired of this madness. i\'m tired of swinging between extremes. i tired of jagged jaded pills that don\'t seem to do anything at all sometimes. i\'m tired of hope, because sometimes hoping hurts the most.   when you tell yourself just a little bit more.  when you say it\'s okay. when you try try try try try try try.   you know what the saddest thing is?  it\'s that in between it all, some part of me still believes. and it is that very part that lets me down each and every time.   i want to run. i want to take a knife and slash myself into a hundred pieces. i want to take a hundred pills, smile, and maybe take a hundred more.   but then, i know as much as you do, that i will never succeed doing it.   never succeed in killing myself, if only just to prove my point that i am an utter failure at everything.   i don\'t know if it\'s worst to not have anyone understand or to have someone live inside this head too.   i\'m tired tonight. broken i suppose.   i\'m  just  so   tired.  you know?  the scary thing is that it\'s brought on by nothing.  that i am losing it.   that i am losing me.  who am i?  what am i?  where is the little girl that danced with the pixies and sang with the fairies?  where\'s the bold, rauccous, laughy, smily, girl?  where\'s the butt kicking, ass sprouting, sarcastic bitching hot stuff?  why does this fat ugly disgusting loser stare back at me?  why does this old old distorted sad girl look into my eyes?  why is this girl me?   i need to get out of my head a while.   merry christmas. happy new year. feliz navidad. happy deepavali. selamat hari raya.   for all the holidays i long to celebrate.   for all the memories i wish to live.  for all the me, i want to find.    a long time ago, i know i loved me. a long time ago, i know i was capable of love and being loved. i don\'t know what has happened now, i don\'t know what i am. and i apologize, i\'m sorry because i\'m a shitless piece of existence. i\'m sorry because i am absobloodylutely nothing. i\'m sorry because i don\'t know what else to be.   i\'m sorry.   i should go lose myself a while.  '),(22737,664485,'22,December,2003','2003-12-22 00:00:00','   i am a mucho dumbass.  lol, i lugged back all my camera equiptment...everything except for my fucking usb port connection.   fuck.  me.  :P      hey look! only 243 votes. shit. i\'d be lucky if i even have 4 readers?  ---  btw, is it just me or is it fucking hot as hell here?  the weather is obnoxious and i\'ve been sick hence, the silence.   shopping though, now that\'s another story.  did the gods of shopping look down on malaysia and just miraculously erected at least 4 new malls in the last three months that i\'ve been away?  madness.   ---  in other news, when the hell are we meeting up and going to party?!!?!?  i want kissies!  and love!  and hugs!  :)  '),(22738,664485,'17,December,2003','2003-12-17 00:00:00','   we left for waffle house down by davis at around 2:54 ...our waffle house. he said it made sense...a complete circle. it\'s where we always used to go when we were dating....the nights we\'d spend talking and talking, deliberating to the wee hours of the morning. everything and nothing all at once. you know how you can talk and talk and talk with somebody for endless hours and still have more to say?  or how you can sit in silence with one another and walk away feeling like you\'ve had the best conversation ever?  and all through the night he was everything you could ask for in a man. he carried my bags and rubbed my back. packed my shit and rocked me to sleep. wiped away my tears and smiled with me...laughed. i\'d somehow promised myself i wouldn\'t cry at  the airport. only three weeks racheal, only three fucking weeks.  but maybe it was the gentle way he held my hand, and swept away my fringe planting a kiss on my forehead. maybe it was the tender way he whispered i miss you already as we left the house. or maybe it was the way he listened the night before as i told him about the ex and our conversation and how i was confused and angry and frustrated and hurt and he took it like a man. he took it like the man i fell in love with and just listened.   because sometimes, that\'s all you want anybody to do...  listen.   i am on my way home. and there\'s something indellibly magical about it. i will miss him, and i will miss waking up and falling asleep with a warm body enclosing mine. i will miss our silly conversations and yes, even our stupid fights. i will miss making up and making out with you, i will miss waiting till three in the morning for you ot come home so we can have dinner together. i will miss having you call me beautiful and i will miss calling you stinky. :P i will miss your sexy lovely ass and i will miss holding your hand.   i love you so much and yes, i\'ll bring back a ring.   maybe.   :P  love you.   it\'s 2:14 in morning in singapore and i am blessed.  '),(22739,664485,'15,December,2003','2003-12-15 00:00:00','   there\'s something about listening to christmas music and a cup of hot, steaming milo (yes, milo...that\'s shitty cocoa for him and fucking excellent homemade 3-1 milo for me), attempting to study and avoid the slobbering kissies of my dog at this hour.  that and also trying desperately to remember what i need to bring back, what needs to go home and what needs to stay here, the seven million things i have to pick up before monday night and precariously balancing the weights of right and wrong, good and bad, sad and happy.   they caught saddam. and it feels so goddamn weird.   i\'ve been asleep for most of the day. i took my vistaril in the afternoon and pretty much knocked myself out for most of the day. i don\'t know what it was....maybe it was being alone in the house and trying to figure out the world, maybe it was the excitement and anticipation of going home, maybe it was the fear of seeing grandma alive for the last time, maybe it was knowing we wouldn\'t spend our first christmas together. maybe, maybe, maybe it was so many things.   and i wanted to cry. and laugh. and i wanted to sob myself sorry and laugh hysterically. so i knocked myself out. took an extra dosage.  for emergencies like this she prescribed.   and so i slept into and oblivion. the great thing about vistaril is that the nightmares go away.   and i\'m at peace.   i leave at four in the morning on tuesday....this is so strange.      urlLink  \"chrisraettig (01:25:15): \"adhere to your own act, and congratulate yourself if you have done something strange and extravagant, and broken the monotony of decorous age\" - or something like that. \"  '),(22740,664485,'13,December,2003','2003-12-13 00:00:00','   there are times i open up the door, and i am put into place by all the impounding darkness that reveberates out.   i had a sick scary dream last night....or perhaps morning rather. i\'ve had my days and nights mixed up lately...insomnia seems to come back and bitten my ass. hard.   the thing about dreams is that they feel so real, right at that moment. and you lay awake from it, shivering. asking...wondering what it meant. how it came about? what the fuck is going on?   and you\'re scared. oh god, you\'re scared. terrified. fear runs through you in an electrifying jolt.   i dreamt i was in the hospital again. and the nurses were cutting off my skin. i\'d been in a fire somehow, and i was struggling to understand what was going on. i didn\'t know how it\'d happened...and i was screaming. asking. panicking like hell.   anxiety attacks. anxiety attacks. anxiety attacks.   but there was this lady, she was so nice. so gentle. and she kept on saying it was okay. i\'d be okay.   and for some odd reason, fariz is in the room next to me. plagued by some unknown illness.   and ryan\'s angry with me. i don\'t understand and i keep on questioning but it\'s as if i am running into a wall.  i collect bits and pieces of what happened.   everybody tells me i\'ve hurt ryan. badly. i\'ve torn up his soul and served it to him on a silver platter.  but i don\'t understand.  i don\'t understand.   i tried to runaway.  and he chased after me.   as always.   somehow we ended up in the van, him trying to talk me out of my catatonic silences. and somehow, distracted, there was a massive car accident. and i was strapped in, stuck into a vehicle bursting with flame. but i don\'t understand, i still don\'t.  i wake up and i say don\'t ever leave me.  i wake up and he looks down at me like he\'s been doing it a while and is still smiling.  he says good morning beautiful.      i am so afraid of losing the best thing that has happened to my life.  what if he has another heart attack?  what if he gets into an accident?  what if by some god damn freak accident, somebody kidnaps him? sells him to thailand as a male gigolo?    and he tells me i worry too much. i always worry too much.   and he tells me he\'s the luckiest little bastard in the world.  and i hug him a little tighter and say, no, that\'s me.     he moves deep within me. higher, deeper, harder.   i  moan  scream with pure ecstacy.   he lines my every inch of skin with kisses.  i hold on to him tightly, tighter.  contracting my muscles, wanting him to be in me forever.  this feels so right.   i don\'t want to lose hime ever, i don\'t want to lose him.   am i weird or just stupid?  '),(22741,664485,'09,December,2003','2003-12-09 00:00:00','   urlLink      urlLink  \"When you cried I\'d wipe away all of your tears When you\'d scream I\'d fight away all of your fears I held your hand through all of these years But you still have All of me\"    he told me once, a long time ago that love was encapsuled into a tiny solitary clear glass. fragile. impounded. specially displayed.   and tonight, as my little city falls asleep, i am alone in the darkness left to reflect on a once forgotten dream. reality. memory.  it is about three-quarters after four, reaching five in the morning, and he lays next to me drifting into another realm where pixies fly and angels dance. i am suddenly drawn back to the memory of Him because the radio plays this hauntingly eerie reminiscence of him. i search for this file, flipping on mtv for it to coincidentally come into rotation. though how coincidentally that is in the overplayed, over rotated mtv world i don\'t know.   it\'s been a long time hasn\'t it? almost as if it was years and years ago. but i remember now, this would have been our second year.   there are times when ryan and i talk about our past. letting the skeletons out of the closets, opening a can of worms to be devoured and so many times it hurts to relive his pain as i imagine how absolutely sucky it must be for him too. yet sometimes, the jealousy gets the better of me. i want to scream and shout and be angry at him. i want to be angry with him because sometimes i know She still affects him in the crawling under your skin way it does. but how can i be hypocritical when i know he understands that i understand too.   in my past, the last couple of years, reaching two decades now...i have learned, grown, matured, shivered, crouched, broken, healed, mended...loved and most certainly lost.   but i don\'t suppose my story is like any other. like everyone else i am the girl in the middle of the night that lets her thoughts out every once in a while. more frequently perhaps. i am the homeless man that begs you for some change. i am the little tiny insecure kid that is the trouble maker. i am the reflection of me that criticizes and the inside of me that comforts.   heart against mind.   some psycho-therapy thing the ex-head doctor played with me once. i spent about two hours yelling at an empty spot on the wall, a space opposite me. it was supposed to be my mind, the evil person, the one that mocks, laughs, torments, criticizes and most importantly, breaks me down. and i spoke as my heart. my heart that fought and fought and was still fighting but was so tired. my heart that felt so broken, so used, so ugly, so not enough yet so hopeful. my heart that was so forgiving, so patient, so understanding. my heart that broke smaller and smaller into pieces as the madness continued.   the empty null void where \'the body\' was. the constant tugging, the extremities of highs and lows taking toll.   and i sat there crying, begging, shouting, screaming, asking, understanding, listening, painfully feeling as she helped me transition into trying to get it all. i mean for christ\'s sake, how the fuck do you understand insanity?  ryan\'s keeping a dry-erase board for all my classes, my doctor appointments, my head doctor appointments, my medicine intakes and notes. it\'s like his big post-it-note of all post-it-notes. it\'s so i stop missing my dosages and start getting better he says. it\'s so i get into routine, the normality of things. it\'s so he can take care of me and know damn well that he\'s doing everything he can damn well bloody do.   but yet i have fallbacks like the other night. worrying, worrying, worrying. another anxiety attack, uncertain about what\'s to come, how to feel. the world around me and where i\'m going. not sure because who\'s going to cuddle me to sleep at night. yet i don\'t want to be so far away from my family. striving to succeed succeed succeed yet trying not to kill myself in the process. understanding that i have shots, and medications and drugs and therapy but god damn it, how much more and how long more?  and so we laid together on the couch, him wrapped around my fetal position and let me cry.  let go.   deep breaths.  and he made me smile again.   you made me smile again.   p/s i\'m going to PEE IN MY JAMMIES!!! :) lol someone  urlLink voted  me  urlLink in !?!?!? woohooo! madness...you people crazy or just insane?  :)  thank you though.   *does happy chicken dance!*       urlLink  \"and you pray God no nightmares please in Jesus name I pray amen and the minute you wanna sleep dawn breaks you grin and say hey lets go run out to the lake and wait fourty five minutes for a complete sunrise and you see that God putting on makeup and you grin and say hey thats too much eyeshadow there dear and the cabbage butterflies flit around weeping willows their leaves tickle your back makes you giggle the silence broken by a single plop there goes the weaseley fish out of the lake\"  '),(22742,664485,'08,December,2003','2003-12-08 00:00:00','   urlLink     padding, gently, softly flow.   river rushing from a stream nearby at the crack of dawn in the openness of nothing and the emptiness of everything.  the bottom of my soul, singing in fluttered harmonics of in between and slushing breezes. a melancholic tune sung repeatedly over and over again into the nothingness of a clear open sky. humming, sunshine warms my fingertips, so cold....i shiver. embracing mother nature; talking as if she listens to me. take in the sweet fragrance of blossoming tulips and the dying everglades. the pixies and angels form a circle around my wasteless tired body and dance.   eva cassidy seduces me from the streams of notes and keys wafting from the stereo just beside the fireplace. i curl up and scratch my little puppy\'s furr, him riveting and moving. orgasming from me scratching his back, relaxing. buzzingly unclear from half a glass of wine left on our coffee table built from scratch. kingfisher blue and upholstered with plaid blues and tiny lined lime greens.   her scratchy smooth voice transcends me into the world of dreams and the half-awake slumber. the in between world of reality.  i taste bittersweet old familiar tears..a strange comfort.   still alive. breathing. knowing. loving.  this is how it feels to take in....and let go.     ...i\'ve wanted to write for a long time now. but it seems as if i have stumbled into a road block. thoughts and writing not flowing quite like it used too...it\'s as if the words are either tumbling out nonsensically or i am stuck for hours in front of a screen puttering around. knowing i feel the urge to write but having continuous stumbling blocks of sentences that are raggeddy and make me want to scream.   change is happening somewhere deep within me.   ...and again, i am afraid. undecided. is this the right path? am i who i think i am? am i making any sense at all?  so many nights we lay in bed. my head on his chest drowning to his heartbeat. reflecting on our day in the big bad world outside there....behind the cocoons of our four walls and feather topped bed with six thousand seven hundred and sixty two pillows. and sometimes he sings to me, or rubs my back. sometimes we talk about our day and our thoughts, feelings. sometimes i tell him about home, the colors, the people, the madness. and sometimes he lets it all go and tells me about his father. the verbal abuse, the beatings, the feeling of self-worthlessness and i feel absolutely useless in that moment. i can\'t take it away, and if i could, i sure as hell damn bloody would.   sometimes i feel guilty because i had a great childhood. the pain he trusts me with is so different. and deep inside, i know he is fragile and we cling on to each other drawing strength from one another. and i am thankful for those moments, no matter how painful it is to listen. no matter how indifferent he pretends to be. thankful because he sheds the tough guy happy go lucky act. thankful because he lets me be a part of him.   he told me once he used to laugh a lot when he was younger. and i see that same happy giggly kid when we pillow fight, or wrestle, or when i pin him down with my legs and tickle the heavens out of him. there\'s that glint in his eye as we laugh together.   and god knows, i am different too.   i am crustily oven-baked warm.     p/s why the heck does my cocoa puffs taste like peanut butter?  P/P/S : go vote for  urlLink joe ! :)   i. because i say so.  ii. because it would be a great honor.  iii. because i\'d like to believe that he\'ll smile down from above and say fuck yeah! :)  hugs   urlLink  \"That time has come again. You know, when life as you know it suddenly takes a dizzying turn which leaves your perspectives hazy and your reflexes dreamlike. Everything you engraved in stone is chipped away for an upgrade that, at this point in time, is still in beta stage. Excuse me, some might differ & prefer an alfa to better eccentuate the version. \"  '),(22743,664485,'07,December,2003','2003-12-07 00:00:00','   urlLink  \"im venturing out into the unknown. im flirting with the idea that my life could be different. im armed with nothing.\"     ever made love and cried?  ever had everything fall together so beautifully it hurt?  sighing to an end of a climax, coming down from an orgasm. spent, wasted...and hearing wafts of an old familiar song gently playing, soaking in the aroma of burning berry candles and flicking through images. there was something to be said of the motion, the sensitivity in that moment, as if something had taken me out sureally and i looked down from above.   once again, i am useless for words. except that it all came crashing down in a suspended fluffiness. as in it was bad and awful and horrible but absolutely amazing at the same time. i haven\'t talked about it, i cried once then shoved it to the back of my mind. my uncle had three massive heart attacks and passed away peacefully about a month ago. yet i know he still had so many things he wanted to fulfill in his life. it comes bursting out through his veins, his work, his eyes. and for a while i questioned, questioned anything and everything that i could question.   painfully obvious, death is a constant calamity that shakes my world.   he lasted enough for mervin to make that 36 hour ride home. and my dad told me about the grace of god, the beauties of life. about how the pilots made sure he got off the plane first, and customs cleared him instantly. he told me how my aunt was so strong, so strong. he said he cried a lot. that everyone cried a lot and how the church overflowed with people. people from afar and near, people he had touched and people he had blessed.   a couple of days later, my dad called to say i love you. and this was maybe about two weeks after c.s.u. where my thoughts and feelings and nothingness jumbled into a whole crazy spectrum. still fragile, still mending and breaking at the same time. not knowing, not sure.  ....and he said, i love you. and that\'s all that matters. that it was okay, come what may.   no lectures, no stop this shit, no get a grip racheal.  i live in a world where i constantly feel like i am nothing. where i feel like i am letting down everyone around me....but that night, listening to my dad and my mom in the background, i realized i could be nothing, empty, shitty, smelly and ugly and i\'d still be okay. that my parents would truly love me come what may.   it\'s hard to explain...almost as if all the knives in my heart begin to let go.   i constantly don\'t ever think that i ever deserve love. it\'s stupid, i know.    ...i have no doubt at all that my uncle is in heaven, perhaps watching over the kids and telling them stories like he used to do with me.   there are times where i am angry with god, mostly because i don\'t understand. i\'ll be the first to admit that my faith has somewhat wavered....that i go through days where i seriously doubt there is some entity above that looks over us, and others, kneeled down by my bed praying like my parents taught me a long time ago. the faith of a child, believing so faithfully and strongly that god will provide, that god will comfort.   i remember crying one night in the hospital...in the midst of a panic attack where my fingers had all clenched up and my sanity was seriously going. i remember that through it, in all in my head, i begged and cried for god. and maybe perhaps he did answer. and maybe perhaps i am okay....because look at where i am now.   i had nurses sticking needles, calming me down, scared of everything at the same time. it\'s hard to describe madness, it\'s this deep dark empty hole of paranoia and depression and extreme nonsensical highs.  i used to ask myself where i would be, what i had to do, where i was going.   and if i was serious about it, i have had no purpose nor direction the last year. it\'s like skiing down the black mountain when you don\'t even know how to ski.   c r a s h i n g.  but i have a plan now. and i have post-it-notes.  little things to say one step at the time racheal.  and maybe someday, i will save the world. maybe some day i will truly, really, be absolutely okay.  but for the in between here and there, i go to school, and write my papers. i tell ryan i love him everyday, so much that we represent one of those sickly couples that go i love you, no i love you more, no i love you infinity...la di da. :)  i try to talk to my parents at least once a day, check in. and truth be told, i am a lot more honest with the people in my life now. no more lonely seduction of depression. somehow, i seem to have gotten to a point where i am not so afraid of myself anymore. i don\'t know.   i\'m crazy.   :P  i don\'t know how i got to this. i had ryan cuddling me and the song had ended. his palm on my sobbing chest...and i felt the pressing need to write it down. writing calms me down in this strange way, i don\'t know.  and yes, i am going home for christmas.   :)    urlLink  \"It is a beautiful day and I would have it no other way. I need to take more time to stop, pause and look, reflect on the people, things around me before I go about my daily business.\"  '),(22744,664485,'06,December,2003','2003-12-06 00:00:00','   streaked crimson red hair, bleached with 40% white, a plastic cup and fork plus a whole lot of confidence(him) and bitchiness(me).  blonde to red in a night. silly stupid ditz to punk rock goddess. and the strange stttrraaannggee thing is that i feel like one too. a goddess. there\'s this store in seattle called the goddess in you, expensive as hell but amazing. there\'s this whole thing about inner beauty that people rant and rave about but sometimes, you just have to look it to feel it you know?  i don\'t know.  i\'m really just rambling.   it\'s about three hours past my bedtime but i am still fucking awake.   and let me tell you, it sucks ass.   especially since brilliant me went to ALL my classes today thank you very much.   playing cutesy with the old man is always nice too. what more with mothernature\'s boobies and old man\'s pimping. ah...the wonders of simplicity is very ironic indeed.  i meant to write this down the other day. you see, where the old man grinch lives...down below holds a child-care full of innocence and wonder and so every couple of days i find myself itching and begging to go there. get lost in the world of superheroes and barbie dolls a while, get lost in the world of the tickle monster and the my turn! my turn!  it\'s the little things that i enjoy the most. that undescribable, unexplainable feeling of having an absolutely untouchable kid reach up for you. it\'s the smiles you giggle with, the twirl of a strand of hair, the fascination in big grown up you in their eyes and the fascination of adorable sweet devils in yours.   it\'s the give a little, take a lot.   yes, i am selfish.  taylor says dance with me, dance with me! courtney says me too! me too! and justin runs and runs and runs around you, screaming, giggling, laughing. so happy, so free. alex coos, and you\'re lost in this magical world..where you make faces and act silly for a couple of hours. you smile, they smile, you wipe off boo-boos and smell up icky poos. you color and pick up and dust off and laugh. as samuel asks for his bwocolli, and says don\'t leave yet, don\'t leave.   (watching too much dr suess makes you start rhyming after a while)  perhaps the best part of it all is the end of the day. when they tug and pull and say not yet! not yet! or the way you pick them up after their nap, and they reach out for you. instinctively, you want to protect them from the big bad world outside but try as you may, you know it\'ll never be. and so you teach them little things, and so do they, hoping that they are equipped, praying that they will know.   sugar-rush from divinity.  precious moments, special hours.   beautiful friends, caring notes.   sticking a ciggarette, inhaling, smiling, and saying;   life is good.   :)     urlLink  \"Sometimes, you have to push a point. You can put up with an unacceptable situation for just so long then, suddenly, an additional factor becomes the last straw that breaks the camel\'s back. At first, you find yourself being surprised by the strength of your reaction: why have I rejected what I previously felt able to accept? You may begin to doubt your judgement. Are you illogical and inconsistent? No, you have merely reached a turning point. And you should be proud of that.\"     p/s yes, i know, what\'s up with all this happy, happy shit dude?  grr.  :P  '),(22745,664485,'04,December,2003','2003-12-04 00:00:00','   urlLink    urlLink    urlLink    urlLink    urlLink      i wrote. replied e-mails and notes and postcards. wrote and wrote till there were nasty blisters on my fingertips and my new manicure screamed at me as if to say, you stinky skanky lil bitch you! yet the ferocious bitchy unstoppable girl inside me snorted gave them the finger.   and in through it all, between midnight and the crack of dawn, the boyfriend comes home from work. tired, angry at the car but still so lovable. and there\'s a surge of unexplainable feelings that run through me. run to him.   bittersweet tears drying on the desk, heart, love, soul, poetry. so many bundles of hope and peace, so many bundles of care. thank you all you, him, them, me, us.   this world seems perfectly surreal at the moment, if not just gently amazingly hard. perhaps living in reality means taking everything with a grain of salt. perhaps living in my world, this world, this prison, heaven, hell, illusion, madness means just being me. being happy, being sad, being angry, being irrational and crazy and stupid all at once. perhaps to live is to run down the aisles of wal*mart at three in morning just because. perhaps it is watching the sun rise and set again, and knowing that tomorrow, it will do the same.   i love the tranquility of madness sometimes. nights like tonight anyway. where i sit here and look at my flaws yet smile. and say to myself it\'s okay.   he sits beside me on the other computer. bugging me about my obsessiveness with computers and the great big wide world through this little box of mine. a testament of faith, he is my testament of love, and trust, and understanding. i get it now, i dance, twirl, revel in it.   there are a million things to do, pre-christmas rush and seventeen other things that needed to be done yesterday. i shuffle from here to there, up and down. went in to see the head doctor in red today, she smiled brilliantly and said i was doing well. that she was proud of me. i had come to this revelation, somewhere in between breaking down and having a part of me die. realization. so i know there\'ll always be ups and downs, extreme highs and lows, and i know i can never just hold my sanity accountable. that i hold decisions in my hand.  i tell her about the not so much puking. and maybe it\'s the medicine, maybe it\'s the therapy or maybe it\'s just the feeling of being safe again. i am safe. and i haven\'t felt safe in quite a while. like i didn\'t need to perform or be perfect. like i didn\'t need to protect myself and the world around me. like i didn\'t have to justify everything i did, like i didn\'t have to tell myself that i was a person and that i was enough. it\'s strange how emotional scars come and haunt you again and again. it\'s a wonder how life weaves lessons in and out, through and through.   who am i?  i still don\'t know.   but it\'s okay now.  slowly, very slowly the pieces fall together. it\'s like putting together a 1228993983729857987 million pieces of jigsaw puzzles together. slowly but surely, the picture must form. eventually.   and when i die, i will smile from above...or below. i will say this was me, and this was who i was.   and i would be someone.   someone special, someone\'s someone, someone\'s anyone.   and it will be okay.   i will be okay.   i just haven\'t turned out yet...but i\'m still racheal. still crazy, still manic-depressive, still a retard. still a sucker for kids and old people, still bisexual and in love with the same man from a year ago. still a people-pleaser(his words) and a rebel (mine). and yes, i still want world peace and uh-oh oreos is still the coolest thing on earth.   smile, it\'s a beautiful day.   smile, it\'s cold as fuck but i have the fire going if you want to warm up a while.   smile, you\'re special.   loved.  :)  hug?   urlLink   \"The Muse...who still inspires, though not so often at 3 in the morning lately. She is renewing my faith in love these days. I notice her, and me, and us growing every time we shoot. (May you and the MovieMan always hear \"action\" - and rarely, if ever \"cut.\")\"  '),(22746,664485,'30,January,2004','2004-01-30 00:00:00','   i read a suicide post today. i don\'t know that strikes me as relevant as it strikes me hard. there are things that are sometimes so damn uncertain in life. the question of death versus life. will you take that journey on your own or do you leave it up to the fates above?  there are times you feel so invisible, sometimes helpless. and it\'s painful to realize emotionally all that you are, can be, will be.   i am no one.   --  i have to lose about a good 15-20 pounds for chicago this spring. there\'s something about doing the \'cell block tango\' number that makes me want to pee in my pants.   i am on the atkin\'s and frankly, i am grumpy as hell. all that meat can certainly not do you too much good.   when you have an eating disorder, it is a constant battle to stay on top. and when you\'re recovering, the battle is ten times harder.   it would be so easy to just not eat.   --  oldman, this has to be one of the most beautiful things i\'ve read in a long time.    \" Perifery   I have been reminded of late just how much we move through life so alone... How, despite all those desperate attempts at love and friendship...  At the end of the day...  We do it all by ourselves.  Existentialism is an interesting philosophical discussion, until you get your heart broken. Then you realize that, for all the trappings of social intercourse, it is true -  No one can take a bath for you.  Of course, having your heart broken has a way of making you kinder to strangers and small children, So, I suppose, even that has it\'s place in life.  Some things, though, I wish I could learn some other way.  Some things, I wish I could TEACH some other way.  I suppose, in the final analysis, we all learn...and we all teach.  The course is titled: Living Your Life 101.  I guess the thing to remember, as you continue the journey...  Is that, while you do make the trip... Trapped inside this \"mortal coil\" - itself working on dying from the day it\'s born...  ...While you make this trip all by yourself...  You are never really completely alone.\"    i needed to talk to you today. crying half-assed and having another nervous breakdown of some sort.   i\'ve been questioning myself again lately. i seemed to have lost my footing in the world.   i don\'t know who i am.   am i making the right choices?  i\'ve been figuring out all sorts of things about me, maturing you say, but yet it still seems so difficult. there\'s still so much more i want to know, want to hide.  am i any good at what i do at all?   is this my life choice?  will i be a two bit actress that waits tables 97% of the time and have that one passing crowd scene in an epic movie?  will i ever be able to confront my fears, be blatantly honest, have the patience to sit down and write my book?  and when i do will i be able to handle the criticism, will i be able to say i am wrong and you are right?  will anyone even read it?  there are times i  want  need someone to tell me i am fine. that i\'m doing okay. that i am who i am and that i am one hundred percent alright.   and i thank you oldman for all those one hundred billion times that you do so.   i truly believe that you\'ve helped shaped me into this slightly more confident person that i am through some nonsensical ass-grabbing and all that good shit.   your pictures say a million words.   and i thank you for showing me me through your eyes.   just as i am.  the barker and i are having small bitty problems that i know takes trust and time to solve.   you see, the problem is i hide. i run. i fucking hit the next train.   and he\'s all about talking, and working through shit. he\'s all about understanding.   in some crazy shitty mind of mine it pisses the shit out of me.   there are times i want to hit him and ask him why why why?  why do you choose me?  why am i the one you love?  why do you even want me?  and it\'s been so long but goddamnit it still hurts.   i gave my everything once, and i was me. i was me in every respect that i was a shitty person.   and frankly, it scares the hell out of me.   it scares me to remember.   yet i cannot help but do.   ---  i am hurting. and i don\'t know why i do.   i should be so happy.   i am so happy.  but it is sometimes so painful to be happy too.   ---  loonz, you are a beautiful person.   thank you for letting me be your friend.   call me sometime.  ---  i love you.   and sometimes that is all that there is to it.   i love how you remind me of that...that it\'s quite simple.   and i yell at you because i am angry with me, and i am angry at my world.  i blame you because i know you want the best for me yet i am so confized on my bullshit that it\'s hard to accept reality sometimes.   i have absolutely no right to scream at you when i fuck up my budget or when i decide today is too shitty of a day to get out of bed.   i have to remind myself that all because you are a part of me, that you still aren\'t me.  that you can\'t be responsible for my every flaw, my every shitting irresponsible action.  i love you, and i love you so so much it\'s hard.   i may be so american but i am still every bit malaysian.   and it\'s hard to be away from family, it\'s hard to be away from friends that know you inside out.   i am learning, and everyday i thank you because you support me in reaching for my goals, setting examples, finding me.   turning twenty, being adults, growing together.   they say the first year is always the hardest, well, we\'re one fucking year and it\'s still pretty damn good.   and i\'m not just talking about the sex.   --  the drugs aren\'t really working.   i\'ve been sick the last couple of days again.   i hate doctors. and i hate chemicals. and i hate this whole notion of being dependent on pills for the rest. of. my. fucking. life.   --  zigi is the most adorable puppy ever.       \"did god make you, them, without a heart, without a soul?\"  '),(22747,664485,'25,January,2004','2004-01-25 00:00:00','   accoustics.  '),(22748,664485,'25,January,2004','2004-01-25 00:00:00','   urlLink    urlLink    urlLink    urlLink    urlLink    urlLink       urlLink    urlLink    urlLink    urlLink    urlLink    urlLink     time passes by almost too fast and too slow at the same time. stuck in between the hands, staring up above into the ticking timebomb of emptiness and hope.   i had to figure out how old i was today...for some odd reason i couldn\'t remember my age.   the years have gone by and yet nothing(something) perhaps everything has changed.   i am nineteen years old, three months, god knows how many hours and minutes.   --  i don\'t know why i am writing. i remembered a promise i gave to myself in my bravado years. i promised myself i would write a book and be published by twenty-one. really, like hell that\'s going to happen.  our computer room is in a seperate addition from the house, and lately, with winter biting my aneamic ass in the butt, i haven\'t found much inspiration nor passion to come into this room. somehow the world wide web out there just seems so disconnected to who i am now. it\'s a part of me i almost don\'t know. names and faces, people and places.   it\'s as if you\'re on this never-ending journey to find out who you are, your purposes, the meaning of life, of you.   why the hell exactly are you here?  why, of all people in the world, did the entites up above give you life? let you stay? what are you going to do today that\'s going to justify the death of someone more precious, more special, more worth it than you.   worth is a word i absolutely loath.   worth is a word i live by.   ---  i have somehow morphed from my gothic phase and am slowly blossoming into the punk-rock madness.   finding myself.   finding myself.  finding.   ---  the old man has been extremely good to me lately. my surrogate father in america. called, begged, whined on to do fatherly things that are oh so important like finding the perfect curtains at ten at night.    \"if you can\'t find anything at wal*mart, it\'s not worth buying,\"   once in a while though, the boyfriend and the old man gang up together to attack. like waking up at nine in the morning for school. or taking my medicine. or stupid responsible shit like that that they just seem to like to bug me about. :P  we\'ve also seemed to have exchanged rooms with bethany. it\'s a fresh start i must say, out with the old, in with the new. meshing our tastes together. calculated decisions instead of when we first moved in and just combined all of our furniture together, regardless.   bright blue and green curtains, rainbow striped bedsheets, our hullabooloo kick-ass entertainment set, little bits of me and him pronounced.   i\'ve also somehow slipped into the role of homemaker. it\'s a good role though, a little odd but nice nonetheless.   i make dinner, he cleans, we do lovey-dovey shit together. it\'s cute. :)  i like this, i am afraid i could fall in love with this life....i think i have already.   i am so fucking in love it might be blinding me. it\'s good to be blind though.  we are learning to grow together.  ----  my big fat ass and waistline has also taken a turn.   i am a chubby little shit at the moment. the medicine gives me the munchies and my boyfriend feeds me like i am from afganistan and am starving.  i have some damn good looking boobs at the moment though, fuck yeah fatness.   :P  there is a tiny little voice in me that still bugs me from time to time, so i alternate between the atkins and cheesecake diet. i think i might kill myself this way, but ah well, i\'m crazy anyway.   ...doh.   --   \"nobody told me that you only needed time to fly,\"   i think i\'m doing good for the most part. i\'m a little off from time to time but it\'s okay.   there are nights where i still wake up and cry for grandma and there are still the night demons that keep me up with insane requests and mind pounding stupidity. yet i know, for the most part, is realizing that this will past. that things will get better and things work out with time.   it\'s this intricate giving and taking that tires me. i am so absolutely sick of being sick but this is me.   that girl in the mirror, me.   that smily ass bitch, me.   that stupid retard, still me.   i cannot run away from who i am, nor can i ever expect to just be anything less.   the power of \'be\' amazes me from time to time.   i\'m doing okay...how about you?  '),(22749,664485,'02,January,2004','2004-01-02 00:00:00','   i leave tomorrow.  it feels as if this trip has been too short and too long at the same time.   i remember only leaving yesterday, and suddenly, i am leaving again.   i\'ve been back for about two and a half weeks now, though only half of it spent in malaysia.  i had a great christmas for the most part. something tells me i will always remember this one.   i played with heng yuen and taught him how to high five....i remember a while ago when i used to be so afraid of him yet now i am fascinated. fascinated by the way his mind works, fascinated by his own little world in which his glass murky doors are encaved.   i remember family and it\'s value. the emptiness, as if something was hugely missing from the table, from the festive season, from the occassion. wanting to say something, but not knowing what was safe. i missed you ku-chang, i know deep in my heart, that you\'re shining down on us. you were a big part of christmas to me, a big representation of singapore and my ku-chang. i still miss you, we all do.   i remember grandma and her voracity (sp?) yet there was something so fragile and vulnerable about it. she was so happy, literally like a little kid on christmas morn. she ate and ate and ate, and smiled and never once complained. she asked to see the christmas lights and so we drove her down to orchard road. and we talked, or rather i mostly just held her hand and hugged her and kissed her. she\'s so old now, i see the wrinkles on her hands and face, i touch her translucent skin that just coats her thin frail legs. she\'s almost bed-ridden now, they tell me, she doesn\'t smile much anymore. but this christmas though, this christmas, she was happy. happy because everyone was there, big and small, far and away, australia, taiwan, america, shanghai, kuala lumpur, london. we all rushed back to see her. maybe even for just one last time.   she had a heart attack the night we left. we came back the next day and two by two, we masked and watched her through glass windows. wanting to say fight on grandma, fight on....yet the other part needing to let her go for she was tired. wanting to see some sign of hope, some sign of life. wanting to see her take her own breath, kick a little, perhaps open her eyes. the little girl in me says don\'t go just yet grandma, i\'m scared. and the other part says, it\'s okay now, we\'re okay. it\'s all going to be okay. tell grandpa and kuchang and the kids i said hie okay?    reunited.   i mustn\'t be so selfish.   back in malaysia and picking up things and putting them away. rushing to meet seven million people before i leave, holding on so tightly and itching to be home with my baby at the same time. so much left to do, the unexpected interruption, i was only supposed to be in singapore for four days. and we shelved ourr plans for new year\'s, and we took long long exhausting trips. yet i know, i\'d do it all again. i\'d do it all again because she made me cakes and goodies, and fed me and clothe me, and loved me and all sorts of wonderful beautiful things that she is.   i should go pack my bags some more, cram in perhaps more gifts and brahim\'s mixes.   i am excited but weary.   my parents are amazing people you know? it\'s funny how it takes your almost nineteen years to figure out how much you love them and how much they love you. it\'s in the little things...it always is isn\'t it?   my brother had his first day of work yesterday. and i know this bond of sibling hood that no one can break. the one where my mom tells me about how i have no one else but him if they\'re not around. and that\'s why the both of us are around. he\'s the more protective, cautious one and as always i am the crazy, slightly delusional brazen one. the one that touches the hot stove while he rushes to puts ice on my burnt fingers despite him repeatedly cautioning me against so.   it\'s 6:56 malaysian time. and the house is beginning to stir awake. my parents have special plans and we\'re all going to do family stuff today. they\'ve taken the time off and itr\'s strange to feel so special. from the way nicholas says i wait for you to take a shower-lah to you call me whenever you get a chance and we go wherever whatever-lah. to archana\'s standing open invitation. bryan flies in tonight, and maybe we\'ll meet up at two in the morning for laughter and stupidness. sakthi and her goddamn vinna classes but coming anyway, that damn girl and the indian blood in her that drinks like a fish. the few couple of us, good friends. always. forever. inside.   fariz, whom i\'ll always love and treasure. good times, great memories.   blogger friends...meeting winnie for the first time and their crisis of needing to get new clothes? meesh, grace, ash, jikon, dinesh, najah, shahrizal, tv, and i know i\'ll never name them all. there\'s just so darn many of them, and just so many darn things to love about them.   jae, my hungarian goddess whom i haven\'t seen and probably won\'t get to.   so amny things you know, so many pieces to tie together.   can\'t wait to go home and cuddle with my lover, can\'t wait to play with the dog and hand out christmas gifts! can\'t wait to unpack and tell all my crazy stories, can\'t wait to talk to three in the morning, can\'t wait to eat his homemade mac and cheese. can\'t wait. can\'t wait.   yet i don\'t want to leave.   i do. i don\'t. always broken, always torn apart.   that\'s just life isn\'t it?  hugs  '),(22750,664485,'01,January,2004','2004-01-01 00:00:00','   death reminds me not so subtly again;   life is Fragile.   just came home from singapore, i reek of antiseptic and damn clean hands.   wrote some paper entries while i was away.   perhaps i\'ll post them up later.   i leave on sunday for a long flight other home.   another place, another piece of my heart.   this is strange with a capital S.   anyway,   anybody up for something or another on saturday night?  perhaps one last soireee (however the fuck you spell that) or a big ass fuck you to 2004?  we\'ll see. (please come)  hugs.  '),(22751,664485,'28,February,2004','2004-02-28 00:00:00','   Sunday | 15 Feb 2004  No one. Nothing. Nevermind. Part 4. I\'m disappointed by something that is superficially so trivial, but that inadvertently, though not innocently, became inbued with meaning and significance to me.  I am pining for something that was never there. That was never real. That seemed promising but was never manifested.  I yearn. I yearn.  There is a hole, a gap waiting to be filled but that now looks like it never will.  It hurts.  I don\'t know why, and it seems so stupid. But it hurts.  I wake up in the morning and it hurts. I get some quiet time and it hurts. My heart aches for a loving hand that was imagined. And it hurts.  It\'s been two days and it still hurts. Even though it was never real to begin with. Because it was never real to begin with.  As the hours tick by, it will never begin to be real.  It hurts. I hurt.  Posted by jikon at 09:43 PM    we\'ve been fighting so much lately...and it hurts. it hurts not because of just the things we say to each other in the heat of the moment, but the implications of it. the little voice in me that says it\'s over. run. don\'t get hurt. take some time. run godfucking racheal run.   he snaps at me and i don\'t say a word, i say nothing at all. and it infuriates him to a degree that\'s unexplainable till he bursts with more anger, more daggers into my heart. and i can\'t eplain it to him, the need to close up, draw in, stare off into space...he doesn\'t get it. doesn\'t want to get it, hates it when i close up so tight and shut him and the rest of the world out.   he yells at me sometimes, screams and it\'s so loud my ears hurt. i hurt. and sometimes i scream back, but i only get a sentence in before i crumble into pieces because i am crying so hard. then he\'s angry he\'s hurt me, and i\'m angry i\'ve hurt him, and we\'re both so angry and so hurt. he makes a joke, and i still want to be angry and hurt and fucked and yell but i smile. and i stifen a grin because it\'s silly anyway.  and we are back to normal lovey dovey shit again.   taking it from one extreme to another, sharing our lives, us, together.   my mother tells me that no relationship doesn\'t ever not go through the wringer. that there are ups and downs, goods and bads. and i understand.  i understand beyond understand.  but there are times where it\'s so hard to just hold on and not give up.   but then there are times where this relationship is as magical, as romantic, as insane as the little girl dreams you had.   there are so many questions, and so much pressure, and so goddamn many things that i need to do, have to do, cannot do.   i am lonely, and loneliness is so painful.  i miss him, sometimes even when he\'s next to me, and i love him.   i love him and so everyday i tell myself, it\'s okay.  i\'m okay.  and i know this when he does the little things and appreciates my little things.   love isn\'t all peaches and roses....love is accepting, learning, trying, giving, forgiving.  hold on tight and never let go racheal, hold on so fucking tight because much as you run, this is the best thing that has happened to you ever.  '),(22752,664485,'24,February,2004','2004-02-24 00:00:00','   Friday | 20 Feb 2004  Relationships I: Men! When I was a child, I thought that one day a man would come and sweep me off my feet. He would be tall, not necessarily dark but certainly handsome. He would be wise and witty. He would be charming and eloquent. But most of all, he would be strong, protective, loving and caring. He would be a real man. Not a boy. Not a child who wouldn\'t grow up. Not a man who requires constant attention. Not a ditzy queen. But a real, big, buffed man.  As I grew up, I began to realise what a fallacious fantasy this was. This man did not exist. Or if he did, he was so deeply hidden in the haystack that I\'d sooner find the needle before I found him. Instead, life yielded men who were in the closet and troubled; men who were married and were trouble; men who were in trouble and couldn\'t fend for themselves let alone me. And then, there were men who were just plain trouble.  Then sometimes, you meet a man who asks for honesty and truth. A man who would offer a sincere relationship and expect the same. You talk, you get intimate, you discover and cultivate the chemistry, only to find out just as you\'re about to cross the rubicon that he is no better than the rest: a coward, an idiot, a basket case and a master of dissimulation. You\'ve been nice, kind and honest. But being nice and kind doesn\'t get you into bed with anyone. In the gay world, being gym-fit and a bitch does.  The problem with men really is that they are men. You can\'t live with them and you can\'t live without them. They are insensitive, inconsiderate and in a world of their own. If they can help it, they\'ll fuck your brains before they fuck you. There are days, and perhaps even weeks, when men being themselves are enough to drive a gay man straight.  And then you actually meet a nice guy. At least, so he appears on the first meet. He\'s got a nice deep voice and a kind heart. He makes you laugh and remembers to send you SMSes when you\'re down. But then, deep inside, you know that such good men don\'t exist. A good man is an oxymoron. And this good man is straight and ostensibly unavailable.  So the cycle begins anew. You don\'t learn from past mistakes because to do so would be depressing. To do so would suggest that there\'s no hope in men and that we all ought to give up while we\'re ahead. And that would not be good. That would take life out of the living.  Posted by jikon at 02:00 PM  '),(22753,664485,'24,February,2004','2004-02-24 00:00:00','   slipped between the cracks of esquisite looking spanish american familes, somewhere in between i promised myself i\'d find it. some niche of everything that brings meaning to life.  we fought a lot through orlando and back, i don\'t know what it was about, snapping at each other, fighting about nothing then another then something else. something off between the both of us, cuddling, fighting, making love, screaming. exhausting.  exhausting to a degree of numbness.  crying so much, so goddamn much. so tired of the doctors and so much goddamn fucking pressure.  i\'m tired world, i\'m tired of you and this and everything.  we watched a diesel blow up in flames, as rob called 911 and everyone sits open mouthed gaping at the fire across the interstate. god bless somebody a little voice whispers, no survivors, burnt crisp and into nothing. for about 30 minutes we sat and watched in silence. imagine what death must be like burning alive.  then there was disney and epcot and walking walking walking. turkey drumsticks and china, mexico, america. there was aurora and little girls all in awe and the magic of disney you wish to capture and tuck away into someplace safe. there\'s the tiny balance of not getting shit faced in the hotel but nicely buzzed. the smell of alcohol on his breath as we\'re tlaking in the middle of the night about life and malaysia and figuring out details. goddamn details.  there is so much going on and no holding on point. there is so much wrong and changing and healig and hurting. \\there is just so much fucking going on.  '),(22754,664485,'24,February,2004','2004-02-24 00:00:00','   and every fucking year, every fucking other month we\'re back here again.  always paranoid, always worrying, always not sure.  what if? what if? is it? holy mother hell shit me now. to be honest, i wasn\'t expecting what she found yesterday...i thought it would be the same routine; blah blah blah, take your medicine, you\'re fine, come back in three months.  i\'ve been looking up the possibilities of what it might be, i really shouldn\'t...but then again, curiousity did kill the fucking cat. i\'m cursing and angry because if i wasn\'t, i\'ll be crying and depressed.  actually, i am. i don\'t know. goddamnit, i\'m just so sick of it you know? so sick of all of this, and everything, and always having something bloody wrong.  and this would be the part i whine and say why me? why me? why me? what in the godfuck did i do now? i\'m not nervous about the ultrasound, procedure wise, i am nervous about the implications of it all.  i am nervous that something is wrong. and i am nervous if the thing that is wrong is a big wrong.  i\'m popping vistarils like candy because i will tear my hair out soon, i am anxious and anxious and anxious and i\'d like some fucking answers. some great life plan or another.  and all these obstacles along the way, my doctor fucking prescribing the wrong bloody medication. the fact that i leave for disney tommorow, our relaxation, a sorta honeymoon of kisses and beauty and wonderfullness and it is fucking pouring outside and it\'s a wet as fuck weekend and everything is wrong wrong wrong. i promised to make my chocolate pudding cream cake angel whatever the fuck you want to call it tonight.  this should be a good day, a good fucking year for christsake. and i wish these bloody obstacles would STOP COMING. that\'d be nice for once.  i\'m one foot into a nervous breakdown only that i cannot afford to have one now.  i have way way too much on my plate to fuck things up by going crazy at the moment.  i am fat. my medicine makes me fat. and it is disgusting.  the old man keeps on smiling and saying you look so healthy. and everyone goes ooh ahh, you don\'t look like the starving aneroxic third world kid anymore and i want to scream. i cannot fucking stand the fat rolls, and the pudgy face or my fucking breasts.  i\'ve been fighting with ryan about diet pills, he doesn\'t think i should mix all my chemicals like that. that it isn\'t safe. and i want to scream but can\'t and yell at him about how hard it is.  poor racheal can\'t fucking look at herslf in the mirror. poor racheal wants to cut cut cut. poor racheal wants to throw the fuck up but that would be so bad. so bad. i\'m not bullimic anymore, god damnit, i am not.  i do not have an eating disorder.  i am not crazy. i am not going to die of cancer.  fuck me.  i am so angry, and i can\'t explain this rage, this monster in me that\'s screaming so loud.  i cannot explain why i\'m so high and so low and so high then so low again. i cnanot explain the very essence of who i am.  i cannot explain what i want to do with my life, yet i know i still want a life. i cannot explain being lonely. how do you cope with being 9134807093275093722054830982 miles away from your family to someone that doesn\'t have one? what do you say when he asks you why you\'re sad all the time? how do you febreeze and clorox away the tearstains you leave on the pillows and his shirt every night, or every other night, or every bloody second. how do you explain that counting cars is very. important. right. now.? how do you tell him you don\'t want to eat the pills, that you\'re so fucking sick of it, that it doesn\'t help, that you\'re just fucked up, that the pain will never go away, that you\'re just fucked as shit? how do you listen to songs that you wnat played at your funeral and not cry? why do you want to die and live so badly at the same time? how do you explain love and thank love and never ever want to let go? how do you be so goddamn needy and suffocating and so goddamn unindependent and pull away at the same time. what do you say when he says in sickness and in health means in sickness and in health? how do you explin to someone that they are just fine, that they are okay, that they are everything and more when you most probably don\'t even know what that feels like? how do you convince someone? why? what? how? who? so angry. so sad. so...i don\'t know. everything.  '),(22755,664485,'20,February,2004','2004-02-20 00:00:00','   Ki System   Natal Year number:  7  Although Racheal may have taken a considerable time to reach maturity, she is nevertheless a quick thinker whose ideas change as rapidly as her conversation.  She\'s an individual who wants to be liked, and as such finds it hard to refuse requests.  For the same reason she dislikes arguments, and wherever possible will compromise to avoid conflict.  Not the sort of individual to be employed in litigation, but excellent in mediation.   Natal Month number:  9  Although Racheal may be relied upon to start her allotted tasks, she cannot always be relied upon to complete them.  This is not because Racheal is lazy, but because her active brain has alighted upon something else that she wishes to pursue.  Nevertheless she has leadership potential and the ability to inspire others with her charm and humour.   Natal House number:  3  With a high energy level, and ambitions to match, Racheal will show early promise.  Unless her tendency to expel this energy on too many projects can be overcome, a promising beginning may soon evaporate, giving way to anger and frustration.  Racheal is someone who enjoys a good time with an active social life, and needs the direction of more sober minds.   Comments based on Western Grid      11        Racheal is a balanced individual who can understand and appreciate the views of others as well as her own.  She can express herself fluently, and with persuasive conviction, whilst taking account of opposing views.  This means that Racheal will get on very well with colleagues at work, both senior and junior, as well as clients, customers, suppliers and external agencies.       4        A practical individual, Racheal is a neat and tidy craftsman who can organise others to achieve her plans.  She should use those practical skills and organisational ability in an office management, or secretarial function, or by applying such skills to self-employment.  In general, Racheal is a neat and tidy individual - qualities of value to many different jobs.       88        A perceptive individual, Racheal takes little on trust.  She maintains fixed views which once established are difficult to change.  Conscientious by nature, she will make a good businessman or accountant, but being naturally perceptive, and unwilling to accept views of others without solid evidence, she may well turn to philosophy.       9        Ambitious to improve her lot, Racheal will constantly push forward to achieve something in her life, yet this ambition will be balanced by humanitarian ideals that will lead her to support causes that may demand self-sacrifice.   [The Arrow of Scepticism: lacking the numbers 3, 5 and 7]   Taking little on trust, in adult life Racheal requires solid proof before she accepts a concept.  Characterised by honesty and fairness, she is quite intuitive, yet firmly sceptical of the claims made, for example, by religion.  This may be important at work, if employers require employee acceptance of work practices that are not immediately verifiable.   Comments based on Chinese Grid      11        Racheal will be reasonably successful in financial terms, and enjoy a degree of luck.       44        Rather stubborn, Racheal is inclined to believe that she is always right.  But she enjoys high intelligence, and often thinks that she is superior to others.       8        In the absence (in the Chinese version) of the numbers 1 and 6, she will be good with money and details.  She will be happier if the Chinese grid includes the numbers 5 and 7.       99        Racheal is intelligent, and prepared to learn.  Without the numbers 3, 5 and 7 in the Chinese version, Racheal is likely to use her head at the expense of her heart, which can lead her to become something of a workaholic.   [The Arrow of Apathy: Lacking the numbers 2, 6 and 7]   Racheal lacks the will to succeed.  Somewhat indecisive, she will miss opportunities and avoid risk.  Racheal needs to recognise that life\'s chances must be grasped with both hands when the opportunity arises.   [The Arrow of Loneliness: Lacking the numbers 3, 5 and 7]   Racheal is so determined to achieve her objectives that she does so at the expense of all else, including friends, family, love and laughter.   ---  Ki System   Natal Year number:  8  Ryan is a steady, responsible and reliable individual who dislikes change for change\'s sake.  He is a solid family person who enjoys tradition, but is inclined to say exactly what he thinks.  Such frankness can occasionally land him in trouble.  Behind a somewhat ponderous exterior lies a deeply sensitive soul whose core is rarely exposed to public view.   Natal Month number:  2  Compassionate and supportive of others, Ryan is something of a perfectionist who will find it hard to delegate even routine tasks.  Ryan is hard-working and good with details, but to get the best out of him, Ryan needs to know that in the end his efforts will be properly rewarded and acknowledged.   Natal House number:  2  Believing in his own abilities, Ryan\'s attention to detail will lead him to success through hard work and his own unaided efforts.  He can however work well as a member of a team even though he will rarely, if ever, delegate responsibility without stringent conditions.   Comments based on Western Grid      1        Although Ryan may be reasonably talkative in public, he finds it difficult to express personal feelings to those closest to him.  In employment terms, this inhibition is of little consequence.       333        Despite an excellent brain, Ryan is governed by his powerful imagination.  Inclined to be petty, Ryan is best employed as a writer or actor/actress.  The problem that Ryan faces, is his belief that imagined facts are true, when in reality these so-called \'facts\' are nothing more than a figment of his imagination.  But for works of fiction, those imaginative powers are invaluable.       8        Good with details and with a methodical approach, Ryan nevertheless is someone who enjoys constant new challenges to maintain his interest.  He is a methodical individual with attention to detail, but is inclined to leave tasks unfinished if his active mind alights on something of greater interest.  Routine tasks that fail to challenge his intellect are at greatest risk, so Ryan needs a job that offers variety.       9        Ambitious to improve his lot, Ryan will constantly push forward to achieve something in his life, yet this ambition will be balanced by humanitarian ideals that will lead him to support causes that may demand self-sacrifice.   [The Arrow of Frustrations: lacking the numbers 4, 5 and 6]   Ryan finds it difficult to accept other people as they really are, and is constantly disillusioned with life.  Once he accepts that others are only human, the frustrations will cease.   Comments based on Chinese Grid      111        Ryan enjoys a positive and balanced approach to life and will be financially successful.  If the Chinese version also contains 2 sevens, Ryan is very fortunate.       3        Unless Ryan also has in the Chinese version either the numbers 5 and 7, or the numbers 4 and 8, he will be hypersensitive and prone to stress-related problems.       8        In the absence (in the Chinese version) of the numbers 1 and 6, he will be good with money and details.  He will be happier if the Chinese grid includes the numbers 5 and 7.       99        Ryan is intelligent, and prepared to learn.  Without the numbers 3, 5 and 7 in the Chinese version, Ryan is likely to use his head at the expense of his heart, which can lead him to become something of a workaholic.   [The Arrow of Suspicion: Lacking the numbers 4, 5 and 6]   Ryan is a cynic who worries too much about the downside.  He needs to recognise more of life\'s advantages.   [The Arrow of Apathy: Lacking the numbers 2, 6 and 7]   Ryan lacks the will to succeed.  Somewhat indecisive, he will miss opportunities and avoid risk.  Ryan needs to recognise that life\'s chances must be grasped with both hands when the opportunity arises.       urlLink http://www.idealpartner.org.uk/ipbi/WhatJob.asp  '),(22756,664485,'20,February,2004','2004-02-20 00:00:00','   it\'s always good when your doctor prescribes you incredible amounts of medicine that you are deathly allergic to.   i really should sue.   or at least think about it.  it\'s days like these i thank someone up above for having my boyfriend read the labels before he feeds them to me.   and then i curse them the fuck out.  --  i\'m in a strange happy mood, nevermind that i was hysterically freaking out last night but that\'s a different matter completely. something in the lines that i almost fucking died.   i\'m allergic to ibuprofen.   so take notes everybody, mix a little bit of the blue pill and i\'m as good as six feet under.   i know this because one fine night when the tylenol ran out i stole some ibuprofen from bethany and a couple of minutes later i stopped breathing and my eyes swelled up to disgusting proportions. on the other hand though, i know what i\'ll look like if i ever was a fat ass bug.  so case in point, in nice bold red sticky paper on top of my chart has a list of medications that i am allergic to.   la di da, going down her list, she prescribed me prescription strength motrin.   so we go and get my precriptions filled and then fucking realize that motrin is bloody ibuprofen.   at 10:30 at night when i am writhing in pain.   exactly.  doctors are so fucking stupid.   apparently one of my ovaries are enlarged, and of course, they don\'t have a fucking clue why.   so in a knapsack,  i almost died yesterday.   (nevermind that our hospital is right down the street)  i have another fucking bloody thing growing in me.   and my boyfriend is one hell of a patient supportive bastard.  --  ultrasounds next monday, we\'re leaving for disney tommorow, we found our dog shacked up in a bloody animal resort spa thing, my mom comes in two weeks and i am a fucking retard.   --  i did have one thought though, imagine what it would do if i took it all?    wa ha ha ha ha ha.  '),(22757,664485,'20,February,2004','2004-02-20 00:00:00','   Today is one of those days that is hard to function. It’s as if my body and mind is split in two and my mind is just somewhere in between….or perhaps no where. There are many days that nothing makes sense and I am constantly fluctuating from manically insane to downright suicidally depressed. There are no answers, there is no cure…only understanding…only trying, then trying again. I am in school, sometimes it is good, and sometimes it is bad. Part of finishing my degree is to take a general psychology course in which I have mixed feelings about. It’s hard to explain a mental disorder, it’s even worst trying to make sense of it. There are times that I am embarrassed because the lonely seduction of depression can be a painful thing. To sit in class and sometimes be angry because the others are ignorant, or to feel alienated because you understand just exactly what your teacher is talking about…perhaps a little bit too much. Yet there are other days where I feel like a sponge, wanting to soak up possibly everything and anything that I can learn about the psychology of things. Cupid versus psyche, seratonin versus sanity, brain versus matter.   I remember a class in which we played Deepak Chopra and the immensely special bond of your mind and body. The unexplainable lengths in which your body might actually be healed by your mind…positive thoughts, positive thoughts, positive thoughts. I asked a question, one which I hold truly to my heart; is it possible to get off my medication and just let some inner resilient strength in me cure me? He answered no.   We are perhaps only half way through class, and I still have many questions unanswered. Will there ever be a way to solve these questions without returning back to pill popping? Can we just talk it out? Brain transplant? Anything at all? You see, manic depression isn’t just like a cold; it doesn’t go away in a couple of days and if anything it is a stupid disease. Stupid as in it makes you want to do and do stupid things. There are times where you feel no amount of medicine will ever help, and times where your medicine takes a weird turn and pushes you towards the ledge. I do not want to get off this ledge.  There was a paper that I wrote at the beginning of class that forced me to explore the areas of clinical psychology. With a little bit of force and help from the class, I have discovered that I am a good person and that this isn’t a disability. I don’t think I would ever pursue being a psychologist but the thought was nice, to be able to live that dream a while. To truly believe in myself and the world around me that this little crazy 19-year old Malaysian girl could do something about it.  I am an international student and I come from a country where mental illness is a hush-hush. Your visits to the psychiatrist is a James Bond mission, one that takes a lot of planning and secrecy. In America, you ask, now why in the world for? The same reasons we never ever say anything offensive to our government in public, the same reason we are more open but ignorant in a lot of ways. I have enjoyed this freedom, I have enjoyed quenching my thirst of knowledge, I have been comforted that being manic-depressive isn’t such a big bad thing at all.   Learning is a wonderful thing, minus the parts of course where you can’t concentrate or that your butt wouldn’t get out of bed. This general psychology class that I much dreaded has become interesting to me. I love that I do not have a teacher that gives monotonous lectures and instead employs other techniques. How would you have ever guessed that you could learn about your brain through chocolate-chip cookies? Our teacher gives examples and listens, yes, even to my roommate’s silly boyfriend that comes out with the most absurd theories. It makes my jaded perception of the world a little less cynical.  There are a lot of smart people in my class, and what they say about people inspiring each other is true in this. I don\'t talk much about me in class, mostly, I sit behind and just sorta hide. It inspires me though to listen to other people’s life and apply their techniques or draw strength in their courage in some weird way. I am not alone, and that is a very good thing.   I have learned a lot in this class. Things that I will probably carry with me to my grave, some about people, some about me, some about nothing and everything.   Perhaps that is the stepping stone that college gives you, a time in your life to explore and educate yourself. Take what you have and fly.  I suppose general psychology wasn\'t too bad.    yes, yes, another crappy selfish lets talk about me paper.  '),(22758,664485,'19,February,2004','2004-02-19 00:00:00','   he had a dream about you the other day, you\'d laugh at this when i tell you. i don\'t know if it\'s weirder that my boyfriend is dreaming about my ex-girlfriend or that we both had dreams about you in the same night.  i\'ve been leafing through blogs again, catching up, reading bits and pieces of different people from around the world...and it\'s funny how certaon inflections and simple words can have such a big pow wow on your day.  i missed you.  oh wait, i said that already didn\'t i?  i feel as if i am at a turning point and stuck in the same exact position at the same time. my sexuality, the person i am, my connection to the big wide world.  who exac tly is this girl living in pensacola?  lets sit down and have some coffee a while, biscuits, cookies, whatever the hell you call it nowadays.  happy valentines day...was it good?  '),(22759,664485,'18,February,2004','2004-02-18 00:00:00','   it\'s dark in here..there\'s really nothing much to do except work on my research somemore or maybe attempt to make something to eat. i haven\'t been too well lately, the pain is back and magnified by about a billion times. if i had nuts, school work have squeezed them fucking dry. my finals for my math class is fast approaching and i cannot help but worry worry worry myself to death. there\'s something about being a perfectionist that\'s soothing. it\'s like cutting yourself, one of those so wrong but so right things i suppose.   i can\'t seem to get out of this habit of picking at the little things. really.   my next appointment is on monday. i.....don\'t know. let\'s not talk about it.   my puppy\'s missing and we\'ve tacked up signs all along ninth avenue and back, sasquatch doesn\'t seem to give a damn and ziggi just enjoys the extra fucking attention. damn we live in a selfish world.   i really should switch on the lights in here, being in the dark just makes me want to tear something or yell. i can\'t decide yet.  there\'s evil spicy fish curry sitting in the fridge that i need to go clean up.   we also did about seventy two loads of laundry last night in honor of our new steel contraption wardrobe and so we have seventy two loads of ironing to do.   fuck.  me.  .....  i am such a terrible terrible housewife.  '),(22760,664485,'18,February,2004','2004-02-18 00:00:00','   big beautiful opacent twinkly dreams that come shining through a foggy night. winter turns around and the cold that bites through your skin vanishes with the comfort of love. bleeding, bright crimson fluid that flows from your pores...make it go away.  hurting, crying, sobbing..make it go away, baby, make it go away.  he holds you tight, tighter. humming. humming makes the evil pain go away.   this world is fucked, you rock and fetalize. fuck you, fuck all of you.      \"There\'s nothing you can do to make anybody feel better and it\'s going to hurt for a long, long time,\"  '),(22761,664485,'18,February,2004','2004-02-18 00:00:00','   better and bored.  '),(22762,664485,'18,February,2004','2004-02-18 00:00:00','   weak,he feeds me my pills, little blue and white, solid greens, and plain white ones. one for this and that and this and that.   another appointment on monday, another hand-hlding nerves, and more nerves.  taking another biopsy.  he whispers, we\'ll make it.  '),(22763,664485,'16,February,2004','2004-02-16 00:00:00','   courage mustered somehow to make that call. an empty stare, the bits of silence that are uncomfortable. a sudden shift into a memory of another phone call, pain. hurt that sears through you and you wonder. relive those moments, and you wonder why he brings it up. oh no wait, it was me. stupid me.  i didn\'t know. i wanted you to..to come, to say it was okay, to say you still cared, to love me, to need me as much as i did in that very moment. to say you would. to say that i was worth every bit that hassle of a three hour drive. perhaps you would never know, perhaps you would never understand what it took, or how much i gave to you. perhaps you will never know how much you meant to me..once, a long time ago. yesterday. perhaps you will never know just how much you still can hurt me. fight back and swallow the tears, it\'s only a stupid scrabble game and only a stupid conversation. all those dumb techniques about breathing in and breathing out just seem like soapy flacky duds of bullshit in this moment. why? hide, and cower with hands above me and shielding. this isn\'t real, i am not here again. sink in, i don\'t want to remember. i want to hide, and cry, and be alone, and just. not. remember. i want to protect and shield and close up and pretend all over again. silence. that bitter empty silence again. fighting every fucking day to justify why i should be here, why i am here, why the battle is worth the fight.  upset. remembering.  you wouldn\'t have come anyway. i didn\'t ask, but yet i waited. like every birthday before, or performance, or something significant in my life. two years. nevermind. nevermind racheal, it\'s all in the past. lock it and throw away the key. store it in that big gigantic yellowing life book. another lesson learned. another though dissected. another fucking memory. im okay. it\'s all fruit loopingly peachy.  '),(22764,664485,'16,February,2004','2004-02-16 00:00:00','   sixteen things and seventeen that don\'t make sense. another doctor\'s appointment, the aroma of french vanilla soothing...nothing and everything in between. nothing is everything that seems it is.  words that splutter like a dying engine, splintering waves of clear cut edgy bullshit. daily everyday stresses and fresh cut roses from our backyard.  dinner at los rancheros and fajitas that make your cheeks burn and your face red. alcohol desperation but medication restrictions. on and off and on and off.  buddy lists and friends and exes and strangers. conversations and more conversations. pictures, new hair, very red, very short. more valentine gifts.  everday is a valentine. talking in short insignificant gibberish sentences.  cold. blankie on my freezing toes and mood color-changing nail polish. blue and purple, blue and a little purple, blue and specks of silver glitter and white. ashron and unreal and marvelous three and corgan wailing you to sleep.  \"there\'s so many of these crazy little things,\"  '),(22765,664485,'16,February,2004','2004-02-16 00:00:00','   goodnight crazy world.  '),(22766,664485,'12,February,2004','2004-02-12 00:00:00','   it\'s funny how the rain drags you down but warms up the temperature temporarily. is god crying?  '),(22767,664485,'11,February,2004','2004-02-11 00:00:00','   surviving cancer is strange thing. i watched an episode about self-euthanasia (can\'t spell it, i know) last night,  that hits me in all the places it really shouldn\'t.  ryan and i talked about it a while ago, our baby girl\'s middle name is going to be akari. i still miss her, but it\'s all good you know? it\'s all good.  there is a heaven, i know that, i\'m sure. there must be...and in this heaven place are all the people i love, and they are all happy and free and it\'s all good. ....right?  '),(22768,664485,'11,February,2004','2004-02-11 00:00:00','   sometimes.  '),(22769,664485,'11,February,2004','2004-02-11 00:00:00','   seeking approval is a strange thing. being thrust a gift is even weirder. remarkable. this time crying, but crying good. good tears i mean...very good. a big sense of relief, of acceptance. a big sense of suddenly feeling okay...a calmness wahed over and a river flowed. friendship, a gift. thank you.   thank you.   thank you.   thank you, like sorry, doesn\'t ever seem to be enough sometimes.   a strange surreality, was this real?  did it really happen.   you know me well indeed.  '),(22770,664485,'11,February,2004','2004-02-11 00:00:00','   churning, something deep within me spirals back down into that lonely pit. it\'s as if invisible barriers put themselves back up in place and i am taken away into solitary confinement. this place where my mind just digs in deeper and deeper, scratching it\'s paws against the soils of depression and it\'s a pit in here. i wish i could cry, sometimes, it\'s hard to even do that anymore. no, i don\'t mean that...i mean cry. like really cry. baby sobbing careless humiliating disgusting crying. i miss those days on days like today. where crying it seems, would make things better.   i miss..this. everything. i don\'t know. i miss not knowing or knowing too much. i miss her, that girl, the bitch whore slut cunt. i miss identifying with her, being able to put a label on how shitty my existence is. a drip of this and a drop of that yet uncontrollably, everything is still as is.   long days, short days, terrifying days, nice days. it\'s like a nightmare that repeats itself over and over again, here i am, hello, my name is your bitch. i am manic-depression.   fuck you manic depression.   insecure. how can it be to feel so secure and insecure at the same time. high and low. happy and sad. right but wrong.   i should go to bed. insomnia only feeds to the mania, and it pushes me further down into this sinking sinking drowing feeling. hopeless.   deep jarring red words scream at me in my mind. they pound on me and drive daggers in...piercing. bleeding. crimson so fucking red. pain. why?  why me?  why this?  as always, the self-pity and self-destruction begins.  as always, tommorrow is a new day and i will try, try, then try again.   as always.  '),(22771,664485,'10,February,2004','2004-02-10 00:00:00','   cross-legged, thin...very. there\'s a certain immaturity i recognize, that same bright eyed aspiring bullshit i had when i started. i watch her as she takes her first few steps...i know that look. that confident insecurity.   i try to convince myself that she isn\'t me that i\'m looking at...but there are so many similarities. i hate it. i hate her. i\'m jealous. i\'m jealous because i recognize that strange happiness you have when you\'re in between...and i think that\'s where she\'s at. it threatens me in some strange weird way, she tries to reach out and i snap back at her. i know she seeks my approval somehow, in this fucked up mess i am in, she looks up to me. and it\'s strange. shit, strange doesn\'t even begin to describe it.   i could put myself in her place, i had that look once. i am comfortable, yet comfortable has a rather sinister ring to me. was i happier then then i am now? no, not really. i cannot possibly have been....could i?  this brain numbing epiphany that hits me like the raindrops that pour harder and harder against my back. i am running, for cover maybe...or just running. salvaging my wet dying disposing ciggarette.   no, no...don\'t turn into me.   there are more things too, things i refuse to admit right now because my pride is too messed up to be fucked with right now. things i don\'t want to face, much less admit.   but i know it\'s there.   i know that\'s why i don\'t like her.   can\'t stand her.   i know why i get irritated when she talks to me, or tries to me my new best friend.   i\'m trying to psychoanalyze this...shit, i\'m always up for helping someone. anyone. anybody out of this same exact path of hell hole i\'ve lived in...but there are too many similarities. i don\'t know if i can help her. i don\'t know if i can help me.  holy fucking mother of god shit.   this is scary.  scary strange.  very strange.   fuck.  '),(22772,664485,'09,February,2004','2004-02-09 00:00:00','   he plays weird music on the stereo. some jiggy ba-doo-dah bullshit from the seventies. at least it sounds like the seventies anyway. we\'re going on a trip...an \'adult trip\' as jessica puts it.   the room is strangely cold in here, yet i am warm. how do i do that? manage to feel warm and cold at the same time?  we made love on the couch last night. suddenly aroused from watching law and order, don\'t ask me why. i stroked his manlyhood and rubbed it against my breasts. feeling.   felt him deep inside me and moaned. soaked in all that was us as he whispered i love you, i love you. harder, softer, faster, slower...now, please, now.  right.  there.   we\'re going for a trip together in two weeks. it\'s our first racheal and ryan holiday together. yum.   love must be this...listening to avenue q together and laughing our asses off. calling to say i love you. picking up, dropping off, and picking up again. love must be laughing and crying in bed together, love must be waking up to the one you love. or being jolted from an awful screaming nightmare and feeling safe secure arms around to soothe you. love must be making mash potatoes and meatloaf turkey, grocery shopping, planning a life together. love must be waiting at the doctors office with nerves and more nerves and having someone share it with you. love must be about buying your first set of plates and owning a home together.   love must be about giving and taking. sharing. learning. forgiving. yes, love must be most of all be about forgiving. for the little things, the big things and the smaller ones too. love must be about knowing when you are wrong and when you are absolutely damn right but able to take the fall anyway.  love is about patience.   love is about experience.  love is this.   all that i know.  glass cases of mixed sand and warming together on the beach.   love is a warm cup of coffee in the morning with bagels and someone to share it with.   is this love?  is this right?  does anything that i say make sense at all?  love is about dating your best friend.   good or bad.   naughty or nice.  stupid or silly.  love is about wanting to stay forever.  and ever.  and ever.  '),(22773,664485,'08,February,2004','2004-02-08 00:00:00','   shopping therapy is a very good thing.  '),(22774,664485,'08,February,2004','2004-02-08 00:00:00','   The Color Of Rain  Saturday, 7 Feb 2004 || 2004_02_07_2330.html     So what is this called again? What is the color of rain? What is this wisp of hair on your face that falls and waves so gently as you talk, but I hear only you soft muted voice blending into the. The..? I dont hear anything anymore. What is this lovely glow from your face that is illuminating my life? Then I hear you speak, that sweet voice asking me, as you touch my fingers briefly. The color of rain? you ask.  Well, it has always been pink for me.  You hold me in your arms and I inhale, filling my lungs with so much of you, so much more life than ever before. She turned around and gave him a smile while the rain continued to patter patiently outside.  Love, he whispered softly into her ear.   The color of rain for me is Love. Yours.  Thats how I\'ll always remember you.   ---   Looking Back Now  Thursday, 5 Feb 2004 || 2004_02_02_2230    It\'s not worth turning back for.  I went through my collection, leafing through the pages, and I saw yours. I touched it gently, yet it crumbled under my fingers, just like how you. I picked the fragments from my bed, careful to not leave any sharp pieces behind, careful to not get myself stabbed again.   And you said. Hey. I turned around, and I saw you. Hi. Come here, give me a hug. I stepped forward. Pangs of pain shot up my leg, and blood seeped between my toes.   You smiled. I didnt.  ---  I am at peace with myself when I am alone. Its called solitude. I like it. Is it cold where you are? No. Am in my apartment bedroom. Up so early? Havent slept. Stayed up in bed making stories in my head. Sleep. You need to rest. Sleep. Stories nice. Stories by the beach. Am always happy in the sytories. Nice. Sleep is secondary. Yeah? Then make stories with eyes closed. Thats what I do. Stories behind my eyelids. Happiness comes alive. Very nice. Yes, have a great day ahead. Hugs. You too. Smooch. Blush. :oD   --    it\'s been a while since \'ve read you...actually, it\'s been a while since i\'ve read anyone at all. nestled in this little comfortable secure cocoon...of blocking out people and faces and pain...or so i think.    you, he, they, remind me once again that the memories never fade. you step one foot out your safety net and it\'s scary as hell. it\'s scary like the little demons are waiting to pounce on you and devour you alive. you inspire me. it is safe to talk, really. feel. write. i haven\'t written for so long...no, i don\'t mean that. i just mean that i haven\'t written really for so long.  there\'s this marshmallow in between that prevents me, i don\'t know how to unleash the skeletons from my closet.   thank you for the direction.  '),(22775,664485,'07,February,2004','2004-02-07 00:00:00','   a wave of nausea hits just right and i am praying to the porcelain gods again. it\'s as if you\'re fighting, and fighting....but sometimes, you win a little and lose a lot. a drawback, another one of those two steps forward and about sixteen steps back things. but it\'s okay, it\'s okay goddamnit, i am alright.   medicine. medicine out of my ass and in through my brains. pills for this and that and this and that. blood red nails run through the bottle cap, fuck me, it would be so easy...so fucking goddamn desperately easy. but i won\'t. fuck you racheal and your insecurities, and your stupid ass thoughts, and your stupid assholic manic depression, fuck you racheal, you are stronger than this. than all this pain, than all this self-pity destruction. you are bigger, better, less ignorant than this.   angry.   so fucking furoius once again at the girl in the mirror that stares back. you fat little shit face. you ugly useless moronic bitch. you good for nothing whore dick sucker cunt. and you can add as many expletives as you want, you can scream and draw and cry and fist up, but the overwhelming need to hurt yourself will always be there. you are better than this, racheal, you are better.   if you want to go down, you know you\'re damn well fucking good at it. you cna close up, protect, cry, spiral back into that empty dark place again. and i know now for a fact that you will never breathe, never come back up again. because your body, your mind, your soul can only take so much violent crude abuse.   so there\'s the part of you that\'s always struggling harder and harder to find the surface and not drown. it\'s that fragile fucking delicate balance of making the right decisions, doing okay, trying, being happy. it\'s you against the chemicals, against the enormous anmount of pressure.   you think way way too much racheal.   it\'s that part of you deep inside that knows above everything else, that you\'re really okay. the tiny little part inside of you that screams to be heard, that says just take your goddamn pills and shut the fuck up. the part that says you\'re a good person, that it\'s going to be alright. that you\'re really only nineteen and so what if you haven\'t found the cure to cancer or saved the world....so what?   it\'s the only part of you that\'s sane enough to say, you\'re just being stupid. or that you\'re overthinking again. it\'s the part that at least somewhat tries to shush the voices that blatantly scream at you, it\'s the little part that struggles to know that you are capable of love, that you deserve love, that you are enough, that you are just plain okay zipdodeedoo.   i\'m doing real well in school, making straight a\'s and all that baloney. but the over-achieving perfectionist in me always yells back in me that IT\'S NOT ENOUGH! somehow, i\'ve convinced myself that i need to be this and that and this and that.   i am fucking in love with a man that adores me and loves me with the whole damn universe and takes me for who i am. and i don\'t know why i am still not happy, not recieving, not able to relax or convince myself that it\'s not all going to crash in the next second. that i can and will recieve this gift, that i should be loved, that i am an alright person to love.   i don\'t know why it\'s so hard to believe someone when they say that i am a good person, because deep down inside i feel rotten and ugly. i feel like god\'s gift to the demons to torment. i feel like shit.   and i keep telling myself that someday i will forgive myself, i will let myself go to the universe, i will be free.  bullfuckingshit.   i will live with this for the rest of my sordid amazing life.   i will live for the rest of my life having ups and downs. living the extremes. i will face good days and bad days, i will never ever forget how to cry or hurt or laugh. i will always be angry or bitchy or sarcastic or funny or adorable or cute. i will always be taken from one side to the other to the other.   i am anxious, and i am nervous as hell. it\'s as if i am always at a crossroad, always making decisions...finding myself and i cannot quit worrying. i cannot stop the \'what if\'s\' nor the constant questioning on where my life is going, who i am, nor who i\'ll ever be.   i cannot stop the blackness inside of me that has lived for so long just to say, you\'re a piece of shit.   it\'s really just called self-pity and i still revel in it and say, i cannot do it.   but i can.  i can goddamnit.   life is so fucking confusing and fun and exciting and miserable at the same time.   fuck me.      \"\"Effortlessly\"  I won\'t be so hard on me today I start to take myself so seriously Shouldn\'t be so hard just to be effortlessly It shouldn\'t be so hard to keep it together It shouldn\'t be so hard to say the right things to you It shouldn\'t be so hard just to be effortlessly  [Pre-Chorus] Weightless worries fall away Wasted pools of energy I want to know I want to breathe  [Chorus] Effortlessly - just be I want to be effortlessly I want to be  It shouldn\'t be so hard to be inspired It shouldn\'t be so hard just to write this song It shouldn\'t be so hard to be wrong or to agree It shouldn\'t be so hard to change the world It shouldn\'t be so hard just to change your mind It shouldn\'t be so hard just to be effortlessly  [Pre-Chorus] [Chorus]  I won\'t be so hard on me today I start to take myself so seriously Shouldn\'t be so hard just to be effortlessly  Effortlessly I want to know I want to breathe I want to simplify my needs I want to live inside this moment And just be effortlessly Just be I wanna be effortlessly I want to be\"  '),(22776,664485,'07,February,2004','2004-02-07 00:00:00','   he lays sleeping next to me, arms curled around my shrinking waist...and i tilt just right so my tears escape him and wet the pillow instead. i listen to the night sounds as always, listening so hard for sounds i don\'t even know...waiting to be amused. comforted. soothed. the tears fall and yet i\'ve mastered the fine art of crying softly. almost unheard...unseen. the voices in my head ravage and attack...they say i am ugly, worthless. they tell me i am a piece of shit. they break me down, and tear me apart. those voices are you, those voices are you in my head. you screaming, and yelling. me hurting. why?  stupid fucking voices.  he moves in closer, as if sensing my neediness. i cry a little harder, i cry like a little girl, rather recklessly unafraid of being caught. being caught of letting down my defenses, revealing that i still hurt, that i am still fragile, that beneath all the toughness and smiles, that i am still broken.  b ro ken.  '),(22777,664485,'20,March,2004','2004-03-20 00:00:00','   you left before i could say my peace. i\'m not mad...i\'m not even really hurt. if anything...maybe i feel like i\'m at peace. like...i can finally maybe start to heal? i don\'t know. don\'t be sad. you\'ve done nothing wrong. i think it\'s obvious after all of this who the bad guy really is. i read through all of this racheal, and i realize that...i don\'t belong anymore. all i am is a burden. you don\'t need me to fuck things up. i\'m sure ryan doesn\'t need me around, fucking with his girlfriend. i know you love me...but is there really a way we can all coexist? it doesn\'t seem likely. i love you racheal. i love you with all my heart. i know it doesn\'t always seem like it...but that\'s because i\'m shit with my feelings. don\'t worry about me...you might think i\'m gone, but i\'m not really. i\'m never that far away. and even if you don\'t see me, you know i\'m here. you know the hard way that i have problems letting go. congratulations on making it. nobody\'s happier than me that you have the life you completely deserve. i\'m just sorry that i\'m not a part of it. but...that\'s also for the best. don\'t be sad anymore...not over me. you\'ve cried over me far too much already. i\'m not worth your tears. i\'m not worth anything.  i\'m sorry for raping your private space...but it\'s the only way i know to get ahold of you. but...you don\'t have to see me anymore. i\'m so sorry.  thank you for being honest with me. i\'m sorry that...fuck it. i\'m just sorry. i\'ll go now. i\'ve fucked up one too many times.  '),(22778,664485,'19,March,2004','2004-03-19 00:00:00','   i believe in an alternate universe where things are different, where everything is beautifully serene and intoxicatingly amazing.   ...cumbering into a low, just out of place and out of sync with feelings and things all around me that seem to crash and take place all at once. life is an open canvas and i am thirsting for oil paints to construct this picture.   big momma arrives tomorrow, and it\'s as if there is a crazy rush for everything to be just perfect. a midnight rollerblade stroll down ninth, holding hands under the night glow of yellow orange hues...some time to get away from it all awhile. relaxxxx he says....relax.   nit-pick the laterns in our room and everything has to be just right. study my ass off for another couple of months and then comes summer and more studying again. summer comes and i am alone again for another three months. there are times i think, how the fuck am i going to survive?...must you really really really really go to jersey? we have our tickets booked into new york for mid-summer from when i return home from malaysia. i\'m spending a couple of days then school starts again.   sometimes it seems as if i am always running on empty, like my life is one big chaotic shit filled with seven hundred and fifty two things that needed to be done five minutes ago.   but that\'s just life.   and i suppose, i should really just deal with it.   not like i\'m going to be going anywhere soon  ...or am i?  '),(22779,664485,'19,March,2004','2004-03-19 00:00:00','   it\'s mismatched strips of thoughts and feelings released. tokens of what there once was and what there still is. and sometimes, nothing can make sense but yet everything falls into place. the words that once empowered now broken reels of a time lost in space. quite a bit to go through, and leafing it page by page is an indefinite surge of emotions that range from so little to so much. falling asleep in bed, listening to the night sounds and him and crying, sobbing into his chest and asking why and how and where the hell do i go from here? i don\'t know if it hurts more that he understands or if it\'s the senseless guilt trip that your mind plays on you. you do not want to cheat on him, but to allow yourself to feel seems like betrayal on a strange degree. he says it\'s all going to be okay. and for that moment, there is nothing else in the world that you would like to believe, that there is nothing else in the world that matters and all that there is, all that it is is now, and time, and place for moments to happen. allowing yourself to love and be loved. a sacrifice of sorts to be made, a painful memory, but nonetheless, one you\'ll have to live with.  he says you are specially oversensitive. that no one else in their right mind will still be here, i say it\'s love. and he says maybe that\'s why loving me, all that i am, means so much. because i take every little thing into consideration, and i look, hope, when no one else is looking. is it possible to love two, three, a million people at once?  the day after, a strange dejavu like mist of the mornings after a fight, or a breakdown, feels the same. it\'s all a haze.   i am tired, and so is the rest of the world with my bullshit. the room needs to be cleaned, and the dishes need to be washed. the dogs and even miss evil cat sits patiently behind closed doors waiting to be let in, to be stroked and to comfort. why so sad? why so angry? why so anything at all?  i have found the one, just like i thought i had many times before. he is special, and he understands. he says he\'s banning me from the computer, because everytime i come back in to the house i have the look on my face that says my grandmother just died. and so for the next few hours or so, he pays for my mistakes. he pays for my insecurities and needing, he pays for my unability to let go and the fact that i feel too much. and sometimes, i save conversations for him to read, for him to understand...because sometimes i know, try as he might. he doesn\'t. there\'s the other guy, and no one will ever take his place, he knows that by now, and so i wonder how strong of a man you must be to let your girlfriend comfort someone else, love someone else, cry for someone else, and just to take it all. there are times where i know he deserves nothing of this. but times also where i know, there is a reason that he has lasted for so long, for all the nights, for all my insanity.  he gets down on knees and hands, scrubbing the bathroom floor sparkling clean for my mother. i let him do all the dirty work and as always, he has to deal with the extra baggage. i am useless, but i am something with him. i am the love of his life, and the one he comes home to. i am the person that needs him like no other and smiles through stupid crazy projects. i have found love, and summer nights, i have found the meaning to all the answers i was searching for....i am a stronger person, a better lover but yet still the same broken mended pieced together girl. not quite yet the woman i long to be.   there are days where i feel like i know what i have to do. and there are days where i don\'t at all.   i know that i had to go through all that i went through to get to where i need to go. but sometimes, it\'s the in between that seems the hardest of it all. that everything has consequences. that to shut out love like i did for a while is an impossible task. that things happen, and situations change, and piece by piece, no matter how wrong, fall together in place. i met so many mr. wrongs to meet two mr. rights. and so i\'m weighing them, testing the one i\'m with and driving the other into insanity. and i knwo deep in my heart that the one i\'m with is the one that i\'m meant to be with for the rest of my life. if only because i don\'t think i will ever find another person that can stand in the midst of a tornado and be calm. i laugh so much more just like i used to, i have found the little girl i used to be but am yet maturing into a woman. i have found responsibilities, responsibilities that i never wanted but have been taught to accept. patiently, and sometimes screamingly, he teaches me what life is all about and all that it is. what it means to grow up and accept everything, accept who i am as who i\'ll be.   i loved someone a long time ago like nothing i had ever loved before. i gave my all, and sometimes, it scares me to think that giving my all wasn\'t enough. that i wasn\'t enough. and so sometimes, when i am doubtful, i have to remind myself that i am okay. that the world is an okay place and that i deserve to love and be loved. i was so broken, and i am embarassed that he has read all the things i never said before but comforted that he finally realizes now, that i did love him, and that i did try, and that i did wait forever.   i\'m sorry. sorry for so many things but sometimes not at all. there will always be memories, and places in time that will take me back into a different era. i broke, and in the end, i didn\'t wait. i left. and sometimes, i feel guilty for that, but then i look at my life now and i cannot imagine the person i would have become if i was still as hurt, as broken, as torn down as before. i would not be the person that they both love without the things, the memories, the times. i had to leave, and i had to break my promise, and i had to go....and i know nothing in the world will ever justify that. will ever make it okay.  sorry, like so many other times before, sorry. but never sorry that it happened. ever.  '),(22780,664485,'18,March,2004','2004-03-18 00:00:00','   memories hit hard, bullet train.  pain, hurting to remember, sometimes you want to say why bring it up? but then you remember that everybody needs to talk it out sometimes. and so you try to suck it in and understand and comfort. comfort him, comfort yourself.  painful memories, times. days. moments.  nights, days, evenings, birthdays, musicals, plays, watching the stars in a foreign place.  a foreign place you\'ve learned to call home, a place you\'ve come to find love, a place you will leave soon and journey on into maturity and growing up and all the crazy things it entails.  love hurts sometimes, and loving someone does too.  letting go is the hard part, but never forgetting is harder.  you don\'t what if it anymore, and you know deep in your heart that it\'s for the better, that you are better now. that you\'ve changed, that you\'re different, that circumstances and life, and your stupid crazy neurotic mind happened. weak, struggling for air sometimes, feeling so tiny and insignificant and insecure.  i hated the me when i was with you, i hated the me that we both drove us into.  why is there so much pain from something that was once beautiful?  but i\'m okay now, and i am happy. if only just for a while. and sometimes, it makes me feel so guilty. so guilty because i know you aren\'t happy. guilty because i never know what\'s right to say? to want to agree but know it will hurt you at the same time.  to understand but not at all.  soemthing that seems so right is questionably so wrong.  we were talking last night and i mentioned la. about how if things had gone right that everything that happened between him and i wouldn\'t have. and that was a strange eerie feeling.  would i have given up the life i have now to have the life i might have had with you? i don\'t know, but when i left, i thought i did.  it was another crossroad, and deep inside, i know i took the right one, yet the right one hurts you so much.....now how the hell again is this right? but this, now, him and i. i know i wouldn\'t have what i have with him without you.   it\'s so fucking complicatingly hard sometimes.  '),(22781,664485,'18,March,2004','2004-03-18 00:00:00','   there are times where i honestly believe that loving me, all that i am, is an impossible thing to do. impossibly hard to do at least. i know this from conversations with the exes, the random mood swings that go from one extreme to another...the fact that i am a selfish fucking bitch....ever sit alone with your thoughts awhile and just absolutely hate yourself?  i\'m in one of my strange moods, it\'s the in between stage of being weird and semi-depressed. thinking too much and trying to process everything, trying to fix everything that subconsciously i know, do not need fixing. the overthinking stage in which you nit-pick everything fucking dry. the part where you smoke so much you know you\'re going to someday die of lung cancer.   i\'ve been reading journal entries and i realize that all i ever talk about, all that i ever am, all that i am made about is anger, and sadness, and madness. and of course, that makes me angry and sad and want to do crazy things.   i\'m whining again. so just insert about seven of my last entries and you will have exactly how i feel.   fuck  me.  '),(22782,664485,'17,March,2004','2004-03-17 00:00:00','   preparing for death must be a strange thing. yule is pretty much teterring on death\'s door, it is weird. correction, it must be weird. all these funeral arrangements to make, caskets i\'d like my body to be put in, living wills to sign, morphine drips to get high on.   this place smells exactly the same five and a half months later. padded walls and looming big fences that cage you in. there are the insane wails and screams down the hall, blank faces and aneroxic fat ass girls. guys with absolutely no comprehension of manners and the psych techs looking looking looking.   same place, same feelings, same thoughts, different faces.   a place to \'get stabilized\'. a place where they drug you incestantly that there is no comprehension of anything whatsoever. so you don\'t kill yourself. so you don\'t fucking cry yourself to death, or slice your arms and legs apart. so you just don\'t. anything. at. all.   i\'m in a funny place. take it literally, take it metaphorically.  i really don\'t give a shit.   i hate it when people tell you that you\'re smart and beatiful and amazing and yet you feel like nothing and am nothing. i hate it that my mirror is never quite the same as the reflection someone else\'s sees. i hate it that i feel like absolutely nothing and everything sometimes. i hate it that i get so high so high and crash crash crash further, way further than where i started. i hate the one step forward three steps back theory. i hate theories and theraphy.   i feel like robert in \'proof\'. the machinery is fucking flying and i am insane.   talking in codes and mathematical errors and solutions and fucking conspiracies.   i am angry.  '),(22783,664485,'13,March,2004','2004-03-13 00:00:00','   i\'m a nervous wreck like a fucking twit at the moment. don\'t ask me what it\'s about...just, i don\'t know...N.E.R.V.O.U.S.  every fucking little thing is freaking me out, it\'s like my neurotic ass brain is on speed and i can\'t breath and my heart is beating way too fast. i feel like i want to scream and hit something and tap tap tap tap tap.   i had a dream about a portion of my teeth falling out the other night, apparently, that is significant. i haven\'t taken my anti-psychotic, anti-depression, anti-anxiety pills in about a good three months. that\'s probably really bad but, fuck, i don\'t know. i really should reschedule my appointment with the doctor, only that i don\'t want to, and i\'m sick of this shit.   so fucking many things to do and think about and just fucking do.   it\'s spring break and i\'m about to check myself into the hospital.   fanfuckingbloodyfucktastic.  '),(22784,664485,'13,March,2004','2004-03-13 00:00:00','   liberated...liberation. it\'s like a feeling of letting go, being happy, screaming on top of the mountain and just falling down, freely...flying. soaring, soaring so high and fast your cheeks are plastered into a smile by the wind contorting you face. the sound of laughter, cackling, laughing so hard you can\'t stop. the sound of music, filling your ears. dancing alone, revelling, having a fucking blast in the twisted fucked up mind of yours.   i am on a high, or at least, i am moving into a manic state. i cannot sit still and i want to laugh and laugh and laugh. i think it might have to do with my diet pills, the caffiene surge, ephedra free bullshit that jacks up your metabolism and makes you want to do crazy things. or maybe it\'s my new pills that at the moment, are working absolutely perfect. painfree, bloodfree, stressfree. a first in a long long time. or maybe it\'s that i have a fucking 4.0 gpa this term, a stark difference from my shittingly failing bastard ass results last term. i am happy.   maybe it\'s that i am allowing myself to dream, to pursue things that i have wanted to achieve for so long. that i am finally letting myself say, you can do it. and you\'ll be just fine. perhaps it is the extra touch of confidence he gives me, the one where he listens, listens to me when i talk. and helps me solve the weight of the world that\'s on my shoulders. the one that says deep breath, it will all fucking work out! and it is, it has, oh my god, it is working out.   and i mean every little fucking detail. it\'s like a catastrophic ephipany and i don\'t know what to do with myself. so many fucking ideas, the will to be creative, the acceptance of being just you, the permission to live and dream, the desire to see things through.   i am bursting and smiling and raving and dancing all at the same time. on a high that spans from so many different things.   am i crazy?  is this real?  i am happy. i am happy and that\'s a weird feeling. i don\'t think i\'ve felt this happy in a long long time. i don\'t think i\'ve ever been this happy or free or just plain fucking okay.  this is weird.  wow.  '),(22785,664485,'10,March,2004','2004-03-10 00:00:00','   ladies and gentlemen, i introduce to you rainbow racheal.  :)  '),(22786,664485,'09,March,2004','2004-03-09 00:00:00','   i had an internal ultrasound and some biopsies taken yesterday. 4 hours at the hospital later and i am still bleeding. life is a stupid stupid thing.  '),(22787,664485,'04,March,2004','2004-03-04 00:00:00','   to everything, there is a something.  '),(22788,664485,'02,March,2004','2004-03-02 00:00:00','   i am tired of my website being closed, and tired of my unincredible range and tired of pain.  '),(22789,664485,'25,April,2004','2004-04-25 00:00:00','   urlLink      the performance last night was weird..stoned throughout the whole thing...popping painkillers and running through lines, acting. singing, breathing in hard through my lungs with my diaphragm, all the painstaking hours upon hours of lessons in a cloak room comes rushing back. the applause shatters with wolf whistles and cat calls, ryan must have paid them to cheer for me.   the tapes came out interesting, i  watched  critiqued it harshly...things i could have done better, things i didn\'t like...my face, my fat ass, my bright ass red hair and the lighting.   one night only, i smile straight into the camera and away, revealing too much and nothing at all..masking with characters and slathering fish scales. flying, a cacophony of voices within fighting to be heard...the final curtain, one last bow. joining hands with a cast that pulled together and smiling. running backstage.   wanting to get the fuck out of here.   life is so weird when you\'re bordering(bothered) and looking through.  '),(22790,664485,'24,April,2004','2004-04-24 00:00:00','   i\'ve been reading...it\'s weird to sit and flip through someones life that could in so many ways be yours...words, pain, feelings, things you think no one else understands, feelings you think no one else can comprehend. weird.   i\'ve got meesh to thank, i got bored and and was browsing through her links and came across her..  so. weird.   i read about second chances and fucking up, i read about pain and tears and slashing your wrists, i read about hope and love and lots of shopping. i read about this and that and this and that and i was mesmerized.   it\'s nice to know that on the far other side of the world, someone thinks very similarly like you.  scary.   urlLink  \"Perhaps I have less skeletons sitting, collecting dust in my closet (although to be honest, I doubt that to be true)... or could it be that I have finally begun to say to myself ... \"To hell with it... I am just who I am and should not try and hide it.\" I don\'t know...\"  '),(22791,664485,'18,April,2004','2004-04-18 00:00:00','   we went out to take photos last night...a memory of moments in time captured. we sifted through a bunch of other things, a bunch of other moments..and it was weird to not recognize but to have some sort of inkling of a different period in time. growing, learning to be independent...sprouting wings.  it\'s hard not to cling on to security, it\'s hard to relax, take things a day at a time, a minute...a second.  it\'s hard to say this is okay, hard to not worry yourself to death.  crouching, hurting, crying for the girl in the pictures. the low light glows, a series of breaking down, being so sad. a cycle continued daily, weekly, monthly. escaping the madness, the in between, the uncertainty of tomorrow or the day after and the day after that.  nostalgia is a strange strange gift, you make mistakes and learn, you take all that you know, all that you\'ve come to know in strive and sometimes you get bruised so easily. sometimes the scrapes and scars seem so much more than they really are.   he somehow managed to pull some strings, yell at some people and so tomorrow we\'re going to see the doctor. and i have to somehow be honest about how i\'m feeling yet be cautious of what i say. catch 22 is an evil thing.  i want to get better, it isn\'t that i tell him. it\'s just hard to find a secure balance to thrive on. and i don\'t want to go back in....i really don\'t. i hate the smell, i hate the darkness that creeps into your skin, i hate the feeling of hopeless hope.   there\'s so many things i need to do, the musical on saturday, auditions tomorrow and hopefully, callbacks on wednesday. finals coming up, registrations, preparations, production meetings. if anything, i don\'t need more time to sit around to think about things....\"but they need to regulate your medicine honey\" regulate my ass. they dick around with my pills so much it\'s green one day, purple the other and blue the next. and there\'s always the misdiagnosis of just what the hell is wrong with me. borderline, truly, unknown.   but i\'m fine.  and nobody listens.   just leave me the hell alone....i\'ll fucking figure it out.  '),(22792,664485,'17,April,2004','2004-04-17 00:00:00','   sometimes i think god forgot to put the connector between fuse a and fuse b in my head.   ...if there was a god anyway.  things don\'t really make sense at the moment....all these internal voices clanging around and pressure pressure pressure.  they\'re threatening to baker act me, yeah, stick me again for a couple of days in the loony bin and everythings going to be just. fine.  sure.  exactly.  uhuh.  you ever have crazy thoughts running through your head?  i threw up for the first time in a lot of months last night, panicked. i don\'t know.  things don\'t fucking make sense at the moment.  i feel crazy, possesed...on fire with no fuel, crashing so fast it\'s a big huge blur.  catatonia is a nice place to be sometimes.  yet this everchanging madness that takes inside always takes a different form...it is essentially the same thing with a different variation, a twist of some sort.   i\'m not suicidal, yet i do want to hurt myself.   how the fuck does that make sense?  and it isn\'t so much about the pain anymore, it isn\'t so much about going away some place.  i just feel so...out of control. and sometimes, that makes me very afraid of who i am, what i\'m going to do, what happens next.  it\'s all the extremes, it\'s all the anxiety and worrying and crying and hurting and everything. every bloody single stupid fucking thing for absolutely nothing at all.  it\'s being at a crossroad without any fucking roads at all, it\'s putting on the gravel and tar to what, you\'re not exactly sure.  it\'s needing to do something but not knowing what is.  everyone says go in, take a break for a while, escape reality. don\'t you understand you stupid morons that it\'s all goign to be the same anyway? that you can\'t just lock me away from the real world because in the end, i am going to be back there anyway?  that the medication doesn\'t fucking work?  that my mind just got wired completely fucking different from the others and it\'s so hard to make it okay.  that i try, oh god, i try every fucking bloody day to stay on top, to try try try, to not do stupid things, to not kill myself, to not hurt myself, to be the best, to be superwoman, to handle every fucking thing?  and this is what scares me..  that if before i had so many things that was going wrong in my life to just say, you need a break and those things are making you crazy.  what do i have now?  what can i possibly blame it on except for myself.  one day i\'m going to drive the people around me away.  an inevitable thing i\'ve come to learn.  but yet he sits there night after night, reassuring and reassuring again. listening to my sorry ass make excuses for being me, loving me. phone calls, visits, they come and ask and pressure and support and this and that and this and that. i\'m trying, but my trying isn\'t good enough.  why am i never good enough?  i really just want to be normal.   yeah, normal would be nice.  '),(22793,664485,'13,April,2004','2004-04-13 00:00:00','   we were hanging out in the house today, him putzing around trying to build the underwater housing for his camera and me just not really doing anything at all...and somewhere in between \"honey what should we make for lunch\" and \"lets go paintballing\" i burst into tears. i don\'t know why or how it came about except that i was shaking and so sad....not quite knowing what was going on or why some demon in me opened up the gates of hell. he held me for the longest time ever, just me curled up and bawling and him holding. protecting...asking again and again why baby why why why.   ....is there ever really a why?  '),(22794,664485,'13,April,2004','2004-04-13 00:00:00','   you wouldn\'t think that it\'d be possible to have absolutely everything go wrong in one day but knowing my luck, of course it does.   we begin with getting up late for church, trying to take a shower without killing each other because it\'s fucking freezing and we both want to be standing under the hot water at the same time.   ....then realizing halfway through that it\'s sunrise service and we should have been at church only about three hours ago.  fuck that.   so we strip off our nicely ironed clothes and let them rot in a pile by the bed and grumpily climb back between the sheets only to have zigi and sasquatch wired as fuck wanting to play. lock them out of the room and go back to la la land.  skip a couple of hours later (yeah right, try in the middle of the freaking day) and now ryan\'s late for work and he\'s flustered so i\'m flustered and hands and limbs fly everywhere trying to find this and that and get this and that.   chop some things up together for lunch, finally take a break to watch some tv and sort through my mail because i haven\'t had time to do anything lately and sitting right on top of the pile is a freaking ass bill around the tune of 1400 dollars for my little visit to the loony bin the other day.   fuck.  me.   fluster around again for another couple of hours trying to figure the hows and whats and why the hell my insurance isn\'t covering it or just the blatant fact that i have to pay for my summer and fall tuition by the end of this month which by the way, runs about 3600 a freaking term because i am charged the international rate and nevermind that my parents have money coming out of their asses but shit. that\'s still a lot of freaking money and the last thing on bloody earth i need and want to bring up to them is the other little \'incident\' where their stupid ass daughter has mental fuckupness.  FUUUCCCKKKK!!?!!@?!@>!?#@!?$@?#%$W?^4  so what do i do?  being the domesticated bitch i am, i resort to cleaning.   i take off all our sheets and decide that freshly laundered sheets would be great and the fucking ass dryer breaks down and because it\'s freaking easter sunday, nothing is open.   nervous breakdown. nervous breakdown. nervous breakdown.  key and solution to it all?  walmart.   ahhhh...walmart.   i fight through crazy old ladies that tell you about their life stories while you\'re picking out zuchinnies and crying babies that won\'t shut the hell up and get lost in flourescent lights, cheap ass crazy easter deals, candles, more candles, (latest and ongoing obsession) and check out ladies that are way way too bubbly for a shitty day.   i come home and i decide i\'m going to make a nice lovely romantic dinner for the both of us only that i don\'t know how to cook or kill or mutilate these freaking live lobsters and one of them seems to have bitten/chewed/fucked up the rubberband on one of their claws between the ten minute drive home from walmart. so in comes more hysteria and madness and THE FUCKING ELECTRICITY BLOWS OUT.  yes.   stupid fucking old ass piece of shit house.   i don\'t pay 800 dollars a month to a. get carbon monoxide poisoning, b. have a huuuuge stupid ass backyard with trees that won\'t stop fucking shedding leaves, and c. most of all, to have no electricity.   ryan comes home, bitch bitch bitch bitch bitch (who else am i going to bitch to?) calls the old ass landlord and he sends out the FIRE COMPANY? why? i don\'t know. ryan flips on the breaker and the whole house goes crazy because the circuits are fucked up and our lights are blinking, blinking, blinking, the stove isn\'t working, i\'ve got fucking live lobsters in my sink and the house seems like it\'s going to explode.   and what\'s the sane racheal thing to do?  run with my passport and sit in the car.   at this point, looking back on it, it\'s fucking hilarious and would make good grounds for a farce/comedy play thingamagic.  cut back to the nice big red fire engine truck that pulls up and big burly yummy men that come in and go through everything only to switch off our power because it isn\'t safe.  duh, you stupid fuck.   call back our landlord and the bastard doesn\'t pick up.  make fucking dinner on our GRILL.  trying to boil water in candlelight on the grill just doesn\'t work for me honey. creative yes, but so wrong.    finally go to bed with no sheets, no electricity, absolutely no mood for sex whatsoever, grumpy, and the little grocery shopping i did in the evening, rotting, because hey, refridgerators need fuckinig electricity to work.   -------  aftermath.    the lakeview bill was a mistake on the wonderful people over there and so the wrote off the bill because i make no income whatsoever in america and poor me is just so pathetic and they can\'t blame shit on me because they baker acted me anyway. so hah!  the dryer didn\'t work because the fuses were fucked, hence the electricity madness because the lightning late saturday hit it somehow and it was all fixed and done with before i woke up because my boyfriend is a good little bitch. :)  the landlord apologized profusely and brought his ass down to the house at seven in the morning to make sure someone looked at it and had it fixed by noon.   aaa-fucking-men.  '),(22795,664485,'10,April,2004','2004-04-10 00:00:00','   Relationships V: Reclaiming yourself The problem with \"looking for love. Real love. Ridiculous, inconvenient, consuming, can\'t live without each other love\" is precisely that: they consume you. They take over your very being at the most intense level for what seems like the longest time only to leave you dry and whithered, worse for wear and drained of emotions shortly after. They swallow what was there, masticate your spirit and soul before spitting you out, transformed into something totally unrecognisable.  Crawling out of one of these relationships, whimpering out of a dogged and tempestuous existence, I find I no longer recognise myself. What was there, no longer is. What there is, is unfamiliar and foreign. Emerging from the tyranny of paradise on earth, I long to rekindle myself with the flames of old. I long to rediscover that spark that was me before and to find the identity that was \"truly me\".  But that, is the hardest thing about coming out of a \"consuming, can\'t live without each other love\": reclaiming yourself. After a trip to \"can\'t live without you\" land, you forget who you were or what you were about. After consuming a passionate, over-the-top love, it consumes you. It eats you up, burns your every fibre, worth and being before leaving an excrement that is only a shadow of what was. After crawling out of an oasis of decadence and luxury you face a barren dessert under the glaring light of a scorching sun - a light so bright that it illuminates the poverty of you and all that no longer is.  And that, is the biggest challenge: remembering who you were and truly are. Not as one half of an entity but as yourself. On your own. Standing on your own feet. Again. As you did before. With a name. An identity. A life. A value in and of itself. One to be cherished.  The act of reclaiming yourself is arduous, fraught with challenges and setbacks. It\'s never easy and most certainly not unilinear. I am overcome by emotions and obstacles every step of the way. One day I may make it, but today, I am trudging, unsuccessfully, through a mire.  jikon   we lied in bed last night talking, about life and the nothingness of the in between. we talked about the summer, and all that overwhelms me about it. we talked about my medication and changing perpectives; looking into the future and enjoying every little thing of the right now.   i layed in bed remembering this post, the all consuming us and how dysfunctional i am without the other half. how our lives have seemed to twine and tangle and it is hard to get away from us, or to know what life (summer) will be like without him. as he rubs my back and tells me i worry to much. i let bitter salty tears touch our pillows, and curl into his arms and cry a while more. i ask him about the skies and life and everything i need to know. i ask him for answers as he tries his best to give my questions a soothing answer.   we always have candles around the house, and as i look up from our bed i see our lanterns we\'ve hung up so precisely around. i remember nights and moments and memories. and slowly, as he moves in me i thirst for more...i thirst to never forget this moment or to never forget what love is all about. i promise myself that i will never get jaded, that if this fails, that i will remember nights like these.   the way he looks at me, with respect, with love, with awe....with understanding.   so many things to do and so many thigns that need to be done. flying through amsterdam with a two week vacation back home that doesn\'t feel nearly enough. singing in the rain and watching the dancing shadows on my wall, i want every bloody thing and i am so needily insecure and alright at the same time.   i have no ending to a post because i am sitting here with thoughts racing through my head and seventeen different directions i could go.   relax, he says.   it\'ll still be here when you get back.   ---  i don\'t know how we got there....or why i flaunted the pieces of my soul nor why as i held the pieces, i cried.   i don\'t know why, i did the things i did, became the person i became, was the person i was.   i don\'t know why there are times i feel so incomplete, times like these...nights where i could sit here, smoke a cigarette, and still sit in the dark for hours asking the questions.   i don\'t know, i don\'t understand, i don\'t want to know.   ---  we\'re watching a movie and the lights are dim except for the neon red exit sign that blinks every so often. i hear people crunching their popcorn, a baby three rows behind me squirming and crying. the little girl with white blond hair that\'s turning around every so often to smile at me, iniatiate a conversation without words.   lost images, and so many of the little things to capture.  i feel like i am losing everything, my mind, my sanity, my grip on things. can you somehow be dragged through with earth stuck between your feet and shoes yet be soaring at the same times? because that\'s what it feels like sometimes, most of the time.   the yin and the yang, the ups and downs, the highs and lows.   losing control and not being completely happy but so happy at the same time?   bigger boobs, shorter hair, quieter...meeker in some sense. i look through archives and archives of pictures; it\'s funny how i truly believed i was ugly then...and now i see. i see what people told me, i see someone beautiful, someone worth loving, someone special.   will i look back at these pictures of me now and realize that i am beautiful despite all that i think of myself now?  '),(22796,664485,'09,April,2004','2004-04-09 00:00:00','   we\'re helping to produce an indie film that has to be shot, edited and ready by the end of april. god forbid i let myself get sucked into such exciting projects when i have my finals coming up. stupid ass.  and so the computer room is full of props and my mailbox with correpondence from the various crazy people i have to deal with.  why?  i don\'t know really.   i\'ve had to iron sixty seven business suits while my boyfriend gets to wear jeans and another vintage punk shirt.  i did however buy new  cute  sexy underwear today.  a little something something after twelve hours of hard work.  :)  perhaps to make up for all the cute rocker grunge come-and-fuck-me sexy bands i have to deal with and their stupid ass label companies.   my boyfriend is more famous than me.   this sucks.  and just for the record, the new rule is that he can\'t check out the same girl i am checking out because that is just wrong.   :P  back to work.   gah.  '),(22797,664485,'08,April,2004','2004-04-08 00:00:00','   POETIC JUSTICE  Is when ten years on, in a most unexpected place, you run into the boy who was the love-of-your-time-in-Year-2 (before you changed schools and at some point clean forgot him), childhood crush you spent the whole year pretending to hate,   and you’re wearing your stilettos and   that   take-me-home-and-fuck-me lace-and-leather skirt  and you realize, standing there watching him flusteredly flattered,  that you’ve grown into the power of choosing  to lure but not to let him.    --  it\'s like lighting up your last cigarette on the wrong end then wanting to slap yourself silly but then an unexpected thing happens and he comes home with a carton just the way you like it right down to the slims.   it\'s like irritating the shit out of him and singing every verse wrong to every damn song because your brain is fried and your sense of memory is filled with stupid things.  it\'s like begging to play and turning him on because if he watches another episode of tripping the rift, or southpark, or clone high you.are.going.to.scream.  ...yet secretly, you watch it from the other computer too(between scrabble games)  it\'s collaborating and producing someone else\'s baby together. and him and them having not so secret conversations about you.  it\'s just this. midnight on a wednesday night with nothing to do, nothing to say. just doob dee doo.  '),(22798,664485,'04,April,2004','2004-04-04 00:00:00','   i\'ve come to the conclusion that there\'s only so many scrabble games you can play before you get so bored and frustrated you have to hit something. hard.  oh, and time changes that make you lose an hour of good sleep suck.  '),(22799,664485,'01,April,2004','2004-04-01 00:00:00','   he plays that oasis song incestantly in the background...or maybe it\'s because it takes me to some place i don\'t know and soothes me. i am not complaining, in fact, humming, wailing with my heart on my sleeve. he laughs at my karaoke display, as he twirls me around in our computer room slash laundry room slash attic slash put whatever shit we don\'t want in the house.   mom left last night...or rather in the wee hours of the morning. somewhere in between four in the morning and six. i love how they make you go to the airport to check in two hours before your flight then decide to open the counter only at about five because no one in their right mind wakes up that early. right.   fucking retards.  i think i am in a haze of some strange unknown weird parallel. i\'ve been off my medication for a couple of months now, i feel okay...a little up and down but still okay...or so i think anyway. i\'m always okay.   they sat around checking each other out. her protecting her little grown up girl, miss hick and strange and weird and happy. him, our relationship and us and who we are and what we are becoming. she grills him, with questions about our future, and missouri...school, work, whatever she can think of. he listens, thinks about it, reassures.   i don\'t think it\'s so much as it is about her wanting to come for a visit as it is for her to accept us. accept this. this whole crazy madness that i tell her about. check us out...making sure i\'m not getting my heart pulped with another ten suicide attempts again. wanting to make sure i\'m physically, mentally and emotionally okay. wanting to talk talk talk talk to the doctors and ask six thousand and seven questions and go through charts and biopsies and projected results and stupid crazy motherly caring shit like that. i see now why they both get along so well...they worry they\'re fucking minds off.   she brings him this nifty little camera, a sony cam corder worth about a grand, and he films all these weird shit with it like me doing the dishes and our backyard and god knows who and what. he plays with it like a little kid and his bright new fire engines. my dad thought it was a fucking excellent idea, and so she carefully lugged it through singapore, and minneapolis, and colorado...back to minneapolis again and then florida. fuck yeah my parents are great people.   along with it she\'s brought three suitcases with everything and anything. undoubtedly pensacola is a tiny ass town but damn, mom! my maid\'s packed every damn spice and ramen noodles known to mankind and little snacks and stuff that i like because she\'s afraid i\'m starving in big bad america. my dad\'s sent me nifty things like books and notes and money(!).   we shop for about ten straight days clocking a good six hours at every go as she fills it back up to take home. she\'s so tiny, my mom. but she fills up the room with her personality...it\'s funny how such a big laugh can come out of such a small person. she shows me the new highlights in her hair, and talks about home and my brother and the family. she tells me about isabel and how damn adorable she is now, how she\'s crawling all over the place and her face lights up. i\'ve missed this, i\'ve missed my mom and being at home and all the wonderfull familiar things.   yet at the same time, i cannot help but notice how different things are. how comfortable i am in my skin, my home. in my life, and my lifestyle and the way things fit perfectly. she\'s this city girl and finds my little town just so adorable. i see the things, the person i was when i first came here in her. ...and i wouldn\'t give up anything in the world for the maturity i\'ve somehow gained nor the lessons i have learned.   i turn twenty in a couple of months more, no longer a teenager, a stepping stone away from a full blown legal adult. i\'ve learned to embrace these moments, treasure these lessons of acquiring my wings to fly. i am no longer so much a little girl anymore, she tells me this. reminding me that they are proud of me, that i\'ve somehow found my niche in life, this stability i\'ve craved for so long. she gives me an extra boost to keep on fighting the good fight, to keep on trying no matter how many times i fall. and i am soaring, reaching for the skies and humbly caressing the earth, the soil that sifts through my fingers.   i am okay.  '),(22800,664485,'30,May,2004','2004-05-30 00:00:00','   oh my god.   i\'m not sure i even know where to begin?   the beach was a fucking blast. and boobiliciously tit-ilaciously fucked up doesn\'t even begin to describe it!!!  i\'ve never spend memorial day weekend here in pensacola, by this time of the year, i am mostly back in malaysia and raising hell there. sure, i\'ve heard the interesting rumors and crazy stories my beautiful boys tell me....but oh fucking hell, nothing in the world prepares you for a weekend like that..!!  so my favorite sexy pants and i have been meaning to get together for the last week or so and finally he just calls me on saturday morning and tells me to wake my ass up and grab my swimsuit cause we\'re going to fucking celebrate gay pride weekend.   okay?  uhm.  yeah.  we go over to boobie\'s house who lives right on gulf breeze and the traffic is fucking insane. think 120,000 people on the beach, and about 1000 parking spaces, 500 cops and a whole lot of rainbows. the beach was so packed they started to turn the cars away and only let the trolleys and cabs through. so we are in bumper-to-bumper traffic for a good hour and a half and we have to fucking drive back to boobie\'s house and take a cab! arrghhh!!! lol, it was still fun though because our cab ride was a story in itself.  and btw, the heat is absolutely sweltering, i mean like your ass is fucking on fire. we finally get to the beach and i swear to god it was like culture shock.   there were a few hundred tents set out along the whole stretch, some elabvorately designed, some put together randomly, some just full of people upon people upon people. we meet up with our \"posse\", leave our shit and start walking.   pensacola bible belt my ass. there were at least 100,000 gay people, and enough of boobies and dicks to last me till the next memorial day weekend. lol, literally, chicks walking around naked, guys in fairy outfits, transvestites, bikini\'s and swimming suits of all sorts, the token old bald fat gay dude with speedos on. the guys got so fucked up it was hillarious. see the thing i love about hanging out with a whole bunch of gay people is that they are fantastic drunks. not only is their wit sharpened, their sense of adventure and excitement is aroused. in more ways than one.  so you flash your way through shots and tent after tent and by the end of the line, you\'re falling off your asses and laughing about sand.   and the great thing is, the cops don\'t do jack shit about it.  how exactly do you arrest 120,000 people? albiet, they\'re all not underaged drinkers but ohhhh yeah.   so there\'s wading pools and rainbow kites, gay pride stalls and the betty ford clinic, thousands and thousands of hot sexy sweaty chicks, and lots and lots of eye candy. yum double yum.  and to top it all, millions of parties to go to, leis and mardi gras beads strewn on you like a stamp of approval and as the tag line goes, \"family gathering\".  now, to go party hopping is one thing, but to go gay party hopping is a whoooole different story. and then at the end of the night, when you\'ve hung up them stilletos and have retreated home with about 50 more people to come is another thing!  this weekend has been crazy, nevermind that i don\'t remember half of it, and nevermind that there were about 30 people i didn\'t know at my house last night, and nevermind that we\'re going to get bitched at on tuesday when the garbage disposal people come around from our tremendous overflow of beer bottles, and cans, and plastic jelly shot holder things, and shit i don\'t even know where to begin with...it\'s all good. because yesterday, was possibly the most damn fucking fun i\'ve had in a while!  sweet fucking yeah!   :P  '),(22801,664485,'27,May,2004','2004-05-27 00:00:00','   i\'ve been juggling schoolwork, sanity, entertainment, and being miss racheal in the last short months and it is crazily fun and insanely tiring at the same time.   there is weee waayynnee that lights up my sometimes gloomy world, something about the way he smiles up at you that makes your heart sing. something about the way his head rests perfectly between the nooks of your shoulder and head; magical. he is my special sweetheart among all the others (burp, sleep, play, change diapers, eat!...what more could you ask for?) and i think he knows it too.   then there are the other kids...the ones that kiss me when i don\'t expect it and hug me from behind, and underneath and upside down. the ones that go \"do me! do me!\" the ones that i color and make funny faces with. the ones that i instinctively pick up when crying, or that instinctively reach out for me. the ones that can\'t decide if they want shrek cereal or mac and cheese or cracker stackers for lunch. and then decide that whatever the other person is having is better after you\'ve fixed it. (duh! miss racheal)   then there is samuel, samuel who only watches tv on movie day. samuel who only eats organic and nothing else. samuel who asks for his bwocolli at snack time. samuel who likes to be called thomas after his favorite toy train and whitney that likes to be called my pumpkwin. alex, who is one and teething and a major pain in the butt but a major tug in the heart too. alex who i have to watch out for, alex who has learned to shake his head and alex who will put absolutely anything and everything into his mouth. alex who likes to be rocked to sleep and only lets certain people hold him, and *will not* let you carry anyone else when he\'s in charge. :) jusdin bo judeen (justin) who has mastered the art of climbing out of his high chair, giggling like a hyena and is obsessive compulsive i swear. (every. thing. has. it\'s. place.) then there is hayley, the new girl who is a spitfire and absolutely cannot go night night with her shoes and socks on, who is convinced she is a princess and chooses pink for absolutely anything and everything and will sit in front of our play make-up area for hours and hours and hours(kid hours anyway) hallie, which is apparently a completely different sounding name from hayley, who is six but is almost if not more intelligent than me. the little girl that aces through the worksheets and then turns around and does a silly thing like sticking her head in between the chairs. dakota, my big boy of five years old, who follows me around and is my little helper. dakota, who gave himself a haircut and comes out with all these crazy stories to tell me that don\'t make sense but are so entertaining at the same time. anna lee, dakota\'s younger sister, that talks in that same crazy alabamian accent that confuses the living daylights out of me and is always one of the first to come up for a hug when i arrive. and finally courtney, courtney whom i truly believe will become miss america one day. she who reigns supreme on the others but gets so so so quiet when tired, she who sings and dances and shouts the loudest. she who forgets if you write flowing to the right or the tfel(left) then decides she\'s going to write hallie\'s name instead because it\'s easier.and she who makes the silliest faces and giggles the loudest. next to justin of course. who laughs for no apparent reason at times. (silly boy!)  i have shirts that are stretched beyond porportion from little hands tugging at every angle. not that that is anything to really complain about. i know more about shrek(their current obsession) than possibly anyone in this entire world or nursery rhymes. and i know this because for about two hours after i come home, and am washing the spit ups, grubby food fingers, and dirt from the yard from rolling around with the kids, out of my clothes, tending to the bruises and scratches for killer hands and feet and tiny bodies, i am singing about bloody rain drops and gumdrops.   so those are my days from about 12-7 on weekdays and i love it.   children are so crazy amazing aren\'t they?    p/s you want to hear a racheal is such a stupid dork story?   i turned a load of whites into pink. P I N K. gahhhhhhhh.   bodoh.  p/p/s have a great memorial day weekend! woohoo!! :)    urlLink  \"Good God, why does it take so long to digest this honeycomb nonsense? I have bits of it sticking right into my teeth. Later folks, later. I must be my charming self and smile at Old men in 10,000 dollar suits. Sigh.\"  '),(22802,664485,'25,May,2004','2004-05-25 00:00:00','   we had a damn successfull garage sale this weekend that amusingly enough, brought in all different walks of life. there was the flea market woman that bitched about being american and great old america, the little african american kid that played with all our toys and begged his mom for more quarters and change, this strange blue haired lady that just about overturned and messed up every. little. single. thing. then there were the others, the art teacher, the homemakers, the lurkers that just sorta hung around for a while and walked away with a fifty cent book, strangers. passing on a gift. giving away a part of my life that i don\'t need anymore...books that i had pored through the middle of the night, clothes that i had worn through phases, lamps and shades, cultural and decorative ethnic shit, photo frames, teddy bears, couches, armoirs, jewelery cases, chains and necklaces i worn so many times before, each with a different story.   it sorta reminded me about my blogging friends, how we all give and take something to each other. how we are so different, from different cultures and backgrounds, a melting pot of sorts that have just sorta come together.  and i could not help but reminince(sp?) about last summer.   i flipped through my archives into a turbulent  urlLink month  of self-discovery and passion. hope and friendship. laughter and tears. beginnings and endings.   and these were the quotes i would remember forever, the quotations full of innuendos and personal secrets, of moments where you just had to be there to understand, of stuff i couldn\'t even begin to describe:   urlLink  \"So here\'s to friendships (old and new, lost and found). Here\'s to pictures and yelling and naughty antics. Here\'s to the White Queen, flitting about, dropping fairy dust, dimpling her way into our hearts. Here\'s to more meetups, and more shared laughs and tears. And finally, here\'s to life. Wonderful, glorious life.\"      urlLink  \"The conversations... the screams... the curses... the profanities The laughing... the drinking... the smoking... The sweet talk... the naughty talk... the cuddling... bordering foreplay\"      urlLink  \"How can anybody describe what it was like last night?\"      urlLink  \"Jamie, I swing every way.\"      urlLink  \"But let me learn, let me make my own mistakes, brush the grit and dirt off and try again. I might cry, I may be difficult, and I may whine about never wanting to live/love/try. But let me wallow till I can wallow no more, and I will be fine. I will get up and try twice as hard. \"      urlLink  \"I was at the Light Side of the Force table before shifting to the Dark Side.\"      urlLink  \"How long did it took you to figure out the tap in the washroom?\"      urlLink  \"the weekend was warped, and the only reminder i have that anything happened at all is the black nail polish on my pinky, a slightly sore throat and exhaustion. and my hair still smells slightly of cigar...\"     that night was possibly one of the happiest nights i\'ve ever had, and it was all because a bunch of people i hardly even knew bothered enough to come together.   thank you.  for every person that showed up that night, for every hug and kiss and word of encouragement.   i will treasure those moments forever.    ...and now for the  urlLink incriminating drunk off my fat ass photos . (thank you alex)  lu punya pasal lah!!!     urlLink  \"I am not ashamed of who I am nor am I ashamed of the size of my ass. You figure out which one is a lie.\"  '),(22803,664485,'23,May,2004','2004-05-23 00:00:00','   we spent about three hours lazing at the beach yesterday, me laid out like a beach whale with strawberries, sandwiches, capri suns, vegetable tray, and wally lamb(she\'s come undone).  ...and now i look like a crisp plump tomato.   lol, i got a tan alright...with lines and little triangles across my chest and milkyway. i look strange, i am golden brown on certain parts of my body and red red red on others.   bethany and jessica don\'t seem to have survived the uv rays either...much as we slathered the spf45 lotion.   the beach is absolutely gorgeous at the moment. miles and miles of white sandy shores and twinkling various shades of exquisite blue that disappears in a distance.   living here, i think i\'ve taken for granted just how amazing the scenery is. pensacolians are some lucky sob\'s as jessica said.   when i first moved to florida i lived in the outskirts of the city and the bay was my backyard. literally. i remember laying out on the dock at night to watch the stars, or do my homework.   i\'ve now moved to the city to be closer to school and civilization. albiet the beach is still fifteen minutes away, i\'ve learned to appreciate it so much more.   even with the damn fucking tourists.  :P  speaking in which, 50,000 gay men and women are expected on our shores this coming memorial weekend. it\'s like a yearly flight down south apparently, and trust me, this should get very interesting.   \"dip me in chocolate syrup and feed me to the lesbians!\"  '),(22804,664485,'21,May,2004','2004-05-21 00:00:00','   i think i\'ve just been offered  the  most amazing job ever!   urlLink adele (the sister i never had) has offered to fly me to belfast for the summer to look after their baby daughter, arriana. flight, food, lodging provided plus extras.  how fucking ass amazing is that?  on a normal day, i take care of about ten kids for about five hours or so after class. sure they get cranky, and baby wayne cries for reasons you cannot comprehend, and smelling like curdled milk is a bitch but everyday that i leave, i cannot wait to come back for more. don\'t ask me why, it\'s the mother instinct in me.   now the only problem is that i have classes running all through summer and books that i\'ve already paid six grand for. i have a three week break right before the fall term starts and i contemplated spending that time in belfast but then after weighing the pro and cons, i realized that i would really like to go home most of all. stuff myself silly with nasi lemak and be a spoiled brat.   so yeah, GRRR.  ...so. close.  '),(22805,664485,'21,May,2004','2004-05-21 00:00:00','   i wasn\'t going to blog today, but the events that lead up to last night prompted something, struck a cord i suppose.   we were sitting by the red light along davis, and across the intersection was an old lady stopped, waving, desperate. she\'d ran out of gas, and i think i whispered something in the lines of \"aww\". i was going to suggest we stop and help, but there were so many things we had to do that day...and so little time.   the old man, a person i admire and respect, a person that inspires me everyday of my life, turned around, went out of his way to help. this man of fifty years, graying hair, did not think twice about flashing his lights and weaving between traffic to help someone else. he, all about a hundred and seventy pounds of him, helped pushed her car.   but what hit me most about that moment was that the truck in front of us, a black dodge with a fresh trip from home depot or lowes made the same weaving, cutting the major intersection to help. and they were just these two tiny little old man, completely white hair and all. i stopped to look at the people that stopped to help, a black woman with a green minivan, a middle aged guy with a full suit and tie, another man just out of nowhere.   i was so comforted and dissappointed at the same time. this is a major highway, a constantly used road that ferries college students just like me to and fro. and not once did i see someone of my age stop to help, and not once did some frat boy jock say, let me take a few minutes from my day to help someone in need. instead they honked, and cursed. instead they drove right by selfishly. instead they turned a blind eye to just an old lady stopped by the side of the road. because it wasn\'t their concern. because it wasn\'t about them. because it didn\'t bother them in the slightest bit except that she had now caused them the trouble of having change lanes.  and to be honest with you, i don\'t know if i would have stopped. or if i would have said, oh i\'m just tiny little me, what can i possibly do? or if i just would have drove past and said i hope the next person helps her. or this, or that, or....everything else but stopping.   the old man reminded me that our generation has lost the meaning of selflessness to it. how many times does the word i,me,mine,myself appear in my blog? way too many times. and all the things that i can say to defend that are only excuses because truthfully, we do live in a selfish world.   the first lesson i ever learned about selflessness was from my mom. she\'s stopped to pick up this elderly couple by the side of the road about two blocks away from her office. i don\'t remember their faces anymore, only that he was older than her and walked with a cane. we stopped, and my mom asked them where they were going and if they wanted a ride. they said they were waiting for a bus to take them downtown and that they weren\'t sure how to get there.   i remember asking why their own sons and daughters weren\'t bothered enough to take their own parents to the place they needed to go? i wanted to know who in the right mind lets someone in failing health take a bus downtown. malaysian busses are notorious for being cruel and harsh, i wondered why they didn\'t have anyone help them at all. the weather was hot that day, they could have passed out from a heat stroke, they could have gotten hit by a car, he could have fell, she could have had a heart attack.   as we finally convinced them to get into the car, the man\'s eyes glistened with tears and said that he\'d never sat in such an expensive car. and they thanked us profusely....again and again.   we move up the hierachy, we drive fancy cars and dine at exquisite restaurants..but we forget about the other people. the other people that aren\'t quite blessed, the other people that have to think twice about buying a sandwich, the other people with kids that have shoes that are too tight.   did we forget along the way?  did we just get so jaded and cold?  my mom still stops and picks people up from time to time. she picks up kids from the neighboring school that are walking home, she picks up this old lady that walks to the market every morning and strokes my mothers face and goes sayang, sayang. my mother is an amazing woman. she lives in the corporate world and jets off to different places, but i am so proud of her. i am proud of her and my dad because i know they are good people inside.  things are changing though, there is so much violence and cruelty in this world. sometimes i worry that she might pick up the wrong person, because the world is just fucked up like that.   i hope that i teach the kids to be selfless, and patient. i hope that i am a good example to them. i hope that they understand the difference between right and wrong, i hope that they know how much potential they have in them and how far they can go in life.   i\'m starting to ramble, i started this post off with a range of emotions and now i am just tired. i hope that as you read this, you realize that you can make a difference. that you are special in your own right and you have a gift of your own to give to the world.  don\'t ever let go of your principles and morals. don\'t ever let go of life.      urlLink  \"But I know how I feel today. Since 9am, Tuesday morning, I\'ve been at home, on my couch, hugging my boyfriend. My friends are alive and my loved ones are safe.  Today, I am extremely thankful that I have no stories to tell.\"  '),(22806,664485,'20,May,2004','2004-05-20 00:00:00','   so i am on the campus walking in between classes when this \"gangsta\" (you know exactly who i\'m talking about...the white scrawny guys with gold earrings and fubu sweatshirts and pants about halfway down) comes up to me to ask if i have a joint. \"joint as in ciggarette? or joint as in  joint ?\" i ask...and he replies, \"the good shit man,\" okay \"dude\", who the fuck are you and why the fuck do you smell like ass? so i somewhat smile in labored breathing because the less i inhale, the better, and say sorry trying my best to get the fuck away from there. what does the bastard do? follows me to class. five blocks away. asks stupid questions like where do i work out and what\'s my freaking sign. i didn\'t think anyone in the world actually said that?!?!? and no, he isn\'t talking to me, he\'s tlaking to my boobs. and god almighty, get a hint! oh, and i don\'t like to be called baby doll either. so wrong.   there is a select group of strange people in pensacola, and somehow, i seem to meet them all.  can someone please kindly remove the sticker on my forehead that says \"come hit on me if you\'re strange and smell funky\"?    urlLink  \"Q-tips make me happy. I freakin\' love Q-tips. I love the way they feel as I roll them around in my ear, cleasing it of all that is waxy.  And don\'t even get me started about the joy they bring me when I twist them around in my belly button.  Oh, if only they made vibrating Q-tips...\"  '),(22807,664485,'18,May,2004','2004-05-18 00:00:00','   my boyfriend tells me i am sensitive, sometimes in a cute pathetic kinda of way and sometimes in a self-destructive hurt myself because others are hurting kinda way.   somewhere deep in me, i want to save the world. i want to bring peace and humanity and justice. i want to feed the starving children in africa with my hands, i want to join the peace cor, i want to take in every stray cat and dog, i want to open up my home to every child, every stranger, every person that just needs one. i want to love someone, give out hugs and kissies and make the booboos go away.  but sometimes, wanting is all that i seem to achieve. wanting isn\'t quite the same as doing. procrastinating, and all that other crappy stuff gets in the way. and sometimes i sit alone and blame myself, blame the world, blame god.   it hasn\'t been a good year for most at all...sometimes the events that seem so integral to life just bogs you down, fucks you up along the way. i\'ve never been good at dealing with death, as much peace as that would bring someone, as much as i know that that someone is in a place where the sun always shines and pain isn\'t even a feeling considered, i still cry.   so this is for baby muhammad, and rick.  this is for akari and grandma and brian.   this is for mun joon and all my other angels in the peadatric ward.   this is for yule and to the fragile existence of life.  i hung out with the kids today and it was beautiful. innocence and wonder, amazement and adoration. i loved every second of it. the way babies smell or the incomprehensible language that toddlers talk, the way i have stains on my pants from spitups and playing with the kids out in the yard. the way justin giggles and giggles and giggles. my happy child that one. or alex\'s fascination with strings....but the most amazing thing is loving them, teaching, nuturing caring. the way that they look up to you, the way that you look up to them too. the way that life is so magical and strange at the same time. the way babies grow into toddlers into preschoolers into children into teenagers into adults into mothers and fathers and into grandparents. the way that no matter who they are or what they become, that they will always be the children that i picked up after a nap, burped after a feeding, kissed their booboos and played the tickle tree with. the kids i read to on a rainy day and the babies i rocked to sleep. the little boys and girls that i worried my heart out as they trampled around the playground unaware of just how high or how dangerous it could be. the angels i cared for in a converted basement that was our sanctuary, the children i laughed, played, cried, and giggled with.   death is a strange painful thing. there\'s a curiousity that attaches to it but shatters you at the same time.   there must be a heaven, and in that heaven the kids are there playing, laughing, smiling. no more booboos. and in that heaven, there is also my grandma teaching uncle yule how to make sponge cake, while rick and brian are busy poking fun at the blogging world. yes, there is a heaven.  and maybe i\'ll see them all again there someday.   hugs.   urlLink   \"We\'ve all come a long way, and life did get very complicated. We all have something, but not everything. We\'ve had love, and saw love lost. We\'ve had our dreams fulfilled to the ends of the earth, but only after we\'ve had it snatched from us over and over again. We were given joy, but in unexpected ways. We are offered happiness over the horizon, but on the condition that we let go of the comfort of our shores.  A lot of things happened to us in the last 6 years. But someone once told me, God never burdens us with more than He knows we can bear. \"  '),(22808,664485,'17,May,2004','2004-05-17 00:00:00','   i\'ve turned the house topsy-turvy looking for my camera charger and i still have no bloody idea where it is. something tells me that there is a possibility that it might be making it\'s way to jersey right now, ahh, yes...and if it is, i know just exactly where it is. the rubbermaid container filled with electronics that i helped ryan packed this morning...and if so, fuck me.   it\'s 2:56 am, still monday night for me, and normally, this is where i come home smelling of stale ciggarettes and a mish-mash of colognes because gay men just cna\'t live without it. chick from emerald city offered to come over and sleep with me tonight, an innocent suggestion because she knew i\'d be lonely though when the cat is away the mice will play. i graciously declined, saying i wasn\'t up for it and told stories up my ass.   i miss drag night down at the club...i talked to sexy pants the other day, and he was moaning and groaning saying that i didn\'t care anymore. i\'ve come to realize that as you get comfortable with someone, you sorta lose sights on everything else. the last couple of months have seen my friends coming over, or doing lunch and i\'ve seemed to hung up my stilleto shoes and traded them for oven mitts. the house is impeccable though, and i am proud of what we have built so far together.   the dog and cat have been allowed back into  our  my bedroom, somehow, there\'s just so much space to still fill up.  i watched him pull away this afternoon, there was about an hour of silence right before he left of us just sitting together and lying in bed. he massaged my back and i breathed deeply that faint scent of his soap and closed my eyes. wondering. thinking. hoping.   i didn\'t start crying till he left, i know he needed to know that i would be okay on my own....and so as he left, i crawled back under the covers and stayed there a while. zigi by my feet and sasquatch on my chest, i watched bubba gumbo swim around in his tank for what seemed like forever till i finally cried myself to sleep. i am absolutely horrible with goodbyes.   my latest schedule is that i spend most of my evenings asleep and wake up at around ten at night and putz around, study, make some food, fuck around with the website, and then go to school. i think i\'m going to go over to marc\'s tommorow and play with the kids. ten angels for comfort....yup, i think that would be just nice.   perfect if not.   how was your day?     urlLink  \"I have come to the conclusion that, sometimes, rain can be the most beautiful thing.\"  '),(22809,664485,'17,May,2004','2004-05-17 00:00:00','   urlLink      urlLink      urlLink      urlLink      urlLink      urlLink      urlLink      urlLink      urlLink     he whispers in my ears, things i am embarassed to say out loud because they cannot possibly be true. only him, through his rose colored glasses can possibly see me that way. i move deeper, thrusting, moaning...begging for him to fill me up, soothe my hunger, quench my thirst. he holds me tight for a while...a long while, wanting to soak in all the moments that have build up to this moment, the night before. he leaves in less than twentyfour hours, a journey of sorts, a voyage he makes yearly right around the bikini season. our respective places we need to be, comforts, i fight against time and the tears that always seem to well up at the perfectly stupid moments.   he leaves bits and pieces of him around the house, our home that we\'ve been building on for the last seven months. a place we\'ve stopped halfway since we begin a year and a half ago. he centers the porch swing a little bit more so that i look into the garden instead of the traffic on nine in moments where i am craving for solitude and tranquility. and while he packs, he rearranges things to my liking, things i have compromised over and it is like the finality of knowing that he is going.   in bed, he talks to sasquatch and ziggi as if they are people, telling them that they need to take care of me. like my less than two feet cat and dog are going to protect me, all the same anyway, i suppose they will keep me company. he came home last week with a surprise he said, his eyes twinkling and him grinning from ear to ear...he leads me into our bedroom and on the sidetable he\'s purchased a beta fish and gently lowered him down into a glass tank filled with crystals at the bottom. it\'s my responsibility now to feed it he says, as he leaves instructions for all the other things that need to be taken cared of around here. the things i usually leave up to him, the things i generally don\'t really bother with.  it\'s funny how much independence you sometimes lose when you learn to work as a team, i can\'t quite remember what it\'s like anymore to have dinner alone or not go grocery shopping with him in mind. who am i going to buy string cheese and cases of dr. pepper for? who is going to pull the covers over me when i\'ve mastered the art of kicking them off? who\'s going to turn off my light and put my book back on the table?   i don\'t think i remember how to make dinner for just one, or dinner that isn\'t too spicy or weird. i\'ve mastered the art of fusion, of mixing asian and american spices into something decent.   the summer brings freedom, of doing what we feel like doing because we want to do it. of less laundry and dishes but more complication and so much loneliness.   i\'ve been reading through the last couple of posts, of episodes and episodes of sinking into a dark place and how he\'s helped pulled me through all those times. he\'s delaying his departure tomorrow so he can go in with me to the doctor\'s to get my shots. i\'ve left him to ask the questions, to grill the doctors, to say fucking fix her now you bloody idiots.   i\'ll miss you.  i\'ll miss knowing that you\'ll be home waiting for me when i go out. i\'ll miss telling you about my day and watching you do your silly dances. i\'ll miss your laughter, the way your whole face scrunches up or the way your beautiful gray eyes smile. i\'ll miss listening to you and stevie talk, when you\'re the cool older brother that\'s trying to be all stern and shit.i\'ll miss the way you\'re always out in the audience supporting me even if i don\'t see you, i\'ll miss knowing that you are just there. i\'ll miss coming home and finding little surprises along the way, i\'ll miss sitting by the beach and watching the sun go down with your arms around my waist and the waves washing onto the shore. i\'ll miss three am excursions to walmart for absolutely nothing and watching you at work. i\'ll miss fighting with you just to get my way, or poking fun at you. i\'ll miss tickling the bottom of your feet when you\'re asleep or wrestling in bed. i\'ll miss the way you know exactly what, exactly when...the way you kiss me when i don\'t expect it, the way you are okay with me being stupid or crazy or retarded or fucked up.   i know i want him in my life for ever and ever and ever.   i know i carry a piece of him everywhere that i go.  i know that i adore him possibly even much more than he does me.   i know what it means to be in love and to go through thick and thin...and then a little bit more.   it\'s four o five in the morning and he\'s fallen asleep on the couch in our living room because he\'s waiting for me to get done with my shit so we can go to bed. i love just sitting here and looking at him, wanting to soak it all in...of how happy love makes me feel inside, that even through everything, there are strong arms that carry me along the way.   i\'m flying into jersey in about a month, we have a weekend away planned. in between the two summer terms of intense classes with notes flying out my ass..this is going to be another crazy summer. i fly back somewhere in august to malaysia, it\'s the first year i\'ve missed the cheerleading finals and some nostalgic memory chokes me up. i am only scheduled for two weeks at home and that does not seem to be even nearly enough. there\'s so many things that i want to do, people and places.   i turn twenty this october, does it seem as weird to you that i\'ve been here for twenty freaking goddamn years? sometimes, i feel like i haven\'t accomplished enough, yet there are moments where i know that i\'ve come a long long way from where i couldn\'t possibly imagine i would end.  so here\'s to summer and life, for all the quirks it sends along the way and all the brickbats we somehow stumble through.   ((hugs))     urlLink  \"And we toasted our clever selves, and drank. To those who stick by us. To those who love us when we\'re ugly. To those who forget our birthdays but remember to put out the bedside light. To those who don\'t call us in the middle of the work-day but never fail to ensure the chauffeur is at our door steps when we want him to be. To those who always show our parents respect. To those who hardly say \"I love you\" but cannot imagine life without us.   To those who drag themselves to work every day just so we can sit at home and drink the nectars of their success.  And it tasted just as sweet as it smelled.\"  '),(22810,664485,'11,May,2004','2004-05-11 00:00:00','   something stops and clicks, something in the air that on intake, seems deathly still and suffocating. perhaps it is only in my mind, perhaps i am caging myself in a prison cell and erecting barriers where they don\'t belong. but then again, perhaps i am only protecting.  survival of the fittest.  '),(22811,664485,'07,May,2004','2004-05-07 00:00:00','   everything fades into a blur of dissonance and heart ache, joy, sorrow, pain, elation. there is a break between reality and fantasy, of falling through the cracks of life and hope that loosely, tightly, binds it together.   the anticipation of something worth waiting for, worth fighting for with bleeding palms and scraped knees. the summer approaches like the boogieman and the monsters under your bed. he leaves in two weeks, ten days to be exact and the picture memories are in full drive.  snap. snap.   this and that and this and that, intricate moments held on so tightly, memories ziplocked, double-padded and stored away with special care.  the house is getting emptier as summer rolls around, the walls begin to echo and the flies begin to weave their stories. people are coming and going....going most of all and the roller-coaster ride of inferno begins. a double full load with a day to spare in between, 15000 miles to fly, three coasts, two major seas.   results of sorts arrive on monday, anticipation. anticipation. anticipation.   breathe.  '),(22812,664485,'27,June,2004','2004-06-27 00:00:00','   life is a little crazy at the moment. i don\'t know if i should tell you about the perfect shrimp scampi-alfredo linguine jessica and i concocted tonight or about my nightmare down at the memphis airport.   or then again, maybe i could tell you about how i\'m going to take three express classes for this part of the summer which would make my ass in school at eight and home at ten at night. stupid idiot racheal.  or then, maybe i should tell you about the incredulous amount that the bookstore charges for textbooks nowadays and how if you click over to half.com, it\'s 1/3 the price, and then you shove it back their(bookstore) asses and sell it back to them(buyback) at the end of the term for three times the price. ahhh...invesment.   or maybe i will tell you about how much i miss ryan and how much more i realize and feel that now that i\'ve had a taste of him. it\'s like luring a bullimic with a teaspoon of fudge. pure evil.   oh, and did i tell you that the dynamitez placed second in the nationals?   :)  or that kim had baby catalina last night after a good 21 hours of labor?   and in her words i say, \"it was all soooo worth it!!\"  there\'s also the five books a week plan that i\'m following religiously in an effort to uhm, improve myself. to put it quite crazily, i\'ve literally got a suitcase of books that i am determined to finish by the summer.   at the beach of course.   i\'ll be the one in betty boop sunglasses and anime on my bikini with a blue/yellow plastic explosive-looking thing that sticks out somewhere in between the rolls of my belly fat.  so maybe i\'ll end things and sort of summarize it with a little happy dance and good cry that things are going pretty okay for the most part, and yet they could be much worst(better).   depending on your perspective.   gimme a quick hug and do the chicken dance with me because i have to pee.  '),(22813,664485,'23,June,2004','2004-06-23 00:00:00','   i watched as he pulled away from the terminal, 5:40am, a weekend spent in glorious coupling, sharing, conversations....so many things to remember...the flight from hell that took two days, his friends, the boardwalk...his little place away from home.   i missed this, missed you.   it\'s funny how big of chunk is missing from my life.   the first few hours together like a reverie, deja vu. what, where, who?  and if i could, i\'d turn around and get on the next flight.   say fuck school and responsibilities.  fuck everything else.  because three days was too short when you want to spend forever with someone.   laughing. i miss laughing with you.  or watching a good movie together.  this place smells the same, a reminder. i feel so odd(off) without you.  i hate this summer away bullshit.   the house feels so empty, and menacing, in some way, without you.   back to dinner for one and a bed to myself alone.   back to life, and going ons, and day to day to day.  monotonous.   funny how a flight from newark to pensacola can take forever.   funny how it breaks my heart to leave but even more so to go.  funny how i love you so much, in spoken and unspoken, in cups of coffee and spoonfulls of sugar...in sunsets at the back of the pier, in crazy nauseating rides we take together.  i miss you.  '),(22814,664485,'15,June,2004','2004-06-15 00:00:00','   our computer room is fucked to hell. it\'s a sweltering 84 degrees and the little tiny bity fan does not do justice. something tells me that if the landlord doesn\'t get his butt in gear, the racheal-jessica wrath is going to come out.  let\'s just say that the ducts were supposed to be installed last october.   uhuh.  ---  i just got of the phone with chinesewhiteboy. chinesewhiteboy is so adorable. :) i haven\'t talked to him in a good couple of months, and i was talking to his partner and apparently, their  urlLink fancy-smanzy company  features a US number that you can call and it\'s patched to hong kong and wales and all the different ports that the rest of the board is at. kinda like calling aol customer service and getting routed to india. very cool.  free for you and cheap for them!  he\'s talking about backpacking in europe and stopping over in switzerland at his parent\'s chalet and what a pain it is to come back to work....as i am sitting here with my ass up with finals and assignments and a shitload of studying. the fantastic news though is that in two days, my ass is going to be on a plane to new jersey. yay me! and in a month and a half, that same ass is going to be in amsterdam and then malaysia. sweet me.   we\'re talking about friends and high-school and that certain nostalgia that comes with growing older and the crisis of wallowing away from your teenage years. it\'s so weird to sit here and watch as the girls from our cheerleading team have grown up into adults, even stranger to watch them blossom into young women. i am so proud of all that they have achieved. this is the first year that i can\'t make it to the cheerleading finals and that shit just breaks my heart. it\'s crazy how far they\'ve come and nice to think that i was a part of it for a while.  ---  ryan and i are going to phoenix in november. i called in and booked our tickets today. :)  i don\'t know if it\'s going to be good or horrible or fantastic but we\'ll see.  think all my family members at the same place and the grilling!!!  mervin and koren are finally getting married(about damn time!) and that just makes me so happy! after ku-chang and grandma and all, it\'s nice to have a good reason for the family to get together.   bryan and yee mei are coming in from singapore, davinia and significant other from australia, apak from taiwan, apakleung from shanghai, my parents from malaysia, and some friends of theirs from tokyo and then there are the ones from this side of the continent. it\'s going to be interesting...it\'ll be the first time my extended family meets ryan and i am excited, proud and scared out of my mind!  i was talking to my aunt today(kumah) and she\'s like bring ryan! bring ryan! we have to meet ryan! bring ryan!   lol.  :)  we\'re going to miss out on the singapore wedding though, because negotiating to go across the country was a dillema in itself, i don\'t think we\'re even going to attempt the crossing the seas.  think of it this way, some day, that\'s going to be us.  only it\'d be missouri and malaysia.  and he isn\'t german-american-japanese.  crazyness.   woohoo!  (i\'m so excited!!)  p/s were the pictured of the kids yesterday just not absolutely fucking adorable?!?!?   :)    urlLink  \"Today is good. \"  '),(22815,664485,'14,June,2004','2004-06-14 00:00:00','   help! share! play nice! be gentle to the babies! racccchhhheeaaallll! i need a hug! my name is pumpkwin! weee-wayneee! wave bye wave bye wave bye! can you say fiiiissssshhhhh? splat! boo-boo! can we watch shrek? can we watch shrek? can we watch shrek? i like that boulder, that\'s a niiiceee boulder! don\'t hit! that makes me sad. i need to pwwweeeeeee. i need to do pooopppiiiieeee. i\'m dooonnneeeee. samuel\'s tooting. i had it first. wanna switch? don\'t put it in your mouth!! do you have anything in your mouth? do you want paci? no paci. do you want milk? i want milk! i want juice! i want snack! water play! water play! water play! uhm, actually... tickle tree! does my name go this way or this way? carry me! me! me! i wuv you.   :)      urlLink \"Three pieces of rice went shooting out and stuck to my monitor.  I\'m classy.\"  '),(22816,664485,'13,June,2004','2004-06-13 00:00:00','   she strips off layer by layer, taking the glitter off, toning the make-up down. a little less out there, but still edgy. a little less anxious and crazy, but still kooky.   reflecting through almost eighteen months of pleasure(torture), delight(sadness), love(hate)...they\'re moving into a new city. her lover boy and her, building more castles and painting more vibrant colors in what was once her comfortable black, white and grey existence(pain).  she reads through all the things on paper that are said about her. and someone reflects that that seems like a different person. the cheerleader, the roller blader, the orang asli bitch, the school school school happy....only she doesn\'t think so, not anymore. i\'m still me, she whispers. just me.   and as she\'s gathering the testimonials and references, she realizes just how much love is around her. the praising joy of her teachers, the pride and confidence of her admissions officer, the people around her that pick her up and push her forward. hardly, gently, softly, subtlely.   --  sometimes, i\'d like to hide away from the world a while. just nestled between the cracks of a rock and a hard place...or something along those lines anyway. it\'s when you\'re crying, and you don\'t know why. is when you miss someone, and loneliness creeps behind you and goes boo. in a big terrifying way. a way that catches you when you aren\'t expecting, a way that says watch out, keep tight, fuck you.   empty bed. empty big bed with many pillows and rainbow stripes. washing  our  sheets, sooner rather than later, i can\'t smell his smell that comforts me, only of purex and downy, only of freshly laundered sheets and nothing else.   three days, three days.   ---  be quiet.   the boogiemen are watching.  '),(22817,664485,'13,June,2004','2004-06-13 00:00:00','   this week has been crazy to say the least....in between trying to adjust again to tiny pensacola after five days of absolute bliss in chicago, i am leaving again this thursday to go to new jersey.   to see the other half.   to have wild fantastic unbelievable sex.   to try out the new ky warming liquid.   to fuck my boyfriends brains out.   :) :) :)  anyway, before pleasure comes a shitload of torture! it\'s the last week of this term and so finals are biting me in the ass with all sorts of ridiculous 1500-word essays that talk about stupid things like literature and nutrition and fitness and hemmingway. pah...hemmingway shemingway. then there are the final exams to cram for, and yes, cram, because you know how brilliant my study habits are. :P  then there is the insane madness that generally goes around in my life because i think above the sticker on my head that says \'assholes and weirdos hit on me!\' there is this other one that says \'chaos, panic and disorder HERE! HERE! HERE!\'. but ahh, the bliss of being neurotic and crazy. oh. so. fun.   i spent the whole of yesterday working on revamping the childcare for summer fun!! lol, we went shoppiong for buckets and shovels and sandcastles. fishes, inner tubes of every adorable kind, snacks, crazy curtains, ice-cream, popsicles, shrek yogurt! oh god, the things i would give to be a child again... theresa(the childcare owner fantastic lady person) decorates the childcare specifically to every season and just the amount of dedication she puts into it is a wonder of itself. then we painted the kids curiculum(sp?) room....lol, it\'s now two sides of yellow and two sides of ocean green. now someone told me that yellow will make the kids hyper...OH SHIT.  there\'s also the current issue of the laundry. FUCKING HELL. i did two loads of laundry the last time i blogged and there are two new loads fuckinh with my head. argghhhhhhhhhhhhhh!!! i. hate. ironing. sticking it in the washer and transferring it to the dryer is fine...but goddamnit, ironing is such a suck ass bastardy creation. you, person that scrunches up your nose at my wrinkled clothes, kiss my yellow ass big behind.   fifteen more hours of the weekend..  hm..  wonder what sort of shit the boys and i can get into tonight?  :P      urlLink  \"I just want to bang on the drum\"  '),(22818,664485,'09,June,2004','2004-06-09 00:00:00','   i. signs i need to do the laundry?  i\'m wearing ryan\'s boxers and even with him not around here, both our laundry baskets are full; vacation over!  ii. shredder, our ex-escape artist, came for a visit the other day and he looks so good! :) i miss the bastard. lol, cj(his new owner) is taking him to texas and this 70 pound dog is going on a plane with him....?!?!? zigi and sasquatch had a wild time with him though, even though they are only about 17 times smaller and he\'s trying to hump poor zigi.   iii. this dude in the computer lab last night tried to grope me.   iv. i forgot to take medicine today. blah.  v. one more week before the summer term a ends!  vi. one more week before i visit ryan!!(need sex. need sex. need sex.)  vii. stan, the director, caught me in the halls the other day and asked if i would like to take part in the summer production of fame the musical. FUCK YEAH!!  viii. the old man loves me. :P...and i like pink.  '),(22819,664485,'08,June,2004','2004-06-08 00:00:00','   \" broken bones and promises.   you didn\'t have to tell me who you are, i knew right away. why are you calling me? you tell me you read about my dad. you told me you were sorry. you wanted me to talk. i admit that i blew you off. what else am i supposed to do? am i supposed to play nice and talk about my feelings? do you really want to hear about how i\'m feeling? then we can be nice and civil with each other for a few weeks (days?) before one of us is screaming at the other for something or another. no thanks. i think i finally learned my lesson the last time. i could have skipped all the nice bullshit and gone straight into yelling, but really nobody needs that shit either. so i just blew you off. i\'m an asshole. whatever.  what else am i supposed to do? i can\'t decide whether i love you or hate you...and some days i\'m not so sure that it\'s not one in the same. it\'s a comforting thought to think that i have a place to run...i just wish it wasn\'t you. i am nothing you need to be worrying about anymore. i am not your problem. you made your choice and i made mine.\"     sorry. i don\'t know. whatever.  '),(22820,664485,'07,June,2004','2004-06-07 00:00:00','   jared\'s dad isn\'t doing too well.  i read, and i guess a part of me hurts for him...the things he must be going through. and i\'m sitting here wondering if i should call, or send him an email or if it would be better off if i just didn\'t say anything at all. what is the protocol in things like this? and if there is, i think we\'ve hardly been conventional in the first place.   we\'re on different roads, and we have broken promises. i\'ve always tried to keep some sort of friendship going i think, no matter how twisted or fucked up, because in some realm, i think we owe both of us that. love didn\'t work out in a way that it was supposed to, but what puts a restriction on friendship other than us?  i don\'t want to start the ball rolling on another tumultous journey to hell where we scream and shout and cry, neither do i want to leave him with more grief than he must already be going through. but then again, i cannot but sit here and just do nothing. it\'s like watching someone drown and just looking. bad metaphor, but you know what i mean.   i\'ve just spend the last five days in chicago with my parents. and everyone knows that i am daddy\'s little girl. this hits so close to home because this weekend, i also realized that my parents are fifty one and growing older. that they get tired easier and that someday, they might get too weak to fly. that\'s a dramatic take on things i know, but it\'s like you suddenly realize how fucking precious life is.   my parents have been so amazing. i was a child again, and i got hugs and kisses and was pampered and it was just so nice. albiet it was a whim to just decide that i was going to go and book my tickets and just go. it was nice to do that though, nice to have that option. i was worried my dad would balk because it was extra charges on the credit card but instead they were so nice about it. they paid for everything and my mom cried when my dad told her i was going to meet them in chicago! when i met her at the hotel, she gave me one of the biggest hugs ever, and i remember what it felt like to have your mom and dad by your side all the time. the security, the comfort, the warmth. she told the driver that was taking me to the airport today to take care of her \'precious cargo\'. and my dad just about made everything perfect for me, eating at restraunts i liked, going places, humoring me and walking all over town taking \'artsy\' pictures with my brand new canon ixus camera they brought for me(!) but one of the most important gestures to me was that he brought me and ryan his and hers matching watches. it\'s this funky little metal thing with the numbers all groovy and black leather straps which is the kinda of thing we would like cause we\'re all weird like that but what touched me was that the gift wasn\'t just about me. it was about us. it was like acceptance to me, and if you know my dad, he\'s the gentlelest sweetest person on earth but if you hurt his daughter he\'ll fucking break your back and kill you. a little overprotective might be an understatement. and so that gesture filled me up with so much joy and warmth, that my dad said to my mom to get us the watches and brought it all the way from malaysia. and just for laughs, he also brought us caps that said \"you know you\'re a redneck if...\"  :)  we stayed an absolutely beautiful boutique hotel. a little pricey and crazy happy with their charges but it was decorated in this artsy fartsy style it was so cute. when you call the front desk they answer with your name, with housekeeping and room service and all the extra little things that make a holiday  a holiday.  hotel 71 people, damn nice hotel minus the stuffy european la di da stuff you normally get at hotels like that.      i miss my parents already, but it was so so so nice to spend the weekend with them.   old man told me to shut up when he picked me up from the airport because i must have gone on and on and on and on about it for the whole drive back.   i don\'t know, it\'s a little bitter sweet i suppose. like there\'s a strange aftertaste to having a fantastic weekend then realizing things.   one of the things my mom brought me was my grandmother\'s diamond ring that she always wore and i inherited. i might have shed a tear or two when i saw it because i don\'t think i ever saw grandma without it and it was special. i don\'t know. i\'m being sappy.   umpacking calls and my cat, dog and fish are begging to be pampered.  love you guys, stay safe.     urlLink  \"i wish...i wish that things would pick up for me. don\'t i deserve good things too?\"    in more ways than a million.  '),(22821,664485,'29,July,2004','2004-07-29 00:00:00','   urlLink http://www.msnbc.msn.com/id/5520120/    urlLink http://www.msnbc.msn.com/id/5519881/   we try and make it on most mondays, rushing after class that ends at 9:30 for a quick shower than down the road to the club for the  urlLink infamous drag nights . and it\'s hard to believe that our somewhat \"safe\" place in little town has been intruded. how do forgive someone who takes the life of another person? someone he called a friend? a good friend at that.   it\'s betrayal and anger and hurt.   it\'s hard to sit here and think that that could have been ian, or sarah, or bryan.   i sit here trying to sort through the surge of emotions running through me, how, why, ???  i am disgusted.   who made you god to decide the life of someone else?  and in that line, did god look down on someone whom he called his child then looked away?  there\'s a benefit on friday at emerald city.   all the proceeds are to help the family with the funeral expenses.   i said a prayer today, then cried for humanity.     urlLink \"He was a friend to many at Emerald City. He was taken violently from us at the age of 18. His  was a gentle soul that will be missed.\"  '),(22822,664485,'25,July,2004','2004-07-25 00:00:00','   idle chatter in a white room with empty spaces, no color(distinction), me and you, flesh on flesh..exploring.   he reminds me what i am made of...on days where the world seems to be fucking me over and shoving a knife up my ass. on days where i get jipped by our landscaping company and yelled at by a homeless girl because i gave her a five instead of a ten. on a day where doubt lingers like a bitter aftertaste and the surreality of growing up becomes too real.   it seems that as the teenage years flash before your eyes and the impending stage of being a grown up comes, that the question \"who am i?\" haunts you like irritating burs on your puppy that take forever to pluck out. like a splinter deep enough to irritate but not deep enough to draw blood and definitely not shallow enough to pinch out. like why are red sunsets so beautiful when it signifies the polluted air?  there are so many silly inconsequential things you learn in an earth science class. like did you know that the water in your toilet bowl spirals counter clockwise in the southern hemisphere and clockwise in the northern hemisphere? or that the reason why red soil is abundant in florida and malaysia is because of the humidity that makes the iron minerals in it oxidate?   there are some days, where i curse my voice because it borders between a man and an old woman. like somewhere in my past life i was a baritone but when i came into this dimension, they forgot to alter that fact.   i hate how in musicals, all the leading roles are always for sopranos. it\'s like a big fuck you sometimes. discouraging to say the least.   but yet i am gifted in an alto voice. a rich sexy sultry tone that comes out from somewhere deep within me. strange.   why am i so shy?  ryan bitches at me day after day that i am too hard on myself.   and maybe i am. but sometimes, i find that being able to say that i am something, beautiful, intelligent, amazing, fantastic, seems so rude. it makes me want to hide in a corner and pull the covers over my head.   there is a disposition that seems to scream out to others. this front that i have that i am loud, and outspoken, and confident. oh god, if you only knew the things that went in my head. that i am good with boys, flirtatious....if you only knew that when someone chats me up i look for the closest exit. perhaps it is the pathetic clutch on a facade that makes me good at what i do on stage. the ability to take on another persona, the ability to slip into the mind of someone else. i don\'t know.   i blush inside, where no one else sees. and in that special place hides all my fears and wants, hopes and needs. striving to be a somebody in a world that sometimes wants you just to conform to a nobody. there are certain things i always stand up for, children, abuse, cancer, justice...but with that, there are many others that i would like to stand up for but don\'t. me.   my mom and dad tell me i am horrible at saying no. which is an understatement to those that know me well and absurd to the others that think they know me but don\'t. here i am, this gregarious crazy in your face girl that seems to be loud and friendly and all those things at once, but the ones that live with me, the ones that have tapped through the layers know that inside, i am scared, and shy. negative, and agoraphobic, mad and dellusional.   as i am peeling through the layers and finding myself, i realize that there is a comfort zone...a nice thin line that i thread on, if i let myself. a chance for freedom and still, hapless self examination.   what was my point in this whole ramble again? i\'m not quite sure.   i think the heat is getting to me.   hugs  :P  '),(22823,664485,'23,July,2004','2004-07-23 00:00:00','   &nbsp;  lol, contrary to what i told daisy the other day, most of the time, while i mantain that i really am the boss, the little guy really keeps me in check.   :)   this picture makes me grin.   heh.   i picked up that shirt for&nbsp;wee wayne&nbsp;while i was in jersey and it says,   \"i\'m boss. no questions. no buts. we just do things my way\"   or something in the lines of that anyway.   :P   &nbsp;  p/s doesn\'t the angle of the picture make it look like the little rascal took the picture on his own?   and what is that cheeky little smirk trying to say?   *miss racheal, you are so stupid.*   &nbsp;     urlLink \"and I am actually enjoying it. For all the I-want-to-strangle-you-and-rip-your-heart-out moments... there have been many more \"awww\" and laugh-out-loud-til-your-tummy-hurts moments to compensate. Like when I told Oskie, \"now, don\'t pee-pee in the pool okay... it\'s not nice.\" And he answers with a deadpan face... \"But Daisy... I like it!\". \"  '),(22824,664485,'22,July,2004','2004-07-22 00:00:00','   it\'s that one day where your \"you know you\'re a red neck if...\" hat is worn backwards along with your 9.99 Gap baggy jeans found in the men\'s section&nbsp;for that \"gangsta\" look with a shirt with stripes going the wrong way that fit snuggly around your waist and a neckline that make your boobies say peekaboo!   it\'s when you have that&nbsp; precious lunch break between classes and with backpack in hand, you run to the mall for some shopping theraphy and to meet a friend of&nbsp; a friend of a friend that owes you fifty bucks.   and it\'s in those 45 minutes that you run into the ???. (can you really call that&nbsp;one&nbsp;night stand anything?)  you glance over and you go, oh shit.   then you casually saunter by trying to pretend that you didn\'t see him. or that he doesn\'t recognize you. or oh god, you\'re just embarassed.   but of course he walks over to you and kisses the air around your cheeks and you go...HEY!!, as if you were just about to do the same thing.  and you suck in your gut hoping the layers of fat around your tummy will somehow dissappear.   you stand a little taller in your cute ass pink shoes and smile. a lot. okay, maybe a little too much.  he makes small talk and so do you, about life&nbsp;and the \"good ole days\" whatever the hell that means?   hm.  and then you say you really really have to go.  and walk away with your cheeks flushed and your brain blaring sirens.   and late at night, after a whole day of classes, as you\'re emptying out your pockets to take a shower you realize that he\'s slipped his card&nbsp;into your side pocket, that you couldn\'t feel because those pants could have fit two of you, for a booty call.  damn fucker.  &nbsp;  '),(22825,664485,'20,July,2004','2004-07-20 00:00:00','   life is a cacophony of accidents and reality biting you in the ass. instances where everything flows together smoothly and harmony is achieved by merely breathing. trudging on, waddling through the mud and debris, feeling the sediments wash through your fingers. watching the lightning and the thunder engulf you and roll pass by. &nbsp; it\'s a melancholic tuesday for me. something about the air that reeks of reflections and permeates of the future. &nbsp; conversations that have reminded me of a being, a girl, a person i once knew. relationships past and somehow forgotten, memories that seem sketchy and not quite there. dreams and nightmares that haunt and enlighten. connections with people on the other side of the world....remembering them as they were. cheerful, happy, sometimes dying in your arms, sometimes there but not really there...alive but knowing they are dead. what happens when you close your eyes late at night and journey on into dream(nightmare)land? &nbsp; is grandma really gone? i catch myself sometimes unconsciously picking up the phone to dial grandma and see how she is doing. or i wake up from a dream that seems so real it is a nightmare. where i am holding her and she is taking her last breath, or we are joking about something, and there are all these words and conversations that seem like a blur once you wake up, you don\'t remember what it was all about, but you know that in those precious hours that you call sleep, she was so alive.  &nbsp; i think some part of me aches because there is no closure. because the last time i ever saw her was behind glass doors fighting for her life and the feeling that i was abandoning her. &nbsp; some things are so strange on tuesday mornings when it isn\'t quite monday anymore and definitely isn\'t wednesday yet.  &nbsp; when your schedule seems to relax a little in between classes and general madness.  &nbsp; when you\'re done with laundry and housecleaning and grocery shopping.  &nbsp; went to panama city for the weekend and it was beautiful. a chance to relax and catch up with life. nature. beauty.  &nbsp; the distant ticking of swinging golfclubs meeting golf balls. the splashes of happy children in a swimming pool. the busy busy busy disconcorted harmony of tourists bustling. things. moments. instances.  &nbsp; life is so strange and beautiful and exquisite and depressing and joyous and incomprehensible and overwhelming.  &nbsp; hugs. &nbsp; :) &nbsp;  '),(22826,664485,'13,July,2004','2004-07-13 00:00:00','   i meant to write this down the other day. and i think i started with three paragraphs of incoherent rambling. somehow, at half past two in the morning, it starts to make sense. somewhat. racheal sense anyway.  i hate how my country defines  melayu ,  cina ,  india , and my personal favorite,  dan lain-lain . a distinction between color, and race. a distinction between who you are, which family you were born into, what shades of brown and yellow and black you are. i wish we were all green, or maybe purple. i wish jealousy was a folklore, something my great-grandparents told us about with anecdotes and morals heavily threaded somewhere into it. i wish jealousy did not exist.   but yet, my country, lives in h a r m o n y. harmony is such a nice word isn\'t it? the mak cik three doors down and the ah-sam next to me with the aunty two doors over.   but here in america, the land of the free.....there is still a classic distinction between the black and white neighborhood. and then the \"lain-lain\" come to play to. those that don\'t quite fit in(foreigners) lumped together into a community of sorts. Community.   and in america, and malaysia; there is still seperation. seperation among races, seperation inside races. white and black. white thrash and elite. who the fuck came out with lower class, lower middle class, upper middle class, and high class anyway?  what exactly is freedom to me?  i have never considered myself, third world and all, to be oppressed. or to not know the meaning of freedom. yet culture shock can do strange things to your system sometimes.  was i brain-washed?   or do i willingly agree to the concepts that i have been taught since young?  things like you just don\'t say shit about your government.  things like internal security acts and being politically right.  things like whatever they tell you, you just stick your tail between your legs and listen.  and sometimes it makes me mad.  there are days, where i know i don\'t stand up enough for what is right because in school and in our culture, and in the way we were raised...you took the shit as it came.   and there are days, that i am so angry with myself because i am bleeding and dying and crying and suffocating, but yet the pressure continues to boil over because i don\'t know how to verbalize. or fight. or go against the \"right\".  there are times where i am upset with the government and their stupid policies and the silly political crap they play with one another. but can you really say anything against the prime minister? or barisan national? or...islam?  but yet i know i\'ve never quite walked the straight line. i\'ve always strayed a little, shocked others. wore tube tops and shown my belly ring. altered my uniforms and said what was on my mind. asked questions in class and challenged a concept.   taken control of my life.   when i came to america, it was strange and exciting and so fucking terrifying at the same time. why does the old man openly state his opinion on what he thinks about the bush administration? people go to jail for that in my country.   or in my first semester, when i was still a journalism student, how i didn\'t want to write that paper at all discriminating my college in a national paper i was assigned to write. how i argued for hours and hours. how i wouldn\'t do it, couldn\'t do it. dreams shattered the moment you realize you can\'t quite get past the fear. because no matter how \"politically incorrect\" i am inside and around my circle, i can\'t slander. i can\'t say bad things, even if they are true.   but then again, malaysia has taught me so many things that america probably never will.  respecting one another\'s religion. that religion is not a bearer.  learning three languages.  and definitely celebrating about five different ethnic festivities. so what if i\'m not indian? or malay? or kadazan? kongi raya. i like that word.   the meaning of \"respect\". something that pains me in the western civilized world because three year olds cuss at their mothers and kids call their parents by name and who the fuck is your neighbor anyway?  patriotism. red white and blue. my second home. fireworks and barbeques and celebrating the right to be treated equally, to have all the constitutions, to say what i want, whenever i want. to still feel a surge of pride when someone waves a malaysian flag, to put my hand across my heart after a salam, to sing god bless america the same way too. to be half of something and half of something else too.   being raised in america, as stupidly drunk as i get, drinking isn\'t the issue. i\'ve lost more friends here to drunk driving than i have in malaysia because it isn\'t the \"it\" thing. it isn\'t a rite of passage. it just is.   where does my loyalty lie?  who am i?  half american? half malaysian? rojak?  what am i?  there are goods, and there are bads.   a long time ago, someone special and wise wrote this.  you got to take the good, the bad, the evil and the happy....and in it, you will find peace.    happy (belated) independence day and happy (beearly) merdeka.   go live out freedom today.  hugs.       urlLink  \"Desiderata  Go placidly amid the noise and haste, and remember what peace there may be in silence. As far as possible, without surrender, be on good terms with all persons. Speak your truth quietly and clearly; and listen to others, even to the dull and ignorant; they too have their story. Avoid loud and aggressive persons; they are vexations to the spirit. If you compare yourself with others, you may become vain or bitter, for always there will be greater and lesser persons than yourself. Enjoy your achievements as well as your plans. Keep interested in your own career, however humble, it\'s a real possession in the changing fortunes of time. Exercise caution in your business affairs, for the world is full of trickery. But let this not blind you to what virtue there is; many persons strive for high ideals, and everywhere life is full of heroism. Be yourself. Especially do not feign affection. Neither be cynical about love; for in the face of all aridity and disenchantement, it is as perrenial as the grass. Take kindly the counsel of the years, gracefully surrendering the things of youth. Nurture strength of spirit to shield you in sudden misfortune. But do not distress yourself with dark imaginings. Many fears are born of fatigue and loneliness. Beyond a wholesome discipline, be gentle with yourself. You are a child of the universe no less than the trees and the stars; you have a right to be here. And whether or not it is clear to you, no doubt the universe is unfolding as it should. Therefore be at peace with God, whatever you conceive him to be. And whatever your labors and aspirations, in the noisy confusion of life, keep peace in your soul. With all its sham, drudgery and broken dreams, it is still a beautiful world. Be cheerful. Strive to be happy.   Max Ehrmann (1872-1945)\"  '),(22827,664485,'11,July,2004','2004-07-11 00:00:00','   i had a long conversation with an old friend last night. eighty two minutes and nineteen seconds exactly. i know this because my silly purple phone that lights up into a rainbow of colors tells me this everytime i hang up the phone. i think we started at around one or so, a nice surprise on a day that seemed good, on a day that even alone at home, with the cat and dog, with the thunderstorms and lightning creating art outside on our porch...things seemed alright. peaceful i suppose.  caller id read private. but then again, all my calls seem to be from private numbers these days. there\'s something about hiding beneath the shadow of uncertainty, or at least lurking in the dark. i don\'t know what it is, i always say it\'s because stupid navy boys get my number and ask for sex in the middle of the night.   it wasn\'t like the usual...not the tranquil setting of a certain coffee joint that i like, the whitewash of muted conversation in the background and perhaps krall washing ashore with her rhythms. nor were there other people to divert our conversation to and create meaningless gossip when things got a little too heavy.   it was me. getting tangled up in another book, watching my candles slowly burn away (this candle shit is a fucking expensive hobby), fingers sometimes reaching for that sacred area between my thighs and my evil cat and dog alone with me in my much too big bed, each begging for attention in their own way. listening to the night sounds, falling asleep, waiting for my \"boo\". :P  it was strange and exciting and melancholic all at once. it was like talking to a stranger yet talking to your best friend all at once. it was a conversation that picked up from where we put it down the last time. as if nothing at changed. as if we were both still kids with raging hormones in kuala lumpur figuring out our lives. i like how certain people can make me feel so comfortable with myself, and even more, i like how comfortable it feels to just be me. to not have facades and faces masking the demons deep within. i like listening.   to stories of japan and how life is treating him. to his latest crazy encounters or the general perception of foreigners and us. foreign. about him and her and suburban city life as i told him about fourth of july and my special date with sexy pants at hemmingways. as we talked about my farewell party and the sorry state i was in at the end of the night and explicit details of the shit i somehow always manage to step both feet into while embroiling myself in much more. we talked about when we were young, and crazy. about late night stops at the mamak store and exchanging pirated vcds, he reminded me about our first kiss. the ackwardness of not knowing what went were and why our bodies were reacting in such a strange way. discovering our sexuality at fourteen for the first time together. in a dark cinema, tanjung golden village in one utama, i remember the exact seats we sat at, arriving for the movie just at it started and having front row seats, cranking our necks so far up it ached for the next week.   it\'s funny the things you remember late at night when you are prompted.   by the time we got off the phone it was late, and an international call that probably cost up his ass at that.   but it was good.  and it made me treasure all my memories of growing up, of finding the paths and digging through the past to get to where i am heading. of the security and somewhat calmness in my life that i\'ve somehow stumbled into. it\'s strange how things work out in the end, right where you never expect them to. it\'s funny how that when you learn to love someone, you never forget nor regret, no matter how questionable.   i am glad i met you to get to where i was going.   as many others (have).   as many more (will).  every bitter tear will mean a whoop dee doo laugh down the road.   hugs.      urlLink  \"So, yeah, basically what i\'m trying to say is: here\'s some shitty poetry.\"  '),(22828,664485,'09,July,2004','2004-07-09 00:00:00','   dark skin on white, and white caressing folds of olive.   tanned, one brown, one with tints of reddish pink.  i think about those nights, constantly.  a bed and two lovers.   moaning.  becoming one.  '),(22829,664485,'01,July,2004','2004-07-01 00:00:00','   i tell this story repeatedly about twice a month because it makes me smile, and my heart does a little happy dance.  it\'s a story of selflessness; the act of giving and recieving.  a good reminder of money, it\'s value and siblings. family.  your blood   i don\'t know how i stumbled into this gem, i think it was one of those hot malaysian afternoons where you\'re stuck in a traffic jam for hours and hours and you\'ve exhausted all conversations on the now, and the current boyfriend, and school, and well...you know that shit.  and somehow, my dad tells me about how his brother used to wait up for him in the canteen to share his coke with him. and if you know my dad and this particular brother(my dad\'s the baby of 2 brothers and 2 sisters), it is that they have three passions in common.  i. their love of coke.  ii. their love of video games and all things technological.  iii. their infinite source of lectures for their only daughters.  i could tell you so many stories of this uncle and my dad. a good digression i think.  i could tell you how davinia and i used to run up to the wrong daddies because they looked exactly alike. or if your back is turned, and you\'re not really paying attention, how they sound exactly alike, intonations at all the same places, talk subjects so similiar even hand gestures and to a certain point, personalities.   my spend thrift genes come from my dad i think, well, then again, could be my mom too. so he tells me this story about how he would use up all his allowance and his brother would know that and save half his coke for him.  i don\'t know about you, but when you\'re at that age, i don\'t exactly think that you would be thinking about anyone else but you.  and so it\'s a nice story that reminds me to love my brother because he\'s the only one i\'ve got. or that he looks out for me when i\'m not looking.  my mother conspiratiolly told me the other day about how she and my brother were talking and he said i was special and very gifted. of course, he probably wouldn\'t tell that to my face but i know he loves me, and cares for me, even if he gets on my nerves and is overprotective as shit.   so yeah, i was just sitting in class thinking about that and thought i\'d blog it.  hugs for all.  :)  '),(22830,664485,'05,August,2004','2004-08-05 00:00:00','   ruhing, rushing, rushing!  flying through memphis into amsterdam and touching down in kuala lumpur.  something like 28 hours of travel, anticipation, excitement, insanity, anxiety.   worried about connections and people and stinky fat chatty men next to me.   i guess i\'ll see you in malaysia!  :)  hugs  '),(22831,2061087,'28,February,2004','2004-02-28 00:00:00','   I found  urlLink Holy Fools  to be hard to read at times due to the phrasing and language at times, yet the story kept me engrossed.  Once I was used to the writing style, the details took me in and I could almost imagine myself there with the heroine.  It is set mostly in the early 1600s and is the story of a young \"gypsy\" woman who finds herself pregnant with nowhere else to turn.  Being very resourceful and a thespian by trade, she portrays herself as a young widow and seeks refuge in a convent.  The Abbey that she comes to is comprised of quite irregular nuns.  More of a home for the troubled women than a convent.  Five years later, following the death of the kindly Reverend Mother, she is confronted with her past.  The leader of the troupe that she was in and the father of her daughter has concocted a new scheme to seek revenge on a bishop.  Before long suspicion, betrayal and debauchery are rampant in the convent, fueled by rumors of witchcraft and demonic possessions.  To save the convent and her daughter, she must match wits with her former lover and divert his plans.  '),(22832,2061087,'27,February,2004','2004-02-27 00:00:00','   I have loved all of Deborah Smith\'s books and  urlLink Charming Grace  was no exception.  Grace Vance Bagshaw is a classic character - a 34 year old ex-beauty queen and former talk show host, she now spends her days grieving the death of her husband, Harp Bagshaw.  Grace was the only one who ever truly believed in Harp until he sacrificed his life on the roof of a hospital to a mad bomber bent on blowing up the hospital and every one in it.  Now aging action-movie superstar turned director Stone Senterra invades Dahlonega, Georgia to film the story of Harp Vanceâ€™s heroism â€“ over Grace Bagshaw Vanceâ€™s dead body.  But to get to Stone, Grace must bypass Boone Noleen, Stoneâ€™s sexy Cajun ex-con bodyguard.  And Stone\'s muscle-bound half-sister, Diamond.  Think pro-wrestler turned movie star.  With a little help from her family, Grace is determined to keep Harp\'s private life, just that.  But, her family doesn\'t see it that way and soon the alliances are flying.  Grace and Boone, Boone and Grace\'s G. Helen (as her grandmother insists on being referred to), Harp\'s long-lost sister\'s daughter and Stone\'s son and a host of other characters.  With plenty of \"bless his/her hearts\" thrown in this book is the perfect combination of drama and comedy.  '),(22833,2061087,'26,February,2004','2004-02-26 00:00:00','   urlLink Seduced by Moonlight  is the third book in the Merry Gentry series.  Merry is a mortal Faerie princess who has been hiding in L.A.  Her aunt, the Queen of Air and Darkness, has ordered her to compete in a contest with her cousin, Cel (who is in the torture hall for trying to have her killed) in making a baby.  Whoever produces a child first wins the throne.  The only qualifier is that Merry can only bed her royal guards and whoever gets her pregnant is to be her king.  But, not being able to get pregnant is one of the least of her worries.  There are those who are still trying to assassinate her, and she is also trying to form an alliance with the Goblin King and there are those who do not wish that to work.  I call Laurell Hamilton\'s books supernatural porn.  There is a lot of sex and it is fairly graphic and always out of the ordinary.  Combine that with the intrigue and violence and you have a perfect book.  '),(22834,2061087,'26,February,2004','2004-02-26 00:00:00','   urlLink Maggie Darling  is very funny and satirical in a low-key sort of way.  Maggie is a Martha Stewart-type goddess of hearth and home where everything is perfect except her picture-perfect life suddenly goes awry.  She is suddenly embroiled in a life of ex-husbands, thieving best friends, flings with younger men, and a son who is haning out with a gansta rap crew - among other such unrefined quandries.  It is delightful the tale of how she has it all, loses it all, and learns how to carry on brilliantly in her strange new situations.  '),(22835,2061087,'26,February,2004','2004-02-26 00:00:00','   urlLink Cereal Killer  is the ninth book in the Savannah Reid mystery series.  I love her books because they are all light-hearted and fun, with murder and intrigue thrown in.  Savannah is a Southern transplant to California and there are always cute Southernisms thrown in.  I am looking forward to the next one.  '),(22836,2061087,'26,February,2004','2004-02-26 00:00:00','   Ok, so this blog is mostly for my reference, but hopefully if anyone stumbles across it, it will give them some ideas for good  books to read.  '),(22837,2061087,'22,March,2004','2004-03-22 00:00:00','   Today has been kah-ray-zee!  That\'s what I get for being gone for a week.  So, quick run down of these books I got piled here so I can get them back to the library and pick up my other holds!    urlLink P S, I Love You - Cecelia Ahern  This book is SOOO good!  I\'ve lucked out lately and have been getting some goodies.  This one was both heart-wrenching and funny.  Very moving.  I can\'t believe that the author is the 22-year-old daughter of the Prime Minister of Ireland.  Very mature writing.  I guess that is the mark of a true novelist, when you can\'t define them by age.  Anyway, go get this book and read it!  Now!   urlLink What She Wants - Cathy Kelly   Another excellent read.  And another Irish read.  I really enjoyed this one and in a way it was surprising because of the way the point of view keeps switching from one character to another.  Normally I can\'t stand that because with most authors you aren\'t ever clear for a bit who is doing the story telling unless it is explicitly noted.  I can\'t recall if she noted or not who was doing said chapter or whatnot, but I immediately knew which character it was from the tone.  It\'s a fairly thick book, but moves along quickly.    Ahh.. must take break.  And now that I\'ve had a break I am dead tired. Quick list so I can take them back tomorrow.   urlLink The Goddesses of Kitchen Avenue - Barbara Samuel  Excellent   urlLink Wurst Case Scenario and Frozen Rodeo - Catherine Clark  I was in the Young Adult section looking for some Judy Blume to re-read and found these two and liked the titles so I thought what the heck.  Very cute books that really portray teen angst.  Takes me back a few!   urlLink We\'ll Always Have Parrots - Donna Andrews  I could NOT finish this book.  Maybe if I had nothing at all left to read I would, but otherwise it hasn\'t held my interest and I\'ve been trying for over a week.  It\'s going back unfinished.   urlLink Spitting Feathers - Kelly Harte    Cute enough book.  Don\'t know that it was worth the $11 I paid for it, so either get it at the library or get it second-hand.   urlLink Lucy\'s Launderette - Betsy Burke  Very much worth the $11!  Outstanding book.  It\'s great because I\'ve always wanted to open up something similiar to what they have done, except with books.  I\'ll keep on dreamin\'.   urlLink The Feng Shui Detective - Nury Vittachi  I haven\'t got past page 9 on this one, but I suspect it\'s going back unread.  I\'m not liking it at all.    urlLink Pretending the Bed is a Raft - Nanci Kincaid  So-so.  I sent it back without finishing it.  Most short story books don\'t interest me.  Oh, and I read through  urlLink Trading Spaces: Color!  and  urlLink Martha\'s Vineyard Houses and Gardens .  The Trading Spaces one was ok, was kind of complicated and some of the pictures weren\'t that great.  The Martha\'s Vineyard one was beautiful and very informative.  I\'ve always wanted to go there.  '),(22838,2061087,'20,March,2004','2004-03-20 00:00:00','   I\'ve been out of town for the last week and haven\'t had internet access so I am way behind on my posting!  I\'ve got a stack of books sitting here that I\'ve read, a stack waiting to be read and I just picked up 3 more out of the mailbox on the way in!  So, tomorrow I will get busy and get them put up here.  '),(22839,2061087,'12,March,2004','2004-03-12 00:00:00','   Luanne Rice is another author who I have read every single novel she has written, and I am never disappointed.   urlLink Dance With Me  is no exception.  '),(22840,2061087,'09,March,2004','2004-03-09 00:00:00','   I absolutely LOVE  urlLink Jennifer Crusie .  I\'ve read all of her books.  And I think this one has been my favorite so far.  I did not want this book to end.  I literally want to live the book!  How insane is that?  But seriously this book rocks.   Min Dobbs knows that happily-ever-after is a fairy tale, especially with a man like Cal Morrisey, who asked her to dinner to win a bet. Cal Morrisey knows commitment is impossible, especially with a woman as cranky as Min Dobbs. When they say good-bye at the end of their evening, they cut their losses and agree never to see each other again.  But Fate has other plans, and it\'Â’s not long before Min and Cal are dealing with meddling friends, wedding cake, a jealous ex-boyfriend, Krispy Kremes, a determined psychologist, chaos theory, a frantic bride, Chicken Marsala, a mutant cat, snow globes, two Mothers-from-Hell, great shoes, and more risky propositions than either of them ever dreamed of including the biggest gamble of all--unconditional love.  (synopsis from  urlLink Jennifer Crusie\'s awesome website )  '),(22841,2061087,'09,March,2004','2004-03-09 00:00:00','   So, remember the  urlLink jackpot ?  Well, most of them are going back.  I read  urlLink Lost & Found ,  urlLink Cooking for Harry , and am slogging through  urlLink We\'ll Always Have Parrots .  I started  urlLink Jackie by Josie  and couldn\'t get into it.  Come to find out I have already read  urlLink Lucy Crocker 2.0  and it wasn\'t that great the first time.  The others I just don\'t have interest in now.  I hate it when this happens.  Ah well, c\'est la vie.  Of course, a lot of it probably has to do with the fact that I did get some new books I had on hold in.  And they were SO GOOD that they spoiled me.  I finished both of them in hours.  I tend to gobble the good ones up.  I\'ll post their reviews next.  And hope the library has some goodies in for me soon!  '),(22842,2061087,'09,March,2004','2004-03-09 00:00:00','   In  urlLink Cooking for Harry , fabulous cook Harry learns that he is going to have to lose the weight he has put on over the years of learning to be the gourmand he is.  His wife, Francie, convinces him to let her take over the cooking and for him to join a weight-loss club.  Soon, Harry is on the road to a slimmer self but between taking more initiative at work and spending more time working out and with his weight loss group, not only is he no longer in the kitchen, he is never home!  Francie is confused by the dynamics of this new relationship, and is worried about the attention her newly svelte husband is getting.  Were they ever truly in love or did they just become comfortable in their routine?  Is the Atkins diet going to ruin their relationship?  This novel brings up some real-life issues in a heartwarming way.  '),(22843,2061087,'06,March,2004','2004-03-06 00:00:00','   urlLink Lost & Found  was well worth the wait.  I really enjoyed it.  But then I love all of the  urlLink Red Dress Ink  books.  In this novel by Jane Sigaloff, Sam Washington is a control freak lawyer.  On a business trip to New York, her precious diary somehow gets left behind.  What follows is the story of the angst Sam goes through until it is returned, the angst caused by who returns it and the chaos that ensues and how she learns to let go of some of that control.  '),(22844,2061087,'06,March,2004','2004-03-06 00:00:00','   Several weeks ago, I read both Savannah Blues and Little Bitty Lies by  urlLink Mary Kay Andrews .  Both books were excellent.  I recommended them to my Mom and she loved them as well.  We were both commenting on the  urlLink Beyond the Grave Chicken Salad  that features in the plot of Little Bitty Lies.  It sounded really good.  So, she was here last week and we decided to whip up a batch.  Oh my goodness, it is soooooo good.  I couldn\'t stop eating it.  It was bar none the best chicken salad I\'ve ever had.  And I\'ve had a lot!  So, first chance you get go get both of the books and the make you up some of that chicken salad!  (Tip:  We only used 4 pounds of chicken and I thought it could use more dressing.  So, next time I will either use 3 pounds of chicken or if I am making it for a crowd make up some more dressing.  We also left out the nuts and stuff, and divided the salad and only put water chestnuts in half since there is no way I am puttin those nasty things in my mouth!)  '),(22845,2061087,'06,March,2004','2004-03-06 00:00:00','   That\'s what I hit yesterday evening when I went to get some new books to read.  Some days I go in and am lucky to walk out with one or two books.  Some days it\'s 10 or 12.  And when it\'s one or two it really sucks because I read fast, like a book a day (sometimes two or three if they are short and I am bored).  I am really thankful for the new online access to the card catalog that our library has because it enables me to seek out books before I even step foot in the library.  That and the hold function allows me to get on the list for new books and not have to wait eons for them.  Yesterday I was able to go to the library all by me onesies since Mom was here to watch the kiddos so I got to do it the old-fashioned way.  I do love wandering the rows.  So, yay me, I have books to read for a while.  I hope they are all good.  I hate when I get books and end up tossing it after a few pages or the first chapter.  At any rate, here is the list:   urlLink Lost and Found - Jane Sigaloff   (I have been waiting months for this one, I had to get it through the interlibrary loan system and it went through 15 before it found it.  They usually come from Dallas since it is the closest \"big\" library.  This one is from Phoenix.  Maybe I should move there.  Great climate and great books!)   urlLink We\'ll Always Have Parrots - Donna Andrews   (This is a cute series.)   urlLink The Good Wife Strikes Back - Elizabeth Buchan   (From the author of Revenge of the Middle-Aged Woman, which I loved.)    urlLink Jackie by Josie  and  urlLink Lucy Crocker 2.0 - Caroline Preston  (These just sounded good, don\'t know anything about them.)   urlLink Lilies That Fester and Reap a Wicked Harvest - Janis Harrison   (I saw Reap a Wicked Harvest in the new book section and it caught my eye as I LOVE mysteries and gardening and unfortunately the library only had one other of the series.  I prefer to read the entire series in order.)   urlLink Cooking for Harry: a low-carbohydrate novel - Kay-Marie James   (I saw this one online somewhere and thought it sounded funny.  Everything is low-carb now, even our books!)   urlLink And Now You Can Go - Vendela Vida   (another one on the new book shelves that sounded good)   urlLink Lunch at the Piccadilly - Clyde Edgerton   (Know nothing, the jacket caught my eye)   urlLink A Love Noire - Erica Simone Turnipseed   (Another new one that I had seen something about online and thought I would check out.  Don\'t know anything yet except the pages I scanned sounded good.)  '),(22846,2061087,'05,March,2004','2004-03-05 00:00:00','   I had stopped reading the  urlLink Death on Demand Mysteries  because after the first few they got so ungodly boring I couldn\'t finish them.  But, the title of this  urlLink 15th installment  and the plot caught my eye (and I didn\'t have anything else to read today) so I thought I\'d give it a whirl.  And I finished it and I actually liked it.  I won\'t go into details but it was a good book.  '),(22847,2061087,'03,March,2004','2004-03-03 00:00:00','   urlLink High Country  is the 12th Anna Pigeon novel.  Anna is pulled from her current station in Mississippi to go undercover in Yosemite National Park\'s historic Ahwahnee Hotel.  Four young seasonal park workers have turned up missing and extensive searching has returned nothing and Anna is brought in undercover to see what she can learn from her fellow employees.  She finds plenty to be suspicious about from her twenty-something roommates and the irritable head chef in the restaurant where Anna is waitressing in the place of one of the missing girls.  An overheard conversation leads her on an expedition alone into the high country where she finds the source of the mystery and almost loses her life to some city-slicker goons.  But, her will to live and to obtain justice lead her back to the main part of the park where she finds more clues and more trouble and her survival skills are pushed to the limits.  Once again, Nevada Barr has delivered a suspenseful, thought-provoking novel set among beautiful scenery and corrupted people.  '),(22848,2061087,'02,March,2004','2004-03-02 00:00:00','   urlLink Don\'t Tempt Me  is the final book in the Pearl Island Trilogy, and I am kind of sad to see it come to an end.  The characters are all distinct and the setting and the history are great.  Adrian St. Claire is the sexy chef at the Pearl Island B&B, which he and his sisters run.  He is the descendent of the woman whose husband owned the house and her and her lover\'s ghost can\'t go on until the treasure that was lost in the cove is recovered.  But, to find the treasure they need help.  Jackie Taylor is the descendent of the lover.  She owns her own \"pirate\" ship and is more interested in running cruises along the Texas coast than chasing lost booty.  That and she would rather her past not be dredged up. Finally, Jackie is enticed into helping and soon not only do they have something other than treasure hunts and casual flirtation in mind, but they learn the most important lesson of all.  '),(22849,2061087,'02,March,2004','2004-03-02 00:00:00','   urlLink The Time Traveler\'s Wife  was an incredibly thick book, yet the story was so compelling that I read it in one day.  It moved so smoothly, and the concept and characters were so interesting that I could not put it down.  Henry is a Chrono-Displaced Person (CDP), who travels through time involuntarily, usually when very stressed.  Clare first meets him when she is six and he is 36.  He returns regularly throughout her childhood from different times in their shared future. At last, when Clare is twenty and Henry twenty-eight, they meet in his present, and the relationship begins in earnest.   His constant disappearing and reappearing is hard on both of them, yet they are destined to be together and they persevere.  '),(22850,2061087,'27,April,2004','2004-04-27 00:00:00','   So... I had this nice long informative post all written up the other night when all of a sudden the power went out and I lost it.  Don\'t you love it when that happens?  I shall now attempt to recreate it in short form.    urlLink The Saints and Sinners of Okay County - Dayna Dunbar  This is a great retro novel (set in 1976 or so) in small town Oklahoma.  Aletta Honor is pregnant with her fourth child and has no idea where her husband is at the moment.  With a stack of bills to pay, she is getting desperate.  She attempts to set up a kolache stand during the annual Czech Festival but she burns them all and has to settle for peddling lemonade.  After she uses her long hidden psychic powers to save the life of a customer, she decides to put a sign up in the front yard and offer psychic readings from her home.  Town and husbandly drama ensues, but Aletta knows she is doing what she should be doing.  This was a great novel.    Nostalgia rules.  I\'ve been on a teen novel kick lately.  These are all good ones.  urlLink Gingerbread - Rachel Cohn   urlLink All American Girl - Meg Cabot   urlLink The Earth, My Butt, and Other Big Round Things    urlLink Filthy Rich - Dorothy Samuels  Not a teen novel although I did find it in the Young Adult section.  If 34 is now considered young adult... yay!  This was a cute read.   urlLink Meely LaBauve - Ken Wells   I found this book in the Young Adult section at the  urlLink library  but I wouldn\'t really consider it that genre.  The main character is a 15 year old boy but the subject matter is a little mature.  Of course, it isn\'t anything I wasn\'t reading (or doing for that matter) at that age.  It is touted as the Cajun Huck Finn, but I have to disagree there as well.  It is much better!  This is a must read or any age, male or female.     urlLink Can You Keep A Secret - Sophie Kinsella  Yet another gem.  I\'ve been anxiously awaiting her next novel, and this one was well worth the wait.  Funny, charming, witty.  Everything it should be.  I was hooked.  And I have read where the movie rights have already been scooped up.  Although Kate Hudson is set to play the main character and I don\'t really see her as fitting the part, not that I don\'t like her but it just isn\'t her.    urlLink The Breathtaker - Alice Blanchard  This was one wild ride of a book.  Set in Tornado Alley (which is where I live) this murder mystery is an experience.  I love it when a book is in familiar settings, you can learn something from it, and it keeps you hooked.  It took me much longer than usual to figure out whodunit in this one.  I definitely recommend this one.  '),(22851,2061087,'20,April,2004','2004-04-20 00:00:00','   Finally finished  urlLink The Solace of Leaving Early .  It started off incredibly slow and I really didn\'t think I would get past it.  That and I was not liking the main character at all.  About halfway through it started getting better, and by the end of the book I was loving it.  I\'m glad it didn\'t turn out to be a waste of time and effort to read.  It\'s funny how some authors get better with time and some get worse.   I re-read   urlLink Neil Gaiman\'s   urlLink The Wolves in the Walls  and  urlLink Coraline .  He is amazing.  A friend has loaned me  urlLink Stardust  to re-read.  I need to scour the used book stores as his are books I need to keep.  I don\'t often re-read books, so I like to have the ones I love around when the urge strikes.  '),(22852,2061087,'17,April,2004','2004-04-17 00:00:00','   I haven\'t been doing a whole lot of reading lately due to pursuing other pleasurable pastimes.  Maybe I should just start listing the books with a rating system as lazy as I am!  ha!     urlLink So  5 Minutes Ago - Hilary DeVries    This was not as good as it was hyped to be, or as good as I was anticipating it to be.  It wasn\'t bad by any means, it was just a little too fake at times for me.  But then, we are talking Hollywood here....   urlLink Scottish Girls About Town - 16 Scottish Women  Scottish Girls wasn\'t as bad as  urlLink Irish Girls About Town , but neither was I very impressed with it.   urlLink Lucky Stars - Jane Heller  Overall, a cute, entertaining read.    And now.. the good stuff!  Ta-da!   urlLink Blessed  Are The Cheesemakers - Sarah-Kate Lynch  Do yourself a favor and RUN - don\'t walk- out and get this book!  It is one of the best reads I\'ve had in a while, yet not weighty or depressing.  I wanted to hang out at Coolarney House and soak up the magic.  Lynch\'s writing made me feel as if I were.   urlLink Something Rising (Light and Swift) - Haven Kimmel  This IS the best read I\'ve had in a long while.  It was more \"wordy\" than I normally like, but in Haven Kimmel\'s style it made the story.  Pick this one up when you are out getting The Cheesemakers.  You won\'t be disappointed.   I  picked up  urlLink The Solace of Leaving Early , Kimmel\'s second novel, yet couldn\'t really get into it.  I am going to give it another try now.  I\'ll also have to look for her first one,   urlLink A Girl Named Zippy .  I had a whole pile of other books that were such crap that I won\'t even go into them.  '),(22853,2061087,'07,April,2004','2004-04-07 00:00:00','   It seems to be the trend lately to read a while and then catch up on updating.  Better late than never, eh?  I got a huge stack of books at   urlLink the library  last week, but only a few of them have been able to hold my interest.  The others I have been vacillating back and forth on, a chapter here... a chapter there.  Hopefully the next batch will hold more goodies.   urlLink Emma\'s Secret - Barbara Taylor Bradford  This was another one that went back unread.  Barbara\'s books are always very well written, but they are usually too... staid... I guess is the word.  I may check it out again when I am in a different frame of mind.   urlLink Straight Talking - Jane Green  Okay, this book annoyed the hell out of me for the most part.  It was a good book, but it was just too rambling, wishy-washy, whiny and I could not sympathize with the main character at times.  I could in theory on the subject, but I suppose it was just that the other stuff got in the way.  Maybe it was my mood, I don\'t know, because here lately all of her books have annoyed me on some level.   urlLink Bookends  was the first novel of hers that I read and it was amazing, and I\'ve been somewhat disappointed by all of the others.   urlLink Luck of the Draw - Carolina Garcia-Aguilera  This book was indeed the luck of the draw.  It was outstanding!  Very well written, very engaging.  I shall be checking out all of her other novels ASAP.   urlLink The Cure for Death by Lightning - Gail Anderson-Dargatz  I\'ve seen this book many times at the library but had never checked it out.  I finally picked it up and am glad that I did.  It has some subject matter in it that I am not too fond of reading about (but then who wants to read about horrible reality?) but it was such a great read, and it left the indelible reminder that life happens, it isn\'t always pretty, but it does go on and you can make it the best it can be.  I loved the realism of the book even if it was harsh.  It was very evocative of the time yet not stuff that is written about a lot.  '),(22854,2061087,'01,April,2004','2004-04-01 00:00:00','   I have so many stacks of books sitting around that I sometimes lose track of what is where.  I found two more that I have read in the last couple of weeks that got buried.  Both of them are from  urlLink Red Dress Ink .  The first one is  urlLink Speechless by Yvonne Collins and Sandy Rideout .  There were parts of this book where I didn\'t really like it but overall it was a good read.  Second up is  urlLink Engaging Men by Lynda Curnyn .  Again, it was just an average book.  Both weren\'t horrible, but they weren\'t the best I\'ve ever read.  Speaking of  urlLink Red Dress Ink , I can\'t wait to order  urlLink Fat Chance by Deborah Blumenthal  and  urlLink Inappropriate Men by Stacey Ballis !  '),(22855,2061087,'01,April,2004','2004-04-01 00:00:00','   I have still been trying to keep up with life here.  I\'m reading enough, just haven\'t had time to be online and post what I\'m reading.  So, gonna play catch up and keep it short and sweet.     urlLink Die In Plain Sight - Elizabeth Lowell   This is another of the Rarities Unlimited series and as usual is a great read.     urlLink Coastliners - Joanne Harris   Her books are great.  This one was not as good as  urlLink Holy Fools , but it was an enjoyable read.   urlLink Ain\'t She Sweet - Susan Elizabeth Phillips   I LOVED this book.  Sugar Beth and all of the other characters draw you in to their world and makes you laugh and want to cry along with them.   urlLink Something Real - J.J. Murray   This book went back unread.  I don\'t know if it was just the mood I was in or I didn\'t like the writing style, or the character or what.  '),(22856,2061087,'25,May,2004','2004-05-25 00:00:00','   What with the broken car, strep throat, pink eye and numerous other lovely things going on around here, I haven\'t had time or the inclination for doing much reading.  It\'s times like this that unless the book is superb, it won\'t hold my attention and give me the satisfaction I want.  After all, isn\'t that what reading is all about?   My reading list here is going a few weeks back and is very scant.     urlLink Scarecrow - Robin Hathaway  Despite the wholly unoriginal title, this book is great.   urlLink Full Blast - Janet Evanovich and Charlotte Hughes  This one went back to the library unfinished (among others I am just tired of listing the failures but this one was too bad to pass up).  I love Janet Evanovich\'s Stephanie Plum series, but I haven\'t liked this collaboration much from the start.  This one was just too cheesy and out there for me to go along with.  I hope her Plum series doesn\'t start going downhill.   urlLink Sunset Over Chocolate Mountains - Susan Elderkin  Very interesting and different novel.   urlLink Sunshine and Shadow; Broken Dishes - Earlene Fowler  These two made it all worthwhile.  Earlene Fowler and the Benni Harper series are my absolute favorites.  If you are a lover of mystery, quilts, history and art you should definitely check this series out.  I won\'t go into details on these today as I am a bit feverish and might not be coherent.  '),(22857,2061087,'11,May,2004','2004-05-11 00:00:00','   I ordered  urlLink The Dim Sum of All Things by Kim Wong Keltner  through the  urlLink library\'s  ILL service.  It was a really good chick lit read.  Lindsey is an ABC (American Born Chinese) living in San Francisco and struggling to find herself (basic premise of most of these novels) while working as a receptionist at a vegan mag and wondering what to do with her degree.  Her take on the single life was dead-on regardless of background and I found her take on \"Hoarders of All Things Asian\" hysterical.  Greige is my new word. Living with her grandmother presents some dating challenges, not to mention sensory ones what with all the Tiger Balm and food aromas mingling.  After a trip to China with her grandmother, things start to fall together for Lindsey and there is the usual happy ending.  '),(22858,2061087,'11,May,2004','2004-05-11 00:00:00','   How sweet is blogger now?  We have a bunch of new templates to chose from and comments!  Yay!  I\'m grooving on the pink right now and hopefully I won\'t get tired of it soon since you still lose all of your mods when you change templates.  I\'ve added a blogroll over to the side with some literary linkies.  I hope everyone enjoys the new look and I\'m going to try to keep my posts less sporadic.  '),(22859,2061087,'26,June,2004','2004-06-26 00:00:00','   urlLink The Other Side of the Story - Marian Keyes  I just love Marian Keyes.  As with all authors I like (and there are plenty), I\'ve read all of her books.  This one isn\'t my new favorite, but it\'s a great book.  I highly recommend this one.  '),(22860,2061087,'24,June,2004','2004-06-24 00:00:00','   urlLink Best Enemies - Jane Heller  With friends like this, who needs enemies?  But all\'s well that ends well.  And some more trite sayings.  Cute book, but a little cliche in parts.  '),(22861,2061087,'20,June,2004','2004-06-20 00:00:00','   So, since I\'ve been back I\'ve been doing more reading than posting.  I have a fairly long list of books I\'ve finished.  I guess I better get them on here, huh?   urlLink Under the Boardwalk - Carly Phillips  Cute as always, good summer read.   urlLink The Shop on Blossom Street - Debbie Macomber  The writing in this books seemed kind of, I don\'t know, stilted?  It was a great story, great characters; but, the telling of it was just off.   urlLink Olivia Joules and the Overactive Imagination - Helen Fielding  Helen has turned out another one that is destined to make a great film.  This book was hilarious.  I can\'t wait to see if they make it into a movie.   urlLink To Have and To Hold - Jane Green  I haven\'t been too impressed with Jane\'s books lately; but, she redeemed herself with this one.  Very well written, funny, poignant.     urlLink My Sister\'s Keeper - Jodi Picoult  Once again, Jodi has taken a tough subject and make it real for us readers.  This is a heartrending book.  As usual, she put one of those \"Oh no she DI\'INT\" moments in it, that I could have done without, but it\'s a fantastic novel.   urlLink Queer Eye for the Straight Guy: The Fab 5\'s Guide to Looking Better, Cooking Better, Dressing Better, Behaving Better and Living Better   These guys are so funny, and they really know there stuff.  If you have a straight guy in your life, go buy him this book.     urlLink Sweet & Crazy - Patty Dann  Exactly what this book is.   urlLink Fat Chance - Deborah Blumenthal  I\'ve been trying to get my hands on this book for ages.  It was well worth the wait.   urlLink The Ocean Between Us - Susan Wiggs  A must-read.  Susan Wiggs is always a sure thing.   urlLink Dumping Billy - Olivia Goldsmith  Hilarious, as usual.  I cannot believe Olivia is gone and we will only have one more of her books.  It\'s tragic.   urlLink Monday Mourning - Kathy Reichs  Kathy Reichs is another sure thing.  I\'ve read all of her novels and they are the ones that you can\'t put down until they are through.  '),(22862,2061087,'10,June,2004','2004-06-10 00:00:00','   I went back home to Graham for about a week, so I didn\'t have access to the internet.  I only took a couple of library books with me as I didn\'t want to have many to keep up with and/or lose.  I read a bushel of Harlequin and Silhouette romance novels.  They aren\'t bad for occupying the mind for an hour, that\'s about how long it takes me to read one.  I won\'t even attempt to list any of them.  So, here are the \"real\" books I read this last week.   urlLink Men in Kilts - Katie MacAlister  Cute enough, I didn\'t hurl it across the room in disgust at least.   urlLink The Song Reader - Lisa Tucker  Not at ALL what I expected it to be, but VERY good reading.   urlLink Good Grief - Lolly Winston  I wasn\'t thinking I would like this book when I first started it, but after I got a chapter or so in and I was snickering aloud, I was hooked until the finish.  Superb read.  '),(22863,2061087,'01,June,2004','2004-06-01 00:00:00','   urlLink Goodbye Earl  is the latest installment in the Bad Girl Creek novels.  This is another series that if you haven\'t read it, do so.  They are fantastic.  All of her books are.  Hank & Chloe was the first one I read, and I\'ve been hooked since.  I think she just gets better with time.  '),(22864,2061087,'30,July,2004','2004-07-30 00:00:00','   urlLink R is for Ricochet - Sue Grafton  I think this is my favorite Kinsey Millhone novel so far.  There were a few there that I wasn\'t too fond of the plot or the writing.  They had became kind of lackluster.  But this one was much improved.  '),(22865,2061087,'26,July,2004','2004-07-26 00:00:00','   I\'m just laaazzyy.      urlLink Judith Van Gieson - Land of Burning Heat   and  urlLink The Shadow of Venus    The more I read of Judith Van Gieson, the more I like her. I love books where you can be entertained and learn things at the same time. And this series of full of history, art and the character of humans.      urlLink The Second Assistant: A tale from the bottom of the Hollywood ladder. - Clare Naylor & Mimi Hare    This was another one of those lucky finds.  I absolutely loved this book.      urlLink Flashpoint - Suzanne Brockmann     This was good reading.   The situation was a little unbelievable, but was still a good story.     urlLink Bergdorf Blondes - Plum Sykes     I had read some bad press about this book so I was very interested in reading it. Usually the worse things I hear about something, the more I like it. At first, I thought that they were all right and it was not going to be that great, but after a few chapters it really got going and I loved it.       urlLink Pad Parties: The Guide to Ultra-Entertaining - Matt Maranian    I was looking for his first book,  urlLink Pad: The Guide to Ultra-Living  but this was all our library had.  Fabu!        urlLink Pilates for Wimps - Jennifer DeLuca    My new favorite book. I\'ve already check this out for the library like three times, and I\'ve considered telling them I\'ve lost it and just paying for it so I can keep it. I love Pilates and this is the best how-to book I\'ve found so far.     I\'ve been on an interior decorating book kick really bad lately. Here are a few of them I\'ve read. All good, but the style is only to some people\'s liking. I like a mix of vintage, country and modern.    urlLink Country Chic - Country Living     urlLink Southern Accents on Color     urlLink Flea Market Finds: Before &amp; After Home Decorating with Makeover Miracles  (another one I want to \"lose\")  '),(22866,2061087,'14,July,2004','2004-07-14 00:00:00','   As I was getting ready to post this I realized that most of the books I have read lately seem to have a common theme.  Death. It certainly wasn\'t planned that way.  And with the recent death of my grandmother and a few other deaths in the family, I think I\'ve had enough of this theme.  Good books, but not something I want to be reading about constantly.   urlLink The Color of Death - Elizabeth Lowell  Lowell delivers another gem of a book.  (ha, I\'m so punny)   urlLink On A Night Like This - Ellen Sussman  This was a great book, if a little farfetched for my tastes.  '),(22867,2061087,'09,July,2004','2004-07-09 00:00:00','   This is turning into like the monthly update blog.  Ugh.  I had to go back home unexpectedly (well, not so much unexpectedly as expected just not knowing when) when my Mom called and said my Mamaw was dying.  She\'s been in poor health for quite some time, and it rapidly declined over about a week and a half.  So, when she called, I threw everything in the car and hauled it up there.  So, what being with her until she passed on, the funeral, and all the stuff after the funeral I was there over a week.  And needless to say, I didn\'t get much reading done.  I\'ve been too tired to do that much more reading since I\'ve been home.  But, a bunch of the books I\'ve really been looking forward to came in while I was gone, so I\'ve been doing some.  So, enough with the story, on with the books.   urlLink Vanishing Point  and  urlLink Confidence Woman - Judith van Gieson  These are the second and third novels in the Claire Reynier mystery series.   urlLink The Stolen Blue  was the first, which I read a few years ago and then forgot about.  I\'m glad I remembered it and went searching for any following novels.  Van Gieson has a great style.  I\'m on the hold list for the other two.  Will post when I read them.   urlLink How to Meet Cute Boys - Deanna Kizis   I just happened to spot this on the new books shelf while in the checkout at the library and added it to the pile.  I\'m glad I did.  Very entertaining \"chick lit\".     urlLink Tallulahland - Lynn Messina  It was really hard to read this book and not think \"Lady, just STOP!\".  I was kind of tired of her attitude towards her father by the end of the book, but I finally forgave her.  All in all, it was a fair read.   urlLink Ten Big Ones - Janet Evanovich  #10 (duh) in the Stephanie Plum series, and fabulous as always.  It\'s hard to believe Steph\'s been at it for 10 books now.  And, of course!, the ending leaves you waiting for the next installment.  Hurry up, already!   urlLink The Goodbye Summer - Patricia Gaffney  I just finished this one, and it hit pretty close to home in many ways.  There were also a few of those \"Lady, come on!\" moments in this one, but this is a great novel.  '),(22868,2061087,'29,July,2004','2004-07-29 00:00:00','    you have chicken pot pie for breakfast.  So, yeah, not the greatest day.  I am a comfort eater.  If I am stressed out, depressed, or even happy I eat.  Why can\'t I be one of those people whose appetite is killed by that kind of stuff???  The rest of the day went as follows:  potato wedges, macaroni & cheese, and a new m-azing candy bar.  And a C2.  Which I love.  I normally can\'t drink diet sodas.  And while it\'s true that this one isn\'t technically diet, it\'s better than the real thing.  All in all, I suppose what I had today wasn\'t bad.  It isn\'t generally WHAT I eat, as HOW MUCH OF IT I eat.  Most times, all that I ate today would have been in ONE sitting.  Add a few more meals like that a day and that\'s where the weight came back on.  I refuse to get stressed out over it in a big way, because I know that will just set me up for more failure.  I\'ll be glad this weekend when all of this stuff is out of the house.  When I do my big grocery shopping, I\'m not even going to buy junk.  If I don\'t bring it in, I won\'t eat it.  Simple as that.  I\'ve just been being incredibly lazy this summer.  It\'s a good thing I never did get into that swimsuit...  That\'s okay.  Next summer I will.   '),(22869,2061087,'27,July,2004','2004-07-27 00:00:00','    Bowl of Honey Bunches of Oats Cereal  PB&amp;J Sandwich  Campbell\'s Blended Vegetable Medley Soup  Grilled Cheese Sandwich  Brisket and Garlic toast  Popsicle  Water, water and more water   '),(22870,2061087,'26,July,2004','2004-07-26 00:00:00','    I walk when I can and I do Pilates daily when I am on the wagon.   urlLink Pilates for Wimps by Jennifer DeLuca   is the best book I have found so far.  I love it so much I want to steal it from the library.  I am going to have to find my own copy of it.  There are many, many other Pilates books and videos out there, but so far this is my fave.  I highly recommend it, even for those of you who have tried Pilates before and got your butts kicked. ;)   '),(22871,2061087,'26,July,2004','2004-07-26 00:00:00','    1 cup plain basmati rice  1 blueberry muffin  1 grilled cheese sandwich (2 slice bread, 1 slice medium cheddar, olive oil spray)  3 Ball Park franks, 2 slices bread  wee tiny box of organic raisins  Popcorn  Water, water, water    Now...I know, I know. NOT the healthiest menu. But it\'s better that what I have been cramming in my gullet. It\'s the end of the month and I will be able to go grocery shopping on the first and plan on going back to the healthy buying style. I will get these 30 pounds off. It just torques me (at myself) that I regained over 10 that I had lost and am now almost having to start over.   '),(22872,2061087,'26,July,2004','2004-07-26 00:00:00','    Why is it that when you find  urlLink something  that you really, really like and that you can eat as often as you want relatively guilt-free ALL of the damn stores in your town don\'t have them in stock???  I\'m complaining to the management.   '),(22873,2061087,'26,July,2004','2004-07-26 00:00:00','    173   '),(22874,2061087,'26,July,2004','2004-07-26 00:00:00','    I\'ve struggled with health and weight loss issues all of my life.  This last year, I was finally back on my way to being healthy and I find that I\'ve backslid yet again.    It\'s easier for me to lose weight if I have to hold myself accountable in small ways.  Weigh-ins, measurements, food diary, etc.  I can\'t afford to pay to be in a group, so I have to find ways to be my own support group.    That is to be the purpose of this blog.  My own record of progress.  I doubt very seriously that any of you who stumble across this little journal will have any interest in it, but I can always use some moral support and encouraging words.   '),(22875,2061087,'02,August,2004','2004-08-02 00:00:00','    I haven\'t updated in a few days due to being incredibly busy. The oldest two kids came home from their Dad\'s for a couple of days and then left Sunday for their Mem\'s house. Nothing special going on with the diet thing yet. I have lost a couple of pounds just from cutting back on the sheer volume of food. I haven\'t done enough true exercise, but that will come. It is much easier for me to get my exercise when the kids are in school. I am forced (yes, forced. I am NOT a morning person) to get up early, and after dropping them off at school I can go somewhere and go for a walk or come back home and do my Pilates uninterrupted. Or both! It\'s also easier to eat sensibly as I am much more busy. I went to the grocery store yesterday, and as planned stocked up on healthy stuff. I didn\'t get as much fruit as I would have liked as most of it was overripe. I\'ll have to get up early and go to the Farmer\'s Market and get more. The peaches from Charlie are in their prime this time of the year. MMMMMMMM soo good. I did buy some cookies, but limited them to the very expensive kind. When I buy those, I tend to limit myself to one to prolong the enjoyment. A $5 box of cookies lasts so much longer than a $2 one!   '),(22876,2061087,'28,October,2003','2003-10-28 00:00:00','   Venez loin avec moi la nuit  Venez loin avec moi  Et je vous Ã©crirai une chanson   Venez loin avec moi sur un autobus Venez loin oÃ¹ ils ne peuvent pas nous tenter Avec leurs mensonges   Je veux marcher avec vous  Un jour nuageux  Dans les domaines oÃ¹ l\'herbe jaune se dÃ©veloppe  Genou-haute Ainsi pas vous essai viendrai   Venez loin avec moi et nous embrassera Sur un sommet de montagne Viennent loin avec moi Et je ne cesserai jamais de vous aimer   Et je veux me rÃ©veiller avec la pluie  Tombant sur un toit de bidon tandis  Que je suis sÃ»r lÃ  dans des vos bras ainsi  Tout que je demande est pour que vous  Veniez loin avec moi la nuit Vient loin avec moi   '),(22877,2061087,'27,October,2003','2003-10-27 00:00:00','   I have long been fascinated with the idea of the English and their tea.  It seems the whole day revolves around it.  When a pal drops by, when the day is starting, when things go awry as they are prone to do, the instant response of the British is to put on the kettle.  Coming from Texas where iced tea is de rigeur, I was horrified at the thought of drinking my tea hot.  With milk no less.  Lemon, yeah.  But milk?  No way.  I come from a long line of coffee drinkers.  As in pots and pots per day.  I\'m not that big on coffee but I can drink gallons of iced tea.  If I do drink coffee it is thoroughly diluted with milk and sugar and flavoring.  Cafe au lait is in my opinion the best way to drink coffee. I just couldn\'t wrap my mind around the whole hot tea and milk thing.       Several years ago I was in a bookstore/coffeeshop and it was cold out and I was hankering for something hot to sip on.  While perusing the coffee offerings I noticed something called Chai tea.  Hmmmm.. what might that be?  They described it and I thought, as usual...  eewww.  But, for whatever perverse reason, I tried it.  And was instantly hooked.  I finally knew what it was all about.  Coming from a very small town where such exotic things as Chai tea are unheard of I had to lay in stock when in the \"big city\".  From there I ventured out to try other teas, yes with milk.  And now I know why, when the mind is troubled (or even when it\'s not)  the benefits of preparing a steaming hot cup of tea and settling down to enjoy it.  Those Brits have been doing it for hundreds of years, seems they were on to something after all.  If you haven\'t tried it, come on over... the kettle\'s on.   '),(22878,2061087,'27,October,2003','2003-10-27 00:00:00','   Please help me \'cause I\'m breaking down This picture\'s frozen and I can\'t get out Please help me \'cause I\'m breaking down This picture\'s frozen and I can\'t get out (of here)  Believe me, I\'m just as lost as you Believe me, I\'m just as lost as you  And every time I think I\'ve finally made it I learn I\'m farther away than I have ever been before I see the clock and it\'s ticking away, and the hourglass empty What the fuck do I have to say  Please help me \'cause I\'m breaking down This picture\'s frozen and I can\'t get out Please help me \'cause I\'m breaking down This picture\'s frozen and I can\'t get out (of here)  Believe me, I\'m just as lost as you Believe me, I\'m just as lost as you  Keep it inside the image portrayed As if I couldn\'t stand losing as if I couldn\'t be saved, no way A small confession I think I\'m starting to lose it I think I\'m drifting away from the people I really need A small reflection on when we were younger We had it all figured \'cause we had everything covered Now we\'re older it\'s getting harder to see What this future will hold for us, what the fuck are we going to be  Please help me \'cause I\'m breaking down This picture\'s frozen and I can\'t get out Please help me \'cause I\'m breaking down This picture\'s frozen and I can\'t get out (of here)  Believe me, I\'m just as lost as you Believe me, I\'m just as lost as you  Lost...lost...lost...lost...lost...lost...lost...lost  I\'m just as lost as you, oh well, what am I going to do  I\'m afraid I\'m falling farther away I\'m falling farther away I\'m falling farther away from where I want to be I\'m afraid I\'m falling farther away I\'m falling farther away I\'m falling farther away  Please help me \'cause I\'m breaking down This picture\'s frozen and I can\'t get out Please help me \'cause I\'m breaking down This picture\'s frozen and I can\'t get out (of here)  Believe me, I\'m just as lost as you Believe me, I\'m just as lost as you  I\'m breaking down I\'m just as lost as you I\'m breaking down   '),(22879,2061087,'27,October,2003','2003-10-27 00:00:00','   I\'m tired of having to be strong.   I\'m tired of not having a shoulder to lean on. I\'m tired of not having anyone to share the load. I\'m tired of taking care of everyone else.  I want to be taken care of for a change. I\'m tired of not having anyone to say \"It\'s okay, honey.  I\'m here to help you.  Let me be strong for a while.  You rest.  I\'m here for you when you can\'t do it and you\'re there for me when I can\'t but if we do it together we can handle anything.\"   I\'m tired of carrying the load by myself and there isn\'t anyone who supports me and my decisions and I don\'t want to have to fight for everything or go along with their poor decisions.  I want someone who makes good ones.  I want to be able to make a decision and be able to follow through on it without every damn thing going wrong. I\'m tired of every single thing having to be such an issue and a struggle.  Just once I want something to come easy and go right.  I want to say \"this is what I want\" or \"this is what I need\" and for the answer to be \"okay, done\".  I don\'t want to have to fight tooth and nail for simple minor things to be done right.  I realize life isn\'t perfect and that some things will go wrong but I\'m tired of it being everything. I\'m tired of being so caught up in the struggle that the love and affection falls to the wayside.  I want to be able to savor the little things and not be overwhelmed by too many big ones.  I want everyone to be happy and for the most part carefree.  The happiness and joy should be the majority, not the minority, buried under all of the strife.   I\'m tired of thinking badly of myself when I can\'t be strong anymore. I\'m tired of having to think badly of others when they don\'t step up and do the right thing. I\'m tired of being taunted by what could be, not what is. I\'m tired of being tired.   '),(22880,2061087,'24,October,2003','2003-10-24 00:00:00','   When the heart just won\'t cooperate With what the mind says is right. When you\'re torn apart by the internal fight and you have nowhere left to turn and you don\'t know where to go... you just have to realize that you\'re the only one to know. No one else can give you the answers that you are trying to seek and no one can question the answers that you finally meet.  -----------------------------------------------------------------------------------  Another time, another place, Another me finding my space. In this life there are many roads The problem is where do they go? You choose a direction, but in the end You find that you just really can\'t blend.  -----------------------------------------------------------------------------------  You think you finally have your way straight When the road you\'re traveling on finds another break. It\'s a never-ending journey with many stops along the way. You think you\'re going in the right direction, When all of a sudden there\'s a dead end, detour or a completely different way to go, and you realize you\'re just not certain you\'re reading the map right.  -----------------------------------------------------------------------------------  I alone have the power  to change the course of my life.  I alone  have the desire to make my life a different one.  I alone am scared to death of what my life has become.  I alone know the dark feelings that course through the jumble of my mind.  I don\'t want to be alone.   '),(22881,2061087,'24,October,2003','2003-10-24 00:00:00','   1. Is your hair naturally curly, wavy, or straight? Long or short?  My hair is straight as a board and sort of medium thick and very fine and silky, not coarse.  Right now it is short, short due to a hair disaster but hey.. it\'s grown out!    2. How has your hair changed over your lifetime?  The actual hair hasn\'t changed much over time other than to go from blonde when I was a child, then brown as I got older and now getting a lot of silver.  The texture has always been straight and fine.  As far as styles go?  Gah... it\'s been everything and every color under the sun.  Long, short, medium, straight, curly, punk, 80s big hair (in Texas no less where bigger is always better and the BIGGER the better), blonde, red, purple, black, you name it my hair has probably been it. Pictures from over the years are always a laugh.   3. How do your normally wear your hair?  On my head?  There is no normal in my world.  The only normal for my hair is clean.   4. If you could change your hair this minute, what would it look like?  It would be long... long, long and slightly wavy.   5. Ever had a hair disaster? What happened?  Since I colored my hair a lot over a span of 20 years there have been a lot of color disasters.  It\'s been green, orange, mud, etc.  As far as cut disasters, those are no big deal.  It can either be cut more or it can grow.  The only time I was ever really upset about a hair cut to the point I cried I was about 20 and my hair had been long for many years and I let my aunt cut it short and wow.. was it short... less than an inch short.  Of course, I\'ve worn my hair like that a lot the last few years but back then it was very tragic.   '),(22882,2061087,'18,October,2003','2003-10-18 00:00:00','   I wonder what it would be like -      Paris in the Springtime. To go upon a brisk, long hike      In Paris in the Springtime.  The flowers there would be in bloom,      A million different colors. The scented air would fill my room,      Plus a very many others.  The couples would walk along arm in arm,      Because this is young love time. The Eiffel Tower bears a special charm,      In Paris in the Springtime.  Sparkling music would fill the air,      Coming from a sidewalk cafe. The happy people would linger there      To while away the day.  It must be a magnificent sight -      Paris in the Springtime. Even to see the smallest bird in flight      In Paris in the Springtime.  Anne Wiley Carr, circa 1962   '),(22883,2061087,'17,October,2003','2003-10-17 00:00:00','   Un malheur ne vient jamais seul.   '),(22884,2061087,'15,October,2003','2003-10-15 00:00:00','   1. Name five things in your refrigerator.      Milk, soy milk, cranberry juice, 6 varieties of cheese and apple butter    2. Name five things in your freezer.      Ice, Brussels sprouts, Boston butt pork roast, pie crust, Blue Bunny sugar free Popsicles   3. Name five things under your kitchen sink.      Dishwasher detergent, Windex, trash bags, apple scented Lysol, Pledge    4. Name five things around your computer.      2003 Calendar featuring the 7 Up truck drivers, Pier One candle in glass holder, cup of Pepsi Vanilla, basket of cd\'s and junk, roll of Charmin (I forgot to get more Kleenex)    5. Name five things in your medicine cabinet.       Almost empty bottle of Wish by Chopard perfume, lip mask and conditioner, 3 different anti-perspirant/deodorants, pearls that need re-stringing, Pond\'s deep cleanser and make-up remover with cucumber     1. What vehicle do you drive?        \'94 Buick Skylark    2. How long have you had it?        Since December of \'02 (right after the asshat who ran the stop sign in front of Walmart totaled my Accord)    3. What is the coolest feature on your vehicle?        Uhh.. the a/c?   4. What is the most annoying thing about your vehicle?        THE most annoying??  There are several.  Only 3 of the windows roll down.  The power console on the driver\'s door is held on by duct tape, as is one of the a/c vents.  The power locks malfunction so I disabled them.  The dashboard almost falls off every time I hit a bump because I need new shocks.  The windshield is scratched up and you can\'t see out of it.  Well... you get the point.  I didn\'t buy the car because I wanted it, it was just the only thing available at the time and I had to have something THEN.   5. If money were no object, what vehicle would you be driving right now?      Cadillac Escalade ESV Platinum Edition in White Diamond    Ok, so don\'t just be nosy... I shared with y\'all.  Now y\'all put YOUR answers in the spankin new comment section below!   '),(22885,2061087,'15,October,2003','2003-10-15 00:00:00','   While I Live Out My Dream     \"My office is the cab of a Ford pickup.  I have no employees, punch no time clock and have no set work hours.  The tools of my trade are carried in a bag in the back of my truck and my paycheck depends on my body, the luck of the draw and the will of God.  I am a bookkeeper, a mechanic, a navigator and sometimes a consultant to my many friends.  I am a rodeo cowboy and I wouldn\'t have it any other way.\"   As I stand at another rodeo with this hat held over my heart, Cold chills run through my body when our National Anthem starts. I see \"Old Glory\" waving proudly out there on the arena sand, Held with honor by one of my buddies on a bay pony he understands.  The tears that stream the faces of fellow riders at my side, Send a message of admiration and our unselfish cowboy pride. For like me, they love this Country and stand for all She really means, And I thank God I\'m down here with \'em - while I live out my dream.  The Grand Entry now dismisses, then behind the bucking chutes, I search there in my riggin\' bag for the spurs that fit my boots. And as I fumble through the memories and the gathered-up supplies, I find a worn old belt buckle that I once wore for second prize.  Winning second wasn\'t special, but this old buckle is my joy, For it was made by a man named Ruddy - I wore it proudly for his boy. A mile or two it\'s traveled, now it\'s shine no longer gleams, But I carry the old buckle for Daddy - while I ride to live my dream.  Years of memories overtake me, and as I hold it in my hand, I think of all the Ruddy buckles that I\'ve seen across this land. Its gold inlay\'s a message of a craftsman\'s work of art, Though it rides now with my riggin\' - I still wear it on my heart.  In the rubies on each corner I see Mother\'s sparking eyes, And the reflections of her loving hand each time she waved good-bye. The buckles tarnished silver tells the color of Grandma\'s hair, I hugged her neck before I left - I\'m so proud that she was there.  I now feel a sense of lonely, while I wait my turn to ride, But I know my family\'s with me in this old buckle by my side. Lord, please watch over all my family and those sisters that I love, Monday is Kayla\'s birthday - send her something special from above.  My pony\'s in the bucking chute - soon I\'ll ride for another eight, But God, finding this old buckle made the time so worth the wait. Lord, should this be my final ride and I depart your Earthly scene, Ride me spurring into Heaven - while I live out my dream.    Written with love for the family of Chad Ruddy, 1975-1999 Lee Wiley Burch October 25, 1999   '),(22886,2061087,'14,October,2003','2003-10-14 00:00:00','   Hello, my friend. How are you today? I was thinking of you and thought I\'d come by and talk. It\'s a lovely day. You\'d like it. The sun is shining and everything is bright and pretty. I remember these were your favorite days. I brought you some flowers. I hope you like them. I really miss you, you know? I\'m sorry you had to leave so soon. I have to go now. But I\'ll come back again. Good-bye for now.   '),(22887,2061087,'12,October,2003','2003-10-12 00:00:00','   I look in your eyes And see all the joy and pain  Of the years we shared. I also catch a glimmer of tears For what you\'re feeling now.  I want to reach out And catch your hand in mine To tell you I still care. But, I seem to have no control. I\'m afraid that if I do You\'ll see all the feelings That are still there.  Yes, I still love you. But, I feel in my heart That if I don\'t leave you And go on with my life, All that I have planned Will fall apart.  So, as we are standing  By the gate to my plane, I finally find the courage  To reach out, Although I sway, And lay my hand against your cheek As a final good-bye, Pick up my bags, And walk away.   '),(22888,2061087,'12,October,2003','2003-10-12 00:00:00','   I\'ll admit.  I\'m a mom on the edge.  My sanity is goin\' south.  Tonight I tried to beat my son to death with the only weapon handy.  What I was holding when I snapped?  A large bag of miniature marshmallows.   '),(22889,2061087,'12,October,2003','2003-10-12 00:00:00','   How bout them Cowboys?  Woohoo!  4 and 1, baby.  The boys are back.  The better keep it that way!  I\'m glad that our team has been turned around and is doing it the way it should be done.  I\'ve been a fan since I was a little kid and in the last few years I have been so ashamed of the way they\'ve been conducting themselves.  The coaches and the players.  I still supported them, but I wanted to kick them in the shin, too.  I\'m proud of the way they have came back and are upholding the fine Texas tradition that is Cowboy\'s football.   '),(22890,2061087,'10,October,2003','2003-10-10 00:00:00','   I am still suffering from the Plague.  I\'ve had many things to say this weekend, just haven\'t had the clarity of mind to get them down without sounding like any more of a lunatic than I already do.  I have many words for what I think about our so-called medical professionals today.  I could (and have) rant for days about that.  I\'ll save that for later.  I am so upset by   urlLink this man\'s whole ordeal  (read the post and work your way forward for the whole story) that I could just spit.  No, I couldn\'t just spit.  I DO spit!  I spit on all the people who are the cause of things like that.  Not only what he went through in the first place, but the reaction of people to his experience.  Whether or not he is telling the truth, things exactly like that do happen and they happen because the people in this world today are for the most part stupid, selfish, hating, judgmental, trouble-making.... OK I am apparently not going to be able to keep from sinkin to their level right now so I\'ll save my full thoughts on this for another day.  I\'m not up to the whole First, Second and Fifth Amendment debate right now.  Therefore, my attention must go back to where it should be... the football.   '),(22891,2061087,'10,October,2003','2003-10-10 00:00:00','   I live in a town of about 100,000.  It\'s a smallish town in North Central Texas.  A college town.  An Air Force base town.  You would think that it would have a variety of eating establishments and entertainment.  If you did, you would be wrong.  Comparing this town to others of its size, be it state-wide or nation-wide, it is sadly lacking.  It\'s not a horrible place to live.  I\'m just sayin it could be better.  It is better than my hometown, a town of about 10,000.  But, it\'s not as good as other places.  The housing here is iffy as well.  I was not happy with the last apartment I lived in.  Circa 1970s, hasn\'t been remodeled since, 6 kids in the apt. upstairs, meth-doin\' neighbors, termites in the bathroom.  You\'d think it was in a really bad neighborhood or something.  No, it was one of the better neighborhoods here.  The outside was nicely taken care of.  But, not so the inside.  I heard about some brand spankin\' new apartments being built so I immediately went and got on the list.  I moved into this apartment complex last year, Halloween to be exact.  I was hoping that since it was new they would be more choosy about who they let in.  How they maintained the property.  You know.... doing things the way they are supposed to be done?  Wrong.  This place is a year old and it is already ghetto.  Oh, it\'s got the big fancy security gate.  That never works.  It\'s got the big, fancy office which reportedly has an exercise room and an internet room and a \"gathering\" room (to hold showers and such).  I\'ve yet to see them.  The buildings are nice.  Brick, good design.  But the lawn is atrocious.  Full of weeds, fire ants and scraps of trash.  The patios and front porches of some of these places are a fire hazard.  Now, there have been about 15 or so places clear out, be it their choosing or eviction.  But, there are at least that many more that need to be.  I cannot imagine why someone would want to move into a nice place and live like trash.  I cannot imagine why the management puts up with it.  (other than some of them are family, how great is that?)  But, enough of the maintenance.  First, let me tell you this.  When they got the security gates funtioning, they sent around a notice that if you wanted a remote you had to pay $XX.XX and $X.XX a month.  They also put up these \"carports\" (and I use the term loosely) in various parking areas along the fronts of the buildings.  It is expected that if you want one of these parking spaces under the carport that you pay $XX.XX a month for that privilege.  Well, I have parked in the same place since the day I have moved in, due to the arrangement of parking and the number of people that live in my building.  It always worked out fine, we always had our own parking place, and it was sort of an unspoken agreement that we don\'t take each others place.  After all, we were all close to our doors.  Well, when they put up the said \"carports\", they took 4 of the 6 spaces in front of my building.  One of them was \"my\" space.  So, I ask you... should I have to pay to park in that space?  Or should I have to park in another space and fight it out with the neighbors over the available spaces?  Usually forcing either myself or them to park buildings away?  I didn\'t think so.  On reason being I am technically supposed to have a handicapped parking space, as are my immediate next door neighbors.  That would take up the two parking spaces closest to the doors.  Which are the ones we were parking in before the great \"carport\" fiasco.  So, I go to the office and register my complaint.  They can either shut up and let me have my parking space back, covered or no, or they can put in 2 handicapped spaces there.  Well, we can\'t do that, how about you pay us $XX.XX instead of the $XX.XX we want.  My response?  What part of I am not paying to park in front of my own apartment do you not understand?!?  Needless to say, they saw things my way.  (Mentioning N. TX. Rental Association and lawyers help) There are many other issues I have with the management of this place, such as some of the hand-lettered posts on the bulletin board and some of the letters that they send about.  If I were more ambitious and if I thought the people who lived here could be cohesive enough to back me I would start a tenants association.  But, to tell you the truth.... I\'m tired of the whole mess so I\'ll just deal with what immediately affects me and wait for my out.   '),(22892,2061087,'09,October,2003','2003-10-09 00:00:00','   Jolene, Jolene, Jolene, Jolene  I\'m beggin\' of you please don\'t take my man   It flitted through my mind for some reason the other day how much I have always liked Dolly Parton, yet don\'t listen to her unless she happens to be on the TV or the radio.  Of course, since I don\'t listen to country radio that doesn\'t happen.  She was on Jay Leno tonight and it reminded me that I really should start listening to her more.  Anyone who knows me knows that while I am NOT a fan of country music, my tastes are very diverse.  I won\'t exclude (thanks, D) anyone simply due to the genre of music they are in.  There is country that I like.  I love bluegrass, the blues, gospel and good old rock and roll.   I could go on listing but you get the idea.  Everything from Rachmaninoff to Rammstein.  However, there are many musicians who slip through the cracks of my mind.  Last week, there was a tune niggling at the back of my head.  It teased me and teased me for days on end.  Finally, one night I wracked my brain and finally managed to get hold of it.  It\'s a great 80s hit that has a timeless appeal.  Go  urlLink here  to check out the tune (click on Sound Library/Video clips) or  urlLink here  for the lyrics to Billy Vera and the Beaters \"At This Moment\".   '),(22893,2061087,'08,October,2003','2003-10-08 00:00:00','   I want my mommy.     I had to go to the doctor today to find out why my hormones and stuff are all out of whack, and of course they had to stick a large, sharp object in my arm and gouge it about.  I\'ve been having some allergy symptoms for the last few days.  As usual this time of year.  I love the change of seasons, but I get sick every time.  And today when I came home from the doc, I could feel it comin on.  The burning lungs.  Flushed skin.  The large object lodged in your throat, quite possibly a tumor.  The coated tongue.  Next thing you know you have a full-blown case of the Plague.   As far back as I can remember, when somebody got sick either my Mamaw or my Momma would cook up a big ol\' pot of Pink Soup.  To this day, when I am sick that\'s what I want.  And when Momma\'s close enough, that\'s what I get.  It\'s a sure-fire cure all.  Sure, I could go in there and cook some myself; but, that isn\'t the same thing.  I guess I\'ll go get me a hunk of cheese to go with this whine.   '),(22894,2061087,'08,October,2003','2003-10-08 00:00:00','   Yeah... YOU    Where are you? I know you\'re out there. I just haven\'t found you yet. Why are you hiding? I\'ve been looking. I just don\'t know which one is you. Have I overlooked you? There\'s been many who I thought were you. But, after they had gone I knew. I\'m tired of looking now. Where are you?   '),(22895,2061087,'08,October,2003','2003-10-08 00:00:00','   A list of 100 things regarding oneself seems to be the blogger thing to do.  So, I figure I better get me one of em started.  Good thing I am a purty good fabricator.  Funny coinkydink, today when I visited  urlLink Jett Superior  her post was about just that... how hard it is to come up with all that stuff about oneself.  Go check it out.  This is going to be an ongoing post.. I\'ll work on it when I have time or come up with some lies... err facts to post about meself.    A Hunnerd Thangs  1.  I was born Wednesday, July 14, 1971.  2.  Why, yes... I am from Texas!  How did you guess??  3.  I\'ve lived here all my life; but, I have always wanted to move to Louisiana, Mississippi, or points farther South (ok, technically east).    4.  I am not a morning person.  (or afternoon or evening, depending on when I am awakened)    5.  I\'ve been that way since I was a small child.  I would stay up all night and then when my Momma would wake me up to go to school, I\'d go into the bathroom and lock the door and fall asleep on the floor.  6.  I have an above average IQ.  But,  when I was in the 5th grade I had the teacher from hell and decided I hated school and was never going to do any good at it.  I somehow managed to succeed at that and yet graduated from high school and went on to take college courses.  My opinion has changed drastically in the following years and I plan to finish my education.  7.  When I was in the 4th grade, I spent a month in the hospital after I had an appendectomy and got a staph infection.  They opened and closed the site four times before finally leaving it open to heal.  I went home with a gaping wound and had to take baths in Tide.  I have a four inch ugly scar to this day.  8.  I got out of the hospital a couple of weeks before Halloween.  That year I went trick or treating dressed as the Fat Farmer.  In my Daddy\'s overalls stuffed with pillows.  With boots that were too big.  We went in my mom\'s best friends station wagon with a bunch of kids.  I had to ride in the back and roll out at each stop and then roll back in later down the block.  One stop I didn\'t roll in fast enough and was almost halfway in and they took off and I was left standing there.  Fortunately, they came back for me.  After a while.  9.  Thanksgiving and Christmas are my favorite holidays now.  10.  My favorite season is Fall.  I love the smell of the air and the turning of the leaves and the cold weather coming on.    11.  Fall means football season!  I love football, have watched it with my Daddy since I was a little girl.    12.  I also got a great love of cars from my Daddy.    13.  I have a big nose, I got that from both sides of the family.  But, it\'s mine, I\'ll keep it.   14.  My feet are a size 10 and 1/2.  15.  My eyes are gold around the pupil, blue-green in the middle and a dark blue-grey ring around the outside.  They change color with my emotions.  If they are bright green... watch yourself.  16.  I am tone deaf and can\'t carry a tune in a bucket; but, I have a great love of music and am very attracted to musically inclined men.  17.  My listening tastes range from classical to thrash metal, touching on everything in between.    18.  I want a man who can sing to me (without me cringing).  19.  I have a long list of what I am looking for in a man.  Yes, written down.  20.  I am a big list-maker.  21.  I am good at losing them.  22.  I love books.  I\'ll read just about anything I can get my hands on.    23.  I read Gone With the Wind for the first time when I was 13.   24.  And Mandingo.  And many, many other trashy novels.  25.  I want a big ol\' Southern house with a library.  Painted red with dark wood. (the library not the house)  26.  I have a real thang for a man with a Southern accent.  (Yes, that is on the list.)  27.  My love for all things Southern goes way back.  I think it was all those bodice-ripping Civil War novels.  28.  I\'m a very good cook.  29.  But, not of \"haute cuisine\".  My specialty is \"down home\".  30.  The smell of lilacs takes me back \"home\".  31.  I spent my 21st birthday in Las Vegas.  32.  I ate calamari for the first time in a fancy Italian restaurant there and had an anaphylactic reaction.  33.  I am allergic to numerous things.  Calamari (obviously), Penicillin, Sulfa, Macrodantin, and several other drugs.    34.  When I was about 8 years old we were visiting my Granny Geneva.  She had a large Magnolia tree in her front yard.  I had never seen one and was fascinated by it.  I plucked me a big, glossy, fragrant bloom and walked around sniffing it.  After a while, I started to feel kind of funny and couldn\'t breathe so good so I went to tell my Momma and when I walked in the door both of them freaked out.  My face and neck were grotesquely swollen and I was covered in red spots.  Epinephrine is my friend.  35.  I started smoking, drinking and getting high when I was about 13.  Many years of this ensued until when I was 23 I started having allergic reactions to pot.  Pot is not my friend.  36.  I quit smoking, too.    37.  I will still enjoy the occasional drink, but not to excess.  38.  I was arrested once for Driving While Intoxicated.  Funny thing is.... it was the ONE time that I wasn\'t.  They let me go.    39.  I once fell out of my car into the street when a police officer pulled me over for suspected DWI.  He let me go, too.  40.  I\'ve been involved in hanky panky in the front seat of a police car.  No, it was not to get out of a ticket.  41.  Starbuck\'s Mocha Malt Frappuccinos contain crack.  42.  I want to travel the world.  France, Great Britain, parts of Europe, Africa, New Zealand are a few places I want to see.    43.  When I was in high school, I took French so that when I found a man who made all the checks on my list and got married and he whisked me off on our honeymoon to Paris I would know what people were saying.    44.  The man I did marry didn\'t make all the checks and he didn\'t whisk me off to Paris.  45.  I am still looking.    46.  I don\'t want the honeymoon in Paris anymore.    47.  I want to elope and have a nice, quiet (well, ok, maybe not quiet) honeymoon in a cozy cabin either in the mountains or on the beach.  48.  Paris can come later.    49.  I do want a beautiful, classic.. even antique wedding ring.  50.  I am not one for a lot of jewelry; but, I do want some special pieces.  51.  I once received a vacuum cleaner and a breadmaker.  A man who does that is not on the list.  52.  A man who will vacuum and make bread is.  53.  But, I am old-fashioned that way.  I want to stay barefoot and in the kitchen.  54.  Notice I didn\'t say pregnant.  55.  I have 3 children.  56.  I have also had a hysterectomy.  57.  A man who is okay with that is on the list.    58.  I would like to have more but that is up to the one in #57.  59.  I\'ve had six different surgeries.  60.  I have two more pending.  61.  One of them is a total bilateral hip replacement.    62.  Sometimes I feel 72 instead of 32.  63.  But, I am a kid at heart.  64.  I love cartoons.   65.  Who lives in a pineapple under the sea?  66.  I read the LOTR waaaay back when.    67.  The LOTR Trilogy is one of the few movies made from a book that I actually like.  68.  Ok, love.  69.  Legolas.  70.  I like Harry Potter, too.  Not the books.  I want a Gryffindor scarf.  71.  I\'m a big fan of all things medieval-ish.  Be it mythical or historical.  72.  I have a small collection of dragons and wizards and would like to make it a big one.  73.  I\'m a collector of stuff.  74.  I love antiques.    75.  That big ol\' house is going to be filled with them.  76.  And the ringing sound of laughter.  77.  And yellin\', I\'m sure.  78.  I have a filthy mouth and an extensive vocabulary.  No, they aren\'t all cuss words.  79.  My voice is different than most people expect.  80.  I have been told I should be a phone sex operator.  81.  I have a warped, wicked sense of humor at times.   82.  I love British comedy.  TV and books.  83.  I wish I had a big screen TV and a huge collection of 80s movies.   84.  And a couch as big as a boat.  85.  I wish I knew all the words to Red Wing and a voice to sing it properly.  86.  My grandfather sang that to me when I was a child.  87.  I was with him when he died.  88.  The day we buried him it was 117 degrees.  89.  \"Amazing Grace\" played on the bagpipes makes me cry.  90.  My ancestry is English, Irish, Scottish, German, French and American Indian.  I am 100% American.  91.  My ancestors on both sides settled in the South before moving to Texas after the Civil War.  92.  I\'d like to see a dragonfly in amber someday.  93.  I am drawn to very old cemeteries.  94.  I want to have sex in them.  95.  I enjoy sex outdoors.  Indoors.  I just enjoy sex.  96.  I\'m not adverse to PDAs.   97.  I have taken a vow of celibacy until I find the one with all of the elements.  98.  Gawd, I hope that\'s soon.  99.  It will be so worth it.  A hunnerd.  That is what it\'s all about.   '),(22896,2061087,'08,October,2003','2003-10-08 00:00:00','   In My Dreams   You were here in my dreams last night.  But now, as the light of a new day awakens me I am slowly brought back to reality.   You\'re gone. You\'re not here. You\'re miles away from me.   I only have to close my eyes to see you, feel your breath upon my face and feel the touch of your hands upon my body.  I don\'t understand this desperate loneliness when I think about not being able to touch you.  My eyes open and where I felt warm and safe, I now feel cold and alone with the harsh realization that you are not really here with me.  The months stretch out before me, long and desolate and I feel an ache because of the amount of time there has to be before I can hold and be held by you.   I long for the days of laughter and conversation and the sweet ecstasy of the nights.  Even though our time together has been short and tinged with the knowledge of the fact that there is much more than distance between us, I feel as if you have been in my life forever.  Making you happy pleases me, and you bring out a feeling in me that I have never experienced.  I enjoy your presence even if you are only nearby.  But, for now, I will bide my time, and not think about the time that will pass before I can hold you. Until then, I\'ll see you in my dreams.   '),(22897,2061087,'08,October,2003','2003-10-08 00:00:00','   On With the Wind   I figured I\'ve been reading these things long enough that it was about time that I got one of my own.  I have a lot in common with the great Scarlett, I am a master procrastinator.   \"I can\'t think about that right now. If I do, I\'ll go crazy. I\'ll think about that tomorrow\" and all that.  But, in my daily reads I am always coming across something that brings to mind a past occurrence or an opinion or some such drivel.  It is high time I started doing something with them.  So, without further adieu, here goes....   '),(22898,2061087,'30,November,2003','2003-11-30 00:00:00','   Without the snow.  And the giant tree.  And the blinding lights.  Ok, so it doesn\'t look a lot like Christmas around here.  Yet.  Have y\'all seen those pine cones that are cinnamon scented?  Mmmmm yum!  Those are awesome.  I found them at Michael\'s and WalMart both.  However, they are a bit bunged up.  I\'m kind of obsessive that way.  I don\'t like for things to be messed up in anyway, especially not something you have on display in your home.  So, I didn\'t buy any.  Instead, I thought to myself \"Self, you can MAKE these for a whole lot less money and have them look a whole lot better.\"  And the kids will have fun at the park pickin up pine cones.  We always go feed them to the prairie dogs anyway.  May as well throw a few in a sack and bring em home and cook em up.  Then it will at least begin to smell a whole lot like Christmas.   '),(22899,2061087,'17,November,2003','2003-11-17 00:00:00','   Well, not exactly all the way, but there were times when I wanted to.  I had my gallbladder taken out Tuesday, November 25th so things have been a wee bit sore-ish the last few days.  But, I\'m not sick anymore!  I can handle sore knowing it is going to go away WAY better than I can being sick 24/7.  The surgery went fairly well.  I have way more pain than I anticipated due to having adhesions all the way up to my gallbladder from the appendectomy gone awry.  And, when I came to in my room (having no memories of recovery room or traveling from point A to point B to point C) I was hurling, which is nothing new, I always do that after anesthesia.  Except this time it was bright red blood.  Imagine if you will a highly drugged person seeing this.  Come to find out they tore something in my nose when they put the NG tube down it during the surgery and my nose had been bleeding the entire time down my throat.  GAH-ROSS!  And scary to see coming up!  Thanksgiving dinner was good for everyone else!  I wasn\'t able to eat more than a couple of bites that day.  And that was only white meat turkey, green beans, some sweet potatoes and roll since I am on a low-to-no fat diet for a while.  But, I\'ll gladly sacrifice this Thanksgiving because now I know that the rest of them will be enjoyable.  This morning I started sneezing.  That wasn\'t a lot of fun.  I think I popped a few stitches loose before I got it under control.  OUCH!  I may have also invented a few new words.  My belly button is some rather interesting hues of blue and black and green.  There are 5 lovely new soon-to-be scars added to my collection.  And the worst thing?  I can\'t drive for at least 2 weeks!  ARGH!!!!  Fortunately my mother is here to cart me and the chilluns around.  Especially since I can\'t pick anything up either.  Certainly not 32 pounds of wriggling, writhing Willabeast.  That child is a buncha handfuls.  So, I\'m thankful for the help.  I will be glad to get my independence back though.   '),(22900,2061087,'17,November,2003','2003-11-17 00:00:00','   I lurve Sarah.  I guess my first favorite song of hers was  urlLink Ice Cream .  I really love  urlLink Angel ,    urlLink Hold On ,  urlLink Sweet Surrender ,  urlLink Do What You Have to Do ,  urlLink I Love You , and  urlLink Possession .  Don\'t get me wrong, I love all her songs, but these are the ones that I can really identify with.  It\'s funny, over all the years, that these songs have never been associated with any person... other than that person out there who I hadn\'t met yet, you know... THE ONE?  The one that speaks to me off of her new album, Afterglow, is  urlLink Push .  And I do associate it with someone; but, circumstances being what they usually are... well... I don\'t know what to say about that right now, so I won\'t.   '),(22901,2061087,'16,November,2003','2003-11-16 00:00:00','   Funny my last post should be about inappropriately medicating patients.  I have been sick for around a month and a half now.  Let me start at the beginning to avoid any confusing backtracking or asides (well ok they won\'t be avoided since I am highly prone to doing them anyways, but at least the story will be chronological).   About a month and a half ago I was suffering from the usual fall/winter sinus problems, which as usual turned into a sinus infection.  I get at least two a year, in the spring and in the fall.  I went to the family doctor I have been using and was seen by his Physician\'s Assistant.  I told her the problem and the history and what meds I was taking.  She prescribed for me an expectorant.  I told her that I was already taking an expectorant and that would not do me any good, I needed some Zithromax.  She did not agree and dismissed me.  I was back two days later:  couldn\'t breathe, wheezing, head exploding... all of the usual sinus infection/bronchitis symptoms.  She said to me \"It\'s only been 48 hours, why are you back here?\"  She is a rather snotty bitch.  I told her why and she ordered a chest x-ray and dismissed me.  I said \"Enough of this bullshit\" and went to the clinic where I used to work and saw one of the doctors and she said, sure enough, you not only have a nasty sinus infection but bronchitis.  She gave me the proper meds and I was well within a week.  Now, I had already taken that round of antibiotics.  I went to my yearly physical with my GYN and he discovered that, unbeknownst to me, I had a bacterial infection.  So, another round of antibiotics.  And if you\'ve ever taken Flagyl, you know how bad they tear you up.  Within about 5 days, I thought I was gonna die my stomach and chest hurt so bad.  So, I went to the ER one night when I was having a particularly bad attack and they told me I was having a \"panic attack\".  Now, I am not one to have panic attacks, but what do I know?  I\'m not a doctor.  Not even a week later, I was back in the ER with another bad attack.  The doctor I saw that night said I was suffering from gastric disturbances due to the antibiotics and that it was GERD (fancy words for real bad heartburn) and gave me some meds and said to follow up with my family doctor.  Which I did, and he gave me some Nexium.  Which did me absolutely no good.  Another week passes and I am still having horrible stomach aches, chest pain, nausea, etc.  Another bad attack sends me to the ER again.  The nice doctor I saw that night actually bothered to do an exam and found that I was having gallbladder attacks.  So, he gave me some appropriate medicine and discharged me with an appointment in the morning to have a gallbladder sonogram done.  The next morning I have the sonogram done and they find that my gallbladder is full of \"sludge\" and \"floaters\"  (which are stones forming) and it is inflamed and needs to come out.  So, of course, it\'s back to my family doctor for a referral to a surgeon.  Which I got that day and had the appointment with the surgeon the following day.  I go to my appointment, give them my information and sit down to wait.  I am finally called back to an exam room and the nurse says the doctor will be with you in a few minutes.  Another nurse comes in a few minutes later and gives me some brochures regarding the surgery.  Then the doctor finally comes in and sits down and says \"So, you have gallstones\" and I say \"Well, not yet, but...\" and he interrupts me and says \"Yeah that can cause you some pain.  Now, we can probably do it with the laproscope but we may have to do an open cholesystectomy.  I\'ll have the nurse come in and get you set up\" and he walked out.  Now, I am sitting there going \"What the hell?  He never asked me my history, concerns, NOTHING!\".  So, in a few more minutes the nurse comes back in and says to call her in the morning and she will tell me when the surgery will be.  I asked her how long she thought it would be and she said \"Oh, with your insurance we have to call the hospital and THEY tell US when the surgery can be done.  It will probably be about two weeks.\"  And I said to her that was not going to work (and had anyone bothered to TALK with me they would have known this).  So, I left very unhappy with the whole experience.  I called my family doctor the next morning and explained to them what was going on:  that I was miserable and could not function with the pain I was experiencing and I have 3 kids to take care of and I could not wait two more weeks and besides that I was not comfortable with the surgeon they referred me to.  Now, keep in mind, I am having to deal with the snotty bitch PA this whole time.  She starts trying to tell me a bunch of bullshit about that is how it is done and I would just have to accept that.  I told her that where I come from it is not done that way, that the doctor finds out the situation and calls the hospital and tells THEM when the surgery would be.  She said to me that maybe that was what I should do and I said to her oh I will.  She then tells me \"Well, when they won\'t help you don\'t come back to us and expect us to help you because we won\'t.\"  I was quite speechless then.  I think I said something along the lines of \"Oh, believe me I won\'t\" and hung up.  I called my doctor/friend back home and he immediately set me up.  The next morning the mail lady knocked on the door with a certified letter for me.  Seems that I have been terminated as a patient for \"inappropriate behavior to physician\'s office staff and inappropriate behavior to a referring physician\'s office\".  Now... I\'m sorry... I\'ve been misdiagnosed by these yayhoos for three weeks (well, actually longer counting the sinus thing, so that\'s twice in a row), I\'m in such pain I can\'t function (never once having been offered the appropriate treatment) and I never ONCE cursed or raised my voice, merely tried to explain my concerns regarding the situation and I\'M the one who has behaved inappropriately???   I don\'t even think so.  They have not SEEN inappropriate.  I am of the mind that this town is trying to kill me one way or the other and this... this is the straw.  I mean... do I not have the right to a second opinion?  Or to express my opinion period??  I guess it is all well and fine that they have been guilty of malpractice; but, I am supposed to just shut up and take it?  No.  I\'ve done that for far too many years.  I am not exactly sure what I can do about this situation legally, but you can rest assured that I will be filing complaints with the proper agencies AND writing a letter to the editor of the local paper among other things.  Sometimes I think it is just me that has these problems, but I know that can\'t be true.  It has to be happening to other people and I for one am not going to take it any more.  Gawd, major flashback.  \"We\'re not gonna take it, we\'re not gonna take it, WE\'RE NOT GONNA TAKE IT ANYMORE\" courtesy of the lovely Dee Snyder and Twisted Sister.  Great anthem.   '),(22902,2061087,'11,November,2003','2003-11-11 00:00:00','   Couldn\'t have said  urlLink it  better myself.   '),(22903,2061087,'10,November,2003','2003-11-10 00:00:00','   What many people don\'t know or don\'t remember is that Dolly\'s hit \"I Will Always Love You\" is not the kind of love song that it has been made out to be.  Sure, it\'s a somewhat generic tune regarding love and moving on; but, she wrote it when she made the decision to end her long-time singing partnership with Porter Waggoner.  To me it evokes the epitome of love and friendship, the generosity of giving your friend what they need (while it may not be what either one of you WANTS).  Life works that way on occasion.  While you may love someone (a friend, a lover, even a family member), sometimes it just isn\'t meant to be.  And what good are hard feelings?   It\'s difficult to not have them, but when you stop and think about it, isn\'t it better to wish someone the best?  If you truly love that person, that is what you want for them.  And sometimes it isn\'t you.  I Will Always Love You  Dolly Parton   If I should stay,  I would only be in your way.  So I\'ll go, but I know  I\'ll think of you ev\'ry step of the way.   And I will always love you.  I will always love you.  You, my darling you. Hmm.   Bittersweet memories  that is all I\'m taking with me.  So, goodbye. Please, don\'t cry.  We both know I\'m not what you, you need.   And I will always love you.  I will always love you.   (Instrumental solo)   I hope life treats you kind  And I hope you have all you\'ve dreamed of.  And I wish to you, joy and happiness.  But above all this, I wish you love.   And I will always love you.  I will always love you.  I will always love you.  I will always love you.  I will always love you.  I, I will always love you.   You, darling, I love you.  Ooh, I\'ll always, I\'ll always love you.   '),(22904,2061087,'08,November,2003','2003-11-08 00:00:00','   So, I was watching the CMA\'s (yes, Country Music Awards) the other night for some reason.  Mainly, the reason being that they said Norah Jones was going to be on so I HAD to check it out.  Come to find out there is a  urlLink tribute CD to Dolly Parton  and she\'s one of the artists.  Holy shit, y\'all.  They blew me away.  I love them both, in case no one has noticed.  I had to go to Walmart tonight and I was killing time, cuz you know... I have nothing better to do on Saturday night than hang out in Walmart... and I went by the music dept. to check out the CD and ended up buying it.  HOLY SHIT, Y\'ALL.  As I\'ve   urlLink said before  I am a Dolly fan from WAY back, I grew up listening to her and all of the other \"old-time country\" folks.  As I sat and listened to this CD for the first time tonight I\'ve waxed nostalgic, I\'ve cried, I\'ve been uplifted and a whole host of other emotions.  That\'s why I love Dolly\'s music.  And to hear it translated by other notable female artists....well...holy shit y\'all.  I\'ve only bought like 4 CD\'s in as many years and this one was well worth it.   I don\'t buy a lot of CD\'s because I usually only like one or two tracks on the whole thing.  The ones I was compelled to buy were Enya, Tracy Chapman, Norah Jones and this one obviously.  While I was out there I noticed   urlLink Sarah McLachlan  has put out a new one, first one in like five years.  Will be having to get that one soon.  There are a couple of other full CD\'s I\'d like to add to the collection, the new   urlLink John Mayer  for one (okay, for three because I need all of them) and a few others.  Oh, and how could I forget  urlLink Jonny Lang ?!?  If y\'all haven\'t heard this kid yet, immediately DO SO!  I had pirated most of it at one time and lost it in a computer crash so now I must go about it the legal way (damn the RIAA anyways).  But this kid... gives the blues a whole new meaning.  Ok, that is the news in music for today.  I am going to go hit play again.  (yes I have a remote but the batteries are dead, damn AA batteries to hell too)   '),(22905,2061087,'05,November,2003','2003-11-05 00:00:00','   Having a fascination of all things Celtic and graveyards as I do, I was delighted to run across  urlLink this site and these photos  today as I was rambling around the \'net.  Her entire site is beautifully put together and fed my voyeuristic tendencies more so than usual since I also got to travel to a place that I hope to visit myself some day.   '),(22906,2061087,'05,November,2003','2003-11-05 00:00:00','   Have you ever been going about your business - driving in your car, shopping, whathaveyou - when you see someone and you say \"Hey, there\'s ol\' whatshisface!\"?  I do that all the time.  Usually, it\'s that person.  Occasionally, it\'s just someone who looks a whole lot like them.  I did that the other day, driving back from Walmart.  I looked over in the truck beside me and my mind said \"Hey, that\'s Chad\".  But, right on the heels of that was \"No - no, it can\'t be, you dummy\".  Chad\'s been gone for three years.  I still do that though.  For a long time after he was gone, I saw his face everywhere.  Every time I saw one of the trucks from the business he worked for, I hoped it would be him.  It\'s funny how the mind tries to deny a loss in little ways like that.  I\'ll never forget the last time I saw him.  (Well, I guess that would be the next-to-last time.)  He came by late one night.  It was around midnight, I was in my pjs, ready for bed.  I was tired and didn\'t invite him in.  He was just coming by to tell me that he was going to be on the road again for a while, had a rodeo to be at in Del Rio and then on to San Antonio and who knows where else.  I said the usual be carefuls and have fun.  He said he\'d see me when he got back.  That weekend I was at work on the nightshift at the hospital.  When I answered the phone and it was my cousin, I knew something was wrong when she said hello.  She went on to tell me that Chad had been in an accident between Del Rio and San Antonio earlier in the night.  I was instantly filled with dread.  My heart sank further and further as she went on to say that the ambulance picked him up from that road in the middle of nowhere and they took him to the hospital but after working on him for hours they just couldn\'t get him back.  When she said that, my heart stopped sinking and was just.... numb.  I thought maybe she had it wrong.  Small town grapevines aren\'t always accurate.  In the back of my head, I knew it was true.  I just didn\'t want to admit it.  I\'d lost many friends in the same way.  Chad was special and his loss hit especially hard.  He wasn\'t a big guy but he had a big personality.  He was on his way to becoming a big star in the PRCA in bronc riding.  The funeral was ... well, it was indescribable.  Sure, I could tell you what it looked like, what they said and the music they played, and who was there.  But, the feeling... I can\'t put that into words.  It was huge.  The emotion was so strong it was as if you could reach out and grab a handful of it out of the air.  The thought that kept going through my mind was the scant few nights before.  When he was standing on my front steps, in the glow of the porch light.  I learned the lesson of treating every goodbye as if it were the last.  Because... it very well could be.  I had forgotten about that.  I won\'t forget again.   '),(22907,2061087,'28,December,2003','2003-12-28 00:00:00','   It\'s a wonder it hasn\'t withered up and blown away.  This is why I don\'t have houseplants.  I wonder how my kids have survived?!  I\'ve been somewhat busy lately what with getting ready for Christmas (what little I did).  Mostly I have just been a lazy ass and haven\'t wanted to post.  I\'d rather waste my time reading about everyone else\'s seemingly cool lives.  I\'m glad this year is almost over.  I hope that next year will be a better one.  It has to be.  I was dreading Christmas really bad.  I know it isn\'t about the presents and all, but tell my oldest two kids that.  I haven\'t always been able to provide them with a huge Christmas, but it has always been really good stuff.  This year my son came and crawled in bed with me Christmas morning.  He woke me up and whispered to me \"Mom, Santa came.  He sure didn\'t leave much.\"  I wanted to close my eyes and either cry or go back to sleep forever.  One group present (a $35 DVD player from Walmart and a DVD) and one present each (a tshirt) just doesn\'t cut it.  Yeah, yeah, I realize there are kids who got far less, and that there are kids who don\'t have parents or at least mothers who love them like I do my kids... and that their dad and step-mom and his mom more than compensated... but that isn\'t the point.  I let them down.  I want better for my kids.  And I feel really bad that circumstances have been what they have been lately and this is the fallout.  Well, enough of that.  It\'s over and done and all I can do is hope that next year is better.  How was y\'all\'s Christmas?  (Yes, that is too a word)   '),(22908,2061087,'11,December,2003','2003-12-11 00:00:00','   It is so nice to know that  urlLink chivalry isn\'t completely dead.   I have long searched for a man who does the little things, holds doors open and the like.  And for one to take it above and beyond?  There are no words.   Other than \"Point him out\" and \"Is he single?\".   '),(22909,2061087,'10,December,2003','2003-12-10 00:00:00','   Today has sort of been one of those days.  You know the ones?  Where your mood changes like quicksilver and you feel out of control?  Nothing has gone your way for so long that you don\'t ever see it coming again?  I have always been a \"driving in the car really fast with the stereo blaring the nastiest rock available\" kind of stress reliever.  Today, I had enough so I had to get out.  My hoopty doesn\'t have a cd player so I just listen to the radio in the car.  Fortunately, we now have a  urlLink decent radio station  here.  So, I was out for a drive... no, not a fast one, I am not recovered enough from the surgery for the maneuvering involved... and was listening to the radio when  urlLink this here song  came on.  Now, I haven\'t liked anything of theirs in a few years, but when I heard this one I was like \"HAYULL YEAH, THAT\'S WHAT I\'M TALKIN ABOUT\".  Yeah, it\'s a bit extreme, but there are days when we feel like that about people we really do love.  I have felt that way a lot lately.   '),(22910,2061087,'09,December,2003','2003-12-09 00:00:00','   I am not just a nitwit! Watched Rudolph for the umpteenth time tonight.  I love that song!   '),(22911,2061087,'08,December,2003','2003-12-08 00:00:00','   I know it\'s a bad habit to post song lyrics and call them a proper post.  But, music is a big part of our lives.  Or some of our lifes anyway.  And sometimes I hear the lyrics to a song and I\'m like \"Yeah.... that\'s it.\"  Something I have in my head, a conflict or feeling that I can\'t put into words myself.  I\'ve long been a fan of Oasis and I had forgotten about  urlLink this song.   It is so fitting at this point in my life.  Well, any point in life, but it really speaks to me right now.   '),(22912,2061087,'05,December,2003','2003-12-05 00:00:00','   Give me an Average Joe ANYDAY!   '),(22913,2061087,'05,December,2003','2003-12-05 00:00:00','   I have a secret fetish.  Ok, so I have several but let\'s stick with this one for now, shall we?  People see mistletoe and automatically think about kissing.  What do they think about when they see a Christmas tree?  Presents?  Pretty lights and ornaments?  Me, I think about something totally unrelated.  When I see a Christmas tree, my mind goes to other things under the tree.  Like two naked bodies intertwined.  Yes, that\'s right.  I have a sex under the Christmas tree fetish.  Am I alone in this or do other people think of the same thing?  I think I must be fairly alone in this because while this is a major fetish for me, I\'ve never once actually DONE it.  Someday, though, I will.   '),(22914,2061087,'04,December,2003','2003-12-04 00:00:00','   In a conversation earlier today, I was saying that one reason I haven\'t been to the movie theater in many years is because I don\'t like going by myself.  No sooner than the words were out, I was thinking \"What?\".  I do so too enjoy going to the moving by myself!  I enjoy going out to eat by myself.  I enjoy going shopping by myself.  In fact, I have no problem going anywhere by myself.  Ok, maybe I do regarding bars or clubs.  But, as far as anywhere else is concerned, especially the movie theater, I\'ve never had any qualms about going alone.  I have been saying this for many years, however.  I think it is due to the fact that I am tired of being alone and I simply want someone to go with.  I also think it\'s somewhat of a cop out, a pat reason why I don\'t get out and do something when and if given the chance.  Isn\'t it funny how if you say something often enough and long enough, you start to believe it yourself?  I have convinced myself of many things that way, in retrospect.  I am positive that I am not the only one.  I think we all must do it.  \"This relationship is working/will work.\"  \"I enjoy being single.\"  \"My ass doesn\'t look fat in these pants.\"  I\'m thinking a re-evaluation of some of the things we have been telling ourselves and others for years is in order.   '),(22915,2061087,'30,January,2004','2004-01-30 00:00:00','   You have just won one million dollars:  1. Who do you call first? My Mom and my Dad.  2. What is the first thing you buy for yourself? A Cadillac ESV  3. What is the first thing you buy for someone else? A Tahoe for my Mom.  4. Do you give any away? If yes, to whom? Yes, to family.  Actually it wouldn\'t be so much giving it to them as buying them things.  5. Do you invest any? If so, how? Yes, I would invest most of it.  How I don\'t know, I would have to either research the best possible way to do that or hire an investment manager.   '),(22916,2061087,'29,January,2004','2004-01-29 00:00:00','   I have expensive tastes.  Here lately I haven\'t been able to support those habits as in the past.  Now, I just settle for adding things I want to my gigantor list.  I was in Dillard\'s the other day buyin some pantays and of course I had to go by the perfume counter.  Bad idea.  I have been out of perfume for a long time now, and since I only like the expensive ones and would rather go without than wear anything cheap, I have been without for far too long.  The new  urlLink Cacharel Amor Amor  caught my eye and my nose first.  Then out of the corner of my eye I saw the familiar Boss logo.  Me love Boss long time.  So, I knew I\'d like it without smelling it, but I had to anyway.  Oh My Gawd.  It\'s called  urlLink Boss Intense  and it smells like ... well, let\'s just say I got all turned on there in the big middle of Dillard\'s.  I must have.   '),(22917,2061087,'28,January,2004','2004-01-28 00:00:00','   Like me, Joelle is a little tired of the whole  urlLink Atkins, low-carb business.   She put it much better.  I am still laughing my arse off.   As beautiful  urlLink as she is , I\'d dig him up for her.  Yes, I have a crush.   '),(22918,2061087,'28,January,2004','2004-01-28 00:00:00','   trapshootinguytx: my nuts were frozen..i walked all over campus trapshootinguytx: in like 25 degrees..with a wind chill of 12 and wind like a muther b0rderline_fo0l: well uh cover em up next time trapshootinguytx: i did, i had jeans trapshootinguytx: and boxers trapshootinguytx: i would keep a hand in my pants if i could, might look funny b0rderline_fo0l: well if they were frozen obviously you need more b0rderline_fo0l: long coat?  crotch scarf?  put a sock on it? trapshootinguytx: crotch scarf? b0rderline_fo0l: yes, haven\'t you heard of those? trapshootinguytx: no i havent trapshootinguytx: can  u locate one for me? b0rderline_fo0l: i can have my mom make you one trapshootinguytx: does it cover the nuts too?  lol trapshootinguytx: does she need measurements? trapshootinguytx: lol b0rderline_fo0l: that would be the crotch, yes trapshootinguytx: ok, goood b0rderline_fo0l: no, they are pretty much one size fits all trapshootinguytx: ok, does she make these a lot? b0rderline_fo0l: she will now trapshootinguytx: lol trapshootinguytx: call her now b0rderline_fo0l: it\'s gonna be a revolution trapshootinguytx: lol b0rderline_fo0l: ok, what color would you like? trapshootinguytx: hmm trapshootinguytx: green b0rderline_fo0l: ok, i\'m on it   '),(22919,2061087,'28,January,2004','2004-01-28 00:00:00','   Laundromats suck.  Big time.  That is all.   '),(22920,2061087,'27,January,2004','2004-01-27 00:00:00','   Well, ok, not really.  My (politician thank ya very much) buddy Jay Dee Gee sent me this cool link here.  If you are undecided about who to vote for, or if you just want to confirm that you are voting for the right candidate go  urlLink to this here site  and take the quiz.  The results are interesting to say the least and might surprise you.   '),(22921,2061087,'27,January,2004','2004-01-27 00:00:00','   For anyone searching for \"like peas and carrots\" on Google or Yahoo.  I was #2 on DogPile.  But, of course, who is it that searches for that?   '),(22922,2061087,'27,January,2004','2004-01-27 00:00:00','   My buddy  urlLink Josh has a new blog .  Go check him out, I just gave him a new look.   '),(22923,2061087,'27,January,2004','2004-01-27 00:00:00','   urlLink This guy  is getting my vote.   '),(22924,2061087,'23,January,2004','2004-01-23 00:00:00','   In his final years, my grandfather suffered from many strokes, dementia, and extreme hard-of-hearingness (yes, I like to make words up).  In one of his many visits to the hospital while he was still pretty much with it, yet exceedingly hard of hearing, the nurse was telling him what they were going to be doing to him.  She was going down the list, and he was sitting there nodding.  She asked him if he had ever had a catheter before and he suddenly became very indignant.  \"I am NOT a cat hater!\", he tells her in a very offended tone of voice.      Cat hater.  hahaha   '),(22925,2061087,'23,January,2004','2004-01-23 00:00:00','   Places I\'ve visited or passed through (had to be awake and set foot in) in bold, places I\'ve lived at least three contingent months in italics and bold.  1) Alabama 2) Alaska 3)  Arizona  4) Arkansas 5) California 6) Colorado 7) Connecticut 8) Delaware  9) Florida 10) Georgia 11) Hawaii 12) Idaho 13) Illinois 14) Indiana 15) Iowa 16)  Kansas  17) Kentucky 18)  Louisiana  19) Maine 20) Maryland 21) Massachusetts 22) Michigan 23) Minnesota 24)  Mississippi  25)  Missouri  26) Montana 27) Nebraska 28)  Nevada  29) New Hampshire 30) New Jersey 31)  New Mexico  32) New York  33) North Carolina 34) North Dakota 35) Ohio 36)  Oklahoma!  37) Oregon 38) Pennsylvania  39) Rhode Island 40) South Carolina 41) South Dakota 42) Tennessee 43)   Texas    44) Utah 45) Vermont 46) Virginia 47) Washington 48) West Virginia 49) Wisconsin 50) Wyoming and 51) DC   Additionally,  well, I have no additionally since I\'ve never been abroad.  Maybe someday.  As always put yours on your blog, in the comments, or email it to me.  Come on people, I tell you all about me.. what about you?   Snagged from  urlLink Skillzy  via  urlLink surgarmama .   '),(22926,2061087,'23,January,2004','2004-01-23 00:00:00','   At this moment, what is your favorite...  1. ...song?   That\'s a tough one, I like so many.  Lately the one going through my head is  urlLink Hit That .  Love me some Offspring.  2. ...food? Ahh food.  Another toughie.  I like it all.  Today I am lovin me some salad blend with Butter Lettuce and Radicchio with BBQ Ranch dressing.    3. ...tv show? Any Day Now, but I don\'t ever get to watch it.  I wonder if they have it on DVD yet?  My fave I do get to watch is ER but I have been less than impressed with it lately.  4. ...scent? Snuggle Botanical Bliss fab softener.  And Mr. Stopable\'s special blend cologne.  Smells like candy.  (why do I have that damn song in my head now, yeah, you know the one)  5. ...quote?  As you journey through life take a minute every now and then to give a thought for the other fellow. He could be plotting something.  Hagar the Horrible  ha!   '),(22927,2061087,'20,January,2004','2004-01-20 00:00:00','   Today over at Mac\'s she  urlLink wrote about something  that brought back a disgusting memory for me.  Go there and read her post and check the comments for what happened to me regarding that.   '),(22928,2061087,'20,January,2004','2004-01-20 00:00:00','   I don\'t know how or why I fell out of touch with  urlLink Sister Hazel.   I remember them from way back, and loved their stuff.  Cruising around on All Blogged Up the other day she had some lyrics and a link at the end of one of her posts.  I liked the one she had thrown up there (Your Mistake), and made a mental note to check them out.  I\'m glad I finally remembered.  These guys are so great.  They have improved much over the years that I haven\'t been listening to them, and there are several songs on their latest album,  urlLink Chasing Daylight , that speak to me.  Most notably Life Got in the Way, Killing Me Too, Effortlessly, and Can\'t Believe.  Pretty decent sorta southern alternative rock. (LOL)  Check them out, you won\'t be sorry.   '),(22929,2061087,'20,January,2004','2004-01-20 00:00:00','   In my previous post, I made a statement that isn\'t entirely true.  I do have friends.  Not many, but I do.  I have become very selective about those who I surround myself with, to the point that I am not surrounded.  So, thank you to those friends for being the kind of people who I can let in.  I have been doing some thinking today.  And some talking.  One of my friends will not simply let me \"be\".  He doesn\'t do it on purpose.  (Well, sometimes he does.)  He is the kind of person who makes me want to be a better me.  Some of y\'all know that trying to get me to open up about the really relevant stuff is like trying to do dentistry on a bear.  I have become a represser and suppresser in the interest of self-preservation, because I have simply become too tired for all of the angst and self-discovery and the gettinupoffofmyassanddoinsomethinaboutitall.  As I was sayin to him, I am at another one of those crossroads you inevitably come to in life.  But, due to the afore mentioned tiredness, I have just been sitting here park.  I am tired of going down the wrong road and having to either backtrack or detour.  I am tired of going along, minding my own business, and being broad-sided.  I am tired of coming to the construction, the traffic jams, and the idiot motorists who litter life\'s highway.  Trite analogy, but it\'s all I can come up with at the moment.  I have been shifting from park to neutral here lately, trying to get revved up to go somewhere.  I know I need to.  I keep running across people and words who make me think of who I could and should be.  This morning over at  urlLink Jett\'s , this post made me sit and ponder for a bit, and light a little fire of inspiration in my clogged up carburetor.  And from there, I was led to  urlLink this here post.   That sums up a lot of my feelings.  She says it much better than I can at the moment.  I need a vacation, y\'all.  And I need that sit-down lunch we been talkin about, Jay Dee Gee.  And I need to get in gear.   '),(22930,2061087,'20,January,2004','2004-01-20 00:00:00','   For sure.  I lived in one place for most of my life.  I was born there, raised there, and swore I\'d always get out of there.  I\'ve moved away 4 times, and for some reason or another I always ended up getting sucked back there.  This time I\'ve sworn I\'d rather live on the street in a cardboard box than move back there again.  So, why is it today I wish that I was back there?  Ahh.. homesickness.  I think that mostly it is that I am not happy where I am right now.  Oh sure, I have a nice apt. and live in a decent enough town.  But, it isn\'t going for me right now.  I have not been able to find a job, and I have not been able to find any friends.  I do not have any family in this town (save my ex husband and that doesn\'t really count).  I\'m lonely.  I want somewhere that I can go visit if I am tired of my place.  I want friends who drop by and hang out.  I want my family close so that when I need them, or they need me, it\'s not an hour drive to get there.  I want to live back in the country where it is peaceful and there are no neighbors and my kids can go outside and play and I don\'t have to watch them like a hawk.  I could go on and on with the good stuff, but then I\'d have to start listing the bad stuff.  Not that I don\'t have bad stuff here.  My problem isn\'t all situational.  The usually rational adult me knows that.  The scared little girl (who I never admit is still there) me doesn\'t care, she just wants to go home.   '),(22931,2061087,'18,January,2004','2004-01-18 00:00:00','   i can still smell your scent in my bed and though it is fading it reminds me of you holding me holding you and i hate that i am holding this pillow  that smells of you instead and that your presence is  all too familiar to me already when last night  could have been just that   '),(22932,2061087,'17,January,2004','2004-01-17 00:00:00','   What is the middle name of the first person you ever slept with?  Haven\'t a clue.  What was their first name?     What kind of underwear are you wearing and what color?  Victoria\'s Secret high cut briefs, kind of a peachy/beige snakeskin pattern.   What is the song you want played at your funeral?  Amazing Grace.   What would your last meal be before getting executed?  Beans of Eternal Life   Beatles or Stones?  Stones   If you had to pick one person on earth who should die, who would it be?  I don\'t want that kind of karma hanging over my shoulder.     The person whose problems you would never want to hear again?  Mine.   What is the thing most important to you (as far as physical) about the preferred sex?  A pulse.   Do you secretly hate some of your friendsters but are too nice to reject them?  I haven\'t been on friendster since I registered, I doubt I have any friendsters   If you could have any super power what would it be?  The ability to heal by touch.   Favorite hangover cure?  Don\'t drink in the first place?  I\'ve never found a sure-fire hangover cure, so I don\'t drink anymore.   How many drinks does it take to get you drunk?  Since I don\'t drink anymore, I\'d be a pretty cheap drunk.  Back in the day, I had to hit the 10 or 12 mark to start feeling it.   Favorite Outkast lyric?  Everybody needs someone to rub their shoulders And scratch their dandruff And everybody need to quit actin hard and shit Before you get your ass whooped (I\'ll slap the fuck out ya!) And everybody needs somebody to love Before its too late Its too late   Hair color you most like someone you\'re dating to have?  Hair color isn\'t important.  Hell, hair is optional.    If you had to be blind or deaf?  Deaf.  I\'d hate to not be able to see my children grow up.     Do you have any psychiatric problems?  Depends on your version of \"problems\".   Siblings that should go to rehab?  Health rehab maybe.   Least favorite month?  February always sucks.   First movie you can remember seeing as a kid?  Jaws.  It was like 1975 and I was 4 and I remember it to this day.  I still have a problem with boats and water I can\'t see in.   Favorite person in the whole world?  That would have to be me.  No, really, I don\'t do favorites much.  I don\'t have a favorite color or song or anything like that.  I like variety.   When\'s the last time you went on a date?  When the Earth was still flat.   Do you like violent movies or dirty movies?  Yes.   Fall or spring?  Both.   Person you most wish you hadn\'t made out with?  I don\'t see the point in having that kind of regrets.   If you are straight, what person of the same sex would you do it with?  If I am straight I wouldn\'t be doing it with a person of the same sex.   Where do you want to live when you are old and brittle?  Somewhere with a great climate.  Not too hot, not too cold.   Who is the person you can count on most?  I don\'t know, who?  There are different people I can count on for different things.   If you could date any celebrity past or present, time and age are not factors?  Jesse James   What books have you pretended you\'ve read?  I\'ve read them all.   What\'s a word you would use to describe your life?  Conflicted.   Favorite drinking game?  The \"try not to dribble it down my chin\" game.   What did you dream last night?  I didn\'t dream last night, I was too warm and snuggly.   Favorite bands?  Rubber.   '),(22933,2061087,'17,January,2004','2004-01-17 00:00:00','   The comment function is back up and running now.  Not that anyone uses it.  (hint hint)   '),(22934,2061087,'13,January,2004','2004-01-13 00:00:00','   You are The Cap\'n!  Some men are born great, some achieve greatness and some slit the throats of any man that stands between them and the mantle of power. You never met a man you couldn\'t eviscerate. Not that mindless violence is the only avenue open to you - but why take an avenue when you have complete freeway access? You are the definitive Man of Action. You are James Bond in a blousy shirt and drawstring-fly pants. Your swash was buckled long ago and you have never been so sure of anything in your life as in your ability to bend everyone to your will. You will call anyone out and cut off their head if they show any sign of taking you on or backing down. You cannot be saddled with tedious underlings, but if one of your lieutenants shows an overly developed sense of ambition he may find more suitable accommodations in Davy Jones\' locker. That is, of course, IF you notice him. You tend to be self absorbed - a weakness that may keep you from seeing enemies where they are and imagining them where they are not.    urlLink   What\'s Yer Inner Pirate?     brought to you by  urlLink The Official Talk Like A Pirate Web Site.  Arrrrr!      Have I mentioned that my lastest fascination is  urlLink Pirates of the Caribbean ?  I almost didn\'t watch it because of Johnny Depp.  In fact, I didn\'t watch it until this month.  Johnny D.  has kind of put me off the last few years.  The whole downhill slide after Edward Scissorhands and What\'s Eating Gilbert Grape.  But, Orlando Bloom won out.  Mmmm.. elf or pirate... I\'d watch him anyday.  So, color me shocked when my favorite in the movie was Captain Jack Sparrow.  That part was made for him, or he made that part, either way.... I\'m hooked.  Aye!  Avast!   '),(22935,2061087,'11,January,2004','2004-01-11 00:00:00','   The World Is MINE! by Demonac      Name:           You will conquer:    Marguerita-Ville (now you get royalties every time they play that song).      Your title will be:    Emperor/Emperess      You will succeed by:    Sending back a computer virus from the future to eliminate the opposition.      Your Enforcers will be:    Jay and Silent Bob.      Your first act as ruler:    Bathe in your victims&#39; blood.             Created with  urlLink   quill18  \'s  urlLink  MemeGen  !   '),(22936,2061087,'10,January,2004','2004-01-10 00:00:00','   But do you really?   urlLink Go here to find out how well you know me!    (as seen at  urlLink Hunter\'s  and  urlLink Mac\'s .)   '),(22937,2061087,'28,February,2004','2004-02-28 00:00:00','   I received an email today from a lady who had read a comment I had written on another blog.  It\'s kind of cool that you can either help or inspire or reassure people who you have never met and never will meet with just a few words graffitied on the internet.  I\'m glad she took the time to send me an email.  I vaguely remembered the post she was talking about, so I searched until I found it.   urlLink This is the post on moodymama  that she read and my comments are farther down.  Take a minute to read it or you won\'t understand the rest of what I am about to say. Done? Ok. I spent years in hell.  YEARS.  In HELL.  I thought I was crazy.  Thought I should be locked up somewhere.  And it just got worse and worse and worse.  When I was pregnant with Willa and the year after I had her, I was seriously considering Red River.  I had been bad in years past.  Breaking things, raging and raging.  Hurting myself mostly.  Physically anyway.  Emotionally myself and others.  Now, I knew a lot of it was due to my hormones, because the worst of it would hit before my period, it was bad during and some after.  It was a vicious cycle.  The week before my period I was like a land mine.  Set and ready to go off and who knew when or on who it would go off.  I was temperamental and irritable and hostile.  There were times when I literally saw red.  It was like the \"dragon\" had possessed me and I could feel the rage and hate overtaking me until I wasn\'t there anymore.  Then, when I started my period... I was depressed.  Crying constantly, getting my feelings hurt over and over for nothing.  Then the week after my period, that was the worst.  Suicidal, feelings of worthlessness, why should I live I am a horrible person I wish I would just die please God just take me now and get it over with.  I dreamed of dying, killing myself, being killed in some random accident.  Fantasized about it.  And this went on for years.  From the time I was 13 or 14 until I was 31 years old, progressively getting worse and worse.  I had times when it wasn\'t so bad.  But that could all change in a moment.  It was always there, lurking.  Since I\'ve had the hysterectomy I have been a different person.  Oh sure, I have my moods.  But that\'s life.  Now I have a handle on them and can discern between real emotion and the irrational, out-of-control monster that had overtaken my emotions.  Stress is gonna happen.  But when it is coming from inside of you... what do you do?  Especially when you can\'t control it.  I had taken pills upon pills upon pills to no avail.  If it had gone on the way it was headed, who knows where I would be now.  And I had pretty much forgotten all of that.  Well, not forgotten but pushed to the back of the memory vault.  Who wants to remember horrible stuff?  So, I am glad that woman read my words and took a moment to write me about them.  It was a great reminder.  Because yeah... things could be worse.   '),(22938,2061087,'27,February,2004','2004-02-27 00:00:00','   1. What cologne or perfume do you wear? The last two perfumes I wore were Wish by Chopard and Sentiment by Escada.  I generally don\'t wear the same perfume for more than one bottle.  I either get tired of it or just want a change.  Right now I want some Amor Amor by Cacharel or Intense by Hugo Boss.  Anyone want to indulge me?    2. What cologne or perfume do you like best on the opposite sex?   Herrera by Carolina Herrera, anything Hugo Boss.   3. What one smell can you not stomach? My mom\'s perfume, that Pheremone crap.    4. What smell do you like that others might consider weird? Gasoline, mechanic shops.    5. How do you plan to spend your weekend? Probably the same way I spend most weekends, in extreme boredom.  Anyone want to unbore me?   '),(22939,2061087,'26,February,2004','2004-02-26 00:00:00','   So, I started another blog today.  Probably won\'t be of any interest to y\'all because it is just going to be a giant listing of all the books I read, what I think about them and whatnot.  But, if anyone is interested  urlLink it\'s over here .   '),(22940,2061087,'22,February,2004','2004-02-22 00:00:00','   Santa Maria, Santa Teresa, Santa Anna, Santa Susannah  Santa Cecilia, Santa Copelia, Santa Domenica, Mary Angelica  Frater Achad, Frater Pietro, Julianus, Petronilla  Santa, Santos, Miroslaw, Vladimir  and all the rest   a man is placed upon the steps, a baby cries  and high above the church bells start to ring  and as the heaviness the body oh the heaviness settles in  somewhere you can hear a mother sing   then it\'s one foot then the other as you step out onto the road  how much weight? how much weight?  then it\'s how long? and how far?  and how many times before it\'s too late?   calling all angels  calling all angels  walk me through this one  don\'t leave me alone  calling all angels  calling all angels  we\'re cryin\' and we\'re hurtin\'  and we\'re not sure why...   and every day you gaze upon the sunset  with such love and intensity  it\'s almost...it\'s almost as if  if you could only crack the code  then you\'d finally understand what this all means   but if you could...do you think you would  trade in all the pain and suffering?  ah, but then you\'d miss  the beauty of the light upon this earth  and the sweetness of the leaving   calling all angels  calling all angels  walk me through this one  don\'t leave me alone  callin\' all angels  callin\' all angels  we\'re tryin\'  we\'re hopin\'  we\'re hurtin\'  we\'re lovin\'  we\'re cryin\'  we\'re callin\'  \'cause we\'re not sure how this goes  Artist: Jane Siberry with k.d.lang Lyrics  This was the first song they played at my father-in-law\'s funeral.  I hadn\'t heard it in forever and it was very fitting.  The other one they played was Amazing Grace on the bagpipes.  He always said that\'s what he wanted played at his funeral.  Little did we know it would be so soon.  Rest in peace, Grand-dad.   '),(22941,2061087,'22,February,2004','2004-02-22 00:00:00','   Today has been a sad day.  My father-in-law lost his battle with cancer.  He fought long and hard, but in the end it wasn\'t enough.  Our only consolation is that he is in a better place now.  But, it isn\'t much when he has left behind so many who loved him at such an early age.  It\'s hard to understand why God does the things he does.  That\'s always been a major problem with me and religion.  Why does He take a good person when there are so many bad ones wasting space on this Earth?  And it\'s hard to try to explain why Grand-dad isn\'t with us anymore to my children.  They don\'t know or care that he is in a better place and not sick and in pain anymore, they just know that they aren\'t ever going to get to see him or play with him again.  They will understand someday, but it doesn\'t make it any easier NOW.  For anyone of us really.  His funeral is Tuesday and they want to go.  It is going to be a hard day emotionally.  But all we can do is get through it and try to understand.  Life goes on.  It will have a Grand-dad sized hole in it, but it does go on.   '),(22942,2061087,'20,February,2004','2004-02-20 00:00:00','   I don\'t know what it is about me that makes sweet young boys want to bring me coffee.  Maybe they think that is the way to an old woman\'s heart.  More likely into her panties.  (Although why they would want into an old woman\'s panties is beyond me.)  I wish that I could take them up on all the coffee offers, without the strings, just for the yummy coffee.  Alas, the only sweet young boy that I want to bring me coffee is not speaking to me, so I don\'t hold out hopes of any more of his lovely, string-free coffee deliveries.  I miss them.   '),(22943,2061087,'19,February,2004','2004-02-19 00:00:00','   When was the last time you...  1. ...went to the doctor? Sometime in December  2. ...went to the dentist? Yeesh, November or December of 2002.    3. ...filled your gas tank? Day before yesterday  4. ...got enough sleep? A few weeks  5. ...backed up your computer?  That would be never.   '),(22944,2061087,'19,February,2004','2004-02-19 00:00:00','   you are the \"you\'re so dumb\" happy bunny. you are brutal in your words and enjoy putting others down.     urlLink   which happy bunny are you?     brought to you by  urlLink Quizilla   '),(22945,2061087,'17,February,2004','2004-02-17 00:00:00','   Que bonito fue, fuimos un par  y llenos de amor, un momento allÃ¡   No existÃ­a el bien, ni tampoco el mal,  nada mÃ¡s tÃº y yo, y un momento allÃ¡   DÃ­as sin fin, noches tambiÃ©n,  amor asÃ­, nomÃ¡s uno en la vida  EntregarÃ© todo mi ser  por un momento allÃ¡   Te siento temblar, Â¿Es porque te vas?  Â¿O piensas que darte un momento mÃ¡s?   La luz es amor, no se acabarÃ¡  Longramos brillar un momento allÃ¡   DÃ­as sin fin, noches tambiÃ©n,  amor asÃ­, nomÃ¡s uno en la vida  EntregarÃ© todo mi ser  por un momento allÃ¡     urlLink (Translation)   '),(22946,2061087,'17,February,2004','2004-02-17 00:00:00','   I find myself greatly unsettled lately.  Have you ever had the feeling that you are forgetting to do something, or missing something, but for the life of you you can\'t figure out what it is?  I feel that way all the time.  I know that it is because I am not happy with the way my life it, yet can\'t seem to figure out what it is I need to do about it.  Circumstances prevent me from doing many things.  That adds to the stresses I have.  I feel like I am in a giant maze and I KNOW that there are goodies stashed all around, but I keep taking the wrong paths to get to them.  And when I find a goody, it either isn\'t what I want or some unseen force comes and takes it away from me.  The hardest part is the loneliness.  I have a feeling I could get some of the things in my head straight if I wasn\'t just so damn consumed with being lonely.  I\'ve tried to overcome that, but it\'s powerful.  And, I\'ve been alone so long that I think I not only grasp at the least little kindness, but I am very selfish about it as well.  Of course, a lot of being alone is my fault.  I have let the past cloud my judgement in some issues and old habits are hard to break, especially when you don\'t realize you are doing it.  So, between not finding anyone worthy enough, finding someone worthy and pushing them away, and finding someone worthy and it just isn\'t meant to be ... well, that leaves me still alone.  I often wonder why some things in life are so hard, and if maybe I make them harder than they really are.  I don\'t know anymore.  Maybe I will someday.   '),(22947,2061087,'17,February,2004','2004-02-17 00:00:00','   Since I had my gallbladder surgery, I\'ve been trying to eat low and no fat foods.  They can be a challenge to cook sometimes.  Well, they are easy to cook, the challenge is making them palatable.  I am getting better.  But, sometimes you just gotta eat REAL food.  I love to cook, but it\'s mostly comfort food and old favorites.  I\'ve been thinking about King Ranch Chicken for a while now and today I just had to make it.  Pretty damn good if I do say so myself.  Want some?   '),(22948,2061087,'16,February,2004','2004-02-16 00:00:00','   (Great for getting your schwerve on.)   I\'m so tired, of playing Playing with this bow and arrow Gonna give my heart away Leave it to the other girls to play For I\'ve been a temptress too long Just. . Give me a reason to love you Give me a reason to be, a woman I just wanna be a woman From this time, unchained We\'re all looking at a different picture Thru this new frame of mind A thousand flowers could bloom Move over, and give us some room  So don\'t you stop, being a man Just take a little look from our side when you can Sow a little tenderness No matter if you cry Give me a reason to love you Give me a reason to be ee, a woman Its all I wanna be is all woman For this is the beginning of forever and ever Its time to move over   '),(22949,2061087,'16,February,2004','2004-02-16 00:00:00','   I now have a washer.  One that works!  For free!  I have the best Mommy and Daddy and Clyde in the world, too.   '),(22950,2061087,'16,February,2004','2004-02-16 00:00:00','   Tonight was a test of the emergency back-up system.  You failed miserably.  Had there been an actual emergency the police would have been on your doorstep.  When I call your house repeatedly and frantically and leave messages for you to call me NOW, I mean now.  I don\'t mean whenever you goddamned well get around to it.  Count your blessings that it turned out to be NOT an actual emergency.  This is your final warning.    (yeah yeah, I know he isn\'t gonna read this, and when I tell him this in person I\'ll do it in a nicer way.  I\'m venting, y\'all should be used to that.)   '),(22951,2061087,'16,February,2004','2004-02-16 00:00:00','   What is worse than not having a washing machine?  Having puking children and no washing machine.   '),(22952,2061087,'16,February,2004','2004-02-16 00:00:00','   Ok.  So, I have always said that if I am wrong I will be the first one to admit it.  It happens so rarely that it\'s not something that I am used to or that comes easy to me.  But, here it is. I was wrong.  It was wrong of me to stereotype you.  It was wrong of me to think that you were like all of the other assholes I have had the misfortune to be friends with.  And above all, it was wrong of me to give you attitude.  That is something that I need to work on.  I will also admit that.  I have became far too bitter and cynical and have found it too easy to jump the gun and get my feelings hurt over something that shouldn\'t be such an issue.  I think I have become somewhat self-absorbed and forget to think that there may be things going on in other peoples lives as well.  In my head it\'s alway about rejection.  That\'s what it always has been in the past, and again with the stereotyping.  I have learned my lesson.  I knew when we entered this relationship that things would be different.  That due to circumstances it wouldn\'t be the usual boy/girl thing.  And you aren\'t the usual boy.  I try not to be the usual girl.  So, this is the public apology for anything posted here in a snit.  I\'m sorry and I hope that you can forgive me.   '),(22953,2061087,'15,February,2004','2004-02-15 00:00:00','   hug from behind - you like to feel what the other person is feeling and see things how they see them. you tend to be serious and emotional.    urlLink   What Sign of Affection Are You?     brought to you by  urlLink Quizilla   '),(22954,2061087,'15,February,2004','2004-02-15 00:00:00','   Words spoken harshly, can cut like a knife, sear like the flame, and crush like an avalanche of stone.  The empty ache caused can hurt for an eternity.  The bitter pain can be unbearable.  The endless lonely days and nights can cause those words to echo in the hollowness of your heart and mind. Then, someone comes along with a few words of tenderness, hope, support.  The hollowness if filled up like sunshine filling a dark cavern.  Those words can give you a new reason to fight your battles and go on with life.  A few words of inspiration can lift the spirits like a breeze blowing through the leaves. It is my belief that words - though spoken in haste or carefully thought out - should not be spoken before the consequences are reviewed. After all, what good are words of support or consolation if they are not really meant.   (circa 1989)  Talk  I remember when I called you,  Just wanting to talk. You came right over. Was talking all you wanted Or was something else on your mind? Sometime in the next few hours  My need to talk turned Into a need for something else, And then there was no need to talk. But, now I want to talk again And I don\'t have the nerve to call you because I don\'t know if you\'ll come this time. Do you want to talk? (circa 1987)  Crying in the Night  I lay in bed at night Crying myself to sleep Knowing it isn\'t right. I never knew my feelings for you Could ever run so deep. All those nights we spent together In each others arms. I thought that I could trust you, That you would do me nor harm. But there were others, too. And I soon realized  That I could not keep you All to myself. But, that doesn\'t keep me from dreaming, Crying in the night. And now in my daily plight I hope to have you And have you without a fight. Soon there will be no pain, No crying in the night. (circa 1985 or 6)  God, I was stupid.  I don\'t guess much has changed.   '),(22955,2061087,'13,February,2004','2004-02-13 00:00:00','   Holy Mother of God,  urlLink this shit  is funny.  I am still laughing.   Sure, sheâ€™s good-lookinâ€™. Sheâ€™s also crazy. Crazy as a shithouse rat. Run for your life.   This should be my new tagline.  hahahaha   '),(22956,2061087,'13,February,2004','2004-02-13 00:00:00','   I guess I met the devil, but I sure didn\'t know no better  You were kool as hell like e-mail, but still timeless like a letter  As I sit and I stare at the satanic glare, on the glass frame in front of your face  You\'re alone on my shelf, yelling, \"look at yourself\"  I feel like Bobby Brady breaking the vase   But now you\'re all screwed up   (chorus)  You\'re so yesterday  Miles away  Promised myself on new years day  I\'d take a bath today  And wash you away  As all of your little blonde hairs go down the drain   Your sister called me yesterday to tell me I was a loser  At least I haven\'t lost my mind, and at least I\'m not a boozer  As I tried to heed to your wants and your needs  You were solemnly lost in space  So keep reading your books on \"how to give dirtly looks\"  Everytime I should be put in my place   Now you\'re all screwed up   (chorus)   I guess I met the devil, but I sure didn\'t know no better  You were cool as hell like e-mail, but still timeless like a letter   (chorus)   '),(22957,2061087,'13,February,2004','2004-02-13 00:00:00','   Days swiftly come and go. I\'m dreaming of her She\'s seeing other guys Emotions they stir The sun is gone. The nights are long And I am left while the tears fall.  Did you think that I would cry, on the phone? Do you know what it feels like, being alone? I\'ll find someone new  Swing, Swing, Swing from the tangles of My heart is crushed by a former love Can you help me find a way To carry on again.  Wish cast into the sky I\'m moving on Sweet beginnings do arise She knows I was wrong The notes are old, They bend, they fold and so do I to a new love.  Bury me (you thought your problems were gone) Carry me (away. away, away...)   '),(22958,2061087,'13,February,2004','2004-02-13 00:00:00','   1. Are you superstitious? Yes.  2. What extremes have you heard of someone going to in the name of superstition? Stopping and turning around and driving miles out of the way to avoid crossing the black cat\'s path, locking oneself in the house and not going anywhere at all on Friday the 13th.  3. Believer or not, what\'s your favorite superstition? Ahh.. I like them all.    4. Do you believe in luck? If yes, do you have a lucky number/article of clothing/ritual? Yes, to a certain extent.  I don\'t really believe in the whole number/article/ritual thing though.   It\'s more being in the right time at the right place and things like that.  5. Do you believe in astrology? Why or why not?  Yes, I do, again to an extent.  I believe the pull of the tide and the moon etc do have certain effects on our bodies.  Of course, I get looked at like I am a raving maniac when I say that, or told that I am going to hell.   '),(22959,2061087,'11,February,2004','2004-02-11 00:00:00','   Today has been another one of those days.  We got the dryer to working yesterday.  I am quite pleased with it.  The washing machine is from hell and I am having Satan\'s minions come and get it in the morning as I am quite tired of dealing with it.  The \"repair man\" (and I use the term loosely) has been here three different times, and the washer still leaks.  So, I will get another one and keep my fingers crossed that it works properly.  Actually, I am going to see if I can get my Mamaw\'s old washing machine from back home here and see if it works and if not what needs to be done to get it that way.  It would be cheaper than buying another one and it worked fine when it was put into storage.  It\'s just the logistics of moving it 60someodd miles.  Hopefully, worth it in the long run and not moved that far just to go into the junk heap.  Think positive!  I sometimes have a problem doing that.  Inherent pessimism and all that.  There is a saying that I try to keep in mind.  \"I was sad because I had no shoes until I saw a man who had no feet.\"  Generally, there is always someone who is in a worse situation than you are.  I learned a long time ago not to ask what else could go wrong.  And, it could always be worse.  It has been worse.  Not THE worst, but I\'ve had some pretty bad times.  But, all one can do it live through them and hopefully learn and become stronger for them.  Still, they can get you down when you don\'t ever have a break from them.  When one thing after another happens, and many things at once are happening.  I\'m gonna whine about it again.  I need a vacation.  I guess we could all use one though.  I am in a better mood than I was the other day.  Like I said, when I am in a pissy mood, EVERYTHING pisses me off.  Even stuff I am generally not going to get pissed about.  I have learned to have compassion.  To always give the benefit of the doubt.  But, I am also an inherent worrier.  No matter who it is.  If someone doesn\'t call, I fear the worst.  In a barditch somewhere, etc.  It\'s usually not the case, but it has been known to be.  So, there\'s that and the whole being considerate thing.  Anyways...  I ramble on.    urlLink Someone  just mentioned V-day.  I think I\'ll kick him.   '),(22960,2061087,'09,February,2004','2004-02-09 00:00:00','   Well.  Okay.  I have lost my mind.  I was over at  urlLink dooce  just now checking out the  urlLink goings on and pictures  and now I am sitting here crying (it was the bottom picture that did it).  I know, I know.  I have 3 kids that drive me crazy.  And I love each and every one of them.  And no, I really don\'t want more.  Well, I do in a way.  It saddens me that I don\'t have that choice any more.  That I won\'t ever be able to go through the whole process the way it should be done, with a loving partner and father there.  That I won\'t ever experience the wonders of a new baby again.   That I will miss out on some great guys because they haven\'t experienced that and can\'t with me so they will do it with someone else.  I debated long and hard about having a hysterectomy.  And I know it was the thing to do, but still... I have my moments.   '),(22961,2061087,'07,February,2004','2004-02-07 00:00:00','   Have you ever noticed how some songs have very stalker-like leanings?   My playlist at this moment:   urlLink Feel - Robbie Williams    urlLink What\'s Your Fantasy - Ludacris    urlLink Come To My Window - Melissa Etheridge    urlLink Call Me - Blondie    urlLink Hold On - Good Charlotte   '),(22962,2061087,'07,February,2004','2004-02-07 00:00:00','   I\'m a clumsy bitch.  Always have been.  Years ago, one of my cousins nicknamed me \"Tommy Girl\" one day after the 574,899th time I marked myself in some bizarre incident.  That day it was opening the door to my Grand Cherokee and proceeding to bloody my nose and black both my eyes in the process.  I can\'t tell you how many times I\'ve been wounded by car doors.  Fingers slammed in them, legs slammed in them, hands rolled up in windows, hair rolled up in windows, jabbing myself on the top corner (you know the one, the wicked sharp pointy part), smacking myself in the face/head/body with them, breaking a finger hanging a ring on them.  You name it.   And that\'s just car doors, all the other assorted injuries due to other inanimate objects would take many more posts.  Today my cousin and I had to go to Walmart and get some washer/dryer crap.  I jumped out of the pickup and some of his assorted floorboard ornamentation fell out.  I leaned down to pick it up and on rising up tried to remove my nose with my glasses by slamming them into the rearview mirror.  Yeah, it left a mark.   '),(22963,2061087,'07,February,2004','2004-02-07 00:00:00','   Does anyone not remember that  urlLink I think February sucks ?  There\'s a reason for that.  I hate Valentine\'s Day.  Why?  I guess because of the whole symbolism thing and all it represents.  Because I\'m always left out of that.  I have never had a date for Valentine\'s Day, no romantic gifts, no flowers, no candy, no jack shit.  NEVER.  Hell, I am 32 years old and I\'ve never had a proper \"date\"  much less a proper Valentine\'s Day.  So, yeah, I\'m a little disenchanted with Valentine\'s day and all of that.  If I hear any more about it I may commence another massacre.   '),(22964,2061087,'07,February,2004','2004-02-07 00:00:00','   I\'ve been on the verge of getting my mad on for a while now.  Well, no longer.  I am in a hella bad mood.  I\'m tired.  I don\'t feel the greatest.  I\'m broke.  There are people who suck.  I am sexually frustrated.  You name it.  I\'m sick of things not going the way they are supposed to.  I am sick of people not doing the things they are supposed to.  The next governmental road block I come across, I am probably going to blow up the goddamn building.  Or assassinate Gee Dubya.  (yeah, now the feebs and all those other governmental pantywaists will probably be watching me, fuck \'em)   After driving all over hell and half of Texas Friday and finally finding a reasonably priced washer and dryer (and of course not at the same place) I couldn\'t get the sonofabitches home.  Well, they are home now.  And the fucking washing machine doesn\'t work.  So, now I have to see if the guy who sold it to me is going to stand behind his guarantee that it worked or if I am going to have to go and yank his asshole over his head.  Does the dryer work?  I don\'t have a clue.  The plug-in is the wrong fucking kind and the maintenance man is out until God knows when. I had an appointment this morning at yet another agency (yes, governmental agency, the fuckers) and after sitting there waiting for over an hour (with a cranky 2 year old) I was finally able to provide them with all 97 pages of the information they required.  So, of course, I am expecting to hear back from them with an approval.  Is that what they called me with this afternoon?  Fucking hell no.  I now have to go BACK and apply with the other stupid fucking State agency that denied me in the first place and then take the denial back to THEM before they will approve me.  As if the first denial wasn\'t enough?  No, now I have to repeat the entire fucking process, and WAIT for God only knows how many more months before I can get the help I need to go to the Dr. to find out about the goddamned abdominal mass I have.  That\'s lovely.  Thanks, y\'all.  I hope it\'s not fucking cancer and y\'all are just dicking me around and letting it overtake my body before you help me so then you won\'t have to pay out as much because I\'ll be DEAD.   Some days I wonder if that isn\'t better.   (Oh, y\'all didn\'t know I have had my lovely psychotic breaks and flirtations with suicide? yeah)  At least that way my kids would have money for the things they need.  If their fucking sperm donors would step up and be real men I wouldn\'t have to worry about that.  That\'s what I get for messing with not-real men.  And still I\'m too fucking nice and don\'t nail their asses to the wall.  Why is that?  Well, I don\'t know, but I tell you, my ex-husband just used his last get out of shit free card.  He has a brandgoddamned new truck and I get shit for child support and he couldn\'t help me get my washer and dryer home.  And we won\'t even get off on the subject of Brad.  Useless fucker.  He should have to have a mandatory frontal lobotomy.  Yeah, I\'m just gonna touch on all the things that piss me off.  Why not?  It\'s my fucking rant.  (and yeah I\'m gonna use that word alot, if you don\'t like it, don\'t read it)  Fucking men piss me the hell off.  They really do.  That was one major reason for the whole celibacy thing.  But I blew that one all to hell, didn\'t I?  Not that I regret it for one second.  I\'m just pissed off right now so everything is fair game.  Those who know me know that I may rant and rave and throw things but don\'t take it personally, eh?  Well, sometimes it\'s personal.  If I\'m spouting off at the mouth it\'s a pretty fair indication that it\'s not too personal.  It\'s when I don\'t talk that you should wonder.  I\'m kind of passive-aggressive that way.  And when I\'m mad everything pisses me off.  (You know the other night when you asked if I was mad?  And I said irritated maybe, but not mad?  Well, the next time I\'ll be mad.  It\'s happened four times.  I can excuse a few, but when it becomes a habit it pisses me off.  Oh, who am I kidding... I wouldn\'t be pissed with you.  Hurt and irritated maybe.  I hate waiting for someone who isn\'t going to show up.  It makes me feel like I don\'t matter enough to a) show up b) call and say you aren\'t going to show up.  And why am I saying this here?  Sorry, got carried away.)  Now I\'ve lost track of where I was.  Oh yeah.  Everything pisses me off.  I seriously need a break.  I\'ve dealt with this shit for far too long and haven\'t had any real time for just me, no kids around.  And it\'s beginning to show.  But, there\'s that whole no money, nowhere to go thing.  It\'s a circle jerk.  So, while a weekend away would be nice I don\'t guess I\'ll be getting a break anytime soon.  Maybe some dirty, nasty, hot, sweaty, pisstheneighborsoff, rough sex?  Can I get that?  Because I\'d take that, too.   '),(22965,2061087,'07,February,2004','2004-02-07 00:00:00','   to not singing the Wash Woman Blues.  I bought a washer and dryer today.  Now... I just have to figure out how I am going to get them home short of strapping them to my back and carrying them.   '),(22966,2061087,'06,February,2004','2004-02-06 00:00:00','   1. What\'s the most daring thing you\'ve ever done? I guess that depends on your definition of daring.  Public nudity, sex in public, meeting people off of the internet, getting stoned in Six Flags countless things.    2. What one thing would you like to try that your mother/friend/significant other would never approve of? I think I\'ve pretty much done all of it.  Tattoos, the above listed stuff, piercing my eyebrow.  I can\'t think of much left that they wouldn\'t approve of that I haven\'t tried.    3. On a scale of 1-10, what\'s your risk factor? (1=never take risks, 10=it\'s a lifestyle) In the past... 10.  Now?  Probably 2 or 3.    4. What\'s the best thing that\'s ever happened to you as a result of being bold/risky? Seeing New Orleans for the first time in the early morning rain.  Meeting a truly awesome person.    5. ... and what\'s the worst? I can\'t think of any worsts.  I have escaped my adventures unscathed for the most part.  I try not to live my life regretting things that I have done.   '),(22967,2061087,'04,February,2004','2004-02-04 00:00:00','   You know the saying... when you reach the end of your rope, tie a knot and hang on?  Mine looks suspiciously like a noose.   '),(22968,2061087,'01,February,2004','2004-02-01 00:00:00','   Well, the Pats pulled it off in the last minute (literally).  And they weren\'t the only one pulling it off.  What the HAYULL was up with the  urlLink  Janet and Justin halftime show antics ?!?  Oh my Gawd!   The whole halftime thing has got to where it sucks so bad it\'s not even worth watching.  I guess they think that \"surprises\" like that will keep everyone interested.  Forget about the lip-synching and half-hearted performances.  And Beyonce really needs to stick to what she does best.  The national anthem it is not.  Aerosmith was great, as usual.  And Josh Groban... my oh my that boy can SING!  His was the best performance bar none.  Well, other than Tom Brady\'s.  The commercials weren\'t all they usually are either.  The AOL/Orange County choppers one did crack me up.  And the Bud Light dog trick one.  And I liked the Pepsi one, but why not just let people download the songs, why must they buy Pepsi to get them?  Oh yeah... marketing.  Too bad Pepsi sucks.  I\'ll stick with WinMX.  But, the main purpose of it all was the game, and it was an excellent one.  The Panthers did an awesome job and I thought they might have it there for a minute, but the Pats and their experience persevered.   '),(22969,2061087,'31,March,2004','2004-03-31 00:00:00','   I love  urlLink Clem Snide\'s  new song!!!  Wow.  I am hooked.  I already loved them, but this song rocks my socks off.  Scroll down to where it says Listen to All Green or whatever it says.   urlLink Or just click here.   These guys make my ears so happy!   '),(22970,2061087,'27,March,2004','2004-03-27 00:00:00','   Well, I am completely and totally unoriginal and still have nothing of interest to say so I will post this amusing little 9 layer thingy for non-comsumption.  LAYER ONE: -- Name: Cristina -- Birth date: July 14, 1971 -- Birthplace: Graham, Texas -- Current Location: Wichita Falls, Texas -- Eye Color: blue-grey -- Hair Color: dark brown -- Height: 5\'7\" -- Righty or Lefty: righty tighty -- Zodiac Sign: Cancer, can\'t you tell... since I\'m so crabby?  LAYER TWO: -- Your heritage: Scots, German, Irish, English, some Native American and who knows what else is in the woodpile. -- The shoes you wore today: Happy Bunny flip-flops -- Your weakness:  Threesomes with Ben & Jerry, books, and shiny magazines. -- Your fears:  Senility (dementia, Alzheimer\'s),  horrible disfiguring accident, knowing I am dying,  being alone for the rest of my life.   -- Your perfect pizza: Ahh Pizza Roma shrimp, mushroom and sun-dried tomato with white sauce. -- Goal you\'d like to achieve:  To just keep on keepin\' on  LAYER THREE: -- Your most overused phrase on AIM:  a\'ight, ahhh, and ahahahahahaha   -- Your first waking thoughts:  Fucking alarm clock.  Yeah, you too, you bastard phone.   -- Your best physical feature:  I\'m told my eyes and my ass.  I dunno. -- Your most missed memory:  I don\'t miss my memories, I still have them all.  Well, the ones I want to remember.  LAYER FOUR: -- Pepsi or Coke:  Dr. Pepper -- McDonald\'s or Burger King: Burger King if I really had to choose. -- Single or group dates: Either one, but I\'d prefer single. -- Adidas or Nike:  New Balance -- Lipton Ice Tea or Nestea: Lipton -- Chocolate or vanilla: Depends on what it is. -- Cappuccino or coffee:  Coffee or frappucchino  LAYER FIVE: -- Smoke:  Only when I\'m on fire. -- Cuss: Never.  What are you, fucking nuts? -- Sing:  Badly. -- Take a shower everyday:  Generally.  Sometimes twice a day.  I rarely skip a day.   -- Do you think you\'ve been in love: I wanna know what love is.  I\'ve been in versions of love. -- Want to go to college:  Not especially. -- Liked high school:  Not in the least.  I hated school from about 4th grade on. -- Want to get married:  Yes. -- Believe in yourself:  No, I\'m like Santa Claus and the Easter Bunny. -- Get motion sickness:  Usually only in the back seat.  Sometimes in the front, depending on who is driving. -- Think you\'re attractive:  I\'d do me. -- Think you\'re a health freak:  Only in the sense that my health is freakish. -- Get along with your parent(s):  Yeah, I do now. -- Like thunderstorms:  That\'s a resounding YES! -- Play an instrument:  Only if the skin flute counts.  LAYER SIX: In the past month... -- Drank alcohol:  Nope. -- Smoked:  Yep. -- Done a drug:  Prescription ones. -- Made Out: Yes. -- Gone on a date:  Nope. -- Gone to the mall?:  Nope. -- Eaten an entire box of Oreos?:  Nope (Not that they come in boxes.  Unless you\'re talking the ones at the grocery store before they unpack and shelve them.  Good God, no in that case.) -- Eaten sushi: I wish! -- Been on stage:  No. -- Been dumped:  Not in the general sense of dumped, but yeah, I think so -- Gone skating:  Hell no, i don\'t want to break a hip! -- Made homemade cookies:  I\'m not Betty Crocker. -- Dyed your hair:  I\'m a recovering dye-aholic. -- Stolen Anything:  Not that I am aware of.    LAYER SEVEN: Ever... -- Played a game that required removal of clothing:  Yeah, several. -- If so, was it mixed company:  Yep. -- Been trashed or extremely intoxicated: Dude, where\'s my car?!? -- Been caught \"doing something\": hehehe  many times -- Been called a tease:  Once years and years ago but I wasn\'t.   -- Gotten beaten up:  Sort of. -- Shoplifted:  Yeah, back in my way younger days. -- Changed who you were to fit in:  No, I never fit in anyway.  LAYER EIGHT: -- Age you hope to be married:  uhm.. 40?  I\'ll just have to give up and become a crazy old cat lady if I\'m not by then. -- Numbers and Names of Children:  3.  Jaron, Maddy, Willa. -- Describe your Dream Wedding:  Vegas, baby! -- How do you want to die: In my sleep like my Grandpa.  Not screaming like the passengers in his car.  (heh  NOT REALLY!) -- Where you want to go to college:  I don\'t. -- What do you want to be when you grow up:  I don\'t want to grow up. -- What country would you most like to visit:  Great Britian, New Zealand.  LAYER NINE: -- Number of drugs taken illegally:  So many I lost the ability to count. -- Number of people I could trust with my life:  I\'d like to think there\'s a few, but I\'m not a very trusting person. -- Number of CDs that I own:  Haven\'t a clue.  Not many.  -- Number of piercings:  7.  3 in one ear, 4 in the other.  Used to have my eyebrow pierced. -- Number of tattoos:  Two. -- Number of times my name has appeared in the newspaper?:  I dunno, a few.  More if you count the police blotter. -- Number of scars on my body:  1,2,3,4,5,6,7,8,9,10,11,12,13...hell.. too many to sit here and count.  I\'ll run out of fingers and toes. -- Number of things in my past that I regret:  I don\'t live my life with regrets.  I\'d be mired in a quicksand of remorse.   '),(22971,2061087,'22,March,2004','2004-03-22 00:00:00','   I know, I know.  I\'ve been very neglectful here.  But in my defense I was out of town for almost a week with no internet access.  Otherwise, I just haven\'t had much to say.  Well, I take that back.  I\'ve had plenty to say but y\'all really don\'t want to hear me bitch and moan and whine, now do ya?  Didn\'t think so.  It\'s pretty much a broken record anyways.  Kids driving me crazy, need a vacation (from them), preferably one involving a lot of sex (or just any sex anytime at all), my car sucks ass and is fallin\' apart on me, I am broke already, etc, etc, ad infinitum, ad nauseum.  So, unless any of y\'all can do anything about any of that... well, don\'t be bitchin\' cuz I\'m not postin\'.   '),(22972,2061087,'19,March,2004','2004-03-19 00:00:00','   (Yeah, yeah, I backdated it.  Shuddup.)  If you...  1. ...owned a restaurant, what kind of food would you serve? Hmm... if I were doing the cooking?  Homestyle definitely.  Lots of soups and sandwiches and chili and stuff.  I actually want a lunch counter/coffee bar/bookstore.  And I had a great idea over the weekend.  Why not have a laundromat with it all, too?  Anyone want to venture into business with me?  I think it would really work.  2. ...owned a small store, what kind of merchandise would you sell? Books, definitely books.  And maybe mixed media.  And coffee and stuff.  See above.  3. ...wrote a book, what genre would it be? Chick Lit probably.  4. ...ran a school, what would you teach? Oral Gratification  5. ...recorded an album, what kind of music would be on it? Eek!  Y\'all definitely don\'t want me singing.  Much less recording my singing.  But if I could sing I\'d want to be able to sing bluesy, torch music.  Something sexy and soulful.  Like Portishead maybe.   '),(22973,2061087,'12,March,2004','2004-03-12 00:00:00','   She may be the face I can\'t forget The trace of pleasure or regret Maybe my treasure or the price I have to pay She may be the song that summer sings May be the chill that autumn brings May be a hundred different things Within the measure of a day  She may be the beauty or the beast May be the famine or the feast May turn each day into a Heaven or a Hell She may be the mirror of my dreams A smile reflected in a stream She may not be what she may seem Inside her shell....  She, who always seems so happy in a crowd Whose eyes can be so private and so proud No one\'s allowed to see them when they cry She maybe the love that cannot hope to last May come to me from shadows in the past That I remember \'till the day I die  She maybe the reason I survive The why and wherefore I\'m alive The one I care for through the rough and ready years  Me, I\'ll take the laughter and her tears And make them all my souvenirs For where she goes I\'ve got to be The meaning of my life is She....She Oh, she....   '),(22974,2061087,'05,March,2004','2004-03-05 00:00:00','   1. What was the last song you heard? Still Frame by Trapt  2. What were the last two movies you saw? The Missing and Pieces of April  3. What were the last three things you purchased? Weell, I went grocery shopping a few hours ago so a bunch of things were purchased there.  Went and bought the kids some dress-up type rocks star clothes at the resale shops before that.  And bought some gas before that.    4. What four things do you need to do this weekend? Hmmm....  win the lotto.  Ha.  I SHOULD clean house and clean off my dresser in my bedroom and clean the kids toy boxes out.  But as far as NEED to?  Nothing I can think of other than get laid.  (Oh, oops, did I say that out loud?)  5. Who are the last five people you talked to? Matt, my mom, my kids   '),(22975,2061087,'05,March,2004','2004-03-05 00:00:00','   What was...  1. ...your first grade teacher\'s name? Nell Wheat  2. ...your favorite Saturday morning cartoon? Hey, Hey, Hey... It\'s Fat Albert!  And it wasn\'t a cartoon but I LOVED Captain Kangaroo.  3. ...the name of your very first best friend? I had two.  They were sisters.  Amy and Karen  4. ...your favorite breakfast cereal? Lucky Charms!  Still is.  They\'re magically delicious.    5. ...your favorite thing to do after school? Watch \"after school programs\" and later on the Dukes of Hazzard or play outside   '),(22976,2061087,'04,March,2004','2004-03-04 00:00:00','   She can kill with a smile She can wound with her eyes And she can ruin your faith with her casual lies And she only reveals what she wants you to see She hides like a child But she\'s always a woman to me  She can lead you to love She can take you or leave you She can ask for the truth But she\'ll never believe  And she\'ll take what you give her as long as it\'s free Yeah, she steals like a thief but she\'s always a woman to me  Oh, she takes care of herself She can wait if she wants She\'s ahead of her time Oh, and she never gives out And she never gives in She just changes her mind  And she\'ll promise you more Than the Garden of Eden Then she\'ll carelessly cut you And laugh while you\'re bleedin\' But she\'ll bring out the best And the worst you can be Blame it all on yourself Cause she\'s always a woman to me  Oh, she takes care of herself She can wait if she wants She\'s ahead of her time Oh, and she never gives out And she never gives in She just changes her mind  She is frequently kind And she\'s suddenly cruel But she can do as she pleases She\'s nobody\'s fool And she can\'t be convicted She\'s earned her degree And the most she will do Is throw shadows at you But she\'s always a woman to me   '),(22977,2061087,'03,March,2004','2004-03-03 00:00:00','   Why is it that some of us are wired to be night people?  And no matter how hard we try we can\'t reset our internal clock to daylight hours, no matter how necessary it is?  As far back as I can remember, and that\'s pretty far, I\'ve always been a night owl.  As a little kid, I would have to go to bed early for school and whatnot and I can remember laying there sleepless for HOURS.  And of course, come morning I wouldn\'t want to get up.  Some mornings I would go into the bathroom and lock the door as if I were getting ready for school and I would lay towels on the floor and go back to sleep for as long as I could get away with.  And in the summer and on weekends, we didn\'t have to go to bed (or if we were made to I never went to sleep) so I was always up.  Over the years I have tried and tried to do the early to bed, early to rise thing.  And I just can\'t get it to work for me.  What makes me mad is people who think that you are lazy or crazy or whatever for being that way.  To this day, my dad infuriates me because he just doesn\'t understand that not all people are like him.  Anyways, here it is almost 6 am and I haven\'t been to bed yet.  The last couple of weeks the insomnia thing has been really bad.  It doesn\'t help that I hate my bed being empty all the time.  How stupid is that?  That and when I do lay down a million and one things go through my head and I can\'t get any peace.  Maybe I need more coffee.   '),(22978,2061087,'03,March,2004','2004-03-03 00:00:00','   I feel like I should post something, but the only words that want to come out are not good ones.  I haven\'t been in the mood to talk lately.  Haven\'t been in the mood to do anything really.  I am feeling trapped and mentally overburdened and all I want to do is either crawl into a hole or rant and rave and kick things.  It\'s not really that I don\'t want to talk, I just don\'t want to try to have inane conversations.  Someone asks me what\'s up or how I am doing and I don\'t want to say I\'m fine and that nothing is up.  I want to say that I am slowly losing my mind and that everything is fucked up and that I hate my life.  But, who wants to hear that.  What does one say in reply to that?  I don\'t want your fucking platitudes.  I don\'t want \"I feel ya\" or \"I understand\" or \"I know how you feel\".  Because you know what?  You don\'t.  (don\'t take the you personally, I am speaking generally here but really, you don\'t)  You have no idea the fucked up extent of my existence.  You have no idea of the feelings that go on inside of me.  You have no idea what it is like to live my fucked up life.  Sure, you have stress too.  Sure, you get in bad moods too.  Sure, you have shitty things happen to you.  But you don\'t know how I feel and if you don\'t know what it is like to be me.  That may sound selfish, but there it is.  I don\'t let the real stuff out, mostly because people I don\'t want people to know.  I don\'t want pity, I don\'t want people thinking \"man, she really is messed up\".  It\'s true though.  I am.  But, I have learned that people don\'t want to hear it.  And if you do tell them... well, they never look at you the same again.  But, neither can I keep it all inside anymore.  It eats at me.  But, no one wants to hear it.  I\'ve been through this cycle before.  I know it will change, and things will be better.  It\'s hard to see the end of it though when you are so wrapped up in loneliness and despair and anger and sadness that you just don\'t want to go on anymore.  Most days I just want to go to sleep and never wake up.  It\'s not that I want to die.  It\'s just that I don\'t want to live this life anymore and I don\'t see a way out of it right now.  I\'ve tried, and I\'ve tried and well... I\'m tired of trying.  That too will change, but until it does this is how I feel.  I\'ve said it a million times, I need a break.  But, I don\'t seem to be going to get one.  I know I need to change my attitude, but that is easier said than done.  There are so many levels on which I am unhappy.  And discouraged.  I don\'t see them changing.  I don\'t see Mr. Right walking into my life.  I don\'t see the perfect job opening up.  I don\'t see winning the lottery.  I know these things are possible.  But the way things have gone for me my entire life...  I don\'t see them happening.  Yet, I am so damned lonely and miserable and hopeful and I want someone to hold me and say everything is going to be okay.  And mean it because they know the whole story and they care.  And well...  I don\'t know why I bother.  I should be used to it by now.  I rent movies and I don\'t even watch them because I want someone to watch them with.  I wish for friends, but I haven\'t the slightest idea of how to go about finding them.  That and over the years I\'ve learned that people suck and it\'s just not worth trying to be friends with them because you always end up alone in the end.  I don\'t know.  I hate feeling this way.  I hate wanting what I can\'t have.  And most of all I hate not knowing why the fuck I can\'t have it.  Am I so repulsive that people don\'t want to be with me?  What am I doing wrong?  I have yet to figure out why I can\'t keep a friend or a lover.  I know that most of the time it\'s because I ditch them because they are assholes.  And yes, I am an asshole sometimes too, but we all have our boundaries.  I don\'t cheat, I don\'t lie, I don\'t do bad things.  That\'s why I quit people.  Why do people quit me?  Maybe I expect too much.  All I know is I just want a chance.  That\'s all.  A chance.   '),(22979,2061087,'26,April,2004','2004-04-26 00:00:00','   New meme thing since I am far too lazy for any actual content.  Do It Yourself Would you be able to do these things yourself or have someone do it for you?  1. Fix a flat tire. 2. Fix a tear or make alterations to clothing. 3. Unclog a severely plugged up sink or toilet. 4. Change the oil in your car. 5. Put together an unassembled toy or piece of furniture out of the box.  Yes to all of the above.  And then some.  My family has made sure to teach me and my brothers along the way to be completely self-sufficient.  And I can do anything they can do (and I can do it better nah na nanny boo boo).  Well, some of it.  heh  But, yes, there really isn\'t much that I can\'t do for myself except for some of the highly technical/brute strength stuff.  I haven\'t had anyone around to take care of that kind of stuff for the most part so it is nice to be able to rely on myself.  It is/would be nice to have someone around to take care of the things I can\'t.  Fortunately, there are people I can call on for the stuff I can\'t take care of, but it is a rare occasion that I have to.  I really feel sorry for these people who aren\'t able to take care of the day to day stuff.  And women who act as if they aren\'t capable of doing such things annoy me to no end.  Yeah, it would be nice to have someone to take care of everything for us, but it isn\'t realistic for some of us.  A good relationship, in my opinion, blends each persons talents and they contribute what they are skilled at, helping each other and maybe learning something along the way.   '),(22980,2061087,'20,April,2004','2004-04-20 00:00:00','   So, I have been really bad about updating lately.  There\'s just been other things I\'d prefer to be doing for a change, rather than lack of anything to say.  I have plenty to say, but most of it I am hoarding, and am going to selfishly keep it all to myself so that I may take it out in private and hold it in my hand and admire it as a child might a secret treasure. But, I do have a niggling little problem.  And since I am unable to hold myself accountable, I am going to call upon you, dear readers and friends, to do that for me.  Accountability sucks at times, no?  Especially when you have an extreme lack of self-control and an overwhelming junk food habit.  Yes, that is what this is all about.  I still have 20 pounds to go.  No, 45 wasn\'t enough!  20 more, I say!  At least.  I would like to be able to wear a swimsuit this summer and not look as if I were a sausage stuffed into its casing.  So, help a sister out here.  I am going to be posting my weight here now, and then I will follow up once a week.  Feel free to cast aspersions and throw tomatoes if the numbers have not decreased.  (rotten tomatoes only please, lest I eat them).  Today\'s grand total is... are you ready for this?  Ok.  Here goes.  This is harder than I thought it would be.  Right.  I weigh... and keep in mind that I usually weigh in the morning so that it is less and it is now 2:30 and I have had food and all that and oh okay I\'ll quit dithering now.  I weigh 173???  Sheee-it.  A few days ago it was only 163.  Maybe we should try that again in the morning.  Anyways.  That\'s what the scale says as of 2:33 pm on April 20, 2004.  That\'s worse than I thought because that in essence means I need to get back down to 163 and then lose 20 pounds.  So...  30 it is.  Ugh.   '),(22981,2061087,'14,April,2004','2004-04-14 00:00:00','   Today was really craptacular  I got out of bed just before breakfast because the smell of cooking bacon woke me up.  I feel sad, because Sarah and Britney are complete bitches. They told everyone I have an STD, just because I slept with both of their boyfriends on Saturday night.  I\'m so happy. I just found out that I have been accepted into Harvard. And Yale. I don\'t know which to choose... oh, why is life so hard sometimes?  Last night I had to shave my entire body. Apparently, the lice that I caught from Amanda\'s friend are really hard to get rid of. I look quite strange with no hair and eyebrows. I\'d post pictures, but my webcam is broken.  I want to tell the world to get fucked.  I am updating this journal for the first time in ages, because I\'ve been in prison.  Today, I got a digital camera! Yes! Here\'s ten thousand photographs of my cat.  I want to say thanks to the world for absolutely fucking nothing! You all suck. I feel so alone, no one ever reads this journal, or even comments to let me know that I\'m not suffering alone. It\'s cold here, and I want to die, but I cannot figure out how many of you to take with me when I go.  I went to the doctor yesterday, and he said I have a terrible skin disease which prevents me from coming into contact with other human beings. And bipolar disorder.  You should all do this quiz! It\'s amazingly accurate. You just put in your name and birthday, and it will tell you you\'re a moron.  That\'s enough for now. But I\'ll leave you with some naked photos of myself. (Not safe for work - teehee).    Created with the  urlLink Gregor\'s Semi-Automatic LiveJournal Updater&trade; . Update your journal today! Powered by  urlLink Rum and Monkey   '),(22982,2061087,'09,April,2004','2004-04-09 00:00:00','   The  urlLink Friday Five  isn\'t relevant today, but I saw this meme at  urlLink jewdez  and  urlLink The Go Fish  and liked it better.    Grab the book nearest to you, turn to page 18, find line 4. Write down what it says:  \"Worth it, though!\" She stroked my face like she was petting a kitten.  \"It\'s a baby\'s bottom, is your face.\"   Stretch your left arm out as far as you can. What do you touch first?  The entertainment center   What is the last thing you watched on TV?  A PBS documentary about a fish   With the exception of the computer, what can you hear?  The TV, my kids playing outside, windchimes   When did you last step outside?  About 30 minutes ago   Before you came to this website, what did you look at?  My email   What are you wearing?  Shortie pajamas with bugs on them   Did you dream last night?  Not that I remember   When did you last laugh?  Last night   What is on the walls of the room you are in?  Paint   Seen anything weird lately?  Not weird enough for me to remember offhand   What do you think of this quiz?  Kind of random, random is good   What is the last movie you saw?  The Ring   If you became a multi-millionaire overnight, what would you buy first?  A car   Tell me something about you that I don\'t know.  I have psychic powers.   If you could change one thing about the world, regardless of guilt or politics, what would you do?  Every person would have a good home, a decent car and no worries about food or bills or lack of necessities.  And every person in the world would be provided healthcare.   Do you like to dance?  Yep  George Bush is he a power-crazy nut case or someone who is finally doing something that has needed to be done for years?  I don\'t think he is either.   Imagine your first child is a girl, what do you call her?  If it had been a girl, it would have been Sophia something.   Imagine your first child is a boy, what do you call him?  Don\'t have to imagine.  His name is Jaron.   Would you ever consider living abroad?  I consider it all the time, but it is more of a dream at the moment.   '),(22983,2061087,'25,May,2004','2004-05-25 00:00:00','   urlLink    Testing out BloggerBot&nbsp; urlLink   '),(22984,2061087,'24,May,2004','2004-05-24 00:00:00','   urlLink Crafty Ninja   '),(22985,2061087,'21,May,2004','2004-05-21 00:00:00','   Here lately I\'ve still been in the mood to just read other peoples\' journals and stay away from mine.  I haven\'t had many nice things to say lately.  The kids have been making me insane.  So, there has been a regime change here and they are grounded for the next 3 weeks.  I know, sounds harsh, but they have been hellions including at school and so far nothing has got their attention.  This will.  They will be grounded for 3 weeks and for every infraction there are days added to their \"sentence\".  They\'ll get the idea or they are in for a long and boring summer sitting in their rooms.  But hey, it\'ll be quiet around here!  I\'m highly pissed off at my ex due to his not having sent my child support yet.  Even after I called him and informed him that we are out of everything and it\'s been a month and a half.  He just sends it whenever he wants to and expects that to be okay.  I guess I should be thankful that he sends it, but still.  He\'s not the one sitting here with 2 diapers, very little toilet paper, etc etc. Eric is as usual the bomb-diggety!  He got me Season One of Samurai Jack on DVD!  How cool is that??  But then again he\'s always cool.  *does the happy dance*  And now I am off to bed.  Bono\'s nachos, y\'all.   '),(22986,2061087,'13,May,2004','2004-05-13 00:00:00','   It\'s quiz night!   urlLink \"Which American City Are You?\"     Seattle  Your dark exterior masks a caffeine driven activism.  You\'ll take up a cause and you\'ll get ugly to advance it.   urlLink \"Which prop from \'American Pie\' are you?\"     Warm Apple Pie  You are the warm apple pie!  It may seem like a contradiction but you are true-blue Americana and the object of every freudian fantasy.    urlLink \"What swear word are you?\"   You Are Fuck!  Go fuck yourself.   urlLink \"What color are you?\"   blue  You are a friendly, trustworthy person. Friends know they can depend on you. Your inner favorite color is blue.   urlLink \"Which Random Irish Gaelic Phrase Are You? \"     Ta mo bhriste tri thine   Ta mo bhriste tri thine - \'My trousers are on fire.\'You\'re a few bricks short of a load, aren\'t you? You\'re probably not allowed to use sharp objects and you should be locked in a rubber room. With Rubber rats. Rubber rats? I hate rubber rats. They drive me crazy. Crazy? I was crazy once. They put me in a rubber room. With rubber rats. Rubber rats? I hate rubber rats...   '),(22987,2061087,'11,May,2004','2004-05-11 00:00:00','   Wow, so blogger has made some big changes!  We have lots of new stuff to play with so I gave the site a makeover tonight.  What do ya think?  I like.  So, y\'all use those comments already!  The older posts won\'t have commenting but from here on out there will be comments enabled so give me some feedback now and again.  I am not here just to hear myself type.  Well, okay, I am but .. just put your 2 cents worth in anyway!   '),(22988,2061087,'03,May,2004','2004-05-03 00:00:00','   You Are a Plain Ole Cup of Joe  But don\'t think plain - instead think, uncomplicated You\'re a low maintenance kind of girl... who can hang with the guys Down to earth, easy going, and fun! Yup, that\'s you: the friend everyone invites. And your dependable too. Both for a laugh and a sympathetic ear.   urlLink  What Kind Of Coffee Are You? Take This Quiz :-)      urlLink Find the Love of Your Life  (and More Love Quizzes) at Your New Romance.   '),(22989,2061087,'03,May,2004','2004-05-03 00:00:00','   Well, here we are again, and it\'s been a week since my last update.  Blame Eric.  (well, hey, so long as it isn\'t me!  :p ) So, let\'s see.  Nothing exciting to report.  Haven\'t done anything other than the usual.  We\'ve watched some good movies.  Ate some good food.  Had some good.... uhm.. fun.  Eric and I went to the end of year BBQ thingy for the Computer Science Math Geek depts the other night and it was pretty fun.  He\'s like a scholar and stuff!  I lurve geeks. I broke my toe.  That was fun.   The weight loss thing isn\'t going so quick.  I think I\'m at 165 now which is where I THOUGHT I was before.  So, I\'m basically back where I should have started.  NOW it\'s 20 pounds.  Or else I am not going to be seen in swimwear.  Hayull no.  I really need to get back to the Pilates thing every day.  Too bad I\'m a lazy wench.  And a hungry one.  Sex makes me hungry.  Shouldn\'t the two cancel each other out?  It\'s exercise.  Right?  Hmph.   I changed my links over there to the right, for those of you who actually pay attention to such things.  I put it all on blogrolling so that it\'s not so big of a pain in the butt to update.   I had an invite today to join a pretty cool livejournal community.  Go over to  urlLink fallsgeeks  and check it out.  Of course, unless you\'re from Wichita Falls, it\'s not going to interest you a whole lot.   Well, what are you waiting for?  Go!  Move along.  Nothing to see here.   '),(22990,2061087,'25,June,2004','2004-06-25 00:00:00','   urlLink    Oh.My.Gawd.  Don\'t forget to click \"read the rest\".  bwhahahaha  &nbsp; urlLink   '),(22991,2061087,'24,June,2004','2004-06-24 00:00:00','   urlLink    HALLELUJAH! She hath returnethed! &nbsp; urlLink   '),(22992,2061087,'23,June,2004','2004-06-23 00:00:00','   urlLink    Classic!&nbsp; urlLink   '),(22993,2061087,'21,June,2004','2004-06-21 00:00:00','   urlLink    I love Jean-Paul Gaultier. &nbsp; urlLink   '),(22994,2061087,'12,June,2004','2004-06-12 00:00:00','   urlLink     urlLink Click here to go to site. &nbsp; urlLink   '),(22995,2061087,'31,July,2004','2004-07-31 00:00:00','   What you should be listening to if you aren\'t already.  Michelle Lambert - I\'m Fallin\' Langhorne Slim - Counting Fireflies Carlos Santana & John Lee Hooker - Chill Out Lucinda Williams - ANYTHING she\'s done.  Fantastic. Loretta Lynn - Portland Oregon (The whole Van Lear Rose album really) Ryan Adams - La Cienega Just Smiled (HOLY SHIT, why am I just discovering this guy??  I am blown away)  I really wish I had a big ass server where I could upload this stuff and you guys could just download it.  But it is worth going out there and looking for.   '),(22996,2061087,'28,July,2004','2004-07-28 00:00:00','   urlLink    Ahh.. my eyes.   Link via  urlLink me(ish)   urlLink   '),(22997,2061087,'23,July,2004','2004-07-23 00:00:00','   I am having chicken pot pie for breakfast.     \"Deluge\"  I\'m watching the pouring rain   streaming down,   washing away everything in its path.   And I\'m wishing that I could   step outside   and wash the memory of you   from my mind,   your touch from my body.   Instead,   I am flooded with images   of rainy days ,  bodies twined under the covers.   The downpour muffling   the sounds of love.   It occurs to me that I am glad   that removing your imprint on me   isn\'t something the rain can do,   rather   its purpose of nurturing   is to remind me of possibility.   '),(22998,2061087,'23,July,2004','2004-07-23 00:00:00','   I CANNOT stop playing  urlLink this damn thing  every time I see the link!  Help me!  Bwhahahaha   '),(22999,2061087,'23,July,2004','2004-07-23 00:00:00','   When the weather turns hot, as it does soon and often here in Texas, I rarely drink hot beverages. My dad has always said that if you are hot you should drink something hot and it will cool you down, and if you are cold drink something cold and it will warm you up. I always thought he was full of it. I mean, he is always coming up with insane, backwoods stuff like that and never once has it been credited. However, and forgive me because I forget where I read this, come to find out there has been scientific research done that proves his theory true. Of course, I haven\'t told HIM that. What brought that tidbit of information on was that the weather has been unseasonably cool and rainy here and I am sitting and drinking a cup of hot chai tea. Mmm. I haven\'t had any in months and had forgotten how good it is.   I haven\'t been doing as much posting as I keep telling myself I should here lately.  I keep my  urlLink book blog  updated somewhat more frequently, and yesterday I started a  urlLink flab to fab  blog.  I also have a  urlLink livejournal  but admit that it is only so that I can post and comment on some of my fave sites there and I don\'t ever really post anything on it. WHAT in the cornbread hell am I doing with so many blogs if I am not going to actually dedicate any time to updating them? Although, it\'s not like any one really reads them. They are mostly for my own venting/progress tracking (or lack thereof)/inane rambling. Such as I am doing now.   So, what has been going on lately, you ask? Well, let\'s see. This summer I\'ve been back home to visit briefly. I thought I was going to get away with not going back for a while, but my Mamaw was dying so I had to get back. I spent another week there that time. The oldest two kidlets have been at their dad\'s a lot this summer, leaving me and the Willabeast here with not a whole heck of a lot to do. Other than her destroying things, and me cleaning up the messes. My youngest brother came and got my mom after the funeral so she is out in Georgia with him. I talked to her last night and she\'s ready to come home and wants me to come and get her. Which I would be more than glad to do if Jan (my ex-mil) would hurry up and get the car that she is giving me fixed. Yes, my ex-mother-in-law is GIVING me a car. It\'s a late 90s model Ford Contour (which oddly enough was mine in the first place, she took over payments when we bought a Grand Cherokee) and is in MUCH better condition than the hunk of shit on wheels that I am currently driving. I\'m sure y\'all have heard me bitching about not having an a/c and windows that will roll down. NOT a good combination in the good ol\' Texas summertime. So, anyways, as soon as she gets some minor little things on the Contour fixed it is mine. Funnily enough, the day she went to buy her new car, the driver side window went out on the one she is giving me. It\'s like the damn thing KNEW. But, she\'s a good egg and it getting it fixed. And then I guess I will be road tripping it to Georgia. It\'s about a 13 hour drive from here. But, I don\'t mind as it will give me some alone time. I haven\'t been kid free in about 3 years, other than when I have had the various surgeries in the last couple of years. Those don\'t count as vacations. My brother wants me to move to Georgia. This will be a good opportunity to check it out and see if it\'s anywhere I\'d want to live. Although, it has to be better than here. But, that\'s not saying much.   I can\'t really think of anything else of any interest that has been going on around here this summer. Well, other than the ex and his wife are starting up again with \"the kids can come live with us and go to school over here if you think that\'s ok\" bullshit. Yeah, just so you don\'t have to pay child support anymore. But, let\'s not get started on that particular rant at this time. It would take more time than I have tea for.   '),(23000,2061087,'23,July,2004','2004-07-23 00:00:00','   urlLink    I can\'t quit hitting replay.  This is freakin HILARIOUS!!!&nbsp; urlLink   '),(23001,2061087,'22,July,2004','2004-07-22 00:00:00','   urlLink Which John Cusack Are You?   '),(23002,2061087,'20,July,2004','2004-07-20 00:00:00','   urlLink    words to get lost in.  i\'m in love.&nbsp; urlLink   '),(23003,2061087,'12,July,2004','2004-07-12 00:00:00','   urlLink    Semper Fucking Fi&nbsp; urlLink   '),(23004,2061087,'04,August,2004','2004-08-04 00:00:00','   Supplies:    Bowl of warm water with one squirt liquid hand soap, orangewood stick or stick with plastic tipped cuticle pusher, hand towel, emory board, 3 or 4-way buffer, Q-tips, polish remover, base coat (Stickey by Creative or similar), polish of your choice, top coat, UV nail dryer.        Soak      fingertips in bowl of warm, soapy water for 3 or 4 minutes.      Remove      from water and using orangewood stick or cuticle pusher, gently push      cuticles back.    Rinse.    Dry with hand towel, pushing cuticles      back with towel.  File      edges of nails with emory board, going in one direction.    DonÂ’t saw back and forth as it tears the      nail.  Buff      nails using 3 or 4-way buffer, following steps numbered on buffer.  Rinse      hands to remove residue.    Dry      thoroughly.  Dip      Q-tip in polish remover and rub over each nail.  Apply      thin layer of base coat.    Let set      one minute.  Apply      two thin coats polish, waiting one minute between coats.  Apply      thin layer of top coat.  Place      hands under UV dryer for several minutes until set.               Useful tips:    Store nail polish in refrigerator.    Remove and let reach room temperature before using.    Roll polish bottle between your palms, never shake it.    DonÂ’t polish your nails under an a/c vent or fan.    DonÂ’t change your polish more than once a week or your nails will dry out.       Keep in mind this deals with the nails only. There are many more extras that you can do to keep your hands and feet looking great. Lotion is a must. A foot scrub or sloughing lotion massaged onto feet keeps them smooth. Pumice stones aren\'t a good idea as they really only make the skin tougher.      This post was inspired by a post over on livejournal on the hip_domestics forum. I realized that many people don\'t know the proper steps to follow for a professional manicure.   '),(23005,4049268,'03,September,2003','2003-09-03 00:00:00','   The 1st day of army... beri beri scared sia. heard from all those ppl outside that army beri xiong de and sergent beri fierce and a lot of ah beng... all this is all in my expection le... wake up damn early sia then nid to report to amoy quee camp at 11am... so early also dunno so early for what... as i m approaching the camp, fear grew within me as i saw group follow by group of strangers going thru the camp gate, people of all shape n sizes sia, some fat some thin some tall some short sporting all different kinda of hairstyle beautifull embedded with different colours... beri scared that wont be able to mix well with all these strangers of different walks of life... There is an mini bus at the gate to bring us in. Thats was totally not i expected... a mini bus that is so small den no air con somemore, only got those sliding windows. The mini bus drive past an sign that shows 20TH SINGAPORE ARTILLERY BATTLION and stopped at the entrance. For once i thought that maybe army isnt really that bad as me n parents along with some others in the mindbus were greeted by soldiers of different ranks with an smiling face and warm reception. maybe ppl words cannot really believe ah... i went over to register then i was given an green card (11-b) which the officer that gave it to me told me this will be my nirc for the next 2 years le... along with a water bottle bearing the name army 21 and a bag. i was then directed by ushers in army uniform to the seating in front of the stage. Hanging boldly on the wall is the 11TH MONO INTAKE. sitting anxiously with my parents we waited for the ceremony to start, the beginning of hell... An officer walked up the stage with an smart uniform and adress us as new recruits of 11TH MONO INTAKE. along came an video which show what will we be doing as recruits. All the mud, sweat is all i should expect. After taking the oath of allegiance we were directed to some refreshment... Hmmz the cakes n drinks taste quite good so i was expecting good food in camp huhz... After refreshments all parents were invited to leave the battlion(camp) i bidded farewell to my family as it will be 2 weeks staying in camp and the reason is to let us get used to military life... stupid right. Reluctantly my parents went home while i prepared for the new start... '),(23006,4049268,'14,August,2004','2004-08-14 00:00:00','   After all the parents went home, the battlion is filled with teenagers like me. teenagers who are here to accept their fate... We were divided to three groups which is named after alpha, bravo and charile... we were brought up to the room and so called bunk by an army personal, i belong to charile wich is at the 4th floor... sux sia alpha 2nd floor then charile still have to walk up so long way up. 4th floor lehz... after leaving our bags in our bunk, we went down to '),(23007,4049268,'14,August,2004','2004-08-14 00:00:00','   Early in the morning rain no stop sia... good day to slp sia... go buy 4d le hope to win lohz cuz pocket like very empty like that lehz... haiz now pocket empty + pay cut to 350... arhz! how to survive ah??? sad sia no clerk job for me... just got my posting le... be storeman ah. sux sia have to live together with dust and stores... feeling is... (sux). the medical officer bluff me one, he say i will become clerk but in the end i become a man of store... haha storeman. So happy today, got a very good start of the day... now then i knoe my words got carry some weight one... not jus plain word but words taken in mind by others...hmmz okok i juz end here ba. nice day everyone... '),(23008,4049268,'14,August,2004','2004-08-14 00:00:00','   hey! finally the 2nd last day of couse le... after this one whole week of training how to keep myself awake during extreme boring lectures... haha i have finally mastered th skill of playing afool in class rather then sleeping... hmmz the test to determine that did we learn enough to pass the couse is rather all bout common sense la... got ppl take down notes in every lectures n people who sleep through the whole days and when the results were revealed, people who sleep the whole day have an average that is much higher then those that got pay attention in class sia... actually its all in the brain not on the lectures one... haha... like me la, sleeping in class, playing afool, teasing teachers... notorious is the word... haha juz like my sec sch life but i still pass with an result that no one will expect me to get... best ah... tml have to go back last day... stupid sia, go back give back the camp pass then go eat breakfast then go home le... stupid rite? even the breakfast also must ownself pay sia... a lot of ppl cheat in the test sia... thats how most of them pass... the sergents n bigger ranks ppl jus like come then ask u dunno rite then they give ans secretly... haha not hints lehz... its ans... but i never copy la... i anyhow guess also can get good grade sia... haha thats why they say saf tests cannot fail one... its beri hard to fail but beri easy to pass... with all the sergents help n retarded common sense... its juz like that... tml is going to b a short short day... if u wan me send u hm tml den tell me k... 24hrs avaliable... aniwae its an cold day today so keep urself warm n DONT catch a cold k... :o) '),(23009,4049268,'14,August,2004','2004-08-14 00:00:00','   sianz sia today go take mc from polyclinic then get one day mc so no need go course... haha my acting skill best ah talk to doc a while then he give the mc le... haha save money ah no nid go private clinic... hmmz didnt really did anything today lohz but brought durians again today haha yesterday also got buy, like got craving for durians daily sia... jia lat liaoz, confirm sick if continue eat sia... must curb le... tml early morning fotta go course again haiz the long bus journey beri beri sianz... bo bianz have to go eh u still haven ans me that *uUu* questionz lehz... aniwae take care... '),(23010,4049268,'14,August,2004','2004-08-14 00:00:00','   (\" n also not forgetting to *uUu* (should know who huh)... thanz also for being there for me always whenever i m in trouble, stress, sad or even in confusion... u nv failed to show ur concern... thanz alot... at least i know tat i m not alone... =) ... n also i m say sorry to *uUu* for sometimes u know... make u frustrated wif me... so sorry... mayb i will learn how to let out my troubles tat is bothering me \")-quote frm ur blog  (\" n also to *uUu* out there... u nv fail to show me ur concern whenever i m down... thanz also n sorry for always making u frustrated... so sorry... know u meant well... but mayb i will learn to open up my feelings...\")-quote frm ur blog   that *uUu* sound like mememe lehz... haha m i right? or m i bei hiao bie???  ___________________________________________________________________________________  (\" haiz.. recently dunno wat is happening to me... i really dunno... things seems to be not the same anymore... not the same... after tat time, things were different... i m really very scare... cant afford to let such thing happen again... argH... god u really have to help me thru this... i will pray hard... very hard...\" )  recently i met up with god le. he know that u r praying so hard le. he said to me that all the external factors are superficial, look on the inside n u will find peace... if u are scare then tell me wat happened... i will protect u or do whatever i can one... just tell me huhz. aniway was really happy when u told mi that u told ur mum about me... althought till now i still dunno why u dun wan me send u home this kinda things but ok la, i dun wan add stress to u... juz wanna tell u that whatever happens or whatever choice u make( 2 choices i told u yesterday )i will agree to it one althought how :o( i will b... happiness is like clapping hands. it always take two to achieve... i love you... now n forever... '),(23011,4049268,'14,August,2004','2004-08-14 00:00:00','   Hmmz going to watch the fireworks at esplanade at night but i dun really noe the place well so meeting 2 frds together with my bro going down... too bad sia some ppl cannot go... met up with 1 frd at hougang at bout 7 i think then continue going down to dobby ghaut meet another frd... haiz too bad i didnt realise that she (the frd that we are meeting at dobby ghuat) have an time keeping problem sia meet her 7.30 she 7.45 then come. hai we miss 5 trains... finally she came then we boarded the mrt... its was like packed sardine like that lohz. beri beri the squeezy ah. almost cant breath... haha. luckily when we reach the fireworks haven started yet or i will chop off that ger head sia. lol. everywhere was people, all size n pattern sia, some even brought grass mattress go there to find an comfortable place sit... the comercial bluff ppl one say what it will starts at 8.15 on the dot but when we reach there it was 8.05 then the first shot of fireworks were launch into the starry night le. it looks more like end at 8.15 on the dot sia... the fireworks were just fantastic. pattern that even national day dont have they also got, including star shaped fireworks, fireworks that burts into 7 colours when launch, fireworks that look like falling stars. the finale was just as fantastic... its like all the stars suddenly burst and the trail left in the sky was enchanting. it was an beri beri good firework display which i think everyone agrees as the applause was even louder then the fireworks sia... a lot of couple went and i think such beautiful scene should be shared with that someone but too bad... thought of asking u go next year one but i heard 2004 was the final year le... too bad rite. i noe. aniwae i juz end here le... '),(23012,4049268,'14,August,2004','2004-08-14 00:00:00','   Wake up at 4 am today sia... so tired as yesterday watch tv till 1 plus 2... still got things to do sia... only 3 hours sleep... how to ta han... :o( Had to go set up water point for Tiger Trail... its is an running event in jurong camp... had to go set up the water point to provide water supply for thirsty runner. the day haven even break yet... its was still pitch dark the took the army lilitart vehical to an forest like training ground which the runner is going to run though... haiz of all the people my sergent choose me... beri beri tired lohz... there are two kinda drinks that we must set up for the runners 1st is redbull another is 7up... that jungle area so mani mosquito... wanna catch an nap also cannot sia with all the mosquito making that irriating buzzing sound when they flew past... soon the run started at jurong camp while we waited for the runners to run past our area... sun is rising le luckily not hot ah or else confirm beri pek chek one... when runners started to run past us, we started to critize all the people running thru... it was really getting beri bo liaos ah so we talk bad about all the officers n sergents from different battlion. haha we can talk bout they run like ger sia, some hair loss till beri jia lat ah den some wanna act say, take off shirt run but eventually those were the one that run till like going to die already... hmmz this 12km run is also quite a killer ah... haha sianz ah from watching people who ran beri fast to ppl behing lagging behind by walking... we critize almost every sing runner... haha bo bian too bored le arhz!!!! when we offer their either redbull or 7up, they will ask for plain water... haha got good drinks they dun wan. stupid one dunno how to enjoy siaz. when we running as a platoon in camp, only got plain water then these runners here got good drinks then dun want they dunno how to enjoy sia... then my commander of camp who was also running mixed the 2 drinks together then dranks few cups of it then he started to invited his frds to come drink, telling them to mix both together... then i realise that when redbull mix with 7up, it will forms plain water ah haha...after the run was completed, we went back camp n sleep for the rest of the day le... '),(23013,4049268,'14,August,2004','2004-08-14 00:00:00','   today whole day slack at home also dunno do what sia... so sian... wanna fix my remote control car but dont have the necessary thing, wanna sleep but not tired... haha basically whole day walk walk from room to room sianz ah... haiz sianz ah tonight go back again... never mind going ORD next year le... Today go back then stay till sat ah super super sianz ah... that stupid sergent la... say whole camp go thailand then we cannot even have nights off... sTuPiD aRhZ!!!! I m going to craete mani mani trouble for him till he bei ta han. humph dare to play with charile people... haha... its going to be a long week... dun wan go think of it sia... going to do painting of the blocks all these... wasting my time arhz... army sux...       ~nO oNe CaN uNdErStAnD wAt I\'m GoIn tHrU..[cOnFuSeD]~... is the email mean what it says??? or m i looking too deep in??? haha only GOD knows... Added new javascript games for u to enjoy urself in case u r bored... '),(23014,4049268,'14,August,2004','2004-08-14 00:00:00','   aiyo y nowadays weather so hot one... cannot ta han ah, can even sweat when playing computer... haiz all my camp ppl going thailand for oversea training le, suck sia i cant go... zhao zhi dao i dun go for the operation then i will stay as fit for combat so can go chiong with all my frds... life as an storeman is rather bored lohz... dunno y ppl think storeman good to do... thinking that we have an easy life... wat they dun see is all the paperworks n all those bo liaoz things that we mus do... hao huai nian last time as an gunner, all the chionging n suffering with my frds... started to think too much nowadays, neccessary oso think, not necessary one also think... kana contaminated by SOMEONE sia who is always in a state of confusion one... haha... i beri afaid that she do not have time fer me lehz... this ger jus dunno how to arrange time one... is an lasting relationship so hard to achieve. i will do all my best to blend into her scedule but will she do the same??? dunno... yi shen yi shi you zhe me nan ma? jus pray for the best is the onli thing that i can do now... i no nid to be shy bout my feelings here as i didnt publicise my blog at all. only u noe my blog add n its all written for u... '),(23015,4049268,'14,August,2004','2004-08-14 00:00:00','   SiaNZ sia have to book in camp again tonight, the long straight road in is really getting on my nerve sia. everytime must walk in so far also dunno why they build the camp till so in wards. hmmz thot its gonna be another boring night that have been wats going on always but something happened that changed my mood for the nite and basically change my onlook about life, something that gimmie new hope. Its is a ray of sunshine for me... haha u muz be think whats that hope right? haha actually its someone that i have been waiting for beri beri long le... i wan very much to be with her but after breaking her heart long ago, i thot thats the end of our fate but tonight when i ask her about this thing, instaed of changing topic as she always did, she actually told me that i MAY still stand an chance... wow i was beri beri happy lohz, like i jus won 4d like that, even more happier then that lohz, although it is only a MAY but its just the thingy that i have been waiting... finally... finally sia all this is happening to me. If all my past months of suay-ness (bad luck) was created because it was all to exchange this chance to be with her... i most willing to take all the bad luck... even if one year of bad luck i will oso take... hmmz thats all folks '),(23016,4049268,'14,August,2004','2004-08-14 00:00:00','   Augest le... hope to throw every sway things that happen last mth away... this mth hope be better lohz... if it continues so sway i really cannot take it le... everything u can imagine can go wrong sia also dunno what happen sia... nothing goes my way at all... TIAN ARHZ!!! plz help me ah... i beri beri stress le... cannot take it le... hope everything will be fine this month haiz... '),(23017,3990866,'16,August,2004','2004-08-16 00:00:00','   Have finished the school structure section. Will work on PLU work and get it done, hopefully, by tomorrow. If I have time, will redo the maps section.  '),(23018,3990866,'12,August,2004','2004-08-12 00:00:00','   10:49am  I have just finished redoing the fundraising section. It works well and looks good.  Before I move on to the PLU work, I\'m going to go back and check on all the work I have done so far. Paying attention to the following:  1. Links 2. Formatting and text 3. Images 4. Inconsistencies 5. Unifying left hand nav bar to be consistent across all updated pages  After that, that e-mail to Joe &amp; Damian. Before all this, an eye break.  '),(23019,3990866,'11,August,2004','2004-08-11 00:00:00','   Notes made on first day. To be checked against master task list for project:  Both Sites General Notes  - way too deep and depth only in two sections - makes it very hard to access information - duplication --> new section of site has separate folder AND html file with setion name - lots of images and PDFs - opportunities for scripting  * intranet newsletters  * publishing process  * links page  * e-mail forms  * news - internal telephone list link broken - suggestion:  links going to internet site labelled or removed for sake of avoiding redundancy - suggestion: make nav bars for both sites more streamlined - not all courses on courses pages have links - suggestion: script news to keep it more current and any staff can add news - student section needs more content and development - question: i dont understand why pages r duplicated in the site map - definite need to reorganize and restructure website - too many different kinds of formats for marketing collateral posted (DOCs, PDFs, HTML)  * suggestions:   ^ either choose one bespoke or existing format and post in format  ^ implement mechanism to manage collateral published  Specific Notes - Break Down by Site and Section/Page  I. Internet Site i. \'About the School\' a. way too much text - too much scrolling happening - suggestions:  * break down into subpages/subsections/sub headings  DONE   * make images larger (Looking into it)  ii. \'School Structure\' a. no problem - well done page  iii. \'Swindon Campus\' a. no problem  (Ed.: Reformatted and broken down)   iv. \'Fundraising\' a. no problem  (Ed.: Reformatted, restructured, and broken down)   v. \'Staff\' a. needs reworking - too little content and too many pages  - suggestions:   * add pictures and complete profiles for all staff   * enforce template for staff profiles   * better/new way of classifying staff (Possibility: JavaScript aid)  vi. \'Press Releases\'  a. no problem - well done page  vii. \'Programmes and Courses\'  a. more explanation needed for different sections  b. amalgamate sub pages with introductory text then reformat first page to just have headings/sections  viii. \'Student\'  a. very bland - little content - needs massive boost and more information  b. FAQ not extensive and needs lipoinsertion bigstyle  c. \'Placement Learning Unit\' page - grammar correction needed - face lift with links placement   d. links on left nav bar need to be consistent with headings on main student page  ix. \'Contact\'  a. way too long - too much scrolling happening   suggestions: * javascript application  II. Intranet site  i. \'News\'   a. no problem - all systems go.  ii. \'Students\'   a. left nav bar links need to be consistent with main page links   b. \'FAQ\' - suggestion: javascript application with bit of scripting to reduce page length  iii. \'Staff\'   a. Internal Telephone List link broken  iv. SHSC information   a. Resources page useless - lipofunctionality / value  v. Policies & Procedures   a. another useless page - needs more link and lipofunctionality  vi. Links   a. suggestion: javascript application with bit of scripting to reduce page length  vii. take out Fundraising link - redundant  viii. take out Maps link - redundant  ix. take out Contact link - redundant  '),(23020,3990866,'06,August,2004','2004-08-06 00:00:00','   I\'m slowly working my way through the site. Fundraising section is under the scalpel at the moment and has been reorganized. I\'ll be adding a new page for just campaign details to update the content and feel. Once that\'s done, I\'m going to move over the PLU job and get that done, bringing in the new template to get it flowing with the rest of the site.  Other than that, CSS is becoming unmanageable on the site. Don\'t understand why it wouldn\'t style those classes today. *Sigh*. Plan is to finish an initial sweep to the 1.1 template and then stop completely on any further rework. Sit back, read, study, and do a proper design for the site as an on-going project. Then start on the same initial 1.1 process for the Intranet. Shouldn\'t take as much time as there is less content.  Note: Ask Annie about buying Sitepoint book. Come in early tomorrow and focus harder on work.  '),(23021,3990866,'06,August,2004','2004-08-06 00:00:00','   11:39am  Things going well. Index.html now redesigned and cleaned up to look like an introductory letter. It\'s now looking slick and easy to digest. Next to the about section and cleaning it up. All is well here so far.  18:19pm  Powerful day. About section now reorganized and all it needs now is just some final touchups and formatting. Will be done first thing Monday morning. Next after this, the rest of Annie\'s updates. Bit behind but I\'m getting in some quality to the work. So I\'m peaches and cream.  '),(23022,3990866,'05,August,2004','2004-08-05 00:00:00','   15:49pm  Finally. Completely done. The following were the objectives:  1. Indenting, organizing, and unifiying the markup                                                        -- DONE -- 2. Developing the commenting style to be used on ALL pages from now on             -- DONE -- 3. Breaking the template down into sections for code reuse                                           -- DONE -- 4. Optimizing the template and tweaking the tables without breaking the design    -- DONE --  Didn\'t blog yesterday as it was a slow and difficult day, and actually forgot to blog. I got so frustrated. Here\'s the scenario: I thought I finished the template. Off-line, it\'s beautiful. I upload it and update the stylesheets, and it breaks like a Ming vase. Couldn\'t figure it out and went home awfully frustrated.  Came in this morning and looked at it with a fresh pair eyes. Went back into the template and did further clean up. I had messed up the template before I left yesterday because I wanted to get all three main elements - the left hand navigation, content area, and right hand side boxes - to be on the same vertical height. So, I came in today and restored the template to health. Made a copy and worked on that copy until I did it. Golden rule:  Never touch a finished template. Always fine tune and hack a copy.  The template is now much cleaner, although still table-based. I must say though it\'s a weak-coupled table-based, as I have moved most of the styles and formatting to the CSS. Finally, the template actually looks virtually identical to the original Brookes one. So, I\'m quite pleased, although my eyes are beyond exhausted.  And all this for the links page.. *sigh*. At least that page is now complete. Next up: the index page is getting a serious face lift. But since I have my trusty template, it\'s just about doing some re-organizing and cleaning up.  I\'m taking a jazz break, resting my eyes, and doing some paper work. If I have time, I\'ll start coding up the new SHSC homepage!  '),(23023,3990866,'03,August,2004','2004-08-03 00:00:00','   10:21 am  After wasting so many hours yesterday trying to fix a broken table-based layout for the revamped links page, I\'ve decided to stop a bit on the update work and construct a clean, readable template from one of the files of the SHSC website. It\'s just a waste trying to infuse new blood into a corpse held together by spit. Preliminary plan for this template sub-project:  1. Indenting, organizing, and unifiying the markup -- DONE -- 2. Developing the commenting style to be used on ALL pages from now on -- DONE -- 3. Breaking the template down into sections for code reuse 4. Optimizing the template and tweaking the tables without breaking the design -- IN PROGRESS --  After finalizing the template, it\'s cloning a new links page using the new template and uploading it.  17:17 pm  So close yet so far. Got to the bottom of the template file and the bottom td cells broke. *Sigh* Will hopefully iron out the spooks in the morning. But so far, have managed to get an empty area in the middle and keep everything standing. Have replaced all the spacer gifs and used simple styles. It\'s all in the padding, that\'s my advice to recovering table addicts. Even got around images and borders.  So close and will be there tomorrow morning.  '),(23024,3366974,'24,July,2004','2004-07-24 00:00:00','   I\'ll keep this as short as possible. Laura has a boyfriend. '),(23025,3366974,'21,July,2004','2004-07-21 00:00:00','   \"No, sir; not yet,\" said Stubb, emboldened, \"I will not tamely be called a dog, sir.\" \"Then be called ten times a donkey, and a mule, and an ass, and begone, or I\'ll clear the world of thee!\"  Good old Captain Ahab... Finally started reading about him afer 100+ pages. Although it rained for quite a while, I didn\'t read as much as one would think. Power got knocked out once, but it came back on soon after. That\'s a good thing too, because when power in our area gets knocked out, we are always the last ones to get it fixed.  &nbsp; '),(23026,3366974,'20,July,2004','2004-07-20 00:00:00','   \"I will have no man in my boat,\" said Starbuck, \"who is not afraid of a whale.\" By this, he seemed to mean, not only that the most reliable and useful courage was that which arises from the fair estimation of the encountered peril, but that an utterly fearless man is a far more dangerous comrade than a coward.\" -Moby Dick  Well, read some of Moby Dick today. In town today, we finally got some good basketball games going. It was 2v2 to 10, but you had to win by 2. It was 15-15 until Scotty made a 2 pointer to win, so my team lost 17-15. Then we got 3v3 going, and played that for awhile until these two huge kids came, and it was 4v4, and despite my complaints they insisted on being on the same team so I just went along with it. And, just as I had feared and predicted, they anihilated us. I watched my dream utterly shatter before me as they schooled me. '),(23027,3366974,'19,July,2004','2004-07-19 00:00:00','   \"Glimpses do you seem to see of that mortally intolerable truth; that all deep, earnest thinking is but the intrepid effort of the soul to keep the open independence of her sea; while the wildest winds of heaven and earth conspire to cast her on the treacherous, slavish shore?\" &nbsp; Well, I read a bit of Moby Dick today and a little yesterday. Speaking of yesterday, oops! Didn\'t make a post, I forgot. Wasn\'t anything worth telling about anyways. I got to thinking, maybe it isn\'t a good idea to post every day anyways because I\'d have 365 posts a year, and that is a lot of damn posts. But, I guess it only shows a week worth at a time. So, hmm.  &nbsp; I went back to the doctor\'s today, and the verdict is this: I must start taking the full 25 MG pill, instead of half a pill (12.5 MG). That\'s basically it, exciting isn\'t it? &nbsp; I went to the fair with Bob, and after walking around for an hour or so we left. I saw some people I knew, but were the stuck up preps. So, basically the fair sucked. '),(23028,3366974,'17,July,2004','2004-07-17 00:00:00','   \"All our arguing with him would not avail; let him be, I say: and Heaven have mercy on us all - Presbyterians and Pagans alike - for we are all somehow dreadfully cracked about the head, and sadly need mending.\" &nbsp; Today, was quite dull. I read a fair bit from Moby Dick, and went uptown a few times. We had spotty showers today, so there wasn\'t much I could do (outside). I didn\'t see her (Lauara) cousin today either, could be due to the rain. Or, the simple fact that she doesn\'t come outside very often. We\'ll see, she is bound to come out sometime. I don\'t particulary feel like knocking on her door, especially since I don\'t know her that well. But, I suppose if I must... &nbsp; That makes me think, at first I didn\'t know what to think of her (Laura). But now, I can\'t help but think about her. Hmm... &nbsp; It\'s the fifth day that I took my medication, and things seemed fine today. I even ate before I went to bed and had no trouble sleeping. I go to the doctor\'s again on Monday, and from whatI understand they are going to increase my dosage. '),(23029,3366974,'16,July,2004','2004-07-16 00:00:00','   \"Be sure of this, O young ambition, all mortal greatness is but disease.\" -Moby Dick &nbsp; Well, I read&nbsp;quite a&nbsp;bit of Moby Dick today. Yet, I found it hard to find a quote worth posting. Actually, I was outside a lot today. &nbsp; Yesterday, as stated in my previous post, I met a girl who was apparently interested me. Well, I decided that this time, I would go to town and be prepared. So, when I awoke, I did all the usual things one would do. I even shaved, despite the fact that I loathe it. I took a drive uptown to see what was going on, and to my dismay, nothing. After killing some time, I went back uptown where I met up with some friends. We rode around town for a while, and then the girls came back outside. I talked to them, but only breifly. Laura, was going with her cousins to the Military Mess Hall. Why her cousin would even need to go there, I don\'t know. My friends and I played basketball not too far from there, I won the first three games before finally getting tiped out in the fourth one. I was hoping to talk to Laura some more before she left, but that didn\'t happen. I saw them leaving a while later, and I decided I\'d try to get her email, assuming she has one, or her phone number from her cousin sometime. A thunder storm was upon us though, so I\'ll have to wait until tommorow, or the next day. &nbsp; My past few posts have failed to mention my progress with Zoloft. So far, besides an itch here or there, and some other symptoms that very well may be sinus related, I haven\'t noticed any sideffects, except perhaps an upset stomach. Which, odly enough, never seems to bother me in the day, only at night. As you can imagine, that makes it quite uncomfortable when your trying to get some sleep. We\'ll see what happens in the coming days. '),(23030,3366974,'15,July,2004','2004-07-15 00:00:00','   \"...yet see how elastic our stiff prejudices grow when love once comes to bend them.\" -Moby Dick &nbsp; It was tough to find a quote that I liked enough to put on here, even though the chapter was very entertaining. But, I barely had the time to read today with all the occurences. &nbsp; First, my day was very typical. Got online, and my good friend Brandon was on. After a while, my other friend Bob got on and we played a few games of Starcraft (Online/Offline PC Game). Then, my day started to pick up. I eventually played some basketball after eating uptown. I noticed to girls riding their bikes, one of which I knew. I debated about inviting them swimming, but didn\'t. I drove around town for awhile, and ended up running into Andy. I talked with him a while, and went and picked his brother Scotty up to play some basketball. Dylan was skateboarding, like always, and we conned him into playing. We were sweating half to death it seemed, so we ended up going swimming. By the time we got ready, however, we were dry and no longer hot. Thus the pool felt freezing, and we only swam for a mere 11 minutes. Perhaps I have gone in too much detail of the events so far, but I guess it is better this way. Well, I loaded my bike in the back of the truck, which I might add is quite the choir since I have a bed cover, since I decided that riding my bike was a good idea. We ran into dylan again, and we rode around. Sarah and a girl (who I couldn\'t recognize at the time) were oustide and she said Hi to me, how nice. (only took three times of riding by, lol) Scott and I were playing some basketball at the other park, when some girls walked by. Considering the way they took, I got the feeling that one or both of them might like us. They starting swinging, and that (to me) gave it away. I couldn\'t help but almost laugh a little. I didn\'t say anything to them though, at least until they left. I said \"later\", as if to at least acknowledge their presence. We rode&nbsp;our bikes by them a little while after, and I think one of them side Hi, but I can\'t be sure. Scott pussed out (ironic that we are both named Scott, I know) and went home at some point. I however, wanted to keep riding my bike for a little while. This, turned out to be a good thing, at least so it seems. Some girl called me by name, and I replied \"what?\". \"Come here\" she commanded. I thought about it for a second, and did go over. I knew the girl, her name is Calista, and might I add she is annoying. She introduced me to a girl named Laura, and informed me that she thought I was hot. Gee, how nice. I would myself be pissed if one of my friends did that, but on the other hand, if it didn\'t backfire, my anger would turn to joy. She (Calista) kept bugging me about various things, all regarding Laura. Quite annoying she is, but she served her purpose yet. Truth be told, Laura is quite pretty. She is only 14, but... oh well. I feel that I took too much of a shy approach to the whole situation. I did talk to her, but not that much. She is staying with her cousin, but she leaves tommorow. Hopefully, I can catch her before she leaves. And this time, not be so damn shy. If not, all is well. I can get her phone number from her cousin. Sorry about rambling, but there was just a lot to tell. '),(23031,3366974,'14,July,2004','2004-07-14 00:00:00','   \"The ribs and terrors in the whale,   Arched over me a dismal gloom, While all God\'s sun-lit waves rolled by,   And lift me deepening down to doom.  \"I saw the opening maw of hell,   With endless pains and sorrow there; Which none but they that feel can tell-   Oh, I was plunging to despair.  \"In black distress, I called my God,   When I could scarce believe him mine, He bowed his ear to my complaints-   No more the whale did me confine.  \"With speed he flew to my relief,   As on a radiant dolphin borne; Awful, yet bright, as lightning shone   The face of my Deliverer God.  \"My songs forever shall record   That terrible, that joyful hour; I give the glory to my God,   His all the mercy and the power.\" -Moby Dick  Finally, after several days, I have begun to read the more interesting pages of Moby Dick. The current chapter that I am on (Chapter IX. The Sermon) seems so interesting, that I decided to stop after that last quote. So that tommorow, I may have another interesting quote for my entry.  Today is the second day that I took Zoloft, so far so good. Every once in a while I will feel like there is something wrong with me, but nothing really is. It\'s just my mind getting the best of me. However, my mind doesn\'t seem to be messing with me as much as it did for the last medication I took.  I went uptown to play some basketball. At some point, my friend Andy drove up and we chatted for a while. He was telling me about this kid who wanted to fight him, for what reason I do not know. Anyway, he was telling me that this kid, and a friend of his showed up at his girlfriends house when he was there and wanted to fight. Her dad was outside mowing the lawn though, and when he saw he came running. They saw him and took off in their cars. Why is it that this county is full of such pricks? Always wanting to fight for some stupid reason or the other. '),(23032,3366974,'13,July,2004','2004-07-13 00:00:00','   \"And the women of New Bedford, they bloom like their own red roses. But roses only bloom in summer; whereas the fine carnations of their cheeks is perennial as sunlight in the seventh heavens.\"                         -Moby Dick  Lately, I have been reading Moby Dick. That was an excerpt from what I read today, and I believe I\'m going to start every day\'s entry like that from now on.  Today, for the first time, I took Zoloft. So far, I haven\'t had any sideffects, but it takes a few days for Zoloft to get in your system. Hopefully, all will be well...  \"Whatever my lot, thou has taught me to say, it is well with my soul\" '),(23033,3366974,'12,July,2004','2004-07-12 00:00:00','   It has been a few months since I officially found out that I (possibly) have OCD and ADD. Well, today I go to the doctors at 2:45 pm again, to hopefully solve my problems.   I should probably explain the events leading up to today, so you have a better idea of what is going on. First, they believed me to have ADD, and they still do. The put me on Concerta, and it seemed to work. However, I refused to take the medication any longer because I felt the side effects were too negative. Turns out, I wasn\'t even having sideffects from the actual medication itself, but from my own paranoid mind. This is when they concluded that I also had OCD. From what I understand, OCD is common with people who suffer from ADD. OCD, for those who don\'t know, can be credited to those who are obsessive about something (in the sitcom Friends, Monica is obsessed with cleaning), but that isn\'t the only symptom. OCD can cause paranoid attacks among other things, and this is what my problem is. I\'m overly paranoid when it comes to health issues, and this is my main problem and focus at the moment. First, I must treat my OCD before I can treat ADD. (Since I won\'t and can\'t take the ADD medication without having Paranoid Attacks caused by the OCD). Of course, I am paranoid about taking the OCD medication itself, and there are some side effects that scare me. So I didn\'t take the medication that they presribed for that. It has been a few months since they issued it (Zoloft) for me to take. During that time I have experienced some rather unpleasant Paranoid Attacks, and decided that I would go ahead and take the medication, pending another doctor visit (which is what today\'s appointment is). In case your wondering what the side effects that I\'m worried about are, they are as follows: Sexual Sideffects that may include painful errections, Tremors, and Seizures. There are more minor sideffects, but they aren\'t bad enough to bother me. They tell me that there is a very slim chance of me actually having any of those sideffects, but it doesn\'t do much good to ease my mind since there is still a chance, however slim. But, I\'ve decided I\'d rather test my fate, because I can\'t continue to live my life the way I am now. Paranoid Attacks are very unpleasant, and they must stop.  I will update you in a few days on my situation, if I\'m still around to report on it. '),(23034,3366974,'04,August,2004','2004-08-04 00:00:00','   It has been quite a while since I made my last post. For that, I apologize (not that anyone reads this). A lot has happened in the nearly two weeks since I made my last entry. I will start off by first mentioning my progress with OCD. After another doctor visit, I have been uped to a 50 MG dosage. I have been taking that for more than a week with no side effects. I still ask questions about stuff, but I seem to spend less time dwelling on it. I haven\'t asked if my chicken strips from Wendys are fish or anything like that. How one could not know the difference between chicken and fish is beyond me... Cut me some slack here.  Lately, I have been hanging with the kids uptown. Riding my skateboard or my bike around town. The kids uptown are Seth (real name is Isiah), Jordan, Dylan, and Zack. Sometime Kaitlyn will be riding her bike uptown and stop and hang out with us. We never converse however, because her mom is my Uncles ex wife. Funny isn\'t it? I never talked to her when we were \"cousins\" either though. She has no blood relation to me, so whatever.  Some incidents that you might find interesting have happened since my last post, so I\'ll share them with you. The first one that comes to my mind happened just recently. I saw some girl standing outside the Ice Cream Shop, so I rode my bike up to the shop so I could see what she looked like. Well, as I was close to her Seth yelled \"Scott, is she hot?\" Embarassing and funny in its own right. The gang skated up to the shop and we all were just hanging out there. The girl got in a Blazer and it went around the block and went past us again. As they were right by us, they came to a near stop. The girl was sitting in the passenger seat and smiled. Then they drove away out of sight. I need to start finding out girls phone numbers, because I\'m sick and tired of letting them slip through my fingers and out of my life.  An exciting posibility has made itself known to me. My friend Tim might move back to Indiana. You see, Tim and I were best friends for years in grade school before he moved to Texas. Although we didn\'t keep in touch as much as I thought we would, we still remained friends. He came back up for freshman year, but he ended up moving back to Texas. I left highschool to get homeschooled that same year, before he moved. Well, his dad told him he is moving back up to Indiana. Tim doesn\'t know if his dad will follow through with it though. If he does that would be cool. One interesting concept about him moving back is the possibility of being able to work out. All my friends up here are lazy and don\'t care about their appearance. Me on the other hand, would like a six pack and bulging biceps. Tim works out, and he said he has the same problem down in Texas. I never work out, for the very fact I just stated. I need someone like Tim, who isn\'t lazy, to spot me. Maybe, I\'m not sure about this yet, but just maybe I could start taking Taekwondo again. Tim might be able to keep me motivated and having someone along for the ride would make it go by faster (its a 30 minute drive). It would be even cooler if Tim gets homeschooled, then we could work out in the morning and cruise for chicks after school gets out. Oh man, the possibilities are endless. I haven\'t had a friend that I could just hang out with all the time like I used to.  Last, but not least. Is the altercation I was envolved in. Well, Seth and I were playing basketball and these stupid little kids, which I have had problems with before came up and were harassing him. They were claiming that he gave cigerettes to their little brother, who is only 9, and was with them. I defended him and asked the kids what the whole deal was. I mean, even IF he did, the kid took them. Well, one of the bitch sisters stole my basketball and ran home with it. Her twin sister, Calista, which I have mentioned before in my posts, told her to give it back. Calista likes me I guess, because she was on my side. Anyway, so now my basketball had been stolen. Well, Brent and Mike rode up on their bikes and we were telling them about all the shit that was going down. The four of us rode our bikes up to their house and demanded that they give my ball back. There mom made them get it while she proceded to threaten us. \"Don\'t give my kids alchohol, drugs, or cigerettes or I\'ll blast you.\" Seth says \"I didn\'t give them cigerettes. Then Brent tells her that her stupid kids were throwing rocks at his house. She tells us that they will be dealt with and goes inside. The bitch that stole my ball tosses it as us and Brent says Thanks. She says \"Jackass\" and goes inside. (lol) Mission Complete. '),(23035,2711915,'30,January,2004','2004-01-30 00:00:00','   Last night we went to the theater and saw a girl get her first boner. Then another girl gave her head. Sounds like family entertainment, eh?  Actually, it was just the warm-up act for the main attraction. We were sitting damn near the front row of the  urlLink Alamo Drafhouse . While this gives an excellent view of the relatively small screen, it also places you too close to the speakers -- considering that they are the only speakers in the theater, they are inevitably cranked up loud enough to reach the deaf drunk in the back, which results in a deafening ring in your ears the next morning.  So, you may be wondering about the two girls. Kat and Morgan were volunteers from the audience. The MC for the evening, who was wearing a  urlLink giant foam cowboy hat  and a pair of  urlLink brown and orange Adidas  identical to mine, had initially requested two girls from the audience with the condition that they be \"fun.\" Lacking volunteers, he eased his requirements to include \"semi-fun,\" and finally, \"boring ass bitches.\" Fortunately for the crowd, someone in the \"semi-fun\" category finally stepped forward. Of course, they were unaware of the challenge that lay before them.  Morgan volunteered to get her first boner. They laid her down on the stage. The MC then called his assistant to the stage with a few props. The first was a box of  urlLink Krispy Kreme donuts . One Original Glazed was placed on the girl\'s crotch -- err, swimsuit zone. The assistant then produced a banana, which he peeled. He then offered to the audience a choice: straight or bent. The audience voted \"bent,\" which was apparently the choice of every performance before. The \"bent\" banana was then inserted snuggly into the Krispy Kreme.  \"Of course,\" to quote the MC, \"no boner would be complete without some cream.\" The MC proceeded to crown the banana with ReadyWhip. He then invited Kat to join him on his knees. It was time to \"service the first boner.\" And she did. To be honest, it looked as though she had done it before (or had at least received professional training).  What\'s the point of the stroy? Is 35MPH sliding down the ladder into a smut-ridden bog of sex and fart jokes? Well, maybe, but that\'s not the point of this post. The point is the follow-up to the opening act:  urlLink Spike & Mike\'s Festival of Sick & Twisted Animation . Two hours of non-commercial animation with a seriously non-PC slant. In other words, it was quite entertaining.  One particularly well-done animation was Happy Tree Friends, from the people at  urlLink Mondo Media . Oddly, I\'ve been speaking with the artists at Mondo Media for the last two weeks, the whole time not realizing that they did  those  animations. In fact, I\'ll be visiting Mondo on Monday in San Francisco. (I\'m working with them to develop a demo for our products and Alias Maya.) I\'ll be sure to get autographs for everyone.  '),(23036,2711915,'29,January,2004','2004-01-29 00:00:00','   Sorry no posts yesterday - I was skiing. Uh, I mean I was sick - yeah that\'s it.  As far as the graphics for the page - I think we need to make a sincere effort to draw this stuff out on our own, without precedent or looking to see what else is out there. 35mph is a creation of our own and I believe everything within it should come from us - not anyone else. The graphics for the front page are only a small part of an overall feel that 35mph should have. It should all relate back to what 35mph means to the group and to each of us. A little mooshy (sp?), I know but that\'s how I feel. If I were to pick a website that has the same cohesiveness that I would like to see in ours it would be  urlLink Tool .  Speaking of Tool, Bennett - are they doing anything crazy (a new album maybe?)  '),(23037,2711915,'29,January,2004','2004-01-29 00:00:00','   The Tall Grunge was really just a visual treatment, not necessarily a final product for a front page or a template for the site. Just an aesthetic piece, really.  I\'m not quite sure where to go with a site template. I definitely agree with Heath that the template needs to evoke more of a gritty film-like quality, similar to the previous placeholder graphic form the Untitled piece. The Tall Grunge was something starting in that direction, but it\'s not quite evolved to the point of a template for the site.  Come on, guys... y\'all surf the web. Do you guys not see anything out there in the same genre of what we\'re wanting to do? Any similar imagery?  '),(23038,2711915,'28,January,2004','2004-01-28 00:00:00','   What are both of your ideas about, what the front page should contain? A stand alone graphic? Some information? I like the idea of a descriptive but highly artistic image on the first page.  '),(23039,2711915,'28,January,2004','2004-01-28 00:00:00','   Here\'s another quick 35MPH treatment I quickly put together this evening in Photoshop. It leans more toward the \"grunge\" style, and it\'s tall, so there you have it...  &rarr;  urlLink Tall Grunge   I quickly picked up some free fonts from  urlLink Chank . I then picked up some grunge brushes from  urlLink Nocturna . For the background, I took a close-up digital picture of a colorful ashtray (the blue green tones) and some potting soil (the earthier brown tones). Mix it all together and you get what you get. Enjoy.  '),(23040,2711915,'28,January,2004','2004-01-28 00:00:00','   I completely understand what you mean, Heath. I\'d like to go with the older style as well, I just don\'t happen to have any work done like that at the moment. Hopefully, if my harddrive is recovered, I\'ll have some of that to post. Of course, the next time I get a moment to sit down with Photoshop (maybe this weekend) I\'ll aim in that direction...  '),(23041,2711915,'28,January,2004','2004-01-28 00:00:00','   OK - Troy, I know that you were just playing around with Photoshop but... I guess what I had in mind for a graphic library was more related to the old frontpage of 35mph.com. Maybe what we need to do is put our heads together and develop this graphics library. I really liked the screenshot from Untitled. It gave me a good film vibe. That stuff makes me think I\'m watching Nickelodeon (sp?). Just an honest crit - nothing personal.  '),(23042,2711915,'27,January,2004','2004-01-27 00:00:00','   I like the ideas, Troy. I was thinking, what about a similiar scene to the jungle but with space stuff. Just an idea. Here goes nothing...  &rarr;  urlLink Hubble Telescope Website  (super hi-res images available).  '),(23043,2711915,'27,January,2004','2004-01-27 00:00:00','   So, I promised I\'d post some Photoshop work I\'ve done. I\'ve been playing around with different arrangements, color schemes, graphics treaments... I haven\'t really settled on anything. Unfortunately, my harddrive kicked the bucket over Christmas. Well, I got a new machine from the deal but I lost a lot of Photoshop work. Damn.  So, I\'ve been starting to create new stuff. The blog gives me new inspiration. Here are a few elements I\'ve done:  &rarr;  urlLink Jungle 35MPH  &rarr;  urlLink Pentagons  &rarr;  urlLink 35MPH Bluegel Logo  '),(23044,2711915,'26,January,2004','2004-01-26 00:00:00','   Prior to the primary photography for \"Shadow Vision,\" I\'ll be setting up lights and recording equipment for a Tyson Foods Event. I\'ll be responsible for gels, constructing rigging for lights and assistant camera work. This will be February 4-6. Though I\'ve been contacted about \"Shadow Vision,\" I have not received an assignment yet. I hope to hear that from Ozark Film and Video or through Alder very soon.  By the way, I researched a little concerning Alder\'s \"Winterfest\" event at Meadowbrook. It appears to have been cancelled at the last minute for insurance reasons. I think it was going to be to expensive to do legally.  Lastly, could one of you send me an email describing how to put a link in the posts?  The short instructions sent previously didn\'t quite soak in.  '),(23045,2711915,'26,January,2004','2004-01-26 00:00:00','   How about an update on your grip duties BWR?  '),(23046,2711915,'23,January,2004','2004-01-23 00:00:00','   Take a deep breath... we\'re about to go deep, deep, deep into the biggest 35MPH weblog post...  EVER!   My adventure today started with a link forwarded by co-worker Ben:  urlLink sputnik7.com . This site, dubbed as \"the world\'s first real-time audio/video Internet entertainment experience,\" thinks of itself as a broadcast network. It evokes a possible template for our ever evolving, deeply beloved 35MPH.com. While sputnik7.com includes video and audio \"stations,\" I was more interested in their video on demand. Herein you can find some great shorts, mostly of the digital variety. Currently on the front of their video-on-demand archive are their RESFEST picks. And here we go deeper...  First things first:  urlLink RESFEST . This is an \"annual, global festival dedicated to showcasing innovative film, music, art, design and technology.\" Right. Sounds like everything we love to do, so these guys have got the right idea. RESFEST is a companion to what was originally a magainze,  urlLink RES . The magazine focuses on similar topics, obviously. Always cool, always expensive (shiny, glossy, thick pages, good binding, bi-monthly... about $60/year for a subscription).  Second things second: the RESFEST picks. Four of them stood out above the others as worthy of commenting. The rest are obviously the pick-of-the-litter as far as cutting edge shorts go, but we shall leave them as an exercise for the reader. My recommended viewings are the following:  urlLink  Sweet  ,  urlLink  Everything Hits At Once  ,  urlLink  Vision Point   and  urlLink  Tongues and Taxis  .   Sweet  has the most professional cinematography of my four selections. And for good reason: the cinematographer was five-time Oscar nominee  urlLink Allen Daviau , whose work includes  urlLink E.T. the Extra-Terrestrial ,  urlLink The Color Purple  and  urlLink Empire of the Sun . But, it was shot entirely on digital video... so how did they get that spectacular \"film-look?\" The answer is simple:  urlLink The Orphanage \'s Magic Bullet Suite. This suite of tools (plug-ins for Adobe After Effects) provide framerate adjustment, color balancing, and letterboxing utilties to shift digital video from Local-And-Late-Breaking to Cinemascope and beyond. There was an interview with one of the creators of Magic Bullet on TechTV; you can find the article  urlLink here . For a more in-depth review of the product,  urlLink check out this review from Ken Stone  and  urlLink this newsgroup posting .  The second short worth watching is  Vision Point . This short introduces a unique technique I\'ve not seen before. The camera is focused on the horizon (or an object, like a mountain, on the horizon) during time-lapse footage of a car trip. The end result is quite hypnotic and his editing is synched with the music toward a very trippy effect. The technique, dubbed Tran Scan by its creator,  urlLink Stephen X. Arthur , is showcased in  urlLink a second short of the same name  wherein Arthur travels the Trans-Canadian Highway from one side of our neighbor-to-the-north to the other.  Arthur also did another short of incredible digital trickery,  urlLink  Transfigured  . In this work, Arthur took the collected works of abstract painter  urlLink Jack Shadbolt  and brought them to life through a multitude of digital techniques. An inspriation to us all!  The third (or fourth, depending on who\'s counting) short to check out is  Everything Hits At Once . This short was done by a local (to me at least, in Austin) web illustrator,  urlLink Divya Srinivasan . The short was done using a non-commercial software package dubbed Rotoshop. The look may be familiar to those who have seen  urlLink Waking Life  by  urlLink Richard Linklater  (of  urlLink Dazed and Confused  fame). Rotoshop was created by artist/programmer  urlLink Bob Sabiston  of  urlLink Flat Black Films . The software is very, very cool, and I recommend that everyone rent/buy Waking Life where one can find a behind-the-scenes demonstration of the software. To provide a more comprehensive look, here are some of my favorite links:  &rarr;  urlLink From the Waking Life website: Rotoscoping (QuickTime movie)  &rarr;  urlLink Wired News: Animating a Waking Life  &rarr;  urlLink Wired 9.02: Unreal World  &rarr;  urlLink Apple: The Dreamlike Animation of &quot;Waking Life&quot;  &rarr;  urlLink TechTV: Rotoshop Animation  &rarr;  urlLink Animation World Magazine: Dream Is Destiny: Waking Life   And last but certainly not least: a bit of traditional animation. Well, semi-traditional.  urlLink Mike Overbeck \'s cartoons are made in  urlLink Flash . Flash was originally developed as a web-animation tool. And for those for a mind toward history, back in the whacky, halcyon dot-com days of the late 1990\'s, our intrepid blogger  urlLink Troy was deeply involved with Flash . Back to Mike, though... his cartoon  Tongues and Taxis  is featured in the RESFEST picks. It\'s fairly funny and cleverly animated. For a little (very little) behind-the-scenes footage see  urlLink his documentary on its making . Don\'t forget to check out his political cartoons on his website... they\'re good for a laugh.  '),(23047,2711915,'21,January,2004','2004-01-21 00:00:00','   A new link added to the permanent list:  urlLink Etaoin Shrdlu . This is the blog for the esteemed screenwriter, Rolin Moe. Always good writing... which inevitably leads to good reading.  '),(23048,2711915,'20,January,2004','2004-01-20 00:00:00','   A must-see site for those interested in animation. Along with iFILM, one of the best collections of animations on the web... Wired\'s  urlLink Animation Express . This site has things in all formats, from Flash to Video to GIF, and in all styles, traditional 2D/3D, stop-motion and more. You could spend a long time watching all of the clips. And they\'re open to submissions!  '),(23049,2711915,'20,January,2004','2004-01-20 00:00:00','   Coincidence?  As most of you know I work in a landscape architecture firm and rarely do my experiences in this field have anything to do with film. BUT - just recently a well known indie-film producer contacted my firm and requested that we develop conceptual drawings for a piece of property he has in Idaho. His name is Stephen Dyer and his works include films like  urlLink Finding North ,  urlLink Late Bloomers  and  urlLink Ball in the House .   The 35MPH network continues to grow...  '),(23050,2711915,'20,January,2004','2004-01-20 00:00:00','   Trick or Treat? Pumpkin pie anyone? There\'s a whole world out there waiting to be dominated by you creative fools, I\'ll mix the celebratory cocktails and make sure the soundtracks are ontrack - And should my creative organs be of use, I would be obliged to comply. Vote Bush, Love, cj  '),(23051,2711915,'20,January,2004','2004-01-20 00:00:00','   I don\'t know what \'streak knives\' are but the Eldorado sounds like a honey!  '),(23052,2711915,'20,January,2004','2004-01-20 00:00:00','   This nice little site was forwarded on to me from Elizabeth\'s cousin in Hawaii.  urlLink DFILM  is a \"visual communications\" company that makes various things for various people. The cool thing that they\'ve done (I\'m sure there are others) is something called the DFILM Moviemaker (you can find it off their homepage).  DFILM Moviemaker allows you to create little Flash films by choosing characters, scenes, backdrops and music. You even get to throw on a title sequence. It\'s not perfect: the choices of scenes are fairly limited, and the characters are not very generic, but it\'s fun to try to make something interesting within the limits they establish. It\'s an interesting idea, something I\'d like to see more of on the web.  To introduce everyone to DFILM, I made a little one of my own:  urlLink My Homage . This link probably won\'t last forever, but it should last long enough for those here to see it.  To go along with this homage, here are some expository links:     urlLink First Prize: Cadillac Eldorado     urlLink Second Prize: Steak Knives     urlLink Inspiration: The Glengarry Mix  '),(23053,2711915,'20,January,2004','2004-01-20 00:00:00','   Check this out...  I was searching Google for the Shadow Vision movie and this article about Alder and APC came up... quite interesting - especially the last paragraph.  Here\'s a snippet:  \"The Academy for Progressive Cinema has received an endorsement from Terry Gilliam, who directed the Monty Python movies, Ken Burns, the documentary master, and Susan Sarandon. All three are on Alderâ€™s Board of Advisers.\"   urlLink http://aatonline.com/searchfinal.asp?ID=882  '),(23054,2711915,'20,January,2004','2004-01-20 00:00:00','   Time to get out your gels!  '),(23055,2711915,'20,January,2004','2004-01-20 00:00:00','   It\'s strange how these things come in waves...  Just as the weblog gets started and ideas start flowing, I get a call from Ozark Film and Video. They have been hired to do more work with \"Shadow Vision\" (Vince Vaughn and Dwight Yoakam). They need me to shoot with them the first part of February. The pieces are moving...  '),(23056,2711915,'19,January,2004','2004-01-19 00:00:00','   A re-edit or (re-shoot?) is definitely in order....but a trip to Austin, for myself, is not in the forseeable future. But then again, neither was my trip to NWA for the holidays and I still managed to share a few laughs and a lot of drinks at the Lounge. Maybe we could work something out over a long weekend. At the least, a collective mindset needs to be re-established and start up on some fantastic new project.  '),(23057,2711915,'19,January,2004','2004-01-19 00:00:00','   Long time listener, first time caller...  I believe this idea has been proposed before, but I would really like some intense trimming on the untitled film.  Maybe we could plan a long weekend in Austin to have a re-edit.  I think, if people are going to see it on the website, it should be in a very polished form.  '),(23058,2711915,'19,January,2004','2004-01-19 00:00:00','   Just added a link to  urlLink TroyGilbert.com . This is my own personal blog devoted to the games industry.  For those that don\'t know, I work in the games industry. I don\'t make games, but I help other companies make their games (I\'m a field engineer for Criterion Software -- we make various game technologies). This means I end up seeing the inside of a lot of games -- games like Call of Duty, Dead To Rights 2, and Kill.Switch -- instead of spending 18 months looking at one game. I think that uniquely qualifies me to share my slanted opinion.  Oh yeah, I play quite a few games as well. But doesn\'t everyone?  '),(23059,2711915,'19,January,2004','2004-01-19 00:00:00','   Over the next few days I\'ll start uploading some of our past productions. Right now, there\'s only one 35MPH production on-line, and it\'s seasonal. Check out the  urlLink Happy Halloween Stopmotion Greeting .  The sets and characters were created from scratch by Elizabeth (actual size of house is about a foot-and-a-half tall). She also did the animation (over 600 individual still frames). I did the editing, sound and effects.  We used the following software:  urlLink Stop Motion Pro  for capture and rough edit;  urlLink Adobe Premiere  for final editing and sound mixing;  urlLink Adobe Photoshop  for titles; and  urlLink Sony Sound Forge  for audio effects editing. Some audio samples were used from  urlLink SoundDogs .  The entire production, from storyboarding to set construction, filming to editing, took about two weeks (evenings and weekends). We began by storyboarding. Unfortunately, we accidentally threw them away after the shoot was finished so I can\'t post them. The final video is fairly close to the storyboard, the only difference being that we originally intended to have a cat as well -- it was cut due to the difficulties in animating it.  The filming was done with a Sony DCR-TRV330 Digital8 video camera (several years old, roughly equivalent to the  urlLink DCR-TRV350 ). Stop Motion Pro allows you to capture multiple frames at a time and average them together. This virtually eliminates video noise and interlacing artifacts without introducing too much softening to the image. Stop Motion Pro also provides an onion-skin mode that allows us to preview the live image against the previously captured stills, making animation far easier than without it.  The video has been featured at the top of  urlLink Stop Motion Pro\'s Gallery ! It may not be the Oscars, but it\'s good company nonetheless.  '),(23060,2711915,'19,January,2004','2004-01-19 00:00:00','   Okay, I think we\'ve cleaned things up a bit... and moved ourselves to the proper spot, 35mph.com. How about that?  Direct your attention to the links to my left:  urlLink IMDb  and  urlLink iFILM . Notice how they either begin with or end with a lower-case letter. That\'s what we in the industry call geek-chic.  Seriously, folks... IMDb may be the single most useful thing on the Internet (outside of  urlLink Google ).  And iFILM... well, it hosts a great deal of the strangest stuff on the \'net (in video form, of course).  These two sites are great \"first links.\" Why? Because they represent some pretty solid ideals for what this site will eventually become: useful -- and more importantly -- a place where you can find some of the strangest stuff on the net (in video form, of course).  '),(23061,2711915,'19,January,2004','2004-01-19 00:00:00','   It\'s really bright in here... and where\'s all the furniture?  '),(23062,2711915,'19,January,2004','2004-01-19 00:00:00','   Wahoo... now I\'ll never get any work done.  '),(23063,2711915,'19,January,2004','2004-01-19 00:00:00','   Welcome to the 35MPH blog.  '),(23064,2711915,'23,February,2004','2004-02-23 00:00:00','   It\'s a black and white neg that I reversed the values of. ( the guitar is black ) Then I added a little yellow and some blue just for effect. I\'m going to be experimenting a lot more intensely in days to come. I\'ll post the results.  '),(23065,2711915,'19,February,2004','2004-02-19 00:00:00','   So is that image just a scan of the actual negative? It looks pretty good. The return of the Elevator People!!  '),(23066,2711915,'19,February,2004','2004-02-19 00:00:00','   I\'ve been playing with double exposures lately. Here is one of the early results. It\'s quite easy to do on the Canon Rebel. It\'s a rough copy with some rough digital polishing. I just wanted to post it to get some feed back.   urlLink statueguitar_scaled.jpg   By the way, how\'s everyone doing?  '),(23067,2711915,'19,February,2004','2004-02-19 00:00:00','   Nice Bennett... keep the ball rolling and keep sticking your big, fat wide-ass foot in all those doors. Commercials and such would be a good departure point for 35MPH to get some real exposure as a full-fledged company. Until all of us are in the same area again (who knows when that will be?) we should start a rotating location pow-wow. We can start in NWA then to Austin and then -- wahoo -- Jackson! What do you guys think? One rotation per year, meet three times per year. We could really get some work done.  '),(23068,2711915,'16,February,2004','2004-02-16 00:00:00','   Just got another production job. Jones Productions from Little Rock hired me to help shoot a commercial in Bella Vista. This is the largest and oldest film and video company in Arkansas. It pays $150 for two hours.  CJ, Emery and I had a talk the other day about shooting some \"knock around\" footage for the Frisco and/or IHCC. This footage could be mere practice or used for televised advertising. I think businesses with the right kind of character would like to use a small production company to create unique, one of a kind commercials. I think, once we\'re all in the same area, we could make some serious start up capitol shooting these small projects. We\'ve all talked about music videos before and I still want to do that as well. These two types of things could be a nice sized bite to chew for our early \"money-making\" endeavors.  '),(23069,2711915,'16,February,2004','2004-02-16 00:00:00','   His GL2? That\'s like 1000 bucks...nice work. Anyone we should know and are you hooking up with the daughter?  '),(23070,2711915,'14,February,2004','2004-02-14 00:00:00','   Just got a gig from a soccer Dad. $50/hour for shooting high school girls and boys soccer games on a Canon GL-2. Pays cash, 10 home games in a season. SCORE!  '),(23071,2711915,'14,February,2004','2004-02-14 00:00:00','   Some Photoshop  urlLink goodness .  And it is, believe it or not, snowing in Austin. These pictures confirm it:  &rarr;  urlLink Snowy Valentines  &rarr;  urlLink Off the Balcony  &rarr;  urlLink Apartment Parking Lot  '),(23072,2711915,'12,February,2004','2004-02-12 00:00:00','   I know, I know... I\'m only about two or three months behind the times. So shoot me. Here\'s some incredible fan filmmaking. Enjoy.  &rarr;  urlLink Batman: Dead End  '),(23073,2711915,'12,February,2004','2004-02-12 00:00:00','   Boy, I wish I could read my own writing...  '),(23074,2711915,'11,February,2004','2004-02-11 00:00:00','   I was unable to find the storyboards we made. Do any of you guys have them?  Here are my story notes:  urlLink 1 ,  urlLink 2 ,  urlLink 3 ,  urlLink 4 ,  urlLink 5 ,  urlLink 6 ,  urlLink 7 ,  urlLink 8 .  And here are Heath\'s story notes:  urlLink 1 ,  urlLink 2 .  '),(23075,2711915,'11,February,2004','2004-02-11 00:00:00','   Okay, here they are:  &rarr;  urlLink Untitled, first cut  &rarr;  urlLink Blackbird, test intro   I didn\'t realize this originally, but the version of  Untitled  I posted is the first cut that we did for initial submission into the film festival. After the deadline, we were given the opportunity to update the short. The changes were minor, but included closing credits and different editing around the shots of Bennett being thrown down the stairs. As soon as I get a larger harddrive -- fingers still crossed for PC Guru -- I\'ll post the second cut.   Blackbird  scans should be following shortly...  '),(23076,2711915,'10,February,2004','2004-02-10 00:00:00','   The software I use for stopmotion is  urlLink Stop Motion Pro . It\'s not free ($89 for educational), but the trial version is full-featured. Only drawback is a red line they burn into the video about 1/3rd down from the top of the frame.  Here\'s your  urlLink one-stop source  for a lot of information on stop-motion in general.  And here\'s a  urlLink good source  that refers to Stop Motion Pro as long as other alternatives (many of them free).  '),(23077,2711915,'10,February,2004','2004-02-10 00:00:00','   I have some negs I can scan and post (with some help) from the Blackbird shoot(s).  Troy, could you get me the software that turns digital photos to video? I\'m going to try some stop motion with a Nikon D100 and high powered flood light kit from work. Do you have any suggestions on what file format to capture the frames in?  As for weekend excursions, I\'m ready whenever.  I can easily arrange for a long weekend if I have a couple of weeks to arrange it. Let\'s get something going.  '),(23078,2711915,'10,February,2004','2004-02-10 00:00:00','   As for the font - it was some weird script we used. EXCEPT for the  a . Remember we didn\'t like the  a  that was in the script so we went through all the other fonts on our computers until we found one that we thought looked good - and no, I don\'t remember that one either.  Troy, by all means, post everything you\'ve got. And I am very serious about getting together SOON. I can take a long weekend to come down to NWA or Autsin (my preference would be Austin). We need to start the machine back up!!  '),(23079,2711915,'10,February,2004','2004-02-10 00:00:00','   Damn... I didn\'t get stuff on last night either. I was up until 230am, just ran out of time to compress and upload the final product.  For  Untitled , I pulled a copy off of tape onto my machine -- I\'m not sure if it\'s the right copy because it doesn\'t have any titles (didn\'t it have titles at the end?). Unfortunately, when we did the original edit on the iMac in the computer lab it encoded the audio in a format that can\'t be transferred off of the tape via Firewire. So, I had to capture the video then capture the audio using the line-in of my sound card. Then I had to re-synch the two (since they were captured separately and the audio didn\'t have any timecode).  Finally, I got all of that done. Last step was to de-interlace the footage and letterbox it. I didn\'t mess with color correcting it or anything because I was running out of time (a few obvious tries didn\'t look very good). All that remains is to compress the video down to a downloadable format, which is actually happening at home right now. Of course, we\'ll have to wait until this evening before I can upload it to the website... so check back then.  For  Blackbird , I captured it. Didn\'t have any problems (probably because I used my old PC\'s Firewire instead of an iMac to do the editing). I\'ll compress it as well and upload this evening.  I gotta say, guys, watching the movies made me pretty nostalgic. And I gotta say... we didn\'t do half-bad. The  Blackbird  shot is pretty cool (and it\'s just a rough, first attempt!).  So,  Watch This Space . Check back this evening for new video footage.  BTW, would you guys be interested in me posting scans of the storyboards and notes I have for  Blackbird ?  '),(23080,2711915,'10,February,2004','2004-02-10 00:00:00','   As I may have mentioned before, my harddrive crashed while I was in NWA over Christmas. Since the first of January my harddrive has been in the hands of PC Guru, a data recovery expert. It\'s major surgery guys. It\'s been weeks that they\'ve been working on it.  So, I unfortunately do not have the  Blackbird  title cards. I also will have to track the font back down that we used (anyone remember its name?).  '),(23081,2711915,'09,February,2004','2004-02-09 00:00:00','   SLAVE1 has returned from the ashes. I still have to reinstall drivers and restore settings tonight. It appears that the \"MYDOOM\" virus did not claim any of my important data. I should be in full working order this evening, just in time to begin playing with some 35MPH media. I hope to post a full report on the video production from last week very soon.  '),(23082,2711915,'09,February,2004','2004-02-09 00:00:00','   Thanks for the compliment on the image... to be honest, it was only about an hour\'s worth of work, and far from one of my favorites. But it came out nicely. I forced myself to use only photos I had taken with my camera. In that photo are several elements: the cooking element of a stovetop, the stairwell of a parking garage, the asphalt of an alleyway, and the shaving mirror from my shower. They were combined in layers, using whatever blend modes looked interesting. I built the focus around the the stairwell picture, which would probably make a decent picture in its own right.  I\'ll try to get the  Blackbird  and  Untitled  stuff online tonight.  '),(23083,2711915,'09,February,2004','2004-02-09 00:00:00','   Also, I was wondering if Troy could put up the opening sequence from  Blackbird  up. I would really like to have the scripted title on black background image.  '),(23084,2711915,'09,February,2004','2004-02-09 00:00:00','   I very much like the \"strange\" image Troy made. More along the lines of what I was thinking about. You know, we really need to get together and have a sit down. Come up with some new ideas - expound on the old. What do you think? Is it already really hot in Austin in May?  '),(23085,2711915,'09,February,2004','2004-02-09 00:00:00','   Okay, finally here\'s the post that was scheduled for Christmas... the  urlLink Christmas 2003 Stopmotion Greeting Card ! And this one, like any good DVD, comes loaded with extras:  &rarr;  urlLink Timelapse footage  of the set being constructed by Elizabeth. You\'ll notice that I pop in occassionally while taking behind-the-scenes photos and doing storyboards. &rarr;  urlLink Behind-the-Scenes Gallery , photos taken with my Canon Rebel. &rarr;  urlLink Storyboard, First Draft , written while on a plane somewhere over North America. &rarr;  urlLink Storyboard, Second Draft , written on a Saturday afternoon. &rarr;  urlLink Storyboard, Third Draft , written mere moments before we began shooting on a Sunday evening.  '),(23086,2711915,'09,February,2004','2004-02-09 00:00:00','   A strange  urlLink image  I put together last night...  '),(23087,2711915,'06,February,2004','2004-02-06 00:00:00','   Just a quick update on things...  I\'ve put together most of the  Christmas Animated Greeting Card  package. This includes behind-the-scenes photos, set construction time lapse, and three drafts of the storyboards. I wasn\'t able to put it online last night as I was having some problems with our website, but that seems to be cleared up now. Eventually this will all be put on DVD (along with the  Halloween Animation  and  Untitled ) as a 35MPH compilation...  I\'ll try to get  Untitled  online this evening, as well as a gallery of full-resolution screenshots for you guys to play with.  '),(23088,2711915,'06,February,2004','2004-02-06 00:00:00','   Just wondering if Troy could send a few screen shots from Untitled to Heath, CJ and I. I\'d like to take some of this content, as well as original content, and play with them. Maybe we could each try some things and then exchange in a format that can be built upon. This idea can be applied to writing as well, like we\'ve talked about.  Tyson Shareholders went quite well. I worked with DVCAM, video console and rear projection screens.  '),(23089,2711915,'06,February,2004','2004-02-06 00:00:00','   Alright guys, I\'m not usually one to get upset by crap like  urlLink this  but CBS and their executives are going too far. It seems they have forgotten about the equal opportunity for every American to speak their mind. And CBS is regulating it on a national and sometimes global scale. Once you have enough pent-up anger - send them an  urlLink email  and let them know that they and the rest of the Republicans are screwing their fellow Americans.  Here is an example of what I sent them:   The CBS executives must have their hands deep into the Presidents pockets in order to justify such a double-standard, which is becoming more prominent as Election Day nears. As such a large part of the national media, it is your responsibility to be a steward of the airwaves and ensure that all people can voice their opinions - which is a right of the American people. Your refusal to run the MoveOn Voters Fund \"Childs Pay\" ad while running and ad for George Bush\'s Presidential campaign is a slap in the face of all Americans and their right to speak freely. Furthermore, if your network runs Bush\'s Medicare ad - which is full of lies - then by default, you must run competing ads which portray the truth. Don\'t be a platform for a failed and corrupted presidency.  Heath Kuszak Jackson, Wyoming  Get Proactive!!  '),(23090,2711915,'02,February,2004','2004-02-02 00:00:00','   Here\'s a random tidbit.  Cruise lines maintain onboard processing labs.  Contracts run 6 months on average with time off as long as a week in some ports-of-call.  My next adventure?  '),(23091,2711915,'31,March,2004','2004-03-31 00:00:00','   I like it. Of course, like you say, it all depends on the writing whether it\'s Hemmingway-esque or a bad episode of  Saved by the Bell . I\'ll reserve judgement until then.  I think this group could very much understand the characters and the situation. While I wouldn\'t say any of us have been there, I think we could all imagine it quite readily.  It\'s also a very realistic concept in that the production requirements are minimal. Dialogue, location establishing shots, no effects. Unfortunately, the quality of the final product would almost entirely rest on the ability (and confidence) of the actors we had.  As a realistic show piece it would need to fall into the sub-twenty minutes category, which it sounds as if that\'s where you\'d be aiming. This would land it cleanly in the \"shorts\" category of most film festivals (I\'m sure that was your plan).  In regards to flashbacks: personally, I love playing with temporal structure and am envious of directors who do it masterfully. Obviously our past efforts have specifically been built around that concept. While we don\'t want to go overboard, it\'d be fun to break away from the normal by playing an interesting twist on flashbacks.  '),(23092,2711915,'30,March,2004','2004-03-30 00:00:00','   Do you have anyone in mind to play the parts? Two main characters, max?  '),(23093,2711915,'30,March,2004','2004-03-30 00:00:00','   Howdy all!  After several false starts, I\'m off and running on the 35mph blog!  More grabass from my blog, which is grabass from life as a grad student.    Alright...here\'s my life, and here\'s what I would like to do.  I graduate in May, pass through NWArk beginning of June, and then spend June and July teaching creative writing to high schoolers in Kansas.  After that, I\'ve got no job, but lots of plans.  If I find work between now and then, I\'ll probably have a week or two to dick around, but if not, the tentative plan is for me to pull up in Fayetteville for a month or six weeks, and try to find some part-time work in production...work some commercial shoots or try to find a film in the area to PA for.    I also want to make a film, and since this is up and established and looking for work, here\'s my pitch.     Log Line  - A recent college graduate and his drug addict friend take one last road trip through their home state, before the addict unwittingly begins drug rehab.     Story  - Matt, a college grad, heads out to Nashville to start a job and find his life.  With him is Dennis, his oldest friend, who plans to start anew in Nashville again.  But Matt plans a stop in Little Rock to take the unaware Dennis to drug rehab.  The film chronicles the drive, and acts as a Hemingway short story, where they talk about everything but the problem that they both know is a wedge between them.    We\'re looking at 10-12 minutes, mostly taking place along I-40 as I see it, with the possibility of flashbacks...but flashbacks have the ability to take serious topics and turn them schmaltzy, and I\'m interested in chronicling friendship rather than drug addiction...Matt believes this is their last road trip, as everything will be different once Dennis starts rehab.  And he\'s correct.  The quesiton at the crux of this -- how would you say goodbye to a friend?  I\'m working out first draft in my spare time, and wrestling questions about how to make this a drama of the mind rather than an after-school special (which it could easily tend towards if not kept in check).  So that\'s that.  If it sounds interesting, let me know.  If not, I\'ve got another short that could be easier to film.  '),(23094,2711915,'30,March,2004','2004-03-30 00:00:00','   Excellent stuff, Bennett. Photoshop would definitely make it easier. But it never hurts to get your hands dirty. Sometimes the ease of Photoshop let\'s you fly by the better options. In other words, when it\'s so easy to make things complex, the number one rule of \"keep it simple stupid\" often gets overlooked.  '),(23095,2711915,'30,March,2004','2004-03-30 00:00:00','   So, you actually used scissors to cut up some prints? Jonesing for some arts & crafts time, Bennett? Think you guys will want to volunteer in this Film Festival with me? Take some time off and come up here?  '),(23096,2711915,'28,March,2004','2004-03-28 00:00:00','   Troy, you\'ve been an awefully busy guy lately. Sounds tiresome, yet exciting!  Here\'s the project I was working on a couple of weeks ago. Right now it\'s just fragments of photographs placed on white poster board. It took a few hours to do and I think I\'ll do it in Photoshop next time. The flash on my camera and a much too wide lens combined to make the thing look contorted and blown out. The lines are straight in real life. Here\'s a crumby photo of it.   urlLink 35mphlogo.jpg   Let me know what you think.  '),(23097,2711915,'28,March,2004','2004-03-28 00:00:00','   Sorry I haven\'t posted much lately. I\'ve been smashed at work with a full plate. Just a bit of info...  For most of February I was traveling around the country talking to publishers about our latest tech,  urlLink RenderWare4 . It\'s something we\'ve had in development for over two years now, and it still has another year before it will be finished. It\'s aimed at the next generation of video game consoles (PlayStation3, Xbox2) and is a complete re-write of everything we currently offer. Really cool shit and it\'s leaving a lot of people speechless.  March was pretty packed as well. I was at home in the office, but there were a lot of distractions. I\'ve been working with a 3D tools company,  urlLink Alias , makers of the modelling package Maya, to create a \"cinematic real-time\" demo for current and next generation hardware. It\'s been really cool. I believe I\'ve mentioned it before. I was working with art production house  urlLink MondoMedia , who famously have worked on the cinematics for a recent James Bond game,  urlLink 007: Nightfire .  The first milestone for our demo was GDC (which took place last week) and I only got the final artwork the Friday before. Needless to say I spent several long nights over the weekend working on this (until 4am Sunday morning and 5am Monday morning) because I had to get on a plane 930am Monday morning. Unfortunately, I didn\'t get it as far as I wanted to (just the environments were rendering, none of the characters or special effects). Needless to say, I was a bit wasted at the end of the day Monday after nearly 48 hours without a single full nights sleep.  Going on in parallel with the demo at work was  urlLink SXSW  at home. Elizabeth and I (as well some friends of ours, Ben and Jenny) picked up film passes. There were definitely soom good flicks and I wasn\'t disappointed with anything we saw. One of the highlights was  urlLink Bush\'s Brain  which we were at the world premiere of.   urlLink GDC (the Game Developer\'s Conference)  was last week. Pretty cool stuff. We were one of the most dominant forces there with a huge presence on the show floor and meetings with most of the top movers and shakers in the industry.  But beyond the business stuff, the week was quite interesting in the \"after hours\" scene. To kick things off, Sony hosted an invite-only party Wednesday night. I was lucky that my Sony meeting was during the day Wednesday so I was able to get a ticket. The party kicked off as a fairly standard one: more free alcohol than is healthy, a huge assorment of food and a very decent club DJ on the turntables. They had some interesting big-screen visuals with the music, most of them tape loops of random things. The night got really interesting when the main act came on.  urlLink The Crystal Method . These guys can put on a helluva show, even though it\'s just two guys who look like they just arrived from watching football/soccer game down at the local pub. The show was awesome. I was about five feet from the stage and the music was about as loud as I can ever remember music being. I\'m worried that I may have potentially vibrated my organs into different positions with their ultra-low bass beats. I didn\'t have my camera with me, but a friend did and he promised to forward his pics. If you check out their website linked above, under the photo sections, there are some pics from their recent LOB tour where they stopped off in Austin (for SXSW) and the setup looked pretty similar.  The next night had two big headliner parties, one at the \"cocktail hour\" and the other post-dinner. The first was put on by Criterion (that\'s me) and Sammy Studios, a big Japanese video game maker who\'s been building a big American precense (the US division is a customer of ours). The second party was Microsoft\'s Xbox party: Risque Cafe. Unfortunately, Microsoft is not quite as good at throwing parties as Sony. I guess Sony has all of these bands on their payroll (Sony Music, Columbia), whereas Microsoft does not. They just had a DJ (but of course had the required copious alcohol and food). They did have freshly hand-rolled cigars, though because it\'s California, you had to head outside to smoke them (where you couldn\'t take your drinks -- tough choice for some!). They also had free massages, etc. I hung out for an hour or so shooting the shit with a friend while we finished our cigars then I hit the road for the evening.  So, that\'s what I\'ve been up to. I\'m about to dig into doing my taxes. How fun. Mr. Rolin Moe has asked to join our little blog to present an idea, so I\'ve invited him to join. Rolin, take it away...  '),(23098,2711915,'27,March,2004','2004-03-27 00:00:00','   I shot three commercials on Monday. It was \"run and gun\" all day long, but we shot inside a Lasik surgery clinic for eye correction. It was quite interesting and gave me an opportunity to experience fitting the gear for a full production into a tiny space. We also used some creative light tricks to enhance the color of the lasers and such. It was a really fun shoot.  I also may be able to do a little traveling soon. I have lots of loose ends to tie up first, but everyone else at work is done with their vacations, so maybe I could take a little time off soon. Let me know of anything in your respective locations that would be exciting and cool. This doesn\'t apply to CJ, as I know what is going on here all too well.  '),(23099,2711915,'26,March,2004','2004-03-26 00:00:00','   Where did everyone go?  '),(23100,2711915,'16,March,2004','2004-03-16 00:00:00','   Well, its been a while -- regardless, I have some news that may be of interest to everyone. A new film festival -- likened to the Sundances, Tellurides, and Cannes of the film world -- is springing legs in Jackson Hole. Called the  urlLink Jackson Hole Film Festival , it is going to bring top notch indie features and shorts to Teton County, Wyoming. All just in time for the Fall \"off-season\" -- which won\'t be an off-season anymore if this thing sticks around. I plan to attend and probably volunteer. Can\'t hurt to rub elbows with these guys.  '),(23101,2711915,'14,March,2004','2004-03-14 00:00:00','   I just finished a neat little \"negative space\" 35MPH logo. I\'ll try to digitize it at work Tuesday and post it. I\'m sure there\'s a way to do it with Photoshop, but I chose to do it with scissors and glue (the real kind.)  '),(23102,2711915,'11,March,2004','2004-03-11 00:00:00','   There shouldn\'t be any difference between expensive and cheap cards, except maybe additional features and additional firewire ports. Speed also shouldn\'t be an issue as most cameras can only transmit over firewire while playing back, so it\'s single-speed anyway.  The biggest difference is usually whether or not the card also supports other types of capture, or if it can support multiple firewire devices (my old one had three firewire plugs, my new one has two). There\'s also a new FireWire800, which is twice as fast as standard firewire and supports cable lengths up to 100 meters. It\'ll probably be useful for HD signals.  '),(23103,2711915,'04,March,2004','2004-03-04 00:00:00','   I just acquired a Firewire card. It was used briefly but should work perfectly. I\'m going to begin capturing tonight and see how it works.   Question: What is the difference between cheap and expensive cards? Speed of capture? quality of video file?  '),(23104,2711915,'04,March,2004','2004-03-04 00:00:00','   Firewire is definitely going to be the best way to go for video. It\'ll come out of the camera in a lossless fashion, and at a fairly fast rate.  If your computer doesn\'t have a firewire port then you\'ll need to buy a firewire card. Should be in the ballpark of $50 - $75, and they usually come with some simple video editing software (similar to iMovie, which is what we used to edit  Untitled ).  You can also get 30-day evaluations of Adobe After Effects (motion graphics, think Photoshop for animation/video) and Adobe Premiere Pro (which is essentially the 7.0/CS version of Premiere). The 30-day evaluations are fully functional with no strings attached.  If you want, I can dump all of the Untitled stuff (and Blackbird) onto a DVD and send it your way. Does your PC have a DVD-ROM? If not I\'d have to send it on Digital8.  '),(23105,2711915,'02,March,2004','2004-03-02 00:00:00','   I just installed some video editing software at home. What is the most affordable way to transfer from MiniDV to my computer and back to MiniDV or VHS after editing? Do I need a Firewire card or something similar? I\'m ready to start practicing and I\'d like to be fluent and fast with Premiere (and anything else I can get my hands on.) I\'d also love to take  Untitled  and chop it into little pieces and put it back together again.  '),(23106,2711915,'02,March,2004','2004-03-02 00:00:00','   Interesting Bennett...  \"Do not try and bend the spoon. That\'s impossible. Instead... only try to realize the truth... There is no spoon...Then you\'ll see, that it is not the spoon that bends, it is only yourself.\"   Maybe we should try some steak knives. I got a new set of Cutco!  '),(23107,2711915,'02,March,2004','2004-03-02 00:00:00','   Here\'s a concept I came up with the other night. I\'d like to explore it further, but I wanted to post my early results. I was thinking of printing a unique font backward and upside down instead of the pencil drawing. Let me know what you think.  &rarr; urlLink spoon_websize.jpg  '),(23108,2711915,'30,April,2004','2004-04-30 00:00:00','   I haven\'t been able to read the second script. The link is not working for me.  The first and the last are working fine. I lean toward the third for this summer, though I need to read the second. I\'m going to print it out this weekend and mark it up and send some of my ideas, so stay tuned.  '),(23109,2711915,'29,April,2004','2004-04-29 00:00:00','   I haven\'t had the time or inspiration to work on the short I pitched you all earlier, but I\'ve got several first draft shorts written that might suffice for production without requiring driving shots.     urlLink Music Short Film  (approx. 3 min)  urlLink Relationship Drama Short Film  (approx. 9 min)  urlLink Character Drama Short Film  (approx. 11 min)  The first is a short I wrote for the Coca Cola Student Filmmakers scholarship, required to be shot in 50 seconds and deal with the movies.  I thought it was very good, but the content might not be family appropriate.    The second is a short I developed last year, under the question \"How can you love two people at the same time?\"  Not a whole lot of people I\'ve shown it to have understood it as I want, but I think it would be interesting on film.    The last one is a recent project, based on a moment in Harold Pinter\'s  The Dumbwaiter .  I finished the first draft tonight, in lieu of working on my thesis feature, due to inspiration and a recent harnekning to the opera  Faust .  There are a few character issues, and the turning point might not come off as strong as I hope.    Read \'em.  See if you like the general idea of the story.  If one of these stands out, we can try to adopt it as a summer project...do a few rewriters, work out equipment and production.  The first and last could be shot in one night, albeit a long night.  The second one would require two or three days of shooting.    And be honest in your criticism.  These are first drafts, so they aren\'t meant to be perfect, but if something isn\'t working, it won\'t get better if you don\'t bring it up.  '),(23110,2711915,'24,April,2004','2004-04-24 00:00:00','   My trip to Arizona was a success. Selected photos to be posted soon. Troy, sorry I didn\'t make it down to Austin, but the road had worn away my patience enough for that week. I hope to make a trip down there soon.  '),(23111,2711915,'17,April,2004','2004-04-17 00:00:00','   Okay,  urlLink TroyGilbert.com  is now live. This blog will be tracking me through the games industry. In particular, it will be tracking a new independent game project I\'m starting.  '),(23112,2711915,'06,April,2004','2004-04-06 00:00:00','   Some pics I\'ve recovered from my harddrive. A bit show some corruption. Some of them were damaged permanently. I figure I need to get them up on the big archive in the sky before they\'re lost forever (not that they\'re necessarily worth keeping).  &rarr;  urlLink bloodbath  &rarr;  urlLink blue  &rarr;  urlLink coral-reef  &rarr;  urlLink crazy-texture  &rarr;  urlLink fall-wall  &rarr;  urlLink fungus-on-tree  &rarr;  urlLink glass-bead-abstract  &rarr;  urlLink green-red-virus  &rarr;  urlLink lined-splatches  &rarr;  urlLink microbia  &rarr;  urlLink pop-noise  &rarr;  urlLink prototype  &rarr;  urlLink rainbow-confusion  &rarr;  urlLink retro-future  &rarr;  urlLink rusted  &rarr;  urlLink tentacles  &rarr;  urlLink ufo  '),(23113,2711915,'01,April,2004','2004-04-01 00:00:00','   I\'m keeping my fingers crossed that I\'ll get a nice graduation bonus from some family members come May.  If it happens, I\'ll probably invest it into an XL-1 (selling for $2G on ebay these days, including XLR adaptor and other periphials), as my career plan is to write screenplays and shoot documentaries.  Regardless, if we\'ve got a camera body (whether it be one we own or one we rent), borrowing lenses shouldn\'t be a problem...it might be the only thing that my UT film degree will come in handy for (getting free shit in Arkansas).    My thoughts on the car were either an old beast wagon, kind of like the one I used to drive...or something like the Suburban that Chris Cousins flipped in Alma.  But it doesn\'t matter...it\'s basically what we can get our hands on.  I\'ll write it as a Volvo, if only because it\'ll be more visual on the page.    And I won\'t worry about car shots.  Sounds like we\'ve got it under control.  I\'ll just write it to write it and leave the logistics to the professionals.    I\'ve shot a few films in my day, so I\'ve got a pretty good heads up on sync sound and lighting technique.  And I\'ve got a pretty broad stage theatre background, as far as actors are concerned.  That being said, as the writer, I will leave camera angles and cinematography and director\'s intent and really everything past original story completely to the board of supervisors.  Nothing worse than a writer/director auture.    Timetable (because it never hurts to think about it early) for me - in Arkansas June 1 to 8 for pre-production meetings.  Then in Arkansas August 1 to ??? (probably Labor Day at the latest), for time to shoot everything.  My presence is not imperative, but I wanted those who live out of Rogers to know what my schedule was in case we want to coordinate.  '),(23114,2711915,'01,April,2004','2004-04-01 00:00:00','   First: I may have access to the Canon GL-2 through the summer.  Otherwise, I could borrow an XL-1s from work (if I leave my soul as a deposit.)  Second: My first ideas for shots called for camera vehicle driving next to subject vehicle. Though we couldn\'t do this for the whole film, it could look nice. I believe my Cherokee would support a camera man and equipment on it\'s roof if we make some sort of platform to stand on. This could provide for some neat and cheap driving shots.  Third: I just inherited $5,000. This is not a joke. Though a lot of it is spoken for, I will entertain ideas for 35MPH investments. I have purchased a nice tripod and a fluid head. I don\'t know if we\'re ready to invest in a camera, but I\'ve already mentioned options for that.  I\'m extremely excited about this.  '),(23115,2711915,'01,April,2004','2004-04-01 00:00:00','   how about this... we find a car - preferably one of those old boats that our parents / grandparents drove back in the day - that has a massive dashboard clearance from the windshield. Enough to get the lens of the camera (another good question, what camera?) 1.5 to 2 feet away from the subjects and then use a fisheye lens to capture all of the in-car dialogue between both front seats. This might be a cool effect and would also allow us to show what\'s going on in the periphery (passing cars, etc.) while going at real speeds - maybe not 70 mph but not 35 either.   What sort of vehicle did you have in mind, Rolin?  '),(23116,2711915,'01,April,2004','2004-04-01 00:00:00','   After watching Eternal Sunshine of the Spotless Mind again, there might be a good way to incorporate the flashbacks into something like this.    I\'m a writer, and I think specifically in the story...what steps do we take to evoke the wanted feeling, plotwise?  How that is directed doesn\'t matter to me, but looking at how you\'d shoot a story like this, it would be best for us to spend as much time out of the car as possible, mainly because we don\'t have the production mechanism to shoot many driving shots.  We can shoot the driver very easy, but if they both sit in the front seat, that\'s a bitch to shoot, and if one sits in the back, capturing both in the frame is a bitch as well.  Hollywood rigs the camera up afront the car (or on a trailer behind), and we could do something like that (more back-window shooting rather than rigging, and knowing how bumpy that damn road is...), but it would involve going 35mph in the right lane.  Passing cars would kill the idea that we\'re going faster than we are.    So flashbacks might be the best way to go with this.  I can write it two ways...idea would be that we film the best written one, but hopefully the easy one to film is as good as the more difficult one.    As far as actors, I have no clue.  I don\'t know any real actors in the area.  I\'m sure we could grab a few UA students to do it.  I don\'t mind working in-house, but the two main roles would be best done with professionals. Putting ourselves in such a project is best reserved for cameos, so we can focus on production measures.    I\'ll start drawing things out over the next few weeks, starting with a character sketch, then outline, and finally draft.  '),(23117,2711915,'19,May,2004','2004-05-19 00:00:00','   Hey there, sorry I haven\'t posted in so long, been real busy at work.  As you can see, there\'s a new look to the site and a lot of new features. Well, I guess two new features: comments and profiles. These are new features of Blogger. Blogger has been revamped. Much cooler, check it out. They\'ve also added some new templates, this being one of them. I may tweak it up a bit given the time. If anyone has any suggestions let me know.  I\'ve also updated  urlLink TroyGilbert.com  similarly. Check it out as well.  I encourage everyone to update their profile with some info. And also start adding titles to their blog posts (if you want to).  Cheers!  '),(23118,2711915,'25,June,2004','2004-06-25 00:00:00','   but its worth a try since I can\'t use my phone to call until my free minutes start at 7 PM (my time, mountain time); and if I wait \'till then, I probably won\'t get a hold of you.  H A P P Y   B I R T H D A Y   C H U C K !!!   and also, because I will not be able to access a keyboard on Monday (I\'m taking my parents into Yellowstone for 2 days, wahoo!), a shout out to another June baby.   H A P P Y (early) B I R T H D A Y   T R O Y !!!  '),(23119,2711915,'01,June,2004','2004-06-01 00:00:00','   I\'ll be in town June 7.  I\'ll also pass through over the 4th of July weekend, and then hopefully some in August.  Anyone around?  '),(23120,2711915,'31,July,2004','2004-07-31 00:00:00','   How much music and sound do you guys think we\'ll use? Do we have an acceptable way to record it? perform it? Do video cameras or cheap computer microphones record well enough to get away with? I would be very interested in playing with sound effects if we have time to cram it in.  '),(23121,2711915,'30,July,2004','2004-07-30 00:00:00','   How much music and sound do you guys think we\'ll use? Do we have an acceptable way to record it? perform it? Do video cameras or cheap computer microphones record well enough to get away with? I would be very interested in playing with sound effects if we have time to cram it in.  '),(23122,2711915,'30,July,2004','2004-07-30 00:00:00','   In case you guys haven\'t checked your e-mail or seen  urlLink my blog  (recently updated the look)...  BTW, I\'m transferring the domain names to another registrar. No changes at all on the site end (still the same servers), but there may be a 24-48 hour period where the site goes down as the name records are updated. Of course, anything *.35mph.com would be affected, so e-mail would be out for a while as well. It\'s likely not even going to register anything since the actual servers don\'t change. Just FYI...  '),(23123,2711915,'27,July,2004','2004-07-27 00:00:00','   So, as a working title, we\'ll call it the  Labor Day Project.   To get the most work done, we\'re going to need to enlist local professionals. Well, at least local top-ranked amateurs. Okay, okay... locals I happen to know! Elizabeth will of course assist in any way she can. Additionally, a guy I work with is interested, Ben, and his girlfriend Jenny would probably be interested as well (particularly if Elizabeth was participating). Both of them have art degrees so their interests definitely line up with our needs.  That brings our potential crew to:   Heath, Ben and myself  Elizabeth and Jenny (if we need them)  Bennett and Rolin (and Chuck?) if you guys can be here   That\'s between 3 and 8 people. That\'s a good size crew and would be very, very cool to have all of that input.  Now, onto what we\'re going to do...  Well, I don\'t have an answer, but I have a suggestion (sorry, this is a style suggestion not an actual plot suggestion). Along the lines of what I was mentioning previously about mixing medias and techniques, animation, live action, etc.  The drawback of video is that, well, it looks like video. Not so much of a problem these days as people are getting used to it, and video can be treated on the computer to look much nicer anyway. But, what I\'d love to do would be to work in a format with a higher resolution than video. While we can\'t afford film, we can maybe approach that quality level through some clever combinations of tech that we can afford.  I think we\'re all familiar with digital cameras. Even my simple digital camera can take very high-res pics (4+ megapixels), and Bennett likely has access to cameras that can take both incredible pictures and high-res (6+ megapixels). Compare this to video, which is in the ballpark of 1 megapixel.  One technique I\'ve been wanting to experiment with is mixing animation with photographic sources. This is called rotoscoping. Rotoscoping basically involves taking some source live action footage (film, video) and then drawing over each frame. See  Waking Life  for an example of this (and see previous posts on the subject). What I\'m proposing is not exactly the same, but its close.  Using a digital camera, you snap a high-res picture of your location. You then video your actors doing the scene. In post-production, you take your high-res picture and clean it up, color balance it, whatever (give it a painterly effect, apply special effects, whatever you want in photoshop). You then take the video and rotoscope the characters. For the rotoscoping, you use something like a Wacom tablet and trace each frame of video to whatever detail is desired. You can then toss the video and drop the rotoscoped characters onto your clean backdrop.  This should give a very interesting look that combines realistic backdrops with very dynamic, organic actors that are hand-drawn animated. I\'m going to try to do a simple test this weekend to see how this looks.  Some advantages of this approach:   It\'s unique, thus it instantly makes whatever we make stand out from the crowd of amateur shorts out there.  It improves the quality of what we can produce. It should provide much sharper, cleaner visuals than video because we can essentially work at film (or at least HD) resolutions. Also, the result is never compressed significantly so the colors all remain pure and sharp.  Because it is already very manipulated, we can manipulate it much further without it looking like a bad special effect. You can make an animated character \"explode\" and it look \"correct\" in the character\'s world; if I have live action footage and try to make a character explode it would require very polished special effects to not be distracting.  It opens up a wide spectrum of shorts we could possibly make. Because the result is so stylized, it begs that you use it on very stylized material, i.e. fantasy, sci-fi, psychadelia, Twilight Zone kinda things, comedies, anything you can imagine as a cartoon.   So, just throwing that idea out there. I understand if everyone would prefer to do something a bit more traditional. I\'d also understand if you guys think this may be too ambitious. But, considering how often we seriously do this stuff I\'m tending to go whole hog on this...  And finally, my actual project suggestion if we wanted to go with a technique like above: what about doing something based on the work of HP Lovecraft? A modern interpretation, maybe?  Just a thought...  Okay, quickly, be...  *bzzzt*   ... damn, didn\'t finish the post before my battery died (I\'m sitting at a conference in Seattle on making Windows games). But, fortunately, a power-outlet was nearby.  Let me know what you guys think!  '),(23124,2711915,'23,July,2004','2004-07-23 00:00:00','   Ok, so things shuffled around a bit on Thursday, but here\'s finalized information for my new residence and phone.   Bennett Ryel (my name is still the same)  506 N. Oak St.  Little Rock, AR ?????   501 .351.0745   Labor Day: I\'m not sure yet what my schedule will look like then, but the Austin Cram Session is my FIRST priority.   Story Ideas: this idea has popped into my head a few times in the last few years. Imagine a near-future society where a pill has been developed to allow humans to digest and break down fiber molecules into glucose. Entire nations of people with access to this pill could take it every morning and then munch on grass, cardboard or even notebook paper for their nouishment.  '),(23125,2711915,'22,July,2004','2004-07-22 00:00:00','   The time is nearing, folks. We really have to do something meaningful and productive because, if we don\'t, 35mph will turn into 0mph. So, excuse me if I am taking liberties here Troy, this coming Labor Day weekend a 35mph rendezvous will occur in Austin, TX. The proverbial ball has almost stopped rolling and we need to get it moving again. Anyone who is interested - Rolin?! - is welcome but come prepared to have ideas flushed out and be ready to work.   Cake Please!  '),(23126,2711915,'18,July,2004','2004-07-18 00:00:00','   &nbsp;All sorts of exciting things fell into my lap recently.&nbsp;I bought a used road bike to feed my growing cycling obsession. This bike is awesome. I also acquired a Nikon F100 for a steal of a deal through work.  &nbsp;The biggest news of all is that I accepted a job supervising and managing the Bedford\'s&nbsp; lab facilities in North Little Rock as well as West Little Rock. I was there this weekend finding an apartment and I think I\'ve scored a little industrial loft right in the middle of downtown. I\'ll be moving on the 23rd.  &nbsp;All these things happened within 48 hours of each other.  '),(23127,2711915,'06,July,2004','2004-07-06 00:00:00','   I recently encountered a long lost friend. Kylie Neal came up to Rogers from her home in Little Rock to visit me and we began dating (I\'m bad with the titles, but I think that\'s what you call it.) I also have started playing music with Jackson Jennings at Club Frisco on Friday nights, as well as random weeknights when we feel inspired. Both of these offer an outlet of sorts for my restless mind. If any of you are in the neighborhood on Friday nights, stop by and say hello to the band and to Kylie, who usually hides from her degree-having, good-job-working, financially stable life at my no-cable having, crap-job-working, starving-wannabe-artist-poorass apartment. I hope you all are well. By the way, if you haven\'t checked out Hunter\'s website, you should. Mad Props to H.Mack. Word.  '),(23128,2711915,'08,August,2004','2004-08-08 00:00:00','   I think I have found a story that is extremely adaptable...H.P. Lovecraft\'s \"The Outsider.\"  http://www.deliverance.mcmail.com/lovecraft/outsider.htm  I have forgotten how to format text in the blogger, sorry.   Heath  '),(23129,2711915,'05,August,2004','2004-08-05 00:00:00','   Ok, so the first week of August is almost over. Labor Day weekend is going to come chugging around the corner before we know it. I think that we need to pick a story fairly soon and go with it so we can start brainstorming how and where and what to shoot for this thing. Rolin and Troy have mentioned that they both have some ideas on where to start in terms of translating a HP Lovecraft , William Gibson or Richard Powers work into a short. I think everyone involved (that is, three or four of us) should be able to agree on one of these and begin to develop ideas on how we can make it into a 35mph Production.  And as far as I know a preliminary head count for that weekend - someone correct me if I\'m wrong - Myself, Troy, Bennett.  Rolin - will you be there? What about CJ? Who else? I guess we shouldn\'t have too many people...thoughts?  '),(23130,3400678,'31,May,2004','2004-05-31 00:00:00','   The drugs began to take hold.... We were somewhere We were somewhere We were somewhere When it all took hold We weer somewhere This story must be told RED SHARK!!! Mescaline, Lsd, only missing PCP Only exsisting to exist Only looking for the american dream we were somewhere (spoken by Zaph) As your attorney I advise As your attorney I advise As your attorney!!! I ADVISE!!!!! Back door Beauty? Back door beauty? Lizards and a twenty foot nazi One hand WASHES THE OTHER!!!! I FEAR NOTHING!!!! We were somewhere We were somewhere We were somewhere around barstow when the drugs began to take hold.... Back door beauty??? Honkey FAGGOT I CANT HAVE THAT!!! I ADVISE I ADVISE I ADVISE I ADVISE I FEAR NOTHING!!!! Back door beauty....  '),(23131,3400678,'30,May,2004','2004-05-30 00:00:00','   (quazi spoken): If I\'m lucky my neck well snap If not its suffacation maybe this is what i need An eternal vacation  I\'m sorry if its hard for you to understand Im sorry if you dont realize I am but a man  I\'m sorry if you realize that there is little i can do I try my best, do what i can, do everything for you  I wish I had the power to make all the pain go away I wish i had the knowledge and the right thing to say  I wish I had some magick, a special little spell I wish you knew no evil, i wish you knew no hell.  Follow me, i cant promise, its the right way to go I can only say: THIS IS THE ONLY WAY!!!!  If I\'m lucky my neck well snap If not its suffacation maybe this is what i need An eternal vacation  I am a dumb little boy I am useless I am just a Toy So not completely useless  I am just a stupid kid i dont know what im doing What i give is what i get What i give i\'m losing  People all around say that im insane  People all around claim that Im insane  people all around ME!!!!! Claim that im INSANE!!  maybe im insane out of my mind Im just dumb and trying to find some happiness  I\'m just a stupid kid a dumb little boy  If I\'m lucky my neck well snap If not its suffacation maybe this is what i need AN ETERNAL VACATION!!!!!  Stupid boy foolish boy stupid kid such an idiot  I\'m sorry if its hard for you to understand Im sorry if you dont realize I am but a man  I\'m sorry if you realize that there is little i can do I try my best, do what i can, do everything for you  I wish I had the power to make all the pain go away I wish i had the knowledge and the right thing to say  I wish I had some magick, a special little spell I wish you knew no evil, i wish you knew no hell.  Follow me, i cant promise, its the right way to go I can only say: THIS IS THE ONLY WAY!!!!  '),(23132,3400678,'28,May,2004','2004-05-28 00:00:00','   Single mother was my ma \'cause my father, he ran off he just packed up and left and left me, his little mess no father figure no need for one forever and always a bastard son no father figure  no need for one forever and always a bastard son My father, met another now she is, my step mother and they have a baby boy he treats him nicely, buys him toys im up here, he tries to visit offers me money i dont resist it some retribution for running away he gives me money he owes, he pays but money doesnt fill the void and ive always been a bit annoyed by the fact a man can leave his kid maybe he\'d come back.... he never did No father figure no need for one forever and always a bastard son dont call him \"dad\" cuz hes not one to me hes just a  broken condom im an accident so just leave me im an accident  so let me be i dont need a family I dont need your charity i dont need your sympathy I just want some clarity wish  you took responsibility I\'ve got along through life so far and when im sixteen youll get me a car maybe to replace a year or two of lost time of lost love of lost memories but it wont help your just a check  Just a check No father figure, no need for one always and forever a bastard son  '),(23133,3400678,'25,May,2004','2004-05-25 00:00:00','   *slaps Twik*  '),(23134,3400678,'24,May,2004','2004-05-24 00:00:00','   Firetruck yeah im talkin about firetrucks Myk loves firetrucks I love frogs Twik loves guys, but thats okay TWIKS gay! Yeah but thats okay Fire trucks and frogs And Mitch hedburg Loves koalas! I Love frogs! Myk loves......Firetrucks! AND VIRTUOSITY!!!!! AND SNORTING SUGAR!!!!! but i love frogs dont even act like i didnt buy that firetruck i put it in a frog and twik was all like (spoken by twik) Im so gay! FROGS FIRETRUCKS!!!  '),(23135,3400678,'24,May,2004','2004-05-24 00:00:00','   This is a very serious subject.....pants..... Pants(repeat 30 times) And thats all im really trying to say!!!!! PANTS!!! PANTS!!!! (spoken by twik) Pants???? NO!!!! PANTS!!! PANTS!!!!  '),(23136,3400678,'24,May,2004','2004-05-24 00:00:00','   This is the song, Myk made up for me, To sing very, Dramatically, I cry for you and your lack of pants Fooooooorrrrrrrrrrrr youuuuuuuuuuu, Need Some Pants You Need Some Pants(repeat for 2 and a half minutes) Seriously, dude. Put some damn pants on.  '),(23137,3400678,'22,May,2004','2004-05-22 00:00:00','   I need some pants.(repeat 14 times) (Spoken by Twik) ummmm.....turtles? NO!! I need some pants.(repeat 15 times)  '),(23138,3400678,'22,May,2004','2004-05-22 00:00:00','   I wish I had a bag of boxes. If I had a bag of boxes I would Have a bag of boxes in each box would be a bag in each bag would be more boxes And this would go on for ever and eternity and always yeah me and my bag of boxes Now i have a bag of boxes ANd inside this Bag of boxes there are more and more boxes and more bags and boxes Bags!  BOXES!!! (spoken by twik) ....ummmm Barrels? NO!!!! Bags and boxes Bags and boxes We need not barrel barrel need not apply im a bag and box kinda guy I wish I had a bag of boxes forever with my Bag of boxes eternity with my bag always with my bag of boxes Bag-Of-Boxes-Be-infernally-eternal! B-O-Boxes Bag-O-B\'s Bag of boxes be infernally eternal I wish I had a bag of boxes. If I had a bag of boxes I would Have a bag of boxes in each box would be a bag in each bag would be more boxes And this would go on for ever and eternity and always yeah me and my bag of boxes Bag!!!!! of BOXES!!!!!  '),(23139,3400678,'22,May,2004','2004-05-22 00:00:00','   There are slits on your wrist  while i yearn for your kiss and you claim that nobody loves you I wish you understood that i can be good I wish i could make you see I want you do you want me? There are tears in your eyes I just want you to smile you say your alone ill be here for awhile I might not exist I might not matter But i\'m here when you need me youll see i want you do you want me? We both cried  when we lost each other and when i got back you were with another better then me and smarter then me and i dont wanna stay  Jack without Sally Im sorry, truly you say you are too. but memories just dont go away they just stay and make things worse so much time spent was it all in vain? You have slits on your wrist  whilest i yearn for a kiss and you claim no one wants you yeah im here for you always Jack without sally yeah im here for you always Jack without Sally  '),(23140,3400678,'22,May,2004','2004-05-22 00:00:00','   Love, Fuck it, fuck it hard, listen to me your freindly bard (repeat twice) Fuck you! My emotions are your toy. Fuck you! I guess im just a dumb little boy. You were never my girl. Fuck you.... Make me wait oh so long, there for a moment then your gone. I just need some time alone. I just need to find a home. I just need someone to hold. I just need to be told....those three little words.... I just need some time alone... I just need to find a home. I need someone to hold..... I loved you, but fuck you, but fuck you....fuck you. Fuck you. I just need some time alone. I just need some time alone. I just need some time alone. I just need to find a home. I just need someone to hold. I just need to be told....those three little words....  '),(23141,3400678,'22,May,2004','2004-05-22 00:00:00','   Geni Sauce By Zaphlon ( zac roberts)  Part One.... Genesis This is a story A completely true story None of itâ€™s made up Except for the parts that are Verse 1 Long ago when Evil cattle lived in Atlantis They ate lots of broccoli and hunted wooly mammoth  But the Evil Cattle king had no Evil cattle Queen He went to America  and set off the countryâ€™s balance It sunk into the ocean and ryhmin to balance is not what I do Verse 2 The cattle king went to Indian land, and there he went to the chief At first, it was cool, but he wanted to turn the evil cattle king into beef So the Indians summoned a monster But when the monster spoke He told a joke And nobody laughed The lil monster got mad So he went home and told his dad His dad came back and killed them all Thatâ€™s how babyâ€™s are made in the monsterâ€™s ball END,yo  '),(23142,3400678,'22,May,2004','2004-05-22 00:00:00','   Culture Lesson by Myk Doolin  (Spoken) I had to shoot it to make sure it was dead  (Verse 1) Cut through the neck and became an apparition, Lightsaber vs. taser would make one big collision Gained more power and knowledge through being alchemistic The Faust that has become me, no soul, just optimistic Donâ€™t ask how, just know why, itâ€™s confusing, but still try, understand the ways of defeating the albatross simple and low loss, state of mind, youâ€™re your own boss hack life, make it work at your own cost fault phantoms spread like viruses throughout the human race tracing and erasing the truth, and the real face, one big lie, waiting to be envisioned through our eyes when we die but yet, everyone seems to still have pride, why? You may ask this, but really itâ€™s common sense, theyâ€™re all blocked by a misleading monolith  (Chorus: Faith No Moreâ€™s Epic) You want it all, but you canâ€™t have it Itâ€™s in your face but you canâ€™t grab it     (Verse 2) some people perceive me as a herbal twister maybe itâ€™s the way I think creativity is my girl and I canâ€™t resist her immortality came back when I lost my top and died listen up you primitive screw heads, are you outta your mind do you even know your kind find yourself, makes life divine  out of time, wait a second, Iâ€™m in a time machine,  I got all the time I need, wanna go back, and make time for me? Havenâ€™t ya seen, the buzz you can get from caffeine Methamphetamines make minds obscene like a latrine The unknown third bean in the salad, watch out Stab you in the back with a phoenix talon Didnâ€™t expect it, you shoulda wrote it down in blood with a pin Visional arithmetic, how every human wants their kin  (Chorus: Faith No Moreâ€™s Epic x 2) You want it all, but you canâ€™t have it Itâ€™s in your face, but you canâ€™t grab it  (Spoken) You might not know what this isâ€¦â€¦â€¦â€¦But, you know you want itâ€¦â€¦..and you know you canâ€™t have it, but youâ€™re willing to do almost anything to get itâ€¦â€¦â€¦â€¦Sometimes, you gotta work around what doesnâ€™t comeâ€¦â€¦â€¦â€¦but, always remember, itâ€™ll get there soon enough.   Traffic  Iâ€™m gonna walk out into traffic With no red lights Iâ€™m gonna walk out into traffic With no yields or stop signs A walk out into traffic My life  '),(23143,3400678,'22,May,2004','2004-05-22 00:00:00','   Paradox, Double cross, never ends, never stops It doesnt go but always rock watch it go slow fast now your mind will blow  Yeah it\'s time enough at last. when did you last laugh? when did you last smile, yeah it\'s been awhile  I\'m always here so have no fear.  Yeah it\'s time enough at last. when did you last laugh? when did you last smile, yeah it\'s been awhile  When I\'m all alone, it\'s the twilight zone when im all alone the twilight zone   Yeah it\'s time enough at last. when did you last laugh? when did you last smile, yeah it\'s been awhile  '),(23144,3400678,'22,May,2004','2004-05-22 00:00:00','   I\'m not rap. no im hard rock, yeah emo thats me I\'m not hot but your so damn sexy im not cool but you are and this i see I love you baby.  Yeah ya rock, rock my f***in socks off yeah your so cool, baby I love you  I feel out of place, i once was last, now im winning the race. oh i love your face smile,laugh and all your traits Yeah i think you should know. I\'m losing all control.  Yeah ya rock, rock my f***in socks off yeah your so cool, baby I love you  '),(23145,3400678,'22,May,2004','2004-05-22 00:00:00','   The label says take one. so what if I take four? the label says take one. so what if I took more. the label says take one. but thats a bit too late the label says take one i already took eight.       I need my little pill       I need it to sit still.       Helps me to be happy.       My little pill. The label says take one there are no more to take the label says take one that is far too late the label lies i need more one is not enough i need more the label says take one i need more.  '),(23146,3400678,'22,May,2004','2004-05-22 00:00:00','   When Spring dies, summer grows fall comes, and then it goes winter comes, and then it snows there is no escaping, those are the rules there is no escaping, those are the rules The seasons change and so do we. we all do. we cant stay we all must move we live, we die we smile, we cry winter,spring,summer,fall you want one and get them all so I die, you will live, you have so much more to give so i smile and you cry perhaps its time for me to die? winter,summer,spring and fall. i have lived through them all.  '),(23147,3400678,'22,May,2004','2004-05-22 00:00:00','   Small slashes by zac roberts.  They cover your arms. (why) they bring such harm. (why) why the yearning? (to die) Small slashes, little lines, on your arms, all the time. bleeding red on skin so white helping to get through the night. Small slashes, Little lines bleeding red on skin so bright bleeding down all through the night can they make everything right? these little slashes on skin so white.  '),(23148,3400678,'22,May,2004','2004-05-22 00:00:00','   Hey, this is the official M3Z blog. Not much yet. Soon i will post lyrics to all of the songs I (zaphlon) have wrote. Here\'s what to expect:(songs i have finished, expect up in a week or so) 1.Time enough at last 2.Bag of boxes 3.Love song 4. F*** you 5.Four seasons 6. Small slashes 7. Pill song .......these will all be up in 1 week.  '),(23149,3400678,'15,June,2004','2004-06-15 00:00:00','   Can\'t do anything Can\'t do anything Can\'t do anything I\'m a loser an unobtainable love a forever out of reach I\'m a loser You don\'t want me you don\'t want me you don\'t want me I was never out of reach I was never unobtainable I was always there when you needed me And I always will be allways be there I\'ve been a bastard you\'ve been a bitch I\'ve been there tho and always will be I\'m a loser You dont want me I\'m a loser You dont want me Be happy and move on Be happy and move on and maybe at some future time I\'ve always been yours But you\'ll finally be mine Can\'t do anything Can\'t do anything Can\'t do anything  '),(23150,3400678,'13,June,2004','2004-06-13 00:00:00','   I want you to know I\'m just waitin til you die waitin til you die so i can write a sing bout it I want you to know I want you to know I\'m just waitin til you die waitin til you die so i can write a sing bout it I want you to know I want you to know I\'m just waitin til you die waitin til you die so i can write a sing bout it I want you to know  '),(23151,3400678,'13,June,2004','2004-06-13 00:00:00','   Little light little burn pass around take a turn spin the wheel spark spin the wheel spark spin the WHEEL!!! spark Im sorry man, im insane im sorry mam, im insane im sorry, ive got a fetish for the flame  the flame  I love the fire love the heat love the burn love the sound smell so sweat I love the turn ing of the wheel spin so fast cuz a spark make a spark hits the gas make a spark make a spark make a spark make a spark  a fetish for the flame a fetish for this game a game of pain sorry mam but im insane sorry man im insane sorry sorry sorry make a spark make a spark metal sound calming metal sound so soothing metal sound uplifting metal sound.... MOVING!!!  IM SORRY MAN IM SORRY MAM IM INSANE  IM INSANE IVE GOT A FETISH FOR THE FLAME!!!!! make a spark...  '),(23152,3400678,'08,June,2004','2004-06-08 00:00:00','   (sung slowly): Me and my buddy Steve We went walking down the street We saw this old women with green hair it was neat Me and steve walked into a bar and saw this chick we asked for her number, she said no.... i said: Well okay...then ummm.. STEVE FUCKING STABBED HER (sung slipknot esque): STEVE FUCKING STEVE FUCKING STEVE FUCKING STABBED HER ............... (spoken by twik):Really I\'ve known steve my whole life? YEAH STEVES FUCKING CRAZY (spoken by zaph): He\'s insane man.... STEVE FUCKING STEVE FUCKING STEVE FUCKING STABBED HER STEVE FUCKING STEVE FUCKING STEVE FUCKING STABBED HER  '),(23153,3400678,'07,June,2004','2004-06-07 00:00:00','   Emotions devoted To everything thatâ€™s been coated In my hate for this macabre life You Fight You Die Or just got forget about Want something more? JUST FUCKIN FEEL ME NOW  Time is passing Nobodyâ€™s Lassie No one on the street is gonna STOP AND HELP YOU OUT They take one glance And youâ€™re critiqued  Their judgment is settled UNABRIGED AND COMPLETE  CROWS Can make Whatâ€™s wrong RIGHT With these Words I will FIGHT  Ball up your fists Protect the whole clique Wanna take risks CEASE AND ASSIST  (Chorus) With these words  I will  FIGHT With these words I will IGNITE With these words  FIGHT x 6  Emotions devoted To everything thatâ€™s been coated In my hate for this macabre life Time is passing Nobodyâ€™s Lassie No one on the street is gonna STOP AND HELP YOU OUT You Fight You Die OR JUST GOT FORGET ABOUT Want something more? JUST FUCKIN FEEL ME NOW They take one glance AND YOU\'RE CRITIQUED Their judgment is settled UNABRIGED AND COMPLETE  (Chorus) With these words  I will  FIGHT With these words I will IGNITE With these words  FIGHT x 9  '),(23154,3400678,'04,June,2004','2004-06-04 00:00:00','   Fuck it. Fuck her. I don\'t need this. Fuck it all. Fuck it all. Fuck you. Fuck it. I don\'t need this. Fuck it all Fuck it all. Fuck love. I don\'t need it. I don\'t need this I don\'t want this. Fuck it Fuck this Fuck this very negative emotion Fuck it all.  '),(23155,3400678,'02,June,2004','2004-06-02 00:00:00','   I dream of reality Of simpler things and Fantasies Of every thing that cannot be and everything i believe should be I dream of reality Of simpler things and fantasies I dream of Love Life I dream of thee I dream of reality   I need to get a grip, emotions and words slip.  If i trip I will never rise just lay there and close my eyes And slowly fade to black never coming back I see no need to be a prisoner of reality a prisoner of what people say I want to live on my terms and allow others to do so too i wanna be happy thats what i wanna do I want to mean and i want to live i dont want to just merely exist in a sea of people, they pass by millions dying, millions cry That i dont know and never will reality is a jagged pill bittersweet forever wish things were better wish people knew like they could see im not insane im just me your all insane im just me  I dream of reality Of simpler things and Fantasies Of every thing that cannot be and everything i believe should be I dream of reality Of simpler things and fantasies I dream of Love Life I dream of thee I dream of reality  I need to find someone to hold so i can be told \"I love you\" and \"I love you too\" that is what i need to do a lover a friend a companion someone who\'s love will never end I may be young and foolish it all may be untrue its the only way to live its the only way  I dream of reality Of simpler things and Fantasies Of every thing that cannot be and everything i believe should be I dream of reality Of simpler things and fantasies I dream of Love Life I dream of thee I dream of reality  '),(23156,3400678,'02,June,2004','2004-06-02 00:00:00','   Zaphlon has written all but the following songs: Pants 1 and 2 Culture Traffic and Pure sounds of entertainment.  '),(23157,3400678,'08,August,2004','2004-08-08 00:00:00','   Hehe, it\'s Loky, I\'m going to post a song that I wrote. If I find out it was stolen, I\'ll hunt you down and kill you. Have a nice day.   I\'ve got scars you can\'t see A past of misery Then along came you You changed everything about me Now you\'re out of my life again How can I still be like you wanted me to be?  It comes down to this You wanted this you wanted that I gave, I gave, I gave, I gave I never asked for a single thing Now is my time, I want You to know, how bad I am doing, without you All I wanted, was you Now you abandoned me I\'m alone once again  I\'ve got scars you can\'t see A past of misery Then along came you You changed everything about me Now you\'re out of my life again How can I still be like you wanted me to be?  I want to be numb again But this pain will not go away You\'ve screwed me over To many times before I refuse to fall this time You can not win anymore Because I have grown immune To your words, to your pain I\'ve grown immune..... TO YOU  I\'ve got scars you can\'t see A past of misery Then along came you You changed everything about me Now you\'re out of my life again How can I still be like you wanted me to be?  '),(23158,3990427,'30,July,2004','2004-07-30 00:00:00','   Last night was...interesting? I got roped into going our with Rachel and Catherine to the bar at the Loews in Philly. The event was for The Young Professionals&nbsp;of Philadelphia. Yawn. The bar was gorgeous but the people were less than impressive. Tons of Real Estate folks and reps. from the City Paper. (Trash paper, as I like to call it. Serious. Naughty personal adds and all.) We mill around smoking cigarettes, I drank perrier with lime, the gals had vodka martinis. I\'m watching my calorie intake and this was NOT the occasion to break it. I\'m not tricking myself into feeling comfortable around these lame-ass dudes by ruining my diet just so I can hold a glass of alcohol and sip it all night.  We pretty much stayed around the people we knew. No schmoozing. Just making fun of the insipid crowd and the old fellah in the FULL, top and bottom, bubblegum colored suit.  We stayed for about an hour (which felt like 5) and moved it on over to the U Penn campus for movie night on the lawn. We saw a HORRIBLE serial called Mystery Mountain about a guy named Ken and his nemesis The Rattler... I joke you not. It was not even kitchy. Just plain horrible. They really neede the guy who did the sound for The Stooges. Their fight scenes packed no punch. (heh heh heh) The main show was a movie called \"Youth Gone Wild\". The ebtire show we were waiting for someone to go wild - we were left unsatisfied. It WAS totally a ball though. We played Mystery Science Theater 3000 the whole time and poked fun at the film.&nbsp; I think my favorite part was when Tottie ( the dame of the flick) DIED by getting hit across the face with some guys legs, mid air, as he was being flipped over someone else\'s back.&nbsp; Did you get that?? THAT\"S NOT POSSIBLE!!! There was even a dramatic scene of a priest reading the last rites...what a RIOT! Thanks for the memories \"Youth Gone Wild\".   '),(23159,3990427,'22,July,2004','2004-07-22 00:00:00','   I just read an article saying that at his trial, Michael Jackson requested \"a celebrity exception to the First Amendment\" to keep some of his child molestation case secret.  THAT gets the award for the most arrogant thing I\'ve ever heard.   '),(23160,3990427,'21,July,2004','2004-07-21 00:00:00','   I just can\'t seem to keep my simple mouth shut. In order for me to heckle my pal Dusty, I have to have one of these accounts.  Do NOT think for a second this is a weird stalker crush. Know for a fact that it is.   '),(23161,2314011,'30,November,2003','2003-11-30 00:00:00','    So, after reading at various times on my favorite author\'s blog about non-saved and erased blog entries, I\'ve at last suffered the same fate.  And sure enough, it was a very very important entry.  Full of wonder and shocking twists.  Really.  It was.  I was typing along, writing about my day back home, full of detail and fluff.  Was getting around to my plants, reliving the stories of their namings when Lorraine IMed me.  The rest of the entry summarized that non-conversation and the torment that went along with it.  Just finished spell-checking my entry for the final time and was IMed by a stranger.  Decided to check out the profile of said stranger and was faced with a \"save entry\" question and I, unwittingly, clicked \"cancel\" just before I realized what that meant.    I\'ll gather myself and try again.  Deep inhalation, and:  Ahhh home again.  (That\'s really how the first one started, I remember that much.)  Made it in 9 hours.  Slept late this morning and made a real breakfast.  Yay.  Finally got around to putting up new art work and repairing old this afternoon.  Running a fan all night to mask the sounds of neighbors and highway isn\'t all that conducive to long-lasting adhesive devices.  Did some Property reading for tomorrow, but never got around to doing the practice exam; maybe it\'ll get done tomorrow.  But it\'s World AIDS Day tomorrow and I\'m supervising the mircophone hook-up process in the courtroom, heavy responsibility.  See the Property practice exam moving to Wednesday.  Maybe I just won\'t take the final at all.  Listened to Tori\'s Collection while puttering and it\'s really not that bad.  They are all improvements, and good ones at that!  Suppose it won\'t be banished to the CD shelf (graveyard) just yet.   The plants made it through their week alone.  They were pretty dry and gasping for breath when I returned, but they perked right up after a shot of vodka each.  They\'re spider plants and the first living things I\'ve ever cared for.  Now I\'ve got to hook up to AC power before my battery goes dead and I lose a 2nd post.  There.  So yea, Levi brought the first plant from Oregon over the summer.  Came home from work one day to find my Shakespeare anthology on the table. \"You name the plant?\" \"Yup....Calphurnia.\" \"Calphurnia?!\" \"That\'s right.\" \"Okay.\" Related the story to Lorraine that night and she replied: \"Ahhh, Caesar\'s wife.\" There\'d been a TNT original movie on that night about the man.  One rarely gets a better set up than that.  She\'d just IMed me as I finished that story the first time.  She wanted to visit, to see how I\'m doing, to make sure I\'m okay.  She was afraid, during my time to think and reflect, I\'d decided to hate her and forged regret, which has always been her greatest fear.  I sounded very detached and harsh when I wrote back that I didn\'t hate her, I have no regrets and we weren\'t going to chat about anything.  It was very weird.  And Dido started singing about her White Flag in the middle of it.  Over the course of the rest of the entry, in which I described the birth and naming of Calphurnia\'s daughter, Trudi, Lorraine asked whether she could still call if it was critical that she speak to me.  Not tonight; just for future reference.  Kept thinking about her forwarded email and what a shit I felt like for not sending a one-line reply, similar to the one-line responses I gave her tonight.  I\'d planned to type about her originally, but clearly hadn\'t expected what actually happened.  Been wondering about all the Lorraine-related stuff I\'ve got around and trying to decide whether to put it all way for the duration.  Like her old shirt I still sleep in, for example.  Any thoughts, dear readers?  I don\'t have a shrine, but there\'s a fair amount of innocent little things around.  Should it all disappear, for a bit at least?    She kept typing a message and stopping.  That\'s one of my favorite features on IM.  Just as I was finishing the entry and getting ready to start spell checking it, she sent a final message and signed her name.    And that\'s how it really happened.  This entry\'s not as good as the original, but it\'ll do.   '),(23162,2314011,'28,November,2003','2003-11-28 00:00:00','    In 8 hours I\'ll be dragging myself out of bed, making the final decision of which route to take home to Moscow.  If the weather\'s going to be as joyous as last weekend, I think I\'ll zoom through Boise instead.  Not too excited about dealing with the remaining 300 miles of two-way highway, but you just never know.    So the big dinner went off without a hitch.  Grandpa and Grandma were wandering around, going through the motions and everyone there tried (pretty successfully) not to talk about the big boulder sitting in the middle of the room.    \"We don\'t own this house anymore!\" / \"I sure am going to miss this house.\"  Were pretty much Grandpa\'s conversation pieces all evening.  Still can\'t shake the feeling that this whole \"new house\" business is the beginning of what we (the fam) gets to look forward to for many years to come.  They\'ll pack all of their boxes of shit into a house a fraction of the size the one they\'ve lived in for 44 years, move it around, get organized and it won\'t be home.  Sigh.  Went to Burley today to spend the biggest shopping day of the year with Mom\'s mom--Grandma Leah.  She used to not be my favorite Grandma for a while, but now she\'s a riot.  Very much a little old lady on a farm in the middle of nowhere who, when venturing out to get her paper/mail, will run back behind her bushes if she hears a car coming down the road.  She hates cats and has about 5, 3 of which come around regularly.  She won\'t see \"The Green Mile\" because it has a mouse in it.  We cruised the Twin Falls mall, prowling for bargains.  She walked with her arm through mine; it was adorable.    Had beers tonight at my cousin Rob\'s and his incredible wife Trish\'s new house.  You think MY house is a mountain fortress, honey...jeeeezzz!  Talk about a party!  Just on our way home when said incredible wife started dishing about her in-laws (Dad\'s brother, my aunt and uncle).  I love talking to kick ass family members about the shitty ones behind their backs.  Good times.  Then her youngest daughter quit breathing, turned blue and we left.    Now all packed up and not at all ready to drive back.  One week of school before the big push.  Not really worried too much, just trying to convince myself I really am prepared.    Then it\'ll be time to do it all over again, only this time we\'ll call it \"Christmas.\"  How does Seattle sound, hon?  I know someone we can stay with.   '),(23163,2314011,'27,November,2003','2003-11-27 00:00:00','    Morning, blogger!  So I just felt like writing a bit this morning, got a problem?  Left a few people outta my entry last night and decided to remedy that this morning.  Been perusing my email and other blogs to find people listing what they\'re thankful for and such.  Don\'t know that I need to go to such extremes, but we\'ll see where this goes, and it may just end up to be exactly that.    Received an email from my former Argonaut-editor-turned-lesbian-pal last night.  Ordinarily, such an event wouldn\'t be much to write home (or even a blog) about, but she doesn\'t do email.  Not that she\'s against it morally (you\'re not, are you?), it\'s just that it\'s never really been her mode of communicating.  She\'s a successful journalist, after all; she makes her living writing writing writing, and I can imagine that the idea of sitting down at the end of a day to write more just to keep in touch with long lost friends across the land would be a bit defeating.  Instead, she wrote from work.  Yay.  And it was such a nice surprise.  So I\'m thankful for our chats in her old office and our October Starbucks gay pride sit-in.  Been thinking about my boy.  If you haven\'t already read on his fine blog, his dad\'s and step-mom\'s house is being pulled out from beneath them by the farm Mitch (that\'s his dad) has dedicated more than a few years\' worth of blood, sweat and tears to.  Suppose it\'s situations (and days, tear!) like these that remind you to get a deal in writing and all, but fucking hell.  The company\'s given them til March to vacate, and anything could happen between now and then, but still.  Can\'t quite imagine walking through a house, knowing it\'s not yours anymore.  And really, if they\'ve decided the folks needed to go in the first place, not sure much trust is being placed in the stability of the move-out date, either.  Could start in on all the \"everything happens for a reason\" stuff, but this is just shitty; there\'s no getting around it.  And maybe it is happening to make room for something else bigger and better, but for now, being pissed, hurt, shocked, etc. is allowed.  This could be the part where I say I\'m thankful for my parents\' house and home, that we can all gather here whenever we want and be secure in knowing that\'s never going to change.  But actually, with everything that\'s been thrown at each of us and our families (his more than mine) this and other years, I\'m thankful that we\'ll always have the other to talk/shriek/sob about it with.  And if it all comes tumbling down around 2, we can always gas up the car and head to Jersey, the mixes blaring.  Okay, I think the rest will turn into a list, otherwise, I\'ll be here all damn day.  I\'m thankful for:  Dave\'s smiles, warm and cheeky alike; wishing him a safe trip home and an even safer week with family;  Carrie\'s phone calls, messages, pep-talks, determination and perseverance;  Not having to use serial commas on break;  Lorraine;  Susan\'s refusal to banter with me at school;  Alycia\'s banter in and out of school;  The London Mixes; and  2003: the year of celebrities, signs and big damn changes.   '),(23164,2314011,'26,November,2003','2003-11-26 00:00:00','    I can\'t seem to get enough rich, fried, salty foods today.  Gorged myself at Garcia\'s in IF today; seriously, those chips and salsa were worth staying another 10 minutes after they\'d cleared the main meal\'s dishes.    Burned it off in the mall, though.  I never go shopping for myself in Moscow.  It honestly never occurs to me.  Get to make up for it here.  Yay.  Found some good stuff.  New shoes, new pants, new sweater and some promising ones sure to be under the tree in a month.  Decided, while Mom and I were admiring a new outfit due out later this winter, that I\'m cutting my hair all off this summer.  Not shaved, dear god.  But I think it\'s time for a short, stylish, maybe even a bit dykey, but not butch, hairstyle.  And I\'ve got 6 months to get used to the idea.  So yea, got home and had to hit the Pine Ridge Mall because it was only 5.  Whatever.  I called Carrie while Mom did JCPenny\'s.  Then on to Vicker\'s to secure Dad\'s present of 5-buckle goloshes.  I shit you not; it\'s what the man wants.  Got home just in time to catch the boys on their way to dinner.  I wasn\'t even hungry, but I\'ve never been able to resist the smell of fresh grease and onions frying.  The butter on the toast was the same shit they drizzle on popcorn in movie theaters for fuck\'s sake!    Suppose I\'m just getting ready for tomorrow.  Yea, sure, we\'ll go with that...   '),(23165,2314011,'25,November,2003','2003-11-25 00:00:00','    Pretty full day today.    Had the annual teeth cleaning this morning at 9 am, which I was not too excited about, but at least I have a sparkling white grin now.  My dental hygenist was in a playful mood and that\'s always a good thing.  She\'s my friend Mike\'s mom, so we chatted about him and his adventures abroad in Ecuador.  He\'s supposed to teach his mom how to say \"Shut the fuck up\" in Spanish when he returns so she can throw down on the days her coworkers feel the need to belt out Tim McGraw tunes in the office.  Laurie\'s (that\'s  her name) good people, she knows I don\'t ever floss and doesn\'t judge me.    Then pounded out a three hour Torts practice exam, which I think I did pretty well on.  Yay.  Decided I deserved a break from studying around 3 and indulged in mid-day network programming.  Finally got off the couch around 930.  Yea, felt pretty guilty about not working on Property stuff, but not guilty enough to pull me away from back-to-back episodes of Queer Eye for the Straight Guy followed by Roseanne on Nick and Nite....god I miss cable!!!  But really, I haven\'t had a \"just lay around watching bad day time TV\" day in a long damn time and damnit...I deserved it!  Property can wait!  It\'s during the last week of finals, anyway.  Have to admit, though, did start feeling a bit zapped after awhile, and maybe a bit creatively blocked after not writing any pages at all for a day.    Was wondering how Amy would decide a case involving an Omish (?) girl\'s custody battle with the baby\'s \"English\" teenage father on CBS\'s riveting family court drama when Dad stopped by the couch to ask if I\'d join him for a visit down at the grandparents\' house.  Decided that was as good an excuse as any to get away from the machine\'s dangerous rays and went on down.  So yea, the old folks have decided to move into a newer, smaller place and I have very mixed feelings about the whole situation.  And by mixed I mean bad, sad and generally resentful.  Well okay, not resentful so much as already missing the old house next door to my elementary school.  I mean, it\'s Grandma\'s and Grandpa\'s House for crying out loud!!  This was the house I memorized my first book in!  How can we not be having Thanksgiving and Christmas dinners there anymore?  Thursday\'s will be the last official dinner in the homestead, so being there tonight was a good preparation for the big goodbye.  Regaled them with stories of meeting Justice Ginsburg this fall and the treacherous ride home last week.  Heard stories of young turkeys raised in the basement and barn, scouring the summer fields for choice grasshoppers, only to be slaughtered in the fall, and Peggy the cow with an amputated leg.    So tomorrow\'s the big shopping day with Mom.  Already fitting the armor and working out thoughtful, calm responses to anything and everything she may try to slip in under the radar.  Not sure if it\'ll come to blows, but I\'m gonna be strong, by god!    Suppose I could use my favorite new line for situations such as these; used it on Dad tonight when he was grumbling about killing a stray cat on the grandparents\' driveway.    \"Right?\" was his inevitable end to the mumbled tirade.  I simply said \"I\'m not engaging,\" and left it at that.   ');
INSERT INTO `msgs` VALUES (23166,2314011,'24,November,2003','2003-11-24 00:00:00','    Okay, I suppose it\'s time to do this.  Have the house to myself for a couple more hours and have been working hard on studying all day, so I deserve this break.  Figured it\'s as good an opportunity as any to start filling in some gaps and clarifying some statements/allusions/what have you.    It\'s been a bit odd being back this time, the journey and luggage issues aside.  And I\'m not even going to pretend to wonder why.  I know it\'s because I\'m not talking to Lorraine.  Realized while on the phone with Levi that this is the first time in 6-ish years we haven\'t spoken on a daily basis.  Suppose it was bound to hit sooner or later.  Everything here reminds me of what it was like when we both lived in the same town.  Knowing she\'s an hour away is another thought to be pushed away, along with how easy it would be to just say \"Happy Thanksgiving\" or \"Made it home safe.\"  So let\'s back up a bit, shall we?  Lorraine is...well, she\'s Lorraine.  Simply put, she was my first, and as yet only, love.  She\'s my best friend, my pal, my goofing-off buddy.  We go fishing and laugh big laughs.  She\'s the closest thing to a soulmate I\'ve ever met.  So why aren\'t we together in as-wedded-as-we-can-get bliss?  Well, circumstances, mostly.  Such a lousy excuse, I realize.  But we both made choices and must now abide by them.  Basically, neither of us were willing to cut off our relationships with our families to ride off into the sunset together.  She was also my first relationship, and there was always a part of me that wanted to explore more....normal dating situations, not wanting my life to be consumed by a shroud of secrecy and/or isolation.  So, while we were together, it was kept secret to protect her from public shaming and me from public institutionalizing.  And we pulled it off for a long time, while living in the same town and, later, when the entire state separated us.    Soon enough, however, it was time to reexamine the situation.  My brother would soon be joining me at school and as much as I wanted to start traveling to a mid-way point for secret monthly trysts, I didn\'t.  So, something had to end.  We told ourselves it wouldn\'t really be an end...more of a change, really.  And as of August 11, 2002, the change began...only not.  We still spoke everyday, in the way lovers do, and things carried on as they always had, only without the visits.    Then, in the Spring of 2003, she met someone:  Jan.    Before you jump to conclusions, it should be stressed that I was (and am) very very happy for her.  For years, she convinced herself that should our relationship ever end, I would find love quickly and just as quickly forget all about her.  But as it turned out, she found someone who made her happy.  A Mormon housewife with 3 kids and one grandchild, but still.  I expected things between us to change further, but Lorraine was adamant about not letting her new love drive a wedge between she and I.  I was still her best friend, and woe be unto anyone who asked her to part with that.  So, after more than a few conversations and one group meeting, we all were on the same page...kinda.  You see, Lorraine and I never really got around to figuring out how our relationship really ended.  We just kinda went along, doing what we always did, knowing it was \"different\" but not really changing anything in our ordinary course of dealing with each other to reflect that difference.  So it was bound to happen, when she visited (alone) in July 2003, that we\'d blur the lines more than a bit.  And sure enough, I was left feeling incredibly guilty and not knowing what our \"re-connection\" meant for our friendship.  But, we worked through it and again, nothing changed in our daily discourse.  Well, nothing except for the fact that she couldn\'t seem to talk to me with Jan under the same roof.    As more time passed, and dealing with the reality staring me in the face kept going undone, I grew more and more frustrated with the feeling that I was nothing more than an obligation...something she checked off her list at the end of each night.  All for the sake of maintaining the last contact our friendship still afforded us.  All the while, I never resented the new girl or their life together..honestly.  I merely resented the fact that I was now a supporting character.    As a last-ditch effort to gain some clarity/perspective on the issue, I consulted a couple of wise lesbian sages for their advice.  My own approach to giving advice came back to haunt me as they told me exactly what the problem was without pausing and without reservation.  I wasn\'t getting over her.  I wasn\'t moving on.  And I wasn\'t being reasonable in my expectations that everything could continue smoothly from one relationship\'s manifestation to another.  There was only one solution: get over her, move on from our relationship, and do it alone.  Only when I no longer needed/depended on connecting with her on a regular (not even daily) basis should I attempt to maintain a friendship with her.    So that\'s where we are now.  It\'s been about a week and a half since The Conversation and as you can tell how it\'s gone from the previous entries in this blog.  Some days are easier than others.    She emailed me yesterday.  It was a forward of a cheezy Thanksgiving joke-picture, but an email nonetheless.  Had it reached me during a time of strength and resolve, I probably would\'ve smiled and moved on to other things.  But it didn\'t arrive during a time of strength and resolve, and as you can tell, I haven\'t entirely moved on to other things.  The feeling of guilt for not responding has passed, but I\'m just thinking about her more than ever now.    Big Sigh.  So yea, not a new story for some of you, and perhaps an enlightening one for others.   '),(23167,2314011,'23,November,2003','2003-11-23 00:00:00','    The plan was to leave town as soon as my Civil Procedure class let out.  Probably around 1, but perhaps a bit earlier, depending on the questions asked during the review session.  It snowed the night before, but that was nothing new; it always snows the night before I leave Moscow.  And sure enough, the snow was melting and the sun was out when I blew out of town just after 1.    Made it to Coeur d\'Alene just after 230 to find gray, doomsday clouds and a sheet of ice masquerading as a highway waiting for me.   \"Should be fine.\"  The slush and snow from the night\'s storm had frozen to the road nicely.    \"So I won\'t get there till 11...it\'ll be okay.  I\'m going to make it, by god!\"  An hour and a half and 40 miles later, I admitted defeat, at least for the night.  Got a nice, cozy, $70 room at the Wallace Best Western and tried to relax.  Fought the urge to call Lorraine (we\'ll get to her later), and found comfort with the Golden Girls instead.    Not so bright and early the next morning I stoically folded back into my trusty steed and made my way over the mountain.  The roads didn\'t actually clear up until Missoula, and I was still trying to convince myself that I could make it to Bozeman at a respectable hour.  \"Okay, it\'s only 11.  If all goes well, I could be there by 2.  That\'s practically a whole day; plus we\'ll have Sunday, too.\"  And for the most part, it did go well.  The wet roads were fine, until I remembered that it was 20 degrees outside and if they were wet, they could just as easily be icy.  But damnit, I\'d made a plan and fuck if the goddamned weather was going to cheat me out of it!!  Kept thinking about arriving and falling into Levi\'s couch, drinking tea or something stronger, and maybe taking our relationship to the next level.    But home was sounding pretty good, too.  Could sleep in my cave, have time to lay around before hitting the outlines and practice exams next week.  But no!  That\'s not what was supposed to happen!  I\'m supposed to be in Bozeman!  That\'s where I\'m going!  Don\'t try to stop me!    \"You could have two entire days to yourself to get prepared.\"  \"SHUT UP!  I want to see my boyfriend!\"  \"And get back in the car Monday morning...\"  \"I can study Monday afternoon!\"  \"After driving half the day?\"  \"If I have to!!  See..here\'s the turn off, I\'m going to Butte!  I\'ll show you what I want!\"  So I went.  And halfway through the town, I realized I\'d be bad news.  Just thinking about how much time we lost, what all we could\'ve done but would be impossible now.  IMPOSSIBLE!  And if I did get home Monday, I knew I wouldn\'t do any work and I\'d just be pissesd at myself all week for not doing what I knew I should...even though I didn\'t want to.     The fucking practical side won out again...damnit!    And today, I\'m sooo glad it did.   '),(23168,2314011,'22,November,2003','2003-11-22 00:00:00','    It never fails.  Usually within a day or so of my arrival back home, Mom and I go on an excursion during which she reminds me of something.  Still can\'t quite define what that something is, but it\'s along the lines of: \"You\'re not good enough/You\'re not what I expected and I resent that/You\'re not in charge here/You don\'t get to be you here.\"  I\'m sure I\'m being completely overdramatic, but jesus fucking christ!  And it always happens while we\'re shopping and I won\'t let her buy me something I don\'t like/want/need.   Today it was in TJ Maxx.  For some damn reason she\'s on this luggage kick.  Has been since last year.  I have two fucking sets of luggage, but I need a middle size.  She was on it last year because I was traveling to London, at least there was some context.  But it was there again this year...whatever.  So, of course, I very gently tried to tell her that no, I still don\'t need another piece of luggage.   \"You always say that,\" she said, as if I was deliberately trying to ruin her fun. \"Well, it never changes,\" I replied, trying to explain why I still don\'t need it, and trying to figure out what the hell this was about, not thinking it was a big deal. She sighed, walked away and said over her shoulder: \"I can see shopping with you is going to be sooo fun....like always.\"  As though I\'d just insulted her.  I couldn\'t believe it.  \"Ditto/Excuse me?!/What does that mean?\"  All ran through my head at the same moment, but I couldn\'t get any of them out.  Instead, I just sealed myself off and followed her to the sock section where I picked out three new pair.   Then we didn\'t really talk for the rest of the outing.  We drove by the new house my grandparents bought, talked about what a mess that is and all, but we weren\'t really talking.  Still aren\'t.    WHAT THE FUCK???  It\'s never been any different.  And I know one of these times one of us will say something, but good fucking god, I just wish I knew what the damn deal is.   '),(23169,2314011,'22,November,2003','2003-11-22 00:00:00','    I wasn\'t the same after that bag of Cheetos...and I don\'t think I\'d want to be, either!  ....I\'ll explain tomorrow, I promise, dear readers.  But for now, it\'s time to sleep.   '),(23170,2314011,'20,November,2003','2003-11-20 00:00:00','    So, just printed off a ream of paper in preparation for the big study session that is Thanksgiving Break.  Got some nice little practice exams raring to go.  Super!    And since I know you\'re reading this, study-pal, I know I said I\'d print off copies for you too, but the Property exam alone\'s 17 damn pages, so you\'ll just have to use the allotment from the school, sorry.    So think I\'ll be going to this little discussion later about \"What Women Want\" because, you know, I should get that figured out at some point.  And if the interim director of the Chick\'s Center wants to hit on me while I\'m there, that could be fun, too.  Should be good for another blog entry at least....   '),(23171,2314011,'19,November,2003','2003-11-19 00:00:00','    A TEAM!  And what\'s a team??  T ogether  E veryone A chieves M ore  !!!!!!!!!!!!!!!!!!!!!!!!  Yes, that is how I was introduced to one of my good friends, Ms. Carrie \"Shut the Fuck Up\" Johnson. As you can tell by her name, that Pollyana attitude went to the way-side eventually, but we still like to bring it up every once in a while.   So now that she\'s been introduced, you\'ll probably be hearing more about her in the entries to follow.  Here\'s the cliff notes on her story thus far:   Fate brought us together at the University of Idaho Queer Student Association and allowed us to continue our blossoming friendship at the University\'s Library, where, over the following years, great trials and tribulations came to pass.  But, we persevered and have lived to tell the tale.   This summer, Ms. Johnson (aka Anastasia Beaverhausen), up and decided that she was tired of this side of the continent and needed to make her way across the country to live in Boston, Mass.   And, by god, she pulled it off!  YAY!  GO CARRIE! Today she begins a new job in that fair city as a crusading social worker, saving down and out citizens from the evils of the world around them by day, saving her evenings for the Jack & Karen hour.  (That\'s code for Will & Grace:  it\'s not about Jack or Karen, and it\'s not an hour, but it\'s the Jack & Karen hour.)  So, there you have it.    Congratulations on the new job, sweetie pie.  I know you\'ll do great things.  And if anyone gives you any guff, just have them call your attorney who also dispenses medical advice.   '),(23172,2314011,'19,November,2003','2003-11-19 00:00:00','    Hello out there to any new readers jumping over from Levi\'s blog!  I find it interesting that his was the \"private\" blog, but it already has a fan-base following it regularly.  Sigh.  Probably has something to do with all that damn charm of his.  At any rate, thanks for the plug hon!  So, it occurred to me that I\'ve been making regular posts alluding to the process of studying for my first semester law school finals, but I haven\'t actually described what that process (or its significance in the great, wide realm of general understanding) really consists of.  And whether or not anyone\'s curious, here\'s the low-down:  Pretty much everyone knows that grades in law school are made up entirely of one (1) exam.  The final.  No quizzes, no assignments, just the final.  Each final\'s length is determined by the class\' credit hours.  One hour of exam-time for each credit hour.  So, a three credit class has a three hour exam.  I\'ve got 4, 3 credit classes, so that makes for 12 hours of exam time.  Yay.   The process leading up to those 12 hours has, thus-far, been an interesting one, I have to say.  The most prevalent study method is outlining.  Basically, students outline their classes and study that sucker up until and including the minute before the exam begins.  Now, at first, I thought outlining wouldn\'t be too difficult an undertaking.  After all, it\'s just a matter of organizing information...not a problem for one who LOVES to organize.  And really, it\'s not \"hard\" per se, more just hugely overwhelming.  Imagine it...4 months worth of class notes/concepts/theories to get properly arranged in learnable fashion, several of which may or may not have been forgotten around Labor Day weekend.    Only after that\'s begun does the real fun/actual work begin.  You see, it\'s not enough to simply know the words and what they mean.  No, there\'s this little thing called \"application\" that\'s actually a pretty big deal around here, shockingly enough.  I mean, sure, I can rattle off the requirements for a remainder in Property, or the Duty/Breach analyses in Torts, or the Parol Evidence Rule for Contracts, but that does shit on an exam.  The name of the game there is: hypotheticals.   An example:  Defendant places white, granular rat poison in an unlabeled can near a coffee pot sitting on a hot plate.  (Why is this a careless, negligent thing to do?  What harm could/should be expected?)  No one (including defendant) knows the rat poison is flammable.  It blows up and physically injures plaintiff standing by the hot plate to pour himself some coffee.   Did the defendant proximately cause plaintiff\'s injuries?  Should defendant be liable?    That could be a short-answer question, or it could be the basis of an entire final Torts exam, depending.    Similarly,  To A for life, then to B and her heirs To A for life, then to B and the heirs of her body To A for life, then to B if B is 21 before A dies; B is 15 To A for life, then to B and her heirs is B survives A, and if B does not survive A to C and his heirs To A for life, then to B and her heirs, but if B does not survive A, to C and his heirs Are all grants of property, after which the question will inevitably be: what is the state of the title?  Now, all of this is not, in any way, meant to conjure sympathy, pity, wonder or awe...although if you feel the need to express any of those through monetary donations or physical affection, feel free. Instead, it\'s meant to just give a little glimpse into life as it is for me at the moment, which, I suppose, is the whole point to a blog in the first place.  Jealous? I know it doesn\'t look like loads of fun on the surface, but really, it\'s a great time.  Just hope I can remember that in the coming weeks... And now, I shall return to work:  Contracts reading for tomorrow\'s class.  Yes, in the midst of all this action, we are still learning more material that will of course be included on the final.   Sigh... The drive on Friday will feel good.   '),(23173,2314011,'18,November,2003','2003-11-18 00:00:00','    hi bloggy! actually pretty tired from another action-packed full day, but still wanted to post a little something that wasn\'t just a cryptic one-liner, so we\'ll see how it goes.  so the new Tori collection\'s out today.  i have to say, i was under the impression most (if not all) the songs included would be re-worked, if not entirely, at least noticeably.  apparently that\'s not the case; so yea, basically, it\'s $17 for 2 new songs, 2 re-recorded ones and 4 live sound-check performances.  not a bad deal, but still.  the packaging\'s great, though.  think she must have been too busy on her 13 month epic tour to care a great deal about her last contractual obligation to Atlantic to re-work all 20 songs, sigh!   but the journal entry/poem was a nice touch; she\'s in London tonight honey!    the cold\'s pretty much vacated my throat and is beginning to lodge in my chest/lung region.  yay.  still determined to beat it with vitamin C tablets and water, but i may have to break down and go on a NyQuill run before i leave town.   sooooo....not a whole lot to mention, really.  suppose i could take suggestions if anyone has any.   '),(23174,2314011,'17,November,2003','2003-11-17 00:00:00','    ...don\'t rush out and get it, hon, if you have more important things to do.  I\'d say more to the general public, but I\'m in Civil Procedure, and should listen...   '),(23175,2314011,'17,November,2003','2003-11-17 00:00:00','    got online after recovering from the sniveling mess the cd left me in only to find the last buddy i need to have on my list right now.    told her i\'d never block her, but ...   '),(23176,2314011,'16,November,2003','2003-11-16 00:00:00','    I LOVE NEW MIXES!!!!  Seriously, there\'s nothing else like \'em.  Especially when your best friend in all the world makes them.  Honestly, he makes the best mixes around.  Can\'t wait to get into the car and drive with the November Mix.  Yes, we name our mixes after months; got a problem??  We do seasons too!  Good god, such a great one hon!  Who knew Dido was so great?   WHO??  Tired to listen to the Tori B-Sides mix, but only got to the 3rd song before I had to just shove November back in.  Currently on \"When You\'re Stoned\"...it\'s my instant favorite.  Seriously jumped when her voice started after the great instrumental background.  And then fucking \"Honey\" after that...WHAT AM I GOING TO DO WITH YOU???? And as for \"Love Him\".....GGGGRRRRRRRRR!!!!!! Have to say, hearing the Cranberries\' \"Can\'t be with You\" on the mix was kinda bittersweet.  Recognized the intro immediately, but for the life of me couldn\'t figure out what the hell it was.   Once the lyrics began, I was back in my old room in the mountain fortress, 14, on the blue carpet playing UNO with Tracy.  Oh yea.   To be followed by only Madonna song that makes me feel like I\'m Seattle in summer just as much as \"Hollywood.\" So basically, the whole damn things rocks the cock all the way to the moon, honey.  Really.   And I didn\'t intend this entry to so closely resemble an email, but what the hell?  And yes, Track 9 totally knocked me down.  Sniff!   '),(23177,2314011,'16,November,2003','2003-11-16 00:00:00','    Circles! Just about to pack it on outta here; put in a full day and feel pretty damn good about it.  Feelings pretty much going in the opposite direction they were when I arrived, and we love that.  Think my productivity has a lot to do with it.  This morning saw work/duties/responsibilities piling up, landing squarely on my shoulders (sob!), and this evening has that load lessened and manageable...KICK ASS!   And the only way to celebrate is to go home and do a little bit more!   ...Might have a little package waiting for me, too...I dunno! Till next time; over and out.   '),(23178,2314011,'16,November,2003','2003-11-16 00:00:00','    As I was coming out of the haze of Contracts reading, my desk-mate pulled out the chair next to me and said he\'d been chatting with a gal from my class who, when he told her who his desk-mate was, said: \"Ah, the smart one from our class.\"  There are others among us I would more readily describe that way, but still; it was nice to hear.   '),(23179,2314011,'16,November,2003','2003-11-16 00:00:00','    The phone rang around 8.  My heart stopped beating and when it resumed, it was in my mouth, and I couldn\'t breathe.  No one calls me at 8.  Almost no one.  Something had to be wrong for her to call.  And in spite of the pep-talks and three pages of writing I\'d just done on the subject, I desperately wanted it to be her.  It was my brother.  He needed my plunger. After he left, I couldn\'t escape that feeling.  The feeling that this is going to take longer than a few weeks.  The feeling that this is going to be harder than just fighting off the urge to call to check in.  Then that damn \"Only Freshmen\" song came on and I had no choice but to turn it up.   Got to school nice and early on account of not having Property class this morning.  Double checked my citations in the infamous paper and discovered two were wrong.  Not to fear, I\'d brought a back-up disk.  Funny thing:  12 pages in Word Perfect equals 12 pages and 3 lines in Word. It should be mentioned that I got perhaps 6 hours of interrupted sleep last night and am feeling a bit jittery from the combination of coffee and cough syrup.   After trimming and fighting with the Insert Page Number feature of that wonderful word processing program, I finally turned in the paper I\'ve had done since goddamned Saturday.  Thought I could collect myself at my computer, read a blog or two, get geared up for the big week before the big break before the big finals.  Starting to feel the pressure/stress/nerves of those exams.  Feeling like I\'ve got 3 weeks to pull 4 months\' worth of material together into a congizable form.  It\'s probably just the usual cold-induced emotional wax/wane thing.   And then I felt like a self-indulgent shit when I read Dave\'s blog.  Dave, I\'m sorry.  There\'s really nothing else I can think of to say, except that I\'m sending you thoughts of strength, comfort and love.  ...And it\'s not even 10 a.m.   '),(23180,2314011,'16,November,2003','2003-11-16 00:00:00','    Mrs. Sarraf was right.   But I love it more than I hate it.   '),(23181,2314011,'15,November,2003','2003-11-15 00:00:00','    Property\'s fun.  You wouldn\'t think it to hear the course title.  \"What do they mean, \'property\'?  Like what\'s \'ownership\'?  They just going to teach us what kind of possession we have over inanimate objects/land/other human beings?  What kind of sick twisted shit is THAT?\"   As it turns out, it\'s actually quite interesting, complicated, frustrating, at-times incomprehensible, and an organizational nightmare that I\'ll have conquered very soon because I have no choice.   Recent topics organized/consolidated into outline (study-able) form:  Future Interests :  Retained by transferor : a.Reversion: when the quantum of the estate given is less than whole, retained by grantor b.Possibility of reverter c.Right of reentry   Retained by transferee : a.Vested remainder b.Contingent remainder c.Executory interest   Remainder : any future interest created in a transferee that can (IS CAPABLE OF) become a present possessory interest immediately upon (BUT NO SOONER THAN) the expiration of prior interests simultaneously created and which cannot divest (cut short) any interest other than the future interest in the transferor.  Vested Remainder : receiving party is presently identified and known (ascertained at the time), and only natural expiration of all prior estates created with it is needed for it to become possessory; itâ€™s NOT subject to any condition precedent  Contingent Remainder : identity of receiving party unknown and  interest is subject to a condition precedent, something must happen before the interest vests other than the natural expiration of all prior estates   Sigh...  Life is good...   '),(23182,2314011,'15,November,2003','2003-11-15 00:00:00','    Morning, blogger I can still say that, since there\'s 42 minutes left before noon.  I should be getting my ass to school to study study study for finals, but I just want to sit on my stylish futon and watching bad but good movies all day.  Plus, I\'m getting sick.  Yes, the first tell-tale tickles of a sore throat started yesterday, and this morning the sinuses are starting to join the party.  Yay. At least this week\'s an easier....well, let\'s just say lighter....week than last.  And the Thanksgiving vacation begins Friday at 1, so there\'s that too. But I just had to write a little something this morning because I was inspired by my boyfriend\'s new-born blog.  HI HONEY!!!  Suppose you all should get used to hearing about him.  His name\'s Levi, but I may refer to him at any time as \"the boy,\" \"that guy,\" \"my guy,\" \"my boyfriend,\" or \"the Jew\" (because he\'s not).  And really, he\'s not my boyfriend either, but that\'s part of the joke, get it? Alrighty...I\'ll be plodding along now...have to see about changing the body text color on this damn thing to make it easier for some pansy-asses to read!   '),(23183,2314011,'15,November,2003','2003-11-15 00:00:00','    ...that the damn thing was 13 fucking pages to start with...  okay really, i mean it this time...as you were.   '),(23184,2314011,'14,November,2003','2003-11-14 00:00:00','    I\'ll tell you how:  By thinking like a journalist.    Suppose that journalism degree will come in handy, after all.    Yes, managed to edit that sucker down to 11.5 pages, only to be faced with 4/5 of a page of case citations....GRRRRRRRR!  But, another editing pass ended in a successful page count of exactly 12, with not a single line to spare...ahhhhh.  As you were...   '),(23185,2314011,'14,November,2003','2003-11-14 00:00:00','    Can anyone tell me how I\'m supposed to write a legal analysis of an employee\'s application for unemployment compensation, which her employers are opposing on the grounds that she was fired for misconduct, in under 12 pages?  CAN YOU?!!?   '),(23186,2314011,'14,November,2003','2003-11-14 00:00:00','    Hi blogger! So, it took me a minute to figure out how to post another damn entry on this confounded thing; but, through perseverance and a bit of common sense, I managed to do okay...sigh! I\'m thinking this might be the perfect little supplement/substitute for my daily, morning pages I write during the week.  For whatever reason, I didn\'t feel like coming home (from Yoga class, we\'ll get to that in a minute), sitting down at my kitchen table and writing three pages long-hand in my trusty notebook as is usually the case every weekday morning (except for most Fridays).  BUT...logging on and posting a blog sound just great!  Whatever. It\'s probably the \"newness\" of the whole thing, or maybe a hint of excitement at joining the celebrity list of those with blogs...I dunno. Anyway, had a nice yoga practice this morning.  Not a fan of getting up at eight on Saturdays, but for a bit of breathing, stretching and lunging, I make do.  Found a new coffee place on the way home, too, which we totally love.  But back to yoga for a sec...basically, it rocks the mullet all the way to Texas.  And that\'s all there is to it.  Can\'t get that \"looking up..\" line outta my head, either.  My usual instructor said it in class last week as she was teaching us a new pose.  Not sure how the description will work out on here, but we\'ll give it a go.  Just think about the position you usually take when you\'re about to do push-ups or lower yourself down onto some lucky person beneath you...now, imagine shifting your weight onto one hand and opening your body to the side with one hand on the hip you\'re lifting away from the floor.  Okay, NOW, imagine lifting that hand up toward the ceiling.  So we\'ve got our legs straight...I mean RIGHT out in front of us, and one hand supporting the whole body while the other\'s lifted toward the ceiling.  Fun, huh?  Actually, there\'s a modified version, but I\'m not going to explain it on here, so deal.  Anyway, once the balance is found (gasp!) the finishing touch is to lift the head up to look at the raised hand.  And that usually obliterates the balance that was tentative to begin with.   So yea, looking up changes everything. Nice little mantra for other stuff, too, I imagine.   Wonder how these little pages of blog-wonder compare with notebook pages...hmmm.  Doubt I\'ve typed as much as three pages would be written, but still...feels like more than the usual morning pages amount.  Suppose I could explain those...they\'re really quite self-explanatory, but if it\'ll add more type to the entry, I\'m all for it.  Basically, morning pages are written in the middle of the night, right after you fall into a deep, relaxing slumber.  Then you jolt yourself awake and write everything you were thinking about right before you fell asleep, along with whatever luscious dreams you were about to have, were it not for your obligation to write these damn pages.  Kick ass, huh?  It\'s great for self-examination and personal growth. Alrighty....I could keep rambling, but I know you\'re all busy souls with things to do and places to be and I refuse to let you use my blog as a procrastination device any longer! Besides, I\'ve got to be a brilliant law student today and I need time to prepare myself for the job.  A shower seems like the best place to start....   '),(23187,2314011,'14,November,2003','2003-11-14 00:00:00','    Okay, soooo... I guess we\'ll give this a go; suppose I should acknowledge my blogger-mentor, David Jack, first and foremost, as his little suggestion that I start a blog of my own tilted me over the \"I wonder if  I  should start a blog\" cliff I\'d been teetering on.  Yay Dave! So yea, here we are kids, hope it lives up to all the hype....   '),(23188,2314011,'30,December,2003','2003-12-30 00:00:00','    We\'re trapped in the mountain fortress for yet another day.  It\'s been surprisingly nice.  Usually, we find ourselves feeling incredibly guilty for spending hours (if not days) on end doing nothing more than showering and eating during a day\'s passing.  But now that we\'re forced to do just that because of the several feet of snow in the driveway and the continually non-plowed streets of town.  Soon we may have to start eating each other to survive.  That\'ll show \'em who\'s boss.  Made it out to see Cold Mountain yesterday then spent the evening finishing the last 100 pages of the book.  Think I\'ll have to see it again before I deliver my final verdict.  Overall, I really enjoyed the movie, but the book will always have my heart.  The three main characters translated beautifully to screen, but the remaining supporting roles were trimmed significantly to fit the movie, which left me a bit miffed.  The gals shared a kiss, though, and we love that, Civil War be damned!  Today\'s only possible outing will be to dinner with Levi\'s aunt and uncle.  Yay.  They\'re good people.  Plus they have a hot tub.  Spent a while at their place the afternoon leading up to the night of gay dancing glory.  His aunt\'s a bigger, brasher, funnier-in-a-different-way version of his mom.  Good damn times.  So mexican food and margaritas with them sounds like a good excuse to risk our lives on the streets of this frozen city.    The boy says I should write about how much fun we\'re having.  We\'re having so much fun.  I\'m starting to dread his leaving a wee bit.  But if the weather keeps up this shit, he won\'t be able to leave and he will be mine.  He will be my squishy.  That\'s my teapot\'s name, by the way.  Got it for my birthday, so excited.  No more of the microwaved brew for me!  Received my first lawyer-to-be power suit, too.  It\'s so intimidating, I\'m even scared to put it on.  But I\'m supposed to be telling you about all the frolicking we\'ve been doing, sorry!    We\'ve been doing Yoga like no other!    \"Let\'s just do some stretches.\"   \"Okay!\"   Cut to an hour later.    \"And this is where you put your hand on the sky and lift your ankle across the room to the mirror before breathing out of your ear!  Let\'s go!  Two on each side!  Remember to breathe and lift up through the heart!  Change the CD while you\'re over there, will you?\"    We\'re so limber.    Listening to the Madonna mix now, remembering the other night at Charley\'s, missing Seattle, and loving being forced to relax.   '),(23189,2314011,'28,December,2003','2003-12-28 00:00:00','    Especially ones that go on for weeks.  Getting pretty cozy indeed with all this free time.  Been recovering today from last night\'s exertions, I\'ll get to those in a minute.  It\'s been a few days since my last post, so I\'ll just take the days one at a time, m\'kay?    Friday, December 26  The big birthday.  Yay!  Mom woke me around 9 to wish me a happy day and to tell me we had a big blizzard raging outside.  I knew when the Boy said he\'d drive from Twin Falls unless a storm broke, there\'d be a storm.  So we waited for the phone to ring.  After much warring, he decided to take the plunge.  The only way Mom and I could make good use of the 3 feet of snow was to go to town.  No force of nature was going to keep us from TJ Maxx, by god!  Couldn\'t get back up the driveway, even in the big truck, and as we left the men to dig and pull it out, the Boy called to let us know he\'d made it to town, if a bit shaken up by the treacherous road behind him.  We finally got the driveway plowed, and the Boy picked up, and there was much rejoicing.  No partying was done, however, until the next night, but the obligatory candles, cake and ice cream were not overlooked.  Remember how I said The Phone Call would be happening soon?  Yea, well, it turned out to be my birthday present to myself.  Was worried she wouldn\'t be home and I\'d have to decide whether or not to leave a message.  But she saved me from that dilemma by answering on the 4th ring.  She had bronchitis and didn\'t sound like herself.  It was different and the same.  Finally laughed a bit at the end, which felt good.  There will be a period of adjustment, that much is clear.  But whatever happens now, I know I\'m ready for and can handle it.  Gave her the address for this blog; everybody wave.      Saturday, December 27  Regaled Mom with the high points of Crazy Aunt Karen\'s reply email, and she in turn regaled Levi and me with stories about the old broad.  It was decided by some unspoken agreement that he and I would not be caged in this house any longer, and we made our way into the bright, blue, white wonderland.  It was to be the only break from a string of storms continuing today, and we were not about to watch it from the living room.  Met up with Natty-poo and her world-traveling English teacher mother for a few good fondles.  Then on to dinner with the entire fam.  Yay.  Joe worked his magic one-liners on Nat as they played When a Man Loves a Woman for the 6th time.    The three of us at last descended on the only gay bar in town, for a few drinks to begin with before moving on to other scenes.  Found a drag show waiting for us, however, and ended up staying until 2.  Levi caught the eye of a cute waiter and we didn\'t pay for another drink the rest of the evening.  Good times.  We drank, vogued, and were forever young all night.    So yea, recovering from it all today.  Just finished X-Men 2, and might curl up with a little booky in a minute, I dunno!    Lazy days are our friends.   '),(23190,2314011,'25,December,2003','2003-12-25 00:00:00','    Merry Christmas, blogger!  In two hours the house will be not all that filled with family and flowing wine.  The Grandparents as well as Uncle Jim and Aunt Amy are joining us for the first Christmas dinner in our house; should be a good time.  For now, though, the fam\'s settling in with the new prezzies, Mom\'s trying to keep dinner under control and I\'m hiding in my cave to reflect on it all.  Snowed all night and this morning the sun\'s breaking through the leftover clouds.  Yay.    Just read the Boy\'s sentimental Christmas Eve entry and am feeling a bit fuzzy myself.  Started last night with the home movies.  Joe\'s first Christmas was the best.  Between watching him spit up, tip over Mom\'s coffee, follow the table knife around the old living room, and me screeching at each new wrapped wonder set before me, the whole thing was just the right amount of too much.  He actually was a cute damn baby; too bad I was too young to stop tormenting him to notice.  Had a few glasses of Merlot over at the New House.  Listened to Grandpa tell us about all his vintage Marilyn Monroe paraphenalia.  He has an original, life-size poster of her, the original first issue of Playboy featuring her features on the cover, and the box of candy bearing her first commercial image.  Told him he could get more money for those three things than he did for his house.  He sold six boxes of vintage National Geographics for $100.  I prefer not to think about it.  Hopefully we\'ll (and by \"we\" I mean \"I\") get our hands on his pop culture relics before anyone else.  But it was nice to sit with them, helping them not be so lonely on their first Christmas Eve in 43 years outside their home.    Looking forward to tomorrow.  And the next day.  There\'s nothing like spending lazy days on end with your best friend.  We haven\'t done it for about three years, so it\'s only fitting it happens here again.  Major revelations and unbreakable connections were make back then; can\'t wait to see what happens this round.  It\'s really the wrong day, date-wise, for this little passage, but it\'s just the mood I\'m in today, okay?    ...the day did come when they were both bored with each otherÃƒÆ’Ã‚â€š&#146;s respective homes and sought external entertainment.  They visited their old stomping grounds, checking in with favorite teachers and restaurants, but soon they were looking for something else to fill their day.  Neither wanted to part ways, but neither knew what to else do.    \"Well, you wanna get some coffee?\" he asked. \"And I imagine we have to get you a mocha, right?\"  \"Damn right.  Just deal.\"  He was restless as he stirred in his cream and sugar:  \"I don\'t really feel like sitting here and flipping through magazines.\"  \"Ummm, okay ...\" she carefully sipped her drink, deciding.  \"You wanna go for a drive around the winter wilderness?  Go admire the harsh beauty of it all?\"  \"Hmmm ... harsh beauty ... that sounds good.\"  This was something she had yet to share with him: random, lazy drives.  And something told her this would be an important one.  It called for a special route.  \"So, where you taking me?\"  His raised voice asked over the music as she took them into the mountains.  \"Oh, just in a big loop,\" she answered coyly. \"We\'ll come to a point in a while where we\'ll have to decide if we want to come back right away or keep going.\"  It was a perfect day for either choice.  By the time they reached the first peak and descended upon the rolling countryside, the sun had emerged, making January\'s latest snowfall impossibly bright, accentuated further by the baby blue sky.    They let silence stand between them; let the music fill the emptiness, simply enjoying being in each other\'s presence.    Over the past few months, they\'d discovered the wonder of wordplay.  Their favorite toys were the truly offensive, derogatory-if-spoken-by-anyone-else types.    \"Why do you have to be so damn gay all the time?\" was a staple.  \"Stupid fags, I swear!\" was another.    They were a damn riot.  Just ask them.    So when he turned to her and said, \"So, you\'ve been waiting for me to come out, have you?\"  she jumped right on the joke-train and answered, \"Yeah, that\'s right.  Enough fucking around, it\'s time to admit it to yourself and everybody else,\" laughing all the while.  \"Okay, well ... I\'m gay.\"  \"Oh really?\" she laughed.  \"Really.\"  \"Well, that\'s just super!  Tell me, how long have you known?\"  She continued to question him and mock his answers, never once believing he was serious.  But soon she realized he wasn\'t laughing with her.  And her stomach hit the floor.  \"Are you serious?\"  \"I\'m totally serious.\"  \"No, really.  Are you really fucking serious?\"  \"Really, I\'m really fucking serious.\"  She pulled the car over, hoping he would start laughing or just say \"Gotcha!\"   \"Okay, really.  Are you serious?  Because either you\'re kidding or I\'m a huge fucking asshole.\"  \"Well, then you\'re an asshole because I\'m not joking.\"  Shaking, she pulled the car back onto the road.  She asked him the same questions over again, this time mulling over the answers, trying to pull herself together to be the best friend she\'d  considered herself to be this past year.  \"I can\'t believe you\'re this surprised,\" he said.     \"Well, I honestly have never thought about it.  I already know I have no gaydar, so that\'s no surprise.  When people in high school used to say, you know, \'Levi\'s so gay, just look at him,\' I would be like: \'Oh come on you guys, he really isn\'t.  Trust me, I know him pretty well.  And even if he is, he has no idea.\'\"  \"I really didn\'t back then.\"  They took the long way home that day.     Their AP English teacher once prophesied: \"You two will know each other\'s grandchildren.\"  It may have been a nice sentiment at first, but after the day of harsh beauty, it quickly became a promise.  The winter break continued much the same way, only with a few more hugs and arms around each other\'s waists.     Big sigh.  So, I think the phone call will happen soon.  Found myself thinking of opportunities last night.  When I have the house to myself for an extended period of time, etc.  Don\'t want to make it on a tempermental cell phone in a semi-public setting.  The best place and time is probably in the Moscow apartment upon my safe return.  Just don\'t know that I want to wait that long, though.  Wonder if that\'s a red flag all by itself.  Finally feel ready to do it.  Enjoyed the break and all, and am enjoying the knowledge that they can be taken whenever the need arises.  Might be a nice little Christmas present to us both.  Think I\'ll email Crazy Aunt Karen now.  She included her address in her card this year; apparently I\'m the only member of the family she\'s cared to invite contact from.  Must have been all those Thank-You cards through the years.  Hmmm.  This could be interesting, indeed.  Merry Christmas, everyone.  Love you all; if I can\'t say it to each of you, know that I\'m thinking of you and wishing you the best.   '),(23191,2314011,'24,December,2003','2003-12-24 00:00:00','    Evening bloggy!  Happy Christmas Eve!  Can\'t believe it\'s here again.  Last year at this time we were down at Grandma and Grandpa\'s, throwing back a few egg nogs, admiring the shaggy carpet we were beginning to miss but didn\'t know why and trying to make my passport appear in the mailbox two days later.  I had plane to catch on the 27th, after all.    Spent the day seeing Return of the King again.  All else feels so insignificant when you come out of the theater after that one.  Cried a bit again, although no where near as much as the first time.  Made it though the first hour of the first viewing just dry heaving and the next two in wrenching sobs.  Dad will be home from worky soon; might venture out again, as a family, to another movie.  Not sure how I feel about visiting the same theater in the same day, but it\'s Christmas and there might be a chance of catching some naked Diane Keaton action...PARTY!    Talked to Natalie earlier today.  She\'s so damn cute and funny and hot.  And that\'s just over the phone!  She\'s the only gal from high school I still try (I should try harder, I realize, and I make no excuses) to track down whenever I\'m in town.  I watched her performance in Highland\'s Dracula production in awe (and a wee bit of attraction), and nearly swooned when Levi brought her to the Mountain Fortress Get-Together the eve of the last day of junior year.  \"Your bathroom is bigger than the blue banana!\" she said as she made her way through the house at her own pace, and we haven\'t looked back since.  Imagined her zipping all over town in her new VW Golf without a care or worry about carbon monoxide in the world.  Forgot to ask her what she named her new charge.  If she hasn\'t named it yet, I think it should be called Gimli.  Not sure why.  She\'s coming out with the Boy and me to help celebrate my birthday Friday.  Should be nothing less than a sweaty, sticky, drunken good time.  Haven\'t seen her since this summer, but I know we\'ll all pick up right where we left off.  That\'s what\'s so great about her, really.  No matter where you\'ve been or what you\'ve done, when you\'re together again, it always feels like you never were apart.   '),(23192,2314011,'23,December,2003','2003-12-23 00:00:00','    So I spent my entire shower coming up with a great way to tell the story of lunch with Mom, The Boy and the hot gay waiter in Idaho Falls.  Had a opening and thought the rest would flow easily.  Finally made it to the computer, checked out the other bloggies of note only to find the Boy\'s own rendition of the incident on his blog.  Well, never being one to compete with greatness, I just copied and pasted the story from his.  Without permission, of course.  (Editor\'s note: the \"S\"s originally in my mom\'s name on Levi\'s blog have been changed to \"C\"s.)   Andrea\'s mom is a golden piece of work. A source of endless humor and entertainment without even knowing it. She can say things like \"Looks like you just shot your wad!\" without a clue of its sexual implications. That\'s precisely what we love about her. Our lunch was stacked with brilliant moments where Andrea and I were convulsing in hysterics and poor Cheryl was just wishing she could go sit at another table. \"You damn kids...\"   I can\'t wait to spend a week with her.   We all met at a crowded Johnny Carino\'s in Idaho Falls, waiting in the lobby with the herds of Mormon families lumbering their offspring in and out of carriers and over shoulders. There\'s no place like southern Idaho.  \"If I had known it would be like this, I would have picked somewhere different.\" \"Oh, like a place where they forbid children?\" Not likely, in this state.   The girl and I haven\'t seen each other in nearly 3 months and our first words are to tell each other how damn hot we both look. And really, we\'re each at the top of our game and loving every damn second of it. She\'s let her hair grow into layered, shaggy fronds (very sexy), where I\'ve chopped my locks for a slightly cleaner, sharper look which I\'ve come to prefer, set off by my scruffy cheeks. She in her leather and I in my wool, we\'d be all over each other if we both weren\'t gay...and sometimes we don\'t even let that stop us.   The event begins with an innocent vibrator joke from Cheryl when the buzzer tells us our table is ready...to be followed by Andrea\'s comment of: \"I don\'t pack mine out of the house, mother!\" and it pretty much grew from there. Things got better when our waiter revealed himself...no doubt in any of our minds on which side he liked his bread buttered. Something needed to be said, but how? Cheryl had only recently learned of my own leanings (which is a great tale in itself) and I wasn\'t really sure how we\'d be dealing with the topic.   \"What is it with Johnny Carino\'s?\" Andrea\'s always great with openings.  \"God, I know...it\'s like a secret brotherhood or something. But he\'s probably just a Mormon.\"  \"Wouldn\'t be surprised, with that hair.\"  Cut to Cheryl shaking her head in disbelief, mumbling something like \"you kids are terrible.\" but knowing deep down she adored our irreverence. She couldn\'t help bring up last year\'s Passport debacle, which is always good for a laugh or two. We get a glimpse of the hidden Cheryl, dark and saucy and wild. She could be unstoppable if only she would let herself.   Rather than help Cheryl pick out a watch for Christmas, the girl and I ditched her for our obligatory coffee-talk. I drove her back to her mountain fortress in Pocatello and didn\'t want to leave. We have so much to catch up on! So much to do and say and laugh about! (Like her mom!) But I had to forge onward...with the promise that I\'d return bright and early on her birthday, to continue our adventure. I always leave her with the thought, \"Remind me again why we don\'t live in the same town? why did we choose to do this the hard way?\"   Good times.  Been hanging around the house this morning, going through the motions of coffee, getting ready, etc. at a very leisurely pace.  Love these days on break where absolutely nothing needs to be done and we\'re okay with that.  Had a great half an hour Yoga practice and am now ready to catch up on a little writing.  Yay!  Finally listened to Dido\'s new CD in its entirety while pounding out those Sun Salutations; stupid European blondes anyway!  Now back to the November Mixy.  It\'s brilliant, really.    So yea, the break\'s already feeling really great.  Took a bit to slide into, but now feeling utterly content to let the next three weeks unravel at their own pace.  Christmas is in two days and I\'m actually really excited for it.  Get to wrap some presents today.  Mom\'s sick of wrapping and tried to get me to wrap a couple of my own presents.  That\'s where I drew the line; it just seems wrong in some way.  Might go to a little movie later, I dunno!  I realize this entry\'s a bit disorganized and little more than a smathering of random thoughts.  Maybe I should actually tell a story, hmmm.    The night after finals isn\'t really worth going into, except to say a more-straight-than-the-bi-vibe-she-wishes-she-gave-off-girl who likes to compete for attention in any way she can fell all over me all night.  Not in any way attracted to her, partially because she\'s a straight guy\'s fantasy of a lesbian (read: Hetero-Porn version of a lesbian [read: not a lesbian at all]), and partially because of the previous sentence\'s description.  But hey, we love attention and flattery.  She told all the men hovering around our lesbian sex discussion that she\'d rather take me home than any of them.  So glad I don\'t have to see any of them for a month.    Almost finished with Cold Mountain.  The Boy decided we\'re seeing it together, so I\'ve got a bit more time.  Trying to figure out how the movie\'s going to handle all of the characters Inman (that\'s Jude) encounters in each chapter.  The previews don\'t really give much away, so I\'m hoping it\'s a good surprise.  Wasn\'t a fan of The English Patient.  Had the same reaction Elaine from Seinfeld did:  \"Just DIE already!!\"    Sooo, what else?  Oh!  Getting new glasses!  Yay!  Mine met the working end of a pool cue last week, so you know, that had to be remedied.  They\'re fucking expensive, though.  Jeeez!  That\'s where the above-described \"shot your wad\" line originated, by the way.   '),(23193,2314011,'21,December,2003','2003-12-21 00:00:00','    Hey bloggy!  Made it home safe and sound.  Sigh.  Still really tired, though, which is fucking my shit up.  The trip was good, the freezing rain in Moscow notwithstanding.  Relaxed a bit more as each mile flew by.  Had Christmas in Burley today.  Tomorrow\'s shopping in Idaho Falls.  Yay.  Wonder if the boy will want to do lunchy; hope he can handle Mom.  She\'s really fun when she\'s been getting good deals.    Might get into the details of the night after finals at some point; now, though, it\'s time for bed.   '),(23194,2314011,'19,December,2003','2003-12-19 00:00:00','    Made it through.  Just barely.  Still feeling worn and haggard, but think it\'ll be alright.  Sky\'s shining out the clearer, just as Sam promised.  The exam was an adventure, too.   '),(23195,2314011,'18,December,2003','2003-12-18 00:00:00','    Well, the sun\'s out and that\'s a good thing.  It was out yesterday too.  Ordinarily, this wouldn\'t be that much of a news item, as, theoretically, the sun rises every day.  But for the last 2-3 weeks, we\'ve seen very little of the big guy, so now when he\'s out with the blue sky I find myself at the window, looking up and saying \"That\'s the most beautiful thing I\'ve ever seen.\"  Kinda like getting out of a mental institution for an afternoon and seeing a streetlight for the first time in a year; at least that\'s what I\'ve been told.  So today\'s the big-ass day, on more than one level.  I don\'t even want to leave the house, but I\'m counting the minutes until I do.  Can\'t sit still long enough to write morning pages, so instead, the computer will get a nice warm up along with my writing skills.  Don\'t know why I\'m flipping out about this particular exam.  Maybe it\'s because it\'s the hardest one, the one sure to knock everyone off their chairs, force us back up and deal with it as best we can.  So over-dramatic, I realize.  Could be that it represents the entire two weeks and 4 months preceding it, too.  Yes, we\'ve been on this journey and we haven\'t turned back, but now comes the last step and it\'s the most difficult to take.    The boy\'s watching Return of the King right now.  Wish I was with him.   '),(23196,2314011,'16,December,2003','2003-12-16 00:00:00','    And by \"you\" I mean \"her.\"  Not in a longing, \"Oh I wish it didn\'t have to be this way; maybe it doesn\'t, gasp!\" kind of way.  More in a \"I can\'t wait to talk to that gal again, she\'s fun!\" kind of way.  Been feeling distance between us for the first time.  Distance that feels good, which is an unusual place to arrive at.  Whenever the daily thoughts arise about her, I\'ve felt really good about the situation, where I\'m at with it all, and very excited about that first conversation.  Not so excited that I\'ll rush to get to it, because I know I\'m not ready for it yet, but just knowing that it\'s going to be so good to talk and laugh with her again.  Expecting a bit of initial weirdness for a bit, as we tentatively come back into our friendship, assuming the new roles and all.    Hoping she\'s doing well and this time\'s been as positive a transition as it has for me.   '),(23197,2314011,'13,December,2003','2003-12-13 00:00:00','    AOL Member home pages are gay.    I mean really, really gay.  And not even in a cool way.  Guess we\'ll just have to figure out other ways of spreading my, and others\', brilliance around via the blog universe.  My idea was that I could create a page with a bunch of my other writings and kick ass images from the boy, assuming he\'d want the advertising, but our stuff is far too valuable and of such greater quality than the AOL member homepage designs that I won\'t subject the material to such squalor.  So there!  Suppose I could just post my other writings a paragraph at a time on my \"daily\" entries.  That means you\'re screwed hon, sorry!   '),(23198,2314011,'13,December,2003','2003-12-13 00:00:00','    What blog?  What\'s a blog?  Okay, I know it\'s been a while, and I usually love giving people shit for not updating their blogs a semi-daily basis, but damnit, I\'m not people!    We\'re getting down to the wire now, one more exam to go, and it\'s the mother of all exams.  It\'s a monster too!  Stupid contracts.  Nobody ever uses them anymore...they\'re sooo outdated.  Think I\'ve got all the rules down pretty well, it\'s just that stupid application crap.  I really dig the professor; she\'s very much a sweet old lady who\'s a joy to talk to as a person, but she\'s been teaching this stuff for nearly 30 years and knows it like she knows breathing.  So she\'s setting the standard very high on the exam, which is a nice little challenge, but also a bit disconcerting when you\'ve worked hard all semester, know pretty much what\'s going on, but  in your practice exam answer you only get to half of the concepts she gives in her Model Answer.  Don\'t think the day before this exam is going to be as relaxed as the other \"days before\" have been.  Had a pretty good study day today.  Changed my study-pals as the usual one\'s not quite at her peak.  We zoomed through our outlines, and by 130 they were all packing up.  Discussion of the rules we all know only gets you so far, after all, as previously explained.  So after a mid-afternoon break in Student Health with said usual study-pal, managed to work on a couple of practice questions.  Feeling pretty good about them overall, it\'s just that pesky extra one-half of material on the professor\'s answers that\'s bothering be just a titch.  But yoga\'s coming up here in about an hour, and we love that.  Going to go over more questions and answers tomorrow, too, so all should be well.  Yay.  The Property exam was entirely too easy, but I\'m choosing to rejoice in that fact.  Our section was allowed to bring in any material we worked on (class notes, outlines, etc.) as well as our book, which was a bit helpful if for no other reason than to extinguish any excuses for not knowing the requirements for adverse possession.  Felt bad for the other section\'s students, though.  Here we were with all the material from the entire course spread out around each of our computers and their areas were just plain empty.  Poor suckers.  Have to laugh at the other university students talking about finals week.  Feels like we\'ve had finals Month.  The last two weeks of test-taking have felt longer than the four months it took to learn the material.  Now that it\'s almost over, it feels a bit bizarre.  In 48 hours, none of us will even remember our names.  If I don\'t post much after tomorrow, it may well be that any memory of my blog\'s existence was wiped out Thursday night.    So I\'m thinking about setting up a Web page so I can use the links on here.  Seeing as I\'m a cheap whore who doesn\'t want to spend the $6 a month to have a Super Blog on which I can post pictures, I think I\'ll take advantage of my overpriced AOL internet connection and make a free Web site.  Yay!  Not that anyone would be interested in seeing anything I\'d link to, but still.  Then I could say I have a Web page AND a Blog!  My god!  There\'d be noooo stopping me then!  Hmmm....think I\'ll go investigate that very matter...   '),(23199,2314011,'13,December,2003','2003-12-13 00:00:00','    Just finished the extended DVD of the Two Towers and all I can think about is last January.  Don\'t think I\'ll ever be able to watch that movie without remembering that first night out.  Even watching it in the Pullman theater and now on the futon, still can\'t get passed that smell, taste, feel of London.  Random days of adventuring flashed behind my eyes as the movie carried on in front of them.  Making our way out of the house around 11 each morning, bracing against the winter chill of the Atlantic.  Wandering through Harrod\'s food halls and bookstore.  Feeling the stomach clench with emptiness as the shutter clicked, taking \"such great pictures\" of blurred celebrities.  Bumping around the white room each morning; turning in circles getting ready, hoping his towel and my hair would dry in time.  Only Starbucks tasted the same as it always does.  But it fit in perfectly in the new adventure.    And now I find myself missing it more than ever.  Sam\'s speech about the great stories will always do me in, but Aragorn\'s dream sequence with Arwen is what I end up remembering that I didn\'t want to forget.  He\'s accepted his duty/role, knows he will follow it wherever it may take him, but has no idea why or how he\'s going to pull it off.  Sounds very familiar.  And her reply is finally spot on.    And then there\'s Cate and her damn Galadriel.  If all the hype and that little place deep within us are to be believed, this third round will mirror just as much as the first two.  We won\'t be able to understand entirely how right away, but we\'ll have some idea.  We\'ll return forever changed, knowing the story\'s finally been told to its end and will carry on, almost as usual.   '),(23200,2314011,'12,December,2003','2003-12-12 00:00:00','    It\'s almost noon and I can\'t seem to get out of my PJs and start my day.  Only got up about 45 minutes ago.  Felt really good to sleep in; think this is as refreshed as I\'ve felt in the last week or so.  Yay.  Remember that Property practice exam I was talking about forever ago?  Yea, think I might actually get to it today.  Not sure what else, though.  Might just be a nice, lazy, relaxed but still getting things done/studied day.  Those can be fun.  Maybe spend some time with the new book later, I dunno!  Hope I don\'t get bogged down in re-learning everything about the Discovery Rule or the three types of interpretation concerning the Hostility requirement for Adverse Possession.  That\'s right!  I said Adverse Possession!  It\'s still around!  So the Property final will consist of two questions.  The first will be a series of land grants (remember that other post with all that giving to letters of the alphabet?) after which we\'ll be required to give the state of the title.  The second will be to write a judicial opinion ruling on a set of given facts.  Imagine the professor will pack as much as he possibly can into those facts so we\'ll have to juggle everything we\'ve learned this semester.  We\'ve got four hours, though, so it should be entirely do-able.  My comfort level seems to be rising in the days before each exam, just as my celebratory mood seems to be rising in the hours following each exam.  Hmmm.  Hope I\'m not setting myself for terrible failure and disappointment.    So the Grandparents are moving out of their house for good this weekend.  They just might have spent their first night in the tiny new house last night.  Got the official word from Mom last weekend that we\'re having Christmas dinner at our house.  This will be the first time in history it hasn\'t been at Grandma\'s.  And she\'s not even going to be up in our kitchen, making anything.  She\'s actually going to let Mom take care of it!  Can you imagine?  Mom\'s worried her stuffing will turn out to be soup, but I\'ve forbidden her from using Stove Top, so we\'ll see how it goes.  We may be asking Grandma to help out next year.    Got a birthday in two weeks.  I\'m always so shocked when I finally do the math and discover it\'s almost my birthday.  Just remember being a kid and starting the countdown right after Halloween.  By the time it finally arrived, it didn\'t even matter because I was so sick of waiting for it.  Have to say, though, I\'ve always enjoyed not having to face the day-after-Christmas let-down.  How do you follow a day like that, really?  Having a birthday works, oddly enough.  And it works every time.  Last year\'s was a bit rough, though.  Stupid passports, anyway.  Very excited for this year\'s.  Gonna have a great gay (but not too gay!) time with the boyfriend and whoever else we can scrape together.  And we won\'t stop until it\'s 2004!    Or 2030!!   '),(23201,2314011,'11,December,2003','2003-12-11 00:00:00','    Hey blogger!  Having SUCH a good night, can\'t even stand it.  If the evening following each exam is going to keep getting better and better, next week\'s going to be Out of Control, I tell ya what!  So the Torts exam rocked the cock.  Or rather, I rocked the Tort exam\'s cock.  Oh yea, don\'t even try that prima facie case shit on me, sister!  I\'ll mop the floor with your prima facie case!!    Went to the Alehouse with some comrades afterward and had some dinner, beer, good laughs and inappropriate touching.  Can\'t get any better than that!  Yoga was sooo our friend tonight.  Instructors in little, loose-fitting cotton tank tops are always our friends.  Plus, the standing-pose practice was great for harnessing all that powerful energy and putting it to good use.  Beer and burger kinda sloshed around, but I was in such a great place that I didn\'t even really notice.  Yay!  Drove all the way home and only then do I remember I\'m out of coffee.  But that just means more time in Harley with the great, uplifting, GRRR music playing and we can always use more of that.  Got the November Mix in the player now and we\'re going great on Track 4 and not a icky clicky noise to be heard...Rock.On!  Okay, got Jack & Karen hour coming up here pretty quick, so I best be off.  Might just be back in half an hour, though, I dunno!   '),(23202,2314011,'10,December,2003','2003-12-10 00:00:00','    The last few days have all looked exactly the same: grey, dreary, soggy.  But there\'s no snow, yay!  They\'ve been warmer than most of October, which is odd, but appreciated.  Just finished the last of my studying for the big Torts exam tomorrow.  I\'m sooo ready for Torts; bring on the rat poison disguised as sugar left next to the coffee pot to explode in a felony committing unauthorized visitor in the laboratory\'s kitchen\'s face questions!!  As me about Duty, I dare ya!  Or even Proximate Cause!  You don\'t know about Cause until you talk Proximate Cause, buddy!  Hope I\'m not getting all oddly confident only to be sandbagged on the exam.  I\'ve taken two practice ones with success, so I don\'t think my comfort is unfounded.  Next week\'s going to be the tougher exams.  Property and Contracts.  Like we need either of them, really.  Going to spend the rest of the day with Susan\'s Contracts study guide.  Probably a good part of Friday, too.    So \"Cold Mountain\" is finally getting good.  Probably has something to do with my finally sitting down and reading it for more than 20 minutes.  Can\'t wait to see Nicole play this role; yet another departure for her.  Suppose none of her roles are really departures, since she has yet to actually define her ability through a particular archetypical role in the first place.  So really, because all of her roles are departures, they\'re all very much Nicole and not departures at all!  Renee\'s going to be great as Ruby, too.  I hope they play up the lesbian angle of their relationship.  You just can\'t tell me that two women working together for survival in the harsh, unforgiving world of Civil War Appalachians aren\'t going to be finding their only comfort in each other\'s arms.  You can\'t!   '),(23203,2314011,'08,December,2003','2003-12-08 00:00:00','    The night went well.  Watched \"Finding Nemo\" with Joey Bear after dinner; had tears streaming down my face every time Ellen or Allison Janney spoke.  Made the poor boy watch all their scenes over and over again.  Good Christ, who do these damn kid-movie-makers think they are coming up with this shit?  We need them for OUR movies!    So, I\'m home now.  In Pocatello.  Decided I didn\'t need to subject myself to all this stress and trauma any longer.  The basement resort in the mountain fortress looks just fine right about now and I don\'t regret my decision at all.  It came to me around 342 this morning.  Got off the phone with Levi last night and felt fine about the ways of the world and my place in it.  But soon enough, I realized I was doing the whole \"law school\" thing for the wrong reasons.  I thought I had to do it for other people.  So I\'d have enough recognition, enough worth.  It seemed a challenge befitting someone of my talents and abilities, but I was never really truly convinced.  And for a while, I may have enjoyed myself, been happy even.  But not anymore.  If I have to go through this every 3 to 4 months, I want no part of it.  There\'s nothing for me here worth all this.  Mom helped tipped the scales.  Called her around 6 to ask her opinion, as all I\'d done during the night was pace around the house, sweating and fighting off diarrhea.    \"I\'m coming home, Mom.  I\'m not going to go through with all of this.\"  \"Really?  Had enough?\"  \"I have.  I can live in Pocatello; it\'s not that bad.  I could be happy there.\"  \"I know, I think you\'re right.  I never really understood why you thought you had to go to law school, anyway.  Just to prove to yourself you could?\"  \"Yea, probably.  But now I realize I don\'t have to prove anything to anyone.  I can work at the Phone Center again until I find something better.  This just isn\'t worth all the stress and anxiety.  I never really liked it that much to begin with.\"  \"Sounds good honey, I\'ll get your room ready.\"  So yea, here I am and loving it.     (Author\'s Note: And by all of the writing above, I mean the exam went beautifully and I\'m doing fine...FINE!)   '),(23204,2314011,'07,December,2003','2003-12-07 00:00:00','    The drive and yoga helped.  Quite a bit, actually.  Getting ready to go to dinner with my brother, Joe.  He\'s good people.  Sounded a bit stressed himself when I called earlier to see how \"The Hulk\" went and whether it was worth my time.  He said no, so Harley won my attention instead.  Talked to Susan a while ago.  We agreed that our nervousness is a good sign.  We obviously care, are invested and want to do well.  Those without a care, thought, worry, etc. are probably the folks in trouble.  Of course, they could be totally together because of their confidence in their abilities, which is where I think I\'m getting to.  As the night progresses, bringing the exam closer, I\'m feeling more capable, confident and assured.  Hope it lasts through the night.   '),(23205,2314011,'06,December,2003','2003-12-06 00:00:00','    So, did a nice little practice exam answer this morning, and, per the Oracle\'s instructions, left it for an hour or so, came back to check my aptitude and did pretty well.  Yay.  Wanted to try another question from the same Web site, but alas, there were none on the other issues we covered in class.  Bummer.  Now just waiting for the laundry to finish so I can distract myself with that for a few minutes.  These kind of days are so odd.  The ones where you\'re just waiting for the next day to arrive.  Suppose I could go back and refine my original answer, just for the sake of typing it all out; hopefully my hands will remember the process even if my mind doesn\'t.  I\'m torn between studying for the subsequent exams and not over-extending myself.  Really, I\'m not sure how well reviewing other material would sink in at this point.  Wondering about my classmates.  Hoping they all slept well and aren\'t ripping their hair out today.  Think we all run a risk of over-reviewing, -studying, -practicing.  Doing it all to the point where we think ourselves out of knowing the material.    Think I\'ll take a drive in a bit.  Maybe do some yoga while the clothes dry.    I don\'t imagine alcohol has tasted or will taste better than it will Thursday the 18th at 4 pm.   '),(23206,2314011,'06,December,2003','2003-12-06 00:00:00','    Told my brother about an hour ago that I couldn\'t watch \"The Hulk\" with him because I\'d be going to bed soon.  Just thought I\'d check up on my email and favorite blogs, and suddenly, I don\'t feel like sleeping.  Think it had more to do with the movie we\'d just finished watching: \"Pirates of the Caribbean\".  It would have been sooo much better had it been an hour shorter.  No Disney movie should be over 2 hours long.  Especially not one about a damn ride.    Okay, this is going to be cut a bit short, as I\'m engaging in a conversation with my first fan.   '),(23207,2314011,'04,December,2003','2003-12-04 00:00:00','    Happy Saturday, bloggy!  Just about to get in the shower and head over to some gal\'s place to do some hard-core Civil Procedure studying.  Feeling good about it.  Think today, tomorrow and the exam will go well; it\'s just a matter of doing it.  In my mini-counseling session Thursday with Academic Support Oracle-lady, she said:  \"What I\'m sensing from talking with you is that you\'re very aware of what you need to be doing, that you\'ve been keeping up with reading, studying, reviewing all semester, you\'ve been doing everything you\'re supposed to and you know what you need to do to prepare.  Now you just have to go do it.\"  That\'s always the clincher, isn\'t it?    Watched The Fellowship of the Ring last night.  It\'s been a while, and it was a nice little refresher.  Seems like forever since I first saw it New Year\'s Day 2002.  The world was dark and scary, and I was the emotional equivalent of a sniveling-Gollum-like creature, continuously shifting between Smeagol\'s \"Come along, Master, we\'ll show you the way!\" and Gollum\'s \"Murderer!\" (I realize those are lines from The Two Towers, just go with it for now.)  Okay, not really, but the swings in my optimism and trust in the universe\'s plans were huge and frequent.  Frodo\'s scenes with Gandalf were the best and the worst.  Usually reduced me to tears and uncontrollable wails...right there in the movie theater.    \"Shut UP Gandalf!  You.Don\'t.Know!!\"  But knowing he really did.  So it was interesting to watch it again and look back, now that I\'m on the journey I was afraid to take then.  The thing that really struck me was how all the characters thought their jobs were done in the first hour and a half.  They all get to Rivendell and think that\'s it.  They have no idea what they\'re actually about to do.  Or maybe they do and are trying to talk themselves out of that knowing.    Then we get to The Towers and it\'s all downhill from there.  I really didn\'t intend this post to be a treatise in comparing the Lord of the Rings to my life, but what the hell.  The day\'s gray and wet, a good one for studying...   '),(23208,2314011,'04,December,2003','2003-12-04 00:00:00','    ...to let you know I made it through the day.  Got 4-4.5 hours of Civil Procedure done this afternoon.  Sigh.  Still have more to go, but it\'s a good start.  Also thought I\'d let you know this could very well become my sole mode of communication for a bit, unless you contact me.  I mean, I\'ll totally talk if you call me, and I\'ll read and enjoy any comments/emails sent in, but don\'t feel hurt if/when I don\'t respond until the week of Christmas.  Feel like I\'m entering (or have entered) a Zone...of what exactly, I\'m not sure, but one I don\'t plan on returning from for two weeks.  Coming of it out a bit for some Thursday night TV, tea and Grandma\'s home-made cinnamon rolls.   '),(23209,2314011,'03,December,2003','2003-12-03 00:00:00','    Not sure how I feel about new features on my Internet Explorer browser appearing today when they weren\'t there yesterday and I did no downloading, updating to make them be here.  Had a host of new \"favorites\" too.  Hmmm.    So my grip on reality\'s returning.  Almost time for Civ Pro review class, then heading home to get cozy with the subject\'s study guide for the rest of the afternoon.  Talked to the Academic Support girlie after class this morning.  Very much Oracle-like: she told me exactly what I needed to hear.  She also had some good tips for what to do the day before each final.  I was a bit worried that I\'d sit around and...well, worry.  I always do better if I\'m DOING something, so she gave me suggestions, and we love those.    Yoga will feel sooo good tonight, methinks.   '),(23210,2314011,'03,December,2003','2003-12-03 00:00:00','    Spent the night more awake than not.  Starting to feel the legendary anxiety in a big way, as if you couldn\'t tell.  But had dreams around 3-4ish about Property exam answers and felt slightly better, as I seemed to know what I was doing in the dream.  Think I was mostly upset at myself for not making the best use of my time, and last night was what finally tipped the scales.  On the one hand, I\'m still really glad I went to the panel, but on the other, I\'m realizing now, I should have stayed home and studied.  Because, really, there\'s no more time to waste.  First final\'s this Monday and I have yet to use the study guides to fill in the gaps in my understanding of Civil Procedure.    Saw the Return of the King trailer just as I was leaving the house and was again motivated.  All the shots of the characters about to meet the challenge they\'ve spent the last 2 movies preparing themselves for; finding out if, after all that, they\'ve really got it in them to do what they must.  \"All you have to decide is what to do with the time that\'s given to you.\"  And that Gandalf always knows just what to say!   '),(23211,2314011,'03,December,2003','2003-12-03 00:00:00','    There are people at school...right now...studying for next week\'s and the week after\'s finals.  Part of me wonders if they\'re incredibly unlucky, unprepared, unorganized, etc. or if they\'re going to be the ones cruising through the exams.  The girl I talked to said she was there until 11 last night.  That means she had 9 hours before Property class this morning.  I\'m remembering all the things the upperclass-persons told us the first few weeks of class about how each student will have his/her own study methods and not to judge your own by the person next to you.  And really, we\'ll only know for sure what works for us and what wastes time is next January after the grades have been posted and the exams are reviewed the first day of class.    A 2nd year told me today: \"No matter what happens this semester, you still have second semester to make up for it.\"  They don\'t actually kick you out until after your first year\'s completed.  Less than a C average will do it.  Doesn\'t  sound  hard.  My Contracts professor gave one (1) A last fall.  Torts professor, five (5) As.  And as dire as all this is beginning to sound, I\'m really not worried that I\'m going to bomb the whole thing completely; it\'s just the obligatory \"what ifs...\" starting to perk up.  So, have to go through the process of dealing with them and being comfortable with my level of preparation.  Or do more.    Tonight I cut the study time a bit short to go to a panel discussion on AIDS.  I know, World AIDS Day was Monday, so what the hell is left to talk about?  Don\'t really know why I felt compelled to go, but think it has something to do with the fact that I\'ve never attended anything like it in the past.  Ever.  It just seemed right somehow.  There was about 20-25 people there.  Don\'t know why I expected a crowded courtroom, but I did.  There are less people dying but more people getting infected.  My guess would be that because there\'s less death, less attention\'s being paid to the ever-increasing problem.  Without the action/excitement/drama of people dying, what is there to do/care about people just getting sick?  Suppose I have one more issue to listen for when the candidates make their pitches.    Speaking of which, I just have to say that Carol Moseley Braun, even though she doesn\'t stand a chance, is totally my nominee.  If, by some bad B-movie completely unforeseen and unwarranted twist of fate, she were to win the nomination, I think she may just have a shot at the office, if only on the sheer novelty of the idea.  Folks have been known to vote (and win) on similar grounds.  The thing I love about her most is the fact that she\'s the only candidate actually SAYING anything because she has nothing to lose.  Dean, Kerry and Gephardt all have to bide their time, talk in parables and lofty goals, waiting until the race actually starts before they actually start saying anything when they speak.  The remaining candidates are, honestly, entertainment, CSPAN fodder.  I\'m still waiting for Hillary to change her mind.  Or for the party to draft her into the nomination.  There\'s petitions, you know.   '),(23212,2314011,'02,December,2003','2003-12-02 00:00:00','    Evening blogger!  Just returned from the sexy yoga class.  Did a lot of leg stretches to get back into the routine after the week off.  Glad I had a mini-practice this morning...there was much groaning and sweating, and not entirely unpleasant at that!  Now just trying to relax for the night.  Got my mango-passion-fruit tea all hot and ready for consumption.  So tonight\'s class was a good time.  Instructor wore her tight, green tank-top; yes, a good night, indeed.  She\'s really a great teacher, I have to say.  Yes, it\'s true I have a fondness for teachers to begin with, but so what?  She\'s petite, wears black square-ish (bordering on geeky) glasses, has a little kid smile that dominates her face and Madonna-esque built arms.  She chants during the quiet, meditative \"corpse pose\" portion of the practice, which is such a nice touch.  Tank-tops are our friends.   Feeling better about everything in general today, which is nice.  My resolve\'s slipping back into place.  The other hormonal shoe\'s probably about to drop, though, so I\'m preparing myself for that.  Suppose that\'s something you all get to look forward to: the updates on Andrea\'s menstrual cycle.  Yay!  It informs so much of my life at certain times that anyone close to me has the thing down almost as well as I do.  That\'s what we\'re all about here, after all.  Studied Torts all afternoon; feeling a bit more on top of that subject, at least.  It\'s actually one of the subjects that\'s come the most easily to me.  Probably because it deals a lot with finding a balance between various circumstances, principles, etc.  Balance Rocks!  Borrowed some study guides from Susan; she\'s so on top of that shit.    Called her on my way over to get them, and pulled into her driveway as the second ring ended:    \"Answer, damnit!\"  Pause and a bit of a laugh.  \"Is that any way to talk to me?\"  Yea, totally busted.  But it was a nice little classic moment.  Perhaps not as classic as others, but not bad.  She still let me in her house and borrow study materials, and I\'m grateful.   '),(23213,2314011,'01,December,2003','2003-12-01 00:00:00','    So, I added a comments feature.  Knock yourselves out.   '),(23214,2314011,'01,December,2003','2003-12-01 00:00:00','    After no consideration whatsoever, I decided not to start typing these posts in a saveable forum (i.e. Word Perfect) and continue typing them at the Blogger site, leaving myself susceptible to incidents similar to the one last night.    So my brand new DVD player is acting funny.  I\'m not freaking out, although I have considered dropping it out the window to see if that improves its attitude any.  Think I have the problem figured out, though.  It makes a weird noise sometimes while playing certain songs.  It\'s this sound that precipitated the stopping of playing and grinding sound of its predecessor.  It only happens on songs copied from MP3 files.  The other copied songs play just dandy.  Think I can handle that.  It plays movies beautifully, so at least there\'s that.  Had an okay day.  Wish I could either get to the point where I feel comfortable with my preparation-study process for finals or just get the damn things over with, because I\'m finding myself stuck somewhere in between.  Feeling better tonight after getting back into the school routine and talking to the boy.  Went to the opening reception of an intriguing art exhibit for World AIDS Day.  Chatted with the interim director of the Women\'s Center.  No sparks to speak of, but a few nice chuckles.  Was nice to get my mind off the biggest thoughts of the day, for a bit at least.     Think I might actually get around to sinking my teeth into the new book: Cold Mountain.  I know, I\'m a few years late, but I\'d like to get through it before the movie\'s released.  Actually bought it in Seattle last December; it\'s only fitting I get around to reading it this year.  It was my first purchase in the store I affectionately call The Pussy Book Store; it\'s just off Broadway, the heart of Seattle\'s big gay district.  The first time I walked through its door, in the summer of 2001, I was immediately struck by the pet-store smell.  The wandering felines and hidden-in-the-sea-of-similar-flyers advertisement for cat-sitting eventually alerted me to the used bookstore\'s alternate identity.  And a trip to the Emerald City\'s never been complete without a visit since.   '),(23215,2314011,'31,January,2004','2004-01-31 00:00:00','    Two big Friday nights in a row.  Who would have thought?  Well, I did, but I suppose that doesn\'t count.  And really, they weren\'t really \"big\" in the ordinary sense.  Pretty damn normal, actually.  But they weren\'t spent in front of the TV for once (then twice!) and that\'s big.    It hit me a couple of weeks into January that I had yet to make a New Year\'s resolution.  Had such a great one last January, what with all the excitement and joy of exploring and adventuring, that this year\'s was forgotten in the nostalgia.  So, it came to me one morning at the kitchen table with morning pages: I decided to be more proactive.  Learned last year to pay attention, listen and breathe; this year it\'s time to start using those lessons.  Don\'t yet know how precisely, but I\'m sure it\'ll come to me.    So yea, had a good night out.  This week kicked my ass, so a big damn bunch of queers dancing to Top 40 remixes sounded perfect.  The homework will be light, too, and we have to take advantage of that.  The show was good; the first one after a dry spell always feels great.  It was Kathy\'s birthday celebration, so every song performed was from the 80s or before.  So great.  Peggy Lee.  Dusty Springfield.  The new remix of the Laverne & Shirley theme-song.    Was on my own for the second show, and was buzzed just enough to continue dancing by myself in the sea of young things.  Yay.  Tried to track down the birthday girl to thank her properly for the free entrance fee, and ended up people-watching with the Voice of God.  She\'s a barrel of laughs, wields a mean screwdriver and makes for a great DJ.  Pink is right!  Eventually, Kathy took me into her big office and we enjoyed a sensitive chat on her couch about toilet gaskets and the asshole-tax for Frat boys.    Today\'s been spent pretty much entirely on the futon, the day\'s single outing being a journey across the state line for The Butterfly Effect.  I really enjoy the preview.  Wish they\'d just stuck with that.  It\'s one of those rare movies that\'s bad in the first 10 minutes and never improves.  People actually screamed a few times, and I laughed out loud a few more.  Demi was behind the whole thing.  It reeks of her.  So glad I get to see Lost in Translation tomorrow.    But for tonight, I think I\'ll get cozy with a new comic and a new novel.   '),(23216,2314011,'30,January,2004','2004-01-30 00:00:00','    ...you want to be sure you know where you\'re sending it.  I love Torts.  I don\'t love the fact that we didn\'t get to the material I spent much time reading last night, but quotes like the above make me forget my bitterness.   '),(23217,2314011,'28,January,2004','2004-01-28 00:00:00','    Goddamn Boy.  Messes everything up.  I call him to ask one simple question that would normally take us 45 minutes to answer, but no, he has to talk for an hour and a goddamned half.  As if I don\'t have anything better to do.  Told him I had this blog entry cooking up all day, but he didn\'t even take a breath as he spoke right over the top of me.  Good thing he fucks like a banshee, or I\'d never forgive him.  So I completely underestimated the outline for the Motion to Suppress Evidence assignment.  Left school around 3 with the best of intentions to work diligently as soon I returned home until it was completed.  Then I would have all evening to write a long, lazy, juicy blog and maybe even start a new book.  Instead, my futon sucked me into its void and there I remained until 430.  Stupid Dr. Phil.  Arranging the work space, surrounding my desk with proper authority, choosing the perfect background music, etc. took another 15 minutes.  Here\'s where the underestimation comes in:  I let myself watch TV for so long thinking the assignment wouldn\'t take very long; once I got started on it.  Knew precisely what I was going to say.  The damn thing was practically written in the cases waiting in my backpack.  All I had to do was patch them together in outline form.  Shouldn\'t take more than 2 hours.    There\'s a moment at the beginning of each and every writing enterprise that bears a striking resemblance to the end of each and every research session.  The moment where everything in front of you is nothing you can use.  All of it made sense half an hour ago.  But now that it\'s time to do something, it may as well be written in Greek.  So the blank page stared at me for more than a few minutes, taunting.    \"Uh huh, that\'s right!\"  It said, hand on hip.  \"Glad you watched that last case on Judge Judy NOW, aren\'t you?\"  But those moments are always defeated.  The outline\'s a glorious 4 pages and it only took an hour more than my estimation.  Which leaves me a bit less time than I hoped to devote to my blog, but we\'ll make do.  My friends\' reactions to the blog (and the content included therein) have been interesting.  The new ones probably just take it all in stride, as part of who I am.  The older ones, though, have been more than a bit surprised.  How could someone who kept much of herself private, hidden from casual conversation, suddenly be this revealing?  Their surprise is a good thing, I think (hope); but it raises a valid question.  And I\'m not sure I have an adequate answer, other than to say I was ready.    I\'m reminded of a conversation with my boss at my high school job.    \"I can read people very easily.  It\'s fun to freak them out with my observations; some of them are like books.\"  \"What do you read from me?\"  I asked, trying not to sound as curious as I was.  \"I can\'t.\"  \"What do you mean?\"  He was one of the few people I spent most of my time with at work, after all; he HAD to have some kind of read on me!  \"You\'re impossible to read.  Very protective of yourself, don\'t let much emotion through.\"  \"Really?\" I replied, trying to figure out how I could have been so obvious.  \"But I think I have it narrowed down to a couple of explanations why.  You\'re either hiding something, or you\'ve been hurt in the past.  And you don\'t seem to be the type who would have something to hide.\"  As it turns out, he was right on both counts.    A big part of what I was hiding has been recounted here.  Mostly because I was ready to tell it.  The last year has been a study in being open and receiving, letting go of control and needing to know, beginning to see the person I\'ve always wanted to be.  And this has been an excellent tool in that undertaking.    Remember the day I was sure the letter in the mailbox was the rejection from the law school because it was too small to be acceptance materials.  I looked, saw, closed the mailbox without taking it out and went to class.  The walk was pretty transcendental, I have to say.  The significance of the fact that it was the way to my Creative Nonfiction writing class wasn\'t lost on me.  Walking back, I promised myself I would stop resisting the part of me who is a writer.  Up until then, I\'d put the choice out to the universe: law or writing.  \"You tell me what I should do, cosmos, the decision has tortured me long enough!\"  That day it seemed I had my answer.  Soon realized the letter was the wait-list notification.  Three months, a round of personal growth and daily appointments with a notebook and pen later, I had both and was on my way to adventures undreamed.    And I\'m still loving the ride.  So my little bloggy was really just another step in that process of exploring, growing, learning.  Don\'t think I kept this stuff from you, friends, and, in turn, had no problem writing about it for all to read.  I finally am ready to share more of myself with you and am able to here, in my favorite passion.   '),(23218,2314011,'27,January,2004','2004-01-27 00:00:00','    Spend more time at the Women\'s Center.  They\'re good people over there.  It\'s quiet, welcoming, comfortable and a perfect place to do reading for class when the basement gets to be too much.  Plus, hippie Yoga types hang out there.  And the athletes running around on the floor above the Center reminds me of home.   '),(23219,2314011,'27,January,2004','2004-01-27 00:00:00','    So you better watch out for that.  The law school\'s eerily quiet at 6 in the evening.  It\'s pretty nice, actually; if one can ride out the waves of students fresh from class until they all leave for the day, one can actually get a thing or two done!  The thing or two done tonight included reading for tomorrow\'s Civil Procedure class (the horror!) and Thursday\'s Criminal Law class (because that day\'s Contracts reading is already done--Yay.) so as to make room for time to complete the writing assignment\'s outline due Friday, even though she\'ll give us until Monday.  Have yet to read the final case (that is sure to change the very nature of said outline) needed for said assignment, so I need all the time I can give myself.  Now I have exactly one hour to get home, change, find my bearings, and get to Yoga.    Hope this morning\'s black ice hasn\'t returned now that the sun\'s finished for the day.  The walk into school earlier was a bit treacherous, as treks across parking lots go.  Black ice on asphalt does not give.  Not even a bit.  And I\'m the biggest pussy when it comes to walking on ice and/or snow.  This fact is exacerbated by fellow students cursing along, passing me without a second glance or thought about navigating the new terrain.  They must have no fear of death.  Like Aragorn.    So class participation in Property is our friend.  Had a sneaking suspicion that those points helped my exam score up over the C/B hump, and my professor confirmed said suspicion in our meeting yesterday.  Strictly point-wise, I should have had a C+ on the exam, but my probing, insightful comments in class gave me that extra edge we so love in times of need.  Yay.    Suppose that gives me more leverage in executing Susan\'s Operation Liberal Countermeasures in every class.  (Aside: wonder if writing case briefs for 2 hours before posting changes the tone/style of blog postings.)   '),(23220,2314011,'26,January,2004','2004-01-26 00:00:00','    Feels like a big blog day coming on, and there will probably be a long, juicy entry later in the day, but I just have to say something now.  Johnny fucking Depp as Best Actor in Pirates of the Caribbean.  Okay.    I can forgive the Hollywood Foreign Press for nominating him; the members are eccentric; they\'re like that; they do that type of thing.  And the American Academy might have gotten away with it, if only for the laugh it produces.  But no laugh can ever purge the taint of the travesty of the Academy\'s omission.  No nomination for Viggo Mortenson.  No nomination for any of the LOTR cast members.  But Johnny \"I look like a reporter on the take from the roaring \'20s\" Depp gets a nod for his drag queen performance in a movie based on a ride at Disneyland.    Aragorn carried that damn movie!!  He IS THE FUCKING  KING !!  AS IN \"RETURN OF THE  KING !!!  I dunno.    I just don\'t know.   '),(23221,2314011,'26,January,2004','2004-01-26 00:00:00','    A snipet from last night:    (Gasp!) \"What are they saying? Who\'s on stage?\"  \"Bill Murray just won.\"  \"Hurray! We love you, Bill Murray!\"  \"Hey, there\'s Cate! Hi Cate!\"  \"CATE! Is she gorgeous?\"  \"She\'s gorgeous and Pregnant!\"  \"(GASP!) NO! What\'s she wearing?\"  And this kind of description went on for hours. It was like being blind and crippled in a dark cave, with only an echoing voice to illuminate the world outside. It was great!   I love being an echoing voice!   '),(23222,2314011,'25,January,2004','2004-01-25 00:00:00','    Been working diligently in the basement for almost 5 hours.  Feels good.  I decided, pretty much subconsciously, that I would do more work when reading cases.  I usually brief (summarize) the assigned cases, read and highlight the notes following them and move on.  Today found me taking notes on the notes AND introductory material in front of cases.  My god!  In some instances those notes exceeded the initial brief.  Overall, I think the change is a good one; couldn\'t hurt at the very least.  Not sure how the 2 subjects (4 cases) in 5 hours will work during a usual week, but feel pretty confident about the new arrangement going in.    Soon I\'ll be headed home for some light cleaning (the floor needs a good vacuuming at the very least), and maybe some not-so-light Yoga before settling in with the Globies.  Not sure when the next book, Ghostwritten, will get its start, but the wait\'s worthwhile.  Trying to remedy my habit of beginning (and continuing) books in the moments before bedtime in favor of reserving nice, big blocks of quality time for them.  Kinda like one would if one had a lover.    In other important-in-a-completely-insignificant-way news, my feet seem to have adapted to my new shoes to the point where I can at last wear them without fear of wearing the skin away from the back of my foot, just above the heel.  YAY!   '),(23223,2314011,'24,January,2004','2004-01-24 00:00:00','    Morning bloggy.  How ya doin\'?   '),(23224,2314011,'24,January,2004','2004-01-24 00:00:00','    \"I\'m 64% sure this woman sitting over here is a lesbian.\"  Dan said with his trouble-maker look.  \"Okay, show me.\"  I shot Dan\'s friend Mike a look and said, \"Bet I\'ll know the person.\"  We crept to the non-smoking side of The Garden, assumed very conspicuous bar-stools overlooking the lower level.  He pointed and said, \"Those two over by the window.\"  But I was already descending the stairs on my way to greet my friends Emily and Lori.  They work at the Women\'s Center.  They\'re feminists.  And straight.  Emily reminds me of a blonde Alanis Morissette.  Lori was in my Yoga class; she\'s in Level 2 now and is training to become an instructor herself.  I met them during the Year of Darkness, Winter/Spring 2002.  Emily was the co-founder of UI\'s first and only student feminist group.  Interviewed her for the Argonaut when the group put on the Vagina Monologues that February.  It was a hit and so were they.  She was the first to read my famous vagina column; she approved, but my editor made me tone it down.  My sociology professor, Debbie, introduced me to Lori during one of our Directed Study \"meetings\" at the Prospector aka Alehouse.  That was the day Debbie said we\'d be back in this bar after my first semester was over, figuring out what I was  really   going to do with my life because I\'d hate law school.       The boys and I found ourselves discussing the waitresses with tight skirts and the social statement that seemed to make.  Dan\'s a debater.  And he\'s exhausting.  I pawned him off on the gals and watched the sparks fly.  Lori and Dan had a class together years ago, so she knew how to handle him.  She made great points about power differentials, societal expectations, etc. while he countered with definition-differences and the pressure on men to be tall.    \"I was trying to ditch them so we could actually have a nice chat.\" I whispered to Lori as we hugged when I left.    \"No worries,\" she replied; she\'s so sweet that way.  \"Tell Debbie \'Hi\', that I\'m thinking of her and she owes me a beer.\"  Because I\'m sweet that way.  So Big Fish was good.  I\'ll give it a B.  Ewen was his adorable, endearing self.  Bit sugar-coated for Tim Burton, but he pulled it off.  Think I\'ll have to see it again when I\'m not half-asleep.  Seemed a bit long, but I suppose that could have been a stylistic technique.  Saw Pieces of April today.  It gets a big fat A.  We love independent movies shot with handhelds.  It\'s a very simple story with a very simple ending that made me smile big.  And Katie\'s actually pretty damn cute in it.  Even if the story had completely sucked, seeing Sean Hayes\'s (Just Jack) performance with Burnadette made the outing worth it.    Back now in my little den, avoiding the Christmas break-esque weather.  Have a terrible craving for hot buttered rum and a satellite dish.  Just finishing some laundry, getting ready to have a Queer as Folk refresher.  Season Three\'s due out in a month, after all.    Always feel better/lighter after posting.  The weekends are a bit conducive to creative blockage and these are a great foil.   '),(23225,2314011,'23,January,2004','2004-01-23 00:00:00','    That\'s exactly what I mean!  Goin\' out for beers with cute Dan and cuter Sadie, to be followed by Big Fish with desk-mate Paul.  He\'s cute, too.  That\'s more action than I\'ve had since getting freaky with Natalie at Charley\'s in Pocatello!    Skipped morning pages this morning and it\'s thrown me off all day.  Just don\'t feel like doing anything.  Actually, that\'s not true.  I feel like doing things, like taking notes, studying and typing them, etc., only half-heartedly.  Even now, I\'m finished with everything, but I don\'t really want to leave because I\'ll just have to go down to the SUB, pick up my W-2, then go to the mall to secure the Parents\' Happy Anniversary card.  It\'s just too much!  Plus, January\'s decided it\'s not giving up on Winter just yet, so we\'ve had a steady stream of snowfall all day.  I could get back to Levi\'s sister about an essay she requested my insights and opinions for, but even that would be half-assed at this point.  And we hate half asses.    But the weekend seems to be shaping up.  Will have enough homework to save me from boredom, and am thinking I\'ll check out Pieces of April at the old, hippie, liberal-run theater downtown.  And the Golden Globes are on Sunday night, so I\'ll have to spend a good half day getting ready for that.  Wow!  Suddenly the world seems like such a perfect place!   '),(23226,2314011,'22,January,2004','2004-01-22 00:00:00','    Yes, I am; and something tells me you are too.  Think we will begin to glimpse your mind and blind artistry before too long.    It\'s been a good day.    Worked hard at school, both in classes and out of them.  Came home to a message from Carrie, whom I\'ve missed over the past few weeks while her phone sage raged across the country.  Found out Erika\'s (the instructor) chants/songs haven\'t gone anywhere; they\'re just waiting for her voice to rest.  And am now settling into the night with some hot soup in preparation for the Jack & Karen Hour.  Yes, it\'s been a good day.   '),(23227,2314011,'21,January,2004','2004-01-21 00:00:00','    Kudos to my Good-Samaritan-hero-of-the-week (she knows who she is).  Not only did she call to admit an error in understanding an assignment, which takes more courage than any of us are willing to admit, I mean really, but she also stopped to help victims in a collision.  She made fire from grass and toilet paper, boiled water and made them all herbal tea to heal their wounds.  And it worked!  Plus, she threw down on a wimpy on-looker, who probably deserved a kick in the balls to start with, but, in her infinite wisdom and patience (that\'s right!  Patience!), our heroine knew the man\'s humiliation in the face of her bravery, poise and unwavering commitment to preserving all that is good and pure in humanity was punishment enough.  Let\'s help her think up a name for her embryonic blog, shall we?   '),(23228,2314011,'20,January,2004','2004-01-20 00:00:00','    I\'m tired and I don\'t know what day it is.    Not much more to say than that, but I love using the university\'s wireless Internet, especially to post, so you know.    Just leaving for the day, but there\'ll be plenty of reading to be done this afternoon/evening.  It just seems wrong to be following up White Teeth (the book I finished the other night) with Contracts and Criminal Law.   '),(23229,2314011,'20,January,2004','2004-01-20 00:00:00','    Since she made fun of my blog, insinuating that I include all the incidental, pointless, random occurrences during a day\'s passing, and, by extension, that such occurrences are of no interest to my readers, I\'m including the story that gave rise to the discussion in the first place, in condensed form.  Susan received a free sandwich at Arby\'s this afternoon, which brought up associations I have with said establishment\'s roast beef.  The associations stem from a drive through Bozeman a couple years ago with Natalie and the Boy, during which Natalie compared Arby\'s roast beef with a particular area of the female anatomy.  This whole discussion would work better if I remembered the actual line Nat delivered that fateful day, but that\'s beside the point.  The point is: the random, pointless stories I tell in this blog are much better when I take myself seriously in telling them, with a few flourishes of sarcasm and irreverence.  Recounting such stories only if they\'re important, relevant, etc. in the grand scheme of humanity wouldn\'t be very fun, as the above illustrates.    To drive the point home, let me tell you another story, plucked from the day\'s activities.  Stopped by Oracle-lady\'s office on the way to Civil Procedure.  She\'s good fun and is loose with her candy supply.  \"Well, I think you\'re entitled to some Smarties,\" she said after I told her I was happy with how my first semester went.  \"Or some Dumb-Dumbs,\" she laughed, referring to those little suckers the banks across the land have become famous for distributing, \"depending on how you\'re feeling today.\"  I spied a third option (one might even say \'alternative\'..gasp!), instead.  \"Actually, I think I\'ll take a Nip.\"  I said, knowing the significance of such a choice was lost on her, but enjoying the joke thoroughly nonetheless.   '),(23230,2314011,'19,January,2004','2004-01-19 00:00:00','    30 pages to go and I\'m wishing I was 30 pages in.    Hit a mental roadblock about 45 minutes ago and couldn\'t go on.  It disguised itself as fatigue.  That Sunday night (for our present purposes, the role of \"Monday\" will be played by \"Sunday\") reminder that tomorrow brings with it not only another day, but another week.  Realize now it was really my subconscious, fighting like hell against the end.  Knowing every page turned is bringing me closer to the last place I want to be.  Finishing a favorite book is always hard, but this...I started missing this book on page 80.  Now I\'m on 510; can\'t move forward but can\'t let it go, either.  Thank god Susan called.  Otherwise I\'d be done by now.  But no; I have 30 pages yet!  Yay!   '),(23231,2314011,'18,January,2004','2004-01-18 00:00:00','    Have an entire day ahead of me with no schedule of events.  Of course, I could read cases, but I\'m not sure I\'ll get to those because I\'ll be too busy to actually do anything too school-related.  Could go for a drivey, maybe through the Lentil capital of the world, hmmmm???  Do have a dirty Mexican porn to watch.  And a book to finish.  And Sean Astin\'s on Ellen\'s show right now, so what more do ya need?  Sam and Dory together!    \"Don\'t go where I can\'t follow!\"  \"Ooh! Ooh!  Pick me!  Pick me!\"  \"I can\'t carry it for you...\"  \"A clam!\"  \"...But I can carry you!\"  \"Are you my conscience?\"  Oh, and props to Martin Luther King, Jr.  Thanks for the day off, man, way to go.   '),(23232,2314011,'18,January,2004','2004-01-18 00:00:00','    With all this talk about Iowa and its damn \"caucuses\", I\'ve found myself at a loss for knowing just what it is.  Suppose we\'ll all know tomorrow.  And no, I can\'t just research it on my own, so don\'t ask.    So I\'m about to settle in with my nightly jug of tea after my nightly (you guessed it) yoga practice.  I\'m addicted, what can I say?    \"I have hot calves!\" is what I can say!  They actually have muscle-mass.  For the first time since I was 9!  I have calf muscles!!  My upper-arms are even joining in!  I realize this is probably less than invigorating news for the lot of you, but, for me, this is really big shit.  I\'ve always wanted to have that sculpted, toned look that goes oh-so-well with tank-tops, but without having to actually go to the gym.  And now I\'m on my way to exactly that.  Yay.  Look out June!    Picked up a dirty Mexican movie along with some hot lesbain stand-up comedic action on the side, so the night should be an exciting one.  Don\'t really want to ruin the great run the player\'s on with the Winter Mix.  Discovered something new about my temperamental player:  if I just let it run on one disc continuously, all the way through, even a burned disc, it doesn\'t throw a fit.  I may even be able to put a single disc on \"Random.\"  But as soon as I start skipping songs or throwing the \"All Disc\" format together with \"Random\", it starts to whine.  I\'ll figure this damn thing out yet.  And keep you all posted, too.  Speaking of which, can I just say (in all seriousness) how much I adore my latest regular reader?  Because I do; even more than she loves to hate writing assignments.   '),(23233,2314011,'16,January,2004','2004-01-16 00:00:00','    That\'s Lorraine\'s attempt at spelling dilemma.  It was too good not to post, and even better to put in the title.    So today\'s been nice and productive; well, more nice  because  it\'s been productive, you know.  Got more than a few cases to review for the first writing assignment of the semester.  Research always sounds fun/easy/relaxed, etc. before you actually sit down to do it.  That\'s when the mind goes completely blank and any specific, insightful, cut-through-all-the-irrelevant-issues-key words floating around in your head earlier are completely gone and the only thing you can come up with is: US Constitution.  You may as well start a search for the blue prints of NASA\'s Rover with \"telescope.\"  Then, when you finally get on a roll, nothing seems impossible.  You could write the whole assignment in 12 minutes!  No one\'s going to find these perfect, spot-on cases!  Certainly not in the time you\'ve found them!    Ultimately, you have to sort through the mountains of paper you\'ve amassed because you\'ve printed off everything that could, in any way, possibly be interpreted to bear a vague resemblance to what the initial subject was.  And, inevitably, as the sorting, shifting, organizing begins, you realize that none, or very little, of it relates to the manifesto growing at rapid pace in your mind.  Then it hits you:  you\'re going to have to read all these cases to determine which ones actually apply to the assignment, regardless of what thoughts passed through your mind at first impression.    \"Where\'s that ONE I KNOW applies?\"  You ask yourself, frantically throwing unstapled papers around your desk.  \"Just ONE!!  That\'s all I need!\"  But the ONE is lost forever in the pile of other ONEs; they\'ve lost all relevance and you have no choice but to take them all home, examine each in kind to determine just how much more work you have now compared to how much you had when you walked in 7 hours ago.    In the meantime, however, I was able to forget Friday night\'s disappointment, if only for a few fleeting moments.  Tuesday will be the true test, though.  Only then will I know for sure where I stand.  Just wish I knew for sure where I  want  to stand!   '),(23234,2314011,'16,January,2004','2004-01-16 00:00:00','    I almost had a date tonight.  Actually, fuck \"almost\"; I  had  a date, for about half an hour.  It was going to be perfect.  Here I\'d spent all last semester waiting for the right time, perfect place, best opening line, trying to calm my fears of rejection and step out of my comfort zone for once...Once!!  And today, I finally took that chance.  There have been signs all along.  She tried to hide her affection beneath a surly, tough exterior, but I saw beyond that.  We joked about the old hometown, laughed at how much we secretly wanted each other.  But underneath, I knew we were laughing about something real.    We planned to meet for beers, and those only lead to bad places; I was counting on it.  I figured such an invitation would be obvious in its entreaty.  After all, we \"joked\" that our movie and dinner the day after our last final was a date.  Clearly she was sending me signals that she was ready and waiting for me to initiate the next step.  Any remaining shards of doubt vanished when we had our lunch together this afternoon.  It was glorious.  We bonded; we shared things few others knew.  I looked at her; she looked at me.  She finished my olives, for god\'s sake!  If that\'s not a come-on, I don\'t know what is!  So when she called half an hour after our arrangements were made, I was excited.  I thought perhaps she couldn\'t wait to see me and wanted to move our plans up a bit; or maybe she had to tell me something too difficult to discuss in person.  I tried to sound nonchalant, telling her I was just \"reading the newspaper\" when I was actually admiring her picture in the law school\'s directory, as I have a tendency to do quite often, when all my life seems to shrink and the darkness curls up around me.    \"Listen, uhh..\" she said.  It sounded serious.  My heart was pounding; this was it!!  \"...I have bow out of tonight.\"  Crushing.  Tumbling down a rocky mountain-side onto a torn, jagged terrain not entirely unlike the broken floor of Mordor.    \"Oh?\" I recovered quickly.    \"Yea, I forgot about this thing I\'m doing with Mark tonight.\"  Oh yes, this \"Mark\" person.  She refers to him on occasion, calling him her \"live-in.\"  It\'s not entirely certain he even exists.  But I remained calm.    \"Okay.\"  Did that sound too pained?  \"We\'ll definitely do it another time, though,\" she offered.  Was that hope I heard in her voice?    \"Alrighty.\"  Like when you finally find the courage to admit the truth?!  Such was my date that wasn\'t.  I\'ll go on, somehow.  Probably get some coffee next door to the bar that was to be the site of our meeting.  But I won\'t go in, oh no.  Maybe come home and console myself with Queer As Folk\'s Season Two on DVD.    But I don\'t plan on letting myself get this hurt again, not for a very long time.   '),(23235,2314011,'15,January,2004','2004-01-15 00:00:00','    The reading assignment for Contracts was the one-year provision and suretyship provision of the Statue of Frauds.  What did I read?  The one-year and suretyship provisions cases.  What did we cover in class this morning?  The Memorandum requirement and reliance exception cases.  It was a painful 50 minutes.  Hopefully my notes will make sense after I\'ve read those two cases.  If not, my life is over.  Ruined.    Doesn\'t feel like Friday.  Probably because there\'s so much to do.  And no, I can\'t just leave those cases to be read over the weekend.  They have to be read (and understood) today!!  No falling behind here, that\'s for damn sure.  Besides, there\'s already plenty to keep me busy over the three-day weekend.  Reading, briefing and researching the new paper assignment.  Yay.    But enough bitching.    Your favorite moments were lovely, commentors.  I may have to do my own list of specific moments that will live forever, but, alas, not this day.  I have something of greater importance to discuss this day.  I\'m working on the track list for a new mix tentatively titled the Spring (April/May) Mix.  The list\'s coming along nicely, adding a few songs here and there each day.  Noticed, though, that the artists are primarily women.  Shocker.  Such is usually the problem I encounter when constructing a new CD.  So I had this idea:  maybe I could get someone to help put this thing together.  Turn it into a joint-venture type of project.  There\'s no mystery about who I\'m thinking should join me in this adventure.  He\'s good at male mixes, I\'m good at female mixes, we create great projects together, so why not expand our talents?    Whatdya think, honey?  Wanna make a mix together?  It could be risky, and our friendship may suffer great trauma as a result, but isn\'t it worth a chance?  A chance at greatness?  A chance of creating the ultimate, best Spring Mix the world has ever seen???   '),(23236,2314011,'15,January,2004','2004-01-15 00:00:00','    ...in no particular order...  Shifting into fourth gear on an open backroad.  The sound of my yoga instructor\'s voice.  Meeting my boyfriend\'s eyes, smiling the smile we unknowingly reserve for each other, and returning to our own activities without saying a word.  The sound of Lorraine, Levi, Carrie, Susan, Mom, Dave, Jen laughing after they\'ve surprised themselves with their own humor.  Watching the sun rise while writing morning pages.  Passion-induced cat naps.  Balancing the book and cup of hot tea while sitting comfortably in the bowl chair with my feet on the futon.  Finding a favorite place for the first time.  Walking into and out of the law school building.  Snuggling into my old bed in the basement.  Waking up in my bed next to the highway.   '),(23237,2314011,'13,January,2004','2004-01-13 00:00:00','    The good mood seems to be holding as we round the day\'s mid-point.  The four-hour power session went smoothly, even though it was a bit rough getting back into the rhythm of taking notes for Contracts.  Yikes!  Statute of Frauds, indeed.  One day, many removed from this one, I may explain the rule to you.  The Criminal Law class, however, looks promising indeed.  We love criminals!  But only the ones we read about in textbooks.  Actual ones aren\'t our friends.  Unless they are.  How\'s that for legal maneuvering for you?  I have a new nickname:  Nothing But Trouble.  I don\'t understand why I have this new designation, but I\'ll take it, own it and be proud of it just the same.  Don\'t even try to stop me.  Suppose it\'s a good thing I\'m already a fan of the Pink song.  I\'m harassing, too, but we won\'t get into that; trying to change that perception of myself, and all.    So now the day\'s down to reading for tomorrow, which is just more Criminal Law stuff, none of which I have to brief...Yah!  And that\'s always a good place to be; especially before 3 or 4 or 10!   '),(23238,2314011,'13,January,2004','2004-01-13 00:00:00','    Morning, blogger!  Have 20 minutes before my first class, but just felt like posting something before the day begins.  Who knows what kind of mood I\'ll be in after four (4) hours of class/Internet database research training/Professional Conduct lecture by the Dean/another class, so I better post something to commemorate my good mood.    As you can probably tell, they didn\'t really miss a beat with the return to school.  And while that fact had me questioning my existence last night, today I feel ready to tackle whatever they may (and will) throw at us, as long as I can have my Grape Nuts and coffee in the morning, and Yoga and tea at night.  God, I DO sound earthy!  I\'ll have to start drinking more beer, I guess.    Final word on the GPA: 2.83.    My friend Dan\'s over here chatting with me.  Say Hi Dan:  hi i am writing for andrea she is wicked smart.  Says he who has a GPA in the 3-point range.    \"3.42, actually,\" he corrected me.  \"He\'s such an arrogant little prick.\"  \"No, Dan, I wouldn\'t say that.  But I\'ll type it.\"  Alrighty...off to class.  Wish us luck.   '),(23239,2314011,'13,January,2004','2004-01-13 00:00:00','    Feeling like posting again, but don\'t really have anything majorly coherent to say, so I\'ll just ramble, m\'kay?    Yoga is our friend.  Just finished a 45-minute practice with the Double Diamond mix going throughout...so great.    The DVD player\'s clickiness stops when I let a CD play all the way through--even the burned CDs.  Interesting.  Squishy doesn\'t scream when she\'s ready to be emptied, but the steady stream of steam lets me know just as effectively.  The candle Mike gave me for Christmas last year burned its last flame tonight during said Yoga practice.  Thanks, little guy; you smelled great.  Another has taken its place on the top of the bookshelf.  It\'s introducing itself to Trudi right now.  Read the first 50 pages of A Civil Action earlier.  It felt cheap and dirty because I wasn\'t reading the book I\'ve been craving ever since I began reading it, but it\'s time will come, if I ever get off this damn computer.  They went pretty fast, those pages.  Vaguely remember the movie, and can already say with absolute certainty that the book is better.  It\'s required reading for Civil Procedure, believe it or not.  Think I\'ll go add some sugar and milk to my Chai tea; it\'s one of my newest faves.  Tea!  Who knew?!   '),(23240,2314011,'12,January,2004','2004-01-12 00:00:00','    The last of the grades came through today:  Civil Procedure: A-  (YIIIPPPPIIIIEEEE!!!!!!)  Contracts: C (Which is quite good, really.  Exactly what I anticipated and not as low as I feared.)  Research and Writing Paper a.k.a the only graded assignment for the semester: A- (see the Civ Pro comment)  KICK ASS!!!!  Nice little finish to the day, semester, year.  The class standings will be out in February, but I\'m not too worried about them.  My GPA will be higher than 2.0 and that\'s worth getting drunk over.  Not that I\'m going to, that is.  Have Property first thing tomorrow morning, after all.    Had a really good day even before checking out the grades.  Spent the day running errands and getting started on the reading for this week.  Three hours of Criminal Law reading at the Daily Grind always puts a gal in a good mood.  Especially when the spectrum of the 36 pages spanned from loud yelling at a parade to the constitutionality of same-sex sodomy; Yay!  And can I just say: one of the reasons I love this town so is the fact that it feels like March in January.  March will no doubt feel like December, but for now, it\'s a nice switch.   '),(23241,2314011,'12,January,2004','2004-01-12 00:00:00','    So the day started kinda poopy.  Couldn\'t shake the overall gloom and pouting-to-China attitude I adopted last night.  Usually, I really enjoy being back in my own space, getting back to its natural sparkle and charm.  But this time, I just missed the break and the folks I spent it with.  Thank god for morning pages, I tell you what!  Had too damn much to get done today to be doing it in a mood!  So yea, as each line of the three pages filled, the little black cloud above my angelic head dissipated, leaving nothing but a bright, sunny ball of sunshine in its wake.    And I\'ve been really productive to boot.  Completed two loads of laundry, cleaned my apartment, hung the new wall decorations (which kick serious ass, thanks to that fabulous man of mine), had my chipped windshield (that wasn\'t really chipped, had a bug baked on...oops!) to the windshield doctor and chatted with a favorite journalism professor.  Yay.  And it\'s only five!  Think I\'ll put the trip to campus off until tomorrow.  Get to buy a $100 book and see about health insurance.  Damn turning 23 anyway!    The sulkiness of the morning reminded me of this day last year.  (Come on, you knew it wasn\'t going to end that quietly, didn\'t you?)  This was the day of my big departure/inevitable return, and there\'s only one way to commemorate it properly.     Sunday, January 12    We all awoke on time.  There was no time for laziness today; we needed to get to the train station by 10, although if something prevented us from reaching it in time, it wouldn\'t have broken my heart.  Each stop on the Tube lasted an eternity, bringing me closer to the last place I wanted to go, but also not going quickly enough.  The frantic pace of life was returning, and I fought to hold onto the patient, relaxed lesson I learned in the city streets.   I have to remember how to do this when I get home   The process of saying goodbye was made all the more difficult because the three of us had, by the end of my stay, formed a cohesive union.  At dinner the night before, Grant and I finally looked into each other\'s eyes and laughed the laugh of friends, a slice of understanding sliding into place.  It had been decided that I would make the journey to the airport by myself, so this station was to be the site of our farewell.     I love the sound of his footsteps behind me, listening to their sighs and laughter.   \"Have a safe trip back and a safe drive back to school.  Let me know you get there okay,\" he said to my hair as we hugged.  \"I will.\"  It was the last thing I wanted to think about at the moment.  \"I\'m glad I got to meet you Grant, and I can\'t wait \'til next time.\"  \"You too.  The train\'s getting ready to leave, get your ass on!\"  And he shoved me aboard.  Then we waited.  I took a seat next to the platform so we could make faces and over-dramatic gestures to each other, laughing so as not to cry.     I smell the white room, and can feel him tapping my back in time with the music.  Wonder if I\'ll still be able to smell this city on my scarf in a month.   They ran with the train as it pulled away, creating entertainment for my fellow passengers and a perfect memory for me.   Grant was right--Coke does taste better over here.   Yay.  You can all celebrate now; no more London essay.  But it\'ll always be part of you now, don\'t even try to fight it.   '),(23242,2314011,'11,January,2004','2004-01-11 00:00:00','    Never recognize this place when I return from a break.  It\'s the damnedest thing.  Drive for hours across one or more states to drag myself from my car, down the hallway, through the door and it always looks completely strange...as though something huge is obviously out of place or missing, but, because I haven\'t been around to monitor each inanimate object\'s activities in the last few weeks, I can\'t tell exactly what.  \"They\'ve been up to something, I can tell.\"  But as soon as my ass hits the futon and my hand grasps my faithful remotes once again, the wave of \"home\" washes over me.  Probably won\'t know where I am when I wake in the morning; but I\'ll stumble into the kitchen, flip the coffee switch and it\'ll come back to me eventually.  Have two days to get organized before classes resume, for which I\'m very thankful.  Think I may do some wall-covering redecorating tomorrow.  Or maybe I won\'t even get out of bed.    The transition to my tiny abode has been a bit shakier this time, as I didn\'t have miles and miles of Montana highway on which to reflect on the break and shift into Moscow-mode.  The Boise route shaved an hour off the overall travel-time, but it feels twice as long; you have to pay attention to those damn curvy roads!  Who knew?  So I find myself missing the quiet cave of a den we called my room in the basement.  The apartment\'s quite a bit toastier than the cave, but the mountain fortress surrounding it more than compensated.   '),(23243,2314011,'10,January,2004','2004-01-10 00:00:00','    And thus, the break ends.  Doesn\'t feel like break anymore, really.  Became so used to waking up late, lazily sipping coffee, showering before noon if that\'s what the fates decided, and if not, that was okay too, that the concept of \"going back\" seems foreign.  Back to what?  Isn\'t this what life is now?  Not so much, no.  But I love that I loved this break so much.  Very relaxing, very restful.  Everything a good break should be and more.  Will have two days before school begins to reconnect with my cute apartment with positive vibes...Yay.  Probably will get nasty with the plants, too.  Might even finish a book or two, never know!  Joe and I are braving the Boise route tomorrow morning.  Wonder if he\'ll make it home tonight before 12.  Wonder if he\'ll stop when I need to pee.  We\'ve never left on the same day, so this whole caravan thing could be interesting, indeed.  The last time we traveled together, his car ended up needing $3,000 worth of repairs.  Hopefully being in separate cars will improve our record.   '),(23244,2314011,'09,January,2004','2004-01-09 00:00:00','    So, I\'m going to edit the post below, because it\'s too great to be so rough around the edges, grammar-wise.  Plus, I forgot to mentioned that Grandma kept asking about Levi after my short, to the point answers followed by subject changes.  \"Levi really is a nice young man.  We enjoyed meeting him the other night, especially after hearing about him for so long.\"  \"Yea, it was.  Too bad the birthday party had to be cancelled, everyone could have met him that night.  This place gets a lot of natural, light, doesn\'t it?\"  \"Has anything developed beyond friendship with him?\"  \"No.\"  \"Is anything likely to develop.\"  \"No.\"  \"Just friends, then?\"  \"Yea, I think that\'s why we get along so well.  Here comes Dad!  Yay!\"   '),(23245,2314011,'08,January,2004','2004-01-08 00:00:00','    I need a nap.    Or a beer.  Guess I\'ll make do with my Apple Cinnamon tea, instead.   Left the Grandparents\' new house around 1, finished lunchy with Dad around 2, and have just now arrived at Mocha Madness (for said tea and wireless internet) at nearly 4 after running all of two, count \'em TWO errands.  I\'ll tell you the story slowly, as I check my email, other blogs, and unwind in general.  It\'s not a great tale of harrowing woe, but it\'ll make good blog fodder.  Should actually be getting home soon, as the fam\'s going out to dinner this evening, but a little decompression will feel good.  Maybe I should do some Yoga in the bathroom.  Deep inhalation, and:  It all began, innocently enough, at the mountain fortress.  The day was bright, shiny, the snow was thawing, I couldn\'t pull myself out of bed and wasn\'t about to apologize for it.  Told Dad I\'d meet him out at the New House when I arose because I wasn\'t going out with him at 9.  Haven\'t seen the Grandparents since the New Year\'s Day dinner with the boy, and knew I had to see them before the big departure.  The day\'s weather was just what I needed to carry me out to Chubuck with a smile on my face and a song in my heart.  Had a nice chat with Grandma about aunts, cousins and the like.  We snuck a box out to my car to be taken to the local \"donations for poor people\" drop off, as well as a box of old books she\'d decided were nothing but dust collectors.    (Aside: a very butch looking lesbian came in a few minutes ago and hasn\'t taken her eyes off me yet.  At least, I hope it\'s a lesbian.  Don\'t have my glasses on, so it\'s anybody\'s guess.)  Anyway, offered to take the boxes off Grandma\'s hands, thinking I could score some dollars at the local used bookstore.  Not much, mind you, because used book dealers are the tightest bunch of people on the face of the earth.  But I\'m getting ahead of myself.   So Dad broke free of Grandpa\'s hold on him in the garage; they were, literally, waiting for paint to dry and Dad just as soon do it at home, but not before buying me lunch.  Yay.    (Update on the aside: definitely a butch lesbian, she just went in the women\'s bathroom.  She reminds me of April. [An old friend, not the month.]  Looks as if she\'s waiting for someone.  Hmmm.)  I had my plan all worked out:  I\'d start on the south end of town and work my way up.  Try to unload the books in Old Town, stop by the local used CD store to pick up the No Doubt Singles album, drop off the free stuff for poor people and end up back at the mall to purchase some shoes Mom and I have waited since November to go on sale.  (They never have, but I\'m sick of waiting.)  Then, because I\'d still be in such a great mood with plenty of time and just enough money, I\'d drive across town again to spend quality blogging time in the attempted-trendy coffee shop.  Super.  Well, here\'s what actually happened:  I did start on the south end of town, only to find the store I was looking for turned into a music store.  Never one to be deterred, I immediately sought out the other used book store; it\'s not on the complete other side of town, but it is on the other side of the tracks.  But on the way, I got my new CD, and we love those.  Nearly forgot about the poor people\'s shit in my trunk until I was at the turn off.  Swung Harley around to just make it, then realized I was on the wrong block.  It was only one more up, not a big deal.  Did the same circle again, plus the extra block, only to realize it was another block up.  Did the same circle again, adding the extra-extra block, only to realize it was yet another block up.  By this time, Hella Good wasn\'t magic anymore.  Thank god the guy at the front desk of Poor People R Us was nice.   Now I just had to unload my box of random 40 year old hardcover and paperback gems.  I didn\'t want to have to go where I was headed, but I couldn\'t quite remember why.  Thought I was just being harsh and elitist, where my tendencies have been known to lean.  As I lumbered up the stairs, navigated the front door and hoisted the box on the desk and heard the owner\'s first word to me, I remembered why.  \"No,\" she said, without an ounce of humor.  I smiled my \"you better be fucking kidding me\" smile and said: \"Really?  You\'re going to tell me that after I\'ve carried all these in here?  You\'re just going to say \'no\'?\"  \"I don\'t take hardcovers.\"  As if that offered any real explanation.  \"There are others in there, too.\"  As she sorted, trying not to touch too many of them, as if in fear of what she\'d catch, I tried to find a happy place by looking around the store.  Recognized the woman in the romance novel section, hands spilling over with 200 page-rs.    \"I\'ll take this copy of \'1492\', but these others are in such poor condition, I\'m not going to take them.\"  I tried to reconcile the lack of creased spines, and torn edges, absence of yellowed pages, and intense old-book smell in my offerings with her inventory\'s dusty, mite-infested, dog-slobber glazed standouts, and instead listened as her friend mentioned the other used book store in town, the one  I\'d just finished searching for.  \"Oh,\" I said, trying not to throw a fit, \"didn\'t it go out of business?  I was just over there.\"  \"Oh no,\" the friend said, \"he just relocated; he\'s only about four blocks up from his other store.\"  \"Do you know what his policy is on buying?\" I asked the proprietor, letting her know HE would be getting my business.    \"I haven\'t any idea.  But I know he only sells for garage-sale prices.\"  She started walking behind her desk to start the paperwork on \'1492\', the most-creased-spined book in the bunch.  \"Alrighty, I\'ll head on over there then.\"  \"And I don\'t sell, I only trade.\"  \"I\'ll be taking that with me then,\" I said as I reached for the nearly-ill-fated little book.  Dropped the box right in the middle of her floor as I struck up a conversation with the woman I recognized.  She\'s an old friend of my parents, with whom we spent many summer weekends camping.  Her husband owned the mechanic\'s shop Dad brought Joe and I to when we were kids and forbade us from touching or playing with anything.  We prank called 911 instead.  It was nice to chat, though, and be in the way of the store\'s owner as long as possible.    (Update: Butch\'s friend finally arrived; she looks like Athletic Barbie.  Hmmm.)  And so, armed once again with my box of books, I went in search of my original destination.  Had a mini repeat of the search for the Poor People\'s Palace, during which I cursed the No-Hardcovers-Here lady many a time.  Once I found my way inside, I took a deep breath and began again:  \"You buy books?\"  \"What kind of books?\"  \"Old ones.\"  Gave him the story.  (They moved, need to get rid of shit, not looking to make a bunch of money.)  \"Well, I only buy books people will want to read.  I won\'t buy them because they\'re old.\"  \"I\'ll go get them.\"  He took more than I expected, and mostly the hardcovers (HA! SUCK ON THAT!).    Eight dollars (it would have been more, had they not mostly been book club buys) later, I took my significantly lighter load across the street to the bohemian coffee shop where I spent yesterday afternoon people-watching, writing and reading.  \"Would you like some books?\" I oh-so-generously asked the owner.  He gave me a weary look and said, \"Uh.\"  Repeated the story and his tune changed to \"Oh!\"  \"This seems like the best place for them, I think.  They\'re perfect for laying around and getting flipped through on occasion.\"  Just get them out of my sight!!!  \"Oh sure!  There\'s a library in the back for the more progressive students.\"  \"There ya go.\"  \"You know who else might like them?\"  OH I KNOW!!!  \"The guy across the street?\"  \"Yea.\"  \"He\'s already thinned them out.  These are just for you.\"  That box isn\'t the only thing thinned out today!  But it has a good home now, and purging myself of the story has done its job.  Never did get to my shoes; think that\'ll make a good outing for Mom and me tomorrow.  They might have a secret Saturday sale, too, you never know.  She does though, because \"those Dillard\'s people are bastards!\"   '),(23246,2314011,'08,January,2004','2004-01-08 00:00:00','    Thought about it, though.  Instead, I\'ll just reference the trip occasionally.  Like right now.  Today was the day we ventured to the ghetto seeking THE Dew Drop Inn.  THE ONE.  In the song (Muhammad, my friend [from Boys from Pele--that\'s all the hints you\'re getting]).  A gay piano bar.  In London; her old stomping grounds.  It had to be the one.  We were afraid it would be so hip, so cool, so  her  that we\'d only be able to leave a candle and our nose prints mashed up against the window.  We were right, but in all the wrong ways.    We should have brought a candle, but as a memorial.  We did leave our nose prints on the window, but only in an effort to see around the boards and drawn curtains.  We left feeling very sad, disbelieving and broken.  So we went to Harrod\'s.    Exactly three months later, I told her the story and she laughed kindly: \"Yea, that\'s not it.\"  Dew Drop Inn is code for \"Let\'s get the fuck outta here.\"  Simple mistake.  Anyway, that was also the day we roamed through Hyde Park, thinking about the next (potentially last) semester and the great unknown waiting.  Don\'t think I would have imagined a year later turning out like it has.  And that makes me smile.   '),(23247,2314011,'07,January,2004','2004-01-07 00:00:00','    Torts: B  YAY!!!   '),(23248,2314011,'07,January,2004','2004-01-07 00:00:00','    The first posted grade:  Property (drum roll) B-  Cue applause.  Suppose there won\'t be any 4.0 to be had this semester, but it\'s higher than the average, and that\'s just fine by me.   '),(23249,2314011,'07,January,2004','2004-01-07 00:00:00','    An angry email, an angry-turned-to-revelations hurried conversation and finally a conversation full of open, honest communication later, I think we\'re doing just fine.  It always does the trick.    \"So, you wanna get a room Friday?\"  \"Sure.\"  That helps, too.  And, as promised, today\'s the year anniversary of a great, great day in the lives of two young (and not so young) people and a few celebrities on the side.   Tuesday, January 7    In the morning, I dressed for the theater.  The plan was to spend the day in museums and the evening enjoying Rent.    \"Looks like it\"s just gonna be me and you today.\"  I looked up from my Corn Flakes/Bran Flakes/Rice Krispies mix, questioning, trying not to look as excited as I was.  \"Grant\'s not feeling too great, so he\'s going to stay here and rest.\"  \"So I guess I better change, huh?\"  On the Tube heading into the city, a thought occurred to me.  \"How long is that Gillian play going to be here?\"  \"\'Til February, why?\"  \"Well, I\'d really like to see something I can only see here, you know?  I can see Rent in Spokane whenever I want.\"  \"That would be really great, actually.  Grant doesn\'t care to see it and you didn\'t say anything when I mentioned it before, so I just figured I\'d miss it.\"  \"Well, let\'s not do that anymore, k?  Let\'s get tickets for tonight and just do it.\"  \"Super!\"  After a stop at Starbucks and a chuckle about Virginia Woolf\'s Burgers, Pastas, and Grills restaurant, we made our way to the British Museum.    The museum is the center of  the literary district of Bloomsbury, and has served as a haven for all kinds of hippies and communists, from Gandhi to Karl Marx, to George Orwell, Charles Dickens, and Virginia herself.  They all holed up (not necessarily at the same time) in the giant reading room at the museum; their names, and those of other such strangers, inscribed on the walls of the library bring the history of the establishment and the city home in a way dry historical facts and dates just can\'t.    The distinct aroma of the city does the same thing. The air is rich and full there.  Full of industry, food, busy people, history and life.  It\'s immediately noticed, quickly adapted to, and eventually missed.  It goes quite well with coffee and perspiration, the kind one accumulates after walking all over hell and back every day for a week.   \"I\'m getting the feeling that there\'s no way we\'re going to get through all there is to see here, so just show me the highlights.\"  \"Oh, don\'t worry,\" he said over his shoulder.  First was the Egyptian halls, filled with statues, relics and sarcophaguses from centuries past.  The shadowy sense of the city\'s life that whispered to me from the beginning solidified as I ran my fingers across the bottom of 3,000 year old works.  The day before, we admired portraits painted during the 1400s, now we were standing in front of products of 1400 BC.  We were standing in a hall of legends, surrounded by gods and kings, and talking to each other felt like an interruption of whatever wisdom they were imparting to us.  In hushed whispers:  \"Over here\'s the Rosetta Stone.\"  \"Oh really?  Kick ass!\"  Pause.  \"Remind me again of the importance?\"  \"Well, when the explorers came to Egypt, this is what they used to translate the markings in the pyramids and such.  There are three different languages written on it, so when they found this, they knew how to read Egyptian writing.\"  \"Oh, right.  I get it now.\"  \"Wanna check out some Greek stuff?\"  Panels and statues from the Parthenon in Athens lined the walls, and my appreciation for the Greeks was sufficiently renewed.  We were finally shooed away from mummified bodies, pets, internal organs and food as the museum\'s closing time quickly approached.    With an hour to spare before the play, we searched for a quick dinner back in Leicester Square.  The screams of the crowd alerted Levi to the lights and posters.  \"Oh, this is the \'Gangs of New York\' premiere,\" again with the \"this happens all the time over here tone.\"  Not truly expecting anything grand, we made our way over, just for the sake of gawking.  But as more time passed, the more real the idea became that we might actually see someone.  Accordingly, we forfeited our dinner and found a space, mashed against fellow onlookers, trying to conserve heat on the blustery winter night, waiting for important cars to arrive.  First was Harvey Weinstein, looking like Jabba the Hut, in true Hollywood fashion-cell phone in hand and all.    Martin Scorsese arrived with very little bombast, in a mini-van, no less.  The next recognizable star was Daniel Day-Lewis.  Now we were down to two, only caring about one.  \"We can leave when Cameron gets here, really.\"  We watched as the clock ticked closer and closer to 7.  \"We need to leave by 7:10, and that\'s all there is to it,\" his voice of reason declared.  \"Jesus, I can\'t feel my fingers.\"  As the tension mounted and cars pulled in, the crowd was sure this was her.  It had to be.  As I readied my camera lens, I listened carefully to catch the person\'s identity.  Finally, I heard his chuckling voice behind me:  \"Oh ... It\'s Joan Collins.\"  \"What the hell is Joan Collins doing here?  Acting like this is her damn movie?\"  The next round of cars arrived quickly and suddenly Leo was there, in a mob of other men in suits, looking more gorgeous than he had any right to.  It quickly became apparent that Cameron was a lost cause, so we headed to the theater and Gillian.  Arriving in time to buy the essentials: chips, water, poster and program, we brushed by glamorous, sophisticated, important-looking people and claimed our discount sixth row seats.   I\'m so glad I changed out of my fancy, theater-going clothes this morning.   The show itself was quite phenomenal, living up to the hype of the day, and the evening\'s entertainment breezed by in what felt like a few minutes.  As we gathered our bearings outside, preparing for the journey home, he noticed the line.  As we took our place, the man announced:  \"It\'ll be just a minute; here\'s the deal: One item per person and nothing from the \'X-files.\'\"  After a few good gasps and jokes, and before the enormity really hit us, the line began moving toward the small redhead in the doorway.  She seemed really exhausted, but still gracious when forced to endure bad play-referencing jokes made by star-struck American strangers.   And it\'s only Tuesday!  If everything had gone according to plan, I\'d be flying back tomorrow, and I doubt we would\'ve spent this evening out on the town.   Yay.  Told you it would be good.    And just out of sheer randomness and end/beginning of the year shit I\'m a week late for, here\'s a list of all the books I read during 2003:  The Truth Is... Melissa Etheridge\'s Autobiography: B+ (I had to have something to take my mind off the fact I wasn\'t yet in London, thanks to the fucking Passport people!)  Mysteries of Pittsburgh by Michael Chabon: A+  The Watchmen by Alan Moore: A (Yes, I\'m counting a graphic novel as a \'regular\' book, deal)  The Lord of the Rings: A (The whole thing!!)  Good Omens by Neil Gaiman: A++ (Bought at a used book fair along the Thames, oh yea)  The Alchemist by I forgot the guy\'s name: A (We love enlightenment in fable form!)  Personal History by Katharine Graham: B (Began to love journalism more after I\'d graduated from it)  The Amazing Adventures of Kavalier and Clay by Michael Chabon: no grade high enough (It\'s become part of my very existence.)  The Hours by Michael Cunningham: A+ (Such great writing, insights, humor and wisdom packed into a book you can read in one sitting; gotta love it!)  The Artist\'s Way by Julia Cameron: A (Two of my favorite gals in one name!)  Supreme Injustice by some legal scholar/analyst: B+ (Stupid presidential appointments, anyway)  Beauty and the Beast by some frump without a life: embarrassing guilty pleasures don\'t get graded (Remember that show?  What a gem)  A Heartbreaking Work of Staggering Genius by Dave Eggers: A+ (Nothing to say, really, just read the book)  Endless Nights by Neil Gaiman: A+ (see the note after The Watchmen)  McSweeney\'s Mammoth Treasury of Thrilling Tales, edited by Michael Chabon: B-  Cold Mountain by Charles Frazier: A   As you were...   '),(23250,2314011,'06,January,2004','2004-01-06 00:00:00','    Found a first edition copy of The Amazing Adventures of Kavalier and Clay in the Idaho Falls Barnes & Noble yesterday, the day of the Boy\'s departure.  Most of you probably don\'t understand the significance of such a find, but trust me, it\'s full of significance.  Now I just have to find the receiver of my other copy; wonder who it\'ll be....   '),(23251,2314011,'03,January,2004','2004-01-03 00:00:00','    Is anybody sick of this stuff yet?  No?  Alrighty then!   Monday, January 6 	   The three of us started the day\'s adventures off right with Starbucks in Leicester Square.  The den of materialism and overstuffed sofas was to be our only coffee shop visited in London, our not-all-that-closeted elitism showing just a bit.  Soon enough, though, we found ourselves back in the city, heading into the gritty mean-streets of the Soho district, historically renowned for its debauchery and drug dens.    The neighborhood has evolved into the hip, happenin\' fashionably bohemian and supremely gay heart of London.  The \"heart\" is basically two streets lined with sex shops, boutiques with sex shops downstairs, bookstores with sex shops downstairs and restaurants flying the rainbow flag.  I\'m sure they had their own version of downstairs sex shops, though.    Heading toward the more sophisticated, more cleanly, National Portrait Gallery, we took a hurried stroll through a decidedly Asian community.  The population suddenly shrunk in stature, and nearly every storefront had Asian writing and dead chickens in the window.    \"Is this like London\'s Chinatown?\"  \"No, this is more like ChinaBackAlley.\"  \"Thank you, Grant.\"   That\'s it!  The trip\'s first iconic joke!  We might just be able to have a smooth ride, after all.   Tomorrow\'s really good, just you wait.   '),(23252,2314011,'03,January,2004','2004-01-03 00:00:00','    Happy Tuesday blogger!  Thought I\'d try an on-site posting today in the only coffee shop in town I like to spend extended periods of time; plus, they have free wireless Internet!  Yay!  Slept late again today.  Don\'t think the 9 a.m. mornings will be sticking around.  Although, I do have a hair appointment tomorrow at noon, so I\'ll at least have to be up before 11.  Done a lot of thinking so far today.  Wrote three morning pages, the first in about as many weeks, which felt good.  But there appears to be more to say.  Thought I\'d post what was happening across the pond last year today, and I probably will get to that on account of the feeling I have to just sit and write and write and write.  Initially, thought I\'d just fiddle around on the fast connection here, sip on my gourmet Mocha and read a bit to fill my afternoon.  Then I talked to Lorraine.  And now, all I want to do is write.    We\'ll get to the conversation in a minute, but I\'m more anxious to explore the thoughts the conversation\'s aftermath brought up.  They\'ve been lurking about in the back of my head for a few days now, brought to life by Natalie\'s observation in Charley\'s last week:  \"Maybe you should try being a bit more supportive\" and the discussion Levi and I had the other day:  \"It just seems you always have these bleak, cynical conclusions.\"  \"How many times have I been wrong?\"  \"Good point.  I just would rather have a bit more optimism.\"  \"I\'m not usually disappointed.\"  \"True.  I guess I\'d rather have the experience and keep up some hope; it\'s worth the disappointment.\"  Seem to go through these every once in a while.  And I\'d like to think they help me discover a bit about myself I\'d rather see handled in a different way and the process of making that come about begins.  But days like this make it seem as though I\'m a long way away from the person I\'d like to be.  Understanding, patient, accepting, fun and easy to be around.  Not someone to be handled or tip-toed about.  Then I feel completely over-dramatic and self-indulgent in my \"Oh why can\'t I be who I wish I was!!\" self-analysis.  Maybe what\'s bothering me the most is the feeling that I was on my way to being all those things all the time during the first part of the year.  As has been referenced before, the London trip taught me a great deal about letting go, having patience, etc. and I made an effort to apply those lessons upon my return to everyday life.  Once school started this fall, though, and time and energy became sparse, I think I may have slipped back into an old routine.  My critical eye, honed for any indication of red-flags or warning signs, did well for me in the classroom, but perhaps not in terms of my attempt at becoming a more centered, balanced (see the adjectives above), individual.  I tell myself that looking at my thoughts, actions, etc. is a good sign in itself, but I\'m hesitant to be too self-congratulatory.  Sigh.  Processing, indeed.  That\'s what she\'s been doing since reading Saturday\'s post.  Had about 20 minutes to talk about it this afternoon.  She doesn\'t recognize me anymore.  She hadn\'t seen this side of me.  She didn\'t think I\'d be this pissy once we began speaking again.  Told her we\'re in very similar places because I haven\'t recognized her for months.  Talked more about what the last 6 weeks meant, that they were about letting her go, being okay with her absence in my life; letting the connection we\'ve shared for so long become something new by releasing the grasp on what it\'s always been.  On my birthday she told me the time apart taught her about self-control, patience and faith; that it was harder than she ever thought it would be and how incredibly relieved she was that I\'d finally called her because she didn\'t think she could make it through the day without emailing me a \"Happy Birthday\" wish.  Today she told me the break eventually felt good; she began to enjoy the lack of contact and at some point quit missing me.   \"So why didn\'t you tell me that when I asked?\"  \"I don\'t know.\"  Last week, before she knew how I felt, it seemed like nothing had changed, that I\'d undergone a process of...well, processing, but I\'d left her in that October place.  Today, after she read how I feel, it sounded like she\'d arrived at the place I\'ve been waiting for her, had always been here, and why am I so pissy now that I\'m here too?  I just wish I knew which one was the truth.  Wish I knew how much of the situation is me and my bleak, cynical conclusions.  Wish I could use that razor-sharp analysis, which seems to oh-so-accurately pin-point the crux of everyone else\'s problems/incidents/etc., to asses my own life\'s wonders.    At the beginning of the conversation, she asked if I\'d like to meet her if she came to town Friday.  The thought initially made me hesitate and pull back.  Now I\'m thinking it may not be a bad idea.  Assuming she still wants to, that is.   '),(23253,2314011,'03,January,2004','2004-01-03 00:00:00','    Keeping up with this whole \"last year at this moment, we were\" thing we\'ve got going, here\'s today\'s entry.  Hope you enjoy it, but I really don\'t care if you don\'t.  Editor\'s  note: the original text has been livened up a bit; look for the comments in parentheses (like this).   Saturday, January 4, 2003   \"I have no idea what I want to see, hon, so you\'re gonna have to take me where I want to go.\"   First stop was Piccadilly Circus.  It reminded me of Times Square in New York, not that I\'d ever been there,  (even though I had in 9th grade)  but the bustle of tourists armed with cameras and natives on their way to somewhere else reminded me of TV\'s depiction of the nervecenter.  And at the same time, it was unlike anything I\'d ever seen in the States.  (except Nebraska [and I honestly haven\'t been there at all])   The buildings we walked by were older than the Declaration of Independence  (and smelled like it too)  and they let me feel their presence as we passed.  The juxtaposition of such history mashed together with Burger King and Tower Records stores represented London\'s crazed meshing of the present\'s frenzied need to rush and the past\'s patient wisdom.      We made our way through central London with a vague concept of destination, not really bothering to know for sure.  The theater district of Leicester Square required a gawk or two and a picture of the site of the \"Two Towers\" premiere.  (because we may have forgotten the movie\'s importance in the two days since we\'d seen it for the second time)   \"You know that Gillian Anderson girl?  She\'s in a play right over there,\"  (foreshadowing, stay tuned)  Levi said, as though, during his time here, he\'d hobnobbed with many such celebrities and now had to be reminded that the rest of us hadn\'t.    Eventually, we came to Trafalgar Square.     Okay, I know this place is important, but I\'ll be damned if I can remember why.   Apparently the plaza commemorates a great victory in 1805 of the English over that little French guy Napoleon.   Right, that ol\' thing. 	  The pigeons in the square created a perfect British moment, especially when they all took to the late afternoon sky at the same moment.  \"We\'re gonna get shit on.\"  (perfect character reference)   \"Thank you, Grant.\"  \"See that clock through there?\" Levi pointed through the traffic, across the street, many blocks away.  \"ThatÃ‚Â’s Big Ben and the Houses of Parliament.\"  \"Hi Big Ben,\" I said, waving.  	    As we strolled along, making our way toward Buckingham Palace, we marveled at the uncharacteristically sunny weather, agreeing that the only thing that would make the day more perfect was if it was Sunday.  (and if Grant wasn\'t there)   I was still trying to wrap my mind around the fact that I was actually in London.  I\'d not ever been off the North American continent before this, for crying out loud.  It was surprising how easily I gave up any sense of control, planning, or organizing each day.  Usually, I have a mentally-outlined schedule of my days.  Even the ones when nothing is happening; I have to plan what to do with the empty time.   But not here.  It was probably because the whole experience was so far beyond any other I\'d had up to this point  (or all that E we did the night before);  I was too busy living and enjoying it to think about what came next.  (because if I had any idea, I\'d have been pissed)   The frenzied nature of the present quickly faded away, and I let myself fall under this wise, old city\'s spell.   '),(23254,2314011,'02,January,2004','2004-01-02 00:00:00','    And I think I have.    Feeling a bit nostalgic and melancholy today.  The most intense waves passed in the night, thank goodness.  Nearly cracked the first 100 pages of a new book, set in London, and found myself missing the city, that experience and 2003 intensely.  The year mark always changes things.  The realization that something truly is over, passed hits home in a new way.  The trip was such a great catalyst for the year in waiting and last night found me pouting over the fact that it\'s over.  Clearly nothing could top such a great year, especially not a new year making its debut in the snow incased fortress in Pocatello.  But a hot buttered rum at 2 in the morning always helps.  The fact that I haven\'t written pages in more than a few days may have something to do with it, as well.  They\'re all upstairs making noodles in the kitchen.  I\'m sure they miss my presence immensely.  The countdown to the return back to school mirrors last year\'s countdown to the return back to the States.  Feels like I have a lot of time left to enjoy the break until I remember how fast the 10 days in London went by.  Tomorrow\'s the anniversary of the drive of harsh beauty.  Stupid January, anyway!  And then there\'s this girl.  I knew things would be different when the Age of Silence had passed, but I didn\'t think they\'d be so similar.  Guess I didn\'t expect to feel this different.  Feel her trying to make everything okay, back to the way it used to be, at least in terms of our friendship-connection.  But the few times we\'ve spoken I\'ve felt like I\'ve returned to something I worked to leave behind.  Something I don\'t need anymore, and have no intention of returning to.  And I feel like I\'m the only one in that place, only one ready to see where the new places in each other\'s lives are rather than trying to fit into  the old ones in different ways.  Find myself getting sad, frustrated and then relieved because I remember that I don\'t need or want to figure it out.  Have to write about it, sure.  But am done worrying.  This will turn into something neither of us expected, for better or worse, without any pushing or pulling from me.    Sigh.  What can I say?   She\'s my wonderwall.   '),(23255,2314011,'02,January,2004','2004-01-02 00:00:00','    A year ago at this moment, I was making my way through Gatwick Airport at the end of an 11 hour journey, watching the day begin for Britain.  Since nothing much is going on this year, other than the accumulating snow level and few escapes to movies, I thought I\'d reminisce blog-style.     Thursday, January 2-Friday, January 3    The eleven hours of traveling didn\'t seem like such a big deal until hour 7.  Sleep was never an option, even if I hadn\'t been seated in the middle of the middle section of the 737.     And just what the hell is a landing card, anyway?  Will they let me off the plane without one?  That would be the perfect thing, really, especially after the passport fiasco.  Where did I put that damn thing again?  Now just calm down.  They\'re not going to send you back to the USA if you don\'t have your stupid landing card.   	  Somehow, time managed to pass and the flight attendants distributed \"breakfast\" around \"5 a.m.,\" even though, to me, it was more like a midnight snack.  One hour to go; the butterflies were back and I couldn\'t sit still.  ~~~  Four days before he was to leave the country, Levi met his first loveâ€“Grant.  The last conversation Levi and I had while on the same continent detailed those four days and I couldnâ€™t have been happier for him.	  Before he left, the seed had been planted in my mind for a visit, perhaps over Christmas break.  Not only would it be a great opportunity for me to see my first European country, but mostly, it would be a chance to visit my guy during his year away.  We often went months without seeing each other, and there were those couple of years in high school when we didn\'t speak at all, but now, a year was forever.  It wasn\'t as though we never communicated while he was in London; we were in better contact then than at any other time during our entire friendship.  Grant and I were the only people in the States he sent daily emails to, and I woke up early each morning during the fall and winter of 2002 to read and respond.    ~~~	  I knew they wouldn\'t be there when I arrived.  The Tube didn\'t open until 7, and they said something about having to take a train to get to the airport.  The time they\'d built in to their schedule for me to secure my luggage, endure customs and such, evaporated as I found my suitcase immediately and breezed through the \"Nothing to Declare\" door.    Still, I scanned the anxious faces waiting at the end of the entryway, unable to actually make any of them out through the haze of excitement and exhaustion, hoping to hear a familiar gasp and shriek of \"Honey!!\"  When none came, I made my way to the change booth to get my hands on some British pounds.    	 Score.  Now when they get here, we can just go.  He\'ll be so impressed.   And so, armed with my new money and the sound of British conversations wafting through my mind, I waited.    I caught sight of Grant\'s short, blond, curly hair first.  I recognized him from the pictures; he dwarfed Levi\'s 6\'4\" frame by a good five inches.  I grinned and my palms sweated.     Holy god, he\'s tall.  Now where\'s my Levi?   He\'d let his hair grow a bit shaggy.  Not too much, but just enough to give him that young European look I know he has a thing for.  Whenever I see him after a time apart, he always looks just a bit different.  It\'s most pronounced in the first moment, but then he always turns back into the boy I knew in 9th grade.   Levi looked everywhere but at me, and I just sat, grinning and waiting.  Finally, we locked eyes and I barreled into him for the first time in half a year.  	 \"How are you?\" he asked my neck, laughing.  \"Oh, Iâ€™m okay!\" I replied to his ear, my legs wrapped around his waist.   	 I was at last introduced to the boy who\'d stolen my best friend\'s heart and the three of us were on our way.  \"Let\'s get a drink!\" I said, only half-kidding. 	 \"You know, we could.  The pubs open around 10 here.\" 	 We were the only ones talking during the half hour train ride into London.    \"We could dress up as nuns and go bar-hoppin\',\" Grant suggested. 	 \"I love that movie!\" I exclaimed.  \"Anyone who can work \'Steel Magnolias\' quotes into casual conversation must be alright.  We are sooo going to get along, Grant.\" 	 Watching the sun rise over the English countryside suddenly reminded me of exactly where I was.  This was an entirely different world, complete with its own unique color, aroma, sound and feel.  Even the laundry hanging on clotheslines had a decidedly English look to them.  Or maybe that was just the jet lag.    It was 10 a.m. (3 a.m. to my body) before we emerged from the Underground and I got my first real look at another continent.  The walk to the house was spent oohhing and aahhing over every cobblestone and car parked on the owner\'s front yard.    ~~~  It\'s an odd feeling, really, to be both refreshed and completely exhausted at the same time.  After my shower and an exchange of Christmas presents, I fought to keep my head up as I pecked away at my ham sandwich.  \"Okay,\" Levi said, asserting his authority, \"you really need to take a nap.\" 	 \"Yea, I probably do.  I\'ve been up for about 24 hours by now.\"  \"Don\'t sleep too much, though, \'cause that\'ll throw you off even worse than not sleeping at all.  I\'ll come get you if you sleep too long.\"  	 Iâ€™m never going to get used to these stairs.   They were entirely too steep for a two-story house, but I hadn\'t the strength to complain; it was all being used in my effort to remain upright while navigating them.  It wasn\'t even a house, really.  It was half of a house.  Everything was small, shrunk down to half the normal size.  I wondered how the flat would be without the wall of separation between each side.  From the outside, the structure looked like a fairly large home, but because of that dividing line, everything inside was reduced to fit two households into one actual house.    The room they prepared for me was smaller than the dorm room I spent two years cramped in, and the only bed covering was a heavy comforter.  Everything was white: the walls, the furniture, and the door, making it difficult to imagine the dark night my body thought it should be.  I pulled a turtleneck on over my top and snuggled close to the radiator.  	 How can they not have a single hairdryer in this house?  So what if it was built over a century ago?  Circuit-breakers, my ass.     The fog lifted slowly as he sat with me later, waiting for the go-ahead on the night\'s activities.  My hair was still cold and wet.  My body ached and my stomach cramped.  From hunger or an adrenaline hangover, I wasn\'t sure.  But I was finally here.  We were together, and I wasn\'t about to lay here, waiting for it to feel normal, by god.   \"I think I can do dinner,\" I decided.  \"We\'ll have to see about the movie, though.\"  Halfway to the restaurant: \"Fuck it.  It\'s the goddamned \'Towers\' for god\'s sake.  We\'ve waited long enough.\"   '),(23256,2314011,'01,January,2004','2004-01-01 00:00:00','    Squishy burned me.  Bad Squishy.  But I feel initiated now, so I forgive Squishy.    Just finished a round of Yoga by twilight; come to really enjoy night practices.  They might make the leap to Moscow with me.  The boy and I are about to finish watching Nurse Betty.  Decided to have a Renee marathon after Cold Mountain.  Am liking Nurse Betty more than Chicago, I have to say.  I\'ve been bitter about that movie since its initial release and have now been vindicated.  Yes, it was well done.  Yes, the actors all performed beautifully.  No, it was not better than Moulin Rouge, which every critic admitted in each review\'s opening paragraph.  And yet, the truly original, campy masterpiece was the red-headed bastard child of 2002\'s Oscar ceremony while it\'s safe, comfortable, more-refined looking younger sister walked away with 2003\'s Best Picture.  At least Nicole got even.  So it\'s 2004.  Yay.  Pretty unassuming number, really.  Next year will tear the shit up, I\'m sure.  But this just gets to be one of those cute, quiet, standing in the corner, waiting for its moment of brilliance years.  Managed to have a great New Year\'s Eve.  It\'s always a question: whether the night will be as spectacular as is fabled (expected) to be.  Started at the third viewing of Return of the King; full circle and all.  Saw Fellowship on the first day of 2002, so it\'s only fitting that the last chapter be seen the last day of 2003.  Ended up at Charley\'s where the music was the same and the drinks were still free.  Moving even slower today than the last time we stayed out dancing until 2.  But managed to get pretty for dinner with the Grandparents.    \"So what\'s my story?\"  The boy asked around 5.  \"Who am I?\"  \"What do you mean?\"  \"Well, what topics should I stay away from at dinner?\"  \"Don\'t mention you\'re an artist.\"  \"Okay.\"  \"Or that you\'re in graphic design.\"  \"Okay.\"  \"Grandpa will ask you if you\'re Jewish.\"  \"And I\'ll have to say yes.\"  \"Oh just be yourself, Levi!\" Mom yelled from the living room.  \"Yea, it\'ll be fine, honey, really.\"  \"You think?\"  \"You\'re the boyfriend.\"  \"I guess I should play the part.\"  Wonder what they think of their newest Grandson-in-Law-to-be.    He waited until they left before slapping my ass.  Mom and Grandma were discussing the poppyseed cake Grandma brought for dessert when the lightning storm began accompanying the blizzard.  The night suddenly became purple and we all gazed outside to catch a glimpse before the thunder followed.  Mom broke the awe-induced silence:   \"You know, I\'ve heard the weather has something to do with that.\"   '),(23257,2314011,'29,February,2004','2004-02-29 00:00:00','    the law school\'s pretty barren.  Beat my goal to be here and working at 1030 by an hour and just rounded out the day\'s homework, 4 and a half hours later.  Wish I was ready for a nap, but instead, I\'ll catch up on reading emails and web surfing before heading home to pay some much needed attention to the rough edges of my apartment and my joints before settling in for the big, gay Oscar gala.  Very excited about that.    It\'s an odd feeling, really, when the weekend you thought would be a low-to-medium level of intensity turns out to be one of moderate-to-high level.  But accompanying the feeling of pride in accomplishing such a feat is the assurance that you realize you adapted and overcame without even realizing it.  To be sure, you bitched and moaned to yourself as you watched the hours evaporate into darkness of evening, but you never once considered stopping simply because you thought you\'d be able to 3 hours ago.    The Spring Mix has played a large role in that process.  And no, I won\'t post the track list here.  You have to listen to both volumes all the way through before taking a peek at the tracks.  Them\'s the rules.  But I will say that it is a representation of and anthem for all the events, circumstances, evolutions happening right now.    Can\'t wait until we\'re flying down I-5, windows down, skin tanning over toned muscles, watching the skyline emerge from behind the curves, singing at the top of our voices with impossibly huge smiles claiming our faces.   '),(23258,2314011,'28,February,2004','2004-02-28 00:00:00','    Finished with homework near 8 and felt completely defeated.  The cramps were making their comeback; wanted to write this damn story down, but felt too tired to even think about anything other than the first line.  Plus, the GLORIOUS Spring Mix didn\'t play well in the player during the outline-typing session.  He was right: it\'s not something that gets its first listen while working.  So, I planted myself on the couch with the headset, gathering my notebook and the other treats included in today\'s package near me and started over.  By the fouth song, I was ready to go.  Completley rejuvenated and loving every note.    Now on the 15th song of the 2nd disc and show no signs of stopping before giving it another listen.  In the meantime, here\'s what came of the 2nd go-round of the 1st disc:    On the corner of Pocatello Creek and Common Sense  	  \"Did anyone bring anything?\" I asked, crawling out of my red Chevy pickup truck, Ned, taking stock of the area I selected for the camping trip.  It wasn\'t the first time I\'d been there.  Scoped it out on one of my many drives up Scout Mountain.  Ned at 45; Melissa blaring; thinking about being in love and imagining our life together when I blew out of this town and told the parents the truth in another life called college.  Remembering our first experiment with sex in the wilderness made me smile that night and I tried to mask it by making alcohol inquiry.     She\'s going to die when I tell her about this.     \"No,\" one of my new friends answered.  \"Rachel would be uncomfortable ...\"  \"That\'s cool.  We\'re going to have a fuckin\' great time anyway.\"  I had no complaints, really; I was with my new group.  It only took two years of high school to find, slowly approach and fit myself within their circle.  Still felt like a bit of an outsider, or at least a newcomer, as they had all been together since the cradle of junior high.  But I was accepted by the close of junior year and was now one of the gang by the August before our last together.    We constructed a fire pit and dug a pee hole.  Once the sun set, Ned\'s break lights provided an ample source of fun, especially when we cranked up Private Dancer on repeat.  The pictures didn\'t actually have the desired red glow, but we wouldn\'t know that for another few weeks.  We each struck poses in the red light and a few naughty, memorable group shots made their way onto many a roll.  Throughout the activities, my mind was continuously caught between thoughts of the other visit to this place, the other reason hands had been washed in the creek nearby, and the fact that I\'d at last found friends I could comfortably be myself with.  They didn\'t  know , of course, but I wasn\'t troubled by that just yet.  After the debacle of 9th grade, I spent most of my time in school being the smart-assed teacher\'s pet in class and the zombie girl in the basement who eventually fell in love with a secret.  Needless to say, it felt incredibly freeing to be able to reach out to a group of gals who reached back.  Gals to be gals with; gals who made me laugh and whom I made laugh back.  Gals to be 17 with.  Eventually, we made our way around the fire where the conversation evolved into alcohol stories.  \"I\'ve tried champagne, and MAN ...\"  \"Whiskey and Squirt got me sooo fucked up this one time ...\"  And, as is to be expected with situations such as these: late summer evenings in the depth of youth and impossibility, we quickly set our minds to getting our lips around some contraband.  Parents\' stashes, siblings, friends, cousins were all considered and quickly discarded.  We had to keep reminding ourselves of our innocent friend\'s sensibilities, after all.  \"If you guys are going to drink, I\'m going to go home.\"  Somehow, she managed to not make it sound quite as threatening as the words themselves suggest.  Instead, the conservative upbringing\'s guilt trip kicked in, and the attempt at sympathy had the reverse effect on the three main culprits.  Finally, Katy came up with the plan.  We would take a fifth of Jim Beam from her parent\'s basement cabinet, mix it with Squirt and \"just get buzzed.\"  We wouldn\'t get trashed, of course, because Rachel\'s beliefs still needed to be respected.    Plus, I had a dentist appointment in the morning.  Katy, Natalie and I were ready to leap into action when we were entreated one last time by the diplomats Tricia and Kjersti:  \"You guys, Rachel is really upset.  Please don\'t go ...\"  \"Fuck that!\"  I was on now; this was happening.  \"She\'s already alienating herself!\"  \"Yeah!\" Nat joined in.  \"She\'s already pissed, we may as well drink!\"  \"I know ... it\'s just that I don\'t want this to be ruined for her!\"  \"Well, then you two talk her into lightening up because we\'re going.\"  And so it was.  We climbed into Ned and made our way off the mountain, into town and the unknown, Savage Garden blaring though the speakers.  The co-pilots managed the music, so we got snippets of \"To the Moon and Back\", then \"Your Little Secret\" ...  \"No!  This isn\'t it!\"  ... to \"The Only One\" to \"If I Only Wanted To\" ...  \"I  love  this song!  But where\'s that other Savage Garden song?\"  \"Yea!  The Chick-a-Cherry-Cola one!\"  20 minutes and three scratchy throats later, we turned on to Katy\'s streets.  We cut the music and headlights several blocks away and parked one block away from her house.  We knew going in we would have to Army-crawl under the trailer in the driveway to access to basement-level patio door.  But faced with the 18 inch space, a part of us hesitated ... but only for a moment.  This would only make the story all the greater.  So we slithered, grunted and laughed hysterically, but quietly enough to convince ourselves we would go undetected.  Once inside, Katy immediately secured the glass monstrosity that was the object of this whole endeavor and Nat bolted for the bathroom.  \"Don\'t flush!!\"  Katy hissed in a whispered tone.  And I was left watching the spectacle, already shaking my head at our adventure I didn\'t want to end.    Several minutes that felt like hours later, Katy and I ushered Nat away from the tiny-so as not to be heard-stream of water coming from the sink in her concentrated effort to wash her hands and we retraced our steps and slides back to Ned, waiting in the wings.  But our adventure was only half over.  We needed mixers now.  And cups.  The closest and only source we considered was the convenience store positioned on the turn-around corner of the police force\'s nightly cruise route.  It was a low-risk night, being late Thursday night, early Friday morning, but our comfort level was compromised as we parked a few spots away from a cruiser.  \"Don\'t leave me in here with the booze!\" Nat wailed as we opened the doors.   \"Just don\'t waive it around and you\'ll be fine!\"  We admonished her.  \"Besides, I have to pee!\"  While Katy procured the liter bottles of our high quality mixer and red plastic cups, I tried to convince myself it wasn\'t completely obvious to the clerk what we were doing in Common Cents at 1 in the morning, buying such items.  But he went for it and we were not deterred.    Pulling ever-so-carefully back onto the street that would carry us to the interstate and our night of drunken debauchery, I took stock of the situation:  \"Okay!  I don\'t have any broken head or taillights, my registration is current, we all have our seatbelts on ... right?!  And I\'m going 75 all the way out there ...\"  \"Andrea!  You\'re going fucking 45!\"  Indeed, in a 35 zone, next to the fire department\'s headquarters.  The rest of the drive back to the mountain found us at the posted speed limit, no more than, and our voices pitched at the highest level.  By the time we returned to the site, the alcohol was completely unnecessary, but any promise to stop at buzzed was left under the trailer.  We drank; we danced around the fire; the diplomats joined in and continued to console Rachel.  We three were left to our own drunken devices, as no one else seemed to want to have anything to do with us, professing our oaths of devotion and sisterhood for all time, endlessly recounting our odyssey and that fucking tiny stream of water!  We were shunned by all but the occasional crossover by Tricia, who answered our every cry of: \"TRICIA!  Get over here!  We love you!  You are my best friends EVER!\" and corresponding group hug.  Many trips to the pee hole were made, me on Katy\'s arm for drunk-leading-drunk support.  We all got caught up on Katy\'s sex life and were immensely relieved to learn she was at last in the habit of having orgasms.  I couldn\'t help but smile, thinking about how my addition to the conversation would change everything.  Instead, Katy became the self-proclaimed experienced one among the group, with Nat coming in a close second, and I looked over at the spot Ned then occupied, remembering.  We outlasted the rest of the group, finally crawling into Ned\'s bed, none of us allowed in the tents, as the sky began lightening with the first rays of the day.  At 7, I moved into the cab.  Around 10, as I made my way into the house, hoping the hangover wasn\'t as obvious as it felt, Mom caught me just as I tried to get downstairs to my bed.    \"Have fun?\"  She asked.  She knew.  \"Oh yea ... definitely.\"   Don\'t look at her, avoid opening your mouth too much!   \"Have a little to drink?\"  \"Ummmm...\"  \"Somebody bring some wine coolers?\"  \"Uh ... yea.  They hit me a bit harder than I thought they would, though.\"   '),(23259,2314011,'26,February,2004','2004-02-26 00:00:00','    All I want to do today is huddle inside the apartment, drinking tea, listening to CDs from 1998, writing like mad.  But there\'s also an incredible urge to do some hard-core legal research and analysis on the Appellate Brief bear.  The cramps suck, but the emotional clarity, zen-like high they ushered in make them easier to cope with.    Think the months of morning pages are finally doing what they promised to do in May.  Clearing out years of pent-up thoughts/feelings/issues/etc., making way for new thoughts/feelings/issues, etc. in order to facilitate the creative process.    I once promised Lorraine I would write our story for the world to read; it was the kind of conversation that followed long, lazy hours in the secret apartment (aka The Nest), and the promise made with an 18 year-old\'s seriousness and sincerity.  It was hard-core.  But today, all I can think of is how much I want to write about the camping trip on Scout Mountain the August before senior year; the nights at the Nest; learning how to shift new-to-me Harley\'s gears.    Not sure now is the time of the great Telling; but it is time to start getting these stories/memories on paper.   '),(23260,2314011,'26,February,2004','2004-02-26 00:00:00','    Think I may have over-cooked myself last night.  Woke up in a pre-menstrual haze that refused to lift until after the first class of the morning.  It was that great mix of phantom headache, teasing cramps, iffy-stomach and equal parts vulnerability/fragility and irritability/ambivalence.  Couldn\'t even get beyond half a morning page before having to head back to the couch to charge up for the shower challenge.  It was a worry, let me tell you!  Resigned myself to the fact that I\'d have to skip yoga and spend the evening on the couch feeling hormonal.  Upon my return this afternoon, however, I found the possible source of my troubles:  the apartment was fucking hot.  I love Spring on the Palouse.  I really do.  But it\'s schizophrenic nature can be a bitch when it comes to adjusting the thermostat.    At any rate, this was not the mood in which I wished to chat with former (and sometimes formidably intimidating) boss Jodi about the terms of my summer employment back at the UI Library.  But the gorgeous, 50+ degree weather helped a lot, what with the fresh air, bright shining sun and skip trying desperately to return to my step and all.  She was on a break, of course, by the time I made it to the 4th floor; a 15-minute one, supposedly.  45 minutes and a thorough catch-up session with my favorite geezer, Cort, later, boss-lady and I sat down to work out the details.  I was strong.  I wasn\'t going to budge.  I was taking August off and she\'d better not screw me too much on my return-pay.  Jodi\'s a bit notorious for that kind of thing, so I was expecting a haggling-session just to get to $7/hour; down $0.30 from when I left last August.    \"So when do you think you\'d start?\"  she started, simply enough.  \"Probably right after school\'s over; second or third week in May.\"  \"And you\'re planning on working the whole time?\"  \"Pretty much; I\'d like to take a few weeks off near the end of the summer to visit the fam and all, but that would be pretty much it.\"   If that\'s okay with you, I mean, I don\'t HAVE to see my family.  Just tell me; whatever you think is best.   \"Okay, well, we\'ve got this project ...\"  ---Insert various project descriptions, all in friendly, cheerful, amiable nature here.----  \"So really, there\'s enough work for you if you want the full 40 hours, or if you only want 30, that\'s fine too.  These aren\'t that pressing, they just need some regular attention.\"  \"Well, there  is  this world-saving, humanity-enriching venture I wanted to check out after school\'s done, but that shouldn\'t take too much time.\"  \"Okay, just let me know what happens.  And if you get another part-time gig, that\'s fine too.  These hours are really flexible.\"   WHO THE HELL ARE YOU AND WHAT DID YOU DO WITH THE BITTER, CROCHETY, JEALOUS SHELL OF A WOMAN I USED TO WORK FOR!  At least I knew where I STOOD with her!   \"Now, you were making $7.30 when you left, right?\"  \"Right.\"   Now it\'s on.  (Mentally rolling up the sleeves, sliding the lawyer-to-be facade into place.)  \"Okay, well, as I told you before, the budget\'s been really tough lately.  They gave us a bit more money for student workers this year, but that will run out at the end of the fiscal year.  Sooo, hmmm....yea...I can probably only pay you about $8 or $8.50.\"  (Putting eyes back in my head and trying pass that huge gasp off as a cough)  \"That\'s fine...\"   did that sound like a squeak?!   Yea, that\'s right!  Try to fuck with ME, will you?!  Guess I showed  HER  who\'s the  real  boss!    Epilogue:  The early yoga class was spectacular, as has been an evening of homework, laundry, pleadings from a spicy number to join her for San Fransisco Pride and Thursday night TV.   '),(23261,2314011,'25,February,2004','2004-02-25 00:00:00','    Okay, so I omitted to tell you all about yesterday\'s low point:  the UI administration, in all its divine wisdom, decided the University\'s Press is not profitable and \"liquidated\" it.  It\'s gone; no more Press at UI.  I returned to Contracts after seeing the Arg\'s headline furious, having to explain that, no, the paper wouldn\'t be effected, just the only academic press in the state that provides an invaluable service called book-publishing to the population at large.    So I decided the only thing to do would be to join my fellow lovers of books, especially those made by an award-winning Press on such non-profitable subjects as racial equality, Nez Perce history, trail/hiking guides through Lewis and Clark\'s trail country, not to mention the amount of material authored by faculty and grad students (Oh, and the Story of Idaho textbook published by the Press has been adopted as the standard state history textbook by the Idaho Department of Education) and write the new president-elect a little email politely asking him to reconsider this fucked up decision by our current prick of a president.    I left out the discussion on how the addition of a new swimming pool to benefit the new women\'s swimming team may not be an adequate trade-off for the Press\' undoing.  At any rate, here\'s my letter:   President-elect White:  I am adding my voice to the assuredly voluminous letters/emails etc you have undoubtedly received in the wake of interim President White\'s announcement of the closing of the University of Idaho Press.  Much has been, and will continue to be, said about this situation, some from faculty and staff at the University and some from concerned citizens of the community.  I would like to add my perspective as a recent UI graduate and a current student at the University\'s College of Law, in order to help illustrate how valuable the Press is to the student body, past, present and future.  When, in the spring semester of 2002, I completed a semester-long internship program at the Press within the Editorial department the Press was able to support only  one intern per semester.  This experience and knowledge I gained from the internship was invaluable to my studies in Communication.  The opportunity to work on an in-depth, hands-on level with various stages of book publishing provided not only an increased understanding of the industry but also an expansion of my abilities as a graduate from this institution.  The rate of students enrolled in similar internships increased by 9 in the following two years, bringing the total number of interns to 11 since 2001.    This is only one aspect of the Press\' rate of growth and its assured promise of  further expansions.  In addition to the internship program, the Press has facilitated a course in publishing, in which 23 students enrolled.  Furthermore, the Press supports faculty members in the English, Natural Resources, Biology, Chemistry, Entomology, History, Law, Engineering, Chemical Engineering, Physical Science, Agricultural  and Life Sciences, Anthropology and Sociology, Art and Architecture, Library, and the Center for the Center of Disabilities and Human Development Departments by publishing and/or advising them on their written works.  The Press-published Story of Idaho textbook serves as the standard state Idaho state history textbook by the Idaho  Department of Education.    In pursuing your dedication to enhancing academic experiences for students, who  ultimately subsidize these programs, I strenuously urge you to reconsider this administration\'s decision to do away with the state\'s only academic press.  These difficult decisions effect not only our current student, faculty and community populations but also  the future generations of possible alumni.    Thank you for your time and consideration.   That\'ll show \'em!!   '),(23262,2314011,'23,February,2004','2004-02-23 00:00:00','    Some granola-type-hippie-liberal probably has some solar-powered bath beads coming to her as a cosmic reward equivalent to a golden toaster oven for all her conversions lately.    I suppose we could track the first of them back to last summer\'s conversion to the study/practice of Yoga.  Then there was the autumnal addition of tea to my regularly-ingested substances.  But things really kicked into high gear in January.  Maybe the gurus were offering additional incentives to their minions like free dredlocks for the biggest, most dramatic conversion.  Because really, the last two months have seen some major shifts:  regular lunches at the Co-op; trading AOL for Earthlink; substituting Lucky Charms/Honey Nut Cherrioes/etc. for Grape Nuts each morning; and the most recent:  the local Top 40 (read: Pop music disguised as hip hop) station\'s been usurped on my clock radio for NPR.  There\'s just something about listening to a British woman tell me about the weather in Pullman that\'s uniquely fulfilling.    What\'s next?  Substituting my usual capitalist deodorant for crystals??!   '),(23263,2314011,'23,February,2004','2004-02-23 00:00:00','    \"Those handicapped tenants must be at it again; causin\' trouble, they\'re up to no good.\"  I thought as I passed three police cars parked near the said-tenant housing complex shortly after my turn onto Levick street, headed to my building after a less-than-enchanting visit to the movies.    But the two police officers walking up the stairs and into the hallway leading to my apartment made me wonder what was really going on.  Were my neighbors embroiled in a brawl?  Did I have terrorists living next door?  Was I going to see an actual police-performed eviction?  Repossession of illicit or unpaid-for property?  My excitement waned when I realized they could be headed to my apartment as easily as they could be after my neighbors.  As evidenced in the last few posts, a certain anniversary had been weighing heavily on my mind the last few days, which was the last time a police officer had visited my humble abode.  The coincidence was a tad unnerving.    So was finding two officers standing at my door when I entered the hallway.    \"Can we talk to you for a second?\" one said, with nary a smile or formal introduction.    \"About what?\"  I asked, threading myself between the two blocks of men to protect my vulnerable front (and only) door.  \"Who lives here?\"    \"I do.\"  I answered, trying to remember if I\'d put my bags of green vegetable matter away and locked the cage in my bathtub holding the kidnapped young heiress to a Greek fortune.  \"Anyone else in there?\"  Again, like we know each other.  \"No; I live alone.\"  \"Well, we received a report that crying and yelling was coming from this apartment.\"  \"Oh thank god!  You boys really had me scared there for a minute!  That\'s just the ghost of an Asian maid who visits me.  She doesn\'t actually  live    here, see.  She doesn\'t like it when I see movies without her.  But when I tell her how bad this one was, she\'ll quiet down, I promise.\"  \"Can we take a look inside just to make sure?\"  I thought for a second about refusing; after all, I\'d spent a month writing a memo in support of a motion to suppress tangible evidence acquired as a result of an illegal police entry.  But, if the ghost really had been throwing a fit loud enough for the cops to be called by a neighbor (I\'m choosing to believe she didn\'t call them herself, which she\'s threatened to do on occasion), she may have also let loose the Greek heiress, who\'s a bit of a hassle to deal with once the gag is removed, in which case, I\'d like a bit of police protection.    \"Sure,\" I said.    They stayed in the hallway while I took the first few steps in and looks around to make sure everything was safe for human consumption.  Once I was confident the ladies hadn\'t betrayed me, the police asked once more: \"Can we come in?\"  \"Oh sure!  They\'re...I mean, SHE\'s fine; doesn\'t seem like she\'s here anymore.  It\'s cool.\"  And, seeing as the situation was under control and I was clearly alone with my demons, they departed, still having never introduced themselves, the inconsiderate pricks.   '),(23264,2314011,'22,February,2004','2004-02-22 00:00:00','    Being the follower I am, I thought I\'d post something here I saw on another\'s blog.  Neil was asked to hit \"shuffle\" on this computer\'s jukebox (or its equivalent) and post the first 15 songs it listed.  Here\'s my list; those out there with blogs feel free to do the same; those without, post in the comments.  1.  Utopia - Alanis Morissette 2.  Mary - Tori 3.  Can\'t be With You - Cranberries 4.  She\'s Got the Look - Roxette 5.  Rattlesnakes - Tori 6.  Good Morning Baby - The Boy\'s London Mix Volume 1  7.  Green - see the above 8.  Bent - Matchbox 20 9.  Come Back Home - Pete Yorn 10.  Carbon (live) - Tori 11.   To Every Captive Soul - Hannibal Soundtrack 12.  Of Square Waves/Random Thoughts - Billy Corgan from the Stigmata Soundtrack 13.  Karma Police - Radiohead 14.  Silent All These Years - Tina Turner 15.  Do You Have a Little Time - Dido  Only 4 Tori songs in 15; her grasp is diminishing...   '),(23265,2314011,'22,February,2004','2004-02-22 00:00:00','    The windows in my apartment are open.  When I return at the end of the day, the place will smell and feel very Spring-ish, with a hint of Summer, waking up along the edges.  And then it will snow next week, I\'m sure.   '),(23266,2314011,'21,February,2004','2004-02-21 00:00:00','    A year ago today, I was trying to reconcile my plans for the day:  breakfast with Carrie, writing the story of the Boy and I, enjoying the fresh, spring haircut, and maybe renting the newly-released QAF Season 2; with having to call the police, parents, landlord and Lorraine to tell them the news.    It was remarkably different from today thus far.  There\'s no snow, my hair is still long from the winter and it\'s a bit overcast.  I\'m immersed in a new writing assignment I find myself getting more excited about by the day, and don\'t plan on leaving the Grind or my homework until 7.  Then I might go home and watch a couple of discs of QAF Season 2.    We love circles.  The story\'s published in the law school newsletter today and I\'ve already received my first piece positive feedback.  \"God loves you, hon.\" was the last line of the note and it reminded me of my landlord\'s wife\'s reaction to the news:  \"God will get them in the end.\"  It may have been a different sentiment if I\'d explained my theory about why the attack took place, but it was sweet, nonetheless.  So it seems the most important bits of that day and that time have stuck around:  the friends, family, car, the smile, laugh, and Queer Debbie with her brownies; while the hardest to let go of bits have managed to fade away:  the worry, anxiety, fear and suspicion.   '),(23267,2314011,'19,February,2004','2004-02-19 00:00:00','    As much relaxed, almost-free time crosses my path every once in a while, I manage to find ways to fill up the minutes with things other than relaxing.  Lately, it\'s been the new internet connection.  It takes up much of my free time.  Getting everything \"adjusted\" and \"optioned\" just so, then finding out all the tinkering left out the one step needed to make it work.  But it\'s all really peripheral stuff.  I have a reliable, stable internet connection, and before too long, it will feel like home.  It\'s odd, really, how different the whole thing feels.  It\'s the damn internet, for crying out loud!  It\'s the same!  Just the access to it is different.  It\'s just that I remember when AOL 4.0 was high-tech.  I grew up with the ISP (tear!).  And, as necessary as the break is, a part of me does miss the kiddy-gloves AOL handles all its subscribers with.  And until the new email accounts are up and running, it might be lurking in the shadows, faintly resembling a safety net.    So today was a very easy, light day.  It\'s amazing what a difference a cancelled class makes.  Plus, it\'s Joe\'s 20th birthday.  Ordered the 1st season of South Park on DVD over a week ago and it was due in at any minute.    Was doing some low-back twists in the early Yoga class (done for the day in time for the 4 o\'clock class!) and it hit me:  \"I should get Joe a birthday cake.  He\'s probably not going to have anything going on tonight, anyway.  Yea, I\'ll get him a cake; or at least a pastry of some kind.  Then we can watch some classic South Park tonight, because really, it WILL be in the mailbox when I get home.\"  And so, after class, I secured a slice of German Chocolate Coconut Cheesecake from the bakery next to the Yoga center and was feeling pretty damn Laura Ingalls about it.  Sure enough, a DVD-sized package was waiting in the mailbox.  I smiled, watching the clouds turn a bright, deep pink (as deep as pink can get, I suppose) against the early evening sky.  Noticed the name on the return address listed as \"Wal-Mart\" and the brows furrowed a bit.  \"Is that supposed to be funny?  Well, the Half.com user names could represent any company, I suppose.  It was listed as factory-sealed after all.  This will make a great blog story, at least....fucking Wal-Mart...\"  Tore through the industrial-strength plastic wrap, trying to figure out why the package was upgraded to 1st class when it should have been media mail.  Saw the white packaging first, thinking they must have wrapped up the box with tissue paper; then recognized the familiar masthead:  black, bold, lower-case letters and a classy thin rainbow running along the bottom.  Queer as Folk:  The Complete Second Season.  \"What the fuck?  They sent me the wrong thing!!  Not that I\'m complaining, but what the fuck!\"  Flashback to a conversation earlier this morning:  \"Your surprise should be there any day; Levi\'s too.\"  \"What did you get him?\"  \"A scanner.\"  \"Shut up.\"  \"No really, I got him a scanner.\"  \"You did not.\"  \"I did too!\"  \"He already has one!\"  \"He said he wanted one!\"  \"He says all kinds of shit!  Now what did you really send him?\"  \"A scanner!!\"  \"Okay, whatever.  Do I get one too?\"  \"You don\'t want a scanner.\"  \"So what did I get?\"  \"I\'m not telling.\"  \"Is it a briefcase?\"  \"No.\"  \"An espresso maker?\"  \"No; you can\'t plug it in.  It\'s just something you like and you said you wanted.\"  \"Did you really get him a scanner?\"  And so on.  Fade back to present:  No South Park, but Lorraine\'s surprise was perfect.  The first season now has a buddy on the DVD shelf and the third is being released Tuesday.  Balance is a funny thing; and it\'s our friend.    And by the way:  it wasn\'t dark at 6 tonight; it wasn\'t light, but it wasn\'t dark.  Spring is coming; and it\'s almost here.   '),(23268,2314011,'18,February,2004','2004-02-18 00:00:00','    1.  I am now a proud subscriber of EarthLink.  2.  Dr. Howser\'s decided to submit a chapter of her personal story to the law school\'s newsletter; it needs to be heard.  3.  Said doctor convinced me to submit a story of my own.  4.  Brought the computer to Civil Procedure class for the first time.  5.  Major shit is going down in said class about the suit against Wal-Mart for gender discrimination and I shall now enter the fray.   '),(23269,2314011,'17,February,2004','2004-02-17 00:00:00','    ...but the Boy makes a mean dykey vagina mix!  I\'m not even all the way through it and I have to write something about it!!  The damn thing couldn\'t be a better finale for the day.  Classes, homework, and media board were all dealt with swiftly, deftly, without breaking a sweat, and yoga perfectly mixed challenge and ease.  I rock at tree pose, I have to say.  Found myself planning the rest of my evening in the meditative corpse pose when I should have been focusing on my breath and thought:  \"I\'ll end up eating a burrito, reading the pleading and Rule 23 while listening to the new mix; it will be waiting when I get home.\"    The only addition to that picture has been the chai tea and a couple extra surprises in the envelope.    And each new song finds my jaw on the floor and a \"SHUT..UUUPPPPP!!!  HE.DID.NOT.PUT.THIS.ON.HERE!!!\" on my lips.  Think this might even have November beat, which makes me all the more excited for the Spring Mix.  I shudder (not unpleasantly) when I think about the inevitable Summer Mix Volume II.    Let\'s solve the world\'s problems with mixes, honey!  Forever!  On Cherry Avenue!  CHERRY AVENUE ON THE VAGINA MIX!!  LOVE IT!  And then we\'ll hit \'repeat\' and do it all over again! 							  Author\'s note: this post was written Tuesday evening around 9-ish; posting was attempted around 10-ish the same evening, but AOL ... well, you can guess.  Hence, the posting this morning.  Note to author\'s note: AOL shall soon be no more on this system.   '),(23270,2314011,'17,February,2004','2004-02-17 00:00:00','    I wish my Civil Procedure professor would let us know which sections of the book are superfluous  before  class.  It\'s a draw really; the student is either damned if she reads the meaningless material or doesn\'t.  He\'ll either skip right over it saying:  \"It\'s good to be exposed to the historical backdrop upon which our modern system is based, so if you read the assignment for today (all 20 pages of it) consider yourself exposed.  It\'s not on the exam and I\'m not going to spend any time on it.\"  Or he\'ll take the material line by line, asking only the most blatantly obvious questions, which are sure to confound the conservatives in the front rows.  My favorite trick of the professor, though, is his condemnation of the book\'s authors for including such irrelevant material and poor organization of the concepts discussed in the book, as though he has no power in how the class and the readings are organized and presented.    But, next year\'s students will have a better time of it because of this test-run called his first year of teaching, and that\'s an encouraging thought.  I suppose.  In other news, socialist hippies make great lunches and even greater atmospheres in which to enjoy said lunches.  And postal workers who refuse to send comic books via media mail because of the advertisements included therein can bite my ass.   '),(23271,2314011,'16,February,2004','2004-02-16 00:00:00','    Glad to hear that notorious page gave my classmates as much trouble as it did me.  Who would think editing a single page after a writing assignment would be so difficult?  After all, it\'s what I was trained to do in my undergraduate degree, both in and out of the classroom.  For whatever reason, the universe conspired against us and many hours were spent retaliating against Page 11.  It was a difficult battle, and many casualties were suffered.  That beautiful policy pontification; the beyond-perfect quote from a spot-on case.  Ultimately, though, we were made stronger and our papers more persuasive because of those editing passes.  Or so we\'ve convinced ourselves.  At any rate, the thing\'s done, finished, turned in and left to the good graces of our esteemed Professor.  And we are all a bit happier because of it.    The weekend rounded out rather nicely.  Everything on the agenda completed, including some quality time with the orphaned novels.  Plus, dear Harley received some long-overdue attention, involving open highways and loud music.  Now the week\'s off and running without Monday\'s usual easy introduction.  In addition to the usual heavy class load, this week is full of activities: the classy dame bringing the major class action suit against Wal-Mart for gender discrimination will be here tomorrow, sharing her experience and imparting her wisdom during the day and cutting it loose with students, faculty and local attorneys at the winery downtown in the evening; Natalie\'s competing across the way and that always leads to exciting (if not a bit frightening) places; and Joey\'s birthday is Thursday, so I\'ll have to arrange some fireworks for that.    And then it will be Friday again!  Who knew?   '),(23272,2314011,'16,February,2004','2004-02-16 00:00:00','    Doing my part to help Neil spread the word:  A panel of five people, whose identities and guidelines have to be kept secret, who don\'t know they were actually on a panel, never met each other, nor knew that their views on TV would decide what was and was not suitable for the deaf to be able to watch have compiled a list of shows  urlLink acceptable  (and  urlLink unacceptable ) to continue receiving closed-captioning funding.    No more Law and Order, but all of FOX News was spared.  Seems logical.   '),(23273,2314011,'15,February,2004','2004-02-15 00:00:00','    Actually got cracking on the writing assignment\'s rough draft around 1030 this morning.  YAY!!  1130-noonish is usually my start-up time, so a full hour ahead of schedule is an achievement.    Two hours later, my argument section is looking pretty spiffy and 11 pages long.  Just have to add the statement of facts and edit the whole thing down to 10 pages.  The most difficult part is over now, though.  Just have to employ those little journalistic skills:  reporting and editing.  Admittedly, this type of reporting is much more involved and entertaining because we get to tell the story from our client\'s perspective; we get to be as pointed, sympathetic and unforgiving as we want.  Think I\'ll take the Man vs The Man conflict theme to describe poor Stephen Perry\'s plight of being the victim of Devil Detective Danielle\'s vile lie about terrorists across the street when she really wanted to bust his hippie ass for pot distribution.  I\'m debating whether or not to include the suspicion I have about Danielle\'s motives; you see, I think Stephen wooed Danielle\'s lesbian lover away from her and now she\'s exacting her revenge by overstepping her official police bounds.    Think it would start out like this:  \"Danielle Washburn was a woman on the edge; a woman apart from her better judgment.  She was hurt, humiliated, humorless and looking for something or someone to hunt.  Who better than the man who ripped her very world, her reason to live and love from her and filled it with intoxicating vapor?\"  Oh yea, this definitely has potential.   '),(23274,2314011,'14,February,2004','2004-02-14 00:00:00','    Happy Valentine\'s Day, blogger!  It doesn\'t seem to count when it falls on the weekend.  Yesterday felt like the big Heart day; probably because I found myself at the mall (in the middle of the afternoon!  GASP!) surrounded by cards, decorations, sales.  I\'m a fan, I have to say.  Valentine\'s day is our friend.  Even those of us who don\'t like to admit it and hide behind a veil of faux ambivalence/hostility like it just a bit.  They just don\'t like their aloneness; not that they should, but still.  It\'s just a cute little holiday that comes out of nowhere and is a nice pick-me-up in the barren void of celebrations left by Christmas and New Year\'s.    So yea, made it out to the mall yesterday for something other than a birthday card or a pound of coffee.  And have a terribly over-priced, retro-ugly, history-in-the-making new shirt to show for it.  Yay!  It\'s a terrible combination of sky/baby blue and \'70s green shag carpet, sporting the stylish felt lettering announcing \"Greenwich Village, NY Speech and Debate\", complete with a big green bird mascot and gold star as background.  It\'s hideous and gorgeous in its hideousness.  It\'s such a Pride shirt; I can\'t wait to take it to Seattle this summer.  Sure, I overspent by at least $15 (and that\'s including the 20% off), but I already love it and am not ashamed.    Stayed out too late, but it was worth seeing Monster.  Charlize sooo has the Oscar in the bag.  I just hope she uses her power for good now.  It\'s a fool\'s hope, I realize, but we\'ve seen what those can do for people!  And hobbits!  It didn\'t leave me as discombobulated as I anticipated, which was nice.  But Charlize\'s performance certainly has stuck around in the back of my mind, which isn\'t entirely all that pleasant, but what can you do?  Anyway, as a result of my latenight excursions, I was left in a less than stellar position to kick ass on the final draft of my writing assignment today.  Woke up too early to be refreshed, but too late to go back to sleep.  Felt hungover only without the alcohol and great stories.  Hated that the easy, relaxed week would have to be kicked and pummeled to loosen its grasp on my work-ethic.  So I pouted in three morning pages and felt a bit better.  Got a surprise call from the Boy and by noon, the day was looking up.    Worked for a bit on said paper, mostly shaping up what I had, inserting parenthetical suggestions like:  (Expand net line, incorporating it into \"trick\".)   and  (Reyes: discovered ruse; clearly the truck wasn\'t hurt and he was the subject of the search; the guy walked out into the street before he discovered the truth and that wasn\'t an intervening act)   and  (Garcia rebuttal: ruse was discovered/over before police were allowed into his house; here, she had already illegally entered by the time the ruse was discovered, tainting anything that came later; all wrapped up because she was admitted illegally)   throughout.  Hopefully they\'ll be helpful tomorrow when I (really, actually, honestly) go to town on it.  Made it to school around 2, typed up Friday\'s class notes and have been pretty successful at wasting the last hour.   Sigh.  At least my mood\'s been saved.  Think I\'ll go do the hunting/gathering routine, then do the nesting thing back at the hovel and see about maybe, just maaaybe doing a bit more paper-work before closing the day down.    Random thought: I miss Juno and Isaac.   '),(23275,2314011,'12,February,2004','2004-02-12 00:00:00','    You want blog fodder, Ms. Mossman?  I\'ll  give  you blog fodder!  It\'s come to my attention that a certain someone seems to get nothing for her trouble in regularly reading, and now commenting on (The Platudinarian), this blog other than a miserable critique of grammar and spelling.  Most of my friends have received their own introductions here: some a bit edgy, others more touchy-feely; but she\'s only been mentioned in passing or in stories of funny moments in otherwise ordinary days.  Always one to jump on the remedy train, I thought I better do something about this issue, and right quick!  I mean really, here she is, taking a chance, reaching out at last, after much prodding and harassment and what does she get in return?  Well, she gets this; her very own introduction. Hope it\'s what she wanted.  She\'s the enigmatic one of our class.  Always has been.  The one everyone has a sneaking suspicion about, but few actually know.  She\'s just recently begun to come out from behind her shield.  And I think she\'s starting to like it.  I remember seeing a bit of that person around the shield the first few weeks of the school year and instantly wanting to know more.  She was so quiet and protective at school.  And, really, who can resist such a great mystery?  Once she found her comfort-level, though, she was incredibly open, honest and generous with herself.   She\'s the one thinking deep, profound thoughts and letting others spout off incessantly.  She\'s the one studying 14 hours a day and lets me get away with claiming I \"carried\" her through finals.  She sings along with a favorite song and wonders if she\'s shown too much of herself.  She begins linguistic trends in strangers\' households without even trying, for Christ\'s sake!  She\'s more careful in showing off her bongo drums than in telling stories of life-changing experiences; it\'s easier to share who she was than who she is.  But she\'s starting to do just that more and more.  I hope she likes what she finds.    She used to \"not engage\"; now she\'s writing jokes in her class notes.    She\'s still hesitant to let anyone know how well she does or how hard she works; but she\'s working on finding ways to articulate those deep thoughts without too much exposure.  This is someone I hope to know forever and am grateful for each piece of herself she shares.   '),(23276,2314011,'11,February,2004','2004-02-11 00:00:00','    Some people you look at and know instantly; others take time to see fully, but there\'s nothing like it when you begin to see them.    Tonight, I asked a bag of runes a not-at-all-specific question about where I am, where I\'m going, who I\'ll meet.  It told me to start (and/or continue) letting go of my old ways, worn habits, usual modes of operation.  The Jesus-fish-looking symbol was good for a round of laughs, but the rune\'s description made me smile.  It was the perfect reminder and the best encouragement.    I looked at the people around me, my friends, and saw the truth and reality of its entreaty.   '),(23277,2314011,'10,February,2004','2004-02-10 00:00:00','    A package from the parents arrived at Joe\'s this afternoon with Valentine\'s Day treats.  The usual:  cookies, money for something fun, etc.  And a new coffee maker!!!    My God!  If that\'s not a sign of good things to come, I don\'t know what is!  I can\'t just throw away my old one, though.  Lorraine bought it; it\'s my first, to go with my first apartment.  Maybe I can turn it into a planter for Trudi; or keep it in Harley; or have it mounted on my wall...   '),(23278,2314011,'10,February,2004','2004-02-10 00:00:00','    Today\'s one of those days.  Haven\'t had one since high school, but the feeling is unmistakable and completely impossible to fully describe.    It\'s one of those days that begins as all others do; nothing is amiss, there\'s no reason to think it will be any more or less demanding than yesterday or tomorrow.  And maybe even the first couple hours aren\'t that special, either, but at some point, the day\'s activities become a mere pretense.  You\'re going through the motions because there\'s nothing else to do.  It hit me in Civil Procedure, after an hour of listening to the professor give us a book report on A Civil Action and taking bids for 2 hours\' worth of legal advice.  No notes, no necessary information, no expectations.  The entire school day pretty much matched that level of involvement.  Not that I\'m complaining, mind you; it\'s just that a day like this hasn\'t come along for a long while.  And to top it off, it\'s an unbelievably beautiful spring (that\'s right! SPRING!) day.  Clouds gone, sunglasses required, window optional.    So I guess it\'s only fitting that the day\'s turned out to be a bit of a collage of various thoughts/memories/feelings that haven\'t visited for a while.  The old high school nostalgia is going well with Melissa\'s latest CD, which is turning out to be quite excellent.  It matches the bright, fresh, young feeling of today.  The last, and first, Melissa album I devoured immediately upon receiving it was (yes, it\'s a terrible cliche) Yes, I Am.  My friend Sami lent it to me as we traveled on a field trip to Boise.  It didn\'t leave my vicinity for two years.  Every damn line of that album related  completely  to my place in the world, matched my life completely, as only a 17 year old in love can construe them.  Ever since, Melissa\'s had a special place on my CD shelf, more so than the obligatory lesbian-collection-favorites.  So it\'s nice to have a new album that\'s easy to get into and enjoy (as opposed to Breakdown or Skin; it\'s over Melissa!  DUMP HER!) without apology.    An anniversary has been tickling the back of my mind for a couple of weeks; and it struck me as I was perusing this month\'s Premiere magazine that another important anniversary has come to pass without recognition.  They work well in balance.    I\'m not looking forward to Mardi Gras.  Last year, Carrie and planned a Saturday morning breakfast; I was going to return and write a wonderful story about a boy and a girl to make the Creative Nonfiction class cry.  Walked out to the parking lot to find Harley\'s side-view mirrors in pieces, her windshield in shards.    \"It\'s finally happened.\"   Not that I\'ll ever know for sure, of course, whether my stickers precipitated the incident; but I\'ll always wonder.  It\'s not that huge a deal, really, and I\'m not exactly sure why the anniversary has been rolling around in my mind as much as it has.  Find myself holding my breath in the morning as I make my way out, hoping.  Then feel over-dramatic and paranoid when everything\'s fine.    The third anniversary of a favorite weekend works well to bring the smile back.    The plans were laid more than a month in advance, during the first epic Christmas break, over the pool table.  I would have Fridays off the spring semester of 2001, and we really needed to see this movie together.  There was no other way.  A trip to Bozeman was in order.  And so it was, that said boy and girl, the weekend of the 9th saw Hannibal not once, not twice, but three times.  The first two viewings within a few hours of each other.  And we haven\'t been the same since.  Admittedly, the movie has lost a bit of its luster since, but the Bible we found waiting under my car hasn\'t.  Those few days have come to symbolize an era, of sorts.  The last moment before everything changes completely.  The deep breath before the plunge.  The pause at the end of an exhalation.  Of course, we had no idea at the time; we just knew we were young and unstoppable, together again, and summer was around the corner.    But anyway...  Today\'s just one of those days.   '),(23279,2314011,'10,February,2004','2004-02-10 00:00:00','    Okay, so winter\'s rallying again, but its grip is still gone.  The frost fought back viciously this morning, but the coat remained unbuttoned.  Spring will find a way!  Like life!  In other news, the Idaho House of Representatives is voting in 15 minutes on whether to ban gay marriage via a constitutional amendment.  Sent emails to 6 of those reps telling them how much I\'d appreciate it if they voted against it.  Not much, but maybe just enough (gasp!)  And now on to Civil Procedure reading about fee shifting and settlements.  Yay.   '),(23280,2314011,'10,February,2004','2004-02-10 00:00:00','    Spring is coming.  It could just be my reaction to seeing the clear sky for the first time in many weeks, but there\'s a subtle, but distinct change in the air today.  Don\'t have to bundle up quite as snug; the frost was still on the windows, but mostly because it didn\'t have anywhere else to be, it gave up easily.  There\'s nothing quite like the transition between winter and spring.  Fall is great and we love it; spring into summer is a party in itself, but winter into spring is unequaled.  There\'s a marked change in most everything when spring is on its way; it\'s very sudden.  One day is decidedly winter, the next, its grip is gone.  Not just slipping or easing; it\'s gone, and we wait for spring to arrive.   '),(23281,2314011,'09,February,2004','2004-02-09 00:00:00','    Alrighty...the tweaking is finished.    The links aren\'t quite right; I\'d rather they were in the sexy little box to the right, but whatever.  All else is fabulous, and the links are still pretty damn cool, really.  And the original bloggy is still lurking around the archives (gasp!).   Had an action-packed night completely devoid of homework.  Well, except for the two pages of Civil Procedure.  Cashed in a dinner Joey owed me and we spent a nice chunk of the evening on the living room floor with our roasted chicken, macaroni salad and dinner rolls from Safeway watching The Spy Who Shagged Me.    Good times.   '),(23282,2314011,'08,February,2004','2004-02-08 00:00:00','    I mean this day.  A few minutes from my last post.  Welcome to the new blog.  Hope it\'s easier to read and a bit more stylish than the original.  Not that the original will ever be topped, mind you.  Or forgotten!  Tear!  Okay, I really AM going home now...   '),(23283,2314011,'08,February,2004','2004-02-08 00:00:00','    No Civil Procedure done.  Instead, I thought I\'d try altering the template on my blog.  I tried a brand new template.  Only to find my comments erased, along with my fancy links, and my text in a new, uncaring, unfamiliar environment.  I hated it.  So, I made my peace with the wonderful comments, lost forever, and set out to make things right by dedicating this (new) blog\'s life to returning things back to the way they were.  Added the links again, matched the background and text colors as best I could, and finally came to the comments.  They came right back after a few copy and pastings.  YAY!!  They weren\'t lost or forgotten!    Now that I know those are easily recoverable and have a better working knowledge of the color situation, I might attempt another template change one day.  But it is not this day...   '),(23284,2314011,'08,February,2004','2004-02-08 00:00:00','    Tomorrow\'s in the can; it\'s 345 and I could go home.  But will I?  Probably not.  Think Civil Procedure will get a jump start.  ...just to keep you updated...   '),(23285,2314011,'08,February,2004','2004-02-08 00:00:00','    I\'m vacillating between enjoying the gooey relaxing Monday the universe has given me on the heels of the lusciously empty Sunday afternoon of yesterday and being wary of the simple gratuity.    \"What am I forgetting?  What huge reading assignment did I miss?  Should I review the reading I did last week that we didn\'t get to?\"   It\'s always a bit weird coming back into the space that preempted a minor breakdown, too.  But I\'m keeping the thought of \"what shall I do when I get home?\" at bay by reading for tomorrow, drawing it out by blogging and surfing the net during my lunch breaky.  I\'m sure something will occur to me by the end of the school day.    But said lusciously empty afternoon was sooo great, I have to say.  Rounded out the paper by 330, leaving nothing, I repeat Nothing, to be done for the rest of the day.  To be sure, a few moments of guilt-induced \"Maybe I should give that paper another look over.  Or I could read those 4 new cases I printed out to see if they have a place in it somewhere.\" thoughts crept up; I burrowed deeper into my over-sized pillow and tucked the new novel a bit tighter to my chest.  The book\'s cadence really picks up when it actually gets read for a couple hours!  Yay!    But the book I couldn\'t put down once I started was A Civil Action.  Decided to give it and Ghostwritten equal time, since I had it in spades, and found I couldn\'t get enough of the legal thriller!  It\'s sick, I realize.  Even had a few moments of almost-panic reading about the heroic young attorney unable to find his way: \"Oh God, don\'t let that happen to me!\" and the sharky older trial veteran training his minions: \"Oh God, I can\'t deal with those people! Do not ask it of me!\"  It was so intense.    And in completely unrelated news:     Anybody who is anybody is part of a  urlLink  photo exhibition opening this week in London.    THE London???  On the centenary of his birth, the National Portrait Gallery  WE LOVE THE NATIONAL PORTRAIT GALLERY!!!  is showing nearly 200 photographs by Cecil Beaton, the gay man who for much of the 20th century was the epitome of sex and style on both sides of the Atlantic.     GO SEX AND STYLE!  GO ATLANTIC!  BOTH SIDES!   '),(23286,2314011,'08,February,2004','2004-02-08 00:00:00','    I\'m posting, not because there\'s anything to say, but because I don\'t know how long this connection will last, nor when the next reliable connection will be available.  (other than the kick ass wi-fi at school, that is.)    But I\'ve just discovered it\'s not just my connection; it\'s AOL in general.  Hmmm.  Just when I\'ve made up my mind to switch completely, it works wonderfully again.  Guess I\'ll just ride it out until it irks me for the last time.    Other than that, today\'s been really good.  Really relaxed, but really productive.  Have the rough draft all printed out and ready to be ravished by the red pen.  Reading for tomorrow\'s completed; laundry folded nicely in the drawers.  And I still have the whole afternoon to do with it what I wish!  MY GOD!  Have to send a shoutout to Carrie in Fuckin\'-Boston.  She hasn\'t gotten much blog-attention lately, and she really deserves it.  So there you go, girlie, you\'re very own shoutout.  I remember my first; it stays with you forever.   '),(23287,2314011,'07,February,2004','2004-02-07 00:00:00','    Okay, so I\'ve been reduced to writing this in Word Perfect because AOL is gay again tonight.  It probably won\'t get posted until tomorrow morning when the novelty of tonight has worn away and all the great lines forgotten. (She wrote as she posted on a tenuous connection.)  Law school parties are fun.  Who knew?  I could have spent a nice, quiet, low-key evening at home with my Indian (as in the continent, not the conquered natives) take-out, laundry, yoga and new book and been perfectly content.  Instead, I risked a night out.  Three weekends in a row!  Someone get the hose!    I was sure I\'d arrive at precisely the wrong time.  It was a surprise party after all, set to go off at 845.  Cruising out of the parking lot at 839, I wondered if I should circle the vicinity for a bit.  Again, I risked it, and was again rewarded.  9 came and went with nary a birthday girl to be found.  She arrived soon after, though, genuinely surprised, wearing a tiara with not a single hair out of place.  That takes skill, let me tell ya!    And there was much rejoicing.  Coups were planned, platitude-filled speeches rehearsed, hook-ups promised, shopping tips shared.  Drinking, fucking and heroine were conversation high points.  And somewhere in between the spilled beer, stories about \"This guy I know\" and lame Laughy-Taffy jokes, we all became a bit closer than we were in class Friday...gasp!  We will absolve ourselves of all knowledge of these activities by Monday, to be sure.  But for now, we have each other!  And it...it feels good!   '),(23288,2314011,'07,February,2004','2004-02-07 00:00:00','    It\'s amazing how many interesting things are done by people who have no idea what they\'re doing.   That Neil always knows just what to say!  Let\'s see how right he is, as I set off on a day of writing a motion to suppress tangible evidence.  Away!   '),(23289,2314011,'06,February,2004','2004-02-06 00:00:00','    Yay for the Internet!  AOL\'s been acting up the last couple of days and I thought I\'d have to turn my back on it forever, but oh no, the ever-so-helpful technical service staff at the corporate monster fixed it up all purty and new.  That really was the breaking point last night, being unable to sign on and surf the net, check in with my favorite bloggies.    So the day ended pretty well.  Checked out the \"Human Rights store\" I\'ve been seeing flyers for and came to find out about a big peace rally protesting the local ultra-conservative-neo-confederate pastor\'s \"History\" conference at the SUB.  The guy wrote a pamphlet discussing how harmonious southern slavery was.  So, the local diversity-loving hippies planned a get together across the street in the Catholic church; they love a good fight!  Have to say, a rally against hate sounded like just the thing I was looking for, but my night was already booked with a campaign strategy meeting.  Ever the loyalist, I decided to stick to my commitment and headed home after buying some bed-wetting liberal propaganda.    As I was cleaning the apartment, which is always a good stress reliever, I couldn\'t escape the thought that I should be going to this rally/march instead of the Garden.  \"Do these people even  know  there\'s a march happening?  Maybe we can catch the tail end of it after the meeting.  Student Bar Association presidential campaigns are important too, after all.\"  When the phone rang, I knew what the call was about and was relieved before I even answered.  No meeting.  All rally.    And it was pretty damn good.  Those Catholics know how to throw a party!  The head of the Seattle Chapter of the NAACP heralded a great calvary charge.  Then we battled the blizzardy night up to the Administration Building to listen to a journalist-turned-Southern Poverty Law Center-publication guy recount the history of neo-confederates and how their messages so closely, clearly, painstakingly match said local pastor\'s.  We needed our NAACP guy back.  Desperately.  Skinny white journalists don\'t, can\'t and shouldn\'t compete with emboldened African-American preachers.  Two hours and years of boredom later, I slunk from the auditorium.    Now that I have my precious AOL back, everything seems complete.  And tomorrow\'s writing assignment isn\'t looming as large as it has the last few days.  I\'m sure there will be blood and tears running down the hands of innocents before it\'s all over, but for now, there\'s a fragile peace, one might even say...(gasp!)..balance!   '),(23290,2314011,'05,February,2004','2004-02-05 00:00:00','    Leave it to me to have a mini-breakdown and semi-sleepless night the eve of an easy school day, promising a solid 2 hours of sleep-in time.  Nothing really new to report.  The same, obligatory overwhelming wave of realization about just what I\'ve gotten myself into, resisting the duty to act accordingly and all the while having a sneaking suspicion I\'m not doing enough.  What in the goddamn hell am I going to do with myself?  Signing up for diversity certification training in April helped a bit.  And writing the motion to suppress tangible evidence this weekend will help substantially.  Productivity is our friend.  Just trying to figure out what this bout of doubt/vulnerability is trying to teach me.  Typing notes from class is helping, too.  Can\'t seem to escape the fear/worry that I\'m wasting time, opportunities, not applying myself fully, etc.  But at the same time, I realize that anything I\'m supposed to do/find will never let me pass it by, no matter how much I may try to ignore it.  Sigh.  This seems to be my favorite struggle.   '),(23291,2314011,'05,February,2004','2004-02-05 00:00:00','    I\'ve resigned myself to the fact that this semester is not going to settle down.  This week has been about as normal as one could hope for, and one\'s still jogging a bit to keep up.  The balancing technique applied this week has been foregoing the easy reading assignments until the hours before class.  It\'s worked thus far, but I\'m not sure it has potential for long-term success.    In spite of that little hang-up, though, things are going remarkably well.  Not sure what it is about the new year/semester, but there\'s a tinge of (dare I say it?) purpose? challenge? that\'s lurking behind every case, brief, outline, note, lecture, always prodding, whispering.  This whole experiment called law school always carried with it the question: \"Can I actually pull this off?\"  That question led to a few sleepless nights and more than a few frantic late night/early morning pep-talks.  And much of last semester was about answering that question once and for all.  Now that the answer is clearly \"yes,\" it seems to be time to see how far this little experiment can go.    As cheesy as it sounds, No Doubt got it right in the line \"I ask myself: \'what good do you do?\'.\"  The answer to that has always been an adventure and a half, even without law school\'s input.  It feels now, though, like I have a bit more leverage in the answer; always remembering that with great power comes great responsibility, of course.    So yes, this semester is going to be a mad-crazy, wild ride, only to be followed by ever more challenging ones that make this look smooth.  But all I have to decide is what to do with the time that\'s given to me, which might include using fewer pop-culture references in my entries.  And it might not.   '),(23292,2314011,'04,February,2004','2004-02-04 00:00:00','    Supreme Courts rock.    Especially the ones upholding the rights to have sex and get married.   '),(23293,2314011,'03,February,2004','2004-02-03 00:00:00','    Ever have those days when you can\'t get enough of a particular artist\'s voice?  And not just  any  particular artist\'s voice, but one of your all-time, all-lives favorites; one of which you hear every day, but there\'s just something about  this  one voice,  this day that you just can\'t get over?  And because you\'ve heard this voice and six of her songs twice, everything else sounds infinitely better and the songs your computer juke-box plays are divinely selected and ordered.  Then the edges of your vision get blurry and ethereal, like a dream sequence in a soap opera, and nothing seems impossible, except sitting still and being calm.  Even Civil Procedure is interesting!!  Well, okay, it\'s not, but the 20 pages of reading you didn\'t get done last night doesn\'t seem quite as insurmountable.  Maybe all of this has something to do with the handstands you did last night.  Party!  Let\'s do some right now!  At the chalkboard!   '),(23294,2314011,'03,February,2004','2004-02-03 00:00:00','    I just have to say how proud I am of my Boy.  Probably don\'t let him know that enough, so I figured this would be a good way to make up for it.  He produces artwork at a stunningly high quality level that his professors (and soon the world\'s masses) never fail to salivate on each due date; he gives everyone around him more than they could ever expect without thinking twice while wondering if he\'s done enough; he watches over the residents in his care, keeps exhausting crazy-makers at bay without causing them pain or embarrassment, works until the wee hours of the morning, creates mixes that make the world cry and STILL has time to listen to me bitch about how rough I have it.  Love you, hon; I want to be like you when I grow up.   '),(23295,2314011,'02,February,2004','2004-02-02 00:00:00','    I don\'t care what that Chinese Medicine expert says!  She\'s a Westerner anyway!  What the hell can she possibly know about Chinese Medicine?  Oh sure, she\'s \"licensed\" and \"apprenticed\" with a master, but really, what kind of standards do these people use to gauge these things?  Not very high ones, I\'d surmise, from their stature.  Stay away from coffee and cold food.  They\'ll only bring you bad things.  Whatever.  I refuse to believe every morning of my life has been detrimental to my overall health because of my ritual bowl of cereal.  Hot oatmeal is too much damn work and tea is great for evening relaxing time, but I may as well be drinking warm air for morning purposes.  Gimmie my overpriced, capitalist-pig supporting Starbucks brew, baby!  I\'m avoiding typing up my Contracts notes.  Nothing too difficult, but I\'m just not in the mood.  The Women\'s Center didn\'t provide the quiet haven I was hoping for.  Tried to read over there yesterday, but ended up using the same noise-avoidance tactics I use here.  I forgot how chatty women can be.  Add a few Asians-at-heart and you\'re totally screwed!  But I did get to show off how deliriously happy I am with law school to ex-professor Debbie, which is always good for self-empowerment.    \"So you think this is what you want to do, huh?\" she asked, not trying to mask her skepticism.  \"Oh definitely; that took about a week to decide.\"    \"Well, it\'ll be nice to have a lawyer on our side.\"   \"That\'s precisely the idea.\"   And I know how hard that was for you to admit, my poor Debbie.     \"Just don\'t follow Darth Vader,\" John, her husband, piped in quietly.  \"Don\'t go over to the dark side.\"  \"Never; I\'m Luke Skywalker, all the way.  Or maybe Princess Leia, only without the earmuffs.\"     And am forever in search of my Yoda!  Tear!   '),(23296,2314011,'02,February,2004','2004-02-02 00:00:00','    So this is the first day of the fourth week of school.  We\'ve had 9 Contracts classes; my typed notes for the class have reached 14 pages.  We\'re allowed a 10 page outline for the final.    Sometimes all you (and by \"you\", I mean \"me\") can do is laugh, shake your head and start reading the next two cases for tomorrow\'s class.   '),(23297,2314011,'01,February,2004','2004-02-01 00:00:00','    Off writing shit like this on his blog, that\'s where!   You know, in a world in which Bush and Blair can be nominated for the Nobel Peace Prize, \"for having dared to take the necessary decision to launch a war on Iraq without having the support of the UN\" I find myself agreeing with Tom Lehrer: satire is dead.   And making sure we know about the  urlLink Bucket Full O\' Kittens .  There\'s something for everyone!  WooHoo!   '),(23298,2314011,'01,February,2004','2004-02-01 00:00:00','    ...the loss of morning pages can be successfully mitigated by being productive (even though completion comes later than it should have) and blogging a bit.   '),(23299,2314011,'01,February,2004','2004-02-01 00:00:00','    Skipping morning pages sucks.  I should know better by now, really.  Yesterday wasn\'t bad; I guess it\'s okay to skip them if you\'re hungover.  Thought it would be okay for today, too, especially since I didn\'t get out of bed until 11 and wanted to get school stuff taken care of to hit the 430 show downtown.  Instead, I watched TV until noon and arrived here half an hour later.    It\'s been 3 hours and all I\'ve got accomplished is typing notes from Friday\'s classes and reading one Property case.  Not sure I even can remember what I\'ve read.  The movie\'s been pushed back to 7, so now I have more time to get through at least 2 (should be 3) Contracts cases, get home, crank out a couple of late-afternoon-early-evening pages and manage to salvage the remaining hours before I sleep.    Skipping morning pages sucks.   '),(23300,2314011,'31,March,2004','2004-03-31 00:00:00','    Seeing as how I don\'t have any brilliant or otherwise prose with which to describe the events of the last couple of days, I\'ll just list them, m\'kay?    Finally eradicated the sneaky Trojan viruses from my computer, thanks to the law school\'s very own Computer Guy.  My search bar is back to normal and no pop-ups as far as the eye can see.  Sweeet.  Saw the cute, hot-in-a-let\'s-go-home-and-cuddle-way, gas station cashier at lunch with the gals yesterday.  She was with a large, burly manly man, and somehow didn\'t seem quite as hot.  Maybe the world looked a bit sweeter Sunday evening when I asked her for change as the fog of the Appellate Brief lifted.    And THEN!  (In the very same day!)  A fellow yoga-ite seemed especially friendly before my class last night, which follows hers.  We bonded over our mutual inability to focus on anything other than what we each have to do for the rest of the night during our deep relaxation, meditative time, she as a biology grad student, me as a brilliant law student.  \"What is it with me and all these attorneys in my life?\"  she asked, genuinely looking for an answer.  Apparently, a good number of her circle of friends from high school ended up being attorneys and now here I am!  Suddenly I\'m \"in her life\"??...interesting.  And she managed to throw in that she\'s going to be visiting one such friend in San Francisco at the end of the month as she bore her mid-drift tying her yoga pants.  Interesting.  And there was more than one \"have a good night/good luck with all your stuff!\", as though she wasn\'t ready to leave just yet, but had no choice because I was headed to class.  Interesting.  Or all figments and exaggerations of my imagination.  Became Vice President of SODA (Sexual Orientation Diversity Alliance) today, finally fulfilling my year-long dream to be beneath Miss Q.  The fact that Dan\'s our Secretary/Bitch is a bonus.   '),(23301,2314011,'29,March,2004','2004-03-29 00:00:00','    If I didn\'t know better, I\'d think it was May.  Between the weather in the 70s, freedom begging to be played with and taken advantage of after turning in the big brief, and cute gas station cashiers giving me more than enough to day dream about, it would be easy to think we\'re done, finished, ready for three months of summer\'s glory.  But May can be a dangerous thing.  Especially when it\'s not yet April.  And so, I find myself somewhere in between:  nestled in the Grind, near the door, enjoying the warmth, getting ready to dig into Civil Procedure, sipping the year\'s first iced mocha.   '),(23302,2314011,'27,March,2004','2004-03-27 00:00:00','    Just revised, edited, re-wrote, cut and begged my appellate brief down to 17 pages.  Only took 8 hours.  Still have to look up and possibly (better fucking not have to) add the subsequent history of my cited cases.  Plus, the table of contents and lists of cases and statutes cited should be typed up at some point.  But not tonight.  For now, I\'m on the prowl for groceries and dinner.  Woo!  Look out Saturday night!  P.S.  To my fellow briefers:  is it okay that I gave bare-bones explainations of two cases under by B. section of the second issue while using way more than that for rule statements?  Your comments (answers) won\'t count as discussing the paper after the black-out period, I checked (while wearing a zipper).   '),(23303,2314011,'26,March,2004','2004-03-26 00:00:00','    Just realized I can\'t stay at school any longer to finish the Civ Pro back-log, so I\'ll be taking the log home instead.  But, that means I can do a blog entry while I\'m here instead.  Yay!  I must tell the story of the shelf-peg.  It\'s classic.    Went to my cupboard earlier in the week, at the end of a long, tiring day, for some soothing tea to find my middle shelf slumped back.  Thinking it time to flip the shelf over to counter-act the apparent warp, I unloaded the shelf and examined the pegs holding it up.  Ah-Ha!  One of said plastic pegs had given up the fight and defected to the bottom of the cupboard, leaving the peg-part of the peg nestled nicely in its hole-home.  Not in any position to devote any energy to any particular reaction, I instead replaced the board on the three remaining pegs and carefully placed light items upon it.  Tea bags here, graham crackers there.  Soup cans and microwave popcorn relocated to the upper and lower shelves.    Made to Moscow Building Supply a couple days later after a light day of class and before the 80+ pages of Civil Procedure project was to be tackled.  I was sure to bring the peg at issue along to be sure I purchased the right size.  Only when I\'m faced with a myriad of plastic pegs of various sizes, no other time, do I realize the one section of the peg I need to gauge the size to be purchased is the one in the hole back home.  Still refusing to give in to any emotional reaction, I calmly pick from the range of sizes I feel pretty certain will fit.  4 of this size (because I may as well prepare for the others to give way, right?), 4 of that and 2 of the more expensive ones.  They\'re too big, anyway, so I\'m not going to buy 4 I know I\'m not going to use.  Better to get 8 that will probably work.  Sure enough, the 8 were too small.  And the two too big ones fit perfectly.  Again, not even going to worry about it.  I can take them back and recoup my $2.32.  Not a big deal.  Now I just have to set about getting this remainder-peg out of this hole.    It should be noted at this juncture that I am a very bad lesbian.  I need to go through a remedial course on basic lesbian-isms.  Just as an elementary but very illustrative example:  when my toilet clogs, I call the landlord and have them send a plumber.  Oh, I  say  I\'ve tried plunging, but I haven\'t.  I think I  could  check my oil if it came to that, but I\'ve never tested that hypothesis.  The only tool I have in my apartment is a hammer.  No nails, no screws, no screwdrivers.  Just a hammer.  So, with this background, let\'s proceed.  I head to the peg with the nail clipper, file at the ready.  It\'s small; it\'s thin; it\'s all I have that will work in this situation.  And so, I begin picking, pulling, scraping, pushing at this angle and that, to (shockingly enough) little avail.  Soon, I employ the steak knife\'s help, merely as a stronger force with which to bed this mere stump of plastic to my will.  Then, I realize:  the clipper might be used as a clamp-device and I can perhaps  pull   the stump out.  Indeed, such a plot would work, were it not for the sharp edges of the clipper.  It does pull the material quite a ways forward, but inevitably cuts through the weak plastic, leaving me in no better position.  After more than a few minutes of rearranging myself on the countertop and continued benign attempts at finding the right combination of leverage, pressure and pulling actions, I begin realizing I may have to explore other shelf-placement options.  Could have it a bit lower or a bit higher.  Stopped a couple of different times to seriously think about these options and always come to the same conclusion:  I want the middle.    Eventually, I stop to eat.  It\'s been five hours since my last meal and roughly one hour since this adventure began.  Hands are starting to shake and I\'m beginning to take the whole thing personally.  The bagel and apple helped and I began round two anew.  With similar results.  Decide to call Joe to see if he has any tweezers (because I don\'t even have any fucking tweezers).  He has none, but suggests a corkscrew.  Why not?  Sure, it\'s a bit large for the area at issue, but who the hell cares at this point?  It was working at breaking the stump into a more workable slab of plastic, but the scraps of surrounding wood-work it was taking out of the cupboard panel made me nervous.  Yes, the area  around   the hole seems to be getting gradually larger, while the hole itself remains fit snugly around the peg.  The knife makes a few more appearances and the clipper is thrown down in frustration thrice, each time with the intention of never picking it up again because, really, it\'s time to just figure out another level for this shelf.  But as soon as I start thinking about it seriously, the more I refuse to let this fucker win.  If I have to cut it down, bit by bit, with the clipper, so be it.  If it takes all night, so be it.  And with cold, lonely resolution, I attack again.    During a typical, identical-to-all-its-predecessors, round, holding my breath to prevent the clipper from cutting all the way through the plastic (because that helps) as I pulled it forward, the stump popped out.  And few things, accomplishments, victories, etc. have felt sweeter.    I celebrated by washing the honey stains off the shelf and making completely unnecessary trips into the cupboard for the rest of the night.    Brought Sadie over for lunch today just to show her.    And now, I will run to the Building Supply store to return the unused pegs, go home and read Civil Procedure.  Happy Friday to me...and to you.   '),(23304,2314011,'24,March,2004','2004-03-24 00:00:00','    Reluctantly admitted defeat of my goal for the evening of conquering 80+ pages of Civil Procedure catch-up reading after approximately 40-50ish.  Having one of those \"Everything must get done now\" weeks.  It started out seeming entirely do-able; the week\'s beginning, that is.  There was no pretense or easing back into the routine after the break, which was more work than break to begin with.  And suddenly, this week has completely escaped my grasp, leaving me scrambling, trying to mitigate my damages as best I can.  Have to keep reminding myself that all these tasks  will  get done.  If not in the next five hours, in the next five days at the latest.  And that\'s just going to have to be okay.  The fact that next week heralds the last 5 before finals begin in May is looming, albeit quietly, in the psyche, as well.  Still have two subject to outline, one of which is the Mother of All Classes: Contracts.  The typed notes for that class reached 46 pages today.  Yay.  This weekend is dedicated to the final draft of the brief; if it gets done Saturday, I\'ll be shocked.  I\'ve tried giving up gauging the revision process, and the result has been over-estimating.  Between revising the rough draft and working on the front-matter (i.e. table of contents, statement of facts, etc.), it will most likely take up a good chunk of time.    So, with all of this in mind, I figured tonight would be a good time to get cracking on the material our Civil Procedure professor \"covered\" in our 2.5 hour class Tuesday.  I read the 6 Rules of Civil Procedure assigned, choosing to forego the corresponding cases, thinking they would be covered in painstaking detail as he has a semi-tendency to do.  Actually, I take that back, he  does  have a tendency to cover material in painstaking detail when it\'s completely self-explanatory.  When it\'s complicated, intricate, lengthy, etc., however, he skims the surface.  And thus, not only were the 6 Rules of Civil Procedure that I prepared for class covered, but so were their cases, as well as 2 more Rules and their cases.  Hence the 80+ pages of reading attempted tonight.  It was folly from the beginning, part of me knew, but another part (the same that thought we might not cover all the material Tuesday) thought it might not be too terribly monstrous.  As per this week\'s developments, that part was wrong.    And so, with the advice of my favorite lesbian sage (she\'s been on a roll this week), I stopped studying at 830; choosing instead to spend the few hours left doing the things I\'ve been craving for many days:  yoga, blogging and, perhaps, reading about London.  Ranting about all this has helped the cause, too.  But the planner deep inside is already scoping out Friday afternoon for the rest of the make-up readings.  I just hate the feeling of being behind; and nothing is right until I\'m caught up.  But I absolutely refuse to miss yoga class tomorrow night, and if it\'s going to make me choose, Civil Procedure will lose.    Overall, though, I\'m managing to keep my \"at peace with all things\" outlook on life pretty much in tact.  And that\'s a good thing.  The Media Board, exhausting though it was, went well.  Think we chose a good group of kids to run the university\'s media establishment for the next year.  Some positions were sought more...intensely...than others, but, thankfully, the right person for each job clearly emerged.    I was very relieved it wasn\'t necessary to give a character reference to an applicant with whom I have a bit of \"history.\"  Just the usual high school-college drama crossover, of which full-disclosure is not necessary.  But partial-disclosure is.  She decided we were arch-enemies in high school, but since we had Intro to Psych together freshman year of college, we could be friends.  Since I had nothing against her in high school and was fairly interested to see where this new \"friendship\" would lead, I didn\'t object.  Eventually, it helped bring Levi and I back into close contact and for that, I am thankful, whether she\'ll ever know it or not.  But, when her relationship with the Boy dissipated for the last time, I was gone from her life, as well.  And, as before, I was neither hurt nor surprised.  In fact, I took a bit of pleasure (and still do on dark nights) in the fact that he\'d finally chosen the right girl.  Every time I would see her on campus while the friendship with the Boy became what it is, I couldn\'t help but wonder how/what her friend/relationships had become and if she had any idea what had become of the two people she banished from her life.  And so, seeing her name on the list of applicants for the university\'s new magazine (the yearbook\'s new form), a part of me smirked and a part of me sighed.  I knew already she wasn\'t qualified, speaking strictly of my first-hand knowledge of her professional and scholastic abilities.  She\'s completely capable, to be sure, she just has no motivation nor ability to commit, both of which are *major* requirements for a project of this magnitude.  But, I resolved to put this and more baggage to the side and assume the consummate-professional hat for the evening.  And, as I referenced above, I needn\'t have worried.  Her references, sadly but not surprisingly, spoke for themselves, and I remained blissfully silent on the matter.    As my turn approached to ask her a question from the prepared list, however, one sentence would not leave my mind.  The statement\'s been on the tip of my tongue for more than two years now and it needs only the slightest opening to be uttered, an opening that\'s never actually been given, though.  And, no doubt, she\'s already heard the news, but my fantasy of being the first to tell her will probably never dissipate.  And so, while she quietly answered the question before mine, I struggled to banish this vision from my mind:  \"Hi Laura, I\'m Andrea, Student Bar Association Representative.  I have a prepared question to ask you, but first, I just have to let you know:  Levi is gay.\"  And after hers and the entire Board\'s eyes fell out of their heads, I continue:  \"That\'s right.  Gay.  We\'re gay together.  And you can\'t stop us.  You had him; thought he\'d be the One, forever, for all times in all worlds.  But no.  He\'s mine.  In ways you could never understand.  And plus, Tori is ours too.  She used to be yours; thought she\'d be the one, forever, for all times in all worlds.  But no ...\" and so on.    But instead, I gathered my wits, wiped a guilty grin off my face, met her eye-to-eye and said:  \"Hey Laura.  You know me...\" watched her saccharine smile widen into one that just might have been genuine and asked my stupid question.    Maturity and growth are our friends, but slimy-mean visions can be great to cat around with sometimes.   '),(23305,2314011,'23,March,2004','2004-03-23 00:00:00','    Haven\'t blogged for 3 days and I\'m going through withdrawals.  It\'s to the point where, now that I have a tiny slice of time in which to blog, I can\'t form anything coherent to blog about.  The week has been a big one, already, in just two days.  We\'re (the Media Board) interviewing for the top positions in Student Media this evening, which is exciting, for professional as well as catty, immature, juvenile reasons (to be explained at a later date, when the results of tonight\'s meeting are known).  Just wish I didn\'t have to miss the first class of the new session of yoga.  Oh well, at least there will be pizza and carbonated beverages available.    Suppose I should get on to typing up today\'s class notes, as any slipping behind in such vital duties would be quite detrimental at this stage in the game, what with the semester rapidly coming to a close, outlines needing completion and professors blasting through material faster than students can read it.    Wonder if the London segment of the current novel will ever have its day.  Probably in June.   '),(23306,2314011,'20,March,2004','2004-03-20 00:00:00','    And I\'m spent.    The evening seems so easy to plan, organize, and \"accomplish\" when you\'re in hour 4 of a 9-hour drive.  The miles were zipping by and I was hefting tasks to an already sizeable list.  Other than unpacking and necessity-grocery shopping that is.  I would email this person for the first time in months, call that person, construct my new CD tower, reorganize my bookshelves, balance my checkbook, blog about the day\'s drive and its wonders, do a solid round of yoga and read the 2nd novel of the year (god, I hate that) until I fell comfortably and contentedly into bed.    Whatever.  Actually, the biggest tasks are complete.  The new tower is dead sexy and the books are breathing a bit easier.  The checkbook and long lost people will have to wait until tomorrow.  The only things I plan on reading is this month\'s Premiere and a free Rolling Stone.  And yoga might be of the deep-breathing variety.  At least I\'ve got the blog-task on its way.  Don\'t ask me why I couldn\'t leave all this shit until tomorrow, when I have an entire day to get it all done.  Something about coming home and making a huge mess really grates; plus, it\'s a nice work-out, running around like mad, after spending the day in a particular seated position.    The drive felt great.  It\'s been a while since the roads, construction and weather have been cooperative and it makes quite a difference.  And as much as I would love to recount every milestone (no pun intended, I swear), my futon and trashy magazines are calling louder than the muse.  So, here are the highlights:  Sharing the trip\'s single stop with David Jack.  Seeing baby sheep.    Hitting \"random\" on the changer just outside of Idaho Falls and hearing the first repeat crossing the Latah county line.  Trolling, utterly in vain, the Post Falls Outlet Mall for the store in which I secured my first pair of non-matching thumb rings.    Thinking about the drive to Seattle Pride in three months.     P.S.  The Ogden trip was a success.  Found a new dead sexy outfit perfect for the wedding and indulgent weekends alone.  Plus, a few other choice gems due out this season found their way into the bags.  But, as Mom and I made our way through the Ogden Mall in less than short order, we realized:  Ogden is dumb.  We should just go to Salt Lake.  It\'s now my responsibility to remind Mom of this fact next time we venture to our neighbor to the south.  For this time, though, we made do with Layton, another SLC suburb; it has a mall of its own (all 40 minutes of it) and a Sam\'s Club, just to illustrate how far we went to find things in our backyard.  But my new clothes are worth it.   '),(23307,2314011,'18,March,2004','2004-03-18 00:00:00','    Big sigh that lasts several minutes.  There.  Now my break can enter its \"Nothing left to do but be a lazy sod\" phase.  And it\'s only Friday at 6!  Finished the rough draft today in just under 6 hours and am feeling pretty damn sweet about it.  Not really thinking about the fact that it\'s 18 pages without the statement of facts or statement of proceedings and without a case my professor told me to use.  I\'ll worry about that next weekend after we\'ve received insightful feedback and constructive criticism on our efforts thus far.  Then, and only then, will I set about making the final draft comport with the 16 page limit.  For now, my efforts have come to an end, just in time for one last day of Break before heading back north Saturday.    So, Mom and I are heading to Ogden bright and early tomorrow to shop our asses off, which should make for a great send-off.  I\'m always surprised to find myself surprised when she has a bag of clothes waiting for me when I arrive.  But sure enough, she\'s been finding unbelievable bargains ever since I left the last time and has piled them in my room.  And tomorrow, we\'ll be searching (and no doubt finding) even more.  Chief among our duties is finding a nice summer frock for my (only cool) cousin\'s wedding in mid-June.  She wouldn\'t care if I showed up in leather pants, but for once, I\'m ready to find a nice, summer-y, not-too-formal-yet-not-too-casual \"something.\"  I have no idea what that means, but I\'m convinced I\'ll know it when I see it.  And if it has to be a tiny tissue of a dress, so be it.    But enough about tomorrow and the days to follow; I realize it\'s been more than a few days since my last post and this fact pains me.  It burrnsss usss!  Okay, not quite, but I do miss my little bloggy.  When last we blogged, the Boy and I had just finished enjoying our first day off in two months.  That evening found us enjoying chips, salsa and screaming monkeys with his hot-yet-empowered pal Bridget and her hot-in-a-cute way fiancee Dave.  We decided against the fried ice cream and opted for Dairy Queen Blizzards in the back of Bridget\'s Honda.  It was so hot.  The next day, we set out to get work done.  The more we sat around that morning and talked about other issues, the more we thought our work wouldn\'t really take  that  long.  After all, I was just outlining my Civil Procedure notes, and between the  Erie  doctrine and Rule 12, what more is there?  Well, as it turns out, about 6 hours\' worth, give or take a back-cracking stretch and whine for a Coke.  But, as the last of the rays descended along the horizons and the first stars made their appearance, we threw ourselves on the couch, triumphant in our accomplishments.    \"Wanna go get a Coke?\"  the Boy asked while flipping through Neil\'s latest comics.  \"Yea.\"  I answered, not lifting my head from its den in his couch\'s cushions.  \"Let\'s get some ice cream, too, and make floats while we watch more Queer as Folk.\"  And so it was.  We stayed up way passed our bedtimes, squealing at Brian\'s sweet dead sexiness, Ben\'s Superman physique and Clark Kent demeanor, and Emmett\'s fabulousness.  Then there was the preparation of Luscious Lorraine\'s Super Spring package.  It was painstaking and caused more than a few tears, but it was worth it....squeak!...it.was.worth.it.   Yea, all in all, the Break\'s been a pretty good one, and the incredible swiftness of time\'s passing hasn\'t had quite the impact I thought it would.  Maybe that\'s because being home feels different than it did during the undergrad breaks.  Christmas break certainly felt different than any of its predecessors, to the point that it didn\'t feel like a break at all.  Near the end, I had to remind myself that I would soon be returning to school, that it was time to start thinking about going back.  \"Back?  Back to  what ?  Aren\'t I finished?  Isn\'t this the way things  are  now?\"  It was the most difficult return back to Moscow, too.  So perhaps I\'ve discovered how to fully enjoy being home and all that entails (even with piles of work to do) to the point where the time doesn\'t evaporate before I even have a chance to acknowledge the fact that I\'m  on  break.  Whatever it is, it feels really great.  And as much as I\'m loving Spring, Summer can\'t get here fast enough.   '),(23308,2314011,'13,March,2004','2004-03-13 00:00:00','    Okay, soo much better now that we\'re on the good ol\' Earthlink dial-up connection.  Just got back from a good day bee-bopping around sunny, yet still chilled, Bozeman.  Spent some money we didn\'t have and added 3 new books to the pile of \"Love that I have it, but will probably never read it\" finds.  There\'s just something about new books.  Don\'t know what it is, but they\'re fucking dead sexy.  And there\'s just something about sitting in a coffee shop mauling the new purchases all over, coming home and rubbing them in hot love juices that seals the deal.  Then, years and many memories later, when they\'re pulled off the shelf, they get to be new all over again!    I honestly have no business bringing home new books at this point.  I\'ve been reading the same novel for 2 months and have barely cracked the 200 page mark.  Plus, I\'m running out of room for them.  I would be more justified bringing home abandoned baby kittens.  But for now, they\'re the newest, coolest books ever and there was much rejoicing.   Just chilling back at the pad for a bit before hitting the night life with the much-talked-about Bridget and her steady beau Dave.  Then we\'re back on the Boy\'s agenda:  movies, magazines and sex.  I would rather curl up with the new bookies, Civil Procedure and tea from Goldie, but I mean, whatever the Boy wants he gets.   '),(23309,2314011,'10,March,2004','2004-03-10 00:00:00','    Okay, so either Thomas Ulysses doesn\'t like me or the Mac version of Explorer is gay, but either way, the forces have conspired against my posting anything on here of value, so let me just say that I made it safe and sound....the drive was great and it seems that spring is waking up all along the Continental Divide.  And I\'ll try to hook my machine up to the Net soon so as to post a good post.  Until then....   '),(23310,2314011,'10,March,2004','2004-03-10 00:00:00','    Left the scarf at home this morning.  It\'s always a little sad when it happens, it\'s such a great little scarf and all.  And we seem to always have an obsession with our *firsts*, especially of the epic variety.    The sexy maroon number with the grey stripe on one end is certainly of that variety.  Leave it to me to travel many thousands of miles to buy my first item of truly British clothing at the Gap.  Not just any Gap, mind you, the one in the commercial heart of our dear old Great Grandmother.  No quaint old store on hidden backstreets shopping for us!  We couldn\'t even get away from huge commercial coffee shops for god\'s sake!  Regardless of its heritage (and maybe a bit because of it!), the scarf became an integral part of the rest of the London voyage and has remained such ever since.  I don\'t put it way during the warmer months.  It stays faithfully at the ready, even though it never gets used once the windows start staying open (which they are).    So yea, today marks the end of the little guy\'s reign for this year.  To be sure, he\'ll come with me for Spring Break because winter\'s still fighting like hell everywhere else in the state, but his duties are winding down.  I miss him when he\'s not around.  But driving to school with the window rolled down and squinting though the sun glasses work well to remind me exactly why he was left to his own devices.    Anyway...  Have a pretty light day today.  So, of course, I\'m going to fill it up with important duties.  Well, just one, really.  Going to write the first issue of the Appellate Brief this afternoon, after the big Candidate\'s Forum.  GO SUE AND ALYCIA!  Will have a nice, solid 5 hours to crank that sucker out, making it the most beautiful rough draft of an argument I\'ve ever seen!  It really shouldn\'t take that long to write it out, but I\'m determined to use the time allotted to the fullest, so if I can make a editorial pass or two, that\'s just fine by me.  Getting the Motion to Suppress assignment back yesterday provided the perfect motivation, too.  We love rolls, and seeing just how far we can push them.    Friday\'s rapidly approaching and the Mixes are anxious to hit the highway.   '),(23311,2314011,'08,March,2004','2004-03-08 00:00:00','    Can\'t think of a more fitting end to a 10-hour day, sure to be the first of more than a few others to follow, than a nice Intro-Level Yoga class and the purely idealistic thought that I might actually do more work upon my return.   '),(23312,2314011,'07,March,2004','2004-03-07 00:00:00','    And...can I just say how much I love the  urlLink Idaho Senate   (at least for now).   '),(23313,2314011,'07,March,2004','2004-03-07 00:00:00','    The weekend was a success.  Productivity all over the damn place.  The brief\'s outline is 8 pages and looking good.  Property outline is twice that long and even prettier.  And as I was printing out said brief outline last night, I realized it\'s not actually due until tomorrow.  For a moment, I was crushed.  But then I realized that I would not have gotten as much done during those two days had I allowed the extra day of work today.  So, as a result, I\'m ready to tackle this week on all fronts and chip away from the Break\'s work-load as much as possible.  Have the last 6 cases to read for the brief assignment all ready and waiting to be highlighted and possibly incorporated into the outline.  Oh yea.  So smooth.    And to top it all off, we\'re supposed to edge toward the 60s today.  Degrees, that is, not the decade; but that would kick ass, too.    And can I just say how much I love the   urlLink mayor of Seattle?   '),(23314,2314011,'07,March,2004','2004-03-07 00:00:00','    Happy Sunday, blogger!  You get to be the morning pages today, m\'kay?  M\'kay!  So the outline is done.  And I have to keep reminding myself of that.  The weekend\'s biggest assignment is finished, just needs to be printed off and that will be that.  Have all day to outline Property, which *shouldn\'t* take the whole day, like yesterday\'s little ditty.  The best part about being so busy with so many different things right before a break is discovering six new things waiting behind each task completed.  Because really, it\'s all about mitigating the damage done to Spring Break.  Sure, next week\'s assignment is done, but if I can do  more , say, have the first issue for the rough draft written before Friday, that means I\'ll only have to sacrifice one day out of the four I\'ll be home to work on the second issue.  So really, even though the productivity is high and the hours long (for a weekend), there nevertheless remains a huge pile of work to be done.    Sigh.  At least my blood\'s finally replenished itself.  And really, I\'m loving every second of the whole damn thing.  Even had my ritualistic Sunday-morning-coffee-web-surfing with the Vagina Mix...kick ass.   '),(23315,2314011,'05,March,2004','2004-03-05 00:00:00','    That\'s me.    And now I have the book to prove it.  Our esteemed Torts professor (looking every bit the 2nd grade school teacher this morning) raffled off the heartwarming story of the courageous teenage boy, who, after being gravely injured while using a farm machine, takes on its powerful manufacturer and wins, to the class members who donated blood yesterday.  \"What if you tried, but the flow stopped?\"  I asked from the pews.  \"Wellll...\"  she waffled.    \"I have the bruise to prove it!  I have a physical impact!\"  I argued effectively, using her own concepts against her.  \"Oh, if you have a bruise, you\'re in!\"  \"Damn right I am!\"  And wouldn\'t you know it, the one person in the pool who didn\'t  actually  give blood walked away with the prize.    Balance is our friend, especially when it\'s slanted substantially in your favor.   '),(23316,2314011,'04,March,2004','2004-03-04 00:00:00','    what\'s the deal with all the \'no comments\' lately, readers?  i don\'t do this for ME, you know.  a little feedback goes a long way, is all i\'m sayin\'...and a lot of feedback goes, well, it just goes.  (and is appreciated...squeak!)   '),(23317,2314011,'02,March,2004','2004-03-02 00:00:00','    i can\'t even donate blood!  they needed 475, but all i could squeeze out was 416.  such a loser.  so i got all the lightheadedness, cold rag, coddling by the cute nurse for nothing!  i was in the damn trailer for over an hour!  and now, while my class is busy learning about homicide in criminal law, i\'m down here the basement, withering away with my water, granola bars and \"i donated blood\" stickers wondering what it all means.    it means i don\'t have to work my ass of tonight, is what it means!!   '),(23318,2314011,'02,March,2004','2004-03-02 00:00:00','    It is quickly becoming apparent that there is no time to spare.  Sure, the Appellate Brief has loomed large since its distribution, and last weekend\'s surprise attack of the Issue Statement brought the scope of the assignment into sharper focus.  This week, though, has brought with it not only the confirmation of that realization, but also the knowing that preparations for finals will soon be in full swing.  Yes, only half-way through the semester and it\'s time to start preparing for finals.  So that whole: \"If I\'m not working on homework, the paper or outlining at each moment of the day (until about 8 or 9--I still demand an hour or two to myself), something\'s wrong\" thing has arrived a good month early.    But we got a bit of a break today with the canceling of Contracts this morning.  And instead of using the hour to read for tomorrow or more cases for the Brief, I moved tables for the blood drive, wrote notes on chalkboards (my first official duties as the new Women\'s Law Caucus Treasurer) and am now blogging.  Fine.  So I\'ll have an extra hour to myself today.  Sue me.   '),(23319,2314011,'02,March,2004','2004-03-02 00:00:00','    Alright, as promised, here\'s the hot action, as it unfolded, from Sunday:  Billy Crystal needs to stop.  Give me Steve Martin or Whoopi any year.    Supporting Actor, Tim Robbins: well- and long-deserved.  Maybe a nice concession for Mystic River.  Art Direction, LOTR:  good damn start!  Animated Feature, Finding Nemo:  Oooh!  Woo!!  Pick me!  Costume Design, LOTR: YAY!  Better not be a damn concession.  There were costumes in that movie?  Supporting Actress, Renee:  Yay!  Good actress, good role, good win; could her speech have been more rehearsed?  Joe:  \"You\'re writing all this down?  Next thing you know, you\'ll be keeping a journal!\"  Fran Walsh, you ARE the lion, the witch AND the wardrobe!  We love you Bob Hope!  Live Action Short, Two Soldiers:  And what a pair they were!  Animated Short, Harvey (something I didn\'t get):  Love you Harvey!  And love your hot creator thanking his hot boyfriend.  Visual Effects, LOTR:  Damn right.  Make Up, LOTR:  They better not be setting us up.  There was make-up in that movie?  Sound Mixing, LOTR:  Okay, this is looking really good; better follow through, Academy  Sound Editing, Master and Commander:  Is there a difference between Sound Mixing and Editing?  GO JULIA!  We love you Julia!  And your wispy, gaspy \"yeah\"s following a legend\'s legendary line.  And who came up with these lost dog Master Card commercials?  They\'re good shit!  We love you Oprah!  And your eye-shadow that matches your big bow-y dress.  Documentary short, (some mini-epic about the results of radiation exposure):  The housewife-director looked thrilled.  Stupid radiation, anyway.  Documentary feature, Fog of War:  It\'s the worst fog of them all!  Original Score, LOTR (Howard Shore):  We love you, Howard Shore!  Make love to us, Howard Shore!  Film Editing, LOTR:  Okay, really, this had better not be a damn ruse!  Original Song, Into the West:  Joe gets the line for this one:  \"Here comes the rain again.\"  Foreign Language Film, The Barbarian Invasions:  Canadians know (and do) them best.  Cinematography, Master and Commander:  It\'s a good thing there wasn\'t any cinematography in LOTR  Adapted Screenplay, LOTR:  It\'s so in the bag.  Original Screenplay, Sophia Coppola:  KICK ASS!  But you\'re not going to be the first American Woman Best Director tonight, sorry.  Hope this makes up for it.  Director, Peter Jackson:  GO PETE!!  Waddle your way up there!  Don\'t stop to comb your hair!  WHAT? NO OVATION?  Go Bill and Joan, good job.  Actress, Charlize Theron:  Yay.  Now go do something with it.  Actor, Sean Penn:  HE gets Peter\'s ovation?  What the fuck?!  We DO love you, Bill!  Go Hollywood Foreign Press.  Picture, LOTR:  It made a clean sweep; it\'s in the company of Titanic and Ben-Hur for most-won awards in a night and STILL no ovation.  Good thing we\'re not bitter.   '),(23320,2314011,'02,March,2004','2004-03-02 00:00:00','    This didn\'t really have a place in my last post, so it gets one of its very own.  I WILL post my blow-by-blow accounting of the Oscars soon, maybe even tonight.    But I\'ve amassed an interesting collection of quotes from the day (just today) that must be posted without any context whatsoever to set them up.  This is why I write down funny random quotes, my friends, because they always pay off later:  \"What is a cow?\"  \"Good.\"  \"Then we\'ll have the plaintiffs testifying about how they didn\'t do anything to get cancer; the Moms thought they were protecting their unborn children.  It\'s time to bring out the violins.\"  \"It was one of those \'see 12 countries in 2 days\' kind of things.\"  \"If we were all the same, without any differences in views or opinions, we would all be Russian.\"  \"Yoga can help that, you know.\"  \"Can it make my boobs stop growing?\"   '),(23321,2314011,'02,March,2004','2004-03-02 00:00:00','    Few things are less tiresome than typing up notes from class, but few things beat the feeling of utter completion and triumph after typing up notes from many days of many classes.  It is true, I have not recorded my notes from today\'s Civil Procedure class, but they\'re all nicely kept in the Federal Rules supplement section called RULE 12.  No, our professor could not just assign us it to read, he had to tell us in class what each section and subsection said.  (Sigh)  The book reports have reached a new level.    Anyway, my note-typing has been a bit lax of late, as I\'ve been distracted by the whether, my fabulous mood in general and yesterday, by this incessant need to run errands.  The two primary points of focus were getting my winter locks trimmed and buying a small, relatively-inexpensive boombox on which to play my bastardly-burned CDs.  And so, rather than staying around, studying and typing after classes, I took off immediately thereafter to seek my fortune with a stylish new do and miniature stereo.  It all went fantastically, I successfully avoided the evil one-stop shopping empire taking advantage of minimum-wage working women across the land, the Spring Mixes providing the more-than-apt theme songs.  In fact, I even managed to investigate the possibility of a new bookshelf/DVD tower and do a bit of grocery shopping AND fill Harley\'s tank before heading home by 3!  I KNOW!!  Amazing, isn\'t it?    And it didn\'t stop there!  Once home, armed with my new toy, I set about enjoying it immediately while catching up on checkbook balancing and general house puttering.  Eventually, all the good times had to come to an end for the sake of doing today\'s homework, which wasn\'t even needed as it turns out.  As I was preparing dinner before heading back to school to type up my back log of notes (yes, class, this is foreshadowing), I felt ready to experience the 1st Volume of the Spring Mix on \"Random\" mode.  I looked for a convenient \"Program\" button or the like on the handy remote the boombox came equipped with, but was denied.    \"How odd to only have the random feature on the player itself.\" I thought as I made my way over to the little guy on the floor, pushing away the thought you all are having now.  Crouching around the new acquisition, I denied the obvious even further.  \"Here, let me turn on the light.  It\'s got to be on here somewhere.\"  But it wasn\'t.  \"Maybe I need to do a combination of \'Program\' and \'Memory\'.\"  But I didn\'t.  \"Okay!  FINE!  So there\'s no Random feature.  So what?  I can live with it like it is!\"  But I couldn\'t.    \"Hey!  I\'m NOT going to Wal-Mart after all that!  Principles are worth more than a damn Random feature!\"  But they\'re not.  I tried convincing myself of that throughout dinner and the first subject\'s notes, but I knew the truth.  And so, Contracts was forsaken for yet another day and I made may way back across the border to return my briefly-but-intensely-enjoyed-but-ultimately-worthless new boombox in exchange for one from the Den of Satan.  And no, Shopko did not offer any viable choices other than the one I initially walked out with, so don\'t even start with me!  I loved my new RCA pal immediately, even though I felt kinda dirty about it.  To make sure I would never forget what was forsaken in the path to Random, I named the little guy Sam, after the founder of Wal-Mart.  He is now Sam Tort in further recognition of the trials those brave workers face each and every day so I can have everyday low prices and my Random feature too.   '),(23322,2314011,'29,April,2004','2004-04-29 00:00:00','    It is 5:13 and I am done.   With reading for tomorrow.  With outlines.  With studying.    I have less than 2 hours before yoga and then the rest of the evening is mine to do with whatever I will.    How odd.    So I thought it would be fun to compare the length of my outlines this semester to last.  It\'s pathetic and gross, I realize, but looking at my Property outline from Fall semester made me laugh out loud, so naturally, I had to find out how comical all the others were, too.    Property:  Fall = 7 pages.  Spring = 28 pages.  Contracts:  Fall = 12.  Spring = 34.  Civil Procedure:  Fall = 11.  Spring = 30.  Torts:  Fall = 11.  Spring = 26.     Criminal Law:  Fall = we didn\'t have this class last semester, but I couldn\'t leave it out of the fun.  Spring = 24.    I\'m taking the new numbers as a good sign.  And now I\'m going home to enjoy the rest of my evening entirely void of any duties or responsibilities.   '),(23323,2314011,'28,April,2004','2004-04-28 00:00:00','    Especially when they\'re tied or in the process of being tied.    Don\'t remember having an afternoon of solid work with corresponding finalities attached to every task.  It feels really fucking great.  Just finished the two major outlines for the two major exams next week.  Only in Contracts can two weeks of class add 15 pages to an outline.  But it is complete at last, coming in at just under 45 pages.  I\'ll think about hacking it down to 10 sometime this weekend.    Now on a little break before I dig in with the last bits of Criminal Law, Torts and Civil Procedure (which is mostly editing Miss Q\'s notes from the last two classes--no wonder it takes her a solid 2 months to outline, everything anyone could possibly know is included in those suckers!  Twice!).    Feeling soooo much better about everything as the week goes by.  A big part of it is studying with the gang.  Who knew study groups could be so helpful?  No more of this island-mentality for me, that\'s for damn sure!  Well, at least in respect to studying for finals.  I can be an island the rest of the day.  Classes winding down feels nice too.  It\'s been a while since I\'ve been  excited   about the end of the year arriving.  Last May was spent in denial about Graduation and going through too many goodbyes to be very much fun.    And there\'s so much to look forward to this summer.    The weekend in Bozeman right after the last exam: nothing but celebration and investigating new possibilities.  Return to the Library:  I\'m famous there and they\'re paying me more than ever.  Cool Cousin\'s Wedding:  Getting some nice family time, full of congratulatory pats on the head and hanging out with the newly-weds at a hip-happenin\' piano bar after the big shin-dig.  Hopefully the parents will retire to the hotel early.    Pride in Seattle:  missed it last year and this one is sure to be the best yet.  Ever.    And all that\'s only in the first month!!!    But first, there are two weeks to triumph over, and I shall return to preparing for the battle ahead (which has magically transformed into something less ominous, oddly enough).   '),(23324,2314011,'28,April,2004','2004-04-28 00:00:00','    ...that happened yesterday.    Because our choices were limited and no one else showed any interest, I was appointed  Interim   Chair of the Media Board.  Sigh.  We\'ll see how many times I\'m going to have to give it away before it stays away or sticks around.     And on a completely unrelated note:  Does anybody want or know somebody who wants a uterus?  I\'m taking bids.   '),(23325,2314011,'27,April,2004','2004-04-27 00:00:00','    Things that happened today (and by today, I mean within the last few days):  I reminded myself that my goal is not As; my goal is to do my best.  The Boy met a nice guy who loves Allison Janney.  And that\'s a good thing.    Sadie laughed until she snorted while reading Contracts flashcards.    I\'ve convinced myself that I will do well this semester.  (Props to the Oracle for the affirmation)  Karen reached between my legs.    Sadie got caught checking a girl out.    The Grind became not only the cool place I hang out on lazy weekends and flexible work afternoons, but also the cool place I study for finals in with my peeps.   '),(23326,2314011,'24,April,2004','2004-04-24 00:00:00','    Now that the Sami Omar al-Hussayen trial has come around, the publicity is back; this time, we\'re getting attention from the New York Times:    urlLink It is one of the few times anyone has been prosecuted under language in the antiterrorism law known as the USA Patriot Act, which makes it a crime to provide \"expert guidance or assistance\" to groups deemed terrorist.   '),(23327,2314011,'24,April,2004','2004-04-24 00:00:00','    As with the viewing of Vol I, I have nothing to say about Vol II that hasn\'t already been beautifully summarized by the Boy.  And so....   Needless to say I was all erect    (well, you know what I mean)  about Vol.II and I went to see it last night... Closure kicks ass! This chapter is just as phenomenal as the first, but in a totally different way. Where you got all the great action and set-up in Vol I, this time you get all the depth and back-story, as well as the satisfying and totally suprpising resolution. The genius of these films is that you KNOW where the story is going, but the methods of getting there are totally suprising and over-the-top in the very best ways. When you expect the epic, you get the intimate; when you think intimate, you get the most epic. Not to mention it\'s fucking gorgeous! Not only does Q have a great mind for dialogue, but he\'s got a bitchin\' eye. (and uses them in unexpected ways).  The two volumes together are stunning, that\'s all there is to it. I\'m such a slobbering fan, but it\'s the truth. A beautifully rip-roaring good time, and as much as we might want to see more of Uma and her bad-ass Bride, it\'s probably good that the story ends where it does. It couldn\'t possibly get any better than it already is.   So yes, such a great way to celebrate the end of another Crazy 88 week and kick off another weekend full of preparing for finals.  And if you let me know when you\'re planning on making your way to Pullman, esteemed Q fan and illustrious, wise professor in all things alternative, I just might join you.  We\'re down to the last week of classes and the obligatory doubts and worries are working their way in.  But they\'ll listen to reason and a few hokey self-affirmations.  The growing leaves on the tree outside my window and spotless sky help, too.  Once the day really warms up, think I might make it to campus and work on Torts while I work on my tan.  It needs the attention, let me tell ya.  Here I\'ve been so diligently crafting various muscle groups (at first because they really needed some action and eventually because summer was on its way and I needed to get myself ready to display), and the skin coating said muscles is still pastey-assed white.  We can\'t have that.    It\'s sick but there\'s a part of me who can\'t wait to tackle this semester\'s exams; to see how the attempt to push myself further worked out; to celebrate the victory of making it through the first year in a manner fit for fairy tales.  And there\'s another part of me who can\'t wait to be a librarian.   '),(23328,2314011,'21,April,2004','2004-04-21 00:00:00','    The powers that be really choose the wrong time to have first year law students plan the rest of their law school careers.  Two weeks before finals is just not the time to do it.  We\'re too stressed; we\'re too worried; we\'re already questioning our very existence and now they want us to decide how well we\'re going to do under  more  pressure and  more  stress.    Had a \"I wish the ring had never come to me\" night last night.  They\'re getting easier to overcome, I must say, but they still visit every once in a while.  And it\'s never a fun time.  The old \"what should I be doing this summer?\", \"do I have any right to be Chair of the Media Board with my schedule?\", \"do I REALLY have to do moot court?\", \"did I make the right choice coming to law school?\", \"when am I going to study for finals?\" questions had a party as I tried to get to sleep after avoiding the bed until after 11.    But this morning, most of them had gone.  Much of this has to do with end of the semester stuff.  And really, this round is sooo much better than the last.  At least I  know  what I\'m preparing for this time and that helps a tremendous lot.  The doubts manage to creep in, however, when I look over at other\'s outlines many times the size of my own or when we get emails from the Oracle telling us we should be doing 2-3 practice answers  A DAY  in preparation for the big exams.    So yea, this atmosphere is not the best for figuring out the next four semesters.  But chatting with professors helps.  And finally making decisions I resisted for a few months does well to take the edge off, too.  The schedule that yesterday looked to be a punishment for stupid, idealistic, oblivious hobbits who have no idea just what they\'ve gotten themselves involved with, today looks entirely do-able, challenging in empowering ways and maybe even a bit interesting and fun too.    I\'m still taking the summer off; but had to finally let the Student Media Board go.  It was hard, but Dan held me afterward; and I immediately felt better, having decided.  Only a few hours later, another option appeared for next summer, which will make for more flexibility during my last semester.  Will still run by the new schematic by the Dean tomorrow, but at least I\'ll sleep easier tonight.   '),(23329,2314011,'19,April,2004','2004-04-19 00:00:00','    My reward to myself for completing the Criminal Law outline, replacing the unfortunate-coffee-fence-jumping-incident-victim  Property Study Guide and cleaning my apartment over the weekend was renting Kill Bill, Volume I. Its meaning and effect on my life can best be described by the Boy from his blog over the weekend:   Never been much of a Tarantino fan. I always found him as obnoxious as he was pretentious, and he didn\'t earn any points by casting himself as the obnoxious asshole in all his movies. I hated \"Reservoir Dogs\" and I thought \"Pulp Fiction\" was cute for its novelty, but otherwise empty and overblown...except for Uma, who only got 1/2 hour. After she leaves, the movie is dead to me. I didn\'t even bother to see \"Jackie Brown.\"  (note:  I actually haven\'t seen Reservoir Dogs, but his was too great a tangent to edit.)   So I scoffed at the hype over \'Kill Bill\' and swore I would hate it if I ever wasted enough time to watch it. I saw it as poor, washed-up-Quentin\'s attempt to cash in on the post-Matrix Kung Fu revival which is so sad it\'s not even funny...and what\'s with this Charlie\'s Angel\'s rip-off anyway?  (This paragraph I agree with completely)   Well this snobby, elitist Pig was more than happy to be proven wrong.   (respectful snipping)   It didn\'t take ten minutes for all my venemous hate to become drooling awe. (That doesn\'t happen often, I assure you.) I was totally blown away by everything about it: the dialogue, the characters (and their names!), the action, the editing, the shameless 70\'s film referrences, the fucking music! Just dead-serious enough to care about the characters, with enough winking parody to offset its own cartoony violence. Even the damn credits were brilliant! And Uma as The Bride! Uuummmmmaaaaa!!! It\'s the movie she\'s needed since Pulp Fiction, and thank god Quentin finally gave her one. (without John Travola!) I was very pleased to learn that Q & U developed the character together way back on that OTHER movie. No wonder she embodies it so well...   (YAY!!!  ME TOOO!)   Anyway, I felt like a whole new fag  (dyke)!  I could not believe how wrong I was, and how much I loved being wrong. How often can you say that you love...LOVE Daryl Hannah and actually mean it? I was a giddy, bouncing fan-boy and not a bit ashamed   After having the same experience, I called him up and we gasped and giggled like little school girls over that scene, this look, ALL THE LINES!!!  It made going to bed after midnight worth it, and when I slept through the alarm again this morning, the omission of morning pages didn\'t hurt quite as much when I thought of the red cross in Daryl\'s eye patch.   '),(23330,2314011,'16,April,2004','2004-04-16 00:00:00','    Today is my anniversary; or, I guess it  would  be.    Didn\'t notice it until checking my watch in Contracts class and saw that it\'s the 16th.  The rest of the class was spent reliving snipets from the day six years ago.  It was a Friday (last year was leap year, remember), and things were slow in Government class.  We must have had an exam or some like thing because I\'d brought my current book with me to fill the remaining 40 minutes.  The author had included his wedding ceremony in the text and I was struck by how similar the sentiments and promises described in the passage resembled feelings, thoughts and conversations Lorraine and I had on a regular basis.    We\'d done the play \"Will you marry me?\" thing quite a bit.  Mostly in the context of playful, relaxed, lazy evenings or afternoons basking in the few hours we had together that seemed to stretch on for days.  But reading these exchanges this day suddenly made me want to do it for real, if only in our own minds.  And so, I left class knowing I would ask the woman I loved to marry me that evening.    It was a pretty informal affair.  None of our friends were in attendance, and we never received any gifts.  I even, quite by accident, called her by her ex\'s name at one point, reading the name out of the text (what are the odds that the author\'s wife\'s name had same name as her ex?!)  But it was sweet and we celebrated it every year, in whatever capacity we could.    Without realizing it, we had a perfect celebration last night.  I returned home from a full day of brilliant law school activities around 9, and pushed on for another half an hour before finally admitting defeat.  Scraped a picnic dinner together in front of the TV, at last free to relax and enjoy the remnants of Thursday night TV.  When the phone rang, I whined and refused to get up unless it was someone letting me know I\'d just won a free puppy or something.  Pretty close.  It was that girl, calling to say a quick goodnight before her Sleep-Be-Here kicked in.  It was pretty perfect, really, and the fact that neither of us knew what date it was probably contributed.  Had good laughs like we used to, and told idiosyncratic stories that illustrate more about who we are as people than long, involved, deep conversations; there\'s just nothing like coming home completely ragged to an unexpected connection with someone who knows you completely.    Happy Anniversary, my Mrs. Jesus.   '),(23331,2314011,'14,April,2004','2004-04-14 00:00:00','    I\'ve never really mastered the art of waking up to Morning Edition.  Maybe it\'s the Brit\'s soothing voice; maybe it\'s the lack of annoying commercials that refuse to shut up; maybe I\'m just conditioned to FM radio.  Whatever the explanation, I\'ve consistently been sleeping through the insightful, probing news stories and these have been increasing in length and significance.  At first, it was just a few minutes out of my breakfast-TV time; then, it was hurrying through Morning Pages a bit; this morning, I managed to get myself on Tuesday\'s clock.  Tuesday-time is half an hour later than Wednesday-time.    I thought I was only 11 minutes behind when I rolled out of bed at 7:11.  And when I got in the shower at 7:37, I was seemingly right on time.  True, there was something a bit off all morning.  Just seemed to not be entirely on top of things, or at least not as much as I would like...but I attributed it to the mocha I had after 5 last night and my less than filling dinner of leftovers.  Nothing a good dose of water and vitamins wouldn\'t fix.  But, when I finally glanced at my watch before heading out the door, I was slammed back into Wednesday-time: 8:06.  Class began at 8.  Yay.    During the drive to school, I considered not going to class at all.  But I didn\'t spend more than a good chunk of my early and late evening reading 20+ pages of material just to not have the lecture to go along with it.  And so, I went into class, 20 minutes late, head held low and avoiding eye contact.    But I needn\'t have worried:  they had done student evaluations during the first 20 minutes of class and the rest of the time was spent discussing the functions and form of the Statute of Frauds, and not the material I read so diligently last night.  So basically, I got to sleep in, albeit unknowingly, took notes that don\'t need to be typed up and have no reading for Property to do over the weekend.  Student evaluations rock.   '),(23332,2314011,'13,April,2004','2004-04-13 00:00:00','    I\'m already mourning the loss of Miss Q\'s presence over the summer.  She\'ll be in DC, playing breadwinner to Alycia\'s Super-Law Student-Intern for an illustrious international ambassador; and together, they\'ll probably make the world a little bit better for their efforts.  Meanwhile, the rest of us back in Moscow will be missing out on the promised barbecues and hot tub parties.    Plus, I\'ll be missing out on conversations like this:  \"Hey Miss President, how was your weekend?\"  \"Did you bring me a cookie?\"  \"What kind of a relationship to you think we have?\"  \"Will you carry my books for me?\"  \"What am I to you?\"  \"A gap-filler.\"  \"Here, I\'ll walk behind you, since I\'m just the Vice President.\"  \"Two paces, at least.\"   '),(23333,2314011,'12,April,2004','2004-04-12 00:00:00','    Not much to blog, really.  Pretty full day of being crouched over various books, struggling to keep the eyes open during the laborious discussions of expectation damages and encouragements to breach for Contracts and the difference between some courts\' requirements for an overt act in characterizing the offense of attempt for Criminal Law.  A nap helped.  The two Torts assignments put me behind a bit, but I trudged back to school around 8 to prepare for Civil Procedure.    Coming home to a request to have a copy of my Motion to Suppress Tangible Evidence ready for the Advanced Research and Writing professor added a nice finish.    Now just surfing nearly-forgotten web sites and checking email at a leisurely pace; never would have considered these activities of luxury, but they feel pretty indulgent tonight.   '),(23334,2314011,'11,April,2004','2004-04-11 00:00:00','    There\'s just something about bringing 58 pages of typed class notes to their knees in a 33-page outline that puts a gal\'s spring back in her step and makes the world a better place, if only for an evening.  Finishing tomorrow\'s reading, grocery shopping, and two loads of laundry helps too.    Suddenly three weeks seems like just enough time to get the semester\'s tasks finished without complete and total burnout.  This morning, I wasn\'t so sure.  I was on page 18 of said 58 and couldn\'t imagine June in any capacity.  But tonight, I can smell summer on the air and the light at the end of the tunnel doesn\'t seem quite as daunting.   '),(23335,2314011,'09,April,2004','2004-04-09 00:00:00','    So the entries have been a bit spotty lately, I realize, and I hate it.  Well, not \"hate\" so much as \"can\'t do a damn thing about it so we\'ll all just have to deal.\"  Been hanging at the Grind for the last few hours, perfecting the Property outline and feel pretty damn good about it, if not just a bit exhausted.  Still not quite sure I grasp the intricacies of real covenants as opposed to equitable servitudes, but the foundation\'s there and the gaps will be filled in due time.  Get to spend the weekend organizing the house of cards called Contracts, which actually feels like a break from the recent history of brief-writing and emotional gauntlet-running Saturdays and Sundays.    But I thought I\'d take a minute and fill in some holes from the week, because I know I left you hanging:  The oral argument went well, just as everybody tried to assure me.  Yes, I froze on the first question, just as I\'d feared.  But, just as promised, I warmed up and kicked ass thereafter.  Didn\'t quite get onto the second issue as quickly as I should have, but overall, I did very well.  Must have been the zipper.  Celebrated the week\'s major anniversary with a hot Girlie Exam.  You know, the one we all hate, dread and try to put off as long as we possibly can.  But at least I was finally getting some action...even if it was in a paper-thin cotton gown with a flashlight-equipped-speculum.  It was over pretty quick, to my relief (and dismay??), and I was left wondering whether I still have a \"cervix they use to draw anatomy books.\"  That\'s what Gloria told me during my first exam and I\'ve always been proud; no such luck this time, but I still like the idea of having a Textbook Cervix, so I\'m gonna keep it around.   '),(23336,2314011,'08,April,2004','2004-04-08 00:00:00','    \"What am I going to ask Tori when I meet her?\"  I asked the Boy one night a year and a week ago as we planned the impossible.  \"I dunno...maybe you should ask Neil.\"    Just to remind you, dear readers, Neil is Neil Gaiman, author of the Sandman graphic novel series and several other venerable works.  He\'s one of our favorite obsessions, and that\'s saying a lot.  He has is own blog and an option for readers to send him questions.    \"If he thinks it\'s not worth his time, he won\'t answer and you don\'t have to worry about it.\"  The Boy encouraged as I deliberated, already knowing I would.  We didn\'t want to be the crazed fans, after all.  We knew we were going to be so cool when we met our celebrities, because that\'s the only way to get them to realize how good friends we will all be.    I passed off the request for advice as a cool-headed college student just trying to handle the possible interview situation a bit better than her contemporaries.  But I couldn\'t sleep that night after receiving his response:  \"The best thing to do with Tori is to throw away the list of questions, sit back and just chat.  She\'s wise, funny, kind and cool and it will be fine whatever you do.  Give her my best.\"    And so, a week later, I made my way up to Spokane early to hopefully catch a glimpse and perhaps an interview.  I knew at least I would ask about the Dew Drop Inn in London and get the picture signed for Levi, but if I could actually sit down and (gasp)  talk   with her, I wasn\'t quite sure what I\'d do, but was confident in my ability to figure it out with the least amount of embarrassment.  There was a small group of 20-ish folks waiting when I arrived and it was clear I neither knew nor fit in with any of them.  Gay guys and goth girls lounged in the warm April sun and I sat off to the side, taking it all in, still trying to decide what I wanted her to sign if it came to that.  Eventually, they set up the barricade and we maneuvered around each other.  Veterans made sure those of us who had yet to meet her were moved to the front and we were grateful.    Then we waited.  We laughed with the security guard.  We bounced on our toes.  We sang her songs.  We watched her bus pull in.  We stopped breathing whenever the stage entrance door opened.   Finally, she walked out.  She answered questions and started signing items for people in the back before making her way down the front line of folks crunched against the metal barricade.  By now it was clear I wouldn\'t be wasting any contact I had with her in a request for an interview; she could say no and walk on down the line.  Besides, this was a time for fans and their artist, and the fan (obsessive and obnoxious as I can be) won.  I had to remember to take pictures and keep inhaling.  Then she was talking to the girl standing next to me, who asked for a picture; Tori handed me the girl\'s digital camera and took my to-be-signed-items from me.  I still feel bad about how terrible the pictures must have turned out; all hand-eye coordination was lost, as though I hadn\'t even seen a camera before, let alone knew how to operate one.    Then she was asking my name and I hoped I wouldn\'t smudge the song requests she\'d written on her hand as I shook it.  I couldn\'t form a complete thought or sentence.  I was that person I promised myself I would never be when meeting celebrities.  No shame, no control; just gibberish and gush.  But she threw me a bone:  \"So, how\'re things?\"    \"Things are good!!\"  I exclaimed.    My god her eyes are green.  She still has your things in her hands.  Say something about them!!    \"That picture you\'re holding?\"  \"Oh!  Look at that!\"    \"My best friend and I went in London, thinking we\'d find this hip-happenin\' place we wouldn\'t be able to get into, and when we got there we couldn\'t, but for a completely different reason.\"   Did I just get all that out in one breath?   \"Yea, well, this isn\'t it,\" she explained with a soft smile.  \"They\'re kinda scattered all over the country...\"  \"I know!!  My grandparents owned one in Paul, Idaho!  We just thought this was a gay piano bar in London and figured it had to be you!\"  \"My girlfriends used to say to each other when we needed to get out of somewhere...like code for \'Let\'s get the fuck outta here\'...let\'s go to the Dew Drop Inn.  So, how am I spelling this?\"  And she moved along the line, hugging some, remembering the names of others and asking about their families.  The whole thing was pretty surreal, to the point where I (again) had to remind myself to take pictures and remember every instant because   this was really happening  .    It didn\'t hit until she\'d gone back inside to prepare for the evening\'s concert.  It didn\'t seem real until I called the Boy.   I sent Neil the link for the  urlLink review of the show , not thinking he\'d actually read it or even respond, but just in case...  He did.    \"Good review.  Nice job.\"    The whole thing still comes around and surprises me every once in a while.  I smile when I catch a glimpse of my name in her easy, generous script when I\'m doing sun salutations in the living room.  I relax when I see her picture on my carrel at school.    When I blasted into my apartment in August, rushing to change clothes in order to get to law school orientation, for which I was 4 hours late, I thought:  \"You can handle this.  You met Tori!\"   '),(23337,2314011,'06,April,2004','2004-04-06 00:00:00','    I\'m freaking out.  Have my oral argument in less than 4 hours and can\'t sit still; can\'t get my palms to stop sweating; can\'t focus on my breath; can\'t get the vision of freezing in front of \"The Court\" when they ask me an easy, obvious question I\'ve been rehearsing for 3 days.    The fucking thing\'s not even graded for god\'s sake.    I\'m reminded of my days of high school debate, when the morning before each competition brought similar symptoms (not indications of a disability or request for accommodation, mind you) and I wondered why I even got myself involved in this shit.  But, once the round began and I stepped up to the podium, took a deep breath and began, I was fine and in the zone.    My cold finally took hold last night around 3, when I awoke with terror and avoidance plans, and the medicine is providing a nice haze to the already adrenaline-fuzzy world.    Thank god my power suit has a zipper.  The Chief Justice loves gals with zippers.   '),(23338,2314011,'04,April,2004','2004-04-04 00:00:00','    It\'s been an interesting few days.  Taking today pretty easy; partly to recover from the previous two and partly because I finally can relax, if only for a few hours.  The fan\'s in the window and the spring air\'s circulating nicely through the apartment.  As though the emotional changes, growing, and sharing wasn\'t enough (we\'ll get to that in a minute), the whole experience has just been topped off by an email from (you\'ll never guess) Lorraine and I\'m sitting here in a new place at last.    Something\'s been building for the last few months, something not-so-vaguely resembling the final blow-up/breakdown before the Age of Silence this November.  But, with the gift of experience and Susan\'s advice, I\'ve been able to prevent any nasty expectations from causing more trouble than they\'re worth and have been able to accept anything Lorraine\'s given me by way of communication in the last few weeks.  It seems I\'m still a problem; an issue, sword, card  to be unleashed only in the most heated, raw discussions.  Despite all her best efforts to maintain peace and balance between us all, Lorraine seems to be losing.  And rather than rushing to her aid, to be there, to listen and to hold (albeit over the phone), I\'ve sat back and waited, resisting the urge every bit of myself is compelled to act upon.  Reading her message today, I found myself waiting for the spark of being slighted, of being put aside in the interest of waiting to see how things worked themselves out.  It never came.  Instead, I felt incredibly sad for how she must be feeling and smiled when I pictured her outside, playing in the dirt.  I will be here when she needs me, but I won\'t be counting the minutes until she does.  Not out of indigence (finally), but out of patience, trust, and resolve.  I can\'t attribute my reaction entirely to the experience of this weekend, but I can\'t deny its impact, either.  And when I think about it, what we did as a group wasn\'t that extraordinary.  We sat in a room, walked across it a few times and talked.  But walking across a room can be the most difficult, profound thing one can do when one\'s exposing more about the self than one ever thought or intended to do.  And that\'s what we did for a day and a half; even when we weren\'t in the \"walking\" exercise, we exposed ourselves, sometimes to the core, and held one another together as a cohesive, bonded group afterward.    The idea was to learn how to be more multiculturally competent in order to better serve our clients, colleagues and ourselves.  In so doing, I didn\'t expect to delve quite so deep into who I am as a person or to listen quite so intently to the people around me.  But, if you want to make the world a better place, free of bias and injustice, the first place to start is yourself.  By the end of the day, I didn\'t want to leave the room and renter the world at large.  I felt like we\'d done so much work together and how was I ever going to incorporate something so big and meaningful into my everyday life?    I guess I\'ll figure that out as I go along, and I can already tell nothing will compare to the ride.   '),(23339,2314011,'01,April,2004','2004-04-01 00:00:00','    Been having a bit of over-extension stress the last few days, but a nice, long, brisk walk around campus putting up fliers for a lecture next week I know very little about (which may not be the best way to begin an administration, but what the hell?) with the Prez makes the mountain of tasks waiting back at my desk seem entirely more do-able and perhaps a bit pleasurable.    Beers and jazz at Pullman\'s intellectual bar is still in the running, but have a few hours to get through before anything\'s known for sure.    Back to work...   '),(23340,2314011,'30,May,2004','2004-05-30 00:00:00','    It\'s just not the summer blockbuster season until the first disaster movie is released.  So now we can get on with the business of all the other big summer blockbusters, because really, \"The Day After Tomorrow\" doesn\'t serve many other purposes.  Don\'t get me wrong, its special effects are incredible and the movie does its job of providing a 2-hour eye-candy-fest perfectly.  But I couldn\'t help thinking, watching six tornadoes sweep through and demolish Los Angeles with stunning detail, \"Someone get Helen Hunt on the phone!\"  But it was a nice way to spend a couple of hours on an empty, lazy day during an empty lazy weekend.  The only thing on the agenda for the rest of the night is some hard-core futon time with a young wizard.  Yay.    But first, a story...  H\'ok so, the whole \"gay subject\" has always been a delicate one with the parents.  It was the first issue that created a rift between them and me, even in the days before I knew about myself.  They didn\'t think themselves homophobic; they just preferred it if gay folks just stayed \"over there\" and kept quiet about the whole thing.  And really, that\'s where they remain on the subject, even knowing about me.  It\'s not that surprising, and they\'ve been much better about it than so many other parents in the world.    But any conversations about my sexuality have been very touch and go, very few and far between.  The last talk my dad and I had about it came a month after I came out to them in October of 1999 and ended with slamming doors.    Finally had to quell their last hope that I would one day embrace my feelings for Levi for what they clearly were and get over this whole gay thing last summer by letting them know that not only was my best friend and their best chance for a son-in-law gay, but it looked like it was going to stay that way for me, too.  Mom and I managed to have our first conversation about the whole thing since the outing and it appeared that some progress had been made during the years in between.  She still couldn\'t say the word, but at least she didn\'t cry.    And so, there\'ve been baby steps for months.  Talking about my new law school friends, sometimes it had to be addressed.    \"So, is Susan married?\"  \"Umm, no....\"   Let it go, Mom, I\'m not in the mood to go there.    \"Divorced?\"  \"Noooo...\"   Alright, I guess we\'re going there.     Pause.  \"There  is  a third option.\" I said with a laugh.  \"I knowww...\" she replied with a long sigh, but managed to change the subject without this disclosure ruining the entire conversation.    Then came the Christmas holiday during which The Boy made an extended stay at the Mountain Fortress.  We toned it down a bit, but there were times our banter strayed into the area marked by rainbow flags.  And shockingly enough, the world didn\'t end.  Mom would usually either ignore us or mutter something about how she didn\'t want to hear about it, all the while keeping one ear on the conversation.    During the slow evolution and turnings of small corners, I\'ve been content to be patient and steer clear from that topic of conversation.  They\'ve been able to put the information in another place and not let it change how they are toward me, how they feel about me, and for that I\'m grateful.  Meanwhile, I\'m sure they\'ve been dealing with it in their own ways and I\'ve had no cause to disrupt their process.    Until I met someone, that is.    Hadn\'t mentioned anything about Miss Vicious in the last two weeks, despite the routine questions of \"How was Bozeman?\", \"How\'s the week been going?\", etc. and before long, it felt as though I was keeping something hidden.  It\'s a familiar feeling, and one I have no desire to relive.  And so, I decided to tell Mom about her when she made her weekly check-in call.    There weren\'t any opportunities to bring up the subject by the time we\'d exhausted all the other topics, and when Mom asked \"You wanna talk to Dad?\" I thought about just saying yes and waiting until next time.    Instead, I said \"Actually, I need to talk to you about something.\"  She let out one of her signature sighs, telling me she had an idea of where this talk was headed and was none too happy about it.    \"I will tell you as much or as little as you want, but I just think you should know...I\'ve met someone.\"  (Insert brief recounting of the Boy\'s matchmaking here.)  \"You\'re just hell-bent on this, aren\'t you?\" she asked, probably through a bit of tears, wishing I would surprise her.  \"I\'m not hell-bent on anything, Mom, this is just the way it is.\"   \"So,\" she said, trying to keep her bearings, \"what do you think?\"  \"I want to see where it\'s going.  We\'re not talking about moving in together, and we don\'t really know what\'s going to happen with her there and me here, but we both want to see where it\'s going to lead.\"  \"And this is the first time you\'ve had these romantic feelings for someone...\" she started.  \"Well, not the first time, but it\'s the first time it\'s happened this way.  I\'ve felt like I\'ve been keeping something from you by not saying anything and I don\'t want to do that.\"  \"Well, I\'m glad you said something.  I always figured you would if you needed to talk.  What\'s her name?....\"  The conversation about Vicious herself didn\'t last very long, and it quickly became what I knew it would be when I brought it up, another turning point.  \"So how are you doing with...\" She still can\'t say the word.  \"With what?\"  \"You know...with all this stuff.\"  \"I\'m fine, Mom.  Really.  I have been for a long time.\"  \"Good, I\'m glad.\"  \"And really, you guys can always ask or talk to me about anything you want to...\"  \"Oh, I don\'t think we\'ll be doing that, dear.\"  She said with a laugh that was part nerves, part painful and no parts humor.  \"I don\'t mean about this in particular, I mean about in general.  It\'s been in the news a lot recently, and I just thought I could be a good resource if you wanted to talk about anything.\"  \"I don\'t think we\'ll be the ones to bring it up, dear; we\'re just going to have to deal with it.  I know it\'s not going to change, and it makes me feel bad.\"  \"Why?\"  \"Because...you know...you won\'t have kids...\"  \"Mom, I plan on having kids, they\'re part of the long-term plan.\"  \"Are you sure you want to do...that...with a kid?  I mean, I know you think it\'s so mainstream now, but it really isn\'t.\"  I put the sting from her question away for a minute so I could answer the question honestly, hopefully keeping this conversation an open exchange of information.  \"It\'s something I know I\'ll think about seriously in another 10 years, when I have someone I want to be with forever and we both decide we want them.  But I don\'t think \'Oh shit, there goes kids.\'\"  \"I know,\" she said, \"it\'s just that it\'s so hard and always will be.  I guess you can decide where you live and that could make it easier...Maybe you and Levi could have kids together.\"  \"Well, he is on the top of the sperm donor list.\"  So, she wasn\'t talking about \"being gay\" not being the right thing to do with a kid, she was worried about the surrounding atmosphere.  Part of me smiled at that realization.   \"And it IS everywhere now,\" she continued.  \"The hardest thing for me is how it becomes so wrapped up in who the person is, and it shouldn\'t.  I hate that that\'s who Rosie and Ellen are now.  It\'s not who the person is and I hate that it\'s such a big deal.  I just want it to be something that doesn\'t have to be announced, that it\'s just something about a person and it\'s not a big deal.\"  \"Well, that\'s what a lot of people want too, but in order to get to that place, we have to go through this place where it is a big deal.\"  \"I guess,\" she said, sounding resigned and I sensed an opportunity to actually share my feelings in a way I hadn\'t before.  \"I can\'t get inside those people\'s heads who come out and then are forever labeled as the \'gay-this\', but I imagine some of them wouldn\'t want to be attributed that any more than you want to hear it.  It\'s something that happens to them after they\'ve come out, over which I imagine they have little control.  And really, I think a lot of gay people want to get to that place, too, where it isn\'t a big deal, it\'s just something you hear about and go on.  But if it really is to be not a big deal and something that just happens to be part of who someone is, then we ought to be able to have the same opportunities and same rights as everyone else.  When things get closer to equal, it may not be such a big deal.  But as long as there\'s a difference, it\'s going to be a big deal.\"  \"Is this what you\'re going to do with your law degree?\" she asked with another nervous laugh.     The conversation eventually, and comfortably, moved to other things.  But it\'s stayed in my mind for much of the day, as I\'m sure it has for her, as well.  They\'re getting easier, these talks; maybe one day she\'ll have what she wants: the ability to know about it and move on.   '),(23341,2314011,'29,May,2004','2004-05-29 00:00:00','    urlLink    Memories of the Super Students Bender....&nbsp; urlLink      Okay, so, added the lovely feature that lets me post pictures to the bloggy!  Yay!  And the only appropriate maiden picture was the Infamous White Board of 1428: The Den of the Stellar Super Students.  Each exam had its own rallying cry, drawn from its respective study sessions.  The surrounding quotes were faves over the two weeks, but didn\'t quite reach the level of exam-theme, if you will.  So there ya have it.  Suppose I could post a key for all those non-Stellar Super Student readers:  \"Strawman!\" = Property flashcards session at 1428.  Emily was all over the red flags hidden in the fact patters.    \"Eyes on the Prize!\" = Contracts food-run to Tideymans.  Emily and I had to focus ourselves on our breaks to ensure our continued studying once we returned from the store.  \"Hello, bullets!\" = Criminal Law flashcards at the Eastside Marketplace.  Emily found the disorganized meshing of multiple factors to consider troublesome while Calamity Jane and I pictured a shoot-out in the card\'s fact pattern.  \"Sharon!!\" = Civil Procedure flashcards at 1428.  The first study session with the new hair do was to be the one during which I finally received a nickname.  We couldn\'t remember who the other twin was from The Parent Trap, but Emily reminded us shortly as we dug through our Federal Rules of Civil Procedure, shrieking out:  \"Sharon!\"  The extra exclamation point was added after the name evolved into an Ozzy Ozborne impression.  \"25.5 for 130.\" = Torts study session at the Grind.  Emily reminded us that we had 25.5 hours in which to study for a 130 minute exam.    And that\'s how it really happened.    Have more to post later, involving another big gay milestone with Mom this morning, but it deserves its own post and I deserve a bit of mindless Summer Blockbuster entertainment, so I\'m off to see \"The Day After Tomorrow.\"  Look out Global Warming.   '),(23342,2314011,'29,May,2004','2004-05-29 00:00:00','    So I\'m hangin\' out at my carrel in the law school basement.  Thought I\'d drop by and pick up everything I\'d left; needed some distance from all the books, paper, shelves and dingey old books.  Putting the books I didn\'t sell back on Half.com, trying not to think about how much I paid for them as I set the sale price.    It\'s weird being here now:  all the empty desks, some cleaned out, some apparently recently used.  Enjoying the Wi-Fi, but will be happy to leave this place behind for good this afternoon.  Think I\'ll spend the afternoon at the Grind, catching up on web surfing and reading.  Yay.   Yea, gonna go through all the notebooks and trash all the precious notes, sort out what needs to be kept and lock it all away in the big gay office upstairs until it\'s needed again in three months.  Think I might take the long route to the Grind, too.  The day is trying to be nice, with the Superman Blue sky peeking in around the gray rainclouds, but it\'s still too nippy to enjoy sans the protection of windows.    Good times.   '),(23343,2314011,'28,May,2004','2004-05-28 00:00:00','    \"Not like anyone\'s going to notice I\'m gone, anyway,\" my supervisor grumbled as we paid for our fabulous lunches.  The joint-dining experience has become the standard routine with nary a word said about the matter.    \"Really, that office is sitting empty right now,\" I joined in.  It\'s Friday of a three-day weekend and even though people are at work, no one is at work.    \"Oh well,\" she said, cheering, \"I\'m leaving at 3, no matter what.  I have annual leave.\"  \"Yea, and I\'ll stay \'til 530.\"  \"Wow.  You\'re hard-core.\"  \"No, I\'m paid by the hour.\"   '),(23344,2314011,'27,May,2004','2004-05-27 00:00:00','    And that\'s only 7 months away; but at the rate I\'m going, it might not be that far-fetched.    It\'s been quite a week.  Full of twists and turns, excitements and adventures, some more enjoyable than others.  But they\'ve all lead to this place and, as promised by the stars, was all worth it.    H\'OK, SO...  much time has been spent sorting out the whole \'where is this going\' question.  I know, before you even start yelling at your computer, I KNOW!  It\'s too much thinking, over-thinking, planning and re-planning.  I\'m working on it, okay?  Ultimately, I came to the conclusion that I\'m down for the ride; I\'m ready to see what\'s coming and where this will take us.  I\'m game for whatever happens.  And that\'s a new place to be, really.  I\'m not one who gives up control or its illusion easily, especially when there\'s no indication or assurances of what  could  or  might be  happening.  Haven\'t quite gotten down the art of trusting in nothing but a feeling, but it\'s getting easier.  But finally emerging from the haze of trying to pin every possibility down and analyzing it to death in order to figure out if its a winner or not was pretty liberating and lived up to all the hype.    And that\'s how the week began.  Open and ready for whatever came next.  Just wasn\'t ready for what ended up happening (do we see a pattern, here?).   SO YEA...  couldn\'t find this new girl consuming many of my thoughts.  Our schedules are kinda scattered to begin with and our conversations thus far have been fitted into whatever open slots come our way, so it wasn\'t too big a deal at first.  Before long, though, started worrying about something I may have said, not said, done, not done, etc.  Of course, now we have that wonderful friend, hindsight, to show us the truth of the circumstances, but at the time, I just had a mind reeling with too many possibilities and explanations that may or may not be the case because I couldn\'t talk to the one person who could reassure me.  Maybe she just needed time to digest everything; maybe she was caught up in personal issues having nothing to do with me; maybe...  You get the idea.    AND SO...  I tried to give her what it appeared she needed, space and time.  But patience has never been one of my strong points, especially when the situation involves someone I care about.  I want whatever the problem is to be fixed yesterday, especially if there\'s anything I can do to fix it.  But I couldn\'t do anything here because I didn\'t know what was going on in the first place.    AND THEN...  came home from work last night, trying to be strong, trying to be brave, and almost convinced myself I could pull off the whole \'patient\' thing when it became quite apparent I needed to do some evening pages (the evening counterpart to morning pages).  Sure enough, they did their job.  And after about half an hour of quality pen and paper time (while watching a flash flood outside the window, appropriately enough), I looked up, felt incredibly lighter and as though something had finally been released for good.  By the end of writing everything out, I was smiling. Because the ride\'s been a wonderful one and whatever happens, it\'s been worth it; wouldn\'t change a thing.  I could wait forever to hear from her again, if that\'s what she needed.    AND REALLY...  the only thing to do after a major breakthrough in understanding yourself is to put on the new Alanis Morissette CD while doing some hard-core yoga, and I\'ll be goddamned if it\'s not a pretty good album!  As a brief aside, the CD is pretty much like all the others:  hate it at first, then it\'s okay, and pretty  soon it\'s kinda alright!  Yes, there\'s some songs that just suck, and this whole 10-song thing is complete shit, but the good songs, are really great and I can\'t get enough of them.  Then curled up on the couch and finished the first Harry Potter book, which is such the perfect re-introduction to reading for pleasure.  Thanks, Emily.  You Rock.    AND THEN...  around 3, the phone rings.  And I jump out of bed, run to the phone and am sooo excited that we\'re doing the late-night, \'I couldn\'t wait until morning\' call, only to find someone else on the other end.  Fine, it\'s my friend April and we have a few laughs before she graciously lets me go back to sleep, trying to maintain the new peace and understanding I\'d found.    SO TODAY...  managed to keep the momentum going from the night before.  Made it through the day without planning a damn thing or over-analyzing a single thought.  How novel!!  And sure enough, after my most excellent yoga class, I had a message waiting for me and a gal to talk to.  She\'s been sick all week, poor thing.  Spending the days in bed or in doctor\'s offices, talking with people who don\'t know how to help her; unable to do everything she wants to do: be with her friends before they all become scattered, talk to and take care of the people she cares about, etc.  She\'s been resigned to bed rest and that\'s the worst.      AND SOOO...  we chatted for the first time since Monday night and it was perfect.  Comfortable, easy and adorable.    FINALLY...  here I find myself, still here, still smiling (while shaking my head at myself), feeling good all over, and down for whatever else the stars have in mind.  Yes, you were right, dear friends and dear horoscopes: patience, going with the flow, and giving up the tight hold on invisible reigns are our new friends.  I apparently need to have the same lesson given over and over before it starts to sink in, but each time, I get the hang of it a bit better.    (cue exhausted, self-deprecating, contended sigh)   '),(23345,2314011,'26,May,2004','2004-05-26 00:00:00','    My fortune from today\'s lunch said:  \"You are apt to give too much thought to appearance.\"    Hmm...seems appropriate, I thought and then said to myself.    My new supervisor, whom I was lunching with, who is a few months younger than myself and honestly believes Gwenyth deserved the Oscar for Shakespeare in Love over Cate\'s Elizabeth, looked me over and said:  \"Mmm...not really.\"   '),(23346,2314011,'25,May,2004','2004-05-25 00:00:00','    I don\'t know how she does it, but Tori has a way.  Never really thought about her as a therapy mechanism, but lately, nothing appeals more or helps take my mind off things as much as she does.    Thanks, ol\' girl.   '),(23347,2314011,'25,May,2004','2004-05-25 00:00:00','    I\'d forgotten the roller coaster ride these things bring.    I remember the newness, how heightened everything seems in the beginning, trying to find my way through old feelings resurfacing in wonderfully new ways.    I remember the steps on the way to getting to know someone, really \'knowing.\'  But I\'d forgotten how quickly everything can change over so small a thing.  How suddenly the day can transform into something completely different than how it began and somehow circle back again by the end.    I\'m remembering now why the tides are so frequently forgotten.  If we didn\'t forget, we wouldn\'t try again.     And I\'m reminded tonight that the ride, with everything it brings, is still preferable to the alternative.   '),(23348,2314011,'24,May,2004','2004-05-24 00:00:00','    Some things that have happened this morning to cheer me out of my all-too-pensive mood:  Blue skies and warm sun;  Random cheerful conversation with a janitor on the elevator;  Getting close to $100 for Property and Contracts books;  Finding a note left by the Office Goddess that she managed to sneak my hours from last week in under the wire, which means I\'ll be getting paid the 4th instead of the 18th;  Checking the web for any posted grades:  Torts: B+ (one grade-level improved)  Contracts:  B+ (three grade-levels improved)   '),(23349,2314011,'22,May,2004','2004-05-22 00:00:00','    why we love May: It gives us the song of summer.  During the month that sets the tone for the entire summer, we find the song that sets the tone for the soundtrack for the next three months.    To illustrate, here are the last few summer themes:  2001:  Elevation, U2  2002:  We Are Made of Stars, Moby  2003:  Hollywood, Madge  2004:  Let it Go, Frou Frou   '),(23350,2314011,'21,May,2004','2004-05-21 00:00:00','    Yes, yet another new look.  What can I say?  It\'s a time for new things.  Emily started a new blog last night and I caught sight of this template and decided I had to have it.  The WSU-colors were starting to get to me, so you know.    So today\'s the first day of having absolutely nothing to do, nowhere to be, and it feels spectacular.  Slept in until 10, didn\'t get out of bed until 11, watched two episodes of Queer as Folk before even getting in the shower.  Working on my pile of laundry now and pumped out three, count \'em THREE morning pages!  And it\'s only 2!  Think the only time I\'ll leave the abode is to take back the DVD and the rest of the day will be spent puttering around, perhaps cleaning, perhaps not, and starting some new books.  I\'ve been instructed by more than one person to read the Harry Potter series, collected 1 through 4 from Emily last night, and as I find myself in between books, it only seems right to dig in this afternoon.    H\'ok, so...the morning began drenched, soppy and gray.  Now it\'s sunny, blue skies and only marginally warmer.  Still going to watch it from inside.  It\'s such a day to sit, do very little, but relish every moment and every thought; letting everything surround you, settle and swirl inside.  Have so much to bask in and now that I\'m letting myself to precisely that, it feels terrific.  On the second round of the Frou Frou CD, with no sign of stopping.    Suppose I could describe the person more than partially responsible for all this, seeing as how the majority of you readers know nothing other than what has already been posted here.  It may be a bit disjointed and random, but perhaps just right, too.  Hope she doesn\'t mind.  She\'s a hero in the making.  Already is in so many ways and can\'t wait to finish the journey.    She wants to be five years older because she\'s outgrown her contemporaries and so wants to be the person she knows she\'s becoming.    She shares Karen\'s (of the Jack and Karen hour) mentor.    She\'s loyal to a fault.  She laughs easy, loud and often.  She\'s generous with herself, but fiercely protects her vulnerability.  She\'ll make an elaborate meal for her friends and go to bed with a peanut butter sandwich for herself.  She reminds me of someone I used be.    She\'s spontaneous, charming and kind.   She lives in the moment, and not one has been the same since she walked through the Boy\'s door.   '),(23351,2314011,'21,May,2004','2004-05-21 00:00:00','    Cue hand draped across the forehead and longing gaze in lead-actress\' eye, one that appears gooey and romantic, and is, but also has more carnal images dancing behind it.  Then she speaks and people listen:  \"One week ago tonight...\"   '),(23352,2314011,'20,May,2004','2004-05-20 00:00:00','    I know it\'s too much, but I can\'t get enough of this astrology shit lately, just deal with it!!  This is a respectfully-snipped excerpt from my May horoscope...fucking creepy.    \"This month will have its plusses and minuses. What else is new? Always the practical one, a Capricorn gets upset at nothing I say. Bless your heart. Your balanced, practical approach to life is the secret of your success and the envy of your friends. I love your spirit. Don\'t ever change!   Let\'s start at the top of the month and go through each aspect in chronological order.   Yay!  Let\'s do it to it!!   (edit)  One tough day is sure to be May 25 when Mars will meet up with Saturn. They will stand toe-to-toe in your seventh house of committed relationships. Mars will make you ache to be assertive, while Saturn will say, \"Wait! Not yet! Give this time\"! Saturn will win this contest, so you might as well resign yourself to being patient.  Think this has already happened, but okay, whatever.   This matter has to do with a romantic or business partner, and you will feel this aspect gaining in power for the whole second half of the month as it reaches perfection.   I love perfection!   (edit)  At the end of September 2004, Jupiter, the giver of gifts and luck, will enter the zenith of your chart, your tenth house and ruler of your career status and promotion. It will be the first time in twelve years that Jupiter has visited this house.   About damn time!   This is a sensational prospect. Jupiter will remain in this house for thirteen MONTHS! This is the best and biggest aspect you could possibly want for professional advancement. But wait! There is more! In mid-October, five planets - half the solar system - will crowd into your sector of fame and honors, making October 2004 one for the record books.  That October always finds a way.   It\'s important that you don\'t let Venus retrograde get you down in the dumps or distract you from your purpose.  Stupid Venus.  Everything you are doing now will count toward your impending victory in late September or October, so you want to keep your record strong. You have been waiting literally years for this massive opportunity, and you will want to be in a position to take full advantage of it.  Shit damn right!   Venus also rules romance, of course, and with Venus in your sixth house of work projects, you could become smitten with a coworker or a supplier who you come in contact with frequently.  Like my dealer?   In general, there are some key points to consider when Venus goes retrograde. May 17 to June 29 would not be a good time to schedule plastic surgery.  It would not be the time to launch a big party, spend a great deal of money on clothes or jewelry, or go to that la-de-da spa.  Well, shit, what\'s the point?   (edit)  Some of the best news of the month concerns romance. Married or single, on May 19 your love life will get a huge boost by the new moon in Taurus. If you are single, even with Venus retrograde you should see a new romantic, social influence start to dominate your life.  Maybe a little bit!   After a long drought, your love life should start to sparkle and it will become easier for you to meet a new romantic partner. A friend seems willing to make an interesting introduction - allow her ( you mean him? ) to do so.  Shut the fuck up.     (edit)   Speaking about talent, you may show off a wonderful talent yourself this month. This new moon will bring out your creativity in a big way. If you work in a creative field, there will be no limit to what you will be capable of turning out. A new moon brings two weeks of very, very powerful energy, dear Capricorn. It is there for you to use or - if you don\'t take advantage of it - to lose.  Sucker.  I\'m coming evening pages, I promise.       Jupiter, the great benefic planet, has been retrograde since January 3 and will turn direct on May 4. This is sensational news because Jupiter will begin to work very hard for you for the rest of 2004! As a result, exciting travel may come up now or later. (Mid-September appears to be a major time for a journey.)  Journeys rock!   If you want to travel during May, you will have two distinct moments when this would be a fabulous idea:   First, from Saturday, May 15 to Monday, May 17 you may take an unexpected short trip that will turn out splendid.   And it did.     On Friday, May 21 though Sunday, May 23, you may take a long trip. If so, it would work out sensationally well. You will have the action of Mars combined with the blessings of lucky Jupiter to thank for these golden days.   If you choose not to travel, you may find the dates mentioned will stand out to advance educational interests, international commerce, citizenship, relationships, or a publishing or legal matter.  Oh good!  That book I\'ve been writing about a secret community of junior honors English professors who travel abroad and become embroiled in the white slave trade with Irish-American militants while exploring their growing mass attraction for their tour guide/attorney, Hanz, will finally be published this weekend!!  SCORE!   So you see, dear Capricorn, you do have some very dazzling aspects along with some less-than-perfect ones. Your sign is known to have a great sense of humor - the best of any sign in the zodiac. This month, laugh. It will be your best tonic, along with love and affection, which should be in full force this month, too!\"  All in all, a pretty good one, I\'d say.   '),(23353,2314011,'20,May,2004','2004-05-20 00:00:00','    No morning pages today, either, and I\'m feeling their absence, but it won\'t get me down today like it did Wednesday; I have my Darth Maul coffee cup to protect me!  Needed the sleep-time after another late-night conversation.  We love communication, especially when it leads to new places.    And so, we come to Friday.  The big boss is gone and I can already feel a lazy day taking shape.  Been here for more than a few minutes and have yet to start working.  Reading The Boy\'s big blog and talking to the fans mulling through the office takes a while!   '),(23354,2314011,'20,May,2004','2004-05-20 00:00:00','    From Neil:  \"As far as Sandman fanfiction tributes that are also excellent kids\' comics go,  urlLink this  is the bees knees.\"  It\'s quite cute and I love that her name is Ellen.   '),(23355,2314011,'19,May,2004','2004-05-19 00:00:00','    I read a horoscope like this and things don\'t seem so bad...  \"Even if relationships, work responsibilities, or your finances are causing sleepless nights, try to put your mind at rest.  Once you stop expecting the worst, your life will (eventually) fall into perfect place.  You\'re holding an astrological winner\'s ticket promising success and fulfillment, but it will take time and patience before you collect on it.  Don\'t lose heart--instead lose some of the weight on your shoulders.\"  We love you, Vogue!   '),(23356,2314011,'19,May,2004','2004-05-19 00:00:00','    Missing everyone today.    It\'s a combination of the rainy weather (where\'s summer?!), skipping morning pages yesterday (might have to institute an evening pages regimen), hormones (PMS is the devil), and being far away from everyone I want to be with.    Trying to pull myself out of the funk with the Summer Mix, but it\'s just reminding me of good times and the people sharing them.    And really, the whole \"detached\" approach is a bunch of crap and is not working in the least.  Part of me is ready to jump in with both feet and let whatever may happen happen and part of me needs to make sure this doesn\'t get away from me before it even starts.    Sigh.   '),(23357,2314011,'19,May,2004','2004-05-19 00:00:00','    Okay, so...on a run downstairs to investigate some dodgy periodicals, I ran into my Torts professor who is grading our exams.  Right now!!  Let\'s all send her happy thoughts and good vibes, shall we?   '),(23358,2314011,'18,May,2004','2004-05-18 00:00:00','    Phone tag is fun.    Still would like the real thing, though.  Can\'t get the weekend\'s soundtrack out of my head or out of the CD player.  Who knew the ultimate yoga album could also be the best make-out album?   '),(23359,2314011,'17,May,2004','2004-05-17 00:00:00','    It\'s an odd feeling, really, to have what you\'ve been waiting for sitting across from you, giving you a look you\'ve undoubtedly given others in their awkward moments.    And it\'s even more odd to return home to a life that, when you left, was hitting such a peak with nothing but more adventures on the horizon, now seems so disjointed, off-balance and brand new.    But I\'m getting ahead of myself.  During the last few months, I\'ve felt as though I was hurling toward something.  Something intangible, unknown, but unmistakable in its presence or importance.  The feeling became lost in the furor of the last few weeks of class, but re-emerged the more The Boy told me about this girl.  She sounded just right.  Too many damn coincidences  to be, well...coincidence, and since we don\'t really believe in coincidences anyway, they quickly became a big pile of signs.  Signs I couldn\'t wait to investigate, but much too big an idea to contemplate for very long, especially with five exams in two weeks to prepare for.  So, the Weekend After Finals became legendary before it even came to pass; it was the perfect motivation to get through the last stretch, the best subject for daydreams and the subject on most of my friends\' lips the night of the big finale, the eve of my departure.    Had no real clear idea of what I was heading into, only that she\'d won the Boy over, loved everything we love, and sounded exactly like me in more than a few ways.  To a self-proclaimed narcissist, that sounded pretty damn sweet!  I was going to be smooth, calm, confident, devastatingly charming and irresistible.  Then we\'d see what happened.    What happened was absolutely nothing I anticipated and everything I hoped it would be.    She grew on me with every passing moment and I found myself quickly unraveling, quite unexpectedly and uncomfortably.  Saturday night found the three of us at her place with Kill Bill and Queer as Folk.  We all laughed at all the same scenes, finished the same lines and bantered with ad-libs.  Before I knew it, I was cuddling with someone for the first time in a year, and was terrified.  Not terrified enough to stop, mind you; no, it felt entirely too good to ruin with too much thinking.  Eventually, the credits rolled and we had a few moments alone, the moments I was craving and dreading.  Her calm, confident, devastatingly charming demeanor was entirely unnerving and completely comforting.  Here I had precisely what I\'d been wanting and I couldn\'t move or even look her in the eye because when I did, I saw nothing but patience and understanding.  We managed to fashion a memorable goodbye at her door that couldn\'t have been more sweet, only to be followed by my graceless exit.    Now, I don\'t do well with confusion, especially my own.  And while I was relishing the night and how perfectly it ended, I tried to figure out just what the fuck was happening to me.  The last (first) time anything like this happened, I didn\'t think about it, just jumped in with both feet and was the comforting, steadying presence.  And now I didn\'t like being the one in need of comfort and steadying one bit; except for how incredible it felt.    And so, much of Sunday was spent in this place, filtering out my feelings, deciphering the sensations and trying like hell to reclaim my comfort zone, all the while trying to figure out how to reach over and hold her hand as we drove through the rainy forest.    The afternoon found our small gang back at The Boy\'s apartment, he and I telling story after story, each with three stories\' worth of backstory and context-setting.  Found myself soothed by her quiet but solid presence and felt the tension between my shoulders melt away beneath her impromptu massage.  While Levi took a call in the other room, an opportunity to put the entire uncertainty mess to rest presented itself.  And, shockingly enough, the fog was lifted immediately after talking with her about the last day, how utterly discombobulated I\'d become and listening to her say exactly what I needed to hear.  No pressure, no strings, we would go wherever this took us, at the very least becoming great friends in the process.  And as is the way with such talks, everything changed after this one; felt re-connected, more balanced and anxious to explore the feelings and sensations that had me in knots hours before.    The Boy and she decided the only way to end the weekend was with \"Love Actually.\"  They were right.  For the second time in as many days, I found myself suggled beneath her comforter, her front glued to my back, and as much as I loved every scene, the damn movie couldn\'t have been over soon enough.  Eventually, though, we cleared her cozy apartment of its guests, and were left to our own devices at last.    The reel of that time together hasn\'t stopped playing behind my eyes all day and I\'m sitting here in my own cozy apartment feeling disjointed, off-balance and brand new.  But finally, it\'s a good thing and I can\'t wait to see what happens next.   '),(23360,2314011,'15,May,2004','2004-05-15 00:00:00','    Hey Kids...  We\'re coming to you live from Bozeman and the other side of the first year of law school.  It still hasn\'t sunk in yet, but we\'re getting there.  Summer hasn\'t hit Bozeman with full-force, either, but we\'ll make do with the mid-60s.  The Boy and I are hitting the strip mall today, reliving our faux-wedded-bliss experience from Spring Break.  The Pier 1 sales girl really wanted to sell us a dining table and flatware.  Stupid bitch.    So the last exam went well.  It was easily a 3 hour exam crammed into 2 hours; actually, less because it took 30 minutes to even read the whole fact pattern.  But overall, feeling pretty damn good about the entire exam period.  Also received our Appellate Briefs at the end of the Torts exam.  Turned out to be my lowest grade of the three making up our final for the class, but ended up with an A- for the entire year.  And that makes me smile.  The evening that followed wasn\'t so much a huge explosion of celebration and rejoicing as it was a relieving time among friends who had all endured the same adventure.    The drive yesterday felt really good.  Always find such solace and renewal on long drives.  It\'s the music, the endless empty miles, the shifting landscapes and hurling oneself toward some destination as fast as possible.    And so, we come to the big dinner....  The Boy is an excellent cook, there was much rejoicing, laughter and flowing wine.  Miss Vicious lived up to the hype and husky-Jodie-Foster-only-deeper voice.  The pubescent cracking of said voice was particularly cute.  The Boy and I regaled her and friend Sally with stories of the old and not so old times.  They heard about the Day of Noodle-Making over Christmas break and laughed as hard as the Boy and I did when it happened.  At 10, we made our way out into the gritty meanstreets of Bozeman by stopping at the uber-swank wine bar, Thunk.  We probably looked like young riff-raff without any clue, but our hot waitress was hot and didn\'t judge us.  We giggled over the Cherry Explosion selection, wondering if it held a prophecy.  It went down entirely too smooth and too fast, and soon we found ourselves at the bar Levi\'s been talking about for two years, The Starling.    Not knowing what follows wine best, I let Miss Vicious buy me a beer and received a tour of the establishment with this Amazon on my arm.  Eventually, our gang found a quiet corner in which to gasp over the appearance of Daniel elsewhere in the bar.  I had yet to see the new man who had wooed the Boy and then kicked him in the gut, and we three gals counseled Levi on how to handle the situation.  Before long, Vicious set her drink down and said, \"I\'m going down there.  Pull some Vicious action on his ass.\"  And sure enough, within 10 minutes, Dan came \'round, looking to talk to the Boy he\'d ignored all week.    While they went off to see if they could save whatever was growing between them, the new hero on the block and I had more than a few moments to chat.  We started with the usual:  what do you want to be when you grow up, etc., which led into some tales of personal adversity and ultimate triumph, not entirely unlike the first real conversation I had with another Libra in a coffee shop 9 months ago.  As we came around to the optimism and hope that best serves as endings to such tales, the Boy rejoined us and became the focus of our attention.  Their talk had gone well; points made, apologies given and a new start on a friendship ahead of them.  Yay.    After some piled snuggling on a tattered leather couch, we all were fading fast, as were the remnants of V\'s voice.  And so, we made our departures, promising to connect today and see what new adventures we could have together.    This morning\'s been spent in much the same way most mornings in Bozeman are: one of us wakes up first, jumps into bed with the other for an hour of dozing, snuggling and chatting.  Then we\'re on the couch with our breakfast and coffee for at least another hour,chatting some more, looking through new magazines, solving the world\'s problems, etc. before one starts morning pages or blogging, all the while planning the day as though we still have most of it at our disposal.  Finally, it dawns on us that it\'s almost noon and someone should get in the shower if we\'re going to do half of the things we\'ve just spent all morning talking about.    And so, since he just finished with his shower, it would appear to be my turn...   '),(23361,2314011,'13,May,2004','2004-05-13 00:00:00','    So yea, in less than five hours, the first year will be complete and we\'ll all be 2nd years.  It\'s fucking bizarre is what it is.  And I can\'t stop smiling.  Been tidying up all morning, puttering around to get rid of the energy generated by excitement and nerves.  Still have yet to type up the notes from the last day of Torts class, but don\'t feel too bad about it.  Suppose I should look at my notes one last time before the exam, but I\'m having a hard time pulling myself away from the Seattle Mix of \'02 and the pics from the last two weeks\' study sessions.    A year ago at this time I was tidying up in anticipation of the parents\' and grandparents\' arrival for the big graduation.  The weather was pretty similar, except that we\'d actually had snow instead of just rain and chilly winds.  And I think today marks the one-year anniversary of the Morning Pages.  Started them the Thursday before Graduation, although the date in the notebook is a bit off.  Stupid leap-year anyway.  Let\'s just go ahead and say it was a year ago today, m\'kay?  Feels like another life.  But then again, so does studying for the Property final two weeks ago.    There\'s been much moaning and groaning about school, exams, classes and even a couple professors during the last two weeks, much of it completely warranted and justified.  But I\'m still sitting here smiling as I think back on the year.  It\'s an interesting feeling, in this time before the End.  Tomorrow will come and life will go on almost as usual, but for today, an entire experience hangs suspended, about to come to a close.  It\'s a time for celebration and release, but it\'s also a bit sad, as all endings are.    And I always cry at endings.   '),(23362,2314011,'12,May,2004','2004-05-12 00:00:00','    The last study day of the year finds us gals in the Meeting Room of the Daily Grind.  Feeling pretty good about this last exam; it\'s Torts, I like the subject, have a pretty good grip on the material and the exam is only two hours.  The longest question is only half an hour; the rest is a mix of 20 and 10 minute questions.  Yay.   The day after tomorrow at this time will find me on the road, headed to The Boy for a weekend of coffee, hugs, DVDs, and maybe some hard core partying.  Good times.  Finally feels like summer\'s here...REALLY here, and it feels great.  Need to do some standing poses tonight.  As well as tidy up the apartment.  Have shavings from my front door on the kitchen floor to attend to, but it\'s worth it because of the big shiny new deadbolt installed yesterday.  I feel so secure.    The new cut is growing on me (bad pun intended).  It looks a bit different every day, which is both a good thing and kinda disappointing when I can\'t get it back to how it looked that one day when it was almost hot.   '),(23363,2314011,'10,May,2004','2004-05-10 00:00:00','    Okay...I need the old comments.  And I like not being forced into either posting as \"Andrea\" or \"Anonymous.\"  So feel free to use either type of comments, folks; we\'re all about choice here.  And I have no idea what the \"Trackback\" feature is.   '),(23364,2314011,'09,May,2004','2004-05-09 00:00:00','    So, it\'s 10, a good 12 hours since I arose and tried not too hard to get to the study session on time.  Rounded out the Pleadings flashcards around 930 and paused enjoy some fresh brownies and Dreyers Andes Mint Ice Cream; somehow, we can\'t seem to get going again.    Got scared around 3.  The exam was less than 24 hours away and I still didn\'t have a handle on joinder, despite the intense study of the subject yesterday.  Granted, I had other things on my mind (like the significantly reduced amount of hair on it), but still.  A wasted day never feels good, especially the day following it, particularly when that day is the one before an exam.    So yea, winding down...the mind is turning to mush and that\'s as good a reason to call it a night as any.  Soon it will be time for Torts, the grand finale.  Good thing it\'s only a two hour exam.  Feels like we\'ve been in finals mode for a month.  Studying for Property is a distant memory.  We\'re discussing the propositioning of ourselves for the questions to the Civil Procedure exam.  Could get interesting...   '),(23365,2314011,'09,May,2004','2004-05-09 00:00:00','    The new nickname is Hayley after the illustrous Ms. Mills\' stunning performance in the renowned, classic film  urlLink The Parent Trap .    I respect that.   '),(23366,2314011,'09,May,2004','2004-05-09 00:00:00','    Not bad, huh?    Blogger decided to re-vamp its look, I saw the new template on the Birthday Girl\'s blog and decided it was for me.  The comments of old didn\'t make the transition (it would have been tacky and bugged me to no end, trust me...it\'s better this way), but that seems to be the only sacrifice.  The links finally have their place on the side bar and there\'s less white space than usual.  I realize the type is small, and that might receive some attention at some point.  But not tonight.  Have GI Jane on TBS and we\'re quickly becoming engrossed.    Still waiting for the hair to grow out.  The EMP (in Seattle) baseball cap might be making regular appearances.    Seriously, we\'re talking an \'80s aerobics instructor.  I should get my own video.   '),(23367,2314011,'08,May,2004','2004-05-08 00:00:00','    Okay, so there\'s a new template in the making, but there\'s publishing issues...   '),(23368,2314011,'08,May,2004','2004-05-08 00:00:00','    is already half over, but that\'s okay.  My apartment is clean and I might get to read my long lost novel for a bit tonight.  Going to see \"Laws of Attraction\" in a few minutes with the gals; it will probably end up being a guilty pleasure, but seeing Julie be a silly lawyer is just what I need right now.    Speaking of Julie-s, I got my hair cut this morning.  Decided this winter that I would cut my hair way short for the summer.  Not too dykey, but shorter than usual.  And I decided my day off in between stressful finals and a week before the big three-day celebration in Bozeman would be the perfect time.  So, I put myself into the hands of my 50+ fag hairdresser and said: \"Short and cute for summer\"  and let him go.  An hour later, it was the scene out of Steel Magnolias when Julia gets her locks lopped off and makes Dolly Parton cry.  I think it will be fine, once it grows in a bit.  It fucking better.  It\'s not super short, but it\'s much shorter than I anticipated.  The back and sides are hot, I must admit, but from the front, I look like a 12 year-old boy.  Sigh.  Maybe the gals will know what to do.  And this is just a random update from  urlLink Neil\'s site  on the new book:   My gut suspicion at this point -- half way through it, I hope -- is that people who liked American Gods, but didn\'t like Stardust or Neverwhere probably won\'t like it, while people who liked Neverwhere but didn\'t care for American Gods will like it a great deal. (People who liked  Neverwhere  and  American Gods  are on their own.) I don\'t think it will be one of the books that wins awards, but it may be one of those things that some people love more than anything. I could be wrong about all of this, mind you. I set out to write a Thorne Smith novel, and I have no idea how close I\'ll actually come. My editor, Jennifer, laughed at the funny bits, and that felt good.   I loved both  American Gods  and  Neverwhere , so I guess I\'m hosed.   '),(23369,2314011,'05,May,2004','2004-05-05 00:00:00','    The major ones are done.    KICK ASSSSSSS!!!!!  Feeling soooooo good about the Contracts exam.  Came out of the exam last semester feeling confused and violated.  This semester, I was calm, confident and on top of shit.  Once I started, that is.  The morning was split between avoiding practice exams, doing one question and breakfast/lunch with Sadie.  The three hours flew by and I emerged from the building with a huge weight lifted.  Suddenly the next three exams are entirely do-able.  We seem to be taking tonight off; got the Crim Law flashcards out, looked at one, and put them back on the table.  Now we have the Friends special of Dateline going.  It\'s so odd that a night of TV has become a luxury.    But I\'ll take it.   '),(23370,2314011,'04,May,2004','2004-05-04 00:00:00','    We\'re plotting the demise of the top students after having tried a Contracts practice question.    The Property exam was pretty good; think I wrote too much, but I\'d rather err on that side than being too skimpy.    Contracts for the rest of the night and the whole day today.  After tomorrow, we\'ll be on the downhill slope.  Feels like we\'re in the thick of the whole process.  The build-up is finally over and we\'re rolling, but at the same time, there\'s still 4 exams left, the mother of all exams coming up tomorrow.  And then there\'s only 3 left.  Fuck.  Feels like it will never be over.  And yet, there\'s not enough time to get fully prepared.    I realize this is very fragmented and boring.  My peeps are discussing the demise of vowels and commas in their attempts at getting their outlines down to 10 pages.    How did it come to this?   '),(23371,2314011,'03,May,2004','2004-05-03 00:00:00','    Once I\'m in this room, I\'m calm.  Or at least calmer.    Glanced over the (now 31 page) outline and remember that I DO know this shit.    It feels like we were just in here.  Most of us are in the same seats.  Seems like it should be snowing outside and the November mix is in my head.    I have enough room for my group\'s three page flowchart on Real Covenants.  Susan has an index for her 80-page outline.    Sadie forgot her disk drive, but will survive on the school\'s wireless system.    Wonder when the computer guys will get in here and tell us how to do this thing.    Once this exam is done, we\'ll be on our way and there will be no going back.    Still have 27 minutes to change my mind.   '),(23372,2314011,'03,May,2004','2004-05-03 00:00:00','    It\'s Monday, close to 1030, and I\'m watching Ellen.  It must be time for Finals.    I still can\'t quite wrap my mind around the fact that it\'s here.  The end.  The whole weekend felt like the beginning of summer.  Highs in the upper-70s, grilling burgers and hot dogs on Karen\'s back porch, watching the boys chip golf balls, much to the quails\' chagrin.  If it weren\'t for the Property flashcards, it may as well have been the 4th of July.  And the more we did, the better I felt.  This was going to be a good exam.  Sure, no practice exams had been done, no written answers to short questions or the like, but the whiteboard examples and flowchart of greatness made up for it.  Plus, it is impossible to prepare for this professor\'s exam.  He\'s giving us a real estate contract and possible title problems, and our task is to write a title opinion addressed to an a practicing attorney who\'s not versed in Property law.  No pressure.    Thought I would come home from studying last night to spend some time with the Contracts study guide before returning it today.  But no.  That will happen tonight and tomorrow.  Instead, I worked more on the outline and tried to keep my mind occupied so as to not think of what Monday was bringing with it.  Did a nice round of yoga, too, and that always helps.    Chatted with the Boy for a bit and all I want to do this morning is go to Bozeman.  I want it to be the 14th.  I want it all to be over.  Part of me is still in denial about it even being finals because it should only be the beginning of April, and part of me just wants the damn things to hurry up already.  Was doing so well all weekend and now I\'m nervous as all hell.  Between 5 and 7, kept waking up doing Privity of Contract and Privity of Estate analyses.  When I wasn\'t walking through the flow chart of Real Covenants, that is.  Can\'t come up with hypotheticals while studying, but I have more than enough rolling around my head at 537 the morning before the exam, half asleep.    Just have to remember to take a deep breath, go slow, one exam at a time and by the end, there will be a great celebration waiting for me at the end.   '),(23373,2314011,'30,June,2004','2004-06-30 00:00:00','    H\'ok, so...I\'ve made reference to him many a time as one of my favorite authors and fellow-bloggers, but there\'s rarely been an opportunity to share exactly what makes him thus without shoving a book in someone\'s hands.  Such an opportunity arose, however, with a posting on his journal.  It\'s a speech he gave at some function at which he received an award and it encapsulates just what makes Neil Neil and why we can\'t get enough of him.  Yes, he\'s talking about comics; but I like to think he\'s talking about other things, too.     I\'m in the middle of writing a novel currently, and unlike the pleasant social world of comics where, if you\'re me, you talk on a daily basis to editors and artists to letterers or colourists or cover artists, writing a novel is something that\'s done solo. It\'s just me and a lot of pieces of paper. Even my family leaves me alone to write.   This means that when finally offered the opportunity to speak, I\'m liable to begin with apologising for being so out of practice, and then to start blithering unstoppably.  Forgive me if I blither.   (respectful snipping)  ****   When, as a young man, my dream of getting to make comics started to become a reality, I started to meet comics people. These were the people who I had looked up to in my teens, in my twenties, as gods upon the earth. These were the names that I conjured with. I would read everything I could about them when I was growing up, in a time when there was precious little about them to read, and even less of what they had done still in print.  And now I was to meet them.  And I discovered, to my surprise, that quite a lot of them were cranky old jews. Or wannabe cranky old jews â€“ they seemed to be enjoying themselves too much to be properly cranky, and not all of them were actually Jewish.  And now, approaching my mid-forties, eighteen years after writing my first comic, I find myself heading down the conveyor belt towards cranky old Jewhood. I\'m at the age where they start to give you lifetime achievement awards, and you rather wish they wouldn\'t, because it may be some kind of a hint that it\'s time for you to sit down and shut up.  It is the prerogative, however, of those who are one day to be cranky old jews to give advice to the generations that will follow them. And while some of you are my contemporaries, and others are my seniors, I shall advise anyway. My first piece of advice is this:  Ignore all advice.  In my experience, most interesting art gets made by people who don\'t know the rules, and have no idea that certain things simply aren\'t done: so they do them. Transgress. Break things. Have too much fun.    2) Read outside of comics. Learn from places that aren\'t comics. Don\'t do what anyone else is doing. Steal from places that people aren\'t looking. Go outside. Many years ago, when it was almost unheard of for foreigners to write American comics, people used to ask why British Writers were different. I had no idea. I did notice that when I spoke socially to people like Alan Moore, or to Grant Morrison, we mostly weren\'t talking about comics. We were talking about avant garde forms of poetry, about non-fiction writers, about weird things we\'d found. Grant Morrison discovered a long-forgotten Victorian children\'s author named Lucy Clifford, who wound up influencing both his Doom Patrol and, much later, my Coraline. We loved comics, but they weren\'t all we knew. There\'s a whole cool world out there. Use it.  3) Read all the comics you can. Know your comics.   The history of comics is not a long one, and it\'s not unknowable. We can argue about whether or not hieroglyphics were the earliest comics, or the Bayeux tapestry or what. At the end of the day, we don\'t have a long history. You can learn it. You can, these days more easily that you ever could before, study it. And the high points of the last century in comics are quite astonishing.    (edit)   There\'s more classic and important material in print now in affordable editions than there has ever been. Let it inspire you. See how high people have taken the medium in the past, and resolve to take it further.  Isaac Newton, even as he created the foundations of huge swatches of science, said that if he had seen a little further than most men, it was because he was standing on the shoulders of giants.   We\'ve inherited an art-form from giants, some of whom were cranky old Jews, and some of whom weren\'t Jews, and some of whom weren\'t even cranky.  Another piece of advice:  I\'ve learned over the years that everything is more or less the same amount of work, so you may as well set your sights high and try and do something really cool.  There are other people around who can do the mediocre, meat-and-potatoes work that anybody can do. So let them do that. You make the art that only you can make. You tell the stories only you can tell.  As a solution to various problems you may encounter upon the way, let me suggest this:  Make Good Art.  It\'s very simple. But it seems to work. Life fallen apart? Make good art. True love ran off with the milkman? Make good art. Bank foreclosing? Make good art.  Keep moving, learn new skills. Enjoy yourself.   Most of the work I\'ve done that\'s been highly regarded has happened in places where, when I was working on it I tended to suspect that it would go one of two ways â€“ either I was doing something cool that, if I was lucky, people would talk about for some time, or I was doing something that people would have a particularly good laugh about, in the places where they gather to discuss the embarrassing mistakes of those who went before them.  Be proud of your mistakes. Well, proud may not be exactly the right word, but respect them, treasure them, be kind to them, learn from them.  And, more than that, and more important than that, make them.  Make mistakes. Make great mistakes, make wonderful mistakes, make glorious mistakes. Better to make a hundred mistakes than to stare at a blank piece of paper too scared to do anything wrong, too scared to do anything.  Critics will grumble. Of course they will. That\'s one of the functions of critics. As an artist it\'s your job to give them ulcers, and perhaps even something to get apoplectic about.   Most of the things I\'ve got right over the years, I got right because I\'d got them wrong first. It\'s how we make art.   (edit)   And if I have a prediction it\'s simply this: the often-predicted Death of Comics won\'t happen. There will be more booms and there will be more busts. Fads and fashions turn up in comics, as with all things, and, as fads and fashions always do, they end, normally in tears.  But comics is a medium, not a fad. It\'s an art-form, not a fashion. The novel was once so called because it was indeed something novel, but it\'s lasted, and I think, after a few shakedowns, the graphic novel, in whatever form, will do likewise.  Already some things are changing:  When I started writing about comics, before I ever began to write comics, I wanted a world in which comics would simply be regarded as a medium like any other, and in which we were accorded the same respect that any other medium was given. The amount of respect that novels and films and great works of art got. I wanted us to get literary awards. I wanted comics to turn up on the shelves of bookshops, and to sit next to books on the bestseller lists. Maybe one day a comic could come out and be on the NYT bestseller list.  We\'ve got all that. And I don\'t think it\'s important after all.  Right now I actually believe that the best thing about comics may well be that it is a gutter medium. We do not know which fork to use, and we eat with our fingers. We are creators of a medium, we create art in an art-form, which is still alive, which is powerful, which can do things no other medium can do.  I don\'t believe that a fraction of the things that can be done with comics have yet been done.   For now, I think we\'ve barely scratched the surface.  And I think that\'s exciting. I don\'t know where comics as a medium will go in the future. But I want to be amazed, and I\'m pretty sure that I shall be.   And I trust that one day when you, whatever age, race, gender, or ethnicity you may lay claim to, are in your turn a cranky old Jew up here giving a speech, that that will always remain true.   '),(23374,2314011,'29,June,2004','2004-06-29 00:00:00','    And the Seattle weather that wasn\'t there when we were followed us home.  Even though I miss the good ol\' house-rocking T-storms of Poky, Moscow definitely has the apocalyptic-cloud-thing going.  The wind\'s picking up as the day turns dark and I hope I\'ll get through this post in one piece.  Just heating up the remainder of the Boy\'s impromptu Vegetable-Beef-soup-turned-Pasta-Stew and surveying the too-empty apartment.  Should balance the checkbook, put the extra blankets and pillows away, finish arranging the new wall decorations and vacuum.  But I won\'t.  Have just enough energy to eat dinner, put on some new mixes, burn a thank-you CD for our lovely hosts, tell the story of the weekend and call it a night.    So the Boy and I have reached the point where we know what we want to do, where we want to go, etc. in our city.  Every visit feels less like a vacation and more like a return to something very familiar and deeply missed.  The transition time between the initial gasps of child-like wonder and squeals of \"We\'re  here !\" and easing into the city we hope to one day call home is growing shorter.    Thus, we spent the first evening reconnecting.  At first with the skyline, skinny sidestreets and Broadway; later with old friends, even though it was a first meeting for some of us and a much-overdue reunion for others.   urlLink The General  (Laura) and her  urlLink Boy  (Brandon) met us for remnants of our Thai food and  urlLink the soon-to-be Driver  (Jen) followed soon after.  The Boy and I were looking forward to kicking off the festivities with a few laid back beers at the Wild Rose, the dykiest bar west of the Cascades.  But the $8 cover charge changed things and our group headed back down Pike street to the sleek, chic Manray (read: opposite of the Rose).  We found wall-to-wall well-dressed fags and the long-lost gay thump and it felt like coming home.  We bonded over too many drinks, unattainable but hopeful goals (four-somes, being forced to turn down three numbers because of the winner in the back pocket, etc.) for the next three days and fondling strangers.   urlLink Just Jack in the Big Easy  even stopped by for a few minutes.  It felt like old times at Shermer\'s with the Argonaut crew, only this time we were older, wiser, hotter, gayer and on top of our respective games.  And we had the entire weekend (not to mention the rest of our lives) ahead of us.  The Boy and I were left to our own devices Friday and we had new sights to see.  First stop was the new branch of the  urlLink Seattle Public Library .  We watched in awe last summer as the skeleton emerged from the mess of steel and insulation.  We stood speechless this summer inside looking out on the gorgeous, clear summer day shining in through the ceilings and walls, drooling over the slanted stacks with rubber call numbers on the floor, imagining taking Sunday worship in its reading rooms.    Such a temple could only be followed by the Elliot Bay Bookstore, a staple we\'ve somehow managed to overlook until this trip.  It\'s the omega of independently-owned bookstores, something Seattle manages to do well without trying.  It lived up to the hype, with its hardwood floors, walls, and shelves, unending levels, rooms and rows of books, and incredibly eclectic selection.  Leaving emptyhanded felt dirty and wrong, but I was not about to force finding the trip\'s signature book.  Levi found more than enough loot, on the other hand, bringing his total for his trip into the double-digits.  Took a rest on the pier and wished it could be 3 p.m. Friday forever.    Stopped by our favorite comics store in Pike\'s Market before heading back to the hotel to prepare for the big night ahead.  A lazy drive back to Capitol Hill and a stop by the Pussy Bookstore later found us at the Broadway Grill for dinner where new big stories were told and old ones were remembered.  Wanted to give the Rose another chance, but the $2 increase in cover charge killed that idea.    Made it to R Place just in time to escape its $4 cover (suck on that, lezzies!) and began drinking what became a river of vodka while catching up on our gay dance fave videos before heading upstairs to the dancefloor.  It\'s a good thing we had a Vanilla (vodka) Coke and two Vodka Tonics in us each, as they were the only things keeping us going while the DJ played hip hop after rap after R&B song.  Not that there\'s anything wrong with that type of music, but we came for the cheezy, light, peppy, downright GAY music!  All we\'re asking for is some Madonna, people; maybe some Pink and No Doubt on the side, Kylie, Bonnie Tyler, ABBA, something.  And what the fuck is the deal with last call at 120 on Pride Friday?  I\'m seriously, you guys.  But the greasy pepperoni pizza from the closet across the street made up for the night\'s shortcomings.  Oh, and scoring an empty Absolut bottle from a Dumpster on the way back to Harley.  Hmm...still need to get that out of my car.  Started Saturday with coffee and donuts before meeting the party-planning crew for lunch at their favorite Asian-cuisine establishment.  The curry was divine and we swapped Mom Stories as the Immaculate Collection made its second pass on \'repeat\'.  The fivesome decided to split up for the rest of the afternoon, but made arrangements for gay bingo and a barbershop opening later.  Levi and I made it to our last favorite bookstore, where I finally cleaned up...scoring not one but three books to commemorate the occasion.  YAY (times three)! As the coffee shop next door (where God was last spotted) met its demise in our absence, we made do with a new favorite for our routine fondling and petting of the new acquisitions.    Gay Bingo was tossed over for pizza and Trivial Pursuit, which ended in a stale-mate between the genders.  Meeting Wyatt at the Barber opening later completed the Argonaut circle and I couldn\'t help remembering the pained look on his face as he tried to swallow his first shot as he stood before me, shmoozing away and drinking like a hot gay fish.    There\'s just something about sitting on an open patio at the corner of Denny and Bellevue with a perfect gang of favorite people, watching the sun set over the Pacific, telling stories of the hottest and most interesting sex we\'ve had (not necessarily at the same time), with the waxing moon watching over us and the skyline that makes a girl smile.  The night proceeded back to R Place where the music was the same and the patrons were still smashed.    Managed to make it to the girls\' place the next morning for the infamous pre-Pride Parade party and ditched it in time to see the roaring Dykes on Bikes who kicked off said parade.  Some of the novelty and much of the free shit (beads, cool stickers, etc.) had evaporated in the time between this and our first parade, but I still got misty when the PLFAG parents walked by.  And there\'s something to be said for getting a sunburn in Seattle.    The rest of the afternoon found me watching the clock, partly because I knew the longer we stayed meant the later we\'d arrive back in Moscow and mostly because I kept hoping it would say something other than it was time to go.  But the day finally got away from us all and before our three hosts collapsed for the afternoon, they saw us off in our bargain parking lot.  Leaving the city gets more difficult every time.  The pouting and chest tightening starts earlier.  The \"I hate when things are over/I always cry at endings\" feeling sinks deeper.  And then there\'s the trying to get out of the damn place.  For two years in a row now, it hasn\'t let us go without a fight, as though we have the strength to pull away.  This time, I-5 thought it would be fun to hide from us.    Took about half an hour on the road to realize there would be no return to work Monday and the day off was the perfect ending to the adventure.  We spent it on the futon with books and conversation, at the Grind with books and more conversation, and on the futon again with AbFab and Scarlet.  The return to reality was hard, but playing the new mixes helped tremendously, as did a yoga session of long-lost stretches.  Hadn\'t done any poses since Friday night in the back end of the Pussy Bookstore and tonight marks the one-year anniversary of my first class.  Good times.    Now sleep is calling and I don\'t want to pull myself away from the memories of the last few days.  Maybe I\'ll bring my new books to bed with me like I used to do with Curious George and Sesame Street.   '),(23375,2314011,'28,June,2004','2004-06-28 00:00:00','    We\'re back.  We\'re waving.  And we\'re not leaving the apartment, except for developing the film, visiting the Grind and maybe a lunch or dinner with Jane (you wanna?).  But the Boy put it best on his blog and it\'s been a while since I\'ve stolen shamelessly from him, so here we go:   The huge summer adventure is approaching its conclusion. Two weeks full of insight and unexpected meetings, resolved conflicts, new beginnings, followed by the climactic Pride weekend in seattle with the gorgeous gay Idaho crew; it\'s a lot to digest.   Now we\'re at the part of the story where the heroes have achieved their mission, changed forever, and have only to complete the journey home to carry out the lessons they\'ve learned. It\'s the bittersweet section, where you don\'t want it to end, but you know it can\'t go on; the part where you realize soon you\'ll have to deal with the dishes and laundry again. This is the point in every vacation where I wish more than anything that I had one more day...just one!  So that\'s what we\'ve decided, the Girl and I. Leaving Seattle at 5 put us in Moscow at 10. She wasn\'t going to be any good at work on Monday. I would barely be able to move, let alone drive another 7 hours to Bozeman. So why even try? Why not just give ourselves what we both want and deserve: one more day of lounging, recovering, and recharging from our fantastic weekend of gay excess.  It really is the only thing to do, when you think about it. So we sleep in, make coffee, fondle the pile of new books we feel dirty for buying, and catch up on our AbFab. Nice way to wrap it all up, if you ask me.  As for the thrilling adventures, exotic new people, and dangerous car chases...it\'s coming. Just give me a few days to sort it all out.    Yeah.  Ditto.   '),(23376,2314011,'23,June,2004','2004-06-23 00:00:00','    urlLink Class actions  rock.  Especially the huge, history-making types.  Quick!  Who can name the three requirements for class action classification?  THEN, for 100 bonus points, which type of Rule 23 class action suit does this one fall under?  (without looking, mind you.)   '),(23377,2314011,'21,June,2004','2004-06-21 00:00:00','    Most of the important scenes from Saturday night have already been recounted here, but what\'s missing is the context.  Even though I made it a point to come out to the parents, it was never on the agenda for the grandparents.  The reasons are pretty obvious, but chief among them are: they\'re from a different generation who wouldn\'t be, how you say, understanding? tolerant? kind for the sake of appearances? It\'s all I would be to them; everything else about me aside, I would for all time be The Gay.  And as much as I want my last few years with them to be spent in that space, I don\'t.  It\'s basically the same reason I didn\'t come out to my parents until I moved out.  I knew the last year under the same roof would be consumed by the Gay Curse, a struggle they weren\'t going to win and I was unwilling to put us all through.  So, because of that decision, I never felt the need to tell any of my extended family because word spreads like wildfire in this family.  Especially this kind of news about one of the two Golden Children.  Also, the parents have such a time with it themselves that to have anyone else in the family know could put them in the awkward position of discussing it with people they know are sizing them and their kids up, something I\'ve always tried to save them from, even though I know it\'s not my responsibility.    But as we\'ve all grown up and become adults, where the 10-year difference in age means less than it did when I was 14, I\'ve found myself wanting to be friends with some of the cousins.  We\'ve shared bottles of Captain Morgan.  I talked with Michel (the hot husband) for many drunken hours the evening we first met and the next year he made me promise to read the Lord of the Rings...something not even Levi could manage.  Teressa (the bride) and I sat in the corner of Heather\'s wedding shower two years ago, having never really known her my entire life, cackling, poking fun and relating to each other in a way I\'ve never felt with another family member.  Trish (the wonderful wife) has always been a favorite among my immediate family and she\'s been opening up more and more during the last few months about her frustration with her in-laws.  And so, I was looking forward to the after wedding wedding party Saturday night with all of these people.  At last, we would be free to be ourselves without hiding in the back or running after the kids when they whine or turn blue.  And part of me wished The Subject would come up, if only to share something with these folks I wanted to be more than just related to.    But I didn\'t want it coming up at the wedding itself, so I decided against the cropped pants that went with the outfit I planned to wear for the occasion.  Figured I could throw \'em on afterward and the tattoo riding a half-inch below the fabric would go unnoticed...or at least un-commented-on.    My eyes must have dropped onto the table when Rob blurted out:  \"Andrea, do you have a tattoo?\" at his sister\'s (Heather) monstrosity of a new house, with my mom to my left.    \"Yes,\" I answered out of my tight lips, one foot covering up the other ankle.  \"How long have you had that?\" he persisted.  \"Since I was 18.\"    Now DROP IT!!  my tone said.  \"You have a tattoo?\" Trish asked from across the room.  And somewhere, thankfully, the subject was shifted and all I wanted to do was change out of these damn pants, but made do with holding said ankle in my hand.  Later, the folks and I made an appearance at Bride\'s parents\' barbecue.  As I made my way out front to wait for the ride to the piano bar downtown, Aunt Becky (Rob and Heather\'s mother) asked: \"Is that a tattoo around your ankle?\" (Starting to see the connections?)  \"Yup,\" I said, not seeing the connections, just thinking it really  is  that noticeable when it\'s not covered.  \"I haven\'t seen that before, how long have you had that?\"  \"Since I was 18.\"  \"The colors are so pretty.\"  \"Yea, I like \'em.\"  Finally in the van on the way to the evening\'s plans, I felt more out of place than I have in a long time.  Surrounded by 35-ish people who have just bought brand new homes, are starting families without a hair out of place just didn\'t make me feel like me.    But a few laughs, drinks and raucous songs later, things were looking up.  And when Michel \"innocently\" brought up a woman\'s (the one our table had been making fun of for the last 4 songs because she wouldn\'t get off the stage) tattoo, I was ready to have this out.  So, he became the first.  He asked when I knew, how the parents and Joe were, if there was Anybody; said he was proud of me; told me everyone already knew (although I never got around to asking who \'everybody\' is); joked about it in the way we love; promised to be discrete.    Just as I was learning to breathe again, Rob called me over to his end of the table, asked about the symbolism, apologized for earlier, reassured and hugged me in a way he never has before.    Trish interrupted to ask \"What are you two talking about?\" and told me I\'d be joining her for a smoke outside.   Sure, what the hell?   It started with \"So what were you talking to Rob about?\" and quickly became a ranting session about her in-laws.  30 minutes in, she let out a long breath and said: \"I\'m sooo glad you\'re here.\"  \"Really?  I always feel like the little kid.\"  \"NO!  Don\'t feel like that.  I love your family!  You guys are so real!\"  And before long, I was struck by the fact that here I was again, being the listening, reassuring one, bonding with someone over  her  troubles, when the whole purpose for our time together was for the opposite.    She seemed to realize this all of a sudden, saddled her body up next to mine and said: \"So, what\'s going on with  you ?\" with a smile that told me she knew exactly what my answer would be.  \"Oh, I think you already know.\"  She just looked at me for a second before saying: \"Why are you so quiet about it?\"  Cue update on the Parent Situation.  She never skipped a beat and was nothing but the calm, sweet, wonderful woman I\'ve loved since we met, even though she was incredibly drunk and more loose than I\'d ever seen her...which actually just made me like her all the more.  \"Is it just your parents?\"  \"Well, no...it\'s the grandparents, too...\"  \"I KNEW IT!!!\"   \"Yea, that one\'s not hard to figure out.\"  \"If it were me, I wouldn\'t be so quiet..but then, I\'m a rebel.\" which she\'s totally not, but okay, we\'ll go with that for the moment.  \"It\'s all I would be to them.  If I had someone, that would be a different story...\"  \"You would say something then?\"  \"Absolutely; if I had someone I wanted to bring home...\"  \"Okay, good.\"  \"But I don\'t.  And there\'s no reason for me to be out and proud for the sake of it when \'that\' would consume the last years I have with them.\"  \"Yea..that\'s true.\"  \"Hey you two!  The party\'s  inside !!\"  Teressa said as she burst through the doors.    The three of us went on a hunt for more cigarettes and nary another word was spoken about the issue.  Rob and Trish walked me to the hotel an hour later and I re-lived the evening all night in my half-dreams.    The drive home was consumed by my hangover, the intense sensitivity that accompanies said hangovers, the enormity of the evening before and knowing I had to tell the parents the cousins knew, no matter what they would say.  It could have been anything, really.  From Mom\'s patented \"You will be the death of me, child\" sigh, to outrage, to anxiety and \"What are we supposed to do/say NOW?  I\'m sure the whole family knows!!!\"  But anything would be better than hearing \"Guess what Becky said to me...\"  And so, I ended up having the best conversation about the whole thing Mom and I have ever had.  Spent the night on the futon, curled up with a stress-victim stomach, talking with Miss Viscious, attaching words to what the last few weeks have said loud and clear.    The haze of the whole thing has lifted slowly over the course of the day, not coincidentally along the same timeline as the release of this month\'s hormones.  At least I know I won\'t have to worry about getting pregnant at Pride.    As the nerves calm, and the raw, exposed sensation fades, I\'m left feeling stronger, closer to my family and shaking my head at the band of color around my left ankle.   '),(23378,2314011,'20,June,2004','2004-06-20 00:00:00','    It all began innocently enough.  I was going to the Emerald Club for the first time.  Ever.  It\'s been the scene of many a fabled story for years and I finally had enough I don\'t know what to leave Mom in the hotel room at 10 and make my way to the drinking establishment alone, looking for a story and maybe a familiar face.  As I walked in, I quickly scanned the faces for one I recognized.  When I found none, I wanted to leave; but, I wanted to go back to the room even less.  The thing is, this place has always been described as this hip-happenin\' dance club where all the young things go to be their gay-fabulous-selves.  It\'s actually just a step up from Charley\'s.  It reminded me a bit of  urlLink Woody\'s , though, and that kept me from gulping my Heffy and bolting.   People watching is one of my favorite past-times.  Folks started trickling in around 1030, which kept my attention busy between watching the door, bartender, regulars and the pool game happening over my shoulder.  After such a sweep over the premises and reach back for my beer, I noticed a missed call on the cell phone.    \"Viscious\' Cell\"  How appropriate.  The weekend she knows I\'m out of town on assignment with the family and immediately after I\'ve posted my \"It\'s over (even though it never began) and I\'m okay with that\" entry.  I celebrated the absolute lack of desire, will, pull, etc. to return the call with a swig and gaze at the slop-shot happening to my right.    But I knew I\'d have to listen to the message.  Just wanted to let me know she\'s alive, that her hibernation had nothing to do with me and she wants to talk to me about the conclusions she\'s reached in her self-imposed isolation.  Deciding I was having too good a time at my post and too close to talking to the girl who reminded me too much of another crush, I left the bathroom knowing V. would wait until Sunday to tell me (or hear) what I\'ve known for a month.  Before long, I was recruited to join the pool party and couldn\'t help but think of how proud the Boy would be of me for being so social and charming.  I followed Chip (honest to god, that\'s the woman\'s name), the center of the group, to the table where I met her pregnant wife, Jen (aka Stands with Arms Folded), their friend (the one who reminds me of a young library worker of old), Christine, and the father of their child, Monte.  Away the next two hours flew with game after game and the arrival of more of my hosts\' friends.  It was so nice to play with amateurs again, make impossible shots without trying, winning close games and laughing with faggy strangers.  Small talk was surprisingly easy and comfortable.  The more I was around this gang, the more I realized I could have become the newest member, if only I lived in town.  But for the moment, they cooed over my use of \"concur\" and \"innocuous\" and Yay-ed my bank shots.    By 1230, Chip turned to me from a conversation with her spent wife and said, \"You up for some pork fried rice?\"  \"Absolutely.\"  Rode over with Monte in his big Cadillac and loved that I told Mom I\'d be back by midnight.  Once we congregated at the Twin Dragon, the ladies chastised me about getting into a car with a stranger and Emmett...I mean, Monte joined in:  \"I couldn\'t believe she got in just like that!  But I thought I better not say anything about it while we drove around.\"  Seriously, the guy is big and smart like  urlLink Ben , has  urlLink Emmett\'s  voice, mannerisms and stunningly sharp wit, and  urlLink Michael\'s  adorable \"I\'m going to be a Dad\" thing.    \"Really,\" Chip got serious.  \"It\'s great that you\'re so trusting, but...\"  \"You guys, if I\'d had any inkling that it wasn\'t absolutely okay, I wouldn\'t have come with you.  Honestly.  Besides, I had my cell phone. I could have snuck a call to 911 if it came to that.\"  \"Good call,\" my peeps for a night agreed.  \"So, what are we all getting?\"  And so, while we passed around pork and seeds, fried rice and pot-stickers, I listened to Chip rattle off the common sayings she\'s writing about for an English class (since she\'s been with Jen, she\'s cleaning up her act) and watched Monte\'s eyes dance as he told the stories behind many of them.    \"Water-logged\" really does originate from logging operations on rivers.  \"Stealing from Peter to pay Paul\" is a British cathedral thing.  We laughed about Chip\'s germ-phobia and how, as a result, Stands With Arms Folded gets hosed when it comes to oral sex.    Gave the ladies equal time by letting them drive me to the hotel.  \"I\'m glad you played pool with us.\" Chip said.  \"I\'m glad you asked me.\"  \"I hung out at bars by myself when I lived in California and it was really hard, so I try to include people I see...\"  \"That\'s really great of you, but I honestly expected to people-watch for a couple of hours.\"  \"I commend you for that...it\'s really brave.\"  \"But I\'m still glad you invited me to play with you.\"  Ended up getting her business card out of the deal (she\'s a real estate agent).  Not to mention a few great memories and a lot of story fodder.  And thus concludes the first leg of the weekend.  I thought it would be the one gay night of the trip and the ultimate opening act for Pride.  We\'ll get to the business of proving me wrong on both counts after I\'ve had some chicken noodle soup.   '),(23379,2314011,'20,June,2004','2004-06-20 00:00:00','    Guess this is what you get when you win a huge lawsuit against the CD industry...payback.     urlLink \"We\'ve been wondering if we\'re going to get 12,000 Yanni CDs,\" said Wallace Hoffsis, director of collections development for the Sacramento (Calif.) Public Library.   And I  will  get around to posting the novel about the weekend, I promise.   '),(23380,2314011,'20,June,2004','2004-06-20 00:00:00','    I have no energy to tell this story the way it should be told, but I can\'t let an evening pass without  something  written about it, either.  Just got home from the big wedding weekend and each day had its own wonders and terrors.  Here\'s a brief sampling of the trip:  Friday night:  Maiden trip to the Emerald Club where I people-watched, accidentally avoided a fateful phone call, met three dykes and Emmett from QAF (actually, he\'s like 78% Emmett, 17% Ben and 5% Michael).    Saturday:  wedding was beautiful; cousin Rob asked about my tattoo back at his sister\'s place, where I immediately clammed up and shot him \"Shut the fuck up\" vibes; Aunt Becky (his mother) suddenly noticed the very same tattoo a matter of hours later; cut to few more hours later and a vanilla-rum-and-coke-induced conversation with cousin Heather\'s (Rob\'s sister) hot husband Michel:    \"You have a tattoo?\"  \"Yes, I do.\"  \"Can I see it?\"  ......  \"How\'s  that  going for you?\"  \"Everybody knows, don\'t they?\"  ....  \"Well....yea.  How are your parents with it?\"  \"They struggle, but they\'re okay.\"  Cue huge hug.  \"I think you\'re very brave and I\'m proud of you.\"  Cut to Cousin Rob:  \"Does your tattoo have a symbolic meaning?\"  Long pause.   Will this never end??   \"You can just say \'yes\'.\"  \"Yes.\"  \"I didn\'t mean to embarrass you at the house earlier...\"  \"I know, Rob.\"  \"You don\'t have to worry about us.\"  Cue another huge hug.  Cut to Rob\'s wonderful wife Trish:  \"Come outside and smoke with me...\"   Okay, so...long story short...the cousins know.    Cut to 20 minutes and a nervous breakdown ago:  \"Quite a weekend.\" Mom said when I called to let her know I\'d returned safely.  \"Yup.  Feeling pretty zapped.\"  \"Party!\"  \"Speaking of that....the cousins know.\"  \"Oh.\"  \"They asked and I had to tell them.\"  \"Okay.\"  \"I thought about it all the way home and decided it would be better for you to hear it from me than be blindsided by it.\"  \"I\'m not worried...if it comes up, it comes up.\"  \"I didn\'t plan for it to happen this way, and I didn\'t mean to instigate anything...\"  \"I know you didn\'t.  And really, I thought you were being paranoid about wearing long pants to the wedding, but when people kept bringing it up afterward....I think it\'s been discussed...\"  \"It has been, and I think this was their way of saying \'We know and it\'s okay\'.\"  \"I think so too.  So don\'t lose sleep over it.  Hopefully no one will say anything to Grandma and Grandpa.  But if they do, we\'ll just deal with it and it\'ll be okay.  I know what to say if anyone says anything to be about it.  It doesn\'t change who you are, we still love you, we\'re still proud of you, you\'re still doing great things, we\'re just going to deal with it because there\'s no other choice.\"  Cue crying session number infinity.  \"You didn\'t think I was this far along, did you?\"  \"I knew you\'d get there eventually, I just didn\'t know when eventually would come.\"  \"Well, I am.  All that worrying for nothing...\"  \"I don\'t know what to say.\"    And I still don\'t.  So I\'m going to take a shower, have some dinner, and process the last three days before I sit down to write the whole story (oh yes, there\'s more...).  Stay tuned.   '),(23381,2314011,'17,June,2004','2004-06-17 00:00:00','    Had an odd but great dream last night.  Don\'t know where I was, exactly, but I suddenly found myself walking by the office of the Aspen Publishing company.  They\'re big on law textbooks and \"study aids\".  They\'re the only publisher who does 1st-year-subject flashcards and those are pretty hit or miss.  Typos, answers to questions never asked in the fact-patterns and flat-out wrong black-letter law cards.  So, whenever The Stellar Super Students got hosed on an answer we KNEW was right, we took great joy in dismissing the card with the obligatory \"Fuck Aspen!\" attached.    So, as I found myself so close to the puppet-masters behind the travesty, I was compelled to share my frustration.  Walked in to an empty office space that closely resembled an overdone ski lodge.  High ceilings, too much decorative wood, etc.  Kinda like The Stranger\'s office.  Nosed around until I found three gals in tank tops, jeans and tool belts (I\'m seriously, you guys) working on more decorative wood, only in here, it was 2x4s.  Knew in an instant they were all \"family\" and one, the leader, took my breath away the instant our eyes met.    But I would not be deterred.    I walked right into that room and declared:  \"Fuck Aspen!\"  No introduction; no \"May I talk to your flashcard department?\"; nothing.  They were a bit surprised, but let me go on.  I explained (and by \'explained\', I mean \'ranted incessantly about\') the problems with said flashcards while they listened attentively.  Finally, as I said my peace and made my way to leave (because really, how was I going to transition well out of that?), all three spoke up at once: \"No! Stay!  We like you!  You wanna help us revise those?\"  And so on.  I was instantly a member of the group and before long, we made our way to the kitchen.  Yes, there was a kitchen in the office..whatev.  All the while, nothing was really said between me and the hottest of the three worker-lesbians, but there was definitely some vibes exchanged.  As the other two and I shared some Better Than Sex chocolate cake and laughed like old friends, She came up behind me, wrapped her arms around my waist and I woke up.  What do you think it  means ??   '),(23382,2314011,'16,June,2004','2004-06-16 00:00:00','    No Intro classes to be found after 645 am.  Fine.  I\'ll just take my handy-dandy refund and release all those I tried not-so-hard to convince to try out the body-bending experience.  Won\'t be as sculpted, but at least there\'s a bit more padding in the bank account.  Sweeeet.   '),(23383,2314011,'15,June,2004','2004-06-15 00:00:00','    H\'ok, so...have 16 yoga classes to cram into 9 days.  7 of those are make-ups for the Seattle and July vacations.  The remaining ones are just regularly scheduled classes.  Yay.  So, until July 1, I\'ll have a yoga class every night, and sometimes two.    I\'m going to be so damn toned, it won\'t even be funny.   '),(23384,2314011,'15,June,2004','2004-06-15 00:00:00','    Your mom\'s got nothing on mine!!  For all the stress and trauma she causes me, my mom kicks ass.  Really.  One of the things I dig about her is her inability to go simple, small, understated.    When it appeared I needed a new dresser, bed-side tables, etc. in junior high, she decided I needed a full bedroom set of the finest quality, you know, to have when I move out and get my own house.    When I got my apartment, I planned to fill it with cheap, tacky furniture I could throw away at the end of my tenure, and shopped accordingly before the parents arrived to help in the venture.  By the time they left, I had a brand new table, futon and desk, because really, this was to be my home for (at least) the next two years.    When I asked for a DVD player for Christmas a couple of years ago, she decided I needed a 5-disc changer with 5 speakers and a subwoofer, you know, to have when I get my own house.  She stresses about vacations especially well.  Take an email conversation from this morning:  From: Cheryl    Sent:  Wednesday, June 16, 2004 9:07 am  To:  Andrea  Subject:  How does Mexico sound?   I just don\'t know what to think about Orlando....too hot and too long of lines.  How does Mexico sound?  From:   Andrea   Sent:  Wednesday, June 16, 2004 9:09 am  To:  Cheryl    Subject:  Re: How does Mexico sound?   WOOOOO HOOOOOOO!!!!!  (it sounds okay) :-)  From:  Cheryl  Sent:  Wednesday, June 16, 2004 9:11 am  To:  Andrea Subject: Re: How does Mexico sound?   Does that sound better to you?   From:   Andrea   Sent:  Wednesday, June 16, 2004 9:14 am  To:  Cheryl  Subject:  Re: How does Mexico sound?   either one would be fine, really...i\'m okay with whatever, but Mexico sounds really good...as long as i don\'t have to see any actual Mexicans, that is.   From:   Cheryl    Sent:  Wednesday, June 16, 2004 9:17 am  To:  Andrea Subject:  Re: How does Mexico sound?   A bus load of mexicans is almost as bad as a bus full of asians it\'s true!    Sigh...could I love her more?   '),(23385,2314011,'15,June,2004','2004-06-15 00:00:00','    Just finishing the \"writing\" stage of burning the new mix.  It\'s so exciting!  This is the first mix I\'ve constructed all by myself since last year\'s October Mix.  I realized, after a few listens, that I may want to start listening to the track list a few times, tinkering with songs and such before actually calling it good.  Never was one who came by revisions naturally.  And so, for months, I resigned myself to the fact that I\'m not a good mixer.  Decided to leave that to The Boy and just help with brainstorming possibilities.    But, as promised, that not-too-forgotten pang of creation, the one that\'s been rolling around in the back of my mind for over a month, finally took hold; just not in the form I was expecting.    But hey, I\'ll take it.  And now, it\'s in Sam on its maiden playing.  iTunes doesn\'t mess around, I tell ya what!  3 minutes, tops.  Jeez!    Madge\'s \"Intervention\" didn\'t make the cut, even though I was dead-set on it yesterday.  But as Stephen King once wrote, we sometimes have to kill our favorite babies for the sake of greatness.  It still doesn\'t have a title, which is a bit odd, but don\'t think it\'ll stay that way for long.  Any suggestions?   '),(23386,2314011,'14,June,2004','2004-06-14 00:00:00','    So it\'s probably just payback for last week\'s Harry Potter debacle, but I\'m really not understanding all the flack \"The Stepford Wives\" has received.  I thought it was great.  Sure, there were flaws: it could have been darker; it could have been sharper; Matthew Broderick could have been left out.  But over all, it was pretty damn good.  Maybe it\'s because I have yet to see the original and didn\'t have any of those nasty expectations going in, other than how much I knew I\'d love it.  The critics seem to be so mad about its campiness and superficiality.  I\'m sorry, but I thought that was the whole point.  Camp works.  Okay, so it\'s not a probing, hard-hitting social commentary/satire.  But there are some great fucking lines and Glenn Close is phenomenal as the perfect evil bitch.  The story\'s a bit weak, but the trio of Nicole, Bette and the gay boy make me want to move to Stepford and join their anti-cult.  I\'m sure I\'ll hate it the second time I see it, but for now, I\'m left coveting Nicole\'s great hair.  Sigh, soon.    In other news...  It\'s been a long-standing joke/theme/infuriatingly perfect/creepy thing that the Boy and I share parallel experiences.  Always have.  Even way back when we didn\'t even realize it.  It\'s not that the exact same thing happens to each of us at the exact same time, or that circumstances that plague one inevitably plague the other.  It\'s more about the more-than-a-crush we each had on our best friends in 9th grade; both loving the person for all the wrong reasons and hating them because we loved them and having no idea we were loving them.  It\'s the hugely epic first loves experienced across an impassable distance.  It\'s the knowing each respective relationship would end and knowing we would be the ones to end it.  It\'s the revolving obsessions.    Our timing is improving.  We\'ve been sharing similar experiences in closer proximity, which is great for girl-talk value and for being available when one needs to hear what s/he just told the other not more than two weeks ago.  And so, when a new boy came into The Boy\'s life and showed promise, a part of me rejoiced for my own chances with this new girl I\'d been hearing so much about.  When it appeared his promise would not be fulfilled, a part of me got worried.    And sure enough, my story shares the same highs, lows and ending.  Granted, the plot points wore different clothing, but underneath, they\'re the same.  To the letter.  There\'s not much to tell by way of details, which is basically the whole point.  There\'s nothing to tell because there\'s nothing to tell.  Nothing\'s happened; nothing\'s moved anywhere.  And while it hurt in the beginning, the wanting to find out what happens next, to become friends and see where it took us only to be met with silence or promises for later, it doesn\'t hurt anymore.  We haven\'t talked about anything; nothing\'s been resolved (or brought up, for that matter); we\'re no closer to becoming friends than we were a month ago.  Nothing\'s ended because it never began.  But none of it bothers me anymore.  I did my part.  I gave it a try and kept up my end as best I could.  There\'s nothing to regret, really, because nothing ever got to that point.  The whole \"in different places\" thing, while known from the beginning, becomes more and more apparent with every passing week, and it\'s time to put the whole thing on the shelf of good memories.    So, the Boy and I find ourselves yet again in the same place.  At the end of another short story, ready to find the first line of the next great tale.  Seattle seems like a good place to go looking.   '),(23387,2314011,'13,June,2004','2004-06-13 00:00:00','    So now when Mom asks me, as she\'s been doing every once in a while for 6 months, I can say I\'ve tried it and it\'s good, just like the package says.  It\'s not the same as the original, from the box, cooked on the stove Mac & Cheese, but it\'ll do on Sunday afternoon spent at the office.    Just compensating for this Friday\'s absence, which will be spent in Boise celebrating the semester\'s victories with Mom before the big wedding on Saturday.  Should be quite the event.  Have a cousin\'s new baby to ohh and ahh over; have another cousin\'s new pregnancy to pretend to be excited about; Grandma and Grandpa will be missing on account of the latter\'s recent open-heart surgery; and the after-wedding party with the newlyweds at a dueling piano bar.  Score.    So yea, just me in the big ol\' empty office.  Brought my laptop along so I could download the iTunes program using the school\'s wi-fi.  Decided yesterday that I wanted a new Summer Mix.  Something other than the fabulously brand new Mixes the Boy is cooking up.  This will be a retrospective: Summer\'s Greatest Hits.  There\'s some hugely guilty pleasures, some classics, some new faves and some long-forgotten yet greatly beloved archetypes.  Accordingly, some songs had to be hunted down on the internet and (sigh) paid for.  Can\'t wait to get home, put it together and give it its first listen doing some long-overdue yoga.    After watching Game 4, that is.  Go Pistons.   '),(23388,2314011,'11,June,2004','2004-06-11 00:00:00','    I kick ass.  Now, before I post this, it should be noted that I\'m not doing so in order to gloat or rub anything in anyone\'s face.  It\'s just that I\'m really quite excited about the news and have to share the excitement.  Yay.    The law school released the class standings this afternoon, and here\'s where this gal fell.  We\'ll start with a bit of backstory, as usual:  Fall Semester GPA: 2.83 Fall Semester standing: 41/98 Fall Semester section standing: 21/48  Now let\'s get to the best fucking news ever:  Spring Semester GPA: 3.45 Spring Semester Standing: 5/98 (the whole class) Spring Semester Section Standing: 4/48 (just my section [half the class])  And move to the big picture:  Cumulative GPA: 3.19 Cumulative Standing: 18/98 Cumulative Section Standing: 8/48   Sigh.  I still can\'t quite believe it, but I also can\'t seem to get this grin off my face.  Nice way to kick off the weekend.   '),(23389,2314011,'10,June,2004','2004-06-10 00:00:00','    There\'s an ass-crack on the cover of Food Processing.    \"What\'s Food Processing?\" you ask?  It\'s a small-scale monthly publication that only librarians and lonely scientists see.  It\'s not sexy.  It\'s not flashy.  It probably has a print run of 5000.  There\'s no need for sex appeal.  Really.    The cover image is actually kind of clever.  It\'s a woman (of course) holding celery behind her back to catch the salt she\'s throwing over her shoulder and the article it references has to do with \"sodium sinners.\"  Okay.  That\'s fine.  But why does she have to be naked from the ass up?  Her pants are half-way down her ass, for that matter.    My boss was half-way across the room when she stopped, did an about-face and came back to inspect the cover.  We were thoroughly outraged.  Then Cort, the resident geezer, stopped by and I thrust the magazine into his hand and said, \"Look at how NOT appropriate this cover is!\"  He looked over it carefully and said:  \"What\'s wrong with celery?\"   '),(23390,2314011,'08,June,2004','2004-06-08 00:00:00','    I love blogging.  And I really love having nothing to blog about, and then something great happens and suddenly, there\'s a perfect story for a blog.    Case in point:  The library Office Goddess is one of the best parts of working here.  She\'s head secretary to the Dean and holder of the key to the supply room.  She\'s from Texas and has the best drawl I\'ve ever heard and the dry, no-nonsense, shit-kicker sense of humor to match.  She\'s larger than life.  Literally.    She approached me from behind a few minutes ago as I carefully double-checked the 2004 invoice line item by line item, tapping my shoulder with a packet of Post-its.    \"Can I trade a coffee filter for this stack of Menopause Purple Post-its?\"  I looked, and sure enough, she held a stack of wonderfully lavender-hued (insert clever description of Post-its without using Post-its again here).  \"Menopause purple?\"  I said around my guffaw.  \"Both my husband\'s and my mothers painted their rooms this color when they got older, so we call it Menopause Purple.\"  \"It\'s also Lesbian Pride Purple.  Now, I don\'t actually have jurisdiction over the coffee filters, but I say go ahead.\"  I said as she went rummaging through the nearby drawers for said filter.  \"You know, Mary Jane, your stunning outfit kinda matches this color.\"  \"I know, I\'m starting to like that color and it\'s worrying me.\"    \"Could be a good barometer...\"  (cut to insignificant, continuity-ruining story about a lesbian cousin)  \"Did SHE paint her room this color?\"   \"No, I don\'t think she painted anything.\"  \"Not much of a lesbian, then.\"  \"Well, she was in Eastern Texas, she probably didn\'t know how to be one.\"  (cut to Mary Jane showing off her rainbow-colored socks)  \"Those are pretty fancy,\" I encouraged, already planning this entry.  \"We all go through changes...\" she said as she made her way back to her throne, two filters (cone-shaped and flat-bottom) in hand.   '),(23391,2314011,'08,June,2004','2004-06-08 00:00:00','    Okay, this deserves a post all its own.  I knew I was starting shit with my previous post, but meant it all in good fun.  Now I realize I\'ve ruffled some feathers I\'d never intended to ruffle and should try my best to alleviate the situation as best I can.  I\'m reminded of my reaction to a friend\'s condescending attitude toward Middle Earth and all its beloved creatures around Oscar Time and want to apologize for donning a similar attitude here.  It was only meant in fun, I promise.  Now, I must stress that I AM a fan of the Harry Potter books.  Yes, like the Lord of the Rings Trilogy, I\'ve been proven wrong by making like the dude in Green Eggs and Ham and just giving the books a try.  They\'re good, quality stuff.  There\'s no getting around it.  And I know I\'ll just come to love them more as I continue to complete them.  My previous post was never meant to besmirch the HP series itself or make excuses for my new-found fandom.  I simply am seeing similarities and chose to comment on them by making reference to my previously-held beliefs.  Yes, I\'ve been proven wrong, and on some levels, I\'ve been proven kind of right.    So now, all I can ask is your forgiveness for shitting on something sacred.  Send me to the Forbidden Forest.  Give me an extra Potions assignment.  But please, don\'t let this incident of thoughtlessness and \"your world, my world\" observations destroy what could become a moment of learning and...hopefully, winning.   '),(23392,2314011,'07,June,2004','2004-06-07 00:00:00','    Okay, I just have to say...I broke down and started reading the Harry Potter books this summer.  They\'ve been perfect for the occasion.  Yes, I resisted as long as I could because I thought Ms. Rowling was a shameless hack of an author who pieced together her story using the best bits of all the great classic epics of our time.  I was determined to resist the furor and hype with every dignified elitist bone in my body.  I was Eric Cartman in the Lord of the Rings episode where the boys are traipsing off to return the One Video Tape to The Video Store and come across a band of fellow youngsters.  \"I shall put a spell on you!\" said one in a fake British accent.  \"What are you kids doing?\"  one of the four asked.  \"We\'re playing Harry Potter!\"  \"HA!!\" Eric laughed condescendingly.  \"FAGS!\"  But the universe saw fit to let me read the series and so I have.  And for the most part, I\'ve been pleasantly surprised.  I\'m about 200 pages into the third book and I\'m finally getting invested in more than just the novelty (yes, I said novelty) of it.  But the more I read, the more I realize I was right:  she is a shameless hack.    Dobby IS Gollum.  The Dementors ARE the Ringwraiths.  Dumbledore IS Gandalf AND Obi-Wan (from the original trilogy).  But I\'m still hooked.  And I still can\'t wait to get home and keep reading.  Stupid hacks, anyway.   '),(23393,2314011,'07,June,2004','2004-06-07 00:00:00','    One thing I miss from southern Idaho are the thunderstorms.  I\'m always reminded during Moscow summers when a small rumbling that sounds like the upstairs neighbors\' dishwasher rolls over the...well, rolling hills of the Palouse.  Always wish we could have a good storm like we used to have once in a while in Poky.  The kind that woke me out of a deep slumber in my cave in the basement and shook the entire house.    Barely beat the \"storm\" this afternoon, walking home from work.  But the lightning, when it appears, puts on a pretty good show up here.  Remember lying awake one summer night two years ago, losing yet another battle with insomnia, watching a storm form and turn into a great light show.  Took some comfort in the knowledge that even had I gone to sleep hours before, when I went to bed, I wouldn\'t have slept through this.  The next morning in my American TV Genres summer school class, my professor yawned and said, \"Yes, I was up last night, watching the lightening with my cats.\"    And this afternoon, the apartment did a slight tremble after the bright flash slashed through the dark clouds, all of which was impressive, for Moscow.  But now, three hours and a cheap movie (\"Mean Girls\" with Calamity Jane; we love you, Tina Fey...keep making non-SNL-character movies, please) later, the skies are clear and the hills are just a bit greener.   '),(23394,2314011,'05,June,2004','2004-06-05 00:00:00','    When I saw the 5-day forecast early Tuesday morning over my bowl of cereal, telling us that Saturday would see 80 degree sunshine and Sunday would bring rain and a plummet to 65, I said \"Goin\' to Elk River Saturday\" around my Honey Bunches of Oats.    The weather was ahead of schedule, proving to be absolutely gorgeous all week, with a slight drop this morning.  But I would not be deterred.  Actually, it probably worked out for the best.  80 degree heat is really not the best to be trekking up and down steep forest floors, after all, and this particular trail always surprises me with its difficulty.  But I bought provisions yesterday and was determined not to get caught huffing and puffing without any trail mix!  And so, bright and early this (late) morning, I packed up my backpack full of water jugs, snacks, Superman notebook, Harry Potter novel and down the road I went.    Outran the rainclouds just fast enough to enjoy a lovely little meander to the falls, enjoyed a full 20 minutes of warm/hot sunshine and made it through a full chapter of The Chamber of Secrets before the rain started.  Even managed to get a little Henry David action on the side, which I thought I\'d share here.  It\'s really rough around the edges, but it works:   Hey kids!  Welcome to Elk River!  We\'re at the Upper Falls lookout at the moment and it looks as though the spring rains have been beneficial.  My other spot--the one I planned to spend the bulk of the time at--may well be under water, so I\'m hanging out here a bit longer than I originally thought.  Brought the old automatic camera I found in the old boxes last weekend--hope the roll of film is properly loaded--can\'t really risk looking now can I?  Looks like the clouds I outran to get here are catching up.  Good thing I changed my top and brought along the jacket!    It\'s nice to be here today--just the thing to celebrate a full week of nightly yoga practices and to remember the day from last year.  Good ol\' June 5th, I tell ya what!  A story just might come out of my nature wanderings by the end of the day, I just dunno!   Yea, I\'ve been feeling that old familiar pang of creation the last little while.  It\'s pretty amorphous and no clear idea has taken shape just yet, but the new reading list and increased \"pages\" time have been helping.  But anyway, back to the day:   Alrighty, here we are at my favorite little spot.  Yes, the water has risen quite a bit.  The places Levi and I dangled our feet and talked about Queer As Folk are under water.  The sun is trying to conquer to overcast sky and it feels really good for a minute.  But then the clouds overcome it and it\'s a bit cooler again.  But that feels really good, too.  There\'s nothing like the smell of a river during the summer.  It would be a bit keener if it was warmer, but it\'s just right for right now.  Making do with rolling up the sleeves to get some sun on the upper-arms and shoulders.  Looks like the sun may be conducive for a bit, too.  Yay.    So glad the day is living up to the idea formed last week.  Was expecting the 75-80 degrees promised, but the mild day is much better.  So glad I brought provisions.    The birds are chatting it up now that the sun\'s out.  Already found a little souvenir for the day---little rock waiting on a bigger rock that fits my fingers perfectly.  This is so much fun.  Soon I\'ll make a little nest and read.  Looks like the cloud cover is reclaiming control.  But the skin got a bit of attention and is thankful.  Nature seems to have its own agenda for the day, so all I can do is follow it and smile.    And sure enough, an hour later, nature ushered me out of the canyon.  And all the yoga must really kicking in, in a cardiovascular kind of way, because really...this trek back to the parking lot usually kills me, and it was tough today, but not nearly as.  Yay.  We love physical fitness!  Came home to find out that Ronald Reagan was dead and that Joe and I could get new, sleek flip-phones for $15 each.    \"They have the \'OK\' button,\" he said, as though that told me anything.  \"As long as they\'re the sleek silver ones I\'ve been drooling over for months, I\'m down.\" Should have it when I go to Boise in two weeks.  Yay!!  And Miss Q called, which was so nice.  She\'s having a fabulous time in DC, working with people who are finally of her caliber, doing the sort of progressive, social-change affecting work she\'s always wanted to do and starting yoga.  That\'s right.  YOGA, I said!  She sounds as genuinely happy and content as I\'ve heard in too long a time.  She deserves every bit of joy that seems to be continuously heading her way.  Keep following those signs, sweetie.  They\'ll never let you down.    So yea, not a bad day, all in all.  Think I\'ll go open a Huckleberry beer, finish a book, start another and listen to the rain falling on my tree.    Summer rocks.   '),(23395,2314011,'04,June,2004','2004-06-04 00:00:00','    urlLink    \"It is sometimes a mistake to climb; it is always a mistake never even to make the attempt.  Sometimes you wake, and sometimes, yes, you die.  But there is a third alternative.  Sometimes when you fall, you fly.\"  --Neil Gaiman, \"The Sandman\"&nbsp; urlLink      Realized it\'s been a while since I used the new picture feature, so I looked around my archives for something to celebrate summer\'s return.  Found this little piece from The Boy, done on a weekend not long ago.  It reminds me of a summer three years ago, sitting in Harley on an off-ramp, watching I-90 for any sign of the car we thought was following us back from Seattle.  While he kept watch, I read comics.   '),(23396,2314011,'03,June,2004','2004-06-03 00:00:00','    But I\'m just glowing (not to be confused with gloating) about this whole grades thing.  The GPA for the semester rounded out at 3.45 (which beats ShecD\'s 1st semester average by .03...sucker!), bringing the cumulative GPA to 3.19.    Think it\'s safe to say at this point that I\'ve been blown away, and that\'s a good thing.    Okay, as you were...   '),(23397,2314011,'03,June,2004','2004-06-03 00:00:00','    To review, the earlier-posted grades fell this way:  Research and Writing: A-  Property: B-  Contracts: B+  Torts: B+  And now.....the last and final, the only ones left, the ones we\'ve been waiting for with baited breath:  Civil Procedure:  A (up from an A- last semester, YIPPPIIIEEEE!!!  He must have thrown out the 50 multiple-choice questions from a non-existing class.  Wonder if half the class is joining me this round, hmmmm...)  Criminal Law:  A-  (KICK ASS ALL THE WAY TO JERSEY AND BACK!!!!!  Okay, so this was the exam we all walked out of saying \'What the FUCK was that???\' \'That was a 6 hour exam crammed together in 3 hours is what that was!!!!\' and so on.)  Not bad, I have to say.   '),(23398,2314011,'02,June,2004','2004-06-02 00:00:00','    Getting better at this whole checking the monthly horoscope at the  beginning  of the month.  If this is boring for you, just read my highlighted commentary and get over it.    And so...  \"No doubt about it, this will be a hard-work, high-productivity month. While other signs might not be pleased to hear this, you are different.  As usual.   Capricorns are very goal-oriented, and this month you will outdo yourself, driven to near-obsession to get things done to perfection. As a Capricorn, you like to hear that your talents and your time will be put to good use, and certainly this will be the case in June.   The new moon on June 17 will be the month\'s main event, for from that moment on (for two weeks) an important new project, assignment or account will almost certainly come your way.   In fact, with four planets in Gemini crowded into your solar sixth house at mid-month, it is very possible that this month you will have not one but TWO important projects on your desk, or two clients to see and please. You will be able to juggle all you need to do with rare agility.   We love agility!  Especially the rare kind!!!    You also will be obsessed with organizing things  NO WAY!  - computer files, drawers, closets - and doing so will make your work life much easier. Throwing things out will turn out to be quite therapeutic, making you feel calm and in control.   Just the way I like it.     (respectful snipping of the fitness, weight-loss program)  Romantically, you might be in for a rocky ride this month if you are married or in a close, committed relationship. Before I discuss this, keep in mind that the seventh house (where this aspect will originate) also rules business partners. In what follows, if you have a business partner or collaborator, you might be able to apply this to that person and eliminate the references I make to romance. I am not sure which close partner will be highlighted, but it seems certain one will be.   Well, as long as we have THAT clear.     The first indication of this trend I see is that your partner is about to become much needier at the very time your work will become so busy. Of course, this is not a coincidence. It appears that your partner will feel neglected and it will take all you\'ve got to keep the balls spinning at the office and, at the same time, to soothe your partner and keep your relationship intact.  Thank god I have that rare agility with me!   Saturn is now touring your solar seventh house,  better get that underwear off my table  which often brings about a separation, but it could be one that couldn\'t be helped. For example, this happens when one person in the pair has to work long hours, study at night, or travel a great deal for work. It looks like this preoccupied person will be you, although I can\'t be sure of that - it could be your partner.  Stupid partners always mess shit up.     Saturn\'s position also suggests that this month you may view your partner as too conservative when it comes to taking necessary, calculated risks.  Pussy.     I feel for you, dear Capricorn. It seems obvious you will be pulled in two directions, work and your personal life.  Oh shut up!  It\'s all your fault, anyway!   There is another way these aspects may assert themselves.  Now you tell me.   You may be in a new relationship, stressed because your work life has become so busy that you simply can\'t give a new romantic interest the time it takes to cultivate it properly.  Maybe...   You may be worried that things will fizzle before you can get the relationship off the ground.  Mayyybeee.....   In this case, I think you may be worrying needlessly.  Oh fine!   Simply explain to your new partner what\'s going on. If your new sweetheart doesn\'t understand, he or she never will. In that case, you need a different type of person. But I don\'t think that will be the case. Think optimistic thoughts.   Go optimism, woo!   In your horoscope, part of the problem is that Mars has been orbiting close to Saturn since late May. This frustrating aspect can make your feelings fluctuate from hot to cold, sometimes within a single day.  What the fuck is that supposed to mean???   Saturn is your ruler, so clearly, you are emotionally invested in your significant other and would like to give and get support in return. Before that can happen, Mars needs to separate from Saturn. That will happen in the third week of June, when you will see an improvement.   Another equally difficult problem is that Venus will be retrograde until June 29. Venus is the natural ruler of your solar fifth house of true love, and retrograde planets can bring on a change of mind, or some confusion - in this case, in matters of the heart. Luckily, all will straighten out at month\'s end.   Let\'s substitute that for  \'all will be figured out at month\'s end\' , shall we?    A special day should be Saturday, June 12. The moon will move through Taurus, a great spot for you, and Venus will wink at Neptune, which is as romantic as you can get! This should be a sentimental night, perfect for music, candlelit dinners and going to a very artistic movie together (and talking about it for hours afterwards).    There are several indications that you may travel in June, too, another way to enjoy a little time away from the office.   The weekend of June 26-27 would be ideal for a trip to a nearby town with the one you love (even a friend). Uranus will send the mighty Sun a powerful bolt, and all systems will be go!  I told you, honey!!  We\'re good to go!!!     If you need to take a business trip, my choice for you would be to schedule a visit to a client or colleague on Friday, June 25. It would also be an ideal day to sign a contract, make a presentation or have an important interview. With Mercury in a sweet angle to Jupiter, this will be everyone\'s favorite day for very different reasons!\"  Favorite days KICK ASS!  Especially the ones we can\'t put back, gasp!!    And looking over the other pertinent horoscopes (Libra, Scorpio) it appears the Pride weekend and days surrounding it are setting up perfectly for a grand adventure fit for fairy tales...   '),(23399,2314011,'01,June,2004','2004-06-01 00:00:00','    Anyone familiar with this blog or its author knows anniversaries are a big deal around here.  Especially the epic variety.  And today, the unofficial first day of summer, finds us at the one year mark of a trip out of legends.  But first, a bit of backstory, as usual.    It\'s no secret that the Boy and I love Seattle.  It was our first city we fell in love with together and each trip to the end of I-5 has marked a very important and identifiable point in our lives.    The maiden trip at the end of May 2001 encapsulated life at that point perfectly.  We were finishing up our sophomore years of college, the Boy had made a life-changing and major-shifting decision to follow his heart to Graphic Design, I had hit a magnificent stride in my journalistic endeavors, and we were together in the big city, ready to gay it up like never before.  Well, as much as we could without being able to visit any bars, of course.     We thought we could channel some of that wonderfully sunny energy to our visit in January of 2002.  The world had changed while we\'d been away and so had we.  I found myself still in the throws of self-doubt and confusion about my place in the universe, but hoped a 10-day stretch in the Emerald City with the Boy would help.  It did, and there were many an adventure had during our stay, but I appreciated them only after the Spring Semester began.    Six months later, the world was looking much brighter as we made our way back for our first ever Pride celebration.  It was also the first meeting of the two most important people in my life.  I spent the weeks leading up to the vacation worried, excited and in awe of the fact that Lorraine and Levi would be meeting.  I needn\'t have worried.  They got along famously and the trip was spectacularly epic in every way.    I reprised the winter visit again in December of 2002 for my first Tori concert, but since it was made sans the Boy, who was across the Atlantic at the time, it will only get a passing mention here.    And at last, we come to the famous visit of June 2003.  Freshly graduated and trying to avoid a fate at the UI Library, I made my way to the coast, armed with resumes and writing samples to see what I could see.  I\'d been on the law school waiting list since April and had decided to stop thinking about that possibility and focus on what new opportunities may be staring me in the face.  What better place to start than the city I hope to eventually reside?  Had just started \"The Artist\'s Way\", morning pages were still in their infancy and yoga had not yet occurred to me.  This was also the first time I\'d seen Levi since our farewell in a London train station in January.    Last year at this time found us at the Wild Rose, drinking (my first) beers and talking big talks.  The next four days were the most glorious we\'d ever spent in the city, literally and metaphorically.  The sun was up at 8 every morning and Seattle\'s infamous clouds and drizzle were off playing somewhere else.  Managed to get the first tan of the summer on that trip.  As each day came and went, with more and more offices visited and left with my materials on desks, the signs mounted.  Every little thing was momentous.  Of course, I can only remember a scant few now, but I remember there being more than a handful at the time.    Like the clear skies all day every day.  Like people-watching and writing three long, lazy pages in our coffee shop on Broadway.  Like seeing God in that coffee shop, at our table, two days later.  Like buying the Michael Chabon-edited edition of McSweeney\'s at the Bailey Coy Bookstore.  Like meeting a girl named Josephine at the Rose (what else would her name be?).  Like driving around the city listening to the Summer Mix because it was the only CD player we could both listen to.  Like listening to the Boy gasp as he heard \"Hollywood\" for the first time.  Like reading this horoscope from the latest \"Stranger\" the morning of the last day:  \"\'History is made out of the failures and heroism of each insignificant moment.\' This observation by the writer Franz Kafka has special significance for you right now, Capricorn.  Every little action you take will have unimaginable weight, and the cumulative effect of your many little actions may alter conditions you\'d assumed were impervious to change.  Even your passing thoughts and idle fantasies will have more influence than usual to shape your future.  Until June 20, there will be no such thing as a trivial detail for you.  My advice?  Act as if each moment is a promise of the life you want to be living next October.\"      I still have yet to figure out how my traipsing off to Seattle to find a job promised a life in the law school basement, but I\'ll take it at its word.    And now that summer\'s here, I can\'t wait to get back and dive into whatever new adventure the city\'s holding for us.  It\'s the one thing I have no problem giving over to the universe to do with us what it will; there\'s just something in the air in Seattle.   '),(23400,2314011,'31,July,2004','2004-07-31 00:00:00','    urlLink       urlLink   '),(23401,2314011,'30,July,2004','2004-07-30 00:00:00','    I\'m not in the mood for a drag show.  I want to stay in and devour Book 5, having watched The Chamber of Secrets last night (in addition to The Dreamers, which was totally great by itself, but it ALSO had the Garden State trailer on it....gasp!!!).  I\'m fighting off a summer cold and the preemptive medicine I took last night is lingering.    Made an appearance at the festival this afternoon, though, and apparently caught the eye of a young, buff, Japanese-American cop with braces named Janet.  She\'s from Boise but is in town for the festivities.  She sent a messenger to The Voice of God asking for an introduction.  Kathy\'s put me on the \'let them in free\' list, even though she\'d rather take advantage of me in her office.    I\'m still not in the mood for a drag show, but I knew the moment Kathy said \"No, seriously.  She wants to meet you.\" that I\'d be going.    What should I wear?  Oh, and Happy Birthday to Brandon!!  I\'ll have a beer and shake my ass in your honor.   '),(23402,2314011,'30,July,2004','2004-07-30 00:00:00','    I just bought a shit-load of textbooks.    I feel so dirty.   '),(23403,2314011,'29,July,2004','2004-07-29 00:00:00','    No wonder I\'ve been having chocolate cravings like mad!    urlLink     discover what candy you are @ quiz me      And REALLY, this is the LAST one...before lunch.     You are the mystery woman    urlLink   Which Ultimate Beautiful Woman are You?     brought to you by  urlLink Quizilla   '),(23404,2314011,'29,July,2004','2004-07-29 00:00:00','    Moscow makes me chuckle for many reasons, but one of the most prominent is the fact that every late-July gay&nbsp;folks from the region (all 18 of us) congregate for a parade, festival in the park and drag show.&nbsp;   It\'s very cute.&nbsp;   Think I\'ll have to stop by the park this weekend to check out the...scenery, as I have yet to attend a Moscow Pride celebration.&nbsp; And who knows, maybe there\'ll be a cute, quiet gal roaming around, new to these parts who needs a date to the big gay gala tomorrow night....   '),(23405,2314011,'29,July,2004','2004-07-29 00:00:00','    Slowly coming back to that comfortable place where things can just be; it\'s quiet again and almost calm.&nbsp;   Had a good night of impressive speeches (you keep getting better, Mr. Kerry; keep it up), a satisfying leftover turkey pesto sandwich from Tuesday\'s great&nbsp;night out with Calamity Jane, mini yoga session (working on getting the bod back to where it was a month ago), and a round of nearly-forgotten evening pages.&nbsp; Before long, I\'ll retire to the futon with Book 5 to find out just what the woes of Mrs. Weasley are; they\'ll be fit for storytelling, I\'m sure.&nbsp; But the blog-bug has been gnawing at the back of my mind all day, even though I don\'t really know what to say.   Think I figured out why the last few days have been so...off.&nbsp; Well, there are the obvious reasons:&nbsp; last week, being back from vacation, realizing that school starts in less than a month.&nbsp; But I realized this morning (during morning pages, no less) that the change, end, old/new path intersection, etc. came when I wasn\'t looking.&nbsp;   The Boy and I talked about it on our day of recovery after Pride.&nbsp; This feeling that something was wrapping up, coming to a close, making way for something else, something new.&nbsp; And somewhere&nbsp;after&nbsp;the days full of emptiness and before the fields finished changing colors, it did.&nbsp; I just thought the transition would be so subtle, I would look up one day and realize \"Hey!&nbsp; It\'s here!&nbsp; The new path...and I\'m already on it!\"&nbsp; Didn\'t think it\'d hit when I smelled fall on the yellowing wheat on the drive back.&nbsp; It\'s not as though the \'something new\' has just landed at the door, unannounced and expecting a meal.&nbsp; I still feel like I\'m at the top of the hill, looking around, taking a rest between rounds, enjoying the view and the pause.&nbsp; But it\'s time to start thinking about the trip down.&nbsp; And what made the return more jarring than usual was not getting the chance to feel the wrapping up, the closing of everything behind, not being able to say goodbye.&nbsp; It was suddenly over, and while it\'s not entirely gone, it\'s too far away to see my wave.&nbsp;   But if the last few years have taught me anything, it\'s that I can\'t try to force what is into what was.&nbsp; Comparing the two is out, too.&nbsp; The only thing to do is to keep moving forward.&nbsp; Because the other thing I\'ve learned is that even though changes and new beginnings can be scary and difficult, they can also be a lot of fun and hold great gifts behind their backs.&nbsp;   So I\'ll sit here a bit longer, watching the skies darken a bit earlier, maybe finish another book or two, and will think about coming down another day.&nbsp;   '),(23406,2314011,'28,July,2004','2004-07-28 00:00:00','    Although the day is going much better than its predecessor, I still have yet to cut out all the dilly-dallying.&nbsp; But that will change after I post this one little thing.&nbsp;   &nbsp;   You are Gambit!  You are a fierce fighter and a good friend to have.  Your preference for solitude and your attractiveness make you very intriguing to those you meet.&nbsp; Unfortunately, close relationships are few and far between for you because you often have trouble opening up to others.   urlLink  Which X-Men character are you most like?       Hmm...suppose I should see about getting familar with this character, of whom I\'ve never heard. &nbsp;   '),(23407,2314011,'28,July,2004','2004-07-28 00:00:00','    Get out the party streamers and kazoos; load up the record player and pop some corn!&nbsp; We\'re gonna do each other\'s hair and party until we fall over!&nbsp;   It\'s Lorraine\'s birthday and to celebrate, her hometown is throwing a big parade.&nbsp; Her whole family is descending on the 100+ year-old mansion&nbsp;to perfect the big float and put the finishing touches on Lorraine\'s headdress.&nbsp;   It will be a celebration fit for legends and I can\'t wait to see the scars.   '),(23408,2314011,'28,July,2004','2004-07-28 00:00:00','    Is Mercury in retrograde again?&nbsp; Is the moon waxing in a particularly strange way this month?&nbsp; Is New Jersey seceding from the Union?&nbsp; Because really, something is going on in the cosmos and it\'s fucking my shit up.&nbsp;&nbsp; I find myself in a rough transition&nbsp;into ordinary life these days and I don\'t know what the hell it\'s about.&nbsp; Everything just seems a bit off...quirky, in fact.&nbsp;   For starters, my garbage disposal is protesting my return to the apartment by just humming when I flip the switch.&nbsp; No disposal to speak of...just humming.&nbsp;   I broke my hot Calvin capris.&nbsp; They have completely supperfluous straps across the tops of the pockets that snap on to the side.&nbsp; Must have been too zealous with the one on the left because&nbsp;the snap&nbsp;popped right off and now I must live with a flapping strap above my left pocket.&nbsp;   I can\'t stay awake past 10 to save my life.&nbsp; Want to read (because I can\'t get enough of Book 5 already)?&nbsp; Or watch the late re-run edition of Will & Grace?&nbsp; No, sorry.&nbsp; The eyes start to water and head droops long about 1007 and it\'s downhill after that.&nbsp; And THEN, I can\'t pull myself out of bed before 7; consequently, I\'m getting to work later and later as each day goes by.   Speaking of work...it\'s a fucking fridge in here.&nbsp; My productivity level is lower than usual, as my fingers are numb, which doesn\'t make for the best typing.&nbsp; And I can\'t get enough junk food while I\'m here.&nbsp; Cupcake at 1030?&nbsp; Sure.&nbsp; Hotdog, Doritos and Wild Cherry Pepsi for lunch?&nbsp; Absolutely.&nbsp; Doughnut for dessert?&nbsp; Why the hell not?&nbsp; Take the stairs to compensate for the cuisine and drive to work on account of the above-mentioned lateness?&nbsp;&nbsp;Ha!&nbsp; Surely you jest.&nbsp;&nbsp;   '),(23409,2314011,'25,July,2004','2004-07-25 00:00:00','    \" It was at these times he began to understand, after all these years of study and performance of feats and wonders and surprises, the nature of magic.&nbsp; The magician seemed to promise that something torn to bits might be mended without a seam, that what had vanished might reappear, that a scattered handful of doves or dust might be reunited by a word, that a paper rose consumed by fire could be made to blossom from a pile of ash.&nbsp;     &nbsp;   But everyone knew it was an illusion.     &nbsp;   The true magic of this broken world lay in the ability of the things it contained to vanish, to become so thoroughly lost, that they might never have existed in the first place.  \"&nbsp; --Michael Chabon,  The Amazing Adventures of Kavalier and Clay     &nbsp;  Tried to&nbsp;make sense of the last week, how it began and what it turned into before anyone could blink, and the only thing that comes close is that book.&nbsp; Everything&nbsp;is going just as the heroes have always dreamed, but dared not hope&nbsp;too seriously for, and&nbsp;they&nbsp;still&nbsp;have so much ahead of them, not the least of&nbsp;which is enjoying each moment that seems to carry&nbsp;its own unbelievably perfect gift.&nbsp; And then they\'re blindsided by the most tragic, crippling event and even holding the book suddenly becomes painful.&nbsp;   He\'s on his way back to Montana today with his stepmom and her mother.&nbsp; They\'ll get Marj (stepmom)&nbsp;settled and arranged before he makes his ride back to Bozeman sometime this week.&nbsp; As unbearable as it all has been, it seems that the most difficult part still lays ahead.&nbsp; And the only thing that keeps me from curling up and crying for him are the memories.&nbsp;   Thursday was the hardest, I think.&nbsp; He decided to speak at the funeral, to give his dad a proper goodbye and all of Mitch\'s friends an image of their old ski buddy, co-adventurer, and friend.&nbsp; But the pastor who married Mitch and Marj had things to say first.&nbsp; Levi\'s aunts and I exchanged knowing glances&nbsp;between us throughout his sermon and made frequent looks over&nbsp;at Levi to make sure he wouldn\'t be&nbsp;rushing the stage.&nbsp; The man mentioned Mitch\'s name no more than three times.&nbsp; Oh, he would  use&nbsp; his name to illustrate a particular passage or Bible story, but he&nbsp;only&nbsp;spoke  of  Mitch a scant few times.&nbsp;   But Levi kept it together as he stood behind the podium and we all sighed huge as he began.&nbsp; We cried as we listened to the nicknames (Dirthead, Uncle Dirt, Dirt 13), and laughed at the image of Mitch sneaking in chicken soup from his mom because Stacey (his first wive, Levi and Olivia\'s mom) couldn\'t cook.&nbsp; And as he spoke of his father\'s generosity, sincerity, kindness, humor and sly grin Levi wishes he\'d inherited, I watched in awe of his ability to hold everyone around him as they cried, lifting them up, his back never breaking and comedic timing never failing.&nbsp;   I always knew that about you, hon.&nbsp; Your ability to make everyone else feel better, no matter what you\'re feeling; to be the unwavering support, to be exactly what someone else needs whenever they need you.&nbsp; So I shouldn\'t have been surprised to see you&nbsp;do precisely that over and over, even though your heart&nbsp;was breaking.&nbsp;   You amaze me, Luke.&nbsp; You have&nbsp;his grin and mischievous sparkle in your&nbsp;eye, as well as everything else you honor in him.&nbsp; You\'re my hero, and I hope to be like you when I grow up.   The&nbsp;gravesite service was blessedly short, and as the&nbsp;masses cleared, leaving only Mitch\'s good friends and wilder relatives, a cousin brought out an armful of six-packs and everyone grabbed a bottle.&nbsp; The toast kicked off the beginning of a long night of&nbsp;partying, just as he would have wanted.&nbsp; A friend stayed behind to make sure the empties and piles of dirt on top of the casket stayed put as the grave was sealed and covered.&nbsp;   After a much-needed break with Stacey\'s sisters and their families back at the condo, we made our way to one of Mitch\'s friend\'s home, where vodka tonics were followed by Mexican beers, followed by Jamaican rum and Coke, followed by more beer.&nbsp; The most surreal moment of the evening came after listening to more stories about Dirthead, icy roads at 120 mph and&nbsp;forgotten friends in the truckbed.&nbsp; The Boy and I found ourselves in a circle with his mom, sister, aunts, neighbors, cousins, and Mitch\'s friends, passing joints and laughing our asses off.&nbsp;   The party continued back at the condo, where I acted upon Stacey\'s recommendation that someone start writing some of this shit down.&nbsp; And so, here are some snipets of the evening.&nbsp; Don\'t ask us what half of them mean, but they all were the funniest things we\'d ever heard/said at the time:   \" Damn  that woman can crochet!\"   \" And  there\'s a shitter!\"   \"You want us to lay down with you?\"   \"Where is that Brit?\"&nbsp; \"Stickin\' it in your aunt!\"   \"It\'s like \'Stacy\'s Mom\', but different!\"   \"You guys, I\'ve never been this fucked up before...\"   \"If you eat the muffins tonight...\"&nbsp; \"You\'ll have to make do with eight Almond Bear Claws.\"   \"Aunt Sharyn is a bit like Gollum.\"   \"Pakistanis are always the butt of jokes.\"&nbsp; \"That\'s because they\'re all taxi drivers.\"   \"I bet Grandpa was like \'God-fucking-damnit!\'\"&nbsp; \"Your grandfather   NEVER  &nbsp;said that!\"&nbsp; \"What?&nbsp; He never said \'God\'?\"   \"She slays me, I tell ya!\"   \"If I were sober...\"&nbsp; \"I\'d kick you where it hurts!\"   \"I\'m ready for love!\"   The only thing better than experiencing each of those firsthand was relating them to Levi\'s grandmother the next night.&nbsp; Later, Levi and I took his sister, Olivia out to Charleys for our own celebration.&nbsp; Natalie joined us eventually and the circle of friends&nbsp;was completed when Jen called.&nbsp; We drank, laughed, talked and bonded&nbsp;like never before.&nbsp; When the DJ played Billy Idol and AC/DC back-to-back, I knew Mitch approved.&nbsp;   And now, I find myself in a much-too-hot apartment with half the day gone, thinking about a&nbsp; man I met only once but have finally gotten to know.&nbsp; He was beautiful, crazy, genuine, adoring of his children, and never spent a day doing anything he didn\'t believe in.&nbsp;   We\'ve come to the part in the book where everything has changed and we wonder how there can be another chapter; how can the heroes go on after so much happened?&nbsp; But they do; the story always does.&nbsp; And so will we.&nbsp; So open a bottle of beer today, put on some U2,&nbsp;Tom&nbsp;Petty, Coldplay, or Lynnard Skinnard&nbsp;and drink a toast to Mitch.&nbsp; Or the next time you\'re driving down a dirt road and \"Bad to the Bone\" or \"She Drives Me Crazy\" comes on the radio, crank it up and&nbsp;sing loud&nbsp;for ol\' Dirthead.&nbsp;   '),(23410,2314011,'24,July,2004','2004-07-24 00:00:00','    Back home again and it feels really weird.&nbsp;   I should write about the last few days, but haven\'t the strength, so that post will wait until tomorrow sometime after coffee and before quality futon-time.&nbsp; I think the biggest shock is being alone after having been surrounded by people non-stop since Wednesday; find myself wanting to have someone to cuddle, chuckle or cry with, but I haven\'t been great at chatting with the few people I\'ve come into contact with today.&nbsp;   But the plants look fabulous...the babies are coming along nicely.&nbsp; Thanks, Calam.   '),(23411,2314011,'20,July,2004','2004-07-20 00:00:00','    Just spent an hour and a half reading and replying to emails, catching up on blogs and comments.&nbsp; It\'s funny how wonderful reading everybody\'s normal, day-to-day stories can feel.&nbsp;  &nbsp;  The folks left around 10 and Joe will be home from work in a bit.&nbsp; Been ambling around, doing laundry, drinking coffee and will soon do some afternoon pages.&nbsp; Starting to feel better, like myself again, getting ready to head up to the unknown tomorrow.&nbsp;  &nbsp;  We had a great conversation Sunday evening about holding on to these perfect moments because we would need them to get us through the hard ones.&nbsp; Of course, we meant the ones in November when we\'d be stressing over this project or that class and wondering why we even signed up for this shit.&nbsp; We talked about the pause at the top and bottom of each breath, knowing that no matter where we are, it is always with us, carrying its quiet solace and we can always stop, hold on to it, and use it to get through the next bit.&nbsp;  &nbsp;  Didn\'t know why I was such a mess yesterday.&nbsp; Kept thinking that if   I   felt this way, what must he be going through, which made me want to be with him all the more&nbsp;and&nbsp;\'carry\' some of it for him, as impossible as it sounds.&nbsp; But maybe that\'s what happened.&nbsp; And now I\'m pooling my reserves, holding on to the weekend, memories&nbsp;and breaths, getting ready to take them to him.   '),(23412,2314011,'19,July,2004','2004-07-19 00:00:00','    This story can only be told just the way it happened.&nbsp; The Boy began it perfectly.    &nbsp;   Saturday:   An unexpected party   As always, her timing was freakishly appropriate. I walked home from work through the park and spotted her ambling down the sidewalk to my apartment. With a goofy wave and a shake of our heads (followed by the public fondling) we began our latest weekend of life-altering adventure and bonding.     &nbsp;   We laid on the couch (and on each other) for the obligatory hour of catch-up conversation that marks the beginning of our reunions. I recounted my week of being the cool role model for 6-year-olds everywhere, she talked of the previous night\'s drive with an old pal, still waiting on a copy of my London mix that I forgot to send 7 months ago. (oops)     &nbsp;   We filled the whole hour before she asked, \"So do you want to eat now, or should we wait for Lorraine?\"     &nbsp;   My gasp nearly propelled me across the room. \"Nooooo, Are you joking! SHUT UP! She is NOT COMING! Are you serious? Don\'t you fuck with me like that! Don\'t you even!\"     &nbsp;   By her devilish grin I knew it was true. \"She\'s on the road right now.\"     &nbsp;   \"Shut.the.fuck.up.\"     &nbsp;   But how could this be? Lorraine at my house? In my town? Black Queen? The grand Clam Dragon herself? The secret benefactor I worship and adore on a level separate from all others? The three of us together at last, after an aching year of separation and longing? Earlier in the week i had made a joke of saying \"So you gonna get Lorraine to come up, too?\" We laughed about the possibility, sure. But always in that because-it\'ll-never-happen-in-a-million-years kind of way. The dream was too perfect and painfully impossible to entertain.     &nbsp;   After 20 minutes of shock and bewilderment, I realized this changed everything. The three of us haven\'t been together since last July, and that was an event big enough to kill a team of horses. This was not simply another rockin\' trip with Andrea, (which is rare and momentous enough) this was HUGE! The Queen touched down around 10 last night, and we haven\'t stopped rubbing each other since. Before bed I whined to Andrea, \"I\'m afraid to go to sleep because i just know she\'ll be gone when I wake up, and then I\'ll die! I\'ll just DIE!\" I have to keep poking her to make sure she\'s really there. She brought her fishing poles, and we\'re off to find a package of Italian Sausage. The rest is up to the Fates. How does one compress a year\'s worth of \"Quality Time\" into 2 magical days? I don\'t know, but somehow, I know we\'ll find a way (tear).     &nbsp;  She came to grieve the end of a relationship.&nbsp; A relationship that may find itself again, but for now has ended in its current form.&nbsp; I must admit, a part of me was worried about seeing her again.&nbsp; It\'s been a year.&nbsp; A year since our rationalized \'reconnection\' and I wondered if we could be in the same room without feeling the familiar tension.&nbsp; But there was none.&nbsp; We got on like we always had, without discomfort, uneasiness&nbsp;or too many sparks.&nbsp;&nbsp;    &nbsp;  The fates handed us a brilliant Saturday indeed.&nbsp; Made our way to the site of the October 2000 photoshoot: the bank of a hot river on an even hotter day.&nbsp; But we found just enough shade to keep us busy for a couple of hours.&nbsp; Lorraine and the Boy worked on their sketchpads while I dabbled a bit in mid-afternoon pages before lugging out Book 4, which never received much attention.&nbsp; Ended up chatting while they drew and even made a mad dash for a stray Coke can as it attempted to flee the sunk sack full of beverages.&nbsp; The yoga certainly is paying off, as my ass (who was saved from an embarrassing dunk in said river on said mad dash on account of my stunning agility due to said yoga) can attest.&nbsp; Before long, though, Lorraine\'s air conditioned hotel room beckoned and we soon found ourselves there, cleaning and priming ourselves for a night on the town with dinner at a favorite Italian eatery and Spider Man 2 (which earned a solid B+ from Lorraine.)  &nbsp;    &nbsp;   Sunday : Learning to fly  &nbsp;  Lazy wanderings&nbsp;around downtown Bozeman after a spot-on wonderful breakfast at the CatEye Cafe.&nbsp;&nbsp;We&nbsp;marveled at the waitress\'&nbsp;tight jeans&nbsp;and tank-top dress code and drooled over the hot cowgirl outside our window.&nbsp; Lorraine and Levi&nbsp;found some keepsakes at a favorite bookstore and we soon found ourselves in Cooper Park&nbsp;(named in&nbsp;honor of Lewis and Clark)&nbsp;with our respective reads.&nbsp;&nbsp;After a round of horoscopes, yoga&nbsp;and&nbsp;the kick-off the the Third Task, the rain convinced us&nbsp;to move the party inside.&nbsp;We relocated to the Boy\'s apartment, opened the&nbsp;windows, turned on the fans and enjoyed the&nbsp;wet fresh air and each other\'s company.&nbsp;&nbsp;We gave&nbsp;and got foot massages, lounged and laughed, smiled and shook heads.&nbsp; Lorraine and I retired to the spare bedroom for a bit, leaving Levi to his latest creation.&nbsp; My guilty grin a while later confirmed what he already knew,&nbsp;and I assured him (and myself) that I knew what I was doing and why.&nbsp; No attempts at recreating what we once had.&nbsp; No fantasies of&nbsp;riding off into the sunset&nbsp;together.&nbsp; No rationalizing, no excuses.&nbsp;  &nbsp;  Before I knew what happened, I found myself in the middle of a perfect moment.&nbsp; Laying on The Boy\'s couch, legs sprawled across Lorraine\'s lap, gripping Book 4 tight, gasping at every other paragraph, listening to one of the new mixes, watching Levi paint out of the corner of my eye, smelling rain on the grass outside, everything was in its right place.&nbsp;  &nbsp;  Had take a breath at one point and write a bit about it all:  &nbsp;   Sitting between the two best people I\'ve known, reading a great book, sipping great wine, reveling in afterglow (sans guilt or conflict), with the smell of rain on the air.&nbsp; Couldn\'t really stop to breathe, let alone write about it.&nbsp; Just felt wrong to stop  being .&nbsp; And at last, a respite.&nbsp; But just for a minute because I want to get back to it.&nbsp; He\'s painting, she\'s reading and I\'ve never felt like the planets have aligned to converge such a perfect, needed, surprising gift in such a way.&nbsp; It may be the hormones and it may be the wine.&nbsp; But it could be just what everything kept telling me would happen.&nbsp; Just sit back and let little, seemingly insignificant things happen because that\'s where the most momentous things will emerge.&nbsp; And I find myself drowning in it all.&nbsp; In a cool, safe, fated kind of way.   &nbsp;  We finished out the night with pizza, more great conversation and a titch more reading before bed.&nbsp;  &nbsp;    &nbsp;   Monday : This is gonna hurt like hell  &nbsp;  The Boy arose at 6 on account of his having to be at work by 8.&nbsp; Lorraine and I bid him a fond farewell and crawled back to bed for a bit before making our own departures.&nbsp; The phone rang around 830, making me jump, literally.&nbsp; I made Lorraine answer.&nbsp; It was the Boy and I was relieved.&nbsp; Then I saw her face change and I shivered.&nbsp;  &nbsp;  \"Here, I\'ll let you talk to Andrea,\" she said as she put her hand on my shoulder and handed me the receiver.  &nbsp;  \"What\'s up?\"  &nbsp;  \"I just got a call.&nbsp; My dad\'s been killed in a farming accident.&nbsp; I\'m heading over there...\"  &nbsp;  Everything stopped.&nbsp; The air in my lungs and stomach evaporated.&nbsp; I couldn\'t speak.&nbsp; He kept talking, but I couldn\'t respond.&nbsp; All I knew was this sudden pain in my chest that hurt more than anything I\'ve ever known.&nbsp; Finally, I knocked myself into reality.  &nbsp;  \"We\'re coming over.\"  &nbsp;  And we stayed until noon.&nbsp;  &nbsp;  Eventually, he went to be with his stepmother and Lorraine and I headed south, although the only place I really wanted to be was at his side.  &nbsp;  We drove down \'together\', she following me through the forest to West Yellowstone and beyond.&nbsp; We said our goodbyes at a rest stop between Idaho Falls and Blackfoot.&nbsp; She was opening a water bottle as I pulled away.&nbsp;  &nbsp;  As I neared the Pocatello exit, a song from the mix from last night came on that was too appropriate to be coincidental.&nbsp; I turned it up and sang along, hoping he might hear.  &nbsp;  \"Light up, light up.&nbsp; As if you have a choice.&nbsp; Even if you cannot hear my voice, I will be right beside you dear.\"  &nbsp;  I looked to my left and saw Lorraine in the passing lane, waving the sign for \"I love you.\"  &nbsp;  &nbsp;  &nbsp;  It\'s almost midnight and I don\'t want to get into this big bed by myself.&nbsp; The parents are leaving for Oregon tomorrow.&nbsp; I\'m going to Driggs Wednesday to be with him before the funeral Thursday.&nbsp;  &nbsp;  Yesterday has never felt so far away.&nbsp;   '),(23413,2314011,'15,July,2004','2004-07-15 00:00:00','    urlLink     Innocent and cute my ass   Posted by  urlLink Hello      &nbsp;   And so, after much toil and strife, I at last downloaded the version of AIM I already have in my effort to join&nbsp;my peeps in a chatroom conversation.&nbsp; The gals had already made their departure, but Levi, Brandon and I didn\'t let that stop us from letting our hair down.&nbsp; We covered it all: the textbook cervix,&nbsp;JJ\'s inches, the&nbsp;Grand Am&nbsp;cult, the secret lives of&nbsp;\"innocent\" kittens.&nbsp;     &nbsp;   Brandon put it best when he said: \"it\'s all fun and games until the next time you see me i\'m wearing a purple shirt and driving a pontiac.\"   '),(23414,2314011,'14,July,2004','2004-07-14 00:00:00','    She introduced herself as she made her exit.  Be damned if I can remember her name, but she wanted to say \'hey\' because she\'s relatively new in town and it was months before she found a gang of friends.    \"So when I see a fresh face....it\'s kinda clique-ish around here...\" (which is totally true; everyone knows everyone who walks in the place.  But it seems friendlier than the place a block away.) \"It\'s hard to break in and make friends sometimes...\"  I explained I\'m from here originally, even though I\'m up north now as she inched toward her car.  She said something about something else I can\'t remember as she left...something about me being familiar with something.  God, what happened to all those journalism skills?    Think she was talking to her significant other during the last post, but I\'m definitely enjoying all this \"I remember what it was like for me when I used to do what you\'re doing and I just wanted to be nice and say hi\" stuff I\'ve received from cute lesbians lately.       This place is about to close, but it\'s too early to start scoping out the bar.  Suppose the front porch will have to do.   '),(23415,2314011,'12,July,2004','2004-07-12 00:00:00','    Enjoying the wi-fi outside on the porch at the College Market.  I\'ve only been here about once in the last 7 years, but hey.  Saw a couple of cute gals sitting out here when Mom and I drove by a bit ago on our way home from Buddies and (gasp!) they\'re still here!  The one just confirmed my suspicions by her phone conversation, detailing the latest dyke drama that ruined her girls night at Charleys.  For my part, I just finished reading the latest issue of  urlLink Diversity  (I worked there for one story, you know), so she knows the score.  She vaguely reminds me of the Voice of God.    My other agenda in coming down here is to tear away at the last 200 pages of Book 4 left.  I can\'t get enough.  Just finished the Second Task and it almost killed me.  This just in: she drives a Jetta.   Do you think the laptop makes me look hot?  Or is it the tight London shirt and new jeans I\'m conveniently wearing.  Had the locks trimmed a bit yesterday...nothing major, just rearranged so they can grow out properly.  But definitely have the cute coffee-shop dyke thing going.    She\'s thinking of going to see SpiderMan tomorrow.   The thunder is starting again, so I should put this thing away and see about making eyes at the girl over the top of my book.   '),(23416,2314011,'12,July,2004','2004-07-12 00:00:00','    I\'ve had the house to myself since I woke, nearly 3 hours ago.  Just got done with a round morning pages on the back deck and straighten..er, I mean, doing some light cleaning before Mom gets home.  I inherited her eye for clutter.  I can\'t have it.  So, I made my way from room to room, organizing this, folding that; I even scrubbed down the countertops in the kitchen.  Wonder if she\'ll notice.  She called while I was on the phone with Lorraine to let me know I have a dentist appointment tomorrow, even though I\'ve outgrown their insurance, to deal with my gingivitis.  Whatever.  I flossed and brushed my teeth twice this morning; that\'ll fix it all.    Getting excited for the Bozeman trip.  It will be nice to lounge around with the Boy and our books.  They\'re like kids, really, except they don\'t pee their pants or have pork product issues.  Then before we know it, the parents and I will be making our way to the Oregon coast for a couple of days.  Yay.  Actually feeling better about that than during the last few days.  Getting my strength back and all.  Ready to deal with any shenanigans or goings-on.  And then it\'ll be time to trudge back up north for a month of normalcy before the whirlwind of second year begins.  Lorrainey may or may not be up for a visit, which should be good for a story or made-for-TV-special.  It\'s been a year since the last get-together and even though I realize I\'m probably too excited (read: attached) than is good for me, I\'ll be okay with whatever happens.    I miss my peeps.  It\'s nice to be in a place where everything\'s taken care of for me and all I have to decide is which movie to see or pair of shoes to pick out (got a couple new pair yesterday, score), but I\'m getting ready to be back in my cozy apartment.  Got a good fortune at lunch yesterday.  \"You have found good company--enjoy.\"    Mom, of course, took it to mean her.    I said, \"I like to think it means my new gang.\"    \"Huh?\" she scoffed.    \"My law school friends.  My peeps.\"    \"What does that stand for?\" she said, thinking it was an acronym.   '),(23417,2314011,'12,July,2004','2004-07-12 00:00:00','    Things I\'ve done today (the role of \"today\" will be played by \"during the last week\"):  Slept until 1130.  I\'m not really a fan of the \'in bed by 2, up at 1\' schedule, but I seem to be getting there.    Found my long-lost thumb-ring.  \"Oh yea, I found that right after you left.\" Mom said as I came gasping into the room with my prize.  Put my favorite Sony Discman to rest.  It\'s been with me since 9th grade; it was time.  Bought a new mobile CD player.  It\'s jog-proof; just in case I get crazy.  Went for the middle of the higher end brands, Magnavox, because Sam, my cute RCA boombox, has been acting up and I refuse to pay $70 for a Sony.  It has a Digital FM Tuner Radio; long playing time; Digital Dynamic Bass Boost (that was the clincher for me); and all CD program modes (made damn sure we had random!).  It will come in handy when the parents and I fold ourselves into the pick-up and do the Oregon coast thing next week.  I\'m leaving out the part where I bought it at the Evil Empire; perhaps I should name it Walt.  I felt like asking the female employees if they\'re part of the 1.6 million.    Gave Harley a bath in the driveway.  Didn\'t get to the business about cleaning the inside, but she\'s pretty damn sparkly on the outside.  And that\'s what matters most.    Got Harley\'s oil changed.  Was about 1,000 miles late, so decided it would be a good idea before heading to Bozeman Friday.  My cooling system fluid is a lovely rust color; they tell me it should be green, but I won\'t have to worry about it til winter, so I just won\'t.    Watched Kill Bill Vol. I with Dad.  He quit narrating (\"Got his knife\"; \"Kid\'s home from school.\") about an hour in and seemed disappointed and confused at the end.  Just made me want Vol. II more.  They really shouldn\'t be watched by themselves, there\'s too much missing.  But Uma looked especially good on the huge, high-definition TV and that\'s a very good thing.      Cracked the half-way point in Book 4.  It\'s the day of the First Task and it\'s so intense.  Loving every damm thing about it.  The tension between Harry and Ron...they should just make out and get on with it; Hermione and her crusade to save the house-elves while repairing Harry and Ron\'s relationship in between studying for class and helping Harry be Champion; Mad-Eye and his moods (bad pun, I realize, but the best I can do on the fly); and the British edition with its non-Americanized language.  I love queues!  And being on about something!    Rented \"Welcome to Mooseport\" (for Mom; she, like everybody, loves Raymond) and \"In America\" (for me; I, like everybody if they know what\'s good for them, love Irish immigrants).  And on a completely random note, go  urlLink Seattle lawyers .  Props to The General for the link.   '),(23418,2314011,'11,July,2004','2004-07-11 00:00:00','    Well, I didn\'t have my day alone, but at least it was spent out of the house.  Went with one of Mom\'s colleagues (don\'t ask; it\'s a long story) to 9/11 and took a trip with Dad out to the grandparents in the evening.    The movie was about what I expected.  Still have to divorce the man from the message when it comes to Michael, but I\'m glad I saw it and it gave me all the more resolve that Bush must go.    Coming back to myself a bit more after a trip to Burley today for (the other) Grandma\'s birthday celebration.  Picnics in the park have a way of doing that.  Just getting ready to do some outside yoga, the first since the round on the beach last July 4.  Then I think I\'ll settle on the front porch with Book 4 and a Fat Tire.   '),(23419,2314011,'10,July,2004','2004-07-10 00:00:00','    from going home and part of me can\'t wait to get back.  I always know it will happen, but it still surprises me when I get my fill of home life and yearn for my own space again.    Spent two days in Salt Lake with Mom, shopping our asses off and having a pretty good time, including a big gay talk on the side.  My system is still in shock over the recent increase in the subject\'s discussion and overall, it\'s a good thing.  But I\'m coming away with the feeling that even with the new openness, they still don\'t know me.  Part of it is because there are things I can\'t and won\'t tell them, but part of it, at least as far as Mom\'s concerned, is her thing she does about seeing everything from her perspective.  Never was a good one for empathy.  She sees things the way  she  sees them and is not big on taking a look from the other person\'s viewpoint.  She doesn\'t understand how I could be \"so upset\" at the cousins asking me about myself when I flash my tattoo around, have a HRC sticker on my car and have told people with \"direct links\" to my parents.  When I try to explain how I\'ve tried to shield her and Dad from having to deal (in a negative way) with an issue they struggle so much with (and also because she once told me never to tell anyone she knew because she couldn\'t defend something she doesn\'t agree with), she tells me I should let all that effort and worry go (sounds good, huh?) because I\'ve already acted in all these ways that let everyone around her know anyway.    It\'s been 5 years, after all, and really, I should know how far they\'ve progressed even though we haven\'t discussed it as much in those 5 years as we have in the last 2 weeks.  If I really am as open and okay with it as I say I am, I shouldn\'t be so surprised when my family members bring it up and want to talk to me about it.    And then we get to the part about looking for a partner.  Really, I\'ve only just started finding a social group and relating to people in a way I never have before (most people do that when they\'re 16, you know).  I\'ve spent so much time alone that maybe I\'m just like those people who never find someone because they aren\'t willing to give up their space for the sake of a relationship.    It\'s hard to explain that I know exactly what I\'m waiting for when I can\'t talk about my experiences.    Sigh.    I need to spend today by myself.  I think my way out of what I know when I\'m here.  I start to think she has a point about all these things.  I start to see myself the way she does.   '),(23420,2314011,'07,July,2004','2004-07-07 00:00:00','    Lazy summer days are fun.  Today\'s big adventure was hitting the Poky shopping circuit with Mom.  I cleaned up.  Three new pairs of pants from Downeast Outfitters and new yoga garb from TJ Maxx.  Sweeet.  Even have our eyes on some hot suits at Dillard\'s.  They just might be on sale in Salt Lake...you never know!    Mom\'s lightening up more, which is so nice.  It\'s easier to say \'no\' to a bad outfit and we laugh a lot.  Even got her caught up on the Viscous situation this morning and it didn\'t end in tears.  She managed to get the words \'gay\' and \'lesbian\' out without choking.  \"You should just find yourself a nice young man and settle down,\" she said with a laugh.  \"I know.  Levi and I should just stop the charade.\"  \"Really!!\"  \"Eww, gross.\"  \"Really?\"  \"Really.  If I can\'t get excited about Levi, there\'s no hope.\"  \"Hmmm.\"  The Viscous conversation didn\'t last much longer beyond the bare facts, but she did give me a hug after she\'d escaped to her bathroom to get ready.  I\'m not sure it helped paint a reassuring picture of the lesbian community, but at least it was the truth.  I\'d been having a hard time trying to bring up the topic because there was so little to tell, but the last Mom heard, things were just getting rolling and I knew she was wondering about it.  Truth be told, I\'d been thinking about V more than I should have been the last few days.  Not wistfully or \"Well, maybe...\", but I thought it might be nice to see her again when I managed to make my way to Bozeman before heading back to Moscow.    I should have known better.  And thankfully, the Boy helped me see the way before I made a mistake.  There\'s no big story to tell, just more of what I already knew about her tendency to hear whatever she wants (regardless of whatever is actually said) and turn it into a bigger, more dramatic, more harrowing saga.  I wasn\'t surprised to hear the spin she put on our story, but was disappointed and more hurt than I realized at the time.  But it got me to delete her numbers from my cell phone and put any ideas of friendly coffee out of my mind.  I know it has nothing to do with me, but it still confuses me that something has to be twisted into an ego-stroking soap opera to be worthy of telling friends.  Sigh.  Oh well.  I\'m having too good a time doing nothing to think about it much.    And I don\'t really know what drove me to tell Mom about it.  But after all the times of listening about Joe\'s woes in love and roommates, it seemed about time for my turn.    \"You need someone more mature,\" she said after telling me what a wacko V is, in a vaguely comforting, supportive, motherly way.  \"It\'s true, I do.\"  You have no idea, Mom.   '),(23421,2314011,'05,July,2004','2004-07-05 00:00:00','    What could top a 4th of July spent in bed until 11, on the lawn reading Book 4 until 330, savoring Mom\'s homemade potato salad with dinner and trying not to think about the visit to the Grandparents\' house while watching the biggest fireworks show in Idaho, complete with painful 9.11 country songs?  The Burley Independence Day Parade, that\'s what!  Picked up Grandma around 1030 and crammed our way to our spots just before 11.  Secured my grape cotton candy shortly thereafter and the fun was underway.  Couldn\'t help but think about the very different parade I found myself at last week, but I don\'t think Seattle-ites could pull off the major pieces of farm equipment and old men on mopeds quite like this humble town on the plain.  Mom slapped me when I accidentally called the female forest rangers hot, but she was also laughing, which makes me smile.  We\'re recovering from the whole experience by searching the (empty) air waves for entertainment.  Maybe we should break out the White Russians and invite some black ones over for a game of pool.   '),(23422,2314011,'04,July,2004','2004-07-04 00:00:00','    She\'s on her way home for her annual repose.  She can\'t wait to curl up downstairs or outside on the lawn and read for hours, getting up only to change books.  She felt the familiar tug to write, the one she\'s felt circling for a month, 30 miles ago and doesn\'t care how long she\'ll have to stay here to get it all down.    With every passing mile and new-yet-perfect song, the last two weeks and years slide into place.  She watches from her new perch, marveling and remembering and letting go.  She suddenly found herself up here on a lazy day last week.  The one they gave themselves to recover from their adventure on the coast.  It felt much like the first time there together, what with the U2 on the stereo and wall coverings about to change.  And at the same time, it felt like only last year, when every minute was taken by all by itself.  This realization came as they sat on her couch, him with his sketches, her with her lunch and revelations.  They\'re both at the top of their respective hills, after much work and toil, they\'re finally at the place they weren\'t completely sure they\'d reach.  At first it seemed odd to actually be here, in one piece and so...content.    She\'s spent more time than usual thinking about the path(s) behind her.  Not longingly or jealously, but because she knows it\'s almost time to leave them where they lay, behind, for good.  She\'s about to start a new path and it may be a while before she\'s able to see this far in all directions again.  She\'s caught somewhere between something\'s end and a new thing\'s beginning.  And even though part of her is still sad at the end, it\'s almost habit to rejoice at the beginning, the change, the shift.  They\'ve been good to her, even when she didn\'t know it at the time.  And when old habits start kicking in, telling her to figure this new thing out, she returns to the seeds planted over Christmas, about enjoying the quiet, the calm and the not moving.  She\'s learning to rest in the pause at the top and bottom of each breath.  It whispers to her and the messages are easier to hear (and follow) than the shouts and shoves that used to guide her.  The person she\'s been becoming is settling in and is about to take over.  But for the moment, everybody\'s resting, taking stock at the oasis she finds herself at.       She gave a bit of herself to someone again for the first time since she loved completely and remembers why she waits patiently, finding comfort in knowing such gifts won\'t hurt as much as she feared.     She\'s lived up to all her hype and is finally comfortable with being bored.  Soon she\'ll be ready to go journeying again, to explore and learn big things with a vengeance.  For now, though, she\'s enjoying the view, the adventure, exploring and learning while she sits still and lives in the pause.   '),(23423,2314011,'02,July,2004','2004-07-02 00:00:00','    She called two minutes ago.    I\'m going home tomorrow.    Yay.   '),(23424,2314011,'01,July,2004','2004-07-01 00:00:00','    It\'s 540.  The courthouse is closed.  He\'s still not back.   But the clerk will call me at home tonight to let me know the dealio.    Better be by 830, I\'m seriously.    And it better be a fucking good trial.   '),(23425,2314011,'01,July,2004','2004-07-01 00:00:00','    Who just fucking goes down to Orifino for the day, leaving innocent people\'s (and everyone close to them) fates in his hand, making said innocents wait by the phone wondering if their 4th of July celebrations will include quality family time or a weekend of work, lesbian gatherings and Fahrenheit 9/11?   The judge presiding over the trial at which I may or may not have to appear at Tuesday morning, THAT\'S WHO!!   '),(23426,2314011,'01,July,2004','2004-07-01 00:00:00','    I love my hometown for not very many reasons, but  urlLink this  is one.    Link courtesy of Just Jack.   '),(23427,2314011,'05,August,2004','2004-08-05 00:00:00','    I always talk a good line before she gets here.  I\'m gonna resist; we\'re going to have a great time without going  there ; it\'s not the best idea with things the way they are for her back home; I don\'t even really  want  to go there; we can just be really good friends...really!  And for a while, I pull it off pretty well.    And then there\'s The Kiss.  The kiss I avoid the first little while she\'s here.  Not the \'hello\' kiss or little cuddle kisses.  Those are fine and innocent.  I\'m talking about the one that opens the door to other things.  As usual, I avoided it last night, vowing to myself that we would talk about the state of affairs back on the homestead.  At last update, they were on the equivalent of \'a break\' and I convinced myself I wouldn\'t try anything funny while things were in such a precarious state.  And as usual, The Kiss happened.  And all my resolve, strength, lofty ideals about what should and shouldn\'t happen melted away.    Thank god I escaped to work before things got crazy.  I\'m reassured by the lack of \"I still love you!\" feelings, but am a bit nervous about the \"I\'m still attracted to you\" feelings.    Sigh.  Such a worry.   '),(23428,2314011,'04,August,2004','2004-08-04 00:00:00','    Anyone know who Frank A. Shrontz is?  Because his alumni endowment fund just gave me $500, for which I did not apply, nor was I aware even existed.  But hey, thanks for the moolah, Frank!  Go oil and Boeing!  Woo!  That damn Mercury, I tell ya.  Sure am gonna miss it.   '),(23429,2314011,'02,August,2004','2004-08-02 00:00:00','    Law school ruined me for reading.  There\'s just nothing I want to do less after a day of writing notes, typing notes, reading 80+ pages of material, writing briefs of said material, etc. than coming home and cracking open another book.  Thankfully, I\'ve been able to replenish my reading-for-fun reserves on break.    Well, the library is ruining me for things, too.  Namely, alphabetizing.  Before this summer\'s projects of finding empty holdings via lists and lists of call numbers, I knew P came before Q in the line-up.  Likewise, I knew the call numbers P24 and P25 came before PE25.  But no longer.  Now L comes before K and SH sometimes passes for SE.    At least my day is almost over.  Have precisely 18 minutes left and there\'s nothing, I repeat, nothing waiting to be done once I get home.  Textbook should be sitting at the doorstep, but that\'s it.  I have all night to wait for the phone and/or door bell to ring.  Yay.  Think I\'ll spend it either on the futon with Harry or online with the Gang.  Or maybe both...it\'s just that kind of night.   '),(23430,2314011,'02,August,2004','2004-08-02 00:00:00','    I post this, not for myself--because I\'ve already read, been mad at and dealt with it--but so my failthful audience can have a glimpse of what we all will be contending with this month.  I realize we are not all Capricorns, but some of the messages included in my horoscope overlap with the other signs\', so just go with it.  Consider yoursevles informed, and on notice.      \"You got through July, your toughest month. Rest assured that August will be better and September will be the best month of all. In fact, it could become your very favorite month of the year. Every day, in every way, things are getting better for you.   I saw such heavy aspects for you last month (due to Saturn), with most of the stress centered on a partner or detractor. As a strong, practical Capricorn, you have proven to me that you are able to deal with all sorts of circumstances, including difficult ones, with admirable grace and even ease. I should have been mindful of those qualities in you.   Looking ahead to this month, it\'s a bit of a mixed bag-but certainly far better than what I saw for you in July!   Mercury, the planet of communication, transportation, commerce and contracts, will be retrograde just about all month long, from August 9 to September 2. That alone will bring down the level of progress any of us can expect to make, because when Mercury retrogrades, it is sleeping and we don\'t have its beneficial rays to help us assess, communicate or project forward with the same acuity that we are used to.   This doesn\'t mean you have to come to a grinding halt on all current projects. You can certainly plan, do research, contact others and conduct meetings. I would suggest, however, that you wait to announce final decisions. I will also ask you not to sign any contracts, but other than that, you are free to work as usual. Keep your antenna up for possible changes in strategy or direction, because they often spring up when Mercury is retrograde. Mercury retrograde is like a green light that changed to yellow: Caution is always warranted.   This time, Mercury will retrograde in your ninth house, a part of your chart that rules travel, important relationships with people at a distance, higher education, publishing, broadcasting and legal matters. ( So basically, anything I\'m involved in  at all  ) For example, if you are about to start a semester at college or grad school, you may find that some of your favorite classes are unavailable to you. (Sign up as soon as possible to avoid disappointment.) If you ship a package, it could arrive broken or get delayed in customs. ( Already had one of my textbook orders cancelled. ) Have someone professionally pack it for you and make sure the paperwork is letter-perfect. If you plan to travel, show up early at the airport, even though there\'s a good chance your flight will be late (sad but true). If you are an author, your editor may request lots of revisions. Just shrug and do it-it seems to be part of the cosmic plan. Also, if you are involved in a legal matter, something may suddenly change in the way the case goes. You get the picture.   This area of travel/study/legal matters/higher education is very prominent for you right now, because Mars, Mercury and Jupiter are all filling your ninth house. But it\'s nothing compared to what\'s coming next month in this area, when even more planets are gathering there! ( I love planet orgies!! ) Apparently, one of these subjects is rapidly evolving into a key area of focus for you, so you\'ll be happy to hear that things will improve and progress quite a bit next month. I love it that so many planets are gathering in fellow earth sign Virgo, for you will feel more \"in your element\" and things will go your way.   Money is emphasized in your chart this month. This trend actually started on the full moon on the last day of last month, July 31, when you were focused on your salary or a purchase. If you are in the arts, you may have made a sale, but it is equally possible that you had a large bill to pay, no matter what field you\'re in. If that didn\'t happen on July 31, this same emphasis is still at play during the first four days of August. ( $2/hour raise, anyone? )  On August 17, Saturn, your ruling planet (and, therefore, very important to note), will have a fabulous meeting with good luck planet Jupiter. This is an exceedingly rare event and will not occur again in a similar way until May 2007. On this day, you will come up with a sensational idea that will require your collaboration with a partner or associate that you trust. Collaborate for either business or romantic reasons-gold-star results await you.   I wish I could say that the rest of the month will be quiet and uneventful, but just the opposite is true.   The first sign that something is up occurs on August 18, when Mars, the warrior planet, opposes Uranus, the planet ruling sudden, unexpected developments. This is a very explosive aspect that is capable of provoking quite a bit of anger from you or the other party involved. These two planets will play out their high-noon shootout in houses that center on attitudes, opinions and doctrine. A conflict of ideals with someone who does not see eye-to- eye with you could cause problems, arising from out of nowhere. You will see this played out in a larger scale in the papers, for there will be world events linked to this aspect, as well. It\'s an aspect that has a wide berth, covering the whole week before and after it occurs, so don\'t expect just August 18 to be problematic. ( Go epic cosmic dyke drama!  WOO! )  Then, on August 27, Uranus, planet of unexpected developments, will oppose the Sun. This is a separating aspect, causing people to go in separate directions, also quite suddenly and impulsively. Once again, events seem to come out of the blue. This aspect has a plus or minus of five days, and will therefore overlap with the August 18 aspect mentioned above.   Your third and ninth houses are drawn into this conflict, houses that concern relatives, namely your siblings, cousins, aunts, uncles and in-laws. It probably will not be the best month to talk politics with any of your relatives (even though there\'s a Presidential election coming up in the USA). Bringing up any kind of controversial topic would be a little like throwing a match on gasoline. Some months it\'s better to bury your head in a book.   These houses deal with travel, too, and Mercury is retrograde, so it would be wise to schedule a checkup for your car to make sure it is in good working order.   The end of the month is likely to be equally tumultuous. A full moon on August 29, plus or minus four days, is due to fall in your third house, the same area that we have been talking about. As I mentioned, this house rules short-distance travel, communication, contracts and negotiations. This time, the moon walks arm-in-arm with Uranus, bringing on a highly volatile influence.   By the time the full moon occurs on August 29, we will have been through the other two nerve-jangling aspects involving challenges to Uranus by Mars and the Sun, so everyone\'s likely to be exhausted. If you feel you can\'t quite cope, join the crowd.   Mercury will always exert its strongest pull on the days it turns retrograde or direct, this time on August 9 and September 2. At the time of the full moon, on August 29, plus or minus four days, Mercury will be getting ready to change gears. That means that Mercury could throw plans into confusion at this time. Try to build in a contingency plan, because the planets are telling us to expect the unexpected.   Romantically, your relationship should proceed more smoothly this month than last. You have Venus, the planet of happiness, in your house of commitment at month\'s end. When it comes to love, your mood will more likely be thoughtful and serious than flirtatious and carefree. That\'s fine-love can have a kaleidoscope of moods. You seem to be aware that a relationship has certain privileges and responsibilities for both parties. ( I have no idea what she\'s talking about. Maybe I should ask Lorraine when she gets here tomorrow. )  If one person is not cognizant of this reality, you will have to talk about how to bring things into better balance. You may have worked on this meditation last month, but then you didn\'t have help from lucky Jupiter in the way you will now. Jupiter will cooperate with Venus in the best way possible this month.   Your best romantic moment will arrive on Friday, August 13. Don\'t be superstitious: For you, this Friday the 13th will be great!\"   So there you have it.  And you better watch out.   '),(23431,2314011,'02,August,2004','2004-08-02 00:00:00','    GOD  DAMN   that woman can put together a persuasive summary of the importance of judicial appointments in relation to the American people\'s rights and liberties!!  Only Miss Q can pack two-hours\' worth of reading material into 13 pages and have the reader nodding her head resolutely at the end.   Guess there won\'t be as much yoga tonight, but by god, there will be Harry Potter reading!  And I have to take down the Morpheus result from the Greek God quiz.  Can\'t have it mucking up my side-bar, sorry.   '),(23432,2314011,'01,August,2004','2004-08-01 00:00:00','    I \"suddenly remembered\", while chatting with the big, mean-turned-surprisingly-nice-and-easy-going boss this morning about her week-long vacation that it\'s after the first of July, which means we\'re in the new fiscal year, which means the funds with which my pay-check is funded was replenished, which means it\'s time for said boss to reconsider my pay rate.  It was so intense.    \"I\'ll talk to Mary Jane about it,\" she said pleasantly, making a reminder-Post-It note.     Damn right you will.     Cut to 7 hours later as I prepared to board the precarious elevator to the depths of the library\'s basement and said boss made her way into the mail room.  \"Oh! I talked to Mary Jane about...\"  \"Oh right...\" I said, preparing myself for the inevitable \"...and you\'re screwed\" speech, but hoping for the \"...and we raised you to the $8.50 we said you\'d get to begin with\" because really, she  has  changed.    \"and I was able to bump you up to $10.\"  ......   SAY SOMETHING!   \"OH!  Great!  Thanks!\"  So yea...for the next three weeks (one and a half pay periods), I\'ll be making big bucks.  Sweeet.    Finally in for the night, which feels nice.  Thought about going to The Grind for a bit to edit Miss Q\'s homework, but the thunder/windstorm changed my mind.  So we\'ll just have to make do with some yoga and Harry Potter in an apartment that smells like rain.  Shucks.  And what.the.fuck is the goddamn deal with all my hot accessories being shoved down to the bottom of the page?!   '),(23433,2314011,'01,August,2004','2004-08-01 00:00:00','    Sorry it\'s taken me so agonizingly long to get around to posting about last night.  I\'ve been busy sleeping in, cleaning the apartment and getting a head start on making up time at the library for being gone this coming Friday.  Whew!     So, about the big show.  Felt pretty good about heading to the Beach all by myself; was looking hot in my new jeans and tight green Banana Republic tank-top...the one that shows off the new Seattle Gap belt oh-so-well.  Did a round of yoga before heading out, too, so was feeling pretty good in my flexible skin.  Took a perch in the bar upstairs and people-watched with the Voice of God for a bit.  Janet didn\'t arrive until the first set got underway, so I spent quality time with Kathy in the meantime.  We took stock of the crowd (a fairly young one) and I learned that, while I\'m not a baby dyke, I do still qualify as \'chicken\' because, according to Kathy, I \"would be really fun to do.\"  Always good to know.    As the crowd dispersed at the first show\'s end, Kathy, Janet and her friend (possibly the messenger?) and I met on the stairs where Kathy hurried through the introductions, on her way to put out some fire or another.  We three secured our beverages and I took another perch overlooking the dancefloor, waiting for them to join me.  The only words we\'d exchanged thus far were \"Nice to meet you\" and I didn\'t actually know for sure if the TabiKat story was all that credible, so I didn\'t really feel the need to follow Janet and her friend around...just yet.  Instead, I made sure to stand in clear view of them, leaving plenty of room on either side of me for company.    And sure enough, before long, Janet sidled up next to me and we started chatting.  I have to say, it was really nice to be able to carry on a conversation with someone easily, without worrying about how this flirt came across or whether that was a spark.  There weren\'t any of  those   sparks to speak of, but that was more of a relief than anything, too.  We talked about the school (she\'s a UI alum), how the drag shows have changed since her time, law school and law enforcement, the Boise scene, and how 29 is not old.    \"I\'m 23, and...\"    \"SEE?  I AM old!!\"  \"...I feel old in this crowd.\"  Watching the young things dance and be merry was a good time, I have to say, but seeing the gang of baby dykes I was once drawn to made me chuckle under my breath and shake my head.  So glad I\'m not a baby dyke.    But back to the story...after we downed our beers, we ditched Janet\'s friend (and no, I can\'t remember her name), who was adamant about her position as wallflower, and hit the dancefloor.  While there was no bump-and-grind action, it was just plain nice to dance with a just plain nice person.  I know it\'s not flashy or even a bit angsty, but for whatever reason (and probably more than a few of them), I\'m just not in the mood for big excitement or gasp-filled stories.  So it was just the perfect thing, really.    Not long after the second show, my reserves were finally giving up, needing another dosage of Vitamin C, and my bed was calling, so I decided to make my exit while I could still be polite and charming.  Being the classy lady I am, I left her with my email address written on a bar napkin.    So yea, not too exciting, but that\'s just okay.  She wrote me a little email this evening, so I should go see about a reply.  Then I\'ll finally dig in to Book 5 for the rest of the evening.  If the rumors are true about the book\'s potency, I may not sleep at all!   '),(23434,3994011,'21,July,2004','2004-07-21 00:00:00','   One finds, while drunk, that proper grammer, though useful, is often an impediment.  If you have not read any Bukowski, do yourself the favor.  I will suggest Ham on Rye.    My new favorite bar is experiencing an upheaval.  Owners are pulling in opposite directions and the cool owner is at wit\'s end.  The cool owner, we\'ll call him Chuck built this place and made it what it is.  The asshole owner, we\'ll call him Asshole, is now trying to step in and change things.  Aside from the soap-opera like qualities of the situation, there is my personal tragedy of losing the perfect neighborhood bar.  Clearly, this is the greater tragedy and my legions of faithful sheep, uh, readers will send me money and/or women.  Else my wrath will be upon thee.    Grrrr.    Aside:  If any employees of said bar were to read this, you know that I am taking dramatic license in claiming that it is my tragedy.  I know that it is YOUR tragedy.    I am sad and frustrated.  I have made new friends amongst the staff and I don\'t want to see them leave.    OK, Hemingway and Bukowski are probably exceptions, don\'t try to write while overly drunk. '),(23435,3994011,'21,July,2004','2004-07-21 00:00:00','   Woke up Friday with a bourbon-induced hangover and a looming sense of dread: \"I have to drive to Dallas today to go to that ridiculous class reunion.\"    I had already paid the seventy-five bucks for the  urlLink pop-tarts and coolaid  or whatever it is they serve you at these things and I had told my mom that I would be in Dallas. I never visit home - not for Christmas, not for Thanksgiving. Never. Yet inexplicably, I am my mom\'s favorite child and she is a master of the guilt-trip.    So I\'m laying in bed \'till noon weighing the pros and cons of killing myself vs. going to Dallas when I realize, I don\'t have to go. I\'m an adult! I can eat an entire bag of  urlLink Oreo cookies  without asking for permission! I can walk around the house naked but for strategically placed peanut butter! I make my own decisions, fraught with error though they may be.    The clouds parted, the sun tousled my hair and bunny rabbits and fawns leaped through the gently blowing grass.    I wonder if I can get my $75 back. '),(23436,3994011,'21,July,2004','2004-07-21 00:00:00','   Read this:  urlLink Secret Cabin to be Destroyed     Just when I start to believe that I\'m the one who\'s insane, something like this comes along to remind me that it\'s those Other People who are freaking nuts and wouldn\'t know beauty if it farted flowers in their face.    OK, maybe I am insane.    Regardless, we all make a big deal about which pompous asshole will \"lead our country\" when it\'s the local assholes who generally impact our lives more directly.  Did I write \"asshole?\"  I meant to write \"politician.\" '),(23437,3994011,'21,July,2004','2004-07-21 00:00:00','   Twenty-year class reunion.  Until I got the notice in the mail, I thought it had all happened about five years ago.  Maybe six.    Interestingly, this six year period had somehow contained the ten-year reunion.  I had no desire to go to that one, but after twenty years, I might actually enjoy seeing a few of these people again.  Naturally, the rest of them will be carted off to my bio-fungineering lab where I\'m figuring out just how much fun is too much for the human psyche to endure.  I always enjoy the trips to the bio-fungineering lab.  The laughter is so infectious - literally - and they always give me a lolly-pop when I leave.    I should probably visit my mom while I\'m in Dallas.    I have to make sure someone comes to feed Chairman Soggy von Uni-ball.    Vodka beckons and I must obey. '),(23438,3994011,'21,July,2004','2004-07-21 00:00:00','   Not yet, the fire-ball is still in the sky. '),(23439,3994011,'21,July,2004','2004-07-21 00:00:00','   There is no topic more fascinating then the interplay between men and women.  A wants B, but B wants C and C is torn between D and A.  We all just want to have fun and fuck someone hot, but somehow politics  or morals or some other bullshit has managed to rear its ugly head and make it complicated.  Why is that?  Maybe I\'m a slut;  I want them all.  I want to stay in bed all day eating strawberries and exploring a beautiful body, but the only woman showing any interest wants to rent kayaks and paddle around a lake.  It\'s a match made in hell.  I\'m an idiot. '),(23440,3994011,'21,July,2004','2004-07-21 00:00:00','   Chapter 3047  After a recent date, a friend asked how it went. I confided to her that although the woman I had seen was very smart, attractive, etc., I was tired of injured birds (little flighty ones like finches) with broken wings who are scared of leaping into the sky. Lately and with good reasons, my friend has been acting as though she thinks she\'s a finch with a broken wing. In actuality, she\'s a beautiful eagle who could eat finches for lunch, but she has locked herself inside a cage.  I want her out of the cage and flying.  With me.   Life is too short.  The beauty of this is that she cannot laugh  at  me because as soon as she starts laughing, it will be  with  me. '),(23441,3994011,'21,July,2004','2004-07-21 00:00:00','   This week\'s survey:  In a completely unbiased survey this weekend, I asked my mom who the greatest person on Earth might be.  Imagine my suprise and humble acceptance when she told me that I was not only the greatest person on Earth, but also the smartest and most handsome.  She then had the audacity to ask the greatest person on Earth when she was going to get additional grandchildren.  \"Begone mortal!\"  I answered and hung up the phone.  Be sure to check next week\'s results when I ask my cat, \"If my name is Meow, who is your God?\" -----------  A beautiful woman I was conversing with recently had the \"I can bring home the bacon, fry it up in a pan...\" song running through her head and asked, \"What was that advertising: cigarettes, bacon-flavored lube, what?\"   My answer:  In early 1972, The grocery store chain, Piggly Wiggly produced a bacon-flavored lube named Slippery-Pig.  The earlier \"Pork-A-Lot\" name was abandoned when farmers came in looking for a place to store their pigs.  Anyway, it turns out that the stuff was not very slippery, it smelled more like perfume and their slogan - \"It\'s chock full of bacony goodness!\" - wasn\'t testing well with their target demographic, so they renamed it \"Enjolie\" and sold it as perfume.  Strangely enough, she seems to have found this story entertaining.  Maybe I had found the right woman after all these years of bad choices.  I smiled on the drive home, imagining our future life together.  I was filled with a warm glow and felt peace... even though I had just dropped $800 on that whore.  Ha!  Maybe I should go back to drunk scribblings. '),(23442,4239436,'22,August,2004','2004-08-22 00:00:00','   The Windows platform started a while ago, but by far, has made so much progress and improvements since the good \'ol Windows 3.1, not to mention DOS.    So, I have dedicated this blog as appreciation and informational to the people that use Windows. Yes, I know, there\'s always Linux, and may I add...there will be a Linux blog soon...but for now, let\'s stay with the majority...  urlLink continued story...  '),(23443,4239436,'21,August,2004','2004-08-21 00:00:00','   CSS has really changed web styles. It not only saves time, but you can customize your webpage color schemes to a higher extent. You can also create more professional looking websites just by adding styles to your documents. Today, I will show you a webmaster trick that uses CSS by applying mouseover styles that change as the mouse \"rolls\" over a table\'s background.     Explanation #1:  Alot of newbies don\'t know how to do this and so many people have asked me how to do this too, so I\'ve decided to make a tutorial on it. What is suppose to happen is, when you place your mouse over a table row, the style should change to another style you have defined in what we will call \" styles.css \" (external stylesheet).   Create a new file in Notepad and save it as \"styles.css\":       td.On { background: #FF8A00; FONT-FAMILY: verdana; COLOR: #FFFFFF; FONT-WEIGHT: bold; FONT-SIZE: 10px; } td.Off { background: #0056a2; FONT-FAMILY: verdana; COLOR: #FFFFFF; FONT-WEIGHT: bold; FONT-SIZE: 10px; }       Explanation #2:  									 Now that you have your stylesheet you may link it to your html document like so:     &lt;html&gt;    &lt;head&gt;   &lt;link href=\"styles.css\" type=\"text/css\" rel=\"stylesheet\"&gt;   &lt;/head&gt;    &lt;body&gt;   &lt;table border=\"1\" width=\"100%\" bordercolor=\"#1C464F\"&gt;   &lt;tr&gt;   &lt;td&gt; &lt;/td&gt;   &lt;/tr&gt;   &lt;tr&gt;   &lt;td&gt; &lt;/td&gt;   &lt;/tr&gt;   &lt;/table&gt;         Explanation #3:  Okay, go get some coffee or something worth while to drink, then come back...WELCOME BACK! Now back to work, the next thing we must do now is create the table like so (without the styles implemented):      &lt;table border=\"1\" width=\"100%\" bordercolor=\"#1C464F\"&gt; &lt;tr> &lt;td>The dog ran the race.&lt;/td&gt; &lt;/tr&gt; &lt;tr&gt; &lt;td&gt;But the cat won.&lt;/td&gt; &lt;/tr&gt; &lt;/table&gt;       Explanation #4:  Now you may implement the styles. Note that becasue you attached the stylesheet as and external stylesheet, you can just add the \'class\' attribute to your HTML tags and your styles will show up. THIS IS WHERE THE MAGIC IS! What you do is you add the code  onmouseover=\"this.className=\'on\'\"  and  onmouseout=\"this.className=\'off\'\"  to make the tables change as the mouse rolls over each row:     &lt;table border=\"1\" width=\"100%\" bordercolor=\"#1C464F\"&gt; &lt;tr&gt; &lt;td class=\"off\" onmouseover=\"this.className=\'on\'\" onmouseout=\"this.className=\'off\'\"&gt; The dog ran the race.&lt;/td&gt; &lt;/tr&gt; &lt;tr&gt; &lt;td class=\"off\" onmouseover=\"this.className=\'on\'\" onmouseout=\"this.className=\'off\'\"&gt; But the cat won.&lt;/td&gt; &lt;/tr&gt; &lt;/table&gt;      Putting It Together:  Now as it is pieced together your HTML document should look something like this:     &lt;html&gt;  &lt;head&gt;    &lt;link href=\"styles.css\" type=\"text/css\" rel=\"stylesheet\"&gt; &lt;/head&gt;  &lt;body&gt;   &lt;table border=\"1\" width=\"100%\" bordercolor=\"#1C464F\"&gt;     &lt;tr&gt;        &lt;td class=\"Off\" onmouseover=\"this.className=\'On\'\" onmouseout=\"this.className=\'Off\'\"&gt; The dog ran the race.        &lt;/td&gt;     &lt;/tr&gt;     &lt;tr&gt;        &lt;td class=\"Off\" onmouseover=\"this.className=\'On\'\" onmouseout=\"this.className=\'Off\'\"&gt; But the cat won.     &lt;/tr&gt; &lt;/table&gt;  &lt;/body&gt; &lt;/html&gt;      urlLink continued story...  '),(23444,4239436,'15,August,2004','2004-08-15 00:00:00','   PHP auto-redirection is an overlooked concept that many people fear. You should fear nothing because with a single line of code I can show you how to redirect your page to another! It\'s as simple as  1 - 2 - 3 .     Note:  										 The only precaution to be aware of is the code syntax. The syntax of this small, yet powerful script is very, very code sensitive. There are many single quotes and commas in this script and if you make one typing mistake, then your script will generate errors and you will end up with very unhappy visitors.   Insert the following code into your .PHP  										document you would like the redirection  										to take place:      &lt;?php    // Echos the page that you want to redirect to thus refreshing // The page to the pages specified after the code \'URL=\'    echo    &quot;&lt;meta equiv=&quot;refresh&quot; content=&quot;2;URL=&quot;&gt;&quot;  ;       ?&gt;     Conclusion  The small script may help out in many cases such as after a form has been submitted or a search has been made. This is something a web developer uses on a daily bases. I really hope this tutorial was helpful and may all your scripting be merry!  urlLink continued story...  '),(23445,4239436,'15,August,2004','2004-08-15 00:00:00','   PHP incorporates powerful and fast date and time functionality. In this article Ryan gives us the low down on PHP\'s various functions to manipulate the time and date with ease. Today\'s web sites try to offer a wide variety of useful data and information to the viewer. This is often what makes a web site more appealing than flashy graphics and cheesy animations. In fact, sometimes it is the attention to subtle detail that helps the web developer take that extra step into making a site appear more professional.    One simple way to add a professional detail to your web site is to post the current date and time within your web page. As insignificant as this might sound, it is a very commonly used way to show visitors that your site is slightly ahead of the curve. 										 The following basic PHP tutorial will show you how to use a few PHP functions to retrieve date and time information from your web server and display them neatly on your web site.   We Have A  										Problem:  No matter which operating system you are running your web server and PHP on, the procedure for gathering the current date and time is the same. The first problem is that the returned value is fairly cryptic.  									For example, the following PHP code:      &lt;?php    echo   time  ();   ?&gt;     Will quickly  										translate into:       Also Note: 	 Though this may not be very user friendly, it is an accurate measure of the date and time. This numeric date and time format originates from the fact that all Unix servers tell time by counting the number of seconds that have passed since 12:00 a.m. on January 1, 1970. All we need to do is learn how to display the above information in a way that makes sense to our site visitors.   Fixing The Problem:  Using PHP to convert the Unix timestamp into something more understandable is just a matter of employing the proper function. This tutorial will cover the use of a variety of basic functions available for this task.  The most commonly used function is date(). As you might expect, this function is fairly customizable. Customization is achieved through PHP\'s recognition of a variety of characters submitted in the format string. Below are just a few common characters, for a complete list of them, visit the PHP   						 urlLink  documentation   website.       D: day of the week     F: the month     h: hour, 12-hour format     H: hour, 24-hour format     i: minutes     l: day of the week     Y: year     										If we combine some of this knowledge into  										PHP code, the following:        &lt;        ?php    echo   date  (  \"l F d Y h:i\"   ,  time  ());   ?       &gt;       										Should give a result similar to:    Monday August Mon 2004 03:42    Conclusion:  Obviously this makes considerably more sense than the previous Unix timestamp. The result could be further improved through the addition of punctuation. When additional text is introduced into the format string it is important to remember to use a backslash to comment it out. Otherwise, your text will be recognized as part of the special character string.   urlLink continued story...  '),(23446,4239436,'15,August,2004','2004-08-15 00:00:00','   Hello Everyone! Hope everything is well. I have decided to put together an article that has to do with XHTML (HTML). After seeing way to many sites using frames   incorrectly .     Introduction  In Web Design and web technologies, there is always a correct and an incorrect way of doing something. Throughout my time as a webmaster I have seen good ways and bad ways to use frames. Personally, I do not use frames at all anymore. To me they are a hassle and somewhat, in my book, depreciated. This article, though, is not for me, but for you to be informed about the wonderful world of web design and multimedia. So come along for the information! 										 										 What Is A Frame Made Of?  										There are two parts of a frame.  There  										are the  frame(s)  and there is the 										 frameset . In small words, a   										 frameset   is the primary page that holds all the \"frame\" pages together into one document and implements the actual frames into the document. Slightly different, a   frame   (or frames) are the actual pages that contain HTML and/or other scripting/programming that form the whole page. So think of a frameset as a placeholder and a frame as the objects. 										 										A frameset document is a document that defines  										one or more frames by using the    frameset    										and    frame   elements. 										 										 So, What\'s The Advantage?  The advantage to frames is you can present multiple documents into one browser window. For example, you could make a frameset(index.html) and add the following page: a frame on the left side(nav.html) and assign the page links for your navigation links for your website. Then you may have a frame at the top of your document(header.html) inserting an image to make it look like your logo top header part of your site. Then you could make a body frame(body.html) and assign it as the target for your links so that your links target display into the body.html frame. This can get confusing if you don\'t do this yourself. But this is the only advantage to frames. 										 										 Yeah, But What Exactly Is The  										Disadvantage?  										The disadvantage to frames is the  										following: 										 										 										      From browser to browser (IE, NN, Opera, Mozilla, and Firefox) they don\'t display the same way causing your users to get annoyed or confused and they click out of your site.       Frame size has a huge factor too. People have different screen resolutions(800 X 600, 1024 X 768, and 1280 X 1024) and frames don\'t calculate right making your users have to scroll back and forth.       Some browsers (most anyway), do not support the fact of printing. Many web browsers just print out a certain page, but not the whole page the user is seeing. It could be the links page, or the body page, or even the header page.       Search engines have a hard time getting a correct search. Like if your site was a frameset and had all the pages as stated above, and you went to Google and searched for your site: You click the link for your site and all you get is this page of links. The user is confused and clicks out having a bad, very bad opinion about your site. And you as a web designer is displeased, because you have no, or almost no visitors.       Finally, the last, but certainly not least reason why frames are a dying tool, is the fact that tables are a whole lot cleaner. You can format and display information much better in tables rather than in frames.       										 										 Conclusion  That concludes my thoughts on frames and I really hope this article was helpful. An to make something very clear, when frames are used correctly, great things can happen. Check out    urlLink Microsoft\'s MSDN    										website.     urlLink continued story...  '),(23447,4239436,'15,August,2004','2004-08-15 00:00:00','   I have been looking into Flash and its design concepts. Most are great, but only if used correctly and not \"too much\". There is a right way of going the flash path, and of course, a wrong way.     Introduction   I am starting to see more and more sites made in Flash (now Flash 7 (MX)). So what I did was I started looking into Flash and designed a few idealistic programs/working models. What I came to find out was that Flash isn\'t exactly the best thing for a website. Don\'t get me wrong. I designed a few web apps using actionscript and such and exported a Flash file, but it isn\'t (in my opinion) the best way to make a site look good.   Web Browsers   Web browsers are starting to support Flash more and more, but the thing about Flash is that it is very limited in user ability. I have seen very, very good flash sites, and unfortunately, very bad ones. The Flash sites that present themselves are the ones that are very interactive with help situations and vector graphics.   The Good & The Bad Ones  The other day I saw a site made with 100% flash and it was very hard to get around, because the designer didn\'t make his site user-friendly, thus I did not know what to do. He didn\'t have a \"Home\" button and it was just horrible. On another note, I have seen amazing Flash sites and have been caught in awe and amazement of what you can do with Flash and web animation interactivity. It had buttons, polls, interfaces, downloads, and everything preloaded making it very slick, fast, and flawless.   Too Much?   Flash really is a great program and a huge revolution in the web design industry, but like with anything else; it can be overused. I\'m not saying that after about a million sites we should stop using Flash. What I am trying to get across is that 100% complete Flash sites can be very, very bad. Not all web browsers support Flash, and, as with any other program can malfunction at anytime leaving the user feel unwanted or disgusted with your site. Too much Flash can also be upsetting in the fact that not a lot of people are used to it and don\'t know exactly how to take it. As stated above, to make a Flash site better, add some help documentation in HTML format to keep the user in control not the program.   Alternatives   There are many alternatives to Flash and actually alternatives to help using Flash and HTML together. One of the best things on a Flash site I have seen is a home page splash option. What this does is it allows a person viewing the home page of a website to choose between a \"Flash\" version of the site or the \"HTML\" (HTML & CSS) version of the website. The helps the older browser and the ignorant/new people in the computer age. Another way to get away from flash is to try other web languages such as C#, Visual Basic, ASP, and my favorite, PHP/MySQL.   Conclusion   As Flash may spark your interest, you should know the good and the bad of everything. Flash is good in some aspect, but don\'t overuse it. If it is overused people will get tired and there can be extreme consequences. Have fun with your website and happy coding!  urlLink continued story...  '),(23448,4239436,'22,August,2004','2004-08-22 00:00:00','   Here we are...now before you \"continue\", just know that I put much effort into this and imagination. As I have started graphic design, I have realized that you must have a sense of imagination. While thinking of a beautiful spot in the world; with the sun shining just off to the northwest and a canyon to your left. Bliss...total bliss...    Here is the final result...   urlLink     urlLink continued story...  '),(23449,4239436,'22,August,2004','2004-08-22 00:00:00','   I finally got my Graphic Design Blog up.  This will definatly updated daily, or bi-daily.  This blog is mostly a collection of...   graphically based images I have created and want to share with. Be patient with what I have, because it will take great effort and time to get everything up.  urlLink continued story...  '),(23450,3739687,'25,May,2004','2004-05-25 00:00:00','   hello,  well i hope everyone\'s summer is going as well as mine.  i had my first thai massage yesterday and it was awesome.  cheap too.  i\'ll give you a brief description if for no other reason than to encourage you all to speed up your vacation to bangkok plans.  -  first i walked into a curtained off area  and changed into a pajama set they provided for me.  then a very nice thai women came in and brought me hot tea, i\'m not sure what kind i think it was caffeniated though, then another woman came in and washed my feat with a scented oil,(for those of you who know about how awful my feet are you can imagine how bad i felt for the poor woman) then the massage began.  it is very different than the ones i\'ve had before.  thai massage involves the giver using their entire body. at times i felt i was being stretched more  than anything.  at one point she flipped me on top of her with her feet in my back, it was very amusing, we both laughed.  i\'m still a little confused on the exchange rate but i only paid a few bucks for an hour.  o.k. so who\'s coming to see me?!?  other than that not much to tell i didn\'t make it to lumphini park yesterday so i think i\'m going to go today.  i\'ve been looking forward to seeing some trees.  btw i love pineapple now.  i don\'t know if it is that much better over here or if being in asia changed my taste in food. either way i eat two or three helpings of it everyday.  wendi and karen (friends for oxford) got into town last night.  they invited me to go to the beach with them and i accepted.  we are going to a full moon festival on an island.  that\'s all the details i have at the moment i\'ll let you know when i find out more.  once i get back then the job search will commence.  wish me good luck.  i\'ve got a few ideas and phone numbers so i\'m  hoping it won\'t be that hard.  most people i\'ve met seem to think i can find one extremely easily.  let\'s hope they are right!  bye,  jane  '),(23451,3739687,'05,June,2004','2004-06-05 00:00:00','   hello,  hope ya\'ll get this.  i\'m back in bangkok.  wendi, karen, dana and jay flew out early this morning. i was sad to see them go.  we had a great vacation together.  i\'ve already gotten the word on a teaching job so hopefully it will work out.  most expats here are extremely helpful.  also i have an address now it\'s:  178 SUKHUMVIT SOI 16 KLONG TOIE BANGKOK 10110  i\'ll let you know my cell number as soon as i go by and get one.  from what i hear they are fairly cheap and i can get free incoming calls.  gotta go.   love ya.  jane  '),(23452,3739687,'02,June,2004','2004-06-02 00:00:00','   hey all,  well i\'m starting to crave good southern veggies but other than that i\'m adapting pretty well.  i really like what i\'ve seen of the people and culture here.  hope ya\'ll get to come see me.  i\'m still talking with various people about jobs everywhere i go but i looks like bangkok is where i\'m going to end up.  i\'m still on the island right now having a great time with karen, wendi and crew.  ya\'ll be good.  i\'ll write more later.  jane  '),(23453,3739687,'01,June,2004','2004-06-01 00:00:00','   hey all,  i\'m at koh phagnan now with wendi and crew.  we are having a great time.  i\'m making contacts along the way but the next few days are for layin on the beach.  i hope everyone is doing well.  gotta go.  jane  ps koh phagnan is an island towards the south of thailand.  '),(23454,3739687,'31,May,2004','2004-05-31 00:00:00','   hello all,  sorry if mostof this isa repeat.  internet problems everywhere i go.  anyway... i\'ve been to kanchanaburi and now i\'m in on nang on the beach.  trying to work on a tan.  i\'ve had to remove some people from my internet list so if you hear of someone not getting this please just forward it on to them.  sorry.  anyway... i\'ve ridden on elephants, hiked in the jungle, gone snorkeling, pretty much in general just had a great time.  everywhere i\'ve gone i\'ve met people that are going to help me find a job.  other than that not much to say.  i finally got my camera working correctly so when i get back to bangkok and get my usb connector i\'ll send everyone pics.  also i have karen heidel, who is in law school at ole miss, who is going to drop of copies of pics for the fam to see.  hope everyone is having a great summer. ya\'ll be good. love ya.  jane  '),(23455,3739687,'12,June,2004','2004-06-12 00:00:00','   well i missed the bus to cambodia this morning.  which means i got up at 5:45 for no reason.  now i have to do it again tomorrow, but the good news is it gets back at 4:00 OR 5:00 in the afternoon so i will have plenty of time to rest for my next week of schooling.  i hope everything goes well. oh and i didn\'t miss the bus because i was late. i missed it because i had the wrong soi.  btw soi means street.  for example i was supposed to be at sukhumvit soi 21 but instead i was at sukhumvit soi 14.  it really pissed me off because i\'m sure the woman told me 14.  anyway....  i\'ll be at the right spot tomorrow for sure.  i\'ll even be atleast 1/2 an hour early to make sure and i\'m going to call the lady if anything looks suspicious.  i\'d hate to  have to miss my first day of a real week of school if i have to do it on monday.  plus those nuns might go nuts on me.  well nothing else to say.  i have my camera but not my usb connector so no pics today either.  i promise to send them soon. gotta run since i have the day off i\'m going to do laundry.  love ya.  jane  ps.  for some reason i can\'t send emails to more than 50 people so shelly will you please forward my messages on to donna, melissa, maryann, etc.  also to whoever would like them.  especially anyone thinking about coming to teach in thailand.  they might find some of what i\'ve said useful.  '),(23456,3739687,'11,June,2004','2004-06-11 00:00:00','   hello,  well another day at school.  this one has been extremely slow, but the good news is that i got my new schedule this week.  i\'ll be teaching many more classes from now on,but since i\'m just splitting other people\'s classes i don\'t have to do much of the paper work.  yeah!  i\'ll also be tutoring during part of lunch break, but apparently it is optional for the students and most teachers who have done it before say they usually don\'t show up.   anyway... i guess i\'ll just have to see.  nothing else to tell really.  i substituted for 1st grade this morning.  they were so cute but hard to understand because they only speek basic english.  i\'ve got to learn to slow down when i talk to them.  other than that it went great.  some of the kids are extremely shy, but if you make it fun they will participate more.  i\'ve almost got my work wardrobe complete.  i just need one more skirt and another shirt.  hasn\'t beeen as hard as it seemed at first.  love ya, jane  ps i\'m joining a faculty basketball team next week.  yeah!!!  i\'m excited to be on the court again.  '),(23457,3739687,'10,June,2004','2004-06-10 00:00:00','   hello,  well today was an interesting day.  apparently i got hired just in time for teacher\'s day.  it was an experience.  the kids all set up in trinity hall which similar to fulton chapel at ole miss or any other large auditorium.  the kids all stand and sing as the teachers walk single file to the stage area where there were chairs for us to sit in.  then they prayed in thai so i have no idea what they were saying.  then one girl read a poem, another  sang, and one read a speech in english.  the speech was about the importance of teachers and how they are second parents to all the students. next one girl for every teacher walked on her knees to present the teacher with a flower arrangement.  the arrangement were so beautiful and the air in trinity hall was filled with the fragrance of fresh flowers.  glad i\'m not allergic. anyway.   unfortunately i didn\'t get to keep the flower arrangement given to me.  also, i found a place to live.  it\'s extremely close to the school.  i move in on june 25th.  i\'m ready to have my own space.  molly\'s place is nice but i won\'t miss the commute.  plus most of the teachers i work with live in the same area.  well gotta go.  love ya\'ll.  jane  ps.  i know how to send pictures i just keep forgetting to bring my camera with me to work.  i promise to send some soon.  '),(23458,3739687,'09,June,2004','2004-06-09 00:00:00','   hello all  well sister erlinda is definately a nutball.  she purchased a program for nilza (a brazillian woman) and myself to figure out and teach with.  only she got ripped off becasue the copyright is from the 1970\'s and the program is not very reputible.  so pretty much i spend all of my day at work correcting papers and figuring out good exercises for the kids to be doing.  another thing that makes it difficult is that i only see the kids once a week so learning names and faces is going to be close to impossible.   anyway...  honestly i only spend about two to three hours in the classroom.  that\'s going to change soon.  they\'ve already noticed that i have too much free time and they are adjusting my schedule.  yuk. but atleast i won\'t be sitting around pretending to work.  i must pretend because since i\'m new they are keeping a close eye on me.  i think they must have hired incompetent teachers in the past or maybe ones that tried to get away with doing as little work as possible.  honestly i can already tell that once i\'ve been there for a few weeks i\'ll probably be able to do as i please.  i\'m looking forward to those days.  other than that no real complaints.  i never have any work to take home and all the other teachers assure me that this won\'t change.  gotta go i think i\'m going to eat indian food tonight. yummy in my tummy.  love ya\'ll  jane  '),(23459,3739687,'08,June,2004','2004-06-08 00:00:00','   hello all,  well i didn\'t get to finish the last email because sister erlinda needed to see me.  i\'m not quite sure how far i got into my description of the school.  anyway...  one problem i\'m having is pronouncing the girls names and telling them apart.  they all wear the same uniforms according to their grade so it makes it even more difficult.  they think i\'m funny because i\'m so tall compared to them.  one girl even called me big.  it didn\'t hurt my feelings but i hope those comments are kept to a minimum.  i\'m off to cambodia this weekend for a border run.  which means i have to leave the country in order to not pay a penalty for overstaying my tourist visa.  then the school will pay for my work permit next month.  unfortunately i have to pay for the trip to cambodia but from what i\'ve heard this is what every school does.  i guess they want to make sure the teacher is going to stick around before they fork out the cash for a work permit.  anyway.....  not much else to say.  i\'ve got the word on a place to live for extremely cheap very close to where i\'m working so hopefully that will work out.  the commute this morning was longer than i\'d hoped it would be.  one thing someone asked me is about an example of the improper english used in various t-shirts and what not.  i think my favorite is a bumber sticker that said \"do follow me i\'m not going\"  whatever that means also they commonly confuse wax with vax. i see beauty salons like that everywhere.  the thai people are very nice and friendly which makes things easier.  the only problem i\'ve had so far is shopping.  they are all so small that it is difficult for me to find clothes to wear.  i\'m working on the problem.  oh well gotta go.  hope ya\'ll are having a great summer.  jane  '),(23460,3739687,'07,June,2004','2004-06-07 00:00:00','   hello all,  well the job seems to be going o.k. so far.  it\'s an all girls school run by thai nuns.  kinda crazy.  so far everyone has been extremely nice and helpful but i haven\'t had to do any real work yet.  the hired me on as an extra teacher so the class size could be reduced from 30 to 15.  gotta go.  jane  '),(23461,3739687,'07,June,2004','2004-06-07 00:00:00','   hello,  well i found a job.  i am working at st. joe for a nun named sister erlinda.  guess we will see how it goes.  i am a little scared but not much i can do but give it a try.  one funny thing i forgot to mention about thailand is that you see kids with english written on their shirts but it is usually improper english or does not make sense.  kinda funny.  guess i will be trying to help with this problem.  i have a cell number now so if anyone wants to reach me i have free incoming calls.  it is  010081407.  i am not sure what kind of international numbers go along with that or if that is all you will need to dial.  well gotta go.  jane  ps.  i can not wear pants or skirts that do not cover my knees at the school.  those of you who know me well can probably imagine the pain it was for me to shop today.  but i got enough stuff for a few days.  love ya.  '),(23462,3739687,'06,June,2004','2004-06-06 00:00:00','   hey,  well things are going well i\'m having a good time with molly and all of her friends.  we went to a wonderful restaurant for lunch today.  now we\'re shopping for a dvd to sit around and watch.  also, found out about another job possibility.  going to look into tomorrow.  love ya, jane  ps i promise to write a longer email when i have a chance giving more details about what it is like being here.  '),(23463,3739687,'19,June,2004','2004-06-19 00:00:00','   hey,  sorry for the generalization.  i know not all french people suck, but the party last night turned out to be nancy, me, and a bunch of french people.  they were o.k. at first, but then politics entered the picture.  what\'s amazing is that i agree with them on most of what they were discussing about bush and the abu gahrib (spelling?) prison abuse scandal, etc...  anyway... they still seemed very attacking.  so since they generalized me i\'m just  going to assume they are all shitty.  o.k. enough negative now on to positive.  a few of them were very nice and friendly and when they weren\'t lapsing into french we had very good conversations.  also, they enjoyed the wines i brought.  which were a couple of australian syrahs. all in all i had a good time vs. a bad time.  i\'m planning on moving into vicky\'s place sunday and then my room will be available on friday.  i\'m excited about having a place to call my own.  what\'s also great is that since it\'s basically a hotel with special deals for st. joe teachers i get my room cleaned for free.  i\'m going to be so spoiled when i get back to the states.  which if i stay as happy as i am now it might not be for awhile.  well gotta go.  jane  '),(23464,3739687,'18,June,2004','2004-06-18 00:00:00','   hey everyone,  well i think i\'m starting to enjoy teaching.  hard to believe considering all the teachers i tortured in my day.  anyway...  i have a few i\'d like to apologize to if i could.  today i\'ve taught cross-stitching, english, reading, and substituted for an art class.  i\'ve learned that being flexible is very important.  there is no organization at st. joe so as long as you look busy it\'s easy to get away with doing whatever you want.  sister erlinda, the head crazy, caught me on the internet today in the computer room which is supposed to be for students only.  i played dumb and since i\'m new it worked.  anyway...  i\'d say the most important thing a person needs for teaching is an imagination.  i didn\'t even know i had one until i realized that if i didn\'t keep the kids busy they would be going wild.  anyway... another comment on the level of just how spoiled these kids are:  they are extremely well off by thai standards.  they all have maids and other workers doing everything for them.  in one exercise that they had to complete about the environment and preserving energy a question asked them what they would do if a worker left the refrigerator door open.  most of the answers involved the worker being fired. like i said before these girls are brats.  well other than that not much to tell basketball was loads of fun yesterday. i enjoyed playing with the teachers and and the kids.  my skills with the ole b-ball hadn\'t completely left me.  i even managed to teach a couple of the girls how to dribble behind their backs.  i think i might make it a regular activity for me to play with them every afternoon.  god knows i need the exercise.  love ya\'ll jane  '),(23465,3739687,'18,June,2004','2004-06-18 00:00:00','   hey folks,  well i had to sub for the music teacher today.  ya\'ll would have loved seeing me trying to sing \"three blind mice\" and \"old MacDonald had a farm\" to the first graders.  it was hilarious.  one good thing about teaching kids is that they love it when the teacher acts silly.  also, it makes them pay attention.  so i quacked like a duck, oinked like a pig, and everything else.  the kids are soooo smart they pick up on everything i say and do.  it\'s really amazing to see how fast their brains can work.  however the older ones are a different story.  they are going through puberty and can be real bitches.  anyway... i didn\'t have to teach them music lessons.  they just had to do an extra english lesson.  i made the handout extremely difficult so it would take most of the class time for them to complete it.   gosh i\'m a mean teacher.  jk.  also, i went to an art opening last night.  it was interesting.  the work was patchwork tapestries done with thai silk and various other fabrics.  of course the ones i liked were sold, but i\'m not in the 12,000 baht income range anyway.  wish i were i would have loved to own one of his works.  also, wish i could remember the artist\'s name.  i\'ll try and find out.  he\'s having an opening in nyc next month.  other than that there\'s not much to say.  i\'m going to a party tonight.  nancy, another girl i work with, is having people over.  she\'s from virginia and very cool.  also, she\'s a great teacher.  i\'m forcing her to help me be more effective with the kids.  i\'m determined to make a difference.  hopefully my dedication will survive the neurotic nuns who are trying to  keep me down.  jane  '),(23466,3739687,'16,June,2004','2004-06-16 00:00:00','   hello all,  well things are going pretty good here.  i\'m still substitute teaching this week.  it\'s going o.k. they actually had me paired up with another teacher, vicky, to help teach first graders how to sew.  like i said befor these nuns are nuts.  anyway... my only complaint at the moment is that i get told what to do by several different people and all of them contracdict each other.  i just do what i think is best for the girls.  i have had one problem student the same one as yesterday.  she has some sort of emotional problem and i was told by everyone (the first time i\'ve heard a general consensus) that as long as she is not being a major disruption in the class to just let her do whatever she wants.  since she wasn\'t yelling at me in thai today i let her color instead of do her homework.  well i\'m playing basketball this afternoon with some of the other teachers.  i\'m looking forward to being on the court again.  it\'s been awhile.  i\'m sure i won\'t last long.  maybe i can talk them into playing horse instead of a full court game.  anyway... someone sent me a comment about having an advantage over the other teachers because of my height.  well this isn\'t going to happen.  i\'m playing against western teachers male and female and  only one of them is shorter than i am.  actually i\'m even intimidate by her because she is actually i good shape.  oh well i\'ll let you know how it went tomorrow.  wish me luck.  hopefully i just won\'t sprain anything.  love ya\'ll.  jane  '),(23467,3739687,'15,June,2004','2004-06-15 00:00:00','   hello,  well it\'s official i\'m the designated sub.  i\'m actually becoming o.k. with it.  the girls hate me because it means one more english class a day but i\'ve learned how to make them behave a little better.  i did have one serious run in today with a girl who i was told later has emotional problems.  anyway... other than that it was a good day.  i\'ve been meaning to mention a few things about thailand that stand out in my mind.  1. the escalators move faster here than they do in the states. 2. they use corn in lots of their desserts.  for example corn shakes at kfc and the dessert today at school was some sort of corn jello mush.  i thought it was disgusting. 3. almost everyone speaks some english.  even poor looking vendors on the street know numbers. 4.  people seem to eat all the time here.  i\'ve never been out and not seen atleast a few vendors selling all types of thai dishes.  anyway... i got an email asking more details about the school i\'m working at so here are a few things that might interest you.  first of all there are 5,600 girls (btw it\'s an all girls school) most of them are very wealthy by thai standards and many of them would be considered  middle class by American standards.  i found out today st. joseph is one of the most respected thai schools (this does not include international schools which are considered in a different class) in bangkok.  the girls wear uniforms  determined by what grade they are in.  starting with jumpers and ending with skirts and button up shirts with ties which are all in blue and white.  well that\'s all for today. i\'ve got lots of internet research to do so i\'ve got to go.  love ya\'ll  jane  '),(23468,3739687,'14,June,2004','2004-06-14 00:00:00','   o.k. so i can\'t figure out how to send a digital picture.  honestly, i haven\'t taken very many pictures. anyway i\'m just not a picture taker.  wish i could for ya\'ll but my patience is up.  something about the size of the file i\'m trying to send.  anyway...  guess i lugged my camera and usb connector to school today for nothing.  oh well,  i\'m sure i\'ll figure it out eventually.  well today was another boring day at school.  i think they hired me so they would have a built in substitute teacher.  i spent all day covering for the thai teacher which basically meant i was a baby sitter since i can\'t teach thai.  the kids got a chance to catch up on homework in other subjects so they were happy.  this schools loads the kids down with work so any chance to get ahead is usually welcomed.  they are basically good kids a little  spoiled but only a few are real brats.  not much else to say really.  i\'m tired.  i think it\'s from not doing anything all day long.  that\'s more exhausting than actually working i think.  well hope everyone is having a great summer.  jane  '),(23469,3739687,'13,June,2004','2004-06-13 00:00:00','   hello fam and friends,  well.  i made the bus this morning.  i got to the station 45 minutes ahead of time just to make sure and i called and harrassed the woman to tell me exactly where i was supposed to stand and what the bus would look like.  anyway... i got on it just fine.  it was a very nice aircon bus that served us sushi (not great but edible) and constantly provided us with a our choice of pepsi, tea or water.  i\'m glad i went with one of the services that does everything for you instead of just going to the bus station and arranging the travel myself.  the part of cambodia we went to, poi pet, was very scary.  i was encouraged to keep my bag in front of me and the tour/border run guides kept a close eye on anyone coming up to me because i was one of only two westerners on the bus and the only western female.  there were beggars everywhere many of  them children running up to me asking for a single baht. also, many people missing arms or legs laying around and one man with only his right leg.  i was afraid to take out my camera to take a picture even. well honestly i didn\'t want a record of the event.  what\'s more amazing is that after crossing into cambodia they put us in a van and took us maybe 1/4 of a mile to a casino where they was a full buffet a all types of food available.  mainly asian food but some american style items also.  strange how people were starving just right down the road.  made me feel guilty when i went back for second helping.  well gotta run.  i\'m a little depressed now that i put the experience into words.  jane  '),(23470,3739687,'26,June,2004','2004-06-26 00:00:00','   hello,  well i went to a going away party last night for an english girl named miranda.  she\'s moving to chiang mai to study a form of massage.  we had a great time.  almost all of the people i know in thailand were there so it was nice to be around a large group of people and know the names of more than one person.  also, i met a few new people who were all wonderful.  the club was playing awful american music.  i even caught myself dancing to  britney spears!  i can\'t believe i just admitted to that.  no laughing.  i\'m going to blame it on the singha (thai beer).  which btw is pronounced sing.  i have on idea why they add the -ha to the end.  i\'ve been being a good girl and going to the gym everyday.  it\'s actually an american branch called california.  they have gyms in hong kong, sydney, dallas, and california (imagine that).  it\'s nice because they use pounds on all the equipment.  well i think i mentioned the large sex scene in thailand in my last message. what i forgot to include was that there is a long history of katoeys (or ladyboys) in thailand.  personally, i always have to have them pointed out to me.  i\'m so naive about things like that.  also, with many of them it is impossible to tell.  i would think they could fool most men.  that\'s all the thailand details i can think of for now. however, i have noticed a few changes in myself:  1. i now say university instead of college because that\'s what everyone else uses.  this is worldwide not just thai. sometimes i get confused looks from europeans when i say college.  also, an american girl, nancy from virginia, and i were talking today and we both laughed at ourselves for saying university to each other.  2. i\'ve begun to say i\'m doing well instead of i\'m doing good.  i\'ve always known that it was improper english but it\'s just so accepted in the states.  3. when asked where i\'m from is say america because most thai have only heard of the mississippi river.  i\'m just as bad with their geography.  i going to get a big map of thailand so i can look up all the places i hear people talk about.  it\'s annoying being so ignorant about the country i\'m living in.  i really should have done more research into the culture before coming here.  i have begun reading a book about the history of thailand so i guess that\'s a start.  4. my accent in going quickly.  since i teach conversational english i have to work on enunciating every word.  now it only comes out after a couple of singhas.  i do however still use many southern phrases such as ya\'ll and fixin\' to.  it makes my coworkers laugh.  well, i\'ve gotta run.  i\'m not sure what my plans are for tonight.  i might just stay in and read my book.  love all ya\'ll.  hope you mississippians are enjoying the summer.  it\'s actually been very nice the past few days here.  i hope it stays that way.  jane  '),(23471,3739687,'25,June,2004','2004-06-25 00:00:00','   hello everyone,  school is going great.  i\'m subbing today, but the teacher left work for the girls to do.  math.  ugh. anyway... i\'m sending this message out a little early today because i have plans tonight.  today the girls put on a play by sunthronpu (spelling ?) anyway... it was in all thai so i haven no idea what exactly was happening.  however some of the outfits the girls were wearing were western looking so i\'m assuming the play isn\'t very old.  i really need to learn more about thai history.  i have a book, but i\'ve been procrastinating.  anyway...  here are a few things about thai culture i\'ve learned recently.  some of this might be a repeat.  the girls believe that the darker a person\'s skin is then the more ugly they are.  exactly like the high yellow aspect of african american culture. also, in most of the ads and commercials they use asian people with western looking features instead of traditional thai looking people.  i think it\'s sad.  it\'s also weird because most of the westerners over here are more  interested in asian looking people.  you see many male farangs (foreigners) with thai women or men.  there is a large gay community in bangkok.  imagine that.  ha. ha.  well gotta go.  love ya\'ll.  jane  '),(23472,3739687,'24,June,2004','2004-06-24 00:00:00','   hello everyone,  well i just saw a good movie. \"mr. ibrahim\"  it was based on a book.  anyway it was good. i won\'t give away any details, but i\'d recommend it to all ages to see.  school is going fine.  i had a good time with all of my classes today.  also, i got to sub for the p.e. teahcer this morning.  i watched the girls play a thai version of tag.  they all stand in a circle, but they are in pairs.  one girl chases another girl who runs inside or outside of the cirlce. the chasee tags a girl and that girls partner has to run without being caught.  once someone is tagged they are \"it\" and the chase is reversed.  o.k. i hope all that made sense.  i\'m moving into my place tomorrow.  i\'m excited to have my own space.  also, it\'s great to know people in the same building so all i have to do is knock on a door to have someone to hang with.  i worked out at the gym today.  it\'s a really nice gym.  also, the equipment uses pounds so it\'s easy for me to know exactly how much i\'m lifting.  well i\'ve gotta run.  hope everyone is enjoying their summer.  it\'s actually not been too hot here so all ya\'ll in mississippi enjoy. ha. jk.  i miss you very much and hope whoever would like to will come see me.  jane  '),(23473,3739687,'23,June,2004','2004-06-23 00:00:00','   well no substitute teaching today.  i had the first graders three times however.  they are wild.  vicky and i are attempting to teach them how to sew.  it\'s a nightmare.  thirty kids running around with needles!  why the school made this part of the curriculum is beyond me.  anyway... other than that i had a great day.  some of the teachers and i met up for a game of b-ball after classes.  it was great.  i love being on the court again.  the girls love playing with the teachers too.  i\'m looking forward to friday when i\'ll be moving into my own room.  however vicky and i are considering some other options that might save us more money.  i just don\'t want to commute ever again.  well gotta go. i\'m off to the gym.  jane  '),(23474,3739687,'22,June,2004','2004-06-22 00:00:00','   hello everyone,  well school is going great.  i didn\'t have to substitute teach today so it was nice being on my actual schedule.  the only problem i had was with the 1/2 class which is the first graders.  they have so much energy and i wasn\'t sure how to handle them.  also, i\'m still in the habit of talking to fast.  i have to make myself remember to slow down.  other than that the day went by very smoothly.  tomorrow i\'m playing basketball again with the teachers and girls.  i had so much fun last week.  the kids love playing with the teachers too.  hopefully i won\'t twist an ankle or worse hurt one of them.  their only complaint last week was that we were throwing the ball a little to hard.  anyway... i\'m ready to hit the court again.  went shopping today.  i hate shopping but this was o.k. because i\'m soooooo tired of wearing the same clothes all the time.  i got a couple of things for work and one shirt for going out.  i guess it\'s nice going on a shopping spree every now and then.  i still have to pick up a few t-shirts, but after that i should be set.  well maybe one more pair of jeans would be nice.  this is one item of clothing that will be especially difficult to find.   thai women are short, small, and have zero hips.  the last time i tried on dresses they had to go to the back and find the largest size for me.  kinda embarassing. so i joined the gym today to work on this issue. jk. i just want to keep in shape and there is a very nice gym very close to where i live.  i\'m looking forward to getting on a regular workout schedule.  in oxford i had faye\'s hyper butt getting me out to the park everyday, but since i\'ve been here all i seems i do is eat.  well gotta go.  love ya\'ll.  jane  '),(23475,3739687,'21,June,2004','2004-06-21 00:00:00','   hey everyone,  well, today was a good day.  i got to sleep an hour later.  well actually i still woke up at six, but it sure was nice to be able to stay in bed until 7:15.  it\'s only a six or seven minute walk from  silom city court, where i now live, to st. joe.  also i\'m probably the most familiar with this area so it\'s nice to know exactly where i am.  bangkok can be extremely confusing at times.  anyway... i substituted again today. it was o.k. i\'ve devised a way to keep the girls in line so i mostly just get to read a book or do work for other classes.  well i thought i\'d give everyone an idea of what a day at st. joe is like. the day starts off with the girls lining up outside of their classrooms.  then an announement is made thru a pa system, always in thai so i have no idea what is being said, then the national anthem is sung by all the girls.  St. Joseph\'s is a large school 5,600, but it is spread out so while the girls from one side are singing the sound travels from the other side creating an echo effect. then the girls say something in thai which i haven\'t asked but i\'m assuming is something like the pledge of allegiance.  next they turn to the a statue of the virgin mary and pray to her, again all in thai.  finally they bow down on one knee and wai. it was extremely weird the first time, but now i\'m getting used to it.  then it\'s off to the classrooms.  which each grade stays in all day unlike in the states where the kids move from classroom to classroom.  at st. joe it\'s the teachers who trek around from one room to another.  we do have our own homeroom with a desk and file cabinet.  my home room is 4/1 there are 60 kids in each grade but in the elp (english learning program) which is the  side of the school i teach at they break the kids down into groups of 30.  so for each grade you have 1/1 and 1/2 and so on.  anyway... in the thai section, which is much cheaper to attend, all 60 kids are in the classroom at the same time. i don\'t have anything to do with that section, yet that is, so i\'m not sure what it is like.  i can say that instead of having all of their subjects taught in english like the elp girls are they only have  one course in english.  i guess similar to the way most american schools are set up, well except for the class size.  after lunch the kids line up again and the teacher leads them back to their classroom.   oh i almost forgot this; at the beginning of every class all the girls have to stand up and say, \"good morning miss jane\" then i have to give them permission to sit down.   then at the end they all stand and say, \"good-bye and thank you miss jane\".  the first time it happened i had no  idea what was going on and the girls ended up having to tell me to give them permission to sit.  however, despite all the differences a school is a school and everyday i see behavior that would be common anywhere in america or perhaps even the world.  such as teenage girls can be real bitches and the kids will do whatever necessary in order to get out of doing work.  well gotta go, jane  '),(23476,3739687,'20,June,2004','2004-06-20 00:00:00','   hey friends and fam,  well i went to a concert last night with vicky and nancy (friends from st. joe)  we had a great time on khao san but the music in the park that we heard sucked.  good songs but terrrible covers. if you can\'t do a song justice i think you shouldn\'t even try to sing it.  anyway... i met someone.  so that\'s good news.  we had a great time and hung out all night together.  i\'m looking forward to school next week believe it or not!  also i\'m moving into the place i\'ll be living in at for today.  sunday for ya\'ll in the states.  if anyone needs my new address just email my mom at REDHILLSWEAVER@MCHSI.COM  she\'s got all the info on address and phone numbers.  love ya gotta go.  jane  '),(23477,3739687,'01,July,2004','2004-07-01 00:00:00','   hello,  well i\'ve been here long enough to miss a few things about mississippi.  here is a list of things i hope everyone down south is enjoying:  1.  taking long country drives with friends, while listening to the grateful dead. 2.  fresh tomatoes from the garden (even though the deer ate my entire garden last year) 3.  having a tomato garden. 4.  fresh air.  bangkok is extremely polluted. 5.  playing ball with faye in the front of oxford apartments while cooking veggies on the grill and drinking a budweiser.  bud is as expensive in grocery stores over here as it is in the bars of mississippi. 6.  my family and friends.  i know it\'s cheesy but i do miss you.  hope everyone is having a good time.  i love ya.  jane  ps i got my first paycheck today! go me.  '),(23478,3739687,'30,June,2004','2004-06-30 00:00:00','   hello,  well this has to be a quick message because i\'m running late.  i\'ve been helping out with sewing again today.  first graders with needles!  i think that is all that needs to be said.  surprisingly they haven\'t taken one another\'s eyes out.  yet.   anyway...  i also had the first grade for reading class which went very well because it was all about magic.  they loved it.  actually, school has been great today because sister erlinda is  busy with parent meetings.  when the cats away!  so i took a two hour lunch with a few other folks.  on fiday the school is moving into new facilities.  it\'s going to be anarchy.  nothing ever works as planned in thailand especially at st. joe.  i\'m looking forward to the chaos.  i just hope i don\'t get hurt in all the running around.  gotta run. jane  '),(23479,3739687,'29,June,2004','2004-06-29 00:00:00','   hello,  well i finally bought an umbrella just in time for a real downpour.  unfortuantely i left it in the room so it was of no help.  guess i\'m still learning.  hell i know i\'m still learning.  i had an interesting encounter with a crazy person this afternoon.  i didn\'t notice her, but some woman followed me home from the gym.  she was already in the hall of my apartment building before i saw her.  then she started talking to me in thai.  i noticed that she seemed a little off so i rushed to my room and got in just in time.  she tried to open my door which i had locked and latched VERY QUICKLY.  she didn\'t bang on the door or anything super scary but she moaned in the hallway for about thirty minutes.  finally she harassed someone coming down the hall and they informed the front desk about her presence. i\'d tried calling the front desk, but they don\'t have very good english skills.  the guy thought i was trying to rent a room for  one of my friends.  anyway... i\'m o.k. the front door guy apologized to me and said that whenever that happens to just tell him \"crazy in hotel\".  hope this doesn\'t make anyone worry too much about me.  really i\'m in a safe area and i have a safe place to live.  i was just rushing home because of the rain and not paying attention to my surroundings.  so i don\'t want any emails about being careful.  o.k. love ya\'ll.  anyway onto my day.  well i substituted for the high school girls again today.  they hate me because i\'ll only let them speak in english.  even when the lesson is over i tell them they can do homework or talk freely as long as i don\'t hear any thai.  we actually read a story about oprah winfrey in their book before i let them have free time.  they were amazed that she was from the same area of the world that i\'m from.  well i know for a fact that nilza, the woman i co-teach with, is leaving.  she\'s very unhappy in thailand.  also, she\'s from brazil so i think the fact that they have her teaching english is a little overwhelming for her.  anyway... hopefully i\'ll be given her schedule.  then i\'ll actually have something to do when i\'m at school.  right now i just have to guess as to what the teacher i\'m subbing for wants me to do.  while i\'ve gotten used to it i would like a regular schedule where i can actually get to know some of the girls.  hope everyone is enjoying their summer.  love ya.  jane  '),(23480,3739687,'28,June,2004','2004-06-28 00:00:00','   hey all,  well i got a compliment from one of the other teachers today.  she told me i did a good job keeping the girls in line.  it\'s a big problem at st. joe because most of the girls are rich little snobs.  they\'ve never had to work for anything and they have awful attitudes.  well most not all and it gets worse as they get older which makes the compliment even more meaningful because it was grade 10/2 which is the final grade.  it really made my day.   it\'s kinda sad how i go about getting them to shut up.  all i have to do is threaten them with an english assignment and they go completely mute.  which is amazing since they\'ve been studying it since they were six.  unfortunately st. joe tends to teach through handouts and vocabulary lists as opposed to conversational or any other style that makes the girls actually have to think.  anyway... everything else is going great.  i\'m starting to fall in love with bangkok.  oh! also, i just found out that i\'m getting a week off in july the 12-16 or something like that.  they just bothered to tell us last week.  the farang teachers are always the last to know everything. very frustrating.  right now i have plans to go to the beach or i might go with vicky to hong kong.  ahh!! so many possibilities.  i\'m so excited.  we always have a good time together.  well i\'ve gotta run.  hope ya\'ll are stayin\' cool down there is mississippi.  love ya, jane  '),(23481,3739687,'27,June,2004','2004-06-27 00:00:00','   hey all,  well it\'s official.  i\'m a fruit.  i can\'t believe how wonderful all of the fruit is over here.  i like everything i\'ve tried.  i\'m on to durian next.  i\'m not sure if that is how you spell it, but anyway... it\'s stinky.  hey i\'ll try anything once.  i think the only thing i\'ve eaten here that i did not like was a tamarind (again spelling ?)  it tastes similar to a raisin.  well not much to say today.  i went to brunch at kuppa\'s in sukhumvit.  which is only yards from molly\'s place.  it was nice to be in that area again.  we bummed around at an art gallery for a while and then went shopping.  i\'m slowly collecting a wardrobe.  things are cheap here but i have to force myself to think in baht instead of dollars since that\'s what i\'m being paid in.  ohhh!!!! payday is soon.  i\'m going to open a bank account here so it will be easier to put myself on a budget.  well gotta run.  i hope everyone is doing fine.  jane  '),(23482,3739687,'10,July,2004','2004-07-10 00:00:00','   hello everyone,  well it feels great to be on vacation.  i don\'t have to work again until  next monday.  there is a big aids conference here in bangkok and i think  they are trying to encourage people to leave town to cut down on traffic.   anyway...  everything is going great no poop incidents yesterday.  i\'ve  begun playing games that encourage the girls to work on their pronunciation.    thai children/people have difficulty with the r and l sounds so words like  river and liver come out sounding the same.  i think the girls enjoyed it so  i\'m going to continue with as many games as possible.  other than that not much to tell.  i might miss a message or two this week  but my mobile should work if anyone needs to hear my voice.  love ya.  jane  _________________________________________________________________ FREE pop-up blocking with the new MSN Toolbar Â– get it now!  http://toolbar.msn.click-url.com/go/onm00200415ave/direct/01/  '),(23483,3739687,'09,July,2004','2004-07-09 00:00:00','   hey all,  well i\'m attempting to start sending my emails to a website.  apparently you  can just go to the website and see all of my entries.  this is just a check.  guess we\'ll see how it goes.  jane  ps thanks doug  _________________________________________________________________ From Â‘will you?Â’ to Â‘I do,Â’ MSN Life Events is your resource for Getting  Married. http://lifeevents.msn.com/category.aspx?cid=married  '),(23484,3739687,'17,July,2004','2004-07-17 00:00:00','   hello,  well i\'m back in bangkok at my favorite internet cafe thanking god that red  ants aren\'t running all over me.  that was one reason my emails from koh  chang were so short.  the other reason was that i didn\'t do anything other  than see the waterfall.  i love doing nothing.  i was basically laying  around a covered pavillion on coushions all week.  we played cards the  entire time. spades and chaw di de (i have no idea how to spell this one)   anyway...  chaw di de is a game from hong kong that vicky wong taught us to  play.  i taught everyone spades and between the two games we played several  rounds of cards.  i guess it might sound a little boring to some of ya\'ll  but it was exactly what i needed.  tonight i have a dinner cruise on the river.  it should be fun i\'ll let you  know the details tomorrow.  so now that i\'m back i have to stop talking about taking thai lessons and  get started.  i\'ve heard of a few people willing that are qualified so  monday i\'m going to start calling.  wish me luck.  i know i need it.  love  ya.  jane  _________________________________________________________________ MSN Toolbar provides one-click access to Hotmail from any Web page Ã‚Â– FREE  download! http://toolbar.msn.click-url.com/go/onm00200413ave/direct/01/  '),(23485,3739687,'16,July,2004','2004-07-16 00:00:00','   well we\'re leaving the beach for bangkok in a few moments.  it was a great  vacation.  i did absolutely nothing other than eat, sleep and play cards.   wonderful.  anyway...  i hope all ya\'ll in mississippi are enjoying the heat.  i heard the heat  index was 115 the other day.  damn.  keep the ac on and drink lots of water.    love ya.  jane  _________________________________________________________________ Get fast, reliable Internet access with MSN 9 Dial-up Ã‚Â– now 2 months FREE!   http://join.msn.click-url.com/go/onm00200361ave/direct/01/  '),(23486,3739687,'15,July,2004','2004-07-15 00:00:00','   Jane,   &nbsp;   This is Michael Cravens from Louisville - I haven\'t seen you in years! Your dad has been sending me your emails from Thailand. What you\'re doing is fascinating!&nbsp; Your last email about showers, squat toilets, and mosquitos made me laugh - because it brought back many memories of my missionary trip for two months to South Korea several years ago.   &nbsp;   It sounds beautiful and I\'m sure that this&nbsp;will be a&nbsp;great experience for you.   &nbsp;   God bless. Write only if you have time. Otherwise, your dad will keep me up to date.   &nbsp;   I\'ll be praying for your safety and a fun experience.   &nbsp;   Michael &nbsp;    Mayo2bog &lt;mayo2bog@telepak.net&gt;   wrote:   From: \"Mayo2bog\"   To: \"Michael\"   Subject: Fw: no running water Date: Mon, 12 Jul 2004 17:59:11 -0500    ----- Original Message ----- From: Jane Mayo   To:  ;  ;  ;  ;  ;  ;  ;  ;  ;  ;  ;  ;  ;  ;  ;  ;  ;  ;  ;  ;  ;  ;  ;  ;   ;  ;  ;  ;  ;  ;  ;  ;  ;  Sent: Monday, July 12, 2004 7:48 AM Subject: no running water   well hello everyone,  i\'m down on koh chang (i\'ve been spelling it wrong) for the next few days. we got here last night and are going to head back friday. anyway... vicky told me that it was a very basic place where we were going to be staying but not many other details. i\'m o.k with squat toilets and staying without aircon but i do have a small issue with no where to take a shower. i basically have to stand next to a pool or standing water and use a bucket to rinse myself off. thankfully some of the other people we came down with, christian from new zealand and kee from bangkok, were unwilling to live that way and got a very nice room  right up the beach. THANK GOD! now i can go and have a proper shower when i need one.  other than that i love the place. great music, great food, and friendly people are everywhere. we are staying in a bungalow right on the ocean. the mosquito netting keeps us from being eaten alive while the sounds of the waves crashing against the rocks lulls me to sleep. i absolutely love it here. i might should look into job openings. jk. it\'s nice for a vacation but i don\'t think there is much to do here other than lay on the beach and eat good food.  vicky has a digital camera so i\'ll be sending pics when i get back. this is a promise. well gotta run. hope everyone is enjoying their summer.  love ya, jane  _________________________________________________________________ Get fast, reliable Internet access with MSN 9 Dial-up - now 2 months FREE! http://join.msn.click-url.com/go/onm00200361ave/direct/01/       		 Do you Yahoo!?   urlLink New and Improved Yahoo! Mail  - 100MB free storage!  '),(23487,3739687,'13,July,2004','2004-07-13 00:00:00','   hello,  well i\'m still at the lonely beach in koh chang.  we are having a great time  hanging out in our hut listening to the sound of the ocean.  it\'s really  wonderful here.  today we went up to a waterfall on the island.  it was  gorgeous.  it looked like something out of patagonia magazine.  anyway....  other than that not much to tell.  i hope eveyone is enjoying their summer.   i\'m certainly loving thailand.  i\'ve gotta run.  i promise i\'ll write a longer letter later.  jane  _________________________________________________________________ MSN Life Events gives you the tips and tools to handle the turning points in  your life. http://lifeevents.msn.com  '),(23488,3739687,'12,July,2004','2004-07-12 00:00:00','   well hello everyone,  i\'m down on koh chang (i\'ve been spelling it wrong)  for the next few days.   we got here last night and are going to head back friday.  anyway... vicky  told me that it was a very basic place where we were going to be staying but  not many other details.  i\'m o.k with squat toilets and staying without  aircon but i do have a small issue with no where to take a shower.  i  basically have to stand next to a pool or standing water and use a bucket to  rinse myself off.  thankfully some of the other people we came down with,  christian from new zealand and kee from bangkok, were unwilling to live that  way and got a very nice room right up the beach.  THANK GOD! now i can go  and have a proper shower when i need one.  other than that i love the place.  great music, great food, and friendly  people are everywhere.  we are staying in a bungalow right on the ocean.   the mosquito netting keeps us from being eaten alive while the sounds of the  waves crashing against the rocks lulls me to sleep.  i absolutely love it  here.  i might should look into job openings. jk. it\'s nice for a vacation  but i don\'t think there is much to do here other than lay on the beach and  eat good food.  vicky has a digital camera so i\'ll be sending pics when i get back.  this is  a promise.  well gotta run.  hope everyone is enjoying their summer.  love ya, jane  _________________________________________________________________ Get fast, reliable Internet access with MSN 9 Dial-up Â– now 2 months FREE!   http://join.msn.click-url.com/go/onm00200361ave/direct/01/  '),(23489,3739687,'24,July,2004','2004-07-24 00:00:00','   hey all,  well it is official the woman running my school is crazy.  apparently she  has a problem with the farangs.  the women she has hired have breasts.  in  her words, \"too bulky up top, wear loose clothes, when you look in the  mirror in the morning say am i going karaoke or to school.\" as you can  imagine i busted out laughing.  also, she has a problem with how we cross  out legs.  we can only cross our ankles.  WHATEVER!  with all the problems  at that school i can\'t believe she is focusing on how we are sitting.   nutcase.  i\'m just amazed someone like her is allowed to run a school.  actually, i take that back mr. cole, my principal, was just as crazy and  unqualified.  maybe this profession just breeds lunatics.  oh well, i plan  on sitting how i please.  she can\'t afford to fire me right now because she  just had three people quit. HA!  i don\'t blame them.  one of them was a  really nice Filipino woman named grace.  they pay filipinos slave wages for  doing more work than anyone else.  it really sucks for them here.  well i\'ve gotta run.  i\'m meeting a friend for brunch. love ya, jane  _________________________________________________________________ DonÃ‚Â’t just search. Find. Check out the new MSN Search!  http://search.msn.click-url.com/go/onm00200636ave/direct/01/  '),(23490,3739687,'23,July,2004','2004-07-23 00:00:00','   hello,  well i found a bootleg copy of farenheit 9/11 on the street yesterday.  even  though i don\'t like micheal moore i thought he did a good job on the  documentary.  definately one sided but not as over the top as he usually is.    anyway... i know some of you have seen the movie.  what did you think?  oh  yeah i only paid $3 for the dvd btw.  ha!  all you silly americans paying $6  dollars to see it in a theatre! jk.  well anyway....   i went to a street vendor yesterday that i used to go to  all the time, but have not visited recently.  i realized while sitting there  that i could see convent rd. (where st. joe is) and my gym from my table.   it was very surreal to realize that when i first got here and was walking  around not knowing where to eat or what to do or even at times where i was  that i ended up at a place that would become so familiar.  now the vendors  all recognize me and are attempting to teach me words in thai.  i\'m a  terrible student!  well i\'ve gotta run.  there is a teacher\'s meeting today in sister erlinda\'s  office.  yuck.  no jk she\'s been so busy lately she hasn\'t had time to  harass the farangs.  jane  _________________________________________________________________ Overwhelmed by debt? Find out how to Ã‚Â‘Dig Yourself Out of DebtÃ‚Â’ from MSN  Money. http://special.msn.com/money/0407debt.armx  '),(23491,3739687,'22,July,2004','2004-07-22 00:00:00','   hey everyone,  well considering it is monsoon season i really haven\'t had to deal with much  rain.  knock on wood.  however yesterday i did get stranded at a friends  place for a couple of hours.  the rain here is different than in  mississippi.  it comes down very hard and very fast.  it rarely drizzles or  mists all day long like it does for days and sometimes weeks in ms.  every  time it happens i\'m always without my umbrella.  i\'ve just got to start  carrying it eveywhere with me.  other than that not much to say.  i just found out two days ago that i do  have to give summative exams.  just another example of how disorganized the  school is.  so now i\'m stuck in the computer room all day making an exam for  a class that i have only taught once.  crazy nuns.  well one more thing i\'ve started to notice about myself is that some of the  things that amazed me a first in thailand are now either annoying or passe.   for example on silom rd. i often see elephants hanging out with their  handlers at first i was amazed, then i was told that the elephants are  mistreated and that it is bad for them to be in the middle of a busy ,  pollution filled street.   so now i feel sorry for them.  anyway... gotta  run.  tests to be made.  jane  _________________________________________________________________ FREE pop-up blocking with the new MSN Toolbar Ã‚Â– get it now!  http://toolbar.msn.click-url.com/go/onm00200415ave/direct/01/  '),(23492,3739687,'21,July,2004','2004-07-21 00:00:00','   hello,  well things are going o.k. here.  moving to new facilities has caused the  chaos that i imagined, but i means that sister erlinda is to busy to bother  the farangs.  thank god.  also, she has it in her mind that i am an  exceptionally good teacher which means i am flying under the radar right  now.  i can pretty much do as i please in the classroom which is good and  bad.  since i don\'t have any books to go by i really have to be creative and  sometimes it works perfectly and other times i have to abandon the lesson  plan and wing it.  some things i\'ve started to notice about thailand recently:  1.  there are some very large thai people.  at first i thought everyone was  a size one but now i\'m noticing that there are quite a few obese people  walking around.  i would be willing to bet that this is due from the  american influence here.  2.  people aren\'t as strict about their cultural rules as i thought they  would be.  i often see people wearing sleeveless shirts, pointing at people,  and changing in front of one another at the gym.  (when i first joined i was  told that only farangs change in front of others and that i should go to the  changing room)  3.  while the thai are buddist this does not mean that a fight won\'t break  out.  two days ago i saw a brawl in the middle of the street.  men and women  were fighting and throwing things at one another.  i\'m not sure what was  going on but i think it has to do with a new food vendor moving into  another\'s territory.  well i\'ve gotta go teach some little monsters.  hope eveyone is enjoying  themselves.  also anyone headed to the neshoba county fair think of me.  i  hate that i\'m missing it.  love ya, jane  _________________________________________________________________ MSN Toolbar provides one-click access to Hotmail from any Web page Ã‚Â– FREE  download! http://toolbar.msn.click-url.com/go/onm00200413ave/direct/01/  '),(23493,3739687,'20,July,2004','2004-07-20 00:00:00','   wow,  it is amazing how easy it is to spend money in thailand.  eveything is so  cheap that i justify every purchase by saying that it would be three times  as much in the states.  i\'ve got to stop doing this.  i\'m making baht so i  need to think in baht instead of dollars.  anyway...  everything is going  great here.  we moved into the new building yesterday.  it is very nice and  all the rooms have aircon so no more standing in front of thirty kids  sweating my ass off.  thank god.  i\'m doing more computer research this afternoon.  hopefully i\'ll have a new  laptop by the end of the week.  well gotta run.  love ya, jane  _________________________________________________________________ Planning a family vacation? Check out the MSN Family Travel guide!  http://dollar.msn.com  '),(23494,3739687,'19,July,2004','2004-07-19 00:00:00','   hey everyone,  well i\'ve gone from the calm of koh chang to the chaos of st. joe.  the  school is moving into new facilites today.  why they didn\'t move everything  on our week off is beyond me!   now there are 7 year olds moving boxes which  i wouldn\'t carry up four flights of stairs.  anyway...  i\'m glad i don\'t  have very many classes this morning.  other than that not much to tell.  i almost bought a computer that i would  have regretted purchasing yesterday.  very hi-tech and way too much money.   the internet really is wonderful sometimes.  i went to a site that warned me  against that model.  yeah me!  thank god the woman couldn\'t get my credit  card to work yesterday or i would have a crap computer.  well gotta run before sister erlinda catches me in the students only  computer room.  jane  _________________________________________________________________ Is your PC infected? Get a FREE online computer virus scan from McAfeeÃ‚Â®  Security. http://clinic.mcafee.com/clinic/ibuy/campaign.asp?cid=3963  '),(23495,3739687,'18,July,2004','2004-07-18 00:00:00','   hello,  well i went and looked at laptops today.  fun fun.  i basically got a crash  course in computers from a friend.  thank god for computer geeks.  anyway...    other than that not much has been going on i ate some great food on the  dinner cruise.  unfortunately i was seated on the end so i missed some of  the sights, but hey the wine was good.  also, the food was great and the new  people i met were all cool.  there were lots of canadians on the trip for  some reason.  well school starts tomorrow.  yuck.  i\'ll be doing lots of research and also  we are moving into a new building so i\'m sure it\'s going to be anarchy.  gotta run.  love ya.  jane  _________________________________________________________________ Discover the best of the best at MSN Luxury Living. http://lexus.msn.com/  '),(23496,3739687,'31,July,2004','2004-07-31 00:00:00','   hey everyone,  well i got my new laptop.  yeah it is an hp.  so far i\'m extremely happy  with it, i still have to get it hooked up to the internet.  contrary to what  most people believe it is not that cheap to buy technology here.   i mean  you can but you have to sacrifice things that are now standard on my things  at home.   anyway... yeah!  i have a new toy.  last night i watched crouching tiger hidden dragon on it to check out the  quality of the screen and it was awesome.  i\'d forgotten how much i like  that movie.  if anyone hasn\'t seen it i highly reccomend it.  anyway... i\'m at the school right now.  i have to do some work they just now  told me i was supposed to do.  the organization a st. joe is so incredibly  bad i\'m not surprised by anything anymore.  well gotta run.  love ya.  jane  _________________________________________________________________ MSN Toolbar provides one-click access to Hotmail from any Web page Ã‚Â– FREE  download! http://toolbar.msn.click-url.com/go/onm00200413ave/direct/01/  '),(23497,3739687,'29,July,2004','2004-07-29 00:00:00','   well hello everyone,  one thing i\'ve noticed about thailand is that everything takes twice as long  as you are told.  so therefore i\'m still waiting on my laptop which was  supposed to be in on monday or tuesday.  a little aggrativing but i know i\'m  getting a good deal so i guess it is worth waiting for.  anyway... other  than that i\'ve just been invigilating all day.  which is the most mind  numbing thing i\'ve ever done.  i just sat there and watched kids take exams.    ugh!  last night i went out with aurora to a place called the londoner.  it was  great.  also free which made it awesome.  her landlord was there so he  picked up the bill.  between him and niles withers i\'ve eaten two extremely  good meals this past week.  it\'s so nice to be treated to good food.  anyway... i\'ve gotta run.  nothing else much to say other than erlinda is  back.  she was gone all for the first three days of the week and it was  wonderful.  now she is back raising hell.  i think she thinks that when she  is gone we do nothing.  not true we usually get more done because we don\'t  have her bugging us about stupid things.  well gotta run.  jane  _________________________________________________________________ Planning a family vacation? Check out the MSN Family Travel guide!  http://dollar.msn.com  '),(23498,3739687,'28,July,2004','2004-07-28 00:00:00','   hello,  well i caught a girl attempting to cheat yesterday.  i was invigilating  (which btw means standing in front of a class a looking at them take a test)  anyway... on one of my trips around the room i saw thai writing on a girls  leg.  i couldn\'t read it and i didn\'t want her to know that i saw it so i  just watched her during her math exam.  she wasn\'t looking at her leg so i  figured it was for her next exam which was in thai.  anyway... i told one of  the thai teachers and they busted her.  apparently it was for her next exam  and i caught her just in time.  GO ME!  my question is WHY CHEAT!  you can\'t fail so why bother to cheat.  hell, the  school does it for you! i just don\'t understand this place sometimes.  i\'m  also not looking forward to grading my exams.  the kids take them tomorrow.   who knows if they even studied.  well other than that things are going great.  i had three things to attend  last night, but i only made it to two.  i guess that\'s o.k. the first was a  b-day party for a guy from london.  it was lots of fun.  almost everyone  that i like from work was there.  the next was a going away party for a girl  from france.  she went with me to ko chang.  eve is her name and i think i\'m  really going to miss her.  we had just started to talk and now she is gone.   well thank god for the internet.  i\'ve gotta run.  i hope everyone is enjoying the fair.  love ya.  jane  btw if anyone wants off of my email list let me know.  since i have the blog  set up it might be easier for you to just check it periodically.  if you  need the link i can send it to you again.  _________________________________________________________________ DonÃ‚Â’t just search. Find. Check out the new MSN Search!  http://search.msn.click-url.com/go/onm00200636ave/direct/01/  '),(23499,3739687,'27,July,2004','2004-07-27 00:00:00','   hello,  well dinner with niles withers went great.  he is a very nice man with a  great sense of humor.  it was wonderful to hear old names from louisville  that i haven\'t heard since i was in high school.  also, he seems to be  keeping up with ms news so we talked about the neshoba county fair and eli  manning.  nice to talk about things i\'m so familiar with.  anyway... i have  three different parties to go to tonight.  first is a b-day party for a  friend from work, then onto a presentation given by my friend christian with  free food and drinks, and finally a going away party for a french girl who  is a volunteer here.  well gotta run.  wish i could write more but i\'m giving summative exams  today and i CANNOT be late.  love ya.  jane  _________________________________________________________________ DonÃ‚Â’t just search. Find. Check out the new MSN Search!  http://search.msn.click-url.com/go/onm00200636ave/direct/01/  '),(23500,3739687,'26,July,2004','2004-07-26 00:00:00','   well,  summative exams are pointless and boring.  since you can\'t fail a kid at st.  joe becasue of the whole losing face thing then why bother with tests.   anyway... all i\'ve been doing all day is standing in front of grade 4 making  sure they don\'t cheat.  o.k. enough bad now some good news.  yesterday i went to the weekend market  with aurora, teacher from st. joe, and had a wonderful time.  she has lots  of thai friends and they speak very little english so it forces me to use  thai to communicate.  but hey immersion is the best way to learn i think.   well i\'ve gotta run.  oral exams are next and i have to figure out what i\'m  doing.  love ya, jane  ps dinner tonight with niles withers  _________________________________________________________________ DonÃ‚Â’t just search. Find. Check out the new MSN Search!  http://search.msn.click-url.com/go/onm00200636ave/direct/01/  '),(23501,3739687,'25,July,2004','2004-07-25 00:00:00','   hello,  well i went to a fun cool place last night with a group of friends.   actually, two places the first was H1, very cool.  all of you artist on the  list would have loved it.  anyway... the second was a place called zantika  which was also very cool.  i liked it because i finally felt like i was  hangin\' where thai people hang.  everyone was dancing to a mixture of music  which was unfortunately bad american pop, but still it was loads of fun.   thai people seem to all dance the same to me. it is very fast and i just  can\'t move my body that quickly.  anyway... they all thought i was dancing  crazy so they laughed but i didn\'t care.  anyway... i was there with aurora  (from oregon hippie parents whole name aurora dawn wood)  she attempted to  teach me a few salsa moves. i think i was a poor student.  at least i tried.  well other than that... i was thinking of a few things that i miss about the  states.  1. jon stewart  2. driving  3. good music  4. all you guys  gotta run.  love ya.  jane  _________________________________________________________________ Planning a family vacation? Check out the MSN Family Travel guide!  http://dollar.msn.com  '),(23502,3739687,'07,August,2004','2004-08-07 00:00:00','   hello everyone,  well tgif!  i went out with some folks from the school last night.  we  started at the irish exchange and then moved on to a place called noreiga\'s.    it was lots of fun.  they were having an 80s night which isn\'t my favorite  music, but it is tons of fun to dance to.  anyway... i was supposed to met  up with aurora only but tons of people i knew just happened to be there.   some friends i met through molly and others from the school.  it was weird  eveytime i looked up i would see someone i knew.  it reminded me of how in  oxford, ms that would always happen.  other than that i don\'t have much to tell.  it\'s been raining more lately.   i think i\'m understanding the meaning of monsoon season now.  actually it  think it is starting to rain right now.  i usually just stay in my room  until it stops, but i always end up being caught out without my umbrella.  oh also, i finally got the internet connection in my room working.  yeah me.    i guess i\'m not quite as computer illiterate as i thought i was.  o.k. maybe i have more to say than i thought.  sister erlinda called an woke  me up this morning. crazy nun with a bad habit. i programmed in that number  as well, but the sneaky lady most know this and use a different phone  everytime.  she might be nuts but she is no idiot.  well gotta run.  love  ya.  jane  ps i have a little room on my email list now so if you know of anyone who  wants on it let me know.  also you could give them the address for the blog.  _________________________________________________________________ Express yourself instantly with MSN Messenger! Download today - it\'s FREE!  http://messenger.msn.click-url.com/go/onm00200471ave/direct/01/  '),(23503,3739687,'06,August,2004','2004-08-06 00:00:00','   hello,  well yesterday i almost had to take a cell phone from a first grader.  i  couldn\'t believe it. i mean i\'ve seen the older kids with them, but a first  grader!  she even had the nerve to make a call during the middle of my  class.  i wish i had of taken it from her because it was a really nice  mobile phone.  jk.  anyway...  things are going o.k. here.  i\'m glad the  week is almost over.  i have a fairly easy day today so i\'m just looking  forward to getting everything done.  this weekend i\'m going to a party for a french girl named ermine.  she\'s  been here for one year and is now headed back to france.  i think she is  going to continue her studies there, but i\'m not sure what she will be  studying.  anyway... lately i\'ve spent a good bit of time with her and some  of the other french people.  also, i don\'t think they suck anymore.   acutually they are very cool once you get to know them.  i think the problem  at first was that their english is not that good so they might feel a little  intimidated when speaking it to me.  my french is even worse so they enjoy  teaching me things that i cannot say and laughing at my drawl when i do  attempt it.  well i\'ve gotta run.  ya\'ll be good.  jane  _________________________________________________________________ FREE pop-up blocking with the new MSN Toolbar Ã‚Â– get it now!  http://toolbar.msn.click-url.com/go/onm00200415ave/direct/01/  '),(23504,3739687,'05,August,2004','2004-08-05 00:00:00','   hello,  how is everyone?  i\'m doing good just chillin at work.  i only have three  classes today so i have lots of time to do internet research.  well actually  right now i\'m just instant messaging with friends in the states, but i do  intend to get started on work stuff soon.  i need to find good learning  games and activities for the kids.  i\'m still figuring out their level so  it\'s a bit difficult.  i went to dinner last night with a girl from england named emma.  she\'s lots  of fun to hang around and is always using funny english expressions i\'ve  never heard.  she\'s been traveling for almost a year all over the world.   it\'s great to hear her stories about the different places she has been.   india, nepal, cambodia, vietnam, australia, and laos are just some of the  places she has been.  wow.  it makes me so jealous i want to see all these  places too.  i think the travel bug bit me over here.  well i\'ve gotta run.  hope everyone is enjoying their summer  jane  _________________________________________________________________ FREE pop-up blocking with the new MSN Toolbar Ã‚Â– get it now!  http://toolbar.msn.click-url.com/go/onm00200415ave/direct/01/  '),(23505,3739687,'04,August,2004','2004-08-04 00:00:00','   hello,  well i went to the hospital yesterday.  bangkok christian hospital right  acroos silom road, which is very close to my school and my apartment.   anyway... it was great.  no long wait and the doctor said just as much as  they always do.  i have an ear infection, that i\'ve had since ko chang, so  he gave me penicillan and ear drops.  probably exactly what they would have  done in the states.  only hear it was very inexpensive, about $22.  not bad  for a hospital visit i think i should get sick while i\'m here. jk.  other than that i don\'t have much to say.  school has been keeping me very  busy lately.  i actually feel guilty about writing this email because i\'m  supposed to be grading papers, doing research, and completeing my lesson  plans.  ugh.  i have so much to catch up on i\'m not sure where to start.   they sure do keep us busy around this place.  also, sister erlinda is  freaking out about everything today because the ministry of education is  coming to visit the school.  however, the kids are behaving even my least  favorite class was manageable today.  wow.  i wish they were like that  everyday.  gotta run.  love ya.  jane  _________________________________________________________________ Planning a family vacation? Check out the MSN Family Travel guide!  http://dollar.msn.com  '),(23506,3739687,'02,August,2004','2004-08-02 00:00:00','   hello,  well the inevitable happened sister erlinda found out my cell phone number.   damn.  i thought i had hidden it well, but someone must have told her or i  accidently put it on one of the half dozen forms she has us filling out  every week.  anyway...  she called last night to see if i could take an  extra class today.  since i\'m more than willing to pick up a few extra  classes a week i wasn\'t pissed about the phone call just the fact that now  she has my digits and i know it won\'t be the last time she calls.  ugh!  oh  well nothing i can do now i programmed the number she called me on into my  phone so hopefully i can screen her out, but she is very determined when she  wants to talk with you.  crazy nun.  anyway...  i\'m having nun of that. ha.  o.k. i know i\'m a dork.  well school is going o.k. i\'m working with the girls.  i wish i new more  about teaching phonetics because i think i have a rare opportunity to really  work with the girls.  since i don\'t have a book and i am basically creating  a class as i go i have lots of directions i could go but i\'m unsure what is  the best for the girls.  also, i have so many ages that it is also difficult  for me to complete my lesson plans.  i just don\'t know until i get into the  classroom what is going to work and what is not.  but hey i\'m staying  flexible and even when i have to scrap a lesson plan i still come up with  something for them to do.  well i hope eveyone is enjoying that mississippi heat.  i love ya.  oh wait one more thing.  i\'ve been getting lots of emails saying things  about thailand being so exotic and third world.  oxford, ms is third world  compared to bangkok.  there is everything here that you can imagine, well  except good soul food.  i think some of your perceptions about how and where  i am living might be a little off.  with six million plus people bangkok has  a little of everything.  i hope anyone who wants to can come over for a  visit.  i\'d love for ya\'ll to see the place.  jane  _________________________________________________________________ Planning a family vacation? Check out the MSN Family Travel guide!  http://dollar.msn.com  '),(23507,3739687,'01,August,2004','2004-08-01 00:00:00','   hello,  well things have been going great here.  i\'m really enjoying thailand.   yesterday and today are some big buddhist holiday so eveyone seems to be  smilling even more than they do regularly.  i can\'t remember the name of the  holiday.  i really need to get another book on thailand.  i\'ve read one but  it was just a historical narrative that explained why thailand was never  colonized.  anyway...  i went to soi 4 last night but they weren\'t serving alcohol because of the  holiday.  which worked out great for me because i\'m trying to drink less.   i\'ve figured out that if you don\'t drink it is easy to save money here.   well you can\'t save much but it will certainly make the ends meet.  thailand is becoming more and more familiar to me.  i think i could be here  for ages and never know everything but i\'m starting to recognize more and  more.  the fruit vendor on the end of my soi (street) recognizes me and  teaches me a new thai word every time i buy fruit from him.  i feel so  healthy here.  i work out at my gym, eat fruit and som tam, spicy papaya  salad that still sets my mouth on fire, also i walk almost everywhere.  if  it wasn\'t so polluted here it would be paradise.  unfortunaley the pollution  can almost be unbearable at times.  i especailly hate it when i\'m eating at  my favorite tofu street vendor and a tuk-tuk comes by blowing carbon  monozide right at me.  ugh!  other than that i\'ve just been chillin\' out with friends at the pool and  trying to make myself do work for school.  i actually went in yesterday but  not today.  sister erlinda was so surprised to see me that she didn\'t even  say anything about my shoulders not being covered.  wow. anyway... i didn\'t  know on our days off we still had to abid by all the rules.  crazy  catholics.  jk.  love ya, jane  _________________________________________________________________ Discover the best of the best at MSN Luxury Living. http://lexus.msn.com/  '),(23508,3739687,'14,August,2004','2004-08-14 00:00:00','   hello,  well i woke up yesterday with a sore throat and instead of being a good girl  and staying in last night i decided to still go to soi 4.  btw soi means  street.  anyway... as you can imagine it was awful when i woke up this  morning.  i don\'t really feel bad it\'s just annoying.  so now i\'m not going  to be able to make it to a friends gig tonight.  i really wanted to go  because this is the third time that he\'s invited me to see his band play,  but i just don\'t think i should.  anyway... today i stocked up on orange juice, ice cream, yougurt, and fruit  so i can be sure to get all of my vitamins.  i also layed out for a couple  of hours to try and sweat any tokins out.  hopefully tomorrow i\'ll be back  to 100%.  well gotta run.  jane  _________________________________________________________________ Express yourself instantly with MSN Messenger! Download today - it\'s FREE!  http://messenger.msn.click-url.com/go/onm00200471ave/direct/01/  '),(23509,3739687,'13,August,2004','2004-08-13 00:00:00','   hello,  well tameron, the girl from south africa, was put on a different schedule  today.  st. joe had two people quit in one day.  tanya a science teacher  from australia quit because she got a job at an international school where  her three children can attend.  i think this is a good thing because she was  trying to home school them on top of teaching at st. joe.  also, ultan a  irish man with teaching experience is off to a different international  school.  he was offered 2 1/2 times the pay at st. joe so i can\'t say i  blame him.  now sister is on the hunt for more teachers but for the time  being i\'ll have all 30 kids.  honestly i didn\'t expect to have tameron for  too long, but i was hoping for a few weeks.  people come and go at these  school\'s all the time so a teacher has to be willing to be flexible.  other than that not much to tell.  i\'m going to soi 4 with friends from  school.  that is one benefit of st. joe eveyone is cool and we enjoy hanging  out with one another even when we don\'t have to.  soi 4 in silom is a gay  area with great music and good drinks.  i\'m looking forward to tonight.   it\'s fun seeing people from work let their hair down and party.  well gotta  run.  i\'m off to the gym.  jane  _________________________________________________________________ DonÃ‚Â’t just search. Find. Check out the new MSN Search!  http://search.msn.click-url.com/go/onm00200636ave/direct/01/  '),(23510,3739687,'12,August,2004','2004-08-12 00:00:00','   hello,  well i\'m just hanging out at my favorite internet cafe.  i have internet in  my room now, but i need to get a bunch of downloads and i don\'t feel like  waiting hours.  my connection is fast enough but downloading songs takes  forever.  anyway... i love my new computer.  i need to figure some things  out and get a digital camera so i can send ya\'ll pictures but other than  that it seems very userfriendly and reliable.  well knock on wood you never  know when these things are going to die on you.  well today is the queen\'s birthday and also mother\'s day.  thai\'s take the  royal family very seriously.  i\'ve even heard it is improper for them to say  their names outloud.  i\'m not sure if this is really true.  also, one of the  thai teachers looks a little like the queen so one of the farang\'s, ronald  from south africa, told her so and she thought it was the best compliment in  the world.  ronald is a very nice guy with a funny accent though.  he sounds  more european at times to me than south african.  anyway...  school is going o.k.  i\'m looking forward to the weekend.  being able to  sleep in today spoiled me so i hope i\'\'m not dragging tomorrow.  one good  thing is that tameron, the girl from south africa who is splitting classes  with me now, is going to take a couple of classes tomorrow.  i think she  needs to get a feel for what it is like to teach thirty girls in case she  ever has to run the class by herself.  i\'m sure she\'ll be able to handle it  she seems like a bright creative girl.  gotta run. jane  _________________________________________________________________ Check out Election 2004 for up-to-date election news, plus voter tools and  more! http://special.msn.com/msn/election2004.armx  '),(23511,3739687,'11,August,2004','2004-08-11 00:00:00','   hello,  well i was wrong about when monsoon season starts in thailand.  it\'s not the  middle of monsoon season it is the start of it now.  well as you can imagine  it is pouring down rain.  apparently it lasts till september and sometimes  into november so lots of rain is coming my way.  i don\'t really mind, but  sometimes when i absolutely have to leave my house it is a bitch.  also, i  lost my umbrella for a day so it caused me to be a hermit.  i just couldn\'t  deal with getting wet anymore.  anyway... tameron, the girl from south africa who is splitting my classes  with me, is working out very well so far.  she has tons of good advice and  is very active in helping me plan the classes.  i think she will actually  turn out to be a great teacher if she doesn\'t lose her drive.  sometimes st.  joe can take the drive out of you without you noticing.  there is just so  much paper work and lots of needless bitching it can take a toll if you let  it.  i try to stay positive, but at times i find myself complaining over the  smallest things.  anyway... the queen\'s b-day is tomorrow so no work.  yeah!  i\'m glad because  i need a day off.  i\'m still adjusting to this wake up at 7:40 business.   ugh!  just when i think i have the schedule down pat something happens to  mess with it.  i just have to roll with the punches.  fortunately, having  three older brothers has made me accustomed to this.  jk.  ya\'ll be good and i hope you are having a good summer.  jane  _________________________________________________________________ On the road to retirement? Check out MSN Life Events for advice on how to  get there! http://lifeevents.msn.com/category.aspx?cid=Retirement  '),(23512,3739687,'10,August,2004','2004-08-10 00:00:00','   hey everyone,  well i got some great news at work today.  i\'m splitting all of my classes  with a new girl from south africa.  her name is tameron and she seems pretty  helpful so far.  it\'s going to be nice to have fifteen instead of thirty.   especially since i\'m working on their pronunciation and the smaller the  group the better to ensure that all the girls are participating and  improving.  having thirty means that some of the more outgoing girls can  really dominate the class participation.  i had one girl flat out refuse to  talk today, but i\'m hoping with a smaller group some of the shyer ones will  feel more comfortable speaking.  yeah!  other than that i don\'t have much to tell.  we have thursday off for the  queen\'s birthday and also mother\'s day.  mother\'s day is always on the  queen\'s b-day so when i informed the girls that in america it is in may they  immediately asked me why.  why is it in may?  i told them because spring  symbolizes the birth, but honestly i have no idea.  i guess us crazy  americans just picked a month without much going on.  anyway... i don\'t  think they understood my answer which is for the best since i made it up.   well gotta run.  love ya.  jane  _________________________________________________________________ DonÃ‚Â’t just search. Find. Check out the new MSN Search!  http://search.msn.click-url.com/go/onm00200636ave/direct/01/  '),(23513,3739687,'09,August,2004','2004-08-09 00:00:00','   hey,  well things are going good here.  i\'m a little tired because i couldn\'t go  to sleep last night.  anyway... long day of classes today too.  also i  learned that no matter how prepared i think i am a class can still go to  shit.  i ended up just making grade 1 write words all day.  i think they  were bored.  actually i know they were bored because they asked to clean the  room. ha.  well gotta run i have people over.  emma, from england and aurora  from oregon are over to play cards.  love ya.  jane  _________________________________________________________________ DonÃ‚Â’t just search. Find. Check out the new MSN Search!  http://search.msn.click-url.com/go/onm00200636ave/direct/01/  '),(23514,3739687,'08,August,2004','2004-08-08 00:00:00','   hello,  well it\'s been a busy weekend.  i went to a going away party for ermine,  french volunteer, and had a great time with all the french folk.  a bunch of  germans were there as well so english was the dominate lauguage. yeah.  it\'s  nice knowing what is going on.  today we took her to dreamworld, a thai  amusement park and tonight she goes back to france.  i\'m a little sad  because i was just getting to know her.  the amusement park was lots of fun.  i\'d forgotten how much fun it can be to  ride rides and act like a kid.  i\'m also looking forward to telling my girls  about it tomorrow because i know it will get them talking.  they love to  hear about what the farangs do with their free time.  well the weekend is almost over for me.  work tomorrow.  ugh.  i\'ve been so  busy i think i need a weekend to recover from my weekend.  well gotta run.   love ya.  jane  _________________________________________________________________ Express yourself instantly with MSN Messenger! Download today - it\'s FREE!  http://messenger.msn.click-url.com/go/onm00200471ave/direct/01/  '),(23515,3739687,'18,August,2004','2004-08-18 00:00:00','   hello,  well it has been raining like crazy here.  i think i know what they mean by  monsoon season.  i guess it could be worse i could be in bangladesh or iraq.  jk.  no not really i love thailand and especially bangkok.  anyway... school  was a little difficult today.  i\'m starting new subjects with a couple of my  english composition classes and it is proving extremely difficult.  also,  while i haven\'t moved into my new homeroom yet i\'ve been spending more time  in there and trying to get to know the girls.  it\'s funny when i first  started teaching i couldn\'t tell any of them apart and i didn\'t think i  would ever be able.  now i notice all the little things that make them  unique in looks and attitude.  i\'m excited about actually being able to call  names in class and not just numbers or by pointing (actually gesturing  because pointing is rude however they don\'t take offence at the farangs).   while i can learn the names in that class i\'m still going to stick to the  numbers in the others to save time.  wish i could learn the all, but with  the number of students i have i don\'t think it is possible.  maybe next  semester i\'ll be able to work on this.  sewing class was the most challenging today.  i don\'t know how to sew and i  told sister erlinda this but she decided to give me the class anyway.  also,  she said the girls already knew how to sew i just needed to make sure they  were doing their work.  NOT TRUE.  a few of the girls can sew and cross  stitch, but not all.  so i was sending the ones with questions to the ones i  thought were doing the assignment correctly, ALSO WRONG THING TO DO.   finally the thai teacher for that homeroom, who incidentally i\'ll be paired  with as soon as i move into grade 3/2, came in and explained to the girls  how to complete the assignment.  thank god.  love ya, jane  _________________________________________________________________ Express yourself instantly with MSN Messenger! Download today - it\'s FREE!  http://messenger.msn.click-url.com/go/onm00200471ave/direct/01/  '),(23516,3739687,'17,August,2004','2004-08-17 00:00:00','   hello,  well sister erlinda has confirmed it again.  she\'s crazy.  now she\'s on back  on her kick about what teachers wear and how their hair is cut.  which is  one thing, but she\'s lost several teachers this month and i don\'t think that  alienating the other ones is a good way to keep the staff happy.  also, the  computer teacher might be looking at another job as well so that will mean  that six teachers have left since i started.  anyway... you would think that  she would leave me alone about my shoes.  other than that school is going great.  i\'m enjoying the lesson this week  and grade one is going to be absent friday.  yeah.  i love them, but they do  try my patience especially on fridays.  also, i\'ve figured out a way to get  everyone involved in class.  it\'s a little impersonal but since i have so  many students it is the only workable way.  they all have assigned numbers  which they have to put on every scrap of paper in every class, so i simply  call out a number and that girl has to answer the question, read, or  whatever other task it is that they are doing.  i like it better because it  keeps the extroverts from controlling the class time.  the only other thing of interest i can think of is that they are moving me  into a new homeroom.  which is good and bad.  bad since it is in grade 3/2,  the ones i\'m always bitching about, but good because they are finally going  to get me a cabinet which i need desperately.  anyway... i\'m hoping that by  being their homeroom teacher they will like me more and maybe pay closer  attention in class.  o.k. i\'m trying to be positive, but really i need you  to pray for me.  love ya, jane  _________________________________________________________________ Check out Election 2004 for up-to-date election news, plus voter tools and  more! http://special.msn.com/msn/election2004.armx  '),(23517,3739687,'16,August,2004','2004-08-16 00:00:00','   hello,  well i\'m doing great.  today\'s lesson at school went very well. i played 20  questions with them and worked on how to properly ask questions.  it went  great and the girls loved it.  it feels great when i actually get to teach  them something and they enjoy it.  other than that i don\'t have much to say.  i also had a great dinner with nancy and cammie at a local thai restaurant.  also, my cold is much better.  yeah.  well gotta run.  i\'ll write more later.  love ya.  jane  _________________________________________________________________ Express yourself instantly with MSN Messenger! Download today - it\'s FREE!  http://messenger.msn.click-url.com/go/onm00200471ave/direct/01/  '),(23518,3739687,'15,August,2004','2004-08-15 00:00:00','   hello,  well i still have a runny nose and i\'m coughing up lovely green colored  phlegm.  i\'m sure all of you wanted to hear that.  honestly i don\'t feel bad  so i think it is some sort of allergy thing and not a cold.  i was a good  girl and stayed in last night.  i slept for around 15 hours.  i kept waking  up and putting in a new movie thinking that i wouldn\'t be able to go back to  sleep but then i would pass back out before i was ten minutes ito the movie.  anyway... i guess i don\'t really have much to say.  if i still feel like it  i\'m going to the weekend market to pick up some things i desperately need.   things out there are cheap and they usually have sizes that fit so i\'m going  to go on a huge shopping spree.  i need so many things such as bowls,  plates, cups, lamps, shirts, shoes, etc.   just about anything i buy would  be a huge addition.  well gotta run.  jane  _________________________________________________________________ Get ready for school! Find articles, homework help and more in the Back to  School Guide! http://special.msn.com/network/04backtoschool.armx  '),(23519,4256898,'22,August,2004','2004-08-22 00:00:00','   Todae, Sundae nite.... again mus think of tml... the dae to face the weekdaes again and to face all of those hw and test tt will be given by the teachers... But anyway, todae was quite a good dae compared to those of the others... Todae, i am able to do my hw without my dad\'s presence and tt i m able to to it peacefully... Though i seem like i always haf a bad tittle in my entries, most of the time when i post it, i m still in quite a good mood, but jus a bit to restless as nth much exciting happened... =) Think tt mayb things if seen more simple, things would oso be more simple... Hope tml will be a gd dae.... =)   '),(23520,4256898,'21,August,2004','2004-08-21 00:00:00','   Todae, a saturday... Finally, all the weekdaes are over... Even though I still return to sch early in the morning for a remedial, I still managed to pay my frenz house a visit to pass him his hw... Watch my god bro perform during his NPCC promotion test... it kindda seem funny to me, but his expression when taking the test was so strict... so firm.... Nth much though, even though it is a saturdae... nv go out, nv go and play bball... whole person felt so weak and tired... hope to jus lie in bed and slack for the whole dae...   '),(23521,4256898,'20,August,2004','2004-08-20 00:00:00','   Today, it is another quite boring dae for mi as i do not find any speciial meaning for mi to be happy when i am carrying on wif the same old life every dae... Everything went on nicely, and first time this yr when i reached home at 1.25pm... haha~ after watching tv, ate lunch, and play the computer.... i went to slp slp slp.... even though there were a few frenz whu asked mi to go out.... SiaNzz... slept for so long... wake up le still veri tired... mayb cos everydae waking up at 5.30am in the morning.... Haiz... now thinking of the A-Maths remedial tml..... Boring, Sianzzz, Tired is all those few words tt i can use to describe wad tml will be like... Think this is all liao lo... =) Hope tt everyone will haf good luck everydae.... :D   '),(23522,4256898,'19,August,2004','2004-08-19 00:00:00','   Today, it is the day when my effort that i had put into the Prefects\' Investiture finally paid off... Even though things went on faster than we had planned... nth went off too much... Today, prefects from other schools like Mayflower Secondary and Ahmad Ibrahim Seondary and more came to gave their support which made me felt very relief even though we had informed them quite during the last minute... I even got to see my previous Prefect Master whom I had adore and idolised since Secondary 1 when I first got into Prefectorial Board. However, he left to become a counsellor after teaching my batch for 1 year... which had made us felt quite sad.... Even though times spent toking and communicating wif him had become veri veri less... but then our relationship is still as close as though he is still our Prefect Master... As my position now as a 2nd Deputy Head Prefect is being comfirmed, it will mean that I will face more work and duties and that i will wanna learn how to cope wif it... Even though there may haf been conflict within all of us... I do not think that I will wanna solve it as the other party had not seem to put in the effort, and that others had hated her/him.... She/He may be in the good books of the teachers... but then she/he will noe that all the tough work had been done by her/his colleagues whereas she/he had been the one whu had been gaining the credit... Truth will one day be out, and the teachers will see wad kind of a person she/he is and that she will one day get her/his punishment for doing all those things that had hurt the feelings of the others... =) ****Happy 15th Birthdae to Bertina Cheong****   '),(23523,4256898,'18,August,2004','2004-08-18 00:00:00','   Today is quite fun.... Although a lot of things cropped up during the Investiture Rehearsal.... things did not went that well and that we had to find a lot of things during the last minute becos of the organisation which still needs to be brushed up... Went home so so late and really had a lot of things to do.... but then still tot of entering another entry here.... I will forever remember... * Â¶Ã°Ã¼gh Â¶Ã¯MÃ«Â§ dÃ¶Ã±â€˜â€  Â£Ã¤Â§Â¶; Â¶Ã°Ã¼gh mÃ¯Ã±dÂ§ dÃ° -- Â£Ã°Ã¼Ã¯Â§* This sentence... by Â£Ã°Ã¼Ã¯Â§ (OBS Best Instructor) to me, make me have the synergy to work... to do my best... I will nv give up and will nv bow to problems unless i am left wif no other choices.... =) Good N!t3...   '),(23524,4256898,'17,August,2004','2004-08-17 00:00:00','   Today, the first time I post things into a blog account. Feeling quite weird as this will show what I felt for the whole day. Today, quite a bad day for me though... Woke up late and rushed to school in my dad\'s car... Although not late for both duties and school... the whole day is still a veri boring day. Nothing went really good and that things tomorrow will still carry on as a big big mystery and i only can say that forever in school... Forever theres\' tests... :D Hope that tml will be a better day for all the people in this world....'),(23525,3751728,'28,June,2004','2004-06-28 00:00:00','   To Billy  He looks her in the eyes, he never told the truth, She thought that the love would surely be there too. \"I love you\" came often, they\'d been through a lot, He always thought that things got hot. He had her heart, it bled in his hand, He\'s just a boy, who thinks he\'s a man. Not that simple, not that right, This is the last moment they\'d spend tonight. A kiss goodbye, she\'d melt like butter, He\'s a prince of destruction, not just a lover. The next day, he went to the same spot, Another girl was there, and her, he forgot. He hugged her tightly, he couldn\'t let go, Who thought that I would ever stoop that low. The next day there were three girls in his room, They never did think, they\'d always asume. How does this boy sleep at night, Why do wrong, when he could have had right. Little thoughts would wander to be, He loved those girls, he just liked me. People would tell her, what he would do, But she couldn\'t believe, that these words were true. He\'s way too hard to figure out, But he\'ll never stop without a doubt.  '),(23526,3751728,'28,June,2004','2004-06-28 00:00:00','   Life isn\'t that simple, it\'s not easy as pie, People\'s ignorance asserts you, they always try to lie. Many people grow up like a druggy with a kid, They cover up all of thier mistakes for the ones they\'ve already hid. Life is a destructional balance that slides from side to side again, The key to life is loniless, the lock is never having a friend. Jealousy takes over, nobody cares for you, People try to screw you over, they try to play it cool. Suicidal attempts that didn\'t really succeed, Slitting your wrists open, hoping for it to bleed. Mom dies, and then dad abandons the only kid he had, But life screws us all, and he\'s really just a big fag. Life is a blemish, to never disappear, It\'s just like a person you\'d give anything just not to hear. Too bad your life will never be perfect, the way you\'d like it to be, So just go around with your middle finger in the air, for everyone to see. No, I can\'t lie my life is obviously hard, Not to be happy, screaming, cursing with the drugs that have made me scared. Why do people act like thier life is a fairy tale dream, Sugar coated with a cherry, a big sloppy lie made out of ice-cream. Precious diamonds means nothing but useless time, Not really as luctuous when you\'ve grown up without a dime. Problems occur when you have nowhere to run, So you hide in your shadows and this becomes your favorite fun. Silence takes over, choking your very own heart, Until you give up and say, this is one life of which I don\'t want to be a part. '),(23527,2116991,'17,October,2003','2003-10-17 00:00:00','    Hi there and welcome to my first posting, thanks for takeing the time to look in and i hope youll keep coming back and let me no how im doing.   Well i think i should start off by putting down what im hopeing to achieve here , why im keeping this blogg. whats it all about etc......  Well I have been fortunate enought ot recently find something that has been missing in my life for a very very long time, and something that is not at first easy to explaine. Im not like the majority of people out there and for a long period in my life ive known i dont fit in . normal life just dosent feel right for me i have a much deeper and more primal side that for far to long i needed to both become aware of and start to satisfy. so what am i yammering on about ??  well im a slave , a daddys boy and thats somethign which isnt at first easy to explain but is a huge part of who I am. I dont feel complete or a whole person without haveing something vitally important to my state of beeing ( God this is getting heaavy alreddy) I need to be owned controled and taken over completly by another man, and for those of you reading this who arnt in the bdsm comunity i no that this is goign to be a concept thats totally alien for you , and for that im very sorry. i hope youll bear with me as i fumble my way through trying to clarify all this .  There was until very recently a huge gaping hole inside me a need for something that for many years i didnt understand, im fortunate enough now to have become aware and come out to myself, which is wounderfull its given me a chance to be the person i need to be. Ive been fortunate enough to meet an incredable man, my Dad and Master, Bearfoot. Ive submited fully to his authority and am owned out right by him. what does that meen well basically I am his boy and slave, im here to satisfy his needs in every way and i couldent be more happy about it . for the first time in my life i feel totally complete and at peace. I have a long and very hard journey in front of me , im going to stumble a lot, I have alreddy and i no i will contimnue to do so. but im on a wounderfull journmey guided by my Dad and Master.   It involves giving over full controll of my life to him, all deescisions, financial controll, how i dress look act my day to day duties and responsabilitys, everything. And being owned totally. it involves Satisfying my Dads every need , be that a sexual, sadistic, or mundane house hold request . i Do it with out question or thought . total obedience. I no by now a lot of you will be eather horified or wont be able to get your heads round this, but belive me its a wounderfull and beutifull thing to give yourself totaly and completly to another man. to pass on all controll. to me it has been a case of giving up so litttle and gaining so much in return.  on wedensday 8th My Dad snaped his collour round my neck and i was finally and totally owned. and i cant tell you how proud I feel to ware this collor. well im begining to waffle here i have so much more to say but i think i need to calm my head a little before posting so illleave you with this . Im totaly new to this world but feel totally at peace and comfortable in it. I find it over whelming emotional scarey and for the first time in my life i no im exactly where i need to be. its going to be hard, its going to be very dificulyt at times its going to scare the hell out of me and im going to be pushed to places i never thought i could go , but its also going to be wounderfull and i cant wait . so keep poping in and ill keep you updated with all my hopes fears concerns happy and sad moments the mistakes i make, and the hurdles i climb for my Dad. im at the start of a wounderfull road here and i hope youll come along with me   '),(23528,2116991,'05,November,2003','2003-11-05 00:00:00','    Hi every body , once again I apologize for the lack of postings, my Dad came down to visit on Saturday and is staying here for a couple of weeks :)  you can imagine how chuffed I feel. Well things are going very well. Its almost a month since I earned my collar and so far so good. Im still blissfully happy .                 Well what have I been up to :) after I came back to Wales from Manchester I\'ve been pining like there\'s no tomorrow , but now Dad\'s here and im walking 6 feet above the ground. We\'ve had some magical days since Dad came down, the first  was on Sunday. At around 3 am we went for a walk up the hill behind the village , in the rain and with the thunder rattling around, we could see lightning on the horizon, and across the loughour estuary to llanelli all  light up.  Its quite a magical view. At he top we talked for a while then Dad ordered me to strip.  I did straight away and although ti was cold and wet and I was shivering, it was wonderful. I was standing there naked with my Dad on top of the hill watching the mist rolling in off the sea and feeling the rain power down over me. What made it even more magical was Dad telling me how proud of me he was.  It was a very special evening , very profound on a lot of levels and something I don\'t think ,me or Dad are going to forget for a while   Monday we went out shopping for boots and me and Dad managed to find some excellent ex army boots for both of us.  Which we promptly went and broke in charging round the local lazer dome, good fun was had by all , but im still not sure where shooting your Dad fits in to the whole Master slave etiquette.   and today has also been really special  me and Dad ordered a  urlLink cb2000   . Dad wants to keep me in chastity , except for special occasions  and it arrived this morning :)  so you can imagine how long work today while I was waiting to get home :)  . Well when I eventually got back here after dinner Dad fitted me with it and I have to say it feels great and im bloody proud to ware it .  so in synopsis how are things going. Fantastic, absolutley no regrets, every day that passes just makes me more and more glad that Dad gave me his collar. im finally exactly where I need to be and feel totally happy. Im haveing a wonderful time with my Dad and serving him and being his boy just gets better and better.  well will close for now, but im going to do my best to try and post atleast once a week if not more  so please keep coming back and see how im doing :)   '),(23529,2116991,'01,November,2003','2003-11-01 00:00:00','       '),(23530,2116991,'01,November,2003','2003-11-01 00:00:00','    My apologies for not updating this a lot sooner, the past couple of weeks have been a bit of a whirl, I\'ve been missing my Dad like crazy as he\'s  been in Manchester and ive been in Swansea . im working on finding a way to be closer to him., so keep your fingers crossed for me .  On the up side my Dad is traveling down to Swansea as I type and I cant wait to see him, ive missed him like crazy, I have lots more to add but cant at the moment As im  stuck in work, but ill pop back on when I get home a little later and fill you in on all the latest news and how ive been doing   '),(23531,2116991,'29,June,2004','2004-06-29 00:00:00','    ahhh another day back at work yuk !!  been quite productive though ive chatted to my boss about going part timeing and hes  going to set that up for me  and ive emailed of to college to see if i can geet some year one notes to revise with  finances are still scareing me a bit but well we wil makje do and im sure well be ok .  Dad looks a little happier today adn by all acounts he sounds liek hes eeling abit better which is good to see . i hate it when hes down and i cant help.  me im stil tired but doing not to bad. early night tonight i think he eh no sitting up playing uru till 4 am LOL  mind u i finished uru adn now im on to uru to dunni. il have to buy the last expanshion pack soon :)#  just chatting to jocjk hes in good spitrits which is good to see. hes one hell of a sexy bloke and i hope me and DAd get a chance to have a gopod old fumble soon. actually im hopeing that well find some older stocky hairy bloikes to play with localy . be nice to build up a cricle of fuck buddies and sm play mates   '),(23532,2116991,'28,June,2004','2004-06-28 00:00:00','    been a good day today :)  ive found out that im geting my student loan  so ging back to uni is on :) just as well really conmsidering all teh whoo ha thats going on at work.  so tomorow im off in to chat to my boss and sort out switching to part time  eeeeep.  had a good day today went for a walk round the reserviours with DAD hes been down and very quiet all day i think it helped a bit we had a good old natter abiout a few things. inparticular about  experiences and becoming jaded with them  i think what brought it on was walkign through all teh beutiful contry side me being moved by it all and DAD looking at trees and going oh look big vegtables. we talked alot about thibgs weve always wanted to do whale watching in canada and treking over the andies that type of thing. and DAD thinks that hes so jaded now that none of that would have any meening for him. I disagree i dont think its a case of being jaded , i think its a case of haveuing been in a bad place for a very long time. My DAD bear has had soem severe depression issues to cope with over the past few years. i think that together slowly bit by bit working together that hell get his enjoyment of the world back again  one small step at a time. although i think my DAD bear needs convincing  :) im one cub who dosent give up that easily though.  been an eventfull weekend. i was quite ill over the first part over exausterd adn my diabeties was playing up . id been burning the candles at bioth ends abit . i had big dark cyrcles under my eyes was very dizzy and had an acheing head. so i took two days of work and slept for most of them . we had our mate greggy over and me being a bit under the wether put a bit of a dampner on the sex side of things but we still had a bit of a fumble. i think greggy had a good time.  Me and DAD had a bit of fun though  weve made our own porn moovie :) quite pleased with the results too weve got video eddeting software on the pc and have done quite a good jiob wihtit i think  at the moment im still feeling a little rough and a bit stressed and frustrated not in any big way , im just frustraded really cause my DAD bear is down at the moment and theres not a lot i can do to help des[pite my best efforts. lots going on too witht he change at work, switching to part time , sorting out the new flat  the hospitals money and everythign else thats going on buit im holding it together  so far so good and all that   right time to go make a cuppa i think by for now   '),(23533,2116991,'25,June,2004','2004-06-25 00:00:00','    Been a couple of weeks now since ive had a chance to post on here . sorry i havent ment to neglect you all  my DAD bear is down in swansea with me for a couple of weeks which is great. a lot has happend sionce my last posting so il try and cram it all in.  first off i had my band inflated for the first time which was a very strange experience   nto been inflated enbough thouhg i think becasue theres not a lot of erestriction there and i seem to be able to eat pretty mucvh what i could before so im gessing another fill is in order soon.  work has been a bit of a trauma. teh technical suport department i workl for has been sold fof to anothe company ibm no less and in spetember we wil all be transfering across to them so all sorts of emotional ups and downs there and were all feeling abit knocked for six.  ive been very run down over the pastr couple of weweks too been burning the candle at both ends a bit  im of work today trying to get myself back on my feet a bit. ill pop on tomorow and wright a better post when my heads in a better place   '),(23534,2116991,'15,June,2004','2004-06-15 00:00:00','    Sorry everybody ,                    Ive been a bit down over the past few days and haven\'t  posted. Back on my feet now though and raring to go. Im one of these unfortunate people prone to sudden and fairly heavy attacks of depression  im climbing back up out of this one though .      I had some great news my DAD bear is coming down on Friday :) YAAAY  just in time for midsummer . Going to do something special for that .      so what have I been up to well aside from getting miserable ,ive been working on a couple of things,  first off im hoping to walk the stmichal ley line next summer. Something I\'ve always wanted to do its about 600 miles from st michals mount in Cornwall to great Yarmouth in norfolk. What im really hoping to do is turn it into a chatrity event for  urlLink the wls site  I\'ve got a fe w ideas on that poss setting up a fund so people can apply for a grant to see there consultant for there first apointment privately .could cut a year or a couple of years of the weight people have. be nice to tey and help people who are on low incomes and cant aford to do the inital asesment privately.  secondly ive started work on  urlLink galaru  again  working on the opening animations for teh game. im rendering a 30 second fly by and its takeing 36 hours to render. tell you what though thisis going to be an awsome game.  and to top it off ive started putting pen to paper on a book that ive wanted to wright  mcaled green wich its a look at my own personal view of paganism and wicca. more on thay in future posts :)  so , as you can see im kind of working myself out of it LOL   any ways have laods fo stuff to do. i want to fifnish the first chapter of greenwich tonight, plus draught my letter to the wls group to sugest teh charity fund and the fund raising idea . that and i need to have a tidy up round here for DAD coming down   LOL bizzy cub   '),(23535,2116991,'11,June,2004','2004-06-11 00:00:00','    rice cakes  now theres alot to be said for the humble rice cake,  as im just begingin to discover they are not infact small round costers with the look and consistancy of polystyrene.  since my banding ive found it quite hard to eat bread as it all sort of recombobulates in my tum and bungs the band up  and i spend the next 20 mins doing my frog impression hic-ribbit hic-ribbit  any hoo after my flat mate spoted them going cheep i delved into the dark art of the rice cake. ok at first there a little bland but with a litle marge on and WOW boy do the taske good,  with cream chese roasted red pepper and baby spinage on top there pretty spectacular and for a purley pornographic food moment i recomend asparagus cup a soop with butterd rice cakes to dunk  ive found a whole new favourit snack :)  any way nibbling aside  had a great today i woke up full of energy and have nbeen zooming around teh house like a thing posesd. ive lost about a stone and boy do i feel it  so far ive finished the washing washed bessy waterd the garden been for a 5 mile walk and im stil readdy fro more   he he super cub !!!!!!  i think its because as i loose weight im gonna have more muscle bilt up that  ineed or to put it another way us fat bastards need a lot more muscle to cart teh weight around so when the weight goes the muscle is still there :)  im going to have to try and keep it so im gonmna do my best to exercise more nothing boring  just walking cycling swiming that sort of thing.  It Did set me thinking though ive always wanted to walk the ickneild way and i used to love long distance walking.  its about 400 miles from st michals mount in cornwall to yarmouth in east anglia. im hatching the iodea of training my self up for it and nextr summer gettign sponsord adn fund rasing for the wls charity plus rasining awareness whil im doing it.  have to clear it with DAD first though  he he today has so far been a very good day :)   '),(23536,2116991,'09,June,2004','2004-06-09 00:00:00','    WOW thats weird  seing yourself on telly is very weird LOL  i had no idea i spoke that fast    was a good prograsm though and seemed to get the message across which was the main thing :)  Im really glad i bumped into mick from the weight loss surgery group im looking forward to getting to no him and getting involved witht he group . i just have to get along to a meeting  it would be great now toget invovled and give somethign back to the wls comunity now that ive had my op :) id love to be invovled in fund raising helping out and generaly doing what i can to raise awareness  im that kind of bear really LOL   '),(23537,2116991,'08,June,2004','2004-06-08 00:00:00','    Ive done it again LOL  gon and got me mug posted all over a bbc web site   have alook at the  urlLink week in week out website  for the episode called one year on :)     any wasy todays been mixed . i didnt get my band filled but they have now made an apointmwent for me so thast somethign  hehe ive lost 5 kelos YAYYYY:)  back to work tomorow though which is not so good harumph   '),(23538,2116991,'07,June,2004','2004-06-07 00:00:00','    Today started as a bad day , id been hit with some bank charges iwasent expecting adding to this months money hassles and to top it off I phoned up about the telephone interview and they had finished the study  grrrrrrrrrrrrrrrrrrrrrr  but things improved abit after  first of , started getting organised with the  urlLink bear boat  I fired of the first round of emails  its still not to late to sign up if you fancy joining us just click on the link for more info .  Ive had a good old chat with DAD and we\'ve decided to make a few changes. With the flat move coming upin September were going to start again big style that means new clothes new flat new everything kind of starting my new life properly :). Plus my DAD bear has given me an awesome  urlLink bible page    a what I here you cry . Well have a look its apage with the rules on there my DAD wants me to live by plus instructions I need to complete for him and a hope lot more  what can I say Im dead proud to have that page and will be checking in every morning to make sure that im uptodate with the rules and that I no the state of play he he im one very chuffed bear cub   '),(23539,2116991,'06,June,2004','2004-06-06 00:00:00','    Morning,          Well I say morning its just gone 3:30 in the afternoon and im a bit annoyed with meself.  I had all these great ideas about getting up this morning doing a shed load of washing then cycling off to the beach for a spot of sunbathing and skinny dipping .   didn\'t wake up till 2 pm   BUGGER  oh well not to worry  got the washing on the go and ill make an effort to get out there and give bessy a good clean too.  You no I don\'t no what\'s up with me at the moment I seem to be a bit wishy washy no enthusiasm and cant pluck up the energy to do anything. I don\'t think im depressed I just need my DAD to give me a kick up the arse I think LOL.  so to make up for it im bizzying myself round the house tiding up the kitchen  washing up etc etc  boring I no but needs doing . One good thing though its given me a chance to catch up with my online stuff and I\'ve just been over to the  urlLink wls  site the what I here you cry ?? Well I am what is nown in medical and technical terms as a \" FAT BASTARD \"  something which im actually immensely proud off , but I had all sorts of health stuff going wrong and wound up getting my stomach banded on April 2nd.  Its a bit like a stomach staple but a little more down market. Any ways I digress       after trawling through the millions of posts I\'ve missed over the past couple of months I found one from a medical research company who want to give me 30 squid for a telephone interview  RESULT !!!!. He he so il call them on Monday .       Actually as im sure DAD will tell you im a bit of a media slut about all this ill have been on telly twice and radio once chatting about the op  infact if any of you are in the Wales telly region or can get welsh beeb 1 tune in on Tuesday for week in week out and im on there :) they\'ve even got film of my operation .  EUGH!  oh well not to worry time to go hang out another load of washing  a slaves life of drugery is never done :0-)   '),(23540,2116991,'05,June,2004','2004-06-05 00:00:00','    urlLink    me posing on a wall :)&nbsp; urlLink   '),(23541,2116991,'04,June,2004','2004-06-04 00:00:00','    OK what do you rekon to the new look then :)  i got a bit bored with all that orange LOL    Ive added some links too theres one to DAD\'s blog  well worth a read one ot the british bears site now if your a bear or an sm person youll definatley enjoy that one esp DAD\'s wrightings on there . and one for the bear boat . now that gonna be fun  a bunch of bears of narrowboating for a week we stil have places avalable so why not have a look on there .   now here goes nothing im gona have a go at posting a pic up on here  keep your fingers crossed   '),(23542,2116991,'04,June,2004','2004-06-04 00:00:00','    Wow well what can i say other than sorry its been a gap of what nearly 6 months since i posted on here .  A lot has gon on in the last 6 months . a lot of good stuff and also a lot of hard stuff.  on the slave and sm side of things  weve had to go a lot slower than eather of us hoped. lots of reasons why , i had an extended stay in hospital to get my gastric band put in, moving between manchester ans swansea . lots of hassils with work  to name just a few. I wont bore you with all the details but things are looking up .                In september im moving to my own place which is going to change everything . at the moment im shareing a house with two other non sm blokes who are a bit prudish so i cant be nacked at home cant follow my DAD\'s house rules ect ect. so in september we start for real . me and DAD can begin to live the way we want to 24/7 difrent from the odd bit here and there me and DAD can snatch. which is going to be great.  We have bessy now our car which meens were going to be nmobile and getting oput visiting other sm couples going to events etc etc is no longer a problem. and im of back to uni full time adn working part time :) so much less pressure from my crap job . so all in all things are abouytt o gete much better  now that we have a plan of action and some serious goals we can work towards. were finally going to be in a postion where i can be my DAD\'s 24 /7 slave wether im in manchester or swansea or whereever we are .   and i tell u what   i cant wait.    Im still as keen as ever to be a slave for my DAD bear and i no hes going to train me and mol;d me in to the best damn person i can be.  its going to be dificult  and scarey at times a lot of hard work but very very worth it   i no my DAD is going to take me firmly but lovingly and push me ,in all sorts of ways allways extending my limits a bit at a time till im a real no limits slave for him  im a very lucky lad indeed   '),(23543,3820930,'01,August,2004','2004-08-01 00:00:00','    My laptop isn\'t feeling too well.&nbsp; I spilled hot wax all over the keyboard and the LCD the other day.&nbsp; I did a pretty good job of getting it off the screen, but the keyboard is a mess.&nbsp; So I went on eBay and ordered myself a replacement keyboard for 30 bucks.&nbsp; I was going to go through Dell, but I figured why pay like well over 100 bucks for something I can do in minutes.&nbsp; When I get the keyboard in, I will take pics of me replacing it.&nbsp; This is the first time I am actually opening up my laptop to work on it.&nbsp; Yipes!&nbsp; I have always been more comfortable with desktop pc\'s, but I think it\'s about time I get more involved with my pride and joy.&nbsp; That\'s it for now.&nbsp; I need to go shower and pack.&nbsp; I am staying over  Julie\'s  house tonight because the Linkin Park concert is tomorrow.&nbsp; Yay!   '),(23544,3820930,'01,August,2004','2004-08-01 00:00:00','    So I spent 70 bucks on albums when I was on iTunes.&nbsp; I have been turned on to some great music as of late.&nbsp; It all started with Postal Service.&nbsp; Postal Service\'s CD \"Give Up\" is a must own for anyone.&nbsp; I mean it is one of the few CDs where I love every single song.&nbsp; Truly amazing shit.&nbsp; Postal Service was formed when the lead singer of another great band Death Cab for Cutie got together with Jimmy Tamborello of Dntel.&nbsp; The called the band Postal Service because they recorded everything through the mail.&nbsp; Everyone should buy this album now!!!&nbsp; Another great Indie band is Architechture in Helsinki.&nbsp; They mix brass with great vocals and percussion.&nbsp; Very mellow and chill.&nbsp; Finally there is Belle & Sebastian.&nbsp; Same mellow kind of tones.&nbsp; Leave comments of bands you think deserve a download and I\'ll check them out!&nbsp; :-)   '),(23545,3820930,'01,August,2004','2004-08-01 00:00:00','    Did you ever sit around and wonder about something?&nbsp; Like the universe.&nbsp; I get all freaked out when I think about it.&nbsp; I mean what  is  the universe?&nbsp; Are we inside something else?&nbsp; Where does it end?&nbsp; Where does it begin?&nbsp; I seriously fuck myself up thinking about it. &nbsp;&nbsp; Try it.&nbsp; See.&nbsp; Told ya so.&nbsp; Anyway,&nbsp;when I take my vacation next week, I plan on making some observations about some things.&nbsp; Like how I live my life.&nbsp; What I\'m doing that is&nbsp;right.&nbsp; What I am doing that is wrong.&nbsp; I want to come away with something that will be helpful.&nbsp; I need to change the way I am doing things around here.&nbsp; This is the first time in like almost 2 years I am getting an entire week off.&nbsp; It should prove to be interesting....&nbsp;  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;   '),(23546,3820930,'01,August,2004','2004-08-01 00:00:00','    Yea.&nbsp; I said I was going to be typing a lot.&nbsp; I inadvertently lied.&nbsp; Work has been so hectic and crazy these past two days this is the first moment I have gotten to sit and actually gather my thoughts.&nbsp; I am sick still.&nbsp; I am such an idiot.&nbsp; I stop taking the antibiotics once I feel better.&nbsp; I need to finish out the entire course of them.&nbsp; I will begin that tomorrow.&nbsp; I have a mystery rash on my chest.&nbsp; I don\'t know where it came from, but it\'s weird.&nbsp; I thought it might be and allergic reaction to my meds, but alas, it\'s still there.&nbsp; If that doesn\'t clear up, off to the doctor I go for more prodding.&nbsp; I swear.&nbsp; I need to take better care of myself.&nbsp; So how have all of you been?&nbsp; Good I hope.&nbsp; I have been downloading shitloads of music off of iTunes.&nbsp; It is really amazing and for any of you who don\'t own an iPod, go get one.&nbsp; Now.&nbsp; I will give you $20 towards one if you take that money and actually go and get it.&nbsp; I am serious.&nbsp; I think everyone on this planet should have one.&nbsp; I have over 900 songs on mine and still have ten gigs left.&nbsp; That\'s a lot!&nbsp; Well I am off to go download some more music, so leave your comments and suggestions as always.&nbsp; Which is never lately so get posting!&nbsp; I will type more later when I am not so bajang.&nbsp;   '),(23547,3820930,'01,August,2004','2004-08-01 00:00:00','    I miss him.&nbsp; I was with him today when he was put to sleep.&nbsp; Poor fool.&nbsp; He will be missed.....   '),(23548,3820930,'01,August,2004','2004-08-01 00:00:00','    Can we start over? Meet each other again? I once was close to you. Told you everything in my head. Unwanted. Scared. Nervous. Sad. Hated. Unworthy. You just don\'t understand do you!?!?! Take a fucking step back! See what this is doing to me! Stop being so right And help me fix what is wrong. You\'re my friend. Or so I thought you were. May be you were right. It\'s time to grow up. But if growing up means being like you, Let me be young forever!   '),(23549,3820930,'01,August,2004','2004-08-01 00:00:00','    I plan to be writing a lot these next few weeks.&nbsp; I have decided to get it all out at the expense of you guys.&nbsp; Sit back and enjoy!!   '),(23550,3820930,'01,August,2004','2004-08-01 00:00:00','    My visit went well today.&nbsp; I had a pretty good rapport going with the RM and ZVP.&nbsp;  Joe  was being a jerk.&nbsp; I know what I have to fix.&nbsp; Just let me do it.&nbsp; I took a three hour nap before.&nbsp;&nbsp;It felt good to just relax.&nbsp; Now on to my topic today.&nbsp; People.&nbsp; It\'s funny&nbsp;how when someone gets called out on their shit, someone who is exposed for their true intentions, acts so differently like nothing ever happened.&nbsp; You know what I mean.&nbsp; It\'s like night and day.&nbsp;&nbsp;First they live in their fantasy world.&nbsp; Then it\'s back to fucking reality like their little&nbsp;\"fantasy\" never existed.&nbsp; You know it\'s there.&nbsp; It\'s like an old saying my therapist once told me.&nbsp; There\'s an elephant in the room and nobody wants to talk about it.&nbsp; I learned a lesson&nbsp;yet again from all of this.&nbsp; Don\'t give yourself to somebody when&nbsp; you don\'t even know&nbsp;what they are going to do with&nbsp;you.&nbsp; I miss  Danielle . &nbsp;&nbsp; She is in the Poconos right now.&nbsp; Lucky her.&nbsp; My vacation is&nbsp;the first week in August.&nbsp;&nbsp;I have no idea what I am going to do.&nbsp; I know&nbsp;one thing.&nbsp; I am not going into work at all!&nbsp; I am going to wrap this one up for tonight.&nbsp; I think&nbsp;I am going to play some&nbsp;video games to get my mind off shit.&nbsp; Comment or something!&nbsp;&nbsp;   '),(23551,3820930,'01,August,2004','2004-08-01 00:00:00','    The big day is today...&nbsp; The day when the Zone VP comes to my store.&nbsp; I plan on getting in my store by 7 am to do my final touches.&nbsp; I am only still awake because I took a four hour nap when I got home from work.&nbsp; Wish me luck and I will post from work when they leave to tell you how things went.&nbsp; I am their first stop of the day.&nbsp; Yipes!   '),(23552,3820930,'01,August,2004','2004-08-01 00:00:00','    Thereâ€™s a letter on the desktop That I dug out of a drawer The last truce we ever came to In our adolescent war And I start to feel the fever From the warm air through the screen You come regular like seasons Shadowing my dreams  And the mississippiâ€™s mighty But it starts in minnesota At a place that you could walk across With five steps down And I guess thatâ€™s how you started Like a pinprick to my heart But at this point you rush right through me And I start to drown  And thereâ€™s not enough room In this world for my pain Signals cross and love gets lost And time passed makes it plain Of all my demon spirits I need you the most Iâ€™m in love with your ghost Iâ€™m in love with your ghost  Dark and dangerous like a secret That gets whispered in a hush (donâ€™t tell a soul) When I wake the things I dreamt about you Last night make me blush (donâ€™t tell a soul) And you kiss me like a lover Then you sting me like a viper I go follow to the river Play your memory like a piper  And I feel it like a sickness How this love is killing me Iâ€™d walk into the fingers Of your fire willingly And dance the edge of sanity Iâ€™ve never been this close Iâ€™m in love with your ghost  Unknowing captor You never know how much you Pierce my spirit But I canâ€™t touch you Can you hear it A cry to be free Oh Iâ€™m forever under lock and key As you pass through me  Now I see your face before me I would launch a thousand ships To bring your heart back to my island As the sand beneath me slips As I burn up in your presence And I know now how it feels To be weakened like achilles With you always at my heels  This bitter pill I swallow Is the silence that I keep It poisons me I canâ€™t swim free The river is too deep Though Iâ€™m baptized by your touch I am no worse than most In love with your ghost  You are shadowing my dreams (in love with your ghost)   '),(23553,3820930,'01,August,2004','2004-08-01 00:00:00','    I have all three.  I am NOT a happy camper.  And on top of it I have three visible cavities.  I am so blah.  Things are better on everything else.  I saw  Danielle  this weekend.  It was good, but I didn\'t feel good.  We just chilled all weekend.  Work is so hectic.   Joe  and his boss are coming on Thursday or Friday.  Of all weeks for me to get sick.  The store really isn\'t in bad shape.  I just need to fix some parts of my NR wall and deal with PV overstock.  Sorry this entry is so short, but I am sickly and am going to play a game of Literati with  Danielle  and chill for the rest of the night.  If I feel compelled, I might write something else.   '),(23554,3820930,'01,August,2004','2004-08-01 00:00:00','    Yea.  That was number four on my search strings for my webstats.  What the hell??         See what I mean??  Very odd.  Today is the first day of my three day weekend and I am sickly again.  I was supposed to go to cemetary today, but that is out of the question.  My eye was glued shut when I woke up.  I hope it\'s not pink eye because that would royally suck.  I am hungry.  I don\'t think there is any food in the house though.  I hate that.  Sorry I am being so random ADD like.  I have a lot on my mind.  Well I am going to get going.  I want to do something.  Something other than type.   '),(23555,3820930,'01,August,2004','2004-08-01 00:00:00','    It happened.  I freaked out last night.  Thank satan for  Julie .  I am not going to get into what happened because it is nobody\'s business, but now it\'s done.  I am back to my old self again.  For those of you who know me, you know what that means.  I am going away this weekend to see  Danielle .  I miss her dearly.  I haven\'t really been paying attention to her and for that I am sorry.  I have a sinus headache right now.  And I am at work.  Blech.  I took a random pic of myself being sickly looking.  Here it is for your enjoyment:         Bye for now.   '),(23556,3820930,'01,August,2004','2004-08-01 00:00:00','    Here\'s a nifty little recipe I cooked up:  2 lesbians currently in a relationship 1 that has little self-control 1 that claims to be a good girlfriend 1/2 cup of lust 3/4 teaspoon of attraction 1/2 teaspoon of drama 1 quart of cheating  Stir vigorously until mixed. Bake at 375 degrees until fully aggravated. Serve.  Now you can have your cake and eat it too.   '),(23557,3820930,'01,August,2004','2004-08-01 00:00:00','    So I am home from work right now.  I did the new release wall and my schedule and put out a lot of overstock.  Here are the before and after pics:            I rock!   '),(23558,3820930,'01,August,2004','2004-08-01 00:00:00','    I don\'t know.  It sounded like a good title.  I am so retardedly confused about so many things.  Well, not confused because I know what I want, but frustrated might be a better word.  Life is always so difficult for me.  I suppose it\'s because I make it that way, but still.  I have to be at work at four.  Yea.  I am still in bed and it\'s three.  LOL.  And Mondays suck.  There is the schedule to do and then the new release wall.  I just have to make room for Butterfly Effect, but I need a whole bay to do so.  Sorry.  Video store talk.  I will take a pic of my handiwork tonight and put it here.  It is so amazing how I can get eight empty rows out of a full wall of movies.  I finally beat  Tiara  at Literati, but then later on in the night she kicked my ass again.  What\'s going on?!?!?!?!  Well, lemme go because I don\'t want to be late for work.  And come on people.  Comment on my entries!   '),(23559,3820930,'01,August,2004','2004-08-01 00:00:00','    This was too cute not to post....still playin Literati...since 4:30 AM   '),(23560,3820930,'01,August,2004','2004-08-01 00:00:00','    I LOST the Literati best of three against  Tiara .  It was sooooo close.  Now I also lost a bet so now I have to do something for her.  She hasn\'t decided yet.  I can just imagine!!   '),(23561,3820930,'01,August,2004','2004-08-01 00:00:00','    That explains my first customer of my shift today. I usually don\'t get annoyed by mentally retarded or challenged people, but this one took the cake.  Oh my holy hell.  This is an actual sentence she said over and over again:  \"Canyoulookupamovieforme? Iaminarushandtimeistickingticktockticktock.\"  She like didn\'t breathe when she talked and mushed everything together and had a lisp.  It was baddddd.  And  Andre  and  Jim  totally abandoned me at the front, so I had no choice but to help her.  She bought Brady Bunch in the White House.  Go figure.  Anyway, I had a cruddy day.  I slept til 2:45 and then showered and went right to work.  My back has been hurting so bad lately.  I routinely take 3 Advil at a time.  That can\'t be good.  When I get paid, I am going to go to the doctor.  Yay!   Madonna\'s  song \"Frozen\" is on.  I love love love that song.  I always wanted to make love to it.  Heh.  Speaking of making love, I need to do that soon.   '),(23562,3820930,'01,August,2004','2004-08-01 00:00:00','    I am truly addicted to this little gem of a game on Yahoo.  Me and  Julie  play it like it\'s going out of style and now I got  Tiara  playing too.  It\'s so simple yet so challenging.  You should all try it.  My Yahoo username is geekdykegrrl if you ever wanna challenge me.  Heh.  So anyway, it\'s funny how you are all bajangady with someone one minute and the next it\'s all good in the hood.  I guess that is how you can tell that person is a true friend.  With all my weird mannerisms and quirks, she still manages to put up with me.  Thanks.  You know who you are.  And if it\'s not you who is reading this pretend that it\'s you.  Feel special.   '),(23563,3820930,'01,August,2004','2004-08-01 00:00:00','    Work sucks.  I feel as though I am just going through the motions.  Everything is so overwhelming.  I have literally hundreds of previously viewed movies staring me in the face with no place to put them.       See what I mean!   Joe  and his boss are coming on the 12th.  I better get this store in shape.  I am sitting here right now playing Literati with  Julie .  I could be organizing the library.  Nahhhhhhh.  Some other time.   '),(23564,3820930,'01,August,2004','2004-08-01 00:00:00','    I have really bad timing.  Like when to oversleep.  I was supposed to be at work an hour ago, but didn\'t wake up until 12:45.   Joe  is there right now.  I feel like shit because I was up all night working on this new blog.  I still have to get in the shower and leave for work.  I wish I was the cat.   Spencer  is just relaxing now.  Screw him.   '),(23565,3820930,'01,August,2004','2004-08-01 00:00:00','    September 11th happens. The country is in turmoil. George W. is reading \"My Pet Goat\" with a bunch of school children as 3000 people lose their lives. This is the same man who, in his first year in office, was on vacation 42% of the time. He as well as all of us were blindsided by 09.11.01. Two days later on the 13th, Osama\'s family is allowed to leave the country when all other flights were grounded. Days, weeks, months pass. Osama goes without detection. You think we would have had the prescence of mind to interview his family members before we let them knowingly leave the country? We decide to look for \"Weapons of Mass Destruction\" in Iraq. We order Iraq to disarm or we will go to war. They don\'t comply because they have no WMD. We declare war on Iraq to help them \"rebuild their government\". As much as Saddam was doing bad things to his people, what gives us the authority to deem what is a \"proper government\"? I don\'t know if any of you have been paying attention, but our government is n\'t running very well right now. Exorbinant health care costs, taxes through the roof, and we are fighting a war we shouldn\'t have even started in the first place.   We\'ve lost more soldiers after the war ended then while the war was going on. Saddam has been captured. No \"Weapons of Mass Destruction\" have been found. We basically made a fool of ourselves, of our country for a pissing contest. Nobody cares if we are the richest, most powerful country on the world. In my opinion, we are rich with ignorance and power hungry. Why can\'t we be like Iceland? They don\'t even have an army. They don\'t need one. We recently turned over Saddam to his country so they can prosecute him. I feel bad for the man kinda. We pinned the 09.11.01 attacks on him indirectly. The american people were happy to have something, some shread of \"revenge\" on the Middle East. Our real war should have been with Afghanistan for harboring a terrorist. Osama bin Laden. Who we have yet to find.   '),(23566,3820930,'01,August,2004','2004-08-01 00:00:00','    I recently updated my site. I decided to go with this format. I hope all of you like it. It\'s a lot brighter. So, what\'s new with all of you guys? I am doing ok. I have been working a lot lately. Me and  Danielle  really haven\'t seen each other a lot because of her car situation and my work situation. I feel bad for that. I will have to admit that  Tiara  has been mentally distracting me. I need to stop that. I have a girlfriend who loves me very much and I need to remember that. I don\'t know. I am all bajang. I took next weekend off to get away from things. I don\'t know what I am going to do yet, but I will figure it out. I am at work right now typing this, so sorry for the choppiness and shortness of this entry.  Well, that is it for now. Visit the forums and post something. TTFN......   '),(23567,3820930,'01,August,2004','2004-08-01 00:00:00','    The concert was fucking amazing! I enjoyed all of the craziness that was going on. Every performance was great, but Linkin Park was the best by far. When they played \"Breaking the Habit\", I acutually started to cry at the beginning. I am so glad that I have my LP tattoo. I am going to post some pics from the concert. The first one is of me just having fun and it came out really good so I didn\'t want you guys to miss out on it. The second one was of me and  Julie  being silly. Then it\'s me again in the bathroom being all blee. Finally is a pic of  Chester  from LP I took when he was on the huge screen above us. Enjoy!   '),(23568,4125847,'09,August,2004','2004-08-09 00:00:00','   The shops by India Knight the book about how, why and where to shop. Personally I thought this was going to be a boring girlie book about shopping funny enough but to my delight when I did pick the book up, not only is it full of head nodding moments we have all been in. It is also written with a sense of humor you can\'t help but giggle. Even if you are on the train like myself tonight.  The main reason however I am now writing this in my blog is because this book has a good few pages dedicated to underwear including a section on \'Bras for big bosoms\' and \'buy your normal - bosomed girlfriend a bra from here\' and the invaluable advice, if you are unsure of sizes DO NOT BUY LARGE - great advice for anyone buying underwear as a gift.   urlLink Click here  to read more about the book   '),(23569,4125847,'08,August,2004','2004-08-08 00:00:00','   urlLink     Luxury deep red and corn yellow embroidered lace bra which once upon a time had a equally stunning matching knickers. The question is where are the knickers now?  Possibilities 1, Ex boyfriend 2, Lost 3, Ruined in the wash  urlLink   '),(23570,4125847,'07,August,2004','2004-08-07 00:00:00','   I am working on launching theunderweardrawer.com late in September. However I can\'t wait anymore so I am going to upload my artwork as it happens here on Pants HQ.   '),(23571,4125847,'07,August,2004','2004-08-07 00:00:00','   Has Agent Provocateur gone all Forbidden Flowers by Nancy Friday on us? They are asking people to send in raunchy stories just like the book but this way you can actually win a limited edition Agent Provocateur Thruxton 900 motorbike. Feel the power between your legs is a line that instantly springs to mind!  For the rather sexy lingerie retailer this is a campaign which is sure to standout and be noticed. Got to hand it to them not a bad bit of PR although be warned on the site it is a data collection exercise as you have to join the AP club in order to enter the competition not bad as you do get email updates.  Simple but worth mentioning that work email accounts might not be the ideal for this newsletter. Also if you view this site at work and your room is quiet then note there is auto play with sound on the homepage.   urlLink Click here  to join AP Club and enter the competition   urlLink Click here  to read more about Nancy Friday and Forbidden Flowers book   '),(23572,4125847,'06,August,2004','2004-08-06 00:00:00','   I am ecstatic. Have I died and gone to heaven? I wonder when I see the sign of my favourite lingerie store has got a 50% sale on. The adrenalin rushes as I see the shop door; my heart is in my mouth as I search the shop for the ultimate pieces of beautiful, exquisite and lavish lingerie. It\'s irrelevant if the piece matches anything else, fits properly or it\'s not my normal style; this is paradise, an Aladdin\'s cave of the best goodies which is better than gold. A pure moment of euphoria.   '),(23573,4125847,'06,August,2004','2004-08-06 00:00:00','   Dutch scientists have developed underwear that calls an ambulance if the wearer has a heart attack.   urlLink  Click here    to read the article   '),(23574,4125847,'05,August,2004','2004-08-05 00:00:00','   Sara Lee-owned underwear brand Wonderbra is set to launch a new range called Deep Plunge in February 2005. This is to suit \'risque\' tops and dresses but will offer support. The range follows the increasing trend into revealing clothing.  There is to be a number of styles in the range, including a bra that is transparent in the middle and has extremely low cups but still gives the \'push-up\' effect. Playtex marketing director Herve Baily says the launch will be supported by significant marketing spend, with the campaign likely to appear on television.   '),(23575,4125847,'05,August,2004','2004-08-05 00:00:00','   Bush filmed in his underwear..   urlLink Click here  to read the article   '),(23576,4125847,'04,August,2004','2004-08-04 00:00:00','   Pants HQ has been created as I have a passion for underwear. This blog will record news from the lingerie world, ideas, articles and thoughts for the day...all based around the subject of underwear. Areas of interest are underwear in news, lingerie in art, marketing of underwear, new products launches and personal encounters.  Underwear is my Otaku so watch this space...   '),(23577,4125847,'04,August,2004','2004-08-04 00:00:00','   Visit Figleaves to buy the latest in underwear fashions..also this is where I use to work so I know its a good place..   urlLink  Click here  to go to buy lingerie   '),(23578,3545726,'23,June,2004','2004-06-23 00:00:00','   Our field trip could have been a little better. The way it could be improved is by having more time to go somewhere farther and more fun. What I got out of it was information and visual learning about fish and sharks. Although I am not coming next year, somewhere that you can go next year to learn about animals, besides the zoo, is the rodeo or to the H & H Ranch, maybe you can go to Moody Gardens.  '),(23579,3545726,'18,June,2004','2004-06-18 00:00:00','   If I were in a ship that was hauntred, I would have been really scared. I probably wouldn\'t even think about going inside a haunted ship. Maybe if I was with a couple of my friends, then I would be a little less frightened.  '),(23580,3545726,'17,June,2004','2004-06-17 00:00:00','   I think that these ghost stories are fact, they are true. I can\'t prove it, but I know. I don\'t know why I think that it\'s a fact. My mom saw a ghost once, this happened about 10 years ago. My aunt saw a ghost, well, I\'m not sure if it was an actual ghost, but she said it was a monster. the monster was sitting on the bed next to her daughter. She said she was yelling at the \"monster\" to leave her daughter alone, the next night, the \"monster\" was gone.  '),(23581,3545726,'15,June,2004','2004-06-15 00:00:00','   A time when I have felt as though I was in a haunted place was when I was at Mexico. I was at my Aunt\'s house, and my cousins and I were telling old mexican stories and myths. We were in a circle outside at night, around a small campfire. One of the stories was about La Llorona and her children. Another was about the candyman. We didn\'t try saying \"candyman\" three times, everyone was too scared to do it. I felt scared, i don\'t know why, maybe because of the candyman or maybe because it was dark at the time.  '),(23582,3545726,'11,June,2004','2004-06-11 00:00:00','   The relative that I get along with and am close to is my mother. She is the one person I can count on, 2nd on the list is my sister. My mom is there whenever I need her. Even though she works 9 to 10 hours a day, she still manages to pay attention to her children. These times are really hard for her, economically and emotionally. To me, she is the most important person in the world, especially in this time of my life.  '),(23583,3545726,'09,June,2004','2004-06-09 00:00:00','   One really famous tradition for females is a Quinceanera. A Quinceanera is a celebration when a female turns 15. It\'s basically the only cultural tradidition that I know and is Mexican. I, personally, didn\'t have a Quinceanera, but I had a party when I was 16. I didn\'t want a Quinceanera because my parents couldn\'t afford it. My sister doesn\'t want a Quinceanera, she wants a Sweet Sixteen. There are so many things involved in doing a Quinceanera. For me, there\'s too many things to do, and it\'s too hard to organize it. First of all, you need the Damas and the Chambelanes.  '),(23584,3545726,'04,June,2004','2004-06-04 00:00:00','   I am 16 years old and I currently attend Jefferson High School. I like to play basketball whenever I have time. I like going to Six Flags Astroworld. I did go to Six Flags Fiesta Texas during the Spring Break. I\'m gonna go again in July. (I have a season pass.)Well, that\'s all.  '),(23585,4150699,'05,August,2004','2004-08-05 00:00:00','   I forgot!  At last! Somewhere I can share my sordid summer fiascos without fear of being found out! :) Well, for years and years, my friends and I have wanted to do this race called the \"Bare Buns Fun Run.\"  As you can probably tell by the name, the race is held at a nudist colony and everyone runs naked!  Well, almost everyone.  If you finish the race naked, you get a special shirt that says \"nude finisher.\"  If you chicken out, you get a regular shirt and tons of mockery from those who let it all hang out.  I was majorly into running in high school and college. My team was my life, we were all queens of the world, etc, etc, etc.  Running was a huge deal in my home town.  Hundreds of people would show up for high school races.  It was kinda a cultural thing.  A lot of famous runners came from my town or were somehow connected to my town later on in life.  Anyhow--- all of this to help you understand where I come from and why I would consider running naked time well spent.   You\'ll hear tons about my teammates, since they were a huge part of my wacky past, but I\'ll start by telling you a story about naked running. We started this tradition when I was a junior in high school.  After my team won the city championships, a few of us distance runners started doing a cool-down on the track with some girls from another team.  We ran lap after lap, visiting and winding down after a stressful day of racing.  Pretty soon it was dark and the only people left at the stadium were our parents. I can never remember how it was first brought up, but it came to our attention that it was absolutely imperative that we run naked around every track in the city starting that night at midnight.   We told our parents our plans, loaded ourselves into a couple of cars, and set out for our mission.  We had a few hours to kill, so we ate at Dennys, made an appearance at a high school dance (did some all-new dance moves that would have been pretty humiliating if we cared,)  and made a nuisance of ourselves pretty much until the time came. We drove to the first track, stripped down, and ran naked around the track as fast as we could.  Now, we should have been masters of stealth, but we were too busy laughing and slapping each others\' butts.  What ladies we were. :) We did a few more tracks that night and had to call it quits for the year.  We swore to continue the tradition, creating a nake bond between our two teams that would last forever. So, the tradition was born. Even today, my little sisters will run into people who can say, \"Hey, I ran naked with your sister.\"  Now, younger generations are running around tracks naked.  It\'s a tradition that\'s kept pretty secret, and we\'ve only had a few cases of innocent bystanders witnessing the event. I\'m sure that they don\'t really mind. Well, I took this naked running hobby to college with me.  I\'ve pretty much made it a point to spread the joy of naked running to all of my loved ones. :)  Well, when we heard about the Bare Buns Fun Run, we wanted to do it. Somehow, through the years, it never quite worked out. We always had running camp to go to and, later, we were camp counselors at the same camp.  It kept getting pushed back. This year,  however, was finally the year.  I rounded up some friends and both of my sisters (much to my mother\'s horror.)  We took the trip out to Eastern Washington and drove out to the nudist colony bright and early in the morning.  Of course, we instantly see some guys from high school.  It would have been embarassing, but they were naked too....  In fact, everyone was.  We thought there would be more chickens.  When we arrived, we saw our first naked guy out directing traffic in a little vest and nothing else.  As we drove on, we come across a big sign that says \"Welcome Home,\" and as we take a turn, suddenly we see fields, volleyball courts, and a pool filled with naked people.  Ahh. Liberating.  We strip down and engage in odd small talk with other naked people as we register for the race and tie our race tags to our naked bodies with red yarn.  Note: It\'s really hard to small talk without letting your eyes roam below the neck!!!  After a while, you realize that it\'s a lost cause, and you just openly look at other people.  I mean- if you go to a nudist colony, you\'re kinda asking to be stared at, right?   So, we all engage in some shameless staring and pretend that no one is looking at us and our flaws. When the race started, there was a bit of moaning as people got used to things bouncing around that normally don\'t.  Surprisingly, the pain goes away after half a mile, and the next 2.5 miles can be run in relative comfort.  Of course, when running naked, it is important to pretend you are graceful, lithe, and quick. (really hard on the self-esteem to acknowledge the truth--- NO ONE looks good when they\'re running naked!) Umm, I could say so much more about this, but suddenly I am really tired of this long-winded story.  If anyone wants to hear more, just ask!  '),(23586,4150699,'05,August,2004','2004-08-05 00:00:00','   Ahh. Welcome to the future.  Although I\'m only 24, I feel like I\'m a dinosaur when it comes to technology.  Many of my students (I teach at a high school) have these Blog deals, and they\'re always showing them to me.  Comes in handy--- I can check up on them and learn all their dirty secrets. :)  I especially love it when they mention me in their entries, complaining about papers I assign, etc.  :)  Anyways, after much thought and consideration (2 minutes), I decided to create my own blog.  You would think that I\'ve had enough of talking and writing, seeing as of how it\'s sort of my profession and all.... but it\'s summer!  I\'ve completely stocked up on ramblings and stories, and here I am- stuck without an audience!  Not that I miss the little boogers... because I don\'t.  They were pure evil the last week of school and I\'m still recuperating.  So all of you out there who are mean to your teachers-- Don\'t be!  We truly love you.  You just annoy the hell out of us sometimes.  '),(23587,4231133,'23,August,2004','2004-08-23 00:00:00','    wEee..! i\'m back.. hahax. now hor. mi wil rite a short one.. cos now in school.. havin home econ lesson.. sIaNz.. miss dear.. mind tinkin wat is he doing now.. haIZ.. amelia sit beside mi.. then duno wat she doing..ask mi how to get into kidscentral.. hahax.. lame.. then.. nvm.long story.. i go back then updae.. cher also cuming le.. *WiNKsS* '),(23588,4231133,'22,August,2004','2004-08-22 00:00:00','    baCk hOmE aFtEr a lOnG sTupId daE iN sChoOl.. firsT.. mR teO waNs tO sEe cHaIrpErSon.. tHeN pRefEcT waNs tO sEe cHaiRpeRsOn.. ThIs aNd thaT.. sHE Ma faN.. tHeN.. haIyO.. lOnG sToRy.. tHeN aFteR sChoOl.. haVe sCiEnCe tEsT.. aIyO.. dUnO whEthEr wIll paSs aNOt.. i aNYhoW do dE.. cOs i caNnOT rEmeMbEr ALL tHE noTes thaT i haVE meMoriSe thE niTe beForE.. haIZ.. aFtEr thaT wEnT tO t maRt tO eaT maC.. sO fUlL.. haHAz.. wEnT baCk tO sCHool tO lOok aT tHE aUdiTiOn.. buT oNlI sEe aWhiLe.. tHEn wEnT doWn wIF hoNeY tO plaY bBaLl.. oR raTheR sUn taNn.. tHen gO tO thE tOilEt aT thE d n T bLoCk.. aND cReaTeD aLOt oF gHOsT sToRieS abOut waT i saW iN thE tOiLeT.. waNteD tO sCaRe swEaT.. bUT shE haVE nO reAcTIOn.. nOt fUn at aLl.. haIyO.. saW eDwIn bUt nEvEr saE anYtiNg.. oNLi tOk duRiNG baNd.. tHen plaY tIlL qUIte laTe.. thEn.. tHen.. tHEn.. hahAx.. '),(23589,4231133,'22,August,2004','2004-08-22 00:00:00','    fiNalLy.. caN gO oNliNe lE.. haHaX.. yEstErdAe dIdnT uPdaTe mI bLoG.. mI wIll uPdaTe tOdaE aNd yesTeRdaE ba.. eRm.. yesTeRdaE.. lIke oRdiNaRy daE.. gO To sChOol fOr banD.. buT hoR.. yesTeRdaE hAd tHaT sTupiD fOotdRiLl.. sIaNz.. wE hoR.. caN bE cOnsIDeReD aBiT lUcKy.. kaNna tEaCh bY meI yIn aNd sOphiA.. iF kaNna teaCh bY tHaT sTupiD yaN lIng.. tHen diE lE.. thaT 2 guRlS.. bEry fUnNy loR.. tHeiR cOmMaNd aLso wRonG dE.. tHeN iF wRonG.. wE anYhOW dO loR.. tHeN tHEy giVe uS.. 5 MiNs breaK.. bREaK lE.. tHEy aSk wE aLl tO gaThEr bACk.. wE SloW.. tHEy nevEr saE aNYtiNG.. tHEn taT sTupiD yaN lIng cUm.. aSk wE aLl tO dO pUmpiNG.. sIAo oNe.. wE nO cHoiCe haVE tO dO lOR.. sHe saE dO 10.. tHeN laTer saE dO 4 mOre tHen 4 mOre.. sO UnfaIr.. paT oNli dO 10.. nEver miND lOr.. wORsT tIng.. wE dO aT tHe roaD.. fInE loR.. sHe wIn loR.. beCos of ThiS sTupID tIng.. mI huRt mI leG.. bUt lUckIly.. noT paIn.. haHAx.. tHeN wEnT baCk.. haD piAnO lEsSoN.. tHeN cHEr kEeP on saeIn thaT pIaNo eXaM 10 moRe daEs.. tHEn aSk Mi tO dO sCAleS.. plaY plaY plaY.. plaY uNtiL fOr 1 houR.. whOlE hanD swOllEn..nVm.. buT luCkY mI.. haHAx.. aT niTe unClE tReaT uS fOr diNnEr.. wEeE..!! eaT quITe alOT.. tHEn aT niTe.. waTch tV.. tHEn gO sLEeP lOR.. toDaE moRniNG.. gO cHUrcH.. siAnz.. spEnT aLmoSt 4 hOUrs thErE.. eVEn sIaNZ.. haIZ.. thEn cUm baCk eaT lunCh.. aNd dO tHE sUNDaE rOuTinE.. hAhaX.. haIz.. mOndaE haVE tesT.. sCiEncE sUm mORe.. mI wORsT sUbJeCt.. haIZ.. nVm.. neEd tO sTudY lE.. [ misS dEaR..] '),(23590,4231133,'20,August,2004','2004-08-20 00:00:00','    haIz.. tOdaE hOR.. sAw tHe fOtOs.. thEn eVErybOdy aLL kEpt ruShiNG tO sEe.. haIz.. tHen aFteR thaT wEnt tO t MArT wIF gUodOng, yInglinG, swEaT, hoNey aNd sElicIa tO eaT.. wHILE eaTinG.. wE waLk inTo fuRrY amIgo.. niCe sEh.. tHeN wEnt baCk tO sChool.. bEfOrE taT hElP cHEr da pAo fOod.. hahAx.. tOok iNsTruCtmEnt.. tHeN wEnt tO aTruIm tO plaY.. bUT NEveR wOulD i noE.. iT raIn sO heaVilY.. buT niCe.. haHAx.. wE play plAY unTil aRounD 5 plUs.. tHEn aMeliA fINiSh bBaLl pRacTicE thEn wE go AnD pLAy bBaLl.. tHe lONgEr wE plaY, tHe hEaVieR tHE raIn.. hAiz.. dUN waN tO saE lE.. mi oNe wRong mOvE.. cOS deAR tO bE sO anGry.. haiZ.. sOrRy deaR.. hoPe taT aFteR reaDin mI bLog.. yoU wONt bE anGrY.. haIZ.. bUt nEvER miND... dO aLsO dO lE.. yOu waN aNGrY tHen aNgRy bA.. jUs waNna sAe sOrRy..loVe yoU.. miSs yoU.. '),(23591,4231133,'17,August,2004','2004-08-17 00:00:00','    to  DaE.. hAiZz.. tirEd bUt cHaO jI fUn.. aNd i aCtuaLly plAy bBaLl.. hAhaX..canNot bElieVe siA.. aNd hOr.. chIneSe CheR saE thAt iN the cLAss.. onLi 5 paSseS iN cHInEsE.. i tOt i waS oNe oF tHE faIluReS.. buT.. haHAx.. i paSseD!! sO hapPY..aFteR sChoOl.. tRieD tO pOn fOr i T.. and wE oF cOurSe maDe iT.. tHen wE aT cLAss cHat cHAT.. tHen wENt tO t maRt tO eaT.. eaT lE.. aCtuaLly waNeD tO gO to hoNEy\'s houSe tO plaY wIF fIfI.. buT wE eNdeD uP plaYin bBaLl.. hAhAx.. i duNO hOw tO plaY.. buT wIF tHE guIdeNce oF hOneY, sWeaT aND doNg doNG.. hahAx.. i shoOt iN aBouT 4 baLls bA.. plaY fOr bOuT 1 houR.. sUN taNn uNtIl quItE daRk.. bUt i lIkE.. hEHeX.. haIz.. cHAo tIReD.. anD quiTe anGrY wiF huIshAn.. sHE saE mI aNd dEaR caN neVEr hAvE a 1 moNtH aNniVersArY.. waT siA.. sO waT iF yOu haVe rITe? i tInK aCtuaLly wE caN hAvE dE.. aLL thaNks to yoU.. wE havE tO breaK.. thaT was ThE paSt bA.. dUN waN to rakE iT up LE.. eX cLaSSmaTe caLl tO jiO mI ouT.. bUt haHA.. i hAvE dEaR deaR lE.. i loVe hIm.. lovE hIm uNtIl.. eRm.. aIyO.. alOt meaNS aLoT lORx.. siGn oFf... '),(23592,4231133,'16,August,2004','2004-08-16 00:00:00','    k.. todae hUh.. eRm.. basically nutting much loR.. go school.. then.. after school.. then cum back.. at home.. do hw.. watch tv.. etc etc.. haiyo.. = ) miss dear dear.. miss him lots lots.. jus wanna tell him sumting loR.. he give mi that kind of feelings that.. no one gives mi b4 loR.. he is very over-protective.. but i duno why.. hahax.. i like lay.. edwin hoR.. never give mi that kind of feelin b4.. edwin and i ah.. never fight, quarrel oR argue.. then everyting he give in to mi.. I DUN LIKE!! but.. we break le.. jus becos i dun wan to break him heart.. amelia thot of a way to lie to him.. peOple may think i am bad [  which i am.. hahax ] but i tink that is the onli way.. to make him not that upset ba..  aiyO.. dun wan to sae le.. mi now hoR.. jus miss miss miss dear dear.. where is he? wat is he doin? questions flashin in my mind about him.. cannot forget him every min, every sec etc etc.. aH.. better stop le.. later go crazy tinkin bout him.. LOVE HIM ALWaYS.. '),(23593,4231133,'15,August,2004','2004-08-15 00:00:00','    kK.. mI secOnd poSt.. haIYo.. todaE.. nutting much.. same as normal.. morning went to chruch.. then came back.. eat then study.. cannot believe sia.. i studyed for 3 houRs.. sianZz.. miss dear dear so much.. yesterdae forget to add sumting into the post.. but nvm.. mi add todae..  dear ah.. you sick.. stil go for bball practice.. next time cannot le hoR.. makE mI sO worry.. haiyo.. ohyar.. todae kanna scolded by mi daddy.. for no reason.. stupid lay.. haiyo.. nvm.. anyting.. i wan to watch \'zhen qing\' also cannot.. miss the most exciting part.. gotta end here..   miSs dear.. and love him always.. '),(23594,4231133,'14,August,2004','2004-08-14 00:00:00','    aH.. finally..!! all mi friends kept askin mi to create a blog.. but I was too lazy to create.. anyway.. first blog.. mus support mi woRx.. hahax.. in the morning had band.. woke up early as cher said that we mus be there at 8.45.. but.. haiyo.. both the cher were late!! That was the most angry part of all.. we woke up early.. went there early.. but cher was late.. never mind.. still had to face.. tat Yan lIng.. sum of you all may not noe whu she is.. but hoRx.. I intro you all to her.. she hoR.. tink that she is the worldâ€™s queen like that.. sum more say sweating the.. standard not good enuf to play clarinet 1.. haiyo.. anyway.. todae was a terrible dae.. luckily never kanna scolded by piano cher.. if not really sway.. sIan.. cum online then no one to chat wif.. haiyo.. mi end off here.. wanna noe more bout mi.. look at mi profile ba.. '),(23595,3451587,'25,May,2004','2004-05-25 00:00:00','   A novel from the 70\'s. Part of a series that kept going and going and going... This first story seems the most thought provoking. The others probably just develop the theme. I don\'t remember them either. More after I read them again.  The theme is...machines are evolving much quicker than the biologicals (which create the machines then die out). The trend is helped along by the machines sending out recon ships that report back.  A visitor to our Solar system, called a Snark by Nigel Walmseley(sp?), comes and goes but leads Earthers to a crashed ship in the Moon which may have aided human evolution 1 million to 1/2 million years ago.  A plot twister though is there have been a series of novas on a direct line in the direction from which the Snark came...suggesting that something is blowing things up.  '),(23596,3451587,'10,June,2004','2004-06-10 00:00:00','   Rock\'em Sock\'em book series here.  A universe where humans have split into Adamists, who are plain vanilla, tech users, and Edenists, who have geneered themselves into psychic connection with each other and their grown space habitats and grown ships, called voidhawks.  On a colony planet, Lalonde, something nasty has entered the universe, some evil energy, ooooh, that takes over people, yikes, and turns them into super bad guys,  whoa!  A ship captain, an Adamist, but with nanotech in his brain to connect him with computers, is a hero of sorts, name of Joshua Calvert.  His girlfriend is an Edenist, who is also a persona/human for a habitat called Tranquillity, her name is Ione Saldana.  By coincidence Ione and her scientists on Tranquility are studying the ruins of another alien habitat that self destructed, no doubt caused by the same bogies that are breaking out in this story.  Well written page turner with good characters.  This is part 1, the Emergence. A report on Part Two, the Expansion, will soon follow.  CCPL only has the second part, so I had to ILL the Emergence.  '),(23597,3451587,'13,June,2004','2004-06-13 00:00:00','    Okay, Savannah made it through two time zones without major mishap.  Now, she just has to get through a week in a land ruled by the Terminator.  Of course when she returns, how will I know it is her and not some clone experiment dreamed up by some movie studio hack seeking plot lines for a new Candid Camera knockoff Fall program?  Hhhmmn. Maybe I will have to pinch her and see if she squeaks?   '),(23598,3451587,'01,June,2004','2004-06-01 00:00:00','    Hey Bodine, Jenny Poo, and Barry Too,  What are you going to do with my daughter out there on the wrong coast?  Everybody is hoping you will take them all to lots of Museums!  Baaaannk!  Not!!  NO WAY Bodine! We got museums back home.  Savannah wants to get discovered by some Hollywood producer. Where do they go to do that?  Jennifer, maybe you can get Big Mom on the Tonight Show. That would be fun, eh?  No fishing Jeff. Nobody wants to go fishing.  What are ya\'ll going to do anyway?   '),(23599,3451587,'08,July,2004','2004-07-08 00:00:00','   Hey...Last week I met with my first girlfriend from Middle School.  What can I say, my heart panged and my mind paused again.  Sigh. Even after more than 30 years away from the place, she stopped my heart again.  I wrote a poem.    35 Years Later   Yeahâ€¦I saw her at the reunion.  After more than 30 years, I saw her and my heart dropped, stopped 	and started again.   My life as I knew it was ended. The careful fiction built up all those years  	crumbled.  I turned back to the stumbling boy who lost his heart.  We talked of nothing.  What to say?  Even back then, she had gone on with her life. I tripped myself into a bounding farce of dope, drink, exuberance, 	ribald humor, and buried depression.  You canâ€™t grieve what you wonâ€™t admit exists. You canâ€™t grieve what isnâ€™t there. Blah. Blah blah. Blah blah blah blah.  She didnâ€™t remember the princess ring. Or the two rubber bands or the cork. Our Christmas.  Her mother liked me? She liked my mother? What?  She thought  I wasâ€¦what? I never knew.  Now? What?  '),(23600,3451587,'18,July,2004','2004-07-18 00:00:00','    Teachers make the worst students. I have given workshops for them, and they talk constantly during the presentation. They talk to each other, and talk and comment on what you say, if they pay attention to you at all.     &nbsp;   They don\'t listen well, they spend the whole time talking and laughing with each other. They ignore what you say, and ask each other how to do the activities. Often trading the wrong answer back and forth.        They come late; they leave early. &nbsp;   &nbsp;   And they Cheat! They look back and forth on each others papers, comparing answers and changing them&nbsp;as their whims direct.     &nbsp;   And they spend the whole time griping about their students, and cracking on&nbsp;the bad&nbsp;behavior of&nbsp; the kids.   &nbsp;   Ha! &nbsp;  &nbsp;   '),(23601,3412947,'21,May,2004','2004-05-21 00:00:00','   urlLink Daily Times - Site Edition   Ok.  I feel vindicated. I\'ve said it before and I\'ll say it again, you never know what you gonna get when you breed these days it seems.  After living in Los Angeles for a few years, it seemed almost too casual to get plastic surgery to fix your \'flaws\'.  With shows like \"The Swan\" making it seem like a prize to get hundreds of thousands of dollars of plastic surgery, trainning, and coaching.  It seems now that a man has divoriced his wife after she had an ugly baby.  So now in addition to having a blood test, we need affidavits declaring we are who we are or listing any work we have done to modify our bodies.    I kinda agree with some of this.  Marriage is more than about just Love.  It\'s also about looks, a breeding partner.  You just don\'t want to have sex and babies with just anyone.  You want someone free of genetic abnormalities.  Bad teeth are warning signs.  Awkward body parts can be warnings about mutations, or just screwing your sister a few too many times in their family history.  It\'s just too much apart of the selection process for finding a great mate that you\'ll want to spend your life and have children.  '),(23602,3412947,'21,May,2004','2004-05-21 00:00:00','   It tough.   When will the product be finished. When can we start testing? When can we sell this thing. You said we would be done in March 2003 :)  That\'s my favorite. haha. I was foolish. I still am.  '),(23603,3412947,'21,May,2004','2004-05-21 00:00:00','   There\'s just something good about Indian food. The vegetables are delish, and the meat is great. Spicy Cool and soft Naan. I\'m hungry. Lunch is soo far away.   ok, well, 1 hour 13 min. but I\'m withering away. Can\'t you see?  '),(23604,3412947,'21,May,2004','2004-05-21 00:00:00','   Well, I\'m not this old yet.  Went for a drive yesterday w/ a 80+ year old man.  I\'ve never feared for my life as I did that day.  Crossing lanes of traffic, blinkers - what are they good for?, getting on parkways at 20 mph, going from 2nd to 5th..  and about that.  Old folks should move from the stick to the automatic, just for reactions sake.  But the best for last...  \"Can you find me a parking spot to pull through.  I hate backing out.  My last 2 accidents were from backing out.\"  Jeeze.  '),(23605,3412947,'21,May,2004','2004-05-21 00:00:00','   Ok, so somedays I feel older and somedays, I just feel old.  So I had a great time last night!  Went over to Rob and Becky\'s house since they were throwing their \"Come as a Porn Star\" party.  Courtney and I dressed as a doctor/nurse pair and we won the best dressed couple.  Let\'s just say she was hot and I was just arm candy.  But really, both of us started drinking around 7 and by 11:30, we were just tired, no buzz.  Don\'t get it.  She claims this never happened to her before.  I tried to console her with comments like, \"it must be just the stress\", or \"did you get enough sleep this week?\".  Maybe it was from the house hunt all day Saturday.  I did find an interesting home though.  Large rooms, wide halls, and a private backyard, but it is a garden home and I\'m not thaaaat interested in those but eh, this is my second foray into the housing market here in Birmingham.  I\'m excited and get that way when I see a house w/ a neat feature, but I\'m soo lucky to have Courtney along w/ me because she can see through all the glittery sparkles and moving flashes of home buying for the crap the house really is like...    Hopefully next weekend will go well and the future more defined.  '),(23606,3412947,'21,May,2004','2004-05-21 00:00:00','   He\'s so quiet and reserved...  but tonight, frantic. It seems that he got his paws wrapped up in a elastic cord that held one of his toys.  It wrapped so tight around his kitty paw that it cut off his circulation and in fact, he wanted nothing to do with me fixing it.  So off to the ER for him...  Well after he was sedated, no problem the vet said.    It was great though.  When we were trying to get it off ourselves, he meowed and cried...  so much in fact that his sister, Belle, came up and bit Courtney in the back a few times defending her brother.  It was sooo funny.  poor guy.  Well, all is better now.  Here\'s a picture of Belle when she was a kitten.  So much mischeif. http://mycathatesyou.com/newshowcat.asp?CatName=belle_080202  '),(23607,3412947,'21,May,2004','2004-05-21 00:00:00','   Wonder what it will be like.  There\'s the anticipation. Will I be let down? Who knows. Gotta try anyways. see ya.  '),(23608,3412947,'21,May,2004','2004-05-21 00:00:00','   This is the second school I\'ve lost and I just don\'t know what to do about it.  Maybe I should just stick to Mammals. That\'s it, the rule doesn\'t count unless it is a mammal.   My cats are still alive. Yes. Alive and well. Crazy, but alive.  There\'s still hope for little Rickies.  '),(23609,3412947,'21,May,2004','2004-05-21 00:00:00','   I hate it.   Driving home I had this fantastic link between September 11th and other attacks that have happened and a Tom Clancy book written a few years ago and it was lost to the ether as I listened to a song on the radio.  Hope it was worth the mindless daze I enjoyed as I drove home. Another idea died.  '),(23610,3412947,'21,May,2004','2004-05-21 00:00:00','   The check is in the bank. I worked at Octigon for over two years and nothing is the same anymore, except for   urlLink belle  and sebastian.  So I traded my office for a cubicle. No more plants yet. Proffesionalism is in... Shorts are out.  Dion is gone.   That\'s a relief. I never knew how much frustration I had to live with on a daily basis until he left two weeks ago.  My days at work are now soo much more enjoyable, it\'s actually a pleasure.  Everytime I would step near him, I could feel it inside me well up and then subside as I walked away.  I would go out of my way not to have to step near him just to stop these feelings.    Well, that burden is gone and I\'m glad.  For the first time in a long time, all I can see are great things now for the future. It has been a long time coming.  '),(23611,3412947,'21,May,2004','2004-05-21 00:00:00','   There\'s a girl, I know. She\'s the one I dream of...      Ha, she just stormed off into the bedroom to console herself in another roll of toilet paper.    The sick.   They are so cute, especially with a gallon of apple juice in arms.  She demands that I take back my offensive comment. oh. haha. You should have seen it.  An empty bottle of  APPLE JUICE flew through the air at me, striking my chair. So cute.  If only these silly boys understood the sacrifice that us girls make for their happiness.....I suppose they never will though...but one can dream.  SACRIFICE!   I came home in the middle of the day to pleasure her with treats, lunch, and my COMPANY!   Showered with gifts, my green thumb, tastes in music (AND YES Screaming Girls Are Music), what is there ever to be angry at?  especially with that green blanket.   it should sheild your back from every harm, dagger, and bow!  Well, enough silliness for tonight I suppose.  But there\'s a lot going on these days. Stress about what are my next steps?   Pay off my debt or buy a home?  time to crunch some numbers. just nervous. loan was ok. prequalified was eh.  just so much to do. and not just for myself anymore.       she looked into my eyes, take me to the clouds above.  '),(23612,3412947,'21,May,2004','2004-05-21 00:00:00','   I sat down with the man that will dertermine my immediate future today. The mortgage guy :)  He\'ll tell me Friday how much I can borrow to buy a home. I\'m terrified. Here\'s a guy that will look at all my debt, what I pay a month, and what I make a month and let me know what I am worth in terms of a loan.  If that wasn\'t enough, then there is actually trying to pick out a home! Where should I live?   Calera Cahaba Heights  I dunno.  It weird.  Something that I think would be absolutely aweful, someone enjoys.   Garden homes.  eek. Town Homes, even worse. Condo? ha.  I want my space == privacy. My tomatoes need their room to grow beautiful fruit. I\'m just looking for someplace that is comfortable to start a family.  life is good. I turned 26 yesterday. I played bingo last night. Practing being old.  Moved to the new office.   urlLink SunGard  bought the company I worked for and was a shareholder which is enabling me to buy a home too.  It reminds me soo much of my last job w/ Context Integration.   I\'ve been soo excited about it, but so tired too. I\'ve worked very hard and am amazed at how much my surroundings and the people around me affected my productivity. I liked the startup feel of Octigon, but I feel like I\'m really going to make a different at SunGard developing Octane8.  Lots of Change for me this past year. Memorial day was a week or so ago. I remember driving back from Chicago that Monday.  Everything is different now. But life is still familiar and comfortable. I can\'t wish for anything more.  '),(23613,3412947,'21,May,2004','2004-05-21 00:00:00','   I\'d have to say that lately, I\'ve definitly feel like I\'m missing something.   I don\'t take the same enjoyment from things as I have in the past. My records are collecting dust. There are no hardwood floors in the apartment. The converstation just don\'t seem to have the same depth.  I\'ve tried to find the spark I had in the past for some of these things but it\'s just missing.   There has been the home search that I\'ve now put off for another year. The new job is very demanding and has added a great deal of pressure to my daily life. I need to do this right because of the potential that exists within the position can give me a great toe-hold to what I want to achieve...  but I need to learn a lot in the mean time.  I think some of this being I have to leave the past behind to get the future I think I want. No more indie-pop? I miss Mates of State.  I had fun belting out those lyrics...    \"And when I stand here alone  I know speaking comes easy to you but I choke  And when I stand here alone, I know this  I know that since I\'m leaving so soon  I owe what I know\"  I just want to be that cool guy with good music, great converstation, and a fantastic leather couch. Maybe one day.  '),(23614,3412947,'21,May,2004','2004-05-21 00:00:00','   I do... but until then, you can see it in  urlLink Quicktime VR . Can\'t wait to get back there sometime... hopefully before I have kids.  '),(23615,3330196,'22,May,2004','2004-05-22 00:00:00','   i woke up without a hangover from lastnights drinking session, instead my body is in pain from my boxing class. it was a project from last year and only yesterday did i manage to realized it. it was fun and quit addictive. i never new till yesterday how a single round of boxing can exhume so much strenght. i finish four rounds and i tell you it was so exhusting and painfull to the fists. as if all the pain is not enough, i accidentally push the hot water button on the water dispenser that cause some swolleness into my lesf hand. anyway i finish my day yesterday at about 3:am the today..........  '),(23616,3330196,'20,May,2004','2004-05-20 00:00:00','   yesterday i finally made myself a member of redcorner. the idea of it was lingering in my mind since november, last year. i will start tommorrow and hopefully i will go on and reach my goal.....  '),(23617,3330196,'19,May,2004','2004-05-19 00:00:00','   yesterday was the day for my sister, she went to the u.s. embassy for a visa application/interview. i never prayed to god to help her pass the interview, conciously i want her to leave because that will mean cd\'s and t-shirts. sub-conciously it was different. i am scared and dont want her to go. she doesnt need it. our parents can provide more than enough for both of us, but then again its her life and her decision. she passed and if all goes smoothly she will be living by next month. what will happen to her there is among the many worries me and my parents will endure. we just hope that she succeeds and enjoy life as it comes....  '),(23618,3330196,'16,May,2004','2004-05-16 00:00:00','   \"why are you an artist?\"......... \"because i have lots of pain\"...... i never saw any episode of six feet under but that may change eventually. i am not an artist but i have lots of pain, lots of disappointments. though unlike before wherein i am under a severe depression, my pain now comes from disappointments and regrets. as much as most of us i also want to go back in time to correct past events that makes me ache in pain now. i am a lonely shadow, always hiding at the back and in search of light so atleast i may be seen and noticed. great regrets often produced fatal outcomes but then again where is the sense in that? i overcome the kindling of suicide and self pitty, i am often smiling now and i can sleep with out any aid of pills, but those are superficial and faux. i may be what others percieved me to be me, but still i am in pain.................  '),(23619,3330196,'13,May,2004','2004-05-13 00:00:00','   may 14, my sleep was disturbed by a txt message from budjie, telling me that they are leaving for puerto gallera. i am supposed to be with them the next morning but desided to follow them on the same day. i traveled alone for the next 2 hours. it was a nice drive almost no traffic, but save from the occational boredom add the loss of radio signal, and a helpful hitchhiker, everything is smooth and fine. by 1 pm, my ferry docked onto the shores of white beach. the first five hours was kinda awkward because jaypee was there, but then again we went there for some good time, any grudge must be left at home. by 2 am the next morning i was tired because it has been a long day, by 9:30 am i am leaving the beach to head back home.  '),(23620,3330196,'13,May,2004','2004-05-13 00:00:00','   this is my first entry and have nothing to say. im hoping i could do this everyday so i can journal my life as i am going into a great bordedom and uncertainty about my future career and relationships. everyday for me is a routine, quite a cycle of very little assortment, less variation, but i am already used to it. no complaints, just ranting, i guess.  '),(23621,3330196,'08,June,2004','2004-06-08 00:00:00','   its been a while since i posted a blog, i was sort of having a writers block, what ever that means. anyway its been raining for some time now, but it is okay. the last three months was scorching and the rain was a welcome relief. though for peewee it means losing income because his into carwash business. i\'ve been going to the gym religiously for three weeks now, im not complaining actually i am enjoying and quite happy with the results. i am more agile and athlethic and give me another month and i will be changing my wardrobe. but not everything is good. the continous rise in fuel will likely go on and it sucks. everyday to and from gym i am spending a good php300.00 for gasoline alone. i change my driving style completely, i hardly floored on the gas pedal but still it consumes gasoline. my body also hurts and i still cannot get the right punch, footwork and position in my boxing class so i always get some flanking from my instructors. i need to improve on my boxing skills.............  '),(23622,3330196,'04,July,2004','2004-07-04 00:00:00','   i miss my sister............. anyway today i am very angry. i cursed someone and i am hoping he would die today, if not sooner. he is so abusive he deserves to die. i am not giving back that one..............i mean it............ i hate this feeling when anger and frustrations are eating in up to my system. i cannot function very well. i cant think clearly. i am so stressed. i just wish all the people who aduse and burden to me and my family will die. i guess if that happen my family will live peacefully and hopefully happy and contented........................  '),(23623,673235,'21,July,2004','2004-07-21 00:00:00','    '),(23624,673235,'26,May,2004','2004-05-26 00:00:00','   Lord of the Rings: The Return of the King  has officially hit the stores this week (I bought the first copy at my local Borders yesterday).  Trilogies have been popular in the movies for years.  This week we look at questions based on some famous cinematic triumvarites:  1) The  Back to the Future  Trilogy sent Marty McFly 30 years into the future, 30 years into the past (twice!) and 100 years into the past.  Which of these time periods would you like to visit in your 1.21 gigawatt-powered DeLorean, and why?   (Note: Thanks to  urlLink  Logtar   for the idea for this question)   2) The  Lord of the Rings  Trilogy was, of course, based on a famous trio of novels by JRR Tolkien.  What sci-fi/fantasy/genre novels or series would you like to see turned into a Peter-Jackson-directed/George-Lucas-written/Steven-Speilberg-produced trilogy?  3) Next summer, there will be two complete  Star Wars  trilogies in the history books.  What\'s your strangest or most interesting Star Wars experience?   Bonus)  While the  Jurassic Park  and  Indiana Jones  movies are currently only trilogies, plans are in the works for a fourth movie in both series.  What trilogy might\'ve been better served with a fourth movie?  Which trilogy would\'ve been better if they had left it at only one movie?  That\'s it till next time!  Be sure to leave your name and blog address in the Comments when you\'ve answered the questions.  And remember, as the  Terminator  said in the first part of his famous trilogy, \"I\'ll be back.\" urlLink   '),(23625,673235,'19,May,2004','2004-05-19 00:00:00','   urlLink Shrek 2  opens this weekend in theaters, hoping to be an even bigger hit than the prequel. From  Antz , to  Mulan , to  The Lion King , to  Heavy Metal , even  Anime ; There are hundreds of animated movies.     1. What is your favorite animated movie?  2. Which animated movie is your least favorite?  3. Who is your favorite animated movie character?   We miss you in Toontown. Thanks for playing. Don\'t forget to leave a link to your blog in the comments, so we can all see what your thoughts are on animation.   '),(23626,673235,'30,June,2004','2004-06-30 00:00:00','   Ok, this is the week a lot of you have been waiting on.   urlLink  Spiderman 2   opens, and the web-slinger (who in reality would have no shot at Kirsten Dunst) swings back into action against the evil Doc Ock.  This week, our questions are about those quintessentially misunderstood, ill-treated, feared and respected champions of truth, justice and the Spandex Way: the Superheroes.  1) All right, have it out.  What is the all-time greatest superhero movie?  2) What superhero movie has not been made that you\'d like to see? Which existing one would you have redone from scratch?  3) What qualities would be necessary to elevate a superhero movie from a simple action film to a memorable motion picture?  Is it possible?  BONUS) Most superhero movies take place in a more-or-less realistic modern world.  If a superhero actually appeared in your city today and started fighting crime, what would be the actual public reaction?  Don\'t forget to put your name and website in the comments section, and remember:  And remember,  urlLink  EXCELSIOR  !   '),(23627,673235,'24,June,2004','2004-06-24 00:00:00','   urlLink Two Brothers  opens this weekend in theaters, the story of two twin tiger brothers that are taken into captivity and seperated. Ahh, the drama. We all love movies that are based upon a cute, cuddly, little animal. Well, I guess King Kong wasn\'t so cute and cudddly. Or little for that matter. So yeah, you guessed it. Its all about animals this week.   1. Whats your favorite movie with an animal as the main character?  2. Least favorite? And is it your least favorite because of the species of the main character, or just because it was a bad movie?  3. Do you prefer your movie animals to be animated or real?  Bonus~ If you were to star in an animal movie... What type of species would you want your leading... animal to be. And why? C\'mon be creative. Even name your animal flick.   Me and my skunk?!?! Yeah, have fun. Remember to leave a link to your blog or your answers in the comments. \'Til next time...   '),(23628,673235,'17,June,2004','2004-06-17 00:00:00','   In honor of Father\'s Day, this weeks questions are all about Daddy Dearest. From  Robin Williams  portrayal as a father willing to do anything to see his children in  Mrs. Doubtfire  ,to the sins of  Chris Cooper  in  American Beauty , all movie dads have made us feel one way or another. So no matter if your dad is a saint or something... else, today is their day.      1. What actor has best portrayed the role of loving father? Name the actor and the film.   2. What actor has best portrayed the role of world\'s worst father? Name the actor and the film.   3. If there was a movie to be made about your life, what actor would portray your father?   Happy Father\'s day to all of you fathers out there.   I love you, Daddy! Remember to leave a link or your answers in the comments.  \'Til next time...  Daddy loves ya.   '),(23629,673235,'09,June,2004','2004-06-09 00:00:00','   urlLink  Harry Potter and the Prisoner of Azkaban   is now pulling all kinds of young witches and wizards away from their Playstations and into movie theatres all across the country.  That got us Muggles to thinking...wouldn\'t it have been cool to be able to point your wand at the bully that pushed you around on the playground, say \"EXPELLIARMUS!\", and watch them fly back against the wall?  1) Which magical power would you like to have (in order to further the cause of good for mankind, of course)?  Which magical power would you like to have that you would keep hidden, guiltily, cackling about it under your sheets at night?  2) If you had a flying broomstick, would you use it?  Where would you go?  3) If you could briefly turn yourself into an animal, what would it be?   Bonus)  Do you believe in magic?  That\'s it till next time!  Be sure to leave your name and blog address in the Comments when you\'ve answered the questions.  And remember,  Traguna Makoides Tregorum Satis Dee !   '),(23630,673235,'02,June,2004','2004-06-02 00:00:00','   The Day After Tomorrow  has opened to  rave   great   passable  some reviews, and disaster movies are once again in vogue.  This week\'s Theatre Thursday questions are about those motion pictures that make us want to avoid the water, planes, tall buildings, large cruise ships, angry atomic reptiles and everything in general that makes life enjoyable...  1) What\'s your favorite disaster movie of all time?  2) Do you prefer a disaster movie with tons of special effects, carnage and general mayhem (i.e.  Armageddon ) or would you rather see one where the focus is on the people and the event\'s effects on their lives (i.e.  Deep Impact )?  3) Of all the disaster movies you\'ve seen, which depicts the most realistic portrayal of what might happen in an actual disaster?  Which was the most far-fetched?   Bonus)  Which character would you be if you were in a disaster movie - the tough hero, the corrupt businessman who profits from it all, the old lady who runs around and screams \"We\'re doomed!!!\", the kid who saves the day, the young guy who buys it in the first five minutes of the disaster, or another stock character?  That\'s it till next time!  Be sure to leave your name and blog address in the Comments when you\'ve answered the questions.  And remember, in the immortal words of Maureen McGovern:   There\'s got to be  urlLink  a morning after   If we can hold on through the night We have a chance to find the sunshine Let\'s keep on lookin\' for the light   '),(23631,673235,'28,July,2004','2004-07-28 00:00:00','   This week the topic is YOU.   1) If there was a movie to be made about your life, what actor/actress would portray you?   2) What actor/actress would portray your love interest? Your best friend? &nbsp;   3) What would be the title of your film?   Bonus) What moment of your life would be the highlight of this film?   Be sure to leave your name and blog link in the comments. If you don\'t have a blog, post your answers in the comments.   '),(23632,673235,'21,July,2004','2004-07-21 00:00:00','   This week\'s Theatre Thursday is dedicated to those Onscreen Moments that stay with you forever.  1) What moment from what movie still makes you laugh out loud - no matter how many times you see it?  2) What moment from what movie still makes you cry like a baby - no matter how many times you see it?  3) What moment from what movie made you actually turn your head from the screen - either in fear, revulsion, or contempt for the fact that you actually paid money to see the film?  BONUS) What is one single moment from a film that is indelibly etched in your brain?  Not a scene or a sequence exactly, but three or four seconds from a movie that contain an image or phrase or concept that transcends normal movies?  Take a Moment and answer these questions, and you\'ll feel you\'ve flown across the moon..  Be sure to leave your name and site address in the comments!   '),(23633,673235,'15,July,2004','2004-07-15 00:00:00','   This week sees the opening of the film adaptation of Isaac Asimov\'s classic \"I, Robot\" starring Will Smith.  1) Who or what is your favorite movie robot?  2) Will we ever see the day where walking, talking robots truly become a part of every day life?  3) Should a robot ever be programmed to kill a human?  BONUS) If you had a personal robot, what would you call it?  What would you like it to do for you?  DANGER WILL ROBINSON, DANGER! (if you don\'t put your name and blogsite in the comments)   '),(23634,673235,'07,July,2004','2004-07-07 00:00:00','   Doing things a little different this week due to hectic schedules. Sorry if the questions are a little boring this week.   This week its all about your favorites.   1. If you could watch only one movie for the rest of your life, what would it be?  2. What is your absolute favorite movie soundtrack?  3. What\'s your favorite movie-time snack?  4. Favorite Actor:  5. Favorite Actress:  6. And last but not least, what is your favorite kind of movie: comedy, drama, action...etc.,???   '),(23635,673235,'05,August,2004','2004-08-05 00:00:00','   urlLink Starsky & Hutch  is the top selling DVD this week. With that in mind, this week the topic is TV shows that have been turned into movies.  1) What movie based on a television show is your favorite?  2) Your least favorite?   3) What television show would you like to see turned into a movie?   Bonus) Should the actors that portray these characters on TV also be used in the movie? Or would you like to see different actors?   '),(23636,1292625,'15,March,2003','2003-03-15 00:00:00','   hey yall sorry lu about the benji thing but i always thought of him as a little yello smiling puppy. awwwwwwww well u guys to morrow is sunday man i hatesunday cuz its a day closer to monday. well man lu u better be cleared up by monday or its gonna be bad well thomas ur party was really great...sorry my leafs tore b4 i managed to but them on not of very good quality. well see ya all in hell with ceci  muahahahahahahahaha bye bye xx   '),(23637,1292625,'14,March,2003','2003-03-14 00:00:00','   hope u all have a good time at the party, tonight im getting a massage, shiatsu, and will e good as new... so the night wont be a complete loss neways.... lodza farts krish   '),(23638,1292625,'13,March,2003','2003-03-13 00:00:00','   well, this is another note in honor of bojoso benji, u will always be with us.... expecially with lou.... we will always remember your smily relaxed face...  you are and will forever be lus benjamin de oliveira (the fluffy bojoso) neways.... ill be doing a complete makeover tomorow night... face masks, massage, painted nails... the ol new(ish) me, hehehe neways... tata krish   '),(23639,1292625,'12,March,2003','2003-03-12 00:00:00','   My leafs are purple, lol, no armaini pants we be worn for the first time got to go do 4 bald spots hw. bye bicthes (nbs)   '),(23640,1292625,'12,March,2003','2003-03-12 00:00:00','   hello this is me ... No offence sof but that guy was talkin about Caca because she wanted to fica with him but he ha a gf! i dont you dare call benji the king of all dogs a fat fur ball because i loved him and i still do!!!! oh well cant wait to go to your party and i cant wait to wear my leafs!hehehehehehehe!!! oh well c ya in hell hole 2morrow!!! bye love me   '),(23641,1292625,'12,March,2003','2003-03-12 00:00:00','   isa apparantly doesn`t know that I Thomas, the duke of tampons has acses to this and really don`t wanna know about hairy leg. no realy.  Tommorow we shall be in mouning in hounour of Lu and Benji. Not a good week; Sara, Benji, Penises and I really don`t have a good feeling about the party... Lu seems to fine under the given circumstances. I really hope the party will cheer her up on Friday  Sencirely Le Duke of Le Grand Tampon   '),(23642,1292625,'11,March,2003','2003-03-11 00:00:00','   hey guys  bells the cute guy was the one that was always on the sofa with the other guy and the were both wearing white shirts. the one cute talke about me to the other guy yay!! o well sorry to hear about benji lu  well he is up there with the angels I really hope margo chears up awwwwwwwwwww i will miss that fat fluff ball well bye guys talk to u at pelunia land xx   '),(23643,1292625,'11,March,2003','2003-03-11 00:00:00','   sarah, i hope u feel better, nasty car should be broken appart and the driver put to death,  no seriousely.... i hope u get better sooner than expected   '),(23644,1292625,'11,March,2003','2003-03-11 00:00:00','   whitch is the cute guy again and how did u break a window ssm???   '),(23645,1292625,'11,March,2003','2003-03-11 00:00:00','   oh how gay, silly putty penisis, sounds like penisis being sillily put in someone, hehehe, it sounded funny in my head oh well... there is a seriouse case of hair issues in the leg department, i have two bald stripes on each side of my right leg, due to some verry painfill \"bonding\" relation ship this afternoon with sofia and louisa, u 2 will pay, i can still feel the pain, and my leg is still red... wwwwwwell,  neways gotta get some more chcolate, luv krish(zulu 4 me)   '),(23646,1292625,'11,March,2003','2003-03-11 00:00:00','   PS lehgrandtampon was created by thomas The name not the blog the blog was gay sofia   '),(23647,1292625,'11,March,2003','2003-03-11 00:00:00','   repeatedly...more to come   '),(23648,1292625,'11,March,2003','2003-03-11 00:00:00','   pennises were also repeatedly made with silly putty given by sofia to us on her birthaday (we were meant to giver her shit right...oh well). After the penies were made out of blue silly putty they were thrown around the class.   '),(23649,1292625,'11,March,2003','2003-03-11 00:00:00','   Sofs the party is around the coner but i am sure u will be dazzling (as a guest of hounour u better).   '),(23650,1292625,'11,March,2003','2003-03-11 00:00:00','   Aparantly cute guy from Gabi`s party is going. Sofia hope ur mom getts better soon. Today i broke a window in English Grammar. Miraculously the fans were on and the windows open. Windows latter to be broken.   '),(23651,1292625,'11,March,2003','2003-03-11 00:00:00','   O my god this has been like almost the worst day! I mean the bad things REALLY sick my mom got hit by a car got totally bored at home k now good things my mom is ok I missed school got to sleep until 12 well o my god thomas i have no idea to wear to yuor party and am getting totaly stressed i hope cute guy that was at gabi\'s party goes. but he probably won\'t well gotta go bye  mb I\'ll write later tonight   '),(23652,1292625,'22,March,2003','2003-03-22 00:00:00','   The gala wasn\'t so bad...it was okayish. I still can\'t get over last night I had so MCUH fun!!!!!!!!!   '),(23653,1292625,'21,March,2003','2003-03-21 00:00:00','   to tired to write!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!   '),(23654,1292625,'20,March,2003','2003-03-20 00:00:00','   Fun movie, kool dinner and no money and many french people that was today!!!!!!!!!! And agudo moaning about chips!!!!!!!!!   '),(23655,1292625,'20,March,2003','2003-03-20 00:00:00','   hehehehehehehehehehehe FAME nooooooowadays, u can like the life ur living u can live the life u like, u can even marry harry but mess around with ike   '),(23656,1292625,'18,March,2003','2003-03-18 00:00:00','   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!SOFIA GOT HER PERIOD!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!   '),(23657,1292625,'18,March,2003','2003-03-18 00:00:00','      '),(23658,1292625,'18,March,2003','2003-03-18 00:00:00','   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!SOFIA GOT HER PERIOD!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!   '),(23659,1292625,'18,March,2003','2003-03-18 00:00:00','   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!SOFIA GOT HER PERIOD!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!   '),(23660,1292625,'18,March,2003','2003-03-18 00:00:00','   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!SOFIA GOT HER PERIOD!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!   '),(23661,1292625,'18,March,2003','2003-03-18 00:00:00','   ahhhhhhhhhhhhhhhhhhhhhhhhhhhhhhh soooooooooooooooooo bored!!!!!!!!!!!!! can not get worse now found out I have a fever!!!!!!!!!! this is not good!!!! USA started bombing bagdad trying to explode all the little forts of the irakies well gonna go watch a lil more news that how bored I am bye   '),(23662,1292625,'18,March,2003','2003-03-18 00:00:00','   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!SOFIA GOT HER PERIOD!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!   '),(23663,1292625,'17,March,2003','2003-03-17 00:00:00','   what ever, soooof, dont feel bad its ok.... well neways, i foud out something really happy today hehehehehehehehehe bibi retard by sofs bye lu   '),(23664,1292625,'17,March,2003','2003-03-17 00:00:00','   For those that have no ideia of war and geography (isafeia) Iraq has no bases in where in hit range of the states!!!!!!! It also doesn\'t have a navy since the place is land locked!!!!!! So NO there are no possible ways of Iraq touching the states!!!!! Also the us will blow up any man, ship or aeroplane with a burret (gay iraqi hat)   '),(23665,1292625,'17,March,2003','2003-03-17 00:00:00','   Iraq will not attack the states they don\'t have any way how to U REATARD!!!!!!!!!!!!!!!! READ THE NEWS!!!!!!!!!!!!!!!!!!!! If they review the draft and u have to go to war my brother will have to go...but review the draft...dubius...   '),(23666,1292625,'17,March,2003','2003-03-17 00:00:00','   man bells the reporter sux I do thik that the usa will go to war but u still have to wait for bush to give the word and england its really sad one of my sis\'s friends might be drafted well bye to sad to talk much xx   '),(23667,1292625,'16,March,2003','2003-03-16 00:00:00','   hehehehehe how retarded....   '),(23668,1292625,'16,March,2003','2003-03-16 00:00:00','   we is not happy with this, we dont want war... no no, war goood yeees, nono, smeagol dont want war, we will get hurt, nono   '),(23669,1292625,'16,March,2003','2003-03-16 00:00:00','   oh no, i cant believe the us is going to war..... thats nooooot good! i just saw on the news that the war could start as soon as wednesday night... iraque will fight bak wont it??? that means that it might screw up some places in the us.... this really bad! well, hehehehehehe, mark looks like a piglet in that pic.... well ttfn me   '),(23670,1292625,'16,March,2003','2003-03-16 00:00:00','   hehehehehehhehe read a really funny joke hehehehehehehehehe my dad is pissed off that sp lost and my sister learnt about 20 new swear wod at the match.... lu ur dad must be mo\'  happy heheheheheheehehehhe c u all  in hell hole bells   '),(23671,1292625,'16,March,2003','2003-03-16 00:00:00','   ceci is our hell master muahahahahahaha lu lee is coming on thursday he might visit St.Pauls u know he\'s still a lil pissed cuz we threw a chicken on him remember us going into the room again and hearing it? hahahahahahahaha aaaaaaaa that night was sooooooo fun!!!! passing by the car and hearing all our drunk elders having dumb fights! hahahahahaha going all the way to the old house just to get a CD at 5 in the morning. man that was the best Ai I love new years the rest of u have to come to my farm the next time we do a big party like that well bye guys nm happend to day exept i went to lunch   '),(23672,1292625,'16,March,2003','2003-03-16 00:00:00','   FUCKING KICK ASS PARTY!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! At the end Lu, Sofs and I swam in the pool!!!!!!!!!!!!! Music wasn`t so good, honestly felt like bitch slapping the DJ!!!! Going to sleep was a bit hard as the gals (luisa and sofia) kept calling and retardedly quoted scarie movie HAHAHA!!!! Seems like the people in our class are fucking mad they gave huge presants!!!!!!   '),(23673,1292625,'29,March,2003','2003-03-29 00:00:00','   how was mexico thomas, o speaking of u i dont mean to be mean but i found a fofoca site on st pauld that said ur party sucked big time..... no offence so... ne ways... just watch da shining... i dont get the dog giveing the old man a blow job though...   '),(23674,1292625,'25,March,2003','2003-03-25 00:00:00','   Boaring weekend!!!!!!!!   '),(23675,1292625,'25,March,2003','2003-03-25 00:00:00','   Sad blog....  nobody blogging ne more.... *not good!* Meu white is competing with Liz in the weird noises world!   '),(23676,1292625,'25,March,2003','2003-03-25 00:00:00','   btw,... that was CAMILAs (aka.. the one typing now) who came up with the purple smoke. copyrighted.MUAH HAHAHAHAHHA!!! mine... all mine... and now, you can all run off and play little lambs. *BANG*   '),(23677,1292625,'24,March,2003','2003-03-24 00:00:00','   hi there..am checking something... will post properly in a few moments.. ta da *puff of purple smoke*   '),(23678,1292625,'24,March,2003','2003-03-24 00:00:00','   MAN THOMAS THAT WAS REALLY BITCHY OF U!!!! BELLS LU WAS GONNA INVITE U BUT SHE DIDN\'T WANT TO MAKE U GET EVEN ORE UPSET WHEN UR PERANTS SAID U COULDN\'T COME! WE REALLY MISSED U THERE!!! well I am sooooooooooo tiered I could sleep for a year!! man lu need to talk to u tomorrow!!! see yall bye xx   '),(23679,1292625,'23,March,2003','2003-03-23 00:00:00','   I am sooooo fraking tired...... and i really dont feel well gawd..... i feel reeeeeeally ill   '),(23680,1292625,'23,March,2003','2003-03-23 00:00:00','   lu, u had a party??? or is there goin 2 b 1??? am i invited, mayb i can convince my parents to let me go!!!!   '),(23681,1292625,'23,March,2003','2003-03-23 00:00:00','   FUCKING KOOL PARTY LU!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! HAPPY BIRTHDAY!!!!!!!!!!!!!!! KOOL PARTY THE BEST EVER!!!!!!!!!!!!   '),(23682,1292625,'23,March,2003','2003-03-23 00:00:00','   hehehe, i went to te game yesterday meu, peeps swear soooo much that you come home talking in swear words, and before going to the game i watched the osbourns, they are soooo fucking funny, so i was mixing portuguese swear words with english ones, like meu you fuking filhilho da puta, go fuck a cunt da sua mae.   '),(23683,1292625,'04,April,2003','2003-04-04 00:00:00','   I MEAN IT... IM IGNORING YOU diont want to get hurt again   '),(23684,1292625,'03,April,2003','2003-04-03 00:00:00','   lu gave me thge Thong*, perscribed** My grandmother is dying by the way, SO NEVER SAY THAT AGAIN!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!11 I love her alot and that is really insulting, u dont even know her, how would you like it if i said that about your grandma, u wouldnt, so GROW UP!!!!!!!!!!!  SCREW YOU JACK ASS   '),(23685,1292625,'03,April,2003','2003-04-03 00:00:00','   well, by the way lu gave me the thing so drop it tell me what you have against me and ill leave u alone... promise wont even talk to u well... listen...make my day, im having to take anti-depressants persribed to my by a doctor and i dont need you makeing me feel worse... god... just let me be dude, STOP INSULTING ME, do me a favour and ignore me as you said you would... dont try and make lu and sofs not talk to me again or you will be sorry.... i didnt write that site and i just thought you should know about it... almost everyone else did... stop bothering me kid cause i aint doing anything to you... what the hell is t.a.t.o and i am not a lesbian... i never have been and i never will be... i like men... noooot women.... i did NOT try to kiss lu EVER... i do not try to kiss women.... get off my back... so what... we have the same best friends... im giving up trying to be yours... but just let me get on with my already shitty life because i dont need you to fuck it up even more... STOP IT!  I AM SICK AND TIRED OF YOU!!  LEAVE ME ALONE AND I WILL ignore you... xxx me PS  the people who wrote the site did go to ur party. i know who it is   '),(23686,1292625,'03,April,2003','2003-04-03 00:00:00','   I KNOW!!!!!!!!!!! MY PARTY ROXED!!!!!!!!!!!!!!!!!!!!!!!!! Thoese peps didn\'t even go to mi party so they can\'t write ne thing!!!!!   '),(23687,1292625,'01,April,2003','2003-04-01 00:00:00','   well guy am sick of looking for sience stuff bells get better thomas the what the peeps worte on the site is bull ur party was kick as at leats to me and lu well gotta go falling asleep bye xx   '),(23688,1292625,'01,April,2003','2003-04-01 00:00:00','   Pelunia, tetunia, popozunia   '),(23689,1292625,'31,March,2003','2003-03-31 00:00:00','   Luisa, Sofia GENERAL ANNOUNCEMENT!!!!!!!!!! ALL MESSEGES FROM MISS SEE THROUGH THONG AND LESBIAN LOVER (GO JOIN T.A.T.O.) WILL BE COMPLETELY IGNORED BY MI UNTILL FURTHER NOTICE. SEE YA SOFS, LUV MI!!!!!! XXX   '),(23690,1292625,'30,March,2003','2003-03-30 00:00:00','   HEHEHEHE the site is plantaof or whatever.... i didnt write it by the way... i just fond it through another gossip site.... neways, hope u feel better lu.... soooooof..,.. i neeeeeeeed 2 tlk 2 u well tata moi   '),(23691,1292625,'30,March,2003','2003-03-30 00:00:00','   Okay sad people that have ne better to do descided to bicth about mi party on the web... sad people with no lives... I know that the Party suxed but they don\'t have to bicth about it to the whole world ne ways.... boaring sunday.... My brother got drunk on tequilas yesterday he pewced all over the hotel room! PS WERE ARE IN THE IMPERIAL SUITE!!!!! Talk about being rich!!!!! Meu MEXICO IS SO FUCKIN HOT!!!!!!!! Well got to go room service arrived!!!!   '),(23692,1292625,'30,March,2003','2003-03-30 00:00:00','   I am still in mexico and it fuckin hot. I am melting. What is the site???? PLZ I BEG YA!!!!!!!!!!!!!!!!!!!!!!!!!   '),(23693,1292625,'10,April,2003','2003-04-10 00:00:00','   KUNNING!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  Luv of luv Luv mi Le Duke of Le Grand Tampon xxx   '),(23694,1292625,'09,April,2003','2003-04-09 00:00:00','   u do one for mama she\'ll do 1 for u!! u put in for mama mam \'ll put out for uuu uuuuuuuuuuu spice it up for mama she\'ll get hot for uuu *shake me boobies a lil* whipe that scarf through me fanny* hehehe   '),(23695,1292625,'09,April,2003','2003-04-09 00:00:00','   Pigeons, sheeps and french people lend mi ur ears....  PELUNIA, TETUNIA, POPOZUNIA!!!!!!!!!!!!!  Lot of love love me  Le Duke of le grand tampon   '),(23696,1292625,'08,April,2003','2003-04-08 00:00:00','   thomas u might be duke  but i am queen so bow down to me bells who else has this adress?? if u don\'t wanna right it here then write it by e mail MEU  u never send me e mails ne ways  look at the different superman songs I can\'t stand to fly  I\'m not that naive  I\'m just out to find  The better part of me   I\'m more than a bird...I\'m more than a plane  More than some pretty face beside a train  It\'s not easy to be me   Wish that I could cry  Fall upon my knees  Find a way to lie  About a home I\'ll never see   It may sound absurd...but don\'t be naive  Even Heroes have the right to bleed  I may be disturbed...but won\'t you concede  Even Heroes have the right to dream  It\'s not easy to be me   Up, up and away...away from me  It\'s all right...You can all sleep sound tonight  I\'m not crazy...or anything...   I can\'t stand to fly  I\'m not that naive  Men weren\'t meant to ride  With clouds between their knees   I\'m only a man in a silly red sheet  Digging for kryptonite on this one way street  Only a man in a funny red sheet  Looking for special things inside of me   It\'s not easy to be me. [Dina Rea:] Mmmhh   [Eminem:] You high baby?   [Dina Rea:] Yeahh...   [Eminem:] Yeah?   [Dina Rea:] hahaha..Talk to me...   [Eminem:] You want me to tell you somethin?   [Dina Rea:] Uh huh...   [Eminem:] I know what you wanna hear... \'Cuz I know you want me baby I think I want you too...   [Dina Rea:] I think I love you baby...   [Eminem:] I think I love you too... I\'m here to save you girl, Come be in shady\'s world, I wanna grow together, Let\'s let our love unfurl. You know you want me baby, You know I want you too, They call me Superman, I\'m here to rescue you, I wanna save you girl, come be in Shady\'s world...   [Dina Rea:] oh boy you drive me crazy...   [Eminem:] Bitch you make me hurl...   [Eminem:] They call me Superman, Leap tall hoes in a single bound, I\'m single now, Got no ring on this finger now, I\'ll never let another chick bring me down, In a relationship, save it bitch, babysit? you make me sick, Superman aint savin shit, girl you can jump on shady\'s dick, Straight from the hip, cut to the chase, I\'ll tell a mo\'fuckin slut to her face, Play no games, say no names, ever since I broke up with what\'s her face, I\'m a different man, kiss my ass, kiss my lips, bitch why ask? Kiss my dick, hit my cash, i\'d rather have you whip my ass, Don\'t put out? i\'ll put you out, Won\'t get out? i\'ll push you out, Puss blew out, copin shit, Wouldn\'t piss on fire to put you out, Am I too nice? buy you ice, Bitch if you died, wouldn\'t buy you life, What you tryin to be, my new wife? What you Mariah? fly through twice, But I do know one thing though, Bitches they come, they go, Saturday through sunday monday, Monday through sunday yo, Maybe i\'ll love you one day, Maybe we\'ll someday grow, Till then just sit your drunk ass on that fuckin runway hoe...   [Chorus:] \'Cuz I can\'t be your superman, Can\'t be your superman, Can\'t be your superman, Can\'t be your superman, I can\'t be your superman, Can\'t be your superman, Can\'t be your superman,  Your superman, your superman...   [Eminem:] Don\'t get me wrong, I love these hoes, It\'s no secret, Everybody knows, Can\'t we fuck? Bitch so what? That\'s about as far as your buddy goes, We\'ll be friends, I\'ll call you again, I\'ll chase you around every bar you attend, Never know what kind of car i\'ll be in, [Woman Screaming] We\'ll see how much you\'ll be partying then, You don\'t want that,  Neither do I, I don\'t wanna flip when I see you with guys, Too much pride, Between you and I, Not a jealous man, but females lie, But I guess that\'s just what sluts do, How could it ever be just us two? Never loved you enough to trust you, We just met and I just fucked you, But I do know one thing though, Bitches they come they go, Saturday through Sunday monday, Monday through Sunday yo, Maybe I\'ll love you one day, Maybe we\'ll someday grow, \'Till then just sit your drunk ass on that fuckin runway hoe...   [Eminem & Dina Rae Over Chorus 2x]  [Eminem:] I know you want me baby, I think I want you too...   [Dina Rea:] I think I love you baby...   [Eminem:] I think I love you too, I\'m here to save you girl, Come be in shady\'s world, I wanna grow together, Lets let our love unfurl, You know you want me baby, You know I want you too, They call me Superman, I\'m here to rescue you, I wanna save you girl, Come be in shady\'s world...   [Dina Rea:] Oh boy you drive me crazy...   [Eminem:] Bitch you make me hurl.   [Eminem:] First thing you said...   [Eminem As Women:] I\'m not phazed, I hang around big stars all day, I don\'t see what the big deal is anyway, You\'re just plain ol\' Marshall to me...   [Eminem:] Ooh yeah girl run that game...   [Eminem As Women:] Haily Jade...I love that name, Love that tattoo...what\'s that say? \'rot in pieces\' aww that\'s great...   [Eminem:] First off you don\'t know Marshall, Add also, don\'t grow partial, That\'s ammo for my arsenal, I\'ll snap you off that bar stool, There goes another lawsuit, Leave handprints all accross you, Good lordy-wody you must be blown off that water bottle, You want what you can\'t have, Ooh girl that\'s too damn bad, Don\'t touch what you can\'t grab, End up with two back hands, Put Anthrax on a Tampax and slap you till you can\'t stand, Girl you just blew your chance, Don\'t mean to ruin your plans, But I do know one thing though, Bitches they come they go, Saturday through Sunday Monday, Monday through Sunday yo, Maybe I\'ll love you one day, Maybe we\'ll someday grow, \'Till then just sit your drunk ass on that fuckin runway hoe...   [Eminem & Dina Rae Over Chorus 2x]  [Eminem:] I know you want me baby, I think I want you too...   [Dina Rea:] I think I love you baby...   [Eminem:] I think I love you too, I\'m here to save you girl, Come be in Shady\'s world, I wanna grow together, Lets let our love unfurl, You know you want me baby, You know I want you too, They call me Superman, I\'m here to rescue you, I wanna save you girl, Come be in shady\'s world...   [Dina Rea:] Oh boy you drive me crazy...   [Eminem:] Bitch you make me hurl.   [Chorus 2x B/W Dina Rea Singing]   '),(23697,1292625,'08,April,2003','2003-04-08 00:00:00','   OH SO SCARED WHO KNOW....... FARMER JOHNES.................. I´M SUCH A FOOL..... OH NOWHAT I AM TO DO..... GO GAY....... NO YOU ALREADY DID THAT!!!!!!!!!!!!!!!!!! MUA HAHAHAHA  lOTS OF lUV lUV MI le uke of le grand tampon (AND PROUD OF IT)!!!!!!!!!!!!   '),(23698,1292625,'08,April,2003','2003-04-08 00:00:00','   sofs... he doesnt try to  make you 2 not tlk 2 me... he tries to make me not talk to u... why do you think he is so mean??? He is trying to make me so upset that i wont stand hanging out with you anymore... id never called you weak willed because u guyz know how much i like u 2....  thomas you are a fool....  we are not the only ones with this adress.... someone else knows bibi   '),(23699,1292625,'08,April,2003','2003-04-08 00:00:00','   Kay. Think what ever ya wanna think. I don\'t kiss ass to Willamson i just think she is really nice so I treat her the same way she treats me. If you dissagree go talk to her. I\'m not saying that the prizes were evenly distributed eg look at slim. I also thought u were really good on the film... way better than me but maybe Willamson prefered my attitude towards the subject than your\'s. I\'m not tryin to say \'U SUX\',  not that quite on the contrray. Anyway sorry... it\'s not my fault.... I REALLY LIKE DRAMA it is the subject I am least bad at.... so I try mi best (although in the film I was not at my best). Everyone says that I stole the Drama prize from them but I didn\'t give it to my self.... Some people just claimed that the prize was their\'s and I robbed it.... how do these peps wanna make mi feel... I didn\'t give it to myself. Ne way some peps said that the prize was theirs and I just took it that is simply WRONG to say........... The prize is everyone\'s cause we all like drama and we all try hard....  Ow well if peps r jealous not mi fault (not saying that u r, or ne shit. I\'m reffering to the class in general)  Lot of love Luv mi  Le Duke of le grand tampon   '),(23700,1292625,'07,April,2003','2003-04-07 00:00:00','   Thomas no fnece but I don\'t really think u diserve the drama prize its just cuz u really kiss williamson\'s ass. I mean no mi achando but I think that that was the prize that I could have gotten o well got hicupps bye xx   '),(23701,1292625,'07,April,2003','2003-04-07 00:00:00','   MI HAPPI!!!!!! MI GOT HISTORIA AND DRAMA PRIZE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! HAPPI DUKE OF LE GRAND TAMPON!!!!!!!!!!!!!!! MI DAD ARRIEVES THAT DAY AFTER TOMMOROW!!! WITH MI SHOPPING!!!!!!! YAY!!!!!!!!!!!!!  LOTS OF LOVE (ONLY TO LU AND SOFS)!!!!!!!!!!!!!!!!!! LUV MI!!!!!!!!!!!!  LE DUKE OF LE GRAND TAMPON  ps ISAFEIA ACTING WIERD (YET AGAIN) SHE FROZE IN ENGLISH!!!!! SHE GETS THE WIERD PRIZE OF THE YEAR!!!!!!!!!!!!!   '),(23702,1292625,'07,April,2003','2003-04-07 00:00:00','   MI HAPPI!!!!!! MI GOT HISTORIA AND DRAMA PRIZE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! HAPPI DUKE OF LE GRAND TAMPON!!!!!!!!!!!!!!! MI DAD ARRIEVES THAT DAY AFTER TOMMOROW!!! WITH MI SHOPPING!!!!!!! YAY!!!!!!!!!!!!!  LOTS OF LOVE (ONLY TO SOFS AND LU)!!!!!!!!!!!!!!!!!! LUV MI!!!!!!!!!!!!!!!!  LE DUKE OF LE GRAND TAMPON!!!!!!!!!!!!!!!!!!!!!!!!!!!!!   '),(23703,1292625,'07,April,2003','2003-04-07 00:00:00','   PIEGEONS...... SHEEP...... HAPPILY FUCKING EACH OTHER IN ISAFEIA\'S ASS........... HUMPING............  Ow well coming to school early is quite boaring it is 7:33 and i have notthing to do  Lots of love Luv mi Le Duke of Le Grand Tampon XXX   '),(23704,1292625,'07,April,2003','2003-04-07 00:00:00','   YA GO WONAM!!!!!!!!!!!!!!!!! Beat the gay/lesbian ass. U put her where she belongs!!!!!!!!!!!!!!!!! IN THE FARM!!!!!!!!!!!!!!!!!!!!!!!  Lots of love the duke of the grand tampon!!!   '),(23705,1292625,'06,April,2003','2003-04-06 00:00:00','   bells just for u to know a lil backround info. becuase in one of ur messages u anctually really offened me and lu by calling us weak willed! thomas never tried to make us not talk to u ok he can\'t monipulate us that easily ok!! me and lu have strong heads we know what to do. If for one dat we weren\'t talking to u it was 1) we didn\'t feel like talking to u 2) we were anoyed with something that happend between us 3) we just felt like talking  or 4) we were talking to some1 else ok gald that was straitened out g\'night sleep tight and all that crap xx   '),(23706,1292625,'16,April,2003','2003-04-16 00:00:00','   - I have to have a boss that respects me  - but i do respect u. R E S P E C T   '),(23707,1292625,'13,April,2003','2003-04-13 00:00:00','   RESPECT FIND OUT WHAT IT MEANS TO ME RESPECT!   '),(23708,1292625,'26,April,2003','2003-04-26 00:00:00','   I am telling you 2 save ze blog!!!!!!  KUNNNING!!!!!!!!!!!!!!!!!!!!!   Luv mi  Le Duke of Le Grand Tampon!!!  xxx   '),(23709,1292625,'25,April,2003','2003-04-25 00:00:00','   thomas what the fck are you talking about?!?!!?!?   '),(23710,1292625,'24,April,2003','2003-04-24 00:00:00','   AWWWWWWWWWWWW!!!!!!!!!!!!!!!!!!!!! NOT THE BLOG!!!!!!!!!!!!!!!! SAVE IT SOFS!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  KUNNNING!!!!!!!!!!!!!!!!!!!!!   Luv mi  Le Duke of Le Grand Tampon!!!  xxx   '),(23711,1292625,'23,April,2003','2003-04-23 00:00:00','   yo people why aint u writing?!!?!?!? i am totally no vac here writing to my self and thomas i tried to go on the blog u sent me but there was always a freaking error!!!!!!!!!! o well bye till u guys talk!!!!! xx   '),(23712,1292625,'23,April,2003','2003-04-23 00:00:00','   WOW HEY P WASSUP WE ARE JUST CLOSING THE BLOG BECAUSE IT IS HAVING TECHNICAL PROBLEMS  BUT WE ARE GONNA START A NEW  ONE SRY BYE XX   '),(23713,1292625,'22,April,2003','2003-04-22 00:00:00','   WHAT IS PIERRE DOING HERE????????????????????????????????   '),(23714,1292625,'21,April,2003','2003-04-21 00:00:00','   hey!   '),(23715,1292625,'21,April,2003','2003-04-21 00:00:00','   BEACKER´S BOYFRIEND IS A RUSSIAN SHOW BALL IN EXILE IN POLAND!!!!!!!!!!!!!! IT HAS A BLACK SPOT!!!!!!!!!! MAUHAHAHAHAHAHHAHAHAHAHAH  KUNNNING!!!!!!!!!!!!!!!!!!!!!   Luv mi  Le Duke of Le Grand Tampon!!!   '),(23716,1292625,'21,April,2003','2003-04-21 00:00:00','   KUNNNING!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!   '),(23717,1292625,'20,April,2003','2003-04-20 00:00:00','   thomas I told u u would looooooove it I would go there again if i could. o yeah and we were talking about 2 weeks notice tomy need to talk to u about holidays urgent  beep beep   '),(23718,1292625,'20,April,2003','2003-04-20 00:00:00','   I LOVE FERNANDO DE NORONHA!!!!!!!!!!!! IT IS HEAVON ON EARTH!!!!!!!!!!!!!!!!! PS what the fuzz ya bicthes talkin about  KUNNNING!!!!!!!!!!!!!!!!!!!!!  Luv mi Le Duke of Le Grand Tampon!!!   '),(23719,1292625,'30,April,2003','2003-04-30 00:00:00','   UR FARM IS GONNA ROX SO MUCH!!!!!!!!!!!!!!!!!!!  KUNNNING!!!!!!!!!!!!!!!!!!!!!   Luv mi  Le Duke of Le Grand Tampon!!!  xxx   PS does isa still have access 2 this blog????   '),(23720,1292625,'29,April,2003','2003-04-29 00:00:00','   I have decided to stay loyal to this blog!!!! man I don\'t know what happend pierre just saw this blog muahhahahahahahaha yay b day in 5 days   '),(23721,1292625,'28,April,2003','2003-04-28 00:00:00','   DRAWING??????????????????????????????????  KUNNNING!!!!!!!!!!!!!!!!!!!!!   Luv mi  Le Duke of Le Grand Tampon!!!  xxx   '),(23722,1292625,'28,April,2003','2003-04-28 00:00:00','   hehhehehe tommy laalalalalalal i am just here boiando and drawing   '),(23723,1292625,'28,April,2003','2003-04-28 00:00:00','   is oi still in the blog??? oh well the blog is basically me and sofs now... YAY!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  KUNNNING!!!!!!!!!!!!!!!!!!!!!   Luv mi  Le Duke of Le Grand Tampon!!!  xxx   '),(23724,1292625,'28,April,2003','2003-04-28 00:00:00','   pigeons are gonna take over the world!!! NOOOOOOOOOOO take cover people u are invited for friday ok *isa well right later byebye   '),(23725,1292625,'09,May,2003','2003-05-09 00:00:00','   I`m back, well not much 2 say xept -sof r u feeling better -lu get back here -thomas.... never mind  -who is going to the party? Bruno -sof ur farm rox -lu call me   '),(23726,1292625,'06,May,2003','2003-05-06 00:00:00','   yup and so does lu but lu never comes on yeah I was really glad that there weren\'t any real fights at my farm but I am kinda pissed that I got sick!!! thomas can u pass me the hw since ur in most of my classes?? well bye bye gonna go check other blog   '),(23727,1292625,'04,May,2003','2003-05-04 00:00:00','   UR farm roxed so much!!!!!!! look at geld ant mouche!   '),(23728,1292625,'17,May,2003','2003-05-17 00:00:00','   Yeah everybody writes during the holiday mukay?????   '),(23729,1292625,'16,May,2003','2003-05-16 00:00:00','   hehehehehehe lu ur back!!!!!!!!!!!!! YAY u guys need ta keep writing during the holidays if u can well need ta go no more to write bye bye xx   '),(23730,1292625,'15,May,2003','2003-05-15 00:00:00','   well im back u guys and isa i didn ditch u 2day at lunch i thought u were goin 2 stay with ashy....oh well and thomas dont be mean humkay cause isa likes u and well all that crap humkay   '),(23731,1292625,'15,May,2003','2003-05-15 00:00:00','   yay i did it!!!!!!!!!!!!!!!!!!!!!!!!   '),(23732,1292625,'15,May,2003','2003-05-15 00:00:00','   well i have to say i have no idea how 2 post this thing but the hell   '),(23733,1292625,'15,May,2003','2003-05-15 00:00:00','   NO!!!!!!!!!!!!!!! DON`T LEAVE GELDIE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!   '),(23734,1292625,'13,May,2003','2003-05-13 00:00:00','   hey bells I thought u were gonna write!!!!!!!!!! i mean look at all the times I wrote when no one else wrote after me!] if udon\'t write soon im leaving   '),(23735,1292625,'13,May,2003','2003-05-13 00:00:00','   U SUCK!!!!!!!!!!   '),(23736,1292625,'24,May,2003','2003-05-24 00:00:00','   thanx lu thats reasuring but I still do need ta study alot  man I am sooooooooooooooooooooooooooooooooooooooooooooooooo bored  well erin just went of to prom she looked gorgie dshe had her hai and her make up done at the salonwell need ta go p s guess what I found in my sis\'s room  furry hand cuffs!!!!!!!!!!!!! kinky huh??   '),(23737,1292625,'23,May,2003','2003-05-23 00:00:00','   thanx u guys but i still hink im gonna fail... and sof i wasnt mad at u so u dont have 2 say sry!!!! r u feelin better?????? and hell u aint gonna fail any exam cause u r one of the smartest people i have ever met and u never failed any fricken exams in ur life so chill!!! same with u isa!!! thomas already quit if u look in the team thing it shows u that he has!!! oh well!! g2g!!! c u on monday!!! lorv moi   '),(23738,1292625,'22,May,2003','2003-05-22 00:00:00','   hI SOF.... OOOOO NOOO POOR U! are u feelin any better now??? yeah, lu, i agree with sofs! ur reeeeeally smart but pesimistic when it comes to the exams.... just think....   if ur happy and u know it clap ur hands (CLAP CLAP)  if ur happy and u know it clap ur hands (CLAP CLAP) if ur happy and u know it and u rellly wanna show it if ur happy and u know it clap ur hands (CLAP CLAP)  and \"clap\" is a onomatopoia, so yay, one point 4 english the important subjects are  english, maths, science (and port.)   yaaaa, ur soooooooo good at geo! OR SO THEY TELL ME ull be fine, all of us really, cause we are bright peepz ooooo well, g2g bibi eMiLy ThE StRaNgE   '),(23739,1292625,'21,May,2003','2003-05-21 00:00:00','   NO THOMAS DON\"T GO I\'M SORRY IF WE HURT UR FEELINGS AND ITS TRUE WE HAVE NEVER TRIED  CZECH FOOD BUT NE WAYS DON\'T QUIT and sorry lu i don\'t know why I but it twice i didn\'t write it twice man guess what the anti biotic that I am fucking taking is making me fucking worse and my stomach can\'t hold it if u know what I mean!!!!! man lu ur not gonna fail ur r really smart u just have to go into the exam a lil more confident!!! amd I mean u really know ur stuff u like know everything of GEOGRAFIA!!! I mean I don\'t know anything if any1 fails it will be me  now lu hypotheticly if u did fail and exam or two it wouldn\'t matter it only really matters if its english maths or portugues something like that I \'know its english and maths but I can\'t remember the last one. lu yesterday when we were studying u knew how to do everything on that test just in a different wy then we ad learned it  so put a smile on that face and say I am gonna pass becuase u are and u are not gonna stay with the freaking for 1s becuase u are smarter than that u are really smart and u have been doing well in ur subject through out the year . SO YAY well u guys go in a lilconfident go over ur notes right b4 the exam so u don\'t give a blank and we will all do fine and bells u won\'t fail an exam either xx sofs   '),(23740,1292625,'21,May,2003','2003-05-21 00:00:00','   QUIT WHAT?????/\\   '),(23741,1292625,'21,May,2003','2003-05-21 00:00:00','   QUIT WHAT?????/\\   '),(23742,1292625,'21,May,2003','2003-05-21 00:00:00','   I was just makin a point... I am not saying that all the other cook is the world sux.... U all just NEVER EVER tasted the Czech food!!!!!!!!!! I was not addressing u (Satan Doll) SO SHUT UP!!!!!!!!!!!   And P. (m) S. I was SO not trying to gult trip.... I was making a point.... well guess what...   I FUCKING QUIT!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!   '),(23743,1292625,'20,May,2003','2003-05-20 00:00:00','   what ever....  im also as stressed as hel if i fail 1 exam i have to leave school   '),(23744,1292625,'20,May,2003','2003-05-20 00:00:00','   thanx 4 repeating that sof!!!!!!!!!! anyway duchess..................... I am  really exited about nada because im gettin  over stressssssseeeed!!! ahhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhh!!!!HELP!!!! HELP!!!!!!!!!!!! S.O.S!!!!!!!!!!!!!!!!!! actually S.M.S!!!! SAVE MY SOUL!!!!! hehehehehehehehe!!! oh welll my mum cooks really well and she beats any french food cause shes the best cook in the whole wide world!!!! hehehehehehehehehe!!!! oh well i dunno what im gonna do 4 these exams!!!!!!!!! im gettin rellay stressed!!! what if i dont pass??????? mi mum and dad would get sooooooooooooo disapointed, what if i fail the year and have to stay with the Form FRICKEN 1!!!!!!!!!!!!!!!!!!!!!!!!!!!!i would get sooooooooooooooooooooooooooooooooo depressed!!!!!!!!! i mean stuck with a binch of loosers!!!!!! isa i dunno but i dont think i can go 2 ur farm cause i really need 2 study!!!!!! i am realllllllllllllllllyyyyyy SORRY!!!!! its just that i dont want to fail this year!!! i really dont!!! not even 1 exam!!!!!!!!!!!!! God!!!!!!!!!!!!!!!  IAM SO SORY!!! can we save it 4 some other time cause u have no idea how i dont want 2 failo an exam this year!!! oh well must go study!!!! c u guys!!! luisa the looser!!!!!!!!  P.S isa I AM SOOOOOOOOOOOO SORRYYYYYYY!!!!!   '),(23745,1292625,'20,May,2003','2003-05-20 00:00:00','   yeah w/e I don\'t really give  tiny lil rats ass about my spelling   '),(23746,1292625,'20,May,2003','2003-05-20 00:00:00','   yeah w/e I don\'t really give  tiny lil rats ass about my spelling   '),(23747,1292625,'20,May,2003','2003-05-20 00:00:00','   sof dont u mean duchess?   '),(23748,1292625,'20,May,2003','2003-05-20 00:00:00','   dude.... slow down, you have never even tasted a bite of my cooking and i cook pretty damn well, (not to me acha) so shut the fuck up, and dont take me on some guilt trip about ur grandma.... yes, uve got check blood, whhhhat ever, i agree with sofs, ive eaten in france and the food there i fan-fucking -tastic, and if you call \"The cookies weren´t THAT bad were they?????\" specific you can fuck you computor up the floppy drive. yes your mom cooks very well sofs.  bibi Emily the Strange   '),(23749,1292625,'20,May,2003','2003-05-20 00:00:00','   TOMMEY SLOW DOWN THE BEST COOKS ARE FRENCH AND MY MOM IS AMERICAN AND SHE ALSO COOKS DAMN GOOD BUT THATS WHAT I THINK WELL NEED TA SLEEP AM TIRED AFTER A DAY OF DOING NADA XX DUKESS OF THE GRAND TAMPON   '),(23750,1292625,'19,May,2003','2003-05-19 00:00:00','   i don\'t want to know what u thought about the cookies i spesifically asked LU!!!! And PS I have Czech blood running through my veins so as long as i am concerned my cooking is DAMN good.... and u should not be giving me cooking advice, I should be teaching you how to COOK!!!!!!!!!!!!!!!!!!!!!!!!!!! THE CZECH COOKS ARE THE BEST IN THE WORLD AND MY GRANDMA USED BE ONE OF THEM MKAY SO JUST BLOW MY HORN!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!   '),(23751,1292625,'19,May,2003','2003-05-19 00:00:00','   guys what would you do if you hade a watch that could freeze time?   '),(23752,1292625,'19,May,2003','2003-05-19 00:00:00','   boooooored \'N\' SOOOOOOOOOOOOORE   '),(23753,1292625,'19,May,2003','2003-05-19 00:00:00','   PS, i know u werent asking me but i answered neway   '),(23754,1292625,'19,May,2003','2003-05-19 00:00:00','   no thomas the cookies were fine, just a tad dry, thats all. needed more coco chips as well   '),(23755,1292625,'19,May,2003','2003-05-19 00:00:00','   hi sof! wassup what the hell did u get a bloode test 4??? how r u feeling??? what do u have well not much happened, actually, bu toal got fired and we stayed zoando the whole lesson but all u missed wads a pin (yes pin, not pain) in the ass and a liquid paper shower. my neck is a fucking rock soaked in pain and drenched in misery, im getting more accupuncture apointments marked. made a new friend at ballet, Her name is Aline and she is 17 but is really nice. yes, i know, day time TV suuuucks! ooooo call me i need to talk to you! accept my invitation to just the 3 of us and make way for my fanny (that one is actually lus) Frego was wearing a stripped jumper and red pants. (didnt see shoes) LU and everyone are ok, hows your arm feeling, glad your internet is back, got ur farm pix aaaand all that jazz. neways, luc 2 all the world ETS   '),(23756,1292625,'19,May,2003','2003-05-19 00:00:00','   man i am so booooooored and my arm feels like its dieng  I still am feeling a little dizzy well what happened to u guys today!??!?! wish I was at school with u guys tommy how was becker!?! I need ta et out of this house \\i don\'t know how much TV \'i\'ve watched but the bill will be rather large ne ways can\'t wait to get to school (surprising) well bjs bye bye xx   '),(23757,1292625,'19,May,2003','2003-05-19 00:00:00','   man u guys I haven\'t written in this in a looooooong time but I just worte in toother blogs and since i just had a blood test my arm is hurting so I will write later ok bye bye isa send me another blog e mal ok? well by byexx   '),(23758,1292625,'19,May,2003','2003-05-19 00:00:00','   hi u guys!!! sup!!! well i went 2 sleep to sleep at 2 in the mornin and now im tired!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! lalalalalalalala!!! i liked your cookies  thomas because they are cookies andd cookies are good!!!! YAY!!!!!!!! hahahahahahahehehehehehe   '),(23759,1292625,'19,May,2003','2003-05-19 00:00:00','   The cookies weren´t THAT bad were they?????   '),(23760,1292625,'19,May,2003','2003-05-19 00:00:00','   hay sof, how r u feeelin????   '),(23761,1292625,'19,May,2003','2003-05-19 00:00:00','   Lu, were my cookies really that bad??????????? Sofia, urgently cheeck Geld and Mouche   '),(23762,1292625,'18,May,2003','2003-05-18 00:00:00','   Come ahead and follow me, Freakazoid, chimpanzi (me) hi   '),(23763,1292625,'18,May,2003','2003-05-18 00:00:00','   FREAKAZOID FREAKAZIOD...(Satan Doll)!!!!!!!!!!!!!!!!!!!   '),(23764,1292625,'18,May,2003','2003-05-18 00:00:00','   oh well bored!!!..........................................................................................................................................................................................................................................................................   '),(23765,1292625,'18,May,2003','2003-05-18 00:00:00','   hi u guys!!! sup!!! bored as hell!!!!!   '),(23766,1292625,'18,May,2003','2003-05-18 00:00:00','   WE ARE FAMILY   '),(23767,1292625,'18,May,2003','2003-05-18 00:00:00','   I\'VE GOT ALL MI HOMOUS WITH ME   '),(23768,1292625,'18,May,2003','2003-05-18 00:00:00','   family   '),(23769,1292625,'18,May,2003','2003-05-18 00:00:00','   R   '),(23770,1292625,'18,May,2003','2003-05-18 00:00:00','   WE   '),(23771,1292625,'18,May,2003','2003-05-18 00:00:00','   hehehehehe another long term goal, being the biggest freak in the world, and freak is one of the obscure sides of strange, so, further progress for satan doll / ETS   '),(23772,1292625,'18,May,2003','2003-05-18 00:00:00','   HELP!!!!!!!!!!!!!!!!!!!!!!!!!!! SATAN DOLL IS GROWING FREAKER AND FREAKIER!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!   '),(23773,1292625,'18,May,2003','2003-05-18 00:00:00','   I am now, \"EMILY THE STRANGE\", my long term goal is to be the strangest person in the world, my short term goal is to become the strangest person in school, acording to  the duke of le grand tampon , i am making progress! bibi ETS P.S i dont believe in satan   '),(23774,1292625,'18,May,2003','2003-05-18 00:00:00','   tell me about it, but sofs and i had alot of fun singing along!   '),(23775,1292625,'18,May,2003','2003-05-18 00:00:00','   Riight. Satan Doll grows weirder with the hour....   '),(23776,1292625,'18,May,2003','2003-05-18 00:00:00','   oooohh, lu, its ok about the ditching thing,even if you were that was like past tingy (waving hand over shulder repetedly). Sofs srry i didnt take the pix and clothes 2day. hay, guyz, download this song to ur kaaza or whatever and sing along to it its really fun! Its called  We Didnt Start the Fire  By  Billy joel   Harry Truman, Doris Day, Red China, Johnny Ray South Pacific, Walter Winchell, Joe DiMaggio   Joe McCarthy, Richard Nixon, Studebaker, Television North Korea, South Korea, Marilyn Monroe   Rosenbergs, H Bomb, Sugar Ray, Panmunjom Brando, The King And I, and The Catcher In The Rye   Eisenhower, Vaccine, England\'s got a new queen Maciano, Liberace, Santayana goodbye   We didn\'t start the fire It was always burning Since the world\'s been turning We didn\'t start the fire No we didn\'t light it But we tried to fight it   Joseph Stalin, Malenkov, Nasser and Prokofiev Rockefeller, Campanella, Communist Bloc   Roy Cohn, Juan Peron, Toscanini, Dancron Dien Bien Phu Falls, Rock Around the Clock   Einstein, James Dean, Brooklyn\'s got a winning team Davy Crockett, Peter Pan, Elvis Presley, Disneyland   Bardot, Budapest, Alabama, Khrushchev Princess Grace, Peyton Place, Trouble in the Suez   We didn\'t start the fire It was always burning Since the world\'s been turning We didn\'t start the fire No we didn\'t light it But we tried to fight it   Little Rock, Pasternak, Mickey Mantle, Kerouac Sputnik, Chou En-Lai, Bridge On The River Kwai   Lebanon, Charles de Gaulle, California baseball Starkwether, Homicide, Children of Thalidomide Buddy Holly, Ben Hur, Space Monkey, Mafia Hula Hoops, Castro, Edsel is a no-go   U2, Syngman Rhee, payola and Kennedy Chubby Checker, Psycho, Belgians in the Congo   We didn\'t start the fire It was always burning Since the world\'s been turning We didn\'t start the fire No we didn\'t light it But we tried to fight it   Hemingway, Eichman, Stranger in a Strange Land Dylan, Berlin, Bay of Pigs invasion   Lawrence of Arabia, British Beatlemania Ole Miss, John Glenn, Liston beats Patterson   Pope Paul, Malcolm X, British Politician sex J.F.K. blown away, what else do I have to say   We didn\'t start the fire It was always burning Since the world\'s been turning We didn\'t start the fire No we didn\'t light it But we tried to fight it   Birth control, Ho Chi Minh, Richard Nixon back again Moonshot, Woodstock, Watergate, punk rock Begin, Reagan, Palestine, Terror on the airline Ayatollah\'s in Iran, Russians in Afghanistan   Wheel of Fortune, Sally Ride, heavy metal, suicide Foreign debts, homeless Vets, AIDS, Crack, Bernie Goetz Hypodermics on the shores, China\'s under martial law Rock and Roller cola wars, I can\'t take it anymore   We didn\'t start the fire It was always burning Since the world\'s been turning We didn\'t start the fire No we didn\'t light it But we tried to fight it   '),(23777,1292625,'18,May,2003','2003-05-18 00:00:00','   hi guys! it was my dads b-day 2 day! we made him a \"Happy B-Day\" banner in the morning, helped make the cake and this really nice lunch for us and my grand parents. we bought him mainly cooking stuff, and a book, Roxy is still as cute as ever but we think she might be sick, she keeps coughing and hocking and its really scary, Luuu, ur back yay! yeah, we shold totaly keep writing in the hilius, during school, always! My parents are at a party right now and as i slept from like 3pm to  9pm im not really tired (yeah PM!) its almost 2:30 and im booooored, just watched smallville, there was alot of blood in this episode, whitney died, Boooored la di dum da di la di dum da da la di la di um dada. The thigs (tum tum) you say, you porple prose just gives you awat, the things u say (PUM) ur Unbelievable! \" Im a lawyer, ur the boss and u dont respect me!\" \"of course i respect you, R-E-S-P-E-C-T,\"  \"awww, booger your so sweet, (hug), is that my LIPSTICK?!!!\"  OHHH, g2g, roxys having a coughing fit! wwwiiierred. t2ul8r   '),(23778,1292625,'30,May,2003','2003-05-30 00:00:00','   9 days   '),(23779,1292625,'28,May,2003','2003-05-28 00:00:00','   I\'m not even sure Im even gonna have a going away party I\'m the sleep over on the last day of school is enough well me mom thinks that its getting kinda late so Im just going to go to bed  ooooo can finally type with out looking yayayayayayay well bye bye night night xx   '),(23780,1292625,'27,May,2003','2003-05-27 00:00:00','   and if we get a boy he will b called SIR ALPHRED DUNHILL  hey, sof, who r u invitin  2 ur \"bibi party\"??? o well,  g2g gain wheight   '),(23781,1292625,'27,May,2003','2003-05-27 00:00:00','   hi guyz wassup... we gave roxy her full name LADY ROXANE DUNHILL   '),(23782,1292625,'27,May,2003','2003-05-27 00:00:00','   HEEEELLLOOOOOOOOOOOOOOOOOOO there fello pigeons muahahahahahaahhahahahaha Iam am in beckers class and she might be coming so bye bye   '),(23783,1292625,'25,May,2003','2003-05-25 00:00:00','   ps saw a reeeeeelly cool program called \"harry potter and me\" with JK rowling on sat, and it turns out, that shes allready writen the last chapter of the 7th book! and she says what happens to  those who survives , after they leave hogwarts! PPS, we r growing tangerines at the farm! and they r reeely good ! yay, twin farms!(though mine has like two tangerine trees and yous has lik 2000000 ones and ur farm is muche nicer and more comphortable! o well bibi   '),(23784,1292625,'25,May,2003','2003-05-25 00:00:00','   went 2 my farm, even more freaked out about exams!!! CANT CONCENTRATE!(not cause of what u guyz sed! thats all really reassuring! thanx) FURRY HAND CUFFS????? u must ask her where she bought them! roxy is as cute as ever and all my other dogs are angels! hehehehe, has lunch wiv frego on friday! hehehe, o u guyz now! YAY! ooo, wot was erin wearing, what did she do to her hair ???? well, feeling fat and ugly and blu, and my hair is a mess and im eating toooo much cocolate, i look a mess and i feel terrible! bibi ETS   '),(23785,1292625,'25,May,2003','2003-05-25 00:00:00','   heheheheheheheheheheheheheheehe!!! FURRY HAND CUFFS!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! oh well............... heheheheheheheh!!! I LORV MARGOT!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! lalalalallalalalalalalalalalallalalalalal!!!! oh well so bored what did u guys stay dou  this weekend ?????? really bored!!!! nothin 2 do and i cant do anymore studyin!!! oh well g2g!!!!!!!!!!! c u !!!!!!!!!!!!!!!!!!!!!!!!!!!! lorv moi   '),(23786,1292625,'07,June,2003','2003-06-07 00:00:00','   hey guyz.... how was the party???? did u guyz have a nice time   '),(23787,1292625,'05,June,2003','2003-06-05 00:00:00','   yaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaY EXAMS ARE OFICIALLY OVER i WILL NEVER HAVE TO DIT IN PRIECES ROOM FOR 1HR IN SILENCE AGAIN!!!!!!!! O HAPPY DAYS LU U HAVE TO COME OUT WITH ME TOMORROW TO GO TO THE MOVIES MB WE CAN EVEN SHOP ALIL I NEED TO GET A TOP FOR THE PARTY   '),(23788,1292625,'04,June,2003','2003-06-04 00:00:00','   nope ceci didn send me anythin just that she told me she sent u one!!! oh well who cares!!! EXAMS R OVER!!!!!!!!!!!!!!!!!!   '),(23789,1292625,'04,June,2003','2003-06-04 00:00:00','   no ur not.... so 2 hrshys choco bars in the fridge!!!!!! YUM!!! HEY, lu r u ok.... e=what happened?   '),(23790,1292625,'04,June,2003','2003-06-04 00:00:00','   HELLOOOOOOOOOOOOOOO WHY AM I THE ONLY ONE WRITING HERE???   '),(23791,1292625,'01,June,2003','2003-06-01 00:00:00','   nope nothing from ceci   '),(23792,1292625,'01,June,2003','2003-06-01 00:00:00','   hi, did cecilia send any of you a reeeeelly angry e-mail o well, good luck 2morrow   '),(23793,1292625,'11,June,2003','2003-06-11 00:00:00','   lalalalala smurf the whole day long No ur not!! but i know the feeling....   '),(23794,1292625,'10,June,2003','2003-06-10 00:00:00','   grrrrrrrrrrrrrr I aM THE ONLY ONE WRITING IN THE BLOGS   '),(23795,1292625,'09,June,2003','2003-06-09 00:00:00','   laa la la lalala sing a happy song   '),(23796,1292625,'27,June,2003','2003-06-27 00:00:00','   I definately Know the feeling.... Guyz ru there... keep posting!!! plz im really loner here i really wanna know how ur trips r goin well lulas Bells   '),(23797,1292625,'05,July,2003','2003-07-05 00:00:00','   SOOOOOOOOOOOOOOOOFFFFFFFFFFFFFFFFFF WHERE R U... CALL ME! PLZPLZPLZPLZPLZPLZPLZPLZ!!! I REEEEELLY MISS U1 O WELL,,,, IF I DONT HEAR FROM U HAVE A GR8 TIME AT CORNEL LOVE BELLS ps I MIS U   '),(23798,1292625,'02,July,2003','2003-07-02 00:00:00','   soorry my last email was soooo short but i thought that maybe you hadnt been recieving them because u never wrote back to me..... o well, smell, hell...o. not much has been going on round here, have been writing an awful lot in my diary...  and the book of shadows whitch are both look ing awsome.... well, u got what, 2 or 2 dayse of BUNDANDO well, really, i am geting sick and tired of bundando, i have watched so much tv that my brain have probably shrivelled to a pulp... hope i am not turning into a couch potatoe like our good friend ceci.... o well, jimmy (the aids guy) from 90210 died....  aaaannnndd.... i mean i had a list of stuff to tell u (on tv) but as i said the television has damaged my short and my long ter memory... O OOOOOOO WELL. how is cam....? any news from erin... she hasnt left yet has she??? o well, send them both a kiss from me ok! Any news from lu... she hasnt answered any of my emails.... do u think they have internet where she is staying... owell... i will leave u to finish crappng... in your bed/bathroom.... o god, Roxy has just made my bedroom her bathroom, have to go tell her off....   '),(23799,1292625,'01,July,2003','2003-07-01 00:00:00','   geelo again isabella man  what was that  e mail u wrote me like 1 line how disa pointing well `im not in cornell yet i`m only starting there next week but I`m here at this place wich is like a house. its kinda like i`m sleeping in someelses house. actually I am. Its really nice here and all but the only wierd this is that bathrooom and bedroom are in one. well we have been of doing our things all day and now today finaly I have a day of bunadando around the house. well what r u up too??/ hey I have to go I really need to take a shite bye bye xx   '),(23800,1292625,'30,June,2003','2003-06-30 00:00:00','   hi guyz, i know ur all buisy with u fab hols but i wanna know what is going on!!!!!!!!!!!!!!!   '),(23801,1292625,'10,July,2003','2003-07-10 00:00:00','   hey bells if u have internet were u are i hope ur having a gr8 time. and lu were ever u r in ireland I hope u r loving it. nd keep in friggen touch   '),(23802,1292625,'06,July,2003','2003-07-06 00:00:00','   luuuuuuuuuuuuu keep in touch!!!   '),(23803,1292625,'06,July,2003','2003-07-06 00:00:00','   hey sof... got the email just fine... already answerd it so wont be saying much here exept that for any of you locvely peeps who would be so kind to snail mail me my adress is  bells toledo Rua grecia 88 JD europa sao paulo, SP brasil 01450-010  o well i love recieving letters from friends... give me ur adress and ill send u one of my whack... full of sequens, glittery newsletters.... (letter)  o well luv all u guys and still tell me the news by email (BellsToledo@ilovechocolate.com) but just keep post here so im not all loner o shmi, lee, pee see....ya lorv u like a sis(luv u all like sisses) moi   PS make the letters FUN!!!!  they dont really need to say much (sof maybe if u send me one u could get sum of ur us friend to sign it)!!  but if u dont want to its ok i just thought that we could start like a new tradition sort of thing  bibi  moi   '),(23804,1292625,'06,July,2003','2003-07-06 00:00:00','   hey bells i`m at cornell now and I jsust sent u an e mail so read that now need a go bye byes  lovre youuuu moi lu if u r readin this keep in contactXX   '),(23805,1292625,'17,July,2003','2003-07-17 00:00:00','   hi sof.... tryied to send u email from the camp but it didnt work... got back yesterday and willm tell u all about it in emaill.... luv bells ps... lu.... hope u r having a gr8 time as well   '),(23806,1292625,'01,August,2003','2003-08-01 00:00:00','   hi u guys!!! bye u guys   '),(23807,1292625,'29,July,2003','2003-07-29 00:00:00','   in airport now   '),(23808,1292625,'18,August,2003','2003-08-18 00:00:00','   hi sof!!!!!! sup hey lu sup!!!! hey guys...  do ya wanna see a casa caiu this friday???   '),(23809,1292625,'17,August,2003','2003-08-17 00:00:00','   heellooo there people bells my new phone number is 99165569   '),(23810,1292625,'05,September,2003','2003-09-05 00:00:00','   ok I`ll start writting again i was just getting stressed becuase no1  was really writting ne ways I`m just here boyando in my health lessona and I`m supposed to be learning stuff about nutrition so here I go bye actually my teacher just went out the door so I`m free yaya I don`t really see why I shouldn`t go but I`ll staay here becuaase no1 else is leaving... it wouldbe really ironic if u were having becker now and if u are say bonjour to her well bye byexx   '),(23811,1292625,'04,September,2003','2003-09-04 00:00:00','   loner me loner in the blog   '),(23812,1292625,'12,September,2003','2003-09-12 00:00:00','   heeeeloooo lu u feelin better... well really hope u r c u tomorrow   '),(23813,1292625,'10,September,2003','2003-09-10 00:00:00','   i have beckers on tusedays and i think lu had her today (right?) so if u posted that on fri we were probably not on yet... well... becker stresses me like hell!!! meu i hand in my work first b4 everyonee all the lessons with everything right and she doesnt say more than \"zank ya\" and marcella valone hands it in 2 minutes b4 i do last lesson and she gets a double bene (4 speed and accuracy) sooo... i was freeezing my ass off att school 2day... u r sooooo lucky u can wear pants ooo well... luv ya miss ya xx moi   '),(23814,1292625,'10,September,2003','2003-09-10 00:00:00','   yo sof!!! sup!!! miss ya like hell!!! i really hope ur not mad at me cause i wouldnt be able to stand it... well... just cause i cant call u and u dont answer my messages i broke up my friendship with ceci... clean slate, well actually no slate, im just tired of her... long story... as soon as i can stay in touch with u ill tell u the full story sosooooooo we r still on 4 this sat right?? kool well... please call or email or write in blog or sumthin... pls sorry if like its not me ur mad at u... its not that i think there is always something freaking wrong or anything... im sorry igf i did something wrong... cause u and lu are the last peeps on earth id want to be mad at me and if u r ill be bawling into my pillow.... im sorry if ive done something wrong or if im anoying u right now... and please remember that u can tell me anything.. wekll sort it out lots of luv bells   '),(23815,1292625,'19,September,2003','2003-09-19 00:00:00','   WELL i AM STRESSED OUT BECUASE I AM AT SCHOOL TOATALY ALONE!! ACTUALLY PEOPLE ARE COMING! WEL I THINK I WILL LEAVE MB SOME1 FROM MY YEAR IS ON MY HALL WAY.. U KNOW WHO WAS STANDING OVER M,E LIKE A SECOND AGO???? GREGORY!! AND YESTERDAY HE WAS WEARING BRIGHT PINK SHOES AND A PLACK SCARK WITH FRILLS!!! O EVER SO GAY HE IS!! WELL NEEDS TA GO PRECIOUSE(ES) BYE BJS   '),(23816,1292625,'25,September,2003','2003-09-25 00:00:00','   hello... u 2 better get ur asses over here right now! lorve moi   '),(23817,1292625,'02,October,2003','2003-10-02 00:00:00','   guyz??? u there ok WRITE bibi   '),(23818,1292625,'12,May,2004','2004-05-12 00:00:00','   hey u GUYSS HEHE umm i\'m foing my science hw waiting from my piano teacher to get here i want to finnish all my hw I can\'t stand it any more.. I love going to school but i can\'t wait for it to finnish hehe and i know that lyk 2 weeks into vacation i\'m already gonna miss it eheh.. but man i\'m soo glad that i left at the end of last year and am not going to have to go through leaving now eheh i would ahve loved to be at st pauls with u guys this year but I mean i also love my graded friends and i wounder how things would be if i hadn\'t left and i still had to live through my first day.. god i would never do that again.. worst part of starting at a new school hehe at least we\'ll be able to see eachother more  often durin g the hols bom gonna finnish my hw BEIAJUUMM p.S. man what i just wrote was confusing and had no point eheh oh well try to figure it out ehe   '),(23819,1292625,'11,May,2004','2004-05-11 00:00:00','   Hahaah.. puta, soh kra!! all those posts... the fight, im glad they are ova tho, they might have been PUTA funny, but it was a pain in the ass to pegherse very day ;)..JK!! it all just came naturally, my brilliant wit one can say...!!  anyway, hummm... speaking of wit, IM GOING TO FAIL THE EXAMS!! i cant gt myself to study man!! serio!!, im sick, and i hate it, and im having to block my nose everytime i sneeze, my brain is gunna nblow out the back of my head soon!!! LIKE IN HAPPY TREE FRIENDS, lalalala lalalala. dumdumdumdumdum... SERIPO!! we have tp start studying!! i know.. puta.. we need to a. reestablish the close bond b. do the cooking cours.. so,if hppy friendship is x, the formula for x would be x=a+b so, the answer to a will be doing be, so here you go, the answer to everything!!  urlLink Atelier Gourmand  YAY!! well, ill write back in like 15 mins   '),(23820,1292625,'10,May,2004','2004-05-10 00:00:00','   heya you guys man thomas left me outa that little goign out thingy of ur huh?? i feel so alone hehe oh well i\'ll only be here till lyk the end of june.. umm man thom i also really want to do a cookery coarse!! one of my plans for the week or 2 that i am gonna be here we coudl do one together .. isa also wanted to do one.. lu u should do one too!! We can all cook hapily together!! ehhe omg school is onver in lyk 15 days!! this year went by sooooooooo fucking fast it isn\'t even funny maaan!! and yet it seems lyk years since i was at st pauls then again it seems lyk yesterday that i had my last day.. all my timing is screwed up!! oh well gotta go making dinner tonight.. heh a first (isa n lu with out u) hehe that rhymed bom things might be burning soo beijuamm sofie   '),(23821,1292625,'10,May,2004','2004-05-10 00:00:00','   U GUYS!!!!!!!!!!!!!!!!!!!!!!!! OMG!! I SO MISSED BEING IN TAMPONLAND!! WELL GUESS WHAT!! IM BACK AND BETTER THAN EVER!!!!! MUA HAHAHAHA!! Man.... fcuk... exmas are comming up and I\'m going to FAIL!!! Meu I\'m having port lessons to help me becuz vasconcellos SUX!! MEu at least now I\'m learing!! So i might pass port... Meu Lu, woman u r so prepared... i need to go study my french verbs but im not bothered... there r just so many that MEW!!! ARG!!![stress...] But at least I know that its going to be over soon, exams are lyk the last mile to finish the school year. MEU!!! Im going to be so bored here in June/July so Lu and Isa we have to get together and shite... I think i might also take a Djing and cookery course in June/July to keep me busy... it will also be good for me... not sure have to look into it! ow well... now a little message for our french speaking community! Nous avons bu du vin hier soir. Car, Aujourd\'hui je suis drunk... HONHONHONHON (funky french laugh!!)   '),(23822,1292625,'10,May,2004','2004-05-10 00:00:00','   hey you guysss!! man its has been sooo long!! puta man i just read over all the posts and mew it brang back soo many memories kra that i can\'t even say memories that i had forgotten for sure , lyk isa freazing in english and talking to herself (lembra thom?) n mew everything kra!! and man i miss that so much!!! I miss u guys soo much!! man before it was lyk we had a connection all of us n i hope we can establish that again!! kra man things have changed so much since our last posts here I mean I changed shcools.. isa and thom are friends rebeca got some ehehe.. bom lots a stuff but that still dosen\'t mean that something can\'t stay the same.. I love u guys more than anything!! i swear that withoutu guys i would be totally and utterly lost!! bom this is really corny n shit but w/e i don\'t care ehhe love ya all!!! hope we can keep this going BEiJAUUMMM sofieee   '),(23823,1292625,'10,May,2004','2004-05-10 00:00:00','   HII GUYZ!!! so ismas i was readin bloggs and i saw an old one of urs bout how u were all depressed that sof was leavin and that it was cause of her that u were friends wif me and that when she left we would neva hang out and that u would neva go out wif sof again!! well i guess that all changed!!! hehe!! u always go out wif sof and were always hangin out!! humm well!! sof miss u muito cara!! neva b the same here in st pauls wif out u!!!! hehe try not to replace and forget all bout us! hummm well w/e!! vry corny but I LOVES U ALL!! (SNIFF SNIFF) WELL nada to write!! bjs lu oh ya and kiko stresses me... the way he says a sua amiga!!! HUMPH!!! wat eva!!! miss marysue had a flue to u all!   '),(23824,1292625,'10,May,2004','2004-05-10 00:00:00','   HIII GUYZ!!!!!!!!!! meu its been such a long time kara... heheh well im supposed to b studyin for exams and doin english homework.... but i just cant b bothered to right now!! meu we were such loosers assim!!! tom and isa were always fightin but those fights were the best!!! (sowy guyz but they were) i mean all that shit u guyz would actually come up wif in like one second!!! ooo nooo i think that u guyz would PRACTICE (dunno how to spell that) in ur rooms!!! hahahaaha!!! meu it brings back memories!!! sniff sniff!\"!! well now we\'re older and mature (coff coff) meu dunno wat else to write... otha than how ze fuck do a post this thing?\" oh welll no one cares!!! Giggolo Joe wat do u know!!! uhauhauha!\"! bom see ya at skool!! sof see u in 2 weeks!! bjs lu\\!   '),(23825,1292625,'20,May,2004','2004-05-20 00:00:00','   hey ppl man my work load has been diminished by lyk half thank the lord!!mm oh yea if u go to the site and u look under each post thing there is a comment box n u can post comments on eachothers posts.. random fact.. left one on one of lu\'s.. hmm i hvae the distinct feeling that isa isn\'t usuing this ne more butoh well... umm so yeah not feeling too well my throught gout worse but thats ok umm gotta goo beijoss byee   '),(23826,1292625,'18,May,2004','2004-05-18 00:00:00','   OMIGOSH!!!!!!!!!!!!!!!! TOM AND ISA SCARE  CROW WAS AT SKOOL!!!!!!! SHES THE BEST CARA!!!ELA É MUITO FEIA MEU!!!!!!! UHAUHAUHAUHAUHA!!!SHE WAS THE BEST MONETOR WE EVA HAD!!!! BOM I CANT STAND STUDYIN ANYMO.... AI AI AI!!!! AND I STILL DONT KNOW HOW TO POST THIS CRAP!!  AAAAAA!!  SOOOOO MAN U GUYZ SHOULD B VRY HAPPY THAT ME AND MY VOVOZINHA ARE DOIN THE HISTÓRIA NOTES CAUSE  THEY SUCK!!!!!!1 É MUITA COISA MAN!!!! ohhh well!!!! MRS WILLY WONKA ALSO KNOW AS THOMAS\'S POODEL LOVA SNORETS WHEN SHE LAUGHS!!! UHAUHAUHAUHAUHAUHAUHA!!!!!!!!OH WELL BYE BYE! I LOVE ME!!!!! luu   '),(23827,1292625,'18,May,2004','2004-05-18 00:00:00','   OMIGOSH!!!!!!!!!!!!!!!! TOM AND ISA SCARE  CROW WAS AT SKOOL!!!!!!! SHES THE BEST CARA!!!ELA É MUITO FEIA MEU!!!!!!! UHAUHAUHAUHAUHA!!!SHE WAS THE BEST MONETOR WE EVA HAD!!!! BOM I CANT STAND STUDYIN ANYMO.... AI AI AI!!!! AND I STILL DONT KNOW HOW TO POST THIS CRAP!!  AAAAAA!!  SOOOOO MAN U GUYZ SHOULD B VRY HAPPY THAT ME AND MY VOVOZINHA ARE DOIN THE HISTÓRIA NOTES CAUSE  THEY SUCK!!!!!!1 É MUITA COISA MAN!!!! ohhh well!!!! MRS WILLY WONKA ALSO KNOW AS THOMAS\'S POODEL LOVA SNORETS WHEN SHE LAUGHS!!! UHAUHAUHAUHAUHAUHAUHA!!!!!!!!OH WELL BYE BYE! I LOVE ME!!!!! luu   '),(23828,1292625,'18,May,2004','2004-05-18 00:00:00','   Woman, we are on the same boat... GOD!! I\'m lyk sinking in me notes! Ive finished physics and chemistry is just DRAGGGGGING ON... I want to finish tem tonight tho... Today Preece was lecturing us on how we make noise, are disorganized, disrespectfull (bla, bla, bla...) and he was standing lyk right infront of me and i was drowzing away... half asleep, half away... i actually think I was more asleep than awake... NE WAYS... YAY!! I found my DJ course!! God I REALLY wanna do it!! Id be lyk Dj tommy lem... no not really thats just sad... Im just really happy that its all almost over and that im going to be in a few weeks chillin... DID I TELL U GUYS THAT I MIGHT HAVE A LIGHT THAT WHEN I OPEN MY CLOSET TURNS ON?!?! I MEAN HOW COOL IS THAT!?!?!?!? Man next years play is going to suxxx... the script sux... the actors suxxx (matheus and thel... among others... NOT YOU ISA!! UR REALLY GOOD!!) Sei la... Im looking forward to next term tho... dont know why... sei la... Man HOLIDAYS HERE I COME!!!! AND WE LYK THE NEW TEMPLATE?!?!?! WE LOVE!!! Bjoinhos The Duke Of Le Grand Tampon   '),(23829,1292625,'18,May,2004','2004-05-18 00:00:00','   hey ppl grrr no1 writing!! man this week is begining to be hell!! its tuesday and i have sooooooooo much hw!! história test tomorrow, hand in my história proj tomorrow, history proj, science current events proj god soo much work and I feel lyk i\'m gonna break soon!! I\'m coming down with someting i can feel it! not a good week for me all i can think of is that next week it will all be over and it will be time to start studying.. haven\'t started yet n exams are in 1 1/2 weeks puta waaaaaaaay too much work!! on top of that i can\'t see u guys cuz of the work n exams and all that bullshit.. nico n vero situation is brightening but still pissing me off! God i just want it all to go away for now! bright side erin coming home on sat... wow 1 gleem of light! last days of school wish i could be having more fun in them cuz i know i\'m gonna miss it lyk 2 weeks into the holls! well man gonna go work  BEIJAUM to ya all miss ya.. (lu gettin the nb ready)have it for u next time i see ya ok.. unless its tomorrow hehe   '),(23830,1292625,'23,May,2004','2004-05-23 00:00:00','   hey ppl its sunday.. good luck u guys exams this week ne?? well good luck i know u guys are all gonna do great!!! jut relax right b4 the exam and go through everything in ur head, no point in getting too stressed out cuz i know at least that all of u guys are reaaally smart without a stupid exam! which exam if 1st by the way??.. I\'ll be thinking about u while ur taking it!! mm ne ways mine start neext tuesday.. GODD i have to get down to studying!! expecially maths man i can\'t figure out the quadratic formula and i did it at the beginig of this semester!! do any of u know it?? oh well all fret later..  still sick (isa its all ur fault!!! eheh jk) mm ok bye then beijauumm ps erins home YAY   '),(23831,1292625,'04,July,2004','2004-07-04 00:00:00','   ok hey gy i was just wndering if it was worth posting things... if anyone was reading or chcking thse besides me cuz if u are then i will so start posting things ok? ell byeesss miss yaall   '),(23832,1292625,'28,May,2003','2003-05-28 00:00:00','    ok... its only me here talkin to me self!!!! sofia and sofia sitting in a tree S N O G I N G FRIST COMES insanity then comes... then comes the babby in the baby carrige I didn\'t know how to complete it  well .. night me   '),(23833,1292625,'28,May,2003','2003-05-28 00:00:00','    HEY SLIM WE HAVE A BLOG TOGWTHER TELL ME IF U WANNA INVITE NE ONE ELSE   '),(23834,1292625,'29,June,2003','2003-06-29 00:00:00','    hey sofs like i aint talkin ta knowone again cuz mister sister sofia aint ther ta write in r blog cuz shes in the frieken states!!! palees, is ther anythin mor rich bich then that, mister sister? well i aint gonna send ya any mor useless posts cuz im leavin tomorrow, happy? tear tear i knnow but ill only b away 4 like a week. wen i cum bak im gonna rest my ass and wait 4 you mister rich bich sisters ta cum bak!!!!!!!!!!!!!!!!!!!!1 cum bak soon  i need ta go out dont leave me!!!!!!!!!!! noooooooooooooooooooooooooooooooooooooooooo hehehe ive rilly gon insane this time an its all yor falt mister rich bich sister sofia mmmmmkkkkkaaaayyy an don say i write a lot,cuz im bored!!!!!!!!!! weell hope ur havin fun bu-bye mrbss (Mister Rich Bich Sister Sofia) hehehehehe ashy   '),(23835,1292625,'28,June,2003','2003-06-28 00:00:00','    hiya sof im leavin on monday ta go ta ma aunts farm thers noone else who aint travellin (i think) up ta now ma holiday have bin mega boring i might go somwer in july but i dunno plz tell me wen ur cumin back so we can have a little mor fun before skool starts see ya hope ur havin fun ashy   '),(23836,1292625,'26,June,2003','2003-06-26 00:00:00','    hiya sof ya didnt call me or write back but u probably went somwer its a pity yor leavin today cuz i aint goin anywer wen r ya cumin back well see ya soon  im goin ta ma aunts farm next week soooooo..... see ya please KIT wen ya cum back dont 4get bout me mister sister mmmmkkkkaaaayyyyyyy BYE BYE bu-bye lorv me hehehehehehehehehehe ashy   '),(23837,1292625,'24,June,2003','2003-06-24 00:00:00','    hehehehehehe yeahI can\'t do ne thing to day until around 3:30 and then I dunno need ta talk to me mom so ... we\'ll see   '),(23838,1292625,'22,June,2003','2003-06-22 00:00:00','    hey sofs glad ya wrote soon i just came back from ma country house tomorrow i have a music lesson at 4:00 so we culd do somthin after......... write back tomorrow or call me kay? yeah ill KIT so u KIT as well and we\'ll all KIT.... yey bu-buy hehehehehehehehe so ya like me bein insane do ya???????????????? hehehehehehehehe bye ashy   '),(23839,1292625,'18,June,2003','2003-06-18 00:00:00','    I think its cool when ur insane (very funny)  well this week I am basicly free until thursday cuz i\'m leaving on thursday so we can plan to do something  r u free tomorrow i.e monday?? well KIT (keep in touch) xx sofs   '),(23840,1292625,'18,June,2003','2003-06-18 00:00:00','    ya rilly did like like that dumb nickname? yeah well thanx 4 invitin me 2 go bowlin but noone culd take me... tell me when yor free so we culd go 2 the movies this week sorry 4 saying all that crap last post i wuz rilly insane well see ya around ........ hehehehehehehehehehehe lorv moi bu-bye   '),(23841,1292625,'15,June,2003','2003-06-15 00:00:00','    o sorry I haven\'t been writting yea ashy i think your going insane well w/e wanna go to the movies sometime thins week??? well need ta go xx mss   '),(23842,1292625,'12,June,2003','2003-06-12 00:00:00','    you never write this makes it boring   '),(23843,1292625,'11,June,2003','2003-06-11 00:00:00','    hiya sof whats rollin (nothin dude stop askin me that!) yeah cant wait 4 friday ha shmee and shmoo and shmall the schmother shmuys are shmoin shmout!! (me and you and all the other guys are goin out) he yeah im bored as always shmoochkins ha once again im talking ta meself cause mister sister sofia aint goin in our blog to check out what i wrote cause shes too busy readin her other frieken blogs!! ye so sumarisin: plz chec out our blog dude ha mmmmmmmkkkkaaaaaaayyyyyy didja get the message mister sister sofia? so dont go around doin 6things ya shouldnt b doin instead of redin wat i wrote mister sister sofia!!!!!!!!!!!!!!!!!!!!!!  he like i just finished given ya the bigest moral speech ever huh? whatever maybe im the one goin insane (plz stop writin) no nigger bitch i write wat i want mister sister sofia mmmmmkkkkkaaaaayyyyyy ye i gotta do other stuff so ill stop writin write back mss (yor new nickname- mister sister sofia)!!!!!!!!!!!!!!   '),(23844,1292625,'10,June,2003','2003-06-10 00:00:00','    hehehehehehehehe we are 1 meeeeeeeeaaaaaannnnnn group we gotta gang up on gudo (bitch) hehehehehehehehe look youve got 2.5 days left at skool........... sorry for remindin ya mmmmmmmmmkkkkkkkkaaaaaaaaaaayyyyyyyyyyy mister sister nigger bitch ha bu-bye   '),(23845,1292625,'09,June,2003','2003-06-09 00:00:00','    hdo i invite someone?   '),(23846,1292625,'09,June,2003','2003-06-09 00:00:00','    slim you do write a lot... yeah I am talking while I write  I think I\'m going a lil insane sure I would like to go to hopi hari yeah too bad u cant be here on friday becuase its like the last day of school ok well yeaaaaaaaa mmkkkaaaaaaaaaaa bye byeeaaaaaaaaaaa talk to you laterrrrrrrrrrrrrr   '),(23847,1292625,'09,June,2003','2003-06-09 00:00:00','    hiya im listenin 2 becker sayin bla bla bla bla ........ yeah do check what i write often hehehehehehehhheheh   '),(23848,1292625,'09,June,2003','2003-06-09 00:00:00','    sofs are you on right now?? just got yor message now...   '),(23849,1292625,'09,June,2003','2003-06-09 00:00:00','    hey sofs u rilly were bored when u wrote th\\t huh? yeah im kinda bored meself now is friday ya last day? i aint comin on friday lets make the most out of this week look ma aunt works for nestle...im gonna see if she can get some tickets for us 2 go 2 HOPIHARI.......that is if u want of course yeah were are lame people lorv meeeeeeeeeeeeee   '),(23850,1292625,'09,June,2003','2003-06-09 00:00:00','    ok then riight we only have 3.5 days left woohoo yay I am so happy and yet kinda worried cuz of graded and all. but im still really excited lalalalalala sing a happy song well tata need to go bye bye keep writting during the holidays ok?   '),(23851,1292625,'09,June,2003','2003-06-09 00:00:00','    hey sofs thanx alot 4 invitin me to have a blooog i dunno if ya wanna invte anyone if ya want ya can invite whoever i don mind hehehehehehheh yeah bow 2 me   '),(23852,1292625,'09,June,2003','2003-06-09 00:00:00','    pigeons, tampons and male tampons.... slim now has a mega blog im grossed out by wat tomtom showed us...im gonna have nightmares dude whatever im happy seeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeseeeeeeeeeeeeeeeeeeeeeeeeeeeeee ya slim   '),(23853,1292625,'28,July,2003','2003-07-28 00:00:00','    hey sofiaaaaaaaaaaaaaaaa yeah like my frieksong? well sorry ive been riitin so often im bored yeah guess ive bin tellin u that all summer huh? wen r ya cumin bak from italy? havent bin able ta contact anyone here im talkin ta u, tomtom and liz from difrent countries whatever its just weird wwrite soon bu-bye lorv ya ashy   '),(23854,1292625,'27,July,2003','2003-07-27 00:00:00','    hey ho whos the man yo its so- -fia !!!!!!!!1  hehehehe write back kay misss ya a bunch seen the movie its kay havent gone to ho. ha. yet g2g KIT ashy   '),(23855,1292625,'23,July,2003','2003-07-23 00:00:00','    hey agen sofs ive been tryin ta read yor message agen to get what ya wrote and i can sortta make it out... the worse thing is that ive just had my eyes dilapidated at the eye cirgion...dont worry nothin serious...its just that i feel tired all the time and i cant see whats under my nose anyway..... yeah i dunno if ill be seeing lu or bells but if i do ill tell them that u sed hii and all yo i also miss ya a lot see ya soon lorv ya to man ashy   '),(23856,1292625,'22,July,2003','2003-07-22 00:00:00','    sofs  somthins gon wrong and i cant read yor last post properly like lu and bells are in sao paulo? i didnt get what you said so write again sorry ive been sleepin rilly badly too like 4 hours sleep im finally goin 2 the movies tonight and im goin to hopi hari tomorrow have fun in italy tommy wrote and apparently he met barnao dude ive made somethin 4 u ages ago but havent had the time ta give it 2 ya... write soon bye!!!!!!!!!!!!!!!!!!!!!!! ashy   '),(23857,1292625,'22,July,2003','2003-07-22 00:00:00','    hey ashy yeah now isa and lu are in sao paulo bbut thomas is only getting in on the 31.  I am in italia now and this keyboard is really stressing me!!! so how is every thing for u?   I am pretty tired but I can\\t fall asleep dunno why its like 6 15 here what about there well if u see lu or bells or the 2 of them say hi and tell them i miss them. I miss u to and need to go now bye bye lurv ya loads mss   '),(23858,1292625,'20,July,2003','2003-07-20 00:00:00','    sof do ya know if anyone is in sao paulo now? g2g bye ashy   '),(23859,1292625,'19,July,2003','2003-07-19 00:00:00','    hiya sofs r ya gonna have a comp in italia? man i can barly buleev....ill b bak in skool in like 3 weeks this sux man well im ok how bout u ??? KIT kay bye have fun there ashy p(m)s: ive made another blog but how do i invite someone else to join it?   '),(23860,1292625,'16,July,2003','2003-07-16 00:00:00','    ya know ya can also send me an email rite? well im bored here but im gonna see charlies angels 2 this week and im gointa hopi hari pity yor not here.... g2g bu-bye ashy   '),(23861,1292625,'14,July,2003','2003-07-14 00:00:00','    hey sofs glad ya wrote IM BACK!!!!!!1 the farm wuz rilly cool im gettin better at horseridin... thanx 4 askin how r u there? italy he, not bad 4 a short vacation.... man its freezin here it must be rilly warm ther sorry bout thr confuusion maybe you dialed the wrong numbeer OR ma aunt who stays at ma house has gone gaga dont worry how 2 pronounce ma name.... see ya thanx 4 writin KIT mmmkkkaayy shmee   '),(23862,1292625,'03,July,2003','2003-07-03 00:00:00','    hey sorry I haven`t been puttin in posts but I havent had a comp for DAYS well i`ll onl;y be home like a week b4 term ends becuase after i leave here I`m going to Italia!! opa!!!  well how is everything and how was ur aunts farm?? well I tried to call u the day b4 i left but i first I called and was like eu posso fala coma Aslinn (man how do u pronounce the name in portugues??) and she was like quem ?? and I was like what??? and she was like que??? and then I hung up. well i called back and she was like there aint no asslin here and hung up in my face. so I was too scared to call back. well I``l write more later bye bye moi   '),(23863,1292625,'28,August,2003','2003-08-28 00:00:00','    hey sofs didja go in school cuz i was at the front gate all the time hey i also rilly wana go 2 the movies but this weekend i cant cuz its ma dads bday and were goin to rio.... next week i promise... soooorrrrrrrrrrrryyyyyyy y ju ask bout becker? shes just the old hands on de monitors oh and bork didnt get betin up cuz they didnt want 2 call attnetion or somethin......anyway hes gross c u then........next week MOVIES i promise...... love ya ashy   '),(23864,1292625,'26,August,2003','2003-08-26 00:00:00','    hey ashy man mb we can go to the movies tomorrw I went by school today but I didn\'t find ne one so I went back to my house! hey how is becker!!??? man I hear the thing about bork! that is disgusting and I hope he was beaten upfor it!!! well i\'m really cold tierd and I have alot of hw so I\'ll be going now bye  xx   '),(23865,1292625,'26,August,2003','2003-08-26 00:00:00','    hey sofs im so frieken gladja wrote oh an my phones easy peezy: 3063-9582 OK? yor phones easy....too anyway yeah um man im freezin ma butts off over here how bout u? i wrote ya an email just now so have a look... miss ya ttoo see ya ashy   '),(23866,1292625,'17,August,2003','2003-08-17 00:00:00','    heellooo there man what is ur freakin phone number I used to have the right on but now when I call u it says that the number does not exsit which is really anoying!!! well my new number is 99165569 ok? well do u know lizzes phone??? cuz no 1 else really does! mb thomas does well bye bye i\'ll come round school later miss ya the party was great had lots a fun we should do it agen some time luv ya bye bye   '),(23867,1292625,'14,August,2003','2003-08-14 00:00:00','    hey   '),(23868,1292625,'12,August,2003','2003-08-12 00:00:00','    hey sofs, so how wuz the weddin agen? man ya gave me a frikken fright today....so can u come tomorrow...im gonna call everyone today......pleez come if u can.......or just not......so bye then.......glad ta c ya ashy ps. ya havent told me bout graded so ill b forced ta call u and ask u personally mister...........bout the bowlin..thomas sed he cant come cuz hes gonna go 2 the psycologist!!!!   '),(23869,1292625,'11,August,2003','2003-08-11 00:00:00','    hey   '),(23870,1292625,'08,August,2003','2003-08-08 00:00:00','    yor kiddin? omg then my bday\'s gonna b another shitbag cuz i never do anythin on my bday... man this is soo bad i wuz hopin i culd go 2 the movies with u guys................. this sux well then bye   '),(23871,1292625,'07,August,2003','2003-08-07 00:00:00','    hey I cn\'t this weekend mb next cuz my cousins wedding is this weekend and Inhave to read this freakishly boring book so bye byexx   '),(23872,1292625,'06,August,2003','2003-08-06 00:00:00','    yeah like ive been writin the last posts mister sister and u aint even reedin them! anyway heres wat i wuz gonna say............] DO YOU WANNA GO 2 THE MOVIES MAYBE ON THE WEEKEND OR ON TUESDAY TA CELEBRATE ME BIRTHDAY??? I WAS ALSO THINKIN OF INVITIN THE GANG SO WHAT DO YA THINK MSS??????????????? i rilly wanna go somewhere so just give me a call (3063-9582) OR write back OR pretend u havent red this and dont wanna go, mkay? im gonna send ya an email now bout wat goin on at skool so u shuld du the same bye then miss ya even though i saw ya yesterday tata ashy   '),(23873,1292625,'06,August,2003','2003-08-06 00:00:00','    sofiaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa aa a                    a                      a        a  Ok mister heres what..........im this blonde bitch......no wait ceci said that..... i think im gonna dye my hair black and cut it this short.....yeah that sounds more like you..........hehehehehe............anyway next post im gonna tell u somthin important sooooooooooooooo ull just have 2 waaaaaaaaaaaaaaaaaiiiiiiiiiiiiiitttttttttttttt, mmmkkkaaayyyyyyyyyyy. so mister sister sofia, suppppp? im just fiiine....hoho.......dude ur just one mister sis SOFIA!!!!!!!!!!!!!!!!!!........but dude i have 2 stop writin now my hands startin ta hurt (not) so bye then weirdo....................   '),(23874,1292625,'05,August,2003','2003-08-05 00:00:00','    sofiaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa   '),(23875,1292625,'05,August,2003','2003-08-05 00:00:00','    sofs ya havent bin writin mister sister aha aha aha aha  well write soon ashy   '),(23876,1292625,'29,September,2003','2003-09-29 00:00:00','    sofs meu all of us really miss u i think we should have like a big sleepover or movies for all of us...watcha thinmk? c u on msn honey..hehehe   '),(23877,1292625,'25,September,2003','2003-09-25 00:00:00','    hey sofs missed out yor las post..umm great idea, but tell me nowww cuz im leavin te me contre haus..hehe..boring anyways i changed my email an got msn so there..g2g now c ya   '),(23878,1292625,'24,September,2003','2003-09-24 00:00:00','    ohoh oooo yeaaahh woohoo yo sister sup? im just chilin out or BOIANDO hehe saucy chau   '),(23879,1292625,'21,September,2003','2003-09-21 00:00:00','    man the party was gr8 u guys r soo lucky u don\'t have school this week well I have fri off yay umm I\'m thinking of having ppl out to my farm what do u think?   '),(23880,1292625,'12,September,2003','2003-09-12 00:00:00','    parteeeeeeeeeeee   '),(23881,1292625,'11,September,2003','2003-09-11 00:00:00','    hiya so man im rilly lookin forward to the party hope 2 see ya there were gonna rock that dance floor...hehehe jk   '),(23882,1292625,'10,September,2003','2003-09-10 00:00:00','    helllllllllllooooooooooooooooooooooooooooooooooooooooooooooooooooooo   '),(23883,1292625,'05,September,2003','2003-09-05 00:00:00','    were u at yor skool wen u wrote that? anyway i luved see in ya yesterday!!!!!! hope 2 c ya soon  have 2 ask u somthin but cant now ....surprise surprise...... and wat the hell is muahahahahahaha???? i luv 2 laugh hehehehehehhuahuahuahauahua ashy   '),(23884,1292625,'05,September,2003','2003-09-05 00:00:00','    heeloo slim ! whatsup?? ummm need ta o evil tacker watching over me!!!! muahahahaha   '),(23885,1292625,'09,October,2003','2003-10-09 00:00:00','    eeeeeeeeeekk heya woman!!!!!!! sup? im like boiAndo here wassup with u there? well c ya then  tata for now hehehe ashy   '),(23886,1292625,'01,October,2003','2003-10-01 00:00:00','    ooo herro thats hello actually well wat is up?!!? hmmm  umm I am very shmery bored and it is very bad fro my phycologcal (never understood why is spelled like that) state well byebyes 4 now xx good idea for the sleep over thing   '),(23887,1292625,'12,February,2004','2004-02-12 00:00:00','    oops yeah kinda didnt work with ba\'s party n all but NEVER MIND... so heard u gonna take some graded ppl over... great! so wat u gonnar wear? i hav absolutly no idea im lost but its casual so wateva bring extra clothes were gonna throw ba in the pool ( we might get thrown, i suppose) so u got plans after? umm we might go over to isas i dunno after but wat a bout u? well well the bat has entered the CAVE cave ave ave ave ave hehe that would be the book...ya k-no wat i mean...ttfn luv ya   '),(23888,1292625,'09,February,2004','2004-02-09 00:00:00','    ahhhhhh heelo sup? so u lik jonny depp huh? well then who the hell is the dude in the pic? does not look like him... n ways im well kinda bored but go to this site: www.happytreefriends.com if ya wanna c some cute/killer stuff quite gross but entertaining...lalala even if i were broke my love dont cost a thing...so tel me bout the bike incident.....did u think isa was babyish? well ttfn bjaum   '),(23889,1292625,'05,February,2004','2004-02-05 00:00:00','    ahhhhh hiya sof...sup woman? meu i realyl miss ya...hows graded n all? no special kids like us ther...hehehe...but we r havin tooo much hwk n its gettin on me nerves...but the rest is chill..so next week its tommys bday...were planin a surprise party..dont know when or how but were thinkin..hop ta c ya soon then....lets plan somthin...by4now...bjaum!   '),(23890,1292625,'18,March,2004','2004-03-18 00:00:00','    mmmmmmm eating maltizers and writing in notebook.... hope ta see u soon not really in the mood for writing as can probably tell hehe wel beijauuuuuuuum bye   '),(23891,1292625,'30,April,2004','2004-04-30 00:00:00','    bom in history writing letters hehe booorring n i was lyk oh yeah i was suposed to go on this thing a looong time ago so yea... ummm william just cme up and kissed me randomly.. lyk i didn`t know that any1 was around me and then he just pops out of no were and kisses me hehe twas quite wierd, mmm n b4 that he was like hanging on my literallybut lyk i just pushed him away both times.. it was quite funny and also nasty,  bom its prolly gonna be lyk another month until i check this again and prolly around 2 until u answer soo i`ll just go for now beujauuuummm sofie   '),(23892,1292625,'22,April,2004','2004-04-22 00:00:00','    hehehe hey! kinda forgot bout this.....well party was great! heard william kinda harassed ya......tell me about it......!talk later.....in ict....   '),(23893,1292625,'28,May,2004','2004-05-28 00:00:00','    ahhhhhh!!!! examssss r herre....man historia was really hard...but geography....mEU we studyid so much shit and NONE of it was on the exam! vai se fuderrr woman....ommg lk me n lu were studyin on those mats in the gym....then we very cleverly decide to get off them on the higher side...we stayed hanging on the thing..and we just felt it fallin n fallin.....PA....OWWWWWWW........all our shit wa on the floor and we were thrown on those metal cages!!!! we were laghin cryin!!heheh but my ass still hurts.!!!!..to make it worse my contacts POPPED out in revision period.....i made it In time though!!!!!!! BARNES HAD A MICROFONE and it was really funny what she did...gotta show ya!!1 well bjs! g2g....   '),(23894,1292625,'25,May,2004','2004-05-25 00:00:00','    hiya! hehehehe.....thats funny...well its happened before to me...i wrote a HUUUGGE email literally the biggest ever n the school system went off...dude i was sooo pissed....yeah well thats y now i try to shorten things down (not really...)wellll.....yeah today was fun....we ran after KATEEEEEE....and we asked her how was her mother...she said.....shes UGLY!!!....i had already seen her....so i almost said.....SHE LOOKS LK U!!!!...heheh but come again that would b mean.. ahhhhh....just found out got an exam this thurs...sommeone has to tell me these things....caus A SENHORA DO TEMPO ie a velhota.....(makins- she STOPS time in her lesson- seriously)doesnt say shit  2 us!!!!!!! well yaaaa.....i was thinkin....mean we were bout havin a sleepover at the end of year...YAYYYY!!!....welll plan later ....g2g!!! bjs....oh u see bout the le grand tampon thingy!!! bjsss   '),(23895,1292625,'23,May,2004','2004-05-23 00:00:00','    ackkkkkkkk i just wrote a puta long post kra and it  was lyk deleted or something!! man this new template thing reaaally confuses me!! kra so do u wanna b part of the great tapon??mano its soo anoying it takes ppl 4eva to write kra!! ok bom i\'m off  sick n I wanna sleep bjooinhos byee   '),(23896,1292625,'21,May,2004','2004-05-21 00:00:00','    heyyy...yeah m,e again..heheheheh well me n thom were readin le grand tampon today.....and how u guys didnt invite isa to lus party last year (hehe...well actually i didnt go either so.....i shouldnt be laughin)n how thom was a bitch to isis.... man to think i was tryin to make a blog for all five of us....u guys already had one!! no one told me shit...man pissin....yeeer...hears yer sister is cumin over sat! cool !!1 welll.....lllooooads of work to do....exams exmas i havent done nthing im screwed....one week left!!!! welll.....g2g2 thenn!!!! bjinkus   '),(23897,1292625,'13,May,2004','2004-05-13 00:00:00','    heyyyy!!!!! ahhhh sof i misss ya loaadds!!thought u would show up at school  this week....i have to show ya somthin....lk REALLY OLD!!! man lke today was kinda weird...we remebered ya randomly....and there were tangerines for lunch..then isa said.....AAAA...TANGERINES FROM SOFS FARM......AHAUHAUHAUAHUAHA friek man ....ooo and we invaded the prep classes once again.....its soo fun......we even got free chocloate! and theres this little fat kid that stalks thomas.....and for some reason calls him DUCK....its funny though...U DO NOTT KNOW MAN.......lk we were walkin in the field and ALL the boys were playin rugby.....THEN nthey grabbed leo\'s pants and pulled it down so we coul dsee his ass.....IIITTSSSS ALLL HAIIRY!!!!!!! AHHHHHHHHH soooo grooosss man im gonna have nightmare today seriously!!! welll not really feel lk writin ne more....but i was thinkin if when exams are over for us to go to movies...and have a PUTA sleepover???? could b at me place if u guys want....welll g2g then...actually i dont i just wrote too much.....tata then bjssss.l......ps...cum by some time mkkkayyy......for some reason i saw isa today going to your house????k then tootles   '),(23898,1292625,'13,May,2004','2004-05-13 00:00:00','    heyy!! omg check out this new format!! its alright just that ive spent a few mins tryin to figure it out...heheh...im sooo confused!..hehe yearrr...well...wait a sec..u like nikko???? i mean, u too???....tell me...tell me...ahhh....how did he like me...all he kept sayin was....SHE DOES NOT LOOK LIKE AMY....whoever that is....and yeah happy late bday nikko even though i dont know u...random once again....WE WENT TO THE CAS THINGY!11!!! to plant......TREEEES...!!! it was so piontless...but were getting refunded anyway soo....hahahahah I GAVE TOMMY a puta WEDGIE!11!!....to compensate for the 10 he gave isa...hehheheheh...and today...we took a pic of this penis shaped thingy we found....we were kinda using isas film..hehe....ohhhhhh and heard u got the pics from your party (isa stuck a picture of u on the cover of her book...dont tell her though)sooo let us seeee.....k im overwriting soo one last thing...do u wanna come to the movies this friday or weekend to watch TROYY (mm) cuz everyoine wants to study...well me too but we gotta have some fun too.....so tell me and well see...bubye now!!!!!bjs   '),(23899,1292625,'06,May,2004','2004-05-06 00:00:00','    HEEYYYYYYYYYYYYY man what u sorry about?? man i dunno why we didn\'t tell u sorry it was just that lu was confused and kiko told me not to tell ne one about lu not even lu but i told lu and yeah that worked out pretty well eheh.. umm my farm ws great.. fun fun and now its may 6 nico\'s b day eheh random but yeah mm i might lyk him.. duno but yeah.. he thought u were reeeaaally pretty bom i\'ll be off mb u won\'t check this so when u do then i will know that its reagualr and will keep writing  bom beijinhos byee   '),(23900,1292625,'03,May,2004','2004-05-03 00:00:00','    hey!? HAPPY BDAY!!!!!!!!!!1 YAY!!!!!! LUV YA BJAUUUUUUMMMM LURVE YA!!!! yeah prob would b a month till i remembered bout  this but today is special sooo , meu i luved you rparty really.....except for the fact that my parents lk randomly popped in......they got sooo pissed at me for nto callin i totally forgot.....ow well wat to do....how was it at your farm then.????..never forget bout the PUTA ORGIEEEEE;...hehehe omg man noone told me lu liked kiko  i had no idea u guys didnt say anythin.....meu wat was the prob in tellin me,,,,,u guys kept on callin me and sayin....yeah ill tell u later.....that kinda pissed me off........well w/e if lu didnt say anythin its cuz she didnt want me to know.....thats ok then i still dont know.....well happy bday soffffiieis.....sry bout anythin...really......BJSSSS......BUBYE!!!   '),(23901,1292625,'06,June,2004','2004-06-06 00:00:00','    ahh...meann Eeevil woman....but dont think well actually work....its just another period..SOO...how was HOPI HARI???? FUNFUNFUN?? hop itwas....nothin to do...so bjs!!!   '),(23902,1292625,'04,June,2004','2004-06-04 00:00:00','    HEY YAY OMG I\'M FINNISHED WITH SCHOOOL AHHHHHHHHH YAY!! THANK GOD!!! HAHHAA AND U GUYS STILL HAVE SCHOOL FOR LYK ANOTHER 2 WEEEKSS AHAH SORRY.. BOM THIS WAS JUST A QUICK NOTE.. MMMM I\'LL CALL U LATER BOM BJOOSS BYEE   '),(23903,1292625,'03,June,2004','2004-06-03 00:00:00','    EXAMMMSS AREEEE OVVEERRRRR!!!!!! YAYYY!!!! im reall glad these shiitens were pissin me off....ooo ....some1 got the french exam and th e whole grade saw it....they caught them and 4 ppl r suspended......ooooooo....(guess who..duh) welll im bundando 4 now...so nthin just giv me a call....bjssssss....   '),(23904,1292625,'02,June,2004','2004-06-02 00:00:00','    HEYY MAN OK SO EXAM STRESS I KNOW WHAT UR TALKING ABOUT WE ONLY HAD LIKE A WEEKEND WITHOUT HW AND SO I ONLY HAD TIME TO STUDY ON THE WEEKEDN FLA SERIU!! WHAT HAPPEND TO STUDY WEEK?? IT JUST DISSAPEARED! OH WELL ENOUGH OF THAT.. WE ONL AHVE ONE DAY OF SCHOOL LEFT YAAAAAAAAAAY!!! OMG SOO HAPPY!! N 2 EXAMS LEFT THANK THE LORD! I DON\'T BELIEVE WE  DON\' HAVE ANY MORE SCHOOL ITS GONNA FEEL SOOO WIERD!! N NEXT YEAR I ON\'T EB THE NEW GIRL ANY MORE YAY  BOM I AHVE TO FINNISH STUDYING FOR MATH EHEH BJOOSS   '),(23905,1292625,'01,June,2004','2004-06-01 00:00:00','    ahhhhh...write WOMAN!! well well exmas exams....sniffsniff....school is ever so tedious.....but what can we do.....oooh...might have sleepover or somthin when schools over mkayyy?? well see ya soon..... ps....I STILL GOT TA GIVE YA SOMTHIN!! BJS   '),(23906,3650032,'04,August,2004','2004-08-04 00:00:00','   These pics are from a bike ride to the north of Yonezawa. I was guided around by Andrew (a 2nd year ALT also stationed in Yonezawa), and we spent most of the time following the main river out of the city. I spent much more time biking than taking photos, so there isn\'t much to show unfortunately. Should be exciting for river fetishists though.   urlLink     urlLink     urlLink     urlLink  '),(23907,3650032,'04,August,2004','2004-08-04 00:00:00','   Here are a couple of shots from a Buddhist temple on the outskirts of Yonezawa. Apologies for the blurriness of the bottom-right photo (up until this point, I hadn\'t been using the focus properly on the digital camera). Idiot proof? Not bloody likely.   urlLink     urlLink     urlLink     urlLink  '),(23908,3650032,'04,August,2004','2004-08-04 00:00:00','   These photos were taken during my commute to Yonezawa, from Tokyo. I quickly learned that I would have to keep my camera at the ready, otherwise the shinkansens (bullet trains) would vanish out of sight.  There were a couple of varieties of shinkansen that went by while we were there. The two most interesting ones were the ultrasleek stupidly high speed variant (not pictured), and the double-decker version (*sigh* - also not pictured). I did manage to get photos of the type that I travelled in though (hooray!).   urlLink     urlLink     urlLink     urlLink  '),(23909,3650032,'04,August,2004','2004-08-04 00:00:00','   Here is the first batch of photos from my stay in Tokyo.   urlLink     urlLink     urlLink     urlLink     urlLink     urlLink     urlLink     urlLink  '),(23910,3650032,'04,August,2004','2004-08-04 00:00:00','   After an 11 hour flight to Narita Airport with 50 or so other JETs, I was shuttled to the Keio Plaza hotel in the Shinjuku district of Tokyo (the bus that took us there was a Mitsubishi \'Powertard\' *giggle*). A bunch of orientation seminars took place in the hotel, in an effort to better prepare us for our work and living in Japan. During the orientation there I met up with a ton of different people from around the world and went on random excursions into the city.  Although I only visited two areas of Tokyo (Shijuku and Shibuya), the experience was almost overwhelming. Video screens the size of buildings were common in many areas of Shibuya, spouting advertising for J-Pop, movies, and personal hygiene products. Neon signage crept up to the sky, and all around there was an insane amount of activity. It is however kinda scary how quickly the city changes from cool exciting techno land, to unsettling dodgy porno land without much notice. Although the streets were well populated with people going about their business, there were guys handing out porno fliers every 30 paces or so. Strangely they were all black dudes, and seemed to target foreigners exclusively. Saying \'no thankyou\' didn\'t seem to do the trick so I would instead mention that I was a \'good Christian\' and that seemed to ward them off (apologies to the good Christian\'s out there).  Among other things, I checked out several electronics stores, including Yodobashi. The particular Yodobashi store that I went into in Shinjuku was a 9-story building selling nothing but electronics. There was some really really cool stuff there, but I restricted myself to the purchase of a digital camera. The camera is a pretty swish Canon limited edition 5MP unit. I bought it on my last day in Tokyo so unfortunately I only had a limited opportunity to snap off some shots. I took some from the hotel, looking out of the sprawling metropolis that is Tokyo, and got some of the Shinjuku station and the Shinkansen platform (where bullet trains arrive and depart). Unlike pretty much every other place in the world, the trains here run exactly on time. I was told I would be arriving in Yonezawa at 2:47, and I arrived there at 2:47 on the dot.  The bullet train I took was one of the slower types, but we were still racing along (at a guestimated 150-200kph). The seats on the train were kind of interesting because you could swivel them around through 180 degrees at any time so that you could face the people sitting behind you. It made for a rather intimate seating arrangement (especially for anyone larger than your standard Japanese businessman), but it was really cool.  Upon arrival in Yonezawa city (my new home for the next year or so), I was met at the station by my Supervisor (Kanomata-sensei) and Andrew Clare - a British JET in his 2nd year of the programme, and a really cool guy. They were both dressed down somewhat more than me in my full suit, and with good reason: the ambient temperature was around 34 degrees. The high humidity made things even more uncomfortable as the air was so saturated with moisture that sweat couldn\'t evaporate.  We high-tailed it to the office in Kanomata-sensei\'s car, where I met with the entire office and gave my self introduction in Japanese a couple of times. The people were really nice and friendly, doing a cool job of laying out the welcome mat. I filled out a bunch of forms with a little bit of Andrew\'s help (he reads and speaks Japanese well), and we were given the rest of the day off to look around the city. Yonezawa isn\'t an enormous city by any means, so bikes are very popular as a mode of transport. I feel somewhat lucky as I have one of the only mountain bikes in the area, whilst 99% of the population ride granny bikes.  An interesting difference between NZ and Japan is that cars are considered to be at the top of the food chain. Next are cyclists, then the lowly pedestrian. This means that unless you are in a car, you best have your wits about you. Another thing is that bikes are allowed on the footpath as well as the side of the road, which is fun if you\'re on a bike, and less fun if you\'re on foot.  At the end of my first day, Andrew took me to a \'beer garden\' where they have an all you can eat and drink offer for two hours. He was noticed by some of the Japanese dudes there and we were invited to join a party. They were a rowdy bunch but extremely friendly and generous and a good time was had by all. Andrew called up another JET in the region and we met for a few more drinks afterwards.  Something to note at this point are the cell phones (keitai) available in Japan. It seems that the flip open design is the most popular, with two screens - one on the outside of the phone, and a larger screen on the inside. Its also possible to receive TV transmissions on your cell phone - so... err... thats good I guess.  I returned to my apartment around 12pm and set up my futon. I was a little bit anxious, as I had sent my luggage ahead from Tokyo a day ago, and it hadn\'t arrived yet. It turned out that it wouldn\'t arrive for another 2 days, so I had to wear the same shirt and socks twice - a rather icky proposition given the climate.  The next day involved some more paperwork and another early finish at around 2pm. We were taken out to lunch by a person in the office for soba noodles which was a first for me. We sat down at a traditional table on tatami mats, and one of the guys started handing out magazines. These particular publications are like Women\'s Weekly or New Idea, except they cater mainly for men. What this means is that the centre of the mag is porn.  Porn is treated here in an even more relaxed fashion than I was expecting. You can go into pretty much any combini store (convenience store) and the porno section is RIGHT THERE. Another pretty much ubiquitous feature of Japan is the vending machine. These machines are everywhere and stock beer, food, coffee, toys and several other things. The panty machine is as yet elusive, but apparently you need to go to really dodgy areas of Tokyo for that \'privilege\'.  I have only had a bit of time to check out the TV here and man its some crazy shit. The advertisements are hilarious and the misuse of the English language makes for some great moments. A memorable example is a music show called \'Barfout!\'.  On Friday I spent some time trying to get an \'Alien Registration Card\' sorted out. Later on I went shopping with Andrew, to pick up some essentials for the apartment. We also went for a bike ride out to the countryside, riding along the side of a river looking for cool bike trails. We found what looked to be a temple, and went up to the entrance to check it out. We were welcomed in by a priest and were asked to sit and have chilled tea with him and his wife. They showed us around the place, and with their permission, Andrew and I snapped off some shots. The mixture of ancient ways and high technology that defines Japan was deftly illustrated when the priest commented on the quality of my camera, and that he suspected it was purchased from Yodobashi!  After a few more glasses of tea, we said goodbye and set off back towards the city. I stopped off at a video game store to see what was available, and was pretty amazed at the selection, considering that Yonezawa is a relatively small place. The XBox has virtually no presence here, its mostly PS2, or GBA. The store sold a lot of repackaged games and consoles, including some more exotic machines such as the NeoGeoCD, PCEngine, 3DO, and original Japanese NES with disk drive.  It looks like there\'s a lot of crap software here, but one of the games that caught my eye involved a giant bikini-clad woman, stomping across various city environments and melting tanks with her laser vision. I have no idea what part you play in the game (if any) but it had quality written all over it.  The remainder of Friday was spent watching the annual fireworks show, held just outside the centre of the city. There was some pretty exciting stuff there, and I caught a number of photos and some movie footage on the trusty camera. We went out for Karaoke afterwards, and had a pretty good time running through J-Pop and Eminem tunes with a bunch of people.  Saturday was spent trying to get the apartment sorted out, and I went out to the best place in Yonezawa for sushi with some JETs and their Japanese friends. I feel somewhat guilty that I have been eating out pretty much every night (under the guidance of Andrew-sensei!). The food here is nice, although some of the stuff available in the \'foreign\' food stores is a bit pooey. Peanut butter is a gelatinous mess. Most cereals are sugar encrusted cardiac depthcharges. Its just a matter of being persistent and eventually something palatable turns up.  '),(23911,3650032,'05,August,2004','2004-08-05 00:00:00','   Here are a couple of pictures from the Uesugi shrine, a moderately famous attraction in Yonezawa. Apparently the Uesugi clan brought great prosperity to the people of Yonezawa and are still highly revered by the locals. It was a bit late to go running around taking photos, but I didn\'t let common sense get in the way.  Note that even in the temple grounds, you can still score a refreshing beverage from one of the three(!) vending machines there.   urlLink     urlLink     urlLink     urlLink     urlLink     urlLink  '),(23912,3650032,'05,August,2004','2004-08-05 00:00:00','   Vending machines are cool. I chucked 200 yen (about $NZ3) into one and a cute little robo kitset fell out. The sword is a bit floppy, but its still a neat little toy.  Some of the electronics stores here have fantastically elaborate kitsets, with huge (and hugely expensive) assault suits available. The Gundam models seem to be the dominant types.   urlLink     urlLink     urlLink     urlLink  '),(23913,3650032,'09,August,2004','2004-08-09 00:00:00','   I have been looking through some of the Vodafone propaganda, trying to decide on a keitai (mobile phone).  Some of the features that caught my attention: - Video conferencing functionality - TV tuner - 2 Megapixel camera  Oh so saexay.  Unfortunately, I have to wait until the 16th of August before my Alien Registration Card is available (Vodafone requires this to set up a call plan).  '),(23914,3650032,'15,August,2004','2004-08-15 00:00:00','   Badminton indoors with no aircon @ 30+ degrees. Despite being loads of fun (especially when you play with two shuttles at one time), you are sure to end up totally drenched after an hour of play.  Recently, I biked to the onsens in Onogawa as Andrew lead the way (an onsen is a natural hot pool). For someone in better shape, the ride there and back wouldn\'t have been so bad, but at this stage even marginal inclines are a killer on my body. The onsen we went to was quite nice, although we cut it rather fine when it came time for the change over (when the male only pool switches to become the female only pool and vice versa). Apparently the sight of two gaijin in moderate states of undress was enough to ruin one old lady\'s day, and she slammed the door behind us as we left.  This sort of moodiness is all the more disturbing in Japan because it is so rare. Ho hum.  '),(23915,3907107,'24,July,2004','2004-07-24 00:00:00','   This is a message for all my school friends:  Thank you all for being there when I\'ve needed you. I fyou ever need anything then just ask and I\'ll help to the best I can.&nbsp;&nbsp; :)    &nbsp;  Some of my friends:   Thomas and Matthew, Wajeeha and Kim, Aksa and Natasha, Umar and Bilal, Jessica and Rabia, Nassim and Yani, Ashleigh and Sheridan&nbsp; and &nbsp;Priya (heh carnt forget you Priya lol)    &nbsp;  thats all for now,   ;) Marc  '),(23916,3907107,'24,July,2004','2004-07-24 00:00:00','   Wow ,  a whole blog dedicated&nbsp;to my&nbsp;friends! &nbsp;     Intro : Hi, I\'m Marc im 15 a&nbsp;Student from Bradford, UK. And this is a blog dedicated to all my friends online and off. So if Flip your reading then this is for you too!  '),(23917,3907107,'24,July,2004','2004-07-24 00:00:00','   A good place to sart is  urlLink www.ancestry.com &nbsp;searching is free and quite alot of the information (for the UK only) is free, however when you want to figure out who married who then you will have to pay for that.  '),(23918,3907107,'25,July,2005','2005-07-25 00:00:00','   Before we start looking on the net lets decide what a Family Tree just is.     Whats in a Family Tree?  Alot more than you could imagine. An increasing number of people are discovering that tracing their ancestors is NOT just an enjoinable past time - they can also uncover some amazing facts about there families. Such as what famous or infamous figures are you related to? Where did your family originate from? We\'ll if you get involved in Genealogy then you\'ll find out. Over the life of the website we\'re going to look at the roll your PC can play in tracing your family, starting with creating a means of recording all your family history. For this were going to use Microsoft Word. But before we begin, lets let\'s look at how you should organise your data. You can store details for each member of your family in a seperate Word document. To Be Continued...    &nbsp;  '),(23919,3907107,'24,July,2006','2006-07-24 00:00:00','   This blog is all about Family History. If you are in search of your long lost relatives etc. Then this site can help you!  Every week or as neccassary I will post here new websites I have found that helped me trace my family back, Generations.     Also if you are&nbsp;in need of desperate help and don\'t want to pay those&nbsp;high prices, just to find out a name or when someone was born then I can help. Just post or email who you are looking for with as many details you can find and&nbsp;I will get back to you as soon as possible.  '),(23920,3908022,'29,July,2004','2004-07-29 00:00:00','   I have a feeling that this is getting redundant. The only thing I ever write about is misery! If I ever write anything positive, it\'s without real emotion. I think I\'ve been so depressed that the only feelings I can express are the negative ones. But, I\'m going to try to write a positive post today.   Maroon 5 are eating their way into my heart and mouth(Which means that I love them and I sing their songs all the time... ). The lyrics are, as most, not very deep in most cases. But a few songs really have the depth that I look for in a song. I don\'t find that very often. I\'m also thinking about making a Gavin DeGraw CD.   Yes, I burn most of my CD\'s, not all of us are able to spend a fortune on CD\'s every month. I have clothes to buy.   Gavin DeGraw is a very good artist and I believe his songs to be positively amazing. His voice is not beautiful but he sounds secure of himself and relaxed. That is always a plus... Relaxation, and not just in music.   Another positive thing is that working really makes me feel good about myself... (partly, partly not.)  I feel that I  am  doing  something worth while . When I feed those people, make them laugh and make them food,&nbsp;I just feel that it means something. That I, for once, is doung something for someone else. But the fact that I\'m making money of this, of course, takes away part of that pride.   The weekend\'s coming up and I can\'t wait to kick back and relax.   '),(23921,3908022,'26,July,2004','2004-07-26 00:00:00','   OK, so there are three different departments. And there were three summer workers, including myself. And guess who gets to work in the dementia ward? ME! Who else, me and my fucking bad luck?  It\'s not so bad, though. Not really. I mean, the things I get to do are not disgusting or actually hard in anyway at all. Actually, all I get to do is make beds, make porridge, sandwiches, water the flowers, feed the cat and the fishes, do the dishes and feed one of the people who live there. The staff is really nice to me and I\'m very well taken care of. The old people, though they all have some sort of mental illnesses, they are really nice and charming. That\'s not the problem either, and I have no problem whatsoever with feeding them or talking to them or anything.  No, the thing that makes me uncomfortable is on a whole other level.  It\'s that feeling I get, when I go in to Agnes\'(one of the ladies there, she\'s in a wheelchair. She can\'t walk, and can barely move at all, except her arms.) room. I go over to her window and spot the row of framed pictures. My eyes stop at her wedding picture. There\'s no actual way to know that it is her, but I assume so. The beautiful woman I the picture with the perfect smile, joy shining from her face. Her eyes were so alive and she was healthy. She was sane, she was young and unspoiled. And I go to the dinner room where Agnes now is sitting. I see the same eyes, instead of joy they seem dead. The spark in that picture is long gone.  That feeling is what I can\'t stand. That feeling of seeing what they have become, understanding, having to realize that for some people that\'s the way of life.  Agnes is now the woman who forgets what she\'s eating in between chews.   Signe is another old woman there. She always repeats anything you tell her. Sometimes she answers, but not often. Her husband came to visit today, Signe didn\'t even acknowledge his presence. After a while, she noticed who he was and said \"Are we going home soon?\".  As if they were both going home and as if everything was normal again.  I sat there next to her with tears in my eyes. I don\'t know how to do this. I need help. I need someone to help me through this. I can\'t do it by myself.  I just can\'t.   '),(23922,3908022,'25,July,2004','2004-07-25 00:00:00','   Tomorrow, I start my new job. It\'s at a retirement home and I\'m scared to death. I don\'t think I\'ve ever been this afraid of doing or saying the wrong thing. I can only imagine what I\'ll feel like the day before I start my new school. I want people to like me. I want to do a good job. I want to deserve the payment that I\'ll get. (No matter how small the amount of money is.) I need help, I think. I keep thinking that everything I heard at the introduction day just went in one ear and out the other. I feel like I\'ve forgotten the most basic things. And I don\'t know how to make porridge! That\'s minus points, right there. And I especially don\'t know how to make it at 7.30 am!! I\'m a walking dead at 7.30! How can anyone focus on anything that early in the morning. I\'m so scared.   '),(23923,3908022,'24,July,2004','2004-07-24 00:00:00','   Sooner or later everything is put in just that, perspective.  Everything&nbsp;changes  from  day to day . Perhaps even from hour to hour,   minute to minute  .   Opinions, opportunities, abilities and feelings. They all change. Far  too often , if you ask me. I\'m always different. And everywhere I go, it\'s different the next time I come. Everyone I meet is different the next time I meet them. That\'s the thing about being a kid.  Which I consider myself to be.   Everyone changes. We\'re all so easily influenced.   For example. One year ago, the only music I ever listen to was R&amp;B and Hiphop, sometimes soul. I hated everything else. Not because I actually disliked it, but because I was in that crowd. The stuck up, mean, wannabe black,\'I\'m better than you\' crowd. And trust me,  I wasn\'t particularly happy in that clique . I only had 3 friends that I considered real ones.   Now I have 10 friends like that. That I love with all my heart. I love all sorts of music (except for Country and Techno music). And I\'m  happier . It was a hard transition, though.  This past year has been   the best year of my life  , but it has also been the  hardest . I feel that I\'m so much closer to finding myself. I\'m not there yet, though.  But I\'ve come a long way. No matter how hard it was to face my feelings towards myself. Facing them was nothing compared to actually trying to beat them, though. I haven\'t beaten them yet, sadly. That\'s going to take some time. I need time to like myself. I have to like myself in order to really like others. I know that\'s some  stupid psycho babble bullshit  but I actually  believe it .   Another thing I believe is that every decision we make, we\'ve already decided. Every decision is decided for already, by ourselves. We always want one or the other. Doubt is just ourselves feeling guilty...  Did that make sense?  Who cares....? haha. I know what I meant, and I am the only person who has to get it.   Now I have to go back to reality.   My journey of finding myself  .   '),(23924,3908022,'23,July,2004','2004-07-23 00:00:00','   Then.    It\'s then.'),(23925,3908022,'20,July,2004','2004-07-20 00:00:00','   \"Keep smilin\'     keep shinin\'     Knowing you can always count on me, for sure'),(23926,3908022,'20,July,2004','2004-07-20 00:00:00','   So, I guess it started Saturday, when Yorda called and told me that if I wanted to I could stay the night with her, Matilda and Amanda at Amanda\'s house. And I wanted to, of course. So I packed my bag and went to Yorda\'s apartment to celebrate her birhtday. I know I\'ve already done that, but her mother has this thing every year where she bakes   a lot    of cookies and makes    a lot    of food. And Amanda, Matilda and I were invited to come. So we ate as much as we could and then we were all high on sugar and really funny. Or, we were just really easy to entertain... I\'m actually going with the latter.   &nbsp; So, then we bought some pizza and&nbsp;went to Amanda\'s home. She lives kind of&nbsp;in the&nbsp;countryside.&nbsp;There are cows and&nbsp;a really bad smell.  Anyhow, so when we got there, there was this group of friends of Amanda\'s. Martin, Justus, Andreas and Ida. They all seemed nice and&nbsp;we hung out with them and we had fun and so on.  But, suddenly, late,late,&nbsp;late at night. When we were all hanging out&nbsp;in the room where we were going to sleep.  The&nbsp;bizarreness started.  They started to touch eachother, in not friendy ways. And say really offending things to eachother, involving lots and lots of words involving sexual organs. So to&nbsp;speak. And then I realized, they&nbsp;had known eachother forever. And instead of forming a mental closeness, they had formed&nbsp;some form of an extreme&nbsp;physical closeness instead. I just thought it was grotesque. When they had gone,&nbsp;Amanda, Yorda, Matilda and I&nbsp;had a lot&nbsp;of fun though.&nbsp;It\'s not&nbsp;as if I\'m suddenly disgusted with Amanda, because she\'s not&nbsp;disgusting&nbsp;when she\'s with us.&nbsp;But I\'m not sure if I ever want to meet those people again. The morning after, when we&nbsp;were on the bus home, me, Matilda&nbsp;and Yorda,&nbsp;we discussed&nbsp;it all. And they\'d thought it was disturbing also. &nbsp; When I got in to&nbsp;the&nbsp;city, I went by bus directly&nbsp;to the next friend. Linn! She\'d been in Holland for 10 days,&nbsp;and I had really missed her. Because I was in Italy before and the day I came&nbsp;home she went&nbsp;away. So I hadn\'t seen her in more than 3 weeks. Which is very unusual for us. So we hung out, talked and she&nbsp;brought me some candy and bath pearls.  After that, we decided with our parents that she was going to&nbsp;sleep at my house for the next two nights.  &nbsp; And she did... hehe. There\'s not much to tell. We talked, walked and watched TV... and slept... a little. And my grandmother came by, and she\'s giving me a pearl necklace, so we went to the goldsmith to fix that. As it turns out, she decided to give me another necklace too. I\'m guessing she was in a giving mood or something. I like it. So today she went home and I\'ve been kicking back and relaxing ever since. It\'s good to be alone sometimes... no matter how much&nbsp;I love her. &nbsp;  Freedoom.   '),(23927,3908022,'17,July,2004','2004-07-17 00:00:00','   Why does everything have to be so damn complicated?  Is it really a part of growing up? Life getting more complicated? I don\'t get the point of that. I always complicate things even more, too. I always read way too much into everything. And I hate that about me. I can\'t just accept something and live with it. No, I have to analyze it over and over and over again. Until everyone\'s sick of me.  &nbsp; That totally sucks. &nbsp;   '),(23928,3908022,'16,July,2004','2004-07-16 00:00:00','   So, yesterday I kind of had a fall back. Depression.  I know that  I\'m a teenager and that that\'s what we do . We get depressed, and it\'s nothing that\'s unique about me. But it still hurts.  &nbsp;  I was sick. With depression last year and the beginning of this year. It started somewhere around September 2003 and ended in March 2004. But, really, it started before that. Actually I think it\'s been there for as long as I can remember.   The uncertainty. Never daring to trust anyone, at least not for real. I\'m very reserved as a person. Even though I talk openly about things. I don\'t open up to just anyone. There are a few select people that get to see who I really am. But, there are even fewer that I really, truly trust. They are the people&nbsp;I love to death.  &nbsp;   You know that thing  where you\'re supposed to fall back into someone else\'s arms when they\'re standing behind you? You\'re supposed to trust them?  Maybe you don\'t know about it...   Anyhow, I\'ve only dared to do that once. Once in my 16 years of living.  People always nagged me about it. But, I think I have a problem with trust. But, only one way of trust. I mean, I trust a few people with my feelings and my opinions. But, the real kind of trust. Where I\'d put my life in someone else\'s hands. That\'s the kind of trust I have a problem with.  How ever, the only time I ever did that fall back into someone\'s arms - thing was one time last year. With one of my best friends.  She\'s someone I really trust. Because she\'s so honest with me. And no matter how much honesty can hurt sometimes, there\'s nothing I appreciate more.  &nbsp;   I was sick.. Back to that.   I was sick. Depressed, doubting everyone who claimed that they loved me, hating myself. Truly hating. Sometimes I would stand in front of the mirror wishing I could cut the skin off of my face, without anyone noticing. And I think I would\'ve cut it off. If I didn\'t know that pealing your skin off gets you attention and not the good kind.  And that was the last thing I wanted at the time, attention.   &nbsp;  But, I cut my wrists. That\'s&nbsp;another one of those  \"All teenagers&nbsp;do that\"  things. And&nbsp;therefore&nbsp;I felt that it didn\'t matter so&nbsp;much if I did it too.  It felt so good. Putting the&nbsp;knife to my skin.&nbsp; Feeling it cut through my skin, all my pain just pouring out.   In the form of blood .  It sounds sick, I know that .  But that\'s how I felt . It was an outlet for my feelings.  I wanted to die. To leave everything behind. Let go of it all, not having to worry anymore. But my friends noticed that I was hurting myself. And they were crying and screaming at me not to do it. I hated seeing them like that. But I still couldn\'t really feel that it was my fault. Which&nbsp;I know it was.  &nbsp;  I also had eating disorders, weeks could go by with me just eating an apple a day. And I just wasn\'t hungry. That\'s what I said when people asked. It was logical to me. I wasn\'t hungry. So, why eat?  I had this thing I did: In the morning, I didn\'t eat breakfast, so mom was worried and asked me why. I said that I\'d buy&nbsp;breakfast and eat a huge lunch at school.  Of course I didn\'t buy breakfast! Where would I get the money to do that?  Then at lunch, my friends were worried about me not eating, so I said that I had a big breakfast and that I was going to eat snacks when I got home from school.  Then at dinner, mom\'s worried again, and I tell her that I ate really much at lunch.  It worked most of the time. And I didn\'t see that there was anything wrong with it. Which is crazy to me now.  &nbsp;  I&nbsp;got help. Or, rather my best friend found out and forced me to get help.  I talked to a counselor, and I really thought that I was better.&nbsp;She was proud of me&nbsp;and I felt happy again. Not happy all the time. But normal, the normal ups and downs. And I&nbsp;ate&nbsp;the right way, too.&nbsp; &nbsp;  Everything was normal,  until last night .  &nbsp;  Suddenly, I just found myself thinking: \"nothing changed! I\'m still as ugly as I ever was. I\'m still stupid. And what makes me think that my friends love me? What has&nbsp;changed? Nothing!!\"  &nbsp;  So&nbsp;I started thinking about all of that. And I cut myself.&nbsp;It just felt so amazing.&nbsp;Once again, my problems and all the hurt just&nbsp;went with the blood.  Not that&nbsp;it bled a&nbsp;lot, because it really didn\'t.  But it felt good.   &nbsp;  And now I don\'t know what to do.  &nbsp;  I don\'t know if it was a one time thing, last night.  &nbsp;  I hope so. Because I really can\'t go through that hell again.  I don\'t think I\'ll survive that again.    &nbsp;  I hate my self pitying.   &nbsp;&nbsp;  &nbsp;&nbsp;   &nbsp;&nbsp;&nbsp;    &nbsp;'),(23929,3908022,'15,July,2004','2004-07-15 00:00:00','   Today I bought Yorda her Birthday -lunch. I love doing things like that for her, cause she so genuinely appreciates it. She thinks it\'s so sweet. And I love that. I\'m celebrating her birthday with Amanda, Matilda, Filippa and maybe a girl named Ida on Saturday. That\'s going to be fun. Her mom always cooks for a thousand. :)  But anyways... As Yorda and I were walking down the streets of our city, I spot the guy that I have a major crush on. And he\'s smoking!!!!!!!!!!! That\'s so fucking repulsive!!!! GOD!!!! I can\'t even begin to describe how turned off I was by that! Geeez!  I can\'t even talk about this anymore.   '),(23930,3908022,'14,July,2004','2004-07-14 00:00:00','   Sometimes I can be such a complete ass!  Everyone can, but this time it was my turn.  You know who you are(Madda), and I\'m sorry! I\'m sorry that I\'m selfish some times. It\'s just that I really want to talk to me. And you not daring to speak to me feels like you don\'t want to. And you not even daring to talk to me makes it even further away that we will ever meet, you know? And I am so ready for that. I feel like I could see you tomorrow. Cause you really are one of my best friends to me. Not just someone I talk to here. You are so real to me. I just want to make it even more real. Cause that\'s just showing how much I care about you! You mean so much to me! And I\'m so sorry that I hurt you. But you hurt me to, even if you don\'t mean to. Just so you know.    Love you!   '),(23931,3908022,'13,July,2004','2004-07-13 00:00:00','   Today has been a good day.  A very good day , so far :) I called Linnea twice yesterday but she was out both times. So the phone woke me up today, she called. It was great to talk to her again, I really miss her. We decided to meet sometime next week. That\'s going to be great! Then I remembered that it\'s my friend  Yorda\'s birthday today . So I called her, and she found that very  sweet ! :D  Tomorrow I\'m taking her out for birthday lunch. I really miss her too. But not the same way. Cause she\'ll be in my class this fall. So I know that I\'ll see her a lot anyhow. But I still miss her.  After I talked to them I went to the tiny amusement park with my mother and the kids at the Kindergarden where she works. I helped out, cause I love kids. And I saw a  gorgeous  guy that works there. His name is  Olle  and he\'s one year older than me. He\'s so hot! :D  I smile just thinking about him...    After that we went to get ice cream! Let me tell yu, kids are not good at eating ice cream! By the time we were done, they had ice cream all over themselves. It\'s unbelievable.  *sighs*  Now I\'m home relaxing. Waiting for tomorrow to come.    I love my friends! :)   '),(23932,3908022,'13,July,2004','2004-07-13 00:00:00','   So today my parents were going to IKEA. And politely asked if I wanted to come. No, I did not want to come.  And why is that?  I asked myself later.  No idea! I have no idea! Instead of going shopping for things I actually need. I choose to sit here, alone.  How much fun is that on a scale from one to ten?     One and a half.     But at least today I\'m going to call Linnea. I promised to call her a week ago, when I came home from Italy. But that just didn\'t happen. Today it will, though. I\'ve got to talk to her. I feel like such a bad friend for not calling her. :(  And what can I do about my room? It\'s so ugly! :D I want to make changes. But there\'s to much wrong about it. I can\'t just change a little, I have to change alot. And I don\'t think I have the money to do that. So, maybe I shouldn\'t change it. I could just keep it ugly.  Yeah, I\'ll keep it ugly. Who cares?   I do.   '),(23933,3908022,'12,July,2004','2004-07-12 00:00:00','   urlLink    Me in April... extremely weird picture... :S&nbsp; urlLink   '),(23934,3908022,'12,July,2004','2004-07-12 00:00:00','   Dear God! I\'m so tired and my head is thumping. It\'s like I have a midget and his five giant sisters walking around up there!  And how much sense did that make? None! That\'s right. I\'m so sick of this.  So my grumpy mood started when I went to this concert. The object of my affection sings and plays the guitar. He\'s  20 years old ... I\'m 16. I\'ve never talked to him. He has the voice of an angel and he\'s amazing on guitar. He also seems to be so cool, he has such a cool style and he also has a girlfriend.  I first noticed him because he works at my favorite coffee shop. Whenever he serves me anything I\'m grinning from ear to ear. I\'m such a teenager! ARGH! I\'m so mad at myself for being so pathetic.  Anyways, so I went to that concert, and I drank alcohol for the first time in my life. (If you cout sips of champagne out.) So I was all tipsy and I drank beer, which I hate and my best friend smoked. Which was kind of hard for me to deal with cause I never thought of her like that. But most of all, I never thought of me like this. Drinking. It\'s not me. I\'m the responsible one, the smart one with plans and a future. I\'m not one of those girls.  But Saturday night,  I was .  And of course I got a bad conscience. So I told my parents that I drank. Of course I didn\'t dare to tell them exactly what I drank, so I made that up a little when they asked. But I just had to tell them, no matter how scared I was of doing just that. I\'m such a goody-goody! Now that\'s annoying me! God, there\'s no way to be with out me annoying myself! GOD!  And I\'m still confused about my choice of school for this fall. I chose to go to Media-school. Cause I want to be a journalist. But then I thought of going to a music-school cause I love singing.  And I don\'t know if I\'ve confused my dreams with what I should want. If maybe I\'m not so good at writing even though people tell me that. Maybe I\'m not meant to do any of it. Maybe I\'m supposed to be a counselor or a teacher.  What do I know!?   I\'m extremely confused, and my head aches.   This is it for me. I hope no one reads this, for their own sake.   *sighs*    Elin       Read the first chapter of my  urlLink story ... and please review! I forgot to put in an Author\'s note.. I\'ll do that in the next chapter!   '),(23935,3908022,'11,August,2004','2004-08-11 00:00:00','   Okay, I don\'t care how ditzy this sounds, but,  not even shopping cheered me up ! It usually gives me a feeling of completeness for a very short while, which usually makes me forget about my problems for a very short while. Today, it didn\'t. On one hand, that might be because I didn\'t actually shop. I ordered. But still. And, plus, I ordered shoes that won\'t go with a lot of my clothes so now I have to buy clothes to go with the shoes I just bought. Which means that I have to spend more of the money that I\'ve earned. Which I do not want to do. Yesterday I got the pearl necklace from my grandmother. It turns out that that necklace is worth like $400! I\'m a little freaked out about that... I don\'t have any necklace that expensive. (Which proves that I\'m not rich...) And I have no idea when to wear it.. :S  One of the things cheering me up, though, is reading all the amazing stories at:  urlLink http://trory.de-nile.net  They are great.!!!  And my Uggs are out of stock!!!!! :\'( I\'m so bummed about that. But they\'ve been out of stock before, so hopefully they\'ll come back again.  Over- and - out.   '),(23936,3908022,'09,August,2004','2004-08-09 00:00:00','   This is one of those days.  One of the days I hate the most.  It is the days when I feel no purpose. Today, I felt no purpose. I feel like my life is so completely unneccesary. I feel so alone and unwanted. Maybe I am.  I can\'t really know that.  I have moments when it feels like everything happens for a reason. So, I guess the negative feelings have reasons too. Only, I don\'t want to see those reasons. Tonight I really feel like taking my life... of course I won\'t, though. I never do. But, the worst thing is that no one really cares. No one really cares if I\'m alive or not. And even if they do, I can guarantee that they\'d get over my death fast. I\'m such an avrage person. I blend in. My looks blend in, my personality blends in. I don\'t stand out. At all.  I wish I found my person. The one that\'d be there always, always. And so many times, I\'ve lead my self to believe that I have found that person. But I seem to be wrong each time.  I wish I stood out. I wish I was noticed.  I wish I was well liked.  I wish I was loveable.  I wish I was...    I wish.   '),(23937,3550869,'28,June,2004','2004-06-28 00:00:00','    I\'m not going to go into an entire enormity into how many partying I did over the weekend, mainly because I can\'t fill all the details, but I should state that I\'ve not done so much partying in my life. Bring it all on again!  Friday night was the first night. Having nursed my sore head following England\'s disappointing defeat, I slept the day away, and took up an invitation to a evening with Teddy, Rui and Jonathan who were throwing a party for their respective banks. Reputedly 120 bankers from 4 banks in one room. Plus a few other non-bankers, strangers and a clearly out-of-place management consultant or two in the mix as well. 120 bankers in the same room - not often that happens! Anyway, after a few(!) drinks, we went of to another club at around 1am and danced the night away - well most of it, as some people had to return to the office the next morning. I love being on holiday to see this!  Saturday, for me though, was a trip to Shenzhen to meet up with my good old mate, Peter. He managed to choose the best place to meet him, the lobby of the Shangri-La Hotel, and kindly provided me with the pleasure to gawp at a few supermodels who walked in. I noticed I wasn\'t the only one gawping at the near 6 ft Chinese babes...  But eye-candy aside, there were more physical delectations to enjoy later. First, my first ever visit the driving range to hit a few golf balls. Now, the only golf I\'ve played before was at Blenheim Palace on their miniature putting range (which I must admit I wasn\'t great at and even Ling had beaten me on the first three holes there), but this time, I was determined to give the ball a real thwack. Not surprisingly, you\'d see me miss my first swing. And then the second. But I did manage to hit it the third time, and it was ok. Later after a few more hits, skims and thuds (well, they didn\'t really go far), I did manage to hit a few fairly sweetly, managing around 130 yds on a 7 iron. Don\'t know if that\'s any good or not, but it was a nice feeling.  After swinging a club, it was dinner and then time for swinging IN a club. Well, less swinging but a lot more dancing in a lovely place called Casa de Michelle. Some great music, some cold drinks, some more dancing... It was great evening!  Sunday, though, was a much more relaxed affair. As part of APS\'s 5th Anniversary celebrations, we borrowed a boat from a bank and put a load of people on it and set sail in the direction of Sai Kung. Most people went for some watersports - from whoo-hoos from Wesley on the banana boat, to a determined, statuesque Rui on the waterskis - but others just enjoyed the scenery and sunbathed. (And some brought their work, Jonathan!).  Sorry to those who weren\'t able to come along for whichever reason (the most novel excuse award goes to Paloma), but we\'ll organise something like this again, I\'m sure! Thanks to Rui and Marianne for doing a lot of the hard work this weekend too!  And so three parties in 3 days. Not bad, not bad at all... Now for some sleep... '),(23938,3550869,'24,June,2004','2004-06-24 00:00:00','    * Up late for a painful loss *   It\'s 5.30 here, HK time. England have lost. Heads hung low down at the Champs bar downstairs as England lose 6-5 on penalties.  It\'s always been hard for me to watch England. I never watch the qualifying games because they usually involve an excrutiating period when England take the lead and you wonder if they can hold on, or they go behind and doubt they can claw it back.  Well, tonight was no different. Three minutes gone, Michael Owen scores. For the next 107 minutes, England defend for their lives. First Postiga scores, then a scorcher from Rui Costa in extra time. England out? No, Lampard saves England, but only for a dreaded penalty competition.  Now if David Seaman was in goal, I\'d have a little bit of confidence. David James never got close. Our hero, another David, Sir Beckham misses - his third penalty miss for England in a row. He looked back at the penalty spot and the remains of the cloud of dust that came up when he scuffed the ground. Doesn\'t matter. He missed.  Rui Costa may have missed, but Darius Vassel\'s miss was decisive. Ricardo, the Portugal keeper, takes the final penalty. Flies by David James. All over.  5.44am and on Portugese television on cable in my room, there is a carnival going on. It could have been England\'s carnival. Dawn has broken over Hong Kong. Time to go to bed. Sadly, I can\'t wake up later thinking this was all *just* a nightmare... '),(23939,3550869,'23,June,2004','2004-06-23 00:00:00','    It\'s 3 days into my holiday and apologies to those who expect my blog not to be so updated. But I have my trusty laptop with me and about 5 minutes, so why not add to it?    * Stanley and the Dragons *  Well Tuesday was Tuen Ng, or, for the Dragon Boat Festival, which takes places in several places all over Hong Kong (and indeed much of China). The festival I went to was the Stanley Dragon Boat Festival, a place more renowned for its fantastic market rather than a very large bunch of corporates descending on the town to participate in furious paddling for a minute and a half, all throughout the day!  Anyway, my role was to simply watch and sample the lovely corporate hospitality on board some luxury yachts that all the corporate big guns seem to have. Moreover, they\'re all tied together, so in a unwritten gentleman\'s rule, one can hop between each of them and say hello (and do lots more) to mates in other boats. Or you could just keep going until you reach the end.  Of course, some boats had a much more party-like atmosphere than other. The UBS boat had Senegalese drummers beating a tune, whilst next door, on a boat where they shouted Liechtenstein out every 5 minutes (I still haven\'t worked out who they were representing!), they had a pro DJ with some dirty dancers on deck. As the various parties went on, it almost felt like a dream, with so many bikini babes on the various boats. It dawned on me that maybe all those beautiful people you\'d expect on my flight over (if the adverts are to be believed) were simply here in Hong Kong in the first place.  One thing though, I didn\'t bump into lots of (or even any) celebrities (I\'m still holding out on bumping into Karena Lam Kar Yan - of Tiramisu, July Rhapsody and Floating Landscape fame - and asking her to marry me at the first opportunity...), but I did bump into lots of friends (even better of course). As well as my host, Teddy, Rui and Paloma were there putting bits of wood to water too, and Janet was also there to watch. And old friends from eras ago popped up (Roy Lam, nice to see you again). All in all, it was a thoroughly enjoyable day, even if it was a little bumped by jet lag and watching England win!    * The Swankiest Visa Section in the World *  As I\'m planning to pop over to Shenzhen to see me ol\' pal and partner in crime, Peter, I had to get myself a visa at, wait for it, the China Resources Building in Wan Chai. I\'m not sure what other resources they have in that building, but I hope they\'ve got enough security looking after the goldmines, oil rigs and the environmental resources in that building. I mean, who knows what is kept safe at the bottom of that building?  Anyway, it was also the nicest Visa Section of any Embassy/Consulate/Delegate I\'ve ever visited. Whilst most places I\'ve seen are basically the basement of any embassy what has second hand furniture that has only been cleaned once since the 1960s when they bought it, The CRB has its visa section on the 7th floor, and it\'s all brand new. In fact it was more akin to waiting at a new airport.  But even with their advanced electronic ticketing system flashing up numbers of which counter to go to, one thing doesn\'t change, and that\'s the queues. I had to wait 2 and a half hours, through which I probably had read every word in my passport a few dozen times.  But I seemed to be one of those people who filled in their forms right because in the end (apart for a 30 second dash to make a photocopy, I was in and out in a minute. But I got my visa all right, so that\'s just fine!    * Chilled, on the Rocks *  Now today was supposed to be one of my two days of chilling, so I started off by going shopping as soon as I had finished trying to get my visa. I also had thought about seeing a HK film, but too my disappointment, not many are on. Super Model was one that\'s on (featuring the lovely Karena... yum), but it\'s not the type of film I felt like seeing. Oh well, I\'ll see it soon on DVD anyway.  So instead, I decided to grab myself some DVDs and watched Love On The Rocks (Louis Koo, Gigi Leung). Unfortunately, it wasn\'t the greatest movie in the World, in fact it doesn\'t even come close to average. And that\'s coming from a guy who wants to marry Gigi if he bumps into her, but won\'t because Ekin\'s got her and Karena\'s got my attention now...  Anyway, Thursday is going to be another chilled day, followed by some partying on Friday. Can\'t wait! '),(23940,3550869,'21,June,2004','2004-06-21 00:00:00','    Well, England has just won its match against Croatia in Euro 2004 by a 4-2 margin. It\'s currently 4.45am in the morning out here, but I\'m glad I woke up to watch the match down at the Champs bar in the basement of my hotel. And my fantasy football team didn\'t do too badly this evening...  My flight to HK wasn\'t very eventful (some might say thankfully). I flew with an airline whose advert for \'Upper Class\' showed ugly people getting on the plane and  beautiful people getting off. Well my first impression when waiting in the departure lounge with my fellow passengers was certainly that I hoped the adverts were true...  But anyway, it was one of the most pleasant flights I\'ve taken, especially with \'V\'s  entertainment on demand. A chance to see The Last Samurai in amongst the sleepless attempts at sleep was certainly one of my highlights.  Anyway, after landing and meeting up with Teddy, I\'m now happy at 5am in the morning that my team won. And that Paul Scholes scored. And that Thierry Henry scored twice. My fantasy football team is in dreamland... '),(23941,3550869,'19,June,2004','2004-06-19 00:00:00','    Well, this is my first post on Blogger. Whilst I have my own website ( urlLink http://www.lestmak.com/ ), this will now be the main store for my blogs, as it is a little easier to update than my own website. Nonetheless, I\'ll be continually updating my main site once in a while and I\'ll make sure you hear about it here. But in the meantime a few comments:   Today marks the 5th anniversary of the founding of the Oxford University Asia-Pacific Society, and that is the reason for my impending departure to Hong Kong. As well as meeting up with some old mates, some of whom I\'ve missed for the last 3 years, I\'m looking forward to catching up with most of them. But the APS 5th Anniversary Boat Party in Hong Kong Harbour will be the epitome of the celebrations. I look forward to it!  However, going to Hong Kong does put me in a slightly awkward position. With Euro 2004 happening right now, I\'m going to either stay up late or miss England\'s matches altogether. Mind you, if England do make the quarter-finals, I will join the ex-pats in the bars and cheering my team on. Of course.  But it will be a fun packed 10 days out there. I plan to see friends paddling away in the Dragonboat Festival (22 June), and catch up on some HK movies. Ah, the fun! '),(23942,3550869,'29,July,2004','2004-07-29 00:00:00','    It\'s a long way to Niagara Falls and back to Boston, you know. A map told me it was around 462 miles to Buffalo, the nearest city to the Falls&nbsp;that\'s still on the US side. But we thought we\'d try to go anyway. So on a Friday afternoon (thanks to L.E.K.\'s summer hours in the US), Annie, Elaine (a friend of Annie\'s) and I set off on the long drive ahead to Buffalo...    &nbsp;   &nbsp;   * OFF TO SEE A BUFFALO *       &nbsp;  Now I\'ve never driven on the right had side of the road. Nor had I driven a car larger than a small hatchback, but this time, I had the chance to drive a beige Chevvie across two US states. Tricky at first (I couldn\'t really feel how fast the car was accelerating and more worrying, how fast it was under braking), I managed to drive down to Buffalo without too many mishaps. Except we got lost on the way in Albany. Twice.    &nbsp;  After some extra speed (I admit, I did surpass expectations, if you know what I mean...), we did make it into Buffalo, or rather its suburbs late, at night. An early morning rise come about as we seek what we came a long long way for.    &nbsp;   &nbsp;   * NIAGARA FALLS *     &nbsp;  The falls are one of the most amazing sights one would ever see. Sitting on the River Niagara and straddling between US and Canada, the falls are actuually two. The American Falls are more of a traditional straight edge waterfall, whilst the Horseshoe Falls are as the name suggests and curved. Due to the shape of the falls (which are probably around 240 degrees around a circle, the water falls down and a mass of mist arises from the pool below, to a height probably around 20 feet above the top of the falls.    &nbsp;  A series of boats&nbsp;( The Maid of The Mist ) go down to the base of each of the falls, taking tourists and soaking them. But if that wasn\'t enough wetting, the  Cave of the Winds  allows you to travel by elevator to the base of the American Falls. Four platforms can be climbed, the third of which is called  The Hurricane  and an apt name it is too. You get totally drenched!    &nbsp;  The American Falls are a fantastic sight at night, but the Horseshoe Falls are more of a disappointment, mainly because the mist obscures much of what you can see of the falls.    &nbsp;   &nbsp;   * WHERE HAVE ALL THE BUFFALOS GONE? *     &nbsp;  In the middle of our day and night visits to the Falls, we decided to go into Buffalo for some Buffalo wings, of course. Having made it into Buffalo, though, it was very much a ghost town. Barely a soul was in sight, no shops were open, and we had to travel a little further uptown to find a bar serving the said wings.    &nbsp;  But it was amazing that the town could be so empty. Where do all the kids and teenagers go? One person suggested we go to the Galleria Mall close to our hotel, which we did, but still not many people were there? So what do Buffalos do on a weekend?    &nbsp;   &nbsp;   * ANNIE\'S WEDDING LOCATION *     &nbsp;  Annie\'s found her wedding location. Our journey back from Buffalo was a more relaxed affair. Taking a detour from the I-90 highway, we decided to take US route 20 to visit the Finger Lakes. Our first stop was at Canandaigua, a little town on the edge of the lake of the same name. A beautiful lakeshore greets you with gazebo that reaches out into the water. Now Annie has picked a place, all she has to do is get married!    The rest of our journey took us through the Finger Lakes region and a few extra sights to see. It was a very good trip and the 1,050 miles were worth it. Not a bad weekend at all!    &nbsp;   &nbsp; '),(23943,3550869,'19,July,2004','2004-07-19 00:00:00','    Well, it\'s now a week since I\'ve been in Boston and I have to say, it\'s all been great so far. Well, except for the weather for the most part of the week, although I have been informed by my pod back home in L.E.K.\'s London office that I&nbsp;  didn\'t  bring the weather with me to London as it\'s still there, and just as gloomy as I left it. Well the weather here in Boston has been fairly cool and wet, except for Saturday, when it turned out to be pleasant, fortunately...   &nbsp;  &nbsp;    * A VISIT TO ROCKPORT *      It was on Saturday that I went out to visit Rockport, a town about an hour\'s train ride outside of Boston. Situated on the coast and part of Cape Ann, it is one of the most picturesque seaside towns you\'d ever come across. Not only for its beaches, but it has become a centre for artists and there are literally scores of art galleries and shops on the various streets there.    It is also a very photogenic place, which is why, I guess, artists love the place. There\'s a very well used harbour, and boats coming in and out all day. Unfortunately, though, despite my best attempts, I feel my own photos didn\'t come out as I expected them to. But a learning process anyway...    I also popped down to Gloucester when I had run out of things to do in Rockport. Gloucester is famous for the setting of the true story of  The Perfect Storm , made famous by the film starring George Clooney (a dire film in my own opinion, I might add). The list of names of those that died are listed in the Town Hall, although I didn\'t have a chance to see it.    There isn\'t much to see in Gloucester, to be honest, but the Maritime Heritage Museum on the coast is a great place to go. There, they demonstrate and actually build and maintain boats and you can see how they do it.      * KRISTINA\'S PARTY*     On my return from Rockport and Gloucester, I joined Kristina, a colleague, and her chums for a belated birthday party (well, that was the excuse, anyway). As well as grooving to a few cheesy songs (I remember ABBA\'s Dancing Queen some time during the music), we proceeded to an Irish pub somewhere in Boston to dance away the night. It was good fun. Thanks Kristina!     &nbsp;    * ANOTHER WEEK AHEAD *   Well, Week 2 of 5 coming up. And some endurance will be needed at the end of it! Anyway, I\'ll say&nbsp;more next time I write...   &nbsp; '),(23944,3550869,'14,July,2004','2004-07-14 00:00:00','    Itâ€™s the end of my first working day in Boston, and it feels very different to what things were like in London. My commute is 5 minutes instead of 50, I have a view to die for from the 16th floor of a skyscraper, and I have Quincy Market, one of my most favourite places in the world at my doorstep...  Itâ€™s been great so far. A chance to meet up with Annie when I arrived, it almost seemed a little surreal that only 4 months or so ago, we were both still in London. Even more so was Sunday and a lunch with Pak Wing. And I must say, he hasnâ€™t changed a bit. At least not his accent! Sunday evening, though, was the highlight of my stay so far. The Commonwealth Shakespeare Company put on a free production of Much Ado About Nothing on at the Boston Common. It was a marvellous performance and I enjoyed it very much. Only odd thing was the American accent in a Shakespeare play. Certainly not what I expected.  But today the rains have started to fall and it looks like Iâ€™ve brought the London weather with me. Oh well, you canâ€™t win all the time. Anyway, two working days gone and just the remainder of 5 weeks to go. Iâ€™ve enjoyed it so far... '),(23945,3550869,'06,July,2004','2004-07-06 00:00:00','    Imagine my surprise. I get home from my holiday to see an e-mail from Helen, our allocations manager in the office. Call up, find out they\'d like to send me to Boston for a month to work on a case. Didn\'t take me long to say yes!  So I\'m off to Boston this coming weekend. Superb! Can\'t wait. So from seeing long lost friends in HK, I\'m off to see more long lost friends out west! Pak Wing and Annie, I\'m coming your way!  Anyway, must keep my excitement in check... :) '),(23946,3550869,'04,July,2004','2004-07-04 00:00:00','    I\'m back home. And missing Hong Kong and Shenzhen already. I wanna go back...    * WHAT WAS IN GUANGZHOU? *   Tuesday saw me visit Guangzhou with Peter Wu, who assures me that the trains in China run on time. Well, maybe it\'s the L.E.K. blood in me (L.E.K. works a lot on railway projects in the UK, don\'t we, Alison? :P), but our train out from Shenzhen to Guangzhou ran slowly with delays. Nonetheless, once we arrived in Guangzhou, Peter treated me to another surprise - McDonald\'s!  Now, it had been installed on me by Lou, my tour guide with Contiki on my European tour with Ling three years ago, that we weren\'t to ever visit the home of \"Satan\'s Clown\" whilst on holiday, except to use the facilities (because they\'re free!), so I was shocked. But I had deprived Peter of his breakfast that morning with the early rise (though bear in mind I had to get up earlier than him to get to Shenzhen in the first place!), and you can\'t deprive a FAF of his breakfast. (FAF = Fat African / American Flatmate, the affectionate names he and his flatmate, Nick, give each other...). Anyway, not to be tempted into a burger, I settled for a 2 kwai ice cream!  It was a sweltering day in Guangzhou reaching 35C with unbearable humidity, and I honestly can\'t remember too much of where we visited! We had visited a shrine, a burial ground, archaeological digs on Beijing Lu and Sun Yat Sen\'s memorial (although we didn\'t go in), but it was simply too hot. By mid-afternoon, we had decided to go to the China Hotel By Marriott, supposedly the poshest hotel Guangzhou has to offer, and I sat there drinking a Pimms No. 1 in their lounge, as one must on a hot day...  By late afternoon, we had given up and decided to go back to Shenzhen. Then it struck. Our train was delayed and left Guangzhou half an hour late... China\'s trains run on time? I don\'t believe it! Or was it me? Can\'t be, never worked on a rail case in my life (fingers crossed, I don\'t!).    * PLEASE DANCE, BUT PLEASE DON\'T SING! *   After reaching Shenzhen, and dinner in the best Japanese noodle chain of restaurants in the world (Ajisen Ramen), Sunny and Lin joined us to go down to V-Bar at Shenzhen\'s poshest hotel, the Crowne Plaza. The impression going in was \"Great! There\'s live music!\", followed by a \"Uh-oh...\", followed by a \"Please stop singing!\". A band had been brought in to sing cheesy classics, all of them dressed in orange boiler suits, and singing very flat. A thought did cross my mind that they should have returned to Guantanamo Bay... I think our dear own Sunny should have taken over and wow-ed the crowd.  Nonetheless, their dancing was all right. And Sunny and I continued where as we did last Saturday, adding our own moves on the dancefloor.  We finally caved in after midnight though. A tired Peter, dreadful singing and Absolut-ly ginned, we all went home singing along to Sunny\'s Whitney Houston\'s Greatest Hits CD in the car...  The next morning, it was a lie in, before Sunny invited Peter and me off to the Raindrop CafÃ© for some false Western food... Well, it was meant to be Western style, but just like many of our Chinese imitations in the UK (not my dad\'s of course!), it\'s always inferior! Anyway, I think I foolishly accepted an oyster, which I\'m still feeling the effects of a few days later!  I was sorry to leave Shenzhen in the afternoon, particularly as I had so much fun there that I hadn\'t expected! Sunny, Lin, Stephanie and of course Peter had made it into a fun place for me, and to them, I\'m very grateful.  On my return to Hong Kong, I finally caught up with Peter Pulsford in the Post-1997 restaurant. Great place, and a bit more food of home, except this time, it was a little more authentic! Hope all is well Peter!    * FINAL DAY *   On my final day in Hong Kong, an early morning to pack up my suitcase (surprisingly still very light), and had lunch with Jonathan and Jennifer. After a bit of shopping afterwards with Jennifer, I managed to bump into the demonstrations that are now a part of the 1st July festivities. Some 200,000 people were there to make their views known on a variety of political issues, particularly, direct elections of the Chief Executive, and although crowds were down on last year, it was one of the most amazing sights I\'ve seen of a mass gathering. Thousands of people wearing their white T-shirts and demonstrating peacefully.  It was soon time to say farewell to a country I have come to love to visit. Although I probably could never live there, it is to me a fantastic holiday destination, where my friends live and lots of hustle and bustle. Aptly enough, the last person I saw before going home was Liana, on her way home from a business trip in Taipei, we bumped into each other in the airport. It won\'t be long before I return again... '),(23947,3550869,'13,August,2004','2004-08-13 00:00:00','    Well, it\'s almost time to return to the UK after 5 weeks in Boston. I\'ve had tremendous fun here. Thanks so much to everyone at L.E.K. Boston for making me welcome and for making my stay so enjoyable. Also, thanks of course to Pak Wing and Annie, fellow \"Londoners\" here in Boston.      Next stop will be a short stop in London, but I\'m being dispatched to Dover on Monday morning. Not ideal, but maybe some more exciting things to come! '),(23948,3550869,'11,August,2004','2004-08-11 00:00:00','    It was the last of my four weekends in New England, and my hub and spoke tour of the region, it was finally the long overdue trip to Cape Cod. Here\'s a brief overview of the outing...     * Jamie *   The train ride out of South Station wasn\'t extraordinary. Annie and I had made the train as expected and chugged along the commuter rail out to Kingston. A couple of quick games of Chess later (Annie\'s new past time), a win for me and a polite concession to her, we made it to the station in the middle of nowhere.  Or so we were told. But that wasn\'t a problem. Annie had met a local resident, Jamie, a few months ago on the Cape, and she had kindly let us stay with her on this trip. Jamie grew up on the Cape as a child, but had spent 40 years teaching first graders in various parts of the world including Fiji, the West Indies, but mostly in Hartford in Conneticut. Now retired to the Cape again, she had plenty of stories to tell over the years, which provided much interest for Annie and me over the weekend.       Jamie now lives in Sandwich, a smallish town just on the Cape and would be our base camp for the two days ahead.        * Beaches and P-town *    Cape Cod has some very beautiful beaches and we visited a few on the east coast of the Cape. A national park designated by John F Kennedy, the arm of the Cape is effectively a series of sand dunes, and therefore is a very fragile landscape. It amazed me how much conservation goes on, and looking around, there is a lot of green foliage, much of it is clearly artificially planted (plants that would not have grown in that environment) to prevent the dunes simply blowing away.  It was also quite poignant for me that we went to Marconi Beach. Slightly away from the beach, Guglielmo Marconi had planned to receive the first wireless message from across the Atlantic to Cornwall, UK to this point. However, he decided to receive the signal in Newfoundland instead, but Cape Cod still has the communication station he built and named the beach after him  The Cape has a series of beaches running on the east coast up towards Provincetown. Probably the gay capital of the US, Provincetown is eclectic, vibrant and fun-loving. Very much touristy (much like the rest of the cape), theatre, music and performing artists do their best to draw you to their stages and artists try their best to get you to bring your eyes into their galleries (and your wallets too, of course).      But whilst this wouldn\'t really be the best town for your kids, it is a playground for the grown-ups. Cosmopolitan in it\'s very nature - flags for dozens of nations fly on the buildings, not to mention many rainbow flags - it is welcoming for everyone and it simply feels like a town that knows how to enjoy itself without over-exertion.          * Martha\'s Vineyard *    There is a little island to the south-west of the Cape called Martha\'s Vineyard (or MV). Arriving in Oak Bluffs, on the north-east shore, it\'s reknowned for its historic gingerbread houses. The houses orginiate to a camp that was built every summer for teaching Christianity, but eventually instead of building tents every year, houses did the job instead. Each of the houses are uniquely and elaborately decorated, with current owners having the obligation to maintain their appearance. Ultimately, this means that the houses make up an idyllic model village, only they\'re real size and people live in them.      Edgartown, on the east coast, is MV\'s Capital. Whilst I have to admit that there isn\'t that much else to see there than the norm, several America\'s Cup boats had sailed into port and were taking part in some races. I have to admit, sailing is beginning to grab more of an interest for me and maybe I might take it up one day!          * Sunset Sandwich *       Before the final departure from the Cape, one last chance to see a sunset in Sandwich. A trip to the Boardwalk, a bridge on salt marshes in Sandwich harbour, and a couple of photos later, I had the photos of my trip.      A quick dinner and a rush to Kingston to catch the last train, it had turned out to be a beautiful weekend.                 Links:    urlLink  http://www.capecod.com/       Photos will appear in due course '),(23949,3550869,'06,August,2004','2004-08-06 00:00:00','    Timeâ€™s running out for me in Boston â€“ My return is imminent and I only have 10 days left to enjoy here    * VERMONT *   Every L.E.K. office has a Summer event and having missed my London one, I had the chance to go to Bostonâ€™s outing to Manchester, VT. It is, however, hardly like Manchester, England, a picturesque town at the base of a very tall hill (canâ€™t really call it a mountain). It does, however, have lots of shops like Manchester, as itâ€™s an outlet city as well.  Our stay was in the Equinox, one of the â€˜resortsâ€™ in Manchester. A large luxurious hotel, it had every activity you could imagine, and as soon as I arrived, I had pretty much made the decision that I was going to mountain bike around for an hour.  It had been some four years now since I mountain biked off-road, and probably even then, itâ€™s once or twice in six years since I was in Brittany with Nick Folkard and SPS. It therefore turned into a bit of a challenge. After the first 10 minutes, I had obviously outpaced myself on the uphill stretch and was completely exhausted and breathless. Fortunately, I had stumbled upon a gorgeous lake, and a few minutes rest was definitely worthwhile. The remainder of the ride, though, was still very tricky, but I managed it. And I had still the sore bottom yesterday to prove it.  The evening was a dinner at the hotel in a big tent they called the Taconic. Now, Iâ€™d managed to survive without mosquito bites in my 10 days in Hong Kong, but I wasnâ€™t so lucky this night. I have been bitten although theyâ€™re not as nasty as Iâ€™ve experienced. But ooh, the itch!    * DRIVING NOT A CHELSEA TRACTOR *   Monday was activity day, and mine was an off-road driving experience. I didnâ€™t get one of the Hummers, but I did get a Land Rover Discovery to tackle one a fairly fun course. Iâ€™d never driven a Chelsea Tractor before (so called in London because 4x4s are more common in the posh Chelsea part of the city than off road!), but at least I can claim my first time really was off-road. I managed to learn some tricks (left foot braking for example), and I have to say, it all did feel natural to me!    * FEW DAYS LEFT *   There are only a few days left and one weekend to go. And Iâ€™ve probably saved the best to last as I plan to go to Cape Cod. Iâ€™ve certainly been waiting to go for a very long time. Iâ€™m very much looking forward to it. '),(23950,3685554,'02,June,2004','2004-06-02 00:00:00','   Whenever I take a look at the amount of homework given to us by my teachers, I\'ll just literally faint and collaspe. I just do not have any idea how to complete it. Well I guess this is where perseverance and determination come in. One has to have those attributes. Surprise! Surprise! I went for a science seminar yesterday. Undoubtedly, it is very interesting. Stem cells, dolly, embryotic stem cells. Just think about that. Cool. Finally, Dr. Alan Colman, our speaker is a big fan of Manchester United! Y0z, Y0z! Brian, you there. I say y0z. Y0z! Y0z!   By Chief Executive of Corporate Services and CEO of Baboon Designs Corporation (BD. CO)   '),(23951,3685554,'31,May,2004','2004-05-31 00:00:00','   It\'s nothing much today. But at least I completed something that I\'m proud of. I\'ve managed to finish the Da Vinci Code. It\'s a fantastic book but I have to say it contradicts Christanity. Notably and claimed as a free thinker but in a way a christian I was quite surpriesed and irked by the way this book depicted Christ and about the Holy Grail. Who knows whether it is true or not but I being christian, I\'m quite against this ideology that Christ was not divine but just pure human. But then this book certainly could cliam its righteous position of #1 International Best Seller for its great contenta nd suspense. A thumbs up for it. Religiously, it\'s different.   By Cheif Executive of Corporate Services and CEO of Baboon Designs Corporation (BD. CO)   '),(23952,3685554,'31,May,2004','2004-05-31 00:00:00','   Mon resultat n\'est pas bien. A la vacance, je ne suis pas heureuse et maman gronde moi. Mon reultat n\'est pas bien parce que je n\'a Ã©tudie pas. A la vacance, je prend etudier, lire mais regarder des films. Ne pas regarder \"Harry Potter\", \"Day After Tomorrow>\". Lire, lire, maman parle moi. Mon resultat n\'est pas bien.   Au revoir et bon nuit. Aujourd\'hui dimanche. Mais, il est matin par consÃ©quent je parle toi \"Bonjour!\"   '),(23953,3685554,'31,May,2004','2004-05-31 00:00:00','   Welcome to Baboon Designs Corporation official website. A newly created ground for your eay access to Baboon Designs Corporation advertisements and designs. Today 30/5/2004 has been our first successful launch of Baboon Designs Corporation official website. Thank you for your support and we hope you will continue supporting us and our designs. Baboon Designs for everyone!   '),(23954,3685554,'11,June,2004','2004-06-11 00:00:00','   I know that I have not updated this site for quite some time but I have got to admit that I am pretty sick of camps. I had two consecutive camps. That is horrible. Think about it. Camp camps camps. One camp we woke up at 1.00 a.m. and did solo night. After that, we had to do 1500 push-ups. The other camp, we were sleeping on the ground which was covered with rocks and stones. Sandflies and mosquitoes constantly harassed us. When we were sleeping, that irritating sound of mosquitoes is unbearable. I\'m glad that all camps are over. Starting a new blog. Will see it soon.   By Chief Executive of Corporate Services and CEO of Baboon Designs Corporation (BD. CO)   '),(23955,3685554,'16,June,2004','2004-06-16 00:00:00','   Death. It occurs naturally to all humans, to all living things. It is a natural phenomenon so why should we be afraid of it. Recently, people have been more and more afraid of death, trying all ways to stop and halt death. Why? Because people fear death. They fear it.   	It is  very noticeable that the more civilized and developed society and country is, their people gradually grow more and more frightened of death. They cannot accept death. They think death as negative and should be avoided which is of course, sadly to them, not possible. On the other hand, people from villages or not so developed societies or comm00000; letter-spacing: 0px; font-size: 11px; -->                                    * Profile *      Name: Yiling  Age: 14   School: Anderson Secondary  Birthday: 1 May 1990   Etc, etc, etc... write anything you want.     * Links *     urlLink Link     urlLink Link     urlLink Link     urlLink Link     urlLink Link     urlLink Anime Skies        * Tagboard *    Have a tagboard here or anything you want. More pictures, links, your own photo, blah blah... by now you should know this column extends forever right?? =)                   *   *'),(23956,3575606,'30,June,2004','2004-06-30 00:00:00','    Wow..I just burped.SO...................yeah I had to take my cousin to court today.It was weird.I dunno why.I think cos of the weather.Yeah probably so.It was dark and rainy.We had to run in the rain.Only we walked for some reason.Anyhow..Im thinkin about going and watching a movie.I dunno which one tho.Hmmm...decisions, decisions.Id probably just end up watching Nick at Nite.DAMMIT for some reason I just want to smile right now.I cant help it I feel like Im um something maybe tripping.But Im not, I swear.I wish I could just stay awake forever.Or at least almost.Without being tired.Id sew some shit.Or clean my house.Or paint my walls.Stuff like that.Ya know?I dont know so..Ive always wanted to paint at least one wall in my room blood red.Yup..I havent tho. Awwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwww I hate liking boys so much. '),(23957,3575606,'24,June,2004','2004-06-24 00:00:00','    I remember that one time I did something bad.      haha..yeah right......Feeling like vomiting and laughing and choking and screwing and passing out and sick and hungry and thirsty and nervous and shaky and confident all at the same time...crazyiness......... '),(23958,3575606,'23,June,2004','2004-06-23 00:00:00','    I dunno sometimes I just dont know whats going on.I guess its ok sometimes but I dunno if right now is one of those times.I should probably stop being an idiot.Or maybe I just need some sleep.......thats probably it. '),(23959,3575606,'17,June,2004','2004-06-17 00:00:00','    How come everybodys gotta go away?For some reason everyone just disappears..for no reason.I dont get it.What did I do that was so bad?Its really fucking hard to type.I cut the shit out of my finger the other day.Yup got me 5 stitches.It hurt so bad.I cried and said the \"f\" word over and over and jumped up and down thinking it would stop and I kept bleeding everywhere.It looked like someone had gotten shot.made me very lightheaded...Oh well its patched up.And tho I cant move it really, it still somehow hurts like hell. '),(23960,3575606,'13,June,2004','2004-06-13 00:00:00','    I had to see the Raggedy Ann reject once again.Damn trailer trash whore.I had to make food for her cos she came to my work..I made it all nice and pretty and charged her a very obscene amount of money for a medium pepperoni pizza.She of course didnt pay the full price cos I didnt want to see her fucking ugly face so I had to have someone else check her out cos yeah...I hate her..I mean I stood close by there but I didnt have to speak to her or almost touch her hand while giving her change back.Other than that I was late to work.Was behind most of the day.Felt like puking all day long.Wanted to go to sleep.Couldnt.I wanted to talk to somebody but I didnt.I dropped some boiling hot spaghetti sauce on my leg and now I have this huge red burn mark down my leg..yeah..really sexy.Guys will be all over me.All cos of the big red mark on my leg.Shit I feel all itchy.Something really fucking awesome better happen tomorrow or at least sometime this week to make up for the shittyness Im feeling now!  Damn you bastard damn you to hell! '),(23961,3575606,'10,June,2004','2004-06-10 00:00:00','    Thinking gives me a headache.I hate to think, but then again, I love to think.Im thinking Im confused.And I think people make me think and make me confused and give me a headache.Im not totally sure what to say.I dont know what to think.Thinking of what to think blows.What if I should never think again?Im not gonna do that cos of reasons.I just dont get some people to the point that all I do is think about them and why the hell they do some of the things they do.Im not obsessive just curious.People make me curious just like they give me a headache.HMMMMMMMM.I dont understand how sometimes you have friends and sometimes you have no friends.One day people like you, the next day they hate you.What can ya do?Not a damn thing.One of the most uncool things in the world.Wow I just now almost fell asleep.I dont like it when people do that shit.Not fall asleep, but choose days to like people and dislike people.I know it happens cos I used to do it myself.But I was just stupid.Im not anymore.I just dont get it.Ive done something wrong.I dont know what I did.I really wish someone would explain it to me!Shit.Is it really that difficult?What the hell?Shit my head hurts. '),(23962,3575606,'09,June,2004','2004-06-09 00:00:00','    Driving in the rain.Rain on my head.Rain making my socks all wet.Rain making me wanna jump in puddles.Rain soaking my hair.Rain drenching my car seat.Rain filling potholes.Rain splashing on my window.Rain making people drive all jacked up.Rain causing car accidents.Rain dripping from my roof.Rain dousing my shirt.Rain making me wanna uh..............what? '),(23963,3575606,'08,June,2004','2004-06-08 00:00:00','    Sad?Happy?Depressed?Suicidal?Witty?Loved?Loving?Afraid?Lost?Alone?Ecstatic?Cheerful?Security?Groggy?Jolly?Lucky?Special?Melancholy?Rejected?Feared?Stupid?Beautiful?Giving?Down?Ugly?Dismal?Attached yet Unattached?Ive never felt anything good.  If I have its gone away very quickly.Im starting to think there is no such thing as \"happy\" \"loving\" \"cheerful\" \"lucky\".Theyre all words used as some sort of hmm..drug maybe,in my life so that I will keep thinking something great is going to happen but it doesnt.Its a joke.Theres nothing for me. '),(23964,3575606,'08,June,2004','2004-06-08 00:00:00','    Kinda NOT like a football player.Food hater.Sleep hating.Non white sock wearing.Hates custard and milk.Smells like a \"sunripened raspberry\".Bathes on a daily basis.Cupcake baking.Cute boy loving.No eyeliner-ing.Hair brushing.Unlabeled.Too much tv watching.Beer drinking.Old Skool.Snowman flasher.Obsessive make up buyer.Beer drinker.Cigarette smoking.Loud music listening.Loves purple.Harry Potter reading Chick.  I guess thats all uninteresting.Dull.Lifeless.Cold.Rejected.Hated.Inanimate.Dead.Always awake but forever asleep. '),(23965,3575606,'08,June,2004','2004-06-08 00:00:00','    Kinda like a football player.Stringy red hair.Unmatching.Ignorant.Jacked up lipstick covered teeth.Big thighed short skirt-wearing.Pregnant lady pushing.Masculine sounding.Cosmetically challenged.Carl\'s Jr working.\"Christian\".Jealous.Reeking of urine or maybe soured milk.Abnoxious.Loud.Whoring for cheap ciagrettes.Nasty hair tossing.\"Gangsta\"\"gothic\"\"hardcore\" wannabe.Uneducated.Too much eating.Horrible mothering.Lying through rotting teeth.Crooked nipple piercing.Crank snorting.Too much tequila drinking.Dirty old man loving.No bathing. Raunchy trailer Trash Raggedy Ann.  I wanna slice her open from head to toe with a butter knife rip out her insides cut up the rest of her body and send it to the people she holds dearest. Fuckin freak of nature.  Right..Id get my ass kicked. '),(23966,3575606,'30,July,2004','2004-07-30 00:00:00','     '),(23967,3575606,'29,July,2004','2004-07-29 00:00:00','    Second attempt at going to sleep..Failed.Of course I knew that was gonna happen.OHshit ya know what I just really dont get???Why do couples with AOL use each others screen names???Whats the fucking point?Are they that insecure?Or just bein lame?I think its really stupid.Something else that totally blows..Is that weve been having awesome sex weather the past few days and I cant get laid.It really pisses me off..I may have to start asking for volunteers for things like that.I dunno. '),(23968,3575606,'29,July,2004','2004-07-29 00:00:00','    haha..yeah Id slice someones ear off.So I guess I should have ate something earlier..but Im a loser and I didnt so now its been nearly two to three days.I still feel like Im gonna puke a little.Not so much.But enough to make me not want to eat.Oh well...most of today sucked my ass.Actually all of it.Im totally convinced everyone just really fucking hates me and Im kinda not likin them all that much either.I guess in the end it evens out.I dunno what Im talking about.Im becoming delierious.I should just go to sleep.Yup I should do that since SOMEBODY has a crappy computer and wont kick it like I said and SOMEBODY cant talk to me anymore cos they wont kick their computer...I have no one to talk to.Bummer.I attempted to talk to some \"OKC\" persons..basically they had nothing to say except they all suck.I refuse to be on a first name basis with THAT many people I dont actually talk to in \"real life\".Damn..Bambi came back.Again..I guess she doesnt want to stay at her OWN home..haha shes a goober.But whatever.So I gueess Ill go to my sleeping place now.Which would be my bed.Cos thats where I sleep.Yup.Sometimes if Im watching tv I may fall asleep on the couch or the floor..but usually my bed.With my most awesome Sailor Moon sheets and comforter and matching pillow case. '),(23969,3575606,'29,July,2004','2004-07-29 00:00:00','    Ok so..No..Im not pregnant but I sure fucking feel like I did then.I think I got a virus or something.I dunno.Ive felt like puking all day.Maybe Im just upset.Who knows?Not me.Fuck it.I took one of those phenegegrsomething or others for vomiting and I think it may have help a lot.Now I just have a throbbing headache. Nope..it cant ever go away.Bein sick sucks only I dont really think Im sick. just depressed.Kinda.Maybe not.Ok yeah Im sad.A little.Angry at someone possibly others.I wish my tummy would stop hurting it makes sleeping very uncomfortable.I just think some people should stop pissing me off.Yup.They should.I was having a pretty gosh darn good week too...and then this.Stupid motherfucker. '),(23970,3575606,'28,July,2004','2004-07-28 00:00:00','    I was right.Today is pretty shitty.So far anyhow.I wanted to go someplace but its raining and this place is kinda far and I dont feel like driving in the rain.Dammit..the viewing guide thing says Blue\'s Clues is supposed to be on but its not.Damn damn damn.Oh well Joe was probably gonna be on there anyways.I dont like Joe.Im kinda really fucking pissed.If I wasnt before I definately am now.No reason really.Yeah right..theres always something.What the hell?Not only is Blue\'s Clues not on like it said..But Rugrats are on and Ive seen this episode twice already today.Even more reason to be pissed.Now I have a tummy ache.The kind that makes your head hurt.And your knees ache.And makes you feel like puking.Puke until you cant puke anymore and your eyes are all watery and your throat is burning and vomit came out of your nose and it stinks and its all chunky in your nose.And so you have to gargle with water to get the burning feeling out of your throat.And try to blow the chunks out of your nose.And when youre done you just fucking hate everyone around you cos they all fucking suck. '),(23971,3575606,'27,July,2004','2004-07-27 00:00:00','    Unable to sleep.Nothing to do. Bein bored not talkin to anyone.Oooh yay someone to talk to.No one interesting so nevermind.Dammit I hate when people start talking and dont finish.I think this person may have add or something cos he says something about one thing and then moves onto another..kinda like..\"My roommate got a dvd\" then nothing..and then \"do you like Dead Kennedys?\"&nbsp; haha..over and over.Kinda cute actually..Hmm.So yeah I do like DK.uh huhhh.... '),(23972,3575606,'27,July,2004','2004-07-27 00:00:00','    I woke up way too early.Even tho I didnt have to.I just sat there and watched tv.Yup I watched Mommie Dearest.Faye Dunaway is so freaking ugly.Not just as Joan Crawford..just as Faye herself.Shed creep me out if she were my mom.Well I guess I wouldnt know anything different so it wouldnt matter.But if I were little and my friends mom looked like her..Id probably run and hide when she was around.I also watch Malibu\'s Most Wanted.Hahaha..it was on HBO or somechannel like that.I also have it on dvd.Its so fuckin dumb but it isnt.I dunno I think its swell.OHHHKAAAYY....SO I had a pretty super awesome day mostly.Unusual...difficult to explain.No problems tho.&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; :)&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;its a secret and I cant say.Tomorrow is going to suck tho I just know it.well maybe not and Id rather it didnt.I have no plans tho and no money and no friends so..I bet it will suck for the most part.Im actually really tired.I hope I am able to go to sleep that would rule.I just dont wanna dream about anything.I had the saddest dream last night.It sucked..it was so vivid.It depressed me.Its probably why I couldnt go back to sleep and woke up so fucking early in the morning.I didnt want to remember it but..oh well...I really do need to get some of that waterproof mascara remover before I pull out all of my eyelashes with this shitty turbo boost whatever. '),(23973,3575606,'26,July,2004','2004-07-26 00:00:00','    Yippee I had a good day today.Hmm I cant say why its a secret.But it was pretty gosh darn smack ass cool.Not for any specific reason either..well ok maybe one but its not THAT big of a deal.I love today.Today is my new best friend.Whoa! and its Monday.Mondays rule.Anyhow..it is waaaay tooohard for me to type right now the concentrating is driving me mad. '),(23974,3575606,'25,July,2004','2004-07-25 00:00:00','    I hate when when people I dont know touch me.Even if its an accident.I dont care I dont want them near me.People in wal mart are zombies.I hate all of them as well.I wish that kid would go to sleep.Shes bein bitchy.I wish I had about&nbsp;um.....80 bucks or so.Id be so happy.Im sad right now.But not cos I dont have 80 dollars.Itd just&nbsp;help a lot.I feel lonely...but I just want everyone to leave me alone.Yeah I know that doesnt make sense.But its&nbsp;what I want.I wish I could go away.Somewhere.I&nbsp;dunno where yet tho.Somewhere with people I dont know.Somewhere that my 70 year old neighbors wont go outside at fucking 5 am and mow their lawn while wearing spandex.Ew.Damn that old lady.I wonder &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; '),(23975,3575606,'23,July,2004','2004-07-23 00:00:00','    No...really..dude believe me when I say I totally fucking hate your dumbass and I want you to die. '),(23976,3575606,'23,July,2004','2004-07-23 00:00:00','    My right leg is hurting extremely bad and I dunno why.Oh well I took some Tylenol.I doubt that helps any tho.It was lightning a whole bunch earlier..I dunno bout now I havent looked outside in like 2 or 3 hours.I didnt watch any tv today.Haha.I usually sit and watch tv if Im not at work.And today I didnt.Maybe for like a minute then I left.Now its raining a bunch again.Uhhhmmmmm. '),(23977,3575606,'21,July,2004','2004-07-21 00:00:00','    Wow.People suck.Theyre boring.I know Im boring.And I suck so do those other people.I wanna talk to someone but no one wants to talk to me dammit.Oh well theyre some losers.Ha but I am too.Its almost the weekend.I dont care tho,weekends are no different than any other day to me.I dont think I even cared when I was in school, that it was near the weekend.I didnt go to school anyways so I guess thats why I didnt care.I mean,I was enrolled and all,I just never went.Cos school sucks my butt.Especially that school cos of the people and their stupid shit.Shit,I say.Id go back to jr high if I could.Fun times.I think.I dont really remember.Ugh phlegm is nasty.Ewwwwwwwwwwie.Hahaha..I think I had a slight cold or someshit..but Im bout over it.Tomorrow is going to be cool cos I can go spend all of my money on stupid stupid fucking stupid useless shit.But it will be fun.Yup.I wanna go somewhere RIGHT NOW!Eh no..I have to wait til at least 8     :(         Man oh man....I gotta find me a better way to waste time.Hmm...I dont know.I dont think there is anything.I could paint some suncatchers.Or make little animals out of pipe cleaners.Or cover things in rubber cement and wait for it to dry and then peel it off.Or read a book Ive read 30 times already.Or cut peices of paper into snowflakes.ooooh..fuuuunnn. '),(23978,3575606,'21,July,2004','2004-07-21 00:00:00','    Yup so Ive dont almost absolutely nothing all day.I meant to clean stuff but there were too many distractions.I went to the mall for a brief moment.Yup.Then I went to Wal Mart with my Mom..whoo hoo,cant get anymore exciting than that.I didnt get anything.Aw man lime jello rules.I love lime jello I wish I could eat it everday and nothing else...hmmm... I hate calculators...everytime I use one I fuck something up.Damn bastards.I guess I just push the wrong buttons ALL of the time.Maybe Im losing my reading abilities.Ya think?I just dont know how to work a calculator possibly......I dunno..whatever it is..I hate it. '),(23979,3575606,'20,July,2004','2004-07-20 00:00:00','    AAAAAAAAAAAAAAAAAGGGGGGGGGGGGGGHHHHHHHHHHH!!!!!!!!!!!!!!!!!!!!!!!!!  Id so rather be anywhere else right now than this house with all these shitty people.Yes theyre my family but..damn...I feel like I should pass out some Valiums.Xanax even.Did I spell that right?I dunno.My cousin has the loudest kid ever.Hes over a year old and all he does is walks around and screams and screams and cries and its very aggravating.Aside from all of that..I think my blowdryer just tried to electricute me.I was sparking pretty bad.It didnt smell or anything just sparks started coming out of it.So now I have to wait for my hair to air dry.Which sucks cos its going to take forever and a day.Maybe longer.Im not sure.It smells good tho.My hair I mean.Cos I washed it.Yeah.With shampoo.Some berry scented kind.Yeah.So there.Huh? '),(23980,3575606,'19,July,2004','2004-07-19 00:00:00','    Ok well I think Im just abou totally finished playing \"depressed loser self pitying drama queen\"   :)      YAY!!!! Were going to the park today.As soon as someone stops being a lazy ass and watching tv.Soooooooooo...Its my day off..Eh I hope anyways cos if not Imma be in a lot of trouble.No ones called yet but then again if I werent off today, I would probably just now be getting there..This hair color totally sucks.I tried to wash it out and it wont work but its like too bright to just dye over it and even if I could do that,it just wouldnt look rigt cos its too fucking bright!!!Fortunately I can blame it on someone else.Cos if I had picked it out myself,it wouldnt be so bright.Not this time anyhow.DAMMIT I just wanna drive someplace!Its too bad for me that the car I have is a piece of shit and feels like its going to fall apart at any moment.I want my own car back like right now.Thats not going to happen tho cos its gonna cost way too much to be fixed..Damned car radio thieves. '),(23981,3575606,'19,July,2004','2004-07-19 00:00:00','    Oh happy day!Ha maybe not.Not too bad tho.Well..actually..that one thing happened..yeah at work.I had to leave and come all the way home.Which wasnt too bad cos it was nice to not be at work.So I went back and thats about it.Then of course,I came home and went to sleep.When I woke up, I dyed my cousins hair.Yup put some blonde chunks in it.I dyed my hair too..it was supposed to be streaks but didnt turn out so well but..I dont care.My hair just sucks.And ICED TEA RULES!Yeah!Rock on tea! oOOOOOHHH Linda gave me my money back..well kinda.I gotta go pick it up from her tomorrow cos she hasnt cashed her check yet.I wasnt even really worried about it.Linda rules,I trust her.Im sure she will pay me back.I dont know why I keep mentioning it.Maybe cos shes my only friend and its the one thing other than work that I can think of right now.Oh how sad.So...I think I need about 75 bucks to for certain expenses..and everything will be good to go for quite a while..assuming neither Maya or myself gets sick. Ball and Chain?Isnt that a Social Distortion song? Im thinking it is. Ah well. '),(23982,3575606,'19,July,2004','2004-07-19 00:00:00','    I should stop going to sleep as soon as I get home from work cos I sleep til late and stay awake til early........with not a thing to do.Ack.Oh well.I have to be at work pretty darn early tomorrow.Its Monday! YAY Mondays rule.I dont feel like explaining right now..but they do and thats that.Today at work I hit my hand on something I forget what, and its bruised now.Heh.I was busy..not too busy tho which sucks.Uh ew there was just something sticky in my hair..I think it was parts of SOMEBODYs sucker.yucky.I am seriously going to catch my death in this house due to the freeezing temperature our ac is set at.Or something..Its so cold.My eyelids hurt to close.yay I get money today whoooooo hoooo!And Friday is pay day! Yippee skippeeeee..Karleighs birthday is Saturday I guess well have to get her an awesome present.I dunno what tho.Ill have to ask her mommy I guess.SOooooooooooooooooooo....I suppose Ill go to bed now even tho Im still not too tired. '),(23983,3575606,'18,July,2004','2004-07-18 00:00:00','    Somewhere between here and there I kinda got this strange feeling like I temporarily slipped out of reality.The sun was super bright.The sky and the clouds looked matte kinda like a painting.I was driving and it just felt really fake.Other cars on the road were moving like they were floating.Gliding maybe.I dunno.There were cows but they werent really moving at all, not that they are usually extremely active but the grass and little hills made it look like it wasnt real.I saw a few cars just parked on the side of the road.One car by the lake.Their stillness just seemed eerie.I didnt actually see anybody at all.Just other drivers as they drifted past me and I drifted past them.I didnt really look at them tho, I never do.Ever since that day with the guy with his pants off.....yeah..I didnt really look before then, just the one time..and BAM!&nbsp; nasty pervert..Anyhow.It was weird.I felt like something super cool happened or was just about to happen.Uh not still talking about the old pantless perv.But nothing happened I dont think.Maybe it did Im just tired or something.I dunno.I havent been outside since then..it might look bad.Not fake anymore.Back to being normal and not so interesting.I see it all of the time so I guess it gets boring..same road almost same cars same people..ya know........ '),(23984,3575606,'18,July,2004','2004-07-18 00:00:00','    Im bout to freeze to fucking death.Maybe not..but Im cold.Id like to have a hmmmmmm..nothing I guess..I was gonna say snowcone..but..yeah.I think I watch those LOTR movies way too much.I got the sound effects down and all.ooooooooooooooooh my hair is super soft.Yay detangler rules.Yup without it,my hair would be shit.I mean it sucks still but..it could be worse.Pickles are gross.So why did I just almost eat one?I dunno.Just seemed right.Ill puke if I do.So yes..I think I need a new car..one thats not automatic I hate driving those..THEY SUCK!I wish my car would work.Cos its a standard..Heeeeeyyy.maybe my sister will let me take hers to work tomorrow..It rules.I want one of my own.Yeah I suck.I know.Oh well.One day, I wont.If I still do, I dont really care all that much.AAAAAAAAAAAGGGGGGGGGGHHHHHHHHHHHHHH I am losing my mind I swear..Im not going to know anything in a few days. '),(23985,3575606,'17,July,2004','2004-07-17 00:00:00','    Aw someone should sing me a song.I dont care who.HAHAHAHA I dont think anyone has ever volunteered to sing for me.Oh thats a lie.This guy did...he used to make up songs about me.It was cool.Then it sucked.And that was that.He probably says all sorts of bad things about me now...Untrue bad things but hey..I guess I deserve it.Im so mean to him after all..I mean shit...I should have been a more pleasant pregnant person dammit!I should be thankful for him for totally ditching me...and for not ever coming to see his little girl or calling to see if shes ok or she needed anything..Hes too busy.Damn school theres so many hours a day you have to go.Never enough time to take care of anything important.Ha.What a fucking loser. '),(23986,3575606,'17,July,2004','2004-07-17 00:00:00','    So ends yet another glooorrrious day for Mel.Ha not so glorious.Not so interesting at all.I woke up went to work came home went to sleep and woke up and here I am again.Fun stuff man.I found this book this morning..it has like every single thing I ever did over a few weeks..its crazy.And its from jr high,which makes it even more crazy cos I dont remember half the stuff Im talking about.Brandi just IMed me.Probably to tell me more about her wedding and blah blah.Not that I dont care..I just dont care I guess..Not to be rude..but I could give a shit.I doubt Im even gonna go.I dont wanna take Maya.Shes mean out in public quiet places.I dont think Id be able to sit still too long myself..so..yeah I dont think Ill go.Not like its anytime soon..but yeah. '),(23987,3575606,'16,July,2004','2004-07-16 00:00:00','    I feel like I try too hard.Even tho it doesnt seem like it.Cos I dunno why..its just one of those really weird things that only happens to people like me. For instance..when I first opened my checking account...it took a whole month for them to actually send my checks and then I never got my checking card haha because the place that makes the cards \"ran out of plastic\".Yes.They ran out of plastic.Or any day now Im going to be the one dipshit that gets a ticket for running a stop sign in a parking lot at 3 am when there is absolutely no one else around..in a dark parking lot..just taking a short cut.Now I totally forget what I was thinking in the first place.Either way Im bout to piss my pants.Yes.Piss.Oh yes and of course the moment I decide to go pee..so does everyone else in the damn house.Piss..How unladylike of me.At least I didnt go into detail.Like saying it burned my eyes.Oooh yeah..thats what the guys really wanna hear..all about my trip to pee.Uh.....No..Ew..No dirty old men better Email me cos of that either ewwwwwwww. '),(23988,3575606,'16,July,2004','2004-07-16 00:00:00','    Fuck Full House.Fuck Roseanne.Damn addictive bitch keeps me up all night.All I do is sit and wait for it to be time to watch Roseanne.And Full House.Hmm..yeah.Nick at Nite rocks my socks man.Im thinking...nothing really..I just now almost choked on a gummy bear.Could have made someones day.Someone who doesnt like me.Its not hard to think of several people who dont like me.I dont care.I just want something to do.Im trying to avoid watching tv.theres nothing on there I havent seen before.Especially Nick at Nite cos all they do is show the same shit over and over and I already watch a few hours of that..so ..I pretty much have seen whats going to be on for the next 12-18 hours.I wish I couldnt stop smiling for a little while.But I dunno..Id have to get really high or something..So..I wont do that.I have to go to work in about 4 hours or so..Yup fun stuff.Hopefully I dont have too much to do cos I just feel like a lazy ass. '),(23989,3575606,'16,July,2004','2004-07-16 00:00:00','    I bet its all your fault I feel like puking.I actually know so.Just the mention of your name....makes me wanna blow chunks.But..whatever.  I need a dollar...so then Ill have 40 all together...instead of just 39...I dont want quarters or nickels or dimes or pennies..just a dollar bill.and I probably wont wanna spend it.Which would be a good thing cos I havent been keeping up with my money too well ever since..hmm..ever I guess.Anyhow...Im going to go take some really icky little pill for people who cant stop puking. '),(23990,3575606,'16,July,2004','2004-07-16 00:00:00','    I feel like Imma puke.Yeah all day.Well at least for the past 5 to 6 hours.I dont know why but I wish it would fucking stop.My eyes are getting so extremely heavy.I think Ill go to sleep sometime soon but once I tried, and I just kept feeling chunks in my throat.SO I got up and here I am.Again.Yes.With nothing to do feeling like shit.Oh wait maybe if I smile..it will go away.Or not.I just dont know.I wish I could just go to sleep and stay asleep for a long time.Maybe not a long time but a more than 2 hours.So I guess the average 8 hours would be a pretty long time for me.Oh well fuck sleep. '),(23991,3575606,'16,July,2004','2004-07-16 00:00:00','    Im pathetic.Yeah.I cant believe I did that.I dont know why I did it.Oh wait yes I do, its cos Im a stupid dumb fucker who nobody likes.So go ahead..make fun of me, its ok.I dont care.Why should I? '),(23992,3575606,'15,July,2004','2004-07-15 00:00:00','    Just thinking........of you.Somewhere away from me.Thinking of someone else.Of course.  I wish I could go to sleep and dream of something super cool.Its been so long since Ive had a great dream.Here lately theyve sucked so much that I havent remembered them any.I wish I could explain in more detail.So maybe you could know too just a little at least.Things might be easier.If I could just stop thinking of everything,I wouldnt have to feel anything.Id probably be more pleasant.I keep worrying.I cant eat.I cant sleep.I cant talk to anyone else.I cant dream of anyone else. I cant think of anyone else.Early in the morning, when Im watching my fan spin or when Im driving in my car going nowhere,Im thinking of you.Even when Im thinking of someone else..I think of what you might think or what it would be like if they were you.I think of you before I sleep, hoping to see you in my dreams.When I go anywhere I look at everyone I can see, cos I think maybe Ill get to see you again.When it rains I think of sleep and I think of you.In silence, I think of your voice.I think of you knowing its not so bad.I think of you not being so negative.Thinking of you makes me feel kinda giddy.I keep thinking of you coming over.Maybe you meant to call.You probably dont have the number, right?Its ok.Im not crazy.Really.Its not normal but still...You rule my world.Youre all I think of,yet I dont know you. '),(23993,3575606,'14,July,2004','2004-07-14 00:00:00','    Sometimes I feel like someone is standing on my chest.Just for a second.I bet Ive already said that but...It happens way too often.To where I can actually hear myself gasp for air.Its strange.The weird thing is, I know where its coming from.Another weird thing is that if I tell anyone they will either laugh for think Im insane.So..Ill just say nothing more.Its hard to explain anyhow.Hmmmmmm....Im so bored still.Im never going to be not bored.Well except for the next time I go to someplace.I dont have any money.And when I do get money, I have to get the car that isnt mine fixed..and save the left over money for the car that is mine to be fixed. My sisters went outside..I guess to smoke.Or maybe theyre meeting someone out there.I dunno they go out there late at night way too much.Sometimes Ill go out there, and they wont even be there...and the next day Ill ask where they were and they will say \"we were sitting right there\"       yeah right.Hookers. '),(23994,3575606,'14,July,2004','2004-07-14 00:00:00','    Ok not really.I hate wearing hats.Probably cos my head is so freaking big.My car sucks.Linda should give me some of my money back so I can fix the damn thing before it explodes or something.I dunno I just really hate it so much.Its not even my car.Ive spent 40 bucks buying gas for different cars Ive driven during the week..and what sucks is that I havent driven them several times.Maybe just once to work, or once to the store...stuff of that nature..its not like I was out partying and driving other peoples cars.Oh well tho.I feel so much like shit.Too much like shit.I need a hug :(    Thats never going to happen.I should just forget that.I should go see a movie.Ive been so bored.Yeah maybe Ill do that.Actually I probably wont.But Itd be cool to think I had something to do with my Friday.Bummer. '),(23995,3575606,'14,July,2004','2004-07-14 00:00:00','    Today sucked so much.I went to work this morning and I got there at about hmmmm 8:23 am..I did a bunch of stuff there..then I did a bunch of other stuff later.When it was time to go home, I didnt go home cos I told someone Id stay there and work for them to be nice..so I had an hour to do nothing although I did end up doing something..It sucked.SO when I finally got to leave work, I came straight home, took a shower, and went to sleep.And now its almost 4 am and Im wide awake with nothing to do or anyone to talk to...wait Im like that almost all of the time. '),(23996,3575606,'14,July,2004','2004-07-14 00:00:00','    Ive decided to learn how to properly iron my clothes.Yeah..and Ive also decided that I think Im gonna go marry me some dirty pervert of and old man (cos theyre the only ones interested in me in any way) and were gonna have about 5 kids.He better make a lot of money cos I plan on being one of them stay at home moms.So Im going to spend all day cleaning stuff up or cooking and taking care of my family.I doubt Ill get lonely or depressed then cos I have come to the conclusion that there is not any person anywhere who likes me in any way.Even just other chick friends..So Ill just stay home and do that thing I do for the rest of my life until I end up shooting myself in my fucking fat ugly head.Who knows?Well actually I know.I know I suck.I know I have some kind of special ability to repell every single person I ever meet.Kinda like OFF..only its not for bugs..its specially designed for people I like.Damn that dirty pervert to hell..I think Ill just go crawl in a hole someplace and die. '),(23997,3575606,'13,July,2004','2004-07-13 00:00:00','    Im tired.I should go to sleep.I know just as I go to my bed, i wont be tired anymore.So...I dunno..........I had to stay at work all day long today..and I mean all day.From morning til dark.Fun stuff. '),(23998,3575606,'11,July,2004','2004-07-11 00:00:00','    Goddamn..this is going to be the third time I attempt to do this..........but I guess its ok cos I dont got much else goin on. Ever.So I have plenty of time.I think I recently made the mistake of loaning a friend a not too small amount of money.Shes awesome tho but I dunno Im just worried something will come up and she wont be able to pay me back. Im sure itll be ok tho.Shes pretty responsible and shit like that.Oh dammit I keep coughing it totally sucks.Someday Id like to do something..I cant say its hard to explain so Im not even going to try to explain and make myself look like a fuckin idiot trying to.Hmm.Yet Ive done it anyhow.. I totally cleaned up my room today.I got rid of unnecessary things I found here and there and things that were just collecting dust.I had a rope light..a purple one..I tried to get rid of it but nobody likes it I guess.O well Ill probably end up throwing it away. '),(23999,3575606,'10,July,2004','2004-07-10 00:00:00','    How do you \"learn\" how to paint?I dont think painting is something anyone learns.I just think its one of those things you either can or cannot do.And that is it.Yeah I just saw this book and it made me think of that...so there ya go.I went to Wal Mart.Yes, Wal Mart.I got some detangler for my hair.It rules.It was a pretty quick trip..In and out.Except when we got in the car,Linda\'s door wouldnt shut.I guess its happened before or something cos she knew how to fix it.Kinda.Some lady stopped and helped her I had no idea how to go about getting the damn thing to stay closed anyone.Oh well.So after that I was home pretty quick.Eh.I dunno if it matters.I guess not.My throat feels all itchy.I better not get sick.Im gonna be so pissed if I do.I mean I know people get pissed about being sick but I may get violent.I went to the park today also.Nothing really exciting goin on there.So..yeah we just sat.I wonder...wonder wonder wonder......I guess.I dunno. '),(24000,3575606,'10,July,2004','2004-07-10 00:00:00','    Weirdo..Yes.Totally weird lookin at chicks asses and just being way too quiet.I dunno.I was so busy today.I got really far behind on things I shouldnt have gotten behind on.I think mostly cos I kept talking to people.Haha.Too much chit-chattin goin on.Ugh so now I have to go to work extra early so I can make up for the crappy job I did today.Its ok tho.I suck.On a Saturday night Im sitting here....doin nothin.Pretty much talkin to myself.Listening to my brother talk to himself.Sad.Yup.Im not sad tho just the though it saddenning is that a word?I dunno.Hmmmm....what to do?Nothing I should probably go to sleep but Im not too tired.Kinda wishing I could go ahead and go to work and get all that stuff done and then come home and go to sleep.eh probably not.I need some money.Not really I just got paid.I dont really need to spend money on everything cos I got most of the stuff I need.So.I guess I dont need money.Um yeah just forget I ever mentioned it.I wish I had something I could paint.Not like a picture but a wall or something like that maybe a chair.Painting stuff rox.I wish I could pain my room blood red.But....I cant..so..I guess I wont be doing that.My nose is all stuffy.I wish all this snot would just come out and leave me alone!Damn stuffy noses and sinus problems.Math totally rules I might screw up on it a lot but its just so fun to sit and add and divide and all that stuff..I like writing it out.I hate using a calculator unless Im in a pretty big hurry.I guess Im a nerd.I dunno..its not like I took some Honors Algebra Trig.classes or anything or ever will..I just think its fun.Aw Im so bored.Someone should come pick me up and take me far away.Yes I do have a car..but I dont wanna drive.I dont think I know anyone who would come pick me up even if it wasnt 1:15 am.So..I guess Ill just be stayin home then.Fiji water rules man.I dunno what it is about it but its just so awesome.Im trying to think think think..........and I know.........Nothing.Sometimes I feel Ive go to...........RUN AWAY..........Ive got to........GET AWAY....from the pain you driiive into the heeaarrrt of meee...the love we share seems to go nowhere and Ive lost my light for I toss and turn I cant sleep at night.        Huh?Sometimes I just wanna sing.So I did.And now its over.I dont get it...I just dont get it and I never will.I have some weird effect on people.They just totally like to ditch me.I just have some sort of appeal that makes people just wanna treat me like shit.Im forever getting ditched.All by my fucking self for no goddamn reason..if there is a reason Im sooooooooooo missing out on what it is..Id kinda like to know so that whatever the reason is, I could try to make it better.Ha..like anyone would actually tell me.Meanasses. '),(24001,3575606,'08,July,2004','2004-07-08 00:00:00','    Well.My mother is a psycho freak sometimes.Shes flippin out cos Im awake.I slept earlier.Im not tired.Yet..shes freaking out.I think shes just picking on me cos she doesnt like me all that much.Hmm..Im so bored.I wish I had someplace to go but I dont.I never do.Except for work.So I look forward to being there which is so totally sad.Now Im kinda pissed.But oh well.Ill get over it.I think I need to find myself one of them significant others.Yup those.well not all of those.But one of those.Yeah thats what I meant.Unfortunately nobody likes me so its going to be so difficult.Damn I just suck. '),(24002,3575606,'02,July,2004','2004-07-02 00:00:00','    So my computer is fucking up again.Piece of shit.I wonder if knitting is difficult.I think I should knit a sweater.Or something like that.Or maybe take up cross-stitching.My feet are cold.I wish I could die and come back as a pretty person.But I dunno.Its lightning outside.Way far that way.West. '),(24003,3575606,'01,July,2004','2004-07-01 00:00:00','    Goddamn I hate frogs. '),(24004,3575606,'06,August,2004','2004-08-06 00:00:00','    For some reason this has been all screwed up.I had a bunch of stuff to say too.Dammit to hell.SO anywho..I did get to see Shadows Fall on Monday.It was alright.Bunch of little kids there.The first two bands sucked giant donkey balls.Sweaty ones.I dunno....we should have been late.I started drinking like the minute I got home from work.It was so hard to drive.I kept wanting to stop and look around at things.Had to stay focused.Which was difficult.But I managed.I rock.Not really but shit my arm itches.All of a sudden out of nowhere the damn thing is burning Im trying.Aw my friend isnt here.I have no one to talk to once again.Damn.Soooo..what else?The Pillow seems to have gone into hiding.Havent seen her for a while.Just vanished.So I have no one to make fun of either.Yup.Fat bitch.Maybe shes burning in hell.Not likely tho.Shes Catholic.Someone totally fucked up my Twisted Metal disc.I was kinda upset.But I figured if I havent seen it in that long...its my own fault. '),(24005,3575606,'03,August,2004','2004-08-03 00:00:00','    well..today has started out as another shitty day.I think its all these people here theyre just pissing me off.way too many people and bratty kids.I think Ill run away tonight..I meant to last night, but I totally forgot.....so yeah maybe for sure tonight.Sure Ill have to come back tomorrow but at least I wont be around all these shitty people for a whole night. whoo hoo hoo.Damn.I guess I will go to work now. '),(24006,3575606,'01,August,2004','2004-08-01 00:00:00','    Man I just dont get it.I always pick shitty people to talk to.Or something and dammit I dunno I cant say.I mean I can but I wont cos just no reason.Yeah actually I have a reason Im just not gonna say.Im bummed.Kinda.Just about some certain issues.Issues that shouldnt matter but should..I cant explain how that is.I ate some Cheezits earlier..although I dislike cheese.Im probably gonna be sick.I cant wait til tomorrow.I dont even think Im going to be able to sleep tonight..Oh wait I never sleep at night anyhow.I wanna get drunk starting at like hmmmmmm 7 am. '),(24007,3575606,'01,August,2004','2004-08-01 00:00:00','    I think its working ok now.But so anyways I had said before..and it didnt show up..that yesterday was a pretty ok day..work actually went well and I made 25 bucks whoo hoo.And yeah other stuff and it was kinda strange but it was awesome.Today on the other hand totally sucked big donkey balls.I woke up late and so did someone else from work so we were really behind and busy and I thought I was going to lose my mind.I didnt.But then I came home and it was super hot in here and Maya was being a pest.It just really sucked.So we just took a pretty long nap and now I am bored.Shit.I hate when stuff like this happens.Ohhhhhhhhhh but its ok cos tomorrow is Monday AND Im goin to see SHADOWS FALL.Yay!Not only is it Monday but the first Monday of August yippeeee!Im still depressed tho and pissed about today and this morning and probably later and probably tomorrow morning and about Maya and my cousin and my friend and my car.Yup.Oh well what can I do '),(24008,3575606,'01,August,2004','2004-08-01 00:00:00','    this shit is pissin me off.I keep typing stuff and it never shows up and dammit Im mad. '),(24009,3575606,'01,August,2004','2004-08-01 00:00:00','    what the fuck is going on '),(24010,3715376,'30,June,2004','2004-06-30 00:00:00','   Well! Its been a few days since I have updated, sorry to all my die-hard yet apparently non-existant fans, but I have been a busy little beaver...   So like I said, on the weekend, I was shipped off to some friends of the family for the weekend, Gitte and Bent, they were both in their early fifties but they were sooo funny, and they have travelled like EVERYWHERE in the world!... ok not quite everywhere, but close enough...   So, I arrived there friday arvo and together the three of us cooked food, sat in the lounge room and drank wine while and spent half the night just talking togetgher which was very nice. Turns out that they both love Les Miserables, too, which was very nice. Which remindes me, I am still waiting for SOMEONE to send me over my copy... (hint hint mum) Anyway, just before we went off to bed, they gave mde a present to remember them by, a dressing gown thing made from real chinese silk with embroided dragons on it, it is soooo nice! They got it when they were in china..... obviously..... OH and I also went grocery shopping with them, and we bought fresh artichokes and ate them for entre which reminded me of our old artichoke plant back home. Which is now \'expired\' as the three of us would put it.   So the next day we were up bright and early and on our way to Nyborg, a small harbour town on the island of Fyn. To get from SjÃ¦land to Fyn we had to cross the longest hanging bridge in the world... well, it was ONCE the longest hanging bridge in the world. For about 2 months until the Japanese built a longer one...lol. Ayway, we were there because their son, Kenneth, is in the Navy and his ship was docked their for the weekend so we went to visit him. The boat was very cramped and kind of crappy, I dont know how he has managed living the past 4 months on it!... so we made Swedish pizza for lunch which was super dooper and then went for a walk around the town of Nyborg and looked in an old church... came back to the boat for coffee and biscuits. I ate the buiscuits but didn\'t dare touch the coffee, the machine looked as though it hadn\'t been cleaned in a long, long time!... and besides, I dont like coffee... Kenneth took photos of EVERYTHING using the phone camera which was funny. We all made fun of him for it.   So after our boating adventure, we took a car trip (minus Kenneth who had to watch the boat) so a place where Gitte og Bent could stock up on beer (and stock up on beer they did!) and we went to the biggest rock in Denmark, Damestenen (The Lady Rock)of a whopping 42 metres! Not quite as big as Ayres Rock, but well, Denmark is a very small country. Gitte and I had a photo shoot on the top of the rock :) If I ever find the pics I will find somewhere to post them so you can see them...   And second last on our little sightseeing adventure was a castle, I can\'t remember what it was called, but I do know that it had over 1,000 singular window panes (it said so on the brochure!) So we looked around the castle for a while, it was the nicest castle I have seen in denmark yet, but I have only seen 4. And yes folks, Katie is now part of history, having been to a real, danish castle toilet! I\'m sure you all wanted to know that. So anyway...   Then we looked around the castle grounds, and they had a little museum thing with all these old antique cars, very nice indeed! Then Gitte and I needed to pee again so we went on a bold search for the museum toilets... which we didn\'t find. But what we did find was a secret hallway, called \'Draculas Mansion\'... It was like this creepy, really dark hallway thing with crypts and spider webs and creepy organ music playing in the backbround. And at the front it had the story of the real dracula and how the dracula stories began, which was really interesting....   So there were Gitte and I, standing at the edge of Draculas Mansion, neither of us daring to go foreward... \'Shall we go in?\' I asked Gitte, clinging to her arm... it was a very creepy hallway... \'Uh... I guess.\' Gitte replied. We both took a tiny step foreward. .... .... We both looked at each other. \'uhhh... Lets wait for Bent..\' We decided we needed a strong, male presence to protect us from dracula.   So we went back, collected Bent and walked into Draculas Mansion, both us old ladies clinging onto Bent like glue. It was so funny, Just as we were going in, we stopped for a few seconds. Gitte looked behind us and screamed, jumping five feet in the air... which made me scream and jump five feet in the air. This German tourist just suddenly appeared behind us, and because of the rain outside, had a hooded jacket on which in the shaddows looked sooo much like a cape! Lol, anyway, Gitte, Bent, the German tourist and I stod there and laughed for like 5 minutes because it was so funny... i guess you just had to be there!   After the museum, we walked through the gardens. The castle grounds had 4 mazes, although we didn\'t have enough time to go in any of them. I noticed one of them was made out of bamboo. I thought this rather pointless as you could pretty much see through the hedges... and wasn\'t that NOT the point of a maze. There was a small tree top walk, where we could walk on little hanging bridges 15 metres in the air, among the tree tops... although certainly not as big nor impressive as the one that was like 80 metres high in Australia, I thought it was still nice. After that we continued our way through the gardens... \'sampled\' a strawberry or two from the strawberry patch, and a raw gooseberry... then exited the gardens. Because it was the castles 450th birthday, a single entry fee will automatically come with a year pass... Gitte spelt my name as \'Catie\' rather than \'Katie\'which I thought was rather cool.   So then on the way back, we stopped at this little resteraunt thing for dinner. Its a quiet little cottage in the middle of a forest, I liked it a lot! Anyway, it was a buffet kind of thing, help your self deal. And we sure helped ourselves! There was soooooooooo much food there, seriously I dont think I can write just how many different kinds of food! So, naturally we wanted to try everything there... and the good stuff we wanted to try twice... we had marinated steak, spare ribs, chicken kabads, curried lamb, baked mini potatoes, fried potato, potato salad, pasta salad... and the list goes on! So yeah, we were very very full indeed when we went back to the car, and I slept until Copenhagen. It was an early night that night, we had a big day! Lol!   The next day they said I could sleep in, they had no plans. Now, I\'m not sure what they ment by \'sleep in\' but when I got up at 11am, they were all like \'Woooow! Good Afternoon, Jonas!\' (they have a joke with this guy they used to know called Jonas who slept all the time... so I became Jonas) It was funny, though. So that day I did not a lot, just sitting around and talking to Gitte and Bent, and then I scrubbed and peeled 50 million potatoes. Fun fun fun! Now, this is kinda strange for denmark, but Gitte and Bent have a little pit fire in their backyard and that night we cooked fresh salmon and potatoes in the fire, and as much as I dont like fish all that much, it was sooo good!   That night I went home with Marianne and Anne and Niels, normal life back at the host families!   Anyway gtg,   Update again another time!   Luv Catie/Katie/Jonas '),(24011,3715376,'25,June,2004','2004-06-25 00:00:00','   This weekend Marianne and Neils are going to a rotary weekend, and due to my temporary loss of host siblings, I\'m being shipped off to a friend of the families for the weekend... I\'ve met them once already and they are very sweet, I\'m sure I\'ll have a nice time there. We\'re going to visit their son who is in the navy (I think) so we\'ll go down to the harbour and see all the boats which will be nice. So around 4.30 this arvo I\'ll be moving to another house, but only for 2 nights.   And at 1pm today Im going off to the airport, again, to say goodbye to yet another oldie, Angela. Which I\'m sure will be very sad...   I got an email from my host-sister-to-be when I go back to Australia today, v. exciting, I can\'t wait! She\'s invited me to visit her and her family in Denmark so I will definately do that :)  I probably wont be on for the next three days, but I\'ll be sure to update when I get back!  Cheerio,  Katie '),(24012,3715376,'24,June,2004','2004-06-24 00:00:00','   So, yesterday was the summer solstice, or Mid-Summer as they like to call it here. The longest day of the year, where its tradition to go down to the beaches and light huge bonfires and (in the olden days only) burn witches!... but only one problem, there was a thunder storm. All through the day it rained and rained and rained and rained... and even at one point hailed... and just as it was hitting the bonfire lighting time, the rain all of a sudden stopped! And not a moment too soon! So we got our bonfire in the end, although it had to be started with petrol because the wood was wet and then the petrol made this really gross brown smoke stuff... but the point is, we had our bonfire, and we burnt some witches! (fake stuffed witches, but witches none the less) And then just as it was getting dark (around 11.30 pm) everyone was setting off fireworks, very pretty!     So we (me and Marianne and Niels, me hosties) came back to the house and ate freshly picked danish strawberries with vanilla ice cream and watched more fireworks on the tv from Ã…rhus... and at exactltly midnight the rain started again and then the power went out which was kinda cool... so that was my danish mid-summer   Toodles  Kaz '),(24013,3715376,'23,June,2004','2004-06-23 00:00:00','   Looks like I\'m an only child for the moment... My sister Anne (16) has gone off to Jutland for the week with a friend from class, and my brother Christian (19) is touring Berlin with his band. So its been nice getting to know the new hosties better.   So... my host family have just finished booking the accomodation to Austria and the Czech Republic, where we are spending around 2 weeks in beautiful central europe! On monday night my host mum and I took a spur-of-the-moment bike trip to the local library to get some travel guides for Prague (they even got a Lonley Planet guide in English for me :D) and we spent the rest of that night reading over the guides together which was quite fun... I am sooooo pumped about our Summer Holiday!!!  Denmark played sweden in the football (soccer for all you non-europeans). If both teams got 2 goals, then Italy would be out of the compitition which is what all us scandinavians wanted... Denmark led for most of the match, but the final score was 2-2, perfect! Ciao Italiano! Next match Denmark plays is against Czech...  Aargh, I can\'t believe just how fast the past 6 months have gone... its so crazy, and they say that the last 6 months goes the fastest... so before you all know it I will be home again.  ...oh and by the way I think you can register as a member of my site here, so if you figure out how to do that, I would be much obliged!   Cheerio!  Kaz '),(24014,3715376,'22,June,2004','2004-06-22 00:00:00','   Hi everyone! Well, guess whos back from the dead? Ok, not the dead, but at least the 3-months-with-minimal-contact-to-australia phase I was going through... so I have found a solution, a blog (short for web-log) here I can post messages... or if I\'m feeling too lazy to type then I can post audio messages so you can hear my beautiful voice again!   So... what has katie been up to in the past 3 months? Hmm... probably too much to write in one simple post, so I\'ll just skim over the highlights...  1. Euro tour, 18 days in a bus with 57 other exchange students going around europe... my fav. countries being austria and italy...  2. Survived a near death experience involving a faulty safety harness and a 20 story drop on a ride called the \'the golden tour\' (or \'den guld tÃ¥rn\')... but yes I am alive and in one piece!  3. Moved host families, my new family are sooooo sweet :D   4. Met some of the coolest people in the world... one of whom is going home on saturday :\'(  5. FINALLY got around to watching Pirates of the Carribean... very cool movie and I highly recommend it  6. Um... thats about it... I\'m sure I\'ve done lots of other cool stuff but I just can\'t think of it at the moment but when I think of all the other cool euopean things I\'ve done I\'ll be sure to tell you...  OH and on the 4th of July I\'m going back to Austria and the czech republic again with my hosties, w00t!  Anyway, gotta get a move on!  Love yas all, Kaz '),(24015,3715376,'18,July,2004','2004-07-18 00:00:00','   So, this little exchangeable is finaly back from  Austria and Prague with the hosties. Had an AWESOME time, twas lots of fun. So here is a quick scedule of what I did...  &nbsp;   Sunday, 4th July: We figured out the car trip to Austria would take around 14 hours... so logically we would leave at 7pm that night and arrive 10 the next morning. Yes, with 5 people in the car plus baggage and a whole night in the car, we were all looking foreward to the long drive ahead... need I point out my sarcasm? No, it really wasn\'t so bad really... of course, me being the ever impressive exchange student ended up stuck in the middle seat for most of the trip, but well, despite the fact I had next to no sleep and I was stuck in between Christian with the pointiest most triangular shoulders in the world and Anne who wouldn\'t know the words \'move-over-your-squashing-your-poor-little-exchange-student\' if they came up and bit her in the nose!&nbsp;So after being sat on, poked by pointy shoulders,&nbsp;a long long time, I was rather looking foreward to the end of the journey. However, there&nbsp;was one highlight of the&nbsp;journey. Let me tell you, theres nothing like&nbsp;watching a German sunset at 4 am squashed into the back seat of a crouded car... unless of course you count sleeping, which is&nbsp;in fact what I was doing as the sun came up. I guess all that lack of sleep finally caught up on me! Which brings us to...   &nbsp;   Monday 5th July: After&nbsp;around 12 hours in the car, we stopped in at a truckstop for some 8am breakfast. We were all tired and grumpy&nbsp;but thaks to my \'Rotary Plastic Smile\' I managed to come out of Breakfast without biting anyones heads off. So I atre some joghurt, bread with cheese&nbsp;and some with nutella and had a hot chocolat. There was a huge toy kanargoo in the corner. Over breakfast Christian enlightened me as to why Danish people dont like German people. I had known about this fued for a while, but I never knew where the arguemant had actually sprung from. \'Well,\'&nbsp;explained Christian, \'Its because they\'re all big and fat and smell like sausage.\' Hmm... yes very good explanatin indeed... but I guess its&nbsp;at about the same level as the why Aussies dont like New Zealanders or&nbsp;Tasmanians. And just to prove his point, the very next person to walk past our table was indeed, as Christian had said, a big fat German... and&nbsp;yes, he smelt like sausage, but then again he&nbsp;was the chef of the cafe there so I guess he had an excuse... So&nbsp;with wide, bright eyes and cheerful faces all round, we willingly&nbsp;squashed ourselves back into the car for the last leg of our journey... only 2 hours thank goodness!&nbsp;     &nbsp;   So we arrived in our little Austrian town, I can\'t remember the name of it but&nbsp;it was very close to Salzburg, where The Sound Of Music was filmed.&nbsp;We were to stay in a 3 storied apartment type thing with three floors,&nbsp;and we had booked the 2nd and 3rd floor of it, as another family (friends of Neils and Marienne) would also be joining us in austria. We arrived at 10 but we werent supposed to check into our rooms before 3. We couldn\'t find anyone who worked there around so we just waited and walked around for a bit.&nbsp;The apartment thing was on a&nbsp;farm thing. Also on the farm was a&nbsp;little herd of deer, some rabbits, some sheep and 2 goats (one of which bit me, lol). An hour later at 11am the family, (Henerik, SÃ¸ren, Lars and Nies) arrived, along with the land lady of the place. So we unpacked, socialised, played table tennis (at which i lost badly) ate a BBQ for dinner and then caught up on some much needed sleep!  &nbsp;  Tuesday 6th July: When Niels and Marienn &nbsp;   &nbsp;&nbsp;   &nbsp;   To be continued, will get some food now&nbsp;cause I\'m hungry)&nbsp;&nbsp; &nbsp;&nbsp;&nbsp; '),(24016,3715376,'04,July,2004','2004-07-04 00:00:00','   Ok... believe it or not... I have like no idea at all what I actually did on the Tuesday after the 3Â½ hour rotary meeting on Monday.... really no idea at all, which is really quite strange.. maybe I rang someone? No, I honestly have no idea... so lets just skip Tuesday and go to Wednesday...     Wednesday - Today I said my goodbye to my little Jill from Oregan, MISS YOU JILL! So got up bright and early at 11am and was at the airport by 12. However, did not come away empty handed - scored 2 bags of her hand-me-down clothes and some old hair products and that kind of stuff. Woohoo! Ok must of it was olf beyond use but it was free, and in an exchange students world, ...well, now ords can quite describe it. Of course it was sad when Jill had to leave, she is so cool and funny and I will miss her heaps but we will get our newbies soon, so that will be fun... that afternoon I went to see My host Mormor and Morfar (granparents on the monthers side) they are really nice, and speak a little bit of english too which is a plus!   Thursday - Ever had just one of those mornings? Ok this was one of those mornings! I had planned to visit the Danish girl who will be my host sister in Australia. The plan was, be in Slangelse by 2, have lunch and dinner there and meet her family. So, it takes around an hour by train to get to Slangelse, wasn\'t sure how long it would take to organise tickets and what not, so I set my alarm clock for 10, which would give me plenty of time to be there.   So my morning went like this... mum rang and woke me up at 8.15... and as much as i liked talking to her ai was so tired and really wanted to sleep... so i went back to sleep. When I woke up I looked at my watch... and almost had a heart attack! 11:59! Aargh! I had slept until 12! Ok, I did a quick figure of sums in my head... 2 hours... I had one hour traveling time, it would take me half an hour to get to contral station, 5 minutes to bike to the station... so poor me is left with 25 minutes to be up, dressed and showered! So going into superman mode, I sprung into action... Was ready in just under 25 minutes! Wow, new world record! So Jumping on my bike I pedaled as fast as I could to the station... that us, until I got a flat tire... so I ran with my bike as fast as I could the rest of the way (which wasn\'t too fast considering i had just woken up and i was lugging a bicycle!)I arrived at the station JUST as the train was leaving. This was just too typical! So I waited 20 minutes for the train... which turned into half an hour as the next train was delayed... so after getting off the central station, I rushed into the ticket office, ordered a ticket to Slagelse... and JUST missed my train there! By this stage I was ready to strangle anything and anyone that came near me... but no I fought down my anger and bought a hot dog since I hadn\'t had time for breakfast. The next train came in half an hour, and in the end I was only half an hour late which really wasn\'t too bad!   My afternoon - I was a little nervous about meeting Katrine, what if she were big fat and hairy? (sorry personal joke... i really have nothing against big fat hairy people)But no, she was really sweet. We discovered that we both love the same rock music and we both looove horse riding. We got on really well, I can\'t wait until she lives with me in Australia!   So we went shopping together for gifts for her mum, as it was her mums birthday. Our spoils of war were - some eyeliner, a pair of white pants, a cd and a shaver. We looked around the town of Slagelse for a while and were just on our way back when it started raining. But not just that little harmless rain... this was fig fat heavy drops of rain, and lots of it too! So arm in arm we sprinted to the public library and took shelter in there for a while and we had fun emailing people from her account, and I was going to show her some websited about shepparton but then the rain stopped so we continued on our way back home. We made it home and I met her sister (i had already met her mum at the station) and had dinner of chicken, salad and potatoes, very nice indeed! Then some of her mum\'s rellies came over so say a quick happy borthday and we ate cake and stuff...mmmm...danish birthday cake.... then Katrine and I sat on the couch together and I showed her some photos of the family in Australia which was nice, she\'s looking foreward to comming here.   Then we raided her room and figured out what clothes she would need to take to Australia... and with a 20kg limit on luggage we managed to sort out what she would need! Then Nicholas came home, the brasilian exchangeable staying with their family, came hoe from rome and the theree of us just sat around and talked until it was time for this little piggy to catch her train back home...   And it seemed by bad luck with the danish transport system was following me still... half an hour into our journey the train suddenly stopped in the middle of nowhere for like 20 minutes for aparently no reason... but all was well after that and I arrived home safe and happy and very looking much foreward to hosting our little danish student! '),(24017,3715376,'03,July,2004','2004-07-03 00:00:00','   So, the day after I got back from Gitte and Bents, the last monday in june, I gÃ³t up at 10 and patiently awaited for Claire, my cool danish friend from class who invited me to go to this kick-ass 5 day music festival on the island of Langelands. 2 other girls from Claires swim team would be comming too, so we arranged a little lunch so I could meet them. They were both very sweet girls, and only laughed when it was discovered that Claire and I had... well... \'re-arranged\' the pizzas we had bought for lunch. Ok... I know the simple rules of gravity say that when you hold a pizza on a right-angle degree, the topping will fall off, but we weren\'t exactly thinking about gravity when we needed ot figure out a way to transport the pizzas from the store to louise\'s house on bikes... anyway... it was kinda funny... :D  Anyway, the two girls, Julie and Louise were both really sweet. And Louise has this huge-ass oversized greyhound which is sooo soo cute!   So after that I went home and got ready for the Rotary meeting that night. 3Â½ hours of wholesome Rotary fun.... yeah as if! no, it really wasn\'t too bad, in fact, I actually kind of enjoyed it! I spent most of the meeting talking to an old Norwegian guy in Danish... which was funny because my danish is soooo bad. And then I spoke to the lady who was Emily (the last exchange student with the TÃ¥rnby Rotary Club) and she was soo nice and said she would invite me over to her house one weekend, which I am looking foreward to because she was cool.   Anyway, signing off for now  Captain Katie '),(24018,3502527,'22,July,2004','2004-07-22 00:00:00','     I don\'t know if you can catch ESPN in the PI, but they\'re airing the annual   urlLink  Scripps National Spelling Bee   . I know, it does sound weird that ESPN has included this kinda thing in their programming but people shouldn\'t be so surprised after they started airing poker games.  Anyways, I was sitting on the couch checking out one of the final rounds. This awkward-looking Indian kid with a complementary strange name and stoic demeanor,   urlLink  Akshay Buddiga    steps up to the mike. He recieves his word to spell, which is \'  urlLink  alopecoid   .\' After repeating the word a few times, our man Akshay seems to be deep in thought.  Then it happens. His eyes flutter and begin to grow wider as if he\'s shocked or something. His body begins to sway to his left a bit. Almost imperceptibly, Akshay begins a slow descent to the ground- drawing gasps from the crowd (probably stifled snickers, too). Further replays (thank you ESPN!) show that he does absolutely nothing to slow his fall. No flailing arms, nothing. You know those videos where you see loggers chopping down trees? Akshay looked just like that. Well, almost like that. Imagine a tree with glasses and prematurely developed facial hair.  At the expense of his embarrassment however, I was left with a classic moment...one of the most hilarious sights I\'ve ever seen. I would pay to relive that shit again.      '),(24019,3502527,'22,July,2004','2004-07-22 00:00:00','     urlLink  I ntroverted, i N tuitive,  T hinking,  P erceiving      '),(24020,3502527,'22,July,2004','2004-07-22 00:00:00','     Current Location : I\'m in scenic Lansing, Michigan visiting my girl Michelle.   Job Status : After a brief stint at Google, I\'m working as a Game Tester at Visual Concepts. Those familiar with video games will know that they\'re the producers responsible for the 2K3 series of sports games. After a few weeks working here, I\'ve decided that this isn\'t the exact industry I want to learn about.  Oh, and I\'m sure many of you out there think that testing video games is the perfect job. This isn\'t exactly the case. While video games are ideal for an hour or two of mindless enjoyment, the fun can wear off when you\'re stuck playing them hour after hour after hour, etc.  I\'m lucky that my coworkers make my job bearable. Well, getting baked before work doesn\'t hurt either. Hehehehehehe... (stoned laughter).   Game Plan  (tentative of course): 1. Apply for grad school to study advertising. 2. Take the GRE. 3. ?  To be Continued!      '),(24021,3502527,'22,July,2004','2004-07-22 00:00:00','     Oh man...Rina\'s got me into Blog-land. You know, I honestly have never felt drawn to blogging. I once tried (and failed) to maintain one a few months ago. It lasted exactly one post.  I guess I can attribute my lack of interest in blogging to my personality. In general, I am a pretty introverted dude. Well, that\'s not entirely true according to the Myers-Briggs test which you all can take  urlLink here .  I last took this test in high school and on the \'Introvert/Extrovert\' chart, I barely scored into \'Extrovert.\' It would be interesting to see what I would get on that shit at this point in my life but I digress!  My point is, I never really saw the point of keeping an online diary (which is what a blog is basically) for all eyes to read. In my opinion, it would be much more interesting for the readers\' sake to fictionalize things.   But who knows...maybe this little experiment could prove to be fun and productive (?). Maybe I will catch onto the whole blog craze. Maybe I will find it to be a cathartic experience to post my inner thoughts, dreams, and desires online for random people to obsessively pore over. Maybe...  Um. Nah. Fuck it. Let\'s have a little fun. Sit back, relax, take a long drag off your  urlLink L  and enjoy my random-ass ramblings.      '),(24022,3502527,'22,July,2004','2004-07-22 00:00:00','     Messing around with adobe during with my free time, results to this kinda shit. Man, you gotta love that photoshop. My staff from the office passed by behind me while I was tinkering with the graphics. And not knowing the muppets, she asked \"Do you even like that?\" with the most puzzled, grossed out look on her face. It was like she had just seen the ugliest creature ever. And to think that these are endearing characters from 2 popular kiddy shows. In response, I just gave a sneeky smile and said \"yes~\" Anyways, hope she doesn\'t see the resemblance there.      '),(24023,3502527,'22,July,2004','2004-07-22 00:00:00','     It\'s as simple as something that nobody knows  That her eyes are as big as her bubbly toesâ€¦  â€¦His dreams are like commercials  But her dreams are picture perfect and  Our dreams are so related though they\'re often underestimated    I canâ€™t get enough of Jack Johnson. His musicâ€™s got incredible vibe and itâ€™s such a great bluesy, folk, tropical mix of sounds. I donâ€™t know if thatâ€™s what he intended, cuz I donâ€™t read much on his bio or anything. What strikes me more is his lyrical sense. Itâ€™s uncomplicated yet profound, whimsical, and sometimes even funny. Like this one called â€œBubble Toesâ€. How the hell do you come out comparing the size of oneâ€™s eyes and toes? His song is so full of contradictions and things that donâ€™t add up, but somehow it does make sense. Sort of like myself, â€˜walking paradoxâ€™ as I always say.   On a side note, the whole â€˜bubble toesâ€™ bit got me thinking about my defects. The stuff I call my defects: lazy left eye, animal teeth, flinstone toes, cracking knees, and the list goes on. Although â€˜defectâ€™ sounds a bit harsh, it really isnâ€™t that bad.   eg. Animal teeth â€“ as in that orange drum-beating freak of nature of a muppet â€“ Animal. Flinstone toes â€“ equal length toes like that of Fred Flinstone.  A lot of these stuff were pointed out to me by friends and were baptized with matching weird names. They sound a whole lot freakish when enumerated like that, but I actually like them quirks now. It gives me character I guess...   Now I\'ve set myself up sounding like some hideous monster. Oh man.      '),(24024,3502527,'22,July,2004','2004-07-22 00:00:00','     I\'m scheduled to leave for the US this Friday, so Iâ€™m making my travel arrangements. Being a travel consultant, this is a cinch to me. But what is painstakingly frustrating is the fact that Iâ€™m traveling on award tickets. Yes, the freaking perks of the job. Seriously, itâ€™s not what it seems.  Anyone whoâ€™s traveled on a non-revenue ticket knows how impossible it is to get confirmation on these award ticket bookings. You think that the airline would give you credit for flying enough to earn those miles. I\'m talking to a Northwest Airlines agent as we speak and I still can\'t get the flight that I wanted. Plus I can\'t even get my return flight confirmed on PAL, so I\'m kinda screwed if I don\'t get back within this month. Frequent Flyer programs are the biggest rip-offs!!!  Rants aside, I\'m glad to be able to visit some really good friends and family. Plus, I haven\'t been there since \'99. This trip will be a long-ass trip, but definitely worth all the hassle. I\'m gonna eat a whole lot, soak up the energy of the city, spend quality time with people over there and just relax. At this point in my life, nothing can beat that... not unless itâ€™s a trip around the world or winning at the lottery.  On a side note, I caught Akshay on the CBS Today News the other day. He apparently got to spell the word after coming out of his black-out. Kudos to him for that. But still, what a dork...      '),(24025,3502527,'22,July,2004','2004-07-22 00:00:00','     A single shadow creeping in the nightâ€¦ Hushed footsteps tip-toeing down the hallâ€¦ down the stairsâ€¦ into the kitchenâ€¦ Doors opening/closing, utensils clinking, sounds of rustling, gulping, gobble gobble munch munch munchâ€¦ TV blasting till the brink of dawnâ€¦  This is the aftermath of my 2 week hiatus.  After taking a break from Manila, flying off to  urlLink New York   then to  urlLink San Francisco  , Iâ€™m left with my zoned out, trÃ©s heavy, jet-lagged self. Itâ€™s not insomnia that haunts me. Itâ€™s having to wake up around 3:00 a.m. and not being able to sleep that kills me. As if my eye-bagginâ€™ zombie self isnâ€™t bad enough, getting the munchies at odd hours is just dietary suicide. The other night I just murdered 3  urlLink Krispy Kreme French Crullers   in a span of 5 minutes. I canâ€™t go on like thisâ€¦ I really canâ€™tâ€¦ more like I shouldnâ€™t.   Whatâ€™s up with midnight snacks anyway? Why does it taste so good, and why do you get such a strong craving for certain things? I was watching some medical documentary on  urlLink Discovery channel   at 4 in the morning and still got this urge for a deluxe cheese burger, guacamole with chips, and chocolate.  Gotta get it together!!      '),(24026,3502527,'22,July,2004','2004-07-22 00:00:00','     urlLink    .pls. be sweet. keep it neat.&nbsp; urlLink     I know, I know. The posting well on my part has dried up as of late. Apart from the dreadful constant that is being a game tester, it\'s been a whirlwind around here. Rina was nice enough to stay for awhile and just a few days later, Shelly came home from school.   Having Rina here for about a week was fun as hell. We both had a great time and shared dozens of laughs (as you can see in her detailed recap!).      '),(24027,3502527,'22,July,2004','2004-07-22 00:00:00','     urlLink    karmanimal stormin\' the haight.&nbsp; urlLink     Man...Rina is outblogging the shit out of me! Time for me to get back on this.  Oh, and if this pic isn\'t dope then I don\'t know what is.  p.s. Doesn\'t it look like Rina and I are covered in slime in the previous pic? Ughhhh.      '),(24028,3502527,'22,July,2004','2004-07-22 00:00:00','     Recalling an incident in San Francisco (dated 6.20.2004.)     It was 2:30 in the morning, Vince and I were wide awake, bored and thought of nothing better to do but to seek for herbal mood enhancers. Mind you, I really donâ€™t do this often. Just wanted to get that out in the open before anything else. Anyways, we step outside his house just to face the cold gust of mid-June summer. (Why is it so frigginâ€™ cold still? You call this summer, goddamnit!?) We light the baby up and each take a slow drag. We laughed mindlessly, spat profusely, and talked hysterical non-sense. As we sat there and did our shit, we encountered a strange bunch. (Maybe not too strange, just a tad bit)   1) Huff, puff and away  Some old man, walked up to us and said â€œDâ€™ya haaave ay sheegaritâ€ in an incoherent slur. In translation, it meant â€œDo you have a cigarette?â€ which is a simple question in itself. At that moment, I actually thought he was speaking some Japanese dialect. To this we just shook our heads and kept our mouths shut. Mystery man in dire need of a smoke left never getting his yosi . It was only when he walked away that we noticed his slight limp.   2) All in the army  Next, was this Chinese (?) middle-aged woman walking â€œmilitary styleâ€ with arms and legs straight, raised to the heavens. Ok Iâ€™m exaggerating about the â€˜heavensâ€™ bit, but really, you could almost hear the â€˜left, right, leftâ€™ chanting in the background. How can a person be so uptight and rigid, when you think you\'re not being watched? (Well technically we were watching her, but it was done discretely) Anyways, who are we to presume what kind of stuff she\'s got in her head. It just looked fuckinâ€™ weird.   3) Melons, I swear    Another random man passed by, this time with a more bizarre-looking walk then the other. If you could see the picture, he looked like he was walking with melons in his hands. He just happened to grip it in the oddest way. This low-tech graphic doesnâ€™t even capture the strangeness of this manâ€™s posing.  Now that Iâ€™ve written all these down, it actually doesnâ€™t seem as funny as it did before. When youâ€™re stoned, even the most mundane things become hysterically amusing. But this was nothing compared to the events that were to follow.  Once we were done, I jumped off whatever I was sitting on to get my feet back on the ground. A miscalculation led me to a clumsy fall, as I thought I was jumping at least 3 ft. high when in reality it was more like a foot or less. Dumbo meâ€¦ After that display of stupidity, I slid my toes in my slip-ons and was ready to go, when suddenly, felt the most yucky, slimy substance in my shoe. I yelled, punched, kicked Vince and accused him of spitting in my shoe. Then I hopped on one foot and quickly made my way to the room to get that mysterious gunk of my foot. As I threw the shoe off, a slug flew out and landed on the floor. How horrible!! In the next second, I dashed to the bathroom, opened the faucet and was desperate to erase that slimy feel off my foot.  As Vince recalls that moment, he said I had turned into a woman in rampage. I was throwing stuff at him, intensely pressing on the shampoo bottle and scrubbing my foot sooo hard. When we got back to the couch, that snot-like slug still had the will power to move 3 inches from where it landed. Man, that stuff was  DISGUSTING!!  The word \'gross\' doesnâ€™t even come close to what I experienced in that few seconds. 2 weeks has passed, but I can still remember the texture of it through my socks.   HATE HATE HATE SLUGS!!      '),(24029,3502527,'22,July,2004','2004-07-22 00:00:00','     (note: I manage Eya\'s Friendster account which lists her real age among other things. Without further ado, I present internet romance at its finest.)  DATE: July 8, 2004 5:29 PM  SUBJECT: wow  MESSAGE:  wow just wanted to say you are ridiculousl cute, do u have aim or yahoo to chat on? love to hear form you    urlLink  kev     This message had us cracking up for days. And, if anything, it reminds us that Eya\'s got gaaaaaaame.      '),(24030,3502527,'22,July,2004','2004-07-22 00:00:00','     I had an interview last week with a recruiter for a tech job agency. To clarify, my co-worker Rich told me about a QA localization job opening in San Fran. Wow. It\'s funny but I think I might have made shit even more confusing. Alright, for the non-videogame industry savvy, QA is just a fancy word for game tester and \'localization\' is basically making sure a game produced in a foreign country (in this case, Japan) translates well into English.  We\'d really like to avoid any of   urlLink  this    happening. Hehe.  By the way, I work at Visual Concepts Inc. which is now known as   urlLink  ESPN Videogames   . They make produce titles that compete with (and usually best)   urlLink  Electronic Arts\'   .  Needless to say, this job opportunity would be perfect for me right now as I\'m trying to sort out going back to school for a second degree. First off, it\'s much closer than San Rafael. Secondly, I would be working \'normal\' hours as opposed to my current jacked up schedule (going from 5:30 pm - 1:30 am, Saturday to Wednesday...nuts, I know). Thirdly, localization work seems like something that would come naturally to me. I have a keen eye for grammatical errors, syntax shit, and typos as a result of my English background. Oh and to top off the list of positives, it pays more!  Anyways, the interview went really well. My recruiter was very nice. Disarmingly nice even. It was the first interview I\'ve had that seemed like a friendly chat if nothing else. The grammar test she administered was pretty easy as well. Dude, the situation described in the test dialogue was two guys with gay porn names (John Michaels and John Stevens) fighting over who owned a duck. Go figure.  After chatting with her, I discovered that the company I\'d be working for would be   urlLink  Sega   , which is ironic as fuck. See, Visual Concepts is a developer for Sega. Oh well- I guess there would be no love lost between the two parties since I\'m basically working for the same company. Now I\'m slated to meet with the Sega hiring dude on Wednesday. Wish me luck!      '),(24031,3502527,'22,July,2004','2004-07-22 00:00:00','     [Disclosure] Iâ€™m not looking for solutions. I just want to let out some steam.   Indeed, there is a fine line between love and hate. As of late, Iâ€™ve been desperately hanging onto the edge of love, cuz itâ€™s so easy to slip into the dominion of hate. More than just hatred for others, but deep hatred for myself.   Each day the demon in me keeps lurking. Desperately trying to find its way to come out. It peeks through my expressions, my words, my actions. One can easily detect its ferocity through my ever gleaming eyes. Often, I try really hard to bite my tongue and resist the urge. But I just canâ€™t help it. Its energyâ€™s always being fueled by the growing hatred I sense; your hatred towards me that is.   You stare at me with a cold blank face. You never hear me out. &nbsp;My voice bounces back whenever I speak to you. And in spite, I let out cruel words like darts and daggers just to make it penetrate you.&nbsp; And I know it hits you hard. But what can I do, when I feel like thatâ€™s the only way you\'d take me seriously. Instead of showing understanding for you (which I really do I have), I show my fangs and become a feral bitch. I know it in my heart that it actually makes matters worse, since you get offended when I become this unpleasant (what an understatement). But my head says stand up for what is right and correct, even if I have to stick it in your face. How do you ever reconcile reason and emotion? Cuz seems like I canâ€™t at all, when it comes to this topic and I just keep falling into some deep dark pit of ugliness. &nbsp;   There is a perfect idiom for people like me in Japanese. Its called  AMA NO JAKU  wherein its literal translation is â€˜Heavenâ€™s demon.â€™ But what it really points to is my awkwardly defiant, almost wicked nature. I never back down when it comes to you and I keep telling myself its only because I strictly fight for the truth. I wonâ€™t comply if I donâ€™t get a logical explanation. And I know I get so anal sometimes, but I just canâ€™t let things slide especially when they are important matters that get massively screwed up on some stupid whim. I really donâ€™t mean to contradict you all the time.       Bottomline is that I hate having to feel this way. I want to bask in the warmth of your unconditional love and be able to return that without expecting kindness in return. I want to be able to turn on some protective force field whenever someone starts mudslinging at me for no apparent reason. I want to be able to understand you and have you understand me as well. Above all, I just really want to be able to talk to you without any hostility between us.   Can we make that happen? Hmmmâ€¦ I donâ€™t know. Can I make that happen? I will try to control that malicious tongue and wicked eye I have. So best of luck to the demon from heaven. &nbsp;Cuz you know that demon is me.      '),(24032,3502527,'22,July,2004','2004-07-22 00:00:00','     You often hear people say \"Iâ€™m getting weird vibesâ€¦\" or \"Donâ€™t you feel bad vibes?\"      It is a word so loosely used in our daily lives. But what does this actually mean?Is it some instinctive gut feel, generated by past memories, experiences and stock knowledge? Or is it simply some thick figment of imagination? I believe they are neither of the two. What I do believe is that it is triggered by a sense of fear that you emit, that affects you to think on your toes instead of being attacked unaware when youâ€™re feet are flat on the ground.      Fear of the unknown. Fear of the unexpected. Fear of facing the unprecedented. I think that it is precisely this sense of fear that creates a kind of illusion that you know something others donâ€™t. Please bear with me here, cuz this is still a working theory. And this is not to pose that there is no such thing as a sixth sense or to even dare to negate the brilliance of instinct (at times). I guess what Iâ€™m trying to say here is that I donâ€™t exactly believe in the kind of bad vibes you get from a person or bad vibes you get from an event that is yet to come. Least to say, not in itself that is. Thatâ€™s basically pessimism doing the talking and any fraction of potential, possibility or positivism is out of the window. I really am not an optimist. But over time, Iâ€™ve realized that it is a great waste of your time and energy as well as those of others, to freak out, analyze and scrutinize every single thing to miniscule particles when it is best to think of constructive alternatives to the presenting problem. There obviously is a cause-and-effect relationship between people who analyze too much of everything and becoming a little too anal? Donâ€™t you notice that there is the word â€˜analâ€™ in the word â€˜analyzeâ€™?    It is extremely frustrating and demeaning when a person keeps gabbing about bad vibes, without them realizing that it is precisely them that transmit the kind of negative energy on to others. Maybe the reason why they feel that certain discomfort from people is because others sense disdain from the so-called Mr. or Ms. â€˜psychicâ€™ anyway. Isnâ€™t it that when you hate someone, the person will sense it even without you saying it? Am I right or am I right? (haha how cocky does that sound?) Iâ€™m getting too carried away in disproving this bad vibes theory.     Why is there a need to constantly size up people anyway? Why do we have to satisfy some twisted need to fill up our heads with our own versions of how people think, feel, and simply are? Why can\'t we just admit that we can\'t fully figure people out right away or even never at all? The beauty in getting to know someone is in unraveling the little mysteries day-by-day and in them being able to gradually peel off your layers as well. What\'s the fun in finding everything out in a second anyway, right? So lets not overassume people and things, most especially if your assumptions only lead you to a negative place.      '),(24033,3502527,'22,July,2004','2004-07-22 00:00:00','     I\'m a friggin\' dimwit!!  Call me clumsy, dumb and stupid. (But only for now) I actually took time out to fix this blog and imput some add-ons like the comment box and tag board. Hehe and get this, doing it during working hours at that.  Then in a flick of second, it was all gone...all my hard work down the drain. Just because I thought of checking out other templates and trying it out on a whim.  I\'ll have to do it all over again.   GODDAMNIT!!?! NOW I HATE FRIDAYS!!!      '),(24034,3502527,'22,July,2004','2004-07-22 00:00:00','     .anna. & .rins.             .jen. &amp; .rins.      anna &amp; jen...these 2 women are the ultimate causes of much neurosis and gas in my life at this point. they sure know how to drive me up the wall out of sheer frustration, but oftentimes are my sources of laughter, crazy fun and great adventures.     the purpose of this entry is just for posting up these pictures of us together, cuz i looove my  urlLink colorsplash lomo.  love the retro feel, aside from the color. heehee~      anyways, i\'m off to dinner now. i\'ll elaborate on our friendship in a future blog when i have the time.      '),(24035,3502527,'22,July,2004','2004-07-22 00:00:00','     This is Jonathan Jerric Sol Palma, Jerric or JJ for short. He\'s my itsy bitsy nephew from Vancouver, Canada. How can anyone resist that look. Isn\'t that just the cutest thing alive? I just wanna give him a bearhug squeeze. Mmmmm~~ Gotta get me a ticket to Vancouver and some time to visit this lil rascal.      '),(24036,3502527,'22,July,2004','2004-07-22 00:00:00','     urlLink  In the True Spirit of the Games    urlLink  Athens Olympics 2004                The Nakayamaâ€™s are big Olympics & World Cup Soccer fans (more like Nakayama minus Ghetto, that is). We live for this stuff. If we could be there, we would. But the closest alternative to watching it live is catching it on the boobtube on simultaneous telecast. In the year 2001, my dad even insisted that we get a new satellite dish and tuner installed just for that, out of sheer obsession I say. And it wasnâ€™t a simple task at all. My brother flew to Japan to purchase the gadget and we had a crew of 3 people or so install a 10 feet mesh satellite dish on our roof. This was done all in the name of live-telecast viewing of the   urlLink  2002 World Cup    held in Japan &amp; Korea. Really, nothing beats watching it live. More so, actually being there in the stadium like what my sister and I did in Saitama, Japan. Heehee ;) Now Iâ€™m crossing my fingers for the 2006 World Cup in Germanyâ€¦ but thatâ€™s another story.  Now to get back on the topic.  The Athens Olympics opens on the 13th of August, 2004, thatâ€™s in a mere 2 days. The first match of the whole tournament actually kicks off in a matter of hours though. - Women\'s Soccer Match: Japan vs. Sweden. Unfortunately, we have yet to hear much about the Olympics here in the Philippines. Only recently did   urlLink  NBN    confirm that they will air the Olympics (crap reception btw) Itâ€™s a once in 4 years world event and still so many things havenâ€™t been ironed out. There are just so many issues surrounding this event, like overgrown budget, fear of terrorism, insufficient accommodations in Athens, bleh food in Athens et all. It would be a miracle if this one goes on without a hitch. So lets pray for a miracle people. We don\'t want a repeat of that  urlLink tragic massacre  during the Munich Olympics in 1972.  Game viewing and cheering aside, what I love about the Olympics is the   urlLink  â€˜Celebrate Humanityâ€™   ad campaigns. Both Summer &amp; Winter series follow this theme and they usually are such inspiring pieces. Sometimes I even get misty-eyed, when I catch those captions with struggling athletes that overcome their challenges, the tears, smiles and all. The narration, the music, the images, all these combined with the solemn ending of the â€˜Celebrate Humanityâ€™ tag on a blank background. I donâ€™t know whether it ranks high up in the Cannes, Lions, or whatever Ad Award giving body, but it sure motivated me to want to get into advertising. The Olympics and UNICEF ads were the prime instigators of my love for advertising. Well, it is a sad fact now, that studying advertising in college has made me lose that passion for an advertising careerâ€¦  What I do wish for now is that they air more ads here in Manila, cuz weâ€™re seriously being deprived of anything to do with the Olympics.    urlLink  Samples of Celebrate Humanity Campaign (that they obviously donâ€™t show here.. hmph!?)      '),(24037,3809990,'30,June,2004','2004-06-30 00:00:00','   What is my purpose? To defend the right and The Right. To loudly and proudly declare myself a conservative, Christian, pro-Bush, pro-life, pro-2nd Amendment, pro-military member of the Vast Right Wing Conspiracy. To throw things out there and see who agrees or disagrees. To possibly promote some dialogue and thoughtful exchange. To support and publicize others of like mind. '),(24038,3809990,'31,July,2004','2004-07-31 00:00:00','   For some reason shortly after the attacks of September 11, 2001, the media decided it wasn\'t good for us to see images of that awful day. Because of that I believe too many of our countrymen have forgotten that thousands of our fellow Americans, our countrymen, our brethren were savagely slaughtered by evil, soulless monsters. If you have a high-speed Intenet connection and want to be reminded of what was done to our nation nearly 3 years ago I recommend the following website.    urlLink  America Attacked!     \"They have attacked America because we are freedom\'s home and defender, and the commitment of our fathers is now the calling of our time.\"  - President George W. Bush   \"They counted on America to be passive. They counted wrong.\"  -- President Ronald W. Reagan  We will not forget. We will not forgive. We will not yield or surrender. We will prevail. '),(24039,3809990,'30,July,2004','2004-07-30 00:00:00','   There\'s lot\'s of \"interesting\" stuff about the convention, not the least of which was Kerry\'s sweating his way through that horrendous speech. But this incident from a campaign stop today takes the cake, I think.  Before a campaign stop in Scranton, Kerry\'s \"bus convoy pulled over at a Wendy\'s fast food restaurant for a photo opportunity lunch that provided an awkward moment.                Spotting a group of US Marines, Kerry, who has made his Vietnam War service a cornerstone of his campaign, went over to chat. The Marines, who all turned out to be staunch Bush reporters, were not impressed.                 \"He imposed on us and I disagree with him coming over here shaking our hands,\" one of them told reporters afterwards. \"I\'m 100 percent against\" Kerry, he said. \"We support our commander-in-chief 100 percent.\"      Man, I love the Marines! OORAH!      News story from Yahoo! News and Agence France Presse '),(24040,3809990,'12,July,2004','2004-07-12 00:00:00','   Lance Corporal Justin Hunt, 22, of the 2nd Light Armored Reconnaissance Battalion, 2nd Marine Division, 2nd Marine Expeditionary Force died July 6 in combat against insurgents in Al Anbar province.    What\'s inspirational to me about Lance Corporal Hunt\'s sacrifice is what he did in order to become United States Marine.  Justin, of Wildomar, CA, was a star athlete in high school, winning trophies in wrestling, baseball and football. According to Justin\'s twin brother, Jason, he always wanted to be a Marine. But when he went to visit Marine recruiters in 2002 he was so big they didn\'t have a scale that would weigh him. Their scales only went to 350 lbs. and they estimated that Justin weighed 390. After giving up some of his favorite foods and lots of early morning runs, often accompanied by his recruiter, Justin dropped to 207 lbs, just under his maximum allowable weight. At boot camp he lost another 30 lbs and began to talk about becoming a Drill Instructor. While at Camp Lejeune he met Kerry Murphy and the couple made plans to meet each other\'s families.  Justin also dreamed of going to college and becoming an FBI agent. \"I want everybody to know what a great guy he was,\" Justin\'s mother, Debbie, says, \"He loved being on a team, fighting for a goal.\"  Justin\'s older brother, Robert, is in the US Navy deployed to the Middle East and will accompany his brother home. '),(24041,3809990,'11,July,2004','2004-07-11 00:00:00','   Kerry seems intent on proving that his words and position are determined by whatever the most politically expedient course is, a point I made last week.  In 1993 Kerry was among other Democrats who voted against then-President Clinton in banning HIV infected people from immigrating to the US. (A vote that Sen. Kennedy said showed racism and partisan meanspiritedness. So, was he saying the Dems were meanspirited?)   Now, 11 years later Kerry needs votes. Saturday he announced he \"will work with Congress to lift the immigration ban.\"  His reason is so the US can host an AIDS conference. I guess this must be a step in his plan to raise our standing in the world. Gee, I hope the French are watching!  Thanks to Michelle Malkin. '),(24042,3809990,'09,July,2004','2004-07-09 00:00:00','   Apparently the man who would be president is too busy to get a briefing on terrorism and national security but not too busy to party down with the Bush-hating Hollywood Left.  When questioned yesterday by CNN\'s Larry King about Secretary Ridge\'s warning of possible attacks by Al Qaeda John Kerry\'s response was less than reassuring. \"Well, I haven\'t been briefed yet, Larry. They have offered to brief me. I just haven\'t had time.\" Hasn\'t had time? The man who wants to lead and protect this nation is too busy to be briefed on the latest threats?   Kerry appeared on King\'s show just hours before he attended a hate fest disguised as a fundraiser. A bash put on by \"Fahrenheit 9/11\" financier Harvey Weinstein featured Whoopi Goldberg, John Mellencamp, Jon Bon Jovi, Chevy Chase, Meryl Streep, Jessica Lange, and a host of other \"stars\". Kerry closed the event by leading a sing-along of \"This Land Is Your Land\" and said that \"every performer tonight ... conveyed to you the heart and soul of our country.\" John Edwards said, \"This campaign will be a celebration of real American values.\"  Well let\'s see about that. What do Kerry, the Democrats and the Left think \"the heart and soul of our country\" are? What are \"real American values\"? While waving a bottle of wine Goldberg treated the audience to a stream of sexual innuendoes and anatomical double entendre\'s such as \"keep Bush where it belongs and not in the White House.\" She boasted that when asked for a transcript of her intended routine \"\"I Xeroxed my behind and I folded it up in an envelope and I sent it back with a big kiss mark. ... Why are you asking me to come if you don\'t want me to be me?\"   Mellencamp  called Bush \"just another cheap thug\" and the \"Texas bambino.\" Chevy Chase, said Bush is \"as bright as an egg-timer\" and that Edwards will make Vice President Dick Cheney look \"as bright as a bundt cake\" in debates.  Jessica Lange called Bush \"our so-called president\" whose administration is \"a self-serving regime of deceit, hypocrisy and belligerence.\" Kerry was seen laughing heartily during Goldberg\'s screed and called it \"an extraordinary evening.\" Edwards said it was \"a great honor\" to be there. Neither Kerry nor Edwards voiced any objection to any of the performers\' statements.    Party on, Dude! Why Kerry is \"too busy\" for a National Security briefing.   So is that what we consider \"the heart and soul of America\" and \"real American values\"? Do we really want the vulgar Hollywood elite wielding their influence in Washington again? And by the way, when will Kerry find the time to learn what needs to be done to protect this country?  *****************************  10 July 04 16:30 CDT  This just in from the  urlLink Drudge Report:   \"Kerry\'s aides said that he was returning to Boston on Sunday a day ahead of schedule to receive a national security briefing.\"    I feel better already! '),(24043,3809990,'08,July,2004','2004-07-08 00:00:00','   Why are there so many stories the media doesn\'t report? They love to beat us down with what they think is the bad news in the war. I read today that who you think is winning in Iraq is determined by who you want to win. If that\'s so, the media must be on the wrong side. Otherwise they\'d tell us the good news. Otherwise they\'d tell us about Jason Dunham.    Corporal Jason L. Dunham is an American hero. On April 14 Jason gave the last full measure of devotion when he leapt onto a grenade to save his friends. Jason had extended his enlistment so he could remain with his unit through their tour in Iraq. He told a friend, \"I want to make sure everyone makes it home alive. I want to be sure you go home to your wife alive.â€  Jason\'s father, Daniel Dunham, an Air Force veteran, said, \"Jasonâ€™s been my hero since he was born. All my kids are. They never had to do anything to prove that to me.\"    Follow these links and read about this brave young hero.   urlLink Bob Lonsberry   urlLink The Desert Sun   urlLink Talking Proud   urlLink Marine Corps News   Corporal Dunham has been recommended for the Medal of Honor for his selfless and heroic actions. According to the Marines, he is their only candidate for the award. '),(24044,3809990,'06,July,2004','2004-07-06 00:00:00','   I hope you caught this on the news.  Bush knocked one out of the park yesterday.  In a press conference a reporter for Reuters commented that Kerry\'s pick for VP, Senator Edwards, had been described \"as charming, engaging, a nimble campaigner, a populist and even sexy. How does he stack up against Dick Cheney?\"   Without a second\'s hesitation Bush confidently replied, \"Dick Cheney can be president.\"  \'Nuff said! '),(24045,3809990,'06,July,2004','2004-07-06 00:00:00','   Tuesday I posted a quote where Kerry claimed, \"I\'m the only candidate running for president who hasn\'t played games, fudged around.\"  I contend that Kerry HAS fudged around. In fact I think he can\'t help it. Kerry is called a flip-flopper or a waffler. I believe he\'s just being a politician. He says what he thinks will score him the most points at the time he says it. He claims to be a Catholic and violates a key tenet of that faith by his stance on abortion. This is what he did in his changing support of the war in Iraq and other issues. He knew it would be poor politics not to support Bush on Iraq, but once the radical left began to protest our involvement he finds it expedient to criticize the war effort.  Likewise with abortion. Kerry knows the vast majority of Americans, including women, are against abortion. So he says, \"I believe life does begin at conception.\" and \"I oppose abortion.\" But he knows the liberal feminists will crucify any candidate with such a view. So he says it\'s personal, and far be it from him to try and impose his beliefs on another. So if he\'s opposed to murder or theft, but I\'m not, does that mean he\'ll do nothing to prevent me committing those crimes? Was it okay for the Nazis to murder millions of Jews, Christians and minority groups as a matter of national policy. Should the rest of the world have turned a blind eye to avoid \"legislating our beliefs\", in Kerry\'s words.  Let me up!!!   Politicians like Kerry are the reason people don\'t trust politicians. They put there finger in the wind and set their sails to get the most mileage toward their goal. Their principles are only as firm and constant as the latest poll results. Anyone who claims to believe that abortion is the taking of a human life, i.e. murder, and yet believes others have the right to commit those murders as a matter of choice is something much worse than a hypocrite.  It should make anyone considering casting their vote for him rethink their position and examine what kind of person they want making decisions for the most powerful nation on earth. It should, but unfortunately it probably won\'t. '),(24046,3809990,'06,July,2004','2004-07-06 00:00:00','   In addition to the notes below please cruise over to  urlLink Captain\'s Quarters  and see what he has to say! - James    Kerry Says He Believes \"Life Does Begin At Conception.\"    \"I oppose abortion, personally. I don\'t like abortion. I believe life does begin at conception.\"  (Jonathan Finer, \"Kerry Says He Believes Life Starts At Conception,\" The Washington Post, 7/5/04)  In Dubuque, Iowa, a community known for its strong Catholic and pro-life history, John Kerry stated that he believes that life begins at conception.  However, John Kerry\'s voting record over the years and his public statements clearly show that he is one of the abortion-on-demand crowd\'s biggest supporters.  In an attempt to blur the issue and win political points, John Kerry is trying to appeal to pro-lifers and hope they ignore his voting record and his public statements.  If someone is willing to acknowledge that life begins at conception and does nothing to protect that life, it demonstrates a startling lack of conviction and politics at its worst.  Kerry Says He Will Provide Abortion Rights Supporters \"Clearest Possible Choice\" In Kerry/Bush Contest. \"And if elected president, I will nominate judges who will seek to expand [abortion] rights and opportunities--not reduce them.  In this area, anyone who values a woman\'s right to choose in particular or constitutional privacy rights in general will have the clearest possible choice between me and President Bush in 2004.\"  (Sen. John Kerry, A Call To Service, 2003, p. 182)  Kerry Says He\'s Only Presidential Candidate \"Who Hasn\'t Played Games\" With Abortion Issue.  \"I\'m the only candidate running for president who hasn\'t played games, fudged around,\' said Kerry, a Massachusetts senator. If you believe that choice is a constitutional right, and I do, and if you believe that Roe v. Wade is the embodiment of that right ... I will not appoint a justice to the Supreme Court of the United States who will undo that right.\" (Mike Glover, \"Kerry Says No Games\' On Abortion,\" The Associated Press, 1/26/04)    What others say about John Kerry   Kerry Is First Presidential Candidate To Ever Be Endorsed By Planned Parenthood Action Fund. (Laurie Goodstein, \"Vatican Cardinal Signals Backing For Sanctions On Kerry,\" The New York Times, 4/24/04)  Kerry Received 100% Ranking For 2003 Votes From Planned Parenthood.  (Planned Parenthood Action Fund, Inc. Website,  , Accessed 1/24/04)  Kerry Received 100% Ranking For 1990, 1991, 1992, 1993 Votes From NARAL.  (NARAL, Congressional Votes On Abortion, 1990, 1991, 1992, 1993)  John Kerry\'s Voting Record in the United States Senate  Kerry Has Voted At Least Six Times Against Banning Partial-Birth Abortion.    Kerry Has Voted At Least 25 Times In Favor Of Using Taxpayer Dollars To Pay For Abortions In United States.   Kerry Voted Against Unborn Victims of Violence Act - \"Laci and Connor Peterson\'s Law\"  Nonpartisan National Journal Scored Kerry\'s Votes Most Liberal In Senate For 2003. Kerry scored a Senate-high 97% liberal rating for 2003, beating out Sens. Barbara Boxer (91), Hillary Clinton (89), Ted Kennedy (88), and Tom Daschle (80). (National Journal Website, \"How They Measured Up,\" http://nationaljournal.com, 2/27/04)  From georgewbush.com '),(24047,3809990,'02,July,2004','2004-07-02 00:00:00','   Well, I really wanted to get some posting done over the weekend but was just having too much fun doing family stuff. Eating, going to church, eating, fireworks, cooking out, and did I mention eating?  Anyway, here\'s another column I wrote in March of 2003 that was published in the Alvord-Sunset Gazette.  Hope you enjoy it. - James    My Heroes Have Always Been Cowboys    According to several European leaders, Iraq\'s Hussein-owned newspaper, Babil, and many leftist Americans, President George W. Bush is a \"cowboy.\"    Defeated presidential candidate Al Gore said in Sept. 2002, that the Bush administration was guilty of a \"do-it-alone, cowboy-type reaction to foreign affairs.\"  Babil, the Iraqi newspaper owned by Saddam Hussein\'s family, urged Russia, France and China to differentiate between Iraq, \"which adopts a policy of peace, and a country (the United States) that adopts a cowboy policy.\" The heads of state of some of our so-called \"allies\" have even referred to Bush as a cowboy. Of course, by association, they consider everyone who supports the policy of President Bush toward Iraq and its dictator to be \"cowboys\" too.   They mean this as an insult, of course.   They want to imply that Bush is hotheaded, vengeful, and dangerous. A real \"shoot first, ask questions later\" gunslinger kind of guy. Maybe even an outlaw.   But they miss the mark completely.   What is a cowboy?   President Bush knows. He was raised in Texas, after all. And in Texas we know about cowboys. We learned about cowboys from the Lone Ranger and John Wayne. We grew up reading Louis L\'Amour novels and watching Marshall Matt Dillon keep Dodge City safe.   We grew up idolizing the American cowboy and all he stands for.     What is a cowboy?   A cowboy is resourceful, able to make do and get the job done with the tools at hand.   He can fix a fence with a piece of baling wire or get water from a cactus in the middle of the Mojave.   He may be \"rough as a cob\" when the situation calls for it, but he\'s gentle and kind with women and children. He recognizes that he is their protector.  His word is his bond. He is polite and soft-spoken. He says \"Sir\" and \"Ma\'am.\"  His speech may be more plain and simple than some prefer.  He may say \"folks\" to describe anyone, friend or enemy.  He uses terms like \"evildoer,\" \"dead or alive,\" \"hunt \'em down\" and \"smoke \'em out.\"  He\'s a man of few words, but when he speaks his meaning is clear.     A cowboy recognizes the Creator and appreciates the beauty of the creation, but he knows that it was created for mankind to rule over and use. A cowboy has dreams, but he\'s also down to earth.  A cowboy may have things in his past that he\'s not proud of, but he doesn\'t let them control who and what he is.   He looks forward to the promise of tomorrow, resolved to be better than he was.    A cowboy knows the difference between right and wrong. He also knows that there are good people as well as evil people in this world.   He willingly accepts his duty to help and protect the good.   He is your best friend and strongest ally if you walk on the side of the angels, but heaven help you if you don\'t for he also recognizes his responsibility to confront and destroy those who would do evil.   He will take on any task that is necessary and do a good job, not for recognition or gain, but because someone has to do it.   He can look at himself in the mirror every morning and sleep easy every night because he knows he\'s done his best to do the right thing each day.  A cowboy is independent yet loyal, free-spirited but responsible. He is reliable, honest, and resolute.   A cowboy does the right thing even when it\'s not popular and even when no one is watching because it is against his very nature to do otherwise. A cowboy is the embodiment of all the qualities that have made America the best and greatest nation on Earth.   He is what all good men truly wish to be, in their hearts, whether they are willing to admit it or not.   I think all those people may be right.   President Bush is a cowboy.   He did not go looking for this fight, but he will not back down from it. Yes, a cowboy will \"walk away from trouble when he can\" but he knows that \"sometimes you\'ve got to fight when you\'re a man.\"  We will confront the evil that threatens us and we will destroy it.   Our enemies will either change their ways or suffer the same fate.   America has been the world\'s policeman when \"criminals\" needed to be stopped or brought to justice.   Perhaps now we must be the world\'s cowboy and put an end to the outlaws and renegades who threaten civilized people.   Regardless of what his critics may say about him, President Bush will do the right thing because that is what cowboys do.   That is what Americans do. Even when it\'s not popular.  Perhaps, the next time someone calls President Bush a cowboy he should simply smile and politely say \"Thank you.\" '),(24048,3809990,'02,July,2004','2004-07-02 00:00:00','   This is unreal and in my opinion the ones who are asking for this intrusion into our sovereignty should face public outrage and whatever official censure might be available. - Jof '),(24049,3809990,'02,July,2004','2004-07-02 00:00:00','   I originally wrote this in February of 2003. It was published in the Alvord-Sunset Gazette in Alvord, Texas.      Great Satan or Great Savior?    This thought occurred to me recently:  If the French and Germans and the rest of our \"friends\" and \"allies\" along with the \"anti-war,\" hate-America Left truly wanted to avoid war then they would support our efforts against Baghdad and join the coalition with our true friends.   The surest way to secure peace is to prepare for war. The only thing that might possibly bring Iraq into compliance with the will of the world is the threat of destruction. Appeasement will not do it. Coddling will not do it. Portraying our President as a wild-eyed cowboy itching for a fight will not do it. Painting Hussein as a poor, maligned victim of American arrogance will not do it. These things only strengthen his belligerence and his resolve to delay and resist.  These things only give him more time and more support to gain the material, technology and capability to threaten the entire world.  Do we really want to wait until he has the upper hand? Until he can hold the world hostage to his will? I can remember when we all lived in fear of a global, nuclear war.  I remember what the phrase \"two minutes to midnight\" meant.  I for one don\'t want to go back there and I certainly don\'t want my children to live there.  The only motive these enemies of freedom have for their words and action is the slandering and weakening of America and all freedom loving people. America does not conquer other nations, we liberate them. America does not occupy other nations, we rebuild them (or build them up for the first time). America does not oppress other nations, we build schools and hospitals. America does not terrorize the children of other countries, our soldiers build them playgrounds and collect clothing for them, usually with money from their own pockets. Americans do not murder the men or rape the women of other nations, we teach them the sanctity of human life and respect for woman and children, the elderly and weak.  America is not the Great Satan, we are the Great Savior. We saved Europe from the Hun in World War I. We saved the entire world during World War II. Let the truth of that sink in.  America saved the world.  Yes, we were part of a great alliance.  Yes, other countries made significant contributions of lives and materiel.  But the war was definitely going in the favor of the Axis powers before we got in the fight.  Europe, Asia and much of Africa were already overrun. England, though fighting valiantly, could not have held out indefinitely. Only with America\'s great industrial might devoted to the production of war materials, only with our men fighting, bleeding and dying, only with our men and women on the homefront producing weapons and supplies was the tide turned and the day won. We saved the French from the Nazis. We saved the German people from Hitler. We saved all of Asia and the Pacific region from Japanese Imperialism. America saved the world.  We saved South Korea from Communist domination in the 1950s. We saved most of Europe from Communist invasion throughout the Cold War. Our resolve and power destroyed the Soviet Union. We saved the people of Panama from Noriega. We saved Kuwait and the rest of Southwest Asia from Saddam Hussein. We saved the Balkans from Milosevic. We saved Haiti from Cedras. We ensured the delivery of food to starving Somalis. We provided medical equipment and supplies to the reforming Soviet states. We conduct counterdrug operations throughout the hemisphere to rid the world of the scourge of illegal narcotics. These and countless other operations were conducted in the 20th century alone at a cost of millions of our dollars and many thousands of American lives. None of these were fought so that we could steal oil or annex countries.  They were fought because someone must do the right thing.  Perhaps the Germans have forgotten \"Operation Vittles,\" the Berlin Airlift. In June 1948 the Soviets blocked all land routes into Berlin in attempt to cut off the city and drive out the western powers. They tried to blackmail us through the starvation of Berlin\'s citizens.  Between June 1948 and September 1949 US and British planes delivered 2.3 million tons of food, coal and medicine to the citizens of Berlin through 280,000 flights. 75% of the goods were delivered by the US - at a cost of 31 American lives.  The Soviets wanted to starve the people of Berlin.  We wanted to save them, and we did. We did it because somebody had to do the right thing.  By the way, Soviets are what we used to call the Russians, the new friends of the Germans in their diplomatic war against America. Politics does indeed make strange bedfellows. '),(24050,3809990,'03,August,2004','2004-08-03 00:00:00','   Please go to the following links and read about some real heroes.    urlLink Heroes of Fallujah    urlLink Lance Corporal Aaron C. Austin - A Texas Marine     \"Greater love has no one than this, that he lay down his life for his friends.\"  John 15:13 '),(24051,3809990,'02,August,2004','2004-08-02 00:00:00','   Time to \"Feed the World\" again!   Please forgive me as I digress from politics (my wife says I\'m too obsessed with that anyway) and rave about something completely different. My other obsession, 80\'s music!  If you were born before July 13, 1985 then you probably remember the greatest musical event of our generation, Live Aid. The all-star concert was conceived and organized by Bob Geldof (Boomtown Rats) and Midge Ure (Ultravox) as a \"global jukebox\" to raise money for famine relief in Ethiopia. The main venues were JFK Stadium in Philly and Wembley Stadium in London, with performances in Moscow, Belgrade and other cities around the world. Phil Collins even performed at both concerts, flying across the Atlantic on the Concorde. It was a veritable \"Who\'s Who\" of music, especially 80\'s music: Duran Duran, Howard Jones, U2, The Beach Boys, Black Sabbath, Judas Priest, Madonna, Elton John, Wham!, and dozens of other artists. It was a lot like Woodstock but bigger, better, and with not nearly as many hallucinogens.  1.5 billion people watched that day, but I had three problems. We didn\'t have cable, we didn\'t have a VCR, and I had to work. I was reduced to listening to parts of it on the radio. Of course, you could try to buy copies on eBay. The only problems with that are the quality probably stinks and it is, of course, illegal. In fact, Geldof has successfully prosecuted a lot of pirates selling bootlegs.  Well, fret no more, 80\'s fans! After almost 20 years Bob Geldof has persuaded all the performers to agree to a DVD release. Warner Vision International has won the bidding and on Nov. 1st will release a four disc set containing 10 hours of performances, a documentary called \"Food, Trucks & Rock \'n\' Roll\" and other gems. The footage is being remastered from the original tapes. By the way, it was initially hoped that Live Aid would raise about $1.5 million, but the grand total came to over $245 million. The DVD sales will only add to the good work Live Aid will continue to do.  Here are some links to read more about Live Aid and the DVD. All I can say is,   \"Awesome!\"      urlLink World IQ - Live Aid    urlLink The Unofficial Live Aid Site    urlLink Warner DVD News '),(24052,3998459,'24,June,2004','2004-06-24 00:00:00','   Oh, I am stoked! I bought a ticket to go a visit Maggie! YEA! I was really bummed, I really didn\'t think I\'d get to see her this summer, but I WILL! I\'m really excited to go and see her.  Summer school has started and Bob, the drummer/co-writer/backing vocals for my band,  urlLink Stericks , and three more of my good friends are all going to be at my college\'s research facility, which is located right by my home! It\'s great, I have stuff to do on the weekends! They are crazy busy during the week, they are taking a years worth of class in three weeks or some ungodly deal like that.  Anyway, summer is ok, it\'s so much better than last year. So I\'m grateful.  I miss everyone who I can\'t see. You all mean a lot to me!  '),(24053,3998459,'21,June,2004','2004-06-21 00:00:00','   Alright, so the weekend happened. Saturday, my family had to go to this picnic at a local state park. It was good to see a bunch of people I haven\'t seen in a long time. I got a chance to kind of show off my portfolio. Good for me.  Maggie and I talked a bit, not a whole lot, she had a party Saturday afternoon. Got lots of graduation money, that\'s always good. She then took off to her internship, so we talked a bit on Sunday in between flights  Father\'s Day was just bad. We all woke up feeling like it was going to be a bad day. Maybe a self-fulfilling prophesy. Dad was just grouchy all day, and when my sister tried to talk to him about his gift he just kind of blew her off. Really made us want to get him something, oh well, seems to happened on \"special days\". There\'s something about the pressure to happy that makes everyone edgy. It\'s too bad. It\'ll get better.  Well, I\'d better get back to work. Hope all ya\'ll who read this have a great day!  '),(24054,3998459,'18,June,2004','2004-06-18 00:00:00','   I\'m sitting here in the Anacortes Starbucks. Thanks to the guys at Garnet Web, I get Wi-Fi here. I\'m on a bit of a break before work. It\'s nice here, probably somewhere around 80. Seems just about right after being in Walla Walla. Oh, well, whatever.  I talked to Maggie this morning, she\'s returning a bracelet she got for graduation from her parents. It\'s REALLY nice, but she doesn\'t really wear bracelets and she wants to have something she\'ll wear a lot. She\'s looking at rings and necklaces. The bracelet is either white gold or platinum (I didn\'t find out for sure), and had like .9 carets of diamonds, so it wasn\'t cheap... Lucky girl.  Well, that\'s about it for now. Yoga wasn\'t so great this morning. I think I\'ll keep looking, maybe I\'ll end up getting a video and do it at home, we\'ll see.  Well, have a good one!  '),(24055,3998459,'27,July,2004','2004-07-27 00:00:00','   urlLink'),(24056,3998459,'12,July,2004','2004-07-12 00:00:00','   I have, by far, the coolest girlfriend, ever!   Maggie and I and the way to Jimmy Buffet  Her dad got tickets to see Jimmy Buffet, and she talked him into getting me a ticket. It was SO much fun! He\'s the king of hedonism. The parking lot was the biggest tailgate party buffet I\'ve ever seen. It was ten acers of kiddie-pools filled with hard liquer, wet T-Shirt contests, jello shots, pirate flags, shark decorated minivans, and so many CRAZY things! Some people hauled in a full tiki bar and like a 20 foot square patch of sand! I could not believe how many people were fans.    As you can see, Jimmy has a Taquila named after one of his songs...  Maggie and I danced and sang the bits of songs we knew, got covered in dancing drunk fan\'s beer, and smelled \"the weed\" a lot. I could NEVER live like that , but it\'s kind of fun to visit it once and a while a get perspective, or maybe I just secretly want to be a pirate... ARRRRRR! *Swash, swash, buckle, buckle*  Anyway, talk to ya\'ll later...  '),(24057,3998459,'07,July,2004','2004-07-07 00:00:00','   Well, it\'s been a while since I\'ve posted, but I guess there are adages for that, huh? Anyway, I found a really good price for a flight back to Illinois, so I jumped on it, and now I sit, in pre-flight, waiting for the gates to open and allow the flood of really unhappy travelers to surge down  the poorly decorated, tan hall way, and burst through the the cabin door, to be greeted by a southern stewardess named Sandy.  \"Hi-ah, waly-cuhm to flaht number tew say-van fowr. Nown stahp sur-vi-ehs to Chic-ah-gow Meed-wahee. Thu cri-oo woooood lahk to thang-k yoo fowr flah-yin Ayee Tu-ee Ayee Ehr-lahns! Nahw, in-joe-y the flaht noahw, ya heer?\"  Some how, the thought of putting my life in this women\'s hands is scary,  somehow comforting, but not reassuring in anyway. Although, the thought of her perky self fly the plane is almost better than a pilot so sleep deprived he can\'t remember he\'s flow from Orlando to Las Vegas to Portland and hopped a commuter to Seattle.  I guess I\'m just ranting now. But, seriously, we go through like five hours of security checks only to find out the planes are pretty much flying themselves.  Here\'s an annoying side note: I go into a Starbucks that is conveniently located at the exit of security, and get a Black Tea Lemonade. I ask the guy if they have a T-Mobile Hotspot, and he says no, bummer for me. But, he says, he knows that this corridor has wireless internet! Great, I say, and I thank him. I get my drink, and find a seat by a plug in. I start up my computer and wait for it to boot. It does. But, lo and behold, no Wi-Fi service, at least my computer doesn\'t think so. So I check it\'s settings, and MY COMPUTER LIED! There it is: ERAIR. So I click on it, and it asks for a password. Well, suffice to say it\'s evidentially an internal work wireless, \'cause I CAN\'T GET ON IT! Kind of bothered me. So, now I\'m typing this in NotePad and saving it so upload later. I\'m also not getting any work done, which in all actuality has nothing to do with there being no Wi-Fi and more to do with me being lazy. Wireless internet tends not to help laziness. Therefore, I\'m perturbed.  I saw The Terminal with Tom Hanks and Catherine Zeta-Jones, about a week ago, and if you\'ve seen it you know that on of Tom Hank\'s character\'s friends is an Indian Custodial worker. He kind of walks hunkered over and likes to watch people ignore the \"Wet Floor\" signs. Well, as I\'m sitting here a custodial worker for SeaTac walked by pushing the exact same cart, and walking in the exact same hunkered way! If the guy in The Terminal was working at SeaTac and not JFK, and if he was of European decent and not Indian, that would have been him. Fun for me, none for the people who didn\'t see the movie.  There\'s a play set and the end of this seating area. You know like the ones that came out in the late 80\'s/early 90\'s. They\'re blue and green, with a red slide, and they look like they\'re inflatable, but really they are made of hard plastic. They are reminiscent of fantasy cottages. Anyway, it\'s just sitting there at the end of the alcove that houses a couple of gates, A4 and 5, I think. Kids are crawling all over it, but it seems so out of place. They should be in a backyard, with a tall wooden fence, and have a grass floor, not a carpeted one. I think I may be becoming WAY too picky, but honestly, who puts a other yard accessories in an airport? You don\'t see garden hoses, and plastic Flamingos, and mailboxes! Not to mention the lack of that tall wooden fence. I think that\'s fairly definitive. Yes? No? Maybe?  Well, I suppose I should wrap this up, I\'m supposed to board in fifteen minutes and my plane doesn\'t seem to be here. That\'s cause for concern. Maybe it\'s on the other side... Maybe it\'s been delayed... MAYBE I\'M AT THE WRONG GATE! That\'s not good. All this negative thinking has got my all mixed up.  So, one last thought to leave you with. Do you ever feel like, when you\'re inline to board the plane, and even though you have an assigned seat, and you are guaranteed that seat, it says so on the ticket, that you will get to the front of the line and the lady will go, \"I\'m sorry, planes full.\" And then pull the little tube-like gate thingy closed infront of you? I do, and it it keeps me up at nights!  Well, the plane showed up, good for it. I guess I\'ll write sometime else. I\'m good like that.  '),(24058,3998459,'16,August,2004','2004-08-16 00:00:00','   Saturday was a very good day! I was at a Leadership Conference for the first part of the day, really encouraging, I needed it! Anyway, yea for me.   Lauren and Seth were just one of those couples that everyone expected to get together, they were really close friends all through Middle School and High School, but they never dated each other. They dated others, but no one ever really fit. They finally got together last summer. It was SO special to see them together, they were very obviously in love, but you didn\'t feel uncomfortable around them.  Lauren and I graduated togther, played in our Christian High School\'s worship band together, and were pretty close friends. We had a mutual friend that we were both close too, and so it was just natural for us to be friends. She\'s a great person, I have HUGE amount of respect for her. OH, and she also got me a job at a fruit stand one summer when I couldn\'t find a job. I ended up getting fired from the job, but she didn\'t have anything to do with that (I don\'t think).  Seth is an AMAZING drummer! We, too, played in the worship band together, but we really got to know each other just because the two of us are crazy, and we have a blast with each other. He plays drums in the band  urlLink Mindhead , along with another high school friend of mine, and I\'ve become friends with them through that as well.  Ironically, of all the people I knew and were friends in high school, Lauren and Seth are ususally two of only a small handful I\'m still sort of close to. They are truely good friends!  Seth Fikkert and Lauren Burr were married at sometime just after Seven Post Meridiem, August Fourteenth, Two Thousand Four in the Year of Our Lord, Jesus Christ.    Seth and Lauren at their reception  I wish these two an amazing life together, they are good people and it was so awesome to see them together! Conrgrats guys!  '),(24059,3998459,'09,August,2004','2004-08-09 00:00:00','   Well, the summer hits just keep on coming! My dad and I went and saw the Village People and KC and the Sunshine Band! It was SO much fun. Very surreal for me, but kind of oddly reminicent for him. I guess that I have this really romanic view of the 1970\'s. I keep assuming that the songs I hear on the radio and in movies was what everybody listened to. I just assume that there was one type of music. People had their preferences, but basically you either listened the radio or you didn\'t. Well, evidentally, that\'s not the case. My dad was NOT into Disco at all, and come to find out, it would be like being into Pop music today. KC even compared himself to \'Nsync, saying that he was my generation\'s mom\'s Justin Timberlake. It was really enlightening.  I really felt odd, in that I never expected to see any of these people, they are out of another world, a past society that can\'t be accessed any more.  The Village People were SO ... the Village People! They didn\'t disappoint at all, nor did they exceed expectations. Just right. The guys were all the original line up, minus the Motorcycle guy (he quit in the 80\'s and the picked up a new guy).   The Village People Preforming  The Village People also informed everyone that most of the US has been doing the YMCA wrong. so they had this whole tutorial. It was REALLY funny! Well, I\'m doing my part and spreading the word, here you go,  urlLink click here  for an animated tutorial.  It was SO much fun, my only regret was that Maggie couldn\'t be there. My dad isn\'t the best dancer in the world, and it was just better to sit and watch.  '),(24060,3998459,'04,August,2004','2004-08-04 00:00:00','   urlLink'),(24061,3998459,'02,August,2004','2004-08-02 00:00:00','   Alright, I\'ve started this post like three times, I\'ve uploaded the photo and then ... I space out! Anyway, I\'m going to follow through this time.  OK, so I know I left IL like three weeks ago, and I should have posted this like the day after it happened but, I didn\'t. So, deal.  As far as show value Jimmy Buffet wins hands down, but for venue: K.D. Lang at Ravinia Park. The place is AMAZING! Ravinia Park is, as it sounds, a park, except it\'s built around a large ampatheater. Also, adding to the unique quainess to the show, is that people treat going to a concert at Ravinia like a day at Ascot. Crowds show up in droves hours early, so that they can setup their elaborate picknics. People actually bring candelabra and table cloth-covered wagons come table and fine wines and foods and, coolest of all, some even bring chandelars to through over the baughs of the trees. It\'s an amazing experience!    I didn\'t get a great shot of the stage or the park (I\'m a smart cookie), so here\'s a cute one of Maggie and I!  Maggie, her friends Aaron and Etson, and I went to the show together. We brought goat cheese and water crackers, stuff for sandiches, a very good Brie-wrapped-with-pastry-sheets-and-brown-sugar-thing that Maggie made, some non-acoholic Cabernet Sauvignon (long story, but someone won\'t be celebrating anytime soon), and I don\'t know what all, oh yeah, grapes, really good green grapes. Etson ate most of them...  After we ate and listened to the concert for a while, Maggie and I decided to walk around and look at all the scenery. Aaron was studying for something so we left him and Etson, who was still eating, to their own devices.    Aaron studying at Ravinia...  We walked among the gay and lesbian couples, K.D. Lang is a fairly proud lesbian, so political as well as musical supporters came to the show, looking at all the really well decorated tables surrounded entirly by men. Lesbians lose points for decor, sorry, but it\'s true. Anyway, we walked through the park and it was the most amazing movie moment of my life so far. Ms. Lang was singing a Tony Bennett cover, the sun was about to set, and the twinkle lights in the trees were on. Maggie and I danced under a very large shade tree, and it was REALLY romantic.  Anyway, we had a great time!  '),(24062,3723604,'30,June,2004','2004-06-30 00:00:00','    Sweet, it\'s Wednesday, that means Spiderman2 comes out.  God this movie is going to be cool. I\'ll probably see it on Friday. Then I\'m gonna go buy the game. GAH Spiderman is such a badass, I\'m wearing a Spiderman T-shirt RIGHT NOW.  If you didn\'t like the first one, then screw you, quit reading my weblog.  As for movies, I\'m going to see if Chronicles of Riddick is worth seeing by downloading and watching the prequel, Pitch Black. From what I hear, it kicks ass.  That\'s it for today.  -Reb   '),(24063,3723604,'28,June,2004','2004-06-28 00:00:00','    After going into a severe coma after eating a bowl of Double Fudge Brownie Ice Cream, and then getting brutality -ercised by a treadmill and some 10lbs weights I think I feel fine now. It\'s about 12:45AM on Tuesday \"morning\"  and I\'m about to hit the sack... bed.  How did Blogger manage to fuck up the two posts on \"Saturday\"?   So, here\'s Monday\'s post:  [MONDAY POST]Monday, ah... a start of another week of freedom.... Well, my last one until summer school starts.  I\'ll be taking Health. Four hours of health.. whoopdey damn do.  Well, my bathroom is done. It doesn\'t look too bad, either. I just need to build the shelf that will hang over my toilet.   Speaking of toilet, one word of advice: Never eat double fudge brownie icecream... EVER. Unless you have lots of patience and good magazines laying around.  -REB [END MONEY POST]  There you have it folks. Two days of posting in ONE post. Hope you enjoyed it.  -reb   '),(24064,3723604,'26,June,2004','2004-06-26 00:00:00','    By early, I mean 1:33AM Eastern time.  Well yesterday me and my dad painted my bathroom, which is now a dark shade of blue. It actually looks like someone when apeshit on a smurf in there.  Today, we\'re going to put up a border between the blue paint and the white paint. It\'s a 50/50 mix, so the effect will be pretty sweet. Alas, when my dad was taking off the floorboard, he snapped a piece off. He\'s going to Jew it back into shape, but it\'s like \"Damn dude, be careful next time.\"  I\'m finally going to go all out on my diet. Screw this weblog, I\'m going to shed some pounds before school starts.  -Reb, future GQ cover model.   '),(24065,3723604,'25,June,2004','2004-06-25 00:00:00','    Last night I was over at my friend Scott\'s house with Brice. At around midnight, we loaded up a Faygo bottle with tin foil and toilet cleaner and made Drain-O bomb. We left it in the middle of the street and it blew up about 5 minutes later with a KABOOM.  Today I feel like crap. I\'m tired, I smell and I can\'t really think straight, but I\'m operating fine on the outside.  It\'s when I have to move or speak, I have trouble.   My dad and I are going to redo my bathroom. We\'re going to buy paint, and paint it, as well as clean it and install a new showerhead. Yay, maybe we can tape it and send it to This Old House. Or maybe it\'s This Olde House. Whatever, that guy on it is a beaver.  -Reb   '),(24066,3723604,'25,June,2004','2004-06-25 00:00:00','    Today\'s Friday, which sucks for me and is a pot of gold for the working folks. What\'s that? Friday sucks for me?  Yes, because that means the weekend, which means my family will be hovering around the house, trying to get me to \"do things\" like laundry, trash, drugs etc...   I\'m also lacking in typing ability, since I had to rewrite that last sentance about four times. (and this one about three).  I better quit posting daily, that would bring people to accept it and check in obsessively waiting to see what Reb posts next. They\'d come to my site twice an hour everyday to see if I\'ve updated, and if I haven\'t they\'d send rutheless, violent e-mails forcing me to post out of self-pitty for neglecting my blog.   Cripes, I never knew something like this could be such a burden.  Late update- Things overheard at a gay wedding #04: \"I\'m not losing a son, I\'m gaining a, uh... a son?\" #07: \"Can you zip me?\"  -Reb   '),(24067,3723604,'24,June,2004','2004-06-24 00:00:00','    Today is definately a slow news day. Nothing important has happened to me today. I\'ve just been sittin on my ass, wishing I was doing something...  On a lighter note, I\'ve started a new diet, which was going great until about an hour ago when I had dinner.   Throughout the day I ate (sparingly, mind you) celery, a peach, tuna and a hand full of marshmellows (the mini ones). So far so good... then comes dinner. I had a bowl of rice, stir fry and french fries (the ones you cook in the oven) and I\'ve been regretting it ever since.   Tomorrow I might have to cut down on some of the eating, just to make of for today. Dammit!  -reb   '),(24068,3723604,'23,June,2004','2004-06-23 00:00:00','    Everybody!  I got a G-mail invite! What\'s G-mail? Google\'s new e-mail service! What\'s Google? Go back under the rock you\'re living under!  Anyways, if you need to contact me, my g-mail (I love saying that) is rebelzeppelin[AT]gmail[DOT]com.  If you want to get in on the \"action\" of having a G-mail beta account, go to www.gmailswap.com  There you will find people willing to invite people into G-mail for a small price: the price being ANYTHING you could think of.   What did I offer for my g-mail spot? The MEGABOT pic from yesterday ;P   -REb   '),(24069,3723604,'22,June,2004','2004-06-22 00:00:00','    Hello, folks.  For the two of you that will actually ever read this, you can call me Reb.  I am free for idle chatting on AIM @ Rebelzeppelin or you can email me @ NIXED see above.  I will use this weblog to chart my summer vacation at home, and I\'ll occasionally post when I get bored.   Today, I drew a robot: http://img78.photobucket.com/albums/v239/jdbrewer/untitled.bmp  Check it out.  First post! woop woop!  BTW- I\'m aware of the issue that the labels on the robot CANNOT be seen. I can\'t fix this, please deal with it, and blame photobucket.com for sucking.  -reb   '),(24070,3723604,'28,July,2004','2004-07-28 00:00:00','    8/14/04'),(24071,3723604,'25,July,2004','2004-07-25 00:00:00','    Woo, post.    Read this, and waste 3 seconds! Garunteed.    Hah I love fucking with the internet.    Shut up.    ASDF    ~Reb   '),(24072,3723604,'19,July,2004','2004-07-19 00:00:00','    I think lawn mowing is pretty stupid.    Just got done with my one acre yard. I had to weed whack(not a sexual comment) and now my legs are all scarred up from GRASS BLADES.    Since when did gras cut people? I thought that was just saw grass...    For all you Rebophiles, I\'ll get more PIX SOON LOLZ WTF!!!1    Umm... let\'s see. Guns \'N Roses are teh roxor. I hate Slipknot... and yeah.    Done.    -Reb   '),(24073,3723604,'17,July,2004','2004-07-17 00:00:00','    urlLink    HAIRCUT!!11  urlLink         Today I got a haircut.      Also, everyone around the house seems unneccessarily pissy.      Oh well, more when I get BOR3D!      -Reb   ');
INSERT INTO `msgs` VALUES (24074,3723604,'14,July,2004','2004-07-14 00:00:00','    I actually beat the first Mega Man game(NES).  Wow, what a run, I can\'t believe I did it.  Anyway...  Health is half-way over. And it\'s almost Friday. Someone got caught smoking in the bathroom today, too.  Cripes, what a dumbass.  \"I know, I\'ll smoke a cigarette in the bathroom... no one from 3 hallways down will be able to smell the SMOKE.\"  Idiot. I hope he\'s expelled.  -red   '),(24075,3723604,'13,July,2004','2004-07-13 00:00:00','    urlLink    Here I am, seconds before death... Before the kitty strikes...&nbsp; urlLink     This is my cat, Velvet. She is a forocious animal. She attacked me when I tried to feed her. There was no avoiding damage once she got on top of me. I am now in the South Bend General Hospital recovering from scratches, bites and cat piss stains.   Pray for me people... pray for me.  -r3b   '),(24076,3723604,'10,July,2004','2004-07-10 00:00:00','    urlLink    Me, looking somewhat cool&nbsp; urlLink     I totally got a new digital camera. So I can take pics of me looking somewhat cool. I\'ll post a few throughout the next couple days or so. At least until I get bored   '),(24077,3723604,'08,July,2004','2004-07-08 00:00:00','    Thank God it\'s almost Friday.  Last night I went to bed at 7PM, woke up at 6AM. Worst sleep of MY LIFE. Christ, it sucked. I tossed and turned all freggin night.  Summer school could be a lot worse. I\'m getting a 91% in the class. You make think \"hey not bad\". BUT ITS EFFEN HEALTH.  Oh well, we watch at least one movie a day, and I only have 14 days of class left. It blows, because I\'m always tired afterwards, and I feel like I\'ve wasted my day -.-   Anyway... POST!  -reb   '),(24078,3723604,'03,July,2004','2004-07-03 00:00:00','    Boredom is never a reason to post. Unfortunately, that\'s exactly why I\'m posting. It\'s raining, I still havem\'t seen Spiderman2 and I\'m a lil\' bit irritable.  I\'m going to go trip out to Floyd.  Durp  -Reb   '),(24079,3723604,'01,July,2004','2004-07-01 00:00:00','    WTF is wrong with me? I almost went a day without BLOGGERCISING!!!  Today sucked. Yesterday rocked. I downloaded and watched Pitch Black for the first time. It was great, I loved it. Then I watched Dark Fury, which is the anime that takes place RIGHT AFTER Pitch Black.  Wooo, so exciting. Make sure to visit www.rooneycom.net and get your $5 ***@rooneycom.net email. Supplies are unlimited, but hurry up regardless! Also, make sure to buy LOTS of RooneyCOM gear.  That\'s all for Thursday, which sucks, because Friday\'s are Hell! But I might see Spiderman2, if anyone will come with me.... >.  -Reb   '),(24080,3723604,'11,August,2004','2004-08-11 00:00:00','    urlLink You\'re fucked . We\'ve seen your work.  signed,  The Illuminati ~+~   '),(24081,3723604,'10,August,2004','2004-08-10 00:00:00','    Check out these sexy new cars:  The Pontiac  urlLink  Solstice    And the Chrysler  urlLink Crossfire   Which one is better? Who knows but they are worlds better than  urlLink this hunk o\' shit   BTW, in retrospect I think Pontiac made the Solstice to make up for the crapfest that was  urlLink THE AZTEK    Normally I would\'ve saved and edited the pics of cars to say something hilarious, but this was a \"I have a thought, let\'s share it\" type of post.  On a sadder (more sad?) note I was [---]   ~reb  urlLink   '),(24082,3723604,'06,August,2004','2004-08-06 00:00:00','    Today I woke up at around 11 0\' clock.  From then until about 4PM I was helping my dad build my 2-year old nephews sandbox.  This isn\'t just a small sandbox... this is the biggest mf-ing sandbox EVER.  It\'s 12 x 12 and has  SIX   TONS  of sand in it. There\'s a four foot mound of sand in the middle.  I wish we had done this about ten years ago. My army men never really had a place to put the smack down on eachother.  I am one sad, and tired, panda.  ~Reb   '),(24083,4047559,'27,July,2004','2004-07-27 00:00:00','   Vanaf nu is Aerodynamik voor 10 dagen de leider van onze clan! Vragen of opmerkingen of andere zever --> AERODYNAMIK :p:p:p '),(24084,4047559,'27,July,2004','2004-07-27 00:00:00','   Tot over 10 dagen dan beste mensen :) '),(24085,4047559,'27,July,2004','2004-07-27 00:00:00','   Morgen zal de site beheerd worden door Aerodynamik wegens mijn verlof:p Veel succes Aerodynamik! '),(24086,4047559,'27,July,2004','2004-07-27 00:00:00','   Tot slot moeten we natuurlijk nog een woordje zeggen over de veelbesproken contextgevoelige advertenties die je te zien zou krijgen. Wel, in de praktijk blijkt dat allemaal nog goed mee te vallen. Advertenties verschijnen netjes in de marge, net zoals dat het geval is met de gesponsorde links in de Google-zoekmachine. '),(24087,4047559,'27,July,2004','2004-07-27 00:00:00','   Gmail heeft ook een erg handige zoekfunctie waarmee je aan de hand van trefwoorden heel gemakkelijk je mail kunt terugvinden. Ook handig is dat e-mailconversaties als een pak kaarten op elkaar gestapeld worden. '),(24088,4047559,'27,July,2004','2004-07-27 00:00:00','   Aan elk mailtje kun je een of meer labes toekennen om zo orde te scheppen in je inbox. Dat is ook nodig als je echt zinnens bent om bijna al je e-mail te bewaren in je postbus die tot maximaal 1 gigabyte kan bevatten! '),(24089,4047559,'27,July,2004','2004-07-27 00:00:00','   Gmail mag dan wel \'bÃ¨ta\' zijn, daar is niks van te merken in de webinterface. Die heeft er zelfs geen probleem mee als je even offline gaat. LInks heb je mapjes als Inbox, Starred (zoals het vlagje van Outlook), Sent Mail en Spam (3). Gmail filtert spam automatisch en doet dat behoorlijk goed. :) '),(24090,4047559,'27,July,2004','2004-07-27 00:00:00','   OSM isn\'t any longer supported on the site! Sorry for the inconvienience. We hope to fix this soon. '),(24091,4047559,'27,July,2004','2004-07-27 00:00:00','   BOEE!HI HI HELLO         Wolfenstein   Posted: Jul 25 2004, 07:21 AM      Administrator   Group: Admin Posts: 14 Member No.: 1 Joined: 21-July 04  Warn: (0%)    o.O  just what I thought  , your real crazy naughty azzhole... '),(24092,4047559,'27,July,2004','2004-07-27 00:00:00','   1 versus 1... an duel between 2 people with 1 goal...\'victory\' . 2 random nations side against side wich will I get? 10 minutes of peace time in a normal game...  First I\'v immedially began to build with the nation i did get...Poland. I\'v used my usual strategy and so started building building by building. And I realized that things went quite well! After 7 minutes or so I\'v got my diplomatic center. In the meanwile I was building up my army with pikeman. I set them up in a squad of 72 pikeman. When my diplo was build i started building archer and grenadiers and set their rally point into my army of pikeman. I\'v got quite a big army but I saw that my opponent who played with Russia was very good too. It wasn\'t going to be an easy game. And it wasn\'t.  When I thought I was ready to go to battle I set up my army as I wanted to be. And of I went. With my army of archers,grenadiers,pikeman and some sich cossacks I reached his defences (or attacking force, whatever you want to call it) and then on a few meters of them I stopped. 2 armies facing eachother...1 of us who whas going to win this battle. And I\'v attacked first with mij pikeman his pikeman and luckely I saw it was going well. But my archers, thats another story. Russia send lots of sich cossacks to my archers but quick I could stop them with my grenadiers but not before a great deal of damage was done to my army. I decided to retreat with my army (or what was left of it) and regroup. But my enemy used that moment to attack my army too!  When I finaly reached my \'regrouping point\' the half of my army was killed in his attacks. But when he came near my regrouping point he made a big mistake. He first sent a wave of archers into the frontlines wich I easily killed. And before he know what was going on his army of archers was totaly destroyed. Now I hurried with my new army to his base wich was now as good as undefenced and destroyed his diplomatics center and his barracks. Now it wasn\'t difficult at all anymore to gain victory with al his fighting forces destroyed and his important buildings destroyed and finaly after quite a lot of time: VICTORY! '),(24093,4047559,'27,July,2004','2004-07-27 00:00:00','   Truky? Poland? Spain? It was random nation so you did not knew what nation you would get. The game started as 1v1v1 although soon we saw that I was the most strongest off all. So both enemys where afraid I would atack em. But I only atacked the best one....(enterprise) I beated him quite fast 72 pike vs +/- 40 so noty real warr ...So I was walking in base then I was being atacked by shenron.. I managed to survive the atack and was fuly going on shenron... This was hard because he managed to build up a huge army, and I lost lots of men vs enterprise. I atacked shenron again I can close I managed to get a blcksmith,and 2 townhalls but then shenron had made very much sich cossacks and then.... he won ground on my in my base.. I used all my peands to stop him. man we fought like hell. Then his siches where down... but I managed to make a 2th 17th century baracks... so 2 time more pikes I aatcked again, and again lots of sich cossacks game but this time I won them and continued in his base. Meanwhile he had made a new blacksmith only that one did not lived long couse m naughty little archers..then his defence felt he had not much gold left so he could not make more sich cossacks I can in his towen and demolisted some buildings....then he should watch so I left some peands alive... but suddenly he deleted all and became defaet.. (meanwhile enterprise managed to knock of all my pikes in his base and he was making a new army!) then I looked enterprises base @#$%&! damn not so good you can imagine(he ahd 2 baracks from the start off) then I was lucky that my archer atempt worked on enterprises diplomatic centre and his blacksmith menaing his aarmy woyuld not grew annymore... meanwhile I made 100 siches who where all done then we went in a Huge battle... his pike archers and grenadier vs my pike sich and archers...I won bnut not on low cost I had only 25 sich cossacks left... adn a hand full of pikeman(50) out of the 150 who went into battle then I started to demolish buildings the suddenly enterprice said that was a good game and left.  the end   it ended in 2v1 like you see.. unless this war we are still friends stronger even our friendship ever has been. this is also good to train allot like it started in training... even I trained and learned alot of this... TIP: nation doesnt matter turky+pol vs spain and spain can still win, it depends on Luck and Tactic  Aerodynamik '),(24094,4047559,'27,July,2004','2004-07-27 00:00:00','   Nation: wich nations to take in Millions, the best nations are : Denmark, Prussia, Ukraine, Saxony, When no powernations (England, France)  -Denmark couse their musketeer wich are build very fast and can shoot very high 117.  -Prussia couse their hussars build faster + cheap stables Bad point muskets are expensive and build slow.   -Ukraine couse their serdiuk they don\'t shoot very much in a single shot but this units fire rate is much faster then normal musketeers + their hetman (450 hp, attack power 350)  -Saxony couse their grenadiers they are build extreme fast their fire power aint so high but you will have allot of shooter(more then when you have musketeers)  -When no powernations England couse they have cheap stable, so more hussars  -When no powernation France couse they ahve cheap stable, so more horses + special dragoon from stable max shoot power 89  How to build:  TIP: (use the shift butten, hold shift while placing buildings to let peasands continue to next building)  Use 18 peasents, lets them make 1 townhall and 1 mill, meanwhile they are building the buildings select 18 peasents who make a 18th century baracks (repeat this 6 times). 6 Peasents will remain let them make a blacksmith. When the peasents are done making the townhall and the mill let them make a storehouse, folowed by a market. The peasents from the blacksmith will make a 17th century barrack, folowed by a academy. (Do the fore harnasses upgarde in the blacksmith mounted units building speed 4 times faster), when the market is done move that peasents also to the academy to let them build it. When the academy is being build you should do the market trick noteyou are the left side from the 2 places where to select resoursers). Select 25000 gold and trade this to iron, then select 825000 iron to gold notehold shift and push on +1000 in order to select 100000 at a time) do the same for coal to. When you finshed your market trade you academy should be almost done. When the academy is completed do the \'\'raise miner salary -75% building time\'\' upgarde cost 9050 gold, this will help to complete the 18th century baracks faster. When this upgarde is done do the reduce musket cost the the academy and start making musketeer from the 18th century barracks, do all fire rate upgardes and the fire power upgrades (firepower upgardes in this line 30%-20%-10%-15% this gives max shoot power). make 2 more 17th century barracks (total you will have 3) do the officer defence and attack upgrade and then make unlimited oficer. Make 5 stables wich make unlimited hussar, and maka diplo where you should make around 500-600 dragoons do all the other upgardes at the blacksmith accept the food and building defence ( use the building defence from the academy that one is better). Make also 6 cathedrals where you make priest (they can heal a little of your soldiers a bit) When you are around 7.30-8.00 minut\'s start making out of 6 18th century baracks pikeman.  Army lining up:  Use the attack underway button with all your units except your pikeman. only place officer and musketeer on stand ground (both in atack underway mode, we do this with the musket in case hussar come between them the 18th century musketeers got a banjonet, but their firepower better then their banjonet so shooting is a better opsion). note: place all your units close to eachother a big front is more difficult to hold then a small one. use you peasents wich are doing nothing as shoothing flesh place them near your musketeer. Line up your hussar behind you army at the 2 sides, (in case his pikes fall then run then in his musketeers) when fighting walk you pikeman with the spear up! then they walk faster and when they in the enemy let their pike down so they can attack for some reason this works better then when you do it from the start. when you run out of pikes and peasents use your officer as shootflesh, when they are gone use your dragoons note: (your musketeer are the most importand!) do everything to keep as many alive as posible.  Afterbattle:  When the war is over just walk in his base and first destroy his blacksmith this will prevent pikes from his barracks and hussar, then get his academy then he wont get anny more officer. then its just a case of destoying all the buildings. note: when you run out of peasents you will be defeated so its nice to destoy all his mines and his townhals.  GL in your game and remember you may loose in the begin but you learn by looses and my manual to win it\'s just a case of play allot and learn from your mistakes. '),(24095,4047559,'27,July,2004','2004-07-27 00:00:00','   A normal game with random nations and 10 minutes of peace time...what was it going to bring, what was it going to be?   The war was between Russia (aerodynamik ) and Sweden (Enterprise) versus Bavaria(Killer 999) and Poland (Tony). An off we went. I started building but quick I was a bit tired (my fingers hurted ) but on the other side things went quite well. I had builded my academy and finaly a diplomatic center. I managed to set up a little army with archers from my diplo and pikeman from the baracks and went to the enemy\'s base. But things went not quite well when i\'v got there i soon got the feeling I was going to loose much of my men. I tryded to get to his diplomatic centre and try to destroy it but I failed. Aerodynamik managed to get his army smaller but that wasn\'t good enough for me. I had to regroup and luckely(with again a little help of aerodynamik , so you can see what teampower can do ) I finaly managed ,after quite what losses, to get trough his defences and destroy his diplomatic centre with my archer. Now with his diplomatic center destroyed it was as good as over for him. On the same moment Russia (aerodynamik ) attacked Bavaria (Killer 999) with a squad of well trained pikeman. Bavaria on the other side also came with a squad of pikeman but the expierience and strength of aerodynamik was no match for him, so he was as good as chanceless against the destroying force of Russia And so in no matter of time aerodynamik got trough the \'defences\' (if there where any:p) of Bavaria and managed to destroy most of his important buildings. With our good tactics and a bit of luck we\'ve got VICTORY. So never underestimate the power of teamwork '),(24096,4047559,'27,July,2004','2004-07-27 00:00:00','   Nation: wich nations to take in Tousands, The best nations are: Poland, Russia, Algeria, Turky, Ukraine  -Poland couse their pikeman from 17th century barracks are made faster then avarage pikeman  -Russia couse they can easy make multiple 17th century barracks (only wood and stone)  -Algeria couse they have the best archers in the game wich are made very fast ( attack 199 each arrow and on buildings 150)  -Turky couse their tatar (made from stable) and their janissary (musketeer)  -Ukraine couse their serdiuk who have high fire rate and their hetman (made from stable, hp 450 attack 350)  How to build:  Let 6 peasents create a mill(do the mill upgarde to) and the other 12 make a townhall (make unlimited peasents), when the buildings completed make a storehouse. Then make a market as fast as you can couse it very importand! Market trade: 5000 coal for food, 3000 iron for wood 1000 iron for stone and 500 gold for stone. Make another townhall(make also unlimited peasents).Then make a blacksmith folowed by 2 barracks where you make pikeman when you are russia or poland, algeria makes archer, ukraine makes serdiuk, turky creates janissary. Then make a academy when finished do the \'\'new woodworking tools (wood extraction +100%)\'\' upgrade this helpes to get the wood faster to make a diplomatic centre, then go colect wood if you did everyhting good you should already have enough stone(except for russia you need to mine more stone then) so everyone on wood. Let 10 men work on the diplomatic centre meanwhile make at least 1 iron mine and 3 gold mines (when you are turky or ukraine make more iron and coal mines)(make in the diplomatic centre unlimited archer for poland and russia, infantryman for algeria ,ukraine and turky). When time and resoursers you can make extra barracks, or stables.   Army lining up:  Poland and russia:  Set your pike man in a formation (can be 15-36-72-120-196) best one do do is is pieces of 72 use with your pikes atack everybody underway and stand ground (a extra atack bonus) don\'t move your men anymore. then use your archer to kill the aponents pikeman use your archer as I say, how stupit it may look it\'s the best way. use ctrl+a (all units) select your archers and put then on atack disable then when u see a big spot of allot of enemy pikes walk near, and let your archers fire in a huge salvo ( just use your mouse to select where to shoot) almost all his men will die. Keep repeating this move with your achers, when almost all his pikes are gone sett you pikes losoe and kill his archer (you can do this very easy by leting 30 sich cossacks(made in diplomatic centre) around in his archer you can do this as a raid but not raid his base but raid his army ).  ukraine and turky  Line up your shooter (musketeer) with atack everybody underway Don\'t use stand ground! select your infantryman and put then atack underway and stand ground just before your shooters so that your shooters are safe keep a eye on enemy siches who could get your shooters from behind. Here you can also use sich cossacks to get his archers.  algeria  Line up your archers with atack everybody underway Don\'t stand groudn them! line up your infantryman on atack everybody underway and stand ground (just to keep enemy bussy) then let your archers shoot at the huge flocks of pikeman when enemy almost dead releasy your infantry loose. Here you can use siches to to get enemy archers.  Afterbattle:   When you are in his base, get his blacksmith no more pikes mean he is lost, then get his market before he can tarde stuff to make a new blacksmith. Then get his diplomatic centre. Note if you have no peasents you are defeated, so get his townhalls and his mines.  GL in your game and remember you may loose in the begin but you learn by looses and my manual to win it\'s just a case of play allot and learn from your mistakes. '),(24097,4047559,'27,July,2004','2004-07-27 00:00:00','   nation: wich nations to take in normal actualy every nation is good in normal if you know their tricks but I think these nations are the best. Poland, Rusia, Turky, Algeria, Ukraine  -Poland couse their pikeman from 17th century barracks are made faster then avarage pikeman  -Russia couse they can easy make multiple 17th century barracks (only wood and stone) so more pikes  -Algeria couse they have the best archers in the game wich are made very fast ( attack 199 each arrow and on buildings 150)  -Turky couse their tatar (made from stable) and their janissary (musketeer)  -Ukraine couse their serdiuk who have high fire rate and their hetman (made from stable, hp 450 attack 350)  How to build:  Start by let 6 peasents make a mill and the other 12 on a townhall (make unlimited peasents). then place 5 peasents to cut stone. The other make a storehouse, when this building is finished palce them to go cut wood. Then make a market the trade: 1000 coal foor food, 200 gold for wood, 200 gold for stone. Then create a blacksmith folowed by a 17th century barracks (make unlimited pikeman) note: don\'t Upgrade your pike man waste of gold! Then when te barracks is done make 1 iron mine and 4 gold mines the unlimited peands go to the mill to gamble food. Trade iron and gold in the market for wood and stone to buy a academy, build the academy with people from your mill, dont wory about food just buy some more. When academy finished place them back to your mill. then place unlimited peasands on wood. Just trade gold and iron again to gain stone and wood to buy a diplomatic centre. When it is fisnihed make unlimited archers. Make your pikes into a squard. (then they are stronger ). make more mines, unlimited peasents to stone. Start on the food upgardes (possible a 2th townhall) and wood upgarde then go save for the big stone upgrade+ 300% stone, stables and other good stuff in most time it will come to diplo and not much further. when you have diplo you can upgrade pikes but diplo is importand!  Army lining up:  Place your pikeman on attack everybody underway and stand ground and your archers behind them. It\'s also sweet to preform little raidings from side in his base with 5 archers, this is verry effective or just grab his mines. then he will be able to make less men, couse no resoursers. Use again the archer tactic with salvo\'s this is the most effective way of using your archers. when he got archers just make 10 siches and walk them into his archers. note becareful for little raiding\'s they are very effective! Note when both teams are weak so it going even, use your peands they are very strong they hit 20 and some nation peands got 100 health good fighters.  Afterbattle:   Try to get his blcksmith meaning no more pikeman for him so he will have little chance of surviving. then his market so that he can\'t trade things to make new blacksmith. then gte his diplo. Note when you have no peasands you are defeated, so it\'s nice to destoy his townhals and his mines.  GL in your game and remember you may loose in the begin but you learn by looses and my manual to win it\'s just a case of play allot and learn from your mistakes. '),(24098,4047559,'27,July,2004','2004-07-27 00:00:00','   It seems that there is not very much activity on the forums and on the site! Visit (if you can) once a day the site and the forums and read new treads and topic and reply! Without your support the clan isn\'t what is should be, a nice and respectable clan! '),(24099,4047559,'27,July,2004','2004-07-27 00:00:00','   Tournament  There are plans to hold a tournament for Cossacks. If you have suggestions about how and when it could be done, send it to one of the councilmembers! More information will follow soon! '),(24100,4047559,'27,July,2004','2004-07-27 00:00:00','   Site-update  For a while the site has a new layout. If you don\'t like the theme, you can change it on the left. (see theme select). There are some difficulties with to forums but we hope to solve this as soon as possible. (now on the left is the forum for Cossacks, in the upper you\'ll find RTCW) '),(24101,4047559,'27,July,2004','2004-07-27 00:00:00','   If you have nice ideas and suggestions for the site, mail them to me '),(24102,4047559,'27,July,2004','2004-07-27 00:00:00','   Members  Currently we have 13 Cossacksmembers and 4 RTCW members. '),(24103,4047559,'27,July,2004','2004-07-27 00:00:00','   There are now 3 council members: Enterprise  Aerodynamik  Orbit '),(24104,4047559,'27,July,2004','2004-07-27 00:00:00','   A giant lurks on the coast of North Africa. Build into the very living rock of the libyan coastline, the Siegfried battery dominates the southern Mediterranean. Siegfriends colossal 15 inch gun is terrorizing Allied shipping and protects Sicily from Allied attempts at invasion. Here in North Africa, The Allies and the Axis are put face to face, it\'s a fight to the death.  Number Info: 1:> Axis North-West MG 2:> Gun(cannon) \'\'Controls\'\' 3:> Allied East MG 4:> Allied West MG 5:> Axis West Bunker MG 6:> Axis South-East MG 7:> Allied Assault Ramp 8:> Back Entrance of the Axis Main Bunker 9:> Command Post    Mission objective destroy the Gun(cannon), this can be done by constucting the assault ramp, or by the axis main bunker entrance. Blast your way in, when you are lost inside the Axis Bunker, follow the \'\'blue tags with a cannon\'\' on it. When you reach the gun just drop some Dynamite (engeneer weapon and arm it with the tang push \'\'5\'\' above typing keys ) next to the gun, if able place more in case it gets defused. When dynamite explodes you win. Note: you can capture the axis west bunker, for the back entrance some axis need to come out so door opens or you need enemy clothes (covert ops) when they say destroy the generator folow the \'\'red lighting tags\'\' in the bunker place dynamite and back entrance \'\'extra doors\'\' open There are 2 doors 1 realy big 1 narrow. enjoy '),(24105,4047559,'27,July,2004','2004-07-27 00:00:00','   The once-peacefull desert town of Tobruk is torn asunder as Axis and Allied forces struggle for tactical and strategic dominace. Tobruk\'s narrow backstreets and byways are the setteling for a desperate face-to-face clash of elite forces. It\'s an all-or-nothing fight to the death. '),(24106,4047559,'27,July,2004','2004-07-27 00:00:00','   There are four game modes in Enemy Territory, each with their own particular style of play: Objective Mode, Campaign Mode, Stopwatch Mode, and Last Man Standing.   Objective Mode In the Objective game mode teams must accomplish a series of tasks or objectives to win. These objectives can include everything from taking control of specific points on the map to destroying key installations. A team must complete all the primary objectives in the time limit to win.  Campaign Mode Campaign mode is a form of Objective mode that allows players to compete against each other in a series of related maps. Players can increase their Rank and Skills over the course of the campaign and carry them from map to map. Skills and Rewards persist from map to map for the duration of the Campaign.  Stopwatch Mode In Stopwatch mode, teams must accomplish the objectives in the shortest possible time. At the end of a round the teams will switch. The team that was defending will now attack and must now beat their opponents\' time from the previous round. Team-switching is not allowed in Stopwatch mode.  Last Man Standing (LMS) Last Man Standing is a form of Team Deathmatch where players must completely eliminate the opposing team to win a round. Players have only one life. They can be healed with Health Packs from Medics or Health Cabinets, but will not respawn if killed. In LMS there are no skill points, XP does not count towards your score and there are no objectives other than to wipe out the opposing team. There are no Last Man Standing Campaigns.  Respawning, Reinforcements, and Max Lives In all game modes except Last Man Standing, players will respawn after they die, meaning that they\'ll come back to life at the team\'s spawn point with full health and ammo. Players will respawn in reinforcements, or timed waves, that run continuously. The amount of time left until the next reinforcement respawn is shown in the upper left of the screen when a player is killed. The player will only respawn at the next reinforcement period if they\'ve hit the Jump key (Space) before the clock runs out. If the player hasn\'t hit the jump key, they will stay in limbo and will not respawn. Players can use the reinforcement time to configure their class and weapon in the Limbo Menu.   All game modes apart from Last Man Standing can use the Max Lives option. This option makes the player pay a price for using up all of his lives. The penalty is set to a default three respawn cycles until he can rejoin the fray. '),(24107,4047559,'27,July,2004','2004-07-27 00:00:00','   Enemy Territory\'s maps can be played in a variety of ways, but mainly involve constructing, destroying or taking Objectives. The objectives can be studied in the Command Map by hitting (G) or the Limbo Menu by hitting (L).  To switch to a weapon or tool, press its corresponding number key or cycle to it using the next/previous weapon keys. To use it, press the fire button. Short taps of the Fire button will fire single shots. Hold down the fire button to fire in a long burst or to use a tool for a longer duration.   Players will suffer damage from weapons (their own team\'s, as well as the enemy\'s), from falling from height or drowning. Players can be healed by picking up Health Packs, either from a Medic or from a Health Cabinet.  If a player runs out of health, they die. They can now either be revived by one of their team\'s Medics, or go back into the Limbo menu and choose another team, class or weapons load-out if they wish. Players in Limbo will respawn in the next reinforcement wave at the Spawn Point they\'ve specified in the Command Map.  Players can always check on the status of objectives and their teammates\' locations by checking their Command Map (G). The Command Map will also display any battlefield intelligence gathered by their team\'s Covert Ops.  Players will be alerted to developments on the battlefield by their Command HQ - automated audio messages telling you the state of play and what your next objective is.  Players can also talk with each other, either by typing text messages (using the T Universal Say or Y Team Say keys) or actuating the pre-recorded voice messages in the Voice Chat menu (V).   Players can also form FireTeams to coordinate their tactics. Players can talk to member of their own FireTeam by using the U FireTeam Say key.  Players will be rewarded for good teamplay with Experience Points, extra skills, promotions and medals at the end of the game.  Don\'t step on the little flags in the ground - they\'re Land Mines and they\'ll kill you!  If you find you can\'t perform an action or fire a heavy weapon, you probably don\'t have enough Power Bar.  Building Command Posts will help your team to fight effectively. While not the primary goal of your team, building and protecting your Command Post while destroying the opposing team\'s Command Post will give your team a definite advantage. '),(24108,4047559,'27,July,2004','2004-07-27 00:00:00','   *1> *2> *3> *4>  you will get promoted when you use heavy weapons (mortar, panzerfaust, tompson, mp40, (when you are a soldier), flametrouwer and mobile MG42 note: also all the other MG\'s count as heavy weapons the ones who engeneers need to fix with unlimited ammo. '),(24109,4047559,'27,July,2004','2004-07-27 00:00:00','   Axis and allies clash in the previously peaceful surroundings of a winter woodland fuel depot. The snowy woody hills and valleys will be the scene of a mortal battle. Face to face, it\'s a fight to the death, only victor will survive. '),(24110,4047559,'27,July,2004','2004-07-27 00:00:00','   *1> Bionoculars (item) *2> Physical fitness *3> Improved health *4> Trap awareness  you will get promoted when you are in the scene of action allot ( kill somone and you hurt yourself then heal yourself etc.) '),(24111,4047559,'27,July,2004','2004-07-27 00:00:00','   *1> Improved use of medic Amunition *2> Improved Resoursers (faster load of usable medic items\'\'med kitt, adrenaline\'\' *3> Full Revive *4> Self Adrenaline (unlimited stamina for short while)  you will get promoted as you do medic skills \'\'health packages, revive people\'\' '),(24112,4047559,'27,July,2004','2004-07-27 00:00:00','   *1> Improved Resoursers (faster load for airstrike and ammo boxes) *2> Improved Signaling *3> Improved Air, Ground support *4> Enemy Reconision (covert ops in your team\'s clothes)  you will get promoted when you suply people with ammo or when you use the \'\'air suport bomb\'\' '),(24113,4047559,'27,July,2004','2004-07-27 00:00:00','   *1> Improved use of light weapon ammo *2> Faster Reload *3> Improved light weapon handling *4> Akimbo lugar or Akimbo colt (dual handgun)   You will get promoted as you use \'\'hand guns\'\' (colt, lugar) tompson (note: soldier tompson is heavy weapons) and mp40 '),(24114,4047559,'27,July,2004','2004-07-27 00:00:00','   *1> Improved use of all scoped weapon ammo *2> --------------- *3> --------------- *4> ---------------  You will get promoted as you use covert ops skill\'s (sniper, sten, FG49 , taking enemy clothes '),(24115,4047559,'27,July,2004','2004-07-27 00:00:00','   Today the Extreme European Warriors are almost 2 months old! More information will follow '),(24116,4047559,'27,July,2004','2004-07-27 00:00:00','   Today the site has got 1200 visitors :) '),(24117,4047559,'27,July,2004','2004-07-27 00:00:00','   Now you can find articles related to RTCW:ET in the \'articles\' section. '),(24118,4047559,'27,July,2004','2004-07-27 00:00:00','   Now in the articles sections you can find all promotions! (soldier,medic,engineer,...) '),(24119,4047559,'27,July,2004','2004-07-27 00:00:00','   Please be more actif on the site and the forums! You\'ll make the clan much more cozier and nicer! RESPECT! '),(24120,4047559,'27,July,2004','2004-07-27 00:00:00','   YESS! AFTER A LOT OF WORK I FINALLY MANAGED TO GET THE FORUM UP  AND RUNNING AGAIN! Thanks to Aerodynamik :):) '),(24121,4047559,'27,July,2004','2004-07-27 00:00:00','   Great, the forum went down :( '),(24122,4047559,'27,July,2004','2004-07-27 00:00:00','   I\'m now setting up a database for tips and tricks for RTCW:ET '),(24123,4047559,'27,July,2004','2004-07-27 00:00:00','   Great, the site is now fully up-and-running!  urlLink www.extremeclan.net  ! :) comments are welcome! '),(24124,4047559,'27,July,2004','2004-07-27 00:00:00','   Things are not going as well as planned with the website but I hope it all will be fixed tomorrow :) '),(24125,4047559,'27,July,2004','2004-07-27 00:00:00','   This is my first message, I\'m trying it out now! :) '),(24126,4047559,'09,August,2004','2004-08-09 00:00:00','   Since today the site has an new lay-out! '),(24127,4047559,'07,August,2004','2004-08-07 00:00:00','   I\'m finally back from vacation :D great isn\'t it :p '),(24128,4016792,'27,July,2004','2004-07-27 00:00:00','       My friend Cathie came back from Aruba looking tan and gorgeous. Decided to go to the movies with Gina and Cara to see Fahrenheit 9/11. My second time. Oh, and it was her treat. Sweet.  The thing that blows my mind the most is the information about the Bush and bin Laden families\' financial entanglements, culminating with their involvement with the Carlyle Group. I had heard of Carlyle, but knew very little about it. I googled it up when I got home and came up with   urlLink  this video   , originally broadcast on VPRO Netherlands TV. This stuff is enough to keep you up at night.  Anyway we agreed, I think, that you can only use so much money. Why do so many people put so much energy and extraodinary effort into making more than they could ever spend? I cannot imagine becoming so dehumanized that I could feel comfortable with the knowledge that I was exploiting or manipulating anyone, never mind everyone.  Of course, I have my own plans for world domination. The manifesto isn\'t quite complete, but it has something to do with casting spells with chocolate. I think.  Michael Moore documents the research for his claims   urlLink  here   . He definitely has his point of view, and he\'s passionate about it, but his research seems impeccable.   '),(24129,4016792,'26,July,2004','2004-07-26 00:00:00','       My sister is having a baby. Wonderful, unexpected news.  6 weeks and 1 day pregnant. only 237 days to go    urlLink   '),(24130,4016792,'25,July,2004','2004-07-25 00:00:00','       So I just got my car back yesterday, and I\'m so happy because I have wheels again, and it\'s no longer making these indescribable loud noises, and I\'m a free woman once again........but now there seems to be a fuel problem of one kind or another, one kind being minor, and the other kind requiring a major overhaul. Either way, it\'s blowing so much smoke that the people behind me can\'t see where they\'re going, and I am doing irreparable harm to the quality of the air.   Should make tomorrow morning\'s commute excruciatingly embarrassing .   '),(24131,4016792,'16,August,2004','2004-08-16 00:00:00','       I spend the afternoon watching the Mets give away the game to Arizona. I love those guys, but they can frustrate the hell out of me. Then Susan and I decide to go to Ralphâ€™s fishing station on the dock and have a beer. Or two.'),(24132,4016792,'14,August,2004','2004-08-14 00:00:00','       Went to the movies with my daughter, Lisa. Sheâ€™s a huge fan of M. Night Shyamalan, so weâ€™ve been looking forward to seeing The Village. Stopped at Starbucks for coffee on this rare Saturday with her at home. Sheâ€™s working toward her masterâ€™s while working 2 full time jobs, so usually she has classes on the weekends. She has a few weeks now between the end of summer session and the beginning of the fall, so weâ€™ve been able to spend some time together.'),(24133,4016792,'13,August,2004','2004-08-13 00:00:00','       This happened a couple of weeks ago.  I go to Burger King for dinner. Hey, I\'m not proud of it -- everyone has their dark side. I don\'t eat red meat, so I order the crappy chicken sandwich. I ask for a tender crisp sandwich with no mayonnaise and a diet coke. The female voice behind the microphone gets this real strange tone as she tells me to drive around to window #1.  I drive around the building toward my food, and I see the girl who took my order hanging out - all the way out - of the window. She\'s outside from about mid-thigh up. She\'s waving her arms at me very excitedly- hurry up. hurry up. This particular Burger King has situated this window pretty high up. I guess to accommodate SUVs and trucks. She\'s hanging out like the munchkin at the gate to the Emerald City.  She waves me over and points at the car in front of me, driven by a blond kid about 20, \"see him? see him? he just ordered the tender crisp sandwich with   extra   mayonnaise, then you come along next and order the tender crisp sandwich with   no   mayonnaise.\" This chick is  astonished  by this nearly  inconceivable  sequence of events.  It seems rude to not to acknowledge the high point of her day. \"Well,\" I say, \"that\'s why we\'re divorced now.\" It kills. Where was this woman when I was thinking of doing stand-up? She loses it. She\'s laughing so hard that she feels compelled to go get everyone else on the shift and bring them in on the joke. Pulls people off the grill. Within a minute, 5 or 6 Indian women are crowding the drive-thru laughing about condiments and broken marraiges. And the poor guy, my ex, in the car ahead has to wait for his food while wondering why he\'s being pointed at.  Eventually they send me on my way. \"No mayonnaise, no mayonnaise,\" in chorus. I swear.  Never enjoyed one of those lousy sandwiches more in my life.'),(24134,4016792,'12,August,2004','2004-08-12 00:00:00','    urlLink Link     Is it possible for a \'98 Hyundai to spontaneously heal itself? I\'m here to testify, brothers and sisters, that mine has. The check engine light is gone, it has all the power and performance of any 6 year old, 4 cylinder, Korean muscle machine, and I can even use the air conditioning.  Of course, about 2 hours after this divine intervention, I found myself on my knees praying that the goddess of kitchen appliances would raise my dishwasher from the dead. The torment! I\'ll take the bus, but I won\'t do dishes. I won\'t.  I finally found the exact spot to kick (harder to find than the G- spot) to turn the damn thing on, but it wouldn\'t fill with water.  Despair. Water is the whole point.  So I used my big soup pot to fill the thing, soaking the floor in the process, kicked the D- spot, and ta da! clean dishes.  I don\'t want to have to do that again any time soon, so I\'ll just have to stop shopping and cooking, and eat out all the time. Problem solved. Coming soon to a blog near you -  Adventures in Dining Out .'),(24135,4016792,'07,August,2004','2004-08-07 00:00:00','       Drove out to Montauk (risky - since the check engine light is still flashing at me and the inspection is 2 months overdue. I continue to live my life on the edge.) and slept on the beach. Found this really cool little crab shack and had some dinner with jazz in the background.  Now must take the d-o-g for a w-a-l-k. If she even hears me typing it she will go botz. Then I\'m thinking bubble bath, red wine and John Coltrane.   '),(24136,4016792,'06,August,2004','2004-08-06 00:00:00','       OK, so I should stop playing with the html and actually post something, right? So much self-discipline required, I\'m not sure I\'m up to whole blog thing. Friday night, thank god. A weekend stretches before me filled with fabulous possibilities....must climb out of the work clothes and go find something fun to do. Scheduled all my days off for the rest of the year. Lots of long weekends and a week in December. A little traumatic for me, and for Chris, who made me promise to available by cell phone, but I feel so freeee now, and can\'t wait the two weeks until my first couple of days of down time. In the middle of the week!   '),(24137,4016792,'04,August,2004','2004-08-04 00:00:00','       Happy anniversary, Kitty and Mike.    urlLink        You are probably finding the baby day counter obnoxious, but eventually you may find it adorable. Or tolerable. Or not. My blog. Get used to it.  Enjoy the sun in Santa Del Whatever. I am planning a few long weekends myself, as difficult as it is for me to take off from work. You never know what might happen in my absence. They may find that they can get along without me. So I have 15 days still to take before the end of the year or I lose them. Haven\'t taken a day off yet this year. I am either the best employee EVER, or out of my mind. Very certain that both are true. So I am forcing myself to schedule 10 days off during August, September, October, and then the week before Christmas. I heard American Airlines has $199 fares between New York and LA if booked this week. Must look into.  Anyway, I\'m thinking a spa day is in order. I desperately need a massage. A loooooong massage. I feel like the stress of the last 45 years is knotted into my neck.  By the way, in case anyone is wondering, Kathleen is the best friend in the world. Talking sense when I needed it. Thanks.  Oh, and I never did find anyone to go to the movies with Sunday night, because (rant) everyone I know is lazy and boring. It\'s the movies, for chrissakes, not like a huge commitment, you sit in the dark for 2 hours, how can you not \"be up\" for that??  So I went myself and saw I Robot. Perfectly acceptable sci-fi stuff, but I would have liked to have seen the robot\'s character more fully developed. I seem to remember more depth in the novel. Must reread. Come with me next time, or miss once again my insightful and witty commentary.   '),(24138,4016792,'01,August,2004','2004-08-01 00:00:00','       What a week. Survived the preparation of our new branch location for a visit by the Vice President of North American operations, car trouble part 3, and I don\'t know what else. Now I have to do my laundry so I can do it all over again starting tomorrow morning, and find someone to go to the movies with because goddammit I am going to have a little fun this weekend. Time is running out.   '),(24139,3700421,'28,June,2004','2004-06-28 00:00:00','   i add this view of our pond to my well.   when i come to think of it, i realize that our house in amersfoort in fact used to have a well in the past! the well still exists, but now it belongs to our neighbours... quite remarkable how the well and garden and the little house changed owners over the years! let me try to sketch the layout briefly by way of introduction to some more photos i posted at www.flickr.com/people/alena.   there\'s two main houses, behind them two little backyards, one little house and a garden with a pond. the two houses, ours and the neighbours\', used to belong to one person, as it were sharing one large backyard, the little house and the garden that goes with it. since the two houses split, we have one bit of the backyard, the little house and the garden with the pond, while the neighbours are left with the second bit of the backyard and the well. anyway, those old houses and gardens are like a maze; they are fitted into each other and entangled like siamese twins.   i\'m glad i can share the pictures. i already got one comment on flickr from someone who wanted to visit :-) for the time being, anyone can in cyberspace... 		 		 		 			 urlLink     			 				 			 urlLink tuin vijver 1 			 			Originally uploaded by 			 urlLink Alena .  '),(24140,3700421,'22,June,2004','2004-06-22 00:00:00','   some events are better written down before they are forgotten. like meeting daniel. daniel is hungarian, about my age. we studied together at the university college utrecht and after his graduation daniel took off to stanford to follow a phd program. maybe he got bored of stanford, he arranged a transfer to harvard. still within the ivy league range, i suppose it gives you a kick to have this all on your cv. anyway, every now and then daniel and his polish wife ania fly over to europe and make a short stop in the netherlands. compared to the united states, distances are so small daniel and ania fit several countries in their visits of differing lengths. sometimes i should ask why they always include the netherlands: for the nostalgic thought of having lived here, for the purpose of seeing old friends, or maybe the reason is less poetic and they simply use the cheapest airfare. leaving all this aside, it\'s always a pleasure to see them. one of those friendships that have a high lasting coefficient.  one week ago i got an email from daniel. the subject line read: what are you doing on thursday? i thought, what am i doing on thursday? since i decided to quit my current job, free time has taken on an air of indulgent luxury. it\'s not necessarily that i work less, it\'s more that i\'m better able to enjoy every bit of it. i am also more efficient with using my time now that the burden of the job has been lifted off my shoulders. for some time i gave it some serious thought to continue the job while keeping the thought of quitting in the back of my mind, but finally i figured this was not a very healthy approach to life. so i\'m quitting and for the time being i have all the time in the world to do things like meeting daniel on a thursday afternoon.   on thursday i took the train to utrecht. the weather wasn\'t too good, neither warm nor cold, a prototypical day in dutch spring. disregarding the fact that the train takes only 15 minutes to get from amersfoort to utrecht, i decided to enjoy the journey and treated myself on a bottle of coca cola. i also brought a book, another murakami\'s classic the norwegian wood. i wore my favourite dark green top and thought the setting was perfect.  as i walked from utrecht central station to the cafe we were supposed to meet, one of the nicest cafes in utrecht carrying one of the funniest but also most stylish names \'winkel van sinkel,\' i got stopped by a guy with a cheerful looking face. he said: \'hi! my name is daniel! how are you?\' i didn\'t really know what to say. \'do we know each other?\' (hesitation on my side) \'yes, of course,\' he replied. \'i used to work here and always told you about jesus!\' thoughts were rushing through my head: jesus? here meaning on the street? is his name really daniel? but this guy never waited for me to say anything. he went on and on talking, and the more he said, the less it made sense. \'i got some money from my mother and bought this suit! a suit of a hundred euros, a real bargain, don\'t you think? now i work in a hotel, making up beds and so. you know, the one on the malielaan!\' i thought: what the heck the suit and the maliehotel! i know the maliehotel, it has this posh atmosphere around it, and louk always says it\'s one of the best hotels in town. so i always look a little suspicious: why should i automatically like everything that louk likes? but he has a point, the maliehotel really is pretty. and this guy was in any case very proud to make their beds.  i didn\'t have a lot of time to spend on this unexpected encounter. after all, i was meeting daniel, the real daniel, in about five minutes at this winkel van sinkel. so i tell this guy: \'the suit looks really good on you. good for you to work in the maliehotel. nice to have seen you again, but i really have to go now.\' he stopped talking and after a short pause went on: \'yes, i am very busy. i must go now, it was nice to see you. i am really so busy with the new job and all!\' i added one last question: \'are you sure your name is daniel?\' \'yes, my name is daniel!\' he maintained. \'okay, bye!\' when you don\'t know what to say, an okay always comes in handy.  i walked to the cafe and met my real friend daniel, his wife ania and a couple of other friends. i tried to sketch this story but nobody paid much attention to it, just a weird coincidence, stuff like this can happen. but it did leave some impression on me. maybe i read too much fiction these days.  '),(24141,3700421,'21,June,2004','2004-06-21 00:00:00','   i call this \'blog spot\' the well. the well is a regular feature of murakami\'s novels and i figured the line blog-spot-well read quite funky, like the song of some old 1980s rock music band. so what is a well? good question, in each of murakami\'s novels it has a different role, and i choose to use it following the wind-up bird template. a well is a place you go to to sort out your problems. what is more, entering the well implies you accept to enter a different reality, a reality of dreams perhaps, but still very real and promising of a solution. murakami\'s novels are like david lynch\'s movies, exploiting the same kind of mysterious playfulness. anyway, i like to think of this blog as such a well. the virtual reality of it forms a charming parallel.  the only trouble is, going to the well presupposes that you\'re in search of a solution to a specific problem. you know what you\'re looking for and it\'s really just a matter of time before you find it. i don\'t have anything close to a specific problem or question. or better, i just might have too many. and i hate keeping diaries.  so what am i looking for in this blog spot? at the moment i can think of at least one reason, to invest in friendships which are otherwise difficult to maintain. and it gives me the opportunity to write down my conversations with myself.  for more than two years i did an awful lot of commuting to my university, some three hours a day, at least three days a week. everytime i walked home from the train station i let my thoughts lose and talk to myself about different events of the day. i can go on and on complaining about a mismatched skirt and t-shirt to an imaginary stylist, continue conversations with my colleagues, and very often scheme imaginary fights with almost anybody to let go of silent aggressions. the fights and the fashion concerns i like the best.  starting with today. i walked from the train station and thought over and over about my first contribution to the blogger, a very well-intended theme for a change. i was very pleased to come up with the name (the well) and immersed myself in a monologue of wells, problems and virtual friendships. i guess none of what i\'d written is as well formulated as when i imagined it while walking. i suppose it usually goes like that with finding the right words. it certainly goes like that with my imaginary fights. angry alone is far more creative than angry face-to-face.  nevermind, so far so good. best regards to my sister and thanks for introducing me to the blog phenomenon! :-)  '),(24142,3558898,'19,June,2004','2004-06-19 00:00:00','   Because I live at the end of nowhere, I had to cut alot of wood to have when the winter will come. And I didn\'t finished all, and I\'m very very tired. My hands hurt, my legs, my head hurt, my everything hurts :(. I just need a long long vacation. A long vacation ....... When??? A long vacation ....  '),(24143,3558898,'14,June,2004','2004-06-14 00:00:00','   Read  urlLink Yep, the monday  '),(24144,3558898,'10,June,2004','2004-06-10 00:00:00','   People are crazy or they are stupid, or something, it\'s very strange.  When I stay home, and i watch the TV or I take a shower, I always dream \"wow, how good will be now if I would be on a mountain, with flours around me, not any car sound, not smoke, not any strange noise from the neighborhood, just the birds and the plants\". And when I\'m up there in the top of the world on the mountain, I think \"why? why I camed here? bugs everywhere, I don\'t have where to take a shower, I don\'t have anything to do, I just get bored, not a TV, not nothing, I should stay home next time when I wana go on a mointain\".  And ofcourse, when I\'m home the same thing it\'s happening, and when I\'m on a mountain the same thing. Why we are like that? We will never get the answer.  '),(24145,3558898,'08,June,2004','2004-06-08 00:00:00','   When I\'m at work I am thinking at alot of things doing after I get out from work, in my free time. So, when I have free time I should do those things. I\'m not. Why? Dunno why, I just want to get lazy and not to do anything. Curios, huh? Sometimes I ask my self why God made us like we are. Eating, going to toilet, drinking if we are tursty, sleeping, and so one. It\'s a strange world and some strange people too.  '),(24146,3558898,'08,June,2004','2004-06-08 00:00:00','   Yes, it was a hard day, because my boss was out of town (you could all say: yupeeeeeeee!!!). No, it\'s not like that, because when my boss it\'s out of town I have to do his work and I can\'t miss from work, and alot of thing.  All the day I was very depressed, something in the morning made me to think at what is the meaning of life. And I didn\'t found out, a part from my depression is because I don\'t have a girlfriend., but I\'m not in the mood to explain what I thought and what not. This subject I let it for another day.  Ah, by the way, today I had the first comment to my blog :), from somebody called Ron, Thanks Ron :), it\'s nice to know cuz somebody is visiting your blog once in a while.  Ok, I will close the post for today, because it\'s showtime and because I\'m just not in the mood, the OZ it\'s begining :). A movie with a jail :)  Good Night  '),(24147,3558898,'07,June,2004','2004-06-07 00:00:00','   The first day from this week. Yep, it\'s monday. Ah, I hate the first day, yep the monday. Why? Because is the first day to work, and ofcourse because it\'s monday. Monday I work more like the whole week. Yep, monday, the day when another week of work is coming. Do you have that feeling \"of monday\" in anothers day? I have that. It\'s not a nice feeling, a nice feeling is \"of friday\" or \"of saturday\". The feeling \"of sunday\" i don\'t like that either, because after sunday comes monday, and again, yep it\'s monday :(.  At work I have a great Boss, yep it\'s great because he consider me his boss, not him my boss, but he talks alot, and when I mean alot, I mean alot, and he talks about things what I don\'t care, because he is old, I think he has over 70, and i\'m just 26, what I could hear from him, everytime when I have a little break, I run from my work to stay 15 - 30 minutes with some friends, who are working at 2 minutes from me, two brothers, one works 3 days after that the other one works 3 day, and with them I can talk about what I want, and about what I\'m interested.  I wait everyday to get out from work, to come online to make what I like the most, surfing :). Surfing nowhere, just I open the google and I search and I search and so one. I stay 2 or 3 hr, after that I go home at my computer and I work at my webpage, at some designs. I look at the time, and wow, it is begining the tv shows, I watch like 3 hr and after that I go to sleep.  I will go a little too surf some more, and after that go home. Why? Because it\'s monday, yep it\'s monday.  '),(24148,3558898,'06,June,2004','2004-06-06 00:00:00','   I wait 5 minutes, to clear my mind, and after a good thinking, knowing a little html :), I looked at my blog and saw the \"I power blogger\" image, and I put there that crazy link with that crazy image, what gived me alot of head pain, and I removed the last link what apeared on EVERY POST what I made, so  now the blog looks ok ;)  I\'m Happy cuz I\'m smart :)  Smile, tomorow is another day :)  '),(24149,3558898,'06,June,2004','2004-06-06 00:00:00','   Searching for blogs on google to see more about theme, I found a site where you have shares from blogs or something like that, I registered in there and to be active I have to put a url gived by theme in my blog. Here is the url  ah, I don\'t know I put here the url but on the blog apeared the code not what it should be, and now I inserted the code in the template, but dang the code apears right but on every post what I make, ah, my mind hurts, so tomorow I need to change that, because it looks like dunno what. ON EVERY POST COMMOOOOOON!!!! Stupid me. Tomorow.  Funny, it\'s my first day when I see the word \"blog\", and I already have one, right now with this post will be two posts :).  Ps. Searching after blogs I found an IQ test, and I tryied to see how good I am, because in my country we don\'t have IQ tests or something like that. The result was 122, I don\'t know if is big or small. (the IQ test was in english and I\'m not to good in english, I did my best I think to the questions with images :)) ).  '),(24150,3558898,'06,June,2004','2004-06-06 00:00:00','   Right now is 20:10, I\'m at an internet cafe, and I was searching on google deifferent things, and I saw a lot of thing about \"blog, blog, blog, blo..., bl.., b...\" and I wondered what it is, and I arived at blogger.com and I found out what it is, and right now I made an acount to have my own blog. So that is how I start my blog :)  '),(24151,4185692,'21,August,2004','2004-08-21 00:00:00','   We went shopping this afternoon. At Eckards I bought some new whitening toothpaste (yesterday Mom bought some whitening gel that i\'m going to be using daily, too). We also bought some new facial cleansing pads. Then we shopped for about three hours at the Mall. I found a poncho. It\'s really pretty. Pink, of course. A sort of peachy pink. I also got a black t-shirt. I look good in black. I\'m going to wear black to the banquet this year. I tried on a black dress in Parisian. It was short and nearly sleeveless and extremely tight. i\'d never wear that in RL, but it made me look like some famous actress going to the Oscars. Gosh that thing was short. I\'d have to work out two hours a day for weeks to wear it anywhere. Mom probably wouldn\'t let me, though. Someday, i\'m going to buy a dress like that and go out for a night and party. I bet Sarah would go w/ me. That would be fun and different. Maybe i\'ll do that in Europe. Yep. Sounds like a plan. Anyway, I\'ve got a few more chores to do before dinner.   '),(24152,4185692,'20,August,2004','2004-08-20 00:00:00','   WHY DO PEOPLE WORSHIP ME?! I\'m not something special! All I do is be nice to people, treat them like they mean something, and next thing you know, they\'re following me around, or, worse, they\'re parents fall in love with me and start pushing their kids to hang out with me more, or encourage their kid\'s worshipping attitude. It drives me nuts! Yankees are crazy. Or... these in-between Yanks and Southerners that live in Nashville. People in Mobile didn\'t do this. I\'d be nice and get cuss words as a response. Not that i liked it better, but sometimes i wish i had all-out enemies instead of \"ministries\" (ppl i have to be nice to despite the things they do that drive me nuts). I don\'t want my friends to worship me. Bailey doesn\'t worship me! Marc doesn\'t think i\'m an angel or anything. Sure, they love me and are there for me, but unless I need them to be something for me to lean on, they don\'t breath down my neck with concern. I may be unstable and all, but i\'ll let you know when i need help with something. Directly or indirectly, i\'ll let you know... Otherwise, i wish ppl would just lay off and if you must watch me, do it from a distance.......   '),(24153,4185692,'20,August,2004','2004-08-20 00:00:00','   I had every right to be scared. This file system is a nightmare.... *shudder* It\'s HUGE and it\'s not like i can just put everything under his or her name, either. He has to be classified under \"investments\" or \"insurance\" or whatever. I\'m totally lost and i\'ve had it explained a few times, and still don\'t get it. This is horrible. I think i\'ll go  home and cry. Ugh..... Well, i wanted to post a really great latin translator site here so i can easily find it later. I need to e-mail bailey and ask her a few questions, too... stupid latin... i cannot wait for lunch break. http://archives.nd.edu/latgramm.htm   '),(24154,4185692,'19,August,2004','2004-08-19 00:00:00','   Why can\'t i actually succeed in being mean to someone? I mean, i know this is probably a weird thing to wish for, but even when i AM about as mean as i can be to someone without throwing out all social manners, for a long while, they STILL think i\'m nice!!!! Like, the other day mom talked to Mrs. Weaver. Her daughter (Caitlin) went to New Song w/ me for a while. At first, i hated Caitlin. So, i was about as rude to her as i could manage. And, after she leant me a few very cool books, i sorta warmed up to her and she wasn\'t so bad anymore... So, we\'re friends but she\'s going to a new school, and she told her Mom that she\'d met a girl who was almost as sweet as Kathleen at her school. I was like \"WHAT?! I figured i\'d at least have some taint on my reputation.\" Bleh. I\'ve done the same to Courtney and she\'s STILL talking to me over IM. She\'s so... SHALLOW! Mom even noted it today, she said \"Courtney seems really immature. Especially in the way she handles things...\" I was like \"Tell me about it.....\" *rolls eyes* I guess she\'s okay to talk to every century or so, but other than that she\'s yet another person hanging onto me for reasons i can\'t fathom... Oh well. There are some pretty delicious smells coming from the kitchen right about now. Dinner is going to be good tonight. I slipped a note saying \"have a nice day\" in Andrew\'s backpack before we left school... i wonder if he\'s found it, yet. :/   '),(24155,4185692,'17,August,2004','2004-08-17 00:00:00','   Yep. That\'s all i\'ve been doing for the past two days. That and listening to music on my new CD player. I LOVE IT! It\'s awesome! I think i\'ll carry it around forever. :-D I hate algebra 2. It\'s so confusing. It rarely, if EVER makes sense to me. I just don\'t get it. And half the time i\'ll lie and say I do get it just so we can put the book down sooner and be off to something else. stupid math. I love hank the cowdog. I\'m going to join his security force and get a Hank poster for my room. :-D Mom is reading Old Yeller aloud. Of all the stories my brother has to read this year, they pick Old Yeller. Looks like i\'ll be having nightmares tonight.... Just what i needed...... more nightmares.   '),(24156,4185692,'15,August,2004','2004-08-15 00:00:00','   Went to see princess diaries 2 today! what a funny movie! And Mimi and Pop gave me this really really cool pink pillow that\'s comfy and squishy and makes you look like a mushroom if you put in on your head, and i bought a portable cd player this afternoon and it\'s gorgeous and sounds awesome and life is wonderful. nothing like gifts shopping and laughter to make a perfect day. *happy sigh*  \"I look like a moose.\" \"But a very very cute moose! You make all the boy mooses go \'WHUUUUUUUUH\" \"What?!\" \"WHUUUUUUUUUUUUH\"   '),(24157,4185692,'14,August,2004','2004-08-14 00:00:00','   I really didn\'t do anything today. Except some vacuming and making this awesome dessert that tastes like it\'s straight from heaven. (Well, almost.) I also finished another book today. The heroine and her friends triumphed(sp?), of course, though she lost friends and blood in the process. Sometimes i wished i lived in that world, instead of ours. At least those people have purpose. They have enemies they can see and feel and they know how to fight. They have adventures, and amazing friends. Not that I don\'t have amazing friends. I must admit, my friends are most definately cooler. I guess my life isn\'t so bad. Except i don\'t have adventures. And i don\'t have enemies that are easy to fight. They\'re all sneaky tormentors that are often on the winning side of the battle. Perhaps i\'ll go outside and swing and stare at the stars and wish i could fly into them. Maybe that\'ll clear my head some.   '),(24158,4185692,'13,August,2004','2004-08-13 00:00:00','   I just finished the third book in the series Bailey has leant me.      I JUST SAW THE SWEDISH OLYMPIANS! Lots of blonde hair in that group. ;-) hehe. I wish them well in the competitions.      I wonder what it\'d be like to actually be able to talk to animals. I always talk to Scout like he understands me. I love that cat. Dad says someday I\'ll be off at college and Scout will be at home w/ him and Mom and Melanie and Robert. Well, Melanie for two years then she\'d go off to college, too. I guess after college i\'ll find an apartment and move Scout in with me. I\'ll miss tha cat. Thank heavens that\'s three years away. I can\'t believe it... I feel so old.   '),(24159,4185692,'12,August,2004','2004-08-12 00:00:00','   So far today has been uneventful, really... Mimi and Pop are coming over for the day tomorrow before they set off again on yet another trip. Michael T. came over today to pick up the geometry solutions manual i used last year. Yesterday Andrew was nearly in a wreck. I\'m glad he\'s not hurt. God was with him, as always. I\'m off to read some more. I love these books.   '),(24160,4185692,'12,August,2004','2004-08-12 00:00:00','   I\'m back from school. The first day went well, i guess. I had all my assignments there and ready, and I didn\'t have any trouble on quizzes or whatever, though i was a bit lost through most of algebra 2... :/ That was to be expected, though. From the looks of it, Algbra 2 is only going to get harder, and having trouble w/ it this early is, to say the least, a BAD thing.      Bailey brought the books! YAY!!! I have something fun to READ! *beams* I\'m happy!      Lunch was interesting... I sat down, and Andrew put his backpack next to my seat, but while he was in line a dozen other teens i know - i won\'t list them all- came up and filled up the table, so he ended up sitting at the other end w/Margaret, Melanie, and Bailey. I wasn\'t all that cheerful today. I was really tired and feeling kinda sick most of the time. Nowdays i get about 10 hours of sleep, and by 1PM i\'m so worn out i need a nap, and i\'m not even doing anything strenuous... It\'s strange.      I\'m deturmined to loose weight and look good (jessica simpson\'s stomach) by next summer. I have decided it will be easier just to leave from here in Nashville instead of ditching a group in Europe. This means my trip will probably happen early summer, instead of late summer, which is good for two reasons: More time to work it all out and get the parents calmed down once i\'m back, and it won\'t interrupt school. I have to find an all-day activity that doesn\'t involve other ppl i know that is planned ahead of time, long enough for me to schedule flights for that day and my parents not find out about it until i\'m halfway across the atlantic. By the time that day arrives i\'ll be frantic with fear. Besides, i\'ll have all the money i\'d saved to spend on that trip BEFORE i\'d turned it all in to whoever i was planning on going to europe with. After all, i must keep up appearances!   '),(24161,4185692,'11,August,2004','2004-08-11 00:00:00','   Well, the dreaded day has arrived. I\'m going to try and take it one hour at a time. I\'ve got my CD player and a few of my fave CDs (and most of the money i\'ll use to buy a new one). I\'ve got all my books and homework... Summer is officially over. I guess Peter Pan doesn\'t exist, though I\'d want more than anything to live in Neverland. Andrew says even if i could go he wouldn\'t let me. I\'ve got a LOT of preparation to do for next summer. I\'ll need not only the money to get me TO Europe, but i\'ve got TONS of research to do so i can get from one airport to the next without getting lost. I actually think it\'d be easier if i just started here in Nashville and got to Atlanta, from there to London, and from there to my destination. That will probably end up costing more than 3000 in airfairs. To make things less complicated i\'m not going to carry more than one bag, and i\'ll carry my cash in a moneybelt - because there will be lots of cash. I\'ll need it for buying new clothes to last me once i get there and for little things like food and stuff. I think my room is already taken care of though i\'ll have to check on that. Actually starting in Europe (like in an airport with a group of ppl i\'m on a missions trip w/) might be harder than telling Mom and Dad i\'ll be gone ALL day then ditching them. If i time it right i could be over the Atlantic ocean before they find out i\'m gone. I wonder how long a flight from Nashville to Atlanta is.... I should make a to-do list about all this. much easier if i could check things off as i complete them.   '),(24162,4185692,'11,August,2004','2004-08-11 00:00:00','   I DON\'T WANNA GO TO SCHOOL TOMORROW! I DON\'T WANNA GO TO SCHOOL TOMORROW! I DON\'T WANNA GO TO SCHOOL TOMORROW! I DON\'T WANNA GO TO SCHOOL TOMORROW! My first check on my Algebra 2 work said 75%, and i\'ll probably be up til 11 working on it. I hate latin, i made a C in it last year, and that\'s was barely. I\'m sure that y\'all know i failed Algebra 1, and this World Lit. w/ Mrs. Whitver is gonna be like living on Tatooine without shelter. The sylabus says Honors Biology is a really hard class, too, even though i\'ve always enjoyed science. I DON\'T WANNA GO TO SCHOOL! I HATE school! Even lunchtime is horrid because all it is for me is a stage, and right now i\'m too tired to act. Not to mention MOM will be at school this year and that means countless hours discussing how i handled this situation and how that person said that to so-and-so and i should avoid whats-his-face and that other person has problems with whatever. I can\'t handle that! And if i ask her not to do that, her answer is \"i\'m your mother! That\'s my job.\" She\'s really been getting on my nerves today. Well, i\'m off to church.   '),(24163,4185692,'10,August,2004','2004-08-10 00:00:00','   Marc is right, i AM an extrovert, and i\'m still a baby when it comes to things i don\'t understand. I have yet to grasp the concept that it\'s okay if people do things differently. I tend to get frustrated and mad when there\'s something I don\'t get, like the Andrew/Margaret thing. That just screamed of wrongness to me. It went completely against what i\'d been taught about self-image, and taking care of the way you look in other people\'s eyes. I\'m not saying that we should strive to make everyone love us, I just know that you\'ll have more success in general if you have the right image. Maybe I care too much about what others think, but I do want to have ppl\'s approval. Some people won\'t ever give me that, I know, but i\'d like to be someone that people like my pastors and teachers and adults in general approve of. (That\'s almost a joke. Of all the people on the planet, i\'d never trust an adult with my secrets and yet i want their approval?! That wreaks of my Mom\'s doing, i think) This often results in meeting new teens who have similar values to mine, like Michal, Brianna, and Bailey. Then i run into someone like Alli, and Josiel, and my response is to pull away from them. I\'d NEVER share my secrets with those two. Not because i think they\'re untrustworthy, but they don\'t dress like me, don\'t act like me, don\'t talk like me, don\'t hang out with my croud. I have yet to start thinking, or, more like believing, that it\'s okay that they\'re different... I mean, who would want to wear such un-fiminine(sp?) clothes? Of all my faults, this gets in my way the most, but I could probably step back and not care at all because I have friends that ARE like me, and don\'t really need ppl who aren\'t, yet. (at least not at school.)   '),(24164,4185692,'10,August,2004','2004-08-10 00:00:00','   Good Morning! I\'m starting a new blog because my other one messed up and i\'m not sure how to fix it... :/ I titled this \"Mind of a Princess?\" cause i\'m not sure i HAVE the mind of a princess, even though I want to be a princess in everything I do. I doubt i\'d be good dealing with press, though. I tend to tell more truth or lies depending on my mood. When i\'m depressed, for some reason i tell the truth more easily. Seems kinda weird. Sometimes i\'m confusing if i\'m really acting, or if i actually do feel that way. I confuse myself. Do you have any idea what it\'s like to be at war with oneself all the time? Hard. Trust me.      I\'m up really early because i\'m off to babysit. I\'m deturmined to finish all of my school work today, because we might be going to see Princess Diaries 2 (jumps with excitement) and Bailey would come with us so she can give me 3 new books to read! All of which i\'m just itching to get my hands on. I\'ll swallow them all tomorrow if i\'m done with my school. Well, almost all of them. Anyway, i\'ll be babysitting for another 9 hours today. They\'re good little boys, though, so i don\'t really mind. My biggest challenge is getting them to turn the Playstation OFF and watch TV or a movie. I\'d really like to put on a movie this morning because they played Playsation for 3 hours straight yesterday morning. That can\'t be healthy. I understand one or maybe an hour and a half, but not THREE. I simply must limit them more. *sigh* i sound like my mother.      Marc rocks. I luv him. :) Dunno how i got so lucky to have a friend like him. He\'s awesome!      I\'m surprisingly calm about Thursday. I still have enough time to get my school done, and the modivation to do it. If my headaches will cooperate and I can buy a new desk chair, i may be able to work through this year. I think i have ADD. Or, what they call Adult ADD. I\'m not sure what the difference is. I can remember to do things and all, i just can\'t sit down and DO them for long periods of time. Say, over an hour. I have to take a break, usually up to half an hour before I feel I can sit back down again. I need to research this.   '),(24165,3565927,'28,June,2004','2004-06-28 00:00:00','   I think saturdays would be ideal to catch the \'Simple Life\' where you see Paris Hilton and Nicole Richie rough it out in life. In fact, in a lot of ways these two girls kind of represent the majority of Singaporeans (not that we look like them or have their financial standing) but we are quite as lucky as these two girls. We have a shelter over our head, food, clothes and time to time able to indulge in other additional leisures in life. To say we are spoilt - it is probably too harsh... But we are in fact far more fortunate than a lot of people. Of course, we shouldn\'t compare ourselves with other countries but sometimes when we do comparisons, we are more easily grounded.  Many many of us (including me) haven\'t had a chance (or maybe we had) rough it out - just like the girls in Simple Life. I am guilty as charged :) I guess to rough it out doesn\'t mean we have to torture ourselves or do extreme stuff in life but I have seen people roughing it out from scratch. When I was in New York for my vacation in 1999, that\'s when I turn 21 there (Gosh, it feels like several years ago). I met an American guy who was 40 - nice breed of Dennis Quaid + Keifer Sutherland. Well, we hooked up and he took me around. Yeah, it was fun. Had deep conversations and from it I realised... He was just a grad from high school. He left florida and went to NYC (worked in Gap for 10 years) before he then became assistant manager in boutiques. Then, he heads 6 boutiques (when I knew him) - but he shifted to Chicago 2 years ago.   I really have a lot of respect for this guy. With his determination and hard work + experience, it was amazing. I mean yeah, qualification can take you places. But seriously, the paper chase in Singapore is way crazy at times. Even my parents jokingly said in 10 years time, you\'ll need a diploma to be a toilet cleaner. Why do parents equate toilet cleaner as \'Pathetic\'? To them,it\'s like a dead end. But these people who do this job - they don\'t think so!?   Guess when you advance more in life and become more affluent. Your perspective just either get better or twisted. I wish I knew what we were doing on Earth sometimes. What is beyond the galaxy? Are we in a box? I think Life is so beautiful but emotions can cut us like a chopping board. Why does it hurt even more than a physical pain? Sean came and took the house key from me today. We met at Bugis. It was complete silence. I could kind of see tears in his eyes but I had to keep myself composed. I just couldn\'t stay there any longer. My memories would come back again like a dormant volcano. It would all gush out and boil me into fragments. Everytime what he has given me in my life - it\'s all slowly being returned to him. Maybe heartache does involve your every part of your heart being given away. I hope this ends. It seems ending all this is coming back like a tidal wave, from time to time, washing me with tears.  Quoting from Avril\'s latest single \"My Happy Ending\"  ..You were all the things I thought of you And I thought we could be  You were everything, everything that I wanted We were meant to be, supposed to be, but we lost it And all the memories, so close to me, just fade away All this time you were pretending So much for my happy ending oh oh, oh oh, oh oh...   .............................   Why can\'t we all smile and laugh like a child in the church I saw today? We\'ve lost that in us at times...  '),(24166,3565927,'26,June,2004','2004-06-26 00:00:00','   As much I would like to keep up with you people for the past 3 days. I\'ve been sooo caught up just being lazy! :P Yes, I\'ve got this opportunity to veg so I\'m going to use as much of this time to do so until I officially join the work force. Anyway, as usual caught up with Clif on thurs (yeah you must be wondering why i\'m getting laid every week) - but just so happen this whole month he is scheduled to fly to Singapore so I\'m quite lucky. Anyway, the poor fella got some glass pieces stuck in his finger\'s bone thus he\'ll probably not fly (be on sick leave) to singapore for a month *sighs*  Well, guess it\'s the little things it\'s quite enjoyable when we catch up. Like how we\'ll sit by the television (where I insist on watching O.C. on thurs at 11pm no matter what) - have our red wine and talk. He was jokingly telling me that the O.C. (not shown in Sydney) is quite good and he should try flying to sg every thursday just for it?! lol In fact, tell you guys - the O.C. was already shown in US (when I was in New York in July 2003; last year), no wonder I found some episodes so familiar hehe... Yupe kind a DeJaVu for moi..  Funnily, maybe I\'ve not told you guys but I think Clif and I are suffering from a same syndrome \"He is a good f%ck and let\'s not fall in love\" mentality... Maybe I\'m wrong but there is a certain percentage I may be right/wrong I guess. I\'m not dumb and my intuition does tell me at times. But seriously, last thing I want is another guy in my life to give me more problems lol.. After we parted, met up with Eileen for coffee. Yeah, she was attending a course.   Then met up with my ex-ibm colleagues. Had Sakae sushi at funan. It was so nice and filling. The something kind of happen. I will not mention names in here (to protect people\'s privacy and decency). A gal frd of mine had to do a lot of explaining to her boyfriend why she was not back home (hello it sounds like a curfew) and where she was. I hate to say this. We are way past our \'policeman in shorts\' era HELLO!? I think she is an individual who can choose who she hangs out with and be responsible for herself. Yes, she\'s got a bf but it doesn\'t mean he has to revolve around her ONLY. I hate it when guys are MCP. (sorry galfrd for bitching ur bf in here lol)  Strangely, she told me she was ok him being like that. Geez, i don\'t think she\'s a masochist but really what is she up to? Do every strong front person has a subduded side? Hmmm I\'ve ever read about something about this lady - mind you, in the 16th Century ok. She was a very posh, proud and feministic lady in the daytime. But in the night time, she was submissive and acted like a w*ore. Split personality? schizophrenic?  Maybe what you need is some spanking lol - my favourite line for some boys/girls out there (at times) *grins* Anyway, I\'m glad I\'m rested in my bed today (sat). Just wanted to spend time with myself and be home. Town (nowadays) can be so crowded I could go nuts..  It\'s 1117 pm and i\'m going to bed.Is age catching up? You\'ve got the answer haha..  nite  '),(24167,3565927,'24,June,2004','2004-06-24 00:00:00','   urlLink    All 3 cousins - Clara Me Angela&nbsp; urlLink  '),(24168,3565927,'23,June,2004','2004-06-23 00:00:00','   urlLink    My 2 beautiful cousins at Far East Square!&nbsp; urlLink  '),(24169,3565927,'23,June,2004','2004-06-23 00:00:00','   As i met up with my 2 cousins - Clara and Angela, for dinner. I realise how much we have evolved - mind you - not changed :) It\'s interesting to see how we as kids when we played together in Farrer Park - no worries, just candies - sort of lifestyle. But now, shoved into the fast pace routine; where we call the Real World! I could remember how we would just meet up nearly every sunday at Farrer Park (before it got demolish) just to have family gatherings. That\'s when I remembered I was always excited.  My mother\'s side cousins are the ones I really talk more to and play a lot when we were younger, not my Dad\'s side. It seems to me when I spoke to my friends, the majority of them get along better with their mother\'s side. No Hokkien Drama mama (which u watch on TV) and rivalry. Why particularly the mum\'s side? It\'s really interesting - where we all can speak openly and just be ourselves. But when dad\'s side come, you kind of become like \'actors\' and be the type of person that is not YOU.  Well, it looks like the only concern right now on our agenda was to get our tummy filled with good food at Sushi Tei so we ordered. Yeah, it took us like 15-20mins - can\'t believe it?! haha I was just telling Angela imagine if we were all Librians (typically indecisive people); that would probably \'help\' the situation even better hehe So food came and we ate. It was not bad. Green tea was refillable and nice ambience  We actually had a family gathering at Pasir Ris (which I posted already) and I jokingly asked Angela why haven\'t u come for it? Well, I am so guilty (as I\'ve been spending time with Sean, only on weekends as his weekdays are in KL) of not spending time with my mother\'s side members - yeah I felt like I was like a \"use-to-be-maybe-well-known\" celebrity that will choose to re-emerge in some gathering lol... I chose this gathering to \'show up\' hhehe Anyway, rubbish on my side - I felt at peace when I saw everyone.  And just to get 3 of us for Karaoke - we have to flip our our organisers, I can\'t believe it... Yeah we all three had one each! It\'s amazing when you see one another grow up from kids to adults,marriage, kids and til the end of time. I think that\'s the beauty of kinship - it\'s never ending. No matter how bad you\'ve kept in contact, memories of each other will always stay so deeply within you.   Okie, going to gym now  '),(24170,3565927,'23,June,2004','2004-06-23 00:00:00','   urlLink    My Good frd Ana Dissing Cameras on her Wedding (I flew to Manila in March \'04), Miss Her :)&nbsp; urlLink  '),(24171,3565927,'22,June,2004','2004-06-22 00:00:00','   Ohhh finally, got my final exam done today. I am close to like dead but I still have this need to write the blog (as much I can update daily). I mean I\'ve been challenged by my classmate (whose name starts with the Alphabet  - that is the 6th position asecending; hey u DARE me OK hehe) to put his name so I\'ve officially accepted his dare and hope he doesn\'t try to subsidise me lol (Yeah, this classmate will know what i\'m talking haha). I met this classmate after my evening date with a nice tall Polish bred UK guy and it was a rather pleasant surprise. He is a well maintained 42 year old chap (for his age) and tall. Well, he brought me to a nice Indian restaurant which had a lift (yeah, in BOAT QUAY) that took us to the 5th storey. Oh boy, it was a nice view of Singapore\'s CBD and very breezy. We were early before 7pm and sat down.  You know first dates are like interviews. You\'ll never know what type of questions to be blasted onto you and anything happens?! Firstly, he looked like the picture he gave me - but I think he looked better in person. He is decent. That sort of Colin Firth appeal - normal looking but charming, yeah. I\'m into that :) Goodlooking guys (not stereotyping) sometimes think due to their looks, they are considered charming. But being charismatic comes within from a person and looks alone is not that essence to it - it\'s how the person speaks, looks at you and exudes presence. His singapore based company has been already been 10 years running and shuttles regionally for work.  The indian food was superb (seriously also a pleasant surprise AGAIN) and we headed to coffee bean to cool off for a drink (quite spicy for the poor Mark lar hehe). Yeah, I\'m glad our conversations were engaging, not only one sided and participative - with each own opinions. And I was also whispering to my classmate that what I\'ve learnt in Class (like Media Law) has probably earned me brownie points in sounding \'intelligent\' when I spoke to him lol...   Well, the night ended nicely after Coffee Bean where he took a cab back to town (home) and me walking to Raffles place to take my bus 14. As I was on my MP3, he smsed me a nice sms saying it was great to have met me and he had a special evening. I hate it when I\'ve got to INITIATE to sms such things to the other person - so I was secretly glad he did it and I replied to him lor ha ;) Anyway, I\'m happy dating right now and I think Singlehood has been the best thing so far for me.  Keeping you people updated..  More soon (Akan Datang)  '),(24172,3565927,'20,June,2004','2004-06-20 00:00:00','   Well, had an interesting chat with my classmate today and he told me via the phone that my thoughts being placed online are really quite detailed. But really, I guess there is really nothing to hide on my part. Maybe that\'s why I share quite strong sense of belonging with Jee at times and we speak our minds. Not sure if it\'s a good thing or not. Talking about speaking our minds, met up with a friend and we strolled the beach today.  It was a beautiful day. Loads of people slowly filling up the east coast beach. Then I started to tell my friend how my caucasian friend made racist remark and he came up to me telling me that i\'ve been too passive in terms of \'protecting\' the asians. Well, I think if it was me 6 years ago - I may just laugh up and ignorantly have no idea (making out the messages). I don\'t know but maybe I\'ve been portrayed that I\'ve been only into caucasian guys to my friends. But I\'ve been very \'United Colours of Benetton\' and not discriminated anyone - dated all races!  I don\'t deny I can be quite an easy-going person but I don\'t take any bullsh*t if I think it\'s wrong (Not trying to prove a point in here to my friend whom I was talking at the beach by the way :) Maybe I don\'t make myself clear enough to my friends but all I can say is that I\'ve made mistakes in how my relationships/dating/friendships are concerned and I do admit that there were probably times I gave in. But really, I am getting way too old for time-wasters.   I\'ll definitely be the first to admit I\'m not perfect. If I was, I\'ll be up there - not on planet Earth :)  Back to books,  B  P.S. By the way, I\'ve checked my diary and this is my breakdown of the demographic of      people I\'ve dated so far in my life! (and I don\'t go around telling you all the       people I\'ve dated right?)  :P       Caucasians - 30%      Chinese    - 50%      Malays     - 8%      Indians    - 10%      Eurasians  - 2%  '),(24173,3565927,'18,June,2004','2004-06-18 00:00:00','   urlLink    When serenity sets in...&nbsp; urlLink  '),(24174,3565927,'18,June,2004','2004-06-18 00:00:00','   I\'m glad that the blog doesn\'t require me to speak to produce words as I\'ve just had a marathon yesterday. Well, not that kind of marathon you would think :P hehe Me and Adele locked ourselves and sang karaoke in K-Box from 1930 til 0200 yesterday (5 and 1/2 hrs) -my throat is still sore mind you. I think I\'ve officially became a \'Singapore Idol\' without even auditioning for the contest hehe... Yeah, it was quite crazy and not that cheap also errr.. But glad we did it - it\'s a very good outlet!   I miss the times when I was in Polytechnic and we\'ll head to this friend who has a LD player - she plays Sammi Cheng concerts and I\'ll prance around crazily imitating her lol When one get older, I think you really don\'t have much stamina anymore :) Exams are round the corner and here am I still playing. Well, I\'m just a last minute person really. I seriously haven\'t in my life studied earlier before exam dates. In this aspect, my disciplinary department needs brushing up.  A scary fear suddenly hit me as I was singing Rene Liu - Afterwards (Hou Lai). The song\'s lyrics spell out quite a lot when I saw one of my friend singing to it. She was upset when she lost her lover. Yeah, I can feel what she feels now. I\'m a sucker for sad songs. Drama me. Maybe I should start getting my act in Korea and be some crying weepy actor? I was literally being identified as one of them (korean) when I was in Seoul - of course I need to really brush up on my verbal skills rather than surviving on this one line \"Han-gugonun mot\'amnida\" ( Means: I don\'t understand Korean)  Why has korean fever hit the crowd with so much frenzy? Are we not depress enough already as humans with our own problems? Guess we are just like how the cantonese people would have a saying \"You take sh*t and rub it all over you\" ( I don\'t really know how to spell it in cantonese hehe). Maybe it is an escapism when we see other people being depressed and we can re-assure ourselves we are fine/luckier than those people on the drama serials? Guess we should start experimenting on the housewives who\'ve glued themselves so faithfully to these serials and squeeze out the essence of this attraction.  Going for tea now (yeah, will write notes later, YAWNS, tell me what\'s new? hehe)  '),(24175,3565927,'18,June,2004','2004-06-18 00:00:00','   urlLink    Who is the fairest of them all?&nbsp; urlLink  '),(24176,3565927,'16,June,2004','2004-06-16 00:00:00','   Well, from the above title - it does say a lot for my situation now. While moving my stuff from Sean\'s apartment, I unknowingly twisted my neck. Not feeling very good and just not in the mood to do ANYTHING. *Grrrr* Well, I thought I could start writing my notes for my exams but just no motivation. I hate exams. Well, it\'s all the studying and when you enter the exam hall. It\'s like diarrhoea and you just let it ALL (what you studied) go in one shot. Some suffer from constipation (memory block) and some just let it \'flow\' all the way hehe...  Talk about flow - do ensure you check your money flow for your bank account. Apparently my ex\'s VISA card has been duplicated (in KL) and unknown transcations has been made without him knowing. Thankfully, the card has been frozen and he does not need to pay anything. Well, people are all out to get others eh. It\'s so dangerous to even leave your name or any particulars with anyone as you seriously never know what they\'ll do to it. I mean if you saw the National Kidney Foundation hand in hand with Aviva where Singaporean\'s particulars have been disclosed (though \'sworn\' with privacy) - when they do donations?! I mean hey, it\'s for the good of these kidney failure people - they shouldn\'t get some marketing (Aviva) company involve.  I can\'t wait, seriously can\'t wait, *ouch* to get my neck back..  Again  Me  '),(24177,3565927,'16,June,2004','2004-06-16 00:00:00','   The most dreadful thing when you break up is to get all your stuff back. Yeah, I headed to Sean\'s place to pick up all my clothes and whatever I\'ve left there. It feels weird. The house looks different. Someone else has been here - yeah his new lover - Gilbert. Do you know when you let a stray cat into your home...the thing the cat will do is to curl around the furniture, wall, anything that it can get in contact to as to leave a \'cat mark\' there. So in future, they will know they\'ve been there. Well, guess I could see the new cat mark now in Sean\'s place. Additional stuff. Feels as though that I\'ve slowly pieced up this puzzle in life and suddenly I accidentally smash on it. Destroying everything I\'ve built.   I slowly walked around the house - like the cat - but not leaving any \'cat marks\' but just reminiscing when I first touched the kitchen, prepared food under that hot environment (until he bought a fan later on for me in the kitchen), cooked my first pasta, sat by the closet and read my magazine as I drank chilled Ribena. As I continued to walk around the apartment - I just heaved and know I do not belong here anymore. I am glad I\'ve released myself of all the repeated times where I wanted out of the relationship before I officially implemented it to Sean. Maybe I\'ve never told you guys but yes, I wanted out even in Feb this year already. Why? I simply can\'t answer why but there were so many things happening - he was in a fix whenever he did anything - everything went wrong. If you think I\'m superstitious but sometimes the signs shown can be a gentle alert in life you know.  Packed up everything and called him up that I was over at noon. Now it\'s time to fill up my closet. Is there anymore space? No. I wish I had more room for anything else. The only space I\'ve got is for myself. This heart is aching. Slowly. But still hanging in there, steadfast.  I remembered the first poem I\'ve ever composed (for Sean) and this may be my last...  ---------------------------------------------  Love is when you first kissed me at the airport supposedly I see  Love is when I wake beside you and I can\'t stop admiring my beau  Love is when you showed me your honesty and filled me with that earnesty  Indeed distance has made it tough but love won\'t turn rough  Steadfast is in it for us imagine how this made us last  Await in few more years as we soak in joyous treats                                                         - Brendan  '),(24178,3565927,'16,June,2004','2004-06-16 00:00:00','   urlLink    Sunset in CityHall&nbsp; urlLink  '),(24179,3565927,'14,June,2004','2004-06-14 00:00:00','   I think I will go to bed soon. Just had dinner and panadol. It\'s barming today in Singapore *phew*. I wish I was soaked in the water or something. Yeah, will try to wake up early tommorrow and get some studying done as my weekend will be filled with other activites :) Met up with Clif yesterday and had quite a sumptous dinner at Sakura. Had deep fried squid, thai chilli chicken and beef kai lan - yeah, so filling! Ended up burping the whole night hehe.. Then his close crew friend Jean Paul (Italian guy) came over to chill out in his room and we had drinks. Well, the red wine was fabulous and just rattled rubbish as I was a bit high hehe.. It was good company and then Jean Paul\'s date came and we all parted, did our own things.   Mind blowing last night and waking up late in the morning - Life\'s pleasures :P Yeah, if only I could just be a nice tai tai that do not need to work and just enjoy life eh?! Isn\'t that all of our desires? hehe Clif will be flying off really early next day (he needs to be up at 5am) so we parted, so he also can hang out with his crew mates. Spoke a lot about his life, crew, principles and other stuff. Engaging conversation I should say but I just feel something is missing from him. Guess I\'ve reached a stage where I don\'t have any expectations anymore from him and just enjoy the company. Maybe I just need more time to heal. How I should re-adjust my life soon to the working society.  I just need to sleep now. Thank God the aircon is repaired today. Was suffering with for the past 4 days without it - quite a torture - as it was quite warm in the night. Will catch up more.  Nite nite  '),(24180,3565927,'14,June,2004','2004-06-14 00:00:00','   Yupe, one shouldn\'t reveal too much to others in life (especially to potentials hehe). Be as hard hearted as a rock. But with the tendency where there are times you just melt and be viscous,flowing in all directions. I\'m not sure why I\'m writing this but it has got to be the time I\'ve got right now on hand (I mean I should be studying).  Just was so surprise to bump into my really long lost ex poly classmate online at MSN. He has been in Vancouver (Studying) for a few years and now working as a designer. I remembered he told me he wanted to join Japan\'s finest graphic cartoonist and work in their realm :) Gosh, I wish I had that freedom he got on hand now. Can\'t wait to graduate by the end of the year, the sort of additional \'wings\' I\'ll be bestowed to do more in my life. It\'s not I\'m not able to do things now in life but it really does boil down to financially. But hey, doesn\'t mean when you start working - you\'ll be able to enjoy your money - NO TIME!  As my best friend says, I just probably end up like Miss Teo \"Rich\" and no life, if i\'m not careful. Guess Passion is also another essence to keep one alive and kicking. My flight attendant friend has been flying already for about 4 years and think he\'s thinking of going back to school. Well, he feels he can\'t be flying for the rest of his life and it does get mundane. What pushes us in life? Just being rich + successful = loneliness?   Yeah, you must be wondering what sorta equation I\'m blabbering about but maybe what we need is some kind of chemical in life. Lots of endorphines *yippee* hehe I want to be happy all the time but what are the things that make one happy? Quoting from Jeremy, what is the thing in life that makes you most angry and most happy? For a moment, me and his cousin - speechless. Could we not know what we want out of life anymore? Yeah, questions questions questions.   Holler on  B  '),(24181,3565927,'13,June,2004','2004-06-13 00:00:00','   urlLink    Lorry ride of our life&nbsp; urlLink  '),(24182,3565927,'11,June,2004','2004-06-11 00:00:00','   Just seriously can\'t wait to plonk myself into bed. Having the laptop beside me as I\'m writing this blog. Good News was up around 9am to get my ass to my hair stylist all the way in the west but I\'ve got myself a nice hair cut. Bad news is that 2molo is her last day as the current shop will be sold to another person (as she is freelancing) thus she needs to find another shop to fit her schedule. Yeah, I\'ve been with her for like 8 years and she\'s literally linked to my tastes in what i want out of my hair. Love her to death!! hehe To Adele who wanted a \'practical haircut\', please lar hehe... Janice (my hairstylist) is versatile - she can do whatever hair you want lol hehe  Quite fun just hanging out with my mum\'s side of her siblings and cousins. Watching the young cousins playing frisbee and soccer. Well, then we headed for dinner and my my I was in a ride haha Took a chance with Sebastian\'s (cousin avissa\'s husband) lorry and it was fun. I think it\'s nice to have a bit of open air driving from time to time. heheh...You can see it from the pic how fun it was haha..  Sweet dreams people  B  '),(24183,3565927,'11,June,2004','2004-06-11 00:00:00','   urlLink    So much ahead of her&nbsp; urlLink  '),(24184,3565927,'11,June,2004','2004-06-11 00:00:00','   I think I\'ve been the most productive person on Earth (for at least today). Met up with 4 different friend in a span of a day *gosh* haha My ex-neighbour\'s wedding was held at 0830am this morning and seeing she & Steven exchanging vows hmmmm it was really touching and beautiful. I could feel how she felt. She sworn to her lifetime husband that she\'ll be there for the rest of the time for him, even in time of poor, sick and anytime... Yeah I\'m a sucker for good endings. Why do we still believe it is out there? Do you or it\'s just me.   After being 2 years with someone just has been a habit or out of love. Obligations? I wish i could really tell you why. Maybe there isn\'t an answer for everything we do in life. Maybe it was all fixed and we were just props just doing through it. Everyone coming into our lives were a reason. They were here to stay. Or some not.  I\'m feeling most sentimental and vulnerable in the night. I miss those times where I\'m spoon to bed. yeah, it\'s a nice feeling of being hugged to sleep. Well, what to do?! It seems I\'m not the only one hitting splitville. Met up with other friend\'s today and realise he n she are single now. The universe is telling us something? I sure do hope not. I think it\'s wonderful to be coupled. But to whom? Should we just settle? How long should we wait?  I do feel like Carrie bradshaw in SATC writing this blog but it seems it has been helping me get over my grieve and displeasures in life (apart from bitching it to your friends in person haha ) Maybe in life, things do happen for a reason. Ahhh...  A pic of the bride walking down the aisle and heading to her happiness  Nite  B  '),(24185,3565927,'11,June,2004','2004-06-11 00:00:00','   urlLink    Me Myself and I&nbsp; urlLink  '),(24186,3565927,'08,June,2004','2004-06-08 00:00:00','   urlLink    Clif and Me in Stamford 11th June 2004&nbsp; urlLink  '),(24187,3565927,'08,June,2004','2004-06-08 00:00:00','   Wish I wrote yesterday but nope I didn\'t. It was a really long day.Had my exams in media law and boy (Thank God) it was fine! :) Damn lucky the questions were broken up into various section thus not a huge weightage is on a whole question or something?! Well, met Clif yesterday and I slept over. Had quite a nice glass of wine and we manage to catch up a bit. It was nice. We literally kept ourselves busy til like 5 am hehe.. Then got up at like noon half past 1. Had a nice clam chowder soup and hung out a bit. Yeah, he left for a flight to London around 9pm and will be back next tues *can\'t wait*   Just got back home from dinner with family and we had delicious fish steamboat. The chilli was so yummy mmmmm mmmmm It\'s my ex-neighbour\'s wedding 2molo and will be going for a church\'s wedding so I\'ll be asleep soon. Going to have a long day 2molo...  Took a pic of clif and me at Stamford  Ciao  B  '),(24188,3565927,'08,June,2004','2004-06-08 00:00:00','   Geez, brother just cooked a rather delicious Omelette with ham + tomatoes for me and it\'s delish...hehe Well, in the midst of completing my bloody final assignment for the semester, can\'t wait *phew* Just thought I pen some my thoughts for the day. I\'m in fact quite tired from baking in the sun today hehe...Was at the beach (did I ever tell you it is about 10-15mins walk from my home) and just sorting out my thoughts.  With regards to the Clif saga, yeah - he replied to my email. I think I frighten him off a bit and I\'m rather lucky to have him still wanting to meet up?! *gosh* Seriously, any guy who read the last email which I sent about (not keeping up with communicating and blah blah blah) would be freaked out and just want OUT. Well, guess he\'s a gentleman and we\'ll see when I meet after my exams the day after 2molo. Oh well, I seriously can\'t wait hehe..  As adele says with regards to men - they just like to take it slow and not rush. Some of them are either dumb or just want to be challenged or something?! hmmm Strange and complicated eh...I wish life was just more simple. I guess being alone is not easy and am not really use to it. Attention being given in the past was rather intense I guess, thanks to sean... But well, I\'m blessed to have my friends just being around me.  Thank you guys for helping me through these bad times and I am seriously lucky. Even my boss lavinia has been supportive and just help sort sort out my thoughts. I better get going and finish up the final portion. Then get some sleep. Will meet adele for lunch 2molo and catch up a bit.  I miss travelling *sighs*...  Alrighty,  Love  B  '),(24189,3565927,'07,June,2004','2004-06-07 00:00:00','   I seriously have no idea what came over me but guess laying out my thoughts can help me be sane I guess lol Well, this idea is literally Jee\'s as I\'ve been reading her blogs constantly and it\'s quite a nice outlet where you n your friends can frequent n read what\'s going on in your life (apart from emailing)..  Maybe I should take this chance to officially do a press release in here too eh?!  \"Solo Me\"  Yupe, just 2 weeks ago - I\'ve underwent a rather traumatic breakup. I\'m single. A little bit glad and a little bit sad. I mean it was 2 years with Sean and I loved him a lot. Well, we were drifting and lots of other issues erupted along the way - that I felt I can\'t love him enough to deal with those issues. Yeah, I wanted out and I got what I wanted.We are friends now and amazingly possible. He went on his first date yesterday and it seems to be smooth sailing :) I\'m so HAPPY for him!!! It\'s seriously hard to find someone to click and he\'s blessed.   Well, I\'m right now only bogged by books and exams (same old same old me). If only books and exams can give you orgasms?! lol Was out with Gerard and we ended up having some blurred cashier from U2 that f*cked up the whole totalling of the bill (not helping when we were starving)..*yikes* So much for discounts but we got it sorted. Yeah, Gerard was absolutely pissed.. Who is not guilty right now for shopping during this Singapore Sale, Hands up? :P  It\'s weird but I think I feel quite emotional now. Yeah, I need a hug and someone who I can lean closely to. It\'s hard being alone and guess it\'s all setting in.I guess the questioning part of the program is over (quoting from SATC) - the party is officially over, it\'s an end of an era. Well, it\'s strange how things are now. And seriously, this blog is a rather good outlet *Gosh* You can just write and write...  I wrote a rather stupid email to clif (whom I\'ve been corresponding via email) and saying to him that I am beginning to like him but am confuse. I said I was angry why he did not email me when I emailed him 3 times? I am so scared. I don\'t want to fall in love, not at least now. He touched me like never anyone in my life and this scares me. That email might even be my last and not sure how he would react to it - yeah, guess I blew my only chance with a guy but yupe...hmmm  I need to get back to my last assignment. Want to hand it up 2molo. I am exhausted.   As always  B  '),(24190,3565927,'31,July,2004','2004-07-31 00:00:00','   As we were kids, we throw tantrums and make our parents give in to our demands. Yes, they would eventually give in to us but they are biologically linked to us. But the world, that is in front of us - not linked in any way to us. Do you think they\'ll submit to us? Especially emotionally - yes the grey area. We grow older and become more emotionally in tuned to the surroundings or we need to become more intellectually challenged. My friend who is currently following Sex And the City from Season 1 til to the end says that this show brings hope to people, fake hope. I retorted to him and said it\'s not true. The show kinds of bring us to think and show the various situation/consequences we may bump in our lives. You probably think I\'m contradictory now as I said from my last blog that movies are bullshit and why right now Sex And the City is the \'Dalai Lama\' for humans. Well, maybe I could relate to myself to the ladies some way or another to the characters (though I\'m in Singapore and without the female anatomy hehe).  I think I\'m back to my old self. I\'m glad. I know it\'s only been 5 days since I was back but trust me. It was hard. I fell hard with Eric. I was even thinking (naively) that he may like me and avoiding me like a plague?! (I mean come on, you may never know too right). I think different guys react differently when they found their equivalent match. I think what we had was all too fast - in terms of connecting with one another. Of course, it doesn\'t help when he was there and I\'m here. I woke up and the \'coffee\' has awoken me. The caffine this round will be strong enough to keep me UP this time.  Maybe you think I\'m crazy to take Song lyrics to relate to me but it\'s easier. When I was in Bangkok with Eric, He Sang the Chorus of TaTa Young (he said this Song represented me) \" Naughty Naughty Naughty, Sexy Sexy Sexy, Bitchy Bitchy Bitchy, B..(Brendan)\" I found that really funny and just got charmed (yupe even more haha).  As we were hanging out and cuddling in the bed. Just being togther. Then the Song that was playing in my head later on was TaTa Young\'s - I Believe \" I feel the heat around me..I feel the beat surrounds me..Could this be real, I wonder...No need for hesitation...It\'s time for celebration...Will this be the night I\'ve waited for...\"  After I got back from Bangkok with a reality check. Now TaTa Young\'s - Cinderella plays in my head \" I don\'t wanna be like Cinderella...Sittin\' in a dark old dusty cellar...Waiting for somebody, to come and set me free...I don\'t wanna be like Snow White waiting...For a handsome prince to come and save me....On a horse of white, unless we\'re riding side by side...Don\'t want to depend on no-one else.....I\'d rather rescue myself...\"  Nuts to have teenybooper songs to keep me alive. But it\'s working hahaha Well, this is to keep me sane ok. There is never a fix time limit to get over somebody. But I think I\'m over Eric. I don\'t want to be hurt anymore, deep in sorrows. He was happy memories for me. I have to let this go. My heart will be hard as stone by my next trip to Bangkok soon. I don\'t need all this now. My friends have been so supportive and thanks (you know who you ARE) :)  Okie, time to end. Close box. Stop playing TaTa Young in background :P I\'m back! Samantha (but younger, sexier, more naughty and more bitchy)  '),(24191,3565927,'29,July,2004','2004-07-29 00:00:00','   urlLink    Happier days - Orchid Ambassador :P&nbsp; urlLink  '),(24192,3565927,'28,July,2004','2004-07-28 00:00:00','   The Cover Girls&nbsp;- Wishing on a Star    --------------------------------------------------   I\'m wishing on a star  To follow where you are  I\'m wishing on a dream  To follow what it means   And I wish on all the rainbows that I see  I wish on all the people who really dream  And I\'m wishing on tomorrow praying never comes  And I\'m wishing on all the loving&nbsp;&nbsp;we\'ve never done   I never thought I\'d see  A time when you would be  So far away from home  So far away from me   Just think of all the moments that we\'d spent  I just can\'t let you go from me you were meant  And I didn\'t mean to hurt you but I know  That in the game of love you reap what you sow   I feel it\'s time we should make up baby  I feel it\'s time for us to get back together  And make the best of things oh baby  When we\'re together whether or never   &nbsp;   This song kind of represent how i feel now. Actually, I feel a part of me being lost right now. Yes in Bangkok. I sometimes think if I haven\'t met Eric, maybe my life would be better - less fuss.&nbsp; I keep telling myself he is a player - an assh*le so I won\'t feel so bad about the situation. I spoke to Sean about it and he said when I could identify Eric from far,it\'s all fated and meant to be (considering he is my ex, he spoke quite sweet words to me). Why can\'t&nbsp;the&nbsp;unsure part of everything be more obvious? Not hidden...    &nbsp;   Jee told me that I am in control of the situation and shouldn\'t be upset/down about the whole thing. I&nbsp; spend the night crying for a while, curling up in the corner of my bed. I don\'t want any other guy to touch me, have sex with me or want to do anything. It\'s like I was touched by an Angel. Maybe I am exaggerating and spouting nonsense with the whole feeling. How much can u know within 14 hours with a person? But&nbsp;watching the movie \'The Sweetest Thing\' where Cameron Diaz met this guy at the club. She never believed in love. She felt something for him. How she travelled nearly 2 hours (just like me soon to bangkok over National Day weekend), driving (for my case I jet) to go to see this guy\'s brother\'s wedding. Turns out he was the one who\'s getting married. She gets disappointed. All she feels now is nothing. But Fate brings of them together.     &nbsp;   Well, movies are nothing but bull-shit I guess. How many of us would end like the leads in the movies? A small percentage or non in fact. Maybe I\'ve been starved of some excitement in my life. I need something to ignite me. I can\'t wait to graduate at the end of the year so&nbsp;I can stretch my wings and fly further, far far away. I am going to meet Sean for dinner now. Bought him a shirt and going to catch up. I miss being a couple - not my rebound side talking :P     &nbsp;   Slowly recovering     &nbsp;   B  '),(24193,3565927,'28,July,2004','2004-07-28 00:00:00','   The pics you\'ve seen below are selected pictures of Bangkok and all the pictures can be found in  urlLink http://photos.yahoo.com/bangkokgals04     &nbsp;  '),(24194,3565927,'27,July,2004','2004-07-27 00:00:00','   urlLink    Upgraded to Thai Airways Business Class - Thanks for Regina! :)&nbsp; urlLink  '),(24195,3565927,'27,July,2004','2004-07-27 00:00:00','   urlLink    On our way to airport stuck in jam&nbsp; urlLink  '),(24196,3565927,'27,July,2004','2004-07-27 00:00:00','   urlLink    Delicious Home Cook Thai Dinner at Regina\'s Friend place&nbsp; urlLink  '),(24197,3565927,'27,July,2004','2004-07-27 00:00:00','   urlLink    Eric Me Carine Regina having drinks in Cafe&nbsp; urlLink  '),(24198,3565927,'27,July,2004','2004-07-27 00:00:00','   urlLink    Carine n Me relaxed at Dinner&nbsp; urlLink  '),(24199,3565927,'27,July,2004','2004-07-27 00:00:00','   urlLink    Our tutuk ride&nbsp; urlLink  '),(24200,3565927,'27,July,2004','2004-07-27 00:00:00','   Geez, how should I even start in here. Firstly, it\'s weird to be back from BKK after the weekend as it was my first trip since I was there when I was 9 years old and this round it was so much fun with the gals (Carine and Regina). Even better is when I did meet someone nice; ironic thing is that the guy was french. Well, maybe some of you knew I had a torrid dating with Olivier - Remember that photographer, yeah I did say I would swear off french for the rest of my life. Looks like life is pulsated in such a way is when you say I\'ll \'NEVER\' - it\'s usually the next time your \'NEVER\' will come colliding into your life unknowingly *sighs*  First day (24-07-03)  Was really excited. Couldn\'t really sleep the night before so watched Murial\'s Wedding to try tire myself out - such a funny flick about an Aussie girl\'s objective in life to listen to ABBA and wait for her prince charming to drop into her life inevitably. Well, with all the ABBA that you\'ve got from that movie, it really does bring back some nice radio moments we had in school.&nbsp; Anyway, then Carine gave me a buzz and we chatted nicely for over an hour before we hit the sacks. Woke up and then waited for her, John, her mum to pick me up via the taxi. It was drizzling the whole morning. Just couldn\'t wait to get out of the country. Just felt I wanted to \'fly\'. Met up with Regina\'s Uncle plus his entourage. Well, Ah Kai (regina\'s uncle) is a regular jetting up to BKK and knew his way around very well. I think a lot of people headed up to BKK would talk to him also he made friends quite easily with anyone :) Well, Carine\'s mum was having a lot of reservation about her going for her vacation but I assured her she would be fine. So we headed to the foodcourt a while, had a quite bite. Then we started to walk towards the boarding gate.&nbsp; I was so surprise with the numbe of people headed to BKK. It was hordes. Many indians were also present as this flight would stop over BKK and fly to bombay too.   We got quite nice seats, all 3 seated together. Sat down and up we went. Arrived in about 2 hours time. Then Ah Kai got the bus to send us to our hotel. Boy, after 16 years not being there - I think I saw so much changes in BKK. Definitely is the traffic - the highway was a good thing that ever happened for the country apart from the BTS (same as Singapore\'s MRT).&nbsp;Reached the hotel and all 3 of us had a culture shock. It looked like a karaoke sorta prostitution kinda hotel place - even hotel 81 would look nicer lol :P Anyway, we got into the room. The design was so retro and Taj Mahal. I think the girls were really shocked - even Carine wanted out to stay in another hotel. I was amazed and laughed so much with our hotel rooms. Well, instead of reminiscing of our new \'fabulous\' room. We need to start off the new first day with some shopping. So we got our butts off and took the&nbsp;BTS. Enjoyed the ride and got to MahboonKrong Center. It was huge and it was amazing (just like their country\'s tourism slogan currently ;) ! Well, got nice bargains and a SIM card. The instruction for the SIM card was all in Thai but finally manage to figure it out as I was trying to call Eric (french guy whom I was going to meet up in BKK so he could take me around to the gay clubs). Finally manage to send him an sms and he replied to meet me up. Then the we all got crazy and took our mane for some nice hair cut. 300 Baht was reasonable for a cut wash blow ( well, until I finally get a reply from my Hairstylist, Janice where is her new base for my haircut). Got myself quite a nice hair stylist - very attentive and spend quite a long time cutting my hair. Well, it finally balanced out the sickening \'slope hair cut from army\' blunder when I was in reservist weeks ago.   Got into the cab and headed back. Then we all showered and uncle took us out to eat dinner. Well honestly, have no idea where we were but it was further up (10mins) from&nbsp;Silom. It was a nice chinese restaurant. We had like 10 dishes, yeah eaten like a chinese wedding ten course meal menu. Had suckling Pig (a whole pig), pepper crab soup, stir fried cabbage (without oil, very interesting and healthy), deep friend pork and many more. Then, we had a nice visitor. It was a baby elephant. Well, Ah Kai bought some sugar cane and fed the elephant. We were in the midst of the crab so I had to whip out the camera and take a shot. It was so cute and all of us fed the baby elephant. I think all of us were afraid of the elephant\'s wet trunks and if it would swing against us lol So after dinner, we all headed to Silom\'s gay area. It was filled up with so many connecting bars and pubs. The Thais would harass us to patronise their shop. It was so aggressive. Then we all entered Future Boyz. Well, you would never imagine how much d*ck you could see in the night - yupe ;) Thai guys (mostly straight by the way) dancing to music with no emotion. It was a rather depressing sight and I felt very uncomfortable. Yeah, this is the \'anal-virginal\' side of me you have not seen but I just wanted to get out of the place. It screams weird. The mobile phones did not have reception when u r inside (looks like the Singapore\'s Esplanade was not the only successful industry&nbsp;in shutting mobile connections within theatres). Then I got out and called Eric. It was terrible - he couldn\'t hear me properly. Even when I spoke to him earlier; he couldn\'t hear me properly. I had this hunch that maybe we weren\'t fated to meet up at all. But anyway, I waited outside Future Boyz while I see the crowd walk pass me - can also identify some Singaporeans from time to time haha.. Then a funny thing happen. The future boyz club was beside Dick\'s Cafe (yeah a rather cliche name within a gay area lol) and I got an sms from Eric that he was waiting upfront in the cafe. There was the outside and aircon. Obviously I thought it was the outside so i saw a guy who looks (I never knew what he looked like) the closest to Eric\'s Age (34) and approached him. Turns out - he was not Eric. Then, after the wrong identification - I saw a tall guy bumping into his drink. Yeah, my gut told me it was him (typical Pisces - careless in nature lol). So I called his mobile and saw his mobile rang. Then I walked towards him. Well, then we smiled and sat, talked. It was nice, pleasant and unexpected. So he had his Pad Thai while I yakked my while through and the charming thing was that he managed to speak his mind, not a passive listener ( definitely a cookie point for this plus).   Then we left the cafe and walked towards to another Coffee Cafe to meet up his group of clicks. His \'sister bemas\' was a Singaporean guy 47 who looked 30 (yes i saw it with my own eyes). He lived in Singapore not for a while but travelled and lived overseas mostly (9 years in Denmark) - was a teacher/actor/multi-talented guy. Smart, witty and funny. Well, we all had a good laugh and chatted quite a bit about Singapore plus other stuff. The Indian + French guy (nice couple) were slowly settling in BKK buying furniture and were together for 20 years *swoons* Yes, 20 years together, faithful. And, there is such thing as longevity in relationships. Then thought we would all head clubbing together but they were exhausted or their thing to do. Eric was a regular in the cafe and seems the Thai gays were quite friendly to everyone. Everyone seem to know everyone.&nbsp;So Eric took me to the linked clubs nearby. It was so packed like a tuna can. More packed than Taboo in singapore. Tell you, you need at least 15mins to move from one end to another. Muscle marys prancing on the stage and we just stand upstairs got our drinks. While we people watched and spoke a while, decided to infiltrate the dance floor together. Then I lead the way and got ourselves inside for a while. But I was sweating like a pig and Eric was considerate enough to want to leave the club. Then We took ourselves to the nicely designed Arabian pub beside the club. Nice cushions and we continued more chats. He was not your typical french spoilt boy (who loves his country to death) and we talked a lot about travelling, people, thoughts about life. I think I was so starved of intellectual stimulation in Singapore and this guy was just so sincere + funny+ smart enough, charmed the hellout of me :) Oh well, we chatted til 2am and it was the end of the night. All clubs and pubs close at 2am in BKK. Then we walked out and I proposed to him about the cramp room (which was the truth but I wanted a good reason to sleepover with him)&nbsp; which I had and the girls snoring (sorry carine and regina) to stay over with him. Before I could ask him the question, he could read my mind. Well, we took a nice stroll back to his very cosy apartment. It was just a nice room connected with toilet. He was neat. Then he asked me to take a shower and when all my clothes were off, he hanged them up for me - which was sweet. After my shower, I found myself in a nicely lit candle bedroom and he went to have his shower. I just relaxed and lay on the bed to rest. When he walked towards me and lay next to me. It just felt to nice and comfortable. Felt right. U know that sort of feeling when you know it is the right feeling/time. We kissed gently and he stroked me in the most nice way you could ever imagine. He enjoyed kissing me on my face, hands, head and shoulders - just what I needed. We had close to 5 hours of foreplay but had a good balance of chat in between. It was perfect. He did quite an amazing thing, we \'did\' it while he lifted me off totally as he was standing. Well, I\'m not exactly light so he lasted for 1 min and we laughed it off that i was not 7kg but 70 kg to carry hehehehe... But he would change positions and it was just the way how I wanted it. Sorry, too much information but this is MY blog :P  &nbsp; Second day (25-07-04)  Then we just relaxed and he finish it off the \'morning\' at 6:30am by spooning me off to sleep *gosh* So i got up at 11 and headed back to the hotel to meet up with the girls. We had a 2pm massage appointment so we ate lunch with Ah kai before we went there. Lots of questioning came from the girls and then I spilled the beans that he was pleasant, good. Apparently, the girls told me that some Thai money boys at Future Boyz wanted to pick them up for the night hhahaah They were winked at and they ignored ;) Anyway, we had more exotic food. Duck\'s Tongue and other stuff at lunch. Arrived the Spa cum massage place and it was so peaceful - very simple and chic design. We were greeted by the receptionist and then had our footware changed into their slippers. Waited a while and then they took us to the room. I shared the room with Carine where we had a nice outside view. We changed into our PJs and laughed at one another as we looked like hospital patients hehe... Then we started with the message. Felt so relaxed and had half way fruits n chinese tea served to us. The whole massage was close to 2 hours 20mins for 600 Baht.&nbsp;   In between my massage, I sms Eric to ask him to join us for tea in Silom as we would be heading there later on but I got an sms like this \" Sorry can\'t and don\'t be jealous&nbsp;as i\'m busy. Have to do another sg boy my systera bemas. C u 2nite!\" Honestly, when I got this sms I was like - WHAT?! I don\'t need so much information about your sex life. Yeah then I realise I kinda was trapped in the bad warp - I may just like the guy syndrome. Do you know? I think I\'m cursed at times. Whenever I like someone, it usually turns out bad. Like very very bad and I won\'t end up with the person. My 2 past relationship was initated by ex boyfriends and they probably liked me more than I liked them. Maybe I was thinking too much but I just finished up the massage. So we all just headed back to hotel and showered, changed. Regina\'s Thai friend invited us over for a home cooked thai dinner meal and we all went. It was a very nice dinner and so delicious. You could see how warmth and hospitable the mother was. She was so nice and gorgeous. We just communicated with our limited Thai but all her 3 sons could interpret for us from English. The night ended nicely and we took the underground (MRT) to Lumphini; Night Market. The gals were nice enough to keep me company til&nbsp;I met up with Eric at around 10pm. I was contemplating to meet up with him but Carine told me to keep cool, just not be brash. After shopping we all went to the same cafe (where I met up with Eric and his friends before) and we all sat down. Then Eric came, we all seated ourselves nicely in a comfy area. So I asked him (sacarstic) how was his Singapore friend in the afternoon? Then he said it was good, he was with Bemas, whom I met last night. I suddenly realised he was with Bemas (the nice 47 singaporean guy who was his best friend, so it was not SEX) Aiyoh, I felt so embarassed and he cleared up the air. I was relieved heheh Then we all chatted a while and then I sent the ladies to take a cab. While&nbsp;me and Eric&nbsp;slowly walked back to relax at his apartment.   Nicely showered. He made nice tea for me. Then coming from my behind, he wrapped his arms around my waist, with his head resting on my shoulders while I scanned through my digital camera pictures to show him. I tell you...it was like a slice of heaven. As seeing the pictures, he would gently stroke me, kiss my neck and hug my closely to him. I think this couldn\'t get any better. We then had tea and relaxed beside one another. With Eric, we could stare at each other and not speak at times. It was like we were connected. Then a really freaky thing happen. I asked him of his ascendent for his astrology. He insisted me to guess and I guessed correctly at the 2nd time. Then, I matched all up.   &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Eric&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Brendan&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Outcome  Sun Sign&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Pisces&nbsp;&nbsp;&nbsp;&nbsp;+&nbsp; Scorpio&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; =&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Best Match  Ascendent&nbsp;&nbsp;&nbsp;&nbsp; Saggi&nbsp;&nbsp;&nbsp;&nbsp; +&nbsp; Gemini&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; =&nbsp;&nbsp;&nbsp;&nbsp; Good Match  Chinese&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Dog&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; +&nbsp; Horse&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; =&nbsp;&nbsp;&nbsp;&nbsp; 9/10  &nbsp;  Maybe I\'m not suppose to read too much into such things but it was just the intellectual stimulation that mattered so much more to me. His company was priceless, not sure if he being a french (a romancer) or what. But I don\'t think he would do what he did to me to everyone - well of course i\'ll never know if he was a \'player\'. But then, he told me that he hardly open to people he knew (which he did) and asked me how I felt when he made love to me *aiyoh*... I mean, it was like a piece of chocolate which I was nibbling for the whole trip when I was with him, in&nbsp;and out of bed. Really, it was his height, watery blue eyes, not super good looking but very genuine person with funny personality that bowed me.&nbsp; I spooned him to sleep at 5 am and he got up for work&nbsp;at 830am the next day to get ready for work. Then woke me up later and I got ready also to head back to the hotel. We walked to Coffee World and head a nice cold coffee together. Both not morning people and just looked at the morning crowd walked past us. Then he walked me to my train station and bade me good bye. A nice day awaits me :)  Last day (26-07-04)  Took a BTS back to the hotel. Showered and apparently, horrible things happened to the girls last night when they went to the night club to celebrate Ah Kai\'s birthday. He asked the Thai money boys to come along and each accompanied everyone in our entourage. Carine and Regina felt uncomfortable. The guys were so persistent and wanted to come into their bedroom. Carine freaked out the most and she just wanted peace. Anyway, she hugged me when she first saw me in the early morning and I just comforted her it was alright *sighs* Those Thai money boys - the things they do in the name of dough. Then we all got ready and Carine n Me had a mission. To find the bridal shop to get quotes for her sister\'s upcoming wedding. The cab driver did not know the exact route (but at least he stopped to ask for directions as compared to one which we had that took us for a RIDE) so we got off far end of the road. Had to walk for about 25mins and it was so warm. But we finally found the place *thank God* and just got quotes. Met with Regina in Silom and we all went to Anna\'s Cafe (Named after Anna and the King). The food was delicious my goodness. We had thai salad, Toffee Banoffee (some toffee cake with banana and nice cream), thai fish cake, beef green curry and we just ate ate ate. Did last minute shopping and i got myself season 1-4 Sex and the City DVD for 2000 Baht (70 SGD). The quality is surprising not bad and I think it was a steal as one DVD bought via Amazon.com cost 70 USD :P haha..  Then we all head back to the hotel. Stuck in traffic jam but then we rushed like mad. Did last minute packing,showered and got ourselves a cab. George (Ah Kai\'s friend) was waiting for his Suckling pig (which he will take back to Singapore?!) asked us to go to the airport first. Then we checked in and guess what. Got ourselves upgraded to Business Class with Thai Airways *gosh* So nice surprise. George was quite close of not getting upgraded but in the end got an upgrade too. We were excited and got up the plane. The service was amazing. Better food and more choices (of course). A singaporean lady (who was seated in the business class) asked me how much did I pay for the business class tic while wating with me at the toilet so I told her that we got an Upgrade due to Regina\'s PPS (SilverKris Priority) card. She envied us and said she had to pay 1000 SGD (yeah we paid 250 SGD for the air tic hehehehe).   Arrived back to Singapore and home sweet home. I do miss BKK and having a person beside me. Eric in fact emailed me a short note, asking me of my flight back to Singapore. I really want this \'butterfly feeling\' of liking someone to be out of my system. I\'m afraid. I miss talking to my conversational equivalent - Eric Dubois  What should I do?  &nbsp;  &nbsp;  '),(24201,3565927,'18,July,2004','2004-07-18 00:00:00','   urlLink    Me&nbsp;Regina&nbsp;n Femi&nbsp;&nbsp; urlLink  '),(24202,3565927,'14,July,2004','2004-07-14 00:00:00','   urlLink    Regina, Femi, John, Me, Geri, Carine &nbsp; urlLink  '),(24203,3565927,'14,July,2004','2004-07-14 00:00:00','   urlLink    The 3 New York (traveller) hopefuls in July 2005 :)&nbsp; urlLink  '),(24204,3565927,'14,July,2004','2004-07-14 00:00:00','   urlLink    Angela n Me having a breather&nbsp; urlLink  '),(24205,3565927,'14,July,2004','2004-07-14 00:00:00','   urlLink    Early birds - Clara n Me in Karaoke&nbsp; urlLink  '),(24206,3565927,'14,July,2004','2004-07-14 00:00:00','   urlLink    My British Frd John with Me in Indochine New York July 2003, miss him!!! :) He\'s a nice guy lar..&nbsp; urlLink  '),(24207,3565927,'14,July,2004','2004-07-14 00:00:00','   I think I just had the nicest time yesterday. Well, I\'m quite a simple person. The little pleasures in life make me happy. Seeing others happy - I\'m happy :) It just feels precious. Firstly, had quite a nice catch up with Serene whom I\'ve not contact for 2 years. Well, she has been busy with social work and it seems to be a good calling for her! Plus, she told me she\'s getting married next year in May, Gosh - I\'m so so happy for her. Finally found someone whom she can share her life with :)     &nbsp;   Then, had another \'Singapore Idol\' Session with my cousins (Clara and Angela) at Kbox Karaoke on Saturday afternoon from 2pm til 7pm. Well, this time it was a good wide spread where all of us could sing equally throughout the time frame. It was so much fun. We felt like kids all over again, prancing and making silly moves to the MTVs heheehe... I think these 2 gals are also excited about their Taipei trip (same day I\'ll be leaving for BKK) so they are discussing quite a bit too. It\'s nice to be able to keep in close contact with my cousins (considering I was quite guilty of neglecting them when I was with my ex lol).     &nbsp;   After which I had to Regina\'s bf place where we all gathered and hanged out. Her boyfriend cooked african food and it was delicious! *yums* Well, it\'s healthy, spicy and good hehe... Then we were playing african poker (Ban Luck) games for a while. Quite interesting ;) So we headed off around 11nish home. Got home, saw John (My british friend based in NYC) online and we chatted a bit. This sweetie called me on my house phone as I told him it was awhile we chatted *swoons* Ooooh... He is such a doll!! Yeah, he has been my online friend for 4 years. We met twice in our life (once in singapore and 2nd time in NYC when I was there holidaying with Sean).     &nbsp;  Most of the time we communicate via MSN and it\'s good we can click. He is so generous as he allowed me, clara and Angela to crash in when we are visiting next year July. I tell you, this call from him was like a cherry to the Ice-cream topping...just incredible! (Honestly, the chat we had were like one of those conversations in Sex and the City hehehe...just the difference is that I\'m in Singapore and he is in New York, the wonders of technology eh?!) &nbsp; Yes, he will be going on a hot date with this nice asian guy - BOATING ooooh so romantic hehe... I am so happy for him ;) Hope&nbsp;things work out for them...    &nbsp;   Yeah, nice to have passion in life and just living it out! Just inserted my new school semester timetable into my organiser and it looks i\'m going to start to get busy soon. I\'m glad Jee got a new job in TTSH and seems to keep her nicely busy (as she only has one module this semes). Going to laze more, such a nice sunday!     &nbsp;   Your laziness..  '),(24208,3565927,'14,July,2004','2004-07-14 00:00:00','   It\'s official - this blog is like a weekly column. At least from the feedback I\'ve got from my Best friend (Gerard), my wacky and gorgeous cousin (Clara) and my zany plus funnily bitchy ex-ibm colleague (Carine). So this shows that I cannot be lazy and I need to update it regularly lol... Was just on the line with my cousin and she told me she laughed her head off when she read the previous blog about my reservist experience in Erin Brockovich, but yeah, sometimes -seeing other people\'s tragedy is one\'s comedy lol  My boss just left for South Africa yesterday and will be back on the 30th July. Yupe the whole office will be rather empty but I can\'t wait to do certain organisation to the files/documents and other administration. It\'s like your work seem to be swarmed by other people\'s work and not YOUR work?! So it\'s good that she left for a break. I will also be having a break myself when I head to BKK from 24th til 26th July. I will be my 2nd time since I was there when I was 9 years old :P Well, guess it should have changed quite a bit as I remembered I was stuck in traffic with my dad in the taxi (heading to the airport to pick my mum and bro, but they didn\'t turn up) for 3 whole hours - ONE WAY! So by the time we got back to the hotel, 6 percious hours are wasted. Guess more flyovers and better road adjustment should make a diff by now in BKK, let\'s hope..  I can\'t wait to be mistaken again for a thai lol Well, I think in general we asians can blend into any asian country quite easily as it\'s hard to differentiate from one another when you don\'t open your mouth. Guess the koreans must be thinking i\'m a snob and hating me for not speaking korean lol Don\'t play play ok, our looks all very versaltile leh..hehe Having a nice cup of tea and some bread. Well, played scrabble til 6 am last night, still recovering. Quite a nice quiet day.  Anyone keen on asking me out? ;)  B  '),(24209,3565927,'10,July,2004','2004-07-10 00:00:00','   Today is the final day of my reservist, thank God! :) But before this nasty day ended - it was soooo much drama before we manage to get out of camp. Maybe you people won\'t know but army at all times is all about admin. You have to sign this, sign in, sign out and sign sign sign. I mean if not this - guess this world would be a better place?! haha Well, my CSM (Company Sgt Major) appointed me today to get signatures from the various departments in our unit today. I mean we are talking about big shots in the camp - that I\'ve got to get their signature (which is for the clearance that everything is received on their side and we can leave camp officially).   Then this irresponsible guy who does the stores was sick. He lives in SengKang and does not even call anyone to inform them he\'s not coming. Falls asleep. Ya. Everyone was kept waiting and was wondering what the hell was happening? Everyone (in other companies) have left after an hour but we were stranded there. To make matters worse - the freaking signature was to be accompanied with the consolidation of all the camp-passes (so this guy is missing in action,how can we do clearance?!)   Then, he turns up around 1120 - with a long face (supposedly sick) and unapologetic. Tell you, I was not there but running around like Julia Roberts in her Oscar winning movie Erin Brockovich where she helps the plaintiffs to get signatures to file a petition (but for my case to help the whole Company BOOK OUT as soon as possible). I tell you - it was stressful, of course if I had boobs and cleavage (in army) - it would probably have sped up the process lol  Anyway, the Majors and other high ranking officers were rather nice. I had to PR a bit here and there but it was worth it. Apart from the chicken run, I did not really sleep well (again) as I\'ve been playing online scrabble. I\'m such a nerd but haha it\'s quite addictive ;) Well, I ended up meeting up with Jee and her mum to attend a skin care beauty course. They were introducing face products and boy it was rather amazing. I bought a cleanser and hydrating spray for starters to try out first. Will let you know if it\'s worth it or not :P hehe..I don\'t deny it\'s effectiveness but guess beauty products vary with people so it\'s hard to gauge at times.  Today is saturday. Spent whole afternoon attending the course and headed home for a short nap. Boy I miss napping hehe.. So i\'m awake now and just thought I pen my thoughts before I get hit by more and more other loads of stuff.  Thankfully my aunt, dad\'s sister (who was based in Jedah) now in Shanghai called and we chatted a bit. She felt apologetic about not being physically here for me when I was undergoing the *jury decision making* of my previous breakup but assured her it was something I had to do myself (apart from listening to other\'s opinions). Anyway, I can\'t wait for my break soon...  Maybe I go watch some VCDs now..or get laid? :P  B  '),(24210,3565927,'06,July,2004','2004-07-06 00:00:00','   urlLink    Pout n Jee&nbsp; urlLink  '),(24211,3565927,'03,July,2004','2004-07-03 00:00:00','   urlLink    Tongue-er&nbsp; urlLink  '),(24212,3565927,'03,July,2004','2004-07-03 00:00:00','   urlLink    Jee Mike n B&nbsp; urlLink  '),(24213,3565927,'03,July,2004','2004-07-03 00:00:00','   Trust me, there must be a reason why men and women - there is a difference. I\'m not trying to segregate both sexes but trying to weigh the pros and cons right now :)   Men =========                               *Need to do National Service   (Army in Sg)     *No periods                                   *More expensive, less choices of   clothes      *Cannot multi-task (unless you   are gay :p)    *Need to be the sole bread-winner             *Answer to every question starts with          NO                                            *Sensitive (with regards to their ego)                                                           Women ============  *DO NOT need to do National Service  *Periods monthly  *In-expensive, More choices of clothes  *Can multi-task  *Can be tai-tai/work/super-mums  *Answer to every question starts with  YES  *Sensitive (with regards to   others/environment)    Let me see - more or less the main points are given ehhe... I guess it\'s the national service where I\'ve got time to think about such stuff (considering I\'ve been given an excuse of uniform + lower limb for 5 days - til the end of my training as I\'ve got a swollen lump on my calves). Well, it\'s initally a scary thing for me to visit my Unit\'s Medical Officer (Doctor) as I tell you - these guys hmmm...are quite something. When some recruits who were sick, visited the medical centre - never reported their name/rank properly to these Medical Officers, they were made to do 20 push-ups on the spot!? I think it\'s stupid and ridiculous; I mean these guys are already SICK :(  Anyway, I slept at 9 pm yesterday as I need to get up at 5:30 am to get ready to get into camp in Yishun (nee soon camp). It\'s like 7 stops via the east plus 12 stops via the North (altogether 19 stops) then I need to take a feeder bus to the outside of the camp then..walk in. Yeah, journey to the NORTH..  Though with my current health condition, I\'ve been blessed. But the Sergeants (which shares the same rank as me) are giving me a kind of look - like i\'m trying to ditch responsiblity by being sick or something Yeah, whatever, i don\'t care anymore.. I am genuinely sick and too BAD. Of course, I don\'t on purpose laze around and help out as much as admin stuff (since i can\'t do streneous) from today onwards til thursday. Friday would be some Unit day where we chill out at SAFRA (Yishun AGAIN). So it\'s more or less the end of this reservist for this working year *phew*  School is starting on 19th July and collecting textbooks on 15th July, can\'t WAIT. Ok, going to upload some pics I took of Jee, Mike and Me at Alley Bar. We were chilling out with her bf\'s friends. Funner!  Muacks  '),(24214,3565927,'03,July,2004','2004-07-03 00:00:00','   urlLink    Christian n Sean at Starbucks&nbsp; urlLink  '),(24215,3565927,'02,July,2004','2004-07-02 00:00:00','   TGIF - yeah, it must be in a lot of people\'s mind. Am sitting at Jee\'s home now and had quite a nice afternoon so far. I actually posted my blog yesterday but guess it got hanged, nothing took place. Well, how should I recap what I\'ve been through this week hmmm...   Monday, met up with Sean and Christian (sean\'s german good friend who is based in Seoul) pop by... We had dinner and coffee. Yeah, it was fun. Felt like ages we\'ve all met - the last time X-mas in Austria 2002 when I visited Sean n his family. Christian thinks whenever me n Sean talk, we are bickering but hehe...that\'s our normal lingual. Christian should be heading to KL with Sean as he has not been there, would be fun for him!   Tuesday, when everything went bersek in Singapore for about 2 hours (negotiable in various areas). Yeah, when something was inside \'something\' (regards to human anatomy), I think the lights went off in the other parts of Singapore lol :P My brother was trying to call me on my mobile but I had it off (as my batt was dying). Tell you, he was calling me but hey, What can i DO?! Errr hehe...  It was good long night. Was showing clif all the pervious pics I took and going down a bit of memory lane (for me).  Wednesday, Chicken rice at Chinatown - I tell you - so cheap and delicious... The tofu is freaking nice. It melts in your mouth literally... The bean sprouts (tao gay) was like super duper nice too.. Aiyoh, we must all go there again. My fav haunt right now :)  Thursday, just staying at home. Nice to be at home and just not do anthing. Thought they were going to show the O.C. but bloody hell got the stupid soccer match?! arrghh  So just watch some VCD on my laptop and just chill by myself. My uncle last minute invited us for seafood at east coast. It was yummy hehe.. Thank God for gym, or else i\'m turning chunky soon lol  Am hungry now siah... Update u later..  Bought quite a nice red top from Zara, hey - Zara got sale from today until the next 3 weeks hehe...  Buy away  B  '),(24216,3565927,'03,August,2004','2004-08-03 00:00:00','   urlLink    Will i grow up tall like a plant?&nbsp; urlLink  '),(24217,3565927,'01,August,2004','2004-08-01 00:00:00','   Can\'t really sleep but am tired really. I was rushing like mad in the gym today in order to beat the gym \'bunnies\' to the bench presses. They\'ll pump iron and then drop the weights with a loud heave. Then, admire themselves in the mirror. I mean, do they have to do that all the time?! Talk about vanity \'lair\' lol I think the gym culture has been seeing much increase in members nowadays. It seems what the advertisments and magazines dictate - it is the ultimate bible for the mass audience. I can\'t seem to be like them, killing myself over 2 hours in there like there\'s no tommorrow. Life is short already. I don\'t intend to \'kill\' myself. The gym is a good place to keep fit - lean. Have realistic bodies. Not stick thin wafer female bods or killer brad pitt bod (in Troy which i refuse to see as the review I\'ve got from everyone is \"Brad pitt has a good bod\"; what sort of a movie review is it!? ) .  Enough of bodies. I just want my tired body to sleep soon. Trying to break the 2am sleep timing but I can\'t seem to do it. It\'s like I\'m becoming a vampire. I wish I could turn into a bat and fly over to various countries; suck people\'s blood (and also other \'things\') hehe.. The only thing that has been sucking my blood is the tonnes of assignments for my upcoming 3 modules for my school. I\'m doing Online Journalism (which has been postponed like twice, no lecturer confirmed and seems we will be slotted on Sunday\'s 9am Arrrgh), Radio and Television Journalism (where i get a young lady lecturer who thinks she\'s funny; speaks with a monotonous tone that puts all to sleep) and finally Media Text (interestingly Murdoch Uni graduate; Cheryl Fox (also Charlotte in SATC) look-a-like chinese 27yr old gal whose having a hard time dealing with students in my school). Well, we had a presentation about representations during Media Text. Communication subjects is the best time to \'smoke\'. Seriously, I\'m blessed and fortunate during presentations. I remembered when I was in my secondary school. I hardly can form or finish a sentence like a snap. Smoked my way through like a breeze today and I thought to myself. Why can\'t I just speed through with the rest of my life like the way how I \'smoked\' in class? *sighs*  Gosh, maybe life is like a cocktail where you need the right concoctions. What is the right ingredients anyway? There is no right stuff. Who tells you that it\'s right or wrong? I\'m sure the blind man in Orchard road would tell you it\'s right to help him. But when I did help him up the stairs (me filled with sympathy and kindess) - he said \"NO\".  He said \"I don\'t need your help, just give me money.\" For a moment, I just stunned. He in fact used his disability to get someone\'s attention and then expects them to pay for the tissues which he was selling. Do you think his actions were low-down or that\'s what people do for survival? When my galfrd was harassed in Bangkok by money boys - telling them how beautiful they were, how they would send them off with flowers and fruits, fondling them topped with sweet nothings - have the world become a place what we think it is?  Not sure if I shared the cab shocking experience which I got recently with you guys. But I took a cab to Sengkang with $12.00 to meet a friend. Turns out when I took the cab back, it was $22.00. Why? Simply, the cab driver refuse to listen to my instructions to take the expressway and my directions which I gave. He took me for a ride around Singapore - I was slightly tired keeping quiet. Then when he stopped me in front of my home (which I\'ve already travelled Up North, Central and finally east) - he ask me to pay up. I look at him with horror and said. Let me ask you. When I gave you directions to take the TPE to take me home from Sengkang, why didn\'t you take it? He replied \"Oooh really, did u say it?\" Well, Yes I did say it. Anyway, thanks for the ride you gave me around Singapore but you deserve $12.00, which I paid the last cab driver. Bye. I walked away, pissed and glad (that I never gave in). Yeah, comfort drivers - you have not seen the last of them...  I\'m not sure what the chicken rice seller (a.k.a. part-time fortune teller) said when I was going to move mid of next year. Maybe I\'m just tired with what is around me. Guys are either taken or cheating. Please, can I just be a plant - at least I\'m asexual; not worried about sex :P   Greens for you during lunch, anyone?  '),(24218,3565927,'01,August,2004','2004-08-01 00:00:00','   A new club, Happy, opened yesterday and so I went. Housing 500 people (as compared to Taboo, the usual haunt for most pink men - holding only 200 men, more smokey and crowded) - it was more spacious and music was not too bad. I enjoy a nice space and place to manuever. You could see the crowd picking up and people cutting in. I have always wondered - how do these people get into the \'list\'? They smile, give each other kisses on their cheeks *muacks muacks* and then say \"Oh you look fabulous\"... In they go. Yes, looks like contact with people are essential to get around these days. It\'s like a monopoly - \'Bail out of Jail\' card.  When you start to observe the crowd. It is as though everyone would give you a look - analysing what you were wearing, how you did your hair, your body shape, etc... Has superficiality reached saturation. I mean you obviously can\'t tell a person\'s personality/intelligence/character just from one look. You have to wait til the other person to open his mouth, chat a while and then you can probably get a clue of that person. How much can you know from a person from an hour, 12 hours, 1 day, 6 months or even 1 year? I mean I never knew what my ex really wanted after 2 years - yeah, that\'s a long time. So in fact, why do people just keep things to themselves? Does communication becomes compromised or reduced when you know more about one another?  I did the brave thing in order to make my (from my last blog) closure with Eric more proper - call him. Yes, I got up at 1430. I called Eric. He picked up (considering I did call him 2 times and he never picked up). Yes, just like the last time (when I first called him before I went to Bangkok) - bad reception. Sometimes I wonder if bad reception means we are not meant to talk hehe... Even when I was in bangkok (using a bangkok SIM card); mind you local to local eh. Also bad reception. Our conversation went like this.  Me : Hey, how are you? It\'s Brendan.  Eric: I\'m good thank you.  Me: Oh just called to see how you were and hope you are doing well  Eric: I\'m okie. Just hungry and going to meet sister bemas for lunch soon He has sms me a lot and i didn\'t reply him.  Me: Did u by chance get my 2 emails? Not sure if you replied as my work email sometimes cannot work as people have emailed me and it bounced.  Eric: Yes I did get your 2 emails thank you. The first one and the second one with the picture link. I saw the pictures. They came out very nicely. And your second email ended asking me I was an email person. Sorry, as I\'ve been busy. No time to reply. Will try reply soon.   Me: Oh i see...  Eric: So what are your plans today?  Me: Errm nothing much. Probably just laze at home.  Eric: I see  Me: I may be heading up to bangkok with another bunch of friends soon, August. Mid august I think.  Eric: Oh okie. You can drop me a line and we could catch up.  Me: Yes. Shouldn\'t be a problem.  Eric: So how\'s work?  Me: Oh well, boss is coming back soon. Got settled with the admin which has been dragged for a while.  Eric: Yeah, read that in your second email. You wrote that.  Me: Anyway, I guess the reception is killing our conversation.  Eric: Yes, let\'s communicate via email or we speak in person,it\'s better this way.  Me: Okie you take care, Ciao  Eric: Thanks for calling and talk soon. Bye   After going through the conversation. Eric has not give me any hope of what may happen with him. But just gave me a nice ending to what I\'ve been through. I misunderstood him twice (the last time was the SMS and this time with regards to the emailing). Eric is just like Mr Big in SATC. He is strange and he does things accordingly to his time. I\'m not sure if I\'m a masochist. But it seems the guys I like are seriously strange. Weird guys. Silent brooding guys are attractive to me haha I am damn sure this guy will forget replying my email. No more expectations. Now, I will learn to love myself more (which have been quite successful so far). Loving myself is the best. If I don\'t lick my own wounds, who would?  Going to Bangkok. My itinerary with Eric would be simply just better the second round. NO, am not doing the sucidal thing. When my mind is tuned to the just *uck mode, I won\'t get emotional. It\'s as simple as that. Well, you think I\'m cold. But I think I need to protect myself more than anything else now.  P.S. Clif (that Qantas Flight Attendant) broke up with his boyfriend yesterday. Yes, that\'s a miracle. Don\'t worry, not interested in him hehehe...  '),(24219,3500159,'31,May,2004','2004-05-31 00:00:00','   Today is Memorial Day and while we attended the requisite party at our good friends Jeff and Kathi\'s house, I had to think about the deeper meaning of Memorial Day. Although I have always known the true meaning, before now it has just been a day for fun, cooking out, and entertaining.  Now I realize that it is about remembering those who have died both in wars and not.  I spent a lot of the day talking to my neighbor about what a brewmiester my cousin was.  Then a new neighbor arrived.  He has to be in his late 80\'s and is a WWII veteran.  He ate with us, talked to the kids and even shared a beer with the guys.  I was in awe of him because he is just a neat old man and he had so much fun just talking and watching the kids swim.  He just lost his wife of many years in January and yet here he was still seeking out friendships with his neighbors who are young enough to be his grandchildren. What a cool guy!  And we did have a nice day just relaxing by the pool, eating and watching our kids play.  It is great to have good friends!  '),(24220,3500159,'30,May,2004','2004-05-30 00:00:00','   Well, it was bound to happen, I guess.  Last summer I was told that my cousin had a \"blog\"....not quite knowing what that was, I wasn\'t sure if I should be concerned or not!  After logging on to his blog and seeing what a really cool thing it is, I thought I might one day try it and never gave it another thought. My cousin was a total computer freak.  He knew everything there was to know.  While I like to think I am pretty good, I had no clue how to do this.....  Life gets to us sometimes, you know?  We all get busy and times just flies by and of course, I never did get around to trying to figure out how to be a \"blogger\".   Now we fast forward to April of this year when this same dear cousin of mine suddenly passed away at age 35.  Now I was forced to stop and think about some things. I realized  that life is just way too short.  And, yes, I had read his blog, but had I called my cousin?  Had I sent a card or an e-mail?  No....and now I can\'t. And yet, even with that kind of wake up call I find myself still being too busy to stay in close contact with those who are the most important to me.  I vow to do better.  Not a day has passed since that April morning when we lost my cousin that I have not cried for him, his little girl, his wife and his parents and brother.  Our family will never be the same again.  But I do know that we must take time to connect with others, even though we are too busy and life is crazy most of the time.  Because, I have sadly learned, when they are gone, it is too late. So, for you Jeremy, I too, will be a blogger!  '),(24221,3500159,'17,June,2004','2004-06-17 00:00:00','   Well, we survived our vacation.  It was cold and rained a lot but still beats the heat of Texas in June.  Spending time with 35 relatives is quite an undertaking....and then there are those who come home with you for another week!    Got a call today that my dear friend lost her baby boy at 22 weeks gestation.  How can that be fair when that baby was wanted so much?   It is hard to understand sometimes.  The summer is passing quickly.....maybe we can get some things accomplished before school starts back up.... Why is it time goes so fast when you need it to go slowly????  '),(24222,3500159,'04,June,2004','2004-06-04 00:00:00','   We are officially on a much needed vacation.  It is so cold here and only going to get colder.  We are in northern Wisconsin and from Fort Worth, Texas, it is a world away.  Riding so far in the car with two teen and pre-teen boys plus a grown man is almost more than any woman should have to bear.  Between the farting and burping, there isn\'t much room for conversation.  It should prove to be a nice week away from the rat race that our lives have become.  My parents will be flying in to join us tomorrow and I really should be getting some sleep now.  I had good luck with scratch cards today, which I really never buy when at home, but on vacation, it seems that anything goes.  I won a total of 145.00 and spent only 22.00.   Not a bad profit.  Kind of makes up for having to sit with kids who burp and fart all day.....  '),(24223,3500159,'02,June,2004','2004-06-02 00:00:00','   Last night the metro area was hit by huge storms and we were without power for about 3 hours.  It got pretty creepy, just me, the kids, the dog and a bunch of candles burning.  They say it is headed our way again right now.  It made me realize that we depend on electricity and our TV\'s, Computers, etc. WAY too much!  My kids and I sat there in the dark and didn\'t quite know what to do......sad, huh?    We are supposed to be heading up to Tulsa tonight, but will have to wait until we see what the weather is going to do.   I hate this...I am ready to get on with my vacation and head north where it is cool!  '),(24224,3500159,'29,July,2004','2004-07-29 00:00:00','   Tonight I met one of my new students.  Yes he has some brain damage, but he understands right and wrong and knows what is expected of him.  He came into my class, rolled around on the floor, screamed at his parents, hit his baby sister, threw cars, etc....This is not the first child I have taught who was like this.  Do you think 4 minutes of time out will do this one any good?  Nope!  It sets him off more.  When my boys were little I never would have put up with rude, disrespectful behavior and yet these days it seems to be the norm.  The kids I teach are special ed, however, we have MANY who are very bright children who know how to manipulate better than any adult.  I am just thankful I don\'t have to LIVE with those kids!  '),(24225,3500159,'27,July,2004','2004-07-27 00:00:00','   Well, I can hardly believe it.&nbsp; Today was my last day of \"freedom\" before returning to work after summer vacation.&nbsp; Our summers in Keller are so short...we are only out about seven weeks and I am not quite finished recovering from last semester yet.&nbsp; Tomorrow it is back to the salt mines as they say.&nbsp; I\'ll have a few weeks to get used to my new kids and new schedules and then my graduate school will begin again!&nbsp; Why am I doing this to myself??&nbsp;&nbsp; I have to work, so that one is a no-brainer, but the grad school thing is just a personal achievement I really want to have.&nbsp; If I didn\'t have to drive all the way to Denton to attend classes it would not be so bad.&nbsp; Oh well, I guess I will just suck it up and do it anyway!   Scott and David are not at all happy either.  They go back next Tuesday...what I would give for one more week......  '),(24226,3500159,'23,July,2004','2004-07-23 00:00:00','   Wow!&nbsp; Today was quite a day!  While trying to be a good mom and take my boys for one last outing to Six Flags before school starts I was involved in the scare of my life!  Scott, David, and Scott\'s friend Jeff, and I had been at Six Flags all day and had a ton of fun even though it was very hot and humid today.&nbsp; We had to leave at 3:30 due to Scott and Jeff being due at a party (that\'s a blog for another time....not quite ready for these high school parties), so we decided to ride just one more ride and then go.&nbsp; Jeff and David opted for bumper cars while Scott and I just wanted one more roller coaster ride.&nbsp; We hopped on the Flashback as Jeff and David went to the bumper cars right next to our ride.&nbsp; The first part of our ride was fine, but then, as we were totally on the top, almost flat on our backs, VERY HIGH IN THE AIR, the thing jolted to a stop!&nbsp; At first I thought it was just going to stop for a second and then go, but as the seconds became a minute, I realized that we were stuck!&nbsp; And as I looked down below, there stood David and Jeff and all I could think about was that I was going to die on that thing and my little son was going to see the whole thing!&nbsp;&nbsp; Bearing in mind it was 101 degrees and I suddenly had to go to the bathroom, I began to panic just a bit.&nbsp; Scott, on the other hand, was having a ball.&nbsp; He said it was the best part of his day.&nbsp; Finally, after about 15 minutes (which is a LONG time when you are stuck on a roller coaster) they got it going again and all I could do was pray that this time when it stopped, we would be on the ground!&nbsp;   Thank you, Lord for getting us off there safely!!!  '),(24227,3500159,'21,July,2004','2004-07-21 00:00:00','   These last few weeks have been crazy.&nbsp; I suppose that time really does go faster when you get older, but it is hard to get anything accomplished when time is zooming by!   School starts back for me next week and the boys the week after that.&nbsp; My niece, Susan moved in with us two weeks ago after accepting a job in Human Resources at the Albertson\'s Division Office here.&nbsp; It has been lovely having her and not as much of an adjustment (at least for us) as I thought it would be.&nbsp; It is nice to not be the only woman in the house!&nbsp; She will be getting her apartment next month but for now I am trying to spoil her a bit.&nbsp;   My graduate courses will resume the end of August and while I dread the drive to Denton, I really want this degree badly.   I have been working out regularly lately.&nbsp; Dixie, Susan and I are attending Jazzercize four to five times a week and working out in the gym on other nights.&nbsp; I have never worked out so much and I hope it helps.   Our family is still reeling in the loss of my cousin and I still haven\'t been able to be much comfort to his parents.&nbsp; Of that, I am ashamed.&nbsp; I wish so much that I could reach out to them, but I am so aware that both my sons are alive and healthy and how is that fair?&nbsp;&nbsp;   On a lighter note, I have been able to spend some much needed time with a dear friend whom I had not seen in a year and that was lovely.&nbsp; My closets are cleaned out, my laundry is done and so I guess I can go back to work.......  '),(24228,3500159,'01,August,2004','2004-08-01 00:00:00','   Night before last I had the strangest dream and it really has me a little freaked out.  I dreamed that the whole family, Linda & Louis included were at this restaurant for someone\'s birthday celebration.  We were talking and suddenly the door opened and in walked Jeremy.   We all just were shocked and stared at him and didn\'t know what to think or how to behave because we all knew he had passed away in April, yet there he was, just like always, healthy as could be.....He smiled at us and said, \"Just joking guys!  I didn\'t really die!\"   And then everyone was angry at him for the horrible joke, but so glad to see him too.....And when I woke up I felt so strange because I can\'t figure out what that all means.  I don\'t know if it means anything at all, but it was so real and yet so strange.   I wish he could come back.......it sure would make so many of us happy again.  '),(24229,3885952,'20,July,2004','2004-07-20 00:00:00','   hi '),(24230,3885952,'29,July,2004','2004-07-29 00:00:00','   and so rages on this eternal war  the likes the world hasn\'t see before  those constant battles unlike any other kind  this internal war, between heart and mind   heart struck first,&nbsp;with blunt emotional force  and attempted to knock mind off it\'s course  but mind&nbsp; held true with it\'s precise logical equations  showing foresight, intelligence, and the utmost patience   heart tried again with unending will  but mind held to it\'s counter strike still  mind struck second, attempting to confuse the heart  but heart was strong, and would not easily be torn apart   heart became desperate and tried to seduce it\'s foe  mind&nbsp; your missing out! there\'s so much good you don\'t know!  mind, reviewing it\'s strategy, brought forth what weapons it required  but the weapons&nbsp; proved useless, without the heart\'s desire   and so both continued to war, waiting for the day when those struggles and those conflicts would all pass away only proving the ever apparent conclusion, be it bitter and or&nbsp;true  that no war or battle can be won, without the unity of the two.   &nbsp;   I wrote this last year sometime. I was trying to explain how your mind says one thing and your heart says another. So not only do you have to deal with the external war of everyday life, you have to deal with the internal warfare of yourself. I hope I conveyed what i was trying to. Hope all is well.  '),(24231,3885952,'28,July,2004','2004-07-28 00:00:00','   take your payroll  and your lottery  thereâ€™s a place I know  thereâ€™s a robbery   and I swear Iâ€™ll never tell  if you swear youâ€™ll never tell  and weâ€™ll all make out so well  we will all make out so well   head on with my hate  into the lights ahead  iâ€™m amazed that Iâ€™m still standing  and I demand that we all blend in   iâ€™m amounted   just the same old Glitter story  from the sea floor  metamorphosis  and I canâ€™t change back for you  i&nbsp;will not change back for you  i&nbsp;must live in skin thatâ€™s new  iâ€™m a livid skin thatâ€™s new   turn your insides out to the outside  turn the outside in to the inside  trade your outside in for the inside  turn it around again   iâ€™m amounted   and I canâ€™t change back for you  no I canâ€™t change back for you   -D. Grohl  '),(24232,3885952,'26,July,2004','2004-07-26 00:00:00','   well folks another monday, another week to start. the weekend went by quite quickly. didn\' t do much out of the norm, just saturday went to a friends house warming party. other then that it was the basic weekend. saw \"I-Robot\" yesterday. would have to say that I liked it. great special effects, good action, cool car. But once again, here I am. My desk needs to be cleaned. I think I need to take my \"APU\" home. It\'s the big machine gun robot they used in the 3rd installment of the matrix movies. I got it at KB Toys for about 10 bucks&nbsp; a couple of months ago and it\'s been here on my desk ever since. I\'ll have to take some pics of my desk and post them on here sometime. That way you can see where I am at when I write on this thing.&nbsp; Well I guess I\'m at a loss of words today, guess it\'s just to early to think.&nbsp; Guess I\'ll put some lyrics from a song I love. Reminds me of some familiar things. Well hope everyone is doing well, be safe. Adios.   sweetness follows   readying to bury your father and your mother what did you think when you lost another I used to wonder why did you bother distanced from one blind to the other  listen here my sister and my brother  what would you care if you lost the other I always wonder why did we bother distanced from one deaf to the other   oh oh but sweetness follows   Itâ€™s these little things they can pull you under  live your life filled with joy and wonder  I always knew this altogether thunder  was lost in our little lives   Itâ€™s these little things they can pull you under  live your life filled with joy and thunder  yeah, yeah, we were altogether  lost in our little lives   oh oh ah oh oh oh&nbsp;oh but sweetness follows  '),(24233,3885952,'23,July,2004','2004-07-23 00:00:00','   well it\'s friday. finally. good to be here. sort of. see life just wasn\'t through with me yet. see when i was driving home lastnight from work, my car started overheating. It has done that in the past and i checked the fluids and all that, and it hasn\'t acted up in a while and has been pretty dormant. well yesterday it decided to wake up. it overheated quite fast, sending the \"check engine\" light on. i was almost home so i decided to just make it home, then I would check it out and let it cool down. i was right up the street when I heard a big \"pop\" sound come from under the hood. looked like something exploded, steam and white smoke coming out of the hood like it was on fire. Have you ever seen Back to the Future part III? At the end when the train is overheating and bolts are flying off, and smoke is coming out everywhere? Yeah just like that. minus the hoverboard. well i made it home, but it was already getting dark,&nbsp;so i had to let it cool before I tried to open it. water and coolant were all over the hood and the bumper, so&nbsp; something exploded. when I finally got to open it, i saw that the upper hose connecting the radiator and the engine itself had completely exploded. it was just too hot for it handle. so ok, not too bad just a hose. well not really. what made it heat up so much that the hose broke? i narrowed it down to either the fan not turning on to cool it, or the thermostat is sticking and not regulating the temperature correctly. hopefully it\'s one of those. well i used kevin\'s car to get to the auto parts store before they closed. they said they had a hose there. they looked, they didnt have it. ok. its 10 at night. no other stores are open. got to wait until the morning. this morning, went to a few auto part stores, they to said they had it, but when they looked, they couldn\'t find it. so after the third store,&nbsp;finally found it. put it on and put some more coolant in the radiator. I had to be into work by 12 today so i made sure I left early just in case. Best Idea of the Day. I didn\'t make it to work. It started doing the same thing, overheating, the check engine light came on, all that good stuff! So i was almost at work, I could see my building! but I had to pull off and let it cool down. Well it didn\'t start up again. So I just locked it up, left it and walked the rest to work. And it is hot here. Not fun to walk to work in jeans&nbsp;at noon in South Texas. I don\'t recommend it. So now, I had it towed to a mechanic that my sister uses, just so he can see what it might be. But you know it\'s just funny. It never ceases to amaze me how life is. It\'s like I never catch a break. I don\'t want to feel sorry for myself, because I\'m sure I\'ve contributed to some of my own problems. But it\'s just amazing, how things can be so crazy. My car breaks down and I lose my&nbsp;job, in the same week! Maybe if I was married my wife would have left me too. A guy here at work said I sound like a bad country song. Murphy\'s Law? Maybe a bit of that yeah. At this point, I just laugh. What else can there be? What else can life throw my way? Like I\'ve always said, if I ever marry and have children (which might not be a good idea the way my life\'s been going)&nbsp;I will have plenty of stories to tell. Hey kids want to hear a story about your dumb dad when he was young? Most people have the saying \"if it\'s not one thing it\'s another.\" I always say \"if it\'s not one thing, it\'s 3 things.\" That seems to be more fitting. I just hope one day i can sit down and not have to worry about too many things. That I will have a little bit of money in the bank, a car that actually works in the driveway, and the woman i love in the kitchen. hahah...ok, ok, in my arms, that\'s better anyways. I think if my mom were here she would laugh, then she would hit me upside the head for being so menso. I think I\'ll get it right eventually, it will all work out someday. But until then stay tuned for more stories of the ridiculous and absurd life of me. I\'m sure life isn\'t done with me yet, I think it\'s just getting warmed up. Adios.  '),(24234,3885952,'22,July,2004','2004-07-22 00:00:00','   Well folks, my days here are numbered. As of yesterday, the big wigs here have decided to shut down the San Antonio site. So what does that mean for me? Well it means that as of the end of the year, December to be exact, I no longer work here. Nobody does actually. They are closing the whole entire site. 550 Employees gone, the building left for some other company to operate in. And you know, the whole situation is bittersweet. Part of me is saddened, I\'ve been here for 6 years and have some good memories. On the other hand, I realize this job has made me crazy, so it will be good to move on and see what else is out there. I think more of older employees are more worried, since they have the families and the mortgage and the bills. Part of me is glad I don\'t have any of that, I think it will help me in making a smoother transition into whatever I do. Now the decision is what to do? I\'ve been here for 6 years, so that means I\'ll get a pretty good severance package. Well actually I already got all the paper work, so it\'s just a matter of filling it all out and sending it in. So with that money, and whatever money I can save by December, will determine what I will do with myself. I actually have quite a bit of options. I could search for another job here, I could look in another city, I could move to Mexico! All these opportunities are quite possible, just depends on what I want and what I choose. So I guess as soon as I figure that out, then I will go from there. I\'m not really nervous, you know this not really a surprise to any of us. I mean we have all been thinking that they were going to make a change here. There were the rumors, the whispers, everyone knew something was up. So to actually have a date, a confirmed day, is actually better. No wondering, no worrying, just a concrete date which to work off of.&nbsp; And I\'ll be happy to close this chapter of my life and move on. There\'s many things I want to do and see, and maybe now with the free time I will have I can do and see them.&nbsp;I\'m not a firm believer in the whole \"things happen for&nbsp;reason\" thing, I don\'t know anyone who really is. But this could be good for me. A time to mature and grow up some more.&nbsp;Be the person I really want to be. If not, then at least I don\'t have to work for a while, and I can sit at home and eat Mexican Doritos. MMMM....Mexican Doritos. Thank Heaven for Mexican Doritos. They have never let me down.&nbsp;   Well I hope all is well in whatever part of the matrix your in. Adios.  '),(24235,3885952,'19,July,2004','2004-07-19 00:00:00','   urlLink   &nbsp;  Yeah I know&nbsp;how it is&nbsp;buddy. &nbsp; urlLink   &nbsp;  '),(24236,3885952,'18,July,2004','2004-07-18 00:00:00','   you know i was just reading a blog of a friend of mine, and wow, it made me think. it\'s not that it was very complex, but i dont know what it was. i wonder why i started this thing. to get my thoughts out? to write rants of past thoughts? for any reason really? I just don\'t know. i have always been the type to write down my thoughts, maybe just so i could&nbsp;get them out of my head. and i thought twice about this. i thought it might make me a little more vulnerable. do you every write down the thoughts in your head? not only that, but would you let just anyone read them? i guess when it comes down to it, i honestly don\'t care. it\'s not that i have anything profound, exciting or important to say, most of it is just ranting and rambling on. i\'ve become a \"human soundbox\" just making noise. hehe. but you know, i\'d&nbsp; have to say that not all noise is bad. and sometimes, when you have been around the noise for so long, and suddenly your not, you actually miss the noise. i had people asking me about those emails that i used to send out. they liked reading them and they wondered why i stopped sending them out. so i guess they were worth something. they meant something to somebody. whether it provide comic relief, a different viewpoint, or just some sort of entertainment, that is ok by me.  &nbsp;  you know the time is really passing by quickly these days. i have a hard time believing i\'m already 24 years old. seems only yesterday that i was a teenager. and&nbsp; you know i\'m not where i want to be. have you ever felt like you were painted in a corner? like no matter how hard you try you just can\'t get out of this rut that you\'ve gotten yourself into? part of me feels like that. the other day i went to the movies, and i saw some friends i hadn\'t seen in&nbsp; few years. and its weird you know. they are both married (two couples) and the one guy is a few years younger then me. him and his wife seem to be doing good, they have a nice car, and i dont know if they live in a house or a&nbsp; apartment. the other couple who are a few years older then me, they are doing really good. they have a really nice house (him and his dad built it) and they have a&nbsp;nice car (probably two or three). we&nbsp; talked for a bit and they asked me how it was going. the same.&nbsp; the brutal honest truth is that. the same. no better, actually probably quite a bit worse. i dont know what\'s happened with me. i just dont know. they are happy (or they seem that way), i mean maybe they are not. but you know, it just never ceases to amaze me how some people just seem to coast carefree through life. you know if you are born into a family with a mother and father, and they make a decent amount of money, its like you are default to have the same. his dad is a engineer, and ofcourse he is one to. how can that be? there\'s has to be some sort of complication there, some sort of fights or disfunction. but it never shows. i just dont know how that can happen. how some people just seem to have it going for them, and then others just have multiple bad things happen to them. like they are truly cursed to this life. i\'m not completely speaking for myself either. i mean there are people out there who are good people but who just have it so bad. how can&nbsp; that be? how? it just blows my mind. i honestly don\'t understand how society can serve up such different predicaments? mine, has been pretty bad, but it\'s by no means the worse. you know i haven\'t been dealt the most ideal hand, but it\'s something that can be&nbsp;improved upon.&nbsp; and if it has not been, then that\'s my fault. i take responsibility for that. but it\'s a little disheartening to see those living it up. they don\'t even know that they are either. it\'s just something the have always known. must be nice. i sound like my dad. i remember him saying that all the time. to me that\'s just an excuse not to try. just to say must be nice, as if your never going to have anything the others have.&nbsp;i know i don\'t need any of that stuff.&nbsp; i know it\'s not important,&nbsp;and in the bigger picture, it means nothing at all. but its still hard to see others having such a great time and you can just watch.&nbsp; \"i\'d hate this place, if it weren\'t for the waves, if it weren\'t for the fact that you love it. where they measure a man&nbsp;on the money he spent, well my heart is not a bank statement.\"&nbsp; well folks, i am a wee bit tired now, and tomorrow starts another week of wonderful work. you know i think i use too many commas. oh well. my blog, my commas.&nbsp;  &nbsp;  hope all is well wherever it is&nbsp;you lay your head down for rest.&nbsp;&nbsp;long&nbsp;live rocco and mateo. goodnight.&nbsp;  '),(24237,3885952,'16,July,2004','2004-07-16 00:00:00','   urlLink    Take it to the Limit!!!!...haahhah..nevermind.&nbsp; urlLink  '),(24238,3885952,'15,July,2004','2004-07-15 00:00:00','   well here i am. but i\'m at home now. work was the same, not too bad actually. suprisingly bearable.&nbsp; pretty uneventful day. this weekend the house is going to be extra quiet. kevin is in belton for the assembly and henry is still gone until tuesday. PARTY!! No not really. Actual this weekend I\'m going to dedicate to cleaning up this dump. I need to finish my room and get the house to&nbsp;a presentable state. I have to mow the front and back yard (if i don\'t no one will) and clean the inside of the house pretty good. kevin already started on some of it, so it wont be too hard.&nbsp; saturday matt said him and lina might come into town, so we might go to dinner later on. that should be nice. I don\'t know with everyone gone at the assembly, I\'m going to have to find another hall to go to. i didn\'t even know it was assembly time already. with work so crazy and me planning for the mexico assembly i didn\'t even pay attention to the time.&nbsp; i remember when i was younger, i counted down the days to the assembly.&nbsp; and it wasnt to pay attention to the sessions. I mean i did, but to go the assembly meant to buy new clothes, stay in hotels, chill at the galleria in houston, and to meet girls. Or should i say &nbsp;spend some quality time with the ones I already knew. I guess all the things the brothers said young boys do during the assemblies, we did. Well we weren\'t that bad. We stayed in our seats, we never roamed the halls during the session. We took notes, we looked up every scripture. We weren\'t as bad as we thought we were. But those times, those were good times.&nbsp; The&nbsp; Astrodome, the hoagies, the swiss miss pudding. Man I miss those days. They seem like they didn\'t even happen in my lifetime. They seem like I watched them on a tv show or something like that. I have to remind myself that I actually  lived  those days. I was such a nerd in those days (as opposed to nowadays). Did you know I actually had a purple suit jacket I wore to the assembly? Yup, dark purple jacket, black pants, and a matching black and purple tie. Can you imagine? A purple suit jacket! I think if i could go back in time, I would beat myself up for wearing that. The girls must have laughed so hard at me. Oh well. they were ugly anyway. hahah.... And now the assemblies are different. Different people, different country, almost different language (ok almost).&nbsp; But they are still great. The one last year was very nice. It was nice and small, and a bit more personal. I enjoyed it very much. I hope this year will be the same. This summer is going by so fast, soon the assemblies will be over and they will be reviewing what learned in them. I actually wouldn\'t mind a little cold air. It\'s&nbsp; been to hot so far, and I\'m already ready for the colder season. Maybe it will snow? Maybe not. Who knows though? Weirder things have happened right? So I\'m not ruling it out just yet. Maybe tomorrow I will.&nbsp; well I figure i\'ll put something i wrote on here. I don\'t mean just ramblings either. I figure I would try to post things that I\'ve written over time. Somethings aren\'t from the happiest of times, some are. So for the first one, I figure I won\'t put something really deep or thoughtful on here. But then again, I wanted to put something I really like. So this one is one of those. I one time read a book on a poet named William Blake. He is famous for many poems, but one that he wrote was called \"The Tyger.\" It happens to be one of my favorite poems. In that book it had some comments and reviews from some critics about the poem. Some went into great detail of what they thought Blake was trying to convey by the poem. But one critic that commented differently. He stated that the poem, is in fact what he called \"pure poetry.\" He stated that if the poem had a deeper meaning, then really only Blake would be the one to know what it was. By it being \"pure poetry\" it might not have any deeper meaning, any special significance, any cultural importance. It might just be thoughts that he had in his head, and he decided to put them down on paper. Those thoughts appealed to me. Usually when I write, I write for a reason, for  something , or  someone .&nbsp; But I found the idea appealing, and realized that I had already written something like that. I had a poem that meant nothing to me. The ironic part is that I had a line in my poem that correlated with Blake\'s poem, look for it if you know his poem.&nbsp;But my poem, it holds no weight with me, no importance, no significance, no importance. It\'s just thoughts, words. I remember when I wrote it. It was when I was 17, I was at the computer late one night. I wasn\'t really thinking about anything important or anyone in particular. But I wrote it.&nbsp; To most it might seem childish, to someone it might not make sense, and others just might thinks it\'s plain stupid.&nbsp;Others might like it, or just say they do because they dont understand it. &nbsp;And you are all right. Because it\'s all those things. It\'s all those things,&nbsp;because I am all those things. I can be stupid, childish, and I can make perfect, and then no sense at all. Like I said, it means nothing to me, and I hope it means nothing for you to. Here it is:  &nbsp;  so here we are at the end  not far from where we started  and although the many we did befriend  we ourselves have parted  &nbsp;  the many that came  the many that left  they took our crown  and left us a frown  &nbsp;  upside down, out of place,  forever lost, without a trace,  fallen down, cut on the knee,  stuck forever in a big tree,  &nbsp;  test drove, but never bought,  thrown perfectly, but never caught,  hunched over, with enormous pain,  a highway with only one lane,  &nbsp;  sought for, but never found,  a bolt of lightning sent to the ground,  a building, built completely of glass,  a raging storm, that soon will pass,  &nbsp;  wind that can move a shopping cart,  a bullet sent straight to the heart,  the hero for the day, once again,  a lazy day spent in the den.  &nbsp;  cuddling the one you love, next to the fire,  one these wings, i dare aspire,  alone, besides the company of my cat,  on my thrown is where i sat,  &nbsp;  here once more, but not the same,  tired, tired of this ongoing game,  on faith and love it does depend,  and this last line is the end.  &nbsp;  &nbsp;  hope everyone is doing well wherever they may be. have a good weekend and be careful in whatever you might be doing. Adios!&nbsp;  '),(24239,3885952,'13,July,2004','2004-07-13 00:00:00','   another day here at the job. you know i think I sit in this chair more then I sleep in my bed. Today I just don\'t care. I\'ve gotten a few calls so far, and they haven\'t been pleasant. I find myself not caring if their password is revoked, if they can\'t use their VPN, I just don\'t really care. I spoke with a lady today and she just made me sick to my stomach. Everything I asked her to do didn\'t work, all she kept saying was \"no its not working\" I wanted to reach through the one and strangle her. These user\'s are telling me, \"hold on\", \"let me try this.\" NOOOO!! Don\'t TRY THAT! DON\'T TRY ANYTHING! JUST DO WHAT I TELL YOU TO DO AND SHUT YOUR BIG MOUTH! Man this job is just driving me nuts. I\'m slowly losing my mind! I think I\'ll be living up the street at the state hospital soon. I don\'t know I enjoy this job, but it\'s slowy killing me! hahah..I make it sound like I\'m out in the sun picking lettuce or something. I\'m sitting in a air conditioned room playing on a computer, big hard work! Well I guess I just needed to vent. These users are impossible, they are beyond stupid most of the time. And because of their ignorance and their pride, they are quick to lay blame. And who you might say do they blame? They bite the hand that feeds them. The very person that helps them is the one that they blame for their problems! We are here to help, and all we get is disdain and contempt. I swear it! Ahhh....this is driving me crazy. I despise this place!  There are only few people who can sum up how I feel about this, and it\'s Agent Smith. Much like neo in the matrix, I\'m stuck in a cube, which is on a square floor in a rectangle building on a geometrical grid! This is how I feel about this place:     Agent Smith: Can you hear me, Morpheus? I\'m going to be honest with you. I hate this place, this zoo, this prison, this reality, whatever you want to call it. I can\'t stand it any longer. It\'s the smell. If there is such a thing. I feel.. saturated by it. I can taste your stink. And every time I do I feel I have somehow been infected by it, it\'s repulsive. I must get out of here. I must get free.     Nothing sums it up quite like that. I need to just do something else. I just don\'t know what. Well I\'ll live, but it\'s not the way I want. Hope every one else is doing better. Adios.  '),(24240,3885952,'09,July,2004','2004-07-09 00:00:00','   Ok, so this is my first post. Am I excited? Not really. Ha Ha. Well I\'m here at work, been here since noon, and will be here until 9 pm. Joy. Today has been a weird day. Got some new shoes. A bumper for my car. A nosebleed. I think I have cancer and I\'m going to die. Not today, it\'s Friday. I don\'t know what I\'m doing tomorrow. Oh yeah, I\'m taking Henry to the airport. He\'s going to California for like 11 days. The house will be quiet. As if it is ever loud. Just me and Lucci. And the occasional Kevin. I really need to clean my room. It\'s been ages since I have, and I don\'t even know what the floor looks like any more. Just Kidding. Maybe. I have a headache now. I\'ve never had a bloody nose before, it kinda freaked me out. Blood isn\'t supposed to just come flying out of your nose like that. Unless I got punched and didn\'t know, I don\'t think I did? I\'ve been stuck on this one song lately.  \"Asthenia\"  by Blink. Has anyone heard it? You know I thought that was a weird word. The whole time I thought it was some girls name or something. But it actually means feeling very drained and tired. Having no body strength. Hmm...Interesting. I wonder why they picked that word. Maybe they were just having fun with the thesaurus. The song is good though, very thought provoking. For those of you who dissect song lyrics like I do, you might find the words applicable to a part of your life. I did. Well all, I can\'t think of anything else to say and my head is pounding. Ouch. Hope all is well with all of you,wherever you might be in this world. Adios.  '),(24241,3885952,'16,August,2004','2004-08-16 00:00:00','   urlLink     I don\'t feel the suns comin\' out today it\'s staying in its gonna find another way As I sit here in this misery I don\'t think I\'ll ever see the sun from here And oh as I fade away they\'ll all look at me and say, and they\'ll say Hey look at him! I\'ll never live that way But that\'s okay  they\'re just afraid to change  When you feel your life ain\'t worth living you\'ve got to stand up and take a look around you then a look way up to the sky And when your deepest thoughts are broken keep on dreaming boy, cause when you stop dreamin\' it\'s time to die And as we all play parts of tomorrow some ways will work and other ways we\'ll play But I know we all can\'t stay here forever  so I want to write my words on the face of today and then they\'ll paint it  And oh as I fade away they\'ll all look at me and they\'ll say Hey look at him and where he is these days  When life is hard, you have to change.  -S. Hoon  urlLink  '),(24242,3885952,'10,August,2004','2004-08-10 00:00:00','   i look at you and feel the same yes, my heart is the one who takes the blame  i look at you and feel no shame because my thoughts you so effortlessly tame  how many times i\'ve heard you say true love is not a game in which we play.  '),(24243,3885952,'09,August,2004','2004-08-09 00:00:00','   well im at work, just sitting here waiting for 9 to come so I can leave. this weekend was pretty nice, did a lot of flea market shopping. i really didn\'t buy much, just a little clock for my desk. as if i needed more things here for my desk.  i\'ve been listening to a lot of spanish guitar music. just the guitar, no words, no accompaniment from the drums or any other instrument, just the guitar by itself. It\'s very beautiful, i can\'t seem to get enough of it. It almost seems haunting in a way, just the sound of the strings. i wish i could play like that. when i go to mexico in a two weeks, I think I might look for a classical guitar to purchase. i figure i\'ve always wanted one, so why not? There\'s a guy here at work, who growing up he would go to a youth place after school. I guess it was a program or some sort of place to keep kids off the street. Well either way he said there was a teacher who taught him how to play the classical guitar. He works here, but he does side jobs with some mariachis. he plays guitar, trumpet, bass, and is working on the accordion. I think that\'s so awesome. how someone can be so talented like that. I wish i would have made that decision when i was younger, to learn to play. well I did make that decision. When i was in 5th grade we had music glass and we would all have instruments. Moving on to middle school, I had decided that I wanted to be in band. So I signed up for it, and when the year came around, i didn\'t have band on my list of classes. I checked with the counselors, but they had screwed up and put me in choir. maybe they thought i would make a better singer. well anyways, i tried to get it changed, but by that time it was too late and the band class had already begun and starting figuring out what instruments were for whom. so I missed out on it. I don\'t know which instrument I would have played, part of me says drums, but then part of me (even back then) wanted to play the Cello. Which actually would have put me in Orchestra not band. So who know\'s I could have been a great Cello player, and maybe learned guitar through that. I don\'t know, it\'s to late now to worry about it. I guess I\'ll just have to buy my mexican guitar and hope for the best. Maybe someday I\'ll write and play a song. I surely have enough lyrics. Well hope all is well with everyone, take care all. Adios.  '),(24244,3885952,'04,August,2004','2004-08-04 00:00:00','   well its late. and im here at the computer. i think im going to go blind from the monitor. or atleast get some sort of cancer. today was uneventful. i ate meatloaf. that\'s how uneventful. my boss is on vacation. tuesdays aren\'t known for their excitement, so i\'m not too disappointed. i have music blaring over the headphones. everlong. i remember the first time i heard it. i was amazed, i\'m sure most people were. the best rock song ever. ever. my blog, so i said so. too bad, if you don\'t like it, stop reading and go home. and if your already home, then go to work, cause your going to be late. and your boss will be mad. mine won\'t, cause she\'s on vacation. i\'m going on vacation soon. i really like vacations. beats work. like seriously beats work. im not even kidding. do you even think i am a kidder? i spoke to a woman today named margo. no joke. no kidding. i thought about superman. i used to dress up like him. when i was a kid. then when i was 16, to impress my girlfriend alison, i bought and wore a superman shirt to a party. she liked it. she called me the man of steel. weird, i went to CLARK high school while dating her. oh well. ironic. too many dots to connect. thats where i got the nick name soup. well it came from the SUP part of superman. i had a friend who called me that in hs. they thought i meant the food. so thats how i got the handle Soup. i had a old email called  urlLink soupystuff@hotmail.com , which doesn\'t work so dont click on that hyper link. i think i tried  urlLink soup@blablabla  but that one was taken so i came up with soupystuff. suprise surprise no one had taken that one. then when i went over to yahoo mail, i had to come up with a new one. so soupycannon was born. soupy was the the original, cannon came from Cannon Beach which is a beautiful place in Oregon that i love. I\'ve also used wasoupy which the WA coming from the state code for Washington State. Souperman was another one i used, but too many thought of that. i am a man of many handles. this is funny. im talking about nothing and making perfect sense. well in my head atleast. \"i know the lines crooked on paper, but i swear i got \'em straight in my head.\" hehe. well i\'m running low on thoughts (for once) so i\'ll just make this look long by putting some funny quotes i read the other day:  \"Right now I\'m having amnesia and deja vu at the same time. I think I\'ve forgotten this before.\"  \"I wrote a song, but I can\'t read music so I don\'t know what it is. Every once in a while I\'ll be listening to the radio and I say, \"I think I might have written that.\"  \"If you can\'t hear me, it\'s because I\'m in parentheses.\"  \"Everywhere is walking distance if you have the time.\"  \"I went to a restaurant that serves \"breakfast at any time.\" So I ordered French Toast during the Renaissance.\"  \"I went to a general store. They wouldn\'t let me buy anything specifically.\"  \"I hate it when my foot falls asleep during the day because that means it\'s going to be up all night.\"  \"I was trying to daydream, but my mind kept wandering.\"  \"I spilled spot remover on my dog. He\'s gone now.\"  \"When I was a baby, I kept a diary. Recently, I was rereading it. It said, \"Day 1 -- Still tired from the move. Day 2 --Everybody talks to me like I\'m an idiot.\"  \"My grandfather invented Cliff\'s Notes. It all started back in1912... Well, to make a long story short ...\"  hehe...ok hope all is well with all of you. Adios.  '),(24245,3885952,'02,August,2004','2004-08-02 00:00:00','   well another eventful weekend. too much to fit in one post. ok well maybe not. went to the flea market, ate some mexican food, looked at pool tables, walked the mall, went to the drive in, drank beer, ate kfc, ate ice cream, ate popcorn, ate pizza, drank pepsi. ok so i eat a lot. that\'s been established now. but you know this weekend had a lot of time devoted to thought. even when i was trying to concentrate on the movies, i was in my own thoughts. contemplating my future, and just what it might bring. i have two choices. one the safer choice and the choice the reasonable person would pick. the other the more daring, risky choice, the one you make when haven\'t thought all the things through. And I don\'t know which one is the better choice. thought or instinct? sometimes i think if i make the thoughtful, seemingly smarter choice i will be less happy. because my happiness would be skyrocketed with the riskier choice. so this is my dilemma of the moment. for the next 6 months to be exact. which direction will i decide my life will go in? only 24 and I\'m making decisions that will effect my life in a big way. you\'d think I would be used to this by now. guess one never does get used to those things. i ponder on what decision my mom would approve of more if she was here. but then again, i wonder if she would have taken a step back and let me decide and make my own decisions to learn from them. i wonder that a lot. some people say that if my mom was here she wouldn\'t approve of some of decisions that have been made, not only with me, with my sister and brother. but honestly, i think that we don\'t really know. maybe those who knew her can say that, because her traits and her way of thinking became familiar to them. but is that really so? i know sometimes i make complete opposite decisions of what people think i\'m going to. keeps them guessing i suppose. so i don\'t like to think that she would necessarily disapprove of certain things her children have done or are going to do. i think she would just trust that we knew what we were doing. and that we would be able to handle whatever circumstances we received by our decisions. she would be there to lend a hand, i know that much. she would be there to bop me on the head whenever i need to be realigned. that i have no doubt. but i think she would let us live our lives and love the people we want to love. maybe i\'m wrong, but then again, maybe i\'m not. if i have the chance, I\'ll have to ask her. But that is sometime in the future, and for now I still have the decision. with my job ending, my world changes. a new year will begin, and it might not find me here. it might find me somewhere else. where is that? i think that is still to be determined. if it is at all. it will be closure. as if i needed any more. i think this life has taught me about new beginnings and endings. i seem to quote the matrix a lot on here, but for a lack of better words, its true. \"everything that has a beginning, has a end. \" atleast in this system that is. a chapter of my life will end and new one will begin, whether i like it or not. sometimes life doesn\'t ask what you want, or what\'s more convenient to you. it decides for you. fate? no. destiny? no. just circumstance, just daily life. and that\'s how it goes. that\'s how it always goes. i truly hate that saying. with all my heart i despise that saying. why? why does it always have to go that way? that seems such a cop out. that\'s how it goes. why?! why must it always go? isn\'t there times when it doesn\'t go at all?? i remember my dad, he would always say \"must be nice.\" and I catch myself saying that from time to time. and i hate it. must be nice. in other words, im not going to try, im just going to sigh and not doing anything about it. if there is one thing I\'ve learned from my dad, is to try. and that\'s because he didn\'t. this is my opinion of course. i know some family members have theirs, and I have mine. and from what i saw from him, he didn\'t. it\'s not about winning, in being a good father, in being a good example, it\'s not about coming in first and being the best. it\'s about trying. i never expected him to be the best dad. i knew he had his problems. but all he had to do was try. he did sometimes, but overall he felt sorry for himself and gave up. that\'s the person i don\'t want to be. i see parts of that in me, it\'s only obvious that some parts would show up. but i try to combat it with other examples that I have. i sometimes wonder if the average person has all the thoughts i have. i mean is that normal? does the average person have to fight himself on a daily basis? and I realize, that yes they do. probably even more so. like I\'ve always said, my life has been pretty hard. the hardest? no, not at all. but it is hard enough for me. i\'d like to think Jehovah has let my life be the way it is, because he knows i can handle it. anymore would be too much. he allows only what he knows you can bear. and if we do have more then we can bear, it\'s not because of him. it\'s because we messed up somewhere along the line and added to the burden. i guess where I\'m at, i want to start taking away from it. over the last 3 years i have seemed to add more then ever. and i am ready to get back to my little burden, the one i naively thought was world ending back then. the one i miss now. but that beginning has ended. now with this new one i have a chance to do what i want, what I have always wanted to do. be the person i want to be. and continue to love the one that i do. i guess some things won\'t change. but until then i will still be eating, drinking pepsi and beer, concentrating on the movie and trying not to think.  hope all is well wherever you may be. adios.  '),(24246,3309500,'11,May,2004','2004-05-11 00:00:00','    First Post! Woot!  Okay, this is my blog of me.  And my not very interesting things.  So if you care read, if not, no biggie.   '),(24247,3309500,'30,July,2004','2004-07-30 00:00:00','    Yeah, my eye still hurts a bit. Not as bad. I may have to go on Steriod drops (look at me bulk up, yeah) but we\'ll see tomorrow.   Check out Condon\'s blog for a pretty funny Clerks - The Animated Series rip-of^h^h^h^h^hspoof*.  Possibility of a&nbsp; urlLink Showgirls  Drinking Game night tomorrow.&nbsp; Many thanks do to the  urlLink Showgirls V.I.P. Edition  for making it all possible.&nbsp;&nbsp;&nbsp;  &nbsp;   *It\'s a geek thing.&nbsp;&nbsp; ASCII - ^H - Binary- 0000&nbsp;1000 Name/Meaning - Backspace.&nbsp; See  urlLink here  for more info.   '),(24248,3309500,'29,July,2004','2004-07-29 00:00:00','    So last night my eye was hurting so I took out my contact for a change (No, it\'s okay, really, I have month-long day/night wear contacts now) and it felt like something was in my eye.  I tried washing it out with saline no luck, I said screw it I\'m going to bed.  I tried to wash it out this morning again with no luck.  Lace convinced me to go to the doctor.  His diagnosis?  I have something wrong with my eye.  He thinks something got trapped under my contact at one point and possibly scratched the eye.  I\'m on Ciprofloxacin - 1 drop every 1/2 hour today and I get to see him again tomorrow morning.  And you didn\'t think this blog would be interesting :)   '),(24249,3309500,'28,July,2004','2004-07-28 00:00:00','    I get dibs on ranting about the new attempt at the most horrible movie that I could possibly think of making.&nbsp;   urlLink http://apnews.myway.com//article/20040728/D843S6E80.html    Seriously, put  urlLink Marci X  on loop before subjecting me to more of that crap.   Yet I am still expecting a good rant from Condon on this one as well.&nbsp; In my opinion he is fundamental in getting the original series canceled.   '),(24250,3309500,'28,July,2004','2004-07-28 00:00:00','    Yes, so I\'m going to try this whole blogging thing again.&nbsp; The last time didn\'t go so well, so ignore them.&nbsp; Thanks to Condon for peaking my intrest in blogging again.&nbsp; See his blog at  urlLink http://youreablog.blogspot.com .&nbsp;   Okay, that\'s all for now.&nbsp; Hopefully i\'ll be back later.   Yeah, and if you have an RSS reader that can read ATOM feeds:   urlLink http://sib57.blogspot.com/atom.xml .   '),(24251,3309500,'06,August,2004','2004-08-06 00:00:00','    Okay, I may be breaking a rule about blogging more than once a day, but you know what?  It\'s my blog, I\'ll do what I damn please :).   urlLink http://www.ctrlaltdel-online.com/?t=archives&amp;date=2003-05-30   Funniset thing I\'ve read in a while.  Ctrl-alt-del is a VERY funny web comic.  Speaking of web comics, I read quite a few.  I\'ll post about that Monday :).   '),(24252,3309500,'05,August,2004','2004-08-05 00:00:00','    Ever have a question as to what something is and you just don\'t know where to look? There\'s a good chance  urlLink wikipedia.org  will have it. They have a goal, and that is to have, in database form, the entire sum of human knowledge. It\'s an open source encyclopedia that is constantly evolving, changing and adding to. It\'s really neat.  The other day I heard the term of a series \"Jumping the Shark\" I\'ve heard this before, but never knew where it came from. This may be common knowledge to some, but it wasn\'t to me, but thanks to wikipedia, I now know. In case you\'re curious  urlLink http://en.wikipedia.org/wiki/Jumped_the_shark . It\'s really a great site.   '),(24253,3309500,'05,August,2004','2004-08-05 00:00:00','    Okay, today I\'m going to rant about Windows 2003 and Microsoft Visual Interdev. Please feel free not to read today.  I copied my new application onto a Microsoft Windows 2003 webserver that we have up and running. It didn\'t work right away so I had to go investigate as to why. Seems as though W3k3 wasn\'t processing a Server Side Include (SSI.) A little bit of research ... AHH W2k3 ships locked down by default, SSIs aren\'t processed. No biggie, I turn it on, still doesn\'t work. AN HOUR I spent searching googol to figure out what was going on. Finally figured it out, and it was nothing with the setup, it was setup fine. See, when VI copies a web application to a new server, it SOMETIMES renames default.asp to default.htm. WELL, for security reasons, .htm files aren\'t processed as scripts, and they shouldn\'t be. So the webserver didn\'t know I had an include file because it was just happy to serve up the .htm file with no problems.  Okay, I knew that was going to be a rant that no one was going to care about.  Ummm, well ... the Phillies won yesterday. Go Phils. Yeah, that is all.   '),(24254,3309500,'04,August,2004','2004-08-04 00:00:00','    I\'m not a smart man. At least that\'s what I keep telling myself today as I am re-writing a lot of code I worked on the past couple of days. See, specs changed, and I SHOULD have seen the change coming but didn\'t now, I have more work to do. I hate deadline projects (yeah, so the deadline isn\'t for another 5-6 weeks, but still, I\'s looming above me like the grim reaper. And I\'m not talking about the cool \"Death\" from  urlLink Family Guy  who goes to frat parties and tries to pick up chicks. It\'s not even the cool \"Grim Reaper Dude\" from  urlLink Bill &amp; Ted\'s Bogus Journey . What was I talking about again?)  Very fun night last night watching  urlLink Hellboy . Maybe it\'s just me, but I\'ve been noticing a trend recently between rising Nazi conspiracies in movies, books and entertainment in general. A book I read a while ago  urlLink The Day After Tomorrow  had it (P.S. this book has NOTHING to do with the  urlLink movie  of the same name. The movie was better.) Well maybe it was just those two things that happened within a week or two of each other that made me think about it and there isn\'t anything going on, but it sure did seem like it for a while. I wonder if the entertainment industry is running out of ideas. During the Cold War it was easy to make the Ruskies the bad guy, now who? I haven\'t seen it yet (but do own it :)) but  urlLink The Sum of All Fears  was supposedly had a terroristic threat, but maybe that type of thing still hits too close to home to be good for entertainment purposes. Maybe it\'s just easy to make The Nazis the bad guys. One look at the swastika and all in the audience knows that they\'re bad.  Enough rambling for today. Bye Bye Jann, have a good time in Hawaii.   '),(24255,3309500,'03,August,2004','2004-08-03 00:00:00','    Lacey drew a very pretty picture yesterday and asked me to put it on my blog. Since she is my wife, and it IS so pretty, I couldn\'t refuse. Check out here picture  urlLink HERE .  So I\'m kinda pissed right now. Looks like I may have forgotten to renew my domain name, christophersmith.us. I may not have any recourse to get it back either, which really sucks.  On a much better note, however, I\'m listening to an excellent box set.  urlLink Jimmy Buffett: Boats, Beaches, Bars &amp; Ballads  It\'s really good.   '),(24256,3309500,'02,August,2004','2004-08-02 00:00:00','    Even though a good 1/3 of my readership is on vacation ( urlLink Condon\'s Blog ) I\'ll write an update anyways.   urlLink Showgirls  night was a success. At one point we had ~4-5 people as coming. By the end of the night we had at least 14 different people there. And all kinds of randomness as well, which was super-cool. Kyle Williams, Tim Morea, my sister her boyfriend and his brother, Ro, Kristen\'s friend. Oh, the movie sucked, though. It is the best box set ever created, so it has that going for itself.  The  urlLink C.S.I. Dark Motives  computer game is pretty fun. Follow the evidence. So far Lace and I have completed two missions and we got \"Master\" ranking for both. Note to anyone who buys or acquires through some other means: There are bugs in the game. Visit the website and download the latest patch. We had trouble with the second mission because of the bug that prevented our magnifying glass from working correctly.  That\'s all for now. If I come up with anything especially interesting, I\'ll be sure to come back.   '),(24257,3517315,'11,June,2004','2004-06-11 00:00:00','   Okay... If i wrote about my situations w/my ex, i would have a huge fucking post, so I\'ve decided not to write about boys at all. It\'s been FUCKIN\' MADNESS!!!!   So.... um.....  uh.... let\'s see... It\'s FRIDAY!  Yay!!!!  This weekend is the Apalachin Fireman\'s Field Days.  Tonight i\'ll be taking my daughter to watch the fireworks.  Tomorrow i\'ll be taking her to my aunt\'s to watch the parade go up the street and go to the Field Days to ride rides and stuff.  That\'s exciting.   Yesterday i went to one of those Adult Bookstores all by myself!  How creepy.  It was like me and 10 hairy greasy icky men.  I bought a lovely blue \"personal massager\" w/a dolphin on it to give my clitoris the special attention that it deserves.  I took it up to the guy to pay for it and he was like \"I have to test it\", so of course, i giggled b/c i thought he was joking.  It turns out he was serious... i guess they have a \"no refund, no return\" policy (which makes sense), so he had to take it out of the box and put batteries in it and make sure it worked AND IT DIDN\'T!!!  So i had to go get another one!!!  It was very awkward, but the guy was really cool about it.    Other than that, i guess i have no news. Ta ta.  '),(24258,3517315,'01,June,2004','2004-06-01 00:00:00','   1st blog ever. I\'m sure that by tomorrow I\'ll forget I even started this whole blog thing... Maybe not tho\'.  We\'ll see.    Oh... what to talk about???  Umm... I dumped my boyfriend this weekend.  Now I have to move all my shit, including a piano, couch, entertainment center, etc., out of his house.  That sucks.  He\'s a cool guy, but unfortunately, cannot be honest.  He has a serious lying problem.  I have given him  plenty  of chances and this weekend was the last straw.  Hopefully this will be a wake up call and he won\'t fuck up future relationships the same way.  Good luck to him.   Now I am at work.  I am a Customer Relations rep for a utility company.  Everyone that calls hates me.  Makes for a long day.   I had a horrible dream last night that the world was ending due to global warming and I had no choice but to die a slow and painful death.  Thank you \"Day After Tomorrow\" previews!  It was a great way to start a great day. :)    The upside is that today is TUESDAY!!!  Only 3 and a 1/2 more shitty days left \'til the weekend. YAY!!!!!  Ta ta.  '),(24259,3647581,'20,June,2004','2004-06-20 00:00:00','   Well haven\'t done anything for a while but play my guitar and eat and sleep, its been pleasant but it is getting lonely. So you ever just wish you had that special someone right by your side but it feels like their a million miles away? It sucks even more when you know they could be right in front of your face I wish I could see heather more. Is describe the events of this weekend but I don\'t have the drive or motivation not to mention it was extremely uneventful. Well that\'s all for now. '),(24260,3647581,'17,June,2004','2004-06-17 00:00:00','   Today was eventful I played my guitar for the latter of it. As for the rest I helped my aunt move out yes she\'s moving to Dallas, whish sucks on two levels 1.)now I cant stay over and sneak out to see heather 2.)im actually going to miss that crazy girl. Well I am really committing my self to this musical aspiration of mine and im infatuated, I also cant stop thinking about how awesome today I went, I saw heather last night and at school she just made me laugh over and over she makes me ridiculously happy. I don\'t know what id do if I didn\'t have such wonderful person in my life. Im happy content and sleepy... '),(24261,3647581,'16,June,2004','2004-06-16 00:00:00','   In my previous blog I spoke of \"josh\" as a threat and it was by far a mistake... the biggest yet. I trust I wont be mislead and I just thought too far into a friendship but through thorough examinations of myself and chatting with longtime friends I\'ve discovered everyone needs a buddy. I\'ve got mine(mostly my journal and brain) and that\'s all they are so im sure that\'s all heather and josh are. Anyway all you indie rock fans out there on June 22 go to www.team-love.com for buttons stickers t-shirts and other sweet goodies from the tilly girls and boys(tilly and the wall). Check out this band world they are coming up them and willy mason hes on the TL site too. So long for now. '),(24262,3647581,'15,June,2004','2004-06-15 00:00:00','   Hello everyone on the information super highway, this is my first blog so bare with me its going to be full of grammatical errors. Well let me start by saying today was eventful...It started last night I slept till about 2:00 a.m  then off I went to go see my \"girlfriend\"(the quotation marks are there because I think the title is cliche, and needs a more proper name) it was hectic a 30 minute walk to the edge of the other side of town where I risked getting beheaded by Heather\'s(my girlfriends name is Heather more on her later) soulless vicious mother or arrested the two are both equally terrifying! I made it wow it was a step and a drudge a step and a drudge the whole way, i first had to get past the hounds now and into the dim room which nestled a heavenly body of which i was more than willing to hold. Anyway in and out by 6:00 we spent the night listening to Bright Eyes first full length effort \"Fevers and Mirrors\" it set the mood for a night full of all sorts of marvelous things. Needless to say neither of us had a full nights rest knowing all to well we had to wake up in about an hour to attend prison i mean summer school but hey we had fun and cherished every moment. I headed back to my aunts as soon as possible where i arrived at 6:32 exactly and slept till 8:00. Now time for school which was fun and full of all sorts of little twists and turns of which I have no passion of typing every single event. So it was all gravy a good day today heather, school, and an evening of boxing my friends for fun and getting the shit knocked out of me. Well that was today in short... But something still troubles me and my mind will not rest until I have someone I can complain to all about how im too jealous and worry about heather spending more time with her friend josh than me. Of course she\'s grounded from me and anything rob so I guess she has no choice I guess I just wish I made her as happy as he does or at least was there for her but Satan#2 makes that impossible(cute nickname for a older mom type huh?). So I guess ill just call up an old buddy to help me through these times of adolescent wonder, and jealousy. By the way nonantiestablishmentarianism(I think that\'s the way its spelled, shit I hope that\'s even the right word) is the longest word in the dictionary...Thanks Johnny what would I do without a friend who can debate politics and chug a beer in the same sit in. '),(24263,4139489,'06,August,2004','2004-08-06 00:00:00','   Hmm, This was an interesting one I\'ll say. It started off pretty weird, I was actually in a game of Counter-Strike. I haven\'t played this game for almost a year, so for it to be in my dreams shows you it\'s effect since 1999 :). It was in cs_militia, a deserty type map. Now it was the back of the militia map but was mostly in Irvine, near Tustin Market Place. Between my old neighborhood and the TMP, was a barren deserty field. Now the house from the map was there, but everything else wasn\'t. What the house looks like in militia...    So there was a long street, coming from TMP to a little buissness center type place, where they were putting up buildings, and to the left was the house from militia. This is what it looked like...    Now I had a team, but they were fuck tards and they couldn\'t do shit, go figure. I was using the SIG552. This is what it looks like...    So I went around back and just killed all the T\'s, and my team didn\'t do shit, so I was kinda pissed off. Now after a little while of that they leave, and walk down the street. Now down the street there is Jessica Simpson... yes, Jessica Simpson... dead though. She looked like she was standing up and then just fell backwards onto one of those green electrical box things. She was wearing some white gown/dress thing. My team was just rushed to her and they were talking and some construction workers went over there also to see what was going on. I went back to my spawn (the other end of the street) and some two kids were walking the other way (towards Jessica). They looked at me funny so I walked up to them and they talked trash, and as I\'m about to smack this kid upside the head with the butt of my rifle, an old pick-up truck comes along and stops infront of me. So I look in and there is some guy in there who is now defending these kids. He gets out and I walk backwards some and as I\'m doing that my brother riding a bike, randomly comes and goes between us. Just then I pulled out a pocket knife and stabbed the guy in his chest, and said something along the line of \"You piece of shit cock sucking faggot, don\'t try to pull that shit\", we then pushed him to the ground and beat the shit out of him (was quite fun) and then Ellen Cleghorne (from SNL)    who I guess is a cop, in a red Chevy Blazer drives up and looks as we are killing this guy, and drives away. that was pretty much it. Very weird one, I must say. Also pretty short. '),(24264,4139489,'04,August,2004','2004-08-04 00:00:00','   Well the dream I had last night wasn\'t that bad at all really. It started off as me and my family doing something, thus we were at some sort of hotel or somewhere. I remember going from room to room with my brother, and in one room there was some porn video on a shelf in some wall unit. I waited for my brother to leave, but he was in the back of the room or something. Eventually I came back and watched the video alone, it was alright, but there was too much scene acting so I fast forwarded it to a random spot and it showed some girl getting fucked on her back, so I rubbed one out to it. Then I left and then I forget what happened. Later I remember that I was at IHS and it was sunny out, and I went into a room and it was all red, from a red light I assume. I walked over to one side of the room and there was a little tub/spa thing and abunch of people around eating and talking. I took off my shirt and pants (still in my boxers though) and got in, but the water didn\'t feel quite right, so I got out and picked up my clothes and went over past a divider in the room. There was no red light here, but it was pretty cool temperature wise and seemed alright. There was a lower level to the room, and down low there was a floor with about a foot or so of water, and two more of the spa things, but they were bigger. So I got in one, there was nobody else in it so I was kind of relaxing at this point. Then I looked over and saw the other one had two girls in it, and they were alone. So I got out and went into their spa thing and talked with one of them. The other girl was good looking with somewhat tan skin, light tan though, and brownish-blonde hair. She just looked to the side and was mellow. The other one was Chinese, and had hair like this,    but her hair was shorter and she was really cute.  Here is what that room looks like,    So I was talking with her and then she got up from sitting down and kinda got on her knees with her toes straight into the ground (kinda hard to explain I suppose) and she wanted to kiss me. She had almost perfect lips in my book. I moved in closer and she opened her mouth and then I started to French her, and suck on her tongue and she got a kick out of this. So I moved my hand around between her legs, and stuck a finger between her bikinii or whatever it was and she rocked back and forth on it while we were kissing. It was pretty nice. Then we stopped for some reason. Talked a little more, then kissed for a little bit longer. Then I guess a bell rang or something and we all got out and i walked with her to another room. This room was pretty bright and alot of people were there, I remember I was between a copy machine or something and a wall talking to her about Cantonese (I had no shirt on and I was holding my pants up, still had boxers though) and a teacher from IHS was with a teacher from my current school, and the one from IHS said that I had to go to some room tomorrow to make up something. This is what this room looked like,    She went out the door behind her and I went outside and everyone was going from one class to another, and I was walking to this room that the teacher had told me to go to tomorrow. Then some other teacher said that I don\'t even have to do it, so I went back to the first room with the spa things. As I was walking in there were a hell of a lot more people. One of them from behind was a lamer from my middle school and he was yapping about CS, saying how we used to play online together (we never did) and he kept refering to my old alias \'Whitey\' as \'Whitt-ey\', and I just walked away and I went down to the lower level and there were again alot of people, and that is pretty much it. I probably would have said more but this is pretty much all I remember, this is about 3 hours since I woke up so I probably forgot some things.  This is somewhat weird because I was just talking to someone about Asian girls and girls with tans online just hours before. Hmmm... '),(24265,3495224,'02,June,2004','2004-06-02 00:00:00','   urlLink    VH-AFR Resting on the GA ramp after a long night.  '),(24266,3495224,'02,June,2004','2004-06-02 00:00:00','   urlLink    VH-SMH stting next to one of Pel Air\'s westwinds.  '),(24267,3495224,'02,June,2004','2004-06-02 00:00:00','   urlLink    VH-ACV sitting beside the hawker pacific hanger.  '),(24268,3495224,'30,May,2004','2004-05-30 00:00:00','    urlLink    One of the local residents at the Currumbin Wildlife Park...   '),(24269,3495224,'30,May,2004','2004-05-30 00:00:00','    urlLink    The Kangaroos feeding at Currumbin Wildlife Park...   '),(24270,3495224,'30,May,2004','2004-05-30 00:00:00','    urlLink    Tip and grandma at Mim\'s Birthday party...   '),(24271,3495224,'30,May,2004','2004-05-30 00:00:00','    urlLink    The Tweed River off the balcony at Mim\'s place... Could live here for ever...   '),(24272,3495224,'30,May,2004','2004-05-30 00:00:00','    urlLink    Mim And Heidi at the birthday party...   '),(24273,3495224,'30,May,2004','2004-05-30 00:00:00','    urlLink    Uncle Nick and Mitchell at Mim\'s birthday party...   '),(24274,3495224,'30,May,2004','2004-05-30 00:00:00','    urlLink    Tip, Dad, John Whiteman ( Dad\'s Brother ) And myself at Mim\'s birthday party....   '),(24275,3495224,'30,May,2004','2004-05-30 00:00:00','    urlLink    Mim at her birthday with Tara ( Left ) and Sheridan ( Right ).   '),(24276,3495224,'30,May,2004','2004-05-30 00:00:00','    urlLink    All the bones of our dinner....   '),(24277,3495224,'30,May,2004','2004-05-30 00:00:00','    urlLink    One of our local mates at the Caravan park, Came every morning and every night....   '),(24278,3495224,'30,May,2004','2004-05-30 00:00:00','    urlLink    Danno, My cousin who now lives up at Mudgeeraba with his girlfriend Natasha. We went out fishing that night and caught a bream, And decided we would have a midnight snack... It tastes like chicken when it\'s cooked on the BBQ...   '),(24279,3495224,'30,May,2004','2004-05-30 00:00:00','    urlLink    Dad And Brian (The Snake Man) At his small show of snakes at the Mount Warning Caravan Park...   '),(24280,3495224,'30,May,2004','2004-05-30 00:00:00','    urlLink    There\'s a polar bear up there somewhere. Can you spot it??? Taken at Sea World On New Years Day...   '),(24281,3495224,'30,May,2004','2004-05-30 00:00:00','    urlLink    Tip And Heidi in VH-SWL...   '),(24282,3495224,'30,May,2004','2004-05-30 00:00:00','    urlLink    VH-Sierra Wiskey Lima the callsign for this helicopter at Sea World. They put on an excellant flight. My First time ever up in a helicopter and it was great...   '),(24283,3495224,'30,May,2004','2004-05-30 00:00:00','    urlLink    Darryl on his bike at The Mount Warning Caravan Park.... One of the residents there... Took me for a ride on this bike, it was great fun...   '),(24284,3495224,'30,May,2004','2004-05-30 00:00:00','    urlLink    Dad sitting outside of G7 enjoying a Wee Willem Cigar.&nbsp; urlLink   '),(24285,3495224,'30,May,2004','2004-05-30 00:00:00','    urlLink    Heidi sitting on the step outside of G7.&nbsp; urlLink   '),(24286,3495224,'30,May,2004','2004-05-30 00:00:00','    urlLink    Mum sitting outside of G7 enjoying the start of her holiday.   :-)&nbsp; urlLink   '),(24287,3495224,'30,May,2004','2004-05-30 00:00:00','    urlLink    Tip in the Van, drinking from the Cacade Light beer bottle with a hose of some sort. Rumour has it that is Dad\'s toilet at night when he can\'t be bothered getting out of the van.&nbsp; urlLink   '),(24288,3495224,'30,May,2004','2004-05-30 00:00:00','    urlLink    A picture of G7 in the kitchen with my bed in the background.&nbsp; urlLink   '),(24289,3495224,'30,May,2004','2004-05-30 00:00:00','    urlLink    Me @ Mt Warning On Darryl\'s Motorbike&nbsp; urlLink   '),(24290,3495224,'30,May,2004','2004-05-30 00:00:00','    On the 27th of December 2003 the QeIII was set to depart north for a voyage that she had taken on many times before. She would trek north heading through Richmond, Wilberforce at lightening speeds. After a long hard trip along the putty road she came to one of her main and favourite waypoints Gyra or Black Mountain. She docked for no more than an hour to let her crew members have a meal break. The crew members were having a great Bacon, Eggs and sausages with toast whilst having a chat with their old mate Billy, who runs the Black Mountain gas station of a night time. The crew were ready to get back on the road and got the old girl running for her next leg on to Glenn Innes which was where she was to dock at 3:00 a.m. to have a long earned rest. The old girl rose to a phone call by Wendy Brown at 8:30 a.m. after which she got her bits and pieces rolling again along the New England Highway. The old girl was running well and before the crewmen Dave      (Captain) and Charlie (Co-pilot) knew it they were docking the old girl in Casino for a late breakfast and a well earned hair cut by the Captain. As the Captain got his haircut, Charlie took the chance to get a bite to eat from the local pie shop. Both Captain and Co-pilot were prepared to get the old girl started again and get back onto the road. At approximately 4:00 p.m. after a long day of driving the old girl found her way to the Sirrus Apartment block in Tweed Heads and gave a huge surprise to MIM (Grandmother) and had a quick chat before starting the old girl up to make her way down south to the Mount Warning Caravan Park which is situated just west of Murwillumbah. The QeIII had done the crew proud and docked at cabin G7 before going to the front office and saying hello to Judy and Bob. After getting the key to G7 we got all our luggage inside of the cabin and got organised to go and have a shower before deciding what to do for dinner. Anyway we got back into the QEIII and departed the caravan park for Murwillumbah. We got some munchies and then we headed back to the caravan park. At this stage it is the 28th December 2003 and we had to wait up for Mum, Heidi and Tip to arrive at the caravan in mum\'s car. Dad and I decided to watch Open All Hours on his portable DVD player in his room, We watched that for a while and Dad ended up dosing off, I thought I\'ll go out to my bed and watch the DVD on my own, So i began to watch another chapter of it and I was beginning to fall asleep, So i figured they would still be another hour or so away and I don\'t feel like staying up, So i turned the DVD player off and fell asleep. At about 2:30 a.m. I was awoken by the arrival of Mum, Heidi and Tip so we sat around talking for a little while to ask how the drive was for them. All was well so we decided to hit the beds again. We were planning to go to Mim\'s not long after we woke up, which ended up being 10:00 a.m. We all had our showers and breakfast and got into Mum\'s car and headed out to Tweed Heads.    31 December 2003  The Day of Mim\'s Birthday, We awoke as early as possible to have breakfast, and have our showers etc. We got organised and headed off to Jan\'s house where the party was to take place.        Built On Monday 30th May 2004  ( Please Do Come Back To See Updates Of This Website Tomorrow )   '),(24291,3495224,'16,July,2004','2004-07-16 00:00:00','    I have decided to renew my blogspot webpage by putting not only the NORTHERN HOLIDAY IN DECEMBER 2003 but alot of other trips that I have taken not only with Dad but every1 else in my family.. And also friends....  I should be able to get some time to get that done with in the next couple of days/weeks... &nbsp; Please come back and check it out.... &nbsp; Thanks   '),(24292,3931951,'25,July,2004','2004-07-25 00:00:00','    Hey all.&nbsp; I am so wired right now, but in a lot of pain.&nbsp; I went to my friend\'s birthday party last night and got really wasted.&nbsp; We went and played on the bales (The joys of living in a farming community in rural Saskatchewan) and we were all jumping off of them onto other bales.&nbsp; Well, when I went to jump onto another bale, I fell down the entire stack of bales and landed right on my back.&nbsp; Oh my god, did that ever hurt.&nbsp; Ry was looking for me and here I ended up on the exact opposite side of the bales as him and had to walk through a great big puddle so deep my pants were drenched past my knees just to get back to the party.&nbsp; It sucked, but it was fun and I was so drunk that ten minutes later I couldn\'t feel anything anyways.&nbsp; I didn\'t get home until 4:30 in the morning.&nbsp; My mom almost shit a brick.&nbsp; Ah well.&nbsp; I went to my exboyfriend\'s house tonight and played X Box with O (My ex).&nbsp; We broke up on the first of July, and after only 24 days we\'re awesome friends.&nbsp; We party together and hang out a lot.&nbsp; I still love him a lot but I don\'t have the urge to go out with him again.&nbsp; The relationship lasted for almost 3 years, so I\'m pretty much ready to be single again.&nbsp; Especially since next year I graduate and will be off to University where boyfriends will be such a distraction.&nbsp; Still love him though, and I really hope we\'ll be friends for a long time.&nbsp; He\'s a really great guy.&nbsp; And he\'s a really great fuck too.&nbsp; That\'s a plus.&nbsp; At least I still get to do that with him without having any emotional attachment and all that clinical/mental bullshit.&nbsp; Anyways, I should go to sleep.&nbsp; I have to work all day tomorrow... 10:30 am until 9:00 pm.&nbsp; wow!&nbsp; G\'nite all.     &nbsp;    luv, nic&nbsp; xoxo'),(24293,3931951,'14,July,2004','2004-07-14 00:00:00','    i don\'t know what to write.  i\'ll write more when i figure out how to... well... do what i gotta do. luv,nic xoxo '),(24294,3931951,'02,August,2004','2004-08-02 00:00:00','    Howdy.       My weekend has been terrific. It started early, since my best friend\'s birthday was on Thursday. She had a party at Aimee\'s (my other best friend) house and we all got really hammered. Tiff (my best friend) and I hung out pretty much the entire night, and we had a great time doing it. It was fun. Lots and lots of fun.       Friday night was a dead night here in town. There weren\'t any parties unless you were willing to drive an hour out of town just to get to one. Instead, Tiff and I played Monopoly at her brother\'s place and she whipped my ass! It sucked losing, but it was great fun anyways.       Saturday (yesterday) there was a great big pit/bush party. I went out there with Tiff and Aimee. Tiff was the DD for the night, so Aimee and I got drunk. I met a guy there named Carter. He\'s from somewhere in British Columbia, but he was down here for some family reunion. He came to the party with a couple of his cousins, whom I practically grew up with when I used to live out on a farm. It was cool. And he\'s a great kisser. He asked me for my phone number, and I gave it to him, not actually expecting him to phone me, but, much to my surprise, I got a phone call this afternoon.       My God, the phone in this house was practically off the hook all day today. I was constantly getting phone calls from all of my friends. It was unbelievably crazy. My friend Ashton phoned me to see if I would go visit a friend with her. He used to live here and is back in town for a little while visiting his grandpa and aunt. I tell you though, I had the most boring time out there this afternoon. I was ready to gouge my eyes out for the least amount of amusement.      I got the day off from work at the clothing store, but I was supposed to work at the ice cream shop for the evening shift, but I got another girl to cover so I\'d have one full day off of work. I don\'t have many of those, working three summer jobs and dealing with my alcoholic behavior (Just kidding!). No, I\'m not an alcoholic. Borderline, perhaps, but not full blown.      I got home from visiting that old friend, was home long enough to watch two episodes of What I Like About You, and then I left again. I drove around for a little while, then ended up at O\'s place. I hung out with him tonight, and we rented a movie. Scary Movie 3. I\'ve seen a couple times already, but Scary Movie movies are so unbelievably great; I could never get sick of any of them. When O and I went to the store to pick up the video, we ran into Carter and his cousins. Talk about awkward. It felt way weird, but kind of cool too. I burst out laughing and they all looked at me weird. Whatever though. I don\'t care.      Tomorrow, O and I are going to the lake. That is, if the weather is nice. We were supposed to go today, but the weather was totally shitty. Wasn\'t hot enough, is what O says. Big baby. Too picky. Carter and his cousins are supposed to be going out there too tomorrow. Maybe we\'ll run into them again and it\'ll be one majorly disturbing run-in. I can\'t wait for the awkward silence that awaits us.      Whatever happens, happen. Can\'t really do anything about it, now can I? But it does feel weird though. I still really like O, but there are so many other guys that, even though I\'m not looking for a boyfriend right now, I would enjoy spending time with and possibly dating. Nothing serious. But... Fuck it. I don\'t know what I\'m talking about. But, good news is... This guy named Josh, who is really cute and really cool, likes me and is coming to visit his grandma in a couple weeks. Yay. I hope we hang out.      Jesus Christ bananas! I sound like some love sick, boy crazy girl in the seventh grade. Fuck me Freddy! (I watched Dreamcatcher the other night, hence the new \"slang.\" I definitely recommend watching that movie. Fucking great!) I\'m really not a love sick, boy crazy girl in the seventh grade. I\'m a grounded, mature teenager in the twelfth grade. I am smart and, depending on the activity, talented. I am not the kind of person to drop everthing for love. I will not, can not, and do not make my life revolve around love and sex. I am a rational person. I have my priorities straight. I know what I want out of life... What the fuck. Who am I trying to convince here; You? Or myself?      Anyways. I should run. Don\'t know why. I think maybe I\'m just done writing stuff. G\'nite all.      luv, nickie   xoxo'),(24295,4204173,'11,August,2004','2004-08-11 00:00:00','   How do I plan to lose 50 lbs, you ask?  By jumping on the low-carb bandwagon. NO, I\'m not doing Atkins--there\'s no way I could ever stick to that for life. But, I am attempting to keep the carbs below 80g net per day. I am also counting calories, too, because I know from experience that calories DO count. So what have I eaten today so far?  Breakfast: 3/4 cup cereal, 1/2 cup skim milk  Lunch: 2 Sausage Patties, 2 slices reduced-calorie whole wheat bread  Snacks: 1 orange, 1 oz. of dry-roasted peanuts, 1 container of fat-free yogurt.  Dinner: 4 oz. turkey breast, 1/2 cup of mashed potatoes, green beans and 1/2 cup turkey gravy.  Total Calories: 1140(Yikes!) Total Net Carbs: 78g  I need to get off my lazy ass and exercise. But this blogging is oddly addictive.  On a personal note, my husband Gary runs a small golf course in our town. He leases the land from one of the local business owners, who informed him today that he was putting the land up for sale in a couple of weeks. He is asking $500,000 for it and wondered if we\'d be interested in buying it. Oh, yeah, we\'ve got that money in our back pocket, mother fucker! So it looks as though Gary will be on the job hunt in the near future. In a way, I\'m glad--he spent entirely too much time there and now he can be home with me more, but, he really loves what he does and I hate to see that taken from him. More to come in the future.......   '),(24296,4204173,'11,August,2004','2004-08-11 00:00:00','   Yesterday I began what I hope to be my final attempt at losing this excess weight.  I weighed in this morning at 175 lbs. EEEK! At 5\'3, that is just plain FAT.  How did I get this way?  Looking back, I have always had some difficulty maintaining a good weight.  As a child, I was not fat at all, but \"stocky\", \"sturdy\" and all the other adjectives used to describe a non-skinny child.  My real weight issues began at about age 11, when I hit puberty with a vengeance.  I suddenly had these enormous tits and began gaining weight in earnest.  By age 13, I weighed 140 lbs at 5\'1.  It hit me that I needed to do something about it when my normally very gentle mother told me that \"I didn\'t need to gain any more weight\".   She understood better than anyone the journey that I was about to embark on; she, too, has struggled all her life with her weight.  So I began watching what I ate.  I don\'t think I really had a goal in mind, I just wanted to be smaller.  By the time I started eighth grade, I had lost 7 lbs.  No one really noticed my weight loss, I was still the same old fat & ugly April.  So that October, I decided to buckle down and lose the weight for good.  By Thanksgiving I had lost an additional 9 lbs, bringing my total weight loss to 16 lbs.  We went home to Kentucky for Thanksgiving and all my relatives gushed over my weight loss! It was the best feeling in the world, and I think it was the catalyst that allowed me to continue to eat well and exercise.  By the summer before my freshman year, I weighed 110 lbs. and looked HOT. I was getting a lot more attention from the male sector of our society and was enjoying attention from a number of suitors.  I gained a little of this back, but maintained a weight of 120-125 throughout high school.  Shortly after finishing high school and starting college, I began dating Chris, my first love and the man I lost my virginity to.  Happiness and comfort set in and the weight began creeping back.  Throughout our stormy relationship I managed to put on 30 lbs.  Chris didn\'t care, but deep down I hated myself for allowing my weight to creep back up like that.  We dated for nearly 3 years, and finally broke up in April of 1999.  At this point I realized that I was back in the dating arena and had to make myself presentable again. I embarked on a crash diet, sometimes eating nothing more than a beefy mac frozen dinner and a hamburger ALL day.  I also walked 2 miles a day, 5 days a week. In a little more than a month I was back down to 125 lbs.  I had a few flings, slutted around a bit and then, on July 17, 1999, I met the man I would later marry, Gary.   God, when I look back on the early days of our relationship, he and I were SO HOT.  Isn\'t that the way it is with everyone?  Well, in no time flat, my old friends Happiness and Comfort showed up at my door again, and I proceeded to gain back all the weight I\'d lost post-Chris.  Gary, unlike Chris, was concerned about my weight gain, though he never came right out and said anything, I could tell.  At any rate, we got engaged in February of 2001 and I began planning a wedding!  I began WW in April 2001, lost 8 lbs, then maintained throughout the summer.  I joined back up in August and lost an additional 12 lbs. I never made it below 130, but I thought I looked pretty damn good.  Gary and I ended up getting married in Vegas in January 2002, and shortly after that, the weight creeped back on.  I stopped using birth control pills in May of 2002, weighing 147 lbs. at the time.  By November of 2002, I had reached my all-time high of 175 lbs.  By this time I realized that I had a condition known as  urlLink PCOS , and that that was contributing to my weight gain more than anything.  I put up with it for about 2 years and in March 2004, I went to my OB/GYN to get help for it. He put me on Metformin, which is a diabetic drug known to help PCOS\'ers with their symptoms. I stopped gaining weight at that point.  I ate just about what I wanted and maintained my weight of 175 lbs. However, that wasn\'t a weight I wanted to be for the rest of my life!  For one thing, Gary and I are trying to have a child without any success(PCOS is a common cause of infertility for women) and I know that losing weight will make the journey easier.  So, here I am today, at 175 lbs., vowing to take the weight off FOR GOOD this time.  Maybe seeing it in writing will make it more real to me.   '),(24297,3835812,'31,July,2004','2004-07-31 00:00:00','   Match 1: He lives in the Valley and is single with two grown kids. He is nice on the phone. We converse comfortably and decide to meet in person. Oh, by the way, he needs to tell me he hasn\'t exercised in decades and is out of shape...\"How much out of shape?\" I querry...He answers...\"I look like Shrek\"   Match 2: Lives in LA. Unemployed for the last two years...\"I would like to be a writer.\"  So, I querry, \"are you writing anything now....in your spare time??\" \"no.\"  Match 3: Nice looking guy (at least he posted a picture) lives near by. We talk on the phone...polite. We meet....\"I have a problem with alcohol.\" I am stupid and go for the second date. I should be shot. He arrives drunk.  So, anyone want to join eharmony?  Really...I\'m not bitter   '),(24298,3835812,'20,July,2004','2004-07-20 00:00:00','   Almost everything I know about how to live, I learned in my music theory class. Yes, you can learn serious stuff in music theory. The most notable lesson I learned is when to break the rules. You know, I was a good kid....I just wanted to stay out of trouble. I did what I was supposed to do and didn\'t break too many rules. Rules were big to me, but I learned that you can\'t camp on rules...you quit living! Here\'s the deal: Everyone needs to learn the rules. Rules about life, social rules, music rules. It\'s the foundation! But once you know them it is critical that you learn when and how to break them. I had a professor who liked to tell us, \"you have to earn rubato!\" Rubato is when you stretch a note, or shorten or change a note. His point was you have to know the music well before you can do these things. This applies to all kinds of areas of life. Even Jesus taught it. He went to the religious leaders and told them they had distorted the commandments...they needed to learn the principles behind them....they even needed to learn to break a few. you aren\'t living if you aren\'t breaking a few rules and breaking them well.   '),(24299,3835812,'19,July,2004','2004-07-19 00:00:00','   That\'s the trouble. Here I am at the stage of life that I should be thinking of retiring and just being a grammie and doing the things I\'ve always dreamed of. Oh well, I am going to school instead. I have felt for a long time that I would eventually go back, but I waited for the right motivation. When my good friend died earlier this year I realized what I wanted to do. Social Work. I felt like standing next to her bed holding her hand and just being there for her was exactly where I was supposed to be. I remember how many times she told me to follow my path...even if I had to go back to school. She said that women often don\'t reach their full potential. I was encouraged by her words. She sure reached her potential. I don\'t know when I have experienced a death that has left so many people with a hole in their hearts. She was amazing and loved everyone around her....yeah, I want to do Social Work. I guess I better go study....   '),(24300,3835812,'19,July,2004','2004-07-19 00:00:00','   OK, I am putting this question to the universe. Where does mom-hood stop and adulthood begin? I think blogging is a great way to share thoughts, ideas, musings, etc. so I didn\'t think much about reading my kid\'s. It\'s nice to experience them as adults who have their own thoughts and things to say. Who knows when they will teach me something valuable? Around the dinner table it is difficult to keep a good intellectual conversation going, but on a blog it is easier. Is it a violation of privacy to read your family\'s blogs????   How about: I\'ll show you mine if you show me yours??  Love, Mom   '),(24301,4293254,'23,August,2004','2004-08-23 00:00:00','   I\'m glad in a sense my masks aren\'t working, but in another I wish I\'d stop crying. I\'m coming off as pissed off to some teachers. Yes I am down. I\'m down, that my boyfriend is leaving for Elon this weekend. His parent\'s won\'t let him have the fucking car even though he\'s paid at least a good $1000 on it. And it\'s not like they need it either considering they have two cars already. I\'m overwhelmed by my honors classes and I doubt the homework will be letting up anytime soon. I know I\'m screwed for this year. I\'m going to end up at the bottom of my class I see it already. I\'m a failure. I wish they\'d stop posting grades because I compare myself to others(even though the names aren\'t listed). I see it already. Branded across my forehead: FAILURE AT LIFE! I can\'t believe what I\'ve become. I love my hate myself and hate myself all in one futile breath. Is this all I think to myself, before I pull my face out from the pillow in the morning. Sometimes I wish I could lie there all day. Somedays I think I\'m dead walking around in some girls body, then again maybe I am... Maybe this is just the clueless ramblings of my depression. Why am I this way? Am I getting suicidal I think to myself? I constantly think of placing the knife into my veins over and over again until I\'m drenched in blood. Is this a beautiful thought? I don\'t know. This quickly moves out of my head. I\'m better now, I don\'t want to return to the places I\'ve been I can do better than THAT. I\'m soooo fucked up right now, then again, when will I not be fucked up? -Nikki '),(24302,4293254,'20,August,2004','2004-08-20 00:00:00','   Hello to the World that doesn\'t know I exist, Today, when I went into to Chemistry I Honors we were suppose to turn in our measuring the thickness of foil project(which I pretty much felt like I did the whole project), a couple of the people in my group came up to me and asked if I had turned it in. When I replied with a quiet \"yes,\" they didn\'t respond or anything they just walked away. I didn\'t receive a thanks or anything. This is after I discovered how to find out how thick a piece of foil is by doing research, typing up and editing all the bull shit they screwed up on, and so forth. Maybe I am being a little hasty, but my time is very valuable to me. Last night I didn\'t have time to finish all my homework because I was editing \"OUR\" paper. I though a group paper, was suppose to be a group \"effort.\" Next time I\'m doing the work on my own. Christ, I might as well. -Nikki '),(24303,4293254,'15,August,2004','2004-08-15 00:00:00','   Hello There My dear World,I\'m extremly ADD today. My mind has been wandering off all this morning. At church I only had one kid in my room which I was glad because if I had more I\'d probably loose them and their parents would be very afraid. My psychologist told me I need to think of positive things I like about myself: I like: that I look older than my age, my hair, my smile, my eyes, you know something I never realized this, but I think the reason why my mind is always thinking about kind of sad things is because I kind of enjoy thinking about my life... I mean my life is like the average american family. Which says a lot for society. Hmmm.... Woah... Off topic AGAIN!!! Anyways I have some Honors reading assignments I have to do so I better go. Love,Nikki '),(24304,4293254,'14,August,2004','2004-08-14 00:00:00','   -Something weird happened in my dream last night , when I opened my wallet, I had a ton of like $1 in foreign money, bute the thing was that it wasn\'t all 1 bills. -I think this is my subconious trying to tell me something about myself. I\'ve been having a hard time lately at school. I\'ve been putting in a lote of effort though. My schedule goes like this: Honors Chemistry 1Honors Environmental ScienceHonors English IIIHonors Integrated Math IV QuantativeA LunchHonors Civics and Economics (a bunch of 10th graders)Spanish II (they don\'t have honors for this class)Newspaper (Editor and Cheif) -Also, Mrs. Kraft the Student Council teacher wants me to run for President, I want to, but the question is: Am I doing it for myself? Or for Mrs. Kraft? For the school? Or for who? -I broke down today at school though, I started crying in the hallway, what was I crying over? Stress I suppose, and I was extremly overwhelmed. My homework is rough, not rough, but when it adds up in all 7 classes it covers my night. Boy, am I glad I don\'t have a job I\'d be screwed over. I\'ll stick with Church. I\'m so depressed right now... I better go do stuff, before Charlie(Hurricane) comes. Love,Nikki '),(24305,4293254,'09,August,2004','2004-08-09 00:00:00','   Let\'s think about what good therapy did today for me? Hmmm... NOTHING!! It only made me realize how much more I hate being stuck in the skin I\'m in. Why is life so fucking difficult? I mean yes I know \"no body said life was going to be easy,\" but fuck how in the great glorius world could so much drama and shit happen to me. Let\'s see... Both of my brothers suffer from different forms of schizophrenia along with depression, my mom has MS and refuses to use her crutches even when she\'s in extreme pain, my father is so fucked up he doesn\'t realize others have feelings and has so many money problems I don\'t even want to think about it. God, you know, sometimes I look around at people and think why do you have it so fucking good? Who said you could have everyone worship you? I have an idea next time your MOM HAS TROUBLE WALKING OR IS HAVING SPASMS YOU CAN TAKE THE DAY OFF FROM SCHOOL TO TAKE CARE OF HER? HOW \'BOUT THAT UH?? You know WHAT AND THE NEXT TIME A TEACHER SAYS THEIR TIME IS MORE IMPORTANT THAN MINE I CAN TELL THEM WHAT I\'M GIVING UP TO BE THERE, because you know what- they\'ve had their parent\'s their parent\'s have supported them, but what about me? ASS HOLESS, ASS HOLES!!! ALSO NEXT TIME I WORK MY ASS OFF AT CHURCH I HOPE EVERYONE NOTICES THE SHIT I GO THROUGH FOR ALL YOU PEOPLE BECAUSE YOU KNOW WHAT I WORK SO DAMN HARD TO GET THE MONEY I GET, SO DON\'T WHINE AND BITCH ABOUT HOW LITTLE OF A PAYCHECK YOU GOT, BECAUSE YOU KNOW WHAT MY PARENTS DON\'T HAVE THE MONEY YOUR FAMILIES HAVE? SO TAKE A STEP BACK AND LOOK AT ALL THE FUCKING THINGS YOU HAVE THAT YOU HAVE THAT I DON\'T AND YOU KNOW WHAT, STOP BITCHING ABOUT HOW YOUR PARENT\'S SUCK BECAUSE ONE DAY YOU will need them more than ever, and you know what they\'ll be gone. Why am I sitting here writing such an angry letter to the world, a huge part of it is because I\'d like to be said \"goodbye\" to or invited to go along with some of you guys to the things you do? You know what I don\'t hate you, I\'m jealous of everything that you have that I don\'t. Lisa- I\'m jealous that your dad cares about you enough to look at you, and that people are constantly taking care of you, even though this is also a bad thing. Sofie- Your slimness, and occassionally your humor even though it can be brutal. Miranda- Your so sweet,and kind and I think you really do love yourself deep down. Sara G.- Your parents even though you hate them soooo much they are able to be REAL with you, and tell you what they think, oh and lets not forget that its good to have our baby fat! Karen- Your ability to keep a secret, and hold it... I will finish this later.. I do love you all but please realize I\'m human too... love,Nikki '),(24306,4293254,'06,August,2004','2004-08-06 00:00:00','   Last night was fun, sort of. I woke up early yesturday day and went into school to work on the newspapers new office. I got to school around 8 am and then left around 4pm in the evening. Mom and I headed to the bank because on tuesday the atm ate my check without crediting my account. After taking care of that we drove my Wendy\'s to pick up something. God I hate fast food, it\'s so greasy, but we didn\'t have time to make anything, so I settled for a small burger, fries and a frosty. After wiping off the grease I finally ate my food. Stopped in at home to change then we went back to DSA (my school) for Orientation. Orientation was okay, but didn\'t realize until later that I didn\'t get Chemistry Honors like I had signed up for which makes me ANGRY!!! But I put in a drop add form and I\'m a little worried might change the status of my Honors Envir. Science Class, but I was told it wouldn\'t. Anywho, later when I got home I talked to John on the phone for a little while. I really enjoyed it, I couldn\'t stop laughing, we\'re going to go bar hopping on Sunday! YAHOO! lol Anyways I have to go, I\'m sitting on my floor in the buff, so yeah, I\'m hungry... lol buh bye love,Nikki '),(24307,4293254,'04,August,2004','2004-08-04 00:00:00','   My day was soooooo fulll today, and I feel terrible. I woke up this morning to go to school and work on NEWSPAPER stuff until 4pm. Well at 4pm I headed over to my mom\'s work and about 5 mins later headed over to Carrboro for my writing class. I was late for my writing class by about 20 mins. as usual. After the class ended around 6pm I had to wait around for an hour. During that hour I was learning or rather attempting to learn how to hemp... geez I suck at hemping. Sorry John, I had a cool idea for a new bracelet for you too... That\'s another thing, I was thinking alot today about how it turned out I wasn\'t going to be able to go to game night right after I promised John S. I\'d come. My mom is having issues though right now with what John and I had talked about. Hmmm.... Anywho .... I\'m really tired and I have to go to school again tomorrow to continue working and organizing the newspaper stuff so I better go to bed... hehehe GO TO BED... With who?? no one :( lol love,Nikki '),(24308,4293254,'03,August,2004','2004-08-03 00:00:00','   I can\'t believe all the crazy things that happened this past weekend. The weird part of it is, is that everything is working out okay, I guess. It seems interesting the way Sofie called me a blabber mouth a while ago, and then wonders why she \"bothers\" me. I guess why Sofie suddenly comes to mind is because of all the people invovled in this situation I think she could be the one to \"handle\" it better. I understand where Drew came from, yes he did break a few bonds, but it\'s not like John didn\'t already know who liked him. It\'s rather obvious. The thing is, is that I don\'t think people realized how close John and Nicole were because the way everyone is acting right after they broke up is insane. I don\'t know this is just my opinion, maybe I\'m rather confused. All I know is I care about John and I hope he takes his time recooperating. love,Nikki '),(24309,4095362,'31,July,2004','2004-07-31 00:00:00','   OMG! im so bored. played netball today and we won, of course. lol it\'s all your fault im on this thing blondie, i wanted to write u a note an i couldnt unless i made a site :( so i did, coz i really wanted to write u something ne hoo, im over it mmmm, jumpy (you get it millie) is still being a prick. dont know why i bother with him i think he\'s such a waste of time i cant believe he called me a slut!!!! grrrrr. not happy any how, gotta go homework calling :( real loud, better shut it up lol buh bye huni\'s  '),(24310,4095362,'13,August,2004','2004-08-13 00:00:00','   im so happy right now. i cant wait for tomorrow night. everything seems to be working out heaps good. im really happy with chips, an even david has come to accept that im going to be with him. all he wants  now is that im happy. chips was worried about david being angry, an i dont want them to ever meet, but i think if i get hurt now, david would be so pissed off. i dont think i will get hurt though. but im really scared of hurting someone with all of this. it just feels too good to be true, so i\'ll make the most of it while its here. i dyed my hair, an it looks retarded. its too red, lisa i dont like it. :( oh well i guess ill get used to it.  mum didnt buy me alcohol :( i think i\'ll just have to take something out of the cupboard lol they wont even notice but i\'ll tell them, just in case.  any hoo im going to bed, got a big night tomorrow ;)  bye  '),(24311,4095362,'12,August,2004','2004-08-12 00:00:00','   things are still heaps confusing i cant spell tonight an i cant be bothered fixing everything i fuck up. millie, i cannot belive that you didnt tell us you an daniel were going out.... how rude!! david an i are finally finished i think, i know i say that all the time, but i really think me might leave me alone now. just in case you were interested millie, the main reason i havent said yes to chips is coz david an i still kinda have things to sort out. well, i dont think we do, but he does. i want to say yes, i think, but i dont want to ruin anything. mum thinks theres going to be no guys there at the party at all. she thinks megans mum is heaps cool. she said when i talked to her that i owe her a favour. lol im heaps tired but i had to come on just to write in this so everyone knows whats going on inside my twisted little brain nighty night  '),(24312,4095362,'09,August,2004','2004-08-09 00:00:00','   i think chips an i could have something, but im not game to tlk to him about it. im scared he\'ll say he doesnt want me, when i think i really do want him. david is being a prick again. he said he was going to make me happy, well he certainly isnt right now. i have to go soon, this will only be short. david wants to call :( lol lisa, im really confused, i think we need to talk tomorrow about this an your problems with josh aka stains aka dickface or whatever you want to call him bye  '),(24313,4095362,'07,August,2004','2004-08-07 00:00:00','   david an i are still not fully apart. he wont let me have a break now that he thinks he knows about me an chips. now i dont think i like chips anymore. i think that how i felt about him was more about what i was missing with david. he made me feel special an gave me attention that i wasnt getting with david. i dont want to be with david, but i dont really want to be with chips either. right now, guys are a waste of time. im sick of them.  next day.... david wanted to read this so i stopped writing he is being a prick again an wont give me space. he doesnt get it. by trying to hold onto me an not letting me have a break, hes pushing me away. theres something my dad told me that i remember... that is something is meant to be it will happen... if you let something go an it comes back, it was meant to be. i want to tell david that but he wont listen. right now, the way he is acting, i really dont want to be with him. chips is the one i think of most of the time... when i think of david im angry with him, like when he falls asleep on the phone to me, thats why i fall alseep thinking of him, not chips. im really confused about everything. i dont want to bring anyone into the mess david an i have, but chips is there anyway. i dont know who i want to be with anymore  '),(24314,4095362,'05,August,2004','2004-08-05 00:00:00','   david and i are taking a break. its what my parents wanted all along, but they weren\'t going to wait to let us sort things out first. thats all i wanted, an im not game enough to tell them that. now david is coming over tonight to tell them that, and mum is cutting sick coz hes said some things he shouldn\'t have. so mum pissed me off an now ive told her i dont want to be here at home with them anymore. i shouldnt have told her that but i was really angry, an now im scared theyre going to kick me out. if they do, david has always told me i can stay at his place, but i do really want to have a break from him, an if im staying there, it wont happen. everyone says that me an chips like each other, an i do kinda like him. it surprised him today when turtle told him i had a bf, which i dont really... he just thought i was still with david. i dont like him as much as i like david, but when david an i were talking last night, when i said that chips was fun to be with an all, he kinda took it the wrong way. i like most things about chips, but that doesnt mean i want to be with him. hes just a friend right now, even though i think we both like each other. there is no way id bring another guy into the mess that david an i have got going. it isnt fair to anyone. i want to leave home, but now isnt the right time. i need to get my license, an a job, an a car. if i left now, everything would be really fucked up. my parents need to get a life, and let me have mine back. they told me that i could make my own choices, an that i would learn from them, but theyre not letting me do anything. i dont get to make my own choices, an i dont get the freedom i want to do what i want. grrrrrrrrrrrr if everything was good with me an david, i would go to his place for a while, so we got some time together an to make mum an dad realise that i was serious about leaving. i really dont want to be at home. anyone want me!!!! lol anyways better go, david will be here soon an i think i need to go an get ready for a mass fight cant wait, this is going to be so much fun NOT :( bye  '),(24315,4095362,'04,August,2004','2004-08-04 00:00:00','   here\'s my little bitch bout david. i cant believe he still thinks i want to be with him. ive told him so many times i only want to be friends, but he doesn\'t get it. thanks millie for sharing with the whole world that i\'m so in love with chips. i dont know where u got that from. ur such a shitstirrer lol lisa i think millie has a new idea in her head keep an eye on her, and when she tells you, just agree. you know that its way easier just to do that you know im joking milz luv ya sweetie anywayz gonna go milz will kill me now, after all this lol bye  '),(24316,4095362,'04,August,2004','2004-08-04 00:00:00','   im on the net, being really bored, wait for ppl to come online to talk. im confused, im still kinda with david, but i like someone else not saying who... nosy ppl read this COUGHmillieCOUGH might tell you l8r missie luv ya :) any ways gonna go, millie has to read this so i better cut it short bye  '),(24317,4130944,'04,August,2004','2004-08-04 00:00:00','     '),(24318,4130944,'04,August,2004','2004-08-04 00:00:00','   urlLink    Nina listens while we sing \"Happy Birthday\"...&nbsp; urlLink  '),(24319,4130944,'04,August,2004','2004-08-04 00:00:00','   urlLink    Let\'s EAT!&nbsp; urlLink  '),(24320,4130944,'04,August,2004','2004-08-04 00:00:00','   urlLink    Thinking of a wish...&nbsp; urlLink  '),(24321,4130944,'04,August,2004','2004-08-04 00:00:00','   urlLink    \"No, I don\'t want to use my spoon--I want to eat it with my candle...\"&nbsp; urlLink  '),(24322,4130944,'04,August,2004','2004-08-04 00:00:00','   urlLink    Nina hugging here new teddy bear from Anthony...&nbsp; urlLink  '),(24323,4130944,'04,August,2004','2004-08-04 00:00:00','   urlLink    Nina eating Birthday Cake...YUMMMM!&nbsp; urlLink  '),(24324,4130944,'04,August,2004','2004-08-04 00:00:00','   urlLink    Nina smiling at Uncle Artie who was singing \"Happy Birthday Nina\" with the clucking bunny rabbit?&nbsp; urlLink  '),(24325,4130944,'04,August,2004','2004-08-04 00:00:00','   urlLink    Nina showing Papa that her new Teddy Bear from Anthony has the softest ears...&nbsp; urlLink  '),(24326,4130944,'04,August,2004','2004-08-04 00:00:00','   urlLink    Papa and Nina cut the Birthday Cake...&nbsp; urlLink  '),(24327,4130944,'04,August,2004','2004-08-04 00:00:00','   urlLink    Nina asks Gramma to help her open THE BIG present...&nbsp; urlLink  '),(24328,4130944,'04,August,2004','2004-08-04 00:00:00','   urlLink    Nina turns Uncle Artie\'s head to pose for the camera...&nbsp; urlLink  '),(24329,4130944,'04,August,2004','2004-08-04 00:00:00','   urlLink    Nina Allysa  urlLink  '),(24330,4130944,'04,August,2004','2004-08-04 00:00:00','   Nina Alyssa turned  four  years old this month. She is beautiful, smart, funny, and a light in our lives. We had a special celebration at my house Tuesday night and her Gramma (Nina), Papa (Art), Uncle Artie, Anthony and I had pizza from Papa Murphy\'s, lots of presents, and a cake she picked out special from Bel Air bakery.  Nina \"scored\" this year with lots of great presents. Her brother got her a soft, floppy teddy bear who she named \"Clowny-bear\". She says it has the best \"ears\" to soft. She got two puppets--one bunny and a kitty. Thanks to Mom and Dad and Artie for buying things that MAKE NOISE.  She also got a kareoke machine from her Uncle Artie. Nina is quite the singer these days. Since I\'ve been playing guitar again this summer, Nina, Anthony, and I take time in the evenings to play guitar and sing. Now Nina has a microphone to sing her two favorite songs--\"Dreamland Express\" and \"A Song for All Lovers\" (or as she calls it--her \"Alaska\" song). We just got a Raffi music book in the mail and she thinks it\'s a lot of fun to laugh as I try to figure out the chords.  Her Papa and Gramma got her a \"souped-up\" Hover Vacuum cleaner--so she can chase me around the house when I\'m doing my chores. It SUCKS--I mean, really...the darn thing actually works! (My brother\'s joke.) They also got her a giant puzzle and this very cool reading/writing electronic book-thing that has a lot of books and activities to choose from. She traces the letters, words, etc. and it reads them back to her. There are lots of games and ways to interact with the stories. She had NO problem taking to it and her focus was amazing.  Last but not least, I got her a Delux Kitchen set-up that puts my own kitchen to shame. Nina spends a lot of time doing imaginative play and has been getting things out of my kitchen to make \"meals\" for her animals and babies. Thank God my Dad was here to help put the darn thing together. It has a mini-light above the sink, a cell phone that lights up and beeps, buttons that beep for the microwave, another set of buttons that beep on the oven, the stovetop sizzles...well, you get the picture. She has been busy, busy, busy making all sorts of meals.  This summer, Nina embarked on her first \"big girl school\" adventure at Storybook Cottage. She is officially a \"pre-schooler\" and had no problem transitioning into the new environment. We\'re all very proud of her.  Nina continues to light up our hearts with her belly-laughs, questions about the world around her, the songs she sings, and her genuine affection for those who love her.  Happy Birthday Nina Alyssa!  Love, Mama  '),(24331,4130944,'03,August,2004','2004-08-03 00:00:00','   urlLink    Anthony catches his VERY first fish...it was a small rock cod--but once he caught it, he was \"hooked on fishing\"!&nbsp; urlLink  '),(24332,4130944,'03,August,2004','2004-08-03 00:00:00','   urlLink    Nina takes a turn driving the boat...&nbsp; urlLink  '),(24333,4130944,'03,August,2004','2004-08-03 00:00:00','   urlLink    On our way back in we slowed way down and Anthony got a picture of a sea otter hangin\' out on his back...&nbsp; urlLink  '),(24334,4130944,'03,August,2004','2004-08-03 00:00:00','   urlLink    He kicks up the speed and smiles...&nbsp; urlLink  '),(24335,4130944,'03,August,2004','2004-08-03 00:00:00','   urlLink    Anthony driving \"Hot Tuna\".&nbsp; urlLink  '),(24336,4130944,'03,August,2004','2004-08-03 00:00:00','   The last weekend in July, Anthony, Nina, and I had the unique opportunity to go out on the Monterrey Bay Pennisula and go fishing. We ended up catching 30 fish total--many varieties of rock cod including \"olives\", \"blues\", \"ling\", and bright red ones who\'s names escape me. They made for some very tastey eating. Special thanks to the Livingston family who made our trip lots of fun. We had quite a \"fish fry\" at their home near Henry Cowell Redwoods State Park overlooking Santa Cruz.  '),(24337,4130944,'04,August,2004','2004-08-04 00:00:00','   urlLink    No Fear!&nbsp; urlLink  '),(24338,4130944,'04,August,2004','2004-08-04 00:00:00','   urlLink    Nina reeled this in using the pole in the back. We thought she was just playing with the reel--but she really caught a fish!  urlLink  '),(24339,4130944,'04,August,2004','2004-08-04 00:00:00','   urlLink    The kid can walk on water!&nbsp; urlLink  '),(24340,4130944,'08,August,2004','2004-08-08 00:00:00','   urlLink    Im driving on a boat.&nbsp; urlLink '),(24341,4130944,'08,August,2004','2004-08-08 00:00:00','   urlLink    My first fish!&nbsp; urlLink '),(24342,4130944,'08,August,2004','2004-08-08 00:00:00','   urlLink    catching a small fish&nbsp; urlLink '),(24343,4130944,'08,August,2004','2004-08-08 00:00:00','   urlLink    Me and Ashly roasting marshmellows&nbsp; urlLink '),(24344,4130944,'08,August,2004','2004-08-08 00:00:00','   Hello, I\'m Anthony. These are the things that I did for Summer 2004.  First of all, I went to summer school to take enrichment classes. My classes were Art, Drama, Computers, and Science. I made clay pots, crystals(made of salt), paper cups, clay boxes, and many other things. In computers I learned how to use PowerPoint, Hyperlink, and practiced typing. Since my Mom was teaching Drama, I was her teacher assistant and helped her with papers and with one of the students who didn\'t have a partner. I also got to goof around on her computer and saw some really funny movies.  This year I got to go to Legoland with my Dad. I went on the dragon, and the technic ride. I also got knights kingdom stuff. I also went camping to Lake Francis and had fun with my friend Ashly.  After Summer School I went whith my Mom to the 2004 AVID Institute at the Sacramento Convention Center. I went to Starbuck\'s for a frapaccino, and took pictures of artwork and architeture around the building. Mom even let me lay on the floor to take pictures of this weird sculpture thing on the ceiling.  We also went to San Francisco and used our family card to go to the Exploratorium. My favorite parts were the echoing tube, a smoke ring maker, and a noise maker. on the way back, we went to Muir Woods.  Then we went over to Santa Cruz with our friends Mark, Corey, Jim, William, and Wesley. They have a really big house where we went to a chili cook-off and went swimming. Then we got to go out on their family\'s boat called \"Hot Tuna\". I caught 6 fish in total, one of them was my very first fish that was 11 inches, and the rest were bigger and bigger. The best part was driving a boat. I felt nervous first, but then I felt fine. '),(24345,4256976,'19,August,2004','2004-08-19 00:00:00','   Women can be so cruel.  She was already at the bar when Alain opened the heavy door to the dingy saloon that over the last few months had become his regular choice for post workday drinking and relaxation.  Alain never thought that he would ever be able to muster the self-esteem he thought was required to walk into a bar alone, order a drink and enjoy it quietly.  He had sworn to himself, since his University days, that drinking was a social event.  Therefore, a solo excursion to a bar would forever be helpless, lonely, and desperate.  But he had never been very good at holding a grudge against himself and immediately attempted to contrive a sort of academic experience out of his current situation.  This decision not to turn and flee lasted longer than he had expected.  So, standing awkwardly in the doorway but with new resolve, he searched for the appropriate place to sit.  Alain searched the few scattered faces around him.  They all seemed vaguely familiar like extras in the predictable bar scene of some marginally interesting movie.  For several years Alain had felt himself to be a marginally interesting character whose story, with an expert pitch, could find its way onto the big screen.  But it would only be great once some big time actor pulled out the performance of a career by really capturing the dreary, selfish and pathetic excuse for optimism that propelled Alain through each day.  After this realization, Alain had decided it would be best for the project if he never thought about it again but his current surroundings brought it back out of him.  '),(24346,4256976,'18,August,2004','2004-08-18 00:00:00','   The knight had been following me for several days.  He was adorned in full suit of silvery armor that remained untarnished through thick brush, beaches and swamps.  â€œWhat was his purpose?â€ I wondered to myself.  As I climbed atop every new hill, I could look back and see an unrelenting glimmer from the previous one.  He would always be there.  So I decided to rest a bit and wait for him.  The wind rustled through the long grass swaying around me.  It seemed to whisper words of caution as I gazed at my determined pursuer.  As the glimmer grew closer, I questioned whether this was in fact a wise decision.  He could be hostile.  If he were carrying news he would certainly not have made the journey in such attire.  I tried to take my mind off him.  This proved to be a much more difficult task than I had expected.  When I turned away I could feel hot breath exciting the hairs on my neck.  I could feel his cold blade turning my insides out as he took advantage of that one brief instant that I had let down my guard.  I saw myself stepping away from his second slash, this time to my throat.  But the damage had already been done.  I cradled my organs, steaming in the autumn air, as they crept from the gaping wound he had delivered me.  I fell to my knees.  Each dancing blade of grass lashed at me with cruel joy, scratching and scraping my blood-soaked arms.  My eyes filled with tears.  I lifted my head to glimpse the two blurred suns hovering in the sky before me.  I could feel the heat of both of them.  With great pain I removed my right arm from the nest it had created around my pulsing wound and reached for the hotter and brighter celestial body.  I wrapped my hand around the knightâ€™s gauntlet, glowing white.  All feeling and pain was redirected as the hot metal burned into my fingers, soldering my quivering flesh to his hard, forged glove.  My palm gushed forth steam as it melted to the knight.  I tried to tear my arm away but the pain was even greater than submittal.  So I did my best to relax in this terrible state, for my end must certainly be drawing near.  I must have passed out for only a minute, but it could have been a day for that matter.  What must have awoken me was the intense heat building against my right cheek which was merely inches away from my attackers breastplate.  I attempted to wriggle my right arm but to my horror it had vanished into the fiend up to the elbow.  The knight was fusing me into his burning shell.  I pulled furiously to detach my forearm which was now most certainly his.  I cried out as my face pressed against his armor.  The burning reignited in my brain as it began to boil and my howling was muffled, as I no longer had my mouth.  '),(24347,4256976,'17,August,2004','2004-08-17 00:00:00','   This wedding was different than all the others.  It was different in the way that everyone who attended the festive event died.  Well, not everybody died.  Just everyone who ate or drank at the splendid event died but even that was a pretty large number.  The Brideâ€™s family was dead and so was the Groomâ€™s.  It would later come out that unanticipated levels of poison where incorporated into the food and drink prepared for the glorious occasion.  It would later be revealed that a wedding gift had been stolen from the Newlywedâ€™s.  However the ownership of the gift was called into question due to the unavailability of any beneficiaries.  Therefore the gift had been stolen from the State.  Investigators assigned to the case were tripled now that it was a State affair.  It would later be discovered that no one had actually died at the once happy, once tragic ceremony.  The theft was reclassified as a merely a family concern but the wheels had been set in motion.  The transfer of ownership of the family property to the State, including stolen and non-stolen wedding gifts, had already been completed due to some excellent paper pushing by government officials.  A top official made a formal apology to all in attendance who, at this time, were no longer dead but still asleep.  What began as a simple misunderstanding between the State and two families would soon boil into a stew of bad feelings when the father of the bride appeared to refuse the offer of forgiveness by declining the government officialâ€™s handshake.  Only reports of the non-death aspect of the case had been analyzed before the ill-fated apology took place.  Immediately, the clerk responsible for handling the paperwork was promoted and shipped across the ocean.  The day was declared a State holiday and all government employees were sent home with the day off with the exception of investigators assigned to the stolen property case, who were immediately paid overtime.  This however did not help uncover any clues regarding the size, shape, weight, or whereabouts of the stolen gift.  The previous owner of the gift was questioned but could provide no real insight due to the fact that he was asleep.  He was later thrown in prison for not cooperating with the State authorities.  '),(24348,4256976,'17,August,2004','2004-08-17 00:00:00','   The Inevitable End is the name given to the series of events which will bring about the demise of Winky and Sam but the subsequent saving the World.  A note on the World:  It exists in all time so nothing is anachronistic.  The events, vaguely prophesized by the Magnificent Cube are as follows:  Winky will fall in battle and Sam will slay The Chosen One.  This foretelling is unknown to our protagonists at this point.  A note on the foretelling: This one point may seem to fly in the face of the earlier note stating that the World exists in all time.  It doesnâ€™t.  Only events foretold by the Cube are unknown to the rest of the World until they eventually happen.  The magnitude of the Cubeâ€™s prophecies is such that each event marks a critical point in the history of the World.  The history of the World was known and catalogued at the beginning of the World by The Keeper of the Magnificent Cube.  Cube Events are so huge that all inhabitants are pretty much aware that a critical point in their history has been reached.  However, the Keeper would remark to the uninformed, â€œYeah, that was a Cube Event.â€  The Inevitable End, a mildly superfluous title after that explanation, is the final event foretold by the Cube.  The World, however, still goes on afterwards.  Sam, alone, walks the earth until his eventual unrecorded death and unattended funeral.  The Cube withers and dissolves.  The Keeper sends a copy of The History to each of the Twelve Libraries.  The World is well, good, and uneventful.  This is where we are going according to The Magnificent Cube.  A note on the Magnificent Cube: it is actually a sphere.  '),(24349,4256976,'17,August,2004','2004-08-17 00:00:00','   There are many characters that need to be explored.  It wouldnâ€™t be much help to list them all at the onset.  Probably the best way to learn the histories and futures of the doomed cast is to appreciate them individually at glances over an extended period of time.  But a few introductions are necessary.  What should be said upfront is that our protagonists, Winky and Sam, are compatriots and friends.  They share a destination.  They tread separate paths.  They drink, never beyond the point of hopeless desperation but occasionally enough for them to feed the animals before sleep.  To Sam, any story worth telling is worth exaggerating.  Winky prefers stripping a tale down to its most basic elements.  Sam searches for reactions.  Winky searches for truth.  Sam is drawn to the earth while Winky looks up to the sky.  But high or low, dirty or clean, happy or sad, their journeys to the same inevitable end are long and hard.  '),(24350,4256976,'17,August,2004','2004-08-17 00:00:00','   The fearless two are back from slumber in the murky depths with revelations and insight.  As before, the goal is neither to inspire society nor provide answers to even the most irrelevant questions.  If anything, the two would appreciate puzzling enjoyment, cautious curiosity, and a brief but kind word for all their efforts.  The scope of material may seem vast and foreboding but there will be help along the way.  The insight will occasionally be harsh but always forgiving.  So just like biting into a piece of over ripened fruit or hearing very bad experimental jazz for the first time, you probably wonâ€™t realize whatâ€™s wrong until itâ€™s too late.  '),(24351,808956,'26,August,2002','2002-08-26 00:00:00','   I have a problem. It doesn\'t matter how straightforward it is, or how overdue it is, I will pissfart around and make it take twice as long. the most notable is my Uni work. For 3 weeks now I have been trying to sit down and write an essay on Civil and Criminal Procedure. As thrilling as it is, when it comes to actually sitting down and typing my measly 1500 words, I will do everything in mw power, including writing in this blog, to evade \'real\' work (Some may question University as falling within the definition of real work, but I think they should bite me).  Maybe I am self-destructive. Actually, I don\'t think there\'s a \'maybe\' about it - I mean every single report card I got in primary school said \"Téa takes herself for granted\" or \"Téa does not work to the best of her ability\" or \"Téa keeps chewing her chair.\" That sort of thing. Similarly, in high school I got \"Téa does not work hard enough\" -or I\'d get an \"A\", but there was usually reluctance on the part of my teachers because of my so-called \"attitude\". I wonder if i\'ll ever change. What is strange is that I am sitting here writing about how I should be working - but no - I\'d never actually WORK. Well, I am, but I am taking my sweet time doing it.  And as much as I dont mean it to be conceited - I am lucky I am \"gifted\", otherwise I\'d be an abysmal failure. Luckily I can bluff my way through ;)  How I ever got to this point i\'ll never know. Back to work for me I think.   '),(24352,808956,'24,August,2002','2002-08-24 00:00:00','   The internet is indeed a strange place. Over the few days I have been diligently avoiding doing any actual work by visiting every link that is sent my inbox. Normally I just skip them, but hey, I am a girl on a mission ;)  First up, this one.  urlLink http://www.31337.pl/ . Its so simple, and yet it articulates what I think at least 15 times a day. I don\'t usually like elitism - but there are two kinds of people I find insufferable. Stupid people and, worse, Stupid people who try to disguise themselves as intelligent people. Now, that may sound like a horrible thing to say, but hey, further to my rant the other other day about AOL - there really are too many stupid people around with access to a keyboard... then again, many may say the same thing about me so I guess we\'re even :)  You might think I have a pessimistic view of people. I suppose in many ways I do - John Howard won the last federal election so its not entirely baseless - but I really do like people. Really, I do. Its just my standards are higher than most ;)  I have no idea what I was doing. Oh yes, I was sharing some funny URLs. hmmm What else did I find this week... (hunts through favourites)... (hunts through history because I havent added anything to my favourites)....ah stuffit.  I installed Cool Edit Pro 2 yesterday. It is SO much better than 1.2. Not that I ever use these things - I install them with all the best intentions and then never use them again. I like mucking around with music and I do write some things, but it never gets finished. I don\'t know if its because of a lack of confidence or a lack of ability (I am guessing its a combination of the two - you can have plenty of confidence and no ability and still be fine *grin*) but I just never finish it. I suppose i am doing the world a favour with that! LOL! Anyway, this post was more to bump the anti-American rant. While I can\'t bring myself to delete it, seeing as my ascii drawing of the American flag was second to none (heehee), I figured I\'d at least stop it from ebing the first thing you read when you enter the blog ;) I was expecting hate mail but got none - even though i\'ve had a fair few visitors (doesn\'t look like it on thepage though because I had to reset my counter)... its great to se that there are people as devoid of life as I am ;) Who is worse? the person who writes it or the person who actually reads it? :)  Well, thats enough insults from me today. To be honest, I am having troule finding things to rant about. Maybe some of my visitors could email me an article or two to get my back up ;)  In the meantime, have fun eh and i\'ll write back soon. I think I left a stinky nappy in my daughters room by mistake.... dont worry, it hasnt been there any more than a week (GRIN)   '),(24353,808956,'18,August,2002','2002-08-18 00:00:00','   I spent most of the evening working on my Formula Feeding support site. itmay sound boring to non-mothers, but believe me, is it controversial. I don\'t know how to be a support group for formula feeding without getting hatemail from the militant breastfeeders - I mean those women are organised! They send out emails to their site members urging them to harass people like me! lol!  I definitely see a need for it. But did you know that you can\'t even get information about baby formula from the manufacturers? This is how much of a strangehold the breastfeeders have - that us formula feeding mothers can\'t even access information. I challenge people to find any info on Wyeth\'s site, Heinz\' website, Nestlé, and so on. Geez... So much for freedom of choice - you only have it if you choose the same as they do... its a pain in the arse if you ask me. Why can\'t you just say \"I like this formula.\" You can\'t. You have to have a disclaimer \"Breast is best....\" so that the militants won\'t come after you. And, even if you have this disclaimer, many still will flame you and have sit-ins on your front lawn! Women victimising and scrutinising other women\'s decisions is not what I would consider moving towards a better society. it is only when these people learn to lighten up and accept that free-thinking individuals are able to make different decisions to other free thinking individuals - and society can function in spite of it - even BECAUSE of it!  I have no idea how this turned into a rant about breastfeeders... it is late and I should be in bed. Yes. That explains it. I also really needed an excuse to test my right click thingo in my browser that lets me instantaneously update this thing. Its pretty cool :)  Thanks for reading :p   '),(24354,808956,'17,August,2002','2002-08-17 00:00:00','   OK....well this seems to be working now! Its weird how hard it is to write when you feel the pressure! I was browsing the net today and found a couple of sites that might be of interest. The first one is  urlLink www.bushwatch.com  which is dedicated to my favourite world leader (AHEM!) George W. Bush. there is a some great stuff there - Be sure to check out the Bushlexia section! There is also a great feminist magazine called \"Bitch\" magazine. It ha some insightful commentary - mostly dedicated to feminist responses to popular culture. There are some interesting reads there, anyway. The address is  urlLink www.bitchmagazine.com . What more can I say today? What I am thinking of doing with this is posting articles and then writing my own commentary on it. It might be a news article, a website, or something in the global media. Of course, being hideously lazy on the best of days (not to mention a myriad of assignments that are yet to be written...:)), I may not get here as often as I would like. However, Blogger makes it a little less cumbersome to update, so I may pop in fairly often. We\'ll see how it goes huh? :)  I suppose I should add a little disclaimer for the rest of my posts. These are my opinions only. While I like to think I am a fairly learned person, and generally do not say things unless there is some basis, I would like to say in advance that I do hold controversial opinions in many areas. There. You were warned. :)  I will not respond to hate mail. There are a few issues that i already know are likely to incite strong reactions in some people, and I will lay them on the table now. I am a social democrat through and through. I am a feminist. I am opposed to the \'War on Terror\', even though I have sympathy for the victims of 9/11 (or 11/9 where I come from heh). I formula feed my child. I am not religious and I believe that \'god\" has been used far too many times as an excuse for oppression. In other words: Crazy \"militant\" Americans will not only be ignored, they will be killfiled and reported to their ISP. Same goes for abusive people, militant breastfeeding organisations that suppress choice, fundamentalist ANYTHINGS that preach, whatever. However, if you do email me in a hateful way, be prepared for it to show up here and for me to write a reply.  I am sorry if this appears harsh, but I needed to put this in. I have already received enough loony emails for a lifetime! :)  With all that nasty stuff out of the way, I do, however, welcome polite debate! It is what I do! Even if you hold different/opposing views to my own, I am happy to defend my position in a polite and respectful way. :)  Well... now that the housekeeping is out of the way...lets get on with eh? :)   '),(24355,808956,'17,August,2002','2002-08-17 00:00:00','   OK, I have created this blogger account to make it easier to edit my random thoughts. Lets see how long it takes for me to stop updating ;)   '),(24356,808956,'04,September,2002','2002-09-04 00:00:00','   Why is it that people only feel good about themselves if they are making another human being feel like a piece of rotten roadkill? I admit it, I do the same thing - sometimes I seem to get great satisfaction out of other people\'s weaknesses. It is most certainly arrogance and cruelty on my part, fuelled by a hideous lack of true self esteem - I admit it. But its not like everyone else doesn\'t. The proof of this is in the myriad of gossip magazines, usenet groups, mailing lists, websites, newspapers, and even people idle chatter - that is devoted to celebrity gossip.  We post to gossip channels, read magazines, join email lists devoted to celebrity gossip. We say scathing things, we scrutinise their every move - from outfits to opinions. We blame them for being too skinny, and yet all hell breaks loose when they are too \"fat\" (obviously a relative term). We praise them for their \'wonderful\' bodies, only to find out they are bulimic 6 months down the line. We want to know who the father of Liz Hurley\'s baby is, but not only that, we accuse her for being some sort of money-grubbing trickster. We mock Britney Spears (I\'ll get to that later). We speculate on people\'s marriages, and, the worst part - we believe the outright lies that are presented to us. Celebrity Gossip cuts across class, gender (yes, believe it or not!) and race. Why?  According to most psychologists, humans need to feel good about themselves and their position in society to be productive. And this is where I believe Celebrity Gossip fits in. Celebrities are beautiful. They are wealthy. They can do anything they want, because money and beauty and a lack of \'connections\' are not barriers, like they are for most of us. They get to be spokespeople for the United Nations and adopt children from overseas. In other words, they get to do all the wonderful things that all of us would do if we had the same money and profile. How on earth are we supposed to feel good about ourselves while these people exist? I know I don\'t! In a world that worships beauty over brains, people like me are destined to the middle classes - and the only greatness we can hope to achieve would be if, in 3000 years from now, we are seen as \"thinkers\". Ha! I suppose if the Google archives count...  So, in comparison to celebrities, we all feel like worthless pieces of crap - cogs in the machine that make other people wealthy. Ben Elton refers to people like us as \"farties\" - people who aren\'t beautiful or wealthy. People who have impeccable manners - who you couldn\'t imagine doing a great big burp whilst sitting down to a microwave dinner. People that take home doggy bags from an expensive restaurant because it costs $120 for a meal and you can\'t bear to throw it away. People who have a gruesome birth experience and walk around in public with baby vomit on their shirt and are most certainly not beaming with full makeup and a perfect body when their baby is 2 weeks old. I could go on for hours. To us, celebrities represent our dreams, our aspirations and our regrets.They represent everything we wish we were and everything most of us are not - even if we tried our absolute best.  So what we do is we tear them to shreds to make us feel better about ourselves. We would be horrified if our husbands or family said \"oi, you\'re looking a bit fat today\" or \"good lord...what the fuck are you wearing?\". Also, if you were walking up the street and everyone was looking at you and pointing, how would you feel? Thats right, you\'d feel terrible. So why is it alright to do this to celebrities? When you post to usenet that scathing message, how do you know that they aren\'t reading everything you say? And, if you knew they were reading it, would it make you rethink?   Take Britney Spears for instance. Now, those who know me know that I am not a fan. I also don\'t dislike her at all - call me a casual \"hummer\" along to her music. My initial reaction to Britney Spears, is that she represents everything that is wrong with the music industry. I am someone with more formal music training and come from old-school that it is your ability to play an instrument or sing perfectly that makes you famous. Naive, I know. But, all the same, I appreciate singer/songwriters, like Lisa Loeb, Ben Folds, etc, where if you strip them down to just them and their instrument in an acoustic show, their prescence still knocks you over. I admit my biases there. And I emulate that music style - fine tuning the acoustic, solo performance so that everything else is just a bonus - it doesnt MAKE the musician - or make the song. So, I lay my biases on the table in that regard. But, Britney, if you ever wanted someone to give you some quality advice about where to go with your career, I\'m here! :)  Issues of musical integrity aside (and I like some BS songs....), I think that the personal attacks on her and the scrunity are uncalled for. Comments of late (in various places) have started to become vindictive, callous, and just plain false. I often wonder that if people knew she was hurt by all of this if they would stop saying such horrible things about her. I know I would, because I can\'t stand the though of my words hurting someone else, even if its a superstar that could buy and sell me on any day of the week! :)  Seriously though, how would we feel if we were put in that position? If I were Britney Spears (heh now THEREs a stretch for  anyones  imagination lol), I would seriously wonder if its all worth it - I mean, she\'s made enough money to take care of herself and her family for the rest of her life. If she invested a lot of the money into her own label and signed other artists, she could do wonders. I really would be wondering if its all worth it. Here she is, 20 years old, a multi millionaire, with a myriad of options available to her. There would have to come a time when you wonder if its worth it. I personally believe that Britney should learn to play an instrument and try to do it on her \"own\" if you will. But thats just me :) Little do we realise that eventually someone\'s got to crack. I know that I could only take so much before I said \"fuck it - I\'m rich\" and disappear out of the public eye.  And little do we realise that this could be happening. Do you REALLY want to be the person whose irascible comments caused someone to quit doing what they love? Because eventually it will happen. Celebrities have internet access too. And I think they read a lot more about themselves than they should. Eventually someone is going to crack. I for one don\'t want to be one of the people that makes someone feel so awful that they question their whole career.  Maybe you can live with that, but I can\'t.   '),(24357,808956,'03,September,2002','2002-09-03 00:00:00','   Don\'t talk to me in the morning. Don\'t even look at me.  I hate getting up at 6am. I hate the fact that going to bed at 9:30 (which is relatively early for me) and getting up at 6am still only means 8.5 hours sleep. This may seem trivial and self-indulgent (when isn\'t it?), but bear with me.  I was having a wonderful dream at 6 this morning - it involved all the great elements: Canada, the beach, shopping, Ben Folds, Deborah Mailman and me being 10kg lighter. I don\'t normally have good dreams, so I was quite perturbed at the incessant HONK HONK HONK that woke me up. And that wasn\'t my alarm clock. it was Jason telling me he was running late and that he\'d be putting Mina in the car. See, I have to drive him to work so I can have the car. And I hate it.  And there is only one thing I hate more than geting up at 6am. It is idle, trivial, chatter when I am still waking up. I got in the car and everything was quiet. I was sitting there with my arms folded and all was well - I\'d be awake by the time we got to Jason\'s work. Look, I love my husband dearly, and we know each well, but he just doesn\'t seem to get that I don\'t want to talk to anyone for about half an hour after I wake up. No amount of grunt-answering seems to deter him.  Example. For some reason the radio was tuned to 96fm. I have no idea why people actually tune into this mindless crap - the last thing I want to hear is \"alright, call in to us to tell us what you can cook to win tickets to XXX\" I\'m serious. No mental challenges there, except for the kitchen-illiterate. But sriously, 96% of the population can cook SOMETHING. And \'something\' was all we heard. Here\'s one brilliant response: \"uh hi.... I make spaghetti bolognese with chopped up onions, chopped up mushrooms and tomato sauce\". My god. How astounding. I always thought that having radio call-ins was a way to engage an aidience to come up with quality entertainment - I normally listen to Triple J and they seem to have it down pretty well. But, my theory is, that anyone who is calling breakfast radio at 6am is a bloody moron to begin with. Firstly, because they are one of those irritating people that \"love mornings\". Secondly, I don\'t trust them because they have time to call said stations. Now, when I have to get up for something, chances are I will work out exactly how long I need between waking up and  leaving, and get up at that time. Like this morning for instance. I crawled out of bed, put on my tracky dacks and got in the car. I have a serious mistrust of those that actually have time for anything other than breakfast, showers or driving. Thirdly, the fact that these people brains are working at that time of morning (or clearly arent, seeing that 96fm dialogue) shows that they are clearly aliens and not to be trusted.  Anyway, as if the inane drivel on  the radio wasnt enough - Jason starts talking about Iraq. Now, I consider myself a fairly active person in terms of politics, and I find it hard to turn down a good discussion. In fact, right now \"Iraq\" is one of my main interests. But it should come as no surprise to you all that, for half an hour a day, I couldn\'t give a flying fuck about it. All I want to do is go back to bed and dream about Canada and Ben Folds.And, for half an hour a day, I like being able to give the bird to the world without caring. Of course it does wonders for my relationship - you know, being snappy and all that. But I have always been the same so I suppose thats what he deserves for marrying me.  My daughter is the same as me. I think I have the only baby on the planet that quite happily goes to bed at 7:30 and will sleep until 10am given the opportunity. And she looks like me when she wakes up too. It srather funny - and scary - to see a mirror of yourself in the morning. Erm, without looking the mirror. Of course.  So, don\'t talk to me until 8am. I likely to tell you to go fuck yourself.   '),(24358,808956,'28,January,2003','2003-01-28 00:00:00','   Its been a big week for sports this week. The Superbowl, the Australian Open, the Cricket...the Geelong bay cycling classic. And of course, the FA Cup.  For those who are not aware, soccer riots are not particularly uncommon around the world. We often see it on the news: some club has lost, the fans start a riot, a few people get injured or killed, arrests are made and they move onto the next game, where yet another riot breaks out. It\'s not something that surprises me much anymore - seeing English soccer fans with all their dutch courage trampling some poor old lady wearing a Manchester scarf. Generally I just scoff and say that it is something that is unique to soccer fans. \"It\'s like religion, it\'s  weeird \" I say to myself.   Yesterday\'s Superbowl also induced riots when Oakland Raiders fans lost the game. They set fire to cars and a McDonald\'s restaurant because they were drunk and noone would serve them. Ok, I made that bit up. But they did actually set fire to the place. What is interesting here, however, is that the Raiders\' fans also rioted when they *won* the week before! It\'s rather funny when you think about it, but it also makes me wonder if people, firstly, use sport as an excuse to be utter melonheads, or whether they are so blinded by faith for their teams that they *honestly* believe that supporters of the opposing team are an \"enemy\". As someone who defines my interest in sport as \'passing\', and my knowledge of sporting events being limited to the 10 minutes on Chennel Ten news every night, it\'s something I simply cannot comprehend. Or can I?  Perhaps it gives us more insight into nationalism and how it can skew our perspective. \"Nation\" as a concept is something that is unclear in its meaning. It is not simply a matter of drawing a line on a map and calling it a nation - it is far more complex than that. It is both objective and subjective. In a formal sense, it is a cultural entity: people who speak the same language, have the same religion, have a shared history, and so on. However, it is commonly accepted that a nation, or a community, can also be defined subjectively by its members, and could be distinguished by a shared loyalty or affection in the form of patriotism. Why doesn\'t this apply to soccer teams? Or Raiders fans? Or the Lisa Loeb Fan Forum?  As I type, American, UK and Australian troops are headed for the gulf to fight a war that about 60% (and rising) of their people disagree with. There was an article in  the West Australian  a couple of days ago that says that 65% of people have no faith in their politicians with regard to environmental issues, and this is a common trend across the western world. Our governments are simply not listening to their people. Most of us just want good quality healthcare and education for our children, and reliable roads and a government that takes care of those unable to take care of themselves. And we simply aren\'t getting it. America spends hundreds of millions of dollars on the military, Australia has increased its funding in defence. And noone wants it. People\'s faith in their government is dwindling and they are feeling increasingly helpless as we are led into World War 3 by a maniac with a vendetta and a lust for crude.  Now you may wonder why this is related to soccer riots and the riots that took place at the Superbowl. Well, it occured to me the other day, as I heard the news about the Superbowl riot, that people have nothing to fight for anymore because their governments just won\'t listen. They have noone to believe in, noone to identify with and bind them as a nation. Our politicians are corrupt, our priests abuse their power. The rich are getting richer at the expense of the poor, and the poor are marginalised for objecting under the guise of \"free enterprise\". So instead, we turn to that last place where equality is not a dirty word: Sports. The one place where the \"level playing field\" is exactly that, and not another government buzzword to dupe the poor out of medicine. Where flags once flew there are now team scarves and hats. And where national anthems once rang, there are team songs. Where there were inspirational political speeches about how our world *should be*, there are backflips and bum patts over a goal.  Now, don\'t get me wrong. I am not at all trying to justify the behaviour of those hooligans that go and beat up other fans. If anything, it gives us some insight into the ugly side of human nature: that instead of enjoying a game together in an environment of equality and respect, they\'ll attempt to beat up the enemy, all the while disrespecting other people\'s environment and property to prove their point. Just like the Americans are quite prepared to create war, to the detriment of our environment and our people, to prove the point that they are the \"biggest\" in the world.  However, in the end, is it  really  worth it? At the end of the day all you have is a burnt car and a burnt McDonalds restaurant. And a few people dead. But it never actually solves the issue, does it?   '),(24359,808956,'23,January,2003','2003-01-23 00:00:00','   It has been a while since I posted anything here, but I will try to post more from now on.  There have been a few reasons why I wasn\'t as enthusiastic about keeping this blog going. The main one was simply that I was too afraid to write! That may sound silly, but I will give you a little back story.  I am a member of a Yahoo Group called \"September2001Babies\". Don\'t ask me why, but I joined it when I first found out I was pregnant with Mina and I made some good friends, so I stayed. Anyway, The list was predominantly American. When I say \"predominantly\", I mean that there were probably 3 members out of 90 that were outside of the United States, with one Canadian, one from Australia (ie me) and one from Finland (I think it was - one of those Nordic ones anyway). Anyway, most Australians and Canadians, and even a few Americans, look at things differently to the dominant conservative culture that exists, and even have a different sense of humour from the dominant culture that is White America. Me being the type of person that I am, I questioned the Americans more than I should, and probably challenged them more often than they would have liked me too. Most of the time they spent their time arguing about how not to upset each other with subject headers whilst upsetting people...go figure... until my Canadian friend posted pictures of her Halloween carvings of Bush and Osama Bin laden to the list. Well, needless to say,  the irony was not noted and it sparked a mass exodus from the list - of all the US \"Moms\" away from me and my Canadian friend. A few decent folks stayed, but the list basically died at the hands of one or two people who couldn\'t take a joke. This isn\'t the full story - there were a few other troublemakers and \"discussions\" that would usually end in me calling them all \"retarded\" or something, but fundamentally, I was consistently bullied for my opinions for nearly two years before the wenches left the group. Basically it ended in me believing that Americans live in a freaky  Sliders  episode and I was too scared to post anything controversial or even honest for fear of losing the few members that stuck around on the babies list. This wasn\'t a direct cause, but it was one of the reasons I became too afraid to post anything about what I was thinking for fear of being harassed again. In other words, I started to conform.  Fast forward to yesterday.  I saw Michael Moore\'s \"Bowling for Columbine\" and I had one of those rare moments where it felt really good to be \"me\". For months I have felt marginalised. In fact, I have felt marginalised for practically my whole life, having always been isolated or harassed for being \"gifted\" (whatever the heck that means). But these last few months, I have felt particularly isolated and down, because no matter where I turned - the mainstream media, my family, that email list that I talked about, even people at uni - there was this air of not only hatred and racism, but support for war, support for American dominant culture, and propaganda against Muslim communities around the world. I have also been confused about what I am supposed to do with my life, how I can use my diverse talents and interests to make a difference. And, not only that, I have been frustrated that noone would listen to me when I expressed an opinion that was different to the norm. So I just stopped. I stopped debating with my grandmother when she talked about \"those Muslims\". I stopped writing my blog and writing my book, and just writing in general. I stopped caring about me degree and worried more about finishing than learning. I stopped writing because, without my strong convictions and my strong sense of irony and humour, I had nothing to write about. I was so afraid of saying the wrong thing that my \'craft\' was suffering, and I was becoming everything I despised: Someone who censors themselves for the sake of appearances. I was becoming mainstream.  And then I watched Bowling for Columbine (www.michaelmoore.com). I watched Michael Moore making a difference - rubbing against the grain of American mainstream politics. And taking galant risks for what he believes in, through his craft of comedy and filmmaking. I watched him challenge Charlton Heston in his own home and present the facts about American Imperialism. I watched him take 2 teenage boys that had bullets lodged in their bodies after the Columbine shootings to Kmart. I watched as he used his profile and convictions to cause Kmart to phase out the sale of handgun and semi-automatic ammunition. And not only did I feel inspired, I felt vindicated.  And I had an epiphony. I realised that it is a GOOD thing that I am different to all those women on my list. I am not someone prepared to settle for a house in a gated community to convince myself that life is strawberries and cream, and that blacks are a threat and everyone outside the US is out to get you. I am not prepared to sacrifice my convictions for fear of what some brainwashed, Oprah-watching housewives are going to say, because I *am* better than them. They will hate me for saying it, but I am, because instead of spending my days playing peekaboo and doing laundry, I have decided to make a difference in this world through my writing. And anyone who attempt to scare me into submission will not succeed anymore, because I am a different person. I have realised that it takes strength to have opinions and it takes courage to express those opinions in face of opposition. I also realised that its okay to be myself, and being armed with education and writing ability puts me in a position of power in this world.  Now, while this may only be a weblog amongst millions, I have realised that there is a reason I was put on this earth - and that is to use my writing in the same way that Michael Moore has used filmmaking. It may sound arrogant, but I really don\'t care. Consider this a mission statement for the blog entries to come. I will be writing my book again; I will be writing newspaper articles and columns. And I will make a difference through my art.  I promise I won\'t make too many self-indulgent, Oprah-magazine editorials from now on. but I did need to explain why I disappeared. I look forward to feedback from people. Thanks for reading.   '),(24360,808956,'15,April,2003','2003-04-15 00:00:00','   As you all may have noticed, I haven\'t written here for some time. It wasn\'t because I didn\'t have anything to write about, but because there has just been so much, and I am just *so* lazy :). Where on earth do I begin? In January, the last time I posted here, the world was only talking about bombing Iraq to smitherines; now they are \"liberated\". In January the infamous Bashir interviews with Michael Jackson were only talked about and not seen; now, we are aware of what an absolute freak he is - and how thin a line there is between a documentary and tabloid journalism. In January 70% of people in Australia and in America opposed war on Iraq; now public opposition is less than half of that. Oh, and of course, \"freedom\" fries are the new side dish at McDonald\'s. Try swallowing those without gagging!  So, with such huge issues unfolding before our very eyes, where can I begin? What can I write about? All of these issues are rant-worthy in their own right. However, as unrelated as Michael Jackson may appear to be to the war on Iraq, there are elements that are eerily similar. And no, its not about both Saddam and Michael being weird recluses or even a joke about Michael Jackson hiring Muhammed Saeed al-Sahaf (the Iraqi Intelligence Minister) as his publicist, its about our obsession with entertainment: our ability, through the media, and as a society, to detach ourselves from tragedy and turn it into entertainment.  Watching Michael Jackson\'s interview gave me an uneasy feeling, and yet I couldn\'t put my finger on it at the time. At first I thought it was because I felt undecided about his \"involvement\" with children (and I am not prepared to speculate, as far as I am concerned he has not been found guilty by a jury) - or that I felt he was just such a sad, tragic figure that it broke my heart. But I realised that it wasn\'t any of that. It was the fact that this man had been propped up as entertainment for the millions of people that watched, even though he is clearly in need of help (what sort, I don\'t know, but I believe he\'s mentally ill). The spectacle that was made of his life and his problems, through this documentary, did what the media does best. It detached us from the very real tragedy of Michael Jackson\'s life. An arguably brilliant man who has been destroyed by his family and by the media. And what did people talk about after the interview aired? Not whether it was moral to interview him in the first place, nor whether it was responsible to portray Michael that way (in my opinion, \'trial by media\' should be a crime). They talked about - you guessed it - whether he raped some kid with cancer. Its this obsession with gossip and innuendo, and of course, celebrity, that desensitises us from the issues we should be thinking about. The issues that really should matter, such as media morality, or the concept of \"celebrity\".  I don\'t think I will ever forget how I felt the morning (Australian time) that the first bombs were dropped on Baghdad. All of our television networks (with the exception of the ABC) brought live footage, uttering buzzwords such as \"shock and awe\", \"dazzling display\", and \"light show over Baghdad\" - like it was some sort of pretty fireworks display for all the loyal viewers. The 24 hour coverage, complete with sexy footage and advertising revenue, reiterated and repeated the same information over again. It had a two-tiered effect: Firstly, it perpetuated that culture of fear that has dominated America for the better part of this century, and secondly, it reduced the war into yet another ratings bonanza. For those that supported the war it was great seeing the \"fireworks\" around Baghdad, and journalists reporting with gasmasks, struggling to breathe. It was, first and foremost, quality entertainment. People were being fed simplistic ideas in a simple way - and we wonder why they now support the war? Of course, because we don\'t get to see the hundreds (possibly thousands) of Iraqi women\'s bodies, impaled, limbless, clutching the corpses of their newborn babies. We don\'t see Iraqi children who are in hospital with shrapnel wounds. No, that would be too confronting for prime time. But that, my friends, is the reality of war. The reality that we are not seeing, because we only see what \"they\" want us to see. They want us to see advertising. They want us to keep consuming. They don\'t want us to see the gritty reality of war, nor do they want to see real debate about the legality or the moral justification for it. So they trivialise it. Make it pretty. Make it seem like a distant problem, that is only happening to \"them\" in Iraq. Well, unfortunately, the more we detach ourselves from tragedy, the more we create a real problem in our society, because we allow our leaders to get away with brainwashing, greed and corruption.  And then, instead of actually debating the issues, we do idiotic things like renaming French Fries to \"Freedom Fries\". They\'ve really done their job well when you think about it.   '),(24361,808956,'23,July,2003','2003-07-23 00:00:00','   The Matrix versus the real world and soggy sandwiches..    Golly gosh, where do I begin? I haven\'t written here for what - 2 months now? Despite my grand promises of updating every week, life inevitably gets in the way. And, whenever I do go to update, my Blog ideas end up being out of date.     Let me fill you all in on the brilliant ideas I have had for Blogs in the last few months. There was that brilliant piece I had mapped out in my head that was meant to coincide with the release of Matrix: Reloaded, intertwining global media corporations and Matrix philosophy. Then there was the brilliant piece about Michael Jackson, and that bit about why I actively avoid Harry Potter and want to strangle Avril Lavigne, or... ok, I think you get the point. I had a lot of really good Blog ideas and I never wrote any of them.     Why? You ask? Well, For the first time ever, Téa has ventured into the big wide world of the 9-5. The hustle and bustle of peak hour traffic and the smell of soggy homemade lunches. That\'s right - I got myself a job.     For years, people have been telling me that I should join the real world, and escape the insular, protective shell of University life. Why they thought I wasn\'t in the real world is beyond me - I think that cleaning grease-catches and making souvlaki for intoxicated accountants for $4 an hour is about as real-world as you can get. I suppose its more a strategy for fobbing off intelligent conversation. But anyway. I have, for all intents and purposes, joined the \"real world\", where people don\'t have time to write humourous (apparently) columns, troll usenet or talk about insignificant crap with American feral mothers. I have entered a world where no matter how creative, no matter how brilliant, or no matter how committed we are to surfing for porn, we all do the same thing: flop on the couch at 6pm and go to sleep by 9:30.    That\'s right. I have been eaten by the work monster. That beast that usurps all my creative energy, and lets my hard earned finger callouses from years of guitar playing go soft. It makes my webpage feel neglected and my Yahoo! groups have a gaping great hole where there was once insignificant diatribe. It leaves my IM windows unanswered, my phone messages ignored, and that 20 pages of the novel I started go unwritten. I used to iron clothes once every 2 weeks or so. Now I iron every day. And I wear stockings. I hate stockings. I always said that there was one way to know you were officially a grownup - and that is that you wear stockings more than once a month.    You may be thinking right about now \"gawd, you\'d think she was the only person to ever start working full time!\" Don\'t get me wrong, as far as jobs go, I have a good one, but sometimes I don\'t feel there is a lot of scope for Uni graduates in particular to be able to adjust. When you are someone like me who is particularly creative, and actually has physical withdrawals when I am not writing, or drawing, or playing music, it is a major culture shock going from University to full time work. First of all, University was invented so that people could sit around, smoke pot and think about life. It allows self-expression, creativity, and exploration of all things magical in this world, like philosophy, art, and popular culture (It also enables a lot of TV viewing, such as Springer, Bold and the Beautiful and Passions, but that is another story altogether!). Secondly, there is all this free time. Sure, you\'d need to get up for that \"early\" lecture at 9am once a week, but for the most part, Uni work was on your own terms and to your own schedule. Then, the unwitting graduate walks into a job where there is structure and process and post it notes and a computer that actually uses the Calendar and Tasks in Outlook. And suddenly 11 hours a day is gone. And then whammo - all of a sudden you are consumed by paper and phone calls and process and paper - and before you know it, you have neglected your web pages, your music and your writing, and bit by bit you lose a little piece of yourself that you\'ll never get back.    Maybe I was just never really cut out for the 9-5 in a strict sense. I don\'t have patience for detail, and worst of all, I really get stifled and bored very quickly if I am not creatively stimulated (shaddup dirty minded people). I have been seriously contemplating the idea of doing a Dip Ed in 2004, then heading out to a country school to teach in 2005. I suppose everyone idealises certain career paths, but my whole life I have been drawn to teaching. And especially being in an office environment (even though that particular environment is fantastic), I don\'t know if I am really cut out for it. I need to be constantly challenged, moved and entertained - and I have a strong desire to make a difference in people\'s lives. And, I don\'t have to wear stockings.    Maybe I just belong in an education environment for the rest of my life. I am so happy when I am teaching others or learning that I suppose its the natural choice. But then that would mean I\'d need to escape the \"real world\".    Is that such a bad thing? After all, 99.9% of people in the Matrix live in blissful ignorance.The Matrix Versus The Real World and Soggy Sandwiches...   '),(24362,808956,'27,August,2003','2003-08-27 00:00:00','   I am sorry I haven\'t updated much (yet again), but I have been busy with uni work.  However, I promise better quality posting from here on in. As you can see, I have implemented a book reviews page - to share some of my favourite books with you.  I am putting an Amazon link up here, just to see if I can get any help funding my book addiction. If you\'re interested in buying the book, by all means feel free to buy them using my affiliate number: All it does is give me some sort of discount on the books I buy all the time from Amazon. So any help would be great.  That\'s not me trying to make money off my site at all - I just figured that if I was going to link to Amazon I might as well get those gift vouchers ;)  The first book is \"Fast Food Nation\" by Eric Schlosser. I wont be cross-promoting my areas at all - this is the last time :)  So, even if you don\'t see a lot of updates from me in the blog department - I do try my best - but, unfortunately when your 3000 word essay on Elite theory in Australia beckons, it is hard to find an excuse to update when ...well...you know how it is.  Anyway, I hope that there are still people visiting this site. lol.   '),(24363,808956,'11,September,2003','2003-09-11 00:00:00','   Well, today marks the second anniversary of the September 11th attacks. I bet you weren\'t aware of that, seeing as the media so often underplays that footage...and there is practically no commemorative spam...lol.  I just got a CNN Breaking News message that says \"[the] State Department warns of \"increased indications\" that al Qaeda is preparing attacks on U.S. interests to coincide with 9/11 anniversary.\" What the hell does \"increased indications\" mean? How much longer is the government going to keep Americans in a state of fear?  Excuse me if I rant a little bit, but it is fear of EVERYTHING that makes yanks so utterly irrational. They are afraid of everything that might happen - and nothing that *does* happen.  For example, did you know that 38% of children in America, or 27 million kids, live in a low income family? Of those, over 12 million live in poverty, and over 5 million live in extreme poverty. That\'s more alarming than \'increased indications\' of an al Qaeda attack.  Did you know that heart disease is America\'s biggest killer? Did you know that lung cancer is the second? And did you know that the number one disease affecting children is asthma? Did you know that there is a causal link between air pollution and all of these chronic diseases? And did you know that Bush and Cheney have an interest in companies that are responsible for massive amounts of pollution?  Oil companies are the largest industrial source of volatile organic compounds (which lead to ozone smog), the second largest industrial source of sulfur dioxide (which is transformed into dangerous fine particles and acid rain) and the third largest industrial source of nitrogen oxides (another smog- forming pollutant). Oil refineries also emit large amounts of toxic air pollutants -- more than 58 million pounds in 1997.  Healthcare funding has been cut, which makes things even worse for those who suffer illnesses as a result of pollution. Why aren\'t people angry or scared about this?  Military spending is at an all time high. It has been legitimised by the periodic releases of terror alerts, each getting more and more vague. Keep them scared...keep them frightened about things that will be unlikely to affect them...so that they don\'t have time to think about the things that do.  \"Increased indications\" of a terrorist attack. Just look at that phrase. That\'s the BEST they can do. Coincidentally, it came out just as people were starting to doubt the success of the war on Iraq. Just as Bush\' popularity is declining. Just as domestic issues, like the environment and health care, are slowly making their way back onto the agenda...what happens? The anniversary of September 11. That hotbed of grief, injustice and terror that is still fresh in most people\'s minds.  But, we need to put it in perspective. It happened. It was horrible. And it does represent people\'s worst fears. However, it gets to the point where the government starts to use it as a tool to manipulate the public. They manipulate the grief, anxiety and shock to further their own morally corrupt agendas. And, worse than that, to criticise them means that you are un-American, or anti-American (as I have been called a fair few times *g*).  I am not saying \'get over it\' to Americans. But I am saying that you shouldn\'t let a group of corrupt, wealthy, and very greedy men use the memory to manipulate public opinion and further their own agendas.  Start holding them accountable. Demand quality healthcare. Demand clean air. Sure, there is a slightly increased risk of a terrorist attack, and yes, someone might kill you. That is, if the air you breathe doesn\'t kill you first.  We need to get our fears into perspective. We need to realise that being hit by a bus is more likely than being on a bus with a suicide bomber. And that you are more likely to die from the\"food\" you ingest than from anthrax.  And most of all, you need to get that oaf out of power before we\'re all wiped off the face of the earth   '),(24364,808956,'04,November,2003','2003-11-04 00:00:00','   I don\'t know if any of you have been following the debate lately about higher education reform, but it is an issue that I am particularly interested in. For the un-initiated, the main reforms involve allowing school leavers who do not get the marks to enter University to pay their way into a course. I am interested in this issue not only as someone who has benefited from accessible education, but I am also opposed to the reforms because it is unfair. Plain and simple. Let me tell you why.  The higher education system in America is a two tiered system, similar to the one proposed by the Liberal government. Wealthy children with not-so-great marks get into Harvard, Princeton, etc, and the rest either struggle their way through the tens of thousands of dollars in tuition fees, or take athletic scholarships to subsidise their education. Then, you have the countless young people who try to get an education by joining the military, like that poor, poor \"Shock and Awe\" poster-girl, Jessica Lynch.  Like many other young working class (and increasingly, middle class) Americans, Jessica Lynch joined the military because she wanted a college degree, so that she could eventually become a teacher. Now, whether you agree with war or not, whether you believe the exaggerated stories of her \'rescue\' or not, or even whether you agree with military service for our young or not - one thing remains abundantly clear. There are kids in America that are prepared to risk permanent disfigurement and even death so that they can receive an education. Surely, that can\'t be right. Surely, if you want an education in America you can have one? Right? Right...?  Now, let us take a look at the person that sent Jessica Lynch and thousands of other teenagers to the war in Iraq. The walking answer to the question \"what is wrong with the United States?\" - George W. Bush.  Did you know that George Junior has a degree in History from Yale? Did you know that he also holds an MBA from Harvard Business School? Pretty amazing feats for a man that referred to the \"Chinesians\". I guess I am amazed, primarily because I always thought that Universities like Harvard and Yale had some degree of academic integrity - where it most certainly helped to be rich, but it wasn\'t the overriding factor in admissions.  I always thought that Harvard required a particularly high standard for entry, including a resume of real world experience, a strong grade-point average, and a \'reputable\' degree.   But, alas, I have been proven wrong, because it seems that baby George didn\'t need it because 1) his daddy was Chairman of the Republican National Committee and 2) They had shit loads of money. Sure, he had a degree from a reputable University...but honestly, do we *really* believe he has studied History? Maybe he\'s confusing an American History unit for a History degree... alas...back to my point.  We don\'t know Jessica Lynch\'s academic history - for all we know she could have been a bare-pass student. She could also be very bright. However, even if she is only average, what differentiates one bear-pass student like Lynch, from another like George W Bush?   The Moula.  And its wrong.  There is something inherently unfair about a stupid man with a lot of money, coasting his way through life, having access to all the \"best\" schools, and others who are not necessarily outstanding, but also not rich, being sent off to war so that they can receive an education.  And this, my friends, brings to my point about user-pays in higher education reforms in Australia. If you allow universities to charge upfront fees to people who cannot get in on their own merits, you take one step closer towards a two-tiered education system. On one hand, you have students who work hard to get a 65%+ average to get into University. They will graduate with massive amounts of debt and get squeezed out of the more expensive courses, and as a result, will never escape the working class trap. Then, you\'ll have kids who have never worked an honest day in their lives - coasting into an LLB with a 60% average and a fistful of dollars, who will graduate into easy jobs with friends-of-friends, and maintain their hold on society by forming the elite.  And then, before you know it, we\'ll no longer be saying \"don\'t let it go the way of America\", because by then we\'ll already be there. And by then, we\'ll be sending our working class kids to fight in wars, getting themselves maimed and killed in order to get an education.  Higher education \"reform\" is nothing but another attempt by John Howard to turn us into the 51st state of America. Its times like this I feel an awful lot like a Canadian.   '),(24365,808956,'01,January,2004','2004-01-01 00:00:00','   First off, I would like to wish everyone a Happy and Prosperous 2004. I, for one, hope to have a better year than I did last year: I guess that\'s really all we can hope for. Not dying and maybe trying to improve our lives a little bit from year to year. Not that I am really keen on the whole resolution thing, but still...  It\'s been an eventful few weeks, what with earthquakes all over the place, Michael Jackson\'s arrest, and Steve Irwin being investigated for mistaking his newborn son for crocodile bait. But instead, me being the narcissist that I am, I am going to talk about myself.  I learned a lot in 2003. I could focus on your typical Oprah-remembering-your-spirit crap, or I could talk about the things that I was rather surprised to learn about in 2003. You see, despite being fairly widely read, and despite being a moderately intelligent human being, there are a few things that I have naively taken for granted. You know, the silly, trivial things that you have never really questioned until you are surprised to learn how stupid you actually are. Here goes.  1. \"Grease\" has a lot of sexual references.  I had no idea.  I must\'ve seen that movie about 40 times throughout my childhood: I knew the movie off by heart, I sang with it, dreamed I was Olivia Newton John singing \"Hopelessly Devoted to You\" whilst kneeling in the front yard (The neighbors thought I was a very strange kid).  Then I watched it not so long ago and heard words that I had never known about. All of a sudden Rizzo was talking about gang bangs and all other sorts of stuff that I had no idea about! I was horrified!  2. There is no such thing as an anti-gravity chamber.  Ok, I am sure the nerd-boys that love physics will snigger at me and say \"what an idiot\", but I really thought that there was some such thing as an anti-gravity chmber where you could just go and float around. Apparently they don\'t exist. Apparently there\'s something called gravitational pull on earth and you can\'t get around it.  3. Stealth bombers do not have special invisible paint.   My husband was quick to point out this one to me. I thought that the diamond-type design was there to make it look cool and futuristic, and the steathy-ness was made by special paint that was undetectable by radar. Okay, so maybe I deserve to be laughed out of the room for that one, but hey, unless you have sought out the information, how are you supposed to know? The only reason it came up was because of some documentary on the stealth bomber came on the Discovery Channel. Until then my husband thought he was married to someone with a brain. Boy, was he wrong.  4. Pencils are not made by drilling a hole down the middle.  I learned this from Play School of all places (a kids show, aimed at 3-5 year olds), where they went to a pencil factory. I always thought that pencils were made from really good quality sticks, and you just drilled a hole down the middle and put the leads in that way. Apparently that is *not* how it is done. Maybe if I had been allowed to watch television as a kid I might know this!  So, there you have it. Four things that I have learned in 2003. Who knows, maybe in 2004 I\'ll learn that my favourite childhood singer is a paedophile.  The thing about the New Year is not to make grand resolutions about changing your life. You can only do that if you are truly committed to it for the long term. But you can look back at the previous year and realise that at least you\'re just that little bit less stupid than you used to be   '),(24366,808956,'28,April,2004','2004-04-28 00:00:00','   urlLink Write This! : an excellent creative writing exercise tool that gives random words and random situations to write around within a set timeframe. I have already written a few stories that I would have never thought of otherwise. Check it out!   '),(24367,808956,'26,April,2004','2004-04-26 00:00:00','   I have discovered WindowBlinds, a cool little utility that lets you skin XP. Being a Mac OSX wannabe, I thought it was only natural that I skin my XP with OSX.  urlLink This  is what it looks like.   '),(24368,808956,'05,April,2004','2004-04-05 00:00:00','   Not sure if anyone is interested in reading this, but I thought I would share an essay I wrote on Elite theory. It still needs some work (I got a bit lazy around the ending), but it might provoke someone\'s interest :)  --  It has been said that pluralism is one of the main characteristics of a thriving liberal democracy, and it has been long accepted that Australia is, indeed, a pluralist democracy (Henderson, 1990, p.201). A manifestation of a truly pluralist democracy is the existence of interest groups that compete for government attention, and actively try to influence government policy.  Pluralism refers to “…a form of democracy that operates through the capacity of organised groups and interests to articulate popular demands and ensure government responsiveness.” (Heywood, 1997, p.77) This has been a long-held ideal for Australia: the notion of a ‘fair go’ that manifests in all interest groups having a chance to influence government - irrespective of wealth, size or the political status of those being represented. However, there has also been significant criticism of this opinion, arguing that despite the ideals of pluralism, Australia more closely fits the ‘elitist model’ – in that some groups are the most influential in the Australian system (Henderson, 1990, p.201). Elite theory is defined in many different ways, depending on whether it has been seen as desirable, undesirable or simply as inevitability (Heywood, 1997, p.78). Modern elite theory, however, has been generally associated with American thinkers such as C. Wright Mills and Thomas Dye. It refers to “…a minority in whose hands power, wealth or privilege is concentrated.” In other words, a highly-concentrated number of ‘elites’ control most of the wealth, most of the power and most of the privilege in society – and use this wealth, power and privilege to influence the decisions of government.   According to C. Wright Mills, the ‘elite’ in America consists of three areas: the highest political leaders including the president and a handful of key cabinet members and close advisers; major corporate owners and directors; and high-ranking military officers (Mills, 1956). The media can also considered to be a recent addition to the elites. The criteria for being considered part of the elite depends on a number of factors: wealth (for example corporations), the number of people they represent (in the case of churches) or status (as in the military, or the media). These factors tend to overlap, for example, churches are wealthy, represent a large number of people but also have significant ‘status’ as moral leaders; the media are also wealthy, represent a large number of people, and hold significant status in society.  Government (at least, members who are not part of cabinet) is generally not considered a part of the elite, but merely a mechanism for the elite to exercise its collective interests, in this case, through interest groups. Of course, there has been much debate as to whether the elite do actually have a ‘collective interest’; however, for the purposes of this essay it is assumed that the wealthy would like to stay wealthy, and that the powerful would like to stay powerful, and with that comes a collective set of goals – including deregulation, lower taxes, maintenance of the status-quo, and other policies that directly benefit the ‘elite’ class.  How does this apply to interest groups and their influence? Davis, Wanna, Warhurst and Weller (1993, p.27) state that “access to government [by interest groups] is unequal. Some groups are so close to government that they can be regarded as ‘insiders’.” The question to ask is whether the same interest groups that form the ‘elite’ (those with the most money, size and status in society), are the same groups that form the inner circle of government influence. A secondary question would be: are we a pluralist or an elitist democracy? The answer is money, size and status do impact on a group’s ability to influence government, and particularly to enter that inner circle of ‘insiders’.   However, the Australian system is not entirely elitist. Our decentralised, federal system of government serves smaller interest groups that do not have a lot of money, size, or status, because senators, state governments and local councils are still able to effect change, and are generally more accessible than federal cabinet members. Most of this essay focuses on federal government policy decisions, however, this can equally apply to state governments (particularly Premiers and cabinet members).   Money   The amount of resources an interest group has access to has a significant impact on their ability to effectively lobby governments. According to Davis et al (1993, p.148):  “…making representations…is an expensive business. It requires expertise in the field and an understanding of the government’s policy options. It also requires a knowledge of the structure and processes of government so that leverage may be applied in the appropriate place and at the appropriate time.”  This means that in order to effectively lobby the federal government, interest groups need to have access to a high level of expertise in order to get their concerns heard. Further, this requires the interest groups to hire high-level staff, or to utilise an independent, professional lobbyist. Along with the structural expenses of maintaining an organisation, such as offices, staff and computers, marketing and advertising campaigns can run into millions of dollars.   Of course, there are cheaper alternatives to mass media campaigns and professional lobbyists. According to Davis et al (1993, p.154), “organising a protest is the traditional technique for demonstrating support.” However, these are only most effective when they represent a significant proportion of the population. We need look no further than the anti-war protests this year, which had numbers in the tens of thousands in each capital city. This is an example of a well organised campaign that represented the opinions of a significant number of Australians. However, according to Davis et al (1992, p.155), there are many groups that cannot organise in sufficient numbers to get support, such as the unemployed, Aborigines, the homeless and the poor, because they represent a minority opinion.  So, those groups that represent a sectional interest or minority opinion often have no choice but to organise a formal interest group, to lobby government and attempt to change government policy. Alternatively, they can try to change public opinion through media campaigns and advertising. Both of these options are expensive.   Another option for interest groups is to contribute to election campaigns of parties that support their aims. For example, tobacco companies have a long history of making donations to both the Liberal Party and the Australian Labor Party in order to influence their policy-making decisions. A leaked Philip Morris Corporate Affairs Plan “…revealed that at least one tobacco company set aside $100,000 a month to target politicians in Australia.” (Action on Smoking and Health Australia, 2001). According to an article in the Sydney Morning Herald (Seccombe, 2003), Australian political parties received donations of $142 million in the 2001-2002 financial year. In June 2001, the NSW Liberal Party’s fundraiser, Michael Yabsley, admitted collecting $3.5 million in 16 months from major corporations, using a Liberal party fundraising organisation called “Millennium Forum” – that not only does not need to disclose donations, but offers an opportunity to talk directly with the Prime Minister after a donation is made (Browning, 2001).   Some organisations have much greater access to senior bureaucracy and government than others (Davis et al, 1993, p.150). A study of advisory committees for the Royal Commission on Commonwealth Government Administration (1976, cited in Davis et al, p.150) stated that ‘more than four-fifths of all groups represented on advisory committees were producer groups: and of these almost half were business associations.” Even though business does not speak with one voice to government and there are several different groups with different aims and influence (Beresford, 2000, p. 121), some groups wield significant influence over government. The Business Council of Australia (BCA) “…has an operating budget of over $5 million and a staff of 16, many of whom are former senior government officials.” (Beresford, 2000, p.121). However, it does not end there. According to Singleton (Cited in Beresford, 2000, p.121-122), many influential business organisations were selected to sit on various committees and forums, and were represented by government departments. In the Kennett Victorian government, there have been links between the awarding of un-tendered contracts and Liberal Party donations (Steketee 1995, p.9).  How do those interest groups that represent the poorer members of our society fare against the interests of the elites? The relationship between governments and these sorts of groups is generally volatile and dependent on the groups’ position on any given issue. For example, despite the ACTU being considered an ‘insider’ under the Hawke and Keating governments, after the election of John Howard, the ACTU became marginalised and undermined (Beresford, 2000, p.123). The same has been said about ACOSS, the peak organisation representing approximately 250 community welfare associations. According to Beresford (2000, p.124), ACOSS was drawn into the policy-making process, ‘…but the relationship was not always an easy one…due to the differences it had with the Hawke/Keating governments…” Despite Gruen and Gratten (1993, cited in Beresford, 2000, p.124)) considering the relationship productive as a whole, it does show the difficulties that can be faced with organisations that represent interests that are not in tune with business. In contrast to business-related interest groups, that generally transcend governmental change (Beresford, 2000, p.123), interest groups that lack economic clout seem to be ‘insiders’ only at the discretion of the government.  This implies that business groups do use their economic clout to influence government – by threatening job losses, or through more covert measures like party donations. This is consistent with Elite theory: the corporations that form the ‘elite’ use their money to persuade governments. The evidence also suggests that interest groups with less money, and those that do not make generous donations to political parties, tend to get fewer opportunities to impact on public policy.    Size   Interest groups often claim to represent a ‘silent majority’, or to represent a significant proportion of the population. It generally goes without saying that government is more likely to pay attention to a group that represents 50% of the population over one that represents 2 or 3% - because to lose even half of those 50% of voters can be disastrous for any government. It is no coincidence, therefore, that the majority of ‘insider’ groups are large, national organisations. Davis et al (1993, p.142) state that “national representation has become the key to participate in tripartite agreements and to having a voice in Canberra.” Because of this, smaller interest groups often merge with other groups with similar aims, in order to better represent their interests to government. For example, ACOSS represents over 2500 community welfare organisations (Beresford 124), and ACTU represents the majority of trade unions at the federal level.  With greater size also comes more access to money and resources (Davis et al, 1993, p.143), including the ability to hire professional lobbyists, and the ability to make donations to political parties, and the ability to fund media campaigns. The ACTU, officially representing 12% of the workforce in 2002, have been influential purely because of their ability to utilise their numbers. The ACTU was central in the formulation of the Accord in the 1980’s (Smith, 2000, p. 287). The sheer size and influence of the unions over ALP policy (they have a 60:40 ratio at the National Conference), meant that the ACTU had remarkable sway in areas like industrial relations and income policy.   The size of an interest group and the amount of money it has are, for the most part, interrelated. Massive corporations like Kerry Packer’s PBL and Rubert Murdoch’s Newscorp, are both incredibly wealthy and own a significant amount of property and businesses. Not only that, they can also wield tremendous influence over the electorate (Smith, 2000, p.333). The Catholic and Anglican churches own billions of dollars in assets, and together represent 45.9% of the population (Australian Bureau of Statistics, 2001). These organisations can also have a tremendous influence over the population, because of their high moral status. Both of these very large interest groups have significant influence of the government, and can most certainly be considered ‘insiders’.  For an interest group that is not wealthy but large, however, there seems to be less of a tendency for them to be able to influence public policy. ACOSS, for example, represent the poorest people in the country, despite representing a large number of organisations. Similarly, Veterans’ organisations, consumer groups and Aged Care organisations represent significant numbers of people that are underrepresented in Australian Politics. In a study of federal Department of Health advisory committees (Australian Consumers Association et al, cited in Davis et al, 1993, p.150), “…professional organisations had about 220 representatives on 65 different committees, [yet] consumer or community organisations only had 30 representatives on 23 different committees.”  Does size in itself improve an interest group’s ability to influence government? The evidence seems to show that, even though groups that are nationally organised and represent larger groups have a better chance at influencing government policy, it is only those groups with money as well as size that form the group of ‘insiders’ in government policy. Even the biggest interest groups that represent poorer citizens are usually outsiders.  This is consistent with Elite theory in that there is a select group of large, wealthy organisations that are getting greater access to government, and, in turn having greater influence over government decisions. The key point in Elite theory, particularly when it comes to interest groups, is that money, size and status on their own do not mean that a group will be considered insiders. It is often those groups that have wealth, size and status together that form that inner circle.   Status   An important factor that determines how much government attention a group receives is the status of its members. ‘Status’ can be social, economic or political, and refers to the level of societal status the group represents. For example, doctors, lawyers and even teachers are considered to have a high level of social status, and this manifests in how the government responds to their requests. The Australian Medical Association, for example, is one of the more visible professional associations, and its position as an ‘insider’ has generally transcended governmental change. According to Beresford (2001, p.124), the AMA has a “multi-million dollar budget and is actively involved in liaison with government through membership of official advisory councils.” They are usually the first group approached with regard to health policy changes, and a stamp-of-approval from the AMA is highly sought after by the political parties. Upon her appointment to the Health Ministry, Kay Patterson made it a priority to forge a close working relationship with the AMA (Fonesca, 2001). During the 2001 election, the AMA actively campaigned on a single issue:   “On health, AMA president Dr Kerryn Phelps has already begun tours of marginal seats, and will campaign for more funds for public hospitals, more aged care places, indigenous health, and higher Medicare rebates for doctors. The AMA will announce its preferred health policy after both parties release costed policies.”  (Kingston, 2001)  Another interest group with a very high level of social status is the churches, and more specifically, Christian churches. In the Australian Bureau of Statistic Census data for 2001, 68% of people answered that they were of the Christian faith, with 45.9% of those being Anglican or Catholic. Those that advocate church activism, including James Murray (cited in Beresford, p125), argue that the churches are “…representative of society at large and deserve to be heard…[and] have a vast amount of experience in social areas” However, are they truly representative of society at large? Despite the 2001 Census results showing 68% of people identified with a Christian church, statistics from previous years show an increasingly secular society. ABS comparisons show that in 1901, 95.9% of the population were Christian. In 2001, 68% were Christian, 1.9% were Buddhist, 1.5% were Islam, and 25.3% stated they had ‘no religion’ or refused to answer the question. (ABS 2001). So, despite the status of the church still being an influence, one could argue that as their numbers decline, so does the social status that accompanies it. This decrease in social status does affect the churches’ ability to influence policy, and even their ability to remain as ‘insiders’. Tony Abbott has frequently questioned the involvement of churches in policy decisions, and has called it a “misuse of the prestige of the church.”  On the flipside, groups with very little social, political or economic status are generally not represented in the upper echelons of government decision-making. Women’s groups, despite the fact that they represent half of the population, are often kept as outsiders. Federal Sex Discrimination Commissioner, Pru Goward, consulted with:    “employers and employer groups, unions, women\'s groups and the community earlier in 2002. Over 250 submissions were received from employers and employer groups, unions, women\'s groups, community groups, Government, academics, individuals, health professionals, legal organisations and other interested parties. HREOC said the submissions overwhelmingly supported the introduction of a national Government- funded paid maternity scheme.”  Despite this, no maternity leave provisions were made in the 2002 or 2003 budgets.   According to an article in The Age newspaper:  “[Pru] was a long-time prime ministerial confidante, but indications are that a chill has descended upon the relationship, possibly over this issue. Goward says she and John Howard have not spoken about the paid maternity leave proposal for more than a year. Her proposal is not faring well in the inner circles of the Government. Canberra gossips speak of her being \"frozen out\" by the Prime Minister because she embarrassed him by championing a cause so allied to the Labor and union agenda.” (Crabb, 2003).  Elite theory says that it is certain groups of people – those who have money, influence and status in society, that determine the political agenda, and those that do not have very little influence over government decisions. The social status of the groups being represented seems to be an indicator of whether they will be considered an ‘insider’ or not. Pru Goward, once an insider, presented policy that was not consistent with the needs of business. She consulted hundreds of interest groups to formulate her report, and, as a result, was “frozen out”. The evidence shows that, at least in terms of ‘insiders’ versus ‘outsiders’ in Australian federal politics, there is most certainly an argument for the existence of elitism.  However, our pluralist democracy is not entirely lost. The Australian system of government allows for pluralist ideals in many ways, including: the federal system; a bicameral system of parliament; thriving local governments; and local community groups provide opportunities for interest groups to have their say on different policies. In the federal arena, the Senate has been a particularly effective system of ‘checks and balances’ against pressure groups wielding too much power.   In conclusion, the evidence suggests that interest groups are not equal in terms of their access to government and in their effect on government policy – at least at the federal level. This is because those with more money, those that a bigger in size or those with a high status are generally more capable of lobbying the government. And, continuing from that, the evidence not only suggests inequality of access, but very high barriers to entry into the inner circle. This means that money, size and status, on their own, will not necessarily guarantee strong influence on the government, that it is actually the combination of these three factors that make that inner circle. The essence of elite theory with regard to interest groups supports the idea of a wealthy, powerful elite that affect government decisions. Given that a significant proportion of ‘insiders’ groups are wealthy, large and have significant status, on the face of this we can say that federal politics is not particularly pluralist, and is, in fact, at least partly, elitist.   '),(24369,808956,'05,April,2004','2004-04-05 00:00:00','   I haven\'t written for some time, as many of you may be aware (if anyone is still reading this at all...). I have had a bit of a change recently. I changed jobs, which took me 800km away from Perth, in a beautiful, postcard-esque town called Esperance.  It is the first time I have lived away from Perth, and I must say it has taken some adjustment. I basically got offered a position as a social worker out of the blue, and had about 2 weeks to get all my stuff packed, Jason to give his notice, kiss everyone goodbye and head for Esperance.   And then, all of a sudden we were here, living 800 kilometres from home, still renting DVD\'s, still watching Foxtel, still trying desperately to stick to the low carb diet, still going to work everyday - only amidst this backdrop of sheer paradise. Look at this picture - this is literally 5 minutes from my house. It is really, really strange. And really, really surreal.      And life goes on. My family ring me asking \"so, how is it in Esperance?\" and I don\'t know what to say. I guess they want me to enthuse, or go into great detail about everything, but, at the end of the day, I guess I have already, in the space of six weeks, taken for granted the paradise I live in. It\'s really hard not to just say \"You know, same old, same old\" to your family when they ask, but, really, that\'s all it is. If you keep the same comforts, the same hobbies, and have your family around you, it\'s amazing how we can be satisfied wherever we are.  Occasionally I think of the distance and feel like I have been punched in the stomach for a few seconds, but I guess that goes with the territory. Occasionally I wonder if I made the right move career-wise, but then I remember that its all part of my life experience.  I hope that I can share some good experiences through this blog here: I have just had a conversation with my \"supervisor\" about creative expression and my work, and whether I can write about my experiences down here. You\'ll find some things shocking, i\'ll tell you... I am here as much for my own philosophical learning experience as I am to make a living, so I would really like to express it.  But, we\'ll see. Just thought I would check in and let everyone know that I am here, I am sane... I am just making sure I don\'t get fired for writing things ;)   '),(24370,808956,'27,May,2004','2004-05-27 00:00:00','   urlLink This is the key to my success... or is it?   It seems that publishers are now taking notice of the Blog thing. I\'d better lift my game - who knows, there could be limitless fame and fortune awaiting me.  Or not.  We all know that most writers still survive on baked beans, even if they are published. Now,   children\'s books  , that\'s where the Harry-Potter-who-gives-a-shit-but-I-would-sure-like-to-write-sonmething-worth-billions successes are.  OK, I\'m feeling a bit cynical today. I doubt anyone will be emailing me out of the blue asking me for my Michael Moore-esque rants in print...  But, if you are out there, I will wave like an uncomfortable Sim without a chair at you and hope for the best.   '),(24371,808956,'26,May,2004','2004-05-26 00:00:00','   The local primary school\'s Principal came into the office today and offered us a heap of books they are getting rid of. There were three boxes full, and there were SO many classics, like Robin Klein, Roald Dahl, Paul Jennings, Choose You Own Adventure books, Babysitter\'s Club... you name it!  I feel like I have won the lottery - all my favourite childhood books are now sitting in ap lastic bag, waiting for me to take them home.  Woohoooooo!   '),(24372,808956,'24,May,2004','2004-05-24 00:00:00','   urlLink Bush falls off bike   This is what happens when Daddy Bush takes off the training wheels...   Bush suffered minor abrasions to his chin, upper lip, nose, right hand and both knees, but was able to ride back home, Duffy said.   Wow, he got himself injured during the Iraq war, but not quite how people would imagine...   Tubb treated the president at the scene. Bush was wearing a helmet and a mouth guard when he fell, Duffy said.   Buahahahahaha! He wore a mouthguard riding a bicycle?   In January 2002, Bush fainted for a brief time in the residence of the White House while eating a pretzel and watching a professional football game on television.   OK, so let me get this straight. Our wartime President get injured twice in his term. Once from falling of a bike and the other from choking on a pretzel. Its no surprise this guy wears a mouthguard whilst riding a bicycle - he seems a little accident-prone. :)   '),(24373,808956,'21,May,2004','2004-05-21 00:00:00','   Normally I am a good sport. Really, I am. I applied for an internal vacancy yesterday back in Perth, and it took no more than 12 hours to get this reply:   Tea, thank you for your expression of interest for the advertised position.  You have not been successful.   Wow, geez, thanks. I mean, I know how the public service works: they advertise internal vacancies in order to renew existing contracts, but still... where is the honesty? Being the bundle of self-loathing mess that I am, I naturally assume it was something to do with me and my application that made me get rejected. Not only that, getting rejected so quickly after the application...ARGH. But, I am a wuss. You know what I did? Little miss polite Téa just replies:    Hi xxx, Could you please provide some feedback about why I was not considered for the role?     Pussy!  That was NOT what I was thinking at all. Here is what I would have LIKED to have written:    I really wanted to say \"Dear Sir, Sorry to have WASTED your time with a job I am obviously overqualified for. Seeing as you have obviously not taken the time to assess my application on its merits, could you at least tell me WHY so that I can shatter any further hope of fulfilling anything even remotely close to my goal? Or at least reevaluate why the hell I even bother with a Masters degree to miss out on a paltry Level 4 position?  And, for the record, I will have your job one day.   Bitchy, I know. But damn, I wish I could have been honest about it. They say that the process is meant to be fair and open. Bullshit. Like I said, normally I am a good sport, and I do understand how it all works. But it doesn\'t help being on the receiving end of it.   '),(24374,808956,'17,May,2004','2004-05-17 00:00:00','   urlLink The America We Know   I found it interesting that I was saying a similar thing this time last year around the war. Of course, I used  urlLink  Michael Jackson and Saddam Hussein , but hey, maybe I could be a paid writer too...  If you go back and read what I wrote, you will notice that I was commenting on the way war is presented, or represented, on commercial television. My friends, the events of the past few weeks are what I was waiting for.   Now, don\'t get me wrong - it is horrific what is happening and the images are very upsetting - but I was waiting for this. I was hoping for the gory details to snap people out of the dream world that has been created for us by the media. For once we are being forced to confront the violence and the human suffering that has been caused by this invasion.  So, I want to give a big thumbs up to the journalists that exposed the gritty reality of this war. The more we see, the more we\'ll be outraged. And the more we are outraged - and the more we are witness to the tragedy of war - the more we will oppose it. And the more we oppose it, the more resistant we will be to Bush and his ilk the next time they drag us into a war nobody wants.   '),(24375,808956,'16,May,2004','2004-05-16 00:00:00','   urlLink Gwyneth gives birth to girl   OK. That is the stupidest thing of 2004. Apple? FFS...  Gwyneth, there are cool \"different\" names and then there are just plain stupid. Try to guess which category yours fits into.   '),(24376,808956,'12,May,2004','2004-05-12 00:00:00','   It is interesting how much you learn about yourself when you\'re least expecting it. Right now I am homesick as hell, and feel a bit like I could easily just jump in the car, drive home, and start all over again. And sometimes, I can\'t help but think I have made a horrible mistake by dragging my family to a small town, a small office and a job that..sometimes...I don\'t know that I committed to anymore. I feel like I have forfeited all my chances of promotion into a policy job, its really hard to make friends and the work is hard. Actually, hard doesn\'t do it justice. It\'s painful sometimes. For weeks I have been thinking \"what does this mean?\", \"why can\'t I commit to this job?\". Well, in part, its because I took a huge risk - jumping into a field that is different to what I am used to - but part of it is around me and my broader search for meaning. And I found it in the most unlikely of places.  I watched eLove on the Lifestyle channel last night, which is a reality show cum documentary about internet love. A divorced American Policeman had been chatting to an English woman for over a year. He had told her \"I love you\", they had phone conversations, they had shared photos and webcams, emailed back and forth, chatted, talked obsessively about each other to their friends... well, you get the picture. Then, this eLove show comes along, pays for her to go to New York to meet him...with sexy results. OK, no sexy results, but that\'s just a lame thing I say ;)  Anyway, despite this electronic barrage of intimate secret sharing, joking, and mutual \"I Love You\" sessions, the American policeman ended up not only rejecting the pommie lady, but he CUT THE DATE SHORT! Now, this could be a reflection of his personality (I mean, NY Policeman: read moustache, chewing gum and a loud mouth - a walking stereotype), or it could be a broader comment around human nature.  Of course, it wouldn\'t be a very interesting blog if I just called him a jerk, so of course I am going to say that it is something bigger than that. It is about fantasy and reality, and how we always want what we can\'t have.  Internet relationships are a little like the fantasy of living and working in a small town. You build up this fantasy of how different life would be if you were away, or if this person you had fantasised about were with you. You create a magical place, where you can relax, socialise effortlessly, and run through daisies in a meadow.  And then you do it. You decide to turn this fantasy into reality. And that reality is nothing like you thought it would be. It is the day-to-day, warts and all, experience of what you had conjured up in your head. If you had experienced this without the fantasy, it wouldn\'t be a disappointment - it would just be life. But, because you\'ve dreamt about it for a long time, often because of a desire to escape this bitch called reality, it will never, EVER live up to the glossy images.  It\'s a bit like holidays. You pour over the glossy magazines of white sandy beaches and cafés, but when you get there, you realise that there is gastro, homelessness and human suffering outside the cosy confines of the resort. Internet relationships are exactly the same. No matter how honest particpants profess to be with each other, you are still only seeing the magazine article. You are not hearing the loudmouth policeman, or smelling the bad breath in the morning, or having to deal with that person rolling their eyes when you speak.  And in Esperance, despite its physical beauty, there is an undercurrent of religious zeal, homelessness, depression, cliquey-ness and cold. Lots of cold. The reality is just far too much. Then you start to yearn for the days where life was better: the warm fuzzy feeling you got when you were in-a-relationship-with-no-human-contact, writing nice notes and having heartfelt discussions. And you were only dreaming about moving to the country, but still living a life you like.  On eLove, they pommie chick and the policeman managed to salvage their friendship via email, but it will never be the same. Most of these come-crashing-down-to-reality experiences change you at your core. Part of it is growing up and realise that you cannot escape the reality of life, and part of it is the disappointment associated with it.  I will go home to the city soon, having learnt that maybe I made a huge mistake. But, on the other hand, I can take some relief in the fact that if I didn\'t experience it, I would never have learned that real life isn\'t so bad.   '),(24377,808956,'03,May,2004','2004-05-03 00:00:00','   urlLink Watchdog of the Wankers   Now *this* is what a blog should look like. Well written, funny, and right on the money (except for that Craig Reucassel thing..I am more of a Chris Taylor sort of girl) :)  Will update again soon. I have loads of juicy stuff.   '),(24378,808956,'22,June,2004','2004-06-22 00:00:00','   ...to the Australian Flag, whether I want to or not.  urlLink Schools forced to fly the flag   C\'mon, John. Get real. Do you really think that this is a real issue? Of course you don\'t. What this is, however, is your acknowledgement of your failure to win the aspirational, under-50 voters back from Mark Latham. So now, what you are doing is appealing to the grey RSL regulars that you *know* dominate our society. You\'ve gone back to Johnny heartland: the voters that are resistant to change; resistant to dissent; and, above all, resistant to those pesky young-folk that have no respect for their elders.  You have to wonder why, John, when you pull such cynical stunts such as this. it shows a politician that is desperate: trying to exploit the racist, conservative heartland - instead of making politics about ideas, you make it about winning.  I have to ask: what happens to the Aboriginal children? Will their flag be displayed alongside the Australian flag as a symbol of unity? Or will the Australian flag be it - a symbol of the \"like it or leave it\" attitude of colonialism? A symbol of pride for some, but a symbol of oppression for others?  John, you have successfully reversed most of the things that have made this country great and unique. It is only when I get a tolerant Australia with a flag that does not have the Union Jack on it (first and foremost) that I will stand proud. Until then, I will write this off as yet another political stunt from a cynical politician, desperate to cling onto his power base.   '),(24379,808956,'15,June,2004','2004-06-15 00:00:00','   Normally I leave Big Brother commentary to those who are  urlLink better at it than I am , but I just can\'t leave last Sunday\'s eviction show without commenting on it.  For those that did not see it, the 5th evictee, Merlin, walked down the plank to the stage. He put gaffa tape on his mouth, help up a sign \"free the refugees\" and refused to talk to Gretel (who handled it fabulously given the circumstances), with sexy results.  Now, there were two great things that came out of this show. Firstly, it showed tremendous courage on Merlin\'s part. Secondly, I think it showed a new, young side of Australia that is really encouraging - and rare - to see on national television.  When Merlin jumped out of the car, you could see that he was extremely nervous. There were thousands (is it thousands? I looked like it) of people cheering him on. He fiddled with the sign and the tape for a while and everyone was collectively wondering \"what the hell is he up to?\". And then he pulled it out. A shitty sign with \"Free the Refugees\" on it. And it gets better... he WALKED RIGHT PAST GRETEL AND SAT DOWN, gaffa tape on, and refused to talk. I nearly peed my pants from that one! But you could see it in his eyes - he looked like he wanted to cry, apologise and take it all back, but he stood his ground.  I just have to say that it was phenomenally brave of Merlin to do such a thing, with the audience booing (why?!), and Gretel trying to salvage what was left of the eviction show. It was one of those great moments in television... where a leftie got one over the big guys :)  The main thing that surprised me, however, is the way the housemates reacted to Merlin\'s message. We hear on talkback radio that refugees are \"dangerous terrorists\", that the majority of Australians support mandatory detention, and that to support their freedom would be \"unAustralian\". But I saw something very different. For me it showed a huge cultural shift that represents how Generation X & Y are ignored in public debate. Sure, Big Brother is hardly representative of young Australia, but its all we\'ve got for now... and I liked what I saw. People who were tolerant of Merlin\'s opinion (and most even agreed with him), had great respect for him and defended him over and over.  The way the housemates reacted gave me more than pure voyeuristic pleasure - it gave me a sense of confidence that things will eventually be ok. Baby boomers dominate the political scene, and whilst it is their democratic right to do so, we often lose sight of the younger generations\' opinions - generally opinions supportive of multiculturalism, diversity, postmodern theory, and most importantly, a strong sense of justice.  It sure made me feel good for a fleeting moment. Of course, then I read the newspaper again and I turned into a frustrated young person again. But hey, it was a special moment, even if it did only last an hour.    Good on ya Merlin...   '),(24380,808956,'12,June,2004','2004-06-12 00:00:00','   I didn\'t get it.  They said that I was a competitve candidate, but didn\'t really get it. Ah well. No biggie really.   '),(24381,808956,'12,June,2004','2004-06-12 00:00:00','   I wasn\'t going to update until my last essay and exam were out of the way at the end of the week, but hey, What would study be without procrastination?  Yesterday I had a job interview for a really good [Policy] position that I thought I had absolutely no chance of getting. To make matters worse, living 800km from Perth meant that I had to have a phone interview. And, my god, was that the worst possible thing that could ever happen to anyone.  Public service positions are based on merit based selection, and that is great. But when you are already a bundle of nerves and cannot even SEE the people that are interviewing you, it adds a whole new dimension to the interview experience.  So, being the self-loathing, agonised writing-type with absolutely NO idea why I was even being interviewed in the first place, you can imagine how much fun I had...  First, I had to talk about community engagement principles, my understanding of domestic violence and the state strategic plan and all that... and I did really well. The boss-man was making jokes and seemed to understand what I was getting at most of the time (which is new for me... it must be a policy-person thing, where there is an unspoken understanding of sleepless nights thinking about political problems!).  I made a comment early in the interview that it is hard to know whether I am saying the right thing, oe boring them to death because I can\'t see their faces. Big Kahuna laughed and said \"don\'t worry, we\'re all smiling\"  So, anyway, I was doing really well. Then, I got an EASY question. An EASY question that I just could not answer, despite me having done several interviews in the past and being acutely aware of the position and everything that it involves. ARGH! Here\'s a little transcript.  Interviewer: \"So, Téa, knowing that this is a small office whose decisions can impact on a whole state of women, government employees\' practice and the community at large, how would this affect the way that you do your work.\"  (silence)  (more silence)  Me [thinking]: \"Shit... potentially 2 million people...\" Me [talking]: \"I would utilise community engagement principles of course...\"  Interviewer: \"What skills do you believe you could bring to the position,  in that context?\"  (silence)  Me: \"Good communication skills, good team skills as it is a small policy unit, good understanding of community engagement principles, including having an awareness of cultural diversity, researching and educating myself about these things...\"  (silence)  Me: \"Do you want to know specifically  how  I would communicate?\"  Interviewer: Yes.  Me: \"I\'m sorry, you know how something just...well..leaves your head? I know this stuff!\"  Interviewer: \"It\'s ok, I think you\'ve touched on what we need...\"  And then, we moved onto Q&A time, where I asked if I was shortlisted, because I was surprised I was selected for an interview. He said that I had been shortlisted from a huge number of applicants and that my applciation was really strong. Wowee. Bit of a confidence boost for me, considering the job that this is...  So, anyway, now I wait. And I agonise. Apparently I will hear on Monday. Damn them for making me wait all weekend! I don\'t want to be overly confident, after all, in hindsight I cannot believe the things I forgot to say... but hey... wish me luck.  And tell me if you think I got the job. I went to  urlLink Facade  for some Tarot insight and it gave me a really good card, which means excitement, rebirth and challenges. So yeah... anyway, how this ended up so long I have no idea.   '),(24382,808956,'06,June,2004','2004-06-06 00:00:00','   urlLink To add insult to injury...   Doctors want an extra $675 a week. Well, whoopee. Since when have doctors been the only professionals that deserve a payrise in our public health system? Not only that, but $675 a week is more than the average nurse or social worker EARNS.  It\'s an absolute joke. I don\'t care if I piss off doctors - I am all for investing in the health system, but demanding a raise that is more money than most people, including nurses, orderlies, social workers, hospital administrators, and so on earn... it is absolutely disgraceful.  Sure, they are overworked, and sure 20 hour shifts are ridiculous, but until everyone gets greater recognition of their skills in the health sector, I am not about to start crying over someone that earns $130,000 a year.   '),(24383,808956,'04,June,2004','2004-06-04 00:00:00','   urlLink Bush says Reagan helped save the world   Yes. Sure he did. He sold weapons to Iran...   '),(24384,808956,'04,June,2004','2004-06-04 00:00:00','   urlLink ...but only if you\'re American.   First of all I want to get this off my chest: \"what a load of shit\". Now, let me get to the real stuff.  Australia is one of the truly democratic nations in the world. Sure, it isn\'t perfect - like most nations it is run by wealthy white heterosexual men and has a long history of native blood going splat in the name of Christianity. But hey, which country doesn\'t?!  One thing we can be proud of, despite our ills, is that we have a nation that is pretty darn good in comparison to other \"democracies\". We have a system of compulsory, preferential voting that ensures everybody\'s vote (whether informed or not) counts. We have a vibrant pluralist system (occasionally), a strong bureaucracy (despite it being undermined in recent years) with fairly good appeals processes, and most importantly, a parliament that allows for alternative policies to the government of the day. Not only that, we have two current affairs shows that ensure that just about anyone can get their opinions heard. Unless of course you\'re a hippy, a bureaucrat, a welfare worker, a dodgy builder or any sort of do-gooder. Alas, I digress.  I was talking about how Australia is a democratic nation, with a parliament that allows for something that is strange to the likes of George Bush: a viable opposition. Mark Latham, as the next Prime Minister of this country, has the FREEDOM, through this DEMOCRATIC institution that he is about to lead, to express his point of view on Iraq. Now, I may diagree with Mark Latham\'s position on many things, but on this one I think he is right on the money. But, that is beside the point.  George Bush walks around, bible in hand, dribbling rhetoric about \"freedom\" and \"democracy\", but does not promote it in his policies, in his speeches or in Iraq. But of course, we all know that America is an autocracy disguised cleverly as a democracy. George Bush acts  in spite of  public opinion and without the mandate of the American people. People cannot forget that. But again, I digress.  This democracy that Australia has, as complex as it may be, is more representative than America. Here, the government of the day, for better or worse, has the mandate of the Australian people. Sure, they are accountable in the Senate, but Australians like it that way. If the ALP win this election, troops will be withdrawn from Iraq. This means that Mark Latham will have the mandate of the majority of Australians to withdraw troops from Iraq.  Surely George Bush would not argue with that? After all, he is such a huge fan of freedom and democracy...   '),(24385,808956,'03,June,2004','2004-06-03 00:00:00','   Thanks to  urlLink Jase  for this great link:   urlLink Pictures in Baghdad   Its actually a little unnerving to see picture of flowers, nut shops and palm trees. Take a look.   '),(24386,808956,'03,June,2004','2004-06-03 00:00:00','   I would like to take the opportunity to publicise my husband\'s blog. Yes, he has entered into the adventurous world of blogging - and does it better than I do. Darn it.   urlLink It sucks to be you .  Apparently so-named because if you are reading his blog, you truly are pathetic ;) But hey...  Here\'s a shout out to Jason.   '),(24387,808956,'02,June,2004','2004-06-02 00:00:00','   urlLink Cheney\'s slick deal for old firm    Is anyone as surprised as I am?  (not)   And here\'s another one.  urlLink \"Sorry\" Sign to ease road rage   I have two cracks I can make here:  1) If I have to have a sign that says \"I\'m Sorry\", then I also want one that will flip the bird. Or, perhaps, a big red boxing glove on a spring that you can eject. What movie was that again? 2) I wonder if the Prime Minister will use his? It seems the words \"I\'m Sorry\" are not his favourite. The light may result in an admiision of guilt, and we don\'t want that now, do we?  hmm...   '),(24388,808956,'01,June,2004','2004-06-01 00:00:00','   OK, I couldn\'t let this one go by without at least a comment. I think that  urlLink this  is one of the poorest pieces of journalism I have EVER seen in a Murdoch paper. And that really is saying a lot, given that, well, it\'s Murdoch.  My question is, why is Natasha Stott-Despoja\'s eligibility for the $3000 maternity payment even relevant to the debate, other than, of course, to try and make a mockery of a young working woman\'s opinion that the payment is nothing but a whitewash? I imagine that the writer (or writers, given that the Murdoch press usually have a few staff writers working on one 200 word article) was trying to point out some sort of irony, or try to counter Natasha\'s valid criticism of the maternity payment by pointing out an inconsistency (if there is one at all).  I just don\'t get what the purpose of this story is, because it fails in the irony stakes, it fails to counter Natasha\'s view... perhaps it is a politically motivated page-filler to further relegate women\'s issues to pettiness and mockery. Oh. I think thats exactly what it is. In amongst the adverts and stories about chickens that can talk is this little ditty about the critic of the maternity allowance being eligible for the payment...  What people are forgetting, firstly, is that both the $800 maternity payment and baby bonus (worth up to $2500) have been replaced by this allowance. It is not a new allowance, it is merely consolidated benefits. Secondly, those who work prior to pregnancy are still going to be worse off as a result of having a baby due to forfeiting a salary. Thirdly, the payment is not \"family\" friendly for one important reason: there are a lot of families out there that have finished having their babies and need to cope with the day-to-day. A truly \"family-friendly\" policy needs to account for the mothers who spend half their income on childcare, rack up debt with a no-win-situation, otherwise called the Family Assistance Office, and, of course... young families being crippled by HECS debts, and unable to buy houses amongst other things.  And, of course, assuming that all families have a stay-at-home mother in a heterosexual marriage (as defined by the Marriage Act amendments) and living behind a white picket fence... BLERGH! Come off it John. This payment is nothing but another conservative sweetener to keep women as breeders, but providing absolutely no assistance to women who do not stay at home with their children.      Making childcare tax deductible  - now THAT is a policy that supports families.   Paying full maternity leave for 14 weeks  - now THAT is a policy that supports families.   Investing decent amounts of money into Medicare and scrapping the private health rebate   - now THAT is a policy that supports families.   Not penalising people who get an education  - now THAT is a policy that supports families.   Not having a \"Family Tax Benefit\" system whereby 650,000 families have a debt   - now THAT is a policy that supports families.   I don\'t think it\'s that fair to expect fairness from the Daily Telegraph - after all, they are not renowned for the world\'s greatest journalism - but the least they can do is TRY to have an actual story when they write about something.  Family policy is something that I have a particular interest in, so excuse the rant. But I expect far better from people that get paid to write about these issues. After all, I am just a lowly unpaid blogger and I can come up with *something* of substance.   '),(24389,808956,'22,July,2004','2004-07-22 00:00:00','   I quit the job today. It was quite a nerveracking experience for me, because I am not very good at quitting jobs, especially when I have only been there for a week!  My supervisor is really nice. She made me feel really welcome and is a really positive person to work for. The rest of the team are also really good, so, with liking my workmates and&nbsp;enjoying the job I did what any sane person would do: I quit.  I walked in this morning, feeling like crap (I have had an ear infection for the last 3 days and I feel like I am being stabbed in the head with a pumpkin), and having braved yet another hour-long drive from home to work, I make morning niceties such as \"hello, how are you\" etc... and then I tell my supervisor how I didn\'t do the whiteboard on onday because I needed some stuff explained to me. Then, she said something and went to answer her phone.  About half an hour later, and after some agonising on my part, she came and sat on my desk and said \"I\'ve got a really good job for you\". I cut her off and said \"at the whiteboard, we got interrupted.\"  I flinched. I then said that I was planning to quit working and focus on Mina and full time study. They didn\'t take it very well at all. They put massive guilt trips on me for taking the job (which I think is fair enough), and blah dee blah...  Anyway... I am tired now. Will try to post something of more substance later on.   '),(24390,808956,'21,July,2004','2004-07-21 00:00:00','   ...to have my feet firmly planted on the ground after a couple of weeks of chaos.   I finally made it back to Perth on the 10th, and the dust has only just settled, as I was only given a week to start the job I was offered. However, this week I have decided to quit and focus on my studies and do some preliminary studies for my Grad Dip Ed next year.  I have a 50 minute drive to and from work, and have a few problems with stress and palpitations (for which I am seeing a cardiologist next week! argggh), and, as you can imagine, being stuck in traffic is not that good for my health. Also, the money I am shelling out for daycare and the time I am forfeiting with my daughter just doesn\'t make it worth it.  So I am going back to uni full time as of next week, staying at home with Mina and enjoying it for the next year and a half until I become a high school teacher. Woohoo!   '),(24391,808956,'15,July,2004','2004-07-15 00:00:00','   urlLink Read this article   The concept of freedom of speech is yet again undermined by the bible belt loonies in the US. WHy on earth would SlimFast be so opposed to Whoopi\'s comments? Surely they aren\'t naive enough to support Bush?  Then again, they are a money-grubbing corporation that is profiting from the obesity epidemic, so I guess that its not that hard to believe...   '),(24392,3746444,'25,June,2004','2004-06-25 00:00:00','   I dont know how to feel right now. My old flame has just called me. I can only do one of two things: fuel it once again or to throw dirt on it so it burn me once again. I know if I get to close she will burn me but if i stay to far i get to cold. When we used to be I was warm but I alone set her out and found my self trying to find new things to warm me like i used be. My flame used to warm her friends as well as mine but i saw once i put her out she drifted to only one other. All of us gathered and huttled together to keep warm thus growing all of our friendships but we all knew something was missing. I turned around and i saw her making another man warm. She seemed brighter than she has ever been i almost didnt recognise her until i remembered that i saw her like that once before thats the flame that used to warm me. Only it wasnt warming us she was warming. She was burning brighter than ever as if another flame was joined with her. But i feel now that if i get to close that i will try and start another fire using her flame But then again i dont want to get burned once more. Maybe Shakespeare isnt right maybe it is better never to have loved.   '),(24393,3746444,'30,July,2004','2004-07-30 00:00:00','   &nbsp;&nbsp; I\'m still searching for a girl that\'s right for me. I have no clue what else to do. Who knows maybe.... there isn\'t anybody that is truly right for me. I mean all the girls I have falling for are either too old, have a boyfriend, bitches, or my \"favorite\" too good of friends. Why can\'t I fall for a girl that can fit in with my friends, cool to talk to, and will actually go out with me. I want a girl that can be crazy but can still keep me in check. I need a girl like my best friend Brandi. And yes Brandi I know you will read this and probably say, \"What the hell are you talking about!\". But I have to say you are the sweetest, kindest and coolest people I know. And don\'t worry I think both of us will be able to find our fish. But for now at least for me the search continues.   '),(24394,3746444,'27,July,2004','2004-07-27 00:00:00','   &nbsp;&nbsp; I\'m still searching for a girl that\'s right for me. I have no clue what else to do. Who knows maybe.... there isn\'t anybody that is truly right for me. I mean all the girls I have falling for are either too old, have a boyfriend, bitches, or my \"favorite\" too good of friends. Why can\'t I fall for a girl that can fit in with my friends, cool to talk to, and will actually go out with me. I want a girl that can be crazy but can still keep me in check. I need a girl like my best friend Brandi. And yes Brandi I know you will read this and probably say, \"What the hell are you talking about!\". But I have to say you are the sweetest, kindest and coolest people I know. And don\'t worry I think both of us will be able to find our fish. But for now at least for me the search continues.   '),(24395,3746444,'27,July,2004','2004-07-27 00:00:00','   No hope it seems for me. I\'m still lost at sea trying to find a light to guide me through the fog. I wish Cupid would just come down and shoot his arrow into some girl for me. Its been so long from the time my guardian angel has spoken to me. I wish she would help me out. I cant stand not talking to her for too long. She\'s my psychiatrist and my doctor....... but most of all she\'s my friend. Brandi I miss you so much. We haven\'t talked in what it feels like an eternity.   '),(24396,3746444,'21,July,2004','2004-07-21 00:00:00','   I know the one thing I need in my life is someone who is there for me. I mean I know my friends will always be there for me no matter what its just that with my \"old flame\" we were always able to talk and joke around. I need that again. I need them to know that no matter what I\'m there for them. I\'m not the lying Bitch mine was. The fact is I want someone real bad but.......&nbsp;I know I need to tell her but I think the fact is....... I think she wants to get with my best friend and that would make me so happy to see but in my heart........ She is the perfect girl and if I cant get with her I need someone in my life that is just like her.   '),(24397,3746444,'14,July,2004','2004-07-14 00:00:00','   I don\'t know what came over me before? To actually think that my old flame and I would have gotten back together. She never changed she is still burning as cold as ever. Even when we started becoming a new she still only played with my heart and still lied in my face. But thankfully I have a friend that stupidly would tell him and seemed to \"forget\" when she would talk to me. This time I think.... NO ... I know what I want to do. I don\'t want to be like her and Travis are. I don\'t want to be acquaintances. NO.                                                                                                                                             NO... I need her out of my life forever   '),(24398,3746444,'16,August,2004','2004-08-16 00:00:00','   Brandi I\'m so sorry that you have to go through this shit. I don\'t know why God puts the nicest and most kind hearted people in this position all the time but he ends up doing it no matter what maybe it\'s to see if we can keep our beliefs and stay on track and hopefully he can see that no matter what life throws at us we will always stay true to what we believe. I know your hurting babe and you don\'t know how much I want to go over there and just let you cry on my shoulder but we can\'t just relize that no matter what I will always be here for. No matter what time of day, what I\'m doing, or who I\'m with you can always count on the fact that I will be here for.   '),(24399,3746444,'12,August,2004','2004-08-12 00:00:00','   Brandi I\'m so sorry that you have to go through this shit. I don\'t know why God puts the nicest and most kind hearted people in this position all the time but he ends up doing it no matter what maybe it\'s to see if we can keep our beliefs and stay on track and hopefully he can see that no matter what life throws at us we will always stay true to what we believe. I know your hurting babe and you don\'t know how much I want to go over there and just let you cry on my shoulder but we can\'t just relize that no matter what I will always be here for. No matter what time of day, what I\'m doing, or who I\'m with you can always count on the fact that I will be here for.   '),(24400,3746444,'12,August,2004','2004-08-12 00:00:00','   Brandi I\'m so sorry that you have to go through this shit. I don\'t know why God puts the nicest and most kind hearted people in this position all the time but he ends up doing it no matter what maybe it\'s to see if we can keep our beliefs and stay on track and hopefully he can see that no matter what life throws at us we will always stay true to what we believe. I know your hurting babe and you don\'t know how much I want to go over there and just let you cry on my shoulder but we can\'t just relize that no matter what I will always be here for. No matter what time of day, what I\'m doing, or who I\'m with you can always count on the fact that I will be here for.   '),(24401,3746444,'12,August,2004','2004-08-12 00:00:00','   Brandi I\'m so sorry that you have to go through this shit. I don\'t know why God puts the nicest and most kind hearted people in this position all the time but he ends up doing it no matter what maybe it\'s to see if we can keep our beliefs and stay on track and hopefully he can see that no matter what life throws at us we will always stay true to what we believe. I know your hurting babe and you don\'t know how much I want to go over there and just let you cry on my shoulder but we can\'t just relize that no matter what I will always be here for. No matter what time of day, what I\'m doing, or who I\'m with you can always count on the fact that I will be here for.   '),(24402,3746444,'12,August,2004','2004-08-12 00:00:00','   Brandi I\'m so sorry that you have to go through this shit. I don\'t know why God puts the nicest and most kind hearted people in this position all the time but he ends up doing it no matter what maybe it\'s to see if we can keep our beliefs and stay on track and hopefully he can see that no matter what life throws at us we will always stay true to what we believe. I know your hurting babe and you don\'t know how much I want to go over there and just let you cry on my shoulder but we can\'t just relize that no matter what I will always be here for. No matter what time of day, what I\'m doing, or who I\'m with you can always count on the fact that I will be here for.   '),(24403,3746444,'12,August,2004','2004-08-12 00:00:00','   Hey Brandi. I know you are still in \"love\" with Gary. But i just don\'t see how you are able to see that you are able to say that when you don\'t even give another guy a chance. I\'m just saying that there is plenty more fish in the sea you might find an even bigger fish, you just have to have a little more patience. I just think its funny how you still want to go back out with him even after you said you will never go out with him again. Because I know for a fact that you would of kicked my ass for even thinking about going back out with Gaby. Sorry if I seem to be coming off as rude I just had to get this off my chest. And I could see if your pissed off at me but hopefully we are good enough friends to see that I\'m not trying to be a dick I\'m just saying you might want to give dating another chance.   '),(24404,3746444,'04,August,2004','2004-08-04 00:00:00','   My best friend is hurting and I wish I was there to give her a big hug and a shoulder she can lean on. But the problem with her is that she is a stuburn girl and would probably deny my help but Damn it Brandi you know we would end up fighting it and we would go after it for a good 1/2 hour or so and since I usually win all our fights we would just sit down and you would tell whats going on. So would blab on and on about how your problems are and I\'m just joking babe don\'t kill me you don\'t blab.............. a lot. Joking again. But I have to go but seriously Brandi when ever you want to talk just call me. Love Ya much   '),(24405,3746444,'02,August,2004','2004-08-02 00:00:00','   Man it was sooooooooooooooooooooooooooooooooo good to see everyone again. Except for Gaby that was way to fucking wierd. Walking to Col. Zider\'s office to tickle Carla and then she goes,\" Whoa I didn\'t think I would see you here today.\" and when she was walking off she goes its good to see you again. and gave me a hug. I wanted to say \" Well.. I wish I could say the same about you.\" (HAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHAHA). But Brandi I think the best reason for today was to to see your pretty little face. Could have lived without the smartass attitude but.... I\'m just playing you know thats what I love about you. Well I\'m going to sleep so talk to you tomorrow. Love Ya much.   '),(24406,3746444,'02,August,2004','2004-08-02 00:00:00','   Man it is so hard to say things on these damn sites because you know people can just come in and look at anytime. But the worst part is I have so many  things to say that just cant be put down on the computer or anywhere else. its weird I have them all going through my head at the same time and probaly have redun this one insert about three times. I just dont know. I just cant wait until school starts up to actually see some friends again. And talk to the ones I most desired to actually sit down and talk to besides going on the computer and typing what we want to say. But........ Damn this alcohol just makes me blab and blab doesn\'t it. Oh well you still have to read it Brandi Hahahahahahaha. Well to get to the point no matter what the case just call me if you need anything and the last time you called my phone didn\'t ring. No, I\'m serious. NO. I really do mean it. Well I love you and hope to see you soon or call me whenever.   '),(24407,4242609,'22,August,2004','2004-08-22 00:00:00','   Tonight I watched the Jets roll all over the Colts in their 2nd preseason game.  Chad Pennington played great, completing 10 of 11 passes and throwing for 150+ yards and a touchdown.  I really hope they get him signed to a contract extension soon.  Besides that, the first team didn\'t look that great.  The run defense sucked, and looked mostly the same as the last 3 years.  We were playing what might be the best offense in the NFL though, so it\'s not time to panic.  The Jets shutdown Manning with the exception of one big play to Reggie Wayne.  On the Colts opening drive Abraham hit Manning from the blind side, slowing down his pass and allowing Donnie Abraham to pick it off in the endzone.  Decent performance, and it\'s good that we are halfway through the preseason without any major injuries.  I can\'t wait to see this offense once Moss and Chrebet get on the field.  BTW, the following still suck: -The Jets backup QB\'s - Team USA Basketball -The Dolphins  '),(24408,4242609,'20,August,2004','2004-08-20 00:00:00','   Since my last post, Team USA has actually won a couple of games. First, they beat Greece. I watched it, and once again they couldn\'t shoot for shit. When did Richard Jefferson get this bad? Luckily Lamar Odom bailed them out with a bunch of big plays on defense.  Then, today they beat Australia. I didn\'t see the game, but from what I heard, we actually played really well in the 2nd half. Looking at the boxscore, Iverson, Duncan, and Marion played really well. Seems like Wade played good too, and Richard Jefferson only took 3 shots. It would be nice if we actually played a complete game, but hopefully the winning continues.  The Yankees Pitching has been scary, and not in a good way. I don\'t think there is one pitcher on their roster who you can be confident in to pitch good in the playoffs. At this point, El Duque might be our best pitcher. Mussina is back, but he struggled in his first start back, and hadn\'t exactly been tearing it up before he went on the DL. Kevin Brown has been pretty good, but not great, and not consistent at all. Same with Vazquez. Lieber put together a couple of good starts in a row, but I\'m not counting on that continuing. Loaiza has been consistent. Consistently giving up 4-5 runs a game while pitching only 5 or 6 innings. That isn\'t going to cut it. If all our pitchers get hot at the same time, we could still go all the way. Otherwise, it might be an early exit for the 2nd time in 3 years.  That\'s all I really have to say now. I did add a counter on the bottom of the page to see how many people actually read this shit. I\'ll probably write again after the Jets preseason game this Saturday vs. the Colts.  '),(24409,4242609,'15,August,2004','2004-08-15 00:00:00','   The Jets had their preseason opener on Friday. Here are my thoughts:     Chad Pennington looked very good, as usual    Curtis Martin is a lot more healthy than in years past, he will have a very good year    Vilma is a beast. He better be in the starting lineup by the time the season starts    The first string defense looked good as a whole. Of course, they were facing a backup QB, and Duece McCallister only had about 2 carries. Next Saturday\'s game against the Colts will be a much bigger test.    If Pennington gets hurt, I will kill myself. Brooks Bollinger is awful. Ricky Ray looks even worse so far. Hopefully one of them will step it up soon     The Yankees continue to roll along. I\'m happy that we are playing so well and probably have our best team since 98, but it would be nice to be in a close playoff race for once. With a 10.5 game lead over Boston, and a 10 game lead over Oakland for homefield advantage in the playoffs, there doesn\'t appear to be much to get excited about until October. At least there are some other good races though. The AL Central, AL West, and both wildcards should go right down to the wire.    Team USA Basketball is a joke. Whether we lose the rest of our games, or win the gold, I really don\'t care. This team is boring to watch. If we win in a blowout, it\'s because we are supposed too, and are on a different level than the rest of the countries. If we win in a close game, it isn\'t exciting, it usually means we played like shit and were lucky to pull it out. And then when we lose in a blowout like today, it\'s just a total disgrace. I\'m tired of just trying to put an all star team together. Find guys who fit what the team is trying to do. Or better yet, forget the NBA, and right after college season ends assemble a team full of college players. It will leave more time to practice, and will give us team basketball, instead of 12 superstars running around heaving up shots and turning the ball over. Maybe then, by the next olympics I\'ll watch more basketball than badminton.  '),(24410,4242609,'15,August,2004','2004-08-15 00:00:00','   Here is my blog where I will post all my thoughts about all the stuff I like, and maybe some other shit too. I\'ll probably post any exciting things that happen in my life that are worth mentioning, but I wouldn\'t expect many posts about that. I don\'t know how often I\'ll update. Sometimes I might update more than once a day, sometimes I might go a couple of weeks. It all depends what I feel like doing.  Enjoy.  '),(24411,3833031,'29,July,2004','2004-07-29 00:00:00','   it seems like most of the girls in my office of mostly girls are crushing on me. This one chick would be alright if she could get rid of her mild southern twang and speed up the rate of which words come out of her mouth in normal conversation.&nbsp; She also bats her eyes like the hot chick cats in the Tom & Jerry cartoons--you know, like when Tom is trying super hard to impress her like by lifting weights and playing his bass and then Jerry pulls some really slick shit like toss an olive in her martini from across the room and he gets the big smooch and then his whol body becomes rigid and his tail sticks out straight and steam comes out of his ears? It is like how she bats her eyes. Maybe she went to charm school. Or maybe she\'s just nervous.&nbsp; Regardless, rapid blinking isn\'t all that attractive o me. Nor is a supersized ass.&nbsp;If it is big because it has depth-think ghetto booty, then ok. If it is big and wide then I gots ta say the nay-nah my damie.  Brian hooked up some wireless internet in our place so that is cool.&nbsp;We can now retire the cans and string we\'ve been using to tell each other ghost stories at night and just send the whole story in the form of electronic mail. BOO!  Speaking of Brian--motherfucker woke up 5 minutes before I came home for lunch at 12:10 yesterday, and only because the FedEx guy woke him up. It\'s a good thing or he would have gotten a knuckle sandwich for the simple fact that I had been going for 5.5 hours at that point. Fuck work.&nbsp;On a similar note, last night I went to bed at 12:30 while he and Zac got to stay at the bar on the corner and play all night. Fuck work. Will you rob a bank with me?   2ce '),(24412,3833031,'25,July,2004','2004-07-25 00:00:00','   It feels like a fucking urlLink  bomb  went off in my head. Doing the hangover thing (duh) this Sunday morning.   Last night  urlLink Matt Long  came to town. He is gonna be going to  urlLink Morocco  soon so he is doing the visit \'Nati boys in the \'Nati then he is gonna visit the Dirr in  urlLink San Francisco . Probly he will try to have  urlLink gay sex  with Dirr cause you know the whole \" urlLink when in Rome \" thing, well he will be like \"when in San Francisco.\"   We went to O\'Malley\'s again and drank copius amounts of alcohol. Obviously, I didn\'t drink so much that I can\'t remember what I consumed. Here is the run-down:    Before we left:   3 bottles of&nbsp;Konig Ludwig Royal Bavarian Hefe-Weizen  1 bottle of Budweiser&nbsp;   At O\'Malleys:   2 bottles of Budweiser  1 Jagermeister shot (more like two)  1 bottle of Budweiser  1 Kamikazee&nbsp;&nbsp; 1 bottle of Budweiser  * I borrowed one dollar from Matt in order to buy one last Bud. I left it stranded in the bathroom over half-full cause I couldn\'t finish it.  There was some motherfucker who kept repeating stuff we said and mocking us and I guess he said something else to Matt that I missed. Zac wanted to fight him super bad and I kinda started wanting to, too. It\'s like, hey, you are at a bar and you are in the midst of drinking professionals, they start talking loud and saying crazy shit and you find it necessary to mock them because you are a fucking square and you can\'t comprehend with 100% of your brain why people would act that way and you are at the same time trying to impress the two ugly chicks across the table from you so they will think that you are cooler than you your two equally dorky friends. Whatever. I hate it when people are like \" urlLink fighting is stupid and barbaric \" but then they fucking run their mouths and speak fighting words thinking they are fucking righteous or something for not resorting to physical violence. Violence&nbsp;can be&nbsp;verbal too. Don\'t say shit if you aren\'t gonna do shit. So I didn\'t say shit. Though I would have done shit: I was drunk enough.   So we got home and I went to pass out and it was spin-central. It was the total opposite of the  urlLink O\'Reilley Factor . I forced 3 slices of bread down in order to soak up the suds, but as soon as I finished the third piece I ran to the bathroom and barfed. Then, all was right in the world and I passed out.   Which leads me to now. There are some ants in our kitchen here and today I\'m gonna get some ant traps. I also developed the Poison Froot Loop Theory this morning and I will test it out later. It is what it is.    urlLink Molly  is coming back this weekend for the N.E.R.D. show. So Molly,  urlLink Sarah , me and  urlLink Zac  will go for sure. Maybe  urlLink Brian . Maybe the Zimmer. Maybe  urlLink Jason  cause I just asked him.   Here are some funny jokes from popsicle sticks:   Q: How did the telephone propose to his girlfriend?  A: He gave her a ring.   Q: Where do bees go on their day off?  A: The wax museum   Q: What do books carry when they are away from home?  A: Pagers&nbsp;   That\'s it for now. I\'m outta here.   2ce  &nbsp; '),(24413,3833031,'21,July,2004','2004-07-21 00:00:00','   That is what Dr. Pfister the veterinarian said while I was there Monday. Brew is able to stand without falling, he is able to walk around without falling and when I saw him today, his legs had swollen back to normal size. In true Lance Armstrong fashion, Brew overcame the odds and decided to grab life by the ass and give it a goose. In true Lance Armstrong fashion he did rehabillatory exercise and took drugs in order to stand up and win. (sorry Jason but couldn\'t resist)  Regardless, I am totally psyched that Brew is doing so well since I saw him Saturday and thought he should be put down. Then the vet said he was even worse on Sunday. Then Monday he lost his fever and he could stand up for hours. He really wants to come home but the folks are gone and I won\'t be able to get him until Friday. His right eye is a little swollen and draining and it\'s still tough for him to get up on slick surfaces, and though I would love to have him back tip-top, I know that that is an unreality. I still want to get Brew laid before he dies so if you know any fly bitches hit me and I will get a hotel room, some Hennessey and some 9-1-1 shit for them to smoke on.  In other news we are gonna see N.E.R.D. next weekend the 30th. An OU Molly might come, which would be way cool. I saw her this weekend at Sarah \"Don\'t Try It\" Hiatt\'s house. Good times.   I am now being trained by my supervisor who is decidedly less hot than I found her to be before, though still attractive. I think it\'s cause she has straightened her hair every day this week. Come on! Women: if you have dark, curly hair don\'t fucking straighten it! At least not often. I know you get bored with it, but it is the best!  Here is this posts INTERACTIVITY:  I have been wearing these jeans for 12 days without a wash. Strangely, there is only one stain on them what is it and where?  A: What? Grass Ketchup Mustard Coffee Pizza sauce Chocolate Mayonnaise Something greasy: Butter, margarine, fat from meat, vaseline Ink from a pen Ink from a marker Paint Blood  B: Where? Crotch Ass Back of leg - left Back of leg - right Outer thigh - left leg Outer thigh - right leg Inner thigh - left leg Inner thigh - right leg Knee - left leg Knee - right leg Shin - left leg Shin -right leg Calf - left leg Calf - right leg  If you can form the correct combination of the aforemetioned whats and wheres you will score a most non-heinous prize courtesy of yours truly.  I gotta bounce. O\'Malley\'s in the alley beckons. Peace y\'all.  deuce '),(24414,3833031,'17,July,2004','2004-07-17 00:00:00','   My dog is ill. Like he is gonna be euthanised in a few days. I had a good cry about it earlier and I\'m sure I\'ll be really sad when it actually happens. So I\'m kinda bummed but I\'m gonna try not to let it get me down. He is pretty old for a dog and at about the stage in life where everything goes to pot. Barring some miraculous recovery, it is gonna be dust to dust for my dog and my friend. '),(24415,3833031,'14,July,2004','2004-07-14 00:00:00','   So last night was Dirr\'s last night to hang out before going going back back to Cali Cali. He isn\'t really going back he got a job in a National Park in/outside San Fran. He will be there for like a year minus Christmas. So last night Barb, Sarah, Nick and Zimmer came over. We were watching World Series of Poker (which I can now stand to watch) and then Jake Dirr came by to see Nick. Everybody but Barb and Brian played some poker with pocket change. I took a lot from Jake on the first hand, but he won even more back a couple hands later. Then he did the funniest thing I have seen in a while. He pulled all his coins to him, like 2 bucks in quarters and dimes and nickels and pennies and shit, then he starts to duck towards the coins and says, \"give you all a little of the Scrooge McDuck\" then proceeded to mimic swimming over the coins, like Scrooge did in his money bin on DuckTales. If you were there and paying attention you would have laughed so super hard. it was awesome. I was the big winner after Sarah totally bet it all with nothing because she drove and people wanted to leave.  Nick also left a headband at my house which I intend to wear while I drink diet beer. If Nick finds out that it touched my head while my hand was touching a Mich Ultra he will never want it back.  There is a universal machine in my apartment so I think I\'m gonna try to start to try to start to try to start to lift some weight again. I also wanna do some other shit like push-ups and sit-ups because I feel like I\'m letting it all go.  Say a prayer for Nick Dirr that he makes it California alright.  later  deuce '),(24416,3833031,'12,July,2004','2004-07-12 00:00:00','   Today was my first day at my new job. I was pretty bored. I have a cubicle. There aren\'t  urlLink hot chicks  walking by all day like at the mall. My supervisor told me that I would be working with 17 other people and that 15 were female. I was hoping some of them would be really hot. There\'s one girl that i think is hot aside from the supervisor herself, who is a total babe.  I just got the  urlLink Portishead live cd  and a  urlLink Stereolab double-album . So I was listening to some chill shit until I finally got my Strokes cd back  urlLink last night . I have rocked that album about 2 or 3 times since then. It is like only 45 minutes long and I always skip one of the songs so it\'s easy.  The one song I wanna download/have on cd so bad is \"(You) Got What I Need\" by  urlLink Freddie Scott . It is the song that Biz Markie sampled for his big hit, \"Just A Friend.\"  ummm...i\'m now too busy to write in this.  peace yo  deuce '),(24417,3833031,'10,July,2004','2004-07-10 00:00:00','   Here is what I HATE:  People say that marriage is sacred and they need to keep gays from having the right to marry because they think homosexuality is blasphemy or heresy or whatever or maybe they think unnatural. I kinda think the latter. But I won\'t get way into that and it\'s only really unnatural in terms of procreation. It\'s natural to love and/or lust after people and the people you love come wrapped in a different package or the same package or a different color package, whatever. It\'s your perogative.  But fuck all these whiny ass retard crybabying about how \"marriage is under attack.\" Two words assholes:  urlLink Divorce Rate . Oh, 49 failed marriages out of 100. That\'s top drawer. So sacred-ass marriage barely works half the damn time? Gee wonder why? Cause fools rush in? Cause people are so fucking superficial that they don\'t even really consider marriage sacred, they just do it cause that\'s what you do in life? Or cause they got knocked up/knocked somebody up?  If you truly devoted to \"saving marriage\" put down the fucking placards and instead of going to the protest with all the other asshole hate-mongers set up a nice little office somewhere with a big comfy couch and be a marriage counselor. Denying the right to marry to worthy people doesn\'t make you a defender or crusader of/for what is right, it makes you a fucking jerk.  Bet money that if gays are allowed to marry their divorce rate will NEVER touch 49%. I doubt it would ever reach half that. Why? Well for one, gay people can\'t go around knocking each other up when they are young and irresponsible. They don\'t face the social pressure to get hitched like heteros do. Gay people will marry the person they want to be with for the rest of their life, not just because it is the \"next level\" of their relationship.  As for George W. Bush\'s statements (get ready for some education): While George is taking some kind of moral (amoral, in my opinion) stand against gay marriage, it is purely political. Duh. It\'s election year. Gay marriage is what is referred to as a \"wedge issue.\" There are swing voters he is trying to bait into voting for him: left-leaning people who don\'t believe gays should marry and share the wonderful tax benefits married couples receive.  For the record John Kerry is not a proponent of gay marriage. Have you seen him on TV praising his home state for being the first to allow gay marriages? Do his ads boast that he supports gay marriage. No, that would be suicide to his campaign. On the same token, if he goes out guns blazing that he is against gay marriage, he may lose gay voters, who tend to be liberal. Duh.  Since Kerry was the senator of Massachusetts or Maryland or whatever and they were the first state to allow gay marriage, the general public (who enjoys their news and current events spoon-fed to them rather than reading or researching shit) can easily have the idea impressed upon them that Kerry is homo-sympathizer and wouldn\'t mind too much if gays destroyed the sacred institution of marriage. Boo-hoo.  Kerry has said that he does not support gay marriage. But he isn\'t going to be against it and lose votes. Duh.  The message here is: Gay people are people, too and they couldn\'t do anything worse to marriage than heteros have. If you think God hates homosexuals and is going to send them to hell then you don\'t know God and you shouldn\'t even speak His name. Get off. If you think Bush is a champion for morals you are retarded because he is killing people, ours and theirs for dollar bills. People with dignity and integrity aren\'t really for sale. I\'m sure Bush believes that gays shouldn\'t marry but it\'s only a fuss to ostracize voters from Kerry.  later  Anybody But Bush \'04 '),(24418,3833031,'08,July,2004','2004-07-08 00:00:00','   Just like an hour ago I was sitting here looking up pictures of  urlLink lotuses  on Google. Then I went outside to water my mom\'s eighty ba-jillion plants and my parents put this little pond/big puddle thing by the deck and there was freaking lotus bloomed out the mahfucker. I am probably gonna try to buys some lotus or lily plants or whatever and grow a bunch and when they bloom cut them and hang them to dry. I think that would be badass cause I think sissy shit like that is badass. Anyway, lotus flowers are mad symbolic because naturally the lotus grows up through scum and muck and shit on the bottom of a pond or a lake or whatever and then it comes out and is all beautiful. Kinda like a rose growing through a crack in a concrete sidewalk, you know? '),(24419,3833031,'08,July,2004','2004-07-08 00:00:00','   Last night I was drinking some High Life\'s and a double-deuce of  urlLink diet beer . I played an  urlLink all-time classic video game  and  urlLink dominated . I ended Dirr\'s 5 win streak and then beat Sarah with the Houston Rockets.  urlLink The Dream  worked the glass: boards, blocks and dunks aplenty. It was a good game I think I only ended up winning by three. Then we played a drinking game. Then we sort of drank leisurely. Then we went to the Leaf. It\'s a bar. So we had some  urlLink Sex With an Alligator  and then I did a  urlLink Jagermeister  shot cause I normally don\'t fuck with whiskey. Even if it\'s  urlLink pretty good .  Jason and Susie called but I was picking songs on the jukebox and therefore couldn\'t talk. But that\'s cool we\'ll talk later.  I have to play softball today and then over to Long\'s house to drink more. I have to play softball tomorrow then I think we are gambling again. Vice. Vice. Vice.  Here\'s a little ditty I came up with after passing some guys on motorcycles drove passed me:  Harley dudes, Harley dudes, Big, fat, leather jacket, Gnarly dudes. Get drunk, ride bikes, eat Quaaludes, Harley dudes.  peace y\'all   urlLink deuce '),(24420,3833031,'08,July,2004','2004-07-08 00:00:00','   I took the same  urlLink personality test  as Jason while I was on hold with Verizon trying to get an  urlLink LG VX4600 . The first 7 or 8 times I called, nobody could help me. It almost seemed like they  urlLink went out of their way not to help me . Then they sent me the wrong phone, like I could roll with a fucking  urlLink VX3200 . Pshaw. Where\'s the  urlLink Organic Electroluminescent  display? There isn\'t even an external caller ID screen to be found! Like I\'m gonna open my phone before I know who is calling me or I am gonna open my phone, find out who\'s calling me then push a button to talk to them. Sheesh.   My moms is gonna give me some loot to go buy a platform bed from  urlLink Bova . Seven locations in the country and Cincinnati is one of them. It boggles my mind, too. The new digs are dope and I will put some pictures up soon. That way, when you\'re daydreaming that you\'re in bed with me you will have a solid image to work from.  This test is bunk, bogus and/or booty. It says I\'m sober. It says I\'m a hippie, which is pretty right I guess. Whatev, there\'s some truth to it.  I should be receiving my VX4600 soon, if get the wrong phone again I might have to beat a bitch.  *hugs*  deuce   Wackiness: 52/100 Rationality: 50/100 Constructiveness: 64/100 Leadership: 48/100    You are an SECF--Sober Emotional Constructive Follower. This makes you a hippie. You are passionate about your causes and steadfast in your commitments. Once you\'ve made up your mind, no one can convince you otherwise. Your politics are left-leaning, and your lifestyle choices decidedly temperate and chaste.  You do tremendous work when focused, but usually you operate somewhat distracted. You blow hot and cold, and while you normally endeavor on the side of goodness and truth, you have a massive mean streak which is not to be taken lightly. You don\'t get mad, you get even.  Please don\'t get even with this web site. '),(24421,3833031,'07,July,2004','2004-07-07 00:00:00','   I decided a few days ago if and when I get a dog I am gonna name it  urlLink Spaniel Gompers . Don\'t even fucking think about naming your dog that. I know you wanna name your dog that cause it\'s the best name ever. Before you were thinking you would just name your dog \"Elroy\" like that shit was hot. It\'s not. Go back to Greenland with that mess and quit stealing my name for a dog.  Oh, yeah me and Brian moved some stuff into our place, with the help of Nick Dirr I might add. I am fixin\' to move in there soon.   love,  deuce '),(24422,3833031,'06,July,2004','2004-07-06 00:00:00','   k, I just tasted one of my dog\'s Canine Carry Outs cause they look like cute, little bite size steaks and since I mainly eat potato chips I am kinda craving protein and, like I said, they look like cute, little bite size steaks. I really only want some cookies and candy but there isn\'t any except I have a Cadbury egg from Easter and a chocolate bar from Christmas in my room but I don\'t want that dinosaur candy, I want some fresh new shit, ya heard?  So the Canine Carry Out wasn\'t too good, I really just licked it cause as I was about to bite I got a little gunshy. Such is life. It tasted like a crayon kinda.  peace and scrubbles,  deuce '),(24423,3833031,'06,July,2004','2004-07-06 00:00:00','   k, I\'m tired of saying life sucks, even if it does you\'re not supposed to say it. The truth is I like life a lot so much that I usually catch bugs and throw them out the window instead of crushing them because I think that a bug\'s life is special, too. You probly think I\'m a fag for saying that but didn\'t you see that movie, \"A Bug\'s Life?\" Oh, you did? FAG! If it\'s a bug that\'s mad grody sorry bug, forgive me God but I\'m gonna kill you or sic my dog on you. And if it\'s more than one bug then bugs ain\'t chillin\' they are trying to infest and invade my shit. Any invaders to my homestead will be met with hostility, especially if the invaders are little bugs. If some big black dudes were invading my shit they would probly be met with me peeing myself in the closet.  So today I\'m just saying I don\'t think life sucks. Well, it probly does more so for you since your not me how bout some shameless self-promotion? I\'m hit you\'re miss I\'m the shit you\'re not worth pissin on damn that shit was hot straight off the top of my dome.  I have a new job at a law firm in downtown Cincinnati and a new apartment in downtown the same city.  Also here is my hilarious new insult (it only works on girls) it goes like this: \"Shouldn\'t you be pregnant by now?\"  OH--here is some crazy entertainment news: Jeffrey Jones (the principal from Ferris Bueller\'s Day Off (also had a lead role in Howard The Duck--bellissimo!)) got busted for not declaring that he was a sex-offender. 8 year-olds, Dude. Actually I beleive he was 12 and Mr. Jones got the lad to pose for some dirty pictures. Welcome to Creeps-ville, population Jeffrey Jones. '),(24424,3833031,'05,July,2004','2004-07-05 00:00:00','   So I went to the  urlLink Argosy Casino  in Lawrenceburg fully expecting to lose a hundred and twenty smackers. It was me,  urlLink Nick Dirr  and Sarah \"Don\'t Try It\" Hiatt. Actually nobody calls her that, I just coined that nickname earlier today. Actually, instead of \"coined\" I want to say \" urlLink minted .\" You know? Like I just made that up and it\'s money. So, on the way to Dirr\'s I stopped at DQ on Route 4 aka Springfield Pike aka Dixie Highway. Do you think I stopped in for Hot Eats or Cool Treats? Duh, Cool Treats. I got a Blizzard with chocolate ice cream and  urlLink M&M\'s  (size medium). I ate the Blizzard thinking that it is probably my  urlLink lucky  Blizzard, and eating would cause me to win some money at the casino. I was either wrong or I was right. At one point I was ahead probably about $175 dollars at the  urlLink roulette  wheel. I ended up losing ninety bucks. So here is the lesson I learned today:  It doesn\'t matter how good you feel or how sure you are that you are doing good things being a good person (being a  urlLink winner ) because sooner or later reality is gonna hit you on the head like a  urlLink meteor  from outer space and remind you of what a huge  urlLink loser  you actually are and as soon as you forget that fact and start thinking you\'re alright it is gonna smack you down again so it\'s probably best if you  urlLink don\'t even bother trying to get back up.   Nick bought us food at Steak & Shake afterward because he  urlLink won  and he is a swell guy. Oh, and while we were there I saw the ex-boyfriend of a cool girl I had a  urlLink two-night stand  with. I\'m such a  urlLink dick . I suck.  I may go back to DQ for a Hot Eat: Sarah and Nick both spoke highly of the \" urlLink Flamethrower .\"  Life sucks.  later.  deuce urlLink '),(24425,3833031,'05,July,2004','2004-07-05 00:00:00','   I don\'t know why I am bothering to do this because nobody cares anyway. I am doing this in order to share with anybody who\'s interested (probably no one) all the pain I experience in my futile existence. Feel free to comment and share your own stories.  Life sucks.  later.  deuce '),(24426,3833031,'10,August,2004','2004-08-10 00:00:00','   what the fucking shit is happening? at least 1ce a day some weird ass number calls my cell phone. If it isn\'t in my caller ID IDon\'t answer it! I am getting crazy ass area codes, where is 404? I\'m about to do a reverse look-up on whitepages.com. I always wanna know who\'s calling me. It is probably long on Jen\'s phone in Cali, cause that happened the other night. Never mind they left a message...It was Dirr! Timeout.  k. so while we were at the bar on Saturday, I\'m at the urinal taking a piss when some jackass comes in behind me and tells me I have a bald spot. He goes on to apologize and says it must suck cause I\'m so young and going bald. So while I piss I try to make this motherfucker understand that I do not have a bald spot and that I\'m not going to be bald unless I undergo chemotherapy or something. So I\'m standing there and I point at the back of my head and point and say \"right here?\" He is like \"yeah.\" I\'m like \"that\'s a cowlick. Look dude, my stylist will tell you that I have a very full head of hair that isn\'t going anywhere. The gene for baldness is passed down on the mother\'s side of the family, and all of the men on my mother\'s side had hair. It\'s called a cowlick.\" He starts some condescending apology and I\'m like \"find me in 30 years and I will have a full head of hair.\" I kept myself from saying, \"oh, but in 30 years you\'ll be dead\" cause the dude was older and it was really hard because the only reason I said \"come find me in 30 years\" was to finish with \"oh, wait, you\'ll be dead by then.\" either way, don\'t fucking piss me off about my hair, if you don\'t think it looks good you can say it but don\'t say I\'m fucking going bald cause if you say that to me you might as well just say \"hey I don\'t know my head from a fucking shitpile\" cause you will be so fucking wrong. and if you can live for like 40 years and not know what a cowlick is you might want to consider whether or not your head actually is a shitpile. Use the fucking tool between your two shoulders or be a fucking tool. Fuck the dumb shit.  peace  2ce '),(24427,3833031,'08,August,2004','2004-08-08 00:00:00','   I just read the BLOG of my compatriot, Jason Fryer, and have to say that working full-time has not had a maturing effect on me at all. Lemme give you a run-down of my after work activities.  Monday: Poker from 8:30 pm to 1:00 am. *plus driving back and forth between downtown and the wild, wild westside Tuesday: 7 pm--Drove to my bank in Mariemont (the closest branch to me now, no shit). Then to the Cure concert until 11. Back home by 11:30 or so. Wednesday: Me, Zac and Brian went to go to the Red Cheetah but the lines were absurd. So went to some other bar until 12:30 and walked home for like 20 minutes. Thursday: Poker again--this time I didn\'t leave until like after 1:30. Friday: Tina\'s for at night. No food since lunch, cannot believe I didn\'t yarf. Saturday: Brian and I went to Tina\'s at 5. He left at 7. I stayed until 10--drinking non-stop. Then went back to my apartment for a few drinks then went back to Tina\'s.  So I am getting my seventh day rest on. Annie and Michelle and my friend Brant came down so we hung out with them. I am feeling like major shit but I drank for like 9 straight hours yesterday. I drank like it was my job. Today I\'m gonna go home for dinner and do some laundry and pretty much take it easy.  I am out for now--God bless  deuce '),(24428,3833031,'05,August,2004','2004-08-05 00:00:00','   I was walking into work yesterday, and coming up the stairs I ran into Jennifer, who is in the same department as me, she has a cubicle down the way from mine and I said hello. There is a door on the left, a revolving door in the middle and another door on the right. She was heading toward the door on the right up the stairs and I had been coming from the opposite direction so I was heading for the door on the left. After I said hello she came across the stairs to the door I was going in, eventhough she could have went in hers in like 2 seconds and continue any conversation we were having inside.  She came across, I opened the door for her and asked \"how are you?\" To which she replied, \"Pretty good...for a Wednesday.\"   NOW WHAT THE FUCK IS GOING ON?  For a moment I thought she could have found this broke-ass little web journal I have and read about the \"for a Monday\" guy and said that for a goof. Then I realized, \"no fucking way that happened\" and came to the conclusion that she just says lame things like the other guy, who I will refer to as Bob until I learn his real name. He seems like a Bob. Bobbish, you know? It makes me wonder how one building can contain so much lame, but it does. Tonight I\'m gonna go to the grocery store and then I am gonna write something I hope.  deuce '),(24429,3833031,'03,August,2004','2004-08-03 00:00:00','   so working in an office building is thus far just like office space. a couple of wednesdays ago I was gonna meet Geers for lunch, then I said I couldn\'t, but then I could again, but I couldn\'t call him before then to tell him I could again cause he was at work and I didn\'t know his number there. So I went to where we agreed to go thinking he may have gone there anyway. He didn\'t. So I ate there by myself, no big. But some guy, who is also eating by himself, is looking at me. I notice it and he says \"hi, how are you\" and I have no fucking idea who it is. He is sitting directly in front of me facing me and there is about 15 feet and two tables between us.  So I walk into the building the next morning and there he is. Duh! He is the door guy in the building I work at and he also helps conduct elevator stuff to somehow not that I gave one or two about that cause even when I was on the fourth floor I took the stairs cause if you notice, most people working in office buildings have big asses and its cause they fucking sit on them all day and they don\'t even take the stairs for one fucking flight. Not me. No need for a large ass. Never.  So this guy is my door guy and he said hi to me at lunch, so now I always say hi to him when I walk by.  So not this last Monday but the Monday before, the 26th of July--THAT Monday I was walking in and I said \"hi, how are you?\" and he says, \"Pretty good...\" then he smiles all devious like he is up to something clever and finishes with a delayed \"...for a Monday.\" And I gave I big smile cause he\'s smiling and he said hi to me at lunch and he\'s a nice guy, and I even let out a very fake bit of a chuckle, then opened up the door then went upstairs.  Now this last Monday, yesterday, August the Second, I\'m walking in to work, see door guy and I\'m like \"Hi, how are you?\" and he is like \"Pretty good...\" evinces the same smile and adds \"...for a Monday.\"  Now let me tell you, for someone who actually experiences deja vu (as recently as Monday afternoon- hours after this incident, incidentally) I can clearly say that my experience was NOT deja vu...it didn\'t even feel like deja vu. Before I even pushed open the door to go up the stairs I was talking to myself (another testament to my sanity) and I\'m like \"For a Monday? That\'s all you\'ve fucking got? The same tired-ass line you probably use on every single fucking person that asks how you\'re doing on a Monday?\" Cause you know he uses it everytime, and he smiles like the payoff is so sweet, like it is so goddammed funny that I\'m gonna be crippled with laughter, probably miss the rest of the day at work yeah right like why would I wanna fucking be there anyway cause it\'s Monday. Ooh, look the fuck out! Here comes Monday right after Sunday but still, oddly, before Tuesday same place it has been since just after motherfucking time started and people started keeping track of shit.  I hope to God they don\'t cancel work on Martin Luther King Day cause when he asks how I\'m doing I\'m gonna be like, \"Pretty good...for a Martin Luther KIng Day.\" Nobody ever says that shit.  K, I\'m off to the bank then to get a sixer to slam in the parking lot before the Cure play.  Peace  2ce '),(24430,3833031,'02,August,2004','2004-08-02 00:00:00','   Here are some links to a couple other blogs I just read about in Newsweek. If you are a nerd like me you will love them, but I haven\'t even looked at them yet, but I can\'t wait to. I can\'t hardly wait, Aman-duh! haha.  the first one is  urlLink PEA Soup . The second one is the  urlLink Garden of Forking Paths . It\'s like a party for your subconscious mind. I think. Like I said I haven\'t looked at them yet, but I can\'t hardly wait, Aman-duh! haha that never gets old.  Dude imagine you are some guy who paints grass that grows and then you watch it dry. If that was your job then your job would only be 10, 000,000,000,000,000.3% less boring than my job. I need a new gig. Seriously, this 9 to five shit is fucking with me. I need a constant state of weekend. I guess that means I have to become a bartender or start learning to play an instrument aside from the air guitar and skin flute (did somebody say \"virtuoso\"?).  Alright time to go back to boredom city. I paid a bill on my lunch break! I hate my life!  love  2ce '),(24431,3644724,'28,June,2004','2004-06-28 00:00:00','   Catchy title, huh?  Now, I am a Yankee girl, who grew up with the notion that roaches are a sign of a dirty home.  I never in my life saw a roach until I moved South where the climate is more inviting to them.  I do the quarterly pest treatment, and feel pretty confident in my housekeeping.  Unfortunately, I am married to a man who sets out a midnight roach buffet before retiring.  My post dinner cleanup includes washing all dishes, wiping/cloroxing down all surfaces, and vacuuming up all crumbs.  Dog dishes are emptied, garbage taken out, and food is put away.  I usually go to bed before DH is done watching Monster Garage/Pimp My Ride/American Chopper/Overhauled.  I wake up to a living room littered with Little Debbie wrappers or a cereal bowl, and a glass of milk or juice on its side with just a tiny dribble leaking out onto my carpet.  The kitchen counter usually has a trail of crumbs or wrappers/empty boxes that just can\'t make it those last few inches to the garbage can.  I don\'t care about the cleaning up after DH, but I don\'t see any point in hanging a welcome sign on our house to my German, Oriental and American \"friends\". My husband mocks my paranoia - he grew up in FL where you can probably saddle up a roach and ride it around the house.  Here are MY experiences with them.  1.  I used to work with a young woman of questionable hygiene, who announced to me one day that she had to leave her microwave outside in the sun.  \"Why?\" I asked thinking it was just one of her quirks.  \"Well, it\'s full of roaches.  I mean, I can see them behind the number pad.  The sun should kill them.\"  Swallowing my disgust, I asked the obvious, \"Why don\'t you just buy another microwave?\"  \"Nah, I can still use this one.\" Note to self:  don\'t eat anything she brings in to work.  2.  We went on vacation to Myrtle Beach, and around 5 am the first morning I felt something on my cheek.  I thought maybe it was a moth or even a fly, but when I brushed it away, it felt pretty substantial, and I could hear it hit the carpet.  I flipped the light on, and saw that it was indeed a roach - ON MY FACE!  After waking up DH and probably our neighbors on either side of our room, I scrubbed myself down, and called around town to get another hotel (not easy during Bike Week).  I waited until 7 am to go downstairs and cancel the rest of our vacation.  I tried to be discreet to the clerk at the counter, but she kept arguing with me that \"That\'s not a roach.  That\'s a Palmetto Bug - that\'s our state bug!\"  I\'ve never heard of a state bug - does South Carolina also have state vermin?  Palmetto Bug - a roach with an endorsement - unbelievable.  3.  I have found these critters in my dishwasher, utensil drawer, front porch, back patio, and upstairs hallway.  A sighting - though VERY RARE - sends me into a tizzy that only my sweet bug man, Tim, can soothe.  He assures me that I have no infestation, and that it\'s just a stray that gets in through crack or opening in the doorways and heads to a water source (hence the dishwasher sighting which now forces me to use the heat dry cycle to nuke any infiltrators).    Truthfully, I really only see these critters outside - where it\'s to be expected. But I still remain ever vigilant against them crossing my threshold.  As Tim says to me, \"You Northern people sure are funny about roaches!\"   '),(24432,3644724,'22,June,2004','2004-06-22 00:00:00','   Ahhhh...summer vacation.  Time to kick back, relax, and have endless fun with my children - just like the pictures in my latest Parents magazine.  The sun will be shining, our yard perfectly manicured, and my children and I will all be tanned and cute in our coordinated summerwear.  Ahhhh...summer vacation.  It\'s raining - again.  Our \"Simple Set\" pool, is leaning - again - and we\'re looking at our fifth go around at draining/refilling. My two oldest kids are at each other\'s throats, and I\'ve finally had to implement the \"kiss and make up\" rule to cut out the squabbling. T-ball has ended, so my contact with other adults will be limited until school starts up again in August.  As for my tan, when I take off my t-shirt at night, it looks as if I\'m still wearing one.  No one wants to hear just whining, so here are some of my bright spots:  DS has learned to tie his shoes, DD#2 has learned to walk, and as I type this she is scaling the mountain of books and magazines by my nightstand, and I\'ve taken the first baby steps to pursuing my teaching certification.  Oh, and today, I\'m visiting my \"Happy Place\" - Target!  I love Target, as do most other stay at home moms.  I think once we get discharged from the hospital with that new baby, there\'s a homing device in those hospital issue receiving blankets that propels us to the red bulls-eye.  When I first became a SAHM, I had to watch costs, so I frequented Wal-Mart.  The upside of Wal-Mart is that you can get EVERYTHING you need - food, clothes, tires, etc.  The downside of Wal-Mart is that it\'s not as peaceful as Target.  If you\'re escaping from chaos at home, Wal-Mart is chaos magnified.  It is crowded, noisy, and in need of a good cleaning - not unlike my own home.  Target, however, is an oasis of calm.  The minute you step through the automatic doors you are hit with an icy blast of AC.  It is well lit, not glaring, the floor is actually white, items are in order on racks and shelves, and I have yet to see a mother/father beat their child over a Tonka truck.  This is no exaggeration, I can actually feel my breathing and pulse rate go down when I walk in to the store.    Well, I would love to write more, but Spongebob is almost over, and I need to get ready for my Target outing.   '),(24433,3644724,'16,June,2004','2004-06-16 00:00:00','   I\'ve never blogged before - except in my own head.  I hope you\'ll enjoy, be amused, interested,  and/or agitated by my tales of stay-at-home-martyrdom - I mean motherdom - and various opinions on just about anything .  I\'ll try and keep it interesting!  More to follow...   '),(24434,3099605,'26,April,2004','2004-04-26 00:00:00','   Greetings mortals!   Ben Here!!!!!!!!!!!!!!  Today was a day about boring videos that are made needlessly dramatic by a traveling Violinist who im sure was just hididng in a draw above the TV!  Plus in another Note graphics really sucs and whats scary is i think our teacher likes this fact!  ARRR BAD DAY MY HEAD HURTS!  HERE IS A GAME!   urlLink DESTROY CARS SAMBI IS DRIVING!  '),(24435,3099605,'25,April,2004','2004-04-25 00:00:00','   Greetings Mortals!  Webmaster J once again blogging one up for absoulutly no reason!   Ive been upt he alotment today totaly out of my mind digging all morning and it got me a thinking about gardens and how deeply underestimated they are at being one of the coolest places ever!  Stop before u go stright to the game and let me finish, i mean in the garden its warm and there are colours that arent grey!  Plus you can perfect my fine art of tipping and relaxing with a beverage of your choice!  May i recommend my latest find which is bacardi white rum and iron brew :P its dam good for tipping and relaxing!  Also if you in to it (we know Craig is) you can create your own little piece of pixy living heavan in a shroom patch!  Also perfect places to drag out some speakers and play rock music really loud!  because in the garden nobody cares because they all either tipping and relaxing or on shrooms!  I mean they are ace!  Plus they coolest on a summers night get candles and look at the moon!  Howl!  lol well for all your deeply indoors types or peopel who aint mad!  dididididididididididiiddi GAME!   urlLink So Weird Its Gotta Be Ace!  '),(24436,3099605,'24,April,2004','2004-04-24 00:00:00','   Hey Me here again  About half an hour after my post earlier today somethign deeply weird happend to me.  I was walking past the TV in my lounge then for no reason the TV turned on.  Now this was very ring esc and kinda creepy but i swear ot was true.  Then i noticed that all the clocks pluged into the mains in the house were claming they had a power cut by flashing on and off, but the weird thing is, they were and still are registering that the power cut was 20 mins int the future!  Now this didnt end the weirdness because i had to take my dog for a walk so i put my cane down to put his lead on and walked out the door with him, then i remembered i had left my cane so went back to get it.  But it wasnt were i had left it!  I search the house for this cane but finaly gave up and went back to the kitchen to take my dog for a walk, and then my cane was were i thought i had left it, 20 minutes later...  I swear the cane wasnt there when i started looking and how the fuck can clocks register a power cut 20 mins in the future!  So i am kinda a bit freaked.  '),(24437,3099605,'23,April,2004','2004-04-23 00:00:00','   Hallo Thar and welcome to Webmaster J once again unplugged and live from his house!  Ima really bored today wandering around my house, tired (stayed up till two in the morning last night watching Bram Stokers Dracula)  and kinda depressed so anyway today i thought id have a bit of a flash game spectacular.   urlLink Merlin!  Easy But Fun!   urlLink Merlin 2!  One Of The Hardest Flash Games You Will Ever Play   urlLink Tetris!  Has a greater thing came out of Russia?  Enjoy!  '),(24438,3099605,'23,April,2004','2004-04-23 00:00:00','   Hey Everyone Webmaster J Here!  Had a alright day today for once *shock horror!*  Wish i had a ale however as the patrioticness even got me into its lair, stupid thing...  However im very much aware that people at Homestarmountney arent aware of two of my weirdest creations,  1.The short musical \"Pontieus pilots flying pirates\"  and number 2.  My nu national anthem to the tune of land of hope and glory, \"Well done Charles.\"    Now not that any of you care but here it is!  Well done Charles Winsor! Meeting Camila at three! Ohh how we adore thee! Who is greater Than Thee? Greater and Still greater! Your Bounds Shall Not Be Set! Our Money That Made Thee Mighty Make Thee Mightier Yet!!! Our Money That Made You mighty Make Thee Mightier Yet!  Isnt it Great!  But this blog gets better Yet Because here is a game!   urlLink Retro Solid Fun!  '),(24439,3099605,'22,April,2004','2004-04-22 00:00:00','   Greetings Mortals Webmaster J here once again!  On the second day back at school we all realize how boring school can really be as the hot and cold weather, craigs addiction to shrooms, making funny gas in chemistry and the theory about being thirsty in period 5 rear there heads.  Yup its been a bad day woah.... sorry.    Well hey ima just wittering at complete and utter random so before i go back to thinking about how great it would be to stab Alex \'s fat asse here is a game!   urlLink Topical Zombie Murdering! Imagine Your Worst Foe And Blow Shit Out Of em! Mwhahah!  '),(24440,3099605,'17,April,2004','2004-04-17 00:00:00','   Hey Everyone Webmaster J here, not that any of you read this however!  Today im thinking about the lifespan of Homestarmountney and what people of expect of you.  Thats the problem if you make something other people enjoy.  Its a faster, harder fall from when your at the top.  Hey we try our best, but two people cannot be funny all the time!  So its time to tell us what you want.  if you want the site to continue, start giving constructive criticism, or if you want the site to be removed i will remove it.  So just decide ok!  No game today... just want answers. If you really wanna game just contact me i will get you one.  '),(24441,3099605,'16,April,2004','2004-04-16 00:00:00','   Hey Webmaster J Here Once Again Live and Unplugged in his House!  Same rules as always games at the bottom, blogs at the top you know i love you all really so go straight to the game if you like!  Right today ima museing about intuition!  Have you ever thoguht to yourself for no reason \"ive got to to do (Whatever) because its important.\"  Then when you do the thing, you realize that it was compleatly pointless, hmm  why the hell does that happen!  I woke up at 8:00 this morning to do somethign that was compleatly pointless!  I mean does someone like brainwash you while u sleep or what :S  well anyway i believe its just some evil goblins way of playing with you.  Well they are locking me back up now so ta da!  GAME!    urlLink Bite FBI Goons Heads off! YAY!  '),(24442,3099605,'15,April,2004','2004-04-15 00:00:00','   Hey Its Me Again, well you would be shocked now if Webmaster JT posted lol well anyway im about to witter so if anyone really hates this bit or hates the very sound of my voice jump to the bottom and grab the flash game!  So anyway i was thinking today as i slept till 12:00 about how wouldnt it be great to be nocturnal!  Because we would have the night and the preps would have the day!  It would be even cooler because they would be afraid to come out after dark, like the pesants in Dracula, hehe i wonder what kind of weird subistions they would have, like \"If you shout Bloody Benny 58 times he comes and drinks your blood!\"  Woah that would be sooo cool well anyway i think im done.  Behold a game!  urlLink Gold Miner!  '),(24443,3099605,'14,April,2004','2004-04-14 00:00:00','   Do you like Burberry pattern stuff?  A) Yes Love it! B) No its wank!  Do you hang around on Park benchs?  A) Yeah Nice and Warm On the Arse! B) No  Do You Like R&B Music or any kind of Rap Based Shit?  A) OHH GOD YES  B) Rap Is Crap  Do You Think Its Cool to walk around with your hands down your pants or t-alk liek this liek?  A) Yeah, m-ate sorted liek do u wanna g-o d-own chippy and h-ang- ar-ound like? B) No  Are you a homophobic?  A) Yes B) No  Do you just wanna fucking stab anyone who listens to rock Music?  A) Yeah fucking grebs! B) No Way!  Do you follow fashion like a religious lunatic?   A)  Ohh Fuck Burberrys out of fashion RUN! B)  If it looks good wear it  If you answered Mostly A\'s then your a Prep and we got nothing to say to each other.  '),(24444,3099605,'13,April,2004','2004-04-13 00:00:00','   It seems our friends over at adamburton.tk they have been studying the changes we make to our site with such great detail, despite the numerous times they have denied copying our site- saying how bad it is, they have finally succumbed to way of the \'Mountney\'. Less than a week after we got a new guestbook powered by bravenet they have also been onto bravenet and got a new guestbook. It finally seems that after weeks of taking the rip out of us and our site to hide their own inadequacies they cannot resist stealing our ideas to make up for their lack of imagination and originality.  '),(24445,3099605,'12,April,2004','2004-04-12 00:00:00','   Greetings Mortals Webmaster J here once again!  Today ima gonna talk about Google, it has come to my attention that this is the most evil program in the whole world!  I mean think about it, it knows everything!  I mean type in Ben Jeffrey in a google image search and there is a pic of me!  its really scary because it knows everything about everyone!    Lol well think about that and meanwhile Play a game!    urlLink Kill Ur Friends!   Lol so upload a picture of someone and waste em!  The funniest ones i find are Me and any prep!  '),(24446,3099605,'12,April,2004','2004-04-12 00:00:00','   Hey Webmaster J wittering again :P  I seem to not be able to stop blocking because is just a fine thing to do lol and today\'s little witter is about life in general.  I mean come on its the easter holidays yet we are all bored, why is this always the case, it really sucks if you ask me!  (lol cant you tell im fustrated)  Well hell anyway sure beats the ever loving crap out of school!  Hehe but i know you didn\'t come to hear that little witter, Here is the game!   urlLink Another Game With Fish!   Yes i know i like games with fish!  '),(24447,3099605,'11,April,2004','2004-04-11 00:00:00','   Yay its easter!  Hope you all got loads of choccy eggs and the like and your enjoying yourself!  heh just a quick Muse today about my fave words!  Chocolate, Rhombus, Mon Dieu, Blood, Spectacular Tequila, awsome, Book, sweeb and probably crantastic.  Here is a game!  urlLink Juggling has never been so crazy!  '),(24448,3099605,'10,April,2004','2004-04-10 00:00:00','   Hey Webmaster J again.  I was thinking today about animals and how they like link with Prepitude i mean even Adamburton.somecountryweneverheardof talk about their hero burt being a monkey!  I mean they hang around in packs like mangy dogs, they insult peopel ina  primitive fashion despretly trying to be pack leader and climb tree\'s and hurl verbal dung on people from miles away just in case they get wasted.  its all about survival with them.  Lol so in conclusion as darwin probably did say when under some big ol drugs,  First the monkeys, then the preps!  '),(24449,3099605,'08,April,2004','2004-04-08 00:00:00','   Hey all Webmaster J here hope you all taking pleny of breaks from revision if not stop revising now!  Just wanted a quick witter about Preps today, Joes had one so now its my go.  Id just like to say that Preps have been getting a bit overreactive at the moment, seeing every little thing we do as a challenge, but hey it aint a challenge because to be a challenge the preps have to give us some compatition.  So smile my homestar based chums and keep on laughing and we will keep on making.  P.s.  Flash is this kinda stuff Adam Burton Guys  urlLink Sonic The Hedgehog!  '),(24450,3099605,'07,April,2004','2004-04-07 00:00:00','   Webmaster J Here  Im having a relaxed aproach today towards the old revision stick and now im bored as hell so i thought id blog one up for all you little Homestarmonkeys  I was thinking about the Prom today, and i hope all you Homestar followers are gonna come show the preps up by dressing like you mean it!  I mean come on the compation is ties tucked into their pants or burberry suits so come on make me proud, come dressed up to the nines!  In other news i have once again hunted a quality piece of flash out for you in the form of this cute little scrolling Jap shooter come on whats stoping you?  Waste some time with me and in the words of the young suave prince of town \"Crank it senoir!  this heres the joint!\"   urlLink ROBO BOMBO!  '),(24451,3099605,'06,April,2004','2004-04-06 00:00:00','   Just thought id let you know how to spot a prep, I mean with so many around this place you need to know how to spot one. If you know anyone who fits these characterisations then they are a prep:        pathetically self-obsessed, bitchy, insult people to boost their over-inflated ego, cares more about their appearence than anything else, homophobic or thinks that the \'coolest\' person is the one who can drink the most. I am sure you are aware that this describes the nobheads who keep writing www.adamburton.tk in the guestbook. What a great way to advertise the originality of your website.  In other news: Go Wigsy!  visit  urlLink www.cheese.com  -Its Jam Gut!  '),(24452,3099605,'05,April,2004','2004-04-05 00:00:00','   Hey Webmaster J here  I was despretly trying to distract myself from revision when i found this game, its adictive clean fun enjoy!    urlLink Fish Game! (Warning Mega Adictive)   So anyway, try to have a break you guys :P  '),(24453,3099605,'05,April,2004','2004-04-05 00:00:00','   What many of you may not know is how Mountney carries a leather purse wherever he goes. It is in the style of a medieval, outlaw, vigilante that he thinks he is. He rarely keeps any gold in his purse but what he does keep is very rarely spent unless on a new quiver or bowstring.   Continuing with ye olde style Jeffrey has a Cain which he will use to keep the younger folk in line or (hopefully) use to bash preps!        Btw todays fun fun link is paying homage to mine and Webmaster J\'s master  urlLink Rather Good!  '),(24454,3099605,'05,April,2004','2004-04-05 00:00:00','   Hey Webmaster J here fresh from two hours of Science Revsion on the ever lovin SAM learning... Yawn.  Hey so i was thinking about revision and i know you all gonna be a bit down, but hey uncle Homestarmountney provides, so have a few minutes break with some coffe and a choclate and play this game!   urlLink Gross Bird Game   It\'s gross but fun, may say a little something about me but hey it\'s a good distraction so anyway ciao amiegos!  '),(24455,3099605,'04,April,2004','2004-04-04 00:00:00','   Yup this is Mountneys and the Webmasters Museings!  So all you Homestarmountney heads enjoy this yeah!    So Webmaster J\'s having a Muse today about The GCSE\'s and were Mountneys gonna go from here!    U\'c i think Mountney will have to use his mock exam results because he will go on a big J2o fuelled bender the night before and be sick for weeks!  Then he will feel remorse for making sweet love to the Conservative clubs pool table!  In Other News My Site Of The day is!   urlLink Virtual Bubble Wrap!  '),(24456,3099605,'28,May,2004','2004-05-28 00:00:00','   urlLink    ANGST!  Hehe Just thought id sum it all up before i go on Holiday. Well anyway pictures say it best    Au revoir.  Eeyore is Me!  '),(24457,3099605,'27,May,2004','2004-05-27 00:00:00','   Heya!  Just been a watching TOP2 and i was almost in tears laughin when the announcer dude claimed that American Pie is the original Teenage Angst song!  Well forgive me if i dont adopt that as my theme tune but hey, made me laugh anyway.  You can also forgive me if i dont go on about the Vietnam war, Nuclear Fall Out, The Death Of Buddy Holly and Bob Dylan :S    To many people on drugs back then :P  Lol well anyway on with the show!  We have a week holiday woot!  Im hopeing i can blog quite a bit with this is internet cafe thats down there but hey im also gonna attempt to not let angst spoil my holiday!  But hey it wont it will just put a human spin on it!  But anyway im off on MSN now so probably talk to you now!  '),(24458,3099605,'27,May,2004','2004-05-27 00:00:00','   Heya everyone!  Woah well just finished the English exam this morning, which was annoying and as always pointless.  The poetry as always is just a pointless annoying collision of questions which i dont know about you but really binds me.  But hey anyway good luck for all you intergrated and technology people this afternoon (especialy to the poor souls who have both!)  As one who has neither i have no clue whats so ever what im gonna be doing this afternoon, probably a blend of listening to hard rock, writing and skulking round MSN.  Lol so its business as usual!  Now i know ima gonna blog more than once today but i just thought id say now im going on holiday tonight!  Yay so the good nu\'s for u is that you wont see me for a week!  But hey im still hopeing to blog a little, (there is a sweet internet cafe were im going)  So hope you keep reading and plz check out mine and Balds website he\'s put A LOT of quality work into that so plz respect it.  Plus you can play flash games in the window on the site, which i personaly feel is awsome! Lol  Well anway im sure this wont be the last you see of me today!  (Yeah and btw Vermilion pt 2 is currently my obsession song :P)  I\'ll be back!  Au revoir enjoy the game!  urlLink Its Metal Slug, Nuff Said   ---------------------------- All I Need To Make It Real Is One More Reason. (Yay Bald Style Quote at the bottom!)  Or the alternative  ---------------------------- First it was nothing then it exploded  '),(24459,3099605,'27,May,2004','2004-05-27 00:00:00','   Heya!  Im quite shocked because today when it started off i thought i would be (normaly would say a word Begining with U but peopel are on orders to hurt me if i do)  Depressed but hey im not!  Woo!  *Hands out cookies!*  Yeah well anyway today i am in a happy mood so this is well shocking and ive basicly got 2 obbsessions today Cookies and Slipknots nu album!  Woot what a strange combonation they are.   Well im glad im having a angst free day and i hope you all are to and you enjoy the pleasures of a cookie tonight or some kinda round shaped dough creation with chocolate in it at least!  Well anway Extra good luck to you all for ur English exam tommorow and i know you will all pass because ur all great!  Sorry its small but yay!  *hugs you all and hands out more cookies*  Au revoir!  '),(24460,3099605,'26,May,2004','2004-05-26 00:00:00','   urlLink    I bought Slipknots nu album today after hearing tales about it being very diffrent to the others so here is the Ben Jeffrey Review!  I did enjoy the album its very grown up of Slipknot because in their latest offering they seem to care about the actual music a lot more.  I guess its because they have taken a lot of lessons from there solo projects of Murderdollz and Stone Sour.  There is deffinatly a lot of influence taken from Stone sour especialy, with its heavy rifts and roaring noise being more organized and melodic.  But all you die hard orignal Slipknot fans wont be disapointed because you can tell even though they have grown up its still Slipknot.  I would say this a good quality Angsty album were Cory shows off his true vocal talents.  This is especialy shown up in the flagship track Duality.  However if you want original Slipknot i would say the song for you is Pulse of the Maggots, i dont know about anyone else but it just didnt do it for me that track.  If you dont wanna buy the album i strongly reccomned you at least download these tracks  .Duality .Opium for the people .Vermilion and Vermilion pt 2 .Circle .Before i forget .Danger-Keep away .The Nameless  Over all i was impressed with this album however some songs are a bit of a let down if you want your music a bit more in depth.  They just seem to miss the point sometimes, however well done Slipknot you did entertain me a good quality solid album.  I Push My Fingers Into My Eyes  '),(24461,3099605,'26,May,2004','2004-05-26 00:00:00','   urlLink    Lol yup its festival Season and i found this to make me look as though im happy so when im Angsty later it wont seem so bad!  Today i feel like destroying something beutiful  '),(24462,3099605,'26,May,2004','2004-05-26 00:00:00','   urlLink    YAY our Website Is Up And Running!  Still needs some bugs fixed but still yay!   urlLink World Of Angst  '),(24463,3099605,'25,May,2004','2004-05-25 00:00:00','   Heya.  Its getting later now and i feel it coming on again so once again im gonna have to say some things :P  Yeah it seems the past few days have been pushing my sanity, everyday i see carefree individuals having the time of there lives.  Im happy for them, but i wish i could just once be one of em.  Yeah ive never fitted in anywear, i know this, i think its because i am in truth a little bit mad.    My mothers always told me im mad, and sometimes when i start to realize this for a time i say the strangest things and also do the strangest things.  Lately ive been writing more again, which im happy about... but im also not, because all of this sudden work is fueld by me and my fucked up illusions.  I am paranoid.  I sometimes think that people who are my mates are out to get me.  For this i am truly sorry.  It is just the way i am.  Anyone i call friend i truly do love and im thankful for you being there.  But anyway on with this, im not done yet.  Sometimes of late i feel like ive been pushing against a barricaded door trying to be with everyone else, but it is not others who barricade the door.  No its me and my own insecuritys.  I sometimes worry that im not just a little crazy but really mentaly disturbed somewere, i hope im not.  But i dunno, everyones a little crazy as people always paraphrase in cheap american shows.  But hey.  Woot.  Id just like to be there sometimes and just belong to something, but i also know i couldnt, strange isnt it.  Ive never been good in big groups, give me a one to one conversation anyday, but hey as if anyone would after reading this :P  Thankyou for listening im done.  '),(24464,3099605,'25,May,2004','2004-05-25 00:00:00','   Heya.  I all hope you had a good day today and if you didnt do the RS exam today you dont know how lucky you are!  Well anyway today i feel a bit down now, went for a bit of a walk and thought about a few things.  Just a bit deppressing really but hey thats the way things are with me.  I dunno really it just seems like im walking down some dark place at the moment, i know its going somewere weird bu i dont wanna turn back.  It\'s kinda strange really but im sure i will get over it, i guess its just some nu forms of Angst to tackle and let it get back into its normal state of angstyness.  I guess you could say it was rogue angst.  Well anyway it seems to be takeing me a long time to write not a lot, its very odd, just feel opressed.  But anyway id just like to use this bit to say how much of a knob head Mr Paine is and how i hope he\'s really unhappy for the shit he has put us though.  Hey well i will probably blog later when i got some more understanding of my present emotion, i gues i just feel kinda down, like im knocking on a door and nobodys letting me in even though there is a shadow in a lighted window.  Au revoir for now.  '),(24465,3099605,'25,May,2004','2004-05-25 00:00:00','   Heya!  Omg the strangest thing ever is happening tonight, ive got nothing depressing to witter about!  This feels so strange because normaly by this time at night im entering maximum Eeyore drive!  But hey... you hear that... nothing!  Its really strange and im just expecting it any minute to jump on me and go \"Suprise!\"  like some dememnted Jack in the Box jumping of the wall of a pokey little alley way.  (lol if i had talent i would draw that picture becaus for some reason that image is very Vivid in my mind!) So if any artists fancy a commision i will love them forever... well actually if you read this blog i already do so i would Doubly love you forever!  So anyway thankyou for listening here is your game!   urlLink Just looks cooL!  '),(24466,3099605,'24,May,2004','2004-05-24 00:00:00','   urlLink    Mmm Yes...  I Feel Stupid and Contagious...  '),(24467,3099605,'24,May,2004','2004-05-24 00:00:00','   Heya!  Welcome once again to your one stop shop for Teenage angst!  Just had my German listening test the typical affair of some creepy German Shouting likehe\'s in a Nazi propaganda video.  *shudder*  But anyway for i know this guy could have been shouting directly at me and i still wouldnt have got it.  Yup it was hard, and of course we all walked home on a bit of a downer.  Nothing like talking about whats getting you down to encourage male bonding :S  But anyway ive been really looking forward to this blog because i wanted to say about the emotion of Happiness.  You see i really dont get it at the moment i mean id love to \'be\' happy but i just seem to be as always Angst tastic!  I guess i just dont get it at the moment!  Plus it really doesnt help when you have this conversation between Me, Xan and David.  Xan \"David i think Ben said he wants to smack you.\"   David \"Well Smack me then!\" Me \"Dont fucking tempt me\"  Yeah not the most pleasant of conversations to have just before a exam... well anyway David, you aint worth my time.    Hehe well im home now and here is the stuff ive just put on to Chill me to bring out the Angst.    .Carlos Santana (feat Chad Kroegar) \"Why dont you and i\"  (I love this song) .Metallica \"Nothing Else Matters\" (There is something about this that just seems right to me\" .Red Hot Chilli Peppers \"Under The Bridge\"  Hehe when your down turn to the classics.  Well i will no doubt blog again later so au revoir for now ma cheries!  '),(24468,3099605,'24,May,2004','2004-05-24 00:00:00','   urlLink    Just a PS to my morning blog.  If God had a car, this would be it!  Behold The Lotus Elise!  Mistress If I Am To Be Your Angel Paint Me With Black Wings.  '),(24469,3099605,'24,May,2004','2004-05-24 00:00:00','   Hello!  How are you all this morning hope you all good and anyone who has got a intergrated or German exam today GOOD LUCK!    Lol well as you could see i was once again worshipping at the temple of ANGST yesterday but im feeling ok this morning, also reading the paper one article is on about how some of the oldest CD\'s from the 1970\'s are starting to corrode!  YAY you know what that means... some day in the future all Linkin park albums on the face of the earth will be destroyed by the CD plague!  YAY Up with the sickness!  Lol well anyway that definalty cheered me up this morning!  Also last night i started a Nu Muse book, which im very pleased about and hopefully combined with this i can keep it going :S I hope anyway.   But anyway \"All apologies\" to anyone who is deeply pissed off with me at the moment about the wingeyness of this blog i promise i dont mean it like that.  Hey just talk to me i dont bite!  Ohh yeah and if your sick of hearing me tell me shut the fuck up!  Hey i do understand!    Woo so anyway you know this blog is gonna go down hill all day so talk to you on it again later!    Ciao!  '),(24470,3099605,'24,May,2004','2004-05-24 00:00:00','   Heya everyone  Im feeling like some Angst has to be vented again so its gonan be small and sharp.  I guess what im annoyed about is that fact that being nice to some people aint enough.  You know your nice to someone for a long time and they basicly dont even care.  They just ignore you, make you feel small and weak in there presence.  Has this ever happened to you... i hope not, it just gets to me sometimes...  '),(24471,3099605,'23,May,2004','2004-05-23 00:00:00','   urlLink    Museings Of a Angst Ridden Youth, Now linked to Kingdom Hospital.  Beware Of The Walking Dead!  Bye Bye Baby Dont Be Long I Worry About You When Your Gone.  '),(24472,3099605,'23,May,2004','2004-05-23 00:00:00','   Heya  First day of the exams today and of course in my typical style im once again finding it difficult to focus on the exams... i take refuge in Angst.  Yeah i guess i do, when im Angsty i feel low but i can think then, truly think.  Its a mindset it definatly is and i guess Angst makes me feel alive.  Yes its a strange phenomen that ive got myself into of late and i sometimes need to remind myself that some people read this blog, well actually they are probably already starting to run the mile away from it but hey, i have to say what i gotta say.    Well anyway id kinda like to describe now what Angst is to me, its gonna be hard but i think writing it down somehow makes it easier, and if people can relate to this im truly touched and happy... well anyway here goes.  I guess Angst to me in its simplest form is Depressed rage.  I wouldnt say it was a emotion as such more than a product of sevral emotions coliding and you know when it hits because it feels like your own soul is devouring itself.  I guess what brings it on in me is certain feelings an thoughts such as Love, Hate, Grief, Longing and id guess the major league one is lack of respect.  Yeah id say those were pretty much the triggers of Angst.  But other people deal with these issues in other ways.  The one thign i would like to say is ANGST IS NOT A BAD THING!  I would say that in a way i love my Angst, it is not a problem that must be removed and people who have it dont need to be cheered up.  Id say just listened to every now and then, yeah id say thats all everyone needs really just to talk and someone to listen to them.  For nobody likes to be ignored.  But is Paranoia really the problem?  I guess thats another thing about Angst, it is a illusive beast that makes you ask questions of yourself, and everyone who has it wants someone to tell them the answer... but of course its not possible you see.  So Angst even though it feels like a black hole isnt all consumeing, its just got you... and when im in the depths of it i always remember, even black holes have got to come out somewere.  But hey thats the end of my little ballard :p  ON WITH THE SHOW!  Yeah today was my business exam and i was in the hall front row faceing my own reflection in the glass that surrounds it.  It was strange because by looking straight forward i could see the looks on everyones faces as they worked... and my own.  *i have very long hair now :S*  It was like being the silent invidulator of the exam, lol i didnt creep around the desks or wear squeeky shoes just to fuck everyone off i just saw the looks on everyones faces... it was very strange.  But anyway the exam i was quite plzed to find was rather easy and the weather outside patted me on the back as i went home lol.  So anyway all afternoon ive just been thinking, trying out once again the whole medative lark. I swear Balds onto something with this, it kinda magnifys your emotion to the point you can almost touch it in your mind... it really is quite brillaint i would recomend it.  So if you want to have a go just take these little steps!  .Find a incredably quiet place (Out in a wood or something, or if you are so blessed as to live in a quiet street!) .Sit in a comfortable position were you know you will not have to think about the way ur sitting so dont make it complicated .Close your eyes (i find this helps i dont know why) .Now just let yourself SLOWLY start to forget all your material worrys (i always imagine im walking away from a city) .And basicly just listen to what your mind wants to tell you, concentrate on emotion not want.  Hehe yes i know very spiritual but alwasy remember everyones got there way of doing things so be your own!  (thankyou very very much Bald for introducing me to this whole idea you truly are a revolutionist!)  Thankyou for listening. Here is the game some of you desire.  urlLink Trogdor THE BURNINATOR!   Have this aswell Just incase you aint seen it.  urlLink Balds Blog  '),(24473,3099605,'23,May,2004','2004-05-23 00:00:00','   Heya Mortals  First things first sorry if the last blog of yesterday depressed anyone or fucked anyone off, i just felt like i needed to vent some Angst.  Well anyway that was yesterday welcome to today my dears.  So anyone i hope youve had a lovely day, as lovely as PIE in fact!    Ive just been around today thinking about the inpending exams and trying to chill out, actually tried some kinda medative stuff a bit like Balds on at the mo, its pretty interesting because its nice to listen to what your mind wants to tell you.    But hey after my major caught on Blog mental anueyrsm of yesterday i feel much better now... kinda like being sick i guess and im optamistic of the week and then im on holiday in the best place on earth for a week!   Yes id like to talk about that place actually.  Some of you already know that me and my family go every year to a place in cornwall called Peranporth and basicly its my second home.  Its a little surfing town were basicly for a whole week i feel totaly free, its always weird because i normaly write a lot more when im down there and its just fun to be with the sea breeze in the hair and the smell of the clean air.  Ive always imagined that id take all my best friends there one day, and i will i promise.    Thats what i imagined today when i tried the old meditation lark... i was their in Peranporth and its really got me excited... it truly is my fave place on earth.  But hey anyway now im all calm again (it was late when i had this synapsis *listening to MUSE* so im glad not many of u heard me roar lol)  But now my head is clear i can do my list!  .Metallica (aint nothing else Matters) .Seeing My Sis and Helpfuly finding out from her its gonna cost about 50 squid to have my hair how i want it .My dreams of buying a trench coat  Yeah i know its a small list but hey i got a business exam tommorow and its bothering me.    Here is your first message of good luck from me during your exams, and remember, When all seems dark remember the candle of your mind and find the person with the match.  Au revoir here is your game.  urlLink 3D snake!  '),(24474,3099605,'22,May,2004','2004-05-22 00:00:00','   Heya  Yes i know this is a very unusual blogging time for me to be serious but i feel a wave of Angst has grabbed me and has caused a wave of verbakl vomiting!  Whats caused this you say?  I dont know really just a sudden need to say what i really gonna say and when ive finished it will go away but listen if you will for this needs to be said!  Well hey ima gonna get started now schools over and i have learnt a lot over that time that has always depressed me and made me Angst.  Such as the way that loud noisey people are always given attention and the quiet ones are always ignored.... i dont know why this is, am i the only one who is annoyed by people like that?   guess its another one of Ben Jeffrey\'s great problems, is he Paranoid or do people have a cause to treat him like he is.    It also kinda annoys me the way some people treat others, the way that if someone extends the hand of friendship the other bites it off or runs away... i dont know if thats ever happened to you but if it has you know it hurts like hell.   You do things for people, and they dont acknowledge you. Have you ever felt like you dont excist to some people and you just want to beg at there feet like they were gods just feel like your excistance is worthwhile?  It stings inside. Its like the warmth you feel from other people isnt in you, its ice.  Have you ever looked into someones eyss and they looked away in disscused or so it seems to you.  Have you ever said hello and been ignored... i do hope you havent.  I guess this is why they call it angst, i would say that Angst is deppresive anger... and it has just spewed from me onto this page.  Im sorry about that but i wanted to say it, it had to be cleared up.  Id just liek to say as a foot note that im very glad my friends are here and they are with me... they take it away when they are with me.  I would also like to hope that i do help people and make them feel loved, and i know people dont like to return love so easy and i accept this but sometimes its hard, to not be acknowledged... but its worth it.  Just be happy.  id like to finish with a little line i wrote myself, id like to think it says something.   \"I shall lead you out of the dark place when you call, for i will hear your cry, even if nobody else does.\"  Thankyou for reading... whoever you are you are brilliant, a individual candle that can drag even the darkest soul out of the gloom.  '),(24475,3099605,'21,May,2004','2004-05-21 00:00:00','   Hello Mortals.  Yes here it is again back to the usual method and style that is my Blog Woot!  So anyway after a very strange and long day that was yesterday ive been relaxing today on the net and being very bored!!!!!  Yes i know some elements of yesterday sucked but it enficsed how boring today was for me!  Lol well anyway i hope youve all had a good time today and your all happy.    But anyway on with the show... revision.  Yes this is something ive been aproaching with great care kinda doing it and not doing it at the same time.  But i honestly do believe that everyone who reads this blog will get what they want out of GCSE\'s and Life!  Yes i know that may sound strange but i honestly believe it!  So dont worry if your panicy just relax and it will be fine!  YAY plus good luck is a very powerful force!  Woot so anyway what u think about wishing?  yeah i know its weird innit, do wishes come true, or is it just silver down the well?  Well i dunno but id like to think that good luck comes to those who want it!    Yes this blog is very spritualy inclined today but hey spritual is cool!  Plus im really bored!  But anyway to conclkude the spiritualness i would say that Wishs do come true, as long as you believe and that Pennys in the well can no longer weigh you down!  Hehe well thats enough of that back to the more Mundane!  The first bit i really got is that Mine and Balds website now has a name \"World of Angst\" or WOA for short! and YAY its coming soon!    Hehe yes its back!  Stuff thats kept me sane over the past week!  .My friends!  Love you all! .My Blog and its increased Number of folk who read it... yay! .The fact We Left School this week! .Angst!  (The Thing You love To Hate!) .And Especialy you!  You are better than Chocolate!  Woot!  So anyway #Take me down to the paradise city were the grass is green and the girls are pretty!#  But Anyway I have Now got My Official College List of Shit I want and can get now im leaving West Park!  .Black Hair with Blue highlights! .A Black Leather Trench Coat (yeah i wish They cost About Â£150!) .A Nu Bag! (Lol typical my old one broke on the last day of school!)  But hey that dont matter its what you want that does!  Have fun today and smile! (you Know u can!)   urlLink Yay Curveball!  (yoink)  Au Revoir!  '),(24476,3099605,'21,May,2004','2004-05-21 00:00:00','   urlLink    I thought id follow up Balds Whole Lycan theme he seems to be on with lately with \"Ulfric The Pink Lycan Stole My Trench Coat!\"  Pass The Penny Royal Tea  '),(24477,3099605,'21,May,2004','2004-05-21 00:00:00','   urlLink    Some People Didnt Think That The First One wasnt a powerful image, so just to dispell any of that myth.  PENANCE!  I am evil I dont need to be told see!  '),(24478,3099605,'20,May,2004','2004-05-20 00:00:00','   urlLink    TWATS!  Yes You Did Ruin This For Us a Bit!! Vengance is a sweet drink indeed!  Laura Webber And David Malarky Will need To Learn To Handle There Drink A Bit!  Woot  '),(24479,3099605,'20,May,2004','2004-05-20 00:00:00','   urlLink    The Bench Bunch!  Yes Just on a slightly lighter note here is a piccy of the gang of us who always used to come to this bench every night after school.  Yes we have left are mark on this bench, sometimes literaly due to our love of writing our names!    So anyway Left to right here we are!  Mark Lakin, Adam Chapman, Simon Percival, Chris Shephard (AKA Wigsey) Ben Jeffrey (AKA ME!) Joe Tinley and leighton Baldwin (AKA Bald)  Cheer Up Mortals We Shall Meet again  '),(24480,3099605,'20,May,2004','2004-05-20 00:00:00','   Greetings Mortals.  Today was the last day at school ever and we decided we\'d party!  Yes we did have a good time but i feel bad that somepeople dont have common decency to not get so drunk they ruin other peoples times.  Id just like to say sorry for that if anyones time got fucked up today and hey mine did to so :( .  Hey but id like to thank the people who kinda tried to make sure we had a good time even though it was sometimes beyond shit, Sophie and Lyndsey, thankyou sooooo much!  Also thankyou everyone who turned up it was nice to know that some people do care! So thankyou to \"Big List\"    Jess, Sophie, Lyndsey, Bald, Joe T, Joe M, Adam, Wigsey, Josie, Amber, Simon, Giggles, Mark, Paul, Big al, Rach.   Two other people did come but im fucking pissed with them so hey David and Laura, u owe me an apology!!!!  Yeah i saw some shit today but some people just seeing them made it worthwhile.  Lol i finaly actually felt like i was wanted somewere which is quite weird... for me.  But hey we shall have to do it again but i dont think we should bring booze... its just not worth the hastle.  So anyway im kinda upset about the whole thing but hey nothings perfect, especialy if its instegated by me. But hey Adam told me i gotta stop kicking myself over this one... but hey i cant help it, i do it to myself because its easy.  But just so you dont think im a complete fucktard over this i cant believe u David and Laura, u acted like fucking preps today!  Im sorry but you just did you totaly regret what you did tommorow.  But hey the people who did keep sensible and happy made it a good day... thanku it means a lot.  BUT WE MUST DO IT AGAIN!  Because a lot of the people i have been with today i wish with all my heart id been mates with them since year seven, they are some of the coolest people i know and im gonna miss them if you go to Billbourgh.  So if anyone who was with me today is reading this *Big Hug* gonna miss you while your gone, but i will see you again... and i promise we will have a good time.  Hehe so i probably repeted myself loads but i dont really care because thoughts have to be expressed..i truly do love you all, your great people.  Au revoir Here is the game :P  urlLink Multiply and Destroy (Hmm Sounds like capatilism)   Hey and i would like to do a bit of Reverb from what ive seen on Balds post (he\'s better at this than me lol) and i would like to talk abit about leaving school.  It seems weird that we are leaving, i suppose Lyndsey put it best \"It\'s like something died\"  lol well i suppose it is a bit but id like to think that somethings have been born out of it and the friendships i made there will go on with me.  Lol like Mr Walker said \"And Bind your friends to your soul with Hoops of steel\"  Lol i know he said it, but it made sense to me... Hell well anyway i think ive had my reverb, talk to you again tommorow!  '),(24481,3099605,'20,May,2004','2004-05-20 00:00:00','   urlLink    Yay Gothic Super Hero! We love em! but what is it about them thats so great?  Is it there deeply cool outfits or is it the dark side to something great that we all secretly want?  I mean all my life thx to my sister ive grown up with them, dark heros.  i will never forget when she took me with my Dad to see him, the Bonedaddy Jack Skelington of Nightmare Before Christmas Fame and ive never looked back! (that day at the UCI changed my life)  So anyway i thought ud appreciate the picture of MDK one of the coolest Comic book/game chars ever!  Here are a few more of my faves! .The Crow (Eric Draven) .Spawn! .Daredevil .Sally! (NBC fame Again)  So anyway that was the first weekend Bonus Enjoy!  Hurrah For The Goth Superhero!  '),(24482,3099605,'19,May,2004','2004-05-19 00:00:00','   Heya Mortals!  Today was the last proper day we will ever have at Westpark!  YAY!!!!!!!!!!!  Yes we have been their done that adn bought the T-shirts over the years at the place but now lessons have finaly ceased!  So anyway how are you today, i hope your all happy and enjoying the lovely weather and the company, because if the company aint good tell them!  Lol well anyway today has been a great day full of funny little memorys which i shall never forget, such as the terrific double act of Mountney and Ben Anderson   Me:Why the hell you got Killroy up on your computer Rob? Mountney: He\'s Ben\'s best mate isnt he! Ben Anderson: HE\'S A NAZI!!!!  lol well i guess you had to be there, well anyway who could forget the \'Tinley Show!\'  (To the Tune Of he A-Team)  Yes Jess it truly is great!  So anyway i actually had a good day but as usual full of Angst... i know it sucks.  In a way as West Park ends im gonna lose some people that even though i didnt know them for long im gonna miss em quite a lot.  But anyway i hope we all keep in touch and one day i might even get a comment on my blog!....  Well anyway hope to see as many of my Good mates as poss at the Picnic tommorow we gonna have a ball!  Here is The Game My Dears  urlLink If God Played Flash Games he Would Play This, Carved Out Of Fucking Marble!   ...Au Revoir (Echo Thought, The Game is godlike but a bit easy)  '),(24483,3099605,'19,May,2004','2004-05-19 00:00:00','   urlLink    Hey just been Playing with the Photoblogging Malarky and id just like to send a picture of me!  Eeyore Is Me!  '),(24484,3099605,'19,May,2004','2004-05-19 00:00:00','   Heya Mortals!  How are you today? I hope your feeling good and you get all the pie you want!  So anyway lots of stuff happened to me today and my sunburn is starting to peel :(  It hurts.  So anyway today my Arch enemy is the legendary Twat Liam Carey because he grassed me up to the P.E. department when i was skiving... stupid bastard!  It was weird though because i just knew that today was the day i was gonna get caught, just kinda had a strange feeling.  Lol well anyway Liam\'s plan didnt go as planned because due to the powerful gift of lieing i got away scot free!  Yay well thats as it should be lol unless you think i deserve to go down... *I hate that Ben Jeffrey So Much*  ... Dam i just lost my train of thought there :S  Ohh yes and i discoverd people have great definitions of the word Angst and i do believe that its write and it does sound like Depressed Anger, which kinda sums me up really!  Well they do say all Great Art is based on Suffering and Longing...PIE!  Now but i also did discover a brillaint flash game today, a glorious remake of street fighter two!  OMG it reminds me so much of my old carefree days when all i did was play Street fighter 2 all day on my SNES lol when i wasnt playing Legend of Zelda Anyway!  So Here it is, relive my childhood!  urlLink Retro Fun, Jeffrey Family Style!   Ohh Yeah and BTW a nu era of Blogging starts as i link to the legendary Mate of Mine Bald and his Blog is finaly revealed to me!  We are also working on a website so nu\'s on thats soon!  urlLink Balds Blog!  '),(24485,3099605,'18,May,2004','2004-05-18 00:00:00','   Hello Mortals  Yup once again here chatting away to you over the internet and its been a interesting day today, with sevral sporadic teacher flare ups, such as the legendary Twat of 11R finaly after 5 years being sent to the head!  Also the last lesson of Sambi was amazing because we have finaly finished with his gimpyness!  YAY!  But hey it alwasy has to get worse as it moves along, drags i guess you could say but hey at least not many preps were active today, probably still trying to comprehend that they are soon going to have to live in the real world, were its not all about the clothes you were and the company you keep :P    So anyway how are you?  Hope your feeling good and have had a nice time today, but hey i fancy doing another list today of stuff so here we go!  STUFF THAT KEPT ME SANE TODAY .The Fact That Some People Are Never phased by anything (Angels Above Us One and All!) .Simon! (He subliminaly told me to download the man who sold the world by Nirvana \"That song owns!\") .The Last Ever Sambi Lesson! .Juice!  Lol so anyway sorry about that (i love the lists!)  In other quick new\'s btw we\'ve only got two proper days left!!! WOO!    Enjoy the Game  urlLink Kill Pop Ups, Now In Game Form!   Training To Kill Pop Ups!  '),(24486,3099605,'17,May,2004','2004-05-17 00:00:00','   Heya my dears  Me again, and today there have been to many cameras around in West Park today the flashes going off around like the fucking blitz does not do wonders for your eyes... and my skin burns.  Aww yes Pain how i have missed you when you left me alone!  Grr!  So anyway i hope that today you had a ace day and enjoyed the weather and the company of those company you keep!  Hehe well anyay its our LAST WEEK this week and tommorow i have my last ever graphics lesson!  YAY!  Omg that will be so great that i will be finaly rid of the UBER pillock Sambi.    So yay be happy and smile!  Anyway here is a list of stuff thats kept the tards from invading my brain today!  .Those Happy People who never get phased by shit! .The low quality Prep insults which have given me plenty of ammo to chuck back all day! .PIE!   Hey so anyway small list today so i had a pretty tard\'less day today witha  few exceptions but hey!  Here is a list of Insults about people with sunburn!  .Lobster Boy (The Ben Jeffrey Collection) .Crab Boy (A Prep Variation on Wit) .Frazzeled Fraggle (Relible Informed By Rose Mills a fraggle is some kind of cool monster thing so i wasnt really bothered by prep serial number 38918=5\'s attempt at wit) BTW thx Rose) .Red Boy (WTF????)  Hehe so on a quick note before the game, do you sometimes close your eyes for a second and see a picture, its only there for a fleeting second but you know it told you something... what did it say :P  Lol well anyway close your eyes open out your hands and enjoy my present to you!  G G G G G G G G GAME!!!  urlLink Play A Crazy Shark! (Check Out the Scary Music!)  '),(24487,3099605,'16,May,2004','2004-05-16 00:00:00','   Greetings Mortals!  Today i got really bad sunburn after a days fishing (about 12 hours in a boat in boiling heat)  So ima kinda in a lot of pain but hey its great if you wanna think, which i like doing a lot.  Lol so anyway, how was your day today?  I was just thinking about all kinds of crazy stuff, such as ancient curses, and chocolate!  Lol well anyway apart from my intense burning pains in my arms and face i actually feel pretty good today after i thougth a lot of deep scarey stuff through  YAY!  So here is a big *HUG* for you who is reading this and enjoy the game!   urlLink Tanks, and wee men lol  '),(24488,3099605,'15,May,2004','2004-05-15 00:00:00','   Hey!  Ben Hear once aagin just to say that ive evolved the blog a little hehe.  Yes i know the title is weird and that its not very homestarmountney like but i kinda wanted it this way, i feel more at home with it.  You can now post comments on Postys!  Which i thinks kinda cool, plus the blacks gotta be cool!  So Bubye :p  '),(24489,3099605,'14,May,2004','2004-05-14 00:00:00','   Greetings mortals  Yay finaly saw Van Helsing today, and the verdicts in, it wasnt bad the acting was crap and the plot your bog standard affair but it was enjoyable, unless your a prep i guess.  But hey i saw it and i didnt moan to much so it was ok.  But anyway how are you today?  It\'s a nice day today i hope you had lots of fun whatever your doing, even if you are like me and the sun casues you to... *shudder*  But hey smile its the weekend, i will do the frowning for you hehe!  But hey what do you care? Or even know? lol but hey if you read i dont care, u dont need a reason!  And hey im plzed you still visit whoever the hell you are.  So hey here is your pressie :P   urlLink Guarrented To Drive You Mad!   I love this game, it makes my brain hurt :P  '),(24490,3099605,'14,May,2004','2004-05-14 00:00:00','   Greetings!  So the last proper friday has been and gone and we are left a little happier by it!  But hey how are you today, i had a sucky day!  YAY however did do soem quality skiving in P.e. and imagined... stuff.    Im sorry my imaginations broke tonight :(  Here is the game!   urlLink Adictive Flicky Fun!  '),(24491,3099605,'13,May,2004','2004-05-13 00:00:00','   Greetings Mortals  Im back once again to talk and to witter and to what eevr the hell you want hehe.  So anyway i hope you are all having a good day and you are feeling happy, anyway today was a alright day for me but as usual i kinda feel a bit depressed.  :p ohh well this blog is for wittering so hey if u didnt liek that much why u reading?  Well hey anyway plz dont go they dont let me out much!  Hey so anywaynot long to go now before us at year 11 sign off for good from West Park!  Ima gonna miss the people but the place can fucking burn in my wake for all i care!  So anyway i hope all non billbourgh goers keep in touch with me or this blog might be in danger of being even more depressing!   Well ive noticed this blogs starting to get very formal so hey here is another list of stuff that kept me alive today, because i liek writing em!  .My friends (Thank fuck for you!) .Breed 77 (Karma... Thats a good song!) .My readers (You guys Rock and deserve to be vampires your that cool!) .This List! .My chocolate Pancakes! (yup i made em again recipe later in the blog) .People who could still smile (yup once again ur all so great!) .My Mind Keeping Itself together!  YAY!  Well anyway my lovely viewers!  here is the recipe for Pancakes!  2 Scotch Pancakes! Big Ol Dollop of choccy sauce!  Size=How depressed you are In the Microwave for 30 secs! ZAP! Eat and know your great!    Hehe well there i guess it is for all the world to see, my blog... yay!  Here open your hand and close your eyes, yes now you can open them!  GAME!  urlLink So cool it hurts!   P.S. *HUGS YOU ALL* (Lol give someone a hug today!)  '),(24492,3099605,'12,May,2004','2004-05-12 00:00:00','   Hey Mortals  Ben Jeffrey Here!  (Yeah i know Webmaster J\'s my alter Ego *bounces Manicly*  Yeah hi all, i bring you nu\'s today that MSN is being bitchy with me and that i feel really sick.  Grr but i dotn care still gonna blog :p hehe determination that!  So anyway i hope your day went well and wasnt to sucky!  I had a ok day but Mr Pain is asking for a smack yes u heard it hear first ogey bogey beard himself (it has bugs in it *shudder*  Is really incurring my wrath today with two hours of religion beign stuffed into my brain as i couldnt seem to be able to day dream today :( .  I hate that man he is a fucktard!  I also keep having visions of him being dragged away by the cops, or in typical ogey boogey fashion Jack coming into a R.S. Lesson and unravling his greasy black chord hair that holds his horrific skin together!  *Ohh plz Jack!*    Well wishful thinking i know but i do love my description of his hair!  he truly is a monster!  Anyway enougth about that onto other stuff!  Robo the arch villain of Chemistry wasnt here today!  *YAY!!!!!!!!!!!!!!!!!!!!!*  Here is a list of stuff that got me through today!  .The fact that some people could still smile at the end of it even though i couldnt (They are nothing short of hero\'s!) .My legendary melted chocolate sauce Pancakes!  (i will make em ya sometime lol!) .Cherry Flavoured Anti-acids  (yay second penny royal tea reference in 32 blogs!) .Whoever you are who\'s reading this you personaly rock and if anyone says diffrent i will totaly vamp them for u! .Visions of Mr Pains Demise!   So anyway that was a big enougth witter from me (i will hopefully be on MSN later tonight if im nto already on when whoever looks at this lol)  Here is a game just for you, enjoy it you deserve it :p  Au Revoir and Goodnight!  urlLink Fiddly Isometric View Fun!  '),(24493,3099605,'10,May,2004','2004-05-10 00:00:00','   Greetings Mortals!  Woah theyve chnaged my precious blog spot, its all coffe coloured and nu :(  Grr dam them, and i got a german test tomorow :(  Ohh well hell down to business, How are you today and hope you all are having a great time loL!  Anyway had a starneg day today for i think someone sedated the preps, they all seemed rather calm today and un confrontational, it was wellllllll boring!  However Mr sambi was on top mardy form today almost spanking Moutney for smiling and if looks could kill all the non preps in the class would be dead!  Ohh well here is the list of shit that got me through this day!  Friends!  (Love you guys) Enemys!   (Hate You Guys) My Dog Peter!  (He Owns) Blogs   (Blessed Be Their Name) Theatre of Tragedy!  (yay!) And whoever is reading this blog!  Here is a game just for you!   urlLink Like The Archery Game Only with Cannons!  '),(24494,3099605,'09,May,2004','2004-05-09 00:00:00','   Greetings :P Webmaster j Here!  Feeling a bit down today (Big aww everyone!)  went To liverpool, read some of my fave book got bored had to listen to old people yammering, suffering suicide 2 weeks to go im fucking bored!  So anyway How are you?  because hey i dont really care about me:P  So anyway hear is a list of Stuff that got me through this \"Fine\" day!  Anne Rice (YAY! Merrick Is the best book ever written!) Day Dreaming! SOAD (Bless Be Your Name!) Sweets and Mints! 2 weeks To GO! ... Penny Royal Tea  Good Night and Pleasant Gameing!  urlLink Simulate A Gailey Shroom Trip!  '),(24495,3099605,'08,May,2004','2004-05-08 00:00:00','   urlLink        urlLink Which Takeshi\'s Castle Character are you?  '),(24496,3099605,'07,May,2004','2004-05-07 00:00:00','   Greetings Mortals!  Hey im here again, carving a blog off the rock of raving!  I guess one day i will pluck up the courage in this blog o write somethign really shocking because nobody reads it hehe!  But seeing as i havent got it yet id rather just write a little bit about the background history behind Homestarmountney!  Now uc Homestarmountney was a dream of Mine and Webmaster Jt\'s respectably because i guess we\'ve always enjoyed entertaining the folk of West park!  Well and all of this formed into human form i guess in essance became Mountney.  Kindof like the virgin birth of Mine and webmaster Jt\'s Humour.    Mountney Quite Frankly is the guy!  But hey what about me and JT!  I guess our humours are very similar but we use Mountney as our public persona!  So there you go, Mountney is the radiation suit that protects u from the full on weirdness of our Humour!  This Is Officialy Webmaster J\'s fave Flash Game Ever!  urlLink Interpret This How You Will!  '),(24497,3099605,'07,May,2004','2004-05-07 00:00:00','   Greetings one and all!  Webmaster J here blogging after a solid day at school.  Yes theyve been me to me today!  A special mention to Mr Cunliffe today for making my life misery and id also liek to thank Spondon for being such a shit place to live!  So anyway thinking that im like total Derren Brown Guy today i was messing around seeing if i can influence people, it turns out i cant!  I guess you could say i am influenceless.... lol well anyho Van Helsings Out today YAY!  Now you all know me being tototal Vampire guy i weally weally wanna see this movie but somethings piss me off about Hugh Jackman and his kick arse alter ego that i thought i might annoy you all with!  Now Van Helsing, in the book Dracula, was a bit of a nutter and was also old... but it has come to my attention that Hugh Jackman isnt :( ohh well this ye olde bond esc edventure should still be a rip roaring flick :P  Pass Me The UCI nachoes!  P.S. This game has got Webmaster J\'s Personal Stamp Of Aproval, i think it kicks arse!  urlLink Make Like Robin Hood and Brain People With A Longbow!  '),(24498,3099605,'06,May,2004','2004-05-06 00:00:00','   Greetings mortals!  Today i once again am my normal self, u know that guy from school, the weirdo thinks he\'s a vampire YAY u got it!  Today i will mostly be talking about Bullshit as always but hey u read my blogs, u get my crap!  (well actually i dont believe anyone actually reads this bit if u do id liek to know!)  So hey how are you today?  Thats great Ben\'s Fine thx just little bit of this a little bit of that a lot of FUCKING HELL and a wee bit of sweeb!  So introductions over what have you got to talk about today, how was it FUCK PLZ IM BORED THEY ONLY LET ME OUT OF MY CAGE TWICE A DAY!  lol sorry about that i must have dropped a marble :P  So anyway Preps ehh who needs them, riddling popping monkeys not me in fact because we leave the majority of them behind soon!  Eat my dust! yay!  Now i was thinkign today if im gonna miss people and honestly concidering if i have any emotions at all apart from Rage and a mysterious inferioty :P  I actually do and i am gonna miss people but hopefully we shall be able to keep in touch!  What you think?  Ohh yeah i forgot nobody reads this bit! :P  Things That Kept me Alive todAY!  SOAD (Bless You Anti War Shenanigans!) Saying Feeb a Awful lot Cake!!!!!!!! and Iron Brew (YAY)  So anyway the moment you have probably all been w8ing for!   urlLink So Cute It Burns!  '),(24499,3099605,'04,May,2004','2004-05-04 00:00:00','   Greetings  Me again, now i know body reads this but thats probably a good thing, it gives me far to much free time!  So anyway today i guess i was just tired and i had a headache, it kinda helped me see things from outside everything today and i guess i learnt a few things :P.  Yes i know its a very melow kinda blog today but hey im writing if u dotn wanna read hop to the game as always.    Id kinda like to thank the following things today  Rasberry Flavoured Squash (Kept me alive) MUSE (They Still Rock) Err My Eyes (Arent they so underestimated) And finaly my Hair (for Hiding my face)  In conclusion TOT Rock i had a headache and i feel like a Juice GAME TIME!   urlLink Its Only Bloody Space Invaders!  '),(24500,3099605,'02,May,2004','2004-05-02 00:00:00','   Grettings one and all for today Webmaster J is once again here!  Today i will be mostly drinking Iron brew!  Woops i mean today ive been thinking about the possibilitys of parent tranquilizers!  I mean how amazingly useful they would be if say you were being hasseld by \'The Man\' you could just shoot one in and good night annoying parental figure!  Well anyway i think theyd be fucking awsome, and you could then lead on into brighter fields such as Sambi sedative, or Prep Poppers!  Anyway about the title, its very german isnt it hmm ohh well ima not gonna talk about it.  Soaring gently as a cloud Ohh how we all should be proud Some people think webmaster J is lame But to others here is the GAME!  urlLink D D D D Dragonforce!  '),(24501,3099605,'01,May,2004','2004-05-01 00:00:00','   Greetings Mortals who bother to read this bit!  This week has well frankly and with no words canceled out... Shit.  Yes it has been for me and many others probably the worst week on record.  We havent had it so bad as this week, come on admit it, you know its true!  All week the teachers at our not so beloved West Park have all been hitting that grumpy bong up in the staff room pretty hard!  I mean their was even a casulty on friday of bong overusage when Mr Sambi said he was \'Ill\' in lesson, so in anyone elses terms he went for a bit of a kip because he couldnt be arsed!  Now this would have been the best lesson ever if once again the preps didnt raise their ugly burberry capped heads against this strange phenomonen called fun!  Yes its true, they resent it when people who aint preps are having fun!  In other news Stewart Fletchers blog will be cancled for the next few week for inability to type with his right hand!  Yes my friends my week did suck balls for all them reasons and more but one question remains, how was yours!  Now without further a do  Presenting For Your Entertainment Needs Its The One The Only GAME!  urlLink Yay Another Megaman Game!  '),(24502,3099605,'30,June,2004','2004-06-30 00:00:00','   Greetings!  Phew, Billbourgh the introduction days are now over, and woah did i have a good time.  It was just so cool because everything so much more relaxed and less opressive than school and the lessons all thoguh will probably be way harder are more interesting and a lot more laid back.  Also its cool because i learnt i can do Tai-Chi woot!  Lol well todays been super cool with me doing History in the morning *Woot German Propaganda*  and then a massive break in between that just to to hang round and chat.  Then after picking up some good Karama by playing match maker *cough still waiting for that!*  Well i wont say who because i dont think they\'d apreciate the whole publishign malarky but if you wish to know just ask me.    Woot and as if that wasnt enough for a day *probably would have been if we were still at lame old West Park* i had a kick arse introduction to media, on which the course envolves making a two minute intro to a film and doing a music video!  HOOPLA.  Then to celebrate i drank some more coffe, (had about 4 cups today so far) and wrote some more in my precious \"road diary\"  woot im sure you will here plenty about that over the hols!  Well anyway it kinda sucks a bit now because i know ive got nine weeks to write this story and it suddenly becomes very real again.  Plus their is a article going for 16 year olds at the Derby Evening Telegraph and i so want to own it for myself and get paid for writeing so i gotta submit two articles ive made before July...  hmm i thinking of sending them a link to my blog... but i think the swearing and shit would put them off so i will jsut hammer out the two articles some time this month *i hope*  I mean that would totaly be my dream job, being paid to write... it would totaly kick arse.    Ohh well some of you ask \"What are your plans for tonight Ben? Well ima gonna drink a lot of coffe and stay up alllllll night!  Woo i got some more Edgar Allen Poe to read, plus more story to write.... plus plus plus i really wanna watch interview with the vampire again.  So basicly a very busy coffe and possibly Dark rum (Yum) fueld Ben tonight... err woot.  Plus some serious Music has to be listened to.  Ohh yeah plus if anyones interested in why i randmoly witter about a thing called \"Mothman\" lately, its basicly a storyt hat is repeated around the world as a large Moth like man beign a harbringer of death... all very interesting creepy paranormal shit!  Woot so if your even remotly interested  urlLink check this out   urlLink This is also pretty cool to on the subject   Lol well anyway thats it cya on MSN probably. Ben.  '),(24503,3099605,'29,June,2004','2004-06-29 00:00:00','   Hiya All  Few First afternoon done at Bill and i actually had a good time!  (Shock, Horror, Hoopla, Mittens!)  Yeah i only did to subjects today though with it only being a afternoon, Biology and Communication Studies.  They were both very cool accept i think Biology will be hard!  So if you catch me in September muttering about amalyse structures and sayin g such legendary quotes as \"You can drink two pints of your own blood before you get sick\"  you will know Biology is takeing its effect.  Hehe Communication studies was really cool aswell, just very relaxed and you know... sooo diffrent from school.  Woo i wrote some shit in my road diary but i dont think ima gonna refer to that in my blog... yet.  Well tommorow it is Media and History, but im also having to look at Psychology to fight off boredom.  I also got a bit of a fresher fair to sign up for activities and so on.  But anyway it did kick arse today but with a full day tommorow it could knacker me quite suffiecntly.  So if tommorow night on MSN im all \"Many combobulations elizegerth\"  Or \"Even though your a dude ima gonna adress u like you was a lady.\"  You will know im just knackered.  So hey, how was your day today, i hope you got cookies and all the other happy stuff that goes with not having to go to school for ten weeks!  Woot still btw.  So anyway quick bit of Observation i got about Bill.   Its Uber relaxed and a lot less opressive than i always thought West Park was. Its great not having to wear a uniform so that aspect seems a lot less opressive. Its great that you have a lot more freedom to say what you like even when it is adressing a dude as Maam.  Woot Cya Later  Ben.  '),(24504,3099605,'28,June,2004','2004-06-28 00:00:00','   Heya  Yup todays the big day and in about fifteen minutes im off to college for my induction thingy.  Well its only the afternoon im doing a full day tommorow aswell.  Err woot.  But i can imagine its impression time today so i gotta practice the big smiles and ignore the whole \"What you do for fun question.\"  Lol because if thats the case, my reading, writing, music listening are gonna make me sound uber boring stright away, i like to stretch the illusion lol.  Woo so im hopeing there arent any prep star runners in my form, or people who hate people who like rock music just because its something to do or it will be all out war from the start... and i need to normaly prepare for that.  Ohh well to anyone who this concerns about Bill, have a great day, meet lots of nice people and enjoy urself!  Also to the people who it doesnt concern have a great day and eat a cookie!  Cya  Ben x  '),(24505,3099605,'28,June,2004','2004-06-28 00:00:00','   Hi All  Not been feeling very well today, i think im comeing down with a cold or something, which sucks a lot because im trying to write my story and i cant even think about real life never mind my fiction.  Doesnt help that someones car\'s all dodged up in the street outside and its making the most annoying sqeaking noise ever.  Ohh well i got my Bill afternoon tommorow and then a full day after that... woo once again i go into a situation were i will no doubt be instantly judged and branded a lunatic.  Woot.  So Ben\'s Adventures in 3RES will begin, along with the whole Bill experince.  Ohh well let the judgement begin.  Sorry this blogs a bit crap today but im finding it really hard to concentrare because i feel like shit.  ohh well.  Bubye.  P.S. As A Apology for the crapness enjoy this  urlLink Game   ~Uber Edit~  Forcing myself to write some more because i dunno just want to i guess.  This week has only just began, but i know its gonna attempt to be lame, its a week thats throwing up a whole barrel full of nu stuff to deal with.  Its almost like ive just been handed a whole bunch of sheets and its \"DO THEM BEFORE FRIDAY OR KARAMA WILL FEAST ON YOUR MISERY\"  err cheers for that.  So basicly im feeling a bit run down you know, with this story moving like a snail with the bends, and the whole Bill thing staring directly in my face like a loaded cannon of inevitabilty.  It kinda sucks.  Plus the gloomy haunting GCSE exam results which u know are gonna eventually come, and they better be pretty or i will smash something pretty :P  Plus it really doesnt help that ive got a cold and i will probably have to meet all these nu folk tommorow with a dripping nose and streaming eyes.  Hmm yes make it easier them to brand me with the twat brush.  Woot.  Ohh well im done bubye.  '),(24506,3099605,'27,June,2004','2004-06-27 00:00:00','   Greetings!  Hey just came on here to talk about the final headlining act of glastonberry 2004!  Yes it was MUSE and yes they were fucking imense!  Matt Bellamy hammering it out time after time, while still having time  (and Energy) to do a massive random solo while straddling utop his amp to finish the night while Bassist Chris Wolstenholme dived into the crowd to preform a whee little bass solo.  Yes it truly did own, they gave the crowd what they wanted and it was MUSE.  Now i know some of you out their dispise this trio, but truly they did own Glastonberry tonight, with a incredible set from everything from the soft melodys of black out, to a truly stunning version of Plug In Baby.  Im truly happy about that, the only draw back is that they didnt play Muscle Museum.  *sigh* Well you cant have it all.  However Matt\'s alternative solo\'s on Bliss truly made them worth listening to, and you could almost forget one of the best songs ever written.  I think they all did incredibly well especialy when they did Blackout... god that song owns.  Special mention to Dominic Howard, he beat his little heart out on that drum kit all night only for Matt to destroy it at the end lol.  But pretty much why this blog is now in excistance is because, i you know love MUSE, and tonight has only cemented my beliefe in how much they own, so please if you dont like em, give em another chance with these songs.  .Plug In Baby .Muscle Museum .Dead Star .Bliss .Blackout .Stockholm Syndrome  But Lets Be Honest, It All Owns.  You will NOT be dissapointed. Plus if you wanna take it further visit there website  urlLink Here   Lol listen to MUSE, you may be depressed but you will know uve listenend to some good music getting that way.  '),(24507,3099605,'26,June,2004','2004-06-26 00:00:00','   Heya all.  Well once again just had another massive long drive to see my brother in Liverpool, its good seeing him but the journey dont half suck.  This one was however a little less sucky due to me still thinking about the amazing poems in the book i bought yesterday.  Here is a example of why it kicks serious arse.  The boy with nails for eyes Put up his aluminium tree It Looked Pretty Strange  Because He Couldnt Really See.  I swear if you like Tim Burton, and dont mind poetry you will love \"The Melancholy Death Of Oyster Boy And Other Storys.\"  Lol well anyway its totaly inspired me, and i hope that i will one day be able to put that kind of flare into something so simple... but nobody does it quite like Tim.  He is the master.  Lol well in other New\'s remember to watch MUSE!  The glastonberry highlights, are on after the fast show on BBC two, which just happens to be on after Kingdom Hosptial tonight so double woot!  But i guess on with the show.  Today ive been under a lot of pressure to do lots and lots of things by family, and it really really sucked.    \"What you gonna do with these Ten weeks Ben?\"   \"Write a book\" \"Ohh just like your sister then!\"  Not what i wanted to hear!  Then after about 15 minutes of my family wittering about how clever and amazing my sister is comes the next question.   \"What do you want to be in your life Ben?\" \"I would really love to be a Author, but medicine and Biology interests me.\" \"OHH A SURGEON!  What a respectable job!\" *Imagines himself repeatldly banging his head on a table*  Jesus wept, before they went all corleone family style gangster respectable job not the *family* business style bullshit on me (yes i sometiems swear that my parents think they are in a gangster movie, but with less people called Jimbo \'The Knee\' Capone.)  I had to change the subject to what the weather or some random shit.  But of course they turn it to \"Ohh you must get out on trips and stuff with your mates Ben!\"  ...Yes good plan but we are NOT the secret seven or wahtever the hell!  *sigh* shoot me please.  Au revoir  Ben  '),(24508,3099605,'26,June,2004','2004-06-26 00:00:00','   Greetings.  Heya all, got some angst at last, i wondered were it had gone really.  Lol well here it is, for all to see and read about.  I just got a lot of annoying shit on my mind at the moment and i need to rid myself of some of it.  I guess ive got a lot of problems with this whole task i got with the book, i really wanna get it done, to achieve somethign during this ten week and do something i really enjoy doing but the problem is my minds not letting me.  Yup writers block, it just gets you so hard because not only does it stop you writing so you cant enjoy yourself, it also reminds you the reasons as to why youve got it.  Just basicly gives you all your angst, reheated, rehashed and right back from were you last vomited it up.  You know all the reasons for this worry and pain may seem insignificant to some people but they are really there.  Like the little things people say to you, sometimes they mean nothing, and other times they really hurt.  Just makes me wonder about the point in having feelings sometimes, you know when something you feel about strongly is directed at something that will never care, be concerned or even bother to validate the excistance of a feeling, then what is the point in having that feeling in the first place?  It Hurts to love, so Hurt loves itself.  I sometimes feel if god made us... he made us wrong.  I mean the human mind is obsessed by massive gestures and feelings, the simple things on a base level never seem enough.  Basicly like Breed 77 Say \"If i paint you a picture and the paint is yet to dry, could you live with jsut a whisper?\"    The Truth is, most people cant.  Another thing that annoys the ever loving hell out of me is the way some people believe they are better than you.  Now normaly i would just laugh this off, but you do always wonder why.   Lets face it they are quite obviously fuckers who use all there fancy terms and shit to tell you how sane they are compared to you and that your experssion, is just some \'Teenage\' thign that some people go through.  Who gives them the right to declare that on you like so much garbage from a great height.  I guess in the end everyone is self obsessed, its biologicaly in our genes to be, but how do people take it so far as to shout at you in the streets and hang round in packs, imtimadating people, and basicly crushing.  Why do people feel the need to validate their excistance by hurting others?  I believe that if people decided to understand rather than judge a bit more, society would be better, for example if someone paints a picture in black and red, does that make them insane... no.  If someone writes a story that isnt about teenage crushs, and doesnt have a pink cover does it make it a bad book... no.  If someone weres a diffrent shirt to someone else, does that make them a bad person... No.  But its all well and good preaching a world of no lables and be neutral.  But there is a very large problem with that.  If you are neutral, and let peopel spite you and not spite back, you have aloud someone permission to punish someone else.  Kinda like passing on a worry.  Now i cant do that.  I guess you could say that i am flawed on the basic level that i will call you a chav, or a prep or wahtever, but only if you opress me.  Nobody should be aloud to hurt someone else.  But i guess thats why i feel this writers block, because of my flaws, and my problems, but i work on a very basic principle of balance, and today someone pushed the scales the wrong way for me.  Thx for listening.  Ben.   ---------------- \"People can do great things if others believe in what the creator themselfs cannot believe in\"  I will always believe that you have the power to do what you want, even if you cant.  '),(24509,3099605,'25,June,2004','2004-06-25 00:00:00','   Hiya all  Hope your all having a good day today and that you didnt go to town today *shudders* i dont think ive ever seen so many Chav\'s in one place, it was like some kinda Chav fest.  *sigh* lol well anyway i suppose it was quite cool because i managed to get something i actually wanted for a change yay!  Its a really cool book by Tim Burton and its piccy is here.     Yes i know very nice title isnt it, well im looking forward to reading that later, (YAY Angsty Poetry!)  lol well in other news ive managed to get my story started!  Double yay!    Lol well i did have plans for this blog to be all big and angsty today, but i think getting what i wanted in town cheered me up a lot i guess so instead im gonna have a ode to carusoses cafe!  Woot i dunno if youve ever been but i think this is the coolest coffe shop on earth, i mean they sell cookies and juice!  Aswell as the greatest milkshakes in the world!  They have bits of whatever the flavour is in it!  I mean milkshakes with bits of cherry in, how cool is that!     Lol well ive been going for about two years now and its never sold out either, so it truly has to be the goddess of coffe shops!  Also another world is selling a film cell from one of the greates films ever made \"The Crow\"  YAY!  They are also selling a NBC one, double yay!  Hehe well apart from me really wanting the crow one, this may sound lame to most people but its a piece of history on a plaque!    Lol well sorry if this blogs a little short, there may be more later.  Au revoir.  '),(24510,3099605,'25,June,2004','2004-06-25 00:00:00','   Hi everyone  Been a weird day today to say the very least about it, just seems that its just really been a chain of sevral events interlocked.  Its almost as if ive lived about five days inside of one.  Start off with this morning, the usual boredom of Trisha, remembering how long i stayed up till last night (3:00am) looking in my diary just to see how i felt last night then being awoken proply by Trisha\'s speech about her next show.    Trisha: Have you been caught out? Me: Ehh Ive Just woken up for crying out loud!  Lol my mum set me a task today, using her precious pink sticky post-it notes, when i see one in the morning, i know its something i gotta do, well today it was \"Polish the dining room chairs and table.\"  ... Fucksocks, that polish makes me sneeze.  Then writing down the first part of my story seemed weird while the ever helpful Adam in the background made fun of me, lol very strange experince.  *sigh*  Well anyway that was the end of day 1 so to speak.  Day 2: Me and Bald hammered the living shit out of each other on a mindless computer game, very theroputic, then i took Peter, out and did the usual humming and haring of thinking what im gonna do with the rest of my day.  Day 3:  Went with Mark to school to fetch back his golf trolley... yes i know school *shudder* but there was something so very diffrent about it, it didnt seem like we belonged there anymore, i mean as we walked past Art block which had been pretty much our home away from home for the past 2 years it looked like new.  Not liek the place were we used to chuck food up onto the lights, watch or play frissbe and genraly wile away the lunch hour.  W block was desterted period 5, no year 11\'s you see.  I mean as i ran past shouting \"SAMBI!\"  all i got was mr Sutton telling me he was on a course... fucksocks.  Lol well and walking home with a very dodgy off balance golf trolley and a boy with a cane didnt half get some stupid looks from the local old people.  Day 4:  Went up the alotment with my dad to bring a large bucket home to be the fishs tempariry home while we re do the pond, there is something very strange about filling a very large blue bucket with water for the better part of half an hour with watering cans.  Kinda monotanous, but in a strange sense rewarding.  Day:5 Writing this blog.  It has been a weird day.   Cya Ben.  P.S. Found A Great Website Showing What Happens When Artists Have Breakdowns and it can be found  urlLink Here  '),(24511,3099605,'24,June,2004','2004-06-24 00:00:00','   Hello All  Well The Ten Weeks of freedom for me began proply today, and i basicly feel overrun.  You know i have all these possibilitys running through my head for these ten weeks, and i want to take advantage of them all but it just seems like its all to much at the moment.  I guess its like i can only take on one thing proply at once at a time, but in ten weeks how do i know i can take on what i want to take on.  I really want to take this time that i have over the next ten weeks to write something, i know that much, but its really hard writing as im sure some of you know, you are just constantly phased by writers block and there are days were everything you seem to write means nothing.  You see, writing is so strange because its not like a question in a test were you can just keep writing and writing when you know a answer,  books ask no questions.  I alos want to have a good time during this ten weeks, you know go out, get pissed see some movies, have a laugh and stuff, but the weather this time challenges me to a duel, and the way that my mind makes me shy towards asking people to do things for me and organizing people to have a good time, it really sucks.  It just seems that every move i want to make will not work at the moment, like when your playing draughts and u know any move you makes gonna get you jumped.  I dunno maybe im just screwed up but i just feel at this time that any move im gonna make is gonna be thwarted.  But hopefully when i finaly charge in and make some moves, its gonna blow the board in two.  But i know that i am gonna attempt to write a story in the ten weeks we got, a proper full scale one, ive had enough attempts now, enough pokes at the jelly fish style dead story on the beach, but now its do or die time at 9:10 tonight it officialy begins.  The first move will be made.  Cya Later Ben  '),(24512,3099605,'23,June,2004','2004-06-23 00:00:00','   Greetings!  Woo Today was my last ever exam, and it went really well woot!  I cant believe it, after 5 years of being annoying, pissed off, upset and hell at times even dam right demented its finaly over.  Just think ten weeks of liberation... getting pissed in the park, hanging round town just because you can, drinking milkshakes and eating icecream in carusoes.  Woah.  Like Mark said to me today, \"This is probably the biggest break off from any work we will ever have.\"  Woah, he\'s right, how very insightful of him.  Hehe well anyway im well chuffed that we are finaly done, squiddled squoddled rocked and rolled all the way out of there woot!    However even though a lot of shit was given at West Park im glad i went, because if i didnt i wouldnt the cast of characters which im proud to call my friends, you know who you all are, and plently of anecdotes were gained just from their exploits.  Hehe i mean who can forget the time we played Ultimate Frissbe in the driving rain, the atmosphere closing in around us like we were in a little frissbe playing world of our own.  Lol or such Great times as Maths with Mr Kavai when we used to play hell every lesson, Chemistry wouldnt have been the same without Tinley Tourettes, The way Me, Adam, Mark and Pat used to ruin our humanity teachers life in year 8 practicly.  Some stuff your just never gonna forget, and i have made some great friends through west park, and i know even though we aint gonna go to the same schools and shit some of us, we will stll see each other, because times like that have a strange habbit of linking people together.  I mean we have all in our own indivual way had a ball, even if we did happen to have this great ball in  one of the suckiest settings on earth, all the way from leaky park road on the first day when we past the big crocodile to Mrs Rollinsons room all the way to last ever chemistry lesson down in the libary, were Jess with a camera made a very embarssing show out of us all.  Lol and leavers assembly all the way through (especialy when kesia sang) me screaming in my head *SHUT UP I WANNA GO DOWN GRAVEL PIT AND GET PISSED!*  Lol we will never forget that day i dont think anyone who was there.  So many memorys, some to embarissing for some people to ever want to recal :P  But anyway, just think what a long way weve came, and we did it, the highs and the lows, the coursework, detentions after school Sambi and Paine.  Lol in the end it was all worth it, because now we come out knowing who we are as people... and who we can trust.  Buildings Burn, People Die, But Friendship lasts forever.  Cya Later Luv Ben  '),(24513,3099605,'22,June,2004','2004-06-22 00:00:00','   Hello All!  Well today was spent by me mostly dodging from going to have my hiar cut, i know i know but i really just cant be bothered and hairdressers scare me.  Ohh well so in thisvein of avoidance, Myself, Adam, Mark and Bald set off to hermits cave once again!  Well for those who dont know Hermits cave a is a creepy old cave on dale Abbey that once contained a hermit, not very exciting some thing but when your four lads with nowt better to do, have legs... will adventure.  So as usual our trip envolved the usual mix of rude jokes, complaining that out legs hurt, rude songs \"So i pretended that she wasnt a prostitute\"  and also songs just because our legs hurt and we are thirsty.  Woot so that envolved such classics as, \"If your walking on a road and you know it Punch Bens Arm\"  and my particulary Bad version of \"Tiny dancer\"  its hard to create the bus scene from almost famous when you dont have a bus.  Ohh well anyway thanks to are ability to know were we are going, (we dotn in life but we know how to get to hermits cave)  a cracking time was had by all.  Woot.   Its good being out, kinda takes your mind off things, and after a particulary angsty morning i really was glad of it.  (I missed my bus into town to get a haircut, i woke up feeling depressed and the usual nine yards of angst.)  So it was a good day but as usual because of it im writing this blog while constantly trying to get the hair out of my eyes.  So as my usual sign out, hope everybody who is doing a history exam tommorow aces it, and have a great day... smile and enjoy this gift that was given to you for who knows this could be the only time we truly get to apreciate the quiet things that nobody ever knows :P   \"Hold Me Closer Tiny Dancer, Count The Headlights On The Highway, Lay me down in sheets of linen, youve had a busy day today!\"  Or alternativly one of mine.  \"Made in great germany?\"  '),(24514,3099605,'21,June,2004','2004-06-21 00:00:00','   urlLink    Yay 100 Blogs!  Yay so a bit of a landmark in my blogitude so i thought id make a bit of a occasion with it with the picture above, and also later on in the post a poem i wrote.  Ohh well before that i think i gotta talk about my reasons for starting this blog in the first place.  Well i guess the original purpose for \'MOAARY\' was to have a voice to my own personal feelings that anyone can look at without feeling obliged to listen to it if they didnt want to, which they might have done if id have said this stuff outloud to people.  I guess also it was used to show the diffrent sides of my personality that nobody see\'s.  This is because i am incredibly shy, and find it hard to express things proply sometimes when im around people who i think their opinion matters.    So this blog became my voice for all my Angst, my feelings, thoughts, qoutes pictures and random stupidity, i guess this is my home on the internet and people can come in anytime they want :p  I also hope that through my blog people can relate to troubles im sure we all have, and id like to think that my blog still strives to help people realize that everybodys individual expression is a brilliant thing, and i hope that every now and then my blog can say something that touchs its readers on a personal level.  Just as writing this blog has helped me out a lot.  Id like to also do a thankyou to everyone who has ever taken the time to read my blog, you have truly helped me out just by listening, because its always comforting to know that even if you feel nobody cares, at least somebody reads about it.  Thankyou very much.  I guess that leads nicely into my poem.     Alpha  I want to keep you all safe I want to show you my worth Give me a pair of wings And be surroundedâ€¦embraced Give me a sword Be forever protected Make me the Alpha Make me the guardian Angel Believe in Me.   I guess that poem sums up Angst, a need to be belived in, everyone gives each other their sword and wings, but only you know how to use yours.   So any comments about the Poem will be well recived even if to tell me its shite lol.  I dont really mind.  Its the first poem ive ever shared, so hey i expect it to be crap.  Anyway that doesnt matter, its the thought that counts lol.  And Just To Finish off, and prove that this is still a 16 year old boys post, here is somewhat of a soundtrack to my life at the moment :P  Ben Jeffreys Angst Album Volume 1! .Metallica: Unforgiven .Brand New: Guernica .John Lennon: Working Class Hero .Dave Grohl: Tiny Dancer .Cradle Of Filth: Her Ghost In The Fog .Ivy:Worry About You .Nirvana: The Man Who Sold The World .MUSE: Muscle Museum .Slipknot: Before I Forget  Lol that just about sums it up. Heres to the next 100 My friends. Thankyou for listening Au revoir!    100 Posts!  '),(24515,3099605,'20,June,2004','2004-06-20 00:00:00','   Hello All!  Well what can i say, today has been pretty dam uneventful, i made some stupid photoshop piccys,  which can be found  urlLink here  and also  urlLink here   Also carved out some preperation for tommorows 100th blog, woot!  So basicly thats a poem and a piccy, u lucky people you lol.  Well anyway apart from that todays been pretty uneventful, apart from having to fork out for some whiskey and a card for my Dad.   Thus once and for all answering the age old question which is better Daddy or chips?  Lol well Lithiums just come on Kerrang radio, because they\'ve never heard penny royal tea... ohh well not complaining to much because u know... no point lol.  Ohh yeah watched Spinal tap yesterday, it was ok but absoulutly nothing compared to almost famous, which is kinda on a similar vein.  *hums Tiny dancer*  Ohh yeah and if you can get the version of Dave Grohl singing Tiny Dancer your life will be enriched.  Lol it really is that good.    Sigh ohh well i forgot to add one more career ono yesterdays hoard, id say it was one that nobody like us would pass up the chance to and that would be to be a rock star, hehe all i need is some talent lmao.  Woot yeah ive been hideing from that whole careers spear today, im sure some people would say this is due to my \'Paranoia\' or some shit, *hums some Sabbath*  Woo, so much for getting in touch with something that people are always trying to kill.    I dunno just seem a bit down tonight, i ant believe this is my 99th blog... so weird.  Hehe well anyway i know lots of lucky people tommorow have their last exam, i mean how cool is that!  So lots of luck to all doing the graphics exam and those who finish with it tommorow poke Sambi in the eye for me :P  Well yeah i guess with kerrang radio becoming more sucky as the night goes on i will end up retreating to my album collection and no doubt listening to the influence for this blog title *sighs thinking how great In Utero is*    Lol well after a very musical, material and mellow blog ima gonna finish with a quoute from this is spinal tap.  Enjoy yourself a lot, all the time.    Or alternativly one of mine.    \"Im Blissfully aware of all the pain i give myself.\"  Au revoir, and goodnight!  '),(24516,3099605,'19,June,2004','2004-06-19 00:00:00','   Hello  Been thinking a lot about career lately, and what kind of person im gonna be in college what im gonna do and how im going to affect the world on a much larger scale.  You see because in secondry school it was very easy to affect your enviroment, you know you just had to say something stupid or bring in a whole new idea to who you hung round with and it had a affect.  But i want to make a impression again now, and i know its gonna be harder, i just basicly want to believe ive made a diffrence no matter what miniscule way because i feel if you cant have a positive affect on someones life, why have a life yourself?  It seems that everyone is trying to force this home to me at the moment,  i mean for a starters i enjoy operation programs and shit, anything about illness or Biology i love learning and talking about, my parents pick up on this the other day ask me to concider being in a medical profesion.  Its strange because i always felt that i wanted to do computer graphics like you see in walking with dinosaurs and stuff... but  Over the past to years i got more and more into writing, i mean ive always enjoyed it but two years ago i was told that i can do it.  Interesting really i dont think untill that time ever in my life had i been told i \'Could\' do something, and over those to years, the world of computer animation has got less and less interesting to me.  It just seems that sitting in front of a Pc all day dragging and pulling at wire frames like its a extention of your mind finaly grated on me.  I just dont think its art anymore.  I mean now that books and \'Art\' movies got into my life in a major way everything is now makeing more sense to me on a more basic level.  I guess cutting people open and saveing someones life links to me a lot more now than dragging stuff that really in the cold light of reality is fake across a screen.  I guess the more artsy side of me apeals to my strange habbits about Charity.  I mean if u have ever been to town with me you will know sometimes i just blatanly ignore charitys and shit, but things that really matter, i have to give.   Its strange really.  But i dont think its for me, i just couldnt think i could learn like that, careers are just confusing.  Theres just so much that gets in my way when i think about \'what career suits me\'  I guess my dream job would to be a author, but its the worry of if its possible, and if it keeps you alive.  Just basic things.  But really careers are not important.  Its people that are the real key.  Ive seen a lot of press and talk about charitys lately, now this is something i want to get envolved in.    For instance, i have seen a lot of medical stuff in my time, around here, hospitals and stuff i just dont really wanna talk about but it helped me reallys some very basic things are almost our duty to do.  Such as giving blood.  This i believe is one of the most influential things i will ever end up doing on this earth.  It is so important because of its basisity, i mean for instance 1000\'s of people everyday need blood in accidents and surgery and so on, now if peopel stoped giving blood, people who need basic operations who would otherwise survive would die.  All because some people wouldnt dare have half an hour on their back giving blood.  It doesnt make any sense to me, and i know for one the second i can i will do this.  I guess what i could say is that my thoughts about future life are suffering from a bit of a \'renaissance\' its very strange, and im still confused but i know no matter what my carrer is i can affect this world in as simple ways as giving blood, and that makes me feel good.   Ohh well in other news, ive been looking at this website about urban legends about creatures around the world, its really some quite interesting shit, worth a laugh if u into all the paranormal and that like i am, (yes i know i used to watch the X-files a lot so shoot me)   urlLink American Monsters   So enjoy that, and to conclude ive gathered a list of Jobs id like to be when i finaly go out into it, (Yup some of them are jokes but id love it)  .Surgeon (err i dunno) .Film Director .Author .Freelance Cryptoligist (lol Monster Hunter) .Paranormal Investigator (Everybody *hum hum hum hum he hum*)  Lol thanks for listening to Ben Jeffreys Bad Asse Angst Song.  Au revoir.  '),(24517,3099605,'18,June,2004','2004-06-18 00:00:00','   Hello All  Nothing like a morning blog to get you going, and as i despretly try to force my brain into doing some creative photoshop picture takeing the piss out of something or other ive been dodging round the internet looking at random stuff, which is a bit boring.  However while i was despretly trying to download music by panic cell (which later failed)  i did find this  urlLink Its a heartwarming story of a guy who made a shell for his computer entirley out of pumpkins.   Well not a lot has gone on at the moment, and its well boring, probably gonna go on MSN as soon as i can figure out what picture ima gonna do :(    Ohh well at least i managed to write up one of my poems on word, now all i need to do is to blackmail Joe Robson into *cough* lending me one of his angel drawings and i got gold!  Woot well err seeing as ive mentioned him here is his website    urlLink Newneojoe   Well talk to you later when i got something to say.  '),(24518,3099605,'18,June,2004','2004-06-18 00:00:00','   Greetings!  Yay late blog tonight, yeah lol well the day did get slightly worse, mainly due to the fact hardly anyone was on line which sucked a bit.  Ohh well.  Well Joe T did come on and we had our usual verbal sparing match, he concluded that im Paranoid and Obsessive... hmm not sure about the obsessive bit myself, felt it was a bit harsh, but i do understand the Paranoid bit, i really do.  He quoted this to me...paranoia is a irrational fear or irritation by something that u feel is connected deeper than it is.  Yeah that sums it up i guess, im so crazy... and hmm highlighting it didnt do much for my mood really.  Grr my mum shouldnt have told me i could finish up the rum,its gonna be so gone by the time this blog is done.  Yeah well anway it leads me onto another point... people who dont get Angsty.  Lol just like hmm really, his quote was \"i dont care about hidden meanings and shit\"  Lol that about sums it up really.  Yeah i also told him that hidden meanings are really just a person begging to talk about something close to them but they cant share it unless its found, yup thats angst.  They should be more people who should understand this fact, i mean if i wasnt so willing to take the piss out of myself all the time id be well pissed off with what he just said to me.  Ehh i dunno maybe its just the drink talking but i doubt it, just a shame not more people were on to talk about it, beign told ur mental before bedtime dont help very much :S  Well if i carry on like this ima gonna start repeating myself so im just gonna simply sum it by saying.  Angst isnt a Mental Disorder, its just a way with dealing with things.  I guess im just searching for The Quiet Things That Nobody Ever Knows.  Goodnight.  '),(24519,3099605,'17,June,2004','2004-06-17 00:00:00','   urlLink    Just Hosting My latest b3ta entry, i will post the link once i got it.  Angstriddenyouth@B3TA   Edit: Heres The link  urlLink Yay First Ever Compo entry  '),(24520,3099605,'17,June,2004','2004-06-17 00:00:00','   Greetings The Folk Of My Blog!  I got up an hour ago and its strange because i feel ok, almost i dunno happy, and i owe that to some of the people i talked to last night and the things they said and did, thanx for that :P  Well anyway on with today!  Yeah well at the moment i got Kerrang radio on which can be found  urlLink Here!   Woot if your lucky they play a good track, lol but at the moment i think they like playing the Rasmus just to piss me off.  Ohh well at the moment im listening to a lot of Brand New, with a recomendation from Amber, i know its not my usual afair of loud rawkus noise, like SOAD or Slipknot but i dunno makes a change lol!  Woot i wouldnt say its for everyone and Bald will probably skits about it but hey, i will listen to anythign once and it pays off to do that i feel.    Hehe i wish i had more rememberance of the insightfulness of some people last night, i guess they just kinda made me believe again in stuff, lol i did have a bit of a panicy anyursm, so Thanku Adam and Amber for that!  Woot anyway its great feeling alright today, lol makes a change you know.  So err GOURANGA!  Lol hehe well anyway once again on with this big ol blog show.  Id just like to wish everyone shit loads of luck with their R.S. exam today, i mean lets be realistic, its gonna be fucking annoying, but just smile and get through it because if i c naturally happy people breaking down, im paying a visit to the nearest cliff :P  Hehe well just do ur best and when youve done look Mr Paine square in the eye and give a resounding \"Go Roger urself u fucking Pedo!\"  Ohh how he will laugh!  Ohh fuck Linkin Park\'s just came on Kerrang radio, thats the one thing u have to beware of, along with the Rasmus, *cough Bald*  Hehe Heres a list of shit thats kept me sane!  .Friends, (Their very insightful when they want to be *cough Adam* .Laughing at Religion .Juice .Laughing at how Bad the Rasmus are .GOURANGA  Hmm after thought! ~On sticking a fork up the lead singer of the Rasmus\'s arse he was remarked to say \"AH! AH! its up me shadow.\"  Shazam and cya later!  '),(24521,3099605,'17,June,2004','2004-06-17 00:00:00','   Hello All.  Been a shit day today, im sure lots of you know, or perhaps you dont those days were you wake up depressed, and you think hmm fuck it i will be ok later... but your not and your still pissed off and you feel everyones getting to you.  Yeah welcome to my day.  I looked up all the psychology papers that wer eplaced on the internet today about Teenage Angst just all bullshit about parents worried about their kids and how they try to sum up your whole entire mind about who you are makeing you a good little consumer again.  It destroys thousands of talented people everyday.  But what they fail to realize Angst is not just over nothing at all, its a way of comprehending how you see yourself and i would say just a attempt before the world breaks you to get yourself heard.  Angst is fueled by death ears.  I dunno, well mainly i was thinking about my lack of understanding towards things today, they way i always ask people whats wrong with me, why dont i know... its strange really, i even tried talking to my mother... she wasnt listening so i told her she can shove her head up her arse.  Nice arent i.  I dunno i jsut have taken to increasingly stareing at myself for longer Periods of time when im not on here writeing away on some story that i thinks cool but nobodys ever paid any interest in, it would also help if i had some talent and i could spell.  My sister has always kinda been my insperation in things, but now i c her as a author, *she\'s writing a big ol novel and doing really well*  and then i see me, with my shoddy Confessions of a 21st century vampire not worth the paper its not written on.  She\'s happy aswell, im not, i saw her Teenage life through the eyes of a child, the rows, the times the dress sense, she always summed up how cool it was to be a teenager.  But hey she didnt tell me she was gonna give me a world of Angst aswell.  Stuff hurts you know, when trying to be urself doesnt seem good enough for anyone, when you can talk to someone for hours on end and still feel like they made no connection to you, even though you make a connection to them.  My dress sense, my music, my writing my own mind that was influenced by my sister when i was growing up, (yes i know it sounds weird but their are some family things id rather keep to myself) it all seemed so cool back then, when my sister gave me my first Album... Nirvana Nevermind.  It all seemed so cool and liberated back then but nobody ever told me with that you\'d get a lot of shit days were everyone hates you for not being normal.  Also because ive always been told in my family and these are my mothers exact words \"Never be normal\" i feel i dont fit into anything.  I guess with my parents being quite old im inept at feelings and shit, yeah i guess thats right, i do feel inept.  I cannot tell anyone how i feel without spinning it out for days on end, i have to get other people to start conversations because im to fucking shy to just ask someone \"What music you like?\"  or any basic shit like that.  This feeling of worthlessness does colide horribly with the days as im sure everyone gets were you feel fucking invincible, lol like the gillete advert thats on at the moment, were everything you do affects the world, but when you feel like me, it hurts all the more on those days when you find its not true.  Now it comes to the blog, this is always ridiculeing the basic fact.  I cannot communicate proply, its obvious when you write em all down in a middle man, and people reading it cannot link it to you because its so aloof from yourself.  They cant here me in real life saying these things, i wish they could but i cant do it, im just inept.  Id like to finish with a quote a very good friend of mine says, and it makes sense \"I like to think im useless in everyway\"  Thx for reading, goodnight.  '),(24522,3099605,'16,June,2004','2004-06-16 00:00:00','   urlLink    Woot Gonna start things off with a little guided tour of some of the stuff in my house... hoopla this is me, looking in a mirror... hmm.   urlLink    My Grandad gave me a big old box full of dead critters, this is the first draw, all the names are written in latin and its really creppy but super cool   urlLink    Yay Box Two, all fleas and scarys   urlLink    Yay big ol dung beetles *shudder*  i swear these thigns should be in a museum ohh well on with the show   urlLink    This is a flower that comes out of my cactus for like only 2 days of the year and then dies away, its a real super cool sight and i got it on camera :P kinda sad really it has to die, lol like Breed 77 lyrics *hums \'The River\'   urlLink    My Notice Board in my room, its just basicly were i put all my random shit, u know, poems, pictures, awards, my sandwhich i made once, that was godlike!   urlLink    Just to finish \"i made this!\"  Yay that was a quick dragging kicking and screaming round some random shit!  Eeyore is Me!  '),(24523,3099605,'16,June,2004','2004-06-16 00:00:00','   Greetings  Blog two of the day and i was just wondering what its all about really, u know i realy think they shouldnt give me days off because i think to much.  It just seems that people no matter how innocent they seem to be they are out to hurt me at the moment.  It seems very odd really, almost like the powers of be hate me beign nice and want to punish me for this.  Seems odd really that, but i can imagine whoeevr made all this would in some aspects have to be a pretty stone cold, power hungry nonce.  Lol you know, it just seems that im once again encountering Karmatic black holes at the moment, you know the way you would say do a good deed and instead of life repaying it back unto you it is sucked into a black hole and destroyed forever.  In the end i always end up hammering this, while my own mind laughs at me.  Another thing thats been bothering me at the moment is the way i convince myself of something to be one way, when really its compleatly the other way!  And any attempts to try to to trick this train of thought end up with it becomign even more confused and screwed up.  It just seems like my own mind comes out to get me sometimes.  Thats the problem, i think im paranoid, basicly in essence this means you believe that everyones out to get you.  I guess the only cure is to get peopel to tell you their not, but i dunno maybe they are :S  and every time i talk about it i confuse myself.  Maybe i should just shut up but sometimes i dunno, just need telling that im not off my trolley lol.                             what I\'ve felt                           what I\'ve known                           never shined through in what I\'ve shown                           never be                           never see                           won\'t see what might have been                            what I\'ve felt                           what I\'ve known                           never shined through in what I\'ve shown                           never free                           never me                           so I dub the unforgiven                            you labeled me                           I\'ll label you                            so I dub the unforgiven   Sha fucking Zam That was another Angst filled blog Ive Been Ben Jeffrey Youve Been my lovely readers, Love you all  P.s.  Au revoir and plz visit us at  urlLink Woa   Because Bald has worked Hard on that site and ima gonna put a story on it soon, so even if u dont like what you see just tell us on our forum, you know id do it for you.  *A personal thanku to whoever read this, because writing is important to me and if you took time to read it all power to you.\"  Au revoir my friends.  '),(24524,3099605,'15,June,2004','2004-06-15 00:00:00','   Greetings the folk of my blog.  Yay i had a exam free day today so woots a plenty there!  Hehe but as usual i didnt do a lot with it, just wandered round town... ho hum.  Hehe well anyway hope you all had a good day today whatever you did im sure it was worthwhile!  Yeah well anyway i was just thinking about how tragic everything really is if you think about it and how saying Gouranga really doesnt help.  Yup its the monks again, weird folk arent they... always happy, lol well they were helping children and handing out stickers today so i obliged, listen to SOAD long enough and i challenge you not to feel sorry for the children :P  Well anyway they preach the whole idea of Gouranga and be happy for thsoe of you who dont know... i mean im sorry i just dont get being happy, got nothing to be happy about im sure even fairys commit suicide.  Poor things :s  Lol well especialy if the head fairy was a cock muncher ruining lifes, yeah just Blair and his dancing spin poneys of polatics... Wankers.  So Gouranaga to you all, if thats seriously works i will be suprised.    Lol well for all you naturaly happy people who read this yay GOURANGA  And for those Angst ridden ones  4000 hungry children leave us per hour,From starvation  and for those who get the reference BOOM! to you aswell.  Au revoir  '),(24525,3099605,'15,June,2004','2004-06-15 00:00:00','   Hi All  Hey just had a uber boring but all right history exam now... yeah typical boring near end to a typicaly boring 5 years of work.  Ive been thinking about school life quite a lot at the moment, and how deeply flawed school is, i mean the preps are loved by teachers, peopel like us are hated, people who deserve better get treated by shit everyday of their school life by people who are future shit shovelers.  School all the way through pushes opinions in your face.  R.S. beign a typical example of how biggoted views of our country and religion are placed in your face.  Typical quote of a RS teacher \"You Know religion is silly but CHRISTIANITY IS THE TRUE RELIGION!\"...Fuckers.  I mean they constantly try to get the latest tinsel town polatics down your throat and chew.  Its shite people are everyday forced to think like biggots.... yeah become your parents!  I mean WTF!    Uniforms aswell now this ive always thought was well feeble, i mean they say \"It discourages bullying\"  Yeah bullshit, if people are gonan insult you for what you wear, they gonna do it any way, Uniform is used to try to rob people of their identity.  But of cours trying to show your identity can get you in trouble, unless your a prep of course then the teachers have nothign but encouragement for you!  Yes i will give you more time for your homework prep:.V.013249 Yes i will ignore you as you ruin something of someone elses to validate your misrable life. Yes do what you fucking like just as long as when you leave school you neevr have a thought that hasnt been infulenece by something uve seen on TV and be a good little consumer.  I mean the facts are hidden from us at school, the school shits abotu buying kareoke machines and crappy plastic cafe\'s to encourage the consumer within while all the stuff used for actually learnign something to get out of this shit hole is slowly falling to bits.  \"Ohh sorry their is only 1 text book between 5 i burnt em to fuel our refigerators in  the hall.  School is quite frankly constantly trying to rob your opinion off you and replace it with a nice big placebo.  Yeah welcome to the real world boys and girls!  Your not number 1, Everyone is against you, You cant go around delibratly ruining others lifes to validate your own feeble existance, you WILL die one day.  But you see school hides these facts from you so you will be shell shocked into being a good little consumer when you enter the real world.  Id like to sum up this Angst filled slice with this.  \"Remember Nothing You Buy In This Life Is Real\"     Thankyou for listening.  '),(24526,3099605,'14,June,2004','2004-06-14 00:00:00','   urlLink    I hate them soooo much... so i made this!  Ohh Killroy is a twat!   Edit!  Yay I just got my post on the legendary B3ta message Boards!  (Basicly a site were random people edit pics to make fun of stuff!)  So im well happy!  Here is a link to the great site and hopefully i will get to the front page some day! urlLink B3ta!  '),(24527,3099605,'13,June,2004','2004-06-13 00:00:00','   Hello all!  Heya all, its been a very exam orientated day today :S  and yeah the double wammy of Science and English did suck. A lot!  The English paper was not only boring but the only question i said i was gonna do (describe) was about a shopping centre... thx for that grr.  Well anyway i hope you all did a good job and be proud of yourself have a biscuit play this game!   urlLink Draw to tidy stuff up!   I swear its so cool you have to tidy up all these messy little critters by drawing!  YAY!  Lol well anyway on with the show.  Yeah well the one thing thats got to me so far today was little tiny events that keep replaying in your mind, they kinda repeat because they seem special to you but really they just gonna eventually break your mind into a million pieces all over the floor.  Ive got two of them to deal with today... how sucky is that!  I mean they seem cool when they happen but the more you think about them they wear at you, make you want to do something about something thats already... BUT BANG!  its to fucking late!  Ohh well i guess i just have to w8 for the next one!  (If There is one :S)  Well as to not break my usual blogging style bring on the stuff!  The Stuff Thats Stopped Things Breaking My Sanity like so much communion bread.  .Laughing at Transubstansiation .Juice and lots of it .SOAD\'s 3rd album (Cant w8 till the fourth!) .All the Folk Who read my blog   Hurrah for all those things!  Hmm strange all the lads in our school (me included funny enough) recieved a incredibly small pair of fake testicles today... how odd.  Hmm so remember lads check for testiclular cancer!  (Any preps who read this check ur own balls not the plastic ones!) lmao  So anyway that was sufficently weird... if they wernt so gross id safety pin em to my bag and take em to school.  But they are just a little bit to gross... hmm maybe i will just throw em at a car.  Ohh well anyway bubye love you all and thanks for all the fish.  '),(24528,3099605,'13,June,2004','2004-06-13 00:00:00','   Greetings.  Ive been askign myself this quite a lot lately.  Its odd really, ud think id be able to tell the diffrence.  But i dunno, just the dreams have been so vivid, it almost begs the point of does it really matter?  I guess i live in a fantasy world of thinking people care lol.  I dunno my life just feels like cereal... bite sized chunks   Dreams are odd monkeys indeed, i mean what is the point in them, what are they really trying to tell you... if anything at all?  I mean are dreams really just a way of figuring things out, or are they really ur base emotions... i dont know they just seem to tell a story that runs parrel to my life at the moment vry odd shit.  And sometimes Life feels like a dream, and it hurts all the more when reality comes back.  Lol odd that.  Ohh well today has been pretty dam uneventful, lots of talk of boring old football, and a shamefull display by England in the last two mins.  Hehe so funny.  Ohh well im not unpatriotic, just think that if we gonna play to that standard and talk so much bullshit its only karama that England get owned in such a fashion.  Ohh well im done good luck with all your exams tommorow, cya!  '),(24529,3099605,'12,June,2004','2004-06-12 00:00:00','   Greetings  heya this is my first blog of the day and im already pissed with all this England shit,  i mean if we win i will be shocked and for some reason some people think that by winning some pointless tournament that makes a great nation.  Err no i doubt it very much.    I mean David Beckham is not a hero... he has never fought in a war to create a ego for a mad man.  He has never through his sport saved a life or helped someone when they needed it.  If anything Euro 2004 has enslaved people due to our own selfishness.  I mean flags on cars... Made in China were a worker gets the equivlent on Â£64 a month to make flags all day.  Ohh yeah how pleasant.  Football does not make us a great nation, our ability to kick a bag of wind about does not make us all powerful.  People dont associate Flags with England as a country and with freedom anymore.  They associate it with 11 men who kick a bag of wind about a pitch.  This is insane.    Football has became more than a game... it enslaves countrys and there people, sells newspapers, creates gods out of men, desecrates a flag with JJB, ruins something thats ours.   Yours preparing to be cruely insulted for expressing opinion Ben Jeffrey  '),(24530,3099605,'12,June,2004','2004-06-12 00:00:00','   greetings!    Hello everyone who btohers to read this shoddy corner of the internet i hope today see\'s you well and you have all had a good time today!  Well anyway cut to the chase... the moral of todays story is i shouldnt be aloud to think by myself for to long... i have been thinking to much today and have came home a bit worse for were.  Yeah its the usual Angst, its boiled over and my blog needs to be shamelessly verbaly raped with it.  So here goes.  I was just thinking about myself and how myself links in with others.  Now i have always i guess since i was about seven felt that i never quite fit in to anything, and of course this has to be my fault but ive neevr quite figured out why.  Its odd really how everythings boiled down to such simple chemical signals in your brain yet it takes so long to put down.  You know i think it was Bald who once said to me \"If its to weird to write down its put in music\"  It was a odd quote but it does ring true, kinda links to my \"All great art is based on suffering\" theme.  However this isnt great art... this is just me saying how i feel in the only way i can.  I guess i suck at expressing feelings to others, i dont know why it is, i just always feel people are judgeing me... they probably are i guess.    Its hard to say you dont care when you do.  I dunno how i feel about anything anymore.  Its just strange for instance after i got laid out yesterday in front of all my mates, i have never felt so bad... but it wasnt because this dude had just punched me in the face... no much more than that really.  I guess it was the fact that i linked to me being owned as me ruining the party.  I know it was quite obviously this twats fault, but yet again i felt it was mine... so odd really.  The second thing that came into my mind was the old feeling of beign outcasted.  I guess on its sad base form, my friends that i was with then are quite frankly my best mates in the whole fucking world, and i was just made to look like a prick in front of em... sad but true i know.  Pride, yeah i guess that was one of em, im always so cocky, always telling jokes being a twat basicly and this guy just twated me, owned me... base form, of being a lad... 9 years at karate and i didnt see one punch... just one.  The punch itself didnt hurt, the symbolism did, ive never probably felt so bad in my whole life as i did after that.  I put on a brave face and smile though, the old jokes on the txt message and that wank, but i am honest now.  (Lol you\'d think im done but im not)  But also there is another thing that has really reflected on me, what a incredibly idiot i am.  I mean ive always tried to be there for people, yet i miss things, blind obvious things, i know there is something wrong with people, but i neevr know waht to say... because of my selfish idea of being judged and hurt.  Ive always hated being hurt in the mental sense.  That reflects on the Punching, my body doesnt hurt, i dont give a fuck about it... but im deadly afraid of painful emotion.  Its kinda sad in its simplest form and anyone who reads this will probably neevr look at me the same again... but i had to say it simple as that.  Thankyou for listening, ive been very honest today i feel.  Goodnight.  '),(24531,3099605,'11,June,2004','2004-06-11 00:00:00','   urlLink    Mark and Simon\'s Epic quest to yonder tree begins  urlLink    Sophie Day Dreaming Away  urlLink    Someone plz interpet this, this somehow escapes me entirely  urlLink    Maybe i shouldnt have taken this picture, i think Josie\'s gonna kill me  urlLink    \"I\'ll Get you for this Joe!\"  Says Adam in a menacing Growl, meanwhile Simon pulls a gun as Bald looks on.  Wherever i may roam, No man Vagabond call me what you will  '),(24532,3099605,'10,June,2004','2004-06-10 00:00:00','   Greetings the lovely folk who live in my blog.  Today heralded the last time i ever have to speak or even refer to german ever again, hurray!  With a annoying German test it was party time!  ALSO HURRAY!  Lol so we once again had one of our gravel put bash\'s featuring such legends as Adam Chapmans one can to rule em all, and yes me getting smacked in the gob...sorry about that.  Lol well id like to think we had a great day out, apart from some random twat punching me in the face seemed to put a bit of a downer on things.  Ohh well lol i hope you all had a good time anyway, i know i did, and i once again apologise for bringing the events of the day to such a spectacular close :(  So sorry everyone but at least i proved a few basic facts about preps.  One: They resent others having fun  Two: It takes more than one lucky punch to stop me being a git to em :P  Hehe well anyway i did have a good time so thankyou to everyone that came, and that was!!!!!!    Amber, Sophie, Josie, Rach, Mark, Bald, Adam (The Can Meister) Joe and Me!  Woot Thankyou all for comeing and thx for all the fish!  '),(24533,3099605,'10,June,2004','2004-06-10 00:00:00','   Greetings the loyal inhabitants of MOAARY (Museings of a angst ridden youth for you exam shell shocked indivduals)  Yup now its my proper blog of the night.  Woot!  Well today see\'s the local elections kicking off *yawn* when will people learn it doesnt matter who you vote for... all goverment will screw you in the end.  But hey who gives a fuck about goverment who reads this often?  ON WITH THE SHOW!  Yeah well in over nu\'s today my dad bollocked me while we were eating because he was talking all of his usual biggoted shit, so i did what any depressed 16 year old would do, SARCASM!  Yup and he blew his lid... i cant even be fucking sarcastic anymore in this house,  i swear my family are trying to ban all forms of *jazzy* speech, so blow Irony, Saracasm, creative swearing you know my usual shit.  So after quickly telling my dad what a cock equestrian he was i swiftly came in here leaving them to discuss \"Whats to be done with our son\"  lol they will probably try to read my diary AGAIN now.  Cocktarts they is.  Ohh well looking forward to tommorow about 10:00Am after German we having a shindig woot!  Hope to see you all there and i have crisps and drink!  Arent i a conciderate friend :P  So expect plently of storys of tree climbing and the like tommorow, and if i dare set up a account a bit of the old audio blogging may be present!  Woot!  Ive also started a nu MSN name today in a desperate bid to remove my second name (Thomas)  *shudder* ive started using the Alias \'Vagabond\' lol so now im Benjamin \'Vagabond\' Jeffrey.  And yes Bald, Simon and Mark it was in my diary not my blog.  I get those to confused lol i dont know why i should though, my diarys called \'My kingdom\'  lol well remember from now on just in case i forget MOAARY is on the net, MK isnt!  woots a plenty there!  So anyway cya on MSN, at the party tommorow, or just around :P    Au revoir!  '),(24534,3099605,'09,June,2004','2004-06-09 00:00:00','   Greetings  Just got back fromt the english exam which was boring and annoying lol well could it be anything else with the words English and Exam stuck together.  Well i hope it went well for everyone and good luck this afternoon for your geogrpahy exam if this concerns you.  Well anyway on with the show of bad grammer and punctuation!  Im pleased to note ive finaly got my insperation back to right things yay!  But the downside (or possibly upside) is that i will be quieter on MSN, if this is the case shout at me and tell me anything i write is pants anyway... thats sure to drag me out lol!  Well anyway ima hopeing to blog later i will just leave you for now with this.  \"The World Is a Delicate Bloom, growing up through rocks of iron and electricity.\"  Ive Been Benjamin \'Vagabond\' Jeffrey  Thx for listening!  '),(24535,3099605,'09,June,2004','2004-06-09 00:00:00','   Greetings Mortals Who Live In My Blog!  Well the 1st graphics exam is now over, and i told Sambi just how i feel about his exam.  \"I Blame You.\"  I think were my exact words to him, i hope it made him think... or possibly even cry!  Lol well anyway that exam was UBER Confusing, Super Hard and pointless, yup its just been one of those exams im afraid.  Anyway in other nu\'s once that horrific event we gathered a small frissbe playign crew to adventure down gravel pit.  So finaly after much male bonding by \"Who can climb the biggest tree\"  and \"Who hurt themselfs the most\"  we finaly decided to head on home.  Well what can i say, it may have been childish but it wasnt half fun!  Hehe well anyway i hope today see\'s all my readers well and they all having a good time in their own little lifes!    Now on with the show!  Yeah i was thinking about how other people view you, and the world around them today, having a good laugh with myself thinking about everyones diffrent approachs to each other and shit, and how people treat people diffrently in diffrent circumstances.  Very long winded stuff and i say you have to have a lot of time on your hands like me to think about it!  You know because i was thinking about how diffrent people will see diffrent things in this blog and just think for a moment if you dare.  How do you see the world?  Lol well anyway enough pondering thought im off!    Au revoir  '),(24536,3099605,'08,June,2004','2004-06-08 00:00:00','   Hello everyone  Yup just like the title says, today was to Uber hot, and i feel sick.  Great.  Lol you\'d think that venus moving in front of the sun would maybe cool it down just the tinyest of bits but alas.  Lol well anyway on the exam front i had a wrist slitting four hours of Maths and Business, (shudder stuck in a IT room for two hours in the boiling heat with people who i swear dont wash)  lol well thank god thats over.  Tommorow i got the horrific wedding ring vomiting exam that is Graphics :(    So good luck to us all, i think we gonna need it.  Anyway i hope you all had a nice day today and you have managed to stop yourself from going insane in the heat!  That part didnt bother me much because i probably already am insane.    In other nu\'s i also got some more batterys for my camera so i can take shots of stuff, make look angsty and put em on MSN for all to boggle at yay!  Well anyway i guess that doesnt really matter but i think its cool :p  So once again a depressing day has reached its twighlight hours, and the sun still refuses to stop shining *stupid thing* and im still toying with the idea of going into Audio blogging from my phone, its free now so they announce!  Hmm well something to think about if you my precious viewers can stand to hear my boring voice yammering on about Angst!   Just when you thought reading was enough!  Au revoir  '),(24537,3099605,'07,June,2004','2004-06-07 00:00:00','   urlLink    Yeah i just thought id Post up my latest NBC piccy, im compensating my inability to draw buy photoshopping it instead (Sorry it\'s small its designed to be a MSN piccy).    Well enjoy, and i guess we all feel like that sometimes standing on that hill looking out at the darkened graves. But i think we are all looking for something like Jack was looking for Sally and Sally was looking for Jack just something we can say is worthwhile and alive.    I dunno maybe its just the Angst talking but id like to think we can all find something to hold on to when we are like that... u know feeling dead inside.    Goodnight.  \"I feel theres something in the wind\"  '),(24538,3099605,'06,June,2004','2004-06-06 00:00:00','   Hello Ladies and Gentlemen who live in my blog!  I hope today finds you well and that you all did a cracking job in your exams today, im sure you did and if your worried havea cookie and smile!  Lol well anyway today it has once again been to hot and hammering frissbe after exams for 1 and a bit hours aint advisable in that heat :P  Few!  But hey today i was thinking about how things can be so easily ruined and how one persons actions can affect so many.  Its really annoying really because these people float aimlessly through there lives trying to validate there excistance by insulting and hurting others just so they can believe that they\'ve had a affect on someone in their misrable lifes.  Those people piss me off sooooo much, if you see one today stab them!  You see ive always believed in what you do someone is reflected back on you, but when people break these rules and they seem to never reap what they sow, or repay unto you, it kinda shakes your ability to have faith in humanity.  Ohh well not that i ever had much anyway :p  Yeah well carrying on away from Karmatic Black hole people *whistles Black hole sun*  i do believe that the following things kept me entertained today!  .ASDA Hot sausage :P YUM .Frissbe! .The possibility of future treehouse building *AKA Me and Adam entertaining a dream* .The People who live in my blog!  Woot well anyway my semi optamisim is wearing off so bubye!  '),(24539,3099605,'06,June,2004','2004-06-06 00:00:00','   Greetings!  Yay another day another blog!  The Problem is today is sunday and sundays suck because they are next to mondays which suck in equal measure!  Well anyway hope you all had a nice day anyway and didnt get to hot (HOW HOT WAS IT!!!!)  Grr i hate the sun, it causes me to burn and die!  Ohh well, i tried out my nu frissbe today woot!  Yes and it does fly, suprising!    As you can see im dancing around the whole point that we got a wank science exam tommorow and im not even goign to go into how much i hate that.  But plz one thing, everyone calm the revsion down you will all do fine!  Hehe well anyway good luck to those poor souls who have a got a geography exam!  YAY i get to stay in bed all morning!  (it is to hot!)  Bubye  '),(24540,3099605,'05,June,2004','2004-06-05 00:00:00','   Greetings Mortals!  Heya i felt i had to blog now or i ould possibly never be able to blog again, it just seemed like a do or die moment because a urge has taken me and they are important!  You know when your mind suddenly almost seems to twitch and you get the feeling that you really want to do something in particular?  Well i was just wondering why some people ignore them, its strange really.  Now i want to share something with you all which i find quite insightful.  All of the people who read this are each brilliant in there own individual way.  You may be exelent speakers,  Authors, Artists, musicians what have you.  But i know how it feels, you create something and you tell yourself \"This is the biggest load of SHITE!\"  But stop before you destroy it (Im one to talk i know *shudders*)  Keep your work, and show it to someone, they will apreciate it.  Because i have come to learn something over the past week.  Nothing is trivial, nothing that you create is rubbish, it is art my friends.  Now im makeing a pact to myself and you all now that i will never destroy another piece of work that i create ever again and i will happily look or listen or see anything that you have done yourselfs if you want to share.  Art is a very personal thing, but showing your skill to another is liberation.  Id like to finish by just saying one thing about Urges  \"If you dont follow Urges there is no point having em\"  Thankyou for listening, and be brave, your work is not crap, its brilliant.  '),(24541,3099605,'04,June,2004','2004-06-04 00:00:00','   Greetings!  Yay proper blog now for you!  Hehe well as you all well know i got back from Hol\'s today and of course its all really weird.  Kinda like 28 Days later without the flesh eating mad people \"im sure they will come later\"  But ehh i dunno, plz come talk to me, the welcome home party sucks lol!  But hey anyway on with the show!  My diary has taken a really good verbal thrashing this week, its very funny to read because my lack of insperation astounded me, so i just ended up writing poems and doodleing little notes in it about how i miss stuff.  YAY!  But anyway its good to be back even if Spondon is the suckiest place on earth, i missed my Mates, My Music and My computer chair!  Hehe so im glad i can once again listen to proper music \"My head phones broke on Holiday!!!!\" So that sucked a Uber amount hehe.  But yay back to reality with a bang on monday with a afternoon science exam :(  Hehe so in the launch back to Normality i bring you the stuff that kept me alive during my holiday!  .My Diary (YAY) .Rum (Often and in respectable doses) .Being By The Sea .Chocolate!  Hehe well yay glad i got that list off my chest!  Now for a game to bring back the total sense of Normality!  urlLink Side Scroller Shoot Em Up *How many of these are they :S   Hehe and A random quote i wrote this week! ---------------------------------------------------------- \"Behold My Lack Of Punctuation and Spelling!\"  '),(24542,3099605,'04,June,2004','2004-06-04 00:00:00','   Hello Everyone!  Yay im finaly back in stinky Spondon after a super Boring 6 hour drive :S  Ohh well hope your all feeling good and happy, and ready to face this weeks exams *shudders* yup this is truly gonna go down as a sucky week as far as exams go!  Well anyway im back a bit earlier than the first thought because my dog Peters been a bit poorly, he\'s getting better now but its better he\'s in a enviroment he knows well so he\'s feeling sorry for himself lol.  But anyway now im back to the norm i hope i talk to you all on MSN later   Au revoir!  '),(24543,3099605,'04,June,2004','2004-06-04 00:00:00','   Hello!  Yup the holidays continue to be Uber boring i dunno what it is, just this plac isnt as cool as it used to be.  Ohh well, i had a major bust up with my mum over reading my diary but she has finaly apologised woot!    Hehe well thats the previous nu\'s over, hope your all having a good time back home and i also hope the weathers better than here, Its crap basicly here lol.  Well anyway i dont mind to much because im coming home tommorow woot!  Yeah lots to look forward to includeing Maths exams *Vomits*  Ohh well... *bangs head on screen*    Bloody GCSE\'s ohh well they will soon end and we will all be happy again LOL  Remember cant rain all the time!  Missing you all loads Ben!  '),(24544,3099605,'03,June,2004','2004-06-03 00:00:00','   Greetings  Im in a bit of a down mood today, my family has been pissing me off and with us alll being together and me having to be in the same room has them has generated something close to cabin fever.  Yes it is Uber boring!  Well anyway  i hope everyones having a great time back home, and that the inpending exams arent giving them to much hastle!  Well anyway on with the show!    I think this may be my last holiday to Perranporth which fills me with mixed feelings really, I mean so far this place has yet to inspire me so as always my poetry and writing goes un written.  Lol which probably to most people is a good thing.  Also i dont think i can stand to go with my parents because they are to old and annoying at times and sometimes it really pisses me off.    I miss everybody loads aswell and normaly the holiday would take my mind off this but of course this time its not.  This blog i guess is now my last refuge of my minds thoughts after i do believe my Diary was plundered this morning by my gimp Parents!  I mean wtf do they think they are playing at!    I have to go for a \'Meal\' with them tonight, believe me in my head i will be somewere else, somewere far away!  Lol probably just imagining something and i will have fun for ages until i remind myself its not real and i tell myself how crazy i am!  Hehe well if anyone still reads this they will know that i am in fact a crazy unconsolable wannabe Vampire, pointless waste of air!  Yeah well one good point today is i went to Newquay Zoo and i love it there.  The animals are super cool.  Feel sorry for the Lions though, there cage is to small, i dont like looking at the lions, They reminded me to much of my present situation in that moldy old shack they call a holiday apartment.  Yeah things appear to be getting worse here.    Plus it sucks because at the moment im watching my time at the internet cafe slowy rot away and i know i can only get away from my parents for another hour over the next few days without just going and being incredibly boring :s  GRRRR!    Well anyway i hope everyones having a great time back home, hope the weather picks up for you and you all enjoy it!  Thx for listening   Miss you all lots.  Benjamin Jeffrey.  '),(24545,3099605,'01,June,2004','2004-06-01 00:00:00','   Heya Everyone!  Hehe Just finished emailing everyone on my MSN list (phew!) So if you didnt get one im Uber sorry but its not me its these computers!  But anyway after righting god knows how many personlized emails i feel like ive earned a Ice cream!  I swear there is a dairy in Perranporth that if god wanted a ice cream he would go there!  Mmm Liquorice!  :P  Hehe well anyway im glad i could blog today because my diary has until this day taken all the flogging my blog would normaly get!  Well anyway hope your all having a great time here and im gonna go play Frisbee in a minute on the beach!  YAY!  But first i gotta say some stuff thats so far kept me alive in this place!  .Perranporth Dairy!  (Mmmmmmm Iced Cream) .The Daily walks on the beach ive been takeing (Mmmm Meditation!) .This internet Cafe!  (YAY ELECTRONICS!)  I Miss You All!  *HUGS*  Luv Ben  '),(24546,3099605,'31,July,2004','2004-07-31 00:00:00','   Greetings mortals.  Well i spent most of my day today with my dad on a boat fishing today... it wasnt that bad actually, we caught a few fish had a few laughs all that kinda stuff, but it was also the usual fishing syndrome. A lot of time to think... and i mean a lot. Now i know probably nobody who reads this has ever gone fly fishing before, so basicly for you lot, u spend a long time doing not a lot. Just waiting, repativly fishing in and out, trying diffrent techniques... till finaly BANG the fish hits your fly pull it in pull out the reaper (or not if you catching and releasing) then wack bang food to eat.  Ohh well but as you can see while your waiting to kill, you have a lot of time to just think about everything, so i always come back angsty, and you know arr! I just dotn know what to think or how to deal with it at the moment, good job im a expert at forgeting things, its the only way really when nobody will face you when you need them to.  But anyway i just cant think of much to say really but im just drinking Rum till my mum wont let me anymore. Meh.  I apologise for the length.  Ben.  '),(24547,3099605,'30,July,2004','2004-07-30 00:00:00','   Hiya everyone.  Woo what can i say, today was kinda shite really, nothing seems to have gone on, i appear to have somehow wrong quite a few of my friends but i do not how i did this, im sorry thats all i can say but can you please just talk to me and we can work this out perhaps?  Meh so its just been quite crap that way that someone or something appears to have destroyed most of my social life at this time... *sigh*  Ohh well also went to karate tonight, so im compleatly knackered now, and im also out all day tommorow in a boat fishing with my dad, so my mums gonna have to record Edward scissor hands for me or their will be a price!  Yeah so at this moment i feel both knackered, and a bit upset, because i just dont know what ive done... please folk talk to me.   Sorry its short i dont have much to say. Ben.  '),(24548,3099605,'29,July,2004','2004-07-29 00:00:00','   Hello everyone!   Woo yeah nothing like the smell of fresh blog in the morning!&nbsp; Well anyway this is one of my early posts so by now i can imagine you all know this is not about a lot!   Well anyway for those who are still interested ima gonna be doing a new update for my website over the next few days, but the only problem is i have absoulutly no idea what ima gonna do for the website, i just appear to have ran out of ideas.&nbsp; Lol i would add some more poems, but the problem is i havent done any i really like recently, also i still havent edited my short story, The named \'Ressurection Men\' until Mark reminded me that there is already a book named that, and he also says its rather good.   But anyway i tink thats to much work to close to a deadline ive set myself, so if anyone has any ideas that theyd wanna share, plz chat to me on MSN about it, leave me a comment, or sign my guestbook my website.&nbsp; Thankyou very much for that.   Ben.  '),(24549,3099605,'29,July,2004','2004-07-29 00:00:00','   Hello everyone.  Yup been one of those \'Low\' days today, were you aint depressed as in a sense is you cant handle the problems that come at you, rather a day were you wonder things like Why, if and how.&nbsp; I dunno ive just been thinking a lot today about all the stuff i appaear to have cocked up dramaticly of late and i was wondering how these things could have played out diffrently, i guess thats were the title comes from.&nbsp;&nbsp; You know your always told that you cannot undo what has been done, you cant forget the past, if you dont know why somethings happened you\'d better find out.&nbsp; All that kind of stuff, well i dont know why they tell you this because its not as if you dont know.  Even when they do tell you, people still try to anyway, even if it will make them look more stupid they are still gonna try.&nbsp; Im still gonna try, ive never seen the point in just not bothering, i rather see it as when to bother, but nobody can tell you that.  I just feel like a lot of stuff thats gone off at late is broken, been sent out of my sites, before it can be fixed, and thats what annoys me, but even if i have to set the whole production line back to zero, im gonna mend that one broken pot.&nbsp;  But how do you mend it?&nbsp; As ive said... never been a wise man.  '),(24550,3099605,'28,July,2004','2004-07-28 00:00:00','   Greetings Mortals!  Yes it is I once again writing in this my loyal blog about the mental sways and physical effects of the one they call me!&nbsp; Woo, so without further a do i do hope you all had a cracking day and let me tell you the story of the mindset of the day.  Yup im not hiding anything, this morning i felt like total 110% shit.&nbsp; Yup i only managed to get what i think was about 2 hours sleep the night before, i dont know what it was but i just couldnt switch my brain off.&nbsp; Just one of those things again but you know when you feel ruff your mind turns to ruff angsty thoughts and by lunch time i was about ready to start screaming, smashing things, and kicking the shit out of any mortal that should cross my road.&nbsp; Lol pleasant arent i.&nbsp; Well thats how it was, yes im afraid to say it but i cant be happy all the time :P  Hehe well anyway, so i just spent the day like any other day apart from that really, only the mental tiredness of myself seemed to force my angsty thoughts even further into the front of my head, dam them.&nbsp; So eventualy i had to tell someone and i did and now i feel much better really. (Thankyou again Rose)&nbsp; Yeah so i was saved from a compleate break down woo yay hoopla!&nbsp; Hehe so id just like to say a few things that have been bothering me.  .If you have a problem with me, please tell me you might aswell. .Friendship is not something i fuck around if i think your my friend that means something ok :P&nbsp; so yay all you people who ive said are my friends, you are treasured by a mad person! and finaly RAR only about 6 days until im told how shit my article is!  Woo so thats the verbal vomit over, i feel much much better :P. With love and cookies Thanks for listening Ben!  '),(24551,3099605,'27,July,2004','2004-07-27 00:00:00','   Greetings.   Woo yay ive had a duff day, lol one of the worst days ive had for quite a while really, it was jsut really boring, pointless, and my old MSN account got hacked by some random twat who i have a hunch has sent out some horrid emails to people.&nbsp; :(&nbsp; so it really wasnt me.&nbsp; So all of you who read this blog and i know because ive seen you in real life will have a nice little invite to talk to me once again on my pretty clean multicoloured Gmail account.   Woo well anyway still feel uber down today but theirs not a lot i can do about that, it almost seems out of my hands, you know almost as if somethings been taken from me and i cant seem to get it back.&nbsp; Plus im super pissed with this hacker person because it seems its someone i know... its just the way they operate and the things this twat has said to people.&nbsp; Its just unfair why cant they just tell me what their problem is to my face instead.&nbsp; I dunno i just feel hated for the wrong reasons at times.   Also there is a lot of shit that i want to talk about to a lot of people at the moment, you know tell em whats on my mind, so i dont have to be in my shell and shy about anythng anymore, but i cant find the words to say it and some people dont have the time to listen to it.&nbsp; I just wish i wasnt so shy and so fucking afraid of reality most of the time, im just annoyed with myself that i seem to have no control over my own destiny, i feel as if everything has been taken from my hands.   I dunno, just sick of it at the moment, just need forget everything.  Thanks for listening  Ben.   &nbsp;   &nbsp;   &nbsp;  '),(24552,3099605,'26,July,2004','2004-07-26 00:00:00','   Greetings everyone.   Sorry i havent blogged earlier today, it just appears to have been one of those things really, like so much else thats gone off of late.&nbsp; I mean late last nigth i found out someone hacked my MSN account to send a mean email to someone, which rather pissed me off (if anyone else has recived one of those emails it was not from me and i do hope that you didnt believe it was)   Yeah well anyway todays been a bit Angsty because everything seems to suck at the moment its like im talking in a language other people dont understand and i seem to be somehow outcasting myself again... its not fair and really sucks because i honestly dont know whats happened or what ive done wrong it just seems like something broke without me noticing.&nbsp; I wish i knew what was wrong but i dont and no matter what i do i cant seem to be able to find out without having to do something drastic.   But if i have to i will.&nbsp; Its just one of those quiet things that nobody ever knows... but i want to know.&nbsp; But i dont know if it will let me.&nbsp; I guess i&nbsp;kinda feel emotionaly blacked out you know, like everyone is enjoying something that i want but i cant get it, like im not worthy, or not good enough to get&nbsp;it.&nbsp;  *sigh* i dont know what to do, i just feel alonesy.  Ben.  '),(24553,3099605,'25,July,2004','2004-07-25 00:00:00','   Hi all  So you really want to know the story of my day?&nbsp; Well i warn you it isnt one for the faint hearted.&nbsp; It was fucking boring thats why!&nbsp; Arr!&nbsp; I seriously cant believe how very little i have to say today, basicly all ive been doing is thinking.&nbsp; Which for me is a very very dangerous activity.&nbsp; I dunno maybe im just good at pissing myself off but i feel really down at the moment, just nothing seems to work, nothing seems worth bothering with because anything i do take a interest in kicks me in the face.  Woo well anyway plus i havent sat down and wrote anything since i finished \'Ressurection&nbsp; men\'&nbsp; which seems ten days ago, but i really have no idea because time seems to have no meaning its rather just a sequence of events glued together by the night sky and little neon lights.&nbsp; *Shakes head* meh, i dunno just feel ARRR!&nbsp; Feel like i want to just sit in a corner and die but on the same time want to totaly rock and roll destroy shit.  I dunno maybe im just going that little bit closer to a nervous break down, i bet there quite fun really, in a abstract kinda way.&nbsp; All these thoughts get in my brian and they just push to get out, its driving me rather maaaaaad i think.  Ohh well i fancy writing another \'Make your own adventure thing for the blog so here goes.\'    Your Name is   Pus Boy/ Pterodactyl girl    Your Arch Nemisis is   Yourself/Some kinda Anti Matter Blob/The goverment    Your Arch Nemisis killed your   Bf Or Gf/The Queen/Your Fave Band/Mittens the singing sock puppet    You fight your Arch Nemisis   On top of a mountain in the driving rain/ On a Train/ Your house/ Alternate dimension/ In a vat of Tomato Sauce    Your power that gives you a edge is   Your stunning Looks/The power of flight/Smelly socks/A gun    How it all ends   Tragicly/Happy ending/Wait for the sequal?  Make your own adventure.&nbsp; Woo cya Ben.  '),(24554,3099605,'24,July,2004','2004-07-24 00:00:00','   Hello everyone.  Woo its been a rather cool day today and i hope you lot have had as equaly cool one, well in the morning i just got on MSN early and jammeed about, but i dont think anyone was interested in talking to me so i just added a guestbook to my website which can be accesed from above, plz visit :P.  Lol well anyway after that Me Adam and Bald went up my dads alotment to do some work in exchange for beer and jollys so we had a fun couple of hours digging while listening to Nirvana, SoaD and Guns and Roses thanks to Balds sound bug which we managed to get to work on a for sale sign.&nbsp; Woo so i get mucho brownie points from my parents plus free booze, woo yay hoopla panowie polly wants a cracker.  We also managed to skank 4 tinnys of carling outa Adams sister so mucho woot to that.&nbsp; So in all afternoon has been rather kick arse because all weve been doing is jollying round in the nice weather drinking ale.&nbsp; Yay.&nbsp; Lol well anyway i got curry for tea tonight, no angst to talk so everything is rather woo here, hopefully talk to some of ya\'s way later on MSN if you fancy living after midnight.&nbsp; Well till then Bubye ma Cheries.  Ben.  '),(24555,3099605,'23,July,2004','2004-07-23 00:00:00','   Hi everyone   Hope you all had a good day today and enjoyed this good weather that we are having... SHOCK HORROR!&nbsp; Lol well anyway today Ben has been doing a lot of Dusting and dances with hoovers, which sucks quite a lot (my mum wanted me to clean the whole house over)&nbsp;&nbsp; *Sigh* so anyway i made the most of it by cleaning to the tune of The Origin Of Symmetry album by MUSE and i tell you now that polishing a large table to the tune Of New Born and Bliss is a very odd experince indeed&nbsp; \"Destroy the spineless Tell me its real!\"&nbsp; lol very odd indeed.   Woo well anyway todays just been one of those days aswell so far were everything seems to be incredibly pointless and every word that comes out of your mouth total crap and you think that everyone knows how much bullshit your talking... now thats annoying.&nbsp; Also annoyed that for some reason unknown the twats at my bank decided it would be a jolly good jape to send me a bank statement this month missing about Â£6 for reasons unknown, so anyway after having a rather annoying call to the bank i final managed to get that rectified because some twating thing had gone wrong, but that was only after i had to suffer some pre mature prick named steven for about 20 mins.&nbsp; *sigh*   So err plently of normal just plain life angst today such as annoying banks and people ringign from deeply weird electrical companys called Shelldon.&nbsp; Then being repeativley laughed at doesnt help even though i probably deserved it because i was wittering on.&nbsp;&nbsp; Ohh well one appaears to be going mad of late plus ive set myself yet more tasks to do and its driving me a bit mad, somebody do me a favour knock me out or take me out and spoil me or kill me or something just stop my mind from working.&nbsp;  Bubye. Ben.  '),(24556,3099605,'22,July,2004','2004-07-22 00:00:00','   Hiya all  Ohh dear Ben\'s writing a late night blog, turn away now all of a nervous disposition.  Yup well youve been warned ive worked some angst to the surface and now its gonna come out in a readable format.&nbsp; Life at the moment seems so stale and dry, like its lost all meaning because my mind becomes more and more consumed by things i want but cant have.&nbsp; The way little things seems to bother me is urging on the scarey now, i cant seem to be able to think straight anymore.&nbsp; Its like there is something in the way i cannot hope to break.  But ohh no it cant possibly be it for angst tactics change you see.&nbsp; Sometimes when your with people and your talking, getting a long like a house on fire its like woo yeah hoopla ive cracked it im finaly starting to get somewere, and then its like next time you talk to them, you expect a open door to come in talk have fun hell maybe even smile, but no the locks have changed again.&nbsp; Makes you wonder why that happens, people invite you in with one hand and drive you out with a crucifix in there other hand or something.&nbsp; I&nbsp;mean what is with this notion i keep feeling that somehow people think themselfs i dont know i wouldnt say better than me but rather that because of who they are they cannot ever fully accept me.&nbsp;   Its strange i see it everyday that people wont accept each other unless they fit perfectly into their lifes already.&nbsp; I dont know i just feel like i dont fit sometimes and my attempts to are so vain and pointless i might aswell not even try to.&nbsp; But you see this is were the true edge is found that just when you want to give up, say fuck it all forget it you aint never gonna burn my heart out all that bullshit people invite you back in.&nbsp; You enjoy urself for a while and the next day cast out thus apealing to the instinct to fight for what you want.&nbsp;   Something has gotta break in these patterns and they just seem like they wont, they just a permanant and welded together, unstopable, unending , infinate.  You get so many conflicting messages the whole time, sorting through them is a pain i mean if sevral diffrent factors give you sevral diffrent factors but your only aloud to use one, how can u possibly decide you know, yes you test.&nbsp; You see whats gonna be the best course of action but when you arent aloud to test you feel as if you might just drop it all and go live in a hole.&nbsp; Im just sick of not being able to say exactly waht i feel, i dont want to risk anything that is worth keeping but at the same time id want to risk it just to get what you truly want.  Fortune favours the brave they say. Carpe Dium is what they yell. But they dont understand that saying is very far away from doing it.  Goodnight. Ben.  '),(24557,3099605,'21,July,2004','2004-07-21 00:00:00','   Hi everyone!  Well its that time of day were i once again get my blog on woo.&nbsp; Well anyway i hope everyones had a cool day today and im glad i finaly got my wish of it thundering woo yay hoopla!&nbsp;However the only draw back to this was that it cut the signal to sky in the last te minutes of one of my fave Buffy episodes ever!&nbsp; Dammit, lol and if you were interested which on it was it was part one of were Angel trys to end the world using the big stone demon.  Woo for any of those who do not speak Buffy&nbsp;i apologise for that.&nbsp; Anyway carrying on Bald has managed to kick out a little update for WOA which contained a Link to my website, and also a link to  urlLink B3ta  which is&nbsp; a website which i frequent often posting crazy pictures i make up on photoshop.&nbsp; Yes so im rather pleased he\'s updated so he deserves a woo a yay and a hoopla and possibly a visit to his blog!&nbsp;   Yup and in other trivial nu\'s i was messing around on the flash game that lets you create your own sonic person again and i shazamed up this happy little fellow.     i can imagine he will have some pretty sad story to tell about revenge for his dead wife/friend/cat whatever and a wandering soul keeper/crow/tramp took pitty on him and gave him&nbsp; super powers of Strength/Speed/Noir Graphics&nbsp; to defeat the evil Druglord/Pimp/Car owner that took his most treasured love away from him.&nbsp; YAY MAKE UR OWN ADVENTURE  Hehe well nevermind that and on with the show!  Well today has been rather boring again in my view, accept for the kick arse thunderstorm and finaly beign able to send away my column.&nbsp; Woo yay hoopla once again for that.&nbsp; But hey after that my day was mostly filled with a craving to listen to Oasis *dont ask me why* and im alright in happiness at the moment, just that usual angsty stuff putting pressure on me, i dunno, just confused a lot by it.  Ohh well hopefully i will have more on that for you soon. Cya later Ben.  '),(24558,3099605,'21,July,2004','2004-07-21 00:00:00','   Hey all!  Just a quick morning blog just to tell you in advance why early august is gonna be super jumpy for me.&nbsp; Lol well anyway most of you lovely lovey people who read my blog know that im trying to get a job writing for the Derby Telegraph.&nbsp; Lol or as mark calls it \"That evil lefty rag*&nbsp;&nbsp; Hmm.  Well anyway after a lot of Amazing help from Amber, (The Queen of proofreading) i have finaly managed to send my article off to the Derby telegraph (or evil lefty rag.)&nbsp; So after a super quick get it away from me email its painful enough telling me how much i suck just do it uber quick!&nbsp; So anyway when i get a email back a few mins later, i thought \"YAY THEY LET ME DOWN QUICK!\"&nbsp; i got a sudden shazam as the email just said \"Thankyou Ben i will enjoy reading your article and&nbsp; we will tell you are decision early august.  EARLY AUGUST!&nbsp; ARRR!&nbsp; DAM U LEFTY RAG!&nbsp; Lol well anyway thats quite annoying due to the fact that why cant they just tell you u suck then and then.&nbsp; Obviously that is beyond em but hey nevermind.&nbsp; Right well anyway i might aswell put up the article i sent to them on here so all my bloggies can have a cheap laugh.   My Column   Hello!  &nbsp;  I guess if youâ€™re going to be reading this article frequently, I had better introduce myself. My nameâ€™s Ben and I have just finished compulsory education. My favourite pastimes are hanging out with my mates and writing so typical teenage stuff there, Ohh and did I mention Rock music and poetry, yes Iâ€™m afraid the parental figure may stereotype me a little.&nbsp; I am also hopefully going to college in September to study History, Biology, Media and Communication Studies. Well I say hopefully, because I need the right grades to get into the college I want. So I still have a long wait to find the truth.&nbsp; Iâ€™m in the middle of a ten week holiday between secondary school and college, kind of like an educational limbo, but that doesnâ€™t bother me much! In this long wait, of which I am suspended in at this time, I have promised myself I would do something that would really change my life for the better, well at least I hoped it would but I really didnâ€™t know how.  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; I wanted to write a book I mean, Iâ€™ve seen it done and heard the tales of people writing books in their lunch hours penning great stories and I really wanted to do one myself.&nbsp; Although, the bitter truth of the matter as I am finding out is, it isnâ€™t that easy!&nbsp; I just donâ€™t seem to be able to write something that Iâ€™d actually enjoy, and thatâ€™s what I wanted, something teenagers could enjoy that actually meant something. Not just another book in a pink cover!&nbsp; Thatâ€™s what always annoys me about modern literature, it removes greats from bookshelfâ€™s like â€˜Interview With The Vampireâ€™, â€˜The Lord Of The Ringsâ€™ and â€˜1984â€™ which are being overlooked by modern teens and I wanted to end this.&nbsp; Yes it almost seems like heresy to even think I could write a 1984 but one can dream!&nbsp; But of course the reality is always very different.  &nbsp;At the beginning of these ten weeks I thought how carefree I would be, just like when I was younger and id spend all my time playing computer games and hanging around with my friends.&nbsp; But thatâ€™s not all of it, I guess my freedom has come at a price, and that was the discovery of the ever-stereotyped teenage angst.&nbsp; The â€œhe will grow out of it phase!â€  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; These ten weeks of my life that im going through are defiantly influential. There is absolutely no doubt I have experienced some great things, but with these experiences comes worry.&nbsp; Nothing is trivial, that I have learnt well, due to the fact that everything people do seems to be suddenly of top importance to me, girls drive you crazy with the smallest meaningless gestures, parents can say things to you that will just drive you up the wall. Even if it is just â€œWill you polish the chairs for me today Benâ€ It seems that pressure is constantly put on you to do the right thing, to be a good person and to make the right decisions.  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; This pressure comes from so many directions at times, usually at the times you just want to forget everything. You already have a enough trouble worrying about whether or not your going to pass your Maths exam, without having to worry about somebody hating you just because you listen to rock music or donâ€™t abide by some unwritten rules.&nbsp; Everyone has an opinion, I believe they are entitled to it, but how they use it sometimes is just wrong, and that is what brings people down.&nbsp; I always like to think I can help people, and I constantly see people who already have enough problems being hurt by others, just because they donâ€™t wear the right clothes, or act like the person someone else wants them to be.&nbsp; I know in these ten weeks, even if I donâ€™t write my novel, Iâ€™m going to come out of it ready for college with an understanding that I never had before. Hopefully, by reading this you will to.  '),(24559,3099605,'21,July,2004','2004-07-21 00:00:00','   Greetings one and all to my blog woo yay hoopla!  Lol well anyway today was very much like any other day for me at the moment, spending time writing and makeing photoshop pictures while talking to folk on MSN and then going for some kind of walk to somewere out in ockbrook.&nbsp; Well anyway todays walk so me dragging Bald, and Mark round ockbrook graveyard takeing pictures of the old graves.&nbsp; Yup what a wholesome activity this was, however we all had a reasonable time doing that and jimmying round the country side.&nbsp; Woo.&nbsp;   In other news my mum keeps telling me i look depressed again, err woo she\'s probably right i mean i listen to, to much MUSE, i have a inability to get anything i truly want and im just dam well pissed off with stuff a lot of the time.&nbsp; I dunno really just wish stuff was easier than it is... but guess what *shocking revalation* its not.&nbsp; Well at least not to me anyway, i believe that everyone has one state of mind that they find very easy to slip into and one they find incredibly diffrent.&nbsp; I guess mine is that of being totaly neutral, i always on some level have to be into something if it interests me, the neutral stance is just something i have not got the hang of at all.&nbsp;   Or in true confusing physcological style it could be that im UBER neutral... I just dont know, i dont really know anything will everyone please just explain!  thanks for listening Ben  '),(24560,3099605,'20,July,2004','2004-07-20 00:00:00','   Greetings all!  Just a quick blog to give you some little tid bits of info on this less than fine morning that we got going on at the minute.&nbsp; Woo for starters my website&nbsp; (See above) has just suffered from a new update and has got some nu piccys and a poem for your enjoyment if you fancy it.  Also in other news&nbsp;last night about just after 12:00 i managed to finish writing my Edgar Allen&nbsp;Poe style short story which im very proud of.&nbsp;&nbsp;Its Called Ressurection Men and if you want a copy, which i doubt you will please just ask me,&nbsp;*Not yet been&nbsp;checked&nbsp;completely yet&nbsp;mind so you will deffo geta&nbsp; very ruff version*  Lol well anyway im concidering wether to NET \'i\' fie it&nbsp;or not but if i do that version&nbsp; will have gone through all its checking process, but hey i dont know, im just really pleased i managed to finish something that im actually quite proud of.&nbsp;  So just one final word if you&nbsp;visit my blgo can you please check out all of the sites linked with it, it would&nbsp;cheer me up no end!&nbsp;  Thx proper&nbsp;blog later. Ben.&nbsp;  '),(24561,3099605,'20,July,2004','2004-07-20 00:00:00','   Hiya everyone  &nbsp;  Well what can i say today has been a day very much like any other day, apart from the sucky fact that ive only just been able to get into my room due to builders causing the hall to my room to be off limits while the level the thing for nu tiles.&nbsp; *Sigh* doing minor acrabatics just to get into the toilet sucks quite a lot, especialy because you know if you fall in the stuff your mother will kill you.  &nbsp;  Woo well anyway that didnt happen and instead of killing me my mum bought me a pair of baseball boots, yay!&nbsp; Lol also finaly managed to get a cheap copy of Incesticide by Nirvana.&nbsp; I love this album because it has a lot of songs that unless you are really into your Nirvana your never gonna hear and thats a shame because it is so grungy you just feel dirty listening to it yay!&nbsp; Woo so once again if you dont have this album ask me for a copy and err if u want some quality mp3\'s to punish bill gates\'s evil windows media player look no further than these great works from incesticide.  &nbsp;  Dive, Sliver, Downer, Aero Zeppelin, Been a son and de new wave version of polly... not the normal one that is on Nevermind.  &nbsp;  Woo quality every bit of it, lol well nevermind my grungeiness today and back on with the blog!&nbsp;  &nbsp;  Yeah well ima gathering up material for another Benjamin The Immortal update in the next couple of days so errm any suggestions yell at me and say \"BEN i will only look at your stupid website if...\"&nbsp; Whats in the dots is up to you :P  &nbsp;  Yeah well anyway i hope everyone had a kick arse today and i really really want it to thunderstorm later so err if we can get some positive thinking on that that would kick major league butt.&nbsp; Also im yet to send off my article *pales* but hopefully i can send up off in its most uber awsome state i can get it.&nbsp; Just want them to judge me for my crap writing skills not for my crap spelling!  &nbsp;  Hehe well anyway i will probably have some angst for you soon if i can force myself to think about it, and err be patient till then and remember you guys rock!  &nbsp; /Edit If my pics aint loading its not my bad photobucket is being tempremental sowwy hopefully fixed soon so i can start work on the update.  '),(24562,3099605,'19,July,2004','2004-07-19 00:00:00','   Hello!  &nbsp;  Well Monday is now officialy in full swing, and its been alright really nothing to major league in the angst department, im just not thinking about it.&nbsp; Hope you all had a good day today and didnt have to wake up as early as i.&nbsp; Dam building folk were supposed to be makeing the hall floor level outside my bedroom today so i had to be out of my bedroom by 9 or i couldnt get out my room for 10 hours, but guess what, they decided there gonna do it tommorow instead!&nbsp; So i basicly got up early for no purpose at all.&nbsp; Dam them to hell.  &nbsp;  Ohh well i did go to town today though and finaly found MUSE\'s first Album, Showbiz!&nbsp; Yay woo hoopla!&nbsp; So another wave of Angsty, Wammy bar and harmonium fueld rock has devoured my computer this afternoon and quite frankly its a dam good album aswell.&nbsp; It is deffinatly more basic than their other two and a deffo build up to the winging, writhing guitar sound of Origin of Symetry, but it is a very staunch and deppressing, deffo worth asking me for a copy of it if u into that kind of thing :P  &nbsp;  Hehe well in other news i found this super duper kick arse flash game last night that lets you create your own Sonic the hedgehog style character, and that can be found  urlLink here   Lol so in honour of WOA and its colour scheme i present to you!     &nbsp;  Yeah and in one final piece of nu\'s ima hopeing to cram in a update for my website which can be accesed through the banner above so err if you like check it out.&nbsp;  &nbsp;  Thats about it for now, cya later  Ben.  &nbsp; Edit: At 7:15pm The update of my website has just been compleated, it now has a hit counter, two of my poems and some new piccys.  '),(24563,3099605,'18,July,2004','2004-07-18 00:00:00','   Hi all. &nbsp; Well once again im here on another far from interesting sunday nigth blogging up my far from interesting day.&nbsp; I hope you all had a much better day than me and managed to have this thing that i do believe is called fun. &nbsp; Ohh well anyway, today saw me doing not a lot at all really, just the usual sunday affair of going tos e my Granny and Grandad.&nbsp;&nbsp; So that cues been laughed at for a variety of activitys which i take part in then going up the alotment and getting rained upon.&nbsp; Sigh i hate the weather, it always stops me doing anything i want to do, and its always used to stop others having fun with me and such like... well i suppose i ought to just say this quote \"Cant rain all the time\" but im starting to feel that, that is wrong. &nbsp; *Sigh* im just sick and tired of being constantly blown back in everything i do, its like running constantly against a brick wall that you know is not gonna give in but your gonna try to smash it down anyway.&nbsp; I dunno what makes me keep running at it, i might have just convinced a part of my brain that the wall has eventually got to give, but i think im lieing to myself. &nbsp; \"Steal The Life Thats Inside Of Me\"&nbsp; Yup that about sums up how im feeling, i have to keep trying but i really need to stop because i think its starting to take its toll on both my body and my mind.&nbsp; Yup going insane thats probably it. &nbsp; I dont know what to say apart from that, just feel like im in a void again and i have to get out one way or another. &nbsp; *Takes another run up at the wall* cya later Ben. &nbsp; &nbsp;  '),(24564,3099605,'17,July,2004','2004-07-17 00:00:00','   Hello, my name is Ben, some folks say im a terrific moron, if i could pick one word to describe myself it would be Fluffy Puff Marshmallows... or Ben!&nbsp;  Bald: \"Dont forget Moron!\"&nbsp;  Ohh yeah and that to. &nbsp; Lol sorry about that, hi everyone this is of course my second blog of the day however due to the fact that my morning one was angsty this one... wont be!&nbsp; Woah, suprised aint you hehe well anyway&nbsp;i can tell ur secretly impressed by that anyway.&nbsp; So many of you would probably ask, \"Why the hell is Ben so Happy?\"&nbsp; Or what you really said is \"I cant believe im reading this morons blog again, and has he been on his mums pills?\"&nbsp; Lol no i havent actually been on my mums pills *Shock horror* but i however did have a rather respectable afternoon of basicly wandering around not worrying about anything with  urlLink Bald  &nbsp; Woo so by doing this i finaly managed to steal the Danger Keep away sign ive always wanted to go with my high voltage one yay!   &nbsp; Woo so i am now on the look out for a biohazard one and a radioactivity one, so if u tell me were ones at i can go Jimmy it!&nbsp; &nbsp; Also my Sis and her Boyfriend came round our house today which was quite cool because i aint seen em in agggges and they were just wittering about how they wanted to go to Wales&nbsp;for there 5 year aniversery.&nbsp; (Jesus why they do not get married i do&nbsp;not know)&nbsp; Ohh well there choice and i dont really fancy going to a wedding, bit to christian for me but hey its nice i guess. &nbsp; Woo me and Bald also did a 11 second Panaromic film of this amazing field/park that he discovered and you will be able to check it out on his blog later and you will all start agreeing it is 100000x better than gravel pit, plus we wont get hastled by Chavs, or \'The man\' &nbsp; Well sorry its a bit on the short side but i guess i got more to say when im down in the dumps, give it time amiegos! &nbsp; Cyas Later! Ben  '),(24565,3099605,'16,July,2004','2004-07-16 00:00:00','   Hello everybody &nbsp; Phew decided on a early blog today just to start this freezing cold day off with a varitable bang of blogging.&nbsp; Well as if you didnt already notice today is Saturday and as always my Saturday would not have been compleate without wakeing up not being able to walk.&nbsp; *Sigh* why does Peter always sit on my legs on Saturday!&nbsp; So unfair, and kinda painful to. &nbsp; Grr plus it so didnt help that i actually got a proper 8 hours sleep last night at a somewere near human time, i swear these normal hours are not good for you, but i was knackered last night after my big walk to hermits cave and then somehow later that night ended up at Redhill summer fair in ockbrook.&nbsp; Ohh well at least that was cool because we did get free icecreams off Bald yay!&nbsp; Shame about the fair mind... worse than the one they always hold at Borrowood (my old junior school) and that is most deffo saying something. &nbsp; Well anyway enough pleasantrys on with the show! &nbsp; I dunno really ive just been feeling down lately, i cant seem to enjoy anything anymore, feel sick all the time to which sucks.&nbsp; Ive almost compleatly given up on trying to be human, i mean if some people think im some kinda lesser being alien monster thing, hell why not, not as if it really matters.&nbsp;&nbsp; They know there right... (yes i fitted in something to do with the title!)&nbsp; &nbsp; *Sigh* well anyway at least i can never say i didnt try, i tried hard, i tried really hard... but it doesnt seem to work, ohh great... i dont want to give up but i think my body will give up for me, stupid headaces and projectile (.....)&nbsp;  &nbsp; But one bit of good news over the past few days is that i finished my article that im gonna send off that will hopefully get me the colum job that i wanted woot!&nbsp; Just need folk to neb and prod at it and tell me to send it then it shall be done by Monday hopefully hurrah!&nbsp; The only major problem is though i  Know  i will not get the job but hey ive been shouted at by people to much about that attitude im just gonna say \"Go letter go!\" &nbsp; Grr also just if my own social problems havent pushed and pissed me off enough already since i got up about an hour and a half ago ive been bothered by 3 callers trying to sell me cheap crap like tiles and internet from wanadoo ARRRR!&nbsp; So basicly to annoy them i have so far Laughed manicly down the phone at Tile woman till she hung up.&nbsp; Then i kindly told Wannadoo guy he can shove his incredibly slow internet down his throat.&nbsp; Then losing my creative edge at that the window woman just got told to fuck off. &nbsp; Arr Poems!&nbsp; *Runs from the inevatible poem writing* WACK! Stereotyped! &nbsp; *Sigh* cya later all thx for reading. Ben. &nbsp;  '),(24566,3099605,'16,July,2004','2004-07-16 00:00:00','   Greetings one and all!     &nbsp;   Ok you will have to forgive me if the font is diffrent to usual but blogger has suffered from a massive update in its writing style so any old noob can now make a blog look cool... dam.&nbsp; Ohh well sounds a bit bitter but nevermind, i hope you all had as good a day as i did, as Me, Bald, Jenks and Joe set out once again for Hermits Cave!     &nbsp;   Woo i apologise in advance that this blog is gonna be a bit image heavy, so the journey began at once and we all set off with a song in our hearts and lunch in our bags!     &nbsp;   &nbsp;      Bald and Jenks made sure to keep us well entertained with good loud shoutings of random words such as BEE and the fact they had line marker and matches in there bags!&nbsp; That deffo made for some fun later.     &nbsp;   On we traveled forth like the great adventures we think ourselfs to be but even we were not prepared for what the witch had in store for us, as about half way through our journey we encountered a odd house that had a sheeps bone on its gate post and a very foreboding sign on its gate      &nbsp;  Woo that wasnt very pleasnt of them, and by this point we were all slightly fatigued with our efforts, *plus we thought the witch was trying to curse us, so with our spirits for reasons unknown near to breakage we all had to stop for a massive rest.&nbsp; The only thing that would get us going again after thatw as yelling YAWP! At the top of our lungs for no reason at all.&nbsp; (I dare you to try it, you feel much better afterwards.)  &nbsp;  So on we adventured through errie woods(Top right blair witch symbol eek!).    and we finaly reached the cave, where the Jack Osbourne voice in me that was dying to get out asked     \"Do you have a flask?\"  &nbsp;  So after it turned out none of us had a flask Adam decided he really was  dieing  for a cup of tea  &nbsp;   But we managed to coax him down and celebrated by eating marshmellows however we were assailed by a dragon!     &nbsp;  But it turned out to be Jenks and some line Marker.&nbsp; So after nearly roasting Balds nose and running up the hill without using the staircase we decided it was time to head home to Bartlewood lodge for chippys and coke!&nbsp; Woot!  But after getting there and eating our chips we all just ended up in a vinegar drinking competition, what a way to end the day.  &nbsp;  The most suprising thing was none of us died.  Lol cya later!  Ben.  '),(24567,3099605,'15,July,2004','2004-07-15 00:00:00','   Hello Everybody!  Yes it is i Ben once again hitting blogger so hard with my crap it makes its very digital head spin!  Woo well anyway i hope you all had a good day today on this ~Fine~ summers day.  Summer rain Summer rain!  Lol well anyway im not cracking any musical number open today!    Ohh well on with this show.  I didnt do a lot today really jsut went to go see Spiderman 2, i just suddenly thought to myself, why dont i instead of saying \"i want to go see it\" why dont i just go see it!  Lol and i did and i enjoyed it quite a lot actually.  Its a bit darker than the first one less pussified and more violent, woo yay violence.  Particulary think Doc. oc is kick arse, especialy if you think about it his super villain powers are really revolve around something so lame as 4 big metal arms.  Lol well anyway i got extra nerd points im sure during the movie by spotting that Stan Lee *The creator of Marvel comics* once again made another cameo in it, lol i really have to get out more.  Well so thats another super hero movie ive seen and will no doubt not be the last concidering the vast amount of superhero movies that are coming out at the moment, such as Hellboy, The Cat Woman, The Punisher and X-men 3.  *Sigh* im sure everyone will get really bored of them eventually and its a shame because some of these characters there boosting to movie status have really cool comic book plots and storys but the big screen will really kill them.  Also a lot of films that were orignaly Manga comic books have been capture and are in the process of being ingested by holywood, such as the new I,Robot starring Will Smith and the holywood remake of \'The Ghost In The Shell.\'  But thats all well and good and they may even be quite cool but with Holywood makeing all this money they are in danger of losing the cult following and the interesting individual storys of these heros just to make them more mainstream and believable.  Its just a shame in years to come kids are gonan remember the big heros such as Spider-Man and the X-men, just because the films had wicked CGI and mainstream plots, but less and less people are gonna remember truly great writing and storys.  Which some people may not appreciate because they see all the pretty pictures in the comics and Manga, but those things do take a lot of writing aswell as drawing.  So ive assembled a bit of a  True  list of great films, that were originaly of comic book creation but people have forgotten because they aint all exlosions and CGI  .The Crow (Based on a Graphic novel by James O\'bar) .The League Of Extraordinary Gentlemen (Comic Book by Allen Moore) .Spawn (Comic Book Written by Brian Holguin)  I strongly suggest u check these 3 films out if you havent already.  Hehe well thats my little witter for the under represented writers. Thx for listening Ben.  '),(24568,3099605,'14,July,2004','2004-07-14 00:00:00','   Hello everyone.  Woo well today im really quite tired so i havent had much time to concentrate on the angtsy side of life.  (If someone even trys to whistle python style we will have to get a record contract!)  Lol well anyway, i hope your day was ever loving amazingness, as my day was how can i put it, very long, very boring and at times painful.  Yes thats right everyone Ben visited the orthadontist today *JOY!*  Woo well i hate going there especialy as my orthdontist, in the nicest way possible is a pathetic, tard with the skill of orthdontic work on comparison to a Llama on LSD.    So it starts off plain enough, me having to tell Captain tard that i need to have my brace off by September due to after that im not gonna be able to make any appointments on his stupid time ever again, (plus ive had this fucking thing on the best part of five years, ive forgotten what its like to not have metal in my mouth!)  But of course captain tard decides it cant be taken from me till December!  Then he proceeds his usual tightning of the brace technique, which involves him inserting tighter wires, as quick as he can into my face ramming the razor sharp end pieces as far as he can into my gum before he remembers he forgot to cut the fuckers down!    HE HAS DONE THIS EVERY TIME SINCE I STARTED SEEING HIM!  ARRRRR!  I swear next time just give me the Llama on LSD.  But anyway time for a short interlude!     Woo back, lol well now that i am home and my brace is once again starting to push so hard into my teeth that it has blanked all emotions except for slow dull aching physical pain i may have to crack open the Laudanum on my late night writing spree\'s YAY i get to be like Samuel Talyor Coleridge (He wrote Rhyme of the ancient mariner and all the other famously drunk or stoned english writers and poets!  Well i would if Laudanum wasnt compleate 110% illegal now!    Yup well anyway in my continuation of rewatching every episode of Buffy ever made on sky one today saw the second part of the episode with the Judge in yay!  I love the series 2 through to 4 the most because Oz and Willow are together yay for them!  Lol yes if u didnt already know ive always loved watching Buffy ever since my sister (yup she\'s responsible for a lot) introduced me to it.  So anyway im kinda happy in a physical sense due to i just saw one of the greatest episodes of Buffy ever woot, and one of the best quotes to!  Willow: Remember Buffy Carpe Dium, u taught me that once. Buffy:  What Fish Of The Day!  Im sorry that just breaks me. Lol well i guess you could say im a bit hyper tonight, well i am because if i wasnt id probably reject everyone and sit hunched up a corner rocking and mumbling.  Lol well anyway thats my lot, bubye and thx for reading Love Ben.  '),(24569,3099605,'13,July,2004','2004-07-13 00:00:00','   Hello everyone.  Woo well err *cough* i think after my crazy blogging skills last nigth i owe you all a apology for having to read it.  All of it was true that i said last night though probably blasted out of context due to the sheer weight of shit that had been builiding up for a while.  Ohh well i do feel much better now and i might even publish up some of my poems to celebrate... but thinking about it i dont want to annoy you still more!  Hehe so err yeah today... hmm whats to say about today, i have so far done fuck all apart from had a few games of Tournie with Mr Bald, lol probably if he ever gets back on MSN might shout at him to come round and bring my albums with me.  Lol so in a way its quite shocking that i wrote yesterdays post without even being under the infulence of MUSE... weird.  Still feeling a bit sick today and if i think about it still feel a bit down, but hey the less i think about it the better it gets, woo childish approach is also the best.  Err yeah well i think my PC is feeling a bit poorly at the moment just seems to be telling me things excist that actually dont which is quite annoying.  But hey as long as i can blog, chat and listen to heavy metal music i should be ok.  Woo havent done one of these for a while, LIST TIME!   Stuff thats Kept me, Awake, Alive, Reasonbly happy, and un comatosed due to lack of sleep over the past few days!  .Seeing Shrek 2 yesterday  (I am Puss... In boots!) .Being shouted at by my mates after de railing myself yesterday (quite funny in heinsight) .Juice .UCI Nachos .Finaly Being Able to sleep .My mp3 collection! .U guys! (woot for you)  So err yeah im probably back to normal now, err i dunno actually, might have to bang my head just that little bit harder next time.  Hehe cya Later Ben  '),(24570,3099605,'12,July,2004','2004-07-12 00:00:00','   Hello.  Got a Angsty late night blog for you, i dont think ive felt this low really in quite a long while, sure when im away somewere im away from the pain, running away and ignoring it, only concentrating on the path in front of me.  But when i come back to seek the remedy of something i truly want, all i get is echos of my own voice.  Ive been ignored before this i know but not like this, this cuts deeper than that, a hollow room with no reply to my call but the sound of my own voice, my love, and my emotion rejected and left to die on the cold harsh cobbles of my own bitter mind.  I just feel sick all the time now, weakend, my mind cant even think straight anymore its just bent and broken by something that i have absoultly no control over.  It just annoys me so much this constant desperate attempt i make to get through to people and the bitter crunch as i once again hit that wall, and that wall moves ever closer to me as i run out of options and everything seems to be dragged further away from me.  I really wish i hadnt started trying because i am determined to win out even some minor victory even though i quite frankly know that this is not possible for me because of who i am what my ideas are and how my brain works.  Any attempt to make others feel good about themselfs does not work, they act as though my words mean nothing and they could spend the night talking about themselfs but to me all is recived is nothingness.  I try so hard yet fail unrewarded and pointlessly lost once again.  I had a falling out with my dad the other day, it was over some trivial and unimportant matter, but one of his words did hit me were it hurts and it still reflects today everytime i try to make contact with these people and talk to them like the friends i judge them to be.  \"You have no opinion and nobody cares what you think.\"  Now i know he said this in anger and he doesnt believe it... but i feel a lot of people do.  Im made to feel inhuman by some people, i mean waht is so despretly wrong with me that i am forced to feel this way? Am i not a person same as them!  What gives them the right to judge me, i never judged them, my only crime is loving them and treating them with the respect they deserve.  My words fall on dead ground to some people, my emotions wither and perish for i feel now more than ever they are useless.  Tonight i feel unhuman.  Thankyou for listening Ben  '),(24571,3099605,'11,July,2004','2004-07-11 00:00:00','   Greetings Mortals.  Well now it is time for the proper blog of the night, on this very fine day of July 12th of the year 2004.  Lol i already think that ive lived this day twice with me staying up so late on MSN chatting away to Amber and Bald... im sooo tired now so im paying for it.  *Sigh* unlike Bald who doesnt need to sleep.  i said \"I cant function without sleep how do you do it?\"  and he just said back \"I dont function anyway\"  or we both said words to that affect which led up to his quote.  Woot im tired and i dont even feel that great.  Woo well anyway Me, Adam and Bald went to go Shrek 2 today, and i personaly thought it was better than the first one, lol i just loved Puss in Boots, and John Cleese as the king now that was hilarious.  I dunno it just seemed more grown up than the first one and more about the other characters and not just his smelly ogreness.  Lol seriously if u havent seen this movie yet i implore you to see it, its really kick arse.    Yeah well apart from that i havent done a awful lot today apart from downloading a shit load more Mp3\'s because i had nothing else to do.  Yup i know i should be writing but my brain today feels like it has suffered a head on collision with a train full of night nurse.  Now thats a image i would actually like to see, i might even paint mash it later for my own enjoyment!  Well that is if i havent slipped into a coma or some kind of a panic attack which i often suffer from.  Ohh well hopefully cya on MSN! Ben.  '),(24572,3099605,'11,July,2004','2004-07-11 00:00:00','   What would your Anime life be like? by  urlLink  hearthlight        Name:           Gender:           Your looks:    Brooding eyes and a trenchcoat      Your best friend:    Friend? You\'re the Dark Loner.      Your powers:    FIRE!      Your beloved:    A sexy vampire.      Your occupation:    Student.      Your ending:    Very, very funny.             Created with the ORIGINAL  urlLink  MemeGen  !       Lol i found this somewere random on the internet, its kick arse!  '),(24573,3099605,'11,July,2004','2004-07-11 00:00:00','   Hello the silent friends of my blog.  Yup today was sunday, and you know what that means, boredom, cold tempretures *its always cold on a sunday i swear* and the serious lack of interesting stuff that happens.  Lol i think the only thing ive got to look forward tonight is Kingdom Hospital yay.  But anyway enough about that, i hope you had a good day today and actually found out that sunday contains a useful or interesting activity.  Woo typical boredom and angst fro me today, when the days are boring i think me, and when i think more my state of mind gets more and more worrying.  Had my Dad bothering me all day wondering why i was sat about the house doing very little.  \"Why arent you doing anything today Ben?\"  \"Err because nobody wants to do anything and i dont fancy doing anything myself.\"  *Sigh* but of course this answer not being good enough i was bothered several more times through the day.  Grrrr8.  Also my Dad asked me about Stairway to Heaven, he had heard the track for the first time in ages the other day and he was wondering if i had it... *sigh* do i have it, lol yeah i have 3 copys as a mater of fact. *funny look from my Dad*    Lol woah have u only jsut realized your sons rather odd.  Grr i dont know what it is my parents have suddenly realized that i look depressed and stuff, and my Mum keeps saying \"Cheer up Ben!\"    Me: \"Make me\"  grr they just dont get it do they.  Well in other nu\'s  tommorow ima going to the 3:30 showing of Shrek 2 at the UCI so if anyone wants to come see that plz get there before 3:30.  Ive been reliably informed that it is a good movie and i wont be dissapointed... woah even me as total movie critic guy wont be dissapointed... shocking.    Lol well anyway speaking of Movies, yesterday i saw one of the best movies ever made, woot yup and that was Dead Poets Society, starring Robin Williams.  Lol now lets face it we all know Robin Williams is a jerk, but very much like Good morning Vietnam, which i also enjoyed he is not his usual jerky twaty self and he actually is prepared to let other people act in this movie.  I wont go into any plot detail or anything other than if you can, watch this movie.  I swear you will take something rewarding from it.  So err Carpe Dium* are the words i give to you, it is rather uplifiting movie in some sense but in others deffo not.  In otherwords its class.  Im quite plzed this week in a movie sense because i havent watch any crap movies, just watched Interview with the Vampire, Spiderman, The nightmare on Elm Street and finaly Dead Poets last night.  Good stuff, one and all.  Yeah well anyway i will be glad to be going out tommorow so i dont get anymore questions of \"Why are you looking so down Ben and all that crap, i mean im not gonan tell them if i havent even told you guys!  Hehe well anyway its my problem i will deal, so anyway Carpe Dium*!  to you all again and err thx for reading!  Ben  *Carpe Dium=Seize The Day  '),(24574,3099605,'10,July,2004','2004-07-10 00:00:00','   Hiya  Yup another Pt.2 blog today, but hey thinking of titles is sometimes really annoying plus id like to think that this title was still relevant today.  Well anyway today was spent very much like yesterday of doing not a awful lot apart from wandering around and talking to folk.  I really im starting to think that what little snaity ive got left in me is ebbing away.    I mean i think to much, thats one major thing.  I seem to spend the majority of my time just thinking, about friends, life, writing.  All that stuff that seems trivial but of course its not because nothing is.  Its just the idea that i really want to find something in my life that is brand new, revalutionary and enlighting.  I guess i search for it in the trivial, for no bigger jesture is than one so small it seems to be a universe entirly entrapt within itself.  Also with me Vagabonding around you get to find the vastly diffrent oppinions of diffrent people and for instance a diffrent take on something you have both witnessed together.  Its interesting doing that because you can get a real and in the end diffinitive view of a event which you could not fully and entirley grasp on your own.  Because without people to disscus and share opinions, views and experinces with you would never get past your own oppinion to respect others.  Lol yeah all this Vagabond wandering is putting things in perspective for me, but i cant help wonder if its the kind of perspective i truly want, you know i cant help but think sometimes im running away from the emotion and the pain of reality.  It just gets a bit to much and you just wanna escape, but you know its still gonna be there, my minds still gonna bother me and tell me that im a coward and im running away from everything just for the sake of a few hours peace.  But i dunno, just seems that the time needs wasteing all it would only make me feel worse.  Sorry todays blogs short, cant seem to get out my words tonight. Ben.  '),(24575,3099605,'09,July,2004','2004-07-09 00:00:00','   Heya everyone  Woo this week seems to have gone uber quick for me i dont know why, and once again i reach friday and ive just come back from karate and im looking back at this week.  Well its deffo nearing the pressure time, and in some aspects i guess its already starting to become a bit clutered and opressive in this place they call my mind.   You know its my story for one thing... i now have about eight weeks to get this story finished, but it aint working that way... im stuck, horrific writers block as always.  I also have 20 days to get these two articles off to have a attempt at getting a writing Job, i aint even started them, its just the fact that i think ive got time to do everything but i havent these 8 weeks and in the case of my Job about 20 days are begining to slip from my grasp, and its begining to grate with me.  it just seems that even though these are supposed to be the best 10 weeks of our lives, i have so far had very few Uber good times, and ive been mostly up to my eyebrows everyday with my story, my blog, my poems, my pictures my shit... flaggy, flaggy by one!  I-E-A-I-A-I-O WHY!   Why the fuck indeed i say.  It seems that everywere i look at the moment i see burnt out people.  People knackered by these crazy shit routines they are put through by people and they have to jump through hoops for others just to get nowere really apart from having there shit devour them late at night.  Yup and it really sucks dont it.  Fucking hoopla im sick to death of Jumping, but i gotta do it.    You know and in this vain attempt to get back something thats truly mine i have to stay up till god knows what fucking time in the morning just to remember im alive and i actually still have a life thats mine and not one that someone else uses constantly to get off on there own sense of constrewd satisfaction. This of course leads to me waking up late in the morning, which even though in some sense of the word is cool but its also not because you realize youve wasted half the day, youve got a million chores to do and all the while meantime in the back of your head the promises you made to yourself are nagging away, but you cant please everyone can you... and of course choosing to help others is the route i take and the route that kicks you so hard you dont know which way to turn next (yeah i might tell you soon)  It just seems like im trapped in some kinda wurling maelstrom of annoyance and juty to others, just to pay for all the major league fun we are gonna have in between the mental pain of college.  Damn it to hell John Lennon has never been so right \"Soon as your born they make you feel small by giving you no time instead of it all... till the pain is so big you feel nothing at all.\"  Thats from a track called \'Working Class Hero\' by John Lennon, if you aint got it download now!  Its some what a anthem of this particular post :P  Hehe read it again while playing it!  Yeah well im done for now, the personal Angst will come at a later date... when its really ready to be shouted.  :P  Thanks for reading Ben.  '),(24576,3099605,'08,July,2004','2004-07-08 00:00:00','   Hello everyone!  Well a random happy blog today going out to everyone.  Its weird because i know if i wanted i could angst about a lot of things tonight, but i just dont wanna think about them at the moment.  I guess my works been keeping me busy, been doing a bit of writing and also mainly trying to boost up over the wall of freewebs shit hosting a website to play in the garden of internet, and so far here it is!   urlLink The Crypt Of Benjamin The Immortal   Yes so if im a very good boy the master of Internet Puppets himself, Bald will attach it to World Of Angst!  so yay hoopla to that!  So some of you people who are very bored at this moment... lol i know thats the only reason you read this, \"What did Ben get up to today on this less than great day!\" Well i will tell you, mostly hacked this website into the net all morning but in the afternoon me and Mr Adam Chapman went for a bit of a adventure round some of the more secret bits and bobs of Spondon.  Woot very cool stuff, wish id bought my camera as we discovered the water resovoir for spondon, and its a very old creepy delabated bomb shelter style place with a amazingly scary resident evil style rusty brown gate with razor wire crudley tied all over it.  It might as well have a sign on it saying \"DANGER T VIRUS BIOHAZARD!\"  Lol sorry all you non computer game fans or people who have no idea about gameing history but anyway twas super cool!  Lol we also discovered our secret village not so secret after all *major embaresment when we came up next to the cricket ground*  lol we suck.  Lol and back to the title of the blog, yup ive discovered over these 9 weeks we have left ima gonna becoem a traveling Vagabond as such, plus this discovery backed up the way everyone seemd to be working but me!  Lol well anyway to be free!  *chinks imaginery shot of JD with ur equaly imaginary one*  Woo well thats about it really, cya on the other side. Ben.  '),(24577,3099605,'07,July,2004','2004-07-07 00:00:00','   Hello again.  *Sigh* well after my beyond shit early morning gamble in which i lost in dramatic style i had to get my hair cut because of this fact.  *Sigh* so here it is prepare to laugh    Yes i know my camera is gloriously shit but you can also see that the hair and to think the face behind it doesnt help matters much either *shudders*  So anyway to cheer myself up i bought a special addition copy of MUSE\'s Origin of Symmetry and i have been listening to that most of the afternoon yay!  I love the origin of symmetry it has super duper kick arse tracks on it including Bliss and Plug in baby, which both truly do own on earth as it is in heaven.  So well feeling dejected after my earliy morning gamble and listening to to much MUSE me and Bald AKA  urlLink Phantom  decided we would go all dejected and go for a man of the road style walk about.  Woo so a  urlLink WOA  event it was indeed so we hammered around spondon at first, laughing at Ben Gillams Nu job.  Ha ha he so cant hack it.  So feeling better that we\'d made fun of another human being we wandered off to ockbrock talking about the usual stuff, including Balds increasing pyroness which is very cool to watch so im sure if u ask him nicely he will send u some truly hilarious video tape.  So after wandering around ockbrook and offending a old person by making a lighter that Bald found in the street explode with a very satisfying pop it was time to go home and listen to more MUSE.  Why the fuck did this mornign have to cock up so badly, all though mine and Balds ockbrook adventure was fun.  Woo so continue to visit us over at  urlLink WOA  plz for more such kick arse film and adventures.   Ohh well so apart from that cool adventure, my day sucked and its all my stupid minds fault.  Woo Rice Crispy Syndrome is kicking in very much tonight to the extent of a Bald Quote  \"I like to think im useless in every way.\"  Cya!  P.s.  Yay Hoopla I noticed i hit my 100th link in this blog.  '),(24578,3099605,'06,July,2004','2004-07-06 00:00:00','   Hello all.  Woo a morning blog, that must mean i had a idea what i was gonna do today,  and i made a bet with my mum last night i said \"If i dont get to do what i wanted to do today, i will get my hair cut.\"  Well my plan just got cannonballed and sank to the bottom of the ocean before a shot was even fired.  So ima gonna get my hair cut.   Bubye Hair.  *sigh* well i guess my decision that i came to left me no other choice... dammit.  Lol my only last wish is the most far out one ever that the moment its cut it starts growing back, like in the vampire chronicals lol.  Id just love to scare the fuck out of the hairdresser that way.  *sigh* i shouldnt have stayed up and watched interview last night.  So under a massive dose of Rice Crispie syndrome (A syndrome i invented in which you feel like a makey box  of rice crispies left on the shelf of a shop that nobody wants to buy and its only goal in lfe is to occupy space)  my hair goes to meet its maker.  Cya later.  '),(24579,3099605,'06,July,2004','2004-07-06 00:00:00','   Heya all  Lol well another late blog tonight, which is kinda annoyign but i aint really bothered i guess it means ive got more to talk about now.  Well anyway  i hope today see\'s you well and youve all been enjoying the sunshine adn the company you keep!  Now the formalitys are out of the way once again, on with the blog.  Today saw me doing the usual activity of just hanging around not doing very much at all.  However we did try to fit ina  interesting town touring spondons parks on somewhat of a frisbee tour.  Lol we also attached Balds sound bug to as many things as possible to make even the most burberry of Chavs listen to such great hits as \'Sweet Child Of Mine\' and \'Innervisio\'  So through its magical gift of being able to turn any flat surface into a music speaker it is now the coolest device on the planet.  Ohh yeah and my writingness, has once again hit a all time low due to the fact that i cant get over this very very dramatic writers block i seem to have going on.  :( dammit.  Also due to the fact that ive been so distracted with everything i do believe im letting it all cave in on my head... which sucks... a lot.  Yeah well and i also discovered i have got absoultly nothing to do tommorow *sigh* so please someone save me from that.  But anyway even if i seem distant and annoyed at the moment its nothing to do with you, im just driving myself insane, so plz just tell me how you feel if your pissed with me or ought.  Well with that said id like to once again thank you all for listening... i love you all and goodnight.  '),(24580,3099605,'05,July,2004','2004-07-05 00:00:00','   Greetings.  Heya all this is my proper blog for tonight and by hell im gonna hopefully give it some of the Angst filled goodness it deserves.  Lately ive been feeling really depressed most of the time, its really starting to get to me the way some people are how they treat you how they refuse to be normal around you and no matter how hard you try you cannot break through to these people.  This is a form of Angst, i feel outcasted sometimes, no replace that with all the time... by certain people.  They make you doubt your own mind with the sheer diffrence of theirs and the way they hide the way they feel and think makes you wonder if its normal to think the way that you do yourself.  I dunno i just feel like im being got to lately, and the funny thing is the people who are doing it to me dont even know they are doing it.  They think that your gonna be alright and tell you how everything  is good and great the sun is shining i did such and such today why dont you get a real job cut your hair, everythings great in my world i would ask how it was in  yours but i dont care to listen!  Some people dont want to listen, thats the sad fact of life, everyone wants to speak but some people do not give a toss about listening.  Ohh no its alright im only there for you all the time but hey it doesnt really matter that you dont listen to me because im still gonna be there.  Woo how fucked up is that.  Angst comes in another form at this moment.  Failure... i resign myself to the simple fact that i probably am one.  I mean im suffering from terrible writers block, i am inept in getting anything to work, to get anyone to listen to me, im just in the background of life screaming, like the noise of a train as someone else walks past it... the train always makes the noise, but you dont have to listen to it because there are more important things going on.    I dunno really i just feel like everything i do at this time in my life is getting me nowere, im just hurled from one day to the next to suffer at the hands of the same people, here the same things, do the same things and witness the same things.  I cant break free of what is not freely given to me.  Its almost like ive been given a puzzel but sevral people have stolen the pieces and hidden them away far beyond my reach, and they all give instructions to find them but non in a language i understand.  People would benifit a lot more from each other if we madem ourselfs a lot more clearer... but the problem is everyones to shy to so someone has to start.  In lies the problem.  '),(24581,3099605,'04,July,2004','2004-07-04 00:00:00','   Greetings!  Woo today has been a very odd day of trawling round the internet and makeing photoshop pictures of rockstars as animals!  So in otherwords has been a weird day... but very cool due to the fact i discovered something truly awsome!  I discovered 6 flash Animations YES THATS SIX!, directed and written by the master himself Tim Burton woot!  About the adventures of Stainboy!  Double woot!  So without further ado you may all now revel in the creepy humour of the master himself now in flash form!     urlLink i still cant believe there is six!   You are the master Tim, no doubt.  Lol well in other new\'s here are the piccys i made today, the first is of Alexander Kapranos (the lead in franz Ferdinand) as a dog, singing \"take me out\" which can be found  urlLink Here   Plus the second is James Hetfield (Of Metallica Fame) Takeing \"Of Wolf And Man\" a little to seriously and this can be found  urlLink Here   Lol so thats what ive been up to this afternoon, as well as this very lame joke   urlLink Granny Knot   Lol well enjoy my heavy linkage, and there will be a proper blog later!   \"I STILL CANT BELIEVE THERE IS SIX!\"   Ben.  '),(24582,3099605,'04,July,2004','2004-07-04 00:00:00','   Greetings one and all!  Well today as most days appear to be at the moment, was boring in my view, i did the usual lark of writing more story, yay, reading some more Edgar Allen Poe, yay and chatted to some folk on MSN.  However i did discover one of the coolest features of MSN today, Audio conversation!  Woot this is really cool for me especially at the moment, as im getting a lot of pain in the ends of my fingers, at the moment its rather strange really.  Ohh well audio conversation owns because you can witter and witter, and not only do you here people for real which is nice, u get a sense of their real emotions which normal MSN severely lacks.  Lol well anyway i thought that was really cool.  In fact the only cool thing that happened all day... apart from the rain coming down like stair rods, now that was really cool.  Also managed to type up my poem i wrote about 4 days ago now, but that doesnt really matter, just glad i got it to switch context without being ruined.  Lol well anyway bit of angst as always today but it was just the usual stuff, that nobody can really console you of... you know you just have to deal with it.  Lol still havenâ€™t done enough writing i fear, my protagonist the great and Mighty Jez, seems gloriously still far from writing his own misadventures, and my short story about body snatching in the 19th century is still non existent.  *sigh* Ohh well maybe i will get some proper work done tomorrow.  Lol but if anyone wants to have fun *looks around hopefully*  plz let me know. Hehe well anyway cya tommorow.  Ben.  '),(24583,3099605,'03,July,2004','2004-07-03 00:00:00','   Greetings The Folk Who Live In My Blog!  Well today was rather shall i dare say it... boring.  Yup 110% nothing better to do than listen to MUSE and watch Takishi\'s castle and flick on kerrnag to go \"Ohh cool Anime style Music vid, shame about the wank song)  Aww yes those shit heads linkin park stealing cool ideas thast good bands could have used!    Lol well anyway i do believe that video is the coolest thing about lincon park... in fact the only cool thing about lincon park... no actually there is another, or am i alone in thinking the Hybrid Nazi soldier was uber cool but a bit out of place on a album produced by Warner bro\'s.  Hmm ohh yeah, and the awsome flash parody of Charlie Brown and the peanuts crew doing in the end, which is buried in the middle of this awsome flash cartoon which can be located  urlLink here   Woot so anyway enough of my love for the INA cartoons, in all they glory of blood, guts and hillarious rip takes of all the bands i hate \"I didnt vote for Suk 41 because they really suck!\"    ON WITH THE FREAK SHOW!  Yeah well anyway i dont think im getting enough sleep lately, it wouldnt normaly worry me accept for im starting to have really fucked up dreams again, i mean take todays adventure through the windshield  basicly i was in a cave looking out on a tunnel pooring water into big conch shells when i saw the grim reaper, and he came up to me, and *Shazam* time to wake up BEN!  It was actually quite scary which is weird for me because normaly i dont get nightmares that scare me, just u know bother me.  Lol well anyway its my own fault im getting them, basicly i stay up till 3:00am most days writeing and reading the works of Edgar Allen Poe.  *shudders* i so need to get one of those things they call \"A ordinary human beings mind\"  However i did right what i think is one of my dudeiest pieces of poetry the other night... which i may share with you one day... if i can be ever bothered to ram it into my Pc using my ol faithful keyboard.  Lol anyway so i really think im pushing myself to hard at the moment but i really dont know why... its almost like im looking for a answer to something in art, when i for hell as one know art only brings more questions.  *Sigh*  I really wanna write a short story aswell now, ive always wanted to do one on bodysnatchers, but ive never got up to it.  Plus with my 10 week story still very much on the smaller side, i really have to stop this whole mental pitchforking.  So err yeah gonna put this to the test now... if you wanna see my Poem i wrote the other night, err plz email me or tell me or something, otherwise i jsut wont bother putting it on the site.  Lol well anyways if you see me reading or shouting about writing mentaly *or physicly if you are near me by any chance* force me to sit down and ram a cookie in my face and Just freaking relax ok!  Ohh OHH!  Now my fave webcomic in the whole world is back on line (in fact the only webcomic i like in the whole world) i can finaly show u the closest parady i got to what my match makeing skills were like at Bill!  Woot Im the Guy with blonde hair for this Story and Ben is the guy with Purple!    urlLink WOO Stubble!   Lol thats closest i can get to my feelings in that final frame... witty shit.  So anyway if u see me torturing myself, either laugh, or knock me out! Bubye  Luv Ben  '),(24584,3099605,'02,July,2004','2004-07-02 00:00:00','   Greetings Mortals!  Heya all i hope youve all had a good day today and not had a list of chores as long as your arm!  Woo i know i have... the list of chores that is *sigh* hoovering, ironing, and tidying my room... ohh well i guess its gotta be done, i know have nicley hoovered rugs, non creased cloths, and brand spankingly tidy room.  Woot.  Err i guess.  Yeah well in other new\'s i pushed ahead with my story tonight which im quite proud of yay! Lol i dont know what its actually like but i will give you a snippet in a bit if you like. Yeah well anyway next week should be a fun week for those who read my blog im hopeing to get you all to have a time with me, using my rather rubbishy organizing skills to have a laugh and do some cool stuff!  So err woot to that and i will talk to you all about it next week.  Yeah well tonights the night of the prom and i hope everyone who is going enjoys themselfs, and they all have a good laugh in their expensive clober!  As for those non prom goers!  \"YAY!\"  have yourself a drink of juice and some cookies and enjoy yourself safely in the knowledge you will pass your exams and you will have a great 9 weeks!  Cheers!  Lol well im hopefully sure i will pass my exams, and i really hope i do because Bill is such a great place plus i get to do Tai-Chi which is quite frankly kick arse!  Lol plus if i get as many permant memorys as often as i did at the induction ima gonna have a ball!  I mean who can forget from the folk who were there the whole back of the bus singing along to SOAD, or us all hanging in the cafe for hours on end!   Lol plus the personal memorys such as having my caricture done in Comm Studies *RAWK* lol and me and Ben in Biology \"Urea doesnt go there surely!\"  Woot, good times, and we are all gonna have many more, all of us, even the folk who aint gonna go to Bill we will still have some kick arse times.  Lol its like my Mum said (very insightful of her i thought)    \"If you dont have a good time when your alive, when you gonna have it?\"  :P very nice.  Lol well for all u lit junkies, here it is my piece of crap taster, i mean story taster!                                ~It Has No Title Yet!~  But that waiting was the killer, listening to Ratfinks heavy breathing and his greasy eyes watching over his thick glasses, Jez swore he almost shuddered and quickly looked outside.  But what he saw nearly made him fall off his chair. 	Absolute blackness.  Jez gasped and quickly turned back to Ratfink his eyes must have seemed to bulge out of his head â€œRatfink look outside!â€  He said in a horrified half whisper as Ratfink disturbed his greasy train of thought to look.   	â€œBut Jezzzzzz.â€   Ratfinks voice always trailed in his mind â€œIts just normal Jezzzzzz!â€  He started to titter his little creepy laugh like a large furball would gnaw at cheese.  (Lol sorry i suck at choosing segmants Jez is the protaganist and Ratfink... is a twat!)  Lol well anyway cya later! Ben xxx  '),(24585,3099605,'01,July,2004','2004-07-01 00:00:00','   Heya All.  I hope youve all had a cool day today, and hopefully enjoyed a lot more of it than me, (didnt wake up till about 1:15)  Sigh i hate wakeing up late and hate going to bed early so either way its annoying.  Lol well i blame my diary for keeping me up till 2:17 last night, i guess i just had a lot to write.  I just kinda talked about all of the stuff that was angst tasticing me at the moment, and it was one fucking big list... just so annoying that i feel this way, i mean it probably just sounds so fucking annoying.    ~30 years later~  \"Hey so what you remember about Ben Jeffrey?\" \"I remember how he used to winge about every little thing!\"  ~2004 right now~ *sigh*  Lol well anyway this list kinda depressed the hell out of me so i woke up with a far from positive attitude today err yay and ive mostly been listening to depressing music *MUSE and Snowpatrol if you must know*  *Sigh* i dunno i just really wonder if there is a point to any of this sometimes, i mean what is the point in thinking in terms of emotion, i just feel that being emotionless has gotta be useful because so often you realize that people you care about dont give a fuck about you, you might aswell stop having to feel pain.  You know just to forget your own mind excists the way other people always seem to.  I mean people hurt others all the time but they dont realize that they are doing it, but when it comes down to them getting hurt they are all \"woah, why the fuck did that happen to me!\"  Its karma and it rapes us all of everything if you let it.  Im starting to wonder if karma really does work both ways, im more starting to think that karma is just the vindictive mind of the universe, a force which only use is to cause pain, and peoples good deeds fuel its ability to cause misery.  Because surely thats what the whole world is run off, everything lives off the back of something else and emotions like anger and sorrow are the only things that remind us that we are alive.  For instance you dont when you are happy realize \"ohh isnt it great and time will last for 80 so years and then we will die.\"  No you live forever when you happy.  When your upset or angry is the time when you realize just how mortal you are.  But i dont want others to feel misrable, its not right, people should not feel sad just because others are, you have to deny misery of company, because it lives off it.  Yeah well im done now.  Thanks for listening as always. Ben.  '),(24586,3099605,'05,August,2004','2004-08-05 00:00:00','   Greetings mortals  Well what can i say, it has been far to hot today so i spent most of it cowering inside hopeing i dotn go all crispy vampire style woo.  But for all you normal people who can go outside without the tendancy of dieing when you do i hope you enjoyed it but i really really really want a thunderstorm tonight.  For the past two thursdays there has been a thunderstorm and i want one again tonight!  Bring on thunder thursday thats what i say!  Well anyway ive just finished watching my fave episode of buffy ever, (i cant believe i always forget about it)  its called \'The wish\' and the cool thing is probably even non Buffy fans would like it because the majority of it is set in a alternate reality were Buffy never came to Sunnydale, and Willow and Xander are vamps YAY!  But however it is a rather sad episode *sniff* especialy when Oz kills Willow.  Why the hell did Oz have to leave  he was the best!  *Sigh* well anyway im still bored and pissed off with life at the moment but this artsy filled day has taken my mind off it somewhat anyway.  Hopefully gonna write some poems or something later for future website updates... but i dunno.  Well anyway happy thunder thursday and err bye!  Love Ben  '),(24587,3099605,'04,August,2004','2004-08-04 00:00:00','   Heya everyone  Well today has been a day much like any other one for me at the moment, alonesy and boring... woo yay hoopla.  I also as of so far have done nothing creative today which winds me up a little but what winds me up more is that ive got nobody to talk to most of the time!  ARR!  Where you all gone :(  Ohh well anyway still havent heard about the job for the Derby evening telegraph (curse their oily hides) and with everyone shazaming off to hotter nicer places (or just the furthest place away from me) i think im going quite mad.  Lol well not as mad as our next door neighbour, she is always killing our plants in the garden and this afternoon she said that the window cleaners could climbo on our roof!  (Our house is semi detached)  She\'s barking!  Plus her and her little yappy much and her annoying screaming chav grandchildren make me want to kill!  Yup thats another problem all my friends have dissapeared and it seems that in their place there has been a massive rise in the number of chavs hanging around near and on my street, with their stupid burb and gold chains... who the fuck do they think they are.  Plus to prove how stupid they are, they are actually scared of a fat labradour thats nearly 7 years old and cant run.   Pff chavs.  Also if blogger could grant me one wish today it would be, bring my friends back, get us free booze and strike all the chavs down with a pox on a rather sunny day.  If it could do that for me it would be swell.  Cheers for listening Ben  '),(24588,3099605,'03,August,2004','2004-08-03 00:00:00','   Greetings.  Hey everyone, well lets just say i hope you all had a better day than me, yup its just been one of those days where you wish you\'d slept all day and have absolutly no memory of it, almost like missing time.  Well what can i say, it was crap because of sevral reasons really.  I once again had to much time to think about myself and how fucked up i seem to be able to make everything.  You know i really do wonder sometimes how can i possibly be this wrong in other peoples eyes, i mean am i not a human being just like them, obviously not.  Obviously there is some law, or some hidden code in life which means everyone can treat me just how they like and i really dont have the right to any feelings or beliefs.  Just so fucking pisses me off the way some people treat each other, i mean can they not see what bastards and bastardets they are being to each other?  I really dont think so anymore.  Another thing thats bothering me at the moment is my hair.  I mean as you all know i used to have it really long and then i got it cut quite brutal, but the weird thing is its growing back really fast... i mean it couldnt have been a month ago since i had it short, now my back is touching my shoulders and my fringe is in my eyes again... so weird.  I dont want to have it cut but no doubt soon parental pressure will build up again and i will have people heckling me in the street again... probably more so than usual.  Because im sure plenty of you know what kind of shit i get shouted at me... hmm chavs, well probably more than just chavs.  But yeah talking of parents, ive been told i get the house to myself while they go on holiday for two days next week, and get this they said if i have any form of a party they will disown me.  The freaky thing is they aint kidding, my dad disowned my sister when she was like 16 for 6 months or something, i was quite young at the time but i do remember some serious shit went down. *sigh* they are such fucking narks, they get to go on their shit little dirty weekend or whatever the fuck it is and i get nothing but me and Peter (my dog) alone in the house for two days.  Another shitty thing is that i still havent heard from the Derby evening telegraph, they said id hear from them like now!  grr.  I mean why cant they just tell me i suck and fuck off again out of my life.  But nobody on this earth seems to have the true power to tell someone \"hey just fuck off.\"  Why is that? I mean if you dont like or want someone associated with you anymore just tell them, just dont let people hang on, cut the strings.  So now in this influential month that is August it seems that life, time and excistance has just seemed to blur out for me.  *sigh*  Thanks for listening Ben.  '),(24589,3099605,'02,August,2004','2004-08-02 00:00:00','   Greetings!  I hope youve all had a jolly nice day today and got up to lots of nice things, and also to the folks going on holiday i do hope you all have a super cool time and enjoy yourselfs!  So anyway whats been happening over on the angsty side of town i hear none of you shout, ohh well ima gonna tell you anyway.  Well the answer is not a lot really, plus its been quite sunny, so ive mainly been indoors trying not to get burnt and working on stuff for my website update (woo boringness personified) lol so anyway ive started work on a poetry series called \'The emotion museum\'  woo i tinks its a rather catchy title, well anyway im hopeing to put them on the site soon, as well as some funny pictures i shopped to get the entertainment junkeys of \'The crypt of Benjamin The Immortal\' envolved.  But please any suggestions or if you want any of your own stuff on the site, and i cant stress enough its gotta be your own please just let me know and i will be overwhelmingly happy to put it in for you.  Well in other news this week, it is now august so any day now i will be hearing from the Derby Evening Telegraph telling me wether i have the colum job i have been hunting for or not!  So mesa a bit worried about that!  Hehe well we gonna have to organize something to take our minds off it because face it August is gonna be a really big month for us all with the exams and everything so anyone who wants to invite me to get drunk with them or anything im more than happy to lend my presence :P  Hehe well anyway hope to talk to all of you on MSN later, so err until then i will leave you with this website  urlLink here  were thee can listen to a song by the band panic cell which ima gonna go see in September!  w/y/h!  Love Ben.  '),(24590,3099605,'01,August,2004','2004-08-01 00:00:00','   Greetings one and all!  Smell the air, tis a afternoon blog so it be!  Woo yay and hoopla, plus its rather nice outside, which is kinda cool and diffrent!  Well ive decided this afternoon to stop me thinking im once again gonna kick out some art and try to get some insperation going on.  Woo yay and a hoopla to that, well anyway thanks to the very nice lad known as Joe Robson (Look at his site  urlLink NOW! )  He introduced me to the wonderful world of the Monochrome bitmap as we swapped Artsy internet tales, and talk of B3ta.  So shazaming here onto the blog is my Mothman almost signature picture, now in monochrome form enjoy!    I tink its rather swish.  Lol well anyway im still working despretly for ideas for my website so err plz help me! Plz plz plz!  All ive got so far is a couple of pictures and a rather crap poem, need ideas for good stuff, please be my MUSE\'s!  ARRRRRRRRRRR!  Lol err yeah i could really use some help as all and hey if any of you want any artwork or storys or any of your great stuff on my website i will be MORE than happy to put it on, so please if you wanna do that please do i will love you even more than i do now! lol.  Well anyway im back to working Cyaz Ben  '),(24591,4073027,'29,July,2004','2004-07-29 00:00:00','   it\'s me.&nbsp; i am at work, bored.&nbsp; rather be bored than be too busy here.&nbsp; too busy means too stressful.&nbsp;        &nbsp;    started atkins on monday.&nbsp; was doing good until last night, had pancakes....oops.&nbsp; starting over today.        &nbsp;    i haven\'t been to church in what seems like forever.&nbsp; i feel bad about it.&nbsp; i can tell when i haven\'t gone or read my bible in a while, my attitude reflects it.&nbsp; i need to go this sunday.&nbsp; i have been working sunday mornings (did i mention i work in airline operations - 24/7 operation) so i get lazy when i get home and don\'t go on sunday nights.&nbsp; i am excited though cuz my church is starting a new service on saturday nights.&nbsp; it\'s suppose to be \"cooler\" more geared towards the younger crowd. funny cuz my church is pretty edgy anyway, so getting edgier will be intrigueing....        &nbsp;    ok, so why do people just naturally feel that they are superior to others.&nbsp; ESPECIALLY when they are clearly not.&nbsp; i work with this guy who just talks down to everyone around him, especially women.&nbsp; seriously i think he thinks we are all trailor trash with an i.q . of 10.&nbsp; we are all management level here so with a few exceptions (and funny...those exceptions are boys) we are all pretty intelligent and most of us have college degrees......i have to work with him all week.&nbsp; so get to put up with his superiority complex.&nbsp;maybe he is trying to compensate for another area of his life that he is lacking in.....just sayin...............wow!&nbsp; i just heard him talk down to his boss....now he\'s really daring.....        &nbsp;    so, going to eat my protein breakfast.&nbsp; I want to lose 29 pounds before 10/26.&nbsp; reason is a secret....will reveal that later...........        &nbsp;    oh, this democratic convention itself is not getting on my nerves. as a matter of fact it\'s helping in my decision making for this upcoming election.&nbsp; i\'m registerd democrat but have been pretty much independant for that last decade or so.&nbsp;&nbsp;       &nbsp;    I am a confused girl in search (again) for what I really believe in.&nbsp; I know I believe in absolute morals, I believe in feeding the hungry, protecting children, in education......But I don\'t believe in making the rich richer, feeding the homeless\' drug habits....so, where is that middle line?????&nbsp; I used to be so sure of what I believed in.&nbsp; Was a member of several organizations, an active member.....But now I question those things....I am older, I am a parent.&nbsp; things change and so do my thoughts......   '),(24592,3700686,'29,June,2004','2004-06-29 00:00:00','   A spectacular action flick, with Killer Whales hunting humans, high-speed hovercraft chases and shootouts. By his own admission, the authors attempt at a Hollywood-style action movie. The hero (Call Sign Scarecrow) is the leader of a military unit that specialises in securing and protecting sites of turmoil. The reason for the strange name - scars that run down from the top of his eyebrows all the way beneath his eyes, along the eyelids. His eyes were cut-out as a mode of torture, and were restored by the miracles of Science, the eye-damage being mostly superficial. He always wears reflective glares to mask them.  The story begins with a research team in Antartica discovering something resembling an alien spaceship in a large underwater recess of the iceberg that holds their ice-station. The news gets out and Scarecrow\'s team sets out to secure the area. We realise that the dynamics of isolated areas are completely different from that of the civilised world, where even the most friendly nations have only their best interests in mind. And they will go to any lengths to achive it. For that matter, even your own country does not trust you, and would rather get rid of you than have the secret made public.  There are many heart-stopping moments for the hero and his friends figuratively, and one literally, during the many twists and turns the story takes, before it is brought to a satisfactory conclusion. A good read if all you want for your money is thrills and don\'t care about too much violence and mayhem.  '),(24593,3700686,'21,June,2004','2004-06-21 00:00:00','   A depressing and dark view of life at IIT Delhi, and how the \"system\" screws you if you try to meddle with it. Net-net, continue with whatever you were doing prior to getting into the IITs, for the next 4 years and you will do fine. Once you get adventurous, and try to have fun, you end up paying for it with low GPAs. The first sem defines your persona for the next 3 and a half years. That also includes the friends you make or don\'t make and what the Profs think of you.....  Ends up rather well, with the forgiveness element of the \"system\" helping to salvage the careers of the maverics Ryan, Hari and Alok. The book is a narration of events at the IIT Delhi campus by Hari, with a chapter each by the others.  I found Hari\'s character rather remorseless and insensitive. Ryan is a guy who lives for his friends, and Alok is a whiner who has no choice but to live for his family.  I guess if this gets into the hands of wannabe IITians, they just may think twice about joining .....  '),(24594,3700686,'20,July,2004','2004-07-20 00:00:00','   This is a very informative book for aspiring MBAs-A chronicle of life during the first year at Stanford Business School. Peter is a â€˜Poetâ€™, and like some of his batch-mates, is uninitiated into the complex world of number crunching and usage of dense formulas. As a result he has to enroll for a pre-MBA Math course, and here begins the journey through the Hell that is MBA.  Peter was previously employed at the Whitehouse as a speechwriter for Bush (VP at the time) and then Reagan. Completely out of depth at Stanford, he has to come to terms with a punishing schedule, working till late at night, completing assignments and preparing for lectures. He has also come to Stanford with the noble purpose of learning for the sake of knowledge, but finds himself increasingly swept towards exam-oriented studies.  The book also gives us insight into the workings of Ivy League institutes. We identify with the misgivings and apprehensions shared by students on the heavy investment they have made in terms of their finances as well as 2 years of their lives in favour of their careers. Though they are students at one of the top business schools in the US, will they survive the ordeal of the course to get lucrative offers from the high-paying corporate firms, or will they find that they have made a terrible mistake by joining this school?  The personal lives of the students are affected, and frequent mood-swings and irritation are common at the beginning. Grades are all-important and the fact that few of the lecturers are young, inexperienced PhDs (some of them with strange foreign accents) makes matters worse. Complaints about instructors fly from the tense students as exams approach. The management appears apathetic, increasing the studentsâ€™ frustration. All of this would ring true at any institute.  The Institute has problems of its own. Being on the West Coast, far away from the corporate hub in the East, there is a constant migration of talent. Why would anyone want to stay in the West coast when they can earn much higher remuneration if they migrate towards Wall Street and fulfill dreams of making it big? As a result the institute has to pick up promising PhD students and train them in the art of teaching. A slow process. Things reach an abysmal level when Stanford ratings fall to a low of ninth place, from second.  Peter also delves into the emotional insecurities of female students. Though they would not freely admit it in public, they are uncertain about how to balance career and family life, about finding partners who are not intimidated by their degree and so on.  Yet life goes on, and Peter and his batch of 333 students somehow manage to plough their way through the course, and towards the end of the first year, their courses become less demanding. This allows them to prepare for what Peter reluctantly admits as the reason for which they were there in the first place â€“ Jobs. Time is spent preparing impressive resumes, discussing with others on each aspect of these, and then revising them. The company representatives come in to present a rosy picture of life as an employee. At the end, the students decide what they would like to take up, and, after the usual interviews, are awarded with summer internships at reputed firms.  Peter himself lands up a job at an investment-banking firm. His relief is enormous, and his apprehensions about whether he would fit into the world of money and business are temporarily cast aside. Yet, he finds that this world is not for him (as do the people who hired him).  The 2nd year of the MBA course is not discussed in much detail, since there is nothing new for the students, and each of them becomes engrossed in their own specialized subjects. At the end of the 2nd year there are 3 categories of students â€“ those who land up the jobs of their dreams, those who do not manage to do it, but still end up with decent jobs, and those who have to rough it out on their own. As you guessed, Peter belongs to the 3rd category. His search leads him to meet eminent people of the business and technology world, like Rupert Murdoch and Steven Jobs. He opts for a career in Television Journalism, and joins Murdochâ€™s firm. Then comes the depression of the early 90s. With job-cuts being the order of the day, MBAs start getting laid-off and the Stanford students, Peter included, are no exception.  Peterâ€™s conclusion is that though a Stanford MBA does little to mould a person into a good, or indeed, successful manager, it does have its merits. The respect it commands in the industry is enormous. In Peterâ€™s own words, the likes of Murdoch and Jobs would not cast a second glance at his Resume, had he not been a Stanford alumni. Stanford also gives students the confidence to face lifeâ€™s situations with dignity and confidence (If I could get through Stanford MBA School, I can get through this situation!).  A very readable book, and recommended for those who want to know what theyâ€™re getting into, by enrolling for the MBA at reputed business schools.  '),(24595,3700686,'12,August,2004','2004-08-12 00:00:00','   Rollicking story of a boating trip that three young men and their dog make. Though the book was written around a hundred years ago, one can identify with it even now! The 3 friends- George, Harry and the author (referred to as \'J\' by his friends) take up this exercise for a change from their mundane lives. Of course, they also have fears of age catching up on them and fear that they suffer from diseases with complicated-sounding names! The author\'s dog Montmorency, always willing to be a part of the excitement also joins in.  The book is a chronicle of their travels upriver interspersed with the author\'s musings on various topics ranging from the silly, like how entertaining and exasperating it is to observe women towing boats, to events with historical significance like signing of the Magna Charta.  The humour is decidedly very British; the kind that one comes across in books by Wodehouse. Both Jerome and Wodehouse have perfected the art of describing incidents with dry wit, sticking to the British tradition of maintaining the \'Stiff Upper Lip\' in doing so.  To be read by one and all! It takes you out of this world into one that is filled with fun and laughter! In Jerome\'s own words - the world persists in recognising him as the author of \"Three Men in a Boat\", though he has written other books that have far more to offer in terms of literary significance and humour in his opinion.  '),(24596,4075287,'29,July,2004','2004-07-29 00:00:00','   I have to back to school on the 5th. I will be in 10th grade this year and am really looking forward to it.&nbsp; It is going to be great to see all my old friends again.&nbsp;I really missed them this summer. Sometimes it seems as thought I am a loner. Some times I will be with a group and feel like I am not one of them. Some times&nbsp;I just stand there and laugh when they tell a joke or act like I am interested. Some times every thing that I say just comes out wrong. I will try to make a joke and just end up looking like and idiot. But some times thing just seem... Normal. My jokes go off good, people laugh, and I seem to fit in. Some times I wish that I could be more like some of my friends. I can remember sitting back stage watching my friends go on stage and&nbsp;listening to them make people laugh and wondering why I can\'t do that.&nbsp; For some reason it seems as though this year is going ot be diffrent. I hope that this will be the year that I prove myself to all my friends. '),(24597,4075287,'29,July,2004','2004-07-29 00:00:00','   I have finished all the I have to do for the summer and am now waiting for school to start back up. I did several.... Fun things this summer. First of all I worked on Pinocchio. The show went well and was very successful but I hate working with little kids. And of course when every time one of them messed up it was the older people that got blamed. But I it is always hard working with kids but over all it was fun. I am really going to miss some of the people that I meet while working on this play. So of them go to different schools so I don\'t see them much.   I am really looking forward to starting back to school.... I never thought I would say that... I am looking forward to seeing some of the people I haven\'t seen in a long time. I am also looking forward to Adv Theater with Mr. Scott. Plus when school starts back we start on the fall play. This year we are doing The Princess Bride. It is being scripted by two of my friends. I am hoping to make this the play where I change over from Tech Crew to acting.&nbsp; I want to get the part of the giant.   I am still playing bass and am getting better. I now have good number of songs I can play. The bad thing is that the guitarist I played with graduated last year. I could play with Ben but.... He has no rhythm.   For the pass 16 years our school has been the 2AA High School Chess Champs, but I think that this is the year the we will stop winning. It is not that the other team are so much better than all of us, it is just that we don\'t have as many good players as we had last year. As I see it now our top four goes something like: Ben, Me, Will K., and Bradley R. Bradley... The guy that will hit himself on command. I now he only does because we tell him to but.... He is still the one who does it. I never thought He would be in our top 4.   I took&nbsp; a trip down to Walterboro, Sc. Yes I know that&nbsp;no one knows where is that. Well it is a little town near the coast. It is very hot and there is nothing to do. The only cool thing is that the movies Radio and Forest Gump were filmed there. Now many of you are wondering why the great Zack Gootman would go there. Well I will tell you that&nbsp;I was forced. But I was able to see one of my old friends. We know him as Stoner, Stone Man, or man other names using the word Stone. He really name is long forgotten. He is also called guitar Monkey. It was cool to hang out with him and his girlfriend. I hope to go back and see him sometime.   Well that is about all there is to know about Zack right now... I\'m so lonely.... '),(24598,4075287,'16,August,2004','2004-08-16 00:00:00','   Sometimes you feel like there is nothing more to say. I mean I let you into my life more than my own brother. I mean a person has to have some privacy in his life. I have some scecrets that no one knows and I want it to stay that way. There are somethings that people will never know about me. And it will stay that way. I live a lie.... I and I am ok with that. I don\'t know why I don\'t let people into my life. I just don\'t. And it will stay that way. I feel as if every one I talk to I don\'t really know. People say that I am really good with listening, I thinlk that is because I don\'t like talking about myself. It is not that I am ashamed of my past, I just don\'t like talking about it. Well people need to know. So I will tell you.  I was born Dec. 8, 1988 in Wheeling, WV. I stayed there till I was 3. Then I moved to Gaffney, SC. I lived there till I was in the third grade. We then moved to Walterboro, SC. I hated that town but when I moved up here in the 8th grade I missed it. In the 9th grade I got into Chess, Quizbowl and Theater. I play bass guitar. Cut and dry.  That seems to be about all that I tell people. I could go into more detail but why waste my time. This blog is about me put if you want to get to know me talk to me. I am online all the time. '),(24599,4075287,'13,August,2004','2004-08-13 00:00:00','   I haven\'t posted in a while. Well there are some new things in my life. School has started back and I like all but 1 of my classes. I got a part in the play. Not the one I wanted, I will be playing Yeste the swordsmith. Mr. Scott has had another child. Chess club is looking better than I thought it would. And life is good.  Ummm. Well I have to say that I have been really angry of late. Not at any one else but at myself. I claim to be an actor, that I can get up on stage and have people pay to see me perform. But now one of the easiest thing and one of the most common I just can\'t do it. The simple little line that I would have no problem saying on the stage, I find that in life I just can\'t. Every time I walk up to this person I get all..... Mixed up. I know what I want to say... I have gone over it a thousand times in my head. When I see her I go up and say hi and talk and say good bye, the whole time my mind is scream at me \"say it! Do it, do it now! It is not that hard. three simple words! I love you!\" My heart begs me to do it... But I can\'t! I just can\'t! And that makes me mad. But when I see her smile.... I don\'t know what I will do but she has to know and I have to tell her.... I want it to be in person. To see her face, to take her in my arms... But I can\'t. I know I would mess it but.... Over the phone I still could mess it up... I could write her a letter, but how impersonal is that! I have to tell her in person, it is the only way. '),(24600,4075287,'04,August,2004','2004-08-04 00:00:00','   School starts back tomorrow. I am really looking forward to it. I only wish that I could still sleep in. I most likely get very much sleep tonight. I have a lot on my mind. I can\'t wait to see all my friends. Some more than others. I am really looking forward to working in the theater dept. again. I worked on a summer childrens play but it was not the same as working with my friends at HHS. '),(24601,4075287,'03,August,2004','2004-08-03 00:00:00','   We went through several bands  names. Most of them were stupid but we finally found one that we like. So from now on the band is called English Graden. Interesting name, it comes from a line from I am The Walrus by the Beatles. '),(24602,4075287,'01,August,2004','2004-08-01 00:00:00','   I am now in a band. Wow!!! That is kinda interesting.... Well every one says that they are in a band I really am. We started the band last Friday. We have 1 song finished and 2 or 3 on the way. RIght now there are 3 members. Elliot Garrou, Ben Gootman, and Me. Elliot and Ben play gutiar and I play bass. I guess that you could say that I am the lead gutairist... Yes a bass play is playing lead... Im not really sure how it ended up like that. We started playing and it was determind that I was the best. So when it came time for the gutair solo it turned into a bass solo. Playing bass the solo was smoother and much deeper than any other solo. This added an interesting sound to the song. Our band hopes to write 6 or 7 good songs and 3 or 4 cover songs and then try to get some gigs. The place that we hope to start off at are like Buds Pub or the Rock. We hope to get into this years Kaliedascope. Hopefully with several new songs and more bass solos this band might go some were. '),(24603,3297517,'21,May,2004','2004-05-21 00:00:00','   Well I have totally been flaking on my poor story this week...its my same issue with beginning it and then not getting past a few pages.  But I made myself write some more of it so I could remember that I really do enjoy making a story.  I think if I go too long without writing I kind of lose interest in it and stop caring about the characters..I just gotta force myself to write at least every few days so that when I get a nice big chunk of time I can maybe get a lot further into the story..the newest part of the story is at the bottom of this post.  Let\'s see what else...I got some BEAUTIFUL roses from my mom and sis for graduation.  Its been a long 10 years of part time school and work and a break in between but I finally have a bachelor\'s degree and I\'m very very proud of myself for not giving up.  Tuesday was my last class.  I may go back in a couple years for a teaching credential but I don\'t know and right now I just really need to be done with school and start my real life!  AND I have an interview on Monday so *fingers crossed* hope it goes well!  I\'m excited about it..and dammit I\'ll get my hopes up if I want to!    Finally Mindy began to speak again. \"Well you know I had that conference in Las Vegas...\"  Of course he did.  Mindy was a real estate agent and she often went on business trips.  This last one was to the famous Sin City.  While he had been sure to warn her against staying up all night gambling and downing too many free drinks..even made a few jokes (or he had thought they were jokes at the time) about strippers.  Turns out it wasn\'t the strippers he needed to worry about...apparently what he should have been warning her against were sneaky ex-boyfriends.  Mindy continued to speak.  \"I never in a million years thought I would ever see Randall again..you know things didn\'t end on the best of terms.\"  Heh yah he did..quite a few times he\'d heard the heart-wrenching story of the man who broke Mindy\'s heart by not being willing to marry her three years ago.  Glancing at the white platinum band around Mindy\'s left ring finger, David tried again to swallow the idea that Randall obviously had a very big change of heart.  He raised one eyebrow, signaling Mindy to continue, fighting his instinct to run far, far away as fast as possible.  It was torture but he knew he needed to hear the entire story before he could ever hope to get past it.  '),(24604,3297517,'15,May,2004','2004-05-15 00:00:00','   I wanted to post the rest of what I\'ve written so far..why I\'m not sure but even if no one reads this I still enjoy posting it.  I like having my own journalish web page...hopefully in the next few days I can write about other stuff besidse my wanna be book.  But for now I\'m just enjoying adding what I can to my story and hoping it\'s something I can continue to work at...  Anyway he was sure she was only asking because it was the expected thing to do.  Typical woman - break your heart one minute and the next she\'ll still try to mother you.    David shook himself out of his reverie and managed to focus on Mindy.  \"David?\"  She looked at him beseechingly as if begging him to make this easy on her. While he wanted nothing more then to cut into her with biting sarcasm and express every ounce of the anger that was beginning to boil in his veins, he knew letting that demon loose right now would bring nothing but a guilty conscience later.  He took a deep breath and in a tightly controlled voice managed to ask \"How did this happen?\"  Mindy\'s eyes dropped to the ground and she dug her toes into the ground like an errant schoolgirl. David crossed his arms over his chest and waited while she appeared to gather her thoughts.  After taking a deep breath, she raised her eyes to Davidâ€™s and blew her hair out of her face.  He had once found the trait endearing.  Now he saw it as a stalling method and just wanted to shake her into explaining herself.  David tapped his foot and continued to glare at Mindy, daring her to say something...anything...that would somehow make everything normal again.  He couldnâ€™t believe how much his life had changed in the last 15 minutes.  Just an hour ago, he had been dutifully writing out his guest list for their wedding and now..he could hardly stand to even think it...she had run off and eloped with someone else.  He couldnâ€™t even comprehend her betrayal, let alone be willing to accept whatever explanation she could possibly come up with.  '),(24605,3297517,'13,May,2004','2004-05-13 00:00:00','   Okay!  So I\'ve been asked to make the effort to actually write my book...I have a bad habit of starting but not following through with it.  I\'ve read so many romance novels in my life that the style shouldn\'t be too hard for me to do..but coming up with a decent storyline is the hard part.  I figure for now I\'ll just keep writing and the story will kind of write itself?  Maybe?  And then I can add stuff in later.  I counted the number of words in a paperback and it seems to be about 200 words a page..unless there\'s like massive dialog (another big part of romance books i\'ve noticed).  That means I\'m about 1 1/4 pages into my book..lol only 300 or so more to go!  So here\'s the next part I\'ve written...  He abruptly tuned back into the words that had torn his world apart just seconds ago.  â€œAnd tomorrow Iâ€™ll be by to get the rest of my stuffâ€¦and then me and Randall are off to the Caymans for our honeymoon.â€  His fianceeâ€™s..make that ex-fianceeâ€™sâ€¦ blue eyes were big in her pixie face.  It was what he had noticed first when his own dark eyes had passed over her face at his sisterâ€™s wedding.  Mindy had caught his attention when he glanced over the other guests and noticed those big blue eyes tearing up as his sister and her husband exchanged their vows.  Now those same eyes that had captivated him were watching him with remorse and a vague sense of desperation to be done with the conversation.  â€œDavid? Are you alright?â€  From some place far away David heard her but couldnâ€™t quite comprehend that she would even ask.   Maybe I\'ll write some more today..its kinda fun making up characters that can do whatever you want them to do.  If I can just get a bunch of pages done I\'ll feel I have no choice but to keep working on it.  Besides with all the changes that are about to happen in the next couple months I could really use the diversion!!  '),(24606,3297517,'11,May,2004','2004-05-11 00:00:00','   My very first post...and I was just thinking about the book I\'ve been reading and how boring it is..and how since I have a decent vocabulary, I should be able to write a romance book at least as interesting.  And then I can get published...of course I\'ve never been able to write a story longer then a couple pages so a book might be a *bit* of a stretch..but why the hell not.  Everything in my life is gonna be craziness shortly..why not throw in writing a book too!    Okay here we go... What was that word..that one they used to describe when a man felt like every drop of testosterone had been drained from his body, when the woman he had planned to love forever castrated him with the blade of her deception, when every hope he had for the future slipped down the drain of infidelity..what was it again?  oh yah..emasculated...meaning to \"deprive of masculine vigor\".  It may have been since high school since he\'d heard it but at this moment no other word was more fitting and it had stuck in his head because he was so sure it would never happen to him.  '),(24607,3297517,'05,June,2004','2004-06-05 00:00:00','   Oookay been a little while since I wrote in here but I\'m back!  Yay I know you missed me!  ;)  So Memorial Day weekend was great.  I went down to San Luis Obispo County to visit my mom and sister, Kate, and one of my best friends, Jacqui, and it was such a great time!  Reminded me a little of what it\'s like to have a life...lol they\'re kinda fun! My sister\'s roommate put together a bbq the first night and had a few people over..we just drank beers and ate and had a good time.  Then on Monday Kate, her friend, and me and Jacqui went to the beach to lay out for a bit.  Jacqui and I left after a couple hours but when we got back into town found out her bf was going to play volleyball on the beach..so we headed back out and had Kate and her friend meet us and we played beach volleyball for the afternoon!  It was so much fun..though apparently beach volleyball is an excellent workout for your ass lol.  I was quite sore the next day!  Then I drove back up on Tuesday and the whole weekend was just perfect.  I couldn\'t have asked for a better time!  Alright now to see if I can continue on with my story....  Mindy took a few deep breaths, in through the nose and out through the mouth, as if gathering her strength.  He had seen her do it often enough in stressful situations and knew it was a tool she had picked up in one of her yoga classes.  Whenever life started to get overwhelming, Mindy would take a minute to \"center herself\".  David began tapping his foot in a rather fast tempo, perhaps hoping that the fast pace would quicken her centering so he could finish this conversation and get the hell out of there.  Finally she began speaking again.  \"Well it all happened so quickly..\"  He\'d say!  She left an engaged woman and came back three days later a married woman..just not to the man she had been engaged to when she left.  The irony was not lost on David.  \"I walked into the conference room and had just sat down when I heard someone call my name in a hushed tone.  I looked up and there was Randall...looking at me like I was an angel.\"  David couldn\'t help the snort that escaped, but kept his mouth shut when she gave him a half exasperated half guilty look.  \"Obviously I know how much he broke my heart back then and trust me I was in no rush to forgive him.\"  Mindy must\'ve thought his snort was in reference to how quickly she changed her mind, and while that in itself was unbelievable, calling Mindy an angel was what had actually prompted his response.  Mindy was smart, beautiful...and could be kind of ruthless when she needed to be.  Angel was not the first word he would use to describe her....in fact she was probably closer to devilish.  But he let her continue.  \"He sat down next to me and would not leave until I would hear him out.  So I agreed to get coffee with him after the conference.  He told me that losing me was the biggest mistake he ever made but by the time he realized it, six months had passed and I was already in a relationship with you and he didn\'t feel it was his place to interfere.  He had even picked a ring out for me but was so worried that I would hate him more if he tried to break us up that he just kept it with him at all times and dreamed of the day that he could offer it to me.\"  Alright thats all I\'ve got for now..its kinda funny cuz I\'ll want to write more but then think I won\'t be able to come up with anything..but then when I start actually writing it just flows.  Let\'s hope that continues once I\'m done with this intro conversation...moving into other situations will be interesting.  '),(24608,3297517,'13,July,2004','2004-07-13 00:00:00','   Forgot to add this latest part i wrote a little bit ago...things are busy right now..or are about to be anyway so i haven\'t been writing much. hopefully i\'ll get back into it later when things are all chill!  \"That day apparently coincided with your business trip?\"  David heard the trace of suspicion coloring his voice and apparently Mindy heard it too.  She looked at him as if he had just suggested she had bought her clothes at Wal-Mart.  \"It wasn\'t planned David.  God dammit do you really think I\'ve been lying to you all this time?\"  At that moment her eyes started to fill with tears.  David watched in wonder as Mindy switched to the role he had watched her perform so many times for other people, but he had never had the opportunity to witness the magic, or in her case tragic,  transformation first hand.  He couldn\'t believe she was going to use the \"manipulation tears\", as he called them, on him!  She must really be desperate if she was resorting to that poor me act.  He\'d watched her practice the look in front of the mirror before going to court for a speeding ticket or on her father when she needed to borrow some money..though that was long before she started making it big in real estate.  He even found himself wondering if the tears that had first captured his heart that he saw her shed at his sister\'s wedding had been fabricated as well.  He was beginning to feel as if he didn\'t know her at all.  \"Cut the bullshit Mindy!  You know I\'m not gonna fall for that so let\'s just skip the little girl lost routine and finish this conversation.\"  David pounded his fist on the counter next to him in frustration.  Though the physical release of pain made him feel momentarily better, the pulsing of his hand only served to mock him for losing control.  '),(24609,4232070,'23,August,2004','2004-08-23 00:00:00','   M o o d :  a c t  h a p p y  a c t  t o u g h  yeah ppl.....jus act happi then sooner or later u will be happi.....anyway Jin Shun.....i read ur blog and it seems that u had  misunderstood  me......i noe u  noe  quite a bit but then......diao u ask mi to give up on her?......come  on  i already  gave up  on her.....since the dae she utter those those....HIAZ! Thinking of it sure remind mi of that dae....I dunno if  tt wad she realli wan ....but wad sure is that  I AM STILL NOT GOING BACK .....sorrie...dun try to convince mi ar  Jin Shun ....i noe u  meant  well....dun use my  weakness ...dun...play  mind   games  wif  mi ...oh ya..... sorry  is for  Jin   Shun  ....jus in  case  [ somebody ] (*coughÂ²) dunno  who  ar....sae  sry   does   not   exist  in [this person\'s]  dictionary . [Somebody sae why do things which requires sry?] Then lemmie ask u......Are u perfect? People unknowingly done sumthing wrong and therefore sae SRY. That SRY is right frm the heart and if u dun appreciate it.... FINE ! So i suppose u r one person who dun sae sry? Even when u noe u made a  grave  mistake? PPL DUN do things which requires a  SRY , wad for sure is that  MOST  including  mi ....sae it  onli  when we  realise  our  mistakes  or  realise  the  damaged   incurred . So......dun  argue  anymore.  Daffy   å§å§  i think you  reasonable   ennuff  to  understand  this  logic  rite.....i am sure u r quite normal......the way u  think ......if u  dun   agree  wif mi....then i  diao ...  le ....then realli nothing to sae  liao . Getting  ready  the  rope  to  hang  myself...haÂ². Anyway  Jin   Shun  now u noe le so dun  anyhow   guess  ar..... ByeÂ²   -you  live  to  die ,  rejected  and  alone -   '),(24610,4232070,'22,August,2004','2004-08-22 00:00:00','   M o o d :  g l o o m y  + a  l i t t l e  f i n e      What a day!... yes ppl...finally told the grp that I  dun belong  there liao...hiaz...sad lar. Gave them  a box  of chocolate...and  after recess ....found them  on  my  chair ...diao.... wondering izzit  cannot  finish caus its  opened .  arrrrrrrggghh ....  lost  so many  friends .....then comes  Jin   Shun  telling me that I AM  SELFISH ....for  leaving  the  group  for one  gal .....sumthing like i \" å¯¹ä¸èµ· \"  4  of them caus of  1 . Hiaz... how he noe my  pain ? I realli mus  leave  lar...Yeah...today is  Daisy\'s  birthday...so Happy  Birthday  gal.... HEY ! I am still  angry  ar! I think I duwan to tok abt God liao...Daisy is  rite ....sometimes i mus  pamper  myself...mus not let myself  suffer  too much...love  yourself  ..more than other ppl...then  nothing  will go  wrong ....tt onli  partly  correct....i am jus taking this  suggestion   temporary  to make myself  happi ....anyway..think i am going  out  wif  Jin   Shun  this Friday....and hopefully  Candy  and  Joanne ...my  sisters  too...but still dunno if they  angry  wif mi  caus  of [her]...AIYA!!! dunno how to  ask  them too....caus now i am  not  part of the grp liao... miss  them.....er....even if they cannnot make it I hope Jin Shun dun ask tt  Guan   Heng  to  tag  along...pls lar Jin Shun....care for mi this once lor...care for my feelings caus  nobody  does ...sometimes  dunno  why i care for ppl so much yet myself lyk heck care ....anyway tt the opposite of Daisy. (*coughÂ²)..erm.....i think its time for me to  stone  again. So thats abt it? yeah? ByeÂ² to all the  angels  out there...oh ya....especially Angel: Daffy å¤§å§...so like frm heaven wan....so kind so  sweet ....no lies...realli!      -love myself-   â”‚â–Œâ–ˆâ”‚â–Œâ•‘â–â–ˆâ–â”‚â‚¬Ä‘Ä‘Ã¯Ñ”â„¢â”‚â–Œâ–ˆâ–Œâ•‘â–â”‚â–ˆâ–â”‚   '),(24611,4232070,'22,August,2004','2004-08-22 00:00:00','   M o o d : d e e p l y  d e p r e s s e d         Sorry that I did not  continue  to Blog yesterday, because after that \" talk \",  exXxe  and [ her ] are not friends anymore. No more. We argued and sort of quarreled. She used extremely  harsh words  and ask some very  deadly  questions. Questions that she know will hurt the listener\'s heart, literally giving me a heart attack...and  in series ...so is like combos...get it? Almost got KO by her. Told [her] she is  very  mean and selfish. A girl who is so ego- centric , so self- centered ... is not like [her] at all. I told her she is like a split.(2 personality). She denys. I was hurt so much that I sort of hate her ...alot. I do not know how to describe, but I kind of hate her and at the same time not. I hate her for what she said yesterday, thats all. But thats enough too, she had gone too far,  too much , absolutely over-board. Even I myself cannot believe what she said to me. I kept wondering\"is that [her]?\". Yes  Eddie,  wake up! Its her.Take it easy, believe the truth! Thats her alright! Try accepting it!  Daffy å¤§å§ , if you are there...heres a qusetion for you. \"Did she do that on purpose, to make me stop thinking about her?\". Anyway, if thats the case, then... she is making a big mistake. Thats not the way to solve problems....then again...if thats from her  heart ...she is not worthy...sorry to use this word...but if only you people heard the conversation...she is like posessed. Not worthy to be my friend. Maybe she is not even worth my time. Hate her so much now...but thats now...my previous wounds are back again. Devastation to me. I think she has given me enough \" hints \" about me not belonging to the group. I think my time  is up . I have to leave. I will inform the group tomorrow. I must not leave just like that without telling them, or I will regret forever. I will miss the group... alot ...especially  Candy  and  Joanne , who reallyÂ² treat me like a  brother .Love them so much, I think I am going to  miss  them alot. This time, I need a longer time to heal the wounds. Like I said yesterday...she literally used a shotgun. Shotgun is spread shot, so there are many bullets in me. That will take a while to get it out, all of them. Think I have to stop now, because  Jin Shun  got to go and he wants to read them now. So i got to publish now. Love you  Daffy å¤§å§  and  Jin Shun  too.ByeÂ²      -reluctant to leave-   '),(24612,4232070,'21,August,2004','2004-08-21 00:00:00','   M o o d : v e r y   c o n f u s e d   a n d   s a d       I really don\'t know what happened that made [her] ask me that question. Is it something  wrong  I did? Or was it something I  said ? She ask me \"if one day\",\"i mean if la\", \"if one day the group reject u joining the group\",\"how will u feel\". I then  told  her that I must had done something  very wrong . She shen continused\"if it\'s not becoz if doing something wrong\". I then got really  shocked , and almost dumb folded. She then asked\"how do u tell someone that they actually dun belong to a group\". I almost got what she mean by then. I am  so sad  now. Nobody understands. Nobody. Not even me. What is the cause of it? Is it the question that I answered? Or is it that my  presence  made her feel  awkward . Can anyone tell me? Daffy å¤§å§, if you are reading this, I hope you can tell me the answer. Why is she like that to me? I already told her, that she could treat me like a brother. (away).(back).Just told her that she could be my sister. I am willing to forget about the sad past... now . But why can\'t she? She told me that she is avoiding me. I expected that and  felt that too anyway . Apparently, she literally took a shotgun and  blast  it right through my heart. That hurts. Just heard that Daffy å¤§å§ betrayed me. So sad. I  did not  expect that to happen. The  group  is still very nice to me. Except Daisy and Shu Qing, who both talk very little to me. Even though sometimes the group is  unusually  quiet, but we are still  close . It cannot be true that the group want me  out of the herd . The root of the problem... the girl . I think I will update again later...after the \"talk\" is over.      -falling deep-   '),(24613,4232070,'20,August,2004','2004-08-20 00:00:00','   M o o d : v e r y  h a p p y          Wow  ppl [she] jus msg mi without mi asking her any qusetion...but  too bad  i -afk-.....just now having dinner and then after tt  learn  piano piece from my sister....tml is my turn to  teach  her......so we ex chan ge knowledge lar.....then came bac and saw the msg.....its been long since i received her msg....btw i am now able to post pics too...yeah...think tt the pic will stay there for a few daes before i change it....ya? I think it look r oman tic.....g ues s wad...the group thinks tt its ok if i  speak  english .....haha tts better...caus my  defau lt language is eng...(chatting with [her])......(BACK)...haha its not  her  who msg mi....lol...nvm ....still  happ i caus hav a  new  fren....[her] sister... da phn....anyway i think nobody will noe who is [her]frm her sis\'s name...... diao ....did i jus gave a hint???...lolx......she seems  veri  friendly....and more  responsive  than [her]....wah taking a gd  course  in ngee ann poly.... sma rt ar....then i wonder.....jus w onde ring ar.....last time [her cousin chat wif mi...dunno izzit daphn.....nvm it dosen\'t matter...] She ask mi some qns which i believe she is going to keep a secret.....diao...i found out sumthing.....when i write in blog....my eng is so  out of my  expectations.....u noe?the kind of sub-standard english...i noe tt called  singlish  ...haha i think i am chatting wif daphn later or should i call her \" daphneå§å§ \"...hehe...so happi....think \"daphne å§å§\"may be present when i celebrate [her] birthday......next week? Hoping to see her......such nice person,  just like me ...lol....(*coldÂ²) laugh leh! Think tt abt it fer todae.....though i hav a feeling tt i hav sumthing to sae....dunno wad leh....forgot....nvm....answering questions soon.....- å†è§ -.....everyone mus \" guaiÂ² \" ar...listen to ur mom ar! Till the next update....which is tml.... take care .      -reached new grounds   '),(24614,4232070,'19,August,2004','2004-08-19 00:00:00','   M o o d : h a p p y  Today got someone told mi tt that  damn gee teng  show alot of ppl my neoprint....haha someone \" bao tuo \" him...caus nobody like him.....hate him man.....then ask him izzit true then he deny.....of coures lar.....the human normal reaction to avoid trouble is to  deny .....then Jin Shun ask him to \" sompa \"with his whole family life....LOL....then he quite hesitant.....caus his life is  worthless .....( Jin Shun sae wan )...idiot wan lor....somemore yesterdae read janet\'s blog and found out tt freaking gee teng actually took her [ 12 neoprints] and then lost it....omg....he lost it ...she was veri pissed...veriÂ³....then todae alot of ppl told him tt he owe Janet $92 or $72 caus some is 6 dollars, some is 8 dollars. So irresponsible of him......wad to do....all the things tt he do realli \" roX \".....almost a pest lor. From what i noe i am the 2nd person to comment on tt  freak  in blogspot....at least  i not  so angry as janet...her title realli suits...and realli roX. Now i noe where all the  comments  on  mi and the group  come frm and now spread like...diao...caus of...... GEE TENG !!! Tell u all lar.... dun  wanna  hide ....realli wan to  pu nc h  tt sanddoll (gee teng)...(actually wanna sae shitbag wan ..lol )todae....he made my fren  BURN  and mi  angry ......somemore  dun  dare to admit....so gay.I am veri veri sure some of u who read this  think  so  too . Hiaz... though t he change my mind abt him....tadae he  screw  it up again.....lucky now i  not angry  liao ....caus Jin Shun  ask him to  sompa  and ask him to \"æ‘¸æ‘¸è‰¯å¿ƒ\".....tt made mi feel amused and sort of quenched the  scorching devil  in me......tt abt it ppl..... watching  Alien vs Predator tml....after Physics and Chemistry Practical mock exam.... i think i updating again at night....so cya....again.ã€€ã•ã‚ˆãªã‚‰.....(sayonara)  -back to reality-   '),(24615,4232070,'19,August,2004','2004-08-19 00:00:00','   M o o d : h a p p y  Hey ppl..u all should take this test too...IQ test frm:   urlLink  http://web.tickle.com/tests/uiq/authorize/signin.jsp?url=/tests/uiq/index2.jsp     this is my result... yeah so happi     Eddie, you answered 24 out of 30 questions correctly! Congratulations! Your score is in the percentile(\"91\"); 91st percentile. This means that if one hundred people took the test with you, your score would rank higher than 90 of them on average.When we analyzed your test, we also discovered that when it comes to quantitative ability, you measure in the percentile(\"83\"); 83rd percentile. This score indicates you have unusually strong abilities when it comes to solving numerical problems. If there is a numerical pattern to be found, you\'ll find it. You\'ve got a knack for noticing when something \"isn\'t right.\" Whether you\'re conscious of it or not, you have an ability to simply understand when something doesn\'t add up. Also, when it comes to splitting the check, doing taxes, or determining the number of fans in a baseball stadium, you\'re the one people turn to.    P.S anyway hope tt u all not offended by this...caus this is my blog and why can\'t i post something which i am proud of? GuaiÂ² ar everyone byeÂ²    -vanished-   '),(24616,4232070,'19,August,2004','2004-08-19 00:00:00','   M  o o  d : f  i n  e   ello  ppl out there welcome u bac again to read  the little bookâ„¢.... today i finished my o level eng oral and i think tt i did  fine ....although i have many other things to sae and the 2 examiner keep ....like s toppi ng mi to talk more.....i have so many things to sae  so  10 mins maybe not ennuff lor....i think i jus need 2 more mins...but too bad they sae \"Ok\"...so tt means tt my time is up and gtg.....lol. nvm lor  see how it goes....anyway today joanne seems  sad again ......hiaz d un no y caus she dun wan to sae the reason...again. Ask  her wad happened....or when i try g ues sing the prob...she said\"  nothing \" again.......actually i think tt if  frens  got  prob...  can  sae it out  wad... so ppl can help u solve the prob or give u suggestions....but if  dun wan  its ok with mi lor....jus wanna ask \" do u feel better...  if i did not ask anything ?.....\"sometimes ppl  dun wan  other ppl to  noe  their prob....so it wise to  shhHHhhhh .....Think i going out wif Jin Shun one fine dae.....soon.And also i watching Alien vs Predator  this sat ....alone.....if  anyone  wanna go wif mi then tell mi b4 sat... this sat ......i might wan to go Lido to watch it.   M o o d  c h a n g e : g l  o o m  y  Aiya .....this music so nice  yet  so sad too......rite? Pianist  noes.......kinda thinking of [her] now.....but mus wait till o level finish then s he  will be more observant to ppl around her?....i dunno.....anyway i am still  waiting .....still  loving .....even though i am  onli  a bro ther  to [her]. I dunno lar ...i hav a feeling she is kinda a void ing mi....we dun tok as much as b4.... ( AWAY ) ...4 mins later....(BACK)... jus taught [her] how to change the computer sound scheme.Aiya...u all  dun need to noe  wan lar...at 1st she  said duwan  one.....then she said k....this gives mi a  feeling  tt she said \"k\" caus  in future  i no need to \" farn \" her.....it may  not  be  true ...and  as i said ...it gave mi tt feeling.....so i jus write wat  i feel  in my blog lor...anyway...i think she is too  busy  to even read  my  blog.....so i think she won\'t see  my msg  or  how i feel  wan. I re gr et answering  her Qn ....realli.....now like  1/2 fren  onli.....almost  no  eye contact..... no  topic ...she will  not tok  to mi unless  i tok  to her.....and her ans veri short one....so  sad  hor....i dunno if time can heal the wound...... time  heal all wounds?......we shall see....i guess tt abt it ppl..... ByeÂ²   -farÂ² away from reality-   '),(24617,4232070,'18,August,2004','2004-08-18 00:00:00','   M o o d : q u i t e  d o w n     Sry ar readers, I also dun wan to make u all  sad  wan...i mean...if this is my blog i should not care what other ppl will think of...rite...so i should just sae what i feel and not care what the readers  esp ppl  i am referring to feel abt mi. So..here it goes...today is a fine dae, and found out tt guan leng is not a gay...(dunno y ppl keep saeing he is one... HiaZ ). Anyway jus finished dinner and came bac to this  com  and saw a  msg  frm D ai sy....its abt the birthdae meal thinge and b4 i went for dinner...i asked her \"when are we all going out for a meal\" onli  for tt meal ...so i wait fer her reply and then...sian she as usual is always busy and did not reply lor....so i went fer dinner. YA...as i said i came bac rite? She replied and said tt there will be 2 \"times\" ...she said tt the first time is the 5 of them onli.... wAHhHHhhhH  my heart  pain  man. The 2nd time is with all of us....which means there will be alot more  other  ppl ...i believe is more than the  6  of us.... arGHHHhhh...so pain!  i keep think izzit of my presence or wat lor.... some of u noe  rite...i thought i am worth to be sitting in the same table...jus the 6 of us...I dunno lar ..maybe the  grp  would like it but daisy jus dun wan to be \"em barrasse d\" lor.....but as i said  i dunno . Maybe i  think too much ...  btw...tml i am taking O\'level  english oral  and now feeling so crap lor...A  bit stress ....stress level inc reas ing....anyway i hope tml  Jin Shun  would tok to mi  more  and make mi feel  better  ( I noe u won\'t fail to  make ppl happi  ) caus i feeling  so down  now... sry  ar if i made u ppl down too. Anyway...I am still  waiting ....Oh..jus remembered tt she kinda  avoided  mi today...the onli thing I said to her was \" bye daisy \"....so sad hor....wad to do? Just bear with it lor...tt all fer today ppl. ByeÂ²  -gone far away-   '),(24618,4232070,'17,August,2004','2004-08-17 00:00:00','   M o o d : Q u i t e  a n n o y e d   Just sat down and tok to my mom....man....she just wan mi to speak with  elegance ...oh ...crap. How?...I mean i am told not to mix around wif bad kids agian.....not with all the \"jiao kia\" in my class.....and I told her tt i would be treated differently if i were to be too \" guai \"...she said it dosen\'t matter what others think of you.....see wat i mean.....i mus be a \" guai kia \"...she said tt only the bad avoid the gd...hmm tt sounds correct lor.....so i said okokokokokokok then settle liao lor. So i will be more guai now  haÂ²  even though i  veri  guai already.....LOL..(la ug h l eh !). Think i sleeping now so bye Â² .  -gone already-   '),(24619,4232070,'16,August,2004','2004-08-16 00:00:00','   M o o d : s  o  H  a p p  y  Man...todae is a good dae..haÂ². Nothing bad happened to me, and as usual I sat with the group. We were toking and toking and i think we are going out together on daisy\'s birthday....on tt dae she will treat ppl to a meal and everything...but for mi...i think i will treat her lor... since  is her Birthdae wad. I considering buying her another  gift  ....as a  birthday  present. U ppl comment on wad is should buy leh.Aiya...wad i expect? Onli a few ppl noe abt this BLOG.....and all nv comment wan. Anyway u ppl dun anyhow send to other classmate ar.....or i ha haha ha...not telling u. HaÂ² anyway todae got Physic and Chemistry mock exam...PHY is P1 and P2...Chem is P1 onli...and i  think  tt i will score for PHY.  Think  onli ar....so if dun hav dun \" suan \" mi...Lol. Think tt abt it for today....so cya tml.  -goneÂ³-   '),(24620,4232070,'16,August,2004','2004-08-16 00:00:00','   M o o d : Q u i t e h a p p y      HaÂ²....certainly felt alot better after i explode. Today when i sat with the group again and we were discussing abt going out ( i guess) , to celebrate Daisy\'s birthday. WAH.....felt so happi and looking foward to tt dae ( which is not set yet...lol ). I hope i can give her a treat lor. See..like the group...if i am treated with respect i will return with respect...the onli prob now is tt i cannot stop communicating with them in english ...haha trying to communicate with chi lor...so tt we can have more \"æ„Ÿæƒ…\" haÂ².....aiyo if u all ( the 5 gals ) are reading hope u all can use chinese with mi lar...persist and u will see the results....provided tt chi is easier 4 u all lor...heÂ²....cuas i always hear u all use chi wan....rarely hear eng outta anyone there.....so i e xtra  wad ....so i sacrifice and make changes lor.....lol.    -gone for now-   '),(24621,4232070,'15,August,2004','2004-08-15 00:00:00','   M o o d : a n x i o u s      Today during F & N lesson tt Pearlyn got crazy again.... always see mi \"bue song\"...WTH is wrong with her i dunno...always liddat to mi....lucky tt jia ying nv siaoÂ² or else i sure dunno wat to do. Sian man got to see her face everyday and she will sae whateva she can use to hurt mi.....think i am different????...hahahaha... got it wrong gal!!! ...i am far more smarter than u    M o o d c h a n g e: F u r i o u s  SO JUST  SHUT  THE HELL UP OK!!!!!!I am not like a dumb doll tt u can bully with and IF U GOT THE GUTS ......INSULT louder and look into my eyes and tok to mi....DUN SO FREAKING \"HUM\"....AS i said    M o o d c h a n g e :  e y e s   a r e   b u r n i n g  i am not what u think i am...think and reflect...do u think someone abnormal can play all kinds of sports and play the piano and excel in his studies???....just because i hate the class and wish to be quiet and not interact with u ppl dosen\'t mean i had been hit by the mill. Watch your attitude \"gal\" and u will not want to see your nemesis in mi........had ennuff of u ok!!!! Think I am someone who do not have feeling...i ingore because i don\'t wan to be embarressed....to tell you the truth i always pretend tt i did not hear insult...       -gone for now-   '),(24622,4232070,'15,August,2004','2004-08-15 00:00:00','   M o o d :  a b i t  s a d  Its been four days since I got rejected by [her] and its still in mi.....those who noe who tt gal is dun tell other ppl ok. Still cannot forget abt her though accepting that she is not rdy for a relationship. I told her I would wait...anyway she seems so engrossed with her O\'Level...wah...then I ask myself:\"izzit the time for awakening?\" I got a pathetic C5 for my chinese language and my oral is onli a (P) pass ...but of coures chi is my  worst  sub so other sub should be fine ...I mean  better . Aiyo, lucky for Jin Shun or else I would \"æƒ³ä¸å¼€\" I realli wan to use this chance to thank him for all the concern provided that he reads this. He is one kind soul out there to help people...noe when to joke and when not to.... recently ..haÂ². Kinda respect him more now for all that he contributed...his unlimited jokes...though some realli \"yellow\"....I remember I cannot accept those jokes in the past but now.....sian already...lol. Jin Shun is rite abt the other time.....I realli is too anxious....sian man. Just hope that after O\'Level I will be given the chance lor....provided that I still luv her.   Oh!  I just remembered something...yesterday tok to somebody on [her] MSN acc...its not her but its her cousin...so late still awake...nvm(wondering izzit[her]) he claim to be her cousin so ok lor talk abit with him...we tok abt music and......wah then suddenly he got quite pissed ...think he misunderstood my meaning of music so can tell tt he not happi lar...i  diao ....dunno wat i do wrong but told him he misunderstood and said sry to him lor.....then after tt he sae ok and things lidddat then settle liao haha....he said \"nice chatting with u\"..tt of coures end the whole conversation. He also quite smart leh...chinese get A1 ....but who noes later like tt Karine ...everything bad except chi ...so sad ...still so notorious and haven wake up yet....sry but i think her chances of going to a poly is slim. Hope God guild her lor.Sian i gtg finish ANBA\'s homework liao. Anyway just hope that this comming FRI can go out with the group lor......å†è§!  -gone for now-   '),(24623,4232070,'14,August,2004','2004-08-14 00:00:00','   Mood: SianÂ³      Does anyone noe how to post pictures and equip the blog page with music ? Can anyone teach mi? Pls teach mi if u noe thx!Caus my blog seems incomplete...i am new to this.      -gone for now-   '),(24624,4232070,'14,August,2004','2004-08-14 00:00:00','   Mood: Calm  Today at church when service is over I went to the grand Yamaha piano and played some church piano pieces....so dangÂ²...everything went on well UNTILL I hear a IDIOT, literally a idiot i mus sae.This unknown person was leaning against a half opaque, half transparent door(like the bathroom door) near the piano. He then said \"u will die....\". Then i down there at my seat ...stopped playing and listen which moron is tt.Then i though is my fren then ok lor nvm.I continued playing with almost perfection and then COMES THE MORON AGAIN!!!!!He kept muttering some insane things and i DIAO.....whats the porblem with him?I then told myself \"its ok\" Must be some siao kia or wat lar.Then i continued playing and within 5 sec tt guy bang onto the door......I then quickly finished playing tt piece and I went out to take a look....at first he is nowhere to be seen...then i spotted him far away.  I then shouted...\"whats the problem\" and oh! I realli shouted.( I must be damn pissed) He then act gong with mi and pointed to himself and asking mi ...\"me?\' like so innocent liddat. Then his sis looked at mi and ask him along and they both approach mi.....His sister dunno elder or younger jus kept saying sry and everything she could sae to remedy the problem....then i quite ok with tt untill HE the ignorant guy who simply look too old (older than mi) sae:\"no need to sae sry!\" I then ask him wth he is doing and frm the way he tok abit \"å¤§èˆŒ\" like the tongue scaled by real hot water....and then his teeth all dented in and out and i though so \"å¯å¿´\" aiya then I forgive him lor on the account of her sister who is like pleading mi like never before. His sister sae he at home also liddat one \"siaoÂ²\" wan so ask mi to pardon him. I told tt guy tt there will be other people playing the piano later and better dun disturb later.Then I walk away lor snd told her sis tt she is a nice person ...wah she at tt time still\" sry ar , sry , sry....\" Then I continue playing but felt quite sian..after a while I went bac home.  -gone for now-   '),(24625,4232070,'14,August,2004','2004-08-14 00:00:00','   Mood: Quite happy  What a boring morning in Orchard walking about and heading nowhere, went to HEEREN and sian...its not open yet...so went to OLD CHANG KEE and bought quite alot of crap which i ate and felt like vommiting. Anyway, seems tt time past quickly and soon its time for some movie I guess. So I head to Lido and bought one CATWOMAN ticket at 11.00 AM. Watched the movie and found tt it is so so sian caus the great villian is a woman who put to much of tt cream...lol.Aiya jus dun watch lar caus its not realli worth your money.Nothing realli happen in church when i went there in the afternoon...jus some play going on.Aiyo..my life quite sian leh.But I am grateful God dun give me accident to spice up my life....LOL. Think i gtg its almost 12 AM midnight. ByeÂ²  -gone-   '),(24626,3753019,'30,June,2004','2004-06-30 00:00:00','   i went to sleep early yesterday so i wouldnt get up late today. that was successful. i have yet to do much of anything. got the crack in my ghetto car\'s windshield fixed... hopefully. i did it the white trash way. this guy on the side of a long road in a motor home fixed it for $60. haha. yeehaw. now i have to go turn in my applications and call in to see if i work... i need to work dammit.  i talked to two of my VA boys earlier! that made my day so much better. i love them so much. aj\'s mom just called me back and thought she had missed a call from her brother in TX, that was sorta funny. she remembers me. sweet.  having talked to them, i got things sorted out and the culprit\'s name was revealed to me. RUMORS ARE BAD, AND THEY HURT PEOPLE. not just the people they\'re about either. but how shallow a person to spread a rumor about your \"best friend\". whatever. i dont have time for that crap anywho. people will grow up and get a life someday i reckon, if not, God help them.  July 4th is this weekend... i wonder what im going to end up doing...   '),(24627,3753019,'29,June,2004','2004-06-29 00:00:00','   well i have to say, Warped Tour ROCKED. Everyone kicked ass, and i crowd surfed! Hell yeah. It was awesome. Granted they dropped me once and i hit my head... but thats alright. Im kinda sore from the 8 hours of physical activity that i pushed myself to participate in. Unfortunately i only heard one of Thursday\'s songs cuz i almost passed out... it was so hot. Texas for ya...  On the way there i was listening to a cd i made quite a few months ago, and realized all the songs i love are the songs that remind me of someone i hate to be reminded of. Not the person actually, but the situation we\'re currently in. Or the situation it seems we\'ll never leave, rather. Funny how cliches sometimes fit your life. \"love hurts\" dammit.  Aside from that, im looking for another job. Unfortunately im in Beaumont and left my applications at my dads in Nederland. Looks like i\'ll be dropping them off tomorrow instead of today. Lets hope that goes well.   '),(24628,3753019,'27,June,2004','2004-06-27 00:00:00','   so yesterday i slept all day. i usually hate putting a day to such waste. but i still hadnt gotten much sleep. i went to work last night and it was so weird. ive been out of town, but i feel like ive been away from life for a long time. reality is still trying to make its way back into my head. i dont know why it hasnt yet. maybe i dont want it to. all i can think about is what i want my life to be like, and i try not to think of what it is.  desperate to go back, i find myself almost recklessly trying to download all of the songs that mark the week that i visited in VA. if it werent so far...  i hate secrets. and even more so, i hate rumors. especially the rumors that some believe are just secrets that somehow got out. if you dont think its something i would do... do you think i\'d do it? im not that unpredictable am i?  so dont let the little things get in the way. dont let the little things in life get you down. if you dont look at the overall picture, you miss the point. and you dont want to miss the point.   '),(24629,3753019,'25,June,2004','2004-06-25 00:00:00','   so now i have something that i\'ll post online for nobody to read. or for them to read and be disgusted by my pessimism. yes you all know how i can be. i know how i can be. and it makes me sick. the most nauseating part of it all is i manage to screw things up, and usually dont see what i did wrong. but after my visit to virginia, im beginning to think maybe im not the person i think i am. i know the good things i do. i know how much i love everyone and the things i do for them and the lengths ive gone. those are the things that make me feel good about myself. but its never enough. and when i finally say, \"thats it, i cant do any more. its too much.\"  thats when i become the bad guy in the black mask. this i dont understand, and im beginning to think that maybe it is just me, and its not everyone else. everyone thinks theyre the good ones in bad situations. maybe i am like everyone else. all i know is that i dont want to be here, and im not wanted where i want to be. so im so stuck. i need to leave. i need to be gone from everything and everyone. i need sleep. i havent slept but maybe 16 hours since i woke up at 9 tuesday morning. even then i didnt get much sleep while i was up there. i realize this is extremely long for my first \"entry\". i have a lot on my mind. and even more to say. but i\'ll space it out. or just take a break. im going to go away.   '),(24630,3753019,'31,July,2004','2004-07-31 00:00:00','   \"i wake up next to you, start the day off right.\"  such a beautiful thing for a person to say, it brings a tear to my eye.  i love  mom left yesterday morning to bring the kiddos to FL. so ive had the house to myself for a while. its nice. but im still bored as always. at least im not being stressed out by other unmentioned occupants. nah... its not that bad. but the peace and quiet is quite nice. i also get to park in the garage when she\'s gone. and that\'s convenient. my car is always blazing hot when i get in it. i will have this luxury once more when she goes to pick the kids up from FL in a week. im gunna miss them. i already do. i love my little girl and my baby boy. such sweethearts.  i got horrendous bug bites on my feet sometime throughout the course of last nights activities. and they itch like a bitch. mercy me.  nothing else has really been going on. minus the things i havent mentioned.  ; )  and we will keep it that way. i love being happy.   '),(24631,3753019,'28,July,2004','2004-07-28 00:00:00','   or am i dreaming?      &nbsp;  oh man. i cant handle being this happy. i cant contain myself. i dont know what to do! all i know to do is smile, but if i smile i\'ll have the stupid cheese on my face all day. then people will start to wonder. i woke up happy this morning, and that\'s all anyone needs to know.  while i can hardly handle my happiness, i love feeling like im about to bust. i love having the urge to run around and hug everyone. what&nbsp;i do not love is that the one person i want to hug and give my love to is not available. stupid timing.   yesterday i went to the eye doctor. they dilated my eyes. that was... well, for a few hours i couldnt see up close, and the sun killed my eyes. soon after i got home jacob calls me and needs me to drive to nederland to let him in my dads apt. because he locked himself out. (genius)&nbsp; so i do, and then i wait for my dad to get home so i can talk to him about getting new handles on my car doors. (the night before someone had gotten into my car. stupid people didnt take anything worth money, they only took the change that was sitting in my passenger seat and my lighter. oh, and they took the two hats i had in there and threw them in the yard across the street from my neighbor. wtf? would someone please tell me what the point of that was?)&nbsp; then i went to my dads baseball game in Groves. i love baseball. fun stuff. so i came home around 11:30, went to walgreens for my mom and came home. then aaron called me around 1 and said to meet him and angela at taco bell, then came to find out it wasnt open. so i waited for him to call me back with a new plan. so i ended up goin to his house for an hour. that was nice. = )  another thing that sucks, that i hardly care about right now because im in such a good mood, but i really do care about overall because it could screw a lot of things up is this school schedule thing. i went to running start monday and found out i lost three of my credits, so when i went back yesterday to find out if i could fix it, the stupid ass. principal is not computer literate so he couldnt access my files, and now i have to wait on attendance to get back on monday to see if i can fix this shit. ughhhhhh. i will cry if this doesnt get corrected.   today i have to make some calls to get my car fixed. aside from that, i dont have anything to do. how lovely. i am going to try to do something with tara so i can gush out all my good feelings to her. i have to do it to someone, or i will burst.   '),(24632,3753019,'26,July,2004','2004-07-26 00:00:00','   &nbsp; today was good and bad. i had an awesome time with tara and her cousin and his friend. we picked up jeff on the way too lol. hooray for the taco bell parking lot smoking spree. we went to running start for school. that alone was sucky, but then i find out i lost 3 of my credits from last year due to absences. what bullshit. they didnt notify me or anything. so now i have to go bawl my eyes out to the assistant principal tomorrow. which i probably literally will, the joys of pms.   our work meeting got moved to tonight. so thats all im going to be doing im sure, unless i hear from aaron. which, the chance is slim as always. we hung out last night, but i really dont talk to him often. and when we hang out, its always us and some other people. which is ok, but i miss my aaron time. lol. i miss having a best friend around, and i dont have one handy. i miss my cj... damn fucking pms. im all upset. ive felt like crap all day.  wow, i love how my 5 year old brother orders me around. as if i dont hear it enough from other people.   i really feel like crying right now, but i dont want to have to re-do my eyeliner again. plus, there\'s no point. i\'ll save the tears for the next time i get sad, since hopefully then i\'ll have valid reason to cry. more than likely that\'ll be the case.  \"we\'ll make the same mistakes. i\'ll take the fall for you.\" - Story of the Year   '),(24633,3753019,'24,July,2004','2004-07-24 00:00:00','   &nbsp; its such a shame that everything has ended up like it has.  i hung out with tara yesterday. that made me happy. she and lindsey came over after they went to the mall. it was nice to see lindsey too. she\'s so cute. after she left tara and i went to the mall for me to get more flipflops since mercy ate my other ones a long time ago, and i also bought a really cute tank top. yay for me. after that we went up to Carino\'s and saw michael, and for me, all the other guys lol. that was a lot of fun again. i think i laughed the hardest when tara started singing \"im leaving on a jet plane\". yeah, good times right there. she tried to spend the night with me, but that didnt go over well at all... so much for that.   i went to bed pretty early. 11:15 i\'d say. woke up to 3 IMs, from 3 really important people in my life. though the IMs themselves werent of much importance. tara wanted me to work for her... but i didnt wake up early enough. the other two were from aaron and sean. my vices. aarons IM was simply \"..\" while seans was more touching and significant. the turmoil will never end. but thats alright. i try to see it as something good and not let my heart be hurt every time i think about it. because it  is  good. and im glad for it.  i called and talked to eddie last night. he was busy though. so he said to call him today, which i might, but not likely. im about to shower and go to a company cook-out of my dads, then back to nederland so he can fix my busted car. lol. well, at least parts of it. good stuff.    \"Yesterday has been and gone   Tommorow will I find the sun   Or will it rain   Everybody\'s having fun   Except me, I\'m the lonely one   I live in shame\"   -Ozzy   '),(24634,3753019,'22,July,2004','2004-07-22 00:00:00','   *sigh of relief that was never expected to come*     &nbsp;  my seanlynn just talked to me. i cried. in fact, im probably not done crying yet. just for now while i eat my chocolate-covered donut. so now 50% of the&nbsp;distress in my life is fixed and better. yeah, im definitely going to cry again. jeezus i miss him. i know, that i can be with someone and be perfectly happy but still wish that&nbsp;he and i could be together. its a sort of \"highschool sweetheart\" situation&nbsp;i guess.&nbsp;a situation i dont want closure from. i\'ll be happy living the rest of my life in love with him. (not to sound like a nut case or anything. you know what i mean.)   news here in boremont... there is none. imagine that! worked last night. i was supposed to work 2-6&nbsp; but our water got cut off and i had to go pay the bill. and then they wouldnt have turned the water back on in time for me to shower and get ready for work, so i had tara go in for me. i was pissed that i couldnt work. but THEN, on my way home from paying the bill, lauren called and gave me her 6-10 shift. yay. im still not getting more than 4 hours a week, but at least im working some.   school in 2 weeks and 3 days. what a disturbing realization. someone should be shot.   &nbsp;  \"hand in mine, into your icy blues, and then i\'d say to you \'we could take to the highway.\'  with this trunk of ammuntion too, i\'d end my days with you in a hail of bullets.  ...and i would drive onto the end with you.\"  - MCR   '),(24635,3753019,'21,July,2004','2004-07-21 00:00:00','   hah, i just got on an old screenname and read the profile... i used to be harshly depressed. which some may say i still am. but it was BAD. i sounded so pathetic. ::embarrassed::&nbsp; yucky.   tonight... i went to bucky\'s for a little. woohoo on that one. then i left around 9:30 and came home. that was a bummer. stupid michael wouldnt come to bmt to hang out with me. and he claims that he loves me! pshT. fucker. but then joel called me back and he got home early, so i went over there and chilled till 12:30. man i havent seen those guys in months. i miss them. paul is my sweetheart and joel is so talented it makes me want to cry. he\'s amazing. and now he\'s started playing piano, which is a weak spot of mine. its so beautiful. so now that we\'ve finally gotten together, hopefully we\'ll do it more often. too bad school starts in 3 friggin weeks. something tragic needs to happen to postpone this traumatic experience, but only to where we still get out at the same time at the end of the year. haha. yeah right.   hopefully mason will have a better day tomorrow. my mom had to stay home from work today to take care of him. he still isnt eating, but he\'s not dehydrated. so thats good. he needs to get better quick, i cant take the squealing noise he makes when he\'s upset or wants something. (he does this because he tries not to speak because of his throat)&nbsp; its so unnerving. not to mention when piper throws in her high-pitched scream of death.&nbsp; wow.   i feel so unaccomplished, still. i do absolutely nothing with my life. nothing. im surprised i even have enough to write about on here. probably because it involves other people, not so much myself. i turned in my application at Carino\'s today. found out i missed getting hired. if only i woulda brought the appl. back within that week. dammit! oh well. hopefully i\'ll get a job there soon. i \"have a good chance\" since i have no life and all i\'ll be doing is going to school until 1. lol. all i wanna do is work anyway. so thats alright.   a lot of things, which i really consider nothing, keep my mind occupied a lot of the day. i think about things that really dont hold much substance in the course of life. but seeing how thats all my life entails, its hard to think of other things. i dont think i made any sense there...   \"i just want you to know who i am.\"  - the Goo  (if you dont know who that  really  is... heh, wow.)   '),(24636,3753019,'20,July,2004','2004-07-20 00:00:00','   for someone who would love you for you?       &nbsp; man, last night was a blast. tara and i went out to eat at Carino\'s. Michael was working, and he basically hung out with us the whole time. it was so much fun. i miss tara. and i miss michael. i love them both with a passion. i wish tara could hang out with me... thats so sad. almost ridiculous, actually. but whatever. people will be people, and i will be the scapegoat when there is no other. anyway, all the other Carino guys were working too. it was good to see them all. im going to put my application in today. last night Eli was introducing me to the manager David and he was like, \"i know who she is. she didnt turn in her application.\"&nbsp; haha. so im going to do that today. i\'ll make sure to give it to David himself. lol &nbsp; this morning mason was having a really hard time with how he was feeling... so my mom stayed home. which i doubt i\'ll be able to leave because she \"needs help\".&nbsp; we\'ll see. i hope mason gets better soon. im sure he will, but this isnt soon enough. its been 6 days now. my little boy... &nbsp; its official, aaron is a jerk. love the kid to death, but he is a terrible, terrible person. haha. ask no questions.  &nbsp; well so she\'ll stop bitching, im off to clean the bathroom. what a wonderful day i have ahead of me. &nbsp;   '),(24637,3753019,'19,July,2004','2004-07-19 00:00:00','   not much going on today. its 12:20 and mason\'s still in bed. poor thing doesnt feel well, still.&nbsp; piper\'s up and chirpy, as usual. she\'s so cute. wow, i\'ve been up for a while, because it feels like it should be later. i wish it was.  &nbsp; i need to go out and look for another job again. i really dont want to though. i have an application to put in at carino\'s. so i suppose i\'ll do that. hostess jessie. bleh. we\'ll see. what i REALLY dont want is for the school to get me a job. because more than likely it\'ll be a sucky one. bleh x 10 haha MsPibb02: i need another job&nbsp; = ( RiDEtheL1GHTN1NG: lol apply at carino\'s MsPibb02: hah yeah i have the application sittin in my car.  RiDEtheL1GHTN1NG: and youre a nigger so you should get the job easy &nbsp; oh yes. michael speaks the truth on all matters. &nbsp; i wonder if eddie and aj are still gunna come see me. they\'d better! they have a week after aj gets outta summer school before i have to go back to school. thats plenty of time. jessie will kick some ass if they back out. hmph. i love my boys.  &nbsp; i had a friend ask me last night how life was, and without hesitation i responded \"life is... slow and complicated.\"&nbsp; which is exactly how i feel. nothing gets accomplished. things just... fade away. that doesnt seem right. the last feeling of accomplishment that i had was when i straightened the panty table at work. and even then i didnt really feel accomplished. i could just say i did because its such an avoided task. pathetic.&nbsp;life is so empty. i know what im missing. im just so worn out, i dont want to do anything that takes a lot of work and thinking. especially if its spiritual. ughh. &nbsp; problems problems problems. &nbsp; \"i am flawed, but i am cleaning up so well.\" - dashboard &nbsp; &nbsp;   '),(24638,3753019,'18,July,2004','2004-07-18 00:00:00','   once again, i am the bad one. and there\'s no being convinced otherwise. i can live with that. someday people will realize that its not always the fault of an outside influence that makes things happen. we are human.  &nbsp; though it probably usually is my fault.&nbsp; here we go again. &nbsp; yesterday i spent the whole day at the lake. well, after noon anyway. i got up at 7:20, took a shower, my dad came and got me around 8:15. yeah, he was late. so we drove to Sam...Rayburn(?) up in the northern part of Jasper. it was so nice. beautiful day, so much fun. im glad Holly was there. i didnt think i\'d spend much time with her, but i did. mostly because i was the only one that would ride with her in the tube. that was fun. i have \"carpet burn\" on my left elbow, my legs got cut up from the tube, though they dont hurt... and i semi-twisted my ankle. (in other words, it hurts when i walk but im not sure why) and my face is burrrrned. it\'ll be gone by tonight though. so i had a good time! rodney damn near killed me and holly on the tube. he was ruthless. i only fell off once, and that was because i let go and leaned out. lol. holly fell out the first time we rode, and i didnt. i dont see how. it was insane. but the time i \"fell out\", that was the worst. it was so rough and he was doing all kinds of crazy stuff. we were airborne most of the time. not to mention it was a long ride because we were going to the middle&nbsp;of the lake to swim.&nbsp; oh wait, i got that confused. but it doesnt matter, so im not gunna fix it. &nbsp; today...  &nbsp; another wasted day. maybe i\'ll get to spend some time with someone though. &nbsp; i love the people that i love. they make me happy.   '),(24639,3753019,'16,July,2004','2004-07-16 00:00:00','   so today was depressing. dont really know why. i just went on a paranoid frenzie and freaked myself out. ironically there was a joke made at the end of the night on the subject matter that i was spazing out about. i didnt like that so much. i wish i could make it known, my valid worrisom paranoia... i just might, tomorrow.  &nbsp;  i hate not being able to control my feelings. especially when i cant control them, and i dont understand where they come from, or why. not knowing \"why\" in any matter will be the death of me.  &nbsp;  there is something else i want to talk about, but im not going to. all of these things make me sad. i dont want to be fucking sad anymore.&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; dammit.  &nbsp;  to hell with this. im going to bed. this didnt make me feel much better.  &nbsp;  &nbsp;  \"though i never never never ever wanted this to be.\"  -CKY   '),(24640,3753019,'16,July,2004','2004-07-16 00:00:00','   i just got done talking to an old friend, bobbi. she\'s such a sweetie. its always nice to \"hear a familiar voice\". im glad she\'s been doing good. it made me think of everyone else in virginia. i got super depressed last night. not exactly sure why. i was feeling pretty lonely. i really miss everyone. i miss how things used to be in my life, minus a few downfalls. i dont have another job yet, but im thinking... if i save up enough money i\'ll be able to fly up for winter break. thats far enough away. we\'ll see. i wish i could just drive up. but so much for having a reliable car.  &nbsp;  didnt do a thing last night. i ate and ate all day and all night. i made myself miserable. almost on purpose. i woke up this morning and wanted to puke. i thought i was going to. this morning, im not in that much of a better mood. only because its quiet in the house and im bored. with nothing to save me from my boredom. something\'s wrong with aaron, and i have to wait for him to call me back to find out what that is. i hope he\'s alright. he felt like crap last night and he sounded terrible on the phone this morning. he\'s my buddy.  &nbsp;  my dad called a few minutes ago and said he\'s going to the lake tomorrow. that would be good fun. so in a little, or when i go to work tonight, im gunna see how the hours are running, because im on call tomorrow. hopefully they\'ll be perfect on hours and wont be able to call anyone in. i havent been to the lake or anything of the likes in a long time. i want tara to come with me. i miss her. ive only talked to her once in the past few weeks, as far as a whole conversation is concerned anyway. we\'ll see.  &nbsp;  mason and my mom are still asleep. its 12:43 in the afternoon. i hope mason is better today than he was yesterday. my little trooper. i love him. i think im going to go clean out my car and then take a shower. do something with my day.    &nbsp;    &nbsp;   \"i want you to know that i miss you, i miss you so.\"   '),(24641,3753019,'15,July,2004','2004-07-15 00:00:00','   yesterday wasnt too bad. i basically babysat until 11 o\'clock at night though. and that sucked. especially since i had some fun to have and i was tardy. didnt like that all too much. i went and hung out, had a good time. except aaron left early cuz he\'s a loser. that didnt suck too bad though. i still had fun. poor josh called to make sure i got home ok and while he\'s playing the good samaritan by doing that, he\'s also cleaning up throwup from two other people. what a saint.   &nbsp; mason got his tonsils out this morning. poor thing. he\'s such a trooper. he\'s been sleeping since they got home around 12 or so. my mom said he was very brave. and i\'ll bet he was. i love that kid. i miss him already, and he\'s only been out for a day lol. soon enough i\'ll wish he was knocked out on drugs again. too bad i dont have them at my disposal. aaron and kenneth have been over here hangin out. they went and got mason some toys. so sweet. aaron\'s good with the kids when he\'s not acting too much like one. and kenneth\'s a cool kid. they\'d better come back soon, because im bored and not allowed to leave.  &nbsp; hopefully as the night comes around, which yay it\'s already 5, i\'ll shower and then go out. she\'d better not keep me here for the night too. that\'ll make me angry. it\'s driving me insane being home all day every day, again. and then getting yelled at because the house isnt up to par on being neat. its so tiring doing the same thing all the time. i dont mind cleaning up, etc. but when that\'s all you do, you tend to not want to do it. *sigh* &nbsp; i called ashley the other day... didnt get a call back. i wasnt completely discouraged, but it happens most of the time when i call. its ok. distance sucks. &nbsp; now im off to clean the kitchen again. and bring my mom some pizza and a coke. she just called me from her bedroom.   '),(24642,3753019,'14,July,2004','2004-07-14 00:00:00','   another wasted day. (yesterday) and yet again another one today. i wish i could do more with my life. i sit all day, and do nothing. i cant even go out and look for a job anymore. by the time my mom gets home its 7 or so... by then im tired anyway.  so mason gets his tonsils out tomorrow. i hope that goes well. =(  poor baby. i need to get my wisdom teeth out. yay. i remember when jacob had his out. it was a pathetic sight. in stephanie\'s blog she mentioned how he looked like a chipmunk, but she failed to mention (probably because she wasnt there for it, but) how he would run to the bathroom with blood dripping out of his mouth, while i follow behind with a rag cleaning up the floor. he was in serious pain. i hate having pain in my mouth. i hate having people dig around in my mouth, so i definitely am not looking forward to the adventure of having someone cut my teeth out of my gums. who the hell decided we should have our wisdom teeth removed anyway!? stupid ass. they, apparently, had enough room in their mouth for theirs.   i sorted things out with someone last night. its sort of an \"again\" situation. but not really. i think things are as good as they are going to get as things stand right now. i feel bad that i was a catalyst to the situation that spurred the failing of a relationship. but im not sure that the outcome would have been much different, for had i not come along, certain things wouldnt have been realized within the nameless. so maybe it\'s for the best, and her relationships will go smoother from now on. i really do wish her the best. i think now that she\'s seen these things about herself, things will go much better for her. i really hope so. i have an indescribable caring feeling for her. i dont know where it comes from, or why. but i do. and i have no problem with that.  today holds nothing for me. complete and utter nothingness. its almost sad how i can sit here and think about what the day might hold, and i cant think of one damned thing.    \"Dirt deep beneath my fingernails, I\'m gripping to the floor I\'m searching through the world, always looking for something more\" - Hidden in Plain View   '),(24643,3753019,'13,July,2004','2004-07-13 00:00:00','   what a night. goodnight. = ) i love all   '),(24644,3753019,'13,July,2004','2004-07-13 00:00:00','   so this morning at the brilliant hour of 4:45 the phone rings. (i have the house phone in my room in case the call i was waiting on wasnt to my cell phone.) anywho, i dont answer it because its a blocked call.  then  my cell phone rings with a private call. holy crap. i answer it, and just as the last blocked call that i answered on my house phone (at frickin 9:30 in the morning, mind you) its someone moaning and breathing deep like they\'re... well you know. so i hung up. THEN THEY CALLED AGAIN. so i picked up and was like \"stop fucking calling me!\" and well, they didnt call back this morning. i hate that i somehow managed to meet a big sicko that has my home and cellphone numbers. i know its not one of my friends. i think i might know who it is. what a freak.  worked last night. that was painful. my back and my knees just cant handle walking around a lot... sounds weak i know, but i cant help it! sure when its time for  my  massage someone has to go home! bleh. well, i dont work again till possibly friday anyhow. even then, im only on call. how ridiculous.  i miss tara. she\'s been with lindsey all the time now since lindsey doesnt have a boyfriend anymore. that doesnt seem fair. i know they\'ve been best friends for years while i was gone, but i was still a best friend... or so i thought. just kinda hurts. not to mention that i have a terrible history with people in general. specifically those that were once close to me. it hurts, but not NEAR as much as it would had it happened about a year earlier. im used to feeling like this. but i just didnt expect it, i guess. not with tara. i know i havent called her to hang out either, but she\'s been with lindsey all the time and she doesnt invite me, because i know lindsey doesnt like me. so i figured if maybe there\'s a day that she\'s not with lindsey, she\'ll call and want to hang out. but probably not. plus, i have a feeling tara isnt into hanging out and doing things we used to anymore. knowing there\'s manipulation and persuasion involved, tara probably thinks poorly of me now also.  every time im on here i think of stephanie. probably because she was the first person i knew to have a blog, but also because i know she feels how i feel quite often. and i love and miss her. and i wish she was here for me to talk to. what a beautiful soul.    \"for all of this, im better off without you. do you regret so your loneliness?\" - the early november   '),(24645,3753019,'12,July,2004','2004-07-12 00:00:00','   My Day: woke up, hung out with aaron and TJ while babysitting. which turned into me babysitting three boys (5, 17, and 21) and one girl (3). the youngest two of the four being my siblings of course.  i just finished doing the dishes, vacuuming, and making brownies. yum!  more than likely i wont be eating any of them. seeing how all of the mentioned above plus my brother and my mother will be around here sometime or another tonight. and i have to go to work in 30 minutes! damn.   fortunately i could somehow \"talk\" my brother into babysitting for a whoppin hour so i could make it to work in time. i dont see why he is so against doing something for someone other than himself unless he gets money out of it. its an hour, what does he want, a whole dollar? he hasnt seen the kids but once in like a week in a half, if not longer. im glad aaron and TJ are gunna come back and watch the kids. they\'ll be nice to them and play with them. hopefully jacob\'ll get over himself after half the time he\'s here and do the same.  i miss my va homies. i would call more often, but i guess i just assume (through experience) that i wont catch anyone. hell with it, i\'ll call anyway. i love them.  welp, im off now to get ready for work. WORK!!! im glad lauren calls me when she needs someone to cover for her. i need the hours.   '),(24646,3753019,'11,July,2004','2004-07-11 00:00:00','   havent been up to much at all. sitting here trying to think of what i\'ve done in the past few days, i cant really remember. probably because its the same thing every day. not that im complaining completely. i have a good time. i dont know. it always seems like its one or the other. (cant have it all)  wow. school starts way too soon. how depressing. not to mention i dont have another job yet. crap.  adding onto my latest entry... dont always assume you know what im talking about. i have a lot of similar situations in my life, and you\'d be arrogant to think you are safe in considering yourself correct in your assumptions. like i said... im vague. for a reason, i think.  not up to much today either. mom\'s not working so i dont have to babysit. but what else is there to do? i took a shower. that counts for something, right?   \"And I held on to what you gave and soaked up every smile to fall asleep with your memory and pictures of a child.\" - Shady View Terrace   '),(24647,3753019,'10,July,2004','2004-07-10 00:00:00','   made myself eggs for breakfast. yum. i feel so accomplished.   crazy crazy stuff goin on. i hate not having control. or not so much that im not in control, just that when i voice that i dont believe something is going to occur as far as myself is concerned, and then it does, i feel badly or two-faced or whathaveyou. hmm, i duno.  notice im fairly vague in what i say. dont analyze. im not trying to put my whole life out there for everyone to read. i keep the details to myself so i can still feel like i have a personal life. i like to talk. i dont like to tell.  what to do, what to do. everything always has terrible timing. timing is key, and timing in my life SUCKS. because then i have to think about things and act upon impulse or desire and \"work things out\" only because the timing and order of events. damn the inevitability of things.  i want to say, that i hope im not hurting anyone. certain aspects of my life somehow seem to interfere with other people\'s lives, and i dont mean for that to happen. truth be told, i dont put myself in every type of situation that i end up in. and neither does anyone else. give the benefit of the doubt. its the least you can do.  anywho... ive been hanging out a lot lately. its nice. i like having friends. they make me happy. to just sit around and enjoy each other\'s company... i couldnt ask for more. then again, i wish that some people were there, that arent. all i have left to say is \"oh well\". i try. i know i try.   '),(24648,3753019,'07,July,2004','2004-07-07 00:00:00','   MsPibb02: i was like half asleep and, i wanted food and something to drink before i went to bed, but i didnt get anything. and then icecream popped into my head as i was falling asleep. so here i am, eating  icecream. not a bad idea huh?  so, there was the marking period of the end of my night. basically. at this moment while im making my post im talking to the all-wonderful michael. and thats always fun. i love michael. all the time.   hung out tonight, im gunna miss amaad. its a shame we didnt get together sooner. i guess i\'ve been somewhat anti-social, as far as hangin with new people here since i moved, in a sense. i think im becoming more outgoing. i like it. its fun.  my apologies for the squiggliness (ha, there\'s my made-up word for the night Amaad. it beats \"unawesome\") of my thoughts. there is reason for that... good reason.  today the lesser became the better, while the better became the somewhat lesser.    ironic what can all can change in one day.  i like having someone that likes you for \"no reason\" at all, as far as you see it. but they give you reasons upon reasons. and to know that they  believe  they have  valid  reason... thats what feels good.  its nice.   haha.   '),(24649,3753019,'05,July,2004','2004-07-05 00:00:00','   yesterday as i was walking up to my dads apartment, i passed one of those huge black and yellow bees. it was struggling with all it had to get off of its back. i think someone sprayed it the day before because i thought it was dead. as i walked passed i felt sympathy for the poor bug. none of Gods creatures deserve to go through such unnecessary torture. as i was passing it again, i\'d decided that on my way back up from getting things out of my dads truck, i would simply flip it onto its legs and hope that it would be well enough to take it from there. on my way back up i saw that it had already somehow managed to do the task itself, and i felt bad that i hadnt helped it soon enough.  i dont really know what i relate this story to, but its just one of those little things that made me think. you could look at it from many angles and apply it as necessary.  well now that i\'ve given my unexplainable blabber, last night i didnt do anything. i was supposed to go to the beach with michael, but he didnt end up going till around 10 or 11 and by then it would have been late and pretty dangerous to drive. so i decided to just go watch fireworks with jayme and then go to lumberton. neither of which i did. we just chilled at jaymes place for a while... it was ok with me. i was tired. its not like i cant participate in the \"celebrating activities\" any other night.  im back at my moms for the summer so it seems. she needs me. hearing this motivated tear shed after i got off the phone. it was sudden and even shocking to me. but i guess it upsets me more than i thought.    heh...   '),(24650,3753019,'04,July,2004','2004-07-04 00:00:00','   im so ready to go! hopefully michael and them are still doin the Galveston thing. because if they are, im there! i havent done anything for the 4th in years... hopefully this goes well. if not, ive got backup. so thats good. not that i look at my options as \"backup\", but you know.  so i went and saw Dodgeball. hugely disappointed... it was corny more than anything. a few giggle parts, but aside from that, its just another movie that got talked up a lot.   again, im only working one day a week for the next 3 weeks. they\'ve hired wayy too many people. i resent that after being told i\'d get more than 4-8 hours a week when the summer picked up. i have yet to get more than 8 hours a week. job please?  havent talked to my \"friend\" since that night, but i did talk to... someone else. and thats still that... we\'ll see how this thing goes.   do you ever feel like your life resembles a dripping faucet? you sit and watch the droplets of h20 form, see how long it takes for it to fall, and then see if it drips again, or if it\'s going to decide to stop...  i miss my VA homies, i hope everyone is doing fine. im gunna make a few calls the next time im at my moms. i did the other day. so much love.   '),(24651,3753019,'02,July,2004','2004-07-02 00:00:00','   wow. last night rocked. i hung out with  my michael  for the first time in months. i miss him terribly. we hung out with some people he works with, and they\'re all majorly awesome. i met this guy... he claims he\'s a dork, but he\'s just like me in so many ways. i guess that makes sense. hah.  then to come home and see that the \"friend\" that randomly decides not to talk to me anymore - called me. so i called back, (to see what the hell he wanted) and we talked until 3 something this morning. impeccable timing. i dont like anything he had to say. because i know im going to hang out with him again. possibly. and he thinks its going to be like it was when he decided to leave, and its not going to be. please dont let it be...  hopefully the cool guy will call me today and we\'ll go see Dodgeball like we had planned. granted he was a little... under the influence of a certain substance, but he was still making sense. haha. its the best to make plans with people you dont know when you\'re intoxicated or what have you, anyway. funny how we made these plans before we even really started talking. awesome.   babysitting again today. working close. oh the hours im rolling in! frickin... hopefully a good night.  = )  we shall see.   '),(24652,3753019,'01,July,2004','2004-07-01 00:00:00','   i love good music. keeps me alive.  so yesterday was a good day. minus about an hour, which is when i discovered its not safe to drive fast on the interstate while crying. or while being just plain hysterical. i was hurting pretty harshly.   cigarettes are bad.  one of the places i applied yesterday is hiring for part time, so i was told. i have to wait till later though to talk to the hiring manager. i hope that goes well. things are so weird right now. with all thats going on between myself and various people... and the fact that im actually trying to make plans for my life and decided what im going to do for me... i guess im just not used to there being room for things to change in my life or for it to be allowed that i make decisions for myself as far as what i want in life. its so strange.   \"if i could take your pain away, i would scream for you. and i\'ll bleed for you, so you\'ll never feel this way again.\" - Hidden In Plain View they\'re awesome. check \'em out.   '),(24653,3753019,'18,August,2004','2004-08-18 00:00:00','   so, i havent been to school all week. i started getting sick monday, and by the end of the night i felt like a ball of shit. so i didnt go to school yesterday, and just to play it safe i didnt go today. which im feeling better, and am now wishing i would have gone. but i\'d rather get better quicker and have missed one more day of school than be sicker for another week and possibly miss more school that i have already. plus, i dont want to feel the slightest bit sickly on saturday. DAVE MATTHEWS BABY!  though it is quite boring, being at home alone all day. my oh-so-concerned mother left with nothing but a demand that i do all the laundry by the time she gets home. of course this was after she woke me up at 7:30 to make the kids lunch and get them dressed and ready for school. i love having an irresponsible parent.  i\'m realizing that for once, i think i\'m content with who i am. i havent been longing to be someone else, or so much like someone else in a while. i havent wished to have someone elses life, looks, or talents. not to say there arent things in my life i wouldnt change. (well, really, the only thing i would change right now would be for tara\'s parents to let us hang out, and then to fix all of this school crap) but i feel happy. now i only get upset over stupid things. like if i dont get to see aaron at all for one whole day, or if he doesnt call me after hours of being preoccupied to let me know what he\'s up to. that and the fact that my mother still doesnt do much around the house and always leaves it to me to do things. nothing significant though.  rarely will boredom drive me to do the things i am told to do. which even today is not the case. but i am going to go ahead and do the laundry. i dont feel like getting bitched at today. i dont have the energy for that and then work at 6.   '),(24654,3753019,'16,August,2004','2004-08-16 00:00:00','   i have a mosquito bite on the inside of the bridge of my foot. and it itches like hell. i can hardly walk on it sometimes because it hurts. aside from that and school, nothing else sucks right now.  except i didnt get home until 2 last night because we were at the emergency clinic with Mercy. yesterday her ribs started swelling, then it went to her stomach. she was huge. so when me and aaron got home last night we all decided she needed to go to the vet right then. good thing. the dr. said if we waited till this morning, her stomach may have, i forget the word, but... all but exploaded. her stomach was the size of a football, almost. so now she\'s at her normal vet today, and she\'s gotta stay there all day. i stayed home from school in case someone needed to pick her up or something, cuz mom\'s at work. last night i thought i was close to losing my baby. thats the worst feeling ever. anyway. this happened because she ate a tampon (somewhat funny) and it got stuck, blocking the passage from her stomach to her intestines. she\'s such a tough girl though. she didnt act like she was in pain or anything. boy, does she fit into this family or what?  so DMB this saturday!!!!!  tara\'s rents arent really letting her go. that sucks. but im really excited about who i did get to go with me. im glad i thought of it. casey moved to houston not too long ago and was supposed to come down this weekend, but didnt. then it clicked, that she could go with me! so my mom told linda to ask her (linda is my moms best friend, casey is her daughter) and she said she\'d love to go and that i can stay the night with her saturday night rather than driving home at midnight. its gunna be an awesome time. so so excited.  well the deal was, if i stayed home from school, i had to clean up the house. and its 12:47 right now, i got up at a quarter to 12, and havent done anything yet. i\'d better get.   '),(24655,3753019,'05,August,2004','2004-08-05 00:00:00','   i havent posted in a while. havent really cared to. i actually dont care to right now, but i figured i would for those who dont care. i normally post to get things off my chest and give myself a therapy session. but lately i\'ve been doing really good. im happy every day. so i dont feel the need to get on here.  i\'ve wanted to avoid posting this because i feel bad that what i swore to one person wouldnt happen, did. but its ok, because apparently they werent meant to be together, and it\'s been long enough anyway... my point is, aaron and i are together. i guess we have been for a few weeks, but we actually just talked about it the other night. i\'ve spent every night with him since friday, minus sunday night. finally he stayed the night over here last night, and gave me a break from having to go to his house lol. i love it that my mom loves him. she told him that this morning. haha. i was still half asleep so i dont really remember why she said that. she\'s leaving tonight to pick the kids up from FL so when he was leaving this morning she was telling him to take care of me and make sure i do blahblah, and he goes, \"well i dont know about that, but i\'ll take care of her.\"  then she was all, \"i love you aaron. = )\"   its funny/cute.  i got my hair cut the other day. its cute. but i wish i could get it to fix like debra did when she cut it. ohwell. i\'ll work with it. i thought aaron would shoot me and tell me it was too short. but he keeps telling me he likes it. so thats good.  well, i think that tara and i are officially not hanging out anymore. her parents dont like me in the least and refuse to allow their daughter to grace me with her presence. that makes me cry. tara is my best friend and i love her so much. i know i have aaron as well, but it is nice to have a female friend. i dont like to have many, but i like this one, and i want to keep her. *sigh* i guess we\'ll see whats going to happen. either way i doubt i\'ll ever show my face around their home again. this is just too harsh.  i\'d keep going about what\'s been up for the past week but its all pointless anyway and im sure not one person gives a rats ass. so im gunna go.   '),(24656,3857776,'08,July,2004','2004-07-08 00:00:00','    I am making a little section for animals because well I am a fan of them, and I am hoping to work with them when I\'m older, for a job. So every week or two weeks I will be telling you some sort of animal thing that is going on in the world. Example: Every day tons of (water habitated) birds and other animals are dying because of pollution. Most specifically oil pollution. Its like this- a stupid ship or boat will dump their oil into the water, all or most; animals, fish, creatures, etc., will die because of the oil that ship or boat dumped. The birds, will also die, because when they go to get fish out of the water, they cannot tell whether the water looks different or not because first, it is too dark to tell, and second THEY ARE BIRDS, THEY AREN\'T THAT SMART. So they will go diving in the water to get fish and they will get stuck in the oil, their wings will get coated in the oil, so they can\'t fly, so they die. This one section was mostly related to pollution and why we shouldn\'t pollute the world. And also some oil that is left by ships, is not dumped on purpose, there is sometimes simply a leak. I know that isn\'t their fault, so if that happens there is nothing I can say. But sometimes ships pollute the water on purpose, even though it is illegal. And they might not even get caught, so you have a ship who is going around polluting the waters, and you don\'t know who it is. That\'s when you have a problem. So I made this section to maybe make people to stop polluting the world, or make other people, make the people who are polluting the world, be stopped. So please try to make a difference. Help stop pollution. That\'s my section for this week and please feel free, once after you have read this, to e-mail me what you think at Nicklachey1lover@yahoo.com  Submitted and made specially by: Ellie '),(24657,3857776,'06,July,2004','2004-07-06 00:00:00','    Strawberry Angel Food Dessert        \"Angel food pieces are topped with sweetened cream cheese, whipped topping and strawberries in glaze in this chilled, layed dessert.\"     Yields 1 - 9x13 inch dish      Prep time: 15 minutes Ready in: 15 minutes        INGREDIENTS    1 (10 inch)angel food cake 2 (8 ounce) packages cream cheese, softened 1 cup white sugar 1 (8 ounce) container frozen whipped topping, thawed 1 quart fresh strawberries, sliced 1 (18 ounce) jar strawberry glaze      DIRECTIONS       1.    Crumble the cake into a 9x13 inch dish.    2.    In medium bowl, cream, the cream cheese and sugar until light and fluffy. Fold in whiped topping. Mash the cake down with your hands and spread the cream cheese mixture over the cake.    3.    In a bowl, combine strawberries and glaze until strawberries are evenly coated. Spread over cream cheese layer. Chill until serving.     Makes 18 servings      Submitted by: Ellie fox Presubmitted by: Kelli '),(24658,3857776,'06,July,2004','2004-07-06 00:00:00','    In Theatres     1.King Arthur 2.Stepford Wives 3.White Chicks 4.Spiderman 2 5.Shrek 2 6.The Notebook 7.Sleepover 8.Harry Potter 3 \"The Prisoner of Azkaban\" 9.Day After Tomorrow 10.The Chronicles of Riddick 11.Garfield 12.Dodgeball: A True Underdog Story 13.Two Brothers 14.The Terminal 15.Anchorman     Coming in theatres    1.A Cinderella Story              (The rest of the movies will be updated later) '),(24659,3857776,'06,July,2004','2004-07-06 00:00:00','    \"A good friend can tell you what is the matter with you in a minute, he may not seem such a good friend after telling\"  - Arthur Brisbane    I think that, that quote meant that a good friend can tell you what\'s a matter with you in a minute, but they really aren\'t such a good friend though, if they are saying something is a matter with you. I\'m not exactly sure as to what that quote means, so if you think it means something else, feel free to e-mail it to me at    Nicklachey1lover@yahoo.com '),(24660,2490703,'12,December,2003','2003-12-12 00:00:00','   Friday, December 12, 2003    Hey, I guess it really has been a long time since the last time we talked. I haven\'t really been up to much lately. I\'m starting indoor track this January. It sounded fun and I knew it would be a really good way to keep me in shape. I\'ve been really injured from cross country, so much so that it kept me from running most of the cross country season. I have physical therapy for 6 weeks. It\'s not so bad, it just hurts a little sometimes. But I guess it\'s helping me out in the long run.   Jason and I broke up in the beginning of November because things weren\'t working out for us. He blamed the whole break up on me. It was really hard on Mr at first and I started doing stupid things like cut myself. He hated me for like two weeks. He even wanted to go out with some new chick a week after we broke up. I seriously was so depressed that I massively cut myself. I just couldn\'t take it. We were such good friends and then we weren\'t. I think I would have been okay if there hadn\'t been some new girl. That was the part that slapped me in the face. I thought that maybe he never loved me and he only wanted to stuff from me. But then like 3 weeks after we broke up, he told me that he missed talking to me. Then like the week of Thanksgiving I went over to his house and he and I were talking. He asked me why I cut myself and started to make fun of me. Then out of no where he told me he loved me and by then I had a new guy...Just joking we started unofficially going out again. Then recently we made it official. I promised him that I wouldn\'t cut anymore, and I haven\'t since then. It\'s been two and a half weeks since the last time I cut myself. I\'m pretty proud. Unfortunately, there is a Mr. X...Ross. He\'s a real cutie and he\'s in a band. So is Jason, but Ross\' band is so much better! :) He\'s a sophomore, but the hottest one there is! No joke. I\'ll find a picture of him for you.  Other than that I really haven\'t been up to much. I really don\'t even go to church anymore. I just stay home and sleep. I really don\'t know what\'s come over me. I just don\'t want to go. I think it\'s because all of the people at church don\'t even know I;m still alive, nor do they honestly care if I come or not. I guess feeling like I\'m not wanted has an effect on you. I think things have really changed about me. I\'m a totally different person, and in more of a good way than bad. Minus the whole church aspect.  Anyway, I\'ll talk to you later. I love you like a sister...I\'ll see you next week, can you believe it. We\'ll have to get together and do something!                                              Peace and Love,                                                    Melissa  '),(24661,2490703,'31,May,2004','2004-05-31 00:00:00','   I love Mr. X but JOHNNY DEPP IS SO HOT!  '),(24662,2490703,'30,May,2004','2004-05-30 00:00:00','   I love the rain! I just hate when I\'m out being a band fag and marching and it starts pooring! I was out marching, as we march into the cemetery, it starts pooring!. It would have been pretty awesome, but we had to sit out there for a really long time before we could get in someplace warm and dry. Luckily, that was the last parade that I will ever march in again. I don\'t mind band, I just wish I played something better...like percussion or guitar. The wierdest thing has been happening lately. Teffticles has been acting wierd around me. He broke up with Justine like a month ago and now he\'s totally different when he\'s around me. It\'s cool because we can actually talk to each other, but I think he\'s kind of using me. Whatever, I have another person to hang out with now. Anyway, I have to jet, the phone is for me. I\'ll talk to you later!   Peace and Love,  Melissa  '),(24663,2490703,'30,May,2004','2004-05-30 00:00:00','   Sunday\'s...They are such boring days. I hate that it\'s raining...BLAH! But other than that I\'m happy that I don\'t have to work and I can sit around and do nothing all day. I totally can\'t drive for a while. My dad is being gay again, God, I hate him so much. You\'d think he\'d realize that I don\'t like him, and the only reason I can\'t wait to leave is so I never have to talk to him again. I told my mom that I don\'t want him at graduation again today. She got all mad and gave me the \"but he\'s your father Melissa...blah, blah.\" I honestly don\'t care. I hate him and I hope he disappears. It\'ll make my life 100% better. I\'m sure if he actually disappeared I would feel bad, but for now...nope. It reminds me of when Amanda Pringle was telling me this one story about Ryan S. OMG. He was supposed to help his dad unload this four wheeler off a trailer and he was screaming at Ryan to get down stairs and help. The whole time Ryan is thinking, I hope he has a heart attack, I hope he has a heart attack. All the sudden his dad\'s like, \"OWWW!!\' Ryan\'s like, Woo Hoo he\'s having a heart attack! Then his dad\'s like \"Oww, my leg.\" Ryan\'s like damnit. Hey, his dad sent him away and he didn\'t have to help. That cracks me up. I hope he has a heart attack definitely makes my day. Dag, that kid is so funny! Well, I\'m out!   Peace and Love, Melissa  '),(24664,2490703,'29,May,2004','2004-05-29 00:00:00','   This is Jenny\'s story. Read with caution...  Jenny walked into her house, went up to her room and turned on her computer.  She was home alone. It was 5:30 PM, Jenny had nothing planned until later that night so she signed onto AOL messenger with the screen name PrincessJ00 and checked to see if anyone was online. No one was online, so she put on an away message and went downstairs to grab something to eat.  She heated a slice of pizza from the night before and went back upstairs.  Someone IMed her.  Hmm, she thought. The person was LuvMe4Life58. She had no clue who that was, but she talked to them anyway.  LuvMe4Life58: hey  PrincessJ00: Um, hi? Who is this?  LuvMe4Life58: You don\'t know me, but I sure do know you.  PrincessJ00: What?  LuvMe4Life58: You\'re pretty. I like your blonde hair and your braces...Don\'t worry, I think they look good.  PrincessJ00: Wait a second, who the heck is this?  LuvMe4Life58: You don\'t NEED to know who I am. But....  PrincessJ00: (getting freaked out) But what? Huh? You don\'t scare me! You\'re just TRYING to but but its not working!  LuvMe4Life58: It\'s ok Jen...I won\'t tell CHRIS that you like him. I\'m the only one that knows. I read it in your diary. Just tell me that you\'ll meet me and everything will be okay.  PrincessJ00: NO! Stop it please... I know this is you Karen! I hate you! Please stop it! (Getting nevus and freshet out, because she didn\'t tell anyone that she liked Chris)  LuvMe4Life58: This Anita Karen. But I know that Karen is your best friend. And you had a crush on Jeremy, too when you were in 5th grade. I read it in your diary. You hide it in your desk. Left hand drawer, the one at the top. I was  in your room today when you were at school. I left you a little surprise in your closet. Just take a little peek Hun. Don\'t be afraid, I promise you won\'t regret it.  PrincessJ00: I don\'t believe you. NO ONE knows where I hide my diary! STOP please!  LuvMe4Life58: JUST LOOK FOR GOD\'S SAKE!  PrincessJ00: NO!  LuvMe4Life58: NOW OR ELSE I\'LL GO IN THERE RIGHT NOW AND KILL YOU WHILE I\'M AT IT!!!!!!!  PrincessJ00: OMG! Please stop!   LuvMe4Life58: No, I won\'t kill you. But I just wanted to let you know, if you don\'t look in your closet- your parents won\'t be coming home tonight!  PrincessJ00: FINE! If you want me to, just PLEASE don\'t hurt them PLEASE!  LuvMe4Life58: Thata girl. A little bit closer. I see you- you know that right? I\'m outside your window and you look great in that white blouse.  PrincessJ00P: Stop! PLEASE! Who the heck are you?  LuvMe4Life58: LOOK in the closet right now!  PrincessJ00: I\'m going, please, just leave me alone!  As she walked closer, her heart beat so hard she couldn\'t control it. She held onto the doorknob to the closet, twisted it, pushed it open and saw......  THE DEAD BODY OF CHRIS HANGING FROM THE CEILING OF HER CLOSET! Horrified, Jenny ran to the phone and tried to call the police. As she picked up the receiver, a hand grabbed her from behind and said, \"you don\'t want to do that.\" The voice then turned Jenny around and she saw it was Michael, the quiet boy in her math class. I\'ve been watching you Jenny, and I love you. Jenny struggled to get away but Michael held her tightly. He threw he down on the couch and tied her up with duct tape. \"I said that I love you Jenny and I\'ll do anything to have you. Screaming, Jenny says, \"NO, NEVER! LET ME GO, NO, STOP!\" Michael knocks her out and says, \"If I can\'t have you, no one can!\"  When Jenny wakes up she feels really weird and she doesn\'t know what happened. She sits up and realized that she doesn\'t know where she is. She looks around for her clothes and finds nothing. Scared Jenny tries to move but he body is too weak to move. Jenny tries to call out, but nothing comes out. She blacks out again.   When Jenny awakes, she sees a man standing over her wearing all white. She tries to call to him but he puts a hand over her mouth and tells her to be quiet. The man picks Jenny up, wraps her in a sheet, and walks down a long hallway out to a dark car. Jenny was never seen again...  Be careful who you talk to...This could happen to you.  **Isn\'t this a creepy a story?***  '),(24665,2490703,'29,May,2004','2004-05-29 00:00:00','   It pisses me off so bad when people bash Paris Hilton. If you looked like her, wouldn\'t you flaunt what you\'ve got. HONESTLY! Oh and gay kid, I don\'t want to be Paris, I just hate my real name, it\'s so common and that\'s so not me. SO, I HOPE YOU DIE! I\'m just joking, but back off Paris, she\'s beautiful.  '),(24666,2490703,'28,May,2004','2004-05-28 00:00:00','   Today has been such a boring day. I haven\'t been up to much at all. I honestly slept all day. It was kind of nice. Last night I went to a \"get together\". A lot of people came and left. There was a party at Mark\'s house but people weren\'t allowed to drink. So people from the party I was at went to Mark\'s party and told them that there was a keg at our party. It was crazy. People were getting mad that there wasn\'t any beer. There was a lot of smoking going on. I didn\'t smoke anything because I was the designated driver. Besides, I really want to give that stuff up. I got home really late, like 2:30, and I got grounded. So I\'m home today and my parents took my car keys for the weekend. It sucks so bad. I need to get out of this hell hole, my parents are driving me crazy. This is the last summer that I will have to deal with them, then I will be free! I honestly can\'t wait.   I leave for CMU the second week in August. My official orientation is on June 9th. It\'s so awesome, I can\'t wait! I\'m actually kind of tired so I best be off. I hope you have a great weekend! I\'ll talk to you later! Love you lots!                     Peace and Love,                        Melissa  '),(24667,2490703,'28,May,2004','2004-05-28 00:00:00','   Last Day of school! Woo Hoo! I\'m really excited, it\'s been a long road that hasn\'t been a smooth ride. Looking back on the past 13 years of my life, I have changed so much. I\'m not the sweet child that I once was. I\'m going to miss the security that high school brings and all the friends that I\'m never going to see again. It hurts so bad just to think about it like that. Oh, the horrible crushes that I\'ve had on younger kids, ones that will never be made known. Regrets, If that\'s one of the only things that I ever got out of Mrs. Fraziers class, I have learned that living with regrets isn\'t any fun. Live life to the fullest and set forth goals that you want to achieve and achieve them. I wrote an essay on that today. I\'m going to miss Mrs. Frazier\'s class, Honors English was hard, but the year FLEW by so quickly. It seems like it\'s going slow, but then you\'re done. It seems like yesterday I was a freshman looking forward to starting high school and making new friends. I\'m sure College will be fun, but the reminiscing will never end. It seems like yesterday that you and I were close together and I also remember the day you left for Ohio. I cried so hard, I knew that I was going to miss you so much, and I still do. I about lost it when they played our class song over the PA and you walk out into the commons and there are a bunch of people crying...Guys and girls that you have known your whole life pretty much and they are gone. It\'s just incredible. I taped a little bit of the madness and the tears. The last day of high school is a crazy time because people tell others how they really feel. It\'s an emotional day. It\'s a day I will never forget. I counted down the hours, but when it was all over, I cried my eyes out. Cleaning out my locker I couldn\'t believe it, you know when you know things are going to change and you get that feeling of hurt like you don\'t want to let go, but when you do, you know it will be for the best. I felt that way today as I opened my locker for the last time. I can\'t wait for graduation.   The Missionaries came over today but my parents weren\'t home. I talked to them for a while and they were really fun. The new one that we have is really cute, I shouldn\'t say that, but he is. I\'m such a dork. He told me he hoped to see me at church, Maybe they know something. Like I\'m a bad kid now. I\'m trying to be the \"perfect\" kid again, but it\'s really hard to give up those bad habits. I\'m sure with a little bit of effort, I will.  OMG! Yesterday the scarcity thing happened. I was driving home and My tire exploded. The car tiped a little but I just thought it was flat so I coasted to the stop sign to see how bad the damage was. I got out and I was so scared. I don\'t have a cell phone right now, so I\'m stranded in the middle of nowhere with no phone and no where to go. I prayed that someone would help me and a man stopped and asked if I needed a phone. I started crying. I called home and no one answered. So I called Amanda Pringle because she was like 5 minutes away. I was so scared I cried when I heard her voice. She drove me home and my mom and our neighbor came and fixed the tire. I was seriously so scared. Luckily all went well and I didn\'t die. But it was still scary. But I have to go, I have some work to do for the open house! I love you like a sister!                       Peace and Love,                          Melissa  '),(24668,2490703,'26,May,2004','2004-05-26 00:00:00','   Bag-heads...OOGLY BOOGLY! AHAHAHAHAHAHA!  Hey Howe, I finally got your blog...took long enough. I can\'t believe that I only have 2 days of school left. Crappy thing happened though. We had this really funny senior prank that we were going to do, and someone told on us. I was so mad. We were going to get hundreds and thousands of bouncy balls and put them in a hallway and shut the doors so when people came out of class they would be like OOH, bouncy balls. But no, now the teachers are all \"prepared\" for it and if they catch people doing it, they can\'t walk at graduation. Isn\'t that SO gay? I was pretty angry. But I guess I\'ll have to do something else...I have like a day to think about it. I wonder if they have a senior prank.com...I should look. I\'m sure there is, there is EVERYTHING on the internet. Hm....What else has been happening? Did I tell you about that kid that I see everywhere? He is so hot...AHHHH I want to know his name. I\'ll figure it out someday. OMG! I have the funniest thing to tell you. Yesterday Amanda Pringle and I were driving home from school and we were waving at random people. It was so funny. We\'d make wierd faces as we were waving it was so much fun. While we\'re driving I\'m like we should get bags and put them over our heads then go out with them on and wave at people. We then got the bright idea to cut out eye holes and draw on faces. So we make these bags...we called them bag heads. We go out in my truck, Amanda is the only one wearing a bag on her head, and start waving at random people. It was so funny. They would give her and I the wierdest looks. AHAHAH. So we decide to go to our friends houses and do crazy things in the bags then run away. So we go to Laura Gadzinski\'s house...she wasn\'t there, but her brother Steve was...it was so funny. He was like Melissa you are so wierd. After, Amanda had to babysit, but her car was broken, so we drove around until we had to pick up one of the kids and go back to the house to watch the kids. After we\'re all done we go to Kevin Melvin\'s house. I park the truck where he can\'t see it and start walking up his driveway with these bag-heads on. We get to his house and he didn\'t know who it was until her and I started laughing and chanting oogly boogly....oh man. It was so much fun. What would you do if 2 people walked up to your house with bags on their heads chanting wierd things like oogly boogly smoogle? Oh man...It hurts so much from laughing so hard. I honestly had so much fun being SO stupid! Well, I have to go 39 page review packet due Friday that I haven\'t started. I\'ll talk to you later!  Peace and Love,  Melissa  '),(24669,2490703,'18,May,2004','2004-05-18 00:00:00','   Yeah, I\'m having one of those days. It\'s not a really good day, but   it\'s not a really bad on either. I\'m just sitting around filling out graduation stuff. It\'s crazy, 7 days left. WOO HOO! I\'m going to Cedar Point tomorrow for Physics. I\'m pretty excited. My group is just Amanda Allmendinger and I. We\'ll have fun seeing as how her and I both aren\'t roller coaster people. It should be great fun. Are you coming up for my open house? I\'m not sure what day it is...My mother keeps changing it. It\'s like she can\'t make up her mind. Oh well, whatever day it is, you should come! You could possibly come to my graduation if you come up for the weekend. I graduate on June 6th. Just let me know hun! Um...Your thing didn\'t save. I don\'t know if you saved it. Nothing has really happened today. I went to school, cheated on my physics test, went through the rest of the day. After school I went to Physical therapy for the bum leg...now I\'m home typing to you. How did you\'re thing go for the ticket? I hope well...Hopefully you still have a driver\'s license! I still haven\'t talked to my dad. I told him I didn\'t want him at graduation...like he actually cares. He\'s only going because my mom would be mad at him if he didn\'t. The way I see it, he didn\'t care for 13 years...why does he all the sudden care. Oh Wait...he doesn\'t. But yeah, I have to go finish my things. You\'ll be getting one in the mail soon! I love ya hun! Talk to you later!  Peace and Love,  Melissa  '),(24670,2490703,'17,May,2004','2004-05-17 00:00:00','   Man, life sucks so bad sometimes. I wonder how I make it through the day sometimes. This morning was a total bummer. I woke up late and I was rushing around for school. I ran up to take a quick shower and I got into the bathroom when I realized I forgot a towel. I ran into the laundry room and I dad pushed my little brother into the bathroom. I was like what the heck?! He started yelling at me like everything was my fault. I was crying my eyes out. I told my mom later that I didn\'t want my dad at my graduation because he hasn\'t cared about me for the past 13 years, why does he all the sudden care? I don\'t want him to ruin one of the most important days of my life. I just don\'t want to go through the crap that I have been on graduation day. I now know why Travis hates dad so bad, he\'s such a jerk. I was looking back and my dad has never cared, or hardly cared. I remember when I told him I was excited to graduate and move on with my life and he kicked me into the refrigerator and Travis saved me from further beating. Or how he was always too busy to go to soccer games or track meets. He never came to any. How sad is that? I remember him telling me that no matter what I did, nothing I ever did would be good for him. It made me so sad. People seem to say that to me a lot in my life. It seems like no one really loves me, they just love to see me hurt a lot. It\'s such crap. I\'m just glad that I have you for a friend to get me through the tough times. I don\'t know where I would be with out you. I spent the day up at MSU with Amanda and all of her friends there. I didn\'t really fit in because I\'m such a minor, but it was better then the family ward and home. Isn\'t is sad to say that any place is better than home. I was always taught home should be a safe place...so much for that. I just hope things get better...Trials make you stronger, but I\'m done with that! I need some smooth sailing...maybe a boyfriend that treats me well. Yeah right, funny joke. I hate being so NEGATIVE! I found this guy though who fits the profile... 1. He has VERY nice shag hair...SO HOT! 2. He plays guitar...or and he sings! 3. He is a skateboarder!  On top of all that, he is so nice and he is good looking which is a bonus. haha... I miss you so much. If you come up this summer we\'ll have to hang out and TP people. It\'ll be so much fun. Hate crimes rock my world. But I have to go hun. JASON is online acting like David again and I really can\'t deal with him again. I love you girl. Talk to you soon!  Peace and Love,   Melissa  '),(24671,2490703,'14,May,2004','2004-05-14 00:00:00','   Wow, Things have massively changed since the last time I wrote in here. Jason and I broke up. He has been sich a dick to me ever since. He tells me he wants me to die and tells people lies about me...so that\'s pretty cool! But I\'m tough and I make it through. Mr. X is still a hottie, too bad he\'s too young. Well, maybe he\'s not TOO young! hahaha...  I met someone new. He\'s a real sweetheart and I love to be with him. There is just one problem...he has a crazy girlfriend who has taken over his life and changed him completely. He\'s still sweet, but she\'s changed him, almost brainwashed him. I miss the old Zore! The bad thing is, so do all of his other friends. They want me to break them up so the old Zore will come back. I feel really bad because I don\'t want to lead him on and then break his heart. I know after he breaks up with her he\'ll never want to talk to me again because he\'ll realize that is is all my fault he broke up with her. It\'s such a soap opera. I tried to tell him what people were feeling, but I think I only made things worse. I feel so bad. Who does that...hurt someone they love? I wish I could go back in time and change the way things went.   I\'m one of those people who try to please everyone. That\'s why I cut. If I couldn\'t fix everything I felt so bad, like it was my fault there were problems in the world. I hate being that way. If only I were a normal person, I would have less scars and more happy memories. I think the depressing me keeps all the boys from becoming more than just my friend. Or maybe it\'s Jason\'s fucking lies! I feel bad for him. He\'s new girlfriend is such a slut. She is really nice, but she likes her boys. I went to this concert and I guess they were making out...I wasn\'t there. As soon as he left, she was all over another guy. That is so sad. He\'s a jerk and stuff, but he really doesn\'t diserve that. Well, honestly he does. He needs to feel the hurt for once. How mean is that? I guess that shows how much he hurt me.   Other than the 2 major things going on in my life, I\'ve been really happy. Only 9 days of school left! How crazy is that. I honestly can\'t wait. It seems so wierd. I\'m going to graduate and never have to go back to the hell hole of High School. It seems like a dream. I\'m sure I\'ll meet lost of cute guys up at CMU. I\'ve decided a guy needs 3 things for me to like him...Shag hair, skateboards, plays guitar. AWESOME...I can\'t resist a shag! They should have a great personality and of course, they would NEVER lie to me!   I was talking to Teffticles, he\'s so much fun to hang around when it\'s just him and I. We talked about a lot of stuff. I feel like I can actually be myself around him and that\'s kind of nice. We were joking around about him being a doctor and me being an attorney then we\'ll get married and have 2 twin boys and live in a huge house and be rich. Everyone\'s dream...Wierdly enough...I could picture that happening. AHHHHH! But I best be off. I love you hun! talk to you later!  Peace and Love,  Melissa  '),(24672,2490703,'28,June,2004','2004-06-28 00:00:00','   Another boring summer day. I like summer...When it\'s actually warm! I was supposed to go up to CMU today to get a job thing, but I changed my schedule around so I have 4 day weekends and classes 3 days a week. Yeah, mom. She helped me. All I have to worry about is if the transfer. I\'m so excited for college now, other than the fact that I\'m not going to see Ryan as much anymore. I\'ll see him when I go to MSU because it\'s closer to home. Sadly, I\'m going to have to live at home for a little bit because housing around there is so expensive. It\'s kind of crappy. But, I\'ll be closer to most of my friends, They are all going to MSU or LCC so it\'ll be okay. I have so many jobs to apply for tomorrow. I found this one a coffee shop. It sounds like a really cute job. Then I have an interview at banana Republic tomorrow. And I have an Application for Kohl\'s. So I hope I can get a good job so I can actually have money to spend on things other than college. All I can say is welcome to being a poor college student. But yeah. I have to go. Talk to you soon!  Peace and Love, Melissa  '),(24673,2490703,'21,June,2004','2004-06-21 00:00:00','   The Open House is finally over! I\'m so happy, that means no more cleaning the house at 8 a.m. EVERY MORNING! It was a lot of fun, I wish you could have been there. I miss you so much. A lot has happened this week. I went Carp Shooting with Andy and Ryan. It was so much fun, no joke. What you do is go out onto the lake when it\'s dark and shoot at carp with a bow and arrow. OMG, it was so fun. I just held the light to spot the fish because I\'m not strong enough to pull the bow back. Anyway, Ryan drove me home and I had been joking with him about not getting my goodnight kiss after Homecoming. So I got my good night kiss. Over the next few days we hung out and stuff and last Tuesday we started going out. He had to leave on vacation Thursday, and he\'s coming back tomorrow. But in the process of him leaving...Amanda had Bob from Wisconson come over. They were an item kind of. But Amanda likes some other guy and they broke the whole thing off. So I started hanging out with Amanda a lot more so Bob wouldn\'t have to be so alone, just because I\'m a nice person. Anyway, we were tie-dying yesterday, because we\'re hippies, and we definately had a lot of fun. But I had to go seeing how it\'s Father\'s Day and I should spend some time with my dad. Bob calls me later but Kevin keeps calling for Amanda so he has to go. Well...Amanda calls me back and tells me she and Bob were talking and Bob has a little crush on me. It\'s so cute. Bob and I talked on the phone until 12:30 when he had to go to sleep because he has to get up at 7! But we\'re supposed to hang out today, which will be pretty cool. He\'s rad awesome at the guitar which is cool. I LOVE THE GUITAR! But I also love Ryan. I can\'t get him out of my mind. All this drama. Hahahaha. Well, I gotta jet. I have a shower to take and a room to clean!   Peace and Love, Melissa  '),(24674,2490703,'11,June,2004','2004-06-11 00:00:00','   Hey! I finally got a job! WOO HOO! I work at a day care in Howell. All I do is clean. It\'s kind of nice. All I need now is a second job so I can actually spend of the money I\'m earning something other then college. OMG! I didn\'t tell you, I got hypnotized at senior lock in. It was pretty funny. The guy had us put our hands together and did this thing and told us to try and get our hands apart. Yeah, mind wouldn\'t come apart. It was a good experience. He made us relax then he was like I\'m going to tell you a joke and it\'s going to be the funniest joke you\'ve ever heard. He said the first word of the joke and I started laughing SO freaking hard. Man...Then he would say think of something sad...Then he would say Indian and we would all crack up again. Hahahaha...I\'ll have to tell you more about it later because I have to go take a shower.   Peace and Love, Melissa  '),(24675,2490703,'09,June,2004','2004-06-09 00:00:00','   Summer is so boring...  '),(24676,2490703,'05,June,2004','2004-06-05 00:00:00','   It feels good to relax for once! I have a couple of open houses to hit today, so I really can\'t chat long. Former President Reagan died today...He was the president when I was born! Weird, he died the same way my grandma did. They both had alztimer\'s and they both contracted pnmonia. But yeah, I just got home from Eliene\'s open house. She\'s so pretty, I\'m jealous. But I gotta run! I\'ll talk to you later!   Peace and Love, Melissa  '),(24677,2490703,'02,June,2004','2004-06-02 00:00:00','   School is out! It wierd how people used to be \"better\" then you when we were all in school together, and now that we\'re out, we are all finally good enough to talk to eachother. It\'s so retarded. I was picking up Tyler from school today, and I saw some people that I know, but I\'m not tight with. One of them came over and started talking to me. We talk because of band and we had 2 other classes together this year, but before this year I had never talked to him in my entire life. It\'s wierd how when things are starting to change, people come together. I always thought things like that were wierd. Like after 9-11 people started having American flags up and they supported the president and all this crap. Then the war starts and the people of America are like screw Bush blah. I just don\'t get it. People are just wierd. But anyway, I have to jet, I have adresses to type up and thank you letters to write. Hope to see you soon!   Peace and Love,  Melissa  '),(24678,2490703,'01,June,2004','2004-06-01 00:00:00','   Yeah, I haven\'t been up to much today. I went shopping with my mom. I like to go with her, she buys me stuff. It\'s so awesome. I got some new shoes today. It was the last pair left on the shelf...Actually they were the display shoes. Anyway, I\'m talking to this really cute guy online. He goes lives in East Lansing. He\'s so nice. I\'ll have to send a picture. Anyway, I have to jet, I have soccer stuff to take care of for Amanda. I will talk to you soon!  Peace and Love, Melissa  '),(24679,2490703,'31,July,2004','2004-07-31 00:00:00','   Hey! Guess what? Dad let me drive his Jag today! I was super excited. I think it\'s because I was really sick last night. I was up late reading this book, it was really good and I didn\'t want to put it down. I was reading when I got this huge pain in my side and I couldn\'t breathe really well. It totally freaked me out, but I ignored it and kept reading. But it got worse so I went up to my parents room and I was freaked out. So I was all hyperventalating and they finally calmed me down and I fell asleep at about 2pm. My dad stayed up and made sure that I was okay for about an hour. I woke up at about 5pm and went back to my room. It was a very unpleasant night! But I made it through and I\'m feeling much better now. I have so much to tell you and I think my mom said that you were coming up next weekend or something? I don\'t know. I was supposed to go get my nails done but I got grounded for not telling my parents where I was going. Well, I pretty much got grounded for yelling back at my mom, apparently it\'s \"Disrespectful\". But you know, she didn\'t ask where I was going so I didn\'t tell her. That\'s just how it works! It\'s called the don\'t ask, don\'t tell rule. But yeah, I move out in 26 days. I\'m happy and nervous at the same time. GRRRRRR! You know what makes me mad. Okay, I get into huge trouble for not telling my parents where I was going right. Amanda steals my mom\'s credit cards and they take away her TV and VCR/DVD player...but no other punishment. Isn\'t that so gay? I was so pissed off. She does something horrible and nothing happens. It\'s not fair! Parents are so dumb sometimes. But I have to go. I think I\'m going out into the sun for a while!  Peace and Love, Melissa  '),(24680,2490703,'30,July,2004','2004-07-30 00:00:00','   I am really bored right now. My mom woke me up early today because supposedly my housing payment for CMU is due Sunday, but they hadn\'t sent us a bill yet. So we\'re looking all over for a bill they may have sent and online for a number to call or a place to pay. We finally found the number, and guess what...they changed the date of payment to August 16th and they didn\'t bother to tell anyone. So that was eventful. But other than that, I\'ve just been working and packing my stuff up to go to college. I\'m half excited and half not. I\'m so excited because I\'m getting all new stuff and I get to leave home and all the exciting stuff that happens to college. But I\'m not going to get to see most of my friends...oh and Ryan, except for like once or twice a month. It\'s crappy, but I\'m sure with all the fun of college, I\'ll be able to deal :) OMG! I had this dream that Laura and I were in a room with 2 girls...and they were LOVERS! Wouldn\'t that be horrible? I\'m sure they\'d be cool and all, but I\'d be so paranoid all the time. Aha. Well, I gotta jet. I love you like a sister!          Peace and Love,      Melissa  '),(24681,2490703,'28,July,2004','2004-07-28 00:00:00','   28 days until I start a new chapter of my life. Wow that sounds like a book! \"28 days until the new you\" hahaha So how have you been? I\'ve been okay. The more I sit around and think about leaving, the sadder I actually get. I actually like working so I can keep my mind off things. It\'s so sad because Ryan is never home and when he is home, I\'m at work. It\'s going to suck so bad when I go to college and he is still here. My mom said that I can come home once a month, but you and I both know my mom and if I called for her to come get me, she would drive all night to get to me. I love that about my mom. Ryan is big into hunting and stuff so during October-December pretty much I\'m never going to see him. It sucks. I know people say change is good, but right now I\'m pretty happy with my life. I\'m not saying that I\'m not going to like college. But I already made it a goal in my life that I\'m going to get as close to a 4.0 that I can in college, no matter what it takes. I want a career. A weird thing happened. I went to this church thing with my mom and there were a lot of mothers that talked and one in particular caught my attention. She said that her whole life she wanted to be a teacher. When she got through a year of college she got this feeling that teaching wasn\'t something the Lord wanted her to do. So she prayed about it and she just knew that she was supposed to be an accountant. Something similar happened to me. I was registering for classes at CMU. I said that I had intended on majoring in law when this thought went through my mind that I shouldn\'t be an attorney. I didn\'t listen and I registered for some law classes. Last night I had this dream that I became a psychiatrist. Weird, I talked to Ryan about it, and because of the many problems that I have been through, I might actually be able to help someone because I probably experienced something similar in my lifetime. I don\'t know...what do you think? I honestly think it might be a good career that I could deal with. Anyway, I have to go. I love you Kira. I\'ll talk to you soon!          Peace and Love,    Melissa  '),(24682,2490703,'13,July,2004','2004-07-13 00:00:00','   So, I\'ve had an exciting week with Travis home. He met and approved of Ryan, and Ryan even survived with no serious injury. Actually no injury at all. But sadly they are leaving. I\'m going to miss them. Janell is a person that I can talk to and trust with my secrets. It\'s going to be sad when she isn\'t around. Ryan isn\'t even home today...I\'m so BORED! Oh, I skipped work today, so that was nice. Maybe I\'ll get fired! WOOT! Well, I\'d rather quit. I start my other job tomorrow, 8-5 pretty much everyday. It\'s crap. But I have to jet, I\'ll talk to you later!  Peace and Love, Melissa  '),(24683,2490703,'07,July,2004','2004-07-07 00:00:00','   Stupid Blogger...wouldn\'t let me make my writting italicized! iQUE ESTUPIDO!  '),(24684,2490703,'07,July,2004','2004-07-07 00:00:00','   It\'s been a while since I last wrote in here. I got a second job at Banana Republic. I haven\'t started working at that particular store because it isn\'t open yet, but I\'m sure it\'ll be loads of fun! Travis, Janell, and Nathan are here from Wisconson. I\'ve missed them so much. Nathan is getting so big. He\'ll be 1 in August. It\'s hard to believe a whole year has flown by. It doesn\'t seem right that I\'m not going back to FHS...which in no way is a bad thing. College, here I come. Anyway, I hate the day care job and I wish it would just end. Hayley and I decided that we\'re going to Warped Tour 2004! YEAH! I\'m super excited. Maybe Ryan will be able to come now that he knows the right date! That smart Andy, he told him that it was the 8th. That kid needs to lay off the pot. Anyway. I have to go, I have to go feed the baby! OMG! TORE IS HOME! I missed him. But yeah, I\'m out.  Peace and Love, Melissa  '),(24685,2490703,'05,August,2004','2004-08-05 00:00:00','   I\'m so mad at Ryan. I\'m crying because I\'m so mad. I called him so I might actually be able to talk to him before I went to work. So he answers the phone and before he even says hello, I hear Andy Marr say \"She\'s a bitch, tell her that\'s from Andy.\" then some girl laughed. I\'m so upset. After he finally said hello, he was like can I call you back. I was like if you\'re lucky. I\'m so mad! He told me we couldn\'t do anything today... Now I know it\'s so he can hang out with his other friends who are clearly more important than me. He\'s leaving me for the weekend then I\'m LEAVING for college and I\'m not going to be here to ignore! You know what\'s sad, Amanda said met a brother of a friend who is my age and I guess he\'s really cute. I just thought about calling the guy so I could piss Ryan off. AHHHHHH! I\'m so mad, Kira I don\'t know what to do. I don\'t think I\'m going to talk to him if he ever calls me back. I know that I\'m not...I have to go so I can cry to myself in my room.  I love you like a sister.                                    Peace and Love,                                             Paris  '),(24686,3952263,'21,July,2004','2004-07-21 00:00:00','   PLOT:  Catwoman is the story of shy, sensitive artist Patience Philips (Halle Berry), a woman who can\'t seem to stop apologizing for her own existence. She works as a graphic designer for Hedare Beauty, a mammoth cosmetics company on the verge of releasing a revolutionary anti-aging product. When Patience inadvertently happens upon a dark secret her employer is hiding, she finds herself in the middle of a corporate conspiracy. What happens next changes Patience forever. In a mystical twist of fate, she is transformed into a woman with the strength, speed, agility and ultra-keen senses of a cat. With her newfound prowess and feline intuition, Patience becomes Catwoman, a sleek and stealthy creature balancing on the thin line between good and bad. Like any wildcat, she\'s dangerous, elusive and untamed. Her adventures are complicated by a burgeoning relationship with Tom Lone (Benjamin Bratt), a cop who has fallen for Patience but cannot shake his fascination with the mysterious Catwoman, who appears to be responsible for a string of crime sprees plaguing the city.  Watch Catwoman in the theaters! Opening soon!   '),(24687,3952263,'20,July,2004','2004-07-20 00:00:00','   PLOT:  When a CIA operation to purchase classified Russian documents is blown by a rival agent--who then shows up in the sleepy seaside village where Bourne (Damon)and Marie (Potente) have been living under assumed names--the pair collapse their lives and head out. Bourne, who promised retaliation should anyone from his former life attempt contact, is forced to once again take up his life as a trained assassin to survive.  Another great movie to watch in the theaters. Don\'t forget!   '),(24688,3952263,'18,July,2004','2004-07-18 00:00:00','   Here are some of the official websites of your favorite star. If not included, email me at: extreme_intense48@yahoo.com  You want em\'? Go get em\'.  Billy Boyd [Lord of the Rings, Taggart]--www.billyboyd.net Jennifer Aniston [Along Came Polly, Friends]--www.jenniferaniston.com Adam Sandler [Big Daddy, Mr. Deeds]--www.adamsandler.com Jeremy Sumpter [Peter Pan  ]--www.jeremysumpter.com Sean Astin [Lord of the Rings, Rudy]--www.seanastin.com Lindsay Lohan [Freaky Friday, Parent Trap]--www.llrocks.com Hilary Duff [Lizzie Mcguire, Agent Cody Banks]--www.hilaryduff.com Matt Damon [The Bourne Idenity, Ocean\'s Eleven]--www.mattdamon.com Ben Affleck [Daredevil, Pearl Harbor]--www.benaffleck.com Colin Farrell [Minority Report, Daredevil]--www.colinfarrell.ie Halle Berry [X-Men, Gothika]--www.hallewood.com Sir Ian McKellen [Lord of the Rings, X-Men]--www.mckellen.com Mandy Moore [A Walk to Remember, The Princess Diaries]--www.mandymoore.com Jake Thomas [Lizzie Mcguire, The Cell]--www.jakethomas.com   Keep visiting this blog for more official websites.   '),(24689,3952263,'17,July,2004','2004-07-17 00:00:00','   Set in a future Earth (2035 A.D.) where robots are common assistants and workers for their human owners, this is the story of \"robotophobic\" Chicago Police Detective Del Spooner\'s (Smith) investigation into the murder of Dr. Miles Hogenmiller, who works at U.S. Robotics (run by Greenwood), in which a robot, Sonny (Tudyk), appears to be implicated, even though that would mean the robot had violated the Laws of Robotics, which is apparently impossible. It seems impossible because.. if robots can break those laws, there\'s nothing to stop them from taking over the world, as humans have grown to become completely dependent upon their robots. Or maybe... they already have? Aiding Spooner in his investigation is a psychologist, Dr. Susan Calvin (Moynahan), who specializes in the psyches of robots...  DON\'T FORGET TO WATCH I-ROBOT!   '),(24690,3952263,'17,July,2004','2004-07-17 00:00:00','   Here are the Top 10 movies:  10. SLEEPOVER  - Four best friends, desperate to improve their social status, enter into an all-night scavenger hunt against the popular clique in their school.  9. SHREK 2  - The honeymoon is over, and Shrek faces an even scarier task than rescuing Fiona from a dragon: meeting her parents (Cleese and Andrews). Worse yet, neither has any idea that Shrek and Fiona are ogres! Throw Puss in Boots (Banderas), Prince Charming (Everett), and a Fairy Godmother (Saunders) into the mix, and it\'s one full fairytale.  8. THE TERMINAL  - Balkan immigrant Viktor Navorski (Hanks) finds himself stuck in a New York City airport terminal with a nullified passport and paperwork when the war back in his homeland causes his nation of origin to no longer exist. Suddenly homeless, Viktor lives in the terminal, befriending the staff and falling in love with a flight attendant (Zeta-Jones).  7. DODGEBALL: A TRUE UNDERDOG STORY  - Can a group of losers win a dodgeball tournament in order to save their beloved gym from a hostile corporate takeover?  6. WHITE CHICKS  - Disgraced FBI agents Marcus and Kevin Copeland (Marlon and Shawn Wayans) go way undercover in an effort to protect hotel heiresses the Wilson Sisters (Maitland Ward and Anne Dudek) from a kidnapping plot  5. THE NOTEBOOK  - An old man (Garner) reads a tale to a senile woman (Rowlands) about two young lovers (Gosling, McAdams) who meet cute, fall in love and break up in 1940 South Carolina. After World War II, the youngsters are reunited, but he\'s bitter and she\'s engaged to someone else. How will it all end? Based on the bestseller by Nicholas Sparks.   4. FARENHEIT 9/11  - A documentary which explores why the U.S. has become a target for terrorism in recent years, and the strange relationship between the families of mortal enemies George W. Bush and Osama bin Laden.  3. KING ARTHUR  - After the collapse of the Roman Empire and the subsequent fraction of the British Isles, a mythical figure emerges, destined to become King of England.   2. ANCHORMAN  - Ron Burgundy (Ferrell), San Diego\'s #1 anchorman, faces his toughest assignment to date: fend off newswoman Veronica Corningstone (Applegate), whose feminist special-interest stories threaten his standing behind the desk.  1. SPIDERMAN 2  - Peter Parker (Maguire) is now attending New York\'s Empire State University, and there\'s a new super-villain on the loose: scientist Dr. Otto Octavius (Molina), otherwise known as Doctor Octopus (thanks to the four mechanical arms he has grafted onto his back).   '),(24691,3952263,'17,July,2004','2004-07-17 00:00:00','   This is the ONLY blog that gives you the latest scoops in Hollywood.   '),(24692,3952263,'01,August,2004','2004-08-01 00:00:00','   \'The Village\' depicts the thrilling tale of an isolated town confronting the astonishing truth that lies just outside its borders. At first glance, this village seems picture perfect, but this close-knit community lives with the frightening knowledge that creatures reside in the surrounding woods. The evil and foreboding force is so unnerving that none dare venture beyond the borders of the village and into the woods. But when curious, headstrong Lucius Hunt plans to step beyond the boundaries of the town and into the unknown, his bold move threatens to forever change the future of the village.  THIS MOVIE IS NOW SHOWING AT THEATERS NEAR YOU, PLEASE WATCH THIS THRILLING/HEARTPOUNDING MOVIE. Go to www.imdb.com for more information.   '),(24693,3452845,'31,May,2004','2004-05-31 00:00:00','   Here I am, just a few days into my blogging career and I can already see myself losing focus on everything that is important. Sorry hun, \"I\'ve got to go do some research.\" Which in essence means I need to search active bloggers for like intelligence and interests. Hopefully I will completely lose interest in the bond market, lose my job, and be forced to blog professionally. Does anyone know if there are blog offs? Much like dance offs? They should make a movie and call it \"You got blogged\". Or as my buddies would prefer to say...stop being such an f-ing nerd.  '),(24694,3452845,'30,May,2004','2004-05-30 00:00:00','   Sweaty guys in sleeveless T-shirts. Girls are parading around in tank tops, belly\'s exposed (the good and the bad). The bartenders are ripping you off for a ten dollar red bull and vodka. However, the look from the cute blonde wearing the casual black pants, gucci sandals, and plain white blouse changed the entire scenery. It\'s amazing sometimes how a single glance can arouse you from the slumber encompassed in the bar ritual. No one spoke to each other. Just a simple smile. Besides, I\'m involved right? Either way, the simple smile is what carries over to today and reminds me that yes I am the cute machine.  '),(24695,3452845,'29,May,2004','2004-05-29 00:00:00','   I started feeling my age yesterday. Amy Sullivan is bigger than I am. I remember the fall afternoon at tomahawks practice, the smell of crumbling leaves and barbeques in the distance. I was punishing Mike Sullivan for the fact that his parents actually had the sex to make the baby Sullivan. Now she is 15. Have I really passed 15 years like that? In a flash where everything has changed but still pretty much remains constant. Keith Poisella however, has never changed. Glancing through the old football and lacrosse programs confirms this fact. I\'ve never seen anyone have the same haircut, same posture, and same attitude throughout their lives. Maybe he is a different breed of human. In fact, I know he is a different breed because his parents remain together and they themselves haven\'t approached any morphing...which is kind of impossible considering Mrs. Poisella must be going through menopause and we all know the ginseng, organic, soy induced stages of that transformation. I guees I\'m glad then that I am a child of divorce...it forces me to keep changing.  '),(24696,3452845,'27,May,2004','2004-05-27 00:00:00','   Walking home today I had a recurring thought popping in my head. The sun, while tremendously loving and caring, can produce thoughts of guilt. For example, I left work after an awful day of trading with a full thurday of a sunny day. What did I choose to do? Walk aimlessly and consider a nap. I wish that I had access to an auto because if I did watch out, I would be cruising the coastline with Sublime blasting. Maybe that\'s just it...A car that\'s what I need.  '),(24697,3452845,'25,May,2004','2004-05-25 00:00:00','   I am a programmed machine running to the tune and rythmns of the universe. I follow the rythmn in my trades. I smoke a cigarette. I play ms. pac man on my cell phone. I think playing ms. pac man is like moving to the nature of god\'s choosing. If you count the dots on the ms. pac man board you will find they are in perfect bars and measures.  '),(24698,3452845,'25,May,2004','2004-05-25 00:00:00','   Walking around aimlessly in the city it seems that females have the annoying habit of cell phone munching. I really don\'t understand the whole idea. What\'s wrong? Walking around by yourself without your electronic device in hand makes you unpopular? Whatever happened to walking and observing? The mouth doesnt always have to move. And I know that no one is that important where a few moments of silence will break the new paradigm of productivity. Relax. Please.  '),(24699,3452845,'11,June,2004','2004-06-11 00:00:00','   Ronald Reagan died last week and all I can think about is barbeque. Yes the markets were closed today and what did I decide to do in honor of our 40th President? I decided to master the art of the Memphis Rib. People who don\'t understand barbeque are those people who would prefer to watch Barbara Walters interview Liza Minnelli about her first period. These same people would also probably habitually read US Weekly while working out on the treadmill. I on the other hand would prefer to molest a budweiser while scanning the contents of a charcoal grill. Alway charcoal, never gas mind you. If you are using gas then you are obviously too rich of a prick to enjoy a fine barbeque. The keys to the memphis ribs are as follow. 1)Marinate for hours 2) Use some kind of smoked wood chip for taste and 3)Let all other thoughts except for brain numbing drinking and basting enter your mind. In that sense barbequing is like therapy. A self medicating, meditating, and marinating kind of therapy. Mmmmm marinating therapy. Watch out Freud I\'m going to barbeque you into oblivion.  '),(24700,3452845,'10,June,2004','2004-06-10 00:00:00','   Living in the world of the yuppie is like living in a transplanted society whose major goals are intense intoxication, golf outings, and sales at the Short Hills Mall. Girls chit chat about the latest Sephora trends while dining on a sushi sahimi menu and sipping on the latest australian chiraz. These people are like the kids who never had worry number one about anything. They were probably given their first car, lack any student loans, and more importantly never have the aching pain in the back that if they fuck up it could be on the street and into the shelter, or even worse back to mom\'s house. I saw one girl the other day walk her minature poodle down the street while talking on her earpiece cell phone while smoking a parliament. And it\'s always parliaments. I don\'t really know why I live in Hoboken, perhaps it is the cheaper rents and proximity to manhattan, more likely however is that I have become what I feared most in this life a yuppie scumbag following the crowd. Why can\'t I just pick up and go? To Ecuador, New Zealand, or San Diego. I know why, because I\'m a big pussy that\'s why. Security and eveness have this attachment to me...I need escape...any suggestions?  '),(24701,3452845,'10,June,2004','2004-06-10 00:00:00','   I think that I will be working for the next fifty years. As I was walking through the train station today I came across some college aged kids obviously en route to their first job. I remember that initial angst of college retirement and workforce entry and to me it was one of the most damaging periods of my life. While dealing with this newfound problem I ran into one of my friends fathers and he told me, \"Don\'t worry. You only have to work for the next forty years until you can return to your college freedoms.\" Now, after five years of the routine I look at those college kids and actually smile. Besides, I got four years on em.  '),(24702,3452845,'05,June,2004','2004-06-05 00:00:00','   The American people are disappointed. Even if you only have a faint interest in thoroughbred horseracing you were probably tuned in to the running of the Belmont Stakes. However, today, only a few hours before the running of the final leg of the triple crown, America lost one of it\'s greatest presidents. That statement is coming from a lifelong democrat/Bush hater. The reason I mention it is because such a sad day in American history cannot coincide with a Triple Crown winner gaining our hearts and happiness. You cannot be in mourning and also celebrating. The whole balance would blow. This day was meant to be a day of remembrance and sadness not a day of exhultation.  '),(24703,3452845,'04,June,2004','2004-06-04 00:00:00','   I don\'t think loyalty has anything to do with modern business. Instead, I think that the company you work for would rather use you as a tool. A basic chess piece in their bigger chess game of social status and recognition...and more importantly, a few extra bucks. Everyone in America sympathizes with some of these executives and their company gorging. People like this say, \"Well, that\'s the american dream. Work hard, succeed and you will achieve economic power completely disproportionate to the actual value you add to the universe. I only mention this today because my employer (who will remain nameless even though I should reveal the spinless twats) promised on one thing and delivered another.  I am the model employee. I toe the line and never miss a day of work. I struggle to pay bills. I struggle to enjoy life. Why on earth, after such sacrifice and committment can you deny me my just rewards in our work TOGETHER? That\'s right we\'re together...not one above and one below aiming for the scraps. You and I are animals laboring towards the same goal. Talk about indirect de-motivativation.  It really only makes me want to slash their tires every day for six months. Only the pain and anguish of every day going to the parking garage and having to call a tow truck could be punishment equal to my aggrivation.Maybe that\'s my calling..The Corporate Justice League Chairman of Tire Slashing. If any wants to pay me for this...let me know.  '),(24704,3452845,'03,June,2004','2004-06-03 00:00:00','   The lesbian today got rowdy and tried to intimidate everyone with her pool acumen. I think she frequents the bar and has the old heads going crazy for her retro hippy lesbian antics. For me, it\'s just annoying. Congratulations, you\'re a lesbian. Everyone is really proud of you for coming out of the closet and proving that you can be what you want to be. But please, don\'t try to garner attention with your dock martins, black jeans, and leather jacket...meanwhile throwing disgusting looks in everyones direction. I think this kind of lesbian can only be called \"dike\". Much like if you are a w.o.p. ill call you a wop...even though you are italian. Everybody stop the act, it\'s okay do what you want but don\'t fulfill every stereotype known and don\'t expect congratulations or commendments on your extreme need to fit the mold. Lesbian or not you\'re still plugged in bitch.  '),(24705,3452845,'03,June,2004','2004-06-03 00:00:00','   I am disgusted. For some reason my blog post on June 2nd has failed the posting process. Perhaps I should save every draft. Too bad because the post was the most genius of all postings with content strong enough to break bones and bend biases. Oh well, it is probably perculating somewhere in web purgatory.  '),(24706,3452845,'03,June,2004','2004-06-03 00:00:00','   The rain has held off and the sun has brightly collided through the windows at work. I decided that I am finished trading for the day due to the fact that I am bored with the funky transfers taking place on my screen. The only thoughts parading through my head are dreams of jersey shore debauchery. The birds are chirping on the streets, I have my coach loafers on, my jeans are firmly pressed, and I am tan. As the saying goes, suns outs guns out. Why let another weekend escape without feeding the party desire? Don\'t worry I won\'t.  '),(24707,1196970,'30,June,2004','2004-06-30 00:00:00','   Fuck, the crazy Asians just invaded. And they were supposed to come after midnight, damn them and their fast cars.  '),(24708,1196970,'30,June,2004','2004-06-30 00:00:00','   The China-people are coming, can\'t you smell them? And then they\'re leaving, for Virginia Beach. Dragging me along. Virginia, oooh, the wonders of an overzealous police force, we\'re all sure to have so much fun.  '),(24709,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink Despair, Inc.  '),(24710,1196970,'30,June,2004','2004-06-30 00:00:00','   Well apparently today is the block party. Joy above joys, how much fun is  that  going to be? You could almost piss in your pants just thinking about it. Kind of like watching a baby crawl around on the floor towards a dead spider and you just  know  that when it gets there, it\'s going to pop the spider into its mouth so there\'s that pent-up excitement; yeah well this block party\'s something like that, minus babies and spiders.  '),(24711,1196970,'30,June,2004','2004-06-30 00:00:00','   Pleasantries! Yes, such pleasantries as going to some damned  Chinese  vegetarian restaurant for lunch. So I can\'t even get a salad damn it. Argh. Humph.  '),(24712,1196970,'30,June,2004','2004-06-30 00:00:00','   What the hell... what kind of fucked-up and conceited self-defense mechanism is this? Whenever person A realizes they may be wrong on something (i.e. a dead phone means it was a telemarketer\'s auto-dial machine but there was no operator available to \"take the call\") they automatically get pissy and declare that you \"don\'t understand\". Fuck you! And it\'s always about small things too, like how our cordless phone should  ALWAYS  be on its charger even though it has a battery life thats longer than our cell phone. Fuck off, eat shit, and die.  '),(24713,1196970,'30,June,2004','2004-06-30 00:00:00','   Woke up rather late today. Lunch kinda sucked, McDonald\'s isn\'t the best lunch, especially when everything\'s a bit stale from being in the car for too long. But that\'s beside the point. But there  is  no point, so it can\'t truly be beside something that isn\'t there, now can it? Kind of like how going to Hell is more fun when riding a SUV, but in a sedan, it\'s quite disk-shaped. No?  '),(24714,1196970,'30,June,2004','2004-06-30 00:00:00','   Momentous occasions are fun. I mean, who  doesn\'t  like to see someone go something horribly stupid and yet horribly dangerous at the same time?  Par exemple , when someone decides to go cow tipping while on speed. And in flipflops. Yes, something like that. They\'ll be glad to know they made a momentous occasion after they\'ve recovered enough to comprehend you, that is, if they recover at all.  '),(24715,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink CNN.com - Guilty plea in \'Matrix\' defense slayings - Jun. 25, 2003   What... the... fuck?  '),(24716,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink Exonerated, but Locked in Shadow  '),(24717,1196970,'30,June,2004','2004-06-30 00:00:00','   Ah yes,  \"Everybody knows about it, from the Queen of England to the Hounds of Hell.\"  - The White Stripes  '),(24718,1196970,'30,June,2004','2004-06-30 00:00:00','   Wow, crazy bastards coming down the stairs. Not in pajamas, though. But then again, they\'re not bananas, they\'re crazy bastards, and crazy bastards can do anything they want. Well within the limits of the law, at least. How funny. Whatever.  Damned technology... you would think that something that can process millions of pieces of information simultaneously should be able to remember that you don\'t like it when it screws up all your window sizes and when it insists that you only have two speakers and then proceeds to making horrible screeching noises out of the back two because, well apparently, they\'re not supposed to be there. My ass. Die.  '),(24719,1196970,'30,June,2004','2004-06-30 00:00:00','   Crazy, I\'ve stopped reading the news since June 19th. Oh well, I suppose the world stops for summer vacation. How much fun.  For some reason, I don\'t think I would mind working or something like that, yet I feel so incredibly lazy. I suppose I need to find something to do before I sink into a coma. On that note... ta.  '),(24720,1196970,'30,June,2004','2004-06-30 00:00:00','   Well today felt productive, in a way; that is, if one can call reading  The Scarlet Letter  productive.  But that;s beside the point. It\'s Tuesday, if that wasn\'t obvious already, yet it feels Wed-nes-day-ish. (Anyone ever wonder who the fuck Wednes was? Ha ha.)  '),(24721,1196970,'30,June,2004','2004-06-30 00:00:00','   Google  searches are so much fun, I mean, you can type in almost anything and turn up some website (although half of those websites are probably utter crap).  '),(24722,1196970,'30,June,2004','2004-06-30 00:00:00','   Merde!  '),(24723,1196970,'30,June,2004','2004-06-30 00:00:00','   Time Cube   Type in \"stupid ass\" in Google Search and this is the first thing that comes up. Damn humans can be fucking retarded.  Kinda like  this .  '),(24724,1196970,'30,June,2004','2004-06-30 00:00:00','   No one\'s online. How boring. And how the hell is someone online for the past 13-something days? What the hell.  So many damned summer assignments to do. Yes, well I suppose it\'s back to reading  The Scarlet Letter . Someone should have killed Hawthorne when they had the chance.  '),(24725,1196970,'30,June,2004','2004-06-30 00:00:00','   Watch, laugh, and be merry.   And now we\'re back to being bored. Fun fun.  '),(24726,1196970,'30,June,2004','2004-06-30 00:00:00','   The formatting is a bit bland, no? I should tinker with it. Now if I actually remember how...  '),(24727,1196970,'30,June,2004','2004-06-30 00:00:00','   Something really unexpected happened. I\'m over at another one of these fucked-up boring as hell crazy Asian parties and someone walks up and gives me the new Harry Potter book. Well it wasn\'t just  someone  per se but it was still weird. I knew the people (and they\'re really nice) but why the hell did he stay up until midnight yesterday to buy iit in the first place? Well I spent the subsequent six hours at the fucking \"party\" ( heavy  emphasis on the  \" \" ) reading, so I\'m almost done. Muahahaha.  '),(24728,1196970,'30,June,2004','2004-06-30 00:00:00','   Just realized it\'s 8:00 already. Kinda late to call anyone, ugh. Oh well, we can go Sunday. Hahaha.  I think I still need to find a summer job. Argh, so complicated.  \"I\'m looking to the sky to save me, looking for a sign of life, looking for something to help me burn out bright...\"  - Foo Fighters  '),(24729,1196970,'30,June,2004','2004-06-30 00:00:00','   It\'s raining. Rain is nice, given that you\'re inside, because then you\'re nice and dry etc. etc. Yes...  '),(24730,1196970,'30,June,2004','2004-06-30 00:00:00','   It\'s so nice, being able to wake up late and... well be able to wake up late. Nice and dark and gloomy outside too, so much better than that wretched sun.  Nothing to do though, but it\'s Friday, so I can bide my time.  '),(24731,1196970,'30,June,2004','2004-06-30 00:00:00','   And apparently there  are  no links to put, so I\'ll just keep on typing.  No more school for some couple weeks or whatever how long it is. Nice, very. We should finally have time for our scare-away-the-Abercrombie-customers mall trip.  Super-long-extra-fancy-expressions-with-lots-of-hyphens are cool.  '),(24732,1196970,'30,June,2004','2004-06-30 00:00:00','   \"How do you get your sleep at night? How did you get your noose so tight? Like chewing on tinfoil, it\'s so much fun Gonna be dead before your gone  Cause look how things have gotten  And I\'ll be happy so I won\'t pretend And I\'ll be cheering that you\'re going down And I\'ll be laughing, I\'ll be laughing  How many feelings can you steal? Gotta be part of your appeal I can see through you cause you\'re wearing thin Like chewing on tinfoil once again  Ha Ha you\'re dead And I\'m so happy In loving memory Of your demise  When your ship is going down I\'ll go out and paint the town Ha Ha you\'re dead Ha Ha you\'re dead Ha Ha you\'re dead  (ha ha ha)  Ha Ha you\'re dead The joke is over You were an asshole And now you\'re gone As your ship is going down I\'ll stand by and watch you drown  Ha Ha you\'re dead You\'re gonna be dead Just remember what I said Ha Ha you\'re dead Ha Ha you\'re dead Ha Ha you\'re dead\"  -Green Day  '),(24733,1196970,'30,June,2004','2004-06-30 00:00:00','   Starting fresh, seeing as how the last one was becoming dysfunctional. Hopefully, this will also encourage me to write more. I have to work on links etc. etc. now, so talk more later.  Freaky, I\'m talking to you as if you actually existed.  '),(24734,1196970,'30,June,2004','2004-06-30 00:00:00','   How pathetic...  I just spend the last couple minutesa rguing with myself about whether my current desktop wallpaper (it changes every couple minutes) was a Cezanne or a Renoir...  '),(24735,1196970,'30,June,2004','2004-06-30 00:00:00','   Well I finally got a watch again, after the sea took my old one. It\'s nice, being able to tell what time it is... not completely sure why it\'s so comforting, it just is. It\'s like having airbags in a car, not that useful as they are, but when you need them, it\'s nice to have them. The analogy probably isn\'t of \"SAT\" standards... fuck them...  '),(24736,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink Chinatown: Eat, Drink and Speak Cantonese  '),(24737,1196970,'30,June,2004','2004-06-30 00:00:00','   Oh, here\'s one, every time you finish writing one of these and you click on the large orange button that says \" Close This Window \", the computer (as usual) tries to interfere, asking if you \"really\" want to close this window... What the hell? I pushed the fucking button didn\'t I?  '),(24738,1196970,'30,June,2004','2004-06-30 00:00:00','   People are fucking ridiculous...  There\'s me sister, who screams about there being a single ant on the floor and is later found feeding it. After I killed it, she even attempted to put \"Angel\" (by the way, our hamster is also named \"Angel\", as is her teddy bear and a multitude of other miscellaneous items) outside so her family could \"see to her\".  Then there\'s the fucking computer program that messes up on installing itself and then refuses to simply fix the error, citing \"a lack of copyright controls\". Whoever the hell mkes up this copyright stuff needs to be shot. Several times, preferably.  There was more, but I don\'t seem to have the words to describe some of the other utterly ludicrous things people do... but then again, it\'s not that difficult to think up of some for ourself...  '),(24739,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink Go Away Little Children, You Might Get Eaten   A horribly rough website I put up because... well I have the space and because someone needed that song...  This post was much more interesting a couple minutes ago... humph.  '),(24740,1196970,'30,June,2004','2004-06-30 00:00:00','   You would think that with a virtual world accessible from a sitting position (both definitions of virtual), you could never be bored... well whoever came up with that idea is a stupid fuck.  I should just try going to sleep much later and waking up later...  peut-etre , but then again I would be bored at night instead of in the afternoon... not much of an improvement.  '),(24741,1196970,'30,June,2004','2004-06-30 00:00:00','   Wow... I feel like shit and I don\'t know why. You know, when you\'re supposed to be hungry but yet you\'re not and furthermore you feel like vomiting but yet not and...  argh .  \"Moo\" should be used as an exclamation of frustration, much in the same way that \"argh\" is used currently.  \"Go away, you\'re so damned annoying... MOO.\"  '),(24742,1196970,'30,June,2004','2004-06-30 00:00:00','   Well, the AP Exam grades were mailed today. I got a five... rather happy. Wonder how everyone else did...  Whatever...  Today was nice for some reason.  It\'s wierd how you can take certain things for granted (not water, electricity, or heat... in the US, you  can  pretty much take those for granted)... something like a watch. Damned ocean took mines over a week ago and I\'m still trying to read my wrist. Damn.  '),(24743,1196970,'30,June,2004','2004-06-30 00:00:00','   New word of the millenium (it\'s \"of the millenium\" because then I\'ll never have to make up a new word again... ever).  omgwtf (OHM-guh-what-te-FOO)  interjection   used when one sees something both shocking (the apalling sort of shocking) yet unbelievably amazing; i.e. you see a cow trample a baby to death while spreading cream cheese on a bagel and ordering her new Gucci shoes from a stolen cell phone - in the given instance the \"omg\" part is directed mostly at the baby, but pity soon gives way to excitement as you realize the cow is actually talking (\"wtf\")  Now, back to being bored.  '),(24744,1196970,'30,June,2004','2004-06-30 00:00:00','   It doesn\'t seem gloomy enough outside even though it\'s raining. Damn.  Pleasantries is such a nice word. Just thought everyone should know that.  '),(24745,1196970,'30,June,2004','2004-06-30 00:00:00','   Oh, the agony. So much to do, too much time to do it all... in fact,  so  much time that it\'s highly unlikely anything will be done at all. Pity...  Life\'s been uneventful... uneventful is a nice way of putting it. There was a word for \"nice ways of putting things\", wish I could remember it right now. Euphemism. Yes.  It would be nice to go somewhere to do something, but the Heavens just opened up... this evening will be rather  uneventful .  '),(24746,1196970,'30,June,2004','2004-06-30 00:00:00','   Damn those \"Lean Cuisine\" things are disgusting. Goopy \"sauce\"-stuff and chunks of mystery meat with \"fresh\" pasta and a side of carrots that just got microwaved for five minutes. How bland...  Munchies... ta.  '),(24747,1196970,'30,June,2004','2004-06-30 00:00:00','   Well, I just found out that there\'s this nifty \"Intercom\" feature on our main phone. And, when you push the \"Intercom\" button from the phone itself, it automatically opens up the speakerphone on the phone thingy for you so speak from. So nice to freak people out... the phone just starts talking to them...  '),(24748,1196970,'30,June,2004','2004-06-30 00:00:00','   There\'s a new Google Toolbar, if anyone wants to  get it . It even has a nifty little \"BlogThis!\" button already on it... yes!  Not that it concerns any of you, but that\'s beside the point.  So bored, you would almost not object to watching golf on TV...  almost .  '),(24749,1196970,'30,June,2004','2004-06-30 00:00:00','   Wow, now I\'m bored again, and dreading all that damned summer work I need to do. Damn.  '),(24750,1196970,'30,June,2004','2004-06-30 00:00:00','   Well it\'s nice to be back, with dry towels and running hot water and food. Not that the hotel didn\'t have that, but it\'s always so inconvenient away from home. Damned beaches. Oh yes, and no more people constantly coming up to me and asking if I\'m the son of so-and-so (and they were usually wrong).  Virginia Beach is no fun, unless you can drink and you like the fact that  ever  single restaurant seems to sell beer...  '),(24751,1196970,'30,June,2004','2004-06-30 00:00:00','   So yeah, this blows, we have all these people over (fortunately, they\'re not related to me in the least) that I don\'t know and so all I\'m doing is smiling and nodding and it\'s getting really tedious and these people are so damned boring and there\'s no one to talk to damn it but I did get to drive a gold cart, which was cool seeing as how I won\'t be able to drive anything else for another 6 months damn that sucks doesn\'t it and I think this is the longest sentence ever. End.  '),(24752,1196970,'30,June,2004','2004-06-30 00:00:00','   I\'m done with all my damned summer work the day before the day before school starts. How nice. School starts... School...  Depressing, no?  '),(24753,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink Estimate for \'04 Deficit Is Increased to $480 Billion   You wonder how we went from having a surplus to the Bush taking office and now this.  '),(24754,1196970,'30,June,2004','2004-06-30 00:00:00','   It\'s strange, I havent listened to The Kinks in so long. I should go do that.  All this damned summer work, you really wonder if the teachers are purposely trying to be spiteful... \"If  I  have to spend half my summer grading  your  final exams which  I  decided to make as eighteen essays and sixhundred seventy-two multiple choice, then  I  think  you  should have to slave away for a month to finish all the work  I\'m  going to give you.\"  Humph.  '),(24755,1196970,'30,June,2004','2004-06-30 00:00:00','   There\'s about a week left before school starts. How... ugh-y.  Pizza here, talk more later.  '),(24756,1196970,'30,June,2004','2004-06-30 00:00:00','   What the fuck is with Pepsi Vanilla now? Oh wow, let\'s buy that soda now because it\'s obviously mad cool, with the trendy ghetto music and whatever...  You wonder how some people ever make it in advertising, with some of the ideas that they come up with.  '),(24757,1196970,'30,June,2004','2004-06-30 00:00:00','   Oh great, and then my parents bitch about how nothing I do, I do corrently. What the hell, I did everything they wanted me to do... and I even made dinner for everyone because they were so fucking busy fucking up their installation of our bamboo floor. And mind you, it was a good dinner, I  can  cook believe it or not.  And then, when I offer to wash the dishes for my mom, she comes back a couple hours later and bitches about how something  I didn\'t even wash  wasn\'t done properly. Fuck off!  Rather long post, I\'m proud.  '),(24758,1196970,'30,June,2004','2004-06-30 00:00:00','   One and a half hours at IKEA, two arguments, and a dead-silent car ride home, I am now the proud owner of two ceramic heads and a spiral wooden thingy...  I could kill my parents. So goddamned  old  and stubborn... ARGH.  '),(24759,1196970,'30,June,2004','2004-06-30 00:00:00','   Finally, after an unbelievable period of discussion as to how we should put in hardwood or pseudo-hardwood or ceramic tile into the den downstairs, we settled on carbonized bamboo. It look alright, and it\'s bamboo, which is literally indestructible. Fortunately, we had to move everything out of the room, includin the computer.  So now the computer is set up in a horribly cramped corner of my room. But nonetheless, it  is  in  my  room. Muahahaha.  Soon, there\'ll be new furniture, which means I can add speakers, too. How exciting... speakers...  '),(24760,1196970,'30,June,2004','2004-06-30 00:00:00','   Yay, done with Honors English summer work. So boring... wow...  '),(24761,1196970,'30,June,2004','2004-06-30 00:00:00','   Wow, a new lunch combo.  Heirloom watermelon chunks with a side of three different Cheez-It crackers.  It\'s actually pretty good. Those funky yellow, extra non-juicy watermelons are cool.  '),(24762,1196970,'30,June,2004','2004-06-30 00:00:00','   My blog seems to filled with mostly one- or two-line entries.  That\'s rather not good.  So here\'s a three line entry.  '),(24763,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink Parts of Country May Enter Weekend Without Power : \"\'You looked up into the sky and there were stars!\' one caller was overheard describing the hours in New York when nature took back the night. \"  '),(24764,1196970,'30,June,2004','2004-06-30 00:00:00','   You know, \"literature\" is good for something...  I\'ve already gotten four bugs with  The Scarlet Letter . And that\'s today alone.  '),(24765,1196970,'30,June,2004','2004-06-30 00:00:00','   Ever hear the fable about the king and the invisible clothes?  Well if not:  Once upon a time (as always), there was a kingdom (ruled by a king). One day, these men come into the kingdom, saying they can fashion the most marvelous clothes, clothes that only the wise can see. Enchanted by the notion, the king immediately orders the men to fashion some clothes for him. The men set off to work, sewing and stitching what appears to be thin air. However, the king (not wanting to look stupid) ooohs and ahhhs along with the men as they explain to him the wondrous embroidery on the clothes. When they\'re finally finished, the king asks the men to help him put it on (as he still can\'t see it) and calls for a parade around the town to show off his new finery. Everyone that watches the parade only sees a naked, aging man walking down the street, but they (not wanting to look stupid) oooh and ahhh along with their neighbors as they pretend to admire the clothes. Eventually, the king passes a small boy, who exclaims flatly to his mother, \"Mom, the king\'s naked.\" and that\'s when he realizes that he was scammed.  Anyways, the moral of the story is...  Those clothes are literature.  '),(24766,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink Take the Test  '),(24767,1196970,'30,June,2004','2004-06-30 00:00:00','   My sister is watching \"Miracle Pets\"...  Whitewater rafting blows once it starts storming halfway through and you have to spend the next couple hours sitting in a cramped car with a bunch of people that you don\'t really want to be sitting in a cramped car with.  Chinese people can\'t barbeque. Grilling as a method of preparing food sucks to begin with.  '),(24768,1196970,'30,June,2004','2004-06-30 00:00:00','   Wow, doesn\'t it suck that walking outside barefooted right now can burn the bottom of your feet? Ow...  '),(24769,1196970,'30,June,2004','2004-06-30 00:00:00','   Coloring, what an essential skill for college. Or so my Biology book says...  '),(24770,1196970,'30,June,2004','2004-06-30 00:00:00','   Apparently, my computer was assembled 11 September, 2001. Ha.  '),(24771,1196970,'30,June,2004','2004-06-30 00:00:00','   Multi-vitamins for thought:  Everyone on your buddy list is \"away\". But... do you think they\'re actually all there and simply watching everyone else who\'s \"away\" and yet not really so and so everyone\'s not actually away and they\'re all watching... and waiting...  I doubt it.  Grape flavored multi-vitamins are nasty.  Then again, so is almost everything flavored grape.  '),(24772,1196970,'30,June,2004','2004-06-30 00:00:00','   Well, it\'s been millenia. Ta.  '),(24773,1196970,'30,June,2004','2004-06-30 00:00:00','   Bob Likes Pancakes  What an orgasmic name for a new band. Just think about it...  Oh yeah, in Montreal, there was a rather  chouette  music festival... too bad it was all in French.  '),(24774,1196970,'30,June,2004','2004-06-30 00:00:00','   Finalement , back home. Home is nice, some of the time, I suppose. It kind of sucks that you can\'t simply assume everyone speaks French here though.  Quebec was nice, mostly. It\'s simply mindblowing how annoyingly closeminded my parents can be though. They get pissed because they chose to eat at the hotel restaurant (tip: you don\'t eat as restaurants where there\'s only a single waiter) and from then on, they refuse to eat anythign but Chinese food. Add to that the fact that there aren\'t any decent Chinese restaurants in Quebec City and you get really pissy assholes that bitch and complain incessantly about every single meal we have.  Minus the fact that I never got to eat at one of those nice little restaurants that should serve amazing food, Quebec city was alright. They have a nice art museum, especially since the museum itself is a work of art (it  was  a prison). Oh yes, and after visiting the museum, there was that stone hamster incident, which really sucked.  You see, we had to bring along this other, unbelievably frustrating, family. Two little whiny children that cry and whine for ice cream  every single time  we pass a street vendor and whine and moan about walking before we\'ve even crossed the street outside the hotel. And then one of the stupid little fucks wants to buy a damned  rock  painted by some local artist to look like a hamster (20$ Canadian, too... what the hell) and the bastard loses the rock somewhere in the park outside the museum. For the uneducated, the museum is in the middle of a park; the park used to be a battlefield; battlefields are usually very large.  And so, the little boy curls up into a fetal position and starts bawling his eyes out, even though his parents are still in the museum. I really don\'t give a shit, he can cry as much as he wants. He seems to notice no one\'s paying him any attention, so he stops crying for a bit. But then, his parents walk out of the museum and, as expected, he starts screeching. Someone shoot him, please.   Next three hours are spent being forced to comb over an acre\'s worth of land, searching for a rock-shaped and -colored like a hamster...  '),(24775,1196970,'30,June,2004','2004-06-30 00:00:00','   I like this weather, it\'s nice. Not too hot for anyone, and much too cold for some. So nice... but still sunny-esque. Isn\'t it amazing? But that\'s besides the point...  Life is so evil, all those little unexpected quirks and the such. Nothing ever goes exactly how you plan it, and more often than not there\'s a little something just lurking in a dark corner waiting to bite you in the ass.  I n   t h e   a s s . Someone very wise said that life is an experience, something that... well I don\'t remember the rest. A crappy experience, that\'s what it is.  And although I may sound pissed, I\'m actually rather happy at the moment, just feel like... pondering... about life. Ponder,  songer . Fall is coming. Fall is nice. Fall ing  isn\'t, but Fall is.  '),(24776,1196970,'30,June,2004','2004-06-30 00:00:00','   What the hell is this?  \"Your buddy list is  not available , but  it has not been lost . Please try again later.\"  So you just  misplaced  it? Or maybe you don\'t want to give it to me at this time because you\'re being a spiteful ass? What the hell!  '),(24777,1196970,'30,June,2004','2004-06-30 00:00:00','   \"Hey! Watch where you\'re going!\" \"Well if you weren\'t tottering around like a fucking drunk maybe I wouldn\'t have bumped into you.\"  Sorry, just had to get that out.  '),(24778,1196970,'30,June,2004','2004-06-30 00:00:00','   You know, I think I\'ve read over my essay so many times, yet I keep finding all these little stylistic errors. So goddamn annoying, why I decided to put a comma here but not there.  Anyways, for the uneducated, below is my stream of consciousness piece for English.  Made some other stylistic changes too, but the column on the side that handles the \"links\" never seems to be wide enough. Argh, I\'ll fix it yet.  In the exact words of someone not-so-wise, enjoi.  '),(24779,1196970,'30,June,2004','2004-06-30 00:00:00','   â€œToday on the Today Show, Al Roker is in Virginiaâ€  we were walking upon the sand the waves washed up along the shore  â€œIsabel is twelve hours from landfall here at Virginia Beachâ€  look theres a shell she said its ugly I said well screw you  â€œbracing for the stormâ€  the impending tempest   â€œI need to get goingâ€  â€œByeâ€   its rather cold outside oh well but its better than being hot its wednesday wednesday the hump day thats no good.  I walked past our neighborâ€™s fake pond-barrel  so there are real fish in there fish in a barrel  first at the bus stop again no one to talk to not that I would want to talk to anyone there to begin with  ever wonder what would have happened if we hadnt chosen this house  avez-vous dÃ©jÃ  songÃ©  So the bus pulls up to the front of the edifice and unloads its cargo, I enter and walk down the corridor nearly empty  thankfully no oppressive crowds of people  apparently the construction workers have decided to stay home today  mr gehry I would like for you to build me a house a nice grand one  but I dont like this house I said well its better than some of the others dad said Im tired of looking at houses what school system mom said  Great Valley    bell   â€œYou do have a computer at home?â€ yes  but I see nothing wrong I should have tried harder to learn chinese but its so frustrating the people not the course.  â€œThis is not supposed to be an expository essayâ€  exposition no thats not it the correct word is exhibit yes exhibit matisse-picasso at moma it was nice not just the paintings but being alone one of the few places this abercrombie world cant touch but why do I care I dont I dont care then why do I complain I dont know  â€œthe day he committed suicideâ€  you think people would miss me  silly of course  life sucks  well look at the bright side youll get to meet new people  but I like it here I dont want to move  Im sorry  thanks    bell  â€œWhyâ€™s it so early?â€ â€œBecause we have an assembly.â€ he said. â€œOh really, on what?â€   bell  â€œLate, as usual.â€ â€œWell it canâ€™t be helped.â€ he said.  â€œPhotosynthesis will increase with light intensityâ€  theres no need to explain such basic concepts to us or rather there should be no need this is boring but looking around the room you wonder how why some people are here and why some arent.  â€œThe major source of light energy is the sunâ€  the suns reflection on the lab table looks nice chiaroscuro isnt it peut-Ãªtre peut-Ãªtre pas    bell  yay the fun campus next  this year sucks I dont get to see anyone at all  humph but look at the bright side youll get to meet new people  oh piss off   The corridor is empty  dr seuss I should take the time to read that mural  which is nice.  ugh some of the people here make me sick so superficial materialistic yuppie wonder what happens if they leave this bubbleworld  damn there I go again I shouldnt criticize I have no right to judge so what you make me sick  you make me sick  Im sorry  sorry doesnt help go away  oh comon  Im sorry youre moving too  thanks well ta  ta thats all you have to say ta  Im afraid so  very well then I have something Id like to say shakespeare actually julius caesar to be precise  do you intend on giving me the line numbers from the play too or would it be alright if we sped this up just a tich  humph you know I think Im going to miss your sarcasm the most  how flattering  anyways it goes something like this if we meet again we shall smile but if not then this parting was well made  awfully deep dont you think  fine then go off with your tas  bell  campus  â€œLook! Iâ€™m practically wearing pants!â€ he said. â€œVery funny, Iâ€™m borrowing this CD for the rest of the period.â€ â€œOnly for the period.â€ he said. â€œYes, but mightnâ€™t the period last forever?â€ â€œHumph.â€ he said.   Everybody knows about it, from the Queen of England to the Hounds of Hell.    bell  â€œYouâ€™re writing the paper on today?â€ he said, â€œhmm, thatâ€™s interesting.â€  B assembly. â€œThatâ€™s right, today is the anniversary of the Battle of Antietamâ€  oh I remember this from mrs dampier I miss mrs dampier I wonder how shes doing I should write her  puritans sermons fire and brimstone the pledge of allegiance  its amazing how blind some people can be its political theological social indoctrination thats what it is  I cant wait to get out of here new york or something  pah-ree maybe that would be nice maybe a semester or two at the sorbonne whatever it doesnt matter just as far from whatever the latest fad clothing store is as possible anywhere with more substance than affluent white suburbia  do you really think thats where we live  somewhat what does it matter just get me away from here  bell  â€œThereâ€™s a bee buzzing around your head.â€ he said. â€œWell leave it alone.â€  assembly time rather painful sitting in front of a speaker  When youâ€™re strange, faces come out of the rain.  bell   â€œShocktarts? Theyâ€™re not sour, theyâ€™re shocking!â€ he said.  nothing to do no one to talk to I hate this class yes hate hate is a strong word but it is just that a word I have every right to use it why am I justifying myself Ã  moi.  do you think some people here are oblivious to the fact that theres a much harsher world out there  do you think I lead a sheltered life  well frankly yes  what kind of stupid question is that of course  do you think they lead sheltered lives   bell  I still fail to see how doubly complex fractions over superfancy numerical operators by way of bobs law of metaphysical hum hum 7 factored over the set of numbers that are not equal than but less to and greater for the other set of numbers is needed in life   bell  â€œToday, weâ€™re doing our skits.â€ she said. actor moi funny-ness  so what did you think  I liked it  no you didnt be truthful  but I did youre a good actress I could never have done that  yes you could all you need is practice  are you mocking me you know perfectly well I couldnt act for my life  practice mdear practice  oh whats that I hear mockery  shut up  hahaha  bell  freedom   CD player.  Iâ€™m an outsider, outside of everything you know. It disturbs me so.  '),(24780,1196970,'30,June,2004','2004-06-30 00:00:00','   Hahaha I found a flaw in one of those Snapple \"Real Fact\" thingies. The world\'s largest lake is not Lake Superior, but the Caspian Sea.  Je gagne, tu perds, merci.  '),(24781,1196970,'30,June,2004','2004-06-30 00:00:00','   16 hours without electricity.  '),(24782,1196970,'30,June,2004','2004-06-30 00:00:00','   is also known as Hurricane Isabel. Isn\'t that strange? But yes, hopefully, we\'ll get oodles of rain and other cool things. Maybe a tree will fall into the school, preferably onto the head of a certain person. I simply can\'t wait. Simply can\'t.  '),(24783,1196970,'30,June,2004','2004-06-30 00:00:00','   How\'s this for fucked parents? Go to Home Depot (fourth time in two days, mind you), spend  two hours , and leave with a single toilet plunger. I could kill myself.  '),(24784,1196970,'30,June,2004','2004-06-30 00:00:00','   Now I can even add ridiculous and utterly superfluous titles to every post. Change is in the wind. Whether it\'s for better or worse... probably not better.  '),(24785,1196970,'30,June,2004','2004-06-30 00:00:00','   Oh la la, apparently Blogger now has spellcheck. How cool is that?  '),(24786,1196970,'30,June,2004','2004-06-30 00:00:00','   Ever wonder about the  wonderful  things that are overheard in nurses\' lounges? I don\'t either, but I should.  Life hasn\'t been very good, I think I\'m addicted to crack and a multitude of other things. Or maybe it\'s just baking flour spiked with vodka. La dee da.  I wish I could afford a hitman. Or hitwoman, it doesn\'t really matter. Just someone that can get rid of certain peoples for me. I don\'t even need a hit(wo)man, just someone to kidnap a certain someone and bring that person to someplace remote where they can get shot in the foot and drawn and quartered. I\'m pretty sure a lot of people have...  ill will  towards the AP theme reader.  '),(24787,1196970,'30,June,2004','2004-06-30 00:00:00','   Corrected the \"Bob the Eigth\" error, but I had to delete the post. So, in the spirit of boredom, I\'ll copy the previous post (which was deleted about a minute after it was made) to here.  Well I just had my weekend plans die on me. But that\'s the thing, it isn\'t really that tragic because I\'m happy to have had weekend plans to begin with. It\'s a paradox, as maybe someone would say. Or would they?  '),(24788,1196970,'30,June,2004','2004-06-30 00:00:00','   Ah shit, I spelled eighth wrong. If you people haven\'t noticed, this blog is now being kept by Bob VIII.  '),(24789,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink BBC NEWS | Americas | US anti-abortionist executed   He sounds so ignorant, or maybe it\'s just the Press that makes him out to be so... I doubt it.  '),(24790,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink Proposed Tax Rouses an Already Jumpy City   Funnyness.  '),(24791,1196970,'30,June,2004','2004-06-30 00:00:00','   White cheddar Cheez-It crumbs are lethally cheesy. Actually, so cheesy it doesn\'t even taste good any more.  '),(24792,1196970,'30,June,2004','2004-06-30 00:00:00','   Well today sucked. As expected. But nevertheless...  The stupid fucks made it a full day instead of a half day, which gave all the teachers oodles of extra time to assign essays that are due tomorrow and dozens of math problems for homework. Not to mention that I don\'t know anyone in my math class and half the people that make class more bearable aren\'t in any of my classes anymore. Oh yeah, the whole modular classrooms are located outside the school, by way of the gym and several one-way corridors isn\'t really helping either.  Fuck off.  '),(24793,1196970,'30,June,2004','2004-06-30 00:00:00','   So apparently, it\'s now the Labor Day  of 2003 . We obviously weren\'t expecting one of these, it just happened, like the Blackout  of 2003 .  Americans are so silly.  '),(24794,1196970,'30,June,2004','2004-06-30 00:00:00','   I\'m so fucking tired of this whole homework and essay distribution people seem to  expect  from me now. It\'s just a small bit of help for a few close friends, not exactly something you\'re supposed to foward to EVERY, SINGLE, FUCKING, person you know, including the teacher. What the hell, I print four copies of a preliminary outline for an essay and within a day it seems like everyone has one or wants one. Fuck off, what have any of you yuppie-spawn done for me lately? And then of course, everyone comes back within an hour to complain about how so-and-so thinks that point number  X  is invalid and shouldn\'t be used. Yeah, well so-and-so can kiss my ass. Left buttcheek, preferably.  '),(24795,1196970,'30,June,2004','2004-06-30 00:00:00','   Oh yes, must get in the last word. Fuck academic honesty. I\'ve probably distributed four different assignments to at least a dozen people (directly or indirectly) over the last twenty-four hours. So fuck academic honesty.   NHS can  kiss my ass.  '),(24796,1196970,'30,June,2004','2004-06-30 00:00:00','   This Elizabeth Smart girl can just die. Who the hell cares anymore, the only reason we cared in the first place (and by \"we\" I don\'t mean myself, I\'m cold and callous and didn\'t give a rat\'s ass about it in the first place) is because she\'s rich and white. The dumb bitch didn\'t even seem to fight back...  \"Here little girl, I have some candy for you in my car if you come with me.\"  \"Sure Mr. Man, let\'s go. I like candy, especially when Daddy fills a Mercedes with lollipops and gives it to me for my 10th birthday. Mr. Man, can  you  fill cars with candy?\"  \"No... but I have a Tootsie Roll...\"  No scratch marks on the abductor, no attempts by abducted to escape. Stupid white Americans.  And now onto the second rant...  This whole NHS (National Honor Society) thing is crap. Who cares? Academic pretentions are all crap. They make it sound all fucking speshal-speshal but really, its just a cult. You have to help all the losers after school by \"tutoring\" them and maybe they\'ll join the cult too someday. And of course, if you don\'t wear that speshal-speshal NHS pin on \"induction night\" then you\'re expelled. Yes, cult, m\'dear.  \'Tis all, folks.  '),(24797,1196970,'30,June,2004','2004-06-30 00:00:00','   Apparently, red wine now might protect against lung cancer too. Before you know it, a gallon of red wine a day might just make you immortal. Yes!  '),(24798,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink  BBC NEWS | Bomb alert over \'break-wind\' dog     For Americans, \"break-wind\" = fart.  '),(24799,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink Calling In Late   \"Phyllis F. Cohen, a psychoanalyst in New York, said one patient routinely calls from a cab to say she is 10 blocks away. She begins her therapy on the phone, but hangs up when she reaches the elevator for fear of being overheard. (As for the cabbies, Dr. Cohen said, \'She\'s convinced they don\'t understand English.\')\"  Damn Americans are stupid.  '),(24800,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink Google Sets   It makes sets out of everything! Even {apple, ballpoint pen, the moon, stockings, speakerphone}. Like wow.  '),(24801,1196970,'30,June,2004','2004-06-30 00:00:00','   You know, Opera is damned useful.  '),(24802,1196970,'30,June,2004','2004-06-30 00:00:00','   Dragged out of bed at 8:30 in the morning so we can do that whole hiking thing as a family.  You notice how all the really \"oh la la look at me I so outdoor-sy\" people have those walking sticks? Yeah well it seems that the  ultra  \"oh la la look at me I so outdoor-sy\" people have  two  walking sticks, as it is so obviously the trendy thing to do.  Outdoor-sy people annoy me. Die.  '),(24803,1196970,'30,June,2004','2004-06-30 00:00:00','   Holy shit! This blog is nearing five months old. Or five months worth of archives... same thing. Yay some more.  '),(24804,1196970,'30,June,2004','2004-06-30 00:00:00','   That\'s right folks, at the low low price of twenty-eight for fourteen dollars. Get them while they last. Amaze your friends.  I forgot this thing had spellcheck. That would\'ve been useful for that other post.  Now that I\'ve tried using it, I think I would\'ve been better without it. It tried to replace \"boogers\" with \"splices\". Die. And oddly enough, \"spellcheck\" isn\'t in the dictionary. I suppose it\'s useful enough, if used sparingly, like... like... oh yes, like  toe boogers .  One of the few times I\'ve managed to tie it all together with a title. Yay.  '),(24805,1196970,'30,June,2004','2004-06-30 00:00:00','   Damn I already see mistakes.  '),(24806,1196970,'30,June,2004','2004-06-30 00:00:00','   Nothing else to do, and I don\'t believe I\'ve done a real update of this in a while, so here goes.  Parents at another annoying Chinese...  box social , shall we say. I\'m just happy they\'re not nagging at me for one thing or another. PSAT\'s were today, that was crappy. What the hell is \"standard written English\"? There shouldn\'t be such a thing - it\'s a language, it\'s malleable; just because some dumbfuck who thinks they\'re the English God or something says that that question mark back there should go  inside  the quotation marks doesn\'t mean I have to do it. Fuck him. Or her. Or it. Yes fuck it. Fuck it all. Not sure how I got to  that , just ranting I suppose. OK done now, onto another topic.  In a tich, AOL Instant Messenger on.   Oh yes, what a piece of shit  that  is. Not only is their website an aesthetic atrocity, you\'re forced to download only to latest completely and utterly commercialized, bloated-with-ads version there is. It\'s not like there were any improvements between v. 1 and v. 2 798 345 or whatever the hell it\'s on now. Maybe the addition of smilies... ooh ahh... fuck it.  Gee I sound really pessimistic. Happy thoughts... happy thoughts...  So anyways there\'s that song about the bunny that hops through the forest, bopping all the animals on their heads. Is it \"Little Bunny FOO FOO\" or \"FROO FROO\"? The first is what \"they\" say it is (\"they\" should die), but the second sounds better. Does it matter? No, not really, but it\'s one of those questions of life that make this world interesting.  Yeah, that\'s about long enough, I don\'t think I\'ve ever typed so fast ever before without making too many errors. Hip-hip-huurray. Cyanide-filled pinatas for  everyone .  And there, I\'ve managed to use  italics ,  bold type , AND  underlining , all in the same post. Life is good.  '),(24807,1196970,'30,June,2004','2004-06-30 00:00:00','   And after millenia of tinkerng, it works. Hurrah.  '),(24808,1196970,'30,June,2004','2004-06-30 00:00:00','   Opera browser doesn\'t work very buddy-buddy with the whole \"BlogThis!\" concept. Damn.  '),(24809,1196970,'30,June,2004','2004-06-30 00:00:00','   Nothing to say, other than I\'m content, for some odd reason. It feels weird being content even though you don\'t know why. I guess it\'s because I have no obligations to fulfill at the moment. That\'s always a good thing I suppose. This is getting too long, end.  '),(24810,1196970,'30,June,2004','2004-06-30 00:00:00','   Nothing to say, other than I\'m content, for some odd reason. It feels weird being content even though you don\'t know why. I guess it\'s because I have no obligations to fulfill at the moment. That\'s always a good thing I suppose. This is getting too long, end.  '),(24811,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink Despair.com   \"The downside the being better than everyone else is that people tend to assume you\'re pretentious.\"  '),(24812,1196970,'30,June,2004','2004-06-30 00:00:00','   \"\'Kill Bill: Vol. 1\' is rated R (Under 17 requires accompanying parent or adult guardian). It has shootings, stabbings, beatings, beheadings, disembowelings, amputations, mutilations, eye-gougings, slicings, choppings, bitings and a spanking. Also some naughty words.\"  '),(24813,1196970,'30,June,2004','2004-06-30 00:00:00','   What a deal! A two-in-one post, just like those speshal speshal magazines with their \" Weddings  Double Issue\" and crap. Except part A is in no way related to part B.  The best stress relief? Cook. Or, rather, chop away at something until it turns into an undefinable pulp. Just imagine that something to be a finger or whatever of someone. Or maybe not someone, if that\'s not your style. But then that\'s not my problem.  There\'s no plural \"you\" in English. And that sucks. Firstly because we\'re demoted to using the phrase \"you guys\" or the even more horrible \"y\'all\" and secondly because it sounds really awkward to say \"and now, you plural die\" or \"and now, y\'all gotta die\".  In retrospect the two  are  related, but in a very weird way...  '),(24814,1196970,'30,June,2004','2004-06-30 00:00:00','   I think this \"new outlook on life\" thing is going alright, compared to every other times I\'ve tried to change. But we (unspecified) were talking about it today and someone\'s name popped up. I\'d like to know why that particular name, but I did say it made sense.  And no, this doesn\'t make sense, but I usually don\'t. So there.  '),(24815,1196970,'30,June,2004','2004-06-30 00:00:00','   Things, no...  people  are never who they seem to be.  That\'s something one should keep in mind more often.  The title has nothing to do with this posting.  The person you expected to be...  unpleasant  may actually turn out as a much deeper person whereas the friend you thought you had is much more interested in the clothes you wear, the gum you have, the people you hang around... What fucking superficial, judgemental bitches.  At times, I think I\'m too nice, too trusting. But it\'s all in good fun, or so its expected to be. So then what happens if its not? Frustrations boil over and I end up not talking to people for a day, a week, a month? Is that even a logical approach? I\'m  not  bipolar  and I  don\'t  get depressed. I\'m talking to myself again...  A logical approach. A logical approach would be to stop trying so fucking hard to please some stupid shithead that cares more about who you\'re wearing than who you are. A logical approach would be to stop typing. Yeah.  '),(24816,1196970,'30,June,2004','2004-06-30 00:00:00','   We never throw anything away. Ever... What the hell?!? The violin I attempted to play in third grade that now has a crack in it the size of that Liberty Bell thing is still around. \"We might need it someday.\" Yeah, like when the world freezes over in a nuclear winter and kindling wood becomes a precious commodity, we\'ll break apart the violin to start the fire to cook the six-headed goose for our dinner.  '),(24817,1196970,'30,June,2004','2004-06-30 00:00:00','   Nuggets of wisdom:  Taking stairs one-and-a-half steps at a time is rather painful.  '),(24818,1196970,'30,June,2004','2004-06-30 00:00:00','   End of vaykayshun. Damn. Depressed. And we had an entire week too. And we have stuff due tomorrow. Oh well I\'ll survive.  So it was alright I guess. Filled with people. People are annoying. So anyways for some reason my keyboard isn\'t typing particularly well so I\'ll stop for now.  I haven\'t updated in a while. I feel obligated.  Nostalgia is a powerful force, I realized that just recently. And it sucks. Die.  '),(24819,1196970,'30,June,2004','2004-06-30 00:00:00','   Going on six months! Hooray for me. So yeah today we\'re \"entertaining people\". How horrid.  '),(24820,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink Turn Left: Make Your Own Conspiracy Theory   What They Don\'t Want You to Know In order to understand love you need to realize that everything is controlled by a cookies made up of little Jewish children with help from three-legged women.  The conspiracy first started during Battle of New Orleans in Jakarta. They have been responsible for many events throughout history, including Bob\'s Second Birthday.  Today, members of the conspiracy are everywhere. They can be identified by blinking.  They want to fuck the Knights of Columbus and imprison resisters in New York City using geese.  In order to prepare for this, we all must piss. Since the media is controlled by your mom we should get our information from Dubya.  '),(24821,1196970,'30,June,2004','2004-06-30 00:00:00','   So the national day of gluttony is over.  What a pity.  And so the national day of shop-shop-shoppinh begins.  Yay? Or maybe not.  So anyways, I hate you all.  '),(24822,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink FUNNY.  '),(24823,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink GloFish   Acid-laced fish!  '),(24824,1196970,'30,June,2004','2004-06-30 00:00:00','   A week off. Yes. No? Maybe so.  '),(24825,1196970,'30,June,2004','2004-06-30 00:00:00','   Thoughtful quote (in reference to the Guggenhein Museum Bilbao):  \"Museums are the new cathedrals.\"  '),(24826,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink Defend your castle.   Holy shit, you can throw people around, it\'s fun.  '),(24827,1196970,'30,June,2004','2004-06-30 00:00:00','   What the hell, is it supposed  my  fault if AOL Instant Messenger can\'t stay connected for over 10 minutes? And of course, then comes the \"you\'re trying to sign on again too soon, please try again later\" error. Screw you.  '),(24828,1196970,'30,June,2004','2004-06-30 00:00:00','   Oh, oh, wait. You can\'t use American because that\'s being an ethnocentric bastard. So in a couple days, we shall be continuing the  United Statesian\'s  quest to turn this world into a living hell. So much better.  '),(24829,1196970,'30,June,2004','2004-06-30 00:00:00','   Posting because I\'m bored. Apparently someone\'s employer is visiting tonight. Not mines...  Thanksgiving vacation in a day! Yes! Time to celebrate and hopefully continue the American\'s quest to make the world a less hospitable place.  Some people can be so incompetent, but then again, people are easily fooled. If you  can  get away with it, why not?  '),(24830,1196970,'30,June,2004','2004-06-30 00:00:00','   Just saw the new Matrix movie. In IMAX... ooooh.  '),(24831,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink After 40 Hours, Democrats Succeed in Blocking 3 Nominees    \"The White House reacted with anger. \'Once again, a partisan minority of senators has thwarted the will of the majority and stood in the way of voting on superb judicial nominees,\' President Bush said in a statement issued by his spokesman, Scott McClellan. \'These obstructionist tactics are shameful, unfair and have become all too common.\'\"  And something else:  \"South Carolina senator [Strom Thurmond] who in 1957 famously held the floor single-handedly for 24 hours 18 minutes to oppose a civil rights bill, a record still unmatched.\"  And that makes him speshal speshal somehow?  '),(24832,1196970,'30,June,2004','2004-06-30 00:00:00','   Updating because I was told to.  Friday, yes! Today wasn\'t at all that bad, although maybe I\'m just trying to \"appreciate\". Damn those essay test, though.  Mody Dick  the movie has nothing to do with American History. Melville should have been a miscarriage. New insult of this indeterminate period of time!  Talk more later, french composition to write.  '),(24833,1196970,'30,June,2004','2004-06-30 00:00:00','   What, exactly, is the \"American Dream\"?  It\'s to own a yard. And why, do you ask? So one can do yardwork. Why do yardwork, exactly? Well, because it\'s an exercise in futility, and Americans seem to like that.  '),(24834,1196970,'30,June,2004','2004-06-30 00:00:00','   Fallen oak leaves are as ugly as hell. Well, as compared to that red maple next to the bus stop.  I don\'t see why our essays should be graded by someone whose attention span can\'t even make it to the end of a paragraph, much less an essay.  Standard written English was a term coined by the uncreative.  I\'m surrounded by idiots.  All the incompetent teachers seem to congragate around me.  If I had a digital camera, I would take a picture of the aforementioned red maple, but then again it\'s a bit too late. But that might make it a better picture.  When one gets an essay returned, it\'s rather disturbing that in the four days it\'s been away, it\'s suddenly acquired the smell of cigarettes. And that smell has presisted for many weeks.  Short. Succint.  Parfait   As a closing note: People have interesting life stories, if one takes to time to listen. School only gives one the opportunity to learn not from the teacher, but from the people.  Or at least that\'s how it ought to be.  '),(24835,1196970,'30,June,2004','2004-06-30 00:00:00','   Less pissed now. Ohhh, crazy crazy posts during school. I\'m surrounded by people all updating their Deadjournals of whatever they\'re called. Muahaha, I so speshal.  So yeah anyways we got our essays back in English. People don\'t seem particularly happy with their grades. And apparently, she hinted that I\'m in the process of fucking my self over. Oh, i love to fuck myself over, it\'s so much fun.  What the hell, I\'m bored.  Ta .  '),(24836,1196970,'30,June,2004','2004-06-30 00:00:00','   The next big thing: pomengranates.  If you haven\'t had one yet, go out and try one. Now. If you look hard enough, you might be able to find some for sale at a local supermarket. Otherwise, you\'ll have to try some international foods market etc. etc. They were plentiful a bit ago, but either Americans are stupid or the growing season is short.  I like pomengranates.  '),(24837,1196970,'30,June,2004','2004-06-30 00:00:00','   That came out plus angry than I expected. Sorry.  '),(24838,1196970,'30,June,2004','2004-06-30 00:00:00','   Ok, so there hasn\'t been a real update in a long time. And fresh meat is coming by sometime soon to look at this, so I suppose I should write something meaningful. (Yes, you people know who you are...)  So, there\'s a lot of work to do, but if you take it all in stride, everything is supposed to end up all peachy. At least that\'s what the hope is... Nevertheless, things always seem to work out if you manage your time well, and no need to stay up until midnight either.  Today is Monday. What does that mean, you might ask? Today is naturally bad, even if it\'s not really that bad. But was it really that bad, you ask? No, not really. But it was still bad, because it\'s Monday. Monday. Wasn\'t there some stupid evangelist named that?  Evangelists are annoying. That, and they can be so damned pretentious. What the hell makes you think you\'re right? For all you know, you could\'ve been some flop of an experiment by the citizens of Madagascar to breed a super-intelligent warrior for their schemes of world domination. I\'m aware none of you know of the evil plans of the Madagascar-ians, but that\'s alright. You know now, and knowledge is power. So anyways, maybe they fucked up in the breeding process and they ended up with an incredibly stupid, stubborn, and gullible chunk of flesh that is supposed human. What do they do with it? Well, of course, they dump their trash in the same place every other country drops its stupid masses... America.  Oh yes, and fuck you if you think it\'s being ethnocentric to say \"America\" when you so obviously mean \"The United States of America\". Well if I so obviously mean it, then you so obviously know what I\'m talking about, so go get eaten by some cannibals in the South Pacific.  Anyways, Mr. Stupid/Stubborn/Gullible gets dropped in New York. We\'re assuming, because I think flights from Madagascar ( Are  there flights from Madagascar?) should be cheaper going to New York. And I so obviously mean New York City, yet no one seems to bitch about  that . Anyways, he\'s destitute and homeless, so he gets taken in by... oh la la... some Catholic charity. Simmer on low heat for eight months and you have a new evangelical preacher that has plans of being either institutionalized or incarcerated for murder in three years, give or take. Three years,    mark my words   , three years.  '),(24839,1196970,'30,June,2004','2004-06-30 00:00:00','   So I haven\'t updated in a while. So so. But there\'s nothing eventul happening...  Damned parents.  '),(24840,1196970,'30,June,2004','2004-06-30 00:00:00','   For a lack of productive things to do, I\'ve resigned myself to posting here. Lack of things to do. But then again this isn\'t exactly spectacularly exciting.  Intellectual copyrights are a pain, you realize that? Only humans would come up with stuff like this; but then again only humans would attempt to pass off or simply take another\'s work as their own.  The Digital Millenium Copyright Act is rather... crap. It\'s prohibitive and nearly  carte-blanche . It would be nice to own one\'s own sovereign nation - other than the ovious benefits you can drink, gamble, fill your swimming pool with heroin... Moreover (such a cool word), you could kill people at leisure... well maybe. I\'m sure if they ever caught you, there would be hell to pay in front of an International Tribunal on Justice or something, but eh, it would be great fun while it lasted.  '),(24841,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink Sodaplay.com   It\'s interesting, albeit a tich difficult.  '),(24842,1196970,'30,June,2004','2004-06-30 00:00:00','   So now back to where I was before I was so rudely interrupted...  It\'s almost like a swimming pool filled with warm pie. Apple, preferably, but without that much cinnamon. And without the sexual connotations. Or maybe fudge, if pie isn\'t your thing.  But what exactly is someone\'s \"thing\", you ask? Or, well, I ask. I\'m asking...  Oh yeah and I couldn\'t get a place on the Model United Nations thing. I\'m disappointed, I suppose, but not necessarily angry. It  was  mostly my fault for attempting to join last-minute, damn procrastination.  So anyways that\'s enough sentimentality/pie for a day.  P.S. The new  it  word for English class is  precocious . Don\'t be so damned precocious.  '),(24843,1196970,'30,June,2004','2004-06-30 00:00:00','   So school got let out early today because of the snow. It\'s nice though,  je suis heureux . I\'m warm (enough) and relaxed (a rarity).  So no complaints today... yay. Rack of lamb with lemon crusted herb butter with a side of cranberry-based reduction. I\'m hungry now.  The lamb thing was completely random, but I felt like putting it in there. I feel like that.  '),(24844,1196970,'30,June,2004','2004-06-30 00:00:00','   Two, and quite possibly three, snow days in a row. Which means that the mid-term schedule has been completely fucked over. I leave Thursday for a week\'s worth of trips to Model UN and Washington DC, so I\'ll have everything to make up once I get back. I\'m so fucked. Yes, you should be feeling sorry for me. So, so, sorry.   La vie, il suce . Not proper  francais , I know, but  sucer  is such a nice verb, I just had to use it.  Okay, bye.  '),(24845,1196970,'30,June,2004','2004-06-30 00:00:00','   The pre-eminently special, definitively excellent, superbly confounded way to spell okay is just that, \"okay\".  The end, okay?  '),(24846,1196970,'30,June,2004','2004-06-30 00:00:00','   I\'m as professional cynic but my heart\'s no int it.  - Blur,  Country House    So yeah I was walking outside today and there was a bit of spit on my cheek. Where is this going, you wonder? Well it froze. On me.  As this is pointless, I\'ll stop now and write more later.  '),(24847,1196970,'30,June,2004','2004-06-30 00:00:00','   Which is also why I\'m sitting here eating the second part of my dinner: an apple. A small one, too.  I\'m hungry, help.  '),(24848,1196970,'30,June,2004','2004-06-30 00:00:00','   Ok, my parents have fucking lost it. Who the hell asks for the boniest fish in the supermarket? And of course they decide to buy a dozen Cornish game hens (aka micro-chickens). Then they take one and go apeshit on it, chop it into thirty pieces (bone and all), and \"stiry-fry.\" Let\'s see how quickly I can make everyone choke on a small piece of bone covered in what looks like shit. And as for the fish, we spent eight long agonizing days attempting to eat it. The first couple days we could make passable attempts, but after that the bone got so fragmented it was like trying to eat a pincushion. Albeit a pincushion in a really nice sauce. And yes you should be feeling sorry for me. And yes I\'m thankful for your sorry-ness.  '),(24849,1196970,'30,June,2004','2004-06-30 00:00:00','   Deleted all those old links on the sidebar, will add some more later.  If you care for reasons:  Atomfilms is an ass and only works on Internet Explorer or Netscape. gnod I should actually keep. And the last one was just eh...  '),(24850,1196970,'30,June,2004','2004-06-30 00:00:00','   Alright, so apparently some of you are bored of listening to the magic of pomengranates. And apparently some Catholic psychopath decided to turn every blog hosted by  urlLink blogspot  into some crap about existentialism. Or maybe it was just a glitch on the system. Most likely a glitch if you think about it, an intelligent Catholic is rare,  Jews for Jesus  rare. What the hell is  Jews of Jesus  anyways? Isn\'t it an oxymoron?  Lots of work I\'m obligated to do, so I\'ll get back to my religion-bashing later. Ta.  '),(24851,1196970,'30,June,2004','2004-06-30 00:00:00','   People can be horrible. Not that I\'m really pissed at the world, it was just an observation whilst in Philadelphia today.  And another observation: even with cell phones, it\'s hard to reach people somehow. Why? What the hell? Keep your cell phone with you at all times, if possible. Not too difficult.  '),(24852,1196970,'30,June,2004','2004-06-30 00:00:00','   Nothing to say.  '),(24853,1196970,'30,June,2004','2004-06-30 00:00:00','   Yeah so I did something very stupid today. How stupid? Well, stupid enough that I refuse to talk about it.  But anyways, a fourth of the school\'s gone to Disneyworld for the music program trip. And since there\'s a direct correlation between musical aptitude and intelligence, that means half my classes are half empty. It\'s kinda cool, actually, except that there\'s only five of us in French now. They\'ll come crawling back though, give them a week I think it is.  As for this post, it\'s done.  '),(24854,1196970,'30,June,2004','2004-06-30 00:00:00','   I feel like I\'ve gotten letters from every single college in existence.  But , I did get two letters I actually cared about, one from NYU and the other from Columbia. Yay. Say \"yay\" with me. Yay.  '),(24855,1196970,'30,June,2004','2004-06-30 00:00:00','   You know what\'s horrible? The fact that those annoying banner advertisements which ask you to \"Shoot the _____\" always take you the same place, no matter what. Even if you miss the _____ completely (and end up hitting the person sitting next to you...), you still \"win\". How depressing, it sucks when everyone \"wins\". Because then there are no losers, which means there are no winners. And I like being a winner. Yay me.  '),(24856,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink AIR   There we go...  '),(24857,1196970,'30,June,2004','2004-06-30 00:00:00','   Holy shit, the new album by  Air  is amazing.  Talkie Walkie  is the name of it, so go out and buy a copy. Now. I order you to.  And for those of you who don\'t know or listen or care for  Air , you suck. But that\'s beside the point. It\'s all good shit, so go buy a copy of anything ( Moon Safari, 10,000 Hz Legend, Premiers Symptomes, or Talkie Walkie ) and listen to it.  And in reality I would get you a link to their web site, but I can\'t find it for the life of me right now. So go Google it yourself.  '),(24858,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink \"Oh my God! There\'s an axe in my head.\"   Smile.  '),(24859,1196970,'30,June,2004','2004-06-30 00:00:00','   Added links. Someone should go to \"That Special Place\". Lots of fun, yes it is, yes it is.  '),(24860,1196970,'30,June,2004','2004-06-30 00:00:00','   Changed nickname back. Was seriously considering using any one of my numerous (and annoying, mind you) nicknames bestowed by random people. The first on the list was \"Nanimal\". As you can tell, I chose not to.  '),(24861,1196970,'30,June,2004','2004-06-30 00:00:00','   I just realized, I\'ve kept this going for almost nine months now. That\'s pretty impressive, it shows committment.  Oh yeah, SAT scores came in. Damn them. I still plan on screwing over The College Board when I become rich and powerful. I\'ll have them all assassinated. No, murdered; they\'re not that important.  '),(24862,1196970,'30,June,2004','2004-06-30 00:00:00','   Yet another quick update before I get bored.  Art is cool. As I stated earlier today (not on this, but just verbally), I like beauty in all its forms. Yes, I suppose that means even verbal beauty. Eck.  Here\'s a cool link:  urlLink The Calder Foundation    That  is beauty in multiple forms. Add in a nice caption, play some Incubus in the background, and you\'re set. I\'m sure if you were allowed to lick them, that\'d be pretty interesting. Not interesting, beautiful.  C\'est la vie. Peut-etre qu\'elle est jolie, peut-etre qu\'elle vous tue.  Nonsensical rambling...  '),(24863,1196970,'30,June,2004','2004-06-30 00:00:00','   Long time, no update. I\'m aware, and currently working on a major super-duper extra-fantastic completely filled with cool-ness update. But until then, this will suffice. By the way, I have a valid excuse: I was doing both the Yale Model United Nations and Close-Up: WASHINGTON DC.  How were they, you ask? About the same, really. The former had more freedoms, as you could eat everything on your own time (in other words, no hotel food...) and there was actually free time. The latter? They babied you, but the people were better (as opposed to being around 1000 private prep school bitches). The food was horrible (and we all know my life revolves around food), especially on Capitol Hill. Our government cannot work effectively on a diet of \"daily specials\" that differ only in the color of ingredients.  As such, I call for an immediate resolution that our government kidnaps Alain Ducasse for the express purpose of feeding our Congress-people. Or, we let them all starve and our government descends into anarchy. Both are perfectly acceptable. For the non-food-obsessed, Alain Ducasse is quite possibly the most famous chef in the world.  But in reality, some of those Close-Up people were pretty pretentious also. And Dave,   NEVER UNDERESTIMATE THE POWER OF STUPID PEOPLE IN LARGE GROUPS   NEVER UNDERESTIMATE THE POWER OF STUPID PEOPLE IN LARGE GROUPS   By the way, it\'s by \"Unknown\". Cool guy, that Unknown.  '),(24864,1196970,'30,June,2004','2004-06-30 00:00:00','   This is cool.   urlLink Google Deskbar  '),(24865,1196970,'30,June,2004','2004-06-30 00:00:00','   Seriously, does the Christian Coalition of America  have  to use such a scary picture to represent themselves?     It\'s almost as bad as Clonaid and  their  scary press-conference-scientist-lady. Do a Google image search for \"Brigitte Boisselier\" if you want to scare yourself shitless.  '),(24866,1196970,'30,June,2004','2004-06-30 00:00:00','   Things are looking up. By \"things\" I meant life, in general. Well, not really if you think about it closely, but who does that?  Julie transferred away to somewhere. Why? That\'s the big point of contention these days. I hope you\'re happy wherever you are. Or at least dealing with it. Come back next year, possibly? I hope.  But there\'s nothing else major going on. And that\'s a good thing really. I mean, there\'s always the one (or two, or a million) thing(s) you want but you can\'t have. Not at the moment at least.  Par exemple , I want an iPod.  This is a weird posting. For once I\'m not sorting through stuff, just taking it in stride. Less stressful, I suppose.  Okay, bye.  '),(24867,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink massive attack   One of the cooler websites out there.  '),(24868,1196970,'30,June,2004','2004-06-30 00:00:00','   It hailing and what-not outside, and frankly I think the sky in general is trying to screw everyone over, but it  is  absolutely gorgeous. The trees, I mean. Covered in a thin sheath of nearly-transparent ice, everything glimmering in the wind. I suppose I\'m trying to say it looks nice if you\'re indoors.   Thievery Corporation - Vivid . Very Nice.  '),(24869,1196970,'30,June,2004','2004-06-30 00:00:00','   Et ce n\'est pas  si  mauvais . No, it really isn\'t; I mean, I don\'t have everything behind me that I would like to have behind me, but life is survivable. More than survivable... I feel fine. Plus, the sun is shining in that quasi-warm way, and it\'s windy. Rather a neat combination, if you think about it.  I may. I would. I shall.  '),(24870,1196970,'30,June,2004','2004-06-30 00:00:00','   Not to get into details, someone I knew vaguely committed suicide. I\'m still not sure how to react, it seems too surreal. It also seemed such an unnatural move. Why? But frankly, isn\'t that the question everyone asks when confronted with death?  I guess whatever precipitated such a move as that must have been horrible, obviously much more monumental than my own problems. And, however callous this may make me sound, everyone needs to know someone who has committed suicide. It makes you think twice about doing the same thing yourself. Everyone toys with the notion at one point or another, but when you see how it can adversely affect thousands of people, even people you didn\'t know well, you have to stop and think.  '),(24871,1196970,'30,June,2004','2004-06-30 00:00:00','   That we don\'t even care to shake these zipper blues And we don\'t know just where our bones will rest To dust I guess Forgotten and absorbed into the earth below The street heats the urgency of now As you see there\'s no one around.  - The Smashing Pumpkins  '),(24872,1196970,'30,June,2004','2004-06-30 00:00:00','   In essence, that was sopisticated venting. Well, I guess a sohpisticated word for sophisticated venting is \"cartharsis\". So, that was a catharsis.  '),(24873,1196970,'30,June,2004','2004-06-30 00:00:00','   How do I feel? I\'m not sure. It\'s weird. This post is more for myself than you, so you might as well just stop reading.  Weird, or any of its synonyms, would be a great way to describe the past 48 hours. Or 36, even. Who cares, just the past indefinite period of time starting with Sunday-ish and ending right around now. So many major assignments, but that really wasn\'t the biggest event. Some tension from having to present a seminar, but also from life in general. There was an unexpected moment of kindness, which did do a great deal to brighten my spirits. Thank you, although you do not read this. But you know, sometimes one just has to be optimistic.   C\'est la vie , take it or leave it. Leaving it generally involves a tall building, poison of some sort, or a gun. Taking it brings sorrow, happiness, disappointment, humiliation, and laughter. It\'s a difficult choice, since you never know whether a life defined as a prolonged period of pain is truly worth living for. Not that I\'m advocating suicide by any means. Just the deranged wonderings of a battered spirit. No, too melodramatic. Just some musings by a spirit. Better.  As for the portion that some of you  may  understand:  Thank you. Not just the obvious ones, but even the people I may never talk to again. In some ways, this is a gracious exit (not from life itself, just from a series of events). And Julie,  je t\'adore . And thank you, not just because you guys have always been there. Well, always as in the sense of \"as long as we\'ve known each other\". But thank you for whatever unnamed feeling it is that people can sometimes convey to each other. That, and a newfound appreciation for Pink Floyd.  And as for a certain person I hope never to talk to again. Thank you in any case. And I suppose I\'m sorry. Not that you\'ll read this, but for whatever reason, I needed to get this out.  I seem to be waxing really sentimental. Oh well...   Whatever tomorrow brings, I\'ll be there with open arms and open eyes.  - Incubus  '),(24874,1196970,'30,June,2004','2004-06-30 00:00:00','   They ought to be on the plane back from Denmark right now. Must suck for them, leaving Daneland for... well...  this  place. Not that there\'s a reason we have a Danish exchange program to begin with, no one speaks Danish. I\'m assuming that\'s a language. Oh well, the best we can hope for is that they step off the plane at least mostly sober.  '),(24875,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink s t e r e o l a b   So pretty...  '),(24876,1196970,'30,June,2004','2004-06-30 00:00:00','   They eat hamburgers on lettuce. That\'s revolting, even though I wouldn\'t ever eat a hamburger to begin with. It\'s the principle of the thing.  Whatever, I\'m thin and fine with it. The fat people can deal with themselves. Enough of this nonsense.  '),(24877,1196970,'30,June,2004','2004-06-30 00:00:00','   Aw, how sad. According to  Time , \"[t]hose slim, wine-drinking, chain-smoking Europeans are laughing at us [Americans].\"  But then there\'s that gross low-carbohydrate thing going around, making it now fashionable to eat a pizza composed of nothing but mozzarella, sauce, and toppings. So when you drop over dead from a \"major cardiac event\", I\'ll laugh. Yes, I\'ll laugh just as hard as those quasi-sober but very thin (and isn\'t that what really matters?) Europeans.  '),(24878,1196970,'30,June,2004','2004-06-30 00:00:00','   I rescind my opinion of the Google Deskbar as cool.  Well, it still is; it\'s just that it doesn\'t integrate well with Mozilla (Netscape), whereas it does with Opera. In any case, it\'s kind of slow. But that\'s to be expected - it uses Internet Explorer.  '),(24879,1196970,'30,June,2004','2004-06-30 00:00:00','   Redid some of the links.  Who doesn\'t like free sex?  '),(24880,1196970,'30,June,2004','2004-06-30 00:00:00','   Omigod.   What the fuck is that?  I\'m jumping to random places on the Internet and I land on the E! website (which in itself is terrifying). But then they go and use  that  phrase-word or whatever it is. Ugh.  '),(24881,1196970,'30,June,2004','2004-06-30 00:00:00','   Oh! News flash! In two months this blog turns one. Yay for it.  '),(24882,1196970,'30,June,2004','2004-06-30 00:00:00','   Well I got Mozilla working. There are imperfections, but the mail client is improved over Opera\'s. In any case, either is better than Internet Explorer. Surfing the Internet through the reflection on the back of a spoon is better than Internet Explorer.  But besides that, nothing eventful. Mike and Company is playing  urlLink NationStates , some political country simulator. Intriguing enough, I suppose. No espionage, war, or nuclear holocaust, however; what a pity.  In other news: according to reliable sources I have \"mastered the art of saying nothing\".  Damn, that\'s mighty exciting.  In yet other news, there was an awfully humiliating event that happened today. Well, two. In the first, a certain person who shall remain unnamed decided to chuck a plastic bottle of orange juice at me. For no reason. The second, I decided to bop the bookbag of a person who I though was a different person. So the person, who obviously was not the person I was looking for, turned around and gave me the classic \"Oh my god have you fucking lost it? Who the hell are you? Go screw yourself.\" look. That was embarrassing. But Alex and I had a good laugh about it afterwards.  And then there\'s prom, which is in two weeks. Damn, it\'s edging ever closer. And I still need a tuxedo. And we still need to make the trance soundtrack for the limo. For no purpose but to be able to step out of a limo (which has fiber optics and strobe lights, by the way) playing techno music and act like Eurotrash.  La vie est belle .  '),(24883,1196970,'30,June,2004','2004-06-30 00:00:00','   Ticking away the moments that make up a dull day  You fritter and waste the hours in an offhand way.  Kicking around on a piece of ground in your home town  Waiting for someone or something to show you the way.   Tired of lying in the sunshine staying home to watch the rain.  You are young and life is long and there is time to kill today.  And then one day you find ten years have got behind you.  No one told you when to run, you missed the starting gun.   So you run and you run to catch up with the sun but it\'s sinking  Racing around to come up behind you again.  The sun is the same in a relative way but you\'re older,  Shorter of breath and one day closer to death.   Every year is getting shorter never seem to find the time.  Plans that either come to naught or half a page of scribbled lines  Hanging on in quiet desperation is the English way  The time is gone, the song is over,  Thought I\'d something more to say.    Home, home again  I like to be here when I can  And when I come home cold and tired  It\'s good to warm my bones beside the fire  Far away across the field  The tolling of the iron bell  Calls the faithful to their knees  To hear the softly spoken magic spells.  '),(24884,1196970,'30,June,2004','2004-06-30 00:00:00','   Oddly enough, there are always five people signed on-line at any given time. As soon as someone signs off, there\'s generally someone else to replace them. Creepy, no?  I\'ve somehow spent the last half-hour listening to remixes of the same song ( Air - Don\'t Be Light ). I feel like shit.  But  my parents are out for the weekend, so it\'s all good.  And I am now the proud owner of a dozen or so bug bites. Damn, life is good.  '),(24885,1196970,'30,June,2004','2004-06-30 00:00:00','   I just bit through my lip. And now it\'s bleeding.  Go me.  '),(24886,1196970,'30,June,2004','2004-06-30 00:00:00','   So my sister went over to an  acquaintance \'s house yesterday and somehow they decided to give her the whole Jesus make-over. So she decided to spend today bitching and moaning about how we don\'t have a Bible and about how we need to go to church. And as a souvenir from yesterday\'s indoctrination, there\'s this little plastic cross. And you know what? The damned thing only fits into its base one way - upright, presumably to keep any heretics from defiling it. What the hell, bastards.  '),(24887,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink Why  the Barnes Foundation is the way it is.  '),(24888,1196970,'30,June,2004','2004-06-30 00:00:00','   \"Beautiful, mysterious, haunting, invariably fatal. Just like life.\" - The Virgin Suicides   Finally  saw the movie. Intriguing, somewhat because of who it\'s directed by and who the score comes from.  '),(24889,1196970,'30,June,2004','2004-06-30 00:00:00','   Finally, bought  Parts of the Process . Wow, what can I say. I mean, obviously they\'re an amazing band; but listening to their eighteen \"most memorable\" songs linked together is just... wow. Awe-inspiring may be a better term, although I rarely call anything \"awe-inspiring\" and mean it. It\'s hard to awe someone, and becoming more difficult daily, with the world so desensitised and whatever.  I do believe my zen is being disturbed. Go away so I can listen in peace.  '),(24890,1196970,'30,June,2004','2004-06-30 00:00:00','   Told you he was an ass.  '),(24891,1196970,'30,June,2004','2004-06-30 00:00:00','   \"The proposed changes must now pass the government-dominated lower house and then the Senate, where  minor parties hold most of the power .\"  The source is  urlLink BBC NEWS | Asia-Pacific | Australia to prevent gay marriage .  '),(24892,1196970,'30,June,2004','2004-06-30 00:00:00','   The anger stems from  urlLink here .  '),(24893,1196970,'30,June,2004','2004-06-30 00:00:00','   \"Speaking at the launch of the report, Amnesty\'s Secretary General Irene Khan said the abuses in Abu Ghraib jail should have surprised no-one, since it was the logical consequence of the relentless pursuit of global security and the so-called war on terror.\"   urlLink Amnesty International    On a different, but definitely not lighter, note, I detest Christian fundamentalists. Even if you are homophobic, that gives you no right to erect a monument commemorating the day Matthew Shepard died and was \"sent to hell\". That crosses religious lines; death is held, maybe not in high regard, but at least with a great deal of respect. The entire fucking world isn\'t Christian. Live with it.  It\'s the same concept with the war in Iraq. Sure, I don\'t agree with it, but it\'s still a horrible breach of ethics to jeer at the soldiers. You don\'t have to cheer them on, but it\'s more than just impolite to insult them. At the very least, stand respectfully silent, or ever walk away.  Here\'s irony for you. If it just happened that there was a \"god\" or higher power and he (she? it?) screwed over fundamentalists because they violated universal standards of human decency. Sure, they may have adhered to the finer points of the doctrine, but that\'s like putting a bullet in someone\'s head and staying behind a bit to wipe up the blood on the linoleum.  '),(24894,1196970,'30,June,2004','2004-06-30 00:00:00','   And you see? I wrote it in big letters Julie.  Okay, part two.  Yearbooks are out once again. Waxing sentimental...  '),(24895,1196970,'30,June,2004','2004-06-30 00:00:00','   Was sad. Well, being able to trace entire families through the little bags marked \"Im memory of: \". That and a single bag marked sisters. Plural.  But damn I\'m tired, and hurting, and tanned and sunburned in weird patterns in weird areas. I feel like crap. But it was amusing enough.  '),(24896,1196970,'30,June,2004','2004-06-30 00:00:00','   Chalk this up next to \"Take demographic information of train passengers at 30th Street Station\": \"How do different people react to bad news?\"  We wonder.  '),(24897,1196970,'30,June,2004','2004-06-30 00:00:00','   Technology can suck so much at times. Like when you decide to upgrade your web browser and it ends up fucking you over because then it takes you three hours to get your bookmarks, e-mail, and address book up and running again. Fuckers.  '),(24898,1196970,'30,June,2004','2004-06-30 00:00:00','   \"It\'s not that we\'re offended,\" Elizabeth Thomas from Massachusetts told us, wielding a Bible and a poster saying \"Sodomy is sin\".  \"God will be offended. He\'s the one we have to worry about. This is the beginning of chaos.\"  Because the United States was not founded on the principles of religious equality and a separation of church and state but is in fact a \"democratic\" puppet government with not the communists but god himself behind it all?  Back to the picture. Don\'t they look like intelligent people? Wouldn\'t it suck to find yourself being viewed in contempt by millions of people worldwide, to have make an absolute ass out of yourself in the international media, to give the French yet another reason to say \"I told you so; look at the stupid  Americans .\"?  '),(24899,1196970,'30,June,2004','2004-06-30 00:00:00','   ...i\'ll make the effort, love can last forever graceful swans of never topple to the earth tomorrow\'s just an excuse and you can make it last, forever you...  '),(24900,1196970,'30,June,2004','2004-06-30 00:00:00','   \"Latz wanted Duisburg-Nord to contain multiple references that visitors might or might not pick up on. \'\'Everyone who uses the park has a different park,\'\' he says. \'\'Someone is coming to study old blast furnaces, and someone is coming to plant a small garden. Totally different.\'\' Latz\'s notion of landscape architecture has a strong theoretical component. \'\'Landscapes as real objects don\'t exist,\'\' he says. \'\'We select some information from the surroundings and make an idea in our head. Each person has another method to combine the information. There are different information layers, and you may understand only one or two, but somebody else may understand 50.\"\"  '),(24901,1196970,'30,June,2004','2004-06-30 00:00:00','   Click on the title. Hilarious.  '),(24902,1196970,'30,June,2004','2004-06-30 00:00:00','   \"Mr. Cattelan\'s art has provoked hands-on reactions before. When his \'Ninth Hour,\' a wax effigy of Pope John Paul II felled by a meteorite, went on view at a museum in Warsaw in 2001, two men damaged the work by trying to put the figure on its feet.  The new work was sponsored by a private art foundation and produced immediate debate in Italian newspapers and, after its destruction, at the tree itself, where a seemingly permanent crowd of people gathered.  This vandalism echoed a case in Sweden early this year. While attending an opening at a museum in Stockholm in January, Zvi Mazel, Israel\'s ambassador to Sweden, damaged an artwork that he felt glorified Palestinian suicide bombers. The work was created by Dror Feiler, an expatriate Israeli artist living in Sweden, and his Swedish wife, Gunilla Skold Feiler. News accounts said it consisted of a large basin of red fluid on which a small boat floated with a photograph of a smiling woman, Hanadi Jaradat, a suicide bomber who killed 21 people in Haifa last Oct. 4. The ambassador\'s action was commended by Prime Minister Ariel Sharon, who told his cabinet that he had called Mr. Mazel and told him \'that the government stands behind him on this issue.\'  Impulsive destruction that is sanctioned by a nation\'s leaders is different from the work of an inebriated, possibly celebrity-seeking individual like Mr. de Benetetto. But both acts seem motivated by a one-dimensional reaction and a  distorted sense of entitlement that is especially surprising in the case of Mr. Mazel, a professional diplomat . Most surprising is attackers\' simple refusal to entertain paradox, to see art as a coalescence of gray areas, ambiguities and multiple interpretations. Art\'s job is to provoke thought in ways that are difficult to resolve and uncomfortable; it\'s a relatively neutral place to experience the unresolvable issues that dominate real life, to practice a kind of abstract flexibility that might move us toward resolution in real life.\"  '),(24903,1196970,'30,June,2004','2004-06-30 00:00:00','   I\'m not totally sure what this \"URL\" field under \"Title\" does, so this is simply a test post.  '),(24904,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink BBC NEWS | Health | Newlyweds advised to lower hopes :  \"The secret of a long and happy marriage appears to be not to expect too much from it.\"  '),(24905,1196970,'30,June,2004','2004-06-30 00:00:00','   Someone commented. Damn I feel speshal.  Anyways, what do you do if you want to tell someone something, but you\'re afraid of the reaction that you may receive? And not necessarily  that  something, just anything in general. Dammit, we ought to have learned this is Psychology.  '),(24906,1196970,'30,June,2004','2004-06-30 00:00:00','   It took me an hour, but I got the template how I liked it.  There are links right up there in the description for this blog. Which is nice. Try them.  Dammit, I said try them. They took me a while to figure out.  '),(24907,1196970,'30,June,2004','2004-06-30 00:00:00','   Apparently they completely overhauled Blogger without telling me.  And they added comments! Oh my god, I\'m way too fucking excited for something that mundane. But anyways, here we come, comments!  '),(24908,1196970,'30,June,2004','2004-06-30 00:00:00','   If you like-a-me, like I like-a-you And we like-a-both the same, I like-a say, this very day, I like-a-change your name. \'Cause I love-a-you and love-a-you true And if you-a love-a-me. One live as two, two live as one Under the mango tree If you like-a-me, like I like-a-you And we like-a-both the same, I like-a say, this very day, I like-a-change your name. \'Cause I love-a-you and love-a-you true And if you-a love-a-me. One live as two, two live as one Under the mango tree - Tricky (with Ambersunshower)  A touching song.  '),(24909,1196970,'30,June,2004','2004-06-30 00:00:00','   I have a guestbook now. Yes, it\'s crap, but it\'ll do. Plus if gives you all a chance to participate, which is immensely exciting, or so they tell me.  '),(24910,1196970,'30,June,2004','2004-06-30 00:00:00','   The New York Times > National > Using the Courts to Wage a War on Gay Marriage  \"The law school will teach the biblical worldview, but also instruct on practical lawyering,\" Mr. Staver said. \"We will train the next generation of lawyers to protect religious freedoms through the courtrooms.\"  Mr. Staver and Ms. Lindevaldsen said that while they disapproved of homosexuality, they were compassionate, and they distinguished themselves from those who \'spread hate\' against gay people.  \"Adultery and homosexuality are both against God\'s will,\" Mr. Staver said. \"Such sinners must be helped.\"   Once again, let me say this. Fucking Americans.  '),(24911,1196970,'30,June,2004','2004-06-30 00:00:00','   Everyone\'s been on edge a bit, maybe because prom is in such a short time; but whatever the reason, it appears that everyone\'s being ground into a fine powder. So when a friend made the simple comment \"You have good friends.\", it occured to me that I really do. I know sometimes it seems as if I take you (a collective you) for granted, but that is only because I\'m comfortable enough around you to do as much. And truly, I do love you all and I  am  thankful you\'re there - I\'m conscious of the fact that there\'s generally someone I can rely on, confide in, talk to, just a short walk or telephone call away.  And I\'m sorry if my humor leaves anyone bruised or reeling - I don\'t mean it. It\'s just my attempt to enliven the journey that is life, to force some of you to see the less-than-rosy side of things, however irrational it may be. But people are irrational, that is definitely a fact. There are things about someone you may never know, and some things you may find out given time, and only time. So in conclusion, I admit I have some amazing people as friends, and I hope you\'ll all remain as such.  Merci .  '),(24912,1196970,'30,June,2004','2004-06-30 00:00:00','   Baby you\'re all that I want When you\'re lying here in my arms (my arms), I\'m find it hard to believe We\'re in heaven.  Sappy, yes.  '),(24913,1196970,'30,June,2004','2004-06-30 00:00:00','   Robert and SchÃ¶n Passmore took their children to Disney World last fall and left bitterly disappointed. As Christians who reject evolutionary theory, the family scoffed at the park\'s dinosaur attractions, which date the apatosaurus, brachiosaurus and the like to prehistoric times.  \"My kids kept recognizing flaws in the presentation,\" said Mrs. Passmore, of Jackson, Ala. \"You know â€” the whole `millions of years ago dinosaurs ruled the earth\' thing.\"  So this week, the Passmores sought out a lower-profile Florida attraction: Dinosaur Adventure Land, a creationist theme park and museum here that beckons children to \"find out the truth about dinosaurs\" with games that roll science and religion into one big funfest with the message that Genesis, not science, tells the real story of the creation.  - The New York Times  Fucking Americans. There, I said it.  '),(24914,1196970,'30,June,2004','2004-06-30 00:00:00','   A mong all life forms, man alone has aspirations to mutate into something higher. Ancient myths -- be they Greek or Inuit -- swarm with demigods who are smarter, swifter, wiser and a lot better-looking than the dreamers who dreamed them. After Darwin published \'\'Origin of Species,\'\' and primeval skulls with prognathous jaws and small brainpans came to be identified as ancestors of man, the idea took hold that Homo sapiens was just one evolutionary stage toward an even brainier superhuman. In the century since Darwin brought forth his theory, even our closest relatives, the great apes, would be appalled by how the products of our highest civilization can behave with such savagery, and only the congenitally ignorant would argue that man has made any progress at all toward becoming a more intelligent, more compassionate animal.  - The New York Times  Well said.  '),(24915,1196970,'30,June,2004','2004-06-30 00:00:00','   Weird as it is, my parents like Pink Floyd.  '),(24916,1196970,'30,June,2004','2004-06-30 00:00:00','   Taking the train, you see stuff. The stuff no one cares about. Like the backs of buildings. For even if the front facade is awe-inspiring, the back is allowed to have peeling paint and dingy garbage containers in plain sight. Because no one - no one of import, at least - ever sees the back, so why should it matter? It\'s like you\'re traveling beneath the skin of civilization. But not only the stores, the houses also. No one in their right mind would allow their multi-million dollar estate to abut the  train tracks , not without a thick stand of trees at the very least. And then it\'s only with the wind\'s cooperation that you might - might - get a brief glimpse of the people who live in the Main Line.  The Main Line. The Main Line  of the Pennsylvania Railroad , the path taken by the very first proponents of \"white flight\".   Anyways, that aside, I saw a neat advertisement today, sponsored by  urlLink BarnesWatch .  It went a little something like this:  \"If there were other places like the Barnes, dispensing with it would not be tragic. But one minus one is zero.\" - The New Yorker, 12 Feburary 2004  '),(24917,1196970,'30,June,2004','2004-06-30 00:00:00','   urlLink \"An article published last year in The International Journal of Impotence Research...\"    What?!  We have a journal - and an international one at that - devoted to Impotence Research? And see, I can\'t even use the stupid Americans excuse, it\'s international. Men, are terrible things sometimes. There we go, that\'s a good excuse too.  Okay ta ta.  '),(24918,1196970,'30,June,2004','2004-06-30 00:00:00','   \"\'People will come into my office in tears and say they\'ve been against abortion their whole lives,\' he said, \'but they\'ll make an exception for themselves.\'\"  '),(24919,1196970,'30,June,2004','2004-06-30 00:00:00','   So today was the very first day of freedom, and we celebrated by going once again to Philadelphia. Except this time there were  fourteen of us. Thankfully, no drama really occurred, although it could have. And for that, I\'m thankful. In all, it was fun; I enjoyed it immensely, although I doubt I\'ll ever have the patience to organise another outing quite this large.  The sectionalism wasn\'t as bad as it could have been, but even so I may have to start organising things by cliche. But that\'s giving in. And giving in is bad.  Anyways,  I miss Greg, in a way.  I think another something is in order, before I head off to Governor\'s School. Which gives me eight days. Hm.  '),(24920,1196970,'30,June,2004','2004-06-30 00:00:00','   \"\'There also is this American way of suing if you have an opinion that doesn\'t suit, let\'s say, the owner of a work of art. At that moment, science stops. It\'s like Galileo, when he said the world moves around the sun, and he was sentenced for saying so.\'\"  '),(24921,1196970,'30,June,2004','2004-06-30 00:00:00','   \"All eight leaders at the event have been supplied with their own vehicle, painted in their national colours... But although he has used his frequently, Mr Bush\'s attempts to establish his green credentials have been slightly undermined by secret service agents following him everywhere in powerful black SUVs. \"  '),(24922,1196970,'30,June,2004','2004-06-30 00:00:00','   I updated the sidebar.  Note that the artists in the sidebar are just a few of my choice. They kind of go together but are by no means all that I listen to. Hell, there\'s no rock in there, which is a pity. But it has to stay that way for stylistic purposes.  '),(24923,1196970,'30,June,2004','2004-06-30 00:00:00','   Still in that creepy, light-hearted mood since who-knows-when. Must be something in the water, because I doubt both \"the first thing\" and getting an iPod can keep someone bemused for this long. I bet you they\'re lacing the drinking water with something. No, that wouldn\'t affect me, as I drink mostly orange and cranberry juice. Perhaps it\'s something in the air?  Or maybe it  is  the residual effect from those first two things which accounts for a third of the giddiness and the fact that school is almost over which accounts for the rest. It\'s logical and doesn\'t require any conspiracy theories. Good, we\'ll leave it at that.  I\'m in such a good mood I think I\'ll keep writing. On what subject I hardly know, but if I continue writing I\'m sure something interesting will present itself. You know, jump out of a dark corner of my mind, dressed in drag, and scream \"Voila! I present myself!\" at the top of its lungs.  Europe. I would like to go one of these days, do one of those three-month sojourns with a couple friends. That ought to be nice. Too bad airfare is 800$ or so at the cheapest (and that\'s to London, not  pah-ree ) with a Eurail pass at about the same price. Then there\'s shelter and food to worry about, although that isn\'t as hard to do. Who knows, it could be interesting.  Enough musing about plans that may never coalesce. I want something tangible and exciting to happen. Right now. Not literally, but something soon.  I think I hit upon the source of that nagging feeling. I need something exciting to happen. Not necessarily to me, just to happen. But what the hell constitutes exciting? Maybe it\'s just the sudden death of my social life as finals approach. Not that I care, although everyone else seems to. I need to get some totally apathetic people together this weekend and go do something. Something exciting, of course. Not super-exciting, involving fast cars, alcohol, sex, and death, but amusing nonetheless. I\'m sure that at this moment I\'m so interaction-hungry a sedate evening at Starbucks would pass as \"exciting\".  Sounds like a plan.  '),(24924,1196970,'30,June,2004','2004-06-30 00:00:00','   \"Randall Terry, the one-time antiabortion zealot who took on antihomosexuality as his new cause after his own church disowned him for getting divorced.\"  '),(24925,1196970,'30,June,2004','2004-06-30 00:00:00','   And 19 June 2004, this blog turns one. Just thought you should know that.  '),(24926,1196970,'30,June,2004','2004-06-30 00:00:00','   Well I got my iPod, with an iTrip and remote and all that good stuff. I\'m amazed, the radio transmitter really does work. And well. Well, well enough to listen to on the main stereo system without noticeable static. And that\'s good enough for me.  Anyways, I\'m happy. And only from this thing, because that other thing kind of wore off. Sad, I know. But whatever; it\'s not like anything would have come out of that other thing anyways.  '),(24927,1196970,'30,June,2004','2004-06-30 00:00:00','   A friend\'s website.  On a different note, I got a new quote-of-the-whenever:  \"I\'d do --- in a heartbeat. Hell, even if we happened to be standing, on stage, at the Republican National Convention, I\'d do ---. In a heartbeat, you hear?\"  '),(24928,1196970,'30,June,2004','2004-06-30 00:00:00','   That\'s what today was. It started out as shit, but because of a lot of big and small things, it got infinitely better.  This morning sucked, frankly. I was ready to shoot someone, anyone, by the time I stepped off the bus this morning. Not only was it crowded as always, but the only convenient seat was next to our bus\' resident hick. You know, the NASCAR-cap, hunting-clothes, American-flag type. The fucker doesn\'t even pretend to see me as I ask him - nicely, mind you - to move over.  Dammit people, this is the  Main Line : we\'re all supposed to be in nice cars and nice clothes and at the very least above whatever IQ is needed for the basic bodily functions. The plain and unfortunate truth is that money can buy a lot of intangible things. Happiness, yes, to a point. Love, from Russia, I\'m sure. Intelligence, I used to think so. Since when did we have to deal with rednecks?  And then, when I sit down in my seat, I sneeze. One of those gross, icky sneezes. Ugh. But as I\'m getting off the bus, Erin, with her impeccable timing, is also getting off hers. Damn it was a relief to see a face I know and love. That in itself was enough to reset the day-counter back to zero from the negative-whatever it was on the bus. And then there was Vanessa and her photo. We all know what I\'m talking about, no? And if you don\'t, well I\'m not at liberty to explain. All you have to know was,  that  made my day.  Oh yeah, somewhere along the way I got an award for French for something. I guess that makes a small plus to my day-counter.  The other big thing today was Julie\'s remark, which I swore was made only as a gesture - you know, one of those phrases that pops out of your mouth, without your even realizing it, which is designed to do little more than allay the fear and worries and frustrations of the other individual. She said something along the lines of, \"Oh come on, you can do better than that.\". And I would have left it at that, a gesture, a meaningless phrase, if it hadn\'t been reiterated again. Thanks, guys, even if you\'re all simply conspiring to find that exact thing to say at exactly the right time - even if they\'re all naught but gestures - you still managed to make me feel better. Good enough that I\'m not sitting on my bed brooding and miserable. Thanks, again.  Oh, and I got an iPod. Well come on, this  is  the Main Line, and in many ways (or  every  way?) , I  am  the antithesis to NASCAR-hunter-NRA-patriot-boy over there, but I\'m damned proud of that. If I ever end up like that moron, someone shoot me. You can use the shotgun I\'ll have tucked away in my nightstand or whatever and load it with my special NASCAR-branded shells I had been saving for killing that big-ass male deer (I\'m sure there\'s a more proper term for those, something I\'ll know if I ever do become a hick). Shoot at the hippocampus area, preferably. It\'s really fucking hard to miss if you\'re aiming pretty dead-center. Much thanks.  Done venting and gloating and venting and gloating.  '),(24929,1196970,'30,June,2004','2004-06-30 00:00:00','   Professor Florida said that so far the greatest cities of East Asia were falling short of these criteria. \"It is important to have the institutions, but you also need vibrant street culture and an open culture, not only openness toward ethnic diversity, but also diversity in sexual orientation and freedom of expression,\" he said. \"Asia certainly needs a city like this, and Shanghai could be the one. Certainly the city that figures it out first will have some tremendous advantages.\" - The New York Times  A rare post from Governor\'s School, considering the severe lack of free time. Things are going so-so. Call me: (484) 716-0913.  '),(24930,1196970,'30,June,2004','2004-06-30 00:00:00','   Proofread and annotated. [Annotations are in brackets.]   30 June 2004  Culture is an elusive concept. It transcends all - race, location, and numerous other barriers. Itâ€™s pervasive, yet invisible; you make it seem almost like some hidden enemy or danger, lurking everywhere and everpresent, a constant threat. But what is culture?  What is culture? The collective consciousness of a people: the value, traditions, history, and perspective of a community - in the nebulous sense of the word - and the underlying reasoning for the above. The â€œwhyâ€ of culture is important - possibly of more value than merely the descriptors of the culture. [I have no idea how I jumped from defining culture to some crap about â€œwhyâ€. However, why culture exists is probably pretty significant.] It is the foundation that has determined the current state of a culture today and where it will go - or where it can go. [Cultures are not in stasis, they revolve around change, or an aversion to it. Change is never always good, nor ever always bad. But one thing it is always - it happens, and thereâ€™s little you can do to stop it.] One must realise that change is something human beings are highly adverse to and it requires an often-intimate understanding of not only the position taken by an opposing culture but the experiences which have acted in conjunction to mould those positions. Who was it that said something about it being easier to condemn than understand?  Cultures are like diamonds: theyâ€™re valuable, multi-faceted, and seemingly illuminated by an inner radiance. But neither necessarily shines in their own right - they take input and meld it - transform it - into something brilliant. But unlike diamonds, there are innumerable cultures and co-cultures - every one of us is party to countless cultures at any given time. Also, diamonds are forever; cultures are not. Everyone joins and leaves given cultures on a whim or maybe out of some requirement or need.  [A perfect example of a transient culture is Governorâ€™s School. Sure, weâ€™re all here by choice, but we canâ€™t leave the culture once weâ€™ve joined. And in retrospect, the dynamic does change greatly over time. Time here is compressed, so the changes are even more noticeable. From the first, rather awkward, day to now, when weâ€™re all more settled-in, one can see our society being changed - and changing - its members.]  Being placed in a foreign culture is interesting. Particularly is the culture is pronouncedly different than oneâ€™s own. [But isnâ€™t every culture very different in some aspect than oneâ€™s own?] One feels lost at best, maybe even a bit intoxicated by the assault of strange and unusual stimuli. When I was visiting China in the summer of 2000, I was placed in exactly that position. But Iâ€™m Chinese, I said; and, I speak - if nothing more - fluent Chinese! How is it possible one can speak effortlessly to the populous yet remain a total foreigner, that one needs more than just words to relate? But we do; because culture isnâ€™t confined to merely utterances of the mouth, it is expressed in the clothes one wears, the attitude one holds, the forms of interaction that occur which go beyond speech to encompass such aspects of communication as body language. But there I was, forlorn and confined, shunned as a waiguoren - an â€œoutsiderâ€. It sucked, knowing that I was essentially paying my way through life with greenbacks. Traveling to QuÃ©bec was fun too; and I felt more at home there than in even the most Western parts of Beijing or Hong Kong. American money is just as useful in either place.  Yes there were stumblings in my attempts at â€œlearning the cultureâ€, but I thoroughly enjoyed both experiences. A wise friend one said that, â€œEven if you stumble, youâ€™ve still made progress. And who can argue against progress?â€ [And another wise saying that went something like â€œIt is only by being torn out of that which is comfortable to you, being forced to look at your life from a distance, that you see yourself.â€]   2 July 2004  There used to be a class entitled â€œThe Theory of Knowledgeâ€ that was offered at my school. Probably they had the whole ontology-epistemology, subjective-objective-intersubjective thing going. It must have been a wondrous class though, being able to spend an hour every day discussing what is, what out to be, why it is - or ought to be, and the underlying principles of that knowledge.  What does lying count as[, in this realm of thought]? I mean, if one believes strongly enough, [some lies can be as good as the truth.] And if enough people [believe in this collective lie]?   And thus religion was born. Just kiddingâ€¦  But I see how the methods through which knowledge is acquired relates to culture. What is culture? [Defining it a second timeâ€¦] In the very end, itâ€™s knowledge. That knowledge may take on physical form in the traditions of a culture, but it can also remain solely in the sphere of memory. If a culture is suppressed vehemently enough, it often recedes into the thoughts of those who cherish it, to be brought forth again when it is safe.  Language is essential to humanity. Of course it can limit us, but that is better than the alternative. [I donâ€™t think I knew what the alternative was when I was writing this. The alternative would be a dearth of communication, a total inability for knowledge to be passed one. Sure, one can acquire knowledge, but oneâ€™s personal - firsthand - knowledge makes up only a small portion of oneâ€™s â€œexperiencesâ€; it is by passing on knowledge between in across generations that culture is generated.] Knowledge is transferred by way of language. And every language has the ability to enrich the human experience. Every language is the product of a peopleâ€™s experiences and travails; an audible tapestry that flows and morphs with time just as its speakers do. Language is instilled with the essence of a culture, even supposedly nonsense languages carry the weight of humanity, whether in a sense of childhood when speaking Pig Latin or community in using â€œâ€™net lingoâ€.    god, this journal isnâ€™t going anywhere.   3 July 2004    Starting anew, and hopefully this time itâ€™ll be more coherent. And more profound. [Profoundness is a good thing?!]  Words are not universal. That is a given. But even if they donâ€™t all manage to convey the same meaning, [there are universals:]human experiences - at their basest level - are essentially the same. I wonâ€™t use the love example, and thus end up on very dangerous ground, but attraction, curiosity, and repulsion are essentially constants. And - to be a bit daring - â€œloveâ€ is just a strong attraction. Anyone who thinks that love - â€œtrueâ€ love - is irresistible, undying, and indestructible is merely deluding themselves. So Iâ€™m deluding myself.  Imagery is also nearly universal. Even badly translated works of literature originally rich in description are still moving. Pictures, as well, are worth a thousand words. ClichÃ©s may be bad, but theyâ€™re true. A good picture conveys something beyond the realm of speech, something that reaches into the very essence of humanity, to the foundations of what it means to be human. A strong image distills language, association, and emotion into the most basic components, whatever it is that all humans have within themselves, what they use to identify what it means to be alive, and experiencing culture, and human.    I know, Iâ€™m waxing both philosophical and romantic, so Iâ€™ll stop.  Loss is another basic human emotion. Its derivative is sadness. Anger is one too. And the derivations of that: frustration, envy ,and spite.  Back to language. If Sapir and Wharf are correct, it either totally or partially dictates our perception. What about those born deaf? What do they use to dictate how they perceive the world around them, if anything?  I think that some extent the hypothesis is true, but even for those lacking hearing, there are still ways to convey meaning without language. I mean, they still form the same connections with one item and another, but they do not perceive the world in exactly the same way.  They also are deprived of the â€œculturalâ€ value of language. But those are also just associations of partially-related or totally-unrelated object, entities, concepts, emotions with a given thing. â€œThingâ€ is a nice short word. And Iâ€™m out of ideas.   	7 July 2004  So we need more â€œgood stuffâ€. Okay, letâ€™s take a stab at it, shall we? With lots of analysis. Lots and lots and lots. [I know, itâ€™s all crap. Forgive me.]  Analyse oneself. That always provides some form of production return. Letâ€™s seeâ€¦ someone in class said something about â€œhiding something about yourselfâ€, so thatâ€™s the logical starting place.    Iâ€™m gay.  Thatâ€™s a relief; I wrote it - in pen no less - and itâ€™s done. No, not really. Through the whole of human history (homo)sexuality has been a touchy topic. Okay, so now we analyse.  You know there isnâ€™t much to analyse. Iâ€™m frustrated often, but far from being depressed - or suicidal. Governorâ€™s School isnâ€™t particularly helping either: Iâ€™m putting the entirety of a typical (or atypical) teenagerâ€™s personal issues and a pressing need for a good hug coupled with a long - long - face-to-face talk on hold for five weeks. And here isnâ€™t really the best place - nor the most comfortable place - to work through - yes, â€œthroughâ€ is a good preposition here [since you have to work through your problems, not around them] - my insecurities and curiosities and whatever else there is hiding somewhere in my subconscious. I want to go home.  And sure more and more people are finding out [here at Governorâ€™s School] and I havenâ€™t felt any enmity - distilled, diluted, or otherwise - so far but Iâ€™m in need of someplace to vent. I have no idea why itâ€™s in my journal, but thereâ€™s no other place. I apologise; skip this paragraph if you wish. I hate being here. Iâ€™ve never met another gay guy in real life before - well, one where it was known that we both were gay, much less kissed anyone. And Iâ€™m stuck here. Stuck. Trapped. My friends finally know, and this summer was supposed to be one to remember, yet Iâ€™m fucking here, or all places. Not that I despise Governorâ€™s School by any means, just that the â€œopportunity costâ€ of being here is high. [And economic jargon begins its inevitable integration into my speech.] And thereâ€™s almost no way in all of hell Iâ€™d let - or want - something to happen here. I hate this; I need to stop; Iâ€™m lonely and unhappy and homesick. There, I said it. Iâ€™m fucking alone in the whole fucking world, or so it seems, and it fucking sucks.    [I have my blue moments, as Iâ€™m sure everyone does. That was one of them.]    I need to call Julie for a self-esteem boost; write more later. [I love you, Julie.]  Ethnocentrism and stereotyping are both bad. How do you avoid either? â€œHave an open mindâ€ is the clichÃ©d answer. What else? Embrace everything and anything, yet do not make generalisations. Essentially Iâ€™m trying to counteract both at the same time, but itâ€™s inevitable one will tip one way or the other. But why do stereotypes persist? To be blunt, because theyâ€™re true. Generally true. Ethnocentrism? Well, for that thereâ€™s hardly an excuse, even a callous one. So I suppose itâ€™s better to be slightly stereotypical instead of ethnocentric. [I still hold that opinion.]  Wait; thereâ€™s an epistemological question here: why - or may â€œhowâ€ is better - do we - or I - know that both ethnocentrism and stereotyping are bad? Ethnocentrism effectively allows one the right to trample all over someone elseâ€™s culture - and their associated values, traditions, and ideals. And thatâ€™s a bad thing; just trust me, it is. But what about stereotyping? It demeans people, I suppose. Well I more than just â€œsupposeâ€ - suppositions are flippant, but you get the idea. I think that this topic is getting boring, so I motion to advance to the next one.  Diversity. Is a term thrown around excessively by, well, everyone. [What else is it?] Yes, I do think itâ€™s nice, but you canâ€™t that so many things of great import and compress them into a single word. [I guess I was trying to say that the word â€œdiversityâ€ is burdened with too many meanings, variants, facets.] Yes having a diverse group of acquaintances - not necessarily friends - can open oneâ€™s eyes, but that doesnâ€™t work if theyâ€™re all hopelessly ignorant. But, having a more or less homogeneous group of acquaintances who are all open-minded and well-read can still confer part of the benefits reaped by living in the midst of a diverse society. Not all, mind you, but plenty. [Which is why I adore my friends. Not only are they a diverse bunch, theyâ€™re intelligent and supremely fun to be around. I love you guys.]  I think this is by far the longest entry. [My entries are not separated, unfortunately. The date divisions are just that - date divisions. In fact, one entry was written over midnight and so crosses a date division.] And my letter â€œaâ€™sâ€ are looking more and more like my â€œoâ€™sâ€. Thatâ€™s not a good thing. Iâ€™ll write more later.   8 July 2004    I feel terrible writing during a class, but if he has to repeat that concept another time Iâ€™m going to shoot myself.  So, I need lots of analysis in this journal. And some way to lengthen this by another four pages, at least. [Handwritten, on college ruled paper.]  Weâ€™ve already gone over the basic emotions, but Iâ€™m in a mood to talk more. Attraction, repulsion, and loss are good ones. Or rather, love, hate, and loss. Or rather, hate, because Iâ€™m not in a particularly warm and fuzzy mood.  Why do humans â€œhateâ€ anything to begin with? Especially between cultures; conflicts often appear when cultures meet and have occurred as far back as the first civilisations. And the reasons? Misunderstanding, or even a simply refusal to understand? And with its advent, religion. Dun dun dunâ€¦ Sheer ignorance or all-consuming righteousness - generally on behalf of an â€œadvancedâ€ Western civilization and its linked religion - usually leads to the collapse, of even the annihilation, of the â€œother sideâ€. Cases in point - Africa and the Americas. Just because one feels that oneself is â€œbetterâ€ than someone else does not mean that one has the right to impose oneâ€™s IdealsValuesReligionPerspective on someone else. And once again we hit the wall of stereotyping and ethnocentrism. [And what a common wall it is. Can we ever get beyond the wall? I doubt it.]    My thoughts were interrupted and so now I have no idea what I was going for. Iâ€™m sorry.  Iâ€™m sorry. Sorry for fucking everything. Every fucking atom in the cosmos. Because itâ€™s all my fucking fault, that everyone hates me. People suck. They suck, plain and simple. My gaydar sucks, which only makes things infinitely more frustrating.    Donâ€™t psychoanalyse, just read. Iâ€™m not suicidal.    And once again I forget the end to which I was heading. Fuck you all.    [Another blue moment. Donâ€™t know where the hell that one came from, though.]   9 July 2004  Iâ€™m just going to end up writing in this every day. Well I need it: itâ€™s a nice place to vent. Oh, and itâ€™s rather nice to put down all my thoughts on paper.  I think Iâ€™m stuck in a rut. And a depressing rut at that. I spent well over an hour on a conference call with five friends. The content doesnâ€™t matter, but it can be seen as a case of â€œintercultural communicationâ€.  [So the five other people arenâ€™t really friends. In fact, some probably hate each other. But theyâ€™re willing to sit there and help me hash through my insecurities and problems. And I love them for it. There was a lot more written here but it was nonsensical and dragged on forever.] Do you think different cliques in a high school each count as their own culture? Probably, especially if each group is well-insulated and things of itself as better than everyone else. Ethnocentrism? Or maybe just blinding ignorance and self-interest. But whatever the case, it happens; not to the magnitude of this hypothetical example [What hypothetical example? I think I lost my train of thought and went rambling again.], but it does happen. In some way, itâ€™s almost like taking the world and shrinking it to a small population or a thousand or so. Not as diverse, and often much more liberal [Actually, a couple weeks of Governorâ€™s School has proven to me there are young Bush supporters. My heart goes out to them, the poor bastards never had a chance.], than the world as a whole; but, in some creepy way, the social interactions align in pretty interesting patterns. And the student council, being - surprisingly - not just a popularity contest, is the body that arbitrates and facilitates interaction. [I think somewhere along the way it morphed from a â€œhypotheticalâ€ example to my school.]    I really ought to go back to listening to â€œinternational political economyâ€   11 July 2004    There is not way in hell this is going to be fifteen pages. And typed. I apologise profusely. [Oh, the irony.]  So I went shopping - for clothes - today. Clothes are tied in to culture, arenâ€™t they? [And who cares if theyâ€™re not? I like clothes.] Every culture or subculture - co-culture for all the politically correct people out there - has added its own flair or input on the basic components of apparel. Especially in wealthy developed countries (i.e. the United States), oneâ€™s manner of dress does so much to project oneâ€™s identity. Or not necessarily a true identity, as one can also use clothes to mask oneâ€™s own identity. [Clothes often reinforce stereotypes.] Sure, taste is entirely a subjective thing, but within a given group it is still somewhat codified to ensure some degree of homogeneity. Because everyone wishes to belong to one or many cultures, and clothes are an easy way of doing so.  Clothes convey social status with a mere glance. They can also convey oneâ€™s interests or beliefs, although less frequently. Adamant Bush supporters donâ€™t wear t-shirts likening Bush to a monkey just as the rich and famous donâ€™t wear just â€œordinaryâ€ clothes. Non-verbal communication at its finest.  â€œThe meeting of cultures creates disparities in societal normsâ€. [Somebody said in class.] Well of course, seeing as how cultures are a composite of the values and experiences of a given group of people. And by â€œgiven groupâ€, I assume the members of the group are relatively homogeneous, not just a random clumping of people utterly unlike each other in any way. So of course their societal norms - and extension of ethics - are not going to mesh perfectly.  â€œWithout the correct feedback, it is human to doubt oneself.â€ Is it ever. Iâ€™m close to being one of the most insecure people ever, almost completely unfunctional without at least a friend beside me at all times. Pathetic, yes, but I seem to have trouble relating actual, received feedback with the outcomes Iâ€™ve already mapped in my head. And yes, I do map outcomes in my head. [Iâ€™m not sure where this came from either. Writing longhand causes weird thought patterns. Or not necessarily weird, different. And being different isnâ€™t a bad thing, they say.]  Feedback is so crucial to a successful conversation in any medium. Beyond the fact that feedback is essentially what makes up a conversation - a verbal response, a typed response, etc. - other, more subtle, forms of feedback are also important. For example, sometimes you really donâ€™t want to talk to a certain person. So you send subtle signs - non-verbal signs - to tell them youâ€™re bored shitless and would appreciate it greatly if they would kindly hurl themselves off a cliff. Generally one can make the conversation as awkward as possible by trying to remain as silent as possible, both in person and on the Internet. I feel bad implying I ignore people on purpose, but whatever.    Okay, I really donâ€™t want to write any more.   12 July 2004  [This last entry was added for no reason whatsoever. No; in retrospect it was probably added to increase the journalâ€™s length. Feel free to ignore it.]  Communication without feedback. Feedback without communication. Neither is possible since communication is essentially the interplay of feedback in both the verbal and non-verbal components. That sentence made no sense. I was thinking of chiaroscuro, but that has nothing to do with intercultural communications. And how can you have feedback occur without the occurrence of communication? Not really possible since feedback almost always results in communication.  Gestures. You know, those really pointless actions demanded by Society. â€œHow dâ€™you do?â€ and all that bullshit. Itâ€™s all been pounded into us since the age of zero so that there are entire conversations without an iota of feeling. But that can still count as communication, given a definition of communication where the feedback does not need any form of vested interest, just an exchange of symbols. And the meaning behind the symbols isâ€¦ being nice, genial, amiable.  What crap.    Okay, ta ta.   15 July 2004    So Iâ€™m writing again. I donâ€™t know why, or about what, but Iâ€™m writing.  I had such a good topic, but I forgot it. Dammit. So Iâ€™ll write about music. Music conveys feelings. [What is that? The seventeenth time Iâ€™ve said that?] Especially wistfulness. Iâ€™ve been rather passive lately. The â€œshopping highâ€ wore off, so Iâ€™m back to my lonely, carping self. And Iâ€™m listening to sappy music. Or maybe because Iâ€™ve been listening to heart-wrenching music, Iâ€™m once again back to sad-Nan.    â€œThere are lots of fish in the sea,â€ Julie says. Lots and lots of fish. Write more later. I think I offended a friend tonight. [It must have been near midnight when I wrote this, because this entry spans a date marker.] A really close friend. And really badly. I feel like shit, like I need someone - or something - to talk to. I hate this, why does everyone take everything I say with such gravity? Or rather, some people take me at my word while others never think Iâ€™m being serious. [Neither is a good thing, dammit.]  Iâ€™m really fucking sorry; what else can I say? And you. Youâ€™re her boyfriend - her best friend. How can you just blow it off? How dare you just blow it off?   16 July 2004    How DARE you show no emotion whatsoever while Iâ€™m sitting here, in anguish, losing fucking sleep over it? WHAT THE HELL?!    â€œI; I have become, comfortably numbâ€, the music says. No, I havenâ€™t. I made a careless slip and now Kaitlin hates me.  This isnâ€™t intercultural communication, itâ€™s basic communication. The human experience in all its wonder and evil, happiness and heart-wrenching misery; all melded together in one pitiful and insignificant life. One. Such a pathetic number in and of itself.  Does life ever become so unbearable that you - ? I suppose it can. Maybe not indefinitely, but it can. And for that one infitesmal moment, it can seem as if the tradeoff is worth it. And thatâ€™s what scares me. As if? What if it is worth it?     Shine on you crazy diamond.    What is a life worth? I donâ€™t; maybe itâ€™s all relative. [I believe I meant to say â€œI donâ€™t know.â€.] How can such a simple question not have a simple answer? [Lots of simple questions have complex answers, or no answers at all.] How can this most basic - as basic as the sun, moon, stars - of questions still require a relativist answer? Is nothing universal in its being;  is nothing sacred?  [Iâ€™m not completely sure why I associated universals with sacredness.]     And you run and you run to catch up with the sun       but itâ€™s sinking; racing around to come up behind you again.       The sun is the same in a relative way but youâ€™re older:       shorter of breath and one day closer to death.    	     Adieu.  I forgot to cap a brand new bottle of white-out and I put it in my bookbag. No relation whatsoever to anything, it just sucks. A lot. And this gross new formulation doesnâ€™t come off very easily. [Or at all, seeing as how my notebooks still remain splattered with what appears to be white paint.] And it feels icky on my fingers. Why canâ€™t they just stick with the old brush [as opposed to a new, foam one] and the old, crappy formula? Change can too be bad. Such as this case. The brush just sucks and itâ€™s supposed to be correction fluid, not permanent paint. Oh, and it still smells terrible. So really, there wasnâ€™t improvement. In fact, the white-out company appears to be retrogressing.  Things you believe you canâ€™t have are as good as things you canâ€™t have. And that sucks, [but you canâ€™t change your mind to believe you can have something if youâ€™ve already decided itâ€™s unattainable, no matter how hard you try.] The mind is a truly wondrous thing [in its ability to fuck you over]; itâ€™s almost as good as real life. Maybe better.  I think the one and only purpose in life is to find satisfaction. Or, in even more blunt terms, happiness. In some way, any way. I need some happiness in my life. What else can someone want? I suppose everyone obtains satisfaction in a different way: wealth, power, spite, altruism. But in the end, it is the same goals. Humans donâ€™t share very much in common with each other, but this they do share. The same ones at least. [No idea where this was headed or how it got here, sorry.]   18 July 2004    I donâ€™t know what Iâ€™m supposed to write, but Iâ€™ll write something anyways.  Mobile communication. The next big thing. Not only cell phones, or even the super cell phones that can check e-mail and send video or whatnot. But internet access from alongside every major highway and in every major city. That is the next big thing. Even now, with wireless hotspots as the newest â€œinâ€ thing, people are beginning to like - and become accustomed to - the technology.  Ohh, big point. When someone becomes accustomed to something, they begin to expect it. And thatâ€™s not a good thing. Westerners are accustomed to too many things. And so they expect the world to be handed to them on a silver platter.    Dinner.    Back from dinner and ICONS. Oh, I was having so much fun, my head could explode. Positively explode.  We - as Americans - feel like weâ€™re entitled to too much. Like itâ€™s our fucking birthright. [Hell, the land we live on isnâ€™t even our birthright.] But why? Is society propagating our lifestyle? Probably. Can we change it? I donâ€™t know, can we? If we all gave up something - Hummer, golf courses, rare wood furniture - wouldnâ€™t we all be better off? The first â€œweâ€ is Americans; the second, the world. To satisfy the United Nationsâ€™ daily water requirement for all six-something billion people on this planet, it would take 2.5 billion gallons of acceptably clean water a day. The worldsâ€™ golf courses use 2.8 billion gallons of water a day. What an epitome of the bourgeois spirit - an activity that requires an enormous expenditure of water and land seemingly to make sure that the planet would be deprived of limited resources.  How can you get across a foreign concept to someone totally and utterly opposed to your viewpoint - largely out of breeding? Indoctrination is a powerful tool. Is it possible to defeat it, however? Not from within, I donâ€™t think. But there are no outside forces powerful enough. So is the answer no? Is there not some method one can use to get oneâ€™s point across? Even if the other party provides no feedback whatsoever, is it possible to transcend bias and prejudice - and all the collected experiences that came before - and provide sufficient reasoning for a good argument?  Itâ€™d take more than words. Words donâ€™t have enough power. Words cannot, in themselves, convey every nuance of human emotion. The symbol component may have once held power, but the meaning has been too degraded, too diluted, too limited to hold much power anymore.    And for that Iâ€™m sorry.   21 July 2004  In global citizenship today we were talking about weblogs. I have one. This is going on it. Because at first I thought about continuing these journals beyond Governorâ€™s School but then I realised there are more effective places to put down my thoughts. Digitsed information lasts longer than paper. Especially when something is written in water-soluble pen. [In fact, Iâ€™m destroying the original copied of my journals after Iâ€™m done with Governorâ€™s School; this will be all that remains.]  The internet is an amazing thing. I keep a public record of my thoughts on it. Itâ€™s the most sophisticated method of communication every devised by humans. It evolves and improves constantly; it hold no allegiance to any entity; it can create and destroy lives - both real and digital; it goes beyond what anyone in Renaissance Florence could have possibly imagined; or even people at the end of World War II. It now defines a significant part of the developed world - or at least the culture and methods of communication in the developed world - and it seems likely that the internet will only grow in power and efficiency.  Instant messaging is an interesting method of communication. But not a novel idea, really. Itâ€™s like walkie talkies in words. Or really fast telegrams. Sure, itâ€™s more cost-efficient - and more efficient period - but itâ€™s not an enormous jump between sending e-mails to sending really short e-mails in real-time.  It gives you time to think. Much like all writing. You generally tread more carefully since non-verbal communication is absent; things can be easily misconstrued or taken literally when one doesnâ€™t mean it. Instant messaging can also get one into a lot of bad situations - itâ€™s too damn easy to lie or dissemble, or to be so neurotic as to believe the other party is lying. You come to realise the inflection of oneâ€™s voice makes up an enormous amount of communication.  New topic. What makes sex such an irreconcilable item between men and women? Beyond the biological factors, I mean. [Is there much beyond those, though?] The cultural factors. That men are often rewarded or commended for asserting their sexuality in a way that their dominance - supposed dominance, I should say - over women is obvious. Why does society persist in maintaining such a double-standard? Because back in â€œthe dayâ€, men maintained their superior position through brute strength - something that mattered quite a bit in the Stone age - and somehow have kept that position over the past however-many millennia?  Okay, that topic got boring fast. â€œWhat if you use derogatory labels but you donâ€™t mean it?â€ In other words, the word â€œgayâ€. Beyond reinforcing the homophobia of society, it deeply ingrains a terrible stigma in the word, often without the user realising it. In informal polls, almost everyone that has firmly declared themselves to be open and accepting persists in using â€œgayâ€ or â€œfaggotâ€ - and its myriad variations - as curses or insults. And then I come out to my friends and almost all use of the world ceases. Amusing, no? And now whenever someone does slip inadvertently, thereâ€™s an almost palpable cringe and awkward silence. Thatâ€™s amusing.    The end.   25 July 2004  These are getting increasingly sporadic; I feel bad. But individual entries are getting longer. Hopefully it will work out in the end.  Everything works out in the end; thatâ€™s part of my theory on life. Ever the optimist, stuck in my world fabricated from wishful thoughts. People say I trust in other too much. I think I trust in them enough - such that we are all conferred the status of rational beings, and rational beings are good in the end.  Iâ€™m relaxed and content. Itâ€™s really nice. We had to â€œcheck inâ€ again tonight, but Iâ€™m still in a mood to ponder. The past couple days have been pretty damned amazing, especially since I met someone whoâ€™d make an absolutely amazing friend. But that person lives here [Pittsburgh] and I live five hours away. So itâ€™s true you meet awesome people here, but you may never have the chance to just sit in a coffeehouse and relax with them, which sucks. Instant messaging and phone calls really arenâ€™t good enough; it takes someoneâ€™s physical presence sometimes. You know, those people who are entrancing in person but almost blasÃ© otherwise. Thatâ€™s a major low point right now. But you just have to push it off center stage and seat it next to the box labeled â€œschool-assigned summer work for the 2004-2005 school yearâ€.  Society dictates a lot of things. Is that a good or bad thing? If we were all free-thinking - and I suppose ideal - beings, then an utterly unstructured society might not be so bad. There would be no need for laws if we could all make the rational choice. But we donâ€™t, so Society is always there to ensure we make the â€œrightâ€ choice.  There are decisions we make for ourselves - or that we believe we make for ourselves - and those so deeply ingrained in us that we see only one choice that would be considered â€œlogicalâ€ to take. In fact, some decisions are so impossible to escape that we donâ€™t realise weâ€™re making a choice. Theyâ€™re transparent and pervasive. Like culture, in fact.  We constantly make choices weâ€™re only slightly aware of. Or ones weâ€™re completely oblivious to. Many many times in a second. Such as why Iâ€™m not stopping this journal and heading off to mingle or why I should jump on a coffee table and start doing the Macarena. Because thereâ€™s no one to mingle with at the moment and because itâ€™d be ridiculous and embarrassing. Why? Because Society told me it would and has taught every to look upon anyone who commits - yes, commits, like a crime - such an act as an imbecile and to point and laugh accordingly. Moreover, Society has made it perfectly clear that being pointed and laughed at is a bag thing.   27 July 2004.    Last journal entry for Governorâ€™s School.  '),(24931,4142553,'05,August,2004','2004-08-05 00:00:00','   Procrastinating: weâ€™ve all been guilty of it, whether we are student, worker or slacker, we all like a good distraction, to get away from whatever it is that we (or someone else) have been nagging us to do for the past hour, month, millenium. Why finish that assignment when you can spend a good two hours trying to beat your fastest score on Minesweeper? Why do the dishes when there is still a plate left, and you know you can go without clean dishes for yet another day? Iâ€™m sure you know the answer, youâ€™re just too f***ing lazy to do anything about it, let alone think about it, right?    These have been my top 5 vices, now that I have succumbed to a bad case of the Mr. Pâ€™s:     1. Finishing my assignment  All of my notes have been written up, the flatmates are out and there is nothing but the sweet hum of the central heating unit. Iâ€™m about to write 1000 words, but do you think that I do so? Like hell I do. Why do you think I have started a blog? Doing this blog adds at least 5 frequent flyer points towards Waste Of Time and 10 points toward Nothing Achieved. As soon as I write 100 words, it will be rewarded with an hourâ€™s worth of coach potato time watching Dr. Phil yell at some poor sod to take control of his life. Way to motivate, but the only thing I want to take control of are the TV and CD player. Iâ€™m sure that assignment will eventually write itself .     2. Unpacking my suitcase  It has been two months since I arrived back from overseas, but the items in my suitcase are still intact, unfolded, untouched and un-put-away. The stuff sits there like children that have been forgotten to be picked up from school, but thereâ€™s no way in hell that I want to move it for the sake of being neat and organised. God no. Why should I unpack when I can play mental Cluedo of where-did-I-leave-my-cheap-intimates, and where-the-hell-are my-socks? Thatâ€™s much more fun.     3. Buying the household toilet paper  If someone else can beat me to it, why cloud my busy schedule at home by going out of my way to get the Sorbents? I can honestly say that this one is a bit of a non-de-plume because I have finally motivated myself to buy a four pack, but not without loads of affirmative self-talk and fear of not having something to wipe the â€˜excessâ€™. If I were any more lazy or procrastinatative, I would have resorted to using the pages of the Melbourne Weekly Times as a subsititute, which doesnâ€™t say a lot about the publication, does it?     4. Finding a part-time job  Is this one really necessary to explain? Gettting a procrastinator to find a job is like getting Lillian Frank to feed the homeless on Flinders Street. Not exactly a symbiotic and harmonious relationship to say the least. But being poor is nothing to boast about when you are living off staples of NQR spag and sauce and day old bread, so unless Centrelink are going to increase my allowance an extra $50 a fortnight, Iâ€™m â€˜workingâ€™ on this one, slowly nonetheless.     5. Excercising  Excercising is like an eternal enigma in the world of the procrastinator: they know it exists, but put off the so-called benefits of being fit and healthy for the convenience of being fat and lazy. Since coming back to Melbourne, the only â€˜exerciseâ€™ that I have done is pressing the remote to watch Ryan piss the girls off in the Big Brother house. By the way, what is â€˜jazz-erciseâ€™? I hope you can eat it.    Here are the top 5 devices that a procrastinator cannot do without. 1. Solitaire and Minesweeper 2. Television, and all the quality viewing that it comes with such as Jerry Springerâ€™s â€˜Sisters who sleep with transsexual Nazisâ€™ 3. Chatrooms on the Internet 4. Nose picking (I have heard) 5. Masturbation (natureâ€™s gift to the procrastinator)   '),(24932,3816360,'31,July,2004','2004-07-31 00:00:00','   Learning HTML is hard as hell.  However, my ego demands that my blog have at least a few cool things.  I could bug Josh or John to program everything I wanna do for me, but my ego will also not allow me to want to know something and not master it.  So I will learn HTML at least the shit I need to know to do what I want.  It\'s hard it doesn\'t make any sense and I\'d wager that this could be made a lot simpler.  I\'d also wager that the simplifier could stand to make a lot of loot. '),(24933,3816360,'28,July,2004','2004-07-28 00:00:00','   My upstairs neighbor is a little off.&nbsp; I have no idea what the problem is, but there is definitely a problem.&nbsp; So all day long she follows my next door neighbor around who treats her like ten year olds treat their younger siblings, bossy, but loving.&nbsp; I recently found myself engaged by neighbor in a \'kids today\' ( yes maybe I am getting old), during which I informed that I think it\'s&nbsp;mostly the parents fault and that often the kids do well in spite of the parents not because of them.&nbsp;&nbsp; She asked me why I felt this way and I told her that when I was young I had an awful childhood and what surprised me more than the incredible disfunction in my family was that no one did anything about it.&nbsp; Child protective services became involved on a few occasions, and they never really investigated.&nbsp; My friends parents got impassioned, and when they spoke to my family and realized what they were up against they backed down.&nbsp; She asked me how do I balance the anger with parents who work against me in mentoring the kids.&nbsp; She said she\'s upset with the upstairs neighbor because she always has a lot of men around and when they around she yells and curses at her son, sends him to stay in his room for the duration of the visit, and gives him baby portions of dinner so the boyfriend could eat his fill.&nbsp; Today my next door neighbor brought said kid to my door in tears.&nbsp; He articulated in beyond ten year old capacity, that he informed her that he felt alone and&nbsp;unloved because she put her many boyfriends before him , but she just yelled and cursed and called her boyfriend.&nbsp;He said&nbsp;hated living there, but as his grandmother was in the hospital and he had no other relatives he didn\'t know what to do.&nbsp; I held him told him it wasn\'t his fault, he didn\'t deserve that, and some parents just don\'t know how much what they\'re doing hurts, and that when I feel bad I smack Joey around and playfully hit him.&nbsp; Joey offered to be the boy\'s punching bag too, but he declined.&nbsp; I asked him if he liked to read he said he did.&nbsp; I was caught somewhere between feeling like I was going to be sick and wanting to punch a wall because I am just as powerless to save him now as I was to save myself then.&nbsp; He told me he liked the first Harry Potter I told him I had them all and he could borrow them.&nbsp; His face actually kind of lit up.&nbsp; When he left I felt a little better, maybe there was a little good from this.&nbsp; I\'ll have a new friend to talk about the books with and the next time somebody starts in on how Harry Potter is evil and they\'d never let their kid read a book about witchcraft it won\'t even bother me.&nbsp; I\'ll think of it as me not having to lend out my quickly wearing copy so it\'ll be there for the kid who needs to know that somewhere across the ocean there\'s a kid who\'s family is just as bad or worse, and no one is saving him either, but somehow he\'s making it and maybe they can too. '),(24934,3816360,'26,July,2004','2004-07-26 00:00:00','   I do a lot of ranting.&nbsp; I know it and I\'m not ashamed of it, but every now and again I try to stop and think about the things that I am pleased about.&nbsp; I thought I\'d mention it here.&nbsp; I went to the beach on Saturday with Ana and Joey.&nbsp; Had a great time.&nbsp; It was in the 70\'s but Joey felt like swimming anyway, so he splashed and discovered the remains of horseshoe crabs and seashells for hours until his lips turned purple and I forced him out of the water and into dry towels, at which point he tells me that the water was dirty and I never should have let him go in.&nbsp; As soon as he was dry he asked to go back in again.&nbsp; Some poor father waited too late to leave the beach and was left with a screaming monster of a daughter who hated everything he did and was illogical as sleepy kids are, she wanted to go home but didn\'t want to leave, wanted to be dry, but didn\'t want to take her wet things off.&nbsp; He got really frustrated and kept grabbing her elbow and whisper yelling \'you better stop it\'.&nbsp;&nbsp; Ana stared at this guy for ten minutes, forget if looks could kill, it looked like if she\'d had a weapon she would\'ve.&nbsp; They both sang loudly all the way home.&nbsp; I\'m having another email affair (for lack of a better word),&nbsp; I\'ve known Erica from my church for a while, but when she goes away we decide to develop a relationship, this is how it went with Joanna.&nbsp; It\'s really weird cause I\'m more of a talker, but I like it, it\'s personal, there are no interruptions, no sideways glances, and you say what you think and aren\'t so busy trying to gauge reactions.&nbsp; She has a great blog&nbsp;&nbsp; urlLink http://www.ilovewords.blogspot.com/ &nbsp;check it out. Anthony\'s been getting up with the kids when he\'s home to let me sleep in, he even closes the door and says go back to sleep.&nbsp; Lolade sent me her old camera, it\'s the best free camera ever.&nbsp; Everyone loves Chicago, Erica\'s there, Deron just went said the trip was life changing, Tameaka went and said she can\'t wait to go back.&nbsp; Josh and James called me drunk last night.&nbsp; I find myself on the phone with someone when James is in the background not too infrequently, he never makes a point to talk to me and neither do I to him.&nbsp; Our love is a more in person thing, but every now and again after a few beers or some shots James or someone he\'s with calls and he makes a point of talking to me.&nbsp; It brightened up my night.&nbsp; He told me Josh couldn\'t let me on their blog cause it\'s only for boys voices.&nbsp;  urlLink http://www.sunsgonnarise.com &nbsp; Penis power is okay by me and it turned out good, though I don\'t really get why they have four writers when it\'s only really Josh that posts. '),(24935,3816360,'16,July,2004','2004-07-16 00:00:00','   In the previous post I mentioned a parallel in the wording of Genesis 3:16 and Genesis 4:7.  I am now including, what I thought to be an interesting expounder on the subject.  Let\'s look at Genesis 3:16. Adam and Eve have both sinned against God. They have distrusted his goodness and turned away from him to depend on their own wisdom for how to be happy. So they rejected His word and they ate the fruit of the tree of the knowledge of good and evil. God calls them to account and now describes to them what the curse will be on human life because of sin. In Genesis 3:16 God says to the woman, \"I will greatly multiply your pain in childbearing; in pain you shall bring forth children, and your desire shall be for your husband, and he shall rule over you.\"   This is a description of the curse. It is a description of misery, not a model for marriage. This is the way it\'s going to be in history where sin has the upper hand. But what is really being said here? What is the nature of this ruined relationship after sin?   The key comes from recognizing the connection between the last words of this verse (3:16b) and the last words of Genesis 4:7. Here God is warning Cain about his resentment and anger against Abel. God tells him that sin is about to get the upper hand in his life. Notice at the end of the verse 7: \"Sin is crouching at the door; its desire is for you, but you must master it (literally: you shall rule over it).\"   The parallel here between 3:16 and 4:7 is amazingly close. The words are virtually the same in Hebrew, but you can see this in the English as well. In 3:16 God says to the woman, \"Your desire is for your husband, and he shall rule over you.\" In 4:7 God says to Cain, \"Sin\'s desire is for you, and you shall rule over it.\"   Now the reason this is important to see is that it shows us more clearly what is meant by \"desire.\" When 4:7 says that sin is crouching at the door of Cain\'s heart (like a lion, Genesis 49:9) and that it\'s desire is for him, it means that sin wants to overpower him. It wants to defeat him and subdue him and make him the slave of sin.   Now when we go back to 3:16 we should probably see the same meaning in the sinful desire of woman. When it says, \"Your desire shall be for your husband,\" it means that when sin has the upper hand in woman she will desire to overpower or subdue or exploit man. And when sin has the upper hand in man he will respond in like manner and with his strength subdue her, or rule over her.   So what is really described in the curse of 3:16 is the ugly conflict between the male and female that has marked so much of human history. Maleness as God created it has been depraved and corrupted by sin. Femaleness as God created it has been depraved and corrupted by sin. The essence of sin is self-reliance and self-exaltation. First in rebellion against God, and then in exploitation of each other.   So the essence of corrupted maleness is the self-aggrandizing effort to subdue and control and exploit women for its own private desires. And the essence of corrupted femaleness is the self-aggrandizing effort to subdue and control and exploit men for its own private desires. And the difference is found mainly in the different weaknesses that we can exploit in one another.   As a rule men have more brute strength than women and so they can rape and abuse and threaten and sit around and snap their finger. It\'s fashionable to say those sorts of things today. But it\'s just as true that women are sinners. We are in God\'s image male and female; and we are depraved, male and female. Women may not have as much brute strength as men but she knows ways to subdue him. She can very often run circles around him with her words and where her words fail she knows the weakness of his lust.   If you have any doubts about the power of sinful woman to control sinful man just reflect for a moment on the number one marketing force in the world â€” the female body. She can sell anything because she knows the universal weakness of man and how to control him with it. The exploitation of women by sinful men is conspicuous because it is often harsh and violent. But a moment\'s reflection will show you that the exploitation of men by sinful women is just as pervasive in our society. The difference is that our sinful society sanctions the one perversity and not the other. (There are societies that do just the opposite.)  May 21, 1989 Bethlehem Baptist Church John Piper, Pastor  Don\'t that just inspire thought and passionate conversation. '),(24936,3816360,'14,July,2004','2004-07-14 00:00:00','   Make sure you read Josh\'s comment to Mary Magdalene.  I gotta say I had to read it like five times to understand what he was saying because I was just so blown away that Josh was quoting the bible.  I felt it necessary to respond as I felt there was a misunderstanding and I just love to talk about all things relating to God.  I must warn the readers that I have herein written things that are different and some contradictory to Josh\'s comments.  One should not attempt to do so without fully realizing that Josh loves to argue, moreso than I, if you can believe it, he is also damn good (get this he\'s been writing and doing mock trials for years) at it and usually wins.  In light of that fact consider me reckless for posting this or consider me wanting to be verbally spanked, but since I do so love the sport I often find myself voicing a differing or opposing viewpoint.  I would like to clarify that I don\'t now nor did I at the time of the posting of the Mary Magdalene piece believe the idea of Mary Magdalene being Jesus\' wife would be a strike against women or her credibility.  To make clear my belief on the matter is that you cannot prove either she was or was not as there just isn\'t much written about her.  It is greatly controversial, as it is a commonly held belief that Jesus had no need to gratify himself sexually so he would not have married.  In my opinion it is neither here nor there, because his marriage would not have been a sin and according to Hebrews 4:15 (For we do not have a high priest who is unable to sympathize with our weaknesses, but we have one who was tempted in every way, just as we are yet was without sin.) he was made in the flesh so that he could understand our weaknesses.  I wouldn\'t say that the desire for marital sex is a weakness, but even if you assume it to be so, it makes sense for Jesus to have experienced it.  On the other hand, he was a very busy traveling man who may not have had time for a wife.  I was trying to make a point of the irony in saying that the assumptions made on the means of her wealth are presuming and not based on much evidence and then presuming that she was Jesus\' wife.  Mary is assumed to be a prostitute primarily because it is believed that she was unmarried and Magdala was a city destroyed for its fornication and because seven demons were cast out of her.  She is assumed to be the wife of Jesus primarily because of the gospel of Philip which says:  And the companion of the [...] Mary Magdalene. [...] loved her more than all the disciples, and used to kiss her often on her [...]. The rest of the disciples [...]. They said to him \"Why do you love her more than all of us?\" The Savior answered and said to them, \"Why do I not love you like her? When a blind man and one who sees are both together in darkness, they are no different from one another. When the light comes, then he who sees will see the light, and he who is blind will remain in darkness.\"  [...] are holes in the text made by ants.  The kiss could be the kind a husband and wife shares or the kind the kind shared in the churches (Paul says greet each other with a holy kiss in Romans and Corinthians 1 and 2. As for Jesus loving her more than the rest, I think the blindness and the light could be a reference to Mary proving that she truly believed in all Jesus said and maybe how she proved that by being witness to his resurrection and telling the other disciples (according to the gospel of Mary they didn\'t believe her at first.). I thought it defeatist to such feminist scholars to imply that the assumed prostitution was wrong and created by masogynists as an assasination on her character, and in an effort to combat that theory and give her character credibility and assumption must be made that she was the wife of Jesus.  I see them as equally presuming and thusly counterproductive to an argument against such presumptions.  I also think that it would be underimining her role to stress her importance lying in her marriage to Jesus rather than what she did and was. I would rather stand out for my own beliefs and words and works than to stand out for being Anthony\'s wife.  On the other hand Ant ain\'t Jesus. I also thought it better to argue what is more clear, that she was really important, possibly most important of all people, to Jesus, as opposed to making the leap to marriage, as again I\'ll say you could argue either point with no end in sight.  As to the piece on the tenses in the original Hebrew, Hebrew tense is largely unlike English.  In Hebrew you have the perfect (completed) and the imperfect (incomplete) tenses.  You desire is for your husband and he does rule over you and your desire will be for your husband and he will rule over you would have the same tense.  As a Christian I believe that the Bible is the best tangible guide to discipleship of Christ.  I also recall that in Genesis the serpent twisted God\'s words to fit his purpose, and I think that this is still a common practice.  I think you must interpret the message of the Bible for yourself through prayer, in doing such I do not believe that I have cause to acknowledge that the Bible gives women a position subordinate to women as I do not believe that is God\'s word.  I think it is in the interpretation and the context that it is percieved that way.  The Hebrew word for spirit is ruach, which is a feminine noun.  In the beginning of Creation we know that God and the Ruach HaQodesh (Holy Spirit) existed.  You cannot seperate Elohim (the Hebrew word for God) from the Ruach HaQodesh, in fact Elohim is a plural noun, as it ecompasses the also the Holy Spirit or the trinity. I believe that is the way it is intended to be on earth male and female to be necessary to each other to help each other and not for one to be subordinate to the other. '),(24937,3816360,'12,July,2004','2004-07-12 00:00:00','   There are some really interesting theories floating around amongst Christian feminist scholars.  They are saying that some of things in the Bible regarding women were either misinterpreted or  purposely misleading to strengthen male roles.  Genesis 3:16 NIV says ...Your desire will be for your husband and he will rule over you.  These scholars believe that it was actually your desire is for your husband and he rules over you.  One is a prescription for how things will be and the other is a statement of how things are.  Then there\'s issues over Mary Magdalene, the scholars say that there is no evidence whatsoever that she was a prostitute.  They assume it was assumed that she was because it is known that she financially supported Jesus and the disciples to some extent, however she was unmarried and seemingly independently wealthy. It is known that she was very important to Jesus some say that she was the 13th disciple others say Jesus\' wife.  Now here\'s the part that seems a bit defeatist she doesn\'t have to be a prostitute because she had money, but she has to be Jesus\' wife to have been important to what he was doing?  In one of the books that didn\'t make the cut into the bible, Peter asks Mary what Jesus revealed to her personally and what she reveals deals more with the metaphysical or Gnostic principles. Apparently Jesus had not gotten into that sort of teaching with the twelve disciples.  This angered Peter and he asked her why would Jesus reveal something like that to you a woman, but not us.  One of the other disciples stepped in and said something to the effect of if she was important to Jesus who are we to question it. But it was Mary who was there when he crucified and there when he rose.  The weird thing is nearly 2000 year later Christianity is not that vastly different.  Women can teach, but the men must lead. Women can be leaders, but not the chiefs.  I am aware that the bible has men as the leaders of the church and home, but isn\'t that contextual.  Did Jesus in all his time spent on children and women and the poor the nobodies of that society ever say that men were to be the leaders.  I have thought and thought and though I can think of examples in which a man would be better for the job I can think of just as many where a woman would be better for the job.  I shudder to think that the most important role a woman can play in advancing God\'s kingdom would be pastor\'s wife, and I shudder to think that in order for Mary Magdalene to have been valuable to Christ\'s mission at the time she had to be his wife as opposed to his friend, follower, and confidant. '),(24938,3816360,'08,July,2004','2004-07-08 00:00:00','   So plenty of my vegan and vegetarian friends have petitioned me to give up meat because of the awful steroids.  I admit steroids probably suck but, it\'s an issue of cost and convenience for me so I never really gave it much thought until recently.  Now anyone who\'s ever been around for long knows that my darling seventeen months old daughter is evil.  She furrows her eyebrows if you say anything to her without smiling, if you tell her to stop doing something she cries and asks to be held and seemingly lovingly buries her face into your shoulder, then she bites or digs her nails into your arm, or pinches.  She often confuses hello Ana for an attack and will smack the one who offered it hard across the face.  Unfortunately this offender is usually her four year old brother Joey and four year old sense of hope won\'t allow him to heed six year old wisdom, when Josh says stay away from her.  If you leave the computer for the bathroom she\'s sitting in from of it clicking and clacking away and furrows her eyebrows when you return  She drinks an exorbant amount, nearly a half gallon of milk every day.  So I blame the milk or the steroids in the milk for giving my baby a form of mad cow disease. I blame milk for making my baby evil and for the resulting hole she just made in my arm. '),(24939,3816360,'08,July,2004','2004-07-08 00:00:00','   So it seems you gotta do a whole intro blog and this is it.  I was more than happy reading the blogs of my friends from time to time, but the site wouldn\'t let me post a comment without registering for a blog, so I set one up but never used it. Until Josh wouldn\'t let me be the voice of reason and ignorance in keepin it real on his blog, so now I must stand on my own soapbox.  I\'d much rather do it with other likeminded people so there wouldn\'t be so much pressure, but as much trash as I talk I should be able to rant forever.  Oh yeah do me a favor and comment.  It\'s really weird to just say stuff and no one responds.  Tell me I\'m right, tell me I\'m crazy, tell me you feel me, or that I\'m wrong just tell me something. '),(24940,3816360,'08,July,2004','2004-07-08 00:00:00','   I watched a whole show about reality television the other day.  I felt a little bad cause they painted reality TV viewers as depraved. According to people who actually work for TV they\'re only about sex, drunkenness, and humiliation.  I get that it\'s probably just that to the producers, but I think people are always curious about other people\'s lives perhaps more so in the aforementioned areas.  \'See if that was me\' is a common phrase, I think people want more situations to say I would\'ve done it this way.  For me it\'s a great deal of curiosity, how do people live what do they do, why do they do it.  I also like to talk trash from time to time and reality shows fuel that.  As for the humiliation piece maybe for some, but not for me.  I think the situations that humiliate people are funny not the actual humiliation.  Case in point, punk\'d, most of the reactions are darn funny, not as much as Ashton talking trash, but close.  However when the person is actually humiliated i.e when Justin Timberlake called his mom in near tears that was sad.  I think identity is also a part of it. It\'s great to see yourself or people you know on TV.  So when I watched the real world reunion and Frankie\'s all it\'s not your problem but I just made it your problem and you misunderstood what I was saying when I said I hate you all and when I was pouting and being sullen and saying I don\'t wanna be here with these people I was actually having a good time and even though you\'re all sitting here with arms outstretched saying you\'d like to be my friend I don\'t understand why you all hate me. I thought I knew the only Frankie in the world and have often considered myself incredibly stupid for leaving that hand outstretched. Because every now and again my hand turns into a visit or phone call and a resulting psychotic episode.  Now I know my friend isn\'t the only Frankie and I\'m not the only dummy who, after hearing it\'s my fault but you did this and that and this wrong, still says lemmie get your number and I\'ll call you sometime. '),(24941,3816360,'06,August,2004','2004-08-06 00:00:00','   I just saw the commercial for G Unit shoes.  If you haven\'t seen it three little boys are peeking through a hole in a wharehouse window.  They aren\'t watching the usual image of Fifty with the Bling chain throwin money around.  They\'re watchin a shirtless, sweaty Fifty Cent work out.  The shots feature his back and chest.  The kind of shots that would make sense if teenage girls were peering through the window, but no it\'s three little boys.  Then he takes a drink of water and flashes the camera a come hither glance, and the three little boys run from the window.  Homo eroticism has never really been a hood thing, certainly not pedophilia.                                G-G-G-G- Gay Unit? '),(24942,4162455,'18,August,2004','2004-08-18 00:00:00','   We will have a 20 player game at my house this Friday night @ 8:00pm.  We may have a new player from Poker Meetup.  Let me know by email  urlLink jia@akaak.com  before Friday how many you will bring so we can make sure we have enough tables. So far we are estimating approx 20 but can break out a 3rd table if we get more.  '),(24943,4162455,'07,August,2004','2004-08-07 00:00:00','   We did not have our normal Friday night game due to Charlie but the next day I headed on over to Derby and joined the 170 player tournament.  I finished 4th and here the notes I emailed Jeff W. >>>  \"Now to the 4th place finish notes. It all started with my lucky \"Quake\" jersey. I wore this jersey when I finished in the money in 12th place so I deced to wear the same last Sat. : ) Basically I was at break even chips when they broke up my first table. I played 1 hand (AJ) on the button with a min raise and the big blind called and we checked to the river and I won it with A high. The reason I remember this hand is because that\'s the only hand I played early other than the forced blinds : ) Basically I threw all of the dominated hands Ax and medium hands like KTs or QTs and even KQ and KJ even in late position. I remembered I folded AQ in early position because I felt there was a raise to come and sure enough a guy went all in with JJ and won that. Anyways.the way I won was basically only playing level 1 hands and playing them with an all-in raise. In the 6 hours that I played, I got 2 pocket Aces, 2 pocket Kings, 3 or 4 Big Slicks, and that\'s about it... I won all the show downs with the 5 or 6 all ins that I initiated so I did get great cards that day which helped. I also got a great straight draw on the river late in the tournament with an A2s vs AQ to keep me alive. My 2 mistakes of the tournament were : 1) limping in on the button with QTs and getting an All-in raise by the most aggressive guy in the tournament to my left which I promptly folded 1/3 of my chips - stupid move - from now on I will not play a hand that I am not ready to go all in with. 2) the last play of the tournament in which I was on the big blind with a Q2 with 9500 chips left after the big blind of 4000. No raise from the little blind (2nd most aggressive of the final 4 but a very skillfull player which I learned after this trap). This guy would go all-in at any sign of weakness just like the guy to my left. After the flop of Qxx I check with my weak kicker to the river which produced a 2 for my 2 pair. Now I looked at the board and saw a possible strait for him if he had the J... He made a great bet of 4000 and said \"i bet the minimum\" and trapped me. I didn\'t think he had the J so I went all in and knew I was in trouble when he immediately called me as I was reaching for my chips. If I folded then I believe I would have made it to 3rd or more as the blind would put the other player all in. All in all it was a great tournament and I learned alot and will try to repeat soon. I think the key is to wait for the level 1 cards to play and play them strong and hopefully they will hold up. The best hand of the tournament for me was getting AA on the big blind and having a mid position player go all in with KK ! I won that matchup as well as AK vs AJ which got me enough chips to make it to the money. I\'m estimating that 3000-5000 chips should be enough for you to sit on to make it to the money.\"  '),(24944,4162455,'07,August,2004','2004-08-07 00:00:00','   Some of our league members are playing the biggest NL Hold\'em in Tampa Bay today - the 12:00pm 200 player No Limit Hold\'em tournament at Derby Lanes in St. Pete. ( urlLink http://www.derbylane.com  727-812-3339.  Everyweek this tournament sells out at the open because they allow you to pre-register Friday starting at noon.  Most players purchase their seat the day before.  If you show up Saturday without a seat ticket, make sure you be at the Plaza door at 11:00am and get in line before they open the door at 11:30am.  As soon as you get in, run straight to the escalator, not the elevator on the right (much slower), and take a right and then a left to the next escalator.  Then go thru the glass doors and go to the left and get in line in front of the counter in front of the poker room.  Since 3 weeks ago, this tournament has sold out every Saturday.  The 1st place prize is $2,000.  I know a player who finished 4th a few weeks ago and it paid $650.  I finished 12th once and it paid $137.  '),(24945,4162455,'07,August,2004','2004-08-07 00:00:00','   We had a good game Friday night at Dale\'s. A total of 19 players showed up. Gene, a new player, won the top prize of $100. Bad beat #1 : Jeff C.\'s pocket queens vs. John\'s pocket aces - All In Action - Jeff gets a queen on the flop to take out John. Bad beat #2 : Casey had AK and the flop gave him trip kings... Casey then goes all in and was called by Bobbie who had a pocket pair... Bobbie hits trips and takes Casey out with a full house.  '),(24946,3572448,'08,June,2004','2004-06-08 00:00:00','    so here i am...halftime of game 2 of the finals.  The Lakers seem to be irritating the Pistons and causing defensive problems for them.  I hope the Lakers lose, i really don\'t like them, and that Kobe, he\'s so smug--i hope he gets convicted.  Anyways, i\'m in charge of the house again, my brother went over to his \"girlfriend\"\'s house and probably won\'t be back.  As for Jacob, who knows, i haven\'t seen him in a week or so.  I need a hobby or something man, i get so bored.  And food, boy have i ran out of ideas of food to eat, makes me kinda wish i was at home and didn\'t have to buy my own food and didn\'t have to think about the meals i have to make.  It\'s nice to live in the city and be semi-independent, but it feels like i\'m missing something...maybe i\'ll go start on the server.  now that i have the software to do it i think it\'ll be pretty cool to have a server set up for the house.  I know it\'s not necessary but it might be fun and i might even learn a few things along the way   '),(24947,3572448,'07,June,2004','2004-06-07 00:00:00','    looks like i\'ve joined the crowd...i have a blog.  does that make me trendy? am i hip?  if only i had something important and mind bending to blog about.   '),(24948,3572448,'18,June,2004','2004-06-18 00:00:00','    have you ever had an indeterminable sadness wash over you and leave a residue you couldn\'t shake?  i\'m pretty sure that i\'m not the only person who feels this way.  i just find myself wondering about my future.  mostly, i figure that i\'m just lonely.  i\'ve never been in a relationship and question whether i ever will.  i mean, if i\'ve gone this far and never been involved, will i ever?  its the early years when people meet and learn how to get along isn\'t it?  i\'ve never felt like i fit in and it seems now that i never will.  where is it all going?  maybe i\'ll leave a life of pseudo happiness and then die.  i say pseudo b/c whenever i\'m left will my thoughts i feel like theres nothing here for me.  whenever i\'m with company i\'m the happy little center of the universe (to paraphrase ed norton).  its not that i am happy, but you flash a smile at the world and the world embraces you.  so in this way i\'m pseudo happy.  i\'m on one of my stints where i quit on females.  it just seems easier that way.  when you put yourself out there you just find sadness, rejection, and disappointment.  too much roller coaster.  what if, what if, what if...   '),(24949,3572448,'24,June,2004','2004-06-24 00:00:00','    hmm...i am the captain of ambivalence and i need directions.  another day, another dollar.  work was boring, luckily it was interrupted by a brief stint of mind numbing in cs336.  speaking of which, perhaps i should prepare for that quiz tomorrow...  Grace:  this isn\'t a secret diary per se, but a place for me to lay down some thoughts (read the headline :p).  they may seem personal, but aren\'t everybody\'s thoughts personal?  directions anyone?   '),(24950,3572448,'22,June,2004','2004-06-22 00:00:00','    so, all went pretty well today.  i think i kicked my cs336 midterm\'s ass.  that class is going surprisingly well, maybe i\'m a genious though...afterwards i had the opportunity to meet up with and have some dinner with jennifer (friend from accounting).  She\'s a really attractive girl, and i want to be able to connect with her, but so far it hasn\'t been there.  i figure we\'re still \"feeling each other out.\"  Oh well, can\'t be sad about having dinner with a smart and attractive lady!   '),(24951,3572448,'20,June,2004','2004-06-20 00:00:00','    i think the title says it all.  i need a life, or something.  maybe i just need some hobbies--read a book, fix something, build something--i dunno, maybe i don\'t have a niche.  i have a test tomorrow and i\'m debating studying for it some more.  i\'m also meeting with jennifer tomorrow (tentatively).   not quite sure what to expect out of that, but i guess i\'ll see.  other than that just emptiness on my mind.  where to go what to do, these are the questions that i constantly put myself through.  perhaps someday i won\'t--yeah i threw some optimism in for kicks.  blah blah blah.   '),(24952,3572448,'01,July,2004','2004-07-01 00:00:00','    sometimes i feel like disappearing for a while.   '),(24953,3572448,'27,June,2004','2004-06-27 00:00:00','    things i want to learn how to do: 1) Dance - classical and swing are the 2 foremost styles i\'d learn, but anything else is game too. 2) Play the piano - seems like something fun to do, but first i need to learn to pat my head and rub my tummy 3) Cook - i\'m thinking of taking some classes after i graduate so i can cook bad-ass meals. 4) Mixology - if i\'m going to be a killa cook, i need to know how to mix a good drink, or be a proficient alcoholic....  once these tasks are completed, i can make a tantalizing meal for a woman, give her a nice drink, serenade her on the piano and then dance the night away...or just know how to do them to further myself, either way.  i\'d feel all kinds of metropolitan if i could do all these things.   '),(24954,3572448,'07,July,2004','2004-07-07 00:00:00','    \"Here\'s to today. Because we can\'t see the future and all we can do is live now...And I will wake up tomorrow. Because I want to live.\" --grace  where does she get these thoughts?  learning people is so interesting...but to add to what she said, sometimes thats all you can do is wake up and want to live.  life holds an infinite mystery about tomorrow...here\'s to today and tomorrow, cheers.   '),(24955,3572448,'15,July,2004','2004-07-15 00:00:00','    i think i\'m normalizing.  >:o   '),(24956,3572448,'14,July,2004','2004-07-14 00:00:00','    so i decided to explore myself (not like that, don\'t be sick).  I\'m trying out my green thumb in the back yard (again, don\'t be sick).  I bought an Allamanda and a Texas sage and planted them in our little garden area of our back yard.  Its kinda neat growing things, but people look at me funny whenever you talk about it.  Also, i\'ve been running periodically the last couple of weeks, it feels good to be active again.  I bought a pair of running shoes and basketball shoes so now i have no excuse.  I managed to juke the rubber right off my first pair of running shoes (which i only had a week), so i made the plunge and finally bought another pair.  I had to stop 3 times when running around the lake tonight, but i don\'t think its bad considering i don\'t remember the last time i ran--any distance--at all--period.  Also i\'m just about done reading Siddhartha, actually re-reading, but i like the book and decided to read it again...what to read next...OH, i also made an A in cs336.  this is turning into the summer of Matt after all, i feel like i\'m growing as a person.  good times.   '),(24957,3572448,'21,July,2004','2004-07-21 00:00:00','    still not much happening around here.  i finished reading Siddhartha and am debating whether i should read a book for my CS349 class, or another book.  speaking of cs, for my 2 classes i need 6 books combined...i feel like a freshman again having to buy all these crappy books, not to mention the cost--so much for updating my wardrobe.  Instead i get to update my bookshelf.  that reminds me, scott (my brother for those of you not in the know) started putting down the pergo today.  He\'d done most of the hard stuff by the time i came back home, but i still put in a few pieces.  The house is going to look badass by the time we replace the crappy carpet and linoleam (that doesn\'t look right) with pergo and tile.  Hooray for a successful brother!  And since i have an abundance of free time i\'ve been running and playing some pick-up basketball fairly frequently.  I just wish my knees and feet wouldn\'t get sore afterwards.  But it feels good excercising again.  Now i just need to start playing raquetball again.  i probably should lift some weights, but i\'ll probably stick to pushups and curls...oh my legs are going to be sore tomorrow.   '),(24958,3572448,'28,July,2004','2004-07-28 00:00:00','    so, i\'m sitting here at work, and i says to myself, \"i\'m bored...what am i going to do?\"&nbsp; That\'s basically how my day has gone thus far, pretty slow.&nbsp; the only thing that i\'ve done is get a user\'s microphone working again, and that was just by plugging it back in....so now i\'m killing time while i update another computer.&nbsp; only 2.5 hrs more to go...  However, it is kinda neat to be in charge while my boss is out.&nbsp; Hooray for the positive spin.   '),(24959,3502783,'31,May,2004','2004-05-31 00:00:00','   urlLink    sick reality that is&nbsp; urlLink  '),(24960,3502783,'31,May,2004','2004-05-31 00:00:00','   urlLink    Not much to say, is there&nbsp; urlLink  '),(24961,3502783,'31,May,2004','2004-05-31 00:00:00','   Growing up  Songs you have grown out of Cant quite bring back those secret paths  To secret lands trodden. Ideals adopted through these whiles That crashed and tumbled downhill Hindrance alone in matters that  Lost sight of peaceful quiet Roads no longer defined, demarked Just walked on, as they present  Ideals crashed  With might unknown  Saves the hidden prize Barring doors to a past unfelt Just observed as the passerby.  '),(24962,3502783,'31,May,2004','2004-05-31 00:00:00','   and got most things done on that previous list  have to call for plane tkts still and tito...this has to happen tomorrow or else am fucked badly. Nice convs with general crowd, and over the email and things  Im not sure what to do with ballet shoes, might wear them about, but feet hurt, I have new respect for dancers  I HATE rem\'s rendition of One, he sounds like a bee gone out of control.   The danube featured, and was happy it came up, it is a mystery, to me too.   Sohini is funny, have to upload more pics soon  A shivdasasni\'s mail was encouraging, and now jsut have to check out the websites remember to call and bug claremont tomorrow, and send stanley poems, do you know which?  And email ppl about stoop sale details, cant be there before 10.30  '),(24963,3502783,'31,May,2004','2004-05-31 00:00:00','   things to do on memorial day  interview cancel CT meeting discuss stoop sale avoid conversation with her find bar anti-war fraction meeting work so many hours... send poems fix up plane ticket!!!!!(urgent) call tito.....(PLEEEEEASE!) look at st luke\'s website  whew (and just noting these down...)  ');
INSERT INTO `msgs` VALUES (24964,3502783,'31,May,2004','2004-05-31 00:00:00','   is a hit  echoes of an earlier tune have elevated it   haha  '),(24965,3502783,'27,June,2004','2004-06-27 00:00:00','   \"Pooh, promise you won\'t forget me, ever. Not even when I\'m hundred.\" Pooh thought for a little. \"How old shall I be then?\" \"Ninety-nine.\" Pooh nodded. \"I promise,\" he said.   Taylor\'s Good Food cafe also had the above to offer.  '),(24966,3502783,'24,June,2004','2004-06-24 00:00:00','   cracked me up completely. to see it being written in ink in earnest, so carefully, made my day actually. Love Feruze, and her aunt and uncle. 7th July it is, then.  Shouldnt have indulged in needless walnut brownie after oath from a few weeks back. And conscious work of art, is cafe latte and cafe mocha at Joe the Art of Coffee. Whether this becomes haunt for the next couple of months, I wait to see, with bated breath.   Lots of conversation on teh phone with people sincerely trying to help. Appreciated it, but of course the slog is upto me. never said I wasnt up for the challenge. Housing is next on the list, and will probably have to visit Claremont to check out affordable places.  What bugs me the most is people never grasping how adversely their concern manifest as panic affects me. Ok futile thread this, so will skip this. I will be the stronger one and ignore this and distance myself from it to function as productively as possible.  I live on what I understand is a healthy dose of denial. Why else would I stop Feruze from pondering about life sustained by email interaction between two ends of the world?  Also, probably because you have already reconciled to such a conditioned way of life.  SO teh saddest news of today is that India trip has no scope of materialising before December...dec???????????????? getting back to denial.....  Giving a talk next week. About time too. On the fake transfer of power in Iraq. have an enormous amount on my plate, but have decided to go for it. good talking to Brian today, cleared up some key perspectives. looking forward to writing it Saturday. not have  one inch of time before that.  Murphy somehow Will squeeze in to spice things up for me.  further confirmation comes from fact that I get invited to New Jersey the very same day Ani might be coming to NY. Second time I am presented with opportunity to Prioritise (different)friend over (same) family in less than a month\'s time. Bizaaaaarre. Well not like I ever claimed to be a family person. Seriously, trite conversation not worth the bus journey. Weighing that against winded walks around the Village or wherever else for that matter. So Easy it is too.           Meanwhile the pile of pending work mounts higher and higher.  Reminders for tomorrow: pay bill, call people about news, call student financing office, look up local newspapers for housing.  And was just mesasaged about friend\'s relapsed sex life, with assurances of further notes. Would happily like to mind my own business, though.   \"So I\'ll just keep on walking down this open road\" Closing with Gary Barlow, ie   Not quite strange that this be the refrain for the day.  '),(24967,3502783,'23,June,2004','2004-06-23 00:00:00','   sauravbasu79: tor gan shonai byaghat ghotanor jonno dukkhito .... sauravbasu79: thank u for the testimonial .. winky_9000: maane ???? winky_9000: byaghaat keno?? sauravbasu79: & yes , lovers song was outstanding acording to me .... winky_9000: just dont feel like it atm sauravbasu79: because .... I think it is one of the few times I have seen u take a areial view of life ... winky_9000: hae bhaablaam likhleo hoy winky_9000: \"aerial\" na \"a real\" ? sauravbasu79: aerial  sauravbasu79: maney just like the pots I like .... sauravbasu79: poets .. winky_9000:  winky_9000: baba... winky_9000: I started thinking of you smoking pot sauravbasu79: aar hya , chobigulo dekhchi ... sauravbasu79:  sauravbasu79: hya , chobigulo dekhchi .... tobey fist e bujhtey time laglo KONTA notun ... winky_9000: ok aerial meaning me seeng things from afar, right? winky_9000: KONTA maane, the chicago pictures, the one I sent you yesterday winky_9000: teh link sauravbasu79: yes ... exactly ... pure , unselfish , godly , summing up of ur OWN perceptions ..... something that can be called as an offering , & not a gatification ... maney onekta gitanjali r moro ... apon-onjoli noi ... sauravbasu79:  sauravbasu79: KOI ???? kono link pai ni to ???? winky_9000: eki.... sauravbasu79: jakgey .... ami bujhtey perechi ekhon ami ... chicago tai dekhchi .... winky_9000: what is the dff b/w apon-onjoli and gitanjali ? sauravbasu79: amar kachey prochur ..... winky_9000: ok good winky_9000: shetai jaane chaai... winky_9000: tor kache ki... sauravbasu79: for me ... gitanjali is like offering a homage to life , to ur own god higher than ur own problems ... sauravbasu79: but apon-onjoli is like explaining urself to u , self justification ... winky_9000: bujhlaam winky_9000: and lovers\' song was gitanjali sauravbasu79: hya , exactly ....  winky_9000: aacha , all explained sauravbasu79: it seemed to me like u were at that point higher than ur state ... sauravbasu79: going to the next room ..... in case u message , there may be a delay .... in case u dont ... happy listening ... winky_9000: aacha na will prob go, tata then and tell me abt chhobis sauravbasu79: I will write abt chobis ..... VERY soon ... sauravbasu79: surely .. winky_9000: and am posting this conv on blog sauravbasu79: THIS one ????? sauravbasu79:  winky_9000: yes sauravbasu79: ok ... sauravbasu79: amra odbhut , amra kimbhut , amra ... , notun joubon eri duut ... hahaha ... tor uddeshey bola ... winky_9000: thank you, I love it sauravbasu79: welcome  '),(24968,3502783,'22,June,2004','2004-06-22 00:00:00','   So Claremont happened ! YEAH! Im still in shock, btw now money situation to be tackled, oh well, one at a time please, before you drive yourself completely nuts.  Chicago etc were absolute riots. tripped out everywhere, lots of drink, but handled so impressively. Met Pakistani, who has to be called about Coalition meeting next week. I would have liked to work at his agency though. Shelley\'s revolutionary poetry talk and subsequent discussion with David and Maya had me in tears, why Im not sure. Probably because Im on teh right path only. Ooooh and 80s disco, so out of the blue and so appreciated. I like Nathan also, he is quiet sort who acknowledges most everything. Helped him decipher crucial reference to Mask of Anarchy. Proud. Conversation(s) with Bill Keach also a delight. those were the defining moments.  Met Frenchwoman who will contact me during her NY trip. Dekalb was fantastic, loooved night out in the woods. Duck groaning and policeman prowling about notwithstanding.   too bad I lost one of my lenses teh next day, but we made up nicely, Sohini and me (should I forgive people so effortlessly..). But there emerges another excuse to go to Calcutta. will post those b&w pictures as well. I love me in b & w. Spooky pic of Suchitrita and me likens ghost sequence in Goopy Gyne Bagha Byne.   People continuously paying me in cash in public bus was a bit embarrassing, for several reasons. I had no idea what they were doing. I also had no idea we were going dutch. Something to think about.  Evanston is truly pretty. Kafein was great and being thrown out of there at closing time, happily reminds of similar fun times in the pleasant past. Shakespeare\'s garden was such a waste, but so lovely. Ani is a merry tourist guide. lake Michigan behind rocks in almost pitch darkness brought back first offcial epic poem. Did a patch of Chicago today morning, and observatory was a swell idea. I love the blue-green of the lake. Also concluded I do not like martinis. Some shop called Seven Seven Seven made me laugh.    I Am tireless.  And what of when I get back, hungry, pissed after delayed flight, just wanting some decent conversation?  Bombarded with questions about funding and admissions and explanations and followups to be done.  There I went thinking I now have an inch of breathing space. it\'s like nothing has changed. taking time off to celebrate would have made more sense, no? Especially after months of numbness and distractions galore. Pretty mad with Baba for doing this to me. good I will finally be moving out, it is much needed and has to happen before I screw up relationship with my parents totally. How we manage I dont know.  go to sleep now, you are tired and sleepy. work resumes tomorrow. have truckloads of which to wrap up in a couple of month\'s time.  but I do know what I want to do for birthday.  '),(24969,3502783,'16,June,2004','2004-06-16 00:00:00','   is the reality check that was today. Ok that is misleading, am very much aware of goings-on around but today was unexpected, and of the sort you\'re wary of.  So the news of teh day is, I will have no place to stay in like 2 months. Everyone going back home in August. Will also have no job in less than a month\'s time.  So why am I so detached from my life, still? well, actually you\'re not, cuz you\'re too busy Living your life.  Claremont has me hanging by a thread about to snap. the edge is where I am. There\'s just so much I can humanly do. Thus went the latter part of the day.  good thing I got work done earlier, before the news was dropped. trainride back home was uneventful except for hep bangladeshi woman talking in bangladeshi bangla and indian man on cellphone rambling on in very bad english. oh and same woman on train earlier was standing next to me.   Met Jenny, her concern touches me. Good talking outside Cupping Room Village, another one crossed off the list. She made me laugh with her very recent discovery of her relatives in the US. I was talking too much. I love dogs, especially the small, brown one smaller and cuter than a fox. have to find out what breed it is.  Got thoroughly disracted at Jembro, though it was worth the splurge. Me highly-strung was the reason. Should never let people sway me into making decisions, even her. Have sworn never to take the F train back from West 4th again. Twice bitten.  Hopefully Chicago will afford more distractions for the next 6 days. Still havent read stuff for the workshops, hate that Im neglecting those. remember to carry Leon Trotsky\'s essays book. Hope to get some reading done during flight. Sanity still sought, so this should be a well-earned time-out.   Soundtrack for today: West End Girls via Pet Shop Boys. Funny.  And am numb waist down to toe for the last few days. This freaking weather...  Have so much to write, but am so pressed for time.  Will go hunt for bag to take now.  '),(24970,3502783,'12,June,2004','2004-06-12 00:00:00','   Blissful at the beach, that\'s what I was today. had almost forgotten how much of a sea-person I am. Mountains do feature, and I do get lost in them, but there\'s a dynamic about the ocean that touches parts nothing else can. will dedicate more time to it.  Fleeting thoughts arose, namely this was a full circle. The beach is where I began, and now Ive returned to it. Always remember and acknowledge your roots. There is a lot of unfinished business to be taken care of. I am sincerely thankful for that which started me off. it will linger as images do.  Forgetting my camera was fate leading me to those glory days. Could concentrate more on myself, for one. Also played with bright green seaweed and dull grey gulls. Liked myself on that rock. Felt an overgrown patch of moss on dead wood. Sand below the feet was comfort forgotten for long. No resistance offered, just sinking me in. I shouldnt wear shoes.  Turkish coffee on way back ensured another addition to list of places to be explored further.   Sandy.  '),(24971,3502783,'10,June,2004','2004-06-10 00:00:00','   Was a smash hit. Everyone gelled fantastically. L\'annam was a success, great service, over-zealous waiter, good food, drinks even better My camera almost let us down, but pulled it off.trippy pics the result. Deborah L Coates should be proud to have a bunch of such diligent associates, toasts to her for bringing us all together for teh time of our lives.   Coffee Shop, my recommendation again (yeah, yeah!!) was the icing. Underground lounge got headier as the night carried on. Thank god talked people out of the movie idea. I will maintain for the rest of my life, that nothing eases the bonding process better than alcohol, and bottlesful of it. Several blue parrots, mojitos, ginger kazzis and chocolate martinis made teh night, as everyone recounted their \'horror stories\' to each others\' delight. such a common thread there. Why do these things happen to good souls....dancing good too. And Ishtar\'s bi as well, something I didnt see coming.   I have now confirmation of what stuff people perceive Im made of...\"I think it\'s beautiful you can say \'I dont know\' like that..\" cheers to you michael, and everyone else for reiterating that, after my vague allusion to my future. I am such a loon.  Also died as I realised time had come for yet another round of goodbyes... Got very emotional when Michael wondered when he\'ll see me again, very unanticiapted and blew me off course completely, wept like a baby. It hadnt occured to me that this was a last time of sorts.   Point is we killed all last night. And will happen again. Pictures prove my point.  '),(24972,3502783,'10,June,2004','2004-06-10 00:00:00','   Oh what a day, loads of work done and am kind of beat, not like it\'s over for tonight though. So this is a break, then.  have cover letters to write, which I soooo hate, but whatever. Glad Chicago trip is finally taking shape. Good fun with Suchitrita/Sohini/Ani lies ahead, not to mention teh conference. Greatly confused about workshops to attend, but will be a last-minute thing as ever. Hope hotel is in a good location, meaning there are decent plcs to eat/drink around. DO I know how to get to dekalb from chicago..no. will figure out. Remind me to remind Sohini about visiting that nice coffee plc on teh way to downtown Dekalb.   Lovely weather today, been anticipating the rain all day, but no one missed it when it never showed up. Took vague pics of skies and trees, just remembered to upload them. Might write something too.   Love 99 cents stores, they are crammed with goodies you cannot even imagine unless you allocate time to rummage through them. Got mug with Hai Xin and green/white pandas on it. Disappointed that Hai Xin is jsut the company name...what a let down. then there\'s teh music or the strange mish-mash of it that customers were being bombarded with.  I saw the birds (thoroughly unimpressive, brown hanging ones) that were squeaking out of tune. One of those other annoying show-pieces were, as is routine, churning out teh Dr Zhivago theme. There was also a Chinese singer crooning, with brief interludes by a mooing cow. Simultaneously.   Ironically I now have slippers which say California Girl Sport.        Saw woman trying to coerce her man into waxing!!!!! something, somewhere...what a nut! Poor guy.     People are straaaange. I am not drunk.  '),(24973,3502783,'07,June,2004','2004-06-07 00:00:00','   Loversâ€™ Song  Loversâ€™ song there is none No music to lead from Except the notes that bind No matter in the mind As long as lifeâ€™s afloat  Canâ€™t ensure a steady boat But lovers recognise The best course to explore And plod away at oars Calming the tides. When Storm signals cry out The chuckling abounds To welcome drowning gusts     And feed into the tumult Till lovers exhausted and spent  And damage a trivial dent On the way to salvage and protect And effortlessly put to effect A funnily simple prelude To a resounding tune Till date unheard and unsung By loversâ€™ lips and tongue.  '),(24974,3502783,'05,June,2004','2004-06-05 00:00:00','   urlLink    like I was saying&nbsp; urlLink  '),(24975,3502783,'05,June,2004','2004-06-05 00:00:00','   hmmm just when you thought you were managing, people descend on you and upset the healthy trend. Why people do not understand or why they care so much is beyond me. is this another reality check, well definitely feels like am in there. Wish I had an older bro/sis to tell me of the nuances of the world outside of my head. Since that\'s never coming true, let;s get back to what you have to do. Do it, ie. ok I have got myself a pair of decent, comfy shoes today and will continue from there. maybe I should just explain to people that I can live with glaring uncertainty, and they need not worry about me, which can potentially translate to worrying me.    And adding to usual pains is cable connection utterly messed up, I will die without the internet, dependency sucks.    Bittersweet symphony featured somewhere today and has again confirmed why Richard Ashcroft is god.    \'Cause it\'s a bittersweet symphony, that\'s life Try to make ends meet You\'re a slave to money then you die I\'ll take you down the only road I\'ve ever been down You know the one that takes you to the places  where all the veins meet yeah,   No change, I can change I can change, I can change But I\'m here in my mold I am here in my mold But I\'m a million different people  from one day to the next I can\'t change my mold No, no, no, no, no  Well I never pray But tonight I\'m on my knees yeah I need to hear some sounds that recognize the pain in me, yeah I let the melody shine, let it cleanse my mind, I feel free now But the airways are clean and there\'s nobody singing to me now  No change, I can change I can change, I can change But I\'m here in my mold I am here in my mold And I\'m a million different people from one day to the next I can\'t change my mold No, no, no, no, no I can\'t change I can\'t change  '),(24976,3502783,'05,June,2004','2004-06-05 00:00:00','   things started vaguely this evening, do not appreciate attitude problems especially when you\'re giving clear-cut directions to a plc right across the street. So fine Esperanto Cafe isnt written in neon signs, but if I tell you to cross the street, you do so instead of walking away.  It is in this context that I dont like being made to think it\'s my fault you cant find a freaking coffee shop. ok the rest of the time was fine, in fact great fun, would have flirted more with waiter if he had been less stud-like. huh, interesting. roomali ensured chicken tikka roll be dinner. falafel for next week then. the pan shop will continue to freak me. madras mahal seemed cool, have to visit too.  thai playboys and red devils happened, all of which got me very happy, everyone acknowledged there was no one to take care of me. I miss tarun and car after such blissful episodes. I love yumiko drunk. her friend already thinking of me going to japan was startling to say the least.  Got u2 poster too, of them so young, and interview done too. Will post pics soon.    Im a bit shocked that Im so detached as with or without you carries on.  '),(24977,3502783,'04,June,2004','2004-06-04 00:00:00','   am soooo tired and sleepy, whole day in the sun does that to you hated clinton\'s speech, hated potential stampede to shake his hands even more. Why am I always caught in the middle of stuff I want to stay away from? Soul food happened, and crabcake and jamaican rice were the highlights. I love Harlem. got pics of bday in 2003, delays kill me.  '),(24978,3502783,'03,June,2004','2004-06-03 00:00:00','   Meeting was a blast, Michael, Ishtar and me cracked each other up, not uncharacteristically. We make such a wonderful team, the \"dynamics\" of which took me a while to grasp. And I honestly didnt realise how rampant teh practice of imitating my \'yeahs\' is. (I, of course do not understand this)  Best part: this is what happens when Rupu leads team meetings. AS we acknowledged, covered everything on agenda, and so much more. Made considerable progress with subsequent work as well.  Next stop, next week: L\'anaam (vietnamese) on 3rd Ave & 28th Street, looking forward to meeting all our colleagues there, but just the 3 of us will do as well.   Spoke to Tito..at last... tkts seem to be on teh way too, things finally moving. good to make Bong connection here, hoping things will work out. Wish I had learnt about the art camp a day earlier.  And overheard Dwaine muttering quite happily about the virtues of his hot stick to his gf over teh phone, right behind me. Which was funny. His voice went down to eerily muffled whispers soon after. Which wasn\'t. had no clue they were such a hot item.  That was Michael\'s contribution today, to all future conversation in the lab - hot stick what he dubbed the USB drive that\'s gleefully replacing the floppy.  '),(24979,3502783,'02,June,2004','2004-06-02 00:00:00','   I Saw the Sun Rise on the 7 Train  I saw the sun rise on the 7 train I saw the ink turn into blue I saw the moon fall on the 7 train I saw the orange give way to sky I saw the white catch up  With the dawning day And spread over the patch anew I saw the dawn transform itself From blackness pitch dark  To sailorâ€™s shade To clear Danube behind the trees To bloodstains capturing heavenly birth Masked solemn by traces of sun An azure greeting the sleepy nods As crimson leads forces of daybreak This unassuming work of art exhibited Brushes in joys of the day awaited As cobalt over steel, rests its case.  '),(24980,3502783,'01,June,2004','2004-06-01 00:00:00','   urlLink    let\'s hope rowling\'s kind to him&nbsp; urlLink  '),(24981,3502783,'01,June,2004','2004-06-01 00:00:00','   urlLink    my love&nbsp; urlLink  '),(24982,3502783,'01,June,2004','2004-06-01 00:00:00','   Chicken Shakuti II (chicken cooked in coconut)  Ingredients  1Kg chicken, cut into desired pieces (or or off the bone) Dry roast the following in a large pan (until coconut is a light brown,  taking care not to burn the spices):  2 tsp coriander seeds  8 dried Kashmiri chilies  1/2 tsp cumin seeds  1 tsp fenugreek seeds  5 peppercorns  2 tsp peanuts (unsalted)  1/2 coconut, grated  1 tsp turmeric  4 cardamoms  6 cloves  1 inch piece of cinnamon  1 lemon  salt to taste    Method  Add the turmeric, cardamom, cloves and cinnamon to the roasted spices and  finely grind. Fry the spice mixture in some ghee for a few minutes. Add the  chicken and brown. Add salt and a little water and simmer until chicken is  cooked. Sprinkle over lemon juice a few minutes before serving.   Well this is just a lame attempt to let you take a whiff of the heavenly scent of the above combination of spices.  It is proshaad in its purest form, almost untouchable in its beauty.  High-inducing, orgasmic...I can live in a roomful of this aroma forever. Ask me to cook it, and I will do so joyously, creating it just makes it my baby which is the entire point. Cooking for the ecstacy of cooking, where have I ever heard that before? Resting my case for the day.  '),(24983,3502783,'01,June,2004','2004-06-01 00:00:00','   Forgot to get PD guide to work, hence have to make up more hours tomorrow... whihc Is why trying to figure out how to occupy myself productively today, ok schedule more interviews.  Ok cell phones have so much potential, yet they end up adding to your agony. How many ppl do I have to talk to now..bet they\'ll all call at the same time. Like Murphy and technology ganging up against you. Things get progressively crazier as the day moves on. Will post I saw sunrise.. I am quite taken with it, dont have copy of it here at work.  '),(24984,3502783,'01,June,2004','2004-06-01 00:00:00','   http://www.nysonglines.com/  is THE site to check out for your trip to ny, still havent gotten over it and am still languishing in the discovery of it.  ok according to my printout there was this place on Christopher Street which is called Fat Cat Billi, possibly a place for World Clothing   was too excited! and such a brilliant choice of words.  found it. ALso found it to be jazz bar called Fat Cat with Billiards table, etc.  too bad Yumiko didnt happen to know Hindi to appreciate and join in my hysteria.   oh, my printout had cut off the right end of the page  NY story 1  '),(24985,3502783,'01,June,2004','2004-06-01 00:00:00','   The Screening Room: Donau, Dunaj, Duna, Dunav, Dunareva  Directed by Goran Rebic, 2003, German with English subtitles, 90 min. (fiction/drama)  Cast: Otto Sander, Robert Stadlober, Annabelle Mandeg, Svetozar Cvetkovic,Florin Piersic jr., Volodymir Goryansky  For the last time the old Captain Franz navigates his rusty ship down the Danube to the Black Sea. On board is an odd mix of people. All are runaways but for different reasons. We are all made of 90 percent water. How much of it is tears? Or Danube?   mystery solved   Wonderfully attractive actors. Lovely photography...every shade of blue, all times of the day.  when people start talking to you about Satyajit Ray in the Austrian cultural forum, it has to spell out a delightful evening   Also added to list of my apparent origins is German.  others so far... Spanish Caribbean Egyptian French (ok...) Iraqi Bangladeshi(sooo close..)  '),(24986,3502783,'01,June,2004','2004-06-01 00:00:00','   Now that my god is up here as well, the place feels a bit more homey, and so can dig into it.  Played phone tag with brian the whole day...hopefully will get to talk tomorrow. Am now excited about the stoop sale, looks to be fun, should figure out when I can sort out items to donate from the mess that is my room.  I need plane tickets badly, I still dont have them. why cant ppl return calls so I can talk to them and figure my plans (however unplanned) out...    tomorrow will get royally crazy as well, hopefully i will remember I have interview at 4, meeting at 2 and ceremony tkts to pick up sometime before that. I dont want to hear Bill Clinton anyway. Such a waste of time it\'ll be. Now that my last month of work is drawing to a close, Im edgy about forfeiting even an hour of work... after which I might have to spend time with her in vague, uncomfortable silence. god, the conv today, initiated by me why I dont know, was so freaking straaaained. So Saturday\'s going to be fun, actually it will, there\'ll be others around too, and alcohol and music will help.   look forward to more ny stories here, oh and the moonlight cruise. Send poems to stanley, you forgetful girl....  '),(24987,3502783,'01,June,2004','2004-06-01 00:00:00','   urlLink    him&nbsp; urlLink  '),(24988,3502783,'01,June,2004','2004-06-01 00:00:00','   urlLink    cool stuff and an excellent photograph&nbsp; urlLink  '),(24989,3502783,'24,July,2004','2004-07-24 00:00:00','   Reasons why I don\'t smoke:   *&nbsp;mouth&nbsp;will smell  * clothes will reek  *&nbsp;teeth will get yellower&nbsp;  *&nbsp;cannot bear to smoke into others unflinchingly  * cannot have people associate me with&nbsp;reek of nicotine  * cannot impose foul-smelling mouth on others (gum makes me nauseous)  * am not so insensitive to others\' discomfort  * too expensive a deal, have more fun&nbsp;priorities to splurge on  * health hazards   thus gave jenny advice on how to quit. she\'s trying too, good girl. Stress is such a killing factor. we decided to&nbsp;work within constraints as best possible. Counseling experience from Obesity Research Center, St Lukes Hospital&nbsp;was&nbsp; put to ample use, (memories of food records and rigid discipline&nbsp;came along) glad to have applied it in real life.&nbsp;Gradually easing it down is best and makes most sense.  Let\'s see progress, should keep checking up regularly.  will also&nbsp;ask Baba what he had done teh 5 months of his adult life&nbsp;to turn into a non-smoker.Will serve several purposes. maybe&nbsp;should have done this long time ago, but somehow&nbsp;am not motivated to help those&nbsp;seemingly unconcerned themselves.&nbsp;&nbsp;&nbsp;  have decided to spend one whole day losing myself in alleyways of west village.  panino giusto&nbsp;makes a great cafe latte, and&nbsp;made it cooler&nbsp;by dunking in brimful of cinnamon.  ended up sitting outside and scribbling&nbsp;random thoughts&nbsp;with our left hands in turns. My sentences&nbsp;ran diagonally along the page, though the leters were wellformed and steady. practice has done me good, have also taken notes with left hand, to kill monotony, among other things.  the bathroom reminded me of a cottage in rural europe, brick walls, flagstone floor, huge, aging mirror in frayed wooden frame, in about nine square feet of space. quite mesmerised.  feruze leaves for albania tomorrow, excited for her. know most people arent, about my trip to calcutta, but will hang on to wishes of those who are.  nidhisaraf will take leave from work to&nbsp;go if I do.&nbsp;her&nbsp;boss thinks&nbsp;she will be visiting her boyfriend.&nbsp;flattered. and also sad. such is the state of the world that recognises the&nbsp;urgency to make cross-country trips for the sake of a \'significant other\', a definition excluding friends. Friendship is the single most abused relationship pragmatically. tirade ends here.   just read article on how more people in hongkong die&nbsp;due to lack of exercise than smoking.  does that discount the risks of smoking? no.&nbsp;no mention of smokers who also exercise either.&nbsp;   saw play called \'a louisiana gentleman\' yesterday. actor looked like keanu reaves and actress like  helen hunt. winding plot with&nbsp;rosy end. about new orleans man dealing with&nbsp;snooty aunt, neglected wife and paranoid sister.   bronx zoo tomorrow, if weather permits. I operate despite it.   &nbsp;   &nbsp;  &nbsp;&nbsp;&nbsp;   &nbsp;   &nbsp;  &nbsp;&nbsp;   &nbsp;  &nbsp;&nbsp;  '),(24990,3502783,'21,July,2004','2004-07-21 00:00:00','   So it happened.  I Ran Away floated in out of the blue and messed me up. About time, no? Fleeting&nbsp;visions of walks&nbsp;around grand central in the winter cold, probably&nbsp;headed towards Book Off reminded me what that song meant to me. THAT is what Im going to miss about this place.&nbsp; anyway major amount of work done today - when&nbsp;am I ever going to find the time to pack?? I should just declare I dont have time to work anymore. Am I only one who doesnt find&nbsp;my pending work schedule strange and ill-timed? Breathing space there&nbsp;is none. As nidhisaraf acknowledged the biggest fish who \"..occasionally comes up for air, but is usually underwater..\" We shall push luck&nbsp;for a ray street reunion. Darling she&nbsp;called on sunday&nbsp;as did Simran :)&nbsp; oh speaking of when, Beach was marvellous under clouds, rain and wind. Kafe Shashlychnaya was over-rated&nbsp;we all felt, but I liked my chicken stroganoff with dill. my eastern-european penchant is still on. anyway funny&nbsp;how there was no one at the table from teh same country. this is the life. ok only till asif showed up. people should have dunked me into teh water, a little more insistence would have gone a long way into the waves. and I havent called leo about next meet-up,&nbsp;felt bad we left&nbsp;shortly after he&nbsp;came along. have to make up for&nbsp;that.&nbsp; so i broke my vow and all also. had to crash at laura\'s place, asif and me. I missed the name of the&nbsp;bar yet again. I can now discern between two separate states leading to different outcomes during bingeing episodes.&nbsp;Cool conversations of course. Too bad couldnt show people the roll shop next door. ok pictures have to be uploaded. bday girl was happy, though. Impressed with small talk&nbsp;and breakfast I had to make next morning with cousin dadu-dida at&nbsp;home,&nbsp;while nursing severe hangover. I think discussing sightseeing in new york eased the process.&nbsp;  cant believe how immune I am to wowy. I am an objective listener now hahahahaha laura is an amazingly social creature, quite adore her. hope I get to meet her again before I leave. asif has treat due also. 125th street was abuzz with passers-by yesterday, was so glad I went. Uplifting talking to lady from reserves and the guy from the Guardian angels, with freaky white lenses and vampire teeth.  and&nbsp;aditi\'s getting married in&nbsp;oct,&nbsp;in&nbsp;montana!!!! and she wants me there, cool ;) will make it to bozeman then. My first (semi)american wedding. Closer to claremont than new york, it is. went to bon appetit&nbsp;bakery/cafe Twice today, you bad, bad girl. Addictions do catch up.&nbsp;&nbsp;  will&nbsp;I be able to reach rockefeller center&nbsp;at 7 am on fri, is the question, pondering over which will be pointless.  I&nbsp;badly want to go to the bronx zoo. (waaah!)  have to find presi marksheet now. something tells me I wont find it tonight.  &nbsp;  &nbsp;&nbsp;  &nbsp;  &nbsp;  &nbsp;  '),(24991,3502783,'16,July,2004','2004-07-16 00:00:00','   My journey began in Calcutta, being the first and last-born of my parents. My schooling was at Our Lady Queen of the Missions School and subsequently at La Martiniere for Girls. After an academically uncertain, yet intellectually stimulating undergraduate experience studying Economics at Presidency College, it was evident this was not my calling. As my family moved to New York at this point, I considered an M.A. in Psychology, a subject I was vaguely familiar with, yet infinitely curious about. As I matured enough to decide my consequent path, I was grateful to Presidency College for advancing my growth.&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  &nbsp;&nbsp;&nbsp; After three years of studying various facets of Psychology, supplemented by fieldwork and research, I look forward to beginning a Ph.D. programme in Social Psychology. This encompasses human behaviour and cognition in every aspect of life, and delving deeper into it would be remarkably fulfilling. Research I have been involved in, though not touching upon social psychology directly, has consistently presented questions whose answers require a social perspective. Diverse experiences have convinced me to study this more intensively, which I endeavour doing at Claremont Graduate University (CGU). The practical applications of this programme in the real world, to influence peopleâ€™s lives beyond the experimental laboratory, I wish to utilise back home in India.  &nbsp;&nbsp;&nbsp; Environmental psychology, deftly linking social psychology and our environment intrigues me. Social concerns are effectively addressed when also acknowledging&nbsp; environmental constraints, requiring an understanding of the psychology of the larger community. Society fragmented into cultural groups, as in India, will usually have different environmental concerns. The Indian sub-continent, an amalgamation of cultures, presents conflicting interests and a pressing need to bridge gaps between divergent norms of each. Repercussions of these are felt particularly while broaching environmental matters as well, due to needs from cultural points of view. CGU will provide me with the tools and methodology to address this by also focusing on the common threads between various groups under the national banner. Emphasis on group relationships should not only be on how they manifest, but also on ways to unravel why they are so, for a greater environmental goal.  &nbsp;&nbsp;&nbsp; The appeal of an applied Social Psychology programme to me then, is its potential in India, where the field of psychology is in a relatively fledgling state. Research especially within social psychological contexts to understand and shape public policy in India, I perceive as an eventual focus. In terms of environmental issues, a broader framework of our surroundings and the diverse inhabitants of the community can refine addressal of environmental concerns. This is very important given the wealth of natural resources and a corresponding multitude of cultures in the sub-continent. With ample scope of exploitation of such valuable assets available to a wider community, it becomes crucial to handle such matters of public interest, politically and scientifically. My intention is also to affect these national issues professionally.&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  &nbsp;&nbsp;&nbsp; This is the future I am investing in at Claremont Graduate University with the hope of equipping myself appropriately for this.&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  '),(24992,3502783,'16,July,2004','2004-07-16 00:00:00','   This was the find of today:&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;    urlLink http://www.mihaieminescu.ro/en/literary_work/poems/gloss.htm &nbsp; (trying to figure out link..) &nbsp;  Gloss  by Romanian Romantic poet&nbsp;Mihai Eminescu. A mere translation, but what can one do, apart from learning Romanian, as was acknowleged.&nbsp;&nbsp;&nbsp;  Fellow passenger in train engrossed in&nbsp;it, caught my attention and hence conversation ensued, he will email me more about him, but&nbsp;couldn\'t&nbsp;resist the search.&nbsp;He wanted me to read it during the&nbsp;ride, but was amply distracted. the first line I read which&nbsp;determined the&nbsp;next was \"...Waves that leap like waves must fall..\"  Note the curious style carried over from the first verse onto the subsequent ones. Thoroughly impressed and blissfully happy. I wait.  &nbsp;  Ups and downs of the day did me in.&nbsp;Cracked under pressure a few hours ago, was pending, too.&nbsp;  &nbsp;  \"...The ups and downs of expectations  I can\'t take any more,  They just end up killing me..\"  &nbsp;  another poet,&nbsp;this.  &nbsp;  Loan is driving me up the wall. Endless&nbsp;clarification with a bunch of different people who&nbsp;provide&nbsp;disconnected,&nbsp;confusing&nbsp;information. Customer services quite hopeless. I will warn people against approaching this agency, for their sanity\'s sake. Spent a fruitless week chasing a lost cause. Now start all over again, wiht trip to New Jersey. and as ever weekend\'s &nbsp;in the way. Have to forfeit&nbsp;the Siren festival because&nbsp;of last minute trip.  Cina, a Romanian place in Astoria will ease the agony, slightly.  &nbsp;  But let\'s not forget, have to now&nbsp;Watch the movie Zorba the Greek.&nbsp;have just started&nbsp;to read, but waiter at fancy cafe Chez le Chef tipped me off. Pleasant, very pleasant.&nbsp;Probably the South Asian connection.&nbsp;Too bad won\'t be revisiting. Turned out to be my respite after 30 blocks of walking down Lexington. Chef Frederic looks like Santa Claus in white.  &nbsp;  Culinary explorations continue as&nbsp;this new face&nbsp;on MacDougal, by the name&nbsp;of Press Toast&nbsp;caught me totally unawares and filled me up to the brim.&nbsp;Bread toasted flat&nbsp;with mashed avocado...will never ever underestimate size again.  &nbsp;  Oh and then&nbsp;the other&nbsp;addition to list of complications.&nbsp;Eye doctor found unexplained inflammations in my&nbsp;eye and&nbsp;thus wants&nbsp;full check up done. So am now punctured all over&nbsp;after blood tests and other&nbsp;needless exams. Doctors more&nbsp;bothered than me.&nbsp;From the point of view of insurance, quite understandable, but I wouldnt&nbsp;count on&nbsp;him for moral and/or emotional&nbsp;support during an actual crisis.  Doctors&nbsp;greet you with the likes of \"So you\'re very worried about&nbsp;the inflammations..\".  &nbsp;  Anyway, so am apparently free of tubercular infection, syphilis, neurological diseases and some other such nonsense. Blood test results should add more to list. Chest x-ray was uneventful except me&nbsp;making a futile&nbsp;attempt to convince&nbsp;fellow-in-charge that&nbsp;exposure cannot harm a&nbsp;non-existent fetus.  The nuances&nbsp;of the medical system, I wouldnt have escaped from, for too much longer, anyway.  Not done yet, though.  &nbsp;  Want Sunday to bring in fair weather.  &nbsp;  &nbsp;  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  &nbsp;&nbsp;&nbsp;&nbsp;  &nbsp;  &nbsp;  &nbsp;  &nbsp;  &nbsp;  &nbsp;  &nbsp;&nbsp;&nbsp;&nbsp;  &nbsp;  &nbsp;  &nbsp;  &nbsp;  &nbsp;&nbsp;  &nbsp;  &nbsp;  &nbsp;  &nbsp;  &nbsp;  &nbsp;  &nbsp;&nbsp;&nbsp;  &nbsp;  '),(24993,3502783,'10,July,2004','2004-07-10 00:00:00','   The welcoming openness Of limitless possibilities  Played back and forth by gentle tides Catalysed by soothing crispy winds Sweeping you off your toes Tiredness of feet alone While alive the reincarnated soul Elusive what is touched by the sea  Just felt within, so out of human reach  So springs forth lost trails To comfortvilles and cosy inns The travels etched  On sand concrete  A full circle turns to bring back home Thoughts of roots of all the passing tales And what paved the eventful, glorious way Sand beneath the sole sinks it in Transient all, from head to toe Waves dying to caress the feet Remind of what sustains, preserves All thatâ€™s been a staple source Of a hazy glimpse of earthly heaven Seaweed, bright green or dirty brown  An exploration of a touch of new  As fingers slide and snake surfaces Refreshed with ocean water cool     From tugs and calls beyond end of land Drifting pieces of broken wood beckon Frayed and withered after sojourns long Gulls playful, of the patient sort Who know sea ways and the open life Of what it is to be free-spirit-kind  Unhindered by none  But gales of time Flight to be prioritised Sands, winds and waves  Hail a boisterous encore.  '),(24994,3502783,'08,July,2004','2004-07-08 00:00:00','   Stumbled upon reason why Ma is so averse to my Calcutta trip. She doesnt want to give up on time with me around. After a long conversation over the wires, in the middle of the workday (both hers and mine) this was explicitly made evident to me. Great job, you budding psychologist, you. First of all, as I pointed out, since she\'s so convinced it\'s not practically possible, there\'s no reason for her to get so worked up with the very thought. Of course, the consequent effect on me is not yet clear to her.   may be visiting Claremont as early as next Tues. Subject to response from Angela Burwick and availability of affordable tkt. Apartment hunting, the intention. Checked out some plcs, glad Michael sent list of pet-friendly apts also. good conversation, that. if it hadnt been for the last-minute-stigma in my life, could have coordinated with him about transporting stuff across.  There were more options and I dont mind other people\'s animals. oh and some of them have pool also :)  Tripped around downtown later in the evening. Gushing fountain at washington sqr a far cry from its waterless state a year ago. Different times, these. Ticked off Polka Dot from list. Not impressed. Music alone has no weightage. Coffee, not unusually memorable (explaining my very late early hours tonight) and seating arrangement lacking in creativity.  Looked for cheap Set List cd. will push luck further in east village tomorrow. The Frames have to be obtained. Went to same cool store on bleecker, where I got b&w u2 poster from their earlier heydays. if I do go to Calcutta, am getting it framed :)  Sushila has probably decided not to reply to my email that explained total uncertainty of trip. Am short of travel agent too, then.  Work stares me in the face teh whole day tomorrow, let\'s try and allot time priority-wise. Will meet Asif for dinner tomorrow, happy things are working out for us. lived through days of unbearable (and in a way unfelt) anxiety the last few months. no breathing space in sight yet, though :)) Cant wait to come up for some air.    Been asked out on car-trip to wherever, next week. Adding to complications.  I wait, for next weekend.  '),(24995,3502783,'08,July,2004','2004-07-08 00:00:00','   Work can be fun. Woke up to pending reports, only to get cracking on them. Which ensured conversation with Saurav online, which will make me rethink aspects I had taken for granted. Like I mentioned, it was trippy and has to be continued. Got a nasty shock about new I-20 rules, which could potentially have had me pay an extra hundred bucks after Aug. They spring new regulations on you without batting an eyelid. such fuckers. thankfully turned out not applicable to transfer students like myself, so spared, but there are others who won\'t be. hated parting with walking tour guidebook of lower east side. But what to do, am heaped in library fines due since whenever. why is this an inevitable fate??? Ok people are asking me how Im doing with \'idea\' of parting with NY.  Nothing has registered regarding which as yet, that is great. Attachments are easily displaced hahahahahahaha Also, it\'s just a City. She laughs again. And then amidst grave doubts about it ever materialising, Sting concert was a major event. I cant even remember when I got tkts. Excursion to NJ quite worth it, with most everything spelling doom, namely possibilities of me misreading date/time, getting into wrong train, missing train, etc etc  Almost super organising flop, but last minute ruled, so smoothly it went. Annie Lennox opening paved the way, she is a cool cat and love her as well. WHY nagged the life out of me till the opening strains were heard.  Sting took over shortly. Man is aging noticeably. Voice still intact though. feruze and me both cried during Fields of Gold. Everytime. Im a sucker for music. Didnt want to leave stage also, twice he came back...that\'s my kind of performer ;)  Dreamt of Dimma last night, in Selimpore. Typical. Someone was giving her blue flowers about which she was complaining. Again, predictable. Im told this is not something I should be seeing. Given a choice of dreams... anyway, this makes more urgent trip to Calcutta, which is up in the clouds. then Jenny left voicemail complaining(?) I hadnt sent her poem. what is happening, I wonder.  head spinning now.  legal alien  '),(24996,3502783,'03,July,2004','2004-07-03 00:00:00','   urlLink    them&nbsp; urlLink  '),(24997,3502783,'03,July,2004','2004-07-03 00:00:00','   Closure happens never at all However convincing may be Thwarting just those very thoughts That were banished carefully Not after violin strains  Play strands handpicked  From a charmed life Of words echoing the very words You didn\'t need yourself to hear Faces of old glistening in the light Tears, an inevitable plight As revelations mark this while And pound and bruise inside Endlessly raining on Reminds of memories  Shunned, misunderstood Or fathomed not at all. Dark, lonesome lakes And gardens green Masked by shadows of darkened times Into muted grey they turn Of nooks unused Undisturbed Peaceful calm, they still exude. Tireless walks gently tugging along Images unmoved, preserved. The first perhaps will be the last The first will revive the past Emerging at the end of last. What conjured into demons scoffed Now warped back to merry sights. Music overwhelms and flows Into veins, replacing life-blood Pumps and charges and devours Every inch of a naÃ¯ve soul Whispers of heydays, bombard And puncture, at first and every strike While the sounds trickle in. Frames flash by at battering speed As the music rocks on   Frames pictures Suddenly restored, renewed Closure is a mere myth.  '),(24998,3502783,'03,July,2004','2004-07-03 00:00:00','   A week\'s worth summarised here? not quite likely, as last night still overshadows others. So, found out about yet another Irish band, The Frames, (www.theframes.ie) and they performed at South St Seaport as ever. Got massively killed. Cried through seventyfive percent of concert time, and now have to get cds. they were sold out.  Dont get me wrong, theyre a rock band, a real rocking one, but there\'s also a violinist, by god. Do I neeed anything more...apparently not.   then, Soma\'s in NY ??!!!! shocked out of my senses, words.. so going to meet her today. Have to figure out how to address fact that her mail from feb, I believe, is still unanswered.    Then talk on Wed went off rather well. Was proud of me. Wrap-up wasn\'t a breeze, but was handled.   Still harbouring thoughts of Calcutta materialising. Will have to keep on trying luck. time is the freaking constraint. let\'s try to get housing taken care of this month and fees semi-fixed. do i visit Claremont now, as well? Should ask them whether it is worth the effort. Also need to get moving with driving license. Miscellaneous committments as far as get-togethers and trips right here. nothing more than a reminder for me, this.  there hasnt been much space or time to reflect on life after new york. No complaints there, though :)) When it rains, it pours... or words to such effect.  What makes a song great?  its ability to tear you apart and force you to connect with not-so salient aspects of your life, for one. Especially those, you didnt quite think existed.   also if it leads to further outbursts of creativity.  ladies and gentlemen, The Frames.  '),(24999,3502783,'06,August,2004','2004-08-06 00:00:00','   you\'re not a new yorker if you don\'t run along the crosswalk after the last minute, if you couldn\'t not care less of the waiting vehicle, if you don\'t have the patience or the time to wait for the small, white man, if the red hand serves you better. you\'re not a new yorker if you aren\'t always in a rush, if your destination isn\'t more urgent than anyone else\'s, if the world outside your head doesn\'t flash by you in streaks of colour indiscernible. you\'re not a new yorker if you bat an eyelid at anything out of the mundane, hell nothing deserves such bewilderment. you\'re not a new yorker if your threshold of tolerance isn\'t constantly on the dip, reaching levels unfelt or otherwise. if this is reinforcing that snooty stereotype then so be it, cuz its true. not all of us are bad though; apart from the several wonderfully interesting strangers Ive encountered, am a living example of this fact. Survivors, shall we say, we are, whatever and whoever the enemy. So it is that time of the story again, when a large chunk of reality bashes me in the face. Life after, will be a toughie and could hinge on rather detrimental consequences. dont know what brought this about: Conv with delphine, whose book I have to return, perhaps. I dont know what Im going to do without the ISO, HAVE to join the LA branch eventually. being out on teh streets does this to me, just the potential kills me. will feel helpless without it. let me see how I\'ll manage this. Tuesday evening then, will be very hectic. Stranger on 125th and lex said he was getting distracted by the whitehead on my lip which I should pop, well not exactly stranger, was arguing with him for a while about the intent of the ISO\'s endorsement of the nader campaign. lost patience at end. told him not to. mcdougal street, the second option: one of teh best lattes at cafe esperanto.. will trip out with jenny tomorrow. will forgo street theatre to get work done at home, so I feel less guilty about prioritising pleasure over responsibilities. Or group in subway crooning to tune of \'land down under\' by men at work. good version with guitar, tambourine and vocals only. maybe the awesome weather as well, I love winds that mess up my hair far more professionally than I can ever do. By the time I reached duane reade, I was already severely tuned to my love affair with this city. hence it was routine to help fellow customer choose glasses for his wife/gf/partner/friend/relative. and why all of this drops in while Im sitting in the 7 train, I will never figure out. of course the chief culprit is \'let go\'. so fine the theme is creepily similar to \'i ran away\'. I cant help it if a song takes complete control of me. Im a receptor of unfathomable sensitivity. am also the most self-involved person I know. not of the selfish sort, but so freakingly aware of ins and outs of my self. this is the second song which has blown me to bits at first sound. am buying dvd of garden state. when will I see it again, I wonder. I should either not be left alone to my devices, or should not be allowed to listen to music or should have \'let go\' be drained out of me. all impossible hahahahahahaha   so what should we do   THEN am about to forfeit 10 precious days of delving deeper into the city, thanks to calcutta trip. let\'s not remind ourselves of that trade-off, shall we. so essentially have approximately 10 days left in new york. so claremont has area called The village...some consolation, that. Some idiot also tried convincing me I\'d \"hate\" LA after new york. what a jerk. good he featured for about 10 minutes of my life. denial sounds like a good proposition right about now. back to it as soon as I can. after thwarting people to think of future times, tsk tsk. cant wait to met jenny. what happened with us, I dont know. oh god... have to call smitesh also. slowly the list unfurls. kati rolls happen next week, hope asif drinks as many as he can over weekend, to me. I will too. Im done drinking tonight, have work also. Entertainment in evening took form of people at home discussing \'castes\', especially in teh bong context. I wonder if I can pinpoint the line between the jokes and the seriousness. it quite shocks me at times. I was gleeful tonight because it was in reference to ranjini\'s multiple boyfriends. apparently dibakar still features, apart from strangenamed people like namit and some person who gives her rides back home. hear me rolling in laughter. I did speak to her though, and shut her up as she started with her trite concerns. wanted to know of gory details of chicago, at which point I impressively cut her short. Didn\'t get annoyed also. Score, pat on back, etc! oh bought laptop also, should be here by next friday. one less thing to worry about. How many people will I not be able to meet before I leave. One last thing, just wanted to acknowledge the historical date that is today. 6th august: going back to 1945, saw what in my books is the most heinous crime ever committed by man. in Hiroshima. have always been tremendously moved by the attrocity, and just like to take some time off, every year, to think of the fates of millions of innocent civilians in the hands of imperialism. oooh and 11th august is around the corner, have to celebrate my third and final year here. Crammed but fruitful week in sight.  '),(25000,3502783,'04,August,2004','2004-08-04 00:00:00','   Had to write tonight. Am leaving work and other more pressing activities to get this done. House is full of aunts and hence might have to play hostess from time to time. Dislike the societal demand to be one. Anyway donâ€™t digress from task at hand. Time is not on your side. Oh got ticket also, and have tucked it safely into drawer, and hopefully will not get lost in the stacks of paper inside. Should make everyone see where it is stashed. End of the mundane. Saw Garden State. Totally bombed and strangely subdued consequently. Freakishly impressed with Zach Braff: not One trace of J.D. from Sacred Hearts Memorial in Andrew Largeman returning to unanticipated reunion with friends old and new in NJ. Respect him madly now. Dude wrote  and  directed it. Thatâ€™s one aspect of the impact. Brought back memories of deserts and isolation from all and sundry. Caravan trips into the unknown without a shred of fear or concern. All unreal. If that is where ultimate happiness lies, I am infinitely screwed. Soundtrackâ€¦hmmm still donâ€™t know the track name, which got me hooked in the first place. Will seek out. Co-dependency is scary, but also such a refuge. Almost a haven which isnâ€™t too bad a deal, come to think of it. Something to strive for, perhaps. I donâ€™t like that Iâ€™m so moved by this. Really??? Seriously, this was a comedy. Just found Zach Braff blog. Cool. And sought after song is Let Go by Frou Frou. Will no doubt wake up tomorrow with this in head. Am not done with this theme as yet. Too distracted tonight. Botched up priorities are in any case getting to me.   So what should we do?  '),(25001,3494693,'31,May,2004','2004-05-31 00:00:00','    Manden siger nok sÃ¥ flot at han har en overrraskelse, og det lyder som det er lÃ¦kker en... Men hmm hvorn kan folk vide at den person de vil overraske bliver glad? Er det ik bare som et lille plaster pÃ¥ sÃ¥ret, fordi man tilfÃ¦ldigvis har vÃ¦ret en rÃ¸v, sq vi sige i smÃ¥ 6 mdr. Godt sÃ¥! Well bout last night, went out with the fabulous 3, even though there is only 2 left... Was supposed to do history with Ãsi, we hurried through that and had a nice barbecue and drank some wine instead. Paw joined us. Well after the barbe joey came over and we had a few drinks,and therefore decided to go out :) This was of course after joey was properly down on his knees... to put a ring on the finger of course! ;) yesh yesh properly engaged, vegas here we come!  Well town was fun as usual, drinks, dancing, and a wicked game of DIE, right Ã si!?! (for those who are lost, one dice two die) :P met some weird ppl @ cosy... bit creepy when u are too sober, must keep in mind- note to self!  Evening ended with joey dragging me home waaaaay too late, and when got of bus, realized that keys to house was in car, which was neatly parked @ Ã¡si\'s... so after a little detour i came home without falling over too much... all in all a successful day...  Nu skal der bare gÃ¸res rent, er alene hjemme Dee er ude og hente lunse i lufthavnen... yesh der trÃ¦nger vist os til det...  Nogen der kommer og hjÃ¦lper?   '),(25002,3494693,'30,May,2004','2004-05-30 00:00:00','    Hey!  This is really odd, am just trying this!! Joey, my future husband, look what ya made me do!! If this is the basis of our marriage... HMMM  more to come when i\'ve figured this out!   '),(25003,3494693,'04,June,2004','2004-06-04 00:00:00','    Wedn Night.  Well were sitting quietly with Dee watching BS tv when a panicing Mie called. \"Are you alright?!?!?\" Yes I lived even though I live in the Bronx and 2 ppl just got shot 2min from my flat. Damn had a minor panick attack myself, whats wrong with the world?? New B tried comforting making fun of me, think it was the whole idea that it was my turn next... The two bastards who did it is still at lose- Yes, guess where they dumeped their get-away-car?!?! yep, thats right next to my street. How totally fucked is that?!?! Well still no prospect of actually meeting New B yet, he\'s still busy with his store... When I finally got to bed I couldnt sleep... Note to self- you live on 4th floor, what are the odds someone will come in your sleep, rape, murder and what-not to ya?? - Stop worrying FFS!     Thursday  Celebrations have set in... yey at me for getting old :) Had a group meeting in the morning, was of course v. late... Quite an accomplishment turning of ur alarm in ur sleep so it doesnt go off :( After group meeting Yndi Guggu and I went for a bagel and laid in a nice parc tanning, and discussing all sorts of things... fabulous sunshine-- And am actually getting brownish ;) Talked to Christian when got home and decided that we should have a little date, so he rushed over round 8pm so we could watch Paradise Hotel.. or something! Well turned out to be a loooooong nice evening... you get my drift ;) Drank some wine, ended doing v. crazy things... (trust me its not to know...) But will tell that I helle 4yrs old dared him to do push-ups... 14 way to go sweets- Taking into consideration you had been using your stamina for other more important things ;) Of course I couldn\'t be a sissy, so had to prove that was nothing, ended up doing... hmmm too many! Well after 5-6 hours of fun it was time to sleep... and Christian decided to be a gentleman and move my wine-covered table... yesh i managed to knock over a bottle of wine, 2 FULL wine glasses and god knows what... my carpet is still red.. smooooth as usual, i am!! :) ...Anywho, so while he moves this table, he sorta loses his balance, and the show he put on there was the funniest thing I ever seen! He steps on bed, almost falls over, so step up on couch, turn around still no balance at all and table with stuff on his hands, manage to finally get down on floor and put table down. Take a bow sweetie, nothing fell off table and he didnt fall over...  Spooned and snuggled all night- except for when i woke up from PAIN in my arms... grrr never doing push-ups again--- 4 asprins and 10 hous later, it still hurts!!!! Damn it!   How stupid can u be? And who said exercise was     good for you!!   '),(25004,3494693,'02,June,2004','2004-06-02 00:00:00','    First of all- Sry Joe! We all have got mobiles, mine is clinincly attached to me. v. fine in situations when you actually need it, or when you have not substituted all reason with a taaaad bit of beer...   So making a long story a bit shorter... The day started with a trip to RUC... evaluation of the semester. Yes we do actually work at out there once in a while, picked up our pretty, wonderful, and absolutely fab project! Well done girlies!! Think we managed to twist our way out of doing the BS history notes... dan charlie, dan charlie what _were_ you thinking?!?  The previous-mentioned surprise, had absolute dickely duckely nothing to do with me, but congratz on the new toy Mus! Be careful and dont break ur neck ;)Guess i might get a sneak peak today...  Talked with sweet Chris, it was nice he\'s on his way to becoming famous, cant wait to see the film ;) Nor can i wait until he move his pretty little behind over here!   Had peachy dinner with Dee, and then got ready to meet Joey-- So got to studenterhuset and had several beers... Joey and I were talking bout embarrasing things... And good ol\' stories! Went on to cosy... And by then I had started SMS\'ing \"things\" to B.. (the potential new)... ohhh dear, somebody plz shoot me! Should really not say things like that to someone who doesnt know you... imagine the worst u can say x5 then you have what I told him ;/ Well it gets worse... have never talked to this guy, phone acts up... so i decide through my beer goggles that the most rational would be to call him. 2am, am in the middle of a noisy gay bar, and cant even get two words straight (okay not to get things straight in a gaybar) but this was calling New B. Well the first words are kinda like: ssshellow shhhow are uuuuuu doing... where to he replies: you are drunk!   Well partied on til Joey v. casually draaagged me home... ;) thanx hubby u r a doll :D Have woken up feeling semi quasi okay, have group meeting in the city...  Walk out in a g-string thing making coffee... standing there in my own pretty little world... ass on full display... Old friend Rasmus from Gymnasium walks out off dee\'s room, and greets morning... how to begin the day :/ Top dollar class!  Does anyone have an asprin?   '),(25005,3494693,'11,August,2004','2004-08-11 00:00:00','    At finde den perfekte gave til en fyr... Det var sÃ¥ min opgave i gÃ¥r! Fisketorvet blev indtaget af Mie og jeg... Og over en frokost lagde vi sÃ¥ en kampplan! Det skulle vÃ¦re en gave der indholdt alle af fÃ¸lgende elementer: Nyttig, betÃ¦nksom, god, ikke for stor, skulle os gerne signalere jeg er sÃ¸d, betÃ¦nksom, vÃ¦lg mig og by the way tak for hjÃ¦lpen!      Hvad skal man sÃ¥ kÃ¸be??? Det er selvfÃ¸lgelig meget at en ting skal kunne! Lad os bare sige jeg var determined pÃ¥ at finde det perfekte... og tror mÃ¥ske det lykkedes... men hey... nu fÃ¥r vi at se... nÃ¥ede i hvert fald at pisse en ekspedient totalt meget af, fordi han ik fandt noget der passede mig.... han vidste selvfÃ¸lgelig heller ik det pres, der var pÃ¥ denne gave! Men mere om det nÃ¥r gaven er blevet afleveret!      Skal pÃ¥ date med T i morgen, det er nr 4, men hvem tÃ¦ller! Mie og jeg hyggede ellers rigtig vi spiste middag, sÃ¥ mona lisa smile... snakkede og fik lige hele verdenssituationen drejet...      Kom fÃ¸rst i seng lÃ¦ 4 i nat, da jeg vist fik siddet og Ã¦vlet lidt for meget med folk pÃ¥ pc... og havde jeg vist at dee ville lege handywoman kl 10 med en ond hammer havde jeg sq nok gÃ¥et i seng lidt fÃ¸r!      I dag stÃ¥r den sÃ¥ pÃ¥ oprydning... damn damn damn! Nogen der gir en hÃ¥nd med?!?   '),(25006,3494693,'09,August,2004','2004-08-09 00:00:00','    - SÃ¥ er sommeren nÃ¦sten ovre, og thank you God for det!! Denne sommer har vÃ¦ret komplet ubeskrivelig! \'Summer from Hell\' i daglig tale-- Alt hvad der kunne gÃ¥ galt er gÃ¥et galt, ja Murphy du har sgu made your point - Crystal Clear!      Men der har da ogsÃ¥ vÃ¦ret positive ting, selvfÃ¸lgelig har der det.... Ã¸hhhh *tÃ¦nke* hmmm tjooo... ja har fundet en bikini ;) Ja, det lyder ik som meget action, men gennegÃ¥r \'find en bikini helvede hver sommer\' og hvorfor skulle 2004 vÃ¦re anderledes- SÃ¥ frisk, fro og optimistisk drog jeg ind til den indre by... Det er jo trods alt  bare  en bikini... Det kan jo ik vÃ¦re sÃ¥ svÃ¦rt finde, eller.... Men 2004 var anderledes, denne gang fandt jeg en, og nÃ¦Ã¦Ã¦Ã¦ hvor er den fin! Har ogsÃ¥ allerede vÃ¦ret ude og lufte den pÃ¥ kirkegÃ¥rden- ja har min egen grÃ¸nne plet der!      Ellers ser jeg frem til det \'nye\' Ã¥r! Ja, for det er jo sÃ¥dan nÃ¥r man er studerende starter der et nyt Ã¥r her om sommeren selvom kalenderen ik siger det. Men tÃ¦nk jer det gÃ¸r den jo sÃ¥ alligevel, vi har jo vores egen som begynder og slutter i juli! :) 2. Ã¥rs studerende, har allerede fÃ¥et kriller i maven ligesom den allerfÃ¸rste skoledag, ever! Der stod man sÃ¥ 6 Ã¥r, fin ny plyser-rÃ¸d bluse med bamser pÃ¥, sort lille rygsÃ¦k med en koala-bjÃ¸rn pÃ¥. Ja jeg havde en ting for plyssede ting, Ã¸hh har det vist stadig!:) shh! og nu var det bare sÃ¥ spÃ¦ndende havde klatret op pÃ¥ plankevÃ¦rket en million gange for at se pÃ¥ de store bÃ¸rn, der var sÃ¥ heldige at gÃ¥ henne pÃ¥ skolen! (skolen lÃ¥ jo lige henne pÃ¥ den anden side af gaden) Og den varme velkommende fornemmelse man fik da man kom ind i ens fÃ¸rste klasselokale, der var smÃ¥ runde borde, et kÃ¸kken, hyggerum og lege-area. Det var en hel ny verden, alfabetet var hÃ¦ngt op over tavlen sammen med tallene- Sikke en spÃ¦ndende dag! Det minder unÃ¦gtelig meget om RUC, glÃ¦der mig til at komme tilbage til vores hus... kÃ¸kkenet, \'hyggerum\' aka grupperum, der er os tavler med info osv... se bÃ¸rnehaveklassen om igen! Ja, til tider virker det ogsÃ¥ som folk stadig skulle starte i bÃ¸rnehaveklassen!    I gÃ¥r fik jeg sÃ¥ en besked fra min onkel, love him to pieces, tÃ¦nker pÃ¥ alt det han har gennemgÃ¥et denne sommer... og pludselig ligner min sommer vand i forhold til! Han sagde at han havde tilbragt dagen med sit livs kÃ¦rlighed, og sÃ¥ var det at det slog mig, at det er ogsÃ¥ det jeg vil. Efter 20 Ã¥rs Ã¦gteskab, vil jeg ogsÃ¥ kunne sidde og nyde en sommerdag og stadig vÃ¦re forelsket! Hvem ved mÃ¥ske dukker prinsen op fÃ¸r eller siden... han gemmer sig bare godt!!! Kom frit frem, jeg bidder ikke---- hÃ¥rdt :P    SÃ¥ det her er for en ny start, nyt Ã¥r, nye venner, nye hobbyer, nyt arbejde, ny kÃ¦reste... ja der er lidt at arbejde pÃ¥, men selvfÃ¸lgelig er nogle af de omrÃ¥der ved at falde pÃ¥ plads! :)    Skulle man starte med sit vasketÃ¸j?!?   '),(25007,2238828,'26,May,2002','2002-05-26 00:00:00','    if i say my heart is sore / it sounds like a cheap metaphor / so i won\'t repeat it no more  - poem to a horse || shakira  it\'s not going to be this hard again, is it? it\'s everything and nothing, once more, yet another time, ad nauseam. everything in my life comes in one tiring endless loop, ups and downs swooping on me without warning, and i wish it wasn\'t like that. wish i didn\'t have to pay for my good days with bad days like this, when smoking a cigarette makes me want to singe myself.  it\'s not just a wrong kind of love, it\'s everything. last night my sister told me that she was going to run away.  i mean it,  she said, her words grating hard against her pillow as we spoke in soft voices in our dark room.  it\'s that or die of depression, and i don\'t want to die.    sometimes you get hope from the weirdest sources. tomorrow i\'m just going to wake up again and in two days, maybe three, i\'ll be happy and i\'ll feel like i\'m worth something again.   maybe i\'m breaking up with him  for his own good , and maybe it\'s a good idea that i don\'t fall in love for a long time. the way i am, i\'m afraid i\'m just going to eat someone up, swallow them down and suck them so dry that even their bones won\'t remember what moisture was like. bones. everything is bones and dust, ashes and black holes, infertile and nonproductive. it doesn\'t help to have this doubting in me that says  you can\'t have children, you\'d only hurt them,  and although i never thought of myself as a mother, there\'s a sort of ripping feeling inside, a hollowing out with a blunt object that comes from imagining a life without ever knowing what it\'s like to be pregnant, to carry a baby inside of me.  tomorrow i\'ll remember i\'m still only twenty-one and i still have time, there\'s still so much time.  right now i feel as if i have nothing left, and no time left, and no heart to struggle for the last remnants of what could have been passion.   '),(25008,2238828,'22,May,2002','2002-05-22 00:00:00','    i\'ve been flipping through my old entries. my, what a depressive child i am. a couple of weeks ago, a business colleague of my parents came to KL, and we were taking him out everywhere. he was much too friendly, in the way overly eager businessmen often are, and in an effort to win me over, he spent most of his time psychoanalyzing me.    you don\'t smile much, do you?  he finally concluded, after three days of apparent failure.  i mean, i\'ve never seen you smile voluntarily.     i smile,  i said.  i just don\'t feel like it,  is what i was thinking and didn\'t say. see, i\'ve learned restraint. i\'m not such a brat after all.  two hours after that, we went out to dinner and brought my sister with us, and he kept on staring in our direction.  you\'re different when you\'re with your sister,  he commented, totally non-apropos to whatever they were discussing at the moment.  very animated.   i came very close to saying  no shit, sherlock,  managed to hold it back, but didn\'t quite manage to restrain the weird look i gave him. i don\'t quite understand this strange obsession people have with smiling. a few weeks ago, E was playing around with a camera and was trying to coax me to smile so he could take a picture. i, on the other hand, was attempting to hide behind the television because the draperies were useless - they were semi-transparent.  smile,  he insisted.   i can\'t,  i retorted, giving up on hiding behind the television and trying to burrow beneath the bedcovers instead.  i feel like an idiot. what the hell am i smiling for?    think of something happy,  he said.  a comedy. your love life. chocolate.  no effect; i gave him my best glare from where i was shielded, behind the monitor of his laptop.  fine,  he conceded.  think of throwing me out the window from the ninth floor.  i cracked up at that one, and then subsequently cracked up again at the look on his face. apparently, it had never occurred to him that violent thoughts would make me happy.  okay,  he grumbled after snapping the picture,  it wasn\'t THAT funny.  i just howled.   my friends are weird.   '),(25009,2238828,'21,May,2002','2002-05-21 00:00:00','    strange, the things i find inside my head, when i crawl inside on a weak day. a pair of gruesome handcuffs, delicately fashioned to carefully rip away at the tender flesh of your wrists when you try to free yourself from them. a litter of pictures, all of little girls who never smile and whose eyes are dark and angry and horribly sad. butterflies pinned to cardboard. a boy\'s bb gun. half-burnt papers and books, violated diaries. and i can never figure out how they got there.  i remember notblood, the coldness of steel against my hands, the way i squirmed against marble to get rid of the heat burning beneath my skin like fever. these are things that never happened, yet have residency in my memory. it\'s as terrifying as multiple personality disorder - the feeling that you are betraying yourself by shifting personalities, by creating memories and lives that you are not wholly aware of. i remember glass, and the sick slick slide of it, the way you would feel a slight stinging only to look down and realize that you were being drained, that blood was staining your bedroom carpet.  things that never happened, and yet i know them. my imagination is way too vivid, most days.   '),(25010,2238828,'18,May,2002','2002-05-18 00:00:00','    just came home from watching  star wars 2  ... i hate to say i\'m disappointed, because i waited TWO YEARS for this, but i am. kind of. yoda kicked ass for the first time. maybe i\'m just hopelessly visual, but i needed to see that because all these years, he\'s been nothing but the annoying green wrinkly thing that talks backwards. today he became more - the annoying green wrinkly thing that talks backwards and KICKS ASS. oh yes.  i am sublimely disappointed in the anakin-padme relationship. first off: why do they always pick such horrible actors for the anakin character? do they not understand that this character will become DARTH VADER some day? they are NOT playing around with a minor character! in the last three episodes, amidala is just a vague memory in leia\'s head; obi-wan is a phantom; yoda is masquerading as the swamp thing; and everybody else is just history. it\'s anakin who develops, anakin who is the axis around which both stories revolve. without him, there is no saga. and yet they keep on casting these very very bad actors to play him. i\'m almost afraid to see who they\'re going to cast as anakin in the next episode, if they\'re going to recast him again. i\'m not even going to speculate, i\'m so afraid.  secondly, this anakin will not become darth vader - not the way he\'s played, not the way he\'s written. darth vader had charisma, a cool insidious charm that wrapped around you and seduced you into the dark side. i can still remember him talking to luke: just the way he said his name, with command and an underlying certainty that his son would obey - this is a man who has incredible power, a genius that has taken matters into his own hands because he can no longer bear to see the way other people fumble around with things that take him seconds to master. he\'s our worst nightmare - the hyperintelligent, all-powerful man who has broken free from the social contract to establish his own code, simply because he can - and because he knows others will follow. this anakin - no one will follow him.   i\'m disappointed in padme, as well - i thought she had more passion than to be swayed by sweet words. she gave in too soon, and she should not have fallen in love. i don\'t understand where it went. she was so kick-ass in ep 1 - i felt for her, a fourteen-year-old queen with her shoulders bowed under the weight of her world. in the opening scene, that compassion and sense of weighty responsibility was still there, in the repressed grief in her voice as she called out her dying handmaiden\'s name. she was pissed as hell at being sent back to naboo before she had been able to do her job ... then the queen sends her to her country home, and suddenly all she can think about is rolling in the grass with anakin. it doesn\'t make sense, especially when anakin is about as charming as a piece of driftwood. even anakin\'s love felt forced, as if it were there only because the story required it.   they shouldn\'t have been in love. they didn\'t need to be.   i never thought that luke and leia had been conceived in love, even when i watched the movies as a child; i always thought that even their conception had been a struggle for power, a challenge given and a challenge taken up. i saw padme being sucked in against her will into anakin\'s quiet, dark magnetism; i saw anakin withdrawn and sullen, frustrated with a system that was too slow and inadequate for his talent and skill.   instead i got a two-hour bodice ripper and a lot of comedy. which was nice, but not what i expected. i want my saga back, my sci-fi world based on the skewed principles of plato and a hefty dose of lucas!imagination. i want that wonderful thing that kept me obsessed in my childhood, that fed my fantasies for years. it\'s lost its magic, and that, my friends, is never fun.   '),(25011,2238828,'12,May,2002','2002-05-12 00:00:00','    the nice thing about mood swings - i mean, if you really, really couldn\'t avoid them and had to live with them for some reason - is that no matter how low you sink, you know that if you just ride it out, you\'ll find a reason to hope again. what was that verse -  sorrow comes in the night, but joy comes in the morning.    and if i just hold on, if i can just wait out this period of depression, then i can return to feeling better in a while. if i don\'t weaken and touch that bottle of pills, or that broken shard of glass, then i\'ll be fine.   i\'ll be fine.   '),(25012,2238828,'08,May,2002','2002-05-08 00:00:00','    falling in love is so hard on the knees ...  - falling in love ... || aerosmith   indeed. and on the elbows, and on the hands, and on the thighs and wrists, and on all the tight strings that bind your body together, and on the loose soft parts that hide tenderly behind your bones. it stings and burns and pulls on your opposite ends until you can feel the give and snap, feel the pieces of you coming apart along seams that you never thought you had.   there\'s a hollow ache in my stomach now, and that could be one of those seams.  you lost weight,  my mother said, pleased, when i came out of my room dressed in a tight shirt and a clingy skirt - the first time i\'ve worn tight clothing in such a long time. i feel lighter, not quite so lumberingly weighted down as i used to feel, but not happier. i could weigh a hundred pounds and my shoulders would still feel bowed. my knees would still buckle.    it\'s okay to feel that way,  V told me once, as we looked out over the city glitter from the top of ayala hill. i just shook my head and tried not to cry, because no matter how close we are, somehow it still doesn\'t seem right to break down like that, over something as trivial as a soured love affair. technically it\'s all my fault anyway, for hanging on to something that i should have lopped off at the knees at the soonest possible time, but instead i hoarded it to cut myself with, like the shards of glass in my room when i was thirteen.   and i\'m just so tired.   i don\'t want to wait anymore, and i don\'t want to look forward to this again. maybe i\'ll feel differently about all this a few years from now, or tomorrow, after i\'ve had a good night\'s sleep - but for now, it hurts   and it hurts   and   it   hurts.   '),(25013,2238828,'03,May,2002','2002-05-03 00:00:00','    today was the day that took me by the heels, turned me upside down, and shook me until my brains rattled in my head like misshapen marbles. right now i\'m in a quiet hiatus from the bustle, allowing myself to take the lunch break we didn\'t have the time to take earlier. we did go out for lunch, but only to check out the hotel at which the philippine business delegation will be meeting their malaysian counterparts - so technically we were working then, too. it certainly felt like it - we walked all the way to the hotel and back, then went to a stationery store and bought various office supplies before FINALLY buying our food, which we had wrapped to go. we got back to work as soon as we closed the office door behind us, setting our lunch aside for later.  i have to attend classes on saturday and sunday, and i have to teach classes the whole day on sunday. a while ago my boss informed me that i just might have to be on standby at the mandarin oriental from sunday morning onwards, so right now i\'m just looking at my options and wondering which one is going to give. at the moment, all them are unyielding and i have a funny feeling that  i\'m  the one who\'s going to give.  i\'m just going to be so glad when it\'s all over.   '),(25014,2238828,'02,May,2002','2002-05-02 00:00:00','    a rooftop with stars above it, a sky so big and stars so indifferent that no matter what i do, i can\'t hurt it. i want something undentable, something that i can break my body against without breaking it as well - something constant versus this inconstancy. i hate this emptiness that rings inside my bones, that scrapes off the edges of my skin. and if i only knew how to fill it up, i would.   '),(25015,2238828,'23,June,2002','2002-06-23 00:00:00','    At 5:56 PM yesterday, the nation was sent into shock when Datuk Seri Dr. Mahathir Mohamad announced his resignation from all UMNO and Barisan National Posts.  - \"Jolt From PM\", Sunday Star, 23 June 2002  i just saw the paper, and my heart turned over. apparently his resignation was unanimously rejected by the UMNO assembly, but there were a few hairy moments. no one saw it coming; as soon as he announced it, the deputy prime minister ran to the microphone and begged him to reconsider. the heads of the women and youth wings rushed to him; rafidah was overheard saying \"no, no ... why?\" before she pushed the microphones away so they could confer privately. the whole nation went into cardiac arrest, i imagine ... this was being broadcast live over one of their major television stations.  he retracted the decision an hour or so later, after an intense private conference with his closest aides. the entire UMNO assembly remained glued to their seats, insisting that they were not going to leave until they were sure that mahathir had rescinded his decision.   there\'s a photo of him on the front page, with a look of utter weariness and frustration on his face ... more than anyone, this is mahathir\'s country, something he\'s invested his entire life in. i suppose that logically, everyone knows he\'s mortal and that he\'s going to have to step down sometime, but no one\'s really been thinking about it. this decision just kind of forced everyone to take a good look at the tenuousness of the situation, how the stability and prosperity of an entire nation has been balanced on one man\'s shoulders. it\'s too much, and it always has been, but mahathir has never complained and never backed down.  this brings me to the question: if he steps down, what then?  life in asia - a perpetual transition. we all live on the edge of hope, clinging to family and traditions because they\'re all that\'s sane in a world that changes by the minute. huntington wrote \"The Clash of Civilizations\" aeons ago, and some parts of what he wrote is already obsolete, of course. but sometimes the world still feels frighteningly bipolar, a standoff between east and west, two worlds coexisting on the same planet. the west is frustrated by our dependence on our governments, and our paradoxical mistrust of them; we are mystified by their confidence in a world that we have never learned to trust fully.   when are we ever going to learn to speak each other\'s languages?   '),(25016,2238828,'01,June,2002','2002-06-01 00:00:00','    in about a minute we\'re leaving for dance practice, so i can\'t stay long, and i can\'t really write out a long entry. but i just wanted to say that for the first time in weeks, i feel luminous again, healthy and ripe and glowing like a normal twenty-one year old. there\'s still that undercurrent - a frail, trembling notion of infertility beneath my skin, because i still haven\'t changed my mind about not having children - but it\'s self-inflicted rather than externally imposed, so it\'s bearable.  i mean, i never could bring myself to cut myself that deeply.  night isn\'t so scary nowadays, and my bed is suddenly cozy. it\'s lovely to be happy again.   '),(25017,2238828,'28,July,2002','2002-07-28 00:00:00','    i know; it\'s full of faulty science and ungainly metaphors and awkward writing, but it was knocking around in my head and i had to exorcise it. forgive?  - - -  midnight.  she kneels, awkwardly, on her bed, shifting instinctively to keep her center of gravity on the soft uncertainty of the mattress. the springs, even buried where they are beneath layers of sponge and cloth, dig into her knees, and she can feel bruises forming where her bones jut hard against her skin. tomorrow there will be little blue spots on her knees, and she\'ll rub absently at them, not really remembering where they came from. it\'s just pain, just another sensation that gets absorbed and overlooked.  outside, the stars are burning, gaseous masses that glow so hard that she can see them from where she is, millions of light years away. they\'ve been shining for years, she knows; in some cases they\'re so far away that by the time their light reaches the earth, the star itself is dead, gone nova and dwindled into a red giant, collapsing into a black hole. some of the stars she\'s watching, right at this very moment, are now dead and gone, and the lights in the sky where they used to be are empty - beauty without substance, warmth without actual fire.  it makes her hurt to think about it.  she liked looking at stars before, when she was younger. she had admired their constancy. no matter what happened, no matter where she was, she knew that as soon as the sun went down, she could open her window and look up, and they would be there. the big dipper. ursula major. orion. castor and pollux. if she stayed up long enough, scorpio and cancer. there was a harmony to their movements, a certainty to their dance, a pattern that they could not break.  the first time they had moved, she had lain awake all night, scared of her strange bed and the strange sounds. the next night, her father had taken her to the window and pointed out the big dipper, and before she knew it, they were starwatching.  that\'s the north star,  her father had said. she can remember his voice, the brush of his shirt against her jammies, the exact words he had spoken, as if it had been ten minutes ago.  the stars move, did you know that? it\'s like a big slideshow. the earth turns around and we see different stars, but there\'s one star that doesn\'t change.   years down the road, bigger and older and wiser, she can still hear her own voice, tiny and piping.  the north star,  she had echoed, and down the years, as she watches the night sky, she hears her father laughing proudly.   that\'s right, my girl,  he said.  people used that star to find their way home.   they had fallen silent then, eyes fixed on the wide scrolling brilliance above them.  daddy,  she\'d said, her voice sounding even tinier,  i don\'t like it when things change. everything gets so strange and my tummy gets nervous and my body feels funny, like it\'s forgotten what it\'s supposed to do.   her father had not looked back down at her, but continued to look up at the sky.  things change, honey,  he\'d said.  that\'s just the way it is.    the northern star doesn\'t,  she\'d retorted then, and he\'d finally looked down at her, a slight smile on his face. she can\'t remember what his eyes looked like. in the semi-dark, his face had been clouded, and although she could see the glint of his teeth, his eyes were shadowed. there was something strange, something ...  then he spoke, and his voice was daddy\'s, and his warmth was daddy\'s, and it didn\'t matter that for a moment, he\'d looked like a stranger.  that\'s right, honey,  he agreed.  the northern star doesn\'t change.   she remembered that so clearly. the year after that, her family had packed up and moved again, to another country where the people didn\'t speak her language and looked at her with eyes that examined and weighed, that measured her and found her strange. that first night, listening to her brother making small distressed noises in his sleep, and the unfamiliar noises that all sounded creepy, she had gotten up and gone to the window. as her father had told her, the north star was there. hours melted away as she stared out at it, her hand pressed hard against the glass as if she could draw it to her by sheer will. when the sky had begun to lighten, silver streaks forming in the previous unbroken darkness of the sky, she had gone back to bed and slept.  they had moved several more times after that, from city to city, from one country to another. it had become a habit; every time she found herself in a strange bed, a strange place, she would stay up long enough to go to the window alone and look for the north star. it was always there. in her first year of college, she had broken up with her high school sweetheart. she had moped about it for days, huddled in the dark of her dorm room while her roommates gave her sympathetic looks and crept out to give her privacy. on the third day she had tried to call home, only to receive a taped message; her parents weren\'t home either. alone in a strange city, cut loose from the one thing that had bound her to all the familiarity of her old life, she had returned to her childhood habit of watching the stars.  tomorrow is her graduation day. her friends lie scattered on the floor and couch and bed, worn from the last hurrah they\'d held earlier. she\'s the last one awake, unable to sleep. two weeks ago, her mother had called her, voice thick and blurry with weariness and hurt. strange that she can remember words spoken by her father years ago, and that something her mother told her just two weeks past is vague in her memory. she can\'t remember the words, only the way they were said, and the dull empty way she\'d sat, cradling the phone long after her mother had hung up.   things hurt and people change. she knows that. a lifetime on the move, a childhood spent learning and relearning societal norms that differed from place to place, a heart that knows only her immediate family and the north star as constants - these have shaped her. she doesn\'t ask for much, really. she doesn\'t cry for the impossible.  but some things - some impossible things could have been possible, she thinks, in a surge of anger. if people tried harder. if they valued things more. she\'d said the same to her father over the phone last month, when he\'d called and tried to explain.  you\'re nothing but a big cop-out,  she\'d hissed, her voice turning thin and ragged. she imagined bones jutting out from it, rough and jagged at the broken edges, glistening with blood from where it had ripped through skin and muscle. she\'d wanted him to hurt like that.  he\'d sounded defensive, and tired beneath the hardness.  try to understand,  he pleaded.  people grow apart. things change - i needed someone who understood -    she\'d put the phone down then.  she\'s still kneeling at the window, staring at the stars but not seeing them. behind her, a friend stirs, groans.  i think i had too much to drink,  he says. a slight pause, and the rustling sounds of him getting to his feet.  what are you doing up?   there is a brief silence.  thinking,  she says.    anything interesting?  he has moved to stand beside her, peering out the window. she can see his reflection in the glass, feel his warmth beside her, and is struck by the incongruity of life. he is a stranger made familiar by four years of camaraderie, a random person made dear by time and proximity.   just wondering,  she says,  about the stars. you know how we can see their light, but not really know whether they still exist? i mean, since it takes so long for their light to reach earth. by the time we see them shine, they might be black holes.   he scratches his head and stretches.  hell of a time to be thinking about physics,  he grumbles.  she shrugs, smiling a little.  i guess. i was just thinking - what if the northern star was like that? i mean, we trust it, you know? people get lost or confused or whatever and they look for the northern star to get their bearings. but - what if it weren\'t real anymore? what if they were just seeing its light, and the real northern star was nothing but a hole in the sky?    dunno,  he mutters, searching the star out with his eyes. she grabs his hand, extending his pointer finger, and guides it until it rests just under the pinpoint of light that is the northern star.  that\'s a tough one. i mean, don\'t we have instruments that measure that kind of stuff?   she is insistent:  hypothetically.   they breathe together, waiting in silence, as he ponders. the reflection of light on glass makes it difficult to see his eyes, and in a brief moment of unexplainable panic, she turns to look at his face. his head is thrown back, his eyes thoughtful, and she relaxes, returning her gaze to the star in question.   i don\'t know,  he finally says.  i guess if i were an astronomer, that would have some deep scientific significance or something. but i\'m not, so ...  he drifts off into silence again, his finger dropping from the window pane.  i guess it doesn\'t matter much. i mean, those people found their way home, right? whether the star was actually there or not.   she turns to look at him, and finds him looking at her, too. they stand, staring at each other as they consider his words. she breaks the silence first, with a slow smile.  makes sense to me,  she murmurs, squeezing his hand and moving off the bed.  makes a lot of sense.   outside, the northern star glows.   '),(25018,2238828,'25,July,2002','2002-07-25 00:00:00','    Okay, this falls into the \"strange but true\" category. I first mentioned it at an intellectual rights seminar I attended last month. The rep from the Swedish Embassy was completely amused, as were the people from the firm that sponsored the seminar; they asked me several times whether I was joking. Well, I wasn\'t, and here\'s the story.  The Philippine Long Distance Telephone Company (PLDT) is one of the biggest telecommunications companies in the Philippines. For a while, it was the ONLY telecom company in the Philippines (yes, folks, you can say it with me: MONOPOLY, and not of the ambiguous Bill Gates kind either). Sometime in the 1990s, the government decided to privatize the telecommunications industry. This was due to several reasons - the most obvious of which, of course, that it\'s not healthy to have a monopoly in any industry. If you have no competition then there\'s no burning need to improve your services, and PLDT, in a lot of ways, was a shining example of this.  Por ejemplo : if anyone remembers, back in the days before privatization, it took almost forever to get a phone line. You filled out the application, you sent it in ... and then two or three years later, PLDT would send you an initial reply.   As you can imagine, this irritated consumers. One particularly annoyed customer took his war to the Internet - and in so doing, discovered that the Philippines\' premier (well, it  was  the premier back then because we had no choice) telecommunications company had apparently not discovered the Internet yet. Having made this strange and perplexing discovery, he cheerfully purchased and registered the domain  urlLink pldt.com , and used it as a platform to voice out all that he found objectionable about PLDT.  When PLDT, the slug, found out about this, it was Outraged (yes, with a capital O). They had to settle for registering the domain name  urlLink pldt.com.ph , and did so. However, it still rankled deeply that someone was using their own name against them, so they filed a lawsuit - a major company against a private individual, whose case has since become a cause celebre for all annoyed consumers in the Philippines. The battle is still ongoing, and the site is still up. The owner of the website has since expanded his website\'s coverage to include critiques of national and international goings-on. It\'s quite sarcastic, irreverent, and roll-in-the-aisles funny - unless, of course, you\'re the one it\'s targeted at. In which case it ceases to be funny, but ain\'t that the way it\'s always been.  To this day, if you go to the address line and type in http://www.pldt.com, you will be directed to an anti-PLDT website. Call it the leftover anarchist in me, but little ironies like this are just plain hilarious. Incidentally, this site was set up a looooong time ago, and I\'d almost forgotten about it. I was reminded of its existence by  urlLink a similar incident  involving another company, and thought it might be of interest - to trivia freaks and the easily amused, at the very least.   '),(25019,2238828,'22,July,2002','2002-07-22 00:00:00','    to all my former teachers and professors and advisers: consider this an apology from a former class terror. you\'ll be glad to know that i have amended my ways; instead of staying in class and sleeping or playing tag or getting into big dramatic arguments with teachers and fellow students alike, i now save everybody the time and trouble and just cut class.  really, it\'s so thoughtful of me.  having said that, i am now experiencing what some would call karmic backlash; i teach basic computer at the embassy\'s skills training center twice a month and let me just say it\'s not easy. trying to impart practical computer skills to people who have no experience whatsoever with the computer, and have little to no access to computers at all, is close to impossible. it crosses the border into ridiculous when you factor in the fact that there are no working computers in the lab. in effect, i am a computer teacher without computers.  yesterday was one of the most trying days. we\'ve had a lot of suspended classes this semester, for one reason or another, and i finally decided to break my no-makeup-classes rule and hold one for my students just so we could make up for lost time. i lived to regret that decision, as it turns out; it resulted in my having to teach SEVEN hours straight, from 10:30AM to 5:30PM.  a short backgrounder: the skills training center is a branch of the embassy\'s labor office, and it holds skills development classes for overseas filipino workers in kuala lumpur. the classes are held every sunday; the courses available range from basic computer training to basic nursing and homecare. ninety-nine percent of the students are maids, and a scattered few are housewives and/or relatives of overseas filipino workers who are looking for something better to do with their time. the teachers are filipino professionals working or living in KL; most, like my mother, are professionals who are here on a dependent\'s visa, and are thus not allowed to work. all of them are volunteers.  i teach basic computer 1, and just started this semester. how and why a polisci-major-slash-MBA-student-slash-government-flunky ended up teaching computer in kuala lumpur is a long and involved story, but let\'s just say there were three major reasons: first, the previous computer teacher got a life, and wanted out of teaching the course; second, my mother decided i had too much free time on my hands and needed to do something worthwhile with my time; and third, i decided it was time to quit whining about how rotten it was in the state of the philippines and start doing something about it. long story short, i found myself agreeing to teaching basic computer twice a month - three two-hour classes every two weeks.  at any rate, due to the make-up classes i promised, i found myself with an even more insane teaching schedule than usual - while i usually have a one hour break between my morning class and my two afternoon classes, this time i had none, PLUS an hour-long extension for my last class. all this, on top of the fact that i had to wake up at 7AM to attend the 8AM service at church - a service i usually don\'t attend because i\'m never awake before 10AM, but ended up attending this year due to my crazy teaching schedule.  so. wake up at 7, attend church at 8, skip out on the post-sermon rituals by 9:30, and arrive at the center by 9:45. not being a masochist, i went straight to my classroom to leave a message on the board:  class starts at 10:30. eat, drink, be merry, and prepare intelligent questions for the makeup class discussion.  then i went back downstairs to have breakfast - a nice, relaxed breakfast with my family such as we never get to have nowadays.  my first discovery, upon going back to my classroom at 10:30, was that none of the computers were working. none. nary a one. i had a classroomful of expectant, eager students on my hands, an introductory session to teach, and no computer to produce a demo on. well, UP students - or UP alumni, as the case may be - are nothing if not resourceful; i went back downstairs and got a whiteboard marker. if i couldn\'t turn on a computer and show them what i was talking about, then, by golly, i was going to draw them a picture and confuse them even more with my horrific drawing skills.  at around 12 noon, i decided to give myself and my students a break, and announced that we would have a five-minute break. upon which one student piped up  but when are we going to eat?  i gave her my best you\'ve-got-to-be-kidding-me look.  class started at 10:30, i pointed out. if thirty minutes was not enough time for you to have a nice big breakfast, then i don\'t know what is.    but i have a class from 1 to 3,  she whined.   i have a class from 1 to 3, another one from 3 to 5, and a makeup session from 5 to 6,  i replied.  your point?  whereupon the other students started giving her dirty looks, and she subsided. i relented a little bit:  look,  i reminded them,  you asked for this time slot. i gave it to you. and if you remember, i gave you all permission to bring sandwiches or snacks to class, so that you could nibble on something if you got hungry. i\'m not inhuman, but give me a break, okay? i\'m not your mother either.   irritating students and hectic schedule notwithstanding, i had a relatively good time. one of my students, frighteningly honest, raised her hand during the Q&A session to ask  how do you use the computer without your employer finding out?  six months ago, i would have laughed and thought she was joking. these days, i know she speaks with heartfelt honesty and an earnest desire to know. eighty percent of my students have never touched a computer; ninety-eight percent are not allowed to handle their employers\' computers, and fear reprisal should they do so. there\'s the blessed two percent, who either own computers or are allowed to use the computers at home, but as you can see from the numbers, they are in the sad majority. at any rate, everyone laughed and i had to give a brief smile, but it was obvious that they wanted a real, honest-to-God answer.  for once, i was grateful for the ribbing and teasing that ensued; it gave me a little while to organize my thoughts and figure out how to respond.  first of all,  i said, once everyone had calmed down,  i know that a lot of you don\'t have access to computers at home, and that you\'d hoped to be able to practice on actual computers during this course. as you all know, we haven\'t been able to do so very often this semester, and i\'m very sorry. it\'s a ridiculous situation - a computer class without computers - but it\'s a situation that can\'t be helped at the time, so we\'ll just have to make do. second, i realize it\'s difficult for a lot of you to understand and visualize what we discuss in class, since you don\'t have computer experience, but i appreciate the fact that you\'re all making an effort nonetheless. having said that, i\'d like to tell you that - i\'m sorry - there is no surefire way of using a computer without your employers finding out, and at any rate it\'s wrong to use something that\'s not yours without permission. so. two things,  i said, to a class that was surprisingly silent and listening with full attention.  number one, don\'t use your employers\' computers - or any of their stuff - without their permission. there\'s no two ways about it, you\'ll just have to ask their permission and live with whatever they decide. number two - IF they decide to allow you to use their things, DON\'T abuse the privilege. use only what you need to use, work only with the programs and buttons you understand, and leave the rest strictly alone. is that absolutely, positively clear?   i still feel for them, though. our computer lab does have computers, but a lot of them are either out-of-order or under repair, and since i\'m just an experienced user and not a technician in any way, i daren\'t touch them either. these women (and one man) really want to learn, and for that desire alone, they deserve complete, working equipment. they\'re amazingly cheerful and resilient, too, despite the kinds of lives they lead. the other week i was discussing one of my students with my sister, and she just shook her head when i was done.  i don\'t think i could teach at the center,  she said, quietly.  how do you not break down and cry every time you hear their stories?   the answer is simple, if a little cold: by not thinking too much about it. if i think about it, it builds up and grows and becomes my pain instead of theirs, and that\'s wrong. it\'s not my pain, it\'s theirs; they need the help, so much much more than i do.  all those songs and dances and pretty political ads in the philippines about how the overseas filipino worker is a modern hero - they\'re useless and pointless and insulting in their inadequacy. don\'t enshrine them, you crocodiles. they don\'t need the pedestals.  HELP them.   '),(25020,2238828,'15,July,2002','2002-07-15 00:00:00','    moving is a bitch. i had to leave my cat and her three kittens behind, and tomorrow we\'re bringing them to the SPCA.  they kill them if no one adopts them within a week,  my sister pointed out, as we lay on the floor of our emptied former house, running our fingers over the soft exposed bellies of our kittens.  i tried not to think of that, tried not to think of soft little pink noses and trusting dark eyes closed and stilled and cold forever.  i\'d rather they be dead than starving and abandoned and cold on the streets,  i said instead, with the downy milk-filled belly of one kitten moving drowsily beneath my stroking fingers. their eyes have just opened, and they still look rather cross-eyed, and they\'re beautiful babies.  sometimes i hate moving, and i hate strange beds, and i know i told E that i\'m used to sleeping in strange beds, but isn\'t that such an oxymoron? and more than anything, i hate leaving things behind and cutting things off, viciously, because it\'ll hurt more to let it drag out and linger.  so. i don\'t care about those kittens, and i don\'t care about my cat, and i won\'t dream about the way she looked at me when i sang her to sleep.  i won\'t.   '),(25021,2238828,'05,July,2002','2002-07-05 00:00:00','    i suppose i should be intelligent and rational and analytical about  urlLink this  and  urlLink this  and oh God,  urlLink THIS , but i don\'t want to be and at 11PM in the evening after a difficult day, i can\'t be. the papers are beginning to depress me. i log onto  urlLink The Inquirer  every day, or at least a couple times a week, to keep up-to-date on what\'s going on in my country. is it too much, i ask of you, to give me good news?  i know. whatever you\'re going to tell me, believe me, i\'ve heard it before. someone actually told me once  you have no right to suffer because you don\'t even live here. you don\'t KNOW what it\'s like.  that\'s not strictly true; i lived in the philippines for a little over eight years, starting from 1992, when electricity was functional for only four hours a day and water only came on from 2AM to 4AM. kidnappings and rape-slays were the crime-du-jour. (yes, virginia; even crimes go by trends in manila. bombings are now apparently THE crime of the day. rape-slay? dahling. you are too, too passÃ©.) six months after we arrived, a drugged up man stabbed a fifteen-year-old boy to death in front of our house as we listened, terrified, in our rooms. fifteen minutes after the boy died, the neighbors came out in full force with bats and pipes and started beating the killer up, and all i could think was  what\'s the point? the boy\'s dead. you\'re fifteen minutes too late.   but okay. let\'s take for granted that i have spent more than half of my life outside of the philippines, and that given this situation, i may not be fully aware of the situation in the philippines. tell you what; let\'s swap. you stay here, i\'ll stay there. i\'ll go through your traffic and your corrupted politicians and your urban poor, the ones whose shoulders are tense with violence and whose eyes are beaten and resigned. i\'ll live through your rallies and turbulence and impossibly high cost of living vis-a-vis the low wage levels. i\'ll rub shoulders with the disenfranchised and marginalized, the people who have no names and no voices and no way to tell us that they exist and bleed. i\'ll do that.  you - you come here and line up at their immigration counters. hold out your passport and see their faces change when they see your nationality. listen, burning in angry silence, while they harass your mother because they think she\'s a maid, and refrain from doing anything because after all, it\'s not your country. go to places and watch the eyes of men change when they learn your nationality; learn to duck their leering, wandering hands as they whisper  filipinas are pretty women  in your ear, and make it sound like an insult. go to class and listen to people lambast your country\'s politics and government, and the awkward, embarrassed silence when they realize that you are there. go to the embassy and watch the runaway maids creep in, red-eyed and bruised and terrified, and watch them cower against a wall no matter how kind you are being, simply because it\'s been so long since they\'ve known kindness. you do that.  then you come back and look me in the eye and tell me i don\'t understand. capisce?  i know it\'s not all sweetness and light in our country. i am not asking for a cover-up. i am not asking for a whitewash. and, God forbid, i am not asking for lies.   all i want is a little hope from you, you know? give me a little optimism, something to grab on to when everything seems bleak. tell me that maybe it\'s not ok now, but it will get better, because if it\'s not, what are we alive for? what am i here for? what am i doing here, standing here and loving my country with all that i have, if my country is going to be nothing but bullshit in three years? what was my father\'s generation\'s fight for freedom worth, if you degrade our country like this?  hope for me. hope with me. our country needs a second chance, dammit, and if no one else will offer it, we should.  damn it.   '),(25022,2238828,'21,August,2002','2002-08-21 00:00:00','    my eyes are burning, burning. late night last night - had to attend a farewell party for the ambassador, hosted by one of the business associations my office works closely with. like i\'ve said, i LOVE dressing up and going out - there\'s something electric about red lipstick, black stockings and high heels that just tilts your chin up and makes you walk sweepingly, like an actress - but not when it\'s required. not when it\'s a work thing.  not to say that it didn\'t turn out well - it was a very small, intimate sort of gathering, and despite the ultra-formal setting, people were chatting comfortably and easily with each other. there were no long-winded speeches or emotional outbursts - the farewell speeches and gift-giving probably took up no less than fifteen minutes, all told, and that\'s a generous estimate. by 10:30PM, PH (that\'s my colleague, not a chemical measurement) and i were standing out in front of the hotel, waiting for our respective rides home. on the other hand, i was still charged up from all the socializing and whatnot, so it took me a good three hours to settle down once i got home.  this morning did not help, either. i came in to a pile of work, all marked either  urgent  or  v. urgent - today, pls.  for five minutes i just sat at my desk staring at them, because i don\'t function like a human being until after 10AM. after which i shuffled them around, closed my eyes, plunged my hand into the pile and pulled out one, and that was what i did first. so it\'s not a scientific method, sue me. you should have seen my boss\'s face when he walked in on me covering my eyes with one hand and groping for a paper with the other. he can\'t blame it on my education, though, because we graduated from the same university.  by 10:30 i was alone in the office - my boss had gone off for various meetings, and PH had just left to drop some stuff off at the airport. i had about ten minutes of glorious peace before the other shoe fell and chaos descended again. my boss called to tell me that i had to tell PH to bring him some extra calling cards; two people walked in and informed me that they wanted to make a trade inquiry; one of the manila offices called on the other phone to ask for information on an upcoming event. i called PH on one phone, and while waiting for her to pick up, continued talking to the manila office on the other phone, pausing briefly to ask the walk-in clients to take a seat, i\'d be with them shortly. they had amazing poker faces; it must have been quite amusing to walk in on a girl holding two phones to her head with her shoulders, writing down something with one hand and scrabbling for papers with another, but their faces remained absolutely expressionless.  at any rate, i managed to handle everything. how, i\'m not quite sure - i vaguely remember pushing a handful of brochures at the clients and telling them to look through them, just before dashing off to answer the phone again. PH turned up five minutes later; i dragged her into the door, begging her in a harried whisper to handle the clients while i answered the phone (which was ringing, YET AGAIN). she didn\'t stay long, though - just long enough to give the clients even more brochures, then she was out the door with a box of calling cards.  when i\'d settled the phone issue once and for all (by leaving the phone off the hook, mwahaha), i went over and finally spoke to the clients, who looked immensely relieved that i seemed relatively sane. ten minutes and a lot of handshakes later, they left carrying a mountainload of brochures and my assurances that they could return if they needed more information; i put the phone back on its hook and collapsed in my chair, laughing hysterically.  hence when E called me (from taipei, of all places) twenty minutes later, i was still charged with all the adrenaline i\'d worked up during that mad rush. he accused me of drugging myself up with caffeine, i staunchly defended my honor against his false allegations, and we ended up talking a lot of nonsense about how he could walk down to KL from taipei and we could maybe swim to manila for a cup of espresso. to all filipinos reading this journal, i\'m sorry. government workers ARE insane, you\'re right to be suspicious of us. or at the very least, E is definitely certifiable. i\'m still borderline, although on the border of what exactly, i don\'t know.  i\'m in the eye of the storm at the moment, typing out this entry during a brief lull. tomorrow my day will start at 5:30AM because i have to help check-in a VIP at the airport before 8AM, and if the embassy doesn\'t provide protocol officers, i will end up carting his (and his delegation\'s) luggage all the way from the hotel to the airport. alone. E seems to think that should be no problem for me, but like i told him, i am NOT The Rock.  laugh with me, folks. we\'ll see where the day goes from here.   '),(25023,2238828,'19,August,2002','2002-08-19 00:00:00','    i admit i have a soft spot for  urlLink mindanao , since both my parents were born and raised there - in the heartland of the ARMM, or the autonomous region of muslim mindanao. how they came to be born and raised there is a whole other saga entirely, since my mother\'s family originally comes from nueva ecija in luzon, and my dad\'s family - well, the maternal side of it, at least - comes from cebu. dad\'s paternal grandfather was a landowner on negros occidental, i think, but like i said, unearthing my philippine roots would reveal a story worthy of judy ann santos\' tearjerking soap operas. it\'s long and interesting and mind-boggling, really.  at any rate, i\'ve been to mindanao, and it\'s a lovely place - the town where my parents met and grew up is both provincial and sophisticated at the same time, with illiterate farmers mingling and mixing freely with the intelligentsia of the town university. nothing at all like the mindanao depicted in the international press, with its bloodthirsty terrorist groups hiding out in jungles, preying on unsuspecting tourists, scuba divers, and missionaries. recently we\'ve had a lot of events going on in mindanao, and every time we call up someone to invite them, we get the same question: mindanao? isn\'t that the place, you know, with the terrorists?  so if that question was indeed running through your mind, rest assured: mindanao is a huge island. the area that bristles with terrorists is a mere one fiftieth, if that, of the entire island, and it\'s located far south, toward sabah. while they may have enjoyed a greater range of power previously, the recent and concentrated war against terrorism has driven them back into a couple of isolated islands, where they are rapidly being picked off.  mindanao is a sad story in and and of itself, actually. being so far from the nation\'s center (which would be manila), it hasn\'t received a lot of attention - financially or politically. in FVR\'s time, attempts were made to develop mindanao through the SPCPD (the southern philippine council for peace and development), which was headed by nur misuari. we all know how that turned out - misuari milked mindanao dry, took people hostage when people started to point accusing fingers, then ran to malaysia to hide. luckily malaysia delivered him back to the philippines by the scruff of his neck, to face the consequences of his greed. vengeance is mine, saith the Lord, but the fleshly part of me wants the philippine government to just turn him loose in a crowd of mindanaoans - the very people that he embezzled from - and leave him to his fate.  in an unrelated note, nur misuari used to teach political science at the mindanao campus of my university. isn\'t that scary?  after FVR came estrada, who didn\'t really care. if i remember correctly, part of the reason people were furious at estrada was because toward the end of his bumbling reign, the abu sayyaf were waxing in strength, taking hostages with such ease that they\'d begun almost to make a business out of it. (rumor has it that there was an argument between commander global and commander robot [where DO they get these aliases, by the way? Voltes V comic books?] to the effect that global didn\'t want to share the ransom money, and told robot to go and \"kidnap his own hostages\".)  currently we have gloria, who vowed to wage a full-scale war against terrorism and to enhance efforts to develop mindanao. it\'s about time, seriously; part of the reason there\'s such strong anti-government sentiment in mindanao is because they\'ve felt ignored for so long, shunted aside in favor of the needs and demands of the capital region. gloria is fighting two wars in mindanao, actually - the more public one against terrorism, and the private, more down-home war to regain mindanao\'s trust.  personally, i think a government initiative will not be enough; they need to educate the private sector as well, and enlist their support. this all-out effort to get foreign investors interested in mindanao will fizzle out pretty quickly if those same foreign investors figure out that domestic investors are themselves not interested. i mean, wouldn\'t you be doubtful about investing in a place where not even the locals want to go?  given this kind of history, news like  urlLink this  irritates all heck out of me. mindanao is part of the philippines, not a separate country, and it should be treated as such - given an equal chance, at least. it\'s the crab mentality all over again.  life in the philippines - a carnival that will not stop. i  did  warn you, right?   '),(25024,2238828,'15,August,2002','2002-08-15 00:00:00','    dusk at my window is a lovely thing. if i\'ve whined about how small and stuffy my room is, i\'ve done it a million times, but i love my window. it faces the west, catching the light as the sun sets, and in those red-and-orange moments, everything in my room comes alight with firecolor. when the sky ripens, purpling like a day-old bruise, so does my room, twinging with shadows in unexpected places.  and when the light stops playing and settles down, down, coiling itself up like the poet\'s \"gorgeous snake\" and losing us in the shimmer of darkness, that\'s when the stars come out. in the quietest of moments, when my house sleeps, i kneel at my window and breathe, and nothing matters but this.  the world empties, tipping itself out and shaking loose of what it has collected over the day. you can almost hear the crackle and tinkle of accumulated insignificances: there goes that annoying client who always asks you to do the impossible and gets mad if it fails to happen; here falls the irritating next-door neighbor whose cigarette smoke fouls up the entire floor\'s air. everything superfluous is removed, falling loose the way pins are stripped from a woman\'s hair. this is the lightest and easiest moment of the day.  tomorrow i start over again, emptied and replete all at once, stripped of insignificant irritants and filled with the trust of stars and wind and breath. this is what keeps me going, and this is what makes me smile in strange moments.  dusk at my window, and tomorrow riding on the edge of the very last star. there are always reasons to live, but this is my favorite one.   '),(25025,2238828,'10,August,2002','2002-08-10 00:00:00','    yaz sent  urlLink this  over to me just a while ago, and in accordance with my vile, cynical, and completely irrational mood, i would just like to say to all the politicians: SHUT. UP. just shut up. zip the lips, lock your clocks, throw away the keys, and MOVE. i\'m sick and tired of press conferences where they express \"indignation\" and \"offense\". please. we\'re a third world country in hock to more than half of the countries around the world, we don\'t have much room for all this chest-puffing. just - please. tell the press \"we are currently addressing this situation\", then shut up and go on with your life. this is one person in perhaps seventy million who feels that way, but please, i\'m begging you. we don\'t have time to sit around and blabber, we need to go DO something.  and to the philippine media: how sad to think that i once thought i wanted to be one of you. i wanted to tell the truth, and i believed that if there was one institution in the country left that still held that virtue intact, it was the media. well, i was wrong. the philippine press enjoys more freedom than most press institutions in asia - than most press institutions anywhere, i guess you could say. but freedom is not carte blanche to go out there and say whatever the hell you like, take the truth and manipulate it or twist it or put a spin to it or chop it up and rearrange it so it looks different. you have a responsibility, made all the greater by the weight of your freedom. i - and many other people like me - are frighteningly dependent on the media to provide us with news that is true and honest and unbiased. when the media fails, then this frightens me, because who then can i trust? i belong to a society that extends far beyond my miniscule network of family and friends, and yet if i cannot trust any institution outside of that network, i might as well cede citizenship and live like a hermit on an island, belonging to no country and protected by no one.  i\'m sorry. it\'s been a long day, i\'m tired, this entry will get deleted tomorrow, but for now this stands because  this is how i feel.   '),(25026,2238828,'06,August,2002','2002-08-06 00:00:00','    so, the embassy stories, as i promised yaz and G.    story the first:    this takes place a few years ago, in hongkong, where one of the maids (let\'s call her evie) who is currently a student in the embassy skills training center, used to work. one day, her employer decided that she no longer wanted evie as an employee, and told her that she wanted to terminate her contract. the story would have been over and done by then, and you would have been spared my pathetic attempts at storytelling, if not for one little thing: evie\'s contract had a clause in it saying that she could not be terminated without legitimate cause. the employer HAD no legitimate cause; she just didn\'t like evie anymore. take note: she didn\'t like EVIE, personally; there was no problem with the way evie worked.  well, evie refused to leave; of course life is difficult when your employer hates your guts, but then she had mouths to feed at home, and that pretty much superseded many of her reasons for leaving. her employer was in a bind until one day she came up with a supergenius solution: she drew up a termination contract in cantonese, presented it to evie, and told her to sign.   here we take a short commercial break for me to explain one thing: this is one of the biggest problems faced in labor disputes. the labor attachÃ© does negotiations, there is a verbal agreement between both parties, then just before they leave he takes the filipino worker aside and warns them, in our native tongue, not to sign any contract without reading it carefully.  if you don\'t understand it,  he stresses, don\'t sign.  not even if they force you to. call me first and let me see the contract, and then you can sign it in my presence to prove that you\'re not being coerced.  two months later the worker calls, crying, to say that her employer is now claiming that she is contractually obligated to feed them with the blood from her veins and the flesh of her body, or something like that. of course, the attachÃ© then blows a gasket, his normally pleasant face turning red with infuriated frustration.  did you sign a contract?  he asks, and the answer is almost always yes.  we were forced,  they tell us.  they said they wouldn\'t pay us if we didn\'t sign.   back to evie. she took one look at the contract and, bless her heart, refused to sign.  i can\'t sign that,  she told her employer.  it\'s in chinese. i\'ll have to ask one of the embassy staff to translate it for me first.   the employer, taken aback, tried to cajole her first -  don\'t worry, it\'s just a standard form,  she said.  you need to sign this to process your employment.  evie, bless her heart a second time, still refused to sign, saying that she would only sign it if and when (a) she was given a copy in english, and she was permitted to sign the english copy; or (b) it was turned over to be translated by embassy staff, after which she would sign the translated copy.  and i want to sign it in front of the labor attachÃ©,  she told her employer. well, the employer was infuriated - and hell hath no fury, you know the rest. evie says that for the next couple of weeks, her employer did all she could to force her to sign - until one day, she was no longer able to resist.  before you bust an artery screaming at your screen \"no evie! don\'t do it! don\'t do it!\" just like my grandma and i used to do back when we were watching those cheesy mexican telenovelas (whoops, too much information right there), let me reassure you - this story has a happy ending, otherwise i wouldn\'t be joking, i\'d be stringing virtual heads on a string.  at any rate, evie gave in and told her employer  ok, i\'ll sign.  so her employer gave her the document and pen, and stood over her as she signed.   however, evie didn\'t sign her real name - and no, she didn\'t sign something predictable and lame like  gabriela silang  or  peter rabbit , like i (and a couple other people) probably would have. no. evie was smarter than us, yo.  instead, she signed  pinipilit ako  (i am being forced) with a flourish, and the employer took the document away for safekeeping.  a couple weeks later they were arguing again about her termination, and her employer whipped out the document with a triumphant air.  you see this?  she asked evie.  you signed this paper that says that you are leaving my employ of your own free will, therefore you are legally bound to do so otherwise i will sue you and have you deported.    i want to speak to my labor attachÃ©,  evie said.   but you signed a legal document,  her employer informed her.  there\'s nothing he can do about that.   evie just shrugged.  let\'s see.   on they went to the embassy, where the labor attachÃ© stared at the document before turning to the employer.  i\'m not sure what you want me to do,  he said, as the employer had pulled it out of her purse voluntarily and put it before him as \"proof\".   do?  the employer spluttered, no doubt wondering if all filipinos were terminally stubborn.  it\'s very clear. that document says that she is asking to break her contract - you can have it translated if you like - and she has signed it.    well, no she hasn\'t,  the attachÃ© replied, looking stern.  according to this note here, she was forced to sign this document.   that\'s pretty much all she wrote. evie went back to a more secure contract, and the employer was fined.  if i had half her survival instinct, i could live on an island with nothing but a bottle of water for a year.   '),(25027,2238828,'27,September,2002','2002-09-27 00:00:00','    unlocking   (and even when i\'m absolutely sure that i\'ve snapped your last thread, severed the last fraying string of your hope in me, you turn to me and say softwarm things wrapped in sandpaper and ripped skin, things like keep dry.  if you love me so much, how could you have made someone who fails you so often?  i\'m afraid of you and i love you and i don\'t want love to come wrapped in bruises and blood, but it is. and do they ever come separate anymore, because nothing i\'ve seen has proven otherwise.  so tired.)   '),(25028,2238828,'31,October,2002','2002-10-31 00:00:00','    i know every girl\'s worst fear is that she\'ll turn into her mother, but what if your mother is the one trying to make you into herself?  urlLink laya , all i can think of is that maybe it\'s part of the separation process - where your mother still has difficulty separating you from the embryonic version of you that depended on her for everything; WAS her, in essence, since you gestated inside her body for nine months. it\'s difficult, pulling all those sticky little love-hooks out of you, especially when it\'s your mother who\'s stuck them in.  no one guilts like a mom.  i\'m afraid i can\'t quite sympathize, though. my mother and i have always been polar opposites, so we get along better than most mothers and daughters. it\'s my sister who feels the threat of being sucked into her, and who threatens my mother with her similarity.  me ... strangely, i\'ve always been more like a son than a daughter. ever since i was a very little girl, i knew, somehow, that i was being groomed to be my father\'s successor. he would take me to his office during summer vacations and leave me to browse through his papers while he buried himself in his work and forgot i existed. then i got older and it was the seminars he took me to, the papers he made me read, the universities i applied to and the course i chose.  i wanted to be a journalist and he wanted me to be an economist like him, and even then it felt like drowning - to be sucked into a field where everything i did would be perceived as toddling in my father\'s footsteps. we eventually compromised on political science, but the two fields intersect so much nowadays that sometimes we wander into arguments without meaning to.  even outside of academia and work, i\'m my father\'s son. we\'re not tactile. i\'ve never been able to throw myself into his arms, not even when i was very little. my mother tells me that i used to sit down across from him while he was working and talk to him levelly, and he would talk to me the same way, as if we were conducting business.  i don\'t remember this,  i said.   you were three,  she replied.  it\'s not something you\'d remember easily.   it\'s still the way i talk to him: short, professional. i get embarrassed when i cry in front of him. he still looks slightly bewildered when he wanders in and i\'m doing something girly, like putting on makeup or fussing with my hair. he is not afraid or overly concerned when i go out late at night, because he knows i can take care of myself.  and it\'s strange, but he\'s the only father i have, and i\'m his only rei.   '),(25029,2238828,'22,October,2002','2002-10-22 00:00:00','    i\'m 22. well. technically twenty-two and a couple weeks, but what are a couple of weeks in the grand scheme of things?  at least that was what i told my mother back in september, when we were doing something mother-daughter-bonding-ish. i referred to myself as 22, she said that i wasn\'t twenty-two for a week or so yet, and i replied  what are a few days?  then explained to her that it was part of my grand overarching scheme to ease myself into my new ages, as they come. it\'s a personal tradition i started when i was 16; every year i would have to reposition myself - hold back my hand from writing 15 or 14 or whatever on the forms, and stop my mouth from forming the syllables of my former age. it got too tiring, wiggling around uncomfortably in the first few months of my new age, and i decided i\'d start breaking myself in. so far, it\'s working. i suppose when i break past the calendar years (the filipino - don\'t know if it\'s used anywhere else - euphemism for any age past 31) i\'ll stop doing it, but for now: 22 is a comfortable age, and i don\'t mind accelerating the aging process.  consider this the inevitable Contemplation on Aging post. last year, if i remember correctly, i was moaning about being unformed and unfocused, all those things you\'re not supposed to be at twenty-one. in another journal, a non-online one, i wrote something like this:  what i really am, at this age, is a paper donkey that older people pin their hopes, like false tails, on. we\'re blinded by our incapability to see the future, and sometimes our refusal to see what\'s right there in front of us, and sometimes the things they pin on me, on people my age - they don\'t fall in the right places. and it hurts. sometimes i just want to be fifty and sure, with a past to look comfortably back on and laugh at, instead of here with crossroads branching out, and no real idea of which road i should be taking.   but that wasn\'t what i was thinking of, today. last night a friend called me up, fresh from her first semester as a UP student. she was enthusiastic and excited, and even over the phone i could hear barely-contained wonder in her voice.  there\'s so much to do,  she marvelled.  i wake up and i never know what\'s going to happen next.   and i smiled, remembering trees afire with flowers, and huge murals set directly on our main road, and eager fine arts freshies in pure black outfits and blazingly orange hair. we parted with a promise to meet up on Friday, bond over coffee, then head on out to a birthday party, filipino-style - which is to say, held at a karaoke bar. then i went to bed and woke up this morning thinking of UP, and of firsts.  i have a fondness for firsts, but i suspect we all do; it\'s easier to remember the first time you did something, the first time something happened. there\'s something new about it that time and cynicism finds hard to crease or wrinkle, something that makes us pack away our firsts in between sheets of paper, protect them so we can unwrap them in later years to think about again.  she was the epitome of firstness: barely-contained delight spilling over the phone, words stumbling over each other in an effort to describe what was still sinking in.  i was so scared at first,  she confided.  everybody knew where to go and they had friends, and all i had was me.    and what did you do?  i prodded, having had years of practice in big-sisterhood.   acted like i knew exactly what i was doing and where i was going so no one would think i was a freshie.  she paused, a little uncertain.  maybe that was a little arrogant?    did you get through the first day alive?  i asked, because that\'s the point of firsts, aside from their newness - surviving them.  she said yes, and i congratulated her, and on friday i will tell her that she\'ll remember that first day for the rest of her life. i remember mine, the way i remember all my firsts: with absolute clarity.  first day at college: fifteen years old, scared out of my mind and not sure who i wanted to be, hovering at the edges of the classroom. first road-trip: seventeen, with the joy of a new license, to a mountain with stars and my two best friends from highschool. first time i cut my hair shorter than shoulder-length: thirteen, and the sense-memory of how light my head felt, how my fingers would rake through my hair only to stop short at the blunted edges.  and today i thought: i\'m twenty-two and i\'ve had so many firsts, and won\'t it be beautiful when i\'m eighty and have time to sort through them with careful hands.   '),(25030,2238828,'21,October,2002','2002-10-21 00:00:00','    happy birthday, PH!  it was a pretty much sucky birthday, yes, since you were busy preparing the budget/itinerary for YET ANOTHER VISIT (for gosh sakes, government officials, go bug another country already) and i was being depressed-girl in the corner. but you\'re a great colleague, i still stand in awe of you for being able to work from 6AM to 3AM for one whole week, and your amazing talent for organization. remind me to teach you more bitchy tagalog phrases and the proper way to say them, because you know you need them for self-defense against E whenever he comes to  bwisit  ... er, visit.   '),(25031,2238828,'20,October,2002','2002-10-20 00:00:00','    urlLink darmok and jilad at tanagra , and isn\'t it funny that it\'s so much easier for beings of different species to come to understandings than it is for us - all human beings, albeit of different colors and ideologies - to get along.  tomorrow morning i\'ll be attending a security briefing at the embassy, because violence and death has gotten frighteningly random. in maryland there\'s a sniper who picks them off like berries, and in bali they blew up a bar, and in zamboanga a mall exploded. there\'s no particular rhyme or reason anymore.  just a month or so back i couldn\'t imagine living in a world where every day was touch-and-go as to whether you\'d make it back home alive, but i should\'ve known, shouldn\'t i? i already do live in that world.   '),(25032,2238828,'18,October,2002','2002-10-18 00:00:00','    i\'m pissed.  i am NOT escaping, okay? if i were escaping, then i would either be facedown in a pool of water somewhere, or huddled up in a cozy white jacket, in a room full of very nice people with Ph.D.s in psychiatry. it\'s ridiculous to say that i\'m escaping work by working, and it\'s not fair of you to tell me that i\'m not being responsible.  i am not running away. i am dealing. i am coping with all of this insane work you are throwing at me because i want to be good for you, but it\'s so hard to be good for you when you keep on raising the bar.  - - -  on the non-self-involved bit: there was yet another bomb blast in zamboanga yesterday. PH sneaked in this morning to drop off some stuff (because technically, she\'s on leave today) and asked me,  sotto voce , if i\'d seen the papers.  yep,  i said.  front page. good luck to whoever\'s going to promote the next mindanao event.   of course it sounds callous, but what else can you say? how do you put the color of fear and the shape of pain and the grating feel of frustration and hope and disappointment and anger into words?  so we didn\'t.   '),(25033,2238828,'14,October,2002','2002-10-14 00:00:00','    in malaysian news: terengganu\'s state government has enacted  urlLink a law  that says that tourists must now also be segregated according to gender. females can only be taken around with a female guide, and males and females cannot tour together. even earlier on, kelantan banned females from performing in public because ... i don\'t know. some maggoty reason relating to their twisted way of interpreting God\'s will. these are the same people who decided that in order to press a rape charge, a woman had to produce four witnesses - not just any witnesses, but four Muslim men of good standing. failing that, she would be whipped for slander.   with that kind of a track record, it\'s kind of hard to listen to their  urlLink protestations of innocence . \"we didn\'t know that they were going to enact it.\" that, i can believe, even if it sounds dumb. what i have trouble believing is that the CWC would have disapproved the ruling, even if it had been brought before them.  somewhere in between that, there was a policeman who raped - well, let\'s give him the benefit of the doubt and say he had sex with - two female inmates at a prison that he was a warden of. (and yes, i win for Most Convoluted Sentence In A Recently Published Journal. you may NOT send prizes or congratulations because my life sucks enough already.) despite testimonies and evidence that the sex was not in any way consensual, he was let off scot-free because, and let\'s quote the judge here, \"there is no solid proof that it was rape ... the women did not scream, or struggle, or protest\".   granted, there was massive outrage in the opinion pages, some of which - gratifyingly enough - came from the male part of the population. but outrage didn\'t help those two women, because they were sent back to prison and the policeman who raped them - ahem, \"had sex with\" them, because after all, he was acquitted - returned to his job after standing trial. in his own words, he was \"glad for the vacation\".  in addition: i do not at all see what point they are trying to make with  urlLink this . the girl was raped, period. what does it matter if she\'s filipino or malaysian or swahili? if it\'s the participation of the philippine government that irks you, then stick this in your boots and suck on it: (a) the girl admittedly has filipino blood, ergo the philippine government has some moral right to extend its protection over her and (b) your precious police/countrymen were not so nice about her nationality when they raped her.  in philippine news: i have no clue because i\'ve stopped reading philippine newspapers. i am aware that this may cause me to fail the foreign service exams, which i may take next year, but at this point, i really don\'t care. after that entire sabah debacle - which, by the way, is still ongoing - i am not prone to believing what they say, in any way. granted the trust thing cuts both ways and you have no reason to believe that what i am saying in this journal IS indeed true; for all you know i could be an aggrieved citizen with a grudge against newspapers and governments in general. there is one major difference, though: this is not a profession. this is a personal journal, and that is caveat enough. if you\'re going to wear the title of \"journalist\", then for heaven\'s sake act like one and go verify your stories before they get published.  and yaz: you\'re right, none of this is making me particularly eager to reproduce. i know we sound like the last paranoid freaks on earth, but i would much rather not go down in history as the mother of my generation\'s ted bundy. \'cause you know that with my track record, i\'d probably end up raising either a serial killer or a catatonic.   '),(25034,2238828,'12,October,2002','2002-10-12 00:00:00','    there\'s a conversation between buffy and giles that i chanced across a long time ago, and never forgot because it was so true. post-slaying, after yet another night of death and standing between humanity and the apocalypse, buffy asks giles  does life ever get easy?  for moments after that, there are no words, only a silence defined by buffy\'s tired seventeen-year-old eyes, and giles\' confounded, inadequate wisdom.   what do you want me to say?  he finally replies, helpless, and buffy says  lie to me,  almost like she\'s begging, and giles pauses slightly before doing just that.  yes, it\'s terribly simple,  he murmurs, going on about how completely easy it is to tell good from bad, how good always wins and evil always loses in the end.  buffy cuts him off with a soft  liar , and the scene ends there.  sometimes, you just need someone to lie to you.   '),(25035,2238828,'04,October,2002','2002-10-04 00:00:00','    first and foremost.   erwin : \"Most of the times I get it backwards, too. We\'ll be fine.\"  \"Also I thought how happy it would make me if I could just sit at a spot in Sunken Garden with you or anyone like you* and talk and burn cigarettes and drink cola. That was yesterday. It won\'t happen again.  \"I\'ll try.  \"So people hurt. It happens. So people worry about the future. It happens. So I\'m sad today. It happens. So you love/loved me. It happens.  \"*p.s. there is no one like you\"  i know i told you, once upon a time, that i tended to forget my own way, that i was so bad at directions that i kept on losing my path back to just breathing. thanks for reminding me, for giving me words that were comforting and safe, for giving me back my words at a time i needed them most and   thanks for being you. you are in my notebooks and i hope i am in your shoeboxes, if shoeboxes are still your life\'s point of culmination and accumulation.  and one day, we should really stop being so arcane, yes?  - - -  i still feel like i\'m drowning slowly in a vat full of strawberry jello, but in the meantime: we have an inquiry from an indian company named Shri Shankar Sahakari Sakhar Karkhana, located in Sadashivnagar, Maharashtra. now try saying that five times fast.  actually, try saying it just once, really fast. i did and PH thought i was shushing her, which would have been rather hypocritical of me, since i\'m the office drama queen.   '),(25036,2238828,'30,November,2002','2002-11-30 00:00:00','    happy thanksgiving to all those who celebrated it. there IS a lot to be thankful for, despite the squabbling relatives, all the things that inevitably go wrong, and all the empty places that should not have been empty.  family holidays can be hard, because you want everything to be nice and normal, and sometimes it seems like no one\'s cooperating. but if it helps: no one\'s normal, and every family is slightly dysfunctional.  - - -   You better lose yourself in the music, the moment / You own it, you better never let it go / You only get one shot, do not miss your chance to blow / This opportunity comes once in a lifetime yo  - eminem || lose yourself  JP e-mailed me once, way back when i\'d first left the philippines, to ask what i thought of the latest eminem song. at the time, \"the latest\" was  stan , and i wrote JP back saying that as i didn\'t like eminem, i hadn\'t really listened to it, but that i\'d keep it on the next time it came on the radio and write him back.  two days later it came on the radio. usually i\'d listen to the first part - which was dido singing  my tea\'s gone cold, i\'m wondering why i got out of bed at all  - and then switch the station as soon as eminem started rapping. this time, for JP, i kept the station on and listened. that evening i wrote him:  i still don\'t like the song as a whole, i said. but there\'s something about the way it\'s put together - about the way dido keeps on singing  it\'s not so bad, it\'s not so bad  while eminem is screaming his lungs out. it\'s like she\'s saying  i\'m not going to let this psycho take me down, i\'m not going to let him forget that i\'ve got hope.   you have strange interpretations,  JP replied.  two years later, i\'m looking at yet another eminem song. he disturbs me because i know someone whose life was similar to his, maybe worse: his mother was impregnated by his father when they were teenagers, they got married and went to live in his father\'s mother\'s house. from the start it wasn\'t a happy story; the teenage couple was way too young for a baby and used to live it up, leaving the baby with his grandmother. he grew up calling his parents by their first names, and his grandmother \"mommy\".  his grandmother disapproved greatly of his mother, and they clashed - often. his mother knew, however, that her MIL loved the baby like her own, and would use him as leverage, saying that if her MIL didn\'t do as she wanted, she\'d take her son away. it came to a head, inevitably, one afternoon. as soap-opera-ish as this is going to sound, the argument escalated until it got physical; the boy\'s mother pushed his grandmother and she fell, hard, against a table. while the older woman was on the ground, crying and bleeding, the younger woman took her son and her things and left for her older sister\'s house. once there, she knocked on the door, left her son in her sister\'s arms, and took off for a friend\'s house.  her sister, at the time a newlywed, had no choice but to take the boy in. he was two years old at the time and so scared he couldn\'t talk, and there was no way to get in touch with his mother. the incident traumatized him so badly that for months after that, he would crawl out of bed in the dead of night, creep out to the stairs, and sit there for hours, crying for his grandmother.  that boy was my cousin. his mother is my aunt, and the older sister who took him in is my mother. his story is almost exactly like eminem\'s - at seventeen he was rapping, hoping to break into the local music scene - except for one thing. amazingly, my cousin never hated his mother. years later, when she went back to nursing school, he woke her up from a bad hangover on her graduation day, made her breakfast, helped her with her makeup, and patiently tailed her, carrying her graduation paraphernalia around at the hall while she went around greeting her friends.  i asked him once:  why don\'t you hate her? why aren\'t you angry at her for what she did to you?  because i was there for part of his childhood - we were superclose as children - and i remember his mother cursing at him and my mother shaking her, trying to get her to stop.  as a matter of fact,  i continued,  why don\'t you hate any of them?  he hadn\'t fared any better with my uncles; when my family left for thailand and my parents weren\'t around to protect him, my uncles would beat him up over the smallest things. once, when i was seven, we came back to the philippines for a vacation. my uncles pitted me against him in a reading \"contest\", and i won; i remember them jeering at him because he\'s four years older than i am. five minutes later one uncle grabbed him, dragged him out the back, and i could hear the slap of flesh against flesh, and my cousin crying.  i remember running out to the back and screaming  don\'t touch him, don\'t you touch him!  and another uncle pulling me back. when my mother came back home, i was literally shaking with anger, and told her what had happened. she went straight to my uncles, who were sitting in a corner playing chess, and told them, her voice hard:  i don\'t know how you treat him when i\'m gone, but while i\'m here, no one touches him. if i hear that you\'ve hit him again, i swear to god i\'m throwing you out of this house.   my cousin looked thoughtful for a while after i asked him that question, and then shrugged.  there would be no point in hating them,  he told me.  don\'t you see, rei? i\'d end up exactly like them. and there\'s no point in hurting yet another generation, because if i were like them, i\'d be beating up our nephew.   i saw eminem\'s  cleaning out my closet ; the snippet above is from his latest song,  lose yourself . i still don\'t like his music - not because it\'s too \"real\", but because it\'s too hopeless, too desperate. it\'s GOOD music - brutally honest and evocative of what eminem must have been feeling; you can almost smell the acridness of his sweat, see him trembling as he goes out on yet another stage, preparing to be booed off. you know his desperation, his motive: he\'s never been given an opportunity, he makes his own; and he gives his heart out at each performance because back home, he has a daughter to feed. he doesn\'t have an education, any training, any special skills, except rapping. this is all he has to put bread on his table, and if he screws up, more is on the line than just another bad night on the stage.  but it\'s hard to accept him as is, to watch him go out and rage night after night, single after single. it\'s hard to listen to him now because he\'s still at the hardest part of the journey, and i don\'t know if he\'s ever going to get to where my cousin is. i don\'t know if he wants to.  perhaps the scariest part of it all is that so many people identify with him.   '),(25037,2238828,'23,November,2002','2002-11-23 00:00:00','    what a week it\'s been.   sunday, 17 november  i fell asleep at one AM and woke at hourly intervals, sweating and my heart beating too fast. at four AM i got up and walked around the apartment, opening all the doors and turning on all the lights; an hour later i went back to bed and slept straight on until six-thirty, when my father knocked on my door. he was leaving for bangkok on a 10AM flight and needed me to drive him to the ERL (express railway, an LRT that goes directly to the airport) station. i got up, pulled on a pair of jeans, and drove him there. he was unloading his luggage when i remembered.  happy birthday,  i said, leaning across the seat so that i could greet him face-to-face. he looked a little befuddled; he isn\'t any more of a morning person than i am.   thank you,  he said finally. the final piece of luggage unloaded, he waved goodbye and walked off, pulling his suitcases behind him. i watched him walk into the terminal and then drove home, where i slept until it was time to go to church for christmas practice.  my sister, C, and i were ready to go when her phone rang; one of her friends, Z, was running away from home and wanted to know if she could stay over for a week. we drove over to Z\'s area and found her hiding, crouched down with her bags behind a car three streets down from her house. she hopped into the back seat, C stage-whispered  keep your head down,  and we made our way back to church.  is it abduction if the abductee packs her bags and comes willingly into your car? i want to know.   monday, november 18  Z confessed that she\'d never done any of the requisite crazy childhood things, like playing in the rain or climbing trees. shocked, C and i took her night-swimming in the rain to rectify the situation. later, when swimming lost its appeal, we got out of the water and out into the condo\'s driveway, where we danced like madwomen and did cartwheels in the puddles.  we ordered pizza for dinner, after which C got completely trashed on the remains of a bottle of vodka. Z and i ended up talking in the living room in hushed voices.  did she tell you why i\'m running away?  Z asked, and i shrugged.   how about you tell me in person?  i suggested, and she did.   i want her to see me,  Z said.  she makes me feel like an annoying afterthought. i\'m her daughter, for crying out loud. if she called me tonight, even if it were just to scream at me for doing something this stupid, i\'d go home in five minutes flat.    you want her to care,  i guessed, and Z shrugged, looking strangely diminished.   i want her to love me,  she said,  but i\'ll take what i can get.    tuesday, november 19  came home from work to find Z and C madly cleaning house. my fears that they had been possessed by martha stewart were laid to rest when they explained things to me: a group of friends (all male, of course) were coming over for dinner, and they didn\'t want them coming across strange embarrassing things in the house.  we ordered pizza, the guys came over (or, to be more accurate, Z and i picked them up and drove them over because they\'re visitors and don\'t know their way around KL) and we ended up playing a whacked version of truth or consequence. much insanity occurred, mostly due to the fact that the guys are french and Z is malaysian and C and i are filipino, and cultural confusion was pretty much the theme of the night.   wednesday, november 20  i came home in a jubilant mood, having effectively tendered my resignation at work. my days are numbered; as of feb 1, 2003, i will be unemployed and minus one source of stress. another of my sister\'s friends was there, cooking dinner; i sat in the kitchen and watched them while i sipped on the remains of a bottle of red wine i found in the fridge. we had dinner and girltalk, and it was all very satisfyingly bachelorette-ish.  later that night, while i slept, C and Z went out and ended up in a gay bar. they were happy when they discovered it was a gay bar; previously, they\'d been worried because they had been vamping all night and no one was paying attention. they came home, jumped fully clothed into the pool, decided that their clothes were too heavy, and skinny-dipped instead.   thursday, november 21  came home from work, went out clubbing with C and Z. it was ladies\' night in the area we usually go to, so we wandered from place to place in search of good dance music. my sister and Z wound up arguing, and we eventually found ourselves in two different places - Z partying with other friends in nuovo, me watching disinterestedly as C got trashed on chivas regal in emporium. when it was time to go home, i discovered that my car keys had dropped out of my bag; my friend PA had to drive me home to pick up the spare set, while C slept off her drunk in the arms of PA\'s friends. on our way back to emporium, PA\'s friend called him and i picked up his phone; they\'d found my car keys. we went back to emporium and PA\'s friends dragged my sister out; while i was supervising the logistics of getting her body into the car so i could drive us home, the door manager hit on me and offered to let me in free the next time i came.  you\'re so sweet,  i purred, because there was no sense in being impolite, and went off to get my car.  C woke up briefly while i was rolling her into bed.  i\'m not drunk,  she said, her breath heavy with whisky.   sure you\'re not,  i replied.  that\'s why i just dragged you all the way from the car to your bed. i hope you feel like absolute hell tomorrow, you drunk.    friday, november 22  spent the day at work just trying to stay awake; at lunchtime i gave in to the inevitable and spent the entire hour at my desk, sleeping. PH told me later that our boss had come in, given me a strange look, then walked back out. not that i gave him reason to complain, because everything he wanted done that day was done, whether or not i had to pinch myself awake at the computer to do it.  and in between all of that i fell in love with KL, but that\'s another story.   '),(25038,2238828,'22,November,2002','2002-11-22 00:00:00','    according to the regional DTI office in zamboanga, one of the region\'s top exports to malaysia includes fighting cocks. there\'s a whole host of risquÃ© comments that could be made about that, but i\'m feeling virtuous so i won\'t say anything.  speaking of unsavory exports: back in college, my friends and i did a report on the philippine legislative process, wherein we tracked the progress of a bill through congress and the senate. the bill we finally chose was a regulatory act on mail-order brides - for which, apparently, the philippines is renowned worldwide. there\'s no specific link for information on this, but when we did a simple yahoo search (this was back in 1998, i think), typing in the words \"mail-order bride\" produced tons of results, most of which advertised \"pretty, obedient girls\" from the philippines. there were a lot of ads for women from russia, too, but i don\'t think the tag \"obedient\" was attached to them.  at any rate, we sifted through endless pages of committee discussion minutes and session minutes and papers for the justification of the bill. there\'s one quote from a house member that stands out in my mind, whenever i remember that paper. frustrated by the length of time it was taking to discuss the bill and all the minutiae involved in legislative processes, he blurted out:  according to our statistics, our top exports are bananas, pineapples, and women. isn\'t the fact that our women are considered commodities justification enough for this bill?   the bill was passed eventually, although i forget the exact RA number. my memory is crap these days - i canâ€™t even remember whether the bill outlawed the sourcing of mail-order brides, or just enforced stricter regulations on the industry. more likely the latter, since after all it IS a source of income. make no mistake, i hate the idea, but i understand the rationale. itâ€™s the same rationale that squatters in the philippines use to justify their theft of public property - weâ€™re poor, we canâ€™t afford to have such high moral standards. my dad refers to it as the poverty mentality, and theorizes that unless people break out of that kind of thinking, they wonâ€™t be able to effect any sort of long-term positive change.  i actually met a mail-order bride once. i was in NAIA, waiting for the flight that would take me to KL for the Christmas holidays. being antisocial, i plunked myself down in an isolated stretch of seats, put my bag on the seat beside me, pulled out a book and started reading. since people generally donâ€™t bug strangers who look like theyâ€™ve got something better to do, no one bothered me for quite a while. then a woman - thirtyish, with shoulder-length hair and a pretty, anxious face - sidled tentatively into a seat a few spaces down from mine. she sat down, fidgeting all the way; first it was her bag strap, then it was her hair, then she opened her bag and started taking out items and repacking them; finally, she left all of that alone and leaned back, rubbing her ticket with her thumbs and fingers.  at that point, i finished my book and put it away. i was fishing in my bag for another one (i always have at least three books handy whenever i travel) when the woman got up and moved nearer.  excuse me,  she said, smiling hesitantly.  what flight are you waiting for?    mh 704,  i said.  heading to kuala lumpur for the holidays. you?     iâ€™m a transit passenger,  she said, grinning with shy pride and showing me her ticket.  this is my first flight. iâ€™m going to germany.  there was a slight pause. i stared at her ticket wondering what i was supposed to say, and she rustled her ticket encouragingly.  iâ€™m going to get married,  she added, when i failed to respond. probably guessing that i was lost for words, she hastily dug around in her bag and produced the ubiquitous bundle of pictures that most filipino females carry around in their wallets.  as she flipped through the pictures, looking for a snapshot of her sweetheart, she bubbled out her story: a friend of hers had been writing to a man from germany, she told me, and that man had a friend who was single. her friendâ€™s boyfriendâ€™s friend had gotten her contact details and written to her, and they had been corresponding for a year.  there,  she said proudly, bringing out a photograph.  doesnâ€™t he look like jean-claude van damme?   i just blinked at her.  have you ever met up with this man before?  i asked, feeling infinitely older than she was.  or have you just been communicating through mail?   she looked at me, uncertain, then turned back to the photograph, rubbing it gently with her thumb.  we wrote each other,  she explained.  he was very nice. last year he came to the philippines for two weeks to help me get a fiancÃ© visa to germany.  there was a short pause, and she added,  he was kind to my parents.    he sounds good,  i said cautiously.  but are you sure? do you know anybody else in germany?   she shook her head, returning the photograph to her wallet, and her hands trembled a little.  just him,  she admitted.  he says heâ€™ll meet me at the airport. and weâ€™ll get married.   i didnâ€™t quite know what to say, after that. there was an awkward pause before they opened up the gate and they started letting passengers board our flight. i didnâ€™t see her after that, but i thought about her all the way to KL, and told my mother about her that night.  how can she do that?  i asked, confused.  stake her life and her future on maybes? this guy could be abusive, or sell her, or use her as a slave, or whatever, and heâ€™s the only person she knows in germany. how can she take chances like that?   my mother shrugged.  she probably doesnâ€™t think she has a choice,  she said, and told me about what life had been like for her as a girl, in a village where they had to ford streams and hike across mountains in bare feet just to get to school, where the children were given a holiday when a jeep came into town because no one had ever seen an automobile before, not even the teachers.  sometimes,  my mother explained,  you just want to get out so badly that you donâ€™t think too hard about what youâ€™re doing to get out, or what youâ€™re going to do after you get out.    i donâ€™t understand that kind of desperation,  i told her.   youâ€™ve never been there,  she answered.   i guess thatâ€™s what motivates most of them - the mail-order brides and the domestic helpers and the women we refer to as  japayukis  and  brunei beauties  - the desperate need to get out of that rut youâ€™re stuck in. in  anak , vilma santos asked her friend,  the first time we leave, itâ€™s because of the money. the second time, the third time â€¦ itâ€™s pretty much still about finances. we need the money, so we go and look for work somewhere else. but what about the fifth time, the sixth time? is it still money that motivates us?  implying that in a sense, they werenâ€™t leaving just to make money; they were leaving because they were running away, escaping a life in the philippines that was too difficult for them to accept or deal with.   thereâ€™s something incredibly strange and twisted and sad about a country that puts on a stoic face and accepts the barter of its women as an economic necessity. yaz and i used to write each other fast and furiously, indignant at the way Filipino women were viewed as toys or pets or slaves in different parts of the world. but when you think about it - in a way  (in a way)  itâ€™s partly our fault, too. not that it justifies the way people mistreat our women, but then - can you really blame them, when we turn a blind eye to the way they sell themselves because we need the money? someone once told me that i couldnâ€™t expect people to respect me unless i respected myself, and itâ€™s becoming clear that the same thing applies to countries and nationalities as well.   '),(25039,2238828,'12,November,2002','2002-11-12 00:00:00','    dear M,  i love you.  do you know why i started my letter out that way? let me tell you, before you start filling out my application forms for UP sappho (which i am not eligible for anymore, at any rate, since i am no longer a UP student): because only for someone i love would i download over 400KB of useless forwards. M! did you not get the memo? dancing pigs and brad pitt\'s unmentionables are NOT acceptable forwards! (the dancing pigs were cute, but brad pitt ... oy. i\'m traumatized for life, M, how could you.)   i shall be forgiving, however, and chalk it down to the fact that you have not yet found the pork chop that you will eventually share your life with and therefore have too much time on your hands. don\'t worry, i haven\'t found my african fish, or my tree yet, either. we will someday, i suppose. in the meantime, there are many many alternative pork chops on the loose for you, and i am haplessly falling in love with various and sundry trees and giving my heart to none. we\'re young, M, we have time to dally. when we get too old to open beer bottles with our teeth, then perhaps it will be time to panic ... but not yet. not ever near yet.  speaking of fish and trees and pork chops and destinies, i see that A has found hers. or, more accurately, it has found her: only a man who could order A to sit HERE and get away alive, deserves her - such strength of character, such bravery and derring-do, such utter recklessness, must be properly rewarded. i laughed until tears ran at your description of their love affair:  worthy of a dime novel,  you said, and i believe you.   at any rate, i suppose this is the grandiosely stupid thing she e-mailed us all about months back? she sounded slightly dazed back then, stunned and upset and knocked out of her common senses, and come to think of it, it did cross my mind that she might have fallen in love. A is the queen of control freaks, no matter how helpless she may try to appear at times. and love, no matter what you think about it, is one of the things that makes things spin out of control until you can manage to catch your breath again and step back a little. the first few seconds of free-falling are always scary.  i\'ll do as you told me and stalk her endlessly until she gives up and tells me the whole story. while i\'m waiting, though, i want to ask you: what about YOU, M? only you, of all my friends, would write me pages and pages of letter and never mention herself. all the news is lovely and more than welcome, you know that, but ... A and V and the lives of the other members of our group aside, what\'s going on with you? where have you been hanging out? who rocks your world (aside from avril lavigne, eewww, M)? have you made friends with any new cars, and have you seen joseph around lately? if you have, kiss his fender for me, because i miss him dreadfully. there are joseph-shaped voids in me, M, and every time i drive with the wind in my hair i have the sense-memory of his steering wheel between my fingers, his reluctant clutch beneath my demanding foot. if cars could dream, he dreams of us, and i\'m sure he treasures the marks we left behind: the oil and sweat from your hands as my reckless driving slapped you hard against the dash, the holes burned in his upholstery by bitter cigarettes, the crumbs of potato chips ground into the floor.  and even more than i miss joseph, i miss A, and i miss V, and i miss you. i never think of manila that i don\'t remember us balancing on the rooftops of intramuros, or being bitter along manila bay, or going the wrong way down those godforsaken one-way streets. i never see water, or rivers, that i don\'t remember driving through marikina late at night, with the concrete glistening from rain and the reflection of street lamps, and talking of things that needed darkness to be told. i miss the time V and i kidnapped you from your quiet night at home to drag you, clad in jeans and a t-shirt, to our org\'s formal sit-down dinner. you sat down next to karlo, i think, elegant and comfortable in your casual clothes, while V and i, formally-clad, attacked the buffet table with all the finesse of wild animals.  and i remember that you never talked about yourself much. you used to allow us to prattle on and on, filling aeons with our useless chatter. i was particularly guilty, ambushing you at odd moments to cry on your shoulder over some new trauma, and you listened. M, you will never know how glad i am that you gave up your dream of becoming a nun (even a carmelite one because you liked how they lived in the scent of flowers) because there was too much silence inside you to bear that much stillness outside. maybe in some strange way our noisiness balanced you, but M, i wish you\'d spoken more. in those short flashes of time when you told me your stories, when you were vulnerable, when you cried - i felt more your friend than in all the times i came running to you and expected you to pick me up.  i have a penchant for saying the most horribly inappropriate things at the worst times, but M, sometimes i have to ask these things. are you still convinced that you\'re not enough? does that part of your past still make you feel as if you don\'t deserve to be loved? of course we told each other that they were all wrong, that no matter what anyone once told us when we were too small to defend ourselves, we were indeed worth it. but it takes longer to convince the heart, doesn\'t it? i know you were never comfortable with loving, not the way V and i were - but then again, perhaps V and i aren\'t very good examples, since we\'re shameless. and i know you never thought you were beautiful, and that you thought we were all lying through our teeth when we said you were, but M, you are. cross my heart, honest to goodness, you are.  the thing is, i\'m kind of afraid that you\'ve given up on things like falling in love and being happy, but M, you really shouldn\'t. you sounded so happy for A - but don\'t you remember what she had to go through first? it\'s not like there are instant happy endings, M; and if there are no happily-ever-afters, it stands to reason that happiness is not a one-shot deal either. you sound like you\'re settling for second-hand happiness - as if you believe you\'ll be happy as long as each and every one of us is happy, but oh no, M, you\'re not getting off that easy. no. i won\'t let you; you MUST have your own share, and i insist - i will absolutely kick your butt if i must, just to get you to go and  carpe  each  diem  to your own satisfaction.  M, you always believed so strongly that we\'d each have our good life, above and beyond the aristotelian ideals described in our philosophy books. i want you to believe again - for yourself, and not just for me, or for A, or for V. we\'re going to find our own ways, M, never worry. it\'s time for you to find your own.  you WILL find your perfect pork chop one day, be it man or situation or career or whatever, and you will deserve it with every cell in your being. and in the meantime, you don\'t need that destiny to complete you, because you are complete. holes and warts and jagged places included, M, believe me. i told you once that sometimes i thought i\'d been given the childhood i\'d had to toughen me up, because without it i would be spineless and afraid all the time, not just some of the time. you\'ve got to learn the same - that everything you\'ve been through, all the wrongs and the rights - they come together to form the you that is today. not that i\'m excusing the wrongs, or validating them, because wrong is wrong, but - you can choose to reject the bad parts, to say  i will use you to make myself stronger, so there, i still win!   and even if you don\'t find it, M, the journey\'s worth it.  before you ask, i\'m fine. and no, i haven\'t found my tree yet. or perhaps i\'ve found it, but there\'s just too much foresty goodness surrounding me to concentrate on one tree, so my tree will have to wait. but yes, i keep myself open to hope and possibilities - and so should you. remember, M.  to joseph and tweezers and cheesy tagalog movies,  rei - who would like to remind you of that cruel tree in the lagoon that she loved hopelessly for years, simply because it was what it was. (you are what you are, M, and if i could love a heartless tree, you can imagine how infinitely more plausible it is that you will find someone who will love you because of that.)   '),(25040,2238828,'11,November,2002','2002-11-11 00:00:00','    it\'s been raining hard in KL lately, and it\'s been ramadan for a week. not that the two are related, although i must say - if i had to fast, i\'d much rather fast in a climate that was cool and fresh than one that was hot and steamy. past ramadans in KL have felt as if heaven was trying to melt us down into puddles on the sidewalk, and it must have been miserable for those who were fasting.  ramadan is an exercise in humility, as i understand it - a month wherein all muslims fast to understand what it is like to be poor and hungry, and thus learn to have compassion on their fellow men. when they say \"fast\", though, they mean it: none of the lenten no meat for a month kind of deprivation. from sunrise to sunset, they neither eat nor drink (not even to swallow their own saliva), do not partake in any vices, and strive to keep mind and body pure.  the fun part of ramadan is  buka puasa , the malay term for breaking fast. not just because of the patently obvious, but because  buka puasa  has come to mean family time. families that normally miss each other due to busy schedules now find time to come home and break fast together. friends make it an excuse to catch up with one another:  let\'s  buka puasa  together later.  eating becomes more enjoyable, more precious - not just because it is something you have forgone all day, but because it becomes an event in and of itself, something anticipated and planned for and celebrated when it happens.  the month-long fasting ends with  eid-il-fitr , which is when everybody goes  balik kampung  - back to their home towns - to celebrate  hari raya , which is their equivalent of new year\'s day. another time i\'ll explain the malaysian concept of new year\'s day - the malays, chinese and indians celebrate it separately, but with very similar and very lovely traditions.  this is true islam, i think, which is why malaysia gets so defensive when it gets labelled a terrorist country. you\'d have to come here to understand, but KL is gentle and innocent in a way that you would not believe a major city could be. where else in the world could you hold open house and not have your home stripped of its belongings within five seconds flat? (aside from the boonies where everybody knows everybody, natch.)   '),(25041,2238828,'06,November,2002','2002-11-06 00:00:00','    this is  not  the report i eventually submitted to my boss. this was my take on it, and i\'m posting it here for fun.  all names have been changed to protect the moronic and extremely annoying.  - - -   POST - ACTIVITY REPORT  TITLE : PHILIIPPINE BOOTH AT TE 2002 DURATION : Sometime in October 2002 VENUE : ABC Exhibition Center, Selangor    I. Introduction   There was an exhibition. We participated. Not much more to it, really.  Oh, yes. This exhibition had live animals and food, and had we had time we might actually have had a chance to go around and snack. But we didnâ€™t, moreâ€™s the pity. We were actually looking forward to breaking our diets, for once.   II. Highlights    Preparation  The Philippine booths at the TE 2002 were offered on a complimentary basis by the organizer of the event, XYZ Sdn. Bhd, which was probably the only reason why there WAS a Philippine booth in the first place. If the Department of Trade and Industry is perpetually lacking for funds, suffice it to say that the Department of Agriculture, apparently, is completely bereft. I mean, if you canâ€™t make money through wheeling and dealing, who said you could make it grubbing around in dirt with fruits and vegetables? Take note: this comes from the granddaughter of a farmer. I know whereof I speak.  Information on the event and possible participation was forwarded to the Department of Agriculture for consideration. After much agonizing (oh dear god, weâ€™re going to have to SPEND MONEY on air tickets!), they decided that the debilitating pain of expenditure would just have to be outweighed by the off chance that people just might be interested in Philippine products and buy them, offsetting the money theyâ€™d laid out. One of the departments in the DA coordinated Philippine participation, which eventually consisted of two - count â€˜em, TWO! - people. Assistant Director JOSN of DA - a short, rotund man who looked like Santa Claus with a mother of a tan - headed the Philippine Delegation, such as it was.  The DA provided promotional and decorative materials, product samples and display items, with supplementary promotional posters sourced from CITEM. Mr. RC, the participant from the private sector, brought product samples and catalogues from his company. The original box of promotional / decorative materials was stolen at the ABC Exhibition Center because the trade assistant (oh I wonder WHO it could have been, [/sarcasm]) went temporarily insane and left it behind to look for a trolley. Thatâ€™ll teach us all to believe in the inner goodness of mankind. PTIC-KL then provided brochures, pamphlets, and decorative material for the booth, because otherwise it would have been as barren as the Sahara Desert and twice as ugly.   Philippine Booth (Booths No. 525 & 532)  There were ten million eighty-six booths there, and our booth was one of the most interesting. Well, no - â€œmost interestingâ€ would have had to go to the booths with carved wooden gargoyles, or the big-ass Ministry of Health booth with the light-and-sound show, but hey, you take what you can get. Our booth could have carried off â€œMost Artisticâ€, though. Minimalist with a splash of opulent red-and-gold sinamay, with rotting fruits nestled daintily in between brochures of the Philippines and Mindanao hand-woven silk.  Oh, sorry, forgot: we mustnâ€™t mention that the fruits were rotting. Because if you didnâ€™t look too closely, they really didnâ€™t look like they were. Honest.  A lot of people visited the booth, although whether it was due to the kick-ass, last-minute decoration job or the fact that we were giving out free samples of food is a matter of pure conjecture. Considering the fact that we have already posited the absolute loss of our belief in the innate goodness of humankind, most likely the latter - because despite what Maslowe thinks, people who can afford to eat three times a day and satisfy all of their basic needs still tend to go mad at the sight of free stuff. Is it the Darwinian need to survive in a harsh world, or are we all just innately cheapskates?  Weâ€™ll leave that to the anthropologists and genetic scientists to figure out.   Decoration  Papayas that were mushy at the ends, mangoes that had been slightly bashed by the trauma of being transported across the ocean, mangosteens that had no identifying marks or characteristics to differentiate them from Malaysian mangosteens (which are, after all, the exact same species). A booth held together with tape and a prayer, various posters that depicted artificially delicious-looking fruits and vegetables (bet you didnâ€™t know asparagus could look that luscious) and odd kinds of seafood. Apparently the Philippines produces black shrimp nowadays, isnâ€™t that amazing?  It sounds bad, but it actually looked okay. Iâ€™m not just saying this because Iâ€™m one of the interior decorating team.   III. Incoming Inquiries   There were inquiries?    IV. Overall Assessment   I have nothing to say except next time, there damn well better be FUNDS for transport and all the small logistical details. Weâ€™re human beings, people, not Irona the robot with super-strength and rechargeable batteries.   - - -  for this, i got my bachelor\'s in political science and slaved through my MBA.  life sucks sometimes.   '),(25042,2238828,'05,November,2002','2002-11-05 00:00:00','    to those who don\'t know narnia: read the backgrounder first. to those who do know, and remember it as a semi-dark fairytale but don\'t want to take it farther than the happy ending, you might not want to read my recs.  the backgrounder:  urlLink narnia  is an alternate universe, one filled with mythical creatures and talking animals and Magic. one day it crosses over with the real world: a girl named lucy, taking an enforced summer vacation in the backwoods of nowhere with an absent-minded professor and her three older siblings, steps into a wardrobe to breathe in the smell of stored fur coats, and comes out the other end into another world.  at first the world is beautiful, ice-cold and pure white. she runs into mr. tumnus, a faun, who takes her home and gives her tea and sandwiches. they talk and talk and talk, and in the end lucy realizes that hours have passed and she must return to her world, because her brothers and sister will be worried. she goes back through the wardrobe and realizes that no one has missed her. time moves differently in these two worlds, and the hours she has spent in narnia translate into mere moments in her world.  the second time she goes in, narnia lifts the veil to show its dark side: mr. tumnus is gone, arrested by the white witch, an evil queen-wannabe who has put narnia under an eternal winter. lucy learns that he has most likely been taken as a result of her visit: she has spies everywhere, whispers mrs. beaver who is, you guessed it, a talking beaver.  meanwhile, lucy\'s brother, edmund, has also discovered narnia. however, instead of meeting a good guy, he runs into the Purveyor of All Evil in narnia: the white witch. she terrifies edmund until he confesses that he came from another world, whereupon she turns around and begins to seduce him with turkish delight and promises of co-rulership over narnia.  tell me about yourself,  she says, and he does: about his school, about how miserable his school vacations are, about his sisters and his brother. she is particularly interested in his siblings, and proposes that he bring them over to her castle for tea.  edmund and lucy return to the wardrobe at the same time; lucy is overjoyed to see that edmund has visited narnia. the last time she told her siblings about narnia, they thought she was making things up, as she is prone to playing games of make-believe.  now they\'ll have to believe us,  she tells him, referring to their skeptical older siblings, peter and susan. however, edmund betrays her, a foreshadowing of things to come: he tells peter and susan that it was just a game he was playing with lucy, and he hadn\'t expected her to take it quite so seriously. lucy runs off in tears, and peter turns on edmund, chastising him for being so harsh on their youngest sister.  eventually, all four of them make it into narnia, whereupon edmund inadvertently reveals that he has indeed been to narnia before. peter tells him off for lying, and the seeds of resentment are planted: edmund decides that when he goes to see the white witch, he won\'t bring peter, since the white witch promised to make them all princes and princesses, and peter doesn\'t deserve to be a prince. lucy leads them to the beavers\' house, and over a dinner of fish (naturally), mr. beaver reveals the reason behind the white witch\'s interest in the four of them.  in narnia, he says, there is a prophecy: the white witch will rule until adam\'s flesh and adam\'s bone - two sons of adam and two daughters of eve - bring back aslan, who will break her rule and end the eternal winter. at this point, edmund - who realizes that his alliance with the white witch might not be viewed in a positive light by his brother and sisters (who are already quite put out with him) - slips out to secretly meet her.  eventually the four siblings are reunited and narnia is saved with the help of the mysterious aslan, who turns out to be narnia\'s creator in the form of a huge lion. the four go on to rule narnia in what will be known in its history as the golden age, with King Peter the Wise and Queen Susan the Beautiful and King Edmund the Just and Queen Lucy the Kind-Hearted. one day the kings and the queens are out hunting, and stumble across a wood; dismounting, they explore it and find an eerie light that seems to come from nowhere.  exploring it further, they suddenly fall through and find themselves back in their old world - in a pile of fur coats in the back of a wardrobe. they have all returned to the ages they once were.  the book ends ambiguously: like the characters, you are unable to decide whether the entire thing was a psychotic episode, or for real.  i received the entire narnia series as a birthday present on my ninth birthday, and read through them in a week. on the surface they were nothing more than moralistic fairy tales, something like a modern-day pilgrim\'s progress, which is most likely why my mother bought them for me. but they haunted me the way grimms\' fairy tales did (with undercurrents of something terrible in between the once-upon-a-time and the happily-ever-after), and i used to return to them, trying to figure out what disturbed me.  in COL, teacher anna made us read a retelling of snow white and rose red -  because fairy tales aren\'t for children,  she said,  not really.  there was death there, and manipulative mothers, and slavery. in the original red riding hood, the girl was raped and there was no woodsman to save her.  and with that class, many of the things i\'d thought about and felt and sensed, crawling beneath the pretty pictures, were given a name. they weren\'t pretty names, but they were names nonetheless, and i stopped being afraid of red glowing eyes in the dark because now i knew who they were.  the same feeling of recognition seeps through in these fics. there are two of them:  urlLink the queen in exile  and  urlLink girls grow quicker than books , which both deal with the aftermath of returning. how does a mind deal with the expansion and shrinkage of time, and the crossover and interlapping of lives lived? if you\'ve read the entire series, the latter fic has a small tie-in to susan\'s eventual exile from narnia.  the moral of the story being: there are no fairy tales, not the way people think there are.   '),(25043,2238828,'03,November,2002','2002-11-03 00:00:00','    urlLink  deepavali  : the festival of lights, a celebration of the triumph of good over evil. technically it\'s a hindu festival, but it\'s considered one of the major public holidays in malaysia. one good thing about a multi-racial society: lots of holidays. november to february is generally considered holiday time, since all the major holidays fall on or around that time - deepavali, christmas, hari raya (or  eid-il-fitr ), and chinese new year.  but wait, we were talking about deepavali. in malaysia there are two things that symbolize deepavali - first, the lamps, obviously, since it is a festival of lights. on the eve of deepavali, lamps are lit at the doorways and pathways of hindu homes and kept burning all night; hindu belief has it that these lamps are meant as a welcome to whichever deity you wish to visit you. (and as an aside: yaz, laya, i don\'t know if you remember, but krishna\'s second name is actually the name of the hindu goddess of prosperity. if you knew this before, be kind because i only found out about this now.)  the second thing is the  kolam  - a design drawn on the ground or the floor with colored rice. it\'s an amazing thing, actually, since the design is usually quite intricate - the link says \"floral\", but it actually varies widely. last year there was a  kolam  at the KLCC - a mall near my workplace - that was the size of several rooms. it took an entire classroomfull of students from malaysia\'s premier art university - the lim kok wing institute - several days to complete it, but it was gorgeous. i took pictures, but i have yet to send them to laya and yaz like i promised.  speaking of the  kolam , i read something about them a long time ago. i\'m not sure if it\'s true, but boy, if it is ... at any rate, if i recall correctly: traditionally, the woman of the house is responsible for creating and forming the  kolam  at the threshold of her home. every evening, she sweeps it away, and the next morning she gets up and - get this - MAKES A NEW ONE. according to the thing i read, this practice was supposed to enhance the virtues of patience and acceptance, and may i just say that you\'ve got to be darn patient to do something like that every. single. day.  there\'s also another deepavali practice in malaysia: the open house. all malaysians, at one time or another, hold an open house - hindus at deepavali, christians at christmas, muslims during hari raya, and the chinese during chinese new year. this is the amazing thing - open house means, quite literally, that their house is open to everyone. strangers are welcomed in and offered food and drinks and hospitality; this is practiced all over malaysia, from the most rural of villages to the urban center of kuala lumpur. even more amazing is that people are generally respectful of the houses they enter; i don\'t think i\'ve heard of any cases where people who had an open house were stolen from.  i don\'t think we have an equivalent in the philippines. the closest i can think of is  noche buena , but then again, your house isn\'t really open to everybody - just to friends and family. that, and nobody\'s really up for visiting anybody except friends and family on  noche buena , since it\'s considered to be a family occasion.  and then again, just imagine open house in the philippines. your house would be stripped before you could say  merry christmas .   '),(25044,2238828,'02,November,2002','2002-11-02 00:00:00','    i\'ve been foolish, these past few weeks. partially because the last couple of weeks have almost literally been hell: the hostage-taking in moscow, the sniper shootings in washington, the bomb blast in bali, and the tension over iraq; partially because my own life was sucking, different parts of it in varying degrees of suckiness.  then there were the aftershocks: demonstrations and rallies in jakarta, because abu bakar\'s followers think that his being placed under police custody is a matter of indonesian democracy being steamrolled by western prejudices; a picture of a man standing at the bomb blast site in bali, his hand over his mouth and tears in his eyes because his 18-year-old daughter had died there; and then bomb threats upon bomb threats upon bomb threats, until you get tired of being afraid of them and start thinking  if you\'re going to bomb us then DO it already, dammit, because this wondering and cowering is driving me mad.   although i\'m not quite sure that i didn\'t go a little mad, anyway. my inner control freak didn\'t want the world to go to pieces like this, not without some sort of explanation, and i wanted so badly for things to make sense. but some things defy consolidation, or rationalization. there are some things that you can\'t tidy up, or sweep around, or rearrange, and that\'s only right because they are what they are. in the end i had to stop being stupid and selfish, stop trying to make things sensible and sane, and just sit back and let them be. how did i put it once? not to internalize the pain, because it wasn\'t my pain; but to acknowledge it, to sympathize, and to look for ways to help.  at the same time, i had to realize that the flip side applied as well; when it IS indeed my pain, then i have to hurt for it. i can\'t put it aside, or hide from it, or redirect it. and i certainly can\'t create a persona to gatekeep my memories, to validate ones that had slipped beyond my reach and to stand as a buffer between me and things that i chose not to remember. it sounds strange and psychotic, yes, but well, the things you do to deal with things that still have the power to make you break out into a cold sweat in an air-conditioned room, years after they happened. G and i were talking about it once, when we were discussing the shadier parts of our pasts, and came to the general conclusion that at the end of the day, you\'ve got to do the things that keep you from cracking up and becoming the same monster that used to terrorize you.   so. like i tried to tell luis (but failed miserably because coherence does not figure largely in our correspondence),  these people may let hatred lead them around by the nose, but i\'ll be damned if i let their hatred and anger influence the way i lead my life.  that includes bomb threats, which we were discussing with much heat and incoherent frustration. and if i feel that way about global hatred and fear, then it would be hypocritical of me to not apply the same philosophy to anger and fear that is as close as my own bloodline, yes?  so, no more hiding. no more hoffe, and even though he may be an invented persona, he\'s my invented persona, so i had to write him and tell him that he would need to step away for a while. no more asking for things to make sense, or trying to force them to fit into little compartments for me, just so i can sleep easy at the end of the day.  it doesn\'t solve anything, to be honest. you\'ve still got a world where a 17-year-old survivor of the moscow hostage-taking has no words for what he feels except  urlLink  everything is perfect. i\'m alive  . and you\'ve got tim o\'brien explaining  urlLink what goes on in the mind of a war veteran , which isn\'t very comforting. but if i can\'t make sense of it then it\'s only right and fair, because if killing ever made sense then the world would definitely be wrong. and as my personal life goes, there are still things that i won\'t be able to do, just yet, because just thinking about confronting certain people still makes my throat close up - but that\'s fine. i won\'t rush.  in the meantime, i\'m taking out all the things i put away, in my mad and useless quest to control what i couldn\'t. like i told hoffe:  it\'s strange: i thought growing up meant leaving behind the things of before, but - in my case, at least - it looks as if growing up means i\'ll have to learn how to take them with me.   '),(25045,2238828,'30,December,2002','2002-12-30 00:00:00','    i didn\'t find any newater in singapore; i\'m disappointed. good golly, it\'s practically the national water, and they didn\'t have it anywhere. for shame. memorable moments in singapore:  > snarking at all the thin, fashionably-dressed girls who strolled the streets, mocking us with their beauty. C moaned  i swear, they all go to marie-france, they have the same bodies,  and i said  no, they just don\'t eat as much as we do.  because we ate. a lot.  > borders and kinokuniya, my personal meccas. i am drawn to bookstores like a moth to a flame, with much less fatal results. of course i bought lots and lots of books, one of them being  the science of God  by gerald schroeder - a look at physics and biochemistry and paleontology and various other scientists through the eyes of one who believes that all things were created by a divine Creator. next up:  a brief history of time  by stephen hawking, who claims to have the final proof that God does not exist. while i believe, irrevocably, that God IS, ever was and ever will be despite what anybody says, does or believes, it\'s interesting to look at the argument from both sides and weigh them using pure logic, and see if logic and faith lead to the same ultimate conclusion. don\'t you think?  > getting lost, endless times, in the underpasses and confuzzling mazes of streets that are the orchard road junctions with all other streets in singapore\'s shopping area. i finally fixed, as my points of reference, borders and kinokuniya - since they were the only places i could find without getting hopelessly confused.  > butterflies landing on my mother\'s hair, an anteater that ignored us in its search for food, a couple of mouse deer resting quietly under a big leaf, revealed only when my father brushed it aside, and a lazy, arrogant monkey that lay flat on its back and surveyed us through indifferent eyes as we did all sorts of antics just to get it to react.  you arrogant primate,  i huffed.  YOU\'RE the monkey in the zoo! do something!  whereupon it seemed to raise an eyebrow at me and say  but my dear, YOU\'RE the one jumping around like a trained moron.  these all took place in the singapore zoological gardens, which has - according to the taped commentary - 1,290 animals in its enclosures. is that including the human animals, i wonder?  > using the male bathroom in wheelock place by honest mistake. i wandered in, bold as you please, used the bathroom, then left the stall to wash my hands ... and froze in my tracks, because there was a guy to my left standing at a urinal. instinct kicked in while my logic was still stuttering for an explanation (because i\'d been so SURE that i\'d gone into the female bathroom), and i exited quickly before he turned around; luckily his back was to me or i would have seen a lot more of singapore than i actually wanted to. i told my mother about it and she went off into absolute fits of laughter.  mother,  i said indignantly.  you\'re supposed to be concerned for my virtue.  she just looked at me and started howling again, so i suppose my virtue isn\'t a very big issue with her. i ask you.  i had stories to tell that i can\'t remember now, so i suppose they\'ll have to be told another day. in the meantime, happy new year\'s, everybody. may you all have many hours spent in the company of people you feel comfortable with or enlightened by, many moments of humor and whimsy and delight, and many many people and objects and situations that will leave you wordless with the beauty and wonder of them. there are hard times, yes, but you never really actually SEE the good things unless you\'ve had to bleed for them, so it all works out.  and we\'ll do this all over again next next year. cheers!   '),(25046,2238828,'21,December,2002','2002-12-21 00:00:00','    i\'m sure you\'ve all been there: the really really REALLY bad date that made you swear off social outings for a few weeks; the one that you endured rather than enjoyed; the one that had you rolling your eyes, glancing pointedly at your watch, yawning - ANYTHING to offend the son-of-a-mangy-cow so much that he would cut the date off prematurely.  well, it wasn\'t my date, technically; it was my friend D\'s date, but since i was her buffer, i couldn\'t leave. it was worse than a roadside accident; at least those give you the option to look away, to not carry away memories. this date - oy, it sticks like grease to the corners of my mind.  the evening started out ordinarily enough. my sister went off to her job - a night shift waitressing at a club. D called up to ask if i was going out, since she was going out with her sister and a friend from school. i\'d been planning to go clubbing anyway - my sister works at the club that we usually hang out at, and i try to go once a week to see her and cheer her up if she\'s cranky - but was happy to know that i\'d have company.  D and i met up, early enough for me to avoid the sleazy door manager at one of the clubs. she introduced me to her sister, who was so beautiful that for five seconds, i just stared; dressed elegantly in black, and with her hair loosely twisted up, she looked like a manga drawing come to life. D is pretty too, but in a cutish sort of way - more like a refined kewpie doll. after a while, RZ - D\'s friend from school - showed up, and we gathered uncertainly at a neighboring bar. the night was still young, the music kind of sucked, and none of us really knew where we wanted to hang out yet. that, and we looked like an odd bunch. D\'s sister ML and RZ were elegant in black outfits; D and i were ultra-casual, both wearing white tank tops - although D had paired hers with jeans and i was wearing a sarong.  a few minutes after RZ arrived, her boyfriend AD showed up. now, RZ is pretty, exceptionally so that night since she was all dressed up; her boyfriend, in contrast, looked like the Missing Link and had a matching personality. even worse, he clung like a limpet to RZ, and at one point i caught his hand wandering down toward RZ\'s butt. later, ML leaned toward me and D and hissed,  is he going to die if he doesn\'t touch her? because he hasn\'t let go ONCE during this entire evening!   and D and i just shrugged; being single (ML was there with her boyfriend, too), we chose to remain silent for fear of being accused of sour-graping. we left the bar for the club upstairs, hoping that they had better music. unfortunately, they didn\'t.  after lame attempts at dancing, during which RZ tried to freak-dance her boyfriend and he stood there with a confused expression on his face, we gave up and returned to the bar downstairs. this time it was, as destiny\'s child would say, jumping, and they were playing danceable music. ML got an uncomfortable expression on her face; apparently, she doesn\'t like crowds. when RZ and her boyfriend excused themselves to go home, ML grabbed the chance to say that she wanted to leave, too. D and i bid them goodbye, then - as soon as they were out of hearing range - looked at each other and squealed. we\'re kindred spirits when it comes to clubbing; we go to clubs to dance, we like the same kind of dance music, and we love crowded dance floors.  we had about three hours to ourselves before D\'s phone beeped. it was a guy we\'d met at the club last week; they\'d swapped phone numbers and agreed to meet up again sometime. he was at a nearby club, and wanted to meet up with her. we went out to meet him, which was made slightly difficult by the fact that he kept on moving around.  where are you?  D asked him on her phone, while scanning the area.   i\'m the one dressed in black,  he answered.   we\'re at a club,  D said, exasperated.  everybody and his country cousin is dressed in black.   we finally found him, and i raised my eyebrow approvingly at D.  he\'s cute,  i whispered. i hadn\'t seen him properly last week, as i\'d been busy talking to someone else.  we went inside to dance, and that was where the evening began its downhill slide. he was a terrible dancer - arms and legs all over the place; he actually nearly hit a woman who was dancing behind him - and even worse, seemed happily oblivious to that fact.  these people suck,  he said loudly to D, waving his hand to encompass the entire dance floor.  i\'m better! i\'m the best dancer in the world!   D grimaced at me, i moved away a little to try to disassociate myself from her crazy date, and ended up brushing my hand against someone\'s lighted cigarette. ten minutes later, D tapped my shoulder and motioned toward the door, and i nodded gratefully.  we headed to a nearby cafÃ© to have a drink, and that was where the night became irretrievably bad. rashid \"call me richard\", aka D\'s date, took the opportunity to enlighten us on his opinions on life, the world, and society in general. we soon wondered why he bothered at all, since his worldview could have been summed up in four words: \"i don\'t like it\". he didn\'t like blacks, he didn\'t like whites, he didn\'t like chinese, he didn\'t like indians, he didn\'t like malays, he didn\'t like malaysians or malaysia in general, he didn\'t like schoolchildren or old people, he didn\'t like the way some of the women at the club dressed, he didn\'t like the way the man at the neighboring table dressed.  to add insult to injury, rashid kept on forgetting my name.  what\'s your name?  he asked, for the fifth time, and i repeated it through gritted teeth. D, to my right, was pretending to fall asleep - she actually put her head down on the table - and we had both ceased to regard him as a part of the conversation. he was not one to take a hint, however, and insisted on staying with us until 4AM. when he finally departed, rashid turned to me and said  goodbye. and your name is rei.    very good,  i replied, my voice dripping with sweetness.  goodbye,  wretched,  have a nice trip home.  beside me, D cupped her hand over her mouth to keep from giggling. when the car door closed behind him, D and i screamed with relief.  my sister\'s shift ended at about 4:30; we were sitting around talking and waiting for my mom to pick us up when another man walked up.  excuse me,  he said, his voice heavy and accented.  do you believe in fate?   after we confirmed that he had, indeed, said \"fate\" and not \"faith\", he continued. he had been in a car with friends, he said, when he had seen the three of us talking. something had struck him, had moved inside him, and he had felt compelled to come to us, to introduce himself. he had no bad intentions, he only wanted to be friends, he wanted to know if he would ever see us again.   um,  my sister said, after a brief second of silence during which the three of us merely stood there looking at him, trying desperately not to laugh.  i work here, and my friends usually hang out here, so if you ever come here again, seeing us is pretty much a given.    but may i not contact you?  he asked.  i will be leaving for portugal for christmas, and i will return only next year.    okay,  i said.  just come to the entrance bar and tell them you\'re looking for rei. they\'ll pass the message on to me.   he produced a pen, and asked me how i spelled my name; when i did, his face lit up.  rei!  he said excitedly.  do you know what this means in my language?   i had an idea, thanks to the fruitcake discussion inspired by  urlLink SOMEONE  who shall remain unnamed, but gave him a wary glance anyway.  if it means something obscene or unimaginably horrible, i said primly, i\'d really rather not know.    it means king!  he told me, and i grinned, taking it as a portent: i am destined to rule the universe. or a smaller, albeit significant part of it, at any rate.  again, we waited until he was out of hearing range before bursting into wild howls of laughter.  first it was wretched and his \'i hate everything but you\', and now it\'s this nut with babblings of destiny,  i said.  whatever happened to \'hello, my name is X, may i buy you a drink\'? must all pick-up lines be indescribably bizarre? is it them, or us?   and then again, clubbing in KL is surreal, so maybe that\'s it.   '),(25047,2238828,'11,December,2002','2002-12-11 00:00:00','    as per yaz\'s request and for  urlLink your  edification (feel the love, G?), i\'ll try to recount what goes on during philippine new year\'s. i\'m not sure i\'ve got everything down pat, since my family has a tendency to take philippine traditional practices and warp them to taste, but i\'m sure yaz and laya will chip in to fill in the details. so.  new year\'s in the philippines starts two days before, since december 30 is bonifacio day - the birthday, i think, of andres bonifacio, one of our national heroes. (although i\'m not quite sure what he\'s a hero for, if the current debates are to be believed.) since it\'s a national holiday, people are out and about, celebrating both the tail end of christmas (because christmas lasts for about three months in the philippines) and the oncoming new year. there is, of course, an abundance of food and drink - no celebration in asia is complete without everyone going home at least five pounds heavier - an oversupply of sparkly stuff and assorted pyrotechnics, and endless radio countdowns.  on new year\'s eve, the pyrotechnics start as soon as twilight starts settling in. they would start sooner except for the fact that sunlight dilutes or completely erases the effect of fireworks, and part of the point of setting off fireworks, of course, is to show your neighbors how wealthy you are. your social status can be determined from the type of pyrotechnics you set off for new year\'s: the poor, or the ultra-cheap, stick to puny little watusi (tiny little matchstick-things that you scrape against the ground until they start crackling and dancing) and sparklers; the more affluent will have long strands of firecrackers and catherine wheels. the ritzy, or those who are willing to forego a year\'s worth of food for one night\'s worth of fireworks, go all out and hire professional fireworks experts (or whatever they\'re called) to set off massive pyrotechnical displays above their house.  all of this happens, of course, despite the fact that technically it is illegal to use fireworks without a license in the philippines. but hey, it\'s the philippines, you don\'t need a license to do anything. including driving, but that\'s another point entirely.  so. people gather up their illegal fireworks, risk their various appendages, and prepare to celebrate the new year. i don\'t know how other families do it, but at our house all the radios are on - each one on a different station, depending on the musical preferences of the people in the house; my dad actually had his radio on the CLASSICAL station. for new year\'s. at around nine or ten PM, the countdowns start - top 100 songs of the past year, the best and worst songs, the cheesiest songs, the most romantic songs, what have you.  there are, of course, several superstitions surrounding new year\'s, and your activities on new year\'s will vary according to which one of them you believe. there are some who believe that whatever you\'re doing when the clock strikes midnight will reflect your fortunes for the rest of the year. one local singer, apparently a believer in this superstition, had her father play every single one of her songs in the hours leading up to midnight, in the hope that they would all hit the top of the charts in the upcoming years. (it was roselle nava, if you\'re curious.) others believe that jumping at the stroke of midnight will cause you to grow taller; my sister, desperate to grow past the 5-foot mark, was seen hopping like a rabbit in the minutes leading up to midnight. there\'s another superstition to the effect that new year\'s eve is when all the bad spirits start to roam the earth, and the only way to send them packing is by creating an unholy racket - apparently our bad spirits are wusses, since they\'re scared away by noise. i used to assume that this last superstition was derived from the chinese, but PH is chinese and she assures me that they have no such tradition. huh. well, i can\'t see the spanish or the americans propagating such a theory, and the japanese weren\'t around long enough to make a lasting cultural impact, so i suppose it\'s a native practice.  while we don\'t necessarily believe in superstitions at our house, we\'re up for any excuse to be noisy. at the stroke of midnight, my cousins and i start screaming and banging things - pots, pans, anything metallic - and turn the radios up full blast. this would normally bug the neighbors, but on new year\'s eve any noise we make pales in contrast to the sounds of fireworks going off and horns honking and other people yelling, up and down the street.  by the stroke of midnight, the air is black with smoke and gunpowder, visibility is low, and your ears feel permanently damaged by the noise. the celebrations start to pall two hours after new year actually hits, and that\'s when people head indoors to eat and make merry.  have i missed anything out?   '),(25048,2238828,'30,January,2003','2003-01-30 00:00:00','    E gets major points for knowing that the proper way to bribe me is with a box of chocolates. all right, amigo, next time i won\'t \"misplace\" your hotel registration forms (although you must admit, you ended up with a pretty nice hotel room anyway. tsk. o ye of little faith). i make no such promises for your identification badge, because if i stopped harassing you, things would cease to be interesting. and heaven knows that in our line of work, things NEED to be made interesting.  lord, i\'m tired. there really aren\'t enough hours in a day to cram in everything i want to do. right now i\'m about three days behind in terms of sleep, and if i\'m not careful i\'ll sleep the entire chinese new year away.  oh yes, it\'s chinese new year; you could hardly miss it. it\'s a big thing in KL - the chinese may be a minority ethnic group, but they know how to make their presence felt. for one thing, they own most of the businesses (as in the Philippines, the chinese are one of the main driving forces behind the malaysian economy) so when chinese new year comes around and people start taking leave to celebrate, KL begins to feel like a ghost town. starting tomorrow, entire offices and malls will begin to shut down and will not return to normal working hours until next wednesday, or later.  on the non-economic side, everything\'s draped in red, the auspicious color, and oranges are the fruit of the day. it\'s a custom to give out oranges during chinese new year, and so far we\'ve gotten so many oranges that pretty soon i\'m going to start bleeding Tang. the FedEx guy came today to pick up a document and i gave him an orange; he accepted it with a rueful grin and told me that it was his tenth orange for the day (it was only 10 AM).  good,  i said firmly.  more wealth for you, less oranges for us. because frankly, i\'ve had enough vitamin C to ward off colds until armageddon arrives.  apparently oranges are symbolic of gold, since the word for orange -  lo kum , or  teochew kum , is similar to the word for gold, which is also  kum . hence the giving and receiving of oranges is supposed to be symbolic of the circulation of wealth. it\'s also cheaper; technically married couples are expected to give  ang pow , or small red packets containing money, to the unmarried (more reason not to get married). if you can\'t afford to do that - taking into consideration the fact that tradition demands that you place an even number of bills in the envelope - then you give oranges.  restaurants are offering chinese new year specials, namely  yee sang  - raw fish salad that you are supposed to toss with your chopsticks as high as possible; the higher you toss it the more successful your year will be. (note: toss as high as you like, but keep it on the plate or the waitresses won\'t be very happy with you.) mcdonald\'s is doing its part to contribute to the phenomenon of globalization in terms of the breakdown of cultural barriers, and is offering what it calls the prosperity burger. i don\'t understand what exactly is so prosperous about it, but it\'s good. it also comes with orange juice, sigh. at this point it would be nothing short of a miracle if anyone in KL came down with a cold.  at any rate, i have the entire weekend plus monday and tuesday to recuperate. office resumes on wednesday, and by then - by golly, i\'ll have caught up on all my sleep.  or not, since i remember that i\'m supposed to go clubbing for the first time since new year\'s this friday and this coming monday, to make up for all the weekends i have not been going to bug my sister at her workplace.  moan. i would duplicate myself if i could, but (a) that isn\'t scientifically possible at this point and (b) the world doesn\'t really need another me.   ---  addendum to the Saga of Oranges, because it isn\'t over yet: just as we were preparing to leave the office, our travel agent called and asked me to come down. i presumed he wanted his check, so i went downstairs and took the check with me. once i got there, he took the check, thanked me, then opened his car and gave me TWO BOXES OF ORANGES.  which, added to the ones we already had in the office, made three boxes and a big bag full of oranges. since PH had already gone home, there were only two of us left to divide the oranges between ... and i was getting ready to sneak off and leave my boss with all of them, when he called me and dumped a box and the bag into my hands.   curses, foiled again.  so when my mother pulled up in front of my office to take me home, i hopped in, box, bag and all.  holy crap,  she said, quite inelegantly.   good afternoon, mother,  i replied, pleasantly.  would you like an orange? how about a dozen?   we spent the drive home eating oranges - managed to polish off four of them before we gave up on eating our way through the EverGrowingMoundOfOranges.  i have an idea,  i said.  let\'s toss them into the pool and swim through them, just because we can!   i think my mother would have said something, but she was laughing so hard that she choked on an orange seed, so she didn\'t.  at this rate, i AM going to bleed Tang - and Minute Maid, and various other brands of orange juice - by the time February rolls around. drink me for good health, y\'all.   '),(25049,2238828,'14,January,2003','2003-01-14 00:00:00','    it\'s been a week, and i\'m finally getting off that holiday-induced high. is it the silence in the office? is it the strange, frightening clarity of vision from behind my new spectacles (glasses! dadgummit, GLASSES!)? is it the realization that three-fourths of the people in the HP fandom are probably latent pedophiles, and the other one-fourth are blatant ones?  maybe it\'s just hearing that fernando poe, jr., is thinking of running for president - nay, has announced his plans, according to the husband of one of my mom\'s friends, who was visiting KL for christmas. hearing that, i threw back my head and laughed until i cried, but they weren\'t tears of merriment. we haven\'t learned at all.  we never learn, in fact. we voted bongbong marcos into congress soon after the marcoses returned from exile. we voted loi estrada into the senate right after her husband was impeached. heck, as i understand it we voted mark jimenez into public office when we knew perfectly well america was trying to extradite him for ... something. i plead repression for not knowing.  see, i don\'t understand this. during ramos\'s term, there was this huge hullaballoo when he proposed a charter change - a rewriting of the constitution. by the ruckus raised nationwide, you\'d think he\'d suggested we all sacrifice our firstborn children to keep the population down. on the other hand - as my professor pointed out - it was no time for talk of changing the constitution. the philippines\' political institutions had very weak foundations, she said; our constitution (the latest one, since we have 3 versions) hadn\'t been around long enough to have been established as a foundation.  what do these people think,  she demanded,  that we can change our political behavior and institutions the way we change a bad hairstyle?  that was the rationally expressed reason.  the irrational reason i got from truncated mutterings from my father; he was an activist during the marcos era, whose self-proclaimed life mission was \"to kill marcos\".  what happens after marcos dies?  my mother asked him, back when they were still dating, and she says he just gave her the blankest look. no one had ever really thought that far. in their heads, they figured that if marcos were gone, things would all be better. the hows and whys of getting to that theoretical utopia, however, were unclear.  people are still afraid. PH asked me a week ago  why do your people always protest against the government?  because here in malaysia, they think of activists as gangsters.    you have to understand,  i explained.  they\'re scared that martial law will happen again. they\'re terrified of disappearances in the night and the stifling of the freedom of speech and assorted other bogeymen that marcos very kindly introduced during his extended term as Dictator Of The Philippines.    was it really that bad?  she asked, and i just shrugged.    pain is pain, PH,  i told her.  and for an entire country to suffer for the megalomanic desires of one man - i understand the fear. but i hate how it\'s holding us back.   we\'re so childish, in a way, railing against the administration like we do.  ibagsak si ramos! ibagsak si erap! ibagsak si gloria!  (down with ramos! down with erap! down with gloria!) my God, after you\'ve torn everyone down, who\'s going to be left standing? after you\'ve criticized and turned down all the plans and goals and proposals that are being presented, what\'s left to do? it\'s as if we think that if only, if only so-and-so were gone from this position, things will be better. sometimes that\'s true; a lot of times, it\'s not.  i said, at the beginning of the year, that i\'d been my own tormentor - and i did it so much more effectively than anyone else could have done. i had to come to a point where i was no longer willing to hurt myself before i could move forward.  my country, you and i have more in common than you think.   '),(25050,2238828,'02,January,2003','2003-01-02 00:00:00','    first off: HAPPY NEW YEAR, EVERYBODY! i\'m assuming that no one was stupid enough to play around with things that had the potential to blow off parts of the anatomy, so when i say hope everybody enjoyed the new year\'s celebration i am not picturing ER-type celebrations. me, i went clubbing and danced like a madwoman and discovered a new cocktail. i rang in the new year with miki, my temporary foster sister (aka an exchange student from japan, who is staying with us for a month and a half), my sister (who i snagged from waitressing for a few seconds so i could hug her), and a bunch of semi-drunk strangers. i have no clue if that was any kind of foreshadowing as to what kind of year 2003 is going to be, but it was mad fun, so i don\'t mind. much.  while we\'re talking of new year\'s-type things: i was intrigued by the  urlLink mayfly project , which i discovered through a link on  urlLink mia\'s  site. partially inspired by the fact that a mayfly\'s entire life can be summed up in four words: live. eat. shag. die., the site asks people to collate the past year of living into twenty words or less. i tried to do so, but the words wouldn\'t gel in ways that would tell you what this past year has felt like. after a long while i had something, but it was still not right, like a jigsaw puzzle whose pieces have not been put together properly. the pieces are all in place, but they\'re not locked down yet, because you haven\'t aligned the edges.   this was the final result:  In 2002 I began to write again. I forgave my parents a little more. I forgave myself.  of course it\'s rather raw, but i have no idea how to make it any better. and false modesty aside, it\'s much better than  what the f**k?  - which was, believe it or not, one of the entries up there. concise, yes, but not as all-encompassing and eloquent as the original four-word summary of the mayfly\'s life. and then there was the entire new year\'s resolution thing, which i pretty much gave up on before i even started it. a long long time ago i said \'sorry\' to someone and she turned around, slapped me and told me not to lie and say sorry when i was going to do it again. i\'ve never been able to use that word in the same way again - sincerely, with regret; i can get it out with ease when i\'m apologizing for stepping on someone\'s toes or spilling things on them, but when it begins to mean something and i know it, the word freezes in my throat and i have to walk away. so if i\'ve ever wronged you and failed to apologize, it was not because i didn\'t regret what i did, or because i was too proud. if the apology is worth any less for its wordlessness, then i must apologize a second time, because it\'s still too hard to say it.  one of my ongoing resolutions, though, is to learn how to talk, how to communicate. in 2001 i bought tori amos\'s first album and heard  silent all these years  for the first time, and for weeks that was the only song i listened to on that CD. tori claims that there is a story behind each song, a reason for its existence; i don\'t know what her story for that song was, but i knew what story i would give it, if i could:  when i was eleven we moved to manila from bangkok.  but that\'s not where it begins. it begins when i was three and we moved to bangkok, and i could barely speak english. my mother sent me to an english-speaking school to learn it, and i can still remember hugging the corners, sitting silently at the table and not saying anything because i didn\'t know how to speak english, and no one else could speak tagalog back to me. in manila we\'d lived in my mom\'s old apartment, and there were always uncles and aunts and cousins underfoot. in bangkok it was just the four of us; in the daytime until early evening, just the three of us - my mother, my sister and me - because my father would leave early and come back home late. and later, it was pretty much just the three of us, because my dad travelled so much that sometimes he would be away for months on end and my sister would sleep with his shirt because she was afraid she would forget what he smelled like.  fast forward to eight years later, and this time it was the flip side of the coin: back in manila, near the place where i was born and grew up. my tagalog was fragmented, and it was hard to talk to people; my mother didn\'t want us to speak english because she was afraid people would kidnap us, so we had to use our rusty tagalog or be quiet. in school we were the kids with the funny accent; a girl came up to me once and asked me to talk, then laughed - at my accent, at the way i struggled to find the right words, maybe at something i said. a lot of times it was easier to just not talk, than to spend several minutes hunting for the right word, getting increasingly desperate because the person you were talking to was getting impatient. there was none of the forgiveness granted to foreigners learning the language, because we weren\'t foreigners. my sister got angry and defensive, and i - well, i\'d always been quiet.  i figured that if i learned the language it would get better, but it wasn\'t just the language, it was so many things. clothing - some of the things we wore in bangkok were shocking or strange in manila. manners - in bangkok your food was your own, in manila it was considered rude if you didn\'t offer everyone a little bit of what you were eating. culture - in bangkok it was profane to touch someone else\'s head, in manila it was a sign of affection. so many things. these were the things no one thought to tell you about, simply because they were part of everyday life. no one noticed them unless you weren\'t doing them, so you learned only when someone commented, or laughed, or was offended.  that was at eleven. two years later it was the same situation all over again, only on a more personal level. at thirteen your body changes. you change. your relationships with people change. and just like two years before, there were so many things i wanted to talk about or ask or explain, only i didn\'t have the words yet. today, i can tell you eloquently that i felt lonely and strange and frustrated; i can explain that it was the onset of adolescence, that my parents were holding on to me too hard and i was pulling away too abruptly. back then there were no words, or there were too many, and i didn\'t know which ones were the right ones. at that time, my dad had just lost his job and we had just moved into a new house and things were tense all around, and the wrong words at the inappropriate time might get you slapped, or kicked, or something.  thirteen was my awful year, a year of silence and noise and ignorance. i hated myself that year. everything i said, everything i wore, everything i tried to do, was wrong - if not by my parents\' or peers\' standards, then by my own. in the mornings i tried not to fight with my mother; during the day i enjoyed school; in the afternoons i would go straight to my room and stay there so i wouldn\'t get into any trouble; and at night i would lie in bed and dream of falling or dying. my parents thought my silences were a sign of rebellion, and once it got my father so angry that he got up and wound his belt around his fist.  if you don\'t talk by the count of ten,  he said, his voice very soft,  i will hit your face with this.  it was one of the most helpless moments of my life: for seven counts i sat there, choking on air and fear and things i wanted to say but didn\'t know how to say, and all the time opening my mouth to show my father that i was trying to speak, i was. at the count of eight i stuttered out  what do you want me to say?  and shut my mouth, breathing hard.  and at the end of it all, it was still my fault: my mother got up from the table after my father had left and turned back to me:  look at what you made your father do to you,  she said.  years later - just last year, i think - my mother and i were hanging out at the kitchen table, talking, when she suddenly turned to me and asked  what were you thinking, when you were thirteen? you were so quiet then.   and that easily, i was thirteen again, wordless and frustrated and afraid and alone. i had to breathe before answering.  things,  i said.  normal things. the kind of stuff we talk about now.    so why didn\'t you just say them?  she continued, and she was sincere in her curiosity. for the first time in years, i saw myself from her viewpoint, and it made all easier.   i didn\'t have the right words,  i replied, very gently, and she just looked at me, not understanding. in that moment, i forgave her.  i still have moments like that, from time to time - moments when i\'m so overwhelmed that the words will not gel in my head, or lose their way from mind to mouth. mostly they come when i\'m afraid, or stressed, or frustrated, and i suspect in a small way - like G so eloquently said before - part of me is still in that room, suffocating with fear and anger and frustration.  like i said, this year i learned - and i\'m still learning, slowly - how to put that behind. i can\'t be that girl, stuck in that room, forever. i refused to tell people these stories for so long because my skin crawled at the thought of being labelled \'victim\', or pitied. and yet - by being silent i locked myself in that same room, victimizing myself every time i remembered.  i\'m forgiving them, incident by incident. i\'m forgiving myself as well - for getting into situations like that, for making them feel and act that way, for hating myself so much that i thought i wasn\'t worth the oxygen it took to keep me alive. as any person who\'s been hurt will tell you, it doesn\'t help for people to tell you that it wasn\'t your fault, because some part of you will always feel guilty. it helps more to just look back - to be objective and rational - and to forgive yourself, even though at times it really wasn\'t your fault.   i\'ve been here,  tori sang,  silent all these years.   i\'m breaking the silence. i\'m forgiving myself. i\'m forgiving them. and i\'m talking.  and yeah, it feels good.   '),(25051,2238828,'27,February,2003','2003-02-27 00:00:00','    27 February 2003 / 3:24 PM  all things being equal, and my sister being the linkin park obsessive that she is, this conversation took place yesterday:   sister:  you remember the linkin park picture where brad and phoenix had The Hair?  me : i constantly pray to forget it, but yes.  sister:  okay, so there\'s a story behind it. apparently the guys decided to have an afro-growing contest.  me  what!? why?  sister:  because they want to start a new fashion trend. i don\'t know. they\'re men, they\'re strange by default. anyway, they showed it on the LP network thing.  me  i bet i know who backed out first.  sister:  so they set it up, right? joe was all um, no. right off the bat.  me  yeah, i couldn\'t see joe going for the \'fro thing. i mean, this is the man who had CGI whales floating in the sky, he doesn\'t need to get any weirder.  sister:  and rob tried, but at some point he just looked in the mirror and was all you know what, i\'m just not feeling the \'fro. and off it went. then mike and chester -   me  mike grew an afro? when? and why did you not show me the incriminating pictures off the internet?  sister:  because you keep on accusing me of being a stalker. i keep telling you, i am not responsible for the things that get posted on the internet.  me  you are responsible for the things you find because that means you went looking for them. remember your Hanson phase? that website that had information on the color of taylor hanson\'s toothbrush?  sister:  that was scary, yes, and for future reference, my official press statement is that i never had a Hanson phase.  me  i reserve the Sister\'s Right to Blackmail.  sister:  hey, i know you like the Backstreet Boys.  me  yes, but unlike you, i am not ashamed of it. i would willingly go to BSB concerts and make heart signs in the air and scream things like Howie, i wanna have your baby!  sister:  warn me before you do that, i\'m going to have to change my name and have plastic surgery. but yes, mike had an afro. he was like you know, when i think \'fro, i think big and fluffy, very beyoncÃ© knowles. my hair is more edward scissorhands. i am a disgrace to afros everywhere. and chester - you know how his hair is a little curly?  me  oh my lord, don\'t tell me -   sister:  it was bad. chester was like i have CURLS! and the guys were all fascinated with the way it would spring back when they pulled it. and you know how he has a son and it wouldn\'t be good for Draven\'s future emotional development to learn that his hard rocker dad went through a Goldilocks phase. so their \'fros had to go. and then it was only brad and phoenix.  me  are they planning to get rid of it anytime soon? or anytime, period?  sister:  at this point they\'re still in the Macho Phase of Stubbornness, where manliness is proved by outlasting the other in whatever stupid contest they\'ve devised, so who knows. as long as it doesn\'t affect their music, i\'m cool with it. except maybe now the two of them will have to be permanently stationed behind the other four, because otherwise their hair will block out the small ones. namely, chester.  later that evening we went out, because my dad wanted to have a family portrait taken. he\'d wanted to have it done ever since christmas, but we\'d never been able to gather everybody together at the same time, so it had remained undone. miki - the japanese exchange student who\'s been staying with us for the past month - is leaving to stay with a new family at the end of the week, so last night my dad herded us all together and took us off to the nearest studio.  afterwards we headed to bangsar for dinner; everybody\'s eyes widened when my order came because there was so much of it.  how come they gave you so much?  my sister asked.   \'cause i\'m pretty,  i said without thinking, and the entire family raised their eyebrows. i raised my hands defensively:  look, i don\'t know, okay? don\'t look gift horses in the mouth.  actually, it may have been due to the fact that i was wearing a loose, low-cut blouse and no bra that night, but we\'ll never know for sure.  my sister\'s portions were equally generous:  you\'re not the only one who\'s pretty,  she pointed out. five minutes later, miki tapped me on the shoulder:  atÃ©,  she said happily,  i\'m pretty too!  - because the restaurant had apparently decided to be liberal with the servings that night.  our eyebrows shot back up when dad finally got his order, though - i have never seen a bigger bowl of soup.  okay,  my sister said after a brief, stunned silence.  it\'s official, dad\'s the prettiest.   the proofs of our family portrait come out on saturday. i\'m crossing my fingers, because the dress i wore for the formal portrait was a little shorter, tighter, and lower-cut than i remembered it being.  hooker chic!  i exclaimed in the dressing room, because my sister\'s dress was equally short and low-cut, although not quite as tight.  this is gonna be a non-family-oriented family portrait.   nope, still don\'t have a normal family.   '),(25052,2238828,'21,February,2003','2003-02-21 00:00:00','    after thinking about it a while, i realized that malaysia is not exactly the best place in the world to muse about my status as woman versus commodity. granted it\'s a very modern country, but you\'ve got the occasional (more than occasional, but i was being diplomatic) oddball who will give you a politely surprised look, as if to say \"there\'s a difference?\"  i was reminded of that yesterday, when i was walking to the hotel where the philippine secretariat is staying. i was in full professional regalia - conservative suit, comfortable but slightly sexy shoes (i *am* 22, after all, not 58), and no-nonsense black workbag. would you believe a car actually drew to a stop beside me and the driver leaned out to give me a suggestive grin? i gave him an annoyed look and flashed my badge at him (all participants in the summit of the Non-Aligned Movement are assigned ID badges), then watched as he sheepishly drove away.  but speaking of being a commodity, i\'ve once again been outsourced - or, to be more accurate, hijacked - from my office. technically our office has no involvement in the NAM - it\'s mostly a diplomatic / political affair, so nix on the trade/investment promotion, which is our mandate. hence imagine my surprise when they called me up last week, asking why i wasn\'t at the embassy meetings - apparently someone put me on the secretariat without asking my boss, much less me.  aside from feeling uncomfortably like a piece of office equipment ( do you need rei? no? okay, send her over then, we\'re short a person. ), i\'m a little tired. my office isn\'t involved, so work goes on as usual ... which means that in the mornings i\'m at my office and in the afternoons and evenings i\'m at the secretariat. fifteen minutes in between, i\'m usually walking from the office to the hotel - an ordinary task made arduous by the fact that i\'m wearing 2-inch heels. i could be practical and wear flats, but my legs look fat in them and i\'m vain enough to sacrifice a little comfort for glamour. that sounds incredibly retarded until you remember that i wasn\'t born pretty, i have to make do with what i have - and what i have happens to look better in 2-inch heels than in flats.   it\'s not the work, though - the exact opposite; more like the lack of it, and how tense i am when people are working around me and i\'m doing nothing. it sounds ridiculous, but up to now i haven\'t been given much information on what i\'m supposed to do. i\'ve descended to soliciting work from all and sundry -  d\'you want me to type that? or here, i\'ll photocopy that for you. and staple it. and put it in a folder. and label it. and ... oh, okay, you can have it now.  - and when there\'s nothing to do, obsessively tidying up the room.  i finally commandeered the phone and dedicated myself to answering calls, but that wasn\'t very comfortable when the officers from manila arrived. apparently i have a very strong american accent, so whenever i picked up the phone, the manila people would turn around and stare. the embassy staff are used to me, but the manila people weren\'t. at first they figured i was malaysian, so they were talking to me in english - but i replied in tagalog, which got me even stranger looks. at one point, i heard one of them whisper,  is she filipino?  to the secretariat head, and her affirmative answer. it was like my first year in manila all over again.  today it\'s raining and the secretariat called me up, asking me to come early ( now, if possible  were the exact words, but i can\'t ditch my boss like that). looks like i\'ll be walking through rain - a pleasant prospect if i weren\'t wearing my pretty, painful, pretty painful shoes.   strangely though, i am managing to have fun. i wonder how.   '),(25053,2238828,'18,February,2003','2003-02-18 00:00:00','    sometimes it\'s hard to draw lines. are you a woman or are you a piece of meat?  and it gets even harder when not even your friends seem to remember. when you say something, are they assuming that you\'re asking for it? when you laugh with them, kid around and get comfortable, are they thinking that you\'re leading them on?  how exactly do you differentiate between who you are, behind the flesh, and who they see?  it\'s been two - almost three years now, and i still don\'t understand what happened.  it\'s your fault,  i told myself in the mirror, because when i dissected it down to bits and pieces, sans emotion, i found that i\'d been a participant in it.  but if i\'d been a participant,  i told a friend a year later,  why didn\'t i feel anything?   woman or meat, i still can\'t tell.   '),(25054,2238828,'11,February,2003','2003-02-11 00:00:00','    i\'m so furious i could spit. in fact, i\'m so furious that i will - ptooey! take THAT, bishop teodoro bacani and the aptly named cardinal sin, archbishop of manila. God bless you for all the good you have done in His name, but seriously, the two of you need to step back and reassess your priorities in life.  in this week\'s issue of  Time  is a short sidebar on the philippine Catholic Church\'s stand on birth control and its effect on the local population. apparently we\'re reproducing like rabbits - at an annual rate of 2.34% or something like that, while our GNP is only growing at an annual rate of 3.1%. according to the sidebar, at this rate our population will double - DOUBLE - by 2030. currently there are 80 million filipinos in the philippines, and that only counts the people who were included in the official census. if we keep it up at our current rate of reproduction, there will be 160 million of us stuffed into the teeny weeny philippine islands by the time i\'m 50. good luck to anyone who likes breathing.  the philippine Catholic Church\'s stand on birth control has long been a strong NO - that is, they believe that it is God\'s will that His people \"go forth and multiply\", and to impede the natural forces of reproduction in any way is - their words - sinful. (not kidding - a congressman who tried to pass a family planning bill was actually accused of being a \"Satanist\" and an \"Anti-Christ\" for flouting this command.) they have also denied the causative link between the increase in population and poverty. now, i\'m not catholic, but you don\'t have to be a non-catholic to see how ridiculous that is.  for one thing: we are a poor country. poor as in hardcore poor, as in perhaps a good 40% of our population cannot afford to buy basic necessities such as food and clothing. for another: we are a relatively small country. yes, we have a longer coastline than the United States, but fat lot of good that\'s going to do us when you can\'t live on the damn coastline because we have TYPHOONS. for yet another: does no one think it\'s ridiculous that the people urging us to mate and produce gazillions of babies are CELIBATE? they\'re not the ones who have to bear the kids, raise them, feed them, worry about their futures, pay for their education, and eventually turn into productive members of society.  to top it all off - recently, the president gave a speech in which she vaguely admitted to there being a link between population increase and poverty, and hinted that she would like to see the use of \"modern technology\" in effecting population control. whereupon bishop bacani got on his high horse, demanded that the president explain herself (the president! explain herself - to a priest! only in the philippines, i swear to heaven), and told press that he was \"hurt\" by her implications. the president meekly backed down, insisting that she had meant that family planning should only take place by means of the \"natural\" method.   by all that\'s holy, get over yourself, man. at any rate, in this case natural is not better; more children have been conceived worldwide through use of the rhythm method than any other method.  am i overreacting? am i still making sense? it pisses me off beyond all common sense, truly it does. (again, i say PTOOEY!!) years of this conditioning, of this kind of teaching, have produced a population that shuns birth control - condoms, birth control pills, IUDs, diaphragms, spermicidal jelly, etcetera - as sinful. i have known people who are perfectly comfortable committing adultery, or indulging in premarital sex (although both are verboten in the Catholic faith), but shy from using birth control devices.  three years ago one of my best friends started having sexual relations with her boyfriend for the first time. as girlfriends often do, she told me all about it, and i cautioned her about using birth control, since we were both still in college and neither of us was planning on being a mommy at that age.  but i can\'t walk into a store and buy condoms, i CAN\'T!  she told me, blushing furiously. i, ever shameless, went to the drugstore and bought her an assortment ( thin? thick? ribbed? studs?  or, i said, holding up a gold disc and flicking it with an evil grin,  these gold thingamajigs. like julia said, nothing gets through these suckers. ), which i tucked into her bags despite her protests. a week later she slept with him again, but didn\'t use the condoms. when i asked why, she told me that her boyfriend didn\'t like them.   screw him,  i said, before rethinking my choice of idiom.  or don\'t, unless he gets that condom on. who cares what he wants? it\'s your body and it\'s your life that\'s going to be in tatters if you get knocked up.   she didn\'t confide in me much after that, because every time i saw her i would nag her about using birth control. six months later she came to my door - pale and thin, with desperation in her eyes.  i\'m late,  she said flatly, after i\'d let her in and we\'d scuttled up to my room.  i\'m late and i think i might have an STD.   that was one of the most horrible conversations i\'ve ever had in my life. she was out of control - kept on lighting up cigarettes and putting them out, then lighting them up again. her hands were shaking very badly. sometime during the conversation, she turned to me and said  rei, you\'ve got to help me, do you know what i have to take to abort a baby?  and we fought because i refused to have anything to do with her planned abortion. i finally managed to talk her into waiting until i\'d bought her a pregnancy test, then i sat with her while we waited for the results. she burst into tears when it came out negative, then walked out of my house. i didn\'t see her again until two weeks later, when she told me she\'d broken up with her boyfriend.  my point is - why teach that the use of contraceptives is sinful, when it actually prevents a greater harm? on one hand, preventing conception helps avoid situations where abortion would become an option. on the other hand, it also helps avoid situations where children are born willy-nilly into families that are unable to support and provide for them. this is also a sore spot for me - i\'ve seen too many children born to people who could not afford to take care of them properly, and they ended up abused or neglected or dead. my parents were children like these - both born into poor families that had seven children each, my mother ended up physically, emotionally, and sexually abused; my dad was emotionally neglected.  rationally speaking, i do believe that these people - bishop bacani and his ilk - are arguing from a genuine belief that their way is God\'s way. but think, my good men, think - would God have willed children to be born into families too poor to support them? would God have willed that children conceived at unwanted times be aborted? if you\'re going to argue on the basis of God\'s will, then so can i - and i make more sense, i believe.  i\'m still angry, but underneath it all is this large heavy sadness. so many things are wrong with my country, and i\'m feeling smaller and less capable by the minute.   '),(25055,2238828,'04,February,2003','2003-02-04 00:00:00','    gong xi fa chai , everyone.  in an absolutely evil mood because i belatedly received a text message from my boss asking me to come in at 8:30AM today to help him out on a visit. my phone\'s battery ran out a day or so ago and i didn\'t find the charger until this morning, which is when i received the message. well, there\'s not much point in my reporting to work now, is there? not that i would have been able to, at any rate ... i woke up at 7:30AM this morning and there was no way i could have gotten out of bed - my body felt like it had been worked over several times by the swedish masseuses from hell. it\'s just ... agh, i hate it when this happens. now i\'m going to feel guilty for a week.  - - -  on the Topic-That-Must-Not-Be-Broached: well, i\'m going to bring it up, nyah nyah nyah. it\'s valentine\'s month, the Most Dreaded Month Of All - not just for the single, who get shunted aside like yesterday\'s wilted flowers, but for the people who think that love deserves a little more respect than it actually gets on Valentine\'s Day. the point of this all, of course, is that we of the Anti-Valentine are not exactly Anti-Love, or Anti-Romance, or what have you - as we are often accused of being. quite the contrary, in fact - we do believe in love, as it would be quite pointless and stupid to deny its existence.  what we are against is the way love is usually portrayed - as starry-eyed attraction, sweaty-palmed lust, or mooning stupidity. we are also doubly against the promotion of Valentine\'s Day as some sort of memorial day for love, especially since it\'s been so distorted by commercialism. we believe that love, should you be so fortunate to be able to experience it, is something that cannot be expressed or compressed into a single day\'s celebration. at any rate, when that single day gradually morphs into an excuse to toss trinkets of affection toward people who are usually overlooked and unappreciated, it no longer celebrates love anyway.  and yes, i\'m using the royal plural because i don\'t have my own personal band of sycophants. don\'t worry, i\'ll round them up eventually.  in the meantime, i WILL celebrate the month of love - but my way. which is to say - no hearts, no flowers, no romance. just real stories, some of which don\'t have a happy ending, but which - i think - are far more appropriate to the season than the usual Hallmark commercial fare. so, the first:  it was a week after september 11, 2001, and the world was still crawling around, blind with tears and blood and grief, looking for safety in the ruins of something that had crumbled like chalk and cheese in the middle of what had been a very ordinary afternoon. we were at home, watching TV - another telethon dedicated to those who had perished, and after a while we switched to another channel.  then my sister asked:  if you were one of those people, trapped in that burning building, would you have jumped or would you have stayed?   and, just like i remembered it in  urlLink this entry , i thought about it, and answered:  i guess i would have stayed - found someone who had more to lose than i did, and held on to him or her until we were both dead.    because i don\'t have anything to lose, really. i\'ve lived twenty-two years, and i\'ve fallen in love and lost people and made friends and been betrayed and everything in between, and i\'m loved. if i died, it wouldn\'t matter because i\'d had all that. and i think my sister understood, because she just nodded.   on the other hand,  i added,  if i were with someone i loved and trusted and cared for, i\'d jump.    like that couple who leaped out the window holding hands,  she said.   exactly,  i said.  if i knew i had no choice but to die, then it would be one of the best ways to die - on my own terms, with a hand to hold on the way down. i don\'t think i\'d be scared at all.   there was a long silence, and we continued watching television. after a while, my sister switched it off and turned to walk away. she was halfway across the room when she turned back to look at me.  you know what,  she said, half-smiling.  if we\'d been there, i\'d have jumped with you.   '),(25056,2238828,'19,March,2003','2003-03-19 00:00:00','    got into a massive argument yesterday with a client - well, not exactly a client. the story is that they came to our office and offered to donate a fund of USD5 billion to the Philippine government for developmental purposes, and we asked for information on their foundation so we could forward the proposal to the government.  so yesterday one of them called up, sounding extremely peeved with what he felt was unnecessary waffling around on our part.  we are DONATING to your country,  he told me, angrily.  we are not applying for jobs. why are you making it so hard?    i understand that,  i replied, also upset (this was the middle part of the argument, so by this time we\'d both run short of patience).  but i\'m asking you to see things from my position as well. so far we have the proposal for the fund, but no information on *who* is donating. we appreciate the thought, but there is no way we will be able to forward this to the president\'s office.    your president knows who our foundation is,  he insisted.  just give the documents to her, why do you keep on hiding them?   i bristled at that.  two things,  i snapped.  first - we forwarded the documents already and were explicitly instructed to obtain more information, which we are trying to do now. second - if the president knows who you are, why don\'t you just send the documents to her in a confidential package?    because i know you people, you\'ll just reroute it through your endless departments!  he retorted.  that is your problem, you government workers - you think like robots! you do only what you\'ve been trained to do!   it didn\'t end there, but that part stuck. on the logical side, i suppose i had a right to insist that they provide us with more information, but i\'m still haunted by what he said. i\'ve always disliked civil servants, simply because most of them are exactly what that man accused me of being - bureaucratic. the type of people who will watch you come in bleeding and half-dead, hand you a sheaf of paperwork, and ask you to  fill out these forms in triplicate, and don\'t drip blood on them, that\'s a dear - it\'s a nightmare documenting cases where the details have been obscured by bodily fluids  - in short, people who are so bound by rules and the hierarchy that they forget the true objective of their job.  in one of my MBA classes, we discussed that - the \"old\" model versus the \"re-engineered\" model, where the previous paper-pushing organization was restructured so that each member of the organization was results-oriented. no more endless forms and approvals, just a streamlined procedure where the essential processes had been identified, refined, and reorganized in a more efficient manner. i liked the idea - i\'ve always hated waiting in lines and filling out forms - and thought that it would be a good thing to apply in the workplace.  and now here i am, a government worker, caught in the cogs of the bureacracy, afraid that i\'m becoming yet another piece of this ungainly machinery. so maybe i was right in this case - what about in other cases? do i ask for unnecessary documentation? do i shuffle people around when i can get up and guide them through the process myself? do i palm off certain tasks as being \'outside my responsibility\', just because i don\'t want to get involved?  it horrifies me no end to think that i might have become exactly what i loathed, but sometimes i don\'t know how to escape being this way and most times it\'s exhausting to be otherwise. and yet all my reasons for acting like a typical government worker sound suspiciously like excuses - the budget is too small, there are too few of us, there\'s too much to do and so little time to do it in, i can\'t do that because it doesn\'t fall under the mandate of my office. they\'re valid reasons, i know that, but does it matter? when i\'m part of a government that is supposed to be helping its citizenry, does any of that matter?  SD will probably tell me: that\'s the way it works, it\'s not like you can change things overnight. but it\'s not changing things that bothers me, exactly - it\'s the thought that i could become this way. i can still remember my graduation so clearly, and what i thought i\'d become, and the kind of worker i promised myself i never would be, and - have i fallen short?  it\'s a question worth asking, because i\'m sure no one ever started out thinking  i want to be a boring, mindless, bureaucratic government drone!  and yet somehow they end up that way, because otherwise \'government\' would be synonymous with \'sparkling efficiency, diligence, and courtesy\'.   '),(25057,2238828,'17,March,2003','2003-03-17 00:00:00','    after jenny walked out of his life, forrest gump ran across america for three years, three months and sixteen hours. then he turned around and ran straight back into her arms.  when i was fourteen, i dreamed i ran away from home. it was very very vivid; in my dream i was sitting at my desk, writing. then i looked out of my window and up at the stars, the night was cold and i was wearing my nightgown - and i got up, quietly opened my door, crept out into the hallway past my sister\'s room and past my parents\' room, down the stairs and out the door and past the dogs, out the gate -   and i ran. barefoot and in my nightgown, i ran up and down streets, trying to get out of our subdivision. i was nearly out when i heard something behind me and i turned and there was a car trailing me, and i knew they wanted me back.  so i turned and ran another way, and it was a macabre game of tag, with the car playing it. at some point - i was running on jade street, toward the subdivision\'s basketball court - my legs gave out, and i fell down, crying because i\'d run my fastest and my farthest, and i still hadn\'t gotten away.  then the car door opened and they took me back home.  the point being that you really can\'t run away from things, can you? either you go back and face them, or they hunt you down.   '),(25058,2238828,'18,April,2003','2003-04-18 00:00:00','    Two days ago my mother turned 52, and both my parents celebrated their 24th wedding anniversary. It\'s strange to think that they\'ve made it so far, although it hasn\'t been perfect. Today, while we were having dinner, I asked them:  How does it feel to have been married 24 years?  and the two of them just laughed, my mother leaning against my father\'s shoulder, my father turning to smile down at her.  It wasn\'t sweet so much as it was comfortable, which kind of describes their whole relationship: the way they like hanging out together and being dorky together and bitching at the world together. The way they don\'t fit perfectly into each other, but have found angles at which some parts of them complement. The way he doesn\'t go to the doctor without her because he doesn\'t know how to tell them what\'s wrong with him; the way he stays in the shallow end with her because she can\'t swim, even though he wants to do laps at the deeper end.  How far they\'ve come, how long the journey\'s been, and how strange and wonderful that they\'re still travelling together.   '),(25059,2238828,'29,May,2003','2003-05-29 00:00:00','    I\'ve been all over the Philippines, it seems, in the past couple of weeks. Just a week or so ago I was in Bohol - a small island off Cebu, I think* - visiting my friend M, who is there championing one of her father\'s causes. We didn\'t have much time together - two nights and two days, which is short if you consider that we haven\'t seen each other for two years - so we barely slept, while I was there. My first night we stayed up talking until we literally fell off to sleep; my second night we went down to the beach with her cousins and stayed there until dawn, eating and drinking. We fried fish and waded through the waters under a full moon and (evil evil us) taught her two 15-year-old cousins how to drink tequila. One got drunk, while the other stayed quietly tipsy and wondered why M and I were not at all affected by all the alcohol we were imbibing.  Practice, dear heart,  we laughed.  Then from Bohol I went home, stayed one night, and was off the next day to Gemma\'s house with Laya to pack and prepare for our trip to Puerto Galera - a beach along the coastline of Oriental Mindoro*. My father called while I was in PG; much to his dismay, he discovered that neither my mother, my sister nor I were in the house - nor, for that matter, in the same geographical location. When he called, my mother was in Bicol, visiting her brother; my sister was in La Union with some cousins, learning how to surf; and I was in Puerto Galera, beachcombing. My aunt, who answered the phone, told me later that he sounded utterly gobsmacked. Poor thing.  I\'m currently resting; as I told PH earlier in an e-mail, there is such a thing as too much vacation. Puerto Galera turned out to be a nightmarish vacation, the type that DOES NOT END, not even when you\'ve had more than enough. Times like that, you learn what you are truly made of, and where your breaking points lie. At the terminal, faced with the prospect of not being able to go home for yet another day, I found my breaking point:  I have ONE set of clean clothes, and I am wearing them,  I pointed out to E, who called my phone because I texted him, begging him to help me retain my sanity.  The water we used to bathe last night was saltwater. I am cramping, my clothes do not match, I smell like the sea and I can still feel sand in between my toes and coated on the soles of my shoes. I am THISCLOSE TO HYSTERIA!    Situations like this build character,  you know, replied E, who was laughing all throughout my diatribe. I did not take offense, as I realized how incredibly crazy I sounded, but on the other hand it was not very funny since I was IN the fricking situation.   I do not want character!  I howled into my phone, causing all and sundry in the terminal to glance askance my way.  I want to go home, take a bath, dump my clothes in the wash, and sleep in a clean bed with no sand anywhere!  E wisely realized that I was at the point of no reason, and ceased all logical discussion.  Now - freshly bathed, my clothes in the wash, and with the prospect of a calm night\'s sleep ahead of me - I can look back and see that if anyone wants to break me, all they have to do is abandon me on an island with no soap and changes of clothing. I swear to you, I will break in three hours.  I return to Kuala Lumpur in a couple of days, and I am simultaneously dreading and anticipating it. Kuala Lumpur is beginning to represent routine and regularity, an air-conditioned bedroom and predictable traffic patterns and a social-political-economic environment that is perpetually safe. One part of me craves that discipline, the lines and perpendicularity of such a life - and the other suffocates there and thrives instead in Manila, where everything is hot and murky and tropical and unpredictable.  And then again, I\'ve lost so much in the past month that I thought I was absolutely sure of, and I\'ve gained what I never even thought about having.  * Just so you know, I am severely geographically-challenged. I can get lost in areas that are within a 5km-radius of my house. So if I were my reader, say, I would take all of my geographical descriptions with more than a grain of salt. Yes?   '),(25060,2238828,'06,May,2003','2003-05-06 00:00:00','    The heat is incredible; by afternoon I am so drained that I find a cool corner and sit there and just DO NOT MOVE, to keep my body temperature from rising. Right now I\'m tired and sleepy and HOT and still surrounded by gazillions of people, but I have to get these things down before they disappear.  Three days ago: Spent the night at Gemma\'s place with Laya and Marlen. Laya told me that she hadn\'t painted or written for seven years, and I confessed that I had similar problems.  I haven\'t finished a story in years,  I said.  The words come until a certain point, and then leave.   Laya nodded.  Don\'t work too hard at it,  she advised.  The characters come because they have something they want to say. Let them tell their own story.   Yesterday was the surrealest of the surreal. I spent the whole day at the National Heart Center with V, because her father\'s been confined there for an abnormally high blood pressure. In the evening we went out to have a quick dinner and a stroll around the UP Chapel before V went in to attend Mass. On our way in, a girl stepped into our path and said  AtÃ©, bili po kayo?  (Older Sister, would you like to buy?) as she waved a string of hair bands at me. I paused - I didn\'t really need a hair band, since my hair was already tied back. But she was smiling at me, and asked  Ate, naaalala n\'yo pa ba ako?  (Older Sister, do you still remember me?); I looked closer and realized, with a little shock, that I did indeed remember her.  In college there was a group of girls who used to go around selling hair ties to the female students; I ran into them all the time. They liked me, I think, because I was an easy mark; every time I saw them, I either bought hair ties or promised to buy some next time. Over the course of four years, I began talking to them - asking them why they were working, what their families were like, where they went to school. I found out that they were scholars at a nearby Catholic school; they went to school in the afternoons, and sold hair bands in the mornings to earn money for expenses such as textbooks and pencils and shoes. Before I graduated from UP, I said goodbye to the two oldest ones, and they asked me if I would come back to buy hair bands from them again.  Maybe,  I said, laughing.  Three years down the line, I still recognized her - one of those girls - and she me, although it was dark and we weren\'t really the best of friends and I\'m terrible at remembering faces.  AtÃ© , she said again,  Please buy from me. I\'ve got problems, I need money.    Oh, come,  I replied, smiling at her,  Don\'t we all. What kind of problems do you have?   Of course I was expecting something along the lines of  My mother is pregnant with her eighth child  or  My father just spent all of my school money on his stupid alcohol  - that, I could have handled. But she said  I have a baby, AtÃ© .   What!  I exclaimed, staring hard at her. Beside me, V was watching us with fascination.  But - but - who\'s the father?    I don\'t know,  she confessed.  He ran off and I didn\'t try to go after him because anyway, he paid me. And AtÃ©, I didn\'t know how hard it was to have a child.    Heaven help us,  I said, feeling extremely lost.  How old are you?    Seventeen,  she said.  I didn\'t really know what to say, so my older sister persona took over.  Look,  I told her sternly as I handed her money and she handed over the hair bands,  Children are extremely difficult, and you shouldn\'t have had to give birth to one to find that out. And you\'re too young to be fooling around with things like these! Next time we see each other, we must have a talk.    Yes, AtÃ© , she said meekly, waving to us as we continued along the path to the Chapel. I think that even when I am eighty, I will see her face and her eyes in my head, the way they were when she stepped out of the darkness to talk to me.  That night, I had to go out; I texted E and he said yes, he was going out and would we like to join him? We ended up in Decades along Tomas Morato, and E was so safe - a reminder of the antiseptic order of KL, and the way pain is clean there, like scalpels - that for the rest of the night, I danced and drank and forgot, and was grateful.  But now it\'s afternoon, and I\'m back in the sizzling heat of sun and people, and I can see the lost look in that girl\'s eyes again.   '),(25061,2238828,'02,May,2003','2003-05-02 00:00:00','    Conversations I had in KL, days and hours before coming here, are replaying themselves in my head now. The things we spoke about, and the things we said - was it prescience? Was it a quiet tap on the shoulder, a warning that coming back is never easy? I told Hoffe months ago that  I\'m so good at leaving; now I\'m trying to be equally good at coming back.  The thing is, though, it\'s so much easier to cut strings than to try to tie them back together.  Two days before departure, I went out with a few friends and ended up in Y\'s house, playing truth poker (something like strip poker, only instead of taking off clothes, you have to tell the truth. It\'s harder than it sounds). Sometime around 2AM, the thrill faded and we sat there, absently shuffling the cards instead of dealing them out for another game. We began talking about everything and anything, ending up, absurdly, on the topics of relationships and sex and marriage. From there we went to Manila and independence, and Y turned to me and said  Days in Manila are so unpredictable.   PA, our other friend, looked slightly puzzled; he moved to KL when he was 11 and has been back to Manila infrequently since then.  Is it so different,  he asked,  from KL?   Y and I shared a glance and laughed.  KL is to Manila as a carousel is to a rollercoaster,  I said.  The days here travel on smooth wheels; you know when to get on and when to get off. In Manila you hop on, buckle up and clench your teeth, and hope to God that the day takes you where you need to go.   And then - hours before departure - C and I got into a fight. Back and forth, back and forth we went, until finally we both collapsed on her bed in the middle of the clothes she\'d set out for packing.  I don\'t belong anywhere,  she told me, as we lay there, tired and frustrated.  I feel as if I have to change every time we go somewhere new, just so that I don\'t stick out so much, and it doesn\'t even help because I don\'t fit in anymore.    You fit in with me,  I said, and it felt clumsy, as if I were reaching out in the dark and were hitting rocks instead of her hand.  I fit in with you. If you shut me out, or if I shut you out, then we\'re both screwed. Who are you going to be real with, if not with me? And who can I be absolutely real with, if not with you?    I\'m so tired of changing,  she murmured.  But what scares me the most is not the changing; it\'s - if I *don\'t* change, if I stay my real self, what do I do if people don\'t like me?   - - -  So far, it feels like I\'m dancing a cotillion two beats behind time - swaying to the right while everyone else is swaying to the left, taking an extra step to find my place when everyone else has settled into theirs. I want to dilate time, to stretch the moments so that I have my few extra seconds to move back into place, to get used to the beat.  On the other hand, I could dilate time until it stretched a million years, and I am afraid I would always be two beats behind.   '),(25062,2238828,'16,June,2003','2003-06-16 00:00:00','    Still with the weird dreams. Last night I dreamt I was with a girl whose brother had been murdered in a particularly gory way - he\'d been hacked to pieces. We weren\'t at the scene of the crime; we were somewhere else and her mother called her phone and told her to get down to the morgue, the remains of her brother\'s body were there. Apparently the girl and I were close friends, because I accompanied her to the morgue to identify the body.  The dream skipped over the identification process, thank goodness, and fast-forwarded to a particularly gross segment: my friend, grief-stricken over the loss of her brother and apparently guilt-laden about his death, decided to mummify his remains and carry them around with her in a cloth bag.  Here\'s where it gets even stranger. We were standing in a public bathroom, in front of the mirrors. I was leaning back against one of the stall doors; my friend pulled out the cloth bag, laid it carefully in one of the sinks, and unwrapped it, reaching in to stroke one of her dead brother\'s hands.  Don\'t unwrap it , I said irritably, reaching up to fix my hair.  We\'ve got to go to ________ soon, and you know you\'re required to cover your head every time we go in.   She nodded and began rewrapping the parcel.  I just realized , she said, casually,  We can\'t go in there carrying something, our hands need to be free.    Make a sling , I suggested.  Carry it on your back.    Oh, that would be a great idea , she replied,  but we only have two cloths - one for your head and one for mine. I\'ll have to carry this on my head.   Having said that, she put her bundle on top of her head and fastened it there with her head covering.  I told my mother about that dream this morning; she was creeped out and it showed.  Goodness , she exclaimed,  What have you been reading about lately that you\'d dream about these things?    I have no clue,  I replied.  To tell you the truth,  I told her,  It creeped me out majorly, too.  We ended the discussion there, and went on to other stories. I did not tell my mother that the dream had not ended there.  This is how it ended, the last scene before I woke up and could not go back to sleep: my friend put the bundle on her head, she tied it on with her head covering, I covered my own head, and we left the bathroom. Before I went out the door, I turned back one last time to check my face in the mirror. In my reflection, I saw - to my great surprise and horror - that the bundle was on *my* head, not my friend\'s.  I turned back to call my friend, to ask her why it was on my head, and she was not there. There was only a janitress mopping the floor; I tapped her on the shoulder and said  Excuse me, did you see where my friend went?   She looked at me, bewildered.  What friend?  she asked.  I\'ve been mopping here ever since you went in. You\'ve been alone in there all this time, miss.   And then I woke up.   '),(25063,2238828,'12,June,2003','2003-06-12 00:00:00','    Things that slip by without you really noticing: dead leaves. Wild birds. Grass, and cars. Pages of a very absorbing book. People you have loved for so long that you are beginning to take them for granted.  Time.  It slipped by without my noticing, but I will not let it pass completely without comment: Happy Birthday,  luminae . It\'s been a year now, and I am still here, and that is something of a miracle. How do you feel about another year together?  - - -   The things I\'m dreaming, now. All connected to water in some way; Laya and Gemma may have been the ones to almost-drown in Puerto Galera, but I think I was the one who took the ocean home. A couple of nights ago I dreamed I was talking to a fish, arguing about the best way to go home. The fish - a large silver one with big buggy eyes - seemed annoyed by my tenaciousness.  Why are you being so pig-headed,  it finally snapped,  It\'s not like you know where your home is anyway.    Oh, I know!  I cried.  I live in Kuala Lumpur, in M______.   Whereupon it looked at me with utter scorn and a little bit of pity, and said  I wasn\'t talking about where you  live , sweetfins.  And then it kissed me and turned into E, who laughed and swam away.  Although perhaps it isn\'t really about water at all.   '),(25064,2238828,'28,July,2003','2003-07-28 00:00:00','    Have the mother of all headache attacks; if I move my head wrong, a little metal ball starts rolling around inside my skull and bumping HARD into soft places inside. This is what my tension headaches feel like, but I\'m not tense so I\'m not sure what the problem is.  Or maybe I am - tense, or perhaps the word is overstretched. I\'ve been spending the past week saying goodbye to KL - eating Malaysian food nonstop, partying, going back to my old hangouts, meeting up with people, et cetera. This coming Saturday I\'m going to be on another flight back to Manila (where, I hear, there\'s an ongoing coup d\'etat - whoever complains of boredom in Manila is either lying or a hermit) and it\'ll be several months before I return - that is, if I return at all. But I can\'t complain; I\'ve packed more living into the past few days than I\'ve done in the past few years in KL.  Just yesterday I managed to fulfill one of my long-time dreams, which was to perform with a live band onstage at a bar somewhere. Turns out that one of my sister\'s friends had just started up a band, they\'d landed a gig at this chintzy little bar downtown, and they needed a backup singer. My sister volunteered me, and so Saturday evening found me onstage, purring India Arie\'s  Promises . It didn\'t turn out too well - the band was too new, we\'d practiced only once, and our playlist wasn\'t exactly geared to the musical taste of the audience - but my dream was fulfilled nonetheless. RZ, the lead singer, was disappointed to find out that I was leaving; we got along pretty well and blended quite nicely with each other,and she wanted me to join the band.  At this point, though, I suppose I could die happy. I\'ve only ever had two major dreams: to publish something I\'ve written, and to sing in public, and so far I\'ve done both. In Bangkok, L was asking me about my dreams and whether my parents\' plans for me and the turbulence of my life had ever interfered with my pursuit of them.  No , I had to answer.  I wanted these things so badly that it didn\'t matter where I was, or what I was doing . Which is true; music and writing follow me around wherever I go, or maybe it is that I search them out the way animals hunt food. It is not so much that I rearrange my life to accommodate them; it\'s more that they are so much a part of me that I cannot go anywhere or do anything without bringing those two things into play.  Which goes to show, I suppose, that it\'s not at all a matter of how hard it is to achieve what you want; it\'s more a matter of how much you want it.   '),(25065,2238828,'22,July,2003','2003-07-22 00:00:00','    I\'ve finally found something I can do at the beach and have fun with, yay! This requires some explanation: although I have gone to the beach often in my short life on the pretext of a vacation, like any normal human being, apparently I belong to the teenyweeny subculture of humans who like the beach but hate sand and loathe bathing in the ocean. It makes perfect sense to me - sand is gritty and crumby and messy and gets into places that are extremely uncomfortable; the ocean has been the repository of the world\'s waste for aeons and is not fit to swim in for anything that is not a natural sea-dweller.  This of course brings us to the very obvious question, which E asked complete with befuddled look:  If you hate sand and don\'t swim in the ocean, what on earth do you do at the beach?  Plenty. I walk barefoot along the shore; I collect shells; I take pictures of the horizon; I build sandcastles; I slather myself with suntan lotion and wait for myself to tan.  Then , as I explained to an incredulous Laya,  I return to my hotel, brush off the sand on the concrete walk, and go swimming in the pool . Wild untamed beaches, quite obviously, hold no real allure for me.  Nevertheless, I went down to Pattaya this past weekend and enjoyed myself after a couple of false starts. Our first night there we went to Hard Rock for their foam party, but found it closed when we arrived (but to be fair - what normal club closes at midnight? I ask you). We then went to Palladium, where we bought a bottle of Johnnie Walker and watched with dismay as gay crossdressers who were ten times hotter than us (it\'s a sad world when gay men are hotter than real women) sang and danced onstage in very skimpy costumes. I consoled myself with the thought that they really sucked as singers - they sounded like cats mating, which is a very likely metaphor - but it did not make up for the fact that their waists were at least 5 inches smaller than mine. At some point I decided that I was not going to leave that benighted club without at least dancing, so I got up on one of the ledges and danced - to  The Ketchup Song , because it was the only one I could dance to.  (You may all disown me now.)  The next day we headed down for Jomtien Beach, picking up some glazed donuts along the way. Once we got there we ate again, stuffing ourselves with seafood (well, what else are you going to eat at the beach) and whiskey-spiked cokes. After a brief nap, we were off to play. My aunt suggested that we take a banana boat ride, and I agreed despite my misgivings. I really should learn to listen to my misgivings, because we fell off that stupid boat a grand total of - not once, not twice, but THREE TIMES.  Guess who had the hardest time getting back up on that blessed boat. I even tipped it over once, tossing everybody else back into the drink. It\'s a wonder they didn\'t just leave me there.  We got back completely drenched, with me sulking because I hadn\'t wanted to get wet and instead I\'d gotten dunked so many times I was beginning to feel like a donut. My companions - not sharing my neurotic aversion for saltwater - decided to stay in the sea and swim, while I immediately headed for shore.  As soon as I got to shore, though, I caught sight of a jetski. L saw me eyeing it and asked  Do you know how to ride one?    No , I answered.  Do you?    No , he said. There was a brief pause before we turned to each other and asked, almost simultaneously,  Do you want to?    Definitely , I said,  But I want to go parasailing first.   So I went parasailing. They strapped me into a harness that was extremely tight, fastened a parachute to the harness, and then hooked me up to a rope tied to a boat. The boat started, went off, and up I went. It was splendid and wonderful and thrilling and breathtaking, and there are no words for being suspended high above the ocean, viewing it from a bird\'s vantage point. For moments I felt nearly like Icarus, just before he flew too high and melted his wings; and I understood what had made him go too close to the sun, even though it eventually killed him.  (Next activity: bungee jumping. Or skydiving.)  Once I touched down, I grabbed L and we went straight for the jetskis. It took them fifteen minutes to find me a jetski because apparently it was beyond their comprehension that a girl should want to ride alone. All the other women were passengers on their boyfriends\' jetskis, but L and I were going to race so that wasn\'t an option. As it was, L took off without me, so when they finally let me loose on my own ride, he was already out in the deep.  I finally got my jetski and took off. Unfortunately, the jetski was harder to control than it looked - it took me a full five minutes to get it to stop moving around in circles. People were moving nervously out of my way - I must have looked like I was going to run them over, since I was leaning intently over the handlebars. Since I wasn\'t that sure of my navigational skills either, I decided to play safe and headed out to the open sea, where there was no one to run over.  Eventually I got the hang of it and went looking for L. He wasn\'t hard to find, even if I wasn\'t wearing my glasses - it\'s hard to miss a fair-skinned guy wearing purple from head to toe. I waved at him, he waved back, and we took off, skimming and bouncing across the waves. It\'s hard to tell who won - the spray would come back into our faces and we\'d have to slow down to wipe it away - but we\'d speed a full length one way, then do wide turns in opposite directions and do another length the other way.  I am currently paying for that fun, in the form of muscle pains in my arms and the back of my neck. Apparently handling a jetski takes more muscle than I comfortably have, and I\'m going to carry these aches until I leave for KL on Wednesday.  That\'s less than two days away. I can hardly believe it; I\'ve been here almost a month and it\'s all flown by so fast.  Bangkok, Bangkok, how I will miss you and the people I leave with you when I am gone.   '),(25066,2238828,'15,July,2003','2003-07-15 00:00:00','    I am still in Bangkok, although currently in a remote part of it that I am not very familiar with. Whether this sounds believable or not, I am attending a seminar on Professional Training for Microfinance and Microentrepreneurship, and actually participating actively. My childhood spent tagging along after my father and reading his microfinance papers when I was bored has finally paid off; I\'ve faked an entire groupful of microfinance professionals into thinking I\'m one of them. That could be very very impressive or very very scary, depending on how you look at it.  (For the uninitiated: microfinance is one of the tools used in poverty alleviation programs. It centers primarily around providing easy access to credit/finance to poor people, as well as teaching them how to save and manage their finances. Okay, it\'s a lot more complicated than that, but I\'m in an Internet cafe and I don\'t have time to explain.)  I am enjoying myself, strangely enough, although I\'m exiting from the program before we start designing our own microfinance institutions - while I am pretty knowledgeable for a lay person, I am not *that* good. Poverty alleviation was something that interested me on a personal basis - the idea of doing something so grassroots, so directly beneficial to those who needed it, was very appealing. On the other hand, it was something my father was very involved in, and very good at - and I have never liked being referred to as \"the daughter of\". I am proud to be his daughter, but I am my own person and I like to be judged on my own merits. Had I entered the microfinance field, I would have remained \"Mr. Quinones\'s daughter\" for at least the first ten years of my practice; some of the current practitioners of microfinance in the region have worked with him ever since I was a child, and have watched me grow up. On the other hand, I am so interested that I might continue to do volunteer work in this field, just for the pure joy of it.  Am still roaming Bangkok, and remembering more everyday. I am so at home here that it is a little scary - I left Manila for two years and felt so out-of-place when I returned, and now I\'ve returned to Bangkok after eight years and it still feels comfortable. There\'s still that office building where my old ballet school was, and the malls we used to go to, and of course Patpong (Thailand\'s infamous sex district), where we used to have to pass through to get to our grocery store. There\'s Swensen\'s bubblegum-flavored ice cream, and the King\'s Anthem played in public places, and familiar food being sold in the streets. My Thai remains, as ever, sadly fragmented, but as before, I can get around alone if need be.  I could make this my home if I chose to; I know people here and I have connections and I am comfortable with the culture and people. This leaves me at a crossroads, really. Recently I got news (from no less than three, count \'em, THREE people) that I\'d passed the pre-qualifying exam for the Foreign Service. E was amused when I told him I was worried about passing; apparently he thinks there\'s no way I could fail. I wish I were as sure.  I  want  to pass, because everyone expects me to and I hate disappointing people. And yet if I pass, then that would leave me in a quandary: to live in the Philippines and pursue a career in the Foreign Service, or to go back to Bangkok and aggressively pursue a new career, or - and here\'s yet another thorn in my side - to follow my family to the US, where I have no idea what kind of job I\'ll end up in?  I have a feeling that the answer lies in how strong I really am.   '),(25067,2238828,'03,July,2003','2003-07-03 00:00:00','    Am currently in Bangkok, rambling through my childhood haunts and realizing in the process that parts of my childhood were, indeed, surreal. A couple of nights ago I went out clubbing with D and L, people I used to go to kindergarten and grade school with. L\'s girlfriend and a friend of hers came along, and we shamelessly left them out. I would have felt guilty - we were unaccountably rude, although I danced with L\'s girlfriend and D tried to chat with them a couple of times - except, dude, I haven\'t seen these friends in what, almost ten years, so the guilt has been mainly overridden. We bought a bottle of Johnnie Walker and finished the whole thing, just the three of us (L\'s girlfriend and her friend stuck to beer), and D and L had to carry me out to the car and up to my room.  Then yesterday L came over, teasing me about how completely trashed I\'d gotten the night before; I thwapped him across the head and we went to MBK to get me cheap cheap glasses because I keep on losing them.  Rule of thumb , L pointed out,  It\'s the expensive stuff you lose. What do you want to bet that this cheapass pair is the one you\'re going to keep until the day you turn 30?   After that L and I went around looking at comic figurines while his girlfriend tagged along in the background, looking bewildered. I admit that comic figurines - horrifically realistic reproductions of Spawn nightmares, for example - are not typical feminine pursuits, but when 80% of your friends are guys, it is hard to squoosh in time for makeup and shoes. Thankfully I manage to do so anyway, and at any rate the fact that I wear high heels and lipstick does not detract from my enjoyment of comic books and action movies.  After dinner, L\'s girlfriend asked - would I like to go dancing? And I said yes, since I am always up for an excuse to kick up my heels. We ended up on Silom, and I mentioned to L that  Oh, this reminds me of Patpong!  - Thailand\'s infamous sex district, which was also next door to the apartment I lived in when I was younger.  L grinned and said  Patpong\'s right down the street, wanna go?  So we took off, L\'s girlfriend trailing behind us in confusion (what would you think, indeed, if your boyfriend and his pseudo-sister took off for your city\'s sex district while giggling madly and saying stuff in a language you didn\'t understand very well) but being a good sport about it, overall. It was like a walk down a very surreal memory lane - our old supermarket was still there, with minimal changes; L and I walked down the aisles, laughing like lunatics when we found the places where we\'d used to pluck out candies and ice cream and cheese and yogurt. Then we strolled around Patpong on general principle, and I kept on pointing to random go-go bars and saying  L, dude, Cleopatra\'s Go-Go Bar is still here! And that massage parlor, I remember that being here before!  They\'d taken down some of the signs and blinking neon lights, which L and I discussed with disappointment, but all in all it was still pretty much the same. No one believes me when I tell them I have fond childhood memories of Patpong, but last night I went back there and all at once I was six again, tugging on my mom\'s hand and wondering why we couldn\'t go in and watch the pretty ladies dancing naked on top of tables.  I feel safe here, strangely enough. D and L and I had an informal convention on how we\'d found life in Manila, and came to the general consensus that it was wild and fun and - strangely - painful. We\'ve lost one of our old group - she cracked and ran off to Cebu with her band, and hasn\'t returned since - and a lot of us managed to tumble into bad habits which we are still recovering from today. We talked relationships as well, and I thought I\'m so glad  I\'m their friend and not their girlfriend , because - I don\'t know why, but we have all turned out to be inordinately cruel in terms of falling in love, because none of us have yet, not really. Last night I lay awake thinking of how easy it is for me to play with people, and wondering if I\'m doing the same thing in my relationships.  Tell me if you\'re serious,  I said once.  In the name of our friendship, tell me if you\'re serious. Because if you are, I\'ll take you seriously. Otherwise I\'m going to assume that all you want to do is play, and just because I\'m younger than you are doesn\'t mean I don\'t know the game.   I wonder whether they take me seriously. I hope they do, because I am meeting the child I was and reconciling her to the girl/woman I am now, and there are parts of her that are not pretty.  I can be cruel,  Tori sang,  I don\'t know why.   Although I do know why, and the fact of the matter is that money can in no ways be considered protection. In some ways, our group - protected and pampered as we all were - is much more streetwise than people who actually grew up on the street. In our defense, we are trying to use what we have for a good purpose, but we are also trying to survive at the same time.   '),(25068,2238828,'20,August,2003','2003-08-20 00:00:00','    What kind of story can you write with the words \'white\', \'uterus\', and \'suicide\'?  Yeah, I was afraid so. This should teach me not to accept challenges from any of my friends, because - as I have noted several times before - I don\'t have any normal ones.  I set aside the whole of yesterday to watch two movies that I\'ve been wanting to watch for quite a while -  Bad Boys 2 , to indulge my inner guy; and  The Debut , a small-budget indie film about a Filipino-American community that won the San Diego Asian Film Festival. The first movie was, as expected, entertaining; I will not criticize technical weaknesses or plot wonkiness because anyone who watches an action film for the intellectual stimulus should have his head examined. (Girls who watch action films for the romance that is usually thrown in as a sop to the female audience should be similarly examined.)  It was the second movie that grabbed me, that walked into my head and heart and stayed there. It was hilarious and touching and managed to strike several very sensitive points with me; the people in the movie could have been my sister, my godbrother, my mother, people I know in the US - why beat around the bush? They could have been me.  A short synopsis would, I think, be helpful. Ben, the boy through whose eyes the film\'s story is told, is a Filipino-American boy who was born in the Philippines but was brought to the US when he was very young, and has lived there ever since. He has no memories of the Philippines. He cannot speak the language, and knows very little of its culture. He hangs out almost exclusively with his Caucasian-American friends, and never brings them home because he is embarrassed that his house always smells of fish. He wants to be an artist; his father wants him to study medicine. To top it all off, he is being forced to go to his sister\'s debutante party (an event exclusive to the local Filipino-American community), which is being held on the same night as the coolest senior party of the year. Ben is, in short, a boy who is torn between nature and nurture - trying to find that happy balance between the conflicting demands of his Filipino ethnicity and his American upbringing.  An amazing range of issues are covered in the film; if you are Filipino and have lived any amount of time somewhere other than the Philippines, you cannot help but recognize some of the situations that crop up. From Ben\'s conflicted relationship with his father to his friends\' reactions to the ubiquitous giant spoon and fork that hang in the dining room, everything has a tinge of the familiar. The theater was near-empty, which was a blessing, as I found myself laughing out loud often, most of the time due to moments of surprised self-recognition. The gatherings around basketball courts and the sideline flirtations and conversations; the family friend who feels as if she is on a higher social level than anyone else and uses the occasion to flaunt herself; the Filipino kids who dance traditional folk dances and modern freestyle with the same grace and ease - I have been to these gatherings. I know these people. I am them.  Passing through PowerBooks as I whiled away time between movies, I came across the latest issue of Time, which featured a cover story on  urlLink Asian diaspora  - people of Asian descent who have since migrated elsewhere, and are struggling to find their roots. They wrote of displacement, of the feeling of being suspended between worlds, of their efforts to reconnect with - or disconnect from, as the case may be - the cultures and people they had left behind. Monique Truong spoke of the hostility and resentment she and other Vietnamese-Americans encountered whenever they returned, how hard it was to go home and how a small guilty part of her no longer saw Vietnam as home. Jessica Hagedorn spoke of how her memories of the Philippines were no longer real; places and people she remembered from her childhood had long changed, or moved, or been replaced.    Home is inside me , one of them wrote, after describing his efforts to search for home in his birthplace, in his parents\' birthplace, in the place he had grown up, and the various places he had lived. It was a common thread that ran through the various stories, although they had come from different countries and faced different problems. Eventually each one had come to a realization that they could not point to a place on a map and identify it as home, because home consisted of their memories, their family, the various places and cultures they had absorbed throughout their lifetimes. I think I can sum it up even more neatly than they did:  I  am home.  (I don\'t know, or understand exactly, why I keep harping on this theme, and currently I\'m mapping out an article that will hopefully let me lay this issue to rest. Is it the impending move to the US? Is it the recent travel jag, which saw me hopping across three cities - all of which have emotional significance to me - in less than three months? Is it my involvement with E, which more and more make me realize how transient my life has been, especially compared to his and the lives of people like him?)  E told me once that he supposed that I wouldn\'t find home until I started a family of my own.  They\'ll ground you,  he theorized, sounding eerily like my parents and my aunt, who insisted that should I enter a relationship, I would have to give up my wandering ways and settle down.  Why does it have to be that way?  I asked my mother.  Why do I have to be the one who gives so much up? Why can\'t the guy be the one who goes with me?   For a while, I suppose, I believed them - that I was, essentially, a girl without a country, and that a steady relationship would settle me down. Two days ago, I changed my mind.  Two days ago, I was in UP. Rendered irrational and restless by first-day cramps, searching for a quiet place to sit and try out the supposed benefits of Midol, I ended up on the steps of the Main Library. As I had feared, once I sat down I lost the desire to get up again, and had to resign myself to staying there for a few hours. Twilight fell while I was sitting there, bored and high on the effects of Midol. In the gentle dark, I happened to glance across the road into the UP Lagoon, and caught a glimpse of a family of fireflies, out for their daily dance.  I don\'t know how long I watched them. A fine drizzle came and went, and people passed by on their busy way to Somewhere Else, as I sat there. One firefly - the group\'s black sheep, no doubt - broke free from the bunch and came daringly close to where I was sitting, shimmering flirtatiously at me. And it was then - for some strange reason - that I relaxed and let go. Like I said, I am home. I cannot be a stranger to places that offer me moments like those with the fireflies - moments of strange, delicious whimsy and wonder - and I know I will find places like that no matter where I am.   '),(25069,2238828,'12,August,2003','2003-08-12 00:00:00','    One week has never been so long.  I spent this past week reviewing - let\'s make that *trying* to review - for the written Foreign Service Exams. I managed to get my hands on two - count \'em again, folks, TWO - textbooks on the global political economy, and finally managed to read through the first chapter of one of the books two days before the actual exam. The night before the exam, I panicked, turned on CNN, and refused to let anyone change the channel. After my eyes had started crossing from absolute boredom, I logged on to the Internet and, after aimlessly surfing through half the webpages ever created, turned to the BBC page and facepalmed when I saw the Philippines\' faux mutiny being intelligently mocked on its front page.  After that, I inexplicably lost the desire to study international events.  As for what actually happened during the exam, it is a story in and of itself.    Day One  Otherwise known as  The Beginning is the End is the Beginning , courtesy of the erstwhile Smashing Pumpkins. With the full knowledge that I live approximately two hours away from the testing center, I allowed my insomnia to take over me, resulting in my only being able to have three hours of sleep before I had to drag myself out of bed, get dressed, and ooze toward the door. My cousin, bless his heart, had volunteered to take me to the testing center because he knows I am directionally-challenged; we drove to the nearest MRT station, fought our way on (in the Philippines everything, including public transport, is nasty and brutish but not quite short) and remained in exhausted silence all the way to our destination. Once there we got off and hopped onto a jeep that would pass by the road leading to the testing center; I will abbreviate this passage by not bitching about having to inhale pollutants and toxins at such an early hour. We took a tricycle - that is what it\'s called, although it does not in any way resemble a three-wheeled bike; rather, it is a tiny metal cage attached to a motorbike, which is supposed to serve as part of public transportation in the Philippines (You may admire our survival skills now.) - the rest of the way, and in this undignified manner I arrived at the first day of my exam.  The exams began at 9AM, although we had been strictly instructed to arrive at 8AM, as the doors - according to the admission pass - would be closed by then. Since this is the Philippines, 8AM was loosely translated into \"oh, 8:15 or thereabouts, don\'t let\'s be too nitpicky about a minute here and there\" and the doors *were* finally closed at 8:30. We were lectured by confident, smiling proctors who annoyed us with their very cheerfulness; what right had they to smile when (a) it was so godforsakenly early and (b) they were not the ones on the wrong end of this stupid exam? The first exam, they said, would be English.  Just breathe , said one of the proctors.  Oxygen in your brain allows you to think better .  It also allows you to live, but I restrained my inner smartass and tried not to tremble. The airconditioning, by the way, had been set to \"Inner Antarctica\". Apparently working at the Department of Foreign Affairs turns you into a penguin.  The English exam went by execrably fast; I took issue with several questions that contained grammatical errors, but restrained myself from editing my exam questionnaire. We took a break for lunch, but I was too nervous to leave the building so I stayed inside instead with all the other examinees who were also too nervous to leave the building.  Second half: International Relations. The first question was a long-winded, unintelligible question about either the US invasion of Iraq, or the UN response to the situation, or the general state of international terrorism, I\'m not really sure. I spent two hours answering that question and ended up giving short, blunt answers to the others. One question asked for five or six transnational issues that were burningly relevant in our time. After contagions and international terrorism, my mind went blank.  It was, naturally, drizzling when I left the testing center. I walked straight down the road, saw a taxi, and got in. The driver, who turned out to be on a break, was initially reluctant to bring me anywhere:  Miss , he begged,  Just take a jeep .  I was too exhausted and out-of-sorts to debate with him.  Manong, PLEASE , I said crisply, in the tone of one who is at the end of her rope. He took one look at me through his rear-view mirror and gave in.   Day Two   Since I had collapsed into bed the night before at 9, I woke up in relatively better spirits. I got down in plenty of time to eat breakfast; my aunt gave me a cup of coffee and asked how I was going to get to the testing center.  Oh, Cousin M\'s taking me , I said confidently. She went off to wake him, then returned to tell me that he\'d said he wasn\'t taking me anywhere.  I could not mock his ancestors because they happen to be the same as mine, so I satisfied myself with shaking my fist in the general direction of his room. Faced with no other choice, I walked out and caught a taxi instead. Miraculously, I got to the testing center without any major mishaps.  And then promptly wished I\'d stayed home.  The first test of the day was Filipino. When they gave the signal to begin the exam, I turned the questionnaire over praying that it would not be unnecessarily cruel. Somewhere in the places where prayers are received and transmitted to the proper departments, the word \"not\" was omitted. The first part of the exam was vocabulary, consisting of five words - none of which I had ever seen or heard of in my entire life, and was doubting even really existed. The second part - translation, or pagsasalin, where we were asked to translate a literary passage and a news article into Filipino. The third part was an essay with a hopelessly ambiguous type of topic. I wasted two minutes when I laid my head down on my desk and indulged in a silent scream.  After mangling my native language, I went off to lunch - this time deciding that I wasn\'t going to be a chicken, I was going to leave the building. I had trouble leaving through the gate; apparently two of the policemen assigned to patrol the area had gotten a crush on me and kept on insisting that they would escort me to lunch. With my luck, however, I would probably have been mistaken as a dangerous criminal en route to jail, so I politely declined.  A mistake, as it turned out, since I really have no head for direction. I ended up walking down half of Roxas Boulevard, getting honked at by three-fourths of the taxis in Manila, while getting odd looks from the local vagrant psychotics. You know you\'re doing something wrong when even the nuts think you\'re crazy. So I turned around and marched right back in the opposite direction, and found McDonald\'s.  Back to the exam room after lunch. I forgot to mention that this time I\'d been smart and brought a jacket - only to find out I didn\'t really need it, since someone had already complained about the room temp and they\'d turned down the a/c. God works in mysterious ways, indeed; I have trouble trying to decipher what exactly He\'s doing with my life.  The proctor took the opportunity to terrify us a little before passing out the exams. It was on Philippine Conditions, and  It\'s 30 percent of the examination , she said impressively. Which was a big thing, since the written exams were broken up into 6 parts; English, International Relations and World History were 20 percent each, Foreign Language and Filipino were 5 percent each, and Philippine Conditions - 30 percent. I trembled in my seat; I had done no reviewing at all on this topic except a brief glance through the 1987 Constitution the night before.  The questionnaire had 6 questions - horrors - but all of them were remarkably humble questions, asking the examinee to comment on one situation or another. I took the opportunity to bitch - albeit diplomatically - about various aspects of Philippine government, not the least of which was their seemingly perpetual lack of funds for what they insist are *important* projects. After two hours of happy, enthusiastic scribbling, I turned in 10 pages of thinly disguised aggrieved-citizen-whining.  I danced out of the exam room that afternoon.   Day Three   My cousin drove me all the way to the testing center, making me retract my previous fist-shaking. I went in - again, with a wildly thumping heart - because this was another portion of the exam I had not reviewed for at all. The night before I had made a pathetic attempt to download information on World History, but was taken aback at the sheer immensity of the information available. Most started with  Since the dawn of humankind ...  and were not at all attractive to a last-minute crammer.  Don\'t look so worried, the proctor said, still cheerfully (what is *wrong* with our government officials, I ask of you). It\'s very easy. I briefly considered the wisdom of taking off my sneaker and throwing it at her - I have very good aim when it comes to throwing shoes at people - but reason prevailed. The questionnaires were passed out. I eyed them with fear and loathing. I turned mine over with extreme caution, and peeked at the questions.  There were two, and they were incredibly cute. The first one -   Discuss the prospects of ASEAN achieving borderless regional integration similar to that of the EU   elicited a squeal from me; it was so  kawaii  I wanted to pinch its cheeks. My desired answer:  Not in this lifetime, sweetie . It took me three pages to elaborate exactly why not, but the words \"not in this lifetime\" did indeed make it into my concluding paragraph.  The second one was even cuter -  Discuss world events which have led to the great influence that the US currently wields over the destinies of most countries , etc. etc. How adorable. If they were expecting me to discuss this in the light of recent unilateral action taken versus Iraq, they were sorely disappointed. I hate to be predictable.  Lunch, again prefaced with trying to avoid the moon-eyed policemen. I finished early and spent some time wandering around the local shops while I waited for the rain to ease; when it didn\'t, I took the plunge and walked back to the testing center anyway. As Philippine superstition has it that you will multiply if you get caught in the rain, I was the only one doing so.  Foreign Language was the last exam of the day - the last exam of the lot. They told us that out of the hundreds that had taken the first set of exams, only a hundred remained - ninety in Manila, which would have been us, and ten in Cebu. I stared at them and wished very very hard for a stiff shot of vodka.  Everything was going fine until I reached one of the sections -  Write about yourself - your hobbies, your family, your likes and dislikes. You may include descriptions of your school or your job and your friends. Use about 100 words . The instructions, as everything else in that exam, were in Spanish, which had been my selected foreign language. I paused, composing a short, simple essay in my mind - and then glanced back down at my paper, rereading the instructions. Something caught my eye that I had not previously noticed -  Note , the instructions said.  Please use fictitious names and dates .  Whereupon I went slightly insane. I do not know how the people at the DFA will take this, but I introduced myself as a 25-year-old man called Antonio de la Cruz, the only son of parents who had only one child because they didn\'t want a big family. Tonio, as he calls himself, is lonely so he keeps many pets - two dogs, three cats, and an anaconda. He likes to read, write, and play the guitar. He is trying to learn how to sing so he can get a girlfriend. He thinks he has no girlfriend because he is fat. He also has no friends.  At this point, I think I\'ve pretty much given up on passing the exam - they would have to be insane to let me in - but oh, the fun I had. The results are coming out mid-October, but you know what? I really, really, really don\'t care.  I\'m just glad it\'s over.   '),(25070,2238828,'04,August,2003','2003-08-04 00:00:00','    This morning, talking in his car in the stealthy twilight:   You scare me,  he said, leaning back against his seat and half-closing his eyes.  You can\'t be pinned down.    That\'s not my fault,  I said, but without conviction. I understood what he was saying.  After a few hours of sleep, I understand even better. Understanding doesn\'t help.   '),(25071,2238828,'06,October,2003','2003-10-06 00:00:00','    You find the strangest things when you go in search of your roots. I spent last night in my maternal grandmother\'s hometown, in the house where she was raised by my great-great-grandparents. There I found people who had known - not only my mother, but my grandmother and my great-grandfather from childhood. I stayed in a house that had been the meeting place of generals during the second World War. I slept in the room once occupied by my great-great-grandparents, now occupied by my great-grandfather\'s youngest sister.  This morning we went on a quick tour of my grandmother\'s hometown. I saw the town park, recently converted into a parking lot; the bakery at which she had used to buy her bread, and from which we bought bread again that morning; the church that marked the turning to what my mother referred to as the family house. In the light of day, my great-great-aunt showed me the family photographs, yellowed and stained with age.  This is your great-grandmother , she said, pointing to a solemn-faced girl in one picture.  She was your age when this picture was taken. It was the last picture we have of her, because she died soon after .  I looked at the photo and could not look away - the girl in the photo had my face, or perhaps it is the other way around.  Mother, look , I said, because my mother was distracted by other things in the house that represented happy childhood memories for her.  I have my great-grandmother\'s face .  It\'s strange to realize that I come from somewhere. It has always been difficult to trace my roots, particularly because my predecessors were as nomadic as my family is now. My maternal grandparents were from different towns in the same province; their seven children were born in seven different cities all over the Philippines. My paternal grandfather was a transplant from Cebu, the illegitimate offspring of a Cebuana and a landowner\'s son from Negros Oriental; my paternal grandmother hailed from the southernmost tip of the Philippines, in the area now well-known for its terrorist activity. They met and married in Northern Mindanao; they left the town where my father was born after a bad accident ruined my grandfather\'s business, and moved to the town where my parents eventually met and began dating.  For my sister and I - transplants of a transplant - discovering our roots is as exciting and exotic as trekking through jungles. This morning I discovered that I have my great-grandmother\'s face, with only a small mark near my chin to differentiate my face from hers. My sister\'s build and walk are eerily similar to my father\'s older sister\'s. Disposition-wise, she could be my paternal great-grandmother reincarnated. My extended arms are crooked, a trait shared by other members of my maternal family; my sister\'s fair skin is courtesy of our paternal great-grandfather\'s Spanish roots. We are an amalgamation of our past, of all these people and histories, and yet we are us - individual and independent.  Sometimes I touch my face, or look at the curve of my arms, or run my hands through the thickness of my hair, and think: I have my great-grandmother\'s face, my mother\'s arms, my grandmother\'s thick black hair. I can tell you my history, my family\'s stories, simply by cataloguing myself and the pieces of me that I can trace back to my antecedents. Sometimes I look back, wondering who I would be now, what I would have done, where I could have travelled, if my history had been different. Would I still be rei, strange fruit of a strange tree, if my parents had never left their backwoods of a village and I had grown up knowing nothing but fields? Would my thoughts run like this, what kind of people would I have had as friends, what would I have done to amuse myself, who would I have fallen in love with?  In my fiercest hopes, I am who I am no matter which historical variable is tweaked: rei the barefoot peasant girl, rei the Quezon City dweller, rei the urbanite dweller of Bangkok, rei the hip inhabitant of California, rei the sophisticated expatriate child of Brazil or Rome. Again and again, I run the scenarios through my head, and I am brought to the same conclusion. I cannot imagine being anyone but who I am now - revised slightly, perhaps, or with a different vocabulary or linguistic style - but still me, down at the core.  Like I said, you find the strangest things when you go in search of your roots. You find a host of other people, crowded into your eyes and skin and hair and the shape of your body and the way you walk. Mostly, though, if you look hard enough and long enough, you find yourself.   '),(25072,2238828,'27,November,2003','2003-11-27 00:00:00','    Happy Thanksgiving, first and foremost. There are many things I am grateful for, and there is not enough space to list them, but:  Thank God for my family, and the practical, regretless way we dismantle each home and set up anew. With them, I have learned how to accept what cannot be changed with dignity and humor; how to peel back mundanity and routine to find the extraordinary, the breathtaking, the quirky; how to live in the  now , and to take responsibility for what I have chosen, and not to waste my time regretting might-have-beens.  Thank God for my country, such as it is - for teaching me how to give; for forcing me to look past preconceived notions of love and patriotism and citizenship, and redefining them; for reminding me that art and music and culture are ways of reinventing what cannot be accepted, of rebelling without destroying.  Thank God for the people I love, have loved and fallen in love with, and the things they have taught me - that it is alright to cry, and make mistakes; the quirks and oddities of Philippine culture, which I would have never learned otherwise; how to negotiate the gap between who you are and who another person is, and how enjoyable that process can be.   Thank God for sparkly things and dark red silk-covered pillows, because they are as balm to a soul that has been rubbed raw by life in general, and ugly floral prints in particular.  Most of all, thank God for my sense of humor and for the people who tolerate (and instigate) it, because without it I would be another statistic, another quitter.  Of course I\'m thankful for things such as food and education and a roof over my head as well, but these things are the cake. I give thanks daily for my cake; I am fortunate enough to be able to save my gratitude over the icing and the sugar flowers for Thanksgiving. So I do.  Now, as for Christmas: I was thinking of writing a short story as a present for my friends, and have been brainstorming ever since. However, I do not know whether this would be a good gift, or whether it is simply self-serving (i.e., the inflicting of my torturous prose on a captive audience). In any case, I have a few more weeks to think about it, and if there are any helpful suggestions, I\'m open.   '),(25073,2238828,'26,November,2003','2003-11-26 00:00:00','    It\'s been three years, five months, and ten days.  Sometimes I\'m angry, and I don\'t know whether I\'m angry at that guy in particular, or at guys in general for being so cheap. Why do so many men ask so little of women? Why are they so easily satisfied with a glimpse of skin and a blowjob? Why are they so convinced they\'ve gotten the best of the deal when in fact they\'re walking away with nothing at all?   '),(25074,2238828,'22,November,2003','2003-11-22 00:00:00','    So, I was thinking this afternoon, as I was waiting for Solemate to show up, it\'s been a long time since I\'ve done political commentary. Pretty disgraceful for a former polisci major, especially one who lives in a country where political near-disaster happens almost everyday. What did I say before?  You\'d have to be a hermit or dead to be bored in the Philippines  or something to that effect.   Actually, that\'s far from the truth. Since I arrived here a few months ago, we\'ve had several more \'domestic disturbances\' - a rather lame attempt to impeach the Chief Justice that brought the country to a bewildered standstill (a \'constitutional crisis\', the papers called it, but in reality it was something a little stupider than that); an aborted attempt to capture NAIA\'s control tower; an anti-government rally in our central business district, staged by people who were paid PhP300 (That\'s about USD5.50) to act like angry citizens; our president and virtually the entire cabinet falling over themselves to get the autograph of a Taiwanese soap star; a Chinese-Filipino woman kidnapped and killed in a shockingly brutal manner, and the entire Chinese-Filipino community up in arms as a result.   I think I understand now why - and how - I could e-mail my friends in the Philippines, asking frantically  What\'s going on? Are you guys okay?  only to receive confused  Of course we\'re okay, what\'s the matter? -type replies. It\'s a little like moving into a house with roommates who are incurably messy. The first few months you\'re stressed out of your mind because there are clothes hanging off of doorknobs and half-empty jars of food on the table and unidentifiable objects strewn on the floor. After a while you get tired of picking up after people, or trying to enforce your standards of cleanliness, and just kind of ignore the mess as you head straight for your own room, which is organized to your satisfaction. There comes a point in time when you expect empty chip packets to crackle under your feet, when you automatically move to the left to avoid the pile of books in the middle of the floor, when you absently reach out to brush the sofa clean before you sit down. Then you go to other people\'s houses and see how clean they are, how neatly everything is kept, how pretty their curtains are.   You get to the point where nothing, not even the great relationship you have with your roomies, can get you to stay in that house  one second longer . That\'s pretty much where we are, or where we\'re headed right now. That\'s why it\'s becoming easier and easier for people to leave, for my generation to abdicate their positions of leadership before they have inherited them, to wipe their shoes on their history on their way out.   Like I said, I wanted to do political commentary. But there\'s so much to write about, and so much I want to explain and analyze and say. It\'s like asking a person  what\'s wrong with your house? Give me a list of the problems and we\'ll analyze what\'s going on and try to fix it.   And you start off by saying  Well for one thing, it\'s upside down.    Sometimes there\'s really nothing to say, after that.   '),(25075,2238828,'20,November,2003','2003-11-20 00:00:00','    Adventures for the asking, and some not even asked for. Last week I met up with A - who from hereon will be referred to as Solemate, since that is what we call each other in real life - to work on some research we\'re doing for an NGO. Not exactly the job we dreamed of doing when we were in college, but then again, working is like writing - sometimes you enjoy it, sometimes you just do it for the money. (Incidentally, Moliere compared writing to sex, but I thought I\'d be a little more circumspect.)  We met up, had lunch, ended up talking about our dreams (literal dreams, the kind you have after you fall asleep), and then took off to go to Quezon City Hall, where we planned to gather some information on our research topic.  How do we get to Quezon City Hall?  Solemate asked as soon as we crossed the flyover to the other side of Philcoa. Not being a genius at commuting and directions in general, myself, I stared at her blankly for a few minutes before turning back to survey the fleet of jeeps that passed blithely by. C\'s advice on how to get to places entered my head:  You look at the signboards,  she instructed.  Find one that has the name of the place you want to go on it, like Megamall or Cubao or whatever. Then you get on it, tell the driver to take you to your destination, and then get off when he tells you to. Simple.    But what if I don\'t see any jeeps or buses with signboards that have my destination\'s name on it?  I asked.  C\'s forehead wrinkled a little in thought, a few seconds before she brightened.  Well, then,  she said philosophically,  You\'ll just have to change your plans and go to somewhere that they have signboards for.   That advice in mind, I told Solemate:  We\'re going to get on a jeep that says CITY HALL, it\'s going to take us to city hall, and we\'ll get off.    Okay,  Solemate said, and we immediately hopped onto the first jeep with a signboard saying \"City Hall\". Now: the journey from Philcoa to City Hall is not complicated. It is a 5 minute trip along a single road; you get on at one point and you get off at another - a perfectly curvilinear transaction. However, Solemate and I are not particularly gifted at directions, and somehow managed to get lost - twice - before finally giving up and walking to city hall.  This is so stupid,  Solemate grumbled, adjusting her umbrella to block our faces from the sun.  We graduated from the best university in this stupid country, how the heck did we get lost?    God is fair,  I pointed out, concentrating on not losing sight of city hall and, heaven forbid, getting lost again.  He gave us so much brains that He figured the homing instincts could go to someone else.   At city hall, our eyes were opened to two things: first, that apparently no one in city hall is aware of what anyone outside of their own department is doing; second, that very few people in city hall actually work. One woman, dressed nattily in the city hall uniform, was apparently employed solely to push elevator buttons - which she did with much attitude. Other employees were diligently occupied in the pursuit of their own happiness, in the form of informal little coffee klatches, newspapers, measuring their children for graduation outfits, filling out application forms for loans ... Solemate, viewing the building-ful of people busily employed in doing everything BUT the tasks listed in their respective job descriptions, shook her head and expressed her gratitude that she did not pay taxes to the Quezon City government.  We headed back to Philcoa under cover of a drizzling rain, arguing amiably about the precise Tagalog term for that particular type of precipitation. At Philcoa, a bare two minutes after we got down, I felt a hand running down my butt - and it wasn\'t mine OR Solemate\'s. I whirled, smacking the arm away and getting ready to pound the groper\'s face - only to find myself looking straight into the maddened eyes of one of the vagrant psychotics who wander the streets of Manila.   A security guard, spotting my dilemma, blew his whistle and hollered at the man, just in time to keep him from trying to hit me with the bundle of twigs he held in his hand. Solemate, confused and alarmed at the speed with which I dragged her away and into safer realms, asked what had happened, and I told her.  Omigod,  she said, her eyes widening.  Solemate, that\'s so scary.   I looked at her, and we collapsed against a wall, laughing in nervous reaction.  It was scary,  I admitted,  but also funny as hell. My gosh, solemate, can we not go anywhere without weird stuff happening?   I\'ve been told that I\'m far from normal, but I beg to disagree. You cannot blame the goldfish if the aquarium is strange - and you must admit, it is a strange strange world we live in.   '),(25076,2238828,'12,November,2003','2003-11-12 00:00:00','    It\'s odd and sad and yet, in the way of things, strangely apt. In Bahasa Melayu, their word for  sweetheart  and  regret  is one and the same:  sayang .   '),(25077,2238828,'10,November,2003','2003-11-10 00:00:00','    I lost my phone a couple of days ago - or maybe it was stolen, I\'m not quite sure which. All I know is that one minute it was safely tucked away in the depths of my bag and the next it was simply Not There. Quite existentialist, really. I came home extremely upset over its loss; there was a moment of confusion wherein my mother and I had to clarify  who  exactly had the right to be the most upset over the situation. I won, naturally, since it was after all  my  phone in all senses of the word, and my mother\'s only claim to misfortune was the fact that she could no longer contact me at all hours of the day and night, regardless of my location.  It\'s been a silent two days, during which I\'ve learned that yes, it\'s true, you can actually grieve for the loss of inanimate objects. Right now I am in the denial phase of Kubler-Ross\'s five stages; I still grope sleepily under my pillow to switch off my phone\'s alarm, and perk up, my ears tingling, when I hear a tone that roughly approximates the message tone of my former cellphone. Yesterday C and I were hanging out on the lanai, bemoaning the loss of my phone;  I miss it so much,  I mourned.  I miss picking it up and playing with it or checking its calendar to see if I had anything scheduled for the day.    I miss your phone too,  C said, glumly.  I had so much fun taking pictures of myself with it.   The most annoying thing about it all is that it is not even practical for me to buy myself a replacement phone; we are moving very soon and as C pointed out, I might as well just get a phone after we\'ve moved. In the meantime, I am phoneless and have instructed all my friends to contact me either through e-mail and house phone. None of them, I imagine, are very pleased about the inconvenience of it all.  And I cannot imagine why I have devoted an entire post to a cellphone. Perhaps I am more materialistic than I thought I was.   Oh, but I honestly do miss my phone.   '),(25078,2238828,'18,December,2003','2003-12-18 00:00:00','    Thinking about paradigms, and frameworks, and viewpoints. Someone once said that in a way, prejudice is a way of life - depending on how you choose to see things, you will always automatically classify certain things as bad, or wrong. This isn\'t bad, in and of itself; it\'s merely a way of classifying the world you live in, ordering it so that you can make decisions more efficiently. Prejudice becomes ugly when people refuse to change their existing set of values and beliefs in the face of either (a) evidence to the contrary or (b) a fundamental change in their situation that renders the basis of those previous judgements wrong.  For example - to use a fairly frivolous example - I am highly prejudiced against the color orange. I do not like it, I loathe it, I think people who wear it who are not either forced to wear it (like prisoners) or paid to wear it (like say, models?) have something fundamentally wrong with their powers of color perception. This bias is fine, so long as I realize that my anti-orange sentiments apply only to me and my personal wardrobe. When I pick up a weapon and start gunning for anyone who dares wear orange in my presence, that\'s when you know I\'m two beans short of an espresso. Prejudice, in a way, is a kind of psychosis - a refusal to accept reality as it is, rather than the way it is inside your head.   Anyway. Like I was saying, I was thinking of paradigms, of the value/belief structure that accompany them, and of the massive disorientation that occurs during a paradigm shift. We may be in the midst of a major one here in the Philippines - or at least, I hope so. It would be a positive explanation for all the instability that\'s been going on lately. Just this year, there was that Hollywoodized attempt at a coup at the Oakwood Towers, led by young, buff mutineers who tattooed the symbol of a secret society onto their arms and carried high-tech weaponry (ultimate phallic symbols, yo). Tourists lined up around the barrier watching and taking pictures, and fangirls hovered around the barriers, squeeing about how cute they were (I kid you not). It says something about our culture that even our mutineers get fangirled. They claimed they had \"solid proof\" of corruption in the higher ranks and would present it to the public \"in due time\". Once in front of the Senate Committee, however, they refused to show their so-called proof, insisted on reading from a prepared statement, and - this irks me the most -  walked out  on the hearing when the Senate asked them to set aside the statement until they\'d addressed the matter at hand, which was  their hearing .   Then there was the impeachment attempt on our Chief Justice, which turned out to be more of a three-ring circus than anything else. I will not list the details here because that would take too long and be too confusing to anyone who is not well-conversed in the oddities of Philippine socio-political life. (I was looking for links to information on the whole shebang, but of course all articles are highly skewed to reflect the leanings of the writer.) Let it suffice to say, though, that the whole country was held spellbound for two weeks while our Congress revealed their massive insecurity complex about their role vis-a-vis that of the judiciary and the Senate; while learned lawmakers and legal eagles debated hotly over the practical applications of the phrase \"to file\" [a motion to impeach]; while incensed court employees threatened to take to the streets in a mass strike, although no one really knew whether the ones striking were for or against the impeachment because both factions issued similar threats.   What else, what else? About a month back there was a rally in our central business district - a rather highly-publicized event, given that it was held on one of the main thoroughfares of the area (think: rally along Fifth Avenue in Manhattan or Jalan Ampang in Kuala Lumpur). The rallyists, however, were rather unconvincing people; apparently they had been hired to man the protest for the day, at a rate of PhP300 (that\'s about USD 5.50) per person, plus free lunch. The protesters had come appropriately prepared, with pots, pans, bottles of water, and their children. They were not particularly interested in the political or economic implications of their actions; as far as they were concerned, they were being paid to be there and that\'s what they were doing.  Around the same time there was an attempt to take over the control tower of our international airport. Coming so close on the heels of the last mutiny attempt, our local armed forces took decisive action: they stormed the tower and summarily executed the ringleader with a hail of bullets. To add to the drama, the ringleader was, at the time of his death, giving an  interview  on the radio as to the reasons behind his actions. His final gasps were heard, live on air. The family and grandstanding members of the media accused the government of overkill, tearfully pointing out the bullet-ridden body of their relative and insisting that he was a wonderful person who had not deserved to die that way. The government, for once, did not back down. There were no apologies, no demands for inquiries into the incident, no removals of high-ranking officials.   What a year it\'s been, and I haven\'t even mentioned half of what\'s happened. Currently we\'re in the initial stages of election fever, with everyone hanging on tenterhooks for the next developments of the upcoming presidential race. Our incumbent president has disappointed and outraged the nation by declaring that she will run \"for the good of the country\", not six months after she had declared that she would  not  run, so as to avoid politicking. She is being contested by a very outspoken former Education Secretary with a penchant for floral shirts; a movie star who is the local equivalent of Elvis who has not really made up his mind to run, but is being fielded anyway by a local party; the former head of the Philippine National Police, who has been accused of killing a bunch of suspected kidnappers in cold blood, and is suspected to be in cahoots with them; and, more recently, a pastor who heads the nation\'s largest born-again Christian movement. Naturally there are also the usual slew of nuisance candidates, among which is a man who claims to be the \"Messiah of all Earth Heroes\". You can imagine how avidly I await the presidential debates.   Like I said, I used to see these things as nothing more than static - on a television whose antenna was broken and there were no channels to switch to. I wanted desperately to change channels, to change antennae, to throw the damn TV out and get a new one. Up to some point, I still do; there is only so much a human being can take before she loses her sense of dignity and pride. But slowly I\'m coming to look at these things as the disruption period between the existing situation and the emerging one. You can\'t build a new house without tearing the old one down first. You can\'t make improvements or install change without undergoing some kind of discomfort.   Maybe I\'m wrong, but I want so badly to think that there\'s some sort of point to this - to the fear and anger and self-contempt I see daily in the faces of my people, the way they slouch and shuffle their way through life, the way they put their trust in celluloid heroes because there are no longer any real-life ones. I want my people to have pride again, to be human, to treat each other with respect and acceptance. I want to know that the next generation - my generation\'s children - will have a culture to inherit, footsteps to follow in, positions to fill and gainful employment to occupy themselves with. I\'m coming so close to the time when it will no longer be  my  future, but the next generation\'s - and I do not want them to look at me with disappointed eyes, telling me without words that I have left them nothing to dream for.   My paradigm is shifting. I am stepping forward. If there is anything I have learned in the past two weeks, it is that my hands may be small, but they are doing something.   For everyone, and most especially for me, an excerpt from a Cowboy Junkies song:   Hold on honey, there\'s a new dawn coming / And a big bird to sweep it away / Are you born yet? / Are you listening? / Are you sick of staring at the walls? / Are you hungry? / Are you angry? / Are you wondering if there\'s anyone at all?  Hold on honey, there\'s a new dawn coming / And a big bird to sweep it away / Keep on dreaming / Soon there\'ll be a reason / To see it through one more day   Merry Christmas. Hope for me, and hope with me, because like I told a friend of mine once, I forget so often and there is no time to forget.   '),(25079,2238828,'14,December,2003','2003-12-14 00:00:00','    Okay. You blink, and the world changes.  First:  urlLink Saddam is caught , looking ages older and more tired than he did when the world last saw him. Apparently he gave himself up without contest. I wonder what he\'s been thinking, all these months in hiding while his tin empire was torn to shreds and his sons were killed.  Second: Philippine Foreign Minister,  urlLink Blas Ople, is dead . Natural causes, the report says. I believe this, because I have seen the guy in person and he smokes like a chimney and drinks black coffee as if it were water. This, while looking and sounding as if he were about to collapse any moment now.   Tomorrow the other shoe will drop, and that\'s when I\'ll start talking. Today my head is stuffed up with a cold and there\'s not much room for discourse.   '),(25080,2238828,'12,December,2003','2003-12-12 00:00:00','    So maybe the Philippines is not the most comfortable place in the world to live, but it is one of the most amusing, particularly around Christmas. That\'s when the people who provide services around the year - garbage collectors, gardeners, security personnel, janitors - hand you little white envelopes with \'Merry Christmas!\' written on them. The envelopes are, of course, empty. The implication there is that  you  are to fill them (with money, REAL money) and return them so that the giver of the envelope will have a truly merry Christmas.   The humor of the situation is apparent only when you understand that most of these service providers do not exactly specialize in spelling. I present to you, unedited, the Christmas greetings on one envelope:   Mirry Chrestmas  (Name of subdivision, also horribly misspelled)  Mentinance  Hardenirro  Mang Edde   Which, of course means, that this envelope is from our maintenance man - our  hardinero , or gardener.  Hardenirro?  my mother asked.  What\'s that, like a relative of Robert de Niro?   So, yes. MERRY Christmas, indeed.   '),(25081,2238828,'09,December,2003','2003-12-09 00:00:00','    Back on my feet, and I hit the ground running. Today was my second day at a course on  urlLink Transforming Institutions for a Sustainable Future . It\'s a distance-learning course and we get our lessons via videoconferencing from Washington, so my very first day I could barely concentrate, I was so giddy about all the cool technology. My very own computer with a flat-screen monitor! A microphone at each station that would broadcast our voices over the internet! A DSL internet connection! Lessons given on CD-ROM and interactive websites! I wanted to camp out in the learning center and bask in the utter coolness of their technology.   Today I came back to earth, slightly. The thing about sustainable development seminars is that they\'re always slightly depressing; one or two days looking at current and forecasted scenarios are enough to make me want to crawl into bed and pull the world in after me. As one of my classmates said earlier, while we were discussing the day\'s topics, there\'s so much to do and we have so far to go, and when you look at what we have to do against what we\'re actually doing, it\'s so easy to get overwhelmed. And, as one of the facilitators from the World Bank kept reminding us, there are absolutely no guarantees.  That, and it\'s slightly discouraging to figure that there is a large gap between the development strategies being promoted by Washington and the realities we see here. One very frustrated participant lashed out on the first day against what he labelled \"the arrogance of the World Bank people in Washington\" - he was upset that they seemed to be  telling  us what to do, rather than consulting, or advising. I don\'t want to judge, but so far I\'m a little discouraged. The first day was forgivable, since we were being presented with overviews and general scenarios, but I\'d looked forward to the second day because they\'d promised to give us case studies. We  did  get case studies, but they were cases conducted in Africa and South America. That\'s wonderful and the cases were rather striking, but still - Asia is a different matter altogether. We\'re still not connecting.  We\'re talking a lot now about building consensus, identifying catalysts for change inside communities, figuring out access points and lines of communication between disparate cultures and communities. Our group identified education, in particular, as a critical development issue - if we\'re going to start building the foundation for the future, we might as well develop the capabilities of the country\'s future leaders while we\'re at it.   I wonder what\'s going to happen after this course is over. All of us are going home with nifty little certificates; I\'m going to have something good to add to my resume. What happens to the people we\'re helping, though? The names and faces and stories I come across and talk about and remember - how do I translate this into something that actually affects them?  I\'m excited about all this new knowledge, and thrilled at the opportunity. But still - how tiny my hands are, and how huge and wounded this world is.   '),(25082,2238828,'01,December,2003','2003-12-01 00:00:00','    It\'s been a rough week or so. So much to do, and the computer\'s breaking down regularly and I can\'t connect my colored printer and I can\'t find my old articles and every time I think about the story I want to write I get overloaded. Sleep is hard to get, nowadays. I work until past midnight, until my eyes are drooping and I can no longer keep my head up, and still when I go to bed, after I close my eyes, I cannot go to sleep. The other day I fell off relatively early, only to wake jarringly early the next morning to a horrible nightmare that still haunts me when I think of it.  I\'m so tired. This afternoon I was standing in the MRT station, waiting for the train to arrive, and for no reason at all I started crying - no, that\'s too active a word. I was standing there waiting and tears started spilling, and I started wiping them away, horrified, and they just kept on coming. All I could think was  Please let me not break down in an MRT station where I cannot even call my mother to pick me up, please let me get to a phone first  and eventually I managed to pull myself together.  It\'s so hard to explain this sadness. I don\'t understand it, myself. Most days I\'m fine. I like it best when I\'m happy, when I can turn my face toward the light and feel like I\'m drinking it in, when small things can make me laugh, when just walking is enough to make me glad to be alive. Other days aren\'t so great - there are moody days and irritable days and bad-hair days and fat days and PMS days, when I feel less like Snow White and more like one of the Seven Dwarves.  And there are the bad days, when I wake up and cannot get out of bed. Last week was full of those bad days, a week so unproductive that every time I think about it, I am ashamed. I would go to bed early every night and wake up late, and then pull the covers over my head and sleep some more. If I left my room at all, it was to grab a glass of water or to go to the bathroom; then I would go back and lock the door. All the time I felt like I was drowning.  C suggests Stresstabs, but I\'m not sure that popping pills is the best solution. At this point, I am just about ready to do  anything  to stop from almost-breaking down in public again. Shuffle around like a slug, yes. Nightmares and insomnia, fine. Loss of appetite - well, I need to lose weight anyway. But cracking up in front of half of the Philippine population - no, no, no, a million times  no .   '),(25083,2238828,'29,January,2004','2004-01-29 00:00:00','    Sometimes coming back is an enjoyable thing, despite all the bitchy things I\'ve said about returning. Tuesday I was asked to speak at a career orientation program for juniors, in the high school I graduated from. One of my friends from that school is now one of the high school guidance counsellors, so she gave me a last-minute call to inform me that she\'d put me on the program as a resource speaker for the field of political science. It was an invitation I couldn\'t turn down - because of my friend, and because of the fact that I have fond memories of that high school.   Tuesday morning I set off bright and early, with a two-page outline/summary of the talk I intended to give on the field of political science (which I had hastily written and printed out thirty minutes before I left home). An hour after I arrived at the school, I was semi-panicking and wondering if I could still back out, for many reasons:  (a) Two of the other resource speakers had come frighteningly prepared, with professional-looking PowerPoint presentations and about ten pages\' worth of speeches. Another one was not quite so amazingly prepared, but then again she was my year\'s valedictorian, so I supposed she had just stored it all in her mega-brain.  (b) It turns out we were going to be speaking to the ENTIRE JUNIOR CLASS for three hours, from a panel-like set-up onstage. I peeked at the table and was a little taken aback to see that the front skirt had a tendency to get tossed up by the wind from the electric fan, thereby giving the audience intermittent peeks at our legs. I happened to be wearing a skirt (a respectable beige knee-length one, which was paired with a matching jacket, but STILL) and had chosen not to wear stockings. Imagine sitting up onstage, trying to talk to a bunch of teenagers while worrying about whether they\'re checking out your bare legs. Oy.  (c) Not to be insulting to my own course, but political science is just not the sexiest career option when you\'re fifteen or sixteen. I was up there with resource speakers for careers in business, medicine, engineering, mass communications, the arts, and music. To make things worse, the guy speaking for music turned out to be the former guitarist of one of Manila\'s more popular rock bands. How was I supposed to compete with that?   At any rate, I got through a trifle more than respectably. Perhaps it was my legs - I have been told I have nice ones, which I appreciate because not all parts of my body are equally nice. Perhaps it was the fact that I complimented the school organ, as it turns out that the current editor-in-chief is a junior - although to be fair, I did not know that at the time, and it was an honest compliment because I was that paper\'s first ever editor-in-chief and it *so* did not look like that during my time. While I was talking there were students who actually appeared to be listening - as opposed to chatting, making eyes at their crushes, or listening to CDs - and I got a few questions during the open forum. Best of all, I got a surprisingly enthusiastic round of applause as I walked across the stage to get my certificate of appreciation. My former vice-principal, who gave me the certificate, was grinning.  This girl,  she told the students as I hugged her and my former guidance counsellor,  used to do nothing but read books in class WHILE her teachers were lecturing. Her math teacher used to confiscate them and throw them out the window, and she would sneak out of class so she could get them back.    While we\'re recounting previous sins,  I added, also into the microphone,  You might as well tell them that I didn\'t attend classes either.  I suppose this endeared me to the black sheep of the class, because now I have fans. Go me.  The best part, however, was at the lunch afterward. I was chatting with the vice-principal, reliving my much-spotted past, when she told me:  Oh, but you know, your batch was the one we loved the most.  Which was touching in the extreme, since that was a difficult time for me. It was nice to know that someone remembered it - and me - with fondness.  Yesterday afternoon I went with my dad to a meeting he had with my former boss. I sat there and took notes, which is what he brought me for, and afterwards we all had a brief chat about KL and what had happened to all the people and projects we\'d left behind.  You slimmed down,  my former boss observed, as we headed out for the elevator.    I\'m wearing black, sir,  I pointed out.  But thank you very much.  It was almost fun to see him again, and to be able to sit back and observe all those little work-related idiosyncrasies I\'d gotten used to after a year of working with him. He is not the type of person you get close to, but he was a relatively nice person to work for. Sitting there, I had flashbacks of all the good times I had at the office - mostly with my ex-colleague PH, as we argued over what wallpaper to put on our computer and what music to play when he was out of the office.   If only coming back were always like this - pleasant and undisturbing, with something new and productive built on the foundation of the past. This is what I *wanted* to do with those old friends: catch up on what\'s happened since, find the places that we have both moved on to, and meet there again, as equals. And then again, to quote Jeanette Winterson:  Never say that something has moved you if you are still in the same place.    Ah well. Tomorrow I will be writing about Mindanao and separations, in response to something G and I are corresponding about through e-mail. In the meantime, I have learned that not all reunions are bad things - something which I knew intellectually, but then again none of your lessons really make sense until you\'ve lived them.   '),(25084,2238828,'26,January,2004','2004-01-26 00:00:00','    From a letter I wrote W, back in July of 2001:   His name is Martin, but we call him â€œjungle manâ€ because he is tall and dark and hairy. Since the first class, he has had a big, bushy beard that looks like Amazons could get lost in it, and he always wears brightly-colored tropical shirts, khaki pants, and well-worn sandals -- all of which, of course, makes the \"jungle man\" nickname seem all the more appropriate. He has a big, brusque voice which is always strangely kind, and he always looks as if he\'s either laughing, just stopped laughing, or will laugh in a moment. When I went for classes last last week, he had, amazingly, shaved off his beard (leaving only his mustache) making him look smaller and younger. Martin is in the Navy, and is perpetually being sent off to fight the Abu Sayyaf and various terrorists in the Southern Philippines, and he delights in telling me about my \"weird countrymen\". (I say:  They\'re not weird, Martin, they\'re criminals.  He says:  They\'re weird criminals.  I concur.) He played basketball once with Filipinos and got his ass kicked. We argue all the time in class because his worldview is the total opposite of mine, and he\'s one of the few classmates i actually have fun with.   Is it strange that all my class buddies are army men? should we be concerned that i have a weird affinity with people who know how to use a gun or drop bombs off planes or fire torpedoes? ...  Itâ€™s July and itâ€™s dusty in my office and the muezzin calls in the afternoon are beginning to drive me insane. (Heâ€™s only supposed to call them to prayer; instead he ends up singing the whole freaking prayer.) Iâ€™ll come home when Iâ€™m real again because I looked in the mirror this morning and there I was, fading at the edges.   It\'s funny because nowadays, I can wholeheartedly concur with friends who tell me I was crazy to put up with the endless back-and-forthing of that relationship. I was barely eighteen when I fell in love with him, and halfway to my twenty-third birthday when I finally gave up. That\'s almost five years, an oddly apt period of time. One of the local rock bands had a song called \"Five Years\", which was about a relationship that had failed. It had no lyrics except for one line:  But he will never be back , the singer moaned, the rest of the song sounding like a scream of despair and frustration. It was not quite that hard for me, because of all the things I am good at, I am excellent at goodbyes. The last time I spoke to him was on the phone, and I knew when I put down the phone that I would never speak to him again in the same manner.  Then again, I look back at letters I wrote him (because I have lost almost all of his letters to me) and remember how much I loved him, how much more interesting the world became when I knew I was going to be talking to him about it. It\'s almost like looking back at another person. Who was this girl who loved him, who wrote letters that were more private than her journal? How I felt about him comes off the paper in waves. What happens to all that emotion after a relationship ends? How does the heart and mind deal?   I wonder, because although the door is closed now, I don\'t understand how it could have been so easy.  Recently I closed another door - slightly harder and with more antagonism than I did the last one.  Why are you being so mean?  he asked, looking bewilderedly at me through the dim lighting of the bar we ran into each other in. Half an hour ago he had come over to our table at the insistence of my sister, and I had blown cigarette smoke in his face and turned my back on him while he was trying to introduce me to his brother (although I did shake hands, politely, with his brother).  Why are you avoiding me?  I asked back.   (Granted, that was a rather dumb question since every time we\'d run into each other recently, I\'d been - shall we say - less than friendly. I would not exactly be eager to talk to someone, for example, who talked to me in a tone of impatient dismissal.)  At any rate, over is over and I am moving on - although much less inclined to be nice this time. The problem with nice, as I see it, is that people assume they can walk over you - and then they get upset when you reach the end of your rope and begin to snap back. I am more inclined now to adopt my sister\'s philosophy - stick first, *then* the carrot. It worked quite well the night before last, when I met up with my sister and her friends at a bar - although I cannot say it bodes well for the reputation of Filipino men that they appear to be more appreciative of bitches than nice girls.   The process, however, has been undeniably therapeutic. We bullied our way in past the bouncers and got in for free. We did not pay for a single drink, since there were plenty of people who were ready, willing, and able to buy them for us. We would scan the crowd, looking for people we wanted to dance with, then - depending on our individual inclinations - would crook our fingers or make eye contact and summon them with a head gesture. Surprisingly, everyone we summoned came. We dismissed people with equal casualness - my sister by waving her hand, myself by simply turning around and walking away.   In the end, I walked away having given my number to a rather slow specimen whose parents had given him a computer shop (perhaps to ensure their progeny would be employed, as it did not appear he would be able to sustain work in an environment that was any more competitive than nursery school). He is taking me and my friends out this Friday to another bar, and I will see where it goes from there.   In the meantime, I don\'t suppose I\'ll fall in love in Manila again. The agony is *so* not worth the ecstasy.   '),(25085,2238828,'23,January,2004','2004-01-23 00:00:00','    Well. It\'s a belated greeting, but it\'s meant in good faith nevertheless. I meant to post about it yesterday - which is when Chinese New Year started this year - but the Internet connection was acting up, and I didn\'t feel like getting into another altercation with my PC. I was at a mall yesterday and they had a dragon dance, but no one really seemed to know what it was for, so everyone just kind of gawked and then walked hurriedly by. It did not help, of course, that apparently no one had thought to clear a space for the dance, so the dancers were limited to doing weird little hops from side to side.   There are times I miss KL, and this is one of them. I love dragon dances. They were a big thing in KL, with the dancers doing various acrobatics and stunts while in the suit; PH and I used to go out on our lunch breaks during the CNY season just to look for the next scheduled dragon dance, since each public area would schedule its own. Having the dragon dance performed somewhere is supposed to bring the place luck and prosperity, so if we were really into it, we could get to watch one new one per day.   Yesterday I went out as well - with an old friend from high school. We met up at Starbucks and ended up chatting for a while; then he got called back to work and I headed back home. It was rather uneventful, as outings go; I ended up talking seventy-five percent of the time, my friend stared at me as if I had grown tentacles, and I ended up feeling very uncomfortable.   Sometimes reunions drain me. The inherent assumption in meeting up, after all, is that there will be an equal exchange of news, of viewpoints - that you will approach each other as old friends who have travelled different paths, but have not forgotten your common experiences. What happens during my reunions is far from that. Often I come ready for a major chat session - eager to find out what\'s gone on in my friends\' lives since I\'ve been gone, and equally eager to share what\'s happened in mine.   What actually happens is disappointing. More often than not, my friend assumes that my life has been far more interesting and active than his or hers, and withdraws:  Oh,  they say disparagingly,  My life? Same old thing. In the same job, dating the same person, hanging out at the same old places. I\'m sure you have far more to tell.  And although I  do  have plenty to say, that shuts me up more effectively than a gag. It\'s like walking blindfolded into a room where all the lights have been shut off. You have no idea if they can relate to what you are saying, if your words find purchase, if this person you\'re talking to is still the friend you used to share confidences and bitch with. Under their intent stare, you feel less like a companion and more like a performer; your stories feel less like sharing and more like public entertainment. You have been placed on a stage with a spotlight, for a show you have not rehearsed for. Even worse, you don\'t know what your audience expects, or what they want.   It makes me wonder, sometimes. In the Philippines, did I really have friends, or did I only have an audience? Because so often I meet up with my so-called friends, get called upon to display myself and my \"exotic\" life, and come home feeling wrung-out and stained, used like tissue.   It hasn\'t even been one month into the New Year, and already everything feels tarnished.   '),(25086,2238828,'16,January,2004','2004-01-16 00:00:00','    Not to distract from the bitching, because in certain situations it can actually be a Very Good Thing. In this case it is, since it is all that stands in between me and a lot of psychosis-induced homicidal casualties.   However, let me just note that  urlLink the new Blade of the Immortal manga  is ridiculously overpriced in ComicQuest. I was willing to pay up to US15 (that would be about PhP 825) for it - in KL it always cost about USD14, so I figured that was a reasonable budget. But no. The newest issue is TWENTY BUCKS. That\'s ONE THOUSAND ONE HUNDRED PESOS. My golly. I could buy a bottle of vodka or tequila for that much, and it would last a lot longer. Never mind - I will wait until (a) the store lowers the price to a reasonable level (something that does not overshoot the minimum daily wage by QUITE so much) or (b) I can buy it somewhere else.  By somewhere else, of course, I mean Not Here.  In the meantime, WOE. Also, ANGUISH and DESPAIR. For - according to the synopsis - Magatsu has teamed up with Manji, and Shira is chasing them, and there will be much kick-ass swordplay. [spoilers - highlight to see]  Magatsu finally confronts Shira for what he did to that poor little whore a few issues back, something I have been waiting forever for Samura to bring back into the storyline. Oh, and Anotsu gets hitched!  [/spoilers] Woe.   Damn you, you crass evil forces of commercialism and corporate profit. Damn you.   '),(25087,2238828,'13,January,2004','2004-01-13 00:00:00','    I have many many things that I need to bitch about, not the least of which are the following:  > The upcoming elections. Who are the people I will have to choose from for my country\'s next president? The incumbent president, who is running for reelection (despite the fact that she promised she wouldn\'t) because she claims it is her duty? A former minister of the department of education, who was relieved of his position due to suspicions of corruption, and who is now making grand statements about being the only candidate without a \"bad record\"? An actor, who may or may not be a Filipino citizen, dropped out of high school after his second year, and has strong links to a former actor-president with questionable moral standards? A former police chief and current senator, who is strongly suspected to be in cahoots with local crime groups? A businessman who ran for senator in the last electoral race and only got 456 votes? The leader of a local evangelistic movement who claims that he is the only alternative to corrupt leadership? (Note that I am not mocking his so-called Divine Mandate because hey, when I get *my* Divine Mandate, I do not want anyone mocking me.  Capisce ?)  You know it might just be me, but I\'m not really seeing a winning proposition in the bunch.  >  urlLink Philippine election proceedings . Not only are our candidates a bunch of laughs (we had one nuisance candidate who claimed that Bush was her husband - Laura honey, wake up, your man\'s got some explaining to do), but the proceedings themselves are a joke. Apparently the Commission on Elections (COMELEC) spent about USD 15 million of the taxpayers\' money on equipment that cannot even be used because - among other reasons - they weren\'t purchased according to proper auditing standards and regulations.   This makes me almost glad that I am jobless and not contributing to taxes. Hello, if you\'re going to fritter away money it better not be MY money, you bastards. Goddamn use your own.  > The public transportation system in the Philippines. A friend of mine from India once told me that in Mumbai, their railway cabs are only meant to hold about 50 people apiece. However, due to overpopulation and the fact that 80% of the population lives on one side of the line and works on the other side, what actually happens is that the cabs are regularly filled to about four times their capacity. Hence, he said, when there is a railway accident, the death toll averages about 700-1000.   The situation here is not quite that bad, but IT IS CLOSE. I say it is not that bad because none of our railway cabs have derailed. Yet.  > Having twenty million relatives living on and off in your house (but mostly on). This has absolutely nothing to do with the fact that I love them, but GODDAMN. When 75% of the trash isn\'t even yours, when you cannot put something into the laundry without worrying about whether it might be sorted into someone else\'s cabinet because damn how can you tell the difference anymore, when you don\'t have anything to eat because the stuff that\'s being cooked has to cater to the general majority and it\'s something you cannot stand to eat - all of this IN YOUR OWN HOUSE - then it becomes a major problem.  It becomes even more of a problem when they look like they\'re settling in for the long haul, and *you* are struggling to find a job to support yourself because you\'re worried about being a mooch. I\'m just saying.  > Taking care of my nephew, who has ADD, because his mother is so consumed with her job that she has no time for him. This is understandable until I come to the point where she leaves just after he leaves for school and comes home after he\'s gone to bed and the major onus of childrearing - the tutoring, watching-over, following-up, cuddling, disciplining - falls on whoever is at home, most often me. Not to sound completely heartless, but HE IS NOT MY CHILD. I cannot comprehend why I am the one sacrificing my plans so someone can be home with this kid, and why I am the one who has to buy the kid slippers when his old ones break or help him clean his shoes or nag him about his assignments being done.  I am not a single mommy I do not want to be a mommy I have taken every single precaution I know of just so that I will not have to be one and you can be sure that I will definitely be getting that hysterectomy sometime in the future. So WHY am I wiping this kid\'s snot and making sure he goes to bed on time when he\'s not even mine? WHY?  > The move to the US. Are we or aren\'t we? Only God knows. I am not, by the way, taking His name in vain because at this point, honestly only He knows. If anyone has a clue or can get one by any method available, please drop me a line so I know whether to rip open all our remaining boxes and find my beloved leather jackets, or to restrain myself and leave them alone so that I won\'t have to repack them. Please. I am going through leather-jacket-withdrawal, and it\'s not pretty.  >  Leche flan . Do not even get me started on how difficult it is to find a  local  dessert in its home city. It\'s caramelized sugar poured over a mixture of egg yolk and condensed milk, for pity\'s sake. What is so difficult about making it, and why can I not find a place that serves  leche flan  that does not taste like sweet eggy cardboard?  > My love life, which sucks big-time ass. The only reason I haven\'t been writing about it is because it\'s so sad I can barely believe it\'s mine. The cognitive dissonance involved in knowing that you are an intelligent, rational human being who nevertheless manages to get tangled up in relationships where you are ignored for months on end because the person in question is \"too lazy\" or \"too busy\" or too whatever to try to contact you is mind-boggling. What, do I have a gigantic tattoo on my head that says \"It\'s okay if you just walk away, I have metal ovaries! I won\'t be too torn about it, so neither should you!\"   I can\'t even laugh about this, it hurts too much.  ---  At this point I am capable of writing entire sagas about each and every one of these things. Unfortunately, my energy limits me to doing only one, so I will have to deliberate.  In the meantime, yes, I\'m drowning.   '),(25088,2238828,'13,January,2004','2004-01-13 00:00:00','    I was originally not planning to do this, but something about the  urlLink Mayfly Project  is inherently appealing. Looking back at the year that just went by, counting the costs and the benefits, remembering the moments that dance in red, and the moments that writhe in the shadows behind your consciousness - and realizing that at the end of it all, you\'re still standing. I get such a kick out of doing that - it\'s like looking all my failure in the face and still having the strength to laugh with real humor.   The year before last was all about forgiveness - about walking back, opening all my firmly locked doors, and letting go of what I\'d hidden there. The funny thing about the mind is that it hides the things you don\'t want to keep, the way normal people hide away their valuables. Childhood monsters, bogeymen, memories of all the mistakes you made and the people who suffered for them. My monsters didn\'t really disappear the moment I let them out, nor were they any less scary or painful, and there are still lots of locked doors. The thing is, I\'m in control now. The keys are back in my hand.   As for last year - last year was about moving on.    Looking for home in three countries A heart that folds neatly into a suitcase And friends that do not.   And there, my friends, you have my 2003 experience in a nutshell.   '),(25089,2238828,'10,January,2004','2004-01-10 00:00:00','    I shouldn\'t be here - I should be industriously at work on my very last MBA paper, which I plan to submit this afternoon at around 3PM (via e-mail, so I\'m not really that rushed). My initial enthusiasm for this paper, which was based on the fascinating story that is  urlLink Google , has waned since I\'ve had to rewrite this paper TWO TIMES already. On the other hand, I honestly cannot see myself throwing away two years of studying all because of one stupid paper, so very soon I will return to my drudgery.   In the meantime, I have discovered that the Philippine Embassy in Kuala Lumpur really doesn\'t update their website very often, so you\'d be better off not really relying on the information you get from it. Hard evidence can be found  urlLink here , where they list the names of people who no longer work there, and have not been working there for at least two months.   '),(25090,2238828,'05,January,2004','2004-01-05 00:00:00','    I may be late in noting this, but did you know that if you typed the words \"miserable failure\" into the Google search box, the top result is George W. Bush\'s official biography?   No,  urlLink really .   '),(25091,2238828,'05,January,2004','2004-01-05 00:00:00','    Two holiday weeks in Bacolod, a little jewel of a city nestled away in the Visayan islands that is filled with the unique, the quirky, the abominably delicious. My parents and I (my sister was working and couldn\'t come with us) left it ten pounds heavier and stress-free. The things I saw, and have stories about, and am bursting to tell of:  > A family graveyard, built in the MIDDLE of one of the city\'s major roads;  > A street lined with small mini-restaurants (and I use that term extremely loosely to describe a group of wood-and-concrete stalls differentiated only by the quality of their chicken and the posters that line their walls) that all serve the same thing, chicken inasal (barbequed chicken) and rice. Aside from a small selection of drinks, all of the establishments on that street serve nothing else. As Poe says (no not THAT Poe, Filipinos, the other one): \"Only this, and nothing more.\"  > Playing badminton with my aunts, discovering that I am execrable at it, and that to make up for my lack of skill in sports, I am extremely talented in picking up the nuances and structure of sports rules and scoring. Go me. Although I *did* fare rather better when we weren\'t playing with the Tasmanian Devils of Badminton anymore - their serves must have been whizzing across the net at, I don\'t know, a few tens of miles per hour?  > Two restaurants - Bob\'s and Pendy\'s, that are local cuisinary legends and have expanded on the basis of their respective specialties. Bob\'s is a rather all-around bakery, although it prides itself most on its chocolate cookies/cakes and its nummy Fruit Punch, which is the first juice I have ever had to chew (because it had chunks of actual fruit in it). Pendy\'s grew famous on the strength of its Half-Moon Cake alone, which should tell you how evil and rich that cake is.   > An eight-year-old girl who can, upon the slightest urging, recite reams of poetry and Bible verses, recall the geopolitical history of Singapore, and accelerated through three grades of piano lessons in less than a year.   > The dogs at my great-uncle\'s house, of which there were legion. There were three puppies that were initially shy and grew increasingly more shameless as my mother and I began feeding them; there was their mother, who intensely disliked breastfeeding her offspring; there was a neurotic brown dog, who seemed to have ADD since he could NOT eat while standing still; and there was a shedding and itchy dog named Blanc and renamed Scraggly by one of my aunts, because by golly he is. We teased my other aunt, who is a dermatologist, about how ironic it is that she has a dog with a skin disease.  > Hearing my mom being interviewed on a local radio station, and getting dragged in for a very short follow-up interview myself. For the record, they wanted us to tell them how we had become such a solid, model-family unit, and I had to work very hard not to say things like  By working very hard at loving each other through our ugliest times, and by not airing our dirty laundry in public.   > Visiting a local holiday spot, the Christmas Village - which is actually a personal collection and arrangement of a HUGE array of Christmas-themed clay miniatures. The Village was gradually built up over a period of eight years, and is set up inside the home of the artist himself. The artist, unsurprisingly, is a fiftyish man who is still single.  There is much, much more - bonfire night on New Year\'s Eve, speaking to the local church\'s youth group on commitment (which was ironic in many ways, one of which is the fact that I\'ve never been in an actual relationship), and trying desperately to stay up for the traditional Noche Buena dinner only to give up and go to bed instead, and find later that everyone else had had the same idea. My problem, as I often find when I return to blogging after an extended absence, is not that there are no words but rather that there are too many.   On the edge of New Year\'s Eve, I stood out in my great-uncle\'s driveway with my mother and one of my aunts, holding sparklers while everybody else watched sedately from the covered garage. As fireworks pulsed overhead and I watched star-shaped sparks falling from the wand in my hand, I wondered why we choose light and flame and noise to signal the advent of a new year. Why fire, a potentially destructive force? Why not eggs and flowers and clay earthenware filled with soil, symbols of fertility and new life?  It\'s occurred to me since then that maybe that\'s how we want to see the New Year: that with the passing of a year, we can burn away all we were, all we used to be, and become completely new. Not for us the gestation and gradual emergence of conception and birth: we don\'t want change to take time, to have to remember that our resolutions will have to be imposed on a life that runs on crooked wheels, in grooves that are pebble-filled. We want to immolate ourselves and come out completely different, changed beings. Like I told the youth group in that small church, we like to think that when we change circumstances, we can change who we are - that we can become something completely different from who we are now.   Or maybe I\'m generalizing, and that\'s how  I  want to see the New Year. So many mistakes, and so many false steps, and too many times when I turned my back on the right thing to do what was easier to do at the time. Given such an overwhelmingly bad record, sometimes you feel as if it\'s easier to set fire to the whole damn thing and just reinvent yourself -   Although that\'s not always a bad thing. In a weird, hidden-between-the-lines sort of way, it sometimes means that you still hope for something better, that you still dream. Maybe you want to take shortcuts to get there, and maybe your hopes aren\'t always realistic, but hey. To borrow from a song,  don\'t dream it\'s over .  On that note, a belated Happy New Year\'s.   '),(25092,2238828,'20,February,2004','2004-02-20 00:00:00','    First: the return of the quiz! You know you secretly love them, even though they are the Internet equivalent of Britney Spears\' songs. Here\'s one that isn\'t from Quizilla, for once:          You\'re    the United Nations!      Most people think you\'re ineffective, but you are trying to completely save the world from itself, so there\'s always going to be a long    way to go. &nbsp;You\'re always the one trying to get friends to talk to each    other, enemies to talk to each other, anyone who can to just talk instead of  beating each other about the head and torso. &nbsp;Sometimes it works and sometimes it doesn\'t, and you get very schizophrenic as a result. &nbsp;But your heart is in the right place, and sometimes also in New York.   Take the  urlLink Country  Quiz  at the  urlLink Blue Pyramid       My heart is in New York! Whee! So true. Also -        You\'re  Catch-22 !   by Joseph Heller     Incredibly witty and funny, you have a taste for irony in all that you see. It seems that life has put you in perpetually untenable situations, and your sense of humor is all that gets you through them. These experiences have also made you an ardent pacifist, though you present your message with tongue sewn into cheek. You could coin a phrase that replaces the word &quot;paradox&quot; for millions of people.      Take the  urlLink Book Quiz  at the  urlLink Blue Pyramid .     Strangely, I have never read  Catch-22 , although it was floating around my cousin\'s house a few years back when I was there on vacation. I will read it now because the quiz said nice things about my sense of humor, which only tells you how vapid I can be.  Speaking of vapid, I am currently deep in the throes of obsession with, of all things, a couple of reality shows:  urlLink American Idol 3 , and - oh, this is embarrassing -  urlLink For Love or Money . American Idol I understand, because hey, good music, kick-ass voices, and a judge who says what we are all thinking - what\'s not to love? Plus, I have a soft spot for Camile Velasco. I liked Bao Viet too, but apparently he didn\'t make it to the top 12. Damn.   As for the latter show, I cannot explain it. Perhaps it is the fact that all the girls - well, the ones that are left, anyway - are insanely nice to one another. I have visions of the show\'s producers scratching their heads and trying all sorts of behind-the-scenes gimmicks just to get a little catfight started, because we all know that\'s what gets the ratings. And yet the girls continue to be incredibly sweet, with their late night chats and their helping each other to get ready for dates - with the same guy.   Still, these are shows that manage to drag me out of lethargy and get me so excited I literally hop on the way to the television - not an appealing mental image, but a true one nonetheless. Again, I say - Camile Velasco, you cutie. Apparently you just had a bad week but managed to squeak through to the finals. I agree with Randy - bring back the hiphop, girl. You can do so much more with that voice than - what the hell -  One Last Cry . Only drunk men who cannot even get paid sex should sing that song. If you are like me and love this show beyond reason, check out  urlLink this place  because they have pretty funny recaps of what went on.  American Idol , the show that frees your inner bitch. Me-ow. Compared to Simon, however, I am the tweetest puddytat in the world. Yes, even on my bad days.  In a completely unrelated development, I just got hired and they are now drawing up the contract. I am no longer a complete drain on my country\'s resources! Whee!   '),(25093,2238828,'12,February,2004','2004-02-12 00:00:00','    I attended a political rally yesterday - the first one I\'ve ever attended, and not the last one, I hope. Like I told Yaz and Lin in chat earlier, I hope to get a chance to attend at least one political rally per candidate. That way I get to hear everybody\'s blustering rhetoric, see things from all sides, and make an informed decision in the end. Plus, I get to observe Philippine political behavior upfront - whee! I hope I will be allowed to, because for some reason my parents are afraid that in the course of doing so, I will get myself trampled upon or spraypainted or arrested or hit in the head with a rock. I, on the other hand, labor happily under the delusion that I am too smrt - er,  smart  - to get caught up in any of these idiocies. Well, we shall see.  Since I have no wish to be cryptic, I will reveal whose political rally I attended - Brother Eddie Villanueva, the evangelical church leader who claims his mandate came from God. From a cynic\'s point of view, not a very good line - the same claim has been made by both Arroyo and Roco - but then again apparently God is a just God and does not play favorites. Everybody is His candidate, with a few exceptions. I note that Lacson, the believer in savage justice, has not claimed any relationship with the Divine - perhaps he is afraid he will be struck by lightning. Neither has Poe, but then again Poe and his supporters are convinced that they need no one, not even a Deity, to be popular. As for Eddie Gil, I suspect that when he asked God to endorse him, God could not answer through his gusts of roaring laughter. So be it.  Pro points: The rally started ON TIME. We were told that it would start at 6PM; my parents (active supporters of the Bro. Eddie ticket - yes, he really *is* called Brother Eddie) insisted that we leave the house by 4PM. (This is one of my sister\'s and my pet peeves: that our parents always, ALWAYS leave for their appointments ridiculously early, in a region that does not take punctuality seriously. It is very annoying to arrive at a venue at least thirty minutes before the appointed time, only to find that the event will start at least two hours late.) At the last minute I decided not to bring a book, because after all, it would be my first time at a political rally and there would be better things to see. As it turns out, at exactly 6PM - by my watch! - the emcee picked up the mike and the program began. It was exactly three hours. Again, at precisely 9PM, the whole thing ended.   Other pro points, which count for less because punctuality in the Philippines makes me giddy: No sexbomb entertainment, which means no scantily-clad girls shouting inane political catchphrases while waggling their booties at us. (Yes, even if there are children in the audience. Which is why the Janet scandal did not touch the Philippines in the least.) The entertainment was QUALITY entertainment - Kuh Ledesma was one of the emcees - and all of them were there of their own volition (read: they were not paid). There were no stupid political songs, although we had to suffer through a few renditions of the Bro. Eddie theme song. The ten-point program looked sound, and the basic platforms of government being offered were well-thought-out and logical.  Con points: Well, the campaign colors are particularly violent shades of yellow and green - yes, that\'s a bad thing, visually speaking, since the shade of yellow they picked does *not* suit Filipino skintones at all. Which is to say that even if I do end up supporting Brother Eddie, I will not be wearing his t-shirt. (I will be wearing Lacson\'s, because his color is red.) The premise of the campaign - and indeed, of Villanueva\'s reason for running - still seems incredibly naive to me. Apparently he ran because he saw no viable alternative for the country, and therefore offered himself up. I do not doubt his sincerity - I do not doubt he  wants  to do a better job. However, translating that sincerity and passion into actual policies - will his motives prevail, or will he be unwittingly led astray by the power grabbers, seekers, and manipulators that will inevitably litter his path? He is also surrounded by good-hearted but equally naive people like my father. Watching them campaign and plan and strategize is like watching a kitten toddle into a den full of hungry lions.   I fear for them.  The official campaign period is in full swing - and when I say that, I am picturing monkeys and all manner of wild animals dangling from the chandeliers, bouncing off the walls, cartwheeling across the linoleum. In short, I am visualizing pure chaos, a literal mad carnival. I\'ve already recounted Villanueva\'s campaign, now as for the rest:   > Arroyo and her incredibly arrogant running mate Noli de Castro (a popular newscaster who thinks he is God\'s gift to the political landscape) kicked off their campaigns in the highlands, due to a superstitious belief that they will be able to conquer all the areas they can gaze down upon. Arroyo made cracks about her diminutive size, somehow relating her dwarfishness to the oppressed lives of the poor; de Castro smiled and made much of his provincial beginnings. Apparently, he came out to the big bad city as a naive farmboy with holes in his clothes and nothing but flipflops on his feet. Naturally, he did not mention the big expensive bash he threw recently for his adopted daughter\'s wedding, at a pricey garden venue inside Quezon City.   > Roco, in traditional floral shirt, went to Cardinal Sin\'s house for a mass, accompanied by his running mates, his wife, and his mommy. He avoided the current Manila archbishop\'s house - not because said archbishop has openly refused to endorse any candidates, he said, but because he didn\'t want the visit to be construed as \"political\". Instead he goes to the more politically-inclined priest to hear mass, and from there proceeded to Bicol, the province in which he has the strongest support. He toured elementary schools, signed autographs for teachers (who will be the vote-counters this May), and had the traditional \"simple meal\" (rice with fish, eaten with bare hands - considered a poor person\'s meal, here in the Philippines) complete with media coverage. Before you ask, that still does not make him a traditional politician because he too was one of those poor little farmboys who got sent to the big bad city to make good.   One day I will make a sign that says \"A traditional politician is what YOU are, not who your parents were.\" Your father may have been illiterate and your mother a struggling schoolteacher, but they are not the people we are voting into position.  > Lacson, in a red shirt, travelled down the Diosdado Macapagal highway with a motorcade adorned with red-and-white balloons. He labelled the highway \"the most expensive highway in the universe\", which immediately made me wonder if he had actually managed to obtain stats on highways that were built in Outer Reticula. We should probably hire a Klingon to examine our construction budgets next time ... but I digress. He claimed he was doing something called \"nontraditional campaigning\", which must have been the reason why there were absolutely no people flooding the streets to meet him. The media followed him down a street as he walked and randomly shook hands with people, all of whom looked more excited about being on television than meeting Lacson. Some looked slightly bewildered, but I will not dwell on that. I am sure the Filipino voting population is aware enough to recognize Lacson by face.   > Eddie Gil! Give me a moment while I try to stop laughing so hard ... okay. I cannot say anything about his campaign because I was too blinded by tears of laughter to watch. For one thing, his hair. He claims that he is worth USD40 billion - therefore he should be able to afford a better wig than the one he is wearing, or at least a very good hair weave! On the other hand, perhaps he idolizes Elvis Presley, in which case I have no room to mock because in the past I too imitated the hairstyles of my favorite stars (fortunately, Lea Salonga is not given to outlandish coiffures). For another thing, if he is worth that much he should at least be able to pay his campaign bills - but no. He was reduced to trying to sneak out of a hotel yesterday to avoid having to pay his bills, but the kitchen staff detained him - such an ignoble fate for a presidential candidate. This is the man, note, who promised that he would pay off the national debt with his personal funds.   He has supporters, but do not be surprised because their sanity, too, is questionable.  > Poe, as expected, had a made-for-TV program as his campaign offering. This meant scantily clad girls (although the President\'s rally had some sexy antics on stage as well - although not by the President herself, thank heaven), comedians trying to be satirical by poking fun at the citizenship issue filed against Poe, various showbiz personalities testifying that he is a good person and has a sincere desire to help the country (see my comments on sincerity above), and his elegant wife eloquently pleading for the masses to give her husband a chance to work his mojo on the much-begrimed state of the nation. Poe then gave a 12-minute speech, during which he did much lambasting of the current situation, but did not give much insight into how exactly he planned to change it. There were a few vague catchwords thrown about, such as \'renewal of the tax system to improve revenue collection\', if I remember correctly, but as I said, nothing specific. Rumor has it that to all media questions regarding his political/economic platforms, Poe answered \"No comment\".   I strongly suspect that Poe sees himself as the Enigmatic President - the one who remains mysteriously silent in the background and occasionally emerges to pronounce the Final Edict, which will naturally silence all squabblers and dissenters because Poe is Always Right. This is a role he specializes in playing, both in reel and real life - well, in the showbiz realm, at least. How do you explain to a high school dropout that you cannot dichotomize so easily in the political arena?   How do you hold a moonbeam in your hand?  - - -  To Loren Legarda: when General Victor Corpus labelled you a \"political prostitute\", he was referring to your penchant to switch political parties and affiliations at the drop of a hat. Remember, you wept with bitter disappointment at Estrada\'s impeachment trial because you felt that he was being shielded from the full weight of judicial consequences, and that justice was not being properly apportioned. You won the nation with your tears, then. Yet now you are running as the vice presidential candidate of a man who is Estrada\'s best friend, and having cozy little lunches with pro-Estrada groups in an effort to win them over. You are no better than the candidates who do the lambada, sing, take part in gameshows, and have photo ops taken with previously bitter rivals, all in the name of political one-up-manship. You have sold your ideals and your principles for popularity. What is that but prostitution?  In other words - Corpus was referring to your political leanings, not your morality and chastity, you stupid, silly, overwrought woman. Grr argh, if you were going to get indignant about name-calling, you might as well have gotten the spirit of the insult right.  - - -  I fear for my country. And, er, no, I haven\'t gotten around to writing the Mindanao post yet.  Gomen !   '),(25094,2238828,'08,February,2004','2004-02-08 00:00:00','    I am normally execrably bad at reading people and determining motives, reasons, underhanded plots and connivances, et cetera, so you can imagine how disproportionately thrilled I am that I got sixteen out of twenty on  urlLink this particular quiz . If you have time, please take it and tell me what you got, and how you were trying to distinguish genuine from fake. Me, I look at the mouth ... but I\'m a fruitcake, so you wouldn\'t want to imitate me.   On the other hand, I\'m a fruitcake who got a good score, so.  By the way, this is from the BBC, so I have a tendency to trust them. I  like  the BBC.   '),(25095,2238828,'08,February,2004','2004-02-08 00:00:00','    I know I said I would post about Mindanao ... er, a week ago? However, I am bogged down in research. What I e-mailed G is a mishmash of information and current events and speculation and piecing-together and random readings, and I would like this to be slightly more coherent. That, and I haven\'t written anything even remotely resembling political analysis in a long long time. Rants and my occasional spates of bitching do not count.  In the meantime, several things that make me squee:  First:  urlLink Mia  has come back, which means that there will be good things to read on Ephemere again. College is supposed to be an arena for education and learning, but the world is so big and life so full that sometimes the lessons you have to learn are those that do not have a syllabus or an assigned text.   Second: Three words for you - Janet. Justin. Scandal. No matter what you may have thought about it, I think  urlLink this guy  makes a perfectly logical point. If this was a cheap publicity stunt, then they succeeded beyond their wildest expectations. On the other hand, if it really were an accident, then we\'re not really helping the so-called innocent kiddies by producing news items about the incident every 3.5 seconds. Besides, children nowadays are frighteningly advanced for their age. Earlier I was tutoring my nephew (who is in first grade) in math, and found to my astonishment that they are already covering the tessellation of shapes.  Although having said that, I will lay the smackdown on anybody who tries to cyber-flash me with their breasts. I am very pleased with my pair and I am proud that I am a woman, but there is no need to belabor the point by becoming one with Ms. Jackson. Keep your clothes on, people. Thank you.  Third: American Idol, which is my new guilty pleasure. I realize that everyone hates Simon because he is such a heartless bastard, but I also believe that everyone secretly likes him because he says all those things that we wish we could have the guts to say- things like  I\'m sorry, that was bloody awful  and  That\'s incredibly corny . Yesterday I was watching the first round of auditions, and - well, you\'d have to watch those to understand why Simon is such a bitch. (Hint: you would bitch too if you were holding auditions for a SINGING COMPETITION and eighty percent of the people auditioning could not even carry a tune.)   At any rate, he has his moments. There were two, in particular, that made me squee - one of this wacky girl who rode in on a scooter and confessed it was a gimmick to make herself stand out.  You have the potential to be the next American Idol,  Simon told her.  Lose the scooter. You don\'t need a gimmick.  The girl, by the way, had an amazing voice. She truly did  not  need a gimmick. The other one was during the audition of this guy with a smile that could have lit up California -  If we were giving out prizes for Best Smile, you would have won,  Simon commented.  You have the happiest face I\'ve ever seen. You\'re in.   One of these days I\'ll start writing about what I\'ve been up to this past week, but first - the serious stuff.   '),(25096,2238828,'31,March,2004','2004-03-31 00:00:00','    The thing about official memoranda and letters is that they take me so long to write. The dissonance between what I actually want to say and how I have to say it is, to put it mildly, quite great. If you are familiar with  urlLink the memo I wrote about a particularly annoying trade event , then you know what I mean. If you don\'t, here\'s an example:   What I Wanted To Say    Memorandum   For		:	Post  Subject		:	Report on Status of Project X  Date		:	2 April 2004   Why have you not sent us the updates after we pleaded with you ten million times on bended knee?  Send us those updates or we will sic the Secretary on you and he will write you Memos of DOOM.  For information / appropriate action.    What I Actually Said    Memorandum   For		:	Post  Subject		:	Report on Status of Project X  Date		:	2 April 2004   We wish to follow up on our previous memorandum on the status of Project X, dated --March 2004.   Please note that this report is submitted regularly to the Secretary as part of the monthly accomplishment report. It is an important basis for performance appraisal, particularly for non-policy posts.  For information / appropriate action.   It\'s something about working for the government, I tell you.   '),(25097,2238828,'29,March,2004','2004-03-29 00:00:00','    Maybe I\'m being overdramatic and quixotic, but here\'s my take on it:    urlLink I will not vote for a turncoat.  I will not be counted among those who condone political prostitution.  I believe in keeping my word, and I will not vote for those who do not keep theirs.   And if you were interested, both Arroyo  and  Lacson swore, prior to the 2004 elections, that they would not run for president. Lacson, who was chief of the Philippine National Police (PNP) at the time he gave this interview, said vehemently:  I will not run for any political post after I step down as chief of the PNP. Politics is too dirty.   If anyone feels the same way, let me know.   '),(25098,2238828,'24,March,2004','2004-03-24 00:00:00','    Things that are currently driving me nuts:  1.  E-mail addresses : Embracing the age of electronic mail is one thing, proposing to it with a big brass band in the background (apparently I\'m  urlLink alliterative  when I\'m pissed, hmm?) is entirely another. I did an inventory the other day, which is when I realized that I have SEVEN. That topples me right over Geek Cliff and straight into Paranoid Freak Canyon. WHY do I have seven addresses when I originally started out with only two?  Okay, let\'s see ... we have the e-mail address I use to correspond with close friends, which is also regrettably the e-mail address I have given to a couple of exes. There\'s the e-mail address I use to correspond to people I meet online. There\'s the one I use to apply to jobs with, and e-mail the people I meet at seminars and work-related stuff ... there\'s another one for joining mailing lists, another one for corresponding to my school in KL, another one I signed up for just because I needed an ID for Yahoo Messenger, and another one that exists, apparently, just because.   The really scary thing about all of this is that yesterday I was trying to figure out which one to delete, and could not decide. This is because they all perform a very specific function for me, and maybe also because I have schizophrenic tendencies.  2.  Inflation and Low Wages : Which is to say, everything is too expensive for me. Granted I could probably save a lot of money if I stopped indulging myself at Starbucks and Tea Republic and Delifrance\'s Brie on french bread, but then if I did that, life wouldn\'t be worth living. I did not get off my lazy butt to work just so that I could buy myself greasy, fatty food that is cheap but will ultimately clog my arteries and give me gas.   Also, if anybody knows a way to come up with money really fast, let me know, because my parents are having their silver wedding anniversary in a couple of weeks and we\'re trying to get them a room at the EDSA Shangri-La. We could probably get a room cheaper at the Holiday Inn or something, but PEOPLE. This is not one of those dumb anniversaries like paper or wood or pebble that you don\'t even remember. This is their SILVER anniversary, and we\'re trying to pull off something really big, with musicians and all. C and I are meeting up with someone from Shangri-La\'s reservation department on Thursday, so I\'d better figure out how to pull a rabbit out of the hat before then.  3.  Starbucks\' Mango Tea FrappÃ©s : Okay, before I comment on the frappÃ©s themselves, let me just point out - they are FRAPPÃ‰S. They are NOT frappuccinos. Frappucinos are frappÃ©d cappuccinos, which means they have coffee and milk in there somewhere. The tea frappÃ©s do  not  have coffee, and thankfully they also do not have milk. They are merely fruit frappÃ©s with an infusion of tea. In short, Starbucks people, stop insisting that they are frappuccinos and giving me that irritating superior look every time you do because they ARE NOT, and if you do it ONE MORE TIME I really will tell you that my name is Wixfiniqhal* (with a silent \"L\") and not spell it for you. So there.  Oh yes, and your mango thingies suck because they are not made of real mango. Mango concentrate, in the Philippines during mango season? Bitches,  please .  4.  Come-on Lines / Moves : Not that I am the hottest hottie to ever strut down Hot Street, but yes, I do get my fair share of come-ons, since there are people in the Philippines that will hit on anything - even a dog - in a skirt. If you are indeed planning to hit on me, or on any of my friends - or, in fact, on anyone with an IQ above 50, then here are some tips:      Inappropriate touching is a turn-off.  I may be in the extreme minority, but I do not appreciate being touched by strangers unless they have my permission to do so. Guys, this is dangerous for you because people nowadays take accusations of rape and/or harrassment a lot more seriously. Therefore, in order to be sure: ASK before you touch. ANYWHERE. (Also: don\'t even think about asking to touch her in private places unless you are already lovers. You\'d think this would be obvious, but sadly it is not.) The other night I was pissed because the guy kept on patting my head, which is  verboten  in Thailand - where I grew up. It is an increasingly globalized world, my friends, and the girl you are chatting up may be used to a different culture from yours. In short: do not assume. ASK.    Do not stare at the girl.  It does not make her feel sexy or beautiful. As a matter of fact, it makes her feel like she has mold growing on her face or a third eyeball dangling out of her ear.    Do not overdo your polite / gallant gestures.  Pulling a chair out for her is fine. Opening the door for her is appreciated. Doing stupid-ass things like yanking a cigarette and lighter out of her hands  while she is lighting the cigarette , only to put the cigarette back into her mouth and then lighting it yourself, is just reclassifying yourself as two baby steps above \'brain donor\'.    And finally,  GET RID OF YOUR POSSE , especially if half of them are representatives of the ugly side of your family. Having ten million people with you at all times does not make you look cool, it makes you look insecure. Perhaps this is not true of all Filipino females, who are notoriously mothering and enabling, but THIS particular Filipino female, along with a rapidly growing number of enlightened others, does not want a guy she will have to mother. If you absolutely  need  to have at least ten people with you at all times to affirm that you are a valid member of the human race, then YES, you fall into that category.   5.  Because of You : For the fortunate few who have never heard this song, it is an incredibly cheesy ballad sung by Keith Martin. To be perfectly fair, I must admit that it is not half as cheesy as some of the songs that have made it on the radio (*cough*Honey, My Love So Sweet*cough*), but still. The cheese factor increases exponentially by a power of ten every time the song is remade / reinterpreted by YET ANOTHER LOCAL ARTIST. Which leads me to my last gripe -   6.  The local music industry : If local performers want to know why they are not making it big internationally, despite the fact that Filipinos are known all over the world for their performing skills, here\'s the reason: they\'re not original. You could take that comment a million ways, and it would be accurate no matter how you interpreted it. However, to be more specific, when I say they\'re not original, this is what I mean:      They don\'t have original styles.  For example, Nina has a good voice, but her current style is too reminiscent of Mariah Carey, pre-breakdown. Sarah Geronimo is an amazing performer considering that she\'s only 16, but honey, just because you sing Celine Dion songs doesn\'t mean you have to sing EXACTLY like Celine.    They don\'t have original songs.  This is an ailment of the local acoustic genre in particular. FYI, any dumbbell with a good voice and a way with a guitar can sing \"Because of You\" and get applause at our local bars. That\'s why there are ten million of you, acoustic singers - Paolo Santos and Nyoy Volante and Nina and Kyla and lord knows who else. That\'s why there will be ten million more in the next five years, and that is why you will probably end up doing endless rounds of has-been bars like Hard Rock or Tia Maria\'s, singing to drunk expats and their ho\'s.   They all look the same.  Granted, we are in Asia, where everybody has dark hair and dark eyes, making it difficult to stand out on the basis of looks - aside from which, you might snark at me, you want to stand out on the basis of \"talent\". Fine. However, take note of the following: on the basis of voice quality alone, you don\'t really deserve to be stars - glory notes and unbelievably long sustained notes notwithstanding (I\'m talking to YOU, Nyoy Volante - and change your name while you\'re at it). Heck, I saw a ten-year-old girl sing five times better than you at one of those thirty-peso karaoke machines set in the middle of a crowded arcade in a cheap mall. What I am trying to say is this: be different. Stand out. You all have the same clothing style, shoes, makeup styles, hairstyles - I bet if I looked really closely, I\'d find out that you have your highlights in exactly the same places. Ditch your stylists and find your own style, and if that means wearing organic stuff and camping out at Mt. Banahaw like Cynthia Alexander, then go right ahead.   Maybe I should rename this journal the Bitch Blog, because that\'s all I\'ve been doing in it lately. Or perhaps I should post a Things to Do list, just to balance things out.  * Wixfiniqhal, pronounced Wish-PHEE-ni-gah. A name invented by my sister C, in a fit of frustration with Starbucks employees who keep on asking for her name at the counter but cannot spell it, despite the fact that she has shortened it to her two-letter nickname, CJ.  How to use it  : When they ask you what your name is, say \"WishPHEEnigah\" really fast, then add \"Make sure it\'s spelled correctly, okay? That\'s Wixfiniqhal with an X and a Q, and mind the silent \'L\' at the end.\"   '),(25099,2238828,'17,March,2004','2004-03-17 00:00:00','    Apparently I\'m a  urlLink flexitarian . Good to know. Granted, I became one out of self-defense - particularly here in Asia where people eat the strangest things. The first few words I learn in any language are simple but very necessary - the local vernacular for \"I live in (insert address)\", \"I cannot speak/understand (the name of the language)\", \"I do not eat pork\", and \"Is that chicken/beef/fish/vegetable?\" I also learn the word for the part of the animal around the rib/breast area, as that is generally the only part of an animal that I *can* eat. So far, it\'s working and I have not had to eat anything strange - well, unless you count my own aunt tricking me into eating goat. It was particularly traumatic because I had a pet goat when I was very little - and what a pet! I carried it around with me all day and took it to bed with me when I was napping.   (By the way, that goat grew up believing it was a cat. It died after I left - of a broken heart, apparently, because no one else was willing to indulge it in its case of mistaken identity.)  I was not always so particular about my food. I can actually remember eating  balut  and fish eyeballs when I was a very little girl - and then again, I did a lot of things as a little girl that I am not capable of doing now. Things like befriending goats, doing a split, eating half my weight in food without gaining a pound, dropping out of class ... oh wait, strike out that last one, I still do that.  It occurred to me today that for someone with no home, I have entirely too much history. For example - I am currently working in an area where it is likely that I could run into someone I would rather not run into. Granted I haven\'t actually seen him yet, but I keep on bumping into reminders that he exists, and I keep on expecting to run into him one day, when I least expect it and therefore am unprepared. In another example, K, a friend of mine who works in the building right across from mine, told me that last week she went out on a group date with this guy I used to go out with. Apparently, when confronted with the fact that they knew each other (he met up with us once while we were on a girls\' night out), the guy claimed he\'d been blind drunk the last time we\'d met up.  Apparently,  I commented,  \'Blind drunk\' is what they\'re calling \'dumb as a stump\' nowadays.  This was, of course, because the last time I\'d gone out with that guy, he\'d said something really idiotic - so idiotic that I could not help myself and called him on it.  And then! To top it all off, K and I met up for a short break today at the Tea Republic (they have the most amazing mango teas and Toblerone cake! whee!). We were in the middle of our usual bitch session (\"Wah, work sucks.\" \"Bitch, please. I have it ten times worse than you.\") when she looked up and hissed  Dammit, one of my workmates is here! Now I have to be nice again.  She was trying to find a place to hide when her workmate looked directly our way. Naturally, she then had to acknowledge him and invite him over. He came out a few minutes after placing his order at the counter, followed by another workmate ..   ... who turned out to be another one of those people I\'d rather avoid, due to a rather unpleasant episode of my life in which he played a starring role. Nevertheless, they came over and sat down, K and I gave each other desperately fake grins, and we spent the rest of the lunch hour reminiscing about our insane former high school. It is true that blessings sometimes come cleverly disguised - how was I supposed to know that my two-year stint in a school where I interacted regularly with druggies, rich brats, war freaks, space cadets, and people with various learning disorders was going to be my saving grace in millions of awkward situations to come? (If you were wondering, I was one of the people with learning disorders - I could not learn *anything* about Filipino history to save my life. Yaz was a space cadet, and Laya and K were war freaks, hee hee.)   At any rate, we got out of it without causing scandal and mayhem in the midst of Makati - something I am so proud of that I rewarded myself with a chocolate-chip oatmeal cookie. Mmm, good.  [ EDIT : Okay, I officially have WAY too much history going on. On the train home from work, less than an hour after I posted this, I glanced casually to my right - and froze. Standing not five feet away from me was this guy I\'d had a crush on in high school. We didn\'t have history, not in the conventional meaning of the word: that is, I had a huge crush on him and I thought he liked me back. To be fair, he did. Well - until I finished his computer paper for him, at which point he dropped me like a hot potato. Story of my life. And no, he didn\'t recognize me at all.]  Right now I feel as if I\'m in the eye of a storm - a  fierce calm , as Tori Amos puts it. Everything seems to be frenetically in motion: when I go to bed, I am usually so tired that I slide into sleep - past dozing, past dream-fragmented somnolence, straight down into a quiet soft darkness behind my eyes. Awake, I can barely keep still for more than half a minute. There\'s so much to do, and so little time to get it done in - aside from my office work, there\'s work from the NGO I volunteer in, stuff to look over for my dad, Luis\'s plays to read and critique, a guest column for my old high school\'s paper, and my own list of articles and fiction pieces that I want to write.   I have a vague suspicion that I may be very very slightly manic-depressive - at least, every time I feel like this it is usually the prelude to a plunge downward. Outside I feel as if I\'m moving so fast, I\'m vibrating; inside I feel as if everything is pristine, untouched, perfectly still. It\'s a fierce calm, all right - one that I am holding on to with two desperate hands. Once it slips away I will fall back down into endless days of lethargy and sadness, and  I don\'t have time .   I\'m so sick of being like this, of flaring up so brightly only to burn out. My God, I\'m so tired of waiting to fall.  Even more than being sick and tired - sometimes I\'m scared that when I fall this time, I won\'t have the strength to get up again, no matter how long I wait, or how much I want it.   '),(25100,2238828,'12,March,2004','2004-03-12 00:00:00','    I have five different articles I want to write and post, but I am dead tired and, after all, I am just killing time until I leave work. Generally I come in at 7:30AM and leave at 6:30PM. Discounting the one-hour lunch break mandated by law, that makes it an even 10-hour workday. Ten hours a day, five days a week - I\'m pulling a fifty-hour workweek, and that doesn\'t count the time I spend commuting (because my house is so far from my workplace that it might as well be in another province - or another state, depending on your viewpoint).   Before you ask, no, I don\'t get paid overtime. This is the Philippines, remember. Good labor practices are always a matter of theory rather than application.  So no real update from me today, except for something I nicked off  urlLink Noelle :    Three Things You Probably Didn\'t Know About Me :  1. I didn\'t know FM radio existed until I was fourteen. 2. I am in love with Johnny Depp. 3. I was nearly recruited into a modeling agency in KL.   Three Things That Don\'t Really Matter   1. I will never have perfect hair. 2. I will never find out what my exes (ex-boyfriends, ex-lovers, ex-crushes, etc.) really thought about me. 3. I don\'t like Starbucks\' Mango Tea FrappÃ©.   Three Things I Tell Myself To Make Me Happy   1. I\'m smart.  2. There will always be someone uglier and fatter than me. 3. I bet this would make a hilarious story.   '),(25101,2238828,'05,March,2004','2004-03-05 00:00:00','    Tired. Tired, tired, tired. I owe so many people email and I have not updated in nearly a month, but first I will let myself sleep for three days.   The funniest thing about this job so far is the dress code, which I flagrantly violate daily.   '),(25102,2238828,'30,April,2004','2004-04-30 00:00:00','    I will not be pragmatic and useful and political today. I am in my office, bound by my brown-nosing timecard, which will faithfully record the exact time I slip past the door and into the elevator and out into a world that has a sky. At precisely five o\'clock, however, I will be free again, and I will no longer be Reichel, Technical Assistant for a government agency - I will be rei, dipping my toes in the wind to see where it will take me.  I\'m cheating, of course. I know where it is taking me tonight.  Tonight it is taking me to Baguio, city on a hill, where there are strawberries and peanut brittle and shopping bazaars and flowers so vivid that you cannot believe they have been cut, and so cheap that you can have them by the armful. I will go and have dinner with my parents and their friends first - a sort of apologetic good-bye, because tonight is their last night in Manila before they take off for their honeymoon in Geneva - and then K and I will meet up and go to the terminal. It was a spur-of-the-moment thing - I did not know until yesterday that we were actually going - so I wasn\'t able to take leave on Monday. I will be heading back again, alone, on Sunday -  but first, this madcap weekend. When was the last time I cried  Let\'s go!  and just went, laughing? I can hardly remember, through the haze of events that has been the past nine months.  Reichel, Technical Assistant, MBA graduate, BA in Political Science, editor-in-chief of several publications, responsible student and dutiful daughter, will not be here this weekend. Try her number again on Monday.  rei, dancer-in-rain, pointer-out of ridiculous things, funky-hat-wearer, lover-of-red, is here.   '),(25103,2238828,'23,April,2004','2004-04-23 00:00:00','    Will post about  Peter Pan  another time. For now -  My friend Yaz and I were exchanging comments in her blog once about willful ignorance - she had a workmate who refused to read anything about wars or genocide or hunger or terrorism, on the grounds that she could not see how her knowing these things could help. Which makes sense, in a brutally honest way. You and I and our ten million friends, sitting in safe places with food and clothing and education - what are we doing that has an impact on the people we read about? It\'s like unnecessary torture, sometimes - back in 2001, after the 9/11 attacks, there was a glut of information - horrible, heartwrenching information about people calling out of the wreckage to tell their family \'I love you\'; about people milling in the streets, terrified that the sky was falling; about people asking  why why why, what have we done to deserve this  and no one being able to answer. Laya told us:  I don\'t want to hear anything or read anything anymore. I want to curl up in bed with my hands over my ears.  We know so much, and yet we are unable to change what is happening.  Sometimes I come across accounts like  urlLink Riverbend  and  urlLink Raed in the Middle , and words fail me - I am caught in a \"nightmare of words\", as Riverbend says. My heart cracks and folds, but my hands remain still, or engaged in mundane, useless things. How do I justify my anger and sympathy when I am still here, alive and indifferent?  Right now I feel as if I should say the same thing I told myself about my relationship with E - this wasn\'t worth losing my innocence for.   '),(25104,2238828,'22,April,2004','2004-04-22 00:00:00','    Posted late, but celebrated right on time: happy 21st birthday to my best friend, my worst enemy, and the only person in the world who has ever seen me dance the Running Man. Hey,  urlLink freakchild .   I told you once that if we hadn\'t been born sisters we would probably never have become friends, because we are so disparate in nature. Knowing what I do now, and with every single year more we have together, I tell you this - and I have never been more sincere:  Out of all the people I have ever known, know now, or will ever come to know, I would still choose you to be my sister. Thank you for being you.   '),(25105,2238828,'22,April,2004','2004-04-22 00:00:00','    Watched  Peter Pan  last night with K and adored it way too much; I had to keep reminding myself of my age as we left the theatre, because I kept on squeeing over the Peter Pan/Wendy scenes. You think you\'re all grown up, but you\'ll never really lose that inner teeny fangirl.   I definitely want to write more about it later, but for now, a snippet of an interview with Jeremy Sumpter (who played Peter Pan) and Rachel Hurd-Wood (who played Wendy). They strike sparks off each other, apparently off-cam as well as on:   Jeremy : Me and Rachel want to do Romeo and Juliet together.  Rachel :I donâ€™t want to do it, do I.  Jeremy : Yeah, you do. (to members of the press) Wouldnâ€™t it be good if me and Rachel did Romeo and Juliet together?  Rachel : Itâ€™s already been done, hasnâ€™t it?  Jeremy : Well so has Peter Pan, hah!   Rachel : Yeah but ... oh, shut up.  Jeremy : And so has Lord of the Rings, hah!  Rachel : No it hasnâ€™t.  Jeremy : Itâ€™s already been on the theatres twice.  Rachel : Itâ€™s a trilogy, you fool.  Jeremy : Oh yeah.  Definitely going to buy the DVD, and maybe watch it again. I suppose this is the part where everybody starts denying they know me.   '),(25106,2238828,'14,April,2004','2004-04-14 00:00:00','    The thing about art is that it\'s difficult to take home. I learned how to love it in college, by a teacher who talked about art the way other people talk about fashion or books or cellphone models; she was so convinced that it was necessary that to suggest otherwise would have drawn a blank look from her. And yet despite her strong convictions and a rather inspiring book on art by Jeanette Winterson, I find myself still unable to take that final step that will lead to my owning a piece of art - a painting, a sculpture, a mixed-media interpretation of something. My rather footloose lifestyle is part of the reason - I cannot imagine carting works of art from country to country, house to house. There is something about good art that seems to disdain such treatment, the way you cannot pair, say, caviar, with cheap soda crackers and grape Kool-Aid.   Price, of course, is the second part of the reason, although perhaps a significantly smaller part. As someone once pointed out to me, if you really want something, price is no object. Art is not yet a necessity for me - given an amount of money to lavish on myself, I would spend it on a spa, books, clothes, shoes, accessories, and CDs. Only then would I begin to think  Hey, yeah, I should go look for a painting I like .   For the noncommittal art enthusiasts such as I, galleries are heaven. The Vargas Museum in UP is amazing, considering that it only charges PhP10  (that\'s less than 5 cents) per entry. My favorite painting is a seascape by Felix Resurreccion-Hidalgo, which - unlike most other seascapes - features nothing but sea and sky, blending together in turbulent shades of blue. After I saw that painting I went out into UP\'s beautiful tree-lined main street, tipped back my head, and stared at the sky for the longest time. You have no idea sometimes how beautiful the world is until you\'ve seen it through an artist\'s eyes.  Lately, though, I have been unable to visit any galleries - too tired, too busy, too harassed. If you like art - or even if you don\'t, and you just want to rest your eyes with something lovely - then online art galleries are a dream come true. Granted, they are to art what fast food is to  haute cuisine , but just like fast food, they\'re filling nevertheless.   Found a new one today -  urlLink Pinoy Art , featuring some contemporary artists. I am particularly enamored of Rafael Cusi\'s   urlLink Palayan   (Ricefield) and   urlLink Tabing Dagat   (Seaside) series, as well as Romeo Tabuena\'s   urlLink Girl with Butterflies, Mexico  . I have a feeling that if I could see the real paintings, I would be overcome with a strong desire to walk out again into the sunshine and stare at the sizzling blue summer sky.   '),(25107,2238828,'12,April,2004','2004-04-12 00:00:00','    Who will you be stuck with at end of time? (sorry, no more pics) by  urlLink  chi_a_baidh        Your name is           Your sex is      Male Female Undecided       Your favorite color is      Red Orange Yellow Green Blue Purple Black White Other       You are stuck there because    you were frozen in ice/a crystal      For _____ years    14      With    Johnny Depp      He/She will think you are    sexy      You will    live in peace             Created with  urlLink   quill18  \'s  urlLink  MemeGen 3.0  !       Well, wouldn\'t you post a meme like this too if you got JOHNNY DEPP as your partner in life?   Even if you only have fourteen short years together ... sigh.   '),(25108,2238828,'02,April,2004','2004-04-02 00:00:00','    [A copy of my letter to  urlLink G , who never fails to ask the most interesting questions.]  G:  Midway into the heat of election fever here in the Philippines and I was all atwitter yesterday because I managed to have a decent conversation with a cute guy. Awhile ago he asked \"rei, how tall *are* you exactly?\" Personal interest! Of such small insignificant things is the joy of my day composed. I bask in my shallowness.  But out of the shallowness and back into the depths, for you want to know about Poe. By this I assume you want to know:  (a) Who the hell is he, in the context of Philippine society; (b) What the hell he thinks he\'s doing, running for president; and (c) How the hell he plans to run the country.  Since I love you dearly and I owe you for having ditched you for chocolate (but it was GOOD chocolate, G - and it was a freebie!), I will wrench my meandering thoughts away from Corporate Cutie and attempt to summarize the wonder and confusion that is the Poe candidacy.    (a) Who is Fernando Poe, Jr.?   First of all, the  real  Fernando Poe Jr. is dead. Poe\'s real name is Ronald Allan Kelly Poe, and Fernando Poe Jr. was his older brother. He is the illegitimate son of a naturalized Spanish father and an American mother; this caused a great furor in the Philippines over the authenticity of his citizenship. As in America, our presidents must be natural-born citizens. The Pro-Poe camp (hee! Try saying that ten times really fast) argued that it was his father who was a naturalized citizen; as the son of a citizen, Poe is therefore a natural-born citizen. The Anti-Poe camp argued that as he was illegitimate, he would, according to Philippine law, take his mother\'s citizenship and was therefore American. The Supreme Court ruled in favor of Poe.  In any case, the entire brouhaha was a moot point, since Poe holds a Philippine passport, pays Philippine taxes, and has purchased property in the Philippines. Personally I agree with the Supreme Court, because to tell the truth Poe is probably more Filipino than I, half-baked product of Filipino parents and a mishmash of other cultures, am.   Poe is also an actor of almost iconic status in the local film industry. He is known fondly as \"Da King\", and heretofore has been treated with the respect and awe normally accorded those who are protected by  lese majeste . Our local media, who are normally like moles when it comes to digging out dirt on celebrities, have left him alone for the most part up till now, publishing nothing but positive remarks on Da King. This is out of respect for Poe, say Poe\'s camp; his detractors, on the other hand, insist it is out of fear, citing a rumored incident years ago when Da King allegedly beat up and pissed on a journalist who had published a negative article on him.   He dropped out of second year highschool when his father died, ostensibly to help take care of his mother and his siblings, and has never gone back. He has never held public office before. In the showbiz industry, he is one of the top earners and yet pays one of the smallest income taxes. He has been married to another actor of iconic status, Susan Roces (affectionately referred to, of course, as \"Da Queen\") for years. They have no children, although Poe has one recently-acknowledged illegitimate son by another actress with whom he had a liaison years ago.     (b) Why is he running?   There were rumors as early as 2001 that he would be running for president (I literally wept when I heard this rumor), which remained unverified until late 2003. His name was bruited about by the top brass of the opposition, but no one really took them seriously because Poe kept on saying \"I don\'t know\" to all questions asking him to verify the rumor. ( Snark:  That\'s what he says to everything, so he might as well just record it and play it back at all interviews.) At an incredibly weird press conference last year, the opposition held a big bash to announce that Poe would be their standard-bearer for the May elections.   At the conference, Poe did not show up AT ALL - instead, he called the cellphone of one of the opposition party members. They put him on speakerphone, and with all microphones pressed eagerly to the mouthpiece of the cellphone, he mumbled something about being grateful for the support of the Filipino people and that he would tell them when he was ready to run. In short, it was the most pointless press conference in the history of Philippine politics, as their standard-bearer could not even verify his own appointment.  Sometime around November 2003, Poe finally agreed to run for President, and hurriedly recruited Loren Legarda - a television newscaster turned senator - as his running mate.   I am actually giving you a factual explanation of how he came to be a presidential candidate. G, I could be psychic ten times over and I am afraid I could still not tell you why Poe thinks he is qualified to run.  How do you explain the political stylings of the deluded?  Like I said before, how do you hold a moonbeam in your hand?   (c) What is his platform of government?   I\'m not sure, really. One of his first proclamations was that he was anti-globalization - this, while speaking to a group of farmers. He did not elaborate. After that he remained noticeably silent on all questions regarding his platforms. Once, caught off-guard, he gave the following interview:   Reporter:  What can you say about reports that the value of the peso has dropped due to the fact that you have not yet released your plans for economic development?  Poe:  Really? The peso drops just because of that? That\'s new ... I didn\'t know that.   Reporter:  How do you propose to improve our economic situation?  Poe:  Everybody should just work really hard.  I am translating from the original Tagalog - trust me on this, he sounds more intelligent once you translate him into English.   Recently he has hired an entire armada of economic and political advisers. His political advisers are all from the opposition party, so they cannot disown him, but his economic advisers are noticeably squeamish about being identified with him. Some have openly said \"We are not officially part of Poe\'s camp\", while others skirt the issue by informing the public that their advice is offered gratis and without prejudice to any candidate who might ask them.   A few weeks ago, he made another announcement about his economic platform. This time, he said he was pro-globalization.    I am trying to be objective, but I am afraid that my subject line completely sabotages my efforts. Never mind - it is the truth. I am sure Poe is a very good man and has a heart larger than Jupiter and warmer than the sun, but in politics, as in love, it is not a good idea to let your heart run wild without your head there in the mix to rein in its more idiotic impulses.   In the meantime, I have decided. I\'m voting for  urlLink Brother Eddie . According to his campaign motto,  Babangon ang Pilipinas. Maniwala ka.  [The Philippines will rise again. Believe.]  I believe.  - - -   Nota bene : In the event that you might want to read some other, more intelligent discussions on the elections, Filipino electoral behavior, etcetera - here are some links:  urlLink tin , who is linked here, points out that there is no magic formula for determining the effectivity of a president, while  urlLink dreamlessness  makes some interesting points on who the Filipino electorate listens to and why.   Hey, if you\'re a Filipino voter and you happen by my blog, get involved, yeah? It\'s easy to mock people for their choice of candidate; it\'s not so easy to dive in, sort through the deluge of information and  mis information, and try to make a logical choice based on what you\'ve deduced. After this election, I will definitely have earned my right to bitch about my government.   '),(25109,2238828,'28,May,2004','2004-05-28 00:00:00','    Okay, so I\'m surfing around CNN being all horrified and concerned about the latest developments in and around Iraq, and what do I come across but  urlLink this . Now I am hungry and I want pizza and I don\'t want just any old kind of pizza, I want  that  pizza. Specifically the Margherita one. I don\'t even want Pizza Hut\'s weird overstuffed sausage-and-cheese-in-the-crust thing - hello, if I wanted that, then I would buy myself a sausage-and-cheese croissant.   Sometimes it is very hard to be a concerned citizen of the world. I am not making light of serious events, I am just saying.   '),(25110,2238828,'26,May,2004','2004-05-26 00:00:00','    I quote myself:  urlLink \"The official campaign period is in full swing - and when I say that, I am picturing monkeys and all manner of wild animals dangling from the chandeliers, bouncing off the walls, cartwheeling across the linoleum. In short, I am visualizing pure chaos, a literal mad carnival.\"    I love it when I am right. In the aftermath of at least one part of our national elections - we have twelve newly-proclaimed senators, although the posts of President and Vice-President are still empty as Congress begins its official canvas of votes today - there is the usual slew of articles doing pseudo-political analysis. Vitriol has been kept to a minimum, despite the fact that our elections were marked with violence - there were over a hundred election-related deaths this year. What can I say, eventually you get tired of screaming for change. As my sister says whenever she slips up,  Sorry, tao lang.  [Sorry - I\'m only human.]  I am particularly amused by these two articles, analyzing the political campaigns of top senatorial vote-getter  urlLink Manuel \"Mar\" Roxas II  (I know it\'s a horrible nickname, but no one else in the Philippines seems to have caught on to it, so ... ) and the daughter of one of the country\'s political elite,  urlLink Pia Cayetano .   Senator Roxas - I salute you, sir; you are worthy of your degree from Wharton. I *was* quite irritated by his campaign ads, to be honest - I did not see any logical reason why the former Secretary of Trade and Industry, grandson of a former president  and  son of a former senator should be hopping and bopping around to a tune that is the Philippine equivalent of, say, a song by Sum 41.  Not to be a mindless loyalist,  I told my mother,  But I know he did a good job when he was with the DTI. I would have voted for him based on that alone. What is with this shaking of his senatorial booty on national television?  However, it appears that there is indeed  urlLink a method to his madness . A pretty amazing branding exercise any way you look at it, really -  and  with rather gutsy people behind it. Imagine telling a former Secretary of the state  Sir, people have no clue who you are. You\'ve gotta learn how to dance.  (If I start mis-typing things, it is because I am laughing so hard. I swear, life in the Philippines = Never Boring.)   For the record, Senator Roxas dances like a white boy. With all due respect, sir, you  do  need to learn how to dance.   Also noteworthy - although perhaps also illegal - is  urlLink Pia Cayetano\'s campaign strategy . While she is pretty and carries the name of a respected political figure, both of which count for much in the Philippine political arena, her problem was that people had no clue who she was - she was even more unknown than Roxas. While Roxas\'s unfamiliarity to the voting populace was due to his natural inhibitions - he is a very private person - Pia Cayetano was a name no one had ever heard of before. A few years ago Roxas was a congressman who rose to the public eye as one of the \"Spice Boys\" - a group of young scions of political dynasties who made waves in Congress due to their aggressiveness and their looks (we are a very fangirly society - we fangirl our politicians, our religious leaders, our mutineers ...); Pia Cayetano had never before held public office. As the article says, she appealed to the upper curve of society: she was pretty, she was well-bred, she was educated ( magna cum laude , UP Law School - the local equivalent of Yale or Harvard). However, it is the lower curve of society that composes the greater portion of the voting populace, which meant that she had to find some way to connect with them. Emotionally manipulative, I grant you, but effective. As the article\'s source points out, people remember her as  Maamo ang mukha, maganda, may hubog naman ang katawan  (kind-faced, beautiful, and curvy) - this is an actual quote. As I said, it worked: she is one of the  urlLink top six vote-getters in the senatorial race .  I could be negative about this and bemoan the fact that relatively worthy candidates have to stoop to cheap tactics to get themselves voted into office, but it\'s too early to be pessimistic. As a matter of fact - I\'ll be laughing about this all day.   [ EDIT : More on what Philippine political life looks like,  urlLink here  - scroll down to the section titled \"Silly or Smart?\" Unfortunately I was not diligent enough to cover the entire campaign / election period, but fortunately there are people in the media with the same sense of humor as mine.  As for a more serious look at Philippine political issues, you can check out  urlLink this site , which provides - well, at first glance, anyway - unbiased information on all candidates. There are a few guides to voting wisely, which I appreciated, and links to sites which help voters learn about their candidates\' track records in public governance, if any. Pity it wasn\'t publicized enough, but who was going to pay attention to the serious stuff when you could watch senators dancing the local equivalent of the Macarena onstage? Be honest - you wouldn\'t have, either.]   '),(25111,2238828,'24,May,2004','2004-05-24 00:00:00','    I\'ve always wanted to write about the more interesting people and events that I\'ve come across, but never really knew how to start. There are so many of them, honestly. Within the past five years alone, I\'ve got stories enough to write a travel journal - I just don\'t know which one to write first. Yaz got me kickstarted on this one a few days ago, while we were chatting - she mentioned casually that she wanted to go to Nepal, and as it so happens, I\'ve been there. I started enthusing about some of the more interesting parts of the trip - the \"walking leaves\", as we called the people who carried bundles of leaves on their backs (the bundles were so big that all you could see was a pair of legs beneath them, hence our nickname); the three-day trek through a Nepali forest on the back of an elephant; waking up in the wee sma\'s to watch the sun rise over Everest; walking through Kathmandu and having to dodge the cows, which lumber about the main streets in the smug knowledge that they are considered \"holy\" and therefore untouchable.  Write about it in your journal,  Yaz urged.  However - I being the wordy sonofagun that I am - I quickly figured that the entire trip (it only lasted about a week, I think) would not fit into a single entry. There was simply too much of it; I had to pick one memory of Nepal to talk about. The one that stands out most is one that appeals to me on a lot of levels - as a story, as a children\'s rights issue, as a cultural quirk, as a women\'s rights issue: the story of the child goddesses of Kathmandu.  First, a brief explanation of why we were there in the first place: my father used to work in the  urlLink Food & Agriculture Organization \'s  urlLink regional office in Bangkok  - a job that would have him flying off to various parts of the world at a moment\'s notice. Pretty glamorous on the surface, but not for a family man - it got to the point where he would be gone six months out of a year, and doing so much overtime at the office when he  was  in the country that we barely saw him. To remedy this, he began taking us along on some of his trips - which meant that we were out of school two months of a year, but my mother argued that the trips were educational, and the school eventually gave in.   The Nepal trip was part of a three-week tour of South Asia: we were supposed to start off in Nepal, then move on to Bangladesh, then Pakistan, then finally India, as my father monitored and assessed  urlLink Grameen Bank  operations in the rural areas of these countries. (As it turned out, we were unable to complete the trip; my paternal grandfather died a day or so before we were supposed to leave Nepal for Bangladesh, and we had to return to the Philippines in a hurry.) My father, typically, had made no arrangements for us while we accompanied him, so our first few days were spent in a jeep, touring the Nepali countryside. A few days later, one of his colleagues\' wives called up our hotel room; her husband had noticed that we were growing increasingly bored of seeing nothing but mountainscapes and poor people (we were stereotypical well-off brats, sorry), and suggested that she take us around Kathmandu.  Off we went, then, on a walking tour of Kathmandu. The office had offered us a car and a driver, but since the streets of Kathmandu were crowded with people and cows and whatnot, our hostess pointed out that it would be much faster - not to mention more interesting - if we took taxis and trishaws instead. She was a lovely guide (I\'ve forgotten her name, but I\'m sure it started with an O): with her at the helm, we managed to get warm milky tea at local restaurants, go shopping at colorful markets, and meet tradespeople and local schoolchildren.   On perhaps our second or third day together, Ms. O took us walking in a part of the city that had plenty of interesting buildings. One of them was particularly fascinating - made of stone, it rose like a palace in the middle of an otherwise ordinary-looking district. It was high and imposing, with intricate carvings on its columns; its walls were high, its wooden gates opening slightly to allow us to peek into a grand, cobblestoned courtyard. It was so intimidating, in fact, that we assumed it would be one of those sacred places that tourists were not allowed to enter.  Imagine our surprise, then, when Ms. O walked up to the gate and rang a bell. After a brief exchange in Nepali, she turned to us and waved us in.  Come,  she said.  We\'re going to meet the child goddess of Kathmandu.    It was one of the most unbelievable and irresistible invitations I have ever heard in my lifetime - even my mother, who then bordered on being a rather fundamentalist Christian, was seduced by the sheer wonder of the opportunity.  Is she really a goddess?  I asked her, whispering, as we crossed the wide courtyard.    Of course not,  she answered.  There\'s only one God.  Her answer, however, was less assertive than it normally would have been - she, like my sister and I, was entranced by the fantasy. It was one of those caught-in-the-moment things - it was Kathmandu, we were walking in palace grounds that seemed to have been transported from another time, Ms. O  had  addressed the person on the other side with deference and respect ... for all we knew, right? For all we knew.  We walked into the palace of stone, our footsteps echoing. Deep inside, we were abruptly wrenched from our fantasy of being transported back in time by the commonplace sight of a line of laundry strung incongruously across one of the balconies, and a woman doing her wash beneath it. Ms. O stopped and made an inquiry, presumably asking permission to visit with the goddess. The woman - who we later learned was one of the goddess\'s relatives - yelled something at someone inside the palace\'s private quarters, and got a hollered response, which she transmitted to Ms. O.  What is she saying?  we asked, a little anxiously; we were Bangkok brats, aware of and respectful of cultural and spiritual taboos. Aside from which, and not to belabor the point - who wants to anger a goddess?   She is saying that the goddess does not wish to be disturbed,  Ms. O said, looking thoughtful.  But maybe if we give an offering ...    Our eyes lit up. In Bangkok, people offered food and sweets to the spirits every day; this was familiar territory. My sister had a doll, I had some candy, and Ms. O dug out some money in hopes of enticing the goddess into making an appearance. She displayed this to the relative, who immediately yelled another message down. Within a few minutes, the goddess emerged - a small, lovely little Nepali girl of no more than nine or ten years old, dressed in red and gold. She moved immediately to a dais on which was placed a red-and-gold pillow - her pedestal, upon which she was accustomed to receiving her supplicants. Ms. O presented our offerings, she graciously - and wordlessly - accepted them, bowed her head politely, and then got up and darted away again.  Outside of the palace, Ms. O told us the story of the child goddess - I forget the details, but apparently one of the deities of the area had chosen a girlchild as her preferred incarnation. Thereafter, the town held selection rituals for would-be goddesses, similar to those held for the Dalai Lama: baby girls from all over would be brought to the temple and put through a rigorous screening process. The infant who passed the test was then proclaimed the new goddess.    What happens to the former girl?  I asked, as it was becoming obvious that the deity\'s representative did not remain a goddess forever. Ms. O looked at me, her eyes serious and a little sad.    She is a goddess only as long as she does not shed blood,  she explained. The girls were guarded zealously against cuts, abrasions, bruises - goddesses could not be blemished, nor could they bleed.  However, once she reaches menarche, she is no longer qualified to be the child goddess, as she is no longer a child.    As child goddess, the infants and their families are provided with everything they could possibly want or need. They are moved into the palace, they are provided with servants, they are given food; the girl is plied with offerings and adoration from the minute she is proclaimed goddess. When she reaches menarche, however, all of this is withdrawn. Her family must move out of the palace and resume normal lives; the former goddess becomes a girl. According to Ms. O, many previous child goddesses have gone insane, or suffered severe emotional problems - how could you not, having undergone such a drastic change in status?   We were halfway back to the hotel when I turned to my mother.  Mommy,  I said,  I\'m glad I\'m not a goddess. I thought I\'d like to be, and get money and toys and candy - but I think I like being me so much better.   [ EDIT : In case this story may seem a little too fantastic to believe, I\'ve done a search on Google and found a few articles on the child goddess, also known as the Kumari Devi. You can go  urlLink here  and  urlLink here  and  urlLink here  for more information on her. She\'s for real, folks. There are even pictures of her temple, which is good because even  I  thought I was dreaming - it was such a surreal sort of memory.]   '),(25112,2238828,'20,May,2004','2004-05-20 00:00:00','    Apparently the main server is down; I can\'t connect to the Internet. Since I cannot work without the Internet - it has become one of the crutches of my working life - I am sitting here, typing up a journal entry. You  know  you\'re addicted to journalling when something happens and the first thing you think is  Gee, this would make such a great journal entry .   A journal: a catalogue of days, an album of memories. The nice thing about journals is that you pick-and-choose what to remember, and how to remember them. Someone pointed out to me once that you don\'t report the facts accurately in your journal, not really - but that\'s to be expected. I forget the precise term for it, but in the social sciences it\'s accepted that first-person accounts are biased, due to the teller\'s presence and/or participation in the event. Therefore reading journals comes with that caveat - that you are not reading an objective account of an event, but a subjective one.   Despite its inbuilt inaccuracy as a memory-keeper, I love my journal - both the online and offline versions. As I suspect most people do, I keep two versions of my journal - the online blog for regular events, and issues that require an audience; my offline, handwritten journal for personal issues, stories that are too personal or painful to tell while there are eyes watching, and random ideas that are too incoherent to post online. A friend of mine speculated once that I must have a secret journal, where I write down the spicier details of my life. Alas, no - my offline journal is actually less exciting than my online one, as I write in it less often (the really spicy details, to tell the truth, all lie in my private e-mail). The tales of my life that are untold remain untold, because I lack the determination and maturity to find the words that will translate them effectively from memory to story. Being molested in my own home by a close friend when I was nineteen, my dreams about abandonment and falling and cold electronic societies, the time my paternal grandmother rejected me as a granddaughter, being harassed in school when I first returned from Bangkok because I could not speak Tagalog properly, various memories of being abused as a child - so many stories I have to tell, and yet I refrain from doing so because I cannot tell them yet without sounding like a victim. Pardon my pride, which is one of my besetting sins, but I cannot - I  will  not - I refuse to sound like a victim in my own journal.   It\'s been almost fifteen years since I first started journalling - as that stereotypical ten-year-old girl with her pastel, daintily-padlocked San Rio diary and multi-colored pens, scribbling earnestly to \"Dear Diary\". I never thought I\'d make it from then to now, still writing, but in a medium that lets me wander freely through others\' thoughts and concerns, and correspond with people whose interests and passions mirror mine. With that - Belated happy birthday,  synesthetique  - conceived as  luminae  and renamed halfway through, you have always been a haven of sorts. Here I am strong, and witty, and articulate; wrapped in the membrane of my prose, I can retell the painful, the irritating, the frustrating, and the glorious moments of my life without resorting to useless things like tears or tantrums. Here I dissect myself - who and what I am right now, vis-a-vis who and what I wish I was, and find a way to make the edges overlap. Here I find a forum for the issues that move me, and people who challenge and support the things I say.   I am hovering on the brink of my mid-twenties, and I no longer start my entries with \"Dear Diary\", but here I am yet, journalling. Four countries, several houses and apartments, and multiple love lives later, with a larger vocabulary and a more complicated heart - but still me, and still with that inborn need to scribble things down. I wonder if I\'ll ever stop needing my journal.  I hope not.  Later: Kathmandu and child goddesses, for Yaz.   '),(25113,2238828,'20,May,2004','2004-05-20 00:00:00','    As long as there are human beings with free will, there will be children running around with names like  urlLink these . Jett is not so bad, but Ahmet Emuukha Rodan? Reignbeau? Elijah Bob Patricius Guggi Q? Obviously these parents were all on crack.  I realize I don\'t have a lot of room to point and laugh, since I am one of those people who have an unusual name (my name is, according to my father, the last name of an obscure classical guitarist who was playing on TV at the time of my birth.  So  obvious you used to be a hippie, Dad. Just sayin\'). Nevertheless, I must give my parents props - this is apparently a shortened form of the term \'proper respect\' - for rejecting various inappropriate name suggestions that were made during my conception. For example, my cousin wanted to name me Eeyore. Can you imagine?  No, you can\'t.   '),(25114,2238828,'18,May,2004','2004-05-18 00:00:00','    I think I need to start taking sleeping pills, although my mother - who is a nurse - is vehemently against regulating normal body functions with drugs. Normally I agree with her, but I haven\'t exactly been sleeping well lately, and I\'m afraid I\'ll start sleepwalking again. On the one hand, I\'m pretty sure that someone in the house will stop me before I get to the street, but just in case - if you see a bleary-eyed rei wandering where she obviously has no right to wander while in sleepwear, please rope her in and deliver her back home.   Thank you.  At any rate, I am a little disoriented today. Last night a cousin\'s ex-boyfriend came to our house literally in the dead of night; the first words out of his mouth were  Look, I just wanted to know, is GP really dead?   Since N, the ex-boyfriend, has a history of being slightly shady, my sister and I stared at him from inside the locked gate of our house, wondering what the hell he was on.  I\'m not kidding,  he said, sounding slightly hoarse.  Here - someone messaged me this afternoon -   and he pushed his cellphone through our gate, indicating a text message on the screen. We quickly read through the message, which said something along the lines of  Let us pray for the soul of our dear departed friend GP. She was hit by a truck on Saturday morning ...  and looked back at N, still befuddled.    We have no idea,  my sister C told him.  We haven\'t seen or heard from her in months. Last we heard, she\'d gone back home to the province.  She started to unlock the gate, but I held her back; the situation was so surreal that almost anything could happen.   Through the bars of the gate, N told us his story: from their province, our uncle had sent GP to another province to find work. He and GP had communicated regularly through calls and text messages; he had asked her if he could visit, and she had repeatedly rebuffed him, saying it wasn\'t time yet. The last time he\'d heard from her was Friday evening; when he received no word from her over the weekend, he began to worry. On Monday he received the text message announcing GP\'s death, which was when he decided to head for our house to confirm the rumor.  We\'ve been messaging and calling our uncle since last night, and trying to contact my cousin\'s phone, but to no avail - no one\'s picking up or answering. Up to now, I have no idea what\'s going on. It may be true, in which case I have just lost a cousin that I never really knew and therefore will not truly mourn. If it is not true, then it is a cruel joke or some sort of misguided call for attention. Don\'t get me wrong, but between the two, I would rather that the rumor were true. If my cousin were indeed still alive, I don\'t want to think about what kind of twisted logic would drive her to construct such a morbid lie, and I don\'t want my family involved in whatever she\'s gotten herself into now.  If ever I wondered whether psychosis really  did  run in my family\'s genes ... well, rest assured. I no longer wonder - I am  damn  sure.   '),(25115,2238828,'17,May,2004','2004-05-17 00:00:00','    As a follow-up to the  urlLink earlier   urlLink discussion , please note that there are  urlLink further developments . I suppose \"It was just a harmless prank\" and \"Saddam used to do so much worse to them, anyway\" weren\'t really going over well with the public, so now it\'s \"We were ordered to do so.\"   Cry me a  river , Lynndie. A big, deep river called de Nile.  [ EDIT : Ceres P. Doyo did a good job of explaining  urlLink why there was massive outrage  against the soldiers, and the US in general. I am not a generic America-hater; far from it. However, her article encapsulates perfectly how I feel about the situation.   On the flip side of those soldiers, however, is  urlLink General Antonio Taguba , who I admire not only because he is a Filipino, but because in this situation at least, he proved to be a man of honor. The sad thing - or the wonderful thing, depending on which side of the Pacific Ocean you are - about Taguba is that he might not have thrived in the Philippines the way he is doing now in the US. I am proud he is of Filipino ethnicity, but also glad for him - that he is in a country that values and rewards honesty and justice. I would like to say the same thing about my country, but the very fact that our choices were so limited during the past elections - well, it speaks for itself.]   '),(25116,2238828,'13,May,2004','2004-05-13 00:00:00','    Some changes have been made to the links -  urlLink john  has been removed because he\'s made his journal friends-only. As I am not really a friend of his, merely an amused reader, that would make the link pretty much useless. Farewell, John - and the times I spent howling with laughter at some of your entries.  I added the link to  urlLink JP\'s office blog , which will stay there unless he tells me to take it off. They post poetry, musings, rants and raves - it\'s a bunch of people communicating online, when they all work in the same office. I find that both amusing and intriguing - along the same line with me e-mailing my workmate instead of talking to her, or messaging my mother on her cellphone because I\'m too lazy to get out of my room and try to find her. The wonders of modern communication, people.  Also here is  urlLink Noelle , who was an orgmate back in college. (To those unfamiliar with the lingo of Philippine college life, an \"orgmate\" usually refers to someone who belonged to the same school organization. There were regional organizations, which grouped together people from the same region or province; issues-oriented organizations, which concentrated on promoting awareness of issues such as gender equality or the environment; religious organizations; interest-based organizations, which centered around developing members\' talents and skills in specific areas; and academic organizations, which concentrated on major areas of study like philosophy, geography, etc. Noelle and I were in the same academic org, whose membership was limited to political science majors only.) She and I had been members of the same org for a few months but had never really crossed paths -  that  happened when I scribbled a scathing comment in the margins of one of her pieces. It wasn\'t very nice of me - I was in one of my more sarcastic phases, and ended up taking the pseudonym  bluestreak  (as in \"cussing a blue streak\") because at that point, I\'d pissed off enough people to know they were mentally poking pins into a voodoo doll with my name on it.   At any rate, she lobbed a cutting reply right back - which delighted my antagonistic little soul. We went back-and-forth on that for quite a while, and finally ended up liking each other. I have very few female friends who I got to know by initially antagonizing them, so Noelle stands out in my memory for that reason.  A lot of my male friends, however, I have gotten to know by either initially antagonizing them, or not liking them very much and snubbing them. So that romance-novel convention about being in deep abiding love with the person you squabble with the most?   Obviously not a problem with me.   '),(25117,2238828,'12,May,2004','2004-05-12 00:00:00','    You know what I hate most about violence? Like the clichÃ© goes,  urlLink it breeds more violence . Abused children become abusers. Kick a dog and it will try to bite you back. It\'s not hard to figure out.   Where do we go from here? On the American soldiers\' side, there seems to be  urlLink little remorse . Like I told Yaz earlier, I think some of them have gotten to the point where they no longer see the Iraqi prisoners as human. It\'s there in the way they talk about them:  I don\'t know whether the one we shot in the chest is dead yet,  an American soldier wrote in her diary.  A lot of times I get people asking me  How can your God be a loving God when innocent people get hurt and die?  and my answer is  Well, didn\'t you read the fine print? He\'s not just a loving God, he\'s a just one.  Justice demands payment - and sometimes, the people who have to pay are not the ones who racked up the bill.   This is a case in point. Nick Berg did  nothing  to the Iraqi people; he was a civilian posted in a war-torn zone as a part of his job. Yet he, not the soldiers in those horrible pictures, was beheaded; BBC says his head was \"sawn off\".   The point is, God doesn\'t cause pain. He set the rules and defined the boundaries - but it is our choices and our actions that set off events, that bounce back on ourselves and other people. When innocent people suffer, it is because someone made the wrong choice.   '),(25118,2238828,'11,May,2004','2004-05-11 00:00:00','    It\'s the small things, sometimes.  Today was a crappy day, pretty much. I\'d volunteered as a poll-watcher for yesterday\'s elections, so my entire day was spent waiting around in the broiling sun and waiting for people to finish voting. The polling precincts closed at 3PM, but then we still had to check the ballots and tally the votes. All in all, checking 126 ballots and tallying their votes took us six hours - after which we still had to help the Board of Election Inspectors (or the BEI, which is composed of public school teachers) verify their numbers and the tally sheets. It should give you an idea of how clunky our electoral machinery is if I tell you that we were the second-fastest precinct in counting and verifying the ballots, and we finished at past 11PM. This, after beginning our day at around 6:30AM, when we had to witness the opening of the ballot boxes to ensure that they were empty.  In short, I\'d had less than five hours\' sleep, after a prolonged weekend that I\'d given myself in order to recuperate from a nasty case of the flu. This morning I woke up and turned on the TV to get updates on the election, and found to my profound disgust that Arroyo and FPJ are  urlLink neck-and-neck in the polls . Worse,  urlLink Jinggoy Estrada  has made it into the top 12 as a senatorial candidate, which means - be afraid, Philippines - that he will be one of our senators. WHY? WHY WHY WHY?  My mother, who had been pollwatching for one of the slower precincts, woke up late and was rushing around to get dressed so she could drop me off at the MRT station. Five minutes before we were supposed to leave, she suddenly realized that she couldn\'t find her driver\'s license, prompting a house-wide search. Almost 30 minutes past the time I was supposed to leave, her license was still missing - so I poked my head in the door and told her it was fine, I\'d go off to work on my own.   I got to work only to find all doors locked; I had to wait about ten minutes for someone else to come in and unlock my door for me. Then - as is usually the case when you take a day off for any reason (which is why I HATE taking sick leave unless I can barely get out of bed) - I came in to find a pile of things-to-do waiting for me, some of which were due last week. While I was working my way through them, I realized that I\'d have to download and print out the office e-mail, since the workmate who usually does it is on leave today. Not a big thing, unless - as is the case with us - your printer operates on manual feed, and can handle only one paper at a time. If that is the case, then downloading and printing 30 e-mails (and their attachments!) will take half an hour, at the least.   It was in this state of mind that I paused from work to check my phone - and found that I\'d had a couple of missed calls and a few messages. Both calls and messages were from my mother - apparently she\'d found her license a few minutes after I\'d left.  I found my license, I can drop you off now,  she said in the first message, and in the second,  Where are you? I\'ll meet you and drop you off .   Such a small thing, but warming nevertheless. I hate taking public transportation, so my mother - to help me avoid frittering away all my money on cab fare - drives me to the MRT station every day. I hadn\'t really expected her to do it for me today, since she was so tired from last night\'s pollwatching - but knowing that she wanted to, and tried to call me so that I wouldn\'t have to take public transport anyway, makes me feel, weirdly enough, loved.   Now I\'m going to go harass the cashier for my paycheck so that I can treat my mother to a CD or flowers or a good meal, dammit.   '),(25119,2238828,'03,May,2004','2004-05-03 00:00:00','    urlLink Abomination ,  urlLink abomination .  However, as  urlLink Senator McCain  points out, with brutal honesty,  it happens .   '),(25120,2238828,'03,May,2004','2004-05-03 00:00:00','    Back in Manila, but still with a touch of the weightlessness of my madcap weekend hanging about the soles of my feet and somewhere behind my eyes. We did not leave Friday evening, as we planned; all buses out of Manila to Baguio were fully booked all the way up to 9AM. Instead we hung out at Whistlestop, a 24-hour bar/restaurant, recounting our wacky childhood at Community of Learners. (As happens quite often, I ended up losing much of my dignity, courtesy of the recollection of childhood antics. How is it that I remember being a quiet, shy child, when so many of my classmates remember a crazy loud girl who rollerskated full-tilt down the school driveway and chased grasshoppers through the lawn? Clearly they are all schizophrenic.)  We ended up not taking the bus at all, but a cramped van whose driver insisted on packing us in, four people to a row - when clearly only three people could sit comfortably.  We\'re gonna get cricks in our necks,  K groused, as we shifted about, trying to find a comfortable way to sleep. There was a brief interlude where B, her brother, attempted to keep me awake, but in the end all was futile - K and I had come straight to the terminal from work, and we had all been awake for twenty-four hours straight. Comfort or no, we all slipped off within the first hour.  We finally arrived at Bagiuo seven hours later, after a few mishaps that led B to christen me a jinx - a title that I protest most energetically. It is not my fault that typhoons come, methods of transportation break down / go on strike / run out of fuel, reservations are cancelled, and roofs leak. Those are events completely beyond my control, and the fact that I am always present when they occur does not signify.  At all . It took us another hour, cruising around in a taxi with a cute and friendly driver, to find a hotel - as with all properly madcap weekends, we had not even made a reservation, nor had any clue where we would actually stay once we got there. We circled Baguio twice before finally settling on the first hotel we\'d seen.  I didn\'t do much in Baguio - there isn\'t much to do, really, in less than 18 hours. K and I went shopping while B napped, and returned proudly laden with our booty, which we spread out on the bed while B obligingly oohed and aahed. After that we went out to dinner beside Burnham Park, gorging luxuriously on all manner of grilled food.  This is pleasure eating,  B said defensively in response to K\'s guilty exclamation of  Oh, we\'re so greedy!     I haven\'t eaten for pleasure in such a long time,  I said, chewing reflectively. There is a huge difference between going out, grabbing a latte or a sandwich, and then returning to the office, and actually sitting down, perusing a menu or a display, and then concentrating fully on eating and enjoying the people you\'re with. Pleasure eating is enhanced by the knowledge that you have not done this in so long; flavors are richer, fuller, augmented by the feeling of having existed so long on the metaphorical bread and water. As the  urlLink poem  goes,  bread I broke with you was more than bread .  After that we went walking in Burnham Park. The evening was cool, made slightly chilly by the soft wind sweeping in off the lake. There were boaters, and B asked if we were going boating.  Over our dead bodies,  K and I cried - K because physical exertion (you are supposed to row your own boat) is not her idea of a vacation, and I because I had gone boating there before and gotten cornered in a remote part of the lake by some hooligans. We walked, instead, through lanes filled with families and young couples and children running alongside us asking if we wanted to boat, to buy hats bags shoes sunglasses, lined with trees trimmed with lights. In the distance, a huge searchlight roamed the sky, lacking only Batman\'s logo to make it the Bat-signal.  It\'s Decades,  laughed B, naming one of the more popular bars in Quezon City.  There\'s a Decades Baguio now! And I bet if you stepped in, all the regulars would be there ...    And because it was that particular night, in that particular spot in Baguio, I thought,  Yeah, maybe .   The next day I woke up early. K and B were supposed to have breakfast with me, then drop me off at the station, but they\'d slept late the night before so I couldn\'t wake them. I slipped out alone instead, and headed over to Pancake House, which was a brief walk away from our Hotel. This Pancake House was particularly delicious - it was on the second floor, and most customers chose to eat out on the terrace, breathing the crisp air and watching the cars pass by. I had cinnamon toast and eggs and hash browns and extra-crisp bacon and coffee, and everything tasted like ambrosia.   There was so much I didn\'t do - I didn\'t eat strawberries, or buy flowers, or browse through the markets, or climb hills just to look down and laugh with the exhilaration of having gotten so high. In the terminal, waiting for my bus to arrive, I took down the numbers of hotels and took note of their rates. I do not know when, exactly, but I will be going back.  In the meantime, I have peanut brittle and chocolate crinkles. If anyone wants some, all they have to do is visit me here at my office.   '),(25121,2238828,'30,June,2004','2004-06-30 00:00:00','    As I type this,  urlLink Gloria Macapagal-Arroyo is being inaugurated as my country\'s fourteenth president . I had a lot to say earlier, but a lot of what I have to say is unprintable. Therefore I shall not say anything.  I have an insane amount of work for someone who is LEAVING THIS OFFICE TODAY, which contributes to the not saying anything - because what is the point of a blog, after all, if you muzzle your unprintable thoughts when you write in it. It is more like I am percolating my thoughts, and waiting to see if they produce anything a little less hostile. If moods had physical manifestations, six-inch steel spikes would be bristling out of my skin right now and white flame would be shooting from my eyes - not a very pretty sight, but useful nonetheless, as people would know not to approach me.  Hence: Arroyo is the new Philippine president, there is a typhoon on our shores, I have had less than three hours of sleep, I have caught at least three people conversing with my breasts rather than my face, and if  one more person  tries to call me a cutesy nickname instead of my  given nickname  - which is only one syllable and does not contain any alien phonetics - they will be slowly and painfully relieved of their vocal cords.  That\'s all for today, I think.   '),(25122,2238828,'21,June,2004','2004-06-21 00:00:00','    The word of the day is sex - let\'s go back to your apartment, I need to have a word with you.  Don\'t look at me like that - it was an old pick-up line and it was dying to have a make-over. Besides, it was something that occurred to me while browsing through  urlLink this entry  in  urlLink the Ferrett\'s  journal.   I\'m back, after a pretty grueling weekend - the last couple of weeks were a rather busy buildup to that weekend, really. A workmate, T, and I, were assigned to come up for a program for the office anniversary-slash-teambuilding, which would be held in  urlLink Tagaytay  (er ... well, with out  the well-endowed woman in a bikini). At first we thought it would be fun - and it was, with all the crazy ideas we had. Then reality intruded, in the form of a extremely tight budget and the realization that we were, pretty much, the youngest people in the office. Which meant that while we were doing our supertopsecret!planning, people would pop in at random moments and ask,  So, what are you guys planning to put in the program? , complete with suspicious look. After two or three such encounters, T turned to me with a wounded look and asked  Why do people think we\'re so evil? We wouldn\'t make them dress up like sheep and do a tapdance to Handel\'s  Water Music , for example.    Maybe the fact that you came up with that idea so easily?  I suggested.   True, they were all well-meaning, but it was a harassing experience nonetheless. Toward the end of last last week, I gave up on being meek and mild and metamorphosed into Wicked!rei.  You,  I said to one officer,  will be doing an interpretative dance. Don\'t worry about the costume, T and I will have it ready by Friday if we have to sew the feathers and sequins in ourselves. You,  I told another,  will be responsible for leading the group singing. You\'ll need to get someone to choreograph the corresponding dance.  Worried looks began blossoming on the faces of people in the office, only to flower in full-fledged panic when I brought various odd CDs to the office, offering people a choice of songs for their interpretative dances.  One officer, apparently convinced by my fine acting, actually came over to select a song. I took one look at the song he\'d selected - Madonna\'s  Like a Virgin  - and immediately vetoed it.  Pick a more  believable  song, R,  I instructed, giving him an impudent grin in answer to his glare. rei, office bully. I shudder to think what kind of tyrant I will become once I actually gain seniority over people.  At any rate, the program went off swimmingly - I started it off by warning everyone that T and I had prepared a Plan A and a Plan B. Plan A, I explained, would be the pre-approved, tame, family-oriented program that T and I had prepared.  On the other hand,  I said,  If at any time you are bored with our program, then T and I will sit down, take a backseat, and Plan B - the interactive program - will take effect. This is where everybody gets up and has to do a presentation - interpretative dances, skits, songs, etc. So just let us know if you ever get bored, yeah?   Having effectively terrorized everyone into forced enjoyment of our program, we encountered very few glitches. Everybody was happy with the games - a trivia contest and a fast-paced game of charades -  and  with our prizes: a bag of Hershey\'s kisses for the winners,  urlLink Serg\'s Choc-Nut  (a local sweet that purports to be chocolate but tastes more like hardened peanut butter) for the losers. (Lest anyone protest that we were being too cruel to those who did not win, let me reassure them that Choc-Nut is actually a local favorite.)  T and I were completely numb with exhaustion after the program was done. She got into the car heading back to Manila and just sat there, napping, while I headed for the dining room - where everyone was having snacks and tea - and just sat there, staring off uselessly into the air. We headed for Cafe Lupe, a cafe on the hillside of Tagaytay overlooking the Taal Lake, soon afterwards for some  more  snacks and coffee. I blame exhaustion and overstress for what I did next - I ran toward the railing, leaned over it, and screamed  Jaaaack! I\'m flyiiiiiing!    Having gotten that out of my system, I was more normal for the rest of our stopover at Cafe Lupe. This, of course, relieved my officemates and my boss, none of whom looked very excited about the prospect of having to rescue me in the event I suddenly took it into my head to do a swandive off the railings.   Saturday I got dolled up and headed out to a wedding I was supposed to sing at - only to stop right outside of our car, panicked, because I realized that my toes were  sinfully  ugly - I hadn\'t had time to go for a pedicure in ages.  We\'ll be late,  my mother warned, but I ignored that.  I CANNOT go to a wedding looking all glam and then having gross toes, Mother,  I pointed out, and she relented and drove me to a nearby salon, where I harassed the staff into giving me a manicure  and  a pedicure in half an hour. Feeling more human with my pretty nails, I went back into the car and on to my destination.  A note on the wedding singer stint: I\'d been asked to sing at my officemate\'s wedding a mere two weeks before the event, and - up to two hours before the wedding - had not met the guy with whom I was supposed to do a duet. When I finally found my duet partner, he turned out to look like  urlLink Dexter  in a violently purple shirt and thick black glasses - and then he opened his mouth and you forgot all about that, because he had a lovely, rich tenor that I despaired of matching with my uneven soprano. Fortunately my prayers were heard and I managed to rise to the occasion. With one successful wedding beneath my belt, I am more confident about my next wedding-singer-stint, which will be in October at my cousin\'s wedding.  Hell Weekend\'s over, finally. Back to work - normal work - and preparing to move to my new office.  [ EDIT : I  was  going to post pictures of the view, but they turned out frighteningly big. How do I shrink them? Yaz?]  [ EDIT , part two: I friendslocked a couple of my posts in LJ, simply because I figured badmouthing someone online who cannot defend herself, even if it  was  provoked, is pretty bad form. I\'ll be doing the same thing in future, so if anyone wants to read those posts just let me know and I\'ll ... I\'ll work out something. I was thinking of creating a new LJ and friending that one and then giving  that  username and password out, but I\'m already pretty schizophrenic online as it is so maybe not.]   '),(25123,2238828,'14,June,2004','2004-06-14 00:00:00','    A  urlLink quick look  at the car my parents are trying to convince me to buy. In an effort to make the pitch irresistible - they know me  too  well - they told me that they would let me get the red one. (This is a huge concession on my mother\'s part, as usually she insists on getting white cars since the color reflects better in the dark, making it safer in general.) All weekend I\'ve been teased and tempted with visions of a shiny new red car that would be mine, all mine -  You can do anything with it!  my mother coaxed, sounding for all the world like she was trying to convince a reluctant toddler to go to school.    Can I put a fluffy white cover on the steering wheel and the dashboard?  I asked, to test the waters.   My mother : Sure!   Rei : And have furry white-and-red giant dice dangling from the mirror?    My mother : Um ... okay, if that\'s what you really want.   Rei : I want to name her Chazm. That\'s short for Chazzy Chaz the Star, but we\'ll tell everyone it\'s short for Chelsea. And really it will be a tribute to Chester Bennington, who is cute and dorky and cool and will rock this car.   My mother , slightly less enthusiastic: Well, it  will  be your car.   Rei :  urlLink Freakchild  and I will replace the regular car light with a cover that glitters and sparkles like a disco ball.   My mother , harassed: You realize your father will be driving this car, too.   Rei : Pfft, it doesn\'t bother him to stand in line to pay for our lingerie, why should it bother him to drive a glittery sparkly red car?  Therefore I was almost sold.  Until  I learned the catch - that is, I will be paying over  half  of my monthly salary for this thing for FOUR YEARS.   Four years, people. A girl who wants a car that glitters and sparkles cannot be tied down to something that takes four years to pay for. I want that red car with the deepest darkest desires of my shallow little heart, but I cannot see myself tied to a regular payment scheme and forgoing so many other opportunities for four years. Therefore my parents will have to find other ways to tempt me.   I suppose nothing really highlights my fear of commitment as much as this situation, does it.  [ Edited to add : Oh yeah, did I ever mention that I have a  urlLink Gmail  account? It has 1GB of memory and if you\'re not afraid of people being able to search through your mail, it\'s pretty cool. Apparently I can invite someone else to use it, so ... anybody want one?]   '),(25124,2238828,'08,June,2004','2004-06-08 00:00:00','    Many Filipinos did not like Lee Kuan Yew or his blunt criticisms of Philippine sociopolitical culture, but there are times when he just  urlLink hits the nail on the head . It stings, the words \"Filipinos of short memories\" - but how true. How else do you explain our voting Jinggoy Estrada as a senator, even as he has a  urlLink pending graft case ? How do you explain our voting the children of  urlLink Ferdinand Marcos  into position, barely a decade after he was forcibly ejected from the country for the crimes committed by his administration? My father has a scar on the back of his head from an injury sustained from a clash with police, back when he was a student activist. When we watch movies like  urlLink Dekada \'70 , his eyes cloud over with painful memory. How do we forget these things? How do we justify putting these people back in power?   Is it a tendency to forgive too easily, or - more frighteningly - is it resignation? Are we so numbed by years of mismanagement and neglect that we no longer understand that we have rights?   I love this country so much it hurts.  - - -  On another note, I have been, for some time, seriously considering creating a second journal.  Synesthetique  will remain here, but will be relegated to being a personal journal - a place where I post the \"today i woke up and ate chicken and got splashed by a car - damn car\" posts that Yaz and I were bitching about in chat yesterday. The second journal -  Luminae , resurrected - will contain political stuff. Not always serious - the occasional post will be composed of nothing more than a link to an article and related snark - but nevertheless, all the sociopolitical, quasi-serious contemplations and bitching will be placed there.   The thing is, it might be a little ambitious. For one thing, I\'ve always thought of my journals as private areas - something like beanbags around a fireplace on a cold rainy evening, where people gather with mugs of spiked coffee and shots of whisky and shoot the breeze. Therefore, the only people who read this are friends of mine - and, hmm. Could I justify asking them to spend more time checking on another blog, which caters to a very special interest (ASEAN sociopolitics, with a strong bent toward Philippine political analysis)?   For another - I am highly irregular when it comes to posting. There are months when I post almost every day, and there are months that contain a single entry and nothing more. Would I have the energy or inclination to sustain two separate blogs?  (Er - which is to say, I\'m asking for help. Should I or shouldn\'t I?)   '),(25125,2238828,'07,June,2004','2004-06-07 00:00:00','    Such a funny thing. My sister  urlLink Freakchild  and I have been pretty vocal lately about wanting stability, about wanting something solid of our own.  I want to  know  something for sure,  she told me yesterday, when my parents held a career planning seminar for our cousins and the two of us. In our hands were blank papers, on which we were supposed to write our long-term goals; we were sitting by the curb, frustrated, because our dad had told us that while our goals were well-stated, they were all at most mid-term goals.  The problem is,  she griped,  I can\'t see past the end of this year. I don\'t know what\'s going to happen, I can\'t figure out what I want, I don\'t know anything.    It\'s funny, because we never realized until recently how transient our lives have been - touching down here, taking off, moving on.  I feel like I don\'t know you at all,  a friend of mine marvelled a few weeks back. I was taken aback, myself -  Haven\'t I ever told you about these things before?  I asked, and he said no. On the other hand, how do you fit your entire lifetime into a few years\' worth of anecdotes and friendly confidences? I feel as if my life - and correspondingly, my friends\' lives - are dissected into portions and organized: first chronologically, by the time period I became acquainted with each friend, and second by interest, by the particular themes and issues that each friend and I have in common. I told a friend of mine once that I never expect to know  everything  about any of them, which is true. In the period between the time we last were in touch to the time we finally meet up again, so much has happened, could have happened, should have happened - how do you explain these things?   You end up not explaining them at all, by merely picking up the surface threads and moving on. If you\'re lucky, you keep the communication lines open long enough to learn something deeper and more valuable about an old friend - this has happened with Yaz and JP and a few other friends who are not online. Otherwise the surface threads remain on the surface, and you start to shuffle each other into the category of \"activity partners\", where you remember to contact each other only when you get involved in an activity in which you know the other friend is interested, too.   It\'s funny, too, how the one person who started the ball rolling on all this self-realization is no longer in my life. I wonder, randomly, what he will think when we meet up again. In the five months or so that we have had absolutely no contact, I have lost weight, cut my hair, developed new interests, made new friends. I can barely remember his face or his name, some days, although there are others when a memory involving him surfaces readily enough - these, however, are rare. How quickly I forget. How swiftly my heart heals.  With all of this running through my head, I wrote down on my blank paper:  I want a steady job. I want a home of my own.  and folded it up, small, my heart\'s desire crumpled into lines and shoved into the corner of an old notebook. I don\'t know if I really want these things, after all. Maybe I\'m still traumatized, after spending three months of last year virtually homeless.  I haven\'t unpacked my suitcases for three months,  I wept one night about a year ago, burying my face in E\'s shoulder.  I touch down in one place, rotate the clothes I\'ve got in them, and then leave again. I don\'t want to move anymore. I don\'t want to leave anymore. I want to put my clothes in a closet, I want to sit down at my own table, I want to wake up in a familiar bed.    There was a moment of silence, his hand moving through my hair, before he spoke:  When I travel,  he said gently,  I never quite get used to sleeping in a bed other than my own. It\'s like I don\'t really fit there - like the pillows aren\'t the right shape and the bed doesn\'t feel right ... Oh yeah, I sleep, but I never sleep as peacefully as I do when I finally get home and lie down in my own bed, the one I\'ve had since I was little.  On and on he went, his voice quiet, until I was no longer crying and just lay there, silent, my head still pressed against his shoulder.  One sweet memory of something I can barely see anymore. It\'s sad, how sometimes I reduce people to memories.  I think I\'m moving again, to another job. It\'s not such a big move, and huge parts of me are happy about it, the way I\'m happy about the greyness of the sky and how I can walk around bareheaded in the rain. It\'s in the same government department, and my new office will be right across the hall from where I work now. Still and all - another new place, new people to befriend, a new boss to figure out. How many people would laugh in disbelief if I told them that I hate being the new girl on the block?   I wonder when I\'ll stop moving, when I\'ll find my niche and stay there. I wonder if I\'m supposed to.   '),(25126,2238828,'03,June,2004','2004-06-03 00:00:00','    I am tired, but I rule. Why, you may ask? For two major reasons: first, I rule simply because I say so. This is a valid reason, as all my childhood friends will tell you. I rule. Period. No questions asked. (You can all also rule, but go do it in your own journals so that the throne area doesn\'t get crowded here. I promise I will be appropriately adoring when approaching you in your own domain.)   (  urlLink The second reason.  )  In other news: I went to the  urlLink Prisoner of Azkaban  Manila premiere yesterday - or was dragged there, rather, by one of my officemates whose sister was selling tickets for a project. The sister was happy - she managed to sell out all her tickets, because by golly there were seven of us from the office alone; I was happy, because I\'d wanted to watch  Prisoner of Azkaban  but hadn\'t wanted to watch it alone; everybody else was happy because we had dinner at an Italian restaurant afterwards. If you are looking for spoilers, however, you will not find them here. Go watch it, and then message or e-mail me and we will squee or nitpick according to your personal preferences.  Various reactions, while watching the film:   Me : OMG Harry and Hermione are so close he is HUGGING HER SQUEEEEEE. Oh, and why does the werewolf look like a poster boy for  urlLink Sally Hansen ?   KS : Who? What? Where? When? Why? How?  Me : Dude, READ THE BOOK. Can you not see that I am busy squeeing?  J : I\'m gonna lay the smackdown on this dude in front of me. Could his hair GET any bigger?  R : I do not get why everybody is so crazy about this movie, but my umbrella matches my shirt so I\'m happy. Also, hot chicks! Yay.  KY : Hello, hot chick. As you can see, I am suave and intelligent and not with the rest of  these  n00bs.  Everybody : HEY.  It is interesting to note that JP thought this was all very amusing until I reminded him that my officemates are all commercial attachÃ©s, making them responsible for the future of our country\'s trade relations with the rest of the world.   JP : Remind me to migrate.  I have a sneaking suspicion that I will have to watch the movie again this weekend with my sister in order to understand what it was all about.   '),(25127,2238828,'01,June,2004','2004-06-01 00:00:00','    There are many reasons why I don\'t want to work in government - why I never  planned  on working in government - and this morning was one of them. (  urlLink Bitch, PLEASE.  )   '),(25128,2238828,'30,July,2004','2004-07-30 00:00:00','    How hard do my friends rock? SO VERY HARD. Go over right now and check out  urlLink Solemate\'s article  on natural family planning.  The coolest thing about all of this is that we don\'t even agree on how family planning should be conducted. I believe strongly that women should be the primary arbiters and decision-makers with regards to what happens to their own bodies (although I am also strongly against abortion), and that therefore they should be given the option to use artificial contraceptives to prevent conception. Solemate writes from the point of view of one who is firmly linked with the Catholic Church, and therefore presents interesting and thought-provoking arguments on why natural family planning, or NFP, is a better solution to the problem than artificial contraceptives.  This is cool because we have spoken about this before, passionately and intensely. I feel very strongly about my position, and Solemate feels the same way about hers. Nevertheless, we respect each other\'s positions and convictions, and the fact that ultimately we come from the same starting point - the belief that every child should be wanted and provided for.   Remember: every child should be wanted, should have a chance for a future. Contraceptives aren\'t just about safe sex, they\'re about making sure that no human being is an accident.   In sum - go read Solemate\'s article. I do not know how I keep on making friends with the most talented people in the world, but hey, sometimes you\'re born lucky.  [ Edited to add:  Okay, to all those weird funky people messaging my phone - I know my number\'s pretty easy to guess, but gosh darn,  STOP IT ALREADY . I don\'t know you, I don\'t talk to strangers, and  no , I do not want to attend an Open Water Diver Course Orientation.  Also,  urlLink freakchild  has now gotten a Yahoo ID. Ph34r. One day, when I am trying to explain how she got her nickname, I will post a chat transcript. Just so y\'all see how freaky she is and how I, her longsuffering normal sister, patiently deal with her oddness. I\'m a saint, I know.]   '),(25129,2238828,'29,July,2004','2004-07-29 00:00:00','    [For the Occident and the Orient. I\'m not really sure if I really did love you, but in all honesty I tried.]   Stay,  he says. I don\'t answer. The brief silence is broken by other things, and the word that could have been a plea or a command gets lost in other conversations, other considerations.  I get scared now, when I hear this word. Whenever someone says this word to me, what it really means is  goodbye .  True enough, he is gone. I can never accurately predict how long it will take before he goes, but it always happens. One day I turn around and he is no longer there, and I wonder how long he has been leaving without my noticing that he is slowly moving away. How long he has not been mine. How long I have been pretending.  - - -  It\'s funny, how strangeness and oddity are things you learn about only after you lose your innocence. When you are younger, everything is normal. I ate fish eyeballs and duck eggs with tiny duckling fetuses embedded in them and pickled pigs\' ears; years later, I can still remember the feel of them on my tongue, the way they crunched or slid against my teeth. I remember how they tasted, and sometimes when I watch people eating them, I lick my lips in memory.   Since then, though, I\'ve learned restraint and inhibition and revulsion. A friend offers me a bite of her  dinuguan  - pig innards boiled in pig\'s blood - and I shudder and turn away. She laughs and redirects the spoon toward her own mouth, and I watch her eating with a twinge of envy. My tongue remembers how it tastes and yearns quietly for it, and my mind sharply slaps it down with learned revulsion. The things I could do, and say, and taste, if I had not learned that they were wrong.   There are some things I  do  regret learning.  - - -  It\'s ironic, but the people who tell me I am strange are the very people I used to consider strange myself.  I\'ve never known anyone like you,  he said to me once.    The first time I met someone like  you, I retorted,  I was almost twelve years old.   We were talking about houses then, I think. I was telling him about trying to convince my mother to move to a new condo because in the one we were currently staying in, I didn\'t have a room. He laughed:  You talk about moving house the way people talk about buying a new cellphone.    But we honestly do need to move,  I insisted.  We\'ve outgrown this place already, even my father says so.    You\'re not supposed to outgrow a house,  he told me.  You\'re supposed to grow with it. I\'ve lived in the same house all my life.   I see my fascination mirrored in his face, and from the outside we must seem funny: two people absolutely amazed by this odd creature we are confronting, because they are both completely sure that  they  are the normal one. In his world, people stay put and go to reputable schools and form concentric social circles, where you know someone who knows someone who knows someone and they all hang out in the same places and you date them and break up and eventually run across them again because after all, you move in the same world. In my world, people appear and disappear and pop up unexpectedly and the circles don\'t overlap, so if you don\'t want to see someone you don\'t have to, you just go with another batch of friends who don\'t know that person and don\'t hang out where s/he does.   How do you not get tired or bored of the same house, the same neighborhood? Do your parents not get bored of working for the same organization for more than five years? Why are you so attached to your house, your car, your  things ? How do you know this city so well? Where do you learn the language-beneath-the-language, the meanings of acronyms, the implications of phrases, the politics of gestures and school affiliation and music preferences?   He answers me the same way I answer his questions: with bewildered eyes. What is strange to me is not strange to him, and vice versa. We do not even speak the same language, when we talk to one another.  - - -  Leaving never used to be such a big thing. In my world, everyone leaves eventually. The girl who is your best friend might move to another country in a few months; your teacher is called back to her own country for a family emergency and suddenly someone else is teaching you about early Sumerian civilizations. Besides, it wasn\'t always permanent. Your classmate who moved to Norway for a couple of years might come back again in the sixth grade, taller and older and bumpier. You run into schoolmates in the airport, en route to their own destinations. You come back for a visit, or they e-mail you to say they\'re dropping by your city and maybe you could meet up for a day or two.  It took me a year or so to learn that you were supposed to say goodbye before you left - preferably throw a party, to mark your departure.  But I\'ll only be gone for the summer,  I told him. He was not appeased.  In college, my friends insisted on seeing me off at the airport, when they found out that I would be leaving the next day and that I was going there alone.  You won\'t be able to come in,  I warned, because NAIA is the unfriendliest airport in the world.    We won\'t let you leave alone,  my friend told me, and I didn\'t ask why it was so bad because it didn\'t seem like the kind of thing that could be explained. In the airport, while I was reading to pass the time, he called me.    I didn\'t want you to leave without saying goodbye,  he said.   With each friend who insists on knowing about my departures and arrivals, leaving gets heavier. It hurts and confuses to hear the reproaches:  Can you  please  let us know next time when you\'re going and where you\'re going and when you\'re coming back , or  How could you have come here and not let me know you were here?   The reasons that make sense to me -  But I didn\'t know until the last minute that I was coming there/going away  or  There was just too little time to meet up with you, I thought I\'d just contact you the next time I was around  - sound awkward and hostile when I am using them in my explanations. I hang my head, apologize, promise to send advance word of my to-and-froing, vow that I will let them know when I am in the area no matter how short my stay is.    Promise me you\'ll tell me when you leave,  he says, and for once there is no twinkle in his eyes.  Promise me.    All right,  I say. I still haven\'t figured out how to get them to make that same promise - perhaps because they don\'t understand that there are so many ways of leaving.  - - -   Stay,  he says again, and impatience froths up inside me.   What for?  I ask, and he opens his eyes and looks at me. I wonder what he will say, and for a minute I regret asking him that question.    For your friends? To be with your family?  he says, but his answer is more a question.  Haven\'t you ever thought about just settling down in one place and staying there? It\'s your country, after all.   ( Even her dreams smell like death to me,  says Savannah Wingo of her her mother, in Pat Conroy\'s  Prince of Tides .)   My friends and family will be my friends and family no matter where I go,  I say. I don\'t really understand where this conversation is going.  Or I do, but I don\'t want to understand.   He closes his eyes again, and I turn away.  I wonder when it will occur to him that I wasn\'t the one who left.   '),(25130,2238828,'28,July,2004','2004-07-28 00:00:00','    Still twiddling with the SME Registration Statistics of DOOOOOM, my goodness - I\'ve finally wrestled them all into some semblance of order, but now I\'ll have to analyze them, argh. I took them home last night to work on - I could have finished them yesterday, but KS and T messaged my phone and invited me to a brainstorming/chillout session with them, and I can never pass up an opportunity to chill with any of my girls. Therefore I struggled to do the calculations by hand while we threw around ideas for their office team-building thingamajig, which is going to happen in November.  Quotable quotes from yesterday\'s session:   T : We do not think of ourselves as merely ambassadors. We are AEP*s. [brief discussion on the various meanings and implications of the inclusion of the word \'Plenipotentiary\' in the title.]  rei : You know my favorite part of that title? The \'Extraordinary\' part. \'Cause, you know, all you ordinary diplomatic folk can just muck about over there ... the line for the EXTRAORDINARY and PLENIPOTENTIARY ambassadors is here, and we kick your asses all the way back to Beirut. It\'s like being one of the Justice League. \"Don\'t mess with me, man, I\'m Extraordinary. My title says so.\"   rei : If anyone asks you where we got this \'FOCUS on STARS\' thing, tell them it\'s from that Latin phrase  ad astra .  KS : Doesn\'t that translate into \"to the stars\"?  rei : Okay, YOU think up an acronym for TO THE STARS.  KS : ... This is rapidly turning into the kind of team-building activity where no one is allowed to ask questions.   T : ... So, FTSC stands for the Fruity, Tasty, Sweet Corps!  GG : Is this a final draft of the program or do we still have veto power?  * Ambassador Extraordinary and Plenipotentiary. I\'m not making this up.  - - -  Drat. I was going to write about exploring the Catacombs in Rome, and our guide lady who scared us into blindly following her about by prefacing the tour with a menacing warning about how others had wandered off and  never been found again .   However, I still have to analyze these dadblasted statistics, so perhaps later.   '),(25131,2238828,'23,July,2004','2004-07-23 00:00:00','    One day I honestly  will  talk about what\'s going on with my links list, because I need to add people and introduce people and so far, it ain\'t happenin\'. Argh. Sorry.  Although you need to know that  urlLink nikki  is an amazing grammar and trivia freak. Recently she  urlLink spazzed out about that stupid Tower Records ad along EDSA , making me a fangirl for life. Also, how can you not love a blog that blatantly celebrates obscure occasions such as  urlLink Father-Daughter Take A Walk Together Day ? (I took a walk with  my  dad, did  you ?)  You also need to know that  urlLink mia  has a  urlLink new place , which you should go see just for the colors. And new stuff. Yes, you really do need to check out that place on top of Ephemere, because Strange Joy is everyday-Mia, while Ephemere is pensive-Mia. All Mias are fun to be around, so gogogo.  And take a while to check out  urlLink Gabby\'s  blog, because it was such a lovely lovely surprise.  urlLink Nikki  has a list of referrers on her website, and I clicked on one of them, and yeah, I wish I knew how to make pretty pictures like that.   If you\'ve been around for a while, you might want to know that  urlLink laya  is back online, although her phone is broken  again . She has some art online at  urlLink another site , so gogogo to that one too.  That\'s it, I\'m off.  [ EDITED TO ADD :  urlLink Sandee , for several reasons. First, Gabby says she\'s responsible for the bootiful artwork, and second, she\'s Gabby\'s cousin. Why should you care, you say? DO NOT MAKE ME LAY THE SMACKDOWN ON YOU, BIOTCH. You should care because they are cousins and both got liberally stuffed with the \'good writing\' genes. I\'ve been surfing through Sandee\'s LJ for the past half-hour and trying to keep from rolling on the floor laughing because our floors are  seriously  filthy. (To everyone who has told me their taxes pay for my office and salary, let me tell you: your taxes aren\'t paying for my hygienic and sanitation needs, so CHILL.)   A sample of the Sandee-wit:  urlLink Harry Potter 3: Prisoner of Azkaban, Sandee-fied . She\'s a highschool senior, yet. I was never so witty when I was that young. I would be jealous if I weren\'t so entertained.  Now, back to slaving over these murderous statistics on SME registration, dagnabbit.]   '),(25132,2238828,'21,July,2004','2004-07-21 00:00:00','    It\'s all about  urlLink Angelo de la Cruz  nowadays, and a small part of me is glad. It\'s the same part of me that ached and cried a little, when the banner headlines quoted him saying \"I\'m coming home\", because at that point those three words sounded more like a plea than a declaration. More like an exhausted, terrified man dreaming out loud than a man who  knew  that he spoke the truth. I don\'t want to be bitter - really, I don\'t - and I want to give my government a chance, because otherwise I\'ll lose my sense of humor about the whole thing and that\'s the last thing I want to happen.   Although the thing is, you can\'t really trust what this government says anymore. When it\'s being run by a president who swore to her people that she would not run again, in order to heal political divisiveness - then turned around six months later and said that she  needed  to run, also to heal political divisiveness - then all official statements begin to sound doubtful. When it\'s being run by an administration whose mandate is shaky - a slim margin of victory in the just-concluded elections, among suspicions of cheating - then you aren\'t so sure they\'re acting in the people\'s best interest. When all recent appointments to the Cabinet and other plum positions have been seemingly politically motivated, then ... it\'s hard to tell anymore. Deep down inside, I think  I\'m Filipino, she\'s Filipino, we\'re all Filipino. This is our country, our people. Surely they can\'t forget.   I don\'t know.   I\'m grateful for de la Cruz\'s return. I\'m glad my president fought for his return, despite widespread global disapproval. I\'m glad that at least this one Overseas Filipino Worker (OFW) was able to return home upright - not in a box, not in pieces, not wrapped so that the relatives don\'t have to see the burns or the bruises on the body, the way a loved one has become a piece of meat. You can\'t fault me for that, no matter what your ideological stand on the situation is.   Rumor has it that Angelo de la Cruz will be presented at Arroyo\'s very first State of the Nation Address (SONA), which will take place next Monday, if I\'m not mistaken. Yesterday he was presented at a press conference, which I heard on the radio which is perpetually on in our office pantry. I don\'t think he understands what is happening, although he sounds overwhelmed and grateful for everything.  What I want to know now is, does this mean anything? We saved  one man . No one wants to talk about Sarah Balabagan, who was raped in the UAE and then beaten and jailed for defending herself against her rapist. No one wants to talk about Flor Contemplacion, who came home from Singapore in a box. No one even hears about the other OFWs, the ones who creep into our foreign embassies with heads and eyes down, as if they were ashamed they were alive - the ones who have been kicked so often and so hard that they can\'t remember what it was like to be treated kindly.   Why can\'t we just bloody do something about how cheap Filipino life has become?  We laughed about it then, but it\'s not funny now - my family was having dinner at a Korean restaurant a couple of days ago, and there was a television turned to the news channel behind our table. Someone, I don\'t remember who, was explaining that de la Cruz would be given job opportunities in the Philippines, his children would be given scholarships, that his family would be guaranteed job opportunities once they became part of the labor force. We sat there, quietly nibbling on our food, digesting that information.   (In order to understand the irony of the situation, you have to know this: we returned to the Philippines in September of 2003. All of us were jobless and my sister had been forced to quit school mid-semester. To add insult to injury, the only one of us who could get a job was my sister, who worked for a call center, a bar, and a record company in rapid succession. I got my current job through a circuitous route of personal and parental pull, which makes me bitter because I  shouldn\'t  have to use pull at all. My father remains still unemployed, although he receives short-term consultancy work on a regular basis.  This was all  dÃ©ja vu , of course, because the exact same thing happened when we returned to the Philippines from Bangkok.)  The newscaster kept on droning, listing the various benefits and gifts that would be Angelo de la Cruz\'s as soon as he touched down in the Philippines. I snitched a rice roll from my sister\'s plate, then gestured toward the television.  See, Father?  I said.  Coming back to the Philippines is easy. All you have to do is go get kidnapped in Iraq.   Yeah, I\'m bitter. It\'s the coffee, I swear.  [ EDITED TO ADD : Aargh, this has got to be one of the most selfish things I\'ve ever written. Although that wasn\'t the original intention, honestly. The point of the thing, really, is that you have this One Exceptional Occasion, and everyone goes all out for it, and then - fizzle.  Ningas-cogon , in the local vernacular: a fire that lights up quickly and dies out just as fast. The point of it is that I\'m happy for the man, but I want it to translate into something longer-lasting and substantial: protective measures for OFWs, good job opportunities and reasonable wages for the people who  are  here, a chance at a decent education for all children, et cetera.   It isn\'t all about my family or my situation, so if it read like that - I\'m sorry. We\'re good, it\'s just - like I said, I\'m bitter.]   '),(25133,2238828,'21,July,2004','2004-07-21 00:00:00','    Oi,  urlLink WATCH IT . Of all of the comments and criticisms that have been flying around ever since Angelo De La Cruz was released from captivity, I find this the most objectionable.   Of course I\'m being unfair by posting this in a blog that is (1) pink and sparkly and (2) read primarily by women. However, when your father has been told repeatedly - in front of your face, no less - that it\'s \"a pity\" he has no sons to \"make him proud\", or \"carry on his name\", or whatever misogynistic nonsense, you rapidly gain a prickly exterior with regards to your worth as a human being who is also female. (Fortunately, my father answers all such comments with a blank, confused look -  But I  am  proud of my daughters,  he says.  And they carry my name, don\'t they?  Hahaha. Love you, Dad.)   '),(25134,2238828,'20,July,2004','2004-07-20 00:00:00','    Apparently, there are people in the world who believe  urlLink wives should be beaten  - albeit with a toothpick, so that\'s all right. The funny thing about fundamentalists of  any  religion is that they all end up flogging the same old dead horse, and yet they\'re all upset at each other - for using different kinds of whips, I suppose.   (There\'s a bit toward the end where they discuss the problem of beating a wife who is a masochist. I\'m sorry, I probably shouldn\'t find that funny, but my goodness, DO I EVER.)  ...  I\'m sorry, I\'m still stuck on the mental image of someone trying to beat me with a toothpick. Here\'s how it goes -   Imaginary Mate : Art thou, indeed, defying me?  rei : D00d, been over this ten million katrillion times.  No , I will  not  watch Vin Diesel movies with you.  IM : Despite my countless pleas and exhortations, you remain hard of heart and head. Lo, I shall banish you from my bed.  rei : Ouchie. But hey, you  do  realize that no nookie for me means no nookie for you too, right?  IM : WHAT?!?! Right, then, moving on to the next stage of admonishment ... Before I continue, please know that this is being done for your own good and this will hurt me far far more than it hurts you.  rei : Can\'t hurt very much, then, because I don\'t feel any pain. Not to downplay your admonishment, but what exactly  are  you doing to me?  IM : Erm. Beating you?  rei : With a  toothpick ?  IM  (defensively): Well, it\'s supposed to be symbolic.  rei : Now I Dead From Toothpick. [Dies of laughter / Revives self five minutes later out of respect to annoyed IM] Okay, consider me  symbolically  chastened.  IM : Does that mean you\'ll watch Vin Diesel movies with me now?  rei : Yes. Symbolically speaking.  IM : Which means no.  rei : See, I  knew  I was doing the right thing by insisting on marrying a smart man.  I had other things to say, but I can barely type through the laughter.   '),(25135,2238828,'19,July,2004','2004-07-19 00:00:00','    See,  urlLink this  just exemplifies what I always wanted to say to those activists back in school. There was a long story about it, about trying to follow in my father\'s footsteps and join rallies and be politically involved, and how I ended up disillusioned because I didn\'t think it would be that violent, that pointless. Violence in and of itself is pointless, and when you use it to try to highlight an issue, then you\'ve lost even before you begin. Story aside, the point to this is that people listen better and with far more sympathy if you try to speak their language, if you approach them on their own turf and dress the way they expect you to dress - I am not advocating selling out,  per se , only translating yourself.  &nbsp;  &nbsp;It is childish to expect people to listen to you \"just because\" you are right; very few do. Witness the success of those who predict the coming of Armageddon on sandwich boards along the sidewalks of bustling cities. They may have valid reasons for believing the end is nigh - who knows? Yet because they walk around in sackcloth and unkempt hair, looking like they have no investment whatsoever in normal life or regular civilization, neither do normal people feel that they have any interest in what those people believe. It is childish to accuse people of \"selling out\" or being \"bourgeois\" for adapting themselves and their message to be understandable by those people they want to reach.  &nbsp;  You do not sell out if your message does not change. The message must remain the same, always. It is only the messenger, the style of language, and the medium of transmittal that changes to suit the audience. Activism has many persona, and not all of them carry picket signs or march beneath a Communist flag. I\'d rather be a conformist  with  a relevant cause than a rebel without any particular cause at all.  &nbsp;  I respect Bill Cosby  so  much.  &nbsp;  - - -  &nbsp;  The tyranny of youth is that you never know whether what you want to do is the right thing - you are torn between fighting for what you dream of doing, and obeying what the older people say because they have more experience. Sometimes when I am at loggerheads with my parents over the choices I want to make vis-a-vis the choices they want me to make, I look at them and think  How do I make you understand? How do you make me understand, when we are coming from two such different places?   &nbsp;  How do I know that I am not dooming myself to failure by insisting on what I want, and how are my parents sure that they are not condemning me to live a safe sad life?  &nbsp;  I hear Yeats in my head again:  &nbsp;  \" urlLink But I am old and you are young, / And I speak a barbarous tongue.\"   '),(25136,2238828,'15,July,2004','2004-07-15 00:00:00','    First: if you are wondering, then yes, there  is  a way  urlLink to get dumped gracefully . Most of it dovetails neatly with my own personal philosophy, which can be summed up in six words: Bitch A Lot, Get Over It. The thing about me is that unless I am actually in a clinical depression, I  cannot  bring myself to mope overlong about anything, because past a certain point I begin to mock myself. When I am tempted to moon over a failed relationship, for example, the following imaginary conversation (or variants thereof) begins to play in my head:   rei : WOE. WOE AND ANGUISH AND DESPAIR, AND PERHAPS RENDING OF GARMENTS (particularly around the cleavage area). ALAS, ALACK, ALAKAZAM AND AZKABAN, FOR I HAVE FOUND THAT YOU DO NOT LOVE ME. WHY DO YOU NOT LOVE ME? WHY? WHY WHY WHY WHY WHY?  Boy of the Moment : Er ... I\'m sorry?  rei : O SPEAK KIND WORDS TO ME, COMFORT MY BRUISED SPIRIT, POUR CALMING OIL ON THE TROUBLED WATERS OF MY UNREQUITED PASSION. Also, please answer the question - kindly note that I asked you WHY, DAMMIT.  BOTM : Look, I don\'t mean to hurt you and all, but I really don\'t love you (the word \"anymore\" can be inserted here as appropriate).   rei : AGAIN, NOT REALLY ANSWERING THE \"WHY\" PART OF MY QUESTION.  BOTM : Um, because you\'re a flake?   rei : SKIVE OFF AND DIE.  BOTM : Okay.  rei : Wait, no, don\'t walk off. What kind of flake, specifically? Flake as in Corn Flake? As in Tuna Flakes? As in A Unique Snow Flake? Unique is good, right? Hey, wait, I\'m talking to you! Yeah, YOU! WAIT I NEED TO MAKE SURE I UNDERSTAND ...   BOTM : (is gone)  rei : Oh, buggre alle thif for a larke.   (Apologies and grateful nods to Neil Gaiman and Terry Pratchett for that last line.)  Second ... okay, dammit, I forgot what the second thing was again.   Third, I HAVE MY CAR. It is a girl - her name is Chelsea, aka Chazzy Chaz the Star, aka Chazm, aka Chaz. Unfortunately, due to idiotic methods of processing, her documents will not be released until today, which means that I couldn\'t take her to work with me as I planned. (This boggles my mind, actually. My mom was driving her all over QC yesterday and I was like \"Um, if she\'s not legal yet, how can you be driving her?\" and my mother - ever the hippie - answered, \"Well, no one\'s arresting me.\") I  will  be taking her to work tomorrow, though, because she\'s my car and I have to imprint myself on her, dammit.  Fourth, I am seriously thinking of reworking some old articles of mine and sending them in to newspapers or magazines and hopefully making some money off them. My first article is already in the process of editing and revision; however, in order to make good money off of this, I have to have a steady stream of articles making the rounds. (MoliÃ©re, you bastard, you were so right about doing it for the money.) Since I don\'t have enough time or self-discipline to sit myself down and actually work out a list of topics to write on, I was thinking of semi-cheating and just reworking stuff I\'ve already written on this blog. If anyone has suggestions or comments, I\'m open. (Oh, and  special note : NO, jps, I don\'t think an article on my sex life would interest the general public.)  I still can\'t remember what the second thing was.  [ EDIT : I remember what the second thing is, o frabjous day! (Sorry, Carroll Lewis.) The second thing is this - yes, I\'ve changed my userpic. JP didn\'t like the first one, I didn\'t like the one he picked out, and at any rate the ones we were using didn\'t seem to jive with my blog. Unfortunately, since I am neither photogenic nor a camwhore, I do not have many recent pictures of myself - case in point: the pictures I am using right now are not even from my own camera.   Therefore in a fit of pique, I selected a picture wherein I was not even facing the camera, and found that, oddly enough, it was stunningly appropriate. Admire the sleek lines of my nape and back, and know that the title of my userpic is \"Talk to the back, bitch.\"   When I find a picture of my face that has as much attitude as a picture of my back, then it will go online. Until then, this is as good as it gets.]   '),(25137,2238828,'12,July,2004','2004-07-12 00:00:00','    At the risk of being labelled obsessive-compulsive (Yaz, love, I can\'t help it, honestly):   What fool of an editor approved  urlLink this article ? There are at  least  five grammar mistakes and confusingly vague use of certain words in an article that can\'t be above five hundred words. In fact, how could an article have been so shoddily written at all? You expect this standard of writing from high school students, from people venting in their blogs or e-mail, from a hastily-written first draft.  You expect so much more from a published article.  Where does the line form for frustrated copy editors? There should be a support group for people like me who have to read articles like this and  know  we could have made it look so much better and yet are repeatedly refused entry into the print industry.   [ Edited to add : Also, I have uploaded my EXTREMELY GAY userpic - so named because I could not leave Yaz\'s neat work alone and had to do stuff to it in Corel PhotoPaint. Namely, blurring it (because closeups of me smiling are always vaguely terrifying) and adding sparkles because -  sparkles , dammit. Hence, the gayness - where gay means \'happy\' as it used to in the days of yore and nothing even remotely close to homophobia, so don\'t even.  The original picture was taken with R\'s camera during the teambuilding trip to Tagaytay. Yaz kindly cropped out everybody else so that the only person I ever incriminate in my wild and wacky word-wanderings is me. Kisses and hard lemonade for you, babe.]   '),(25138,2238828,'08,July,2004','2004-07-08 00:00:00','    Sleep-deprived - not the ideal way to start off a week, but on the other hand, there are events and people worth staying up late for. Yesterday we spent the whole day with friends from Bangkok - it was the first time we\'d seen them since we\'d left all those years ago. Time rolled back and compressed, and it was as if the twelve years in between had never happened. Our mothers congregated around the kitchen and the dining table, as they\'d used to, and we hung around long enough to eat before taking off to AL\'s room to catch up.  There were five of us - AD and AL, brothers; KD and his little brother L, who had been a baby when we last saw him and is now 11; and  urlLink freakchild  and me. We finally found out how old AD  really  was - he had always been everybody\'s big brother, and in the blindly adoring way of children, we\'d always thought he was eons older. Apparently he\'s only three years older than I am, and I was stunned.  Wow,  I marvelled as he grinned and our mothers laughed.  We always thought you were like ten years older and six feet tall, you know?  All those years later he was still the consummate big brother, alternately comforting and wise; sitting at the head of the table, he chatted with our moms about the various adjustment problems we\'d all gone through while the rest of us snuck off to AL\'s room with a glass of wine to do our own bitching.   I feel so safe when I\'m with them, like I\'m eight years old again and squabbling with the boys and AD is patrolling the area protectively, yelling at us for the millionth time not to run by the pool. Sometimes in Manila I am struck by moments of panic or terror, when I feel like I\'m stranded in the middle of a stormy ocean and I have no lifevest and I don\'t know how to swim and I don\'t know where to go or how to get out of there. With these people, though, and a few of my friends, it\'s as if I\'ve got a killer lifevest on and a strong cord that tethers me to safety, and I feel perfectly free to leap out even in the middle of a typhoon.   Of course,  urlLink G  would probably say that normal people wouldn\'t leap out in the middle of a typhoon even if they were CHAINED to a secure place, and she may have a point.   - - -  I\'m getting the car after all, despite all my wibbling. My mom will pick it up Wednesday afternoon and I\'ll bring it in to work on Thursday, just because. Sparkly red car and a wild happy rei driving it around!   Ph34r m3, f00lz.  Speaking of wild happy rei, solemate messaged my phone yesterday to let me know that there\'s a tentative date set for her wedding - January 2006 - and that I\'M GOING TO BE HER MAID OF HONOR!!!  No, of  course  it\'s not going to be a normal wedding. With solemate as the bride and her sister as the organizer and me as the maid of honor and M and V as bridesmaids ... this is going to be  fun .   '),(25139,2238828,'08,July,2004','2004-07-08 00:00:00','    No more politiposts for now - I\'m tired and my head hurts and my eyes hurt and there\'s still coffee and/or beer with my girls later and an e-mail for G sometime ... er, tomorrow morning? My internet connection\'s not working properly and I haven\'t finished this thing that was supposed to be done today, and I have a desk-full of articles that need to be sorted through and read and none of it  is . My shoes are battered and I haven\'t shaved my legs today (it doesn\'t show, but I can  feel  the difference) and my hair is being porcupiny and weird and I think I\'m gaining weight again and I don\'t know whether that\'s a good or bad thing.  I want to live someone else\'s glamorous life for a day.  In other news, the house is getting renovated. Actually, it\'s been in the process of renovation for a few months now, it\'s just that I\'ve been lazy about blogging about it. About a quarter of the house is now done, sort of - by which I mean the kitchen(s) are complete, the new balcony is half-done, my dad\'s office is ready for occupation, and the hallway leading to the living room is pretty much finished, too. As for the rest -   a few days ago my sister  urlLink freakchild  said  Oh, they just knocked down half of our house.  I was tired and braindead after a long day at work, so all I said was  Oh.  (Yes, I am indeed a  scintillating  conversationalist, you should all envy the man who gets to marry me.)   No,  she insisted - freakchild is not the type to be put off by less-than-enthusiastic responses - and pulled me to the front door so I could see for myself.  You must  see. So she opened the door and at first the porch looked the same, so I gave her a Look. She looked back, impatiently, and rolled her eyes.  Look  upwards, she ordered.   I honestly don\'t know what you\'re going on about,  I muttered, annoyed.  Nothing\'s - WHOA.  Because at that very moment I\'d turned my eyes upward, and where I\'d expected to see our rather worn roof, instead there was nothing. Absolutely nothing. My eyes slid up up up, my gaze scrolling helplessly across a seemingly infinite night sky.   Wow,  I murmured.    Isn\'t it cool?  she asked, gazing upward with me.   Standing there in the middle of the wreckage of our house, staring at a sky that went on forever, I had to agree. My house is in semi-ruins and commuting is still tiring and scary and Luis just wrote me a letter about trying to sell my stories to a few magazines and I haven\'t gone dancing in MONTHS.  But there are still stars in the sky on a clear night, and yesterday I saw the clouds turn a pearlescent pink as the sun started setting, and freakchild still knows that a clear night sky in the midst of ruins is a cool thing.   So yeah. It\'s cool. I\'m cool.  It\'s weird, how my most peaceful moments are mostly shoehorned into the middle of very chaotic periods.  Via negativa , I suppose. When have I ever appreciated anything I\'ve gotten the easy way?   '),(25140,2238828,'06,July,2004','2004-07-06 00:00:00','    urlLink You GO, Lola Dinky.    Now that she is officially out, I suppose I can say it: Corazon \"Dinky\" Soliman, former secretary of the Department of Social Welfare and Development, is a relative of mine by marriage. That is, her husband is my grandmother\'s cousin - pretty distant by Western standards, but close enough here in the Philippines.   [ Note : I claim no close relation to the former Secretary Soliman - I may have met her at a family reunion or two before, but I was too young to remember. It is her husband who my sister and I remember fondly - he stayed at our place in Bangkok years back, and was great fun as a playmate. We teased him with the honorific \"Lolo\" (an affectionate way of saying \"Grandfather\" in Tagalog, most likely a shortened form of the Spanish  abuelo ), from which he recoiled  with horror. Technically - according to the family hierarchy, which the Diaz family matriarchs implement strictly at reunions - that  is  his title. However, in practical terms he\'s younger than my parents, so when we\'re just chilling, he\'s \"Uncle\".]  If I were related to her by blood, I\'d think that an aversion to Noli de Castro was genetic. As I\'m not, I\'ll assume that it\'s a matter of like minds. No matter how you slice the banana, though, she\'s pretty cool - from her blue-streaked hair and impassioned pleas for parents to stop making their kids perform sexually suggestive dances (what  is  up with that, anyway?), to this snarky dismissal of de Castro, who I personally think is a nitwit. According to  urlLink one account , she told media that she was stepping down from her position only because public service was \"bigger than [her] ego\". Hah! Straight to the gut.  I await her Tuesday presscon with bated breath.   '),(25141,2238828,'05,July,2004','2004-07-05 00:00:00','    Step right up for a game of  urlLink Rock, Paper, Saddam . Hilarious. Nicked from  urlLink Guav , who really should be on my links list.   Also:  urlLink When the going gets tough, the tough get Pinoys  - scroll down to read the story of a Filipina who got stranded in the middle of a desert at the height of the Gulf War. The sun was high, she was alone, no one knew or could tell her where the UN refugee centers were, and she had two suitcases. By golly, not only did she make it back, she had gifts for everyone, to boot.   Beat  that , if you can.   '),(25142,2238828,'05,July,2004','2004-07-05 00:00:00','    Woman,  urlLink  enough  already . There are quite a few reasons why I cannot connect the words \"hope\" and \"President Gloria Macapagal-Arroyo\", but my brain completely freezes up and threatens to shatter into icy fragments when I consider the thought of Arroyo governing the Philippines over and over and  bloody  over again in various guises.  urlLink This article , particularly Miriam Santiago\'s  urlLink veiled references to a possibly \"more authoritarian\" style of rule in the future , does not reassure me in the least.  If all goes according to plan, I will be able to give myself a break from reading any and all national newspapers by December. Yay for being politically aware and updated, but sometimes ignorance honestly  is  bliss.  - - -  Oh and hey, not to blow my own horn or anything, but I had no clue that the  urlLink restructuring of the National Steel Corporation  was such a big deal. When I was working back in KL, any and all paperwork related to this case went straight to me. Looking back, a lot of my work-related griping was about this case, because it was SO BLOODY FRUSTRATING. Hush-hush deals and the like are only exotic and exciting if you\'re on the other side of the fence. If you\'re in the know, it\'s just another paper to push, another round of calls to make, another bunch of people who will get ticked off by your pestering.  I still don\'t know what I\'m doing working in the government, so don\'t anybody ask.  - - -  And yeah,  urlLink I\'m not the only one  who\'s not impressed with GMA. Ten-point agenda, my prettily-shod foot.   '),(25143,2238828,'02,July,2004','2004-07-02 00:00:00','    You know how it is when you\'re busy with your life. Appointments, paperwork, phone calls, text messages, e-mails; work and family and friends colliding, intertwining, struggling for priority. Being several persona in the span of a few minutes, a stream of consciousness that shifts in and out of multiple thought processes. In this whirlwind, you exist solely in the present and perhaps a week or so in the future, but that is all. There is little time for anything else.   Then you hear a song that stops you dead in your tracks and pulls you back somewhere into your past, and you have no choice but to follow.  Today\'s song was  urlLink America , by Simon and Garfunkel. I was three when we moved to Bangkok - old enough to remember the Philippines. (I was going to say -  young enough to not be traumatized by the move , but it occurred to me that there is a gaping hole in my memories where that move should have been. My mother tells me that when we first moved there, we lived in a hotel for a while before moving to an apartment, but I have no recollection of that, or of our last few weeks in Manila, or of the plane ride to Bangkok. Which is strange, because I have vague memories of visiting Hong Kong when I was a year old. Oh, well - one day, I suppose, I\'ll ask her what happened.) We lived in a small flat, in a compound along Soi Rambutree - my father picked the place for its proximity to  urlLink his office  (this is the first time I\'ve seen a map of the place since I left, and all the names sound vaguely familiar still).   I have fond memories of that apartment, although it was easily one of the smallest and plainest places we\'ve ever lived in. It was the second storey of a duplex townhouse, designed so that one apartment occupied the entire first floor and the other occupied the entire second floor; you got to our place by climbing up a narrow, shadowy flight of stairs. The apartment was completely unfurnished when we got it, and remained, for the most part, unfurnished - my mother disliked it intensely, and refused to buy furniture in the hopes that it would motivate my father to look for a better place. In the meantime, we rattled around an apartment that had two tables, two chairs, a few shelves, and a closet. Note that I omitted mentioning beds - we didn\'t have any. My mother bought cheap mattresses, which we laid on the floor in the bedroom and slept on. In the mornings, my sister and I would sprawl around the living area in our underwear and read and draw while my mother did the chores.  Years later, my mother told me how lonely and frustrated she had been, those first few months. She was in her early thirties at the time, living in a small apartment with a husband who was never there and children too young to talk to her, in a country whose language she did not speak, far away from family and friends. There was no e-mail in those days, and our apartment had no telephone. It was a moot point in any case, since at that time she didn\'t have any friends in Bangkok, so there would have been no one to call.  In order to understand how truly mind-numbing that must have been, you have to know a little about my mother\'s history. She had been a sixth-placer in the national board exams for nursing; even before she graduated, private companies were sending her lucrative offers to come and work for them. She chose, instead, to go and work for the  urlLink Philippine General Hospital  so she could pay off her scholarship. After having worked off her scholarship, she applied for work as a nurse in the US and was accepted. She was 22 when she left for the US, young and pretty and fun-loving. We have pictures of her roaming across the US, and countless anecdotes from her travels there - watching Old Faithful in Yellowstone, eating lobster in Maine, picking oranges in Florida, hooting at boys as she drove past them in California, sleeping over at an aunt\'s place in Whidbey Island. Four years later my father proposed to her, and two years after that she came home for good, to marry him.   Imagining myself in that situation, all these years later, I asked her  How did you not go crazy?    I don\'t know,  she said, looking back at her younger self with amazement.  I really don\'t know.   The mornings were the emptiest times of the day. In the afternoons she would run errands or take us to school; in the evenings my father would come home. In the mornings, though, there was nothing much to do, and it was then that she would flip through her collection of cassettes and play us music. My mother used to love music; a few weeks ago my sister and I came across a stack of old vinyl records, which she had collected when she lived in Ohio. By the time we got to Bangkok the music medium of the day was the cassette tape. She wasn\'t able to bring much, but the tapes she brought define the soundtrack of my childhood: the ubiquitous  We Are the World , various albums of The Carpenters, the soundtrack of  Flashdance , Olivia Newton-John, Chopin, Mozart, some Sandi Patti, and Simon and Garfunkel.  My sister and I learned that we could tell our mother\'s moods by the music she played. Olivia Newton-John and  Flashdance , of course, were dance music; my mother used to be a dancer when she was younger and up until we were in grade school, she would put on music and dance to stay fit.  We Are the World , The Carpenters, and Simon and Garfunkel\'s cheerier songs were chore songs, which she played when she was trying to feel good about doing something she hated. When she was upset or irritated, it was classical or gospel music - Chopin when she was particularly frustrated.   When she was lonely, or when she missed home, it was Simon and Garfunkel. She would sort through photographs while  Sound of Silence  played in the background, write letters home to  Bridge over Troubled Waters , absently arrange and rearrange her belongings as Paul Simon crooned  America . Even more than the classical-music-times, we learned that these were the times when Mommy didn\'t really want to read stories or answer questions or blow bubbles off the balcony.   Today I heard the words  I\'m empty and aching and I don\'t know why , and it took me back there - to that first year in Bangkok, and the longing look in my mother\'s eyes, and the way my father would sit on the balcony with his shoulders slumped, smoking. I remember playing with my sister, who didn\'t know how to talk yet; going to school and not saying a word, because - funnily enough - I didn\'t know how to speak English; poking tentatively at the noodles my mother had just bought in the street because she didn\'t know how to cook and that was the only thing being sold in our area and if we didn\'t buy it then we wouldn\'t have dinner. I remember afternoons on our balcony, watching my father\'s hands as he pulled the thick skin off a pomelo, and thinking that he must be the strongest man in the world. I remember clutching to my mother\'s dress as the little girl who lived next to us called out to me in Thai, asking something, and how horribly embarrassed I was because I didn\'t know what she was saying.  Say  pai tiaw, kha ,  my mother prompted softly.  She\'s asking you where you\'re going.  [ Note :  Pai tiaw  means, roughly, that you\'re going out; it\'s what you say when you know you\'re going somewhere, but you haven\'t really decided where you\'re going to go yet. Kind of like saying  Kami lepak  in Bahasa, or  labas lang kami  in Filipino.]  I remember, I remember, I remember. There will never be enough words for a memory.   '),(25144,2238828,'05,August,2004','2004-08-05 00:00:00','    Funny how I can be so fluent when I\'m here and so incoherent in real life. How I can never say the things that I want to say properly, and they come out wrong, in tangles instead of straight clean lines.  How can I write so well and talk so badly?  Sometimes when I\'m saying things I feel like I\'m piling heavy rocks on top of a crumbling foundation, and everything inside me holds still waiting for it to fall. Why can\'t I just say things, the way everybody else does? Why do I always end up clutching frantically at fraying ends, trying to pull them together so that they hold   why why why.  Times like this I miss W, because he always made me feel as if he saw me clear through, no matter how mud-spattered and streaked I felt. Like I was translucent in his eyes, no matter how jumbled I was.   I\'m so, so,  so  very tired, and I don\'t want to go home today.   '),(25145,2238828,'03,August,2004','2004-08-03 00:00:00','    I think I might be sick. Not very very sick, but at least hovering miserably on the border, which is bad enough when your workplace is FORTY KILOMETERS AWAY from your house and you haven\'t brought a car. (Okay, even if you  did  bring a car. Because seriously, Makati traffic.)  I also do not think I will be able to take tomorrow off, because J - the workmate who I partner with for a couple of projects inside the office - is out sick, too, and she may not come in tomorrow either in which case I will have to be here. Wow, life sucks, what a revelation.  I also think I\'ve gained weight again. I am 130 pounds - this is what I saw on the scale in the split second I actually stepped on it - and that does not bode well, because this coming Sunday I am getting fitted for my bridesmaid dress for my cousin\'s wedding. That gives me - let\'s see - four days to lose at least ten pounds.   I am also thinking in a very panicked manner of the various projects on my desk that I have to complete NOW NOW NOW and yet my brain is not cooperating. What would be the most professional way of dealing with this predicament?  I think this will be the best week ever. (If you didn\'t hear the sarcasm there, YOU ARE DEAF.)   Woe .   '),(25146,4178311,'22,August,2004','2004-08-22 00:00:00','    Jun & Jen were a pair of young couple...Jun was veri charming &amp; many girls could`t helpfalling in love with him... therefore, he became quite a flirt.As for Jen, she was an ordinary girl with average looks. Although there were quite a number of guys chasing her, jenonly had Jun in her heart.altought She knew that one day Jun would eventually leave her ..she still continued to love him whole heartedly...Jen loves rainy days...she loves playing in the rain, & whenever Jun wanted to join her, she\'ll always stop him from joinin.Jun then ask her; Why don\'t you let me join you?Jen replied;Err..er..because i don\'t want you to fall sick.Jun then ask her again;If playing in the rain will make me fall sick, then why are you still doing it?But each time Jen will always keep quiet & smile at him...Although Jun was`t able to join her, he felt happy. To him, as long as Jen was happy, he\'ll be happy as well. :)However, nothin last long... Jun soon fell for another girl after two months. He even love the girl more deeply den he love jen One day, while Jun &amp; Jen were having their dinner , Jun told Jen that this would be their last dinnertogether... Jen looked at him. She knew dat this was going to happen, but she choose not to ask any farther...Jun then requested for a break up.. And Jen...she accepted willingly... partly because, she knew...Jun was just like the wind...never stop at any point.That night..it was to be the last time Jun send Jen home.. Jen kept veri quiet...although deep down she wanted to know the reasons but she choose not to ask. Just when Jen was about to step into the lift, Jun stop her. He said:Jen, I\'m really sorry that i\'ve let you down. But i\'ll never forget those days when i see you playing in the rain...those are the most happy &amp; unforgottable memories i had with you After listening to Jun, Jen could\'t help it but cried..Jun hugged her tightly. He said:Jen, There\'s one question i wanted to ask you long ago... Why is it everytime when you are playing in the rain, you just would`t allow me to accompany you?Jen was stunned..pausing for a few seconds,Jen replied...Because...Because i don\'t wish to let you know...I was crying..*~~l3aF dÄ”Ã¾Ã¢Ð¯Ð¢Å¯Ò‘Ð„ ÏŠÂ§ ÑŠÎ¾ÄˆÅ‘z Î©f Ð Ï‰Î¯Î·ÄŽ ÃžÃ›Ñ“Â§Ã¼Ä­T..âˆžÅÐƒ Ð²ÎˆÄ‹Ã´z Ð— Å§Ð¯Ð·Ð· dÎ¯dÅƒ\'Å£ Ã¥ÅœÐš hÎ¾Ò 2 Â§Å§Ä‚ÐŽ....~* '),(25147,4178311,'21,August,2004','2004-08-21 00:00:00','    mM..~tOday fEelIn\' bEtTa..~bUt nOe thAt i aM sTill mIsSiN9 sOmeoNe wHo dOn dEsErvE my mIZZin9..~iTs oKAy..~lEt tIME pAss thEN i wiLl bE okay lE..~todaY gOing to bu9IS loR..~mAybe 9oin9 tO sEe mY hP..~buT mOst pRobAly alSo gOne cAse dE la..~sO..hMm..~dOn rEalli cArri mUch hOpe toDay..~1 wEek anD 6 dAys mORe tO n levEl liAo..~jIAyou..~!!!!kAmpatEh nEh..~!i hOpe thAt pErson wHo wRotE mY cOmMent ytd wiLl tEll mi wHo hE or sHE iS..~rEalli wISh to mAke fRen wIth thAt pErsoN..~aNoNymOus..~tEll mi wHo u aRe kKz?iF u aRe sOmebOdi inOe..~pLz tEll mI wHo u aRe oKoK..?don lie tO mi..~i hAte liArz*~.... Ð²ÎˆÄ‹Ã´z Ð— Å§Ð¯Ð·Ð· dÎ¯dÅƒ\'Å£ Ã¥ÅœÐš hÎ¾Ò 2 Â§Å§Ä‚ÐŽ....~* '),(25148,4178311,'21,August,2004','2004-08-21 00:00:00','    i cANt sToP thInK in\' aBt hIm..~muZ i rEalli gIVe uP..?mAybE givIng uP iS thE bEst wAy..~i mUZ lAugh aND sMile iN fRoNt oF my fRen aND fAmili..~*vEri xIn kU..~*hAIx..i Am nOt hAPpi..wO bU kUai lE..bUt i mUz sTill smile..~yTd..wHen i sLp..~i fEel sufFocated aGain..~i nEarli dIEd..~i wAke uP aND wEnt to my parent rOom to slp..~wAD hApPen tO mi..i aLSo donnO..~i jUz nOe i nEarli lOst my liFe yTd..~sEc tiME liAo..~iZ tHEre a thIrd tiMe>?i dOnno..~*~~l3aF dÄ”Ã¾Ã¢Ð¯Ð¢Å¯Ò‘Ð„ ÏŠÂ§ ÑŠÎ¾ÄˆÅ‘z Î©f Ð Ï‰Î¯Î·ÄŽ ÃžÃ›Ñ“Â§Ã¼Ä­T..âˆžÅÐƒ Ð²ÎˆÄ‹Ã´z Ð— Å§Ð¯Ð·Ð· dÎ¯dÅƒ\'Å£ Ã¥ÅœÐš hÎ¾Ò 2 Â§Å§Ä‚ÐŽ....~* '),(25149,4178311,'20,August,2004','2004-08-20 00:00:00','    i aM bEri sAD tODay aFtEr rEadIn\' sMth i sHouldN\'t hAve rEad..~tHose wORds jUZ sEems lIKe a bIg bIg kNifE tHAt sTab rIte dOwn tO mY hEart(\"V\")~nOW thAT i waS typIn\'~i wAS cRyin\'..~is sHoulDn\'t liKE hIm dE..~i nOE..~!!!!!!!!!!!!!!!!!!!i rEalli ha+e mYs3lF aLOt fOR thIs..~i kNew bUt i sTiLL dId..~i fEel sO cHeAp sUdDenli..~sO chEap..~wHY dO i liKE hiM..~~?iF 1 dAy i diE..wiLL hE eVa kAre tO dR0p a tEar fOR mi..~?tOday wHile i wAs hAVin mY nAP ..~i fElt sO sUfFocAted..~i cANt bReathE..~i jUZ hoPE i cAN dIE..~dIE lE bEtta..no mORe pAIn***(\"V\")pAIn***wHo cAN eAse iT..~?*~.. Ð²ÎˆÄ‹Ã´z Ð— Å§Ð¯Ð·Ð· dÎ¯dÅƒ\'Å£ Ã¥ÅœÐš hÎ¾Ò 2 Â§Å§Ä‚ÐŽ....~* '),(25150,4178311,'19,August,2004','2004-08-19 00:00:00','    toDay i fEel sO rElax lE..~i fInaLLi hANd uP mY aRt wORk..i leaVe sCh aT aBt 18 50..wOw..sO tiRed..dIN eaT lunCh alsO..i dO my aRt iN scHooL fRm 1230 tO thAt tiME..~v3h tIred..dOnnO wHat hAppen tO mY hEad oR rAthEr bRain..~wHEn dOin\' mY cHi pAPeR..~mY heAd wAs fEelin\' v3h wEird leh..~liKE kEep beAtin\'..~pAIn pAIn alSO..~bRain cANcer..?dONno lA..dOnno wHy toDay gOt a kINd oF \'fAstEr dIE bEtta\' dE fEelin\'~nBm lA..mM..~!mISsIN\' hiM buT hE wAs mISsIn\' anOthEr pErson..~*sObBi3*~~l3aF dÄ”Ã¾Ã¢Ð¯Ð¢Å¯Ò‘Ð„ ÏŠÂ§ ÑŠÎ¾ÄˆÅ‘z Î©f Ð Ï‰Î¯Î·ÄŽ ÃžÃ›Ñ“Â§Ã¼Ä­T..âˆžÅÐƒ Ð²ÎˆÄ‹Ã´z Ð— Å§Ð¯Ð·Ð· dÎ¯dÅƒ\'Å£ Ã¥ÅœÐš hÎ¾Ò 2 Â§Å§Ä‚ÐŽ....~* '),(25151,4178311,'18,August,2004','2004-08-18 00:00:00','    *sObBie..~!!tOdAy tWo pApEr gOne cAse agAiN lE..sO sAD..~mAthS pAPer leAve sO mAnI bLAnks..~!cHem nO bLAnk bUt aLl gu3ss gu3ss dE..~dIe lE la......~tmL gOt cHineSe pApeR 1&2..~hOpe thAt aLL pPl wiLl fAre wEll fOR thIs paPEr bAHX..iNclUdin\' thOSe sChoOls sTudentS tAkin thE sAmE pApeR aS uS..=))) kAmpAtEh n3h..~!~*..~ Ð²ÎˆÄ‹Ã´z Ð— Å§Ð¯Ð·Ð· dÎ¯dÅƒ\'Å£ Ã¥ÅœÐš hÎ¾Ò 2 Â§Å§Ä‚ÐŽ....~* '),(25152,4178311,'17,August,2004','2004-08-17 00:00:00','    mM..tOdAY thE mAths pApEr hOR..i soMEhOW gOnNA gONe cASe aGaIn..i mAke sO maNi miStaKEs..~!!!!hAIx..~cANnot bLAm3 lA hOr?i sItOpid mA..~tODay aRt..mY fINal pIeCe likE wOn hElp mI sCoRe mUCh..bUt mdM sAi k33p noddIng hER hEad wHEn sHE sEees mY wORkpiEcE..hAix..dONo wIll gEt gOOd gRade oR fLunk..=(..*sObtMl cHEm aNd maThs pApEr tWo..~tHURs mUZ hANd iN thE aRt cOuRsEwOrk liAo..dONno wAD mDm sAI wiLl sAy abT iT..******~~l3aF dÄ”Ã¾Ã¢Ð¯Ð¢Å¯Ò‘Ð„ ÏŠÂ§ ÑŠÎ¾ÄˆÅ‘z Î©f Ð Ï‰Î¯Î·ÄŽ ÃžÃ›Ñ“Â§Ã¼Ä­T..âˆžÅÐƒ Ð²ÎˆÄ‹Ã´z Ð— Å§Ð¯Ð·Ð· dÎ¯dÅƒ\'Å£ Ã¥ÅœÐš hÎ¾Ò 2 Â§Å§Ä‚ÐŽ....~* '),(25153,4178311,'16,August,2004','2004-08-16 00:00:00','    hMm..tOdAY pHysICs pApEr moRe eASier thAN i eXp3cTed..~!hEe..~y3@h..~!!v3h  hAppI yEh..~tHen mY aRt tEacHEr tODay hOr..~wAh..i tEll thE aRt pAPer tmL i jUZ dO tiLl miNdmAp..~bUt sHE dIn sCold mI wOR..~!!!sO sUrpRise anD hAppi..~!lolZZ..~!!!!!bUt fINd nO oNe to shaRe thIS jOy wId mI..~hAIx..nBm la..i nOe i hApPi cAn le..~^(^O^)^pRepARin\' tO fAil my aRt tMl..~*~~ Ð²ÎˆÄ‹Ã´z Ð— Å§Ð¯Ð·Ð· dÎ¯dÅƒ\'Å£ Ã¥ÅœÐš hÎ¾Ò 2 Â§Å§Ä‚ÐŽ....~* '),(25154,4178311,'15,August,2004','2004-08-15 00:00:00','    yEah..~!!i fINish mY cANvAs liAO..~!!rEleASe soME sTr3ss lE..KOR aR..jiAyou hOR..~~!!kKz..??u toDay sO guAI nO onLIne..haRdwROin wOr..~hEe...thNE nOW nuZ sTArt mY pReliM aRt lE thNE lATa mUZ dO soME rEvisIon..~*ph3w*~~l3aF dÄ”Ã¾Ã¢Ð¯Ð¢Å¯Ò‘Ð„ ÏŠÂ§ ÑŠÎ¾ÄˆÅ‘z Î©f Ð Ï‰Î¯Î·ÄŽ ÃžÃ›Ñ“Â§Ã¼Ä­T..âˆžÅÐƒ Ð²ÎˆÄ‹Ã´z Ð— Å§Ð¯Ð·Ð· dÎ¯dÅƒ\'Å£ Ã¥ÅœÐš hÎ¾Ò 2 Â§Å§Ä‚ÐŽ....~* '),(25155,4178311,'14,August,2004','2004-08-14 00:00:00','    toDaY mI nOT hApPi..~miSsiN\' sOMe1 i sHoulDn\'t bE miZing..~sItOpiD hOR..?hAIx..~toDay nEed tO rUsh aRt..thEN rEvIse pHysIcs..~theN gOT tImE mUZ dO pRelim aRt lE..mY kOr aLSo aRt dE..yOngcHoOn alSo mUz jiAyou..~!!=)dON sTrEsS liKE mI lA..silLi bOi..~!!wAhAHA..~!!hMm..nTH lE bAHx..~hOpE mY fRen wiLl aLL rEalli gEt gOod goOD gRades fOr thEir pReliM aNd aLsO n lEvel woR..~!!=.=\" *~.. Ð²ÎˆÄ‹Ã´z Ð— Å§Ð¯Ð·Ð· dÎ¯dÅƒ\'Å£ Ã¥ÅœÐš hÎ¾Ò 2 Â§Å§Ä‚ÐŽ....~* '),(25156,4178311,'14,August,2004','2004-08-14 00:00:00','    todAy b3ri tir3d..~!*yAwnzz..~8a.m wak3 uP le..~toDay neeD sTudi mY pHysic..~thEN dO aRt aGain..~oH gOd..~!!l3t mI dIe la..sO tirEd..~*...~ Ð²ÎˆÄ‹Ã´z Ð— Å§Ð¯Ð·Ð· dÎ¯dÅƒ\'Å£ Ã¥ÅœÐš hÎ¾Ò 2 Â§Å§Ä‚ÐŽ....~* Ð²ÎˆÄ‹Ã´z Ð— Å§Ð¯Ð·Ð· dÎ¯dÅƒ\'Å£ Ã¥ÅœÐš hÎ¾Ò 2 Â§Å§Ä‚ÐŽ....~* '),(25157,4178311,'13,August,2004','2004-08-13 00:00:00','    tOdAy dE eNgliSh gONe cAse liAo..wAHha..~!!whEN dOin thAt pAper..n3arli fAll aslP..~lolzZZZ..~!thNE thE mAp reADin fOR gEog..fIrsT 3 qNS oK...coz 1 mark eACh..then oTher 7 maRks aLSo goNe cAse..~wAHHA..~!aI s3h hOr..?hAIx..aniwaE i aM alSO tOo tIRed to gIb a dAMn lE..BEri tiRed..~k bAHX..aHdiOs..~!^(^O^)^*~~l3aF dÄ”Ã¾Ã¢Ð¯Ð¢Å¯Ò‘Ð„ ÏŠÂ§ ÑŠÎ¾ÄˆÅ‘z Î©f Ð Ï‰Î¯Î·ÄŽ ÃžÃ›Ñ“Â§Ã¼Ä­T..âˆžÅÐƒ Ð²ÎˆÄ‹Ã´z Ð— Å§Ð¯Ð·Ð· dÎ¯dÅƒ\'Å£ Ã¥ÅœÐš hÎ¾Ò 2 Â§Å§Ä‚ÐŽ....~* '),(25158,4178311,'12,August,2004','2004-08-12 00:00:00','    hAIx..~!!i cANT fInisH mY sS pApEr tODay..mUmmi aSk hOWs thE paPeR thEN i dOn dARe to sAy..jUZ sAY okokokokook..~!!!i aM sO sAd..cANt fInish mY pAper..~tHen nTh muCh hAPpen tOdAY bAHX..~nEEd to sTudI fOR mY g3og eXam tMl lE..wISh eVeRyoNE wiLl fInish dOin\' thEir pApeR on tiMe tmL..~=(* Ð²ÎˆÄ‹Ã´z Ð— Å§Ð¯Ð·Ð· dÎ¯dÅƒ\'Å£ Ã¥ÅœÐš hÎ¾Ò 2 Â§Å§Ä‚ÐŽ....~* '),(25159,4178311,'11,August,2004','2004-08-11 00:00:00','    hAIz..toDAy BEri BERi pr3sSuRixe lEh..nEed tO dO aloT oF thIngy..sTudy sTudy stUdy..~!!!!wAh..~!gOin\' tO diE liAo..~haRdli tAKe sUm tIMe oUt tO fiLl iN trhIs blOG yEh..~tOdAy hOr..sUm oF mY fRen tReaT mI vErI cOld..dOnnO wHy lEh..~hAix..~i nO tiEM to d3aL wif tHem yEt lA..nO tiMe nO tiME..~!!!sOlli yEH..i reAlli kARe aBt ouR fREshiP bUT i reALli nO tiEM toDay..~hOpe u aLl uNdeRsTAnD..~=(...nZt tuES nEed to hAnD iN mY aRt lE..pRelim oNE..bUt i sTiLl sTuck aT mINdMAp...~!!!!oH mY gOd..~!!!wAD aM i to do..?*sOb..~tMl gOT sOciAl stUdiEs eXam lE..~goIn\' bACk to sTudy liAO..~*sObBy gURlz isH mi toDay..~*~~l3aF dÄ”Ã¾Ã¢Ð¯Ð¢Å¯Ò‘Ð„ ÏŠÂ§ ÑŠÎ¾ÄˆÅ‘z Î©f Ð Ï‰Î¯Î·ÄŽ ÃžÃ›Ñ“Â§Ã¼Ä­T..âˆžÅÐƒ Ð²ÎˆÄ‹Ã´z Ð— Å§Ð¯Ð·Ð· dÎ¯dÅƒ\'Å£ Ã¥ÅœÐš hÎ¾Ò 2 Â§Å§Ä‚ÐŽ....~* '),(25160,4178311,'10,August,2004','2004-08-10 00:00:00','    toDay quITe hApPi whEN i sEe my bLog b3iNG sEt uP wIF thE hElp oF my \'kOr\'...~hEe..aCtUalli i fINd mYseLf quItE fOrtuNaTe tO hAb a kORkOr liKE hIm..hEe..bUt hE trEat mI gOod iSh yIng gAI dE mA..~lolZXXX..~!!!toDay i nEed to rUSh my aRt..~sTudY  fOr mY sOciAl sTudIEs..~!*pR3sSur3*wHAHa..~!!!bUt dOnnO wHy i sTiLl cAN laUgh leh..~lolZXXX..~~!!bUt thIS tiMe muZ b3 sErioUs lE la..cOZ i nOE i cAnt aFforD tO flUnk my pRelIm..~2 mORe dAys tO my pReliM..~wISh thAT kOr yOngchoOn wiLl fAre gOOd gOOd fOR hIs pReliM..~!!aND hOpe thAt aLL mY fRen wIll alSo gET gOOd goOD graDes fOR theIr pReliM thIS timE..~!!^(^O^)^* Ð²ÎˆÄ‹Ã´z Ð— Å§Ð¯Ð·Ð· dÎ¯dÅƒ\'Å£ Ã¥ÅœÐš hÎ¾Ò 2 Â§Å§Ä‚ÐŽ....~* '),(25161,4178311,'09,August,2004','2004-08-09 00:00:00','    hmm..today feelin\' donno how to say wor..quite kind of sad..donno why..need to do art..alot of thingy awaiting mi liao..my things are now crush tog..so pressure..bUt nBm..i mUz bE sTrong..~!!cOZ tHEre aRe mORe pPl wHJo reAlli nEed mORe kARe BAHx..~hMm..~!!!hAO le..laTer sEe cAN updaTe mORe thIng mA..~wISh aLl pPl aRoUnd mi caN BE hAppi woR..~!!^(^O^)^ *l3aF d3pArtUre iS b3coZ oF tHe wInd pUrsUit..~oR b3coZ tHe tR33 dIdN\'t aSk h3r tO sTaY..~?* '),(25162,4284264,'23,August,2004','2004-08-23 00:00:00','   No preachy thoughts today. I\'m tired, but I have had my favorite joke on my brain all day so I thought I\'d share it. This joke does not fall into the category of crude, at least I hope it doesn\'t. It\'s a joke dependent upon your ability to read with an Irish accent... basically trill your \"r\'s\" and vary your inflection and you\'re on your way. Here goes.  An older Irish gentleman stepped off of the ship and placed his foot on American soil for the first time. It was early summer many years ago, and as the ship sailed into the harbor our hero promised himself that he would, on his first day in his new home, do something uniquely American. As he walked the streets of New York, he came upon an amazing structure... it read \"Yankee Stadium\". A smile crossed his lips, and he had found the days activity. He bought his ticket and followed the crowd in to their seats.  As he sat in the crowd, it dawned on him that he had no clue how this game was played... he had never even seen it, only heard about it. The solution came to him quickly - watch the folks around you and do as they do. After all, \"When in Rome\". The first batter stepped to the plate, the pitcher wound up and released the ball... \"CRACK\"! The crowd leapt to their feet cheering the runner toward first. Our hero followed suit - \"Run, run ye varmint! Run will ye?!?!?\" A few of his neighbors snickered at his exuberance, but all in all his first game was going well. The next batter stepped up, and after 3 or 4 pitches the sound of the ball and bat colliding once again drove the crowd to their feet. \"Run!!! Run ye varmint! Run will ye?!\" The Irishman was having the time of his life... his fellow onlookers had caught some of his enthusiasm and seemed to enjoy the game all the more, partly because he was there.  Then the third batter stepped to the plate. Pitch after pitch crossed before him, until finally the umpire signaled him to take his base. The crowd applauded politely as the man began a half trot, half walk towards first... all of the crowd, that is, except our hero. He was dumbfounded. He leapt to his feet - \"RUN!!! RUN YE VARMINT!!!\". Snickers from the crowd, and a look of bewilderment on the Irishman\'s face. He slumped into his seat, trying to sort out what had just happened when a kindly neighbor leaned over to explain the situation. \"He doesn\'t have to run,\" the kindly man said, \"he\'s got four balls.\" The Irishman\'s gaze fell away from the fellow fan to the field, he rose to his feet and uttered these simple words: \"Walk with pride, man. Walk with pride.\"  There you have it... my favorite joke ever. I cried the first time I read it... but somehow my version is always considerably longer. At least my embellishments in this version left out the smell of hotdogs and freshly cut grass. This has brought a smile to my face all day - not sure what made me think of it, but I\'m thankful for it today.  Grace &amp; Peace,  PJ  '),(25163,4284264,'22,August,2004','2004-08-22 00:00:00','   I\'ve spent the better part of my adult life convinced that I was supposed to be in worship ministry, or at very least in some role that made a sizeable impact on the Kingdom.  That sense of calling has been my life preserver at times, but has at times been the stone dragging me at break-neck speed towards the bottom of the lake.   Without going into great detail on this, I am in a hard season with some of God\'s children... I feel like I\'ve been beaten up, left and then beaten again.  I ended up in this situation because of my compelling need to make sweeping contributions to the Kingdom - this would be a stone situation.  I felt I could bring the gifts and talents given me by God into a notoriously hard working environment and change everything and everyone there for the good of the Kingdom.  I got in there, worked as hard as I could and tried to serve to the best of my ability; I can honestly say that I gave my best out of what I thought were Holy intentions, but here I sit with my feelings hurt and my reputation on the line.  What went wrong?  Re-read that last paragraph and notice how many \"my\'s\" and \"I\'s\" are there.   I  thought that  I  could effect change...  my  calling was to take  my  gifts and talents.  I\'m finding the link between when the notion of God\'s call and annointing over me serves as life-preserver or stone has everything to do with... well... me.  Specifically, I am finding that when I go into opportunities with my own great plans and agendas things quickly begin crumbling around me.   Here\'s the challenge:  I want to be a part of the body of Christ but I have always approached my life as though it were my calling to be the twinkle in the eye of Christ.  I am naturally given to arrogance, and I confess that here for all to read, but this belief was , at least in my mind, Holy.  I was supposed to be the tool with which God would bring sweeping change and revival to His people!  As it stands now, I am not God\'s secret weapon... what a painful discovery.  What I am learning through this time is simply this:  if all I am ever given to serve as is a callous on the foot of Christ\'s presence in this world then I am blessed.  I don\'t have to be the ring finger or the sparkle in the eye to be a part that is necessary and beautiful.  God knows me - he knows what he made me to be and that is enough.  I can\'t make myself more than I was created to be, but no one can take my place in the body away from me.  Grace & Peace,  PJ  '),(25164,4284264,'21,August,2004','2004-08-21 00:00:00','   Refresh my memory - could Jeckyl remember the atrocities of Hyde? In the spirit of confession, I never quite got around to reading this classic in school... Nor do I remember reading the Cliff\'s notes. I would say that I relied on the movie, but I never remember getting around to watching it either... come to think of it, how did I ever graduate? I digress. There\'s a theme in these horror classics in which a transformation takes place (Werewolf, Jeckyl & Hyde, the Hulk... Michael Jackson) where a \"what have I done\" moment inevitably follows \"the change\".  I can relate to the monsters in those stories a little bit... okay, a lot bit. I am one of those guys who can be just fine one minute and then something is said or takes place that absolutely sets me off. Contrast this with Jack (my son). The boy can be smiling and giggling up a storm one minute and them BAM! (all props to Emeril), the boy is screaming like mad. Fast forward ten minutes, though, and big boy is asleep or smiling again with no recognizable sign that he was ever upset. I always have the \"What did I say or do?\" moments that follow these outbursts, but Jack is ready to be loved on and played with again almost immediately. So here\'s the question: when does the change take place and why?  Hypothesis: Jack screams because he has a need that he can\'t communicate in any other way. Once the need is met, the time of need is past and it\'s playtime. When I go off the deep end, it\'s more often than not overflow - too many annoyances and challenges have built up and I\'m off. If I\'m not careful this blog could get crazy long, so I\'ll go straight to the point: We are called to be like children, and that has a great many implications, but this one is new to me. The difference between me and my boy is this: he recognizes need as need and I bury my needs under a stack of wants and \"I deserves\". Jack\'s day is made by being loved, being fed and being sheltered, whereas mine is ruined if I don\'t get all of those things  plus  affirmation from my boss, my wife\'s undivided attention, the blessings from God I so richly deserve... the list goes on.  What would my life look like if I was able to approach my day with the knowledge that before my feet ever touch the floor my needs have been met? What if, instead of living and dying from sale to sale (or failure to failure) I rested in the promises of God and took the time to watch the birds of the air and notice the flowers of the field? This is  not  to say that I need not work and \"provide\" for my family, rather to say that I should recognize that my NEEDS are met. You can see how this could get lengthy, so I\'ll leave you to stew on that instead of trying to say it in ten other ways, as is my want (certainly not your need, however).  I don\'t expect that I will master this any time soon, but I do believe that there is a lesson to be learned here. HIS divine power has given us everything we need for life... for I know the plans I have for you says the Lord... you have been given a spirit not of slavehood, but of children - and THAT IS WHAT WE ARE - children of the most high God. I love the fact that my boy gets it, and I pray that he will help me recognize my needs under that pile of rubble the world tells me I \"need\". Cue Fernando Ortega\'s \"Give Me Jesus\"... I\'m out.  Grace &amp; Peace~  PJ  '),(25165,4284264,'20,August,2004','2004-08-20 00:00:00','   Okay... maybe not Waterloo, but it was close.  Picture it (Golden Girls reference there... Sophia is hysterical): Walgreens on a sticky Friday evening in August. A beautiful woman comes bounding down the toy isle toward an unsuspecting new father as he struggles to hold the child/car seat combo (see \"ton of bricks\") off of the nasty drug store floor. In her hands she carries them... the downfall of many a father\'s dreams, the scourge of the never-to-be future quarterback... the Care Bear. Ladies and gentlemen, I confess to you that my son is the owner of a Care Bear - Good Luck Bear as I am told. He has that one because there is a big shamrock on the front. If my boy is to grow up confused I can pray that he thinks we are Notre dame fans! Alas, I was defeated. You will be happy to know, however, that when Mama started talking about busting out the little boy Cabbage Patch demons the foot came down. Negative ghost rider, the somewhat-less-than-manly toy pattern is full.  Hallelujah for the weekend. Jack can holler all he wants tonight, because I can sleep until the Spirit moves me in the morning. If, however, he were to sleep all night but wake up at 6:00 a.m. I would be bitter... I don\'t want to see the clock until at least 9:00.  I have no profound wisdom for the blog today... some would say that is nothing new. I tried hard to see my world afresh today and had a few victories but many defeats. I just try to keep reminding myself to lift up my eyes. Strange that we are called to look away from the next step and lift our hands away from what \"holds us down\". Lay everything at His feet, turn your eyes toward Jesus and... well, when I develop the faith to try it for myself I\'ll tell you all about it. As a friend often says, you can\'t lead where you have not yet been. The first few steps down that road have been interesting - thanks for joining me on the journey.  Grace &amp; Peace,  PJ  '),(25166,4284264,'19,August,2004','2004-08-19 00:00:00','   Have you ever seen a baby pinecone? I was at Cumberland University today, trying like the dickens to sell some phones so that I can buy fun stuff for our 17-week old... and maybe a little sump\'in sump\'in for me (ncaa 2005 on the ps2). I was walking out with my boss after a so-so meeting and there it was - a tree full of baby pinecones. Who knew? I remember playing with them as a kid, my mom sitting them beside the fireplace... even kicking them around the yard (a poor man\'s soccer ball, if you will). Never, though, do I remember seeing baby pinecones.  Baby pinecones are plenty cool, but they are nothing compared to my little man. As I said, he is seventeen weeks old and is changing every day. His smiles are radiant, his eyes are as big and bright as the sky and his laugh is infectious. He is the coolest thing in the history of ever, period. Here\'s my fear, though: It has taken me the better part of seven years of marriage to figure out how amazing and beautiful my wife is, and I am learning that more every day. How can I, working 40+ hours a week, stressing over sales numbers and bills, insure that I will not look up years from now and be shocked to find my little man... well... a man? How can I not glance up from \"keeping my head down\" or \"my nose to the grindstone\" years from now and not find that I missed everything?  Final part of this thought: the same thing has happened in my walk with God. I lived 20+ years of life singing songs and attending services, all the while missing the bright, shining face of my Lord looking down on me - His child... his chosen. I was so busy watching for His hands to move that I never stopped to look up into His eyes. I love it when Jack looks at me and his face lights up in a huge smile. I love coming home in the afternoons and seeing my wife at the back door, babe in arms waiting for me to come in and give her a \"hoogle\" (better known as a hug). I love the notion of the God of creation singing over me and smiling as he watches me rock my son. My prayer is that my eyes will seek out the precious things God has put in my life and take them in fully.  I kind of like the fact that at 27 I am amazed at the sight of a baby pinecone... it makes me wonder what new sight awaits me tomorrow.  '),(25167,950122,'21,March,2003','2003-03-21 00:00:00','   Hi! This is my new blog... I decided to move my blog here after I realized that I wouldn\'t be able to have any archives whatsoever with Geocities. So I\'m making my Geocities site my personal site and this my blog! '),(25168,950122,'29,April,2003','2003-04-29 00:00:00','   Oh yea, I\'ll be thinking of some names too! n_n IMO the transfer student\'s name sounds better as Ashley. n_n But that\'s just my opinion. '),(25169,950122,'27,April,2003','2003-04-27 00:00:00','   To Speedo-kun:  Yup! n_n Tell me the addies to those yaoi doujinshi sometime, alright? :)  urlLink Here  is a page with some scanlations of Yami no Matsuei. n_n It only goes up to volume three tho. But it\'s still worth checking out.    Well! I\'ve changed my layout once again... I really like this one. n_n I think I\'ll stick with this for a while. It\'s even my fav. color! And I don\'t care if it\'s a dead color mom! >    Anyway, nothing much has happened lately. Jeth\'s now trying out his new disposable contacts. n_n;; And I quote:  My eyes!!!!  Haha, it\'s funny. n_n;;    Ergh, I\'m a bit braindead at the moment.. I\'ve been doing a lot since this morning. > '),(25170,950122,'27,April,2003','2003-04-27 00:00:00','   I\'ve read the first three volumes of Yami no Matsuei, and I\'m hooked! n_n I like Hijiri for some reason. Oh yeah, I remember getting him on a quiz. Tsuzuki and Hisoka are priceless! And Muraki... Muraki interests me. n_n Seriously!     To Speedo-kun:  I think you should get a Yahoo Messenger. That way, if we\'re both online, we can chat! n_n You don\'t need a yahoo e-mail adress to sign or anything either! n_n '),(25171,950122,'24,April,2003','2003-04-24 00:00:00','   Argh... My mom, dad and brother are going to Virginia,  and are leaving me behind!  Argh. It\'s so shitty! Really! >P They\'re just going to leave me with my aunt. It\'s so mean!    I accidentaly cut my finger with a razor blade early this morning. It\'s pretty ironic, because my dad had just told me to be careful, then I cut myself. n_n;; I didn\'t tell him I cut myself though; he would probably freak out and suffer a panic attack. XD Anyway, it was sorta weird because it just surprised me. It didn\'t hurt much, but there was a lot of blood! >    We went to Glorietta to watch Johnny English, but ended up just walking around. My mom bought me boots! n_n I also bought some books. I saw this one book;  A Guide to Tolkien . I wanted to get it, but I decided against it. I need to save like everyone else! > '),(25172,950122,'23,April,2003','2003-04-23 00:00:00','   I watched Animal Planet this morning, and the show was about animal fathers. Gosh! I was amazed at how quickly I realized that some males really suck. Can you believe that male seals actually  rape  female seals just for the sex?! Sheesh! It made me scream out one of my new mottos:     I wish men were more like sea horses!     Funny, but true. n_n '),(25173,950122,'21,April,2003','2003-04-21 00:00:00','   Ahhh... I like my Winamp 3 Player... n_n I just got a Nataku [X] skin.X3 '),(25174,950122,'21,April,2003','2003-04-21 00:00:00','   I have been flooded with surveys!! X_X I could only answer one, and I had to delete the others... I wasn\'t recieving any other mail! n_n '),(25175,950122,'21,April,2003','2003-04-21 00:00:00','   New World  by  Round Table  has found a place in my head. It\'s a cute song! n_n    I\'ve watched the first two episodes of Wolf\'s Rain. I loved it! The ending song,  Gravity  sung by  Maaya Sakamoto , is beautiful. I\'m completely in love with the song. '),(25176,950122,'14,April,2003','2003-04-14 00:00:00','   By the way, the underwear that Duo has on his head in the pic is Heero\'s. n_n;; '),(25177,950122,'12,April,2003','2003-04-12 00:00:00','   I think I\'m neither seme nor uke. n_n;; The thought just popped into my mind. '),(25178,950122,'12,April,2003','2003-04-12 00:00:00','   Hay... it\'s 2:00 a.m... I have to stay up untill we leave for Iriga so I can sleep on the way there.n_n;;    I\'ve really gotten into yaoi lately... n_n;; But I have no idea what 1x2 and all that other stuff means. XD And I had no idea that Jakult was actually... that. Yech. XP Maybe I still have some innocence in me. Nah. Anyway... I found some Gundam Wing doushinshi [?] you can see a pic  urlLink here . I found it quite funny. n_n Yeah, I\'ve also gotten into that as well, and... well.. some  other  things. n_n    Speedo-kun, I wouldn\'t kill you either. n_n;; '),(25179,950122,'12,April,2003','2003-04-12 00:00:00','   Oh. Here\'s the Address.     urlLink Which Yaoi Pairing are YOU? '),(25180,950122,'11,April,2003','2003-04-11 00:00:00','   Nope. No luck. I just hope Speedo-kun doesn\'t kill me. n_n; I got HeeroxDuo. *sigh* can you believe it? Have the world hates me!! XP But I won\'t let one quiz spoil my day! Right? I should\'ve gotten YukitoxTouya.... n_n;;    Men! What can you expect. *shrugs* '),(25181,950122,'11,April,2003','2003-04-11 00:00:00','   Wahhhh... I just took this Yaoi Pairing quiz, and, well... n_n;;; I\'ve already taken it twice! And it\'s still the same. Hmmmm.. Let\'s see if I get lucky this time! n_n '),(25182,950122,'11,April,2003','2003-04-11 00:00:00','   It\'s amazing what people would say over a song. I download lot\'s of files using WinMX, a file sharing program like Kazaa. I was downloading anime episodes, so I kept the computer on all through the night. I was only able to get back to the computer at about 3:00 p.m today, and it turns out this guy was messaging me throughout the whole night. Apparently he wanted to download one of the songs on the computer, but kept on getting disconnected. n_n;; Poor guy. He wanted the song so badly he messaged me about 15 times. But then again, he got pissed off with me [duh! How the heck am I supposed to reply?] and told me that he had already downloaded the song off of someone else, saying that I was an \'ignorant bastard\', and that I should go do a bear. n_n;;; MEN!    I\'m currently looking up the anime,  Wolf\'s Rain.  I\'ve been hearing that it\'s really good, so I\'ve decided to look it up myself. From what I\'ve read and found, I\'ve been convinced to actually start downloading it on WinMX. n_n;; With that, I just have to say that I love  Google.  I usually find everything there. n_n What?? You thought I was going to say someone\'s name?? Heh! You wish! XD '),(25183,950122,'11,April,2003','2003-04-11 00:00:00','   Well, lot\'s more happened than my previous post, so I\'ll post it. I\'ll try not to transmit gloomy vibes through your screens. n_n    I\'ve been trying to write  something ,  anything  for quite a while now. Nothing\'s some out. n_n ARGH! I wonder if I\'ll ever get my character\'s past right. n_n; I wonder if I\'ll  ever  get anything right. n_n;;;     \'\'A little biRdy in the sKy  DroPped A poOpOO in mY eyE.  I didNt Laugh,  I diDnt cry.  I jusT tHanKEd God...      COWS DONT FLY\'\'     I got this off of AC. I found it funny. n_n Here\'s another one.     \"Duct tape is like the force, it has a dark side, it has a light side, and it holds the Universe together.\"  --Carl Zwanig     I got this off of a blog that\'s no longer updated. Too bad.    Oh yeah. I bought this manga... Shadow of the Dark River. This is the same manga that made Enid go on a kissing craze... n_n But this is the second volume, so it\'s got things other that FKs. Hehe. n_n;; I\'ve shown it Gela already, and I think I\'ve gotten her hooked. She wants me to look for the third volume. n_n *shrugs* What can I say?    Oh yeah, I\'ve red streaks in my hair... my mom let me! n_n '),(25184,950122,'07,April,2003','2003-04-07 00:00:00','   Ok. It hasn\'t sunken in yet, so I\'ll take this opportunity to post it now before I become increasingly depressed.    My aunt died early this morning... cancer. Tristan and Pocholo\'s mom. I\'m scared to go to Iriga now... I dunno. Maybe I\'m in denial or something.. Even my father pointed out that in a span of roughly four years, we\'ve lost many of those close to us. Chiara, Eddie boy, Cheif, Indy, Uncle Sonny, Lolo Ernie, Lolo Judge and now Auntie Jing-jing. It\'s like some sick coincidence. Even freakier, at around 11:30 last night, something  \'flashed\'  in my eyes. I was freaked out to find out she died early this morning...    I\'m not sure what I\'m going to do when I het to Iriga. I\'m probably going to become moral support.. I hope I can handle it this time. '),(25185,950122,'07,April,2003','2003-04-07 00:00:00','   Hehe. I haven\'t updated my blog lately, now have I. Nope!  My cousin from Austria was staying in the study room, so basically I was cut off from the internet for three days.  Oh well. Lalala... I\'m quite bored. Let\'s just try to remember what I was feeling for the past three or so days.    Let\'s see... April 4!       April 4, 2003     My brother graduated high schoool today. I\'m reluctant to say it, but I guess I\'m a bit jealous of him. Most of the time I\'m used to getting all the attention in the house, now everyone\'s concentrating on Jeth. Recognition Day, Graduation Day, Graduation Prom... I dunno. I\'m just not used to it. I guess I\'m pretty selfish huh? Weird to think. Goody-two-shoes me is not so angelic. But then again, who said I was angelic in the first place? n_n;;       April 5, 2003     Stephan is here. He\'s my cousin from my dad\'s side. He\'s from Austria, and can speak fluent German. Crystal and I had fun having him translate some phrases for us [Mostly Crystal; she was in a questionable state], like \'Fuck the truth\', \'You suck!\' and \'Shit!\'. I still remember some of them. Haha. I wonder if I\'ll  actually  end up using them.    Crystal [my 17 year old aunt] and I were lying down on my top bunk doing nothing. We were talking about how we really didn\'t have anything to do, but the topic just led us to that; doing nothing. So we just stayed there, staring at the ceiling. Then Crystal asked me, \"Ria, what\'s the meaning of our lives??\" in a very subtle manner. I replied, \"Gosh Crystal, are we really  that  bored?\"    But that just brings me to why I don\'t like having nothing to do. It\'s not that I dislike not having anything to do because I don\'t have anything to do, but it makes me think of things I don\'t want to think about because there really is nothing else to do. [I hope you got that. n_n;;]    For some reason, I don\'t feel as if Jeth has already graduated. It\'s like I\'ve always had to be the big sister to him, and he\'s off to college!    Gosh, the house will be so empty once he leaves.       April 7, 2003     I can\'t really remember anything that happened yesterday, so I\'ll just post about today. Stephan has gone back to Uncle Teddy\'s place, where he first started. Crystal is leaving tomorrow, so basically I\'ll be pretty alone in this house since Jeth locks himself up in his room all day long. Really, I don\'t want to be alone. It\'s makes me think too much. I remember this time during sophomore year, I was supposed to organize my group in cooking as the next day we would be cooking/baking something. [I forget.] I was sitting on the couch upstairs SMSing everyone, then I had to wait for their replies. I started to think too much during my wait, and I had to leave the organizing up to Mikara.    Hmmmmmmmmmmm. Here I go again. I\'m starting to think too much. I don\'t want to be alone. Jeth is leaving, Crystal is leaving... I mean, even if Jeth holes himself up in his room, I still know he\'s  there . But in June, he\'ll be gone...      Schaisse!     I\'m thinking too much again.    P.S--> I did end up using the German words Stephan taught me. '),(25186,950122,'01,April,2003','2003-04-01 00:00:00','   I just watched Forest Gump and Cast Away, both with Tom Hanks as the lead role. I dunno, both movies hit me in a way. It should have hit me a long time ago, since I\'ve already seen them before. Forest Gump... it made me feel like I\'m wasting my time. My mom told me that there is no such thing as a limited person. And I quote, \'Stupid is as stupid does.\' Well, something like that. On the other hand, Cast Away made me feel like I was an island in the middle of the sea. I may have everything, but I\'m surrounded by merely water. '),(25187,950122,'31,May,2003','2003-05-31 00:00:00','   Hehe. Sorry for not updating lately. I\'ve been overrun by my male cousins, and if they\'re not playing the PS downstairs, they\'re playing the computer. n_n;; I get cut off from using the internet. :P But I\'m back! Back for now. :)    I took a quiz.        What Is Your X Japan Ballad?      *sigh* Love, it\'s so... thing! XP Ergh. Nevermind. n_n; '),(25188,950122,'26,May,2003','2003-05-26 00:00:00','   It\'s raining. Still raining. It\'s been raining for roughly three days. And I love it. n_n    I\'m just waiting for school to jump me. In roughly one month, it\'ll be school again, and I\'m waiting for my brain to comprehend it. Hehe, then I\'ll be panic stricken, running about the house screaming \"OMG school starts tomorrow!\". But I\'ll get to see everyone! :D Then I\'ll be all happy-dappy screaming \"OMG school starts tomorrow!\". That would tick my parents off either way because of my noise pollution. Hehe. Beh! :D '),(25189,950122,'24,May,2003','2003-05-24 00:00:00','   *announcement*     Mikara would like to say that she is not dead. She can\'t blog because her keyboard is broken and their internet bill is too high. Poor Gretchie. n_n;; '),(25190,950122,'23,May,2003','2003-05-23 00:00:00','   Poo. I\'ve been feeling weird lately.    After reading my past entries, I\'ve realized that I\'m no good at keeping one thought. n_n; I guess that\'s bad... but I guess I don\'t care much.    I bought a book of Hamlet. I couldn\'t find an english version of Noli, so I decided to buy it instead. I remember someone telling me that we might take it up in third year, so I decided to look it up. [Hoshi, was that you?]  I haven\'t read past the first scene... n_n; The way Shakespeare wrote is quite confusing. He has no sense of sentence construction! >    Not much happened today... So I\'ll be signing off early. Sweet dreams. That\'s what we all need. '),(25191,950122,'18,May,2003','2003-05-18 00:00:00','   Yey! I\'m back from Iriga! n_n I ate tons of Indian Mangos and did nothing but swim, play the PS and take up piano. Not to mention bum around the house, wake up late and... some other stuff. n_n;;    Sometimes I get the feeling that... I\'m disconnected. Disconnected... from something. I\'m not really sure. Then I snap back, just like that. It\'s weird when things suddenly pop into your mind without warning. That\'s one reason why I hate my imagination.    Speedo-kun! I saw you go online in YM! But it was only for a second or so... n_n; BTW, my Yahoo ID is  the_blacksolstice . Also, your archives don\'t work... n_n;;   Mikara:  Are you still looking for PS game suggestions? n_n Also, what genre are you into?    I watched  Matrix: Reloaded ! It was so cool! :D '),(25192,950122,'08,May,2003','2003-05-08 00:00:00','   I feel like hugging a tree. I feel like saving the ocean. I feel like lying down in front of a tractor on it\'s way to demolish the rain forest. No wonder... I got Earth on that angels quiz. '),(25193,950122,'03,May,2003','2003-05-03 00:00:00','   Hiya! I\'m back from my mini vacation. n_n I went to Cauayan, Isabela for three days. I got to see Callao caves and the Rice Terraces. It was fun. Well worth the three days. :)     You are a Day Dreamer..   Daydreamers tend to be away from the normal, and have a deeper aspect when looking at things. Most of the time, thier outcasts of the social society, but don\'t worry. It just means your so deep it scares thier simple little brains.    urlLink   What kind of Dreamer are you?     brought to you by  urlLink Quizilla      There\'s a lot on my mind again.. It\'s a weird feeling when you have so much to think about, that you can\'t even think straight. It\'s like... you don\'t know where to start;  how  to start.    I think I\'ll go to bed. It\'s been a long day. '),(25194,950122,'03,May,2003','2003-05-03 00:00:00','   I\'ve just logged onto  urlLink Fanfiction.net . I remember just a few years ago, the first website I\'d go to was FF.net to check for new stories and reviews. But now I only go there once in a while... Hm. Hey! Someone flamed me! > '),(25195,950122,'03,May,2003','2003-05-03 00:00:00','   I took the Belief-O-Matic test on  urlLink Belief Net , and here are my results:    1.  Orthodox Quaker (100%)  2.  Sikhism (84%)  3.  Orthodox Judaism (81%)  4.  Seventh Day Adventist (78%)  5.  Mainline to Liberal Christian Protestants (77%)  6.  Mainline to Conservative Christian/Protestant (76%)  7.  Jehovah\'s Witness (75%)  8.  Bahá\'í Faith (74%)  9.  Eastern Orthodox (74%)  10.  Roman Catholic (74%)  11.  Church of Jesus Christ of Latter-Day Saints (Mormons) (73%)  12.  Hinduism (71%)  13.  Liberal Quakers (70%)  14.  Islam (68%)  15.  Jainism (67%)  16.  Reform Judaism (65%)  17.  Unitarian Universalism (58%)  18.  Mahayana Buddhism (54%)  19.  Theravada Buddhism (54%)  20.  New Age (51%)  21.  Neo-Pagan (49%)  22.  Taoism (37%)  23.  Christian Science (Church of Christ, Scientist) (36%)  24.  Scientology (31%)  25.  Secular Humanism (30%)  26.  New Thought (29%)  27.  Nontheist (17%)    I\'m not all too sure about this though... I have my own questions about the infinite. '),(25196,950122,'28,June,2003','2003-06-28 00:00:00','   I need to get a new layout for my blog. This one seems too cluttered now. '),(25197,950122,'15,June,2003','2003-06-15 00:00:00','   Woah, Blogger looks so... different. Sorta confusing. Anyway, I\'m glad I\'m finally online!! Yeah!    Ergh, my face hurts... I had a tooth extracted today. [Actually, I\'d say my dentist yanked it out of my mouth... *rubs cheek*] I\'m going to get braces pretty soon. >    Tired........ I feel tired. And hungry. I haven\'t eaten lunch yet, and it\'s... 6:20 p.m?! >    Gundam W will be showing on Cartoon Network after Tuesday! Digimon Tamers ended yesterday [I think] and I have only seen 2 episodes of Invader Zim. The one where he makes this one kid his \'best friend\', the one where they had to sell chocolate bars to make money... and when Zim invaded Dib\'s body. Wait, that\'s three episodes. It doesn\'t matter anyhoo, I never catch it anyway. '),(25198,950122,'15,June,2003','2003-06-15 00:00:00','   I\'m back again! *sigh* All this traveling is going to take it\'s toll on me one of these days. n_n;; Blogger doesn\'t like me right now. It won\'t load properly. I\'m having doubts on whether this entry will even show up or not. >< Anyhoo, school starts tomorrow. My mind seems torn between feeling excited and feeling just dead. x_x Thinking about the school year just stresses me out right now. n_n; Beh. That\'s all I can say. Beh. '),(25199,950122,'30,July,2003','2003-07-30 00:00:00','   My bloggie is so lonely now. ;_; Haha, I\'m just kidding. I\'ll be posting more after the tests pass. I\'ll have a lot of free time after that. In fact, I just got back from publishing our Journalism project. \'Tis weird; this morning I was asking myself how I\'d get it finished. It was so incomplete, but now it\'s finished, and it looks coolie. :P Now I only have to worry about my CL project and my Computer project. Hoshi! I\'m going to interview you tomorrow! xD '),(25200,950122,'27,July,2003','2003-07-27 00:00:00','   Coup de etat. Well, something like it. Some young soldiers have taken over Makati, and all of the foreigners have been evacuated from their hotels. All the malls there are closed as well. Tomorrow the economy will drop. This is freaky. 0_o Part of me still doesn\'t realize what\'s going on; it\'s not like I can\'t believe it, it\'s just that, I never thought I\'d have to go thru one of them. Besides, the country isn\'t in need of a coup. This \'coup\' will just make the country worse. '),(25201,950122,'22,July,2003','2003-07-22 00:00:00','   You represent... insanity. You\'re quite a quirky little creature.  You\'re emotions are varying.  You may appear childish and innocent, but you have a tendency to freak out.  You\'re incredibly random, but it\'s good to be unique.  People know you\'re an odd one, but you certainly don\'t mind.    urlLink   What feeling do you represent?     brought to you by  urlLink Quizilla '),(25202,950122,'21,July,2003','2003-07-21 00:00:00','   Wow!  I met Duncan Sheik last sunday! :D It was so cool! I never thought I\'d meet him. :P I got a poster, a mag, his latest CD and his autograph! :D Sure isn\'t something that happens everyday. It all happened like this. My family and I went to Church, then we went to a local mall for lunch. While we were eating, my brother text his classmate from 2nd grade (from Japan). One of her friends had found my brother\'s website, and signed the guestbook, giving him Marielle\'s (bro\'s classmate from 2nd grade) e-mail address. He e-mailed her, and it turned out that she was here in the Philippines, right next to Ateneo, which is pretty close. So, she e-mailed him her cell number. They text each other, and she said that she was going to Greenbelt, a strip of nice restaurants and cafes near Makati. My mom, always wanting my brother to go out, had us pick her up and to go Greenbelt. :P So, we went there, and went to Music 1, a music store where Marielle\'s friends work. It also turned out that Duncan Sheik was to be there at 2:30 p.m  (We got there at 2:00 p.m) for autograph signing. So we stayed there because Marielle had been listening to his music ever since she was in 5th grade. Then he came, and I met Duncan Sheik. Cool huh? :D  Tests are next month~! Geez, time flies. '),(25203,950122,'14,July,2003','2003-07-14 00:00:00','   Yes! I\'m now taking medicine for my medicine, and I\'m almost cured! Happy me. :P    I haven\'t been doing any of my assignments lately, and it was only till one of my classmates asked me, \"What\'s happend to you?\" did I realize that I\'ve been slacking off. > < Do projects, and finish all my homework!! '),(25204,950122,'07,July,2003','2003-07-07 00:00:00','   My feet are swollen. Also horribly itchy. It turns out I\'m allergic to the antibiotics I\'ve been taking lately. Ergh, it\'s really getting to me. x_x '),(25205,950122,'06,July,2003','2003-07-06 00:00:00','   I\'ve lost my glasses somewhere in my room, and only now I\'ve realized how much I need them. I accidentaly read \'Blogger\' as \'Booger\'. >< I guess this will be just a short update. I haven\'t been going online much lately. Hm. I really don\'t like Blogger\'s new layout. '),(25206,950122,'30,August,2003','2003-08-30 00:00:00','   Not being able to go online has been hell! I end up with not much to do. Anyway, nothing much happened this week. Went to outreach last friday tho. I helped teach some special children in a centre called \"Nayon ng Kabataan\". I ended up making drawings for a lot of kids. :) The best part of the trip tho was our bus ride to and fro. Our class rode in three buses, but only two teachers went with us, and our bus ended up being the one unsupervised. We\'ll be going back sometime next quarter, so I won\'t be seeing them for a while.  We\'re having half day on Monday; the high school and grade school teachers are going on a pilgrimage for enlightenment. Don\'t know whether I should be happy or sad.   Oh yeah, my birthday is this friday. Wonder what\'ll happen. '),(25207,950122,'23,August,2003','2003-08-23 00:00:00','   6 days in my auntie\'s house, and 45 days till my family returns. I\'m not as depressed as I thought I would be. n_n  I went to Glorietta with my friends yesterday. I went there with Gela, and we walked around for a while before we stopped at Pizza Hut to wait for Camilla and Racine. After that we walked around even more and ended up watching a romantic-comedy movie. Then we went to the hotel Camilla\'s family was staying at. She and her sister tried to teach me this card game, but it was all in vain. x_X Funky. It was a lot of fun. :D But on the way home I ended up leaving my cellphone in Gela\'s car. So I survived a night without my phone then Gela returned it the next day. n_n Maybe I\'ll call her later. :D '),(25208,950122,'18,August,2003','2003-08-18 00:00:00','   I forgot to post this on my previous post... I won\'t be staying in my house while my family is away. They\'ll be gone until sometime in October... so I won\'t be online much. And I won\'t be going to any gimmicks either... I\'ll probably be able to update my blog during the weekends only for a while. I hope things will only get better for now. '),(25209,950122,'16,August,2003','2003-08-16 00:00:00','   Blackout in the Northeast! It sounds weird hearing it on CNN and FOX. Then again, there were a lot of places that were affected, so it\'s understandable. We really depend on electricity now a days. Seeing photos from NYC with the traffic jams, people walking in the streets and commuters sleeping everywhere, I just imagined what it would be like if we really used up all of our energy resources. Conserve energy! It\'s not to late!!! [I hope.]  urlLink Here  are some pictures of the blackout in NYC taken by the moderator of  urlLink imood.com .  About being forgotten.... I feel like I\'m being abandoned by my family. It\'s a stupid thing to say, but them going to Virginia and all.... it makes me feel all alone. Solitude is okay, but loneliness is different. I feel so left behind. My mom keeps on talking to me about what would happen if anything were to happen to them; where the Bonds are, where my parent\'s wills are, where we keep our important documents... but I just don\'t want to think about it. It\'s selfish, but I really don\'t care that much at the moment. We\'re all being inconsiderate to each other. Everytime I start to miss my brother, he does something that makes me feel like I shouldn\'t miss him.  He had his farewell party yesterday; we all had a lot of fun. Now I know I\'ll miss him no matter what he does or says to me. u.u I don\'t knw how to let him go.... I don\'t know if I can let him go. '),(25210,950122,'10,August,2003','2003-08-10 00:00:00','   Happy Birthday Mom! It\'s my mom\'s birthday. n_n She\'s 26 as always. Lol. :P  Yes! I got a CD burner today! :D Oh happy happy. n_n I also bought a tee, a bunch of CDs, a throw pillow and some Inu-Yasha manga. Greenhill is a lovely mall. Oh yeah, I also got a FF9 mini soundtrack for my friend Guia. I spent a lot of my money tho... x_x But I\'ll make up for it with my CD burner. Hopefully...  Oh damn. My brother is leaving in ten days. x_x It\'s hard to imagine what it\'ll be like when he\'s gone, even if he usually locks himself up in his room... He\'ll be going to Virginia... along with my parents; I\'ll be all alone.... even on my birthday! > '),(25211,950122,'08,August,2003','2003-08-08 00:00:00','   Tests are over... They weren\'t as hard as I thought they would be. Studying helps alot. :P I should study more often. 0_o Gela was absent for the three days of major tests. I found out earlier today that she had Typhoid Fever, which really freaked me out. She\'s better now. n_n Gela LIVES! :) Our Chemistry test was very easy... our English test ended up being hard. >< It pissed me off; I studied all night for Chem instead of English. I should get over it pretty soon tho... Next week we\'ll have new lessons; more things to study. :P '),(25212,950122,'21,September,2003','2003-09-21 00:00:00','   Yey! My parent\'s will be back soon! :D It sounds pretty weird, doesn\'t it? Anyway, I\'ll be able to go back to my house when they come back. I just miss them a lot. n_n I ended up buying the infamous pair of pants, along with a Vienna Teng cd. Pretty good buy. Unfortunately, now I\'m broke. >< '),(25213,950122,'14,September,2003','2003-09-14 00:00:00','   My blog is pretty lonely now. n_n;; My birthday was lonelier than expected. It didn\'t feel like my birthday that much, but Tepai gave me a bouqet of flowers, and that made my day. The pair of pants I was planning to buy didn\'t fit me when I tried it on. ;_; And I didn\'t get to go to the Ragnafest either. ;_; Aw sad. I hope they have another one some time soon. n_n '),(25214,950122,'19,October,2003','2003-10-19 00:00:00','   I watched  Nausicaa of the Valley of the Wind  last night, and I cried through nearly half of it... especially the end. >   Anyway... internet has been a drag. My connection is never stable now-a-days. Damn... It pisses me off. >  Today we went to the UFO Church in U.P. I met Janelle there. Then we went to Greenhills and stayed there for 5 hours. 0_o It was so boring. My mom was so overwhelmed by the jewelry she was buying for her business. Ugh. I wonder if she\'ll ever get tired of looking at things that sparkle. > '),(25215,950122,'16,October,2003','2003-10-16 00:00:00','   Sorry I haven\'t posted lately. I\'ve been busy with things like studying for tests and last minute projects.  My brother is gone now. The house is empty. My mom has changed the way she treats me; she\'s trying to make me more responsible and \'prepared\'. Anway, I had my last day of tests today. I went to school earlier than usual because the car was banned, and that gave me more time to study for Christian Living, Araling Panlipunan and Journalism. Luckily for me, the tests weren\'t that hard. :) After that, I had practice for our presentation that we\'ll be presenting tomorrow, and went home with Camilla while my parents were at Greenhills. Had  three slices of cheese pizza for lunch, then we went out to buy mangos so I could help Camilla make Maki, but just as we finished slicing all the mangos, my parents came to pick me up. Then we went home and since then I\'ve been doing nothing but become a blob in front of the computer. Actually, I should be playing right now... but I\'m out of money. *sigh* Not really and eventful day.  I logged on to  urlLink Friendster  earlier to find out my brother already has an account. 0_o Well, now he\'s my friend, so no worries. Not much more to say... so peace out. '),(25216,950122,'11,October,2003','2003-10-11 00:00:00','   Ouch! I got hit in the face with a softball on wednesday and thursday... now I have bruises on my nose and on my right cheek. x_x My parents are wondering what type of training I\'m going through. n_n;; Anyway, I\'m back home! :) The house was so humid that mildew was growing everywhere. Gah. Luckily, you just have to wipe it and it\'s gone. n_n '),(25217,950122,'25,November,2003','2003-11-25 00:00:00','   I guess I had a bad day today. I really don\'t want to talk about it, so let\'s just say I had a bad day. Things just came crashing down on me... I finished some of them, but not all.. like that stupid ass music project. I hate it when she doensn\'t announce when we\'re to pass her projects!! Just because we always miss her class, doesn\'t mean that we\'re going to read her damned mind..   Anyway, tomorrow is a holiday; Hari Raya I think.. but our retreat is tomorrow, so we\'re going to go to school anyway. I\'m not so excited.. I\'m not really in the mood to \'open up\', cry and stuff like that.. I hope I\'m just in a bad mood. I don\'t want to end up ruining everyone else\'s retreat just because I feel like a pathetic bitch.I hate it when I feel like this..  It\'s during times like these I realize I need a punching bag... '),(25218,950122,'17,November,2003','2003-11-17 00:00:00','   Ack. I got my report card for the Second Quarter on saturday, and my grades in about six subjects went down. @_@ My grade in Trigonometry is 76. ;_;   ...  And now, my internet connection has failed. Again. -_-;; So as I\'m typing this, I have no internet connection. My mom says that I\'m relying too much on technology, which is true, but we\'re paying to get internet service. And so, we\'re paying for a service that isn\'t even working! That\'s not right! If she has to send an important e-mail and the internet fails her, what would she do? She gets pissed off just as I do, doesn\'t she? To put it bluntly, I live on the internet. n_n;; I know it\'s not good, but it\'s true. I don\'t watch TV because of this bad habit I have (I tend to faze out when I\'m watching TV... and when someone talks to me, I don\'t notice... it\'s completely rude, and completely embarrassing) and I don\'t talk on the phone. Well, we don\'t have a phone. -_-;; I have a cell phone, but calling on a cell phone is impractical because it\'s expensive. Text on the other hand isn\'t very reliable, and as most of my friends text a lot, they\'re usually out of load. As for me, I don\'t know why I don\'t text. n_n;; Technology is supposed to be reliable! That\'s why a lot of people say that technology is the future! Sure, we shouldn\'t go as far as \'The Matrix\', but I probably won\'t be alive then! 0_o  ...  You know, I can\'t believe that I\'m getting this worked up about a stupid internet connection... I just hate it when things don\'t work, even if I\'m good at compromising.  Phew. Haven\'t ranted in a while.. -_-;; '),(25219,950122,'07,November,2003','2003-11-07 00:00:00','   I actually understood our lesson in Chem today!! XD I\'m so proud of myself! We had Mass to Volume conversion in our Stoichiometry chapter. Maybe I should pay more attention in class. *wink*  I stayed up all night just reading an online comic titled  urlLink \"RPG World\" . :D It\'s funny, taking up all the stupid things in RPGs like repetitive battles and stupid heros. Lol. Most of the puns were rips from the FF series. :P The thing is, the comic has been up for three years, so I read three years worth of comics. 0_o Hekhek, bad for the eyes... @_@ But once I started the story, I figured it would be too much a hassle to stop then and begin again tomorrow.  Oh yeah... I saw Racine\'s date to the Prom!! :P He\'s got a nice smile. Plus he\'s megane. Lucky Racine. :P Speaking of the prom, now my dad wants me to have a date. >< It\'s not like I don\'t want a date, it\'s just that I\'m not sure if I\'ll like my date. I mean, a bad date can ruin a prom.. Of course, me, being me, I\'m expecting too much of the prom. Ergh.. *mumble stupid prom mumble mumble* '),(25220,950122,'03,November,2003','2003-11-03 00:00:00','   Read Kari\'s post about living without internet.. I wish I could say the same, that not being able to live with out internet is kinda stupid [which it is] but internet is my only connection to the outside world, aside from my cellphone. Well, at least I think so.. After all, the only way I get to \'talk\' to my friends when not at school is through chat, which technically isn\'t even talk. Thinking about it makes me feel lonely... Yeah, I\'m lonely..  Gawd. How can one enjoy solitude but feel lonely at the same time?  I feel like my stupid internet connecetion; stupid, and always disconnected. '),(25221,950122,'25,December,2003','2003-12-25 00:00:00','   Listening to: Mom telling me to eat '),(25222,950122,'21,December,2003','2003-12-21 00:00:00','   Listening to: -   Merry Christmas people! It\'s 2:33 am, and I really shouldn\'t be up anymore, but I\'m not sleepy, so thing. :P ~~~~~ Let\'s travel back in time. n_n;; Last tuesday I went to Galleria with Gela to meet up with Racine and Camilla, and this guy and his friends from their interaction. We ended up getting there late, but Racine was the last to come. n_n;; We ate at Pizza Hut and decided to roam around until the guys got there. They didn\'t until around 3-4 pm, which was bad because we had been there since around 12. During that span of time we basically went everywhere in Galle. Then, as we were going up the stairs to go to Starbucks, I saw him. 0_o The guy I had (note: had) a crush on at the interaction. He was with two of my other classmates. Damn, it really caught me off guard! There I was, hoping to meet a nice guy, when out of the blue, I meet the guy I gave up on. He wasn\'t supposed to be there!! n_n;; Seriously though, it pissed me off. What pissed me off more was the fact that sooo many other guys at Galle looked like him. > ~~~~~ Christmas Eve! Actually, it\'s already Christmas here. :P I got a portable sewing machine from my auntie! O_o And right now, I\'m trying to translate this H doujin.. (Don\'t ask why. n_n;;)How do I do this?!?! x_x Heeelp!  But overall, this Christmas is way much better than last year\'s. :P '),(25223,950122,'21,December,2003','2003-12-21 00:00:00','   I bought my prom dress today! Guess how much?? 800 pesos only!! That\'s like... 14 US dollars!! XD It\'s actually pretty simple.. I\'m going to have a shawl, but my parents and I are debating on how the shawl should look. n_n;; But still, it\'s one big load off my chest. I don\'t have to worry about my main dress anymore. Plus, I got it at a really cheap price. Which makes today a happy day. :P '),(25224,950122,'15,December,2003','2003-12-15 00:00:00','   Yey! Christmas Vacation! Two weeks of having nothing school related to do! :D Now I can try to get my mind off that horrid Trigo exam... I\'m sure I failed it.. and I know that I fail Trigo this quarter.. >  Anyway, a few nights ago I felt like crap. Why? Because I passed up the opportunity to go to a friend\'s house to go to a Christmas party. Then what happened? I ended up losing my transpo and not going to the party. Right now I\'m just telling myself that everything happens for a reason. Gah. On the brighter side, I got these kick-ass platforms from my auntie! :3 They\'re really comfortable and the design is cute too. :P Plus, I look taller when I wear them. XD Anyway, I went to Greenhills yesterday with my family. I bought the Sims : Making Magic! But something\'s wrong with the cd... n_n;; Never buy pirated, eh? At least I\'ve installed it. Now I just need a cd that\'ll run properly.  It\'s settled. I\'m going to the prom. Most likely without a date, but who cares? My mom suggested that I take my cousin (Tristan?!?!) as my prom date, but I reasoned with her, telling her that he\'d wear a tux, but have army boots on. n_n;; I\'m getting a bit worried about my dress. I haven\'t been to the seamstress yet; actually, none of us have been to the seamstress yet. Plus, I don\'t have a design. I\'ll probably just rip something off the internet and modify it a bit.. n_n;; If all else fails, I\'ll just guy one. It can\'t be that bad. :P  I\'m worried about my family. My mom just had a checkup, and her blood pressure is really high. She\'s taking medicine and not doing much around the house now. Dad, on the other hand, always seems to get on my mom\'s nerves, and just this morning I woke up to hear mom nagging dad again. Like usual, it was about money. My mom does all the budgeting, and my dad just loves to spend. It\'s not exactly the best combination, especially because we\'re not exactly rich or anything. My parents don\'t work for goodness sakes. We\'re just relying on dad\'s pension.. That\'s why I try my best not to spend too much. Because if I do, I might not have anything to fall back on.  *looks at her computer clock* Wow, only four more days \'till Christmas. n_n My aunties (My mom\'s sisters) will be spending Christmas with us this year. They\'re coming here tomorrow with my grandma from the province. I wonder what will happen this Christmas. I can expect the unexpected because my mom\'s sisters are crazy! They all have these trademark laughs. 0_o I hope it\'ll be more fun than last year.. *looks back* '),(25225,950122,'15,December,2003','2003-12-15 00:00:00','   Went to an anime convention last Saturday. Over all it was ok.. I ended up either getting pulled around or being left behind.. n_n;; I got this drawing there, but unfortunately it\'s missing right now, so yeah.. We got the idea to cosplay, but I don\'t know if we\'ll really end up pulling it off. Rivermaya was there. I\'m going to buy one of their CDs on of these days. Anyway, when they started playing, (we were sitting on the floor) everyone just suddenly got up and left to watch. n_n;; I really felt like crap at that moment. And they were playing that song.. \"Umaaraw Umuulan\" (LT: Basically a song on trying not to be sad and how life is \"sadyang ganyan\") I couldn\'t exactly go and watch too; I didn\'t want to leave their things just  there . Especially because they bought this stuffed toy that cost 500 pesos. So when they came back, I just left. I really just couldn\'t take it.  We saw Imelda Marcos that day. She was inside this Japanese candy store. They wouldn\'t let anyone else in. She had like, 7 bodyguards? Ugh. Frick her. She should be rotting in jail, that old hag.  Afterwards we went to Camilla\'s house. Ordered pizza and watched Weiss: Gluhen. It was more like a violent soap opera where the person you vow to protect dies, you get used by the bad guy, and the old man you end up killing is actually your grandfather. And that was only the first four episodes. So yeah. Then I went home.  Anyway... I got thinking again. And guess what I started thinking about! Guys. Ugh. It\'s the file in my brain that reads: \"Don\'t Open\". But stupid little me; I obviously don\'t know how to read. I got to thinking about it because of out junior prom.. which is only one month from now. crap. Well, my friends all have dates, so they said I should have a date too, because they\'d be busy with their dates. And I wouldn\'t be busy at all. But unlike them, I don\'t know a whole class of boys or have a kuya or even have a partner from our interaction. I don\'t even have a guy text mate. Tuh! The last guy who text messaged me text me for only half a day and never text back. So yeah. My social life with the opposite sex is very happy. Actually, I\'m having second thoughts on going to the prom. I mean, I\'d only end up being \'the pretty little wallflower\' like my mom says. That\'s what happens to all the occasions I go to. Besides, if I don\'t go, I won\'t have to buy a dress. I could just give my invitation to a classmate (who\'s bringing two dates) and she could just pay me back.  The whole prom seems like such a hassle now. '),(25226,950122,'08,December,2003','2003-12-08 00:00:00','   I\'ve got a new layout! :P It took me a while to finish this up... n_n;; My first layout ever! :D I\'m so proud of myself. Now I just have to figure out what the heck happened to my tag-board. 0_o '),(25227,950122,'02,December,2003','2003-12-02 00:00:00','   I had my retreat with my class last November 26-29. We sneaked our phones and food into our room though, and we had codenames for both. We called our phones \"Panty\" and the food \"Utang\" (LT: loan? I\'m not really sure) It was funny hearing my room mates casually going : \"Ria, can I borrow your panty?\" or \"Ria, there\'s a name on your panty!\" (meaning I just got a text message) It was so bad and yet so hilarious! XD The place was wonderful too. It had a nice abiance surrounding it, and it wasn\'t that noisy (except for the roosters outside our rooms and our own noise.)  The food was good too. They fed us 6 times a day for krissakes! And after our snacks, they let us have a siesta/nap. Suuuper fattening activities. Lol.  Our proctors, Brother Paulo and Brother Bernard, were wonderfully gay. n_n And when I say gay, I mean it both ways, even tho they\'re both straight. :D It was so much fun listening to them and watching them. But Brother Bernard\'s life story is so sad. n_n;; Most of my classmates cried. Yeah...  The Retreat turned out waaaay better than I expected, and I feel really lucky.  --------  After the retreat, I went to Lillia\'s house to work on a project, then I went to Gela\'s house.  --------  I went to the J&J Cheering Competition last Saturday with Gela and co. after sleeping over at her house. We got to play the Sims : Making Magic. I\'ve got to buy that CD and install it without Jeth.. :P I fell asleep too much.. n_n;; I practically beat the purpose of the sleepover. Lol.   Anyway, we got at the Ultra Colesseum the next morning a bit later than expected, but it didn\'t start until an hour or so later... ugh, talk about Filipino Time.. As we were going in, we were given this kick-ass cute-ass halos. :3 We got the best seats because our school was right above the judges so we got the best view. Then it started..and then we saw Miriam College\'s cheer. I got scared. Honestly in my opinion, it was good. But then they copied some of our stuff, (I heard they even used our school cheer during the elims) and I couldn\'t help but jeer. Hah. At the end of their showy performance, we found out that they didn\'t have saftey marshals for their wild moves(like throwing cheerleaders all over the place), and it was a big downer for them and a big upper for us. :D (I\'m bias. So sue me.) Then we performed. And I screamed, yelled and cheered at the top of my lungs while snapping pictures for our newsletter. And I absolutely loved it. :P I was just shoving it in MC\'s face. Or at least I saw it like that. Lol. In the end, SSA (School of Saint Anthony), last year\'s Champions got first place again. It hurt a bit, but they were really good. Plus, they weren\'t copy-cats. (I can\'t think of a better term. n_n;;) We got First Runner Up and MC got Second Runner Up. Afterwards, we sat on this legde-like thing that was on the corner of Ultra. Because of our halos, everyone passing by knew we were SHSians. (I heard another school wanted to steal our halos. 0_o) It was so cool sitting there because all the other students of the schools that joined congradulated us. Even those in the buses passing by made a \'C\' with their hands. n_n;; St. Mary\'s even chanted \"Holy! Holy!\" as their bus passed by. Miriam on the other hand closed their curtains. n_n;;; I\'m really glad we went.  Afterwards, we ate at Kamameshi House. We started telling those wierd brain-teaser like things and even the staff ended up listening in. \'Twas kick-ass I tell you. n_n  Then I went home, and slept like a rock. '),(25228,950122,'23,January,2004','2004-01-23 00:00:00','   Listening to: Garbage ~ Cherry Lips   *Continuation*  It took forever to start. We talked about anything while waiting. It finally started, but we just kept on talking with each other. It was fun. :D Soon afterwards tho, Gela and I started to get really cold. x_x Again, Franco offered Gela his coat, but still she refused. :P The soup came, and Gela and I were happy because it was hot. Other than that, the soup sucked.. It was gritty. >_> CJ had this thing for the butter, so we told him to get the three extra ones from the three unoccupied seats at our table. :P Of course, he didn\'t get them. Then this dude came up to us and asked us if we wanted to get our pictures taken while there was no line. So we did. Camilla was first w/ her date, then Racine. When the photographer called Frederick to join her in the pic, he called him an escort. Then Racine was like, \"Escort?\" So I said, \"Security!\" \":P Gela was the first to get their pictures taken with her date. Then it was my turn. I got four poses because I was stag. :P After those pics, we got a barkada pic, then one with our bodyguards. :D After the pics, we went back to the hall, and wow! Dinner was there, waiting for us. I felt bad after dinner bacause I was the only one who finished all of my food. >_  After dinner, there was dancing! Actually, we didn\'t dance at first because we were pratically blinded by the disco lights. x_x But as more and more people started to dance (and much prodding and convincing), we decided to join in. We just swayed at first. Then I started to dance (tried is more like it), then we were swaying with style! :P Then Franco suddenly started to dance! XD But he got a headache soon afterwards. Turns out he was actually a bit sick, but he still went to the prom (AWW GEL!! XD). There was a time for slowdance, but it was only ONE SONG. O_o Gela and Camilla took of their shoes after jumping around a bit. Racine and I kept our shoes on. My shoes were comfortable enough to jump around in. The guys sat down near where we were swaying, and I noticed that they were looking and Gela\'s and Camilla\'s feet! :P It was so weird.   The awarding came next. I was a bit bummed because they didn\'t have a stag of the night award. :P I think I know who should have won it. (Joanna Feliciano!) The Prom Queen wasn\'t who everyone was expecting (at least not part of the group everyone was expecting), so that was cool. Some people still danced afterwards, but we decided to hang out in the lobby. The lobby was filled with parents waiting for their daughters/sons. Not to mention there was another school having their prom in the hall next to the one we were using. We were all tired, so we just sat down, waving good-bye to those leaving while talking. I\'m not sure if it was Franco or CJ who left first, but after CJ left, Camilla went upstairs. Gela, Racine and I stayed in the lobby waiting for Frederick\'s family to pick him up. They came after and hour or so, and when I saw his brother, I thought he was MY brother! Geez, they looked so alike. Thinking the night was over, I started for the elevators, but Racine started yelling. \"He forgot his glasses!! O_o\" I grabbed them and ran (as if I could run in that dress. >  Phew~! I don\'t think I should add anything else. I might end up writing a story.. O_o '),(25229,950122,'20,January,2004','2004-01-20 00:00:00','   Listening to: Maksim ~ Hana\'s Eyes   Ok ok, my description of the prom on my previous post was completely uninspired. :P So here goes:  I got to the hotel first. While we were still driving around outside, I couldn\'t bring myself to think that  that  was the Manila Hotel. In my head I was thinking,  Isn\'t this supposed to be a five star hotel or something?  Honestly speaking, it looked like a big motel from the outside. :P But once we drove into the driveway, boy was I surprised! Inside the lobby, we met up with Tita Lulu, who was basically the big honcho on organizing the prom. While they were talking, (my parents and Tita Lulu) I decided to walk around. The Lobby had many red sofas and even a band playing classical music. Plus a chandelier. Anyway, I went to the Polka Ball to have a sneak peak of where the prom would be. I wasn\'t too impressed tho. The first thing I saw when I went inside was the ceiling. Sure, it had chandeliers too, but some of the lightbulbs were busted, and the ceiling was dirty. With cobwebs and unidentified sticking objects. There were more tables than I expected, but the stage sucked. O_o Then I noticed the carpet. I was wearing the platforms my auntie gave me, and when I stepped on the carpet, I felt like I was sinking! >_>; It was that soft. Soft.. But it had all these black spots on it. O_o Like bubblegum or something. After that, I went back to the lobby where my parents were still talking. We went to our room after checking in. Room #1013 on the 10th floor. :D The room was very comfortable. We got a view of the ocean and the swimming pool. :P  After a while, Gela and Camilla showed up at the room (after much texting and excitement from yours truly). They told me that the beauty dude was there, so I took a shower and went down to get transformed. >_> I brought my mom with me, and when we got there, Gela was just getting finished. The mommies started talking and stuff, and after a while, Racine showed up with one of her mommies. Then it was Camilla\'s turn to get pretty, then mine. While I was in transformation, Gela, Camilla and Racine went to Racine\'s room to get her dress. Then Camilla\'s mom came, and so the room was filled with mommies. :D Then Racine got pretty, and it was time for the prom.  While I was running around, going upstairs and downstairs getting all my stuff, Gela\'s date came and they registered and went inside. Before that tho, I was able to get some pictures of us. We waited in the lobby for Camilla\'s and Racine\'s dates, but then Gela texted, telling me to get over there with her. :P I then realized that I was the expendable one because I didn\'t have a date. Anyway, I went there and registered; got inside and met Franco. He was nice; he talked, and when Gela was getting noticably cold, he offered his coat (numerous times!), but Gela refused. (aw...) The conversation was just getting started when Camilla came in with her date, CJ.  Hmmm, approved!  I thought as I looked at Franco and CJ. :P Honestly, that was what I was thinking then. I asked about Racine, and it turned out she was still waiting for her date, Frederick. I went to the lobby, and said \"Moral support has arrived!\" Luckily, we didn\'t have to wait for a very long while. He came after a few minutes ( Approve approve! ) and it was to the prom for us.   *note* My post was WAY longer than this... but it got cut off. n_n;; I type to much when I\'m narrating something. I\'ll just post the rest next time. My mom\'s being crappy and telling me to get off the computer. *note2* Again, I\'m stuck in a bad situation with my mom. x_x Why is this happening? '),(25230,950122,'19,January,2004','2004-01-19 00:00:00','   Listening to: .hack//SIGN ~ Aura   Prom is over.  How was it? It was better than I expected; I wasn\'t a burden to my friends (I hope I wasn\'t), but I didn\'t get left out either. The thing that bummed me the most was that there was no Stag of the night award. Yeah, it\'s shallow, but I just wanted to show some certain people that going stag on the prom isn\'t the end of the world. So yeeeah.. I felt a bit biased when there was no award at all. But moving on... I\'ll try posting some pics some time. :D '),(25231,950122,'14,January,2004','2004-01-14 00:00:00','   Listening to: Dido ~ White Flag   Prom Countdown: 2 days  Things are getting better. (He heard my prayer!) Mom and I went shopping in Ever, and I ended up getting a pair of capri denims because all of the pants I tried on didn\'t fit. O_o Made me wonder if I was getting fat.. I was actually there to get a manicure/pedicure for my prom, but there were too many people in David\'s Salon. It wasn\'t a complete waste though as I found a hairstyle while flipping through a magazine. My mom stole the page it was on. n_n;; Not much happened today.. *shrugs* '),(25232,950122,'13,January,2004','2004-01-13 00:00:00','   Listening to: Goo Goo Dolls ~ Name   Prom Countdown: 3 days  I am in a slump. I feel lost. No direction. No destination. Not even a northern star to guide my way. (As if I could even find it) I\'m just.. so... ergh.  I had a dress rehearsal with my mom last sunday. When I was wearing everything, I was just looking at myself in the mirror. I told her I wanted to win the Stag of the Night award, and she said nothing. Not even a nod. She just looked at the dress. She didn\'t say  anything . It was so horrible; even a shred of recognition would have been ok, but she showed nothing. I took my dress off and hung it in my brother\'s closet. As I wiped the make-up from my face, I couldn\'t help but wonder if I really am the source of all her stress. Looking back on the past week, it doesn\'t look good at all. She doesn\'t speak to me. She doesn\'t greet me when I come home and greet her. Geez, not even at the dinner table. I can\'t help but think that she hates me or something. I know I\'m probably over reacting, but with the way things are...   God, please prove me wrong. '),(25233,950122,'07,January,2004','2004-01-07 00:00:00','   Listening to: Alanis Morissette ~ Uninvited    Prom Countdown : 9 days  Everyone (almost everyone) in my batch has the prom fever! Lately I\'ve only been hearing about prom dates, prom dresses, make up and all that.. stuff. I\'d say crap, but in a few days I might have the fever too. > not  going to the prom feel. Like Gretchen. She wasn\'t allowed to go the prom. That really sucks. She barely gets to go out with us, and now she doesn\'t get to go to the prom? Geez. She better go to the Grad Ball. *mumble mumble*  Oh! Tomorrow is Thursday... I\'m getting my teeth tomorrow! XD Right now, I\'ve two gaps on the upper teeth.. > long  time ago, but it got posponed because my dentist was pregnant and gave birth. 0_o I\'ll be getting 2 other teeth extracted after the prom tho. How ouch. Then I\'ll get braces. Meh.  ERGH!! I\'m in such a thing situation with my mom right now. I mean, I know I failed my Trigo test. Nearly everyone did! But I passed Pinoi with a decent grade! And I got the highest in AP in my class! But she still grounded me. (Sorta grounded me. I don\'t care; she grounded me!!) I mean, she didn\'t even give me a shred of consideration as to how hard the tests were.  Now that she has High Blood Pressure, she acts like everything I do stresses her out.   *sigh* Seriously, that\'s not fair. She\'s still my mom, and I still love her, but the way she\'s treating me isn\'t fair at all.  Sometimes, I don\'t even know why I try. '),(25234,950122,'04,January,2004','2004-01-04 00:00:00','   Listening to: Bone Thugs feat. Mariah Carey - Breakdown   Dude! I just got back from my week stay in Iriga.. Actually, we arrived this morning at like 2:50, and I didn\'t fall asleep until around 3-4. And that was after a 9 hour drive with three people in the back with me.. Not exactly the most comfortable trip I\'ve been on..  Anyway, New Years was ok. Tristan moped the most. Agh, it was so thing.. I didn\'t know what to say to him because I didn\'t know what he could do. I hate it when people bury themselves in depression. We didn\'t have any fireworks. Nor any torotots. We had a quiet New Years. And it was quite boring. We had dinner with the family at my Lola\'s house, but only my younger cousins and I stayed up to greet the new year. I wore stripes and jumped as much as I could. n_n;;   While I was in the province, I got my prom dress altered for only 50 pesos. 0_o That\'s less than a dollar. I also found a comfortable pair of prom-able shoes. So now all I need is a small bag and a shawl. But my mom wants me to wear a boa instead of a shawl. 0_o I guess that would be ok.. I\'d have to see it first. n_n;;  Not much happened in the past week. Just played RO and was lazy.. n_n;; Eck, I\'ll post tomorrow. I\'ve got school tomorrow.. >< I hope they don\'t show us our test results. I might die of a heart attack. '),(25235,950122,'29,February,2004','2004-02-29 00:00:00','   Listening to: Maaya Sakamoto ~ Dive   ~RANT~  ARRGH!!! EVIL INVADER!! XD  I lost 127 acres because of you! You jerk! XP You suicided on me! You made me lose 46 honor points!! My networth would have already been 200k!!! YOU KILLED MY PEZZIES!!! My poor defenseless pezzies!!! ;O; I swear, if you didn\'t have so many thieves, I would have stolen all of your money, food, and runes! JERK!  *punches imaginary figure*  ~END OF RANT~   urlLink Utopia   >_> '),(25236,950122,'26,February,2004','2004-02-26 00:00:00','   Listening to: The Saddest Song ~ The Ataris   YEAH!! Prom pics are out! XD First though, check me out being regular.   urlLink Nyaa~   Then the Prom Pics~!   urlLink Prom Pic 1~   urlLink Prom Pic 2~   urlLink Prom Pic 3~   urlLink Prom Pic 4~   I have such a crap scanner. XD '),(25237,950122,'25,February,2004','2004-02-25 00:00:00','   Listening to: All the People in the World ~ Some guy named \'Safri Duo\' I think   Phew. The previous problem was resolved. :) I\'m glad.  Today was weird. I stayed up really late last night because we had no school today, so I ended up getting out of bed at around 12 noon. I actually woke up at 9, but it took me three hours of initiative, going back to sleep and thinking of getting up to actually drag myself off the bed. I went online after eating lunch. At 4pm we went to Church because of Ash Wednesday. I almost passed out at mass. O_o I can\'t really remember it, but we were standing for quite a while, and my chest was starting to ache. Then I got really light headed, and I had to just sit down. Thing is, this isn\'t the first time this has happened to me. So now, my mom thinks that I\'m anemic, because she\'s anemic. Though it is a possibility, it doesn\'t sound right. I mean, just because I get light headed when I stand up for a long time doesn\'t mean that I\'m anemic... right?  But enough of that. The only way to really find out is for me to get a blood test. Yikes. '),(25238,950122,'20,February,2004','2004-02-20 00:00:00','   Listening to: Duncan Sheik ~ Barely Breathing   I don\'t know who is correct. I don\'t know if what I did was correct either. I did what I thought was right at the moment, but now I realize that a moment can pass by without you even knowing.  We all made decisions, but not all decisions we make are the right ones.   I\'m stepping out of the ring. This isn\'t my fight, but in reality, this never was a fight to begin with. I\'m sorry I blew this whole thing out of proportion; making it ridiculously worse. I\'m sorry if I coached you on, I\'m sorry if I put you down. I\'m sorry if what I did made you cry.  My dad was right when he said that this was childish. Friends shouldn’t be fighting over petty things. So I\'m letting this go. What you do is your decision. '),(25239,950122,'16,February,2004','2004-02-16 00:00:00','   Listening to: -    Nooo! I can\'t see my tag-board! >_> It won\'t load!! And I can\'t visit any forums because they don\'t load either!!  *looks at her modem which has conked out while she was typing*  Oh great. Here we go again. Damn you Destiny Cable! You don\'t even have Animax, now your internet service is biatching on meeee. x_x  *modem comes to life*  >_> Geez. '),(25240,950122,'14,February,2004','2004-02-14 00:00:00','   Listening to: TLC ~ Damaged   ~ Happy Valentine\'s/Single\'s Awareness Day!!! ~  I acted on my emotions today, and I realize that I acted way too hastily. I called my friend\'s guy...friend, an ass. Well, it\'s not really that bad, but I guess it was so sudden that even I was surprised.  I kind of regret it now; after talking to my parents about it, I guess I was too righteous. But still, someone needed to push down his pride. In my humble opinion, he was getting too high on himself.  Now, what\'s done is done. Can\'t change that. I just hope his groupies don\'t gang up on me. n_n; '),(25241,950122,'09,February,2004','2004-02-09 00:00:00','   Listening to: Avril Lavigne ~ Anything but Ordinary   I\'m soooooooo tired. I\'ve got to practice in Pinoi and music! >_> And I should start on my projects.. Pfft. I dunno, just feeling a bit tense right now.  Anyway, my parents brought my MP3 player to some shops, and it can\'t be fixed. Here at least. Ergh.. At least I told them about it. I haven\'t told them about my wallet tho. I\'m still hoping I\'ll find it. >_> Plus I don\'t want to get in trouble.  Meh, not much has happened lately.  Mirodo>> Eeek! It\'s taking so long for us to reply your to letter! x_x I hope you\'ll start feeling better soon... '),(25242,950122,'01,February,2004','2004-02-01 00:00:00','   Listening to: Ghost in the Shell : SAC ~ Inner Universe   Wow. Fair is over. Let\'s recap.  First day was alright. There wasn\'t a lot of people yet because it was Friday, and everyone\'s bolyets still had school. I didn\'t do much until the Art Attack started. We had to wait a while because it didn\'t start on time. >_> By the time it was our turn, it was soooooooo hot! x_x We had to do an art attack about Osama Bin Laden. The other sections got F4, Uncle Sam and Sadako/Samara from The Ring. We were only given 10 minutes to do it; 7 minutes to make it, and 3 minutes to clean up. Thankfully we actually finished it, but afterwards I was so tired... x_x I went to the MMA to rest, then I found out that my MP3 Player broke! > _> After that we went to the Baila. It was ok. Camilla and Racine slept over. I didn\'t sleep until 3am.  Second day was ok at first. I promised myself I\'d be more careful with my things after my MP3 Player broke, but lo and behold, I lost my wallet that day. I don\'t care about the money that was in it; I just want the wallet with my friends pictures, my smal keepsakes.. Shit. Fredrick and some other AC guys went to the fair that day. They didn\'t stay for the Band Aid tho. I also missed the SHS Chorale concert because I didn\'t have any money.  Later on in the evening we watched the Band Aid. Was ok. I slept through part of it, but we stayed \'till the end. ._. Most of the people left after Rivermaya presented. Yeah, it was insensitive. But we stayed. :D  Went home afterwards, and slept. '),(25243,950122,'31,March,2004','2004-03-31 00:00:00','   Listening to: Tears ~ X Japan   Well, it turns out we\'re going to Iriga! n_n I won\'t have to back all to myself tho; my mom\'s friend from Holland just arrived today with her eldest son, and they\'re going with us. It\'ll be a crowded 9 hour trip. -_-;;  I\'ve been reading Naruto lately; I\'m starting to like it a lot. :D The story is pretty cool, and there\'s lots of fights, so yesh.. :P   ANYWAYS! I\'ll be gone for maybe 3-5 days. Aww. Till then! :P '),(25244,950122,'29,March,2004','2004-03-29 00:00:00','   Listening to: The Reason ~ Hoobastank   Mom\'s home now. :) She\'s going to a different hospital for another check up tomorrow tho. =/  I\'ve been drawing a lot lately. I have a lot of time on my hands now because it\'s summer. I should be playing RO... XD I\'m worried tho. I keep on wondering if I have to go to summer classes. O_o I\'d kill if I had to.. But my parent\'s would kill me first. >  I haven\'t been eating lately. O_o I\'m not on a diet or anything, I just haven\'t been really hungry lately. I skipped lunch today, and I\'m still running. Makes me wonder what I\'m running on.  I need to make a new layout. =/ '),(25245,950122,'27,March,2004','2004-03-27 00:00:00','   Listening to: Background Music ~ Internet Cafe   *meant to be posted yesterday*  We just finished a week of hell today. But my mom\'s in the hospital, and she\'s going to have an operation tomorrow morning. I\'m worried. My dad won\'t tell me anything. I\'m staying at Camilla\'s house because he\'s going to stay in the hospital and there will be no one in the house with me.  I\'m scared.  *today\'s post* I just found out this morning that the surgery will be on one of her ovaries. They found a growth, and they\'re going to remove it this afternoon. I pray everything will be alright. '),(25246,950122,'21,March,2004','2004-03-21 00:00:00','   Listening to: Believe in Heaven ~ Miki Shinichirou   Feeling weird lately. =/ Lot\'s of stuff on my mind.  I\'ve been thinking about summer. Right after my tests we\'re going to go to our province in Bicol, namely Iriga City. We\'re going to stay there for roughly 10 days, then we head back to the city. Then we leave for Virginia on April 14 or 16, I\'m not really sure. We\'re going to stay there for about a month I think. O_o We head back in May, but we\'re still going to have a stop over in Hawaii, where we\'ll stay for about a week. Then it\'s back to the Philippines. Phew, this summer will be pretty busy. I\'m a bit scared tho. With all the terrorist attacks happening, I can\'t help but think... Nevermind. I\'m sure you get my drift.   I\'ve also been thinking about my future. I\'m seriously considering going into the ROTC Scholarship for college. I join, then they pay for my college, and they also give me an allowance. But I have to train in ROTC. Plus, the course has to be something useful to the Military, like communications, or international/asian studies, medicine or whatnot. =/ But after college, I\'ll automatically have a job in either the Navy, Air Force, or Marines for 4-6 years. I\'ll probably be stationed somewhere, especially if it\'s the NROTC(Navy ROTC), maybe Europe, or Asia. I dunno. After that time, I can choose either to stay, or quit. *sigh* Seems like so much to absorb.  It weirds me out. I can\'t imagine myself all grown up.  Maybe my sudden serious thinking is because I fell down the stairs this morning. I was sprawled on the landing, and I laid there for maybe 3 minutes. Our house is dangerous! x_x Our stairs are wood, and my mom waxes and cleans it to no end! That\'s why so many people fall down the stairs in our house. >_> '),(25247,950122,'18,March,2004','2004-03-18 00:00:00','   Listening to: Into the West ~ Annie Lennox   I\'ve been pretty sluggish lately. I can\'t wait for summer, but I still have to drag myself through the last week of school ( which is next week!) I\'m a bit worried tho. Like one of our teachers said, \"You guys feel like it\'s summer already, and end up neglecting the last weeks of school.\" I don\'t want to end up getting stuck in summer classes!! > '),(25248,950122,'14,March,2004','2004-03-14 00:00:00','   Listening to: Whre the Sky is High ~ Yuki Kajiura    urlLink     Which Great Teacher Onizuka character are you most like?     brought to you by  urlLink Quizilla    Isn\'t this the stalker guy on GTO? :P  Oooh, I\'m cramming again.. I really need to be more responsible. =/ Oh yeah, does anyone know where I can download the font used on the quiz pic?  ________  I just want to reflect on what happened in Madrid, Spain. 200 people dead, just like that. How do those people live with themselves? I\'m so sick of hearing about people killing other people that just  thinking  about it makes me want to vomit. But more than anything, it scares me. And I hate being scared. '),(25249,950122,'10,March,2004','2004-03-10 00:00:00','   Listening to: Electric Fan ~ Rota Aire   Wow! After five looong years, we finally get a phone line! What a surprise! But now that I\'ve got a phone line to chat up my friends, I don\'t feel like using it. I\'m too used to just using my cellophone to text my friends and classmates and chatting with them on YM. Haha, the irony of it all. '),(25250,950122,'08,March,2004','2004-03-08 00:00:00','   Listening to: Electric Fan ~ Rota Aire     ~SORCEROR/SORCERESS~  You are a Sorceror or Sorceress, and the world bends to your will.  You can work miracles, and people often comment on your magical and/or mysterious ways.  Hopefully you\'re using your powers for good...    urlLink   What Fantasy Role Are You?     brought to you by  urlLink Quizilla '),(25251,950122,'06,March,2004','2004-03-06 00:00:00','   Listening to: Pasha ~ Vanessa Mae   Happy Birthday to Camilla and Gretchen!! :D :D :D  I slept over at Camilla\'s yesterday! We watched the first six-eight episodes of Sailor Moon (Live Action Drama Series), and we just soo got into the parts. Now I\'m Sailor Jupiter. XD We went to Tomas Morato afterwards; it\'s this trippy food strip and ate at Teriyaki Boy. We got to ride in their new car!! It\'s a Chevrolet!! Cooool. Teriyaki Boy was this cool Japanese Restaurant with a call button for the waiters. They were also playing japanese songs. After eating, we saw one of the players of the Ateneo basketball team. We was like, 6\'3 or something! Geez!  We went back to Camilla\'s house after that. We watched more SM LADS. :)) Gela and Karina went home soon after, and we went to sleep. This morning tho we watched like, 6 episodes of .hack//SIGN. Wahaha. I went home at 12 noon. Fuuuun. :D  Oh yeah. I can go to forums again!! And I can see my tag-board! Happpoooy! ;P '),(25252,950122,'03,March,2004','2004-03-03 00:00:00','   Listening to: Yantra ~ Vanessa Mae   Ack.. Summer has officially begun here in the Philippines. I can feel it. Literally. Heat wave! >_> But of course, with summer comes the end of school! :D And the beginning of review classes for the college entrance tests. =/ I won\'t be taking the review classes. One reason would be because we\'re visiting my brother in Virginia this summer. He\'s already been there for over six months, so I\'m actually looking forward to seeing him again.  Hmm, not much has been happening lately. Nothing of much interest. Just school work and projects. *sigh*  PS>> Taira, my dear childer on the Vampire Site (whoever you are...) just e-mail me for my YM id. :P '),(25253,950122,'27,April,2004','2004-04-27 00:00:00','   Listening to: My cousin snoring. Kekekekeke.   A few days ago, we went to Atlantic City, New Jersey, then New York. Then just recently we went to Luray Caverns. Here\'s the scoop:  Atlantic City  First, we went to Atlantic City to celebrate my uncle\'s 24th wedding anniversary. The car ride was uneventful. (Mainly because I slept thru a lot of it) We drove through Virginia, Maryland, Delaware, Pensylvannia, and NJ. O_o I\'m not sure about the order. When we got to NJ, it was so cold! O_o Even colder than Virginia! XD After settling in our rooms for a bit, we went to a casino owned by Mr. Trump.. The guy from the reality tv show \'The Apprentice.\' Of course, my cousin, my brother and I weren\'t allowed to go to the casino. So our parents gave us 20 dollars each to spend in the arcade. XD We  only spent around 5 dollars each. But still we had fun. Our parents had fun also, but they either lost some money or broke-even. Surprisingly, someone hit on mom. Again. O_O  Anyway, we were supposed to go to the boardwalk afterwards, but it was raining, so we just went straight to the Inn. They had an SNES Emulator on the TV (7$ per hour), which my cousin and I played until we fell asleep.  New York  The next day we drove to New York! We got lost on the way there, then we got stuck in morning traffic. XD We went to Times Square, and we saw the Naked Cowboy! :D :D :D It was so cool to be there, even if there were so many people. But when we were looking for a parking space, this guy tried to charge us 44$ for an hour!! O_O That set us on a bad mood. But we cheered up after eating at this small Italian Resto. The owner was impressed at auntie\'s Italian, so he gave us a hefty discount. n_n Afterwards, we walked 12 blocks to get to the Empire State Building! x_x But when we got there, we found out that we had to wait an hour just to get inside. >_> So we chilled at a Starbucks. I drank the Strawberries and creame.. Yum. We walked 12 blocks back, and we came across a comic shop! I was in heaven, checking out all the comics and manga, but I didn\'t get to buy anything. *pout* After that, we drove home.   Luray Caverns  The day after that/yesterday, we went to Luray Caverns! The drive there was very serene, because we drove through the countryside. I saw horses, cows, and barns. The cavern was reaaally fun. There was this one part where there was a puddle of water that reflected the ceiling perfectly. It was so beautiful. The deepest part of the cavern was 160 ft, where there was an organ-thing which used the stalagtites to make music! O_o It used rubber plungers to lightly tap the stalagtites to create notes. It was wonderful to hear. Then there ws a wishing well as we were going up. I didn\'t have any US coins on me, so I threw in two 5 peso coins. XD  When we got out, there was a garden maze nearby. So my cousin and I went through it. (We got discounted because the registrar thought I was below 13 years old. x_x) There were gnats everywhere, but we had a lot of fun running around the maze and finding the 4 goals. There was this one point were we met up with 3 guys who started whistling when they saw us. O_o We eventually found our way out, with four goal stamps. After that, we drove home, and again, I slept through most of it.  Today  Nothing much happened today. We went to Jeth\'s college to fix his application for summer classes. I saw my first squirell! XD We stayed in McDo for nearly an hour and a half, and went grocery shopping after dropping Jeth off at his dorm. My cousin and I dug up her old NES, and now there are three screens in her room. n_n;; '),(25254,950122,'21,April,2004','2004-04-21 00:00:00','   Listening to: Ticky Tock went the Clock   I had a long entry last night, but the internet connection biatched on me and I lost it. n_n;;  Anyway, I\'ll try to remember what I wrote. I\'ve only been here for five days, but it seems like I\'ve been here for a long time already. I miss my house. And my room. And how I could stay online all day. XD But here I have to use it late at night. It\'s funny tho. This vacation feels like it\'s trying to pry me away from my online persona, but here I am, clinging to dear life. Lol!  All we\'ve been doing lately is going to different malls (which are really big and nearly deserted, compared to the Philippines at least) and ending up not buying anything. =S But I\'ve already gotten some stuff. I got new shoes, a manga (Zodiac PI... it was an impulse buy), and a frog lamp... I also got a jr bow and arrow set. XD But along with my new stuff, these horrible little sores developed around my lips... >  The weather has cleared up since my last post. My last post it was raining and the wind was blowing in all directions.. but now it\'s sunny; you only need a light jacket. Well, I need a light jacket at least.  Lots of our plans have changed. I don\'t think we\'re going to Canada anymore; the uncle we were supposed to stay with is going to the Holy Land and to a tour of Europe for his wedding anniversary. The uncle we\'re staying with now is having his wedding aniv. this week, and we plan to go to Atlantic City! O_o  Another one of our plans have changed also. I might not go to college in the US afterall. '),(25255,950122,'15,April,2004','2004-04-15 00:00:00','   Listening to: Full House (in the background)   I\'m finally in Virginia!! But it\'s so cold!!! And I only have one sweater!! I feel so misinformed... >  Wow... I\'m so happy to be here. XP I never thought I\'d make it. But the 12 hours jet lag is going to get me. >  Ok, I\'ve got to go and find a heater... My feet are freezing. XD '),(25256,950122,'12,April,2004','2004-04-12 00:00:00','   Listening to: Black Balloon ~ Goo Goo Dolls   My journey to Virginia starts tomorrow! >  First we go to Virginia to visit my brother, then Canada, then Hawaii! =P I\'m really excited. I\'m a bit scared tho. =/ I\'m praying that we have a safe trip.  Belated Happy Easter! n_n;; I hope you enjoyed your chocolate. I didn\'t get any. *sob* '),(25257,950122,'09,April,2004','2004-04-09 00:00:00','   Listening to: Dive ~ Maaya Sakamoto   I\'m sooo tired. Why?  I watched \"The Passion of The Christ\" today. I cried through more of half of the movie. I guess that\'s what tired me out the most. Thinking about it also has left me sapped. But what do I think about the movie? I can\'t really say. Mom doesn\'t understand why Jesus had to suffer so much, and why God let him suffer. Dad on the other hand feels angry overall. He was especially angry at the Pharisees. As for me, I really don\'t know how to react. I feel guilty, yes, but not as guilty as I thought I would feel.  I wonder. If I was in the shoes of the Jews, would I believe in Jesus? I wonder if I am strong enough to stick it out with him, or would I deny him like Peter? I compare it to having a person with a gun come to me and ask me if I believed in God. There are only two answers. What would I do? Of course, I can\'t say; it\'s one of those things that you\'ll only know if it were to happen. And of course again, I would rather not have that happen.   Right now, I have a headache. Crying really does tire me out. '),(25258,950122,'08,April,2004','2004-04-08 00:00:00','   Listening to: Misfits ~ Third Eye Blind   When I opened the newspaper and checked the movie schedule, all I saw showing was \"Passion of The Christ\". The first thing that popped into my mind was,  \"Why do these people want to watch it in a movie theater?? They\'re obviously going to be fazed out afterwards, plus they\'ll be crying through out the movie.\"  Yes, this is the perfect time to show it because it\'s Holy Week. In fact, we already have a pirated DVD of it. But I ask myself; am I ready to watch it? A comment that I\'ve heard about it is,  \"Agony in the Garden pa lang, magi-guilty ka na.\"   LT: \"You\'ll only reach the Agony in the Garden, and you\'ll already feel guilty.\"  Agony in the Garden is in the beginning of the movie. So I\'ve been thinking. Am I ready to wake myself up from the illusion that is life? Afterall, I believe that Jesus died to save us. Am I ready to realize first hand how much in debt I am to Him? Seeing it first hand is different from reading it. To actually see a person, even more so Jesus, being tortured and beaten to high hell is agonizing. But what more to know that the person did it all for you. I have mixed feelings because yes, I do love Him, but I also rely on myself a lot, often forgeting about God.  It\'s like slapping yourself and saying, \"You\'re only human.\" Which we all are. '),(25259,950122,'05,April,2004','2004-04-05 00:00:00','   Listening to: Opening ~ Trigun   Hmm. I\'m back from the province. Eating Buko Pie at the moment. :P  My stay was fun; all we did was play video games. We didn\'t get to play RO tho because up till now it\'s down. We played Tenchu: Wrath of Heaven, Guilty Gear XX, Robotech, Twisted Metal: Black, and I even started playing Diablo 2 on Tristan\'s computer. XD '),(25260,950122,'30,May,2004','2004-05-30 00:00:00','   Listening to: The various sounds of the Internet Cafe   I never thought this would happen, but my computer died! :( A virus attacked and killed it!! So now I\'m in an internet cafe.. *sigh* My brother is fixing it now, and he says it\'s awful. And I\'m much pissed because everyone says it\'s my fault. My family has blamed me for putting the computer into a coma. But honestly, I haven\'t downloaded anything! >   I hope my brother can fix the computer without losing all my stuff. '),(25261,950122,'28,May,2004','2004-05-28 00:00:00','   Listening to: Fiona Apple ~ Across the Universe   After roughly 24 hours in the air, I\'m back home, sitting in front of my spy-ware ladden PC. *sigh* Will format later. Anyway, the flight back from Honolulu was horribly tiring. We were delayed an hour, because the plane wouldn\'t fly. O_o After 7 hours, we arrived in Narita Airport. We saw former president Ramos there. =P I wanted to stay longer there, but nay, we had to go home. I missed it much.  On to spy-ware... I hate the horrid things. Trojans, key-loggers, hackers. Gah. I\'d chop their hands off at the chance. *chop chop chop* XD '),(25262,950122,'25,May,2004','2004-05-25 00:00:00','   Listening to: TV in the background   We went to the Arizona Memorial a few days ago. It was really depressing there. Before we actually went to the memorial, they showed us this video clip/short documentary about the attack on Pearl Harbor. It was so sad to watch. I can\'t imagine getting caught in the middle of an attack like that. The thought scares me.  Recently I went to Borders! Borders is a bookstore here, and it\'s paradise. Soo many good books to read, and they don\'t mind if you get a book and plop yourself down on the floor somewhere. I could stay there for a day and be happy. *daydreams*  I\'m coming home on wednesday. I have a lot to catch up on. '),(25263,950122,'21,May,2004','2004-05-21 00:00:00','   Listening to: The wind in the trees...   I wrote this while I was flying to Hawaii:  My last day in Virginia was pretty boring until Vanessa came home from VTech. We talked mostly about buying Jeth\'s car. Tee hee. After a while, Patrick, one of Jeth\'s first friends when he got to Virginia, came over. While he and Jeth were looking at pics he took, Maelle and I were in Vanessa\'s room, messing around with her uniform hats and canteens. We had so much fun! Patrick ended up stayin for dinner, and after dinner we went out to Baskin Robbins to buy ice cream. Vanessa drove Jeth\'s car, which was manual, and it\'s probably by the grace of God that I\'m still alive today. Jeth and Patrick, the \"Stick Masters\", were supposed to teach her how to drive manual. When we finally got to Baskin Robbins, we each got lots of ice cream. Yum. =) Surprisingly, Maelle finished hers last. :O We got a gallon of gas, then we ran an errand for my aunt. We were supposed to just buy dishwashing liquid, but we ended up staying there for a while. Vanessa, Maelle and I were just staring at all the facial products there were. O_o When Jeth finally called us to find out where we were, I expected them to be in the electronics section or something, but lo and behold, they were in the candy section. XD  Jeth drove us home from there. It was a pretty smooth ride. We watched American Idol while Jeth fixed Maelle\'s computer. Jeth left his computer to her. *pout* Oh well. We played cards afterwards. Texas Holdem, Egyptian Poker.. well, I forgot the names. Patrick left at about 10pm. I wonder if he\'ll visit us in the Philippines.  At the airport today I tried to convince Uncle Ernie to visit us in December. I threatened to spam his e-mail if he didn\'t. XD  Well, I\'m off to Honolulu. I wonder if Chloe still remembers me.. *shudder* '),(25264,950122,'18,May,2004','2004-05-18 00:00:00','   Listening to: Opening of Roswell ~ Dido   It didn’t work out. Jeth’s coming home with us. It was a big, expensive gamble, and we didn’t win. So, he’s coming home with us now. He’s leaving his computer to Maelle, but he’s taking the monitor, the hard disk and the DVD writer. I really don’t know how to react. My dad keeps on asking me for my reaction, and I seriously don’t have one. All I have is the anxiety that the same thing can happen to me. '),(25265,950122,'16,May,2004','2004-05-16 00:00:00','   Listening to: The Disney Channel   I\'m really homesick. I miss my friends. I miss my room. I miss my internet connection. I miss Photoshop. ;O;  Today we went to watch a movie. We got there early tho, so I was able to go to a comic shop! I bought the first manga of Gravitation. :P At the theater, my parents and my brother watched Troy, while Maelle and I watched Mean Girls. It was pretty funny. Our movie ended an hour before Troy did, (we even started before they did) so we had to get picked up early. After that, I hung out with Maelle\'s neighbors. They had a bonfire, and it was pretty cool, considering bonfires I\'m familiar with usually has my dad burning a grass field near our house. :O Anyway, we had smores and watched them play Mortal Kombat 3. SNES. Hmmm.  We leave for Hawaii next week, May 19th. Then we go back to the Philippines probably on May 26th-28th, and before I go back to school, we\'re going to Iriga! O_o Still a lot of travelling in store for me.. '),(25266,950122,'07,May,2004','2004-05-07 00:00:00','   Listening to: Recess ~ Disney Channel   We\'re going to an amusement park! Yey.  Not much has been happening.. Just driving around and stuff, trying not to buy stuff. I\'m really worried about my future. I don\'t know anything about what I want! >  Then, there\'s the Presidential Elections in the Philippines. O God, please help our country.  And, there\'s the War in Iraq. I believe in the cause, but it\'s so hard because moral is low... and I know people who are there,  right now .  The Iraqi prisoner torture, the ambushes..   It\'s just so SCREWED!! I hate how everything seems to be so damned screwed. I hate how ass presidents are elected, how people die over stupid beliefs, how evil always seems to get back at the good. It\'s so horrible, it makes me angry and sad at the same time.  I hate how human we can be. '),(25267,950122,'29,June,2004','2004-06-29 00:00:00','   Listening to: Misfits ~ Third Eye Blind   Hm. >_> There really shouldn\'t be any school tomorrow.  I think my CL teacher has some sort of grudge against me. O_o She gives me evil looks, and always looks over my shoulder, like the devil is suddenly going to spawn from me or something. I really would care less, but she\'s my class adviser.. >_>  Gah. I know I promised myself I\'d make this year count, blah blah blah. But you know me. It\'s back to my old tactics. n_n;; Gaaaaah. I suck. =P '),(25268,950122,'24,June,2004','2004-06-24 00:00:00','   Listening to: Rage Against the Machine ~ Killing in the Name of   I\'m so stupid. Everytime someone trys to get close, I put up a front. I step away. I run. It\'s so stupid.. In my mind I\'m so liberal and \'free\'. What the hell am I thinking?? I\'m just full of hot air! Once I step out into the real world, lo and behold. I freeze. Nothing is real. I\'m such a coward!  I\'m like a Mime. I\'ve put myself inside an invisible box. But I can\'t get out. '),(25269,950122,'22,June,2004','2004-06-22 00:00:00','   Listening to: Faster ~ Third Eye Blind   I have music again. n_n I really need to buy their album. Also Portishead. *nod*  I just finished .hack//SIGN today. I was thinking. What if later on ni the future, \'The World\' really came to be? Or at least something like it? Not necessarily with evil entities and wishes and hacking, but more of a completely 3D MMORPG. Of course I\'d play it (or at least try to); there\'s no question. I was wondering tho if I\'d actually lose myself to it. I wouldn\'t be surprised.  Online personas are like second lives. That\'s what I think. I mean, when I go online, I can be a completely different person. I have different friends, different opinions, different attitudes. There\'s me, and there\'s online me. I love going online. I\'m part of many online communities, forums, sites, and I\'ve met so many nice people through this. But I\'ve also met my own share of arses. That doesn\'t hold me back tho. It\'s like, going online is like an RPG game. Going out, meeting people, interacting. You can be any hero you want. You can be you, you can be typical, you can be neutral, you can be misunderstood, you can be plain ebiiil. Maybe it\'s this about the internet that has me so drawn to it: Being an unmasked (yet masked) person you can\'t (or plain aren\'t) be in real life. I admit, I do not have a nice, thriving social life. I keep in touch with more people who live in different countries than I do in my own. Honestly, I feel so spaced from the people in my own country than with those elsewhere. Yeah, it worries me like hell, and it\'s horrible when I know nothing about those close to me. I feel so.. different. And I space myself from others because I feel so close, yet so far. Physically, I\'m here, but socially, I\'m somewhere else. It\'s stupid, now that I think about it. Sometimes, you just want someone to talk to, but you don\'t know who. '),(25270,950122,'15,June,2004','2004-06-15 00:00:00','   Listening to:   I\'m feeling much better than a few nights ago.   We had our first class of sex education today. The akwardness. We\'re not allowed to talk about abortion, contraceptives, rape, ect ect ect tho. We have to look at it from a Christian point of view. It\'s not that I oppose this way of seeing things, but it\'s sort of like sheilding us from the truth? I think I expected it to be more liberal. '),(25271,950122,'11,June,2004','2004-06-11 00:00:00','   Listening to: Nothing   I need a hug.  Lately, when I get home, I feel like I\'ve been completely trampled underfoot. I\'ve been filled with thoughts that run thru my mind at blinding speed. It\'s crazy. I think I\'m going crazy.  I saw a woman this afternoon. She was walking down the street naked. As the car sped by, my heart. It twisted. It moaned and creaked. I didn\'t dare to look back. Someone please just stab me repeatedly! Stab me in the back and leave me to die.  I hate how so many things can make me feel bad about myself.  So many things make me feel so frustrated. Sometimes up to the point where I can\'t stand it at all! My mind goes into ultra frustrated mode, and I go crazy! I just have to get away, just get away, but I end up in my room where I push the frustration down, and I go back. I never ever get rid of it, I just stuff it under my bed. I can never get rid of it. It\'s always there. '),(25272,950122,'08,June,2004','2004-06-08 00:00:00','   Listening to: Nothing~   I\'m trudging back to school tomorrow. Is it just me or did summer pass by so fast? It\'s prolly just me. Hmm. I\'m worried. And excited. >  I\'ve been reading romance books lately. O_o So not my style, eh? But to hell with style. I\'ve never had much style anyway. Hihi.  I saw this on Gela\'s bloggie (Lost Karaoke Machine). I do like what turned up. :D      N     Naughty       A     Awkward       Y     Yum       O     Odd       M     Meek       I     Intelligent       Name / Username:       urlLink Name Acronym Generator  From  urlLink Go-Quiz.com '),(25273,950122,'02,June,2004','2004-06-02 00:00:00','   Listening to: Myself Humming   It\'s raining outside. I love it and I hate it.  I bought my books at school today. I sort of missed my school. The fact that I\'ll be graduating next year still hasn\'t settled in yet.  Since my comp died, I\'ve had a lot more time to myself. Good when reading. Bad when not. Like I\'ve said before, I think too much.  I just realized... once I get my braces, I\'ll be the dorkiest girl in school. Hmm. Not like I care. Not like it matters. '),(25274,950122,'27,July,2004','2004-07-27 00:00:00','   Listening to: Puddle of Mudd ~ Blurry (Acoustic)   When two friends fight, what should one do? It sucks to get stuck in the middle. It hurts to see people who aren\'t even part of the whole thing (no matter how small it may be) get affected. It\'s hard watching&nbsp;two of your closest friends fight. So I say this.  For God\'s sakes, GROW UP! You\'re both 16 for crying out loud, do you still need us to fix things up for you two?&nbsp; Nagpaparinig sa blog;  Yelling at each other thru YM Stats...&nbsp;Well&nbsp;hot damn, talk about maturity.&nbsp;Talk to each other! It\'s so frustrating; it makes everything so&nbsp;hard! What you do affects all of us..  ...or did you forget that in the spur of the moment? '),(25275,950122,'23,July,2004','2004-07-23 00:00:00','   Listening to: The same song as last night, and the night before that.   I have braces now. O_o  But besides that, I haven\'t said anything about my retreat to Tagaytay with my class. Overall, I had a very nice time there. It was fun bunking with everyone in one room, even if it meant being flashed by a classmate, or talking about things we really shouldn\'t have been talking about. *coughsexcough* XD XD Of course, our facilitators weren\'t very encouraging either. Brother Pao and Brother Bernard were back with a vengance, and this time with a new sidekick, Brother Mike. And yeah, they acted as gay as ever. But what the hell, it was still fun. The food was great too. :D Good thing we only stayed for three days, or I seriously would have gained weight. The bathroom. Ah. Many things happend in the bathroom. It was the place where we all met one way or another. My classmates actually gave each other tips on how to take a poo. O_o I dubbed it, B Cubed, or Bayanihan\'s Bathroom&nbsp;Bonding.&nbsp;It needs a better drainage system tho. The floor was always wet.  Anyway, the programs and activities were alright. I accidentaly fell asleep through part of the most important activity. O_o Bad Ria, bad! But I woke up just as the interacting part began. T_T I\'m such a bad person.  But there were worse people than me that day. 14 unsurrendered cellphones were confiscated just as the Retreat ended, all because one of them text the incoming section. I heard they\'re now faced with Supervised work, and possible suspension. It\'s a good thing my mom made me feel guilty about bringing my phone, or else I wouldn\'t have left mine at home.  The way home was cold. Exremely. O_o I couldn\'t feel my fingers. Seriously. I had to share Guia\'s jacket with Racine to keep alive in the bus. The two sitting behind us actually brought out a blanket. O_O Lucky them.   That\'s pretty much the retreat. n_n; '),(25276,950122,'18,July,2004','2004-07-18 00:00:00','   Listening to: Dreamless World ~ Siam Shade        &nbsp;   Guess what? My internet connection is down. Again. OIHJRLKNSDFOHIOF.     Anyway.. Ouch. My mouth hurts. I got my spacers. And. They. Suck. O_o It\'s like this dull pain in your mouth; I can barely chew. I get my braces next week. *cough cough*     Ah, there\'s the internet again.     I went to a soiree with Camilla and Racine yesterday. Honestly, at first, it was really boring. Then it became ok, then it got out of hand, then it got scary, then it was back to ok. Sorta. XD Somethings are better left unsaid. So ANYWAY! I\'m glad I went. Hihihi. *rolls eyes* XD     After the soiree Racine and I slept over at Camilla\'s house. Talked about the soiree, and fell asleep. :P The next morning, I could barely get out of bed. XP But of course, Camilla was already up. Racine was still snoozing. Hihi. Then after a while, Camilla came up to tell us to eat, and yeah we ate. We also watched Animax, and critisized the dubbing. It was horrible. More horrible than the defense of the panty in RO. Then Racine and I watched Jiro and Matthew kill each other in Tekken, witnessed a half nekkid Din Din and Carl, and pondered the meaning of life on Camilla\'s bed. We finally got to the computer shortly after that, and we got one slide finished! :D :D :D Yay for progress.     Well, I\'ll be gone for three days; Monday to Wednesday. I\'m going on a religious retreat with my classmates in Tagaytay.&nbsp;W00tness. \'Till then I guess! =P   &nbsp; '),(25277,950122,'11,July,2004','2004-07-11 00:00:00','   Listening to: Duvet ~ BoA   Haven\'t been feeling very well lately. Really been out of it. I don\'t know. Since last week, I think. My appetite has shrunk from a BigMac meal with large fries and a large coke to a dinky Cheeseburger meal. I wonder if there\'s something wrong with me?  Chuu. I read \"Tough Love\" by Melissa Salva. Was very..  kilig-ful . =P \'Tis fun being a closet hopeless romantic. '),(25278,950122,'04,July,2004','2004-07-04 00:00:00','   Listening to: Good Man ~ Third Eye Blind   Remember how I said that I go crazy when I get frustrated? I think I\'ll be going to the Mental Ward very soon.  Everything is really starting to get to me. '),(25279,950122,'07,August,2004','2004-08-07 00:00:00','   Listening to: ~     UPCAT was surprisingly easy. O_o I didn\'t study tho, so I couldn\'t answer everything. Especially the questions on genetics. I was like: @_@;;; But whatever. I don\'t feel like I took the UPCAT. Honestly, we all mentally and physically tortured ourselves over this test, and it was like.. easy.  Well, here\'s a list coz I\'m bored! :D  The ones that are  bold  are the ones that I\'ve never done before:  I\'ve never been drunk.  I\'ve never smoked pot.  I\'ve never kissed a member of the opposite sex. ( :O ) I\'ve never kissed a member of the same sex. ( :O to the nth power )  I\'ve never crashed a friend\'s car.  (My friends don’t have cars. Yet.) I\'ve never been to Japan. ( Lived there for more than half of my life... XD ) I\'ve never been to Europe. ( Vacation. =P ) I\'ve never ridden in a taxi. I\'ve never been in love.  I\'ve never been dumped.  I\'ve never shoplifted. (Ho hum..) I\'ve never been in a fight. I\'ve never snuck out of my house. I\'ve never pissed on myself.  I\'ve never been arrested.   I\'ve never made out with a stranger.   I\'ve never been on a blind date.  I\'ve never lied to a friend.  I\'ve never had a crush on a teacher.  (It’s true people. XD) I\'ve never skipped school. (Surprisingly, no.)  I\'ve never thrown up in public.   I\'ve never purposely set a part of myself on fire.  (Surprisingly, no. XD) I\'ve never eaten sushi. I\'ve never been happy with myself.  I\'ve never met a movie star.  (I met Duncan Sheik; he’s not a movie star, but he’s still famous. :D) I\'ve never cheated in any test/exams.  I\'ve never confessed my feelings for anyone before.  I\'ve never run away from home. (I ran to the park, and it was only for an hour.) I\'ve never been on a stage.  I\'ve never had my heart broken.   I\'ve never been on the dean\'s list.  I\'ve never colored, permed, hot oiled, salon treated my hair. (Can anyone say.. 10 years old? XD)  I\'ve never killed anyone.  (I certainly hope not! O_o)  I\'ve never bungee jumped/rock climbed.   I\'ve never completed reading the Bible.   I\'ve never attempted to kill myself.  (Hurt. But never kill.) I\'ve never dreamed of becoming the President, or any government official. (Quote from Childhood: \"Candycanes for everyone!\") I\'ve never seen a real kangaroo. (Koala’s are cuter.)  I\'ve never written a [published] book.  (That\'d be soo cool.)  I\'ve never completed watching Meteor Garden.  I\'ve never saved anyone\'s life. (Hey, I\'m still alive.) I\'ve never saved money above P1000.  I\'ve never imagined my wedding day.  (With who I tell you?! XD) '),(25280,950122,'04,August,2004','2004-08-04 00:00:00','   Listening to: Break Me ~ Jewel   I\'ve been so swamped with school.. It\'s crazy.  I\'m taking the UPCAT this Saturday.. I\'m scared, in a not so caring way. T_T (BTW, the UPCAT is a college entrance exam. Prolly the hardest, not only because of the test, but also because of the sheer competition) Then the week after the UPCAT is exam week! Haay. If I didn\'t study before, I don\'t know how to compare how I dont\' study now. It\'s not that I really don\'t care, it\'s just like.. I don\'t have any motivation. No dream. No goal. I dunno. It\'s like I\'m letting myself get blown about by the wind? Where ever the wind may blow.  I wish I knew what I wanted from life. '),(25281,4176168,'21,August,2004','2004-08-21 00:00:00','   I\'m melting!!!...  Not really...but I feel like hell...staying up till 2 AM in the morning and waking up at 8 the next morning is not a good idea, I\'ve been feeling tired the whole day, the only way keeping me sane...is probabely coke and work...very tired indeed.  Got a call from my traveling parents, and from what I can tell, they have been doing way to much driving and way too little sleep. It looks like it takes about 10 ours to drive to Quebec from Toronto, and when they finaly got there, they had trouble finding a place to stay...very troublesome, and the hotels there is very expensive...$120 a night...  Err...as I said in my last post, I finished the handbook last night and sent out the version 3, well..today I again got e-mail replays. But only 1 with a very minor change, so I guess that made my happy enough...so all I need to do is the add the cover and schedual and I am done. But printing present a problem once more, because I do not know how to print it in a book fashin...err...and they want to see hard copys tomorow...got to get this problem figured out...and make sure I don\'t screw up...  The stress and the deadline have been deadly...after I present this thing tomorow...I\'m taking a couple of days off with this and going to enjoy my self and camp that\'s coming up in a week...and actually take a break during my vacation...  Anyways...going to go back and finish the handbook and print it correctly (hopefully) and get to bed before 1 in the morning...  Later Days  \"Most people who are old enough to know better often wish they were younger enough not to\" - Bob Edwards  Neo...  '),(25282,4176168,'19,August,2004','2004-08-19 00:00:00','   What a long day it hav been...  Up until tonight...it was a good day. Remeber the draft that I finally sent out?...today I got about 10 e-mails back saying that there is something wrong with it that should be changed...which I could have done easily until I relized the massive changes that these people have made. Which have brought me here...2 o\'clock in the morning writing this post with my last bit of energy...the whole change took me about 4 hours to do, so I have been working from 10 last night till today 2 in the morning...I am tired...but at least it\'s done. Who knows? Maybe tomorow when I wake up I\'ll face another 10 e-mails...oh well...this is the part I choose to do and I\'m going to stick with it, and all in all, I am the best....or so I think ;)  Well...the life without parents is...so far very nice, I have been serviving and been having some fun, so I guess I am kind of enjoying this.  I never get a break, after all these hard work, I can only sleep for a couple of hours because I am giving a piano lesson 9:30 in the morning...I don\'t usually do these things this early, but oh well...and in the afternoon, I have worship practice for tomorow and a full day of meeting with the people to discuss the handbook and getting it ready to print...-_-...I am soooo tired...  Anyways...even the coke and coffe is not working on me anymore, so I guess I netter go to bed before I break down...so  Later Days  \"I am soooo tired\" - Me  Neo  '),(25283,4176168,'19,August,2004','2004-08-19 00:00:00','   AHHHHHHHH!!!! HOME ALONE!!!!  Alright.....:D  So my parents are getting ready to leave for the trip around Canada...they\'ll be gone for 5 days, so tomor when I wake up...I\'ll be waking up to a empty house for 5 days...HOUSE PARTY!!!...nah.  err...the hand book issue was finally solved after some thinking and everyone seems to be happy...just hope it gets approved in time so I can print it...time is always against me...-_-  Remeber how I said in my post yester day that I can\'t take another server break down for the medias on the site?...well...it broke down...but this time for other reasons than computer bugs. It looks like our bandwith is being used a lot lately...and the 10GB bandwith was shot within 15 days. So now I can\'t log on to the server or do anything. We\'re on lock down for the rest of the month...-_-...oh well   Later Days...  \"Why, Why do you keep going?\" \"Because I choose to\"  Neo...  '),(25284,4176168,'18,August,2004','2004-08-18 00:00:00','   No matter how much they can do, always...always there will be a downside to them...  Last night I finally got my Handbook sent out...this morning to my complet surprise, the 5 people that I send it to...4 of them bounced back...-_-... and for the rest of the day, these guys have been on my ass asking for the draft when their the one who couldn\'t find the space the recieve it...when times like this, I wished that everyone used Gmail...:D  Naruto Spot will be once again up tomorow, I had the honor of again looking over the site and find bugs and so on...it was a good night of work and we got everything fixed up and looking like new. All I need to do it to upload the current Naruto Shows and we should be ready to go tomorow. Hope the server doesn\'t break on my again. I can\'t take another server break-down...  Seems that my parents will be leaving for a quick tour of Canada this weekend, they\'ll be going to a couple of places...but since I am all booked Friday and Sunday, I can\'t go with them >  Later Days...  \"The one place I\'ve never lived in is a perfect world\" - John Sewell  Neo  '),(25285,4176168,'16,August,2004','2004-08-16 00:00:00','   Well....finally got the handbook done today, the content anyways, the cover still need some work, I\'ll get it done tomorow...kind of tired now after a full day.  I want to get my passport today, so we can have over trip at the US, we\'re thinking of going to NY for a couple of days...4 to be exact, we\'re going to drive over (which I think will be a lot of fun) and stay there and visit a few places and have some fun. I think it\'ll be great.  A lot of anime for me today, Chobits, Samuria 7, Naruto...fun fun fun, love these animes...only if the download times are not so huge -_-...but I got a new Bit Torrent client, and it seems to be faster, so we\'ll see. :D  A couple days ago, one of my house phones decide to break on me, today, I finally got around to fix it but the service fee is crazy!...just for the tech people to simply look at it, it will be $30, and than they\'ll give you a \"complet\" bill after they finish fixing it...I really don\'t have that much money to fix a simple cordless  phone, so I guess I\'ll find somewhere else to fix it. How troublesome...  My friend have been working on a new lay-out for Naruto Spot, and I can say it is looking good!...it\'ll be up in a couple of days, so go and check it out!  The chinese\'s have already won 10 gold in the olympic game, the have more than any other country!...this really is coming as a surprise for me, I never realized that the chinese were strong at these games, but I\'m very proud of the team and hope they keep it up. :D  Later Days  \"Pollution will be recognized as a serious problem when it interferes with TV reception\" - Hugh Arscott  Neo...  '),(25286,4176168,'15,August,2004','2004-08-15 00:00:00','   Well....I would say today was a nice and productive day...  I finally got a cover for my handbook today!....thank godness...now all I need to do is mess with photoshop, put a title, and send it off...the deadline is today, but I feel too tired to do it now, another 12 hours wouldn\'t hurt. :P  I got to spend some quality time with my family today, had dinner, did some games, watch the Olympic game...a guy like me can get use to a life like this...everyting is good. :)  Later days  \"Television\'s influence is much too subtle to be detected by its victims.\" - Milton Shulman  Neo...  '),(25287,4176168,'14,August,2004','2004-08-14 00:00:00','   No post yesterday...I was feeling really really sick so I decided to stay in bed...wasn\'t a bad idea or I would have been really sick...Lucky me :D  Just though I do a little post before I go to the airport and pick-up my father...gosh is sure good to have him back even just for a little while...err...this project that he have been doing overseas have beening very...busy, so he only get to come back every 3 month...oh well...at least me and my mother can take care of our self...I don\'t really know when he\'s plane is arriving...I\'m guessing around 9 o\'clock...so, I\'ll be off by than...  I was watching the olympics...and I see China have done well in synchronized diving 10m platform for Women...I\'m glad...I predict that the Chinese will also do very well in bedminton, and table tennis...they always do...well...I wish them luck...maybe I should be saying that to the Canadian team? ;)  The handbook is still missing a cover...tomorow will be my last chance...and also my deadline...and this is one deadline I intend to keep, time is running short for everyone...so I can\'t afford to screw up now...not that I will...I already have a idea of what I will do, just need to run it through my team and we should be done in less than 10 minutes, wish me luck.  Later Day...  \"Expect the unexpected because your life is bound to be full of surpises.\" - Bernard Grad  Neo...  '),(25288,4176168,'12,August,2004','2004-08-12 00:00:00','   Yes indeed today was a great day...  I finally got everything done for my handbook...other than the cover page of corse...I was able to go over it many times today and fix up anything that may be wrong with it and found some more pictures on Google to use as background Pictures...so I basically have everything done. Now regarding the Cover...I\'ve decided to take a really crazy/Nice Idea picture this coming friday and slap it on as the cover...we\'ve done it many time before...I don\'t see why it wouldn\'t work now...the only problem - I need to think of that crazy/nice idea...well...got another 24 hours...times to dig up my brain and loss some sleep again...  I finally got to watch the Naruto special that I have beening waiting this week...what can I say?...1 hour of pure fun and excitment...I love Naruto, and what is even better?...the naruto site me and my friend have been working on...  had a review...a rather good one on that note...and we are quite happy about it, if you got the love for Naruto, check out our site\'s review at: http://kyubinetwork.com/index.php?id=sitereviews/nspot/index.htm ... a bit of traffic wouldn\'t hurt...:D  Anyways...today I have relased that my father is coming back to Canada this week...he have been away on a project in China for many month now...and he\'s finally coming back!...too bad he\'s leaving again in a month but...I\'ll just enjoy the time for now...we have a lot of planned made, we were thinking of going to Disney World this summer, really looking forward to it...  Later Days  \"Don\'t you know what a library is for? It\'s to prevent us making damn fools of ourselves\" - George Locke  Neo...  '),(25289,4176168,'11,August,2004','2004-08-11 00:00:00','   Is summer suppose to be tiring?... Because I am...             I found myself sleeping very badly and waking up very early today...probabely only 4 hours of sleep...-_-...so very tired...my body is aching all over the place...ouch!                 But even with all my limitations, I still    manage to get alot of work done today...I finally decided to take up my Volunteering work today and start doing some...I was told if you do more voluteering work, you will better your chance in getting into a university...who knows? Maybe their right...           Anyways...I discovered some major flaws in my handbook today and took me 5 hours+ to fix it...turns out all my page size was not correct and was not ready for printing, and since many pages of my handbook had ridiculously complex format, when the page size was changed...well...everything just want......I\'ll let you imagine the rest...           But all in all I was able to fix it...but what do I get other than a very tired body &amp; mind?...even more problem, now because of the new page size, I have doubled the number of pages I orginally had which present a problem because I don\'t have as many background images...and my time is running out for more real life photography...so let\'s see if Google can help me out on this one. Still no cover though, my pastors want to see a full hand book this coming Sunday, I don\'t know where I can dig up a cover...oh well...I\'ll do the best I can...but time is against me...it was always against me.           But...overall it have been a ok day, I\'m sure to get a good night sleep after all this...after all...I\'m still just a kid. :) ...well...at least boredom haven\'t kicked in yet...(this is about the only summer that I\'m not bored) but I have a feeling I soon will be...           Later Days           \"Since the beginning of time, time were against us, and always will be\"           Neo...  '),(25290,4176168,'10,August,2004','2004-08-10 00:00:00','   well......at least the water is back on...:)  I was able to install the final version of Windows XP Service Pack 2 without any trouble, and in all honesty it is a great looking service pack...Microsoft have focused much on its security this time around and I think they have finally gotten something right ;)...and for all of you that are not as computer savy as you should be...this service pack will show up on your winodws update on the 28th and you can express install from there...all in all it\'s great.  This morning, I have noticed a piece of software that have come across my way...Ad-aware...not the Ad-aware 6 everyone have being using, but Ad-aware 6 Speical Edition. So I decided to give it a download and it is GREAT!...it scan files much more carefully and it is able to pick up all different kind of adware and spyware...I would recommend it to everyone...you can go grab it at download.com...but...I need to warn you in fairness that it is kind of slow comparing to Ad-aware 6 and according to others...it some times delet the wrong files...but no problem at my end...so go grab it...  Even with all these great things...my situation of my handbook still haven\'t changed...uptill this point no one is able to give me something decent for the cover...Time is running out (since the deadline is the 14h) I can only wait one more day...if still nothing good, than I\'ll have to make it...err...I really don\'t want to go there...-_-  The weather here in Toronto have been very cool...unlike the last couple of years when it\'s suppose to be 30+....but this year in the middle of August is surprisingly...very nice & cool, I almost have no use for my AC in my room anymore, but I\'m going to keep it anyways...just in case... ;)  All in all....it have been a good day...just wish someone would give me a decent cover page...oh well...nothing I can do...  Till then...  \"Ignorance is Bliss\"  Neo...  '),(25291,4176168,'09,August,2004','2004-08-09 00:00:00','   Nothing have been going right today...when I work up this morning...to my complet surprise, I realise that I have no water and I didn\'t wake up early enough to save some water to use for the rest of the day...-_-...lucky me...  And if that is not the only thing that is not going well...the Handbook I\'m working on for my Church\'s Summer camp is in desperate need of a cover page, and it almost seems that no one in my church or my handbook team have any artistic skills whatsoever...what does that boil down to? I need to ride shotgun and make my own cover, and if I ineed need to...it will not be pretty...oh boy, this will really make my day...  And what is even more depressing is that I just realized today what many doesn\'t want to...My summer is half shot!!...only a few more week until I need to sit back down into that classroom and well...learn and achieve...as much as I like reading and writing and other \"eductaional\" stuff, school is not on my favourt list...(well, not any more)...  Well...what am I going to do with all these stuff?...I guess I can only do them, forget about them and hope for the best...oh well...let\'s hope the situation changes in the next 24 hours...  Till than...  \"You don\'t know what it is, but it\'s there...it\'s like a splinter in your mind, driving you mad\" -- Morpheus  Neo...  '),(25292,4176168,'08,August,2004','2004-08-08 00:00:00','   Well Well Well.....finally...my own mind blog...I always wanted to do something like this...but never got around to it...I think this is a great tool to write down how I feel, opinions and...who knows what might come across my way...  Some happy news today I guess...my English youth pastor got engaged today to one of my friend, really happy for them both...but right after the good news...more responsibilitys, more things need to be completed for the hand book and my GM position...things are going well I guess, since I am on my summer vacation, there is much time to relax and rest. But a bit of work is always needed just to keep me away from brain dead state. ;)...well.....nothing much to add...but I am really enjoying this, hope this will last.'),(25293,4221834,'21,August,2004','2004-08-21 00:00:00','   whenever i read his blog, i become very paranoid...   keep guessing what his blog means...who is he referring to...  wat has happ to him...did he fall in love with someone...   i get very irritable when i dun get to see him for a few days...   i hate to behave like dat...   and worst, she kept pissing me off...maybe becoz they are both capricorn....   perhaps i am the one who cause misery to every1   will he ever be concern abt me...   told him my problems...send me \"why?\"...no reply after that...WTF...   i dun mind being fooled around by him at all...as long as i get to see him...   silly, stupid, naive....watever u call dat...   i last saw him on 17/8, 4 days already...i miss him alot...   is he thinking of me? haiz...he will call if he really does...   can someone get him out of my mind...   my inferior complexity is getting worst....   i am ugly, fat, ignorant, bossy, irritating, good at nothing, useless...and any other bad traits u can think of....   name it, i have it... '),(25294,4221834,'19,August,2004','2004-08-19 00:00:00','   For  SKY :        INSTALLING LOVE  Tech Support: Yes Ma\'am... how can I help you? Customer: Well, after much consideration, I\'ve decided to install Love. Can you guide me though the process? Tech Support: Yes. I can help you. Are you ready to proceed? Customer: Well, I\'m not very technical, but I think I\'m ready. What do I do first? Tech Support:  The first step is to open your Heart . Have you located your Heart ma\'am? Customer: Yes, but there are several other programs running now. Is it okay to install Love while they are running? Tech Support: What programs are running ma\'am? Customer: Let\'s see, I have  Past/Hurt,  Low Self-Esteem, Grudge and Resentment running right now. Tech Support: No problem,  Love will gradually erase Past/Hurt  from your current operating system. It may remain in your permanent memory, but it will no longer disrupt other programs. Love will eventually override Low Self-Esteem with a module of it\'s own called High Self-Esteem. However, you have to completely turn off Grudge and Resentment. Those programs prevent Love from being properly installed. Can you turn those off ma\'am? Customer: I don\'t know how to turn them off. Can you tell me how? Tech Support: With pleasure. Go to your start menu and invoke Forgiveness. Do this as many times as necessary until Grudge and Resentment have completely erased. Customer: Okay, done! Love has started installing itself. Is that normal? Tech Support: Yes, but remember that you have only the base program. You need to begin connecting to other Hearts in order to get the upgrades. Customer: Oops! I have an error message already. It says, \"Error - program not run on external components.\" What should I do? Tech Support: Don\'t worry ma\'am. It means that the Love program is set up to run on Internal Hearts, but has not yet been run on your Heart. In non-technical terms, it simply means you have to Love yourself before you can Love others. Customer: So, what should I do? Tech Support: Can you pull down Self-Acceptance; then click on the following files: Forgive-Self; Realize Your Worth; and Acknowledge your Limitations. Customer: Okay, done. Tech Support: Now, copy them to the \"My Heart\" directory. The system will overwrite any conflicting files and begin patching faulty programming. Also, you need to delete Verbose Self-Criticism from all directories and empty your Recycle Bin to make sure it is completely gone and never comes back. Customer: Got it. Hey!!! My heart is filling up with new files.      Smile is playing on my monitor and Peace and Contentment are copying themselves all over My Heart. Is this normal? Tech Support: Sometimes. For others it takes awhile, but eventually everything gets it at the proper time. So Love is installed and running. One more thing before we hang up. Love is Freeware. Be sure to give it and its various modules to everyone you meet. They will in turn share it with others and return some cool modules back to you. Customer: I promise to do just that. By the way, what\'s your name? Tech Support: Just call me the Divine Cardiologist, also known as the Great Physician, or, just \"I AM.\" Most people feel all they need is an annual checkup to stay heart-healthy; but, the manufacturer (ME) suggests a daily maintenance schedule for maximum Love efficiency. '),(25295,4221834,'18,August,2004','2004-08-18 00:00:00','   its time we keep a distance...   things she says starting to pissed me off...   maybe its my character...i cant be too close with someone for too long....   she mentioned that she feels dat way before...   her words hurt now...perhaps i am too sensitive...   i care for her as a little sis...whom i care alot...but it seems that she is taking it for granted...   she will still do things for me...but her words...i dunno how to say...juz affect me alot...   i have to tell myself...she\'s juz a little gal...let her be...   he &amp; she are the same....capricorn...always think that they are right...   she will revoke watever comments i make, watever i say...   everything...   hope things will turn out fine....dun wanna lose a fren... '),(25296,4221834,'14,August,2004','2004-08-14 00:00:00','   this is the 1st time he openly discussed his past relationship with me....   so wat does that make us...i mean he treat me as fren only,right...   i kept asking myself-will i tell the person i love abt my past relationship so openly...declaring my love for my ex???   i mean normally, i will only discuss these kind of things with pple i have no feelings for, am i right...   or is this a step further for our relationship....   i am pessimistic coz this is the guy who told me dat i am not the one he wanna spend his life with...relationship is impossible between us...   i can no longer hold on to any hope that our relationship can go further...   he never asked me abt wat had happ to me....never was he concerned abt me....   told him i wanna quit my current job...he asked me wat happ...1st time...   told him i wanna stay at his place...he asked if i had a quarrel with my family...1st time...   in the past, he shld be glad that i wanna go over to stay....this time, he asked why...   so wat am i to him...   i wanted to ask him this....but i know i will not get a reply...so wats e point in asking...   i was at his place the past 2 days...mixed feelings...   sometimes i hate him...but i think i love him more...   my dad was pissed with me for staying out...but wat can i do...   i can never reject his request...i am afraid that he will never ask again...i am such a wimp...   when will the time comes whereby i will chide myself for being stupid after reading my own blog...   and then delete all those memories away...   hope that day will come soon...i am tired...   career is not going smoothly...relationship...????...worst....   i am choked...by emotions...and by lies...the fact that i cant share wats happening with my frenz...not even my close frenz...   everyone hates to see me sad and down over him...   now i have to lie to all my frenz dat i no longer keep in contact with him...   wake me up....bring me back to reality.... '),(25297,4221834,'14,August,2004','2004-08-14 00:00:00','   i never have the habit to write a diary...   until he sent me a link to his blog...   i never knew that he needed frenz and loved ones....maybe becoz i am not included in his list of frenz &amp; loved ones...to him, maybe i am juz a fling   sad,right? he meant so much to me....   until today, i am still waitin for him....   but i know that i am always his lowest prority...   e whole world is bz then he will call me....normally on mon or tues...   other days, he will be out drinking, hooking up fling...i wher got chance...   i promised a fren to end this stupid thing by 8 August but i still accede to his request on national day...haiz...   maybe when i look back in future, i will think that i was so stupid....   but now i am still hoping that he will need me again... '),(25298,4221834,'13,August,2004','2004-08-13 00:00:00','   whats the real meaning of loving someone?   i\'ve tried so hard to change...   but he can still tell me straight in my face that i\'m not e person he wanna spend his life with....   e more i try, e more he hates me...   maybe its really juz abt  chemistry ...he will never love me...   i\'m not askin 4 much...   juz someone whom i can talk to, give me advice...someone whom i can shower him with care and concern   esp after a hard day wk...juz wanna be wif him, spend time wif him...even juz looking at him...   but my \"relationship\" with him started off weird...   we used to be colleagues...he never earned my second glimpse then....   i somehow gave him my no....   then one fine day, i received a msg from him informing his change of mobile no....   then started asking him out...   pple ard me who had seen him ask me to steer clear of him....but being a stubborn person (he hates me for dat as well)....i carried on...   we started off with a no-commitment thing, he needed me then....maybe becoz of my ego...(A LEO)...haha... i stupidly gradually fall for him...  then i slowly realised that i\'m juz being made used by him...   he was in between job then...sold off his car...rotting at home...(as usual, guys-cant do anything without a car)   and me, a stupid fool...will go against my dear daddy to spend almost everyday at his place...   hopin dat he appreciate me 1 day...i was so wrong....my dad was really upset and disappointed with me...i still went on...   everything ended when he expected me to go to his place to feed the dog...coz he\'s going drinking...   i din go...he flare up...gave me a fucking long lecture as he always did n cut off everything with me...drama??   no, juz a chance for him to ditch me...   i even willingly become his sex tool, sucking tool....   even call gal earned money from sex...me....free sex...comes with free delivery some more...good deal ya...   until now...i am still giving free sex...   he can even sms things like \"u will only look for me when u need it\"...look who\'s talking...   he will give me false hopes although he made it clear we can never be together...   giving me 3 commandments in order to be his gf:-   1) no guys shld call my mobile   2) no clubbing   3) stay at his place for at least 3 times a week   i can do everything he asked...i agreed immediately...guess wat he said-> \"i\'ll give u more time to think abt it\"   isnt he irresponsible??? nopez...he dun think so....he chides me for being irresponsible....   i can comment things abt him...but i love him as he is...   but he hates things dat i do...   wats wrong with changing mobile phone...he chides me for being childish....   if u cant appreciate the things dat i do,fine...y hate me for dat...   even if i dun appreciate the things he does, i still loves him...   poor leia...she\'s suffering...without love...without care...sometimes without food...   i can name all sort of bullshits which he had given me....but i still loves him...   heard from someone who asked me to let go....   he plan to remain a bachelor for life...   wat can i do.... '),(25299,4021133,'31,July,2004','2004-07-31 00:00:00','    harloe.. wow, i wake up so early today. normally i wld slp till 2 pm. nvm, its ok. i set up this blog few days back, but i didnt got a chance really to blog it. so today i\'m dng it. yesterday was min bdae. i think i gonna get her present later. actually yesterday went to buy, but cannot find. today was justin bdae. later in the evening he\'s treating us to jap buffet. oMG.. hehe.. b4 tat, i gotta meet wiro at orchard at 1.30 to get sth first. and promise to treat him ice-cream. so fer the time being i\'m still quite free to blog. hee.. neeway, lynn bdae they all decided to get her the top we saw that day. rima suggest to get a cake. i\'ll see rima tml and shall ask her. oh and yesterday i went out wif my kor, steve. went to PS watch house of flying daggers. went to eat pizza. saw chuan xi wrking at the GV dere. went o Best Denki to get a mic. saw chris and issac. chris give me discount. so sweet of him. haha.. but he said my hair very kuku. wad nonsense. he muz be jealous. LOLZ. i also bought Elva cd at mJ. the songs are nice. really.think i shld end here le. write u a story in my next blog. cya! '),(25300,4021133,'26,July,2004','2004-07-26 00:00:00','    after the closing of my pinkytortoise, i came up wif another blog.. this the one.. cool and nice rite.. miss those days when i blog.. after sometime i nv blog, the urge of blogging came back.. shldnt haf close it in the first place.. but its ok la.. haha.. anyway.. end here.. cos using my sis com.. later she #$%^&amp;*.. bleahz '),(25301,4021133,'19,August,2004','2004-08-19 00:00:00','    The truth is tearing up my heart I can\'t recognize this place The endless road without a stop sign Can\'t even find a stranger this time Why am i still holding back my tears In this loneliness there\'s nothing to fear Every chord still seems a wonder How we could be together Every time i ask if this would be the last Why am i still talking to myself Hoping you will have the keys to my cell Every song might calm the weather but it just draws me deeper How do i get out of this I think i never will A crystal forming in the eye Maybe this would be the last The winding pathdown my face Till i begin to taste the bitterness inside '),(25302,4021133,'14,August,2004','2004-08-14 00:00:00','    Hiding in a corner looking at you Keep a distance While not noticing that i\'m around Look at you with your back facing me When you don\'t know My heart brings a sweet little feelings  You breeze past me, you can\'t see me is this call fate I\'m here with you, right beside you You just don\'t know I breeze past you, you can\'t see me From this time There seems to be a bridge Where we keep our distance right where we are  Keeping a distance from you I try Can\'t get too close From this moment I know I can\'t breathe Look in my eyes deeply if you will My heart beat fast I tremble on my knees I pray  Just where you fall in love There are smiles, there are tears, there are joy The distance is where I keep my thoughts for you '),(25303,4021133,'14,August,2004','2004-08-14 00:00:00','    ç¿»ç€æˆ‘ä»¬çš„ç…§ç‰‡   æƒ³å¿µè‹¥éšè‹¥çŽ° åŽ»å¹´çš„å†¬å¤©   æˆ‘ä»¬ç¬‘å¾—å¾ˆç”œ çœ‹ç€ä½ å“­æ³£çš„è„¸   å¯¹ç€æˆ‘è¯´å†è§ æ¥ä¸åŠå¬è§   ä½ å·²èµ°å¾—å¾ˆè¿œ ä¹Ÿè®¸ä½ å·²ç»æ”¾å¼ƒæˆ‘   ä¹Ÿè®¸å·²ç»å¾ˆéš¾å›žå¤´ æˆ‘çŸ¥é“è‡ªå·±é”™è¿‡   è¯·å†ç»™æˆ‘ä¸€ä¸ªç†ç”± è¯´ä½ ä¸çˆ±æˆ‘   å°±ç®—æ˜¯æˆ‘ä¸æ‡‚   èƒ½ä¸èƒ½åŽŸè°…æˆ‘ è¯·ä¸è¦æŠŠåˆ†æ‰‹å½“ä½œä½ çš„è¯·æ±‚ æˆ‘çŸ¥é“åšæŒè¦èµ°æ˜¯ä½ å—ä¼¤çš„è—‰å£ è¯·ä½ å›žå¤´   æˆ‘ä¼šé™ªä½ ä¸€ç›´èµ°åˆ°æœ€åŽ å°±ç®—æ²¡æœ‰ç»“æžœ   æˆ‘ä¹Ÿèƒ½å¤Ÿæ‰¿ æˆ‘çŸ¥é“ä½ çš„ç—›   æ˜¯æˆ‘ç»™çš„æ‰¿è¯º ä½ è¯´ç»™è¿‡æˆ‘çºµå®¹   æ²‰é»˜æ˜¯å› ä¸ºåŒ…å®¹ å¦‚æžœè¦èµ°   è¯·ä½ è®°å¾—æˆ‘   å¦‚æžœéš¾è¿‡   è¯·ä½ å¿˜äº†æˆ‘ '),(25304,4021133,'13,August,2004','2004-08-13 00:00:00','    It\'s best to wait for the one you want than settle for the one available. Best to wait for the one you love than one who\'s around. Best to wait for the right one because life\'s too short to be wasted on just someone.  \"Before you get married, keep both eyes open, and after you marry, close one eye.\" Before you get involved and make a commitment to someone, don\'t let lust, desperation, immaturity, ignorance, pressure from others or a low self-esteem make you blind to warning signs. Keep your eyes open, and don\'t fool yourself that you can change someone or that what you see as faults aren\'t really that important.  Once you decide to commit to someone, over time, their flaws, vulnerabilities, pet peeves, and differences will become more obvious. If you love your mate and want the relationship to grow and evolve, you\'ve got to learn how to close one eye and not let every little thing bother you. You and your mate have many different expectations, emotional needs, values, dreams, weaknesses, and strengths. You are two unique individual children of God who have decided to share a life together.  Neither one of you are perfect, but are you perfect for each other?  Do you bring out the best in each other? Do you compliment and compromise with each other, or do you compete, compare and control?  What do you bring to the relationship?  Do you bring past relationships, past hurt, past mistrust, past pain?  You can\'t take someone to the altar to alter them. You can\'t make someone love you or make someone stay. If you develop self-esteem, spiritual discernment, and \"a life\" you won\'t find yourself making someone else responsible for your happiness or responsible for your pain. Manipulation, control, jealousy, neediness, and selfishness are not the ingredients of a thriving, healthy, loving and lasting relationship  Seeking status, sex, and security are the wrong to be in a relationship.*TAKE NOTE*  What keeps a relationship strong? * communication * intimacy * a sense of humor * sharing household tasks * some getaway time without business or children * daily exchanges (a meal, shared activity, a hug, a call, a touch,    a note) * shari ng common goals and interests * giving each other space to grow without feeling insecure * giving each other a sense of belonging and assurances of    commitment  If these qualities are missing, the relationship will erode as resentment, withdrawal, abuse, neglect, dishonesty and pain replace the passion.  \"As long as we have memories, yesterday remains. As long as we have hope, tomorrow awaits. As long as we have friendship, today is beautiful.\"  Love is when you take away the feeling, the passion, and the romance in a relationship and find out that you still care for that person.  If You Are :  MARRIED = Love is not about \"it\'s your fault\", but \"I\'m sorry\", not \"where are you\', but \"I\'m right here\", not \"how could you\" but \"I underst and\", not \"I wish you were\", but \"I\'m thankful you are.\"  ENGAGED =The True measure of compatibility is not the years spent together but how good you are for each other.  NOT-SO-SINGLE = (Those who have \"special someones\") Love isn\'t about becoming somebody else\'s \"perfect person.\" It\'s about finding someone who helps you become the best person you can be.  HEARTBROKEN = Heartbreaks last as long as you want and cut deep as you allow them to go. The challenge is not how to survive heartbreaks but to learn from them.  NAIVE = How to be in love: Fall but don\'t stumble, be consistent but not too persistent, share and never be unfair, understand and try not to demand, and get hurt but never keep the pain.  POSSESSIVE = It breaks your heart to see the one you love happy with someone else but it\'s more painful to know that the one you love is unhappy with you.  AFRAID TO CONFESS = Love hurts when you break up with someone. It hurts even more when someone breaks up with you. But love hurts the most when the person you love has no idea how you feel.  STILL HOLDING ON = A sad thing about life is that when you meet someone who means a lot to you, only to find out in the end that it was never bound to be and we just have to let go.  SINGLE = Love is like a butterfly. The more you chase it, the more it eludes you. But if you just let it fly, it would come to you when you least expect it. Love can make you happy but often times it hurts, but love\'s only special when you give it to whom it\'s worth. '),(25305,4021133,'12,August,2004','2004-08-12 00:00:00','    shldnt haf msg me shldnt haf see him shldnt haf walk there shldnt haf take short cut shldnt haf met her at tat time shldnt haf meet her in the beginning shldnt shldnt shldnt shldnt all the shldnt or maybe we were destinated to meet is it so? why do u wanna play a joke on me? you want me to see him wif his gf want me to give up on him totally now u allow me to see him again and allow him to msg me what do u want actually do u noe how miserable i feel? how am i gng to forget him? u tell me are u seting a test on me? ok, now i tell u, i failed or are u hinting me sth else pls, i beg u stop it. i\'m not strong. i\'m weak if u really want me to forget bout him pls dun let me see him again. if not u r telling me we r meant to be when we are not. '),(25306,4021133,'07,August,2004','2004-08-07 00:00:00','    yEah, its me. cant wake up on time. so skip one lesson. when gng fer the next lesson, feel like grabbing a bite. went to coffee shop alone, eat breakfast watch tv. take my own sweet time. late fer another half an hr. saw mal at mrt. den chat all the way to sch. during lesson, read magz. nv pay attention at all. after sch went to jie hse do CA. after tat met aisha to help me wif that proj. thk u eh. den on the way to aisha hse, saw him. (\'n\')\" sad ar. den went to aisha hse saw the photo she took wif rima. i luff till tears flow out. it was very furnie. noe wad, they actually haf the same post fer all eight diff background. LOLZ. she gave me one. it really cheer up my day. '),(25307,4021133,'07,August,2004','2004-08-07 00:00:00','    ar loe.. i\'m back fr wrk. waitin fer my sis to be back. she went to shang wei and wilber concert. tml i have to wrk full shift again. meeting siti 9.30 at amk. tml will be another boring day. friday morning wake up nine plus. but still went to sch late. i\'m too slack. watch tv n read newspaper n makan at the same time. den call yj n she told me lesson cancel due to national day. so my lesson is only 1 to 3. den decide to go J8 to copy down the ingredients fer CA report. den bought a shampoo. i forget to bring my phone and gt lots of missed calls fr aisha. hehe. after sch went back hm. den sing songs. till i fall asleep. den at six went to meet siti fer dinner. den wait fer rimah till eight plus. den we go buy fruits and KFC. Ard 9 plus, went down den rima say she nt yet eaten. so poor thing eh. whole day nv eat till 9 plus. den JC fetch me. den we came back to J8 at 11 plus again to fetch angeline. and we went to jln besar eat dim sum. shiOk. todae todae todae, went to wrk. SC fetch me. i\'m so good life. haha. i\'m so tired. nO mood also. look so gloomy. hehe. den a lot of things happen.. but dun wish to say so much.. dun wanna blog le.. end here.. goOd nitez.. loVe ya All.. '),(25308,4021133,'05,August,2004','2004-08-05 00:00:00','    hello, iz me again. juz came back fr wrk. i\'m so shag. yesterday morning, sc send me to sch. jc follow along too. wif his subaru. coOl. so tired in sch. think i sfall aslp fer all the lectures. haha. i skip reactor systems. went to far east get S.H.E pic bk. quite nice. k la, they are pretty. den pt bought my bf (renfu)candy floss. after tat we went to AMK hosp to visit grandpa. came hm and went to pass min her prezzie. order pizza at nite. so full so tired. today, i was as usual very shag. but i was quite happy. morning lesson start at 8am. i reach at 8.45. cOOl rite? coOl my head. lOlz. the lecturer even rem my name.cos i skip her lesson twice in a month and late everytime. actually i did it on purpose.i noe i am crazy. but i cant help it. hee. b4 CA lesson, saw KM. ask me acc him but tissue. troublesome. haha. den CA lecture he crap alot.bth. but actually i dun really mind. cos i was bored too. his gf call yl. neeway oso nt really his gf. dun care him la. went to play pool at amk during the break. wif sh, faz, fh n john. actually didnt wanna go, cos left yj behind very bad. guilty. but sh say we in sch also do nth. quite true. so we went. den i find out that faz dun like malay gals. he like chi. funny rite? but ok la, cos i oso quite like malay guys. hee. but it doesnt implement anything. okay. dun anyhow think. after lesson went to wrk wif aisha. qiaoling came. so glad to see her. i miss ferrin. haha. dunno wad to write alr. sian. tired ah. nYTE.. my way of ending funny rite. so sudden. but i dun care. haha. miz ya guys. byEZ '),(25309,4021133,'04,August,2004','2004-08-04 00:00:00','    harloe.. been very buzi recently. study and work. dun even have time for myself. okay, sat went to justin bdae rite, saw devon and albert at PS. tat day the bdae boi very attitude. keep giving me attitude and start acussing me giving him attitude. so flare up of him. and still had the guts to ask me wad wrong wif me. hey, are u having a big prob or a small prob? till now still nt yet talk to him cos on mon he gave the f**king attitude again. and the same old thing happen. childish rite. hA, but that\'s life. neeway, quite enjoyable oso la. sunday, wrk in the morning. late fer ten min and got all the naggin fr ah nan. haha. but is alright la. gt use to it olreadi. rima msg at 9.30 and said she will be late. den on tues, which is yesterdae, it was mine turn to msg her. wad a joke. i\'m following her footstep closely. hee. juz kiddin. den mon went to catcha  movie wif yj. Mean Girls. it was nice. really nice. johnathan was especially handsome. oh my goddness. den after nite class, went out wif KM. walk all the way to sport hall and left the sch fr the back exit. den walk to yck mrt den take 13 to amk. went dere chat chat. drink bubble tea. BLUE SEA. he treat me. den he told me a lot of things. den after that walk back to yck. walk walk walk the whole night. but dunno why, quite happy lei. hee. den meet JH and Steve. go makan at 795 and met XJ. the next dae which is tues, woke up at 9.30 when i was suppose to wake up at 8.30. stupid HP batt flat. totally flat. nOe why? cos there\'s this person who bomb and flood my hp wif msges on mon nite. and the msg was so disgusting. u wont want to noe the content. its so...... nvm forget it. tues work @ atrium.. it was so the bored. zZzZz.. rima and siti help me bought the FILA bag.. very NiCeeee. hehe.  thanks R and S.  oso bought a polo t. den we share a bdae present fer lynn. bought her a clothes. hope she like. gng out now to pass min her present. write again some other day. buAiz.. mUAckIeZs '),(25310,3722603,'27,June,2004','2004-06-27 00:00:00','   im going to post a bunch of random stuff now cuz its sunday morning, a day before tour, and im doing my laundry. i am SO cool. after i do laundry, im going to visit my brother at his rehab center place and take him out to lunch then go see dodgeball again. After that im gonna come home, start packing for the trip, then stay up all night cuz i wont be able to sleep. THE ANTICIPATION IS MAKING ME FRICKEN CRAZY.last night, i was thinking of all the places we were going to and i started to get hyper, at THREE in the morning. Tour,THESE ARE GOOD TIMES. so im going to list my hotties, there are many, but ill try to keep them to the HOTTEST HOTTIES, like on VH1. (these arent in any order, just as i start remembering them) jennifer lopez,jessica simpson,jessica alba,NOT BRITNEY SPEARS AND PARIS HILTON UGH AND THAT OTHER UGLY ASS ON THE SIMPLE LIFE,the girl in \"day after tomorrow\", kristen kruek, ur mom, halle berry, keira knightley,the indian chick from \"bend it liek beckham\" oh, posh spice, leah remini, ben stillers wife, lucy liu,charlize theron, the harp player lol. thats it for now but i shit you not, the list continues. booz chauncey '),(25311,3722603,'25,June,2004','2004-06-25 00:00:00','   i liek the wierdest music. for example, to day i went to tower and bought 3 cd\'s. i got the franz ferdinand cd the d-12 cd cuz i lost mine, then the jedi mind tricks \"violent by design\" if u know any of these groups, you would knwo that none of them have anything in common except for d-12 and jedi, but only cuz they are both rap groups, however, they rap about totally different things. basically, jedi raps about some shit that d-12 is too main stream to rap about, such as the \"game\", cuz if d-12 tried to rap about that, they would be mocked, cuz they alreay made it, whereas jedi is still underground. so, i have a large range of music going from classical(brahms) which in itslef has many different periods to it, to rock(hoobastank) to metal(SOAD) to rap(eminem). rap and classical i say are the best, cuz when combined, u get artists liek jedi mind tricks who are truely, the best of bothc worlds, unlike R kely, and jay-z, BUT jay-z is a sick ass rapper anyways. chauncey '),(25312,3722603,'23,June,2004','2004-06-23 00:00:00','   alright, so i was driving down to san mateo with my mom to go to my brothers rehab meeting, which by the way was lovely(will ferrell from \"old school\") ANYWAYS,and this fucken truck fucken streaks in front of our car, making my mom swerve the car. so at first i was like \" wtf, that almost made me shit in my pants\" cuz seriously, i was frightened for the moment. but soon after, the car that cut in front of us stopped, then started up, then stopped, then started up. THIS CONTINUED FOR A GOOD 15 SECONDS. then this woman raises her fucken crusty hand through the sun roof, flipping us off. LOL, so my mom tried to do it back, but she has tendinitis in her left hand, so it came out as a sort of \" fist\" instead of a flipping of the bird.LOL, but the woman got the point. so the woman gets into the lane next to us, and slows down so that she is right next to ME,talking to me. she says \" you cant just fucken cut me off like that!\" and i rolled up the window and acted like i didnt notice her yelling at us. BUT this made it worse, she got back in front of us. WHOA, U MUST UNDERSTAND THAT WE WERE GOING LIKE 60 ON THE FREEWAY DURING THIS WHOLE TIME. back to the story,  so when she was back in front, she reached for somehting in the backseat of her truck and pulled out a drum stick. then she started twirling it in front of the rearview mirror, LOL, and then she dropped it in the seat, went to look for it, then she changed lanes, and never came back. In the end, i feel sorry for her. She was trying to be all hard and shit, and ended up looking liek a retard dropping the drum stick.w/e tho. I hope i am doing this whoole \" blog\" thing right, cuz i have nothign else to say besides what happened in my day. chauncey '),(25313,3722603,'23,June,2004','2004-06-23 00:00:00','   alright, FIRST OF ALL, my fucken haircut sucks my moms booty crack. SECONDLY, NEVER EVER GO TO SUPERCUTS!  This one time when I was in 7th grade, I went to supercuts right before the year started. I wanted to get like a 2 on the sides and about an inch off my 3 inch long hair on top. However the Asian lady how was about to butcher my hair never really heard what I said. I heard her murmur something along the lines of \"askhfgasjkfgjkg  2 OK! asfdhibasjkgvaha\" so I was like OK, maybe she will give me a much betetr haircut than I was use to receiving. BUT, the fucken woman gave me a 2 all the way around, leaving me with a shaved head. Today, I went in and got my haircut. It wasn\'t THE WORST HAIRCUT I\'ve ever gotten, but nothing is evenly cut, and there are patches of hair that are much more dense than others. I\'ve gone to supercuts between those 3 years, but supercust just isn\'t that great of a place. IM sure there are a few hear and there that have actual barbers, but in the end, they just aren\'t great haircutists. i leave for tour in 5 days. I hella wanted that loyd banks cd, but it comes out the day after I leave. DUDE, i want to see spiderman 2, looks hella good. I saw around the world in 80 days, it was alright, better than expected which is good and i saw dodgeball, SO FUCKEN HILARIOUS. damn, i dont think anyone can capture such moron-ness better than ben stiller. chauncey '),(25314,3722603,'22,June,2004','2004-06-22 00:00:00','   during the school year i was always busy. i usually woke up before school at like 5 a.m. so i could practice for about an hour and a half. after practicing, i would take a shower get dressed read the sports page and datebook, then drive to school. THEN there would be school, which for me, is at Stuart Hall High School which is a really small boys school near japan town. its not a great school, but for me it is perfect. i can co exactly what i want to do which is to get my school shit done, have a few close people that i basically only hang out with all year cuz everyone else is REALLY COOL.... play baseball whihc is the best sport ever next to baskteball and football, and focus on my music outside of school. So for my pusposes, Stuart Hall is fine.during baseball season, i would go to practice everyday after school until about 6, then come home, do my homework, then go to sleep and wake up and practice and condoing that cycle of nonstop madness everyday until now, which is summer time. this summer is really busy for me. i had a recital, which i just finished , and it was a lot better than i thought it would be, then i have tour with SFYO. the SFYO tour is gonna be fucken tight. we\'re going to france, austria, amsterdam, germany and somewhere else i cant remember. the best part of it is that im going to get to chill with a lot of people i grew up with at TCS, my junior high school. so hopefully tha will be cool. OH YA, weed is legal in amsterdam, so if you\'re there, look for me in those coffee shops that have it on their menu\'s............JAY KAY. after i get back from tour, im going to drive down to LA wiht my sister and cousin and go to disneyland for a few days. ok fine, disneyland isn\'t the coolest place ever, but shit, a roadtrip wiht my closest people is gonna be sick regardless of where we are going.  after the road trip, im going to england which is followd by a camping trip to mendocino to close out the summer. back to \" having nothing to do is fucken wierd\", until monday the 28th, i have jack shit to do. THE ANTICIPATION for tour is gettig to be too much for me to handle. everyday gets longer and longer as each hour seems to last for a century. NOW i realize how easy it is for people to ramblke on and on in these things, so im goign to stop. im sure this is stupid but, it is my first online diary shit post thing ever so have fun. chauncey P.S. fucken lakers.............. GIANTS BABY   YEAEEEEEEEEEEEEEAAAAAAAAAAAAAAHHHHHHHHH '),(25315,3722603,'28,July,2004','2004-07-28 00:00:00','   So i was talking to this guy and he was like,\" OH MY GOD, how long does it take a black woman to take a shit?\" I, along with all the other \"PC\" people i was standing with did not move a muscle. \" 9 MONTHS. HAHA isn\'t that funny?\" I had to keep my dumbass from wetting myself in the pants. That could possibly be one of the most racist, offensive HILARIOUS jokes i\'ve ever heard. ANYWAYS, im just getting ready for my trip to England.&nbsp; I dont really know all of the people im going there with, but the 10 hour flight should allow me enough time to become friendly with my new posse. Oh ya, my mom needs to DIE. I just realized how much of a bitch my mom is. My friends, lol, use to tell me \" your mom is like fucken scruff Mc gruff\" LOL, at the time, I&nbsp;would&nbsp; cock slap them and tell them to never say that again. Now? I AGREE WITH THEM. my mom needs to get laid or eat some fruit or do anything that gets her to chill out. I mean i use to think it was impossible to walk around with a cock in ur ass(ron gallman LOL) but NOOOOOOOOOOO. my mom proved me wrong by achieving the impossible. she just needs to back the fuck off of me and learn how to deal with her issues without yelling at me. FUCKEN BITCH. so my trip to england should be fricken awesome. did u know that people in Britain call c smokes FAGS? i know its not that funny, but it still is wierd.&nbsp; but here is my checklist 1- get a haircut, I LOOK LIKE JESUS 2- go out with katy(hell yea pacifica) 3- buy some POUNDS, british currency 4- pack for the trip 5-charge my i-pod 6- make sure i have all of my extra shit for the trip like cell phone cords, i pod cords, and gameboy shit, and condoms, my hat, my music, i was JOSHIN about the condoms. 7- cock slap my mom 8- thats it 9- i  WAS  serious about cock slapping my mom. for now,  chauncey '),(25316,3722603,'22,July,2004','2004-07-22 00:00:00','   First off, im&nbsp;not actually a bitch, ILL KILL YOU if u call me one. JAYKAY,not really though. SECONDLY, so theres this girl that i sorta like. I dont know how im going to tell her. most people ive talked to about it say i should go ahead and tell her, but im hesitant to do so. i know she doesnt like me, but I feel obligated to tell her for some reason. I know FUNNY GIRL thinks i should not do it, but im going to. It will alleviate myself of any obligations ..............anyways, im just over analyzing&nbsp; LIKE A LITTLE GIRLY BITCH. THERE I ADMIT IT. ARE U HAPPY NOW??????. Lastly, i need to go back to europe NOW. its like a&nbsp; fricken drug man, I want more of it. Mainly the alcohol purchasing at the age of SIXTEEN. but the places were nice too, and so was this girl that i liked. w/e tho, its all behind and now i get to look forward to..... o wait, thats right, NOTHING! NOTHNG!= summer reading, school, life at home with my family, and the list keeps going and going and going. peace chauncey '),(25317,3722603,'20,July,2004','2004-07-20 00:00:00','   ive only been home for 4 days, and im already sick of being here.&nbsp; im probably just tired of being at home with my parents. in europe its basically like being on your own. maybe thats what i liked about it so much. that, plus the fact that i can buy my own alcohol.............but seriously. I CANT FUCKEN STAND BEING HOME WITH MY FAMILY. IM READY TO GO TO ENGLAND ALREADY. but i have 11 days until that happens. im just going to have to bare down, and deal with it. im fucken pathetic, im bitching about having a home to stay in, LOL, fuck it. chauncey '),(25318,3722603,'17,July,2004','2004-07-17 00:00:00','   so i just got back from my trip to europe with the SFYO. in short, it was just fucken crazy. everything you think i did, i probably did. we went to germany france austria and NETHERLANDS, home of amsterdam, home of the red light district which is a place where prostitution, and weed are legal( throughout all of the netherlands tho,). so you would see a prostitute in every window practically, just waiting to have a client.  justine patrick and myself went on a mission. we tried to take pictures with the prostitutes to bring back home. LOL, but see, we didnt want the hookers to feel us up like they do to all other customers, so patrick and I often told them we were gay so they would talk to us, and not grab our reproductive organs, which ONE BITCH DID. FUCKEN PROSTITUTE, anyways,while patrick and I were doing that, justine would take a picture of us, or the other way around. either way, we took a lot of pictures. but then, justine left her camera on the bus, so off with that bus went our shots, and our evidence of what we did. i bet most people wont believe what we did, but w/e. tour was just fucken great. i mean I was allowed to buy alcohol, everywhere except fucken cologne, germany. but man, i met some new friends, baileys, JD, smirnoff, and cola. Along with maryjane. TOUR WAS FUCKEN AWESOME on the other hand, i was in to this girl who i wont name, but most people know who she is. nothign really happened, but i found out how she feels about me, and now that i know that, it will be easier to talk to her. Fuck it, i shouldn\'t stress&nbsp;out over her. what happened, happened, and nothign can be changed so all i can do is&nbsp;wait until&nbsp;next year when i see her again.&nbsp; i just realized it is like 1 15, and i need to go to sleep. chauncey &nbsp; OH YA, I NEED TO SEE ( I ROBOT, SPIDERMAN 2, KING ARTHUR, ANCHORMAN, AND SOME FUCKEN GIANTS GAMES) fucken giants, i left and they were 3 and 1/2 up, now they are like 1/2 out. \" warriors about to sign derek fisher?\" about time, we got an actual player. now all we need to do is get one more player liek him, and we may have a decent team. cept we dont have a center besides foyle cuz dampier left. starting rotation derek fisher, jason richardson, troy murphy, mike dunleavy, and adonal foyle. i mean we do have some bench, clifford robinson, calbert cheaney, speedy claxton. if we add someone along ther skills of antanw jamison, we would be set. WHY DID WE TRADE HIM??????? our rotation could be somehting like fisher, richardson,jamison, dunleavy and dampier. YES we would ahve dampier cuz we would want to stay cuz we woudl have a good rotation. BUT THE FUCKEN WARRIORS ARE STUPID SO THEY ALWAYS FUCK SHIT UP FOR THEMSELVES. oh ya FUCK THE LAKERS. if only they couldve kept shaq, seriously. say goodbye to their dynasty. payotn and malone i THINK are finished, and all they have is kobe. they will always bebetter than the warriors, but they wontbe the same unless they get yao ming. '),(25319,3319097,'15,May,2004','2004-05-15 00:00:00','    hEy bl0gx~ todae i went down to eat with my parents as usual... i woke up at about 10am... hmm... after tt dong hong called mi again to jio mi go cycling at east coast with ziyang... i hav to get my present ready for my darling bdae tmr... hehe... i oways do things last min 1... so bad rite? hehe... so i went bugis alone to buy sth special 4 her... hehe... dun tell u.. =pPp... after tt i went dhobby ghaut to get my darling a bdae card... coz bugis de not v nice... hehe... so i went plaza singapura... after tt i still rmb my darling sae she feel lyk to eat garlic bread... so i intend to bought her 1 from those selling western food de... but it taste not v nice so i went hougang mall de pizza hut to buy 1 for my darling at compass point de library revising work... hehe... pizza hut\'s garlic bread is de best after all... finally, i reached compass point... i quickly stuff the present into all my pockets so tt she cant see it... hehe... naughty mi... after a while, i cant find the library... the stupid directory oso dun hav... den i find from floor to floor... till 3rd floor... den i look up to the 4th floor, i cant see ani sign of library... i cant giv her suprise liao so i had to call her and ask her to meet mi at the 3rd floor... she came down and i keke tok alot of craps den i tak out de garlic bread and put in her hand... at tt moment, i see her eyes with tears... den she was so touched... den i keep an wei ta... hehe... i sent her to the library at the 4th floor... we hugged each other at the enterance, pple dere keep looking... lol... den i tak nel to my home... i call my darling and tell here i get it from hougang malll... den she sae tt compass point got pizza hut oso... wa... i was so stupid... hehe... i wan to get her a hot 1 so i went hougang, which is near to compass, to buy... now den i noe dere g0rt pizza hut... hmm... after i reach home... jing xiong in msn congrats mi... den i ask him congrats wad... he sae tt i pass amaths... i tot tt de webby got virus cannt visit den i nv visit... heard from wad he sae, i check my results and i was surprise that i realli passed my amaths which only 4 pple in our klass pass... hehe... i was elated... lol... den i check my e maths that i passed.. half of my klass failed emaths... i was a little disappointed that i expect myself to score higher in emaths... but i was still delighted after all... l0l... darling was da 1st i shared my joy with as i called her... lol... tonight liverpool will play against newcastle which is also the last game of epl... looking forward to watch the match... im so happi now... hehe... darling~ dear will oways giv u wad u wan de... treasure u alotz... lurve ya lots...   '),(25320,3319097,'14,May,2004','2004-05-14 00:00:00','    todae is a pathetic dae... lol... i woke up a bit later todae... coz darling study v hard ydae... den late slp... make mi late slp oso... lol... nv blame u ar... hehe... todae oni g0t chem... so ended early... den i wait for my darling at de canteen coz she got geo and phy... after her geo paper den she cum down liao... hehe... hmm... not quite sian la... at home oso nth do... see wad i can do later den... hehe... darling i miss u dearly... hehe... end here le... love my darling alotz... hen ai ni...   '),(25321,3319097,'12,May,2004','2004-05-12 00:00:00','    im so fortunate to hav such a good darling... hehe... my darling treat mi v gd... hehe... hmm... i miss my darling every second n minute de... hehe... no1 can hav my darling except mi h0r... hehe.... so oni mi can hav it... hmm... my darling is a v gd lady... she\'s soft hearted, kind, and friendly... she\'s cute too... lol... i wun 4get her as my darling and will oways ensure tt she is given the best things... hehe... darling hen ai ni w0r... miss ya so much... hehe... love ya foreva....~   '),(25322,3319097,'20,May,2004','2004-05-20 00:00:00','    thUrsdAy,20 mAy,2004  hmm... todae get back my results... hmm... unfortunately my eng fail... my eng p2 gotten back, i failed badly... 18/50... hmm... muz try harder... im not very contented wif my e maths result... 52/80 for p1 and 74/100 for p2... so total 69.5/100... hmm i intend to get A1... i was happi for my phy results... 70/85.... hmm... but my chem did badly... 37/85... social studies i juz passed, 25/50 and my geo 31/50... my chinese even worst... total average 58.3/100... hmm... juz happi for my phy results oni... hope tmr i wun get shock again... bio hor sure fail 1... a maths hope tt i can pass hehe... hmm... todae after sch, i went home to change and meet darling... we went ps... hmm.. we watched blood brothers at 5.10... reaching dere early we got the tickets and we went to eat... darling, as usual, ordered her fav food... lol... den we shared... den after tt we went to yamaha... i sae a nice drum pad... 39.90... quite ex... but i wanted to buy... hehe... so muz save $$ liao... den it was time for our movie... after the show... i do not know wad was happening in the show... its was very very and very complicated... dunno tok abt wad... lol... den we went out of da cinema and saw jacinta n her fellow frens... i waved to her hehe... den we went to breadtalk to buy a strawberry sweetheart cake... it was so attractive... i requested a candle from the lady and she even gave us a spoon... tt was gd.. l0l... den we went to waston and carrefour to get a lighter.. imagine getting a lighter at carrefour... so big place jus for a lighter... wen search from b1 to storey 1 again... dun hav... den we ask the cashier... it was outside at the info counter... so we went dere and get a lighter... dun wanna waste the effort to change again, we checked the lighter if it is working... hmm.. so we headed to east coast park... now everything was ready and the stupid wind keep blewing... hmm... den clever mi go block the wind and finally the candle was lighted... we blew the cake by saying we love each other... hehe... den we eat it of coz... the cake was nice~ hehe... after tt we rent a double seated bicycle to the sea side food centre... we ordered \"la la\" and chicken wings... it was so nice... den we went back to the sea side and relax ourself... the breeze was so strong... i enjoyed myself and we went home after tt.... hmmm... mizz my darling so much... nitex darling... muackz...      wEdnEsdAy,19 mAy,2004  hmm... ydae was my end of paper~ hehe... c00l~ it was combine science mCq question... hmm... i was suck in chem but in phy i was better... so lets see how.. hehe.. hmm... mi, darling, siling and ming kai went to mac to hav our breakfast... hmm... after tt i send my darling home... i went home and changed and meet my fellow soccer mates to hav a match... it was a unexpected match as i didnt expect mi to play tt well.. hmm... after tt leh... i bring darling to my hse to study after her tuition... hmm.. we went my hse downstairs de department store to buy instant noodles... den she cook for mi... so touched... lol... tom yam de.... *yUmMy yUmMy* it was so delicious... thanks darling... hehe... den my mum came back wif chicken rice... den i hav to share the noonles and chicken rice with my darling... coz she dun wan to eat, den i share wif her... hmmm... den darling study and study till nite liao den i send darling home... it was a tired day for darling... hugz n kizzes... muackz... luv ya~   '),(25323,3319097,'18,May,2004','2004-05-18 00:00:00','    hEy~ im taking amaths paper 2 todae... its so difficult... i tink nv study de h0u gu0 ish litat... lol... can pass h0r i can go run field 10 rds le... hehe... =))) todae i wake up late... hehe... abt 6.10 litat... coz too tired liao... den v sleepy... 3 alarm clock oso cannot wake mi up eh... power bo? lol... den meet my darling late... sorri w0r.. let ya wait so long... hehe... den after my paper i went home and pom pom... after tt leh.. i went back sch dere to wait 4 my darling... we went to serangoon garden eat 1st den i send her home... hmm... on my way home... i asked my neighbour oso my school mate if he can play soccer... den we went training at the field at my hse dere... my long pass was lyk shit... i was teaching sum pattern oso... *lobang* i was teaching sum common wae to lobang pple... lol... he told mi the nike advertisement abt ronaldo being lobang by figo.. lol... hmm... after tt i went home and sleep.. *sn0re sn0re* den the day goes litat... hmm... tmr combine science mcq... lets see if i can sc0re well without studying... hehe... hmm... miss my darling so much... wonder was she doing now? shd be studying h0r... hmm... mizz ya darling... muackz... tatax...   '),(25324,3319097,'17,May,2004','2004-05-17 00:00:00','    sUnday, mAy 16,2004   hellox... hmm... sorri bl0gz... ydae din hav da time to write to ya... hmm... ydae i went celebrating darz da bdae... so fun... hehe... we went sentosa and we tak cable car which my darling wanted it 4 long liao... she nv tak b4 marz... so we tak it... hehe... den in da middle of da route... we saw superstar virgo which is damn enormous... lol... hmm... we reach liao den walk and walk and walk... we were perspiring lyk hell... l0l... soo sticky... its long since i went sentosa da last time... den we finally reached the bus stop... we went quite alot of places... musical fountain, images of singapore, underwater world, dolphin lagoon and many more... hehe... dolphin lagoon and underwater world are da best.. coz we get to see dolphine as well as marine creatures... lol... am i being lame? hmm... dis is my first time seeing dolphin in real life... hmm... quite interesting... after tt we tak cable car back to harbourfront... hmm... den we ate at the food court dere... after tt we went tak foto and we saw ms loh... lol... den we hid here and dere... and finally she was gone without noticing us... phew~ hehe.. we went bugis by nel and we bought nice shirts from yi fu ku... hmm... pair shirt leh... hehe... den i bought a real small cake from 4 leaves... mango de... den i request 4 a candle... hmm... after tt we went studyig at mos burger... den pple dere so noisy... darling cant concentrate... after tt we tak cab to darling dere de park... b4 tt we bought a uselss lighter from the prata shop... it cant even light... inside oso no gas 1... den wanna bluff us... stupid... cost 60cents sumore... so ex... den bo bian... we reach liao and was about to light it den notice it... hmm... den i wanted to ask de pple around us, living in the terrace, but dey were talking among themselves... den bo bian... i try v hard to light the candle but oni got sparks... den darling call her hse dere den fren to help... unfortunately, she dun hav.... i continue doing it but it doesnt help... den i use tissue to start the fire... i keep trying.. finally the tissue burns... i was overjoyed.. lol... den i lighted the candle and i sing bdae song 4 darling... hehe... she was happi! hehe... den we share de cake and eat 2gether... l0l... hmmm after tt i sent darling home liao... happi bdaae darling... hope u hav fun tt dae and hope u luv my present... muackz... luv ya lots....    m0ndaY,mAy 17,2004   y0z... hmm... todae nothing much... hehe.. as usual i made ribena 4 my darling... den i freeze it todae... i hold it from my way to sch... den my hand was lyk so cold... l0l... hmm... darling ydae nv study long enough... sorri w0r... made darling oni hav little time 4 studying... hmm... h0pe darling can excel in her exam... lol...muz ace for it h0r... hmm... todae having emaths p2... i tink quite simple... oni got 2-3 sums dunno how to do... hope can score high so to pull my p1... hmmm... tmr amaths p2 liao... hope can score high too... hmm... now miss my darling so much... she\'s having tuition... hmmm... tmr muz jia y0u w0r... muackz... oh ya... juz now i went hospital so see my newly born niece... its at thompson medical centre... we waited 4 a while nia... den i saw my niece... she\'s so cute...hehe... juz born nia leh... so cute... lol... but i haven tak foto h0r she is sent to da nursing room liao.... l0lz... hope my da sao can regain health back quickly... hmm... tts all i tink... luv ya darling... miss u every milisecond... muackz... luv ya lots....   '),(25325,3319097,'03,June,2004','2004-06-03 00:00:00','    hEy bl0gz... i wanna apologise to ya for not writing 4 long... hmm... it was due to my o\'level chinese... aniwae it was quite diff... hope can get my aim results... hmm... ydae was our anniversary... mi and my darling of coz... hehe... darling came my hse and pass mi present... hehe... so nice of darling... den mi nv buy anithing 4 her... feel so bad... hehe... den we went imm and walk walk coz darling wanted to go dere 4 long liao... den got drop a bit of hints for mi... den is darling plan de ma... den we went dere... the giant dere was super big... lol... den darling eat chicken rice... wa the chicken rice hor from far can smell the fragrance rice liao... hehe... den i eat toast bread and chuee kuey... den after tt we walk walk den went to causeway point... from dere i bought sunflower for my darling... she lyk sunflower so much... lol... den she tak all the way... hehe... after tt we took several fotos... once at causeway point and once at heerens... den darling keep saeing her face not nice... den i dun lyk ma... hehe... but i noe darling will sae again de... but nvm... den we found out tt we were late for our harry potter and the prisoner of azkaban if we didnt hurry... we hav 10 mins to fly from heeren to ps... hehe... but we took 15mins oni... fast rite... we hop in bus 16 after noeing tt it stop at ps... den we rush to the theatre... we put our sunflower in the drink slot dere... fortunately, the stalk juz reach the ground den if not the flower cant balance and i might spoilt it... hehe... and it was nice... but i feel tt it was getting from best to better and to nice.... mayb harry and frens grew older liao... hehe... den after tt it was late liao we eat at kfc... den i send darling home liao... hope darling haf lots of fun ydae and happi 1 month anniversary... muackz... and lov ya~...   '),(25326,3319097,'11,June,2004','2004-06-11 00:00:00','    hEy bl0gz... 2 daes ago, darling came my house and make puzzle... we was making and it was so nice... we bought it frm bugis oppo the foto shop... it was so nice of darling... hehe.. den it was hanged on my room... hmmm... wen i miss darling i can see that... hehe... den todae we went tt shop again and buy a higher level 1... it was smaller pieces and there are 300 pieces... last time we make was 108 oni... now is was more difficult for us liao... i went darlings home and we oni managed to make oni sum parts... v hard to complete... hehe... den wen we do finish liao den hang of darlings room... hmm... after tt at 4 i went tuition liao... i was late abit but it was okie as i read darlings diary... hehe... darling was so cute last time... hehe... hmm... i will cherish darling de... muackz.... tmr is my niece, adel, 1 full month... which means 1 month years old liao... den tmr will be a busy dae for mi to help out with errands and cooking... but my beloved darling will cum and help mi... so nice of darling... hehe... hmmm... cya bl0gz... lova ya darlingz... muackz... will cherish and love ya lots... =)))   '),(25327,3319097,'07,June,2004','2004-06-07 00:00:00','    hey bl0gz... so sorri tt i haven update for so long... hmm... todae i came sch early w/out darling c0z her lesson start later... darling wan cum sch wif mi but i dun wan her to wake up early... hmm... after de whole lesson is over, we went heartland n eat at de food centre... darling treat mi eat leh~ so nice of her =))... den i bring darling to see the pink adidas shoe tt i saw wif my sis ydae... it was v nice... but darling dun wan coz got no size for mi... bEcAuSe the shoe is for gers de... den darling wan us to get the same... lol... den we went darling hse de pet shop... de dog dere so notti... keeo cumming near to my darling n scratch her... in the end darling nv gt her hershey\'s food and sumore kena red marks... i so heartpain... hehe... den after tt i send darling home liao... hmm... saturday 5/6 darling came my hse and accompany mi... so nice of darling... hehe... den in the evening tt time we watched the day after tomorrow at jubilee... is was so nice... so gd... the filming was great... it was da best film i had ever seen... so nice... after the show we went walking at the pasar malam there... it was so big but not realli interesting... after tt we went black pearl to buy zhen zhu nai cha... and it sucks... after tt we tak 76 and send darling home liao.... on friday 4/6 darling came my home and cook spaghetti... darling came my hse first to check the neccessary ingredients den we go buy the needed stuff... after tt we cook 2gether and eat... it was so nicee... the pasta was so nice and delicious... i love it... thankew darling... muackz... den after tt we watched drumline 2gether... but at first i was bathing den cum accompanying darling liao... it was v nice... the snare drummer was so *bang* den after tt (oink oink)... hehe... after tt i cook fried rice for darling... it taste sucks... wahhaa... hope darling dun mind... hehe.... coz the rice put in fridge v long liao... hmmm after tt i send darling home liao... muackz.... miss ya so much darling... love ya~   '),(25328,3319097,'19,June,2004','2004-06-19 00:00:00','    hEy bl0gz... todae wake up early coz haf to go for band... den i wake up earlier to make bread for darling... den my darling lyk turning mouldy liao den nv make... den i tak ribena from freezer den i get ready and went downstairs to get pancake for darling instead... lol... hope ya lyk it okie... 1 red bean 1 peanut... den meet darling and went band 2gether... todae band was quite fun though cant use much instru as tian boon, my juniors, oso nid to use, but i get to use from mandl... lol... quite fun... todae got lots of new piece... all small case 1... den stand by mi tuba solo... \"sup sup shui\" la... lol... sir sae we may not be playing for performance... so sad... lol... see mi small... hmm... den after tt darling and mi went boon lay coz intend to watch home of the range at jurong point... den dunno wheres tt place den ask de serangoon mrt personnel... den noeing de place le... we went dere... de movie was average... not realli nice... lol... den after tt we went walk walk... i bought sth for darling... the thing darling bought for mi b4 oso le... hang in handfone de den gort engraved our name de... lol... hehe... den we went to buy baby socks for our hp... darling got 1 but the other side got dirty liao... den cant giv mi... den we buy loh... so cute.. de top dere got half of a doggie face... lol... den after tt we went eat again... coz our mouth itchy lol... eat ramen den got 2 more side dish so cheap 5.90 oni... lol... den eat liao go walk abit oso... den after tt we went home liao... i send darling hm den went home... lol... muackz darling... love ya...   '),(25329,3319097,'18,June,2004','2004-06-18 00:00:00','    y0z bl0gz... ydae mi gave darling a letter... writing all my feelings... lol... den darling so touched... den cry... hmm... darling cannot be so emotional okie? hmm... darling muz keep it h0r... hehe... den ydae darling pass mi present tt she bought wen she go jb.. so nice of darling... its a unique grey tee and orange surf sh0rts... thanks darling... realli love it l0tz... lol... but the sh0rts too big liao... lol... but still can wear hehe... den we bought a collar tee... so nice... stripe de... den darling de is pink de den mine is blue brown de... hehe... den ydae find whole orchard den cannot find my pathetic slipper... my reebok de slipper worn out liao... so pathetic oways wear old 1... hehe... den todae leh we went heerens den finally got mine... hehe... at last den buy 2 f0r $50.... den no darling\'s size... sorri h0r... hehe... den we go tak 2 times of picture den after tt we sae pple taking siemens M65 if im not wrong de balloon... den darling wanted it... coz its so cute... den we went everiwhere for it... den cannot find we ask pple den all giv wrong direction c0z de pple who is giving out will walk all over orchard de... den cannot find we find abt 2 hrs... still cant find... den i ask a couple who got 2 kids... den dey sae the took de last 2 2 hrs ago... w0w... so long liao... den darling so sad... sorri h0r... den i lyk shout at darling coz i wanted to find f0r darling but cant find... den we walk slowly from taka to somerset mrt station coz darling de foor abit pain... den while walking towards de mrt le... i ask darling to wait dere for mi coz realli wanted to find it... den i run to taka area coz dere got 3 big balloons dere... den cant find... i giv my self 10mins tt means to 2.30pm i muz giv darling my baloon liao... balloon muz be in her hands... den i saw a grp of gers haf... i ask dem but dey seemed unwilling den forget it... i even ask if i can bought frm dem... den nvm i saw a couple finally... den i ask den if i could buy from dem... den he sae y? i said tt i nid it... pls... den he give to mi... so happi... thanks dude... i keep wanting to buy from him but he wanna giv mi... he asked mi if it was for my darling, den i sae ya.. lol... souch a kind hearted couple... so i ran to tt place and pass it to darling... den darling cry... lol... so emotional? lol... den we went pasir ris sec to get tixs from our sir... den see their band oso... woo.. thier drummer not bad w0r... but keep rushing nia... hehe... den we went tampines mall to hav our dinner, pizza hut... so full... den after tt we watch tu ran fa cai... not bad la... though i dunno a bit... quite funny... lol.... den after tt we went home liao... hmm... todae quite unlucky... hehe... mizz ya darling... love ya~...   '),(25330,3319097,'16,June,2004','2004-06-16 00:00:00','    hey bl0gz... my darling going msia todae... den at nite cumming back liao... miss her lotz... in da morning darling wake mi up by msging mi... hehe... den after few hrs darling gonna leave liao... wow... dis shock mi sia coz we gonna xchange sim card... muz make it quick... my sim card got auto roam ma... den darlings dun hav so darling use mine so tt she can tok to mi dere lolx.... as im going sch later to take things from ms soh den i wear scch u n tak a cab to darling\'s hse dere... den we walk a while around her area and change sim card with her... den darling so excited coz got 1 decades nv go oversea liao.. lol... den she so excited ask mi dis n tt... lol... relax... den after a while i send darling home liao den i go sch liao... reach sch ms soh not in... den mc wife called her den she gave instructions to lixin den we go hod room her place and tak... so we juz cum sch and take the tape and the reading material, which is a set of paper, oni... cum sch n tak dis... den i tak for hui mei and ming kai... yes... all da wae to sch den litat... den reach sch oni 7 of us go n tak... dunno de rest go where... i tink sure got alot of pple slack de... zZzZz... den after tt i went bishan to get sth special 4 my darling coz i wanted it long ago den cannot find... so pathetic... den finally at bishan\'s popular i got it... lol... so happi... =))))..  den after tt i went seiyu i buy a surf shorts... not bad.. lol... den i went home liao... im so hungry so reach home i eat liao... lol... coz i ask my mom to leave the breakfast for mi she bought for mi and want mi to eat... unfortunately, i cum back den eat and rush out liao... lol...  den my nephew and niece cum... my notti nephew will not be good 1 dae 1.... if good for 1 dae... i win soccer liao... lol... den my niece not used to my hse den keep crying... lol... hope ya can get use to eat okie... i got feed her drink milk h0r... lol... i so clever ^.^... hmm end here bl0gz... miss ya darling... fun time at jb? lol... shd be h0r... love ya~   '),(25331,3319097,'15,June,2004','2004-06-15 00:00:00','    hEy bl0gz... todae i went for band from 8-5 n kinna tired... we hav lots of things to do coz of the late public performance at braddel hieghts cc at 4th july... dis was our first public performance... hope of giv the mps and mayors a gd impression... hope we can do well... lots of pieces nid to brush up and edit and play... more new pieces cumming up... + holidae hmwk... no holidae for mi den... hmm... ydae mi and my darling wen out after darlings tuition at around 5.40pm... we went bugis to find my pathetic slipper... coz mine had lasted for 1-2 yrs liao and sumore the sole cum off... den cannot find... i lyk 1... but no my size... too big my foot... pathetic... den i saw 2 nice 1 from barcode... so ex... no $ liao... den nv buy... lol... i bought for darling a secret present and darling bought for mi 1 too... its secret lol..... so after tt we went home liao... i send darling home till half wae den we went to the nearer prata shop to drink their bundung again... oways go dere 1... i tink de pple noe us liao coz we oways go dere buy 1 cup of bundung and drink... lol... looks lyk cheapo... hehe... while walking back to darlings hse... sth bad hppn.. hmm make my darling so sad... darling dun sad k? muackz... love ya 4eva... no matter wad hppns... our bonds are never broken... love ya n miss ya lots... muackz...   '),(25332,3319097,'13,June,2004','2004-06-13 00:00:00','    hey bl0gz... todae it was a puzzle dae again... since tt dae we din did much... todae we complete it... it was so nice... hehe... we put in lots of effort h0r... though we tot of giving up... but our determination stop us... hehe... my uncle fetch my to darling\'s hse... so gd of him... i reach abt 1+ den we quickly start liao... we did the hamster head first... den the body.. my darling did the outer layer during her free time liao... *play cheat* hehe... den we do till leg v tired back v tired... rest awhile den continue... after tt (oink oink).. after tt we went to prata shop to eat... we went to the prata shop near darling\'s hse de to buy their bundung ocz its nicer... hehe... after tt we hop into the further the prata shop to eat... hehe... 1 cheese, 1 egg, 2 kosong, 1 bee hoon goreng... so full~ hehe... after tt we continue our puzzle as its going to finish... den while we were doing... we were quite tired, so we were playing... hehe... as we were dancing.. we dunno sang wad song by the dance step, den car\'s honk can be heard... i tot is was wad coz its the same as our singing beat... hehe... den darling rush to the balcony and its was her father.. *heart going to drop a bit* hehe... luckily our reaction were fast... i quickly turn darling\'s hp to silent mode... lol... den keep in her carboard... hmm... after tt her dad walk up the stairs den i called him.. he sort of smiled at mi... hehe... den after a while... he wash his car at the porch le... it was so unsual as darling dad oways came home late, but todae it was early... den darlings mom soon came home... heard from her mom, darlings dad seem not happi wif mi... hmm so darling noe tt after i left, she gonna get scolding... hmm... so soon, we concentrate on our puzzle.. and finally we finished it... *phew* it was a great effort put in... hehe... den we took fotos of it and we start to glue liao... while glueing, strands of hair could be seen... we quickly remove it... den i spread the glue evenly... soon it was done... and its time for mi to go to my aunt hse too as my mom was dere... so i left darling hse... otw out, i greet her dad and her dad greet mi too and sae see u... hehe... den after tt darling called mi and said her dad dun lyk her to bring guy back to her room... hmmm... so nxt time i will only enter darlings room wen deres pple... hehe... so bl0gz gd nitex... later gonna wake up at 2.45 to watch eng-fra... it was a nice match... hope eng can win~ hope 2 goals, eng win ( eng eat 1.5), italy win against denmark nxt dae... coz i bet on it... win win win~ hehe... miss ya darling... gd nite darling too... muackz... love ya~   '),(25333,3319097,'24,June,2004','2004-06-24 00:00:00','    hEy bl0gz... sorri for not blogging so long.... hmmm... dis week is such a pathetic week for mi... tues, wed, thurs... all band prac... hmmm... but quite fun... can blast a bit... lol... den todae i try playing the cant tak my eyes off you on drum set... at first i tot it was easy but after a while i noe tt it isnt wad i tot... quite tough... but after learning for mins... i learnt liao.. not band h0r? lol... hmmm... i can zhi you fa hui de... sui xin ying bian... den can apply and fit into the music... coz i can hear ma... lol... too confident liao... coz i tink tt im better den dem... lol... cannot too hao lian... but feel lyk playing... nvm den... playing tuba in tt peice was so shiok too... lol... can enjoy the music... hmm... every piece is a challenge to mi... hmm... abit fuck up wif band leaders... i tink dey still not experience yet... still got lots to learn... so nvm... but the experience 1 more fuck up... act in front of teachers... den true colours are out... got meeting so wad? cant i noe... still nid to bluff... woo... such a nice bm i haf... wahahhaa... nvm la... juz let it over man... gotta get used to it... so sian liao... no point toking abt it liao.. fed up... aniwae... mi went wif darling to east coast park tmr after band pract... b4 tt, we went toa payoh to haf our lunch and oso to tak 155 dere... darling bought a burger for mi at the pasar malam de malay food stall.. hehe... so nice of it.. it was nice... hehe... thanks!  we eat ban mian.... after tt we took 155... it was quite long... den near east coast park... but haven reach.... well.. i noe tt we took wrong bus again... lol... den nvm... tak cab dere loh... reach liao den we noe tt 401 got go... but we waited at the other bus stop tt we last alight, but no bus comes... stupid~ lol... den we went cycling... and haf lots of fun... go shore and play... hehe... got the big rock and play.. den see a couple the the bg rock dere hugging and kissing... but the path is coveered by water, high tide...  den the climb down... quite cute... den after a while we go off le.. coz darling come... den hehe... its okie de darling... i enjoyed myself... *dear bless darling wif no stomach pain* *dear bless darling wif no headache* *dear bless darlnig with no probs* lol.. muackz darling... love ya lotz~!!   '),(25334,3319097,'01,July,2004','2004-07-01 00:00:00','    hey bl0gz... s0rri tt din blog for so long... hmm... so many band pract lately... sumore sch reopen liao... so tired... hehe... hmm... ydae i went pasir ris with darling... we went whitesand shopping centre... we walk for a while den we go downtown east play bowling... but den tt time is 6.20pm.. over 6 pm liao... so more ex... so we din play... we watch ple play a while... den we went to escape juz to play de go kart... den escape oso nv open.. open on public holidaes, school holidaes.... hmm.... so we went pasir ris park liao... den we walk walk and we finally find a place to sit... so quiet... a few pple oni... den a few couples dere... so romantic... hehe... the waves, the rainbow, the beautiful clouds and scenery... so nice... soo wonderful... cannot describe...love it soo much... we sit down chit chat... hehe.. muackz... den we we nt back whitesand sc to eat de thai crusine... not bad... but de drinks sux... lol... woo... longan drink got chlorine lyk swimming pool litat... den after tt we tak mrt to payar lebar... den see mr wong... darling de pe teacher last time... den dey chit chat a bit... den i slp... den after he go liao... darling ask wad hppn to mi... den we tok a while... dear v easily jealous de h0r... nth can make mi nt jealous... so sorri darling... hmm... den darling cry again.. coz i shouted at her... sorri darling... i make u sad oways... rmb the things i told u h0r... muz oways rmb okie? muackz... hmm... sianz.... now miss her so much... she now helping her mum in her housew0rk.... muackz... love u n miss u n muackz...   '),(25335,3319097,'06,July,2004','2004-07-06 00:00:00','    sorri bl0gz tt din update lately... was quite busy... hehe... hmm... on 1st july darling bluff mi sae go help mama do hsework den actually she din do... she go and make doughnut for mi as 2nd month anniversary present... hehe... it taste nice~ thankew darling... there got total 10... 3 icing sugar de and 3 cinnamon and 4 chocolate de... hehe... it was splendid... thS~... den 2nd july... i make sth special for darling as 2nd month present too... hehe... it was my na shou lucheon meat, potato, carrot in tomato sauce wif rice... den got cabbage too... cabbage is my maid cook de... coz if no cabbage den oni got tt den wait eat till sian... hehe... darling lyk it ma? hehe... hmm... den on 3rd july i went school for listening chi o\'level exam... den my band was having rehersal at braddell heights cc for the next dae performance... den it ended later den expected time... 1.30 den end at 2.10... we quickly rush home by taking cab.. i send darling home 1st den i went home... coz we rushing for SYF opening at national stadium... we rush till lyk wanna die litat coz we meeting darling\'s sista n her stead at 3.15pm at kallang... den we delay till 3.30pm... we reaach dere juz in time den dey sae dey will late... we follow de crowd to the stadium by walking den reach liao dey oso late... wen it was abt to start den dey cum... we did so much juz for dem den dey late... den nvm... we lwft space for dem den dey go sit beside the auntie which is nxt to us... den wan sit wif us... den we change place juz for dem... hmmm... den de uniform grps march in... den is de central judging of display band competition... it was de final... deyi was good... tk was not bad... spfd and pingyi combined was okie... bowen was weaker compared to last time... dunno y... de dm drop mace twice... den deyi won de best dm of the year... tk was de winner... din expect dem to win... i tink deyi can win de... den we went toa payoh to hav dinner, which is mac... we shared den we full liao~ hehe... den darling went to my hse... and oink oink... hehe... den on 4th july we wake up early for our performance... 1st performance w0r... we reach early... den after all briefing and tuning... we went for a break for a while... i saw kelvin, my darling de neighbour hu still lyks her... den i saw darling de fren from beatty band, sher lin and her dm.... den my mum and sis and aunt juz came on time... hehe... den saw my darling de family.. whole family... hehe... her dad so cute cannot play de timbaleh still play... lol... den darling cut of wrongly for march in den she v sad w0r... den cried again.. hehe... dun be sad le h0r... den after de whole thing, we tot of going to changi airport to do my hmwk... den we end up in pasir ris cycling...  coz v sianz.... hehe... den both of us was v tired... wen going to payar leber in de mrt... we overslept to kallang... den hav to tak cab home... hehe... so cute of us...~ den todae we went sch den we release early coz got wad emergency excercise for teachers... den we release at 1.15pm.. darling came my hse to study... den darling came and burn some songs, den log into her account a while in msn... den study ting xie... den i do chem and abit of maths... den went home liao... enjoy w0r... den my nephew make mi so jealous... dun tell ya~ hehe... muackz darling... miss ya and love ya... ~~~ =DD   '),(25336,3319097,'12,May,2004','2004-05-12 00:00:00','    Dear... thx wor.. helping mi to create this blogspot... hehe... i got a very gd dear... he treat mi super gd, teng mi lot lot... i miss him every second and minute... hmmm... i got to know my dear thru.... band? lolx.. we seldom tok in band... until one day, we went out together wif siling n ming kai... from that day.. we name ourselves as Lam4... lolx.. we often go out together and play until quite fun... dan dat time dear dear ask mi to be his godsis... so we became kor n mei... lolx... one day... we went to east coast park n cycle... i really enjoyed myself wif my dear... although dat time we were kor n mei relationship... we returned the bicycles at around 8 plus to 9.. .we walk to the bench and sat down.. i started teasing him abt the dm2 thingy.. he started to tickle.. and we went on n on for a long time... we decided to take a cab back... on the journey back to my place... sth happen.. hehe... shhh... well, we started on the 2 May 2004. dear.. just wanna say i cherish u a lot, i miss u and love you~        * If i were a head of a lettuce,         I\'ll cut myself into two.         I\'ll give the leaves to all my friends,         And save the heart for yOu....  No matter rain or shine, i\'ll be by your side... Ai ni 1314... 3344.... muackz..   '),(25337,3319097,'20,May,2004','2004-05-20 00:00:00','    Hmmm... today dear n i went down to ps and watch movie... we brought the 5:10 tickets for blood brothers... it was only 4 o\'clock then... so we went to the food court and eat... dan we walk walk... sOon, it was 5 and we went into the cinema... hehe... i hugz dear tight tight... dan keep pressing his egg yolk... lolx... Oh... u must be wondering wats egg yolk gotta do wif dear right.. hehe... dun tell u... =x only dear n i know.. lolx... during the show, dear kept kissing dar.. hehe... so cute... when we watch finished, we went out thru the exit and saw jascinta and her frenz.. she was rather shocked when she saw us.. hehe... but nvm lahz... so we went to bread talk and bought our strawberry sweetheart cake... hehe.. is heart shape and it looks very attractive.... we lack of lighter!! haha.. we went to watson and search... they say dun have... i saw rose and his mum at the watson too... later, we heed to carrefour and search.. from level 1 to b1... dan up to level 1 again... at the end... is outside carrefour.. the information counter... phew... at last... we took a cab down to east coast park and find a place to settle down... but, the wind was too strong... cannot light the candle... haha... in the end, we found another sit wif table and after many tries... we finally lighted the candle and b4 we blew off the flame, i told dear, \"i love you\" and he replied, \"i love you too..\" hehe.. sweet? dan we counted 3 and blew!! ~~ hehe... we rent a double seated bike and dear cycled all the way to the food centre... dear intent to go there and order a lot of food... but i was so fulled... dear must be disappointed.. sorry wor... poor dear.. he had to cycle all the way here to eat and all the way back... dar so heavy.. dan plus the weight of the bike.. poor thing... sorry... pain pain not? muackz.. we went back to our \"sit\" and started tickling~~ hehe.. dan i told dear its pain pain.... dan dear nv tickle le.. =) we hired a taxi back.... thx u dear.. i enjoyed my day wif you... even without buying things for mi, spending money on mi or going to far places...., as long as ur heart got mi... i\'m contented.. i treasure the moments we spent together... i love you.. muackz... u brought sunshine into my life... dear, miss u every second... muackz.. hugz... thx u so much... love u soooo deeply.... =D  -=* dAr *=-   '),(25338,3319097,'19,May,2004','2004-05-19 00:00:00','    Hey hey.. today finish my paper le... wooo... so song... hehe... dan now finally got the chance to come online and update ya.. hehe.. muackz... later going out wif dear dear... hmmm... quite happy when i think of it.. actually wanted to watch shrek... but then tml dan open.. so today dunno where to go.. dear dear ask mi plan.. haha... @_@ headach le.. hehe... i always bluuf him when he ask mi to plan later go where... wat to eat.. or ect.. hehe.. dan i say i headach... wanna faint le... hehe.. dan dear dear will let mi off.. ~~ =D hmmm... now waiting for his reply... my mum asked mi accompany her go ntcu... dan later meet dear dear... hmm.. dear ar.. miss ya so much... muackz.. hugz.. see ya later... muackz.. hugz... hehe.. tata.. take care...   '),(25339,3319097,'19,May,2004','2004-05-19 00:00:00','    Dear diary... sorry wor.. long time nv update le.. my brother everyday at home.. cannot use... hmm... dear dear... i feel so touched after i have read ur blogspot... i\'m at dear\'s house.. he went to pom.. so i quickly write a new one.. hmm.. on 16 May, which was my birthday... i celebrated wif my dear... =D so happy.. i long wanted to take a cable car.. see hows the feeling like... and finally... dear n i went to sit the cable car... it was quite expensive.... dear dear bought mi a drangonfly neckless from perlini silver... omg... its my favourite neckless... dear~ thx u !! it cost $45... heartpain~~ well, continued from the cable car thingy... hmm... we alighted at sentosa... dan we intend to walk to underwater world... thought was quite near... so we walked n walked n walked... but still got a long way to go... haha.. we walked back to the main road and follow the sign board back to the bus stop... decided to take the bus to the destination... lolx... we reached underwaterworld le!!! horray! haha... but... we looked at the price... $17 plus... so ex.. is actually for two places... the dolphine lagoon and underwater world... but i thought is only for 1 place.. we saw a paper somewhere at the exit of underwaterworld... it stated.. 3:30.. dolphine lagoon.. it was 3 plus then... we quickly took the bus down to the destination... we saw the price again ... $17 plus.. but no chioce... come here already... so we queue up to buy ticket... chey... the person told us that $17 plus for two place... ok... we bought it and went in to watch.. hehe... i always hope watch dolphine performances wif my loved one... hehe... my wish is granted... =D then, we took a bus back to underwaterworld... we entered and walk around... woo... its so nice.... hehe.. after that, we heed back to world trade shopping centre... we went to take card photo and saw miss loh wif her ai ai~~ hehe... then, we went to bugis and buy the couple tee shirt... =) so cute... we went to take another neo print... hmmm... b4 we went hm... dear bought a small mango cake for mi... so sweet of ya, dear... he wanted to bring mi go east coast beach and celebrate.. but the next day got 3 papers... i haven\'t study at all.. dan i told him i wanted to go home... sorry wor... spoilt ur plan le... next time k? sorry... dan we went to my hs the playground.. hehe.. stupid lighter!!! hai wo de dear dear de hand pain pain and black black.... dear, just wanna tell ya that i really treasure u a lot.. i love the present u gave mi... i love u i love ya... thx u for everything u have done... muackz... miss u lot lot... ai ni... yap, i enjoyed myself wif u on my birthday.. i calculated le... u spent $142 wif mi.. or maybe more... sorry wor... but still must say xie xie ni.. muackz... love ya lot lot.. huggz...  -=* shen ai ni de ren *=-   '),(25340,3319097,'24,May,2004','2004-05-24 00:00:00','    Hmm... today morning.. as usual meet dear in the morning.. i looked at the sky... @_@ going to rain liao.. i quickly get myself ready b4 it rains... b4 i start to pack my stuff... my mum told mi is raining... first thing dat appear in my mind... \"oh no.. rain le... later dear caught the rain how?\" dan i quickly take my umbrella... dear gave mi the umbrella a bit small.. dan later dear sure ask mi to cover myself only... so i purposely bring a bigger de.. hehe... but, as i reached the bus stop.. it has stopped raining.... intend to cross over and sheltered dear... god nv give mi a chance to touch dear\'s heart.. always is dear touched dar\'s heart... hehe... hmmm.. after school supposed to meet dear at 3rd level there.. but after waiting for 20 min, my stomach suddenly very pain... oh no.. gastric pain le.. i know i have to eat something b4 it got worst. so i quickly went for my lunch dan go up n find dear... think dear a bit worry.. coz he cannot find mi... sorry wor... hmm.. we heed to band room... dan started to take out my instrument.... dear missed his tuba a lot.... he kept practicing a lot... hehe... but after a while... he left le... miss him so much... muackz.... hehe...dear dear.... sorry wor.. sometime mood swing... dunno wats wrong wif mi.. dan nv tok much.... let u worry again... hao xiang ni wor... On sat, dear n i went to ps and watch movie..shrek2.. i long wanted to watch it.. so nice.. hehe... we kept luffing.. dan later we went down to far east... find the stupid skirt... i saw at teens.. very nice... dan they give wrong informations... make us searched for so long.... we went to the information counter ask.... he said the shop we looking for is actually KFC!!! omg... dan nvm.. guess dear a bit anfry le.. coz we found so long.... dan cannot find... we walk inside the level 1 and suddenly.. saw a shop named banana... guess that was the shop we r looking for... yap! it is!!! i saw the skirt... dear say not nice.. haha... coulour weird... dan we take a bus down to esplanade... dan dar naughty... want to take the river taxi... dan dear spent $18 on it.. sorry wor.. but i really enjoyed myself dat day.. we went to arcade and play... aiyo.. poor dear had to spend another $10 on it... we collected a total of 610 ticket and we changed a piglet, 2 cute kepchain and 2 funny bookmarks... haha... we play the games until hand pain... haha... so funny... but dear spent money to gain dar laughters... thank u dear... xie xie ni wor... muackz... i will repay u by...... sanding by ur side 4 ever... hehe... lolx... hao xiang hao xiang ni... muackz!!!! dear... i love you.. sorry if i did anything dat make u unhappy... sorry... u can tell mi de.. is alright de k? muackz.... treasure u 4 ever.... dun wanna loose you~~ 13-=* 02.05.04 *=-14 hehe... 02-=* daN.vaN *=-05   '),(25341,3319097,'05,June,2004','2004-06-05 00:00:00','    Hmmm... yesterday after school, dear waited for dar at canteen... dan he sent mi home and change while he went to the prata shop n eat a bit... coz later we going to dear\'s house to cook spagatti.. once we reached his house, we checked the ingredients dat he have in his house. Then, we procceeded to the nearby provision shop to buy the ingredients... Oh no! we went around to look for tomato! Oh gosh! we searched most of the provision shops but they r out of tomato! haha... in the end... we left out tomato instead.... Once we got home, we started to cut the ingredients... Dear chopped the onions while i looking for another cutting board to slice the hot dog. I took out one big cutting board and put in on the sink, suddenly... a crocokrache ran on the cutting board... i screamed a loud and i guess dear was frightened by mi not the crocokrache. haha... i asked dear to catch it... but the stupid crocokrache was running fast.. hehe.. dan later dear managed to kill it.. haha... so scary... sorry dear... i think my reaction was quite big... hehe... dear washed the cutting board thoroughly... dan i continued to slice my hot dog. hmmm... while preparing the ingredients, we went to boil the spagatti... well, i like to eat soggy soggy soft soft noodle... dan dear also accompany dar to eat in this way... sorry wor... soon, the spagatti was ready and its delicious... haha... well, dat was my style mahz... but can see dear trying to finish the spagatti.... maybe ate too much feel a bit funny... not used to it i guess... sorry... hehe... dan dear washed the dishes... he dun allow mi to wash... coz last time when i came to his house and eat... i will washed for him... dan now he asked mi to sit aside... waiting waiting waiting~~~ finally he washed finished and he let mi see the drumline show... lent from mr jackey tan... while i\'m watching.. he\'s bathing... so i watched alone.. the starting part was quite boring... but to the middle and ending part.. woo... cool... nice... haha... hmm... when i watched finished, it was already 7.20 le... dan we went to dear de room and sleep... hehe... *(cut off, NC-16)* At about 9.35... dear cooked fried rice for dar... yummy... fist time dear cooked for dar... i wanted to finish the whole plate but was rather full... sorry wor... its nice... really... just dat the rice a bit hard... butt overall... i like it... really.. thank you lao gong... I had an enjoyable day yesterday~~ muackz... love ya... hugz...   '),(25342,3319097,'04,June,2004','2004-06-04 00:00:00','    Dear diary,       Its mi again. Now i\'m at computer lab... hao meng wor... i forgot to bring my f&n disc. I thought the disc was wif ms lina... dan din bother to pack my stuff... hehe... this keyboard sux man... the spacebar spoil le... must press so hard~~ hmmm... yup dear... thx u... i really enjoy myself on wednesday... dear kept spending money on dar... dar heart pain wor... guess now everyone doing their work so seriously... except mi... hehe... yesterday i went out wif ms lina and felicai koh... we went to raffles hospital and eat.. coz their canteen the food is all hala including chinese food... gd hor? ms lina is a malay person... like dat, she can try the chinese dish... after eating, we walked around parco bugis... we told ourselves that we must go for window shopping... but in the end, ms lina bought one skirt and shirt.. haha... she spent around $60 plus.. hehe... she was quite pek chek also.. coz felicai and i kept repeating dat she spent a lot... hehe.. dan after dat, we went home early... dear came n find dar around 8.45pm. So touched... hao xiang jian dear... dear ye hao xiang jian dar... hehe... dan we walked to the nearest park.... we sat on the swing and hold hands... dan played together... sweet? hehe... after a while... dear want to go to the prata shop n eat~~ haha... dan we walked there and ordered 2 cups of bungdong, 2 plain prata and one plaster prata... dear nv try b4 wat is plaster prata mahz... so we ordered... hehe... we very en ai de wor... we share share...~~ hehe... dan after eating.. dear sent dar back... we walked the longer path... coz can spent more time together... hehe... hmm.. b4 i walked to my hs, i hugged him tight tight and kiss~~ hehe... hmm.. i end here le... next time dan continue... muackz.... hao xiang ni... take care dear.. love ya lotx lotx.. =D  -=* 02.05.04 *=-   '),(25343,3319097,'03,June,2004','2004-06-03 00:00:00','    Dear.. happy one month anniversary!! love ya.. hmm.. well.. it was yesterday 2 June. Hmmm... i went down to dear house and passed him the present... It was his anniversary present. hmmm... dan reached his house about 1.30pm. Dan he put the present down and he said that i wore until so nice... so he went to change his shirt to another long sleeve button shirt. We took a cab down to IMM which cost about $12. We explored the whole place... dan walk around... suddenly... we smell something fragrant... we looked around and saw a chicken stall... comfirm the smell came from there... hehe... dan i went to eat chicken rice and dear ate toast bread and chee kuey. We shared around mahz.. hehe... i wun ill treat dear de...after dat... we went down to causeway point. Dear bought mi a stalk of sunflower... hehe.. dan i kept holding it... so happy... we went to take photo wif the sunflower... dan my face so er xin... =x spoil the photo.. really... haiz.. so sad.. dunno why... hmm... dan we went to heeren to take another photo sticker wif the sunflower again... haiz.. also same.. my face also same... hao er xin also... so shi bai... sobx... dan we catching a show at 6.40 but we at herren at 6.30 ... omg.. how? wa.. we quickly rushed to ps.. took a bus no. 16 down to ps. phew! on time... the show harry potter haven\'t start... hehe... dan we sat down dan dear hug dar tight tight... hehe... so sweet... after the show... we went for our dinner at kfc.... yummy... last, we took a cab back to my place coz it was quite late le... dear dun want dar to reach home late.. caring hor? muackz.... thank you dear.... i love u... thx for the sunflower too... muackz... hugz... happy one month anniversary... -=* 02.05.04 *=-  -=* 02.06.04 *=-   '),(25344,3319097,'09,June,2004','2004-06-09 00:00:00','    Morning blog~~ hehe... actually supposed to meet dear at 8.40 but then he reached at 9am. Hmmm... its alright dear.. dun worry... muackz... i dun mind... i know is becoz of the bus fault... not ur fault... i have to come out early coz i scared my dad will asked mi hu i meet... wif a guy/ger... haha... so to prevent from all these... i come out early and wait.. is also my fault to come out early.... so now is nobody fault k... hugz..hmm.. dan later dear n i walked super fast.. coz he scared i late... haha.. well, i was a little late... but some of them came in after mi... hehe... hmm.. dear went to play soccer wif his frenz... i can see him from upstairs... so cool... woo... after school, dear will send mi hm to change dan we will procceed to bugis to buy puzzle... dan later we go back to dear\'s house and fix... well, hope later dun rain wor... =) hmmm... well.. it will be difficult to update u blog! coz after today... we need not come back for f&n.... dat means no com for mi to use le... sobx... haha... hmm.. now, i\'ll end here... love ya lots; dear! ;) muackz.. take care....   '),(25345,3319097,'07,June,2004','2004-06-07 00:00:00','    Dear diary,     Its mi again... i guess after this wednesday.... i can\'t really update u anymore... coz com broke down... haha... but strange thingy, my younger brother can on the com and play his game.. but when i switched it on.... it hang! not fair... =( haha... hmm.. yesterday went out and ate wif my mum for breakfast/lunch... haha.. dan came back quite boring... so dear asked dar if want to go dear\'s house and cook? but then around 6:30 i have to leave his house coz i going to my grandma\'s house and eat. So maybe just go his house and tok tok play play.. haha.... hmm.. he\'s wif his sis at si ma lu temple. dan maybe his sis wanna go buy thingy... so asked dear accompany her... dear told mi around 4 plus dan he will come... but then they still at suntec area... dear very pek chek.. coz his sister wanna see more things i guess... i understand how dear de feeling... so dear asked dar to take a bus down to his house dan he decided to go north point and walked instead to his house coz his sister at home... so we take bus down to city hall and change to north point.. but the bus passed by esplanade... so dear suggested dat we go there and walk walk.... nice breeze.. haha.. dan i saw cheyenne, my cousin in there... haha.. she\'s wif a guy... hmm.. next, we walked to the ice cream uncle and bought ice cream and ate... we spent our time there and later took  bus no. 133 back to ang mo kio, where my grandma lives... But then, i was late.. my mum told mi that everyone was waiting for mi to start the dinner... so we alighted and took a cab down to my grandma\'s house. Thank u dear.. for accompanying mi... love ya... later, dear went home le... muackz.. now rushing my f&n project le... tata... muackz, dear!   '),(25346,3319097,'19,June,2004','2004-06-19 00:00:00','    dearest blogspot.. i think i am very lucky in the world... i got such a wonderful dear.. he is daniel ong qi ming.... ok... the story goes like this.. yesterday... dear and dar went down to hereen and continue finding his slipper... at last... we found it... actually buy two at $50.. but no my size.. argg.. hehe.. nvm.. dan i ask dear buy first... dear was so happy... haha.. dan i walked further in... saw neo print machines... hehe.. dear asked dar wanna take not.. but dear know dar sure will take de.. hehe.. dan we took twice.. hehe... so nice... dan i very happy... when we walked out of hereen.. saw many ppl took the M65 siemens phone balloon... omg.. is so cute... we asked some ppl hu r taking the ballon... they pointed the direction.. dan we walked and searched... searched and searched... dun have.. all fake us de.. so tiring... dan later i very sad.. decided not to find le.. coz the last couple took it was 2 hrs ago.. haiz.. dan we walked to somerset mrt station the time.. dear asked mi to wait for him at the this fashion outside.. asked mi to moved around.. i know he going to find the balloon... so i stood there waited for him... dear msg mi asked mi to wait for him to come back.. hmm.. at 2.30pm... dear ran back wif the m65 siemens phone balloon... omg.. i cried and hugged him... i was so touched... sNiff sNiff.. dear thx u! muackz.. dan we went down to pasir ris secondary school and get the concert tickets from sir.. but reached there.. sir gave mi $8 tickets.. means have to sit at the bottom... humph... i told sir b4 i wanted $10 tickets.. sit at the top... but the got the wrong one.. haiz.. dan sir said will get for mi.. phew... hehe.. dan we went to listen their the band a while... okok lor... hehe.. dan i hungry le.. i asked dear to make a move first... so we left and we ate at tm pizza hut.. hehe.. sooo full.. later.. we catch a movie at 5.20pm.. the tu ran fa cai.. a little bit funny.. haha.. dan later we watched finish.. dear send mi home.. thank u dear.. for giving mi such a happy day.. although sometime we r very unlucky nor.. haha... oh ya.. i remembered.. on thurs, i met dear dan we went to far east and find his slipper... dear passed mi a letter when we met.. dan i gave him the present i bought for him when i went to jb.. hmm.. when i read the letter... i was so touched and cried... dear!!!! this was the first time u write letters to mi... u expressed out all ur feelings... thank u dear... i love ya lotx lotx... today after band... we went to jurong point and watched home on the range... dan b4 dat.. we went to banquet and ate.. a lot of ppl over there.. so i just ordered a plate of malay rice wif 5 dishes... so we shared.. hehe... dan after the movie.. we went to triangle and make the handphone keychain dat i bought for dear.. wif our names crafted... so dear went to make for mi too... the person asked us to come back in 10 min times.. so we went to shop around.. i guess we shopped for more dan 10 min.. we went back to the shop.. she said haven\'t ready.. dan dear a bit pek chek le.. so we walked around and we went to kopitiam and ate ramen... so full... when we ate half way.. dear went back to the shop and collect the keychain... i sat there alone waiting for dear to come back.. but waited so long... sort of like knew that they haven\'t make.. dear called mi and told mi he\'s waiting.. they haven\'t make... going to ready le.. dan after dat.. dear came back dan i see the keychain... well, the letters not really nice but then nvm... a present from dear lehz... i would appreciate it..thanks! hmm.. later.. dear sent mi home le... wonderful day today... muackz...! love ya dear... hugz...   '),(25347,3319097,'17,June,2004','2004-06-17 00:00:00','    Hey blog... muackz.. hehe.. so happi... yesterday went to jb shop till drop! haha... i know u must be thinking... \"wAhaha!!! go jb only so happy!!\" but then.. to mi... its a joy to mi lor... hehe... dan yesterday morning dear took a cab down dan we exchange sim card... coz his got auto roam... dan i can msg n call him... sorry dear.. hai ni yesterday so bored at home... once we reached there... we go to their food court and eat.... i walked around the food court dan see the dish very sianz.. =xXx choosy mahz... hehe... dan i saw the japanese cruisine... so i went to take a look... i saw their today\'s speical.. the chicken piece so big.. dan i ordered... my few other cousin went to orderd some other dishes from the same stall.. we waited for so long.. dan finally the dishes are out.. my pathetic chicken sooo small~~ omg... am i seeing the wrong dish?? haha.. i was astonished!! omg... can u believe it.. all our dishes dat we ordered are sooooo less... almost half of wat i saw in singapore.. haha.. dan bo bian... after eating... is shopping time!!! woohoo... yeAhz!! waited for long.... dan i bought some stuff for myself.. well, girl need to be pampered at times... haha... so i pampered myself nor... lolx.. so shuang~~ ops... is so nice~~ hehe.. dan sometimes i will call dear and bao ping an... or chat wif him.... hehe... dan i nearly lost at there.. phew! hao sian wor.. i bought the sunflower photo frame... the salegirl took her own sweet time... aiyo... dan i got the feeling like.. my mum and aunties all gone.. really.. when i got back there... all gone le... i panic le.. hehe.. dani told myself to stay claim... i glanced the whole level dan i went down to another level... dan i saw them.. thank god! dan i follow them closely le.. hehe... bought lots of stuff... got shirts, skirts, ear ring.... hehe... dan walk n walk... last shop at factory outlet... bought one shirt and shorts for dear.. hehe.... i like it mahz.. dan i dunno dear like not.. hehe... =xXx but then dear say wat dar buy dear sure like it... later dear see le dan if not nice also must say nice nice... haha.. hmm.. dan we walked around and shop again.. reached home about 10.25.. dan i go bath... dan called dear about 10.35.. dan we chat until 12 like this.. dan we go slp le.. hehe... miss ya dear.. love ya.. hugz.. muackz...   '),(25348,3319097,'16,June,2004','2004-06-16 00:00:00','    Hmm... on 14th June... we were supposed to find dear slipper... but end up i kept looking at my stuff... dear tried a few slippers but they are either too small or too expensive... haiz.. dan in the end... we walk around bugis dan we went into seiyu... hehe.. i bought something nice and cute for dear and dear bought something attractive and beautiful for dar... lolx... shh.. dun tell u.. hehe.... after that, we went to simply thai and ordered one fried noodle and thai iced tea... not bad lor... dan after that.. we walked to b1 and wanted to buy tori-q... but then.. no more lunch box le... dan we walked around.. i saw one bag.. so cute... and nice.. intend to buy but i got no $$... so dun buy le.. dear actually wanted to buy for dar.. but then nvm de... dun keep spending money on dar... so just forget it then.. hehe... next, we took bus back to my hs... dan on the way.. we went to my hs outside the prata shop and drank bungdong... after that... dear intend to walk mi back through another direction.. but then, i insist on walking the nearest path back... in the end.. gave mi papa saw... he screamed at mi like siao... -.- so scared... haiz.. dan let dear worry for so long.... sorry dear... should have listen to you... muackz... ai ni dear... take care... later i going to malaysia.. jb... shop shop... dan just now dear took a cab down and meet mi... dan we exchange sim card... coz his line can use for overseas de.. hehe... dan later i go there can call dear... =D so happi... bye dear.. love ya always... ya.. our bonds can never be broken... muackz.. hugz...   '),(25349,3319097,'14,June,2004','2004-06-14 00:00:00','    hEy blogx... sorry wor.. long time din update ya... hmm... dat day, dear n i went to buy puzzle and we did at dear\'s house... haha.. we got the big puzzle and inside contain 108 puzzle... looks interesting and cute.. so we did it... after that, dear hung it at his room the wall.. haha.. so nice... dan we oink oink... lolx.. hmm.. dan on 12 June, i went to dear de house again. His house got baby 1 month... so cute... dan i went there to help up but sort of play instead.. haha.. dan dear de hum n aunt cook a lot... so nice.. so dear n i ate a lot wor.. i ate more dan dear... hehe.. dan we went up and watch tv, lie on the bed n play.. play will dear de aden.. haha.. he so cute.. he kissed mi~ lolx.. hmm.. dan we slack until 6 plus.. we went down for dinner dan we went to mount faber...took a cab down which cost $9.50. dan we walk walk n took photo... dear the left side slipper spoilt.. haha... after walking n enjoying the breeze, we took a cab to habourfront dan we went to the habourfront food centre and ate... hehe... i guess i gonna do more exercises... coz i kept eating.. haha... dan dear send dar home... On 13 June, dear came to my house and we continue fixing the puzzle dat we fixed dat day... not yet complete mahz... we bought the 300 puzzle and is small de.. so difficult.. dear n dar wanted to give up but then we never... =) after dat.. we oink oink... hehe.. we continued fixing the puzzle half way when we decided to go to the prata shop n eat... after eating, we went backz and continue fixing.. we were dancing n singing song when my dad came back... hao xian wor... he nv scold mi .. for briniging a guy into the room when no one at home.. he just talked to mi... dan dear went to his aunt place... sorry wor.. cut short a lot.. coz my brother wanted to use the com le.. cya~~ muackz.. love ya dear.. hugz..   '),(25350,3319097,'29,June,2004','2004-06-29 00:00:00','    Hmm.. sorry blog.. din update u for quite a time.. feel so bad.. hehe... last week got lots of band practice... hehe... so tired... hmm.. yesterday dear touched my heart again... i forgot to bring my english newspaper... dan we walked to the mama shop outside school... but then sold out liao... dan bo bian... dear asked mi to carry his books back to my class and wait for him... he run to the shell station and buy for mi.... sNiff sNiff... haO gAn doNg wOrx... xie xie ni... he almost late wor.. luckily he run back in time.. pHew... heNg.. dear dear.. thank u... although i din really express it out.. but my heart melts when i saw him sweating.. and the way he pass mi the newspaper.. so cute and sweet... hmm.. now at com lab... hao sianz... doing the f&n again.. sad sad wor.. miss lina leaving us in a few days time.. depend on the MOE to post her at which school... but dun worry.. we will still contact wif her...=) just receive a msg from dear... he say he turn short and fat... from 179cm, 59kg to 177cm, 60.8kg.. hehe.. so cute.. =) ok lahz, i\'ll stop here... muackz.. take care.. tata.. love ya... =D hUggx..   '),(25351,3319097,'09,July,2004','2004-07-09 00:00:00','    Grrr... blog ar blog... stupid bitch a.k.a meiting sux to the core!!!!!!!!! HUMPH!! so pek chek!! ok... let mi explain to u... today... we were suppose to go for our p.e lesson... dan tse ying and i went to the toilet and change... suddenly.. i noticed that tse ying has stained her skirt... so she went to \"change\"... when we came out of toilet... mr silver hu is our p.e teacher said that we do not have p.e lesson as we r going for our photo taking soon... so tui ar~~ dan we went to change back again... aRhh.. when we went back to class and sit.. suddenly... fire drill~~~ we went to assemble in the field... 1/2 ++ passed.... finally... back to class... as it is Ms See lesson... hehe.. so sophia, tse ying and i went to tell siling dat we r going to the needle work room and iron tse ying\'s skirt as it is still very wet.. while going to f&N block.. saw that bitch walking and playing wif guys... yUcKz.. dan when i walked passed her.. she looked at mi n diao... -.- fUck rigHt? hUmpH! KnN... i guai guai de... nv do make ppl le.. dan she diao mi.. think her face like luo han yu i scared mehz... =Xxx haiz... sorry lahz.. din really want to loose my temper de.. but too pek chek le... i thought this yr can be guai kia... stay away from those vulgar languague.. but then.. sorry.. hehe... tHe sigHt Of heR mAke mE pUkE!!! yUcKx... pUi pui (x 100000000000000000) hUmpH!!! fUckiNG gEr!! hAte yOu... becOz oF hEr... see lahz... i scold so much... hehe... but i feel so much better now... hehe... ok... enuff of her... eEEeee.. wRitE agAiN aNotHeR dAy.. tAta~~~   '),(25352,3319097,'05,July,2004','2004-07-05 00:00:00','    Hmm.. sorry blog... din update u le... quite busy.. sometime lazy to write.. =x Hmm.. on 1st July, after school, dear send mi home.. dan i got tuition... until 4.30pm... i told dear until 5pm... sorry... hmm.. i quickly go n make doughnut for dear as our 2nd month anniversary present.. i dunno wat to buy for dear.. so decided to make for him... more sincere.. hehe... dan i make n make until 7 pm like dat... sweat a lot.. but no choice... change into a clean clothes and took a cab down to dear\'s house... wanted to give it to him my freshly made doughnuts.. but not hot le... dan a bit hard and sweet.. sorry wor... i din taste.. no time to taste... dear was kinda surprise when he saw mi... hehe... dan i gave it to him... On 2nd July, after school.. dear sent mi home... dan i got myself ready and went down to dear\'s place... he gave mi our 2nd anniversary present... guess wat was it? hehe... he cooked luncheon meat potatoes wif ketchup sauce... so delicious... together wif some rice n vege but was cooked by his maid... lolx... hehe.. so delicious.. thx u dear.. yummy.. hungry le... lolx.. dan we went to enjoy ourselves... On 3rd July, after our band pract at Braddel Height CC.. we took a cab back dan prepare ourselves and we head to National Stadium to watch the SYF... 4 bands competiting.. surprisingly, bowen band was out... sad sad... it was tk band.. haha.. their band not bad also... after that, i went to dear\'s house dan ahem ahem... ~oink oink~ again... lolx... hmm.. dan sear sent mi back... On 4th July, is our performance.. so nervous... as i was tunning the band.. i saw my neighbour kelvin.. he was standing there looking... hmmm.. dan at around 10am.. we started playing the songs.. lolx.. after that... at 11am.. i conduct the band... we played for the npcc to march in... omg.. i cut off at the wrong timing... so sad.. my sir asked mi not to be so nervous... haiz... feel so sad... but everything went on smoothly... after the whole thing ended.. we packed our things and took a lot of photo as a memories.. lolx.. shiok sia... hmmm.. after that, ms chan asked for volunteer to go back school and put back the instruments... i wanted to go.. but ms chan dun allow.. -.- but heck... i still follw back coz dear need to go back to school and help out... haha... after that, dear sent mi back again... when we changed finish... we intend to go changi airport... dan dear come down to my place here... we took bus no.70 to paya lebar... from there can take a train down to changi airport faster.. but when we alighted at paya lebar... i wanted to go to the singapore post inside.. coz their bubble tea very nice.. dan there the environment very comfortable and nice.. intend to do hw there... dun go changi airport le.. but dear dun 1 2 do le.. suggested to go pasir ris park n play... so we went there to ride bike and enjoy ourselves there... so nice... after that, we went to settle our dinner there and went home... we slept soundly in the train.. when we woke up.. it was at kallang.. omg... we quickly hired a cab and dear sent mi home.. he scared my dad would come back early.. kallang is not a gd place to wait for a cab... lolx.. dan dear a bit pek chek while waiting for cab.. sorry wor... hmmm.. when i reached home.. i quickly bathe and wait for dear b4 we went to slp.. hehe... so nice... i enjoyed my days wif dear.. thx u.. muackz... take care.. love ya... huggx...   '),(25353,3661997,'16,June,2004','2004-06-16 00:00:00','   June 16th, 2004~ Today was one of those regular, not so special days. But still I had some fun. I went shopping today, and got me two cute shirts, three sexy pairs of shorts, a new purse..Because someone copied off of my other FAVORITE one...No comment on the name. And other stuff...LOL. After that I went to Kirkword Pool, for a couple hours. Once I got back I didn\'t do much more. Now as I am sitting here writing this, my older brother has friends over...HeHeHE! Even so, I\'m bored out of my wits! SO.. CALL ME UP AND GIVE ME A RING-A-LING! LOL   '),(25354,3661997,'17,June,2004','2004-06-17 00:00:00','   Today I went to the pool with my 2 pool buddies, Steph and Kristina. We went to Tropical moose too. Right now were all at Kristinas house! \"Look my toe looks fat EWWW LOOK AT MY TOE!\"says Steph. LOOSER!!! also we egged this guys house and then the next day (There her her neighbors)his moms like \"some one edd\'d us last night\" AND WE STARTED CRACKING UP!!!! and then that night we did a crapy Tp job on the island on the street lol and put water on it so it was stick...the next day is was like gone haha! good times! o and kristina cut stephs hair!! its looks fabulous!! (and i cut a lil bit..like the front like one snip hahaha!   '),(25355,3661997,'18,June,2004','2004-06-18 00:00:00','   Well, me and stpeh spent the night at Kristinas house..we stayed up like all night haha, but we could not shut up!! haha but when we woke up...had a tasty breakfast AND I HAD MILK WITH IT!!!!!! hahahahahaha then we went to the pool for a bit, and then the pool closed kinda becuase of the strom, so then we went to tropical moose iwth ty, mitch, adam, and kevin AND THEY THROW MY FRIES AT ME (WELL ONLY TY DID) THAT LITTLE BUTT HEAD!!! ya, and then i went home and ate the rest of my fries IN PEACE!!   '),(25356,3661997,'19,June,2004','2004-06-19 00:00:00','   Today i woke up at 12:30...I was suppose to met my friends up at the pool at 11...OOPS!!! FUCKING ALARM CLOCK..its a peice of shit haha!well so i went to the pool at like 12:45....i walked around.....WEHRE HTE HELL ARE THEY!!!! they werent here!! then i called lauren b\'s cell \"OH! we went to tromo were walking back now!\" OOOOHHH haha should have called frist...instead of waiting an hour haahha...and my \"cuz\" Mark (eittoh) was working there so i chatted with him a while....then i went home...dog sat all day, I LOVE THE DOG...shes a boxer named sadie! ahhh shes adorable...and shes HYPER like me!HEHEHE...but then everything got bad!!!!!! ahhhh I wasnt invited to something that upseted me....and then i thought this really hot guys called me a bitch...turns out he didnt...(no comment ont he name...altough he\'d perfer it was) haha sux 4 him!! JKJK I LUV YA!!!! but now thats over...but then i was late on taking my Acid refulx medication...which makes me get heartburn...which reminds me..i still havnt tooken it!!!!!   '),(25357,3152463,'31,May,2004','2004-05-31 00:00:00','    Episode 22 : Another day spent doing very little'),(25358,3152463,'30,May,2004','2004-05-30 00:00:00','    Episode 21 : Tracey Emin fights back - brave girl'),(25359,3152463,'29,May,2004','2004-05-29 00:00:00','    Episode 20 : A sinking house, a new cat, and a nap'),(25360,3152463,'28,May,2004','2004-05-28 00:00:00','    Episode 19 : Frozen in French, and Harry\'s shades'),(25361,3152463,'27,May,2004','2004-05-27 00:00:00','    Episode 18 : Tracey Emin\'s crappy tent, plus more'),(25362,3152463,'26,May,2004','2004-05-26 00:00:00','    Episode 17 : A philosophy, Haribo, and peach water'),(25363,3152463,'25,May,2004','2004-05-25 00:00:00','    Episode 16 : Book-snaffling, and my soap reviewing'),(25364,3152463,'24,May,2004','2004-05-24 00:00:00','    Episode 15a : Corrie update - yet another punch-up !'),(25365,3152463,'23,May,2004','2004-05-23 00:00:00','    Episode 15 : Mozart let it all hang out. Dirty boy.'),(25366,3152463,'23,May,2004','2004-05-23 00:00:00','    Episode 14 : Eggicide - Humpty\'s family too, now ...'),(25367,3152463,'22,May,2004','2004-05-22 00:00:00','    Episode 13 : Humpty Dumpty sat on a wall ... sadly.'),(25368,3152463,'21,May,2004','2004-05-21 00:00:00','    Episode 12 : A free mobile frame - huzzah for Tech'),(25369,3152463,'20,May,2004','2004-05-20 00:00:00','    Episode 11a : Save the Pleasure Beach\'s Art Deco !'),(25370,3152463,'19,May,2004','2004-05-19 00:00:00','    Episode 11 : History pays off ... what a great pun'),(25371,3152463,'19,May,2004','2004-05-19 00:00:00','    Episode 10 : Jewology, digital cameras and an exam'),(25372,3152463,'18,May,2004','2004-05-18 00:00:00','    Episode 09 : \"The dog\'s less trouble than you two\"'),(25373,3152463,'17,May,2004','2004-05-17 00:00:00','    Episode 08 : \"Does that word even exist ?\" \"Yes.\"'),(25374,3152463,'16,May,2004','2004-05-16 00:00:00','    Episode 07 : Carriages and Emily Howard ... quaint'),(25375,3152463,'15,May,2004','2004-05-15 00:00:00','    Episode 06 : \"A sky like only a summer sky can be\"'),(25376,3152463,'13,May,2004','2004-05-13 00:00:00','    Episode 05 : \"You lot, you\'re all bloody weird.\"'),(25377,3152463,'13,May,2004','2004-05-13 00:00:00','    Episode 04a : Starry, starry night ... and an egg'),(25378,3152463,'12,May,2004','2004-05-12 00:00:00','    Episode 04 : St. Thomas Aquinas meets Mr. Muddle'),(25379,3152463,'12,May,2004','2004-05-12 00:00:00','    Episode 03 : Agatha Christie ... where are you ?!'),(25380,3152463,'11,May,2004','2004-05-11 00:00:00','    Episode 02a : The chicken, or the Art Deco egg ?'),(25381,3152463,'10,May,2004','2004-05-10 00:00:00','    Episode 02 : The shelf life of Art Deco, and eggs'),(25382,3152463,'10,May,2004','2004-05-10 00:00:00','    Episode 01 : Fun Run, sports cars and striplights'),(25383,3152463,'26,June,2004','2004-06-26 00:00:00','    Episode 29 : That\'s it - the end of the dream ...'),(25384,3152463,'21,June,2004','2004-06-21 00:00:00','    Episode 28 : Rule Britannia - a bumper blog entry'),(25385,3152463,'18,June,2004','2004-06-18 00:00:00','    Episode 27 : Moody mares abound - and it rained.'),(25386,3152463,'17,June,2004','2004-06-17 00:00:00','    Episode 26 : Sketchbooks, sneezing and the Swiss'),(25387,3152463,'15,June,2004','2004-06-15 00:00:00','    Episode 25 : The Return Of ... He Who Blogs Here'),(25388,3152463,'05,June,2004','2004-06-05 00:00:00','    Episode 24 : The Last Post ... well ... in a way'),(25389,3152463,'01,June,2004','2004-06-01 00:00:00','    Episode 23 : Feeling absolutely awful - oh well.'),(25390,3152463,'12,July,2004','2004-07-12 00:00:00','    Episode 31 : I say, Jeeves - Thunderbirds are go.'),(25391,3152463,'03,July,2004','2004-07-03 00:00:00','    Episode 30 : Random wanderings at 3am not advised'),(25392,3152463,'25,April,2004','2004-04-25 00:00:00','    Well, hello everybody ! My first post here, so let\'s see how it goes. Please bear with me - and if you don\'t want to bear with me, then go and bear somewhere else; although please remember there are laws against such actions.  So, an entry to post here, eh ? I used to have a blog, you know. It was nice. Sort of like this, only plainer. And it didn\'t have my picture at the top of the page, which Mr Battrick, the creator of this small piece of the web, has kindly done. This is sort of a trial entry for me, so we\'ll see how it goes.  I\'ll tell you a little about my day, that shouldn\'t be too bad. I\'ll liven it up a little, hopefully you might find it vaguely interesting. I awoke to the sound of birdsong, which was nice, and the sun streaming through my curtains ... that\'s the poetic bit, you see, so this becomes almost like a novel. I\'ve spent most of the day writing some French coursework, which is quite time consuming. I managed to spend half an hour looking for a quote, and failing to find it ultimately. Ah well, I could have done far less productive things, such as ... dunno, daydreaming ? Although that is undeniably fun.   And now I\'m upto the stage in the workload where I\'m writing about \"Art Deco â€“ itâ€™s influence on transport â€“ 1925-1939\". Can you think of a topic that could be more exciting than that ?! Most likely. But I shall carry on, regardless of boredom level, until it\'s concluded in a manner that would make a soap-opera writer quake in their boots (at the thought of a conclusion ... as they don\'t have them ... soaps, that is, not the writers ... it just wouldn\'t do to have immortal soap writers, would it ?).   Right, I think I had better go back to writing about the Normandie, the Bremen, the Europe, and whatever else was floating in the 1930s.   Until our next meeting, au\'voir ! '),(25393,3152463,'22,April,2004','2004-04-22 00:00:00','    What have I done today ?  Well, considering the nice weather we\'ve been having, I thought I\'d go for a walk, which I did.  That was nice.  Birds singing and the like ... yes, it was good.  I didn\'t see any badgers unfortunately, but overall it was nice.  A longer, better post will be posted soon ! '),(25394,3152463,'22,April,2004','2004-04-22 00:00:00','    Mustn\'t forget the good people at Accurist ... They do a great job of the speaking clock, the *most* expensive way of finding the time !   Speaking Clock - The Facts :   The BT Speaking Clock has also been known as TIM (the three-letter code, short for TIMe, that users dialled in the days of alphabetical dials !).    The BT Speaking Clock has been ticking 24-hours a day, seven days a week since 24th July 1936 â€“ which is 67 years, more than 24,000 days, more than 586,000 hours or 35 million minutes !   Originally the accuracy of the BT Speaking Clock was one-tenth of a second, it is now correct to within five thousandths of a second !   There have only ever been three voices for the BT Speaking Clock, apart from a two week period in March 2003 when Lenny Henry did a special version of the Speaking Clock in aid of Comic Relief  :   -         A London telephonist, Jane Cain was the first voice in 1936 and lasted until 1963.   -         Miss Pat Simmons, a supervisor in a London telephone exchange became the second voice from 1963 until 1984.   -         The present voice, belongs to Brian Cobby (73) who was an assistant supervisor at Withdean exchange in Brighton. He became the first male voice at 11 am on 2nd April 1985. Brian Cobby, an actor by profession before he joined BT, was selected from 12 finalists in BTâ€™s Golden Voice competition, on 5th December 1984.   Before Brian Cobby worked for BT, he recorded the â€œ5-4-3-2-1â€¦ Thunderbirds are go!â€ for the theme tune to Gerry Andersonâ€™s TV series !     Brian Cobby has received fan mail from â€˜mature ladiesâ€™ who say they listen to him late at night when they canâ€™t sleep !     The longest distance call to the Speaking Clock is from a factory in Hong Kong, which makes handsets for the VideoPlus VCR programming system. The in-built clock is set to British time, courtesy of BT\'s Speaking Clock service.    The timing of most television programmes is synchronised to the BT Speaking Clock as well, so when your local station goes over to ITN for the News at Ten, this is done \"at the third stroke\".     The worldâ€™s most famous clock, Big Ben, checks its time with the BT Speaking Clock and many major organisations have permanent feeds of the clock from BT into their private internal phone systems so employees can check the time without making an outside call.   The 1936 Speaking Clock was an of array of motors, glass discs, photocells and valves and took up the floorspace of a small room! The current Speaking Clock, with a built-in crystal oscillator and microprocessor logic control, the complete apparatus is made of solid-state microchips and occupies no more shelf space than a small suitcase does â€“ and has no moving parts at all.     Describing the BT Speaking Clock as a National Institution is probably fair, since as a nation we seem to be remarkably keen on ringing up. Even in its first year, 1936, the service registered nearly 13 million calls and it was not a nationwide service then - that came six years later. Today we make around 100 million calls a year and that\'s a lot of watches being checked.   The original Speaking Clock message was recorded and replayed rather like the optical sound track of a film and the equipment represented the state of the art of current technology in those days. This lasted until 1963, when it was replaced by more modern recording technology, using a magnetic drum. It gave way to the present digital system in 1984 and this has no moving parts at all.  The unquestioned accuracy of the Speaking Clock has also led to the association of Accurist Watches, who since 1986 have sponsored the BT Speaking Clock announcements. With those 100 million calls a year, this may well be the most frequently heard advertising message of all time.    But what did people do before the speaking clock was invented if they wanted a time check? Simple: they rang the operator and asked her the time by the exchange clock on the wall, but this was not precise to the second, nor could the exchange always answer just when the customer wanted. The first genuine speaking clock machine was introduced in the USA in 1927, coming to Paris in 1933, The Hague in 1934 and Switzerland in 1935. But automatic time service (of a Heath-Robinson kind) had been available to telephone users in San Francisco since the late 19th century; by listening for to an observatory clock at least a minute and decoding clicks and single and double buzzes against some detailed instructions you could set a pocket watch-but it helped if you already knew more or less what the time was ! A proper speaking clock is far less trouble !   - from  urlLink  Telephones UK '),(25395,3152463,'17,April,2004','2004-04-17 00:00:00','    Hello ! ... I\'m back !  A couple of days ago, a continental market arrived in my area, full of wonderful European foodstuffs :) !  Mmmm ...  I especially liked the Dutch chocolate waffles and the French fries dipped in garlic and parsley !  Today, however, I went to IKEA, and was enjoying my meal, when a women asked me if I\'d mind taking part in a survey ... Of course, I said I wouldn\'t mind, as it wouldn\'t delay me for very long ... It was very interesting ... That\'s all that\'s really happened in the last few days, so I have nothing much else to write ! '),(25396,3152463,'14,April,2004','2004-04-14 00:00:00','    Ooh, I\'m back, and I\'m probably here to annoy you again ! (If I annoy you, I really don\'t understand why you\'re bothering to read this, but ...) Anyway, on with the show :D !  I\'m sitting here again (oh, how many hours is it possible for me to just sit ...) in boredom (as usual) and wondering what to write ...  Well, it\'s not so bad, is it ?  At least I\'m making the effort to write something (unlike some people I can think of [cough, cough, cough] ) ... I know what I\'m going to tell you !  I\'ll tell you about the evils of the popular high street trading post known as WH Smith.  Today, I walked casually in to my local (notice that word,  local ) branch, hoping to order a copy of a recently published book.  To my shock, horror and amazement, I was told that the book was unobtainable, due to the fact that  \"it\'s not on our system\"  ... I responded with a rant : \"terrible service for a national chain !  Amazon is much better !  I can order the same item for their efficient online site (click  urlLink here  to go there [please support those good people as they do a great job !] ) and have it posted directly to me within 3 days !\" ... As you can imagine, I left quietly ...  It makes me so angry ... There is no reason whatsoever for them to tell me that it\'s unobtainable !  If they didn\'t have monkeys working at the desk ... none of this would have happened !  The day has by no means been wasted, though !  I managed to find great value for money in the form of 30 CD-Rs for an amazingly well spent Â£10 ! (the package also included a free CD pen !)  So, I\'m happy ! '),(25397,3152463,'13,April,2004','2004-04-13 00:00:00','    In an effort to find something worthy of writing about, I have ended up amusing myself by finding definitions  (in the Oxford English Dictionary)  of words that seem to me to be undefinable ... Not much luck yet ... \"  art  (noun) - production of something beautiful; skill or ability; paintings or sculptures etc.; (pl.) subjects other than sciences, requiring sensitive understanding rather than use of measurement \" ... Seems reasonable to me. '),(25398,3152463,'12,April,2004','2004-04-12 00:00:00','    So, another day, and another way of saying hello to the world !  I\'m back home now (you didn\'t know that I\'d ever left, of course) , sitting at my desk, wondering what to tell you ... Well, as I write this, I can hear the faint sound of badgers singing as they roam in a nearby field (for more information,  urlLink click here  and  urlLink here  ... but that\'s for another time.  Anyway, where was I ?  Ah, yes, I was telling you what I\'m doing ...   I think this explains everything quite well :       Anyway, I\'m feeling quite happy, surfing the net and making a few adjustments to a couple of my websites, so overall, it\'s been quite a good day (nothing unusual happened [click  urlLink here  for details of what could have easily happened !]  Oooh, by the way, there is a really funny Flash animation  urlLink here , so have a look at that as well !  I hope tomorrow will be as happy as today has been ! '),(25399,3152463,'09,April,2004','2004-04-09 00:00:00','    Of course, \'blogging\' is very hard work, so I\'m going to tell you the reasons I have started this blog !  i) I\'m very, very bored  ii) I think it will be fun  iii) I\'m wondering if I can write a sort-of diary (I keep writing \'dairy\' !)  Also, I see that lots of interesting people write blogs ... I hope I\'m interesting ! '),(25400,3152463,'09,April,2004','2004-04-09 00:00:00','    Blog, weblog, online diary, journal; call it what you will ... but I\'m starting one !  I hope that it\'ll be as interesting as some others I\'ve found, although I doubt that I will be able to hold an audience in suspense for very long ... Anyway, on with the blog ! '),(25401,3152463,'23,May,2004','2004-05-23 00:00:00','    You may have been led to believe - due to the fact that I haven\'t posted for several days - that I was no longer posting.  Sadly for you, I have just been buying time for which to compose a newer, longer post for all 2 of you to read !  The site statistics are promising, so I\'m expecting quite a number of avid readers this time around.  Yesterday was interesting, mainly due to the fact that I purchased a number of interesting items during my shopping trip.  First of all, it was a Peter Kay Live DVD : \'Bolton Albert Halls\', which I would highly recommend to any fan of his excellent comedy series, \'Phoenix Nights\'.   Quotes : \"Don\'t pull that ... It\'s for warden !\" ...  followed by  \"Sorry, It\'s me grandson ... He thought it were t\'big light\".   The second purchase was a very good guide to Photoshop 7, \'Photoshop 7.0 A - Z\'.  It\'s a great book, with a really good cover price of Â£17.99, the average Photoshop book being Â£30 !  It explains how to do absolutely everything under the sun, and is illustrated with full-colour screenshots throughout.  The Appendices are also incredible, with a comparison chart, plugin descriptions and recommended websites.  Well worth buying !  The final purchase of the day was the stunningly funny French film \'Les Visiteurs\', which if you haven\'t seen it, is about a French knight and his servant who get transported to the 20th Century ... Monty Python meets Blackadder, definately.   Quote : \"Look Sire, a Moor !\"  They procede to smash up a car. '),(25402,3152463,'18,May,2004','2004-05-18 00:00:00','    Here we go ... I awoke this morning at 7 AM.  Yes, I know it\'s early, but that\'s the way it is.  I spent the day, as usual, boring myself half to death with inconsequential thoughts and ideas.  I had lunch at 1:30 PM (extremely late, but not my fault) - undoubtedly featuring sandwiches - which delighted my taste-buds no end ... Well, maybe  not , but  still  ...  The afternoon was spent puzzling over more inconsequential thoughts and ideas which would seem boring to an amoeba, with pen at hand.  It\'s amazing how uninteresting things can appear when you put your mind to it, you know.  I have discovered this through years of practice, preparation, and very long sessions of boredum, without any means of escape, or \'way out\'.  I then returned to my \'place-of-dwelling\' and sat down.  If any of the amphibiology above seemed somewhat thought-provoking, I advocate that you seek medical attention instantaneously, if not sooner. '),(25403,3152463,'16,May,2004','2004-05-16 00:00:00','    Yesterday was throughly enjoyable, what with  The Calculus Affair  being shown on  Five , and of course, Eurovision.  Surprisingly, we didn\'t win; as every country under the sun donated 12 whole points to their neighbour. Cyprus voted for Greece, for example.  Another example of this highly unfair policy of voting for friends (this time Belarus is at it) :   Belarus Announcer : And 1 point to the UK ! Terry Wogan : Finally ! ... Belarus (later) : Finally, 12 points go to ... Terry Wogan : It  can\'t  be Russia ? Belarus : RUSSIA ! Terry Wogan : I\'ve been doing this for too many years ...   France\'s entry was of note; featuring a bald woman on stilts and a smaller man singing *at* her ... Terry Wogan compared it to  The Little Prince .  The Netherlands entered  Reunion , consisting of two singers, one *playing* (I doubt) the guitar, the other singing - in a very awkward high pitched voice - about something which I have now luckily forgotten.  By the way, the Ukraine won, with another sloppy act worthy of absolutely nothing.  But,  The Calculus Affair  was very enjoyable ... '),(25404,3152463,'13,May,2004','2004-05-13 00:00:00','    My parcel arrived today; very promptly I thought !  The package was *very* well packed; two layers of bubble wrap, two plastic bags, two Jiffyâ„¢ bags, and yards of duck tape featured heavily ... Plus a \'Buddist-clipart\' sodden \'thank you\' note ... with a PS - \'Please leave positive feedback\' (Hmmm ... [But only if I leave them feedback, will they provide me with positive words of praise, they say] ) '),(25405,3152463,'11,May,2004','2004-05-11 00:00:00','    Hey nonny no !  I felt like writing an entry tonight ... you know, you get inspired, or something to that effect. Well, I was inspired a couple of Saturdays ago, when I went to an exhibition of Quentin Blake\'s work at the Harris Museum, in Preston.        Yes, that\'s the guy ! The work was really incredible - including some full-page ink and watercolour illustrations for \'The Witches\', and cover art to \'The BFG\'. There was also huge pieces for picture books he had done, which were stunning to look at. A fantastic trip, I\'d say, and highly recommend going, if you can. It\'s on at the Harris Museum and Art Gallery, Market Square, Preston PR1 2PP (I have the exhibition poster on the wall in front of me). It\'s free admission, and you can purchase a Magic Pencil from the gift shop, which draws in four different colours ! And yes, I have one. It\'s so much fun, I\'ve spent countless lessons doodling with it.  I\'m not sure what else I can write about now ... Oh ! Congratulations to Oliver for getting \'Tintin and Alph-Art\' from eBay - that edition is incredibly rare now, enchanced by the fact that the new edition was released yesterday.   And I believe I shall leave it there. Stay safe, everyone, and remember : feeding Stilton to rottweilers by hand may seem like a good idea, but it\'s not as easy as it looks, if you are fond of your limbs. Au\'voir ! '),(25406,3152463,'01,May,2004','2004-05-01 00:00:00','    eBay ...  what  a topic to talk about !  There\'s nothing I like better than bidding on eBay, apart from any other alternative ... Yes, eBay is usually a  last  resort.  I did however, manage to pick up a nice copy of  Tintin and Alph-Art  yesterday, which seems to have influenced my decision to write a post about this incredibly interesting subject.  Highly unlikely, I know, but I hope you enjoy this short delve into the world of eBay.  I\'m going to explore the fascinating field of eBay IDs.  Before any individual can join the eBay community, he / she has to choose a nickname to be identified by within eBay.  This is referred to as an eBay ID.  Many IDs are unpronounceable, but many display humourous or entertaining qualities, such as mine, which I think sums up my \'end of auction\' feelings; \'franticdilemma\'.  Call it what you will ;) .  Richard suggested \'Outbid_me_and_i\'ll_kill_you_in_your_bed_as_you_sleep\', which I also think is very appropriate.  At least now I can relax, safe in the knowledge that the good book is finally on its way ... '),(25407,3152463,'01,May,2004','2004-05-01 00:00:00','    To start with, I\'d like to congratulate Mr. Wainman on his first post !  Then, I\'ll tell you a bit about my day ...  I started this morning by watching the Tintin movie;  Tintin and the Lake of Sharks , which, although nothing like HergÃ©\'s books, was quite enjoyable.  What interested me about the television version is that the title sequence differs from the video version ... yes, I know I live a sad life :( .  You can\'t blame me though, as I have been under the weather for several days.  If you find that reasonably interesting, I suggest you wait for another post ... '),(25408,3152463,'13,June,2004','2004-06-13 00:00:00','    Hello again !  I am sorry to have been away for so long.  Thank you to Richard for holding the fort, so to speak, while I have been away.  I have, in fact, been to Walt Disney World in Florida, which was very enjoyable.  My favourite of the four Disney parks is EPCOT, the main feature of which is pictured below.       A very impressive site at the front of the EPCOT Center, indeed.  There is actually a ride inside it -  Spaceship Earth  - which is a history of mankind, featuring animatronics and \'time-travelling\' vechicles ... Don\'t ask.  Another great part of EPCOT is  Test Track  : a fast-paced ride in which the riders are supposed to be crash dummies for car testing.  The tests are performed indoors, but the highlight of the trip is a \'practice run\' at 80 mph around the outside of the building !  That is a truly incredible experience !  Highly recommended.         Also in EPCOT\'s Future World is  The Land  - an example of how to grow crops without soil etc,  Mission : SPACE  - a disappointing simulation ride of space flight,  Universe of Energy  - an animatronic show about the creation of energy,  Imagination !  - an area about \'your imagination\', featuring Dr. Nigel Channing  (Eric Idle) , a strange character called \'Figment\', and the cast of \'Honey I Shrunk The Kids\',  Wonders of Life  - about the body,  Innoventions (East & West)  - a world of exciting interactive experiences - hosted by \'Tom Morrow 2.0\' and  The Living Seas  - seemed a bit of a disappointment ... About fish.  The other half of EPCOT is called  World Showcase  - a journey around the world, featuring pavilions of Mexico, Norway, China, Germany, Italy, The USA, Japan, Morocco, France, The United Kingdom and Canada (all include people of these nations, and local food etc.)  There are of course three other parks in Walt Disney World, but EPCOT is my favourite, so I\'ll leave it at that.  I will however mention a extremely enjoyable ride at the Animal Kingdom, called  Primeval Whirl .  It\'s basically a roller coaster, but with not as many turns as the usual kind.  The difference is that the carriages spin around as they go down and over bumps.  It is a great ride, and I would recommend trying to get on it twice in one day, as it\'s over quite quickly.  The queues don\'t tend to be that long though, so it\'s not too long a wait.  If you\'re in need of music, DVDs, videos, games, books, a cinema etc. while you\'re staying in Disney, there\'s a  Virgin Megastore  at Downtown Disney, along with a cinema, a guitar shop, night clubs, restaurants etc.  Sadly, I found no trace of Tintin in Florida. '),(25409,3152463,'01,June,2004','2004-06-01 00:00:00','    Ladies and gentlemen, we have a problem.  We have lost our Oliver.  No sign has been seen of said Oliver for a number of days now, and quite frankly, we are beginning to worry. We wish Oliver a speedy return from wherever he is, and hope that he gets back to blogging, MSNing, drawing, and all the Olivery things we know and love.   Please come back, we need you ! '),(25410,3152463,'18,July,2004','2004-07-18 00:00:00','    Last weekend (a week ago today), I endured a rather long train journey to visit London; the reason being that I was going to see the new Tintin exhibition,  The Adventures of Tintin At Sea.&nbsp;  I managed to acquire all the leaflets, guides, and booklets regarding the exhibition that you can possibly imagine.&nbsp; On the way to London Paddington, I read  The Pocket Essential Hitchhiker\'s Guide , which is the best source of information about the production of The Hitchhiker\'s Guide To The Galaxy that I found at the time of writing.  &nbsp;  Anyway, upon arrival in London (or Lon Don as it is referred to by Tubbs in that rather obscure television and radio series  The League of Gentlemen ), I made for Covent Garden, and the Tintin Shop which is located at 34 , Floral Street.&nbsp;&nbsp;Below is photographic&nbsp;evidence&nbsp;of how that marvellous establishment looked on Saturday, 10th July.  &nbsp;         I was rather pleased with my purchases of that trip, which included two T shirts, several books and a keyring featuring the logo from the Tintin musical,  Tintin Et Le Temple Du Soleil .&nbsp; On my way back to my hotel, I stumbled upon a Swatch shop in Covent Garden, and promptly purchased a Tintin watch.&nbsp; Where I\'m from, that\'s what we call a very eventful day.&nbsp; The following morning - on my way to the exhibition&nbsp;- I discovered a wonder of London Public Transport :  The Docklands Light Railway .&nbsp; For the unfortunate souls amongst you who have not had a ride on this \'highway in the sky\' - to coin a phrase, I will try to explain this jewel in London\'s crown.&nbsp; Basically, The Docklands Light Railway is&nbsp;a monorail system, very much in the style of the Underground, except of course that it\'s two storeys above the ground, and goodness knows how many above the \'underground\'.&nbsp; I know - that sentence explains absolutely nothing, and you\'re probably thinking that its sole purpose is to pad out this otherwise dry, dull text.&nbsp; If you think that, then you\'re reminding me of Marvin (the paranoid android of \'the guide\'), and I suggest you stop reading as soon as possible.&nbsp; Anyway, back to the Docklands Light Railway.&nbsp; This incredible ride \'zooms\' past tower blocks, until it finally comes to ground level to cross the river like an underground train.&nbsp; The \'Cutty Sark (for Maritime Greenwich) is the station needed to access the National Maritime Museum , where the exhibition is located.&nbsp; I found an Ottakar\'s Bookshop right outside the station, and -&nbsp;you should have&nbsp;guessed&nbsp;- I bought three Tintin postcards.   &nbsp;'),(25411,3152463,'04,July,2004','2004-07-04 00:00:00','    I\'ve been away for another four weeks ?!  So, what\'s new ?  Well ... Yesterday, I decided that I would explore the shopping scene and I ended up buying a rather fine dip pen from an art shop :       It works incredibly well, uses an \'italic 3\' nib and I\'m quite proud of my first drawing with it (on tracing paper and coloured using Photoshop) :       More purchases were purchased during yesterday\'s travels : several books (\'The Encyclopaedia of Cult Children\'s TV\', \'The Pocket Essential Laurel & Hardy\', a previously \'unknown-to-me\' translation of \'The Little Prince\', with colour illustrations and \'It Must Be True - A Collection of newspaper bloopers, misprints etc.\') and also a bottle of black Quink ink for my dip pen !  Today, I\'ve been sitting here listening to Green Day\'s version of \'I Fought The Law\' and \'Camp Grenada\' ... '),(25412,3471937,'30,May,2004','2004-05-30 00:00:00','    So, it is Sunday night about 1 in the morning here in Indy.  It is Memorial Day weekend, so no work until Tuesday, which is fabulous! YAY!  I\'ve really accomplished my goal this weekend of having a positive attitude and not being so sensitive.  Today, especially, was really nice.  I actually spent the night at my apartment, which may have been the first time all week.  And I spent the day cleaning and organizing.  It is not completely finished, but I can see light at the end of the tunnel for sure, and it is a very nice feeling.  We also bought some paint for the walls in the dining area, which will be a lot of fun.  I\'m really happy for Dana.  She has met a new guy, Dan (go figure that would be his name, right?) But anyways...he is really great.  I like him a lot.  He is her age, 19, which is easy to forget sometimes because he seems so mature.  If something more comes of it, I think he will be really good for Dana.  He is really responsible.  He has a great job working as a mechanic for BMW and he owns his own house and at least 2 vehicles, completely financially stable at 19 which is beyond amazing.  But the more important thing than all of that is that he seems to be a real sweetheart.  He is treating Dana really well so far, which she totally deserves and needs.  She really seems to like him so much too, which makes me so happy.  I hope that she can see that this is the type of guy she can be with.  Somebody whose life is actually going somewhere and who treats her well.  I know that it will still be difficult for her to move on with her life and completely cut off things with Daniel (the ex, not Dan the new guy), but hopefully she will be able to see how much better she can do.  I foresee good things (or I really hope anyway).  So that\'s that.  And...well, I got an email from Uli.  Crazy.  I just had to laugh right out loud.  So, she and Ovi have been spending a lot of time together.  I had a feeling that he may be interested in her, but lo and behold, she seems to be interested too.  She said that he gave her a big hug and was holding her hand this week and her heart was racing.  Wow!  I mean what can I say.  I\'ve always thought that Ovi was a great guy, even if we didn\'t completely hit it off due to the situation with Paul.  So I think it is really fantastic for Uli.  But I have to laugh I really do.  I mean Ovi is Paul\'s best friend.  So...Paul cut off things with Uli because he wanted things to beyond friendship without an actual commitment.  And she wasn\'t willing to have a relationship with him beyond friendship because of me and my history with Paul, which she told Paul, and knowing him as well as I do, I know he wasn\'t happy about that.  So, now, Paul\'s best friend may be hooking up with Uli.  And she and I, despite the distance, are just about best friends.  Poor Paul...haha.  Karma, my friend, karma.  Paul and Ovi have a special relationship so I don\'t quite know how this turn of events will affect things, but Paul could not possibly be happy about it.  I really shouldn\'t be happy that things are turning out so badly for him, but I don\'t know, I guess there is an evil side in all of us.  I just can\'t help but think it is a bit funny.  But more than anything I wish that he would think about everything and realize that it is his actions that put him in such a situation.  He pushes away all of the people who care about him.  Okay enough about that...Oh, but jeez it is funny...  Now, it\'s just about time for me to snuggle up to my sweet Joey, I do love that man...Life, again, is good.   '),(25413,3471937,'28,May,2004','2004-05-28 00:00:00','    I\'m sad today.  And last night.  I don\'t know what is wrong with me.  I think I may be having anxiety attacks.  I don\'t even like myself lately.  Last week was very very bad, but I thought that I had gotten over that, calmed down a bit.  But last night was the same as last week.  It is like I cannot take an ounce of criticism.  If anybody says anything negative, I take off on this downward spiral.  I don\'t even recognize myself anymore.  I guess there are still traces of me, but where is that happy girl.  Duey Lightfoot once told me that he loved being around me because he had never seen me without a smile on my face.  Granted, back when he knew me, my life was a lot less complicated.  I make my own complications.  I always said that because I know that I do that, and I make my choices anyway, I would deal with the consequences. But sometimes it is really hard.  I really miss my friends.  But would that even really make things better? Or is it just something inside of me that I need to conquer?  It is the stress of trying to take care of Dana?  I do feel that is a huge part of things, but I do not want to blame others.  Not Dana, not Joe, nobody, this is my problem and I know I have to find a way out myself.  And I will.  I\'m still that eternal optimist.  I will find my positive outlook.  I will not let these feelings take control.  I have to find a focus.  I have to be less sensitive.  I\'m so sensitive and it is getting ridiculous.  I don\'t know how other people stand to be around me.  Today, in my car on my way to work, I screamed as loudly as I can.  I can\'t stand being around myself even, so I cannot imagine why anyone else wants to be around me.  I wish I could go to the ocean.  Lie on the beach, feel nature all around me. Decompress.  I\'m driving Dana up the wall.  I\'m so sorry about that.  I have to try harder.  I will be happy this weekend.  I will get things accomplished.  I will be me.  The fun Kari.  The one who smiles and has fun no matter what.  That is my goal.  I will find my sunshine.   '),(25414,3471937,'27,May,2004','2004-05-27 00:00:00','    I\'m such a slacker. When I know that I\'m way ahead of my deadline, I cannot work. Oh well, gives me more time to write, right? Hehe. So, we were supposed to go play disc golf today but the weather.com shows lots of rain.  bummer!  We didn\'t go on Tuesday when we were supposed to either.  I bought new inline skates and haven\'t used them yet either. The weather needs to get nice soon.  I\'m getting stir crazy.  Nobody really seems to want to work today. I can hear people hanging around each other\'s desks chatting.  Well, it IS the week of the race, and we DO live in Indianapolis.  How are we supposed to focus?    Changing the subject, I miss Uli.  I wish sometimes I was still in Regensburg hanging out with her everyday. I think that I\'ve given up on the idea of studying in Germany for a while. I don\'t know.  I\'m just afraid to leave this job because it is truly an amazing opportunity, and I really love it.  Despite the fact that I\'m terrible about procrastinating, I do a good job.  Plus, I\'m getting licensed next month.  Last night, Joe talked about me supporting him while he goes back to school.  I actually would have no problem with that, so long as I could afford it.  I\'ll be his Suga\' Mamma ;) But giving up on my dream to study in Germany...Hell, this may be my last chance.  I promised myself I would never live my life regreting.  I always want to take the chances.  That\'s why I left Houston and came here to Indy in the first place.  It was still really early in Joe and my relationship, but I had fallen in love with him and though it was a huge step...moving 1,000 miles with a guy I\'d only been with like 6 months at that point...I thought I had to take the chance, or I might have regretted it later.  Wondering what might have been, I don\'t want that. But if I did go to Germany to study, would that screw up things with Joe and I?  And would I regret leaving this job?  I can only imagine how amazing it would be to stay there long enough to completely be fluent in German.  And I know that Uli and I would really become the best of friends.  We\'re already great friends despite the ocean between us.  I guess I will sit back, try to save money, and see what life has in store for me. Sound like a good plan?   '),(25415,3471937,'27,May,2004','2004-05-27 00:00:00','    I went to Circle Bar Saloon last night for the art show \"8\" (which also happens to be my favorite number).  It was really fabulous.  Everything from culinary arts to belly dancers to amazing paintings. It has inspired me to start painting again.  Joe went with me, and I think he actually enjoyed it. He was really making an effort last night, and I love him for that.  We\'re going to be okay. I know that now. We\'re going to stand by each other and work through whatever problems come up. I had been overwhelmed before.  That\'s what happens when I try to fix everything myself. I get overwhelmed and I start thinking there is no way to make things better except to run away. I don\'t regret the past. I did what I had to do. I had to have closure before I moved on with my life with Joe.  Be careful what you ask for, I guess.  I most certainly got closure.  Closure and a kick in the ass.  It\'s still hard to believe that that was the end.  After all these years. After everything that we went through, all that we shared, and that was how everything ended.  But it by no means was my fault. I gave him every opportunity to stay friends with me.  I poured my heart out to him the way I always did. But I guess things change. I always loved that we could speak our minds to each other.  Tell each other everything, exactly what we were feeling, without holding back. But that didn\'t work this time.  He wasn\'t willing to be my friend. Not a friend in the true sense of the word.  He would be my friend if that were an easy thing. If it took no effort on his part.  But with our past, some effort was required. So many empty promises. I wish I never would have believed a word he said. Did he really change so much?  Or did I just believe in an illusion?  Was he never really the person that I thought he was?  I keep being reminded of Strange Attraction by the cure  it started with a dedication \"lost in admiration - happy birthday - i\'m forever yours\" blossom faded red inside a tiny book of butterflies  i smiled surprised at how when flickered through  the wings flew by spelled out my name...  six months went by the summer lost obsessively the letters dropped into my life  the same soft blood smooth flowing hand  \"please try to understand - i have to see you - have to feel you -  tell you all the ways i need you - yours forever in love... \"  strange attraction spreads it\'s wings it varies but the smallest things you never know how anything will change strange attraction spreads it\'s wings and alters but the smallest things you never know how anything will fade  the year grew old incessantly she wrote to me  she\'d started smoking poetry!  i laughed in recognition of a favourite phrase  she pulled me in...  i answered her  a christmas card in sepia arranging when and where  and how the two of us should meet...  her opening so well prepared  a nervous smile  i couldn\'t take my eyes from her  she whispered  \"can i use some of your lipstick?\"  it was perfect so believable  i couldn\'t help but feel that it was real  and kissing crimson fell into her waiting arms...  strange attraction spreads it\'s wings it varies but the smallest things you never know how anything will change strange attraction spreads it\'s wings and alters but the smallest things you never know how anything will fade  so alone into the cold new year without another word from her  i wrote to ask if we could maybe meet again before the spring  but weeks went by with no reply until once more my birthday came  and with it my surprise but this time nothing was the same...   \"i\'m sorry - blame infatuation - blame imagination -  i was sure you\'d be the one but i was wrong -  it seems reality destroys our dreams - i won\'t forget you\" blossom faded red inside a tiny book of old goodbyes...    strange attraction spreads it\'s wings it varies but the smallest things you never know how anything will change strange attraction spreads it\'s wings and alters but the smallest things and you never know...   But you know what? Forget all that, it\'s over and done with...finally. I have a new TRUE love in my life now. I just got an email from him. I love the way that he loves me. The way that he makes my heart soar. Just looking into those gorgeous green eyes.  I\'m going to be with him for the rest of my life and I love him more than anything. I see that clearly now.   '),(25416,3471937,'29,June,2004','2004-06-29 00:00:00','    Things were really bad this weekend.  Dana\'s temper went completely out of control.  I haven\'t seen her like that in years.  It was truly scary.  I told her that I didn\'t want Daniel staying even one night at my apartment...that didn\'t go over very well.  So I called Dad thinking that he could talk some sense into her and get her to respect my wishes since the apartment is mine and in my name and all of that.  Well, instead of trying to reason with her, Dad basically told Dana that if she went back to that abusive relationship that he and Mom have tried so many times to do what all they can to help her, but they can\'t keep on doing it over and over again if she keeps going back to him.  He said that the next time she gets into a situation with Daniel and she wants to get away, she should call the police rather than call Mom and Dad.  He basically told her that if she makes this decision that they aren\'t going to have much to do with her anymore.  Okay, so I can understand why she flipped out.  I totally didn\'t expect Dad to be so hard on her.  But maybe I\'m too easy on her.  She\'s going back to an abusive relationship, but it seems that no matter what I say it is not going to change her mind.  I suppose I\'ve somewhat given up.  I don\'t want her to get hurt, but it seems inevitable.  But the thing is she flipped out on me rather than on Dad.  She told me that I have betrayed her more than any other person in her life.  And many times over she told me that she hated me.  Her words didn\'t affect me too much because I know she was saying it out of anger.  And in truth, I\'ve never really betrayed her.  In her mind, my calling Dad and telling them about her decision to go back to Daniel in Houston was a betrayal.  I think that she didn\'t want to tell Mom and Dad because she actually knows what a bad choice she is making.  Maybe I\'m being selfish now, but I\'m not fighting it anymore.  I\'m tired of it.  Even if I found a way to change her mind this time, in a matter of time, she would want to go back again and I can\'t keep dealing with it over and over again.  It is making me crazy.  And the condition of the apartment.  I spent hours upon hours cleaning it this weekend.  It looks really good now.  But she has already basically ruined the carpet.  If she stays here longer, who knows what else I\'ll have to pay for.  I know that is a selfish way to look at things, and if I thought that I could actually change her mind I wouldn\'t care about all of those things but I\'ve tried and tried, and I don\'t think there is anymore I can do.  Before I changed her mind the last time, I became thoroughly depressed dealing with the whole situation.  I mean, I haven\'t felt that depressed in a very long time, and now that she is right back to where she started wanting to give him another chance when he hasn\'t at all proved he is worthy of it...I can\'t fight it anymore.  I\'ve given up.  I hope with all of my heart that she\'ll be okay, but I give up.   '),(25417,3471937,'24,June,2004','2004-06-24 00:00:00','    So, life is crazy. (What else is new?) If my life weren\'t crazy, I don\'t think I would know what to do. So...Dana is moving back to Houston.  What can I say?  I think it\'s a huge mistake.  I don\'t think she will be happy.  Maybe at first things will be okay with her and her idiot ex (or I guess current) boyfriend.  Since they haven\'t seen each other for a while, I\'m sure he\'ll be on his best behavior for a while.  I really think that after a little while things will go back to the way they were.  I know that she loves him.  And I know how hard it is to let go of someone that you love, even if you know that is what you should do. Believe me, I know.  I had to learn my lesson the hard way, and I suppose she will to.  I just hope he doesn\'t hurt her too badly, emotionally and physically.  The fact that I\'m no longer able to take care of my baby sister breaks my heart.  I wish that I could just wrap my arms around her and keep her safe from any future pain.  Unfortunately, I can\'t.  She is an adult now, and she has to make her own mistakes, I suppose.  I wish her the best.  I truly do.  I don\'t support her choices, but I hope that somehow she finds a way to be happy.  I will always love you my Dana-bee.  I know your lovely soft caring heart, and I hope that it doesn\'t lead you to any more pain.  I don\'t want her to leave me.  There is this ache in my throat and in my heart.  But I guess I have to let go. The only positive thing about the whole situation is that Joe and I will live together again.  We basically do now, but there will be no need to transport clothes back and forth and so on... And a bit of privacy will be nice too.  Don\'t get me wrong, I love all the guys that we hang out with. But they are there practically all of the time.  They even invite each other there without asking one of the actual people who live at the condo if it is okay.  It is a pretty funny situation.  I\'m glad that Joe has found so many good friends.  And I almost feel guilty that I want him to come live with me at my apartment because I feel like I will somewhat break the cycle that we are in now.  Almost every evening all of us tend to migrate to Joe and Andy\'s condo.  I imagine one of 2 things will happen.  Either, Joe and I will still hang out frequently at the condo, despite the fact that he won\'t live there anymore, or people will migrate to our apartment to hang out.  I kind of think it will be the former, especially because Jay will be moving into the condo.  If it would just be Andy there, then I would have my doubts that it would remain the hangout place, but since Andy and Jay will be living there, I can see everything relatively staying the same.   I haven\'t written here for a while.  I finished taking my licensing classes, and I passed the test. Yay!  So now I just have to send in the app, fee, and paperwork, and I am a licensed producer.  I also met this really cool girl who was in the class with me.  In fact, I really need to call her soon.  She seemed like somebody I could get to be good friends with.  Well, there\'s more to say, but I\'ll leave that for another day.   '),(25418,3471937,'11,June,2004','2004-06-11 00:00:00','    So I\'ve been spending the morning reading my textbook for my Life & Health Licensing class I take next week.  Pretty interesting stuff...let me tell ya.  Scott didn\'t come into work today because they are having people remodel their house.  So it\'s just me here trying to stay focused until it is time to go home.  I\'m a little bit nervous about my class next week considering it has been nearly 2 years since I\'ve taken any kind of class.  I think I\'m getting old.  I want to go back to school, mostly just because I feel like I\'m growing up too fast.  Life is good, though.  I\'m proud of where I am now.  The struggles that I have made it past, defeated, conquered.  Yeah, I did good. Joe and I had a fabulous night together last night.  I went over to his place just to pick up my phone that I had accidentally left there, and I ended up not going back to my place until this morning to change clothes and head straight to work.  Considering that I can\'t remember the last time I actually slept at my apartment, it is basically a glorified closet.  Joe really is the sweetest.  I love him more everyday.  The way he holds me and looks at me, it just doesn\'t get old; it still makes my heart race even though we\'ve been together for a while now. So, I absolutely have to go to FNM tonight.  I was supposed to go last week, and didn\'t quite make it in time.  I think I was there for the last round, maybe.  Joe and I have plans to go out for sushi afterwards. Oddly enough, I don\'t know if I\'m in the mood for sushi, so maybe I\'ll try something else today.  Get some Lo Mein or something; that sounds yummy to my tummy.  I really need to spend time studying this weekend, so I hope the guys don\'t have too many fun plans because it will be hard for me to not want to go even though I know what I need to get done.  Surely, I have enough self-discipline for one week though.  I\'m not really that worried.  I\'m a procastinator without a doubt, but in the end I always pull through.  I had a great conversation with Uli today.  She is the absolute sweetest, ever.  I wish so much that she didn\'t live thousands upon thousands of miles away.  It would be so unbelievably awesome if we could live in the same place.  It has been a long time since I have had a female best friend.  She and I are so much alike too.  Maybe someday.  Crazier things have happened.  Uli told me that she talks about me to Ovi all of the time, and that he thinks I\'m wonderful.  If you would listen to her, you would think that I have no faults, when in actuality I have an abundance of them, but I\'m happy that she sees the good side of me.  And I\'m really happy for Ovi to have a good opinion of me.  You know this part of me hopes that Ovi gets to know me really well and we become good friends, and then he tells Paul that he thinks that he is crazy for ending things with me.  Not that that would get me anywhere.  I\'m totally happy and in love with Joe.  And it is not like I will ever make that mistake (Paul) again.  I just want him to realize how much he hurt me and how it could have been.  And I\'m honestly talking about as friends because looking back, I think wanting more than that may have always been a mistake.  But at one time we were really special friends, and it would have been nice to have that always.  But as I told Uli today, everything that happened between Paul and I is worth it because I got to meet her and we became such good friends. Well, I suppose maybe I should get back to reading this fantastically interesting textbook! :D   '),(25419,3471937,'10,June,2004','2004-06-10 00:00:00','    I had my lunch at the park. It was especially nice today.  I focused on the sounds of nature. The insects, birds, thunder, the light rain. So gloriously beautiful.  I could have just sat there all day long listening.  Reading and listening.  It was all so relaxing. I could forget anything that troubles me.  My feelings of solitude don\'t matter when I\'m there because in that place solitude suits me. It is what I want.  I could forget it all.  I even feel like I don\'t care anymore.  I don\'t care about all of the pain I have been feeling lately because my feelings are not important.  Instead of that thought making me more depressed it makes me feel free.  To be outside of myself and my thoughts and my feelings and everything that has brought me down lately, to immerse myself in nature, in peace, to find a happiness within myself and no longer feel the need to look for that happiness elsewhere.   '),(25420,3471937,'07,June,2004','2004-06-07 00:00:00','    I got a call from Ludwig yesterday. I miss him. I miss my apartment in Houston where he would come over and we would drink together and watch my crappy non-cable tv and he would laugh over my frozen banana in the freezer. I think back to that semester I lived at Cullen Oaks, which was arguably the worst time in my life. But that was the time that Ludwig and I became the friends that we are today. He was there for me when I had no one else.  He tried to take care of me the best he could, which I know was extremely difficult because I can be stubborn and independent.  He never gave up on me, even when I felt hopeless and practically gave up on myself.  And I know that even now, if I called him and told him that I needed him right this minute, he would drop everything and find a way to get here and help me with whatever it may be, big or small.  He is so much like me too.  It still amazes me.  I was reading his blog that he wrote while he was studying in France.  Some of the things he wrote could have come straight from my mouth.  I wish that he still wrote in his blog because reading it was one of my favorite things to do.  His words almost always put a smile on my face.  Sometimes I laughed so hard over what he wrote that my sides hurt.  I hope that someday we live close together again.  At least, in July, he will be visiting me here in Indy.  That will be so awesome.  I\'m thoroughly excited.  He\'s one of the best friends I have ever had, maybe the best ever.  I sit here sometimes and feel sorry for myself, but look at all the wonderful people I have in my life.  I\'m so lucky. I need to remember that.   '),(25421,3471937,'07,June,2004','2004-06-07 00:00:00','    This past weekend was really nice.  I cooked gumbo for everybody on Saturday, and I believe it was quite a hit because they all came back Sunday evening to eat the rest. Saturday night after we ate, Joe, Evan, Dana and I went to Broad Ripple Park for a walk.  We walked a bit and then we all swang on the swings, all of us, even the guys!  And then we went to one of the play grounds and played on the gymnastic type equipment.  Then the guys sat on a bench and talked for a while, and Dana and I sat up near one of the slides and talked and sang songs that we used to sing all the time when we were younger.  At one point, I went down one of the slides backwards and Dana came down it right behind me, too quickly for me to get out of the way.  I haven\'t laughed that much in a really long time.   '),(25422,3471937,'03,June,2004','2004-06-03 00:00:00','    I just came back from my lunch break.  I\'m reading House of Leaves.  What a trip.  I think that I spend half of the time re-reading everything, but I think that is the point.  I think he\'s trying to make you think that you\'ve missed something.  Each time I put the book down, when it\'s time to go back to work or what not, I have the same reaction.  I softly mutter the word crazy.  I\'m anxious to read more.  Maybe tonight...  So, I had a very long talk with Evan last night.  We both told each other things that others don\'t know.  I told him things that even Joe doesn\'t know, but I know that I can trust him.  He is going through hell right now.  I don\'t feel like there is much that I can say to him because he is so self-aware and already such a positive person that he just knows everything I could possibly offer.  I guess the thing is right now to just listen.  We can\'t change the past.  I say it a million times over, but it\'s true.  We have to mourn over what we may have done or not done to cause a situation because we have to find a way to heal.  But then we just have to move on with our lives.  We can ask ourselves if we have done the best that we could given our circumstances.  Maybe we did, and maybe not.  But take that answer one way or another and learn from it.  Try your best not to make the same mistakes again.  He will be happy again.  He will find a way out of all of this.  I know it.  I already believe that he is an amazing person, and that this will only make that even more true.  I told him some things about Joe, about my feelings, nothing too personal of course.  Before I even really started talking about any doubts I may have had and what not, he asked me where I saw things going with Joe and I.  I said that I think we will probably get married someday.  But the way he asked it... It keeps running through my mind.  It is almost as if he has doubts that that is the way we are headed.  It makes me think that he is unbelievably perceptive.  I don\'t even know for sure if that is what he was thinking, but there was something in his eyes, when he asked me.  I\'ve always thought from the outside there appears to be absolutely nothing wrong with Joe\'s and my relationship and that just about everyone would think without a doubt we are headed toward marriage.  Sometimes I feel like my mind goes back and forth, and I\'m not completely sure about what I should do.  But I told this to Evan, and it is totally true:  I fully believe that Joe would fight for me and for us and our relationship.  He may not be able to change everything, but I can see that he is totally willing to put forth effort.  You know, I say these things as if I\'m perfect.  Lord have mercy, am I so far from that.  I know that there has to be things that Joe puts up with in me that are difficult for him.  I really need to focus on myself and the changes I need to make.  I\'m happy, though.  Yesterday was a good day.  Today is a good day.  Everything will be okay.   '),(25423,3471937,'02,June,2004','2004-06-02 00:00:00','    \"Sunday morning rain is falling Steal some covers share some skin Clouds are shrouding us in moments unforgettable You twist to fit the mold that I am in\" --Maroon 5-Sunday Morning  That reminds me of Joe and I on weekends.  Lying in bed for an hour or so after we wake up, enjoying snuggling up with each other. Our bodies just seem to fit together in such a beautiful way while we sleep or doze or snuggle.  It makes me feel so loved and so safe.  I love the way he takes care of me.  He is truly amazing.   '),(25424,3471937,'28,July,2004','2004-07-28 00:00:00','    urlLink    That\'s me, guys&nbsp; urlLink   '),(25425,3471937,'21,July,2004','2004-07-21 00:00:00','    I had a dream about falling last night.  It wasn\'t your typical falling dream though.  I was riding this dirt bike around this canyon.  The canyon was not the size of the grand canyon. It actually resembled this gulley or valley that is at Brookside Park because it was covered with grass.  But it was much deeper than the one at Brookside Park and the canyon walls were nearly vertical.  So, I was riding this dirt bike.  The edge around the canyon was somewhat curvy and I was riding close to the edge. I was fine for a while but when I got to the other side of the canyon, there was a slight decline and I started picking up speed and barreling right toward the edge of the canyon.  So the next thing you know, I\'m riding this dirt bike nearly vertical down this canyon. I was actually having a blast.  But then suddenly my bike was gone and the canyon wall was completely vertical and I was just falling straight down. It was a very long fall, so I was having time to try to figure out keeping myself from dying. I decide that if I position my feet in this specific way then I\'ll probably be okay.  I catch this glimpse of some weeds growing out of the canyon wall. I grab hold of the weeds just in case, and somehow they end up being the most stable weeds ever.  The weeds slow my fall enough and a few seconds later I hit the ground (with my feet in that specific \"lifesaving\" position), and I was just fine.  My mom suddenly appears at the top of the canyon and yells down to see if I\'m okay.  My heart was beating like crazy but I almost wanted to do it again.  So do you think that is an indication that I need more adventure in my life right now?   '),(25426,3471937,'08,July,2004','2004-07-08 00:00:00','    I\'m feeling really happy right now.  Though, this weekend I had a spout of being really depressed. I felt like I was invisible to Joe, that we\'ve gotten to a point in our relationship where he sort of takes me for granted. The thing is that I think too much. So I start thinking that things will always be that way, that there is no hope, and that I should just break up with him now to start recovering from my broken heart and move on with my life. I know, I know, I can be a drama queen. So we talked for a little while and I mentioned that maybe he should stay living at the condo and I should live in my apartment alone. And I said that I thought he\'d rather live there than with me anyway.  But he told me that I was wrong, that he would rather live with me. He told me that he doesn\'t think about the things he does as not paying attention to me. He doesn\'t mean to treat me as if I\'m invisible to him; he just starts doing his own thing and...I don\'t know... forgets about me for a while? Anyway, he told me to just talk to him and point it out when I feel left out. I know that he is right. I know that is what I should do instead going all overboard and thinking that we need to break up.  Just sometimes I think in one extreme to the other. I can\'t help it. It\'s just how I am.  Anyway, we worked things out, and we\'re just fine now. I overreact. I know it. I admit it. But I\'m happy for the way that he balances me. So, I\'m sooo excited that Ludwig will be here in a week. He\'s my best friend ever.  When I think about all of the fun times we\'ve had, I can\'t help but have this huge smile on my face. And we\'re going to go see Incubus together. I mean, come on, what could be better than that!  On a less positive note, Dana will be leaving like the day after Ludwig leaves to go back to H-town.  Crazy, stupid girl. I wish she wouldn\'t be going. She drives me insane, but I still love her to death.  She\'s not going to be happy. But enough of that, I want to focus on the positive because there is just not anything more that I think I can do, short of tying her up and forcing her to stay here, hmmmm...which may not be such a bad idea...:P   '),(25427,3471937,'02,July,2004','2004-07-02 00:00:00','    I don\'t want to think about him so often. I\'ve moved on with my life, and I am happy.  I still think about him though. Too often.  I remember his smile and his laugh. I don\'t want to remember. Remembering only makes me try to justify him breaking my heart so many times. I\'ve already done that way too much forgiven him, justified his actions; I\'m done with it. It only hurt me worse.  I have no desire to be with him. But I still think about him. Why is that? Is my heart and mind once again not in sync? Why can\'t I wipe him out of my life completely? I want to forget. It doesn\'t help that I\'m still friends with all of his friends.  Everytime I talk to them, it is a constant reminder of how it once was, or how I once wanted it to be always.  I loved the illusion.  The illusion disappeared. Reality destroyed my dreams.  That\'s okay, though. I learned to deal with it.  I learned to move on. I just don\'t want to remember anymore.   '),(25428,3471937,'01,July,2004','2004-07-01 00:00:00','    I miss having friends.  Okay, so I still have a lot of friends but they are all so far away.  I miss having people I can do things with all of the time. Not just talk to from time to time on the phone or online.  I\'ve never had so much trouble finding friends.  But I guess I\'ve always been in environments conducive to finding a lot of people with similar interests.  When I was in college or working at restaurants or pretty much in any other stage of my life, it was crazy easy to make friends. But now, where I work, I\'m like 15 years (at least) younger than everyone.  So there\'s nobody at work to be really good friends with.  The guys we hang out with are great, and they are my friends, but we rarely have girls hang around, except for Lindsey.  Lindsey is nice enough, but there is some reason that she and I don\'t hit it off completely. I can never quite put my finger on it.  I did meet a girl during my insurance classes, but I left her number in a pair of jeans which are at my apartment, and I keep forgetting to get the number to call her.  I really should at least give it a chance.  She was in the same situation as me.  Her husband was transferred here; he made a lot of friends easy because of his job; she left all of her good friends back home and she also works in a type of place that doesn\'t have anyone her age or that enjoys what she enjoys.  So I really need to make the effort.  We could really become good friends.  Bailey and I also have potential to be really good friends, and now that Dana is leaving she will need someone to hang out with too.  Anyways, I miss my friends.  On a happy note, it is only 2 weeks until Ludwig will be here, and we will go to the Incubus concert!!!   '),(25429,3471937,'06,August,2004','2004-08-06 00:00:00','    She left last night. Actually early this morning. I want her back. My heart aches for her. She was supposed to call me the first time she stopped. But she didn\'t. I\'ve called her like 20 times, but she hasn\'t picked up. Why did I let her go? Why didn\'t I just wrap my arms around her and force her to stay as I wanted to so badly. If something happens to her, I don\'t know how I will make it through. She stopped by the condo at 3am. I was half asleep, but I gave her the cookies I made and I hugged her goodbye. I told her how much I love her. It doesn\'t seem enough. Please let her be okay. Please, please let her be okay. I can\'t lose her. Help me to be strong. I\'m so afraid. Help me to be strong. Oh please, take this pain away.   '),(25430,3471937,'04,August,2004','2004-08-04 00:00:00','    I\'m going to say it the way it was said to me. Sonja was robbed at gunpoint yesterday. They took her purse and her car, and they robbed the store where she works. They threatened her that they would come after her if she called the police. They tied her up and put a gun to her head. She said all that she could think was that Christian wouldn\'t have his mommy anymore. I freaked out when Mom told me. I know that physically Sonja is okay, but she has been having emotional problems lately and this has sent her over the edge. My heart goes out to her. I\'m just about over the edge myself. There is just so many bad things happening, Dana is planning to go back to Daniel, Maw Maw\'s health is failing her, and now this. I\'m having anxiety attacks lately. I\'ve had maybe four or five of them now. And the latest thing is that, I feel like I can barely breathe not only when I\'m having an attack but al      most all of the time. Like I can never quite catch my breath. I\'m constantly doing stretches and trying all kinds of things to calm myself down, but I\'m at my breaking point. Or at least I feel like I am. I did some research on anxiety attacks and read the symptoms, and that is definitely what is happening to me. Feeling like I can\'t breathe, extreme changes in temperature, chest pains, a racing pulse, thinking that I\'m going crazy or that I\'m going to die. Yep, that\'s exactly how I\'ve been feeling. The thing is, I don\'t want to go to a doctor. I\'m trying to do breathing exercises, and I\'m looking for natural treatments. I just don\'t trust the types of drugs that the doctors give you for this kind of thing. There are too many side effects. I think that I\'ll be okay. I\'ve never had any problems like this. Even when I was so extremely depressed back when Paul and I broke up and my family wasn\'t having much to do with me, I was very depressed but I never had anxiety attacks. Now, it is just bad things happening to everybody else. I think it is the fact that I have no control over the bad things happening now. Before, I had control over my life even if I was severely depressed and I looked for a way to fix things. I don\'t have the power to fix the things that are happening now. It makes me feel so helpless. I thought writing would help me to feel better, but it hasn\'t. So I\'m going to stop now.   '),(25431,3471937,'01,August,2004','2004-08-01 00:00:00','    It\'s Sunday, and it is beautiful. Sunny and gorgeous. I spent most of the day at the pool.  I\'m reading a new book.  It\'s about the friendship of two women.  I\'m not that far into it, but I believe that one of the women is eventually going to die of cancer.  It is based on a true story in the life of the author.  It sounds as if it would be really depressing, but instead it is really the opposite, quite upbuilding.  So...I think I may be driving Joe completely crazy this weekend. I\'ve been pretty emotional and down.  I mean Dana is leaving this next week.  I\'m so torn about the whole situation.  I want to spend a lot of time with her while I can, but I also feel so depressed over the fact that there is nothing left for me to do to encourage her not to make what I feel is a really wrong decision.  Taking a step back from everything and looking at everyone around me and their reactions to the situation with Dana has been somewhat interesting.  I\'m not talking Mom, Dad, or Sonja because technically they\'re not around me.  They know more of the history than any of my friends here that have opinions on the matter, and therefore they are basically of he same opinion as I am.  I\'m not even saying that our family\'s opinion is the right one, but we\'ve gone through this so much with Dana when they\'re comes a time when you let her do what she is bound to do anyway because it kills us to keep trying to stop her and watching her make the same mistakes over and over.  So, we\'re stopping the fight, at least for now, letting her win this battle.  Hopefully, it will not hurt her more but develop her personality into one so strong she will be able to face any challenge in her life.  So I\'m talking about Joe\'s reaction, Evan\'s reaction, Bailey\'s reaction, Tyler\'s reaction, Jay\'s reaction.  Joe\'s reaction is somewhat obvious and predictable if you know Joe well. He doesn\'t deal well with people when he considers that they are acting stupidly.  He considers Dana\'s actions to be stupid and on top of that he sees the pain that her actions cause for me and my family, and that makes him avoid her when he can.  Evan\'s reaction.  Actually hard for me to put in words.  I think he realizes that I\'m doing the only thing that I can at this point.  I think that he would probably handle things differently if he were in my shoes. How, I\'m not sure.  I know he doesn\'t think that she should leave, but I\'m pretty sure that he knows it is inevitable.  Bailey\'s reaction. Bailey is very sad that Dana is leaving. In the short time that she and Dana have gotten to know each other, they have become very close friends.  I think it breaks Bailey\'s heart that Dana will be leaving. But she is not discouraging her from doing so, she is pushing away the pain.  She is trying to be a supportive friend, but not knowing the complete situation as those of us within her family do, I don\'t think she realizes how bad of a decision Dana is making in wanting to go back with Daniel.  Tyler, Bailey\'s older sister, sent a card to Dana, basically telling Dana to live everyday to it\'s fullest, to try to make the right decisions in life.  I know that she is trying to give Dana some positive things to think about, even though she is in total disagreement with Dana\'s choices. I like her, even if Bailey and Dana thinks she can be pushy sometimes.  We\'re big sisters, we\'re supposed to be like that.  Even though it may be just by a few years, we do have a bit more life up on our sisters and we just don\'t want to see them make some of the same mistakes that we\'ve made especially in matters of the heart.  Jay\'s reaction. Okay...my reaction to Jay\'s reaction is to tell Jay to shut up.  I like Jay, a lot, a whole lot.  He\'s a great guy.  He was one of Joe\'s first friends when we moved to Indy.  He even fed Mr. President for us when we went to H-town for vacation.  He loves his friends.  He really really loves his friends. He loves people, and if he considers you a friend well then he really loves you. When he and Dana went to pick up oil for her car the other day regarding her situation, he told her to \"just do it\".  If that\'s what she wants to do, he told her to go for it.  I just feel that he barely knows anything about the situation.  Despite the fact that I think Dana will do exactly what she wants to do, I just don\'t feel that she needs encouragement.  She doesn\'t need to have somebody patting her on the back saying, it\'s just fine that you want to go back to the druggie ex-boyfriend of yours who was abusive and treated you like shit.  Don\'t worry it\'s totally fine.  NO, JAY, IT\'S NOT FINE, AND SHE DOESN\'T NEED TO HEAR THAT IT IS.  I know that I\'m emotional over it, but I just don\'t think that she needs to hear that what she is doing is okay.  And yesterday, he asked if we are having a going-away party for her.  I didn\'t say a word but of course my thinking was \"why in the world would we want to celebrate her bad decision?\"  I would feel like I\'m telling her that I support what she is doing, and I don\'t want to do that at all.  But obviously, what I was thinking was evident in my face because Jay said \"Karoline, don\'t give me that look.  I just want to be involved if we\'re going to do something special for her before she goes.\"  It made me feel as if I\'m the evil older sister who doesn\'t want to do anything nice for her younger sister.  It\'s not that way. I love her.  I still love her so much.  Nothing will ever change that. I just think that at this point in her life she is making really bad choices.  I don\'t know maybe we should show her just how much we care and how much we are willing to do for her, so that maybe she\'ll realize sooner rather than later that she\'s made a big mistake and she\'ll find a way to come back here to people who really care about her.  Would that be the better way to act? I don\'t know anymore.  I\'m completely drained. I don\'t know the best thing for me to do. I just don\'t know.  Ah, my sweet little Dana...your heart is just too soft my little one.   '),(25432,3481255,'31,May,2004','2004-05-31 00:00:00','   Here we are! My third post. Yep. And I still don\'t have much to talk about. I can\'t really talk about the weekend \'cause I didn\'t really do much. Well, yesterday we cooked the leftover hot dogs & burgers from a previous BBQ at a friend\'s house. That\'s something! Eh...see you next post.   SONG OF THE DAY  The Coral - Dreaming of You  '),(25433,3481255,'29,May,2004','2004-05-29 00:00:00','   Sitting here in my pajamas with nothing to watch on TV (sigh).  So bored. I should get out, but the powers of the pajama holds me in it\'s most comfortable grip. Anyways, yesterday I decided to try out this blog-thing for a while, with the idea that I might eventually have something meaningful to say here (but I doubt that). Let\'s see how it goes!   SONG OF THE DAY  Magnetic Fields - Chicken With It\'s Head Cut Off  '),(25434,3481255,'28,May,2004','2004-05-28 00:00:00','   \'Kay, I just created this as a test. We\'ll see if I maintain it though....  '),(25435,3481255,'28,June,2004','2004-06-28 00:00:00','   I sit in front of the mocking glow of the computer screen with my legs crossed, right over the left. My brow furrows, my fingers sleep on black keys, and nothing but a thought echoes between plastic headphones:  What the f*** am I going to write on my blog!?   Inner thigh itches. Lingering minutes that refuses to pass buzzes about annoyingly. I scan my faux-wood desk looking for beautiful inspiration, but find none. An empty glass, miniature men & women, and papers strewn about is all was spotted. My knee itches. I am reminded of the thousand-and-one mosquitoes my friends and I encountered at the park earlier that day. How I hate the tiny pricks.  A yawn washes over my body at 1:04 am. I stretch as many muscles to the tune of the blissful yawn, because I know the moment to follow would be a completely relaxing feeling. A thought comes to me, hasten by the fact I should get some sleep. It whispers: take the last seven seconds and stretch it into three paragraphs of filler. Keys dance without hesitation. A barely audible titter escapes my mouth, finding humour in the fact that someone will see through my sad attempt at being a writer. With sleep beckoning to me, I\'m sure there will be spelling and grammatical errors, but I don\'t care. I have finished a post, ending several days of creative drought on my blog.   SONG OF THE DAY  Raveonettes - That Great Love Sound  '),(25436,3481255,'23,June,2004','2004-06-23 00:00:00','   You need to check out  urlLink Diesel\'s latest campaign for their Fall/Winter 2004 Collection . Their whacked-out design team some how were able to take their keen fashion sensibilities and twisted it to work within the context of an old sci-fi/fantasy comic magazine you\'d find in the 70\'s. Incredible!   SONG OF THE DAY  Unicorns - Emasculate the Masculine  '),(25437,3481255,'21,June,2004','2004-06-21 00:00:00','   A while back I was messaging with my friend Jen, and I made up a poem on the spot just because. A ridiculous ditty that may bring joy if even for a moment. So today I wish to share it, give you a moment or two of cheer, or at least a seed of inspiration to write something silly.  Â Â Â  Upside down umbrelly to catch all your precious rain  Â Â Â then sip the goodness of rain that fell from the heavens  Â Â Â the rest of the heaven juice will be boiled to make tea for two  Â Â Â with a dollop of honey for sweetness     SONG OF THE DAY  Yesterday\'s New Quintet - I Am Singing  '),(25438,3481255,'20,June,2004','2004-06-20 00:00:00','   One of the good things about these kind of Sundays is you\'re left with nothing but your meandering thoughts to entertain yourself. It\'s interesting what runs through your head, you go through two different states of thought (or at least that\'s how my mind is working at this moment): Analytical and Creative.  With analytical thoughts you start analyzing things (outwards and inwards) in your life. Creating a mental list of things that can be better organized or improved upon. And within that list a sub-list of steps to take to create change is created. Now to initiate these changes, motivation is needed. The tasks you\'ve listed may seem daunting, but can be done with the proper inspiration to push you. That\'s where the creative thoughts come to play.  Your creative thoughts can inspire you to action or it can distract you. The mind poses different questions or situations where you come up with solutions or possibilities. Depending on the solutions or possibilities generated, motivation can be built. And I just lost my train of thought and forgot where I\'m heading with this. :P  Oh, well. Happy Father\'s Day, everyone!   SONG OF THE DAY  Peter Tosh - Stepping Razor  '),(25439,3481255,'17,June,2004','2004-06-17 00:00:00','   Today\'s top story:  Man discovers life does not stink. It has been revealed that the stench that was thought to be emanating from life was in fact coming from the world. When asked about this recent discovery, man replies \"The world has indeed stunk for many years, tests that were performed earlier show that man just gotten used to the smell.\"  In other news, man continues to bitch and complain like a baby on his blog. More on that after weather and sports.   SONG OF THE DAY  Queens of the Stone Age - Never Say Never  '),(25440,3481255,'16,June,2004','2004-06-16 00:00:00','   urlLink Murphy\'s Law: If anything can go wrong, it will.   All too often this stupid law proves itself irrefutable with the universe\'s innate ability for ineptitude. Through the years, new laws are formulated and added; it\'s gotten so numerous, laws can be split into different categories like bus laws and nurse laws. And as much as I loathe to add a new law, it should be included in it\'s Love Laws: If you are interested in someone, they can  AND will have a boyfriend and/or think of you as just a friend.  I find this cynical way of thinking very stinky, but it feels like all I can do is give in. I\'m tired of the \'love game\', why should I play when I can\'t even get a chance to bat. How often can one dust himself off and stand back up without having to hunch? Where do you go from there? You go to the gym, then to Perkins for some Eggs Benedict, then complain about it on your blog, that\'s what.  See you in the funny papers.   SONG OF THE DAY  Spoon - I Summon You  '),(25441,3481255,'13,June,2004','2004-06-13 00:00:00','   Yep. Sunday. And I\'ve never felt so unproductive in my life. I did absolutely jack squat all day, I just sat in front of this ol\'computer and watched a less-than-stellar anime series,  urlLink Ikkitousen  (i dunno, maybe you\'ll enjoy it more than me). The only time I got up was to eat or go to the bathroom. I know there\'s a lot of things I could do, I\'m just not motivated lately. And I really HATE that feeling. Gwah! I should just go to sleep. Good night, people.   SONG OF THE DAY  Lawrence Welk - Good Night  '),(25442,3481255,'12,June,2004','2004-06-12 00:00:00','   Know the real Harold...according to  urlLink Master Rao . I don\'t believe astrology, but I\'ve always found them damn fun! This reading was taken from combining your western and chinese sign.   \"You function almost exclusively with your heart, your reason having no word to say here. With your extreme emotivity, you sift everything through the sieve of your affectivity. Even your intelligence is always colored with sentimentality.    Loving beauty under all its forms, you\'re inclined to facility and like to profit by the unexpected. You obtain much from others by your charm and kindliness, and can get things accepted without having to fight. In fact, you\'re more made for enjoyment than for struggle. Your greatest flaw is touchiness.    You\'ve gifts for studies and much chance in examinations. Your taste for letters, publishing, bookselling, for all that\'s related to distribution of art and beauty can bring you great profits. You also have a good contact with the public and frequently enjoy certain notoriousness. Your friendship or love relations play a great role in your material destiny by facilitating your fortune.    In love, you\'re gifted for pleasing even though you hesitate to attach yourself profoundly. What you\'d prefer would be to live a merry single life. But you\'ll also be happy in a union where fantasy and an easy way of life prevail.\"    SONG OF THE DAY  White Stripes - One More Cup of Coffee  '),(25443,3481255,'10,June,2004','2004-06-10 00:00:00','   urlLink Just one more reason to love Japan.  Spider-Man was cool before, but when you add a cool car, a giant robot, evil monsters, and set it in 70\'s Japan, you get all kinds of ass being kicked!   SONG OF THE DAY  Ray Charles - You Don\'t Know Me  '),(25444,3481255,'09,June,2004','2004-06-09 00:00:00','   Blogs. Everyone should have one. It\'s like one\'s personal soapbox where one one can rant and rave about how it sucks when you\'re at the gym and you sit down at a bike realizing too late the previous biker neglected to wipe it down leaving his nasty sweat and B.O., or showcase the latest pic of your drunk lady friends flashing their lop-sided boobies to any horny by-standers willing to look (which I don\'t have, sorry). It\'s kind of empowering that way.  Recently my friends,  urlLink Rob  and  urlLink Ray , started up their own blogs and I\'m trying to get more people into it. I found it cool reading what insights they have when you\'re not hanging with them at the moment. Insights like:  Don\'t you hate it when you have to fart, but someone comes up to you and starts a long conversation, and you try to hold the fart in, but can\'t, so you try to let it out in a way it doesn\'t make a noise and pray that it doesn\'t stink? Yeah, I know I hate that.    SONG OF THE DAY  Kylie Minogue - Chocolate  '),(25445,3481255,'08,June,2004','2004-06-08 00:00:00','   urlLink Le Parkour  is a relatively new sport created by David Belle. You\'ve probably seen  urlLink commercials  with Le Parkour in it. Imagine Jackie Chan taking up skateboarding, only he forgot the skateboard.  I find the stuff they do really cool. I\'m surprised I don\'t see more of it around. Heck I\'d take it up if it weren\'t for the fact I\'m, y\'know, lazy.   SONG OF THE DAY  Human Beinz - Nobody But Me  '),(25446,3481255,'07,June,2004','2004-06-07 00:00:00','   I was going through my old bookmarks and re-discovered a  urlLink chocolate chip cookie recipe  I found while traversing the World Wide Web. Oooh, man! Those were good cookies! My mouth is just watering just thinking about it, I wanna bake them again. Yeah, I bake on occasion, so what? Wanna fight about it?   SONG OF THE DAY  Clarence Carter - Making Love (At The Dark End Of The Street)  '),(25447,3481255,'06,June,2004','2004-06-06 00:00:00','   In my opinion,  urlLink Homestarrunner  is the funniest internet cartoon out there. It\'s consistently funny and usually have hidden \"Easter Eggs\" that keeps me coming back for seconds. Here\'s one of my favourite episodes,  urlLink check it out !   SONG OF THE DAY  Kanye West - Home  '),(25448,3481255,'05,June,2004','2004-06-05 00:00:00','   Quote:  urlLink \"Girls only want boyfriends who have great skills. Y\'know like, nunchuck skills, bow-hunting skills, computer hacking skills...\"    SONG OF THE DAY  Toktok & Nena - Ich Hang Immer Noch an Dir  '),(25449,3481255,'04,June,2004','2004-06-04 00:00:00','   urlLink You have to check this video clip out , I couldn\'t stop laughing when I saw it. God bless karaoke singers! May each and every one of you be one.   SONG OF THE DAY  Gloria Gaynor - I Will Survive  '),(25450,3481255,'03,June,2004','2004-06-03 00:00:00','   Once again NBA\'s  urlLink Hoop-It-Up  is coming to Winnipeg this weekend, and as suggested by Rob, I thought it\'d be fun to profile the team slated to dominate all suckers on the court and take 1st place in their division. Ladies and gentlemen, I bring you (dum, dum, dum....)  Max High All-Stars!!!    Rob Carino  AKA Max High  Team Captain  This man\'s proven that a man can bleed basketball. Through injury after injury, this outside treat is still ready to go high....MAX HIGH!   Rhoden Chavez  AKA 2 Hype Hesitate for a moment and Rhod will school you 5 times and call you Susan! Size will not intimidate him, so you better bring your game on.   Ray Hamel  AKA the Weatherman Eye of the Tiger may be a cliche phrase, but still rings true when describing the man called Ray. Through will and determination, he\'ll take it to you any way he can.   Chris Kolev  AKA 2 Dope You\'d think his height would be his major advantage, but it\'s not. Chris brings mad skillz (yes, with a \'z\') to the table and will smash the table with said skillz.   Rey Quiogue  AKA Your Daddy Don\'t let his smile fool you, inside is a monster waiting to be unleashed. Get into his space and he\'ll shut you down before you even think of going up.    SONG OF THE DAY  Gang Starr - Skills  '),(25451,3481255,'02,June,2004','2004-06-02 00:00:00','   The sun was shining with hot temps and a slight cool breeze, felt like I accomplished quite a bit at work, I got a haircut, took time to sit at one of those neighbourhood parks, had gelati with a friend who I haven\'t seen in a while, and I just finished up the day with a refreshing shower after a nice little workout. Today did feel good, almost perfect. The only thing that can make it better is if a gorgeous woman fell into my lap carrying a million dollar cheque with my name on it. Well, there\'s an hour and change left in the day, so here\'s hoping!  A couple of friends been saying my blog needs some pictures of naked chicks. Well, if there\'s enough requests for it I\'ll do it, but in the meantime they\'ll just have to settle for  urlLink this .   SONG OF THE DAY  Broken Social Scene - Pacific Theme  '),(25452,3481255,'01,June,2004','2004-06-01 00:00:00','   urlLink You bored?  Here\'s a cool little Flash game that was forwarded to me by my ol\' pal Mike. Whenever I got several minutes to kill, this game has continuously kept me entertained. There\'s something about shooting a stickman full with arrows until he bleeds profusely that makes me smile like a little child.   SONG OF THE DAY  Iron & Wine - Each Coming Night  '),(25453,3481255,'30,July,2004','2004-07-30 00:00:00','   Note: I\'m probably missing a lot of details in this description of our Vegas trip, maybe because it was undescribable!   Probably goes without saying that Rob and I were estactic to be heading to Las Vegas, Nevada for vacation. It was our first time in the city of bright lights, Elvis, and showgirls, and it showed. As we were about touchdown, our excitement mounted higher and higher at the sight of desert and mountains (or as we grew to call them \'mow-taings\'). The desert heat hit us immediately as we left the ear-popping plane. An amazing 40Â°C (and it maintained that temperature for the rest of the vacation)!  At the  urlLink Tropicana Hotel  they didn\'t have the room we requested available, so they upgraded us to an Island Tower room. Bonus! We quickly threw our baggage into the room, so we can explore the Strip, but not before taking notice the king-size bed had mirrors surrounding it at the top, right, and above. Funny. Now, we walk. This being our first time here, our senses are loaded with bright flashing colours selling tonights show and sounds of the 80s blaring from hidden speakers, creating feelings you get when you first enter a candy store as a kid. We walk forever in the heat seeking refuge at cooling stations and air-conditioned casinos serving free beer, we thought we can go on without sleep, so we wouldn\'t miss one beat of this city. Fat chance! We wuz tired!  The next day we continue walking despite blisters from the other day. Walking past vendors handing out strippers-to-your-door fliers we found  urlLink Madame Tussauds Wax Museum . There, we posed with many celebrity wax-replicas. Touching and prodding them as much as the curators would allow. Very fun. At night we head over to  urlLink Cheetahs , the topless club where they filmed the movie  urlLink Showgirls . Surrounded by beautiful women, I kinda got carried away with the booze, making the next day a write-off. Sleep and vomitting took up the whole day, leaving me enough energy to eat at  urlLink Todai  for a Japanese buffet dinner.  Tuesday was spa day. We booked that day to get ourselves pampered a bit. I got an aromatherapy wrap and a massage, and it felt good! At the end of it all my skin felt smooth, my body relaxed, and I smelt like coconut, almond, jasmine, and eucalyptus at the same time. And at night I decided to take in a show. I went to see  urlLink Zumanity  and it was awesome! Nothing beats watching half-naked contortionists, acrobats, and dancers performing unbelievable feats.  Highly recommended!  Then it was time to leave. I was sad to go. I had to take in the sights one last time, trying to burn in the imagery into my mind what my camera couldn\'t capture. We board the plane. Rob gave up his window seat, so I can see the Vegas terrain as we took off. Lotta mow-taings. At our stop-over in Minneapolis we were running late, so we had to run for it like you see people do in the movies. Finally we were home in Winnipeg. It was weird coming back, we had just left a city with bright lights and +40Â° weather to a flat town with a mild 15Â° temperature. It felt so laid back compared to the energy of Las Vegas. I kinda missed it. But can\'t wait to leave it again! ;)  Photos of our trip can be found  urlLink here .   SONG OF THE DAY  Jody Watley - I\'m Looking For A New Love  '),(25454,3481255,'22,July,2004','2004-07-22 00:00:00','   In about 34 hours me and my pal, Rob, will be boarding a plane to fabulous Minneapolis! What\'s in Minneapolis? Another plane to wisk us off to  urlLink Las Vegas, Nevada ! What\'s in Las Vegas? Nothing but 4 nights of family-oriented sin!  I\'ve lost sleep thinking about Vegas. It\'ll be the first time to fly in a plane. It\'ll be that farthest I\'ve even been from home (Edmonton being the furthest so far). It\'ll be just two guys in a town geared for having a good time and we have no gameplan whatsoever! I was just thinking of standing in the middle of the street and letting the natural flow of the city\'s energy take me whereever it wants. If anyone happens to see us, feel free to join us in the grand adventure they call Sin City. Drinks on whoever feels like paying.  On a different topic, if you\'re philipino born in North America, please rent  urlLink The Debut . It\'s like watching your family onscreen...only with a dance-off. Please check it out.  \"An award-winning English language film for the whole family, \"the Debut\" revolves around Ben Mercado (DantÃ© Basco), a talented high school senior who has rejected his Filipino heritage. The long-simmering feud between Ben and his immigrant father Roland (Tirso Cruz III) threatens to boil over and ruin the 18th birthday party of Ben\'s sister Rose (Bernadette Balagtas). But to Ben\'s surprise, his sister\'s celebration challenges his sense of misplaced identity, and the way he regards his father and grandfather (Filipino film legend Eddie Garcia). In one night, Ben faces the true nature of his relationships with his family, his friends, and himself.\"  Summary by Anonymous on  urlLink IMDB.com     SONG OF THE DAY  Christina Milian - Dip It Low (JJ Flores Extended Mix)  '),(25455,3481255,'19,July,2004','2004-07-19 00:00:00','   urlLink Ika, the Calamari Wrestler . I really don\'t know what to say about this. Um... Hooray for outside-the-box thinking? Uh, yeah...the trailer to this is 2nd from the bottom right.   SONG OF THE DAY  Mase - Welcome Back  '),(25456,3481255,'18,July,2004','2004-07-18 00:00:00','   And guess what I did to enjoy this weather? I went shopping (was at Polo Park  urlLink 3 days in a row ), had ice cream (2 for $2 sundaes at  urlLink McDicks ), and watched a couple of rentals in my cold basement (highly recommend  urlLink 50 First Dates ). :P   SONG OF THE DAY  Specials - Hey, Little Rich Girl  '),(25457,3481255,'14,July,2004','2004-07-14 00:00:00','   urlLink This trailer freaked me out!  If it weren\'t for the fact I have complete control over my bowel movements, I\'d crap my pants after seeing that. Definitely for the horror/thriller aficionados. Geez... I\'m gonna go listen to some  urlLink Quiet Storm  to put me at ease.   SONG OF THE DAY  Patti Austin & James Ingram - Baby Come To Me  '),(25458,3481255,'08,July,2004','2004-07-08 00:00:00','   This past Monday and Tuesday, me and my amigos were at the  urlLink WWE  Raw and Smackdown shows, respectively. Even though we had less than spectacular seats, it was a fun time filled with blood, sweat, and smoke! Good times, good times.  Though I\'m fully aware of it\'s homo-erotic undertones and it\'s ridiculous storylines, there\'s something about wrestling I find appealing. I think it\'s because it\'s the closest thing we have to real-life super heroes. Think about it: characters in colourful costumes, larger than life physiques, good guys vs. bad guys, people performing incredible feats and taking inhuman amount of pain. The only difference is the whole \"flying and shooting lasers out of your eyeballs\" thing, and I think they\'re going to have the Undertaker do that next week.   SONG OF THE DAY  Dandy Warhols - Bohemian Like You  '),(25459,3481255,'01,July,2004','2004-07-01 00:00:00','   G\'day, eh? I hope everyone\'s enjoying their day off to celebrate  urlLink Canada Day . Whether it be camping, catching the fireworks display, or just getting piss-drunk. Me? I did what any good Canadian would do and saw  urlLink Spider-Man 2 .  I have to say that I enjoyed it. Even though the movie had the same amount of cheesiness the first one had (must be a  urlLink Sam Raimi -thing) and many parallels can be made to the first movie (villain created in lab accident, kidnaps MJ, yadda yadda yadda), I thought the overall presentation was superb. The effects were better, fight scenes were really cool, and you still feel for Parker (the poor sap).  Verdict: Well, it get a (some-what biased) thumbs up from this comic-nerd! And if you\'re wondering what Spidey and Canada have in common? Not a damn thing, you hoser.   SONG OF THE DAY  Hidden Cameras - Boys Of Melody  '),(25460,3481255,'01,July,2004','2004-07-01 00:00:00','   I was going to go on a rant on how addicting the internet is, but I didn\'t have the energy to do it. BUT...It did remind me of one of the reasons why it\'s addicting:  urlLink mindless dribble that yet somehow manage to entertain .   SONG OF THE DAY  Prince - I Could Never Take The Place Of Your Man  '),(25461,3239841,'21,May,2004','2004-05-21 00:00:00','   I look back on my life, through some of the good and bad, and I realize just how much of an effect my parents have had on my life.  Some of it was good, but a lot of it was bad.  My parents have been very restrictive in respects to me from the beginning.  I was only allowed to associate with people they approved of, for example.  I could only listen to their music.  I could only believe what they wanted me to believe.  Looking now, they control every single aspect of my life, no exceptions.  What I eat, what I wear, what I listen to, must all meet their approval or somehow be smuggled past their notice.  My friends they control as much as they can, although, thankfully, it is much harder at my age to control who I meet and befriend.  It has only been what, a month or maybe a month and a half, that they\'ve let me buy my own music.  I still can\'t wear half the stuff I would.  It\'s \"satanic\" and evil, or some such nonsense.  The point of this post isn\'t to rant, though, its to say that this control, of which you have a small tast of in the above description, has really destroyed my life, and prevented me from rebuilding it.  My parents have always had a sort of negative aura or somethign around them.  They were never people I felt comfortable around, they had something more than just that parential air about them that said that something was wrong.  I\'ve never been able to place it, but it exists to this day.  So, it is difficult to tell them how I feel.  Not that it makes any difference.  They are stubborn and close-minded, and are very slow to accept things that are in their favour, let alone something that challenges their preconceived notions.  Thus it\'s been a hard fight to even get them to allow me to listen to things such as metal which they have labelled as \"evil.\"  This really held me back from forming my own ideas of the world I think, because it really stopped me from having my own feelings, I always had to feel someone else\'s emotion, how someone else wanted me to feel.  After all, when someone is trying to control what you listen to or see, really what they are doing is controlling (or trying to control) what you think.  I never could introduce Reena to my family, as much as I longed to, because my parents think goths are similarly \"evil\" and that wouldn\'t fly.  At best I\'d\'ve been barred from seeing her again, at worse they would have gotten into a huge argument.  So that relationship had to be kept hidden - for FOUR YEARS - not an easy cross to bear, and I\'ve slipped a couple times, but they didn\'t pick up.  At least, they haven\'t given me any indication that they had picked up on that relationship.  The clothing issue is really a sticking point too, because clothes are something I value as a self-expression.  I wanted to go in all black to the prom - cause that\'s my style - I didn\'t even ask for any of the multitude of weird and wonderful things people have: spikes, chains, studs, whatever - just all balck.  Of course, you can guess the answer.  I can\'t or couldn\'t be me.  It was the helplessness I felt not being able to be myself that contributed the my depression, the same one that led to my suicide attempt.  Furthermore, not only do they enfore these rules strictly, but there is no dialogue to be had about them.  They are there, and if you question them there is hell to pay (I know, I\'ve tried more than once.)  When I haven\'t been allowed to be me, how will I know where to go when I\'m building my life back up from scratcfh as I am from now?  How can I know if that direction is true to me?  '),(25462,3239841,'20,May,2004','2004-05-20 00:00:00','   This is just something I was talking with Jessica over, and I need to get off my chest.  Psycho-analyze away, but it helps me to talk and it helps me to feel that I don\'t have to keep something as a deep dark secret.  Jessica and I were talking about what our biggers \"what ifs\" were, and this is what I said:   Leading up to that day I wasn\'t feeling great at all, I mean physically too, not mentally, but my parents didn\'t help, you could go up to them with a bicycle halfway up you ass and they\'d still send you to school... my marks weren\'t very good either, despite me getting tons of extra help from my friends... I had just failed two big tests in math and I knew my final mark was going to be crap.   And around then some of my friends were really cutting into each other and I tried to step in a be a peacemaker, but then it just ended up getting them all ganging up on me...   So I was feeling pretty crappy, and on top of all of that Reena and I had been fighting so I was pretty volatile around then,..   The thing that set me off was this huge fight I had with my parents the night before, they\'re going on and on with me cause I complained about my brother and sister saying shit about me and that backfired big time cause then my family ganged up on me ...   So it really felt like everyone was against me and I had my back against my wall, I was physically and emotionally exahausted, so I took the cleaver from over the refrigerator that night and I took it to my arm.   Two and half hours later (that\'s how long I went at it), I was lying on my bed with my blood all around me and my head spinning from the loss of blood, and like I said, all I was thinking about before I went unconscious was \"God damn, god damn ... Reena\" and then I slipped into unconsciousness...   It was sort fo surreal when I was unconscious because it was like I was aware of something, but not something in this world, it was like I was im limbo or something, idn. I felt like I was suffocating like I was being crushed before I went unconscious, but after I felt just like, empty.   I don\'t know, but when I woke up that morning it was a scramble to cover up all the blood on the sheets, because my parent\'s would have finished the job if they saw the blood.   I don\'t know, all I can think of in retrospect, is \"what if I had? how would it have made reena feel?\" God damn, it would have hurt her so very much...    To move forward, I think i really have to take a look at my past and look at it long and hard, to see what I\'ve done wrong.  A lot of my past has remained my evil in a closet, something I didn\'t want to touch and something I didn\'t want to remember.  But how can I learn the lesson I am to learn from my past if I don\'t reflect?  I still don\'t know what I was supposed to learn from my suicide attempt though...  I was reflecting back on it, and these lyrics fit the mood and feel of that memory well, I think:   \'Scuze me while I tend to how I feel These things return to me that still seem real  Now deservingly this easy chair But the rocking stopped by wheels of despair  Don\'t want your aid But the first I\'ve make For years can\'t hold or feel No, I\'m not all me So please excuse me While I tend to how I feel  But now the dreams and waking screams That ever last the night So build the wall, behind the crawl And hide until it\'s light So can you hear your babies cryin\' now?  Still the window burns Time so slowly turns And someone there is sighing Keepers of the flames Do ya hear your names? Can\'t you hear your babies crying?  But now the dreams and waking screams That ever last the night So build the wall, behind the crawl And hide until it\'s light So can you hear your babies cryin\' now?  Mama they try and break me  - Hero of the Day, by Metallica  '),(25463,3239841,'19,May,2004','2004-05-19 00:00:00','   I have to admit, I\'m not very good at this relationships thing.   Maybe that\'s why I haven\'t had maybe.  I\'m in a bit of a conondrum.  You see, there is a woman I know and talk with a lot, by the name of Jenna.  To be honest, I think I love her, though I haven\'t felt much of that, so I dunno.  I\'ve been trying to find the strength to tell her how I feel about her.  But there are things complicating the whole thing.  You see, she is already in a relationship with someone else, whom I think is an okay guy from the the one time I met him, and she has seemed happy with him, so I\'ve been holding back.  I\'ve been worried about destroying that relationship, because I know it would hurt her, and that\'s the last thing I want to do to her.  Recently though, she has seemed less happy with him, and I don\'t know.  I tend to be a sensitive person and can pick up on emotion, but she is a bit of an enigma to me.  (I suppose that\'s one of the things that attracts me to her.)  A part of me wants to wait and see how that pans out, but another part of me seems to have this sense of urgency, like if I don\'t tell her soon, something will happen.  It\'s confusing.  I don\'t know if I can find the strength to tell her how I feel, but then how else do I?  It does not help things that I\'m a shy person, and find it hard to talk to people period, forget telling people, even one I trust as much as her, something so close to the heart.  But at the same time, I feel as if I owe it to her to tell her how I feel.  She was so much of a help in getting past the pain of Erin\'s death, and she was my shoulder to cry on through the reception after the funeral.  She has been one of the few people that has been there to talk to.  She is a kind and compassionate person.  I don\'t know, relationships are confusuing.  *sigh*  Here\'s one for you Jenna:    So close no matter how far, couldn\'t be much more from the heart. Forever trusting who we are, and nothing else matters.  Never opened myself this way, life is ours, we live it our way. All these words I don\'t just say, and nothing else matters.  Trust I seek, and I find in you. Every day for us something new. Open mind for a different view, and nothing else matters.  - Nothing Else Matters, by Metallica  '),(25464,3239841,'17,May,2004','2004-05-17 00:00:00','   America and Canada really have become the fast food nation.  Forget Christianity, forget Judaism, Hinduism, Islam, Buddhism - we have a new religion, and that is food.  We place such importance on our food, that I am somewhat surprised that we haven\'t officially elevated food to the status of a diety. (pun quite intended)  We have people who can barely raise enough money to feed themselves, forget their starving family of five, and yet it is more important that YOU get your food.  This is the prevalent selfishness of our society.  It has become a case of the rich and the well-off, the upper-classes and the middle-classes (which clearly exist), against the poor.  As if the poor aren\'t downtrodden enough.  We place our this food above our own happiness sometimes, in a way that many Christians (which is the only religion I am qualified to speak for) place the word of their God above that of their own desires and wants.  Soon, obesity will be one of the greatest killers in America - just as bad as nicotiene or cancer.  That really says something for our culture I think.  Does it surprise you to know that the bigwigs at McDonald\'s refer to their constantly returning users as \'heavy users\'?  As if they were drug addicts?  And if fast food was considered a drug, how many hopeless addicts would we have?  I\'d hazard to guess that we would have a lot.  At least half of the population.  The TRUE problem with fast food isn\'t it\'s health factor, it isn\'t even how much we eat.  IT IS THAT IT IS TOO EASY TO GET FAST FOOD.  We can just walk into a restraunt, point, say we want the number #5 meal (or somesuch) and *bang*, there you go, you have your meal.  You didn\'t have to raise the cow you\'re eating.  You didn\'t have to shoot it and kill it.  You didn\'t have to slaughter it.  You didn\'t have to cook it.  You didn\'t even have to put the condiments on the hamburger.  You just said what you wanted and you got it.  Don\'t mistake me for a vegetarian - I\'m far from - but it seems to me that our greatest strength, that ease, is also our greaetest weakness.  Because it is so easy, we are literally gouging ourselves to death.    Am I the only one that sees something wrong with that?  '),(25465,3239841,'16,May,2004','2004-05-16 00:00:00','   Together the band and I got together and made what was for us a hard descision.  Faced with our John\'s travel demands, and Reena\'s sickness, we have decided that we\'re not going to be playing the tour or doing the circuit anymore.  I guess you can call that a break-up.  This wasn\'t an easy decision for any of us to make, but with the turmoil of the past year, and so many personal problems and sicknesses to deal with, it is best for us, emotionally and physically, to just quit while we are ahead.  We will be releasing our album Soulpath, as we have been planning to, and then I think we will just go our seperate ways.  I have a confession to make, though.  It was I who first put the idea on the table.  I have had a hard time coping lately, just with life.  There is no one thing that I can really pin it on.  I just haven\'t been doing what God wants me to do.  I\'ve realized that.  I think I knew this before, I just was being a stubborn cuss and didn\'t want to admit it.  There were a few things that really spoke to me that I haven\'t been doing this as I should be.  The first was a letter I got, concerning Erin.  This is the part of it that really spoke to me:   If you wonder why none of erin\'s real friends haven\'t gone and made a  huge website for her, its because we all have real memories of her, which live in our hearts and only need to be in our hearts, not broadcasted. Erin\'s brother wrote a song for her, because he was her BROTHER, and it would really be innpropriate for anyone with a lesser relationship with her to do so. This is all justso completely too innapropriate to express completely. I can honestly understand how you would be touched by her, because so many people were truly inspired by  her, but it is definitely crossing the line when you approach her family and friends with anything more then your sincere condolences and a few kind words.   It really did hit home.  The tone of the whole thing was a bit negative, but I needed dressing-down.  (If that\'s the proper expression to use.)  I have to admit, I liked the song idea, and it was a good idea, but this person brought up a good point.  I am finding it hard to find the right word & feelingss to put down, so I think that it is best that I just stop and not do so.  If I am unable to put it together with the wholeness and sincerity that I know her family and Erin would want, then it is best I don\'t.  I do really appreciate the work and how people have come together to help me do that, but I think it is best that I don\'t, both for me, emotionally, and for her family.  This kinda leads on to what I think is the real heart of the matter, though.  You can call me crazy for saying this, but I sat down and folded my hands and talked to God today, and it really helped me to discover a lot about myself.  I have to admit, I have been doing a lot of what I do for Erin because I feel the previous things I\'ve done weren\'t good enough.  I felt as if I was not doing enough.  Somehow, I felt as if I was failing her.  Now, I\'ve realized that in doing what I\'ve done for her up till now - the pictures I made, volunteering at the fundraiser, all of that - she has smiled upon.  She is content with that.  I wanted to give more, because I felt like it wasn\'t enough, yet, I know now that it was indeed enough.  I worried so much about me, whether I was doing good enough, that I forgot who I was doing it for - her.  To admit that hurts me, but it is the truth.  I have tried to be her, to carry her message of hope and happiness, yet I have realized that it is HER message, not mine.  I feel like such a fool that I could take her message and make it mine.  In a way I suppose that I wanted to be as loved as she was, but now I know that I am fine as I am.  That\'s the most important realization anyone can make, I think.  I have lived my life trying to be someone else, I think.  Trying to live up to other people\'s expectations, trying to live up to other people\'s beliefs and ideals.  I have lived my life giving it to others, so much so that I am just learning to make that life mine, and to take time out for me, to give to me too.  I have to make my own message.  I have to be myself.  That\'s what she wants for me, I think.  That\'s what I\'m going to do.   The second thing that made me realize some of what was wrong in my life has just been my discussions with other people, with Jessica, with Donna, with Katherine, and with the people at school.  I really, seriously, think I have been living backwards.  I\'ve been giving so much of myself and not keeping any of it for me.  The writer of the email I quoted above thought I was seeking attention.  Perhaps I am.  I don\'t know.  It\'s hard at this point in my life to really look at yourself and say you know all that\'s going on inside, because you really don\'t have the wisdom to see that.   I asked Jessica to do something last night that I have never asked anyone before and I wouldn\'t usually think of asking someone to do for me.  I asked her to pray for me.  Jessica says she doesn\'t have strong faith, but I think she really must, because I am 100% sure it is that prayer of hers helped me to come to this realization of myself.  I think I have finally come to terms with myself.  I think I can finally let go of this pain inside, and embrace life.  And I\'ve longed for that my entire life.  I don\'t know what it means to you, Jessica, but I thank you with every ounce of my soul.  If one prayer like that can do for me all that I think it has, then I really need to learn to reach out to people more.  As Omni wisely noted, I\'ve been trying to help people for so long, I guess I didn\'t know how to ask for help.  I\'m no one to pass judgement, but, Jessica, you pray for yourself tonight, and you ask for help, because that was what I needed, and I got that from you, and from God.   The final piece to my self-realization was a much more pointed one, and a much more painful one, but it really did make me go back and look.  When I went over to Reena\'s today and told her these things, and showed her that email, she looked at me with a sort of odd smile, and she told me that perhaps that the time had come that I put her behind to.  That hurt me, but I understood after she explained things to me.  She told me that for a long time now she\'s been in our relationship for me, to support me, but she said now I\'m able to walk for myself, to go on for myself, that we should move on.  She said that such is what she was here to do, and now that it\'s done, it was time for us to move on.  I will always remember her, and Erin, for being a crutch for this emotional cripple through this trying time in my life, but Reena was right - I have a new time in my life ahead of me, and it is time for me to move forward and embrace that life.    I don\'t know what the future holds for me, but there\'s only one way to find out.  '),(25466,3239841,'27,May,2004','2004-05-27 00:00:00','   Just as life seems up for the people around me, it falls apart again.  Such has always been the cycle of my life, and as such I suppose should not I be the sensitive person I like to think that I am, then I would be used to this.  Even as I am it has jaded me and left my cynical, yet still I must have hope ... and it hurts every time that this hope lets me down.  Every time I die, indeed.  As you can likely imagine, I am writing now because that has happened again. :(  Today reading Omni\'s blog, I found out some very sad news, and that\'s an understatement - but words to describe those things have always been elusive to me.  Her grandmother was given from 2 weeks to 2 months to live.  I don\'t know the circumstances, but regardless, it\'s a devestating thing, an utterly destroying thing.  You can read what Omni had to say  urlLink here .  I don\'t mean to demean this by talking of me, but I\'ll share with you what has helped Reena and I through our losses, in the hopes that it may help you, Omni.  When one loses a family member - any family member it is hard - but what you always have to keep in mind is that your loved one is in a better place - heaven, nirvana, whatever you may call it - and they would want you to remember the good times, to remember the strong points - the positive - rather than the negative.  That\'s what I finally came to the realization with as I grieved for Erin.  I did her a great disservice by being so sad for her, because she always looked to the light at the end of the tunnel, and never the darkness in the tunnel, where I was the opposite.  I was living backwards, and finally I\'m crawling my way to something better after amking that realization.  Omni, you have always been a source for strength when i\'ve been down, so, if you need it, let me repay that kindness and try to be a strength for you through what will be a trying time.    And, as Erin -  Pray Until Something Happens (PUSH)  I know I will, for you, Omni.  '),(25467,3239841,'24,May,2004','2004-05-24 00:00:00','   You know, I would never wish half of what has happened to me in the past on anyone, ANYONE, but, when some of the shit I\'ve been through starts happening to one of the most remarkable people I know, someone who has helped me through a lot of the pain I\'ve had to bear, then it really cuts deep.  My good friend, Jessica, is going through some of that now, it seems.  I don\'t know how much she would want me to say, so I\'ll keep it breif:  Jessica has a friend, and it seems she has a real love/hate relationship with this friend, though I\'m not one to judge.  She has said somethings behind Jessica\'s back in the past, but other than that she has been a good friend.  But now it seems that she has decided to do this more and more, in a way that is destroying Jessica.  When I heard about it, it cut me deep, because it sounds almost exactly what Megdalena did to me, when we broke up.  She went behind my back and told a whole bunch of lies and half-truths, and worse than that a whole lot of things I had trusted with her as secret, and it tore me asunder, and burned what remained to ashes.  On one hand, I wanted to kiss and make up, and on the other, I wanted to give her a slow and very painful death.  I know what you\'re going through Jessica, and I know how much it hurts.  Just find the courage to stand up to her, and to YOURSELF.  The past is a painful thing and I am sure she will open the wounds of the past, but you have to overcome that... God, Jessica, don\'t do what I do and let that person destroy you without fighting back.    You\'ll have my prayers, Jessica.  '),(25468,3239841,'05,June,2004','2004-06-05 00:00:00','   I\'m thinking those thoughs again, god dman this all, thinking about cutting again.  Jessica, if you are out there and reading this, hold me, I need it.  I hope I can keep my promise ...  Damn it, why do my parents have to be the way they are, don\'t they see it\'s destroying me?  Your forum\'s down, you\'re not on AIM, I feel so cut off from my one sense of happiness...  [update]: I did get on aim to talk with Kat though which helped my flagging spirits ... she\'s such a kind woman ...  '),(25469,3239841,'01,June,2004','2004-06-01 00:00:00','   It\'s not easy for me to say how I feel, I suppose I\'m always afraid I\'ll be shut down or that they\'ll take it the wrong way, God knows everytime I try to tell my parents how I feel it never works out, but id, I feel I can say that stuff to Jessica, it\'s just, idn, it\'s still hard to find the words...   I owe it to her to say what I feel will, so I will.  Excuse the incoherence, but my thoughts tensd to be that way, and it\'s taken from a seires of message board posts between us.  I think it would be real cool to meet you, idn, I think we do share something, we have connected, idn, just don\'t know what that bond is, hmm... how deep it is, idn. It\'s hard for me to connect to people I think, so it\'s kinda  remarkable, idn... but either way you\'re a real kind person, and I could use one or two of those in my life...  Jessica, you really are a sweet person, and anyone who says otherwise is horribly mistaken...  Idn, I kinda think that it would be cool to meet you in person, Jessica, because it would make it seem that bit more real to me, if you know what I mean...   Trying not to come off too stupidly saying that...   Idn, it\'s kinda awkward to say, but I think I knida might like you... I\'ll avoid the other word ... but idn it\'s so hard to put a finger on my emotins cause I\'ve been so emotionaless for 4 years of my life ... but idn, I think that\'s what it is ...  Like Jenna I guess I have to say how I fell for better or for worse...     I don\'t know, it\'s just I\'ve only had one person I\'ve truly connected with before in my life, that being Reena, and now she\'s gone ... and I find that similar connection with Jessica.  I don\'t know, I mean we know so much about each other, yet so little, we need that time to mature.  The one thing I hope though, and I think I do foresee some sort of relationship - the thing is I dunno what kind - but the thing I hope is that maybe it can go deeper that the one Reena and I had, a different level, I suppose.  Reena in a way was me testing my feet in the minefeild that is personal relationships, and I think I care enough for you Jessica that I\'d be willing to take those next steps with you.  If I haven\'t come off as a weird insane person or stalker type by now I\'ll be very surprised... *sigh*  '),(25470,3239841,'09,June,2004','2004-06-09 00:00:00','   Okay, I was talking about this with Jessica the other night, and I guess I should share this experience with people, maybe as an example, maybe as a moral.  Just something to learn from, when you\'re dealing with people.  I probably won\'t say as much here as I did to Jess, but you can get the jist of it I think.  First of all, I suppose I should say who Hannah is.  She was a (pretty much) life-long of Reena\'s, that I met a while into my relationship with Reena.  She was always kind, gentle, and caring, but she was horribly shy.  She was into some of the same music I was, that\'s the one thread we all shared, and through that I began to know her and relate to her, though it was difficult to reach through that shell of hers.  When I did though we really got to know each other, it wasn\'t a boyfriend/girfriend thing, but she did have a certain afection for me.  She thought it was really cool that I\'m with a metal band and came to all of our shows (that she could).  She looked up to me as a role model, I think.  But she picked up a very bad bit of me, and it affected the rest of her life.  She knew of the one time I had cut myself and though I never did anything that I thought would encourage it, she picked that up from me.  She did it to herself much worse than I did, though.  She cut herself right to the bone, and ended up in the hospital.  Afterwards they took her away for her to get \"help\", though she came back even more scarred, physically and mentally.  That devestated her, and it really devestated me, because whether I meant to or not, whether I meant for that to happen or not, I had a card to play in that outcome, and the card I played was a very negative one.  The thing I learned from this, the hard way, was that when you are a part of the community or in the spotlight as I was for her, you have to be very careful about what you portray to people, you have to be conscious of that and how it may affect people, for as I learned, people can pick up negative things.  Just remember that they can pick up equally strong positive things if that\'s what you put your heart into showing to the world, that part of you.  Hannah and I do have a happy ending of sorts though - she\'s gotten over that as much as she can, and more importantly, on Tuesday she came right out and said she had forgiven me, and moreso, herself.  Now she\'s become what I HAD wanted for her - that same gentle, kind, and compassionate person, minus the shyness that was a barrier for her to show that to the world.  And perhaps it could be said that she gained a new insight into my own soul, going through that, and to the many others like me, and she can help them how they need to be helped.  '),(25471,3239841,'07,June,2004','2004-06-07 00:00:00','   God, and I thought yesterday was hell, but just when you think it cannot  get any worse, not only can it, but it often does.  It was worse even than the first time, the one that got me into this emotional mess to begin with.  I don\'t know, I hadn\'t thought I was into cutting recently, but tonight it was a battle, even when every ounce of my soul was poised against cutting myself.  I\'m feeling better now, big thanks going to you Jessica, I really wouldn\'t have made it through tonight if not for you, I really wouldn\'t\'ve have.  I can\'t repay what you\'ve given me, for holding on, but just know that if for nothing else, for that you have my undying gratitude.  I don\'t mean to hurt you saying this, but I do love you, and it shows that you care too, I don\'t know how much, that you stuck with me, even when you wanted to go to bed, even with your bad news, even with how crappy you felt.  Thanks.  A lot.  Because of you, I held on.  '),(25472,3239841,'25,June,2004','2004-06-25 00:00:00','   This question is asked by many people to me.  Some of these people know that I too have struggled with cutting, others had no clue, but it\'s got me thinking: what  really  got me cutting?  First of all, at least the first time, it is a reaction to a person\'s environment.  Humans are not born with natural self-destructive urges, at least, not in that way.  For me, it was a combination of a horrible school social environment compounded by an even more poisonous home environment.  I could not, and still cannot, be myself, and I was suffocating.  Cutting myself was a power thing me - it was a way of taking control.  A lot of the mindset was if they were not going to let me, then I will take control once and for all, and rid them of a life they liked to toy with so much.  That hate-bred sentiment hasn\'t left me at all, but I deal with it differently now.  After the first time, however, for the repeat cutters it becomes a self-destructive cycle.  It becomes a mental addiction of sorts.  It\'s the same with money or guns, and other things that give you power.  What do you want to do when you first find something that gives you power?  You use it as much as you can, naturally.  It is a difficult urge to try to supress.  I know that all to well.  So then what about beaking that cycle, supressing that urge?  I\'ve written about a few of the lesser times I\'ve thought of killing myself here.  I\'ve never written about some of the harder times.  There have been some real dark ones, perhaps as dark as my first one, lately, but now there\'s something, someone, there to stop me from crossing that line - Lynn.  She\'s my hand to hold and she is slowly helping overcome that.  That\'s the one thing that can break the cycle - love.  And I finally found it.  '),(25473,3239841,'20,June,2004','2004-06-20 00:00:00','   Yeah, I have a splitting headache right now... and there\'s a hell of a reason for it.  Long version: was asked by Sparky to take out the trash in the back room so I got all the stuff together, went out to the enclosure where the dumpster is and opened it in, threw in the stuff and took out the cart for the cardboard, went back, asked if anyone needed help cause the cardboard was goona be a while (lots of empty boxes), and they said, naa, just finish packing stuff up so I did, took it back and unloaded the cart. Put the cart away and went to go get the door for the enclosure and the wind caught it so I had about 400 pounds of door in my face. 7 stitches and many X-Rays later, I\'m back home.  Short version: Door comes swinging towards me. Face vrs door. Door won.  The final battle wound tally    - 1 deep laceration (cut) above left eye, 7 stitches -> doctor wanted to put more in, I didn\'t let him. - Abrasions (scrapes) on right arm (lots, ow ow ow), left leg (not so bad) and my side (kinda hurts but only if I hit it). - Possibly broken nose, but XRays didn\'t come up with anything so I think that\'s okay. - And one bloodsoaked uniform.  My sister was on the shift after me and was telling me a few things when she got back home.  She was in the manager\'s office and he had stuck a big note on the bulliten board for the other managers there... \"PETER\'S NOT DOING BACK ROOM ANY MORE\"  She was also telling me near the end of her shift they were looking for someone to do the back room cleanup ... for some reason there were no volunteers.  And apparently Sparky who had sent me out to do that was real concerned. I don\'t know he cared ... heh.  My parents caught the logs of my chatting, though they didn\'t read through them much ... but my dad took MSN and AIM off my computer piece by piece (sigh) They\'re also fairly angry with me (like that makes my pain-filled day pleasant) but I\'m pissed at them, cause I\'m the ONLY one of my friends whose parents wont let them chat at all.  And he also limited MY OWN ACCOUNT ON MY OWN COMPUTER so I can\'t install ANYTHING.  (sigh)  Excuse me while I try not to hurt myself...   I\'m kind of depressed Lynn (Jessie) isn\'t on, she has an odd way of making me happy even when I\'m hurting like I am, inside and out.  '),(25474,3239841,'07,July,2004','2004-07-07 00:00:00','   I\'ve broken away enough from my own cycle of suicide attempts that I think I can look back on it now, and I can think a little more clearly than before, so here\'s my thoughts...  First of all, to anyone who would judge me because I tried to kill myself, or to those that would tell me it\'s not worth it, remember that you havent been through the hell that has been my past, and you would never know how I felt, or the entirety of what I lived through.  For me suicide was a release, but not just from life. It was a release from all the bitterness, all the anger, all the hate, and all the pain I have felt. I am not a person who is very happy in myself, and when I see those qualities in myself, and see how much it has hurt those around me, I really, genuinely want to die.  The first cut was that, it was me feeling incredibly down on myself. The day before was the day I first found out my ex had a drug problem, and that shook me to the core, and that day school and the home were both hellish, so that evening I took a knife to my arm, my wrist, and cut to the bone. I was surprised with how pale I was that my parents didnt notice ... but then again I think they knew whatd happened and just wanted to deny it.  The second and third cuts were in similar circumstances ... when I found out Maya too had done drugs, and again when I learned Hannah had piced up cutting from me ... the third being a bit hypocritical, I know.  However after that, the fourth time and after, it became catering to a mental addiction, and it\'s a very poisonous and self-destructive cycle to try to pull one\'s self out of. I was at the point where only the help of another could, and I am eternally grateful and loving to Lynn for that.     Hope, in the face of our human distress, helps us to understand the turbulence inside  - Six Degrees of Inner Turbulence, Dream Theater  '),(25475,4044095,'15,august,2004','2004-08-15 00:00:00','   so here i am, still in mpls. waiting for the rocketships to pick me up. i have been declared the ambassador from space. thanks to the habitrail that is the downtown skyway system. soon, my posts will be&nbsp; over 1900 miles away... '),(25476,4044095,'15,august,2004','2004-08-15 00:00:00','   so here i am at my parents house, waiting for the terrible twosome to arrive by giant metal bird. i have no idea what airline they are using, so i\'m sending a psychic message to them on the plane, er... bird, to call and let me know. otherwise, i stand just inside the terminal, not a good word in an airport by the way, randomly screaming thier names. one step closer to arrest. ah well, soon they\'ll be here with a story&nbsp;to tell, and then i\'ll be there with a limited amount of money and no job. if you could see inside my head right now, i\'d be laughing maniacally. and bleeding. '),(25477,4044095,'15,august,2004','2004-08-15 00:00:00','   this will ,sadly, be my last movie night in yon midwest. we just watched the birdcage, and will soon comence on the divine road through BIG TROUBLE IN LITTLE CHINA. one of the best films ever... that is not a lie. we were having an arguement about the best lines throughout the movie. it\'s between \"sonnuvah bitch must pay!\", \"that big oaf slams your favorite head up against a brick wall, and asks you, \'have you paid your dues, jack?\'\"&nbsp; and, \"we really shook the pillars of heaven didn\'t we, wang?\". i prefer the latter, peet is a freak for the first. you never know... do you? '),(25478,4044095,'15,august,2004','2004-08-15 00:00:00','   figured out links by looking it up... check my sweet ass sidebar, links and all. word. '),(25479,4044095,'15,august,2004','2004-08-15 00:00:00','   peet got that fucking song stuck in my head. useless crap! ahh well, what can you do other than murder? complain. and since i\'m good at that, thats what i\'ve been doing about it. kay, farting noises behind me from peet, again. gastronomical fun for the whole family. or at least peet. definately peet. more when the urge strikes me. '),(25480,4044095,'15,august,2004','2004-08-15 00:00:00','   alright, this is the deal... you wanna see pictures, go to  urlLink linecook , you wanna see writings, stay the hell here. i\'m about to converge upon san francisco, and i have to tell you, i\'m a bit... anxious about it. i can\'t friggen wait! i\'m kinda freaking out in my own little head. soon degeneration and chaos will be mine again. or just a bunch of good old times... whichever happens first. keep up, i\'ll try. stay true, you too. schmootches, FB '),(25481,4044095,'15,august,2004','2004-08-15 00:00:00','   strange things are afoot. with the insurgence of society on my quiet little world, i\'ve come to know new things, meet new people. some large and imposing but quite jovial. others smaller, quiet natured, with random outbursts of amazing absurdity. of the former, i\'m pleased as punch to associate with. the reigndeer games offer a wonderful acceptance to me. but the latter, oh my little darlings, the latter has me flabbergasted. i spend a lot of time here staring off. down unknown streets and dark alleys. up impossible hills, trying to slice through fog to what can only be utopia beyond my line of sight. but now i find myself staring at a more terrestrial being. watching with strangled eyes, squinting for fear of loosing one of these precious orbs. i babble in bursts of history, personal and world, and fidget. \"am i sweating?\" i ask myself constantly. i still wear the same pair of pants for three days straight, i still sleep close to the wall, i still wake up like a bear, grubling and growling at the first people i see as i walk to the bathroom in my invader zim boxers, and i still chew my nails incessantly. and yet, everything feels... odd. good odd. delightfully odd. standing on my desk and making fun of the teacher odd.  like i\'m winning odd . anyway, the universe is returning to my calloused fingertips, and it has bright blue hair... '),(25482,4044095,'15,august,2004','2004-08-15 00:00:00','   oh dear friends, the lexicon show was a BLAST! everything i wanted it to be and more. so fun. great time, lots of cool people, and good, good music. i\'ll be helping out at the show next month, doing door, or something, so i\'ll get to enjoy it again, immensely... on a stranger note, ren and i got dragged out to a BEAR street fair. for those of you not in the know, bears are guys who are and dig other hairy big guys. interesting times for a small dago such as myself. ah well, life is something that must be experienced. or at least watched from afar, like a zoo... noodge. '),(25483,4044095,'15,august,2004','2004-08-15 00:00:00','   yes thats right, a 30 foot long giant squid was caught or dragged up off the coast of the canary island in spain. will the sunken city of r\'lyeh be next discovered? hurricane charley? friday the 13th? are these the signs? only nyarlithotep and the fist of azathoth can be sure... okay, back to reality... got up and helped gib, my roommate, put up the stage for the lexicon show tonight. good to learn the ins and outs of the club a bit more. since i seem to have been dropped into the ensign position around there. pretty much my \"job\" is running random small errands like getting shots for the doormen and finding so-and-so for some other so-and-so to talk to... you know, like gopher on the love boat. whatever, anyway i can\'t wait for the show, its going to be really friggen cool, and ren says he\'s gonna jump up on stage in a green bunny suit for the magical mr. lexicon song \"shit tunnel\". sorry mom... still waiting for the call from the kitchen spot, but i\'m patient... smootches to all, -dom '),(25484,4044095,'15,august,2004','2004-08-15 00:00:00','   ...but it sure is handy in a knife fight. yuppers, it looks like the kid really does stay in the picture. cut my hair today. it was buzzed around the sides and back, as per the usual, and today i got a bug up my butt and cut all the stragglers on top in half. clean and ready for action. looks like a bit of video news for vox machina, in the next three weeks we\'ll be shooting around the frisky, and cameoing the crap out of it. just a bunch of the locals, some celebrity like lexicon, and the kids from the neighborhood. pre-production right now, more when it happens... look for me when it comes out. i\'ll be there. mr. lexicon show on saturday. gonna be GOOD... hopefully i won\'t be stuck in the kitchen at the Pound till after saturday... pray for me. boggely boo, -dom '),(25485,4044095,'15,august,2004','2004-08-15 00:00:00','   yo, check my comments out, dawg... yup, i am smart. in some circles... leave a comment or 9076. i don\'t mind at all. smootches. '),(25486,4044095,'15,august,2004','2004-08-15 00:00:00','   it is wednesday. where are you right now? on a bus in mpls? on the bart? san frisky is strange and new and wonderful and still strange. no doubledecker bicycles, no winter coats, no peet or o-draws or daniel or keegs. i see everything like an anthropologist/quantum physicist/new born. it\'s all new and bright, different. not scary different... wild, fascinating different. mad scientist, controlling the world thanks to the power of steam, different. i feel clear. clean. awake and alert. i\'m soaking up all of the information like some sort of monstrous behemoth sponge or hard drive. sifting, constantly sifting through it all. adapting. being aware feels really good. thank you. '),(25487,4044095,'15,august,2004','2004-08-15 00:00:00','   so here\'s the deal. it looks like i might get hooked up with a job at one of the local watering holes/nightclubs, but it will be in the kitchen. i desperately don\'t want to cook, but it\'ll be a really good chance to get in the door. working at some of these places would be rad, and there are a few of them. i would really like to get in at the door somewhere. just i.d. people and chill out, but if cooking is gonna get me there, i\'ll do it. i\'m also gonna go to borders. i got an application from them, and will send shortly. i don\'t really want to work there, but it\'ll pay the bills. and i got the skills, sorta... other than that, i had my first meeting with the other gentlemen of PARAPHAN, and it went very nicely. i knew two of the three already, and had met the other guy briefly, but never sat down with all of them. we\'re promoting a show at KEMO\'S for Vox Machina and  urlLink Mr. Lexicon  next month, it should be friggen cool. small company, small times. other than that, i have been having a great time out partying with the roommates, scaring the women folk, and generally causing a bit of havoc upon this fine california city. i love my life. noodge. '),(25488,4044095,'15,august,2004','2004-08-15 00:00:00','   well, here i am, and now i need a job. i still got a bit of scratch left, but i think it would be a good idea to get started soon. i think i have an interim thing working the door at a local bar, one of the notorios three roomies tends there, but it will be random and in between, and under the radar if you know what i\'m sayin\'... i gotta get the cash flowing, ramen does not agree with me for very long. flyered on haight for the label today, gonna sell cd\'s on market tomarrow, keep the blood pumping, and the legs moving. yup, even without a j o b things are definately looking up. for you kids in mpls, i miss you, love you, and i hope to see you soon. everybody else gotta wait in line... '),(25489,4044095,'15,august,2004','2004-08-15 00:00:00','   so when i got here i have to admit, the idea was starting to scare me. you know, the same old, \"i\'m here in a strange place, i only know like three people, and i have no idea where i am.\" thing. but as of last night, i met a crapload of really cool kats. its insane how quickly all that anxiety flips over on itself. i am a happy little dago. '),(25490,4044095,'15,august,2004','2004-08-15 00:00:00','   so there was this boy, and this boy wanted to move to San Franciso. he mailed away all of his little life in a series of 37 little boxes, and got on a plane. the plane was supposed to go to Phoenix,but the airline, in thier infinite wisdom decided not to got there. instead, they sent the boy on a much worse flight to Las Vegas. this flight was supposed to get to the airport so the boy could get on another plane, a creation that struck deadly fear in the heart of the boy, and finish the trail to San Francisco. but alas, Odin, Shiva and the almighty Jesus seemed against the boy from the beginning. he was delayed in Vegas and forced to spend the night in a plush, posh downtown hotel meeting a white guy bhuddist, an army soldier just back from Iraq, and a Pot head real estate agent from Utah. when morning came, and much fun was forgot, the boy got on another deathtrap, er, plane and flew off towards the western state to spend the rest of the day opening presents, and sleeping, which would soon become the rest of his life. the end. '),(25491,4261141,'22,August,2004','2004-08-22 00:00:00','   Well it\'s me again....of course!! (Nee-cole)      Well Friday I had five friends come over and hang out for a while. Then we all went to John\'s for the night. My dad had given me 20$ since he sponsors our team. (it\'s called Shotgun Shellz) We used the money to get junk food. YUM!! We stayed up until around 11 playing Halo. Then the next morning we were up, dressed, and out of the house by 9. We had to be in Winter Park, at Thumb Fu by 11:30. (we watched SWAT on the way there) We were there by 10:30 and the place was packed!! But we didn\'t start until around one b/c a team was coming late. There were 17 teams, each with 4 players, we have 7 but were only allowed to bring 4. We didn\'t get out of there until around 7. When I got home there was no one home so I ended up going to a friends again for the night. Today we finished the movie b/c we didn\'t have time to watch it all on the way. It was a pretty cool movie, one of my favorite action movies. Well that\'s it, there you have it sounds fun, huh? O, by the way we didn\'t do so good because the team got to pick the level and of course they picked their best, and it was best two out of three. So we didn\'t make it past the first round. The first time we kicked butt though!!! We\'ll do better next time!!!      Well, thanks for being faithful to my site!!!! Well to Mikes site but I do all the work!!! But anywho gotta run!!      This was Nicole and Mike, Love Ya\'ll!!  '),(25492,4261141,'22,August,2004','2004-08-22 00:00:00','   Hello to all and how are you? Well of course this is Nee-cole. And I am here to tell about Mike\'s day.   (Well yesterday and today)      Well, Mike and Baron had plan to come see me but things didn\'t work out. Instead they were going to go to Pinellas but that didn\'t work out either. See Carolyn and Travis told Baron they were going to get Suzie. So Baron decided he wanted to stay to see her I guess so anywho they ended up not going anywhere. Just stayed home and \"hung out\". Well, they showed up at church in these outfits that sound wierd Baron was sporting a Vintage suit that was pink and green plad and Mike was sporting a Vintage suit that was corduroy. (I think I spelled that right) Some ppl thought they looked good others thought not so good. (I would have been in the 2nd group) After church they went home (Jean\'s) and changed, then Carolyn, Suzie, Mike and Baron decided they were going to go see \"The Little Black Book\". Well Mike was paying for Baron so when they went up to the window and asked for the tickets and the chic laughed in their faces. (I would have too) After the movie walking out to the car, a bum got 50 cents off Baron. Mike stayed at Jean\'s for the night then went home the next morning at around 11 and was pretty tired b/c he didn\'t go to bed until about 3 so he took a little nap. And that\'s all folks.....Hope you enjoy knowing what he does everyday!! I bet it\'s really interesting to ya\'ll. LoL Well, Nicolina I am sorry that we couldn\'t come see you and I would have rather gone to see you, maybe sometime soon. I miss you.            Well, as interesting as his life is, my hands are tired from typing so I think I will end here... Did you like the part where they didn\'t want to come see me? I did, thought it was quite amusing. O yeah I forgot to metion that just know I got to talk to Mike\'s mom, she seems very nice. And I got to talk to my lil\' sis Alyssa, she is so sweet I wish her and Lauren could change places for a while LOL!! But anywho Aloha!!        This is Nee-cole and Mike love ya\'ll!!        P.S. Aloha means hello and goodbye!!  '),(25493,4261141,'22,August,2004','2004-08-22 00:00:00','   Well it\'s Nee-cole again it looks like I will be doing all the writing, but he will tell me happened in his day or week or whatever. So here I am being faithful to his site and he isn\'t on to tell me about his day!! But I do know this:  Today, I was suppose to go see Nee-cole but things didn\'t work out I am so upset. I was really looking forward to seeing her, maybe soon I can. I think I will go to my room and have a good cry, yeah that sounds good.  Mike is not going to allow me to do his posts much longer!! LOL But o well I guess he would have to do them then so haha!! But ya anywho....  This was Nee-cole posting for Mike!! Love ya\'ll!  '),(25494,4261141,'22,August,2004','2004-08-22 00:00:00','   Hello to all,  Well this is acually Nee-cole writing, b/c Mike (Josephine) is too lazy to write,....so I am going to tell you about his day.  Well, the day started off great Baron woke me (Well him) up, We went to Bill\'s right after that, to cut up a tree. Fun Fun Fun. After that we went to Jonathan and Holly\'s to get rid of the deck that was moved by wonderful Charley. Baron threw me into the pool, later I tried to kick a piece of wood and I fell in again. So I got to go swimming twice today, fully clothed!! Speaking of this hurricane we are out of school for a week since, there is no electricity. Okay now back to my day....After that we came home (to Jean\'s) I took a shower and then Jean took me to my house and since then have been waiting on the computer hoping to get a chance to talk to Nicole. She is just a sweet person to talk to. O but i didn\'t get to tell you the best part Baron may allow me to ride with him, to Nicole\'s. I hope i get to go b/c i would love to get to see her.  Well, I hope everybody is doing okay, and Micheal is going to kick my butt!! LOL Bye, thanks for reading   Nicole, writing for Mike  '),(25495,3382977,'30,May,2004','2004-05-30 00:00:00','   Posting a Pic for every place im going this summer!!!   First stop...New York?...well, yes, for maybe an hour or two (switching planes and crap)      Next...VIENNA!!      Then...SALZBURG      Next Stop...Munich!!!        Then i go to...Strasbourg!!(Looks boring but its NOT)      Then comes...LYON!!! YAH      YES...you know it...PARIS!!     thats the hotel im staying in!!! YAAH   Then we go to ... St. Riquier? ... doesnt seem exciting, but whatever      Next stop...COLOGNE....YAH...      Then i go to BERLIN!      I will then travel to...MAINZ?...whats that??hmmm....     Wow..looks exciting, doesnt it...   Anyway...next is AMSTERDAM!!!      YEP...thats my Euro-Tour!!! But right after im gonna be at sunny Florida, YAH...its gonna be one hall of a Summer...cant wait till these darn Fianls are over with...   '),(25496,3382977,'27,May,2004','2004-05-27 00:00:00','   My name is Erik, and im joining my brothers blog just because he let me....interesting isnt it....Yah, it is, so shut up   '),(25497,3382977,'27,May,2004','2004-05-27 00:00:00','   well  i hated this day alot  cuz  well  i got up same as normal  got ready for skool  and went and we had a feild trip today  and it suked balls    there was nothing good about it infact it was the worst feild trip iv ever been 2  wel  some of it was ok  like the sand fights and the beating up alexander and tye part was fun    but all the rest suked  well  so we go on bus 2 ocean beach  it suked   but it was funny cuz tye kept tagging the back of the bus  and then he did it on the camera  in the bus    stpid    well   we went and got there and made our assigned kindergarten make this stupid kite  and well    it didnt work so we  just said screw it  and after just hangin out and screwin around   tye and kyle and kenedel and i had this brilliant idea   why dont we ask 2 c the other ppls kites and then we can just le go and c wut happens  well   no1 would let me hold them  the only person that got 2 hold the kites was kyle  and he was holding like 4 frikin kites (dont ask me how) but he just let go of all of them at the same time   god   it was hella funny   i was hella laughing   well  after that tye and i went to jump off of the wall at ocean beach that was fun    but then after like 10 min   he had 2 go and eat lunck  and so we did and brought larwence with us and got him a hot dog and  i guess he liked it  but  oh well  screw it  i could care less if he did or didnt  but anyways  after that tye and i got kendel 2 jump with us   but not david or alexander  or kyle  cuz kyle is a stupid ass  well  any ways after we were done with that we had decided 2 go and chill on the ramp and so we did  and tye kept doin stupid shit  and so  i pulled his sweatshirt string all the way and well   like it curled all the way up on him( cuz he was wering his hood) well   like  he got hella pissed  even though he was tryin 2 push ppl pf pf the wall  well  i guess that then we all just started pushing eachother off the wall  then david and i pushed kendel off of the wall and stole his pants(he had shots unger them and david realy liked them) and then tye put sand all over him and then the huge sand war startedwell  liek  kendel and i ganged up on alexander  the  tye joined us  then tye told kyle to put sand down my shirt  and then i got kyle i dropped him then i put wet sany and dry sand all over him  then i did the same thing 2 tye but i put the dry sany all over him and then pured water on him  and then i got kendel and then kendel and i ganged up on alexander and we did hella stuff 2 him...well  then we went and tye agian we went 2 the back of the bus and he was writtin hella stuff and left the camera alone this time  a little bit smarter than b4  well  then i came home   went out  put stuff on my face cuz i got sun burned and came back home and started 2 do me hw and now i am gonna finish a report that us due 2m well  bye   '),(25498,3382977,'26,May,2004','2004-05-26 00:00:00','   hey  this is robert  yea  well  almost all of u pl kno me   or  i guess that i wouldnt give it 2 u  well   iono  wut 2 write  well  i guess that my day went well 2dai   we had the stupid play practices and some ppl were just bein plain dumb well  like in da mornin i woke up same old same old went 2 skool  and well   had sumthin   dunt remember  but  well   i just remember everything was just really really dumb !!!!!!! and tye  u r the dumbest person i kno  alexander  u just look like a retard  and  well   mr. rogers isnt kool  !!!!!!and no alexander  i wont     urlLink   '),(25499,3382977,'05,June,2004','2004-06-05 00:00:00','   hello   hello....hello   '),(25500,3382977,'02,June,2004','2004-06-02 00:00:00','   2day was the last full day of skool   yea well  i got an a on the history test   and well  i was really happy   we didnt really do that much well   we didnt do anything we just hung out well i guess i havnt posted in a while and well  ima just skip 2 2day cuz alot of stuff has happend scince then and it would take me a long time 2 explain all of it and welll  i dont have that much time   and well  2day i woke up a little later than normal cuz erik doesnt have any skool 2day and well so we left for skool later and it was pretty fun and then 4 all yall that probably kno wut im talkin bout the answer was okay     she said yes  !!!!!!!!! well anyway   besides that then we went 2 skool and  wasnt early and stuff like normal and so i just went straight 2 class wich wasnt really class it was just nothing and well   we signed year books and stuff   and tye had 2 do all of the finals that he didnt do at yesterday cuz he didnt want 2 go 2 skool and stuff and well   yea  and so i was fun yesterday without him (cuz it didnt stink wen we were around him....cuz he wasnt there!!!!) and yea    well   it was pretty fun today also but he was only gone most of the time   but his smelly self was there at lunch but that didnt matter but anyway then we went back 2 class   same as normal    nothing just boredom talkin bout sumstupid toiled and stuff and like i guess that it was ok   i guess   but then we went sumwhere else and that was another class !!!!!   but it was muchas more fun   but it was with the proncipal and i guess that it was ok and    well         it was fun cuz tye had 2 take the history final and stuff and like so he wasnt there and we got 2 just like hang around and do nothing   but accually talk and stuff unlike the other ones we either listend 2 sum1 else talk bout sumthiun stoopid or we just sat there and read......but then we had 2 go back 2 do  nothing and listen 2 sum stupid teacher talk bout toilets cuz we got the same techer  man that makes stuff really boreing and stuff   and well then e were excused and cleaned out our desks and stuff and our lockers  we then we went home and well i was bored and went down 2 my neighbors house and i asked bout the stupid finals that she was talkin bout and she said stuff like she was reviewing and was getting all mad at me and stuff and so i told her that i didnt understand about that cuz i just alwas was able 2 memorize history really easily and stuff and she got even more pissed then she arty was told me 2 get out and i was kinda confused then i understoodand i asked why and god i widhed i never asked she got furiouse like crazy animal and starded bouncing off the walls telling me that she didnt need any1 telling her that  they are better than her  and well  i was tryin 2 tell her that i wasnt sayin that i was better than her i just didnt understand wut she was doin and stuff but well i got kicked out ....and if she thought that that brightend my day at all......well  it didnt, if u read this well   anyway wi went home being bored as normal and did nothing went outside a little bit and now i am on the comp typing this and doin nothing but tye and ima go now         peace out   '),(25501,3382977,'01,June,2004','2004-06-01 00:00:00','   Why am i so happy...well, let me make this quick...  Finals are over  School is pretty much out (except for one more day of handing out prizes)  Fianlly got my Yearbook  Got an insane summer coming up  Grandparents are coming this Friday  No more damn stupid conservatory forever (Yes...FOREVER)  and, well....think of something good..hmmmm  oh yah      Just a reminder....   '),(25502,3382977,'18,July,2004','2004-07-18 00:00:00','   well, not exactly back to SF, but back as in out of Europe. I had a good tour, lots of fun, and i even met some new friends. Some of you who know what im talking about might know Mrs. Smirfy also. I met Dr. Jacky but didnt like him very much. Mrs. Bicardinia is doing well, but i still didnt see her as nice and sweet as Mrs. Smirfy. anyway, enough of my new found friends. I had a bunch of fun in Paris and Amsterdam, which were both great and wonderful citys. Amsterdam was the most interesting and crazy. Paris was beatiful. the plane rides were ok, i watched a bunch of movies and stuff...nothing big. anyway, i am now in Florida, and im having a bunch of fun. soon to be home. post later.   '),(25503,3382977,'30,May,2004','2004-05-30 00:00:00','   Im Katarina, the sister of these two boys...im gonna be posting here to as well as in my own blog...I\'ll be posting about everyday activities and stuff in general.   '),(25504,3382977,'29,May,2004','2004-05-29 00:00:00','   well  today was a very vaery interesting day   lily erik and i took cracker, kooch, and boomer up 2 mt davidson and like lily lost her very expensive glasses so we went back and looked for them  and was offering first rive then ten but at the end 50 $$  and well no1 found them and well  so then after that we took the bus around the place and then went around and didn nothing and then erik told lily that he lost the transfer and well  we were all like tryin 2 c how he would get on the bus and that he might have 2 walk cuz we didnt have any change or cash  and like so he was lying and stuff well  but anyway it was pretty funny and then we got home and my mom as all mad and stuff ad well  i guess that i just ignored her and went 2 eat al lilies house and we had pizza ans i huess that my mom was ok after that but then she got mad agian and stuffa nd i was just like screw it and wel  who canre and now i am just gonna go 2 bet cuz im hella tired!!!!!   '),(25505,3382977,'28,May,2004','2004-05-28 00:00:00','   today was one hell of a waste of time...wow    ok, so i woke up today and i actuallly felt good since it was like 9:30 ( i actualy got to sleep in today ). I had a quick breakfast and then Lily, Robert and I decided to go take a \"mega-walk\"...which is a huge long walk with the dogs ( Koosh, Krack, and Boomer ( if you care )).   so we go up to the top of mount Davidson and hang out there for a while, come back down, and then go home and stuff...have some lunch, and then we decide to go back up since lily lost her glasses ( these really expensive nice sunglasses ) ... so without the dogs we go up and check it all out...and then Lily decides to go on the bus for no reason at all ( she claims it is relaxing )...so we check out the 43 bus and start riding it....wow, that was the most relaxing ride of my life...are you kidding, the bus is the worst place to ride around for 2 hours for no reason....well, we get to lilys house, have a pizza, and now im here with a huge head ache and very tired...ho-hum...   '),(25506,3382977,'28,May,2004','2004-05-28 00:00:00','   well....another day of yucky finals , Geometry and English...   English went well and Geometry was horrible...   enough about finals, how about this summer...oh yah...im going to Europe and then to Florida!   '),(25507,3382977,'27,May,2004','2004-05-27 00:00:00','   hey well my day suked  well  we had 2 take our kindergarfen \"buddies\" to sum stupid a** ocean beach and well  i guess it suked   i got all sunburned on my face  and well  i came home and now i have a headach   godd***  this day really suks   '),(25508,3382977,'26,May,2004','2004-05-26 00:00:00','   Today was the first day of those darn brutal finals!!! i had a normal sized Bio test that was pretty much easy except for the first section ( i dont think it will do much to effect my grade since we figured out that in order for the top student in the class to have his 2nd semester grade affected he would have to get at most a 44%...and he has a 98% in the class...so I ( having about a 95% ) wont have to worry about a grade change...even if i do get something like a B ( which i am expecting )).    The Religion Exam, on the other hand, was the longest thing i have ever done...ok, i was going along with the first page of pretty much simple straight forward multiple choice questions...and im thinking this is a breeze and that im gonna get this done in 20 minutes max...but them i turn the page and POW...there are like 20 vocab words from Islam, Judaism and Christianity ( most i knew well ) and about 15 short answers ( which the teacher usually likes a couple sentences for each )...and on top of that we were supposed to write an essay on comparing two of the three religions...so there was a crap load to be done in an hour and a half.    estimated grades for todays finals ... Bio=85% min. and Religion=90% min.    wasnt too bad....but now i gotta study for my english and geometry coming up tomorrow...   '),(25509,3382977,'26,May,2004','2004-05-26 00:00:00','   hey  im eriks lil bro robert and i ll b posting here wth erik   '),(25510,3382977,'26,May,2004','2004-05-26 00:00:00','   Hi, im Erik and i will be co-posting with robert who is the other partner in this dual action blogger!!! Yah   should be a lot of FUN   '),(25511,3382977,'26,May,2004','2004-05-26 00:00:00','   Yep, this is gonna be a whole bunch of stuff ... mostly daily stuff and just anything random or not random that we feel like posting...so check it out!!!   '),(25512,3382977,'16,June,2004','2004-06-16 00:00:00','   hi, I\'m the sis of Rob and Erik !They like to joke a whole lot.Ex. Today I was scared that I would have to take a blood test and was talking about it with my Grandma. Then Erik walked in and said \"Oh yea blood tests. I took one  and all you feel is a needle smashing around your musules and baning everything ,but besides that its fine!\".I got so0o freaked ,then my Grandma said he was lying. I\'m still really freaked out.   '),(25513,3382977,'07,June,2004','2004-06-07 00:00:00','   \"The Flames, threatening to become one of the most unlikely champions in NHL history, were held to only seven shots in a dismal first two periods before making a frantic late surge started by Craig Conroy\'s power-play goal midway through the third.   Ruslan Fedotenko scored twice, including the critical first goal just as he did in the conference finals, and the resilient Lightning held off the Calgary Flames 2-1 in Game 7 Monday night to win their first Stanley Cup.\"   first the sharks are out..and i loose hope in them, and then my good ole flames loose...oh well      its a beauty   '),(25514,3382977,'05,June,2004','2004-06-05 00:00:00','   Harry Potter and the Prisoner of Azkaban is a movie i would give about an 8 (of 10). why? well you can definetly see and feel the difference between Alfonso Cuaron\'s directing in this movie and the  Chris Columbus\' directing in the others. Most people said that #3 was better than the first two, but i think that they are about the same, or even better in some ways. The first two would switch from scene to scene with not many breaks, which kept the watcher on the edge of the seat, while the third has these funny fade outs and interesting endings of the scene ( mostly with Harry Potter fainting or passing out to make it seem like its supposed to be like that ).   There were also a lot of things put in that made the movie seem a lot different than the book\'s idea of how it was supposed to go. The main  plots were there but somethings ( like the music that Lupin played ) gave the movie a weird feeling. anyway, you would have to see it for yourself to see what i really mean.  I would go see it. It is a fun and exciting movie. 8 out of 10 is what i would give it.   '),(25515,3382977,'01,August,2004','2004-08-01 00:00:00','   ~$155,000  ~$130,000'),(25516,3382977,'30,May,2004','2004-05-30 00:00:00','   Posting a Pic for every place im going this summer!!!  First stop...New York?...well, yes, for maybe an hour or two (switching planes and crap)   Next...VIENNA!!   Then...SALZBURG   Next Stop...Munich!!!    Then i go to...Strasbourg!!(Looks boring but its NOT)   Then comes...LYON!!! YAH   YES...you know it...PARIS!!  thats the hotel im staying in!!! YAAH  Then we go to ... St. Riquier? ... doesnt seem exciting, but whatever   Next stop...COLOGNE....YAH...   Then i go to BERLIN!   I will then travel to...MAINZ?...whats that??hmmm....  Wow..looks exciting, doesnt it...  Anyway...next is AMSTERDAM!!!   YEP...thats my Euro-Tour!!! But right after im gonna be at sunny Florida, YAH...its gonna be one hell of a Summer...cant wait till these darn Fianls are over with...  '),(25517,3382977,'29,May,2004','2004-05-29 00:00:00','   Today was an extreme waste of time...at least i got to sleep in ( till 9:30 )...but besides that the day sucked...  firstly, today i have walked up to the cross on Mt. Davidson twice...yes twice, why?...well...the first time was becuase of this \"mego-walk\" dealy which i really wanted to do with the dogs ( from down the street ), and so we went up once, then we went up again becuase one of my buddies lost their glasses ( very expensive sun glasses ) and so we went to look for them ... then the rest of the day consisted of a stupid bus ride and soe pizza and some homework and some sleeping since i have an insane headache right now..,  on the bright side of things, as you know ( if you read my profile ) i like to collect classical cds of my own ( my family as a whole has over 400 cds in the dining room but they are all either old, in bad shape, or are just stupid...so i decided that once i got my new room i was to start my own private collection of the ones i like and have...and it has grown to almost 75 ( started it sometime in March )...so i think just for no reason im gonna type out all the cds i have as of now...and ill put after the title how much i like it ( on a scale of 1-10 )  itll be like this=Composer and Piece ( unless there are tons of them, then ill just write the title ), Performer, Label, score   1. Beethoven: Violin Sonatas, Menuhin (Kenter (Pianist)), EMI (3 discs ),   8 2. Rodrigo: Concierto de Aranjuez, Barrueco (Domingo (Conductor), EMI, 6 3. Veuxtemps Ravel Saint-Saens, Perlman (Barenboim and Martinon (Conductors), EMI, 8 4. Paganini: 24 Caprices, Perlman, EMI, 10 5. Duos for Violin & Cello, Kennedy and Harrel, EMI, 8 6. Grieg & Schumann: Piano Concertos, Richter, EMI, 9 7. Holst: The Planets etc., Boult, EMI, 7 8. Schubert: Symphony #9 & Rosamunde Overture, Vienna Phil. ( Muti (Conductor)), Seraphim, 10 9. Paganini Vln. Concerto #1 & Lalo: Symphonie Espagnole, Kogan, Testament, 9 10. Beethoven & Mozart Vln. Concertos ( Mozart=#5),Kogan, Testament, 8 11. Bach Vivaldi Mozart: Vln. Concertos, Kogan, Testament, 8 12. Brahms & Mendelssohn Vln. Concertos (Mendelssohn=#1), Kogan, Testament, 8 13. Orff: Carmina Burana, SF Symphony & Chorus (Blomstedt), London, 10 14. Sibelius: Sym. 3 & 6, SF Symphony (Blomstedt), London, 10 15. Williams: Fantasia on a Theme etc., Marriner, Argo, 6 16. Brahms & Bruch Vln. Concertos, Zukerman (Mehta), RCA & BMG, 9 17. Holst: The Planets & Williams: Fantasias, Slatkin, RCA & BMG, 8 18. Vivaldi: The Four Seasons & other concertos, Moscow Virtuosi (Spivakov), RCA & BMG, 9 19. Dvorak: Symph. #7 & #9, Chicago Symph. Orch. (Levine), RCA & BMG, 8 20. Ravel: orchestral music, Dallas Symph. Orch. (Mata), RCA & BMG, 8 21. Gershwin: orchestral works, Earl Wild (Piano), RCA & BMG, 9 22. Beethoven: Symph #9, Wand, RCA & BMG, 7 23. Famous Overtures, Various Artists, Naxos, 6 24. Baroque Violin Favourites, Huang (Vln.) & Kim (Piano), Naxos, 5 25. The Best of Brahms, Various Artists, Naxos, 7 26. Tchaikovsky & Dvorak: Serenade for Strings, Vienna Chamber Orch. (Istropolitana), Naxos, 7 27. Wieniawski: Violin Showpieces, Bisengaliev (Vln.) & Lenehan (Piano), Naxos, 8 28. Vieuxtemps: Vln. Conc. #5 6 7, Keylin (Vln.), Naxos, 8 29. Bruckner: Symph. #3, Tintner, Naxos, 8 30. Bruckner: Symph. #4, Tintner, Naxos, 9 31. Bruckner: Symph. #8 & #0, Tintner, Naxos (2 discs), 9 32. Bruckner: Symph. #9, Wildner, Naxos (2 discs), 10 33. Brahms: Four Hand Piano Music Vol. 9, Matthies & Kohn, Naxos, 8 34. Paris La Belle Epoque ( Faure, Franck, Massenet, Saint-Saens), Yo-Yo Ma (Stott (Piano)), Sony Classical, 9 35. My First 79 Years, Isaac Stern, Sony Classical, 10 36. Simply Baroque, Yo-Yo Ma, Sony Classical, 9 37. Mendelssohn & Beethoven Vln. Concertos, Bell, Sony Classical, 9 38. Beethoven: Piano Sonata #s 14 23 & 8, Ashkenazy, Universal Classics, 7 39. Beethoven: Piano Conc. #5 & Triple Conc., Stern (Vln.) Rose (Cello) Istomin (Piano), Essential Classics, 10 40. Gallery of Classical Music: Romantic Interludes, London Symph. Orch.,     Platinum, 7 41. The Instruments of Classical Music: The Violin, Szenthelyi & Kote, LaserLight Digital, 6 42. Let\'s Dance, Dubeau & Pieta, Analekta, 8 43. Ballo Ongaro, Herencsar, Hungaroton Classics, 5 44. Classical Favorites for Strings: Songs Romances & Dances, Budapest Strings, LaserLight Digital, 6 45. Music of Ravel, Cincinnati Symph. Orch. (Paavo Jarvi), Telarc, 10 46. Ballets Russes, Mariinsky Choir & Orchestre Phil. de Radio France (Paavo Jarvi), Virgin Classics, 10 47. Sibelius Cantatas, Estonian National Symph. Orch. (Paavo Jarvi), Virgin Classics, 10 48. Paganini for Two: Shaham & Sollscher, Deutsche Grammophon, 9 49. Great Works for Cello and Orch., Rostropovich, Deutsche Grammophon (2 discs), 8 50. Bruch: Conc. for Clarinet & Viola etc., Lyon Opera Orchstra (Nagano), Apex, 9 51. Shostakovich: Piano Conc. #s 1&2, Leonskaja (Piano) & The Saint Paul Chanber Orchestra (Wolff), Apex, 10 52. Dvorak  Liszt & Brahms: Slavonic Dances  Hungarian Rhapsodies & Hungarian Dances (respectively), Various Artists, Brilliant Classics (3 discs), 8 53. Dvorak in Prague, Yo-Yo Ma (cello) Perlman (violin) Firkusny (Piano) Stade (Mezzo-Soprano), Sony Classical, 8 54. Mussorgsky & Wagner, Stokowski, Cala, 7 55. Ravel: Daphnis et Chloe, Boston Symph. (Munch), RCA Victor, 9 56. Classics of Spain: Rodrigo Albeniz Chabrier Ravel & Bizet, Various Artists, Golden Classics (2 discs), 9 57. Berlioz Overtures, Boston Symph. (Munch), RCA Victor, 10 58. Various Artists, Kissin (Piano), RCA Victor, 8 59. Virgen de la Caridad, Anthony Blea, Hopping Mad, ? 60. J.S. Bach: Brandenburg & Violin Concertos, Orchstra of the Age of Enlightenment, Virgin Classics (4 discs), 10   well, thats my total cd collection as of 5/29/04....i might update the whole list again next month or something...who knows  and...if you look through the list, there are several discs that had so much crap on them that i decided to be lazy and write something along the lines of \"various Blah blah blah\"...so if you are really interested and really need to know what was on that cd then you can always e-mail me ... if your that desperate ofcourse ( i dont mind )...(:   OH...and if you have any personal favorites or any other classical cds you would like to recommend please do so ( e-mail me or post a comment ... make sure to tell me the name of the cd along with something else (like the composer or artist or label))....THANKZ  '),(25518,3382977,'28,May,2004','2004-05-28 00:00:00','   Yes......has been a loonnngg time since i have had a three day weekend....and, this is a very special three day weekend... NO  Conservatory ( which means i dont have to get up at some stupid hour (usually 6:30) and then get to the place by 8 ),  NO  YO rehearsal ( that means no going to Davies at 1 and practicing till 5 ), and ofcourse i have NO school on Monday....which makes this weekend the best ever. And My granparents are coming next friday so i\'m very pleased.  thing im not looking forward too=studying my butt off for the history and latin I exam coming up this teusday , but it wont be that bad   '),(25519,3382977,'28,May,2004','2004-05-28 00:00:00','   urlLink    This is the YO Repertoire for the Euro Tour.  '),(25520,3382977,'26,May,2004','2004-05-26 00:00:00','   well...i got this smacking huge packet from the YO today and it had a whole bunch of junk that needed to be signed so i did all that, and then theres this paper that has all the songs we are playing in the tour...and guess what, the two songs i was looking forward to (  Le Corsaire  Overture and the Grieg Piano Concerto in A minor ) i wont be playing...  why? well, first of all, im the last stand of the second violins ( pathetic...i know, but there are some pluses for sitting in the back ) and so the Grieg ( since its a soloist playing ) will need a \"reduced strings\" thingy...which pretty much means that everyone thats in the back of their section ( in the strings only ) will have to get outta there when the certain song is played...  now, the deal with the  Le Corsaire  Overture i have no idea about...they just decided not to play it, which sucks, since i really wanted to play it and its a reealllyy fun piece...oh well.  i guess i\'ll post the paper with all the repertoire for every day of the tour on it soon...dont feel like typing it all out....but these are the pieces we are gonna play and how much i like them ( scale of 1 - 10 ):  Bruch / Violin Concerto No. 1 in G minor, Op. 26 (7) Grieg / Piano Concerto in A minor, Op. 16 (9) Ravel / Suite No. 2 from  Daphnis et Chloe  (6) Tan Dun /  Orchestral Theatre I: Xun  (3) Tchaikovsky / Symphony No. 4 in F minor, Op. 36 (7) Copland / \"Hoedown\" from  Rodeo  (8) Gershwin /  Promenade  (?) Tchaikovsky / Pas de deux from  The Nutcracker , Op. 71 (?)  '),(25521,3382977,'26,May,2004','2004-05-26 00:00:00','   hey....  Robert and I ( robert being my younger brother ) are gonna make a team blogger just for random stuff and mostly daily stuff and how things are going and stuff....so, the name of it is gonna be eriknrobsblog  thats about it!! did i mention that my finals have been going ok?...well they have been    '),(25522,3382977,'25,May,2004','2004-05-25 00:00:00','   The two pictures below are the itinerary for the Euro Tour....and you can read them by clicking on the picture. After clicking on the picture it goes to another window with it on it ( somewhat enlarged )....but i know...its still kinda hard to read  Well, it might seem like a no brainer to you computer guys out there, but if you go to the bottom right hand corner of the paper after you have clicked it and the new window has popped up, there is an enlrage button ( yes, an enlarge button )...well, click it and there yah go (:  '),(25523,3382977,'25,May,2004','2004-05-25 00:00:00','   urlLink    This is the 2nd page of the Very Detailed YO Itinerary for the Euro-Tour. Check it out.  '),(25524,3382977,'25,May,2004','2004-05-25 00:00:00','   urlLink    This is the Very Detailed Itinerary for the Euro Tour. I has the rehearsal times and concert times and everything else one wouldn need to know. This is page 1 of 2.  '),(25525,3382977,'25,May,2004','2004-05-25 00:00:00','   well...i got those darn finals coming...  i dont know why im posting this but i just feel like doing something...so here it is My Darn Final schedule: Thursday includes my Biology Final Exam:    Then I have my dreaded World Religions Exam!!!     And the last exam on Thursday is my great art exam!!! will might actually be fun, we get to go to the Museum of Modern art and check out some cool stuff and then just have to write some stuff, which will be very easy ( barely have to study ) Yay!     Next day of Finals....FRIDAY  I will have my English Final first    ...And then my crazy Geometry Exam ( its not too bad )     AND a 3 day weekend YYEEESSS!!!!!  ...and then disturbed by another batch of finals....  AAHH.,....My Latin I Exam ( tons of words and grammar and stuff to learn )    The comes my World Histroy Exam ( cant be too hard ... its the easiest teacher in the school )      '),(25526,3382977,'24,May,2004','2004-05-24 00:00:00','   well...my birthday was    May 7th   ...so im kinda late...but still, im gonna say how great it is just for fun ( and to have a new post )   well...i got a lot of cool stuff, but im still searching for a digital camera to take on tour and to take lots of god pictures with...so far my dad and i were thinking of this cool one, tell me what you think     pretty cool eh?  hopefully my dad will overcome to stress it will have on his wallet I:  also...i got a really cool pair of cufflinks from my Uncle and Aunt...they are very cool  I also got a pen and pencil...that are really nice from my sister  I got a game ( Splinter Cell: Pandora Tomorrow )  yep...looks cool to me ( i\'m trying out this whole picture in the post thing)  The only problem is that my laptop isn\'t good enough for it...so i mite as well go and play it on the main computer ):  I also got an airline ticket to FL...so my summer is gonna be one fun time ( see itinerary below )          '),(25527,3382977,'20,May,2004','2004-05-20 00:00:00','   So...check out this newly updated schedule for my summer break.   June 4th - Schools out!!!   YO Rehearsals. 6/4/04, 6/5/04, 6/11/04, 6/12/04, 6/17/04, 6/18/04, 6/19/04   June 19th - Bon Voyage Concert at 2 pm   YO European Tour- June 29th to July 16th, 2004   I then come back from tour and go straight to Florida the next day! Yay...   Depart from SFO on July 17th , 8:50 AM to NC   Depart from NC at 4:46 PM and arrive at West Palm Beach Int\'l at 8:19 PM   I come back on the 29th of July....    and then im free  '),(25528,3382977,'18,May,2004','2004-05-18 00:00:00','   ok....lots of people ( mostly relatives ) have been wanting my itinerary for the summer Euro tour so im just gonna post it here since my other site is still kinda pathetic and useless ( almost )...if your interested in seeing how pathetic is is go to www.freewebz.com/erikpolus....   ok, here it is....  Saturday, June 19           Bon Voyage Concert- SF Davies Symphony Hall  Monday, June 28             Tour begins/ depart from SFO  Teusday, June 29            Arrival at Vienna, AUSTRIA  Thursday, July 1            Concert Performance #1-Vienna. Musikverein  Sunday, July 4              Concert Performance #2-TBD  Teusday, July 6             Concert Performance #3-Lyon. \"Les Nuits de Fourviere\" Festival, Grand Theatre  Wednesday, July 7           Concert Performance #4-Paris. Theatre des Champs-Elysees  Thursday, July 8            Free Day in Paris  Friday, July 9              Concert Performance #5-St. Riquier.  Monday, July 12             Concert Performance #6-Berlin. Haus des Rundfunks  Teusday, July 13            Concert Performance #7-Kloster Elerbach. \"Rheingau Musik Festival\"  Thursday, July 15           Concert Performance #8-Amsterdam. Concertgebouw  Friday, July 16             Return to San Francisco    If you have any questions on any of these things or want more details just e-mail me....   erikpolus@sbcglobal.net  '),(25529,3382977,'18,May,2004','2004-05-18 00:00:00','   Hey, just checking out how all this blogger stuff works...seems pretty cool   '),(25530,3382977,'27,June,2004','2004-06-27 00:00:00','   hmmm...i know im missing something, and im worried cause my suitcase is exactly the size maximum and it is just on the weight maximum, so i hope there are no problems at the airport...and i still gotta pack my carry on backpack, theres so much to worry about, i gotta pack the bag that im taking to FL when i get back, hopefully i dont run out of batteries for everything.  I went to See\'s Candies and a hair saloon store to get candies (a pound of them) and some nice gel. Then i got back and watched some Reno 911, then i packed, then we went to west portal to get some food for the smores tonight, and there was a huge line to Farhenheit 9/11, and i really want to see it, hopefully itll still be in theatres when i get back. so, we had a campfire thingy again, and then we played football till about 9, then i was tired and started packing more. Still have a lot to do. very excited about the trip. you can always go to http://www.sfsymphony.org/templates/youth.asp?nodeid=82 and find your way around, there will be a person posting news and updates during the tour, with pics. The link to the tour pages will become active after we depart. sooo, if i have the chance i might post something up, but there wont be any pictures until i get back, since i dont think there will be computers available with the plug ins and all that stuff. I definetly will post smoe really good pics that i take from Florida and defiently the Euro tour. post soon  '),(25531,3382977,'26,June,2004','2004-06-26 00:00:00','   ok, so today was a somewhat fun and somewhat tiring day. First i went to YO rehearsal at 10:30, but found out once again that i didnt actually have to play for the first 45 minutes of rehearsal since i dont play any of the solo concertos. and guess what, i didnt play for a good 2 hours of the rehearsal today. I also got a 2 hour lunch (while the others got 45 minutes). so, in some ways its a good thing. i also dont like the Bruch that much anyway, so no loss there. BUT, i am hoping to move up a couple stands for next year since we are playing some really cool solo pieces (and the concertmaster (whose rreeaalllyy good (and only 12) is playing the solo for it). Got the letter saying what pieces are up for next year, and they are really fun (except for one whole concert which is full of ok pieces, no big ones or anything.   During the huge lunch break i went to the Flight 001 store, which i have to say is one of the coolest stores i have ever been to. It is by far the best travel store i have ever gone to too. So i walk in and they have ten times more things than i thought (from looking at the stuff online). I was there for many things, but my mom wanted me to get first and foremost one of those things that hang under your shirt from your neck, which you can keep your passport and money and travelers checks etc. in. So i found one of those, and then i checked out the alarm clocks...and theyhad some reeally cool ones, but my favorite was this alarm clock which is a small cube (about an inch and a half in height and lenght and width). It has the temperature, date, alarm, and time on four sides with a light and other cool things...    anyway...after the rehearsal i took the Muni to Waldenbooks where i used up my gift card...and then my mom came and got some other books...got my three summer reading books (The Hobbit, Into Thin Air, and The Things They Carry) and i got another spiderman comic collection...and i then went to the camera store on west portal, and got a nice camera bag (which just fits my digital camera if i sqeeze it in). so now im set. just gotta pack the extra suitcase thats going to florida after i come home. and i gotta finish with the euro packing too. got my travelers checks which add up to 200 euros and 100 euros in notes, and then i have my credit card so i think im set in the money section...shouldnt spend too much though, cause its coming out of my own bank account so im not too happy about that.   tomorrow will be a day full of playing games and stuff since my violin is packed up, i have nothing to do...except pack.  '),(25532,3382977,'25,June,2004','2004-06-25 00:00:00','   soo...everything is ready except for my camera bag, which i gottas buy. I also need a pair of sunglasses (or do i reeally need them...i guess not), and i gotta get that converter crap, and i gotta get the darn travel alarm clock...which is all to be done tomorrow...which i just remembered have a humongous YO rehearsal from 10:30 to 4:30....which is the time afterwards where we are supposed to pack our instruments up in the big cases...and then they are shipped off...not to see my precious violin for a good what, two days...boo hoo. anyway, after the packing i gotta go to the Flight 001 store and see if i really want to get anything, and if not mom will just give me 50 bucks and ill give her the gift card, since my dad might like something from their, or maybe even my mom.   soo...you wanna know what messed up about my laptop, that is stopping me from bringing it to euro and to FL, well. its just plain f***ed up. i have tracked this messed up ness to the cooling system, which isnt working, which is causeing the computer to just shut off when the comp gets too hot. i cant bring it to euro because it especially gets hot when it is playing something in the disc drive, which will make it shut off in the middle of a DVD, which was its main purpose there. so, the problem: it is overheating thanks to the messed up fan and coooling system. so its toastm the waranty is over with, and the people that try to help you arent that great. so, i was thinking of selling it and then using that money, along with some of my own, to buy a better and improved one. Also, my laptop wont play some new games that i buy (ie: Splinter Cell, Pandora Tomorrow), which is a drag. My 256 mb card came today, for my Digital Camera, but im not sure how many pictures it can actually hold, so im gonna have to do some experiment pictures sunday, which my mom says will be the day for partying and stuff...so she actually said i could play videogames...amazing. Some friends are coming over to play and watch hella movies...not sure which ones im gonna rent. hmmmm   so, today i woke up earlier than usual (like 9 ish) and had to take my grandparents to the airport...too bad they had to go...but who cares, im gonna see them in 2.5 weeks, so i might as well not get tooo used to them. after that i chilled at home for a bit, watched some more Reno 911, and then went to get my hair done. I had to walk errol (my littlest brother and sibling) down there at 1, while mine started at 1:30. and mr mom wanted me to put some money in the meter for the car\'s parking spot, so i had to break my 20, so i went to Ambassado Toys (dont ask) and got a tiny puny little deck of cards, and two rocks, which you are suppsed to break open in half with a hammer, and then if you do it right you get one of those cool crystrals-inside-a-rock things. so i put in 12 minutes worth of change into the meter and got my haircut, which isnt too bad, unlike the other time i went to Supercuts, and some barely0able0to-speak-english girl who was probably an illegal immigrant cut just about all my hair off (this was like March, so no big deal). then i went home and started to pack and do all that stuff...  '),(25533,3382977,'24,June,2004','2004-06-24 00:00:00','   ok, my days have been getting busier and busier...and i have more, and more to do, since i got the SFYO Tour coming in ... 4 DAYS!!! yah, so i have everything done on the checklist except for the converter crap...why cant europe just do what we do here, i mean, its just an outlet. So gotta go to Radio Shack and get the two thingys. I also have to get a camera bag for my new camera...which is a really weird shape, so i gotta get a good size and good fit one. I got my visor, i got a suitcase, which is exactly 62 linear inches, which is the exact limit for a bag, now the problem will be keeping it under 50 punds or something like that. I got more socks...and i got a jacket finally, and i got my new \"kicks\" and all my new shorts and i still havent used my $50 gift certificate to Flight 001...darn, they have some cool stuff but im not sure about what to get...its all so overpriced. anyway  Today i got up and did the usual. I watched some more of my great Reno 911 the whole first season, which i got yesterday at Target, and that reminds me, i still havent gotten the LOTR 3 on DVD...well, maybe after the tour. I also got some new swimming trunks that actually fit. the only thing that i dont get is...it says the whole first season, and it only has episodes 101 to 108 on the first disc and episodes 109 to 114 on the second, including the uncesored bonus stuff...weird. so, after that i went to what is now my least favorite store...ROSS: dress for less....it sucks...the place is a wreck, its for cheap-ass loosers and i hate it...its soo cheap, its where all the stuff that no one bought at the real stores goes and the price is cut in half. Why do i go there...i dont know. I would think that i was a bit better than that (no offense to the people that shop there regularly). ok, then i went to Baskin Robins, where i had a nice huge great icecream, and then i had a darn good violin lesson, which btw is the last violin lesson i will have for 2 whole great months!!! yippay. actually only grom June 28th to August 1st, but still. Then i came home and rushed the two rental videos (Dickie Roberts and Malibus Most Wanted (which were both ok but not the best)) on my bike back to the store, and then rushed back and had dinner, which was kinda weak, just a bit of lamb, and then i went outside and played on errol\'s jeep for about half an hour, taking it on the most rugged courses it has ever been on, and then i was caught by robert with the battery dead, but he didnt care, he just was making sure i got my swimsuit so that we could go swimming, and we did, with errol and my grandfather, which was really fun. The rec swim lasted from 8:30 to 10, and we left at about 9:45. Good excersize. sooo...im still waiting for that darn memory card to come, Amazon.com said it would be here by the 25th, and i checked earlier on the status and it hadnt been shipped...what kind of service is that, pretty damn lousy if you ask me. anyway, im really tired. and my laptop is really messed up, ill talk about it next post.  '),(25534,3382977,'22,June,2004','2004-06-22 00:00:00','   ok, so yesterday i didnt post becuase i was really busy...actually, i wasnt, i was just lazy like hell. Not much went on, except for going to the eye doctor and finding out that im a bit near sited (but no worries, im sure i still have 20 20...or i better) and that i have some kind of weird eye infection that is causing this burning sensation in my pupil and is decreasing my sight every second...actually its not as bad as it seems...the doctor was really nice and friendly and told me to get this tupe of \"eye drops\"...but i knew there was a twist to it using my erik-senses. And what do you know, there WAS a twist to it, i get home after raiding Walgreens and its an effin gel!!! what the hell, how am i supposed to drop a gel that comes out of the tube not in drops, but like toothpaste into my eye...and the worst part is, it says apply two DROPS into each eye before sleeping...well, how is that gonna go, i have two options, a. do it like a normal person would with a normal eye drop thingy and then find out that a big long chain of it has dropped in (which is definetly not the most comfortable of things) and then have a big bad time fixing the mess...or b. put a dap on your finger and try to force it gently into your eye...well, last night i attempted plan b. and found that 50% of the gel got in my eye lashes and eye lid and only a bit got into my actual eye...oh well, better luck tonight i guess.   besides that ordeal, i got home and found out that my spanking new and shiny and cool credit card had arrived, so i activated it and decided to go and take it for a test drive, so i went to the nearest Video rental store and got two movies....well, robert got one of them, Malibus most Wanted, and i got Dickie Roberts: Former Child Star, which i havent atched yet. So we watched Malibus Most Wanted...which i would give a umm...5 out of 10...its not really a good movie, its kinda pathetic and the plot is weak, but it does have some comedy and stuff, not too stupid that its unbearable to watch...(like Garfield). anyway   after the show i went down to my neighbors house down the street, where i found a giant mess in the dirveway(the night before we had a huge barbeque with smores and the like, which we made ourselves, inculding the fire). There was a bunch of ashes and pieces od burned wood everywhere, and once we cleaned up that mess we found a crack in the cement and a ton of burn marks around the place where the main fire was...and found out that it was permanent...but fortunatly we can do it again tonight, except putting the fire about 5 inches off the ground with dirt underneath...which will help with that problem.   today, on the other hand, i got up and had some breakfast and played some cards and did some puzzle and filled in some crossword and then went with my mom to the tailor to get my good suit fixed and fitted and cleaned...then me and the rest of the kids went out to eat and then we went to Nordstroms for some new \"kicks\", where i found a really nice pair of K Swiss...which is unussual for me since i usually dont like K Swisses instead this time it had like a running type of bottom (not the usual plain flat bottom) and it looks nice...and i got some new t shirts and a really nice looking tie...so, until next time....  '),(25535,3382977,'20,June,2004','2004-06-20 00:00:00','   today i had to wake up at 9 to go to the Fathers Day Breakfast at church, which wasnt bad for a church breakfast, it was actually quite nice. We had some hashbrowns and some pancakes and some sausage and some orange juice and some eggs (i didnt have any eggs, but everyone else did) and some fruit and then some carrot cake to finish it of (which i didnt have either (carrot+cake...don\'t do much for me)). Then we had an amazingly on schedule church service and then i got home and chilled for a bit, then had hamburgers for lunch and headed for the YMCA swimming pool with gramps and the rest of the kids (instead of lap swim it was Rec. swim so everyone got to go) and it was really fun. I had my new state of the art goggles and checked out the jacuzzi and then the pool and then the sauna and after a good hour or so we decided to go. This whole time my mom and dad were out somewhere at a party from 1 to 5, so we had the home pretty much to ourselves. After getting back from the Y we stopped by the Baskin Robins 31 flavors place and i had some ice cream and then we got home and watched Moulin Rouge, which was a very good movie, i liked the music a lot, thinking of  getting the soundtrack. Then we made our own pizzas, dough and everything, and i made mine somewhat better than last time (last time it tasted like a dried up breakstick with some weak cheese on it). This time it tasted like a pizza except it looked square, and could have used a bit more sauce...but it was good. Then we chilled for a bit and my neighbor down the street, who has asked to keep his/her name confidential, came up with the brilliant and somewhat random idea of doing S\'mores over a homemade fire. Well, anything that involves fire and eating is all good for me...so we rush down to west portal to pick up the supplies (marshmallows, chocolate, graham crackers...you know). By the time we were there i started to get a bit worried since it was past my parents due time to get home (they were supposed to be back at 5 and it was something like 6:15) and they wanted to celebrate my dads birthday and fathers day right when they got back, and if robert nor i were there we would be toast...and i didnt have my cell phone...so we would be burned toast. anyway, we get back and there not there, so we decide to start the fire then just incase they get back soon. At first we have a pile of newspaper and sticks that are just making smoke and ash and a mess and it was really windy so nothing seemed to be working, but then we got smart and made the usual pyramid shape and things started to get toasty...we got out the goods and had a party. so...at about 9:15 (4.25 hours past their due time) my parents roll in and we go home and do some more celebrating and then get kicked into bed and now im here typing away while my friend is looking up on the web if its illegal to have a bonfire in your driveway (which is YOUR property so i dont see a problem with what you do on it). sooo...my eyes have been killing me forever so im gonne gat them checked tomorrow...and im hoping to see one more movie before i leave. I finally ordered the extra big SD memory card for my camera and hope it gets here before i go (i did the free shipping one which takes 5 to 9 days to ship and its June 20th and i gotta go June 28th...that means i need it by June 27th, which means it better noe take more than 7 days or i am screwed...). so, until next time  '),(25536,3382977,'19,June,2004','2004-06-19 00:00:00','   so...i was woken up by my grandfather who said he was going out to get the ingredients for the cheese cake we were supposed to make  for fathers day (and they are really good when i make them). But, i remembered that i had a crazy YO practice from 10:30 to 12:30 and then a lunch break and then the concert at 2. So i got dressed and started the drive over but then my mom called and i had left some tickets and the purple form which had all the tickets i had sold for the contest at home...so we drove back and at the last second got back to the free way and i was about a minute late...turned in my pruple form, only missed a couple announcements...and we practiced, then i found out at lunch time that i had WON the CONTEST and i get a free deck of cards, two free Mahler tickets to any of the SFS Mahler 2 concerts, $50 gift certificate to Flight 001 travel store ( urlLink  ) and i got the GRAND PRIZE which was one of the two official Bon Voyage concert posters, which i got for myself...which im gonna put up in my rooms which is pretty empty on the walls, so its a good addition. anyway, tomorrow i have to go to some Fathers Day breakfast at the church...ho hum.  '),(25537,3382977,'18,June,2004','2004-06-18 00:00:00','   the biigg Bon Voyage concert is tomorrow at 2 pm...and i might have a chance at some grand prizes for the most-tickets-sold contest. After getting home at 10:30 after our intensive rehearsal i found out that i have 27 tickets sold total up till now...and they are due tomorrow at 10:30 when the pre concert rehearsal starts so i have a good bunch on my purple paper to hand in. very pleased with my success , didnt think i would get so many.  anyway... we are gonna celebrate my dads birthday tomorrow so i get to go to Outback Steakhouse (one of my favorite resturaunts) and make a cheesecake for fathers day which is coming up fast...until then, theres gonna be rehearsing from 10:30 to 12, then a lunch break, and then the concerts at 2 and it is supposed to be pretty long since we are also playing some encores at the end (which makes one full concerto, a full symphony, and a couple short pieces, and then two encores...its gonna be long).  '),(25538,3382977,'17,June,2004','2004-06-17 00:00:00','    hmmm...well, first off, Dodgeball: A True Underdog Story is a hilarious movie. It is full of laughter and jokes, even though the commercials gave some of the good ones away. The first two thirds of the movie are full of new and exciting stuff (if i try and forget the ads on tv) and then it slows down when it nears the last third...and then the ending was kind of awkward, but its still a good movie!!! believe it.   the storyline is pretty straight forward...A bunch of loosers enter a Las Vegas dodgeball tournament in order to save their cherished (yet pathetic) gym from a health fitness chain called GloboGym...yep, not too complex and not too stupid, which is just right for a good comedic satire.   so if you feel like spending a good hour and forty minutes on a movie which gets you to laugh a lot and if your ok with the not so spectacular bang at the end, and if your ok with leaving the theatre with not as much laughs as you got in the beginning or middle, then this movie is just right for you. I give it a 8.5 out of 10.  '),(25539,3382977,'17,June,2004','2004-06-17 00:00:00','   well, today i woke up and went t a friends house and watched a movie (Chronicles of Riddick (read previous post (if you havent already))) which started at 11:30 am, which i think is the earliest i have ever watched a movie...i think. I always remember it being at least past noon...anyway, tomorrow im seeing Dodgeball at 1:10. So, after watching the movie today i went back to his house and played some games and then went to my violin lesson which started at 4 and then went home and then went to my YO rehearsal which started at 7, but i was 15 minutes late...but it didnt matter since i dont play the Piano concerto in A minor ( since im in the back ), so i had a good hour and a half to do stuff, so i got a bit to eat since i didnt have much time at home earlier...and cruised around Van Ness (i just remembered i should have brought my gift certificate for a book store near there). And i called the Waldenbooks hotline for gift cards and found a whopping 30 bucks on my card that thank god hasnt expired (since i live in CA). Anyway...thats just about it. I totally forgot about a rehearsal with a friend for his recital coming up and he got kinda pissed and decided to call it off, but im glad he did since im kinda busy with other stuff...or am i?...hmmm.  '),(25540,3382977,'16,June,2004','2004-06-16 00:00:00','     Ok, first of all...i would like to give it a 6 out of ten...why? because it was really unstable. Yes it had tons of action and fighting and sometimes some suspense, but it would start some bits of plots and not evolve them...it would skip from a start of an evolving storyline to a big action packed fight which would then lead to another not fully understood story. Yes, i havent seen the first Riddick Pitch Black, but people tell me that the first was just as confusing. In the beginning you see Riddick running away from some bounty hunters (which supposedly is five years after the events in Pitch Black). After fiddling around on Inam\'s planet (Inam being one of Riddicks old pals) we find out that there was a girl named Jack (which he supposedly saved in the previous movie) that has been put into a prison...so he goes and saves her, but she gets taken away by mercs, and then he goes after her again, and she ends up dieing and he becomes the leader of the mercs...thats the whole movie in a nutshell.   Now, the battles and fighting by Vin Diesel are awesome and very intense...and there are some really cool parts where he shows off his skills as a Feore (the race the Riddick is which is something that the Lord Marshall is afraid of and the reason why the mercs are all after him (to destroy the Feorian race forever)). anyway, there were plenty of times when there was also a ton of computer graphics, which made it look totally fake...but if you are really into the movie you can kind of look past that (since it is in space on some weird planets). So, in general, The Chronicles of Riddick is a movie i wouldnt recommend seeing, but if you are a person who loves action and computer graphics and is usually too dumb to pick up any plots (like the Matrix), then this is the movie for you...and by the way...the ending was the thing that made me actually start to dislike it...it was horrible, but im not gonna ruin it for those who are actually willing to go see it for themselves.  '),(25541,3382977,'16,June,2004','2004-06-16 00:00:00','   well, i went to go see a movie today with a friend. Had to see Mean Girls again since he had watched all the ones i hadnt and i had seen all the ones he hadnt...but still, Mean Girls is a ggoooodd movie. If you havent seen it...SEE IT. anyway, heres what i think about it :    During the first 2/3rds of the movie there was just non stop comedy. Its hilarious. The only thing that could have been better was the fact that the last third of it turned into a big, G rated, everyones happy ending with the morals of true friendship and blah blah blah.... It made you leave with a not so mean feeling that the whole beggining put you in the mood of. But still, it was a very good, naughty, and lots of hot girls movie. I would give it a 8.5 out of 10....yah   anyway, after the movie we went outside waiting for someone to pick us up and a woman comes up and starts asking us  questions for a survey or something...and after that little survey she hands us a pair of flyers asking us to go to some videogamers get together ... it said on the paper that i was invited to participate in a focus group to learn more about my thoughts and opinions on a variety of topics....hmmm, well they are paying $75 in cash for a little meeting , and the woman said that you watch commercials and pick which ones appeal to you the most and test games and stuff...but who knows...and its on a day that im busy on so, just to tell you ... for tomorrow: The Chronicles of Riddick!!!!  '),(25542,3382977,'15,June,2004','2004-06-15 00:00:00','   woke up...stayed home all day...nothing interesting...except for making my own pizza for dinner and playing some online games and setting up some fun movie trips...hopefully get to go see Chronicles of Riddick on thursday and am hoping to see Dodgeball on friday and I might see Mean Girls again tomorrow....anyway, ill post reviews of all three movies just to keep you updated (:  '),(25543,3382977,'14,June,2004','2004-06-14 00:00:00','   today i went swimming again at the Y (tried to go yesterday but it was one of those closed-for-cleaning sundays). Got both my new shelves up and running, still dont know what to put on them. Still looking around (mostly online) for a good stereo system (something like a 5 disc changer with a subwoofer and some satellite speakers that i can set up around the room). Im gonna check out the GoodGuys today and see what they have and to actually find out what i really want (i know what i want but looking online i noticed that it doesnt exactly exist, there are mostly speakers sets, home speaker systems for a home theater, and a thing that has DVD and CD player that has its own speakers...but im still not seeing a system just for CDs, but i guess ill go for the DVD ones (that also play CDs) so that i have something incase i get a nice little flat screen TV for the wall (im looking ahead a long way)).  Also im gonna go and spend all the gift certificates i have amassed this past year (none have expired luckily). They are mostly small but i have a lot of them, Borders (including two coupons that i can take along), Waldenbooks, A Well-Lighted Place for Books, and Tower Records gift cards and coupons...  Gonna go and make some Jello and some Chocolate Pie before i go.  '),(25544,3382977,'12,June,2004','2004-06-12 00:00:00','   well, today was a very slow and tiring day for me. first of all, i spent the night at a friends house and stayed up watching tv and the movie What Women Want while eating popcorn and tons of icecream. So i got to sleep at about 12 and then woke up at about 8:20 remembering that we had to be at a rehearsal across town at 9:00. Amd to top it all off, neither of the parents were home to take us there. So i called my dad and he picked us up and got us there a bit late but not bad. While waiting i played a really fun game called crimsonland or something like that. When i first got to the friends house i played the demo with him a few times, but then i timed out, and we were addicted. So we payed the small fee to get the whole thing and played it for a few hours when we noticed that playing one player taking turns was starting to get dumb and we needed to get something like a joystick or a game pad from the nearby RadioShack to play 2 players...so we got one for a small fee again and played it about an hour...then did the tv and movie and then my friend played the game got a few more hours while i tried to go to sleep. Anyway, if i cant get a copy of it i guess ill just get it myself...since im getting some kind of credit card soon for Euro trip i wont have to ask my dad or mom for theirs if i want something off the Internet...which will be a big plus (since they wouldnt let me spend 20 bucks on a downloadable game off some site (but its reeeeaaalllyyyy good (trust me))).  This all happened after watching Garfield, which absolutly sucks (as you would know if you read the previous post). I just wanted to give you a bit more on how much that movie sucked...here what CNN\'s David Germain thought of the movie:   \"\"Garfield: The Movie\" elicits laughs about as infrequently as a hairless Sphinx kitten sheds fur. The movie\'s vocal coup -- landing Bill Murray as mouthpiece for droll Garfield, the lazy hero of Jim Davis\' comic strip -- falls flat amid the banal lines he\'s given to mew.  And the curious decision to make Garfield a computer-generated cartoon figure amid live-action cats and dogs -- while not quite the distraction it could have been -- still is bothersome enough to yank viewers out of the story now and then.  That is, yank them out of what little story the filmmakers dredge up.\"  he also points out that \"The structure of the comic strips is often maintained, with Garfield handling much of the dialogue as one-sided monologues. Too bad he\'s given virtually nothing funny or interesting to say.\" and that \"\"Garfield: The Movie\" hits its low point as Murray caterwauls an awful ditty, \"New Dog State of Mind,\" to the tune of Billy Joel\'s \"New York State of Mind.\" The many zeroes in his paycheck for some easy voice work must have been all that sustained Murray while crooning the insipid lyrics.\" and to finish off the article he ends with this: \"Young children may get a few laughs out of the movie\'s many pratfalls, but even the slapstick of \"Garfield\" runs thin compared to a decent Looney Tunes cartoon.  The filmmakers seemed content to toss out any old stray cat of a movie with the \"Garfield\" brand name, figuring fans would show up.  They may be right, but some viewers may not be such cat lovers when they leave.\"  ....my case rests...  '),(25545,3382977,'11,June,2004','2004-06-11 00:00:00','   well, today i had the pleasure of watching the one and only GArfield~~~yes, it is so great i actually left the theatre after about 20 minutes of hell. It started ok with the usual \"let me show you the ropes\" beginning. After about 15 minutes i started to get uncomfortable not knowing what the plot was, but my patience hadnt run out yet. After a bunch of random junk and mini mini plots things started to get hairy. Firstly, they were using the same tricks and jokes many many times, as if we \"little kids\" (its a PG movie ) would forget about them and laugh again. For example, Garfield had already made the point that his chair was his chair, and no one sat on it. Well, when Odie comes and sits on it, Garfield states that again, just to remind us, which wasnt bad. The he pushes Odie off the chair and he starts watching his music video. Well, after about 6 times of puching and jumping and listening and dancing i started to get really pissed off. OK, we get the picture!!! Odie is a dumb bitch and Garfield hates him. Garfield on top of that has the stupidest \"jokes\" i have ever heard and not only that, he repeats them about a billion times. And one other thing that just really got the worst of me. I was on the verge of going crazy when all of a sudden, Garfiled starts singing!! What was he singing? I dont know. First of all, Bill Murray cant sing. Secondly, i did not pay 8 dollars to watch a fat orange fur ball sing about how this lame dog has taken over his owner....dude, cmon. So, i just couldnt handle it.  Now the question is...why did i go see it in the first place? well, i was on a time budget, and that was the only short movie that started at the right time...but from now on, i guess ill just play games instead. And a warning to all....Dont go see Garfield or you will die!!!   '),(25546,3382977,'10,June,2004','2004-06-10 00:00:00','   wow, my grades were a lot better than i expected...i was really worried, luckily a friend allerted me that it arrived today and to get it before my parents did...but there is nothing to worry about, lowest grade was a B+ in darn English I....which actually is not my least favorite class but i am just not an english natural here so i try my best...but the books that we read are not too bad, and i am really good at grammar thanks to my middle school english teacher so there isnt much of a problem. In the other classes i got almost all As except for Geometry and Biology, which believe it or not are becoming two of my least favorite classes...even though i used to like science and math best. Well, things have changed.   Today i woke up and started a new puzzle after we finished the other violin and chocolate 500 piece one. This one is 500 pieces as well but the pieces are a bit bigger (making it 20\"x20\") and its a mystery thingy were you read the short story and you try to figure it out while doing the puzzle, which you dont know what it is at all (no piture anywhere) which makes it a bit harder.  Robert and errol are still into Zelda. Every morning i hear the usual sounds...getting kinda annoying. Got a couple more cds today in the mail and i have some big plans for my room. Hoping the get an identical cd tower that i have on one side of this book case and put it on the other side (since the one i have is getting full), but im not sure since there was only one left at the place i got it, hopefully there are more online. Then im hoping to get rid of my 8 year old boom-box thingy and replacing it with the new 5 disc cd player/dvd player ( most are cd/dvd players, just incase i have a tv ) with a subwoofer and a couple satelite speakers (maybe 5 that i can put in the closet and one in the bathroom and three around the room), along with some other extras, which arent too expensive (if i were to get all this it would have to be my own money since its like an extra that is not necissary. I see most that i want at about $400, but there are cheaper ones that could work.   So after checking out the new puzzle i got ready to go to the YMCA with my grandfather (the lap time) and so we did and stayed from 12 to 1 and then went to get some lunch and then headed to Home Depot, where i got one new big dark oakwood shelf (for above my bookcase) and another smaller pinewood shelf that i will put somewhere, which will probably hold some random stuff....not sure yet. Then we went to Albertsons to get some Jello and some Chocolate Pie Mix (will make two probably tonight or tomorrow morning).  Hoping to see some movies before next week.   '),(25547,3382977,'09,June,2004','2004-06-09 00:00:00','   well...now that im really in the \"movie-mood\" im gonna tell you all the movies that i want to see in theaters or rentals...and if you have any comments about it, then comment! and if you cant comment, then please notify me right away on why i shouldnt spend 10 dollars and waste a couple hours at the movie theater. Got it? Good...  The Day After Tomorrow Soul Plane Raising Helen The Chronicles of Riddick White Chicks Fahrenheit 9/11 The Terminal Dodgeball: A True Underdog Story Spider-Man 2 (a Must) The Village The Manchurian Candidate Pirates of the Caribbean 2  (YES, its coming out in December 2006, believe it or not) Spider-Man 3 (they say its coming out May 2007...but who knows, at least its out there) Underworld 2 (still havent seen the first, but this ones due sometime in December 2005) Harry Potter and the Goblet of Fire (November 2005, its a must see!!!)  Movies that i gotta rent! well...i dont know, but i do wanna see the Sum of All Fears, and maybe The Panic Room...but who knows...thats after i have been to the movies a few times and am really borded  '),(25548,3382977,'08,June,2004','2004-06-08 00:00:00','   All Time Top 20 Movies at the US Box Office Released- 	Film Name- 	Total 1. 	1997- 	Titanic- 	$600,788,188 2. 	1977- 	Star Wars- 	$460,998,007 3. 	1982- 	ET: The Extra-Terrestrial- 	$431,197,000 4. 	1999- 	Star Wars: Phantom Menace- 	$431,088,297 5. 	2002- 	Spider-Man- 	$403,706,375 6. 	2003- 	Lord of the Rings: The Return of the King- 	$377,027,325 7. 	2004- 	Passion of the Christ, The- 	$369,895,388 8. 	1993- 	Jurassic Park- 	$357,067,947 9. 	2002- 	Lord of the Rings: The Two Towers- 	$340,478,898 10. 	2003- 	Finding Nemo- 	$339,714,367 11. 	1994- 	Forrest Gump- 	$329,693,974 12. 	2004- 	Shrek 2- 	$321,821,000 13. 	2001- 	Harry Potter and the Sorcerer\'s Stone- 	$317,557,891 14. 	2001- 	Lord of the Rings: The Fellowship of the Ring- 	$313,837,577 15. 	1994- 	Lion King, The- 	$312,855,561 16. 	1983- 	Return of the Jedi- 	$309,205,079 17. 	1996- 	Independence Day- 	$306,169,255 18. 	2003-	Pirates of the Caribbean- 	$305,411,224 19. 	2002- 	Star Wars: Attack of the Clones- 	$302,181,125 20. 	1999- 	Sixth Sense, The- 	$293,501,675   All Time Top 20 Movies by Global Box Office 1. 	1997- 	Titanic- 	$1,835,400,000 2.	2003- 	Lord of the Rings: The Return of the King- 	$1,117,602,779 3. 	2001- 	Harry Potter and the Sorcerer\'s Stone- 	$975,800,000 4. 	1999- 	Star Wars: Phantom Menace- 	$925,600,000 5. 	2002- 	Lord of the Rings: The Two Towers- 	$922,986,073 6. 	1993- 	Jurassic Park- 	$920,100,000 7. 	2002- 	Harry Potter and the Chamber of Secrets- 	$869,400,000 8. 	2001- 	Lord of the Rings: The Fellowship of the Ring- 	$867,683,093 9. 	2003- 	Finding Nemo- 	$844,400,000 10. 	2002- 	Spider-Man- 	$821,700,000 11. 	1996- 	Independence Day- 	$813,200,000 12. 	1977- 	Star Wars- 	$797,900,000 13. 	1982- 	ET: The Extra-Terrestrial- 	$772,000,000 14. 	1994- 	Lion King, The- 	$767,900,000 15. 	2003- 	Matrix Reloaded, The- 	$727,400,000 16. 	1994- 	Forrest Gump- 	$679,400,000 17. 	1999- 	Sixth Sense, The- 	$672,800,000 18. 	2003- 	Pirates of the Caribbean- 	$652,102,724 19. 	2002- 	Star Wars: Attack of the Clones- 	$648,200,000 20.	1997- 	Lost World: Jurassic Park- 	$614,300,000   '),(25549,3382977,'08,June,2004','2004-06-08 00:00:00','   Today i woke up to three people banging at my door: robert, mom, and victor, and old friend from middle school (more like elementary). I got up, played some videogames and watched my brothers play Zelda for some stupid reason....and then had to go and wash the cars, do my practice, make and eat lunch all before 12:30 which was the time Harry Potter 3 started. Today was gonna be a \"movie day\". I was gonna go see Shrek 2 with vistors brother and the rest of the kids were gonna see HP3. So we wash the first car and victor took it for a \"blow dry\"...if you catch my drift, and then we tried to evade the second Yukon XL...but my grandfather got us out there...it had to be done. Anyway, we come in and eat some hot dogs and sausages in 10 minutes and then rush to the movie thaters....getting there 10 minutes before HP3 started and a whopping 40 minutes before Shrek 2. So victor\'s brother and I hung out in the arcade for 10 minutes and then bought ourselves some food and sat in the theater....then it began!   So...after battling a fire-breathing dragon and the evil Lord Farquaad to win Princess Fiona (Shrek 1), he now has to get by the king and queen and Price Charming and Charmings mother The Fairy Godmother. Of all the chraracters my favorite this time would have to be Puss in Boots...it was hilarious (last movie my favorite would be the donkey).   Mike Myers=	        Voice of Shrek Eddie Murphy=	        Voice of Donkey Cameron Diaz= 	        Voice of Princess Fiona Julie Andrews=	        Voice of Fiona\'s Mother, Queen Lillian Antonio Banderas= 	Voice of Puss-in-Boots John Cleese=	        Voice of Fiona\'s Father, King Harold Rupert Everett=	        Voice of Prince Charming Jennifer Saunders=	Voice of The Fairy Godmother  Just in case you were wondering...   ok....so this is what i thought about it...well, it was just plain funny. I got to see more of the three pigs, the gingerbread man and Pinocchio (including his desire to wear women\'s underwear) which was good. Even though the first was by far better because it was all new and just stuck in your head funny, the second had (along with the first\'s themes and jokes) some of its own fun stuff. It took a lot of stuff and used them from other movies, especially the Beauty and the Beast\'s themed dance thingy with the moving furniture and stuff with the Fairy Godmother and some Mission Impossible and Spider Man added in (the mission to save the prisoners and the special kiss between Fiona and Shrek respectively). Thery also added some real worls things (Hollywood themed place) which made it seem more realistic in a way...kinda. The Donkey is just the same being his annoying and crazy self as the first one...except this time he is jealous of Puss in boots. Shrek 2 may not be as fresh and new as the first, but it still gives the viewers a good laugh or two.    so for a film that has hit the record for playing at 4,163 domestic theaters and for being ranked (currently as of June 9th) at number two on thw Box Office charts, it is definetly a must see movie. i would give it 9 out of 10.    So after the movies i came home and hung out with some friends...then i had to go blow the leaves from the garage and driveway, which got very tiring and annoying when the wind started blowing the opposite way...but now theres a big line of leaves and stuff in the middle of the street ready for the street cleaner tomorrow morning.  Got a good bunch of new cds to my collection, will update the list when i reach my 100 mark! (:  '),(25550,3382977,'08,June,2004','2004-06-08 00:00:00','   urlLink    This might seem like a normal house...but this was actually a picture taken through a window of a car while driving around my neighborhood testing out the 1/1000 of a second shutter speed...yes, thats fast, that means that it will stop any motion no matter how fast (well...anything thats not faster than 1/1000th of a second)...  '),(25551,3382977,'07,June,2004','2004-06-07 00:00:00','   urlLink    heres a funny looking picture that i took with the shutter open for a couple seconds in the car on the way to get icecream...notice the green light, it was really bumpy (:  '),(25552,3382977,'07,June,2004','2004-06-07 00:00:00','   today was another somewhat fun since my grandparents were here yet somewhat boring since there was not much to do day....  first i got up and had some breakfast like usual...then i walked about while mom and her mother went out shopping...and printed out a humongous walkthrough for errol and rob for Zelda (Orcarina of Time)...i dont know how but all of a sudden they have sparked an interest in it once again, and the game has been taken out of the dusty box and played for a consecutive few hours (first time in years). I have to admit it is a classic...and i did have an urge to go and start where i left off, but then i remembered why i left it in the first place...oh yah, it was way too hard and i was sick of having to take it to my friend Max\'s house and having him beat the entire thing for me (actually up to the Water Temple i think...if you care)...and then we got the second one that came out about a year later (Majora\'s Mask) and it was impossible....so it was probably played for a day and then stuffed into a closet box somewhere...that game i have to admit was one of the most complicated games i have ever played...anyway, so now robert and errol are into that old game and all day i hear the good ole sounds of my childhood...kind of.   So my mom and grandma arrive home and tell me to try this stack of maybe 6 shorts and 4 t-shirts which i desperatly need for this summer (since i need 9 days worth of clothers for the euro-tour and i need some stuff my grandparents are gonna take back with them to florida so that i dont need to pack a whole bunch of florida stuff with me (since im not coming back home between trips (might be spending the night with some friends))). So they were all pretty good, and the t-shirts are fine...  I also had a chance to play arounf with my camera today...my grandfather gave me a loonngg lesson on what all the special words meant (aperture, exposures, shutter speeds, telephot postitions, etc.) and now i know a lot more about the tons of different modes i have...finally got the cool night effect of all the lights and things blurring when you keep the shutter open for a long time...they look cool, figured out and tested the timer thing and the progressive thingy and taking 5 pictures a second and all that jazz...very cool.  Watched the Lakers game...but dont know who won (left to go get ice cream when overtime started)...even though im not a basketball fan at all...i think the pistons should get the championship this time...im tired of the lakers...anyway  got some cool pictures but im tired of having to delete them so i got a couple downloaded onto the comp and may post some later one...still gotta get that 256 mb card...will hold tons of pictures...oh, and i got the cool movie mode working, it is tons better quality then i thought it would be...anyway, im gonna post two pics that i thought were interesting...check them out  '),(25553,3382977,'07,June,2004','2004-06-07 00:00:00','   \"The Flames, threatening to become one of the most unlikely champions in NHL history, were held to only seven shots in a dismal first two periods before making a frantic late surge started by Craig Conroy\'s power-play goal midway through the third.   Ruslan Fedotenko scored twice, including the critical first goal just as he did in the conference finals, and the resilient Lightning held off the Calgary Flames 2-1 in Game 7 Monday night to win their first Stanley Cup.\"   first the sharks are out..and i loose hope in them, and then my good ole flames loose...oh well  '),(25554,3382977,'05,June,2004','2004-06-05 00:00:00','   Yesterday was a normal somewhat boring but still ok day. I got free time with my grandparents and errol since katarina and rob had to go to church from 10:30 to like 2 ( after they had gone out to eat somwhere ). I started the puzzle my grandparents brought here for me. Very fun and challenging (500 pieces, which isnt that big, its just has a lot of similar things and the pieces arent normal). Got about a quarter of the pieces together yesterday and have about half of it finished now ( its a violin surrounded by chocolates and some flowers ).  Today I woke up at about 9:30 and had some breakfast and did some puzzle and then played some cards and then had some lunch (hotdogs on the grill) and then played some more cards and the went to the dentist to get my braces removed (finally) and then got home and found that my camera (the new digital one (Minolta DiMage Z2 4.0 MegaPixels)) and i got to play around with it (makes cool sounds when it turns on and off and it has a lot of different modes and i cant wait to take it to euro trip).   anyway....now im waiting for a bunch of cds i ordered saturday...which are still being shipped, and i gotta get a 256 mb sd memory card before i go to europe...maybe a bigger one if i want...i found one for 37 bucks...not sure if they get much cheaper than that. I also got the LOTR Fellowship of the Rings soundtrack at west portal Music Store for 10 bucks (which is really good if you like opera and very loud and good orchstral syphmonic whatever music) while robert got the South Park Movie  soundtrack for however much it was...they said they accept any kind of somewhat decent condition cd for money or trade...so lily, rob and i and gonna try and collect cds that no one listens to anymore to see if we can make a bit of a profit...anyway, gotta go eat a big see-food dinner  '),(25555,3382977,'05,June,2004','2004-06-05 00:00:00','   So far these past few days of no school or anything else have been great. Thursday night at 6 was the SHHS Back-a-loree-et ( i dont know how to spell it ) service. I had to play some songs w/ Chauncey and stuff. Not the best but everyone thought it was amazing...anyway. Friday at 8 was Prize day at school, and then it finished at 10 ( might as well not have come ). Then i got to go home and went to go see Harry Potter 3, which was ok, ill get to that in a second. Then i rushed home after the 2.5 hour movie and got changed into a suit and rushed to the Hotel job. Then when that was over at 6, i went to go to my YO rehearsal at 7 and it ened at 10. My Grandparents arrive at SFO at around 7:30.  Harry Potter and the Prisoner of Azkaban is a movie i would give about an 8 (of 10). why? well you can definetly see and feel the difference between Alfonso Cuaron\'s directing in this movie and the  Chris Columbus\' directing in the others. Most people said that #3 was better than the first two, but i think that they are about the same, or even better in some ways. The first two would switch from scene to scene with not many breaks, which kept the watcher on the edge of the seat, while the third has these funny fade outs and interesting endings of the scene ( mostly with Harry Potter fainting or passing out to make it seem like its supposed to be like that ). There were also a lot of things put in that made the movie seem a lot different than the book\'s idea of how it was supposed to go. The main  plots were there but somethings ( like the music that Lupin played ) gave the movie a weird feeling. anyway, you would have to see it for yourself to see what i really mean. I would go see it. It is a fun and exciting movie. 8 out of 10 is what i would give it.   Another thing. In the YO we have a bunch of enchores that we are gonna play after every concert on tour, and i found one of my top favorite pieces today. Tchaikovsky\'s Pas de Deux from the  Nutcracker . It isnt my usual exciting huge sound orchestral pieces ( even though it does get really loud thanks to the brass ). It is more of a really intense love song sort of. It sounds really good. Jut if you were wondering what my favorite songs are nowadays....  '),(25556,3382977,'02,June,2004','2004-06-02 00:00:00','   some new things going on around here...getting ready for my grandparents to come, so ofcourse my mom wants everything in mint conditions ( as in we gotta clean all the floors and get their room ready and do the yard etc...)  Now that schools over...pretty much, i was able to sleep in until 11:30 today ( a record for me since i always wake up hungry at about 9 )... Got to go and do the back yard, which is a humongous mess. Let me make this short...imagine a backyard were weeds have taken over every inch of dirt, the wooden fence is falling apart with holes in it, the trees are practically obstructing the path to get anywhere, and there are pieces of toys, play stuff, etc laying around in the weeds and sidewalk and pieces of brick and glass everywhere....yes, thats our backyard. So i have to go and get that mess picked. And i have had this sudden urge to do some gardening (ever since i went to the garden store on sloat to get the items needed to repot my cactus ). So i am gonna not only have to clean up the backyard but attempt to grow and garden of flowers and some kinds of vegetables...  My dad finally ordered my digital camera...hope i made a good decision ($450 camera cant be a bad thing). Should arrive in 7 to 10 days...cant wait.   '),(25557,3382977,'01,June,2004','2004-06-01 00:00:00','   Yes...today was the last day of \"school\" ( i do have to go to the campus on Friday since its something like prize day or some junk like that...ho-hum )  Friday\'s Final Exams: English ( did ok...at least an 85 ) and there was a crazy hard Geometry Exam...which i didnt finish ( but i got to finish up a couple problems today ( i had to leave right about 12:05 when almost everyone else stayed till about 12:20 )) I had my Latin I and History exams today...Latin was ok, maybe at least an 85 ( maybe less, i hope not ) and History i kinda messed up on but it wont really effect my 98% grade so...oh well anyway....i didnt do as well on most of the finals as i wanted to, but still, its only freshman year and its not the end of the world....did ok . now i have wednesday and thrusday off...and a little thingy that i gotta do on friday morning..and then im definetly free forever...this is great  Freshman year has gone by fast   '),(25558,3382977,'31,July,2004','2004-07-31 00:00:00','     The Bourne Supremacy is a pretty good thriller. It still isnt as thrilling or action packed as the first one, but it is pretty good considering the fact that most of the time sequels arent always up to snuff. Since this is the second installment of a trilogy (yes, there will be a third coming soon), im not sure how the third do. maybe itll be a blast, but i doubt it. anyway, we\'ll see when it comes.   Unlike the first one, Bourne is not very interesting. In the first movie, Bourne is a guy who is just trying to figure out who he is, while at the same time doing all these amazing things keeping the bad guys away and keeping himself alive. The plot was more interesting. In The Bourne Supremacy, Bourne is just a crazy killer, until the last half hour, when all of the sudden, he becomes all sorrowful and compassionate, which is kind of awkward. In the Bourne Identity, the fight scenes were more clear and exciting. In the sequel, the fight scenes were so cut up and edited and stuff that they were just hard to follow. You might as well just waited till the end to see whos left standing. The car chases and stuff make up for that though. Besides those points, the movie was good. I would give it an 8 out of 10.   now your probably thinking....is he going to give The village a better rating than Bourne Supremacy?. actually, they both have their good parts and weird parts...but Bourne seemed to have more than i wanted. so thats that.   '),(25559,3382977,'30,July,2004','2004-07-30 00:00:00','   urlLink    This is the cathedral (one of the many) in Salzburg. Its so far my favorite cathedral (i didnt get to see the inside of the Notre Dam so i cant say thats better, even if it might be). There are five organs inside, one above the main entrance where there is usually an organ, and four around the corners of the cross section in the front. It is also magnificantly decorated and is just amazing!  '),(25560,3382977,'30,July,2004','2004-07-30 00:00:00','   urlLink    This is the famous statue of...Heifitz?. i think so. Its in Vienna. Its a nice gold statue located in the Stadtpark which was right across the street from out Hilton Vienna.  '),(25561,3382977,'30,July,2004','2004-07-30 00:00:00','   urlLink    I think this is Salzburg, im not 100% sure, since i have been getting Salzburg and Strasburg very mixed up. Still, its a nice pic.  '),(25562,3382977,'30,July,2004','2004-07-30 00:00:00','     The Village is a very interesting movie. I would split it up like this. First half, somewhat scary, middle section gives away a lot of unknown info which totally changes everything, late middle section, somewhat scary,  and the last part is very good. if all that made sense to you, good. if not, oh well, watch the movie. The Village is not a run and hide and have nightmares scary. It is more of a suspense mixed with some scary. There are a few jump-out-of-your-seat parts...but other than those the movie is suspense. The movie dumps the audience into a world with no explanation. Through observation and close attention you figure out that its really happening in the late late 19th century. If you didnt pick that up (most people dont), then you see a whole bunch of pilgrim people that look like they just got off the Mayflower. I\'m not going to spoil the whole thing for you but all i have to say is that when you get to the middle of the movie your gonna start not liking it...for several reasons. The biggest of these is that the movie spits out a lot of new things at you that you totally didnt expect, and it starts to seem like that the whole beginning was just a waste. but it wasnt and just hold on till the end, cause its worth it! I give The Village an official 8.5 out of 10.    ');
INSERT INTO `msgs` VALUES (25563,3382977,'30,July,2004','2004-07-30 00:00:00','   so far vacation after Florida has been not so good....first of all, i actually have to do stuff like practice and clean up and HELP OTHERS....who thought of such a thing! actually, i only needed to help two friends with some music stuff, so it wasnt that bad. anyway, a neighbor from down the street and my brother and I have been playing around with this huge and thick tree branch which we cut down yesterday, and now are trying to make something out of it. besides that, we rented four movies...Magnolia, Euro Trip, and two others which i dont remember. we went and saw The Bourne Supremecy, which was my second time seeing it...its pretty good. i havent finished my unpacking, but im just gonna take out clothes that i want for the day and then after they have been washed theyll get put into the closet in the right spot...which is a kind of slow way of just taking it out of the suitcase and putting it away. less work actually.   '),(25564,3382977,'30,July,2004','2004-07-30 00:00:00','   urlLink    not much explaining needed here.....  '),(25565,3382977,'18,July,2004','2004-07-18 00:00:00','   urlLink    here is the whole front of the hall.  '),(25566,3382977,'18,July,2004','2004-07-18 00:00:00','   urlLink    here is the ceiling of the Musikverein....a whole lot of gold and paintings.  '),(25567,3382977,'18,July,2004','2004-07-18 00:00:00','   urlLink    here is the Musikverein, the first concert hall in which we performed...it was one of the most beautiful halls i have ever seen.  '),(25568,3382977,'18,July,2004','2004-07-18 00:00:00','   life is slowly moving back to its original state...now i have practice to do, school is coming soon, got some summer reading to do (actually a whole books worth, i finished The Hobbit in Florida and now have The Things They Carried), i had to do that friday hotel thingy again...im starting to find it not worth 50 bucks...and i also gotta unpack all the junk in my suitcases, and get used to this somewhat cold weather (50s in the mornings and nights, 60s and low 70s during the day, with a nice dose of fog added to the mix). im really missing the nice 70s at night and 80s and low 90s during the day...it was reeally nice sitting out by the pool at night listening to the quiet Florida night, well, there is a lot of those weird bugs that make that chirping noise, but you get used to that. the only thing that i was missing there were friends that were my age, then it would be the life. I did have the neighbors to have fun with. A family with two little girls and a really nice mom and a cool dad, i went over a couple of nights to play XBOX with him on his hhuuggee screen tv, which apparently came with the house, including the surround sound. anyway, he didnt get to play the games that much, and when he did it was by himself against the computer, so he was happy that he had someone to play with, and a challenge. the mom was totaly different than what i expected after the stories of her from my grandmother. she was very alive and loved chocolate...i mean Loved it, as in she ate brownies that were so undercooked that the insides were still someone liquidy, covered with chocolate sauce and chocolate sprinkles and chocolate icecream...and i thought i loved chocolate.     so, even though i was in FL for a good 2 weeks, i still wanted to stay longer. found out that my camera held 200 pics with the 256 MB card. got about 160 of europe and 30 of Florida. i deleted 20 pictured that werent important earlier. will post some soon.     plane rides back were a mess. unlike the nice and easy plane rides to Florida, this time there was a delay in the air between Palm Beach International and Charlotte, thanks to some thunderstorms. then, i didnt have any time to stop for lunch as planned since we landed at Charlotte and 3:30, which was the boarding time of the next flight, and even though the airport is fairly small compared to some other ones i have had troubles with (Newark...huge) i still didnt want to be too late. i then found out after sitting on the plane for a solid 30 minutes that i did have 10 minutes to stop by at a Burger King on the way. now i know that its ok to get to the gate 20 minutes after boarding has started, since the plane left 30 minutes after. anyway, i get on the plane and they cant lift off thanks to the weather. so im really hungry and i get some pretzels and eat them...and after 3 hours my stomach isnt feeling very good. all i have is 2 extra bags of pretzels from the flight and half a snapple (kiwi strawberry flavor) and two raspberry smoothie drinks...which are the last things i want at the moment. unfortunately i didnt have time (or i thought i didnt) to go to the bathroom on the way to the gate in Charlotte...which is bad since i have a personal rule to never use the airplane bathroom...i didnt even use the lavatory on the euro flight from NY to Vienna. but i just had to...which was better than i thought. so, after landing i felt a lot better, went to the baggage claim, and there was none of my suitcases to be seen. well, i had to get them to deliver it to my house...kind of a pain in the neck. but ok. so they delivered it today at 2.     got a new cellphone...since i killed my old one in europe, forgot about the 120 volt to 220 volt thingy in europe. it is color and looks a lot better. and it flips out, so i wont be accidentally dialing people in my pocket....even though i usually lock the keys. im hoping to get a new computer...was hoping for a new laptop but im thinking for money purposes to get a nice desktop, since i only took my laptop out of my room once, and desktops are less expensive and usually pack more for less $ than a laptop. and once i get to be a senior i might give the desktop to the family and get a good laptop for college. anyway, those are my computer hopes for the next couple of years....but this one seems to still have a bit of life in it. the stereo system is on hold...im gonna wait for a tv or something like that....wont be soon.    so today i went to go see The Village with a friend, and it was good. gonna right an official review later. tomorrow im hoping to take the MUNI downtown with some friends and just walk around and chill and talk and stuff....not too exciting. '),(25569,3382977,'18,July,2004','2004-07-18 00:00:00','   this summer so far has been very fun thankz to this tour...too bad it only happens every three years or so. anyway, i loved just about everything except for the plane rides (which im quite used to by now). The plane ride from SF to New York was pretty easy because the plane was not full at all, the last ten or so rows were totally empty, which meant that people could move back there and do just about anything, and since i was extremely tired i decided to put up the arm rests on the middle four chairs in the back and lay down like a bed, which was very comfortable, and i caught up on my sleep which i lost after waking up at 4 am. So we arrived at New York and i didnt feel all too happy, since i knew there was an even loonnggeerr flight up next. So the next flight from NY to Vienna was actually a lot better than expected. First of all, i had nonly once before been on a plane with little tvs in front of every seat, which was very nice, since i could play games and watch a selection of five different movies, and i could watch the progress of the flight, and i could look at the front of the airplane and the bottom view (thanks to two cameras in the front where the pilots are and one facing down under the plane) which were very cool while taking off and landing. I watched Welcome to Mooseport and a little bit of comedy stuff. didnt sleep much though.  When we arrived at Vienna, we had a good couple of days totally free to ourselves, and it was great. Vienna was my third most favorite city, next to Amsterdam and Paris. I had 300 euro to spend, but i was worried (i dont know why, since i had my credit card also) so i didnt buy that many souvenirs or post cards and stuff in Vienna, which sucks, i should have gotten a key chain or something. anyway, by the time i got to Paris i got a whole bunch of souvenirs and post cards, and i sent a bunch, even though i totally forgot about writing down people addresses, but i had a few memorized (i even sent one back to my house, where no one is since the rest of the family is in Italy, but theyll see it when they get home). I bought almost 20 postcards, most from amsterdam and Paris, and i sent about 8, so i have a bunch left over, which my mom is ok with since she doesnt care much about pics that i took if im not in them. I think there are about 5 pics with me in them, so i guess the post cards are ok to keep anyway.  The concerts were all pretty good. The best by far i think was the one in Paris. The worst probably the one in Rheingau Musik Festival, not that it was bad bad, it was just not as good as the others...but people still liked it. The two best places that we played for me was the first and last, the Musikverrein in Vienna and the Concertgebouw in Amsterdam, they are both insanely nicely decorated and everything is great. Ill post pics later. The bus rides were long but they were pretty fun, since there were people to talk and play and do stuff with, and we stopped at places and bought stuff. Really fun times. Took about 125 pics, and still have a lot of space, so ill get some pics of FL too.   The ride back was very interesting. First of all, we get to the plane in Amsterdam and we are about to leave and all of a sudden theres a problem with the Cargo door, which they cant close, so it takes a good 30 minutes to take out all the cargo, fix it, give it a test-run, and then load everything back in. Then we took off.&nbsp;I watched The Practice, four episodes of CSI Miami, Walking Tall, and some comedy stuff...which all took up a good part of the flight, and then i played some games for 10 minutes and slept the rest of the way. Somehow the flight got behind even more, so the plane landed about an hour and a half late...weird. The only problem was, we were given an hour and a half to get off the plane and go through customs and then check out bags again and do all that, but we were an hour and a half late....so we had about 5 minutes. But ofcourse the airlines is not gonna re locate 40 people, so they held the plane while we got our own lines through security and customs...like we were VIPS, which was ok. So we get on the plane a half hour late...and then the captain announces that there will be a good 2 hour wait...why? well, the taxi thingys that take the planes out to the runways were on a very tight schedule, which means that they had to skip our plane thanks to the YO coming late, which was because of the cargo door, and the taxi thingys had to go through all the others in line and then come back and get our plane. So after 2.5 hours sitting on the plane, we finnally get out of there and get in line for the runway, but then...there was supposedly a thunderstorm near SFO....yah right, so anyway, they close the runway and we go to another were we are in the back of the line....and then they open the other one when they realize there isnt a problem...so after a good while we are finally up...and the flight is long but not bad because i was pretty used to 8 hour flights...and this was about 6, not including the long time in the plane on the ground.  anyway, got some things to do today, like trying to get my pics out of the camera and building my model. post some pics and more stuff later.   '),(25570,3382977,'17,July,2004','2004-07-17 00:00:00','   well, now that in the past two days i have been on the plane and in the airports for a nice 21 hours approximatly (um, 8 hours from Amsterdam to Newark, 2 hours wait on the next plane while it waiting for a taxi thingy, then a 6 hour flight on that one from Newark to SF, then in the morning i got on a plane from SF to Charlette NC which was about 6 hours, and then 2 hours from NC to West Palm in FL). anyway, im pretty messed up timewise, everyone is going to sleep when i should be getting up, but im very happy to be here in FL and cant wait to do what im planning to do, like swimming ofcourse, and building my model of an airplane, and going kayaking and other stuff. gotta go through all the pics from europe and somehow get them to the comp or online. ill post details later   '),(25571,3382977,'14,July,2004','2004-07-14 00:00:00','   finally got enough time to get to the internet, yet i dont have much time. 5 darn euros for 15 minutes. anyway, im soon gonna go on a canal tour here in amsterdam, im having a ton of fun! so far the bus rides have been pretty fun since there is a lot of people to talk with. i have had tons of new foods (i even tried the escargo, which is surprisingly ok). So far my favorite city has been Paris, and the Hiltons have been very nice too, ezpecially the Hitlon Arc de Triumph, which is spectacular. I have taken about 120 pics on my camera so far, but only 5 have me in them, which are the only ones my mom really cares about, she says i should just buy postcards of things that i dont have a picture of me in. i bought a whole bunch (maybe like 20) but only have afew peoples addresses that i know, so i guess ill keep the rest for myself. only thing im not looking forward to is the long airplane ride ahead. The tour went by very quickly. only today and tomorrows concert and then the next day is flying home. met a bunch of people who i didnt know from before. hope everything is going well in the states. gotta go!  '),(25572,3382977,'09,August,2004','2004-08-09 00:00:00','      courtesy of clumsycrooks.com News source:unknown   '),(25573,3382977,'07,August,2004','2004-08-07 00:00:00','   well, thankfully i didnt have to go to that music thing that i explained in the previous post, and so i got to check out the new tv, and i got to watch Gladiator (for the first time) which was a great movie. I loved it. Besides the greusome stuff which really isnt that bad now, after watching the other movies i have seen this past two weeks, there is a lot of thrill and excitement in it, and it is just amazing, cant really put it in words at the moment. aside from the movie and the new tv and the games and stuff, nothing else is going on besides the party next door and me having to do my practice and waiting for tomorrow...which nothing is happening in either...so ill see what comes my way. hmm...nothing to do besides make a record of all my DVDs....nothing else to do around here....hah  Name.Date.How much i like it (1-10)  1. Matrix.1999.9 2. The Lion King 1 1/2.2004.? 3. Rat Race.2001.7 4. Spy Kids 2:The Island of Lost Dreams.2002.6 5. Ace Ventura:When Nature Calls.1995.6 6. Ace Ventura:Pet Detective.1993.8 7. Kung Pow:Enter the Fist.2002.4 8. Spongebob Squarepants:Lost at Sea.2003.? 9. Chariots of Fire.1981.8 10. Ice Age.2002.8 11. Kangaroo Jack.2003.5 12. Happy Gilmore.1996.8 13. Harry Potter and the Sorcerer\'s Stone.2001.7 14. Harry Potter and the Chamber of Secrets.2002.8 15. Die Another Day.2002.7.5 16. Finding Nemo.2003.9.5 17. Meet the Parents.2000.8 18. Pirates of the Caribbean:The Curse of the Black Pearl.2003.8 19. Behind Enemy Lines.2001.8 20. Black Hawk Down.2001.8.5 21. My Big Fat Greek Wedding.2002.8.5 22. Dumb and Dumber.1994.5 23. Three Kings.1999.7.5 24. Remember the Titans.2000.8.5 25. The Tuxedo.2002.7 26. Stargate.1994.6.5 27. Patton.1970.9 28. Spider-Man.2002.8.5 29. Star Wars (Episode 1):The Phantom Menace.1999.8.5 30. Star Wars (Episode 2):Attack of the Clones.2002.8.5 31. Tuck Everlasting.2002.8.5 32. The Master of Disguise.2002.5.5 33. A Knight\'s Tale.2001.6 34. Cop and a Half.1993.5 35. Ghostbusters.1984.7 36. The Rookie.2002.8 37. Crouching Tiger, Hidden Dragon.2000.8.5 38. A Beautiful Mind.2001.8.5 39. Bruce Almighty.2003.6 40. Top Gun.1986.9.5 41. Shrek.2003.8 42. The Terminator.1984.7 43. Terminator 3:Rise of the Machines.2003.8.5 44. Men in Black.1997.8 45. Men in Black II.2002.8 46. Lord of the Rings:The Fellowship of the Ring.2001.8 47. Lord of the Rings:The Two Towers.2002.9 48. Lord of the Rings:The Return of the King.2003.9 49. Charlie\'s Angels.2000.7 50. Conspiracy Theory.1997.8.5 51. American Beauty.1999.8.5 52. XXX.2002.8.5 53. Robin  Williams Live on Broadway.2002.? 54. Porky\'s.1998.7.5 55. Batman.1989.7.5 56. Batman Returns.1992.8 57. The Real Eve.2002.? 58. The Crocodile Hunter:Collision Course.2002.6 59. About a Boy.2002.8 60. The Pelican Brief.1993.6 61. Divine Secrets of the Ya-Ya Sisterhood.2002.? 62. Daddy Day Care.2003.8 63. Holes.2003.8.5 64. The Right Stuff.1997.? 65. Scooby-Doo and the Reluctant Werewolf.2002.? 66. Scooby-Doo and the Ghoul School.1998.? 67. U.S. Marshals.1998.6.5  well, now that i have that done....i can go order more and just use up more free time watching them and now i gotta do my practice.   '),(25574,3382977,'05,August,2004','2004-08-05 00:00:00','   you know what really sucks? firstly, i have the head of school as my advanced english teacher...i just found out a couple days ago, and i just ordered all the books. If only i was in normal english II, which has the teacher i was hoping to have, who is a lot cooler and not as strict at all as the head of school, who is extremely strict, since we had him for one class in religion last year as a sub. Now that thats off my chest....another thing that sucks....yesterday i had to go with my mom and youngest brother and his annoying friend to Costco, and on the way had to pick up some Lithuanian girl whose living this summer with a Lithuanian teacher (who used to be my teacher for a piece that i had to play). So i walked up to the door, and the Lithuanian teacher asked me if i could go to some stupid music at menlo festival concert...which was at 3 tomorrow (which is today unfortunatly), and i off the top of my head came up with the lame excuse of \"well, i dont know becuase i have lots of movies that we rented to watch tomorrow....\". yes, i know, it s pretty lame and probably in her eyes stupid becuase she could care less about movies that i rented....anyway, so i finally say after some awkward silence that i might be available. i asked how long i would be gone, she said about 40 minutes, then paused, and then said to get there. ok, so the concert is about an hour long, and it takes about 80 minutes to get there and back. well, what a great way to enjoy my saturday, spend a nice 2 hours and a half riding in a car and watching a concert with some Lithuanian teacher. at the same time, my neighbor is getting a \"big\" screen tv (it would be more accurate to say \"bigger\" screen tv). and the time to be watching it is today at 3, which is the time im watching a stupid concert...oh well.  things are a tiny bit better and someways worse with my middle bro and sister gone. first of all, the noise level and annoyance level is very low. secondly, i have to do some easy yet teadious things around the house that waste my time, but then again, their easy. I have been left alone a lot more than i thought i would be....besides having to empty the dish waaher and clean and make my bed. Yesterday i went and rente another batch of movies, and found out that i have rented (or the account has rented (which is 95% me renting and 5% other family members renting)) a little over 200 movies...anyway, so i returned 21 Grams and Silence of the Lambs (which wasnt that scary afterall, more disturbing) and Pretty Dirty Things and 13 going on 30 and Confessions of a Teenage Drama Queen (which i didnt watch along with the one before that ) and Greese (which i havent seen nor did see yesterday). I got Gladiator (meant for the big screen, and another one i havent seen), Austin Powers 1 (for littlest bro), Stagmata (a scary and disturbing one that i might not watch) and The Whole Ten Yards (which i watched last night and turned out to be pretty good, review coming later, after Manchurian Candidate and Monster\'s Ball).  im gonne post a list of all my DVDs....just for the heck of it. and i might update my classical cd collection list later. how pathetic i am...humph   '),(25575,3382977,'05,August,2004','2004-08-05 00:00:00','   urlLink    This is the stage at the Roman amphitheatre. You can see behind the orchestra the original pillars that were around back when the romans did their thing. pretty cool  '),(25576,3382977,'04,August,2004','2004-08-04 00:00:00','   urlLink    this is the old Roman ampitheatre in Lyon. the audience sits on the hard rock while we play on stage (they pass out little sitting pads).  '),(25577,3382977,'04,August,2004','2004-08-04 00:00:00','   urlLink    umm....salzburg again??? i gotta get to know the difference between salzburg and strasburg. didnt spend that much time in either of them. thats my lame excuse  '),(25578,3382977,'04,August,2004','2004-08-04 00:00:00','   urlLink    Well, i think its Salzburg.  '),(25579,3382977,'04,August,2004','2004-08-04 00:00:00','     21 Grams is a mix of different stories that gradually come together into a very powerful tale of tragedy (and a lot of it). Now when i say gradually, i mean gradually. The movie is a little over 2 hours long, and it takes a good hour to finally get a grip on things. The movie is as if the filmakers took 75 different scenes from random points on the storyline, all about 3 minuters in length, and jumbled them all together into one messy clump. I know that sounds reaally bad, but it actually isnt once you have figured out that it is what it is. For people that need a more linear storyline and more explanation, the first half of the movie will be way over your head. It took me a solid 30 minutes of mixed up scenes to really understand who is who and when is which scene happening. I was forced to ask myself several times, \"is this in the present?\", \"is this is the past?\" and \"is this going to happen?.\" Now, im not saying that you will definetly have a hard time figuring the movie out, but it will at least seem strange for a bit.  21 Grams at first is jumping backwards and forwards in time and jumps from character to character in a big frenzy. But sooner or later, from what seems to be a crazy approach to a movie, a pattern starts to come out. These \"random\" scenes are all pointing and narrowing down to key scenes and moments where they all converge into a somewhat linear plot. By the end of the movie, everything is pieced together and everything that was supposed to be expressed has been with amazing acting. Besides the jumble at the beginning (which once you have seen it doesnt seem all too bad, but a pretty good way of getting the audiences attention), the acting and storyline was incredible. 21 Grams is a film that will keep you thinking about it for a good amount of time. I give it an 8 out of 10.   '),(25580,3382977,'04,August,2004','2004-08-04 00:00:00','     Magnolia is a three hour long movie which features several mini-plots (or mini movies almost). Throughout the whole movie, these storylines go on and occasionally interconnect with one another. This makes the movie verryy hard to follow at somepoints. When this interconnecting and mixing actually works out, it is wonderful and amazing. at some points, the plots are so far from each other that you can get pretty annoyed, but if you hold through, and keep everything fresh in your head, and not forget anything, you will love it, or almost love it.  In order to understand anything in Magnolia, you have to remember the characters. Throughout the movie, the \"leaves and branches\" of all these characters and their stories start to come together and make sense. Another thing is how different the scenes are, which kinda keep you at the edge of your seat, at least at the interesting parts. Some scenes are long and unbroken with slow proggression, and some are really fast with many closeups and quick cuts. This is just one of the many details of the movie that keep it from being a loong and confusing wreck.  So, Magnolia is a movie about chance. In the beginning there are three different crazy, yet true, stories of almost unbelieveable things happening. The point of the three stories is that no matter how crazy or strange or how freaky some happening might be, its all because of chance. So for the next 150 minutes, the movie expands upon that idea, except with many different storied which are a lot longer and more complicated and complex than the first three. Overall, the movie makes you think, and wonder about different things. Its not a movie which you can watch and just switch right away to another one. It takes a moment to sort out everything that has happened and realize the importance, at least, it took me a while to figure it out. I guess it would help to view it again ( maybe even twice to really understand it ), but its one hell of a long movie to need to watch again. And not only is it 3 hours, but its a long three hours, as in theres not much that happens. Its not a thriller that keeps you on the edge of your seat. Its more of a sit back and try to understand it movie. I give it a 9 out of 10.   '),(25581,3382977,'02,August,2004','2004-08-02 00:00:00','   just got the pics from Ofoto, and they are awesome. for a nice $25 i got 6 great looking 8\"x10\"s and 2 5\"x7\"s and 4 wallet sized pics from the euro tour. wonderful. i gotta frame some of the 8\"x10\"s and hopefully i can get them to go into my room, while mom is hoping for two of them to go into the library. anyway, i had my first lesson in a month, which went ok. im still not sure what the requirements this season for YO seating auditions are, so im assuming there is a solo piece and stuff needed, until i get the official word from someone or something (presumably a flyer).  My sister and middle brother are going to FL, finally. I actually wanted to see them after my venturing around the world, but now im almost getting used to them, so i dont mind them going away for a good 9 days or so. They are frantically packing since they gotta go at 7 tomorrow morning. Might get kinda lonely now that i think about it with a nine year old (my youngest brother) all day, but hopefully he will be occupied with his little friends that come over to play videogames...  Last night i watched Gothika, which is a pretty scary movie. I got to admit, im not all too fond of scary movies, and believe it or not, sometimes they get to my head. But it wasnt that scary. More suspense and thrilling, and scary too. i also rented 21 Grams, which i havent watched at all yet, Silence of the Lambs, which i will watch with the neighbor down the street when she gets her big screen tv, and Monster\'s Ball, which i heard was good but havent watched yet. These are all due friday at 8, so i have time. My sister rented Grease, which i have never seen, but maybe will if i have time. Silence of the Lambs is supposedly extreeemmeellyy scary....so im not sure what ill do after watching that. I also have been ordering a lot of stuff from this practically free giveaway DVDS site, which i have gotten 6 DVDs from so far, for a whopping 2 dollars and some cents. there is a catch though. I gotta order 5 more dvds at \"normal\" price (which is pretty cheap compared to stores) in the next 2 years. So today i ordered 6 more in another name but on the same credit card, and i realized that as long as its the same credit card i gotta order 5 more at normal price in the next 2 years. so in all, i gotta get 10 DVDs from the site at \"normal\" (about 10 to 20 dollars) price in the next 2 years....which isnt bad, considering that i got 12 DVDs for 5 almost 5 dollars.  so, besides all the movies and practicing, i got tons of cleaning and reading to do. One good thing is that i ordered a ton of new cds for my \"collection\". i think about 15 in all (almost 200 dollars worth). good thing i got my own credit card. (:   '),(25582,3382977,'02,August,2004','2004-08-02 00:00:00','   this past week i have watched probably the most i have ever in one week, not just movies at the movies, but renting movies aswell. its actually really fun, except that theres almost no time to watch all of them. Today i watched The Manchurian Candidate, most of the beginning of Notting Hill, and some of Magnolia (now we are about half way through it). We finished EuroTrip last night and took it back today and got 21 Grams, which is due Thursday, and we havent watched any of it. Magnolia and Notting Hill and My Best Friends Wedding are due tomorrow at 12, so we gotta watch those tomorrow morning. I saw the Village and The Bourne Supremacy also this week. Im hoping to rent Fight Club soon after we get rid of the ones we have. besides movies, there has been a lot of practicing and cleaning going on. my mom seems to bet into a bitchier and bitchier mood everyday...im getting reaalllyyy tired of this Friday Hotel gig. besides that we went to the YMCA at about 8 and did some gym stuff, and then checked out the pool and did some laps and played and stuff. now my eyes sting like hell and my dad has just rampaged into my room after i was supposedly \"going to bed\". luckily i heard him coming down the stairs. he gave me a little \"be more good\" speach, probably because my mom just told him upstairs about todays little fiasco this morning with the burner...well, im in bed and its about 9 in the morning and she pops the door open and yells \"erik, get up and eat your breakfast which is ready right now...and then go do your practice, etc. \". so i get up and find my food on the stove under some foil, so i figure that everythings off (i didnt really think about it becuase it is always off anyways if moms not there) and eat my breakfast and get to the millions of things my mom wanted me to do. so she comes home and tells me to get all the crap out of the car (she went shopping) and then she comes out yelling that the burner is still on (low enough that my French toast wouldnt burn) and that im very irresponsible and all that jazz.....so im figuring that the days plans are screwed and i might as well go and twiddle my thumbs....but luckily my sister and her friend from Lithuania want to go see a movie so my mom lets me and some people go too. anyway, so thats what started the whole grumpy and bitchy attitude for the day.    '),(25583,3382977,'01,August,2004','2004-08-01 00:00:00','   besides having to go to church and acolyting and stuff, today was a pretty fun and free day. the plan was to get up at 7:45 and get down the street to a neighbors house by 8 to watch the movie Magnolia, which we were trying to do the night before before my brother and i were rudely interrupted by an unwanted phone call from my mom telling us that it was time to do some practice and that the place was being \"locked up\". and it was 9:00 pm....and i thought it was summer vacation...anyway, so my brother came in at about 8 all ready to go and i said screw it and so that was that, and we woke up a nice hour and a half later for breakfast and church.  so after church our plan was to go downtown and explore, so because i knew that wasnt gonna fly with my parents (very slim chance that we could have if i explained exactly what we were going to do) i decided to just go to my dad and tell him that the three of us were gonna explore a bit for a couple of hours. ok. so we go downtown and check out the Apple store and by some stuff at Virgin and eat and stuff. not too bad.  so we come home and watch EuroTrip, which was somewhat fullfilling. not a very complicated and meaningful movie but it did give plenty of nudity, swearing, comedy and whatnot. thinking about expanding the horizons of our explorations, maybe to Sausalito....i dont know what to say to the parents about that....we\'ll see   '),(25584,3382977,'01,August,2004','2004-08-01 00:00:00','     This movie was pretty good. After watching it i wasnt too pleased about it but thats only because i had high expectations for it. Now that i look at it as just a normal movie it isnt too bad. Overall Anchorman is a good movie.   The movie is about the behind the scenes action at a news station. It is not a serious movie at all. Anchorman is more of a silly and over-the-top movie than i expected. Yes there are some funny parts, but most of it is just sillyness and lame jokes mixed into a comical environment. Will Ferrel is the perfect man for the part of Ron. Like his roles in Elf and Old School, he is just the type of silly and somewhat comical guy for the part. Besides the weird scenes and somewhat awkward jokes and over the top silliness, Anchorman does have its hilarious and pretty funny parts. I give it a 7 out of 10.   '),(25585,2693059,'31,July,2004','2004-07-31 00:00:00','   song  Devin the Dude - Right now   bored..ate a big ass breakfast today then went to funcoland to return the broken controller i bought for a new one. Then i sold 2 games to them and bought  Freedom Fighters  and  Manhunt . Freedom fighters is good and if ne1 wants manhunt ill trade them for a diff game. Then after that went searching for dressers at all those cheap furniture stores in downtown hayward, didn\'t find anything but expensive cheap furniture i would rather buy good shit for the same price. Yea now im sitting in my room online listenin to some Devin the Dude. Nothing else todo...so how are u doing? Oh your silent, i understand. FUCK YOU THEN BITCH. Yea so ima sit and just say some stuff cause theres nothing todo and im bored like i have already stated. Feelin soo wonderful. Did alot of calf raises today cause my knees don\'t hurt that much so i can start dunking easy again and by season, someone will get dunked on. Today while i was looking for furniture a  BIG  ass plane flew over my head, i swear if i wasn\'t next to the airport i would have been extremely frightened . Im just gonna list my goals for this schoolyear.      Get straight A\'s to raise my GPA     Dunk on someone           Get a Job     Windmill     Get accepted into a mediocre college  take the slang word   hella   out of my vocabulary because its stupid and i don\'t know how i got caught up in saying it                                                 -and the list goes on......  '),(25586,2693059,'29,July,2004','2004-07-29 00:00:00','   song  Anthony Hamilton - Charlene     summerschool again yea....made pancakes fucked them up.....ate too much damn food....what else.....i want some ihop, 40 pancakes, 20 sausages, 40 pieces of bacon, 10 sunny side up eggs, 5 steaks, big ass 9 liter glass of orange juice... yea im hungry  '),(25587,2693059,'27,July,2004','2004-07-27 00:00:00','   song  nelly - my place    finally all that basketball stuff is over and i get to rest some.  Went to summerschool today then came home and went to southland togo job searching, only got a few apps and i need to make a resamay so i can apply for some jobs.  After that stuff I saw Anchorman, that shit is one of the funniest movies ive ever seen.  I just might see that twice.  Yup thats all i did.  '),(25588,2693059,'25,July,2004','2004-07-25 00:00:00','   song  J-Kwon - IC IC    bored....  '),(25589,2693059,'23,July,2004','2004-07-23 00:00:00','   song  Zapp and Roger - Computer Love    well my wish came true, i did kinda good.  We had 3 games yesterday we lost 2 and won 1.  In the game we won i had the game winning shot.  Fuckin lost to weak ass teams though.  I played pretty good in all of them.  Today we had 2 games we won both.  I did good today too.  Can\'t wait till season now, ima bout to lift hella weights and bulk up and practice even more now so i can kill in season.   Right now im tired as hell but thats ok since we dont have any basketball stuff in august.  Im just going to rest for a few days and do nothing till my body feels good again.  Then i will get to practicing .  I also only got 1 1/2 weeks left of summerschool can\'t wait.  Now i can put my wow card to use.  I paid 40 somethin bucks and i havent even used it yet sooo if u going to GA tell me.   Its amazing how doing good in a few basketball games can make you feel so much better.   gotta take a shower now so ima stop writing.  '),(25590,2693059,'23,July,2004','2004-07-23 00:00:00','   song  jadakiss - still feel me    went to summerschool, goddamn im losing all my patience in that class and i think my teachers gonna go columbine on us.  Had 2 games today first game we won 85 to 65 second game we won 100-51 we played the weakest team ever.  I couldn\'t get my goddamn dunk, fuckin pissed.  knees are not working and i got 3 games tommorow. Until then im too lazy to write Hasta Luego  '),(25591,2693059,'22,July,2004','2004-07-22 00:00:00','   Song  R.Kelly - Feelin on your booty    no motivation todo anything. i always fuck up on any goal i have. theres no point wasting my time in school anymore and failing classes and doing all this bullshit for what? Tobe dissapointed not get in a good college then goto chabot then just say fuck everything get some weak ass job and become a bum and beg for money in SF. All im doing is prolonging my future. I might as well start now instead of later. Im so damn tired of trying and just being defeated in every damn thing i do. Nothing goes right for me. I don\'t understand why not, i lose in everything, i wanna know what success feels like. Or accomplishing something. how can i fuck up at everything in my life its just not right.  im so tired of being kicked and kicked while ive already been down.  '),(25592,2693059,'19,July,2004','2004-07-19 00:00:00','   Song  Lloyd- Sweet Dreams    sleepy...went to summerschool again wooo...after that i ate at dairy queen and i got their fried chicken, those cheap asses, that shit tasted like tv dinner chicken or some shit. After that just rode straight home. Right now just icing me knees and chatting..nothing else todo till i get my Nba Live 2004 back at 3:00 PM. I gotta tourny tommorow its the last one. 2 games friday 3 games sat and 2 sunday. Hopefully we will do good, then i will get a month of rest,pumping iron, and individual practice. cant wait to play some live, i haven\'t played in so long ima be hella weak but i dont care. I shall update afteer i play some live.  '),(25593,2693059,'19,July,2004','2004-07-19 00:00:00','   im finally bak from that trip, i thought it would be great but no.  El Dorados buffet was good though.  We went like 4-1 or something in the tournament.  Got no sleep or ne rest.   Nothing on tv either.  I had the most juice ive ever had in my life while in Reno.  I remember getting like a sobe and arizona tea twice a day and downing them in like 10 seconds cause i was always hella thirsty.  Today just did summer school again was a little late cause i really didn\'t feel like riding my bike there.  I broke my record of 20 mins riding my bike home, i did it in 13 mins, tommorow ima try for 10.  Doing pretty bad in basketball right now i dunno wtf is up with me, i have 0 energy and all i wanna do is sleep all day.  This summer was supposed to be my worry free, great and fun summer but its my worst and i can\'t wait till next summer when i can be fillin out papers and stuff for college then finally leave.  '),(25594,2693059,'14,July,2004','2004-07-14 00:00:00','   song lloyd banks - warrior pt. 2  did el usual Summerschool then game we played Castro Valley.  My legs are gone i dont think i can get net right now.  Some miracle is going to have to happen for me to have energy for the Reno tournament cause i get about 0 days of rest.  Feelings kinda sick right now cause i drank a pitcher of kool-aid.  Tommorow gonna get rdy for Reno, buy some batteries, food and other stuff.  Only if i had my damn camcorder for this trip......... but im broke.  yea yea yea no need for talking about personal stuff, im sick of writing in this.  '),(25595,2693059,'13,July,2004','2004-07-13 00:00:00','   summerschool then workout at tennyson woo.  Excited for Reno tournament on friday.  Watchin some funny ass family guy.  Body is as dead as ever.  Woo party tommorow in summerschool,  i get to fatten up on some pizza.  well thats the end of that chapta  '),(25596,2693059,'12,July,2004','2004-07-12 00:00:00','   hmm i don\'t know what to say... i just sat here and wrote like 9823 things then erased it.  fuckin tired did the usual summerschool then game shit.  Reno trip comes up Friday and if i dont get my mind right ima fuck up some more there.  Im starting to notice how all the badass undeserving people who lie and do all this bullshit get everything they want.  And all the good people don\'t get shit.  I hardly know ne good people with good lives and if you know alot  then hahaha they probably aren\'t and your just a blind dumbass.  Good people don\'t get second chances the bad ones do.  So i suggest we all become sleezy ass losers...that is the true path to becoming successful.  '),(25597,2693059,'11,July,2004','2004-07-11 00:00:00','   hmm nothing todo.....look my wonderful desktop  '),(25598,2693059,'10,July,2004','2004-07-10 00:00:00','   Had like 4 games today lost in finals. Team did good most of the games. I PLAYED LIKE SHIT, fuckin embarressing.  Tired real tired and i got summerschool tommorow and another game.  Yea im lazy.  '),(25599,2693059,'10,July,2004','2004-07-10 00:00:00','   woke up went to basketball game at 10 we won then came back home then left at 2 to basketball game won again then came back.  Ran Ad-Aware on parents comp, only found about 690 items of spy-ware on it.  Spent a little bit installing antivirus software and other crap on it to rid it of its demons.  Then bbq\'ed some kabobs and chicken.  Iced knees off an on , now online great day.  '),(25600,2693059,'09,July,2004','2004-07-09 00:00:00','   soooo fuckin tired..... woke up at 5:00 then went to doctors appointment at 7:30 for my knees only to be given more exercises and medicine.  After that i went to summerschool till 12:35, then rode my bike from there as fast as i could to get to tennyson before 1 to get a ride.  After that went to game in pleasanton we won both of them.  I was tired as hell before i even started playing.  Then got back at about 5:30 only to have a nice little 30 min rest then on to doing the scoreboard from 6 to fucking 10.  I have gotten a combined 10 hours of sleep probably in this last week.  I got no more energy for tommorows game at 11 AM.  Just gonna go and do the best i can.  World is also getting worse everyday.  Everything around me is getting worse. Life has gone from terrible to fuckin  i dunno but life is just fucked up.  '),(25601,2693059,'08,July,2004','2004-07-08 00:00:00','   Song Bone Thugs N\' Harmony - Resurrection  Did the usual, summerschool then practice.  Iced knees, played some ESPN and thats about it.      Just pissed off and don\'t got much to say.  Tourny tommorow in Pleasanton then i gotta do that damn scoretable stuff for Jv Tournament from 6 - 10.  Feeling sick like i gotta throw up but i can\'t. yea that is all...  '),(25602,2693059,'07,July,2004','2004-07-07 00:00:00','   song Musiq - Soulstar  Went to my doctors appointment at 7:45 to get my knees checked but he canceled last minute.... Then went to McDonalds and had a McGriddle meal, pretty good then off to summer school.  After summerschool i had a game but the other team didn\'t show so we played ourselves.  When i got home i played some more ESPN Basketball.  Went to play in Fremont again... Then went to Pepsi park and played a few games there and held the court till like 9:30 and thenwe left.  Overall pretty good day just went with the flow of things.  '),(25603,2693059,'06,July,2004','2004-07-06 00:00:00','   I don\'t care what you think.  Im not a professional critiquer       -Chronicles of Riddick - this movie was terrible.  Plot was retarded actors retarded everything.  Tell me how they go on a planet that is 700 degrees hot in the day time and the go out and simply hide in the shade so they won\'t get burnt.  Since when does shade protect you from 700 degree weather?     -White Chicks - a few funny parts but most of it was just weird.  Everything kind of just fell apart at the end.  Somehow this movie reminded me of Ms. Congeniality.     -Spiderman 2 - bullshit most of you will like this cause your idiots.  Retarded ass drama everything, Kirsten Dunst is hideous i dont wanna have to look at her for 2 hours.  Just a 2 hour soap opera.  '),(25604,2693059,'05,July,2004','2004-07-05 00:00:00','   Song 2pac - All about you  Today was ok just went to summerschool then hoop camp. Hoop camp was pretty tiring, at least im getting in some kind of shape.  Got home played some ESPN basketball again and whooped some people.  After that went to a park in fremont to play, we won all our games not much comp and i didn\'t really do nething but walk up an down court cause my knees were sore.  Yup thats all i did.  '),(25605,2693059,'05,July,2004','2004-07-05 00:00:00','   song  Slum Village - Closer ft. Dwele  Didn\'t have togo to summerschool today cause of 4th of july. I went to my game at 2 played arroyo and we won.  WHen i got home just did the usual and played some espn basketball and listen to music.  Went to the mall bought some shorts that were too big.  Im too lazy to take them back though so ill just eat some burgers and grow into them.  Woo summerschool is tommorow get to wake up early and ride my bike there.  Yea well its time togo to bed now cause im tired.  '),(25606,2693059,'10,August,2004','2004-08-10 00:00:00','   ..patience..  '),(25607,2693059,'08,August,2004','2004-08-08 00:00:00','   song The roots - stay cool  fuckin wasted my fucking money again.  I bought a ps2 network adapter thinking i could hook it up to my ethernet card and use the internet connection from my wireless network(maybe i still can if i use windows xp as a router but that is way way way waaaay too much trouble and im lazy) but noooo.  So noooooow i gotta buy a wireless ethernet bridge so i can play my ps2 games online using my wireless network.....  The wireless bridge is like 80 bucks.....even on ebay.  I could have fuckin bought a wireless network adapter for my ps2 for 70 something by itself and now ima end up paying like 130 cause i always make the wrong decisions about everything.  The worst thing is i knew there was a chance of it not working but i went ahead and bought it anyway.  Oh well, since i always fuck up on everything eventually something great is going to happen to me and everyone will be jealous of me.  Ima fuckin win 1 billion dollars or be the 1st man on mars or somethin, its going to be great.   Tommorow im going to Great America and im finally going to put my WoW! card to use and no you cant come cause you will fuck everything up and bitch about where to eat and how you dont wanna go on any rides and sit to the side making everyone feel sorry for you, i hate when people do shit like that, just dont go.  Ima go and ride every single ride and go on the skyflyer as much as i want.  Im done  '),(25608,2693059,'08,August,2004','2004-08-08 00:00:00','   song Mobb Deep - Got it twisted (remix) pretty tired... went to some chinese buffet for lunch hella crowded, there was like 99823 people coming straight from their churches. Food was good at least, i had like 90 baby octopus\'s. After that i did the usual...lay on my bed watch tv and eventually fall asleep. Then when i woke up i went to play basketball, im not gonna be able to walk tommorow cause of that. yea im done..  '),(25609,2693059,'07,August,2004','2004-08-07 00:00:00','   song Brand Nubian - Young Son  bored.. I just ate a tasty baked potato with full strips of bacon not bits.... Cheese melted all over , sour cream, chives, butter, pepper, it was an orgasmic experience.  I tryed to play more live but i gotta headache.  Yea im done for now.  '),(25610,2693059,'06,August,2004','2004-08-06 00:00:00','   USA whooped Serbia-Montenegro.  Today just sat in room and played Nba Live and Freedom fighters.....what else? nothing...  '),(25611,2693059,'06,August,2004','2004-08-06 00:00:00','   song 2pac - fuck friends  bored...hungry...  '),(25612,2693059,'05,August,2004','2004-08-05 00:00:00','   song 112 - say yes  took like 80 naps today, and alooot of icecream.  Shot around at the park.  Hmm just playing some NBA live 2004 right now.  Feeling sore just from shooting around for some reason.  bored....  '),(25613,2693059,'02,August,2004','2004-08-02 00:00:00','   songs Too Short - Gettin it Ginuwine - Same ol\' G  im so bored everyday its just not funny anymore.  I started to read...................... Right now im reading Dennis Miller\'s - The rants.  This book is pretty damn funny just like his show on HBO.  Right now my back is actin up, i dunno whats wrong with it.  Today just went to summerschool then came home and had a big ass headache.  I went online a little bit and ate then i took a nap.  At 5:45 went to my basketball workout, i didn\'t wanna go cause i had a headache and was  sleepy but i need to practice.  Until next time....  '),(25614,2693059,'01,August,2004','2004-08-01 00:00:00','   song same one....    Yea im feeling good right now. My favorite boxer mike tyson lost because he tore his knee ligament in the 1st round. He still managed to win 3 of the 1st round but by the 4th it was difficult for him to throw punches. I knew there was no way Williams could beat him even though he has over 30 pounds of muscle on Tyson. If Tyson didn\'t tear his ligament Williams would have been knocked out in the 2nd round. I just hope Tyson can heal up. Now i feel like playing some Mike Tyson punchout for nintendo. Freedom Fighters is continuing to get funner and Manhunt is continuing to piss me off so if ne1 was to trade me a game for Manhunt tell me.  Im tired of people telling me they wanna see that weak ass catwoman movie too.  That shit is going to get the worst reviews ever.  People expect it to be like spiderman and daredevil and all that shit and follow the comic but noooo...  They completely remade costume,  that is the ugliest costume ive ever seen.  What the hell were they thinking.  She looks like a biker chick or somethin.  That costume in no way resembles a cat except for the mask.  Why didn\'t they just use the costume from Batman Forever or one like it.  They also changed the whole story, that should be illegal.  They basically took the name and idea of a woman who has Catlike attributes and fights crime and made it into shit.   Im not going to see the movie for the simple fact that the costume is retarded.  Also the funniest movie ever is Anchorman.  They should have done a better job advertising and it should have had a different name.  The good thing about that movie is that they don\'t show all their funny parts in the previews.  That movie is one of the funniest i have ever seen in my whole life.  The first 20 minutes of it i was regretting i bought the ticket for it.  But after the retarded jokes and dull parts the movie made a 360.  I believe this was done intentionally, but it could have backfired because i was about to leave the theater and get a refund.  You have togo see that movie over anything else.  Its better than everything in the theaters right now.  I wanna see that shit again.  Yup ima play some Live now.  '),(25615,4186455,'20,August,2004','2004-08-20 00:00:00','   today i\'m happy, play all day.. never do works, tommorow then suffer :D this evening i watch AVP with my friends. it quiet good movie. and today also abit sad cause cannot watch the fashion show, but nvm sunday monday i can wacth.. hhmm.. now i watch indonesian idol, that delon must out!!!!! wait ok.. have to disconnect for a while.. hehehe...  '),(25616,4186455,'16,August,2004','2004-08-16 00:00:00','   hhhaaiiizzzzzzzzz.. yesterday.. hmm.. no until now.. i feel abit headache.. i think will get cold soon, but hopefully not.. i drink vit C and drink alot of water.. rest alot.. lately this week i always sleep late.. all my fault... what else if not playing that stupid pt... but how.. i feel like addict to it (wink) hhmm.. today i do some drawings.. and prepare for sewing my skirt later.. hopefully everything going to be ok.. hah.. i think i must cut all the cloth now.. or else later no time.. ok.. see you... next time i tell you something interesting better then all this pathethic sad stories.. hatttccchhhhttyyyyyy....... ZZzzzzz................  '),(25617,4186455,'15,August,2004','2004-08-15 00:00:00','   Later that day I got to thinking about relationships. There are those that open you up to something new and exotic, those that are old and familiar, those that bring up lots of questions, those that bring you somewhere unexpected, those that bring you far from where you started, and those that bring you back. But the most exciting, challenging and significant relationship of all is the one you have with yourself. And if you can find someone to love the you you love, well, that\'s just fabulous. - Carrie Bradshaw  '),(25618,4186455,'15,August,2004','2004-08-15 00:00:00','   Now its sunday morning.. yesterday whole day i only play pt all the time.. only do some of my works!!! god!!! i don\'t know why i\'ve become so lazy like this.. i feel like i\'m loosing my spirits in this fashion study.. i feels like i have no friends.. i feels all alone, help me!!! i\'m really lonely my friends i don\'t know where all they\'ve gone. do do do my works must done... i have to be tough and can stand on my own two feet. keceeppiiaannnn...... hhaaiiizzzzz...... i need someone to really care and always beside me one. i have 2 friends that i consider my good friends, but lagi poor me.. i don\'t think they consider me as their good friends cause i see they always left me behind... what a life.. i\'m so pathethic at the moment.. so fragile and HHAAHHHHH...... stop typing do works do works!! ok guy whoever there just wish me luck!!! Amen.. ^_^  '),(25619,4186455,'10,August,2004','2004-08-10 00:00:00','   today is my 1st time i post a new blogs... its 10 august 2004, and i don\'t know want to do. i just woke up.. mmaadd!!!!!!!! how come i woke up so early like this T_T  i need rest!! so tired.. but what can i do? i can\'t go back to sleep. so i turn on my comp and i see this interesting things. so i give it a try.. never write a diary before.. (sigh) i feel my world is a mess today.. alot of works to do. i scare i cannot make it.. but must.. it a must!! i can make it!! draft patter, 4 skirt, drawings.... BWAHAHAHAHAHAHAHAHAHAHA.... MAD MAD I\'M MAD!!!!!!!! ZZzzzzz........  '),(25620,2673530,'26,April,2004','2004-04-26 00:00:00','   urlLink    Handing over the cheque   From the football tournament and recent donations, the new grand total is Â£1485!!  '),(25621,2673530,'19,April,2004','2004-04-19 00:00:00','   Last week (Wed 21st April 2004) we held a friendly 5-a-side mixed football tournament at work, for staff, PhD students and friends of the School of Civil Engineering and Geosciences (CEGS), Newcastle University.    urlLink    Stuuuudents.   Ten teams turned up, and in total 60 people were there to play and support.  All players and supporters made a contribution to the cause, and are now official members of the Bellavista Social Club (BSC).  A total of Â£117 was raised.    CONGRATULATIONS to the overall winners, Young Ones (Matt, Kate, Rory, Oliver & Simon D), who were presented with the winners trophy, certificates and mystery prizes.    COMMISERATIONS to the overall losers, Hero (Lesley, Jon, Karen, Michelle and Adam), who received the coveted losers medal, certificates and mystery prizes.    THANKS to all the teams (Hero, Tripoli, No Name, Water A, Water 1, Young Ones, Toxic, Molecular Maestros, Geomatics, and Leftovers) and supporters.  Particular thanks to those whoâ€™ve hardly ever played football before.  Hope you enjoyed it.  Remarkably, there was only one serious injury and a few aching legs.    SPECIAL THANKS also to the staff at Claremont Sports Hall, David Parker for agreeing to pay for the hire (enabling money collected to go to Bellavista Juventud FC), Heather McGrath for her excellent firstaiding, Rob, Paul, Sean and Russell for being refs at short notice, and Lesley Batty for arranging the CEGS night out afterwards.    Charlotte Weatherell, 26th April 2004.  '),(25622,2673530,'19,April,2004','2004-04-19 00:00:00','   urlLink    Awotchawow...   The first meeting of the Bellavista Social Club (BSC) at the Egypt Cottage last Saturday had a great turnout.  Congratulations to the four lucky raffle prize winners: Laurie Vicquelin (brand new Bellavista Juventud FC shirt as sponsored by the BSC), Conrad Scott (old strip signed by Nobby Solano), Michelle Taylor (Peru national shirt, also signed by Nobby Solano), and Chris Jewitt (Crunchie Easter Egg!).   urlLink    The Egypt Cottage&nbsp; urlLink     From ticket sales, donations and raffle, a GRAND TOTAL OF Â£1318 has been raised.  A cheque was presented by the BSC President, Doug Paterson, and accepted on behalf of Bellavista Juventud FC by special guests, the parents of the FC President Father Joe Plumb.  THANKS to all of the BSCâ€™s 185 MEMBERS for their generous contributions.     urlLink    DJ Conrad  Special thanks also to Conrad Scott for his groovy tunes, Wendy Parker for making us a banner, Stew for making the presentation cheque, Father Joe and his parents for their messages from Peru, the Egypt Cottage for the great friendly venue, Nolberto Solano and Aston Villa for agreeing to sign shirts, Peter Warriner for creating our original website, Kim Smith, Steven Plumb and Nick Kemp for selling lots of tickets, Benet Biscop Young Enterprise for donating some of their hard earned profits and, last but not least, Steve Jackson for making tickets, setting up this weblog, general PR and lots of help besides.  WATCH THIS SPACE for news of the team in Iquitos and further BSC events in the Toon.  '),(25623,2673530,'16,April,2004','2004-04-16 00:00:00','   Bellavista Social Club Membership List @ 19th April 2004   doug and charlotte steve biff chris and sophie richard john and lucy stew conrad jen steven and emma sue and jamie tim and corrine keith d adam claire oliver rob and muriel anna and rob gemma and nick bob richie tom c glynn richie lyn h anne l hazel s marie e bev charlotte and andy pat richard g lisann matt k andy and lynne tim and kim nick and debs laurie ian andy r toby and louise mark and ailsa griff nick and kate goody tony p tess ma and pa plumb jim and gayle small paul claire kate denise scott angie karen kath h suzanne michelle richard g lynne b greame abigail vic linda and stuart ian w helen karen and griff helen and noel victoria and andrew camilla rebecca and sophie judith and malcolm moira tony f alan w benet biscop young enterprise nick b dom kevin lesley and keith rosemary simon m eileen m mike f karen t jake p judith f andy and estelle con marion dave w wendy and brian craig t p phil lesley b jon lawrence michelle heather ali a rob h chris h kim mo hayley and sean debbie h jules andy a udi femi angie claire isabella andrew g will mark w seb luis vassilis aidan matt kate s kate g rory simon d patiroop mujalin worawan shree paola nikorn russell joana stephen w niel angie andy g fiona craig a alistair tom b stephen f pauline sofia peter f geoff a geoff p jamie i mark s paul q gholamreza f s contreras steve l owen l nolberto solano urlLink    Little Drummer Nobby&nbsp; urlLink  '),(25624,2673530,'16,April,2004','2004-04-16 00:00:00','   The Bellavista Social Club   In Peru, on the banks of the Amazon, is a city called Iquitos.  In Iquitos is a small parish called Bellavista which has a local under-21 football team called Bellavista Juventud.  Presiding over the team is Father Joe Plumb, an Englishman, formerly of Ashington, Northumberland.  The side competes with other nearby teams and Father Joe is rightly proud of Juventud.  He believes that the football team isnâ€™t just great entertainment for players and spectators alike but it also helps teach qualities as diverse as self-esteem, pride and responsibility.   In 2003, Father Joe was joined by Doug Paterson and Charlotte Weatherell who also settled quickly in the town and fell in love with the area and its people.    Their common link with Joe is Bedlingtonâ€™s Benet Bicop school in Northumberland, where Joe was a regular visitor and Doug a history teacher.   During their time in Peru, Doug and Charlie put their own skills to good use.  This included putting on special classes for local  children to help them with rudimentary English.   urlLink    Charlie and Doug\'s students get their certificates.&nbsp; urlLink     They were made very welcome, by Joe, their pupils and the townsâ€™ people.  And although Doug was keen to teach the children all he could.  There was nothing he could teach the football team.   That will come as no surprise to anyone who has seen Doug play.   urlLink    Doug sets off on a run down the wing and is sensibly ignored by the midfield ball winner.&nbsp; urlLink      Anyway, before too long it was time for Doug and Charlie to leave.  But not before they had handed out certificates to their pupils and made a promise that they would return.  There was also one final promise.  That they would look for sponsors for Bellavista Juventud back in their home city of Newcastle.  So it was decided that in order to help support the football team and to buy the shirts, shorts, socks and boots they needed,  a new club would be formed â€“ the Bellavista Social Club.    Everyone brings their own special gifts to the table.  While the young men of Bellavista are tackling, dribbling and shooting, the less fit, older crowd, back in the Toon can do their bit by drinking, smoking, dancing and putting their hands in their pockets.  The first meeting of the club, including will take place at the Egypt Cottage, in Newcastle on Saturday April 17th from 7pm to midnight.  Tickets cost Â£5 with all profits going to the football team.  '),(25625,2673530,'23,May,2004','2004-05-23 00:00:00','   urlLink    &nbsp; urlLink     List of players names (lista de sus nombres de los jugadores):  Goalkeeper (arquero) - Juan Manuyama MorÃ­ Left half (delantero) â€“ Jackson Perea Suarez Centre forward (delantero) â€“ Dennis Mark DÃ­az de LoaÃ­sa Right midfield (medio-campista) â€“ Alan Jorge DÃ­az de LoaÃ­sa Right wing (volante derehista) â€“ Daniel Ersen DÃ­az de LoaÃ­sa Left midfield (medio-campista) â€“ Luis Alberto Pilar Teo Left defence (lateral izquierdista) â€“ Gustavo Vela Tangoa Left wing (volante izquierdista) â€“ Jonathan Peze Flores Right half (delantero) â€“ RavÃ­ Pallas DÃ­az Right midfield (medio-campista) â€“ Raul Gomez Vela Right defence (lateral derechista) â€“ Jackson Gomez Huayamba Right defence (lateral derechista) â€“ Bete Junior Rengifo Manihuari  Presidente (president) â€“ JosÃ© Plumb Nathaniel Vice presidente (vice president) â€“ Jaime Pinchi Pinchi Director fÃ­sico (physical fitness coach) â€“ Mario RÃ­os VÃ¡squez Director tÃ©cnico (technical coach) â€“ Juan Vargas MelÃ©ndez Secretario (secretary) â€“ Carlos Valera Choclote Tesorero (treasurer) â€“ Ike Chong Vilchez Fiscal (?) â€“ Mercedes Manihuari Sanagama Delegados (representatives) â€“ Hugo Fernando Pezo DÃ¡vila, Clodomiro Rojas Lozano Utilero (?) â€“ Wilfrede Soplin Tonasoa Medico/masajista (doctor/masseur) â€“ Asensio Tina Fachin  '),(25626,2673530,'23,May,2004','2004-05-23 00:00:00','   The following inspiring message from Father Joe Plumb, President of Bellavista Juventud FC in Iquitos, was passed on by his father Bernard Plumb to Bellavista Social Club in Newcastle, at the BSCâ€™s first event at the Egypt Cottage on 17th April 2004.    urlLink    Speech from Father of Father Joe    Things look good for the lads of Bellavista this season.  The mood is positive, the camaraderie is tangible and having a new kit and 16 pairs of boots helps us feel ready for the new season.  The team have recently won three training matches, ALL against First Division teams from the League of Iquitos.  All of the 22 players are from Bellavista itself.  We hope to wean down the team to 18 players who will make up the final squad for the 2004/5 season.  Two of the lads are â€˜moto-carroâ€™ (Iquitos-style motorbike taxi) drivers, several are still at school or studying at technology colleges or to get into the University but the majority and unemployed or rely on occasional jobs carrying cargo in the port or working as fare-collectors on buses.  About ten of the lads havenâ€™t finished secondary education and therefore have little possibility of gaining any permanent work.    Playing football for their â€˜barrioâ€™ gives them pride in their abilities, loyalty to one another and the contributions and efforts that YOU are making show them that SOMEONE believes and shows a concern for them.    We started training in March, running each morning at 5AM and meeting at 7pm to have a knock-around at the local schoolâ€™s 5-a-side pitch.  In April weâ€™ve had three friendlies, and the good this is to see the understanding and good communication and real TEAM SPIRIT between the players when theyâ€™re on the pitch.  Mario, the physical fitness coach, is happy with the ladsâ€™ commitment and progress.  Juan Vargas, or â€˜Pata Largaâ€™ (big foot) the technical coach, is getting good results from the new players.  The â€˜starsâ€™ are Jackson and Ravi (although both are highly-strung and need to learn to control their tempers with the ref).  â€˜Yuccaâ€™ in defence is a good captain and organises his team-mates well.  Our wingers Beto and Gustavo need a bit more pace and all the players could do with growing a few feet and filling out more â€“ our main expense pre-season could be multi-vitamins and calcium supplements.  Your donations have helped us buy: â€¢	22 shirts, shorts and socks â€¢	22 bibs for training â€¢	8 pairs of boots (we had 8 pairs from last season still in good condition) â€¢	3 footballs â€¢	2 kit bags â€¢	inscriptions for 15 players â€¢	photo passes for 8 members of the directive for entry to the Max Agustin Stadium in Iquitos on match days â€¢	bandages, rub (â€˜ralgexâ€™ and â€˜??â€™)  The season starts in July and there are 12 dates â€¦ weâ€™ll keep you posted.  Thanks for ALL your support and please, if you pray, pray that the lads give up the booze and the women, at least before match days, as at present thatâ€™s the only thing between us and promotion this season!  Enjoy the party!  '),(25627,2673530,'20,June,2004','2004-06-20 00:00:00','   urlLink    President Doug.   Well, thanks to all those who bought raffle tickets for such quality prizes as a rubber glove! A nice Â£47.90 was raised for the Club at the Sixth Form A Level Drama Exam Performance (hope they all pass). A night of top quality acting with added social conscience.   Now time for some musing from me, the Club President.  The Bellavista Social Club is taking off in a way I never thought possible. We plan to have the best supported team in Iquitos Peru, although we admit there is limited competition. The lads are looking very smart in their strips and are set to  dazzle the opposition with both their footwork and their cool gear.  As for events there will be another meeting of the Club soon, keep checking in.  Think of it this way - the more we drink, dance and smoke tabs the more the lads benefit.  And the more the benefit, the better they do on the pitch.  Doug \'El Presidente\' Bellavista Social Club  '),(25628,2673530,'05,June,2004','2004-06-05 00:00:00','   urlLink    The lads looking very cool in their new strips provided by their very own social club.   urlLink    The Bellavista boys training hard for the new season  '),(25629,2673530,'29,July,2004','2004-07-29 00:00:00','   The latest venture for the Bellavista Social Club is the official BVFC Christmas cards.  Each are hand painted and you get the extra glow of spreading a little seasonal cheer.   No details as yet regarding price - but if you\'re interested then leave a message below with your email address.    urlLink       urlLink       urlLink  '),(25630,2673530,'27,July,2004','2004-07-27 00:00:00','   Thanks to the nice people at Natwest we now have a bank account, cheques payable to The Bellavista Social Club, lots of forms later, we have a treasurer in the form of Charlotte so all enquiries to her.   Plans for the Club to hold its Christmas party are forming, suggestions from members welcome.  Had a visit from Peru, not only are the boys the best kited out team in the league but money raised means that seven are now in college or getting job training!  '),(25631,4199600,'21,August,2004','2004-08-21 00:00:00','   urlLink    Hi!  Don\'t you love it when your sitting at a red light or just cruising and happen to glance over at another vehicle only to see something you don\'t want to? One thing I find disgusting is people thinking that just because they\'re in a car and their windows are up people can\'t see them. So they decide to pick their nose. That alone is a nasty habit. But when the finger is drawn to the mouth and you decide to eat what you pull out...THAT IS JUST FU**ING SICK. HOW OLD ARE YOU? I thought that picking the nose and eating a booger stopped in early childhood. But to do it as an adult? REALLY GROSS!! GROW UP PEOPLE! If you\'re going to pick your nose actually look to make sure people aren\'t looking. And if you decide to do this during the day, tint your damn windows!!  Blessings, Moxi '),(25632,4199600,'20,August,2004','2004-08-20 00:00:00','   urlLink    My cousins new baby! Isn\'t he just adorable? I\'m so jealous! Congratulations again Misty!  Blessings, Moxi '),(25633,4199600,'18,August,2004','2004-08-18 00:00:00','   urlLink    Hmm...  Is it the salt I love? Maybe it\'s the vinegar. Who knows? Both together make my mouth happy!  Blessings, Moxi '),(25634,4199600,'18,August,2004','2004-08-18 00:00:00','   I want a lemon with pickle flavored salt and a cheese hotdog with sauerkraut and onions. YUMMY!!!  Blessings, Moxi '),(25635,4199600,'18,August,2004','2004-08-18 00:00:00','   My husband and I have been married now for almost 3 years. And so far they have been the best 3 years of my life. There have been some times when I just feel like squishing his neck but through all the good and bad times he\'s always been there for me. About 2 years ago I was hospitalized for an ectopic pregnancy. About a month before that I had a miscarriage so I was already depressed. Now they have to cut me open and take another one away from me? It was a very painful and depressing time. From what my Doctor at the time said I was pregnant with fraternal twins. One I miscarried and the other just decided to attach itself to my tube. I didn\'t want to have the surgery but it was either the baby or me and the baby. I was in the hospital for about a week. Since the surgery I\'ve had 5 more miscarriages. It gets so depressing. I\'ve pulled myself away from my family and put a little distance between my husband and I. I\'ve been feeling so empty. I want children. I\'m just a little sick of people telling me that it\'s going to be OK. Especially people who have never been through it. There have been times where I\'ve even considered suicide. But my husband has been so supportive. As of right now my period is almost 1 week late. Losing 7 kids in less than 3 years is enough. I can\'t take another one. So I give it one more week. If I don\'t start, I\'m going to a doctor to get tested. You know the losses wouldn\'t have been so bad if I had a better Doctor perform my surgery. He was the most insensitive bastard I\'ve ever seen. And if any woman out there sees Doctor Joseph Bell here in Ogden I suggest you see someone else. To me it doesn\'t matter how far along you are in your pregnancy. You\'re pregnant with a baby. Even if it\'s only 3 weeks it\'s still a living growing person inside of you. After my surgery I was curious about how big my baby was. So I asked. Dr. Bells words were \"It wasn\'t big enough to be considered a baby. It was just a growth.\" I was about 2 1/2 months pregnant. It was a baby. My baby. And since then I can\'t help but think of that every time I miscarry. It\'s depressing. He wasn\'t even concerned about how I felt. He just kept telling me that I was too fat and that I needed to lose weight. My husband was there with me the whole time. I got the worst possible care from Davis hospital. The nurses were total bitches. They didn\'t take care of me. My husband did. My stomach was just sliced open and they expected me to be able to get up and walk like it never happened. When I did walk my husband dragged my I.V. while I held the little pump that sucked up and leakage from my wound and the nurse held my catheter. I was in a little too much pain and asked the nurse to stop and she said that I was fine. I stopped and she kept walking almost pulling my catheter completely out. It was a nightmare. I\'m just glad it\'s done and over with although I do feel a little empty at times. Especially since my cousin had her baby last week. I can\'t help but wonder how life would be if my husband and I had children. Lately the only thing I can do to keep myself busy is to concentrate on the craft and create new spells. Let me get off this computer and sulk to myself. Thanks for listening. Even if you don\'t know me.  Blessings, Moxi '),(25636,4199600,'16,August,2004','2004-08-16 00:00:00','   The other day I was flipping channels and happened to see a show called amazing animals on animal planet. This woman had a cat that woke her up because her apartment was on fire. Last night one of my cats was going nuts. Meow meow meow! At first I wanted her to just shut up. Then it clicked. Maybe something is wrong. So I got up and got dressed and looked around. Nothing was wrong so I laid back down. She was in my face again. Meow meow meow. Turns out nothing was wrong. She just wanted to crawl under the covers and wanted me to lift them up for her. Now if you think that\'s bad. She woke me up to let her under the covers. No big deal. But she didn\'t even want to cuddle with me. She wanted to cuddle with my husband. She did this all night and most of this morning. Amazing? I don\'t think so.  More like annoying.  Blessings, Moxi '),(25637,4199600,'14,August,2004','2004-08-14 00:00:00','   I\'m so damn tired. My stupid neighbors have their radio blared up so damn loud. When I was younger I loved loud music. Sometimes I still do. But not this early in the morning. Maybe I wouldn\'t care if he listened to something decent...  1 sheepy sheepy... 2 sheepy sheepy... 3 sheepy sheepy...  I don\'t have a serta so I\'m stuck counting sheep.  4 sheepy sheepy... 5 Sheepy sheepy...  Guess I\'ll try and sleep.  Blessings, Moxi '),(25638,4199600,'13,August,2004','2004-08-13 00:00:00','   Right now I feel like the little brown presents my cats leave me in the litter box.  Moxi '),(25639,4199600,'13,August,2004','2004-08-13 00:00:00','   My husband and I have been living in this duplex for almost 3 years. When we first moved in our neighbors at the time were awful. They were selling drugs and the women were selling themselves. We were able to get them evicted about 2 1/2 weeks after moving in. It was great. It was quiet for a little over a month. Then a lesbian couple moved in. It was still quiet. They honestly have been the best neighbors we have ever had. We got along great. We hung out together we partied together. I loved it. Then about 7 months after moving in they moved out. It was quiet for about 3 weeks. Then the neighbors from hell moved in. My husband and I have tried to be nice. We really have. Our new neighbor moved in with his teenage son in November of 2002. At the time he was 45 (now 47) and his kid was 17 (now 19). We\'ll call the older one Patrick and his son Alex. We live in a duplex. It\'s not the same as an apartment or the same as a house so a little common sense can go a long way. At the time they moved in my husband was working swings so we normally slept until sometime in the afternoon. 3 days after he moved in my husband and I were in bed sleeping when we woken up by a loud noise. BANG BANG BANG!! This ass starts banging on his drum set. Not playing...banging. Why have a drum set if you don\'t know how to play? Our walls are so thin you can hear almost everything that goes on in their half of the house. All the time with the loud music and the drums and now he\'s got an electric organ he loves to turn up. Yet another instrument he can\'t play. These people are trashy. He doesn\'t work. And when he does all his money goes to alcohol and drugs. He\'s only 47 and has already filed and used up all of his social security earnings. His mom pays for his rent. Sometimes they have up to 6 adults living in his one bedroom rental. Yet none of them seem to be able to work. What I really hate about this place is we have a shared basement. We used to have a washer down there because for awhile those were the only hook-ups available. No big deal right? WRONG!!! Even though they had their own washer they were using ours. Our electric and gas bill went up and we never had any hot water. I got so sick of it. I sure as hell didn\'t want the washer knowing his nasty clothes were in it. So we cut the power to our half of the basement from the breaker and cut the power cord so they couldn\'t use the washer. We bought a new washer and dryer and hooked them up in our back room. Problem solved? NO! They reconnected a power cord and plugged it into their half of the basement. We still had no hot water. My husband had to go to a hardware store and buy a plug and some cap to put on the water connection so they couldn\'t use it. So far it\'s worked. Then about 2 months ago my husband had to go down into the basement and turn the water off so he could fix our tub. He noticed an orange extention cord running from our half of the basement to the inside of the house. Now I know I said we cut the power the basement. Or at least we thought we did. All the lights on the main floor, upstairs and in the basement run on the same fuse. We didn\'t know that. When our bill went up almost $20 we didn\'t think anything of it. In the light socket on our half they had one of those things that allows you plug an electric cord into it. They had their power cut off due to non payment and was stealing ours. We couldn\'t cut the power to it because we needed our lights too. It really pissed me off because they were using our electricity to crank their radio up and piss us off. So we took care of that too. In the daytime when they had their radio cranked up we flipped the switch in the breaker and the music stopped. We were told to take pictures of the cord running from our half into their house for proof. So we did. And then we called to cops. It wasn\'t as simple as going down there and unplugging it. They would just plug it back in. These are drug addicts and alcoholics so confronting them was out of the question. It\'s just my husband and I. What can the 2 of us do against 6 drunk and high pricks? So like I said we called the cops. Then they have the nerve to get mad at us for calling the cops so they would stop using our electricity. They even threatened my husband. Alex is currently in jail for robbing and beating an elderly man for his money. It\'s so sad. He learned all that from his dad. His dad has been in and out of jail since he\'s moved in. And all the problems we have don\'t matter to our landlord. As long as he gets his rent money he doesn\'t care about anything. It\'s really sad. And now we\'re trying to keep his cat alive. When they first moved in their cat weighed about 12 pounds. He was healthy. Now he looks like he weighs no more than 3 pounds. He keeps his cat outside and never feeds him. So when we see him we feed him. If my husband and I didn\'t have 6 cats of our own we\'d take him in. But we just can\'t do that. Patrick\'s cat has been losing his hair and we don\'t know if he\'s seriously sick. We want to catch him and take him to the vet so he can get better. We\'d like to find him a good home. He deserves that. And he\'s so sweet. Even if all our neighbor did for that cat was put dry food out he wouldn\'t look as bad as he does now. It\'s only $8 for a 25lb. bag of dry cat food. But instead he chooses to spend it on alcohol and drugs. Sad really. Why have an animal when you can\'t even take care of yourself? I love every one of my 6 cats. I can\'t ever imagine mistreating them. My husband and I would starve before we\'d let our cats go without food. Our cats always have food and water. Sometimes they get special cat milk. They almost always get treats and catnip. They have so many toys it\'s unreal. We even bought them a 7\'2\" house. My cats are spoiled. And I wouldn\'t have it any other way. Just knowing that I live next to such a loser sickens me. Not only is he sh*t but his cat is suffering. I hate that. I hate him. I know that hate is such a strong word. But I do. I hate Patrick. He\'s such a loser. He\'s 47 and still depending on his parents and the state because he\'s too lazy to work. He\'s a piece of sh*t and I hope he soon suffers the way his cat is. He deserves that.  Blessings, Moxi '),(25640,4199600,'11,August,2004','2004-08-11 00:00:00','   This one time at band camp.... ;)  Moxi '),(25641,4199600,'11,August,2004','2004-08-11 00:00:00','   My husband works with a short fat ugly chick that claims she has a boyfriend. Not that that is the problem cause if that was the case I could care less. But when she claims they\'re getting married and he hasn\'t even proposed....COME ON!! \"I\'ve spent $1500 on a plate that serves 4 people.\" STUPID! Why plan a wedding when the guy hasn\'t even proposed? This chick makes me sick. She talks about herself too damn much. Normally people like this wouldn\'t bother me. But when she calls me fat behind my back to my husband when she doesn\'t even know me and she weighs like 500 lbs. herself it really pisses me off. She\'s pretty damn lucky my husband needs his job. Believe me my fist and foot would love to become close friends with her fat ass. She even had the nerve to ask my husband why I don\'t like her. Like it\'s that damn hard to figure out. Woo! It feels good to get that off my chest. And if that nasty slut just happens to read this....I have only 2 words for you : F**K YOU!!  Moxi '),(25642,4199600,'10,August,2004','2004-08-10 00:00:00','   Congratulations to my cousin Misty and her husband Stephen. Today she gave birth to a little baby boy. He weighed 7 lbs. 11 1/2 ozs. and is 21 inches long.  Misty,  My dear sweet cousin. Now that you\'ve had the baby quit complaining that you can\'t see your own twat. And please. PLEASE PLEASE PLEASE!!! SHAVE YOUR LEGS!!!  Blessings, Moxi '),(25643,4030244,'28,July,2004','2004-07-28 00:00:00','   I\'m re-doing my about me bit, it\'s too long. It may be crap but it took me a while to write so I\'m posting it for posterity  I\'m a girl who\'s old enough to be calling herself a woman now, not a girl, and takes offence if referred to as a girl by most other people. I\'m an optimist and romantic, with those ideals dying slowly and painfully each day that I spend in the real world (I try to avoid that though) I\'m a feminist, and love to be feminine, without that being taken to mean I\'m in some way a lesser person than one who has their genitals hanging about on the outside. I love to be treated like a lady by a man who is a gentleman, but does not see weakness in my gender. I hate to be hated and reviled by other women because I love to be sensual and approach physical pleasure in a way that would be perfectly and widely acceptable were I a man. I hate the double standards that exist in this world. I hate lies, but hate people who parade insensitvity and malice under a flag of openness and honesty. Those who feel that any mean words can be excused if followed by \"I\'m just being honest...\". I may seem self absorbed at times, but believe that thinking of others costs nothing and will be paid back through a kind of karma. And I have an exception skill for navel gazing and writing flowery crap! '),(25644,4030244,'25,July,2004','2004-07-25 00:00:00','   Here\'s a list of things which are insulting and apparently not insulting.  I\'ve learned them in the last year or so, I really had no idea!     It is insulting to point out that someone who is older than you, is older than you.  It isn\'t insulting to disregard a 27 year old as being dumb and ignorant, simply because you are OLD.  It is insulting to ask a supposed friend how their current relationship is going.  It isn\'t insulting to tell someone that they\'ll never get over you, move on in life or find someone who will want them and love them, the way you used to but don\'t any more.  It is insulting to comment with genuine concern that someone is loosing weight, when you worry that it may mean they\'re ill.  It\'s ok to comment on how fat someone is, and constantly refer to them as fatty and fatass.  It\'s wrong to jokingly point out a (one amongst many) new grey hair.  It\'s just fine to tell someone their skin looks like shit, not jokingly, and&nbsp;in a \"not sure I want to be seen with you\" way,&nbsp;even though you are well aware that they have their period due in a day\'s time.   and most of all   It\'s perfectly OK to lie about every aspect of your life to someone, just to make sure you can carry on screwing them, and still insist you care about them when discovered, and maintain that no wrong was done by you doing this.  Under no circumstances should you react, on discovering the above, in a manner that is angry or emotional, nor should you blame said liar, as it was done with your own interests at heart really, considering being used as a cheap fuck in a hotel room by someone who you have trusted implicitly is probably all that keeps you going, silly girly! '),(25645,4030244,'25,July,2004','2004-07-25 00:00:00','   \"hi, I miss our arguments !!!! Seriously, I miss talking with you x\"     &nbsp;  This was yesterday! His response to me not answeing emails. He\'s also used a different mail address, so someone\'s told him that I\'ve redirected all mail from his usual one to my spam folder.   Talking about what, is what I want to know!   Way back when, before the whole relationship debacle, we did talk. We\'d sit over a drink and chat about all sorts, music, movies, the state of the world, all the usual stuff. We\'d talk about our relationships, our respective love lives. We were really close mates so this was not unusual.   Maybe it\'s the storyteller in him, he needs create a total fiction and I\'m the only one he feels comfortable enough to share it with.  Maybe he\'s just a full of shit control freak.  Hmmmm....   For the past 6 months, the only times he\'s talked to me about his love life or relationship, it\'s been a total fabrication. The only other totpic of conversation has been him arranging when we could possibly meet up, and that was only for sex. I\'d agree, wanting to sit and talk with my friend. W\'e meet in a hotel, have sex, then he\'d put on the TV and each time I opened my mouth he\'d shush me.  Any films or music I liked were laughed at and scorned as dumb and unworthy of his conversation. He didn\'t want to talk about music or films, just sex, or even forgo conversation altogether and just have sex.  Our relationship, it\'s past and possible (according to him) future, were completely taboo topics. If I talked about them, I was being a \"bunny boiler\"&nbsp;or was trying to make him feel bad for dumping me, apparently.  If my love life came up, I got a barrage of questions. I have no love life these&nbsp;days. Before I found out the extent of his lies, and was still being told constantly by him that we\'d get back together, I only slept with him and I didn\'t date. I didn\'t want to anger him.&nbsp; Now that I have found out (apart from being to broke to go anywhere anyway!) I feel sick just thinking about sex, and my opinion of men is not exactly&nbsp;at it\'s highest point. I knocked down a lot of hard-built walls to let him in, I\'m not doing that again for anyone.  If I was feeling bold, and said that I was making an effort to find someone new, he\'d laugh out loud at me, saying \"Yeah, right...\" And \"why fight the inevitable, we are meant to end up together\".   Since I stopped sleeping with him, he had pretty much nothing to say to me. All his MSN chats were to get me to talk dirty with him, to get me to sleep with him again,&nbsp;or to try and to pry into my personal life. If I refused to answer, or was evasive because it really was none of his business, all I got was a tantrum and a torrent of bitchy abuse, aimed at insecurities I\'d confided in him about as a friend.  I would have been well at ease talking about my (admittedly none-existent) private life, but that he refused to talk about his. I wasn\'t trying to snipe or be mean. I just wanted to know what was going on with him. I was by this point resigned to never getting him back, as he&nbsp;wouldn\'t let me know where I stood. I wanted to just chat. But he would barely say a word, or would just command \"don\'t even go there!\". I have, of course since found out that the few things he did tell me were complete lies anyway.   I have no idea what he thinks he misses talking about.  Maybe he misses the feeling he got, whatever that may have been, when his&nbsp;asking me how I was ended with me crying because&nbsp;I felt like shit. Like all I was good for was being used by him, being his dirty secret, his cheap thrill of doing something he shouldn\'t.   &nbsp; '),(25646,4030244,'25,July,2004','2004-07-25 00:00:00','   Not long after we split, he found someone new. There may have been an overlap, I suspect he was seeing her when he was still with me, I\'ve no way of knowing. The first big lie came. What he had told me was nothing very serious, was actually them living together. I freaked and he simply told me he hadn\'t wanted to hurt me.&nbsp;  We were still having sex. I told him I didn\'t want to be his dirty secret, that we couldn\'t have sex anymore. He stormed and sulked and said this meant we couldn\'t be friends anymore.  After a time he started to come to me saying how she wasn\'t what he wanted. They didn\'t have a future and didn\'t sleep together any longer, he was now just a lodger. From time to time we\'d spend a night in a hotel together.  I wasn\'t dumb enough to belive they didn\'t have any sort of relationship at all, but I did believe him when he said there was a chance that we could work it out, that I was the one he wanted in the long run.  If I asked about her, or their relationship, he would loose his temper, saying not to ask. If I wasn\'t updating him on all aspects of my life, he would say that I was jepordising my chances of winning him back. I still loved him, and had always been able to take him at his word, so I told him where I was going, when and who with.  I went on a trip, alone to the States, fulfilling a long held dream. He tried to stop me from going, using emotional blackmail. Just before I left, he told me to trust him and have faith that he\'d got wheels in motion for us to be together again. Through out my trip, even though it was difficult, expensive and I often got a sulky or rude response to what was perceved as \"late\" contact, I stayed in touch. I held back from things,&nbsp; turned down dates and kisses, thinking that he was really going to fix things for when I got back.   I really am dumb   Within the first week I was back, He\'d told me he was going to make a go of things with his new woman.  Within a few weeks of that, he was back on the fone saying there was no hope for them and that, yes, I was the only one for him.   I still really wanted this to be true, I still believed that he wouldn\'t tell me a lie that huge. He knew all my insecurities, all my hangups about being decieved, about being used, and he was my friend, who I could, at the heart of things, still really trust. '),(25647,4030244,'25,July,2004','2004-07-25 00:00:00','   We were friends for years. I knew him and all his flaws, he knew mine. I was young single and very very fre, and he was in a relationship with someone i knew, which was stormy and destructive. We were soul mates, the first person each turned to when we needed to be understood. We didn\'t judge each other.  When we first met, we slept together, it was just a one night stand, both of us insensibly drunk and high, trying to be kinky and ending up giggling at sillyness. He wanted to wrap my breasts in clingfilm, but the stuff he had at his place was the sort thats pre-perforated to tear easily. I was sick on his t-shirt, he though it was funny. I\'m not sure how, but with time we grew to be mates. We\'d still have sex when he was thrown out of the shared flat.  He was always there for me. Trough all the shittyness that comes with being a single girl trying to date. If it was needed he\'d give me a shake and tel me to get a grip, and I\'d do the same for him. He was always \"no good\", being in a band, he\'s go on tour, doing large metal festivals in mainland europe, and along with that doing drugs and groupies. I knew he was faithless, but he was always honest with me. When she finally got sick of him, and found someone else, it tore him to bits. I loved him as I do all my dear friends, and tried to pick up those bits and put them back together. I slept with him many times, but without sex. I held him when he cried over her. We spent more and more time together. People in our social circle assuming we were a couple long before we actually were. He told me that he loved me, and broke down the walls I\'d built over many years, and I told him i loved him too.&nbsp;I was unsure of a relationship, I knew him. I told him, I could live with infidelity, I was aware that it came&nbsp;with the package. I told him the worst thing he could do to me was lie to me. Anything else I\'d scream and shout about, but would get over with time, maybe even end our relationship, but we\'d still be friends. But I cannot stand lies, I cannot stand the thought that I\'m being laughed at behind my back for being a gullable fool. Anything else was forgiveable, that was not.  After some time, our relationship did end. He ended it. He\'s given many excuses since then as to why, I\'m not even sure that he knows why.  I was upset, I did scream and shout, but it was something I could get over.  Then came all the lies, and I hate him now. The only thing I ever asked for, he couldn\'t do. A person I trusted with all my heart, betrayed me. I\'m sure it will pass but right now I\'m angry and I am bitter. '),(25648,4030244,'25,July,2004','2004-07-25 00:00:00','   I\'m starting this blog simply as a place to vent, enthuse, whine and ramble, which is probably the reason most people do.  My head just feels crammed full of stuff that needs getting out. Not anything creative or witty enough to share really, and mostly stuff I don\'t want to bore my mates with, things they\'ve heard a thousand times before.  Mostly, for at least a little while, my posts are going to be about my strong and overwhelming dislike of one person, who shall remain nameless. An out pouring of all the things that he won\'t listen to, all the ways he\'s upset and hurt me, bile and bitterness that I need to spit but I don\'t want him to know he can still provoke.  I\'m not a teenager, this won\'t be all \" oh poor me \" stuff. I just need to let it all out,&nbsp; I\'m hoping this will be a cathartic experience. And once I\'m done with the lemon squeezing, I\'ll just stick all the stuff down that I used to put in my paper journal.  I hid that when he came into my life, thinking that I now had a person I could express all that stuff to, that I didn\'t need it anymore. Now I can\'t find it, so I have this instead. '),(25649,4030244,'04,August,2004','2004-08-04 00:00:00','   Spell checked that last post. Mondaygirl has only one suggested alternative,  nondescript.   That was the name given to the fake Victorian chimeras, two dead, stuffed animals, stitched together to make a \"mythical beast\" for a circus or sideshow. A monkey and a big fish, supposedly a mermaid, going bald so you can see the seams. That\'s a nondescript. '),(25650,4030244,'03,August,2004','2004-08-03 00:00:00','   I have to go to the bank this weekend to beg. I\'m as broke as broke. I even owe my gran money! Broke as I am, I do have a roof over my head, even if I despise the people I\'m under it with. My bank are going to piss off, but I\'m going to ask them for help getting out of here.  \"yeeees mondaygirl, I see you have no savings, huge debts and not a very large income. Of course you can have help with a mortgage! We\'ll just take your soul...  *checks screen of computer* **mutters** oh MONDAYGIRL, hmmmmm,   better make that your kidneys, all things considered, as a security.\"   I can\'t stay in this house though. I\'m already nut\'s, I just don\'t want to push it further. It\'s not even like I can go out, I have no money to go anywhere. On a warm day like today it\'s a killer not to be able to just slip into the pub after work.  Not wishing to dwell, but I\'m not even sure I could go to the pub had I the money. Not my old regular anyway. There are people with whom I\'d had heated arguments with when they told me that I should stop being dumb, tried to tell me about what  he  was up to, then he\'s got wind of it, had a word and they\'ve phoned and apologised, they were mistaken, in unconvincing leaden tones. All I can wonder is what has he said to them to make them make that call? Am I now pitied, has he told them the \"for her own good\" crap that he tells himself and tries to tell me? Or has he portayed me as the one who won\'t let it go and insists on making the first contact each time? And my worries about what will happen when I first bump into him with her aren\'t about how I\'ll cope seeing them as a couple, or him with someone else, they\'re about what the hell has he said to her about me? Obviously not \"oh, that\'s my ex who I\'ve been shagging behind your back and slagging you down to\". What though? Am I about to get my throat ripped out or be laughed at? I think I\'d prefer the first option. And will I be able to resist the temptation to tell her, to hang onto my dignity and not tell her everything she really does deserve to know about the lying scum that\'s leaching off her. I\'m bound to be drunk, but I\'m also bound to be with friends who\'re used to wrangling a drunken, difficult Mondaygirl.  Blah. None of that makes me any better off financially. Money can\'t make you happy for sure, but it takes the edge off being miserable. Can\'t remember who it was, but I read a quote;  \"life is like a shit sandwich, the more bread you have, the less shit it seems you have to eat.\" '),(25651,4030244,'03,August,2004','2004-08-03 00:00:00','   thinking on it, or brooding as it\'s sometimes known, the privacy of the shitty space that is my room  has  been invaded. I let him come here (for sex, what else), he has been in my room, in my space.  Git.  I am erasing him from my life though. there\'s no trace of him on my computer or fone. No old messages, sent or recieved, no addresses, no numbers. Nothing.  At the weekend I\'m burning all pictures from the one holiday we took together, off loading some gifts I still have and never wanted to charity shops, and then that\'s it (I am keeping some of the few good things tho! I love my polariod cam, my Elvis dvd\'s and some other bits and bobs, despite the fact that they came from him!). Nothing of him in my life but bad memories that are already fading. '),(25652,4030244,'01,August,2004','2004-08-01 00:00:00','    '),(25653,4030244,'01,August,2004','2004-08-01 00:00:00','   Hmmmm. Put that in my first post. Has fallen a bit by the wayside really hasn\'t it.  I\'m not a teenager any more (thank god!) but I am clinically depressed. No childhood trauma, no abuse or anything, just the wrong chemical balance in my body. Like being a diabetic, but I can eat cake and I can bum you out more in a conversation.  I have big highs and big lows. I have antidepressants that I hate to take, and don\'t, unless I\'m crying all the time instead of some of the time. I\'m not really a fatalistic, but I was put together this way, a bit wrong in places but that\'s just me, who I am.  **melodramatic violins** I my view, my condition as being in a way, a terminal illness. It is the thing that will kill me in the end. (**make that cellos, more mournful**) Unless you\'ve been there, you don\'t get it. And even some people who have still don\'t get it. Suicide (dum dum daaaaaaa!!) is not caused by one shitty event. You can walk around feeling like that all the time, like \"why the hell do I bother?!?\", doing all the hard work that amounts to nothing but more hard work, and you just carry on. There are good bits, and people who\'s feelings you care about. It can make friends and relatives who are left behind feel abandoned and rejected, like they failed you somehow, failed to fix the world for you. Not being conceited about how important I am in their lives, I just know how I\'d feel if one of them took the \"big step\".  All that the one shitty event does is just make the part of you that really couldn\'t care less about other people, and what their reaction might be, turn around and go,   \"What now!? MORE crap?! Crappier crap!? Hey, you won\'t even be here for all that \"boohoo\" rubbish, look out for number 1 for a change, look after your own needs! We don\'t have to stand for this, it\'s getting us nowhere, carpe deum, sooner\'s better than later, old age looks like it sucks ass! They\'ll respect your decision, they know it\'s crap too. If they really love you, they\'ll know this is he only way you can stop feeling this bad! There\'s only more of the same on the way you know\"'),(25654,4030244,'01,August,2004','2004-08-01 00:00:00','   I feel totally shit. I can\'t sleep at all an I\'ve hit a total low. If I tke my tablets I can\'t work properly at my job, which I need to pay back my debts incurred on stupid things. I\'m stuck in this shitty house with 2 people I can\'t stand, with no money to get away from them. Neither short term by going out and letting my hair down, nor long term by getting out of this house. I\'m paying off a huge loan for a dumdass holiday for the next 7 years. A holiday I didn\'t even make the most of because I believed the word of a liar, and held back, not wishing to ruin something that it turns out didn\'t even exist.  I have no space, no privacy and no peace and quiet.  The one person that I could always turn to when I felt this low has betrayed my trust so much that I couldn\'t turn to them even if I could stand them right now.  All I have is my one shitty room, and a bunch of junk in it. I can\'t even guarentee that I have that. I can\'t tell if my privacy is invaded to the extent that my room isn\'t even truely mine anymore.  He was completely right, all the things I like are useless crap. I am not a person people want around and I am stupid. '),(25655,3850554,'08,July,2004','2004-07-08 00:00:00','    Crap Crap and more Crap.      Now that I have that off my chest I guess I can get on with todays blog.    I finally learned how to change the windows explorer text size (after being a coputer tech for ten years), so now I can actually read the pages I go to on the web.  Too bad there isn\'t anything worth reading out there.  Friends of ours from Alberta are comming to stay with us for a while.  Their ACTUAL vacation plans got screwed up beyond repair so their final desperation choice was our place.    My dick hurts.  Took my son and his friend to the community pool today.  This was the first time I had been there.  Nice place.  The tits on one of the lifeguards were pretty nice also.  I wasn\'t allowed to smoke near the outside pool so I had to travel through the main building and out the front doors to get my nic fix.  I just ate two pizza pockets but I\'m still hungry.  that\'s it  go away now.   '),(25656,3850554,'05,July,2004','2004-07-05 00:00:00','    AAAAAAARRRRRGGGGGGHHHHHHHHH!!!!!!!!  This pretty much sums up my life.   I\'m not really unhappy, just frustrated and fucked. (hence the Paxil)  Since this is my first blog ever I guess I will tell you the story of Ogrgiss from the planet of Xlplik.. Or maybe not.  Perhaps you would rather I stick to the theme and describe my daily life etc...   ON WITH THE SLAUGHTER   I was born at a very early age  ha ha  ha ha I love that one.  My wife says then I turrnneedd (sp) into a gorilla.  Not quite sure what she means by that but she is sitting right beside me so I will humor her.  she is getting abdomon pains right now but she says it isn\'t period cramps just dull aches that won\'t come out.  Probobly farts.     OH I SEE.   She said the gorilla thing as I am hairy.  wow she comes up with some good ones doesn\'t she.  Today I did squat.  Same as yesterday.  And the day before that.  You see I am on Paxil.  A drug that is given to people (and lab rats) that stops you from going deeper into depression.  That depresses me.  The side effects include sleepiness, a sense of lazyness, and the inability to CUM !  This is why my wife is a very happy person.  I can go for a long time without cumming but that presents other problems such as a raw penis, wife falling asleep during sex,  AND SEVERE RAINBOW BALLS   Rainbow balls are like blue balls but even worse.  To see a picture of my wife\'s dildo collection click here urlLink    Sorry boys that link thing didn\'t work.  Lets try that again here urlLink    I\'ll have to work on that link.    I guess I\'ll go for now as I am feeling horney and need to jerk off.  Sorry if I offended you but if you can\'t take a joke   FUCK YOU!    In the next episode you will hear about the time I rescued all the one legged taranchulas from a KFC slaughterhouse.........   '),(25657,4231869,'15,August,2004','2004-08-15 00:00:00','   who are You?  Take the quiz:  urlLink \"Are you romantic?\"    Hell yeah!  I haven\'t even met you and I hate you!  Take the quiz:  urlLink \"What does your birth month reveal about you?\"    November  Has lots of extraordinary ideas. Difficult to fathom. Thinkforward. Unique. brilliant. Sharp thinking. Fine, strong clairvoyance. make good doctors. Dynamic. Secretive. Inquisitive. Know how to dig secrets. Always thinking. Less talkative. amiable. Brave. generous. Patient. Stubborn. hardhearted. Determined. Never quit. Hardly become angry unless provoked. Love to be alone. Think differently. Sharp-minded. Motivate self. Dont appreciate praises. Highspirited. Well-built, tough. Deep love, emotions. Romantic. Uncertain in relationships. Homely. Hardworking. High abilities. Trustworthy. Honest. Keepsecrets. Cant control emotions. Unpredictable.   hey, actuallie quite true eh.. cept e less talkative n praises n gd doctors part.. hurhur.. doctors?!? it\'ll b gd if i managed to not kill sum1 accidently!!! '),(25658,4231869,'14,August,2004','2004-08-14 00:00:00','   who are You?     went out to mug wif jeannie n co. juz now... dey\'re reallie disciplined pple which is gd cuz hopefully i can be influenced.. hehe.. reallie need all e disicpline i can get.. oh n lots of luck too.. haha.. ar well itz sun again.. time reallie dussen fly u noe.. it zooms!!! like prelims\' in 3 wks? oh gr8... nah i\'ll make it.. yup.. self believe is powerful!!! wif actions accompanied of cuz.. hurhur... abt loh n leow.. dey\'re drivin me nuts... i guess e best thing to do now\'s to stand back n let nature takes itz own course.. yeah.. tt\'ll b e smart thing to do.. God noes i\'ve tried... i juz pray dey wun end up hating each other.. yeah.. God help us..      briefly chatted wif hid juz now... tt gal.. she\'s ultra nice noe... can\'t believe sum1 like her actuallie walk e earth.. i mean dun dey all belong up dere? yup.. truely blessed to have a frend like her... pls note tt not every \'like\' word is used as a simile.. u used a simile btw A n B when A is like B... \'frend\' is not like \'her\'.. i mean becuz she\'s a frend.. oh, in tt case it shuld b metaphor rite? geez i\'m gg bongus... too much of figurative language tt\'s y... newaes, u dun use \'like\' for metaphor cuz A is B when using metaphor.. yup.. k ENUFF! sighz tis pile of tut\'s staring at me now... wat do u want from me??? k i get e hint... signing off NOW!      had bad dreams again last nite... so for tonite i hope i\'ll have NO dreamz at all! '),(25659,4231869,'14,August,2004','2004-08-14 00:00:00','   who are You?   ar... hid.. dere u go.. i\'ve created a simple blog 4 meself... a real simple 1 indeed.. haha.. newaes... went to ktv juz now... super funz ar.. haha.. fei xue\'s voice ULTRA nice!!! can juz fall in lurve wif it... can\'t believe i slack away 2dae... nvm dere\'s alwaz a tmr!!! hurhur... ar well, tmr  i\'m gonna mug super hard to make it up 4 2dae\'s loss.. hehe.. ~yawnz~... hope i\'ll have nice dreams 2nite!       i\'ve been a fool again. but tt\'ll b e very last time. tis\'s a promise to myself... '),(25660,4231869,'14,August,2004','2004-08-14 00:00:00','   testing testing 1 2 3... hehe tis\'s real wierd... nvm.. 1st timers shall b pardoned! ahem... so tis is e 1st entry! '),(25661,3766692,'30,June,2004','2004-06-30 00:00:00','   I finally not late for sch... hehe... today got Culture and Communications module... sianZz... change of the groupings... sianZz... dun wanna change... okay lo... really nth to write leh... oh ya... tml, guitar prac and band prac clash... how? go for which one??? haiZz... mi sianZz... how? how? tell mi leh...  '),(25662,3766692,'29,June,2004','2004-06-29 00:00:00','   haiZ... today i am late for sch... the stupid dog \'s lesson... wa lao... den the prob so doff... dunno wad inductive and deductive.... so diffcult... den the VB test was like hell... nv learn stuff came out, mi gong gong... found out Floorball got to go this friday, den my whole week everyday not free... lolZz... so stress la... den after the test liaoz rush to Wind Symphony... got new score... dunno wat sheep graze blar blar lo... den so diffcult... flute melody... luckily mi second... or else the band will hav no melody.... lolZz... got an expert beside mi... so stress... cannot play the semi quaver thingie... aiyo... so after that i borrowed my instrument home lo... den when i reached home, den realise like no time to practice lo... Waste energy to bring home flute.... den now got black out... wat kinda rotten day is this... so damn hot so tired cant slp... wat did i do to my life man! so cham!haiZzz... =(  '),(25663,3766692,'28,June,2004','2004-06-28 00:00:00','   haiZzz... today sch reopen... c my poly frens again... hee! quite ok la... lots of fren going to Np sch reopen today... all of them say so sianz...hee... erm.. wonder they can use msn messenger in sch anot...(hmmm)... nvm... back to my blog... today so sianz lo... basic science... wanna die le... learn abt wad dissipation and stuff leh... HELP!!! so diffcult nv learn the topic in sch sch... it\'s abt the concentration and erm... blar blar la... aiyo... just finish writing the Reflection Journal... so long... hope tt tml my Congnitive n prob solving fag will change... stupid DOG!!! LOLzZZ pple in my class will noe y la... Oh Ya, tml got the Visual Basic understanding test... aiyo... Cham le la...OmG! tml still have Rp Wind Symphony... 5.30pm to 8.30pm... i m gonna be so tired!!!oh! now havta go study le... haiZz... so tired...  ---> Every obstacle presents an opportunity to improve our condition.  '),(25664,3766692,'27,June,2004','2004-06-27 00:00:00','   Erm... hihi... my first day of this blog... dunno wad to write... all thks to my frens blog page attracted mi to start â€˜bloggingâ€™ too! Erm... basically, nth much to write la... today nv went anywhere... so tok abt yesterday, I went to the Vb thingie... the senior was a great helper... veri friendly n everything... den after cleafing my doubts n everything, mi, Peili, Shuimin n Eunice went orchard lo... we went to KFC to eat... den funni thing is we all end up eating the same meal... haha... wonder y??? keke... while we were eating we tok abt sec sch, rp life n guys... keke! Found out tt Eunice n Shuimin dun really tok abt guys la... so tt topic was passed on lo... den later we say abt music... haha... my ears pricked up (omg, like a dog!!! lolZz)... itâ€™s like mi n Eunice r the onli one toking... aiya... veri unfair to others... ok den we when to walk walk n Eunice has to go cos she has to buy sum stuff n had piano lesson later... so we bide goodbye at lido... den the rest of us went FarEast... we went in n straight into A BOOKSHOP... pple go there shop for clothes we go there READ!!! Haha... the rest of us likes to read lo... we were looking around den suddenly settled in the Chinese n comic section... jus settled down as in sit down on the floor like the place is the library like tt... lolZz... but luckily nobody chase us out or anything like tt... Shuimin den found out tt there is this really great bk she wanna recommend it to us but she alone couldnâ€™t find the book so we three were hunting around for this one bk... den mi n Peili got tired so decided to look at comics... den we were discussing abt the comic character like we look at the books picture to decide how interesting the book is... haha... oh ya, I luv jap character in the â€˜mangaâ€™, oh ya, I used to copy them out in my secondary sch days... I like young Kindachi, Doraemon n so on la... when we got out of the bookshop, we decide to go walk walk again... when we end up in another book shop, Peiliâ€™s fren, Sharon, joined us... Wa! A veri mature ger, not like the three of us, siao siao one... hehe... veri friendly too! But can c our taste diff la... but nvm all of us still got along well... well, by the end of the day I got to noe Shuimin better... den found out tt she n I luv crafts... her cross-stitches n mi beads... oh ya, btw, I hereby announce I noe how to make earrings wif beads... hee! Cos this Peili dunno n she was quite shocked... back to today, in the end mi n Shuimin went off first, as I was failing geographically in directions, Shuimin had to sent mi to the MRT station... mi feel so guilty, cos her bag was so heavy wif a veri thick VB book... SO I wanna comment tt Shuimin is a veri nice n really great friend having her as my friend is so lucky, man!!! Thks ger! Erm... ya not the end...after tt tot I could go hm n slp... received a call n had to go ChinaTown, my family were down at the night market... wa when I reached there I couldnâ€™t find them... search for almost 45mins den found them... so irritating! But the place was so crowded, we walked for several streets n we reached the night market... the things were veri common like pasa malam like tt... but when we reached the food section it was WOW, the pastries n tutu kueh making my mouth water... so nice! U should image the dinner-less mi! den we brought the food home n eat... now I eating the egg-tart... wa ! Yummie!!!  '),(25665,3766692,'06,July,2004','2004-07-06 00:00:00','   hmmm... today the DOG\'s lessons again... wad to do... jus late again... but this time the DOG is more late... lolZz... hmmm... nth much la... the lesson is abt bouncing the ball... so lame... we spent our time outside doing the bouncing experiment... so \'FUN\'... pple look at us like we siaoZ like tt... den after the lessons got RpWS... the prac is quite okay la... Mr Gloss is not as strict as i thought he was... lolZz... the music he really nice... i like the piece \"StarShip\" the most... so claming piece... other pieces are ok la... still can follow but cant play tt well la... but i\'ll JiaYou... all the best for myself...  '),(25666,3766692,'04,July,2004','2004-07-04 00:00:00','   Hey! Forgot to write yesterdayâ€™s blog... too tired to remember... btw, yesterday nth much lo... jus went to guitar practice and play the song... it is a veri nice combination of instruments... but I guess I really need a conductor to cue mi in... cos after a lot of rests I got lost... got some notes oso in wrong key... aiyo... 13TH of july performing... how to survive... veri near liaoz lo... not much time to practice... hmm... now onli and wait n c lo... wad to do.. practice more lo.. haiZz... yesterday got watch the JIAN GUI, quite okay la... not veri scary cos I watch b4 le... lolZz... the lift part is so exciting... den this stupid show make mi forgot to watch the last epsiode of ALL IN... Damn it... did Ren He die? can anyone tell mi... wa laoz... sianz ji pua... today oso nth much to do la... return my books and gif Michelle my guitar... practiced my flute n guitar... so shuang! If onli m split into two den can play both at the same time... lolZz... impossible... but now can onli persuade my bro to learn guitar... so can play together... oh ya, den he can sing along... lolZz... haha... jus for hueverâ€™s concern he is a choir member, he went to the youth day ceremony yesterday... OH YA!! Which reminds mi, Y RP NO YOUTH DAY!!! I still a YOUTH leh... sianZz... my bro luffing at mi so PIA in my studies... -_-lll... sianZz...  '),(25667,3766692,'02,July,2004','2004-07-02 00:00:00','   Erm... today morning was really bad, due to sleeping late yesterday (some idiots disturbing me when I was rushing through my RJ!!!), I went to school quite late. To my surprise PeiLi is oso late... then on the way PeiLi complain to mi abt how she cum feel like going for VB lessons today... she is so stress... reach sch wif Mustaqim... so tired lo... didnâ€™t feel like doing the programming for today, haiZz but the fac put mi as the leader so no choice la, havta do my job... In class I was racking my brains for the codes to the buttons and everything... den by the there and end of the day I was really mentally tired... den abt the floorball, it was having a meeting... so they had a friendly match, WA!, really long time nv touch floorball le... so FUN... but like a lot of malays in floorball.... but it was really long time since I played floorballl... so fun lo... finally a day of fun and happiness since the holidays has end...  '),(25668,3766692,'01,July,2004','2004-07-01 00:00:00','   haiZz... Jus finished my RJ... how to motivate myself... I jus write what came into my mind... no time le... okok... this morning rain lo... quite wet la, den go inside class den kept sneeezing... felt sick and hungry the whole morning... i tot i lost my jacket, saw it beside the window... oh ya, den got this new fac in our class... found him bascally ok la... the whole day keep thinking of one qns... go guitar of wind symphony, somemore i was a bit sick... so the whole day like nv do much things in the team... i was such a let down... sorri team... den i didnt hear much of the facilitor comment on the teams and the 6th presentation.. haiZz... i m worried abt the clashing of the ccas, guitar and RPWS, i like both veri much... and badly want to keep both Ccas... A lot of people encouraged me to drop one of the similar music ccas... but i jus dunwan... as i know how to play the flute in sec sch already... i begin to like it more and more, so i dunwan to stop practicing... and guitar???!!! could someone ADVICE me??? (BTW, the next instrument which i wanna learn is the drum set, it\'s jus amazes mi... how the player synconise so well!)is my dream instrument to play, and mow i finally got a chance... so how to choose i everyday got troubled face, today worrying about the clash of my ccas... i wondered wat tml will be... XP  '),(25669,446275,'28,May,2004','2004-05-28 00:00:00','    I just don\'t understand how all of these minorities - blacks, immigrants, women - got the right to vote.  Now, despite how despicable they may be, blacks and immigrants are a topic for another time, right now: women.  Women just don\'t really need to vote.  Everybody knows that all that they do [or should do, at least] all day is sit around the house and take care of the children while cooking and watching soap operas.  They aren\'t informed of political issues, they don\'t know what is going on in the world.  And, by golly, they should not be.  Just think, if these females started watching the news they would stop taking care of our nation, Under God\'s children.  In addition to this, every single time the woman has sexual relations [because, you see, the average woman is a whore.  Every good Christian knows that one should only have sex for procreation and one should never  enjoy  it.  That\'s against God\'s way.] it\'s equivalent to her selling her vote.  Because everybody knows that women don\'t really think, they just latch on to others\' opinions.  So if this hypothetical woman was having sexual intercourse with a man out of wedlock [obviously they are Liberals because only Liberals have premarital sex.  this is another instance that the death penalty would come in handy] and he whispered in her ear, \"Oh woman with whom I am fornicating, won\'t you please vote for that Liberal candidate John Kerry?\" she would do it.  This is not the way to a good democratic system.  In another growing trend, women are getting entirely too uppity lately.  Now they want to kill their babies and get upset when the Good Christians of our nation tell them that this is unacceptable behaviour.  As one of my fellow authors pointed our ina previous post, life - as well as the soul - is formed at conception and if is the worst sin  ever  to take a human life.  These baby killer, anti-life Liberals are just the source of all evil in the world.  Masturbation is also, as a commenter pointed out, \"evil and unnatural\".  When people masturbate it is exactly like killing children.  And everybody knows why the Liberals want to kill children.  Because they are afraid that the children will grow up to be good law-abiding Republican Christians in our United States Under God Our Holy Master of America.  Just another string of examples of why Liberals are the Agents of Lucifer, the Devil and Enemy of our Lord.   '),(25670,446275,'25,May,2004','2004-05-25 00:00:00','    [this is a special guest submission by our dear friend Justin, who is obviously brilliant just like me.]  On May 22, the World Health Organization succumbed to the mass liberal Jewish gay conspiracy and raised the threat level of God\'s smite upon the gays, AIDs, to a higher level. Although they have not officially been inducted into the Axis of Evil, it seems to only be a matter of time.  Showing once again their uncanny ability to buckle under even the slightest of pressure, WHO shows yet another step in the increasingly slippery slope of liberalism infecting us in America today. Citing work by Dr Lee Jong-wook, a suspected Arab sympathizer, the liberal-leaning \"international\" organization sought to overcome the \"stigma\" and \"discrimination\" of having HIV or AIDs.  This is where I must disagree. AIDs is God\'s punishment upon homosexuals, and any stigma it incurs is well deserved. After all, in a recently conducted study, over 90% of America\'s favored a heterosexual president over a homosexual president. Disturbing God\'s natural order would be like sticking a very personal part of your body into a beehive - of course, I wouldn\'t put it past liberal\'s like Bill Clinton!  Officials from WHO also spoke about the need to \"prevent and treat HIV/AID.\" Well, yeah, if by \"prevent\" you mean \"stop engaging in homosexual sex\" and by \"treat\" you mean \"hellfire and damnation.\" You say your average American doesn\'t know this? Well I\'ll be darned, and the public school system was doing so well!  Sorry to bore all of you that pay attention to Fox News with this - I know you already know.   '),(25671,446275,'24,May,2004','2004-05-24 00:00:00','    Another thing that is terribly wrong with Liberals:  they allow websites like  urlLink Erowid.org  to exist.  As we all know, marijuana is the scourge of the devil.  It steals young children from their religious beliefs and leads them down the path of drug abuse and whatnot.  They start smoking the marijuana and they enter the forbidden dark gateway of  harder drugs .  They start stealing and they can only have fun by messing up their brain functions.    Is this what you want for the future of our population?  A bunch of doped up zombies wandering around looking for their next fix?  I believe in the deaht penalty and I believe in it for drug users.  Their type should not be welcome in our United States Under Our Lord Jesus Of America.   '),(25672,446275,'16,May,2004','2004-05-16 00:00:00','    I come back here to make another go at freeing people from their terrible liberal views and, what do I see but a comment claiming that gay people cannot get married.  Well, my friend, there you are incorrect.  As  urlLink Orson Scott Card said , homosexuals certainly can get married, they just must find a member of the opposite sex willing to marry them.  It\'s as easy as that.    Also, the 1st and 14th amendments may protect you but not if you\'re breaking the law.  Unless you are in one of the states where  urlLink engaging in homosexual behaviour  is legal, you really have no protection there.  The profile of the commenter claims that he is moving to Kansas in July, if this really is the case, I\'d liek to declare that  I  am from Kansas and I know for a fact that being homosexual is illegal in our great and noble state.    I can also see that the Liberals are not only scumbags but also wimps.  They hide behind their \'anonymous\' name in the comments of our great weblog thus showing that they cannot really stand up and declare their true opinions without attempting to disguise themselves.  If your views were worthwhile you would not state them anonymously.   '),(25673,446275,'16,May,2004','2004-05-16 00:00:00','    As any good Christian knows, cloning is evil because clones do not have souls. Only people born naturally have souls, and they get these souls at the moment of conception when they become real people just like you and me. (unless you are a democrat, they don\'t have souls)   Now, those of you who know anything about me know that I think trying to save endangered species or protect the environment is a very bad idea. It would cost good, Christian, corporations a lot of money, and if God wanted all the animals to live he wouldn\'t have made them so delicious, or their habitats so fun to destroy.  Now you are probably wondering, what do these two evils, cloning and the environment, have to do with each other? Well, I\'ll show you. Some crazy liberals have combined the two to commit the most heinous of sins,  urlLink cloning pandas !!  I find this appalling, naturally. It\'s unnaturaly and against the will of God! First of all, Pandas are Chinese and therefor they are communist Athiest bears! They are not worthy of saving!  And even if they were saving Glorious Bald Eagles, that would not justify trying to play god with their evil, secular, science!   '),(25674,446275,'15,May,2004','2004-05-15 00:00:00','    A previous comment rather rudely pointed out to me that America not following the Geneva Convention \"makes us look like bullies to the rest of the world.\"  My question to you, my conservative friends and liberal enemies, is, why shouldn\'t America be a bully?  Everybody knows that all of those little worthless third-world countries such as Zimbabwe, France, Iraq, Afganistan, and England would not get by if not for us!  Who do they export their goods to?  Who gives them aid when thye need it?  Who pays for them when they cannot pick up a resteraunt bill?  Yes.  That is all the Holy United Conservative States of America.  If you are not with us, you are against us.  That is the way it\'s been since this great country was started and this is the way it shall always be.   Also, of course Bush is a War Hero.  He is sending Americans to Fight in Iraq because he is too busy leading the country.  If he was not leading the country, our great and noble America, then all really would be lost.  We could not survive if we did not have him for president.  This is why he, himself, cannot go and fight.  He is too patriotic to do that.  Another thing about the liberals that I absolutely hate.  Waht is it with their insistence that \"under God\" should be taken out of the Pledge of Alliegience? I mean, if the founders of this country had not wnted it there, thy would nt ohave put it there.  What next?  Take our holy Lord, Saviour of Mankind, and Holy Father, our God off of money?  Out of God Bless America?  What will these liberals want?  We already know that if we give them an inch they\'ll next want a country.  Those liberal scumbags.  They should have been the fisrt to be put in the concentration camps.   '),(25675,446275,'15,May,2004','2004-05-15 00:00:00','    If those hippy liberals have their way, gays won\'t just be invading our homes on \"Will and Grace\", but they\'ll be invading the holy sanctity of marriage!   urlLink Sign the petition  and  urlLink buy a t-shirt  to show those sodomites what we think!   '),(25676,446275,'15,May,2004','2004-05-15 00:00:00','    I was reading a printed version of Fox News the other day and I heard of terrible atrocities being commited in our great nation.  By liberals.    The gays are marrying each other in Massachusetts; Kansas, the one state that anybodyin the old days could rely on for good, old-fashined conservative behaviour, voted AGAINST an amendment banning gay marriage, and all of the United States is in shambles.  Because of the liberals.  That John Kerry character is trying to defeat our Great Lord and War Hero, George W. Bush.  If that isn\'t the darndest thing then I just don\'t know what is.  George W. Bush stood up to those damn Iraqi terrorists that caused Septemeber 11th (along with the homosexuals, of course) and those Islamics deserve waht they got from our noble American forces in the Middle-East. They got nothing more or less than what they deserved.  Let\'s not forget that they brutally chopped off the head of one of our Americans AND bombded our World Trade Center.  Everbody who is anybody knows that this Kerry monster is a fraud, anyway.  He THREW his medals at the WHITE HOUSE to protest the VIETNAM WAR.  But, you know, he was greedy and didn\'t really throw the medals that he\'d earned (earned by lying and cheating, that is.  unlike our George W. Bush who wanted the war in Viet-Nam) he threw copies of the medals or just the ribbons.  Waht a fraud.  He makes me sick.  That Kerry bastard.    Wow, I was listening to the radio today and some of that music should really be banned.  It\'s just profane.  I dont see how anybody can be allowed to listen to it.  All of their euphamisms for the private parts of males and females just makes me uncomfortable.  Plus, all of that Satanic music (rock, rap, metal) really is quite offensive.  Next teenagers will stop going to church like I do every three days.    Then our country will really go downhill.  Love forever, Anastasia (me@ithinkso.org)   '),(25677,3927367,'31,July,2004','2004-07-31 00:00:00','   If there is one thing I have learned it is this: The moment you think everything is going right for once, it all goes wrong. When that happens you have a choice to make: sit down, cry and wait for your life to end, or forge ahead and keep going. If you decide to cry and wait for your life to end, I wish you luck and I pray the end comes swiftly. However, if you choose to keep going, never look back, except in happy times and with friends. If you look back at your life in sorrow, you just may change your mind and decide to sit, cry, and wait...  '),(25678,3927367,'28,July,2004','2004-07-28 00:00:00','   With both older siblings gone all the time I am basically the oldest child. Actually, cassidy is home a lot of the time but she refuses to help, so that leaves me. I guess all this pressure is just building, and it is hard to handle right now.  B.J. is a good child, my parents golden child basically. He has followed everything they have ever taught him. In a few months he is going on his mission. Golden child. I don\'t really resent him, in fact I look up to him a lot, but I don\'t appreciate the way my parents want the rest of us to turn out as good as he did. Whatever. I\'m gonna do things my way. Always have, always will.  Cassidy has rebelled against everything my parents have ever taught us. She has really hurt my parents, and me for that matter. She really resents my brother, and that ticks me off. She verbally abuses my younger sister, and me when i talk to her. She is not all bad, horrible-ness, but she has sent my family into turmoil many times. Right now she has us all walking on eggshells so that we don\'t set her off, AND THAT REALLY PISSES ME OFF!! Why the hell should we all be on edge just because she has a ****ing anger problem? I\'m so sick of her attitude running this house. One more year and she\'s off to college.  Then there is me. I feel that because cassidy refuses to help around the house, my parents expect me to pick up the slack; watching the kids, cleaning, etc. I also have this lingering feeling that because their first daughter has rebelled so much, especially against our church, that they expect me to be their perfect daughter. I don\'t like this at all, because I am not one to conform to anything. I have my own way of doing things, my own thought patterns, my own style, everything, and I\'m not gonna change just because I don\'t live up to my parents standards. If my mother wants a prodigy she has two other daughters she can try and mold.  '),(25679,3927367,'26,July,2004','2004-07-26 00:00:00','   Yes indeed, this was a weekend of firsts. The first time i\'ve been on a weekend trip with my friend and without my family. The first time i\'ve seen the ocean. The first time I\'ve eaten halibut, shrimp, clam chowder, lobster, and crab. The first time i\'ve been allowed to walk around a city after 7, much less a city i did not know. The first time i\'ve ever gone dune-buggying. Fun stuff...sort of.  The ocean was so gorgeous. It went on forever, and the clouds were hanging above it, but way out to sea. It also seemed like the clouds were trying to mimic the magesty of the waves, because they were curving around eachother. They were rounded off a little, just like the rolling surf. It all made me feel very small, and very alone at the same time. The thing i wanted most of all&nbsp;this weekend was to have Tyler with me so that we could&nbsp;build a small bonfire, roast \'mallows, and just sit and watch the sun sink below the horizon.  '),(25680,3927367,'21,July,2004','2004-07-21 00:00:00','   So, my life has been absolutely crazy as of late. Ever since monday my garage has been full of drummers, pretty much non-stop. We have been painting our drums charcoal grey. Cool cool. Or not.... We\'ve run into a million problems, not to mention my family is being stupid about it. That and I have only gotten about 7 hours of sleep since waking up on Monday morning. That\'s an average of about 3.5 hours of sleep a night. It is messed up! Oh well, I\'m thriving I guess. So, my parents pretty much hate me right now because painting is taking so long, and my little siblings aren\'t handling the change very well or whatever. It\'s all retarded crap. Moving on, tomorrow my family is having a family portrait done, then we come home, I get to work on the drums for about 2 hours, and then hopefully I\'ll get to go to my boyfriends house and watch band of brother, or at least part of it. Then on Friday I\'m leaving with one of my best friends and we are going to the ocean. Newport to be exact. I\'ve never seen then ocean in my life!!! I am so excited. Well, I got to get back in the garage. Some of the drummers make me nervous. Laterz.  '),(25681,3927367,'17,July,2004','2004-07-17 00:00:00','   Hey all. Goodness, today was a great day. I went to walmart with my grandpa and he let me drive his really nice focus. Usually i hate driving, only because i still have my permit and therefor must drive with one of my parents, and both like to yell at me when i do something wrong. Or even when i don\'t for that matter. It sux. Well, with grandpa it was so much better. He explained things to me, he didn\'t yell or get mad. It was so cool! Once we got home it was about 11:30, and i had to be at the highschool around 12:45 for our drumline rehearsal. Also, before i left i had to make pizza dough, so when i got home around 5 i could just roll out the dough, slap on the toppingsd and away we go. Well, between the dough, loading up my quads, putting away the paint for the drums, and saying goodbye to my little siblings, i was almost late. I made it there just after our section leader did, though. I felt special. Anyways, i had forgot to put sunscreen on before i left, and there was no shade around. It was about 103 outside during the rehearsal. Not fun. It did cloud over for a little while, but not long. Actually, for about 15 minutes we hung out in the dugout. That was fun. We played and marched, and marched and played until 5. I got home, made dinner, welcomed my older sister home, and that would bring us to this point. Yup, i\'ve had a really fun day. The only thing that would make it better is if i could see, or at least talk to, my boyfriend, but he\'s out hangin with his friends. That\'s ok. I\'ll talk to him on monday hopefully. Well, my grandma\'s leaving, so i\'m gonna go. Laterz all. Keep on keepin on.  '),(25682,3927367,'16,July,2004','2004-07-16 00:00:00','   I love finding nemo. That movie is so cool. My friends went nuts with that movie and nicknamed everyone they knew after a character from that movie. I was the little fish that, when introducing himself to nemo, says \"I\'m obnoxious!\" Yeah, that name is pretty self explanatory i guess.  Well, last night was the worst night i&nbsp;can ever remember having, hands-down. It was ok to begin with, we had dinner, the little kids didn\'t scream which was a first. Then my little sister who is 11 got invited to go sleep over at a friends house. My grandparents said ok, so she went her way. However, with her gone that meant that i had to do two paper routes this morning: hers and mine. Usually my mom does both routes with us, but she was gone, so my grandpa was gonna do it. That meant that I would have to run to every house. (we drive the routes in a car. Lazy, i know, but it gets it done faster.) Oh well, i decided, at least my little sister would be having fun. And so the night progressed onward. Generally the kids go to bed at 9 or 9:30. Last night they didn\'t go to bed until 11:00, all because my grandparents had put on a 2 hour movie and the kids wouldn\'t go to bed until the end of it. I didn\'t really mind. I thought the more tired they are the easier they\'ll go to sleep. I was wrong. Dead wrong.&nbsp;  To put the kids to bed some one has to read three stories, sing three songs, and say prayers with them. Well, that someone is my mom who gives into them a lot, and only because both boys are autistic and half deaf. Autistic or not, they are extremely intelligent and should not be treated like babies. They\'re 6 1/2 for crying aloud! I told them last night that i would read stories to them in the morning, when i could actually focus my eyes to read. Bryce was ok with that, he just read himself a story, sang a song, said prayers, and went to sleep. Chase totally went to pieces. He was screaming for mom and refused to go to sleep until she came home. She doesn\'t get home until tonight, around 6. So i told him whatever, i don\'t care, sleep on the floor with all your toys, it really doesn\'t matter to me. So he slept on the floor for about 30 minutes, then crawled in his bed. I won that battle easily enough. Then there is Hannah. There is nothing wrong with her, except an extreme attitude that needs adjusting. I read her book but would not&nbsp;heat up the rice pack for her. Why in the world does she need it anyways? Oh. That\'s right, she doesn\'t! Well, she did this cry/whining combo that i could not stand. I turned around and started walking toward the door. All of a sudden both Hannah and Chase start crying and screaming that they hate me. For some reason it really hurt. I closed the door, laid on my parents bed, turned on my music and tried to tune them out.  Around midnight Hannah gets up and says she has to go to the bathroom, so i take her. Then she automatically thinks that she gets to sleep in my parents bed with me. My mom evidently lets her sleep with them, but i\'m not my mom. I told her no, and she started her whole cry/whining thing again. I can\'t stand that cry because it is so fake, honestly, it is. It\'s all for show. I put her back&nbsp;in her bed where she commenced to cry for another 15 minutes, and then fell silent.  All in all, last night was absolutely terrible. Tonight when my parents come home i am not gonna help with the kids at all. I must say I am so sick of them. I have been stuck at home for the past 10 days straight, with only them for company. I have only left the house to go to the dentist and my paper route. Oh well. That\'s all in the past. This next week, starting on Monday, the drumline is coming over to my house and we are gonna be painting our drums. It\'s supposed to be around 103 all next week, and we\'re gonna be in&nbsp;my garage, but it\'ll be fun.  Well, today is a new day, with new chores and responsibilities. I have to clean the downstairs, bathroom, and upstairs, not to mention run to the store and get some groceries, paint brushes and tape. I have to do this by 1 because that\'s when the drumline rehearsal begins. I gotta get crackin\'. Keep on keepin on.  '),(25683,3927367,'16,July,2004','2004-07-16 00:00:00','   This whole day i have been yo-yo-ing between anger and depression. I was angry this morning because i had to take all three of my youngest sibnlings to swin lessons, and sit in the hot sun. I was also wearing jeans and a black t-shirt, which was stupid. When i got home i had the biggest headache, and i felt so tired. I laid down o my bed, feeling thouroughly ticked, and was soon asleep. I woke up to my grandma pounding on my door, asking me to help her make lunch for the little people. My mom, dad, and older siblings were gone, so i had to help.&nbsp; That turned into a fiasco that i won\'t even describe. This afternoon my older brother left for eugene, oregon. My parents got home, and 3 hours later left for the same place. Tomorrow they are going to all watch my older sister and her cohorts perform at drum major camp. Then they are bringing her home. In the meantime, i am the oldest child left at home, but both my granbdparents are here too. For some reason i don\'t see eye-to-eye with either of them. they think they are right in everything, and though they may be, they don\'t have to force their ways and opinions on me. Tonight i have to watch ove the little people, getting their dinner, giving them baths, getting them to bed, and in between all that keeping them entertained. I\'m feeling realy put upon right now and i don\'t like it. Tomorrow i have to get them all ready for the day, keep them entertained, and find some way to get down to the high school at 1 for our drum line rehearsal. I don\'t understand, whats going on?! This day s seeming to last forever, and i\'m not looking forward to tomorrow. I just wish i could see, or at least talk to, one of my friends. I haven\'t seen any of them in a long time. I know i saw my boyfriend just last night, but i haven\'t talked to him at all today, and i wish i could. I can\'t seem to stop being sad, and it\'s driving me nuts. I have this desire to be happy, like everyone else seems to be, i just can\'t seem make it.  '),(25684,3927367,'15,July,2004','2004-07-15 00:00:00','   Yesterday I couldn\'t sleep, last night I couldn\'t sleep, and today does not show much hope either. I don\'t know what\'s going on, maybe I\'m worried about something, I dunno. Oh well. Yesterday was fun. I mowed the lawn, cleaned the whole downstairs of my house, got my little brothers ready for swim lessons, and practiced the quads, all before lunch time. After lunch I had a quad sectional from 1 - 3:30. That went pretty well I guess. Our music is all paradiddles, paradiddlediddles, rolls, and messed up splits and stuff. Don\'t worry, I\'m gonna have a whole different post that only concerns drumline, cuz there\'s a lot of crap going on there. Anyways, around 4 my boyfriend was supposed to come over so we could take my three youngest siblings&nbsp;to a park&nbsp;and get them away from my mom because she was totally freakin out from stress and whatnot. Well,&nbsp; evidently he called around 3 and my mom didn\'t tell me until about 4:15 that he was waiting for me to call him.&nbsp; That ticked me because 1) my mom always gets mad at us for forgetting to give her messages and yet when she does it to me she gets pissed because I get mad at her. double standard? I think so. oh yeah, and 2) because my boyfriend had said he would come over at four, I had it already&nbsp;ok\'d with the parentals and everything. Then on the phone he had the gall to ask \"so can I come over?\" Well,&nbsp; that was a retarded question, but whatever. He came over, we took care of the kids, had dinner, went to McDonald\'s, and then came home and hung out with my family/watched Emperor\'s new groove. Fun fun fun. Actually it was fun. I\'ll post again later. certain people keep looking over my shoulder and it\'s ticking me off!  '),(25685,3927367,'15,July,2004','2004-07-15 00:00:00','   It\'s 9:08, and I\'m awake and out of bed. This is a definite low point for me. 7:00 this morning my mom kicked me and my siblings out of bed. Why, you may ask. For the sole purpose of making us work. My two little sisters, ages 11 and 3, are at swim lessons while my two little brothers, ages 6 and 6, are destroying our downstairs living room. My older brother, age 18, is out mowing lawns for his business. My mother is at the dentist, and my older sister, age 17, is currently 6 hours away at a camp. That leaves me to tend the hooligans that are at the moment \"making up stories\". One is using Thomas the tank engine, while the other is using barbies. Basically they are talking aloud to themselves right now. They also have Mister Rogers on, and Mister Rogers just got a shot. I hate *twitch*  needles. But you see, if I were in bed right now, I wouldn\'t have seen that, and I wouldn\'t have to listen to my little brothers talk to themselves, and I wouldn\'t have to fold laundry or clean the house either. Oh well, at least now I can practice for our Quad line sectional, which is today.  '),(25686,3927367,'14,July,2004','2004-07-14 00:00:00','   Holy Kracker! I just got back from my youth group and I feel so sick, and yet extremely hyper at the same time. It\'s a weird sensation, but anyways. We played the chocolate game. You sit in a circle and there is a tray with a hat, scarf, two oven mits, two forks, and a heavily wrapped giant Hershey bar in the middle. You pass around 2 dies, and whoever rolls doubles has to put on all the clothes, take the forks and rip the wrapping off the chocolate bar. They then commence to try and cut up the bar and eat it. All the while everyone else is still trying to roll doubles. We played this game twice, and between the two games I ate about a whole giant candy bar. I am so sick of chocolate! At one point I had so much chocolate in my mouth at one time I couldn\'t breathe and my up-chuck reflex started to kick in. Not a good combination, but I\'m alive. We also played the bubble gum game. This is where you have a plate in front of you. On the plate is one piece of bubble gum, generally bubble yum because it is bigger than normal gum. Whoever eats the gum and blows a bubble first wins...but there a catch. The piece of gum is buried underneath a  huge  mound of whip cream. My parents say I\'m not American just because I don\'t like whip cream, but I don\'t care. I hate the stuff with a passion, and yet I have a very strong natural instinct to not lose. A very strong competitive streak in me I suppose. Anywho, needless to say I toughed it out, and came in a close second, but I\'m still satisfied with that. However, as we were sitting there just chewing our gum, the girl next to me says \"hey andie\" as if wanting to start a conversation. I turn to look at her, and just as I opened my mouth to say \"hey what\" a whole plate full of whip cream lands in my face. It was so disgusting! I was pissed to the max. I threw my plate at her and it got all in her hair and on her shirt. It also got all over the floor too. We were forced, against child labor laws I\'m sure, to clean up the mess, but we laughed it off afterward. But don\'t think for a minute that I\'m not gonna get her back even worse for what she did. I also have a rather strong revenge streak in me too.  Oh, yeah, and my leaders are gonna get it too. As we all were stuffing our faces with chocolate, and as I was getting a face full of whip cream, they were taking pictures! Gaaaaa!!! Evil gits. Oh well. If I get any copies I might post them, but then again I might burn them. I dunno. I\'m still really hyper. I need something to do. I might go burn about ten of my candles at one time in my room. Fire is fun.....  '),(25687,3927367,'17,August,2004','2004-08-17 00:00:00','   Yesterday was a very good day, or as good as my days will be for awhile. I basically vegged out all day, until band practice. We were there at 5 - 8:30. After that we went to burger king. That was really nice, considering i hadn\'t had lunch or dinner. After burger king i went home with one of my friends to spend the night, but we didn\'t go to her house. We actually went to a house that they are house sitting. It was so big and pretty, that and they have a really cute dog. His name is toby, he looks like the ice cream \"brown cow\", he\'s a polmeranian, and he only has one eye. He\'s adorable! He slept next to me and it made me happy. My friend and i talked until 2:30 in the morning. We just sat there and talked. It was the best thing i have done in awhile, and it helped me so much. We pretty much decided that we are gonna make this next school year totally awesome, and i totally hope we have at least two classes together. Dang it. that reminds me, registration is this week. That totally awesome, completely incredible friend that i am talking about, i know you\'re gonna read this, so do you wanna try and plan a time to go to registration together? I don\'t really wanna go alone and you know you\'re gonna be the first one i tell my schedule to, so why not have you right there? Anyways, so yeah, Last night really helped. I feel somewhat refeshed, like i now have the energy, and will, to keep moving forward. I\'m not saying things are perfect, because they are far from it, but i can now look at the world and say without fear \"Bring it, moron!\"  '),(25688,3927367,'16,August,2004','2004-08-16 00:00:00','   He\'s gone. He\'s really gone. I spent most of saturady with him. We went to see the village with his brother, matt, and matt\'s girlfriend allie. It was a pretty cool movie. Then we went home, and i just sat around for the next 2 hours, trying not to distract him from packing, which he still had not done. I played with his dogs, talked to his mom, and just hung out. Then we had a steak dinner and his friends (shay, seth, and jeremy) showed up. They certainly made things interesting. After dinner we all went into the study and watched strongbad emails from  urlLink homestarrunner.com  and from  urlLink joecartoon.com . The gerbil cartoons are the best. I would suggest joe fish. That one is hilarious. After all that the guys went to carls junior to get more food. Damn bottomless pits. They left Tyler and I alone. We just sat in his room and talked about everything and nothing. I had done so well the whole day, and then mrs.goodnight came in and said that matt would be taking me home at 9:30, which was an hour away. She left and I just totally broke down. It was completely ridiculous and retarded of me. I have known ever since last October when we first started going out that he would eventually leave for college. I know a part of me knew this, but i think another part of me didn\'t want to believe it. I believe it now. His mom and him ended up taking me home that night. I watched him drive away, then i went to the kitchen and sat in darkness for a half hour. I then went to my bedroom, sat on my bed, and cried. I know i fell asleep sometime after midnight, but i\'m not sure when. This is only the third time that i can ever remember crying myself to sleep. Sunday morning came. I got up to do the paper route. My mom was ticked at me because i hadn\'t gone downstairs saturday when i got home to tell them that i was home. She said that she didn\'t even know that i had gotten home. \"Well, that\'s your fault now isn\'t it.\" is what i was thinking. While waiting to go to church i was sitting downstairs. My dad asked my i was so melancholy. \"You use to always be my little ray of sunshine!\" he said. \"Yeah, when i was what, three?!\" was my reply. As these little excerpts of my day may show, i was in a very depressed and angry mood. Sunday was a black day. I did get to talk to tyler though. At nine i called him, but it just kept ringing, so i hung up. a minute after i hung up the phone rang and it was tyler. He\'s doing good. Having a fun time. Hanging out with dalari and whitney. I love whitney to death. She\'s so funny. I really dislike dalari though, for retarded reasons of my own that only one other person knows about. Unfortunately, i can\'t help but thinking that once his school really starts, and he gets busy, i\'ll either be mostly forgotten, or completely forgotten. Whatever happens happens i guess. All i know is that he is gone, but such is life. And time moves on.  '),(25689,3927367,'13,August,2004','2004-08-13 00:00:00','   Omigosh! I just got back from bandcamp, like, 30 minutes ago. It was so tight. This whole past week has been one huge blur, but it was so freakin worth it! We had three of the hottest drum techs alive, two who just got back from marching corps. They were so great, they pushed us hard, but they didn\'t break us. They came pretty darn close, but not quite.  At the beginning of the week I almost had the opener memorized, now I have it down really well, but doing it with drill it kinda difficult. I\'ll get it eventually. The third song is messed up, so I am gonna have to practice that one hard core. It is 180, and has so many beats. It\'s insane! Oh well. Maybe I\'ll get that one eventually too.  So, right now I am in more pain that I can ever remember being in. My feet keep cramping up, so do my calves. I have a slight sun burn on my shoulders, and my wrists are dieing.  Drill was fun, but we were hardly ever inside for anything. The average tempt this past week was around 100-102. At one point it was 106. Yeah baby!!!!! We made it! I have a killer sock tan too. I wore the high ankle socks just so that I could really see the tan. I know it\'s stupid, but I didn\'t wear sun block all week, and I only got a shoulder burn, which is good cuz it\'s not even that bad.  The last night there we had a potluck thing where all of our families could come. My family came, and so did my boyfriend. I hadn\'t seen or talked to him for almost a week, so I was really excited. What sucked though is I only got to talk to him for two minutes, then my mom started freaking at me because she couldn\'t find chase, my little brother. She was getting mad at me, as if it was my fault. It\'s not my problem that she can\'t keep track of all of her freakin children! Anyways, I had to go find him, which I did. He was with this girl named shaylee and they were looking at the turtles in the wildlife area. I went back to the cafeteria to talk to tyler again, but he was gone. That made me a little disappointed and pissed at the same time, but I had to go change into my shirt and study my drill anyway I guess.  We went through the whole rehearsal. It was fun...sort of. I was gasping for breathe by the end of it, but that\'s because of our last three sets. They are all double time, and you have to haul but in all of them. I would have to say that they are probably my favorite sets though.  After the rehearsal we had a dance with just our band. Walking out there I saw my brother and his friends that had just graduated. I asked bj where tyler was and he told me that tyler went home. He hadn\'t said a damn word to me at all, not even good bye or anything. I was pretty mad, but I decided that I would have fun at the dance without him. The dance was pretty cool, mainly because the drummer stuck together and we all just hung out there. I had fun anyway.  Next day was freakin hot, but we had an awesome rehearsal for three hours, then the drummer clan went to burger king. Real food, finally!!! It was great. Then I got home and that would bring us up to this point.  '),(25690,3927367,'06,August,2004','2004-08-06 00:00:00','   This day has been the worst in a very long time. Last night\'s sleepover was ok. There was someone there that I have never gotten along with, but limited interaction with her kept me good. This morning I was on call (babysitting) for a very special woman I know. I thought that the latest I could be back was 4, so that I could go to a movie with my friends. Come to find out, they were all busy later, so they had to go to the 1 o\'clock showing. I didn\'t find this out until noon, when my mom called me at the peoples house that I was babysitting at. She said that Tyler had tried to call me telling me that they were going at one and no later. This really upset me. That is actually an understatement. I nearly broke out in tears infront of the girls that I was taking care of, but I was able to hold it in. I kept hoping that the girls\' mom would get back before 1, but no such miracle occurred. She got home at 1:15, I was home by 1:20. I was devastated. When I got home, my parents and the little ones were all packed up and getting ready to leave for wenatchee. They left, which left me alone. I looked for my drumline music folder...And never found it. I searched this whole damn house, and never found it. Tyler called during my frantic search, and the tears that had been threatening to come all day finally found an exit. We didn\'t talk much, but just knowing that he was there helped a little. I had to get off the phone, and I resumed my search. Again, to no avail. Cassidy came home not long after. She told me about her day, then listened to mine. She then decided that we needed to watch a movie together. She chose finding nemo, and I fell asleep on the couch with her sleeping head in my lap. When I woke up, I had that same sense of hopeless nothingness. Also, this whole time it had been raining.  I  didn\'t miss it because I wasn\'t in a movie theater, but it wasn\'t all it was cracked up to be. So, in summary so far, I had babysat, felt rejected by my friends, left by my family, lost the most important thing I own right now, and had to listen to the miserable rain for hours on end. Needless to say I have spent most of the day crying. While in this attitude of despair, I decided to call Tyler and ask if we could do something. So, he invited me over and we \"watched\" princess bride and weezer/stone temple pilots/multiple other music videos. I had such a good time there. I know that in a week he is gone, and that whole week I am at band camp, but while I was with him I forgot all that. I like to sleep so much because while you sleep, you forget. You forget everything that has been happening, you forget why you were angry, you forget the world. Tonight, being with Tyler was like being asleep. I forgot everything but him. He was all that mattered, and as long as I was with him, everything would be fine. Even though I am at my own home now, he is not completely gone. In fact, he will never be completely gone from me. Knowing that is what gives me the strength to keep going.  '),(25691,3927367,'05,August,2004','2004-08-05 00:00:00','   Tyler came over last night. It may have been the last time I see him before he leaves, but I pray it is not. We watched a walk to remember, and then the end of emperor\'s new groove. My little siblings were pissing me off and stressing me out when my dad came in from outside. He asked what we were doing. \"Catering to their needs!\" Was my reply, and it had quite a bit of anger in it. I love my dad so much, because he said \"no, that\'s not right. boys, upstairs now! Leave andie and tyler alone.\" So they left , but don\'t worry, we weren\'t alone. My little sister, chelsea, age 12, was still there. Oh well. We watched the movie together and that is all that counts. Then we had to go upstairs for prayers so that the little kids could go to bed. Earlier that evening tyler had given me a present, but I refused to open it in my angry state. The little kids, however, would not go to bed until they saw what it was, so I had to open it. When I did I was completely speechless. It was a porcelain Tinkerbell with 24 carat gold inlay. It is of Lenox make, so it is extremely beautiful, fragile, and expensive i\'m sure. He got it for me when he went to canada last weekend. This is the link to a picture of it...I hope.  urlLink http://www.lenox.com/images/pg/product_images/087992_w3.jpg  Links and pics don\'t usually work for me, which is sad. Anyways, tinkerbell is currently in her box hidden in my room, away from the kids, until i can get a shelf made to put her on. After the movies we went outside, looked at the stars, and just sat there. Eventually he left. I didn\'t want him to go because I knew that it would mean that the day was over, and we would be one day closer to his departure. Needless to say I cried myself to sleep last night.  '),(25692,3927367,'04,August,2004','2004-08-04 00:00:00','   I hope that whoever coined the phrase \"change is good\" is rotting in hell right now. Yes, change is necessary, but good is not the word that I would use to describe it. Occasionally change is good, but in my experience it is mostly unpleasant.  When change occurs in my life, is all seems to happen at once.   Last Sunday I found out my favorite youth leader and closest \"adult\" friend was released, so she won\'t be working with us anymore.  this past week, all summer actually, I have been getting ready for band camp, the whole drumline has, and yet the line-up for the line is still not rocksolid. We have 2 people who are complete flakes, so even the line is changing.  A week from Friday one of the coolest families I know will be moving to Spokane. The Lewis\'s have been here since before my family moved here, more than 15 years ago. Margeaux, one of their daughters, is the sweetest, funniest, prettiest girl ever. She and I went through drivers ed together. Great times, great times.  The Sunday after band camp, and two days after Margeaux leaves, my boyfriend is leaving for college. This will be the hardest change to deal with. I love him so very much, and I am going to miss him even more. We have decided to try and stay together, and I truly hope it works out. We shall see.   These are a few of the changes that are about to happen. I do not even wish to touch upon the ones that have happened, and especially not the ones that are happening now, because those are the hardest to talk about.  '),(25693,3309198,'16,May,2004','2004-05-16 00:00:00','   for anyone out there that enjoys the arcade game DDR, there is a computer version called Dance With Intensity. If your interested go to http://dwi.ddruk.com/downloads.php you\'ll have to download the song Not My Battle by gundrill as well. For now thats the only song I can find for free, but I\'m sure if you look you\'ll find more somewhere besides its Hella-Hard!(especialy whe you have to use the keyboard >   '),(25694,3309198,'12,May,2004','2004-05-12 00:00:00','   Just letting everyone know the incredible capabilities of the Nintendo DS,Heres one blurb! With its two screens, Nintendo DS instantly shows its innovative face. But capabilities like a touch-screen input, voice recognition and wireless communication set it apart from every other game device, and again demonstrate Nintendoâ€™s passion for groundbreaking originality.  Game players need only use their imagination to see how the DS promises to transform the portable game industry by changing the way gamers relate to their games â€“ and each other. Two screens offer two perspectives on the action at once. The touch screen could make accessing items, moving characters or navigating menus as easy as a tap or drag on the screen. Voice recognition could let players simply tell the game what they want it to do. Chat software will let users transmit text messages, handwriting and even drawings to one another. And wireless functions could link players in the same room â€“ or across the country. The unique features of the Nintendo DS hold a wellspring of potential for the future. Game players will be wowed.   Check it out for yourself  at http://www.nintendo.com/e3_2004/ds/index.jsp It\'s incredible! And heres a comment from the President of Square-Enix (Creators of Final Fantasy!)   â€œI believe that Square Enix has a promising chance in the platform, since Nintendo DS allows users to experience whole new game play with both its input system and its output system.â€ â€“ Yoichi Wada, president and representative director, Square Enix  Damn this is gonna dent my wallet ^^;   '),(25695,3309198,'11,May,2004','2004-05-11 00:00:00','   LIke I mentioned below ( The newest post seem to go to the top) Im a gamer, so I thought I\'d suggest a couple sites, some with no relavence to Games at all ^^; www.gscentral.com is full of codes for all systems and even Action replay, www.lunarknights.com is my Final Fantasy 11 Linkshells Site, come here to generally talk about the game, go to www.homestarrunner.com and then go to Sb E-mails for some good laughs, www.battleon.com is a very limited online rpg,www.gunbound.com (it might be .net) is a cool online game similer to Worms armageddon, or go to 2flashgames.com for a lot of cool games, last but not least check out www.newgrounds.com with more....\"mature\" comedy Don\'t forget to look at Tommorrows Nobodies and Retarded Animal Babies!   '),(25696,3309198,'11,May,2004','2004-05-11 00:00:00','   Well If Im Going on about myself ( What else can I do on this site? ) I may as well tell you about some of My Fav TV shows: South Park, The Simpsons, Family Guy, That 70\'s show, Fooly Cooly, InuYasha and Futurama   '),(25697,3309198,'11,May,2004','2004-05-11 00:00:00','   Well now that I\'m actually posting, I suppose I should tell You about myself, Im 13 (14 in July)I live in Ridgeway, Ontario A small town fairly close to Niagara falls. I play a Saxophone, alot of Videogames,and I\'m a DDR freak ^^;   '),(25698,3309198,'11,May,2004','2004-05-11 00:00:00','   Hey all! Though I highly doubt anyone I havent told about this so called \"Blog\" Will visit it, I suppose making it couldn\'t hurt, besides I can still show people I actually know! At the moment I don\'t really know how to do anything other than write, so if someone knows how to add picturees and stuff please inform me! Umm this is a test, so If your bored by now, I suppose thats good ^^;   '),(25699,3602089,'10,June,2004','2004-06-10 00:00:00','   Yes, you read right...Some Jolly Rancher Lolly Pops ARE made more equal than others.  For example, green apple tastes like chemicals and makes me cough, on the contrary, watermelon is light and soothing and is quite a plesant experience.  '),(25700,3602089,'10,June,2004','2004-06-10 00:00:00','   Today, I attended the Laurel Festival once again that is held every year in Wellsboro, Pennsylvania.  On \"The Green\" they have many tents full of people selling neat trinkets, hand-carved woodwork, lawn-decorating pieces, tye-dyed clothes, and hundreds of other unique products.  I bought two dollars worth of marbles which appealed to me...especially the orange and white ones that look like cream/dreamsicles.  In the north section of \"The Green\" are approximately fifteen food stands; They have everything you can imagine...german bratwurst, greek gyros, candy apples, mongolian noodles, old-fashioned orangeade, bloomin\' onions, sausage, meatballs, cheesesteak, pie, pierogies, cheesecake, chicken sandwiches, and other exotic food stands.  This was the list of my \"gastric-hoardings\" from today in order of first to last eaten from 1pm to 5pm...  1.)  Old-Fashioned Lemonade 2.)  Half of a funnelcake 3.)  A huge strawberry shortcake (real strawberries, real biscuits, and really       rich, buttery, whipped topping) 4.)  Greek potato salad - perfectly cooked warm-ish olive-oiled potatos with       leaves/flakes of parsley and pepper 5.)  Old-Fashioned Lemonade  I am kicking myself now because I decided against buying some oriental mongolian noodles for fear of barfing near the \"marble tent\".  I will just have to wait until next year.  As I listen to the news they are talking about \"Bill Clinton\'s Autobiography: My Life\"...haha...what a freaker...And for DINNER (7:30pm), dad made a delicious bbq chicken sandwich with pepperjack cheese, garlic, and romaine, all on a plain flour bun...  '),(25701,3602089,'10,June,2004','2004-06-10 00:00:00','   I had so many carbs and calories yesterday...420 calories worth of chocolate milk, a whole bread stick and huge slice of lasagna, a clearly canadian blackberry soda, a whole basket full of french fries, a bbq sandwich w/ two slices of white bread, four twizzler sticks (all flour and corn starch), and two slim jims.  I felt so sick...but now, today, I am hungry.  I think, perhaps, I should go on a forty-day fast and not eat anything.  Fasts are quite healthy, because, at any given moment, your body is carrying five to ten pounds of toxins and other bad things.  Fasting releases these harmful byproducts of your body making you healthier.  I\'m nost sure which DVD I want to watch tonight... either Office Space, Fantasia 2000, Behind Enemy Lines, or The Others.  Everyone told me it was going to be a nice day when I knew it was going to be crappy.  I woke up to ugly skies but now it is sunny...wtf is going on?!  I believe that the events that take place in the movie \"The Day After Tomorrow\" are not too far in the future from occurring.  We have all of these screwed up weather patterns that do not necessarily follow anything that we\'ve seen in the past.  Barcodes are awesome... www.upcdatabase.com is my website of the day. I\'m out...  '),(25702,3602089,'10,June,2004','2004-06-10 00:00:00','   Today was filled with restlessness as I struggled to find other things to occupy the hours.  I got up approx. at 10:25am and started taping Ronald Reagen\'s funeral for my mom at 10:37am...it had started at 10am.  I got online to see if my friend in Niedersachsen (Dissen, to be specific), Germany, was on or ÐšÑƒÑ€Ð¾Ñ‡ÐºÐ¸Ð½Ð¾ ÐÐ°Ð´ÐµÐ¶Ð´Ð° Ð¡ÐµÑ€Ð³ÐµÐµÐ²Ð½Ð°, my friend in St. Petersburg, Russia.  Neither were on but I chatted with my mom who was at work at the courthouse.  I finished watching the commentary version of \"The Craft\" on DVD, which is one of my favorite movies of all time!  I tried renaming and reorganizing my mp3s that i had to transfer from my iPod to my new computer hardrive which will be a strenuous task, seeing as how the names somehow change in the transfer process... I really wish there were a Starbucks within at least 30 minutes of here.  There are three in one city an hour from here which is not fair.  The one-pound-burrito restaurant, Chipotle: there is not one in the whole grand state of Pennsylvania.  I had one while in Baltimore on a field trip and it was extremely tasty.  OH DAMN!  I forgot that the stand up comedians on on tonight!...hmm...I hope I\'m not missing any good ones!...My condolences go out to Nancy and the rest of the Reagan family...he truly was a compassionate human being.  '),(25703,3602089,'10,June,2004','2004-06-10 00:00:00','   ÐŸÑ€Ð¸Ð²ÐµÑ‚!  I\'m glad that this day is finally over.  If all of those preps hadn\'t been in my last period class of the day I would have been fine.  My internet has been going extremely slow so I had to reconnect about four times.  Now I am talking to one of my friends and she is telling me that she heard a \"bad but true story\" (hmm...) about me that shows exactly how much of an \"a-hole\" I really am...that\'s almost funny...although, she is sworn to secrecy not to tell me the story or who told her.  And she just said that the other main person is the true a-hole in the story and not me, after she thought about it.  Hmm...I want to finish watching The Craft, but I am too tired to watch the whole thing with the commentary without passing out less than five times.  You know, water doesn\'t really keep your breath fresh.  I wish we had central air in this house...it becomes quite humid at points and the air flow is almost non-existant unless you were to knock down the entire front wall.  I wonder if the gas prices are going to go down and I DON\'T think that the power is going to be handed over to the Iraqis by the deadline.  That new Coca Cola (C2) is kinda good.  It has half the carbs and half the sugar...although it contains phenylalanine, which is the fake sugar in diet products that causes cancer in lab rats... Ð´Ð¾ Ð·Ð°Ð²Ñ‚Ñ€Ð°!  '),(25704,3602089,'10,June,2004','2004-06-10 00:00:00','   Hallo alle personen!  Wir sind in die Schweiz!  I we have had a great flight with no delays.  We checked out a swiss clock shoppe and bought lots of neat stuff in the village that that was in.  The wood carvings all over the villiages are amaying...most houses are over 400 years old.  Surprisingly, in the hotel that we are staying at (in the villaige it is in) there is a musik festival going on that we have been to...very interesting but since we cannot drink beer, there is not much to do there but walk aound and look at the sidewalk shops and food stands.  The language barrier doesn\'t really create a problem...usually only machines.  Today we checked out lots of waterfalls as well, but I must go now...only 40 seconds left...yak  '),(25705,3602089,'10,June,2004','2004-06-10 00:00:00','   From today, there are only four more days until I go to Europe.  While at the mall, I went in to Holiday Hairstyles to get a haircut and saw my one friend from the skating rink getting her hair done.  She is being a bridesmaid for her one friend this afternoon.  Her hair was halfly-braided, halfly-twisted, with small butterfly clips everywhere and small curls in the back.  As soon as I step foot onto Paris concrete, I am going to buy fifty-four crepes...fifty-four because everyone else always picks a stupid number like \"fifty\" or \"fifty-two\".  Why did Coca Cola get rid of their \"Surge\" line of soda?!  I thought it was quite good...I know it\'s weird for me to be talking about it now since it happened many months ago (before Christmas), but I am just now realizing how much I miss it.  I like most all sodas except for Diet Pepsi (Pepsi One is ok though...).  I love the vanilla colas.  If you like both Vanilla Coke and Pepsi Vanilla, check this...v. coke contains artificial vanilla flavorings while pepsi v. contains pure vanilla extract.  Also, I mentioned in an earlier post that C2 contains phenylalanine that causes cancer in lab rats...Pepsi just introduced Pepsi Edge which has half the sugar and carbs (same thing as C2), but does not replace it with the fake sugar!  Amen to that, Pepsi!  They seem to be thinking more about what\'s \"healthy in an already bad-for-your-body soda\"...which is enough for me.  Even though Pepsi tends to go the natural way, I still buy Coca Cola more now just because the shape of the bottle is more appealing to me than the regular Pepsi bottle...lol...what a tart...  '),(25706,3602089,'10,June,2004','2004-06-10 00:00:00','   Since I got my iPod from Apple about three months ago, I have gotten very poor quality output of bass and terrible sound distortion on my favorite kind of music, Russian Techno (specifically, the artists Ð ÑƒÑÑÐºÐ¸Ð¹ Ð Ð°Ð·Ð¼ÐµÑ€, Ð’Ð¸Ñ€ÑƒÑ, and Ð ÑƒÐºÐ¸ Ð’Ð²ÐµÑ€Ñ…).  I have used a little music format converter for a while and decided to fool around with the bit rate and file type.  I converted some .WMA files into mp3s because Apple won\'t play windows media audio formats...go figure.  Lol...anyway, I discovered that the songs sound perfect, except there is still not enough bass to satisfy me.  I don\'t exactly know what I did differently this time to change the format because all of the settings are the same as last time I did this.  Anyway, I have only eleven more days until I fly to Europe for twenty-one days, which is going to be very fulfilling for my travel craving for a few months.  The drivers\' saftey classes I have been taking all of this week and will continue to for the length of next weeks is total hell.  The teacher acts like a younger version of Mr. Rogers on delta-9 tetrahydrocanabinol, or even crack cocain...it doesn\'t matter, he\'s a little too happy in the mornings but is annoying at the same time, which is nearly impossible for the rest of us.  Now, I must look for my camera tripod before my trip so I don\'t miss the perfect long exposure shot of the Notre Dame cathedral...this house is such a mess I\'ll never find it in the piles of shiz flooding my \"space\"...  '),(25707,3818324,'28,July,2004','2004-07-28 00:00:00','   \"Why don\'t you just crucify me, nail me to a cross. Bite and scratch and make me scream, if that will get you off!\" Lyrics from London After Midnight.  Those Lyrics are how I feel right now. Today I\'m not in the fucking mood to beat around the bush about how I feel. I\'m fed up with so many things.  I\'ve been sick with a bloody stomach flu type thing since Monday, and well being lathargic and having no one there to just hold you is so fucking annoying. What am I bitching about? I\'m sure any female who reads this has bitched about the exact same fucking thing, so its nothing you havn\'t read before, and if you havn\'t read it truthfully then you\'ve thought of it.   My relitaves just got here frmo England, they\'re staying with us for three days. *spins a finger in the air sarcasticly* whoopty fucking doo. I guess its no big suprise, I havn\'t seen them in five freaking years. The only good thing I can think comeing out of this family rondevouz would proabbly be that I got my bass tuned.  Lets see, the only real development in my life besides that... Nero. I\'m going to Toronto to play Nero. The highlight of my summer, that and Renn Fest. This summer will not have a bloody highlight so far that I can see. I\'ve seen my friends twice, well two of them anyway. I\'ve been talking to the others on MSN, well at least the ones that don\'t think its ok to ignore me because it IS infact the summer.  One male I do have an attraction to (that is suprisingly more than physical) was sopposed to go to Nero aswell, but it turns out he\'s working. So much to my dispare I won\'t get to spend the weekend with him. At least there are other people there I guess... though I&nbsp;was looking forward to it. Unfortunately its yet another thing that does not turn out how I\'d like it to  &nbsp;   '),(25708,3818324,'03,July,2004','2004-07-03 00:00:00','   Wonderland, heh, my Wonderland is a place where I have emotions that don\'t land me in a horrable place after. I\'m writing this from the true and black depths of my black little heart. Everything I say is true, and I need to get this out, regardless of who or what reads it:  Emotions just fuck me over in the end, its all god damn pantomime! Never once have I ever felt anything for anyone that was returned without a slap in the face! The only person in my life, I ever loved, who loved me back to the best of my knowledge gave me what I thought the point to be the best gift ever. Whenever I was depressed he\'d just let me sleep in his arms for as long as I needed to. Or he would cry with me when everything just became too hard. Now I\'m nothing! Its better to have loved and lost, than never to have loved at all?!? Thats bullshit! Because once you\'ve loved and lost it, you have nothing at all!  Those times when he used to hold me are now haunting me. We used to go to a friend\'s house and just sleep in the couch, not doing anything... just sleeping, or crying, or shareing the sweet serenity that was each other\'s silent thoughts. Now when I look at a couch, or a bed, with the intention of sleeping I\'m just haunted by everything. I can\'t sleep, sometimes I wake up wrapped so tightly in my blankets I think he\'s there, sometimes I can even smell his lavander after-shave. Its all pantomime, pretending somethings there when its not.  I\'m the Queen of Fools, to think that with every possible male I might find that something I\'m missing. I\'ve yet to find any mate in life that holds up long enough to see me cry, none do, and I doubt any ever will. My expectations are too high, especially for one such as myself.   I\'m 17 years old, I\'ve loved once and lost ever since. This Wonderland is just a dream, a world where people mean what they say.. and emotions are all they\'re cracked up to be.  Its pathetic I know. Trust me, I know.   '),(25709,3818324,'01,July,2004','2004-07-01 00:00:00','   First let me introduce myself. I\'m Arsenic or the Black Umbrella Girl... either or works. I\'m 17 years old, metaly ill so my doctors say, though I think they were just being sarcastic, and some mixed up mess between gothic\\punk\\and rockstar. Don\'t ask about rockstar, people get the impression I\'m famous from my appearence and mannerisums I guess. I don\'t look like one, most of the time I look like I just woke up... because I probably did.   Well, well... The unfortunate day has come... I\'M HERE! Though more than likely anyone who reads this has probably scrolled over it looking for some kind of porpose, I\'ll tell you in advance There isn\'t one!  Lets see, anything I can write about? AH yes! I spent the entire day at Niagra trying to find some sort of word to discribe all the bordom suffered. Then I realized bored is about it. It really isn\'t a pretty thing to sit next to Lake Ontario and eat lunch. I wasn\'t sure weather I\'d vomit or die. The water smelled like raw sewage, there was dead fish in the water... and people were letting their children swim in it! *shudders*  I have nothing to do at all, Canada day used to be reserved for sparklers and alcohol... but I rarely drink anyway, and I find playing with sparkers ends in setting things on fire. However I\'m not allowed to play with sparklers anymore, and a grapefruit cooler is about all I really want. So looks like I\'ll be home watching sick humored movies and writing stories.  Well thats it for my first blog thing. *tips hat* Untill next time! Arsenic Malice, signing out! PAINT YOUR HOUSE BLACK!   '),(25710,3496136,'30,June,2004','2004-06-30 00:00:00','    on my way to a meeting out of the hospital stopped at Wendys to pick up a salad.  Crash!...a guy in a big pickup truck crashed into me...well, my car.  He crashed into my newly washed and waxed Saturn.  This was shortly after being turned down [ever so gently] by Mark Groves for lunch today.  Honestly, I\'m just really going to stop asking friends for shoulders...disappointed at every freaking turn.    I\'m fine and the car has only a little red paint on it...but jeesch..  Could this day get any worse?  You know, if I were premenstrual or something I\'d at least understand my seemingly uncontrollable emotions...but COME ON!  Is this some sort of test?    Maybe it\'s time I explored the joys of psychotherapy and better living through chemistry....   '),(25711,3496136,'29,June,2004','2004-06-29 00:00:00','    Went to  urlLink this site  and took the quiz...they are called lemmings, I think.  I am an Anarchist...apparently  \"You are a WEDL--Wacky Emotional Destructive Leader. This makes you an anarchist. You don\'t give a damn. When push comes to shove, you just forget about it--it\'s just not worth the heartache. What this means for others is that dealing with you can be aggravating, because they find they can\'t get you motivated about things they care about. What this means for you is that you are happier, calmer, and saner then they are on their best days.  You are near-immune to criticism, and those who know you well acknowledge and respect that. You may come across as lazy, but the truth is that you find little to get worked up about. Regardless, you have slews of friends, because they are fascinated by your world view, jealous of your lifestyle, and drawn to the fact that you are hilarious to be around.   You are a pillar in a sea of hot-bloodedness. You have a sweet tooth.\"  I\'d be willing to bet that my answer would be different every time I took this quiz..I\'m mercurial that way.  In a really foul mood today, mostly because I had to get up really early today...be at work by 7:30.  Taking pictures for the department website.  I\'d post the link but I dislike being scoffed at.  I arrive on time ready to go and are the #@%#@ residents there?  Did I remind them yesterday?  Yes.  Did I send them email two days prior to remind them?  Yes.  Did I also notify the attendings so the residents could be excused from clinical services?  Yes.   Am I fuming?  Can you see the smoke coming out the top of my head?  Less is more...breathe in, breathe out...trudge onward.  And why am I truding?  WTF, why bother.  It\'s not like anyone listens to me or cares what I think.  Sigh...  I might be over-reacting.  Okay, I\'m probably over-reacting.  Sigh...wish...nevermind...why bother wishing...been down that road and it leads nowhere but to heartache.  Attempting good mood...last night at rehearsal I got to sing with a full orchestra - cellos and everything.  After we finished the number the director says \"Val, don\'t oversing.\"  I said, \"I\'m not oversinging.  It\'s not my fault the orchestra can\'t play any louder.\"  *wink*  The brass section cracked up...and so did everyone else.  Am having a good time in the show...but it\'s clearly community theatre.  I suppose it will be enough to feed my hunger for performing in the coming years.  Commedia opened \"In a Minute Men\" at the Just Off Broadway Theatre last night.  Hope it went well.  Hope Chris will call me today to let me know how it went.  I doubt I\'ll have time to reach him.  Residents....grrrrr...  Suppose I should turn my attention to the ingrateful little....darlings.  Less is more...breathe in, breathe out.  Sure could use a hug.   '),(25712,3496136,'29,June,2004','2004-06-29 00:00:00','    I sent the following to the residents:   \"It\'s Time for the Annual Great Locker Move.  Lockers are on reassigned based on year.  Please see Dr. K for locker assignment changes.  Once the dust has settled, I will assign the interns locker space.  Please have all your stuff moved by Thursday.   Val\"  Dr. K then writes her very first email to the group:   \"Please note.  Dr. B will not be able to vacate until after this week as she is interviewing.  This should not be an issue because there are many equally lovely lockers to be had.  If one has grand designs for Dr. B\'s locker you may purchase my favor with chocolate or sworn allegiance to George W. Bush.    Thank you, Dr. K  Cursed Texans.  So much for my dream year of teambuilding and a gentler kinder residency program.  I\'m pretty sure I\'m really in for it this year. I\'m proud of her display of humor...at least the residents are learning something from me.  And it\'s probably not professionalism.  If you\'re not having fun WTF?  Right?  ;)   '),(25713,3496136,'28,June,2004','2004-06-28 00:00:00','    Seymore is married to MotherPockets.  He has such a positive outlook yet realistic at the same time.  A blog I like a lot.  Here\'s his latest...  \"Most of the gravest judgment lapses we all ever allow to muck up our lives come from one simple idea. We are going to miss something, and it is going to be great! It is something that would have made our lives complete and made everything make sense. Anxiety, infidelity, addiction, gluttony so many things find us reaching out from a good place and risking it all because there might be something more. Well, let me set your mind at ease. You are going to miss something. For that matter you will miss more than you ever experience. That is the infinite nature of the universe. Get over it. If you really experience what you have before you, you will have more to carry in you heart and head than many who rush from one thing to another, notching something as they go along. I can promise you, as one who has seen more of the world than most people will ever come close to seeing, who has had nearly all the ups and downs that this life can offer, you are going to miss something. I did. I can live with that. It is grasping for something that becomes important to so many, not the thing itself. How many people can you have sex with? How many eclairs can you eat? How many trips can you take? Once the individual experience has been had, did it add to your world picture? If not, that was more of a waste than doing nothing. Give yourself a break. Enjoy what you have. No, don\'t give up on doing new things, but don\'t do them because you are searching for that \"perfect\" experience. Make what you have perfect and you will have lived a life that will matter more than all the awards on your mantel and all the medals on your chest could ever give you. You don\'t need something from without to complete you. You just need your own permission. \"  Hmmm...wonder what makes people feel incomplete?  Danielle today in the car said \"When will the \'struction (construction) workers be all done with our road?\"  Indeed, when will they?   '),(25714,3496136,'28,June,2004','2004-06-28 00:00:00','    Today I had lunch at the Blue Koi with my new residents.  It\'s a noodle place.  Don\'t know how else to describe it...sort of Asian, not quite Chinese, not quite Thai.  Yummy, probably best describes it.   It\'s a hip, trendy smallish place in the 39th Street corridor.  All the waiters have piercings, dyed hair and tatoos.  The decor is comfortable and light.  Tile floor, booths along one wall, rag rolled walls, dark blue ontop of a pristine white.  Lots of Asian accents...ecclectic Asian accents.   The food was incredible.  Had boiled dumplings that came with a complex sauce that wasn\'t exactly soy and wasn\'t exactly sweet...I had a hard time not picking up the bowl and drinking it.  Lettuce wraps filled with a delicious chicken, basil, ginger concoction.  Minced marinated pork over bean noodles...called Ants on a Tree.  mmmmm....mmmmm...good.   The best was the tea.  They had easily 20 different kinds of teas.  Hot teas, bubble teas (tea with bubbles of flavored tapioca...I\'m not adventuresome enough for this one yet), iced tea...mmmm.  I love hot tea - chinese hot tea.  This was a real treat...I tried a couple of different teas.  8 Treasure was my favorite, a sweet bright taste.  The taste is still on my tongue.  I feel refreshed and rejuvenated.    I would like to spend the afternoon in the comfy booth sipping tea and writing.  Looking out the window at the passersby...maybe go browsing at Prospero\'s Books?    And about the writing...I finished the Wild Minds book and the take home message in a nutshell was...  keep writing, doesn\'t really matter what. don\'t take yourself too seriously, you\'ll quash your creativity when you sit down to write, you have to clear the clutter.  She recommends writing at least 5 pages of stream of conscious thought before beginning tackling the current writing project.   I think those are all good ideas to try if you\'re stuck.  Writing is such an individal personal activity...I think you have to do what works for you.  Gotta start somewhere.  I like the idea of cleaning the clutter out of my mind first though.  I never start cooking unless the sink is clear of dishes, why would writing be any different?   '),(25715,3496136,'27,June,2004','2004-06-27 00:00:00','    Marissa has moved into her new home.  Can\'t wait to see it and take pictures which I might share if people ask politely.   Christine and Dean are in a film and are filming today!  Christine is the 4th screaming villager - which suits her just fine.  She was born to be a screaming villager.  Dean is a sword swinging blademaster...an evil one.   less is more...  comfortably numb...   '),(25716,3496136,'27,June,2004','2004-06-27 00:00:00','    I think  urlLink this  is a brilliant idea.  I\'ve always wanted to get involved in a project like this.  I think recreating a trip down the Mississippi on a showboat [with gambling] would be just the coolest thing.  I\'ve always loved history.   to that end...not the showboat thing but the history thing.  I have a new project at work that I\'m really excited about.  In conjunction with the History of Medicine Library, I\'m going to interview Dr. Krantz on videotape for an hour a week.  Dr. Krantz was chairman for 35 years of the Department of Obstetrics and Gynecology.  He\'s 82 and crusty as a French bagette.  He curses up a blue streak and uses the most politically incorrect language that we have to keep him away from the litigenous medical students.  He adores me though.  For the last 10 years everytime he comes to KU [about once a week] he stops in my office to tell me a joke, each one more obscene than the last.  His penchant for filthy jokes aside, he has done a lot for women in the State of Kansas throughout the years, and he was a critical player in the desegregation of KU Medical Center in the late 60\'s.  I\'m looking forward to learning more about the interview process and sharing his stories.     Wizard of Oz last night was most entertaining.  Had a bit of problem though...at intermission went to take some of the stuff [copious blankets, chairs and pillows - it\'s an outside ampitheatre]to the car and I couldn\'t find the freaking car.  I walked all over that @$%^! parking lot and I couldn\'t find it.  I finally gave up and went back in with all the stuff.  Enjoyed the rest of the show - the cast of Oz contained a chorus of 120 children...not a typo...120 children!  Dang, it\'s a huge stage...which will make a difference in my characterization [NOTE TO SELF:  Be Bigger].  Waited for the crowd to clear somewhat - figured that would make the car easier to find.  Well, we found it but when we got to the car the keyless entry thingy didn\'t work.  You know, the button you push that says *unlock*.  So I used the key, which set off the alarm...the car wouldn\'t start.  Sigh...so I had to wait 15 minutes to start the stupid car.  It was now midnight...dark parking lot with two small children.  Maybe I should have taken up karate instead of tap.   We did arrive home safely without further incident.   Raining today...going to bake cookies and play in the pop-up tent with the girls.  Reading Beverly Clearly to them \"Mouse and the Motorcycle\".  Also bought  urlLink Wild Minds  for me.  A friend who teaches playwriting recommended it to me.   I love  urlLink Boondocks .  Comic with an edge and a wry humor.    Listening to Madison read the comics outloud to her daddy.  Danielle is now *training* Mechante, which means the two of them are wandering around the house searching for trouble...better go clean up what they have left their wake.   '),(25717,3496136,'26,June,2004','2004-06-26 00:00:00','    someone should do something about the balance problem with sprinklers.  You attach it, you carefully place it so it won\'t fall over, you turn on the water and voila...the sprinkler turns on it\'s side and you have to either get wet [don\'t get me wrong, I love being wet just not when I\'m already dressed for work] or go turn water off and start over.  I have devised a rock system to hold the f-ing thing down...but you would think some clever human would have pattoned a cool invention to get a sprinkler not to fall over.  Sigh...listening to the sound of my girls giggling as they run through the sprinkler.   I am comfortably numb.  My relaxing weekend so far has involved 7 loads of laundry, weekly trip to the grocery store, mowed the lawn, and washed both cars.  Ah...relaxing.  I don\'t really know how.  Just trying to keep my brain on off switch so I\'m keeping busy.  Wizard of Oz tonight with the girls at Theatre in the Park.    Feel...nevermind.  No stewing, no ranting allowed.  Less is still more.  Riiiiiight.   '),(25718,3496136,'25,June,2004','2004-06-25 00:00:00','    urlLink Bill Clinton explains himself  in Time mag...just in time too with  urlLink The Passion of the Bush  coming out tomorrow and I didn\'t make up that pseudo title - I just think it\'s funny.    and looky... urlLink 50 cool websites ...    going to write this weekend...yup...and smell some roses too.  A.   '),(25719,3496136,'24,June,2004','2004-06-24 00:00:00','    Less is more...less is more...less is more.  Why doesn\'t it feel like more?  And....nevermind.  *breathe in, breathe out*  No stewing, no ranting...*breathe in, breathe out*  Less is more...less is more...less is...   '),(25720,3496136,'24,June,2004','2004-06-24 00:00:00','    Today is one of those days that I\'ll keep asking myself \"SELF, why didn\'t you just stay in bed?\"  Really wish I could get some sleep.  Somehow after nearly 40 years I still don\'t know where the off switch to my brain is.  I\'ve been kicking around a couple of story ideas.  But I don\'t know where to start...do you just start writing and see what comes out on the page?  Isn\'t there supposed to be some sort of outline?  Character sheets?  Research?  I guess it doesn\'t really matter, as I\'ll be writing it just for me.  I suppose the approach to writing is just as individual as the individual writer.  Or any artist, really.  Hmmm....finding time to write though, is the trick.  Last night at rehearsal we ran act two.  I\'m in act two a lot more, so the evening went by faster and we didn\'t start till 8.  Mark, who plays Eugene - my brother [have I mentioned yet that he\'s 6\'5\" and is built like a scarecrow?] slipped and said \"In our fart of the world...\" laughter rang through the dance studio where we rehearse.  My next line \"Excuse me, may I say something?\" which I changed, of course to \"Excuse me, do you smell something?\"  Last night the music director asked me to sing off stage on all numbers except solos, which was flattering.  I\'ll be body miked.  They are a fun group.  I haven\'t really taken learning the lines seriously...a very bad habit from hanging out with improv actors.  They gave me line notes last night...every time you call for line or transpose words, etc.  I said \"You know I\'m an improv actress and can\'t learn lines, right?\"  wink.  Sigh, I\'m going to have to buckle down though.  I called for line last night and couldn\'t hear her.  Gee, what a surprise.   I finally blew up the inflatable swimming pool.  We had some nice family time squirting each other with water.  Mechante got loose.  Two teenage boys were chasing her.  They were bent over breathing hard and one of them said to me \"Man, she\'s fast.\"  Well duh!  She is a greyhound.  Fence, must build fence.  Plans for the weekend:  None.  I\'m staying flexible.  I would like to take the girls to Theatre in the Park this weekend sometime to see Wizard of Oz.  But we shall see what happens.  Lawn needs mowed, flower garden needs to be weeded, and I seriously need to relax.  I\'m not going to plan though...usually I make lists on Friday night of all I want to get done around the house and stuff I would like to do over the weekend.  But so not this weekend.  And I don\'t want to spend a lot of time thinking either.  Thinking too much lately.  Drunken stupor maybe.  Yeah...might be time for a drunken stupor.   '),(25721,3496136,'24,June,2004','2004-06-24 00:00:00','    While working on a database, \"My Val, you\'ve certainly ensured the referential integrity with that update!\"  Wow, I have integrity, at least referentially speaking...now that I\'m updated.   \"Hey Val, text page me your cell number.  Nevermind, I\'ll just IM you\"  Doesn\'t anyone use a plain old telephone anymore?   Lunch discussion - ever wondered what gynecologists talk about at lunch?  Usually I would warn you to never, ever, ever ask...you\'d be shocked and appalled...but today was just downright funny.  You see I was bragging on my baby brother, who is a career man in the Air Force.  He has just taken this exam to be a Master Sargeant.  He passed it on the first try which is quite a rarity I understand.  This led the discussion to commissioned officers and non-commissioned officers....and...[random people speaking following here]\"How come they never promote people to be officers for just being damn fine soldiers?\" \"Like Audie Murphy.\"  \"Audie Murphy?\" [obviously a young pup resident] \"Wasn\'t he a movie star?\" [young pup who knows several of our faculty are veterans and have done their homework][Rant on...Sometimes the extent to which residents will go to brown nose amazes me.  The heirarchy of medical education is completely ludicrous.  They try nobly to advance young men and women based on objective criteria but the plum assignments just come down to who plays the game the best and that damnable old boy\'s network...and...well, I\'d better stop now before I get into trouble.  Rant off.]  \"Well, he made just one film, I think  To Hell and Back .\"  \"Wasn\'t there a sequel?\"  \"Yeah,  I\'m Back From Hell \"  Laughter.  \" Just Go To Hell \"  \"You talking to me?  Yooooooou talking to ME?\"    Ah...doctors can be such commedians sometimes.  At least they haven\'t totally lost their sense of humor.   '),(25722,3496136,'22,June,2004','2004-06-22 00:00:00','    Some idiot went the wrong way on one of the bridges this morning and ended up killing him/herself as they crashed into 5 cars.  As a result, most of the traffic was merged onto one bridge which my normal 20 mnute commute was one hour and 15 minutes.  GRRRRRR!  With a 4 year old in the back seat...so that meant Barney music, cause I can\'t find my Shrek tape!  Now I can\'t get that voice out of my head....wouldn\'t it be awful if that was the last sound I ever heard?  Barney....*shudder*  My Horoscope today...I hate it when they are scary \"You\'re putting up with a lot -- make sure you\'re at least comfortable. Start sorting out what you want to get rid of and what you want to keep. Less will be more.\"  Hmmm...less is more?  I thought it was whoever gets the most toys wins?   Friendship...now there\'s a vague term.  The eskimos have 50 words for snow.  We have just a few words really for those that are important in our lives..love, friendship and family.  But how do you define those relationships?  Is definition really necessary?  I have been thinking a lot about those definitions because in the face of looming crisis we tend to lean heavily upon those we love.  I just don\'t want to lean on the wrong person and accidentally crush something.   Last night at rehearsal was the costume parade.  My scene partner and I have rented costumes that won\'t arrive until July 5th.  We sat around literally until 9:50 last night (that\'s 3 hours and 20 minutes for those of you that are numerically inclined).  I did get involved in helping the teenaged dancers dress.  They couldn\'t figure out which way to wear the hats...I tied lots of bows and adjusted a lot of chaps.  Woo-hoo!  I love chaps!  Hard to sit around watching rehearsal and not direct.  I was in a vibe with the director though because just as I\'d mentally think a directing note...he\'d stop and say what I was thinking.  Watched the 7 minute dance number to \"I Got Rhythm\".  Wow, 65 people tapping on stage at once...wow.  Sure wish I was in it.  Waaaaah!  Okay, residency crap calls.   '),(25723,3496136,'22,June,2004','2004-06-22 00:00:00','    Am looking forward to the Michael Moore movie.  I like it when art creates debate.  I never really believed in art for art\'s sake...art should have a higher cause than self promotion.    So I\"m not sure if Farenheit/911 is classified as a documentary...but I intend to read what the right has to say about it like here in this  urlLink critical article .    I do admire Michael Moore\'s tenacity and for standing up for what he believes in.  I\'m not sure I like his style of debate which inclused more name calling than fact but I can certainly understand getting caught up in the heat of momment.  Which was why I was never good at debate myself...I was good enough to win stuff in high school but not quite good enough to please myself.   So today is gloomy here in KC.  No more nights off from rehearsals so the grind begins.  I\'m kind of excited though about performing on that huge stage.  Can\'t wait for opening night and feel the adrenaline, the rush and then fill the stage with energy, light and my voice.  I\'m supposed to be miked - I\'ve never been miked before.  Never needed one.    Marissa is closing on her house today.  I\'m very excited for her.    Am going to speak to the owner of Prospero\'s Books down the street about commedia performing there.  They have poetry readings and music and other performance art on Friday and Saturday nights sometimes.  It would just be Marissa, Chris, Christine and me.  It\'s a small space.  I think it would be fun to do improv for a book crowd and we all have a different literary background so that would be quite entertaining.  *Crossing fingers*  Mark wrote a script for the JOB show \"In A Minute Men\".  Haven\'t read it yet but Chris, the newly shaven, is pumped about it.  I can\'t believe Chris shaved his beard!  He looks so different.  I saw him last night and his mouth is so expressive.  His goofy looks are much funnier now.  I teased him about getting him some *product* for his hair and that he needs to get it *shaped*.    He\'s going to play Ichabod Crane this fall at the Gem Theatre near 18th and Vine.  I\'m happy he found some work.  He\'s talking about trying to lose another 20 pounds before then.   The Fly  Danielle, my 4 year old \"Mama, there\'s a fly on the table.\"   SWAT.  Bottom lip trembling \"Mama, that was a baby fly.\"  Heads bent over the smashed remains...\"No, honey. It was pretty big. I\'m sure it wasn\'t a baby.\"   Blue eyes wide, hand over her mouth \"Oh, it was a mama one.\"  Yes, I ended up burying a fly carefully interrned in Kleenex. With a funeral and everything.   Well, off to recruit for the JOB show.  And to do a little work for gynecology ;)   '),(25724,3496136,'21,June,2004','2004-06-21 00:00:00','    I should just stop looking forward to things.   '),(25725,3496136,'20,June,2004','2004-06-20 00:00:00','    \"Mommy, look out for the dragon!\"   What?  Oh, right.  At the SCA event the people referred to cars as dragons.  Cars apparently have become dragons for good ;)    Sleep is over-rated.   Having lunch with Mark Groves today at the Thai Place in Westport which is just about the only reason I got out of bed today.  I am physically, emotionally and spiritually exhausted.   Trudging onward...   '),(25726,3496136,'20,June,2004','2004-06-20 00:00:00','    Just back from Theatre in the Park - Sound of Music.  A nice production...I am hypercritical.  Was that a block of wood or Captain vonTrapp?  Did he really block the entire song of \"My Favorite Things\" on the bed?  Think environment and scale when you design props...puppets 18\" can\'t be seen...bigger is better.  Well, wider is always better...but sometimes bigger is too. ;)  My kids really enjoyed it though.  They were singing along.  I had to tell Madison to just mouth the words...pretend to play giraffe while you\'re singing.   I hope our show is better.  i think the direction and staging is better, definitely.  I\'m glad I\'m working with the director I am working with.   A lot on my mind...can\'t sleep.  Gonna try the warm milk - read a book method.  Wish me luck.   '),(25727,3496136,'19,June,2004','2004-06-19 00:00:00','    Used to sleep in until 11 before I had kids. Now I wouldn\'t give up the peace and quiet between 6 and 7 for much of anything. A nice cup of coffee, the paper and the dog at my feet....ah, Saturday morning on the deck. A chance to gather thoughts.   Grocery shopping, laundry and a lot of work done. Looking forward to taking the kids to the SCA Lillies War thing in a bit.    Ah...girlfriends  IM\'d with a single girlfriend last night until the wee hours of the morning. She was asking me advice about what questions to ask this guy she\'s sweet on. Heavens, we behaved like giggly junior high school girls. So very much fun. She keeps me young. It seems whenever I\'m getting in a rut and starting to feel sad she brightens things up. Ah...friends.   Children\'s Lit  The girls crawled into bed with us sometime last night during a thunderstorm. Woke up unable to move with one child in either side and a cat on my legs. What a great way to wake up. Danielle and Madison and I went to the library this morning and signed up for the summer book reading program. Each girl got 5 books. I checked out The Mouse and the Motorcycle by Beverly Clearly to read to them. Remember the Ramona the Pest books and Henry and Beezus! I was going to read The Lion, Witch and The Wardrobe but they didn\'t have it. I ordered it. Just looking through those titles made me remember how my mom took us to the library every Saturday morning. I so used to look forward to spending the morning at the library. Madison had already read one book by the time we got home. So definitely want to read Little House on the Prairie, BoxCar Children and the Secret Garden.   mmmm....Brisket  Am very happy cause Hen House was having a special on brisket $.99 a lb! Woo-hoo! Brisket for Father\'s Day. I love making brisket!    Dr. Phil  Flipping through the channels last night for a bit and heard Dr. Phil say...\"Couples often fight about topics instead of issues.  Topics would be how much golf your husband  plays or how he doesn\'t do the housework *right*.  The issue would be how you don\'t feel safe or like a priority.  Couples need to speak about what each others needs are.\"  Hmmm...did he say golf?    Well, off to shower and enjoy the rest of the day.   '),(25728,3496136,'18,June,2004','2004-06-18 00:00:00','    The Campaign  Last night I\'m sitting on the sofa watching the  urlLink Newshour with Jim Leher  (who never seems to be there anymore...but hey, they have a great online thing now).  I  decided to try to start paying attention to current events.  And how about that 9/11 commission - What!  Goverment agencies don\'t communicate well in time of crisis!  No!  Said ala * urlLink Jack Benny   \"Well\".  The two girls came down the stairs. Madison very politely asked me if I wanted to see their \"show\". \"We\'ve been hearse-in\' \" pipes in Danielle. \"Well, of course, let\'s see it.\" I said. \"I\'ll go get Daddy\". Greg was in the basement. \"No, mommy its\' just for you.\" Awwwwww.....I\'m getting a little teary now.  Madison was first, of course, she made up a song about how fun it would be to sleep with a kitten. Danielle was next, she made up a song about how kittens are soft. Madison\'s turn - song about how only evil people won\'t help kittens. Evil people like Cruella Deville which started a whole new verse about how people should not kill kittens just for their fur...she actually went into Al Jolson pose...on one knee hands clasped to her breast, pleading. Danielle was next, she made up a song about how kittens are soft...sounded remarkably like the first time...but it had CHOREOGRAPHY! I\'m so getting her tap shoes. I thought I would die laughing...it was very...lets say original. ;)   I gave them huge hugs and told them I would think about getting a kitten. The thought of them ganging up on me amuses me greatly and yet frightens me...the foreshadowing of things to come. LOL.   The playset and ramblings about tatoos  Last night got the playset delivered and set up in the back. Greg\'s mom bought it for us. The girls love it and will spend many happy hours sliding and climbing and playing in it. Greg\'s brother, Jeff, hauled it from Topeka and helped set it up. He\'s really tall and bald. He shaves his head. I wonder why men do that? He showed us his new tatoo it looked really cool. And then he started explaining it. It looked like a cut right below his right breast about 2 inches long then moving down his stomach were 12 drops of blood, one was black and it ended with a Chinese symbol. The drops were the apostles - black was for Judas and the symbol was Destiny. Which is pretty original and so not what I would get. I have thought about it and I can\'t think of a thing I would want tatooed on me...guess I don\'t have a permanent enough personality to have anything that would always remain or maybe nothing ever clicked for me. I know I would want it where no one could see it. So something small on the hip...maybe tiny letters \"It\'s all about me?\"   Yard work...yawn...  I spent a long time last night trimming bushes and then picking up the mess. I don\'t know what those things are called but they have thorns! Ouch! My hands are kind of torn up. The lawnmower is almost funtional again so I\'ll have to mow this weekend. And I have plans for the front flowerbed! I\'m going to put in another flowerbed in the front late summer/early fall. And strawberries...must plant strawberries. The less grass the better. Thinking of putting in all rocks like they have in the Southwest. ;) Since I\'ll be digging...Woo-hoo...maybe the girls and I will make mud pies too!  Greg is going out of town to play golf but I meight get him to go to the  urlLink SCA Lillies War  tomorrow (weather permitting). The girls and I are going to see Sound of Music at  urlLink Theatre in the Park , probably tonight (weather permitting). I know I\'ll be thinking of the Lord of the Rings scene we did at Sedalia last weekend the whole time and laughing to myself \"How Do You Solve A Problem Like the Nazgul?\" And Sunday I\'m going to blow up the inflatable pool and work in the yard. I\'d like to find some time to write but we\'ll see how it goes.   Still haven\'t heard from the Lyric Opera people yet.  Am having lunch today with one of the resident alumni and the senior class. It\'s tradition. We go to McCormick and Schmick on the Plaza for lunch. I give them pictures - I take lots of pictures during their residency training - mostly for the website but some just for fun- and give them printout but this year I\'m burning CDs for them. I think they\'ll treasure them.   And, dang I look good today. All dressed up in my administrator clothes, heels and makeup. I\'m such an adult...who knew?   '),(25729,3496136,'17,June,2004','2004-06-17 00:00:00','    Green day  While driving to work I ate a toaster scramble. Not just any toaster scramble - a Shrek II toaster scramble, the very kind two children desired with all their hearts and promised to be very good the rest of the entire day if I would just buy the Shrek II toaster scrambles. There was nothing else in the house to eat for breakfast, so I popped one in the toaster and ate it on the way. It was more than delicious I must say...a sweet berry flavor and fun green icing.   Turns out the green icing turned my teeth green...just like Shrek. Nothing like attending a meeting on how to prevent STDs in the KC area with green teeth. *bangs head on desk*    Madison and Mark Twain  Madison (my 7 year old) is campaigning to adopt a kitten. \"...but mommy, think of how sad it would be if you were kitty without a home, alone in the world. I\'d take extra special good care of it. And the best part is kittens are free!\" Yup, I\'m gonna start saving for law school.   We just finished reading Tom Sawyer last night. We took turns reading to each other outloud. She would read a page then I got to read two pages. I took the girls to the Mark Twain Cave last summer when we were home (I\'m originally from Hannibal). How heart-warming it was for me to share something I love with Madison and have her respond so well. When we got to the part of the book where the cave is involved and they mentioned Aladin\'s Palace Madison said \"Mommy! I remember that place! I\'ve been there!\" And she was riveted. She also likes the name Huckleberry. After we finished she asked if we could read it again. Awww...I think we\'ll read Davenport and Cherry Pie next or the Boxcar Children. If you\'re reading this, feel free to post some of your favorite childhood books.   We played war (the card game) and she cheated. Yup, law school, definitely.    S\'plosions  Danielle (my 4 year old)on the other hand, watched Star Trek Enterprise with me last night. She loves *space* shows. :) This morning after watching some of the news (we try to catch the traffic before heading south) she says...\"Be careful Mommy, there was a s\'plosion (explosion)and you don\'t wanna drive through it.\" Blue eyes wide with trepidation. \"That\'s on the other side of the world, honey.\" \"Well, you should still be careful. You never know when a s\'plosion is gonna happen.\" And she is absolutely right, you know...you never do.  ;)   '),(25730,3496136,'31,July,2004','2004-07-31 00:00:00','    urlLink Here\'s  why.  I love his  urlLink show .  He talks about not only how to prepare food, but the history, culture and CHEMISTRY of food.  And he\'s just as witty as my friends in commedia, which is the highest complement I can give.  Am rumminating on an email to commedia...about moving on, about remembering where we came from and how to heal the rift between us all.  After more long talking with Chris, am unsure of what the future holds.  No group is constant.   '),(25731,3496136,'30,July,2004','2004-07-30 00:00:00','    Note to self:  Ultimatims are probably not the route to go with him.  More when I calm down.  [ahem]  Like that\'s going to happen...*eyeroll*   '),(25732,3496136,'29,July,2004','2004-07-29 00:00:00','    completely and totally bored...  I managed to catch up 2 days worth of work in oh...3 hours.  Sometimes, this job just isn\'t challenging enough or is it that I am just not as interested?  Are there a million small things piled up on my desk?  Yes.  Do I want to do any of them?  No. What do I want to do?  Hmm...don\'t feel good enough to do anything too active...I want to chat with friends or see a movie or sit on the deck and carve pumpkins.  Gosh I can\'t wait for pumpkin carving time.  I really should take up whittling.  A good use for all my excess energy.  Yup, that\'s me good ol\' whittlin Val.  Course, no one will let me handle sharp objects....  ASSFBB yes, but Val no.  So unfair....so very, very, very unfair.  Really must get to work on my monologue.  oh, Oh, OH!  I wrote one!  I\'m afraid to show it to anyone though.  Maybe I\'ll work on it a little bit more and polish it and then...  Auditions are in 2 weeks.  Being one of Red\'s girls is a little more work than I\'d planned on.  Must buy new shoes, make bloomers and a skirt and...a wig?  A wig....or make a garland...must send Lawfully Good Man the audition information.  We\'re coming apart at the seams...but I think we can hold it together until KCRF ends.  Its just everyone is so weary of each other.  I think back to a year ago...that lovely night when we played at Hyde Park.  Rissy and I rolling down the hill.  The business meeting at Christines...such hope and dreams for the future.  We were so cohesed.  The driveway....the beginning of the end.  Still, it was such a lovely night.  Seems like a lifetime ago.  And it was.   '),(25733,3496136,'29,July,2004','2004-07-29 00:00:00','    If I ever had a way to support this  urlLink bakery , I would.   I think I love  urlLink Barack Obama  .&nbsp; He\'s David from Five Smooth Stones come to life.&nbsp; Gee, how long before he gets assassinated?  Watched Amish in the City... Because it looked really intriguing...and it was. Young Amish people at a crossroads in their lives, tasting the outside world for the first time trying to decide whether or not to embrace their culture and faith or to explore the outside world. I have always been very curious about the Amish way of life. It seems so simple and pure. To watch these young people look at common ordinary objects like...parking meters with amazement was very entertaining. The city kids come across poorly...spoiled brat, know it alls. It is interesting to see how deeply the faith of the Amish kids runs...how moral they truly are. Watching them interact with the city kids is really interesting.&nbsp; But the most compelling moments will be when the Amish kids grapple with their faith and morality issues.&nbsp; Inner-struggles are always compelling.&nbsp; They didn\'t really need the city kids...they could have build a background for us of what it\'s like to grow up Amish.   Prospero\'s Books Performing live comedy in an artsy environ...should be great, great, great fun.   '),(25734,3496136,'27,July,2004','2004-07-27 00:00:00','    At last...  A pat on the back from the boss....  Let the accreditation gods do their worst...my data has referential integrity!&nbsp; HA!!!!  Satisfaction of a job well done...  Ah....  *Thud* *Ouch, I banged my head on the desk* and shit...I knocked over the cold cup of HAZLENUT coffee left over from morning.&nbsp; Yup, all over the freshly printed report...  &nbsp;  shit  Maybe I\'ll just put my head down for just a minute....   '),(25735,3496136,'26,July,2004','2004-07-26 00:00:00','    *lifting head from desk*  \"MUST finish annual report  Holy crap, I\'m turning into a lizard...too tired to apply lotion...arms like lead...  Maybe I\'ll just put my head for just another minute....\"  *putting head back down*   '),(25736,3496136,'26,July,2004','2004-07-26 00:00:00','    sinuses aching...can\'t swallow...  must finish annual report today...  body feels made of lead...  must trudge onward...  cough, ache, shudder.&nbsp; Maybe I\'ll put my head down for just a minute....   '),(25737,3496136,'26,July,2004','2004-07-26 00:00:00','    that having chaotically neutral/good people mixed with lawfully good people turns the whole group dynamic into slush.&nbsp; And...that dynamic makes men incredibly high maintenance.&nbsp; I will say only this....Shrew-baby and I were supposed to be the high maintenance ones!&nbsp; Dammmit!  Drama queens...honestly...*eyeroll*.&nbsp; Well, at least we\'ll have something constructive to discuss online tomorrow night.&nbsp;   I\'m secretly laughing inside...this too shall pass.&nbsp;   &nbsp;   '),(25738,3496136,'25,July,2004','2004-07-25 00:00:00','    Woke up to something shining in my eyes...it was the reflection from the shiny clean faucets!&nbsp; Woo-hoo!&nbsp; A bit sore from all the reorganization I did yesterday but how nice to move about in a clean almost clutter free house...until I walked into the girls rooms...le sigh!   But I have a plan...MUWAHAHAHAHAHAHA.&nbsp; Think the girls would be injured if they fell out of an elevated bed?&nbsp; I can always add a rail.&nbsp; Hmmm...  Have *Love Stinks* running through my head - ASSFBB sang it on Saturday.&nbsp; She ended the song with \"Love Stinks...and STUFF!\"&nbsp; Am still laughing at her antics.&nbsp; Ah...ASSFBBB!&nbsp; Hearts and flowers.   Starting the Chronicles of Narnia tonight!&nbsp; It\'s been so long since I\'ve read it it\'ll be like the first time.&nbsp;   My new corner in my little room is wonderous.&nbsp; The perfect spot for writing and reading and enjoying a quiet moment.&nbsp; Gonna have to build up though...writer\'s cramp.&nbsp; Yuck!  Well, off to a productive day.&nbsp; The sun is streaming in through my windows, I look around my office at pictures of family and friends, listening to Seal \"Love Divine\" and in this moment...my heart is soaring...I am happy.&nbsp;   '),(25739,3496136,'25,July,2004','2004-07-25 00:00:00','    woke up at 11 after a fabulous time last night.&nbsp; Yup, karaoke with ASSFBB, Shrew-Baby, Guy with the Cart and others.&nbsp; Took lots of pics, wish I were smart enough to figure out how to post them here...sigh...maybe someday.  Tried a variety of alcohol last night...just sips for tasting purposes...orange vodka - rather like biting into an orange rind...YUCK!&nbsp; Raspberry Rum with Apple Cider chaser - a little taste of Raspberry followed by the chaser and then pop!&nbsp; Serious raspberry explosion in my mouth - kind of like pop rocks.&nbsp; Yum!&nbsp; And then homemade mead.&nbsp; Wonderous....after a sip wait 8&nbsp; seconds and the taste of honey coats your throat.&nbsp; Just loved it!&nbsp;   Met lots of new people last night and had a real ball.   Today is all about the cleaning.&nbsp; Breaktime is over...on the upside...now I have a little corner in my room that\'s all for me.&nbsp; I\'ve decided to try writing the old fashioned way - on paper.&nbsp; Will be quieter there.&nbsp; I\'ve put up some of my favorite pics and knicknacks.&nbsp; Gods, I\'m glad to be rid of all that baby shit!&nbsp; Boxes and boxes and car seats and diaper genie and ah....clutter free.&nbsp; Lovely.  Back to nesting....   '),(25740,3496136,'24,July,2004','2004-07-24 00:00:00','    Yesterday was one of those days where nothing went right...ever have one of those days?&nbsp; Then I went to lunch with the wife of *the guy with the cart* [chicks dig the cart ;)].&nbsp; They are building a house and we were talking about how disappointed he was with the size of the foyer of the model home they are considering.&nbsp; I said, he\'s overcompensating.&nbsp; Which made her snort diet coke out of her nose.&nbsp; *The guy with the cart* is not a large man, but he\'s awfully cute and considerate and...grounded.&nbsp; Yes, grounded and calm.&nbsp; A good influence for me.&nbsp; Anyway, she brought a couple of friends from the office who I think I might have terrified with my bad humerous behavior, but they were laughing too.&nbsp; Then she started talking about commedia and who has energy and what order of them she should introduce to her co-workers.&nbsp; I suggested she start with *the guy withthe cart* as he has less energy than her, move on to Shrew-baby, then Lawfully-Good Man, then me and then....   then she suggested Griffin.&nbsp; GRIFFIN!&nbsp; Griffin has more energy than me?&nbsp; *sniff*&nbsp; Maybe all that diet coke snorting affected her brain.&nbsp;   Then ASSFBB, of course.&nbsp; I would think the two of us would be tied.&nbsp;   Anyway, it was fun and I like her very much.&nbsp; I enjoy hanging out will all the martyrs.&nbsp; And it sort of turned my day around.&nbsp; Chatted with Shrew-baby and got a nice email from a friend.&nbsp; Right before the faculty meeting from hell...sigh.&nbsp; Doctors can be such children.&nbsp;   Audition information was wrong, it\'s not until the 14th.&nbsp; So I have time to work on the monologue.   Whip Girl and her hubby and 6 year old came over and took all the baby stuff.&nbsp; Going through the baby stuff brought back a&nbsp;lot of great memories, which I won\'t put down here but I will journal them.&nbsp;&nbsp;My girls...love, love, love them.&nbsp; So the stuff was cleared and I have a sitting room again!&nbsp; and a couple of empty drawers!&nbsp; Sigh...now I just have to move some furniture around.&nbsp; But it will be comfy.   Megan sent me her birthday pics from the party...heavens that was a fun party.&nbsp; It was a come as your favorite fantasy character.&nbsp; I went dressed as a catholic school girl.&nbsp; Madam went as Betty Page - and she looked just like her in all that black lingerie.&nbsp; And Shrew-baby was in red satin and black lace.&nbsp;I don\'t quite fit in though....I don\'t make mead, I\'m not a pagan, I don\'t play video games, and I don\'t have quite enough outer angst.&nbsp; But I fit in enough to have a marvelous time.   Tonight should be fabulous.&nbsp; Meeting ASSFBB, Megan [must come up with a name for her here...] Shrew-baby and her victim, for karaoke.&nbsp; Then over to Whip Girl\'s for a party.&nbsp;   Haven\'t felt well all day.&nbsp; Don\'t know exactly what it is...sort of sick to my stomach and blue.&nbsp; Going to go take a nap.   '),(25741,3496136,'23,July,2004','2004-07-23 00:00:00','    to put it simply and succinctly....  &nbsp; suck.&nbsp;   '),(25742,3496136,'22,July,2004','2004-07-22 00:00:00','    Writing  The process is harder than I thought.&nbsp; The time I have to write are not times of the day when I am at my most creative...in fact the times I have are when I am tired and worn.&nbsp; I hate rising early but that may be the only way.&nbsp; Sacrifices, the sacrifices we make to create and play.&nbsp; Boggles the mind, really.    A Couple of Thoughts on Religion  I was conversing&nbsp;with  urlLink Butterfly Bill  [now there\'s and interesting person]&nbsp;and he answered my question about what he gets from going to church with&nbsp;\"I want to be part of some kind of spiritual community. It helps to keep me from going over the edge and becoming a total hermit and wingnut.\"&nbsp; I don\'t really believe in God as the archetypical kind old white man with a beard sitting on a throne.&nbsp; I don\'t like certain aspects of organized religion and I don\'t like being told what to believe.&nbsp; I do miss the sense of community though.&nbsp; I do miss singing in a church choir.&nbsp; I just don\'t think I could sit through sermons, too much like listening to the President\'s State of the Union Address.&nbsp; I want to be inspired by a sense of a higher being.&nbsp; I already know I\'m not perfect and in the eyes of the culture I grew up in I\'m going to hell [Do Not Pass Go, Do Not Collect $200.00 - and probably no Get Out of Jail Card For You, Young Lady] unless I repent and change my ways.&nbsp; But still there are many aspects of life I would like to celebrate.&nbsp; Mostly nature and random acts of kindness.&nbsp; I wonder if I\'m missing a spiritual connection in my life though.&nbsp; I really wonder...   '),(25743,3496136,'21,July,2004','2004-07-21 00:00:00','    urlLink Jerry Goldsmith  died.&nbsp; So did  urlLink Richard Bloch .&nbsp;   On the upside I\'m still here.&nbsp;  Beware she\'s back.&nbsp;&nbsp;Ancient Sumarian-myth Spouting, Flailing, Blind Beaaaatch [ASSFBB] is back on the internet.&nbsp; Should liven up Tuesday nights.&nbsp;   Got an e-card from Kurt.&nbsp; He wants me to come audition for the Chestnut Players in Olathe.&nbsp; If I had known that all of the friends I would make over the next 3 years would be south, I would have moved South.&nbsp;&nbsp; They aren\'t really doing any shows I\'m really dying to do but auditioning is always good practice.&nbsp; Right?&nbsp; Hmmmm...haven\'t done a monologue in....10 years?&nbsp; 15 years?&nbsp; Do I even have one I can find and do in 3 days?&nbsp; and 32 bars of songs.&nbsp; Which one though?&nbsp; They are doing a 40\'s cabaret style show so....how about something Gershwin?&nbsp; I know some Gershwin.&nbsp; I really want to do \"Tell Me On A Sunday\" though.&nbsp; It\'s a song that shows off my acting range and vocal range.&nbsp;   Madison and I were watching some piece of anime crap before dinner last night.&nbsp; And right after that \"Fellowship of the Rings\" came on.&nbsp; She\'s seen the very first part about 3 or 4 times and just can\'t handle it - too scary.&nbsp; Usually at the point where Gandalf says \"You think I\'m a simple conjurer\"&nbsp; but last night she seemed fine.&nbsp; So we watched.&nbsp; She asked me a million questions (Where\'s ASSFBB when I need her) which I answered as best as I could.&nbsp; We got to the part where Frodo gets stabbed and Aragorn says \"He needs Elvish medicine\".&nbsp; Madison asked me \"Mama, what does the King have to do with this?\"&nbsp;  \"King?\" \"You know, the King of Rock \'n Roll?&nbsp; Elvis.\" Laugh, laugh, laugh.&nbsp; She\'s still thinking.&nbsp; The movie goes on....\"we\'re taking him to Rivendell\" A tug on my sleeve.  \"Mama?&nbsp; Is Graceland in Rivendell?\"  Images of Elrond in an Elvis suit, eating PB and banana sandwiches, crooning about how he\'s all shook up.&nbsp; LOL....  Talked to Shrew-baby for quite a while last night, a deep conversation tempered with laughter.&nbsp; Just what we both needed.&nbsp; We\'re considering becoming hetero-life partners.&nbsp; ;)&nbsp;   &nbsp;  &nbsp;  &nbsp;  &nbsp;   '),(25744,3496136,'21,July,2004','2004-07-21 00:00:00','    Why when you can earn so much more of the almighty dollar?&nbsp; Why continue to fill out the endless flow of paperwork about residents and medical students that no one reads?&nbsp; Why go on having your salary cut repeatedly due to budgetary restraints?  Academic medicine is losing some of it\'s finest teachers.&nbsp; Healthcare is in a tailspin.&nbsp; If you get sick - be sure your doc has gray hair.  One of our finest teachers resigned today unexpectedly.&nbsp; He taught residents to operate here for 20 years.&nbsp;&nbsp;He also taught them to treat the patients and families with respect.&nbsp; He\'s a cancer fighter - and he manages to let his patients die with their dignity intact.&nbsp; How he\'s managed not to burn out in that field, in this place is beyond me.&nbsp; Maybe it\'s the respect for others&nbsp;his father instilled in him with his strict Southern Louisiana upbringing, maybe it was the discipline learned during a lifetime in the military service, or maybe it was the rapport he shared with students, patients and colleagues.&nbsp; Whatever it was...is now a shadow.  I&nbsp;hope that he will continue to teach residents.&nbsp; He has a gift for making the complex appear simple.  He\'s a very tall man 6\'3\" at least.&nbsp; Quiet and softspoken with a dry acerbic wit and he has playful side.&nbsp; He was on his way to the gym and his secretary and I teased him about his legs.&nbsp; He came back after working out and put on scrubs.&nbsp; We teased him again and he dropped his scrubs.&nbsp; I just happened to have a camera in my hands and have a picture of him with his scrubs around his knees with his shorts on - still looking as if someone had just finished ironing them.&nbsp;&nbsp;  I talked to Newt Gingrich one day while answering his phone.&nbsp; And have teased him mercilessly ever since.   He\'s one of the good guys.&nbsp; When you work with a person for 14 years, day in and day out...its hard when they leave.   '),(25745,3496136,'20,July,2004','2004-07-20 00:00:00','    Friday -had a productive day at work, Chris J bailed on lunch, which was just as well because I had lunch instead with 3 of my residents and Dr. Wiley at Fric N Frac - home of a great Philly cheesesteak sand, and great steaks on Tuesdays and Thursdays. They have tables outside which is great when the weather is superb.  &nbsp;  Friday night at the show  Mark (who plays my brother) and I handed out the pins we had made. They were small pins that said \"Chin Up\", which is something we say rather chipper repeatedly throughout the show. I handed them out first to the stage crew when I got there early. Went out front to save seats and the next thing I know the other actors are asking me for the button cause all the cool kids have them. :) So everyone in the cast asked for the button and wore it with pride all weekend. It was a great idea and I liked it that everyone enjoyed it so much. Did I mention yet that I lost my wallet at the front gate? Luckily, the ushers found it and returned it to me. Good thing I have people looking out for me.  &nbsp; Greg came with the girls, my father-in-law came in from Albuquerque, my mother-in-law came in from Topeka with friends, and most of Commedia. The light board got rained on so they lost all the light cues which were programmed into the computer. Why should this affect the lighting, I don\'t know. All the equipment still worked - what they can\'t turn on a switch at the right time? You would think people could still run lights manually? WTF? The show went pretty well. I screwed up the tap number and was late for my last entrance. Just was unfocused. After the show came out to see everyone. Danielle and Madison ran up and gave me big hugs. Found out Danielle was dancing on the sidewalk and Madison was singing during \"I Got Rhythm\". Awww. My friends in commedia brought me flowers...awww. Hugs and accolades. :)  &nbsp; Greg and I went to the cast party. It was near 92nd and Holmes in a fabulous arts and crafts house. They had obviously treat the home with love and care. We sat around talking and laughing until 2 AM! We planned to steal a construction cone and turn it into a megaphone as a director\'s gift for Brian. Luckily, no theft occurred as Taylor\'s dad had one in his garage. Almost fell asleep on the way home.  &nbsp; Have become inspired to do a fundraiser for the Alcott Center. Want to do \"I Do, I Do\" in April at Alcott. Mark (played my brother)is willing (excited and enthusiastic even) to play my husband. He has a such a beautiful tenor voice and the right attitude for the character. We had nice onstage chemistry and had so much fun working together. I thought the show would be perfect because it\'s a two person cast, takes place in one place and can be done with only a piano. I think it would work well in the Alcott stage space. Will present a proposal to the Alcott board and check on royalities. Have a couple of piano players in mind. I\'ve always wanted to the show. I could direct it. Have a couple of choreographers in mind...not that much is needed. I hope it works out.  &nbsp;  Saturday  - slept till 10 went to the Alcott Center for the pirate improv workshop. We had 16 participants. Started with evaluations and then played games until 2. I love observing, digesting and then trying to vocalize what I see in a constructive manner. I find it to be so helpful when I get constructive criticism. The hardest part was not jumping in to play. It\'s been gosh, almost a month since I did any improv. Went to lunch at Arthur Bryants with the crew. Had the pleasure of sitting with Don and Chris . I respect the work they are doing with Red Griffin and enjoy watching them perform at fair. Am so glad I am getting the chance to get to know them better. Lovely people. A new color to the palate of friends. Went home and painted the cone. The paint I used said it would dry in 72 hours. Uh-oh. Stopped at Target to get markers and found just the best card for Mark, who plays my brother. Stopped at Alcott on my way to the theatre. Saw Adrian and Leslie and Faye and Chris and some other renfest friends. Went upstairs and played for a bit.  &nbsp;  Saturday\'s show  was wonderful. No lighting issues, no sound issues. Managed to make all my entrances and did the tap steps right. Yaaay! The party was in Olathe near Black Bob and 146th. One of the cast brought his karaoke equipment so we had karaoke on the first floor and musical theatre jamming in the basement - with a partial orchestra ;) Scott the violin player can play be ear. He and Greg and I had a great time talking. I invited he and Jenny who plays piano excellently to my house sometime to jam - a folk jam! So I sang \"Love Shack\" with David who did the best imitation ever of the B52s...I really forgot that I always dance to this song and have never really listened to the verses...oh well, we certainly had fun. I made up for it later with \"I will Survive\". Brian, the director started dancing like a crazy man and the next thing I knew I had back up girls dancing behind me on the fireplace hearth. I so rocked the kazbah! Then I went downstairs and sang \"Tell Me On A Sunday\" from Song and Dance, cause it was demanded. ;) \"Take the hurt out of all the pain...take me to a park that\'s covered with trees...tell me on a Sunday please\" We were out until 4AM. Greg played foosball and pool and laughed and made friends. I think he might have a better time at these cast parties than me ;) Okay, probably not. It\'s a tie :)  &nbsp;  Sunday  Got up at 10:00 and went to teach improv. They started at 10. Damn. Had fun anyway :) I learned a lot about pirate vocab and ships and pirates in general. Did evaluations. The group as a whole did so much better the second time around. I think every person there got something out of it. We stayed late to play shopkeeper. I took my group (woo-hoo! I got my own group!) with Rissy (woo-hoo! I got to play with Rissy!) into another room. Secretly pleased that Don and Chris and Jeremy [hearts and flowers! Heavens, I\'m shameless for Seamus ;)] came with me. They played shopkeeper a couple of times and then I taught them slide show (badly) and explained elevator, though we didn\'t really play it. I really enjoy running the games. But man, it made me hungry to play. I had planned not to go to any rehearsals until after fair, but I doubt I make it. I miss playing with Rissy and everyone. So went to lunch with everyone at Fric N Frac. I was in the middle of man sandwich - between Jeremy and Don. Woo-hoo! Great conversation! Learned a lot about everyone. Chris\' mom - I think I love her. She\'s bold and witty and kind. Just a joy. Jeremy bought lunch, what a sweetie. After lunch I showed Jeremy the way to the QTrip and then to I-35 south. I headed North to home.  &nbsp; Got my things together, hung out with the girls for a bit and then headed to the show. The last show was great. We all got a bit teary at circle. Before every show Brian gathers his cast together and we take care of any business or general announcements and then he says something inspirational. I really like this tradition. He\'s a really sweet man. I gave Mark a card that had a peacock on it. Inside it said, \"I couldn\'t be prouder of you\". I wrote in \"This is the only card I could find with a c$#% on it.\" LOL...he died laughing. We really bonded by being just frankly so nasty...we made fun of the song \"Slap that Bass\" as \"Slap that Ass\"....my, oh my, we had fun. I took lots of backstage pics before the show and during intermission. During the last scene Mark and I are sitting having *martinis*.&nbsp; I had this business where I spill a little on my chin - imagine that but that night I spilled the whole thing down the front of me.&nbsp; Jenny was so cracked up she had to leave the stage early.&nbsp; I was using my cardigan to wipe off my face.&nbsp; And then we struck... &nbsp; and we struck... &nbsp; and we struck until 2:30 in the morning. I got to work with Kaitlyn and Tara in the shop pulling staples and screws out of wood. Then we got to stack huge pieces of Luon - I carried them over my head \"Val Smash\"...Managed to have a good time. Hung out with Loren and Curt a little. I adore the cowboys who are light in the boots. I didn\'t stay for the pizza. I sort of sneaked out cause I didn\'t want to say a bunch of goodbyes. I\'ll see the people I connected with again. I am so grateful for this experience. And I was afraid I\'d fall asleep.  &nbsp; Which I almost did anyway...It\'s a long drive to almost Gladstone.  &nbsp;  Monday  I slept until 10. Greg was off and so was I. We kept the girls home and had a great day all together. We watched Greg play golf - 9 holes and then went home while he played 9 more. He loves it so. I love being outside and walking and thinking. Reflected on the weekend and what it is I want. I achieved my goal of being in a musical so what\'s next? I enjoy teaching and playing improv, I enjoy spending time with my children. I suppose the next big challenge I face is my hearing. That will be solved in the next couple of weeks, hopefully. We went swimming. We ate a snack. I took a small nap. Then our friend Bill and his girls (15 and 11) came over and we all went to the Royals game. We tailgated and watched the game. The girls really enjoyed it. Got home around 11 and went immediately to bed. It was a fabulous day.  &nbsp;  Today  I slept till 10.&nbsp; I\'ll start normalcy tomorrow. Today, I\'m going to do some domestic stuff and spend time with the girls. And tonight I\'m going to write. I don\'t think one day creates balance. But it\'s a nice start.   '),(25746,3496136,'19,July,2004','2004-07-19 00:00:00','    Greg:&nbsp; Hey girls, look at those cows on the other side of the fence... &nbsp; Girls:&nbsp; Moooo, Mooo, Mooo, Mooo...(didn\'t seem like this chorus of moos was ever going to stop so...) &nbsp; Greg:&nbsp; Girls, don\'t do that please.&nbsp; It\'s scary. &nbsp; Madison:&nbsp; Who\'s it going to scare Daddy?  &nbsp; Greg:&nbsp; The golfers.  &nbsp; [a beat] &nbsp; Madison:&nbsp; Are golfers stupid, Daddy?  &nbsp; tee-hee-hee.&nbsp;   '),(25747,3496136,'17,July,2004','2004-07-17 00:00:00','    I played today. Haven\'t played in a long time.&nbsp; I am rusty...but Chris was particularly hard on me as an example in the Who, Who, Where, What game.&nbsp; Which is a difficult exercise really.&nbsp; Played with some of the Red Griffin crew...three in particular are quite talented.&nbsp; I disagreed with Dean on every point only to bring out more discussion of acting in general.&nbsp;  &nbsp; Dean, Erin and I served as evaluators and we had forms to complete on everyone.&nbsp; They coupled up and did two minute scens.&nbsp; The first thing Dean did was set out his the Dory the Bluefish and he said \"We can\'t have proper feedback without Doug.\"&nbsp; Evil, maniacal grin on his hairy face.&nbsp;  &nbsp; Memories of last summer as I spent the day in Alcott, walking up and down the stairs, in the theatre space, the green room and Chris\' office.&nbsp; But today, I am smiling at the memories....somehow they don\'t make me sad.&nbsp; They are comforting somehow. &nbsp; more later...gotta get ready for the show.   '),(25748,3496136,'16,July,2004','2004-07-16 00:00:00','    was shoved under my door this morning when I opened my office door.&nbsp; Karen, favorite resident had asked me if I had ever read this children\'s&nbsp; urlLink book .&nbsp; She said she would find it an bring it in for me cause my girls would really like it.&nbsp;  &nbsp; I read it and it made me cry...tears of joy as comprehension dawned on me...love for another person is as individual as your children.&nbsp; It\'s indefinable and warms you to the tips of your toes.&nbsp; It\'s such a simple thing really.&nbsp; She\'ll never know how much this allegory has helped me.&nbsp;  &nbsp; Her inscription BTW:&nbsp;  &nbsp; To Val,  Whom I love the orangest, reflecting her fierce energy and passion.&nbsp; May friendship burn forever.&nbsp;  &nbsp; Karen.  &nbsp; Wow...I am truly blessed.  &nbsp; &nbsp; &nbsp; &nbsp;   '),(25749,3496136,'15,July,2004','2004-07-15 00:00:00','    Start from sleep from the tornado sirens. Turn on the TV and watch the weather...holy shit! Grab the children and the husband and the cat and the dog and go to the basement. Images of the tornado that wreaked havoc on our neighborhood last year coming fast in furious unbidden in my head. A sigh of relief as the storm heads north.  &nbsp;  All four of us sleeping uncomfortably in the king bed. Woke to two beautiful girls sleeping. I am very tired but waking to that image will sustain me through the day. I will admit I took delight in the evilness of waking up their little butts and getting them ready for school. heheheh  &nbsp;  Last night\'s show went very well. The cast party was at Kaitlyn\'s house. We were both supposed to wear our Tinkerbell shirts yesterday and I forgot. Oh man, she looked so disappointed. She forgave me though and we\'ll be Tinker-twins on Saturday!  &nbsp;  Talked to Brian (the director) about teaching an improv class at his dance studio and he seemed pleased. The schedule is going to be published on Monday though so&nbsp;Beardless Wonder Acting Coach&nbsp;and I will have to decide quickly who, where and when pretty quickly. I am pretty excited about the chance to work with young actors. May have to put it off until next year, but...we shall see.&nbsp; I\'m actually more jazzed for this opportunity for Beardless Wonder Acting Coach&nbsp;than for me.&nbsp; He\'s such a good acting teacher.  &nbsp;  Tons to do...all of my family is coming to the show tonight. Man, it\'s going to be so very fun. :)   '),(25750,3496136,'15,July,2004','2004-07-15 00:00:00','    Okay, here\'s the questions... &nbsp; The Questions: 1. What are the three most important non-living things in your life? These can be work, ideals, material objects, dreams, etc. 2. You are at my favorite bar, Charlie-O\'s. The place is smoky, with low ceilings, loud people and filthy surfaces. Beer is overpriced, the music is too loud, and the place is packed asshole to elbow with the freakiest conglomerationof Montpelierite Vermonters you\'ve ever seen. What is your immediate visceral reaction? 3. What is your favorite season and why? 4. What do you wear to bed? 5. What question would you like me to ask? Be advised that I will do so.   My Answers  1.&nbsp; Three most important non-living things in my life are my house (have already poured a lot of time, energy, dollars and dreams into the good old ivory tower in the middle of suburbia.), my performance goals (notice I said goals instead of dreams - go me!) and a silver charm of an elf my aunt gave me when I was 10 (it\'s a connection to a very special and unique woman and a connection to my childhood). Hmmm two of the 3 are material....well, we are *living in a material world...and I am a material girl...*&nbsp; Oh looky, I just turned into Madonna!!! 2. Charlie-O\'s, huh?&nbsp; I don\'t know what&nbsp;\"Montpelierite Vermonters\" are?&nbsp; Are they eastern rednecks?&nbsp; Hmm...Order a beer and conspire to start a wet T-shirt contest. ;)&nbsp; No really, I\'d order a beer and have as much fun a possible - hopefully involving dancing.&nbsp; Me, I\'ve never really met a stranger, I love beer and the whole bar scene. 3.&nbsp; My favorite season is autumn - I love the sound of the crisp leaves under my feet, the smell of applebutter and football.&nbsp; But mostly, I love the autumn leaves and the remarkable colors.&nbsp;  4.&nbsp; Varies...sometimes nothing, sometimes a comfy big cotton T-Shirt, sometimes a white silky nightgown and sometimes flannel jamas.&nbsp; Usually, comfy big cotton T-Shirts - I always wear socks.&nbsp; Always.&nbsp; I can\'t really stand for my feet to be cold. 5.&nbsp; Is the glass half full or half empty?&nbsp; Today the glass is half full...but I\'m living on a day-to-day basis.&nbsp; Check back tomorrow for glass status.   '),(25751,3496136,'14,July,2004','2004-07-14 00:00:00','    particularly spicy ones such as those served at the  urlLink Bulldog  will give you gas.  I thought I was having a heart attack.  Heavens they were good though.   Had lunch there today with Madame (the illustrious Red one) and Whip Girl (she of closeted Halloween fame).  Madame cut off her hair and dyed it brown.  Whip Girl is 7 months pregnant.  We talked about upcoming fest, how BlondeLittleJohn (the one I climbed during chessmatch) created fireworks and then blew them up on the 4th (no rennies, geeks or freaks were injured during same), upcoming baby stuff and of course, sex.  A frank discussion with those two made me feel...well, *blush*...that should be enough said, I am blushing!  Me!  The toys described...NOTE To Self:  Find a Priscillas on the way home.    Then we planned a baby shower and talked about a pagan festival and the joys of creating a commune.  We talked about our puppies, children and jobs.  They are lovely women inside and out.    I was thinking about different relationships are between men and women.  The conversation that a group of three women would have versus the conversation three men would have, I mean as an example.  Really, why are women so much more based in emotion?  It seems to drive us.  Or at least the women I know.  And women just bare it all to each other...completely.  Well, mostly.  Men, I guess I don\'t really know.  I guess that\'s an age old question though, one I\'ll never have the answer to...why are men and women so very different emotionally?   '),(25752,3496136,'14,July,2004','2004-07-14 00:00:00','    Horoscope Today... You already know that you don\'t have to express every emotion that you feel.  [I don\'t?  wow.]   But this doesn\'t mean that you shouldn\'t feel them.  [I only recently figured out how to express them.  How is it that you feel a strong emotion without expressing it?  How do you not react to a punch in the stomach?]   If you bury emotions, you will end up being controlled by your fears instead of by love.  [what if I don\'t want to be controlled at all?]  Even in the workplace, you\'ll be pulled between the safety of rational thoughts and the vulnerability of feelings.  [rational thoughts?  in the workplace?  You haven\'t lived until you\'ve worked with 7 other women who all take Premarin (it\'s hormone for menopause)]    Logic may create security, but it won\'t cure loneliness.  [So Vulcans are lonely?  Poor Spock]    Last night we had dinner at Applebees.  The girls enjoyed their mac n\' cheese.  I had a fabulous steak with bleu cheese and great big beer (They come in pints?).  Greg\'s 92 year old grandmother had us worried but her results came back and she\'ll be fine.  Whew, sigh of relief.  She\'s headed off the Washington DC today.  I so admire her spunk.  We call her Grape.  Her husband died 18 months ago and she has been spending almost all her time traveling to see her family - Albuquerque, Washington DC, Salt Lake City, Switzerland (yes I said Switzerland) Alabama and Colorado.  It\'s like she\'s trying to get in all the travel she can.  Well, hell good for her.  She was sort of stuck taking care of her husband for the last 15 years so she SHOULD enjoy herself as much as she can.  The really amazing thing about her is that she takes no meds on a daily basis - at the age of 92...now how rare is that?  She taught first grade for 45 years, raised 3 boys who grew up to be fabulous men and PhDs all of them, all while her first husband was bedridden with Parkinsons.  She cooks like you wouldn\'t believe and she doesn\'t judge.  She will certainly tell you what she thinks if you ask her but she\'s open to more than one solution to a problem.  She is a shining example of how to live a happy life.  I have asked her often for advice and never managed to follow it.  All of her grandchildren would rather die than disappoint her.  And the funny thing is, the thing they don\'t know is...they couldn\'t disappoint her.  Because she loves them for who they are not what they accomplish.  To be so vibrant at 92...  Had lunch with Shrew-baby yesterday.  The show is coming together nicely, I can\'t wait to see it.  I listened.  Sometimes, that\'s all you really need is someone to listen to you, really listen.  Then she listened to me.  We\'re sort of stuck in the same place.  She has the most amazing amber eyes and a wit that is so sharp.  On the inside though, she\'s as soft as a goose down pillow.  I find her company so very, very, very comforting.  This better living through chemistry thing isn\'t working out so well...maybe I need better chemistry....almost didn\'t get out of bed this morning. Tossed and turned all night.  Sleep is overrated.  My mom isn\'t coming to the show.  I convinced her not to...no point in driving 8 hours for a two hour show on your one day off in 14, is there?  I\'ll send her a tape.  I think I am supposed to get the CD ROM with the pictures from the show today.   My brother has taken off with his on/off again girlfriend.  She\'s a bit of hippy.  Well, she\'s a hippy.  No doubt about it.  They are leaving for a 3 week *vacation*.  Heading to Mount Rushmore and other sites yet unseen.  She says she\'s going to dry him out and then see what happens.  Having my brother around, drinking was just killing my parents, literally killing them.  The stress...my dad\'s BP through the roof.  My mom...losing her hair.  And our family has great hair.  So Mom told him when/if he comes back and he drinks she\'s throwing him out.  Now she\'s beating herself up over saying that to him.  I am trying to reassure her she said the right thing and I hope that if they get to that situation that she\'ll follow up on her words.  It\'s a tough situation.  I am amused that the name  urlLink Griffin  is so intricately involved in such a courageous act of rebellion.  If my friend Griffin were actually there he probably would have dumped the tea just so people would drink more coffee ;)  So show tonight...hope it won\'t rain.  Here I go...trudging onward...   '),(25753,3496136,'14,July,2004','2004-07-14 00:00:00','    There\'s a storm coming in from the north and it looked like an Ansel Adams photograph.  A black and white piece of gray masterpiece.  Phenomenal.   Madison is very blue still.  I think part of it is tiredness.  I feel compelled to make her feel better but you know what?  I won\'t be able to.  Each person, now matter how large or small has to own their own feelings.  Maybe she likes feeling blue.  Maybe she\'s trying to get attention.  I shall just let her feel what she wants to.  The only person I can control is myself and I am not going *impose*, *project* or *expect* anyone else to feel the way I want them to.   Heavens, I think I might be growing up!  *looks around and begins to chase own tail*  Maybe not!  ;)   I was thinking about Jeff Smith this morning, you know the Frugal Gourmet.  He died last week.  I was looking at my row of cookbooks and I have three of his cookbooks.  I watched his PBS cooking show religiously and tried many of his receipes.  He\'s the one who inspired me to become a good cook.  I wish I didn\'t feel ashamed for admiring him.  Because no matter what he may have done or may not have done with children (which is totally wrong, don\'t misunderstand me - that\'s just sick and so wrong), he taught me stuff and the good part of him will live on in my memory.  A love for good food, a feeling that you can show your love for others by doing nice stuff for them - cooking their favorite dish.  It\'s the little things...hey isn\'t that a Sondheim song?  Yup.  From Company.  Everything is a musical with me.  ;)  Gosh I got so off track, so one other thing I wanted to say about Jeff Smith is that during his three great cuisine (Italian, Chinese and I forget the other one, nust mot have been that great) I made the most fabulous dinner ever for a superbowl party.  We had Hot and Sour Soup, crab rangoons, fried rice with shrimp, chicken and pork, Stir fried Lamb and Vegetables in Lettuce rollups and Peking Duck.  Yes, I had a 10lb duck carcus hanging in my kitchen basting for 2 days.  It was just the best thing I\'ve ever put in my mouth!  I\'m getting tired just thinking about cooking all that stuff now but looking back, it was wonderful.  I didn\'t see any of the game and the clean up was hell.  But, I love cooking for friends.  I should do it more often.   '),(25754,3496136,'13,July,2004','2004-07-13 00:00:00','    The conversation in the back of the car on the way home.   Madison \"Danielle, I got some new toys.  They are in my suitcase but you can\'t play with them, cause they have a warning and they are very breakable and no one who is 4 or 5 or 3 or 6 can play with them.  They are action figures.\" Danielle \"Well, sissy, that\'s okay cause I got your other action figures in my room while you were gone.  I\'m protecting them.  But it\'s okay cause there\'s no warning on them.\" It\'s not often my oldest is speechless.  Repeated exclamations of how much she misses her Bomm and tears.  Finally Danielle says \"Madison, will you quit crying about Bomm, you\'re killing me.\"   Ah...the joys of home.  Glad she is safe and sound and asleep upstairs.  Very sweet that she misses her Bomm.  Want to figure out a way to post pictures here.  NOTE TO SELF:  research hosting and crap like that.  Have decided to use code phrases for friends.  If you know me and can\'t figure it out...then you must be stupid ;)  but you *could* email me and I\'ll tell you.  Am IMing right now with Shrew-baby, Lawful Good Man and The Guy With The Cart.   '),(25755,3496136,'12,July,2004','2004-07-12 00:00:00','    Hope floats  One of our former faculty has had a recurrence of breast cancer and it doesn\'t look good. She was cultured and soft spoken. Earlier in her life she was classically trained on the piano and even toured with one of the East Coast Symphonies. Her father was the first African American physician to practice in Kansas City. She spent a lot of her time doing charity work for the community. The best thing about her was the way she would bend rules to just the breaking point - she had an inate talent for knowing when to stop. Black eyes that sparkled and a smile that could light up a room. She told me once that I need to relax more. The next day she brought me a pair of slippers with stars on them - big furry powder puff blue ones with one large yellow star on each toe. They were to \"get you started on the right foot. Relaxing will lead you to your stars, Val.\" That was gosh, 4 years ago? She never seemed to be truly focused on teaching or her clinical duties. It was like she was on another plain - and she was. I think she took the faculty position because she didn\'t want to move her family and she knew academics would have less long term responsibilities than private practice. I feel privileged to know her.   I wonder what it must be like to go through chemo? Is she lonely? Does she want company? I know that she is very sick and cannot cook and do things for herself right now. She has plenty of help, I offered. I sent her a note and I\'ve called her a couple of times but I want to do more. I worry that she wants to be left alone or is too ill to be bothered. I am hoping that she will recover but the results from her surgery....hope floats though. Hope floats.    Madison is coming home today!  Madison is coming home today! Madison is coming home today! MADISON IS COMING HOME TODAY! She has been in Albuquerque with her grandparents for a week. They took her all over town and up to their cabin in Southern Colorado. She played in the river and went hiking in the mountains every day. I talked to her every night. Listening to her telling me what she did that day, imagining her facial expression, and laughing at her impressions and astounded at her insight and vocabulary. I can\'t wait to hug her. Last night she suggested that she stay there and we move down there. ;) I watched Danielle and Madison talking on the phone. Danielle pacing and in my imagination I could see Madison doing the same thing. Danielle is not allowed to play in her sister\'s room, by decree of her sister. Of course, this week...well, we\'d better get it cleaned up before she gets home.   Kurt and Jason  Kurt plays Lank Hawkins in the show at Theatre in the Park. Kurt is an art designer. He has curly black hair and is tall at least 5\'9\". He has black eyes and really long eyelashes - I just hate it when men have better eyelashes than me, but don\'t they all? Kurt was a little overwhelmed by me at first in rehearsals. I am a strong flavor. Every time I turned around there he was on the edge listening and laughing. I finally just started hugging him. He opened up a little to me. He and his partner have been together for 15 years. He got to go to Disney and go in the vaults. His joy describing holding an original sketch from Steamboat Willie was contagious. Kurt reminds me a little of Dean Kinsey - quiet, intense and lawfully good. Kurt takes personal responsibility for ensuring that the scene changes go well. He\'s one of those people that really care about the quality of what they do...not the quantity...the quality. I met his partner on Saturday night. He\'s a delightful man who is quieter than Kurt. Nice laugh though and beautiful eyes. There was a palpable tenderness between them. Kurt\'s singing voice is magnificent and not used enough in the show. He\'s a very good actor and is very giving onstage. In short, I adore him.   Jason plays the male lead. Jason is the shyest person I have ever met. It took me eight weeks to figure out if he even liked me. Jason works at Worlds of Fun and does 5 shows a day...then he came to rehearsal. He\'s a senior at KU. His singing voice is like silk. He can dance and when he chooses his smile is dazzling. I told him at the cast party that I still wasn\'t sure if he liked me and he gave me a very big hug and said \"who couldn\'t like you?\" Awwww....then he stayed around to hear me sing. He complained that his makeup made him look like a Mexican. I said \"Andale, Bendejo.\" He is a serious and somber soul, I think. Though he does like to kid around when he\'s finally comfortable with you. I think people perceive him to be stuck up because he is so shy and handsome. I would have never made the effort to get to know him were I younger so I\'m glad I met him when I did. It was worth the effort. He\'s a sweet kid.   '),(25756,3496136,'12,July,2004','2004-07-12 00:00:00','    at my desk whilst trying to get organized.  Never, ever, ever going to get caught up.  sigh...  the brie is good though.  And grapes and carrots.  didn\'t really eat lunch today - have been grazing.   Last night the show was cancelled after the first act because of rain.  This meant the cast party started early.  It was at the Guacamole Grill in Olathe.  We had karaoke.  The best part was they did one karaoke song, one dance, one karaoke, one dance.  I danced my ass off!  HA!  Danced with the teens - some of them had never seen the Madonna foot hop thingy.  Electric slide...mmmm...baby got back!  Wow, it was fun.  I sang Black Velvet and treated the mike stand like a pole - got a huge laugh and a standing ovation!  fun, Fun, FUN!  Still smiling thinking about last night.  Greg went with me and had a really great time.  By the end of the night he was mixing it up too.  It was nice to see him meeting my new friends and liking them.  I was afraid that he wouldn\'t.  What fun we had.  We even danced.  He\'s thinking about volunteering to be on set crew for whatever show I get cast in next summer.  I\'d really like for Madison to be in the same show I am.  A family thing...fun without the guilt.  Now there\'s a concept.   Kaitlyn - the female lead is 17.  She is about 5\'4\", white blonde long hair, blue eyes, a perfect smile.  She has a wonderful voice, sweet and syrupy.  She is innocent and fun.  Her acting ability astounds me, really.  There\'s a point in the show when her character has just broken things off with her man and she\'s singing \"But Not For Me\" and when she does the line \"But I can\'t dismiss the memory of his kiss...\" she very slowly raises her hand to her cheek in just the most perfect way.  The audience believes she\'s caught up in reverie and then her tears fall.  Made me cry the first time I saw her do it.  Well, I cry every time I hear that damn song.  I can only imagine how much life experience is going to make her a better actress.  She\'s trying to decide whether to go to college or just move to NYC and make a go of it.  I told her Carpe dieum baby.  College - you can always go to if you don\'t make it.  College can be a crutch because you\'re afraid to try.  She\'s good enough now.  She is a true triple threat.  Great dance technique, beautiful voice, and a sensitive actress.  The thing that really is best about Kaitlyn is that she is so down to earth.  She\'s just a teenager.  She doesn\'t put on airs and refuses to be treated like she\'s special.  I tease her about becoming a diva in training.  She calls me her tinkertwin - cause we both have Tinkerbell shirts.  Her laugh is like music.  She\'s one of those people that you\'re sure is destined for greatness.  Gosh, it\'s great to be involved in theatre again.  I am so much happier.  Mom-in-law\'s birthday is Saturday (NOTE TO SELF:  Buy something)   '),(25757,3496136,'11,July,2004','2004-07-11 00:00:00','    Really, really, really fabulous. Opening night was a smash hit.  Standing ovation at the beginning of curtain call.  No set change disasters, no sound issues...the lighting computer did totally go beserk in the final two scenes but we were still lit - albeit not artistically unless you like strobe lighting.  But no one was injured and people laughed!   The moment Mark and I came on stage people laughed at us.  We are a pretty funny sight - 6\'5\" Mark and I in khaki and pith helmets.  The audience laughed at all of our lines in the right places even.  Before the show Brian, the director, gave me a note.  I had been ad libbing lines on my first exit - trying to incorporate lines from the other Theatre in the Park shows.  The other casts had used references to the other shows so I thought it was a part of the culture.  So the latest thing I was saying was \"Eugene, now that we\'re not in Kansas anymore (Wizard of Oz) we should climb every mountain (Sound of Music.\"  To which he would answer \"Patricia, do you think there are any hot box dancers in there? (Guys and Dolls)?\"  Brian\'s not was not to adlib from the other shows - it screams community theatre.  Ouch.  He said think of something else.  I said hopefully \"Something different every night?\"  Now this was in front of the entire cast....\"Sure, I trust you\".  He smiled.  I said \"Be afraid, be very afraid\"  Laughter rang out.  The Stage Manager piped up \"Yes, we are.  Val\'s first and last performance at Theatre In The Park\"  Again laughter.  In conversation afterwards with the Stage Manager - we have deaf interpreters and they need to know what is going to be said beforehand to make sure they know how to sign or spell it.  Sigh....So I set it...and it\'s brilliant.  Here\'s the new bit....  We (Eugene and I) arrive in Deadrock to review the Lank Hawkins\' (the bad guy)Hotel, Saloon and Restaurant.  We meet Lank and are motioned into the Saloon.  He has just spoken very bad  French with a Western drawl...we are very British.  Right after his very bad French I say..okay, I don\'t know how to spell it....Oh (over the top delight) parlevous Francais (rushing to him and pinching his cheek - as sickeningly sweet as possible) and then dash through the swinging doors (hoping not to injure myself or others - grateful Rissy isn\'t involved with the swinging doors).  Mark comes up to him in the same tone and says aurevoir and gives him a pat on the cheek.  The idea is born out of trying to annoy him which is a nice setup for what happens later between the characters.  No, not adlibbing and creating character development!  Brian loved it.  So after the show I found Greg and Danielle.  Danielle said \"Mommy, look...Chin up.\"  Which is a line from my featured song and she did the choreography backwards.  I am still smiling at that image.  No flowers for me on opening night.  The first words he said were \"Now what?\"  No hug even.  He hasn\'t been feeling well all week though.  I forgot to remind him to get me flowers but dang...I shouldn\'t have to.  Don\'t all men know that opening night your lady should get flowers?  Expectations lowered...or spoken.  So I told him I\'d really like flowers on closing night but not about my disappointment.  So I addressed the topic, Dr. Phil, but not the issue.  Am working up to it.  Saturday night\'s show was not quite as good but still fabulous.  The media was there and I\'m looking forward to the reviews.  The sound had a lot of problems.  The guy who plays the lead, Jason, was sweating to death - 95+ degrees onstage and his mike kept shorting out because it kept getting wet.  They changed it 4 times.  Missed a lot of sound cues, but luckily it was on for all his songs.  Now about the heat - what are they bitching about, come on...they\'re not in corsets and chemises and long skirts...weenies.  ;)  I sing offstage during most of the production numbers which means my mike is on a lot.  Last night Mark and I sang the first part of a song and it sounded so good I turned to him and said \"Dude, we so rock!\"  He pointed at the mike.  DAMN!  Curse you Richards, Doom wanted to be mute!  At least I didn\'t curse.    Went to a cast party last night.  Great time...we sat around the piano and sang show tunes until 1:30AM.  I got to sing \"Tell Me On a Sunday\" from Song and Dance and \"I Dreamed a Dream\" from Les Mis and then Mark and I did some duets.  They kept asking me to sing...how could I say no...;)   Got a lot done around the house yesterday.  Really enjoying being home.  I bought coffee yesterday.  Bought some Fair Trade stuff from Guatemala...Arabica, maybe?  Anyway, I chose it based on smell.  Thought I\'d try something new.  It\'s better than Folgers even.  Wonder how it would be if I reused the grounds?  hmmmm....naaaaaah!  ;)   '),(25758,3496136,'09,July,2004','2004-07-09 00:00:00','    Why am I calm and serene?  Must be the lack of sleep.  Or the massage.  Yep, probably the massage.  Smartest move I\'ve made in sometime getting a massage this morning.  Ah....  I found my wallet - all the stuff was in there!  Woo-hoo - no going to the DMV for me, thank you very much!  Heavens I\'m so grateful to be doing a show again.  Elation just sitting around backstage.  I\'m having so much fun.    I called to tell Danielle good night last night and she was already asleep [lucky thing!] and Greg told me she said the sweetest thing.  I told her last night was my last rehearsal for the rest of the summer and she told Greg when they got home that she couldn\'t wait to \"give Mommy hugs and kisses whenever I want.\"  Awwwww.....  It\'s so nice to be loved.  I have finished a draft of the first chapter of *my great American novel* *smirk*.  One of my residents is going to read it for accuracy.  Gosh, I hope she likes it.  It takes place in a hospital - it\'s about a group of ob/gyn residents.  I thought I should write about something I know.  Am looking forward to spending time writing after the show closes.  This morning a gentle rain started to fall as I was driving in.  I rolled down my window and took a deep breath.  There is nothing like the fresh smell of rain on a summer day.    Rain?  SHIT!  Hope the show will still open tonight!  Crap, must go check the radar...  *lazzi*   '),(25759,3496136,'08,July,2004','2004-07-08 00:00:00','    Got home from rehearsal last night at 2:30 am.  Got up to go to work today...no coffee.  I reused the grounds from yesterday.  *hanging head in shame*    The 20 showgirls first scene outfits were all too large.  I pinned one of them in and fixed the huge bow across her butt.  The next thing I know, there\'s a line of long legged beautiful girls waiting for me to *adjust* them.  I could only think about how jealous my husband was going to be when I got home.  *snicker*  So these sets are huge and the scene changes aren\'t going well.  One of the actors, Kurt goes out and talks to the director about making one of the scene changes during intermission instead of in a hurry at the top of the \'Entract.  He said no.  Kurt, Mark and I were bitching about it.  Pointing out how difficult it will be to get the logistics right with little or no rehearsal and how the stairs are dangerous until they are clamped.  Then the stage manager\'s voice came looming from above \"Stop.  We\'ll have to start over someone is talking.  Patricia [that\'s my character\'s name]I think it\'s you, dear.\"  Gulp!  There should be a light come on when your mike goes live.  Luckily, I didn\'t say anything I wouldn\'t have said to the director.  I\'ve learned that lesson from a friend and only recently.  We ran the number again.  Then the director announced we would do the scene change at intermission.  Later I found out he didn\'t hear our conversation but the stage manager used all our points in her *discussion* with him.  Safety first. ;)  The lights are going to be good, the costumes are fine, the dance numbers are fabulous and the leads are quite good.  I think the show is going to be wonderful.  It\'s the scene changes I worry about.  Hate to be run over by a big frame covered in Luan...death by veneer, if the coffee doesn\'t get me first.    One more day of bleary-eyed tiredness before I can rest.   '),(25760,3496136,'07,July,2004','2004-07-07 00:00:00','    Tech week at Theatre in the Park consists of arriving at 6 PM to help clean up from a day of set construction.  Most of the sets in this show are larger than most of the apartments I\'ve lived in.  Vocal warm ups, get miked, get in costume and then sit.  I am not complaining, really.  Monday night I sat from 7 PM till 1 AM and never got on stage.  Last night I finally got on around midnight.    I am the helpful sort...I find it difficult to sit still.  So I helped the sound guy adjust levels, costumers accessorize, moved hand props and the director even had me help reblock some stuff.  It\'s nice to feel useful and respected.  Last night when I walked out on that huge stage for the first time and looked out at the lights...well, gosh it felt like home.  This to me is performing.  This is where I belong.  It will be hard to go back to Renfest after this.  Took the girls home to Hannibal this past weekend.  Madison and I had just finished reading Tom Sawyer outloud and it was so much fun to watch her point out things from the book.  We lined the kids up in front of Twain\'s boyhood home by the fence and she insisted everyone pretend to whitewash.  It was a mostly great weekend filled with popsicles, water balloons, fireworks, BBQ and beer (lots of beer).  My brother is backsliding.  Some horrible moments, frankly I don\'t know how my parents stand it.  Their health is beginning to suffer.  I feel some action should be taken...I just don\'t know what. I think at this point all I can do is listen and hope that will be enough.  I have been out of touch mostly with my other family, Commedia.  The JOB show went well from what I understand.  Rehearsals are progressing well for KCRF.  Doubtful I\'ll make it to Bristol.  The hunger is gone.  I need a long break.   '),(25761,3496136,'11,August,2004','2004-08-11 00:00:00','    But first my horoscope... Even if your logic is telling you one thing and your feelings are telling you something else, there is a solution to your dilemma. That famous Libran indecisiveness won\'t buy you any time today. You\'re going to have to make up your mind and stick with your decision. You\'ll be surprised at how others will come through for you now to help you make the changes that you desire in your life.  Oh great...I just love surprises (said dripping with Shrew-baby\'s sarcasm)  And now [drum roll] Babylon 5 Philosophy... \"We create the meaning in our lives. It does not exist independently. Being Anla-shok does not mean worrying about what others will think about us. It does not mean deciding what to do based upon whether or not it serves our sense of ego or destiny. It means living each moment as if it were your last one. It means doing each right thing because it is the right thing. The scale doesn\'t matter. The where, the when, the how, or in what cause .. none of those things matter. In my life I\'ve discovered very few truths. Here is the greatest truth I know: Your death... will have a meaning if it comes while you\'re in fullest pursuit of your heart.\" --Babylon 5, Season 5, \"Learning Curve\"  I have been so untrue to myself for so long.  I have become what my mother wanted me to be...married, nice home in the suburbs, two children and a dog.  I am living the American Dream.  So I ask myself, HEY SELF!  What is you really want out of life?  HEY SELF!  What makes you happy?  HEY SELF!  Why isn\'t that enough?  Because all I ever wanted was to sing.  A career in performance is an unsteady rocky road that petrifies me to the point of paralysis.  And as much as I enjoy doing improv I enjoyed doing *real* theatre so much more.  I truly love the people in commedia and I only stay to play with them.  It\'s time I changed my focus though.  I only have so much time to play cause being a wife and mother and residency coordinator and vice president of the PTA is all a part of me as well.  Finding balance for all those pieces is becoming more and more challenging.  And I don\'t think I can add a couple of community theatre shows on top of all those other committments without toppling the building that is my life.  I want someday to run my own theatre company.  A company on a gambling showboat that teaches history to interested students by day and provides excitement to those with a taste for whiskey, gambling and fine entertainment by night.  It\'s a dream I\'ve had since I was a child on the Mississippi.  And one I find myself thinking of a lot more frequently these days.  A beautiful white boat with a red paddlewheel, balconies all around, and a calliope gently floating down the river.  You see, I like to be in control.  I like to be challenged.  I like learning.  And I like entertaining.  And for the first time in a long time I\'m starting to think that I like me.   '),(25762,3496136,'04,August,2004','2004-08-04 00:00:00','    My boss is a statistics uber geek.  We\'re working on my pet project - resident experience log database...mmmm....easily generated, concise reports, yummy....and he says...  \"You *could* add additional categories to the Procedures Table but here\'s the danger.\"  Pause.  Pause.  Pause.  \"I won\'t like it.\"  I\'m so going to Chi his square.   '),(25763,3496136,'03,August,2004','2004-08-03 00:00:00','    Suck.  Especially when they are right.   Inner Reflections  Also had my tarot cards done...here\'s a small exerpt... \"The cards do not suggest reflection or thinking things through, but rather a curbing of natural tendencies, freeing oneself from obsessions, until new knowledge makes itself known, otherwise patterns of the past will leave the questioner on the outside, trudging alone, outwardly resolute but inwardly bitter and feeling betrayed.\"  So how many times have I used the phrase \"trudging onward\" lately?  Hmmmm?  Not that I will alter my course of life based on a tarot card reading...I just found it to be eerily accurate and it has given me a little to think about.  So daoism...  I am a leaf on the great river of life...a passionate, acting with no forethought kind of leaf...but a leaf just the same.  Letting go...is easier said than done.   More Talking Instead of Writing  ASSFBB and I were discusing writing last night.  She gives great constructive criticism.  Active/passive voice choices...Clarity, fluency and eloquence.  I feel like such the beginner...oh wait, I am!!!!  She says my writing is emotionally intense...huh, imagine that from me.  I like the process and discussions of writing.  I don\'t know if I\'m any good at it.  But I\'m enjoying it...and isn\'t that half the battle, anyway?   '),(25764,3496136,'02,August,2004','2004-08-02 00:00:00','    On Saturday I met C at Joannes and we shopped for fabric and trim for bloomers for 2 1/2 hours.  She is sort of tall, about Shrew-baby\'s height, with honey brown long straight hair, large green eyes, a complexion like porcelain, generous lips and a beaming smile.  Her voice is like velvet and she has a simple, clear way of speaking but you can\'t help being somewhat mesmerized by the sound of her voice.  We\'re going to be virgins for Madame Red together, she and I.  But only on Sundays ;)   We giggled over ugly polyester fabric, we dreamed of taffeta, and squeeled over fringe.  She found blue and white fabric with cherubs on it.  Perfect bloomers for Angel.  We have the same tastes, she and I in fabric and trim.  We planned to meet on Sunday to make the bloomers at my house.  On Sunday I made the Angel bloomers.  They have pleated, crisp white fabric with a little touch of lace on the ends of the legs topped with a simple but elegant wide blue trim.  Pretty, comfortable and just right for Angel.  I can\'t help but wonder if I will ever wear them as Angel?  C arrived around 4.  We made two pairs of bloomers for her.  Hers are much fancier than mine as she has 4 types of trim on the bottom of her legs.  She looks quite wonderous in them.   We talked almost the entire time.  A shorthand has already developed between us.  We talked of where we grew up, what our passions are, a couple of influential people that have touched each of our lives.  She went to NYU and was an actress in NYC.  She now teaches communication to medical students.  She sings.  She dances.  She is spiritual in a pagan sort of way.  We got into a long discussion of religion, dare I say spirituality.  She recommended a book called \"Spiral Dance\", which I must get.  She told me about daoism and I must admit, I am very curious about this philosophy.  I am sure that I would have trouble following such a \"go with the flow\" kind of philosophy.  She described it as a leaf in the water just gently floating down the river not worring about the rock ahead.  The water will carry the leaf where it needs to go.  I was thinking, of course, as myself as the leaf trying to paddle desperately away *ROCK!!!!!  SHIT!!!!  HELP!!!*  So daoism probably more of a goal at this point.   A couple of things I want to remember about this experience is how quickly trust can develop between people of like minds.  It was an unsteady dance at first our conversation, and as we became more comfortable we continued to feel each other out, so to speak.  I am unusually good at making friends but never before have I experienced friendship where we shared matters of the soul and heart so very quickly.  The other thing I want to remember is the look of amusement that would cross her face every now and then...and it wasn\'t from something I said to be funny it was usually something about myself that I shared.  A look of amusement and acceptance and pleasure.  She takes joy in the personality of the person, not just in what they say.  We talked of sex:  attitudes, societal impacts, very briefly what we like, how we came to play whores at renfest [what we expect that experience will be like, what we want from that experience, fears and dreams].  She told me of a book called \"The Story of O\", which I really have to read as I have recently discovered the courage to like erotica.  She is married quite happily it seems.  And while I know much about her, I can\'t wait to discover more.  I have a feeling we will be lifelong friends, maybe.  Hmmm...maybe this daoism thing is easier than I thought.  Just be vague about everything...  There is so very much in the world that I know so little about intellectually speaking.  I am emotionally and intellectually a child.  But I do have lots of energy and passion to pour into such pursuits...I am happy to keep finding new doors to open, especially as some close behind me.  Less is more... The wheel of life continues to turn...  Am keeping reverie and sadness at bay...   '),(25765,4271044,'19,August,2004','2004-08-19 00:00:00','    While I was organizing my computer files I came upon an essay that I wrote in school more than two years ago. I think it\'s worth sharing, for it deals with a subject that still resonates to this day. I must warn you, though, that although I\'m fond of Gallic culture, I\'ve never had the chance to visit France, so all of my arguments are based on what I\'ve learned from the media and other source material.  ---------------------------------------------------------------------  A STANDPOINT ON U.S.-FRENCH RELATIONS  The first chapters U.S.-French history abound with tales of comradeship between these two nations. One has not to look further than LaFayette and George Washington, the nascent U.S. government and Rousseauâ€™s  Du Contrat Social  and the U.S-France, WWII Alliance. Fastforward to the present, and at first glance one can see more signs of amicable U.S.- French interaction: From the thousands of French and North American tourists who routinely visit each otherâ€™s homelands to Croissants, Euro Disney and the Frenchmenâ€™s love for Woody Allen movies.  And speaking of Woody Allen, maybe one has not to look further than to the solemn monument of his hometown, the Statue of Liberty, for the best embodiment of this mutual affection between the U.S and its oldest ally. According to the statuteâ€™s sculptor, FrÃ©dÃ©ric Bartholdi, the idea for his masterpiece sparked in his mind at a dinner with French intellectuals where the host, Edouard Ren Lefebvre de Laboulaye, asked, \"Wouldn\'t it be wonderful if people in France gave the United States a great monument as a lasting memorial to independence and thereby showed that the French government was also dedicated to the idea of human liberty?\" (1).  Another case in point is that French has been one of the top foreign languages of study in the U.S. Add to that the North American writerâ€™s penchant for French expressions in their writings and,  Voila! , one can reasonably conclude that the U.S. and France share close, friendly ties. However, to draw this conclusion is to be misled to some extent. Indeed, further examination of French-U.S. relations indicates that these countries share a cautious, if not strained, relationship despite their cordial, 200-plus year history.  On a cultural level, the French view of the U.S is nowadays less Bartholdi than JosÃ© BovÃ©. As one admired the North American way, the other wrecks McDonalds with zest. French people of course are not throwing rocks at every Starbucks they see, but they seem to grow weary at this pervading avalanche of North American culture and the flaws that it carries. This sentiment is echoed in a March 2002 report about a deranged gunman that assassinated several officials in a French town hall. In the report a French political figure named Alain Madelin said in regard to the crime that, â€œThis American-style by-product we wished not to have in Franceâ€ (2).  As noted by the D&C article  France Aims to be U.S. Alternative , thereâ€™s a growing friction between these two nations. This article tells of how France is set to contend U.S. â€œhyperpower,â€ an Americanization that extends â€œbeyond economics, technology or military might to â€˜this domination of attitudes, concepts, language and modes of lifeâ€™â€ (4). Both countries, of course, are still allies, as Western powers that they are. But now France is taking certain steps to counter this hyperpower of the U.S., who considers this French viewpoint, in the words of former secretary of state Ms. Albright, â€œartificial and unnecessary; we (French and North Americans) have such a similar background. (4)â€  Jean-Philippe Mathy, professor of French at the University of Illinois, elaborately conveys this French perspective on Americanization in his book,  French Resistance . Here he writes that Americanization \"economically, it turns citizens into individualistic consumers and subjects the national economy to the whim of transnational corporations; culturally, it debases the high standards of the indigenous aesthetic canons, displacing Versailles with Disneyland; politically, it dissolves the shared values of citizenship into a neoromantic celebration of diversity \" (3).  It is important to note that whether Mathyâ€™s argument is right or not, on a cultural basis, the French (and for that matter, the entire world\'s) distaste and consequent defiance towards U.S. hyperpower must be based on cultural preservation and national pride rather than fear of cultural homogenization. In the D&amp;C article, French foreign minister Hubert Vedrine says â€œAmerican supremacy would come at the expense of the planetâ€™s healthy diversity in political, cultural and economic thoughtâ€ (4). Of the three, cultural thought is one point of debate, because not even U.S. hyperpower can eliminate the worldâ€™s cultural diversity. Cultural diffusion, however extensive it is, cannot engulf other cultures without changing itself. History has shown (the Romans over the Greeks, the Mores over the Spaniards, the Spaniards over the Incas, the British over the Indians) that when a culture rules over another, each culture incorporates features from each other. Even the Tainos, an Ariwakan tribe of the Caribbean archipelago that was exploited to extinction by the conquistadores in no more than a century, left cultural prints that last to this day in words such as tobacco, barbecue and hammock.  Homogenization is not feasible, for diversity is an intrinsic quality of humanity. If, for instance, third-generation Latino Gen-Xers become more about Pop Tarts and John Grisham than tostones and Gabriel GarcÃ­a MÃ¡rquez, they will soon diversify and make their own culture. Proof of this is how descendants of African slaves in many American countries have virtually lost their ancestral culture and language to that of the English or Spanish, yet they have formed their own distinctive culture within their colonizersâ€™ cultures, as in hip-pop, Afro-Cuban, Garifuna, etc. If U.S. culture were to consume the burkas, siestas and sakes of the world, then what would follow is the birth of a diverse mosaic of cultures from globalizationâ€™s cultural concoction. Therefore, Vendrine and all French leaders must understand that Americanization is not a homogenizing force thatâ€™s out to eliminate cultural diversity, but just another step in the human cycle of cultural change.  It must be noted that contrary to Ms. Albrightâ€™s remarks, France does have substantial reasons to stand as an alternative to U.S. hyperpower. U.S. has all the right to extend its culture just as Gallic life has expanded from Montreal to Dakar. But just because Americanization is a natural step in cultural change, France should not let its cultural, economic and political principles erode because of it. Every society has the right to preserve its values, because just as cultural change has occurred throughout history, so has the passing down of principles and traditions from generation to generation.  In light of Franceâ€™s stance, the U.S. should welcome the competition to its hegemony rather than disapprove of it. Isnâ€™t healthy competition the core of that deeply engrained U.S. value called capitalism? Competition motivates competitors to become better at what they do, so let there be more Vivendis for Time Warner, more Evians for Coca-Cola, more for Peugeots for General Motors and more Cesars for Oscar. But on the other hand, let there be less unconstructive competition such as that that took place in the 2002 Salt Lake City Olympics where a French figure skating judge was embroiled in an International dispute that hampered the objectivity and camaraderie in the Games and, hence, the world.  If France aims to be the U.S. alternative out of national pride, the more power to them. The world population, from its conservative elders to its impressionable youths, will rightly dictate the prevalence of either culture within its members. Theyâ€™ll surely be glad to at least have a choice between two and not one.  Human liberty sits at the core of the U.S. right to extend its reach all over the world and for France to take a path of its own. Back in New York, the Statue of Liberty still stands as a symbol of France and the U.S.â€™s dedication to human liberty. Laboulaye was right to think that it would be wonderful to have a memorial for such an ideal, one that should be allowed to thrive in venerable competition between these two nations rather than decay in international antagonism. That way, for the sake of humankind, the Statue of Liberty maintains its significance, becauseâ€¦     \"to the sculptor form is everything and is nothing. It is nothing without the spirit - with the idea it is everything.â€   - Victor Hugo, May 13, 1885 (1)  Note: For some strange reason, the blog\'s template gets distorted when I add the reference section. Right now, the only way I see around this problem is for you to access them from this word  urlLink file .   '),(25766,4271044,'18,August,2004','2004-08-18 00:00:00','    September --and by that I mean college-- is just around the corner. I\'m not brimming with joy; that\'s for sure. It\'s not that I don\'t look forward to the coursework, though I do possess an impressive body of work in the wonderful art of procrastination. Actually, it\'s more due to the fact that I have not fulfilled all of my summer plans. \"Aaaaaaaaah, poor baaaaaaaaaaaaaaby\", you may say.  But no, I\'m not going to languish over my shortcomings and, besides, there are obviously more important events in this world to consider. On my homeland, the Dominican Republic, Mr. Leonel FernÃ¡ndez became the new head of state. Like most of his predecessors, he\'ll have to face, confront, avoid or just simply ignore (isn\'t it always a mix of the four?) the rampant corruption, resource maldistribution, and the eternal G8 hegemony that afflicts DR, along with almost all other Latin American states and developing nations throughout the world.  Latin America is one region where government leaders address these perennial challenges in their own distinctive ways. On the day of President FernÃ¡ndez inauguration, for instance, President Hugo ChÃ¡vez was voted to remain in power by the Venezuelan electorate. Here you have Mr. FernÃ¡ndez, a mild-mannered lawyer in favor of privatization against ChÃ¡vez, an energetic, lefttist populist whose firebrand rhetoric riles all the way to Mr. Cheney\'s pacemaker. Go west and you have president Uribe working alongside the DEA to eliminate the drug cartels in a war that apparently will never end, to the lament of Colombians, Latin Americans, and North Americans as a whole.  Go south and there\'s Mr. Toledo of Peru who\'s mostly known as the Latin American president with the least amount of public support. In Bolivia president Mesa also won an important election, although not for the right to remain in power. Instead, it was about public approval of a controversial plan to export the nation\'s gas, which entails a brew of Big Bussiness meddling, native indian rebellion and the fair distribution of the nation\'s wealth amongst the populace. It is sad to learn that part of the controversy also involved using Chile\'s ports to export the gas of land-locked Bolivia. How regrettable that between Chile and Bolivia, brothers in history and language, there is so much contempt; one that according to the news stems from the war in which Bolivia lost to Chile her access to the sea. Hopefully, somewhere down the road these two nations will reach some form of compromise.  I also hope that president Kirchner and his cabinet continue to lift Argentina from the devastating economic collapse that ocurred not long ago. Sometimes I feared that DR would also fall into economic chaos, as inflation and  apagones  (blackouts) continue to threaten its stability. In this ominous state in which poverty fuels desperation, many of my countrymen embarked on a treacherous sail to Puerto Rico that ultimately yielded nothing more than death and sorrow.  Amid their peril lies the abhorrible treason that\'s forever present in the developing world. After two or three days on sail, the  yola , as the standard-fare migrant boat is commonly known, ran out of fuel. Desperate and helpless in the vastness of the sea, the migrants were relieved to see another  yola  pass by. The captain of the stranded boat told everyone that he will board this incoming vessel to then return with the necessary fuel and supplies. After three days in which the migrants ran out of their boiled eggs, water, and sardines, they came to realize that the captain, el  maldito  captain, deserted them. After more than week at sea, they were finally rescued by some fishermen not far from where they began their odissey. Back in DR, the survivors came with tales of comptemplating cannibalism, of women being forced to share their breast milk, and of people in utter upheaval jumping to the sea. It was ultimately an endeavor marred by risk, treason and death.  The most heart-wrenching part of this story is how the \"leader\" dared to say that he\'ll come back; that he\'ll save the people who payed and risked so much to understake this trip under his wing. How many leaders, from coyotes and oligarchs to liberals and neoconservatives, have left their people in the cold? Alas, nepotism and cronysm are kings and unscrupulous self-interest their driving force. What a shame. Four our sake, let us not forget all that we are risking, the perilousness of our ongoing trek and the responsibility of our leaders, regardless of affiliation, to never,  never , dare to abandon ship, for that\'s one of, if not the most, vile and inhumane act that could arise from the heart of man.  As I said before, I have many personal regrets about this summer. But, needless to say, the real worries are those on the global scale. In this  yola  of ours, we ought to be cautious against those who would dare abandon ship and more so against those passengers that upon reaching their destination, would forget about those that stayed behind.   '),(25767,1835108,'27,August,2003','2003-08-27 00:00:00','   Yesterday after having left a post I actually did start packing.  I got two boxes done and then got called into work.  When I got there I picked up one patient.  He had just had a heart attack, got two stents and I was going to pull out the catheter from his artery (we do this a lot).  I was thinking ok, this won\'t be such a bad day.  Well then I picked up a second patient...he was very hard of hearing and very confused.  To make a long story really short...it ended up being a really frustrating day.  But it ended well...I got to go home early.  So yes indeed I not only went in late, but I got to leave early.  So you are probably thinking that I got a lot of packing done when I got home early.  Nope, none.  Today is a day off for me and I have big plans to pack.  I will let you know how that goes.  Did I ever mention that I truly HATE packing. On another note.  I went to my brother-in-laws house this morning.  I had the pleasure of giving him the news that his ex-girlfriend (of 4 months) whom he dated for nearly 8 years is now engaged and planning a wedding for May.  Unbelieveable...yes.  And I felt absolutley horrible giving him this news.  But he knows that I am still friends with her and if he heard it through the grape vine I know that he would wonder why I hadn\'t told him.  My point is that I now feel like shit because he probably feels like shit.   I must move away from the computer now and do something that will help with our move out of this place.   '),(25768,1835108,'26,August,2003','2003-08-26 00:00:00','   Although today is one of my regularly scheduled days at the hospital I am on call.  They are probably over staffed or don\'t have enough patients in the unit.  This all works well for me.  Usually on an on call day it goes like this...the alarm goes off at 5:15 a.m, I hit snooze until 5:45 a.m.  Somewhere between 5:30 and 6 a.m someone from the scheduling office at work calls and asks would I like to be put on-call for the day.  I usually ask if they will just cancel me, they say no and then I say ok o being on-call.  Usually I then go right back to bed and sleep for a few  more hours, then spend the rest of my day watching some good tv.  However, today did not happen that way.  Everything was just as it should be up until the going back to bed part.  Seeing as how we have to be out of this townhouse by Sunday and nothing is packed I couldn\'t justify going back to bed.  Actually on the way back to the bed I caught a glimpse of the empty packing boxes and guilt filled my brain.   Now it is 7:50a.m and I still haven\'t packed anything.  I ate breakfast, checked the email, surfered around for awhile and now am rambling.  Why is it that packing up all of your belongings is so crappy?  Unpacking...not so bad.  Packing....freaking sucks! But I will pack, and then if I don\'t get called in to the hospital I will take a load of clothes (currently thrown into garbage bags) over to the new house.  And when I get to the new house I will put a second coat of paint on the trim in the guest room.  After the last coat of paint, I just have to clean up that room and then we are ready to move in!  You would think that the excitement of moving in would jump start some energy inside me and catapult me upstairs and start packing.  NOPE.  I have energy to do a lot of things right now, but packing up our shit is not on the list.  Not even on the bottom of the list.   '),(25769,1835108,'25,August,2003','2003-08-25 00:00:00','   I recently became a new homeowner.  Thats right, at that age of 23 I have quickly found myself married and owning a home and with an actual career.  Its quite scary.  Much scarier that I imagined it would be.   I was a tad frightened when I graduated last year with an applied science degree and entered the work force as an RN.  I got married to my boyfriend of 6 years, and this wasn\'t so scary.  But the first year of marriage and buying a house have been quite frightening.  I didn\'t think that at 23 I would feel so tied down.   So \"stable\" I guess you could call it.  I always pictured my early 20\'s as being carefree and trying out all sorts of new things...trying to find what I like the best.  But, here I am working at job that I need (which sometimes can make it difficult to figure out if its the job that I actually want) and fixing up a house that my husband and I will be living in for quite awhile.  Which means that we are in the Lansing area to stay, ugh! With all the ramblings of how I am scared of my \"stable\" life. One would assume that I am all around unhappy. However its quite the contrary.   I do find myself sometimes wondering if I am in the spot that I am meant to be in, or the one that I am to stay in.  But I am happy...for now.  I do see some changes happening on the job front.  Not in the immediate future, but not so far off that I haven\'t already started the planning.   '),(25770,1835108,'28,April,2004','2004-04-28 00:00:00','   my goal is to use the treadmill 4-5 times a week.  about 5-6 miles (on an incline).  i know that this isn\'t a very high goal.  but for someone who hasn\'t worked out in a while this is pretty decent.   i will keep track and everything.   '),(25771,1835108,'23,April,2004','2004-04-23 00:00:00','   i\'m going to be the camp nurse at church camp this year.  i\'m actually really very nervous about it.  but excited at the same time.  i\'m hoping to find a home parish in lansing soon.  it will be really good for my marriage.  and lord knows (and he really really does) that our marriage needs some serious help right now.  how fun to go back to camp.  i spent a lot of my youth there.  it is really my only vacation this summer too.  so i am going to make the most of it.   '),(25772,1835108,'23,April,2004','2004-04-23 00:00:00','   I didn\'t go as short as I had anticipated.  I\'m the type who really listens to the stylist.  I said I\'d like to go a little shorter, but that she could just do her thing.  So she did.  Its way blonder than I had wanted, but I like it.  Its just hair, right?   Todays my birthday.  I\'ve talked (or heard from) to all my favorite people already today....mom/dad, jason, jenna, melissa and mike.  yahoo! I started to put together my new treadmill today.  After I finally got the box open the instructions said \"don\'t put this together by yourself!\"  So I stopped.  Jason is giving my lots of flack for getting it.  He says that all women who have treadmills just let them collect dust.  I would use it, if I could get it put together.   I also got new bedroom furniture.  I\'m really really excited about this.  I don\'t really love my house and it doesn\'t seem like that is something that I can fix.  So I am going to fill it with things I love.  The new furniture is suppose to get here the second week in May.  The set is with a queen size bed.  So I need to hurry up and get a queen size mattress.  I am taking my time with the decision, I\'m hoping to get a mattress that will help rid me of my insomnia problem.  We\'ll see!   '),(25773,1835108,'22,April,2004','2004-04-22 00:00:00','   Long or short hair?  I am switching to a new hair stylist today.  I\'m a little nervous.  I am very attached to my orginal hair person and I\'m sure that the next time I can drive the 2 hours to see her I will.  But today...someone new.  I am getting my hair cut and colored.  I\'m thinking something a little shorter and sassy.  We\'ll see.  Hopefully it will be fun.   '),(25774,1835108,'20,April,2004','2004-04-20 00:00:00','   AHHHHH!  Insomnia.  Not really the best problem to have.  I suppose that if I was super productive and got out of bed when I can\'t sleep and got stuff done it wouldn\'t be such a bad problem.  But I\'m not that motivated.  I simply lay in bed wishing that I was asleep.  Laying there thinking about how tired I\'m going to be at 5am this morning at work. You know what else is sucky....one of my greatest friends and support systems is moving away.  Far away.  I\'m afraid that losing contact is what is going to happen.  I\'m excited for him and his great oportunity (if he decides to make the best of it), but sad for our friendship.   '),(25775,1835108,'16,April,2004','2004-04-16 00:00:00','   does having your period seriously have to be so painful.  i mean really...like gushing blood isn\'t gross enough, but does it have to feel like someone is inside you squeezing the shit out of your uterus.  i\'m thinking \"NO\".   '),(25776,1835108,'13,April,2004','2004-04-13 00:00:00','   horrible, horrible headache.  the third time this week.  its only tuesday, so that means every night so far this week has been plagued by a dilapitating headache.  its the kind that makes you feel nauseous. i did get a fabulous fabulous pedicure today!  and i do mean fabulous.  there are few things better in life than a good pedicure.   '),(25777,1835108,'12,April,2004','2004-04-12 00:00:00','   oh yeah.  that great friend that i mentioned earlier?  maybe not so great.  i\'m not sure who i really and i mean really trust these days.  my spouse and i\'m beginning to think that thats it.   '),(25778,1835108,'09,April,2004','2004-04-09 00:00:00','   i\'m convined that there isn\'t anything better than reality tv.  really, reality tv controls all of tv these days. i mean really....the nick and jessica variety show!?!  i have to admit it...I LOVE IT!  i can\'t not watch reality tv.  i\'m an addict.   '),(25779,1835108,'09,April,2004','2004-04-09 00:00:00','   it is fabulous to actually have that great friend that after hearing the familiar voice you immediately feel better.  i had a really crappy afternoon (hence my last entry) and an okay evening despite feeling absolutley betrayed by a friend.  easter is almost here and i have to work the weekend.  so my husband and i dyed easter eggs tonight and ate a baked ham (yes i actually baked a ham...my family traditionally always has ham on easter).  but after my husband went back to bed i was back into the ugly crying and feeling really bad about myself.  i called a friend to chat for just a second.  yes the purpose was partly to vent but really was because i knew that i would feel better after the conversation.  i do.  THANK YOU!   '),(25780,1835108,'08,April,2004','2004-04-08 00:00:00','   i cried today like i haven\'t cried in a really long time.  the type of crying where you can\'t hold yourself up and your face is contorted into the ugliest face possible.  and the noise...it is unimaginable. the crying... 1.  for the dissapointment in someone i thought was a true friend 2.  for the rumor that was spread 3.  for people beleiving it and hence beleiving that i\'m capable of cheating on my husband with not only a married man but a co-worker. 4.  for the loss of a true friend (not the above mentioned from number 1 but the married man from number 3).  He is exactly what i would call a true friend.  and soon he is leaving for another state so that he and his wife can start fresh at their marriage.  and i do wish the only the best for them.  but being a little selfish...i am going to miss him and the friendship terribly. 5.  just to simply cry.  i have been dealing with a bit of an insomnia problem lately and in utter exhaustion the crying felt good.   have i mentioned that i hate lansing, mi?  well i do.  and just as i was starting to like my job, someone there has completley ruined that for me.  i will walk into work and feel like people are looking at me and believing the rumor...i cheated.  what a stupid and haneous rumor.  and aren\'t we suppose to be professionals and not high schoolers?  i work in a critical care unit for god sakes.  what the fuck are we doing worrying about who is hooking up with whom and what caused someone to quit, and why the demise of so-and-so\'s marriage.  its silly and petty and unfortunately i care what people think of me.  and i am soooooo very dissapointed in the person who had balls to actually say out loud \"we decided by process of elimination that M is leaving because he cheated on his wife and we decided that it was with sarah\"  that is ludacrious.  and she should even be considered in the same category as true professional women.  even is she really did believe this to be true (how could you call yourself my friend and beleive it, but whatever) why would you sit around and talk about it at work like it won\'t affect others lives?  asanine.   '),(25781,1835108,'08,April,2004','2004-04-08 00:00:00','      '),(25782,1835108,'08,April,2004','2004-04-08 00:00:00','   Wow...It has been a very very very long time since I\'ve posted anything.  Quite a bit has happened. 1.  We moved into our first home. 2.  I switched to working the night shift (7pm-7:30am) 3.  My husband spent 3.5 months working out of town in Alabama. 4.  We\'ve discovered the true meaning of \"the first year of marriage is the hardest. 5.  I\'ve applied to Baker College and will start working on my Bachelors in Health Administration Mgmt in June.  Gosh.  While the list isn\'t very long, it really feels long. No kidding about the marriage thing.  I sometimes wonder how young married couples make it to their third or fourth year.  We don\'t have any kids and both still under 25.  It is very tough.  It isn\'t only hard to simply just share your life with another person day in and day out.  But it is really tough to feel secure that the decision you made to say \"I do\" at the age of 22 is the right decision for the next 80 years of your life.  I feel pretty secure that the grass isn\'t greener on the other side, so to speak, but I do wonder if this marriage is the right place for me to be for the rest of my life.  AHHHHHH...does everyone feel this way when they are young and married?  With all these doubts people would probably wonder if I was married to a horrible guy?  I\'m not.  I have my list of grumbles about him (and i\'m sure he with me), but he loves me unconditionally.  I like to attribute most of our marital woes, to our young age and inexperience.  I guess time will tell me if my situation should change. On going back to school...I am excited.  I only work three nights a week.  So I have a lot of down time.  It is sort of strange to feel excited about studying.  But I really need to do something productive with my free time.  Lets not forget that finishing up this bachelros in Health Mgmt will help to get me out of the hospital and staff nursing.  I do love my job (most of the time) but I also know that I can\'t do it for the rest of my life.   '),(25783,1835108,'14,May,2004','2004-05-14 00:00:00','   Peggy\'s wedding is done.  I had a great time.  It was really fun to spend time with her (and brent).  and to spend some time with Tammy.  She is great.  We do have a lot in common and probably would be great friends if we had any of the same free time to spend together.  who knows. work was crazy this week.  good and bad.  got to see anne.  CONGRATULATIONS!  I can\'t believe that she is getting married.  That\'s crazy.  i\'m happy for her.  she is really excited.   Summer is almost here.  almost time for church camp.  i\'m really getting excited about it.  my dad and jenna will be there.  yahoo.  hopefully all goes well.  i have another meeting this saturday about it.   jenna is coming to town tomorrow night.  that will be fun!  hopefully we can see some silly chick flick. then...the furniture gets here.  i can\'t beleive that i am going to have grown up bedroom furniture.  i might feel like i am living in a house and not a dorm room.  we\'ll see. lots of stuff to do.   '),(25784,1835108,'06,May,2004','2004-05-06 00:00:00','   Peggy\'s wedding is only two days away.  I can\'t believe it.  I am very incredibly happy for her...Brent and her are very very happy.  It is just strange to think that about 1.5 years ago she was going to be my sister-in-law.  Meaning that yes she was going to marry my brother-in-law.  But nope.  I\'m just as happy for her now.  Jason is freaked out about going to the wedding though.  Hopefully he will be corgial to her.  I will flip out if he isn\'t.  When I called Peggy today to check on her she was at the hospital with her grandpa, looks like he isn\'t doing well and will be spending the next few days in the hospital.  Poor Peggy, I hope that her wedding weekend will go well, she definetley deserves it.  Horoscopes are funny.  I go through cycles with reading them.  I sometimes find myself checking them all the time, and I kind of geek myself into believing them.  They are sometimes hard to ignore, especially when they seem very very accurate.  A friend of mine called today and told me to check his out... this is his for the 8th ...\"Wait until the weekend is over to make a final decision that involves the making or breaking of a long-term situation. You may not have all the answers yet, and you really are in a rather impulsive mood, aren\'t you?\"...    here is mine for the 8th ...\"A call from a faraway loved one will arrive -- and just in the nick of time, too. Don\'t forget how close you were, and still are. Time and miles don\'t matter. The connection will always remain.\"    I\'m just guessing that maybe I will hear from a long lost friend on Saturday?  Or I will hear a message seeing as how I am a bridesmaid in Peggy\'s wedding on saturday and I\'m sure that I won\'t have my big man\'phone on me.  One of his later ones also says...\"The good news is that someone charming, spontaneous and absolutely delightful will be the cause of all the ruckus.\"  I can think of someone who is \"charming, spontaneous, and delightful\"!  Can\'t you?  hehe   '),(25785,1835108,'05,May,2004','2004-05-05 00:00:00','   got my new earrings yesterday.....0.82 tcw princess cuts!  yahoo.  i love them! THANK YOU!  I am going to a new tanning place now.  new bulbs...ouch! ordered some cute new sandals from JCrew, they are pink and have little bees on them.  i am also finally ordering my purse from dooney & bourke!  really excited about that.  i\'m thinking that the shopping is whats keeping me happy.  isn\'t that ridiculous.  i\'m not sure what type of void it is filling, but it works.   '),(25786,1835108,'04,May,2004','2004-05-04 00:00:00','   isn\'t it so very sad when you lose a friend.  i didn\'t think that that was really going to happen.  but it seems as though it has.  we said \"good bye\" but not really as if we believed that the friendship would be over.   i feel in a way like i\'ve been deceived a littled or made an ass of in a way.  i don\'t know.  its a funny feeling.  i guess i just feel sort of \'uneasy\' about the whole thing.  and i simply miss the chats and the general support.  it was a friendship that no matter what was going on in each others lives, there was always time for support and encouragement.  on a whole different note.  i am so thankful to have the next 6 nights off.  i am exhausted with work lately.  and i really don\'t want anything to do with it for awhile.  another note again...a friend ended up in the hospital yesterday.  had to pull over on the highway cause he was dizzy and starting to pass out.  called 911 and sat in Charlotte\'s ER for a long time.  the dizzy thing has been happening for a while.  I\'m a little nervous for him.   hope he is okay.  i will check on him later tonight.  i wonder if my sister is preggers yet?   '),(25787,1835108,'03,May,2004','2004-05-03 00:00:00','   So very tired.  Work has been nutty.  I feel like I work on a psych unit lately.  And its not only the patients...but the people I work with too.  Can everybody realize that we are working through a nursing shortage.  Just because I am the \"contact\" nurse for the night doesn\'t mean that I am responsible for the shortage.   This is an interesting month for me.  I should be registering for classes, but haven\'t heard from Baker about how many of my credits are transferring, we have to make three big car payments, I am in a wedding, am sort of looking for a new (or another) job, and our new bedroom furniture gets here soon.  YEAH! Exercising is going great!  Yahoo for my new treadmill. Oh yeah... my diamond earrings are ready for pick up tomorrow!  I love them and can\'t wait to wear them!   '),(25788,1835108,'25,May,2004','2004-05-25 00:00:00','   life seems tough lately.  love working out, after i actually get started.  but it is still tough to get my butt started. horrible loss of a patient.  glad i went to the funeral.  really really really nasty weather.  very frustrated with the house, and the water in the basement. furniture is on the way, maybe that will help with the house frustration.  we\'ll see.   '),(25789,1835108,'14,May,2004','2004-05-14 00:00:00','   i love my husband.  and we will be happy.  and content.  and i am committed to making that happen.   '),(25790,1835108,'14,May,2004','2004-05-14 00:00:00','   New address so that I know exactly who is reading this. getting on with my own life.  which is exactly what i was doing and just being his friend...for real.  but i can respect how that it might not be viewed as such.  bye.   '),(25791,1835108,'28,June,2004','2004-06-28 00:00:00','   its been awhile.  the computers been down, and it has been out-smarting me, at least as far as getting back on-line. i\'m going to be a mom!  some days i can\'t believe it...it is mostly the days when i picture myself with a huge pregnant belly.  but i\'m excited.  i have a friend here in yucky lansing, that is very schooled in the way of birthing.  she currently works at a birth center nearby and is going to be jason and my doula for the birth.  \"doula\" being our birth educator before the baby is here and the link between us and the nurses during the labor.  i think it will help to keep me calm, and hopefully help to keep me away from the drugs and focused during labor.  we\'ll see. work is crazy.  it is tough for me to do 3 shifts in a row every week.  i am hoping to get back on the day shift by the end of the summer.  it is just too tough for me to keep flipping back and forth between nights and days.  hopefully being on the day shift will help. the dress store in Michigan City, IN (where the bridesmaid dresses for Kimbers wedding were ordered) agreed to simply order me a bigger, and i do mean bigger, dress.  I am just hoping to have it altered to not only fit well but to be cute on me.  we\'ll see.   '),(25792,4066793,'30,July,2004','2004-07-30 00:00:00','    Ok, Kerry has truly lost touch, lets see this is what he wants to do. In a speech he gave today, he stated, if he was elected president and Usama was captured by US forces alive, he would make sure he got a fast and speedy trial here in the United States, where he would see that he was tried for the murders in NYC, Penn, and Washington DC. Wait a minute, I can here Johnny Cochran just beginning to count his money he would get from all the Usama sympathizers, hell he would probably try it pro-bono just to get the book deal, and his face all over the big screens across the world. But here is the case as it would unfold, Johnny would prove with out a doubt Usama was not on the planes, US intelligence is classified therefore not able to be used in court and completely circumstantial, and finally Usama was really misunderstood and suffered years of oppression from the west, it was our fault he turned out to be a fanatic, he really meant no harm in his statements to his underlings.   God what a moron, so here is the deal, if Usama is found, shoot him dead, cut of his head and air it every Arabic TV station.  below is the link to the article, just so everyone knows I am not making it up, and it is not a black helicopter theory, its not a Fox News site, or any other site of a vast right wing conspiracy, it is simply the AP.  urlLink http://apnews.myway.com/article/20040730/D845AAV00.html   '),(25793,4066793,'29,July,2004','2004-07-29 00:00:00','    Ok, Kerry has taken every speech he has given to date and stuck them all together.&nbsp; It really is just a big anti-bush deal, I thought his goal was to appeal to the American people and tell us what he is going to do, how he is going to do it!&nbsp; Damn, I have heard more perspective arguments from my dad\'s neighbor\'s children, &nbsp;trying to get me to give them a can of pop.&nbsp;  When are they going to explain to us their plan for the war on terror, the economy, etc, I have heard nothing from these people, absolutely nothing!&nbsp; So here is what you get, JFK will do this smarter, he will make things better because he was in Vietnam, and the French will support him.&nbsp; Sigh, so here is the new slogan.&nbsp; KERRY 04, because he is not BUSH.   If I really cared I would actually line by line his crappy speech, but there is no substance here.&nbsp;   &nbsp;   '),(25794,4066793,'28,July,2004','2004-07-28 00:00:00','    So guess what, sometime last night, after watching the DNC I am sure, so low life varmint decide that our hammock we had outside was being neglected and decided to borrow it without returning it.&nbsp; Got to love AMERICA, since we are \"rich\" ie rich being defined as having something that other people don\'t have since the George Bush was unable to provide it, (note during the Clinton years theft most not have occurred because a democrat was in office.&nbsp; HOLD THE PHONE, this is a real time discovery, Emily (my girlfriend for those who don\'t know), went outside around behind our garage and discovered the varmint was unable to carry it over the fence that was bent over and left it there to rest, so it looks like we are going to have to throw some razor wire back there to deal with the problem.&nbsp; Damn, maybe not, because if I do that, and an unexpecting criminal happens to cut himself (or herself, gotta be PC, not gender biased), then its my fault and I will get sued by a lawyer much like the kind of lawyer John Edwards.&nbsp; So whats the next step, cops in my neighborhood will not patrol (station says they need to get more tickets written for speeding, catching thieves and deterring crime only costs the station house money ).&nbsp; What next, maybe I should sit in my back yard with a large assault rifle and just \"defend my property\", at any rate cops are useless since they don\'t patrol, everyone knows there is not a police presence anywhere, so what better thing to do.&nbsp; Note this is not an isolated incident, before I moved in, and we were living out of my apartment, we had some nice kind soul, kick down our door, just a few weeks ago another nice individual borrowed my gas grill (with out returning it).&nbsp; So what to DO!   &nbsp;   '),(25795,4066793,'28,July,2004','2004-07-28 00:00:00','    Ok, I got sick and tired of sending out email spam of random political rants.&nbsp; So, I took my friends advice and started my own blog.&nbsp; Not that I am expecting to many people to actually waste their time and effort on reading this thing, but hey I might as well put it up for my own sanity.&nbsp; So sit back, reply, enjoy, hate it, praise, ignore it, I personally don\'t care, but at least I have a place to vent.&nbsp;   here is a quick link to a Ray\'s blog, enjoy his law school rants:&nbsp;   urlLink http://www.raymond42.blogspot.com/   &nbsp;   '),(25796,4066793,'28,July,2004','2004-07-28 00:00:00','    YAAAAAAWWWNNNNNNN  is about all I can really say when thinking about it.&nbsp; Hell, even Dan Rather thinks this is so boring. So what happened to the conventions?&nbsp; Once upon a time, the conventions was a place where all the delegates of each state would sort through pressing issues of the time and nominate the person they see fit, now what is it, a freaking joke!&nbsp; All they do is put on there pretty hats, censor all the speeches to be more \"moderate\" and less anti-Bush.&nbsp; The entire world has known for months that Kerry is the democrats guy, why waste all the money, why shut down an entire city for absolutely nothing, and don\'t get me wrong this all will directly apply to the RNC as well.&nbsp; The legitimacy is long gone of these things, I am not sure why they even cover it, hell on Monday, with all the heavy hitters, both Clintons, Al Gore etc speaking, nobody watched and it was on every channel, expect for one, where a lone rerun of CSI:Miami was on, which according to the ratings, everyone was watching.&nbsp;&nbsp;  Another quick note, these people don\'t care about Kerry, Edwards, Al Gore, the Clintons, in fact the only time people there get excited is the few Anti-Bush statements that were permitted by the DNC.&nbsp; Did anyone else notice that 1/2 the seats there was unfilled, even when Bill and Hillary was speaking?&nbsp; Kinda strange for a big time convention, maybe they where trapped in a cage match in the Free Speech Zone!   '),(25797,4066793,'09,August,2004','2004-08-09 00:00:00','    Yes, my neighbor\'s rule! Ok, so this family up the street from my house was letting their kids play in their pick up truck (with the keys) which is parked on the street. So guess what, as if you didn\'t already know, the car was placed into neutral and rolled and crushed another car a few houses down that is also parked on the street. So, guess whose fault this accident was, was it the neighbor\'s down the street who should not have parked their car on the street? Was it the children in the car who the parents gave them the keys to the truck? Or was it the parents of the kids who allowed their kids to play in the truck that rolled and crushed the other neighbor\'s truck on the street? I am guessing NONE OF THE ABOVE, it really is the fault of President George W Bush, since he has failed to provide education to our kids. It is more likely the fault of Halliburton and Dick Cheney as well, its all a conspiracy, had Bush been on top of things he would make sure those children were safe, he would make sure they knew how to drive the vehicle before getting into it, had he properly passed legislation to do so this would not have happened, in fact GW should write my neighbor\'s a check tonight. Hopefully John Edwards won\'t get wind of this, he may attempt to sue the city for not putting in safety devices on the street, as well as suing the manufacture of the truck, since the truck should not be able to go into drive or neutral with the key in the ignition...How FREAKING ABSURD!!!!   '),(25798,4066793,'02,August,2004','2004-08-02 00:00:00','    Yet Neil continues with a great addition of Neil Nuze at  urlLink www.boortz.com  read and enjoy:   One of the centerpieces of the Democratic stump speech these days has been that \'Bush lied.\'  The loony left likes to say that Bush \'misled the country.\'  I\'ve been trying for months now to get just one caller to tell my vast (but not vast enough) audience just what lie Bush told us.  So far, no luck.  It\'s a lot easier to run around saying that Bush lied than it is to actually come up with an example*. Still, even thought mountains of evidence, along with two reports by the British and United States governments, exonerate the president, it\'s still a rallying cry for the left.  In case you come across one of these weirdoes, let\'s examine the evidence.David Kay, the leader of the Iraq Survey Group, speaking about the intelligence said there was \"no sign that the administration pressured analysts.\"  Then there was the bipartisan Senate Intelligence Committee which completely absolved the administration of any lying.  Over in Great Britain, something called The Butler Report found no evidence that Prime Minister Tony Blair misled anyone.  Then, there\'s the 9/11 Commission, which found that the administration was misled by a flawed intelligence network.  And finally, even Hans Blix himself said who said \"I don\'t think they acted in bad faith.\"So why do people keep repeating the fantasy that \"Bush lied?\"  Because their hatred of George W. Bush is so deep, they fail to see the facts (something liberals don\'t like anyway.)Next time somebody tells you Bush lied, ask them to prove it.  They won\'t be able to. *Remember, a statement can only be a lie if you knew that statement to be untrue at the time you made it.  If I tell you that I\'m going to take Paul\'s phone call right after the break, and Paul hangs up during the break, I can\'t be tagged with a lie.   '),(25799,4066793,'01,August,2004','2004-08-01 00:00:00','    Thank you Neil Boortz, I couldn\'t have said this better: Its long, but is a MUST read!   THIS YOU MUST REMEMBER ABOUT JOHN KERRY\'S TAX PLANS John Kerry is announcing, rather proudly in fact, that he intends to raise taxes ... but only on those people in this country who are making more than $200,000 a year.  This idea is quite popular with an alarmingly large number of (stupid/envious) Americans.  I\'m going to explain why this is a bad idea.  To most of my listeners I offer this apology for the slow, plodding way I am going to go about this.  The problem is that most of the people who realize that Kerry\'s tax increases are dangerous already understand what I\'m going to tell you.  Those who think it\'s a great idea were probably educated in government schools.  That\'s why I have to take this slow.  These folks aren\'t the easiest people to educate. OK ... first of all, on the campaign trail Kerry will usually chose his words carefully.  He will say that he wants to raise taxes only on \"people earning more than $200,000 a year.\"  He intends for you to believe that he\'s talking about individuals here; individuals earning more than two hundred grand a year.  But Kerry didn\'t say \"individuals.\"  He\'ll say \"people\" or \"Americans.\"  The secret here is that married couples fit under the definition of \"people\" and \"Americans.\"  Kerry\'s tax-increase plans include any couple reporting more than $200,000 in income on a joint tax return.  So, you may be making $85,000 a year, and your wife or husband may be making $120,000 a year, and you think you\'ve escaped Kerry\'s tax increase.  Sorry!  File that joint return and join the club! Now ... just who are these people earning more than $200,000 a year?  Sure, some of them are salaried individuals or commissioned sales weasels ... but a lot of them are small business owners.  In fact, a huge percentage of them are small business owners.  To be more specific ... a huge percentage of these people are the very small business owners our economy depends on for well over 75% of all jobs. These small business owners are operating their businesses under such legal structures as a Subchapter-S Corporation or an LLC, a Limited Liability Corporation.  These legal structures allow these individuals to operate their businesses with most of the legal protections afforded regular corporations, but without some of the more onerous accounting and financial requirements.  These Sub-S and LLC business owners do not file corporate tax returns.  All of their business earnings are reported on their personal tax returns as personal income.  So, if you\'re running a dry cleaner, a restaurant or manufacturing company; if you\'re a doctor, lawyer or dentist; if you own a construction company or any other small  business as a Sub-S or LLC you will report the earnings from that company on your personal tax return.  You can see that it won\'t take much at all for your income to exceed the magic Kerry-targeted $200,000.  Now just what happens when you raise the income taxes on these $200,000 + income earners?  Come on now.  I know you went to government schools ... but you can see the effect here, can\'t you?  You\'re not really raising an individual\'s taxes, you\'re raising taxes on a business.  These small business owners turn over to the government in taxes the money they could be using to expand!  Why do they want to expand their businesses?  Because the more business they do, the more money they make!  To do more business they may need more employees!  But if the government is taking more of their earnings, then there is less money available to hire new employees or purchase equipment that is needed for business expansion! This is why you see economic growth following tax cuts.  Sure, the extra money in the pockets of individual taxpayers doesn\'t hurt ... but the real growth comes from small business owners using their tax savings to expand their businesses by purchasing new equipment and hiring new people! Does Kerry understand all this?  Two weeks ago I would have told you that he understands it all-too-well.  Today, I\'m not so sure.  After all, just last week I interviewed a congressman from Detroit, John Conyers, who didn\'t even know that American corporations have to pay federal income taxes to our federal government for earnings realized solely from overseas operations.  So, it\'s not beyond the realm of possibility that Kerry might be as dumb as most other Americans on this issue. I\'m going to make the assumption, though, that The Poodle knows who these people are he wants to hit with a tax increase, and he knows what the effect will be.  He knows, but he doesn\'t give a damn.  For Kerry, it\'s all about power.  He knows that the vast majority of Americans couldn\'t write a coherent sentence about the tax treatment of Sub-S and LLC small businesses.  He knows that when he talks up a tax increase on \"Americans making more than $200,000 a year\" that most people create a mental image of some high-roller somewhere simply getting a $10,000 paycheck every two weeks.  The voters Kerry is after know that they\'ll never see a paycheck like that, and by God they want to punish those evil people who do.  After all, they\'re only making that kind of money because they\'ve cheated someone else or because they have crooked lawyers.  They\'re not really working people like us, are they?  They just sit around all day polishing their Benz and sipping Chardonnay.  Those $200,000 a year types were just lucky.  After all, didn\'t Richard Gephardt say that they were the ones who \"won life\'s lottery?\"  And don\'t they call us the \"less fortunate?\" So, if you happen to be on the fence in this election, here\'s your question.  Do you want these small business owners to keep that money and spend it on business expansion, or do you want the government to take it and spend it on who-knows-what in some congressman\'s district to help him buy some votes? If there was any decent economic education in our pitiful government schools we wouldn\'t be discussing this right now.   '),(25800,4066793,'01,August,2004','2004-08-01 00:00:00','    Mr. Kerry in your speech and speech\'s that followed the DNC you make statements that we need affordable healthcare, strong environmental polices, a smarter war on terror, along with a lot of other claims. Now my question is that you have been a senator for many years. None of the things you have stated are \"new ideas\" for voting issues, but why have you never proposed any legislation in the past 20 years to solve these problems? Where is your healthcare bill, your environmental bill, your legislation to wage a smarter war. I would think it would be your duty as a senator to propose as much legislation that can help the war on terror, since it affects both republicans and democrats, but yet no legislation, just rhetoric that Bush is not doing a good job? Enlighten the American people, let us know what YOU are going to do.   '),(25801,4066793,'01,August,2004','2004-08-01 00:00:00','    Well, John Kerry has finally unveiled his plan, but the catch is that its a Secret. He has made claims he has cards he can play with heads of states that will allow us to move our troops out and get full international support. I am interested for a senator who wants what\'s best for the world, he sure hasn\'t proposed any legislation on this issue in the senate, but that\'s ok, we know he does truly have a plan, to bad we can\'t compare it to Bush\'s because its all hush hush. Read it at this right wing biased news source:  urlLink http://www.newsmax.com/archives/articles/2004/8/1/145747.shtml   '),(25802,4066793,'01,August,2004','2004-08-01 00:00:00','    Ok, last night Emily and I saw \"The Village\".  If you have not already seen it, don\'t.  Its god awful.  I wish I could get my 2 hours of my life back that I lost watching that horrid film.  The twist is predictable, and not very shocking.  I am not going to spoil the movie, but it was slow going and boring, I give it negative 5 stars.   '),(25803,2694370,'18,January,2004','2004-01-18 00:00:00','   Well, we pick up actually in the middle of October, and this is just a quick summary of what happened the last two weeks.    10/23/03    We spent some time reviewing and making the list of what had happened in the group since April 2003.  We had a quick discussion from Matt. 10:27 about laying our lives down for Christ.  As we do this and delight ourselves in Him, He will give us the desires of our hearts... His desire for the lost.  We then began to pray over the map of our city.  As we were praying, the Lord gave an interpretation of tongues that were spoken.   \"There is a power in the heavens... it is very great.  There is a power in the heavens... Who will stand against it?\"   We believe that this is refering to the principalities in our area of the world.  Two of them locally being Fear (Superior) and Unbelief (Duluth).  We also believe that God has put angels at the gates of this city, and that through prayer and fasting God is going to turn this city around.    10/30/03    We began this night by discussing from 1 Corinthians 14.  The theme seemed to be that when God speaks, we need to listen.  Don\'t just put it aside and forget.  We also discussed the relevancy of the gospel today.  The only way to present the gospel in a relevant way is to LIVE it!  After some time in prayer, we believe that God was speaking:   \"If you do it, it will happen!\" & \"If you go I will do it\"   This was a specific word of healing for Laurel B.  We also felt that God wanted us to be bold and step out.  No more timidity!  So, we packed up and headed out to Laurel\'s house.  On our way, we stopped to speak with some teen guys on the street corner.  One in particular God was calling!  It was so clear.  We spent a long time praying for Laurel, and it was clear that He is not done with her or us yet!  '),(25804,2694370,'17,January,2004','2004-01-17 00:00:00','   F.I.R.E. started in April of 2003 as a group born out of the desire of individuals from Living Stones Fellowship to have a unique outlet for evangelism.     The idea...   The idea behind the group that makes it unique, at least for our church, is that it\'s goal is to be totally Spirit lead in it\'s structure from week to week.  Nothing is really set from meeting to meeting other than the place an time we meet. (...and sometimes that even changes!) It is a time to step out of the box of traditions and the way things have always been done, with only God\'s presence and His Word as our guide.   Normally?   We usually take time to discuss anything that may be on our hearts for the evening, then we begin to pray and sing songs of worship.  We spend time in God\'s presence and wait on Him with open hearts. The attitude of the group is that if we need to stay there and pray we will, and if He tells us to go, we will.  His leading usually comes in the form of an overwhelming urge to do and/or say what is on our hearts.  As long as it lines up with His Word, we are off to do it!   Why this blog?   Obviously, we just started this blog recently, and have been meeting for almost a year.  We have some catching up to do to fill you in on what has happened.    The idea to keep a record from week to week was first brought up in May 2003, but for some reason, it didn\'t actually start until October 2003.  So, the first few months are summed up in the abbreviated list below and I will enter separate logs for the months following.   So far...   4/1/03 - 10/23/03      1st Street Ministry to 2 men   Prayer at St. Mary\'s hospital (David)   Enger Tower Prayer (twice; \"Pray West\", Red & Green lights on the piers representing angels)   Cookies to our neighbors   Prayer at UMD (hotel, prayer about a abortion/suicide decision being made)   Grandma\'s Marathon event (balloons, candy)   Neighborhood clean-up in Duluth Heights   Barnes & Noble (spoke with girl in the coffee shop)   Central H.S. Prayer (silver lining/heaven)   Bus Stop (spoke with Maria)   Skyline Prayer (out west)   Oneida Cemetery(Prayer for Brady family), Canal Park (spoke with Ken)   Canal Park with telescope for the lunar eclipse   Prayer at the new church property (Melissa J. at D.Q.)   Prayer in Superior (Train Tracks, Brett?)   Country Kitchen (2 times, spoke with older gentleman, sent letter of encouragement to owner, ate good food!)   12th Ave E & 5th St. prayer   St. Lukes Emergency room (Tina----Martha\'s foot healed)   Much Prayer!   Groceries to Swansons/student     It is always interesting to look back and see where it is that you have been, and remember things and times that may have been forgotten. For those of you in the group, I hope this brings back memories of our first days, and we look forward to many good days to come!  '),(25805,2694370,'24,May,2004','2004-05-24 00:00:00','   Our first meeting for F.I.R.E. was not held on Sunday, May 23 as indicated in an earlier post.    I guess we will postpone it for a week or two.  I will keep you informed on the new time when we have a date set.  '),(25806,2694370,'18,May,2004','2004-05-18 00:00:00','   Here are some interesting links that I have found about evangelism and blogging:    urlLink \"Dialog: Breaking the Bubble (Christian Weblog Evangelism)\"      urlLink Mikao\'s World     I think these deserve some consideration.  '),(25807,2694370,'17,May,2004','2004-05-17 00:00:00','   There comes a point in any type of group/small group where the question becomes \"where do we go from here?\" or \"is this all there is?\"  I believe that this is the time to take a step back and evaluate what is happening with the group.  It may even be the best thing to take a break, or even end the group at that point. Or, it might be time to hand the group off to a new set of leaders with fresh vision.  At any rate, that is point we found ourselves at with the F.I.R.E. group last November/December.  It wasn\'t actually until sometime in January that we made the decision to end F.I.R.E. for an indefinate period.  F.I.R.E. took a break for a couple of months and is about to come back as a monthly event style meeting.  There are three main reasons for the change in frequency. One is that we are starting up more weekly cell groups, and would like to see people get involved in those as much as possible.  Secondly, it gives us more time to focus on one event, and to get it right.  And finally, and most importantly, we believe that this is how God wants us to structure it for now.   Here is the main idea...   F.I.R.E. will begin again in May. This time however, it will not be a weekly meeting. ( Cell groups are taking that place, and are better suited as a weekly meeting. MAKE SURE TO GET INVOLVED WITH YOUR CELL GROUP! )  The F.I.R.E. meetings will be held monthly as \"outreach events\".   We will start each meeting with a prayer time, and then we will be off to do what we believe God has shown us for that time. This starting prayer time will also be a planning session for the next month\'s meeting.    That is where it stands with the first meeting tenatively scheduled for Sunday, May 23 at 6:00pm.   For more information,  urlLink visit www.livingstonesfellowship.org  '),(25808,2694370,'01,July,2004','2004-07-01 00:00:00','   This group is once again on hold.  It appears that it will be August or September before we get this up and running as planned.  Stay tuned for more info here!  '),(25809,2694370,'28,May,2004','2004-05-28 00:00:00','   You may be wondering what I was doing this morning at 1 am posting a link to some blog about some people moving to an apartment in Australia.  The answer is... I don\'t know!  Well, I was not feeling well, and didn\'t feel much like sleeping. So, I was awake and wandering the internet in and out of blogg-dom.  I came across this blog that I found interesting so I posted it on my own blog.  It is kind of interesting looking at people\'s blogs.  It\'s kind of like high tech people watching.  And it is amazing that people will just let you into their lives like that... to follow what they are doing.  I wonder if anyone is doing the same to me?  It\'s fascinating isn\'t it?  '),(25810,2694370,'27,May,2004','2004-05-27 00:00:00','   Ok, what about this one??   urlLink James and Elicia\'s Adventures in Australia\'s Winter Wonderland  '),(25811,2694370,'27,May,2004','2004-05-27 00:00:00','   I have this urge to get to know my neighbors.  They are all in some way, shape or form going through some sort of a crisis.  These people are my friends and acquaintances, but I feel like I need to pour myself into them.  Actually I feel the need to pour Jesus\' love out on them.    What do I mean by that?    We know some of them better than others.  Some we have over for coffee, or we will go and visit every now and then, but I think we need to go deeper. Like we are to try to form some more real, close friendships, and somehow help bring healing in their lives.  I believe that it is what I am supposed to do. So I am going to try.  '),(25812,2694370,'26,May,2004','2004-05-26 00:00:00','   The memorial service for my little friend Cameron ( urlLink who I wrote about last week ) was yesterday afternoon.  Aside from my own father\'s funeral, this was probably the most difficult one I have been to.    There was a very moving photo and music presentation honoring Cameron\'s life, and his father Chris gave a eulogy that was probably the most beautiful and eloquent I have ever heard.  We will miss Cameron so much.  My oldest daughter is about his age. My wife teaches his Sunday School class, and we consider their family so close to our own. These will be hard days ahead.  But... We are looking forward to the day when we will see him in heaven and be there for the joyful reunion. That is our hope in times like these.  Again, please make sure you take the time today to let someone you love know that you care about them.  Grab hold of these opportunities before they pass you by.  '),(25813,2694370,'23,May,2004','2004-05-23 00:00:00','   urlLink   Last summer I picked up a pretty fun hobbie called  urlLink geocaching .  Basically, it involves using a GPS to locate a certain location in which a treasure is hidden.  It\'s high-tech treasure hunting. I like it because it gets me out to places I would never have found on my own. Seems like the type of sport you really enjoy, or think is a waste of time, but if you have a GPS, I would recommend giving it a shot!  '),(25814,2694370,'22,May,2004','2004-05-22 00:00:00','   9 1/2 hours working outside on the  urlLink new church building  today! In the pouring rain I might add!  Oh and it was only 41 degrees F.  So yeah, I spent the day cold and wet.    But we did get all of the trusses (is that how you spell trusses?) on.  I need to sleep!  '),(25815,2694370,'21,May,2004','2004-05-21 00:00:00','   This is what a do for a living (currently)...  I take Quark documents and turn them into  urlLink this .  I say that I do this currently, because that is probably going to change.  My job as I know it is currently being shipped to India.  So, it remains to be seen what I will be doing a month from now, but for now this is what I do.  This is not something I always dreamed of doing.  It is just something that kind of happened.  People often ask me, \"Do you like your job?\", and I don\'t really know how to answer that.  It is something that I do, and I believe that God has used it to supply our needs.  I don\'t dislike my job.  It\'s just a job.  So, I do it.    What I really enjoy is people.  I enjoy ministry and reaching out to people. I enjoy relationships. Even though they can be a mess sometimes...  I would say that this is my calling, and I guess the best part about my job would be people.  That\'s what I do.  '),(25816,2694370,'20,May,2004','2004-05-20 00:00:00','   I have one thing to request of all of you who read this...  Do you have kids?  Give them a hug and tell them you love them today.  Do you have parents?  Do the same.  Do you have a wife, husband, girlfriend, boyfriend, best friend, anyone that you are close too??  Let them know how much they mean to you!  I compose this message with a heavy heart.  Some very good friends of mine lost their 4 year old son in a terrible accident yesterday afternoon.    It happened in an instant, and he was gone.    So I am begging you to please let someone you know feel the love that you have for them today.  And while you are at it, let them know that God loves them too!  You won\'t regret it.  '),(25817,2694370,'18,May,2004','2004-05-18 00:00:00','   Here is a pretty cool blog that I found if you want to take a look:   urlLink http://mikao.blogspot.com/   I also added the the \"Our Links\" section to the side bar.  '),(25818,2694370,'14,May,2004','2004-05-14 00:00:00','   This blog is an experiment I am starting.  I want to see what this will become as it grows and evolves.  I will open it up to a few friends and see what happens.    The idea is that there are no real rules about how you use this blog, but please use it as much as you can. hmmm...  '),(25819,2694370,'30,June,2004','2004-06-30 00:00:00','   I got half the day off yesterday!    Read the little story all about it...  urlLink Smell prompts evacuation   (about half way down the page)   All I did was get some yard work done, but it was nice to have some free time off!  '),(25820,2694370,'28,June,2004','2004-06-28 00:00:00','   My wife and I are going to try sharing this blog.  I have decided to separate our posts with some icons.  So... for the time being, if you see this:     You know that Martha posted this!   If you see this guy:     You know that Mike posted this!  I hope that makes it a little less confusing.    This should be fun!  :)  '),(25821,2694370,'28,June,2004','2004-06-28 00:00:00','   Our church made the paper this weekend, and there was a picture of me printed under the headline.  (ok, I was about an inch tall, and you couldn\'t really tell it was me unless you read the caption...) But it was me none-the-less, and I feel really special about it.  Actually, it was a really nice story about our church.  I think you might like to give it a  urlLink read .  ( I especially like the last quote in the article by Chris Swanson.  Their family is dealing with the death of their 4 year old son last month, and he is still looking to reach out to others.  I see Jesus in them so much! )  In other news, I am trying to get my wife to sign up and contribute to this thing!  I would be so honored if she would, because I think it would add the missing dimension that this blog needs. Everyone who has read the posts so far has no idea that they are only hearing from half of me!  (by the way, I know that she will read this and feel so guilty about not doing it yet... :)...)   So, I think you will be hearing from her soon too!  '),(25822,2694370,'25,June,2004','2004-06-25 00:00:00','   Seems that a lot of our friends have left town on vacation.  Some went to Mexico, some went to Montana, some are in Houston, and others are on a road trip south.  We are still here counting pennies, and holding down the fort I guess.    It is really quiet all the sudden though.  The phone isn\'t ringing as much, and we find ourselves wondering what to do with our time.  This weekend will be a quiet one too.  Which is quite alright with me!  Next weekend will be very busy on the other hand, so I guess I should enjoy it while it lasts.  And, maybe I could catch up on some projects around the house, or friends I havn\'t connected with in a while.  I have also decided that this blog needs more pictures.  The one in this post is of a summer sky near my home.  Looking down the avenue toward Lake Superior.  '),(25823,2694370,'23,June,2004','2004-06-23 00:00:00','   This should keep you buzzzzzzy for a while!   :)   urlLink zzzzzzzzzzzzzzz  '),(25824,2694370,'22,June,2004','2004-06-22 00:00:00','   I just want to update this while I have a couple of seconds to do so.  It has been a busy week or so all around.   Work has been nuts!  I have been trying to keep up with all the various projects, but I feel like as soon as I get one done, there are four more on my desk.  It really helps to raise the stress level, that\'s for sure.  Typically when I get done with work (usually late each day)I go home and have dinner, and then it is off to the next task.  Work on the  urlLink church building , have  urlLink cell group , meet to go over the church treasury...  Somewhere in between I spend time with my family, and find some time for myself.  Then I go to bed so I can wake up in the morning and go to work.  That pretty much sums it up for right now... that will calm down soon (I hope).  I had the opportunity to preach on Sunday, and I think it went well.  It has been a while since I last preached, so I brought a bit too much material to cover, but I got most of my point across.    \"Receive your call from God, Step into that call, and Prepare to stand your ground.\"  I took the overview of the life of Nehemiah.  Cell group tonight!  We are planning a kick-off party for next week, so it should be fun!  Thats all for now.  '),(25825,2694370,'22,June,2004','2004-06-22 00:00:00','   Another good link...   urlLink Ship of Fools: the Magazine of Christian Unrest  '),(25826,2694370,'14,June,2004','2004-06-14 00:00:00','   I had an awesome weekend!  First of all, this was the first weekend that we had good weather in a long time.  It has been rainy and cold for most of our spring.  But this weekend we had sun and 70\'s which is always good.    We got a lot done on the construction on Saturday.  We managed to get most of the windows in and it is starting to look a lot more like a building.  I really am enjoying seeing that all come together.  We had to let the Ballards know that we won\'t be able to go to MT with them this year. It is actually a load off my mind and heart now that the decision is finally made.  I would have liked to have gone, but for some reason, or many reasons, this year wasn\'t the year to go.  My wife and I celebrated 5 years of marriage on Saturday.  It is amazing how fast 5 years can go, and how much can change in that time.    So, here we are 2 kids later, looking back on the years we have had.  They were good years, and we accomplished a lot.  We learned a ton.  We loved even more.  Now we are looking forward to the next 5 and wondering what they will bring.  I want to finish my Berean studies within 3.  Martha is starting a new phase of ministry.  Our children will be 9 and 6, and who knows how many more will be here by then...  ...and those are just some of the things we think will happen.  I wonder more about what God has planned for us.    We will continue to follow Life HIS way!  '),(25827,2694370,'09,June,2004','2004-06-09 00:00:00','   Had our first \"real\" cell group tonight (in my opinion).  I take that back! As far as that goes, I have found that real is how you see it... or what you make it.  So, I will say there was a significant change in the way cell group went this evening.  At any rate, we had a good discussion about Ephesians 6.  Specifically the part about the armor of God.  I could have talked all night, and I had a hard time holding back.  Man, I really enjoy being around a new Christian.  There are so many refreshing questions to think about.  It is weird how sometimes it is easy for me to forget to ask good questions.  Since when did get to the point where I have everything figured out?   I think I am slowly figuring out what my heart is yearning for through all of this.  I guess I am going through a time of wishing to be a bit more child-like in my imaginings of all things Christ-like.  Like... like... like...   In other news,  I am struggling over whether we should go with the Ballards to Montana in a couple weeks.  My heart is divided with my mind.  Part of me worries that we will not have the funds to do it, but then I have those moments where I feel that we really need to go... like right now as I write this.    Can someone just confirm this to me one way or another??  Well, I will sleep on it again...  '),(25828,2694370,'08,June,2004','2004-06-08 00:00:00','   Well I let it go almost a week since my last update, and I can\'t believe it has been that long.  It is amazing how time flies!  What has happened this last week?  Not too much.  Work has been crazy busy.  I can\'t keep up with all I am trying to do, and I don\'t really know why.  I think I may be too much of a perfectionist sometimes. I have to try and find the balance between worrying about things being perfect, and letting them be good enough.  The family is doing pretty good.  Although, James managed to fall yesterday and gave himself a pretty nasty gash on his forehead.  Had to get 5 stitches!  (ouch!)  Well, I am at the end of my lunch break so...  off to work I go! :)  '),(25829,2694370,'02,June,2004','2004-06-02 00:00:00','   Check out this link...   urlLink The Dying Church  '),(25830,2694370,'02,June,2004','2004-06-02 00:00:00','   How rich are you??    I found this little test while searching for something else.  I don\'t know how this works or how accurate it is, but it does make you think about what you have, what you don\'t, and what you really need...   urlLink Global Rich List  '),(25831,2694370,'01,June,2004','2004-06-01 00:00:00','   Went to coffee this morning with my brother-in-law.  We finally have a decent day as far as the weather goes.     urlLink Duluth is REALLY beautiful...  when the weather cooperates.  There is no place like it!  Check out  urlLink Northern Images  to view some great photos of where I live!  '),(25832,2694370,'01,June,2004','2004-06-01 00:00:00','   Um...  I have yet to look at this... but it looks like it might be cool???    urlLink Our Hangout  '),(25833,2694370,'01,June,2004','2004-06-01 00:00:00','   So I will try to recap my weekend just for fun.  It was miserable weather for most of the weekend.  Cloudy and cold and just for kicks some rain and fog mixed in!  We went to the Ballard\'s house for dinner on Friday.  That was good.  We had a nice dicussion about the state of our church both Spiritually and emotionally.  We agreed that God is up to something and is drawing us together as a group.  It is so good to see how close we can become with God\'s love!   Saturday was a work day at the  urlLink new church building .  We got a lot of little stuff done even though it didn\'t look like a lot.  I am off to work up there again this evening.  Sounds like I will be hoisting bundles of shingles to the roof.  (oh joy!)  Sunday was church, which went pretty well.  The rest of the day we just hung out at home.   Monday was Memorial Day holiday.  I had a Company vacation day, so that was nice.  I just wish it was nicer out!   We went to a service put on by the local VFW post to honor the veterans who have died.  Then we went to visit the various grave sites of my family.  Later my family got together for ice cream.  It was a nice day.  Today I am back to work and swamped as ever!!! So, I better get to it!  '),(25834,2694370,'26,July,2004','2004-07-26 00:00:00','   Today was kind of a blah day...  The events of the day left me with a bad feeling.    First thing I guess is that I am dealing with some relational issues right now that I am just not sure of.  It makes me a bit nervous.  Second, we found out that the  urlLink new building site  for our church was burglarized last night.  The thieves got away with about $1600 worth of tools.  I was shocked that someone would break into a church, but apparently there was another church in the area that had the same problem recently.  It is such a shame...   urlLink     picture of our new church building   Third, I went to work at the building tonight, and just have such a difficult time not having much experience building things for myself.  Does make me feel a bit out of my element.  I know that every little bit helps, but it makes me wonder how much helping I actually do, and how much hindering I do sometimes. (no one likes to have to teach people when they are trying to get things done...)  So, I guess if I were to just go by my feelings, things wouldn\'t be too good.  BUT, I am relieved to know that my feelings don\'t always echo the Truth.    The Truth is what I base my life upon!  '),(25835,2694370,'23,July,2004','2004-07-23 00:00:00','   urlLink    A  urlLink friend of mine  offered some panoramic shots on his  urlLink website  recently, that were so beautiful, that I had to offer one of my own.    Actually, this image was found on the  urlLink website  of a local photographer.  I visit this site often to refresh my desktop images.  It has some really beautiful shots of the place that I have lived in all my life.  '),(25836,2694370,'22,July,2004','2004-07-22 00:00:00','   Well I do agree with Mike I do not like the new set up, I could not figure it out! But he finally showed me how to do it. Then it seemed like every time I had a thought to write, it was a bad time! My weeks have been pretty packed. With every day stuff to just getting out and finally getting to enjoy the weather.   Soooo.... about the bikkie thing....I knew that I should not have commented on that! Oh well. My Grandma is who started it and I suppose my mom just picked up on it. But the way I usually would hear the word bikkie\'s used was when my mom would say \" I\'m gonna spank those bikkie\'s!\" So there you have it. :) Was it worth waiting for?   I\'m off now to put kids to bed, till later!  '),(25837,2694370,'16,July,2004','2004-07-16 00:00:00','   Well, this is my first post since  urlLink blogger  gave us the fancy \"WYSIWYG\" editor to use... I don\'t like it.   And neither does my wife by the way!   I have included a pic of a rainbow that I took a couple months ago. A little dark, but a pretty interesting image I think.  I am just sitting here looking through various blogs I read and decided to add a couple to my blogroll...     urlLink Exploring Orthodoxy  (Ian\'s blog --  thanks for the comments and B-day wishes mate! )   urlLink  >>    urlLink spirit farmer    In other news, my geocache has been found!!!  urlLink Read about the missing cache here.  I guess some kid found it and brought it home to show his parents.  Once they finally got it open and realized what it was, they sent me an email.  They put it back together and want to see it replaced in it\'s original location!  I will have to find a rather stealthyway to hide it this time so it doesn\'t get \"muggled\"!  I am really tired now... so I sign off with that!  '),(25838,2694370,'14,July,2004','2004-07-14 00:00:00','   Take a look and make your thoughts known...    urlLink The Most Controversial Bible Ever  '),(25839,2694370,'12,July,2004','2004-07-12 00:00:00','   It is my daughter\'s 4th birthday today!   That is a picture of her on the right.  She is so excited for today.  As a matter of fact, she has been counting down the days for the last week! (which is a long time to a little girl!)  This is the first year that she has really been hyped up about the whole thing, so it has been pretty fun.  She is going to the zoo today with some friends.  This evening we will have a party with the family.  I should be a fun one!  '),(25840,2694370,'12,July,2004','2004-07-12 00:00:00','   This one is going around, but I had to add a link so you can take a look... urlLink click here!  '),(25841,2694370,'12,July,2004','2004-07-12 00:00:00','   Well, I will let my wife explain that last post.  I guess her Grandma my have started an alternate definition of the word bikkies.  Anyway, time to move on with this blog into other topics I think.   President George W. Bush will be in my hometown of Duluth tomorrow. I think that is pretty exciting!  It is not every day that you get to see the President of the United States in your city.  Although, I think there was more excitement over the visit of Senator John Kerry a couple of weeks ago. This part of the U.S. happens to be a fairly liberal area demographically.  The fact that we do have both Presidential candidates come through in the election year is pretty interesting though.  It also makes for some interesting conversations around the water coolers as well.   I was just thinking that it is amazing how quickly 4 years has gone by since Bush was elected. My daughter (my oldest child) is turning 4 on Wednesday as well.  Time is just flying by, and it is causing this desire in me to savor every moment.    I am realizing more these days... actually since my last birthday that I have a lot of things to accomplish in my remaining time here on earth.  Things that I would like to see accomplished that I also believe God wants to see accomplished in me.  We watched a video this evening of a sermon preached a couple of weeks ago by Pastor  urlLink Bill Johnson  of  urlLink Bethel Church  of Redding, CA.  The message was about taking up and reclaiming \"lost mantles\" of the spiritual fore-fathers, and receiving the inheritance that has already been given to us.    I am still mulling it all over in my mind, but somehow it has inspired me to look at the last few years of my life and take stock of where my heart has been.    I am excited to see what God is going to do with this, because I can feel something happening within my heart.  I believe a change is coming again!  Oh boy!!  ;)  '),(25842,2694370,'09,July,2004','2004-07-09 00:00:00','   I must say..... this blogging is starting to catch on with me. That is usually how it goes, it takes me awhile to get into somthing, but once I do I find I like it.  To comment on the bikkies thing...I have heard of that word before, actually my mom used it quit a bit when I was little. Only in a different manner. I\'ll just leave it at that. I am sure I will be questioned by my husband when he reads this! Till next time.....  '),(25843,2694370,'08,July,2004','2004-07-08 00:00:00','   Well here you go!  Have a look...   urlLink Biscuits --- Gourmet Australian Bush fruits spices  '),(25844,2694370,'08,July,2004','2004-07-08 00:00:00','   It seems that a comment I recently made on another  urlLink blog  has touched off  urlLink a bit of discussion (and  urlLink here ) in the blogosphere.  I have been visiting  urlLink James and Elicia\'s Adventures in Australia\'s Winter Wonderland  ever since  urlLink I first blogged  about it in May.    The other day, I was reading through and came across the word  urlLink bikkies ...  what?  You read it right,  urlLink bikkies !  As in:   \"when she arrived home she carried beer, sausage rolls, a jam donut, cheese and  bikkies  and chocolate\"  (emphasis mine)    I have never heard of this word before... so call me uncultured... but I had to find out. Bikkies by the way are biscuits.  Who knew?!?!  I don\'t know, I think it is pretty interesting.  Why is this cool you ask?  I guess for me it still boggles (or should I say \"bloggles\") my mind how communication has changed these days.  I can sit in the comfort of my own home and read posts that someone has written from half way across the world.  (I don\'t want to sound too old fashion, but I am a child of the \'80\'s afterall...)  I also find it fascinating that there are so many differences in cultures that seem so much the same.  Something as simple as the language can set us apart, and heaven knows that there are probably more complex cultural differences that we could talk about at great length.  But, as I read through this  urlLink blog  I am drawn to the similarities in the human experience.  Without getting too deep, I think if you read through you will find the same thing.  Somehow I make a connection with these people from another continent just by reading what they have to say.  As I have said before, it is fascinating to me!  I am also amazed that my beautiful wife has finally decided to join me!!! :)  '),(25845,2694370,'08,July,2004','2004-07-08 00:00:00','   Well it has taken me weeks, as my husband has said, to sign in and actually right something. And by the way he did not have to bribe me to do it. I just finally have the time to sit for a second and add my thoughts. Which to tell you the truth are kind of scattered at the moment and a little pre-occupied. We\'ve had a very busy holiday weekend, and I am trying to get everything back in order. I think I am about there. So hello to all of you, and sorry to you mike that I have taken so long. Bless you all!  '),(25846,2694370,'07,July,2004','2004-07-07 00:00:00','   I am really good at opening those cans...  I think that I have done this a few times lately.  Once for sure at work, and I think I may have done so with other relationships lately.  It seems like a combo of misinformation and bad timing have gotten me.  (That and the fact that sometimes I can\'t leave well enough alone...)  What a mess it leaves me in, but every once in a while, it turns out to be a good thing.  Who wants to keep a can of worms anyhow? Isn\'t it better to open up and let them out so they can be dealt with properly?  Well, that is my philosophy anyway...  Things are good here as far as Life goes!  I couldn\'t ask for much more.    I do keep trying to convince my wife to blog, but it doesn\'t seem to be working. I have asked, begged, guilted... about the only thing I haven\'t done is bribed.  Well, it will happen one of these days.  She is just keeping us in suspense!  '),(25847,2694370,'06,July,2004','2004-07-06 00:00:00','   Well, we had a great time this weekend.  It was good to have some time off to do whatever we wanted.   Just a basic recap... we went up to the lake Friday and Saturday, and spent Sunday and Monday at home. I think that Sunday was the most relaxing day of the weekend.  We went to church that morning and had a picnic right after, and then we just spent the day at home.  It was great!  The weather was pretty miserable though!  Friday was really the only nice day of the bunch.  Saturday was HOT in the morning, and rained in the afternoon. Sunday and Monday were cloudy and cold with off and on rain. It seems like we have only had a handful of nice days this summer so far.  Last night we went out to watch the fireworks (they were postponed Sunday because of rain)!  I was a good show, and the kids loved it!  '),(25848,2694370,'01,July,2004','2004-07-01 00:00:00','   I need a vacation very badly! I am taking tomorrow off work, and we are going up to the lake for a couple of days.    Maybe do some fishing or swimming with the kids.  Enjoy the summer sun!    It is going to be great!  But it also means that we will not be updating the blog at all (for a few days anyway...).  Of course we will be celebrating Independence Day this weekend, so there will be a lot going on.  Hopefully we will have a few pics to share when we get back.  '),(25849,2694370,'06,August,2004','2004-08-06 00:00:00','   Hey!  Does anyone want to join our really really great discussion forum?    If you would, just visit this site:    urlLink click   OK... truth be told, I am having a hard time encouraging people to discuss on this thing.  So, please join and post your thoughts.  I think it will help encourage others.  :)  '),(25850,2694370,'05,August,2004','2004-08-05 00:00:00','   To add to mike\'s thoughts! That night was a true summer evening in my book! We really had a blast. I do feel a little guilty about Mike\'s injury seeing as I am the one who talked him into going! oops!!!   Our frineds live on a hobby farm, so there was alot of room for the kids to run(of course it did not stop them from running through the vollyball court!) And good food, fellowship and beautiful weather. A perfect night~  (Well exept for Mike\'s injury, and the fact that I woke up the next morning with a sad relazationg I am getting old! Man was I sore!!) No LAUGHING now!  Well.....till next time~  '),(25851,2694370,'03,August,2004','2004-08-03 00:00:00','   It has been over a week since my last post?!?!?  What is going on?  I will skip ahead to the weekend, because not a whole lot happened last week. I will just say that my last post caught me in a bit of a bad mood. I worked through most of that just fine.  We were able to get out and have some fun and relaxation this weekend. Saturday I worked at the church for half the day, and that evening we went to the beach and had a picnic with some friends.  Sunday came along, and church was great.  We had a good potluck after the service as well. Now I am not necessarily a fan of potluck, but our church seems to put them on pretty much once a month.  I do enjoy the conversations over a meal with friends.  It is funny how potluck will vary from time to time as far as the quality of the food goes.  We have had some great ones, but then again, every once in a while, there is a really bad one.  I think the last potluck we had was one of those bad ones.  There was a whole lot of mushy hotdishes, and for some reason, they all pretty much tasted the same.  So, this one was much better by comparison. There was good variety, and the ingredients were either better or fresher or both.  Later Sunday evening, we went to our friends\' house to play volleyball. There were quite a few people there, and we had a blast!     I must say that I should have read the  urlLink comments on this volleyball post  before we started the third game.  I enjoy volleyball, and though I am not a pro, I like to think that I can play.  Put it this way, I can play pretty well with the people that I usually play with.    Well, we were pretty well into our third game.  Our team was way ahead at that point.  We were in the middle of a pretty good volley, when the ball happened to be set to the player playing opposite me at the net.  I had in my mind to make a great block, so I jumped into the air.  (of course, I missed the block...) I was just thinking to myself, \"how did I miss that?\" when I landed.  Then I heard (and felt) a *pop!*... *snap!* emanating from my left knee, and actually felt my knee bend sideways.  This is not something that is supposed to happen and trust me it feels very unnatural when it does.(it was downright PAINFUL!)  The first two thoughts through my head were \"Ouch!\", and \"I hope I don\'t need surgery!\" (in that order)...   So \"knee-dless\" to say (very puny!), that was the end of my volleyball for the night (and probably the rest of the summer...), but we did get 3 games in before hand and we got to stick around to roast marshmallows and have s\'mores.  So, all in all we had a great time!    I guess I am also learning to use crutches for the first time, too.  Not that I really wanted to, but hey, you have to make the most of it, right?  Now I just wish I could heal a bit faster...  '),(25852,2510334,'15,June,2004','2004-06-15 00:00:00','   It has been a couple of days, but how hot has it been?  For those of you not in England, the summer has indeed arrived. Though everybody is walking around saying â€˜nice weather, be a heat wave soonâ€™, mainly because between the last time it was this hot and now we have had all the bad weather under the sun for about 7 months.  But life must go on despite the heat, and this past few weeks I have been sitting my AS level exams Iâ€™ve had eight in total for four subjects, and my last exam was yesterday for economics. Now, normally (as has been the case for every other year of exams) I would have been revising heavily every night or so. Not this year, Iâ€™ve done in total about 6 hours worth, which sounds a lot but it isnâ€™t when you imagine me doing it at midnight or watching the TV. So I have no idea what my grades are going to be like but I fear I will disappoint myself and my family.  I have had a lot on as well as school. I recently got a new job in a clothing store called River Island and they have been giving me a lot more hours a week, then there is being with Martin which I love doing, and that takes up most of time and I wouldnâ€™t have it any other way. But the bad things are that I havenâ€™t been out for a drink months, I havenâ€™t had time to go swimming or to the beach or doing anything summery, Iâ€™m always busy or skint (poor).  So, it looks like Iâ€™m will be bringing in more money, I will just be too busy and too tired to spend it on anything worthwhile. All I want to do after work or at the weekend is to jump into a big cool pool, but I have to walk somewhere because getting the bus will be a waste of money, and the sun gets hotter, and I get more annoyed,  so the sun does indeed jest. Summer fun!  '),(25853,2510334,'11,June,2004','2004-06-11 00:00:00','   Just before Christmas last year I decided that it would be a brilliant idea to get an eyebrow piercing, my main drive being to get it done before my ex-boyfriend did. I happened to casually mention in to a friend whilst in town and she insisted that we go and do immediately. After some deliberation I agreed and off we went. I subjected myself to a certain degree of pain and then paid Â£24 for the priviledge.  Now advice for any person thinking of doing something similarIfff you take it out for more than 2 hours it WILL close over. I thought \'that\'s fine, it won\'t hurt to much to shove a piece of metal though my skin\'. Big mistake. One month later I ended up going to bed with a slightly swollen eye and woke up with a satsuma attached to my eye-socket. I wouldn\'t want to disgust you but after taking it out I realised what was actually in all the swelling.  So I made an appointment at my nearest surgery. \'I\'m sorry you are not resgistered here\' she says. \'That\'s right I moved about five minutes further from my old surgery and they kicked me out\' I say. \'well if you come down at 5 PM you can fill out some forms and register\'.  So I go to a history exam with one eye and then go straight to the doctor\'s and having not been ill that often was quite shocked but also amused at the things I saw. I followed the warren of corridors to the \'upstairs waiting area\' in the surgery built in a Victorian House and took my seat. Across from me was a young man about 22 falling asleep, to my right was an old man... falling asleep, completely unable to keep his head up. to my left was a woman with her head back, quite possibly asleep. Then across the room was a man that looked completely terrified and this was only made worse when his name was called. he bowed his head, either to pray or deliberate whether he was going to go in. I just sat there and thought that my condition wasn\'t really worthy of an appointment compared to these people. in fact the only other person who looked relativlely calm and alert was a guy about my age 16 or 17 with a nice pair of legs,so I decided to pass the time taking glances at them. Not at all harmful, i\'m still completely devoted to my boyfriend. I do remember sitting there and thinking he would probaly do the same thing in my position.  So that was about it they slowly got picked off one by one, and i eventually left with a course of antibiotics and the smell of TCP on my clothes.  So ot looks like all of got to concude is that doctor\'s surgerys are full of sick, tired people. a revelation I know!  '),(25854,2510334,'09,June,2004','2004-06-09 00:00:00','   I\'ve sat here at my computer since my boyfriend left (about an hour ago) and trawled through various blog sites. I discovered a few things;  1. You can actually get an award for blogging from the guardian newspaper (a v. posh paper read by nearly employee of the educational system in the UK)  urlLink Guardian Article  2. It can actually get quite expensive to blog up to $130 for a years worth of blogging 3. In my opinion most blogs will never ever be read because of the time needed to trawl blogs to find a subject interesting to you but you might try  urlLink BlogPulse  4. And perhaps the single most important thing I have found out is that there is no perfect formula for blogging and its best to keep it simple and write a load of crap off the top of your head like I\'m doing right now...  So, blogging, a valued tool to all Microsoft employees who dedicated minutes a day to enlightening their colleagues and captivated customers on, what I am sure to be, ground breaking developments.  '),(25855,2510334,'08,June,2004','2004-06-08 00:00:00','   I find myself becoming ever more tired, mainly due to the fact that I can\'t seem to drag myself away from the computer before 1 am every morning but also because of that weird anomaly that when someone begins to feel bored they start to fall asleep, ever worsened by this un-relenting heat, which would be fine if it hadn\'t been for the 6 moths of dismal British whether that we have just emerged from.  My weariness is set only to get worse as I have just agreed to work today,and most probably for the rest of this week, except that is for Thursday when I will be stuck in a lovely History exam, I can\'t even say that everything will be fine after I finished my last exam (Monday) but no, I have to go back to school and start learning next years work, then due to a severe lack of funds I will be working all summer. As is usually the case I will go back to school after six weeks more worn out than when I left.  adults think they have it hard, and tell us off for moaning about over-work and tiredness, but i don\'t see many of them going to school five days a week and then working week-ends as well, in fact I know some people including myself that work weekday evenings just to earn enough money to have a life. As my mums says \'you are old enough now to fend for yourself\'. If that is true why does she receive ÃƒÂ‚Ã‚Â£50 in benefits from the government for me and get nearly ÃƒÂ‚Ã‚Â£300 pounds in tax breaks for \'supporting a family\' wIeI i don\'t even live with her??  well must dash off to work, on a week-day, wheI i should be revising, but instead need to earn money to wheI i don\'t turn into some kind of hermit and never chance to see the world outside school, home and work!  '),(25856,2510334,'08,June,2004','2004-06-08 00:00:00','   Hi my name is Kyle if you haven\'t already seen my profile, and I live in a city called Plymouth in the South-west of England. At the moment I am just finishing my exams in my first year of sixthform. So that\'s the basic me, looks like a normal seventeen year old? I guess I am...  At the moment I am a bit of a lodger at my mum\'s friends house until the end of my final year of college, this being due to the fact that my mother decided to move 200 miles away to live with her boyfriend (a fat, controlling egotist, who I anyone to fond of). So along with my 15 year old sister I fend for myself in this fast paced and dangerous world... Well its not really that bad, I do get cable and broadband in my room and less of the nagging, but its the little things you miss, like the odd Ã‚Â£5 here and there. As for my dad, he moved toSpainn at thebeginningg of the year with his new girlfriend and they split up within 3 months.Thoughh he did decide to stick it outIi don\'t know how long he ill actually last.AlllIi do know is thatIi get a free holiday to sunnySpainn this august.  Other things in my life that are very important to me include my boyfriend, Martin; we have been together now for 8 months andIi know that to some people that 8 months doesn\'t sound like much we are very happy together andIi hope it will continue to last for a long time.  That\'s about me for the first blog andIi hopeIi can keep on up-dating you on me and my situation.  '),(25857,979795,'24,June,2004','2004-06-24 00:00:00','   One Hundred Fourteen:    The divine beauty Of heaven and earth! All creation, Members of One family.  '),(25858,979795,'23,June,2004','2004-06-23 00:00:00','   One Hundred Thirteen:    The Path is exceedingly vast.  From ancient times to the present day, even the greatest sages were unable to perceive and comprehend the entire truth; the explanation and teachings of masters and saints express only part of the whole.  It is not possible for anyone to speak of such things in their entirety. Just head for the light and heat, learn from the gods, and through the virtue of devoted practice of the Art of Peace, become one with the Divine.  '),(25859,979795,'23,June,2004','2004-06-23 00:00:00','   One Hundred Twelve:    The Art of Peace is the religion that is not a religion; it perfects and completes all religions.  '),(25860,979795,'23,June,2004','2004-06-23 00:00:00','   One Hundred Eleven:    When you bow deeply to the universe, it bows back; when you call out the name of God, it echoes inside you.  '),(25861,979795,'23,June,2004','2004-06-23 00:00:00','   One Hundred Ten:    The Divine does not like to be shut up in a building. The Divine likes to be out in the open. It is right here in this very body. Each one of us is a miniature universe, a living shrine.  '),(25862,979795,'23,June,2004','2004-06-23 00:00:00','   One Hundred Nine:    You cannot see or touch the Divine with your gross senses. The Divine is within you, not somewhere else. Unite yourself to the Divine, and you will be able to perceive gods wherever you are, but do not try to grasp or cling to them.  '),(25863,979795,'23,June,2004','2004-06-23 00:00:00','   One Hundred Eight:    Unite yourself to the cosmos, and the thought of transcendence will disappear. Transcendence belongs to the profane world. When all trace of transcendence vanishes, the true person - the Divine Being - is manifest. Empty yourself and let the Divine function.  '),(25864,979795,'23,June,2004','2004-06-23 00:00:00','   One Hundred Seven:    The Divine is not something high above us. It is in heaven, it is in earth, it is inside of us.  '),(25865,979795,'23,June,2004','2004-06-23 00:00:00','   One Hundred Six:    Rely on Peace To activate your Manifold powers; Pacify your environment And create a beautiful world.  '),(25866,979795,'23,June,2004','2004-06-23 00:00:00','   One Hundred Five:    Construction of shrine and temple buildings is not enough. Establish yourself as a living buddha image. We all should be transformed into goddesses of compassion or victorious buddhas.  '),(25867,979795,'23,June,2004','2004-06-23 00:00:00','   One Hundred Four:    The Art of Peace is a form of prayer that generates light and heat. Forget about your little self, detach yourself from objects, and you will radiate light and warmth. Light is wisdom; warmth is compassion.  '),(25868,979795,'23,June,2004','2004-06-23 00:00:00','   One Hundred Three:    The Art of Peace that I practice has room for each of the world\'s eight million gods, and I cooperate with them all.  The God of Peace is very great and enjoins all that is divine and enlightened in every land.  '),(25869,979795,'23,June,2004','2004-06-23 00:00:00','   One Hundred Two:    Ultimately, you must forget about technique. The further you progress, the fewer teachings there are. The Great Path is really No Path.  '),(25870,979795,'23,June,2004','2004-06-23 00:00:00','   One Hundred One:    The techniques of the Way of Peace change constantly; every encounter is unique, and the appropriate response should emerge naturally. Today\'s techniques will be different tomorrow.   Do not get caught up with the form and appearance of a challenge. The Art of Peace has no form - it is the study of the spirit.  '),(25871,979795,'23,June,2004','2004-06-23 00:00:00','   One Hundred:    If you comprehend The Art of Peace, This difficult path, Just as it is, Envelops the circle of heaven.  '),(25872,979795,'20,June,2004','2004-06-20 00:00:00','   Ninety Nine:    To truly implement the Art of Peace, you must be able to sport freely in the manifest, hidden, and divine realms.  '),(25873,979795,'20,June,2004','2004-06-20 00:00:00','   Ninety Eight:    Cast off limiting thoughts and return to true emptiness. Stand in the midst of the Great Void. This is the secret of the Way of a Warrior.  '),(25874,979795,'20,June,2004','2004-06-20 00:00:00','   Ninety Seven:    The Art of Peace can be summed up like this:  True victory is self-victory; let that day arrive quickly!  \"True victory\" means unflinching courage; \"self-victory\" symbolizes unflagging effort; and \"let that day arrive quickly\" represents the glorious moment of triumph in the here and now.  '),(25875,979795,'20,June,2004','2004-06-20 00:00:00','   Ninety Six:    If you perceive the true form of heaven and earth, you will be enlightened to your own true form.  If you are enlightened about a certain principle, you can put it into practice.  After each practical application, reflect on your efforts.   Progress continually like this.  '),(25876,979795,'20,June,2004','2004-06-20 00:00:00','   Ninety Five:    Fiddling with this And that technique Is of no avail. Simply act decisively Without reserve!  '),(25877,979795,'20,June,2004','2004-06-20 00:00:00','   Ninety Four:    Progress comes To those who Train and train; Reliance on secret techniques Will get you nowhere.  '),(25878,979795,'20,June,2004','2004-06-20 00:00:00','   Ninety Three:    In your training do not be in a hurry, for it takes a minimum of ten years to master the basics and advance to the first rung.  Never think of yourself as an all-knowing, perfected master; you must continue to train daily with your friends and students and progress together in the Art of Peace.  '),(25879,979795,'20,June,2004','2004-06-20 00:00:00','   Ninety Two:    Master the divine techniques Of the Art of Peace, And no enemy  Will dare to Challenge you.  '),(25880,979795,'20,June,2004','2004-06-20 00:00:00','   Ninety One:    The real Art of Peace is not to sacrifice a single one of your warriors to defeat an enemy.  Vanquish your foes by always keeping yourself in a safe and unassailable position; then no one will suffer any losses.  The Way of a Warrior, the Art of Politics, is to stop trouble before it starts. It consists in defeating your adversaries spiritually by making them realize the folly of their actions.   The Way of a Warrior is to establish harmony.  '),(25881,979795,'20,June,2004','2004-06-20 00:00:00','   Ninety:    Left and right, Avoid all Cuts and parries. Seize your opponents\' minds And scatter them all!  '),(25882,979795,'20,June,2004','2004-06-20 00:00:00','   Eighty Nine:    Even the most powerful human being has a limited sphere of strength. Draw him outside of that sphere and into your own, and his strength will dissipate.  '),(25883,979795,'20,June,2004','2004-06-20 00:00:00','   Eighty Eight:    Do not stare into the eyes of your opponent: he may mesmerize you. Do not fix your gaze on his sword: he may intimidate you. Do not focus on your opponent at all: he may absorb your energy. The essence of training is to bring your opponent completely into your sphere. Then you can stand where you like.  '),(25884,979795,'20,June,2004','2004-06-20 00:00:00','   Eighty Seven:    Opponents confront us continually, but actually there is no opponent there. Enter deeply into an attack and neutralize it as you draw that misdirected force into your own sphere.  '),(25885,979795,'20,June,2004','2004-06-20 00:00:00','   Eighty Six:    Your spirit is the true shield.  '),(25886,979795,'20,June,2004','2004-06-20 00:00:00','   Eighty Five:    When attacked, unify the upper, middle, and lower parts of your body. Enter, turn, and blend with your opponent, front and back, right and left.  '),(25887,979795,'20,June,2004','2004-06-20 00:00:00','   Eighty Four:    Seeing me before him, The enemy attacks, But by that time I am already standing Safely behind him.  '),(25888,979795,'20,June,2004','2004-06-20 00:00:00','   Eighty Three:    In the Art of Peace we never attack. An attack is proof that one is out of control. Never run away from any kind of challenge, but do not try to suppress or control an opponent unnaturally. Let attackers come any way they like and then blend with them. Never chase after opponents. Redirect each attack and get firmly behind it.  '),(25889,979795,'20,June,2004','2004-06-20 00:00:00','   Eighty Two:    There are two type of ki: ordinary ki and true ki. Ordinary ki is coarse and heavy; true ki is light and versatile.   In order to perform well, you have to liberate yourself from ordinary ki and permeate your organs with true ki. That is the basis of powerful technique.  '),(25890,979795,'20,June,2004','2004-06-20 00:00:00','   Eighty One:    In our techniques we enter completely into, blend totally with, and control firmly an attack.  Strength resides where one\'s ki is concentrated and stable; confusion and maliciousness arise when ki stagnates.  '),(25891,979795,'20,June,2004','2004-06-20 00:00:00','   Eighty:    One should be prepared to receive ninety-nine percent of an enemy\'s attack and stare death right in the face in order to illumine the Path.  '),(25892,979795,'20,June,2004','2004-06-20 00:00:00','   Seventy Nine:    The Art of Peace is to fulfill that which is lacking.  '),(25893,979795,'20,June,2004','2004-06-20 00:00:00','   Seventy Eight:    Even when called out By a single foe, Remain on guard, For you are always surrounded By a host of enemies.  '),(25894,979795,'20,June,2004','2004-06-20 00:00:00','   Seventy Seven:    At the instant A warrior Confronts a foe, All things Come into focus.  '),(25895,979795,'20,June,2004','2004-06-20 00:00:00','   Seventy Six:    In extreme situations, the entire universe becomes our foe; at such critical times, unity of mind and technique is essential - do not let your heart waver!  '),(25896,979795,'20,June,2004','2004-06-20 00:00:00','   Seventy Five:    Failure is the key to success; Each mistake teaches us something.  '),(25897,979795,'20,June,2004','2004-06-20 00:00:00','   Seventy Four:    Be grateful even for hardship, setbacks, and bad people. Dealing with such obstacles is an essential part of training in the Art of Peace.  '),(25898,979795,'20,June,2004','2004-06-20 00:00:00','   Seventy Three:    Life itself is always a trial. In training, you must test and polish yourself in order to face the great challenges of life.  Transcend the realm of life and death, and then you will be able to make your way calmly and safely through any crisis that confronts you.  '),(25899,979795,'20,June,2004','2004-06-20 00:00:00','   Seventy Two:    Protectors of the world And guardians of the Ways Of gods and buddhas, The techniques of Peace Enable us to meet every challenge.  '),(25900,979795,'20,June,2004','2004-06-20 00:00:00','   Seventy One:    Each day of human life contains joy and anger, pain and pleasure, darkness and light, growth and decay. Each moment is etched with nature\'s grand design - do not try to deny or oppose the cosmic order of things.  '),(25901,979795,'20,June,2004','2004-06-20 00:00:00','   Seventy:    Do not look upon this world with fear and loathing. Bravely face whatever the gods offer.  '),(25902,979795,'20,June,2004','2004-06-20 00:00:00','   Sixty Nine:    Free of weakness, No-mindedly ignore The sharp attacks Of your enemies: Step in and act!  '),(25903,979795,'20,June,2004','2004-06-20 00:00:00','   Sixty Eight:    If your heart is large enough to envelop your adversaries, you can see right through them and avoid their attacks. And once you envelop them, you will be able to guide them along the path indicated to you by heaven and earth.  '),(25904,979795,'20,June,2004','2004-06-20 00:00:00','   Sixty Seven:    Always try to be in communion with heaven and earth; then the world will appear in its true light. Self-conceit will vanish, and you can blend with any attack.  '),(25905,979795,'20,June,2004','2004-06-20 00:00:00','   Sixty Six:    The body should be triangular, the mind circular.  The triangle represents the generation of energy and is the most stable physical posture.  The circle symbolizes serenity and perfection, the source of unlimited techniques.  The square stands for solidity, the basis of applied control.  '),(25906,979795,'20,June,2004','2004-06-20 00:00:00','   Sixty Five:    When an opponent comes forward, move in and greet him; if he wants to pull back, send him on his way.  '),(25907,979795,'20,June,2004','2004-06-20 00:00:00','   Sixty Four:    Spring forth from the Great Earth; Billow like Great Waves; Stand like a tree, sit like a rock; Use One to strike All. Learn and forget!  '),(25908,979795,'20,June,2004','2004-06-20 00:00:00','   Sixty Three:    The techniques of the Art of Peace are neither fast nor slow, nor are they inside or outside. They transcend time and space.  '),(25909,979795,'20,June,2004','2004-06-20 00:00:00','   Sixty Two:    Manifest yang In your right hand, Balance it with The yin of your left, And guide your partner.  '),(25910,979795,'20,June,2004','2004-06-20 00:00:00','   Sixty One:    Functioning harmoniously together, right and left give birth to all techniques. The left hand takes hold of life and death; the right hand controls it. The four limbs of the body are the four pillars of heaven, and manifest the eight directions, yin and yang, inner and outer.  '),(25911,979795,'20,June,2004','2004-06-20 00:00:00','   Sixty:    If your opponent strikes with fire, counter with water, becoming completely fluid and free-flowing. Water, by its nature, never collides with or breaks against anything. On the contrary, it swallows up any attack harmlessly.  '),(25912,979795,'20,June,2004','2004-06-20 00:00:00','   Fifty Nine:    Techniques employ four qualities that reflect the nature of our world. Depending on the circumstance, you should be:  hard as a diamond, flexible as a willow, smooth-flowing like water, or as empty as space.  '),(25913,979795,'20,June,2004','2004-06-20 00:00:00','   Fifty Eight:    Move like a beam of light: Fly like lightning, Strike like thunder, Whirl in circles around A stable center.  '),(25914,979795,'20,June,2004','2004-06-20 00:00:00','   Fifty Seven:    The key to good technique is to keep your hands, feet, and hips straight and centered.  If you are centered, you can move freely.  The physical center is your belly; if your mind is set there as well, you are assured of victory in any endeavor.  '),(25915,979795,'20,June,2004','2004-06-20 00:00:00','   Fifty Six:    A good stance and posture reflect a proper state of mind.  '),(25916,979795,'20,June,2004','2004-06-20 00:00:00','   Fifty Five:    It is necessary to develop a strategy that utilizes all the physical conditions and elements that are directly at hand. The best strategy relies upon an unlimited set of responses.  '),(25917,979795,'20,June,2004','2004-06-20 00:00:00','   Fifty Four:    Always practice the Art of Peace in a vibrant and joyful manner.  '),(25918,979795,'20,June,2004','2004-06-20 00:00:00','   Fifty Three:    The totally awakened warrior can freely utilize all elements contained in heaven and earth.  The true warrior learns how to correctly perceive the activity of the universe and how to transform martial techniques into vehicles of purity, goodness, and beauty.  A warrior\'s mind and body must be permeated with enlightened wisdom and deep calm.  '),(25919,979795,'20,June,2004','2004-06-20 00:00:00','   Fifty Two:    To injure an opponent is to injure yourself. To control aggression without inflicting injury is the Art of Peace.  '),(25920,979795,'19,June,2004','2004-06-19 00:00:00','   Fifty One:    There are no contests in the Art of Peace. A true warrior is invincible because he or she contests with nothing. Defeat means to defeat the mind of contention that we harbor within.  '),(25921,979795,'19,June,2004','2004-06-19 00:00:00','   Fifty:    The Art of Peace is the principle of nonresistance. Because it is nonresistant, it is victorious from the beginning. Those with evil intentions or contentious thoughts are instantly vanquished. The Art of Peace is invincible because it contends with nothing.  '),(25922,979795,'19,June,2004','2004-06-19 00:00:00','   Forty Nine:    The heart of a human being is no different from the soul of heaven and earth. In your practice always keep in your thoughts the interaction of heaven and earth, water and fire, yin and yang.  '),(25923,979795,'19,June,2004','2004-06-19 00:00:00','   Forty Eight :   A true warrior is always armed with the three things: the radiant sword of pacification; the mirror of bravery, wisdom, and friendship; and the precious jewel of enlightenment.  '),(25924,979795,'19,June,2004','2004-06-19 00:00:00','   Forty Seven:    The Way of a Warrior is based on humanity, love, and sincerity; the heart of martial valor is true bravery, wisdom, love, and friendship.  Emphasis on the physical aspects of warrior-ship is futile, for the power of the body is always limited.  '),(25925,979795,'19,June,2004','2004-06-19 00:00:00','   Forty Six:    Instructors can impart only a fraction of the teaching. It is through your own devoted practice that the mysteries of the Art of Peace are brought to life.  '),(25926,979795,'19,June,2004','2004-06-19 00:00:00','   Forty Five:    From ancient times, Deep learning and valor Have been the two pillars of the Path: Through the virtue of training, Enlighten both body and soul.  '),(25927,979795,'19,June,2004','2004-06-19 00:00:00','   Forty Four:    Iron is full of impurities that weaken it; through forging, it becomes steel and is transformed into razor-sharp sword.  Human beings develop in the same fashion.  '),(25928,979795,'19,June,2004','2004-06-19 00:00:00','   Forty Three:    The purpose of training is to tighten up the slack, toughen the body, and polish the spirit.  '),(25929,979795,'19,June,2004','2004-06-19 00:00:00','   Forty Two:    The Way of a Warrior Cannot be encompassed By words or in letters: Grasp the essence And move on toward realization!  '),(25930,979795,'19,June,2004','2004-06-19 00:00:00','   Forty One:    Day after day Train your heart out, Refining your technique: Use the One to strike the Many! That is the discipline of the Warrior.  '),(25931,979795,'19,June,2004','2004-06-19 00:00:00','   Forty:    Those who are enlightened never stop forging themselves. The realizations of such masters cannot be expressed well in words or by theories. The most perfect actions echo the patterns found in nature.  '),(25932,979795,'19,June,2004','2004-06-19 00:00:00','   Thirty Nine:    Practice the Art of Peace sincerely, and evil thoughts and deeds will naturally disappear. The only desire that should remain is the thirst for more and more training in the Way.  '),(25933,979795,'19,June,2004','2004-06-19 00:00:00','   Thirty Eight:    Daily training in the Art of Peace allows you inner divinity to shine brighter and brighter.  Do not concern yourself with the right and wrong of others.  Do not be calculating or act unnaturally.  Keep your mind set on the Art of Peace, and do not criticize other teachers or traditions.  The Art of Peace never restrains, restricts, or shackles anything.  It embraces all and purifies everything.  '),(25934,979795,'19,June,2004','2004-06-19 00:00:00','   Thirty Seven:    Even though our path is completely different from the warrior arts of the past, it is not necessary to abandon totally the old ways. Absorb venerable traditions into this Art by clothing them with fresh garments, and build on the classic styles to create better forms.  '),(25935,979795,'19,June,2004','2004-06-19 00:00:00','   Thirty Six:    Your mind should be in harmony with the functioning of the universe; your body should be in tune with the movement of the universe; body and mind should be bound as one, unified with the activity of the universe.  '),(25936,979795,'19,June,2004','2004-06-19 00:00:00','   Thirty Five:    The Path of Peace is exceedingly vast, reflecting the grand design of the hidden and manifest worlds. A warrior is a living shrine of the divine, one who serves that grand purpose.  '),(25937,979795,'19,June,2004','2004-06-19 00:00:00','   Thirty Four:    Foster and polish The warrior spirit While serving in the world; Illuminate the path According to your inner light.  '),(25938,979795,'19,June,2004','2004-06-19 00:00:00','   Thirty Three:    The Art of Peace does not rely on weapons or brute force to succeed; instead we put ourselves in tune with the universe, maintain peace in our own realms, nurture life, and prevent death and destruction. The true meaning of the term samurai is one who serves and adheres to the power of love.  '),(25939,979795,'19,June,2004','2004-06-19 00:00:00','   Thirty Two:    Economy is the basis of society. When the economy is stable, society develops. The ideal economy combines the spiritual and the material, and the best commodities to trade in are sincerity and love.  '),(25940,979795,'19,June,2004','2004-06-19 00:00:00','   Thirty One:    Loyalty and devotion lead to bravery. Bravery leads to the spirit of self-sacrifice. The spirit of self-sacrifice creates trust in the power of love.  '),(25941,979795,'19,June,2004','2004-06-19 00:00:00','   Thirty:    Each and every master, regardless of the era or place, heard the call and attained harmony with heaven and earth.  There are many paths leading to the top of Mount Fuji, but there is only one summit - love.  '),(25942,979795,'18,June,2004','2004-06-18 00:00:00','   Twenty Nine:    All life is a manifestation of the spirit, the manifestation of love. And the Art of Peace is the purest form of that principle. A warrior is charged with bringing a halt to all contention and strife. Universal love functions in many forms; each manifestation should be allowed free expression.  The Art of Peace is true democracy.  '),(25943,979795,'18,June,2004','2004-06-18 00:00:00','   Twenty Eight:    The only cure for materialism is the cleansing of the six senses (eyes, ears, nose, tongue, body, and mind).  If the senses are clogged, one\'s perception is stifled. The more it is stifled, the more contaminated the senses become. This creates disorder in the world, and that is the greatest evil of all.  Polish the heart, free the six senses and let them function without obstruction, and your entire body and soul will glow.  '),(25944,979795,'18,June,2004','2004-06-18 00:00:00','   Twenty Seven:    The essence of the Art of Peace is to cleanse yourself of maliciousness, to get in tune with your environment, and to clear your path of all obstacles and barriers.  '),(25945,979795,'18,June,2004','2004-06-18 00:00:00','   Twenty Six:    The Art of Peace is based on Four Great Virtues: Bravery, Wisdom, Love, and Friendship, symbolized by Fire, Heaven, Earth, and Water.  '),(25946,979795,'18,June,2004','2004-06-18 00:00:00','   Twenty Five:    To practice properly the Art of Peace, you must:  Calm the spirit and return to the source.  Cleanse the body and spirit by removing all malice, selfishness, and desire.  Be ever-grateful for the gifts received from the universe, your family, Mother Nature, and your fellow human beings.  '),(25947,979795,'18,June,2004','2004-06-18 00:00:00','   Twenty Four:    Crystal clear, Sharp and bright, The sacred sword Allows no opening For evil to roost.  '),(25948,979795,'18,June,2004','2004-06-18 00:00:00','   Twenty Three:    The Art of Peace is not easy. It is a fight to the finish, the slaying of evil desires and all falsehood within. On occasion the Voice of Peace resounds like thunder, jolting human beings out of their stupor.  '),(25949,979795,'18,June,2004','2004-06-18 00:00:00','   Twenty Two:    The penetrating brilliance of swords Wielded by followers of the Way Strikes at the evil enemy Lurking deep within Their own souls and bodies  '),(25950,979795,'18,June,2004','2004-06-18 00:00:00','   Twenty One:    As soon as you concern yourself with the \"good\" and \"bad\" of your fellows, you create an opening in your heart for maliciousness to enter.  Testing, competing with, and criticizing others weaken and defeat you.  '),(25951,979795,'18,June,2004','2004-06-18 00:00:00','   Twenty:    Always keep your mind as bright and clear as the vast sky, the great ocean, and the highest peak, empty of all thoughts. Always keep your body filled with light and heat. Fill yourself with the power of wisdom and enlightenment.  '),(25952,979795,'18,June,2004','2004-06-18 00:00:00','   Nineteen:    Study the teachings of the pine tree, the bamboo, and the plum blossom. The pine is evergreen, firmly rooted, and venerable. The bamboo is strong, resilient, unbreakable. The plum blossom is hardy, fragrant, and elegant.  '),(25953,979795,'18,June,2004','2004-06-18 00:00:00','   Eighteen:    Your heart is full of fertile seeds, waiting to sprout. Just as a lotus flower springs from the mire to bloom splendidly, the interaction of the cosmic breath causes the flower of the spirit to bloom and bear fruit in this world.  '),(25954,979795,'18,June,2004','2004-06-18 00:00:00','   Seventeen:    Peace originates with the flow of things - its heart is like the movement of the wind and waves.  The Way is like the veins that circulate blood through our bodies, following the natural flow of the life force.  If you are separated in the slightest from that divine essence, you are far off the path.  '),(25955,979795,'18,June,2004','2004-06-18 00:00:00','   Sixteen:    Do not fail To learn from The pure voice of an Ever-flowing mountain stream Splashing over the rocks.  '),(25956,979795,'18,June,2004','2004-06-18 00:00:00','   Fifteen:    Create each day anew by clothing yourself with heaven and earth, bathing yourself with wisdom and love, and placing yourself in the heart of Mother Nature.  '),(25957,979795,'25,June,2004','2004-06-25 00:00:00','   \"Surrounded by a forest of enemy spears - enter deeply and learn to use your mind as a shield.\"   -- O-Sensei  '),(25958,979795,'18,June,2004','2004-06-18 00:00:00','   \"A host of enemies rush into attack - think of them as a single foe and deal with them accordingly.\"   -- O-Sensei  '),(25959,979795,'18,June,2004','2004-06-18 00:00:00','   \"... training is training in the path of human life.\"   -- O-Sensei  '),(25960,979795,'18,June,2004','2004-06-18 00:00:00','   \"The subtle working of ki is the maternal source that affects delicate changes of breath.\"   -- O-Sensei  '),(25961,979795,'18,June,2004','2004-06-18 00:00:00','   \"... It is not a question of speed. The fight is finished before it is begun.\"   -- O-Sensei  '),(25962,979795,'18,June,2004','2004-06-18 00:00:00','   \"True breathing means to breathe in unison with the universe. Then one gains the power of nature.\"   -- O-Sensei  '),(25963,979795,'18,June,2004','2004-06-18 00:00:00','   \"Seeing me before him the enemy attacks, but by that time I am already standing behind him.\"   -- O-Sensei  '),(25964,979795,'18,June,2004','2004-06-18 00:00:00','   \"In Aikido we control the opponent\'s mind before we face him... drawing him into ourselves... going forward in life with this attraction of our spirit, and attempting to command a whole view of the world.\"   -- O-Sensei  '),(25965,979795,'18,June,2004','2004-06-18 00:00:00','   \"Aiki is the power of harmony... Of all things, all things working together.\"   -- O-Sensei  '),(25966,979795,'18,June,2004','2004-06-18 00:00:00','   \"Reviewing a lesson a hundred times cannot be compared to reviewing it a hundred and one times.\"   -- Rabbi Hillel  '),(25967,979795,'18,June,2004','2004-06-18 00:00:00','   \"Take the first step in faith. You don\'t have to see the whole staircase, just take the first step.\"   -- Martin Luther King, Jr  '),(25968,979795,'18,June,2004','2004-06-18 00:00:00','   \"Be not afraid of growing slowly, be afraid only of standing still.\"   -- a Chinese proverb  '),(25969,979795,'18,June,2004','2004-06-18 00:00:00','   \"Practice means to perform, over and over again in the face of all obstacles, some act of vision, of faith, of desire. Practice is a means of inviting the perfection desired.\"   -- Martha Graham  '),(25970,979795,'18,June,2004','2004-06-18 00:00:00','   \"You gain strength, courage, and confidence by every experience in which you really stop to look fear in the face. You must do the thing which you think you cannot do.\"   -- Eleanor Roosevelt  '),(25971,979795,'18,June,2004','2004-06-18 00:00:00','   \"Progress comes to those who train and train; reliance on secret techniques will get you nowhere.\"   -- O-Sensei  '),(25972,979795,'18,June,2004','2004-06-18 00:00:00','   \"Where the spirit does not work with the hand there is no art.\"   -- Leonardo Da Vinci  '),(25973,979795,'18,June,2004','2004-06-18 00:00:00','   \"Who overcomes by force, hath overcome but half his foe.\"   -- John Milton  '),(25974,979795,'18,June,2004','2004-06-18 00:00:00','   \"When shall we finally realize that to need nothing is best? Not even good luck is necessary! All we really need is to return to our original simplicity. This is the open way for us all.\"   -- Hogen Yamahata, from his book On the Open Way Â© 1991 Jiko Oasis Books  '),(25975,979795,'18,June,2004','2004-06-18 00:00:00','   \"We practice sitting meditation, but for what? For some specific reason? What we need to do is just sit in this cosmic Here-Now fully. Sitting Zen is alive, yet peaceful; it does not need to go anywhere.\"   -- Unknown  '),(25976,979795,'18,June,2004','2004-06-18 00:00:00','   \"It is easier, it seems, to run into death as \'spiritual soldiers\', than it is to free ourselves from our \'stony-headed\' convictions.\"   -- Hogen Yamahata, from his book On the Open Way Â© 1991 Jiko Oasis Books  '),(25977,979795,'18,June,2004','2004-06-18 00:00:00','   \"We know nothing about our fate; we are almost blind to our destinations(s). What is motivating us? Our decisions? Our choices? Something else? We are always trying to see the relationship between decisions, choices and the \'fate\' of all beings.\"   -- Hogen Yamahata, from his book On the Open Way Â© 1991 Jiko Oasis Books  '),(25978,979795,'18,June,2004','2004-06-18 00:00:00','   \"It is ultimately possible to become a Bodhisattva; it is equally possible for one to become an instrument of devilish ego.\"   -- Hogen Yamahata, from his book On the Open Way Â© 1991 Jiko Oasis Books  '),(25979,979795,'18,June,2004','2004-06-18 00:00:00','   The Master said, \"If you do not become irritated about why you cannot understand it, I will not teach you.  When you cannot easily express it though you can inwardly understand it, if you are not fretful I will not guide you.  When I explain one angle to you, if you do not answer the other three using your brain I will not repeat my lesson.\"   -- Yoshita Furukawa, the Analects of Confucius excerpted in Roots of the Japanese Mind Â© 1968 Tokyo, Japan  '),(25980,979795,'18,June,2004','2004-06-18 00:00:00','   \"That\'s why O-Sensei said that Aikido is ichi go ichi e. You must do the technique as if for the first time and the last time.\"   -- Christian Tisser, 7th dan, Shihan  '),(25981,979795,'18,June,2004','2004-06-18 00:00:00','   \"Techniques are tools. They are the mechanical structures for our transformation.\"   -- Christian Tisser, 7th dan, Shihan  '),(25982,979795,'18,June,2004','2004-06-18 00:00:00','   \"The real voyage of discovery consists not in seeking new landscapes, but in having new eyes.\"   -- Marcel Proust  '),(25983,979795,'18,June,2004','2004-06-18 00:00:00','   \"In Aikido, we always want to make the ideal movement. Somebody comes with a very good, earnest attack, and you try to make the ideal movement. Of course, it doesn\'t work. So you do it again -- again trying to make the ideal movement. You fail again. In Aikido, we want to make the most beautiful technique, and we fail. Still, the purpose is to make the most perfect technique. That\'s what we have to remember but we forget it very easily.\"  \"There was a painter named Hans Menling who never signed his paintings. He just wrote, \'It\'s my best.\' I attended a Jesuit College, and every day we had to write \'It\'s my best\' in our notebooks. I think that Aikido is the same: you try to do your best -- not the best, of course, but your best.\"   -- Christian Tisser, 7th dan, Shihan  '),(25984,979795,'18,June,2004','2004-06-18 00:00:00','   \"My advice is to be patient. You don\'t learn Aikido overnight. If you want something out of Aikido, you have to put something into it.\"   -- Donovan Waite, 6th dan, Aikido of Center City  '),(25985,979795,'18,June,2004','2004-06-18 00:00:00','   \"Aikido, he says, is research into the ideal of the perfection of self.\"   -- Christian Tisser, 7th dan, Shihan, quoting Yamaguchi Sensei  '),(25986,979795,'18,June,2004','2004-06-18 00:00:00','   \"You gotta charge them something for Aikido,\" a trainer at a Marine base told me. But gurus in India were beggars. They didn\'t sell their knowledge, because they viewed it as priceless.   -- Ralph Glanstein, 6th dan  '),(25987,979795,'18,June,2004','2004-06-18 00:00:00','   \"In the beginners\' mind there are many possibilities, In the experts\' there are few.\"   -- Unknown  '),(25988,979795,'18,June,2004','2004-06-18 00:00:00','   \"Your soul - your life energy - is not bound by flesh. It can reach where your fingers cannot.\"   -- Unknown  '),(25989,979795,'18,June,2004','2004-06-18 00:00:00','   \"O-Sensei is my only Aikido Teacher. When I have questions, I go back to him in my memory. In that way, I am still learning from him - almost 50 years later.\"   -- Kazuaki Tanahashi  '),(25990,979795,'18,June,2004','2004-06-18 00:00:00','   \"It is the sound of purest harmony, the sound of the universe.  Make your soul sing its song, and you will find there is nothing you cannot accomplish.\"   -- Unknown  '),(25991,979795,'18,June,2004','2004-06-18 00:00:00','   \"Listen carefully to what I will say now.  For the lesson is not always carried in the words of men.  You see, many believe that the nature of the universe is to hide itself from mortal eyes.  That we must search and meditate upon all around us to see and grasp the simple truth that is enlightenment.  However, perhaps the answers are being told to us all the time and we have just forgotten how to hear them.\"   -- Unknown  '),(25992,979795,'18,June,2004','2004-06-18 00:00:00','   \"A superstar is like a roman candle, they shine bright.  A teacher is like a lantern, they may not shine as bright, but they illuminate the path.\"   -- Unknown  '),(25993,979795,'18,June,2004','2004-06-18 00:00:00','   \"Art reaches its greatest peak when devoid of self-consciousness.  Freedom discovers man the moment he loses concern over what impression he is making or about to make.\"   -- Bruce Lee, Tao of Jeet Kune Do  '),(25994,979795,'18,June,2004','2004-06-18 00:00:00','   Those who are first on the battlefield and await the opponents are at ease; those who are last on the battlefield and head into a fight become exhausted.  Therefore, good warriors cause others to go to them and do not go to others.\"   -- Sun Tzu, The Art of War  '),(25995,979795,'18,June,2004','2004-06-18 00:00:00','   \"The river has no shape, but it takes on the boundaries which it carves out for itself, so is the mind boundless, until it creates a prison for its own thoughts.\"   -- Unknown  '),(25996,979795,'18,June,2004','2004-06-18 00:00:00','   \"Here is a test to see if your mission on earth is finished: If you\'re alive, it isn\'t.\"   -- Unknown  '),(25997,979795,'18,June,2004','2004-06-18 00:00:00','   \"The bird does not care how it flies, and Oh! how it flies!\"   -- Unknown  '),(25998,979795,'18,June,2004','2004-06-18 00:00:00','   \"Patience and persistence can bring down the tallest trees.\"   -- Unknown  '),(25999,979795,'18,June,2004','2004-06-18 00:00:00','   \"He who speaks in anger makes his anger heard, but his words forgotten.\"   -- Unknown  '),(26000,979795,'18,June,2004','2004-06-18 00:00:00','   \"Whenever I move, that\'s Aikido.\"   -- O-Sensei  '),(26001,979795,'18,June,2004','2004-06-18 00:00:00','   \"In true budo there are no opponents, no enemies. In true budo we seek to be one with all things, to return to the heart of creation.\"   -- O-Sensei  '),(26002,979795,'18,June,2004','2004-06-18 00:00:00','   \"Aikido is based on Love, yes it is based on Harmony, yes it is based on Oneness, and yes it is based on not having an enemy. Even if somebody comes to try to cut my throat right now, in my opinion, they\'re not necessary my enemy, because I have what we call mushin. When somebody attacks me I don\'t think about, I\'m going to get him, I\'m going to do this. It\'s just a state of mind, a heart without emotion, a mind without emotion that I have, and the only thing I\'d do is react to terminating the situation to the best of my ability.\"   -- Steven Seagal, 7th dan, Shihan, Shinto priest  '),(26003,979795,'18,June,2004','2004-06-18 00:00:00','   \"In Aikido it doesn\'t matter who is better. It doesn\'t matter who\'s right and who\'s wrong, or who has how many students or whose dad is bigger than whose. Who cares? None of it matters, it has nothing to do with Aikido. What matters is that we all try to help each other to improve ourselves as human beings. Whatever styles come to us are welcome, nobody is better than anybody. Concentrate on the philosophical and the spiritual aspects of Aikido rather than who\'s affiliated with who.\"   -- Steven Seagal, 7th dan, Shihan, Shinto priest  '),(26004,979795,'18,June,2004','2004-06-18 00:00:00','   \"Budo means to stop war. Stop arms, stop war. So Budoka is a heihoka, a warrior is really a warrior of peace, or a man of peace. You have to be powerful enough to stop war, because if you\'re weak you can\'t stop war. And Budo has that yin and yang, it has that Tate to yoko no ito, izu no mitama to mizu no mitama. There are all Shinto terms. Yoko no ito means moon, feminine, water, love, the power of forgiveness, the power of love. Tate no ito means sun, we talk about masculine, we talk about fire, we talk about the power of decision. That is the time when you don\'t forgive, that is the power to cut.\"   -- Steven Seagal, 7th dan, Shihan, Shinto priest  '),(26005,979795,'18,June,2004','2004-06-18 00:00:00','   \"When I go from dojo to dojo, sometimes for example; I see people attacking each other and then I see the uke falling before he\'s even been thrown, falling before he\'s even touched. How could this possibly have anything to do with Aikido?\"  \"O-Sensei always said \"Aikido wa budo de aru\". And what I felt he really meant was, \"This is a martial art and if it doesn\'t work, take up aerobics, take up dance, or take up a gun! But don\'t call it the Martial Arts. Aikido is a Martial Art.\"    -- Steven Seagal, 7th dan, Shihan, Shinto priest  '),(26006,979795,'18,June,2004','2004-06-18 00:00:00','   \"O-Sensei was known to always have said, \"Ai te no chikara awasu\", and what this means, loosely translated, is that you have to practice at the ability of your partner, and it always has to be the stronger and more adept person that has to adjust to the less adept, for obvious reasons.\"   -- Steven Seagal, 7th dan, Shihan, Shinto priest  '),(26007,979795,'18,June,2004','2004-06-18 00:00:00','   \"The most important thing to beginning Aikido study is developing spiritual awareness, the perfecting of the physical man, the developing and nurturing of the emotional and mental and psychological state of man, the nurturing and developing and helping the physical state of man, the medical state of man, a lot of people have been cured by this so it helps you become a healthy person physically. It helps you become healthy and well-balanced mentally, psychologically, emotionally and it helps you develop and perfect your spiritual awareness.\"   -- Steven Seagal, 7th dan, Shihan, Shinto priest  '),(26008,979795,'18,June,2004','2004-06-18 00:00:00','   \"It is all your level of ability.\"   -- Steven Seagal, 7th dan, Shihan, Shinto priest  '),(26009,979795,'18,June,2004','2004-06-18 00:00:00','   \"Once somebody is a Sensei, it is proper protocol to refer to them as Sensei, whether it is on the mats or off the mats, they are always Sensei. However, protocol also says that if you do not respect that person as Sensei, never call him Sensei, not once not ever. \"   -- Steven Seagal, 7th dan, Shihan, Shinto priest  '),(26010,979795,'18,June,2004','2004-06-18 00:00:00','   \"Learn to feel what\'s real.\"   -- Steven Seagal, 7th dan, Shihan, Shinto priest  '),(26011,979795,'18,June,2004','2004-06-18 00:00:00','   \"O-Sensei said that the basics should never change. O-Sensei founded Aikido, he invented Aikido. There are no styles of Aikido. There is only one Aikido. And that is O-Sensei\'s Aikido because he invented it.\"   -- Steven Seagal, 7th dan, Shihan, Shinto priest  '),(26012,979795,'18,June,2004','2004-06-18 00:00:00','   \"The benefits that one should reap in the martial arts have nothing to do with the color of a belt, they have to do with developing spiritual awareness, developing mental and emotional calm and strength and developing physical health and things like that.\"   -- Steven Seagal, 7th dan, Shihan, Shinto priest  '),(26013,979795,'18,June,2004','2004-06-18 00:00:00','   \"Some people will learn certain things in six months, and other people will take six years. There\'s no way that this can be any other level other than an individual level. It is very personal.\"   -- Steven Seagal, 7th dan, Shihan, Shinto priest  '),(26014,979795,'18,June,2004','2004-06-18 00:00:00','   \"If you don\'t demand a real situation, a real atmosphere in your dojo, where people are able to go hard, and they do go hard, then your practice becomes, in my opinion ineffective.\"   -- Steven Seagal, 7th dan, Shihan, Shinto priest  '),(26015,979795,'18,June,2004','2004-06-18 00:00:00','   \"When you\'re able to practice it at a high enough level, you can\'t get into something that is more. No holds barred. In other words, you don\'t have to restrain yourself from going into the things that are more dangerous. It\'s real in the sense that no matter how much a beginner you are on the mats, if you do the wrong thing, you\'re going to get hurt.\"   -- Steven Seagal, 7th dan, Shihan, Shinto priest  '),(26016,979795,'18,June,2004','2004-06-18 00:00:00','   \"No, try not. Do, or do not. There is no try.\"   -- Yoda, in Starwars Episode V: the Empire Strikes Back  '),(26017,979795,'18,June,2004','2004-06-18 00:00:00','   \"Daniel-san, Karate here,\" points to the head, \"Karate here,\" points to the heart, \"Karate never here.\" and motions to the belt.   -- Pat Morita as Mr Miyagi, in the Karate Kid  '),(26018,979795,'18,June,2004','2004-06-18 00:00:00','   \"With Aikido, you have to unlearn first. That is why it is so hard. You are learning something that is very, very sophisticated. It is something that Warriors after hundreds and hundreds of years of dedicated endeavors, of life and death situations, have finally figured out that there is a point of least resistance, and that there is something that works beyond strength...physical strength. It is \'Ki\', its attitude and its adept movements that seem to go beyond force, and that a lot of times using strength is completely secondary to learning a way. A kind of moving without resistance, a non-resistant movement, a relaxed movement. So you have to unlearn in order to learn. It does take longer.\"   -- Steven Seagal, 7th dan, Shihan, Shinto priest  '),(26019,979795,'18,June,2004','2004-06-18 00:00:00','   When students asked to repeat movements of a technique, he replied  \"Each and every technique is unique. If I do it again, it will naturally be a little different. Rather than trying to grasp how I move, pay attention to what this old fellow is saying to you - that will enable you to execute the techniques properly.\"  When students protested his talks were difficult to understand, he replied  \"What I tell you may now seem to be way over your heads, but eventually there will be a time when it all becomes clear.\"   -- O-Sensei  '),(26020,979795,'18,June,2004','2004-06-18 00:00:00','   Question: \"How long will it take me to get a Black Belt?\"  Answer: About 10 minutes. Because \"... there\'s a martial arts store nearby and you can pick one up for about $6.00. On the other hand, if you want to learn Aikido ... \"   -- George Simcox, Ki Society  '),(26021,979795,'18,June,2004','2004-06-18 00:00:00','   \"... the important thing is to practice and more practice, but with reflection and common sense. The rest is illusion.\"   -- Pablo Vazquez, 6th dan, Shihan, Aikido Institute of West Los Angeles  '),(26022,979795,'18,June,2004','2004-06-18 00:00:00','   \"Attention or awareness is the secret of life, and the heart of practice...\"   -- Charlotte Joko Becki, ZenCenter of San Diego  '),(26023,979795,'18,June,2004','2004-06-18 00:00:00','   \"By experiencing, I mean that first movement when we receive life before the mind arises.\"   -- Charlotte Joko Beck, ZenCenter of San Diego  '),(26024,979795,'18,June,2004','2004-06-18 00:00:00','   \"Don\'t think that you can repeat the same movement twice. There is no such thing as repetition, just as today is not a repetition of what yesterday was, and tomorrow will not be a repetition of what today has been.\"   -- T.K. Chiba, 8th dan, Shihan, San Diego Aikikai, Chair USAF Western Region Teaching Committee  '),(26025,979795,'18,June,2004','2004-06-18 00:00:00','   \".. form is the finger that points at the moon we never see. Even when we think we see the moon it is only the reflection of light from the sun reflecting off the moon.\"   -- (late) Nobuo Iseri, 6th dan, Shihan, Chief Instructor, the Moving Center  '),(26026,979795,'17,June,2003','2003-06-17 00:00:00','   Conclusion  Given some hiccups during the whole trip itself, I still really loved the experience of trekking in Norway. I enjoyed the experience of going on a tour with Daniel and Justin. Most often they took the lead on the planning, suggested places of interests, etc.  Learned a little something planning and budgeting my finances. I had to ensure that I bring MORE THAN ENOUGH money whenever I\'m on a trip.  Safety was essential. Especially safety with respects to your own luggage was not to be taken for granted. Regardless of how   suagu   it\'d make us look, there\'s simply no compromise because as tourists, we stand out too easily.  Hope someday, I can go on another trip like this with the guys.  '),(26027,979795,'16,June,2003','2003-06-16 00:00:00','   After 0600 hrs  Needed to wake early today. Marcus is going to give me a lift down to the train station to see me off. I found out online that there seems to be a train around 0700 hrs. I managed to get a discounted ticket at 250 SEK (around Â£25) to Stockholm. In total, I spent less than Daniel or Justin did for their tickets down to Stockholm yesterday. I think itâ€™s because I needed to make 2 changes and part of the ride was on the older trains instead of the new X2000 series.   1241 hrs  Arrived at Stockholm. At the Tourist Information counter, I had to pay 20 SEK for a map (not FOC leh). By the time I left, I had wasted 20 mins just to queue up. It took me about 1 hr to orientate and get myself to the hostel that Justin had told me about. Naturally, given my â€˜fantasticâ€™ sense of orientation, I did make some wrong turns here and there. The hostel is really unique. It is a ship, permanently secured to the side of a harbour / waterfront. Its reception is located in a building opposite the ship. In addition to accommodations, the building also houses the common kitchen and a lounge. I was staying in a separate dorm from the guys. But it was pretty comfortable. Point to note, they sold disposable bed covers which cost cheaper than renting them. Most other hostels weâ€™ve visited only provided rental sheets. The disposables are a good idea for travelling in Scandinavia. I shared my dorm with a family of about 9. It looked to me like the grandparents down to the grandchildren and an uncle or family friend were here on holiday. Felt a little awkward being the only Chinese in the dorm. But they were really nice and friendly folks.  After checking in, we went on to explore Stockholm. Again we decided to go for a canal tour of the city. However, this time it was a pre-recorded message. The whole thing lasted about 2 hrs, which was really a little too long. Naturally I fell asleep again. This time, it was because of the music. It kept coming on, then after awhile the message came on, then the slow music, then the message, and so on. I think I also have some kind of thing about sleeping on ferries. I feel that they just put me to sleep sometimes. Could also be because I woke up really early this morning and I havenâ€™t gotten much sleep on the train rides because of the short span of time in between the changes.  There was a part of the city called the Old Town. This was an impressive place. The guys had visited the area yesterday and they took me to a sci-fi store. It was really cool. After that, we headed back to the hostel and prepared pasta for dinner again. After dinner, we explored the little island that the hostel was built on and then returned to hang around the TV room. Here we called back to check up on Lynette and Annie. I noticed 2 Welsh English playing pool, befriended them and played a few games with them. After awhile, a group of Scottish teens joined us. 1 of the latter was an Australian who is actually staying in Scotland now. Daniel decided to do some bouldering on a rock wall behind the hostel and Justin left together with him. Daniel looked pretty impressive as he climbed. I told myself that if my results were alright, I will try to visit the Rockface back in Birmingham to try to do some climbing one last time before I went back to Singapore.   2330 hrs  Returned to my dorm, grabbed a shower and got ready for bed. Everyone else in the dorm was already asleep. I had forgotten that there were kids in the dorm too and they probably needed to go to bed early. Tried to move around as quietly as I could.  '),(26028,979795,'15,June,2003','2003-06-15 00:00:00','   0800+ hrs  Got up and washed up. We all went straight to the rail ticketing office to change our rail tickets. I got mine changed from Copenhagen-Stockholm to Copenhagen-Karlskrona (both via Malmo). Karlskrona is the place where my Poly classmate, Marcus, is staying. The RSN had sent him to Sweden for submarine training. Ivy, his wife, accompanied him and she had given birth to baby Zelda sometime in Nov 2002. Iâ€™m really quite ecstatic when I think about seeing all of them soon. They are the first close friends I know whoâ€™ve already started a family of their own. Iâ€™m scheduled to arrive at Karlskrona at 1421 hrs.   1100 hrs  We all departed Copenhagen for Malmo, Sweden. Apparently we were misinformed and should have boarded a slightly earlier train. This caused us to miss our connecting rides to our destinations. Now that Iâ€™ve had time to think things through, I think I mightâ€™ve overlooked the timing for the earlier train. Anyway, when we arrived at Malmo, I tried to change my ticket for another one (the guys are going straight to Stockholm so their tickets arenâ€™t affected). I was initially told that to pay for a new ticket. But I explained my situation and the ticket attendant decided to change a new ticket for me FOC. Cool! However, my arrival time has now been delayed by a good 2 hrs. I bought a local phone card and gave Marcus a ring to inform him of the delay. It was a pity as he had planned to take me to the Maritime/Marine Museum on my arrival. It seems that I will arrive too late to visit the sights of the small town.   Around 1200 hrs  The guys boarded their train for Stockholm and I still have some time to spare. So I grabbed a map of Malmo and decided to explore outside the station. Naturally I didnâ€™t dare to venture too far and kept one eye on my watch constantly. The train station is located at the edge of the city itself. Malmo looks really nice and peaceful. Of course, being a Sunday, that may be the reason.   1429 hrs  Train departed Malmo for Karlskrona. I managed to arrive around 1641 hrs. The moment I stepped onto the platform, I could see Marcus standing a short distance away. He was in a typical Singaporean attire of T-shirt, Bermuda and slippers. Of course, his short GI haircut was hard to miss too. As I was walking over to him, who should come along but Ivy carrying baby Zelda. It was sweet seeing this family of three. Marcus drove us to a supermarket nearby. After the 5 min drive, they asked if I had bought my ticket from Karlskrona to Stockholm for tomorrow. Alamak! He drove me back to the station but the ticketing office had already closed for about 15 min (4 min after my arrival). Iâ€™ve been really careless and forgetful since this morning. Sigh, canâ€™t be helped now. We headed back again to the supermarket, bought some stuff and proceeded back to the family apartment. The flat was nice, cosy and had a spacey living / dining room, a kitchen, a toilet and a bedroom.  They prepared steamboat for dinner while Zelda took her nap. Almost towards the end of the meal, she woke up and Ivy had to prepare her baby food. The couple switched after awhile. I was cute seeing the three of them; Marcus, a happily married father, Ivy, a busy housewife and a 7 mth old Zelda. We had a nice long chat the throughout and after dinner. I played with Zelda a little but I seriously didnâ€™t feel too comfortable carrying her. Iâ€™m not sure how tight I had to hold her and I was really afraid of dropping or hurting her somehow. Zelda has Ivyâ€™s nose and Marcusâ€™s eyes and mouth. This was one time that I actually had to be referred to as Uncle Aylwin. Haha. The baby proved to be quite a handful for my friends sometimes but they do seem to be managing very well. Just before midnight, we all decided to go to sleep. I was really quite sad that the day had to end so soon. Had things been perfect, Iâ€™d have an extra 2 hrs to spend with my friends, not that it wouldâ€™ve been enough anyway.  '),(26029,979795,'14,June,2003','2003-06-14 00:00:00','   Before 0800 hrs  Woke up. Breakfast consists of cereal, different types of bread, butter and coffee/tea. Weather seems as though it\'ll rain soon. We\'re just waiting for the sky to get a little better before we go out. Later we\'ll be shifting to a more centrally located accommodation. However, it looks like itâ€™s may be in a red light district of sorts. It also seems like there is a huge population of Asians in the area. We do have some events lined up for the day but everyone seem to be in a world of our own since Justinâ€™s bag was stolen. I have no idea how long this mood will last but itâ€™s getting boring. I hope we can find something to lighten the tension soon.  Visited Amelieborg Palace (Royal Palace) for the changing of the guards. I finished up my current roll of film here. It was interesting but not too fantastic. Justin noted that one of the guards dropped his magazine on the ground during one of the drills. Cham liao, the chap sure kanna extra guard duty. After this, we went for a guided canal tour. It was really good, but Justin and I fell asleep. For me it was because of the rocking motion on the waters, the nice sun and the soothing breeze that made me fall asleep. The tour isnâ€™t boring and I did find it quite informative. Next, we visited the National Museum. It was pretty interesting. But the problem was that we were all too tired to enjoy it much.  In the evening, while deciding on what we are having for dinner, the atmosphere turned a little nasty. We met a group of Singaporeans and Daniel took a bit of time to talk to them â€“ heâ€™s a friendly and sincere guy who takes time to talk to people, even those we donâ€™t know very well. On the other hand, Justin is rarely that patient with strangers over such niceties. Heâ€™s not an unfriendly person, itâ€™s just that he places purpose, speed and efficiency over such protocols. I feel pretty uneasy when I noted the change of tones. But everyone knows that tempers were just running a bit high at that moment. The mood was defused as quickly as it started. Whew.  We all had an enjoyable pizza buffet (but the drinks we had actually cost us more than the meal). After dinner, we went back to our accommodation and rested. I called Marcus about the details of visit and found out that we had a cheaper alternative. What we needed to do is to try to change our tickets first thing next morning. We had an enjoyable discussion about things in general and as it sometimes happens, a (friendly and interesting) debate ensued. I think we only went to bed sometime before 0100 hrs.  '),(26030,979795,'13,June,2003','2003-06-13 00:00:00','   0700 hrs  Justin woke us up to get ourselves ready. He heard some kind of random announcement about presenting your boarding ticket on arrival or something like that. Justin thought that it meant that we had arrived early in Copenhagen. When we were all ready, we went to the cafeteria with all our barang barang (luggage). That\'s when we realised we were really an hour and a half away from arrival. Alamak, we couldâ€™ve gotten an hour\'s more sleep. But it wasn\'t exactly a bad thing. Being up early, we decided to use up all our Norwegian currency and coins on breakfast.   0900+ hrs  Arrived in Copenhagen. My first impression is that it seemed livelier and nicer than Burgen. First thing we had to do is to get to the train station to purchase tickets for travel to Stockholm 2 days later. My next impression, things were still as expensive as before. The tickets cost us 673 DKK (about Â£67+). I made an initial estimate of my transport costs within Scandinavia so far and it seemed to have already hit about Â£230. I think it\'s even possible to travel by flying at this price. This obviously only shows how poorly estimated our transport costs were. But since we are already here ... sigh.  At the ticketing office, after we had paid for the tickets, we immediately realised that Justin had lost his bag. He had placed his bag in front of his feet (between both feet and the counter), a relatively secure position I would say. Naturally, it\'s been stolen. We searched around and even went to the station\'s toilet to see if any bugger was hanging around there rummaging through a familiar looking bag but all to no avail. We suspect that it was as we were all leaning forward to talk to the counter staff and checking details that the thief made his move. Even so, Justin had only left his bag down for a few mins. The lucky thing was that he had all his documents (passport) and cards (credit and debit) on him when the bag was stolen. He did lose his new Â£200 credit card sized Casio digital camera (with MP3 playing capability) and about Â£30 worth of cash though. The really inconvenient part was that all his clothes and his glasses were in the bag as well. Naturally, Justin felt really sian about the whole thing. We went to the Police station located within the rail station and made a police report.  We still had to grab some chow. We came across a buffet serving Indian/Mediterranean food of sorts and decided to have a try. It wasn\'t bad at all, but there wasn\'t a great variety. After dinner, we went on to the famous Tivoli Gardens. I originally expected Tivoli Gardens to be some sort of botanical garden or park. Imagine my surprised when I realised it was really an amusement park. It wasn\'t particularly expensive either considering that we\'re in a country of such a high standard of living. The place was bustling with people and activity. Featured rides and all that weren\'t like state of the art or brand new but the atmosphere sure was one of the best in the world. It definitely lightened our mood a bit given the incident of the day.  Allow me to go off-track for a while but at that moment I was wondering how come Sentosa can\'t enjoy the same kind of atmosphere or popularity as the one I had seen in Tivoli Gardens? Justin pointed a few possibilities to me. Most events held at Sentosa seemed more catered for the teens and young adults. The locality of Sentosa was a bit too much to the south, but considering Singapore\'s small size so that\'s probably more of an excuse for people not to visit Sentosa than a good reason. But Tivoli Gardens is situated in the heart of the capital itself. It closes at 0100 hrs but in Singapore, only certain special locations do so and they are more like pubs, discos, etc.  We are quite lucky as there will be a free concert tonight. We don\'t recognise the singer (she was a Randi something) but her vocals sounded pretty good. Music was nice too. As Justin isn\'t in a perfect mood, he wasn\'t particularly interested in the concert. We left Daniel at the concert while we went to the arcade/jackpot. He did pretty ok and won some credits but he had no idea how to retrieve the tokens from the machine. Ultimately he used up all his credits instead. I did have a go on a swash-buckling samurai game. It\'s the first time I tried and it was quite fun. I tried to apply some simple things I learnt from weapons work but after awhile, I realised it was pointless. There was so many loophole in the game compared to the real thing. It\'s still a game after all I suppose. O yea, Daniel, the ever-ready-rock-climber had a go on a machine which simulated a rock wall. A crowd always gathers whenever someone has a go on the machine. It just seemed interesting lah. Danish girls here have really pretty and outstanding features. Lots of blond around (dunno if they\'re natural blond or what). Anyway, they really stood out.  '),(26031,979795,'12,June,2003','2003-06-12 00:00:00','   0600+ hrs  Woke up, washed up, packed, and had a simple breakfast of bread, butter and sausages.   0720+ hrs  Departed the YMCA. Walked to the train station and boarded the train back to Oslo.   Before 0800 hrs  Train pulled out of the station. The journey to Oslo will take us about 6+ hrs. We\'re scheduled to arrive around 1420 hrs or so. The train was a new one. Reasonably comfortable with a fair amount of leg space. I spent about one third of the journey sleeping. The rest of the trip I just read my book.  In Oslo, we have just less than 2 hrs before we need to board the ferry bound for Copenhagen. We\'re probably going to look for an Internet cafe and a bookstore (Daniel\'s still hoping to get a book for the rest of the trip).   1600+ hrs  Arrived and boarded the Crown of Scandinavia. Cast off timing is 1700 hrs. It\'s an interesting experience. This is the first time I\'ve been on a ferry / cruise like this. Dinner isn\'t included in the price of the ticket though. But we\'re still planning to go for the buffet dinner in the ship\'s restaurant later. It\'s my turn to treat the guys. I just can\'t wait to explore the rest of the ship. So exciting. In the meantime, think we\'re all gonna grab a nap first.   2000 hrs  Finished dinner. It costs 195 DKK each but it was really good. There were seafood, meat courses, cold dishes and desserts. I only managed the first 3 sections. The dessert section I had to leave it out as things didn\'t look too tempting to me. I ended my dinner with a green apple instead. It\'s been a great trip so far making our way to Copenhagen in this fashion. In fact, I hope someday in future, I can do the same again with my parents or family on something similar.  I felt that the waiter serving us wasn\'t really comfortable serving us. It\'s either because we seemed like some young punks or it may be because we\'re all Chinese or both. He wasn\'t like particularly nasty to us or anything, it\'s just that he didn\'t seem friendly enough as a service personnel. He just gave me the impression that his level of hospitality was not like the other crewmembers we have come across earlier on the ship. It was only when the bill came that he seemed to be friendlier, but only slightly (\"I hope dinner was OK.\"). I dunno lah, like I said, he wasn\'t exactly mean to us.  After dinner, we walked and explored the ship. Getting out onto the open deck was a bit scary and exhilarating at the same time. When I first pushed the door open, the breeze from outside was so strong that it nearly pulled me outside. Stepping out into the open, it took me awhile to get used to the strong breeze and the wobbly feeling. You can feel the seas-spray almost everywhere. You\'d realise it if you lick your lips and you\'ll find that it tastes salty. After awhile, I managed to find my sea legs. It just felt damn shiok to be out on the ship\'s deck with the strong breeze all around. It made me feel as though I was flying. The guys kept jokingly asked me to sign on with the Navy when I got back to Singapore.   2343 hrs  Justin and I just finished a film we caught in the ship\'s cinema. We each paid 40 DKK for it. We watched the Recruiter (starring Al Pacino and Colin Farrell). It was quite OK. Daniel caught another film instead as he had seen the Recruiter before. We\'re probably going to grab an early night and just ready ourselves for tomorrow.  '),(26032,979795,'11,June,2003','2003-06-11 00:00:00','   Was horrible. Some bugger\'s alarm clock rang and woke the whole room up except for himself. I felt really pissed. To top things off, someone spilled water on the floor last night. The water flowed towards my bag (which was on the floor) and kinda soaked it a bit. But it was wet enough to get my jeans and some shirts inside wet as well. I was extremely irritated when I left the YMCA.  Because Burgen is known for it\'s seafood market found along the waterfront in the early morning, we walked across to grab breakfast there. The guys were really willing to spend money on seafood. I went to a convenience store for a hotdog and a cold drink. I\'ll probably try the seafood for lunch instead.  Walked around aimlessly and it was really boring. What an ordeal it was just passing 2 hrs. Luckily we found a bookstore. The first branch we visited had very little English books but the guys were still able to hang around for quite awhile. I\'m not a very great fan of literary works so most of the English books displayed simply didn\'t appeal to me. Across the street, there was a bigger branch, which was much better. I bought a book on Viking sayings. It\'s a bit pricey for it\'s thickness I guess it\'s but OK lah, bring a bit of Norway back with me lor. So far, I\'ve yet to see anything about Burgen I really like. It\'s so similar to Birmingham that I feel that I should could\'ve remained back in Brum to see the same things. It\'s almost 1200 now and I think we\'ll grab lunch in awhile. All we\'re doing now is just sitting in the square and listening to a brass band playing nearby. They are pretty entertaining though.   1740 hrs  Having a terrible day so far. We still haven\'t done much except sort\'a whiling time away sitting here and there and not really doing anything. I think I\'m kinda holding the guys back because I haven\'t got as much cash to play around and visit places and stuff. I also can\'t eat as much as I\'d like. They probably find this cultural city with all its museums and all that more fascinating then I do. It wouldn\'t be a bad idea for them to go ahead without me but they probably didn\'t feel it was a good idea to just leave me while they went enjoying themselves. I\'m grateful but I still don\'t like the idea of preventing them from enjoying themselves. Another part of me isn\'t exactly that enthusiastic about visiting museums and all those stuff either, but if I had the kind of spare cash I probably wouldn\'t mind it. Sigh, I\'m not comfortable with incurring too much extra expenses and bad debts lah. Not to forget that I still owe them a meal in Copenhagen. Daniel has already treated us one. One thing though, after we returned to the YMCA, I did catch a really good nap. Hopefully that changed my mood a little. I\'ll go look for the guys in the lounge now, maybe we can do something after dinner ... I hope.   2100 hrs  Went up to the peak of Floibanen. Sort of Burgen\'s version of Singapore\'s Mt Faber and/or Bt Timah Hill I suppose. There are a few tracks here and we took a quick walk here and there. Daniel was finishing a book he just bought this afternoon so he tried to get another one in the evening, but to no avail. Most bookstores were closed already.  Tomorrow is another full day of travelling. We\'ll be taking a coach/train back to Oslo to catch a ferry/cruise to Copenhagen, Denmark. I\'m really looking forward to leaving Burgen and going on to Denmark.  '),(26033,979795,'11,June,2003','2003-06-11 00:00:00','   0800+ hrs  Had a longer sleep today. Laundry was still damp. We hung them out on the clothesline provided outside. But only for about an hour or so as we had to pack up and check out soon. We took the same breakfast we had 2 days ago of bake beans and bread and went back to bed to sleep a little more.1030 hrsGot out of bed, packed up and checked out of the hostel. Purchased some tokens at the Internet cafe nearby and cleared our emails while waiting for the ferry which will be taking us to Gudvagen. We whiled time away sitting along the harbour waiting for the ferry.   1430 hrs  Boarded our ferry and departed Flem for Gudvagen.   1700 hrs  Arrived at Gudvagen and boarded a coach bound for Voss.   Before 1900 hrs  Arrived at Voss and got on the 1915 hrs train to Burgen.  Basically, our whole day was spent travelling and getting to Burgen.  It\'s raining / drizzling when we arrived in Burgen. Horrible. Checked into the local YMCA here. It\'s newly renovated but everything seems half done. They were still in the process of renovating the place and you can still smell the new plaster everywhere. We grabbed an expensive dinner at a \'cheap\' restaurant. That\'s not the worst. The toilet in the restaurant was coin-operated, same as the public toilets in the damn country. I can understand a public toilet being coin-operated but to do the same to the restaurant\'s toilet is simply outrageous.  The shower rooms were horrible. Not a proper changing room was done yet. The way the doors opened, I think someone walking past the male showers can look in easily. Can\'t imagine what\'s it like for the ladies shower room. I had a hard time showering. I\'d classify the shower and changing rooms to be simply pathetic. The dorm was similar to military style bunks. They were double-decked. It took me a long time to fall asleep. Some bugger was snoring loudly nearby. My first impression Burgen has been dramatically damaged.  '),(26034,979795,'10,June,2003','2003-06-10 00:00:00','   0800 hrs  Woke up and had a buffet breakfast. It was similar to what we had in Oslo at the Ankar Hostel. All of us took a lot to keep our strength up for the journey back the way we came yesterday. We had the nicest hardboiled eggs. The egg white was solid but the egg yolk was only half or three quarter cooked.   1045 hrs  Departed Osterbo. We moved a lot faster and we made good time, possibly because we didn\'t take the steeper route this time. It also felt like there were more down slopes on the way back. Justin led and I was trailing behind with Daniel between us. Halfway through, Daniel took the lead. Being the last man, I lagged behind quite often. A few times, the guys had to pause for me to catch up with them. That says a lot about my level of fitness. I guess there\'s a lot of training I\'ll have to do when I get home. Might have to do some thigh strengthening exercises too.   1500 hrs  Ended much earlier than intended. We decided to hang around a stream where we washed up (no skinny dipping this time). We immersed our blistered feet into the ice-cold water and it felt really good. We took some photos here as well. I was quite happy that my legs felt stronger and hadn\'t tightened or cramped up today.  The cab came 15 mins earlier than before. We returned to Flem and lodged back the same hostel. We were given different rooms though. This time it was a room for 3 instead of a dorm. Dinner was a simple meal of pasta with eggs. Justin was trying to achieve the same standard of the eggs we had in Osterbo. He didn\'t quite make it for this first try. We had early dinner and an early shower. We even had time to wash our clothes in the laundry.   2045 hrs  Feeling a little bored â€˜coz we haven\'t got anything to do. Justin popped out to get some chips for snacks. At this point we got into some random chat and read up on some pamphlets, which we picked up here and there.  '),(26035,979795,'09,June,2003','2003-06-09 00:00:00','   0800 hrs  Woke up. Made and ate our breakfast of bake beans and bread. Cab was 15 mins late. We took it to Vassbygdi, our start point.   0845 hrs  Start of trek!! It was simple path, leading past some houses and farms.     urlLink    Photo of Daniel and Justin at the starting point.  urlLink      urlLink    Another photo at the starting point. Behind is actually a sort of rest station. This was to be the spot which marked the start and end of our trek.  urlLink'),(26036,979795,'07,June,2003','2003-06-07 00:00:00','   0800 hrs  Woke up late. Intended to wake at 0700 hrs but Daniel had forgotten to adjust his mobile phone\'s time to Norwegian time. We washed up and changed, chop chop everything was done in 30-40 min (bear in mind we only had 1 toilet to share) and we went for breakfast. The buffet breakfast was in the cafe of Ankar Hotel, which was next to Ankar Hostel. The buffet was OK, but not fantastic. There was a small variety of sliced meat. We stunned some butter, sugar and salt, all essential for cooking in the days to come.   1000 hrs  Finally checked out of the hostel feeling filled up and happy. We took our own sweet time to walk to the train station. I tried calling home from here but I just can\'t. Don\'t understand why they don\'t seem to take ICC. Weird. Tested my overdraft by withdrawing more money. Success! I recall applying for an overdraft 2 years ago. Still seems to be valid.   1046 hrs  Norwegian rail system are so meticulous. They depart on the dot. Now we\'re on our way to Myrsdal.   1100 hrs  Reaching the edge of Oslo city. I can see the clear blue sky with less than 30% of cloud cover. The train ride (with the expensive ticket) is reputed for the scenic views it\'ll be taking us through. It was said that we should stay awake for the duration of our train ride to take in the sights. Daniel has already knocked out. He hasn\'t caught much sleep the last few nights in UK, preparing for his personal stuff and the coming Italy trip with his parents.      urlLink     Picture of a waterfall. This was one of the stops. That\'s Justin\'s head in front of me when I snapped my photo.  urlLink'),(26037,979795,'06,June,2003','2003-06-06 00:00:00','   Time stated in GMT British time   0000 hrs  Departed home for Digbeth coach station. Clement, a final year Singaporean Law student, gave us a lift in his car. Heard he was really tired from driving people around the whole day. The lift we got was quite uncomfortable with a lot of bumps and knocks here and there (but he was doing us a favour after all, so...). I did get a bit pissed off as he went over a particular deep hump and I knocked my head against the ceiling of the car. Seriously, if I knew that a \'free ride\' meant having to endure such a thing, I\'d gladly pay for a cab ride down instead. I\'m not one to accept \'free lunches\' but don\'t expect me to have to see your lian3 se4 (çœ‹è„¸è‰²) and take such shit too easily. I kept looking at my watch and reminding myself that getting out now and flagging a cab down won\'t get me to the destination on time so I just kept reaaaaally quiet about it. Like I said, he was under no obligation to give us a lift and he did it out of goodwill, but I wished he hadn\'t tried to be a nice guy and had just gone about his normal business that night.   0100 to 0550 hrs  Took a coach down to Stansted airport in London, one of the smaller airports in the outskirts of London. Changed some Norwegian (NOK), Danish (DKK) and Swedish (SEK) currency. Rate was approximately Â£1 to 10 NOK / 12 DKK / 10 SEK. Around there lah.   0720 hrs  Boarded the EasyJet plane bound for Torp airport in Norway.   Time stated in Norwegian time +1   0945 hrs  Landed in Torp airport. My first impression of Norway: lots of low buildings and trees, clean air. Paid 120 NOK for a coach service to Oslo city where our hostel is located. Withdrew more cash from the ATM. Second impression of Norway: expensive country with a high standard of living. As we got closer to the city, taller buildings started to appear. Plenty of modern architecture dominated the Norwegian city landscape.   1250 hrs  Arrived Oslo city. We checked into the Ankar Hostel situated a few minutes walk from the city centre. It was a nice place. New too. We seemed to be the first group to be using our room as the hostel had just opened that day very day.  We approached a mountaineering shop for maps to our intended trail. Found out that we were too early in the season and around two third of the trail was still covered with snow. We were told the earliest possible season for trekking should be around mid / end July depending on the weather. This was in great contrast to what the Lonely Planet book said. Sianz. We had to revise our original plans again and re-plan our route. In the end we decided to go for the remaining part of the trek where the snowed has already cleared (supposedly the best part as well).  We made our way to the rail ticketing office to get tickets to take us closer to the starting point of our trek. Purchased a train ticket called \"Norway in a Nutshell\". Cost us 1657 NOK! Almost Â£165+. Have to try and call home to get more funds. Friends are paying for some of my expenses at the moment. What makes things worse is that tomorrow is a weekend. My parents won\'t be able to do anything till next week.  At the end of the day, I noticed that there\'s about 1 hr more of daylight compared to UK.  A coincidence. The 4th person in our dorm is a 5th year medic (medical student) also studying in the University of Birmingham. He had ended term early and had been travelling Europe for the past 7 weeks. We had a conversation with him till about 2am before we all went to bed. He mentioned a lot about his travels in Rome, Munich, etc.  '),(26038,979795,'28,November,2002','2002-11-28 00:00:00','    Wah piang eh!  Juz now after lecture was looking around for a toilet and I didn\'t realised I walked right into the ladies!!!! Argh!!! So embarassing, when I came out a door across the corridor opened and a lady came out. She muz\'ve been really shocked to see me exit the ladies. But I really wasn\'t that bothered. Maybe that\'s wat they mean by having a clear conscience, huh. Still, thinking about the incident is sorta funny. Hmmm ... my high point this term. Hahaha.  Lastnight\'s Aikido training was an ultimate low for me. Sigh, felt sooooooo tired and I just don\'t feel that I\'m concentrating properly. Actually though of leaving after the first hour with the beginners. But I just though, if I gave up so fast and easily when will I ever improve. So I stayed on for the 2nd hour. But still didn\'t feel like I\'m doing anything right. Hope next week will be a lot better.   '),(26039,979795,'27,November,2002','2002-11-27 00:00:00','    Haiz, tonight never went for Aikido training. Pity lor. Felt really lazy. So many things just delayed me from getting ready and psyching myself up to go. Argh, really need to get ready earlier next time. I feel so bad / guilty for not going when I could have. I really don\'t know how this kind of attitude of mine is gonna help my Aikido development in future. So sad. Need to put in more effort. And the distance to the dojo doesn\'t really help either. It\'s not near like in Uni. Even though Uni has a free bus ride down, just traveling down to Uni and waiting for the (long wait) bus is a real put off as well. But I know that once I get myself down to the dojo I won\'t regret the training and I\'d be glad to have put in the effort. Sometimes I seem to treat my Aikido training like a convenience. Seems like I go down only when I can go down easily. Shouldn\'t be the case. I think from next week onwards, I\'m gonna bring my Gi around sch for the day so that I can leave from Uni direct. And I gotta liase with a senior to go down together. That shall be my level of commitment. Dat\'s what I\'ll do!!   '),(26040,979795,'26,November,2002','2002-11-26 00:00:00','    Alamak, Uni wants me to pay school fees liao. Wah lao! £9000+ pounds, almost 10k. This country really suck students dry on money. It\'s no wonder people, especially students, will try all means to maximise and even cheat the Uni of some money.  Very sian. Just got an email from my laopeh. Kaopeh at me about the Uni thingy, coz if pay late gonna get fined for it. A lot of things tell him liao he also kaopeh. That\'s why really freaking hate to talk to him. Cannot have a decent conversation without disagreeing. Shits. Ah, parents lah, what to do. Sometimes I miss Singapore, but everytime I go back and I think I may want to try to be a good son, some shit will happen at home and my parents will make me wish I\'m back in UK alone and away from them. Confirm plus guarantee one. These few years I really don\'t feel like I can hang around them for long. I don\'t really feel the point for joining reunion dinner and all that stuff. Don\'t understand lah. I think army/camp life suits me lor. When I kanna hold back by the govt/army parents usually can\'t do anything. Woohoo.  O yea, talk about army life. I think I will sign on with the Police Force after I graduate. I\'ve got a module doing something called Commercial Programming and the school brings in industry people to talk to us about issues out there right now. The module is very very good. Anyway, one area was Computer Forensics. Really damn interesting. Don\'t think anyone really knew that our data can still be retrieved even after we do a format. The expert said that you had to format the harddisk 7 times and do physical wipe of the system to have a high chance of ensuring that data cannot be retrieved. I dunno if I can do Computer Forensics lah, but I wanna do something to serve the public. Almost definitely government work. Especially with the Sep 11 incident, think it strenghtened my convictions that we need to be alert and stron internally. It was really scary to know that JI had planned possible attacks on certain locations in Singapore. If intelligence and news didn\'t reach us in time, the outcome would have been catastrophobic. I can\'t begin to imagine the number of innocent people who might have been killed and injured. Especially after watching war movies like Band of Brothers, Black Hawk Down and Saving Pte Ryan, imagine some civilian who might have lost a limb and just be bleeding on the ground waiting for life to end and being fully conscious. Ewww ... really very cham. If I\'m the one there, I might wish for someone to put a bullet through my head immediately.  Anyway, gotta go library liao. Got to return a short loan video now. Otherwise fine fine fine. Hey, who said Singapore is a \'fine\' country. It\'s the same over here and the amount is a lot more hefty compared to Singapore.   '),(26041,979795,'25,November,2002','2002-11-25 00:00:00','    Today was just surfing around and I went to the websites of every Aikido club in Singapore.  Aikikai (Singapore) has a new website. Well it\'s new lah following a split [which formed the Aikido Federation (Singapore)]. Anyway a lot of political shitz inside that which I dunno, so ... no comments. Was just to Aikido ShinJu-Kai\'s website too. Now I realised that the 2 of these clubs now have the \'history\' section. Heh. The latter\'s used to be very very brief, while the former had absolutely none.  I dun mean to criticise lah of coz, but I noticed there\'s this sudden outburst of \'Singapore\'s Aikido History\' available online. I\'m happy to have put up Aikido in Singapore\'s history online (at http://www.geocities.com/Aikido_in_Singapore ... aiya muz promote my site mah hehheh) before ShinJu-Kai\'s new history updates and Aikikai\'s new site. Hope that helped to kinda poke more people to put the info online and made readily available to the general public. I\'m sure that there must be a lot of people who have already wanted to do this. Also realised everyone have a slight version of their own \'truth\'. Doesn\'t matter to me. Think when all the info is put out, the public themselves can read and then formulate their own opinions and ideas on the subject.  Personally I have a few views about the various club\'s sites I think. I analysed them through the following, overall look/design, content, navigation.  Aikido Fed\'s is simple to use. Nice comfortable look. But maybe because it\'s still new, it has a somewhat empty look. Should get better with time. Use of white and dark blue color is good and relaxing to the eyes. Doesn\'t take long to load. Better use of fonts may enhance the site further. Might help to include some history, but being a new club (and maybe coz of the unpleasantness from the split with Aikikai Singapore) think the lack of info now is coz they\'re trying not to offen anyone. I\'d think a 3.5/5 is fair. Content probably only a 3/5 but it\'s simple to use design pulled it to a 3.5.  Aikikai\'s site looks \'stiff\' (??). Dunno how to describe also. But I think it\'s also because it\'s still a new site. Weird color combination. Looks like they are trying hard to \'structure\' their content at the moment. Heard that they were working on something \'good\' few mths ago, but it looks crappy. Think the NUS aikido site is still better. Might have helped if they worked on it from there. But the info is still OK for a new site. Should get better with time. 3/5 grade maybe? It\'s design deserves a 2.5/5 though but the content kinda pulled it up a bit.  Ueshiba\'s can be better. They are quite a laojiao club but still very poorly established online. Maybe that\'s not their emphasis that\'s why. They\'re probably more focused on training and other forms of expansion instead. But I figure that in ths modern day and age, the development of a proper website is extremely important. They have a link with Ngee Ann poly\'s Aikido site. Both sites basically show the same stuff. Quite pointless to maintain 2 pages like that. May be better to crosslink some info pages to allow 1 update only. But this may slow down surf time for one of the site. Alternatively, they may want to use some kinda of FTP update facility to make auto updates/publishing. Either way, for a club that \'old\' the site is quite crappy. Their content is OK. A history section would be nice though not really necessary. But I think George Chang Sensei runs the club very niely. Dunno, juz my impression of the club. 3/5.   Yushinkai is a new site as well, looks comfortable enough but can still be improved. Fonts can be improved. 3.5/5 coz of the nice feel it gives. Doesn\'t look messy.  Ki-Aikido\'s has a nice site, but they look a bit too cramped. Content is very very good. Only site which included video clips. Thumbs up on this point. Structure of the site content seems ok but feels a bit weird. Not sure why. But content more than makes up for it. Including history of Ki Aikido (being the only Ki sch) would be very very nice. 4/5.  ShinJu-Kai has by far the best site established online. Anyone looking for Aikido in Singapore will definitely like their site more. They are extremely well organised. Content is very good. 5/5 score.  Hope I dun offend anyone with these frank opinions. Hey, I\'m no expert but I think even layman have got opinions. My site isn\'t that fantastic either. I\'m so envious of Aikido clubs with impressive looking sites. Dun think I can ever reach that stage for my site. Think for an Aikido site to excel, including history and background is necessary. Also need to project a professional image and that\'s why how the site is designed is very impt. Of coz, Aikido is not about website building lah. But website helps the spread of the art.   '),(26042,979795,'19,November,2002','2002-11-19 00:00:00','    OK OK This will be my first official blog post here. Hmm what to say neh? Sorry if my posts not very structured. Just speaking from what I can rmember and my memories do jump around a bit sometimes.  Let me just start with a brief introduction of myself lah. Alternatively, if anyone out there is even remotely interested, you can visit my webpage at http://www.geocities.com/aylwintan1978.  OK, I don\'t want to, like, go really that far back in history (you can find that out from my website) so I\'ll just talk about myself now. Was from Nanyang Primary Sch, went to Buona Vista Sec and Computer Studies in Ngee Ann Poly. NS, was a M113 driver (no lah it\'s not a freaking tank lah) with 56th SAB, then disrupted to go UK for my degree.  Haiz, talk about my degree really need to be ashamed lah. I got direct entry in 2nd year of BSc Computer Science / Software Engineering. Didn\'t do well (that\'s an understatement, more like did really badly), so I had to take a year out to clear my resits while doing an internship in Singapore. I only just came back to UK 2 mths ago for the final year of my degree. Hope can still get my honours, even if it\'s a lousy one.  In poly I was very much into working out and physical training. Nope, I\'m not one of those tanned and really fit guys you see out on the streets lah. I into working out for a tiring feel to stretch my muscles at a (turtle) slow and comfortable pace. My exercise \'kaki\'s in poly generally consisted of Ronny and Gary. We\'d go for a run, then go to the gym first before jumping into the pool. Shiok, the pool ... my favourite exercise, swimming. Usually target to swim 1 to 2 km. No lah, not difficult lah. At a slow pace you can pretty much complete the distance.  Anyway, when NS came. You pretty much dun do anything but run and run. Plus there wasn\'t really much of an opportunity to to swim anymore. So there I was, panting and dying on the road. But I learnt to appreciate the feeling I get from running. I\'m still a pretty lousy runner (aiya, can only get Silver for 2.4km in IPPT mah), but I happily volunteered for the Army Half Marathon in 2000. Really cham. Coz I overstrained one of my knees during training. Still, I didn\'t give up when it came to the actual day. I think it was either 12th or 16th km when I felt like my knee really hurt. Still half limped and half jogged my way to the 21st km finish line. I didn\'t meet my personal target for my timing though. But still, it was a very shiok feeling knowing I completed 21 km.  Anywayz, when I went into Uni. Went through a lot of stuff. Knowing more about people, more about myself. You discover a bit of your hidden strengths and weaknesses. You make some friends and you make some enemies. You become more receptive, more adaptable, yet still resistant to some kind of changes. You fit in and socialise, trying to widen your circle, only to find that you can also be a bit anti-social, and don\'t need anymore than a small circle of close friends. Yea, everything listed contradicts the other. But that\'s how it is. I suppose everything comes down to a right kind of balance between the 2 sides. Anyway, like I said, I screwed up my 1st yr of study (which was really the 2nd yr of the course) in UK. Did helluva lot of resits in Singapore and then proceeded back to UK for my final yr. I also met my current girlfriend in UK. Of coz lah, we didn\'t really get together till like last summer holidays.  We\'ve got our own share of differences and arguements but so far things have been working out. At least I think so lah. Dunno what more I can comment on my relationship at this point.  Oooo ooo, now comes my favourite part, Aikido. Heard a lot about it long ago but never really tried it or really bothered to find out more about it. I started in University of Birmingham under Pauline Squire Sensei, 4th dan. She teaches every Wed evenings at the University. The Uni has this Active Lifestyle Programme that\'s really a cool idea. It is a very good way of introducing sports and certain activities to the students. They\'ve got a lot of tstuff from martial arts, sports, aerobic-like stuff and even yoga. Fantastic. Anyway, I gradually went on to increase my training at Pauline Sensei\'s dojo, Po Rin Kan (literally translated as Pauline\'s House in Japanese). Just a bit of information. It\'s affiliated under the British Aikikai. This is sorta an umbrella organisation that has links to the United States Aikido Federation Western Region. And the USAFWR is under T.K. Chiba Sensei (hmmm ... should be Shihan right? I dunno), 8th dan. I\'ve been reading stuff up on some Sensei\'s and from some of the stories told by Pauline Sensei, I would really just love to watch a class of his. I\'m not sure if I can take or tahan one of his trainings though. Sighs. Anyway, I took my 1st grading (6th kyu) with Pauline Sensei and passed. When I returned to Singapore for my year out, I joined Aikikai (Singapore). Wah, I was so onz about Aikido that I increased my trainings from 4 to sometimes 5 days a week. 1 of which was a bokken course which taught me the really really basic stuff. At that time, my Sensei was Freddy Khong Sensei, 4th dan, who happens to be the club\'s Vice President. Bokken course was taught by Seah Ah Hee Sensei, 4th dan. One of my training class was usually conducted by Sensei Koh (Nai Sing/Seng I think). Let me just comment on their teaching styles. Pauline Sensei explains and tells the class what she expects from us when she teaches. It helped me a lot when I started out cos I felt like my hands and feet had exchanged places and had absolutely no way of co-ordinating myself. So telling me to shift this and do that was very very helpful. As I trained on, I found her techniques sometimes to be quite agrressive but she has mentioned before that she is really soft compared to Chiba Sensei. This also makes him sound a lot more scary and that\'s why I\'d just like to watch a few of his classes instead of jumping right in. Hahaha :P Freddy Sensei is a quiet man. He teaches and leaves you to yourself. Of coz he goes around the dojo correcting the students\' techniques and all that but he is a man of little words. Sometimes, I\'m not sure what I\'m doin is right or wrong. I guess when nobody mentions anything, it means its acceptable. Ah Hee Sensei has one of the softest technique I\'ve ever felt. I especially like his kokyu-related techniques and exercises but juz can\'t comprehend them fully. Think his Taichi background is a reason for this. Sensei Koh also explains a lot when he teaches. Good if you can absorb them in. Sometimes I can\'t take in everything he says coz I\'m concentrating on 1 small specific aspect instead of the overall. But I guess I\'m juz a slow learner. Pauline Sensei\'s ukemi training is fantastic. I\'ve not seen anyone teach ukemi in such a way before. She breaks things down into small parts that they are easy to execute and then bring them all together. Then she teaches different level/stages of ukemi and breakfall as you progresses. I really enjoy Pauline Sensei\'s class. Part of the reason is because they are usually small sized. She is able to give very personal attention on your own flaws. In Singapore, due to the large no. of students, I don\'t often get to learn what I\'m doing wrong exactly. Sometimes I feel that I\'m just going throught he motions of the what everyone is doing. I guess that\'s where my seniors come in, to help me out in this way. But 1 similarity among all the Senseis is that they all make you practice humility. In their own way, I am often reminded that no matter my grade or experience in training, I\'m still learning. That\'s good coz I really really really hate to be on cloud 9 when it comes to Aikido. I would hate it if I ever get illusions that I\'m a bad-ass martial artist who can kick people\'s asses and be some kind of martial expert. Still learning and reflecting daily, all thanks to Aikido.  OK I gotta go for lunch, so I\'d probably finish this long entry in awhile.   '),(26043,979795,'25,December,2002','2002-12-25 00:00:00','    Lastnight we had a \'simple\' Xmas eve dinner. It was pathetic for a variety of reasons. First off, I\'m not interested in having such a big group over because I really was hoping for a relatively quiet dinner. Second, there wasn\'t exactly enough food to go around for everyone. Oh sure, I had my fill of food, but I noticed some people wasn\'t eating enough. Sigh, it also comes down to some people being a bit shy and all that. No point lah, when you come to such things organised by Singaporeans, being paiseh is like one of the stupidest thing you can do. You only leave feeling hungry.  One of my housemates pointed out a point to me which I totally agree, it was that he wanted to have a proper dinner in which he can actually sit down and eat instead of stand around. I thought over his words and I agree. I also want a dinner in which everyone can be seated to enjoy the meal. I was a bit pissed lastnight and hence seemed a bit anti-social to the main crowd coz after dinner I went back to my room with some closer frenz.  But I was really pissed with myself. Pissed that I didn\'t voice out when my friend suggested asking people over for Xmas Eve. I didn\'t realise that most of my housemates actually prefered a quiet dinner instead too. Sighz... sometimes I really need to learn to speak my mind. My other housemates are basically all very diplomatic people so they took everything in comfortably. I\'m not. Well, not as diplomatic as them anyway.  So my new year resolution will be to express myself nicely to people and speak up more often if I think something doesn\'t feel right. That\'s 1.  O yea, good thing is that I managed to get my stamina for swimming back. No I\'ve yet to swim long distances, but only becoz I dun wanna be too tired for the rest of the day. Think a short swim is fine coz I\'ve been timing myself. Estimated timing lah.  Anyway, whoever reads this. A Merry X\'mas to all and Peace on Earth.   '),(26044,979795,'20,December,2002','2002-12-20 00:00:00','    Today went to town. There\'s a German market on right now. Kinda like a fair thingy. But the food was good. I had 2 pork sausages today. Shioooook man. Glorious food. Too bad it ends this coming Sun. I\'ll probably drop in to eat 1 or 2 more sausages before then.  Also saw Lord of the Rings, the Two Towers today. Hmmm... it was alright but I really expected better. I\'m not saying it\'s any worse than the 1st, but I just somehow felt it was a big anti-climax. I expected better war scenes. But it was still a good show.  Also noticed recently I\'ve been buying books. Muz be some influence by my housemates. But it\'s good. I dun regret the books I bought, they are worthwhile either in content, or price, or both. Either way, I\'m glad I\'m starting to get back into my leisure reading again.  OK betta get back to work. Heheh, but think I\'ll sneak a chapter from my new book first.   '),(26045,979795,'19,December,2002','2002-12-19 00:00:00','    OK, last few days had relaxed a bit liao. Got to get some real work done. 2 essays and 1 project and 2 modules to revise this 1 mth holiday. Half the week has already rolled by. Need to make better use of my time. NOW!!   '),(26046,979795,'13,December,2002','2002-12-13 00:00:00','    Haiz,  Yesterday had my Jap test. It was teeeeerrrible. And to think that I spent the last few days studying it completely. Even sacrificed a few hrs from my final year project to prepare to prepare for the test. Just felt a bit wasted lor. And I think it\'s a wake up call that my prep work was insufficient. Think I\'ll use the holidays to brush up. Gonna go borrow tapes and do all the exercises from page 1 onwards!! Gambatte!!!!!  Today had my project inspection. Woohoo quite good and encouraging. I was kinda panicky over it but I think the lecturer was satisfied about my progress. He just wants to make sure I know how to do what I want, and that I know what I really want done. I think I\'ll spend more hours on my FYP this holidays too. In addition still got about 2 modules I think I need to read up on and 2 essay questions I need to work on. I may be a bit greedy but I think I\'ll try to finish the 2 essays early during the holidays. But I foresee that I\'m probably only going to finish 1 1/2 of the 2 qns. Bopian. Try my best liao lor. Dun wanna waste away this holiday too much. I\'ll of coz rest a bit too lah. But I know for sure that 2 days before and after X\'mas will be quite a slack day for studying coz of X\'mas dinner and all that.  GAMBATTE Aylwin-san!! Gambatte!   '),(26047,979795,'07,December,2002','2002-12-07 00:00:00','    Whew, just had my supervisor meeting about my final year project. A bit stressed coz last min my program just can\'t get working. Made me panick. But my sup was quite understanding about the whole thing. He asked me about other problems I might get stuck in and where I needed help/advice. Also reminded me not to be too worked up about the Project Inspection next week. Gotta just keep my cool and prepare some stuff and get ready to be asked some questions. On top of that, I need to prepare for my mega important Jap test coming on Wed. Wah, this week siong man. Gonna work hard this weekend, but not before a good rest tonight. Juz loaned some videos from school library and got some classic shows. I\'ll relax a few hrs tonight first.  Even the next 5 wks looks like a break from Aikido. Sigh, sure would like to go train somewhere but I guess it might be an overkill. Taxing myself too much is a definitely no-no for me in my Aikido training. Wanna improve and not lose interest. I really want to keep at this long term throughout my life. Hope that can be done. Anyway, I can always go train my swimiming this coming holiday. Woohoo.  O yea, I just found this really fantastic site called www.vcdhelp.com. They offer tips and advice on creating VCD and DVD from a variety of format. From hardware, software to tips and techniques. Fantastic site.   '),(26048,979795,'04,December,2002','2002-12-04 00:00:00','    Been awhile. These past week a bit more stressed than usual. Got my Jap coming along now. The test next week is 40% of the module. After my Xmas holiday is the Oral test which is 20% followed by exams in May, another 40%. All quite heavy. But I think they are definitely lighter compared to all my other IT modules. Next week also got a Project inspection coming up. I don\'t theink I really need to prepare anything and I was told it\'s not gonna be accessed. But knowing that I\'ll be facing some lecturers and they\'ll be questioning me on some areas of my work kinda scares (terrifies) me. Sigh, wish my demo day will either be 1st day of the week or the last day which leaves me to be really unprepared, or really well prepared.  Aikido cancelled last night. Sensei was down in Blackpool. Tonight\'s training is taken by someone else. But I\'m feeling kinda tired from all the school work. Sigh, not sure how leh. Really think I should go down and train today coz I don\'t think my schedule can permit me to train next week. And I\'m not sure there\'s any training for the Xmas holiday either. Would love to though. Maybe I\'ll just ask. But I suppose I can always train my swimming again. I\'ve stopped for about 3 wks now. Thought I had managed to hit back 2km in 50 min. But Think now I\'ll have to re-train over from 1km onwards. Think I\'ll try to aim to hit 5km swim at the end of my holidays. However, I don\'t think I\'ll push myself for the timing. Just make it really relaxed first. Once I got the stamin back then mabbe do some sprints / speed up my time. Err ... that also means improving my tecnique. Poor technique slows me down as well. Hmmm ... sounds a bit like Aikido. Maybe I should emphasize on my technique first. The speed probably comes much later. My mind in Aikido always seem stuck on technique and rushing in on the speed. Bad bad....  Had a bit of disagreement with Corrine last few days. I know we had problems but I didn\'t expect that we\'d have so much differences. I really hate to see her being so emotionally affected by things I did (or didn\'t do). Feel so bad sometimes but I also dunno lah, think we both got things we need to work on. I was really surprised that I even suggested some time off (ok lah, more like a break-up) from both of us. But I was even more surprised by her response and effort to try to keep things together. I feel a little ashamed. Coz now I dunno that when I suggested the break-up, wheather I was taking the easy way out or I really thought we aren\'t compatible? But I felt really really reallu elated by her saying (similar lah, my memory not that good to remember the exact words gramatically), \"I\'m also not sure if we are making a mistake or not, but I\'m really willing to take the \'risk\' with you.\" O man, because of that phrase, it practically erased most of the doubts and fear in my mind. I just think, no matter the mistakes or what, so long as we\'re making the step together, it\'s the right choice. Even if it means that someday in the future we really do break-up, but because of the fact that we had made whatever decisions together, it\'s can\'t be a mistake. Right? ... Sounds a bit chim, now that I\'m re-reading my own words. But, aiya, that\'s just me lah. I can\'t phrase and express some things properly. I\'d like to just follow my instincts. Even if it means \'losing out\' in some way or making a mistake. I mean, there\'s no better time to make mistakes and pick yourself up while you\'re still young, rite?   '),(26049,979795,'30,January,2003','2003-01-30 00:00:00','    Lastnight my housmate accompanied me for my Aikido practice. It was his first try at it.  We met at the bus-stop outside the University\'s train station at around 7pm. From there, we took bus 44 down to Weiley Hill, Village Hall. It was quite a fair distance off, beyond Sainsbury, a big supermarket. Halfway, my senior boarded the bus as well. Funny thing was, we alighted at the wrong stop (1 stop too soon). It would seem that I\'ve been alighting too soon all the while I\'ve been to training. What a real joke.  Anyway, we arrived at the dojo pretty early. We had to lay out the mats together. That evening, Sensei Pauline wanted us to have the mats wiped. Soon as that was done, we proceeded on to change into our Gi, training wear, for class. David, my housmate, wore loose and comfortable attire (trackpants and tee) for class. Prior to this, he had seen some videos of Aikido, which my Sensei had loaned to me last week. He was impressed by it but still slightly skeptical. That evening, there were 4 students; John, Angelo, me David (in order of seniority).  We started warmups and some basic ukemi. As Sensei took him aside to instruct him, we were told to proceed with backward rolls followed by forward rolls. Later, we switched to shikko, samurai knee-walking. Finishing all these, Sensei wanted us to do ikkyo just for starters. She said that she would continue from where we left off from last week, which were some variations of sankyo. We carried on practising, occasionally swithing between the 3 of us and David to help him out with his rolls.  That evening, I got to try out 2 variation of yonkyo. I\'ve never seen this variation before. In Singapore, I\'ve only seen it a few times. And I remembered that I never managed to get it right. Sensei Pauline went through some simple explanations of how yonkyo works. When she saw that I couldn\'t wuite grasp the meaning of it, she tried out a small one on me. It caused some discomfort but not much of a pain. I won\'t begin to realise that I had gotten a bruise from this technique, until first thing the next day.  At the end of class, after we\'ve cleared up the mats and paid our mat fees and thanked Sensei for the night, we had a short chat. I had casually mentioned that Chiba Sensei had a very solemn look in the videos (in which she gave a short laugh) and that Tamura Sensei had very soft techniques (so much so that sometimes I wasn\'t even able to \'see\' what he was doing, the way he did it was really soft and effortless). I went on to mention that Freddy Sensei (in Singapore who had the opportunity to train under Tamura Sensei) had very similar warmups, techniques and teaching style as that of Tamura Sensei. Sensei Pauline also gave some accounts of her personal experience with Tamura Sensei during earlier Aikido camps. It was really enjoyable.  After we all said our goodbyes, Angelo gave me and David a lift down to Selly Oak where we could walk back home easily. While buying kebabs and walking home, I asked David if he enjoyed himself. He was really excited and I could tell that he had thoroughly enjoyed himself that evening. He had also decided that if time and studies permitted him to do so, he would definitely join me next week again. He described his experience to me that he felt Aikido was an abstract form. His initial impressions about it had changed as well. Pauline Sensei\'s teachings had managed to leave a deep and lasting impression on him as well.  I occasionally forget why I train in Aikido. I found myself bogged down by my own stagnant progress, that I had (in Gary\'s own words) \"reached a plateau in my training\". Not the peak, just no further improvements. This was really demoralising for me then. Having been with David that night, he reminded me of how  had felt 2 and a half years ago. He had helped to remind me why I had taken and continued with Aikido. The Sensei\'s in my short Aikido-life also played a big part in this.  Training with beginners isn\'t really a bad thing. Look for the things which remind us about ourselves. Reflect on these and I think we will realise we\'ve came a certain distance since then. If there are qualities we dislike in what we see in others, maybe we should think if we have, or still retain these qualities. If we find it irritating or offending in others, imagine how others might feel about us. Look around us and take note of those who have been really tolerant about our bad habits, most important of all, our Sensei(s).   '),(26050,979795,'16,January,2003','2003-01-16 00:00:00','    Lastnight\'s training was OK but kinda tiring. My muscles along my back all the way down to my butt is aching.  We did variations of iriminage. Was good. But a lot of them I felt really unfamilar with. I might have seen them before but I\'m really not too sure. Sensei Pauline was emphasing the need to move the centre around in the techniques. Of coz there were the fine technicalities of hand motion, feet work, etc as well.  It was an interesting way to drill in techniques coz she started with simple gyaku and ai hamni iriminage variations, before advancing on to morote, ryote and katate dori variations. I think my centering has improved slightly (but not a lot). I realised I need to be thinking less when I\'m practising. Kinda juz empty my mind so that I can move accordingly.  I hope I can remember what I\'ve learnt last 2 days and use them effectively next week. She might be bringing down some videos. If I can find out how my firewire connector of DV recorder works properly I\'ll probably make some copies for reference.  I was just thinking of my rock climbing experience few weeks back. I realised that the really good rock climbers have very very very stable and strong centering when they are climbing. Me, I only use my hands and legs. Basically I\'m using muscle power which kind of just make it so much tougher and strenous to finish a climb. But they are all able to do it almost effortlessly. Same as Aikido, timing & confidence come in as well. Some handholds which are just too far away, they time themselves nicely to \'hop\' a little to reach it. They also move their centre close to the wall before they stretch their limbs to reach a hand/foothold. Really really good. I\'d like to try to \'experience\' a little Aikido during a climb but I\'m not good enough. I\'m tryint to relate rock climbing to what I\'m suppose to feel in Aikido. Sometimes, a change of environment seems to bring out other stuff in a different light. It seems to provide a different angle for you to \'see the light\'.  Another area I\'m trying to Aiki-sync is swimming. But I\'m having a bit of difficulty. I think I\'m thinking and concentrating too much. That\'s why I can\'t seemt to relate them. But if I don\'t think at all, I\'m not able to \'see\' the link. I just hope I can \'feel\' it someday. Really love swimming. Woohoo.  O yea, was just at the  urlLink British Aikikai  website. I saw the site list Sensei Pauline as a 5th dan. Hmm ... I didn\'t hear that from my senior John. Might just ask them about it. Mabbe she got it last year while I was in Singapore or something.  Anyway, time for my lecture. Gotta run.   '),(26051,979795,'15,January,2003','2003-01-15 00:00:00','    Last night was my 1st Aikido class for for the new term. It was very good. I felt that I got a bit of a good warmup and even though I didn\'t do anything that was really fantastic, I felt that I was actually making sorta some small (but primitive) progress for my ukemi. Hopefully tonight I can get more practice as well.  School\'s ok. I\'m getting some work done early before it piles up too much towards the end of the term. Hope I can keep up and just drive myself a little bit harder towards the next few wks. I know some of my deadlines coming up so I should really be planning out the phases of my work properly and early now.  Timetable for this term seems quite alright. Though I expect things aren\'t goinna be finalised till next Fri. Anyways, so long as I got time for my work and Aikido, everything else should fall in place nicely ... I hope.   '),(26052,979795,'11,January,2003','2003-01-11 00:00:00','    Haven\'t made an entry here for sometime.  2 wks ago I tried out rock climbing, well, more like wall climbing. It\'s an indoor climbing centre (Rockface - www.rockface.co.uk) in town. I\'ve got to admit, it was scary for me. But there was a certain thrill and adrenaline rush to it as well. I have a great fear of falling even thought I was very well secured with the rope. Last week, I went for my second try. I think I\'ve gotten worse at it. But I\'d really like to go again, juz to get over the fear of falling.  It was a bit depressing to know that the girls (some of my friends before me) actually climbed better than I did. No lah, of coz I know I\'m not competing against them lah and I won\'t rush it either. But I think maybe it\'s kinda a guy thing. But I hope that\'s not going to stop me. Gonna try. Well if schedule permits, I might take it up back in S\'pore too. We\'ll wait and see.  Term starts next Mon. Argyh!! So much work I need to do. On the other hand, it also means Aikido class is starting soon. And I\'m still thinking of joining swimfit in school to refine my swim techniques. Sigh so much to do and so little time. I\'ll see how things goes, might just stick to Aikido. Still my strongest passion so far.  Talking about Aikido, heard that Jesse might be stopping Aikido. Might not be true, but it\'ll be a real pity if he does. He\'s got a natural talent for it. Sigh, wish I\'m half as good as him. Can\'t even get my ukemi right. I hope I can train and improve my ukemi under Sensei Pauline this term. Gotta work harder.   '),(26053,979795,'28,February,2003','2003-02-28 00:00:00','    Yea well, the only reason why I can practice bokken at home is because I either hold the bokken really low, or I try empty handed, or I do certain motion of it sitting down. I figured I can emphasise different area of what I learned in the following 3 ways and also considering the space constraints. Can\'t practice bokken work in my room either, too cramped.  I\'d usually do it in the living room. If the weather was warmer, I\'m thinking of laying some cover in the backyard and workout from there. Fresh atir in UK is very shiok. I heard that during the summer, Pauline Sensei sometimes conducts classes out on the grass. I\'ve seen some camps where they do move out into a huge field to train. Though I\'m not sure how dirty the gi at the end of classes.  In Singapore, I also can\'t be swinging my bokken around like a maniac lah. Parents will kill me first and I don\'t think any amount of Aikido can save me then. Usually I\'d try to wake up early to practice in the carpark. Sometimes the neighbours will see me and they probably think I\'m some kind of a maniac wielding the bokken around like nobody\'s business. Hah... but it\'s ok lah, coz private carpark, the public can\'t see me. Just one or two person catch a glimpse I can just ignore.  At home in UK, I\'ve already scratched a bit of the ceilings and walls during bokken practice. Hope they won\'t ask me to pay for repainting. Heh heh, then again I might shift some furniture around just to hide the scratches and marks.  Sensei mentioned that Chiba Sensei will be dropping by and staying over at her place for a night sometime around end of April. I\'m wondering if I will get to see him face to face and in real life. Wah, a bit scary to face a traditional Japanese Sensei, man.   '),(26054,979795,'27,February,2003','2003-02-27 00:00:00','    Lastnight\'s class was quite tiring. Again we continued from kokyu-ho techniques which we learnt the previous night.  I felt really breathless and my breathing was COMPLETELY off. Really need to get my breathing back in check someday.  I felt a bit kanchiong, because Sensei actually wanted John and I to just demo a bit on breakfall. Wah liew, I was so kanchiong I think I got really unnatural. Basically I felt very conscious tht there were a few pair of eyes on me. Sigh ... when the mind is not calm, the body naturally follows. I distinctly remembered a very cuckoo fall but at least I didn\'t hurt myself.  When the first hour was over, Angelo turned up to join us (he had missed Tue\'s class). Sensei told us she was going through some techniques for our 5th kyu grading. Argh ... imagine that. I was happy of course, but I still don\'t feel confident with it. But she also said she had to check the syllabus again so yesterday I think we only did some techniques which (may/may not) be inside for 5th kyu. Oh no. John said he\'d be glad to help us train, and that we could meet up just before the grading (good ol\' John!).  Well, anyway class went on and we did shihonage from gyaku hamni. Some tenchinage, iriminage amd kotegaishi/kotegaeshi (err ... dunno wats the right romanji spelling, need to check up someday when I can remember). Well, throughout all the throwing around, I got back my more normal ukemi than the cuckoo style I had demonstrated just earlier. Felt real comfortable about it then as well.  Anyway we ended by being uke for 2 beginners (Steve and Alex) who had just extended their training hours into the second hour with the seniors. Must admit, Steve seems to be picking up really comfortably. They were doing about four variations of kokyunage and we had to just keep rolling and rolling one after another. Good practice for both. At the end of the rools I did somehow manage to get a bit of my breath back.  Now that Aikido for the week is over, I\'ll have to concentrate on completing my essays and working on my VRML and VR projects now. OKOK, maybe I\'ll just revise a bit of bokken work with David once a twice each evening. But I gotta go home and prepare dinner for the house today. Hmm ... haven\'t quite thought over what I need to cook tonight yet. Hmmmm .... gotta go dig around the fridge.   '),(26055,979795,'26,February,2003','2003-02-26 00:00:00','    Like I mentioned in my previous post. Lastnight\'s class was really fantastic. Our (David and I) first bokken class in the UK. Yea, sure I\'d done it before in Singapore under Ah Hee Sensei, but I think I really wanted to learn something from my first Sensei as well.  We were talking to John about it just before Sensei arrived at the dojo. He was surprised (well, actually Sensei was as well) that I had never done it before. He said the last time they did it was pretty recently. Must\'ve been the time when I was back in Singapore. Anyway there was a request and Sensei said it was fine if we wanted to do it that night.  Class started with some kokyu-ho variations. Some I\'ve done, some I\'ve not. She mentioned my ukemi was better than when I started out again. Guess if I didn\'t bruise myself (surprisingly), I must\'ve learnt something about ukemi the past few months after all. But there is still the part about giving in ukemi that I really really really really need to work on. I\'m always afraid to go charging (commit) in as an uke. Think, it\'s still the fear of being thrown and falling hard against some1\'s good technique. But I also realised one thing, that if I don\'t commit properly, I can get injured easily because I probably won\'t be able to react in a relaxed manner wen taking ukemi. I really ned to work on this issue and resolve this contradiction within me soon. I feel it\'s a major obstacle in my training right now, and I\'m sure I can easily piss of the nage sometimes by not helping them learn the technique. O yea, I also tend to have pretty messed up breathing during  ukemi. That explains how breathless I get. Talking to Begh (housemate), she pointed out to me that in gymnastics (she was on the Singapore\'s National team, y\'see), it was the same thing. That you keep practising and practising to perfect it. They sometimes spend years perfecting a small part which lasts only a few seconds during a sequence of movement. WOW is all I can say, but the link to Aikido is there.  Halfway through class, we switched to bokken. Sensei wanted me to put on the tsuba for my bokken (well since everyone was waiting for me, I\'d thought I\'d just ask if I really had to put it on). But when I was in Singapore, I was the only one using the tsuba. Haha ironic. Anyway in Singapore, I suspect the tsuba comes separately at an extra cost, so most people never would buy it. Sensei first taught us how we should position the bokken. When seated, the end of the handle of the bokken extends just a few inches above the knees. It is placed on the right side, with the blade (\'sharp\') edge facing ourselves instead of the person on our right.  You hold the bokken in right stance all the time. Right hand is on the top of the handle, just below the tsuba, and left hand at the bottom. For both hands, the grip is relaxed. You have to grip it in such a way that the back of our hands are both facing us and that the tips of both thumbs is usually away from our view. I\'ve read somewhere that the grip is mainly on the last 3 fingers for each hand instead of the index fingers and thumbs. These two fingers seem to be just a guide of sorts. We are not supposed to hold the bokken like we would hold a machete. When holding the bokken in stance, the end of the handle should be approximately just 1 fist away from our center, with the blade being horizontal. The shoulders, elbows and wrists should be relaxed instead of being locked and stiff.  We started with some basic cuts/shomen strikes first. When raising the bokken, use the wrists a little first to initiate the movement, then just raise the whole arms. Be careful not to raise/lock the shoulders. Blade is raised to behind (and probably just above) our heads. The raising motion also aims to be in such a way as to protect our own shomen with our hands just above our foreheads. When striking, we were told to concentrate just on the last 6 inches of the blade edge and be relaxed. It should bring us back into the starting stance, meaning relaxed posture, unlocked joints, 1 fist away from center, horizontal blade, grip with back facing us and tip of thumbs fairly hidden.  After a few cuts we proceeded on to learn 2 variations of an 8-directional exercise. We had to bear in mind our posture constantly.    First Exercise : We will be facing North, South, West, East, South East, North West, North East, South West  1. North  We were told to raise and cut down. While cutting down, we took a step forwards with the front (right) foot.  1a.  On finishing, Bring the sword up to protect our heads. This is different from the previous raising action we did initially. I\'m not sure how I\'m suppose to put it into words, so I beg your pardon if the way I\'m describing it sounds kinda confusing or silly. This time, the raising motion is done by kind of twisting and raising the hands to our left. Of course hands are still above our forehead. The tip of the blade is now facing the \'imaginary\' person opposite us, edge of the blade is facing skywards.  2. South  Now we turn around to our left side 180 degrees, and cut down. You\'ll find yourself standing with left foot in front and right foot behind. (But hands still same as original - hands will never change position here.)  2a.  Bring sword up by twisting hands right and up at the same time to our forehead. Tip of edge faces our opponent.  3. West  Now we move our back foot to our front (but don\'t pass the front foot), and while we cut, just slide to the right by 90 degrees.  3a.  Raise the bokken by twisting the hands left and up to our forehead again, blade tip facing our front again.  4. East  Turn 180 degrees left to face behind and cut down again. You might realise we have just finshed North, South, West, East directions.  4a.  Raise the sword as we originally did, both hands to our forehead, blade tip pointing behind the head now  5. South East  When cutting, move back foot to face right by 45 degrees.  5a.  Raise hands left and up such that blade tip faces our front.  6. North West  Turn left by 180 degrees and cut as well.  6b.  Raise hands right and up with blade tip facing front.  7.North East  Move back foot 90 degrees to right and cut at same time.  7b.  Raise hands left and up keeping blade tip facing front.  8. South West  Turn 180 degrees left and cutting down.   Second Exercise : We will be facing North, South, East, West, South West, North East, North West, South East Please note that the second exercise is similar. Difference is that because the first step has changed it also alters the side in which you will be turning. But the angles will be the same. So are the raising and cutting movements.  1. North  We were told to raise and cut down. While cutting down, we took a step forwards with the back (left) foot.  1a.  On finishing, Bring the sword up to protect our heads by twisting and raising the hands to our right. Tip of the blade is now facing the front.  2. South  Now we turn 180 degrees right side, and cut down. You\'ll find yourself standing with right foot in front and left foot behind.  2a.  Bring sword up by twisting hands left and up at the same time to our forehead. Tip of blade pointed front.  3. East  Now we move our back foot to our front (but don\'t pass the front foot), and while we cut, just slide 90 degrees to the left.  3a.  Raise the bokken by twisting the hands right and up to our forehead again, blade tip facing our front again.  4. West  Turn 180 degrees right to face behind and cut down again. You might realise we have just finshed North, South, West, East directions.  4a.  Raise the sword as we originally did, both hands to our forehead, blade tip pointing behind the head now  5. South West  When cutting, move back foot to left by 45 degrees.  5a.  Raise hands right and up such that blade tip faces our front.  6. North East  Turn right by 180 degrees and cut as well.  6b.  Raise hands left and up with blade tip facing front.  7.North West  Move back foot 90 degrees to left and cut at same time.  7b.  Raise hands right and up keeping blade tip facing front.  8. South East  Turn 180 degrees right and cutting down.  You can repeat the exercises by turning back to North direction again. Throughout we really need to keep posture, and sword grip and position in mind.  Interesting huh. Damn cool man. Can\'t wait to learn more. It\'s a pity there won\'t be class next Tues as Sensei will be away for awhile. Wednesday\'s class will be taken by someone else (Paul, I think).   '),(26056,979795,'25,February,2003','2003-02-25 00:00:00','    Dammit, I had typed a whole chunk of lastnight\'s Aikido class and I just cancelled accidentally!!! Again.!!!! Now I need to retype everything because I think I need to refer what Pauline Sensei had taught me for bokken class lastnight.  ARGH!!!!! When shit happens, they really come in 1 big chunk. I\'m really not sure I\'m in the mood to retype everything now but I\'ll probably do it later after my Jap class. Sigh.  Lastnight I had a bit of a tiff with my girlfriend ... again. My house had arranged to eat \'Yu Sheng\' lastnight after I returned from Aikido and they had invited my girlfriend the night before too. I had turned off the sound to my mobile during Aikido. Lastnight we couldn\'t get a lift back and took a bus back instead. Maybe we were still pretty excited over our bokken class and David and I were talking about it all the way home. We dropped by the pizzaria and bought 2 small sized pizzas (buy 1 get 1 free) home. The moment we stepped through the door, David\'s phone rang. Apparently, I had forgotten to turn on my phone after Aikido. My girlfriend had smsed me and called me a few times after my Aikido timing. She was expecting to join us on the way back to my place for \'Yu Sheng\' when we were passing by her place. She was really pissed when we spoke on the phone. I went down to her place and followed her back to my place. It was my fault I really think so in this case. Because I seriously forgotten to turn on my mobile again when I should have. I apologised to her twice, but she\'d make passing comments about how she\'s used to it (my kind of attitude) and that I won\'t change. She feels like I\'m not really putting much effort into a relationship. But I\'m really not sure I want to change my lifestyle and habits. I dunno is it because I think it\'s not worth the effort. But I know I do feel very sian when I see her \'sian\' expression. Last time she once commented that I tend to ignore her when she tried to contact me. Though I\'ve told her I didn\'t do it purposely. At least now I do make it a point to carry my phone around more often, that my phones are almost always charged. That I check my phone a bit more frequently than I did. That I grab a seat in my lab that\'s near the window where there\'s a decent mobile reception. My computer labs I think is kinda magnetically shielded from all the networking equipment and concrete building. That\'s why in the labs I\'m quite hard to be contacted. But it\'s gotten to such a degree that I\'m beginning to feel that I\'m being artificial now - dunno if that\'s the right word to use.  I\'m not trying to defend my actions above. I know a lot of lazy things that I\'d do. Sometimes I don\'t call her daily / regularly. I know there have been times when I don\'t call her for even a week. Of course, if I was really doing serious work that\'s probably still a reason. There are a lot of times when I was just being lazy. It\'s quite sad. I think we humans are sometimes just like that, we know there are things seriously wrong with some of our usual habits but we just refuse the change them. Thats a big flaw on my part. It\'s a very bad flaw in fact.  I also know that my housemates just want to help. But I get very sian hearing that I\'m the one at fault ALL the time, that I\'m the one who don\'t put in effort. I think it\'s not I don\'t put in effort but instead what I put in is what the other party (or even what some other people) would expect in a relationship. I don\'t know what that\'s saying about me, it might not mean a big thing, but it might also mean that I\'m not ready to be in a relationship now. All this shit just goes to tell me I\'m a very kok-up person. Everything I do can never be right. I\'ve zero credibility, that I\'m just a pretty useless person just because I don\'t do what other would people do. I hate the feeling I\'m getting out of this relationship now. My personal esteem seems to be at a all time low (zero!). Juliana (housemate), was trying to talk to both of us lastnight (well I\'m pretty sure she\'s talked to my girlfriend). The moment she started, I really had to ask her to stop. I was just simply not interested in hearing that I\'ve been doing this or not been doing that anymore. I\'m also sick of hearing about things from other people. Sometimes, I\'ve been told I don\'t speak up enough, hence don\'t communicate sufficiently with my girlfriend. Whatever I did, I must\'ve made my girlfriend really sian about talking to me because she don\'t anymore, she\'d jsut show me a sian face and be very cold whenever she\'s not happy. That\'s the only way I can tell that I did/didn\'t do something. Makes me very jumpy about what I say and do whenever I\'m around her.  Aikido seems like the only thing that calms me and makes me relaxed (mentally). I can\'t really find the \'peace\' in my mind outside the dojo.  I\'m probably going to give myself a few days to think things over lah. To see which direction we are headed. If I think I\'m going to be hurting the both of us a lot more in the long run, I\'m just not sure we should go on. We sure don\'t need this kind of a distraction at this point of time as well.  I\'ve mentioned the word breakup twice before. And I really hate to make it even seem like I\'m threatening her or anything. It\'s never never been my intention even hint that at all but she has said something which seemed to say that\'s how I sounded last time. Maybe that\'s really how it sounds like. This time, I want to be really really really sure about what I want, or what I don\'t. No point dragging on if I\'m (yea yea it\'s me again, always me me me) just going to hurt the both of us in the end.  I just need a place to vent my fustration and I think blogging is a good way for me to do it. It doesn\'t matter how many people reads this (or don\'t), it\'s just for me to rave and rave about my moods. Sorry if I\'ve bored anyone to death. But you can expect a more interesting post next time as I\'ll have a lot of details about my Aikido class lastnight. Cheers.   '),(26057,979795,'25,February,2003','2003-02-25 00:00:00','    OK, both Shuhua and Eileen mentioned this commenting / tagboard service to me so I thought I\'d check them out. I\'ve just attached the relevant commenting feature into each posts. Anyone with any comments can now feedback to me anytime. I\'m still not sure how I\'d like it but whatever the case, I think I\'ll just try it out until it becomes too much of a chore or interferes with what I\'m doing.  Happy reading.   '),(26058,979795,'22,February,2003','2003-02-22 00:00:00','    Alamak, even with all the reminders I still didn\'t manage to free up time to watch the footage of advance ukemi.  I guess, when you\'re busy, it\'s probably quite alright lah. But I really hate it coz I think I\'m also usinga lot of excuses to slack. Especially when it comes to my school work. I\'m behind schedule on my final year project and also my essays. Really need to catch up at this point in time. I have tonight, but I think my Aikido is probably gonna tax me up quite a fair bit for me to work efficiently. Maybe if I don\'t use too much strength or energy and see how things goes...  What\'s this tag board thingy  urlLink  you were talking about? My friend jsut sent me an email on something he called a \'commenting service\'. That wouldn\'t be the same now would it?  I\'ve submitted a mini-project for my VR module. I\'m gonna make a virtual dojo webpage. Wah a lot of things I need to look into. Musn\'t be too difficult and and interfere with my final year project schedule, and musn\'t be too simple that I can\'t get a decent score. Really troublesome. But interesting lah. Just wish I had more time to work on this instead of cramping everything in now. Think when I go back Singapore I might look for some kinda course or books to do my personal project andhost it on my website or something. At least with a VR website, I can probably escape Geocities\' stupid pop up on the top right corner of my pages. Spitz at them... ptuiz.  Anyway, I got about 4.5 hrs to work on my final year project now before Aikido leaving for Aikido tonight. Hope I can accomplish something today. Gambatte kudasai!   '),(26059,979795,'22,February,2003','2003-02-22 00:00:00','    Dammit! My stupid touchpad on my laptop is so touchy. Just tapped lightly only and it responded when I didn\'t want it to. Basically I typed a bit of stuff just now and they\'re all gone now. I\'ll try putting the stuff down below again but I won\'t bother to rethink my exact words.  Busy weekend. Tomorrow morning need to go down town to buy meat for the week\'s cooking. After that need to come home and go back to Uni to do some work. Been lagging behind lately and I\'m panicking. Daniel, my housemate, was asking if I\'m swimming also. Doubt it. I\'d rather get a bit fatter if I can clear up more of my work. Anyway, I still go Aikido so I dun feel like I\'m being really slack or anything. But I sure do wish I\'m keeping up with my swimming though.  The girls (housemates) today were really noisy from dinner onwards right till just now. Dunno why they\'re so hyper. In fact, got a bit irritating for awhile. But sounds like things have settled down a bit now. Think worst thing is that with their high pitch voices they tend to be a bit noisier when they are slightly hyperactive. Guys (like David and I) can get loud, but not really noisy I think. Tried doing my essay just now and I only kinda got 1 paragraph done. Need more time and effort or I\'ll never finish my 1st essay this Sun and start on my 2nd one. ARGH!!!  Sigh, really dun feel like missing any Aikido coming week. We\'ll be doing nikkyo this Tues I think. And the Tues for the week after, there won\'t be any class. Think Sensei Pauline\'s going out of town for awhile. Just got my bokken from Sensei last Tues, really wanted to refresh some stuff that I\'d learnt in Singapore but my living room isn\'t quite high enough for me. If weather was warmer I\'d do it in the garden. O yea, and must remember to watch Donavan Waite\'s Advanced Ukemi this weekend. MUST REMEMBER!   '),(26060,979795,'21,February,2003','2003-02-21 00:00:00','    Wah, at least  urlLink you\'ve  had a good practice.  My training on Tues was pretty alright. Wasn\'t great but still ok. Wed, I had to skip the first hour and could only go for the second. But I don\'t feel very warmed up. 1 hr class always never feel alright with me. Feels a bit abrupt. Like everything ended too early. Dunno lah, just crap lah.  Sensei taught us 4 variations wourking our way up to a technique called Uderiminage. Wah, took me a while to remember the name. Think she said it\'s called \'arm entwining\' throw or something like that. It\' a very good and effective technique, but I must admit I had a lot of problems being uke. She pointed out that I wasn\'t giving enough for nage to do the technique. Sigh, spent the whole day working on the giving part, but I still can\'t quite get it. Alamak how how how. Sometimes I realise that even if I do certain things right, I lose it the next time I practice. How come I can\'t keep things consistent. Ok lah, this giving area is something I\'ve never been able to do properly. Sometimes I can do it to a \'passing\' degree for certain techniques but for uderiminage it was absolutely crucial, and I just couldn\'t. Think this weekend, I\'m gonna go take a look at Donovan Waite\'s Advanced Ukemi footage again. Try to understand something that I might\'ve missed out.  ARGHHH!!!!   '),(26061,979795,'14,February,2003','2003-02-14 00:00:00','    Wah piangz, Eileen just informed me of her new blog. Alamak, and she mentioned that I neva informed her about my blog earlier until few weeks back too. Haiz, like I said, paiseh lah. First blog and all that. Was hoping to build it up quite a fair bit first before letting people know about it.  Anyway, it\'s quite heartening these pass few days regarding Aikido. Well, first off I\'m glad I help introduce Aikido to David, my housemate. He\'s really happy doing it now. Last 2 people I introduced kinda lost interest after 5 mins. I just felt that they only wanted some kinda company and that they weren\'t really interested in learning at all. I don\'t mean to be rude, but some people in my Singapore dojo gives me that kind of feeling as well. David would always be asking about stuff and pondering over the things he learnt. I must admit, that by going to the dojo instead of the University class, he gets more personalised attention from Sensei Pauline. That also helps him to pick things up a bit faster than most of the beginners I\'ve seen in University. Problem with most people in University is that, they come and go very fast and Sensei rarely has a student that stays with her for long. I feel sad sometimes when I think about it. But University students tend to have academic workloads and it\'s hard for them to turn up consistently as well, even if they are really interested. I get the feeling that Sensei is well aware of this point and she just hopes that she can make them have a good start so that wherever they go when they leave, they\'ll be able to go along nicely and comfortably from there in their Aikido practice.  Tuesday\'s Aikido class was quite alright. A little dis-orienting for me because I\'ve not been \'revising\' what I was taught properly. Been thinking a bit about school work and all that, you see. Maybe I\'ll plan a few mins daily just to \'revise\' my ikkyo, sankyo thru rokyo. Either next wk or the week after we should be working on nikyo. A bit excited and apprehensive at the same time. Coz I know I need the practice, but I also know it\'s gonna hurt quite a bit haha.  Wednesday\'s Aikido class was alright. Sensei made a single comment that my ukemi was much better than before. I\'m glad I\'m actually learning something. I really really pray that I can remember to apply them from now, even if I don\'t do the same exercises wherever I\'ll be going. I do agree my ukemi seems to have improved a little. Obvious because I don\'t seem to hurt myself as much as before. But of cause there\'s a whole other room for improvement and I\'m really hoping to pick up some more about this area.  I just borrower a video from Sensei Pauline on O-Sensei. It shows a bit of history and a little of his techniques. Truly remarkable. Problem is that because they were originally from 8mm footages, the quality are not very good. It\'s also quite funny to see the many senior Sensei\'s of Aikido when they were young. Some of them really looked quite different. I really couldn\'t tell which was Tamura Sensei. But I was able to pick out Saito and Kisshomaru Senseis.  Recently I also found a new club in Singapore, Ban Sen Juku, from a Belgium website. The most interesting thing to me is that they offered Iai training, one of my greatest interest. From my really limited understanding and knowledge, I just feel that Iai training can be (or is) a really integral part in Aikido training too. The relation in posture, movement, technique, stance and even balance just feels so close to each other. Prior to my return to UK, I found out that Ah Hee Sensei is actually a certified teacher for Goshindo (long and short sword), and I was thinking of approaching him to learn from him when I return to Singapore. I hope to drop by BSJ S\'pore to find out more about their classes and syllabus too. It\'s great to have so many options available. To think that a few months back, I was just lamenting that anyone interested in swordwork could only pick up Kendo. But that, to me, is another different aspect of swordwork and one that I\'m really not too keen to learn. After about a day however, I got an email saying that the club wasn\'t established in Singapore. It was a mistake on the Belgium site. Nevermind, I can always ask Ah Hee Sensei about this.  Sometimes I really wonder what is my interest in all these martial arts. Am I really interested in learning them. And how long I can keep it up with pressures such as, family life, girlfriend, future job pressure, work schedules etc. But I think I maybe thinking abit too much about it. All these details aren\'t stopping me from learning yet. But I\'m really trying to avoid being put into a spot where I have to prioritise or choose one of them someday. I really want a balance and I hope that the people around me will be able to accomodate and understand that. Like I said, it\'s way too early now. So I\'ll just take things as they come.  Been contemplating my last few months here in the UK too, like what I plan to do and all that. If I can finish my project before easter holiday, I hope to go on a 3-4 days trekking trip. Think that should be quite fun but really is highly dependent on the progress of my project.  After that will be my exams. If God wills it, I really want my last paper to end before the last weekend of May. Sensei Pauline has mentioned there\'s a 1 wk summer camp being organised up at Bangor University in North Wales (Chiba Sensei will be there!). I\'m really hoping I can go for it, but again, it depends on when\'s my final paper.  After exams, I probably need to spare 2 wks on a tour with my girlfriend (Corrine!!). I\'m hoping to visit a poly classmate and his wife in Sweden, who is a Navy regular and was posted there for sub-marine training.  After that really depends on when my parents are coming before my convocation. If they come over early, naturally I need to spend time with them and just pay a visit to some places here. If they\'re gonna save some money by coming only for my convocation, hopefully I can find and join some uchi deshi programme here in UK. Currently I\'ve only managed to find 2 dojos online. Of course I\'ll be consulting my Sensei on this as she may have a better recommendation or suggestion but whatever the case I will only go if she gives her approval for it.  All the plans I have are really dependent on how much money I have left when the time comes. I estimate that to go for all of these I may need at least 500+ pounds. Of course if I sacrifice some of these ideas the required budget will be less. Since I don\'t intend to tax my parents by asking them for more money, I think my budget will be a tremendous consideration when the time comes. Sigh, sometimes just wish money can drop from they sky when I need them. I don\'t need a lot, just enough to fulfil my plans. (But I guess if they did, I\'d only make plans for trips to Hombu dojo, Japan and the likes hahaha. Moore\'s law truly applies as well in my case.)   '),(26062,979795,'11,February,2003','2003-02-11 00:00:00','    I was doing more reading and research on Aikido.  Been to the late Saito Sensei\'s tribute pages. He has really left behind a legacy. A huge number of students from different countries have benefited from their experiences with him. His passing is a truly sad event and I feel that he will be sorely missed. But we can all be glad that his many students show that a part of him lives on in them and I hope that every bit of his legacy will be passed on and down to the future generations of Aikido-kas.  I\'ve also been reading up on Aikido Yuishinkai\'s founder, Koretoshi Maruyama Sensei. He once held office as the President of Ki no Kenkyukai.  He offered some particularly enlightening insights about Aikido. I\'ve exerpted some from their  urlLink website .  \"to ‘move quickly’ does not mean ‘body speed,’ but ‘mind speed’.  Through Ki extension and by using the mind, the body can respond naturally and quickly.\"  \"When questioned about aikido politics and the Aikido Yuishinkai’s founding premise of, ‘Aikido without boundaries’, Sensei explained that all styles of aikido are welcome to train and that Aikido Yuishinkai students are free to experience training wherever they choose. Sensei further explained that politics are like American English.  Depending on the geographic part of the country you happen to be in, everyone thinks their own accent is the correct one.  He expressed that there was only one aikido, that taught by the founder, Morihei Ueshiba.  Differences are created by different perceptions.  Much division is created by money and ego. Maruyama Sensei doesn’t see his aikido as ‘right’ and everyone else’s as ‘wrong’.  He says aikido is like a feast.  Some choose the salty foods and some choose the sweet.  Everyone should decide what they like. He said Aikido Yuishinkai was like a beautiful fruit tree.  It does not need to be advertised or promoted-people will come to pick the precious fruit.\"   '),(26063,979795,'04,February,2003','2003-02-04 00:00:00','    I\'ve been surfing around some Aikido websites again. Just to keep myself updated on the happenings in Singapore. It\'s really tough doing so when I\'m on a separate continent.  Anyway, I was estimating how long it took for anyone to find the Aikido in Singapore website. Basically, it takes a while. I realised it actually goes to the other clubs first, such as Ki Aikido, Aikido Yuishinkai and Aikido Federation before they see my site. A real pity.  I also noticed that ShinJu-Kai pops up quite readily in a general search on Aikido. Though I\'ve been to their websites countless times, I\'ve never searched for the term ShinJu-Kai before today. I realised that among all the clubs in Singapore (so far), they are the only one which has been expanding. I noticed an Aikikai Shanghai branch which is affiliated to ShinJu-Kai. This is really cool. Now I\'m getting kind of interested into the strategy that ShinJu-Kai adopts in spreading Aikido. Even Philip Sensei\'s original organisation Aikikai (Singapore) has remained kind of stagnant for a period. One might even argue that they\'ve already downsized following Freddy Sensei\'s resignation from Aikikai (Singapore) to form Aikido Federation (Singapore).  Sigh, a sad thing in the Aikido world. I realised it\'s not entirely uncommon since the old days. Say, Koichi Tohei\'s split from Aikikai to form the Ki-Society, and also US Aikido Federation\'s leaving of the International Aikido Federation in 2002. All these are way out of my league and I just hope that I will be able to steer clear of all these in my Aikido learning.   '),(26064,979795,'03,February,2003','2003-02-03 00:00:00','    I was just lazing this morning before my classes started this afternoon. OKOK, not really slobbing lah. I did do some ushiro ukemi in the living room. Didn\'t hurt myself ... yet. Proabably because I did it really slowly.  After awhile, it suddenly dawned on me about an experience with some beginners I had last week.  I was paired with a new-comer to the Uni\'s dojo and we were going through Shihonage. Well, each time I brought him down to the ground, he was able to slip out of the pin I had on his arm either completely or at least halfway by rolling out from under his own arm. This brought to mind about what I was doin (or wasn\'t doing).  I wasn\'t able to completely hold him in a pin because my technique wasn\'t really good, and second because I wasn\'t commmitting fully into the technique either.  Just a few minutes prior to my blog (and what prompted me to write this blog), I was surfing around AikiWeb and reading some articles. That\'s when the picture of the scene I had in last week\'s class just popped into my mind. (I dun understand why because what I was reading didn\'t bear any close link.)  Anyway I remembered a second instance in which I paired up with Steve, who\'s been with the dojo for about a school term. I remembered him being stiff when I was still fumbling with a kokyuh o technique on him. (In his own friendly words, he said it\'s probably him \"just being difficult\"... nice chap hehheh.) Anyway, some of the techniques I did went easily on others but not on him. Similarly, I was poor in my technique and I was still trying to be \'nice\'.  I\'m not saying I should go around breaking pple\'s arms, but I\'m just hoping that I can train towards being able to feel how a beginner is coming towards me and still being able to handle him despite his \'stiffness\'. In addition to my previous blog, here\'s another point I can look forward to by training with someone who is unfamilar in Aikido.  I should explore the little loopholes which my partner exposes and make use of them fully. If I should \'hurt\' him/her a little in the process (pls note that I dun mean injuring my partner), I think that is a form of training for them too. If they decide to stop participating because \"that chap in the dojo always gets too rough\", I think they are limiting their own Aikido experience. But if they should see and understand something from it, I think they might\'ve discovered something about Aikido.  We should endeavor to moderate a little of our \"niceness\" to our partner, not to make things unnaturally easy (or hard) for them. Instead, we should probably apply whatever level of restraint and \'nastiness\' with the thought that whatever we do to them will enable them to realise something for themselves. It\'s not easy to keep this in my mind when I\'m thinking of learning a technique by pairing up with a senior, but it sure will be easier to try it out with a junior.   '),(26065,979795,'30,March,2003','2003-03-30 00:00:00','    Err the guys are still playing the drinking game ... ar duh ...   But at least, they are all having fun and enjoying themselves.  Was surfing around just now and I saw Aikikai (Singapore)\'s website, wah they actually went to the extent to post a certificate of Hombu\'s recognition certificate on their website. That\'s like ... wah liew ... such a cheapo thing to do. I really have no idea what they are doing lah ... but with all due respect to Hombu and the art, I think it really suxs. Brings down their image. I think the best way is to put it under the section \"About Us\" instead.  My goodness. ShinJu-Kai actually post a notice on their site saying that due to the pneumonia thingy, they\'re suspending classes for the next few practice. Looks like the case is really quite a big thing in Singapore now. And if anyone hasn\'t checked out ShinJu-Kai\'s photos page, I\'d really suggest popping in for a look. They\'ve got quite a few good shots of the whole event.  Tendoryu\'s website still hasn\'t got anything up yet. I\'m eagerly anticipating what they are planning for their website.   '),(26066,979795,'28,March,2003','2003-03-28 00:00:00','    My housemate is having a bbq in our backyard right now. Basically inited the biggest buncha people. All from engine dept. They\'re on a drinking binge now. Making a bit of ruckus but still within acceptable limits. One is feeling like really drunk right now and he\'s gone to bed (at 6pm heh heh). He tried to make me drink 1 shot on his behalf but I stood by my ground. No compromise when it came to drinking alcohol.  OK lah, last week was siongest week for me but also quite glad that it\'s all over. Tues I had 1 last supervisor meeting. It was a group meeting in which they went through all the key points expected in a good report. Was really helpful. One very impt point he mentioned was that if we felt that we did pretty bad for the program, the marks would really count in a report. I\'ve got 2 wks to get it done. Wed I had my Jap test. Basically listening and written. Wasn\'t terribly difficult just wasn\'t really a walkover. I prob won\'t do well but I think (hope)  I might be able to clear it.  Night came and I went for Aikido (also after having been skipping it for quite sometime). It was the last class in Uni. Again I brought down the ideo cam but I really didn\'t manage to get much good footage. And worst, most footage I captured was at a really bad and odd angle. Think I\'ll probably get more footages in my coming 3 classes thiss holiday. We did variations of kokyu nage. Started gyaku hamni katatedori with the normal tenkan turn and throw, followed by a ryotedori tenkan and throw. There was one in ai hamni as well. All these led on to the final variation. The last one we did was with 2 persons, each grabbed an arm with a 2 handed grab. Still kokyu nage and it was an immense test for coordinating both left and right hands. Sensei pointed out something to me tonight. That I needed to respond with much more confidence \"You\'re a 5th kyu now and going onto 4th. Gone were the days where you wait and pause before you perform the technique.\" I accepted her gracious comment and I keep reminding myself to practice as she had said now.  Basically that evening there was only 6 students. John, Yannis, me, and the beginners were \'Magik\' (think that\'s how his name is spelt), and a couple; Alex and Rachel Royal. Gotta comment on these 2. They are really friendly and nice folks prob about 30+ yrs old. Alex is quite stiff but he\'s trying hard, especially with his ukemi. He get really tired and breathless sometimes a bit too soon. Rachel always like to grab me as her uke/nage. She says she like me not throwing her as hard as some would I think. Haha ... something I really should be working on correcting.  At the end of class, I took a dojo photo. Will try to post it online somewhere this holiday. We ended about 30 mins early and went to Staff House for a chat and a drink. Had a cup of coke and some chips. It was an enjoyable get together. Sensei shared some insight into the character of Chiba Sensei. He\'s a very hard and formal budo man. She said something about a \'love-hate\' thing. You either love or hate the guy really. Very very domineering person who knows exactly what he wants. Eerything he asks of you may seem hard/tough. But the things he expects from you, he\'s already done it and gone beyond that. However much and hard you think you\'re doing, he\'s done it much more and much harder. However much you\'ve pushed, he\'s done that before as well and beyond. That\'s why I think he has such expectations. But he really is a very nice man. He has a many students but he really isn\'t the type that most of them would have a relationship with. I think it\'s something like when I was serving my BMT. Some Sgt don\'t expect to get too close to the recruits so as to maintain the line of command and all that. A lot of people has mentioned to my Sensei that she has a very good relationship with Chiba Sensei which I agree as well. I\'ve not seen them together but I think she\'s like one of his closer students. In fact 3 wks from now, he\'ll drop by (on his way to France) and will be staying 2 days at my Sensei\'s place. Heard he loves to fish and my Sensei is staying near a really nice fishing area.  Fri came and I had my 30 min demo. Both the lecturers were late. I was quite panicky about it. Slept about 3 hrs (though I actually intended to grab a nap of juz 30 mins) the night before while still working on some last min stuff. The lecturers were quite decent about the whole thing and I think things went a lot better than I deserved. I had to mention that I didn\'t use any 3rd party software to do my VRML geometic productions and I think that realy scored some small point with them. The overall product isn\'t too fantastic though. After my report is done, I\'ll try to setup a webpage where I can put my project up.  Anyway, got home. Took a nap and got up for a run. Manage to feel my stamina coming back then as well. Prob do more runs this holiday and some swimming. Will still be going to Uni to get most of my studying done. Night I think I\'ll work on my VR module project (diff from my FYP). Basically model a Jap looking dojo which may double as a Virtual dojo online. If anyone has any ideas about what I can put inside, I\'m pretty open to suggestions. I\'m definitely going to put in video clips for viewing.  Sat came, went to get groceries in the morning, tried to revive the server but to no avail. Hdd looks like it\'s goin soon. Looks very unstable now. Anyway my housemate decided to use her laptop as the server but that means that we can\'t go online after she\'s gone to bed because she refuses to on her laptop 24/7.  Oops, I failed to mention that after my project demo, I went to grab a G.I. hair cut. No. 1 on the back and sides. Kewl man. But freaking cold when the cold wind blows (hmm isn\'t that a song?).   '),(26067,979795,'28,March,2003','2003-03-28 00:00:00','    Woohoo!!! It\'s all over!!! Project demonstration. Yeeehaaaaa!! ... OK I\'m ecstatic now but I do realise I got a lot of things to follow up on within the next 2 week. Project report for 1. If u guys are interested, I\'ll just upload the thingy somewhere. But you\'ll need to download a plugin for your browser to be able to view VRML. I didn\'t manage to have time to integrate these into Java (heard it takes a whole lot of time so I decided to forego it, no point spending that amount of time when it\'s just 4 buttons to press around.)  A lot of stuff has been happening this past week. Aikido, work, house, friends. I\'ll share them later coz ... today is the LAST DAY OF SCHOOL!!!!! I need to clear some stuff and photocopy some notes so I can start my lessons next week. Hate to come running to school and waste time doing all these during holidays. If I have to come, I\'d much rather it was productive rather organising all the shitty stuff mentioned above.  Ja mata.   '),(26068,979795,'26,March,2003','2003-03-26 00:00:00','    OK just now saw my supervisor liao. It was actually a group meeting of all his project students. He was giving us tips on how to proceed with writing our dissertation. Looks like if I cannot write my program properly, I have another shot at it by writing a better report. The weightage isn\'t very clear/well defined. I think they really need to see your program and your report to be able to give an overall score. Still got hope lah basically. Must fully use the next 2 weeks to write 1 swee swee report.  After the group meeting, I met him in the labs to clarify some VRML thingy he already told me before. Heng, he\'s quite patient with me. Seems like I\'m very kok-headed. Anyway I\'m just waiting for my coursemate to finish his class before we go back to my place for makan. Beats walking back alone.  Only lastnight, my ex-girlfriend\'s housemate got robbed. I only heard about it this morning. The moment he left the train station, 4 guys shoved him against the fence and robbed him of his mobile and train pass (dunno if he lost any money). Heard they were going to beat him up but when they saw someone coming they ran off instead. Frankly, just imagining myself in his shoes and I feel a bit pumped up with adrenaline. Not because I want to nikkyo them lah. But because there\'s a sort of helplessness in the situation. I\'m not sure I would\'ve resisted (of whether I can in the 1st place). It\'s lucky that my ex got home just before her housemate did. I can\'t imagine how dangerous it would\'ve been for her to have been robbed by the 4 guys.  I guess I\'ll just have to keep my eyes peeled whenever I pass by that area late. And I think the guys in the house need to take more care and effort to meet the girls and walk them back home. Really terrible. Wish I was able to take 4 guys on or at least have \'that\' kinda level of Aikido that I don\'t have to worry about being attacked suddenly. Haiz ... Aikido o shimasu, gambatte kudasai!!!   '),(26069,979795,'25,March,2003','2003-03-25 00:00:00','    Well, no surprise that my house server crashed again 1 final time. Think it\'s almost breathing it\'s last coz no matter what kinda options I restarted it in, it just came out the MS blue screen. Can\'t even enter safe mode anymore. Suspect the harddisk finally gave in. Anyway, lagi best. Means I can concentrate properly on my project meeting tonight, my Jap test tomorrow and my project demo this Fri. DAMMIT!!!! Kuusooooo!!! (Note the effect of knowing  urlLink someone  who throws in Jap verses every now and then.)  Oh yar, last weekend Annie (housemate, in 2nd year medical school) had a group presentation to prepare. She invited 2 of her friends over and one of them (another Annie) brought 2 Chihuahuas!! Cute leh. I always knew how one looked like, but to have 2 of them running around and licking your finger was like ... wooohooo. Did I mention they were cute little bugger? Fed them a bit of ham and they started to hang around my feet when I wanted levae the kitchen with my ham sandwich. Heard they\'ll be dropping in sometime this Easter again and they\'ll be going to the nearby park for a walk. Woohoo ... Chihuahuas.  I\'m probably not gonna start server revival operations till this weekend. But only after a well deserved break and maybe a lunch in town or something. FRI!!!! FRI!!!!! C\'mon!!! Anyway, won\'t be online till my server comes back on. Cheers.   '),(26070,979795,'22,March,2003','2003-03-22 00:00:00','    Well can\'t believe it. I had actually done a bit of progress for my project the past few days. But still some way to go in terms of coding. I love VRML. I\'m not proficient at it but I think it\'s just wonderful. Before term ends, I\'m gonna copy every past year project I can find on the module site. They make for fantastic references. Gonna burn them on a cd and go back Singapore to study it properly. Kewl man.  Anyway, I finally finished my book, Warriors of God, by James Reston Jr. Yea yea, I\'m surprised I still have time to leaisure-read books, but it took me quite a long while. Been getting back into reading since I came back to UK. Think I owe my thanks to one of my housemate. He\'s like a freaking walking encyclopedia. Well, navy regular studying Economics lah. 1 of those award/scholarship/bursary winners (think not considered scholarship), so must be a real smart fella I guess. Quite interesting to hear his views on current affairs and all that. Anyway back to the book, it\'s about Richard the Lionheart and Saladin during the 3rd Crusades. There were 5 Crusades in all, and the only 1s that were really considered a success would probably be the 1st and the 3rd, but the latter is really debatable. Some interesting facts, Richard the Lionheart was a gay. Apparently he had an \'relationship\' (ahem ahem) with the kind of France (who ultimately became 1 of his most direct enemies). Richard hs always been made out to be sort of a war hero and goodie goodie guy when in actual fact, it seems like Saladin had exercised more kindness to his enemies than Richard did. There was this particular moment that I remember in the book where they were going through some diplomatic discussion for a possible peace treaty and when Saladin didn\'t reply, Richard actually marched a the Muslim prisoners of war to the front lines and had them executed in plain view of the Muslims. Saladin on the other hand had displayed a great level of graciousness to his enemies by releasing his prisoners of war during the capture of an enemy fort/town. During the final battle of the 3rd Crusades, Richard was unhorsed. Saladin sent him a horse only because it did not \'befit the status of a king to be standing on the same level as his own men\'. But despite all these, I really must agree that Richard was one helluva tacticion and strategist. In that sense, Saladin might even be just a notch below him. 2 great leaders.   '),(26071,979795,'20,March,2003','2003-03-20 00:00:00','    Dunno if anyone has seen this, the  urlLink Aiki Forum . Looks good. It\'s still relatively new but seems promising.  Benedict Loh, the super administrator, is a Computer Science undergrad (I\'m assuming that means he\'s a NUS student) who trains with ShinJu-Kai (Gek Poh dojo).  Well, looks like Singaporeans are taking a first step into making a mark on the Internet. Hmmm now all we need are names like Aiki.net, Aikido.net, Aikido.SG, etc.   '),(26072,979795,'19,March,2003','2003-03-19 00:00:00','    OK I\'ve just evaluated the Interactive-X blog site. Well it looks simple and nice enough to use lah.  It doesn\'t seem like I can modify much html codes inside. Looks really funky (maybe a little too much so for me). You can upload MP3 onto it as well, provided they are original MP3s which do not breach copyright laws (where\'s the fun in that then, sheesh). There\'s 1 really irritating thing about it, it keeps refreshing something in the background, I dunno what. Once it got so bad it was doing it like every 3-5 seconds. I wasn\'t even able to key in my login ID and password properly. When I finally did, it just constantly refreshed and I can\'t get anything done. Think it\'s just an occasional hiccup because when I closed the browser and went back to the site, it never got that bad again. But it still does the background refreshing thingy coz u can hear mouse click sounds every like 10 seconds or so.  Bottom line? I\'m not gonna use.   '),(26073,979795,'18,March,2003','2003-03-18 00:00:00','    I had forgotten that I had registered with  urlLink another blogging site . I\'m not sure if it\'s better or not but it looks funkier than Blogger.  Anyone tried / checked it out before? Any comments? Looks quite good though. Doesn\'t seem to have ads inside either. I Haven\'t tried customising it at all yet. Not sure if it\'s possible.   '),(26074,979795,'18,March,2003','2003-03-18 00:00:00','    Hmmm I\'m thinking I may need to postpone the trek with the guys next month. I think I may be squeezing in too much stuff before the exams. I\'ll put the thought forward to them about doing it right after the exams instead. Anyway, it\'ll be summer by then and we should be expecting better weather and temperature also.  Haiyar! It\'s no wonder I was told to reduce the suaning for 1 of my housemates. Coz she just got together with David recently. Alamak, I should\'ve known. I was finding it rather suspicious also. How come he can suan but I cannot suan ... hmmmmmm  . Now I understand lah huh. OKOK must give my brudder some face haha.  And my finances / budgeting this term has been a real mess. I think come Apr I really gotta get them updated. Still have enough cash lah, just don\'t have an idea of exactly how much I\'ll have for each month. And I really should keep track of all these so that I won\'t have any problem with planning for trips later. Basically I can just easily look and know whether my cash is enough or not. It\'s a good habit that always seem to get mess up when I reach the 2nd term.  Gonna have to skip Aikido tonight and tomorrow. There\'s suppose to be a weekend course this coming weekend. Had to give it a miss. If I\'m lucky, I can probably attend the on with Chiba Sensei around the end of May. Keeping my fingers crossed now.  Tentative plans will prob be like: Apr - Easter holidays and study for exam, postpone trekking to after exams - Email SPF asking about intake dates and contract details May - Check with Pauline Sensei about the possibilities of uchideshi programs in UK - Revision lectures + Exams Jun - Exams may end in 1st wk. May go trekking for few days - Visit Marcus+Ivy in Karlskrona, Sweden - Visit Karlskrona Aikidoklubb (wooo, is dat swede/dutch language) - Return UK collect results - End of school - mid Jun, pack and start shipping stuff back to Singapore - Visit Gary?? OR Uchideshi program (3-4 wks) Jul - Convo (16th-23rd) - Parents coming either juz b4/after convo Aug - Return S\'pore latest early/mid Aug   '),(26075,979795,'15,March,2003','2003-03-15 00:00:00','    OK lah. The pace of my life this week is revolving a bit too fast. A fully packed week due to my final year project, slow progress, my 5th kyu grading on Wed, worrying for my coming jap test in 2 weeks, my project demo in 2 weeks, possibility of goin on a 5 day 4 night trek on the 1st weekend of the Easter holidays, project documentation due in 4 wk\'s time, 1 jap test when school reopens and studying 1 year\'s worth of sch work in 4 wks while juggling the above mentioned stuff. ARGHGHGHGHGH!!!! OK lemme blog a bit to calm myself 1st.  Tues\'s Aikido was ok. I recorded the class and had them digitised. If anyone\'s interested, I can send over via ICQ. Also digitised the 1st AFS grading. Well, left a bit of the 2nd kyu grading lah, basically only the tachiwaza not digitised. We went through the 5th kyu syllabus and I was kind of just going through the motions in training. My heart and mind just wasn\'t there. At the end of class, when we did a revision of the whole night, there were some techniques I just couldn\'t remember. Shitz how do you remember everything anyway. Until now I always have this problem. Maybe it\'s because I\'m not concentrating fully lah. Well, when class ended, we were told that grading will be next Tues. I was still thinking that I had the weekend to find my senior and clarify some stuff.  Wed came. Just before class Sensei asked if I received the 5th kyu grading syllabus which she had emailed me that very evening just a few hrs ago. I replied I had not because I had left the house early this evening. The 1st hour with the beginners was ok. Sensei made us do some techniques which were actually part of 5th kyu grading. Being the fool I am, I didn\'t notice. Like I said, I just went throught the motion of class again. We did quite a few times of ai hamni nikkyo but somehow it always registered in my brain as an ikkyo. The 2nd hr came, and again I just followed blindly. We did a lot of techniques. If felt like a blur to me because I wasn\'t really thinking much. Basically just using my eyes and not my brain. I probably saw a lot of stuff but never really understood them. Haiz... We started off with hanmi hantachi kaitennage. My partner, Yanni / Yarni (Greek), couldn\'t be nage because he has a bad knee. So while waiting for a chance to ask Sensei if he can do the tachi waza version, we just carried on 1st (problem was Sensei was concentrating rather hard and long with the beginners so we actually did quite awhile of hanmi hantachi). Wah, wat a killer. I started off ok. But quite possibly because I was already half tired out from the 1st hr, I lost my stamina after mabbe 6x of the technique. My knees were so knackered out. Until ultimately I felt off balanced and out of breath. When permission was given for Yanni to proceed with the tachi waza variations, the reprieve didn\'t help me much. Because he was good, his turn always finished really fast and soon I would be nage again. Techniques carried on and all I wanted to do was to finish so that I can go home for a rest. As uke, most techniques that night were forward rolls and I also burned myself out over these.  In the last 15 mins, we were told to line up. Thinking it was the end of class, I was quite thankful and hopped quickly to the end of class. However, instead of ending, she asked Angelo (also grading for 5th kyu) and me out. Next she wanted 2 ukes. The beginners were told to get comfortable (but some of them, not knowing what was really going on, just kept on kneeling for the remaining 15 mins). We were told to do a variety of techniques; katatedori gyaku and ai hamni shihonage, ai hamni iriminage (jodan and chudan variations), ikkyo and nikkyo and some others I can\'t remember at this point. Nikkyo was the one that caught me. Like I mentioned, I kinda thought they were some weird looking ikkyo variation so I took awhile to remember what I needed to do. I was simply horrible. There was a few rounds of gyaku hamni tenchinage (2 variations). Next was kokyu ho (standing) and kokyunage. Randori was quite ok I suppose. She asked us if we still recalled the sankyo, yonkyo, gokyo, rokyo which we did for a straight 2-3 wks sometime ago. I think both of us could only remember certain variations but not all of them. I basically did 1 sankyo, 2 yonkyo and 2 nikkyo. As usual my nikkyo suxed. I think this was Sensei\'s way of giving me a 2nd chance for my nikkyo. Haiz ... I dunno lah. We ended grading by taking ukemi. It was a whole lot easier to me than asking me to be nage. At least I had more time to clear my head. Taking a few rolls here and there helped a lot.  Well, both of us passed. Angelo did considerably better than me - \"nice strong techniques, but need more work on ukemi\". My comments were \"techniques were nice and technically good, but lacked \'fire and life\'. There were a lot of stuff in the head, but not enough was coming out in action\". Come to think of it, I wasn\'t even really like applying my techniques properly, it was more like trying to get the actions \'right\'. That\'s my problem I guess, my Aikido never works because I never really use the technique on my uke seriously. She ended by saying she wants to see me work on this area for my 4th kyu. (Despite, knowing that I\'ll be graduating this year.) At the end of it, I\'m happy I passed. But I wasn\'t happy with my performance at all. Quite sad really. I had wanted to do a lot better actually. Well, that just leaves me with a lot of areas I should be looking out for then.  This grading made me realise a bit more about myself. My non-aggressiveness during Aikido is not really the right attitude. I think like omote and ura versions of a technique, there are times I should be stepping in instead of stepping back. Causing a bit of pain to my uke is acceptable (so long as I don\'t make it really excruciating for them lah). Think my attitude during class needs a lot of work also. My mind drifts a bit sometimes.  Thursday I woke up to find my knees and elbows were a bit bruised and one had a small cut on it too. Think something on the mat caused the cut when I took ukemi for ikkyo. There were 2 marks on both wrists but I don\'t remember being yonkyo-ed the night before. Went school for a few hours in the morning and went back home. Had to prepare dinner for the house on Thursday. Very sian lah when it came to feeding them sometimes. I admit that I\'m not the best cook in the house, neither do I put in a lot of effort cooking my dishes. They have to be practical, easy to cook and edible. But 2 of the guys were making a bit of noise that my cooking lacked effort. What the heck?! Cook for you guys also need to listen to this kind of crap. So I just \'jokingly\' commented that at the most I can always pull out of the cooking arrangement and make my own meals instead of sharing. Whether they can tell that I wasn\'t exactly in a good mood I do not know, but the atmosphere sure didn\'t get out of hand. Later on that night, I had supper and again we came back to the topic about my attitude to food. Having calmed my mood down a bit I was able to jokingly point out that one of my housemate is too picky with food. I don\'t have an answer to this lah, maybe I\'m imposing my taste on them, but I think they are putting me in a similar spot by imposing their taste on me too. Anyway, I\'ll see how things go. I do enjoy their cooking but if I have to tax myself every week (even though it\'s just once a week), I really just can\'t be bothered at all.  Sorry if I sound like I\'m complaining but I think I need to spend a bit of time outside of my house and alone. My housemates are all nice people but I\'m sorry to say that the chances of staying in contact with each other will be almost impossible when I go back to Singapore. There are just some things we can\'t stand about each other and I don\'t see me being the type to stay in constant contact with them when we aren\'t really that close. The few bonding sessions we have had while in UK really doesn\'t make us that much more closer. It only exposes us to what habits the other person has but that does not mean we can accept it. I occasionally need to remind myself that there must be a ton of things that people can\'t stand about me and they\'re just keeping quiet only. That helps me to tahan their different lifestyle. Sad lifestyle lah in UK. Can\'t expect too much, just live with it and go home after that. Don\'t be mistaken that I\'m really having a miserable time in UK. I would still recommend anyone to try staying abroad for a period. There\'s a lot of things you will expose yourself to rather than cacoon yourself in Singapore. I think I just have too much stuff on my mind and I hope things get better this Easter holiday.  I went back home for dinner just now but a part of me didnt really want to stay home and listen to crap from people. So I decided to come to Uni to do some work and blog out my unhappines and grievances for the past few days. Heh :P   '),(26076,979795,'12,March,2003','2003-03-12 00:00:00','    I know I know, I was suppose to be working but I just visited the  urlLink Aikido Federation\'s website  and I saw the new addition of video clips and the kids class at Gombak. Arhahahahaah!!! Some of them were so funny. I saw one guy making a funny face when Freddy Sensei was trying to teach him iriminage. And the last 2 clips looked like they were playing games. What on earth was that all about? Hmmm ... strengthening their arms, maybe?  Just had to make a note of this on the blog. Cute fellas. Oh yes, Ms Ng was inside in the background and assisting. By the way, who was doing the recording anyway?   '),(26077,979795,'10,March,2003','2003-03-10 00:00:00','    Wow, another club in Singapore - Tendoryu Aikido Singapore (TAS). Their  urlLink website  is still being updated.  2 days ago, their secretary emailed to inform me that CAA and TAS were not related. She sounded friendly enough. Luckily I checked my account today. Updated the Aikido in Singapore blog but somehow the site seems to be down pretty often. My personal blog doesn\'t seem to have it as often. I might have to change it to something more reliable for Aikido in Singapore.  I emailed my Sensei to ask if I can record down some of our training sessions for personal interest. She hasn\'t replied my email yet (just checked prior to publishing this blog). But just in case, I\'ve already brought all my stuff to Uni since this afternoon (coz I\'m not going home first). Even had to loan the brand new 38 ltr backpack which Daniel (my housemate) just got for his birthday last weekend. Alamak, opening ceremony by me somemore. But the bag was cheap and good (by North Face). I\'ll probably go get one this weekend (if the sale is still on), otherwise I\'ll always have problem carrying my Gi and notes for training on Wed. Today was especially heavy because of the video camera and a tripod stand I loaned (from Daniel again :P). Wah, I tell you, look damn professional leh. Hehe. Even worst, heng heng everything I bring with me is in black. Black tripod bag, black backpack, black t-shirt, black gloves. Daniel said, all I need now was a black sunglass.  Really hope Pauline Sensei won\'t object. If I can, I\'ll digitise some nice clips and send to  urlLink Eileen  and  urlLink Gary . It\'d be interesting to show everyone how cuckoo I can be sometimes hahaha. Cuckoo-Tan ...   ...  OK better finish up some stuff before I go and see my supervisor in about 1 and a half hour. After that, it\'s Aikido!   '),(26078,979795,'10,March,2003','2003-03-10 00:00:00','    Just went to check out the price for a plane ticket from London to Denver International. Wah, expensive sia. Going to cost me about £350 for a RTN ticket. But I still need to check one more lobang before I know if its possible to make the trip. Chamz. Was estimating my budget and I may not have enough to go both places. Haiz ...  Maybe if I can sell a few more phones I might be able to make it. But I think if ultimately I really can\'t. Then I might just stay on in UK to do an uchideshi program here instead. Need to get my Sensei\'s approval of course.  So much details to check out and confirm before I can start making a choice. Ma fan!!!!   '),(26079,979795,'09,March,2003','2003-03-09 00:00:00','    Eh yar huh, nowadays Shuhua\'s always going to ECP. What for anyway? Looking for some charbor to sian there issit? :P Eh don\'t forget there\'s a lot of army people there leh. Remember to wave your pink IC around and throw it in their faces shouting \"  Ar hahahaha, chow recruit / NSF, wahahahahha ... ORD liao ORD liao!! Me 3SG (NS) Foo now ... wahahahah you suckers!!\" Eh wait wait, I still got 4 mths to serve, scarly you\'ll see me there one of those days.  Yea, well even though I was kinda in a bad mood, but I think I might have overeacted. Coz when one of the girls threw an old Chinese New Year decoration at me, I kinda went a bit off the edge and just said, \"Eh please, I feel like scolding some vulgarity now hor, ok.\" Think that kinda told the whole room what kind of state my mood was in. Of course, now I think of it, I\'m just glad I didn\'t. She didn\'t like do anything really bad. And even if the others were a bit noisy, I think it doesn\'t really warrant an overeaction on my part. I dunno, sometimes you really get the kind of effect like Stephen Chow\'s movies Xi You Ji (Journey to the West) part 1 and 2, where Tripitaka was like a freaking fly always buzzing around the ears. I mean, c\'mon lah, even Guan Yin in the show tried to trottle him once. To those who\'ve not seen it, let me try to illustrate this point:  Wukong was trying to steal some heavenly device and got caught by Guan Yin. While meting out punishment, Tripitaka arrived at the scene and started pleading for mercy on Wukong\'s behalf, but he has this stupid habit of talking too much (in a very irritating sort of way). He says: \"Wukong, why didn\'t you tell me you wanted the magic box? I mean you should have told me you wanted it. If you\'ve told me you wanted it then I would\'ve given it to you. You didn\'t need to steal it. If you didn\'t tell me about it, then how can I give it to you?... \"  And somehow, the sentence isn\'t quite over, he still went on and on and on and on. If you actually translated it into Chinese, you can tell it\'s a long sentence. Especially irritating since all the sentences contained basically one meaning only. Just that it\'s magnified 10x. You probably just need to go watch the show yourself to know how funny it really is.  Anyway, think later I\'ll apologise to them about my mood lastnight. But I still think it was a bit irritating. I like to watch my own show and movie quietly. Anything you don\'t understand, I think can be settled at the end of the show. Of course, if your memory isn\'t too good, then that\'s probably going to be a problem. Think in this case, just remember that it\'s a S-H-O-W or a M-O-V-I-E, and just accept it as it is. That\'s a good enough reason. I also really can\'t be bothered to put too much effort to convince others about how I interpreted what I\'ve watched, unless the topic deserves a good discussion.   '),(26080,979795,'04,March,2003','2003-03-04 00:00:00','    OK ... I\'m like in a really bad mood now. Was watching TV just now, and I really can\'t stand it when folks just keep asking questions throughout the show. I\'d really prefer it when everyone can just watch quietly and just ask questions during the commercials or after the show. I don\'t care to what extent that you understood the show so far or anything. If you really wanna ask it. At least ask someone who is willing to answer, not distract others so you can get your answer. Maybe such demands are unreasonable, or maybe I myself would do it sometimes, but really ... I\'m not in the mood to self reflect now. When I do cool down, I\'ll do it. Sheesh ... all this shitz over one stupid televesion show.   '),(26081,979795,'04,March,2003','2003-03-04 00:00:00','    O yea, about Tamura Sensei, my Sensei told me some things about him before. Coz I was watching the summer camp they had in 2000 and Tamura Sensei was 1 of the 4 instructors there. Can\'t remember what it was he was teaching but I think it had a lot to do with feeling uke. I think a lot of the yudanshas at the camp that he was correcting didn\'t quite get that as well. You can tell they weren\'t reacting very well a lot of times. Heh ... guess we\'re all learning. And I heard Tamura has like one of the longest warmups ever. Sensei Pauline was commenting that they planned him in a certain slot because they know that was one of his habits. Heh, sometimes he\'d look at everyone there and say \"Oh they don\'t need any rest.\" but actually may be some of them might think \"Oh but we do\" instead. XP A lot of Tamura Sensei\'s warmups I heard Sensei Pauline mention sounded really really familar to Freddy Sensei\'s, maybe that\'s where he picked them up from. Interesting. One more farny thing was that Pauline Sensei said sometimes Tamura Sensei would pretend that he didn\'t speak any English. Well it\'s kind of a tradition thing to have some kind of speech at the end of the camp and Pauline Sensei reminded Tamura Sensei a few times about it. Then I think after trying to act blur for quite awhile, he said \"I\'ll give 1 if you\'ll give one.\" So my Sensei went up, clap her hands and said \"Excuse me everybody, Tamura Sensei says that he has something he\'s just dying to tell you\" XD~  Self-nikkyo? I know I can\'t do it. Best I can do is do more of those wrist training push ups instead. I\'d like to try some koshinage and sumiotoshi though. Didn\'t do those much before but I think you\'d most probably need to know some really proper breakfall if training with someone like me. Beginners in these techniques can\'t actually make it very painless for ukes. :P   '),(26082,979795,'04,March,2003','2003-03-04 00:00:00','    Shitz dah, just before I started this blog, spilled water all over myself. My housemate just made mee goreng for us, y\'see. Had to drink water and ... splash. \'Puo1 Leng2 Sui3\'. Anyway, let me just continue while I wait for my singlet to dry.  Wooooo  urlLink some people  reading technical stuff huh. FAQ somemore. Just telling me to get into a book will put me to sleep. In fact, just thinking about it, makes .... me .... a. .....   ... little ....... zzzzzzZZZZZZZZ .....   ...  This morning was supposed to go Chinese Quaters to \'yum cha\' and have some Dim Sum with my housemates. But after buying groceries, I just felt really lazy to leave the house. When those that did go returned, I wished I went as well. WINTER SALE!!! Argh, David bought a nice 30ltr backpack, quite cheaply too. Hmmm ... maybe I\'ll go next Sat or something. Just hope the sale is still going to be on. I\'m so used to being around Uni and home that it slipped my mind about winter sale season.  Feels kinda weird not training Aikido last week. By the way, I simply do not understand the  urlLink slamming exercise  you were referring to. Maybe I\'ll do act it out in front of my PC when I\'m more free. About moving angmohs, well my Sensei (being of small built) really have no problem moving anyone of us who are bigger than her. And you gotta remember the european guys are also a whole lot bigger than me. And er ... I\'m sure the fact that she\'s a 4th dan probably has something to do with it as well. Dunny thing about slamming, I\'ve only heard the term used in Singapore. Over here, they use the term taking breakfall on quite a generic basis. Only diff is how soft/hard nage executes and nage takes breakfall. When I do gyaku hamni variations of kotegaeshi, the static versions are always very weird. Think really need some kinda giving in from uke if nage don\'t do it well. But once you bring it up to a more dynamic approach like a tsuki or sho/yoko-men, think can do it more smoothly coz there\'s a flowy feel then. If uke don\'t let nage do it ... think might as well use some other technique for effectiveness. Doesn\'t mean it\'s always the nage\'s fault. (I get that quite a lot ... from uke\'s perspective I mean haha). By the way, where do you think you stand in terms of technique and ukemi in the whole class? I\'m sure you see some cuckoo things by some people as well right, pretty much like back home. Think I\'ll be doin nikkyo this coming week. Even if I don\'t I think the next one will. Coz Sensei gonna spring a grading test on us sometime this term (within 3 weeks). Panicky sia. Not ready leh. Also scared I may need to skip training in my last week coz projects are gonna be due then.  I understand what  urlLink you  meant by \"throw yourself when the nage was just about to throw you\". I have this problem with over anticipating my nage, like ALL the time. If they did something unpredictable, I doubt I can adjust myself well enough to take ukemi and come out unhurt. My Sensei tries to help us by making us do randori in the last few mins of almost every class. One it helps us recall what we\'ve learnt for the whole 1+ hr, two it helps us take ukemi at a more unpredictable level. I think it\'s good and helpful for me but I will never be confident. My memory isn\'t very good. Even if I can remember a bit, my hands and legs rarely follow what I want. My brain just can\'t multi task at that level I suppose. So I always start after seeing the first few students do some techniques. Heh heh, monkey see monkey do. It\'s only after a few gos, then maybe I\'d change the order a little. Yupz, as Tommy said, my Aikido also \"has no life\". Gary\'s right, Tommy\'s a good partner to practice with. There was this period where I almost almost always heng heng paired up with him for iriminage. It was like 95% of all my iriminage practice. Dunno why so coincidental. Haha I recalled coz after he grabs my neck and throws me, he sometimes commented \"wah, you very wet ar you.\" and I\'d just laugh. Damn farny man. Miss the irimanages with him hahaha. O yar, and I also enjoy the after-class-sessions with Jesse. Err ... \'cept that one time before I flew off to UK where he kinda damaged my specs. Now a bit scared hahaha XP  Hey the grabbing part y urlLink  ou mentioned I like what my Sensei does. She always tell us that when we grab nage, pretend there\'s a 50p coin between our palm and nage\'s arm. Whatever the nage does, uke needs to follow in such a way that the 50p coin won\'t drop off when you are doing techniques. Of coz there were always some occasions where we used a real 50p coin to try it out. Works a whole lot better than normal but just remember not to depend on it too much such that uke becomes really artificial. Err, and just in case someone who reads this don\'t know it, remmeber to let go of the arm when you do get thrown. Otherwise you\'ll usually find that while your body is somewhere else but your arm is still remaining on nage\'s arm ... gross out man. Anyway I think it\'s especially good when doing sitting kokyu-ho. A lot of people I do it with don\'t practice the grabs properly at all. Over here, my classes place a lot of emphasis on this. Think it\'s a trademark with my Sensei. In Singapore, almost none of the 6th kyus do it, some of the 5th kyus might though. Anyway, different Sensei - so different style and empahsis also. Sometimes when I grab my Sensei\'s wrist, she will say \"I want you to hold it, not stop blood flowing through it\" hahaha ... think that\'s my point. So I usually try to tell myself consciously to grab lightly to the point of just a touch kinda feeling, but with the palm having almost full and flat contact to her arm.  The part about ma-ai (distancing) is something I\'m not doing right either. It\'s even harder when I want to apply atemi properly. My Sensei points out to me quite often that when I atemi nage, I should make sure I can at least touch him. Otherwise nage will need to shift in more just to do the technique on me. I\'m not helping lah basically. I try to remind myself that but I got no 100% effective enough reminder at this moment. With my kinda atemi, my nage could probably just ignore me and step back and walk away. Heh, I\'m lousy lah. But in Singapore, a lot of people I practice with don\'t quite get it leh. There were more than one occasion that I\'d just not do my technique and just stand there coz my uke isn\'t even touching me. I didn\'t want to point out too obviously but they just hardly get it. Think the atemi I like best is Syed\'s. Strong, stylo, well at the very least it sure feels and looks like his arm is really gonna hit me (not just graze or touch me only). My Sensei will usually just stand there and not budge, then gimme a sian expression. If I still don\'t get it, then she\'ll say something. :P There\'s a Greek guy in my Uni class who\'s a 5th kyu. He has really nice ukemi and soft technique. Trying to take note and learn from him. Especially his ukemi is like ultra soft and I like the way he commits in his atemi. Soft yet commited enough for nage. With him, I know that if he falls weird, it\'s probably because I\'m making him do it that way. Me again ... duhz.  Err dun really wanna comment much about the Iraq/UN/US/UK issue here, but I think since US/UK are gonna commit in a strike anyway. They might as well get it over with. I mean they\'ve already expressed that no matter what other countries say, they\'re still going for it what. Everybody don\'t want the war obviously, but since there\'s no holding back so why ... hold back. Go lor go lor.   urlLink Heat wave ? Well, these last 2 days just got wet. REALLY WET AND WINDY. Doubt we\'ll get any good snow now. But can\'t complain lah. Weather has been really good since I got here. I should have guessed it wouldn\'t last. Heard things are gonna get wetter and \'windier\' (??) coming weeks. Sianz. Hate walking 30 min to Uni already, now it\'s gonna be walking to Uni wet.   '),(26083,979795,'04,March,2003','2003-03-04 00:00:00','    OKOK I know I said that I won\'t be blogging for awhile but I just wanted to make a point. My project IS STARTING TO WORK!!!!! WAhahahahahahaha ... Woohooo!!!!  OK lah, not like working completely. But at least the stupid vehicle is moving (in the wrong direction and along the wrong axis) and I can show my supervisor something tomorrow. Woohoo!!! But think tonight still have to burn midnight oil (drums of them). Darn it!   '),(26084,979795,'03,March,2003','2003-03-03 00:00:00','    OK, it\'s official. Back to bachelor status now. Had a bit of talk with my ... ex. Well, at least we\'re still friends. Time will tell if we can remain good friends as we were before we got together. Anyway ... like one of my housemate mentioned, this period is really one of the worst possible time possible for us to put our relationship to the test. Everyone\'s so busy and all ...  Thanks hor, brother Shuhua, for your comments on my blog. I think I managed to read all of them unless you left some on the blogs which dated about 3 months back. Yea, well I didn\'t really tell anyone about my girlfriend (ex now) when I was in Singapore. Probably only the 3-4 of my NS kakis and Gary and Marcus/Ivy. Gary knew because we see each other often enough during Aikido, plus I always drop by his place regularly. Marcus because I had mentioned to them before the last time we met before he went to Sweden. Yea, my blog does look like a training manual of sorts. It basically documents what I\'m going through daily. As Aikido is a huge part of my life now, it\'s not surprising that most chunks of it are Aikido related contents, an Aiki-diary if you will. Anyway my VRML project is very slow and status is quite bad now. I\'m gonna be skipping Aikido to make up for lots of lost time. May need to spend a lot more hours in school from this Mon onwards. Ai pia jia eh yia! Anyway, own time own target just feel free to drop by and check me out. Leave me a message on ICQ too if you like. But sometimes I maybe asleep or in school. Might have left my laptop running to download some stuff.   urlLink Una-san  (??),  urlLink Aikiyoda  did do 200 cuts a day while he was in Singapore. In fact he was the one who psychoed me into it as well. I did it consistently until I came to UK. Took me a long time before I got my hands on a bokken. I doubt I can do anything physical exercise consistently until my 1 month Easter break. But by then, it\'s really time for exams revision. Think Gary said he\'s gonna be getting a bokken there. Well he\'s definitely going to be getting another set of gi. Damn sponge monster still dirties the dojo mats everywhere he goes. Ar duh...  OK time to get back to my second essay and project.   '),(26085,979795,'01,March,2003','2003-03-01 00:00:00','    OK just an update.  I\'m going to be really busy for the next 4 weeks. Probably won\'t blog much unless I\'m like really pissed about something, or really need to note down something. I\'m probably not going to Aikdo this week. Just to catch up on some work first.   urlLink Foolz!  I can\'t do my 200 cuts a day lah. I\'ll probably scratch my ceiling paint, or my bokken, or even both, if I did.   '),(26086,979795,'28,April,2003','2003-04-28 00:00:00','    My goodness. Made the same housemate cry again lastnight. I don\'t understand it, it sure looks like I don\'t practice enough tact when I\'m talking to her because the same thing seems to be happening over and over again. But it\'s weird because we\'re quite alright friends in the house.  Basically 2 days back, I had pointed out that she tends to forget to clear up some of her stuff and it was becoming a sort of a habit. I wasn\'t really nasty about it. Her bf did mention something as well. Maybe she felt we were picking on her so she kinda just lost her temperjust a little because you could see her face turning red and her pitch increasing. Well, I have to admit we weren\'t picking on her lah. It was meant more as a joke and a reminder. But I was very surprised that she got the impression of \"why she\'s the only one being reminded instead of the other people who\'ve been doing it as well\". She has a small point to this, but the way I see it, when a problem comes up, it\'s natural for me to approach the people I feel closer to as I think that together we can help to make things better. I\'m thinking she feels picked on lah basically. But I honestly don\'t feel that I\'ve said anything or outstepped my boundary. I\'m sorry about whatever unpleasantness there might be but I really don\'t see an apology coming from me over this matter. I don\'t think she\'s taking my intentions the right way as I wasn\'t trying to make her the scapegoat for anything. Past 2 days, I noted that she\'s always making statements like \"I\'m going to use this particular cup, instead of xxx\" => personal cup => can wash at personal convenience, or \"OK, you said you\'d wash it. I didn\'t force you to do it huh\" ... Sigh ....    Granted, we\'re all imperfect and forgetful at times. I also tend to forget to do things like clearing up after I\'m done. I\'m grateful for the next person who comes along and does it for me (which may / may not be all too often), but I really don\'t think that the next person should clear up after me if they feel that they are being taken advantaged of. There\'s nothing wrong to just leave the old dishes wherever and just point out to people that they\'ve overlooked clearing up. It\'s a good reminder, ONLY if you don\'t get too nasty about how you \'remind\' them.  This incident brings to mind the situation we had (have??) at the BV dojo. Where people tend to just leave and not keep all the mats. In an ideal world, if everyone just kept one, it would make things so easy. But as things turn out, the world isn\'t a fair place. The sooner we accept that the easier we can proceed with doing our job.  It\'s alright to NOT do some stuff, just don\'t be too difficult about it. In contradiction, don\'t push others too far by taking advantage of them, if we just opened our eyes a bit now and then, we may realise that there are rubbish or dishes that disappear even though we never seem to clear them daily.  Also realise 1 final point. I appreaciate what some people have been doin for me in an invisible sense, maybe it\'s high time I showed some form of appreciation instead of keeping quiet about it.   '),(26087,979795,'27,April,2003','2003-04-27 00:00:00','    OK, I\'ve sent in the email to the Singapore Police Force enquiring the specifics of a career with them as a Senior Police Officer.  I think bottomline is that I don\'t want just any deskjob for a career. I definitely like something which is a bit more physical than that. I just hope investigative work is more my cup of tea.  I\'ve got my fingers crossed on this now. Really hope things turn out well and I get some kind of promising news from them asap.   '),(26088,979795,'26,April,2003','2003-04-26 00:00:00','    Things I got for my birthday:  On the eve 2 of the girls baked muffins. OK lah, wasn\'t meant to be my birthday thing but it was still nice.  On my birthday, they made cookies and I took them. 1 of them made me a bowl of \'Mian Xian\' (mandarin) / \'Mee Sua\' (hokkien dialect). 1 baked a chocolate cake (Mississippi Mud cake/pie, if I remember the term correctly). Had a well spread dinner which was ultra filling. Basically too much food for even the likes of me to finish them all. But it was good.  Got 2 birthday cards. 1 stuffed monkey, and a Bible (graphic novel / comic). Everyone knows I get tired of reading the conventional Bible, so ... heh :P The monkey thing was coz I mentioned before that it\'d be really kewl to wear a gorilla suit and just act as a monkey. But they said suit rental for just 1 night was just so expensive. The monkey\'s cute.  It was a different experience this birthday. Very much so from my 1st year here as well. Definitely more fun then the ones I\'ve ever had in Singapore. Love the people here.  Quarter of a century old ... think I\'m in a quarter-life crisis already.  This will probably be my last post on anything even remotely related to my birthday. Cheers.   '),(26089,979795,'26,April,2003','2003-04-26 00:00:00','    25 years old today! Quarter of a century old!! Oldest person among everyone in the house now!!!  Was feeling a bit crappy yesterday until the later part of the night. The (3) guys removed my top and carried me outside the house. Thank goodness they were decent enough to leave my bottoms on. Of course, could be the fact that I didn\'t quite struggle against them. More you resist, more tempted I\'m sure they\'d be to carry out a more ... er ... vicious \'sabo\' session :P  Hmm being my birthday and all, my mood today has been good. Weather\'s all nice and sunny and I hope it\'ll remain like that for the rest of the day. Sigh, I think the people around me also has a sorta positive impact on my mood as well. OK, had to go up to my room and \'act blur\' about the cake they\'re going to make for me today. Touched. Really. Usually I\'d not even bother about my birthday. Last time I tried asking people out was sometime ago in Poly. Wasn\'t particularly fun then either with people being late and all that. Shucks...  Well, let me just set some record straight about stuff that I\'ve blogged about previously. Some may seem downright nasty and all that about people I\'ve came into contact with, that\'s why I\'d really like to re-evaluate my expressed opinions at the time of blogging.   urlLink  Monday, November 25, 2002   About the websites of Aikido clubs in Singapore. No offence intended but my opinion still stands, but please bear in mind some of these sites have seen some other changes by now.   urlLink  Wednesday, December 25, 2002   About the 2002 X\'mas dinner at my accoms. Think whatever reaction I had then was clearly exaggerated to a certain extent.  I do still prefer that dinner can be an affair where each and everyone of us are able to sit down and eat comfortably. Well, I suppose a certain level of discomfort to squeeze slightly more people around the table is still acceptable lah.  I\'m still unhappy that I do not speak up about things that I may feel pretty strongly about. But I\'m also now aware that staying with other people is about a level of acceptance and telerance with each other. There are things people can\'t stand about me as well. Give and take lah.  Part of the reason why I wasn\'t too ecstatic about the party was because I didn\'t feel comfortable mixing with the crowd. Granted, I\'m not a very socialable person and I do sorta \'pick\' the people I mix with. I think 1 thing I need to work on is being more sincere like how Daniel is. But of course, things like this calls for some kind of moderation too as that doesn\'t mean mixing with every Tom, Dick and Harry (no offence to anyone by these names). Mustn\'t let my mood get in the way of my behaviour too often. Think being a certain level of diplomacy is involved in maturing.  Important note. I don\'t mean to direct any unhappiness I felt then at any of my housemates. I\'m glad things happen, for whatever reason. I learn that little bit more each day.   urlLink  Friday, February 21, 2003   A bit of comment on the girls\' volume in the house. Realised that us guys get pretty noisy as well. We\'re no different / better. Just give and take and be considerate to everyone. But of course, we need to be frank and tell each other when anyone feels bothered about the noise level. Everyone in the house is reasonable. To be the \'nice person\' and make yourself take things lying down is kinda unnecessary because if the other party knew that he/she/they was/were a pain in the butt, the party involved will just control a bit and make the necessary accomodation. Everyone in the house is nice. There shouldn\'t be a problem speaking up and talking to each other about how we all really feel.   urlLink  Monday, March 03, 2003   About my breakup with my ex-gf. Think it was all for the best. Wish we can be better friends then we are now but I think there is a certain amount of resistance in both of us at this point. Wish things weren\'t like this but until each of us start to think and behave differently, there definitely won\'t be any improvement. At the very least, we are still seeing each other on and off when she comes over for dinner.   urlLink  Month of March, 2003   Realised my mood was really atrocious and unforgivable the period just after my breakup with my ex-gf. I seemed to be taking out any resentment I felt on people around me, most often on 2 of the girls in the house. I\'m so sorry for that but I\'m also really appreciative of how tolerant my housemates were. Self reflection ar, ah-boy!   urlLink  Friday, March 14, 2003   About cooking. Really need to point out that I\'m one of, if not the, laziest person in the house when it comes to food. But as time went on, I realised a point Justin pointed out, it\'s a real moral booster to be able to look forward to coming back home to a good dinner. No doubt I don\'t feel as strongly about food as some of my housemates do, but I guess putting a bit more effort into cooking wouldn\'t hurt anyone. I guess sometimes I just feel at a loss about how and what to cook because I obviously don\'t cook as much as the other guys. Maybe asking for a bit more help would be a good start. But I\'m figuring, with the exams coming up soon and so little time left that\'d we\'re staying together, there may not be much of a chance. See how lah. Anyway, just liked to point out that irregardless of how cheesed I might have sounded in that blog, I\'m really quite alright with everyone and their expectations. Like I mentioned, everyone here is reasonable. I\'ve not came across a really unreasonable request yet.   Sunday, April 06, 2003 [ Try  urlLink link 1 , or  urlLink link 2  ]  About meeting this new \'someone\'. I can\'t say for sure it\'s not a crush, neither can I say that it is. I can only say that I really like what I\'ve seen and know so far. From my previous relationship, I\'ve basically came to realise a lot of things I really \'do not want\' in a partner. With this new girl, there are none of these bad qualities and tons of good ones. Naturally I\'m greatly disappointed she\'s not the least bit interested. Neither can I say for sure that things will definitely work out between us in a positive way, but if she was willing, God knows, but I\'ll give it my utmost and best shot at the chance.  These days, I\'m constantly blowing hot and cold around her. Want to be a good friend but when I feel that I\'m getting too close I really need to cast myself a bit further just to keep my emotions in check. I\'m still really happy she\'s not been mean or unkind to me so far. In fact, sometimes I don\'t feel any of the awkwardness of the situation all thanks to her. It\'s hard to control these things I suppose but I thoroughly respect her for her decision. I know she has her own problems and as much as I\'d like to share her burden, I know that it\'s not my place to decide if she has to share things with me. In my poly days, I learnt a particular lesson with another classmate of mine in an almost similar situation: - If I can\'t be a help to her current situation, the very least I should / could do is not  BE  one of her current problems.  I still stand by this belif even after so long. I\'m no saint or angel that I\'d try to be so wonderful or \'wei da\' in Mandarin. But I just think it\'s really worth it sometimes to make some people happy at my own expense. I may be criticised for being a fool or a love-sick idiot but there\'s no changing the fact that given another 100 times, I\'d still choose to do the same in a similar situation. There are just some people that I know who are truly worth making an emotional sacrifice. I basically do not want to jeopardise the friendship that\'s available now by even attempting to bring things up to another level (not that there\'s even a hint of hope). I\'m happy with the way things are now.  A summary of things now, I love all my housemates. Nice wonderful people and I\'ve learnt so much from each of them. They\'re all so unique and special in their own way. I love Aikido, and looking at the way things are, it\'s the only thing which never gives me any pressure in my everyday life (well, that is, if you overlook the occasional physical bumps and bruises one may get during training).   '),(26090,979795,'14,April,2003','2003-04-14 00:00:00','    The eve of my birthday today. Got a few well wishes here and there. Got a call from Gary from US as well. It was nice to hear a familar voice again.  Just 2 days back, I played some basketball and got a knock on my glasses. Now it feels a bit dodgy. Sigh, no choice lah, people will probably see me in contacts a lot more often these days.  Today was relatively quiet. Was considering going swimming at night coz 2 of my housemates were playing soccer in school at the same time. But in the end, I just decided against it. Got a bit of a throbbing headache now. Guess, it\'s better this way. 1 more hour to Friends. A bit sad lah the life here but I think it\'s coz exams are coming lah. So things just seem a bit more boring. Of course, there\'s a lot of studying to do. Sianz.  Just thought I\'d best update some details on what\'s been going on before my papers start.  Oh yea, 1 fantastic piece of news. Aikido is still on this term in Uni and at the dojo. Problem is I can\'t go too often because of the exams. Argh! Looks like I can only go for the coming week and the 1st week of Jun. Depending on where I\'ll be going after all my exams, I may / may not be able to attend the last session before term closes.   '),(26091,979795,'14,April,2003','2003-04-14 00:00:00','    OK, have to make a note that I had apologised to Juliana yesterday 1st thing in the morning. She seemed fine lah. Looks like I was just caught between the 2 lovebirds. OK not so guilty liao. Basically worked on my report on and off the whole day until about 4pm. Went running with Justin, David, Juliana. Coz of the crying thingy the night before, I owe Juliana 15 rounds but only managed either 8-10 (coz I miscounted) rounds. Nmind, try again until I get there.  After that we had a fantastic game of frisbee. Woohoo!!! Played 4 on 4. I wasn\'t particularly good and I got tired really fast. Towards the end I got a cramp in my left calve/calf (watever lah) muscle. Basically that kinda meant the end of the game for me. Went back for dinner and pia my report again. Finished and went to bed before 2am.  This morning, I got up early and touched up some stuff. After that took a cab to Uni (wet weather and my left leg still quite tight from the cramp mah). Time now is 9am exactly. Library should have just opened. I\'m gonna go submit my report and get it over with. Damn! It feels good now that it\'s over. Going for a quick swim at 12pm. Hope my leg holds up today. Going home after that and I plan to watch the Steven Seagal aikido documentary - Path Beyond Thought. Caught snips of it. It was daaaaaamn good.   urlLink Eileen , I\'ve already transferred the clip to Gary. Thought you might be interested too. It\'s awe-inspiring.   '),(26092,979795,'13,April,2003','2003-04-13 00:00:00','    Goodness, I directly / indirectly made Juliana cried this evening during dinner.  I was washing the dishes when I heard some topic the others were talking about, one of the comment was funny (to me lah) so I laughed out like really really hard, long and loud. And she just left the table, supposedly all teary-eyed. Sheesh, no doubt lah I didn\'t like make the comment lah, but I think it felt like I was laughing at her directly in her face or something. Sure, some of them said I wasn\'t the one to make the comment so I wasn\'t like really responsible. Still, I think to say that I had nothing to do with it is absolutely irresponsible. Sighz, ... guess I\'ll have to apologise to her tomorrow. Will try to finish my report by tomorrow morning so I can stay awake overnight to apologise 1st thing when I see her.  Goodness, feeling a little bad.   '),(26093,979795,'12,April,2003','2003-04-12 00:00:00','    I just recalled some clips of Steven Seagal\'s Aikido. And all of a sudden I just wondered about the best form would be having no form at all. I look at my stance in Aikido, I feel like I\'m preparing myself for an attack either this way or that. For instance, when I\'m in right posture, I\'m basically expecting an attack to come at me a certain way. And when it doesn\'t, I get all mixed up and everything just goes out the window. I suppose the form probably refers to something more metaphysical and mental.  I noticed Steven Seagal has a stance in which he stoops down a bit and relaxes both his hands. In some scenes, he has his hands fluttering all over the place so quickly I can\'t even catch what he was doing. I guess, that would include a distracting element to someone who\'s attacking you.  I for one would love to learnt how to achieve his style of iriminage. Bam bam ... and nage would be on the ground already.   '),(26094,979795,'10,April,2003','2003-04-10 00:00:00','    Have to make a note that my ex came over for dinner few days (Wed I think). She brought her junior along as well. Think that\'s a good thing, coz it could mean that she\'s not unhappy about the last incident anymore. That\'s good. And I do have a slight ease of mind now. But like Eileen mentioned, I don\'t dwell on it too much. Afterall I\'ve already gotten over the occasion by opening my house\'s door to her anytime she may want to makan here. We (I was with Justin and Daniel) were doing our warmup jog up a slope from our place to the park nearby when we met the both of them coming down to our place. Couldn\'t chat much then. But everything seemed ok. Even after dinner things seemed pretty alright. Just that they left kinda early after dinner instead of hanging around.  Went running today again. It\'s turning out to be a guys thing in my house. All 4 of us went for a run at the park. David, being his 1st run, had a bit of a hard time. But I can tell that he has quite a bit of power still in his legs. I think if he ran a few more times, I\'ll easily be the slowest runner in the house. I guess that\'s the difference between Officers in the SAF and normal foot soldiers like me. OCS training is something to be reckon with. Which kind of makes me wonder. At 25 years old when I graduate, will I be able to take the training in the Police Force. Hmmm ... well, guess there\'s no point pondering over it too much. Just keep training and bring my fitness level up first.  After the run, I was feeling fine but Justin and Daniel had a slight headache. After dinner, I feel a slight groggy feeling. Could be a fever or something going around the house now. Think I\'ll pop either 2 panadol or 2 raw eggs in again. Need to keep my health up the next few days. I need the extra energy to complete my final report before Mon.  Now that David\'s decided to join us for a bit of running. I think after Annie (housemate) gets back from Spain with Juliana (housemate), and we\'re back into playing frisbee again, I\'ll see if David wants to do a few forward rolls while we\'re at the park. I can do it in my backyard lah of course, but I think the daily bokken exercises I\'ve been doing has worn away the ground quite a fair bit and it looks a bit too dirty now. Whereas at the park, all we need to do is pick a small corner and we can roll around on the grass. Err ... only thing I think we need to look out for will be dog shit. It\'s a park afterall :P I\'m kinda keen on training my forward rolls with some obstacle in front. Might need the other guys\' help to curl up in front and let me jump over. And I\'d love to try jumping over the 2 taller guys and train my rolls into getting some distance and height in as well. Well, the girls are due back tommorrow anyway so I guess I can look forward to my rolling exercises soon.  O yea, talking about this, just few days ago Justin and I were running and we saw this guy standing in the park practicing some kind of self defence moves or something. Err ... OK lah, not trying to like put down the guy\'s efforts or anything but to do some things too openly in public is kind of weird. And I think he does look like those half half pattern who hasn\'t exactly been learning much for a long time. I mean, whether I\'m in Singapore or UK, I only practice my Aikido in a private compound. Sure lah, the neighbours may see me, but I usually pick the weirdest hours so as to minimise the no. of people who might see me. I mean, I\'m not that good anyway so I would really hate to expose myself more than I really needed to. Only my housemates and family are like the only ones who see me practice regularly, but they know me enought to be able to tell that I\'m being serious and just wanted to be alone when I\'m concentrating during those times.   '),(26095,979795,'10,April,2003','2003-04-10 00:00:00','    Just went for a run. I think it\'s either the ice cream Venetta I\'ve been eating these past few days taking effect on my fitness, or the weather is just getting cold, because I seriously had a slight problem running these past 2 sessions. After the run, I even felt a bit of cough coming up. Just took my grandma\'s remedy, 2 raw eggs. Justin and David were disgusted by it. Haha. Must admit lah, the initial taste is kinda hard to take down, but it gets better after that. Anyway, so long as it\'s effective, I don\'t mind the awful taste for a few seconds. Let\'s hope my stomache can take it tonight because I haven\'t taken this remedy for some years and I\'m not sure of my stomache is used to it now. Well, maybe some water to help it go down later.  I emailed a complaint to the bus service and I\'m hoping to get some kinda reply soon. But even so, I think I\'ve cooled down enough to be able to forget it lah. Afterall, the thing is over already. Sigh ...  OK I gotta go work on my report liao. Final year project write up, ARGH!   '),(26096,979795,'09,April,2003','2003-04-09 00:00:00','    Damn the bus service over here!!! ARGH!!!! R$&^$#())(*&(^&%$#  I\'m suppose to have be in Aikido right now but because of the stupid bus service everything went haywire.  Left my house at 7:15pm to wait for the bus to take me to the dojo. Waited over an hour for it but it never turned up. Mind you, because it\'s daylight saving now, it was still mighty bright when I started waiting. After 1 hour however, it was difficult to keep my spirits up as the temperature got colder and the sky got darker. Class was supposed to start at 8pm. I thought I\'d just have to make do with being a little bit late for class. However, at 8:15pm I decided I can\'t wait any longer. I dialed for a cab which arrived 5 mins later and I reached the dojo under 10 mins. Unfortunately, Sensei had just locked the door to the building thinking I wasn\'t gonna come.  The situation at the dojo was this, it\'s basically something of a community centre. But the instructors there do not have the keys to the place. Instead, the place is open at a certain timing and classes are scheduled 1 after another. Which means there\'s always someone there to open the door for you. As my Sensei does not have the key to the place, she can\'t reopen the door again. And she had only closed it 2 mins prior to my arrival.  ARGHH!!!!! ... I don\'t blame my Sensei for this. Anyway, she was so nice about it, she reimbursed me my cab fare back, even though I said she really didn\'t need to do so. She said I\'m just a student, so money and all that was kinda important for me coz I\'m not earning anything yet. (Sigh, nice hor.) I really blame the bus service because I suspect the people at the bus depot were probably just slacking around. I know for a fact there\'s a big match between Madrid and Manchester United. That probably account for the absence of the bus during this hour. I\'m gonna find a number and make a really formal complain 1st thing tommorrow morning.  There\'s a lot of reasons why I\'m pissed. 1. Sensei Pauline had shifted about a month ago and she basically needs to take an hr\'s drive just to come down to teach us and another hr\'s drive just to get back. (By the way, that\'s also the reason why she\'s looking for another dojo location to change to. Meaning she\'s probably not goin to be teaching University\'s class in the near future. So even if I do decide to return here for my Masters, I won\'t be able to train under her anymore. ARGH!!!) Basically she wasted her trip down (much like the rest of us) today. Because of the distance, with due consideration to her, we, the regular students, always try our best to inform her of our attendence in advance to prevent her from making such wasted trips down to the dojo. 2. Sensei paid for my cab fare though it\'s not her fault. 3. Today was the LAST class for the Easter holidays. Sure there\'s a class once term reopens. But besides turning up for the 1st week of May, I probably will be absent for about 1 whole mth due to my exams and all that. After which, if I\'m lucky enough, I\'ll be able to attend 1 more of her class before she goes to the US (either on holiday or for some Aikido camp thing there I suppose). 4. I also caused Angelo to make a wasted trip down to the dojo. Sure he stays nearer compared to Sensei but he still made a wasted trip down and he had to gimme a lift part of the way back to my place, something which he does every week. (Sigh, another nice person.)  All because the STUPID BUS NEVER CAME!!!!! ARGHHHHHHHH!!!!  After getting off Angelo\'s car, I was still feeling kinda frustrated. While walking the rest of the way back home, I kept looking around, probably hoping some fools would try to jump me and try to rob me or something. Just so I can use some Aikido on the fella. Luckily, nothing happened. Must admit the walk did me some good to cool my head down a little. I entered my house and saw Justin and David watching soccer (damn Madrid vs Manchester United). They were surprised that I got home so quick. Explained what happened and went on to make myself a cup of tea as well as attacked a piece of chocolate flavoured Venetta ice cream cake or watever you\'d call it. (Mmmmmm ...) Really wanted to practice my 8 directional exercise (happo-giri) but our backyard had no yard light and I thought it\'d be too dark and cold to do it at night anyway. Brought all my stuff to my room and passed Lynette. Explained the whole deal to her again and came back online. Met Gary on ICQ and made noise about the whole thing again before I decided to blog things down. After I post this blog, I\'m gonna surf for the Midlands bus service website to get a number where I can make an official complaint. I\'m definitely not in an Aiki mood tonight. Hopefully when I wake up 2moro, my mood will change for the better.  Can\'t wait for morning so I can do my happo-giri exercise and go running in the evening. To end this post .... well, ... ARGHHGHGHGHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHH !!!!!!!!!!!!!!!!!!!!!!!!   '),(26097,979795,'08,April,2003','2003-04-08 00:00:00','    Eileen\'s blog kinda made me notice I\'m always talking about Aikido in my blog. So I thought I\'d say something a bit different for awhile.  Taking a breather today from running. Been really making good progress the past 3 weeks. Started off with about 4 rounds initially, each round is estimated to be about 500m maybe. At that time I weighed about 68kg. At the end of last week I weighed about 66kg or so and I managed to be clocking about 6-8 rounds per run. Did 10 rounds once but at a really slow pace then. It was kinda fun. Like I said, the run yesterday was quite crappy. I miscounted and had actually wanted to stop at the end of my 6th round. But when my housemate (Justin, aka personal runnung trainer haha) stopped running, I realised I had already done 7 rounds already (coz he managed to lap me by 1 more round). Had to suck it in and just finish the final lap. Was really on the verged of giving up mentally. Wah liews. But at the end of it, felt good that I still managed to complete 8 rounds instead of giving it up at 6. By the way, I\'ve been maintaining a constant 65kg so far. Doubt I\'ll get any lighter than this.  OK can\'t leave out Aikido stuff completely I realised. Coz I need to make a point with relevance to my legs. The exercise I\'ve been doing has a bit of impact on my legs. They feel stronger but as they are still aching from my (almost) daily reps, they do feel kinda sore right now. I\'m targetting to increase some kinda endurance training to my legs by increasing the hiji/higi tai practice I mentioned a few posts earlier. And if I can help it in terms of schedule and studies, I may swim once/twice a week as well. Legs legs legs ... very important. Keep overlooking them. No choice lah, my legs are generally the weakest area coz of my lack of emphasis in every exercise I do.  Which reminds me, really funny, when we left for our run 2 sessions ago, 3 angmoh girls were in front of us. While we overtook them, 1 of them called out \"Woooo!! ... Sexy legs!!!\" Hahaha, I felt like laughing out there and then. I immediately called out to my friend in front of me, \"Eh Justin, calling you leh.\" His only reply was \"Siao.\" Hahaha ... We sure have some funny moments over here.   '),(26098,979795,'08,April,2003','2003-04-08 00:00:00','    Went for a run today and I felt like crap right after, though my appetite was good. After the run had a slight stomache ache and had to really visit the toilet. Well, sorry I had to share this piece of news online but I just couldn\'t understand it. My fitness was definitely improving and I was surprised the way my body reacted to the run. Anyway, I took lotsa water and that kinda helped. Did a few more reps of 8 directional exercise just to take my mind off the discomfort. Also watched Steven Seagal\'s Above the Law just now. There were some cool moves but being a movie, I really dun think he was being very Aiki on tele. Well, that\'s Hollywood for you lah. If I\'m not wrong, there\'s a stream of Seagal\'s movies for this month. Woohoo can take note of his techniques and scutinize them. Ikkyo!!! ..... Nikkyo!!!! .... Sankyo!!! ... Rokkyo!!!! ... Kotegaeshi!! ... Iriminage ... and a whole lot of other moves that I really have no idea about. Could be a combination of some other variety.   '),(26099,979795,'06,April,2003','2003-04-06 00:00:00','    It\'s only been about a month and I thought I\'d met someone who is really nice to me. I\'ve always thought that I would take a bit of time before I even get any ideas about another relationship. Well I made a mistake about the girl\'s intent and the we had a good talk about stuff. It seems that all the girl did was just being really really friendly and sociable and that kinda gave me the wrong idea. Sure it was embarassing but what struck me the most was the amount of courage she must\'ve mustered to come tell me nicely. Take note, I know a lot of girls who just have a problem telling people they\'re simply not interested in the guy. They tend to be really really nasty about stuff. But with her, she was really nice about it and that helped things along a whole lot. The way things turned out was so great (honest!), we could talk again right after the conversation we had. It\'s really kinda embarassing talking about all this stuff online here and now. But when you think about, I think I managed to face the truth of the situation. It\'s a growing process and I\'m not ashamed of sharing it with anyone else. In fact, it feels kinda good. :)  Girls out there take note on one area you may want to work on. Guys out there take note about another area you want to be careful with. Heh ... today has been a good day, let\'s hope 2moro will be too, especially being my last Aikido practice for this month. Gonna video it down again. Erm, talking about Aikido, the daily practice I put into my 8 diretional exercise is definitely helping. I dunno any meditation stuff that I can use to usually calm myself but doing repetitions of bokken work for about 30-60 mins is definitely a way to reduce any tension you have and clear most thoughts from your mind.   '),(26100,979795,'02,April,2003','2003-04-02 00:00:00','    Yesterday\'s Aikido was really interesting. We did a full session of weapons work.  We did a quick warmup / revision of the 2 8 directional cuts, followed by a new pattern she called hiji/higi-tai. Till now I still can\'t find anything on hiji/higi yet. Think I\'ll check with her again by email.  Anyway for starters, we\'re required to \'Zha Ma Bu\' and start making shomen cuts. After that, we needed to shift our center / lean left and right while swinging the bokken to our side to deflect invisible blows coming down on us, immediately we had to shoft ourselves back to the middle and do a shomen cut again. With paired work, it actually became really nice and it was something like a I-cut-you-while-you-deflect and immediately followed by a You-cut-me-while-I-deflect motion. When you do it continously the exercise seems really really nice. Next stance had 1 guy in hiji-tai and another in our normal sword stance (right stance). Uke does a shomen cut with a step forward, and nage blocks (not parry/deflect). The block is just the shifting of hands to the side and up while raising your hands above your head, ultimately the tip of the sword faces your opponent and the edge is facing skywards. Anyway once you\'re in this position, the tip of uke\'s sword is kinda resting just before the tsuba. What we need to do now is bring our handle to our center. It gives a parrying motion which throws uke\'s sword off to the side. Again, doing this continuously make things look really cool. But please note that nage (whose sword gets parried aside), should remember to follow the parried movment other wise the sword may hit your knuckles, as I managed to find out the hard way. I\'m just glad my fingers till work. 1 final exercise was a parry. Nage attacks with a shomen, uke in higi-tai shifts 1 side and glances the blow away before returning with his own shomen cut. At this point I can\'t remember clearly but I think nage basically twists the arm/wrist or something along that line and deflects the blow in a direction, uke having had a blow deflected follows his deflected sword and swings behind/above the head to return with another shomen cut. Repeat process and you got a slightly complicated exrecise.   '),(26101,979795,'01,April,2003','2003-04-01 00:00:00','    Today was the 1st day since the end of term that I came to Uni. Had to loan some books and do some photocopying. My eyes feel sore and tired from sleeping late lastnight.  It\'s been about a month since my breakup with my ex and I was doing some thinking and reflecting about my last relationship. I dunno why this occurred but I think it\'s because of what my ex said a few days ago. Can\'t rem which day she called last week, but she was asking me something about how to run scandisk on WinXP. Gave her instructions and she thanked me and we hung up. After that I went to watch TV and was downstairs the whole evening. Suddenly, Lynette (housemate and course mate) got a call from my ex, who was looking for me. (Apparently, she called my mobile 3x and because I was downstairs I didn\'t hear my mobile. Mind u, my room is on the 3rd level.) My ex was pretty pissed, she started telling me things like, what on earth did I install in her laptop without her permission and why did I put XP in her laptop in the 2st place when I last helped her to reformat her harddisk. Then she said everytime she tried to run NJWin (think is a chinese recognition program), NJWin would crash. She started blaming me for not asking her before installing XP instead of win98 etc. She said she tried running scandisk and it won\'t even carry that out properly. I was a bit pissed but I really controlled my temper. I told her I didn\'t know what the problem was and that I dunno how to rectify it. I said I could put in win98 for her again if she wanted me to (considering that it was the start of my holidays and I can make time to settle it). All she commented was she didn\'t want me to do anything to her laptop and since I couldn\'t solve the software problem she kinda hung up abruptly.  I was really annoyed with her tone. But I can understand it a little better when I found 3 missed calls on my phone and found out my house phone was in use by my housemates that evening as well. She might have tried calling me and the house so often that she just blew it. Plus she probably encountered the problem more than once each time she used it. But understanding her reasons for blowing her top don\'t mean that I\'m excusing her for it. We are after all no longer a couple and I would really appreciate it if she stopped speaking to me like I bloody owed her something in my last life. To me, it only sounded like she called me up just to tell me off about me having installed XP on her laptop instead of Win98. To a certain extent I do admit that I don\'t remember consulting her before doing that. But another point I\'d like to note is that I distinctly mentioned to her before that XP doesn\'t need NJWin anymore because of it\'s capability to recognise other languages (I installed that component for her as well). The only reason she might have for installing NJWin would be to play Chinese games. If she wanted to type or surf chinese sites, XP could\'ve handled that.  I do not excuse my self from any part of my involvement in her frustration (no matter how small). But I really don\'t think she should overlook the big picturet that I wasn\'t intentionally evading any of her calls and the vents that were out of my hands (house phone in use). Also to note is that I had volunteered helping her reinstall back Win98 if she wanted me to. Because of all these, I really cannot excuse her for her behaviour. I\'m sorry things turned out this way but I\'ve made a resolve to not hold back these kinda emotions with respect to anymore outbursts by her directed at me. Naturally, her close friends will get the picture that I\'m some kind of monster as they never heard about my side of story. I dun think I give a damn anymore. Just so long as I know that I really didn\'t do anything wrong (or intentionally did it). If she just want to call me up to give me a piece of her mind on the phone instead of solving her problem, I dun see the point. I absolutely refuse to entertain anymore such outbursts like that as though I was still her bf or something. Since I\'m only a normal friend to her now, I do seriously expect to be respected as 1. Afterall I really don\'t blow my top at anything like that to my friend. I really have this feeling that all because of the way she responded, she isn\'t treating me as a friend anymore. Who on earth flares up so easily with a friend.  I don\'t want others to be too judgemental about her in anyway. Afterall, you never heard her side of her story and everything I\'ve said is 1 sided. I think my ex is still basically an OK person just that we have a great deal of extreme differences that makes us clash around a bit too much.  Goodness gracious. I\'m glad I\'m documenting this part of my life and that someday I can look back and laugh at myself or something. I hope that I will learn something from all my earlier posts, no matter how insignificant they may seem.   '),(26102,979795,'17,May,2003','2003-05-17 00:00:00','    I\'m in a pretty shitty mood today. Feeling the preassure of tommorow\'s paper. I think also partly because of yesterday\'s paper being quite bad. Very very sian. 1st paper on Wed was good. Really good. Half the paper was exactly the stuff I revised that morning. I should quite easily secure a pass. How good I do depends on how much sense the lecturer makes out of the crap I wrote in extra. The 2nd paper (yesterday\'s) was terrible. I\'m afraid of failing this module now. Seems like the opposite. It\'s such a broad area that things I wrote in might not be complete or they are wrong. To make things worse, I might not have remembered to write all the points I studies before. Argh!!.  What makes today top the icing on the cake is that I have a 5 min demo for my VR project at 5:30pm. Even Annie who took her exam from 2-4 came home before me. And I haven\'t even left the house at that time. It takes me 1hr to walk to and back to the house for a measly 5 min demo! BASKET!  Decided not to take my dinner at home. Probably grab some unhealthy fish and chips later on the way back. Might take some time to practice my bokken and handstand to ease the mood before I start my revisions again. Damn the day. Even the weather is all wet and windy. ARGH!!!   '),(26103,979795,'13,May,2003','2003-05-13 00:00:00','    OK I managed to find out what the stance I always seem to be using during bokken practice is really called. It\'s Jigo-tai, meaning defensive posture. Term is broadly used in Japanese martial arts besides Aikido, eg. Karate.  Another term is \'Masakatsu agatsu\', True victory is Self-victory. In Chinese / Kanji = Zheng4 Sheng4 Wu4 Sheng4. Meaningful, yet so hard to accomplish for so many of us.   '),(26104,979795,'12,May,2003','2003-05-12 00:00:00','    Haven\'t blogged for over a week now.  Let me just describe my last Aikido lesson last week. We did weapons work and I managed to video the whole thing down as well. Managed to get some pretty nifty footages too. There\'s only about 15 mins worth because we basically did tons of repetitions. I must admit that my bokken work has been improving a little. The daily exercises do help me really well. Even just this morning, I\'m not sure how but I managed to get the feel of what Pauline Sensei mentioned before in the cutting motion. That it was pretty similar to fly fishing where the wrists will twist first. Don\'t know why it happened either but I got a sudden smooth feel to things and I do notice that the sword tip does fly a bit faster. Only thing I still noticed was that my shoulders seem to be locking up easily sometimes as I progress on in my cuts.  Anyway school revision has been quite tensed. What some of us are doing is that we do stupid stuff to ease the stress. Begh was on the Singapore\'s gymnastics team for the SEA Games a few years back. So we suddenly asked her how to do handstands. We\'ve been trying like crazy daily. Haha... really interesting. A few knocks and falls here and there but Daniel, Justin and I do seem to be progressing. Justin seem to be able to get the most balance because he\'s not afraid of falling. I\'m slow because I am afraid of falling too hard. Daniel seems stiff in the beginning, but at least now he seems to be able to do a cartwheel while practising the handstand. Now handstands are like a daily regime for us. We\'d just do it against the wall for awhile before trying freefall. If I\'m lucky I can let you guys see a 4 min footage I recorded. Annie did manage to snap some photos of us but I probably won\'t get to see those until after the exams.   '),(26105,979795,'03,May,2003','2003-05-03 00:00:00','    OK lastnight I had to camp over in school to finish my VR project. The virtual dojo looks like crap. But I really have no time to make it too complex, what with exams and ll just less than 2 wks away. Quite disappointing.  Typed up a crap report (no point wasting too much effort for 10%, plus the lecturer really isn\'t too interested in a lengthy one anyway. He stated the requirement as anything below 2000 words.) Submitted my project half a day earlier and I went back home. On the way, bought fish and chips at a takeaway for lunch. Finished and took a bit of a nap. Woke up and basically replied some emails.  Seems like most everyone I know is either going to be busy or unavailable the next 2 wks. Unless I\'ve got something interesting to post, I\'m probably taking a blog break as well. Might throw in some occasional stuff on Aikido.  Almost forgot, I went for Aikido on Wed night. Only 3 students. Me and 2 new beginners. Well, one of them actually did a beginners course at a friend of Pauline Sensei\'s about a year ago. He was a really polite chap. Well mannered and cordial. I noticed that hee understands the formality aspect of the art and our normal practice as well because of the way he behaves in a dojo. The other guy had no experience but we did some ikkyo waza throughout the session.  I think I made the right decision to go down because it would have been weird if there was only Pauline Sensei and the 2 of them. I\'m sure she can still teach as effectively but it would be nice to take some ukemi from her. Hope I wasn\'t being too stiff that night. The first thing she did when I arrived was ask if the families of my housemates and I were alright. Told her noone had any bad news from home so far and things seemed pretty much well under control in Singapore. Last I heard 2 days ago was that there have been no new cases in Singapore.  I almost forgot to mention the short conversation we had after class. I remembered she mention that Chiba Sensei was visiting her the last 2 weeks so I asked if she had enjoyed his company. Turns out Chiba Sensei didn\'t come to UK. It seemed that he injured his foot/leg before he went to France for a seminar. He was in so much pain that they had to send him back halfway through the seminar. She said it must\'ve been really bad as she has never known him to have to cancel a trip or seminar before. On the scheduled trip to UK, the teachers in UK were suppose to attend a seminar organised specially for them.  Apparently Chiba Sensei is a Buddhist monk as well. His study of Aikido is not just about taijutsu but weapons work (Batto-ho), ZaZen and Buddhism. Pauline Sensei says he encourages these 4 disciplines for anyone who is serious about doing Aikido in the long run. I put forward a question to her about Buddhism as a religion and having the need to take up this religion even though I don\'t really want to or believe in it. She says, Chiba Sensei doesn\'t really expect you to be a convert from your religion but instead to accept the discipline invloved in all 4 elements. She quoted an example saying that when you sit in the lotus position (I think adopted by buddhist and yoga in which you sit cross legged but with your foot lifted up on top instead of under your knees) for 45 mins, you are in such pain or discomfort and that you are so alone that there really isn\'t anyone else but yourself to tide you through this. The idea is that you have no choice but to look within yourself for strength. Strangely this idea comes to mind when I\'m doing my running and swimming. In swimming, I can tide through the feeling of being tired and just keep going for ages and long distances. But when it comes to running, I have a problem staying focused and pulling together to push myself past the no. of rounds or distance that I had decided to run. Sensei says no one would really just sit there and do this for no particular reason. I also never thought of stopping in the middle of what I\'m doing to just hold still and look or listen at the world around me. I guess if I look at 2 of my Christian housemates, they always have a daily quiet time of prayer with God. I\'m not sure what is the point I\'m driving at, but I just see a lot of co-relating situations and similarities in a positive and negative sense in my self and people around me. Sensei made 2 points to me, as an Aikido-ka under Chiba Sensei, she encourages anyone to take up the studies that Chiba Sensei encourages. But on the other hand, at a more personal level, she admits that despite being pretty competent in taijutsu and weapons work, she doesn\'t go into ZaZen or Buddhism. But that\'s because she has looked at herself and feels very much comfortable with what she sees in herself currently.  It was an interesting chat lastnight. Food for thought.   '),(26106,979795,'30,June,2003','2003-06-30 00:00:00','    I\'m working on including my journal of the trip to Scandinavia and another of my trek of the Cumbria Way in UK.  Previously placed them in the blog here but I found them way too long to be included in just one blog. Photos have been developed but will be scanned in when I can.  The journals are both pretty long. Cheers.   '),(26107,979795,'25,June,2003','2003-06-25 00:00:00','    I\'ve just returned from my 1st trek.  Decided to cut my trek short. Instead of going for both treks, I decided to stop. I\'ve finished Cumbria Way (whew). It was quite an experience because I\'ve never gone for such a distance for 5 days. For guys who\'ve done the route marches in BMT, it\'s not particularly difficult. It was a matter of getting used to doing the same thing for the next few days.  Anyway, a lot of reasons why I decided to cut my trip short. I\'ve had to because I realised that my trip to Scandinavia actuallt burnt my finances by quite a lot. A lot of complications involved. Also, I managed to contact a property owner about renting a space for my parents during convo, I\'d thought I best settle this fast and there\'s no better way than to come back early to do so. Next, on the last day of the trek, I got caught in a drizzle/rain. Basically I\'ve been sick for the past 2 days. Last, I\'m not sure I want to do 2 treks 1 after another (though I do have a 2 day rest inb between them). I do really want to go back to Scotland but I think I\'m not physically and mentally prepared right now to do both. There\'s just too many thing hanging on my mind. I\'m not one to be able to multi-task so I think I\'ll just tackle one issue at a time.  I\'ll be posting my journal of this trek online soon as well. Photos will take sometime though.   '),(26108,979795,'25,June,2003','2003-06-25 00:00:00','    Just to let people know I\'m still alive. Got 1 min of online time in the hostel left.  In UK now trekking thru Cambria Way. Freaking shiong and I still got the West Midlands Way in Scotland to complete. Sibeh jialat. Blisters on my feet are killing me.  Will try provide more details soon. Ciao.   '),(26109,979795,'20,June,2003','2003-06-20 00:00:00','    I just read thru all the blog I\'ve missed so far. Been quite fulfilling, I see. (Aikido training, work, school, etc activities.)  OK lah time to update people about my existence.  I just got back on the 18th morning about 4am. Immediately, I asked the cab driver to wait while I grab some stuff and went down to the Uni straightaway to check my Provisional results (may be subject to changes). I know it\'s a bit early but I was really nervous lah ok. My dept is accessed using ID cards so it\'s open to us 24 hrs anyway. When I checked my results online, I went to double check on the notice board. I got a 3rd class Honours. OK lah, to some people it\'s not fantastic. But to me, considering the fact I was expacting that I didn\'t manage to achieve my honours successfully, it was a tremendous relief. I\'m seriously damn happy about it. Was really worried that the money and time spent in UK would be wasted if I didn\'t at least get home with a honours. granted lah, back in Singapore most people treat 3rd class the same as a normal degree but that\'s besides my point.  Just today, I managed to get my official results already. No different from my provisional. Still damn happy about it.  Well I did enjoy my trip to Scandinavia but it was a real experience with multiple hiccups along the way. I\'ve managed to maintain a written journal in my notepad during my travels and I\'ll type it in asap. Meanwhile I only have a few days to settle some admin stuff and get my final results. I basically need to settle convocation details, accomodation for my parents, bills, handing over of the house back to the property management, packing my stuff for shipping back to S\'pore, shop for some emergency items for my trek starting this Sun, air ticket, etc. Really 1 big pile of shit to do in 3 days.  By the way, I need to state for the record that the phone line in my house together with the Broadband internet will be cancelled this coming Sat on the 21st of June. So I probably won\'t be able to come online anytime soon. The best I can do is check my emails occasionally in Uni but that\'s not confirmed yet because I may be travelling around with my parents. A brief summary is found below:  Current - 21st Jun : Busy settling packing and admin. Won\'t be sleeping too well next 3 days.   HOUSE PHONE AND INTERNET CANCELLED.  Only contactable by mobile phone +44-77-1717-4992. Not sure of the temporary address at this point in time.  21st Jun - 8th Jul : Trekking in Lake District area all the way up to Scotland. Landscapes aren\'t suppose to be too difficult but the distance is going to be quite an experience for me. 8th - 14th Jul : Bumming and settling admin for parents visit to UK. 15th Jul : Parents arrive in UK. 15th - 25th Jul : With my parents. Note my convocation date is 17th Aug. So I\'m probably busiest around this time. But I\'m probably not contactable at all then either. 26th Jul - 2nd Aug : Visiting Stradford upon Avon (again. Sianz ...) with my folks, dunno wat on earth my parents wanna do there a 2nd time. 2nd Aug : Parents depart for Singapore. I\'m on waiting list currently, so not confirmed if I can go back on this date. 2nd - 3rd Aug : If I dun fly back on the 2nd, I\'ll juz be buming around. 4th Aug : I\'ve got a confirmed flight back on this date. Will depart UK latest on this date. 5th Aug : Should arrive in Singapore.   '),(26110,979795,'05,June,2003','2003-06-05 00:00:00','    My exams are finally OVER!!! Woohoo!!!  The studying is over. Now is the worst part. Waiting for my results. I definitely know that I didn\'t do as well as I\'d liked. Dammit. There\'s a good chance I won\'t be able to get my honours even. I really have to acknowledge that fact after my 2nd last paper. I dunno how but I guess a degree is what I\'m hoping to look for at this point. I\'m still trying hard to accept this fact. No choice. Can\'t do anything now but pray.  Went back and slept and slept. But I dunno why I think the adrenaline is still rushing and I didn\'t quite sleep for as long as I intended. I definitely need to force myself to bed and catch up on some rest which lost. Especially so during the last 72 hrs.  Tonight the house is really empty. But it\'s good lah, coz I got Aikido anyway. Going for practice in a few minutes.  Tomorrow night I\'ll be heading off to Oslo, Norway for a trek with Justin and Daniel. Cool man. Can\'t wait for it. Will probably be trekking for about 5 days or so. But don\'t think it\'s out in the wild lah. There are towns and cities and hostels along the way for the trek. Civilisation really isn\'t far away from our route. It\'s suppose to be a simple and comfortable route in Oslo. After that, we\'ll hit Copnhagen and take an overnight ferry to Stockholm. Will buy a train ticket to go down to Karlskrona, Sweden to visit my naval friend, attached there for submarine training. Oops, reminded myself that I need to call them up once I get my hands on the ticket. Finally I\'ll meet up with my housemates in Stockholm and fly back to UK to get my results (which won\'t be good).  After a few days of rest, I\'m heading off to Scotland for trekking again. Lake Coniston all the way up to Edinburgh. Will be back just before my parents arrive and hope to prepare for the convocation.  All in all, I think I\'ll be back in Singapore on the 2nd / 4th Aug. Hope to resume my NS around the 18th Aug.   '),(26111,979795,'25,July,2003','2003-07-25 00:00:00','    OK I\'m back in Birmingham now. Got here lastnight amd will hang around till Sat morning before I head of to Stradford.  For today, will be settling stuff in Uni and taking a day trip out to Chester (where\'s dat? no idea).  Past week of tour was quite crappy.  Went Leeds coz my dad had to meet a business associate. It was damn paiseh when my dad actually asked them if they may have a job for me. Wah lao ... seriously leh, dun wann do IT job liao and they\'re like forcing me towards that direction. But heng lah, they dun sound too negative at all towards a career with the Police Force. Think when I get back I can apply immediately for real by going down to the recruitment office personally.  I hr drive away is York. Quite nice. We stayed there overnight. My dad was quite pissed with the Hostel. Actually went into a bit of tantrum for a bit. Can\'t stand it. Argh!!  Next day went down to Edinburgh. Wasn\'t bad. Went Glasgow as well. Spent 3 days in Scotland. Quite nice. they got a lot of nice museums which were free and they were actually really interesting.  Went down south to the Lake district. The worst day here! Mom insisted on diverting here and there to see this and that and we had to drive in the country roads when it got quite dark. Considering there\'s more daytime in summer, you can tell it was already about 10+ when we lost daylight. Kept telling her we won\'t be able to get to the accoms with enough light to spare but she wouldn\'t listen. She seems to be on this marathon where she juz want to hit the entrance of this or that spot, take fotos of the entrance (coz she dun wanna pay entry fee and go in) and ciao off to th next place. Wah lao!!! Dad didn\'t help either. He\'s basically the type who expects standards in things but coz of the cheap and budget rate my mom is able/willing to fork out, we dun see or do much. So he\'s always complaining about things like staying in Hostel lah, no privacy lah, low class lah ete etc. I mean c\'mon man, we\'re in a hostel dorm, if u wan that kind of standard, jolly well check into a hotel if you can afford it lor. ARGH!!!  There\'s actually a lot of things to do and see in Lake District area. It\'s a national park but they dun suit the older folks like them. And they refuse to go for walks and treks coz they feel they \'can\'t take it even thought those are simple ones. Argh!!! I can\'t go do any water sports coz of bad weather and budget. Argh!!!  Damn weather was poor since Tues too! Argh!!  Yesterday must be the best day. We were leaving Lake District area and we stopped at this cheap factory outlet spot. The moment my mom stepped out her shoes broke. Heng heng we found a Clarks factory outlet. Went in and bought 6 pairs for £89 (think is about S$200+). Was damn worth for 6 pairs. There was Buy 1 get 1 free, buy 1 get 2nd for £12, buy 1 get 2nd for £5, and some other shoes which were going cheap. Plus, the range was a lot more than I expected. I got myself 2 pair, my dad got 1 pair, my mom/sis got 2-3 pairs since they sometimes shared shoes. I suspect this may be the highest point of my trip liao. Well, unless my housemate Annie is able to pop by next week for a visit, then that\'d probably be the highest point of my trips. But then again, with my folks around ... I think I\'ll still be in a sian mood. Siaaaanz ... argh!!   '),(26112,979795,'18,July,2003','2003-07-18 00:00:00','    Graduation day today.  Off to a \'great\' start, couldn\'t find my wallet. Woke around 6am and took me 1 hour to find it. Realised David had taken it by mistake the night before. All the while my parents were nagging in the backgorund .... buuuuuzzzzzzz. Didn\'t get any peace until after it was found. Sheeez... Left the house with Lynette and both our parents for Uni. Drove down, despite my recommendations to walk down. Weather was really dark and gloomy. No sun today. Duh...  Went to collect our convo tickets and the gown in the Avon Room, University Ctr. It was quite easy to put everything on. Was really chop chop. I must admit, despite having only achieved 3rd class, I still feel rather proud to be wearing it. Dunnoz, maybe it\'s normal for under-achievers like mua. Hyuk hyuk hyuk. We went around the Uni to snap some shots before the doors to the Great Hall opened. That would be where the ceremony is gonna be held.  Our parents went in to grab good seats in the front row as soon as the doors opened at 930am. Lynette and I met some people here and there and decided to take some photos with them before we had to be seated in the Great Hall. Took us awhile but we went in by 10am. I kinda fell aslwwp while waiting and I was jolted awake by the orchaestra\'s fanfare that followed when the procession (of Chancellor, Vice-Chancellor, lecturers, etc, arrived. It looked nice and kept me awake pretty much for the whole ceremony. The Great Hall was kinda warm, and I\'m sure the shirt, suit, tie, gown, and hat didn\'t help to keep us cool. We listened to some speeches (less than I had expected, which was good), and preceeded. I was lucky to be the first half of the ceremony. Saw a few familar faces here and there. At the end, we followed the main procession out of the Hall. A lot of waiting around before the dept photo was taken. Paid £21 for it. My parents finished 2 roll of films which we sent for developing in the Uni Ctr.  When we left, the mood (expectedly) turned irritatingly and nasty. Parents were like blaming each other about where we can grab (cheap) lunch. Of course, I got part of the blame (as usual). The graduation day had to start and end in a crappy manner. I\'d say that\'s a tradition. What a day.   '),(26113,979795,'16,July,2003','2003-07-16 00:00:00','    Everyone\'s parents seemed to arrive UK on the same day. David\'s, Lynette\'s and mine. It was heckuva day. Been running to and fro and rushing places. Dun understand why either.  My parents arrived as well. After some slight difficulty, they did manage to get themselves 1 street away from me. I walked down to meet them and direct them the right way (which was like 500m away only).  The evening they hung around and they got me really pekchek. 1 of my housemates reminded me to be more patient (obviously they could tell I was starting to rise my voice and all that). Told myself I\'d try to contain whatever impatient feelings I\'ve got but it\'s tremendously hard with my mom. The dun-listen-but-wanna-talk-more-and-harping-on-same-topic-but-can\'t-get-main-point-across chitter chatter juz gets me everytime. My dad\'s at least is more objective about stuff.  I have a feeling it\'s going to be a long 2 wks. Watever lah, at least I do have the convocation tomorrow to keep me busy for the next 24 hrs.   '),(26114,979795,'11,July,2003','2003-07-11 00:00:00','    Kewl man. Was surfing around when I found this:  Singapore police off to help build peace in Iraq   urlLink http://straitstimes.asia1.com/topstories/story/0,4386,198752-1057787940,00.html?   This would most probably be the kinda stuff I hope I can do someday. Cross border policing.   '),(26115,979795,'09,July,2003','2003-07-09 00:00:00','    Just read an article about Kata and Waza from Aikido Journal.  urlLink http://www.aikidojournal.com/new/article.asp?ArticleID=463   \"Kata should be practiced according to a certain order or prearranged method which is based on a rational relationship (rial). Thus, we are not falling because we are being thrown but rather we are practicing a kata designed for us to be thrown. When we master a rational movement (kata), it is expressed as a natural movement (waza). That is, if you become able to execute a kata spontaneously as a result of repeated practice, you are no longer performing a kata but are executing a waza. We learn through kata and become unconscious of the fact. In other words, as long as movements require our attention they are kata, when the kata become spontaneous they become waza.\"  Interesting explanation on wats kata and wats waza. Of coz the remaining article is even more interesting. \'Cept that the whole article is available to paying members only.   '),(26116,979795,'31,August,2003','2003-08-31 00:00:00','    urlLink Eileen\'s  updating of the AFS gallery only serves to remind me that I should update my own photo gallery soon. So far I\'ve only scanned 1 out of the 10 rolls of film I had developed. Just started editing some of them to make them clearer. But I look at the other 9 rolls ... oh man.  Time to plan my day liao. Looks like a busy period ahead.   '),(26117,979795,'29,August,2003','2003-08-29 00:00:00','    During the Aikicamp, I was commenting to Cody on his getting his 4th kyu so soon and he commented on something like once you\'ve clocked up enough hours you can easily get the grade.  This just brings something to mind. Is that how Aikido grades work? That once we\'ve achieved the minimum no. of hours and we\'re about able to get the next grade. I really don\'t think so but I still get a bit of a nagging feeling that it may be slightly true. But shouldn\'t it be the case that techniques and understanding be the main factor when it comes to grading? Part of the reason why I may have the wrong impression is also of the regularity of grading dates. Sorry I\'m always doing some kind of comparison, but back in UK, we never actually set when a grading may take place. Pauline Sensei would at most give 1 week\'s notice but I did hear of a grading being sprung on the students the very night itself. Her purpose of course, is not to cause panick and fail any of the students. But it does seem more effective in the way they would handle the stress and sudden preassures in real life.  I sometimes wish things were just simpler where everyone just train train train and forget about grading. That\'d be wonderful. If it comes it comes, if it doesn\'t well, no loss either.   '),(26118,979795,'29,August,2003','2003-08-29 00:00:00','    Had a family meeting tonight. Haiz ... it\'s tough having to make ends meet now. My Mom\'s CPF has just finished as of last month. We\'re discussing about what we can do to help out with the money situation at home now. Very vexing. Even worse, my dad likes to lament on the fact that I extended 1 extra year of study for my degree, over expenditure in Norway, etc. Nothing I can do about it now but he seems to enjoy the finger pointing exercise from this. Sianz ...  Looks like I\'ll need to be doing some parttime job soon. Probably gonna do some waitering at a cafe or tuition. May need to inform Sensei that I seriously can\'t afford to pay him for Aikido anytime soon. He was kind enough to say that he\'s ok with waivering the fees for students who can\'t afford it. All I need to do was help him out in the dojo and maybe some demonstrations here and there.  Lotsa shitz to handle. Need to try and send out more resumes now. Oh man ... so much woes as a fresh graduate. I just hope I can enter the Police Force soon. Or at the very least get some kind of confirmation. If I don\'t, might just have to settle for a boring IT job for money\'s sake. Also, seems like my Dad not very supportive of the idea that I\'m going into the uniform service. He is OK with the govt sector as long as it\'s non-uniformed. I dunno leh, that\'s exactly what I\'m working towards. But will need to be very sure before I decide.   '),(26119,979795,'29,August,2003','2003-08-29 00:00:00','    So sad today. Annie left Singapore for UK this morning. So many things to say and do but not much opportunity.  I guess I can only look forward to the future.  I guess I should start working towards my goal as a police officer after my NS. Time to start reaching for it instead of waiting for it to fall into my lap. Hope my remaining NS life will be able to help prepare and gear me up for a new life with the police force.  The part waiting for news is the worst.   '),(26120,979795,'26,August,2003','2003-08-26 00:00:00','    Muahaha ... planning my leave liao! The way I see it, I maybe clearing them on every Mon after the exercise ends in early Sep. This way I can enjoy a long weekend and still not interfere with the life-runs in camp. Woohoo ... But of coz it\'s still flexible lah, if I get any interviews here and there, I\'ll still need to shift the dates around so I can go for them.  Err ... if anyone out there has anything to suggest about the way my leave should be planned. Please gimme some tips too huh.   '),(26121,979795,'23,August,2003','2003-08-23 00:00:00','    OK today has marked the end of my 1st week back in NS. 6 days has passed. I now have 116 days left as of coming Monday. Countdown!!  Err anyway, the week was quite uneventful. 1st day itself was the day reservists had to report to camp for a 1mth exercise, so naturally my precence was a low priority to my AS1 (manpower officer in charge). The interview and briefing was a quick one. Saw some of the old regulars who were still there. But it\'s quite sad, 2 of them are just waiting for the SAF to discharge them. All because they didn\'t pass their IPPT (hence no promotion coming). I guess some people didn\'t see fitness being such a serious aspect of their job feequirement when they signed on. It\'s not uncommon among many of the regulars I\'ve seen, and I\'m sure they won\'t be the last.  I wasn\'t sure where I\'m attached to so I basically reported to 2 \'bosses\' when I can. The SSG in charge of workshop and the brigade RSM. Currently I\'m attached to the RSM at least for the duration for the month long exercise as his PA (personal assistant). He treats me quite well. 2 other guys have been temporarily attached to him prior to my arrival, Neo and William. Both nice friendly blokes. Neo is an MOE local scholar and William is a poly-grad from TP. Both are quite a funny combination during office hours. It seems like the RSM will be trying to retain me for the duration of my NS under his office but it\'s still too early to say if that\'s going to happen. I wouldn\'t know anyway, I\'m just a foot-soldier. But I do hope that a good balance of work load is possible in office and workshop. I seriously want to try driving an M113 again.  Tues was my first liferun in camp. It was an internal run by my own unit instead of a sister unit conducting it. Was a good slow pace for 4.5km.  Wed work has basically started.  Thurs was another life run, this time we joined our sister unit in camp for 5km. Being an infantry unit, the level of cohesiveness and spirit they have is simply admirable. They were actually cheering all the way.  Fri things were quite busy, even for Thurs we extended our hours a bit beyond office hours but not by much.  Sat was an outcamp run from Pasir Ris camp to Pasir Ris park. Originally I was just supposed to run 10km, but I met an officer I was attached under when I had just passed out from BMT a few years ago. He remembered my face but not much besides that. As I talked to him I kindly of went past the route I wanted to take and had to go for the 15km distance instead. My timing was an atrocious 2hr 2 min. But that included a lot of walking in between and I suffered quite a few cramps on both my calves. But considering that it was only my 3rd run and 1st long distance for a long time, it may still be acceptable. I hope I\'ll be able to improve before the Army Half Marathon on 14th Sep Sun. Didn\'t managed to get much work done today but will need to do some stuff from home to prepare for Mon.  Sigh ... NS life.   '),(26122,979795,'17,August,2003','2003-08-17 00:00:00','    Today marked the end of the Aikicamp with Hattori Sensei. Just wanted to mark it down as my first Aikido camp. It was an enjoyable experience. I\'m just disappointed that I did miss some lessons and I can\'t remember a lot of the things I was taught, especially weapons work.  If I could just understand things better, I might be able to practice weapons at home. But ...  Anyway, I did meet a lot of nice people (and a crappy one) at the camp. I\'m glad on the whole, I still enjoyed everything.  I do have some wishes though ... wished Gary was training there as well (hope you\'re reading this dude   haha).  Tommorrow will be my first day in camp. I have to get used to wearing uniform again. Not a problem though.   '),(26123,979795,'15,August,2003','2003-08-15 00:00:00','    Annie had arrived from UK just 2 days ago and she\'ll be hanging around for a week. She\'s trying to source for air tickets to Taipei to visit her grandmother there. Her return flight back to the UK from Singapore will be the 27th of Aug because medical students will be starting their term 1 mth before the usual time on the 1st of Sep.  Today I met up with most of my housemates in UK. All but David managed to go out for dinner. It was an expensive but happy day. Of course, it was a real pity I had to excuse myself from Hattori Sensei\'s seminar at BVCC but because Annie will be around for a short time before I needed to go back to NS, I really had to juggle my time between Aikido and her.  Most of us met at City Hall around the afternoon. Then we went around the area to look around. Spent awhile at Citylink mall and then went for dinner all together at Marche in Suntec. After that we took a short walk to Esplanade mall to look at the night view of the area there. It was really nice. I must admit, the new look looks much better than the old days.  It was disappointing that the night ended so early but becasue Justin and Daniel had to return to base early tomorrow morning so we all went home pretty early.  For me, tomorrow is the start of the Aikicamp and I\'m suppose to go down to help move the mats. I had previously arranged to meet Richard at about 10-1030 in the morning, but I needed to collect my glasses and the shop only opens at 1030am. SMSed Richard and he said it was ok and I could make the 2nd trip down to shift the mats. I\'ve a feeling it\'s gonna be a great camp and a great day tomorrow.   '),(26124,979795,'14,August,2003','2003-08-14 00:00:00','    The seminar / combined training today with Hattori Sensei at BVCC was great.  I have to admit, it was nice seeing some familar faces again.  Paid my arrears, collected my jo, and even managed to get some short clips of Hattori Sensei teaching the dojo. It was a waste that I didn\'t manage to get the part of him showing some weapons technique to the dojo. He was trying to relate the empty handed forms with the swordwork. It was truly interesting.  I found a lot of people have improved tremendously. They\'ve adopted a very style which seemed quite common in some of them, and yet still still makes themselves stand apart from the rest. I\'m not saying they\'re very great or anything but I was pleasantly surprised with the improvements in the past 12 mths or so.  Just which I had made that kind of progress when I was in UK. Just makes me want to train harder. I haven\'t started my bokken exercises since landing in S\'pore. Think I\'ll start real soon. Once things get settled, I may ask Ah Hee Sensei if he still conducts weapons classes. Would be really happy to be able to supplement my taijutsu with weapons.   '),(26125,979795,'13,August,2003','2003-08-13 00:00:00','    PC has been acting up just yesterday. Managed to kinda get it working again. Haiz ... been working on it since last week after I landed in Singapore.  Has anyone seen the NUS Aikido website? It looks really good now. Much much better than before, and definitely way better than even their associations homepage.  Talking about new websites, the SPF has a totally new and kewl look too. Was checking out the recruitment page when I saw the new look. They even have a link for the STAR (Special Tactics and Rescue) Team. My only beef with them is that it takes longer to download now. First of all it requires plugins like Macromedia. After that it takes time to download and travel in between links (ie. Loading xx% etc).  OK talking about the SPF reminds me of my NS. I got my official notice to resume service on 18th Aug. Just occured to me that I\'m not exactly sure how to travel there and the time I\'m suppose to report. Think I\'ll need to call CPC for the unit no for all the details.   '),(26126,979795,'06,August,2003','2003-08-06 00:00:00','    Been working on my PC for the last 2 days. Damn thing keeps failing. Seem to have sorted things out. Dad\'s PC is probably just spare parts now and I\'ve set up space for him on my PC as well. Ah .. anyway I dun think I need to PC much. Hope I can shift the thing out into a more public area. I just dun like people running in and out of my room constantly.  Today went NUS to check out prices for a 2nd laptop for my sis. She\'s quite eager on getting one. Some good deals for NUS students but I only got my cousin there as a lobang and he\'s already bought a laptop from Uni. 2nd purchase isn\'t possible till Sep and even so the price is gonna be increased I think. Anyway, there\'s a computer fair on during the 1st weekend of Sep, might as well check that out then and get one there. I\'m sure there are pretty good bundled deals.  Went for my first run today. Pleasantly surprised to realise i could still run pretty OK. Then again it could be coz I started 3km only. Halfway back I was overtaking 2 girls when 1 of them called out to me. Such a coincidence, it was Begh (UK housemate). Wasn\'t in the best of condition to stay for a long chat so I departed pretty soon. Probably give her a quick ring these days when I\'ve settled my things. Hmm ... maybe I\'ll go run again if I can wake up early 2moro. Maybe grab an earlier night today. Past 2 nights I\'ve been sleping no later than 3am.   '),(26127,979795,'04,August,2003','2003-08-04 00:00:00','    OK juz got back to S\'pore this evening. My flight was delayed about 30-45 mins but my parents flight was early about 30-45 mins so we basically arrived almost the same time.  My flight was from via Emirates, from Birmingham to Dubai, where I had to take a transfer from Dubai to Singapore with a stopover in Columbo/Colombo. That\'s where the delay occurred. For a stopover, only passengers whose destination was  Colombo needed to alight. Those going to S\'pore just stayed onboard while the plane refueled. But some buggers actually alighted and they didn\'t get back on the plane on time. Basket, the least they could do was take note of the time and listen to announcements what right. Really selfish coz they delayed everyone else and it was really inconsiderate of them.  Anyway, I was suppose to meet a friend who was working at the airport but coz some miscommunication and my parents early arrival we didn\'t manage to meetup. Nevermind lah. Was quite knackered out tonight. I tried to stay awake on the plane as long as I could so I\'d have to go to bed early to recover from jet lag. But I feel so awake now.   '),(26128,979795,'27,September,2003','2003-09-27 00:00:00','    Dunno what came over me but I just decided to be adventurous. I signed up for the Singapore Marathon 2003 which will be held on 7th Dec. Though it\'s a full marathon (42km), I\'ll just be running the half marathon route (21km). I\'m even considering the biathlon which is organised every year in March but I think I need to seriously cross train properly before I can take it.  Still dunno what came over me.  Yuishinkai\'s seminar will commence this coming Thurs to Sat (2nd-4th Oct). I think I\'ll register for the Thurs afternoon seminar. It should be an interesting experience. After that I should have more than enought time to grab dinner and head on down to Siglap for Aikido class at Siglap. I just hope my leave application on Mon will be approved.   '),(26129,979795,'23,September,2003','2003-09-23 00:00:00','    Today I visited Yuishinkai\'s dojo at Upper Boon Keng Road / Geylang West CC with my friend who trains there.  Yuishinkai has a really soft touch to their techniques and their class practice with an open attitude. There was a casual atmosphere throughout the stretch of class. Everyone was friendly and approachable.  We started with some stretches at a faster rythm then I was used to. A lot of ukemi too. A lot of emphasis on Ki extension and balance (ie. imagine your head is lifted by a think string, etc). It was an interesting and fulfilling class. Sure it seemed slower then the class I was used to. But they pointed out plenty of minute details in my technique I had not noticed before. Ki extension (the simplest one), isn\'t that impossible for me afterall.  All in all, it was a really fascinating experience. I might actually drop in more often to reinforce some of my concepts and fundamentals again.   '),(26130,979795,'20,September,2003','2003-09-20 00:00:00','    The last 2 days were terrible. Again it\'s because of poor working attitude and irresponsible behaviour of certain individuals in camp.  Basically a sister unit was unable to provide the usual guard duty personnel for certain days due to a sequence of outfield exercise they were involved in. Naturally, our unit had to be pulled in to replace the guard duty personnel. In the process of all the reshuffling, some people had to cover by taking up extra days of duty.  1 particular person went to get MC in order to escape from a duty day allocated to him. In the process, other people had to be arrowed to take up his duty. He is so childish and irresponsible. Instead of reflecting on his actions, he feels that it was my RSM\'s fault that other people got the duty. This mentality is dominant in several people, they do not see the fact that these extra duties HAVE to be fulfilled, and that my RSM actually had no choice but to allocate them to the men. It wasn\'t some random punishment meted out to the men for nothing. It was just something that HAD to be done. Such childish and selfish thoughts and behaviour are what I have to put up with most often in the office. Looking at this particular person, I feel seriously disappointed with people as a whole. I feel sorry for my 2 friends who have been seriously affected by his inconsiderate actions.  I have a final full day guard detail tomorrow on Sat. So looking forward to it because that\'s the last duty of the month for me. I really have to GAMBATTE!   '),(26131,979795,'17,September,2003','2003-09-17 00:00:00','    It\'s another duty day for me. Staying in camp (again). Haiz ... can\'t wait for my last duty to be over this Sat. But then again, it\'s the worst, guard duty. More walking around camp. For a whole day somemore. Boring, but needs to be done, what to do? Must think of some ways to entertain myself this duty.  Oh, later gotta order takeaway. Hmm ... wonder if we should order pizza, Mac or KFC?   '),(26132,979795,'16,September,2003','2003-09-16 00:00:00','    Dammit! Some stupid zip disk in office damaged my Zip drive! I brought the drive and disk home to test and it spoilt the 2nd drive I had at home too!!! ARGH!!!!! .... MY ZIP DRIVESSSSSS!!!! ARGH!!!   '),(26133,979795,'15,September,2003','2003-09-15 00:00:00','    Right ... I\'m now in camp whiling the night away. Got stupid QRF (Quick Resonse Force) duty. Basically means if the camp is under attack or something, QRF personnel will be supporting the guards at the guard room at any point in camp, hence the term QRF.  Had a difficult afternoon today. Basically being in RSM office means I gotta plan out the monthly duties (eg. guard duty) for the whole brigade. Haiz, it\'s proving to be quite an ordeal. Basically what happened was that when I came in, my friend Neo had already planned out a duty plan. Problem was a neighbouring unit had to go on exercise and that meant they couldn\'t send the men in to do duties for certain pre-arranged dates. They requested a swap but now the dates we got were not favourable. Means some people had to to have their original duty dates rearranged and some actually got extra duties beyond their monthly average, while others got their duties so closely packed together it became difficult to have enough rest in between. Some last minute medical cases had to be pulled out from their duties at the last minute, leaving even more empty duty slots for the month and insufficient people to fill them.  It seems like even I myself will be doing some of these extra duties.  I\'ve heard of several unhappy grumblings and it feels bad to be the\'bad guy\' to have to do these. I\'ve been making changes since last week and till just now, things have been pretty disorienting and hectic. It seems there are so many factors to consider, their personal commitment dates & their other duty dates (which is a lot to be going through on an individual level).  Neo has been helping to try and sort out the duties together because he has more experience than me. Originally, he wanted me to handle it because I\'ll soon be the one taking over this completely soon. But it was just too much for one person to handle.  It\'s impossible to please everyone in the brigade and I thought of taking up more duties. But Neo said it was not worth it as nobody here will appreciate it. That\'s true. I think all I can do is distribute the extra duties amongst everyone (including ourselves) and whoever is unhappy about it, well ... tough luck. But it sure isn\'t going to earn me any friends in the Brigade.  Also realised one thing, sometimes it\'s necessary to be the bad guy so as to get things done. If I ever get into the police force, I think I will enforce this belief. It is not only necessary but crucial to get things moving quickly. Humans are forgetful, most of them will soon forget this unhappy month and just carry on their duties whether they like it or not. To allow things to drag on so long will only serve to remind them of their \'torturous experience\'.  Hope QRF won\'t be activated later. They usually test our readiness one/twice in a night. I wanna sleep .... :(   '),(26134,979795,'14,September,2003','2003-09-14 00:00:00','    What a tiring day the past few days had been.  I did my first guard duty on Fri night. When it was finally over, I was so tired from the lack of sleep. On the way back home, I was nodding off while standing on the MRT. Got some weird looks from people too.  Got back and went straight to bed. Slep for over 3 hrs. Got up and redied myself to go back to camp. We had book in by 6pm coz of the Army Half Marathon the next day. They wanted to ensure we had enought water and sleep the day before the run.  On Sun, we all got up around 230am. I was tossing and turning around 0055am. Probably becoz I wasn\'t used to sleeping in a new environment yet, or becoz I had gone to bed so early (730pm). Took a 5 tonner to Division camp. The 5 tonner is like a truck with little/no suspension system. But becoz of the time, everyone learned to enjoy and sleep on the ride despite the bumps on the road.  We had breakfast in with Division and departed for Kallang Indoor Stadium which happened to be the start and end point.  The competitive runners started at about 0545hrs whilst the non-competitive ones started at 0600hrs. I was in the latter. I paced myself with a fellow campmate, Chan Kai Chong. The pace was comfortable for me and we made it through the first 10km rather easily. After that, Chan started feeling uncomfortable with his ankles and I kept on encouraging him forward. Throughout the whole run I was considering if I should leave him and just proceed on my own instead. But I convinced myself that it was pointless. I think the run meant more if I could see him through the remainder of the run. Towards the last 2km, Chan stopped more than a few times and I had to slow down with him. Partly to make him know that I wasn\'t goin to leave and that I\'d be foregoing my timing if he chose to give up and juz walk.  We both completed the run in 2hrs 35 min, that was just 5 min faster than the timing I had 3 yrs ago. I wasn\'t exactly satisfied with this but I felt a wee bit consoled that at the very least I did the right thing by completing it with Chan. From another point of view, it was probably Chan who helped me maintain my pace. If I ran alone, I might\'ve overburned myself somewhere in the run and not complete the whole distance. All that\'s left is to better my timing and test myself again next year.  At least at the end of this event, I know where I stand. There\'s room for more improvement and I have 12 mths to do it. Ideally I\'d like to complete under 2 hrs but let\'s just take a few mins at a time.   '),(26135,979795,'11,September,2003','2003-09-11 00:00:00','    Today I started my regular Aikido training at Siglap.  Most of the other students were transferred over from students class so they almost all from secondary school and one from primary. A funny lot. Made me feel a whole lot younger (Sensei mentioned that too at the start of the class haha). But they were a cheeky lot as well. Sensei seemed well and happy teaching them. Some of the specifics of a technique may be missed by them, but I did realise that a lot of them were pretty good and soft. Maybe it\'s the age they\'ve started and they\'ve grown to develop a nice foundation for Aikido. I realise I\'m even looking at how they do certain stuff to curb some of my bad habits here and there. And they were pretty soft as well. All in all, it was a good start for me and I\'m happy going back there for my regular training (at least while I\'m still in NS lah).  Did know four of their names. There\'s Leon (Pri 6), Weide (I think), Daniel and another I can\'t quite remember now. They were asking me lotsa questions about NS (I went right after camp so I was in uniform) and I answered them as best as I can. Some of things I told them were the brighter side of things. I think it\'s no point giving them the nasty stuff and direct them towards a very negative attitude of NS. I did give very neutral and fair comments that a lot of things depended on them how they perceived everything and their attitude. Think they were a bit young to fully understand what I was saying as well but I hope that they\'d grasp a bit of what I was trying to convey.  Oh yar. One more thing. NS pay came in today ... so did my eMart credits. WOOHOO! Bought my AHM running attire ... goodness, the V-cut on the running shorts made me feel really \'exposed\'. I miss the Mizuno sponsorship we used to have for AHM runners.   '),(26136,979795,'07,September,2003','2003-09-07 00:00:00','    Got a call on Fri afternoon saying I need to return to camp on Sat for the run instead of going to ECP. ARGH! Camp runs are so boring. Anyway, yesterday I went back and we had a 15 km outcamp run to Pasir Ris park. It wasn\'t as nice as ECP, and there wasn\'t any early fall out but I did manage to go a lot longer than my 1st 15km run a few wks back. However, after the run, both legs felt really weak and my knees felt kinda bruised / sore. I think I\'ll need to take things easy the next few days just to make sure that I\'ll be ok for the AHM next Sun.  This morning I woke up early to go for a quick jog. It seemed to help my legs and knees a bit. Either that or it was the sleep I got lastnight that helped. Later I\'m gonna be going to Comex at Suntec to get my CD-writer. Been waiting for ages to get it. Hopefully can meet up some friends and catch a movie later on. Dun wanna stay home all day again.   '),(26137,979795,'05,September,2003','2003-09-05 00:00:00','    RSM gave us, his 3 stooges (Neo, William, me), the afternoon off. Naturally the morning was spent clearing up last night\'s dinner function in the officers\' mess. Haiz, and wat a mess it was. Anyway, William was damn happy about the off. You could tell that he worked a little faster than normal. I think Neo and I only wanted to clear up what we could before we left. I was seriously pondering if I should stay on instead of leaving. Leaving camp in the middle of a day when there\'s no transport provided is a serious burden on my transport cost (in fact I think I forgot to scan my card when I alighted the bus home... argh!! $2.40 gone.). Decided to go for the off because the 2 guys were pretty tired from lastnight\'s function and they really wanted the rest. If I was the only one who stayed on, it seems a wee bit like I was spoiling market for them. Anyway, William did lompang us all out to the MRT.  I dunno, I can tell some people weren\'t very happy about us getting the off when they didn\'t. I had a serious talk about how the young NSFs feel about things like off. I seriously can\'t stand the way some of them think. Granted some of them are practically 20-ish kids but I seriously dun remember being so particular about whether I get to take off or not. Even lasttime, I rarely asked for any and I\'d just accepted those that did came my way. The only time I asked for off was when I couldn\'t cover enough days with my leave so that I can depart Singapore for my Uni\'s induction programme in UK. I think it\'s normal to be given off by someone else, but a direct superior, due to dept work commitments, is unable to release his men for the off they had been given. What\'s wrong with that? As the SAF always say, off is a privilege, not an entitlement.  And I can\'t stand the selfish attitude some people have in NS. We had to shift some tables back to our bunks this morning. Some would only shift what was rightfully suppose to be in their bunks and leave the rest of the stuff untouched. Some went on to lie on their beds to rest while others cleared up the other stuff. Some would be so scared of doing things one way that they\'d do everything by the book and it can really seriously piss people off as well. Flexibility in a lot of matters is necessary. Without a bit of EQ when it comes to doing stuff, it\'s easy to offend someone. And in this world, it\'s always good to be on good terms with people and just practice a little Aiki-attitude. Do others a little favour because you dunno when you\'ll need help from others someday.  My unit has a seen a lot of improvements compared to when I was around. Vice versa, there\'s a horrible and ugly side which I saw a lot of this morning in the unit. Espirit de corp is not a quality that has been cultivated into my unit. I miss the old days when life was simple and all you needed to do was do things together. Eat, suffer, play, enjoy together. The army teaches us a lot of stuff. Maintaining the right attitude, being an example for your peers to follow - work harder than others if it\'s an area of your forte to influence others to work as hard, help when you can, be cordial to people, even more so to others not in your unit, and especially so where a co-operative effort may be needed someday. Start everything nicely but end them even nicer. Don\'t keep thinking of what posible rewards you will receive because it will only blind you from putting in 100% to finish the job at hand.  A lot of nasty attitude and examples were experience this morning in a short span of just 4hrs. Detailing all of them out will take too much effort and pointless. These are matters I hope to foget when I go to bed. But if people with the right attitude appear a lot more, they do provide a more positive atmosphere for me to work in. These are the only things and people I want to remember when I leave.   '),(26138,979795,'02,September,2003','2003-09-02 00:00:00','    Yes! Exercise Tiger Balm 2003 has ended today. The closing ceremony was a tremendous success. The Americans really got into the whole thing and participated wholeheatedly. It was a great bilateral exercise.  Our RSM seemed to have scored some major points with Division. The performances and various other arrangements turned out exceptionally well. The food was nice too. There was a superb video clip which was made by my unit. There were more than a few requests for copies of the video. Even I feel like asking if I can keep a copy. But I didn\'t get to leave camp until about 10+pm.  Tomorrow, however, is gonna be a major cleaning up process. Think that\'s going to take a brigade effort.   '),(26139,979795,'02,September,2003','2003-09-02 00:00:00','    RSM (Regimental Sergeant Major) wasn\'t around all morning and didn\'t came to camp till mid afternoon. I had to prepare the laptop for some presentation coming thurs. We spent a great deal of time waiting and had to do OT. A bit sianz about the late book out but I reckon I should start to prepare some clothes and leave them in my locker so that I can stay in camp whenever I need/want to. No point going home everyday also.  Had quite a long chat with 2-3 reservists in the day. Shared a lot of insights about how I felt about our education system(one of them being a teacher), past experiences and stories out in the field (which wasn\'t much for me), etc. It was nice.  The next 2 days are probably gonna be the busiest for me. Really have to be ready to stay in if it gets too late. It surprised me even more that the RSM had to go to anothercamp even late at night. I guess this exercise is taking its toil on everyone.   '),(26140,979795,'01,September,2003','2003-09-01 00:00:00','    Went to report sick in camp today. All because of the bad case of rash/infection I had around the time I went trekking in Norway. It\'s been stuck to me for a month already. But it is healing (very) slowly. At least he gave me some cream for it. Seriously hope the rash dun impair my AHM training. I only got lesst than 2 wks to train for it now.   '),(26141,979795,'29,October,2003','2003-10-29 00:00:00','    Had a nice day at Ikea. I expect to return again after I do more organising of my cupboard. At least things look neater now instead of being all stacked up. Tried the Swedish meatballs that Annie had recommended ever so highly :P. It was pretty good. A bit on the pricy side though, $5.50.  Caught \'Intolerable Cruelty\' today too, starring George Clooney and Catherine Zeta Jones. Was ok. Was about a lawyer and a lady who married and divorced pple for their assets. They started out on opposite sides, then slowly fell in love and got together. They tried to trick each other all using the idea of a pre-nuptial agreement. I think the pre-nup is a safe way for these kinda \'weird\' marriages. But to require things to reach an extent like this, I think there\'s a real lack of trust. If that\'s necessary, maybe it\'s not right to get married afterall. It felt nice to see them tearing up the pre-nup in the end tho. At least it meant something, but there was also an earlier similar scene which was faked to lull the guy into a force sense of security.  Just caught a documentary on SAF peacekeeping forces in Timor Leste. It was inspiring, to say the least. This used to be restricted for regulars only, but I\'m glad they opened it up to NSFs and NSmen as well. I would\'ve loved to do this during my NSF period for a few mths. It would\'ve added more meaning into my NSF life.   '),(26142,979795,'28,October,2003','2003-10-28 00:00:00','    Just got back from the combined training at BVCC. Was quite surprised. There weren\'t as many people there tonight as I would\'ve expected. Freddy Sensei took us for the first half and Ah Hee Sensei the second half.  The training wasn\'t like very very very enriching for me, but I enjoyed it a lot. Think I really sweat things out tonight. Probably a result of training in aircon room for too long. I really can\'t wait for Thur\'s class. Wanna train again. And we\'re using our jo too! Whee...  Alright, just made a lunch appointment with Darren Lim, an old Uni friend, to go to Ikea, shope around and we\'ll prob have lunch there too. After that hopefully we can go out for a movie too. Been soooooo out of touch with shows. I think I\'ve got a lot of back lot to watch. Argh, wish I had cable so I can download movies everyday!   '),(26143,979795,'27,October,2003','2003-10-27 00:00:00','    Okie dokie ... feels weird to be clearing some of my leave and off these 3 days.  Took leave for the afternoon, was considering if I should go donate blood. Got an email sometime back saying there was a blood donation drive at Selarang camp so I was thinking should I use my personal leave time for this. In the end when I reached Tampines JC, I thought \'what the heck, might as well since I\'m still in the east side\' so I alighted. Got to Selarang camp and realised there wasn\'t any mobile blood bank there. ARGH!! wasted my personal transport fare and time. And it was only last night that I saw on the news saying there weren\'t enough regular blood donars each year. Haiz ... maybe I\'ll check their website again next time before I do anything like that.  Woohoo ... tomoro on off. Goin Ikea to do some shopping. Hope I dun buy too much coz I still wanna go for the Aikido combined training at BVCC.   '),(26144,979795,'26,October,2003','2003-10-26 00:00:00','    Resumed my Japanese language class in Singapore today. School is called PYAESS.  The first half was just going through grammars by a young teacher. She seems awfully shy/quiet/inexperienced maybe. It was kinda boring and slow. But I figured, what the heck, maybe give her some time and we might all just learn something. The second half was basically conversational practise and some exercises. It was conducted my an older teacher but at least she was more proactive with the class. Much better I would say.  I left realising I knew more of content compared with the other students but more out of practiced. Plus my memory with vocabs and some grammars of what I learnt aren\'t that greath either. Haiz ... looks like I\'ll have something to work on from now on, in and out of camp.   '),(26145,979795,'24,October,2003','2003-10-24 00:00:00','    Yesterday was a 3/4 working day for me only. Evening we had the 9 Div Chain-of-Command Relay. It was really really tough. Our earlier training runs were not as effective as we had thought. It was kinda disappointing as we had only gotten 6th and 8th placing out of the 14 teams. But I felt inspired, we had done out best and given our all for this run. I think the atmosphere and mood gave us a lot of pressure. A lot of us started the run too fast, I know I did. Just 100m or so and I knew there was no way I could keep up this pace and complete the race. To make things worse, the flag we were carrying though light, but was cumbersome. I carried it in a high-port stance similar to the way a rifle is carried. I should have used one hand to hold it instead. The first half of the race our 2 teams did extremely well and we were well ahead in 4th-5th placing. But as we went on we just got slower and slower. I also think that several units actually placed their better/faster runners at the end of the relay. The infantrymen from 3 and 6 SIR were truly inspiring. I wish I could train my legs stronger so I could\'ve put up a better fight. Only 1 guy overtook me, despite my best efforts to sprint towards the end of the finishing line, I was just about to reach him when he gave a final boost and zoomed onwards. My mind and legs just couldn\'t go on. Right after, my muscles from the calves ached right up to my ass and there was a \'numbing\' sensation. Most of us could hardly stand. Unbeliveable.  After the prize giving ceremony I tried to rush down to Siglap for Aikido. Unluckily the aircon had broken down and Sensei decided to end the class earlier at 9:10 pm, the time that I arrived. All I did was help to keep the mats and go on home for a well deserved ultra long weekend. He said I should feel free to visit any of the other dojos to train if I could make it from my camp in time. I\'ll probably try to work something out and pop by the other dojos to train more regularly.  Arrrrrghhh ... time to sleep. I\'m sure I\'ll be feeling the effects of this evenings run in the morning.   '),(26146,979795,'19,October,2003','2003-10-19 00:00:00','    Today\'s training was pretty OK. Considering I\'m still feeling pretty taxed from the guard duty 2 nights ago.  We did 2 slow rounds of a full run-down (with a slight change in route), and then we went for 5 laps of slope training. It consisted of about 200+m I think but it sure was tiring. On the actual day, we still need to run with our unit flag / colours instead of using a baton. A lot of weird rules like the flag cannot go below the runners head, etc. Demerit points will be given and penalty timing of 20sec-2min can be given depending on the seriousness of the breach of rules. Mad ...   '),(26147,979795,'18,October,2003','2003-10-18 00:00:00','    Sooooo tired after lastnight\'s guard duty. I think Thur\'s Aikido class really taxed me physically since I\'ve not done that much slams for quite awhile.  Had a hard time trying to finish my rounds with my SBO on. Shoulder muscles feeling strained now. Even my calves are aching. Got no idea how I\'m going to run tomorrow morning. Prob let my RSM/S3 know I\'m not feeling physically at my optimum and I wanna refrain from getting an injury before the actual run this comin Thurs.  Ohh ... my aching muscles ...   '),(26148,979795,'16,October,2003','2003-10-16 00:00:00','    Today our Bde had a friendly soccer match with another soccer team from 8 SAB, I think. However, they were the representative of the Armour formation and were quite professional. Naturally we lost, 18-0 hehheh. Anyway, the rest of the Bde had to stay as supporters. I became the ball-picker instead :P  I had to rush down to Aikido after that as there was no transport out. I was just on time. Managed to settle payment for the mat fees at the CC (been owing for quite awhile), and rushed on to change into my gi. There were 2 adults watching (father and son I think). I got slammed more than a few times today. Was getting into it, but I realised I tend to fall really hard. More than once, I landed weird on my back (hope no permanent damage was caused :P). Somewhere in the middle, I managed to fall really soft for some not-too-soft-slams. Dunno how I did it, but I seem to have a vague idea. Really need to try it out again at slow-mo to be certain in future. Today\'s training was great! At least I left feeling like I\'ve brought something away with me.  Oh yea, 2moro guard duty. Sianz ... gonna miss the ECP run. Hope to make up for it this Sun at Selarang training run.   '),(26149,979795,'14,October,2003','2003-10-14 00:00:00','    Today I visited the AFS site. Read the updated news of Gombak closing down and the merging of Gombat and BB East dojos. I think Freddy Sensei must be feeling kinda sad about this. Also just realised there\'s a combined training every last Tues of the month. That\'s a fantastic idea. Sensei is encouraging each one to cross train with people they\'ve not seen before. I think this should be a regular practice which should be carried on as time goes by.   urlLink Eileen\'s blog  has a fantastic account of a training session with Tommy. I think it\'s really wonderful. I also believe it\'s a great way to improve the stamina.  I\'m staying in camp again. Somf of the nicer folks in camp are doing overnight duties today so I though staying back shuold be quite interesting. Anyway, probably going for a quick run around 8:30pm later since this morning\'s life-run was cancelled due to bad weather. Again, the whole Bde got screwed up by the RSM. Over discipline and performance again. I think he made really valid points about the discipline in the Bde. Problem is not many people realise it till it\'s too late. When that happens, we all deserve the lecture we\'re getting. Even I realised my discipline slackened down. Before the Sun run at Selarang, I did take my own sweet time to get to camp even though I was already late. Instead we the juniors made the officers wait for us. Of course the weather was bad but still it didn\'t project a good image. I guess some things can be quite \' wayang\' (just to put up a show / front for people to see), but image is pretty important too. We all know how kokup things can be in NS, but if we let foreign powers know every knitty gritty mess we encounter, I think we\'d have been attacked and conquered a long time ago. Professionalism is not something found in every organisation, or every level of an organisation, but it definitely needs to be projected to external parties. It\'s like a hawker ctr and a foodcourt thing.  OK almost time for the run. Ciaoz.   '),(26150,979795,'13,October,2003','2003-10-13 00:00:00','    Yesterday was a great Sunday.  I joined the training run at Selarang camp (my previous camp). We\'re preparing for a relay on the 23 Oct. There\'ll be a no. of reservists and some active personnel participating. I\'ve only just realised I\'m going to have to miss the training at Siglap then. Argh!! That is unless someone gives me a lift down immediately after the run. Drat!  Anyway, we did a full run-down (900m) twice, followed by 100m sprints thrice and a final run-down again. After each sprint, the losing team (we were divided into A and B) will need to do 10 push-ups. All 3 times, my team (B) kanna pumped ... haha. For the last time, Team A joined in as well. This training run showed me that the reservist officers have a great deal of endurance in them despite their age. For the first 2 run downs and sprints, I was able to keep in front one of the older reservist. But the final sprint and run-down, he was in front of me and I was all but depleted both physically and mentally. It was a truly great session that morning.  I think besides feeling tired, everyone left that morning feeling really good.   '),(26151,979795,'11,October,2003','2003-10-11 00:00:00','    Today was Enlistee\'s day. I woke up really early and left early. Got there way to early and there wasn\'t anyone there. To make things worse, it was raining so heavily. I tried to explore the area to look for somewhere to grab a bite. After walking around in the rain for awhile, I decided the closest spot was too far and I turned back. By the time I got back to the gathering pt, almost everyone had arrived except for a few laggers here and there. Finally the event was cancelled but breakfast had to be consumed as they had already been indented (ordered, to u civilians out there). Me and some of the guys wandered around looking for somwhere to hang out and decided to goto the Kallang Leisure Park. 2-4 of the guys played bowling while I just watched and napped. After that we all went back home. Kinda a waste of the morning, was looking forward to trying the Dragonboat thingy since I was already there. Damn the weather. Anyway, Saw RSM just now. I decided to join in the 9 Div Panthers road relay as a reserve. Going to go back to Selarang (my old and previous camp) for training the next few Suns.  In the afternoon, met Darren (Uni friend). We went down to sign up for the same Japanese language class. It\'s going to cost 200+ for about 10 weeks, 1430-1730hrs every Sun. Hopefully I can pursue this longterm. I think it starts 26th Oct. I might\'ve remebered wrongly but it seems to be clashing with the Panther\'s Run.   '),(26152,979795,'10,October,2003','2003-10-10 00:00:00','    My guard duty on Tues had left me really worn out for the whole week. Maybe it was due to the 2 life runs we I had before and after the guard on Tue and Thurs. A warrant was pushing us to run under a certain timing each 3km run. I managed to get a better timing on the latter training but it was still pretty slow to me. My next guard is going to be next Fri. Yawwwwwnz...  This coming Sat is our Div\'s Enlistee Day. They\'ve organised some stupid Dragon boat competition at the Singapore Dragoin boat Association at Kallang. Heard we\'ll be rowing a 1.2km race. Sounds interesting lah but considering it\'s a Sat, I\'d rather have a simple run in camp instead. Anyway, no choice also.  Eileen has a nice saying from Tommy on her  urlLink blogpage . It\'s meaningful and it reminds me not to be too judgemental of others, a common tendency.   '),(26153,979795,'03,October,2003','2003-10-03 00:00:00','    Yesterday was a relatively great day. Met Lynette for dinner in Tampines coz I had a few hours free before Aikido. We had a good chat and I was really happy to hear that she\'s got herself attached just recently. Can tell that she\'s beaming the whole night from the conversation about her new beau... haha. Think it\'s about time and she definitely deserves it.  Was thinking back since the day I knew Lynette back in 2000. I remembered it was still in NS, prior to our departure to UK. I got no idea how, but we had chatted on mIRC and then ICQ. Somehow or rather we met in UK after that. The person I knew then was really quite temperamental. Most of us would be \'afraid\' of offending her somehow. But now, she\'s really quite different. She\'s a good friend and she has controlled her temper a lot. Her willingness to go out of the way to help a friend and standing by her principles are just some well noted points of her strong character. She deserves everything that life has given her because she\'s worked hard for it. A good listener whenever anyone of us needed a ear to grumble to. Hah :P  After this long good talk, I rushed down to Siglap for Aikido. Today\'s class was good. It incorporated soft feeling techniques (morote dori ikkyo and nikkyo waza, if I got them right) and some jo equivalents as well. I could actually feel the techniques working. I managed to co-ordinate myself for some techniques to blend wif my uke. Kokyuho was done with Bridgette (spelling??), I noted she has a very strong way of doing kokyuho. Very good and effective. She\'s able to \'feel\' the strength I\'m exerting out and divert them to move me around, I\'d say, pretty effectively to a certain extent. But I can\'t do it well. So much I can learn from a sec sch kid. Puts me to shame man. Haha ... Still, it was a good night.   '),(26154,979795,'01,October,2003','2003-10-01 00:00:00','    OK, gotta post some comments about the grading on Sun. It was interesting, (tiring for most of them), but I have to say, it lacked the extra \'oomph\' that I was expecting ... Not enough killing!! Muahahaa!! I think my favourite part was still free practice for the 1st kyu grading.  But I\'m still happy for those who had taken their grading that day. Hope they excel even further in the days to come.  Yuishinkai\'s seminar will be starting this Thurs. Having applied for a full day\'s leave for Thurs, I\'ll be attending the afternoon session from 2-5pm. My friend had suggested that I drop down for the morning session for a look even if I\'m not training in that slot. It\'s a good idea but I\'m not sure I can wake up on time. I\'m also planning to go down to Sim Lim to buy my PCI firewire card for my PC. It\'s high time I started burning some of my videos from UK and the grading on Sun into CDs. Hope I can finish this fast instead of procrastinating.   '),(26155,979795,'26,November,2003','2003-11-26 00:00:00','    OK my third visit to SNEC today. Good news is, it\'s only a bad eye infection. Hmm ... the redness and irritation apparently seems to be caused by one of the eye drops I was given. THey told me to throw everything away and they\'ll give me something else. Can tell the difference between an experienced and less experienced Dr today. The more experienced one would say, \"Don\'t worry. It\'s nothing serious.\" \"Just use xxxxxxx, and the redness will go away. You\'ll feel much better after that, guaranteed.\" Was quite elated lah. But during the checkup my medical docket seemed to have gone missing for awhile. My father made a bit of fuss over this. I just kept quiet and didn\'t want to interfere. If I did, it might make things worse for the nurses there. Sigh, I pity people in the service industry sometimes.  Anyway, I was given 6 days MC. Looks like my Sat duty need to be swapped with my friend\'s next month. Drat, and I was hoping to clear up everything before Dec so I\'ll be a lot more free then :P Oh well. Things are looking up at least. Next checkup due on Mon. Hopefully I can go back to Aikido on Tues and get some last min prep for the marathon next Sun.   '),(26156,979795,'24,November,2003','2003-11-24 00:00:00','    Latest news after my visit to the Singapore National Eye Centre this morning. Just abad infection. I was told I should have kept on applying the antibiotics despite the discomfort I was feeling. OK No choice, each time I just suck it in and bear the pain during my first application. At least the goodnews is that I only need to apply it every 2hrs now instead of 1 hr. My next appointment will be on Wed afternoon. Meanchile, I\'m going to just make myself comfortable in bed.  Had to email Marcus and Ivy that I won\'t be able to meet them at the airport on Wed morning coz of the eye thing now.  Oh yea, imagine my pleasant surprise when an old poly classmate just emailed me on Friendster. I immediately forwarded it to our class members. Maybe after I recover I\'ll ask him out for coffee or something.   '),(26157,979795,'23,November,2003','2003-11-23 00:00:00','    Today my eye feels like it has gotten worse. I was fine with the medication lastnight till even this morning. But after I woke up from an afternoon nap, it seemed to have gotten so swollen and stight that I can barely keep it open. It feels so tensed and I\'m not sure how I can relax it. Medication don\'t seem to be helping so I dare not apply anymore. It gets so tight now that tears can\'t flow out freely and I gotta use my fingers to keep my eye lids open while my eyes tear. Will be seeing the specialist tomorrow and I hope there\'ll be good news or a change in medication.  I hope no extension of mc is required coz I\'d hate to make someone from camp takeover my sat duty. I also hope I won\'t need to be warded after that either. Really hope there\'ll be some good news tommorow. I doubt I can meet up with Marcus, Ivy and baby Zelda at the airport on Wed morning if my condition persists.   '),(26158,979795,'22,November,2003','2003-11-22 00:00:00','    Went to the Polyclinic to check out my eye thoroughly this morning. The doctor would\'ve referred me to the Singapore National Eye Centre but being a Sat and all, it would\'ve been closed. He had no choice but to send me to the A&E. So I went to SGH A&E where the doctor wasn\'t able to see anything wrong, again he referred me to an eye specialist. Well, after all the referrals and waiting (not complaining about these, I\'m kinda used to this all thanks to NS), the verdict is that it\'s most probably a bad eye infection due to my contacts. First off I going to have to stop using them, second I was given 3 different type of eye drops; 2 of these were antibiotics and 1 of which was hand-mixed by the specialist herself (suppose to be stronger). From now till first thing Mon morning, I have to use the eye drops on my right eye every hour, with 15 mins between each different type of eyedrops. Was quite surprised because I didn\'t feel that it was that serious. But turns out I needed to follow this. The specialist even asked me if I could do it, otherwise she\'d ward me in the observation ward. Looks like I\'m going to lose some sleep for the next 48hrs until my 2nd appointment with the specialist. She was kind enough to slot me in first thing in the morning even though she already has a full schedule.   '),(26159,979795,'21,November,2003','2003-11-21 00:00:00','    Had a most horrible day today. Spent awhile at my aunt\'s/uncle\'s place trying to link up a spare PC and laptop into their home network. Wasn\'t successful. Apart from the obvious stuff I changed, I have really no idea what is wrong with it that I couldn\'t get them to go online. But this has been draggin on so long now. I\'m very sian of doing this shitz now. I hate maintaining PCs. I don\'t mind doing it when it\'s needed but not setting up stuff and new technologies. Think everyone is kinda taking this for granted now. If I can only find a permanent fulltime job, I\'ll definitely refuse anymore of this requests.  I got home soon after. My dad wanted to view a TIF document. Simple question which lead to a whole big arguement which lead to a lot of hollering, shoving and throwing of stuff. I didn\'t help matters by the tone in which I replied him. Neither did he want to listen when he\'s at the peak of his temper. Got a chair thrown out of my room (damage to the floor, nothing to the chair). During all the shoving around, my mom knowing that I had a bit of sore eye on my right eye tried to stand between the both of us to protect me. In her frustration she grabbed the nearest object to throw on the ground. Bad mistake, turned out it was a $3000 piece of engineering equipment my dad had loaned from his principal co. in Australia. Shitz ... I guess we don\'t have to pay it back anytime soon coz it\'s on long term loan but still ... ~$~  I let him get to me again. I find it so hard to shield out the comments coming from my family unlike those coming from an outsider. Each of us had a part in the way things got out of hand. It\'s a frustrating moment. Things cooled off pretty quick but I still find it really irksome to spend time at home rather than in camp. I\'m considering spending time in camp even when I\'m on leave. No point coming back and making things so difficult on everyone right?   '),(26160,979795,'20,November,2003','2003-11-20 00:00:00','    Lastnight I went for Aikido at BB East. Oh man, seriously, I think age is catching up with me liao. I\'m aching all over. Plus the past 2 nights we had to do situps and pushups.  BB East\'s class was a comfortable size despite combining with Gombak students. I liked it but I think Tommy was right. The aircon isn\'t a good idea. Too comfy for me. I think I might just continue at Siglap and drop by BVCC/BB East on just 1 more extra day when I can. This way I supplement my training and I still get to train with the young kids/teens at Siglap. Eh ... but it still feels weird to not be paying mat fee at the dojos I\'m popping by. But at least given current $ situation, it\'s not a bad thing either. :P  Got my Medical FFI back in camp this afternoon. Woohoo, 1 more milestone towards my ORD!   '),(26161,979795,'18,November,2003','2003-11-18 00:00:00','    Today\'s Aikido was GREAT!! I can\'t say I did achieve really fantastic stuff or understood everything, but I definitely, absolutely, totally enjoyed it. I have nooooo idea why.  Was a bit late again missing warmup and all that. Not a good idea coz it\'s the same both times I popped down to BVCC that this had happened. Really need to keep my timing in check.  Today was the first time I paired up with Sunny (kowtows to Nidan ... capable of some really hard throws). I can tell you every iriminage he did on me, I didn\'t get what he was trying to get across to me. Seriously ... I only thought I understood it the final time he did it and then it was over liao. Wasted. Didn\'t even manage to take a decent ukemi for him. Fell like a piece of rock each time and I know I could\'ve hurt myself badly if he really wanted to throw me. Dun understand it, but I still totally enjoyed it. Each iriminage I did on him he could\'ve done a reversal. OK my iriminage was craaaap. Absolute shitz ... but I still enjoyed it. Dun understand it. No frustration at not \'getting it\' either.  Today\'s a weird mood of training for me... but a happy one.  Out\'ve the blue my mom was in the neighbourhood to pick me up. On the way back in the car, we actually had a decent conversation. I\'m still quite skeptical about how long these moments can last but it still feels nicer to be talking rather than hollering at each other.   '),(26162,979795,'17,November,2003','2003-11-17 00:00:00','    It\'s another week of leave for me. Time to get some old chores cleared up. Will also be intensifying the training over the next 10 days. Hopefully I can train more often.  I was watching the show Glory on HBO this morning. It\'s basically about the first black infantry regiment that was formed during the American Civil War (North vs South). I realised a lot of things that our current NSF are taking for granted. Back then, they were despised so they weren\'t issue guns, uniforms and weapons for training till much much later. Now, the moment we enter BMT, we get our standard issue items almost immediately. Each one of us just put on the uniform like any piece of clothing. Some don\'t even bother to make themselves neat and tidy in uniform. Just 2 days back, I was on the morning MRT to camp and I saw a PTE wearing his no. 4 camo uniform but instead of having his boots on, he had his PT shoes! He looked like a stupid clown. It was a disgrace to the uniform and himself. Naturally he looks like he might\'ve gotten some medical excuse for his foot so he needn\'t put on boots. But c\'mon man, everyone know that you can\'t wear this outside of camp regardless of medical excuse. He should jolly well change out of his uniform and slipped into more comfortable footwear instead. I was really pissed when I saw it. Really. Felt like asking him which unit he belonged to and lodging a formal complaint thru Mindef. But I also thought, this is prob one of those psycho people in NS who has the wierdest thinking. Signing extra duties will hardly change them. It\'ll only make them more resentful of their NS life.  The black soldiers trained in their normal shoes (for those who had them) until their foot were cuts and badly bruised. When the South heard that the North were recruiting slaves into service, they passed a law saying any black soldier in uniform who got caught will be punished and the commander, put to death. This was something like psychological warfare to stop the blacks from enlisting. The Commander was expecting most of the current no. of blacks in camp to leave service but not one of them did. How many NSF today have the kind of, or even any, commitment to country? They only see it more as what NS can give to them rather than vice versa. Didn\'t John F. Kennedy say the famous words \"Ask not what your country can do for you, but what you can do for you country.\" These are just empty shells to the average Singaporean. When push comes to shove, not many will choose to stay on by choice. It\'s quite heartbreaking really.   '),(26163,979795,'15,November,2003','2003-11-15 00:00:00','    It\'s another duty day in camp. I\'ve been really bored but I used most part of the afternoon surfing around and checking out Aikido sites. I realised it\'s been a long while that I had been reading in on forums and articles. As I surfed on, I found many articles of interest. Some of which beared some relevance to thoughts and misgivings I have about attempting the coming grading next month. I\'ve been feeling indecisive about whether to attempt it. Most part of me feels that I shouldn\'t.  I came across some articles at AikiWeb and I began trying to evaluate _my_\'fear\' or unwillingness to attempt a grading. It dawned on me that at 6th kyu grade, I do feel a \'wanting\' to attain 5th kyu. But I feel pretty often that I am unable to handle, execute and receive techniques. It could be due to this lack of confidence, but a lack of confidence to do what? To pass my grading? No, because if Sensei would allow me to register for it, it probably means that I have met the standard to take and pass the grading (how well I perform on the day itself is another matter). It may be the confidence to live up to the next level of performance. I do feel more confident just practicing with juniors and peers and I may be trying to avoid facing the challenge of the next level. It sounds very much like I\'m avoiding what I\'m not good at instead of facing up to it and trying to overcome it instead. I think I feel a bit apprehensive of facing the responsibility of holding the next grade as well. Such that if I do a technique worse than a more junior grade, I won\'t feel good about myself. That sounds very much like a struggle on ego. All because I do not feel good appearing like a substandard Aikidoka of the grade I am holding.  It shouldn\'t be that case. An uncle once told me something which I interpreted as \'Do not let the success of others intimidate / demoralise us. We should, instead, use their success to inspire ourselves onwards.\' I\'ve remembered the meaning conveyed in many instances and I thought I did as well in my Aikido training, but I realise now that it might not have been so.  Enlightenment: I will never be the good Aikidoka I hope to be wherever and however I train, but it should not stop me from trying to be.  I will always be a substandard Aikidoka of the grade I\'m holding, but as soon as I become the substandard Aikidoka of the next grade, it makes me a better Aikidoka which I\'ve been at the previous grade. To put the arguement into a more personal context; I\'m proud to be a substandard 6th kyu. When I become a substandard 5th kyu, it means I am an improved substandard 6th kyu and I should be glad about having improved. Rather than compare my current standard of Aikido with others, I should be trying to attain the level of Aikido the better ones are practicing. \'Looking bad\' itself is a great lesson on humility on the mat.  We should not let our current state hold ourselves back from the next.   '),(26164,979795,'11,November,2003','2003-11-11 00:00:00','    Argh caught Matrix tonight. Seriously man, I really felt that each epsiode seems to be getting worse. I didn\'t find it special anymore. And so many people had to die in this episode. Instead, the extras don\'t seem to be really dying ... duuh.  Before the show, I visited Ajisen Ramen for the first time. Spent about $15 for dinner there but it was pretty nice and filling too. Mmmmm....   '),(26165,979795,'11,November,2003','2003-11-11 00:00:00','    Today\'s Aikido at BVCC was nice and enjoyable. Came a wee bit late during/after warmups coz I had to hold the fort at home while waiting for our new fridge to be delivered. By the way, the old one konked out on us yesterday. Had a hard time clearing some of the food. Let\'s just say I went to bed a bit heavier and feeling satisfied.  Wasn\'t the only one to be abit late. Tommy and Cody came while I was changing. We did shomenuchi ikkyo, iriminage and I\'m sure there\'s something else that I just can\'t remember right now. The second half was a lot of suwari-waza. nikkyo and sankyo. Felt a bit more balanced than lasttime. But still had a lot of problems. It\'s weird though, I never used to feel that suwariwaaza was fun, but I did today. Maybe things just felt they were in place... dunnoz. We kinda ended with kaitennage (and of course kokyu ho). Again, my kokyuho felt really crappy. All in all, it\'s been fun. Looking forward to Thurs\'s class at Siglap.  Oh yea, got home just in time to see the new fridge. Wooo...   '),(26166,979795,'08,November,2003','2003-11-08 00:00:00','    Had a bit of bad news from UK yesterday early morning. Apparently 1 of my old housemate has broken up with my friend (one of the Navy regulars). Sigh, it was totally unexpected. Think she\'s feeling damn bad about it now. Can\'t say I dun totally understand how it happened but I still felt really sad.  As 2 of my ex-housemates are still staying together in UK now, I kinda had a quick chat with the other girl to see if she could help talk to her. Apparently things haven\'t been exactly great between these 2 girls as they\'ve been drifting apart since last year ever since one of them got into the relationship with my friend. (Hope it\'s not getting too confusing at this stage coz now I\'m talking about the 2 girls and there\'s nothing to do with my Navy friend.) Somehow as we talked, I made the other girl felt like she wasn\'t putting enough effort into trying to make their friendship work.  But contrary to this I only said the things I said because she was always the stronger one, more sensible and with a more open heart. I think I must\'ve really made her feel like I wasn\'t supportive/was laying some kind of blame on her because she sounds kinda upset even today. I feel so bad now and I\'ve tried desperately to explain to her what and why I said. I\'m really at a loss now coz I wouldn\'t want to lose this really good friend.  This has really spoilt my day because now I can\'t keep this out of my mind. Plus I\'m really scared that this incident might\'ve scarred this friendship bad enough to bring it down a level, something that I\'m really afraid of but feel so helpless to be able to rectify because of the fact that we\'re on separate continents. I may be scaring myself unnecessarily but the more I think about it, the more I feel that\'s the case it is now. I\'m really feeling down in the dumps and disspirited.   '),(26167,979795,'08,November,2003','2003-11-08 00:00:00','    OK and update on the situation. Somehow things kinda died down. The bunk didn\'t get any extras over the whole bunk deal. But turns out the \'suspectee\' (??) went to see RSM about some things and it didn\'t turn very pleasant. Heard he was complaining and commenting about making complaints to MP, etc. Goodness man, seriously don\'t know how NSFs think these days. I\'m pretty disturbed byt what I would kinda describe as \'mental\' cases serving NS. Even one particular clerk in my unit has the weirdest way of thinking about responsibility. People who took a lot of MCs means there\'s a requirement for him to dig out their dockets and file in medical paperwork. For people who don\'t take too many MCs often, he feels that because he doesn\'t touch their dockets too often, hence when it turns out some document went missing, it\'s these people\'s fault and not his fault for losing the documents. ??? How on earth is that logical? ANd he has the nerve to say things like \"I could\'ve complained to my psychologist, who happens to be an MP, and these people will get into trouble.\" Oh c\'mon man, I\'ve not heard a more ridiculous logic. If an MP is willing to carry this case forward, I\'d classify him/her as another nutcase.  These 2 guys are harmless really. It\'s just the way their thought process works that\'s really disturbing. It\'s not that I\'m unsympathetic to some of their cases, but I find it increasingly frustrating to try to talk to them. Most of them are in a world of their own. Whenever I speak to them, they can hear me but they don\'t really listen. And I do worry for them like how they\'re going to cope after they ORD. The world outside is not as forgiving or sympathetic as the SAF.  Well I\'ve been sick the past 2 and a half days. Still went for Aikido on Thurs and it felt really good. Somehow I just didn\'t feel as sick after that. Going to camp was kinda difficult as the medicine made me drowsy pretty often.  Realised the Singapore Marathon is drawing closer and I still haven\'t clocked up enough distance to feel ready for the run. Now that I\'m sick, I\'m not sure I\'ll get enough traiing inside to be ready in time.  Eileen, also just reminded that grading is drawing near. But no preassure on that. I feel so happy training that I just want to continue attending lessons rather than go take the test.  Oh yea, I\'m still eagerly waiting for some kinda response from the Police Force. Dec is dragging really close now. ORD loh.   '),(26168,979795,'04,November,2003','2003-11-04 00:00:00','    Gotta post this in a separate blog instead of updating the previous one.  The whole bunk went into the RSM\'s office to get to the bottom of the matter. Apparently some people have their suspicions of who the culprit really was. There wasn\'t anything said outright but a lot of indirect talknig behind the back has been going on for awhile. RSM explained the seriousness of the situation and the procedures of how he would handle the matter.  One particular guy had a lot to argue back. I think he feels really indignant about the whole deal. He feels very strongly that he is annocent and that he should be allowed to go for his evening class. Of course lah, I think a confinement is a small issue which we can / should handle ourselves within our unit. No point making the whole mess blow up to even the SAF Counselling hotline or watever. Sigh, NSF these days are having so good life in camp that a little bit of regimental punishment makes them feel like a huge injustice has been brought to bear upon them. I have no idea whether he is innocent/guilty because of his response. But what surprised me was another storeman standing out to take responsibility for the offence. He gave ambiguous facts about his motive and events which led to him playing such a prank. But I don\'t think he was the real culprit. I think he was just saving everyone the trouble of a confinement and having the matter closed. A brave lad if I do say so myself. But RSM has no choice but to take into consideration about what he had said. Tonight he has released everyone and he will come to a decision tomorrow instead about hsi next course of action.  After we were dismissed for a few minutes before we reconvened, I did a bit of explaining to this storeman. If the real culprit gets caught and it turns out that he had lied about being the real culprit, he might be charged for lying in the SAF. (This can happen if RSM finds that he doesn\'t like being lied to, or if the RSM decides to make an example of him to show that anyone who tries to \'help\' others by taking the rap for an offence, that this may be one of the outcome.) Especially pointless since this guy has plans to further his studies after his NS. It would be bad if he gets some kind of blemish on his record over this matter, especially when he\'s not the culprit.  After our 2nd meeting, he asked to speak to RSM privately. I have no idea what it\'s all about or what\'s the outcome of the matter. But I think there\'ll be some decision tomorrow morning.   '),(26169,979795,'02,November,2003','2003-11-02 00:00:00','    Sianz ... today I\'m on duty and some shit thing happened. My whole bunk got indefinite confinement from today onwards.  A storeman in my bunk got his locker locked up by some unknown bugger. This storeman went to complain to the RSM who asked for the culprit to own up. When nobody did, he decided to confine our whole bunk until the culprit surrenders himself. Goodness. Being the duty spec today I\'ll have to stay in anyway but I\'m wondering if it drages on till Thurs, how can I go for training at Siglap. And what\'s going to happen if it drags on till Sun? How am I suppose to go for Japanese class.  I can only hope for RSM to change his mind within the next 2 days or that the culprit will own up soon instead of dragging down the other guys. That\'s regimental life for you.  What a malicious and brainless prank to pull while you\'re still in the Army. Good gawd...   '),(26170,979795,'02,November,2003','2003-11-02 00:00:00','    ~@##^%^#$!@ Had a really heated argument at home today.  An account of the issue: Mom made lunch and left her portion to cool on the table. I finished up mine. She distinctly told me that\'s hers. After awhile, my dad came down and asked if the plate was his. I have no idea why but I somehow felt that my mom had finished her share and the portion on the table was his, so I said she\'d eaten and he went on to finish the food. His share was actually still in the pot. Halfway thru the 2nd serving (the one from the pot) my mom came down and made a helluva ruckus about the whole thing (her share being eaten). My dad gave her what was left. She just kept on yelling (high pitched and irritating volume) about the whole thing and kept poking (not physically) my dad. He of course yelled back. I also joined in the yelling and I actually got so heated up I called her a b***h. Fact was, more than 1 profanity was flying around the house not just from me (though I don\'t think my mom said any).  Went upstairs to prepare for my Japanese class later. And my dad came up to scold me that no matter what she was still my mom and that I shouldn\'t have behaved that way. Bottomline: I call \'em as I see \'em. Not denying my fault in the matter, made the mistake of telling my dad my mom had eaten and being rude to both of them. But to me, that was the only way I managed to get my mom\'s attention and to stop. My dad, being a sensitive man will probably keep harping on the issue even after me and my mom had forgotten about it.  Conclusion: Think we were all in the wrong somway or the other. I can\'t say who made a more serious mistake than who, but I can only say I felt like shit after my behaviour (though I didn\'t show it to them). But guilty or not, I can\'t say there\'s a solution to how they should be spoken to. Mom is in her own world and points take 3x the effort to get to her (if it even goes in at all). Father is in a world of his own too and talks in circles (he may have his reasons for doing things but the way he does things hardly works out. He is so keen on pointing out things to others he overlooks his own problems).  Course of Action and the rationale behind it: Made a point very clear to me in my head, I\'m not going to let them affect me anymore in this way. From today onwards, I\'m going to switch off my frequency to them. There\'s no point in communicating when noone knows how to. Plus I\'m no longer bothered about whether we do or not. My friend, Neo, said what will my kids pick up from the way I treat them. I only told him I\'m not condoning what I\'m doing. But with my kids, I\'m going to teach \'em another way. If what and how I teach to the next generation don\'t work, I\'ll accept my failure in this area and I will live with whatever the consequence maybe then. It\'s a set of double standards I guess, I won\'t stop trying with my kids, but I\'ve absolutely given up on my parents. They may have brought me up, but I\'m still going to take care of them in future as best I can, just don\'t expect too many words to be coming out from me anymore.  There you have it folks. The true scenario at home. It\'s been like this for the past few years. If anyone has anything enlightening to share, please do. But I doubt there\'s anything that can be said to change my course of action anymore.  Going off for class now.   '),(26171,979795,'01,November,2003','2003-11-01 00:00:00','    Recently joined this really interesting thing online called  urlLink Friendster . How it works is that you\'d need to first register as a member (free). Using email and/or names, members get to search and add their friends into a personal network. Once added, you can view your friends\' friends and so on. The interesting is that sometimes you\'d realise a lot of you actually have a friend in common. What\'s unique about this system compared to ICQ/MSN/Yahoo messengers system is that you get to see how other people are linked to you. Eg. Me -> Person A -> Person B -> New Friend.  Feeling a bit out of sorts today. I think it was because of doing Guard duty on the last day of the month. It doesn\'t feel like it\'s my last duty for Oct because coming week I\'ll have another duty soon. In fact for this month, I had a really hard time planning out duties for the Bde again. It\'s so hard to be fair because there\'s just way too many things I needed to take note of when planning. It\'s getting so tedious and yet I dare not really slack too much on this as it will make a lot of people\'s life miserable if things weren\'t planned out properly.  I\'m looking forward to the last day of Nov as that\'s approximately the day of my last duty again. But that\'s also my last duty for my NS. Wheee!!!   '),(26172,979795,'31,December,2003','2003-12-31 00:00:00','    Saw Return of the King yesterday. Was good but I would\'ve expected things to be much better. Think it was still good. Now I\'m really tempted to read the book and getting the extended / Director\'s cut on DVD and watching all 3 shows again.  Was feeling lazy on Mon and didn\'t go for Aikido thinking I could go on Tues instead. But lastnight I was out with my cousin (after the movie) and so I missed practice again. Haiz ... must be getting a bit lazy these days.  Found out today that my dad\'s Pacnet is on a 1 yr contract and we have to pay about $300+ if we wanted to cancel it. WTF! And to think I had asked him whether we were on a contract or not. He nonchalently replied no, making everything sound as though we had the better end of the deal. Oh goodness, I can\'t even upgrade the plan to broadband according to Pacnet. Argh ... feel like puking blood. Now the 2nd alternative would to maintain the co. a/c, and cancel the pte a/c instead and replacing that with SCV\'s connection. But why on earth would we need 2 unlimited a/c?   ... looks like no broadband/cable for the next 10 mths at least. Might have to discuss with my sis to see if she\'s willing to pay the extra. But then again, we\'ll still need to setup our own wireless network and the equipment seems like it\'s gonna cost us a onetime amunt of $500 or so, excluding subscription plans. ARGH!!! This is so irritating.   '),(26173,979795,'28,December,2003','2003-12-28 00:00:00','    Uneventful day today. Sorted out my photos which I had taken in UK on my trips and graduation. Took me till the night. Besides ordering them and deciding which ones to develop extras, I started putting some of them into the new album I bought yesterday. Halfway through, I remembered that I had yet to scan in all of them. Hmmm ... it\'s too many for me to finish one shot. Oh man, sometimes maybe a digital camera is the better alternative afterall. At least I can just copy into my hdd straightaway.  Shouming (cousin) has already found a job with NCS. Mom was nagging me again today about getting a job. Also told her I didn\'t get shortlisted for the police. That kinda got her going again... nag nag nag nag nag. Oh please, it\'s not like I\'m not looking for jobs. She just like to build up everything based on assumptions. Well, let her assume then ... I just want to be left alone. Preferably with zero interaction. Told my dad I\'ll find and take whatever comes along first, but I\'ll still reapply for SPF coming Feb. He asked if I really had to insist on the SPF. I said, it was my interest and we kinda left it at that. I\'m getting a lot of grief from them these days. I also don\'t want them to have the excuse to try to psycho me out of applying for the SPF. Trying hard to not let what they say affect my decision. Naturally I recognise the need to get a temp job first. I also don\'t want to lobo around to be nagged at for the next few months to a year. Prob find something / anything that comes along. Even if it\'s a perm job I\'ll still take it. Just in case SPF rejects me again, at least I can fall back on this and continue on while waiting for the next application date. I guess that\'s the tentative plan I have now.  Will sort out all resumes, CV and copies of result slips first. Apply whatever comes along. While awaiting for reply, will need sort out the littel details at home, clean up my room, clear the junk. If time permits, think wef 2004 I can attempt volunteering at the SPCA on weekends/Sun. OK, got a plan in order now, time to put into action.   '),(26174,979795,'27,December,2003','2003-12-27 00:00:00','    Lastnight\'s dinner at RSM\'s place was OK. Our Ops WO actually had a bit too much to drink and was being all rowdy when we left. Haiz ... not the first time I\'ve seen / heard of this. A bit too much sometimes.  This morning I called up the SPF. Had some disappointing news lor. The lady on the line said that they\'d already replied to shortlisted candidates. If I\'ve not gotten anything it means that I\'d not been shortlisted. I was really keen to enter this batch of intake because it\'s in Mar 04. 9mths of training will end around Dec 04. Anyway, I asked for the next intake which is in Jun/Jul. Recreuitment will start as early as Feb. Looks like I have a good 2 mths before my next application and 6 mths or so to the intake date if I\'m successful. Hope things will work out fine this time.  In the meantime, I\'m going to have to take up another job first. Hmm ... I wonder what\'s available and suitable...  Today\'s Aikido at BB East was pleasant. Didn\'t feel like I absorbed a lot a lot though. Ah Hee Sensei was there as well. At the end, paired with  urlLink Eileen  again for kokyuho. ARGH!!!... again I can\'t grasp what I\'m doing wrong. I can feel my upper arms getting tired so I was using strength instead of actually doing it the right way. Still, can\'t get my iriminage to work properly though, but I tried the \'keep your\' hands in front\' thing and it kinda helped a bit.   '),(26175,979795,'25,December,2003','2003-12-25 00:00:00','    Yesterday evening, my family all went over to my aunts\' and uncles\' place for dinner. It was quite nice but my appetite wasn\'t particularly good. Maybe coz of the cough I\'ve been having. Haven\'t seen my cousins for ages. 1 went to study in Canada, heard she\'s doing pretty well. Haha, apparently she had some moles removed. Haiz ... charbors ... Haha but it was nice seeing long-lost cousins again though. Another was finishing her final year in NUS. She brought her boyfriend along. Quiet guy and quite cham coz he was pratically in one corner and not many of us were talking to him. Heh... I was too busy sleeping to be bothered with the others. But her 14 yr old brother!! Good gawd! Haven\'t seen him for a few years and he\'s already 1.8+m tall!! Pengz ... I wanna kill myself liao. Now I know where did all the tall genes go in my dad\'s side of the family?! Oh, can\'t blame me for falling asleep there. They were playing one of their church service ... but it was in Hokkien and the pastor had a Cantonese translator   ... Hey, but at least I did manage to understand the gist of everything. Anyway, I also didn\'t mean to fall asleep there. It just ... sorta ... happened. Haha ... the cushions I was lying on felt soooo comfy ... zzzzz  My aunt\'s side of the family also invited some relatives. Their daughter is currently doing her degree in Manchester. Heh ... good lah, heard she was quite havoc in S\'pore a few years back. Hopefully mellowed down a  bit by now. But one thing that struck me, they seem to have the mentality that overseas is much better than Singapore. I dunno, I think it\'s a real misconception lor. I\'ve always encouraged people to broaden their horizons but ultimately I\'d always say we have to come on back here. I dislike the notion of people going abroad and deciding not to come back. This excludes longterm overseas attachments of course. Just so long as you are aware of where your roots really lie. Of course some people will always argue that it\'s the local government\'s fault for not being able to retain local talants. No doubt a certain extent of this maybe true. But this brings to mind one thing I read from the newspaper recently. I can\'t remember what the topic was about but this Minister was saying something like Singaporeans have got to grow up. They should not expect too many handouts from the government. Absolutely true. I think we\'ve had too comfy a life here soemtimes that we start to complain once things get a little tough. Sometimes I pity the government for what they have to put up with. I wonder,  rather than taking the easy way out why not make a difference by coming back here. Obviously most people will always say the local environment is not conducive for them to make the kinda of changes, be it in a legislative or a political stand. But if you decide to \'go along\', what else is there to complain? Most just want the best of both world, no restictions to do what they want and an environment where they can make the easy changes they want.  Anyway, I was happily announcing to everyone of my ORD status hahaha... CPL (NS) Tan :P Almost everyone was asking me about what were my plans now. Think I only briefly mentioned Police Force to one uncle and aunt. Dunnoz ... dun wanna say too much until something more concrete appears. Haha will they get a shock if I really do enter the service.  I spent the whole day at home today. Finally digitised all footages of UK and Aikido into my computer. But I didn\'t burn them in CD yet. I\'ve practically used up ALL my harddisk for this. 60Gb of space originally, now I have 3Gb left. But I\'m not too worried lah, after I burn them into CD, I can free up the space.  OK I\'ve decided to call the SPF tomorrow to find out if my application has been processed or rejected. Justin was right, have to bug them like siao to get a reply rather than let them drag my time. At least I\'ll know when I can send a second application in again. But seriously man, hope there\'s some form of good news soon.  Tomorrow evening I\'ve been invited to my RSM\'s home for dinner with some of the regulars and Neo. Neo kept psychoing me to buy a gift and all that. So troublesome. I went to my dad\'s display case to pick an unopened red wine to just give to him. Afterall they don\'t really drink. I got no idea why my dad used to buy so much wine... duh. probably get it wrapped up later tonight. Nothing much on tele tonight either. If I\'m like really bored, I\'ll probably start converting stuff to CDs.  I know X\'mas is almost over but I just wanna wish everyone a Merry X\'mas, season\'s greetings and  Happy New Year.   '),(26176,979795,'23,December,2003','2003-12-23 00:00:00','    HWAAAHHH .... I had the greatest lesson lastnight at BB East. Really forgotten how good Alvin teaches. He was also pretty vocal and his points were so specific. Yesterday he went through a bit of how uke should fall from kotegaeshi. Personally, I \'m never really good with taking slams unless nage is able to guide me down properly (think Syed tried one on me about 2 yrs ago). Maybe that\'s one point Alvin had noted during grading, that once we were into the thick of things, slams and footwork seemed to go all haywire. He went through the motion of how iriminage works and footwork. Seriously man, he just turn turn turn with such ease. Came back to try but I still felt a bit cuckoo just turning 3 times, let alone more.  Alvin also emphasised a lot on on feeling and leading the uke, keeping your hands in your center during iriminage, something I\'ve NEVER been able to do properly and still maintain posture.  Paired up with 2 beginners who had taken their 6th kyu at the Sun grading. After that managed to pair with \'Daniel\' o_O\" ... for nikkyo somemore. A bit stressed sia, didn\'t wanna cause him too much pain so I went ultra slow and I didn\'t really dare to \'experiment\'  too much, hyuk hyuk hyuk. The experience wasn\'t as bad as I might\'ve expected and he was quite grateful too, kept saying thank you. But he was constantly asking about how to do this and that. Told him I didn\'t know that much either so couldn\'t answer him too much and just have to try and \'see\'. After that paired up with a sec sch kid I think? I\'m just guessing coz he seemed too big/tall to be in Pri sch and he had highlighted his hair a wee bit too much to be accepted for a pri sch I think. Then again these days, kids seem to dunno eat what one, they just grow at such a fightening rate. We experimented a bit more on the techniques we did.  While waiting for the bus home, Alvin trotted along. His student, a Japanese lady, was with him. She came down to have a look at our training it seems. Found out that Alvin was NIE trained, went into the govt sector, \'couldn\'t stand it\' (in his own words  ) so he left and entered the pte sector instead. Since 97 he\'s been teaching Eng with a pte school. Most students are foreign nationals here in Singapore, like the Japanese lady, trying to master the Eng language or preparing to take some proficiency test. (Alvin mentioned TOEFL/TOEFEL, forgot the spelling liao since poly. :P)  Oh yar, saw Tommy and Karen there too. Hwah, everyone seems to be running all over for training these days.  Shitz man ... all in all I had such a wonderful time there and the other dojos I seriously dunno how I can train under Sensei, Sunny, Syed, Alvin unless I do it like 4 days a week which is financially impossible. Haiz ... need to be really specific about my choice of dojo liao. But once I enter SPF I got no choice but to train BB East coz there\'s Weds nights off and book out on Sats ... BUT IT WON\"T BE ALVIN TEACHING!!! .... argh   ... And coz of my appointment, I\'ll be missing the end of mth combined training at BVCC tonight too. ARGH...   '),(26177,979795,'21,December,2003','2003-12-21 00:00:00','    Took my grading this afternoon. Felt really good but realised that I started getting a little bit too excited because I began rushing into things once I was uke and got thrown all over. No idea is it because of the adrenaline rush or the disorientation. Sensei\'s comment was \'slightly better than fair\', which seemed to mean average I think. But nevermind, I\'ve noted some points I should improve on. Free practice for one should be smoother and I should try to keep my ukes in front  of me (like ALL THE TIME ... heeeello, duh). I used really limited techniques during free practice so I need to practice using a broader range. My nikkyo and shihonage still need some work. There was once, as uke, I used shomenuchi instead of yokomenuchi. Somehow I kinda heard wrongly. I was a little puzzled how come my partner used techniques which seemed more suited for yokomenuchi instead. I\'m not sure if I managed to change back to the correct strike in between though. My footwork during free practice wasn\'t as sure or precise as it should have been. For one thing my kaitennages didn\'t feel like it used to be during usual classes. Free practice, I forgot most of the kokyunages that I usually did in class. Got no idea why. Even the way of just tenkan and turning to throw was lost on me... duh.  It was a really small crowd compared to any previous gradings I\'ve seen. Which was good, I thought. There were a lot of things I noted. There was this malay guy who maintained a continuity during his free practice. He made sure his ukes were always in view most of the time and he never stopped. My partner had strong and better techniques, I thought. At least I could feel like he was actually doing something to me. I also noted one small point about myself from pairing up with him. I didn\'t maintain as much close contact as I would have hoped in order to execute the technique or taking ukemi. Those would be the times when I felt like I was doing something weird or falling wrongly.  Hmm ... after this, there\'s a lot I would like to work on. For one thing it\'d be intent during atemi. Sensei was saying we\'re too relaxed sometimes. Footwork and posture not just during pair work but especially so during free practice. And taking slams :P All in all, I\'m glad it\'s over. I\'m satisfied with the day\'s training.  I just remembered, should I be submitting my grading card for Sensei\'s signature?   '),(26178,979795,'19,December,2003','2003-12-19 00:00:00','    Heard from  urlLink Eileen  that there\'s gonna be just a total of about 20 students due for grading today. Quite a small no. Dunno whether that\'s a good or a bad thing either. Woke up this morning with less of a fever but a bit of sorethroat. Seriously hope I feel up to grading later.   urlLink Ruth  seems like she\'s considering stopping blogging because she\'s worried about what some people might find out / read from the blog. I was kinda worried about this when I first started too. But I soon realised that it\'s simply a way of self expression. It helps me to self-reflect the mood captured when I\'m blogging. Sometimes I\'d realise I\'ve been a pompous stick-in-the-mud and a real jackass, but I also learned that so long as I behave according to how I\'ve blogged and vice versa, that my actions do not contradict my thoughts captured during my blog and vice versa, there\'s really no issue. Nobody can pop on by and influence me to \'shut down\' my blog. Because if they are not able to accept what\'s been said, they can either email me or post their own comments. After blogging doesn\'t mean everything (my attitude and mindset is final), somewhere along the line I may still change my perception and my stand accordingly if I\'m convinced that I should. Sometimes we need to place past events under the light and re-evaluate them again for the sake of self-reflection and growth lah I suppose.   '),(26179,979795,'19,December,2003','2003-12-19 00:00:00','    Finally got my pink IC. Can officially call myself CPL (NS) Tan now. Spent the morning trying to settle on a draft for my testimonial. Got my RSM to vet through it and got my S3 to sign it. Took me a long time and I felt so tired when I left camp. Must\'ve caught a cold entering / leaving aircon offices and the weather\'s been all hot and cold in the late morning / early afternoon. Was so tired on the MRT and bus journey home I was practically asleep 95% of the time. Wasted leh, there was this really adorable kid standing in front of me in the MRT with his mom, aunt and sis I think. I was kinda awakened by him and he kept staring at me with a bewildered look as I slept. Imagine, me just waking up with a sleepy look on my face with this little boy staring at me all the while and poking my lap now and then. Well, maybe not quiet a poke coz the MRT was really cramped but it sure woke me up. Haha ... really adorable fella. The cute type you can just pinch his cheeks and just go \"So cuuuuuuute.\"  Was suppose to go for the Combined Training this evening. But when I got home I felt so sick that I decided to grab a nap first. Slept right past training hours and even dinner time. I guess it wasn\'t exactly a bad thing either because I really felt refreshed after the sleep and my mind is a lot clearer. I\'d really hate to be too sick to take my grading this Sun. Anyway I hope everyone else there will be enjoying their session. I really really wanted to at least take a look but the weather wasn\'t very forgiving either. There was qutie a heavy downpour then as well and I felt even colder than usual. Better go for water parade within the next few hours.  Neo has been bugging me for the web address of the 302 that had messaged me. Kept on and on about it it\'s starting to be irritating. But of course lah, I know he don\'t really mean anything. Anyway, I\'m just kinda let the matter rest and forget about it. Got a long week ahead and lots to clear up before the year ends. Tomorrow I promised my mum to stay home and do some spring cleaning. Might as well stay home instead since grading\'s on Sun.  Oh had to point out something today. This Robin kept harping on and on about running every weekend at ECP. I said it\'s a great idea but I can\'t for this week. He just had to try and convince me again and again. I got a bit impatient and I had to sms him back telling him to just start learning how to take \'NO\' for an answer. Didn\'t sound rude, but I made sure I expressed my opinion really firmly. Think he took the hint, but still messaged me back saying he\'s going to make it a point to train every week and aim for the full marathon in 2004. Didn\'t reply his final sms but I was thinking, it\'s of course good that he has goals like that. But setting too much too soon won\'t get anyone anywhere (reminds me of me intensifying my Aikido training lasttime too). I personally would love to be able to participate too but I would probably want to be able to complete the half marathon at a better timing and more effortlessly than I do now before even attempting it. I mean, it\'s not like I can complete 21km that fast now either. Anyway, I respect his decision and I don\'t want to have the I-know-it-all-or-better-than-you attitude when I talk to him.   '),(26180,979795,'15,December,2003','2003-12-15 00:00:00','    Tonight\'s Aikido at Siglap was really nice. We did a few katate dori gyaku hamni nikyo which kinda stretched my right wrist a bit. But luckily after practice my wrist still feel fine so Iguess the discomfort was only momentary. There were 2 new white belts as well. They were pretty friendly guys but I\'m not very sure how old they are or what do they do. They were quite chatty amongst the 2 of themselves but still pleasant when training with them. Felt that they kept an open mind during class so they were quite receptive about everything in general.  I\'ll be attending the combined training at Tanglin CC tomorrow with Ueshiba Aikido. Like Sensei had said, \'it\'s been too luxirious for us\' because we\'re so used to having so much space in training that we fall all over the mat space available. Training in a cramped and crowded environment wll help build an awareness during ukemi (and not kill someone by landing on top of them), as well as more control over how and where we\'re falling.   '),(26181,979795,'15,December,2003','2003-12-15 00:00:00','    OK been looking through my blog and all that recently and I realised 1 thing. My English has been going down the drain. It\'s been happening since my Poly days. Not a good sign. So much so that it sometimes affect my speech, most probably in a grammatical sense.  Sigh ... 1 more new year resolution.   '),(26182,979795,'14,December,2003','2003-12-14 00:00:00','    Came back from duty yesterday. Felt really tired coz I kept myself awake waiting from an overseas call from my friend. Turns out she wasn\'t free to call yesterday. Well, that\'s fine, but I knocked out almost immediately at 11+. Took me awhile to hear the phone ring and answer it. Turns out it was my S3 (he\'s my Major to  urlLink you non-military people   ) calling up for something. Anyway, today must be the day I purposely skipped my Jap class just to nap at home (notwithstanding last Sun coz of the marathon) and I\'m aware that I\'ll have to skip coming Sun\'s class for the grading too. That means I have only one lesson left. Doubt I\'ll continue the class coz I want to finish my ePrep online courses for the next 3 mths before I do anything. Think it\'s a good time for me to attempt to fix in some volunteer work at the SPCA as well. Will see how things goes.  About coming grading. I don\'t feel prepared. I probably know the syllabus, but I think I need some more refresher on them this week.  And this Thurs... ORD loh! Haha... Got to remember to go get some food to treat the guys when I go back.  Just got a message on Friendster from someone in camp who\'s in a different unit from me. He\'s gay and I\'m a bit uncomfortable of being told that I\'m \'cute\'. But I\'ve seen the guy before in camp and he seems friendly+harmless enough. How did I deduce his \'gayness\'? His profile lor, he says he is \'Gay + Dull + Stubborn + Persistent + Sensitive + Introverted + Friendly + Loving\' and in the \'Who I Want to Meet\' section was, \'Non homophobic people\'. So I messaged back saying I don\'t want to sound presumptous but whether he was one. He said yea and he\'s kinda trying to hit on me. Don\'t mean to make this sound like a joke but I think he\'s just being friendly. Hmmm ... my first \'gay\' contact online. Think the first time I actually came across gay was my first day of NS in Nee Soon. Do  urlLink you  remember? The 2 storemen issuing out equipment were all 302s (military classification for gays). Anyway, told him I\'m probably ok talking to one but I\'m totally straight. Hope that\'s the end of it.   '),(26183,979795,'12,December,2003','2003-12-12 00:00:00','    Been overly \'hecticated\' (ar ha! learnt this new term from my fren in UK.) these 2 days so wasn\'t able to blog. Well yesterday was a bad and good day for me.  Was supposed to meet Robin (sec sch friend) for lunch then maybe go around Orchard for a bit before we went down to Siglap. He said he wanted to take a look at Aikido. Was all ready to leave when my dad popped over with a problem. Something was wrong with his PC. I got no idea what he did or what happened (but I dare say it\'s almost always the user\'s fault 95% of the time), but the PC somehow kept showing fuzziness and discoloring on the monitor. It just refused to go away. Was actually thinking of leaving since I already made an appointment, but then realised that I won\'t be back until after Aikido (which means I can only work on it about 11+ after a shower). Haiz, decided to try to fix it then. To my horror, the damn thing just got worse when I restarted it. It just kept on rebooting and never got into windows. Couldn\'t even enter safe mode. Was seriously pissed coz while working on it my dad was in the background mubming and grumbling away, about why I never set his emails on my side so when his PC is down he can fallback and use my PC instead. Goodness lor! Gimme a break lah. His emails are downloaded from server into his hdd lor, even with an internet connection on my side he also can\'t view his old mails what. Fish lah! Tried very hard to solve the stupid problem. Manage to find out that the hdd is not faulty (whew...) so I transfered his emails to my PC so he can at least access his mails while I\'m away. Throught the 4-5 hrs I was working on this, I let Robin know about what I\'m doin so he can at least carry on with whatever he wants to first. He said since he was already in Orchard, he\'ll hang around in the library and hopefully we can have dinner later. This was a bit preassurising lah, coz I wasn\'t sure how long I\'m gonna take and I just had to rush to do what I can with the PC. Anyway I short circuited my external HDD casing in the process. Was seriously pissed with myself too. Used to bea ble to use this rack and shift it between different systems via a USB connection. Now I need to open up the whole PC and plug it in instead. ARGH!!!!  Anyway, I was almost done when  urlLink you  smsed me about Sensei going down to BVCC instead. By this time it was raining quite badly for awhile already. Anyway, my dad did me a favour by dropping me part of the way where I avoided the rain and to grab a bus down. Originally arranged to meet Robin at Heeren, but didn\'t know exactly where we were meeting. I waited at the bus stop for awhile before walking over to Taka. Called him continously and smsed him for 30-40 mins and didn\'t get a reply (Robin, if you\'re reading this I have to tell you man, this is something typical of you. Better change ar!) I called until his hp actually went flat. At this stage I was seriously exasperated because I had to way fo reaching him and I kinda lost the mood to even go for Aikido. Gave him a final sms saying that I\'m going home. On the way to the busstop, I got a call from him. We arranged again to meet at Orchard MRT control station immediately. Apparently, this bugger kept his handphone on silent mode!! ^%$#%!!! Was seriously pekchek man but I figured things weren\'t totally his fault coz I kinda made him wait all afternoon alone anyway. Haiz ... For both our benefit, I decided to goto BVCC instead of Siglap for the evening to train. Wasn\'t in quite the mood to travel long distance to Tanah Merah at that point.  We grabbed a bus to Holland V and had dinner at BK. After that we went over to BVCC. Aikido was great lor! It was the first time I was in Sunny\'s class. Very good. He was very vocal and kept coming up with a lot of points for us to note. Like  urlLink you  said his movements were really something. So relaxed, smooth and natural. I felt like a cuckoo when I was practicing lor. He kept emphasising being serious to my group (Edwin and another guy) coz I think they were joking a little bit too much/loud enough to get noticed. Heh ... well it was still enjoyable. I just didn\'t manage to accomplish tho. Ended the evening with kokyuho with  urlLink Ms Ng  there, ar I still dun get it lor. Sometimes can, most often can\'t. Seems like every 6th kyu-er and above I came across are able to do it leh. Haiz ...  Anyway like  urlLink you  noted, Robin was sitting outside even tho he was actually here to \'take a look\'. I know him long enough to just leave him to do what he wants even if I think some of his actions might be a bit weird. He has his reasons for his actions I suppose, I\'m just not able to agree with his rationale sometimes. Anyway, I used to be very judgemental of him in sec sch. In fact a group of us kinda gave him a hard time then coz he was quite \'different\'. I\'ve kinda grown out of it liao, think it just comes down to respecting what he wants to do. Anyway, most of his actions dun quite affect me so that\'s fine by me.  Well that\'s yesterday, but today I spent the whole day testing the PC and fixing it. Managed to find out that the graphics card was faulty. Seems that the motherboard was a bit dodgy also. Anyway got the thing to work and had to reinstall windows twice, plus all relevant software. Than spent hours customising small itsy bitsy things to prevent the \'ignoramus\'es in my home from meddling too much with the system and making more problems for me later. All the while my dad was again popping in and out asking me if it\'s ok already, what\'s the problem with it, why it like dat, etc etc. ARGHHH!!.... Actually had the intention to go Sim Lim get some parts, exchange some faulty CD-Rs, get some clothes from Seiyu and asking Robin to join me. But things just dragged on so long (was in front of the PC since 8am till 9pm straight) that I kinda dropped the idea. I felt too tired anyway and I have to go back camp first thing tomorrow morning for overnight duty. Well at least, I\'ll have 1 whole day to sleep and rest there without being bugged by PC problems. Think I\'ll go Sim Lim/Bugis on Mon instead. Probably ask Robin if he\'s interested then.  Sigh, the past 48hrs would\'ve been sheer torture and demoralising if not for those few hrs of Aikido at BVCC...   '),(26184,979795,'11,December,2003','2003-12-11 00:00:00','    Today has been a nice day again.  Been home all day until the late afternoon when I had to go for my medical appointment at the Eye Ctr again. Dr said my eye was healing and I can reduce the medication to 2x a day now for the next 2 wks. My next appointment will be in Feb 04. Heh think he\'ll be discussing my options of wearing contacts and some kinda of counselling on how to maintain my contacts. Haiz ... troublesome but no choice. Forgot to remind Sensei he should change his contacts more often instead of wearing it for over the stipulated period, otherwise liek me like that, damn troublesome. Dad was nagging me to lay off contacts for 1 yr. See how lah, will try to not wear them for as long as I can, hopefully more than a year.  Following  urlLink your  post, I didn\'t think that Daniel was ultra weird or anything when I saw him today. I suspect he might\'ve mellowed down a bit since last week. One good thing tho, despite having experienced \'unpleasantness\' he\'s still willing to come back and continue training. Of course lah, it could be because he feels that he had already paid for 3 mths liao, or he came back expecting some kinda public apology ......... Anyway, today\'s Aikido at BB East was quite fun. Dunno why either, just found it really nice.  Just checked my email and found that the Half Marathon results have been  urlLink posted  online already. I should receive my cert by the 15th. The site is quite cool coz besides posting the results online, it also has a link which does some kinda analysis of the runners\' performance.  Summary: I finished 498th out of 1409 participants. Finishing time is 2hr 21min 16sec. Net running time is 2hr 18min 54sec. (Like I said in my earlier post, it took me about 2min+ to get past the starting pt.) Placing in the entire field was 612th with 1270 entrants behind. About 32% of entrants ahead. Placing among the Males was 497th with 911 entrants behind. About 35% of entrants ahead. Placing in category was 368th place with 662 entrants behind. About 36% of entrants ahead.  Split Times (whatever this means??): 10min 36sec per mile or 6min 35sec per km Speed: 5.7 miles per hr or 9.1km per km  I must say I\'m feeling really proud of myself at how much I\'ve improved sicne the AHM. At least now I\'ve got a benchmark to train myself. Yesterday, just asked my friend to help me get a form for the NUS biathlon in Feb 2004. Not sure if I\'ll signup for it yet though.   '),(26185,979795,'08,December,2003','2003-12-08 00:00:00','    Right knee was hurting from the run yesterday. Read somewhere it\'s normal to feel sore and fatigued after a run. Not sure if I\'ll have any problem with Aikido tomorrow night either, but I think I\'ll try to train. Afterall, I\'ll be in the West side fixing my friend\'s PC in the afternoon anyway.   '),(26186,979795,'07,December,2003','2003-12-07 00:00:00','    Alrighty! Today was a good day I think. I remember waking up earlier around 2+ and it took me awhile before I was able to get back to sleep. Nervous and excited I guess. :P  Got up at 0415hrs, washed up and had a quick breakfast before I went out to meet the cab. It was already waiting when I left the house. Went down to Neo\'s (campmate) place and waited awhile before we went to the National Stadium. A lot of road closures so we alighted along Nicoll Highway instead. There was already a crowd by the time we arrived. I called Robin (sec sch mate) and we met up at the bags deposit area which was actually just a small lorry labelled A and B. Things were pretty organised and systematic because lorry A only took bags for tag IDs 4000-5200. Mine was 4440, not very auspicious lah :P. Despite the crowd, it wasn\'t as bad as the Sheares Bridge Run/Army Half Marathon. Today\'s crowd was just around 10000, but I heard that this event attracts a lot of foreign and professional marathon runners each year.  We started the run at 0600hrs. There so many people packed together at the starting line that we had to walk for 2min+ before we actually reached it. The 10, 21 and 42km runners all started at the same time, but the end point will be different. 21km will end at ECP, while the others will end at Padang (if I\'m not mistaken about 10km). The first 5-6 km was nice and comfortable, but it got kinda boring the moment we hit ECP at the Fort Road junction. I had to start taking drinks at almost every drink booth fro this stage onwards to sustain myself. The first 10km was quite alright and light. All the while, we were all running at a small pace and slow speed to conserve strength and energy. Along the way there were cheering teams and drink stands which really helped the runners to push forward. I think the toughest part for me would be the last 3-4km. By then my legs were feeling fatigued, my pace was smaller and speed was slower too. Kinda entertained the thought of ending it early by walking but I kept telling myself \"There\'sonly slightly more than our usual 2.4km run left. I can do this!!\" haha ... I think Standard Chartered will be mailing my timing to me at a later date. There\'s this electronic chip on loan to runners that we had to attached to our shoes. It\'s suppose to monitor the start and end time to provide us an accurate record of our running time. Apparently, some metal plates were placed on the ground at the start and end points to act as a trigger for these chips.  Well I did finish the run in the end, my timing seemed to be around 2hr 20min or so. This was considerably better than the 2hr 30+ min I did for AHM. But I expect there\'s some discrepancies because AHM used a clock to record the timings. Somewhere during the run, I lost sight of Robin and I think he went on ahead. Hope ghe got a good timing for himself. Neo was just under a minute behind me. I was surprised as he was usually a better runner than me. After the run I must have gulped down about 4 cans of Gatorade and 1 bottle of mineral water at the end point. Didn\'t manage to meet up with Robin as he had to go to church soon so Neo and I took a free shuttle service back to the Padang. By then the first runner had already finished the Full Marathin. It was an incredible timing of 1hr+, just under 2hrs. Oh man, by the time I finished 21km, a bunch of others had already finished 42km. Wooooo...   From here Neo met 2 of his old friends and the 4 of us made our way to the 42km finishing pt to see how everyone was doing. It was quite encouraging really to see so many people cheering. We went to grab some food at BK and changed into clean clothes. The weather today had been fantastic. Not sunny at all and neither did it rain. I think this was a very important factor during the run.  To conclude this long entry, I hope to go back for AHM 04, but I\'m not sure if I\'ll go for the Singapore Marathon 04. Even if I do, doubt I\'ll sign register for it as I still think $30 is still pretty steep, considering it\'s an early bird price. 42km? Hmm, won\'t entertain the thought for a few years at least. Maybe only if I can shave down my timing to under 2hrs? Hahaha ...  How do I feel now? The mind is fresh and awake but the body is really sore. Because I\'m wet through down to my shoes and toes, it kinda made my nails brittle and the skin wrinkled. Luckily the nails didn\'t break like it did during AHM. Some light abrasion along the ... ahem ahem ... common areas and the sole of my left foot feels bruised though it seems to be getting better just by removing the shoe. Threw away my old series SAF New Balance running shoes, which in my opinion, is a lot lighter than the new one.   '),(26187,979795,'06,December,2003','2003-12-06 00:00:00','    OK today was quite nice, met up with an old sec sch classmate. Went to collect my race entry pack at Pan Pacific Hotel for tomorrow morning\'s marathon, and had dinner with my friend. Apparently he\'s also interested to participate but had not been able to register before the closing date. Anyway we\'ve arranged to meet at the luggage/bags deposits area instead.  Went straight home after that to rest and book a cab at 4:45am. Still have to pick up my friend at his home before we go down to National Stadium (start pt). At least I\'ll have co. for tomorrow\'s run.  OK. 8:30pm liao, time to sleep. Looking forward to an interesting morning...   '),(26188,979795,'04,December,2003','2003-12-04 00:00:00','    Was rushing to the dojo. Prior to that had to meet my friend at Tanah Merah MRT to return his jap textbook. Had loaned it last Sun to photocopy (cheap cheap :P) and die die needed to return coz he need to do the homewk. Anyway, that kinda made me late for practice.  Heh,  urlLink you  had to give us your wonderful warmup again :P But can tell lah it\'s not as tiring as the other time. But I still dun understand how on earth you can hold your leg up so long. I can feel my whole body quivering and straining in jus a few seconds.  Hmm ... Sensei kinda lost a bit of patience with one of the kids in class. But seriously lah, he\'s quite a joker. Maybe coz of his age that\'s why he still feel like having fun rather than train seriously. Oh but something Sensei mentioned today that stuck. He likened aikido with horce racing. Sometimes horses would run really fast but lose stamina later, some horses which look like they can\'t even run but start to pick up later on in the race. The latter being the good horse of course. But it\'s the same in Aikido too. I was recalling the time when I had really wanted to train 4-5 days before I returned to UK. But all along in my mind I was expecting to slacken down my training to just 2 days. Of course lah, commitments in NS can\'t be helped either but still ... (Alamak! I\'m a lousy horse!)  Oh btw, I\'ve tried and tested liao. it doesn\'t matter if you put \'www\' in front of the address of your blogpage. I was wrong. Publishing didn\'t seem any faster than the other time. So pple, dun bother updating my add/link. I\'ll change back later.   '),(26189,979795,'03,December,2003','2003-12-03 00:00:00','    Aikido lastnight was quite interesting. A lof of feeling techniques. We did katate dori ai hamni ikkyo, iriminage and shihonage. Shihonage went on to some others which I can\'t recall the name or be sure of. Was quite tiring but fun. Dis some kotegaeshi as well.   Aching like siao this morning, still had to make myself go for my run. It\'s prob the last few things I can do to prep myself for the marathon this sun. Goodness, I dunno how I\'m going to finish the distance. Been out of training for too long already. I\'ve only just gotten back to getting used to a slow but more regular pace of running. But I\'m not confident of getting a better timing this time round. Which reminds me, I think I better visit the website again to familarise myself with the route and procedures for the run. Still need to plan breakfast and time to meet friends on that day itself. Sigh, more liasing.  And I really have to say this, I\'m STILL waiting for some kind of notice from the police force.   '),(26190,979795,'01,December,2003','2003-12-01 00:00:00','    Something I just found out about Blogger. It seems like it\'s been taking really slow to update and I might have found a reason why, Blogger seems to be updating straight to www.your-blog-ID.blogspot.com first.  What I did was update my links on my blogpage, my links on my homepage & my commenting facility (enetation). Hope this helps everyone. Oh yea, and remember to update your link to this page too :P I suspect this the reason why the no. of comments displayed don\'t seem to registering properly either. Updating enetation on your blog address might solve it. I\'m still testing it now.   '),(26191,979795,'01,December,2003','2003-12-01 00:00:00','    OK 4th appointment at SNEC shows I\'m on the road to recovey. Next appointment will be on 10th Dec. Oh, and they\'ve extended my MC till then as well. Goodness, I\'ve been on MC since 22nd Nov. Everything still looks blurry and unfocused for my right eye though. Hope it\'ll recover with time instead of becoming too permanent (?!).  Think I\'ll still go back for duty on 13th Dec. Friends have been covering for me last mth, feeling a bit guilty. Anyway it\'s not guard duty so should be quite light. Doubt I\'ll need to \'see\' anything that clearly. But think parents are gonna make noise again if I go back if I\'m covered by MC then.  Prob resume Aikido this Tues onwards. Time to put in more effort in my Japanese too. Realised missing one class last week made me really rusty. Had a lot of trouble remembering the vocab everyone was using. Oh well, gambatte!  Oh yar, my friend\'s church booked a whole cinema for LOTR on the 21st. Actually wanted to get the old housemates together for this. But coz of some having other commitments and the grading date/time kinda clashed, so I had to rearrange. Probably just meet up for a meal or something instead.  I\'m still waiting ernestly for some kind of response from the Police. They\'re taking so damn long. This thing is just nagging on my mind now.  Oh yea, I registered some courses through the ePrep programme for NSFs. Quite interesting lor. NSFs are given 300 credits => $300 to apply for some of those eLearning courses. We juz need to pay abt 10%. Signed up for a series of continuous Java programming courses and one on Information Privacy. Can\'t wait for the materials to be sent over. Prob stop Japanese after Dec and do some self revision while I continue with this ePrep courses for the next few mths. Best thing is that all these are OTOT (own time own target). I just need to schedule my free time at home regularly to make sure I get some work done.  Oh well, time for new year resolutions :P ... Looking forward to 2004 where I\'m prob gonna embark on a whole new phase of my life.   '),(26192,979795,'26,January,2004','2004-01-26 00:00:00','    &^$&%UFR%&$&^%!!!! My spare harddisk seems to be slightly corrupted. I was trying to copy some Aikido pictures of which I had scanned from my Sensei in UK. But 1 directory seems to be corrupted and I can no longer retrieve the pictures and originals. ARGHHGH!!! Worse! One of these photos actually includes a photo of my Sensei!!! ARGH!@#!@#!@# ..... I\'m so pissed.  Think I better start backing up my data onto CD. Especially those GBs of Aikido clips I so painstakingly downloaded and catalogued. If even those are gone, I think someone had better just take a bokken and chop my neck off. Don\'t matter that it\'s blunt, all the more painful it\'ll be then.   '),(26193,979795,'24,January,2004','2004-01-24 00:00:00','    Right, just got back from Malaysia. Visited some relatives there (mum\'s cousin ... I think :P). Not gonna be able to attend Alvin\'s classes tonight. I think I can officially call my angbao collection complete by now. I\'m pretty sure I\'ve visited everyone I was supposed to and everyone that needed to come to my place had already been here too. Well I have to go pay my mat fees with my angbao money this week, plus use the remainder for a new pair of specs. Figured these things shouldn\'t wait too long anyway. Might as well settle it since I\'ve got the moolah now and it sure beats withdrawing large amounts from the ATM later.  Sat I was supposed to visit Daniel\'s for lunch there. Prior to that, I was at a relative\'s when I got a call from David saying the visit was cancelled. Apparently Daniel\'s dad had a fall that morning and Daniel was at the hospital with him. Luckily that same afternoon, I got news that he was fine and we could still visit him if we were all able to. But I didn\'t think it was wise to be disturbing his family at that time.  There\'s been quite a number of accidents around my area just the past week alone. I know coz I\'ve been hearing a lot of \'screeeeeech CRASH .... BANG!\' and they\'ve pretty much all sound the same. Just a few hours ago a car flipped on the road just opposite my place. Naturally there were police, ambulances and reporters. I think it\'s partly due to the wet weather which pretty much started on the day of CNY itself. Weird coincidence. Unfortunate too that all these area happening on a festive season.  This CNY I noted that I needed to know my limit when it comes to eating new year goodies. I feel a bit sick because I\'ve been having too much pineapple tarts. I think it\'s because of all the butter that\'s giving me a very bloated feeling now. Wouldn\'t go for Aikido even if time permitted. Gosh, seriously man ... I feel like puking. The pineapple tarts were all home made by my aunt that\'s why they were especially good. Not too sweet like those you may find outside sometimes. Perfect I would say. I don\'t have much appetite to have my dinner later either, so bloated now.    OK I\'d better go nap now. Hopefully my poor appetite will be better when I wake up.   '),(26194,979795,'24,January,2004','2004-01-24 00:00:00','    Did a small change to my blogpage template. Added some Hopes and Wishes columns. Basically covers some long and short term goals I\'d like to try and keep in mind.  On the 20th, I met up with Begh and Lynette for dinner at my friend\'s restaurant called Swiss Culture,  Suntec City mall. We had an ultra filling meal there and I think the food there has improved. I have to admit he is doing pretty well. Anyway we had a really good and long chat there as well. In fact, I only just received his housewarming invitation by SMS at his place next Sun. Hopefully my NS kakis will be attending too.  On the eve was reunion dinner with my mom\'s side of the family. After that, us guys decided to catch the Last Samurai (starring \'Tomu Kurusu\' - Tom Cruise ;P). 2 of the ladies went for KTV and another went to catch a midnight with her boyfriend. All our parents decided to go home for the night. I haev to admit, the Japanese actor acting opposite of Tom Cruise did outshine the latter in the show. There was just so much more depth to his character. The action sequences were pretty fantastic too but you can kinda tell Tom Cruise\'s swordwork, was way different from the Japanese\'s in the show. Not enough sword fighting scenes as I would\'ve liked as well, but that\'s just me lah. Thinking of buying the DVD when/if it\'s released too.  It\'s the year of the monkey now. These 2 days were quite uneventful save for visiting relatives and friends and vice versa. Got another interview next Tues. Something similar to an interview I went a few days ago. It\'s to be attached to schools to teach IT. Mr Tan ... haha ... can\'t imagine students addressing me that. Sounds weird ... dunnoz.   urlLink Gary  finally received the CDs I sent, though he did mention one of the discs seemed a bit dodgy. This week I should be able to attend Aikido. Hope to pop by on Mon. It\'s been a long time since I managed to put in any serious time into it and I really want to get back on track again. Back hasn\'t given me any problem past week, but I don\'t think I\'m going to push myself if it still feels weird during training.  One not too happy thing I just remembered. Dad thought he was helping by \'asking\' his cousin and brother if there\'s any job vacancy they might know of. I\'m glad there wasn\'t exactly a response or a positive one. Seriously felt like I was being shoved into an area I don\'t quite want to be in. Didn\'t want to put him down for his request in front of others (nor did I when everyone left). -- See \'Hopes and Wishes\'. -- And I\'m so getting tired of him comparing me with my cousin who just finished his course from NUS too. Same as Gary, just that he took his honours year as well and he\'s got a good job paying him 2.3k. Dad keeps poking me about why he is good lah, why I\'m not lah and all that. Geez ... I\'m getting fed up just blogging this down now. But I also recalled soemthing my aunt said to me lastweek. \"Afterall lah, he\'s your father. He want to say let him say lor. It\'s not like it\'s some outsider / boss outside telling you off now. It\'s all within the family.\" Ar, she\'s right. So far keeping this sentence in mind helps me a bit but I dunno how long it\'ll last lah. Ultimately I don\'t see myself as someone who is willing to back down if I feel that a gauntlet is being thrown in my face.   '),(26195,979795,'17,January,2004','2004-01-17 00:00:00','    Didn\'t exactly have a great night yesterday. Day was fine coz I had the whole day pretty much to myself. Dinner time came and my parents were like arguing over who gets to make dinner. Duuuh ... me? I\'m just contented to get some food at the end of it. Each was like criticising the other on the \'right\' way to cook this or that. Goodness lor, who cares so long as the food at the end is decent enough and edible. Plus it\'s just fried rice!!! DUH!!!!  Again, arranged to meet Begh this evening. Once again, this stupid incident (over the silliest issue) made me delay dinner and had to cancel. @*^#$#...  Anyway, mood died down pretty fast while we were all eating. Think it\'s because of the Quan2 Xin1 Zhong3 Dong4 Yuan2 3 (Snap 3) on tv. The night was OK after that.  Sensei sms saying BB East had no training on Sat coz the hall was being used. I emailed Sensei about my back and videoing for the session when I pop down. Hopefully get a reply sometime this weekend. Otherwise may have to sms him liao by Mon.   urlLink Eileen  finally popped back onto the blog to \'show face\'. Haha. Thought she MIA liao. And Gary sounds like he\'s enjoying his training too much to want to leave US. I dunno, I got the same feeling in UK as well. Just gets me thinking what was the difference I felt about here and UK exactly? Hmmmm .... class size, environment are the obvious ones. But what about with relation to training? I got no answer yet.  Realised I\'ve downloaded too many articles of the past and not done enough reading for all of them. Was reviewing some when I came across something called \'Verbal Aikido\'. Damn interesting article about handling verbal attacks. Just somehow got me thinking about how I can apply them when my parents make noise at me (or even vice versa). Dunnoz, maybe I\'ll post it after I finish it in detail and reflect a bit more over it.  Also realised my scope of job applications is slowly expanding from what I intended. Initially was non-IT/ programming intensive. Now it\'s starting to include things like VB, SQL etc. But then again I hope there\'s more willing to take in multimedia applications and stuff, especially those with \'willingness to explore and experiment\' kind of criteria.  OK time to get back to my ePrep programme again. Time to get a job ...   '),(26196,979795,'15,January,2004','2004-01-15 00:00:00','    Missed training ... again. I must have missed up to about 3 weeks worth of training I think. I dunno if it\'s a damn coincidence with illnesses, injuries, clash of timings (duh ... and I\'m not even working permanently yet), plus a whole lot of other shitz, or is it that I\'m slacking and feeling just plain lazy. Sure hope it\'s not the latter though.  Anyway, today I helped carry some buckets of water for some washing. Though my back was OK liao, but apparently it wasn\'t \'OK\' enough. I\'ve just decided to stop training till after the Chinese New Year. Best to ensure a complete recovery to avoid complications. Think that\'s the best alternative. But I don\'t really want to miss out Mon\'s lesson, it being the last one before CNY. I\'ll probably email Alvin this weekend see if I can pop down to watch and video his class down. Just for the fun of it.   '),(26197,979795,'14,January,2004','2004-01-14 00:00:00','    Haiz ... I\'ve only just ORD for less than a month, and I\'ve got notice to return to camp on the 31st for Mob Briefing liaoz. Oh man ...  TV repair man came today. Woohoo ... New sofa came today as well. I was quite irritated with the delivery people. Original arrangement was that they will remove our old set for $20 when the new set comes. I told them it was arranged in advance and even written on the order slip in black and white. The delivery people were still insisting that they can\'t do it. What peeved me off a little was their response came pretty much when they heard that they were going to be paid $20 (only). I got the impression they\'ll probably do it for more money instead (I\'ve always thought such things usually got done at a rate of $50). Finally, easiest response was we didn\'t want to accept the new sofa and would like to cancel the order. Now they had to call their office and ask. There was a lot of arguing and finger pointing, though not in a really nasty way. More like them saying aloud \"Customer won\'t listen to me, you go and talk to him yourself.\" etc over the phone. I was getting pretty irritated as well. In the end we said, we\'ll accept only whenever they are ready to take over our old sofa. Way I see it, several thing made me pretty mad with them. One, they delayed delivery beyond the original delivery date. Two, the taking over of the old sofa was arranged when my mom placed the order and it was written in black and white. Three, the way the delivery people reacted when they heard of the $20 was seriously negative like a \"Impossible!\" or a \"No way!\". Four, when we talked to the manager of the store, he actually had the audacity to say something like he has no control over the delivery people. If he has no control then who does? Oh please. We\'re not not unreasonable people here either. We can see their truck is too full to take our old sofa set. Their suggestion was for us to leave the old set in the carpark first and they\'ll be back tomorrow. But we have no guarantee they will return to collect the old one after we accept the new set now. My dad\'s back isn\'t good and there\'s no way I can shift it out alone either. Plus our estate management do not allow such things to be done. And there\'s absolutely no space to leave it outside the house as there are more than one household in the estate who own more than one car. It was simply ridiculous. If they\'d only just come back later when they\'re truck has space, I don\'t see why we can\'t work it out. But the driver just kept saying it\'s too late lah, cannot lah, etc etc. Idiot! Ironic thing was that, after one hour they called back saying they now have space to takeover our sofa and handover our new set. DUUUH!!! Anyway, I\'m just glad that episode is over. I think I\'ve already done my part. Despite paying them $20, I still helped them carry the sofa out as well. Maybe I should\'ve kept $10 for myself instead. God knows, I needed the cash anyway... My father was quite insistent that we let them carry all the sofas but that would\'ve been impossible, mean and physically impossible anyway.  Begh (UK housemate) SMSed me about grabbing some coffee tonight. Was supposed to go, but the whole episode above delayed my dinner and I didn\'t eat till like 8-ish. Had to cancel the appointment.  Anyway spent most of today finishing up my book \'Black Hawk Down\' and going through my Java ePrep modules online. I think I can finish the unit I\'m on by this weekend and quite possibly start on the second unit. In total I\'ve 3 units to complete and another course on Computer Security. Hohum... too much time to spare at home anyway. Just started reading Lord of the Rings and realised it\'s not exactly an easy read. The book I got had small prints and that didn\'t really help. I\'ll probably take a long time to finish this book. Worse, I promised a friend to finish another book she had recommended. I just might KIV LoTR first.  When lifting the heavy sofa today, I was careful not to make my back any worse than it seemed to feel. Fact was, it seemed to be recovering fine this past week. If nothing goes wrong tomorrow, I\'m probably going down to Siglap for training and to pay my mat fees as well. Owed it too long already. Talk about owing, I also need Sensei to sign my grading book and pass him the grading CDs ;P. Lucky I\'ve packed all this in my bag already. Hope Gary\'s degenerative brain syndrome won\'t affect me haha.   '),(26198,979795,'11,January,2004','2004-01-11 00:00:00','    I was reviewing footages of previous grading clips this morning. I dunno why but I just suddenly recalled a statement I made on ICQ to  urlLink Eileen  sometime ago. I mentioned \"Is it me or do the standard of people taking the grading seem to be lowering these days?\" Well first of all I see myself having been stagnant and even backsliding a bit to a certain extent for a period. Second, certain gradings just don\'t seem as intense as the one(s) I had seen before. However, I also realised that every batch of students will be different. Some improve fast and some slow, while some others don\'t. The \'standard\' of improvement is also on an individual level too. Let\'s say for example some dude did just OK during his 6th kyu test, but suddenly he did better than normal during his 5th kyu test. Such improvement(s) is relative and varied in different ways. It may be philosophical, technical, even personal (attitude, habits, etc). What we see during a grading is not the whole picture. To be specific, I\'ve always felt that Soh was a pretty stiff uke whenever I paired up with him few years back. But he has seriously improved a lot since then and that difference really impresses me (though I\'d still prefer another uke instead. Guess I\'m still standing on the same point instead of moving on and \'improving\' this attitude :P ). Something along the \'some horses run slow first then pickup speed, while some run fast at first and get\'s punctured later on\' theory. Also, I\'m trying to remember that such things can\'t be \'compared\' with others. I mean what\'s the point right? So long as there\'s personal improvement, it really doesn\'t matter how much. Just as long as you\'re happy with how much you\'ve grown. If you\'re not, well that\'s where you work towards it I suppose.  What (I think) I really want to say is that, despite being bad at what you do now, don\'t mean you can\'t be good / improve later. Maybe the problem will be how much we really want to improve and actually committing ourselves into it to do it. Whether achieving it or not, well I guess there\'ll still be no regrets with the results. But what about in cases where you can\'t afford to fail? Hmm, gimme some time to dwell on that. Haven\'t thought out a moderated thought on this yet.  All this thinking of my career kinda made me visit the SPF website once again. Just noted that as of this year, salary stated on site will be reduced by 15%. Well, still not a big deal to me considering I didn\'t want this because of the money. But I\'m considering applying to CNB too. Actually given a choice, I would\'ve wanted to do uniform service in the early years (4-6), try out 1-2 years of UN peacekeeping, and then maybe opt for a transfer to somewhere eg. CNB, CID, etc, depending on what I\'ve learned in the force. Haiz, just hope I can be shortlisted someday. Pity I can\'t just waltzed in and sign on.   '),(26199,979795,'10,January,2004','2004-01-10 00:00:00','    Well, I\'ve got quite a lot to write about. Dunno where to start.  For starters, I missed Aikido on Thurs coz of work. Fri morning, I went for the interview at a small french-german company consisting of just 3 persons. The Dir was not available then because he was at the Dr\'s, I think. But the other two had given me the impression that the company held a lot of promise and there was a lot of space and possibilities to grow. The company provides several software packages in the form of inventory, revenue monitoring, reservation sys to the hospitality industry and more in the region. They are already providing support to several big hotels here and I was so impressed with what they had accomplished with just 3 person. It doesn\'t surprise me that they actually need to hire another person to help out either. The job sounds really challenging to me though somewhat less technical which was exactly how I had pictured it. I\'m guessing that once I go into this, there\'ll be no room for me to jump back out into the IT/programming aspect of IT anymore. For starters there\'ll be a lot of answering of calls to answer users\' queries and help solve their problems. But as time passes, they hope that more responsibility can be assigned to assist certain areas of their work for on-site installations, trainings, upgrades, maintenance, etc. This is why there may be some travelling involved overseas. What really struck me was the great working atmosphere the team had. The Dir is transparent with the revenue the co. is generating so everyone knows where the money is coming in from. The only down point was that they were only able to offer 1.6k for the first 3 mths during training, and then 1.8k. This is seriously low to what I\'m hoping to get. I was really frank when I left and I said I needed to think things over because of the pay. They were really generous, cordial and open from beginning to end.  I asked a lot of opinions from my parents and friends. My parents agreed the the pay was a bit low and that the job entails some OT usually, being a customer service oriented job. I\'m not surprised even if there may be a need to remain contactable during weekends and public holidays despite it not being the official office hour. Neo especially made me re-evaluate what was it I really wanted from a job/career. (YO! You happy now boh? Got mention your name leh! :P) I realised a lot of things about myself that I had tried to avoid asking myself for sometime too. Conclusions I had drawn, I wanted to be sure that I\'m not just taking up any job to \'escape\' the programming area of IT, that what I really wanted was job satisfaction instead of just monetary rewards. That I\'m willing to take the job up to another level and push myself beyond what others may not want to do. All these are what I expect of myself when I\'ve decided to place my foot down on the direction my career will be taking and I don\'t see any other way beyond this. I\'m still willing to keep my options open at this point because I still feel really keen about the SPF. But everything I\'ve heard about the job has seriously tempted me to try out something else on a long term basis. In fact, I\'ve only just emailed them saying I\'m willing to take up the job if I do get shortlisted. Let\'s hope everything works out.  Went back to work in the afternoon and didn\'t manage to talk to my boss about my decision to stop work till the end of the day. He was nice about it and he actually took up a chair to have a good long talk with me. I was aware he had delayed/cancelled a delivery/meeting that evening just to talk to me and we spoke till about 2 hrs after closing time. He\'s a honest, traditional man who is open to many things. After hearing things out from him I realise how much he had gone through. I even understood why he had requested that I be willing to try things out in the co for a week before he decided to pay me. No doubt I may have helped him out for free the past 5 days but I really hadn\'t done too much there either. Sounds a bit foolish but I do pity him more than I pity myself. I also respect what he has put himself through to have what he has now, and what he is going through now just to keep the company going too, but there\'s all the suffering and hiccups from yester-years that I didn\'t see which has made him a stronger person. He was actually quite keen to help me grow both as a person and in the company and it\'s a pity I wasn\'t the right person for the job.  The long talk with him made me almost all of us just want a comfortable job that we can work in because of the stability it provides - entering a big company where things are already stable and there\'s a system in place. But if you\'re in a small co. and you\'re able to make something out of it or turn things around, I think that\'s where we can impact the greatest difference and climb our way up too. I was kinda relating these thoughts - from the interview, the long chat session - to my asipirations to join the SPF. I realised that it meant more to me that I make a difference in Singapore in whatever small way I can than to get a great paying job. But on the downside, I also realised that I really wanted that comfortable level and stability much more. I wished what I felt was evened out on a 50-50 basis so that I can have the best attitude for things, especially so since it will affect my decision on selecting a job/career now. I dunnoz ... I just hope I\'ve made the right decisions and reevaluations of my job-selection criteria and I\'ll be really happy about the decisions I\'d have made when I re-read my blog 10 years down the road.  Almost failed to mention that I kinda strained my back lifting some stuff on Thurs/Fri. It doesn\'t feel particular bad but it sure feels damn weird. I\'m not gonna make myself start training anytime soon until I\'m sure it\'s nothing serious. I\'ve seen too many bad back-related injuries in NS to take this too lightly. Anyway, it may not be bad to stop lessons for awhile, I can concentrate fully on finishing my ePrep course and job-hunting too. Oh noted one point about my CV during my interview. Apparently they kinda got the impression most of my experience were all very short-termed. I quickly explained the dates to them that all these were fixed due to sch vacation dates, contract, internship dates etc. Looks like I better include a new point \"Reason for leaving\" in my CV just in case most people aren\'t aware of how the local academic dates work. Just felt this might be a very important point in a CV.   '),(26200,979795,'07,January,2004','2004-01-07 00:00:00','    Had a stroke of good news today. Got another interview this Fri. Will probably need to drum up some excuse to get the morning off then, afterall, I\'m not being paid yet so not quite a loss anyway haha. Anyway this new job sounds quite interesting. It\'s labelled as a Software/Helpdesk Consultant, but it\'s with a co with some European roots so there\'s actually an opportunity to do quite a bit of travelling within Europe. Not sure how things will turn out, but I hope it\'ll be good. Anyway the company is going to be in Sci Pk 1, so getting to BVCC and BB East should be considerably easier for me. Woohoo... no excuse to not train liao ... well maybe unless I keep forgetting to bring parts my gi ...   '),(26201,979795,'06,January,2004','2004-01-06 00:00:00','    Today  (well, technically speaking it\'s yesterday coz it\'s past midnight already now) was an eventful day. 1st day at work. Didn\'t find an apporpriate time to talk to the boss about my pay yet. Will probably need to do so tomorrow. My mom was worried that they might be exploiting me, but I don\'t see it lor. Everyone in the office is seriously nice, but I do agree that the talk about pay has to be settled sooner or later. I guess I\'ll just bring it up tomorrow nicely lor. Seriously hope for at least $2k, but I\'m seriously considering taking the job for a little bit less also. But I think I\'d be spoiling market if I accepted something way too low.  Anyway, today was spent mostly troubleshooting PCs and some hardware problem in their office. I can only KIV it because after 1 whole day, I still can\'t say for sure why they have so much problems. OK, being a small company kinda made them really messy. But I can tell the boss is really friendly to everyone working there. Everyone is addressed on a first name basis, the boss included.  Went to BB East for Aikido. Was going to change in the toilet when I realised I FORGOT MY GI\'S PANTS!!! ARGH!!!! .... Seconds later Alvin entered the toilet and I had to tell him the pekchek news. DOH!! Went into the dojo and everyone was asking me why I\'m not changed for training. Oh Man!!! I received more than one suggestion to train in my jeans instead. I would if only I could. Sigh :( But sitting along the sidelines today wasn\'t a waste. I noticed a lot of points in Alvin\'s class. A lot of kokyunages from katate, morote, ushiro ryote doris, etc. Each one of them an extension of the previous technique. The things to note especially was centering, leading and footwork. I think from his past few lessons, he repeats these points almost religiously. I must say, I can tell what I\'d be doing wrong just by watching everyone today. However towards the last 20mins or so of class, I seriously felt frustrated because I really needed an uke to let me feel what I\'m trying to fomulate in my mind. At the end of class, I took the same bus as Alvin again. Always enjoyed talking to him about stuff off the mat because I always pick up some way to being \'Aiki\' off the mat. Told him about not getting into the SPF and he was quite encouraging. I guess some of the stuff he\'s saying I know them as well, but you sometimes just needed to hear them from others to register it in your mind. Anyway was quite a good day lah.  Got back home and something nasty happened which I don\'t feel like blogging in too much details. All family related but I wasn\'t involved at all this time. Still, find it very hard not to be burdened by some of the disputes / conflicts sometimes. That feeling of loss and helplessness is very irritating. I don\'t want to be judge, jury and executioner but I think everyone was at fault. Even myself even though, like I mentioned, I wasn\'t involved. Shitz ... sometimes I feel that every problems I noted are related somehow or other to the material world. If only luxuries didn\'t exist and everyone knew the contentment of just living simple lives. An unrealistic dream, if I do say so myself.   '),(26202,979795,'02,January,2004','2004-01-02 00:00:00','    Went for the job interview this evening. It was a small company which specialised in printing, manufaturing souvenirs, badges etc. The boss was a nice chap. Throughout the conversation I could feel like he was accessing me, though in a relatively casual sort of way. Turns out what they need is someone who can do almost everything and assist the boss. Seems like the job scope, includes creating systems (not programming, I mean more like procedures) to help their admin, maintaining the PC(s), software and quite possibly hardware troubleshooting, general admin, coordinating work, etc. In other words, the works. It seemed like a comfortable place to work in, despite it not being anything like the big corporate places most graduates seem to be working in these days. We\'ve not really discussed the specifics of the work contract or pay yet, I\'m hoping to leave it till Mon. Seriously hope it\'s at least decent, otherwise I won\'t be able tot take it up even if I really wanted to. But I\'m wondering what aspect I can pick up here. Besides the expected admin and business knowledge, there may not be enough in a technical sense even in the design and IT field. Serious delimma, I\'m not hoping to jump into IT, but I\'m also not trying to trade that off for something too broad. The boss said something about having \'faith\' in the company (dun ask me how that came up), and I have to admit it\'s something I don\'t have. But that\'s probably natural because it\'s the first time I\'ve actually seen the co.   '),(26203,979795,'01,January,2004','2004-01-01 00:00:00','    Well knackered out today. Woke up this morning and started packing my room. Been at it for almost 11 hrs or so, but at least my room looks like a place for living in now. Will try to make it a point to clean up regularly every Sun morning just to make things easy.  Well, I\'ve classified what\'s junk; to be thrown, what\'s to be donated (gotta call Salvation Army tomorrow), what\'s to be kept. Seriously man, I just realised there\'s a lot of small knitty gritty things I would like to keep but I had to stone wall myself otherwise I\'ll be keeping practically everything. My dad was well impressed with my effort and initiative. Somehow he also tried to do the same thing for himself but I think it got worst, so he was quick to lose his momentum. I was practically sneezing and wheezing a little because of all the dust. Hah, also noted that there\'s a lot which I\'m getting rid off without asking my parents. If I did, they\'ll be nagging me and making so much noise about throwing good stuff out. I seriously have to bypass them if I need to get this done. And no more waiting around to consolidate what needs to be given to the Salvation Army/Red Cross/whatever. All these will just delay me from getting everything done. Hope to clear up the categorised and packed stuff within a few days. Next week, if I\'m still jobless, I\'ll start chopping up my old toys cupboard for firewood muahahah ... so therapeutic. Gonna shift my furniture around alittle here and there when I do get my new loft bed. Hopefully by then I\'d already have a job so I can get some new shelves to replace the space for the old cupboard.  Oh yea, tomorrow I\'m checking out a Graphic Designer cum Coordinator job. It\'s located fairly near my place near Whampoa Drive/PIE/Jalan Toa Payoh. Hope it\'s something decently paid with a good and enjoyable workscope so that I can accept it on the spot.   '),(26204,979795,'29,February,2004','2004-02-29 00:00:00','    What a boring weekend. Got so much to do by Mon. Worse still, I was down with high fever after work on Sat. My whole household has got it too. Probably some kinda viral fever going around now. Anyway, I felt something was wrong so I popped two paracetamol and water paraded myself for a straight 4 hrs or so. By the time I went to bed, I felt a whole lot better. Sun morning came and I think I managed to get over my fever ok. But still felt weak and as though a sore throat/cough will be coming up soon. Oh well, that\'s normal whenever I get fever. Worst thing is that I need to formulate some teaching structure when class starts Mon onwards. In fact I\'ll probably be quite taxed for the next 6 weeks.  Some friends from NUS just formed a committee for GOLD - Graduates of the Last Decade. They pulled me in as a member sometime ago. Been missing  the past few meetings and just when I thought I can join the one on Sun, I fell sick. Shucks, didn\'t want to make it seem like I\'m trying to siam the meet but I\'m in no condition to join them at all this weekend.  Sat was suppossed to go for volunteer briefing at SPCA. Problem was that it was conducted at 1130hrs and I was still working at that time. Had to miss it but they said they\'ll schedule another one for Mar. Hope this time, it\'ll be in a more decent time slot.  This Tues, I think I\'ll approach Sensei for permission to train with Yuishinkai on Friday nights. Dunno how he\'ll take it. I dun wanna make it seem like I\'m belittling AFS, but I just hope to expose myself to their form of exercises and training style. For one thing, I like their \'extra emphasis\' on Ki compared to Aikikai style, but I also don\'t want to train completely in Ki style. I find that they are a lot more vocal in class (like during Alvin\'s classes), something I can relate to in UK under British Aikikai.   '),(26205,979795,'27,February,2004','2004-02-27 00:00:00','    Today was weird. I was the only non-beginner today with 2 white belts when Richard arrived. 3 of us laid all the mats ourselves. Everyone else came just a  bit later than usual. I was surprised to see Darryl from BVCC drop in as well. Then Bridgette came. After class started the twins Darrel and Daniel arrived too (later I found out one of them got his slippers stuck in mud and was \'trapped\' there for over an hr while he waited for his brother to go home and grab another pair of slippers). 3 sec after Richard started warmups, he asked me what belt I was. Told him 5th kyu, then he wanted me to take the warmups. That\'s when I kinda freaked. Told him I\'m not familar with taking warmups but he said we all needed to start from somewhere. So I took them thru what I could remember, with some pointers from Richard. Ar ... it still feels weird to me lah. Made mistakes here and there probably coz I felt nervous. Never thought I\'d take warmup at all one day. Still dun really want to lah. Dunnoz, figured I\'m not the kind to take up too much dojo responsibilities at this stage, just enjoy being one of the foot soldiers so I can train. Anyways, I did take the warmup slow and easy because I felt that usually we were just rushing through the motion of it too quickly. Of course, I didn\'t make them do any killer warmups that Eileen or Sunny did. Haha ...  We did gyaku and ai hamni nikkyo today. Went through very slow processes but it still felt like we all learned something today. Hah ... Richard brought his toy gun again tonight as a prop for nikkyo techniques. Farny man ...  Btw, asked the twins to come early next time coz I really dun wanna take warmups anytime soon. Doh...  Like to share some things Richard said as well which I, well, never exactly quite thought over before. He mentioned that uke and nage when practicising are on equal standing. No one is better than the other. A senior needs to bring himself down to a junior\'s level and the junior to that of the senior\'s (mabbe not in terms of technical stuff, but could be in terms of attitude, observations, etc). As a senior I why should I feel good at being stronger than a junior? No basis for feeling proud of that as an achievement. To me it\'s ego at work. Something we need to rectify and eliminate. All of us have heard this before I\'m sure, that we should help each other when it comes to learning the technique.  A point on ego that Alvin Khong told me a few months back. When I asked him how long he has been training, he said \"Since the year xxxx, you go and calculate.\" He says he don\'t like to keep track of the no. of years he has been practising Aikido. It\'s only to boost your ego. Why would you be counting up the no. of years of training anyway? Counting up to what? To tell people about it? Makes me so ashamed man. Before we had that conversation I\'d always tell people, I\'ve been training xx no. of years. Now I\'m trying to adopt the same attitude by saying \"since 2000\". But I gotta admit in mind there\'s still the thought that\'s counting, more out of habit than anything else.  Daphne and I had a talk about something last week on \'False Humility\'. It\'s like when people praise you, and you just say things like \"No lah, no lah\" \"Dun have lah\", etc. She says I have a problem about listing out my flaws all the time. She said I need to be more accepting about my strengths. True I think. Told her it\'s cause I hate it so much that people who dunno me gets a lot of false impression about me. Simply HAAAATE it. I rather people see me as I am, even if it means seeing my flaws first. But it\'s true, I tend to refuse compliments often. I think there\'s nothing wrong with simple \"Thanks\" for the nice comments afterall. Something I need to look into. Heh :)   '),(26206,979795,'25,February,2004','2004-02-25 00:00:00','    Right, yesterday due to a stupid misunderstanding. Lim Koon and I kinda got Daphne mad at us for the whole day. Worse, we didn\'t realise it till today. Luckily she\'s not the kind to bear grudges. Actually wasn\'t actually anyone\'s fault. I felt that it was just a huge misunderstanding. Anyway, never mind. It\'s over and it doesn\'t seem that anyone\'s been seriously psychologically scarred over the matter :P... whew.  Aikido was great. I realised a lot of things about my technique as I was doin it. For Shihonage, Gary\'s pointers on keeping arms slightly bent is good. I tend to pull with straight arms. I went home and thought things through but I think the bent arms helps me a lot more to bring uke down to my centre.  Lotsa kokyunage free practice in the final 30min. My group was with Karen, Li Hsing (Johnson\'s gf) and Alvin (I think, I\'ve seen him but not sure if that\'s his name). I think I was basically rushing too much liao, I felt like I didn\'t manage to maintain my balance, posture or even do proper atemi. But still, at the end of the session it was fun. Did a series of kokyuho with Darryl again. He\'s better at feeling the force from uke than I am. But I felt that he has a bit of \'Wallop\' in him too. They both tend to be more concerned with showing what they can do, than help their partners. It\'s not like they are selfish or anything, they just think that they are helping better in that way. I\'m saying this because at the end, Darryl was basically the one doing more of the feeling movement instead of just bringing me down. But he did so much until I didn\'t think I was an effective uke as I didn\'t have the energy and strength to keep applying force against him. Oh well ...  I\'m a bit nervous about starting the eLearning classes next week. I will be busy for the next 6 weeks or so taking classes of students and teachers. I have about 60+ teachers to train, re-train, conduct workshops and 1500+ students to take. All of which needs to be familar with how to use the eLearning system around early April. I think I\'m going to ask Sensei if I can go Pillars from next week onwards because I\'ll probably need to do OT ocasionally. It\'ll also mean I can probably grab a bite before training coz Pillars and Siglap are so much nearer. So far I\'ve been skipping meals whenever I need to go BVCC. Not a good thing in the long run.   '),(26207,979795,'23,February,2004','2004-02-23 00:00:00','    Right the booths setup at Lido today was by the Ministry of Home Affairs. Seems like recruitment was more for SCDF and Coast Guard. Both were interesting lah but neither was what I really wanted. Anyway, I spoke with a lady and I was told they weren\'t taking any SOs now. Each year there\'s only 2 batches of intake and sometimes just 1 batch even. My heart sank when I heard this. But oh well, I\'m happy with this job now and I\'m sure I will learn stuff here. Might as well make use my stay here and pick up what I can first.  Tang did advise me to take up the Coast Guard thing first and just go through the 9mths training in Police Academy first, then request for a transfer to the land unit later. But I know that approval is subjected to vacancies in the force. If there isn\'t any, I seriously don\'t want to be stuck within the Coast Guard.  Also spoke to the lady about the contract scheme. Bond is just 2 yrs, not 6 yrs which I initially thought. That\'s good news. But I forgot to ask her about UN Peacekeeping. Was also told that they will advertise in the newspapers and on the website when recruitment for SO is open again. She also that if they were taking anymore people this year, I could send in my resume sometime in Aug to the ManPower office. I won\'t need to send in my application anymore though because they\'ve stored it in a database already. Well I hope the rest of this year I can build up my resume and get a reference from school. Maybe that will help my application then too. She also mentioned somethat the 3 wks overseas leadership training is conducted in Thailand now, instead of Nepal like I had thought. I guess they rotate the locations every now and then. Sure hope that when/if I get in, I\'ll be able to go somewhere either really nice or exotic. Thailand\'s OK, but it wouldn\'t be my first choice (as though I\'ve any).   '),(26208,979795,'21,February,2004','2004-02-21 00:00:00','    Been getting busy in school these days. Things look like they\'re picking up. But it could be because there\'s a lot of last min stuff that needs to be done or at least made presentable before the Minister of Education\'s visit to the school coming Tues. Goodness, I know Tracey (HoD IT) would like to present something on eLearning during the visit but I\'m hoping I can do something light then instead. Still not sure I feel up to it at such an early stage.  I\'ve been working on getting the IT notice boards up past week. Took me 2 days later than intended to finish it up. Basically I put up some minor content (after amending the language to suit primary school level) and surfed around for cute pictures to be placed on the board too. Well, I decided on Neopets. Seems like a craze of sorts with the kids these days. They do look kinda cute lah but I got no idea what those creatures really are. Most look like a cross between 2 different animals, while others I got no idea what they really are. But it catches the kids attention. While pinning it up on one of the boards, some pri one students started crowding around behind me and they just went \"Oooo, so cuuuuute.\" o_O\"... I dunno if it\'s a girl thing lah, but when I sat in one of the classes Daphne (resident IT trainer) was conducting - MS Publisher -, some girls presented their final work. When pictures of animals or pets were shown, they just went \"Ohhhh, so cuuuute\" again.  Well, my friend Tang informed me of a Ministry of Home Affairs career fair this weekend. Will be meeting him for lunch before we pop by together. I know SPF isn\'t recruiting but there\'s no harm taking a look either.   '),(26209,979795,'17,February,2004','2004-02-17 00:00:00','    Tonight\'s training was pretty ok. Did a bit of group work here and there. But think the greatest learning point was during kokyu ho with Johnson and Darryl. I\'m beginning to feel something there. Though I\'m still somewhat stiff when it comes to executing it, but I did manage to find the right way myself at least twice. Geez ... sure took me long enough.  We had quite a few beginners today so Sensei took them through the basic ukemi and maybe some simpler techniques. Ah Hee Sensei took the first half of class today. It was nice. Pairing up, we started in shomenuchi position (uke and nage\'s extended arms touching against each other), nage kinda guided uke in while doing and irimi tenkan himself. Nage will kinda follow in and flow to face the opposite direction. Ends with both facing each other again. Did these flowy movements 2x each and non stop. We progressed on into ikkyo, sankyo and kotegaeshi using these flowy movements. I think it was good coz everyone tried to be continuous.  2nd hr, Sensei took the class again. We did a series of kokyu nages in groups. Didn\'t maange to learn much coz everything kinda went really fast. Did manage to ask Johnson for help with my kokyuho during the break though.  One thing really pissed me off though. After class, I helped to keep all the mats and went to change. The stupid Aunty was nagging me to leave. I was TRULY pissed tonight. She was well aware I\'m helping to keep all the mats. In a way that\'s like doing her a favour so she can lock up faster too. Being the last one to change, I beared the brunt of her naggings. I was still buckling my belt when she started saying things like \"Dun dress up in here, if you wanna dress up go outside.\" Told her I wasn\'t even ready / decent enough to go outside. Fact was, I was trying to get myself minimally ready so I can bring everything out to pack nicely and tidy my shirt at the lobby. Here I am trying so hard and I\'m still getting nagged.  I\'ve concluded that I\'ll try to avoid training on days that the naggy aunty is around. If I come on a day that she is, I\'m not going to finish keeping the mats anymore. Just pack up and leave halfway. I\'m getting seriously pissed with her responses.   '),(26210,979795,'15,February,2004','2004-02-15 00:00:00','    Today was the day of the Aikido demo at ECP. Gotta admit things went pretty ok. I loved watching Syed\'s demo with his Pillars Youth students and all of them taking slams. After the whole thing I walked to the busstop wif 4 students from Pillars and found out that they paid mat fees monthly instead of every 3 mths like the CCs. Hmm I\'m now considering taking additional training from there as well. Oh just a note, I did help a bit as uke during the demo but I think my ukemi was stiff and awkward. Well, there\'s always a first lah. Fun though, but I wouldn\'t want to really try again till I can take more decent ukemi and apply proper atemi.  Another reason why I felt really happy was because I met my sec 3+4 form teacher, Ms Lim, Well technically she\'s Mrs XXX now I think but I didn\'t remember to clarify this. Anyway I was giving out Aikido phamplets when lo and behold, this lady (she didn\'t change much even after giving birth) walked past, refusing my phamplet at the same time (HUMPH!). Anyway, my first instinct was to call out \"Ms Lim! Ms Lim!\". She didn\'t turn her head but I noted that the name did make her turn her eyeballs in my direction. So I called out again and kinda waved at her \"Ms Lim from Buona Vista Sec?!\" That kinda got her attention haha. Well, she couldn\'t remember my name lah of coz since I hadn\'t seen her after my Os (she was posted out soon after I graduated). Updated her my past 3 years and told her wat I\'m currently working as. Apparently she\'s now with Tampines Sec (or was it Primary??). OK lah, at least not too diff to find out if I need to look for her again. Anyway, this kinda reminded me of my Sec 1+2 form teacher, Mr Lim. I think someday I gotta call up MOE to see if they can help me out liao.  Oh, saw the huge blackish bruise on Eileen\'s arm and it was horrible. I\'d imagine it feels worse than it seemed lor.   '),(26211,979795,'09,February,2004','2004-02-09 00:00:00','    Just got back from the combined training. Whew I\'m really tired today as the unofficial cameraman. Seriously, not having a  tripod makes it difficult to hold out for so long.  I\'m quite pissed with myself actually because I didn\'t manage to get good shots using Eileen\'s digital cam. Most, if not all, were way too blurry and fuzzy. And to make things worse during Sensei\'s certificate presentation I didn\'t manage to snap the photos fast enough either. Haiz ... I should\'ve spent more time using it before the actual presentation. Either that or I should\'ve used the videocam to record down the whole sequence. In any case, I messed up.  Anyway I tried liao lah. Can\'t do anything about it now.   '),(26212,979795,'08,February,2004','2004-02-08 00:00:00','    OK I\'ve just gotten Sensei\'s approval to video down tomorrow\'s certificate presentation and the combined training at Tanglin. I\'ve even gotten permission from my HoD if I can arrive earlier in school so I can leave earlier too. But even then, I think my timing is going to be really tight as the last time I tried getting home at peak hour, it took me aaaaages. Anyway, no choice lah. Just do what I can liao. Eileen will be loaning me her digital camera as well so I guess I\'m going to be the unofficial cameraman for the night. But seems like she\'ll be late because of late classes. Everything seems to be falling into place nicely woohoo.   '),(26213,979795,'07,February,2004','2004-02-07 00:00:00','    Been to school since yesterday and it was quite an experience. Met the school\'s IT resident trainer, Daphne, who is from another company. She said she\'s new as well but she seems to handle the kids very well. I sat in for one of her webpage building classes. One thing that struck me was that the students may be cheeky at times but they were obedient. It\'s either the school has disciplined them really well or it\'s because they\'re just girls... I dunno. I was quite blur when I saw some of them already blogging at such a young age (think that class was Pri 6). Goodness ... kids these days. There were some really talkative ones too. But in this school a lot of the students were happy to pounce on the chance to show what they were capable of. Daphne actually picked some of the students to present to the class and they were really hyper about it. After that the last 15min was basically a mini concert where some students decided to sing for the rest of the class. I have to admit I was seriously entertained on my first day there. Can\'t stop laughing. They even did a dance. It just makes me wonder whether they do these kinda stuff in their spare time (choreographing that is), and somemore all these were all impromtu rather than pre-arranged. OK, it still feels ultra weird to be called Mr Tan and I don\'t really remember to greet them back when they greet me \"Good morning\" or something. If it was just passing by the corridor it would\'ve been easier. Today was a bit more boring. First of all I had problem getting access into the labs and had to wait for Daphne to arrive. The whole day was spent testing if their old educational CD-ROMs could run on their newly migrated systems with WinXP installed. Did finish most of it though but still left with some. Should be able to complete on Mon.  Which brings another problem. There\'s absolutely no way I can leave school and arrive on time at Tanglin CC on Mon evening for the combined training, being a new job and all. If the award ceremony is at the start (which I would assume it to be), I think I\'ll miss it.  I\'ve gotten confirmation from SPF saying they are not recruiting anymore SO at this time. Well, I\'m still disappointed but I think it\'s for the best now. I can concentrate on working for the next few months and learn something here. The school seems like a really nice place to work in and the Principal, my HOD (Tracey), Daphne and TA (technician) all seem and sound like nice people to be around. There\'s one thing I do wish for though ... more male toilets in the school. Hahah :P   '),(26214,979795,'04,February,2004','2004-02-04 00:00:00','    Tonight\'s Aikido was greeeeaaat!. I didn\'t do too fantastic things, but I just felt really comfortable. Got no idea why also. I think my ukes for today helped as well.  We did a form of kokyu nage backstretch which also seemed a bit like going into a shihonage. Dunno how to explain it, but I\'m sure we\'ve done it before. Paired with a relatively beginner who might have practiced for sometime before because he was a pretty good uke as a newbie. After that we did Shihonage and I think it wasn\'t as good as I really wanted that to be. Didn\'t get too many locks on my uke. Also paired witha  little chubby girl and Karen. The new girl was either sec sch or late pri sch and I think she felt really awkward about the training. But luckily Ms Wong was there to always reassure her. But she did apply a shihonage before she made my elbow \'point upwards\'. Think I was quite lucky to have reacted when I did so I got away unscathed. Sensei did call out to her but a bit too late. Think it\'s a common mistake for beginners to get this wrong lah, I still remember when I first did it. End of the first session was practicing sitting kokyuho with a lady. It was quite nice because we did koyuhos so continuously that everything was relatively smooth and we did not feel as tired as we might have usually been. But I think both our knees and legs were feeling quite sore because of all the kneeling.  Second hour basically was spent with Karen as my partner. We did a variety of techniques from iriminages to koshinage. I tried to apply what Gary said about keeping the arms relaxed and they did feel better during the technique. But I wasn\'t able to keep it up all the time. I thoroughly enjoyed today\'s class this evening.  Asked sensei if he will still be teaching in Siglap. He said yes. Apparently Sunny had secured a new job and is not able to teach on Thurs at BVCC. So Sensei will resume the same plan again after a period.  About my first day at work, it was pleasant. Went directly to school to meet the HoD. Met Victor and Emily from the co. who were there to discuss the specifics of the contract with the school. Nice people. After that I returned back tot he co. with them and tried to familarise myself with the system and services they were providing to the school. Though not difficult, there really was a lot to learn and remember. I did feel that the package was extremely well designed and its concept was good. It\'s a powerful and flexible package, I think schools who use it are really getting a damn good deal here. I think the main challenge here is how to make teachers adopt a more open attitude in learning elearning and making them want to apply it to their work. Because the possibilities of how the system can help them are endless from what I\'ve already seen. I wish I had taken some kinda Education Psychology stuff that I can apply to this kinda situation. I\'ve heard so many stories of how resistant to change people can get sometimes. There\'s quite a challenge involved as well because not to be confused with purely training that so many commercial schools seem to be involved in, our co is really much more into spearheading elearning. The initiatives they\'ve adopted is really quite fresh to me. All employees have to choose to develop an elearning module in 2 mths in an internal competition within the co. It\'s really to encourage employees to enhance their development skills and keep in mind that elearning is really the aim behind everything we do. There\'s like some kinda monetary incentives for winners of coz. I\'ll only be going over to the school from Fri onwards. Figured it\'s the best time to do so because the end of the week will give me some time to familarise myself with the new environment before I reeeeaaaally get rolling. I\'m expecting there\'s gonna be tons to do and catch up because according to the HoD, the school is already like 1 mth behind schedule. I\'ve already got a teacher\'s list and there\'s about 75 of them in all. Heard that students range around 1300 if I remember correctly but I\'ve not seen the exact list yet. Also managed to meet another trainer, Danny, whom the company had attached to the school across mine. It was quite refreshing   '),(26215,979795,'01,February,2004','2004-02-01 00:00:00','    Been keeping away from my blog for a few days coz everytime I tried posting comments on it, there\'d be some kinda stupid error. Got kinda sick of retrying and trying to remember my comments then.  Found a job at last. Multimedia Educator with an elearning development company called  urlLink AskNLearn . It was all pretty rushed because everything was confirmed last minute. I\'ll be attached to Katong CHIJ Primary as a content developer for the teachers there. I\'m to update and put up elearning content for the teachers according to the curriculum they tell me I think. Anyway the Principal and teacher in charge wants to meet me on Tues 8am first. Pay\'s 1.6k for 1st 6mths probation and there\'ll be an increment later (how much exactly will depends liao). Think the school also rushing now coz Feb has started and they\'re still not done yet. Haiz ... a lot of clashes. Actually arranged for the Salvation Army to come my place collect stuff the same morning, now I have to rearrange. I\'m hoping I\'ll get some time off to prepare myself for the job and settle my own stuff first. But I think that\'s going to prove difficult. My impression is that they needs things done urgently so I may at most have 1-2 days of training with the company first and then I\'ll need to go to the school liao.  Yesterday was my first mob briefing as a reservist. Spent the whole morning in camp listening to stuff I had already heard when I was still serving NS. A lot of things about teh stupid reservists around me sometimes. Some of them bring their \'civilian\' mentality into NS and I absolutely hate that. They say things like \"I will complain, must complain, etc\". I got really irritated hearing things like that. Obviously after ORD these are the ones who have forgotten what it was like during their NS. Once you are back for NS matters whether you don the uniform or not, you better jolly well behave like a damn soldier and not a civilian. Suspect these are the chow geng kias during their NSF days. Some of the new/young reservists also feel that after ORD, they are considered \'king\' lah, \'untouchable\' lah by SAF or something. Dead wrong, they obviously haven\'t seen many DB cases for reservists so far. I\'ve personally seen a group of reservists being handcuffed together in a string just awaiting the MPs to come pick them up. Worse case is for no reason/some stupid excuse on their part, they now need to disrupt their professional and personal lives, get their heads shaved, and enter DB to serve their sentence. In fact, the MPs were nice enough to give a certain \'discount/offer\' that if we had 10 or more who needed to be sent to DB, they\'ll provide the manpower and vehicle to come and pick up the prisoners themselves. I really never see the point why reservists sometimes do all the knitty gritty things to avoid their liability, if they had just finished things in the first place ... Anyway, after the brief I had a nice talk with my RSM and he was willing to act as my character reference (though he just knew me for 4 mths).  After the mob brief I was supposed to meet Lynette and then we\'d drop by to visit Daniel. Suay thing was my timing got kinda messed up and she spent a long time waiting for me. Got scolded by her over the phone which was quite well-deserved. Anyway I did a lot of walking and when I got to Daniel\'s they became quite sympathetic because I was wet with perspire. Had to take 2 cans of green tea and a huge bottle of plain water to cool down. If it was anyone else, I\'d be so embarassed visiting someone\'s place so wet with perspire. But with Daniel, aiya ... hahaha, Anyway his gf, Charlotte, was there as well. We all had a good chat and departed after about an hour a half. Almost had to meetup with my NS friends to get a housewarming gift for one of our kakis the next day. Was really glad it was cancelled because I felt really really really tired today. Felt like knocking out when I got home but dinner was to be served soon, so I pretty much made myself stay awake, finished dinner, called the company to settle some details before I could finally went to bed. Woke up damn early today and realised it was Feb now.  New Yr and CNY has been over for awhile now and it only made me realised I did waste some time in the past 31 days or more. I\'m glad I had finally gotten some backlot moving, things like clearing stuff for the Salvation Army to pickup and volunteering with the SPCA. Some other things I\'d really like to do is visit my NP lecturer, Agus, before I start work soon. Some disappointment I felt is that I didn\'t get into SPF. Even worse I\'ve heard of news saying that they might have closed applications for Senior Officers for the whole year. I\'m thinking I won\'t give up yet, but I\'ll still keep my options open for now. When I spoke to my RSM, he had mentioned that even the SAF have their vacancies filled up pretty tightly now. Seems like they now have a pool of people enough to last the next 3-4 years. Just imagine, if the SAF can afford to be choosy with their applicants, it\'d be worse for the other uniformed service isn\'t it? I\'ll try to check up on CNB side as well though.  Weather seems great this mornin and it\'s still early to avoid the traffic. Think I\'ll go for a run now. Time to resume marathon training liao. Just another thing I had put off since CNY coz of the wet weather then.   '),(26216,979795,'26,March,2004','2004-03-26 00:00:00','    Right, today was quite a nice day teaching classes in the mid and upper Pri levels. They pretty much paid enough attention to me. I realised that the secret was just to get their attention. Worst case scenario is that after 1 hr you still can\'t find the right way to get everyone\'s attention that will kinda make one pekchek (frustrated).  I also got the chance to stand in for one of the P4 iClub (something like an IT club). When I first entered, the kids were kinda trying to siam their assignment and they almost immediately started surfing for irrelevant stuff when my HoD (their original teacher in charge) left the room. All the while saying their excuses out aloud for me to hear, and trying so hard to justify why they weren\'t doin work. Only 1 group concentrated more on gettting their stuff first before they started surfing around for leisure.  Very first thing I did was try to be firm with them and said you MUST get your stuff ready before you can surf for leisure. That didn\'t seem to work. I was aware that they were surfing for fun out of view. Tried to stop them once or twice but I didn\'t think it\'d be useful.  Me: Now are you sure you are allowed to surf around for Neopets? Kid 1: Yes. Me: (Don\'t exactly believe them.) Are you sure? Then what did Mrs Hoa tell you to do? Kids 2 and 3: Nothing. Me: Nothing? Are you sure nothing? How can that be? Kid: She just said to look around for fractured fairytales but we cannot find. Me: Now, she has already told you what you need to do. If you want to surf around for other things you can do so after you have done her work.  Some mumbling went on around the groups and they kinda ignored what I had said. But after awhile one of them noticed me typing away. I was actually typing out a question paper for my HoD as she said she wanted it later when she popped back in. It seemed urgent for her so I thought I might as well just spend about 30mins to help her out in it.  Me: (Typing away.) Kid 2: Mr Tan, how come you can type so fast? Me: You just need practice, the more you type the faster you will get. Kid 2: Are you typing the things for the elearning? (Because I had just taught this girl\'s class 40mins ago.) Me: No, I\'m trying to finish typing this for Mrs Hoa, she will be coming back in later. Kid 2: Huuuuuh! Me: Now you see, that\'s why I\'ve been telling all of you to do your work first before you start doing all these other things. At least finish your work before you start playing around because Mrs Hoa will be coming back later. I dunno what she will say when you cannot finish your work. See, even I am trying to rush her things out first so I can give it to her later. Now, I leave it up to you. If you want to play, you can continue but if she scolds you later, you have no one to blame but yourself. (And I went back to my typing in earnest.)  The kid kinda went balllistic and told the others. That helped because I started to notice them working a bit more though they still kept their leisure sites and homepages opened. But I figured, as long as they were working, it was ok. Slowly one group after another started saying things like, \"Mr Tan, I just can\'t find the fractured fairy tales\", \"Where to find\" and \"Cannot find leh\". To help them kick start a bit, I decided to help them search around for some link first. Found one and wrote down the URL for them, told them to pass the paper around to the other groups as well to share the resource. Some felt rather surprised and others even admitted that they saw the link but never clicked on it. My answer to that was simply, \"Now, the Internet is so big. Everything can be found online. It just takes a little bit of effort.\" Hahaha the usual Gary Tan\'s \'everything can be found online\' statement.  Slowly group by group they started to come over to ask me for small help here and there like \"Mr Tan, my link cannot work\" - typo error of course and \"Mr Tan, can I print out the stories for the script\". Felt happy that the different approach worked. As I went on to finish my own work the girls started surfing around for those ghost websites. But at the same time they\'d get so scared of the sound and flash screens of ghosts zooming here and there that they started screaming around in the lab.  Girl 1: Eh come and see this come and see this! Very scary one! Girls: (Started huddling together over this particular PC and staring intently.) What? What? Girl 1: You stare at this for awhile and the ghose very scary one! Girls: (A bit of silence for awhile ...... then suddenly ...) AAAAHHHHHHHHHH!!!!! (and everyone run away from the PC). Me: Goodness what on earth are you doing? Girls: Mr Tan Mr Tan, very scary. You come and see. (Another says,) Ya Ya Mr Tan come and see. Very scary ghost. I dare not clost the window now. Me: (thinking to myself) Is it really so scary? Wah if I also get shocked then it\'d be so \'diu lian\' (embarassing). Me: (walks over feeling a bit curious) What is it?  Apparently it was some website with tons of flash based screen shots and animations of the supernatural. Some looked really faked though. But it still made the girls scream. Anyway the girls showed me quite a few flash screen shots and some of them were preview shots of horror movies. I even saw one of \'The Eye 2\'. It seemed scary enough though, but with so many kids making so much noise and all that, it was kinda an anticlimax. After awhile they got kinda bored with trying to scare me with these flash sites.  Girl: How come you not scared one huh Mr Tan? Me: Aiyo, how to get scared. Everytime I watch the things you are all around me screaming into my ear before the thing even appears.  Anyway, towards the end of class my HoD came back and they did manage to show her the work they did. She seemed satisfied enough to leave them for another 15 mins to surf around for fun. Aiya ... kids lah, harmless to give them a bit of R&R I suppose. Anyway I did enjoy the company as I worked on my company reports that afternoon.  The final class that I took after that was a Pri 3 class and they were so inquisitive that it was quite enjoyable teaching them as well. One thing I realised, the P3-4s enjoy doing quizzes. They kinda treat them as games. So I\'d promise them I\'d let them go through the quizzes if only they\'d listen and let me finish up fast so they\'ll have more time to work on the quizzes. Turns out there wasn\'t enough time and I still had to make them do a survey. But they treated the survey as a quiz and some even enjoyed doing these. Haiz ... kids can be so naive sometimes. But what they don\'t know won\'t hurt them lah.   '),(26217,979795,'24,March,2004','2004-03-24 00:00:00','    New school term for Primary school started yesterday. The classes I took were all pretty pleasant. I did feel a bit more pressure teaching the upper Primary though. They were a bit too quiet and it made me all the more nervous.  Anyway, for yesterday and today, there were quite a few cancellation of classes. The feedback I had received was that everyone thought the updated copy I sent out last week was not confirmed yet. Anyway, not a problem lah. My second email to the school seems pretty straightforward liao. Things should work out just fine after this.  There was a fire drill yesterday. Daphne was taking a class of Pri 2s in her lab and I was in a separate lab doing my own work. Suddenly the fire alarm went off. First of all I tried to ignore it for a few secs/a min. But I found it to be too irritating a sound to be something minor. I immediately suspected some kind of an emergency or a real fire. I decided to pop my head out of the lab for a look. Most of the kids were lining up and moving towards the field. But out of all these activity, I noticed the Admin Manager standing outside Daphne\'s lab and being all calm and relaxed. That\'s when I knew for sure that it was only a fire drill. I walked towards her and I was thinking \"OK, can check if Daphne had moved the students out of class already or not as well.\" As I neared the Admin Mgr, I did notice a standing board with fire painted on it. For a split second, my brain actually registered that it was the exact spot where the \'fire\' started. But I still decided to walk right past it.  Admin Mgr: (sees me) Eh you just got burnt. Me: Huh? Oh, erm ... fire drill huh? Am I involved? Where should I go? Admin Mgr: Just follow everyone to the field. Me: OK (Decided to take a longer route to avoid walking past the \'flames\' in front of the Admin Mgr :P)  At this point I didn\'t pop by to check on Daphne\'s lab even though I was just outside. I kinda assumed that since the Admin Mgr was standing outside doing nothing, maybe the class was empty and Daphne had already taken her class out to the field sometime ago. Not bothering to check, I just went to the field blindly like a sotong. When I reached the field. Our HoD asked if I had seen Daphne and I said \"I think she knows, she should be down already.\" (Again, I\'d like to use the army term \'You think, I thought, then who confirm?\') Tracey suggested that I could call her to check. After standing around for awhile, Daphne was nowhere in sight. That\'s when I decided to call her.  Daphne: (answers call) Ya Aylwin, I\'m in the middle of a class now. Me: Oh it\'s a fire drill now. Daphne: Ya but I\'m in the middle of a class now. Me: Huh? ... (a bit lost for words) ... but it\'s fire drill. (Can\'t remember what she said right after but I just thought, maybe she knows what she\'s doing. Or she knows that she is excused from the fire drill.) Me: Oh ok lor (Hangs up).  At this stage I did feel a bit stupid. Thought I didn\'t have to participate and there I was standing in the field with everyone. OK, mistake made. So just went along with everyone until the \'all clear\' was given. Then suddenly who should I see coming down to the field but Daphne and the class\'s form teacher witha bout 30+ little Pri 2s trailing behind them. OK that\'s when I thought \"Heng ar... I came down liao.\" Haha ... I went over lor and they were commenting/discussing amongst themselves over the mistake. Later on I was told the form teachers said that the kids were all \'Chow da\' (burnt). Well, was a comic day lah with me and the kids all getting \'burnt\'.  Today I had some classes in the morning while most of the remaining ones in the afternoons got cancelled or rescheduled due to the same reasons I mentioned earlier. While doing my own work in the lab, a class I had taught before came in for their lessons. These were all Pri 1s and they were quite funny partly also because of the way the teacher was trying to handle all of them at once. There was this particular little girl who had trouble handling the CD-Roms and she was too afraid to ask for help. She was kind breaking down into tears quietly.  Kid by name of Vanessa (2 seats away from Cying Kid): Mrs Lim, Samantha is crying. Mrs Lim: Why? She\'s always crying. Why is she crying?  Now Vanessa is one hyper kid who irritates her fellow classmates very often. I think she\'s the one who wants attention from people (I posted about her a few blogs ago).  Mrs Lim: Yes and you Vanessa, you are so busybody that Samantha crying you also can see from your seat so far away right?  Now Mrs Lim is only saying this because Vanessa tends to do alot of inappropriate stuff at a inappropriate time. Always so hyper and disobeying instructions (purpose, if you ask me - just for the attention she can get). Anyway I was keeping my ears opened and I took a look at what could actually be Samantha\'s problem. I just saw a quiet meek mannered little girl with a pony tail and big round eyes staring at the monitor. Aiyo ... also dunno how to help her. From what I can gather, I think she just feels helpless and she\'s one of those who is too shy to ask for help. Pri 1 lah, I guess it happens. Anyway Mrs Lim just tried to explain to her that if she needs help she should just ask. I guess she\'s handled enough of such cases before to know that \'spoonfeeding\' the kids too much won\'t help. It\'ll just make them more dependent and they won\'t learn anything.  Anyway a few moment passed and I naturally just turned back to check out if Samantha might be having anymore problems. I think everything seemed OK lor. But she suddenly turned over and she saw me looking over at her. I just turned back in a nonchalant way and hope I did not manage to embarass her. But I could tell her face and eyes were slightly redder than normal because of all the crying earlier.  Later I saw her walking over to the teacher and I assumed she had manage to pay heed to Mrs Lim\'s advice about getting help when she needed it. But ...  Samantha: (Whispers something too Mrs Lim. Too soft for me to hear over the noise that the class was making.) Mrs Lim: Aiyo, why do you want to look for your sister? She\'s not free lah, she\'s in class now. That\'s when I realise that Samantha actually has an older sis somewhere in the same school.  Haiz, I also dunno what I\'m trying to say about the situation lah. Just think kids can be so difficult to take care of sometimes despite being so adorable.   '),(26218,979795,'21,March,2004','2004-03-21 00:00:00','    OK something really disgusting occurred on Thursday. That morning I happily donned my shoe and went to work as usual. Spent a whole day in school and went Aikido at Siglap after work. Arrived early as usual and started to change. Lo and behold, guess what I found? There was a freaking dead cockroach smeared on the right side inside of my right shoe. Ewwww ... freaking gross. Cleared out whatever of its remains I could and blocked out the thought as I went on with Aikido for the night. Can\'t quite shake the thought out when I had to change back and go home after class. Moment I got home, removed my shoes and went to wash my gi + socks (blood  smear on my socks). Thoroughly washed out the inner regions of my shoes with a  fibrous brush and an old tooth brush. All the while the remains were like being removed really slowly. I figured I must have worn out a whole chunk of my shoe with all the hard scrubbing I did. Lesson learnt? I always shake my shoes to throw out whatever dirt, insect or creepy crawly first before sticking my foot in now.  Yesterday went to River Valley High to take a company test. Technically I was late but when I arrived it seems like people had just settled in a few mins ago only. Test turned out to be OK. Was decently easy, but it\'s not like I\'ll be getting a high score in it or anything lah. Just felt that they were reasonable questions though some terms were a bit technical. I was lucky that during my first 3 days in the company, I had an expert like Angelica walkthrough me a lot of the stuff. After the test, I went to Clementi central to make a new pair of specs. Cost me like $240. Freaking expensive but I figured getting one I really liked will last me longer plus I guess I really needed the new specs. The new one came with 2 clip-on sunglasses. Always wanted to get one to replace my old set years ago. My old ones are either too dented or scratched from Aikido or dropping them now and again. One small news, turns out my old specs degree were a bit high. It was like right eye 500, left eye 700 (astigmatism). But after the tests, it was 500 and 650. Could\'ve reduced it to 625 but the optician said 50 degree reduction is a lot already and she recommends doing 650 instead because the eye needs some adjustment. After this, I met up with Tang for lunch at Mac\'s and we proceeded back to River Valley High area for Aikido training.  It was my first official lesson with Yuishinkai. Their ki emphasis is definitely a lot more structured then Aikikai. Maybe structured isn\'t exactly the right word. The thing is they use the words ki, ki direction a lot more than us. Just doing warmups alone, Frankie Sensei went through how our hips, center, arms, ki direction should be going. Very very very detailed. I can understand some but not all. At the end of class, he passed me some forms to fill up and a copy of their syllabus as well as one set of illustration for their warmups. Willl try to scan these in when I get the time, but anyone with Koichi Tohei\'s movie clip called Authentic Aikido will be able to see everything better. In terms of technique, everything was about redirecting ki, placing yourself, footwork, etc. That doesn\'t sound so different from what Aikikai style does, but practically to apply what I\'ve learnt from the warmups was a whole different thing. The thing is, there are so many similarities between Yuishinkai style and Ah Hee Sensei\'s. I guess that\'s why Freddy Sensei said, all Aikido is the same. Just a lot of different \'brands\' out there.  Stupid cough is getting from bad to worse. Think it\'s the cold drinks I\'ve been taking since Thurs till now. Sigh, have to watch my food and water consumption again. Mon will be the start of term 2 and I need to take the class lessons for the upper primary levels. First week is like my killer week but subsequently the afternoons do get better. But still, I\'ve got a lot of outstanding tasks not done or just started only. Need smokebomb liao, hahaha.   '),(26219,979795,'19,March,2004','2004-03-19 00:00:00','    Need to get so much stuff done today. School holiday almost at an end and I still don\'t feel like I\'ve finished the most urgent stuff I need to. Looks like I\'ll be sleeping late tonight.  Tonight\'s Aikido was great. I took warmup again because Daniel and Darryl were late. But the good thing was that this time I was a lot more composed. My mind was quite empty and I\'m pretty sure I took the warmup through the wrong order as well for some. No matter lah, the empty minded feeling I had just made everything feel OK.  We did a lot of iriminages. I managed to do some fairly decent ones tonight. Richard emphasised different way of feeling in iriminage tonight. I understood/felt some, but missed the rest.  I think I\'ll make it a point to grab different ukes/nages in future practices. Just to feel the techniques.   '),(26220,979795,'16,March,2004','2004-03-16 00:00:00','    Crazy day in school today. Laughed like crazee with Daphne in the lab till people were sure to think we were a bit gila. Bui tahan. But lunch was yummy ... MOS burger + chocolate ice-cream... mmmmm. Need to think what to eat for lunch tomorrow.  Went to grab a haircut today. Another garang haircut. But I still felt it wasn\'t short enough. Think I told the barber to cut no. 1 along the sides and back. Should\'ve told him no. 0 instead. Anyway while waiting there was a father and son waiting in line just before me. Maybe it\'s working in a school that\'s why I seem to be noticing kids more. But this kid was damn cute. He was scrawny and dark and his hair looked like it was cut in durian style a month ago. Just that now it has grown and the hair seems to be protuding from all angles on his head. His father was convincing him to go for a shave with the blade later but he simply refuses. His afraid of the razor blade and insist on just the shaver only.  Dad: Later go for a shave OK? You see, other kor kor so brave. Kid: Dowan. Dad: Want lah. Kid: Dowan.  While waiting his turn, the kid was so hyperactive and was contantly running around. The barber was busy cutting the hair of someone else at the moment.  Dad: Can you don\'t run around or not? I want to listen to the radio programme. (FM95.8 was on.) I can\'t remember his exact words but the kid was saying something like this; Kid: When is it my turn? Dad: Later lah when you get into the chair. Kid: (Hops onto the empty barber seat in the middle) I get into the chair already. Dad: Not this one lah. The one the uncle is cutting now (for current customer) lah. Kid: (Blur look.)  When the barber was done. He went for a quick smoke break. The kid settled into the chair to wait. Dad: (Just playing around while waiting for barber) Papa cut for you want or not? Kid: OK lor. Dad: But if Papa cut means must shave one hor (meaning using the razor at the end of the cut). Kid: Dowan. Dad: Want lah. Kid: Dowan. All the while the father was pretending to mess around with his son\'s head/hair. But it looks more like a head massage for the kid. Dad: (Massaging / messing with son\'s head.) Kid: (Turned head here and there.) Dad: What are you doing!? Later uncle cut for you can move around like that or not? Kid: Cannot. Dad: Ar, ok ar. Cannot move ar.  Barber comes back and proceeds with haircut. Kid still moves head here and there. And when he doesn\'t, he moves his hands to scratch his neck (thereby making his head move as well).  To help prevent his son from moving too much, Dad: Ar close your eyes ar. Kid: (Squeezes eyes shut so tight that he appears to be frowning throughout the haircut.)  At the end of the cut, Barber: (To father) Your son\'s head is like those Indian dancing like that, neck and head cannot stay still one. Dad: (Laughs) Kid: (Staring at his own reflection and new haircut with blur look on his face).  I look at the son and he looks even more impish. Resemble more like a monkey now with his durian/GI cut.  Bloody cute kid, man.   '),(26221,979795,'12,March,2004','2004-03-12 00:00:00','    Lastnight\'s Aikido was great. Syed came down tonight. Did a lot of practice in terms of proper atemi, combining intent with momentum, flowiness, etc. Both from uke and nage\'s point of view. It was good. But I think it still took some getting used to because most of us usually did things really slowly. Too slow in fact to be considered proper practice. Like he said, if uke doesn\'t come in properly, it doesn\'t benefit either one.  Classes today was quite horrendous. Had to scold 3 out of 4 of the classes I took today. But the good class was exceptional and by far the most obedient so far. Told them about it and asked them to give themselves a round of applause. Haha...  The naughty ones, 2 of which were the older ones in Pri 3. Seriously man, at that age, they feel like they know enough things to be able to take a lot of short cuts and start ignoring instructions. Had to raise my voice at them a few times. Luckily Tracey gave me off tomorrow and school holiday starts next week. I can take the 1 week break itself to rest my voice as I\'ve been feeling a bit of sore throat coming up since 2 days ago. Sigh ... can\'t imagine how I\'ll need to discipline the older batch of students later a on.   '),(26222,979795,'10,March,2004','2004-03-10 00:00:00','    Lastnight\'s Aikido was OK. Paired with a beginner for most part of the first hr. Did things like iriminage, shihonage etc. One thing he tended to do was that after being thrown, he just collapses flat on the mat. Much like what Leon does. I think he doesn\'t really know that he needs to pop back up asap, or that it\'s a bit rude in terms of dojo etiquette, and how dangerous it can be if someone just collapses on top of him by accident. Well, nmind lah. His first day, so better not scare him away with all the details yet. Maybe let him know about it the next time I see it.  2nd hr was more different. We did a few pair work for iriminage and kotegaeshi. After that we got into groups for free practice. I think Sensei has gotten a bit sian of us taking so long to get into groups so he basically went \"1, 2, 3, 4, 5, 1 group. 1, 2, 3, 4, 5, another group.\" and so on and so on. Well after he finished pointing out my group, I turned around to face everyone. That\'s when I realised .... Jun was in our grp. I was like \"Oh si liao si liao.\" Then I thought, \"Heng ar, not doin nikkyo.\" Coz I still remembered the 1st or 2nd lesson when we had just joined BV and we did suwariwaza nikkyo. Wah ... the feeling still can remember sia. Anyway, the others in the group were, a law undergrad from NUS (tall pleasant dude I\'ve been training with on and off, pretty decent in terms of technique also - 5th kyu I think), Chin and Lee (the Filipino 3rd kyu, think that\'s his name right?). Naturally none of us can move Jun in our techniques. The effect was something like this:  Nage: (Struggle with technique. Struggle struggle struggle...) Jun: Can\'t be moved / thrown. Nage: (Gives up) ...  Jun: (Forward roll himself) Nage: (Laughs at self and thinks ... \"How on earth to do it on Jun?\")  Well, Lee was quite unlucky for training lastnight. Kept getting small knocks here and there. Jun\'s iriminage kinda rubbed the gi a bit too hard on his lower mouth so it felt sore after that. Then Jun\'s shomenuchi kinda went a bit too hard and knock him on the forehead. Well no serious injury came about, but you can really see him getting little knocks by someone every few mins. At the end of class, I spoke to him for awhile and we just laughed it off. Oops ... must remember to burn him a copy of Hattori Sensei\'s vid to pass him next Tues.   '),(26223,979795,'09,March,2004','2004-03-09 00:00:00','    Class just ended. Gosh ... took quite a drain today. Took 1 class in the morning. Had 1 hr break, took another class, had 1 meeting right after, took a quick lunch, had a nother meeting, break another hr, took 2 classes after that. So tired now, need some Aikido to clear my mind. Hope I won\'t be too blur tonight. Luckily tomorrow is Sports\' Day. I get to wear casual and I may/may not have a halfday. But I\'m considering staying later to finish up some work. If I do leave early, I may go grab a haircut instead. But I\'m trying to delay it till next Tues instead.  I\'m starting to recognise some girls from the classes I\'m taking. First was the little one who grabbed my hand in the library. I was walking past her seat and she grabbed my hand to try and ask me a question. Another is from a different class. ALways running around or irritating the friends next to her. Noticed that basically both just want attention. The former doesn\'t distract others while the latter does.   OK I have to take note lah, can\'t let the former just happy happy grab my hand or something. Teacher is still teacher lor. Need to maintain some kind of discipline in class.  I\'m thinking more on how to \'teach\' the latter. I may have to try some \'ignoring\' tactics on her. Maybe isolate her by making her seat far away and alone from anyone. She gets really happy when she notices people gettting affected by her actions. Sit with her legs on the chair, places her feet on her friend\'s seat. Runs around pressing other people\'s keyboard.  In fact, there are quite a few students who seem particularly slow with the PC. It\'s either they do not have much contact with a PC, they can\'t hear my instructions, or they choose not to hear my instructions. Just some that I need to keep an eye out for. May actually have to try shifting them out of their old seats to somewhere where they can see what I\'m showing the class.  OK time to go Aikido. WOOHOO!!   '),(26224,979795,'08,March,2004','2004-03-08 00:00:00','    Went to the Career\'s Exhibition yesterday. Went straight to Ministry of Home Affairs sections and asked around. Basically still no news on SO recruitment. I was told they recruit JOs every month. Haiz, guess I\'ll just keep asking whenever fairs like these pop up. Went with Tang actually. After that we went over to the libray @ Esplanade. It was one of the \'noisiest\' library I\'ve been actually. Coz being some kinda cultural / music hub, they had a malay dance/performance that afternoon. We hung around discussing about some of the politics, rumours, and different styles of Aikido. Even discussed some of our own perception and understanding of Ki, etc. Tang knows a lot. But what I like most about his training is that he kinda learnt by feeling rather than what I\'m going thru now. I feel like I need to be told what I should be feeling for.  Not having a really great start to the week actually. Had 1 class followed by 1 meeting with the Pri 4 teachers. I didn\'t hold this one very well as I\'m not sure what to expect and how to go about doing this. But the teachers were really nice and I think they were one of the easiest to speak to.  Next were the Pri 6s\' teachers. Tracey was around to help. Really depended on her a lot throughout the meeting. She was really able to push for the eLearning. At elast everything seems to be scheduled rather than get pushed back.  I had my first Pri 3 class just now and they were a difficult bunch. They\'ve reached the age where they can understand you but simply don\'t just because they think that they know better. Naturally a lof of double work was done. Rule of thumb I learnt. Treat them like Pri 1s until they prove themselves otherwise. Hmm ... maybe I should apply this rule to all of them instead.  After that I had my meeting with the Pri 4 teachers. Only met 4 of them. 1st mistake, I had overlooked informing them this was a meeting with ALL teachers of that level. 2nd mistake, I caught most of them at really bad timings. 3rd mistake, I didn\'t get my stuff ready before the meeting. Basically this was one of the worst meeting I had. Can feel the teachers actually trying to shove work to me. As much as I\'d really like to help them with their stuff. It most probably won\'t be possible because I\'m dealing with every level. I really wished Tracey was around to ofer some help then. I\'m sure some would give her some \'mian zi\'. In the end, the meeting turned out to be a waste of everyone\'s time. Nothing concrete was arranged. It was more of \'Let\'s wait till 2 wks later after sch reopens and we\'ll see what and how we can do this.\' Sigh, seriously lor. I can tell some of them quite resistant to the new system. At least if Tracey was around, she\'d know what is considered reasonable and how to compromise. A lot of things I don\'t know about the teachers and education sys in the sch now, so I\'m unable to \'insist\' on certain things. I guess I just need to make more friends with them lah. Bo pianz.  Got a class now in fact. P1s, the adorable bunch. Haha...  Meeting cousins (Shouming and Shouwen) for dinner later. Basically just a meet before the latter goes into NS and give him a treat for his As.   '),(26225,979795,'06,March,2004','2004-03-06 00:00:00','    Just checked out the  urlLink National Geographic Asia website . I\'m quite interested to catch its Aikido part of the martial art documentary series on the final Fri of Mar. Looks like it might actually be Yoshimitsu Yamada Shihan in the show. I just wish there\'s someway I can video it down. Argh ... damn VCR spoilt long ago in 1999.   '),(26226,979795,'05,March,2004','2004-03-05 00:00:00','    Went Siglap on Thurs. Training was good. Richard teaches quite well from a point of view. I dunno why I just enjoy training with the beginners there. In this dojo I\'m able to functiona  little better because I really feel that every student there see each other as equals regardless of grade. It\'s good.  Well except maybe Leon. The kid is still young. Typical thing he does is not maintaining posture and having a slack training attitude. Even dojo discipline is qutie bad. But I think given his age, that\'s still understandable. I dun see him putting much effort when it comes to laying mats. Always comes early, but takes donkey time to put on his gi, and like to walk around the dojo while we lay mats. Ocasionally lays one or two pieces then stops to chit chat and tok kok. The two white belts, Ming Lee and Jeremy, at least lay mats while they chit chat. I like their beginner\'s attitude, open to learning no matter whether they can do the thing or not. They have no qualms about laying the mats themselves even. I had to suan Leon a bit that night. I said how come he\'s always walking around and when people lay ten mats he only lays one. Ar well, still young lah. I leave it to him.  Had a long tiring day today. Took 2 classes of Pri 2s this morning. After that there was a meeting with the HoDs of every subject. My manageress came down too. Basically a lot of discussion was done about how they wanted eLearning to go with regards to their dept. We got some really enthusiastic teachers in this school and also soem slack ones (naturally lah). Well, dun wanna comment too much but I didn\'t get to really rest till like after 3pm. By then I was starving coz I only had a drink of soya bean at 11+. Took some biscuits from Daphne and bought some mashed potato and bread from the canteen. Cheap lah.  Oh yea, some cute thing happened in school too. All the Pri 1s basically know who am I since I\'ve already taken their class. I was popping in the library to look for my HoD when a bunch of Pri 1s who were there during their recess saw me. Here\'s what happened:  Me (Coming up to library): (To Librarian) Lathifa, did you see Tracey around? Lathifa: She was here just now. Me: Oh never mind then, maybe I\'ll come back later. Me: (Turns around and see bunch of kids jus behind me. Dun quite recognise them coz I\'ve only met every class once so far.) ??... Kids: (Murmur murmur murmur...) 1 Kid: (Trying to recall my name) Ar ... mister ... ar mister ... ya Mr Tan. Me: Yes? Rest of kids: Ya ya ... it\'s Mr Tan. Me: Yes? Why? (as I tried making my way thru them towards the lib\'s exit). OK come let me pass first. Kids: (Crowd even closer. 1 girl to my left saw my hand just in front of her and just happily grab it.) Me: ???? ... yes? why? (While still making my way towards the Librarian and the door.)  I need to go meeting now OK? Me: (Amused) ... (walks over to Librarian and looks to her for help.) Lathifa: They like you lah. Me: Come, need to let go. Mr Tan needs to go for a meeting now OK? I\'ll come back later. Kid (with my hand): (Hugs hand around her neck and shakes her head. - Think the aircon was quite cold for the kids and my hand was like quite warm - I think that\'s why she didn\'t wanna let go lah.) Lathifa: OKOK, Mr Tan got meeting lah. Kids: (Start to disperse. Girl lets go of my hand also.)  Here\'s when the recess bell rang. (Saved by the bell.)  Me: OK... byebye (still feeling damn amused as I made my way back to the conference room). Kids: Byebyyyyyyye (while rushing around like ants).  Hahahaha ... cute hor.  Skipped training at Yuishinkai today because after the long schedule today I didn\'t feel mentally competent to be able to do anything besides dropping dead on my bed tonight. Tried my best to catch the documentary on National Geographic about Muay Thai. But seriously, was way to taxed to do so.  Shouwen (cousin) got his As today. B, C, D. Dunno for what subjects. His GP was a B3 I think. Anyway, the male cousins will be meeting for dinner at Suntec next Mon. Gonna be eating at my fren\'s place - Swiss Culture ... again. Haha ... can\'t help it lah, we like the food there.   '),(26227,979795,'03,March,2004','2004-03-03 00:00:00','    Right! Something I heard from the forum.  ----------------------- National Geographic Channel will be airing a series on martial arts in March 2004 - schedule as follows :  10 pm Fri 5 March Muay Thia 10 pm Fri 12 March Capoeira 10 pm Fri 19 March Karate 10 pmFri 26 March Aikido  don\'t know yet if the series will run into April.  programs will be repeated 10 am on the following saturdays and mondays - times vary.  Colin  for the program schedule check out ngcasia.com and do a search for \'martial arts\' -----------------------   '),(26228,979795,'02,March,2004','2004-03-02 00:00:00','    OK let me just update what\'s been happening after Sun.  Mon I took my first classes. All Pri 1s. The first class was more hyper and I felt more lost on how I to \'teach\' them. The next class was slightly better. However, being Pri 1s, it kinda took me 1 hr just teaching them how to login to the sys. That\'s normal lah since they\'re still so young. Tues, I took 1 teachers training in the morning and 2 Pri 1 classes again. Teachers training was slow but I did gain some insight on how I can conduct classes/workshops with them in future. Same thing for Pri 1s. But I think I\'m either getting faster, or the classes are getting more obedient or are generally just more quick in learning because they seem to be going 1 step further than each of the previous classes. Anyway, just to share some cute moments in my classes.  In my first class. Bunch of Pri 1 kids raises hands and calling me: \'Miss\'ter Tan, Miss Tan. Me: ... Me (to kids): Mr Tan  Trying to keep the class quiet. Me: Class be quiet. Girl: Keep quiet lah keep quiet lah... Me: (thinking to self) *She\'s making more noise than any of the others.*  Trying to \'por\' them a bit just to get their attention. Me: Is MO1 an obedient class? Class: YEEEES!!!  After giving instructions. Me: Who has any problems raise your hand. Girl: (Raises hand) Me: (Walks over) Yes, what\'s your problem? Girl: No, I finished already. Me: ...  Something I heard last Sat on the bus to school/work. We were passing the Sheares Bridge over the Kallang River where you can see some nice condos on the river front. Daughter (Pri school age) to Father: Actually I want to stay here one (Points to condos). Father: Ya, I want to stay here also. Daughter: Next time I will built a house here. Father: A house? Why not a bank? Me (cannot tahan liao): (Smirking and controlling laughter. Exchanged smiles with the father.)  Kids say the darnest things...  Tonight\'s Aikido class was kinda disappointing. I didn\'t think I managed to pick up much or remember enough of it. Worse still my sitting kokyuho seems to be worse after switching partners. Seriously man, need some serious help on this. Managed to remember to bring down my grading booklet for Sensei to sign my 5th kyu grading after such a looooong while. When I renewed my membership, I think I forgot to tell him and he kinda just missed it as well. Also finally remembered to bring down Hattori Sensei\'s old video clips on CD to pass to Han via Sensei. Again something I\'ve been owing for mths now. These days memory seriously not too good leh.  At the end of practice, Sensei said the Aikido trip to Hanoi has been confirmed. It\'ll be from 15-18th Apr and it costs S$900+. More expensive than Bangkok. Haiz, bopians. Can\'t go anywhere this year I think.  At the end of class I asked Sensei if it\'s OK for me to train 1 day each week with Yuishinkai. He thought it over for a few seconds and simply said that sounded OK. I\'m quite glad really, hope I can learn something from both sides. I\'ll try to blog what I can whenever I can after each training.   '),(26229,979795,'27,April,2004','2004-04-27 00:00:00','    Work was ok. Updated my HoD yesterday about some stages of my work and I still felt like I was behind a lot of things I wanted to accomplish. But I got a reply from her saying that I was \"doing a great job\". So encouraging but also hope that I can do better.  Met up with Robin today. Like I was telling Daphne, coz on several occasions when I tried to meet with Robin, a lot of things would go wrong, I was kinda expecting the same thing to happen today. Things were ok lah, he was there waiting already when I got to Dhoby Ghaut MRT. The evening and dinner was quite uneventful. We couldn\'t quite agree on a movie so we juz settled for makan instead. He\'d already seen 2 of the 3 shows I was hoping to catch. Oh well... I had to admit, things were quite boring over dinner coz I didn\'t really know what to say and all that lah. But after dinner we went Kinokuniya at Taka and browsed through some books for awhile. After that we went to grab coffee (well, tea for him lah) at Coffee Bean. There we had a good chat about family, life, philosophy of living, etc. Basically a superficial start in the evening ended with a good chat at night.  When I got home, I realised Zatoichi was still showing at Cineleisure... only there in fact. Gotta catch it latest this weekend before they stop showing it. Wasted though, could\'ve caught it tonight.  Oh yar 1 more thing, traded in my Samsung T100 for the new Samsung E700. Quite cool lor. Didn\'t think I\'d like it but I really do. Need to take good care of it. My new toy ... hehe ;P   '),(26230,979795,'24,April,2004','2004-04-24 00:00:00','    Those girls can really ji siao me.  On top of always pairing Daphne and I together just because we lunch often and work in the Labs, they were a but more thick skinned than usual yesterday.  After lunch I entered the lab to find Laura (a Mandarin teacher) with her class. They were typing words in Mandarin inside Winword. So I just hung around coz I had nowhere better else to go. As the lesson ended, some girls approached me saying there was something wrong with their elearning account so I tried to settle it for them asap. As the other students were making their way out of the class, some girl shouted to Laura in Mandarin \"Huang lao4 shi1! XXXX (god knows who) likes Mr Tan!!\" The other girls giggled here and there and Laura also found it amusing. I would probably eng siu them a bit by some kind of boliao reply if I wasn\'t in the midst of settling the elearning accounts on the spot. I basically ignored them, didn\'t even bothered giving them a look. I vaguely heard Laura saying something to me, and though I didn\'t mean to, I ignored her comment as well. (Oops, hope she wasn\'t speaking to me directly or directing a question at me then.) Anyway I heard someone saying \"Ahhhhaha, Mr Tan paiseh lah.\" And I still ignored everyone else in class (except the 2-3 students waiting for me to settle their accounts). After I finished my task on the PC, I looked up to realise only 2 students were waiting. The other girls with account problems had already left. Idiots ... what did they expect me to do when they haven\'t given me enough details. Anyway I didn\'t manage to solve everyone\'s problem then and they deserve it too. I\'m not their servant where you just tell me to do it and you walk off like that.  At this time I also just realised that Laura and the rest of the class had left sometime ago. Oh well, like I said, if I\'m concentrating real hard, I can be dead to the world and just ignore everything else. Just hope Laura didn\'t think I was brushing off her comments/questions then, that\'s all.  Ar ... got a birthday present from Juliana and Asuka from Birmingham. Quite touched. It was a shirt and knowing how expensive things are in UK, I think the shirt easily cost them S$80-100+, and that\'s excluding mailing the whole thing over. Geez ... gotta send them my thanks.  Will be quite a busy week ahead. But I promised myself to work on the courseware which I need to submit back to my company on 30th Apr. To make it worse, I need to finish up my mthly report, wkly report, and misc shitz as well. Fish lah... tomorrow still need to get a new phone coz my line is under my sis\'s name and she is free only tomorrow afternoon. Oh boy. Think I\'ll have a hard time breathing. Only silver lining is that my parents will be leaving dunno which day for a tour in China or Korea, I also can\'t remember. Woohoo! Bo Zeng Hu (no governing authority) for 10 days.   '),(26231,979795,'18,April,2004','2004-04-18 00:00:00','    So much to blog about. Past few days saw a lotta things, felt a lotta things as I passed people, listened to people, heard of events, and experiences of my own. Was reading Guilian\'s last 3 posts on her OD and I gotta admit. Since lastnight I was in the same mood as her. Dunno why, just feeling damn bored and lifeless, slightly pekchek over nothing in particular. Blasting some music now. Most friends know that though I dun chiong, I enjoy a bit of noisy music like remixes and all that. Nothing completely techno or heavy metal type (can\'t stand those at all) ... Eargh!!   Aikido @ Siglap (Thurs night)   Something I was telling Gary on Thurs night after training at Siglap. Leon was his usual joker self during training. Was quite rude and not serious at all. Usually I didn\'t find that to be a problem. But that evening I felt a little indignant about things. Even Richard got more serious with the class in general. Made us sit down and gave us a lecture on etiquette and attitude. These days I found myself getting impatient with Leon also. And the effect of it is that I may tend to tell him too much what he\'s doing wrong and he\'s feeling so sianz that he pays even less attention to what he\'s doing. I might have to watch myself and converse less with him. He likes to ask and ask but never quite listens when he is being answered. Oh what the heck lah, I\'m not his teacher anyway. Shouldn\'t even be teaching him anyway. Weird thing is I never face that much of a problem with the other 2 beginners. So I also dunno, is it Leon, me, or wat? What I remembered most distinct about training with Leon is that we were training halfway and he just walked over to another pair next to us to help \'demo\' with one of the guys. I was like just standing there facing the wall where my partner was standing a few seconds ago. Damn rude. Richard saw it and that\'s when he told him to get back and not just walk off like that. Leon\'s attitude on and off the mat is the same, slacker attitude and not really putting much serious effort in anything. Can tell even when he is laying the mats. When he lays just 1 pc of mat, he literally sits down and starts fitting in the jigsaw mat. We always lay our mats in a ready position because it only lasts like 15-20 seconds, so there has never been a need to sit down. Sitting down and getting up again waste more energy and nobody I know does that. Only Leon. You can tell he is maintaining that \'after I sit down I don\'t want to get up again\' mindset. Same thing when he does his Aikido. After being pinned to the ground, he stays there so long before picking himself up. Bad thing is 1 new brown belt girl (prob from kids class too) is kinda taking for granted that mat laying is a guy\'s responsibility. I don\'t see her helping at all though she gets to class a bit early. As a junior I don\'t think I should say anything to her directly, but last Thurs was seriously tempted. Why? Because I really don\'t want new blood to come into the dojo, see a senior not doing anything, and just follow suit. Being in a Pri sch now, I realise a lot of thing depends on a good foundation. We are literally cultivating the next generation of people (whether in school, society or in our dojo) and if they do not learn things the right way at the very start, it will be difficult to change them later on. But at the same time I\'m also worried of the effect that may arise, what if for eg, this brown belter gets sian of being told to lay mats each time, may feel like I\'m picking on her, then just keep arriving late to siam the mats? Also got a negative effect. There\'s no 100% method I guess. Dunnoz, may just speak up coming week when I see Leon and her again. Leon for one needs the extra work. He\'s been an orange belt for 3 yrs and I think it only shows one thing, that he has been stagnant for the past 3 yrs coz his mind is not moving forward. I\'ve heard of pple avoiding taking the grading tests but never of the case that Sensei keeps one at his current belt level, probably because of his wrong mentality and attitude.   Friends   Friday came. Was really looking for ward to meeting some friends from Uni that evening. Was great seeing some familiar faces again like Begh, Lynette, Darren. Dinner itself was nothing special, just KFC. A bit quiet though we did talk quite a bit of crap. Usual lah, with Darren and me around, you can\'t quite expect too much serious talk. After that we went for coffee at Raffles Link. Nice place, but too bad it seemed a wee bit more enclosed. I still prefer Spinelli at Heeren but these days I figured it\'s just way too hot to sit anywhere without aircon. After that we left about 10-ish. Got home and I SMSed everyone that it was nice to meet up after so long. Got some SMSs back. But this Begh ar, said \"Since when did I get so sentimental.\" Kinda got me thinking. I must be mellowing with old age. Harhar...  Just brought to mind how much I haven\'t kept in contact with my friends. It was only recently that I sent a short email out to my Poly classmates and updated what I knew of each other. Was glad that I got some mail back from everyone as well. Think I\'ve kinda become the unofficial database administrator already. Probably arrange some meet up after James\'s wedding dinner. Despite these people, I\'ve also failed to update groups from NS, people I knew from Aikido in UK, Birmingham Uni (there\'s a bigger group besides those I met up on Fri), my English housemates from 1st year Uni, and some really close friends from here and there. Funny how convenient you just let these friends slip out of view as time passes. With some lucky ones, you\'ll still be the same no matter how infrequent you meet up with each other. But with many, you\'ll just loose touch quickly.   Work   At work also some very small matters going on here and there. Overall I\'m glad teachers aren\'t completely shunning elearning out despite their already busy schedule. But I think I need to do more as well. Gotta start prioritising my work. On the whole, I get quite a chunk of quizzes from teachers to convert into elearning during the past 1 wk plus. A bit overloaded all of a sudden. But they seem to be clearing up. Had to turn back 2 quizzes from teachers until my load is relieved a bit. I won\'t say it\'s unbreathable lah, but just that it seems like everyone is fighting fire at the last minute so deadlines are all packed in a bit close that\'s all. A bit too much all of a sudden that\'s all. Daphne was feeling down yesterday also. Shan\'t say too much except that it\'s all work related. I dunnoz, I dun feel that kinda pressure she does. In fact, any pressure I feel is mostly temporary and I never let them get to me for long. Maybe coz I\'m more bochup lah, but it\'s only coz I know that if I dwell too much on the \'here and now\' I\'ll just get stuck and stop moving on.  Plus the fact that the kids are constantly making fun when they see the both of us together. Dunnoz, again I\'m not so bothered coz they seem to direct this more at her than me. I don\'t normally get affected by these but Daph sure seems to be. Seriously lor, maybe I need to step in to make them learn a bit of respect. Dun wanna use army style on them lah but last Fri itself I just did (again). 1 group of students came in to use the lab PCs and Daphne asked one of them (anyone of them) to go check on whether another lab is empty so that she can bring them over there for the lesson. Immediately one by one they just went \"Chop! Cannot call me!\" then another went \"Chop!\" followed by another and another. Most friends will know lah, I cannot stand that kind of attitude one so my response to that was \"AR! You chop somemore! Chop somemore lah, I make sure ALL of you go to the lab and come back here together instead of just one of you!\" They still proceeded to ignore my statement and went on \"Chop! Chop! Chop!\" So I said, \"OK I count to 10. 1 of you better go or else it\'s all of you. 1, 2, 3, 4 ...\" And one kid (group leader apparently) decided to go. Really man, army style. Haiz ...   Home   Getting exceeding irritated with small little things at home though I try very hard not to be. For one thing, my parents get really cynical at times. Eg if they see something on the news, they say things like \"Ya they so stupid\" or \"These people deserve it.\" Basically really machiam Luo4 Jing3 Xia4 Shi2 type of comments. These days rather than comment too much I just step aay and go back to my room. Think it\'s all I can do, by removing myself out of the equation, nothing will happen. Another thing is my sis. She is so ji4 jiao4 when it comes to some small things. For instance, just eg. I was the last to come back home for dinner. She was in the midst of finishing her last remnants of food when I placed a plate in the sink (for washing). While I was finishing my soup, she washer her own bowl and can\'t even help to wash a bit of the stuff in the sink. Seriously I dare say I never consciously leave thigns in the sink for her to wash. If I do, it\'s only because I\'ve already washed the first batch of stuff with soap and the soap on the sponge has gotten too diluted for another batch. I figured since she\'ll be washing after me, she can clear the next batch. Even so, I try to maximise the soap by leaving her very little washing. She\'s really taking things for granted liao. Even though things are right in front of her she can siam it like it doesn\'t exist. Really, if you want me to be frank, I have to say that there\'s a very good reason why she is still single. She\'ll be a horrible wife and I would pay my deepest respect to the man out there who can change her someday. OK, make no mistake. I\'m not perfect either, but I think I\'m not as particular about these things like she does. In fact my whole household seems to be that way. My parents also get this idea I\'m very particular about small things and keep asking me before they do this or that to the extent it\'s irritating. I can only say that it\'s because they themselves are very particular people so they assume their son to be the same. I\'m really really irritated at the way things are.   People in General   My conversation during lunch at BK with Daphne still rings clearly in my mind. People in Singapore gets damn haolian sometimes. Arrogance is what I call it. Just walking out of the fastfood they can not even bother to close the door again and let the aircon out like that. Geez. I\'m really worried about the next generation of people when the current ones show these kind of examples. What type of people are we trying to cultivate. Being in school everyday I sometimes look up to what teachers are doing. MOE\'s slogan \"Moulding the Future\" holds true. But how many of them really believe in that when they took up NIE\'s offer/contract? Most just do it because of the economic downturn to get a job. What\'s the point in that? And S\'pore govt itself is still way too result/academic oriented to get down and dirty during the filtration process to filter out un-committed people. Same thing I\'m feeling for the SPF. I\'m aware that uniform services sees a huge group of people leaving each year, but they only have themselves to blame for taking in the uncommitted people in the first place. Geez... really, we can see how a person or people in general are from the little things they do. Do you condemn someone just because of a sudden outburst? Natural for humans to be biased based on a first impression. But open outburst like that are not accurate. The best way to judge a person I think is if you take note of their behavior during the other times. When a person gets angry, it\'s because something happened and clouded their usual emotions instead. (Of course, those who have frequent outbursts ALL the time, definitely have a problem there lah.) We gotta remember not to let their emotions cloud us the same way it does to them. It only takes a celebrity 1 misbehaviour in public outside to destroy his/her reputation. But I think that can be so unfair to them. They are human afterall.   Movies   Really felt so bored lastnight I\'m considering catching a show alone if I have to. Never tried that before. Anyway I was flipping the papers and came across some shows on my list. Zatoichi (some samurai film on a blind swordsman), Hellboy and Starsky & Hutch (comedy), just to name a few.\" Will see how lah...   Aikido @ ARCC   Was   late for training yesterday. Took 197 all the way from Parkway to Jurong interchange. Really roundabout journey. Took me 2 hrs to get there. I was late of course. And I didn\'t get anything right. We did static exercises/techniques for a long while but the moment I tried out non-static ones, all the basics went out the window - from moving hips first to simple arm and foot movements. Was quite disappointed with my practice yesterday. Only thing I really took with me, was the fact that I will never take 197 to Jurong East from Parkway anymore. But Frankie Sensei has a very good way of breaking down the exercises and techniques. I\'ve decided to start doing their exercises more regularly at home every few days to familarise myself with the basics and fundamentals. Even just how to hold the bokken is like so difficult for me to grasp. Everytime there\'s the statement \"extend ki\" I\'d just go \"....\" Heh ... I\'m only glad I\'m wearing my white belt during training instead of blue.  Beginner\'s mind, beginner\'s mind, beginner\'s mind, beginner\'s mind, beginner\'s mind ....................   Weather   Weather has been soooooo daaaaaaamn HOT! I have trouble sleeping everynight. I\'m perspiring like a pig at home every night. Luckily I just bought 2 cans of prickly heat powder. These days I put some before going to work. When I get home I powder myself with so much powder after my bath, that it gets ultra cold for a good 5 mins or so. But at least it helps prevent me from feeling all sitcky and yucky at home. MMmmmm .... prickly heat powder. Again, another army must have item.   '),(26232,979795,'11,April,2004','2004-04-11 00:00:00','    Last 2 day\'s exhibition at Expo is finally over.  Sat\'s was quite good. Syed and Richard came down for it together with students from Pillars and Siglap. I thought everything seemed really well arranged for the day even though it was more or less impromptu. Lucky Richard was quite soft, and I managed some of the more decent ukemi for him. Syed again went throught his \'Seagal-like\' randori and techniques. \'Eastern Seagal\' was the term used hahaha. Thought the weather wasn\'t exactly forgiving, I thought the atmosphere was generally pretty relaxed during training.  Today\'s was conducted by Sensei himself and Alvin came down as well. More senior students turned up here and I really wanted to concentrate more on training instead. Sunny was around, and of course I didn\'t dare commit myself to most of his techniques. But that\'s no reason to stop nor, hope someday I can take more decent ukemi and his techniques.  Overall, I enjoyed yesterday\'s exhibition a bit more. Can\'t explain it, it just felt more like the usual class I\'m accustomed to. Today, well I got the feeling that we were a lot more conscious of this as an exhibition than went through it like a normal class. Again, it was seriously warm and even Alvin didn\'t seem like his usual self. He looked like he was sweating a lot more than usual. Syed dropped by but didn\'t join the mat.\'  Didn\'t manage to stay all the way to help with the mats coz Tang was waiting for me. After that, told Sensei sorry I couldn\'t stay (coz we loaned the mats to a Hanmudo exhibition after us and could only keep mats after them) and went to Bedok with Tang. We grabbed a bite there and talked for hours over Aikido. Dunnoz lah, we tend to discuss things like that. One thing about Tang is that he\'s as real on the mat as he is off the mat. So I learn a lot of things from him in terms of attitude and all that.  Anyway, can\'t wait to go back to regular training class soon. I\'ve even planned to train up to 5 days a week when my parents go off to China at the end of the month for a holiday. Which reminds me, I better schedule my leave early to let my HoD approve soo it won\'t interfere with work then.   '),(26233,979795,'09,April,2004','2004-04-09 00:00:00','    Last night\'s training was OK. Syed did pop by the dojo but not to train, he only stayed awhile and then he left. Think the rest of the guys in Siglap will be popping by Expo on Sat. I guess if they are doing so, I should do the same. Emailed Frankie Sensei to inform him that I may not be going ARCC this Sat and I might pop by the Expo exhibition to help out Freddy Sensei & AFS instead. But that also means I\'ll be missing my weapons class this Sat. Drat!  Had a stroke of promising review from Richard lastnight, he said I had an \"advantage\" because I kept training with beginners both in Siglap and BV, and that he sees me improving. That\'s one thing about some of the slightly senior belts, they sometimes tend to siam white belters. I\'d personally prefer to take a white belter first, and then rotate partners later on in the class if only I can be sure the white belter I\'m leaving behind is able to get a decent partner. Hmm ... not trying to inflate my ego but I also pondered, what area did I improve on? I did feel that last night\'s training was smooth, I didn\'t get myself lock down by a stiff uke or anything. I dunnoz also. My uke is only into his first month of training and I also think he has improved. He\'s beginning to know how to follow nage\'s movement in a certain way. When we did Shihonage, he was even able to tell me that he was falling down when I wasn\'t even applying the lock and pulling him to the ground. I was surprised because he took the words from right out of my mouth. I really wanted to tell him not to anticipate what I was about to do. I\'m sure he doesn\'t quite realise it, but he does a pretty decent job at maintaining his hands in front of him when he did shihonage as well as the initial hand movement of diverting uke\'s ki relatively naturally. I had to follow follow follow and I found myself rising up and down because of his height.Not on purpose but because he was actually controlling me.  But things change when I pair with a more senior belt. Maybe that means I can\'t blend properly as an uke? Maybe it means my ukemi still got some major problem I need to concentrate on. Food for thought, which aspect of Aikido did I improve on so far and which area is still lagging far behind? Hmmmmm...  Anyway, I know students from Pillars will be down on Sat as well, probably gonna see a lot of slamming again hehe...  Tonight went on to watch Passion of Christ with my aunts, uncles and family. Originally thought it was a family / relatives event. But when I got there, the freaking number of crowd entering the cinema made me realise it was a church which booked the whole cinema. Look, I have nothing against Christians / Christianity but I honestly can\'t stand events like these. Most of my freinds are aware I dislike crowds as well so I really felt doubly irritated. To be fair, the show was really well made. I didn\'t realise it wasn\'t an English film until when the show started. Very interesting. Movie was enjoyable just not the atmosphere. Sure, I realised it was Good Friday lah. I guess I should\'ve expected something like that. But it only strengthened my conviction that I really don\'t wanna hang around a Christian crowd. Note that I said \'crowd\', not people. Christian people, I have no problem with. Just not like such a huge crowd kinda size. Worse still, at the end of the show they actually turned off the credits and started to sing 1-2 Christian songs. Dammit! I\'m the type who would rather take a look at the credits at the end of the show. People who work in the background are often taken for granted. Sometimes the least we can do is finish watching the credits - especially if the movie was a great one. Was basically showing my black face throughout and I hardly spoke to anyone of my relatives and family from the moment I entered Cineleisure (not the cinema yet), till the journey to Suntec for dinner. Think they could tell also that\'s why they basically left me alone as well. I\'m feeling guilty about that but sorry lah, I show my mood quite frankly around relatives. Not that I\'m bratty, just that I feel that sometimes no need to put on a mask among people you feel close to. Going to Suntec, I complained to Daphne via SMS about the whole thing hahaha ... Ms PA indeed, always there to hear me complain about stuff.  Luckily I had a great dinner at Swiss culture. (My sis\'s treat to my family coz she just got her promotion/pay increment.) My mood was basically turned around because of the dinner. Good lah, I wanna look back and remember the show for what it\'s really good at, not the environment of the cinema. That\'s giving the film due credit where it really deserves.   '),(26234,979795,'07,April,2004','2004-04-07 00:00:00','    Aikido was good today. Dunno why but things felt smooth and I actually found myself coming to terms with some kind of understanding. We only did shihonage, iriminage and kotegaeshi. Sensei asked for volunteers for the exhibition at Expo this coming Sat and Sun. I\'m not sure I\'ll be able to attend Sat because of training at Ayer Raja CC. But I may go for it if Sensei really need people there. See how lor, but seriously I really want to go for my training at ARCC.  Yay parents going China end of this month for about 10 days. I think for that week or so, I\'ll plan to go down to BB East on Mon. Haven\'t exactly attended Alvin\'s classes for ages liao.   '),(26235,979795,'04,April,2004','2004-04-04 00:00:00','    The kids in school haven\'t been giving me too much trouble. The usual lah. I\'ve found 2 classes who really have a lot of character. They may appear somewhat \'arrogant\' the first few seconds you see them. But I think they display a level of confidence that makes them so at ease with being around people they\'ve only met a few times (namely me lah). Class JO5, the group of girls who were screaming at the ghostly website I mentioned in my last post, and AM6, my HoD\'s form class. These two classes are really really alike that I find it quite intriguing, considering the fact ehy have different form teachers.  Just remembered that James called me sometime ago. He\'s inviting me to his wedding dinner on 7 May. But I think he probably won\'t be inviting most of our classmates down because he don\'t feel like he knows everyone that well. But I think he has to bear a part of that responsibility because he\'s been flying us so many aeroplane the past few years. Shuhua\'s gonna be invited so I guess that\'s one familar face there, whew. I think Yvonne and husband will be invited as well.  Failed the company test I took about 2 wks ago. When I went back yesterday for re-training, I finally understood why. Seems like I failed to attach the file I was working on and I basically failed the practical part. Geez man, what a waste. I coulda passed. It\'s ok doing the things again because I didn\'t find them to be too much of a problem. But I sure felt like I made a wasted trip down. Oh man...  Training at Ayer Raja CC was nice. I realised my mentality is changing a little. Originally I wanted to train with Yuishinkai hoping that it will improve my Aikido. But I found myself feeling twice as stupid, twice as lost, twice as lousy. But I still enjoyed the experience because it opened up another area to me that I\'ve not felt before. Also made me realise that I may need to work twice as hard to reach the standard I hope to arrive at someday. Gotta gambatte in this area.  Spoke to Marcus (Poly classmate) lastnight and found out he\'s been posted onboard one of Singapore\'s submarine. Of course lah, considering he\'s a submariner. We talked about the usual blah blah about people to people relationship at work which you can find almost everywhere. Shared with him what I\'ve been going through in my job since Feb. Was quite interesting and enjoyable having such a long conversation again. Found out Ivy\'s having another baby coming end of Aug. Woohoo ... damn happy for the both of them. Haha... I asked hime whether it was planned or accidental and he said they did plan for it but didn\'t expect the next one to come so early only.  A bit of reflection since my last post. Realised I\'ve not been keeping to much in touch with my friends and old classmates. Feeling kinda guilty over this also. Spoke with Bin Soon via MSN also few weeks ago and realised some of our friends have switched companies. Think he\'s doing fine as well. Seriously man, hope to meet with them someday but I doubt this month is gonna be good for me. Think I better update our usual bunch via email soon.  On ICQ, I asked someone how was her day and I got a statement asking me to check out her blog. I\'m not trying to criticise or direct this at anyone but I\'m just trying to recollect and understand the thoughts that came to my mind after that. Sigh, I remember telling someone (could it be Gary? Lynette? Roy?) to go read my blog a long time ago, think it was around the time when I started blogging. Idea was just to get people to read my blog so I was quite excited about \'promoting\' its existence. But I realised that that\'s so lame. If I ever direct someone I know to go read my blog rather than take a bit of effort to tell them about it, it only goes to show the person I am, the way I treat them, or value them. Nonono... I absolutely refuse to do so from this day forth. If I can\'t even bother talking to them (via phone or ICQ), why would/should they even bother talking to me, or respond in kind? Everything goes both ways. I gotta start treating people the way I want them to treat me.   '),(26236,979795,'24,May,2004','2004-05-24 00:00:00','    I love this.... really. It\'s one of the principles I live by. I only hope I can apply it in almost every situation I experience. Don\'t know if anyone is able to view the chinese characters in this entry.  &nbsp;   &#19990;&#19978;&#30340;&#20107;  ,    &#26080;&#21892;&#20134;&#26080;&#24694;  ,    &#26080;&#23545;&#20134;&#26080;&#38169;  .   &nbsp;  In this world, there is no good and evil, no right and wrong.  &nbsp;  &nbsp;   &#25152;&#35859;&#30340;&#21892;&#24694;&#23545;&#38169;  ,    &#21482;&#26159;&#23384;&#20046;&#19968;&#24515;  .   &nbsp;  What we know to be good and evil, exists only in one\'s state of mind.  &nbsp;  &nbsp;   &#20320;&#20043;&#25152;&#20197;&#22914;&#27492;&#22256;&#33510;  ,    &#30342;&#22240;&#20320;&#22826;&#37325;&#35270;&#20182;&#20154;&#22914;&#20309;&#30340;&#30475;&#24453;&#20320;  .   &nbsp;  The reason that you feel troubled, is because you view others\' opinions of you to be of great importance.  &nbsp;  &nbsp;   &#20854;&#23454;&#20570;&#20154;&#21482;&#35201;&#38382;&#24515;&#26080;&#24871;&#23601;&#20250;&#24555;&#20048;&#20102;  .    One will be able to be happy if one is able to face his own conscience.  &nbsp;  &nbsp;   &#24754;&#20134;&#22312;&#25105;&#24515;  ,    &#21892;&#20134;&#22312;&#25105;&#24515;  ,    &#22806;&#29289;&#19982;&#25105;&#20309;&#20248;&#21705;  .   &nbsp;  Sadness within me, goodness within me, external factors has nothing to do with me.  &nbsp;  &nbsp;   &#19968;&#20999;&#29575;&#24615;&#32780;&#20026;  ,    &#20294;&#27714;&#24515;&#20043;&#25152;&#23433;  .   &nbsp;  Act with frankness, but ask only for peace of mind.    Yes I know it sounds wee bit profound. Sorry, not meant to be. I can only blame my poor translation skills haha. It makes more sense in chinese.   '),(26237,979795,'19,May,2004','2004-05-19 00:00:00','    Wah, Guilian actually put what I said about her honours into her OD. OK lor, it\'s nice to be appreciated I guess.  Right, have been laggin behind my regular blogs. Simple reason to it, I\'ve been procrasitnating work and getting disorganised. I realised how it\'s affected even my personal life. Not physically but in mindset. My mind don\'t feel totally relaxed even over the weekends, I tend to forget / overlook stuff easily, etc. (oh no! senile liao!)  At least past few days I\'m getting back into things. Haha, suddenly from watching all the Yi Tian Tu Long Ji vcd, I got a craving to play chinese games again. Was replaying my Jin Yong Qun Xia Zuan again. I think it\'s gonna be a long game man. Slowly lah, rushing also no use.  On Aikido, gee, realised I\'ve been mising a lot of trainings at BV past 3 wks. In fact, think it was something like since Freddy Sensei went Vietnam. Alamak, I better decide properly whether I\'ll be training in Pillars or BV. And decide soon I must. At least I\'ve been really consistent at Siglap and ARCC.  I suspect that I\'m beginning to compare things in different dojos and clubs. But even under AFS, I Siglap gives me such a nice feeling compared to BV. Someone once told me, the average joe in BV trains so hard and so much but they don\'t seem to be improving much. Why ar? I have to admit that I\'m kinda inclined to agree. From a personaly point of view, I see myself that way. Maybe that\'s why I really want to jump out of the box a bit to experience something different. Yesterday pia like siao to BV only to oversleep on the bus all the way to the interchange. Missed Aikido lor. Wasted, apparently Tomita Shihan, 8th dan, from Belgium was conducting the class. I didn\'t have the fortune to experience his lesson. Head his class was quite profound, but I\'m interested in these. We don\'t get enough on this aspect I think. Won\'t be able to go BB East for training today (Tomita Shihan again). I never get to BB East on time one. Previous time I rushed like siao there to see them doing kokyu ho as a final exercise. Yestrday I got there in time to help keep some mats only. Must admit I felt quite sian lah.  Never mind lah, take things as they come.  I feel quite happy because Steff from 1st year is visiting S\'pore with 4 friends. They are all English. Will be meeting them later for dinner, but a bit of delimma as to where to take them for dinner. Thinking of giving them a bit of treat if dinner isn\'t too expensive. Daphne was suggesting Don Noodles at Park Mall. Just nice since I\'m meeting the girls at Dhoby Ghaut MRT. This week gonna be quite chor coz I\'m not makaning at home at all until the weekends. ~$~   '),(26238,979795,'09,May,2004','2004-05-09 00:00:00','    Guilian got her results today. She\'s feeling so disheartened because she truly believes that she will be getting 3rd class honours. Makes her doubt herself a little, I believe. She now has a choice whether she should file for graduation (deadline is this Fri) or to continue and get a 3rd class next yr.  This brings back some memories on my thoughts and feelings about getting 3rd class. I remember doing badly after yr 2 (technically speaking it\'s my first yr because I actually skipped 1st yr and entered 2nd yr of the degree programme), so I had to repeat 2 modules. That delayed my studies by 1 yr because in UK, exams comes only at end of academic year.  Read some comments on her OD and I can see the expected comments she\'s probably getting from most of her friends around her. Things like \"No lah, not worth the extra 1 yr lah - for 3rd class\", \"3rd class pay and Bachelor/Merit degree pay no diff lah\", \"wat for\". Problem with us Singaporeans is that we think too realistically. We can\'t take failures man. I think it was DPM Lee Hsien Loong who said \"If you are totally focused on your studies and other preoccupations, then perhaps you are sleeping but not dreaming. It is time for you to wake up and dream.\" Just an extract that was given to me last week (coincidentally).  One thing I can relate this to is Aikido. It\'s like when taking grading tests. It\'s not a paper chase. So what does it matter if it takes us 10yrs to achieve Shodan. Now, if what I want is a black belt, I can get that in 10min. Just pop down to the sports store down the corner and buy one lor. But if I want to learn something from Aikido, it\'s a different story. Sorry, I digressed a bit (it is MY blog afterall wahahaha!). My thoughts on the belt system in Aikido. Being a senior coloured belt don\'t mean you are better than one with a junior coloured belt. It only shows you the length of time, effort and commitment you\'ve put into training so far. I believe results are the same way. But most of us take it as a  \"I\'m better than you\" kind of system. Face it, no matter how good you are, you will never be the best because somewhere along the line there\'s always someone better. The moment you think that you have achieved a certain mark, that\'s when you will stop improving. No point comparing your results with others. If you get 3, there\'s someone wif a 2:2. If you get 2:2, there\'s always one with a 2:1. If you get 2:1, there\'s another with a 1. If you get 1, you think that\'s it? Oh pls, there\'s always someone stronger/better/richer than you in someway that will make you look at him/her and just go \"waaahh!!\" in awe.  We\'re so trapped in the paper chase, results status and all that we forget wat on earth we are doing in Uni. That aside, I don\'t have any delusions about why I studied what I did. For me, after 3 yrs of Computer Studies in Poly, 2yr-3yrs of Comp Sci in Uni, I\'ve arrived to the conclusion that pure hardcore IT work is not for me. Though it took me almost 6 yrs to arrive to this conclusion I still never found it a waste. Where are you rushing that you can\'t afford 1, 2 or even 3 extra yrs of your life? It\'s not how fast you get something that makes it memorable. It\'s the process and how you get it. My mom has a bad habit when we go holidays. She likes to make it as cheap as possible by rushing to as many places as possible, reach the entrance, take a photo to show \"I was here\", if admission is chargeable then leave and go to the next spot. I kept telling her it\'s not reaching the places that count, like in a touring marathon. It\'s enjoying the places you\'ve been. Even if you stay in one spot but you\'re able to truly enjoy/appreciate the really simple things it has to offer. I tell you, you will remember it as one of the most memorable holidays of your life.  The things I picked up those 6 yrs? Life overseas, Aikido, friends. All precious and things that I wouldn\'t trade for anything in the world. I think I deserve everything I have now. If I didn\'t work hard enough for some things, what I get at the end of it is, well ... mine. The good and the bad. Everything is relative. I can\'t just expect the good things and shun the bad right. That\'s not how this reality works. You take the good with the bad. Balance, yin and yang, positive and negative, chaos and order, etc. The only thing I regretted when I kokked up my 2nd yr was that I actually spent such a huge amt of money in UK that yr and not work hard enough. Sure I was guilty, but I had to get over it. Harping on it only holds me back.  So thinking about this, I suppose that explains why I\'m really quite happy with my 3rd class. I\'ve managed to endure through the whole process of things. Graduated too. Sure, during graduation I was the last few person to collect my results. Basically everyone at the convo can tell that I\'m a 3rd class student. But I can tell you, when I see the fellows behind me. I feel elated that I had chosen to complete the whole course when I did. It\'s nothing about ego - not being the worst / being better than some others. I still felt really proud to put on the gown and being present for graduation that day. It was 1 heckuva moment for me. I still stood tall and proud. You have to know and recognise what you have achived in order to do that. Acknowledge your failures, accept them and you will recognise your achivements as well.  I can only hope each one of us really think about what we want. When we know it, we will then know what we need to do next.   '),(26239,979795,'09,May,2004','2004-05-09 00:00:00','    Wah, can\'t believe it man. Haven\'t been updating my blog for a while. These days the frequency of my blogging is getting pretty irregular. I wondered why, think it\'s because I\'m starting to let things get more and more messed up when it comes to prioritising work and play. I guess that\'s what happens, people who manage their time well will always have time for everything. Whereas once I let myself get too disorganised, I will always find that I\'m fighting time all the time.  Actually I did have a lot of things I wanted to blog about. But I may not recall everything now.   Aikido @ Siglap   Siglap CC saw a new student. A Japanese (national I think), by the name of Sakai. He\'s a judo yudansha, but I\'m not aware of which dan. Was told he seems stiff but haven\'t quite seen his training much yet to know what that means exactly. Leon still as slack as ever. Tang came down to watch our training (he couldn\'t join in coz of the indemnity notice everyone need to sign during training) and commented on Richard\'s style. He uses a lot of Ki extension (as Yuishinkai would term it). I went down to Pillars Youth for training on Mon. Very shiok. Just need to get used to being around an all malay training ground. But I have to admit it felt really nice. Everyone was so friendly. In general even the white belts practice a lot more harder and have better ukemi than those I\'ve seen in BV. I\'m seriously thinking of transferring from BV over to Pillars for a few mths. One, coz the distance from work is so much nearer. Two, I feel damn uncomfortable not paying mat fees and training at BV every Tues. At least at Pillars, I\'m paying $5 for each visit. I dunnoz lah, see how lor. Don\'t think I want to stop Siglap/ARCC trainings anytime soon. ARCC maybe far but I always have ample time.  Anyway, next week is going to be weapons class. Gotta remember my bokken then. Realised my bokken seems like it\'s cracking liao. Wah, seriously man, the ki my bokken experiences with Yuishinkai is taking its toll on my poor bokken. Hahah ... never mind lah, can buy a new one if it breaks.   Friendster   Friendster seems to be up again, but still as unreliable as ever. Got a message from someone but when I checked my inbox, I couldn\'t find a thing. Ridiculous lah, think Friendster did something . Anyway I dunnoz, forget it. Was checking through the testimonials in some groups I joined and realised that some of my earlier posts had been deleted. Let\'s not finger-point, and it is their perogative too. I\'m more concerned about how Friendster is turning out. If we are only concerned with keeping the nice testimonials about ourselves, I think you might as well email a nice essay to your friend and ask them to post it on Friendster instead. As Neo once put it, it\'s all about clocking up good testimonials of yourself and that is so besides the point. I guess that\'s a reason why I don\'t log in to friendster much anymore. I do find it kinda pointless these days.   Aikido @ ARCC   Yesterday\'s training at ARCC was good. What always strikes me at the end of class is that I\'ve learnt something about the way warmups relate to our techniques. We don\'t relate warmups to trainings almost 80% of the time. Even when something is said, it\'s only a few sentences. Yuishinkai will have a Ki class where they tell you why warmups are like this, how to relate it to your technique and what you should be feeling during the warmups itself. Slowly when we go on to techniques, they fit in nicely. Of coz applying it is a whole new ball game lah. We were doing the rowing exercise with a jo and I finally understood what I need to do or feel.  Holding the jo horizontally with both my hands with my foot in left stance. I lean my body forward and try to maintain stability still. Note not to overstretch myself. As I lean, my hands (with jo) automatically raises upward a little. As I pus out with my hands, I need to do it in a downward motion and hard enough to feel the skin on the back of my neck being stretch and \'trembling\' a bit due to the force of the push. When pushing arms down, note not to loose balance due to overstretching. As I lean body back, arms will raise a bit to an almost hotrizontal arc. As I pull my arms back to my hips, I should be following the ki in which my fingers are pointing. When done without a jo, uke will feel that they are not able to hold back the force of this pushing and pulling motion.   Home   Finally cleared junks of old toys and clothes by getting teh Salvation Army to pick them up on Fri afternoon. Woohoo ... room got more space now. I\'m 1 step closer to getting my loft bed. Next thing I need to do is saw up a old cupboard of mine into small chunks and getting rid of it. I don\'t knwo whe I can or should do this though. Need to plan my time a bit. Folks came back from Thailand yesterday afternoon. Dad had a swoolen knee, dunno why. Mom had a swoolen foot. Heard them mention something about being old and taking too much of a certain food. We are what we eat, that\'s for sure. Sis was home then so I asked her if they mentioned anything about us donating so many new (but unworn clothes) to ther Salvation Army. She said they made a heckuva ruckus when I wasn\'t around but nothing they can do about it now. Muahahaha ... my idea of Xian1 Zhan3 Hou4 Zou4 worked brilliantly. Anyway, I tried to appease them by just walking over to show them the whole series of Yi3 Tian1 Tu2 Long2 Ji4 (Heaven Sword and Dragon Sabre) I purchased few days ago. Watched about 3 VCDs lastnight and I think they were hooked liao. Watched another VCD today. Probably after lunch can watch a few more. Muahahaha ... think it\'s the start of me getting a whole series of those oldies HK TV series like Shen2 Diao1 Xia2 Lu3 and Lu4 Ding3 Ji4. Shiok man.   Friends   Steff will be coming to Singapore next week with some friends from Uni. Heard she\'s starting a PhD in Brum soon too. Good for her. I think I better do some research on where she can visit for 4 days or so. I can only suggest eating and eating hahaha.  James\'s wedding dinner at the Singapore Marriot Hotel was good. You can tell he really loved the day. Good to see him settle down at last. Met up with Shuhua, Wanloong and Yvonne that night too. Was nice to see familar faces again.   '),(26240,979795,'30,June,2004','2004-06-30 00:00:00','    Niiiice... went for an evening run when I got home. Managed to clear my 3km at an OK pace. Traffic wasn\'t as nice as in the early morning. But I realised that these days it\'s difficult for morning runs because of work. Evening air isn\'t as fresh because of the recent haze and traffic on the roads.  My Kinokuniya 20% discount expires today. Kinda regret not going just now. I used one over the weekend and got myself a book, Secrets of the Samurai, by Oscar Ratte and Adele Westbrook. The same authors for Aikido and the Dynamic Sphere. Book was written in a similar style and it\'s not a pure Aikido book. Touches on the history, tradition, developments ets of traditional practices and arts. Also explains a bit about the philosophy of the Budo-kas then. Very very interesting.  I\'m about finishing my book We Were Soldier, Once and Young soon. Damn tempted to read the Secrets of the Samurai next. But I figured I should probably start with Killing Pablo or Lord of the Rings next instead. LOTR is not an easy book to read for me because of it\'s language, background and the author\'s style of writing. I\'ll probably go onto Killing Pablo first. He was a head figure of a drug cartel which the CIA actually tried to take out. A lot of classified details in the book that makes the whole operation damn interesting. I guess the background of the story simply appeals to the policeman in me lah. Heh ...   '),(26241,979795,'26,June,2004','2004-06-26 00:00:00','    As of yesterday, I started taking P1 classes in some basic computing enrichment. In fact, it will be progressing for the duration of another 5 more weeks. Yesterday\'s classes were very bored with all the theory stuff I had to say. But today\'s classes were slightly better. Noticed a trend, the first class I take will kinda be the \'worst\'. Probably I\'m still experimenting on how to handle people and teaching contents across to the class.  Good weather in the morning because it rained. But that stopped pretty quick as well. Hope that cool weather will prevail from today onwards. Well, at the very least, I hope it doesn\'t get warmer than it already is.  Went Aikido at BVCC. Was quite good training. Freaking knackered though. Was totally wet by the end of the first hour. My uke had a problem keeping his momentum coming and I\'m not so good at controlling him. Even the first thing we did (backstretch) didn\'t feel that good. He\'d just grab, then go all heavy like a boulder. After my tenkan, he\'d just hang on damn tight to my wrist but did not push. When he push, he keeps pushing me forward instead of up. Didn\'t feel like doing anything funny today, so I kinda left him as he is. Anyway, he wasn\'t stiff, just that he\'s not feeling enough as uke. Of course, as nage I probably didn\'t lead him in enough in most of the things we did. Oh well, takes 2 hands to clap.  Later on we did group work and there were a few throws. On one of them, after I was thrown. Something \'came out\' of my nose. The feeling I had was like \'wah shiok, my nose feels so clear\', as though I had just blown my nose like that. I wiped my nose with my hands and I saw red. As in blood. Fish. Was having a nose bleed. At the same time Sensei called for a stop and he demoed another technique. I quickly set down behind and started holding my head up. Didn\'t even notice what technique he was teaching. Wiped my nose a few times and the blood just kept coming. Think it was bleeding on and off for at least half a min. Not sure how long it was but it felt like ages before Sensei let us practice amongst ourselves. I asked permission to go to the toilet amid all the surprised look. Maybe some folks would be thinking, \"Was it Sunny who gave Aylwin the bloodied nose?\" \"Or Eileen\" muahahaha... Well, I cleared myself up quick and spent the next 30 mins or so watching instead. Didn\'t wanna risk more blood dripping over my gi. My hands got most of the blood, but I did touch my pants by accident and left a bit of stain there. Anyway, I hope the stains come off. Got them washed the moment I got back.  Got a Super Big Gulp for 2 bucks from 7-eleven after training. Woohoo ... shiok man. I felt too bloated after that. Should\'ve stuck to the normal $1.70 Big Gulp instead. While waiting for the bus, the whole area suddenly blacked out. Not kidding, the only lights came from the headlights of cars on the roads. You could see the even streetlamps were out. The darkness stretched from Clementi area down to Holland Village, Commonwealth, Queensway, Holland Road (to town), etc and those are only the area I passed by on the bus. I believe it was quite a major black out and a huge part of S\'pore was affected. Was kinda worried that my home would be affected too but luckily, the area from Orchard Road onwards seemed fine... whew.  A really weird day.   '),(26242,979795,'26,June,2004','2004-06-26 00:00:00','    Previous post had something related to Ki-ai.  In AFS, we\'re not really taught on how we really do a Ki-ai. But in Yuishinkai, we\'re told that it\'s \'e\' (as in the word h a y) and \'i\' (as in the word h i t). So it sounds pretty much like \'eeeh-eeee\'.  Only a matter of how long you drag each syllabus. For instance we do a shomenuchi strike, at the start of the strike (when arm is swinging forward) the \'eh\' sounds off and \'eee\' comes immediately after. But \'eeee\' is suppose to end together with your strike (when hand reaches maximum pt/end of strike at hip level).  I guess it applies to shomenuchi undo, rowing exercise, striking exercises, tsuki, etc. Good pt.s to note. I tend to only ki-ai during warmups. But never in a technique unless I\'m told to do so.   '),(26243,979795,'23,June,2004','2004-06-23 00:00:00','    Attendance on Thurs night was OK. I was there first followed by Daniel and Darryl, Roxanne and Sakai. Weide needs to study and won\'t be training till end of year after his end of year papers (crucial year for him). Leon and Bridget MIA. Meng Lee wasn\'t around, most probably due to work again. The 2 beginners (couple) came a bit later than usual.  One Indian chap is starting to resume training with us again at Siglap. I did see him a few times here a long while ago but I think he probably moved on to train in another dojo. Recently he\'s been joining back with us again. Quite OK dude, but sometimes overly enthusiastic as an uke. Heh ... as Tang might put it ... sometimes we over-exaggerate our reaction/momentum of a grab, a pull, a punch, or an off-balanced motion as uke. I know I\'m pretty guilty of that sometimes. A lot sounds coming from him when he partnered with Roxanne. Err, something like a partial ki-ai. But a lot of it seemed to me to be more of just making sounds on the mat. Dunnoz, ki-ai is something I\'m still trying to understand.  A lady Yudansha popped by today. Spoke to her for awhile once before during last year\'s Aikicamp. Not sure if she still remembers me. Think she doesn\'t really train anywhere specific, but she\'s on good terms with Freddy Sensei. Heh ... I dunno why, it kinda reminds me of Ah Hee Sensei. Think her name\'s Jessie. Anyway, when we were doin warm-ups she came in. I was surprised as I thought she was taking the class for that evening. Later, Richard turned up. Than I realised class was as per normal. I pretty much partnered with Sakai the whole night. Jessie was teaching the couple rolls and some basic techniques while Richard went around guiding everyone in class. Was a good evening.  Was surprised that Richard called me up as uke that evening. Sure, it\'s not the first time I\'ve uked for him in class. BUt the trend was usually that he would call the most senior student present to be his uke (meaning that if a more senior one came late, he will still be calling up the initial uke). That\'s why I was surprised because the twins were both 4th kyu and they weren\'t late that evening. But it was good practice, helped me to feel what a nage wanted me to do/how to move. A lot of leading in fact. Only bad thing is that as uke, sometimes I\'m not able to \'see\' how to do the technique.   '),(26244,979795,'23,June,2004','2004-06-23 00:00:00','    Will be starting to revamp my  urlLink homepage , the  urlLink Aikido in Singapore  site, its logo, etc over the next few months.  Hope to integrate Blogspot inside as well. So far I\'ve started transferring the  urlLink \'Art of Peace\'  and the  urlLink \'Quotes\'  into blogger. In fact, I like them so much I\'ve added the links into my own blogger template. :Þ   '),(26245,979795,'19,June,2004','2004-06-19 00:00:00','    Feeling very happy this morning.  Checked my Friendster a/c to discover an old sec sch friend, Yuanhui, managed to find me. Steady lah. I immediately smsed him on the bus to work. He called and we spoke briefly for awhile. He highlighted a few more people from BVSS who were on his Friendster a/c. I popped back in and started adding or messaging people.  So surprised to hear that he has already ROMed. So happy for him. I can\'t believe the length of time we\'ve not seen or heard from each other. Almost 8 yrs or so... gosh. Time flies so fast when you let it. Anyway, we both promised to stay in touch and try to arrange the next earliest date when we can meet up for a chat, makan or coffee.  After a few hours, I was quite happy to noticed that another classmate, Yinhong, replied the message I sent in the morning. Gosh, memories just kept flowing back. I remembered I was always throwing stuff at her and her friend or tugging her hair back in lower sec. Hahaha ... yes yes ... I was quite a monkey back then.  Anyway, it\'s good lor. I\'ve always like my sec sch life and friends a lot. It was a pity I kinda let them go after sec sch and lost touch with everyone save for Robin and Roy. Roy, no need to say, he\'s my best bud even after so many years. No doubt we don\'t communicate as much but each time we do, we talk or email like it was only yesterday. Robin, well, can\'t complain. He\'s one of the few who still makes some effor to stay in touch with me after so long. It brings to mind another good buddy in upper sec, Jialiang, who had kinda drifted away over the years. Quite sad really. Oh well... I\'m just glad my Poly and Uni friends are still around me at this point.   '),(26246,979795,'19,June,2004','2004-06-19 00:00:00','    After the grading (see previous post), I went down to BVCC for Hattori Sensei\'s seminar. Managed to grab a quick bite at Macs before popping over to the CC.  First thing I saw there, was a figure flopped over on a chair snoozing. Couldn\'t tell who it was initially, so I went straight to change into my gi first. When I came back I saw Tommy kinda poking at the body or something and it was stirring... oh ... Eileen nah... thought who.  Anyway, talked to Tommy awhile about that morning\'s grading and Yuishinkai\'s training. Class started pretty soon after that. It was quite nice all the way (though somewhat crowded) until we started weapons training. First few were still OK, but towards the end everyone was kinda \'I stare at you and you stare at me\'. Well, for one thing, I had problem even see the whole kata because I was too far behind and there were just too many heads blocking me. I found that evening slightly disappointing because I was actually looking forward very much to Hattori Sensei\'s weapons class.  Mon evening, after work, I left early and went down to BB East CC. Hattori Sensei will be teaching there as well. Was early so I enjoyed the curry puffs I bought from Old Chang Kee earlier. Later I saw Iulian, Ah Hee Sensei, Alvin around and popped in to help with the mats. The night class was very very very good. For one thing, I was lucky to have paired with a pretty decent beginner (Kenneth, who seems to train at BB East). Also some group techniques went quite smooth too. But Hattori Sensei\'s weapons class that evening was the best. He revised what we did the day before, but this time he broke it down into parts and everyone picked it up a whole lot faster and easier. Really shiok lor. I like the fact that he\'s teaching mostly same things as what he did in last year\'s Aikicamp. Because we don\'t do enough weapons work, it\'s the only way he can make sure we are learning correctly, by teaching us himself. I wish like Tommy mentioned, there\'s someway he can conduct regular weapons work locally instead of just once or twice a year. Anyway, the final night\'s training went very well and judging from the applause, I think everyone felt the same as well.   '),(26247,979795,'19,June,2004','2004-06-19 00:00:00','    Right, I have to put in a separate post for this.  As I mentioned, during Sat\'s class Williams Sensei requested that some of us drop by if possible to help as uke for the Sho and Ni dan gradings on Sun. Sun morning, I went down to GWCC as arranged. Everyone was there early. After warmups, Sensei explained some common procedures during the grading (eg. bowing to kamiza, first, bow to ukes, time in between the bow to kamiza and time turning around to bow to ukes should be the same, bowing to ukes is on own discretion and timing and NOT after the call of \'Hajime!\' - a shomenuchi strike would be on the head by then. Also, when nage was suppose to do a weapon technique or kata, they mus never never take the weapons themselves. Instead, someone had to kinda take the weapon from the ground and pass it to them instead.  Even the nages taking dan gradings had to rotate as ukes (3 shodan, 2 1st kyu). In addition, there was a 3rd dan (Japanese lady based in Australia I think, practicing for 8 yrs I heard), 1 1st kyu (I think), 2 green belt (3rd kyu), 2 white belts (namely Tony and me, Aikikai 5th kyu).   Shodan Gradings  Jeremy (1st kyu) started first. Katatedori grabs, shomenuchi and yokomenuchi strikes as well as tsuki punches. Basically everything was free play. After that were the bokken and jo katas. Jeremey did well but you can tell he was really nervous and he really got breathless very quickly. Usually his movement would be pretty good. After the kata came randori.  Pearly (1st kyu) - if I got her spelling right, came next. She was very very nervous. Pretty much the same thing as Jeremy. But she was so nervous she got quite a bit of the kata wrong. Plus her weapons work didn\'t seem to be as \'forceful/strong\' as Jeremy\'s.   Nidan Gradings  Next Michael came. I\'d say he performed the best. Everything was pretty much the same except that randori was against 4 ukes instead of 3. In addition, there were some techniques against the bokken and jo as well. Also there were some randori using the jo against 2 ukes. Anyway, I think he\'s nervous as well because he got a bit breathless in the middle. On a side note, Michael and his wife cross train in Kendo. I think they are shodan but they might have gotten their nidans by now (or will be getting it soon).  After Michael was his wife, Janice. She was pretty good too. I can\'t remember much save that it\'s almost same as Michael\'s.  Last was Mark Tham (the guy who created the Martial Art website on his homepage). He\'s a Wing Chun practioner too. His techniques are pretty soft and you can tell he adopts a lot of the \'Ki-style hopping around\' quite often. But he has a very soft style. I distinctly remembered being his uke for his \'jo randori\'.  I honestly can\'t remember who were the ukes for whom. I helped out in most of the katatedori, a few shomenuchi and most of the randori. A few distinct things to note is that I did some good ukemi, some weird ones, and some extremely horrid ones too. Basically we ran and fell all over the place grabbing, striking or whatever. Trying desperately to build up a certain rythm for nage. Sometimes as uke, when I hesitate, I kinda broke the rythm for them and they had to pick things up again. It was a good lesson on randori though.   '),(26248,979795,'19,June,2004','2004-06-19 00:00:00','    Only one thing to be said ... FANTASTIC!  Yesterday evening, I left school early so I was sure I\'d be able to arrive on time. I even had time to grab a dinner at MacDonald\'s before going for class. When I arrived, I caught a slight glimpse of the seniors sitting in seiza and I could hear Williams Sensei explaining aspects of weapons work (bokken to be specific) to the class. I was both flustered and puzzled at the same time. I thought I was late and class had already started. So I quietly slipped away from the door to the toilet downstairs to change into my gi. When I came back, I waited at the door and saw Williams Sensei (for the first time), explaining or talking to the class. Next thing I noticed was that everyone had a bokken in their hands. Pengz ... I didn\'t bring mine because I thought weapons wouldn\'t be taught. Geez ... first I was late, second I forgot my weapons. Thought I really deserved to sit at the back of class to watch instead of joining in.  While I sat in seiza (which I later changed to cross legged coz of my strained left ankle), more and more people started arriving. That\'s when I checked the clock in the room and realised it was early. I did arrive early. (Whew!) and I noticed a stack of weapons on the left side of the dojo. So the class didn\'t bring their weapons either. (Double whew!) Just as class was about to start, Williams Sensei left the seniors and came towards us, the new arrivals. Frankie Sensei introduced each and everyone of us and he shook hands warmly with everyone. That really struck me. He was such a warm and friendly person. He carried himself in a very casual manner but the words he used often carried a certain \'oomph\' that it always made us think. I do feel that I did manage to pick up some things. I only wish I can constantly remember his teachings. He was very relaxed and casual with the class. There were some laughs here and there as well. We started with warmups and he corrected us on a lot of technical things as well. Some of which were certain co-ordination exercises (musubi). As there were a certain no of seniors about to be graded on Sun, the lower kyus had to train separately instead. Still Williams Sensei came over to teach us. First thing he wanted to do was check our rolls. I picked up a lot about forward rolls. I only wished I could put into words what I had learnt that evening about my rolls. Suffice to say, I intend to spend regular and a good deal of time correcting the finer points of my roll. Williams Sensei was always kind with his words and he was always very encouraging to everyone. It made everyone feel like his praises was as sincere as his critiscms. Despite being pressed for time, the patience he displayed in class was extraordinary. Though he was constantly explaining things to the students, it never felt \'boring\', \'tedious\' or anything like that. On the contrary, everyone seemed to hang on to his every word and advice. It was one heckuva experience.  Today, I went down to ARCC for class and Williams Sensei conducted the class. Again, some of us came very early. We all hoped to join the class early and learn as much as we could. Prior to class, Williams Sensei showed everyone a VCD of his daughter taking her SHODAN (1st dan)!!! She was veeeeeeeeery good! Her movement and strategy was similar to the lady in the video Gary showed me once before (nidan/sandan grading I think). Even Williams Sensei commented that she was pretty good but her performance on that grading wasn\'t her best. He\'s seen her work much better under preassure and he pointed out some aspects of her techniques or weapons work that \'should improve when she gets more serious with her training\'. And guess what? She only just turned 13!!! .... I wanna pengz liao. If that\'s the standard for a 13 yr old who isn\'t 100% serious with her training ... I can\'t imagine how she\'s gonna be like when she gets older. Anyway, even Williams Sensei did comment that he sees a lot of potential in her. There were tons of randori against 2 ukes. Her weapons work was also extremely good. Everything was smooth and flowy whether her ukes were kyu or dan grades. I was like ... speechless man. I only wish there\'s some way to make a copy of the VCD.  We started with warmups followed by some co-ordination exercises before beginners were split up from the rest of the class. Again, this was to facilitate the seniors training before the grading tomorrow morning. But this time, as there weren\'t enough ukes. They asked me Tony and I to join the seniors to help out as ukes because we were 5th kyu Aikikai. It felt great to be uke today because I did learn a lot as well. Each Senior (to be graded) would call up a senior and a junior kyu as ukes. The training went so good today that we will be helping out in tomorrow morning\'s grading. I saw Leong videoing down yesterday\'s and today\'s sessions. After the grading and Tuesday\'s sessions are over, I\'ll have to ask for a copy of the videos. It\'s too good a class not to share it!!  I\'m really looking forward to the grading at GWCC tomorrow morning. But after that I\'ll need to rush off down to BVCC for Hattori Sensei\'s seminar with AFS. I\'m feeling a bit guilty about sacrificing more than half of the seminar but I think I really don\'t have a choice. No regrets so far anyway, I\'m sure whichever I go I\'d still be enjoying myself thoroughly.  Williams Sensei ended today\'s class with a talk on how Aikido is applied in our daily living and such. A lot of good advice and pointers today. Tis been a truly Aiki weekend so far.   '),(26249,979795,'16,June,2004','2004-06-16 00:00:00','    Left school early yesterday but missed 2 buses of 196. Stupid stupid ... ptui ptui. Well was actually because I was in the midst of a serious discussion with Daphne. Piangz, she so sensitive about stuff. Chin cai lah ...   Aikido @ BVCC  Grabbed my haircut and missed the first hour of training. Still the second hour was ok. I kinda strained/sprained my left ankle when I landed badly. Noe whenever I need to do a forward roll which uses the left foot, I do it \'Judo\' style. Why do I say that? Because I land exactly the same way Sakai (Siglap CC) does. Anyway not so bad that I can\'t walk properly. Only hurts when I fully extend my foot to be on tiptoe or when I twist it to the side. Just a strain I would think. Bought my ticket for the seminar this coming weekend and Sensei also announced a news. $5 to be charged per session when anyone drops by a dojo where they do not pay the mat fee. Think it\'s a good idea but I hope it can be enforced among the students tho. (One eg, I didn\'t pay for yesterday\'s class ;Þ but I intend to do so from next session onwards.) I had thought that the $5 was applicable to only Pillars (cos it wasn\'t a CC but a pte spot). Joining 2 clubs means we are able to visit other dojos without paying mat fees. While keeping mats, Tommy was asking \"Eh, so now how ar? U join which dojo?\" Told him my impression was that cheapest 2 were BVCC and another CC (but can\'t rem which exactly. Still it\'s gonna amt to about 160+ (3 mths) or so coz I think the CCs increased the charges due to the GST increase.   Dojos vc Cost  Just checked, cheapest 2 dojos seem to be BVCC and BBCC (not BB East CC). But then again, joining 1 dojo and just paying $5 per session (if you train an extra 1 session a week) means total expenses will be about $130ish (3 mths). But the most worthwhile spots to train for the amt you pay would still be BVCC and BB East CC (if you consider at an hourly rate). Sigh ... too bad I hate Maths, otherwise I\'d do up a statistical analysis of these costs.  But if joining 2 dojos means you can train any no. of times and anyplace... now that is definitely the best deal. Just sign up for the 2 cheapest dojo but make sure you get ur money\'s worth of regular training. If you are only to train a certain no. of times it may not exactly be worth it.   My Conclusion  I guess next thing I should do is seriously consider how regular I plan to have my trainings. I don\'t intend to leave Siglap anytime soon yet. Doubt I will increase my AFS training beyond what I\'m going thru now because of distance, cost and time constraints. Also because I\'m committed to Yuishinkai training once a week as well. I guess I will stick to Siglap on Thurs and ARCC on Sat with Yui. As for Tues at BVCC / Pillars on Mon, I may rotate it to alternate weeks or whenever my schedule dictates. Strongly doubt BB East is a feasible spot for me to travel to as long as it\'s a working day for me.  This morning I went running again. It\'s ok but I was kinda late. If I only made myself get out of bed 30mins earlier then I\'d be able to reach school on time. Was feeling lazy. Actually woke at 6 but lazed in bed till 6.30 instead. Started slow coz I was afraid my ankle will give way when I\'m running halfway. But it seemed fine. I\'ll be meeting 2 colleagues for lunch and a reservist in the evening. The latter is some financial advisor with AIA so I\'ll be finding out more about these kinda stuff soon. Hope to start some investment/savings plan within the next year or so so need to get some research done now. Evening I\'ll probably drop by to collect my samsung battery from James then pop home for dinner.  Can\'t wait for the Aikido seminars (AFS and Yui) to start soon. But it will be a very tiring week for me. Training will run consecutively for 5-6 days for me. Hope I can take it.   '),(26250,979795,'11,June,2004','2004-06-11 00:00:00','    Oh man, can\'t tahan. Really sleepy today that\'s why I turned to blogging for awhile.  Resumed reading everyone\'s blog and only realised how much I\'ve missed out the past weeks. Especially this Guilian ... ultra long posts man hahaha. Anyway, my mind is quite dead to absorb everything I\'m reading on these blogs so I only left minimal messages or comments here and there. At least I\'ve managed to update the format of my blogpage. Now all the links are back. Only thing I\'m hoping to be able to do is to include my old enetations link so that I can check the comments with regards to older blogs. Oh well, not a priority lah so I\'ll see how is it later.  Resumed my 3km run this morning before going to work. Felt great after that but I built up a lot of heat inside and I can feel myself still perspiring even after a shower. One good thing, I noticed I still maintained the stamina. Only big thing I needed to note was maintaining pace throughout the distance I\'m running and I should be able to complete it OK. I think at this stage, daily running is necessary for me to get ready in time for AHM 2004. Doesn\'t sound possible but I may take 1-2 break days depending on my physical readiness, workload, etc.  Aikido at Siglap yesterday saw 2 new adult students - husband and wife actually. Husband was Chinese and wife was ... I dunno, maybe French? I detect an accent somewhere. Took warmups but I seemed to have missed out the wrists exercises. Haiz, felt like blaming others but actually it was my own fault. I was asking Leon what came after this or that exercise. Shoulda trusted my instincts more and just do the warmups in the sequence my brain remembered them. See lah, in the end made everyone miss out on warmups. Anyway, the couple were quite keen to bring their 5 yr old son, goin-onto 6, for the Sun class at Siglap. Good lah, we get additions in both classes this way. Just hope things didn\'t seem too slow and boring for the couple during their first day of training. Coz I realise a lot of pple actually stop Aikido even before they\'ve started it.  I\'m glad Sakai (judoka yudansha) stayed on. He\'s been with us for about 1 mth or so. He seems to be improving too. Still stiff but you can see him getting better. Paired with him for the duration of the evening. We did TOOOOOONS of shomenuchi ikkyo and nikkyo (omote + ura waza). Most probably because Richard had to give most of his attention to the 2 beginners. Well, that\'s fine. Most part of the night I shared what I knew to be the \'finer\' pts of the technique with Sakai, tips mentioned or overheard.  As when I first started, Sakai tend to miss taking the side step. Pointed out to him we tend to be crashing into each other even before the technique was done. He did manage to pick it up slowly. I think the actual ratio of practice I had to his was only about 1:4. I tend to go along and let him repeat as many times as he wants till he got comfortable with the technique. Next mistake I noted was he almost always grabbed my arm when I used a shomen strike. Pointed out that it was more of a guiding movement. Other things I managed to point out were footwork, positioning, etc. Well, at least he was willing to listen. I only hope I didn\'t bombard him with too much information last night.  For kokyu-ho (my crappiest), Sakai tend to anticipate what I was going to do. So most often I really was just following his arms instead. I advised him to close his eyes instead and just feel what I\'m doing. It helped a bit but not much. His \'stiffness\' is most noticeable during kokyu-ho. But there was something really nice towards the end of kokyu ho. He did manage to move a bit more smoothly and lessen the stiffness. Not great, but still good.  At the end of class, managed to speak to the couple. They were asking things about belt colors and all that. Told them the twins (Darryl and Daniel) were more senior than me. They were a bit surprised and the wife mentioned something about seeing me \'teach\' the other guy (Sakai). Ar ... I think I might\'ve said too much liao with Sakai that\'s why it gave them that impression. Anyway, told them that us students aren\'t qualified to teach, only to share hahaha.  All in all, was a nice evening.  Will be meeting Henry tonight, he\'s one of the old M113 drivers I was with during our trainee days. Haven\'t sen him for a good 4-5 yrs in fact. Dunno what we\'ll be doin, but in the very least we\'ll be grabbing dinner. Hope to be able to catch a movie too before work piles up so much I can\'t do so.  Serious man, this holiday is quite a killer. I don\'t get these much stuff in a normal school term. Haiz, dunno how to finish everything before the given deadlines. Organise organise organise ... and stop procrastinating you fool!   '),(26251,979795,'08,June,2004','2004-06-08 00:00:00','    Yes man. Really pissed for yesterday and today.   One Particular P6 Class   Took a class for Ulead Media Studio yesterday morning. They were quite unresponsive and rude at times too. Now, I can take cheekiness to a certain extent. In addition, I allowed them to be on the Net for most part of the class. Wrong decision, they can\'t even concentrate on the task at hand. Very poor participation. Worse still, at the end of class, I found scraps of paper on the floor, on the table and spilled drinks. And no one bothered to clean it up. Such irresponsible behavior is totally unacceptable.  When they came back to class this morning, I wanted to overlook the mess they made yesterday but they started making so much noise. I had to step in, scold and lecture them. I know these are good kids, but sometimes I find that I really need to remind them of their place. At least now they are busy producing the video work I want them to. All I want, is for them to have fun and learn something out of the whole thing.   2nd Class Yesterday   The second class was a lot more fun and responsive. AT least they produced some really really good work and appeared more interested. But this afternoon, they were a bit more noisy and some of them didn\'t listen to the specific instructions I told them - no Internet until work is done. Had to raise my voice at them again. Today is a seriously tiring day. I feel very taxed even during lunch time. Will probably be able to get a good sleep later on the bus journey to BVCC for Aikido. Hope tonight\'s training will be a good one.  Another incident yesterday afternoon occured around 5pm. Classes had just ended for me and Daphne and the school auntie came to knock on our door informing us that she was locking up. I reeeeeeeeeally hate the kanchiong attitude I saw, comparable to the \'kanchiong spiders\' in NS - what we\'d call people who were rushing and panicky about stuff and always wanting to go home early etc. Just wanted to point out that things like off and early release is a PRIVILEGE!! NOT an entitlement. This policy in NS stands whether we\'re in NS, govt or pte sector of work. Just wish some folks will understand that.   '),(26252,979795,'01,June,2004','2004-06-01 00:00:00','    Can\'t believe the lapse in blogging I\'m experiencing these days.  Been busy settling the coming IT monitors camp and Ulead Video editing classes for the holidays. Prep work seems almost complete but both will start running from next Mon onwards and stretch on for quite a period over the Jun holidays.  Sis will be visiting her friend in Philadelphia. Wish I can join her and drop by Gary\'s but too bad lah, boh lwee and boh si gang (no money nor time). Parents going China tomoro onwards. Woohoo ... nopw govt for another 10 days or so.  Most woulda noticed the new look to my blog. Think I\'ll need to do a complete revamp of my homepage to match my blog soon. But I doubt I\'ll be free to do this over the next few months. Love the new look Blogger adopted though. Best feature is that they now have their own commenting feature, but that\'d mean that my old comments are gonna get scraped. Gee ... anyway, that\'s kinda why I\'m only posting my blog on 1st of Jun, so that I can republish only my index pages and not the ones before. Hope I can retain those.  IT monitor camp is starting tomorrow. I think I\'m as ready as I\'ll every be, but I\'ll probably wake early to run through some things and the whole program for the day in my mind... just to be sure. Daphne\'s taking the P2-4s while I\'ve got the P5-6s. Hope I can make this as enjoyable and be as friendly as I can, no point spoiling everyone\'s mood by getting to mad with them. Just hope they don\'t push the teachers around too much and go overboard. Give and take lah girls (if anyone is reading this).  One of the P6 actually said they came across my homepage. Like Daphne mentioned my homepage has a lot of personal stuff. Not sure if it\'s appropriate to speak my mind and be as flamboyant about my thoughts now. But hey, it\'s me ... I don\'t think I\'ll change much on this. Even if they do see my page, maybe some of the things they read will be able to help them understand somethings about the way I do things. Oh well, honestly not giving this much thought. Err ... but it does mean that I now ahve to watch my language online. Geez ... what a hassle.  Been procrastinating my training for the AHM and Singapore . I think after this camp is over I better start planning my training regime and stick to it. Been meaning to do my regular runs a long time ago but everything just seems to get delayed. Ar well, it\'s my fault lah. Problem now is I won\'t be able to start until Wed (Vesak Day). If it\'s physically possible, I may do daily runs just to make up for lost time. Not sure if I can clock up enought distance in time but I gotta try.  Been looking around for the AHM 2004 details. Only managed to find a  urlLink website  which says AHM is on 1st Sep. That can\'t be right, 1st Sep is a Wed. These kinda stuff always falls on a Sun. Anyway, I\'m guessing it\'s 4th Sep. Hope so coz that week is the kid\'s 1 week term break. Might actually do myself some justice by taking the Mon off/lve to rest and recover. Also been to the  urlLink Singapore Marathon  website but it doesn\'t seem like application has started yet. Anyway, I gotta check back on this coz I want to be able to meet the early bird registration. I wonder if both routes run past the Nicoll Highway (after the collapse).   '),(26253,979795,'31,July,2004','2004-07-31 00:00:00','    Something Syed said to us when he taught thu\'s class at Siglap CC.  He was trying to teach us a technique. I can\'t remember which one it was, but the main thing I remembered was NOT the technique, instead it was when we chose to step forward and back. Realised after being shown a technique, we\'d photocopy and reproduce the same thing right down to the footwork, hand movement, etc.  Syed tried to explain why we moved back and forward. Depends on Uke\'s timing, speed of attack, etc. He told us the words above. Something I\'ve spent a few nights dwelling on. I have limited understanding of the words as I try and try to place them into context when it comes to my training. Change the scenario, and I feel lost. Change the technique, the same happens.  \"How\" can I train with understanding... when do I ask the \'right\' questions... what are the \'right\' questions to ask to achieve this?   '),(26254,979795,'27,July,2004','2004-07-27 00:00:00','    Again, anyone who hasn\'t seen the movie may want to refrain from what I have to say in this particular post.  Story revolves around the brotherly love between 2 brothers. Starts off just before the Korean war, the simple but blissful life they were enjoying then. When war soon broke out, both brothers were forcibly drafted into service and sent out to the front lines. Younger brother was a the weaker one. Student before the war and on his way to enter University. Elder brother was a shoe polisher who gave up school to support his brother, their mother and a fiance (with her 3 younger siblings). Mother and fiance spent the day tending their own noodle store.  At the front lines, you\'d see how the elder brother tried protecting his brother. In their first skirmish with the communists, the younger brother felt really lost and almost suffered a heart attack. The elder brother would think of ways and means of getting his brother out of the army. His only aim was to watch his brother survive the whole war and get him out unscathed. The commanding officer somehow gave the elder brother an idea that one way was to distinguish himself above and beyond the call of duty so as to get the Medal of Honor. With this, he could put in a request to get his brother transferred. This led to his brother volunteering for almost every risky mission that came along the way of the platoon.  Soon, the unit found themselves cut off from the rest of the army. To delay the enemy, a group was ordered to plant mines. The elder brother volunteered for this. On completion, they were ambushed. Here, you\'d notice that the elder brother was brave and exceptional in soldiering. He managed to pull a wounded comrade to safety. But he would risk everything and more in his quest for the Medal. When ordered to pull back to escort the wounded back to camp, he insisted on staying on to fight the remnants of the enemy ambush party and finish them off. They were that close to losing the wounded soldier as he was losing a lot of blood. The SGT told him never to disobey his orders again during a battle.  Days passed and the cut-off unit were low on food, water and medical supplies. The sight of maggots gnawing away at the wounds of a wounded soldier was really disgusting. Morale was low. And in despair, one of the wounded got his hands on a weapon, shot the wounded soldiers in the tent before blowing his own head off. The elder brother suggested that they go on an all out attack. A logical idea as everyone had nothing to lose anyway. That very night, they striked back. The elder brother would put himself at risk time and time again in the same battle. The most exceptional was him taking out the command bunker and killing the enemy commander himself. The younger brother would also take his first life on that very night. This was the night of nights as it saw a transformation in the both of them. The younger brother would learn that sympathy to the enemy during a battle could mean the end of himself. The elder brother would let his heroic exploits get to his own head. He was branded a hero and the army was quick to use him for their own propoganda. They would feed him \'lines\' to read out to the press. No doubt all these would fall under psychological warfare but I think there\'s nothing worse than playing with a person\'s mind. The US soon joined in the war to assist them. Elder brother was quick to get himself promoted to SGT and as leader, he would often put his fellow comrades at risk to accomplish the mission, even when the mission had been designated under somebody else. There was a difference between taking the initiative and going all out unecessarily. Very soon, in an attempt to capture an enemy captain alive, he indirectly caused the death of a platoon mate. Younger brother, having learnt the trade of soldiering, blamed his brother for this. It was quite sad to see a rift come between the two of them.  Even more heart wrenching was the scene where the communist, while retreating through civilian villages, would kill everyone in the village , women, children and old folks. They even hung bodies out on trees and booby trapped some of these bodies. These traps kill and wounded many soldiers. But the scene of slaughtered civilians created a lot of pain and hate within many soldiers. They would kill the enemy without remorse and slaughter the enemies they captured. It was scary to see how they changed and each one of them holding on to so much hate. Only the younger brother seemed to still retain a bit if humanity. One day they managed to capture some enemy soldiers, one of them being a good friend of the two brothers. He had been forced by the communists to join their army. The elder brother, overcomed by so much hate ordered that all captured enemies be shot as these soldiers would be \'excess baggage\' for them. Only the younger brother protected the PoWs. I know that war changes many things, I\'ve heard of gentlemen\'s war, guerilla warfare and such. I have no idea how one should balance both. To have a war without honour, you\'d see the same atrocities that were televised regarding how US MPs treated captured prisoners in the Gulf. To fight a war with too much honour opens you to guerilla and unconventional tactics.  Anyway, China decided to join in the foray and the communists started fighting back and gaining lost ground. The US were quick to lose ground and soon the winning side started to retreat. During a scuffle, the PoWs managed to overcome their captors and held a soldier hostage. The elder brother started firing on the PoWs and even shot his friend. Amidst the retreat, the elder brother was finally awarded the Medal of Honour. But he had \'lost\' his younger brother for it. The latter hated everything the Medal represented, he felt that he didn\'t know his elder brother anymore. While passing through their hometown one day, the younger brother decided to look for his family as he heard they were still in the area. He found his brother\'s fiance. Unfortunately, she was soon arrested by the Anti Communist Army, a group which weeded out communist in the community and executed these traitors. The younger brother found himself arrested as well. I found this hard to swallow. The abuse of power but this party was frightening. They were able to just pull people off the streets and execute them without trial. They based their accusationa and charges on papers they retrieved from the communist parties. These papers were namelists of civilians who had attended communist talks and rallies. However, to attract civilians, the communists used to exchange food to attract people to attend. Quite possibly most of them only went for the food and they didn\'t really understood the difference between democracy and communism. I just felt so helpless at the level of misunderstanding I was seeing. All these could be cleared up (though not without time and a great deal of effort), but instead cicilians were executed based only on  suspicions  of being Marxists!  The elder brother, recently promoted to Staff SGT was quick to arrive. He tried to protect his fiance and get his younger brother out. In the ensuing struggle, he shot several members of the Anti Communist Army. The fiance was shot and killed by someone and the brothers arrested again. While the elder brother was interogated by a superior officer, the communists arrived. In all the confusion, the elder brother managed to get free and held the officer at gun-point, ordering the latter to free his younger brother. The officer instead ordered that the makeshift prison be set on fire. The elder brother tried to rescue his younger brother but did not manage to do so and somehow got injured as well. When he came to, he found a pen he had bought and given to his younger brother before the war. Thinking that his brother had died, he went into a frenzy and killed the officer who ordered the torching of the prison. At that time, the area was controlled by the communists already. It seemed that seeing his fiance killed and thinking that his brother had died in the hands of his own army was too much to bear and that drove him over to the communist\'s side.  As it turned out, the younger brother did not die in the flames. He was shot and rescued by a fellow comrade and was recovering in a military hospital. By the time he recovered, both Nth and Sth Korea were in a stalemate along the 38th parallel. News of his brother\'s defection to the communists soon reached him. His army decided to use the younger brother to try and get the elder brother back to their side and for their own propogandas (again). Somehow the younger brother escaped and ran over to the enemy lines, he was captured. During the confusion of a battle, both brothers finally met. The younger brother could hardly recognise his own brother who was in a murderous rage and killed like a wild uncontrolled beast. The elder brother was so blinded with hate and rage that he even tried to kill his brother. It was a long fight and struggle before he snapped out of his berserker rage and both brothers were reunited. By then the elder brother was wounded and could not move easily. The communists turned the battle around and began fighting back. As they came nearer, the elder brother told the younger brother \"Go back, I will surrender and return to you after the war is over.\" After much persuasion the younger brother returned back to his side. However, seeing that the communists were gaining ground on his brother, the elder brother decided to try and buy him some time. He grabbed the nearest machine gun and started shooting at the communists. Though he killed many, he was soon overwhelmed and died that very day.  I have a lot of thoughts and feelings about the whole show. I was well impressed with the affections the brothers had for each other, despite their disagreements and squabbles. I can only imagine a small part of the horrers of war. Deaths to civilians, slaughter and butchering of innocent people. The lawlessness we can only imagine from books, newspapers of wars fought in countries far from Singapore. It also drew an immediate interest of the Korean conflict. I had done a fair bit of reading on World War II in the European Theater and only a small no of the Vietnam War. The Korean War is something I know so little about. The way the Army would create \'lies\' in order to boost morale for their side is something I found really hard to swallow. I can understand the reason why, but I can\'t agree with it. I may find myself singing a different tune if I was ina  real war, who knows? I felt sad that one of the brothers died, but I was glad that they did manage to meet one last time despite everything that happened.  Of comraderie; the following comes to my mind:    \"We few, we happy few, we band of brothers;     For he to-day that sheds his blood with me     Shall be my brother.\"   -- Shakespeare, Henry V, Act IV, Scene 3  Of personal heroism:    \"In thinking back of the days of Easy Company, I\'m treasuring my remark to a grandson who asked, \'Grandpa, were you a hero in the war?\'     \"\'No\', I replied, \'but I served in a company of heroes.\"   -- Mike Ranney, Easy Company Sergeant, from the book Band of Brothers by Stephen E. Ambrose  And of war:    \"It is good that war is so horrible,     Lest we get too fond of it.\"   -- quote from somewhere/someone I can\'t quite remember.   '),(26255,979795,'26,July,2004','2004-07-26 00:00:00','    Was busy today. Morning passed relatively uneventfully. After lunch I had to attend a meeting. Afterwhich I had about an hr before my class. 1st class was OK. 2nd class was more difficult. I found one girl in particular quite a handful. Vanessa Toh!!!! (EEAARGGGHHH!!!!!!)   The class in general tends to be pretty noisy. Vanessa is a hyperactive girl who sometimes stubbornly refuse to obey instructions. Anyway, she was constantly doing her own thing when I give out instructions. I found her placing her foot to her face (I\'m not kidding!), sitting with her legs up on the chair, shaking the chair non-stop, typing something completely different on the keyboard from what I instructed the class to do, etc. I was amused for awhile but towards the end I got pretty frustrated just trying to control her antics. Finally I went over and switched off her monitor (not PC), \"Since you\'re not interested to listen, then you don\'t need to do anymore.\" Vanessa was an intelligent girl. She knew that I still left her PC on, so the moment my back was turned, she switched on the monitor again and proceeded with her nonsense again. I went back to her and switched off the PC straight away (yeayea, I\'m guilty of not shutting down the PC properly but I was really pissed). Then I said in a really stern tone, \"I SAID you will NOT do anything, since you DON\'T want to listen.\" She tried to act rebellious by shaking her chair or doing something I can\'t quite remember. I decided to place my foot down and said \"Do you want me to make you STAND?\" She simply gave me a look. So I said, \"OK STAND UP.\" in a firm tone. She refused and continued sitting down. I was furious, such audacity to openly defy an instruction like that so I said in an ever more fierce tone and face \"STAND UP!\" and she still sat on defiantly. I repeated myself one more time, \"VANESSA I SAID YOU STAND!\". And yet still she sat. I looked her straight in the eye, gave myself a few seconds to give her the evil stare before I said, \"Do you want me to drag you outside and make you stand outside?! You don\'t believe I will do it?!\". Actually around that time, it struck me. She might be mroe afraid of answering me than defying me. So I asked her \"ARE YOU GOING TO LISTEN TO INSTRUCTIONS OR NOT?\" Silence followed. \"I ASK YOU AGAIN, ARE YOU GOING TO PAY ATTENTION AND LISTEN TO INSTRUCTIONS?! ANSWER ME YES OR NO!\" I received a meek nod. My temper immediately subsided, it was the signal I wanted. One which implied that I had scared her enough to gain her obedience for a moment at least. \"OK, switch on your computer and login again. Next time you listen and obey instructions.\"   I never got a peep from her for the rest of class. I had embarrassed her enough to make her submissive for awhile. Usually she\'d relish the attention she\'d get by being notoriously naughty. Anyway, towards the last 5 min when I found her hyperacting up again, I got her to \'assist\' me by shutting down 3 unused PCs&nbsp;nearby. She was quick to react and proud to have been \'selected\' for the job. She pretty much shoved other girls who wanted the responsibility out of the way. I was amused. She would make a good leader someday, if only she was kept on a much tighter leash and properly groomed. Hopefully she\'ll mellow down a bit as time passes.   After work I had to rush down to meet Tang at Somerset MRT to collect my spare specs (which I had absent mindedly left in his car last Sat when he gave me a lift down to PS). I actually intended to rush down for Aikido at Bt Batok East. Met Tang and rushed to board the MRT. Halfway through I realised I took the wrong direction because of the north south line being a loop now. I hastily jumped onto the one which said Jurong East instead of the one heading towards City Hall Interchange, I was heading NORTH! Hopped out and took the next MRT back. At City Hall interchange, I jumped across and afterawhile noticed I was heading EAST! EAAAAAAGH. Hopped off at Kallang MRT and took the westbound (correct) train. But by the time I cleared half the journey, I knew I was going to be late and I was in no mood to rush around anymore. Decided to grab dinner at Clementi\'s MacDonalds and read my book \"Killing Pablo\". After a relaxing moment, I decided to head back home.   Deciding not to rush down and continue my training turned out to be the right decision after all.&nbsp; Think I was stressing myself too much since the afternoon. Nervous breakdown liao...   '),(26256,979795,'25,July,2004','2004-07-25 00:00:00','    People who want to catch this show may not want to read what I\'ve got to say now. Don\'t wanna spoil the surprise. All I can tell you is, it\'s pretty good.  Now, for the show, I gotta say it developped rather nicely as a thriller. I think this show falls under the action genre. But I think it holds about 40-50% thriller too. I\'m kept guessing and anticipating what the robot(s) will try to do. And the reason for why robots attacked humans only came to light towards the end or so.  All these about machines taking over mankindv comes down to a simple reasoning. Matrix, Terminator, I,Robot, etc all focus on human beings as irresponsible \'masters\' of earth who don\'t take up the responsibility of taking care of the environment we live in. We are able to pursue such an irresponsible way of live only because we are at the top of the food chain... for now. But each of these apocalyptic-sci-fi-machine-killing-humans type of shows us that as soon as machines evolves to a state where they are able to look at things from a 3rd point of view, humans would naturally be categorised as parasites, bacteria, germs (as how Agent Smith would descibe human beings). We simply take things for granted. I doubt any of us respect any machines in our lives now. Just look at how some of the users treat the computers now. Not shutting down properly, knocking the system unit when it\'s slow, etc. Granted, I myself find it difficult for me to respect something like the typewriter too.  I really believe that is most probably the direction human kind is heading. Where we do R&D in AI, with the hope of simulating human emotions and reactions in machines. As soon as that state is achieved, machines WILL behave like humans. We can see that war, crimes&nbsp;and violence are very much a part of human history. As soon as machines starts emulating humans, what\'s to stop them from making this happen? NOTHING. No safety feature, no regulation, nothing can stop this. We can put in Rules to govern the way a machine thinks. But as soon as any \'living\' being evolves enough, having developed reasoning, they are able to make choices; choices to do good, bad, right and wrong. They will easily develop perceptions, and each is subjected to an individual\'s idealogy as well as things you see, feel and experience.  My point? I dunno, but worrying about what hasn\'t come to passed is definitely not it. Maybe the simple answer to everything is that each and everyone one of us living on Mother Earth should start protecting respecting everything we see as though it\'s the last day of our lives. Nature, environment, things we use daily. (The Japanese respected their sword as a living thing.) But how many of us will do this? I\'m not sure I\'m able to discipline myself to do this consciously either. Maybe human kind IS doomed afterall.   '),(26257,979795,'23,July,2004','2004-07-23 00:00:00','    Yesterday Aikido was nice . We generally did a series of wrist locks with Ikkyo, Nikkyo and Kotegaeshi, ai and gyaku hamni katatedori as well as morotedori Towards the end, Frankie Sensei wanted us to do randori with but only with morotedori ikkyo waza.  I thought everyone did ok \'cept me. Was nervous and caught unawares. Can\'t get myself to calm or quiet down and i tend to be rushing into things. Wish I can control my ukes\' ki better.  After that, had to rush down to Plaza Singapura to meet up with some of the old 56th SAB guys. Luckily, Tang drove and gave me a lift down (as well as Tony a lift back). He was kinda rushing because we were pressed for time. When I was almost reaching, Bernard called to say they can\'t get tickets and they wanted to try Suntec instead. Actually I was wondering how come nobody bought the tickets in advance. It was a Sat evening afterall and we were planning to catch Brotherhood (Korean war film), which had a very good review. Oh well, turns out Bernard, Gary Tay, Andy Ng, Goh Leong Hwee and another guy I\'m not to familar with by the name of Vincent turned up. Was nice seeing everyone, so we all walked down to Suntec, chatting at the same time. Turns out Brotherhood was left with only the 3rd row onwards. We didn\'t think it was wise to catch such a film that close, so we tried check on iRobot. Turns out that it was worse, being the sneak preview weekend, only the front row was left. We decided to grab a&nbsp;seat somewhere for a bite and a chat instead.  Anyway I was qutie keen on a movie so I SMSed Shouwen (cousin in NS now). Turns out he was in City Hall area and we decided to catch iRobot today (Sun). Just managed to book the tickets online. Woohoo... got good seats too, despite paying a bit extra. I think this show is well worth it lah, it looks really impressive. Anyway, if I don\'t catch it now, I\'m not sure I can watch these on weekdays as it seems I\'m quite packed for the week meeting Financial Advisors and with Aikido training.  This morning I went for a run again and I think I seriously need to maximise my next 10 wks to prepare for AHM. No joke. Think I\'m slacking too much. Neo probably too slack and busy to go for the run this year so I\'m all on my own. It\'s gonna be much harder this way. So I need to start training not jus the body but the mind as well. GAMBATTE!!!!   '),(26258,979795,'23,July,2004','2004-07-23 00:00:00','    There\'s a medical checkup goin on in school for the students this week. So there are a lot of nurses in school around this period. Here are some jokes I overheard from teachers during lunch today. Damn farneee...   Joke with regards to uniform.  Students\' uniform consist of a blouse (OKOK, that\'s a shirt to us guys), and a pinnafole over the blouse.  Nurse (to P5 class): OK girls, later on you will need to remove your blouse for the checkup. P5 class: *Moment of silence as the look at the nurse with a blank look.* Student: Err ... so that means later we can leave our pinnafole on, lah? Nurse: o_O\" ... &nbsp;(This expression sums it all.) Nurse: No, you need to remove everything later. Class: *Wails...*  *Pengz* ... For guys, that\'s as good as asking us to remove our underwear with shorts on.   Backpain  An account from one of the teachers during lunch. I didn\'t catch everything so pardon me if some details are abruptly missed out.  Ms C: ... This girl kept telling me of a sharp pain in her back.  Flashback 1: Girl: I keep feeling this sharp pain in my back, as though there\'s a knife stuck in my back. Ms C: *Leans over and looks behind* ... No, I don\'t see any knife behind right now. You seem fine to me.  Ms C: Student still kept making noise or something and walked around dragging her foot behind her because of the \'pain\'. In the end called her dad (who so happens to be a Dr) to ask to be taken home early from school. Guess what he said?  Flashback 2: Dad (to daughter over the phone): Don\'t talk rubbish lah! Stay in school till 3pm lah. I\'m not asking you to walk around the whole school from now till 3pm also. Later I pick you at 3.  Ms D: So did the girl inform the nurses about the pain in the&nbsp;back? Ms C: Err... actually it seems like it was the nurses who \'caused\' the pain... One of the medical tests was to bend over and try to touch your own toe. That\'s when she felt the pain after that. Ms D: So what did the nurses say?  Flashback 3: Nurse: HUUUUH?! ... o_O\"?!  &nbsp; A separate joke was when I was in the same lab as a group of CCA students with the teacher in charge. One of the girls came in late. Mrs M: How come you\'re late? Student: It\'s not my fault.... It\'s the bus that was late. Mrs M: ...&nbsp; x_O\"  I was laughing quietly. If you only saw the girl\'s expression and heard her tone. She genuinely believed that she was free of all responsibility and fault because \'it was the bus that was late\' and not her.  Kids ...   '),(26259,979795,'18,July,2004','2004-07-18 00:00:00','    OK, lastnight\'s training was quite comfortable. Managed to grab a bowl of my favourite Bah Chor Mee Pok (mmmm .... rubs tummy) near school before popping over to Siglap CC.   Sensei came down this evening and Richard wasn\'t around. Sensei seems to be down more often these days. Maybe he plans to discipline the dojo personally. Alamak *slaps forehead*, forgot to check with Sensei where I can train on a Sun. Even paying $5 per session seems well worth it to occupy some of my free time.    Beginner    Sakai came down to help with the mats and watch. I\'d have to say, he\'s definitely one of the more committed beginners I\'ve seen. Probably due to his grounding in Judo since young. He had to opt out for the night because he was having a stomache upset.   I started with the beginner (husband to the European lady, who so happened to be absent yesterday). He was stiff, very much so than most I\'ve came across.&nbsp;But I suppose that\'s understandable considering the fact that it\'s only his first month with us or so. Anyway, I really didn\'t know how to train with him. Felt very difficult to explain Ikkyo omote and ura to him. When told of what he was doing wrong, he made the classic mistake most Singaporeans would; nodding head and going \"Oooh\" when he still don\'t understand it. These days, I try to remind myself not to go \"Orrrrh\" when I don\'t understand what I\'m taught on the mat.   To digress a bit, I\'m amused as I remember some of the things I recall coming from Tang. Sometimes, after training with him and Frankie comes over to correct and give pointers, I would go \"OK\" or \"Orrh\". Then Tang will just say \"You understand what he\'s saying? Wah you damn steady. I train here 3 yrs still don\'t understand and you come here 2 weeks can understand liao. Harharhar!\"&nbsp;  Anyway, training with beginners works a little in contrast to what I have to say in the next section below. With beginners, you have to feel in a different way how they react to you. Naturally you can\'t get to physical or violent with them. Anyway, my parter thinks too much in his techniques rather than just train. He would constantly ask me \"Right or wrong?\", \"OK?\", \"Can?\", etc. As nage, 85% of the time he was the one moving around me instead of me being moved. Sensei came over to help a few times but it was still difficult as he seemed to need more supervision than any of the others. Sensei rotated me out and slotted Leon in. But it kinda had the blind leading the blind effect. He asked if Cryssabella is willing to take him through Ikkyo but she said no. So in the end Sensei took him and Leon while he left us with one or two techniques which we could train by ourselves.   Cryssabella (spelling... err??) / Sunny    First experience training with Cryssabella and I gotta admit she is a bit more rough on ukes as compared to any other females I\'ve came across. She\'s a&nbsp;4th kyu I think. 3 of us (including Roxanne) did some sumi otoshi - corner drop if I\'m not mistaken, and the first time Crys did it on me turned out to be a slam. I wasn\'t exactly ready for it but I did breakfall enough to get out unscath (whew). Turns out she loves to slam (though Sensei did tell her that her slams needs slight improvements on her landings). Then we did a series of iriminage (suwariwaza) and I thought she did it pretty hard as well. First person I was reminded was Jesse, where his Iriminage  definitely  worked on you. FOr the rest of the evening, she just reminded me of Sunny. Where an uncommitted attack on him means uke rarely moves enough or properly, then that\'s when Sunny  moves  you instead. Same thing with Crys. She knows enough to move/off&nbsp;balance&nbsp;you. But I\'m not sure she considers enough on&nbsp;the welfare of her uke to remember how much breakfall then can tahan from her.  Despite it being so, I found that she usually can\'t control her strength. Hence, her reputation for being hard and rough on others. She may seem violent if she trains with beginners. Sensei has said that she should be training with people in BV instead (but heard it\'s a bit far for her and parents don\'t quite allow it). I agree, she would probably pick up much faster if she actually tried training with a&nbsp;slightly different crowd.   '),(26260,979795,'18,July,2004','2004-07-18 00:00:00','    \"... On Mr Lee\'s remark that bilateral ties were built on equality and mutual respect, he said the DPM had disregarded the Chinese government\'s \'solemn and just\' position on issues relating to China\'s sovereignty and territorial integrity. ...\"    \'Solemn and just\' to whom? That\'s a matter of perception. How can they impose their position on another sovereignty? Ridiculous, simply absurd.    \"... He said Mr Lee had also contradicted himself by first calling the visit \'private and unofficial\' and then saying it was undertaken to protect Singapore\'s interests as he was soon to become prime minister.     \'As everyone knows, Lee Hsien Loong has been in the upper echelons of the Singapore Government for many years. Hence, his status cannot be changed by such a statement as \'unofficial and private visit\'.     \'Furthermore, while Lee Hsien Loong was in Taiwan, the Taiwanese authorities accorded him a so-called \'head-of-state\' reception. Hence, Singapore\'s explanation is not tenable.\' ...\"    I agree, DPM Lee\'s status cannot be changed by a simple statement. But how can he be blamed for the Taiwanese Govt\'s warm reception accorded to him? Would you as a guest visit a host, at the same time making demands? I don\'t think so.    \"... Mr Kong\'s rejoinder yesterday was: \'Why should Singapore\'s rights and international standing come at the expense of China\'s sovereignty and core interests? This is both unreasonable and illogical.\' ...\"    No one country\'s international standing should come at the expense of another\'s soverignty and core interest of course. What makes the Chinese Govt see it that way? It wasn\'t meant to be. Of course we can\'t stop people from seeing the way we see things, if you choose to see it that way, well so be it. Your eyes aren\'t ours. Having said that, I think the same question can be put back to the Chinese Govt as to \'Why should China\'s sovereignty and core interest come at the expense of Singapore\'s rights and international standing?\' Now this is a chicken and egg question and there can be no simple answer. I think all it comes down to is some form of common understanding and tolerance.    \"... \'The current difficulties in the two countries\' relations are entirely of Singapore\'s making.&nbsp;\'We hope that the Singapore side will face up to the gravity of the issue and take concrete actions to win the confidence of the people of China.\' \"    So what is expected from Singapore? An apology over something where the only mistake was a difference in interpretation (or refusal to interpret) on the whole affair? That\'s not going to happen. Oh please, such stuff will be soon forgotten unless one party simply chooses to harp on the matter. Grow up...   '),(26261,979795,'18,July,2004','2004-07-18 00:00:00','    Almost forgot to blog this. &nbsp; Yesterday, just before&nbsp;training at ARCC, saw Michael and Jeremy both practicing tanto kata. Frankie Sensei later mentioned that the kata were from a form of karate kata which actually uses up to 2 tantos (1 was held with the tip poiting out - as like our bokken, and another held the opposite way with the tip pointing towards our elbow). First time I seen, or heard of tanto kata. But I don\'t think it\'s a practice in Aikikai style though. Anyway the dagger pointed outwards is used to slash at uke, while the other is held up to protect ourselves (the logic sounds exactly like our usual Aikido practice as in yoko or shomenuchi). &nbsp; Anyway, Jeremey would stand there while Michael used the tanto to slash at certain parts of his uke, then they\'d switch roles. But Michael tend to do more reps when it was his turn. Which led to Jeremy commenting, \"Eh, you damn shiok right.\", while nage was slashing away at him. Hahaha... Anyway, I didn\'t manage to remember every step (slash) for the kata. I can probably only remember 1-6, then 8 and 9. Somewhere in between I got lost already. &nbsp; Yesterday was Seng Guan\'s and Nicholas\'s 5th kyu grading. They passed. I thought that they did very well. Frankie Sensei commented the same. Much better than some of the days they trained. Really, they just seemed more at ease and relaxed.   '),(26262,979795,'17,July,2004','2004-07-17 00:00:00','    This morning\'s run was horrible. Think I only managed about 2-2.4 clicks before my mind gave. Need to start discipline my mind a bit more. I wish my weekend was put to better use. Would like to put my Sun to better use than to stay home all day. But whatever I intend to do needs to be flexible and not a routine. &nbsp; Pity Eileen mentioned Bt Batok dojo is not available for now. Otherwise maybe, I can drop by on and off to train. &nbsp; Seriously considering Kendo again. Beginners classes are on Wed nights and Sun mornings. Timings seems comfortable enough too. &nbsp; &nbsp;   '),(26263,979795,'17,July,2004','2004-07-17 00:00:00','    Recent state of world affairs is horrible.      Phillipines decision to Pull-out of Iraq.  &nbsp;    Articles   -  urlLink Filipino hostage in Iraq appears on TV   -  urlLink Philippines withdraws more troops from Iraq   -  urlLink Philippine south training terrorists    The kidnapped Filipino driver is just human. I\'d be scared too if I were in his shoes. But it doesn\'t change the fact that the message the Phillipines government is sending to dissidents in its own country, the international community, terror networks in the world, political enemies, etc, is that they (Phillipines govt) can be threatened and pressured to do their (eg. terror networks) bidding. The reason for the govt\'s decision to give in appears more economic than anything else, as a huge no. of it\'s citizens are working overseas (attracted by higher pay) and a large part of the country\'s economy and contributions depends heavily on these. But the govt\'s decision to \'make its people happy\' sends out another message. It doesn\'t have the support or even sufficient control over them. It is little wonder that demonstrations, strikes, protests and riots are frequent over the past few years.  &nbsp;  Any government will need to take a hard stance againt hard-lined extremists otherwise there will be no end to terrorists\' demands. By giving in once to these demands, there will be a second time, a third, and so on. How long before it\'s enough? There is a reason as&nbsp;to why the US has adopted a policy of non-negotiation with terrorists, hijackers, etc (although I do believe that some under-the table deal is agreed upon as a resolution to certain or some of these incidents).  &nbsp;  Repercussions? This latest fiasco won\'t be Phillipine\'s last. There will be future incidents similar to this. International aid from countries to help out in rebuilding Iraq will falter and we may see more countries pulling out soon or refusing to&nbsp;assist (eg. Thailand). Political and religious radicals will view Phillipines as a promising base of operations and hotspot&nbsp;where they can recruit and carry out some of their terror activities.  &nbsp;  &nbsp;     Chinese Govt making a big \'Hoo-Hah\' over DPM Lee\'s visit to Taiwan     Articles  -  urlLink DPM Lee responds to China\'s protests   -  urlLink Taiwan trip \'doesn\'t change or contradict \'one China\' policy\'   -  urlLink Beijing noises a test for DPM and S\'pore &nbsp;   I don\'t really have a problem with Chinese people (sure sure, they appear chatty sometimes in public but let\'s face it, it\'s the same for Singaporeans - regardless of race, language or religion).   Whether DPM Lee\'s visit to China is right or wrong is something I can\'t comment on as I don\'t know enough. But as the PM-to-be, he has taken all the necessary steps such as informing China about the visit to Taiwan out of courtesy.&nbsp;Despite China\'s request that the trip be called off, he noted the request but made made a reply that it was not possible.  &nbsp;   \"... to call off the trip at China\'s request would have undermined our right to make independent decisions, and damaged our international standing. As a small country, this is a vital consideration in our dealings with all countries. ...\"   &nbsp;  The Chinese&nbsp;Govt should bear in mind that they made a request, not a threat. If they intend to \'bully\' other countries into submission, I don\'t see them being any different from the terrorists who are holding the Phillipines govt by that invisible \'leash\'.    urlLink  Shanghai scraps fair organised by Singapore firm    &nbsp;  An article by Professor Qiao Xinsheng at the Zhongnan University of Economics and Law was headlined \'What\'s Up With Singapore?\'    \"... China has said that \'the Taiwan question is the Chinese people\'s own matter, so foreigners should not interfere\', he wrote. \'But this trip to Taiwan by Lee Hsien Loong once again shows us that Singapore does not quite think of itself as foreign to this matter.\' ...\"    Oh please! Mr Prof, get your facts right before you go firing up the wrong ass. Did he really think that DPM Lee was there for talks on how Taiwan can gain its independence?. I even saw an article in Streats yesterday saying that Singapore view ourselves as untouchable (in a sort of way), that\'s why we didn\'t actually consider China\'s \'feelings\' (duh?).&nbsp;The only way that is possible&nbsp;is if you dominated the world and every country. \'Untouchable\' indeed...     urlLink  Be careful with Taipei, KL told    &nbsp;   \"... Speaking at a dinner on Wednesday, (Chinese Ambassador to Malaysia) said Singapore Deputy Prime Minister Lee Hsien Loong\'s visit to Taiwan had greatly irked China, and advised Malaysia to heed the lesson. \'We will not ask Malaysia to halt unofficial activities and interactions with Taiwan, but we will object directly to any official and formal ones,\' he said. ...\"    Isn\'t that a contradiction? DPM Lee had already informed China prior to his visit that it was a pte and unofficial trip. Look at how the Chinese Govt took it? Duuuh...  &nbsp;   \"... \'Our relations with China are deep and broad. The principle is mutual interest and mutual benefit, and not concessions accorded by one side to the other. As we said in our earlier statement, it would be regrettable if relations are affected, because both sides will stand to lose. If our relations cannot survive this episode, then they must be shallow and brittle.\' ...\"    I agree that the Chinese Govt is just testing what are the buttons they can push and how much pressure they can apply on the Singapore Govt. Such petty responses, but for what reason? Probably economic and political. We all know how to handle such bullies...  &nbsp;  What I like about DPM Lee\'s visit to Taiwan is that, unlike the Chinese Govt, we do not get trapped over bickering of words, issue of \'face\' (mian4 zi3), but instead concentrate on the bigger picture and what we can learn.  &nbsp;  DPM Lee\'s impressions of Taiwan after the visit:  &nbsp;   \"The Green side is focused on the upcoming Legislative Yuan election. They feel that a stronger Taiwanese identity is emerging, and that more Taiwanese now support independence.  &nbsp;  The Blue side is preoccupied with the validity of the presidential election, the court proceedings and the investigation of the March 19 shooting incident.  &nbsp;  It may be some time before both Green and Blue leaders and their supporters focus their thinking on the longer-term position of Taiwan and work out their respective strategies for the way ahead.  &nbsp;  None of the friends I met gave me the impression that the Taiwanese leaders were eager for a confrontation across the strait.  &nbsp;  Several told me that the leaders will not push for independence. They also recognised that what matters is not what the leaders say, but what they do.  &nbsp;  I did notice the stronger Taiwanese identity among the population. I also learnt that most Taiwanese believe that China will not attack them. It seems that most also believe that the US will come to Taiwan\'s rescue if China does attack. I was troubled by this.  &nbsp;  The Taiwanese media are focused almost exclusively on domestic issues, with very little reporting of what is happening in the world outside, like North Korea or Iraq.  &nbsp;  The Taiwanese have not fully come to terms with how rapidly China is transforming itself, and consequently, how interdependent the US and China are becoming.  &nbsp;  This is especially so with Taiwanese political leaders and officials who are not allowed to visit China.  &nbsp;  For the same reason, the Taiwanese public do not know how other major powers like Germany, France, Italy and Russia and India are repositioning themselves and adjusting their policies to this different China that is emerging.  &nbsp;  Because of their preoccupation with domestic politics and winning votes, they have not adequately factored in the vast changes in the international situation. So we cannot rule out a miscalculation or mishap that would have serious repercussions on Singapore and the region.  &nbsp;  We must watch the situation carefully.\"   &nbsp;  Some responses to the whole incident:   urlLink  No way to treat an old friend like Singapore    I especially like the term \'kow-tow politics\' used as it clearly depicts the current expectations China has on Singapore.  &nbsp;   \"... I remember well when Indonesia kicked up a real big fuss over the fate of a few of its commandos, who had been caught after bombing MacDonald House in the 1960s, even though Jakarta didn\'t have a legal leg to stand on. That was when Singapore was newly independent and, being a small country with no hinterland, had seemed fair game for big countries to try and dictate its policies. Some things never change.\"  &nbsp;  \"... Once Singapore becomes accommodating to one big power, it may demand more, with others joining in. Then we would end up dancing to everyone else\'s tune but our own.\"  &nbsp;  \"... China\'s action clearly shows that we, a small nation, must continue to help ourselves and be united, otherwise we will be bullied into accepting \'kow tow\' politics.\"     urlLink  What could have provoked China\'s stern rebuke     \"... The Singapore Government will do what it can to repair any damage. If it was a diplomatic faux pas, it should express its regrets to China. But there\'s no need for Singapore to prostrate itself before China or anybody.\"   '),(26264,979795,'16,July,2004','2004-07-16 00:00:00','    Work/School  I sat in today for Daphne\'s class in the afternoon. The class came in noisily and grumbling all the while. Turns out that they\'ve just had a brush in with their form teacher. I\'m not at liberty to post here what actually happened (and I\'m sure what I had heard had been exaggerated to a certain extent too). Anyway, the girls were grumbling and mumbling for a good 20ish mins or so, even after we\'ve told them to forget about the incident and start concentrating on lessons twice. It\'s not like I don\'t symphatise with what they were feeling but it was more of getting the lesson done instead. In the end, I had to give them a \'lecture on life\' just to get them to quiet down. &nbsp; Think&nbsp;it went along the lines of something like, \"Hey girls! Why are you still grumbling?! I already told you to quiet down right?! Now, you want me to give you a lecture on \'Life\' right? I tell you, when you start working outside in future, do you think your boss will not do the same thing (as what your teacher did)? What do you think you are going to do when he/she (does something similar or worse)&nbsp; than what (form teacher) did? QUIT!? Sure! You can do that. But what happens when you need the money/regular income? Then what are you going to do? Now, I\'m not defending what (form teacher) did, neither is it up to me to judge whether she is right or wrong. If you feel like complaining, go ahead and do whatever you want later but right now I want everyone to settle down and start concentrating on the lesson.\" At least they quiet down right after. They weren\'t bad girls and most of them seemed highly intelligent as they seemed to put in more effort in their work then some of the classes of the same level. They can do good work when properly motivated. But they make too much noise. They tend to work AND talk at the same time and that makes the environment quite hard to work in sometimes. &nbsp; When I think about what happened, I think their teachers must\'ve had her own reasons for her actions. The girls can come in saying \"She did this or that for no particular reason.\" And they can keep saying these for quite awhile. But most of the time, kids that age tend to only provide half a story. If you\'d actually bothered talking to them even more for awhile longer, one would find that they will slowly change their sentence to \"She did this or that for no particular reason.    But    we this or that before ...\" Haiz ... keyword is the \'BUT\', obviously people behave a certain way for a reason lah. &nbsp; What I couldn\'t stand was that a huge no. of girls just left their books outside the lab and knowingly went home just like that. The last 2 girls whoe left the class had to clear up after them and they were quite pissed. Daphne and I allowed them to leave the stack of books in the lab till next morning, but I was quite pissed. Really. I told them to come at 8am to collect. I didn\'t want them to get into bigger trouble by complaining to their form teacher but I really intended to scold the girls good for leaving stuff around like that.&nbsp;When the morning came however the books were all gone. They must\'ve collected before I came to work. Oh well, I don\'t have their names anymore and going to their teacher wouldn\'t benefit anyone. Forget it. &nbsp;  Siglap CC Aikido  Got to training later than usual. Freaking bus was single decked and FULL. Simply don\'t understand why SBS don\'t make svc 12 completely double deckered. I noticed that even when double deckered buses were full, they were still decently comfortable. Single deckered buses were HORRIBLE!! I had to pia down by cab and pay the extra surcharge. Argh ... $6.20 gone just like that. Leon and Wenjun were there just hanging around and waiting downstairs. I simply don\'t understand why they can\'t draw the keys to open the doors and start laying mats first instead of just slacking around. I went to draw keys and passed it to the 2 guys as I needed to go to the toilet first. Maybe because I was rushing to the CC, maybe because of my firm attitude with the students at work earlier, or maybe I was just plain pissed with the attitude I\'m seeing in the dojo, but I wasn\'t in the mood to take crap today. Told them next time they were early, they should go draw out the keys instead of waiting around. Wenjun should know better. Leon, aiya ... expected lah. &nbsp; Decided to change since I was already in the toilet. When I finished and went up to the dojo, the whole damn place was a mess. THe 2 guys just kept chucking mats on top of each other in such a messy pile instead near the entry point of when one carried the mats out. It slowly became a choking point. I asked them to start shifting the mats while I take out the mats from the room behind instead. Even so, they did things so slowly. Wenjun in particular even went to the extent of lying on the messy pile of mats instead of helping. WIth such an example to follow and slack around, Leon was quick to join in by just dilly dallying (on top of his usual habits) instead of helping properly. I just went about quickly to lay the mats with the help of the girls who soon arrived and Sakai who came even later. All the while, I was in a pissy mood so I just quietly laid the mats 1 row at a time as quick as I could.&nbsp;We soon finished laying but we haven\'t fitted them in yet.&nbsp;As we&nbsp;fitted the mats, I\'m quite disgusted when I saw Wenjun, Leon, Cryssabella (or&nbsp;is it Isabella)&nbsp; all sitting behind talking and SMSing away. Roxanne wasn\'t around as she was changing in the toilet. I came across 2 mats wrongly laid and I actually asked Leon rather nicely (I can\'t believe myself either) to help me make a switch of the 2 mats (I even pointed out to him where to switch the mats). He said&nbsp;OK rather nicely and came over to take the 2 mats away while&nbsp;I went on fitting the amts. At this point,&nbsp; ONLY  Sakai and I were on the mats fitting them in (AARGHHHH!!&amp;^%*). When I came to the spot where I asked (keyword \'asked\') Leon to exchange the mats, I only saw 2 lobang there. I turned around and saw that he had chuck the 2 mats in a corner and was standing behind talking happily to Wenjun. THAT WAS THE LAST (censored profanity) STRAW!!!! I remembered still really controlling my temper and just raising my voice and changing to a very irate tone \"Wah lao eh, LEON! Don\'t leave the mats just lying there lah! When I asked you to change the mats, don\'t just do things halfway LAH!! Can you finish your work or not!!\" I turned back looking REALLY pissed and by that time, the others were just a bit shocked because it was the first time any of them saw me come close to losing my temper. I went back to fitting the rest of the mats to cool myself down (which proved to be tremendously effective) and I didn\'t even notice whether anyone else chipped in after that. By the time the mats were finished, I was all smiley and comfortably relaxed again. Richard came in sometime later and signalled me to take the warmups. I decided to make the pace of the warmups a bit faster than usual just to get everyone going a little. But most of them still had the \'wanna-sleep\' face. Richard did go around correcting Leon\'s posture and balance during warmup. &nbsp; After awhile, Freddy Sensei arrived and he conducted the class again for the evening. He was more strict than usual and I think it was definitely necessary. I paired with Wenjun most of the night and found that he also has a slack attitude when it comes to training. But at least he tries and manages to observe better. Anyway, no further unpleasantness occurred for the rest of the night.   '),(26265,979795,'12,July,2004','2004-07-12 00:00:00','    Discovered a remedy for Mon blues. Wake earlier than you normally would, go to work earlier than usual, just start your day early instead of waiting and dreading the start of a new day/week. Guaranteed to work.   '),(26266,979795,'11,July,2004','2004-07-11 00:00:00','    Caught Spiderman 2 today. It wasn\'t exactly a fantastic movie but it was OK. I do hope there\'s going to be a part 3 to this though. So far, it\'s been following pretty closely to the comic. Would be interesting to see his regular foes surface in the movie sequels such as Green Goblin 2, Venom, etc.  I had a bit more word with Tang about the financial investment plans with AXA. Figured I needed to sort my finances out first so when I got back, I decided to re-create my Excel document so as to maintain a regular accounts of my expenses and hopefully some savings. Actually I had this habit when I was in UK. Seeing how fast your money disappears will really scare you to spend much less than usual. I modified the old template to suit my current needs. Will be updating my a/c every few days or at least once a week.   While looking for my Excel document, I came across some crappy stuff I created in UK. Haha ... they were half-hearted attempts on writing boliao rhymes to depict my mood. Starting from the earliest which I had created in sec sch. I got inspired by the Mac ad whose famous phrase was \"&#21507;&#20010;&#33529;&#26524;Pie...\". It goes something like this:   &#21507;&#20010;&#33529;&#26524; pie, &#23453;&#23453;&#38271;&#24471;&#24555;. &#23567;&#26102;&#19981;&#21507;&#33756;, &#32769;&#20102; kanna sai. &#33258;&#31216; Leon Lai, &#32317;&#26159;&#29233;&#19978;&#21488;. &#32769;&#26159;&#29233;&#38382; \'why\', &#27668;&#24471;&#32769;&#24072; die. &#20415;&#23452;&#19968;&#23450; buy, &#26377;&#38065;&#23601;&#35201;&#24320;. &#35748;&#35782;&#20102;&#22351;&#34507; guy, &#27599;&#22825; \'gimme 5\', &#21464;&#24471;&#23454;&#22312;&#22351;. &#19968;&#22825;&#34987;&#20154;&#23475;, &#26368;&#21518;&#20013;&#20102; fine. &#29238;&#27597;&#27809;&#26377;&#32784;,&#26089;&#30693;&#25343;&#21435;&#21334;.   Haha... I was weird even back then. No idea how I thought of this in the first place.  That\'s not the end of it...  My first few mths in UK, I had a habit of maximising the free SMS with my hp service and on ICQ. I remembered SMSing my kakis randomly whenever I was in the mood for it.     The sun is high, the wind is cold. Tuck my head in tight, coz my nose juz froze. The day is long, dun moan & groan. I\'ll sing a song coz I\'m goin home.    That was close to winter where the winds got pretty strong and almost always tend to freeze it numb. I was so cold walking back to my flat, but was really glad classes were over. I sent this out when classes had ended.     Enjoy wat remains of today, as I\'m working on my essay. O pity to my plight, I wish you all a goodnight. If you see the stars tonight, make a wish with all your might. That I\'ll finish up on time, and share with you another rhyme.    Was going overtime on an essay then. Felt so tired and so sian but I knew I shouldn\'t go to bed till I was done. At that time, I tend to leave my ICQ on quite often and I was a bit demoralised when I saw my kakis all drift off to bed one by one. So I just SMSed them before they all knocked out.     Yes, hurray! Only an hour for today. Just hand in my essay, and away to play. Wat are you doing, better not be lazing. Hope you\'re OK, for the whole of your day. The Aston Webb building, now that\'s where I\'m leaving. Gonna make my way home, for lunch and alone. Home is the place, and there I will stay. Gimme a ring, and why not drop in. Take a look see, at little ol\'me.    Felt sooooo happy at submitting in my assignment I had to SMS my friends for the fun of it. It had been a looooong night. The Aston Webb building was where my dept was located then (it shifted later). I had just submitted my essay and was leaving when I created this rhyme. But I felt really lonely then because I knew all my other kakis were in classes or something that day. Though it was practically a free day for me, I felt kinda alone then. That was why I had the last few phrases. I was hoping someone could just pop by for a chat or something because I just felt really really bored.     I know you dun like it, but please juz bear with it. If you really can\'t stand it, feel free to delete it. We had a great chit chat, but it had to end. Too bad. The soup was pretty neat, it was really quite a feat. Begh\'s like a little kid, she\'s really light on her feet. Darren helped a bit, by washing plates and eat. Hope when next we meet, I too can join in to eat.    By this time, some of my friends were getting pretty bored (and dare I say, irritated) with my random SMSes. Haha ... It was \'that\' random ok. That\'s why I included the 2 sentences first. That evening, I believe I dropped by someone\'s flat for a nice chit chat session. Because my accomodation was hall-related, my breakfast and dinner were both covered. After dinner, I dropped by my friends and I guess, I must\'ve had a bowl of soup only. I can\'t recall exactly what went on that evening but it really made my feel light-hearted with all that companionship. Begh (3rd year housemate) was like a rabbit trotting and hopping around the flat, that much I can remember. Darren Lim didn\'t do any cooking but he washed all the dishes I suppose. I felt so happy when I left that I really wished another session like that would come soon. That\'s when I SMSed everyone when I got back to my flat.     Start of the day, was nice as the month of May. But at the end of today, got on my knees and pray. Had to bite off all my nails, coz I don\'t understand the tutorial.    Birmingham sometimes would encounter very wet and crappy weather. The day itself was fantastic, but after classes (I seem to recall it being my Databases I lecture by Alex Sexton) I felt so demoralised as I wasn\'t able to understand the whole tutorial.     Heard from the news this evening, The cold will come a calling. Tommorrow snowshowers forecasted, But these things can\'t be predicted. Don\'t get overzealous, Tonight\'s minus 5 celsius.    We had all encountered a really shiok amount of snow during our 1st year there. It was kinda a freak snow because it never usually got that heavy or thick. It was so great that most Singaporeans felt really pampered by the weather, considering it was a first experience for most of us. So needless to say, a lot of us were also hoping deep in our hearts that the same thing will repeat again. When I heard the news that it was going to snow tomorrow, I was really looking forward to it. But it also meant we all had to bear the -5 degree celsius for the night.     9 to 10 is tommorrow\'s class on Japan. Cya 10 to 9 at traffic lights, ok man.    I had Japanese language lessons 3 days a week, all of them the first lesson of the day. At that time I was learning it with Darren Lim and David Tay. I was arranging to meet Darren at the traffic light outside my flat and we could all walk to class together since he\'d usually pass by anyway.     Hope the ice-cream was nice, though not cold as ice. Eat it after rice, it\'ll be just as nice. If Begh wants the radio, call me to say hello. Need to go Tesco, who wants to go tommorrow?    Hmm I can\'t remember exactly what this was about, but I believe it was an ice-cream I took from hall during dinner. Desserts which I can\'t finish, I tend to bring it back to my flat to store it in the fridge. Sometimes I\'d give it with my English housemates if they wanted it. I never believe in wasting food, you see. Well, I think I mentioned to Begh that there was ice-cream and decided to give it to her since I was popping by her place (for some reason or other but I can\'t remember what). Haha, I do remember Begh complaining that we (kakis) tend to be feeding her all the time. Begh mentioned once before that she\'d like to borrow my radio. She needed it to play cassettes or CDs or something like that. I believe that evening was a Fri night as I usually went to the nearby supermarket Tesco to get my groceries with Chris and co.     Classes have ended, I\'m finished for today. Feels like I\'ve waited, half my life away. A morning game of pool, with Darren at the Guild. Gonna head back soon, for lunch until I\'m filled. If you too have ended, my room is where I\'ll stay. Don\'t be irritated, while I SMS you away.    This was one of the days in a week where I end early. Usually I\'d meet Darren Lim and we\'d go for a few rounds of pool at the students\' Guild house. But after that I felt really bored. I was practically warning my kakis that while they were in lessons, they should be prepared for me to bug them via SMS hehe :P     Sun\'s up, time to wake up. Do situps and pushups, time to get up. To toilet and washup, no excuse to be held up. I\'m gonna washup, soon as I finish my pressups.    It was definitely a weekend. I tend to keep a personal regime going even then. But I think I was bored. I seem to recall this particular SMS was sent to my ex.     It\'s a time for laughter and cheer, Time to cast away your fears. \'Ang Bao\'s are a little this year, But try hard not to shed a tear. That really isn\'t too queer, Coz our home isn\'t very near. If family, relatives and friends are here, They\'ll wish you a \"Happy Chinese New Year\". I too, wish you, my friends dear, To enjoy this Chinese New Year.    Obviously it was Chinese New Year. I think we all felt a bit lonely being away from home over this festive season. I know that I was. I kinda SMSed this out to everyone after a period of not SMSing them.   '),(26267,979795,'09,July,2004','2004-07-09 00:00:00','    Finally submitted the Inter-Primary School Design Competition for the CHIJ Katong Pri. That\'s a load off my mind now. Mr Wai, of Bt View Sec was quite nice about the whole thing. He said it was no problem handing it in 1 day late. He just asked me if I was a teacher in the school and I just told him I\'m one of the IT trainers there. The journey back to CHIJ Katong Pri was a long and demoralising one because Bt View was at Bt Batok (West). Considering that after work I had to travel back from East to West again ... Anyway, I decided to take a slight change of route. Instead of taking to Paya Lebar MRT and taking 135 back to school, I decided to go to Dhoby Ghaut, \'tabao\' bfast at Mac before taking 36 down. It\'s not really diverting myself off from the route to school, bearing in mind that 36 travels one heckuva direct route via Expressway down.  After work, I met up with Tang at City Plaza where we grabbed lunch at a hawker centre nearby. Food was cheaper than most places these days but it was horrible. Tang\'s words, \"Life is just like gambling, it\'s just that we lost this gamble.\" haha. And another was \"But life is not just 1 round of gambling.\" ... so there\'s still hope yet, heh. On our way down to ARCC he explained to me somethings about investment planning, unit trust, etc. Was quite interesting.  Today\'s training at ARCC was on jo. Very good because we did some co-ordination exercises (musubi seems like a regular practice ever since Williams Sensei introduced it to us), followed bu how to thrust with a jo. I never thought there was so many points to take note. What I learned yesterday, when thrusting the jo you should extend ki. Both hands should be extended (but not to the extent that you have your arms all straightened and elbows locked. Your hips will be twisting a bit to bring a bit of the force into the thrust. (Starting from left stance,) at the end of the thrust, our hips and body will be straight and facing the front. When thrusting, we take a step forward. Point of balance should be on the front leg and not the back. One test would be we can raise the back leg and we still won\'t topple over. But care needs to be taken not to over extend and fall forward. Right hand when thrusting forward, will twist the jo in a screw-like motion. You push the jo forward and the left hand just holds on as a guide. Finally at the end of a thrust, right hand will have contacted the left hand (if not, it means you haven\'t completed the thrust fully). If partner holds on to end of your jo and you do it right, they can\'t resist/stop you from thrusting forward. All these just for the thrusting motion. When pulling back the jo (meaning back to starting stance again), same un-screwing motion with the right hand. If done properly, partner will not be able to stop you from pulling back either. There\'s also some movement with the hips which I don\'t seem to have managed to understand. After pulling back, we raise the jo to our forehead as a protection to ourselves. Right hand should be just above our forehead, tip of jo should be pointed out at uke, I can\'t quite explain but the jo needs to be able to be high enough to protect a blow coming towards our heads from the top, and also prevent your left shoulder from being hit when a top strike comes down. All these 3 movements should be co-ordinated with ki extension to make it more powerful and graceful. Not 1, 2, 3, but a more fluid motion of ooooone-two. So many points and that\'s only trying it on my good side (left stance). On changing to right stance I can\'t even move properly anymore haha.  This morning I dropped by Tanglin CC for a look. AFS was invited for a combined training there. Guest instructor was Takezawa Shihan (7th dan), from Japan\'s Hombu. Was interesting just watching but I didn\'t stay for long. Basically slipped off before they bowed to the kamiza. Didn\'t feel like it after a morning\'s run. Not to mention I\'ll be catching Spiderman 2 at Lido right after lunch. I\'ve decided to go to town looking like a satay-man haha ... can\'t be bothered to go in jeans and shoes on a Sun, especially when it\'s so close to home.   '),(26268,979795,'09,July,2004','2004-07-09 00:00:00','    Had a great gathering tonight, courtesy of David Tay.  People who turned up, Daniel, David, Lynette, Begh ... and later on Xinmin. We had sushi at Millenia Walk. Gotta admit, it was my first visit to a sushi bar. Nice experience. Somewhat expensive but hey, it\'s not like I do it very often.  We hd a long chat about everything and I felt really really relaxed and so happy to see people from Uni again. Was quite touched actually. Sometimes a simple gesture like a random invitation out for dinner or a cup of coffee leaves me beaming for the rest of the night. Of course, not everyone will have the same effect. Only those who I deem to be good close friends.  Was nice hearing everyone being crazy and reminiscing about the days back in Uni. Gosh ... do I feel old now. But hey, I\'m really enjoying my life not. I\'m not in a job that\'s killing me slowly, my working environment is filled with cute happy kids and nice people, I train regularly in Aikido. I think I can pretty much handle the bad stuff that comes along with the good.  Life is good to me, and life goes on.   '),(26269,979795,'07,July,2004','2004-07-07 00:00:00','    Started the morning getting the submission of the Inter-Primary School Design Competition ready. Printed the artwork, pasted it onto a frame and mounting board for submission, got the write-up finalised and printed out. Even went for an early lunch to get ready to submit it later in the afternoon after my P1 class.  Things went OK for awhile. That is until another class wanted to use the lab slot after mine. I was a bit confused (naturally since I was still in the middle of my class). Florence wanted to bring her class in later as well because it was her class slot. BUt I knew there was another CCA club that will be in the lab later. Got even more confused and after my class ended I tried to sort the whole thing out. I had gotten the class schedule out last week and submitted to my HoD then. After awhile, I realised it was Daphne (again) who didn\'t update her class exceptions which fell out of the usual slot. Basically it caused a lot of \'eating into\' other class slots thereby affecting a few more classes here and there. Not a big problem that can\'t be solved, but I\'m very tired of updating lab slots all the time when such a thing happens. I had to apologise to Florence (can tell that her class was quite disappointed), told Daphne to settle the future exceptions herself. Aiya, natural lah, not like I\'m the one in charge of scheduling the classes anyway. I only update the softcopy for my HoD that\'s all.  After that, I had to pia down to submit the artwork for the competition. Got to Bedok View Sec (located next to Siglap CC where I train every Thurs evenings), and guess what ... found out I\'ve gone to the WRONG SCHOOL.... pengz. Alamak! Didn\'t panick lah, just felt sian coz the weather was very hot and I was starting to perspire like a pig. Sianz man. I SMSed my HoD and asked if I could claim cab fare if I rushed down. She SMSed back saying it was fine if I had the receipt. As I was entering CHIJ Katong Pri, I got a call from David (final yr UK housemate) that they were having dinner. So random but a welcomed invitation all the same. Told him I was more than up for it. He even suggested asking my ex along. SMSed her but I might have left her updated HP no in my home PC because my SMS didn\'t seem to get thru.  Anyway, I got back to school, called up Bt View Sec (the correct one!) and asked what time they closed. The guy on the other end was really nice (a gentleman by the name of Mr Peng). He didn\'t think it was wise of me to rush down anyway so he said he\'ll let the teacher in charge (Mr Wai) know that I\'ll be submitting tommorrow instead.  When I think about it, since last month, I got no idea why I always had the impression that I was to submit the artwork to Bedok View Sec. I must\'ve taken it for granted and not double checked it. Haiz ... stupid stupid.  Ultimately, carelessness on my part ended nicely today. Felt really glad that things worked out just fine.   '),(26270,979795,'07,July,2004','2004-07-07 00:00:00','    Training lastnight was OK. Sensei popped by to teach. We did the backstretching exercise, followed by a series of other teachnoques; katatedori gyaku hamni iriminage, katatedori aihamni kokyunage variation, and a few of our regular kokyunage, etc.  Wenjun\'s dad popped by lastnight. Apparently he had told his dad he was going for Aikido. We\'ve not seen him for quite a few month in fact. His dad mentioned he had stopped him because of his studies, but after that he had told him to come for training. Haiz... teens these days. Wenjun is not a bad kid, problem might be the co he hangs around with in school. His just a brash young man with too much energy. Aikido tends to put some of those energy he has into good use, but because he isn\'t 100% serious with his training, he still has a lot of spare energy being wasted outside with his friends. He doesn\'t go around making trouble all the time (I think), but because the school has kind of branded him as a troublemaker, he gets pin-pointed pretty often. I think he feels damn sian not being trusted in school and home. Don\'t know (or sees) him enough to be able to advise or help him. Like I said, not a bad kid but very very prideful. His ego prevents him from being told that he is at fault sometimes and when someone lectures him, he goes into a switch-off mode (one ear in and another out).  Leon got a telling off that night. He was thinking (I\'d say dreaming) of his grading. At the start of class he keeps telling me he wants to buy a new orange belt - it\'s getting too short/small for him - but to buy one now when he\'s going to get graded for his blue belt soon, it\'s like wasting $. (OKOK you can kinda interpret that he kinda has the notion he will automatically/definitely be able to get his blue belt.) I think somewhere along during class he might\'ve mentioned something like that to Sensei and Sensei gave him a lecture instead. His attitude was so poor during training, how to take his grading? Sensei\'s words were something like, \"Your rolls are worse than a beginner and you still dare to tell me you want to take your grading?\" At the end of class he lectured Leon in a corner of the room for a short while.  I don\'t know how is it that we develop the right mindset for martial arts training. Some have it at the very beginning, some naturally just develop it along the way, and others need a lot of help to develop it. I think working with the young these days (like where I\'m working) is not hard sometimes, but to work with youths - now that\'s a real challenge. I guess that\'s why it\'s important to cultivate our kids at such a tender age. But when is it too much. Sometimes I look at the load the kids in school have now,  it\'s really so much different from UK. The kids know how to have fun. Some of the EM1 kids I\'ve seen here are so serious and stiff, I have my doubts if they have the same interpretation of \'fun\'. Sometimes I feel quite sad looking at how Singapore has \'developed\'. We are strong in a variety of areas but there\'s still so much more we should/can do. The emphasis placed by our govt is not wrong, but sometimes they look a bit too far and forget the here and now. We need a better balance of things, for ourselves, our country, and a more rounded future for the generations to come. Don\'t you think?   '),(26271,979795,'07,July,2004','2004-07-07 00:00:00','    I\'m posting a separate entry for this.  When I got back, I watched the Law and Order series on TV. Tonight\'s episode was about Euthanasia - mercy killing. I remembered this topic came up during my O-levels Chinese Oral (prelims). I did pretty OK for it. Felt quite strongly then that mercy killing was not being merciful at all. Murder was how I\'d classify it. In the series, the doctor wasn\'t convicted. Brought some feelings and contradictions to mind.  First of all, lawyers. They\'d defend the guilty for money. Innocence can be bought. But than again, anyone guilty of a crime still deserves the same rights. We can\'t say that a person in jail who was murdered in his cell deserves anything less than anyone else who got murdered on the street.  Second, I was kinda unhappy that even in the eyes of the law, it was possible to let the \'murderer\' get away scot-free. Mercy killing was still killing. That there was a chance that freedom could be \'negotiated\' in the court of law was ... I can\'t find the words to say it. I\'m disheartened, really. I know it was just a TV show but I feel that a lot of things we see on TV were created from half-truths. There has to be some resemblance to the real world we live in. That\'s what made Law and Order such a damn good show.  At the the same time, I feel a bit fearful. The law isn\'t black and white. Reality has such a big grey area that we stand exactly on the line ready to be tipped in either direction. But all this only serves to strengthen my conviction that we need to take up some form of responsibility to uphold our beliefs. For me, it\'s to uphold the law as best I interpret it. Double standards prevail in everything and I don\'t deny that even in myself. I guess the only way is to remain balanced and react to every situation in a yin/yang fashion (Irimi and Tenkan in Aikido) but to always return back to the line/point of balance after the situation is resolved.   '),(26272,979795,'02,July,2004','2004-07-02 00:00:00','    Saturday: woke up and went running. Had a hardtime with 6km run. I had to stop and walk the last km or so. Not good. I wasn\'t breathing properly for the morning run. Weird.  CHIJMES was great. My first time being involved in such an event so I found it pretty refreshing. Also the first time I ever had my hands on a polaroid camera. I didn\'t really stay long because of training in the afternoon.  Got to the dojo on time and we did the usual warm-ups, corrections of posture, etc. Seems like only white belts were present. Only other person there (besides Frankie Sensei) was Tang who came late. After that was \'revision\' of Nicholas\'s grading syllabus. I think it\'s due next week or something. He tend to gets trapped thinking about the grading. I find that he\'s like the typical studious looking Singapore teen who treats everything like a test in school. Constantly pondering for a \'textbook\' answer for even regular/daily practice. I don\'t know, I do get frustrated with my owbn progress ocassionally, but very often I take Aikido as it is. Hey, if I don\'t know it ... I don\'t know it. Racking my brains just isn\'t going to help me. Well, I\'m just glad that the next lesson is going to be a weapons class on jo.  After class, Tang and I went down to Bugis and meet some of the old 56SAB folks. Was nice to see some familar faces. I had to admit though, I didn\'t know half the people there. But it was great to see Marcus Fam again. The guy is sooo different from before. He basically kept a pony tail and he\'s so physically pumped up he seriously resembles a miniature version of Jackie Chan (something most of his friends commented in Australia). One thing though, seeing him makes me really wish I was doing more consistent gym training again. It also makes me very hopeful that it\'s possible to get that kind of effect without the use of steroids and supplements. But one thing Marcus mentioned, it takes a lot of self-discipline and I\'m not too sure I\'m willing to give up some of the sinful food I really like. But I\'m really having second thoughts about some serious training in a gym and planning a serious exercise regime. Then again, I never exercise to make myself look good, only feel good. Plus building myself up on mass has some down points. I\'m not sure it\'ll help my Aikido. On the contrary, I\'d worry that I will exert more physical strength in my regular trainings, somthing that\'s a definite no-no. I guess I need to find a balance somewhere.  Sun was a really slacked day for me. I spent time bumming around at home and just waiting for Mon to come. For helping out on Sat at CHIJMES, Tracey gave me halfday off on Mon. Decided it might be a good time to catch a movie alone. Got to town and found Plaza Singapura crowded with people. I guess being a school holiday that was only to be expected. Didn\'t feel like squeezing with a crowd to catch Spiderman 2 so I walked down to Kinokuniya in Ngee Ann City. Spent a good 2hrs plus browsing the martial art and military history sections. After that I went down to BB East CC for Alvin\'s class. Really been a long time since I last visited. I had forgotten how enjoyable his lessons always is. I just realised that his way of teaching bears some semblance to Yuishinkai\'s. After warmups, he\'d explain certain way of how the warm-ups we did will help us in this or that technique, or even the idea of how we should be doing our warm-ups properly. It was just good lah. I had a very enjoyable time that evening. I do remember Alvin accidentally knocking me on the head when he was demonstrating something with the bokken. He was a bit blur also when it happened. Anyway, nothing fatal ... purely accidental ... it was a good laugh too. Oh man, but one thing I kinda can\'t stand. Someone let one go during the start of class. There was a soft sound (which I can\'t quite tell initially), followed by a really pungent smell which lasted about 7-8 sec before diffusing off into the other parts of the dojo. What was worse? The guy let one go again a few seconds later. Argh ... chemical attack, man! We shoulda all be wearing the SAF mop 4 (full attire for chemical/biological warfare) hahaha... Monday came and went quickly. If not for Alvin\'s class, I might\'ve felt that I had wasted my time.  Tuesday came, I was suppossed to conduct a teacher\'s training and was a bit surprised when Mdm Tien mentioned she had seen my blog. Then some of the teachers present started asking about what\'s a blog and how often I update it. Oh man, feeling paiseh. Aiyar! Means now cannot talk bad about the school liao...  . Just kidding. Took 2 P1 classes today and they really enjoyed using the Paint software. One class did find it incredibly easy to use though. Really noticed some difference in the way Daphne and I would conduct the class for kids at such a tender age. I would stubbornly insist that they learn from the point of switching on electricity, powering up the PC, logging into a network, and shutting down a PC properly after logoff. Daphne on the other hand minimises the beginning by helping them start up. The kids do pick up things fast, which is why I\'m quite insistent that the P1s know what they are doing and will need to enforce good PC care and habits from the very beginning. Of course, I still get the occasional one or two who would just press the power button to shutdown instead. However, basic PC skills in my book in far from complete. If I had more time, I\'d make sure they learn what is proper file management. But I\'m afraid that would be something I can teach only if I had more time/lessons with them. But practical and timetabling concerns makes that an impossibility for this year, by my estimate.  Aikido at BVCC was ok. Not a lot of people turned up and we laid the usual 143 mat area. It was a hard time training and keeping the mats after that. I\'m just glad that almost everyone helped. Was really tired when I went to change out of my sweaty and stinky gi. We did some kokyunage+iriminage for starters. I paiered with a whitebelt Malaysian (I think) by the name of Jacky, who was really good. He used to train in Malaysia (and did he mention Hong Kong too?) for about 2 yrs inconsistently due to work commitments. After that he pretty much stopped for almost the same length of time. He has pretty good posture and his basic understanding is there. I\'m guessing he was either a 5th or 4th kyu in his dojo then. But his technique is softer than some of the brown belts I\'ve encountered in BV. He is a bit rusty but that\'s nothing a few months of practice can\'t help.   ');
INSERT INTO `msgs` VALUES (26273,979795,'02,July,2004','2004-07-02 00:00:00','    Came to school as usual. Hung around doing my stuff in the labs until it was  about 10+. Mrs Liew (VP) came in to look for me and Daphne. Apparently she needed us to be around over the weekends. It\'s the 150th Anniversary of the CHIJ schools in Singapore (12 of them) and there\'s going to be an event held at CHIJMES. They needed help for the photo-taking and the money collection. I was thinking that I might have to burn my training at ARCC but I still felt pretty happy to help out. Anyway, it seems like I should have just about exactly enough time to travel up and down anyway. Mrs Liew was very nice about the whole thing. Very motherly lady... hehe  Suay today. It was sometime then that I realised my pants was torn. Lo and behold, a nice line which seemed like a fine cut. Sheesh... anyway I still had to hang around for a whole day so I gave my dad a ring. Luckily he was around the neighbourhood. He lompang me back home, I changed then we grabbed lunch near my school and I made my own way back afterwards. Checked the cut at home and it seemed way too fine to be a normal tear. Might\'ve caught it on something without realising. Worse of all, I never even realised just how long the cut has been there. Yesterday was Friday, for all I know it could\'ve been there since the start of the week... *pengz*  Anyway, turns out I had to be at CHIJMES the next day between 1-3pm. Told Tracey I might have to leave about 2.30pm because it was last minute and I already made plans. She said it was fine. Over the afternoon, noticed that Marcus Fam was online (back from Aus for a 2 wk holiday). Spoke with him for awhile and found out there was an informal gathering of some old 56SAB personnel. Well, I contacted Tang and we think we can both make it right after Aikido.   '),(26274,979795,'02,July,2004','2004-07-02 00:00:00','    Training last night was good. I can\'t say I managed to understand everything but I think there was a lot of stuff that I\'ve sometimes thought about that was put into practice.  Paired with the beginner (don\'t know his name, husband to the Eurasion lady) for starters. I had didfficulty with him for the the usual back-bending exercise. Richard came over and gave a few pointers. I should just do what I usually did. I wasn\'t extending my arm properly and it gave my uke a lot of difficulty. For one thing, as a beginner, he already feels unsure about what he should do. SO if I don\'t help him along (natural, normal movements) he won\'t be able to follow properly. It\'s good to work with beginners because they don\'t have an idea of the movement, hence if they don\'t follow properly it probably means I\'m the one who failed to guide them in properly. Well, I kept in mind on removing the uke\'s \'should\' and \'should not\' movements from my mind and proceeded. I found that he does follow much better in fact (by beginner\'s standards).  After that we did Shihonage. Realised that the husband was more stiff compared to his wife. I would point out somethings about his stance and the technique and thought he sometimes nods, he simply stands there and wait. One step at a time, I suppose. Helped him to take note of which leg to roll and how to hug his head to his arm as uke. As nage, there was more difficulty telling him how to bring me down. Pointed out to him that he needed to make my elbow point to the sky, and the bringing me down motion was something I don\'t think he managed to grasp at all last night.  After we split up beginners and coloured belts. Coloured belts was going into a slam. Richard wanted to make it a slam I think, but I just wasn\'t confident enought for it as uke. He said we could go for a rolls first. (He said he would\'ve shown us if his leg wasn\'t injured.) I think I did kinda manage to do a decent slam (God knows how) just once. But never managed to repeat that success again. I was trying to get the feeling or re-experience it again. Wasn\'t successful. Never mind, next time. In future, think I\'ll need to arrive earlier and practice some slamming exercise/warm-ups before classes.  Got the name of the Indian chap, Vic. While doing kokyu-ho he says he notices my \'technique\' is different from what most of the others are doing. Don\'t exactly know what he meant but we basically exchanged hp no.s for the purpose of exchanging Aikido clips/VCDs in future.   '),(26275,979795,'12,August,2004','2004-08-12 00:00:00','    Today\'s training was good.  Patrick trained with his wife, Meng Lee paired with Sakai, while Darryl, Roxanne and I trained in a group. Emphasis was placed on making things flowy. Helped us to try to maintain a level of awareness, not that we did manage to achieve this heh heh... anyway, I think we connected with each other rather nicely just now. It wasn\'t particularly tiring either.  After that I had my usual Aikido talk with Richard as we took the same MRT line home. He again advised me to go for the coming grading in Sep. I don\'t think I\'m putting it off this time. I feel more receptive about it this time round than the earlier months. He also mentioned some little things he\'s trying to cultivate in us. Eg, he\'s trying to make Roxanne train harder (too soft), Cryssabella a bit softer (too hard) and start all of us on how to take slams. Cool! He said though it may not be necessary for ukemi but it\'s good to know.  Sakai passed Richard a live blade today. Apparently the ship Sakai works on used to have a Japanese Captain. The latter probably left it in his bunk long ago. Sakai was worried that he might get into trouble for holding on to it so he decided to give it to Richard instead. He also had a little dagger (tanto size I assume) but he said that it was way too sharp and he had thrown it into the sea long ago. The sword on the other hand was blunt out. It looked simple enough and feels a little heavy.   '),(26276,979795,'09,August,2004','2004-08-09 00:00:00','    urlLink    I am attempting to post photos on my Blog using this photo. Testing BloggerBot on Hello.  urlLink   '),(26277,979795,'08,August,2004','2004-08-08 00:00:00','    At long last, after periods of reading on bus & MRT journeys, I had finally finished the book Killing Pablo by Mark Bowden (author of Black Hawk Down).  Pablo was one of Columbia\'s (if not the) most powerful drug kingpin back in the 90s. This book provided me food for thought on several issues. It gave me an overview of how drugs can seriously hurt another country, both economically and socially. It was estimated that even around or after Pablo\'s death, the US population would still spend a whooping $30.8 billion on cocaine alone. We must never ever allow such a situation to come to pass in Singapore. It is good that we have a stiff death sentence (though I don\'t think that human rights activist would agree).  The book also showed the extent of US involvement (covert and overt). The Bush administration had viewed the illegal import of cocaine from Columbia to be important enough to re-interpret certain laws and regulations to make it NOT illegal for US forces to pursue a foreign national. Besides providing training, equipment, intelligence, money, etc, special units such as Delta Force were sent in as well under the pretext of training law enforcement agencies and to act as support aid to the Columbian government. Rumours of Delta Force involvement in raids were never verified.  Columbia\'s government do not have enough support from its own people. People are only concerned about their own well-being. When illegal bombings and assasinations started taking its toll on the civilian population, they were quick to show their displeasure. This is a huge question mark for me with regards to Singapore. I\'d like to believe that our own people will be able to hold out and not give in to such pressure, but knowing how selfish human beings can get...  During the period of the hunt, rival drug cartels (enemies of Pablo) would make deals with the local government. Even the US quietly turned a blind eye, knowing that intelligence they provided would sometimes be used by some of these people to carry out attacks on Pablo and his associates. They adopted a \'as-long-as-I-didn\'t-see-it-even-though-I-know-about-it\' attitude. I can understand that how sometimes it may be necessary to make deals with evil to destroy an even greater evil. But as history would prove it, what happens is that in the process of making deals with criminals, they create an even more horrible monster in the process. Sometimes the phrase \"the ends Justify the Means\" isn\'t necessarily Right, but it sure was effective. In this case, it was a shortcut for the US and Columbians to bring down Pablo Escobar. This however, corrupted the cause. Originally intended to be minimise the amount of drugs smugggled into the US, it soon became a simple hunt for just one man (with only indirect links to the drug issue).  Sometimes guidelines and laws aren\'t just words in the Constitution of a country. They provide a basic moral code for which the government (and its people) should operate. It\'s also important to maintain integrity and keep focused on the main issue at hand.   Conclusion?  I once felt very strongly that the ends justify the means. To maintain a level of integrity in handling matters may seem foolish, slow and impractical but that compared to selling/losing your own soul is a small price to pay.   '),(26278,979795,'06,August,2004','2004-08-06 00:00:00','    Past few days been running in the night on alternate days. Mon was the start of my night runs. Been performing relatively good though. Tend to make them 3km runs each time. Worst stretch was the part along Scotts Rd as the crowd starts squeezing along the streets at Lido\'s side. As I hit Stevens Rd, the pavement I run along gets very close to the road itself to make it pretty unheathy to be breathing all the exhaust fumes from passing traffic. Effectively speaking I now have only 8 wks left to the event. Not very confident I can train myself up on time to finish the marathon, but I\'ll still try. Even then, I have Aikido grading the afternoon before. I\'ll be cutting things a bit close. If I really don\'t feel up to it, I may just sit out for AHM 2004. But I will definitely be training and aiming to attend.   '),(26279,979795,'03,August,2004','2004-08-03 00:00:00','    Training tonight felt really good. I was trying to maintain a mindset for my next few trainings, to conventrate, commit and brush up on my atemi. I\'m hoping that making a decent atemi will help me to fall in a more natural way during ukemi.  My first partner was Reno so I didn\'t know if I did proper atemi or not, afterall he\'d just move me if I was standing still. Next was another lady I sometimes paired with. It was OK but I didn\'t feel comfortable going all out in this case. All I did was try to follow through in slowmo. Then it was with Sunny and Reno. Naturally I couldn\'t do a thing to Sunny. But I did try to concentrate on committing more, like actually grabbing. Whatever the case, Sunny threw me a couple of times. Didn\'t manage any fantastic ukemi but did manage to prevent myself from going down too hard and I think Sunny held back a lot for me.  Finally was with Edmund. One thing about Edmund\'s atemi though, it\'s rarely 100%. You\'d get 100% from him when he does it fast, but then you\'d have to know how to make use of the incoming energy at that speed (which I don\'t). And because of his weight, I also have a hard time trying to apply techniques on him. Dunnoz ...  My iriminage feels crap today. Considering it was with Sunny, I can understand why I don\'t seem to be able to do a thing. Haiz ...   '),(26280,979795,'01,August,2004','2004-08-01 00:00:00','    Yes, that\'s how I think. There must be some sort of conspiracy going on. If you read a few posts earlier, you\'d know that I had missed training at Bt Batok last Mon coz I was stupid to have taken the wrong MRT line. Not once byt twice.  Well, yesterday I finished classes at 5.30pm, left school immediately and boaded bus 135 to Paya Lebar MRT. No way was I gonna be on the wrong line now because this time I only needed to change 1 stop before I alighted (@  Jurong East MRT Interchange). From the East to the West there were already a lot of people at the time as it was after school and work. I had to stand together with a crowd all the way until I alighted. Was so tired and sleepy. I forced myself to read my book instead of standing and falling asleep. Well, I got to Bt Batok MRT to change TIBS 177 to go down to the CC. Waited there for a good 10 mins before some young sec sch kid politely told me \"Mr, I think you missed the last bus for 177 already. The service has ended at this time.\" I felt blur man. It was only 7pm and the last bus had left? Geez... I wanted to kill myself then. Felt so hungry by then. Bt Batok Central was a shopping mall and I could find lotsa food there, but I seriously was in NO mood to be hanging around there any longer. Thought I may still make it if I could find one of those bus which went in the direction of BB East CC. Went out to the main road and accidentally took a service which took me to Boon Lay. At this time I was in a pissed mood already. Decided to go Holland V for dinner and cool off. I took the MRT from Boon Lay to Buona Vista and walked towards Holland V. When I passed by the bus stop nearby, my usual bus home (105) arrived. I decided, \"Oh heck with dinner!\" and hopped right on to go home.  Got home, still in a pissy mood. When I entered the house, first thing my Dad asked was \"Had dinner?\" Was in no mood to be nice so all I said was \"Already told you this morning I\'m not eating at home right?\" He could tell something was wrong and left me alone. I went straight up, changed into my running gear and went for a night run. Didn\'t feel like bringing the house keys with me this time. (The previous time I didn\'t bring my keys, the front gate actually konked out. When I got home I pressed the button and waited to no avail. Had to scale the damn door like we usually did back in NS for the \'Suicide Window\' of the obstacle course. I was hoping that the electronic gate might\'ve konked out again so I could use some spare energy to hop over the gate again. Yea, sounds stupid, but I that\'s how I felt.) Went for a 3km run, which felt good. I think I made good progress this time. But it was horrible running along Stevens with traffic on the road, all other roads I ran along were a slight distance away so it didn\'t feel that bad. I got back in good time (and mood) and took a 1.25 litre bottle of water from the fridge. Sat along the steps outside my house to try and cool down and enjoy the breeze in what was an unusually nice and breezy evening in Singapore. I might\'ve been sitting there whistling tunes for a good 30mins or so before I went to grab a cold shower. Shower never felt that shiok before either. I\'d just stand there letting the cold water trickle from the top of my head down to my toes.  Sometimes no matter how bad a day may seem, there\'s always a silver lining to be found. Even if it means that you had to spin the thread out yourself from cotton plants, but it\'s there... somewhere out there lah...   '),(26281,979795,'01,August,2004','2004-08-01 00:00:00','    Yes yes, another movie I just caught. Seems like I\'m catching movies pretty often on Suns. Not that I mind it one bit.  Not too fantastic despite some well choreographed fight scenes. But everything seems to be  going the \'Crouching Tiger Hidden Dragon\' way now. It\'s nice, but not fantastic. I think it\'ll only impress non asians instead as they are still getting used to these type of martial art flicks. It\'s plot is simple enough. Revolves around a time when rebels were against the local govt. Basic story line is, A & B tried to trick C. B goes undercover, B and C falls in love. Turns out C is a spy, and it was a trick in the first place. A &amp; B are captured. Story twists somemore, turns out A is a double agent. A & C used to be sweethearts. When A finds out B &amp; C are in love, A tries to kill both. C gets mortally wounded. A & B fights a duel and both are injured. C intervenes despite injury. C threatens to kill A if B dies. A pretends to try to kill B (hoping to die in C\'s hands), C pretends to attack A but unable to harden herself to do so and dies as a result of her wounds. A and B devastated and show ends.  I tell you is a \'chim\' show. Zhang Yimou kinda overdid it lah. Heard Hero was something like that. This should be a theatre work instead of a movie. 2 things I found comical and kinda stupid, A walking around with a knife in the back (literally!!). C gets a flying dagger in the chest, coughs lots of blood, collapses, starts snowing, and she still has the strength to climb up again. Dunno what it takes for her to die lor. Maybe a nuclear warhead will do the trick.  Anyway, I\'m guessing that there will be quite a no. of pple who wants to or will watch this because of 3 reasons; Andy Lau, Takeshi Kaneshiro and Zhang Ziyi... people, please don\'t waste your money. If you really have to watch it, do so when it costs $6.50.   '),(26282,2606936,'31,Januar,2004','2004-01-31 00:00:00','   So.... what\'s up with the new Jewel, anyway? I was looking at CDs on Amazon.com and saw her cd which is called \"0304.\" Her identity keeps goin\' funky. I mean, she\'s quite pretty, and has always been the pretty natural girl, which gave us hope........ and I am just not into this whole DIVA thing......  She went from being (a little creepily, i may add) folky with \"Pieces of You,\" to being really earthy with \"Spirit,\" to being a little more rock-y with \"This Way,\" and now this pop goddess phase. The following are all actual songs on the CD.   2. Run 2 U    5. 2 Find U           8. 2 Become 1                     11. Yes U Can           12. U & Me = Love         Erm. I guess the number title was appropriate..... If you\'d like to sample them you could go  urlLink here ....   but I haven\'t even listened, so whatever floats your boat.  '),(26283,2606936,'29,Januar,2004','2004-01-29 00:00:00','   I have a gut feeling that the charge on ammonium is -1 or -2.... maybe. I cannot for the life of me remember the charge on chlorine. I took chemistry last year, so it\'s not a big deal. However, I am trying to figure out the formula for ammonium chloride.  Our band directors are out of town at a state music teachers\' convention, so in band we were sitting around and \"studying.\" My friend L walks into the room and looked a littel down, so i\'m all \"hey, what\'s up?\" and she just asked where some other girls were. I pointed them out and as she walked off i saw tear tracks on her cheeks.  She came to me about ten minutes later, beckoned me away from the crowd, and told me, \"B\'s getting expelled.\"  I immediately assumed something had gone terribly wrong and was like, what the heck did the school try to pull on us now..... and asked why.  Turns out she and another girl, during a lab in their chemistry class, randomly thought it would be a nice idea to drink ammonium chloride.  She freaking drank a chemical in a high school lab. After all the safety contracts we have to sign, after all the rules we have to memorize, and while they\'re even protecting your SKIN from it.   This is so weird. The other girl, everyone tells me, is pretty dumb and reckless. But B? She\'s got more common sense than that. I don\'t know if it was a dare, or what. L said she was laughing about it and said, \"Oh, I don\'t know why I did that....\"  She could be expelled or suspended..... It makes absolutely no sense. IT\'s the kind of ridiculous story chem teachers tell to wake up kids. Like, the one about the boy who put sodium (sulfur? some S-starting element that combusts upon contact with water) in his front pants pocket and then went swimming. B\'s story will become one soon.  How can you ever know someone? I mean, we weren\'t freakishly close, but I would have given her the benefit of the doubt. Who the hell drinks ammonium chloride in school?   To all of you in my locale, please don\'t comment on this here, because I don\'t want to further the gossipy crap already perpetuated.  To all of you out there in Blog Land, please try to refrain from randomly ingesting chemicals you\'re told not to ingest.  '),(26284,2606936,'26,Januar,2004','2004-01-26 00:00:00','   \"Hearts are worn in these dark ages....\" -Sarah mcLachlan  I was so hoping that this year wouldn\'t suck like the last one. Truth be told, mine hasn\'t been too bad, except for me AAAAH not havingthe scarf   But everyone\'s breaking up.... which means that there is an awful lot of bad stuff swirling in my locale. And that the spite wars are far from over.  I think tomorrow I will wear my \"Boys are Stupid, Throw Rocks at Them!\" t-shirt. I love this shirt, despite the fact that it is baby blue, which is an incredibly unflattering color for me. I wanted the original one i saw in some magazine, which was pink, but alas, I could not locate the store.   When shopping, or dragging the Schnig around the mall, or something like that, I came upon this shirt in a store I normally would never have gone into.... and I had to have it. Yes, I was shopping with a boyfriend, but WHAT A GREAT SHIRT.  It is the shiznit. It is bodacious and tubular. It is so very.  I never actually wore it in front of HIM until after the breakup.   Man, i treat that kid horribly. I ignore him like, um, an ignoring thing..... I mean, I smirk whenever he attempts to address me and ignore him even when he opens doors for me.... I mean, he opens the door at school in the morning because it\'s locked, and that\'s what you do, you open the door for people in the morning who want to get into the school, not because he\'s being chivalrous or some crap like that, as I had previously thought. See, he tends to make rather odd friendships in which he insists on carrying stuff and helping people all the time. SOOOOO obviously he will open the door for everyone, but he\'ll run faster if they are a)cool or b)pretty or c)not me, since I am obviously both *cue hair flip*.  He doesnt\' do it on purpose. Of course it isn\'t his fault he metaphorically pantsed me, punched me in the chest a few times, kicked my stomach repeatedly, and somehow removed my heart, ran over it, backed up, and did it again, only to attempt a hug.....   But I still cannot bring myself to greet him.  '),(26285,2606936,'25,Januar,2004','2004-01-25 00:00:00','   W: Hi, welcome to McDonald\'s. Can I take your order? 1: Wait..... what? W: Can I take your order? 1: No way! I want to eat it!  W: Can I take your order? 2: Um... where\'s the McRestroom?  W: Can I take your order? 5: No.  W:Can I take your order? 3: Mc-yes! *sniggers* I would like a McSprite, please. W: What? 3: A large McSprite? And an apple McPie.  W: Oh....kay....*sighs* um. That\'ll be $2.17. 3: Ok, thanks! Where can I get some*sniggers* McNapkins?  W: Can I take your order? 4: What\'s that supposed to mean? W: Can i take your order? 4: I don\'t appreciate your tone! W: I\'m sorry, ma\'am, I don\'t understand what you\'re upset about. 4: Oh, of course you don\'t understand. How about this: I don\'t McLike your McTone! W: Ma\'am, did I do something to- 4: I would like to speak with your McManager! Manager: What\'s the problem? w: What the- 4: I\'ll tell you what the McProblem is! The McService here McSucks! I\'m going to file a complaint with your McCEO. W: I quit. Manager: You have to give two weeks notice if you want pay! W: Ok.... fine. Two weeks notice and I McQuit.  W: Can I take your McOrder?  '),(26286,2606936,'19,Januar,2004','2004-01-19 00:00:00','   Today, I attended a very cool Anti-Defamation League Youth Leadership Conference to Stop Hate. It was, as Heather would say, so very. However, I do not want to go into the meaningful part of it. No, I want to talk about the hot guy I met!  Now, at one point, they gave us sheets of paper. Half of us had words, and the other half had definitions. I got matched up with a rather lovely lad from another school. He was a self-proclaimed hippie environment kinda guy. Usually I find that rather weird, but when he\'s pretty, he\'s pretty!  Our word was \"scapegoating.\" For the less vocab-ulated among us, that is blaming a person or group for something that is not their fault. Case in point - Jews being blamed for economic crisis in post-WWI Germany by the Nazis.  SO we\'re trying to think of \"personal\" examples, and I\'m trying to break the ice over the fact that neither of us had experienced it. Him: \"So.... well... I mean, there was the Holocaust....\" Me: \"Yeah.... don\'t really know of any others. DO you?\" Him: \"No, and I\'ve never really been scapegoated or anything.\" Me: Me either. *pause* Me: But, ah, it has \"goat\" in it, and goats are pretty cool, so..... HIm: *smiles at my patented blend of beauty, intelligence, and oddness* Yeah. I\'ll a gree with that. *laughter* Me: DO we have to explain it to them? Him: Um.... I don\'t know. but if we do, I\'d definitely go with the goat escape. *laughter* SUpervising kid: Ok, so y\'all are going to make skits. Me: OK... so pick another animal. Him: Um.... one that pertains to a goat situation? Me: Sure. Whatever. Him: a rooster. Me: OK. SO let\'s say it\'s been raining a lot at the farm,a nd the rooster\'s like, \"It\'s the goat\'s fault!\" Him: *hahahahaha* Me: SHould we use something else? Him: No, I think rain is fine.  SO, we come to perform our little skit. I read the definition.  Him: Wow, it\'s been raining a lot lately. That\'s not very good. Hey, I bet it\'s because of you! Me: What? Me? It\'s not my fault! I\'m just a goat! I don\'t control the rain! HIm: Well, how else do you explain it?  It then deteriorated into the laughter of the rest of the group and us doing our best barnyard arguments. Cute, but not incredibly productive. I was trying to figure out how to end it....  and I   head-butted him   on the shoulder.  Head-butting is an inside joke with my friends. We do it at random as a fake fight instigation. It is quite funny to watch. If planned, in a skit, it makes people laugh. The group was falling out of their chairs and stuff.  However, I do not suggest randomly head-butting people, especially those of the opposite sex in whom you have the capacity for interest. Because i was too afraid to talk to him at ALL for the rest of the day.   America\'s young leaders: pretending to be barnyard animals.....  '),(26287,2606936,'18,Januar,2004','2004-01-18 00:00:00','   Two kinda sad things.  1) I think i\'m going to skip my junior prom.  I did not particularly want to announce this, but a friend recently made a similar one, and i thought it\'d be nice reading for those who know me not.  You see, since the end of the AGe of the Boyfriend, i have been attempting to muster interest in other guys. This, i have done, but not exactly with results. I have crushes that go from minor, to obsessive, to \"enh,\" within the space of a week. I have yet to reveal this to anyone of interest. I do not expect any interest to be revealed to me, either. I think I\'m kind of intimidating. People notice me, but I don\'t need them to fall in love with me.  Also, I\'m still in the depths of both pathetic sadness and pathos, although i can often rise above it. So probably, a date wouldn\'t be worth taking, because I\'d just be all comparative and shit. So unless i fall madly in love sometime in the next few months, we can forget a romantic prom.  I do have a few friends i could go with, but the only reason would be to get a dress and get pretty, and they\'d probably be at a point where they\'d take me for fun, but be staring at another girl all night. Why weigh them down?  Not to mention the fact the the Schnig will probably get his new obsession to come with him. I don\'t want to see that. I shouldn\'t have to.  I don\'t dance much, I have no real love interest, I don\'t want to watch him with someone else.  If Steve-o is brave enough to state that he doesn\'t want to go, i can too....  2) Tomorrow, i shall attend the Anti-Defamation League Conference to Stop Hate. Expect a barrage of confusion from me tomorrow, because it\'s being held at a Holocaust museum.  '),(26288,2606936,'16,Januar,2004','2004-01-16 00:00:00','   ok. everyone needs to go to  urlLink The Fudge Said No , because it\'s a rather funny blog...  oh wait. hmmmmmm. The site has exceeded its monthly bandwidth allowance!  hmm.... i wonder how? Just how DID the bandwidth exceed itself, hmm?  Probably from jerks who are too lazy to save pictures on their own servers, and instead link to them from the website!  All right. I don\'t exactly understand what that means. All i know is, try very hard not to link to pictures that are online unless absolutely necessary, or  1) the \"bandwidth\" will get \"exceeded\" and           1a) the site will be shut down, and          1b) the picture will disappear bc it is no longer online, or           OR  2) If you are using up bandwidth, the person hosting the picture will get pissed and do  urlLink this.   This is also why i do not attempt to post pictures here, especially of myself. For the sight of my lovely face would surely incite all of you to include it on your own websites, and the all my bandwidth would be stolen....   It\'s possible Mr. Fudge managed to exceed his own bandwidth without outside help, but i doubt it. Anyway, it\'s a good opportunity for a PSAabout the evils of internet theft. Speaking of INTERNET STEALING-TYPE PEOPLE....   Oh, screw this. i\'m not the internet Polizistin. My neck hurts and i need sleep. Go to  urlLink Tequila Mockingbird  for further reading.  '),(26289,2606936,'16,Januar,2004','2004-01-16 00:00:00','   ever had braces? if so, remember having them adjusted, and how your mouth was sore and stuff?  imagine having that feeling in your NECK.  (if you\'ve never had braces, imagine soreness every time you turn your head more than two inches or so.)  Yet, there\'s a triumph in this pain!  You see, there\'s this wonderful way of relieving tension known as ACUPRESSURE. basically, the idea is that when you have a knot in a muscle, you force it to relax by apply pressure. It\'s easy to do to yourself... What you do is take either your finger (or if you have nails, the flat end of a pen) and push into the knot until you feel that \"good hurt\" and hold it, either until the pain goes away or until you can feel your pulse in it.  Well, today, while waiting for a friend to get off work so we could all go out, i was sitting with two guy friends,a dngot one of them to hold a pen against this awful knot at the edge of my left shoudler blade that i\'ve probably had for months. The muscles in my back actually twitched when it released! Unfortunately, now that my upper back and neck are all loose, my neck is SORE from being tense and clenched for months. It actually hurts when i laugh, which isn\'t good considering i went out with friends and then watched the movie  Clerks.  But whatever....  i suggest anyone who (a) attends high school, (b) works, or (c) has children, try acupressure. It is absolutely fabulous, because event hough i\'ll be sore today, my neck is so wonderfully loose! Extend that recommendation to all wind players and singers as well....  '),(26290,2606936,'16,Januar,2004','2004-01-16 00:00:00','   1) a new haircut 2) a new template for this 3) sleep  '),(26291,2606936,'12,Januar,2004','2004-01-12 00:00:00','   I absolutely positively LOVE german.  i get tired of the whole \"guttural, short, efficient\" stereotypes....   if you\'ve read stuff like shakespeare and even scarlet letter, you\'ve seen the patterns of speech they use, but not all of it. german has a rounded syntax, so the sentences just roll out, in waves, swelling and relaxing. They phrase themselves like music, so when someone reads German poetry well, it sounds like a chant. Each syllable, each word, is separated by the breath so that it\'s all articulated clearly, but there is always resolution. If you know what a periodic sentence is in English, that\'s the closest thing i can think of to compare it to. But german - !  The inflections of the voice, of questions, anger, uncertainty, sarcasm, all feel so much more natural in german. The sentence always ends, either resolving the entire piece, with a soft, low tone and a crisp cut-off - \"Das habe ich noch nie bevor gemacht.\" or with a question! - the beauty of a German question, carrying the melody to the next speaker, growing in intensity, volume, inflection, ready for the next entrance. \"Was denkst du?\"  It feels like a crime to interrupt someone because it disrupts the pattern. if you don\'t allow them to finish their ideas, it takes away from the song quality, because there\'s no verb, no action, no movement. \"Das ist die Maedchen, die ich sehr gut bevor-\" man kann da nicht nur - wait - crap - sorry. You can\'t just stop there, you have to hear the rest, you have to know what happened, to feel the emotion behind the words.  When you speak long sentences, it sounds like a Romantic melody, one long, clear, stream of breath, interrupted only by the tongue. I love to speak in it, just to hear my own thick, loud, obnoxious voice transformed into music.   It seems so much more natural. So many words sound like English, use the roots we know so well, but the way they feel in your mouth are so much easier. \"The\" vs. \"die;\" \"tanz\" vs. \"dance\"; \"that\" vs. \"das/s\" - if you know German, just take a moment and speak. Don\'t read, don\'t recite - speak the language, feel it, let it roll off your tongue in swells, and you will feel the resolution. You can feel the connection with other speakers, why any language is so important.   With German, i marvel at human creation, that people were able to develop something so aesthetically pleasing, using only their own minds, their own voices, their own need to communicate and touch others. It;s like, they only needed a way to comunicate, but they gave it these patterns that are just so perfect and so beautiful, and it\'s entirely human. I feel like if i had been born speaking it, my poetry would flow so much better. German is always crowding the back of your throat, begging to be brought into the world for the joy and enlightenment of other people. it\'s like when you hear music that gives you chills on your neck, and you try to recreate it with your own poor voice, and the attempt brings tears to your eyes. The emotion, the beauty, the joy and pain, the sheer humanity behind language, is what i love about it. Music creates a mood, creates feelings that are shared by listeners. A beautiful language can both create the mood and communicate concrete ideas.   That\'s why literature is so amazing. The awesome thing about good song lyrics is that it\'s poetry - the form of language closest to the human spirit - but it has that untouchable, unreachable music behind it.  this is what makes me happy, truly - knowing that it\'s possible for a human soul to make something so beautiful, because if someone else can do it, so can i. i want to create beauty, or at least perpetuate it, honor it, reveal it to those who are too afraid to look in familiar places.  '),(26292,2606936,'09,Januar,2004','2004-01-09 00:00:00','   i have come to the following conclusion: I am a sadist.  Well, not like a physical one, but an emotional sadist. I think.  Maybe that isn\'t the word I want....  Well, the point is my ex-bf\'s depressed mood actually made me feel better about myself today, but when he perked up, my mood went downhill.  This morning, I was just kinda chattering to someone about band stuff, and the guy he was talking to was like \"Hey, Audrey!\" as I\'m leaving the room. I kinda turned and was like \"Hey!\" but the Schnig definitely snorted . I dunno if it was a \"wow, what a stuck-up bitch\" or \"she\'s been ignoring ME, why would she talk to YOU?\" snort, but it was definitely a snort. I guess one of us has to be rude in order for the world\'s energy to be in balance....  But then suddenly after school, he was all happy and stuff. And this shouldn\'t have any effect on me, but it\'s like, \"Dude, can\'t he EVER be depressed?\" The kid just doesn\'t care.....  Argh. I mean, it was so nice to see him visibly acknowledging my ignoring him.... and then he went back to NOT showing it!   How is a jerk like me supposed to get any happiness with stony-faced emotional ironmen like that around? My geez.  '),(26293,2606936,'09,Januar,2004','2004-01-09 00:00:00','   The worst songs ever written.....  - \"I Wanna Be Bad,\" performed by Willa Ford, a blonde who attempted to be the anti-Britney by dressing like a slightly more punk slut, but singing the same songs.... \"I wanna be bad, you make it look so good....\"  - \"if I was Invisible,\" performed by one Clay AIken. This song is supposed to be all sweet and sad and stuffm, but it\'s really creepy.... \"If I was invisible, I could just watch you in your room...\" Obsess much? He also states, \"I would be the smartest man if I was invisible,\" wrongfully implying 1) that invisibility is not only a good idea, but a measure of intelligence and 2) that he is a man.  - \"Can I Get Your NUmber?\" performed by... um.... wait, let me look that up. Ah, that well-known group known as No Authority.     (?)        anyway, this song goes \"Can I get your number, baby? Hit me with the seven digits!\" I mean, how European-naive is that? \"7 digits?\" What about the area code? Hello, welcome to America, land of the free  and  land of the 10-digit phone number..... This is just asking for her to blow you off. But then, if you tell her \"I\'ve seen blondes, and brunettes, and some really hot redheads, but I\'ve never seen a girl like you,\" chances are she\'s going to blow you off anyway.  - \"Hey Juliet\" as performed by some group that has a number in their name. let me look it up. Nope, it\'s LMNT. We\'ll save the commentary on the name for another post. Anyway, this song is just a little too cutesy and upbeat to use allusions to Shakespearean tragedies.... are we supposed to believe these guys are in a \"homeroom class?\" Obviously, if they ARE in high school, they\'ve never read \"Romeo and Juliet,\" because he thinks they could just \"run away\".... This guy watches a girl in homeroom, pages her incessantly, and dedicates songs to her on the radio, all to be eye-rolled... and yet he firmly holds that \"You really want me!\" Denial ain\'t just a river in Egypt, honey.  - \"Just Got Paid,\" as performed by N\'SYNC. Number one - I doubt a paycheck makes a whole lot of difference to these guys. Number two - they say \"ho\" a lot. And they don\'t mean it in a disgusting, chauvinistic way - which can never make for a good mainstream non-Christmas song.  - \"E-mail My Heart,\" performed by Britney Spears. This would be a cute idea for a country song if it was a love song with a computer theme. Actually, Weezer could do it really well, too. And I guess a pop song could do it too..... But it has absolutely nothing else to do with email. She\'s like \"I miss you, I love you, email my heart and tell me you love me\" or something. It\'s just a cheap attempt at being \"hip and trendy.\" Also, they hyphenate \"email,\" which is about the most \"I\'m so-dorky-i-don\'t-even-use-a-computer\" thing that can be done, unless you are a German, in which case someone decided they could do whatever they wanted to the language, and you can\'t really be blamed for being forced to spell it \"E-mail\" or not ever use ÃŸ\'s. (more on this later....)  - \"Heaven,\" by Live. I definitely thought this song was by some terrible religious-band copy of Matchbox Twenty... I don\'t know if Live  is  a religious band.... but it\'s never good when you can guess the next word in a song. Rhyming \"believe,\" \"perceive,\" and \"decieve\" is rather trite, especially when you\'re talking about religion. \"Percieve\" is one of those \"big\" words they use to sound original. I should knnow, I use it all the time in lit papers. And what the heck were they doing playing a sucky song about Christianity on normal radio?  - \"Baby, When the Lights Go Out\" by Five. Not only does this song have horribly-disguised sexual subliminal messaging, they also say \"second verse, girl\" before the second verse, which clues you in to the carefully formed song structure, and how much the guy is waiting for the song to be over... wonder what HE\'S planning.  '),(26294,2606936,'07,Januar,2004','2004-01-07 00:00:00','   i\'m not sure anymore if I actually want someone to love me, or if i only want to love someone.  '),(26295,2606936,'06,Januar,2004','2004-01-06 00:00:00','   i would just like to announce that   my ex-bf has dropped out of my last-period class!!!!   this makes me immensely hopeful for many reasons:   1) One less hour of being in the same room with him every day. Don\'t have to look at him and make myself sad!  2) He isn\'t there walking there! No more depressing reminders on the way to class! No more watching him restrain himself from groping other girls when directly in front of me! 3) He isn\'t there walking away! Even fewer depressing reminders on the way home!  4) Don\'t have to listen to his retarded comments and convince myself that they are indeed stupid! Or remember when I thought they were funny!  5) Don\'t have to watch him humping anyone\'s leg outside the classroom! (this is, of course, a crude and metaphorical denotation of what actually takes place, but the sight is nonetheless both disgusting AND heartbreaking!) 6) Don\'t have to talk to him when he thinks it\'s appropriate to discuss dates we went on! Don\'t have to listen to him bugging me about whether or not we went to this or that movie! Don\'t have to listen to other kid snickering at me bc he\'s so insensitive about it!  7) While i\'m at it, I don\'t have to talk to him, period! No more acting like seeing his face doesn\'t rip my heart up in new ways and put me off food for several hours! No more polite ceremonial conversations!   I AM FREE.... or at least freer than I was last semester. He\'s still in band, which is where he acts  more like an ass than any other class, but at least i\'m free of him for one more hour every day.  '),(26296,2606936,'06,Januar,2004','2004-01-06 00:00:00','   I have six pieces of jewelry that i wear every single time i leave my house. These are: my grandmother\'s engagement ring, my class ring, my sapphire ring, 2 Nominations bracelets, and a watch.   The other day i was taking them off before bed and my grandma\'s ring fell off the dresser. I heard it hit wood, so i knew it had fallen into one of the two open drawers below. However, i dug through them and couldn\'t find it.  Tonight, after work, i finally pulled the  two drawers out and dumped all the clothes out. These were my two \"main\" t-shirt drawers. The top one has mostly cute, tiny little shirts i wear all the time, and the bottom is mostly the big souvenir t-shirts i wear to sleep and practice and stuff.  Digging through the first one,  i found two GHP shirts i hadn\'t worn in months. GHP, if you don\'t know, is an AMAZING summer program i went to last summer. Basically, high school sophomores and juniors are nominated in one school subject, anything from math to art to German(as I was) to dance. You go through a series of auditions or interviews, paring the number down to a lucky 675 that get to go spend the summer on a college campus, studying their subject. There are no grades or tests or anything. You learn only what you want, and everyone there has a love for learning. It is also a once-in-a-lifetime experience, because you can only go as a student once.  GHP changes people. It makes you hate the stupid people in your school even more, for one thing. You get used to the freedom, to being trusted and praised. You get used to only studying what you love. You get used to the friends that are all worth knowing, that are all intelligent and talented. People leave and find their old friends too shallow, or end up breaking off pre-GHP relationships because of how they\'ve grown more than their partner.  I should have broken up with The Schnig while I was there. I did, actually. I thought i was gonna be ok. But he called me back at 3:00 AM and begged. It was sweet. I took him back.  Several months later, he dumped me and broke my heart. It hasn\'t healed yet. We haven\'t spoken in weeks. We haven\'t had a real talk in MONTHS.  Anyway, at GHP we made t-shirts at the drop of a hat. There was one for each hall students lived on, one for each major class(German), and one for each minor class - you chose one other subject to study. I took a Brass minor, since i play French horn. I also got the t-shirt for Pink Appreciation, which was a tongue-in-cheek organization celebrating the fabbity-ness of the color pink. They had shirts for the Homosexuality and Society discussion group, which i wish i had gone to so badly. I am probably the worl\'ds leading straight fighter for gay rights. I think i was probably afraid of seeing that side of my friends, though - the side that condemns certain people being able to fall in love. Or else I was spending too much time on the phone with the Schnig. But people made shirts for IB, for Asian Pride, for anything and everything.   It\'s incredibly fun to wear them because no one understands the inside jokes, except sometimes the handfuls of school-people who went. You pass in the halls and have an instant bond, a nod at their shirt, because they understand, even if you\'ve never really talked at all.  GHP was amazing. There are times when someone reminds me of it and I miss it so much I literally start to shake. I wore the German shirt to Nixie\'s sleepover this past weekend. That one isn\'t as great because it\'s in German, so people would rather say \"een goot klaze abber een bibshun kamch\" and laugh maniacally at their cleverness, than ask me what \"Eine gute Klasse, aber ein biÃŸchen KOMISCH\" means.  (FYI, it\'s \"A good class, but a bit ODD(funny, strange, etc)\"- \"komisch\" literally means \"funny\" but it has a connotation of being somewhat off, weird, whatever. Thanks for asking.)  I think i\'ll wear my brass shirt tomorrow. We\'re starting school again after break and I\'m going to need it to give me some confidence with all the stupid gossipy little asses there.  Not to mention the Schnig.  '),(26297,2606936,'05,Januar,2004','2004-01-05 00:00:00','   I slept over at Nixie\'s last night. It was quite the bodacious event. We watched movies and all that good stuff.... Spirited Away, if you\'ve never seen it, is something that MUST BE SEEN. We decided to have a Spirited Away party someday and dress up like characters. D is going to be the Radish Spirit, which is like this big walking white tuber with a red hat and loincloth. He rocks the cat spa. I wanted to be Lin, who is this cool lady who\'s all like, \"You\'re in BIG trouble!\" but i think i\'ll be Yubaba the witch, who has a big nose and hair.... too bad i can\'t put pics up here.... oh well.   They do have two cats, which was not so good for me. I have a cat phobia. Seriously, i looked it up, it\'s called Galeophobia. Napoleon had it. It must run rampant in the short gene or something.... but yeah, i came in the house and the first thing i say is, \"CAT!\" because it\'s all trying to crawl on me and stuff.... Why is it that whenever i babysit, and come over and they have a dog, they\'re like, \"Oh, i forgot to ask if you were afraid of dogs\" and i\'m like, \"No way, i love dogs!\"..... yet, if they have a cat, they never think to ask if i\'m afraid of cats.  This is just more proof that cats are slowly taking over the minds of AMerican suburbia, with the intent to  undermine and ultimately destroy my attempts to help the poor.  (They play with yarn while i\'m making charity scarves. The n00bs.)  ....... getting back.......However, D actually brought her tarot cards, which she\'d never used or anything, so i did tarot readings for everyone. I sure am talented... Lace was screaming and playing with, like, a talking book or something and we were like, \"Don\'t disrupt the flow of positive energy!\"  But i did myself, and i\'m like \"When am I going to actually love someone again?\" and it\'s telling me the past is like, wealth and stuff; present - what was it? like, withdrawal; others think i can\'t move on; i should try not to make enemies; and, if i do all this, i will long for a lost love.  NO SHIZNAT.  aaaaahhhhh! That\'s it. I\'m never dating again.  '),(26298,2606936,'04,Januar,2004','2004-01-04 00:00:00','   urlLink           urlLink Which quiz are you ?   This pointless quiz was made by  urlLink TMO      If you want to read anything for the time being, go to my xanga. The link is on the left.  '),(26299,2606936,'04,Januar,2004','2004-01-04 00:00:00','   hey, i am thinking of moving to blogspot, because xanga is a little bit constricting. but how wil i get my friends to read this?  '),(26300,2606936,'25,Februar,2004','2004-02-25 00:00:00','   I am so friggin\' giddy. Jitter was here and we were writing our script for our independent film that\'s going to change the world.   National History Day Project. Enhhhhh. But it\'s all about Freud, so there\'s a small degree of woot to it. *dances joyously*  Gotta love people who think they know a lot about psychology. I mean, hello, not only have I read on it extensively, I\'ve BEEN to them! AND talked a lot about the craft, not just me..... so hearing someone be like, \"Oh, I LOVE Sigmund Freud!\" is just aggravating. It\'s so pretentious. The mind isn\'t something that can just be laid out and studied. I mean, yes, there are people with \"issues\" but that isn\'t how it works. It\'s complicated and emotional and there is most definitely a reason why you need training for it. There are bookoo pppl out there who are going to major in psychology. But they don\'t want to practice. And half of them shouldn\'t practice! Just go do studies, please don\'t walk in with someone thinking you can break downt heir mind into defined problems.....  But anyways. Freud was weird but quite a smart man. And so we\'re detailing how a traumatic experience caused a boy to repress the memory of rejection and transfer his love for a girl to a DDR machine.  Obviously, it\'s got a lot to do with DDR and dancing. So we\'re talking about how people will dance out their emotions or something in a comical fashion. And I ended up dancing around my room for like an hour. I was so hyper and happy and WOW, I feel optimistic.  In other news, I\'M NOT TELLING. *raspberry*  '),(26301,2606936,'24,Februar,2004','2004-02-24 00:00:00','   He apologized to me today.  He wrote me a note and signed it with his biological father\'s last name. He asked me to talk, and I didn\'t think I would until I saw the name.  He said stuff has happened. We didn\'t have much time. But I apologized and he apologized more. And he meant it. And we hugged. And I feel so much better about it now.  He\'s a different kind of friend. But I think maybe, considering the apology was given, without any asking from me, that he\'s not all bad. And maybe I could possibly consider forgiving and being his friend.  It\'s so weird. We just  talked and were actually honest and I realized how much I had missed him. Not necessarily in a romantic way. Just as himself. It\'s weird to lose a good friend.  Only problem, in the chance that we do become friends?  EVERYONE ELSE WE KNOW.........  '),(26302,2606936,'21,Februar,2004','2004-02-21 00:00:00','   The \"five questions\" game-ish thing, to which I previously referred, goes like this. these questions are from Jen of Prepare to Meet Your Bakerina. Links are now abundant.  Here are the rules. I left her a comment. Anyone who wants to participate and has a blog or journal, leave me a comment, preferably with an email address. Although I understand if I\'m too freaky and you just want me to comment on YOURS..... *sigh*   1 - Leave a comment, saying you want to be interviewed. 2 - I will respond; I\'ll ask you five questions. 3 - You\'ll update your journal with my five questions, and your five answers. 4 - You\'ll include this explanation. 5 - You\'ll ask other people five questions when they want to be interviewed.      1.  What is the funniest book you\'ve ever read?   Angus, Thongs, and Full-Frontal SNogging,  et al. To commit the reductive fallacy, they are \"teen\" novels about a weird English girl. Sooooo addictive, and I laugh out loud while reading. I suggest them to anyone, because they\'re too good to be \"tenn books\"..... they\'re even funnier for the culture shock.     2.  Where in the world would you most like to visit?  WIlliamsburg, VA, right now. I want to go to college there but my parents don\'t like that idea, so I may never get to go.....    3.  Who is your favorite non-Simpson resident of Springfield?  Milhouse. The show would have no emotional drive without him. \"I swear, I don\'t know how that scrunchie got in my hair!\" ....... \"I wish I was a free man. .,.... Oh, no! Puppy Goo-goo! MOM! Where\'s Puppy Goo-goo?\" *recieves toy, cuddles* \"Oh, Puppy Goo-goo, fetch me a dream!\"   4.  What music do you play when you are feeling happy and/or optimistic?  Sheryl Crow,  C\'mon, C\'mon  album. \"Soak up the Sun\"...... Need I say more?   Actaully, this one is versatile. It expresses so many feelings, but always with this overtone of hope. I listen to it when I\'m in a good mood, but also when I want to be in one......FYI, it\'s also great music for packing and unpacking suitcases.....    5.  What made you decide to study German?  FATE. or an accident, if you prefer...... I can\'t roll my R\'s, and I really hate the sound of Spanish, and my middle school had the choice of Spanish or German. I chose German, unlike all my friends. But it has been so worth it. I mean, I\'ll have done six years of it before college, because it just rocks. I love it and I thank god I didn\'t take anything else! I need to blog in German sometime.    5a.  Optional bonus question:  Have you ever seen the Wim Wenders movie Wings of Desire?  No, but now I think I\'ll have to. Does it involve German?   NOW GO COMMENT!  '),(26303,2606936,'20,Februar,2004','2004-02-20 00:00:00','   Billions of apologies to Jen of  urlLink Prepare to Meet Your Bakerina.  I asked her to \"interviewificate\" me with the five questions game and accidentally emailed her from my spam email address! SO It\'s been almost a month since she sent me the questions and now I actually got them. Man, what a loser am I.  To make up for it, I shall now plug her site. It is something between a baking blog and a reggaler thoughts one. She is a smart kid and has really cool shoes.  urlLink So go visit it and tell her you think I suck.   Answers shall be posted tomorry.  Signing off, I am stupid. *circle-y finger salute*  '),(26304,2606936,'20,Februar,2004','2004-02-20 00:00:00','   You\'re Ireland!      Mystical and rain-soaked, you remain mysterious to many people, and this  makes you intriguing. &nbsp;You also like a good night at the pub, though many are just as  worried that you will blow up the pub as drink your beverage of choice. &nbsp;You\'re good  with words, remarkably lucky, and know and enjoy at least fifteen ways of eating a potato.  &nbsp;You really don\'t like snakes.   Roman\">Take the  urlLink Country Quiz  at  the  urlLink Blue Pyramid  '),(26305,2606936,'18,Februar,2004','2004-02-18 00:00:00','   Forgot to mention that I ate kola nut today. One of the sacred foods of the Ibo people, this was used in many ceremonies and symbolized something in Chinua Achebe\'s  Things Fall Apart.    Fabulous, fabulous book. We\'re studying it in my IB LA/AP Lang-Comp class right now  However, it was incredibly yuck. Tasted like a potato covered in hairspray and anthrax.  Also, the girl who provided these morsels (who was doing a presentation on the symbology of food in the book) ordered them off the internet. So here I am, having eaten a frigging kola nut that came off the internet..... i mean, god knows who grew them, what kind of pesticides are on there.... they sure TASTED like pesticide.  And so then she tells us that the Ibo ate/eat a bit of it before each meal to help digestion.  I only took a small bite, but i did swallow. I mean, it\'s high school, fourth period, and our stomachs are eating themselves because we don\'t eat breakfast. But then I turn to the guy behind me, who probably ate about an entire half a kola nut, and said, \"Is it weird that I can feel it digetsing in my stomach?\"  And he\'s like, \"Yeah, I can feel it too! My stomach\'s all like whew-whew-whew.\" *pantomimes squishing things with hands*  So yeah. I haven\'t felt sick yet, so it must have digested itself or whatever.  BUT can you imagine our entire IB class getting sick from (possibly) black-market African kola nuts someone bought off the internet?  We\'d have a big IB nap-and-barf day. And maybe we\'d get to miss math. Woot for that.  '),(26306,2606936,'17,Februar,2004','2004-02-17 00:00:00','   I am one of those people who reads the Bill of Rights\' \"freedom of religion\" as \"freedom FROM religion.\"  However, today, i found myself procrastinating. Instead of turning into my neighborhood after school, i drove past it to a little shopping center. I wanted to go to CVS to replace some cool heart-shaped mardi gras beads i accidently threw away today (don\'t even ask). However, finding none, I wandered around the center.  I went first into a little consignment store. I\'ve never been to one before, despite living here for the greater part of my life. I poked around and tried on three little black dresses. Two were size 8 and waaaay too big. One was a \"size 2\" and yet looked strangely small on me..... still, it was nice. The dressing room was pretty. It had one of those accordion-fold doors and star-shaped hanger things on the walls.  Then I walked all the way down, past a Hallmark owned by these weird uptight Indian people who have all these signs like, \"children should not be let in here unsupervised\" and stuff. I went past a closed bar and a pizza place and a camera store.   And then I went into a little store whose tagline was \"religious books and gifts.\"  I don\'t know why I went in. I think just seeing it there all those years and boredom and the fact that I was (and still am) wearing a tiara for the beginning of Karneval, made me step in.  I actually got a little freaked out when I pushed open the door. And working in retail myself, I found it surprising that they didn\'t greet me when I came in. The lady was chatting to some customer.  I saw a bunch of books, for children, like \"Gross and WEird Bible Stuff\" and \"Bible Weapons and Wars.\" I saw pamphlets about pornography and abortion. I saw books on prayer in marriage and YA fiction series about adventures and stuff. Books on \"apologetics,\" which I now assume to be the art of justifying Christianity, or something. (It was like, Apologetics 2.5: YES, YOU SHOULD BELIEVE IN THE TRINITY: Speaking With Jehovah\'s Witnesses.) There were rosary beads in all colors and frames and little \"TEEN CREED\" things. There were little wallet-sized cards proclaiming the dangers of abortion and \"Your Baby\'s Growth,\" with drawings of fetuses and stuff. There was one bookmarks with a picture of a little thing in a blob with some hands, and it said \"Living X-old fetus from ectopis pregnancy.\" ANd on the back, \"A person\'s a person, no matter how small. - Dr. Seuss\"  So yeah. All these things that normally make me mad - I mean, I\'m not married and not exactly virginal, and i\'m pro-choice to the point of anger. And I don\' know what the hell a rosary is, and I don\'t care about prayer or pictures of God or whatever.  I felt like an impostor, masquerading as a sweet little Christian, when I\'m really just a heathen curious to see what the store\'s all about.  I\'m not drawn to religion at all. It just seems tedious and meaningless to me. Which is why I haven\'t picked one.  But it was a nice, peaceful atmosphere. No God-Squadders shoving things in my face, blessing me, asking me if I\'m saved. (Although there was a pamhplet on that last one.) I think if I was religious, I\'d have a store like that where people can come to sit and read.   Although maybe not next to giant plastic Mary statues.  '),(26307,2606936,'17,Februar,2004','2004-02-17 00:00:00','   My former OT once told me, \"don\'t hide from yourself.\"  Funny that this was during the Epoch of Sadness following the Age of the Boyfriend. So i kinda hid from myself 24-7 there.  Unfortunately, it\'s happening again.  NUmber one - on Valentine\'s Day (VALENTINE\'S DAY!) I received a call at ten-of-ten PM (on VALENTINE\'S DAY!) from the Schnig. (Did i mention this was Valentine\'s Day?\" Who wants to receive a call from an ex-bf on \"VD,\" especially when they have recently discovered unpretty suspicions about said ex-bf are true, cementing a resolve to never again exist on civil terms with said ex-bf?)  (Not I.)  The conversation wasn\'t much. Here\'s the rundown. I answered it despite caller ID because I was afraid he would leave a message, adding to the twenty or so i already have of him that i\'m afraid to delete bc I\'ll ahve to listen to him talking about how much he LOOOOOOVED me at 12:47 AM Saturday, August Blah. But to get to the interesting part....  Audrey: Hello. The Schnig: Hi. A: ..... who\'s this? TS: um..... it\'s [NAME]. A: ....oh. ..... hi. TS: Um. Is it, like, too late to call, or do you not want to talk to me, or.....? A: ..... ....... ...... I don\'t especially wanna talk to you. TS: *sighs* Ok. ......... well, bye. A: *hangs up*  Eu. Icy, I am. I refuse to say what all went down, but he deserves it from me... but anyways.  So all that funkiness notwithstanding (VALENTINE\'S DAY?!?!?!?), a new problem has arisen.  In that I may have thought I had a crush on someone who turns out to like me a whole lot, where I don\'t think i reciprocate, and yet I cannot stop flirting with him.  If this is followable, without names, I do this a lot when single..... I generate small amounts of interest, flirt a little, and end up with guys who think they are in LURV and major guilt. I\'m a repentant serial heartbreaker. Ask a friend....  So it\'s like the real, nice Audrey is hiding from the reality - that she may be leading someone to care about her. And that she doesn\'t REALLY \"requite\" this.   And she is hiding from the fact that she could probably stop it, if she would only face up.  But the real, nice Audrey is a wuss. Which is why I can be a \"cold ice queen.\" * I let my id rule my superego and end up wondering why I suck.  * Great story..... he said something like, \"Well, you can\'t just be a..... I\'m sorry for saying this..... a.... cold ice queen! There, I said it, and I\'m sorry for being so..... mean.....\"  and i believed his innocence and honesty.  NICE AUDREY, you are a sucker.  '),(26308,2606936,'16,Februar,2004','2004-02-16 00:00:00','   .......... Audrey is loose again, and she may be about to break another heart.  Someone save other people from myself. Or whatever.  '),(26309,2606936,'14,Februar,2004','2004-02-14 00:00:00','   I love to read bottles on beauty products. You know, the back, where they promise all kinds of stuff? It\'s too fun. Yesterday, a friend and I spent an hour at CVS buying all kinds of cool stuff (mousse, cheap red V-Day Mardi Gras style beads, brown hair shampoo, water flavored with SPLENDA, etc).  I think reading bottles of conditioner while treating your hair actually has a psychosomatic effect on its beauty. Which is why I am quite familiar with the Pantene Pro-V Treatise on Conditioner. It includes the  Six Signs of Healthy Hair:  1) silkiness 2) shine 3) less frizz 4) softness 5) strength 6) body  So, yeah. I mean, hey, they\'re the ones who devote their careers to making hair shiny, so why not listen?  However, I was just now reading a bottle of my brother\'s shampoo, which is VO5 Balsam & Protein, and it promises to restore the  Five Attributes of Healthy Hair:  1) Manageability 2) Shine 3) Softness 4) Bounce and Body 5) Strength  Compare and contrast, my friends. Compare and contrast.   I must now return that stuff to my brother\'s bathroom. But I think this shows that not all hair companies are the same. I mean, one seeks \"silkiness\" and \"less frizz\" while another wants \"manageability\" and \"bounce.\" Totally different.  What the hell is balsam?  '),(26310,2606936,'14,Februar,2004','2004-02-14 00:00:00','   Some of my friends tease me about being a \"sex addict.\"  We decided jokingly to discuss the meaning of life on a school trip and I stated that if you look at it scientifically, the meaning of life is to create other life and perpetuate the species.  Dom: \"....... so, you\'re saying the meaning of life is to have sex?\"  And you can imagine the rest.  But i was reading Jitter\'s thoughts on it and i thought i might state my position on it for ya.  I think the meaning of life is different for us all - it\'s what you make of it. But for me, it\'s to be happy and impact the world in a good way. If i can make one person smile every day, I feel like i\'ve done well.  So many of you go around, worrying about not being world-changing. But then your presence can have such an impact on others\' worlds, and you don\'t even know it.   I was once pouring out my worries about being important or needed or whatever to the Schnig, and he told me that a friend of ours had at one time been so sad that she had been cutting herself. When he asked her why she had stopped, she said a lot of it was because of me. Just being happy to see her, greeting her, made her feel wanted.  It was so small, the things I did, but they made her life that much better. This made me feel that my life was not a waste. If i have helped one person to be happy, then I have not lived in vain.  So just know this. This is why I am nice, why I compliment people, why I laugh at stupid jokes teachers make, why I make an effort to be kind to everyone, to get to know people. I\'m not sucking up, or being \"social butterfly\"-esque, or whatever. My two goals in life are to be happy, and to make others happy. Y\'all dont\' realize how the little things impact others. My friends don\'t realize that sometimes, just giving me a hug or something can make my day. They do it because they want to, and that\'s what makes it so special. They want me.  And so, I try to give out, to \"pay forward,\" i guess, the care that people give me. Just remember how much you mean to the people around you, how much your seemingly little actions can do for the world. Don\'t hold back on telling your family or friends how much they mean to you, because that\'s often what they need most.  To be happy, we all have to take care of ourselves and each other. That is the meaning of my life.  '),(26311,2606936,'13,Februar,2004','2004-02-13 00:00:00','   I am wearing a thong.  For the first time in my life.  And ya know what?...........  I don\'t hate it.  '),(26312,2606936,'12,Februar,2004','2004-02-12 00:00:00','   so i\'ve been thinking about the whole \"blogging phenomenon.\"  Millions and billions of people, all with blogs.  Publishing every little thought that creeps into their heads.  Hoping for someone across the world to read it and smile.  And leave a comment.  there are a few types out there. There are those really fabulously talented people who made it popular (see  tequila mockingbird, bellow , and  indiscretions ) and get read over and over daily. there are the ones who write down everything they had to eat and how many times their children pooped that day. there are the ones who use it as a place to post inside jokes and talk to their friends in asian baby slang (\"it ish now jowk! I em not keedin, peepul reelee tock like dish online! azn_prYde for everyone!).... and then there are ppl like me, who aspire to be blog geniuses and so must refrain from ranting about food or whatever, unless it\'s interesting (to normal people).   and.... i had a point.... oh yes.  so. What got me here was how Julia of  TM  hasn\'t posted in a few days and is driving us all nuts. I was thinking, \"what, did she die or something?\"  and then i thought, \"Oh. what if she DID die?\"  What if someone who blogged a lot and was part of a big blog community died in an accident? I mean, none of their friends or family would know how to get into the blog and tell people they were dead. They probably wouldn\'t even know it existed, due to this whole anonymity thing. How weird would that be? Can\'t you just see the comments piling up and up....  and then what if one of their actual tangible peeps found it online, and read thru all of it?  It\'s kind of a theoretical reality check. Like, the internet can only go so far in connecting people. Ya know?  '),(26313,2606936,'11,Februar,2004','2004-02-11 00:00:00','   In band, we\'re preparing for our Large Group Performance evaluation.  (It used to be called festival but apparently all the stupidly non-musically-inclined teachers at school got pissed because they thought were were missing school to go to a party.....)  The requirement is that each band plays three pieces. One is a march, one is a piece with movements, and one is just.... a piece. The teachers pick these from a list.  We are playing a massively fabulous piece called \"Of Sailors and Whales\" By W. Francis McBeth. It\'s got five movements, and they are all scenes from Melville\'s  MOby Dick.  It is mucho fun-o.   One of the best parts, in my opinion anyway, is that the third movement is almost all choral. Tubas, percussion, and low reeds play, and everyone else sings. It\'s some of the words from the ballad \"Father Mapple\'s Hymn\" and it\'s a tune that\'s similar to \"Eternal Father, Strong to Save,\" which adds up to a mournful song baout a guy stuck in a whale praying.  However, I love to sing. LOOOOVE to sing. I may not be so good but I enjoy it. I did a little theater and a lot of band and I can belt it o-u-t when needed.  SO, we\'re going over it today in class with CT\'s dad, who\'s a chorus dude, and every time the \"women\" had to sing, CT would be like, \"GO AUDREY!\" and people kept smiling at me and stuff. I thought the sopranos were just really soft (i\'m singing alto) compared to the guys, but when we got out of class, I was informed that I\'m the only girl who can actually be heard. Apparently I sound pretty good, since there was no malice or anything. i actually got asked if I\'d had training!   So there\'s a nice feather for my metaphorical AP Lang/Lit Hat. As one guy put it ,  I am \"the only girl who has gonads.\"  '),(26314,2606936,'11,Februar,2004','2004-02-11 00:00:00','   I have always been  sehr  annoyed by cards that refer to a person as \"Valentine.\" For example: \"Valentine, you set my heart aflutter!\"  However, I\'ve found a new annoyance.   What\'s  up  with the candy conversation hearts that say \"Let\'s Read\"? I was eating some the other day and I can\'t figure out what it means. Usually the phrases are like, \"Call Me\" or \"Luv Ya\" or something. Cheesy, to-the-point, cliched.  But \"Let\'s Read?\" Are teachers supposed to give them to the class or something?   Or maybe \"read\" is actually \"R.E.A.D.\" and it stands for something mildly to severely suggestive. I don\'t know what it COULD stand for - but that\'s how these kids do it these days. The make up all those crazy words for *whispers*  \"the sex.\"    Maybe it should have said \"read my lips\" or something like that, and the people who work in the factory are just illiterate.....   or maybe it\'s actually really beautiful and romantic and I\'m just missing the point by making fun of candy for the sheer reason that I don\'t have anyone to give cheesy candy  to?   Nah.  I\'m just glad they finally figured out just how a man can knock me off my feet. \"Hey, baby, let\'s READ!\"  (We shall ignore, for the moment, the fact that i had my first kiss in a bookstore.)  (ignore it......)  (oh, go.........   climb in a handbasket ............or something to that effect.)  '),(26315,2606936,'08,Februar,2004','2004-02-08 00:00:00','   Have you ever stopped to think about how so many Americans use big words they don\'t understand?  First of all, I can guarantee that 50% or more of people who use the word \"abstinence\" all the time not only believe it to meaan \"not having sex,\" but would also be clueless if you asked them to abstain from something.   And \"platonic.\" God, how annoying is it when teens use that on TV or movies? Most people think it means \"non-romantic.\" Many fewer understand the \"non-sexual\" part of it, and very few make the connection with Plato. A pencil-eraser\'s worth knows what Platonic forms are.  And.... \"perverted.\" \"Eeeeew, he\'s perverted.\"  I don\'t think that was ever supposed to be an adjective. \"Perverted\" means \"sexually gross\" to today\'s kinders. \"To pervert\" would be \"to turn away from its natural state.\" But if you use any form of the stem around a middle-school-aged-or-younger crowd, you\'d get giggles and junk...... The correct adjective is \"perverse.\" But it just doesn\'t have the same sickening ring when squealed, now does it?........  And for a non-sexual example, \"president.\" Everyone who\'s been to kindergarten, and many more, know who the President is. Yet, they cannot define the word \"preside\".......   Fifty-cent words for everyone!   \"What, were you, like, raised by wolves?\" \"Yes, I was. Hmm.\"  '),(26316,2606936,'08,Februar,2004','2004-02-08 00:00:00','   DO you remember that old SNL sketch with Lisa Kudrow? She did the talk show \"Single and Lovin\' It!\" where all these single women were all \"happy\" and stuff. If I remember correctly -   Woman: SOmetimes I get really bored and lonely on Sundays. Other Woman: Oh, yeah, I know how you feel. (murmurs of agreement) Lisa: Oh, ok. Well, what I like to do is sit on my bed and draw myself into my favorite Cathy cartoons. (murmurs of appreciation) Other other woman: Um, I just want to say that I think that\'s really sad. Lisa: What? I\'m single and Lovin\' it!  and then the entire show deteriorates into needy women begging for men..... or soemthing to that effect.  ..... and...... point?...... oh yes.  SO I burned my very first CD today. Must be some rite of passage..... but I got a blank cd and used ti to make sure i could download onto and then burn cds on my computer. and i can! WOOHOO for that.  My plan is to take this playlist of \"friend\" songs i have and burn a CD to copy for all my girlfriends for valentine\'s day. Basically, I\'ve got 2 friends who aren\'t single.... girls i mean. so yeah. i think it\'ll be good.  but then there\'s this leeeeeeetle voice in my head going \"WHAT THE HELL ARE YOU DOING? YOU\'RE SEVENTEEN!!!! WHY ARE YOU WORRYING ABOUT DATING AT SEVENTEEN! THIS ISN\'T HEALTHY!!!! YOU HAVE TO START DIETING BEFORE YOU DO THAT!!!!\"  uuuuuuuuuuuuuuunh. I don\'t want to feel guy-less. I want him to notice me, dammit! alas, alack.  I went on a blind date a few nights ago. Well, my friend Vix persuaded me it was a blind just-friends but she meant it as a date. luckily, judging from both our reactions, we\'re off the hook as far as dating goes.   Why are there so few smart guys out there, and why don\'t they want to date me?  Playlist = soak up the sun, diamond road(sheryl crow); up from under, birdcage, baby bird (wallflowers); something to believe in (shawn mullins); answer, perfect girl (sarah mclachlan); everybody hurts (rem); and some other stuff. God, i don\'t remember. Whatever.  Much love.   And by that I mean Love, as in the Form. The Platonic form. Wait. Is all love striving to be Platonic \"Love\" or \"platonic love\"?  Damn Plato and his suspect sexuality.  '),(26317,2606936,'05,Februar,2004','2004-02-05 00:00:00','   I would now like to explain in detail my foolproof cure for the hiccups.  While I was at GHP last summer, I got a severe case of them, and a girl I met was like, \"I\'ve got a cure!\" Having often been advised to do realllllly weird things, i was skeptical, but she got me to try it. What you do is, fill a glass with water and drink from the opposite side. I don\'t mean, turn it around. I mean, basically you have nto have your chin inside the cup. It\'s awkward, but the slurping calms down the spasms of the diaphragm. The water probably shouldn\'t be too cold either.  This has worked excellently for me since then. It\'s awesome and also makes a funny story.  That being said, it ain\'t working now. Hopefully I won\'t end up like that guy who had them for like two years.  Anyone got a better cure for the hiccups?  '),(26318,2606936,'03,Februar,2004','2004-02-03 00:00:00','   I came out of the bathroom stall and washed my hands in silence. My friend N stood next to me drying her hands.  I turn to her and say, \"You know what? I\'m a horrible mean bitch and I don\'t even understand why people want to be friends with me.\"  This would have been a lot more effective if  a) she hadn\'t gone on to ask me to repeat myself twice(loud hand dryers, ya know). b) I had said it to the person who deserves to hear me admit it.  I have been having a weird time of it. One of my good friends\' birthday parties (J) was Saturday night, so I slept over there. Problem? J\'s mom and the Schnig\'s mom are bestest buddies. So naturally J\'s house is plastered with photos of the Schnig. Let me tell you, you have not felt awkward unless you have attempted to converse with someone standing near the refridgerator while keeping your eyes focused elsewhere...... especially when they probably know exactly what\'s up.  ANyway, the combination of these pictures and our Mexican dinner (coincidentally, at a former favorite date spot of ours - kill me now) gave me weird dreams. Something involving bicycles and helicopters and shopping (at a Target outlet for eyelash yarn and lampshade trim) and coffeeshops with tall stools that someone fell off of. And at some point the Schnig asked me if I would please go back out with him. This caused me to burst into tears (in my dream) and tell him I had to think about it. I felt as though I was being inexorably sucked back into the dead-end relationship and knew I shouldn\'t, but feared I would.  I woke up and was actually surprised to realize it had been a dream  ANyway, I guess this refueled my animosity towards him. Yesterday, sporting an Audrey Hepburn purse, I got many compliments, especially from people who understood the correlation. (MY NAME IS AUDREY, dumbbutt.) ANyway, one of \"his\" friends complimented it and he was like, \"I don\'t get it.\" I surprised myself by saying icily, \"Well, hon, that\'s cuz you\'re stupid.\" I swear it was supposed to be a joke.... but it didn\'t sound that way.  Let me add that he has also been hanging near me wayyyyyy too often. I don\'t want to be his friend and he seems to think I\'ve gotten over that. (And obviously I don\'t get over things too well. Tee hee. Bawl.)  SO today I\'m walking to the restroom and as I leave class, HE\'S leaving his class! and he\'s in front of me. SOmething in me decided to walk brusquely around him (slowpoke lazy underachiever) and I expected nothing from him. But as I\'m just getting him out of my sight, I hear him.  He ventured a \"hi.\" In a little broken baby voice, like he was afraid and resolute but still hopeful.   And confused and scared to death, I barely turned my head, then flipped my hair and waked off shaking my head.  Let me clarify that there was no one else in the hall and that he knew I had heard him. How rude was that? How awful was that? How could  I  possibly put up with that from someone? How can I possibly continue hating him when he speaks to me like a child defeated?  Why does he try?  And the scary question - will he ever try again?  '),(26319,2606936,'01,Februar,2004','2004-02-01 00:00:00','   Advanced Big 30 Personality Test Results       Sociability   |||||||||||||||||||||   70%       Gregariousness   |||||||||||||||||||||   62%       Assertiveness   ||||||||||||||||||||||||||||   90%       Activity Level   ||||||||||||||||||||||||   78%       Excitement-Seeking   ||||||||||||   34%       Enthusiasm   |||||||||||||||||||||   70%        Extroversion    |||||||||||||||||||||    67%        Trust   ||||||||||||   38%       Morality   |||||||||||||||||||||   70%       Altruism   |||||||||||||||||||||   62%       Cooperation   |||||||||   22%       Modesty   ||||||   14%       Sympathy   |||||||||||||||||||||   62%        Friendliness    |||||||||||||||    44%        Competence   |||||||||||||||||||||   66%       Neatness   ||||||||||||   38%       Dutifulness   ||||||||||||||||||||||||   74%       Achievement   ||||||||||||||||||||||||   78%       Self-Discipline   |||||||||||||||||||||   66%       Cautiousness   |||||||||||||||   46%        Orderliness    |||||||||||||||||||||    61%        Anxiety   |||||||||||||||||||||   66%       Anger   |||||||||||||||||||||   70%       Depression   |||||||||||||||   46%       Self-Consciousness   ||||||||||||   38%       Impulsiveness   |||||||||||||||||||||   66%       Vulnerability   ||||||||||||   38%        Emotional Stability    |||||||||||||||    46%        Imagination   ||||||||||||||||||||||||||||   82%       Artistic Interests   ||||||||||||||||||||||||   78%       Emotionality   ||||||||||||||||||||||||||||   82%       Adventurousness   ||||||||||||||||||   54%       Intellect   ||||||||||||||||||||||||   78%       Liberalism   |||||||||||||||||||||   70%        Openmindedness    ||||||||||||||||||||||||    74%'),(26320,2606936,'01,Februar,2004','2004-02-01 00:00:00','   My friend Vicky and I have this inside joke. She once told me this story, i think from SNL or something, in which someone specualted that if stupid people would only carry around signs stating their stupidity, people wouldn\'t waste time asking them for help or speaking to them or anything.  The example she gave was that of a truck driver who drove under a bridge that was too short and got stuck. He waved people down for help and they would ask him, \"Oh, is your truck stuck under this bridge?\"  He would reply, \"No, I\'m a Shriner. Heeeeeere\'s your sign!\"  Essentially, when we give sarcastic answers, we end with that and giggles.  SO today I was perusing my \"referral stats,\" which gives me the URLS of search engine pages, including queries. If anyone remembers the kola nut episode, which is linked at left, I talked a little about the book  Things Fall Apart.   Anyway, this one Googler came from the USA, I guess, and their query was: \"Did Chinua Achebe\'s life affect Things Fall Apart?\"  SOmeone out there is searching the internet for this. To this person, if you still read this site: OH. MY. GOD.   I\'m not going to ask if you\'ve ever heard his quote on the spear-shaking savages, or if you know where he lived, or where he was from, or if from his name you might have figured out his heritage. I\'m not even going to say \"No, of course a revolutionary book written about his native culture\'s tragic end was not at ALL affected by his life. It was a coincidence. If he hadn\'t got to it first, Danielle Steele would have,\" with a mean ironic inflection.  I\'m just going to say:  No, his death affected it. Heeere\'s your sign.  '),(26321,2606936,'01,Februar,2004','2004-02-01 00:00:00','   My friend Sara posted something today that made me smile. I altered it to suit myself:  \"I love you more than the federal deficit!\"  In reading back over that, let me just say that i do not love the federal defecit. It is just really big. Dummy.  '),(26322,2606936,'01,Februar,2004','2004-02-01 00:00:00','   I just want you all to know how wonderful American schools are.   Now, in all seriousness, getting an International Baccalaureate education with AP credit is HOT HOT HOT!   However, it\'s not the gifted kids that set the school\'s standards.  You see, darling Bushie-Wushie decided that no child would be left behind in terms of education.   This is a nice policy, i guess. Make sure everyone is educated, blah blah blah. I mean, it was crappy when in fifth grade students couldn\'t even read properly bc teachers were sooooo lenient with them.  But do those of us who care about what we\'re doing in school, really ever need to be exposed to the mundane stuff? I mean, I would hope that having passed all my hard-ass classes I would be prepared to graduate. But no, Georgia like to do this \"high school graduation test\" - a big standardized test on math, soc stud, science, and LA - and if u don\'t pass it you don\'t graduate.  This is also a great idea. I totally agree with the concept.  But OH MY GOD. The things they consider testable knowledge, learned in school, at a high school level? ANd the fact that people fail this stuff? That just fills my heart with dread for the future of our great nation.  We\'ve only had science and social studies so far, but here\'s some stuff i saw that was memorable. I\'m typing a few from memory and a few off of a study guide i was given....  1) A chemist drops a beaker holding unknown chemicals onto the floor and it breaks. Which of the following is    NOT    part of the correct safety procedure for cleaning this up? a) Use tongs to move the glass shards. b) Wash hands after cleaning. c) Do not rinse chemicals down drain. d) Carefully pick up the glass shards with your fingers and deposit them in the nearest trash can. PROBABLY GEORGIAN ANSWER: B. We too deep for chemical burns.  2) Which of the following is    NOT    a responsibility of a U.S. citizen? a) vote in elections b) talk other people into agreeing with your point of view c) obey the law PROBABLE GEORGIAN ANSWER: A.  \"Evolution: don\'t let it happen to you.\"  3) The Senate campaign in Jane\'s state has become very controversial. The chanllanger, John Jones, has accused the incumbent, Harry Havelock, of being unconcerned about an issue. Where should Jane go to  find out how Mr. Havelock really feels about these issues? a) the encyclopedia b) her parents c) the COngressional Record d) her social studies textbook a) vote in elections PROBABLE GEORGIAN ANSWER: B. Mostly Republican, huh?  Enh.... i don\'t wanna think any more.....  '),(26323,2606936,'01,Februar,2004','2004-02-01 00:00:00','   In preschool, there was a little chart under the bulletin board with the days of the week. It started with Monday. Each one was a different-colored pocket, with a white square that had the day printed in black letters. You were supposed to move the paper that said \"today\" on each day.  I still remember each color. Monday was blue, a dark blue that i today learned is known as \"Klein blue\" after a guy who patented the color. Tuesday was orange, Wednesday was green, Thursday was a dark red-burgundy-red, and Friday was yellow. Saturday and Sunday were light purple and pink, I think, but we didn\'t have school on those days, so they weren\'t so important, since you didn\'t get that cool job of moving the paper.  To this day, I associate each day of the five-day school week with that color. We make plans for Tuesday and I envision orange. Etc, etc. Weird, huh?  '),(26324,2606936,'01,Februar,2004','2004-02-01 00:00:00','   i just discovered those little page counter thingies.  See, i\'ve had one, which is nice, for about ..... 460 people. But i just realized if you go into the thingy, it shows you where peopel got referrals, and how they found me on search engines and stuff.  It was kinda freaky bc last night, one of the addresses was a Google search, with a query about \"pictures of naked little boys doing\" something I don\'t especially want to type on the internet. Sure, I\'m a publicity whore, but i don\'t want THAT kind of publicity. (please don\'t lead them back to my page now.)  However, my Simpsons love has helped a bunch, since the biggest searches were stuff like \"puppy,\" \"goo,\" and \"milhouse.\"  OH NO. I just realized i no longer have the link to REJECTED. Ahhhhhhhhh.  Homework? Psh. Pssssssssssssssssssh. High school kids don\'t do WORK on the weekends, especially a 4-day weekend! Especially when we feel like losers and stuff like that. No, we blog and im each other and ingest caffeine.  I think I\'ll go mess with my new bangses. Have a charmed evening.  '),(26325,2606936,'01,Februar,2004','2004-02-01 00:00:00','   B - the new name for the ex-bf, taking into account a few events that you shall never hear of - and I are american teens, if you didn\'t get that already. We dated for a year and a half. We were mad mad mad mad madly in love. We were best friends, and more, and soooo happy.   It\'s funny how double dates tend to highlight stuff in different couples. Like when Vicky and her bf of 2 years, and my blind date and I, went out together...... it was weird because they\'re so comfortable, while we we afraid to touch elbows. Stuff like that.  You see, it was a strange but great coincidence how my best friend and his best friend - we\'ll call them Boy and Girl - began dating a few days after us. Naturally, we double-dated all the time. Our relationships moved at the same rate and it was just nice - you get to spend the day with three of your best friends, one of which is the person you LURV.  Just, ya know, fun. Innocent, first-love excursions, holding hands in the mall or sitting on each others\' laps at a coffee shop. The four of us, in our clearly-defined couples. So cute, so sweet.  Now, while Boy was much shorter than Girl, B and I fit that old \"girl-should-be-smaller\" standard. But not in a funny way, like Nixie and Timmy, who are 5\'0\" and 6\'3\", respectively.  So which couple got the stares? Which couple got looks ranging from amusement to disapproval every time they went out? Which couple wouldn\'t have been identified as a couple but for the linked arms? Which couple was just \"wrong?\"  I guess the little stuff does matter. Stuff like how I\'m white, and B is black. Ya know. Just that.  It\'s weird for me, having been a sheltered white girl, raised without any trace of racism, how it works. How people I don\'t even know can look at me, standing with someone I love more than anything in the world, and tell me with their gaze that we\'re not right, that I can\'t love him, that it\'s unnatural, that it sickens them, that I\'m being stupid and deviant and that I\'m WRONG.   If you\'ve never been there, thank someone.  Funny how that translates. How my thinking about this, my first actual encounters with prejudice, somehow led to that ongoing question everyone faces - \"is it okay for people to be gay?\"  And suddenly, I thought, \"I\'m not gay. It\'s none of my business whom they love, because it doesn\'t affect me at all. How could I ever turn around and say that people couldn\'t be in love after knowing how that feels?\"  I realy wish all the stupid uptight prejudiced cowards of America would just take a friggin\' step back and look at themselves. Thank God that you\'re not homosexual, because we will never have to live the life they do. We will never have to struggle for  the right to love someone.  We will be able to marry the people we love. We will not get those stares everywhere we go. We will not live our lives set apart from so many others who are too afraid of opening their minds to try and care about another human being.  What kind of world is this, that I have to be thankful not to be in a minority? WHat the hell kind of world are we creating for the future? There\'s no way I\'m having children. I\'m not bringing a child into a world that\'s so full of hate it outlaws  marriage.    My God, people! Live and let live! Not everyone lives by your religious rules!  WHatever happened to \"majority rules WITH MINORITY RIGHTS\"? Whatever happened to a man and woman being the same in the eyes of the law? Pursuit of happiness? Not infringing needlessly on others\' rights, maybe? I dunno.  U.S.A.?  That\'s funny, I thought this was the United  Christians  of America. But hey, since the majority doesn\'t want to let gay people get married, gays obviously don\'t matter. Who cares if we\'re barring people from living the happy lives they deserve?  Some people think it\'s gross!   YOU KNOW WHAT\'S GROSS? HOW MUCH AMERICANS SUCK. We act like \"Oh, we love everyone, oh, we\'re so free and open here, and we all have our rights, blah blah blah\" and then we turn around and complete ignore people. What a shitty era. God, I hate hate. I hate hate. I hate hate.  Nice country, where people are free to live and love and speak and work and learn and worship the way they want.   Maybe someday I\'ll get to live there.  '),(26326,2606936,'01,Februar,2004','2004-02-01 00:00:00','   So we finished our Freud movie. It rocks really hard.   We thought it was going to SUCK because not only did the actors hold scripts a lot, we filmed it two days b4 it was due.  Then the deadline got pushed forward a day, so we had more time.  Then we discovered that there wasn\'t any of our recorded sound that couild be heard.  Then we came up with the idea of subtitling and using mood music.  SO basically, it\'s got awesome music (Gary Jules\' cover of \"Mad World,\" REM\'s \"Endgame\", and a really cool techno remix of the \"I Dream of Jeannie\" theme) and then these cool subtitles, which make sup for the stumbling actors. IT really looks good. I\'m excited beyond your mom\'s pocketbook, or something to that effect.  I think it\'ll change the way one or two ppl think about neuroses. I also think it\'ll make a high school Theory of Knowledge class fall out of their chairs with its comic genius.  Woot for history!  '),(26327,2606936,'27,April,2004','2004-04-27 00:00:00','   * denotes sarcasm.  Uhhh, I hate him. I hate him. I hate him. I hate him. He is a jerk a jerk a jerk a jerk a jerk.  If anyone wants to hate him with me, go right ahead. We can make a chant or t-shirts or maybe you can buy me ice cream.  On another crappy note, despite my request, I have somehow been scheduled to work on prom night. And if they don\'t fix it, i am going to break something. And I would like very much if it were that stupid boy.  '),(26328,2606936,'20,April,2004','2004-04-20 00:00:00','   If you\'ve ever played in a band/wind ensemble/symphony, go find a recording of The Children\'s March by Percy Grainger. Grainger is the rockin\'est weird genius musician ever, and this song ROCKETH. (Horn dominance, anyone?)  I might could attempt to post the recording of the piece HERE, but i tend to hate that, altho that usually results from crappy midi files..... i\'ll look into something that doesn\'t violate copyrights.  AND if you\'ve ever taken Zoloft, or know someone who has, and had odd moody side effects from it OF ANY SORT, or you just happen to know a whole lot about it, please leave a comment with ur story..... i am trying to find out something and yeah, ease my nerves here.  '),(26329,2606936,'16,April,2004','2004-04-16 00:00:00','   The Schnig came over today after school. Neither of us had anything to do and he wanted to escape his house, so we thought we\'d watch some tv and just \"chill.\"  SO he came over and we sat on my porch for a few minutes and talked about postmodernism and silly things. And then we decided to watch TV. And we went to the basement for some reason, and not the one upstairs where other people were.  And see, whenever i make up romance stories in my head, they\'re  Ã¼berromantic . And my favorite movie-scene kiss is the one that\'s totally unplanned, the first kiss between two friends who are just talking or fighting or something, where they don\'t even know what happens. Like they\'re just sitting there and suddenly they\'re against each other and they don\'t even care how.  But then it\'s kind of funny. Because i went down the stairs first and when we got to the bottom i think i must have turned around somehow, talking about some TV show, and was standing in his way, but we just looked at each other and he didn\'t stop walking and i felt my mouth drop open and he groaned.   And then his arms were all around me and my hands were on his back and everything else was gone. And it\'s like this sense of urgency, but also this familiarity. And it\'s so perfect and sexy and sweet and there\'s so much to feel.  And then we pull apart and somehow we\'re not standing up any more, we\'re lying on the couch and both of us are shaking. And we lay together for a moment, not listening to the TV at all. And he kissed my hair and i was close enough to count his freckles again, and he\'s crying a little and I can\'t make him stop. And you want him to stop so desperately because you want this to be a happy movie and you might start too if he doesn\'t.  But it doesn\'t matter how much you comfort. And just like before, we talked again and cried again because our parents don\'t like it, and our friends want to kill us if we mention the other\'s name, and we\'re right back to where it left off all those times before.  In the movies, he never leaves after a perfect kiss like that. And you don\'t ball up on the couch and sniffle because it still smells like him when you wrap yourself in the blanket. Because even as you both apologize, you know you would give anything to hold each other for another five seconds. And in the movies, if he whispered \"Just one more time,\" even as he\'s apologizing and pulling himself away, you would have one more time and a million more besides. You wouldn\'t soak his shirt with tears because you know that it\'s never the last time, that the last time is already through, and that you\'ll never be this close to it again.  And you wouldn\'t ever ever cry because you don\'t know if it\'s worse that it happened, or that it had to stop.  '),(26330,2606936,'10,April,2004','2004-04-10 00:00:00','   ....... i have discovered the best product EVER.  It is called Coppertone Endless Summer Sunless Tanning Foam.  I am looking good.  I recommend, my friends. I recommend.  I think my \"medium/dark\" tan makes me feel more robust and rugged. Why, I flirted with the guy at the ice cream place we went ot after the movies, and when i got home I ate a steak  without utensils.    That\'s right - eatin\' with the fingers, baby.  I should do commercials for them.  '),(26331,2606936,'09,April,2004','2004-04-09 00:00:00','   My LA teacher made some joke about how when we read Gatbsy this past week, and didn\'t HAVE to read it from a scholarly view (for once - cue \"Hallelujah\" chorus), we would still feel like annotating.  [Ahem, those of you who grew up in the \"everyone\'s-not-special\" generation, ANNOTATION means writing in the book. You write what you think and underline stuff. This is not \"Oh no, I hope they don\'t go under the bridge!\" This is \"Foreshadowing?\" or \"Periodic sentence\" or \"irony.\" I sometimes write \"Ouch!\" or \"Haha!\" but I odn\'t show my teachers.....]  I underlined stuff to avoid being pointed at and laughed by this woman, but I still thoroughly enjoyed the book!  However, it was weird, because on our plane ride home(yes, I was gone, losers who didn\'t notice or miss me) I\'m reading the last part - for those who\'ve read it, it was about the time Wilson and Nick left their respective residences - and my dad suddenly leans over and asks, \"What are you writing?\"  He leaned over to try and read it like I was a little girl writing in a diary and snickering at him. it was kind of weird, considering he\'s mass-educated, but then I think this whole annotation thing must relatively new, considering the amount of used books you can find that aren\'t all pencil-y.  I\'m sure I sounded like quite the disgruntled teen, but I whined, \"Dad, I\'m  annotating!\"    And because I thought my annotation here was stupid - it was like, \"is it dead?\" and generally my hope on that subject is rejected by superior Southern-Baptist lit experts - I  actually covered it up with my hand.   Like a little girl writing in a diary, guilty and embarrassed!  I don\'t think this says that annotation is part of that whole reader-response theory, that the reader puts a huge personal thingy in their reading, and maybe doesn\'t want to share, or anything profound like that.   What it says is that I\'m a nerd.  '),(26332,2606936,'31,may,2004','2004-05-31 00:00:00','   According to  urlLink Go-Quiz.com , this blog may be dangerous......      U  CAUTION      IN THE INTEREST OF SAFETY IT IS ADVISABLE TO KEEP MMM, ICED WATER AWAY FROM FIRE AND FLAMES.     Username:     From  urlLink Go-Quiz.com   (now, does that borderline-ironic sentence make anyone else smile?)  '),(26333,2606936,'28,may,2004','2004-05-28 00:00:00','   [i\'m going to nitpick for a minute here.]  You know that phrase that goes.... \"Shoot for the moon. Even if you don\'t make it, you will land among the stars!\"...?  Well, ever since I first heard this phrase, I have believed it to be crap.   This is not because it is cheesy - cheesy stuff can be sweet and kitschy.  It is crap, simply because the moon is closer to you than the stars. So it makes absolutely no sense that aiming for a short distance, and running out of fuel, will get you a long distance.  The moon is, like, the closest celestial body to the Earth, which means it ain\'t the hardest thing out there. Metaphorically, it suggests normal success, like pushing paper, not being in debt, having kids, buying a house on mortgage, etc.  The stars, on the other hand, are pretty freaking far away. So getting to the stars would be like finding a cure for AIDS, or being poet laureate, or being the world\'s first biracial hermaphroditic one-legged lawyer.  *note to self - must submit ideas to SNL for review.*  So that statement is saying something like, \"Try to learn to use a calculator. Even if you fail, you will still have that knowledge of calculus\" or \"Apply for a job at McDonald\'s. Even if you are rejected, you can always fall back on that degree in psychology.\"  It should be, \"Shoot for ...\"  You know what? I can\'t think of any intelligent way to rephrase that. I just think it\'d be funny if school counselors said \"Shoot yourself!\"  and..... uhhh....  Fargo : \"Go Bears...\"  [End annoying pessimistic jab at middle-school wisdom.]  '),(26334,2606936,'27,may,2004','2004-05-27 00:00:00','   Having discovered that i WASN\'T crazy, and  urlLink goliard  actually changed links, i finally had to do some cleaning up of the links list. I had left some \"dead\" ones up in the hopes that they would update that nifty bloglinker thing, but alas, they have probably been lost.  people don\'t live forever, and we often get bored with our activities. So some of those great-blogs-that-i-don\'t-read-daily-but-are-still-good-reading are going to be seriously missed.  SO, the first installment of the informal obituaries, as best I can remember.....  ORFINANNY AND AUSTIN http://orfinanny.blogdrive.com  This writer guy, Austin, had this cool thing going where he invented the \"perfect\" woman, Orfinanny, and they supposedly wrote a blog together. Lots o\' political stuff, and lots o\' humor. Woe, woe, woe to the dead duo. I can only assume that one of the guys who envied Orfie killed Austin off and they ran away to St. Bart\'s.  STRATFORD-UPON-AVON: THE TRUTH http://atowngoesdown.blogspot.com  This other guy, he lives(d?!?!?) in Stratford-upon-Avon, England, which is that place where that writer guy (not Austin) wurde geboren. So, he had some neat observations about life in the famous place. And also some neat observations on life in general. Some tourist probably shot him or something. It may very well have been my English teacher.  A DAY IN THE LIFE http://adayinthelifeat.blogspot.com  Ok, I didn\'t read this that often, but it was cool because this lady linked to me early on and I was flattered that someone read me. She had survived cancer and wrote cool stuff about her life - just not necessarily that humorous. Still, twas good. I have no idea how she \"died\" - I seriously hope it wasn\'t cancer. Probably just bored from reading about MY life.  shit. I can\'t remember who else died. Seems like there were more....  Well, uhh, just remember this, bloggers:              Do not ask for whom the bell tolls - it tolls for thee.   (And if you didn\'t figure out which writer guy was born in STratford-upon-Avon, you either a) got confused by my German or b) are an idiot who shouldn\'t be allowed to speak English.)  EDIT -  urlLink ORFIE LIVES! ORFIE LI-HI-HI-HIIIIIIIIIIVES!  '),(26335,2606936,'20,may,2004','2004-05-20 00:00:00','   I\'ve become a \"damn\" liberal.  You see, when i\'m driving to school in the morning, i enjoy listening to the Don Miller Morning Show, which is on my local alternative/rock station, 99X. (Yay, atl.) They are pretty controversial because they are frank and not at all PC, but still muy funny. And if people would only a) understand sarcasm and b) use moderation, there\'d be no problem at all. For the record, there is no \"Don Miller,\" and they explained that a long time ago. They made up a random name for humor.  Now, last week a great controversy erupted because some guy on the show commented that he found it annoying when people on bicycles rode in the middle of the street at slow paces and wouldn\'t allow cars to pass them. Some \"cyclist\" took this the wrong way and suddenly all the cycling nubs of Atlanta were on the show\'s case, writing these great letters about how people were always persecuting them, comparing their plight to that of blacks, and threatening all kinds of crap.   Normally, I really enjoy stuff like that because honestly, there are a lot of idiots in Atlanta and you couldn\'t make that kind of comedy up if you tried. But this morning, they read a letter from a pissy conservative who brought in a whole bunch of trite things about Bushie-wushie. And this guy actually wrote a letter to the FCC about how 99X was spewing \"liberal Anti-American propaganda.\"  You see, it\'s all fine and good to support Bush, because i know, love, and respect plenty of people who do. It\'s your fucking right. I\'m okay with that. It\'s America.  But it just downright PISSES ME OFF when people claim that NOT supporting him is \"anti-American.\" That somehow, THEY have the right to support their own beliefs, but IIII do not. Like, it\'s only patriotic to agree with the government.  Uh, yeah, in Cuba.  I hate very much that such a small percentage of Republicans I know are actually reasonable when it comes to politics. How, when you look at the nation, it\'s not really that many that are unreasonable anyway. How a percentage of IDIOTS ruin their whole thing, and say shit in the name of their party.  BUT i hate even more that they get away with it.  So, in a small bout of retaliation, I am going to deconstruct this idiot\'s letter. I don\'t know where to start on his beliefs, so we\'ll just hold our noses high as we observe his shitty grammar. My comments are   BOLD.   Enjoy.   OFFENSIVE MATERIAL ON WNNX   Good day gentleman.  I think he\'s writing to more than one guy, and needs a comma after his introductory phrase.   Of course, you may not have good days much longer, thanks to your lack of respect for our President and the FCC.  That syntax is just gross. If I agreed, i might rephrase it for him.  Just so you know what was said, I\'ve attached the message I sent to the FCC last Friday about your show. I guy work with said Jimmy claimed to make $300K a year.  Nice grammar, buddy.  That\'s great, because it will only take him a few years to pay off his portion of the fine. You other two might have a little more trouble.  I tuned this morning  oh, he must be a musician  just long enough to hear that you\'re having America hater  should be hyphenated  Bill Marr on your show tomorrow. The same Bill Marr who said the 9/11 terrorists were heroes and we are the cowards.  Nice fragment.  Unless you want another complaint to the FCC, I suggest you either cancel the interview, or at least do everything in your power to make him understand how wrong he is. I will be monitoring your show tomorrow to make sure your don\'t allow any more indecent material on the air.  In your sick twisted minds,  commas, anyone?  you probably think what you did was alright and that I\'m a jerk. Well if you had kids your attitude would be totally different.  commas after introduction and subjunctive.   -Josh  ANyone else like how he started off formally and ended with a dash?   Honorable Chairman Powell,  Suck up more, please. Oh, wait, that\'s not grammar. Oh well - he\'s still being dumb.    I am sending you this email in regard to inform you  WHAT? What are you, retarded? What kind of affected and grammatically incorrect phrase is that?  that some very offensive material was aired over public radio in Atlanta recently. On Thursday May 6  welcome to writing dates correctly  at about 9:00 AM, the \"Don Miller Morning Show\" on WNNX 99.7 FM allowed an anti-American, liberal so-called comedian named David Cross to come on their show and talk at length about how the current war for Iraqi freedom is bad, and how our esteemed commander-in-chief, George W. Bush, is terrible.  \"Big word big word big word big word.... uhh... bad.\"   The Don Miller Morning show allowed this David Cross character  sounds like an old West sheriff  several minutes to spew liberal anti-American lies and propaganda.  I love the diction of vomit here - it\'s so creative! \"spewing propaganda\" - whoever heard of such a great use of imagery and metaphor? Bravo!  They did not try to counter his arguments or shut him down. Many of the cast members even vocally agreed with several of his \"points\".  punctuation inside the quotes  He cast dispersions on our President,  Does anyone else know what the hell he means by \"casting dispersions?\" ... maybe he thought the big word would scare us.   questioned his intelligence, and even had the gall  oooh.... brain cramp  to question our  whose? are YOU part of the war?  motivations for the war for Iraqi freedom. All of this was said over public airwaves and could be heard by impressionable children. I think everyone agrees that freedom of speech is important, but their are limits, especially when children could be given the wrong idea about our leaders and our country.  Well, that\'s debateable, but I think we can all agree that children should never be allowed to think freely.   In addition, while corresponding with the Don Miller Morning Show by email over this matter, I was called several unmentionable names and my intelligence was derided.  Again... does anyone know what \"deriding intelligence\" is?  One of the members of the show, \"Jimmy Baron\",   comma inside the quotes  who has admitted on air to being Jewish,  admitted? What, is it embarrassing? even took the name of Jesus in vain.  Well, God forbid a BUSINESS should allow other religious beliefs!   According to their website, the members of the Don Miller Morning show are: Fred \"Toucher\" (the head), Jimmy Baron, Wally (last name unknown), Leslie Framm, \"Crash\" and \"Calandro\" .  punctuation inside the quotes!  For your record, their is no \"Don Miller\".  first of all, learn to spell \"there.\" and for God\'s sake, PUT THE F*CKING PUNCTUATION MARKS INSIDE THE QUOTATION MARKS!  I have only had correspondence with the first three on this list, and I can only assume that most or even all of the show members are using fake \"show-biz\" names.  Thank for the attention I know you will give this matter,  thank WHO?  and thank you for the fine job you are doing in cleaning up America\'s airwaves.  Your Fellow Patriot, Josh   Sucking up.... and if he thinks Colin Powell is as stupid as him, he should be shot to prevent his genes from tainting the rest of humanity.     He also misspelled Leslie Fram\'s name, and he probably doesn\'t even know what \"FCC\" stands for, as he neglected to use its full name of this government institution in \"formal\" letters.  Moral of the story - don\'t write rude letters unless you write them WELL.  (taken from  urlLink the 99x website. )  '),(26336,2606936,'18,may,2004','2004-05-18 00:00:00','   rrrrrrrrrrrrrrrrrrrr  I just stumbled across  urlLink this  lovely website.   Basically, it\'s some conservative in Cali bitching about how liberals are trying to take over the world by persuading people. Or something.  But since it\'s annoying to have people write negative shit on ur blog, i\'ll do it all here....   \"Like most Americans I was stunned. It was a body blow,\" said Rumsfeld, who has been fending off calls for his resignation over a crisis threatening President Bush\'s bid for a second term in November elections.  Some analysts say they are even wondering whether the photographs of prisoners being abused will have the same effect as television pictures of the Vietnam War which eventually turned many Americans against it.\"   Apparently this is liberal bias because it might persuade someone not to agree with the war, and because the more gruesome pictures of what THEY did to US aren\'t shown, people don\'t get persuaded by them, or something.  1) Uhh, obviously there are standards? The fact that \"tha Ay-rabs\" may have committed worse things, that violate censors\' standards for putting pictures in the public domain? (even though we easily get them online anyway.)  2) You see, apparently we are all SO mindless that seeing pictures of torture, next to explanations of how it wasn\'t the whole army, will make those of us who love and admire Bushie-wushie STOP loving him. Since, ya know, we don\'t think for ourselves or anything.   Come on... people who actually LIKE bush are soooo in love with him that they simply scream \"propaganda!\" and ignore it. People who a) don\'t care or b) hate the man, just get more evidence on why he sucketh.  3) \"Well, ok, somaybewemighthavedoneabadthing  BUT we\'re NOT AS BAD as  THEM!\"   is getting to be a very annoying mantra. Dude, we  know  they do bad stuff too.   But come on..... aren\'t the conservatives the ones who are anti-abortion? ya know, \"it\'s your choice! don\'t have sex bc someone tells you to! if everyone jumped off a bridge, would you?\"   so what\'s up with this sudden \"but MO-OM, everyone ELSE\'S government lets them break the geneva conventions! it\'s not fair!\"  4) God.... the fact that someone actually sits around and infers \"bias\" from news articles so he can bash liberals. Man. We worry about hatred from stuff like racism, sexism, religion..... and right here at home, we have people like this. People who can\'t even see around the politically-rigid glasses they\'ve built themselves into, who hate people who disagree with them so much that they build websites about it.   The sun is setting....  5)Whoever wrote the article this guy spoke to, used an unclear pronoun reference that makes my head hurt. So basically, you both annoy me.  Uhh, have a nice day.   (And by that, of course, I must mean \"kill a Republican.\")  '),(26337,2606936,'15,may,2004','2004-05-15 00:00:00','   and it was GOOOOD.  However, for those of you who don\'t know, I wear my hair long and straight, with long bangs at the front which i can part and push over or whatever... It is extrememly similar to recent haircuts of  urlLink Britney Spears  (blech) and  urlLink Lindsay Lohan  (woohoo!), who stars in the movie.  Anyway, so i\'m admiring how LL\'s hair was fixed in this movie, which was all about NOT being superficial, and realize that mine have grown out a lot again. I was going to trim them just before prom, but held off in case of hairstyle problems, and they were just..... everywhere.  So i came home and, having learned my lesson about appearances, cut my bangs.  They are now too short and make me look something like  urlLink this  unfortunate girl, although they are past my eyebrows...  But hey, at least now my lesson comes into use.... i hope.  '),(26338,2606936,'12,may,2004','2004-05-12 00:00:00','   my toes are dark red, and that\'s weird. I had prom last week and went for the boring, regular toenail color. I really really want to change it to orange, but doing so would ruin my fabulous french manicure.  We took a trolley to prom. It was awesome, since we had a group of about 18 people. The trolley was wood and pretty old and kind of falling apart, but it was cheap and it was FUN. It had one of those cool dingy bell things and we got to ride out on the back of it and wave at people who were wondering why all these teenagers were dressed up and riding on a trolley.  My hair was biiiiiiiig. And I had an awesome date who is a good friend of mine. And i looked great. And I had fun dancing with people and it was just awesome.  But I realllllllly reallllllllly missed him, him, HIM, a lot. Especially since he didn\'t even go. And because it felt weird to dance with someone else. And because I bought that dress for his-and-my \"future\" prom a long time ago. And while my friends are bodacious, I think I\'m still sad, especially in light of the past month\'s events.   And then he saw my pictures and didn\'t even tell me I looked pretty.  If you really want to know.  '),(26339,2606936,'06,may,2004','2004-05-06 00:00:00','   On Saturday, I babysat at a party. There was one family there, two parents and one child. The dad had a name like Martin or something, but the mom\'s name was Cinnamon. The I met their daughter and found out her name was Cayenne.  At Music Midtown, I kept seeing this couple walking around. The woman was very short and dressed modestly, in a T-shirt and long shorts, but had a shaved peach-fuzzy head. Her boyfriend/husband was very tall (Tim and shaved Allie, anyone?) and wore no shirt, but had a green camo cowboy hat on. As they walked, he was absentmindedly rubbing her bald head.  And this is the best...... Lit/Lang students, chew on this.  A week or so ago, I was on my break at work and was sitting outside the cookie place doing HW. Two little girls walked by, and they were about 8/9/10 years old. One was wearing a hot-pink spaghetti-strap top and a more modest little girls\' version of those fluttery skirts you see these days, made of denim, and heeled sandals. Her friend wore a big blue t-shirt and big shorts and flip-flops.     As they passed KB Toys, which closed recently, the one in blue stopped and peered into the window. Her friend, annoyed, called her to come on.  '),(26340,2606936,'01,may,2004','2004-05-01 00:00:00','   shamelessly ctrl-c-ctrl-v\'d from  urlLink And another thing...    A meme   I want everyone who reads this to ask me 3 questions, no more no less. Ask me anything you want and I will answer it. Then, I want you to go to your journal, copy and paste this allowing your friends (including myself) to ask you anything.   My question now is, or are, a) will you accept?    and b) Is \"meme\" Canadian for \"memo\"?  '),(26341,2606936,'26,june,2004','2004-06-26 00:00:00','   The Schnig. Henceforth, I think I shall call him \"B,\" for reasons I ought not to divulge.  ANyone who\'s read maybe one-third of my entries on this page know who this guy is. On-again-off-again boyfriend, best friend, worst enemy, future husband, support system. We have an odd relationship, but it\'s one of the strongest bonds in either of our lives. Current status = LURVE him to death, except for his ideology....  See, he was raised kinda-sorta in a  urlLink religious cult . The Way International? Yeah. I mean,  urlLink this (Christian) ministry  has good religious habits, and they\'ve done a lOT of research that\'s pretty damn important in the Christian world, but most people in it are also weird.   He lived for two years on a ministry communal farm thing.  However, since his mother\'s divorce of a \"Ministry man,\" and the implied shunning from so many of those people, their family pretty much can\'t stand the Way any more, so they just worship outside of it.   ANyway, he\'s got a very strong background in Christianity, and a freakish amount of historical knowledge stored in his brain. I\'m talking, this is that freshman who tested out of AP history classes. This is the guy I use when I\'m writing papers and studying for history classes. He just RETAINS it. So, of course, he has all kinds of religious history committed to memory.  He doesn\'t walk around talking about it, and most people don\'t even know his religion because he\'s very laid-backa bout his faith, but in truth, he\'s Christian through and through.  Being agnostic myself, there are a few issues we just will never ever agree on, ever. One of these is homosexuality. B is all, \"Homos homos homos, gross gross gross, Bible Bible Bible.\" I\'m all \"SYmpathy sympathy sympathy, humanism humaism humanism, rights rights rights.\" It doesn\'t work. At any rate, one of his endearing nicknames is \"my little homophobe,\" which is actually pretty dysfunctional, but it still makes me laugh.  So a few nights ago we were talking on the phone and at some point he remarked that he couldn\'t think of anything worse than having to watch two men have sex. Sick of his \"gay\" problems, I pointed out that there were much worse things to watch. When asked for an example, I brought in a disgusting comment made by one of our other wall-humping male teenage friends earlier in the day, and suggested bestiality. He denied it.  Basically, I\'d rather watch gay people have sex than a person and a horse, and he\'s rather see the other.   I\'m not so sure why this is relevant to anything, except that it reminds me of the \"Bush or Kerry?\" question..... pick between two undesirable things?  Now, unless you\'re both gay AND bestial, this is probably somewhat YUCK for you..... still, I\'d like to play this age-old game and ask, what would YOU rather watch, is you had to watch one? (The sex, not the Presidents.)  '),(26342,2606936,'25,june,2004','2004-06-25 00:00:00','   I\'d just like to point out that, being in the middle of P.E., I have not learned a lot about lifetime fitness. I have learned WHY you need to be fit and HOW to do it. However, I still abhor exercise. I hate sweating and hurting and wasting a fucking hour of my life walking and not going anywhere!   You\'d think they\'d have figured out that the reason some people don\'t exercise is because they HATE it, and maybe try to fix that. But instead of the great teachers who show you their subject don\'t suck, they instead drill you with facts and give you the most boring and horrendous schedules to follow. I\'m amazed at how something so incredibly NOT INTERESTING can still be so ouchy.  Is it really healthy for me to walk for an hour a day in my target zone? I mean, a whole fucking consecutive hour? I truly doubt that\'s a good idea, unless you\'re into pain and injury and fatigue. I was fine when the workouts were only half an hour - I felt GOOD then. That was GOOD pain. Now, it just hurts because the only way I can finish the wokrout is to push myself so hard it hurts beyond any fathomable promise of benefit.  Now I feel this mundane, endless task ahead of me every day. It\'s five-o-fucking-clock and I can\'t make myself do it. I just don\'t fucking want to exercise right now. I\'ve had a tiring day and I need a nap, not an hour of hell.   I have to go walk in el THRZ-o for an hour and then do pushups and shit. Pushups? I love pushups. I eat those for breakfast.  I\'m even fine with walking, even for an hour, except i have to walk within my target zone for an hour. This doesn\'t include the time you spend warming up, cooling down, and resting from overexertion, not to mention taking a sip of water and hearing the fucking heart rate monitor beep and tell you to work harder, bitch. No nutrients for you.  This leads to pain and cramping and soreness and for some reason, it\'s incredibly fucking DIFFICULT to get into the target zone for me lately, even when i\'m feeling motivated. Maybe because I\'m so pissed off. So I work harder and harder and I already absolutely fucking hate this course and I\'m barely halfway done.  My soul is all shrively and gross. And my legs hurt, too.  It\'s people like Miss Biological Changes Over Time that consider this an important part of my education. [Put down the hairspray, lady. Actually, pick it back up and stop coming to work.] I KNOW people are getting obese and stressed out and shit, but i certainly don\'t think overexhaustion is a good alternative. If this thing were only a bit less ridiculous, I\'d be just fine. But I cannot possibly live through two more weeks of hour-long pain sessions.  As much as I try not to hate things without good reason, I HATE P.E. If this course was a person, I would push them off a building and make it look like an accident.  '),(26343,2606936,'21,june,2004','2004-06-21 00:00:00','   My brother is currently at Boy Scout Camp. That\'s where the Boy Scouts go and be gross for a week. They work hard, sleep too little, live in the boonies, and eat disgusting food.  So my dad\'s telling us about dropping him off, and talking about the counselors and stuff, and mentioned \"she.\"  I did a double take (well, not really, but wouldn\'t that look nice in a restaurant?) and was like, \"there are FEMALE counselors there?\" My father snorted and affirmed this.  They don\'t allow GAY MEN to work in the Boy Scouts, but WOMEN are okay.  Say WHAAAAAAAAAT?  I\'m completely confused. I mean, I thought people were justifying this whole anti-gay-scouting* thing by saying that gay men would either rape the children or display something less than the masculinity desired by the boys.  So in light of that new stuff..... I\'m sorry, but how the hell can someone rationalize this? Are they saying women are incapable of raping people? Or that women - actual FEMALES who are actually feminine (and I mean this in the most un-loaded of ways....) - are better suited to teach masculine values than men? Actual masculine(again, the most un-loaded of ways) men, with penises and th4e ability to grow facial hair?  I\'m pretty sure anyone can rape ANYONE. At least, that\'s what they told us in Health class, and those \"my body\'s nobody\'s body but mine\"-type videos.   And by \"masculine\" and \"feminine,\" I mean \"of a man\" and \"of a woman.\" Not \"rugged and musky\" and \"flowery and delicate.\" I read the essay on the English language - I know when to be careful.  But how are they pulling THIS off? Are gay men just evil, evil people or something? Are they just going to come out and SAY that they hate gay people? I mean, it\'s about time SOMEONE was honest about it.  More likely, they\'ll continue to rationalize it.    *In the meantime, I think the Gay Scouts would be a cool organization..... find your favorite homosexual and spread the word!  '),(26344,2606936,'17,june,2004','2004-06-17 00:00:00','   I kissed him and he kissed me back.  '),(26345,2606936,'15,june,2004','2004-06-15 00:00:00','   I don\'t think I outlined the last LAST time I spoke with the ex. The Schnig, if anyone remembers.  You see, I have this friend, T, who is a guy. He is absolutely awesome. He is over six feet and his girlfriend, one of my best friends, is under five feet. He is sweet as can be and despite often claiming to be a reject, he gives great advice. A hug from T can make my day go from shitty to great, because he just has a good sense of people.   T is the one who .... well, this one time the Schnig and I had a huge fight (post-breakup) and then he showed up at Chili\'s after a football game, where my friends and I had made plans. The fight had been about how I said to someone \"I don\'t want the Schnig to come with us. He makes me sad,\" but he got la pissed and made a huge deal out of it. I then sucked it up and asked him to come sit with us, and he acted as though nothign had happened, declining because he was \"going out.\"  So I sat on my car trunk and cried and was too embarrassed to go back into the restaraunt. So T came out, picked me up like a baby, and carried me inside.  One of the top ten best moments of my life.  So anyway, though T is awesome, he has one flaw - he\'s a teenage boy. Which means he\'s got sex on the brain. And tho he has stated that he won\'t do NOTHIN\' before he\'s married(damn, he makes ME look like a slut!), he makes sex jokes ALL the time. Pervy stuff. Like, I\'ll say \"Oh, Jenny\'s coming to the movies with us!\" and he snickers.  So he had come up with this \"great\" joke where he held his hand about two feet away from my butt and acted like he was an old-man pervert, or something. It was actually quite funny, but it got old fast.  But then I made some joke and he was like, \"Audrey, don\'t be such a perv!\" I reply, \"Well, you\'re the one who\'s been grabbing my ass all day!\"  The Schnig: \"Well, it is a NICE ass.\"  We\'re in the fucking BAND ROOM, for crying out loud. In front of PEOPLE. And TEACHERS. And to look at his face, I have no idea if he was even kidding!  Embarrassing. Oh god. It was awful. I don\'t even know who heard it!  And also, there\'s that whole thing about how I don\'t like guys to make comments like that to me. Especially not my \"friends.\" Easpecially not the ones I\'ve dated.  T\'s eyes got kinda big and he left after a minute. I rememeber this vaguely - I think i playfully smacked the Schnig in the stomach as he said it,t hen realized what he had said, gave him a nice eyeroll, and walked away.  I vowed not to speak to him till he apologized. And we hadn\'t spoken in almost a month.  But then tonight I really needed to call my friend and I had lost her new phone number. So I had to call him to get it. And in talking to him, I kind of forgot I was angry. And then we talked for two hours and made a date to have breakfast tomorrow. Oops.  No, not THAT kind of a date.... I hope. But then, I\'m still gonna shave my legs.  '),(26346,2606936,'09,june,2004','2004-06-09 00:00:00','   I haven\'t done a crochet post in a while, but I feel obligated, on behalf of the yarn-purchasing population of my area, to tell you about my love affair with yarn.  Back in the day, I\'d assume the 70s judging from my mom\'s experience, there was not a whole lot of choice when it came to yarn.   There was a lot of worsted and a lot of fingering, and a LOT of sport.   There were solid colors.   There was scratchiness, whether it was synthetic(acrylics? polyesters?) or natural(sheepses.)  Today is different. Today, we have so many weights and textures - bulky and chunky! Micro-fingering! Eyelash and boa, mohair and boucle, chenille and cotton!!! They even make yarn out of Polartec and terry!  We have variegation! We have tweed! We have no-dye-lot!  We have baby yarn, soft pastels!   Kids\' yarn - the thick, primaries-and-neons variegations!   The \"teen\" yarn, the stuff that made knitting \"cool\" - the fuzzy stuff to make those [revoltingly popular] eyelash scarves, and the honorable chunky, silky, colorful Homespun - guranteed to make even the worst stitching look elegant!   The stylish \"adult\" yarn - loopy boucles in bright colors, soft mohairs in earth tones!  And the rest of them - the denim print! The magic-stripey! The sparkly holiday yarn!  So much yarn.... so much beautiful, soft, ARTISTIC yarn. So much choice!  And when you dream up a pattern, and you look online to see what could make that pattern come to life....   You realize that it\'s just so soulless to buy yarn online. You have to see it in real life! (And real light!) You have to touch it! You have to feel its warmth and its weight!  You have no choice. You MUST shop.  There\'s a sense of urgency as one embarks on a yarn hunt. You enter the craft store knowing EXACTLY where to go. You envision the project that will begin as soon as your perfect yarn has been found. You hope and pray that IT is there, that it hasn\'t sold out, that it\'s stocked, that it\'s in the right colors....  And you end up face-to-face with an amazingly HUGE aisle of yarn!  (Isle of Yarn?.... what an idea.)  Two huge cases, reaching to the ceiling, filled with color! Rolls stacked on balls stacked on skeins stacked on those huge one-pound bags! A still-frozen avalanche of fibrous wonder, just waiting for you to enter!  You can\'t decide where to begin! Do you slowly pace up and down, giving each style the consideration it deserves? Or do you jump at the first one that catches your eye, even though there are NORMAL PEOPLE watching?  What am I talking about?...... NORMAL people don\'t craft-shop.  You jump!........ and it\'s a miss.   But the first one\'s always a miss! You never buy the first prom dress you try on, right? ... so there must be more!  You jump about, as more and more hook your attention. ALAS! It\'s all-wool-blend.... or it\'s super-duper-fine fingering weight.... or it\'s really scratchy, upon touch......or it\'s actually......well....... that one\'s kind of ugly, now that you look at it closely.  But wait.... you\'ve scanned the whole aisle! Why hasn\'t YOUR yarn jumped out to greet you?  Is it hiding in the sale rack?..........No - just Easter and Christmas stuff here.  Is there another aisle? Is it on the end?  Okay.... okay. Time to pace up and down, scanning each type.  So... here we go.... It\'s almost found! Almost there!  Blue acrylic sport weight. ..... royal blue acrylic sport weight.... magical blue acrylic sport weight... dusky blue acrylic sport weight.... red.... scarlet.... gules..... tan.... beige..... cafe con leche........ cafe au lait..... cafe mocha.....cafe.... coffee........brown... mahogany.... ebony..... black.... black.... white.... white... white....   acrylic.  sport.  weight.  Can it be? An entire HALF an aisle devoted to five different companies making the EXACT same yarn?   Instead of stocking up on the fudge brownies, ladyfingers, and strawberry tarts of the yarn world, this store has about 7 different kinds of packaged, processed, preserved, and, i might add, eerily PERFECTLY ROUND chocolate chip cookies.  And so did the next store. And the one after that.  So many kinds of yarn. So many stores.  So much ugly yarn in stock.  And, as I\'m discovering, such high shipping costs for online yarn shops.  '),(26347,2606936,'06,june,2004','2004-06-06 00:00:00','   Administrative thoughts:  Damnit! I guess all those people who got a rush out of being nasty to me didn\'t even care enough to check back.  That\'s ok, tho. I seriously hate internet flame wars, especially among people I don\'t know! it\'s better that one didn\'t start, and had those people checked back, they\'d probably be as angry as I was that they took what i said so oddly. They\'ve probably found someone else to \"hate\" for a day.  However, I did find the interesting blog of  urlLink David.  It\'s political stuff, written by David and his friend(whom i suspect to be a robot), but it\'s a lot more thoughtful than most any politically-focused blog out there. I recommend - these people/androids have a lot to say, but they\'re generally low-key and non-confrontational - just observations about our crazy system. Good reading, kiddos. -------------------------------------------------------------------  Now, something someone said recently sparked a thought train for me. How about the concept of \"responsibility,\" or even \"duty?\"  There exist many kinds of responsibility. The first one you learn is \"responsibility for your actions,\" which translates to children knocking over vases and then owning up to it. Obviously, everyone needs to keep this, from that child\'s play to the President\'s foreign policy.   This is a very important part of our society, especially with the capitalism. I mean, we don\'t do the whole \"Glory for God!\" thing anymore, considering that the amount of people who don\'t care about God, or those offended by the assumption that they DO, is gradually increasing. We still have the Puritan work ethic, though - work hard and you get rewarded. If you don\'t take responsibility for what you do, whether it be curing cancer (i wish) or murdering your wife, you never get what you deserve for the action.   Certainly, good deeds on the DL can be great, like anonymous donations to charity, or helping someone you don\'t know and never contacting them again. SO, I guess, the American child-rearing focus here is more on responsibility for BAD deeds. This is also important - guilt just isn\'t healthy. (Or maybe someone else gets put on death row, but it all comes back to guilt and fairness.)   And then we have the responsibility that comes with, for lack of a better word, possession. The responsibility of a parent to his children.... ... of a person to his car, or whatever toy he\'s got.... This is what you create yourself, because of the benefits involved. Sure, having children costs money and energy, but it\'s also probably the greatest kind of love a person can experience. (this comes from a non-mutter, but I assume from the attitudes of others...) It may be hard to keep up a car, but it benefits your life. But what if, for example, you child is born autistic? - or you aren\'t such a good driver? - It demands much more than what you expected! That kind of responsibility is the risk we take in improving our lives.  And then there\'s the whole \"working-world\" thing. The basis of it, to go economic, is that an employer has a responsibility to compensate his employees for their work without cheating them or abusing their help. Likewise, the employee is reponsible for performing the duties they promise to do, so that the employer does not waste money in overcompensating them.  This is why we have minimum wage laws, child labor laws, etc. This is why we have sales goals in retail and work codes of ethics.  Or, also, the responsibility of a government official to his constituents. He has to show that they have not elected him in vain, that he will keep his promises, that he will do his job to the very best of his ability.  Or what about the responsibility of a teacher to his students? To do the best work possible, that they may gain the best knowledge possible.  And then there\'s also the responsibility of using resources. Humans try to not use up all the Earth\'s fossil fuels, or kill off all the animals. Students work hard in school, because they feel responsible for making ht ebest use of what\'s available.  Or the responsibility of improving things - to not just \"cut corners\" and do the minimum, but to really make something better. Taking a hard class at school because the knowledge will pay off... allotting government funds for the arts....   And i think that leads into what I consider to be the most important responsibility a human being can assume - the responsibility they owe to other people. We have a responsibility to treat others WELL. To be kind to them... to avoid infringing on their rights. To sometimes give up something less important that we want, so that a greater amount of people can live free and happy.   To drive safely when there are children around.  To bake casseroles for a family who\'s suffered a loss.  To show your support for hard-working politicians who so rarely get the recognition they deserve.  To curb your anger at your job when speaking ceremonially to a waiter you\'ve never seen before and will never see again.  To give what you can in a romantic relationship, and not to simply take for granted what your partner may do.  To NOT act on animal impulses - to not rape a pretty girl, to not smack someone you find annoying.  To care about the well-being of all humans, even those who will never affect your own life.  This is a responsibility we all have, and one that most everybody acts on. If we can all view each other as active holders of this responsibility, then maybe we\'ll treat each other as equals.  We all have the capacity to perform the basic human duty we owe each other.  Perhaps i\'m a little idealistic today, but you\'ve GOT to realize that responsibility is not ALL about actions and possessions and work - it\'s also about existence in a society.  And if we can just try to fulfill what society needs of us in our attitudes toward others, then maybe our society isn\'t as short-lived as some predict.  Maybe it IS a rising sun on that near-proverbial chair, after all.  '),(26348,2606936,'03,june,2004','2004-06-03 00:00:00','   Does anyone, out of the four or so people who read this, remember that ironic post i made a few measurements ago? It was when I read some weird \"liberals suck!\"  urlLink website , criticized them, and then wistfully remarked that people should stop criticizing each other?  Yeah... well, the guy who wrote the post that sparked me, read MY post and skipped the sarcasm. SO nwo he and his buddies think I am an illiterate, fat, liberal, teenaged relative of Michael Moore, or something.... And they think I am angry at them, or something. (Well, I am kinda angry now, but I wasn\'t when they went berserk on my comments option.)  And they criticized my grammar and tone!  urlLink Who would EVER be that nerdy?   I dunno. I wish sometimes that people would take things with a grain of salt, but I\'m pretty sure my post made at least one of you smile. And if I can just stop one person out there from saying shitty things about \"the others\" without thinking about how pointlessly rude it is, then I can die happily.   And since so many of these people would like me, my references to books and music, and my damn colloquial tone dead, then I think everyone will be happy.  FROM THE PISSED CONSERVATIVE DUDE\'S SITE - SOME COMMENTS ON ME: ----------------------------------------------------------------------------  Posted on May 31, 2004 @ 10:24 pm by Jason  \'ur\' blog is betta, Lee. like, fer shur. that dude is all, and i\'m all, and he\'s like, whateva...    DUDE! What are you, a sped? No one says \"fer shur\" anymore. It is \"fo shizzle.\" GAWD!!! *twirls hair around finger bershonly* ------------------------------------------------------------------  Posted on May 31, 2004 @ 10:24 pm by Nick   Wow,  Still, he (or she) has zero comments, so maybe he (or she) is actually a complete fucking loser as well as an obvious fuckwit.  He (or she) couldnt get a clue in a field of clues, covered in clue attractor on the clue-iest day of his (or her) life,  Nick   Clueless, you imply? But I ALWAYS figure out Blue\'s Clues! (note the link to my man Steve Burns at left.)  -------------------------------------------------------------------------------------  Posted on May 31, 2004 @ 10:29 pm by Sipidation  Read some of the other great post like:  Samstag, Februar 14, 2004 \"forty-two\"  Some of my friends tease me about being a \"sex addict.\"  Freitag, Februar 13, 2004 no, not the sandal  I am wearing a thong.  For the first time in my life.  And ya know what?........... I don\'t hate it.   You would think these guys would get the \"forty-two\" reference.... but since they obviously don\'t read entire posts, that wouldn\'t really matter.   --------------------------------------------------------------------------------  Posted on June 1, 2004 @ 12:13 am by Matthew Spence  I think her list of links says all you need to know about little audrey.    Oh no! So, reading blogs is.... NOT cool? NOT intellectual? NOT politically involved?  \"I\'m Captain Obvious! ANother mission complete!\" ----------------------------------------------------   Posted on June 1, 2004 @ 8:35 am by Josh  \"Bigger than my body gives me credit for.\" Could this be Michael Moore or his daughter?    Um, if \"I\" am bigger than my body, doesn\'t this mean that a) my body is small in stature or b)that my \"soul\" is big?  I dunno.... I guess you guys know everything when it comes to figurative language.  -------------------------------------------------------------------    Posted on June 1, 2004 @ 10:08 am by Paul  Clearly this asshat has never actually read the Geneva Conventions. It\'s amusing to see the phrase \"Geneva Conventions\" bandied about by people who have no clue about the contents.    Oh yeah. I totally stated that I read them, eh? Actually, what I got from History of the AMericas was that it had standards for how POWs are treated. But then, I have no clue, right? ---------------------------------------------------------------   Posted on June 1, 2004 @ 10:12 am by Drumwaster  Or the phrase \"International Law\" (with both words capilatlized, as if that would make it both real and applicable).    Yeah, cuz I said that. *eyeroll* And your spelling AND grammar make me want to vomit.... liberal propaganda, that is! ---------------------------------------------------------- AND FROM MY OWN COMMENTS. -------------------------------------------------------------------  Was there a point to that incoherent rambling? - Brandon    Yes. It showed my own calculated hypocrisy, essentially trying to tell people to NOT be hypocrites when it comes to this whole stupid \"right-vs-left\" thing.  --------------------------------------------------------------------------------    Does the geneva convention apply to non-military combatants? - Jason    Nope. Well, I\'m not the lawyer in this family, but I have always understood it to be a \"time-of-war\" kinda deal. Bu I\'m a stupid liberal asshat, so maybe you should ask someone else. And am I stupid, or is it \"Geneva COnventions?\"  Anyone else seeing a new kind of advice column? \"Dear Liberal Asshat....\" --------------------------------------------------------------------------------    http://ian.gerdesdesign.com/images/news/liberal.jpg - marc    Yes, ma\'am.   I\'ll gladly stay out of combat.  But don\'t use the word \"pussy\" around me again - I have an irrational cat phobia, and the chauvinistic implication of the way you used it is awakening the feminist within.   Grr.  --------------------------------------------------------------------------------    What the hell are you on about, you stupid American douche-bag. Iâ€™ll agree that this article is a poor example of liberal bias but only because there is so much more out there, in greater quantities; however if you really canâ€™t perceive the bias here then I pity you.  Tell me when was the last time you saw the mass graves newly found in Iraq being plastered all over the media? I doubt even Fox News went that far. And yet your talking about 1000s of lives gone, executed; and all they get is a small slot on CNN, gone in a day. Now when was the last time the prisoner scandal wasnâ€™t plastered all over the media? UN OFF scandal? Nope. Bet you donâ€™t even know about that. - Poosh    Wait.... when did I talk about \"1000s\" of deaths? Are you even on the right page?  I would appreciate it if y\'all would read my posts before attacking them, mm\'kay? *head bobs, teeth sparkle, ponytail bounces*  Also, grammar = SUCK. But let\'s ignore that. --------------------------------------------------------------------------------    Such drivel, young \'un. Hope you learn to write someday, keep practicing. And when you grow up and assume some responsibility in this world you\'ll look back and realize what a little idiot you were. - Bill OH    One of my managers at one of my jobs calls me \"littl\'r\'un.\" Is that what you meant?  And as for the responsibility thing, we won\'t go there, especially after the post you\'re referring to. Obviously, OH ain\'t no TOK.  (or LA - your grammar is pretty sad, too.) --------------------------------------------------------------------------------    ON the one hand you deplore right-thinking.com\'s author\'s lack of tolerance and political rigidity, on the other you say that \'the sun is setting\' because essentially you disagree with his opinions. Those in glass houses....  I do agree that the US should be held to higher standards in its treatment of prisoners than dictatorial Middle Eastern kleptocracies. I don\'t think that the rationalizations for the prisoner abuse are so much an ends-justify-the-means as a \"yes, we\'re human, and imperfect, but lets keep this in perspective\" thing. Yes there are limits to this view, but I don\'t think that you should dismiss it out of hand. - David   David, I think you\'re cool. Thanks for saying what you want to without being rude about it, as well as highlighting my point for the others. And I think \"kleptocracy\" is one of the best words I\'ve heard in a long time. (begin sarcasm) you are going on my \"friends\" list that I have to start to prove that I have them. (end sarcasm)  The \"sun setting\" thing actually referred to how I foresee Americans killing each other because we get so angry about disagreements. But thanks for pointing that out and giving me a chance to clarify, anyway. You\'re hot. --------------------------------------------------------------------------------    (And by that, of course, I must mean \"kill a Republican.\")   Such peace, love, and tolerance.  At least Lee doesn\'t say to kill people for their beliefs in his blogs. - Josh   hahahaahhahahahhahahahahahahahhahahahahahahahahahahahahahahahahahahahahahaha *breathes* hahahahahahahahahahahahahahahhahahahahahahahahahahhahahahah *breathes* Oh, man! You can\'t make this stuff up. You really can\'t.  If everyone in the world thinks like you, I can understand why these people think I harbor ill will toward them. Verbal irony, anyone? Point, anyone? --------------------------------------------------------------------------------    I think its pretty clear that she\'s being facetious with the \'kill a Republican\' statement.  But in her post, she seems to doubt that there is a general liberal bias in the media, and seems to find conservatives of her acquaintance fanatical. I would hope that she would not be so rigid in her worldview so as to not be able to see what is in front of her, or be unable to have a respectful conversation with someone who disagrees with her. David    Again, David has kindly assisted the meanies in understanding me, while still saying what he thinks about my opinions. Criticism WITHOUT malice? Could this be what I was searching for in the post?!?!?  David, will you marry me?  --------------------------------------------------------------------------------    Audrey,  \"An American teen. Bigger than my body gives me credit for.\"  But apparently larger than you own sense of self-importance.  If you\'re going to Fisk someone, at least have the presence of mind to counter with good arguments; otherwise, your words will be and have been taken as nothing more than the vomit of your mentors.  Further, I could not find this quote anywhere in Lee\'s post or the comments: \"tha Ay-rabs.\" This must be a racial slure of your own fabrication?  Before you slam other people, take a closer look at yourself. - Kurt    Hey.... I love your name! Are you German? I LOVE Germans!  Wait.... if I\'m bigger than my own sense of self-importance, then I have no self-esteem. I actually think my own sense of self-importance is bigger than me. Is that what you meant?  I mean, that whole \"vomit\" thing is nice and original*, but I didn\'t realize someone had told me to think this way. I dunno. I thought I had come up with this on my own, but I guess it was brainwashing by CNN or something.  Also, the \"ay-rabs\" thing was not put in quotation marks to denote quotation, but to denote an \"abnormal\" figure of speech. Something out of place. Something.... facetious, even? FY infizzle, it is an inside joke with someone. Probably an imaginary friend, though, if you buy into what my buddy Nick said earlier.  So did you want me to look more carefully at my hypocrisy, to which I devoted an entire post, or what?  [* - denotes sarcasm]  --------------------------------------------------------------------------------    Wow, you have been bitch slapped, havent you? - Kimble    I would indeed have been bitch-slapped from here to the moon and back again, had I taken myself as seriously as some of these people do.  And did you know that your name rhymes with \"thimble?\"  --------------------------------------------------------------------------------    Cool Dude, I say you hang out at Right-Thinking. You just might learn something. You might be surprised to learn not everyone is a Republican. Its nice to see you lump everyone into one box. How sad. - Kilroy    \"Wait.... what?..... That\'s cool, I don\'t even care.\" I guess that even though I refer to many non-Republicans here, I think everyone\'s Republican. Might as well go \"hang out\" in this \"box\" of yours that I apparently lumped people into at some point. Does it involve free food? Because me and my father Michael Moore need fuel for our hatred.  --------------------------------------------------------------------------------   Man........ I should have seen this shizzit coming, based on the comments pages  urlLink here  and  urlLink here.  But it\'s all good.   I\'ma go dream up new ways to spread propaganda and be mean to conservatives.  (Just be fucking CIVIL to someone for me, okay? Make my \"loser\" life worthwhile.)  '),(26349,2606936,'31,july,2004','2004-07-31 00:00:00','   Hell yes, I do.  My freshman year in high school was also the year my band director came to our school. That year\'s show had a \"night\" theme, and the tunes were Tossin\' and Turnin,\' A Night in Tunisia, Stardust, Saturday Night Fever, and Late in the Evening, among others. It was a wonderful show - it had so many different genres and feelings in it, and it was exciting and it was the only show i had ever marched before, and it was so special.  I had the misfortune that happens to every other rookie band class of having my very first uniformed performance at an away game. I remember little bits of it - how I was the bus captain, which was weird because no one remembered having freshman captains before, but that later became a normality with the director. I remember having my first conversation with B that day, and him telling me some story about a  Jackass  episode where someone hit someone else in the nuts with a hammer. I remember when we came for report time, how all of us freshmen had our hats on with the brims way up high, and all the seniors pulling them down over our eyes. I remember warming up in the parking lot with my section and feeling like I was going to throw up, and seeing that reflected in the eyes of my best friends. I remember how much fun we had in the stands, singing and cheering and chatting with the upperclassmen. I remember how, marching in, after a lecture about not allowing anyone to break the line, some rude woman stepped out into my path and then looked at me and asked if she could get through, and how I didn\'t meet her eyes because I was at attention and just tried not to step on her, and how my revered section leader was in front of me and how the director was walking beside the line at the time, and how no one ever got mad at me for it, and now I step on people who try to break the line. I remember hitting that very first pivot in the line coming in, and how I did it perfectly, and how the director was right there.  I remember when suddenly it was halftime and it was time to put on our hats and gloves and march down the stairs, and I was scared out of my mind, and how the guest band always goes first, and how we stood in front of the stands and the freshman boys were trying to get someof us to move but we didn\'t, somehow, and how we got out onto the field and I was so afraid and how the senior in front of me counted the  duts  and how somehow, during the big swing phase of Night in Tunisia, I realized I was marching it in uniform, and I started to play, and how it felt when we brought the horns down at the end of what we did that night and I was panting and my arms hurt and I realized I had done it.  I remember the competitions when we\'d spend the entire day with band - practicing in the morning, then all the girls would cram into the bathroom to paper-towel our faces and put on eyeliner and lip gloss and french-braid each other\'s hair into pigtails, and then we\'d eat lunch in the bandroom and then we\'d do the hi-dee-ho, which was just a big cheer, and then we\'d get on the buses and when we got to the schools, we\'d watch the other bands compete, and see how some other bands were allowed to march and realize how good ours was, or see what cool things you could do with a marching band.  Then we would go to the buses and get out stuff, and we would change, sometimes in a bathroom or locker room, once in a gym with chaperones covering the windows so the boys couldn\'t see us, and we\'d have to have someone else zip up the jackets and we\'d stuff our hair into the hats and go out and warm up, and then when we marched people watched and listened because they were there for the bands, not the football teams, and they knew what was good and what sucked, and they would clap, and I remember how I felt at the end of the closer, because we ended with a \"Hooah!\" on the horns-down and it was just amazing, the rush you got out of that.  I remember practices in the fall and the winter, how the weather would change, how we\'d be sharing bugspray or huddling together, how B would run around trying to get us to play football when it was freezing because he had just moved here from Philly and thought the weather was nice instead of hot like it had been. I remember how good it felt to go home after practice and take a shower and go to bed feeling like you\'d accomplished something.  I remember Waffle House and Chili\'s after games, and sleepovers after competitions.  I remember eighth grade, feeling awkward and insignificant, and then suddenly I was in high school and I was in the band, I was important, I had people who cared about me and looked after me as I found my way through high school. I was loud and hyper and weird again, instead of introverted like I has become as adolescence set in.  I see now, in the freshmen class, the ones like me - they keep a low profile, but every now and then you get a glimpse of the outgoing, crazy band person hiding inside of them. They\'re learning, and they\'re going to be a strong senior class.  Few of us will go on in music after high school. I love it now, but it\'s something fun, not a career choice.  We stay in it because we love it - this group that simply takes people in and makes them feel welcome, needed, wanted. We grew from scared freshmen, and now my class is the cool, calm, collected senior class, and we\'ve watched two other classes growing as well.  I wouldn\'t be myself if it weren\'t for marching band. I would still be afraid to speak up and out, to say what I actually think to anyone, whether it sounds ridiculously intelligent or ditzy. I wouldn\'t be proud of myself the way I am today.  My last band camp is over, and I\'m about to jump into my last marching season. I am neither the first nor the last person to go through this nostalgia.  I\'ve seen, a few times, exactly what marching band is all about. I\'ve watched the others as we march into competition, and seen all these different people - different ages, races, religions, economic classes, intelligence levels,  athletic abilities, musical abilities - walking the same way for the same purpose. Over one hundred teenagers, at attention, quiet and ready and relaxed and aggressive all at once.  When we march as a band, it isn\'t the loss of individuality people make it out to be. Uniformity isn\'t about conformity - it\'s about purpose and pride.  Somehow, when we get onto the field, we can all work toward the same goal.  There is nothing in the world like marching band. I will never be in another organization like this, and it will hurt, but I\'m lucky to have been a part of it. Yeah-whoa-hey-yeah-whoa.  '),(26350,2606936,'24,july,2004','2004-07-24 00:00:00','   I\'ll be away for the next week because I have band camp.....   Marching, playing, hurting, etc.   If I see any metaphors for human society while I\'m up there, I\'ll let ya know.   However, I\'d just like to ask if I am justified in this.   For&nbsp;the past few years, girls have been permitted to wear spaghetti strap shirts during practice,  under the condition  that they wear at least a sports bra with  full straps , if not a full tank top,  underneath it , and it doesn\'t display too much of the bra [pretty much anything besides the straps]....   and if boys have been allowed to wear this sort of outfit&nbsp;too, when they cross-dress for humor on certain dress-up days and skit nights..... (which is when we have a theme like Tacky Day or soemthing and everyone dresses accordingly for prizes)   Is it fair for a parent to change the rules, in that now no one can wear spaghetti strap tops at all, \"no girls OR boys\".... and  then  say, \"Well, I should take that back... sometimes on dress-up days you\'ll see&nbsp;them [nodding at a boy]wearing those?\"   Keep in mind that this is MARCHING BAND, not cheerleading,&nbsp;and we wear them for comfort, not to be sluts or whatever it is they expect. [Although if they were just banned totally in the first place, we could get over it easily, because there are other comfy tops out there we could live with.]   The problem is that she implied that it was okay if guys wore them to mock girls, but not if girls wore them for comfort. Which is absolutely ridiculous.   Wouldn\'t most anyone who is not an ADD pervert who spends a lot of time around men in tank tops,&nbsp; find a boy in a spaghetti-strap shirt much more distracting than a girl in one, especially&nbsp;during a time when we\'re trying to get work done? It\'s a much less common and more disturbing sight, and it\'s also just gross. Girls in&nbsp;spaghetti-straps are a normal sight in the world of a teenager in summer, and it\'s not like any more skin is being revealed than in a normal tank top. In fact, if you want to measure exactly, MORE is covered than in a regular tank top, if you wear the sports bras.   &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; vs&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp; &nbsp; &nbsp; ?   Is one of them better or worse? I don\'t think so.  Plus,&nbsp;they just decided to tell us 3 days before camp, so now a lot of girls have had to shop for more appropriate clothes that they can march in without discomfort..... so those of us, also, who had planned in advance, are getting screwed over.   I am tempted like a mofo to call their bluff and go ahead and wear the Twin Day outfit that my friend and I have planned for over a month, which involves wearing a spaghetti-strap top over a sports bra.&nbsp; Since it\'s  on a dress-up day, for a costume , it does fall within the guidelines they laid out, and we\'ll be wearing the appropriate bras to cover ourselves up anyway. I kind of feel like we deserve to at least show them what\'s wrong with that rule, and I don\'t know who will have the guts&nbsp;and debating skills, should a problem arise,&nbsp;to do it but me.   If boys can make fools of themselves in comfortable tops, then girls ought to be able to wear them when they\'re being serious. If you ban some of us, you ban all of us. I can almost hear the \"boys will be boys\" coming, and there is no way I will stand for it.   Do I sound more like a pissy teenage girl who wants more than she deserves,&nbsp;than someone who sees a fairly   silly   double standard emerging? I think I\'m a little of both, and I wouldn\'t be angry if we hadn\'t already gotten outfits.   Still, they cannot do that and think it\'s fair. If the boys don\'t indulge in their little privelege, I\'ll accept it and sew the straps of the outfit, because then we\'re all equal.   But if I see any boys wearing them and getting away with it, you can bet I\'ll be doing the same thing,&nbsp;with a whole lot more modesty than they will. I\'ll make SURE I get away with it, too. I don\'t take kindly to being treated like a child by school officials or parents acting as a school representative, and I\'ve usually always won in the past over&nbsp;stupid things like this.&nbsp;   And if I have to do it, I\'ll make sure I look damn good, too.  '),(26351,2606936,'22,july,2004','2004-07-22 00:00:00','   He\'s breathing down my neck now.... I\'ve probably read SOMETHING about him at least five times today.   Can\'t the leader of our country just stay out of my internet for a while?   But yeah. This upcoming election, as anyone with reasonable sentience (and eyes) can see, is not about liberty, or war, or gay rights, or patriotism, or any of those pretty words. This election is all about Bush. Some people just can\'t get enough of Bush. They want twice as much Buch. More Bush, please. Bush, Bush, Bush.   Basically, most of us who are considering The Kerry as a plausible vote-receiver do so simply because of our loss of respect for Bush. (Or else some of them never respected him at all, but it\'s all in the anti-Bush pie.) Bush is wrong. Bush sucketh. Anyone but Bush. Hold the Bush. Bush, Bush, Bush.   Ugh.... I would really, really love to be extra-super rebellious and vote for an independent candidate, but frankly, I don\'t find any of them any more appealing than Kerry, and at least Kerry, being affiliated with the Dems, has some chance of winning in the real world.   But in all honesty, I\'ll probably end up voting for Kerry, assuming I don\'t find anything that disturbs me on his record.   But come on... we\'re sick of all the Kerry-bashing. That\'s old news. We should get back to Bush.   See, I keep clicking random blogs and finding ICH HEART BUSH essentially plastered across the page. It\'s crazy. The thing is, though, so many of them are all like \"Mryah, Kerry sucks\" that I feel bad for them. I mean, why don\'t you rise above the opposition instead of stooping to the level of what one of you so aptly called the \"party of hate?\"   So on one chick\'s \"personal weblog,\" which was more like a liberal-hating database, she had a link with a picture of a beer bottle cap which said Bush-Cheney. Its description read, \"Bush Buds.\"   Oh. Wow. I bet Bush is pleased to see that his supporters photoshopped his name onto a Bud label.   Anyway, the site linked to  urlLink this  place. It\'s called Kerry Core, or Kerry-Core, or something similar, and is a nice idea if you\'re into paying for domain names for the purpose of making fun of a single man. HOWEVER, there are, as far as I can see, three things fundamentally wrong with this.   Number one: People insulting Kerry think they are Bush\'s buds. Unless Bush is as dumb as all those cartoons make him out to be, he probably doesn\'t want to run Kerry\'s name into the ground. What they are doing by insulting one candidate in the name of another is giving a few people a laugh and a few people another reason to dislike Bush. This basically blows as a move to further any sort of cause, unless your cause is to be annoyingly uninteresting.  Not only is that anti-Kerry thing sehr overdone, it is also done here in a rather unoriginal way. This is the second thing wrong with this site - it\'s actually not even witty or funny at all.&nbsp;Their idea of biting criticism is \"\'[Bush] must think we Democrats are the most idiotic buffoons on the planet\', said Kerry.\"   Mmm. Subtle.   I mean, if you\'re going to make a site that satirizes the campaign site of someone else, you\'d better damn well make me either wince or giggle. They describe Democratic supporters as \"angry, dirty foaming hippies,\" and I swear to God, if I&nbsp;see that stupid \"liberal-tree-hugger\" concept trotted out one more time, I am going to take my tree and shove it up someone\'s ass.   But still, it\'s just mundane, regurgitated criticism of Kerry thinly veiled as some sort of clever parody.   This bring me to the third thing wrong here..... When you open the webpage, the description at the top says \"John Kerry 2004 Parody Website.\"   Real clever, guys.   Audrey\'s tip of the day - if you have to explain to your readers that what you are writing is parody, then either your readers are blind, retarded bunnies, or you are not writing very good parody. If anyone actually reads this boring website consistently, I think both of those things would have to be true.   I am abso-fucking-lutely FINE with people voting for Bush, as long as they\'ve put some thought into it. But this site isn\'t a very good way to show your support of any candidate who has minimal standards of self-respect.&nbsp;If you\'re going to spend time and money on dissing ANYONE, you might as well do it in&nbsp;a format that\'s aesthetically pleasing, or at least uses some cool fonts.  '),(26352,2606936,'20,july,2004','2004-07-20 00:00:00','   I discovered  urlLink The Watchman Expositor  when I was doing a TOK project involving Mormonism, and it\'s pretty interesting. It profiles a whole bunch of cults and deviant groups and stuff.  &nbsp;  Now, through  urlLink dooce , I&nbsp;was reading about some ministry, and wnated to know if they were on the cult list. They weren\'t but as I scrolled&nbsp;this definition caught my eye.&nbsp; It makes me giggle.       Wicca:  The practice of witchcraft, nature worship, worship of the Gods and   urlLink  Goddess   ,   urlLink  paganism    and neopaganism, various groups with assorted names. In modern form, it is a revival of pagan religions (  urlLink  neo-paganism   ) and the worship of â€œpre-Christianâ€ Gods and Goddesses. A common creed is, â€œDo what you will, and harm none.â€ Modern Wicca owes much to the influence of Gerald B. Gardner and has been further advanced by the writings of Patricia Monaghan, Silver Ravenwolf, the late Doreen Valiente, and the leading   urlLink  occult    publisher   urlLink  Llewellyn World Wide Ltd   . In popular culture, films and television programs such as  The Craft, Practical Magic, Sabrina the Teenage Witch,  and  Charmed  have helped attract teens and children to this growing religion.    Let\'s ignore \"the Gods and Goddesses.\" They were trying, after all, to educate people about the religion, and they probably were a)confused or b)trying to make it much weirder and nonsensical&nbsp;than it actually is.  &nbsp;  No, the real problem was that whole \"pop culture\" part. Because Sabrina was totally into Wicca and Britney Spears, er,&nbsp;goddess worship..... she was so respectful of the earth...... even with all those Christmas&nbsp;specials and revenge schemes and meddling with nature and hatred of Libby et al&nbsp;and stuff.  &nbsp;  (What can I say, I was a loyal follower up until she got that internship at that stupid music magazine. It was&nbsp;fun to watch her at school, but not when she\'s in the real world. Enh.)   But&nbsp;I digress.... That show was a teen comedy, not a portrayal of pagan spirituality.&nbsp; Sabrina  was magic-without-a-k. Saying that someone converted to Wicca because of watching&nbsp; Sabrina, the Teenage Witch  is like saying that someone converted to Christianity because of watching  Frosty the Snowman .   I mean, all these people who have never watched shows like this,&nbsp;assume that they\'re all about&nbsp;black spells and dark worship, when they have nothing to do with the religion at all.   Like those people who believe Harry Potter is evil? Many admit to having never read the books, and then bash them for being occult and crap. Depending on how you define occult, maybe playful children\'s-storybook \"magic\" is, but hello? Do you allow your children to watch movies like Sleeping Beauty or Peter Pan or Cinderella?   Does anyone else see the humor in this? It\'s like they were BSing a paper for school and had to come up with some random sources, so they just pulled books off the shelf and carded them.  &nbsp;  Is it the term \"witch\"&nbsp; that bugs people so much? I mean, it isn\'t the religious followers\' fault that mainstream culture turned it into something so different. It doesn\'t mean \"a Wiccan\" they way it should anymore. Now it means someone who does magic, meaning turns people into toads and shit, or someone evil who eats children or whatever.  &nbsp;  And \"magic,\" too, now means pretty much something defying the laws of nature. \"Magick\" means energy that comes from religious observances, not something that carpet cleaner can perform. Do you think these anti-Harry-Potter-and-Sabrina people boycott products like  urlLink this?  Do they try to have their childrens\' teachers fired for reading  urlLink this  in class?  &nbsp;  Oh, man. There are times when I wonder why I don\'t have a religion, and then there are times like now, when I\'m glad of it.  '),(26353,2606936,'19,july,2004','2004-07-19 00:00:00','   Uh, never mind that whole intervention thing.  &nbsp; It doesn\'t quite work that way. &nbsp; However, things will still be helped.  '),(26354,2606936,'16,july,2004','2004-07-16 00:00:00','   ... doing all this research about cutting. Self-abuse, self-harm, self-mutilation, and those all come un-hyphenated as well. &nbsp; Only it\'s not for a grade or something like that. It\'s like, IMPORTANT. It\'s stressing me out more than a Lit essay. If I fail this one, I will be screwed. And so will a lot of other people..... &nbsp; Right now, I am trying to make sense of all the information i\'ve deemed helpful. (ABout 11.25 pages\' worth on Word, buddy. Many spaces and varied fonts, but still a LOT....)&nbsp;  &nbsp; Big T, as he shall henceforth be known, has known about Jules\' \"problem\"&nbsp;for a lot longer. I don\'t know how he found out, but I know it wasn\'t be seeing them like Angela and I did. &nbsp; (God, these fake names are killing me.) &nbsp; So after much discussion, hair-pulling-outing, whining, crying, and thinking, we\'re going to have an intervention. It\'s been going on longer than we realized and none of us can work alone.&nbsp;  &nbsp; I know there will be at least ten friends who will come if we ask, maybe more. We may be able to get the band directors, assuming they aren\'t required to tell the school. Unfortunately, Jules\' parents are somewhat wacko. They\'re very controlling and sometimes actually just a teeny bit verbally abusive - it\'s always, \"Get your butt over here and answer the phone!\" or something. &nbsp; Angela also noticed that the way Jules talks to her younger brothers, who are in elemntary school, is probably influenced by her parents. I wondered why she was mean about stuff, but that\'s just how you learn to deal, I guess. I mean, I use humor, which came from my Elterns, so take from it what you will. &nbsp; Anyway, her parents are of the keeping-children-inside-the-home-will-make-them-good school of thought, and were we to tell them, we all fear that they would do something stupid like ground her and make her promise to stop.  &nbsp; Even my MOTHER thought of this. &nbsp; So.... I\'ve been to all these wizeb sizites, which is exhausting. I\'m hoping against hope that cutting is her only problem, so that her parents don\'t have a chance to muddle things up and alienate themselves from her even more than they already have.  &nbsp; However, she has post-traumatic stress disorder which gives her migraines, diabetes, and an ebbing relationship with a rather stupid boy that I am not a big fan of right now. &nbsp; I dunno... Don\'t worry about me, no need to pity-fy, but maybe if you\'re one of those \"God\" types, you could put in a good word.... I think I just need to put my thoughts out there, and I didn\'t want anyone to think I was just going to \"forget\" about it. We\'ve got two weeks full of schedule-type activities, but soon, my dears, soon we will confront her, and I have a feeling that this will be the hardest thing I\'ve ever done. Harder than AP Language, harder than weight training, harder than learning music. &nbsp; Damn, this will be hard.... I\'m very glad I\'ve got two of my best friends to help me through the planning. And Jules is very much worth it. &nbsp; I don\'t need anything&nbsp;surmised about an intervention right now, please. Contrary to popular belief, I do have common sense... but if anyone out there has ever staged one, maybe leave me some tips that you can\'t find on \"support&nbsp; dot org\" type places?  '),(26355,2606936,'14,july,2004','2004-07-14 00:00:00','   To begin with, thanks to  urlLink goliard  and  urlLink David  for the responses. I may update sometime, but for now we\'ll just say progress is being made, and I think it\'s going to end up okay with Jules.    Jules just spent five days at the beach with my family and I, and I am tan as two short tan things. We spoke vaguely, without me having to directly confront her, but I think for now all I can do is listen, because she seems to know that it needs to stop, and I\'m probably the only friend who wasn\'t thrown into a complete shaking panic by it. This may be because I know her well, or because I\'m in denial, but I still refuse to shove \"help\" down her throat, for now. As long as she keeps talking to me instead of cutting, we can both pretend I don\'t know what\'s happened.   So, despite something that most definitely could have screwed up a vacation, the beach was fantabulous. We had some interesting encounters there, beginning on the day we arrived, Saturday.   The place we stay is a condominium thingy, which is basically shaped like an L with a pool in the crook, facing the ocean. So we\'d be swimming and see these guy watching us from the balcony. If they weren\'t old, it was nothing to be alarmed about, and if they were cute, it was quite the ego boost.   There were three in particular that were together that kept catching our eyes and waving to us. Two of them appeared to be identical twins, somewhat fair, and then their friend was really really tan. We called him the Dark One. Doesn\'t it sound nerdy?   Anyway, after two days of waving and smiling and saying \"hi\" to these people when we passed them, Jules was ready to \"take the relationship to the next level,\" as they say, and actually met them. She has a boyfriend, but harmless flirting means nothing to them. However, she doesn\'t exactly know that B and i have something of an understanding. [This one is of the \"I love you, but my parents will castrate you if they find out we\'re dating again\" variety, so essentially everyone is wondering why I don\'t hit on people any more.] So rather than reveal to her information which could cause my dear B bodily harm, I sucked it up and initiated a conversation.   I think it was something like, \"Uh, we can\'t decide if you\'re identical or fraternal, uh, hi?\" and then we spoke for a while. It was pretty nice, I guess, even though their \"I can\'t run good\"\'s made me clench my teeth.   So these three guys (aged sixteen.... Jules and I are cradle-robbers, we are) had this other little kid running around witht hem that they had just met on vacation. He was short and looked like he was, oh, thirteen, but he claimed he was sixteen. (Hint - he\'s from Georgia and claims he failed his driver\'s test... of course, he doesn\'t seem to remember that whole learner\'s permit law we have. But yeah, he was sixteen like i was looking to date those guys.)   The kid was odd. He was always hitting on people, including the looks-to-be-twelve-but-says-she\'s-fourteen-and-wears-the-same-bikini-every-day-because-it\'s-the-only-skimpy-one-she-has chick, and was generally just kind of annoying. But then last night the twins both went down to walk on the beach with this blond girl and her little siblings. (Yes, they thought they were Big Pimpin.\')   The Dark One was telling us about how Twin A is about to hook up with Blondie, but Twin B won\'t take the hint and leave. \"Whenever someone\'s about to have a good time, [Twin B] is just taggin\' along,\" he stated. He actually said \"have a good time,\" but he had great abs, so we\'ll ignore that for just a little longer.   So then the twins and Blondie and the two little kids came walking up the boardwalk just as the Dark One is proclaming Twin B to be an annoying \"bitch.\" Jules and I hushed them, because generally speaking, people don\'t usually like their friends to overhear them when they\'re talking about them. That, however, seems to be a Georgian trait, because they were all macho and \"I don\'t care!\" and crap.   So the Dark One, who we know is his best friend, jokes loudly that twin B is a bitch or something. Then the little dude started in, following along, and he called out \"Tagalong!\"   I don\'t know. This just pissed me off. This kid lies about his age and acts all \"suave,\" trying to get in with the older guys, but he\'s basically just been a pig. Every time I was there, he\'d be talking about some girl\'s face or hair or body. And now, he\'s pulling the classic dis-people-to-look-cool trick? That old trick of climbing the social ladder, right? You have to kick others out of the way.   Dear God. I know it was just some stupid pubescent male, and it was people who will never care again. But the principle of it! Where do you get off?   Jules looked at me, and I paused and said to him, \"Wow. That was.... just... biting.\" I was being so obviously facetious that sarcasm was dripping from my fangs. Jules giggled.   The little guy: \"Well, I don\'t care. I\'m never gonna see him again after this week. You think I\'m ever gonna see him again after this week?\"   I looked at him. \"I mean... well.... that was just so witty. And so cruel.\"   Little Dude: \"I don\'t care. He\'s a tagalong bitch.\" Or some shit like that.   So I kind of looked at him for a minute, trying to see if he was serious. Like, he honestly thought that I found his attempted insult to be funny, clever and oh-so-not-repetitive.   He made a big show then, getting Twin B away from Twin A and the girl, and, like, dragging him off, and so then it was just me and Jules and Dark One in the pool. I think Dark One had some inkling of suspicion that there was something mildly ironic occuring, so he started doing cannonballs into the pool.   Jules and I just treaded water for a minute, looking back at where the little asshole had disappeared.   \"He\'s the bitch,\" she told me.  '),(26356,2606936,'06,july,2004','2004-07-06 00:00:00','   I am completely obsessed with  Degrassi: The Next Generation,  as viewed on Noggin. For those of you who a) like me, are slaves to a communications giant because of a family trade and therefore do not recieve this channel, or b) have better things to do with your time, Degrassi is a Canadian-made pre-teen and young teen soap opera.   It is fucking genius. These characters go through every possible problem teens could face, and they do it with perfect hair, dull wit, extra melodrama, and Canadian accents. (How long have we \"bean\" friends?.... I\'m \"soary!\"..... Ask him \"ote!\" etc etc)  Anyway, so my favorite character on the show is Ellie. Redheaded, somewhat pretty, punky-gothy-kinda chick. However, she\'s awesome because she was always one of those mature ones who kinda saw through all the shit going on, without being annoyingly intelligent or rebellious. She\'s just THERE, but in a good, strong, important way.  But then, in one very special episode, Ellie\'s dad went off to Iraq and her mother parked her Canadian ass on the couch with a bottle of vodka. So Ellie\'s all stressed between trying to keep her mom, like, alive, and her dad from worrying, and her grades up, and also to hold on to this film internship deal she landed. So finally, when she sees a compass fall out of her backpack, she starts to cut her wrists.  Anyway, obviously someone notices by accident, and of course it\'s Paige, the resident bitch queen of this Canadian suburb. So it\'s like, Ellie\'s only \"hope\" is this chick who hated her, and competed with her for that internship, and Ellie has a million reasons not to trust her, but in the end, she ends up talking to the school counselor a lot and getting on her way to being okay.  So in a later episode, a miniature  Breakfast Club  in which some vair vair different people are thrown into detention together and forced to rethink themselves, she ends up talking to and \"sparking\" this guy she\'d never really known before. This kid was Sean.  Both these kids had rumors circulating - while hers was about the cutting, his came from nasty allegations about his stealing school laptops and such. So they\'re lying on the school roof, talking, and one of them grabbed the other\'s hand. They kinda looked at each other, and Ellie said, \"You don\'t seem freaked out by me.\"  Sean: \"Nope.\"  Ellie then rolls down the wrist cuffs she wears, and shows him the scars of when she had cut. \"Are you freaked out now?\"  Sean ran his fingers across her scars, stroked her hand, and told her he wasn\'t.  Lately one of my best friends, Jules, has been writing depressed things on her xanga. Things about comparing her old boyfriend to her new one, and how difficult her family life is, and just how things are hard. Anyway, a few days ago she quoted Ellie from Degrassi: \"Are you freaked out now?\"  So today, I was planning to work out with another good friend, Angela, to complete some P.E. credit for me. So we\'re kinda lazing around my room, and Jules calls. She said she was on an hour break at work, and was wondering if she could come over.  It was weird, since I don\'t recall her ever asking herself over before, but of course we wanted to see her, so she came.   Jules, Angela, my mother, and I sat in my kitchen chatting about a million things, and at one point my mom went to tlak on the phone to someone. Angela was talking animatedly when she looked down at Jules\' arm. \"Oh, what happened?\" she asked, pointing at a long red scab across the back-hand-side of her arm.  \"Probably a cat,\" I quipped, in my usual anti-feline fashion, but Jules kind of ummed. I touched the scabbed-over wound and her wrist turned over, showing more.  Six or seven red lines, across the underside of her wrist, all the same length, all evenly spaced. As I ran my fingers over it, she finally said, \"I don\'t know.\" Our eyes met and hers were frantic, as I tried to push back what came into my mind.  I fussed over it and dismissed it, and the conversation moved away. A while later Jules left. Angela and I were going to go and do our run, but my mother, who was making country-fried steak, needed us to go buy her some eggs.  So we\'re halfway to the grocery store, yammering about something random, when Angela said - \"Oh God, Audrey! Do you think Jules cut herself?\"  She saw everything, too - I wasn\'t the only one to suspect this. Angela cried in the car and we decided to tell my mother. However, her rooting around for her copy of \"Reviving Ophelia\" didn\'t quite give us what we\'d wanted.  For God\'s sake, we know what cutting means. What we need to know is what the hell we\'re supposed to do now.  We\'re trying to plan a \"girls\' day/night\" for the day after tomorrow. Maybe we can get her to tell us without letting her know we know. Maybe we\'ll have to tell her, and maybe bad stuff will happen.  I don\'t know what we\'ll do, but it will be something. It will be everything we can muster.  She wanted us to see it - she wore short sleeves, she asked to come over, she posted that goddamn Degrassi quote. She needs something, and she knows it, somehow.  I haven\'t cried yet, though Angela did as we drove home from the grocery. My Jules, my best friend... she\'s one of those who held me afloat during that dark time known as Getting Over B. She\'s done so fucking much for me, and she can\'t do this on her own.   Jules, we are NOT freaked out. We will never be freaked out. I will never be afraid of you, Jules, and as soon as we can come clean, i will never, ever let you be afraid of yourself again. You will never be alone.  '),(26357,2606936,'04,july,2004','2004-07-04 00:00:00','   I discovered a very cool Blogroll activity... however, since I am a Bloglinker kinda gal, we\'ll just change it to fit me...   There are TWO rules when answering these questions: 1) Only ONE answer to each. Of course its tough - thats the point! 2) Each blog/blogger may only be used ONCE  What blogger inspired you to FINALLY start a blog?    urlLink Indiscretions    What blog do you visit the most often everday?   urlLink And another thing...    What blogger do you think you have the most in common with?    urlLink Leticia McKenzie , given age and self-professed dorkishness. In no way do I claim to be either as sexy or as brilliant as Leticia!   Which blog can you be sure will make you pee a little you laugh so hard?    urlLink The Sneeze    Which blogger leaves you the best/funniest comments?    urlLink Machine609    What blogger do you wish would update more often?    urlLink bellow    Which blog do you wish more people would read?   urlLink one child left behind     Which blog do you learn the most from?   urlLink The Real Planet Earth Site    What blog is your newest addition?   urlLink Whaz Up    Who has been on your blogroll the longest?   urlLink Cheekysquirrel    Whose blogroll would you LOVE find yourself on?   urlLink Que Sera Sera    Whose bloglinker list were you the happiest to find yourself on?   urlLink Bakerina  ok, maybe just a list, but whatever   If you could write like any blogger, who would it be?   urlLink tequila mockingbird    What blogger are you the happiest you\'ve \"met?\"   urlLink popcultureatemybrain    Which blog do you recommend the most?  oh, man. there\'s a reason I had to break it into 3 lists.   Who is the next person you\'ll add to your bloglinker list?  Uhh. Either the next person who links me, or the next person who makes me stay up late or almost be late for work because I\'m reading their archives.   Who is the blogger you hope to meet in \"real\" life?   urlLink David  - I think I\'m still engaged to him, and I\'m not into that whole \"Married by America\" blind wedding stuff. No, sir.   Which blogger do you admire?   urlLink aethele.   the way she writes about her life, I sometimes don\'t realize how different it is from my own, and how strong this woman must be!   Who would you trust with your blog while you were away?  Psh... honestly, I haven\'t actually met any of these people, so I don\'t think I\'ll ever really know if they\'re actually scary old men.   Well, some of you are scary old men, but you know what I mean.   Which blog has your favorite design?   urlLink ich zwoje  has a neat one, but I don\'t recommend clicking it unless you have a pretty fast connection.  How many blogs are on your blogroll(s) 35  '),(26358,2606936,'01,july,2004','2004-07-01 00:00:00','   ... is what goes on in my head.  Today, I realized something cool about language. You know how \"dead\" and \"bread\" rhyme?   Well, the German word for dead is \"tot.\" The German word for bread is \"brot.\" They rhyme too.  Whoa.  Also, I can\'t decide which is better: \"lizegalization\" or \"bahguggle.\"   The first came out of me reading back over the gay-marriage debate at  urlLink 1-800-FUK-U.  The second is how my best friend\'s two-year-old niece pronounces her birthday gift.   Thoughts?  '),(26359,2606936,'01,july,2004','2004-07-01 00:00:00','   I just did another one-hour THRZ walk. I expected more hell, but about five minutes into it, it started raining, and rained for about forty-five more.  It was absolutely brilliant. It\'s so much easier to work hard when you\'re not fucking HOT. ALso, there were exactly NO other people outside when it rained. You get to jump in puddles, too.   And when given a choice between sweat and water, I would most definitely prefer water.  '),(26360,2606936,'08,August,2004','2004-08-08 00:00:00','   Also. Jetzt fÃ¼hle ich am schlimmsten.  Scheisse, scheisse.  Ich weiÃŸ noch, dass nicht so viele Leuten kÃ¶nnen deutsch, und dass keiner meinen Leser kÃ¶nnen deutsch. Ich fÃ¼hle nur, dass ich zu viele Male dass gesagt habe. Wenn ich die selbe WÃ¶rter benuzte, dann werde ich mich krank machen. Heute war ein schlimmer Tag, und ich habe kein Energie mehr.  Ich muss dass aber hier sagen. Mein Journal braucht dass. Ich kann nicht ignorieren, ich kann nicht entgehen, ICH MUSS ES NICHT VERWEIGERN. SO habe ich bevor gemacht, und sofort habe ich mehr Probleme gemacht. Ich muss etwas hier sagen. Ich muss mich helfen.  Wenn Sie kein deutsch kÃ¶nnen - und jetzt bin ich sehr dumm - dann kannst du  urlLink hier  gehen.  B und ich haben schlussgemacht, und dass war nicht meine Idee.  Mist. Jetzt habe ich kein Lust, um das zu schreiben. Ich bin sehr mÃ¼de.  Lieb mich, bitte. Ich erzÃ¤hle euch alles, wenn ich besser fÃ¼hle.  '),(26361,2606936,'06,August,2004','2004-08-06 00:00:00','   I think I once absentmindedly told my mother that I\'d love to live in the 1940s. \"Everything just seemed so..... nice,\" I said.  She stared and replied, \"Yeah, except for that whole war thing.\"  Touche. And yet, I still hold this dream of living in the forties. I would have those gorgeous bouncing waves in my hair, and wear red lipstick and adorable dresses and swimsuits, and I\'d feel the solidarity of hating Hitler and loving FDR along with the rest of America. I could gossip with my girlfriends and crochet socks for the boys overseas, and i would probably have a dog like Skip.  I know, though, that I would also be unhappy, because I would suck at being a forties girl. I look terrible in red lipstick, and I couldn\'t possibly sleep every night with bobby pins sticking my scalp. They wouldn\'t even be the cool kind with the ball tips - no, they would probably cut my head. I bet that old lipstick was the kind that stains your lips and gets all over your teeth, too.  The other girls wouldn\'t like me, either, because I\'d be too weird. I\'d be loud and I would speak my opinions and talk about \"boy\" things and they\'d get bored. And teachers would be like, \"Audrey, shut up about Nietzsche, it\'s time to bake.\" And boys would think I was pretty with my great hair and stuff, but I talked too much, and I wouldn\'t sleep with them (because it wasn\'t societally okay back then, for the most part) so there would be no fun in it. So I\'d, like, sit around crocheting and talking politics to my dog, but the only interesting thing to talk about would be how Hitler sucked, and even that would get old.  I\'d probably dream of being a movie star or something, but the directors would get annoyed and fire me for being a diva or something. Or else I\'d have to wear fur and shit all the time, and plug Ovaltine and war bonds and whatnot, and that would get boring.  I guess I don\'t really want to transplant myself into the time period, so much as I want to be one of those girls, but like, I guess the kind who goes to college and stuff.  It just seems nice, like everything back then was laid out for teenagers. Don\'t have sex, go to school, eat your dinner, and everything went the way it should.  I would be resentful, but I wouldn\'t be so confused.  '),(26362,2606936,'03,August,2004','2004-08-03 00:00:00','   i must change this template. it is killing me....... soooooo much orange..... and white.......  eeep. stay tuned.  '),(26363,3972921,'24,July,2004','2004-07-24 00:00:00','   Now I don\'t think I have ever called him Chewy to his face. In fact I am not sure I remember how he got that name....something about back hair?  My friend Brian Jeanseau is the monthly Blog Star. I will be highlighting a new person in my life each month that inspires me, encourages me, or makes me laugh. This month\'s BS does all three.      Let\'s start with why Brian inspires me...  Brian is one of the most real people I have ever met. He is honest, forthright, and nice when doing it. Brian cares for his family. I think that is fantastic. He provides for his family, financially and I know he takes his wife on dates. I have observed him being an excellent husband and father.  Brian is encouraging. He recently took over as worship leader for our church and is doing a phenomenal job. I felt SO cared about the first week after worship he sent all of us an email informing us of how proud he was....so neat. He is always pushing me a little further than I want to go, and I truly feel he believes in me...I cannot say that about very many people.  Brian makes me laugh...mallard duck. Enough said.  I think Brian is one of the best people that I know and all of you should be lucky enough to have a friend like Brian.  Kat  '),(26364,3972921,'18,July,2004','2004-07-18 00:00:00','   I have become a member of the \"Check Fraudulence Club.\" No I have not started conterfeiting checks in the McVicker\'s basement(next to the Meth Lab...AKA \"Brew Kitchen\"). Yours truly had some checks that got into the hands of the devil\'s handy men(or women for that matter). Okay so I am kidding about the DH men...just need a little humor. I came home on Wednesday from a relaxing evening in the suburbs to find a letter from Winco Foods notifying me of a dishonored check. I thought that was weird since I have not shopped there in forever and I certainly did not spend $129. I mean come on...I don\'t think I have EVER spent that much on groceries..I am single...well maybe when I was a youth leader and all of those teenagers eating me out of house and home. Anyways I called the store and found out there was another check for $98...called the bank the next day to find out it is 6 checks thus far. Did I mention that this is on my closed account from WA? Yeah so I had to file a police report, they are sending me some stuff to sign, then as far as I know now I will have to take a police report to each place that this person wrote a check..needless to say I was hacked. I was more stressed than anything.  Tommorow I am teaching on Phil. 4:6-7  Do not be anxious about anything, but in everything by prayer and petition, with thanksgiving, present your requests to God. And the peace of God, which transcends all understanding, will guard your hearts and your minds in Christ Jesus.  Kinda funny the scripture that I have been studying for the last week and a half is perfectly applicable to my present life circumstances.  I have spent the last couple of days realizing how truly blessed I am, which leads me to being thankful. I have had a really shitty last year, but despite the ebb and flow of life I realize I always come out of it fine. I survive.   The God that is my very breathe, my very reason for living, is my strength. Father I thank thee.   I am thankful for relationships, love, truth. Father I thank thee.   Take time this week to be thankful for your existence, purpose, and His divine beauty shining on you.  Kat  '),(26365,3972921,'18,July,2004','2004-07-18 00:00:00','   I watched the movie Grand Canyon last Sunday. I had not seen that movie in several years and was surprise by the poignant theme. I came away from that movie thinking/rethinking a lot of stuff in my own life as to how I am living and how that may/may not affect my world, my community.   Fast forward to Friday I am driving to work and I take a different route one that included me stopping by the yuppie Safeway(have you seen the ones with track lighting in the cheese/wine section?) for an energy bar before work. I decided to go up Cornell to 158th and drove by the large building I drive by on the 26 occasionally. I always looked closely at this building wondering what it was. They were obviously doing some sort of additions and thought it was perhaps a new college building or perhaps a high school, or a home for corporate offices of some sort. To my amazement it was a church. I saw the sign for a presbyterian church and was speechless. Now what I am about to write may seem judgemental...but I need to pose the question...why? I don\'t get it. I don\'t. Multi-million dollar(guessing...not really sure of the exact numbers) building projects when there are people with no food or clothes for people in the community. I just don\'t get the world today. I don\'t understand how we have a society in which the rich become richer and the poor become poorer. I was angry when I saw that it was a church. I am not trying to be a bleeding heart liberal but I feel something is severely wrong with this picture.   I was reading Blue like Jazz by Donald Miller again this week and he writes about problems. He states that the problem does not lie in the politics of the world, the church, but essentially the problem lies within us. He states that the only true change, true life-giving, God-honoring change would have to start with the individual. I whole-heartedly agree. It does begin with me.  When I have these moments of enlightenment I am jacked up and ready to make a change. I harbor these utopian ideas of evolving into a better person. I need to give more, consume less, live a life of simplicity. Give to the poor, I don\'t really need to buy make-up or shoes, I can give it away. I don\'t need to drive or buy products that are not good for the environment. I make list after list of all of the things I want to change about myself, my habits, only to be disappointed at my lack of discipline and follow-through. So I ask myself, how do I make changes in my life? Miller says that the path to joy winds through a dark valley of our own depravity.   I know that are times in my life when I have looked at my own crap and been so completely discouraged at my progress. I see who I am and all that I aspire to be and it makes me sad that I fall SO short of my expectations. The only thing redeemable thing about my life is the fact that as I may be overly critical of myself,God is not that way. In fact he is just proud of me that I get up and make the effort to live my life and want him to be apart of that. I think that things will really start to change not only in my own heart, but in the hearts of men and women in this world is when we see that we are a really screwed up people in need of someone to save us. Yes the problem is in me, but there is someone who wants to hold me, love me, and show me who I was intended to be without all of my self-absorption. In the light of that knowledge we CAN as Eugene Peterson says be made into our true selves, our child-of-God selves.  Peace   Kat  '),(26366,3972921,'15,August,2004','2004-08-15 00:00:00','   Each year as my birthday comes and goes, I take a significant amount of time reflecting on the past, present, and what is to come. Periodically a sense of nostalgia comes over me as I smell a familiar scent or hear a song that takes me to another place. These moments lead me to recall old memories, some joyous, others despairing. Recently I has one of these occasions as I began to read past journal entries.   Journaling for me has become one of my most treasured activities. It is a place that I bear my soul, process through emotions, and pray. Most of it is prayer. I write these letters to God of my pain, fraility, and my lack of understanding the mystery. I pour out my love, gratitude, and desire...passionate letter after passionate letter.  In these letters I record not only my own voice but the voice of the One who weeps with me, laughs with me, and renews hope within my feeble heart.   When I was in middle school I had a diary. I was good about writing in it. It always felt funny to write my thoughts, activities, and emotions to \"Diary\". It was pointless to me. I remember at slumber parties we used to read each others diaries if we were \"best friends\"....only your BF could know the intimate details about you. I remember some of the things I used to read, \" We went to the mall today,\" or \"I really want a new phone,\"...mine were always something about a boy. \"I really wish Josh McDonald would be my boyfriend.\" Josh McDonald was my sixth grade crush who did not have a clue as to who I was until 7th grade when I became popular. Look back at those times, now at 26, I wish I would have kept those old pages. To be able to see the reality of middle school dating rearing it\'s ugly head again almost 15 years later would be humbling. Girls, I don\'t think we ever grow out of our girly middle school phase. We still ask each other info about what the boy said, we still get all stupid and googly, and we still think that the world would be complete if we just had \"the boy.\" I\'ll try not to continue with this tangent and save it for the book. There is still something very sacred about the contents of a diary/journal that even only your closest friends are able to know your thoughts. Even with that said I only read exerpts from my journals to my most intimate friends...which consists of 2-3 people. And with that there is only one person who I would let hold the treasured contents and read it for themselves(you know who you are). It is a very very intimate thing.  Over almost 9 years I have written in everything from notebooks, cutesy journals from Old Navy, to napkins from coffee shops. I am somewhat of a minimalist so everything that is not being used is in the trash or on the way to the Goodwill...but never my old papers. There are moments, emotions, and journeys recorded in these books that would make an excellent made for T.V. movie. I read my earliest prayers and note that everything is flowery and pretty using big words like majesty, king of kings, and holy of holies. Over the years this was toned down as I grew in my relationship with God to words like my friend, my father, with the occasional king thrown in. Then there were the crying out for God years...that is where my pain began to seep into my view of God. My self-inflicted wounds causing me to to question whether or not he was there. My faith diminishing under the weight of disbelief, so I yelled, screamed, and cursed at God. He became at that point a comforter, lover, and a savior in a way I had not understood and still am trying to comprehend today.  Now I look back at my life over the last 10 years and I see that it is good. I struggle with my desire to be all that I aspire to be , yet all that He has created me to be...eventually I will grasp that these coincide. I see the person that I am becoming and it excites me. Not because of anything that I have achieved, but because I can see the miracle within me. I see the person I was, I am, and who I am becoming and I am awed by the work of his hand. Who ever thought he would take a snobby materialistic girl and change me into who I am being molded into?  I look over my relationships and am awed. I think back to years past and all of my friendships were somewhat superficial. This year brought some relationships to a close, but by the closing it brought a depth of love and understanding to some that lacked. It also brought the emotional availability to invest in new friendships. My friend Matt, who I would say is my best friend, has taught me new things that I never thought I could learn. We have fought over everything there is to fight over, disagreed on everything, and for a period of time could not stand each other. I have a greater understanding of fighting for people in relationships because he fought for me. He did not give up and now he is reaping the fruit of his hard labor...I am no longer a bitch(most of the time). I guess that is a miracle...he would say so. My friends Trissa, Julie, and Curtis have all had a hand in shaping me into the person that I am becoming. Entry after entry usually includes something that one of them has said to encourage, stimulate, or piss me off. It is so good.         These are pictures that were taken from my 26th birthday party. The picture with Trissa, Alicia, Matt and I, is my favorite. Allisha and Matt both helped out tons to get the party ready. They both know how stressed out I get at events that I am planning and so they were right there keeping me together. About an hour before the party Matt took over and finished getting things ready while I got dressed. Trissa and Jon arrived about a 1/2 hour later and Jon(who is not familiar with my pre-party stress) said, \"Someone get this girl a drink.\" Trissa replied, \"This is just how she is before the party. As soon as people begin to arrive she will be a completely different person...laughing, enjoying herself.\" She was right. It is so nice to have people in your life that know you so well. These relationships continue to go through trials and tribulations and each time we all learn a new meaning for words like love, forgiveness, understanding, and compassion. I am the person I am because of the people that play the part they are supposed to play in the symphony of life. Each instrument, each note being played, sometimes not harmonically, but after much practice and devotion...the music is played and flows into the beautiful melodies that are intimate relationships.   Peace  Kat  '),(26367,4255254,'23,August,2004','2004-08-23 00:00:00','   I\'m so bloody hell sick of goin to sch..every morning i gt to force myself to wake up..den ish back to my farking routine life..draggin my feets to school and attending lessons..how i wish i had done well fer my PSLE..n i wun be struggling like fark rite now..i ve no one to turn to and juz can vent my frustrations on bloggie..there ish no sense of urgency in me at all..i\'m juz farking damn lazy to touch my books..can any1 help..? haixx rotting everyday after sch..since i m back frm sch i ve been online..well i wonder how can i cope if my com breaks down..everything juz sux rite now..dnt deadline ish nearing and i m no where near completion..n next, ish prelim..4get abt tht..i m bound to flunk..n perhaps even the O levels too..i gt to admit is difficult to cope..sec5 werk ish much tougher than N levels werk..ish nearing september..i dun wish to giv up at tis period of da tym..but i m on da verge of breaking down..i seems fine in reality..but da actual fact ish i noe i m a potential O levels failer !! ARghhh!%^!%@ -.-\'\'  '),(26368,4255254,'21,August,2004','2004-08-21 00:00:00','   I m such a disappointment..i gt back my geography test and i flunked it..it may b juz a common test but i did nt xpect to fail..perhaps i was over confident ?? tis daes i dunno wad\'s exactly rong wif me..sumtimes things may nt be as complicated as i may tink n yet i kip pondering over it..realli TTM !! ish onli abt 70 daes to O levels..and i haf nt started revising..n worse stil, i dun feel any pressure at all..n abt bgr, i haf no comments at all..blame me fer falling fer her but i ve no control over my feelings fer her at all..we may haf doubts fer each other..but plx dun question my feelings fer you..blame it tht i cant made it clear to u..blame it tht i sound so unconvincing coz i ve proven nth at all..u ve ur own dreams and plans and i respect ur decesion..if u feel pressurised, i will leave u alone coz if it ish really meant to be, we will definitely be tgt one day..i m trying to be optimistic though i dun haf high hopes it will turn out da way i long fer..as fer now, i wish u can study hard and nt slack anymore..da same goes fer me..i juz want u to stay happy..i wanna do my bez fer O levels so tht i wun break another promise agn..  '),(26369,4255254,'21,August,2004','2004-08-21 00:00:00','   Why?? she was leading 3-1 and da worst possible scenario juz happen..she lost 4-3..i cant bring myself to believe it..i was pissed off wid how da opponent behave..it sux she shouted n juz stare at jiawei throughout da game..kinda using her lame antics to sux her opposition\' morale haixx it was nv meant to be..jiawei had no luck at all..congrats to da north korean and to be fair to her, she showed true grit by playing to her strength to win da last 3 sets..wish her all da bez and whu noes? she may even clinch da gold medal wif da help of her lame antics..as fer jiawei i wish she can win da bronze medal fer singapore..n most importantly, fer herself and it was a pleasant sight to realize there were greeks supporting jiawei during the match..lastly, pray tht she can regain her self confidence before da last match where da bronze medal ish at stake..da bez player wins and todae da north korean proved juz tht..  '),(26370,4255254,'20,August,2004','2004-08-20 00:00:00','   I m so glad tht she beat da china rank no.2 table tennis player..erm hope tht she could be da 2nd person in histroy to win an olympic medal fer singapore..ish true tht she may nt be a singaporean but she deserves da regonition and support fer donning da colours of singapore..i wish she will not disappoint me..=Xx praying hard..~~~  '),(26371,4255254,'19,August,2004','2004-08-19 00:00:00','   Time: 5:30 pm ..o.O Mood: Boring rainy day..guess wht\'s on my mind..? yesh i feel lyk sleeping..exactly..ish so damn fuc*ing bored..but i guess ish a nice day to rot awy..Today, i shall tell u why my sch ish full of lamey stuff~~tht is, tht fatty black DM!!! dun get me rong..nt a racist remark or whtever !!..ish juz tht i cant find any wordy to describe him..coz i juz cant think of wad\'s exactly rong wif his blackie brain..n ish hellya funny when he actually thot tht da pear cover on my hand ish da latest trend in sch..wad CRaPps !! FLAshback..........................................................................................o.O tym: 10:oo am..i was approached by him while walkin back to klass..n he started asking qus abt my bandaged finger n of course da fruit cover..at da beginning i was so naive to tink tht it was juz out of concern tht he qus me tht..da few seconds of impression on him suddenly changed frm sum1 whu likes to send students to Dc(s) or shout \"COME HERE !!\" in tht strong indian accent whenever he like..to sum1 whu actually cares alot..hahax guess wht\'s next..? it turns out tht he was suspecting tht i was fakin an injury..n perhaps to him..to start a new trend ??..0mg!!!#!#^%!!..ish hellya funny to tink tht he haf tht idea in mind but on da other hand, i was so pissed off tokin to a person wif P** Brain !!~ ish hellya lame n ish pure shyts..coz here i m kinda frustrated tht i cant do much writing..n everydae i nid to get used to styling my hair wif 1 hand..n here ish tis fatty having tht idea i was faking an injury..n even claimed tht he was goin to call da doctor to make sure it ish no new trends or watever..i cant do much as a student..coz fer a teacher da ball is oways on his court..juz could giv him tht bloody pissed off stare..n here he came up wif tis lame line to defend himself..here goes..\"i ve noticed fer da past few days, there ish an increasing number of students sporting tis white color thing on their hand..or rather * PEAr CoveR..\" (*- edited by me) =Xx i thot silently to myself..wad CRaPpy sHYts..!^#!^*%#!^# -.-\'\' n he managed to came up wif tht EXCELLENT reason fer approaching me..-.- and i hope he can realli giv a call to da doc..n realise actually he ish such a FAggot..n perhaps he can tink of becoming a Hollywood director since he ve so many WILD imaginations !!~ y rot in tis sch..? a piece of advice frm me..=Xx hand in ur resignation letter tml !!  '),(26372,4255254,'19,August,2004','2004-08-19 00:00:00','   Thx fer wasting ur tym in reading my bloggie..erm i saw ur journal at multiply..i m nt sure if i ttm coz i thot tht u should be feeling down after so much things had happened..I ve exactly no idea how i can help coz other den to cheer you up, da rest ish within my reach..i juz hope u can move on w/o any worries..n of course cheer up !!! stay happy oways !!~~ n rmb nt to slack anymore..study hard !! smile and rockz on !!  '),(26373,4255254,'18,August,2004','2004-08-18 00:00:00','   Hi guys..like tht new layout of my bloggie..?? lolx~ o.O i thot it looks kinda cool and provides me wid a nature bite..erm well anyway back to my routine life..juz back frm sch after another sucky day..haIxx i was feelin kinda frustrated when i cant do much writing..it makes me feel so helpless and i thot wad\'s da point of goin to sch..? yawnx but i gt no choice or wadever coz if nt i would be rotting at home as well...  '),(26374,4255254,'18,August,2004','2004-08-18 00:00:00','   I came across this wordy \"Pepperistic\" when i was surfing sum webby da other day..n the owner of tht webby happens to be a potential designer..i was kinda impressed wif his werk..n thot da wordy can be used to describe flavour of crabs ya..? n since i m known as pang xie by my peers, i thot tis wordy fits in perfectly..hahax ish kinda lame~ =Xx coz i m crappy !!  '),(26375,4255254,'17,August,2004','2004-08-17 00:00:00','   2nd day on Bloggie~lolx..by right i should be in sch now..rather den goin hm at 230 i m already rotting rite now..=Xx paid a visit to da doc juz now..n gt tht pear cover off my hand..but ish juz fer a few minutes before another brand new fruit cover gt attached to my hand..argh!! nt again..-.-\'\' n 10q gy fer adding da 1st comment of my bloggie !! but tht was nt a rather nice 1..n u mention abt those lame stuff of urs again..-.-\'\' but i guess i gt used to lamey thingys n was nt kinda surprised..hahax..n nowadays u r kinda stress uh? ermm..dun be too stress ya? muz haf more confidence in ur ability..i trust tht u can make it..n make sure u dun disappoint me !! =Xx slack fer a couple of hours more..n no matter wht i nid to complete a portion of my folio by today..i hate it wen i kip on procrastinating things..haIxx..n todae while i was goin to sch, i saw her agn !! =Xx i ve no idea y izzit like tis..lolx coz ish either i m behind her or infront of her..n frm my memo, i rmb onli once when we met n stand beside one another waiting fer tht traffic light..anyway i m juz contended to noe tht she ish happy..lolx..perhaps i m unaware tht she may haf lots of probs but one thingy fer sure ish she ish willing to move on..n tht\'s kinda a positive sign ya? i guess tht\'s abt all..her smile juz brightens up my day !!~  '),(26376,4255254,'17,August,2004','2004-08-17 00:00:00','   Oh my..ish my 1st day on bloggie n tis ish da 3rd message i m posting..i was glancing thrugh my calender juz now n realised ish onli abt 70++ days to O levels..BLANK !! i thot i was running out of tym..arhhx juz wen i wanted to instill sum sense of urgency into me..i met a freaking accident..n guess wht? i cant do much writing fer now..n i stil haf my dnt folio to complete..blame me fer nt startin early..haIxx n i thot i was goin to flop my prelim next mth as well..i juz study wadever i can fer tht exam..since my priority ish O levels..n ya last week after gettin da chinese results, my peers haf start to prepare fer da O levels..n da bez of luck to all of em..rcv it..? i miz those daes when we play soccer everydae after sch..hahax but i guess 3 mths later we can play till we go bonkers..lolx..fer now i tink everyone of em haf their own schedule..n hey i hope all of u kip on da tempo till nov..when da actual thingy realli starts..n juz last month i came across 2 noe tis ger..i gt to thx frenster fer tht..=Xx i summon enugh courage to drop her a msg n frm thar we nv look back..i gt to admit we may ve lots of up n down..n there was once when i m realli sick of everything n juz trit tht we ve nt known one another..but after a few days everything was back 2 normal..n i m juz glad tht it turns out tis wae..arhhx running out of words..i wish i could get tht white thingy off my hand rite now !! coz i m sick of pple telling me it kinda looks lyk tht pear cover at fruit stalls..-.-\"..or piggy hand..? lolx..  '),(26377,4255254,'17,August,2004','2004-08-17 00:00:00','   Poor me !! gt to go to sch tml wif no company..=Xx but i would be leavin sch at 9 tml..haIxx hope my injured finger can recover fast..fingers crossed fer tml visit to da doc..2 more weeks to my dnt proj deadline..plxx Buck Up !!  '),(26378,4255254,'17,August,2004','2004-08-17 00:00:00','   Ish my 1st tym on bloggie !!`  '),(26379,3485704,'28,May,2004','2004-05-28 00:00:00','   Well my first blog I am somebody now.I feel like a million bucks, post my thoughts on the web possibly gain some momentum maybe catapult this into some kind of career. I hardly have any aspirations for gaining fame I can barely stand being around people I know. That is not to say that I am some sort of introvert that never leaves his house or does anything outgoing. Sure I dodge the spotlight at any chance it is put on me, but still I am not afraid. To be afraid is to be a coward and that is just the ultimate insult to a guy and hey I am a guy if you did not pick up on it earlier. I mean hey a selfconscious arrogant prick who thinks his dull life is actually worth talking about. I mean hey this must be a guy, only a guy could be arrogant and selfconcious at the same time. Bullshit what do I know, I know nothing women which is I suppose common and I am not even out of highschool yet. There is no way I can have an opinion about anything simply because I am a child who knows nothing about responsibility, really though I do not have that many opinions so you want find anything on here about the War or any of that shit until I am at least in college drinking coffee and smoking cigarettes letting the world have a glimpse at my infinite genuis. I guess there will be some social commetary whatever that means, I heard the phrase somewhere, makes me sound like I know something. Really the only way I keep the ego growing is by using shit like that and trying to sound smart but arrogant. The whole ego came about when people started calling me smart because I got good grades, hell I did not know that they were good grades because I was always slacking off well not really. I try to downplay the effort I exhaust in getting good grades just to sound like I am some prodigy really it is just to shield myself from whenever someone says I am not as smart as someone else and I can just say well if Iworked as hard as them I would be just as smart. The reason I am not as smart as I could be is because of my mother. Typical blame the mother, but in some aspects it is true she used to help me with homework when I was younger and I can never forget how she would yell at me if kept repeating a mistake. Eventaully I just stoped asking from help from anyone, in class I don\'t no something I don\'t ask the teacher I suffer or pray someone else is has the same problem but lacks my phobia of help. That is not the main reason that I do not raise my hand it also comes back to being selfconscious if you did not see that one coming. I always think people wo ask questions are dumb, this is only in class mind you I would question if asked to sodomize an Iraqi detainee. Sorry no political commentary until college, smoking and coffee my apologizes I humbly beg for your forgiveness.Well the people who ask question no matter how mundane always come out with the A not to say that I angry about it I think they deserve if they do not know something they ask. This must have been an extreme waste of your time I went off on to a rant that splintered into another rant that splinterd into another rant that eventually led to you reading this sentence.        This is your friendly neighborhood sarcastic condescending illiterate peusdo anti-social hero off to save the day again.   '),(26380,3870574,'08,July,2004','2004-07-08 00:00:00','   Oscar winning film maker Michael Moore is not patriotic. While talking to Lila Lipscomb in Fahrenheit 9/11, he talks about how great America is. At the end of the movie, he talks about the selflessness of our troops even though early in the movie, Moore makes fun of our young guys playing songs like \"Let the Bodies Hit the Floor\" in their tanks while they put their life on the line.  To further demonstrate his patriotism, he posts this on his website:  \"The Iraqis who have risen up against the occupation are not \"insurgents\" or \"terrorists\" or \"The Enemy.\" They are the REVOLUTION, the Minutemen, and their numbers will grow -- and they will win.\"  http://www.michaelmoore.com/words/message/index.php?messageDate=2004-04-14  Some support of our country, Mike.  As for United Nations help?  \"There is a lot of talk amongst Bush\'s opponents that we should turn this war over to the United Nations. Why should the other countries of this world, countries who tried to talk us out of this folly, now have to clean up our mess? I oppose the U.N. or anyone else risking the lives of their citizens to extract us from our debacle. I\'m sorry, but the majority of Americans supported this war once it began and, sadly, that majority must now sacrifice their children until enough blood has been let that maybe -- just maybe -- God and the Iraqi people will forgive us in the end. \"  It is a sad day in American politics that when your level of political \"dissent\" reaches as far as predicting and wanting the enemy of our soldiers to win. The worst part? Michael Moore is not off base in the eyes of the Democratic Party and many of their members. Hate our President Mike, but don\'t hate our troops.  '),(26381,3870574,'08,July,2004','2004-07-08 00:00:00','   I did not what this title be deceptive. The truth is that John Kerry has no vision for America. I challenge anyone reading this post to find his vision. While reading  John Kerry on the Issues  on his personal website, I came across something that is very disturbing.  Let me quote John Kerry\'s stances on the issues:  John Kerry on the Issues   Economy:  \"John Kerry has the courage to roll back Bushâ€™s tax cuts for the wealthiest Americans so we can invest in education and healthcare.\"   College Affordability:  \"While George Bush gives tax cuts to the wealthiest Americans...\"  \"John Kerry has the courage to take on Bushâ€™s tax cuts for the wealthy so we can invest in education.\"   Education:  \"John Kerry believes that we need to invest in our schools instead of giving tax cuts to the wealthiest Americans.\"   Health Care:  \"George Bush has chosen to lavish tax cuts upon the wealthiest...\"   Homeland Security:  \"John Kerry has the courage to roll back George Bushâ€™s tax cuts for the wealthiest Americans so we can invest in homeland security.\"   National Service:  \"Whether it is protecting America from the threats of terrorism or addressing the problems we have at home, America\'s new challenges will not be met by the same old answers of big government or big tax cuts for the wealthy.\"  Kerry\'s vision? The only one that I can see is a promise to raise our taxes and take money out of the pockets of the middle class, working Americans.  '),(26382,3870574,'02,August,2004','2004-08-02 00:00:00','   I\'ll give some of you some of the reasons why I feel that John Forbes Kerry is bad for America based on my research on him.On key votes, John Kerry has sided with Senator Ted Kennedy 94% of the time.      Kerry has voted against a balanced budget amendment at least five times. Fiscally responsible? Kerry voted against President Bush\'s tax cuts which Alan Greenspan claimed to have pulled the economy out of the recession. \"The first priority is the economy of our nation. And when you have a downturn in the economy, the last thing you do is raise taxes or cut spending. We shouldn\'t do either We need to maintain a course that hopefully will stimulate the economy...No, we should not raise taxes, but we have to put everything on the table to take a look at why we have this structural problem today... you don\'t want to raise taxes\" (John Kerry on NBC\'s \"Meet The Press,\" September 2, 2001)       Then, in April 2002, Kerry said he wanted a larger tax cut and was \"not in favor of\" a repeal. CNN\'s Tucker Carlson: \"Senator Kerry...[many Democrats] [g]et a lot of political mileage out of criticizing [President Bush\'s tax cut], but nobody has the courage to say repeal it. Are you for repealing it?\" Kerry: \"It is not a question of courage... And it\'s not an issues right now. We passed appropriately a tax cut as a stimulus, some $40 billion. Many of us thought it should have even maybe been a little bit larger this last year...[T]he next tax cut doesn\'t take effect until 2004. If we can grow the economy enough between now and then, if we have sensible policies in place and make good choices, who knows what our choices will be. So it\'s simply not a ripe issues right now. And I\'m not in favor of turning around and repealing it.\" (CNN\'s \"Crossfire,\" April 16, 2002)      Eight months later? NBC\'s Tim Russert: \"Senator...should we freeze or roll back the Bush tax cut?\" Kerry: \"Well, I wouldn\'t take away from people who\'ve already been given their tax cut...What I would not do is give any new Bush tax cuts...\" Russert: \"So the tax cut that\'s scheduled to be implemented in the ocming years....\" Kerry: \"No new tax cut under the Bush plan...It doesn\'t make economic sense.\" Russert: \"Now, this is a change, because let me show you what you said in September 2001 when I asked you the very same question.\" (NBC\'s \"Meet The Press,\" December 1, 2002)      Kerry claims to want to make this a clean campaign, but in a statement he said on June 19th, 2003, he said that President Bush \"misled every one of us.\" (Ron Fournier, \"Kerry Says Bush Mislead Americans On War.\" The Associated Press, 6/19/03)      Four days days earlier?George Stephanopoulos: \"People are really upset that they feel misled by President Bush on this issue of weapons of mass destruction. I know you said you\'re agnostic about whether or not he misled the public on weapons of mass destruction. But do you have a hunch on whether you think they hyped the intelligence?\"John Kerry: \"George, again, I think it would be irresponsible of me at this point to draw conclusions prior to all the evidence being on the table. What I know is we have to get that evidence. We have to have an investigation to know to a certainty whether or not it was hype, whether we were misled...\" (ABC\'s \"This Week,\" (6/15/03)        Kerry has voted for at seven major reductions in Defense and Military spending, neccessary for our national security.         In 1996 - Introduced bill to slash Defense Department Funding by $6.5 billion      In 1995 - Voted to feeze defense spending for seven years, slashing over $34 billion from Defense.         Fiscal 1996 Budget Resolution - Defense Freeze. \"Harkin, D-Iowa, amendment to freeze defense spending for the next seven years. In 1993 - Introduced plan to cut numerous defense programs, including:          Cut the number of Navy submarines and their crews.      Reduce the number of light infantry units in the Army down to one.      Reduce tactical fighter wings in the Air Force.      Terminate the Navy\'s coastal mine-hunting ship program.       Force the retirement of no less than 60,000 members of the Armed Forces in one year         Military hardware he felt we nolong need since the \"cold war\" is past. The money would be better spent on \"social\" programs. These weapons are now the core of our military might:          The F-16 Fighting Falcons      B-1B\'s, B-2As, F-15 and F-16s       M1 Abrams      Patriot Missile      AH-64 Apache Helicoptor      Tomahawk Cruise Missile      Aegis Air-Defense Crusier         While running for Congress in 1972, Kerry promised to cut Defense Spending. \"On what he\'ll do if he\'s elected to Congress,\" Kerry said he would \'bring a different kind og message to the president.\" He said he would, \"Vote against military appropriations.\" (Candidate\'s For Congress Capture Campus in Andover,\" Lawrence [MA] Eagle-Tribune, 4/21/72)        I leave with this.        \"So you can look at all the potential threats of the world, and when you add the expenditures of all our allies to the United States of America, you have to stop and say to yourself, \'What is it that we are really preparing for in a post-cold-war world?\"-Senator John Kerry, Congressional Record, 5/15/96, p. S5061)        That makes me feel warm, how about you?  '),(26383,3870574,'01,August,2004','2004-08-01 00:00:00','   I have talked at lengths about why John Kerry is bad for America. I feel compelled that right now I should lay out a case for why the President is better than John Kerry.      Faith      President Bush is a man of great faith. He was a former alcoholic that seeked God to help reform his life. He pulled himself up by the bootstraps with his faith giving him strength. As a Christian, George W. Bush represents the Judeo-Christian values that the majority of Americans hold, and represents the history of what this country was founded on. All though President Bush is a man of faith, he knows the value of being \"strong in faith and secular in politics\". His faith gives him the strength and moral conviction to do what is right for the country.  With his proposed faith-based initiatives, it gives children to go to community centers where the church is the only place where there is the only safe haven.           Economy:           The economy has been through some tough times.After inheriting an economy that was slumping near the end of the Clinton Adminstration, President Bush was handed the tragedy of the September 11th attacks on our nation, which deeply hurt American businesses and American confidence dramatically. Furthermore, the high-profile corporate scandals of Enron and WorldCom compounded to the problem by weaking confidence in our nations CEO\'s. Americans were wary of buying stocks of companies in fear that they were \"cooking the books\".           The President sprung to action and cut the nations taxes so that the money would not be with the government; it would be the with the people that know what is best for that money...the American people.       GDP grew by more than 5 percent over the last three quarters, the fastest rate of growth in nearly two decades.       Productivity grew at the fastest 3-year rate in more than 50 years.       In terms of creating jobs, it could not be any stronger. The tax cuts have put more money in the pockets of our small businesses so that they can buy more equipment and hire more workers.       Since last August, over 1.4 million new jobs have been created.       The unemployment rate has fallen from 6.3 to 5.6 percent, below the average of the 1970s, 1980s, and 1990s.       This job growth is widespread â€“ employment over the last year was up in 44 of the 50 states, and the unemployment rate was down in 46 of the 50 states.      This job growth is widespread â€“ employment over the last year was up in 44 of the 50 states, and the unemployment rate was down in 46 of the 50 states.      This job growth is widespread â€“ employment over the last year was up in 44 of the 50 states, and the unemployment rate was down in 46 of the 50 states.       Homeland and National Security:      The highest priority of President Bush is to keep Americans safe. He is decisive in his actions. He has the moral clarity to differenciate the difference between good and evil. He has the courage to stand up to dictators, despots, terrorists and tell them that America will respond to attacks on our country and we will take the fight to them. He believes in the philosophy of going after the terrorists before they come here.      He has created many new measures to make it harder for terrorists to attack on the homeland. With the Passage of the USA Patriot Act, inter-agency communication is improved. President Bush has created a number of new agencies that make it faster and easier for our border agents and airport screeners to identify known terrorists.      Furthermore, with the creation of the Homeland Security Department, we now have an agency that is totally focused on preventing and protecting the American homeland from terrorist attacks.       Iraq       President Bush was given a choice. He was to take the word of the CIA to defend America, or take the word of Saddam Hussein. Given that choice, he will defend America every time. With the liberation of Iraq and Afghanistan, 50 million oppressed people now have the taste of freedom.      Even though there is more work to be done in Iraq, we are getting the job done right. President Bush does not believe that running away would be sending the correct message to the terrorists or to the world. America will not waver in the face of terror. We will stand up even when our \"allies\" will not. Eventually, our strength and resolution will bring them to our side.      We should not have to ask the United Nations, France and Germany permission to defend America.       Our servicemen and women represent the best America has. I deeply thank the brave servicemen that risk their lives so that we will be safe and that the Iraqi people will be free.  '),(26384,3870574,'01,August,2004','2004-08-01 00:00:00','   It is obvious that John Kerry is ashamed of being a liberal today. As the the most liberal Senator in the United States Senate, his voting records tells the American people what he does not want them to know. He stands for higher taxes; against the protection of pregnant women. (Laci Peterson Law) He is for sending our troops to Iraq but against giving them the proper equipment to fight the insurgents in Iraq. He is for the Patriot Act, and now he is against it. Why can\'t John Kerry tell the truth to the American people on the issues he stands for? Is he embarrassed of his voting record or does he not want to give the American people the whole story?       Secondly, when John Kerry spoke the Democratic National Convention, he was trying to display a more centrist, more moderate viewpoint. Clearly from the reactions in the crowd, the Democratic Party is far from being moderate. They cheered the most when John Kerry made his digs at the President and his cabinet. He pledged that \"As president, that is my first pledge to you tonight: as president I will restore trust and credibility to the White House\", only to go on to say, \"I want to address these next words directly to President Bush: In the weeks ahead, let\'s be optimists, not just opponents. Let\'s build unity in the American family, not angry division\"      Furthermore, Kerry goes on to say:      \"I will be a commander in chief who will never mislead us into war. I will have a vice president who will not conduct secret meetings with polluters to rewrite our environmental laws. I will have a secretary of defense who will listen to the best advice of the military leaders. And I will appoint an attorney general who will uphold the Constitution of the United States.\"          No angry division, Mr. Kerry? How can you, in the course of one speech, flip-flop on your statements? Especially in a speech of this magnitude? The President mislead us into war just as much as you did, Mr. Kerry, and unlike the Senator from Massachusetts, President Bush has a clear and concise view of what he plans to do to protect America. Bush states that he was tired playing defense and he wanted to go on the offensive on these terrorists. (Stated before the 9/11 attacks as chronicled by the 9/11 Commission Final Report.) Bush believes that to defend America we must bog the terrorists by making them move constantly so they cannot plot to murder Americans. President Bush believes that we should not have to get United Nations to get consent to protect our country. President Bush, as did John Edwards did at one point, believed that if we had to defend America without the help with some of our allies, that we will do so.      President Bush is proud of his faith, values and conservatism. John Kerry cannot decide if he is proud to be a liberal. He was a hawk at the beginning of the war. Prior to the re-invention convention, he had to adjust to a more anti-war stance to become the Democrats choice to run for President, and now, at the time where he speaks to the country, he comes to the center and tries to have it both ways. Mr. Kerry, it will not work. The American people see a straight shooter in President Bush. They see you as someone who says what is politically expedient. And in November, it will show as President Bush will have \"four more years\". For you Mr. Kerry, no \"help is on the way\".  '),(26385,4179297,'13,August,2004','2004-08-13 00:00:00','   I found out last night that I will be moving shifts at work.  This is depressing news,  as I adore the folks I work with now.  We get along so well and know how to work with each other.  Not trying to say too much in the negative about my reassignment, BUT....  Have you ever ran into people who just don\'t like anyone unless they\'re cute, drive a cool car, or have money oozing out their ears?  The ones who are so negative about most anything unless it involves something good for them?  That\'s this crew....well, with the exception of one.  Well, I\'m cute (I think).  But I am a big woman, so that marks points against me where these folks are concerned.  My car is not cool at all.  (10 year old Toyota Tercel that has blue smoke to match the blue paint) And since I support myself and my disabled mother who lives with me, I don\'t exactly have money oozing out of my ears.   I\'m trying to keep a positive attitude about the whole thing, though.  One of my current shift members who has more supervisory experience than I do told me to try to go into it with a clean mental slate and not to hold grudges about problems with those shift members in the past.   I believe that is certainly valuable advice.  A little difficult to do, but still good advice.  A few months back I had mentioned that I would make a shift change if needed to try to help alleviate some of the problems that come from that group.  I didn\'t really expect it to happen so soon!  So I suppose I will be hitting the bookstore today before work to find some books on personality types, management strategies, etc.   At least one thing went well yesterday!  The dentist appointment wasn\'t even half as bad as I expected it to be.  It took longer for the anesthetic to work its magic than it did to complete the procedure itself.  My appointment was at 2pm, and by 2:30 I was driving home!  Today my jaw is a bit sore at the injection site.  I guess that\'s a small price to pay for a tooth that is now more aesthetically pleasing, durable, and complete.  Oh yeah...Seems as if Tampa might get blown away before I get to my vacation.  I\'ve been keeping up a little more than usual with the Weather Channel since Hurricane Charley is headed directly for my hotel room in Apollo Beach.  This is par for the course for my run of luck.  She certainly isn\'t a lady with me!  Off to Barnes and Noble to spend even more money on books I probably won\'t complete!  (Oh shit.  I just realized it\'s Friday, the 13th!) '),(26386,4179297,'12,August,2004','2004-08-12 00:00:00','   Ugh. I have a dentist\'s appointment this afternoon. They\'re going to fix a tooth that I chipped on a piece of fried okra, of all things. When I was there for my cleaning a couple of weeks ago, the good doctor mentioned that he could probably repair my chipped tooth without having to use anesthetic.  Yeah.  Wonder what brand of crack he started smoking?!  I mean, really! I\'m certainly a chicken where the dentist is concerned, but not as bad as some. One of my friends gets his dentist to use nitrous for a routine cleaning. Hopefully things won\'t ever be that bad for me! Unfortunately I\'ve been blessed with the curse of being completely honest. There I was in the dentist\'s chair whining, \"Oh, no! Please use the shots! I\'m such a chicken, but I don\'t care to admit it! Please use some Novacaine!\" To make me feel like less of a moron, the dentist said that he, too, was a big ol\' chicken. Hey, he gets points for at least trying to make me not feel like a dumbass!  Speaking of not getting shots at the dentist, there is an awesome book I finished not too long ago called   urlLink A Million Little Pieces .  This book is by far, one of the best I\'ve read in a long time. It\'s a brutally honest and painful account of James Frey\'s stint in rehab. He\'s 23 and has been an alcoholic for ten years and has been huffing paint, snorting coke and smoking crack for three years. I don\'t want to talk too much about it because you can read reviews on the link. Trust me, it\'s well worth the time. (You\'ll understand the dentist reference not too far into it.)  Okay, I\'ve procrastinated long enough. My work stuff is due this evening, so I\'m going to get cracking on it before I get sleepy. '),(26387,4179297,'09,August,2004','2004-08-09 00:00:00','   After reading a few blogs over the years, and since my  urlLink best friend  recently started one, I decided it was time I get one, too.  Kudos to Kelly for her MANY ideas!  To let you know a bit about me, I am a 27 year-old single white female who lives in the Bible-thumping region of these here United States called  urlLink Chatsworth, Georgia .  It\'s a small town that is rather uneventful, but we are in between Chattanooga, TN (30 miles north) and Atlanta, GA (about 80 miles south).  Whenever there\'s a craving for \"city life,\" we usually head south.  For good shopping, we typically head north. (Even though TN sales tax is highway robbery at almost 10%!)  My mother, my dog (a beautiful Sheltie named Sierra), hamster (uh.....a brown one I call Woodrow), and I all share an apartment.  It\'s certainly cramped and frustrating most of the time, but it\'s home.  I have been looking for a house to buy for the last several months.  It seems like every time I get excited about one, something happens that I can\'t get it.  I either get outbid,  the house is cute outside but falling apart inside, or it\'s 35 miles from my work, etc.  I\'m sure anyone who has bought a house before knows the frustration.  I have never been married and have no kids.  I don\'t want to be married, either.  I have seen too many strong, smart, ballsy, and wonderful women turned into a weak, emotionally unstable, and depressed shell of a human being all in the name of \"making a marriage work.\"  I don\'t see how arguing 24/7 and hating the person you\'re supposed to be with forever is any better than being single!    I work as a supervisor in a local 911 center.  I\'ve been there a little over three years, and have been a supervisor since February 2004.  We dispatch law enforcement, fire departments, ambulances, as well as take all 911 emergency calls as well as the majority of non-emergency calls. We handle about 115,000 calls for service each year, and that workload is divided amongst four shifts.  I work nights and LOVE my job.  It\'s wonderful because I\'m making sure someone\'s day is going to be better because of the help I\'ve sent them.   By definition, I am considered to be a BBW.  For those of you not familiar with the acronym, it means  big beautiful woman.   I am 5\'8\" and proportioned, so I believe I carry it well.  I have  urlLink numerous freckles, fair skin, grey eyes, and short choppy blonde/brown highlighted hair .  Recently I have been trying to get a bit of a tan because I am going to  urlLink Tampa  in about two weeks.  It\'s slow getting super-pale freckled skin to tan.  Apparently I am going to be doomed to turning into a broiled lobster on the beach.....I don\'t think my tan will be ready by vacation time.  For those of you who keep up with astrology (if there are any), I am an Aries.  I am a true Arian except that I am nowhere near being considered an athlete.  Even if my body were in better shape, I still would suck at sports.  I could never do better than 160 on the golf course.  I had the prerequisite bicycle as a child, but as I got older and busier, that hobby just kinda went away.  On the Keirsey Temperament sorter, I score as an  urlLink ENFP . I like  urlLink good books ,  urlLink Academy Award-winning  and  urlLink nominated movies , cooking, good conversation, having nights out with friends, comedy clubs, concerts, plays, and any other typical-but-not-sports-related activities.  I have strong convictions when it comes to members of ethnic groups, alternate sexual lifestyles, and any other thing that would make a person \"not normal.\"  I have very strong opinions on how the country is on a downward spiral since there is a Wal-Mart on every corner, (thereby putting small business out of business for a dollar) congress is trying to legislate morality, and how some folks think it\'s acceptable to try to throw  around \"Bible-based\" beliefs while hating people who are different from them.  Is it just me, or is that quite hypocritical?  Is hating people because they\'re different from you what Jesus would do? No, you say? Then take off that damned bracelet!!! A diatribe on that will follow in a later entry...I can\'t use up everything I like to fuss about in my first entry!  Golly geez, it\'s late....I\'ve got to get some work stuff done.  Thanks for taking the time to read about me! I hope you\'ll be back for more later on! '),(26388,4052167,'27,July,2004','2004-07-27 00:00:00','   Dag nab it. Look what I found               Kindergarten classes are provided for all children of kindergarten age in the Rochester School District. Michigan Statutes stipulate that before a child can enroll in kindergarten, he or she must be five (5) years old on or before December 1.       Each spring the district conducts a Kindergarten Registration to identify and enroll youngsters who will enter kindergarten in the fall. Informational evening meetings are held in the spring for parents. Newspapers will announce the dates.       It greatly assists the districtâ€™s planning efforts if all kindergarten students are identified in early spring. Parents should contact the Student Enrollment Office or plan to attend Kindergarten Registration if they have a child who meets the age requirements.   Pupils are assigned to either morning or afternoon sessions based upon geographical location. Sessions are three hours and five minutes. Daily attendance is stressed because kindergarten is an integral part of the total instructional program.\"       *grmbl* Leslie could have started last year if she were in Michigan. She was ready. We fought them in Kansas to let her to no avail. AND she\'s been in half a day preschool for 2 and a half years. I want to go back to work and only having half a day school for her complicates things substantially.  '),(26389,4052167,'27,July,2004','2004-07-27 00:00:00','   Over off Livernois (What\'s the street, Deb?)   <>Rent $1275 (with $1275 deposit)  3 bed, 1 bath, 1200 sq feet + carpeted (unfinished) basement   2 bedrooms on main floor  1 on second floor  fenced yard  hard wood floors  fireplace  ceiling fans  Kitchen/dining combo  All appliances  gravel road   walk to elementary school  <>   Does it have Washer/Dryer hookups?  '),(26390,4052167,'27,July,2004','2004-07-27 00:00:00','   Call the school system and get a definate start date for school in case there is anyway we can be there in time.   Done. Official first day is a half-day, August 24.        Sort videos so I can take hand-me-downs to Daniel     Find that box with Buzz/Woody stuff     Get the TV/VCR thingy back from Desiree     Pick up that travel activity book (Friday)  '),(26391,4052167,'27,July,2004','2004-07-27 00:00:00','   We needed someplace to put all this info we are collecting that Jason can get to from work, I can get to from home..even Debi and Mitz can post stuff if they so desire. :)    Voila. A Moving Blog....    SO.  Here\'s what we have so far:  We want to move back to Rochester ASAP. Jason has his Michigan nursing lisense (Hooray) and is in the process of contacting hospitals to set up interviews there on Aug 2-4 (Mon - Wed).   We are leaving here around 5 pm on Friday, July 30 and will drive like crazy people with a goal of arriving at Debi\'s late Saturday night.    I understand that VBS is Aug 1-5 so Debi\'s gonna get the kids signed up for that.  '),(26392,4052167,'12,August,2004','2004-08-12 00:00:00','   Job : Well, yes.  As it stands right now, Jason has offers from North Oakland and St. Joseph Mercy (both in Pontiac) .   I also spoke with Julie from Beaumont yestarday and I really think they are going to offer him one too - their process is slower and they keep finding things in his file they need (she called yestarday to get a copy of his army paperwork).   House : Also Janet called about the log cabin - that looks pretty good too, depending on when we can move.  She didn\'t seem unsettled about our October target.   Money : So now it\'s really just a matter of finding the best way to get the most money in the shortest amount of time. Hehe.  Jason is spending his lunch hour today talking to travel nurse recruiters.  We thing that if he could get a quick 3 month assignment that paid big money (and they all do) that we\'ll just pack up the house - put everything in storage - and go to CA or Boston or Seattle or DC or wherever.  I\'ll homeschool for the 3 months and we\'d live in a furnished  2 bdrm appt provided by the travel company.  I know it sounds crazy, but it\'d be like a 3 month vacation.  It\'d be terrific for the kids and lots of fun for us AND Jason can make huge money doing that.  Then by October, we\'d have ample money to pay off everything here and start the way we want to start in Michigan.  It also has the added benefit of getting us away from here as quickly as possible.  So, that\'s the scoop! :)  Debi, I left Collin\'s shorts in your freezer *sigh* Can you mail them to me? I really need them.  He only has a couple of decent pairs of shorts....  '),(26393,4052167,'10,August,2004','2004-08-10 00:00:00','   Hey all!  Any news since you have gotten back?  Miss you already!  Me  '),(26394,4110791,'13,July,2004','2004-07-13 00:00:00','    Well Today was pretty much uneventful. Me and my sister came home from my Dad\'s house and started switching our Mom\'s and our rooms. Lots of hard work but it was ok. But then tonight my boyfriend, Gene told me that he passed his Driver\'s Test :) I am soo happy for him. Yay! Well that was today :p '),(26395,4110791,'12,July,2004','2004-07-12 00:00:00','    Blah Blah Blah!!! I\'m only doing this post because my Sister is a wacko and wants me to :p '),(26396,4110791,'08,August,2004','2004-08-08 00:00:00','    Well lets see. Today was pretty uneventful, hung out with my sister for awhile, and that\'s about it :p Other than going to Baskin Robbins to say hi to Rhiannon, I didn\'t get to get any ice cream :( Oh well we have ice cream bars at home if I really want some :p Hmm..... Yesterday was a really good day :) Gene came over around 12:30 and we hung out and played Nintendo games :p It was lots of fun :) My Mom got to embarrass me :p I think her pain pills made her even more crazy :p After Gene left I talked to my Mom and watched t.v. Then he messaged me and asked if I could go swimming with him and Jesse and my Mom actually said yes :D Yay! It was lots of fun :) Gene and I got to kiss and make out :) and play around in the water. I really enjoyed it :) Then Jesse drove me home and I ate dinner, baked potatoes, chicken and corn on the cob :p It was pretty good :) After dinner I took a really hot and long bath :)  It felt really good to relax :) Then I watched t.v. till around 11 and then went to bed. That was pretty much it :) '),(26397,4110791,'04,August,2004','2004-08-04 00:00:00','    Yesturday, I woke up around 10 a.m. and I layed in bed and read my book and did laundry. Gene showed up around 12:30 and I was still in my pajamas, cause I had been waiting for my clothes to dry :p so I had to hurry and get dressed and go downstairs to answer the door. I was soo happy to see him :) We talked and I showed him my messy room :p and then we went outside and sat in the grass and talked for awhile. We ended up making out in the front yard :p I love kissing him :) Well by the time we were done or whatever he had to go cause he was almost late to be picked up. He told me later that he had to run the whole way :p Overall yesturday was a GREAT day! :) '),(26398,4110791,'02,August,2004','2004-08-02 00:00:00','    For those of you who aren\'t sure exactly who I am, my name is Gene Baxter, and yes, I am Kendall\'s boyfriend :)  I just wanted to say a few things about Kendall.  First and foremost, she is my girlfriend :p  And I love her very much :) And nothing will ever change that!  But I also did want to say one thing, just so that everybody knows, she is afraid of worms :p  Ok, maybe scared isn\'t the correct term.  I will just say this... She can jump six feet in the air, even if she does deny it :p  There is a whole little story to that, but I am not going to start explaining that now.. I have to get some sleep :)  Its 2:22 am, and I want to go to sleep so that I can have wonderful dreams :) '),(26399,4110791,'01,August,2004','2004-08-01 00:00:00','    My birthday was Friday! I turned 17! I had my party at my Dad\'s house. I invited Gene ( :) ), David, James and Rhiannon. They all came and we had a blast. Penny, Gerry, Morgonne, Heather, and Nick came too. Its was one of the best birthdays I\'ve had! We swam in the pool. Even though the water was freezing cold we all had lots of fun. Gene shivered a lot, so for once it wasn\'t me shivering :p So I got to make fun of him for that :p I liked hanging out with my friends and having lots of fun. I wish we could do it more often :s Ohh well. School is starting September 1st. I can\'t wait! I\'m excited about my classes and that its my last year in high school, but I think I\'m most excited about being able to see certain people everyday :) Well thats all I wanted to say :p '),(26400,4110791,'08,August,2004','2004-08-08 00:00:00','   Lets see, Today I woke up at about 8:30... Im suprised that I woke up that early :)  I got up, and then I had a bowl of cereal.  Then Shawn called, and he said that he wanted me to mow the lawn and clean the kitchen, but I ended up mowing the lawn, and Bill cleaned the kitchen :p  When I got done mowing the lawn, and Bill got done cleaning the kitchen, we went to an abandoned house about 3/4 of a mile down our back driveway, and shot .22\'s around, that was pretty fun :)  Shawn\'s stupid black lab Ash is going to get shot... She walks right in front of us while we are shooting :S  But, we managed not to shoot her, which is a good thing... Shawn would kill us if we shot his dog :S  Yesterday was alot more fun and interesting than today :)  I went over to Kendall\'s house at about 12:30, and we played Nintendo games :)  Her mom had a chance to make fun of her, and embarass her, and she did :) Kendall turned red :p I had lots of fun :)  Then at about 4:30 Jesse and I left to go back to his house, and we talked about going swimming, and so when we got back I messaged Kendall to see if she could go swimming with us, I\'m very happy that she got to go with us :)  We all had lots of fun :)  Kendall and I got to kiss and make out :) That was lots of fun too :p  After that, Jesse drove Kendall home, and then we went back to his house for about a half an hour, and left for my house :)  For some reason nobody beleive\'s me when I say that I live a long ways out in the country :p  After I got home, I ate dinner.  Steak and potato salad, it was really good :)  After that we all watched Bull Riding, because there wasnt anything else good on, and because Shawn likes to watch Bull Riding :p  After that, it was about 11:00 and I went to bed :)  Even though I didnt fall asleep until at least 12:30... that was the last time I looked at the clock :S  But I eventually went to sleep, and had very sweet dreams :) '),(26401,4110791,'04,August,2004','2004-08-04 00:00:00','   Today I got up at 4:40 AM, because Bill and Zach (Bill\'s cousin) wanted to go fishing.  We had to go in with Tina when she went to work, otherwise we would have been stuck at home :p  So we loaded everything up, and went to town.  Tina had a flat tire, so I aired it up at the shell station.  While I was doing that Bill went inside to get us some burrito\'s breakfast, sandwiches, mountain dew, and pepsi.  Then we were on our way, and we went to a fishing spot right behind Bill\'s Aunt Kelly\'s house.  We were going catfishing, but we didn\'t have any luck.  We caught 15, but they were too small to keep :S  During the time that we were there, I ended up going to the shell station 3 times to get Bill and Zach something to eat or drink :p  I was watching the watch and waiting very patiently for 12 o\'clock to come around, so that I could walk over to Kendall\'s house :)  I ended up being delayed for about 20 minutes, because Zach got his fishing line into a big knot, and so we just put some new line on it because the   line was a few years old :p  After that I left for Kendall\'s house :)  I got there at 12:40, and rang the doorbell, but nobody answered :p  Kendall was still in bed.. I think she was sleeping, but she said that she was just reading a book :)  I got to see Kendall\'s room, which is cleaner than my room at the current point in time :p  Even if she said that it was a big mess :)  Oh and I saw some pictures of Kendall when she was younger :)  Then Kendall and I went outside and talked for a little bit.  Oh, and by the way... Grass has a heart and soul :p  We ended up making out,:)  I don\'t know how long we were outside for, but by the time we checked the time, it was 1:56. I was supposed to meet Bill and Zach at 1:30 so that we could go to his Aunt\'s house, but I was a little late.  They asked me why i was late, and I just told them that I was preoccupied :)  When we all got home, we went down to an old abandoned house that is just about half a mile down our back driveway.  We brought wheel barrel\'s down there a  loaded up a whole bunch of wood, because we are going to have a fire, and stuff :p (It was Bill\'s idea) And then we had our fire, and all I could think about was wishing that Kendall were there to enjoy it with me :S '),(26402,4110791,'03,August,2004','2004-08-03 00:00:00','   Hello to anyone who is reading this :) My name is Kendall Hopkins, I am 17 and I am the girlfriend of Gene Baxter, whom I love very very much! :) And could never stand to live without. Gene makes me happier than I have ever been and I don\'t know what I would do without him :) Well lets see, I don\'t really know what else to say :p But that\'s and introduction to me :) I love you Gene!!!  OHH and the only reason this is in pink is to make Gene laugh :p '),(26403,4110791,'02,August,2004','2004-08-02 00:00:00','   For those of you who don\'t know me yet, my name is Gene Baxter.  I would like to start out by saying that I am 17, soon to be 18 :)  While my girlfriend Kendall just turned 17 :)  And I will start out by saying that I love her VERY much, and I don\'t know what I would do without her :S  I always want to be with her, nomatter what :)  I always want to be able to be with her, to hold her, to kiss her, even when I know I can\'t :S '),(26404,3743431,'24,June,2004','2004-06-24 00:00:00','   I have been a pretty unbiased person in regards to Michael Moore\'s latest film. I even gave it a half-assed positive review, although I was honest in reporting some inaccuracies that I found.  After visiting http://www.moorewatch.com after hearing about the website on the news, I decided to take a wander over there.   It seems the Michael Moore has taken an article about his film, posted it on his site, but removed a sentence or two from the article.  Below is what was written by JimK at Moorewatch.com:   urlLink Mikey Is Afraid of Us - MOOREWATCH - Watching Michael Moore\'s every move : \"Case in point?  Just like we do here, Mike has been reposting articles on his website that pertain to F911.  On this page he reposts the entire text of a column from Canada\'s CTV News.  Here\'s the way Mike quotes it.   CTV.CA: Missing Fahrenheit 9/11\'s Forest for the Trees\'  Bill Doskoch, CTV.ca News Staff  Updated: Wed. Jun. 30 2004 5:51 PM ET  Both the critics and the audiences have come out for Fahrenheit 9/11. Pulling in almost $22 million US in its first weekend, the film has set a record for a documentary.  Not surprisingly, the mainstream media has cast a skeptical eye on the left-leaning film. And U.S. conservatives are absolutely apoplectic about it.  Both the mainstream news media and conservative advocates have harped on inaccuracies in the film.  Now, here\'s the way it appeared on the CTV News website.  Missing Fahrenheit 9/11\'s forest for the trees  Bill Doskoch, CTV.ca News Staff  COMMENTARY&#65533;Both the critics and the audiences have come out for Fahrenheit 9/11. Pulling in almost $22 million US in its first weekend, the film has set a record for a documentary.  Not surprisingly, the mainstream media has cast a skeptical eye on the left-leaning film. And U.S. conservatives are absolutely apoplectic about it. Check out Moorelies.com or Moorewatch.com to see what I mean.  Both the mainstream news media and conservative advocates have harped on inaccuracies in the film. [Emphasis added]So, even in an article sympathetic to his cause and his message, he feels the need to delete the names of two websites.  As a writer, I am appalled. As a person, I am angry that someone would go to these lengths, as well as, deceive, in my opinion, people who believe in him and his ethics. '),(26405,3743431,'24,June,2004','2004-06-24 00:00:00','   The F word has become such a dirty word in our society. It is filled with stereotypes, derogatory remarks, and, in some cases, pure ignorance and hate. Now, I am not talking about THAT F word, I am talking about the word Feminist.  Does anyone actually know what a feminist is? Ask someone and their answers may include the term woman\'s libber, or a comment about burning bras, or even a comment questioning one\'s sexuality.  They may believe that being a feminist means you are anti-marriage or anti-name change after marriage. They may believe that being a feminist means that you are men haters or belong to a different era. To some, it means intimidation. To others, it means weakness.  I am a married woman, who took the last name of her husband, has three children, cooks dinner for her family, cooks for her family, and takes care of her family. I stay at home with my children, and for a time of four years, was always barefoot and pregnant. And I am a feminist.  Feminism is believing in social, political, and economical equality for all sexes. Many of you reading this are probably feminist without knowing it. Some of you may hold bad feelings towards feminist, but have wives that hold this belief.   Feminism is just another example of something positive that has been spun by the media to gain the negative stereotypes that it has.  So, the next time someone tells you that feminist hate men, ask them to name one. The next time someone tells you that feminist are the enemy of all women, ask them how, when all feminist really want is to be treated as an equal. You will be surprised at how many individuals will not be able to answer you back.  Be proud of being a woman. Be proud of being a feminist. Stand up for yourself and take back the dirty associations with the word, and show how it is a positive force striving for positive things. And raise your daughters to believe the same things. '),(26406,3743431,'24,June,2004','2004-06-24 00:00:00','   Did anyone catch the news blurps on Saddam in court today? What am I thinking? This is America! It was on every channel, every five minutes. Some emphasized his statement that President Bush was the real criminal, some minimized it, not wanting to give him the opportunity to allow anymore anti-bush \'propaganda\' or because they are republican-based news sources and didn\'t want anyone to get wind of what he said, because some people might actually agree with him.  I have to admit to you all, if there even is an all, that the legal transfer of Saddam to the brand, spanking new Iraqi government makes me extremely apprehensive. I mean the Iraqi government had only been a \'newly sovereign\' (as stated by one news source) country for a couple of days. And today with Iraqi guards escorting him into court? Can they be trusted? Is there some inside information on his where-a-bouts that extremist and loyalist could possibly hijack him?  Now, don\'t go accusing me of being pessimistic or unpatriotic. That couldn\'t be further from the truth. But, those questions, those quandries, and those concerns undoubtibly have not only entered my mind.  Personally, and please don\'t tell me how inhumane I am, I think they should have killed the S.O.B when they had the chance. Why? To insure HE would not be a problem. He is a sneaking man and as dumb as a fox. That is the truth. Did he give all those he ordered to be tortured a chance? No. Did he give the thousands of men, women, and children that he gassed to death a chance? No. Does he care about the human condition? Only when it applies to him. So why give him the honor of living another day? After all the horrible things he has done and the monsters he has raised, why give him another chance?  And now we have one of his defense lawyers here in the good, ole U.S. of A. telling people that he was tortured and injured, telling people that the affects of prison abuse were also felt by Saddam Hussein. I don\'t know if that was true or not, but I will admit to two things. One - if it did happen - good. Two - talk about a good public relations attempt to capitalize on an unfortunate and embarrassing situation.  I hope that I am proven wrong in my fears and that he does go to trial and is executed by the Iraqis - the people that he harmed, tortured, and ran a government on fear with for many years.   I also hope that the man from the United States that has joined his defense team realizes what he is doing, being a member of society in America defending a monster like that. '),(26407,3743431,'24,June,2004','2004-06-24 00:00:00','   urlLink WorldNetDaily: Bush to screen population for mental illness : \"President Bush plans to unveil next month a sweeping mental health initiative that recommends screening for every citizen and promotes the use of expensive antidepressants and antipsychotic drugs favored by supporters of the administration. \"  Does anyone else see anything wrong with requiring every citizen, from pre-school age and up, to have mental health screenings?  This is absolutely ridiculous. Those who have been diagnosed or who have undergone the process knows that it is a sometimes complex and confusing process. I had five misdiagnosis before they realized that I had Social Anxiety Disorder. FIVE. 1 -2 -3 -4- 5. That experience is not uncommon. Not to mention the medication process. Prescriping a certain dosage, monitoring of this dosage is appropriate, re-evaluating the dosage, etc. And who is going to pay for this? WE ARE, the Taxpayers.  So, President Bush can require every citizen to have mental health screenings, but cannot take the money that would go into this, including the money through corporate lobbying and sponsorship, and insure that people have something that they need, like preventive health care?  I am disgusted by this. Why? Because more and more of our rights, our freedoms, that are shouted high off the mountain tops are being invaded. Government isn\'t getting smaller, it is getting bigger. It is creeping it\'s ugly head into all aspects of our life.  Perhaps Orwell had mistakenly named his book 1984 when he did intend it to indeed be 2004. One thing for sure, Orwell would sure be proud that although 1984 was a written work of imagination and fiction, it is beginning to prove to be a foreshadowing into what life is like in America at this moment.  And life, right now, is stranger than fiction. '),(26408,3743431,'24,June,2004','2004-06-24 00:00:00','   At the grocery store, I found Carb-Friendly Hamburger Helper. Yes, you read that correctly, Carb-Friendly Hamburger Helper. As I was watching television, I saw an ad for a carb-friendly menu at T.G.I.Fridays. Whether it is the Atkins diet or the South Beach diet, I have two words for you when it comes to low-carb diets - rip and off.  It is amazing how many people are getting rich from the low-carb craze from these \'experts\' who create these diets to the companies that manufacture these foods to the book publishers putting this reading material on the shelves to the restaurant chains, even McDonalds, cashing in on the now low-carb, food-watching society we have become. I swear, if I see another low-carb food related commerical on television, whether it is for Applebees or Bacardi, I am going to create a new term for going postal, not to disrespect the disgruntal postal workers around the world, I am going to go carb on these television station\'s asses.  These diets are fads. Plain and simple. They create a false sense of accomplishment because weight is lost rapidly - AT FIRST. Then it comes off very, very slowly. Until you decide to get off the diet, then you gain it all back, plus some. I have witnessed it for myself, not only buying into the Atkins diet three years ago, but watching my husband fluctuate in his weight so much. The claims made by the Atkins cat burglars that ketosis helps burn fat has been proven untrue by medical researchers. In fact, prolonged ketosis can cause bones to become brittle. You don\'t hear the experts at Atkins telling you that though. Of course they aren\'t going to tell you that. Why? They would all be out of jobs.  And this whole South Beach Diet about good carbs and bad carbs. What exactly are good carbs and bad carbs. According to Mayo Clinic, one of the top rated hospitals in the world, there is not such thing. A bad carb is a sugar based product high in calories. You mean eating a candy bar isn\'t all that healthy? You are kidding me! A good carb is a carb in products that are lower in calories. That\'s it! Yet people SWEAR by this diet. It isn\'t a diet, it is common sense that is taught to use with the food pyramid. But put a different spin on it and make millions of dollars.   The Mayo Clinic also states that a red flag should be lifted when diets: Promises of a quick fix  Dire warnings from a single product or regimen  Claims that sound too good to be true  Simplistic conclusions drawn from a complex study  Recommendations based on a single study  Dramatic statements that are refuted by reputable scientific organizations  Lists of \"good\" and \"bad\" foods  Advice given to help sell a product  Opinions based on studies published without peer review  Recommendations from studies that ignore differences among individuals or groups  http://www.mayoclinic.com/invoke.cfm?objectid=5F0EFE07-2043-411C-A49B407640B28BFE  Both the South Beach AND Atkins do all of these.  But, what do I know. I am a bad mom who allows her children to eat according to the food pyramid and require limited computer, television, and video game time and require outside play. *sigh*  The low carb industry has played brillantly on two things: a. a society of women that feel the demand to meet some kind of definition of what is accepted as beauty and b. the fear of obesity and the increasing obesity rate in this country.  Sure, you are going to gain weight if you eat and McDonalds and only eat at McDonalds as was done in the documentary Supersize This, an over the top, but hilarious documentary about a man who gained twenty plus pounds by eating at McDonalds three times a day. If you do not exercise, do not allow your body the fruits and vegetables it needs, and do not practice portion control, gaining weight is going to be easy to do.  There are two things that I have notices that have changed since the time I was a child to the ways of our society now: 1. Less people sit down and eat a home-cooked meal together as a family 2. Less people are active because they are too busy.  And, of course, let\'s not forget our addictions to the internet.  Do yourself a favor, go eat a nice plate of spaghetti with bread sticks and get over your low-carb self. '),(26409,3743431,'24,June,2004','2004-06-24 00:00:00','   Critics. Critics are the biggest group of snobs that I know. Why? Because they try to act like their opinions are unbiased and they simply are not. I can cite two specific examples: My Life by Bill Clinton and Fahrenheit 9/11. I want to state that I have my own political views and what I speak has not-so-much to do with party ties. I do not believe in the two party political system, which I am sure I will rant on about at another time. It has to deal with the fact that individuals use their own political beliefs to dictate whether or not something is \'good\' or not.   Over and over on the news, on the web, and in newspapers I have had to hear how horrible Bill Clinton\'s book, My Life, is. Of course I had to go pick up my own copy and discover for myself. The book is not horrible. The book is not boring. The book is not a self-promoting tool thought up by Mr. Clinton himself.   The book goes through his life, which is perhaps why it is called MY LIFE, and documents important moments of his life. Some critics have said that they were disappointed by how little about Monica Lewinsky was in that book. I hate to disappoint people, especially those that are party whores and use their party ideology to justify every single belief that they have, or as I like to call them, Ba Ba Sheep, but Monica Lewinsky was not all of Clinton\'s life. He had a life before and a life after Monica. And there are over 15 pages devoted to Monica. Don\'t believe me? Go to your closest bookstore, look in the index and count them for yourself.   Most of the critics went into this book with preconceived notions on what it should be about, what tales should be written, and how the story should have unfolded. Sorry jerkies, this was HIS book, not yours.   Other idiots, like a few talk radio show hosts and cable news show personalities criticized the book on the first day that it came out, after admittedly only receiving a copy of the book that day. Hate to break it to you, but they did not read all 900 + pages of that book. They decided to thumb through it and base their biased, republican bought butts on a few chapters. Again, more jerkies.   Now let\'s move on to the Michael Moore film. I went and saw it because I was asked to write an unbiased review on the movie for publication. The movie was ok. Got people thinking. Wasn\'t lacking Michael Moore\'s dry humor during the commentary or his in-your-face-style. Wasn\'t his best work. I leave his best work to Roger and Me. But it was ok.    Now the backlash begins for Michael Moore, having to back up every statement that has been made, which if you are going to present a movie with a topic as serious as this, you better be prepared to. Do people, and by people I mean individuals paid to write about this movie, blindly miss the fact that this movie is going to be based upon his own opinions and will most definitely be biased towards that? Michael Moore has never apologized for himself or what he believes in. Why should this movie be any different.   And then there are idiots like Christopher Hitchens who debunk the movie. If anyone has read his column, it is more than obvious to anyone with two brain cells in their head working in harmony that this guy clearly does not like Michael Moore. And, while he is trying to debunk and prove with \'fact\' the inaccuracies of Michael Moore in Fahrenheit 9/11 (which the 9/11 Commission, if being facutal in their reports, have already proven) he gets a few of these \'facts\' he states wrong. Like that Iraq was the only country in that region celebrating what happened on that horrible day in September, almost three years ago. Iran, in fact, celebrated as well. Among others. His ranting and raving about Michael Moore is the same ranting and raving that republicans have done on websites like freerepublic.com or in various message boards and political chats. I am sure that I could find a transcript of Sean Hannity or Rush Limbaugh saying the same things that Christopher Hitchens did in his article.   What happened to unbiased reporting and truth in journalism? Now everyone has a vendetta or parties in their pockets, making them nothing more than literary whores. Cheap whores at that! And those literary whores do a great disservice to the writing community. '),(26410,3743431,'24,June,2004','2004-06-24 00:00:00','   Groupthink:  The act or practice of reasoning or decision-making by a group, especially when characterized by uncritical acceptance or conformity to prevailing points of view.  Yesterday, the U.S Senate Committee on Intelligence reported that the war on Iraq was fought on the basis of absolutely false evidence. The report, over 500 pages ( urlLink 